# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author latif@netapp.com, thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::SolarisMP::MPXIO;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::SolarisMP;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable MPXIO daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, $new_stmsboot_format, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SolarisMP::MPXIO::enable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # handle stmsboot. Setup, but do not reboot, will be hanlded later
    my $filename = "/tmp/stmsboot_input.txt";
    my @stms_cmds = ( "y", "n" );
    &ClientUtil::create_remote_file($conn, $filename, @stms_cmds);

    # Solaris 10 U5 changed how stmsboot is called, but we cannot look
    # for just the minor revision since the kernel may be upgraded, but the full
    # system has not been (therefore minor revision could be < 5).
    # Hack alert, run stmsboot, and look to see if "-D fp" is in the
    # output and that case, also make decision to use new format.
    $cmd = "/usr/sbin/stmsboot";
    ($status, @output) = &ClientUtil::execute($conn , $cmd);
    if (grep(/\[\-D\s+fp/, @output) > 0) {
        $new_stmsboot_format = 1;
    } else {
        $new_stmsboot_format = 0;
    }

    if ($self->{parent}->{parent}->{parent}->{revision} =~ /5\.10/i and
        ($self->{parent}->{parent}->{parent}->{minor_revision} >= 5 or
        $new_stmsboot_format == 1)) {
        $cmd = "/usr/sbin/stmsboot -D fp -e < $filename";
    } else {
        $cmd = "/usr/sbin/stmsboot -e < $filename";
    }
    ($status, @output) = &ClientUtil::execute($conn , $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SolarisMP::MPXIO::enable() Error while enabling MPXIO");
        return $ClientUtil::FALSE;
    }

    &ClientUtil::remove_remote_file($conn, $filename);

    # Verify that MPXIO daemon is running
    $self->_is_mpxio_running(-conn => $conn);
    if ($self->error) {
        $self->_set_err("Error: SolarisMP::MPXIO::enable() _is_mpxio_running reported error");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       disable MPXIO daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SolarisMP::MPXIO::disable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($self->{arch} =~ /sparc/i){
        # handle stmsboot, don't reboot now, will be handled later
        my $filename = "/tmp/stmsboot_input.txt";
        my @stms_cmds = ( "y", "n" );
        &ClientUtil::create_remote_file($conn, $filename, @stms_cmds);

        $cmd = "stmsboot -d < $filename";
        ($status, @output) = &ClientUtil::execute($conn , $cmd);

        if ($status > 0) {
            $self->_set_err("Error: SolarisMP::MPXIO::disable() Error while disabling mpxio");
            return $ClientUtil::FALSE;
        }

        &ClientUtil::remove_remote_file($conn, $filename);
    } else {
        # Return true as no-op for x86
        return $ClientUtil::TRUE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       verify proper enviroment for MPXIO daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SolarisMP::MPXIO::verify() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### Verify that package for MPXIO is installed
    if (($self->{parent}->{parent}->{parent}->{revision}) !~ /5\.9|5\.10/) {
        $cmd = "pkg info system/storage/multipath-utilities";
        ($status, @output) = &ClientUtil::execute($conn , $cmd);
        if ($status != 0) {
            $self->_set_err("Error: SolarisMP::MPXIO::verify() Failed to verify packaged (system/storage/multipath-utilities) is installed");
            return $ClientUtil::FALSE;
        } else {
            $self->{parent}->{parent}->{parent}->sanlog(-msg => "SolarisMP::MPXIO::verify() Verified that package (system/storage/multipath-utilities) is installed",
                      -msg_level => 9);
        }
    } else {
        $cmd = "pkginfo -l SUNWmpathadm";
        ($status, @output) = &ClientUtil::execute($conn , $cmd);
        if ($status != 0) {
            $self->_set_err("Error: SolarisMP::MPXIO::verify() Failed to verify packaged (SUNWmpathadm) is installed");
            return $ClientUtil::FALSE;
        } else {
            $self->{parent}->{parent}->{parent}->sanlog(-msg => "SolarisMP::MPXIO::verify() Verified that package (SUNWmpathadm) is installed",
                      -msg_level => 9);
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       setup proper enviroment for MPXIO daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SolarisMP::MPXIO::setup() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # no-op for Solaris

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: config()
# Objective:       verify proper enviroment for MPXIO daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output, $protocol);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SolarisMP::MPXIO::config() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: SolarisMP::MPXIO::config() Invalid value for argument '-protocol' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Get the HU toolkit version
    my $hu_version = $self->{parent}->{parent}->{parent}->get_hu_toolkit_version();
    unless (defined($hu_version) and $hu_version ne "") {
        $self->_set_err("Unable to determine the Host Utility version");
        return $ClientUtil::FALSE;
    }

    # Get alua support
    my $alua_support = $self->{parent}->{parent}->{parent}->get_alua_support();
    unless (defined($alua_support) and $alua_support ne "") {
        $self->_set_err("Unable to determine the Alua Suppot state");
        return $ClientUtil::FALSE;
    }

    # Get the HU bin path
    my $path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();
    unless (defined($path) and $path ne "") {
        $self->_set_err("Unable to determine the Host Utility bin path");
        return $ClientUtil::FALSE;
    }

    # Configure ISCSI.
    if ($protocol =~ /iscsi/) {
        if ($hu_version < 6.0) {
            if ($alua_support == 1) {
                # Cmode config. Make sure mpxio symmetric option is disabled
                $cmd = $path . "/mpxio_set -df";
                ($status, @output) = &ClientUtil::execute($conn, $cmd);
                if ($status != 0 and $status != 1) {
                    $self->_set_err("Failed to execute command ($cmd:" . @output . ")");
                    return $ClientUtil::FALSE;
                }
            } else {
                # 7mode config.
                $cmd = $path . "/mpxio_set -ef";
                ($status, @output) = &ClientUtil::execute($conn, $cmd);
                if ($status != 0 and $status != 1) {
                    $self->_set_err("Failed to execute command ($cmd:" . @output . ")");
                    return $ClientUtil::FALSE;
                }
            }
        }

        # Make sure mpxio is enabled.  It might have been disabled by running a veritas config.
        my $sanclient_obj = $self->{parent}->{parent}->{parent};
        $sanclient_obj->_modify_iscsi_mpxio(-state => "enable");
        if ($sanclient_obj->error()) {
            $self->_set_err("Error enabling MPxIO", $sanclient_obj->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    # Configure FCP
    if ($protocol =~ /fcp/ and $hu_version < 6.0) {
        # We need to turn symmetric access back off for FCP configs
        $cmd = $path . "/mpxio_set -df";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0 and $status != 1) {
            $self->_set_err("Failed to execute command ($cmd:" . @output . ")");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
#
#                  NOTE: This uses the command mpathadm to get information
#                  about LUNS/PATHS. This command is available with Solaris
#                  10U3 and above (and will fail with earler update levels)
#
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($cmd, @mpath_output, @output, $status, $conn);
    my ($line);
    my ($validate_mode);

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Error: SolarisMP::MPXIO::get_path_info() Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my $per_lun_hash = {};

    # get connection from object
    $conn = $self->{conn};

    # call mpathadm list LU to get list of MP devices on system
    # Since this is MPXIO that will be the case. Then look though
    # each LU for NetApp|IBM VID, otherwise ignore.
    $cmd = "/usr/sbin/mpathadm list LU";
    ($status, @mpath_output) = &ClientUtil::execute($conn, $cmd, 180);
    if ($status != 0) {
        $self->_set_err("Error: SolarisMP::MPXIO::get_path_info() Failed to run mpathadm list LU");
        return {};
    }

    # filter to just /dev/rdsk representation, example output:
    #       /dev/rdsk/c5t60A98000486E2F4B6B5A517653563272d0s2
    #           Total Path Count: 4
    #           Operational Path Count: 4
    @mpath_output = grep(/^\s+\/dev\/rdsk/, @mpath_output);

    foreach $line (@mpath_output) {
        # remove whitespace in line
        my $device = $line;
        $device =~ s/\s+//g;

        $cmd = "/usr/sbin/mpathadm show LU $device";
        ($status, @output) = &ClientUtil::execute_retry($conn, $cmd,
                                                        180, 5, "-warn");
        if ($status != 0) {
            $self->_set_err("Error: SolarisMP::MPXIO::get_path_info() Failed to run mpathadm list on $device");
            return {};
        }

        # If this is not a NetApp or IBM (N-Series), ignore
        if (grep(/Vendor:\s+(NETAPP|IBM)/i, @output) == 0) {
            next;
        }

        # store number of paths
        my $num_paths = scalar(grep(/Path State:/i, @output));
        $per_lun_hash->{$device}->{num_paths} = $num_paths;

        # store number of paths that are enabled (alive)
        my $num_alive = scalar(grep(/Path State:\s+OK/i, @output));
        $per_lun_hash->{$device}->{num_paths_alive} = $num_alive;

        # store number of paths that are disabled (dead)
        my $num_dead = scalar(grep(/Path State:\s+unavailable/i, @output));
        $per_lun_hash->{$device}->{num_paths_dead} = $num_dead;

        # store number of paths that are primary
        $per_lun_hash->{$device}->{num_paths_primary} = "NA";

        # store number of paths that are secondary
        $per_lun_hash->{$device}->{num_paths_secondary} = "NA";

        # store protocol (look at portname to determine)
        my $protocol = (grep(/Initiator Port Name:/i, @output))[0];
        if ($protocol =~ /iqn/i) {
            $protocol = "iSCSI";
        } else {
            $protocol = "FCP";
        }
        $per_lun_hash->{$device}->{protocol} = $protocol;
    }

    return $per_lun_hash;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $obj_template);

    $self={};

    bless $self,$class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents and return
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        # store parent
        $self->{parent} = $parent;
        return $self;
    }

    $self->{parent} = $parent;


    ($self->{version}, $self->{revision}) =
             $self->_mpxio_version(-conn =>
                                   $self->{parent}->{parent}->{parent}->{conn});

    # Return MPXIO object
    return $self;
}

################################################################################
# Subroutine Name: _is_mpxio_running()
# Objective:       returns ClientUtil::TRUE or ClientUtil::FALSE depending on
#                  if mpxio is running or not.
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_mpxio_running {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SolarisMP::MPXIO::_is_mpxio_runing() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check status of mpxio service
    my $cmd = "modinfo |grep -i md_sp";
    my ($status, @output) = &ClientUtil::execute($conn , $cmd);

    if ($status == 0) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: _mpxio_version
# Objective:       To version and revision of Sun Volume Manager package installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _mpxio_version {
    my ($self, %args) = @_;

    my ($conn, $cmd, @output, $status, $version, $rev);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SolarisMP::MPXIO::_mpxio_version() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->_clear_err();

    # For solaris 11+ use new pkg info commands and parse out Version and Branch. Otherwise use
    # older pkginfo command.
    if (($self->{parent}->{parent}->{parent}->{revision}) !~ /5\.9|5\.10/) {
        $cmd = "pkg info system/storage/multipath-utilities|egrep Version\\|Branch";
        ($status, @output) = &ClientUtil::execute($conn , $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SolarisMP::MPXIO::_mpxio_version() Failed to execute command $cmd");
            return ("", "");
        }

        foreach my $line (@output) {
            if ($line =~ /^\s*Version:\s+(.+)/) {
                $version = $1;
            } elsif ($line =~ /^\s*Branch:\s+(.+)/) {
                $rev = $1;
            }
        }
    } else {
        $cmd = "pkginfo -l SUNWmpathadm|grep VERSION";
        ($status, @output) = &ClientUtil::execute($conn , $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SolarisMP::MPXIO::_mpxio_version() Failed to execute command $cmd");
            return ("", "");
        }

        $output[0] =~ /^\s*VERSION:\s+(.+)\,REV\=(.+)/;
        $version = $1;
        $rev = $2;
    }

    return $version, $rev;
}

1;
