# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::WinMP::MPIO;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::WinMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable MPIOMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    # NO-OP for Windows
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       disable MPIOMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    # NO-OP for Windows
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       verify proper enviroment for MPIOMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    # NO-OP for Windows
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # no-op for Windows

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       setup proper enviroment for MPIOMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    # NO-OP for Windows
    return $self;
}

################################################################################
# Subroutine Name: config()
# Objective:       verify proper enviroment for MPIOMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    # NO-OP Windows
    return $self;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($cmd, @format_output, @output, $status, $conn);
    my ($line, $win_client, $details);
    my ($validate_mode, $takeover_state, $num_ports_disabled);

    # parse arguments
    $details = 0;
    foreach (keys %args) {
        if (/-?details/) {
            if (not length($details = $args{$_})) {
                $self->_set_err("Error: WinMP::MPIO::get_path_info() Invalid value for argument \'-details\'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Error: WinMP::MPIO::get_path_info() Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?takeover_state/) {
            if (not length($takeover_state = $args{$_})) {
                $self->_set_err("Error: WinMP::MPIO::get_path_info() Invalid value for argument '-takeover_state'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?num_ports_disabled/) {
            if (not length($num_ports_disabled = $args{$_})) {
                $self->_set_err("Error: WinMP::MPIO::get_path_info() Invalid value for argument '-num_ports_disabled'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check to see if details was defined and set to TRUE
    if (defined($details) and $details == $ClientUtil::TRUE) {
        $details = 1;
    } else {
        $details = 0;
    }

    my $per_lun_hash = {};
    my ($have_ontap_dsm, @non_ontap_dsm_disks);

    # get connection from object
    $conn = $self->{conn};

    # get the winClient object
    $win_client = $self->{parent}->{parent}->{parent}->{WinClient};

    # Issue a rescan (devcon::rescan disks) to "wake up" disks with no I/O
    # and allow MPC to complete on takeover w/o exception. Only perform
    # when in takeover state or if any port is disabled
    if ($takeover_state == 1 or $num_ports_disabled > 0) {
        my $rescan_error = 0;
        foreach my $retry (1 .. 3) {
            $cmd = "$win_client->{cmd_disk_rescan}";
            ($status, @output) = &ClientUtil::execute($conn, $cmd, 600);
            if ($status != 0 or (grep(/An error occurred/i, @output) > 0)) {
                $rescan_error = 1;
                $self->sanlog(-msg => "Error occurred when calling ($cmd), retrying ...");
            } else {
                $rescan_error = 0;
                last;
            }
        }

        if ($rescan_error > 0) {
           $self->_set_err("Error: WinMP::MPIO::get_path_info() Error while calling $cmd");
           return {};
        }
    }

    # check the DSMs defined to see if ONTAP DSM is there. If so,
    # we be checking for which DSM is managing which LUNs and
    # and check appropriately. For systems with ONTAP DSM, we
    # will simply call mpc mpio show lunpaths.
    if (defined($win_client->{dsms})) {
        my @dsms = @{$win_client->{dsms}};
        $have_ontap_dsm = 0;
        foreach my $dsm (@dsms) {
            if ($dsm->{dsmname} =~ /ONTAP/i) {
                $have_ontap_dsm = 1;
            }
        }
    } else {
        # if the dsms array is not defined, assume no ontap DSM
        # Case with DSM outside ONTAP or MS
        $have_ontap_dsm = 0;
    }

    # check DSM version and gather information appropriately
    # First case is not ONTAP DSM installed, call:
    # mpc mpio show lunpaths
    if ($have_ontap_dsm == 0) {
        $per_lun_hash = $self->_mpc_mpio_show_lunpaths($conn, $win_client);
        return {} if $self->error();
    } else {
        # We need to get the LUN information to know what DSM
        # is managing which LUNs. If we are all ONTAP DSM then
        # we call mpc ontap dsm show lunpaths, if we are mixed,
        # we need to call both.

        my $have_another_dsm = 0;

        # first call mpc ontapdsm show lun paths and get back hash
        $per_lun_hash = $self->_mpc_ontapdsm_show_lunpaths($conn, $win_client);
        return {} if $self->error();

        # get show lun output to know which DSM is managing which LUNs
        # add retry logic since, mpc show luns can fail
        my $mpc_error = 0;
        foreach my $retry (1 .. 5) {
            $cmd = "$win_client->{cmd_mpc} show luns";
            ($status, @output) = &ClientUtil::execute($conn, $cmd, 600);
            if ($status != 0 or (grep(/An error occurred/i, @output) > 0)) {
                $mpc_error = 1;
                $self->sanlog(-msg => "Error occurred when calling ($cmd), retrying ...");
            } else {
                $mpc_error = 0;
                last;
            }
        }

        if ($mpc_error > 0) {
           $self->_set_err("Error: WinMP::MPIO::get_path_info() Error while calling $cmd");
           return {};
        }

        # determine if we have another DSM other than ONTAP
        # managing LUNs, and if so, which LUNs.
        foreach my $key (keys %{$per_lun_hash}) {
            $line = (grep(/^$key\s+/, @output))[0];

            # Example line of output:
            # Disk  LUN DSM    Policy Paths  Serial       CFMode  Filer        Proto    LunType Path
            # ----  --- -----  ------ ------ -----        ------  -----        ----     ------- ----
            # Disk1 0   ontap  DLQD   4      Vqn9GJHe58cG SSI     fas960-rtp02 FCP      windows  /vol/nate_1_1/nat1
            my $dsm = (split(/\s+/, $line))[2];
            my $disk = (split(/\s+/, $line))[0];
            $per_lun_hash->{$key}->{dsm} = $dsm;
            if ($dsm !~ /ontap/i) {
                $have_another_dsm = 1;
                push(@non_ontap_dsm_disks, $disk);
            }
        }

        # If we get here, we have LUN that is not managed by ONTAP DSM and
        # we need to call mpc mpio show lunpaths to get the information
        if ($have_another_dsm == 1) {
            my $generic_mpio_per_lun_hash = $self->_mpc_mpio_show_lunpaths($conn, $win_client);
            return {} if $self->error();

            # for the devices that were no ONTAP DSM, delete from
            # from the per_lun_hash and replace with the
            # generic_mpio_per_lun_hash hash (from mpc show lunpaths)
            foreach my $disk (@non_ontap_dsm_disks) {
                delete($per_lun_hash->{$disk});
                $per_lun_hash->{$disk} = $generic_mpio_per_lun_hash->{$disk};
            }
        }
    }

    # traverse the per LUN hash extracting the pathing information
    # and storing in the same hash
    foreach my $key (keys %{$per_lun_hash}) {
        my ($protocol);

        # store number of paths, account for DSM and protocol
        # being on the hash.
        $per_lun_hash->{$key}->{num_paths} =
                     (scalar(keys %{$per_lun_hash->{$key}}) - 2);

        my $alive_cnt = 0;
        my $dead_cnt = 0;
        my $proxy_cnt = 0;
        my $primary_cnt = 0;
        foreach my $path_key (keys %{$per_lun_hash->{$key}}) {
            if ($path_key =~ /protocol|num_paths|dsm/i) {
                next;
            }

            if ($per_lun_hash->{$key}->{$path_key}->{path_state} =~ /1/i) {
                $alive_cnt++;
            } else {
                $dead_cnt++;
            }

            if ($per_lun_hash->{$key}->{$path_key}->{path_type} =~ /primary/i) {
                $primary_cnt++;
            } else {
                $proxy_cnt++;
            }

            # don't need again and want to clean up ending hash
            delete($per_lun_hash->{$key}->{$path_key});
        }
        # store number of paths that are enabled (alive)
        $per_lun_hash->{$key}->{num_paths_alive} = $alive_cnt;

        # store number of paths that are disabled (dead)
        $per_lun_hash->{$key}->{num_paths_dead} = $dead_cnt;

        # store number of paths that are primary
        $per_lun_hash->{$key}->{num_paths_primary} = $primary_cnt;

        # store number of paths that are secondary
        $per_lun_hash->{$key}->{num_paths_secondary} = $proxy_cnt;
    }

    # generate additional device/path details if requested.
    if ($details == 1) {
       # add retry logic since, mpc show luns can fail
       my $mpc_error = 0;
       foreach my $retry (1 .. 5) {
            # get additional per device details
            $cmd = "$win_client->{cmd_mpc} show luns";
            ($status, @output) = &ClientUtil::execute($conn, $cmd, 600);
            if ($status != 0 or (grep(/An error occurred/i, @output) > 0)) {
                $mpc_error = 1;
                $self->sanlog(-msg => "Error occurred when calling ($cmd), retrying ...");
            } else {
                $mpc_error = 0;
                last;
            }
       }

       if ($mpc_error > 0) {
           $self->_set_err("Error: WinMP::MPIO::get_path_info() Error while calling $cmd");
           return {};
       }

       # add additional per LUN details
       foreach my $key (keys %{$per_lun_hash}) {
           $line = (grep(/^$key\s+/, @output))[0];

           # Example line of output:
           # Disk  LUN DSM    Policy Paths  Serial       CFMode  Filer        Proto    LunType Path
           # ----  --- -----  ------ ------ -----        ------  -----        ----     ------- ----
           # Disk1 0   ontap  DLQD   4      Vqn9GJHe58cG SSI     fas960-rtp02 FCP      windows  /vol/nate_1_1/nat1
           my (undef, $lun_id, $dsm, $lb_policy, undef, $serial,
               $cfmode, $controller, $proto, $lun_type, $controller_path) = split(/\s+/, $line);
           $per_lun_hash->{$key}->{lun_id} = $lun_id;
           $per_lun_hash->{$key}->{dsm} = $dsm;
           $per_lun_hash->{$key}->{serial_num} = $serial;
           $per_lun_hash->{$key}->{cfmode} = $cfmode;
           $per_lun_hash->{$key}->{controller} = $controller;
           $per_lun_hash->{$key}->{lun_type} = $lun_type;
           $per_lun_hash->{$key}->{protocol} = $proto;
           $per_lun_hash->{$key}->{controller_path} = $controller_path;
       }

       # add additional per PATH details
       foreach my $key (keys %{$per_lun_hash}) {
          foreach my $retry (1 .. 5) {
               # get additional per device details
               $cmd = "$win_client->{cmd_mpc} ontapdsm show lunpaths $key";
               ($status, @output) = &ClientUtil::execute($conn, $cmd, 600);
               if ($status != 0 or (grep(/An error occurred/i, @output) > 0)) {
                   $mpc_error = 1;
                   $self->sanlog(-msg => "Error occurred when calling ($cmd), retrying ...");
               } else {
                   $mpc_error = 0;
                   last;
               }
          }

          if ($mpc_error > 0) {
              $self->_set_err("Error: WinMP::MPIO::get_path_info() Error while calling $cmd");
              return {};
          }

          foreach my $line (@output) {
             chomp($line);
             $line =~ s/^\s+//;

             # skip lines not important to us
             if ($line =~ /^\s*$/ or $line =~ /DsmID/i or
                 $line =~ /show lunpaths/i or
                 $line =~ /----/) {
                 next;
             }

             # Example line of output:
             # Disk  Policy  DsmID     PathID   State   Proto   Proxy   Primary Reserve Weight
             # ----  ------  --------  -------- -----   -----   -----   ------- ------- ------
             # Disk4 DLQD    02000103  02000101 Active  FCP     False   False   False   000000ff
             my ($device, $policy, $dsm_id,
                 $path_id, $state, $protocol,
                 $is_proxy, $is_primary, $reserve,
                 $weight, $target_port) = split (/\s+/, $line);

             # populate hash with per path details
             $per_lun_hash->{$key}->{per_path_hash}->{$path_id}->{is_proxy} = $is_proxy;
             $per_lun_hash->{$key}->{per_path_hash}->{$path_id}->{protocol} = $protocol;
             $per_lun_hash->{$key}->{per_path_hash}->{$path_id}->{reserve} = $reserve;
             $per_lun_hash->{$key}->{per_path_hash}->{$path_id}->{weight} = $weight;
             $per_lun_hash->{$key}->{per_path_hash}->{$path_id}->{state} = $state;
             $per_lun_hash->{$key}->{per_path_hash}->{$path_id}->{dsm_id} = $dsm_id;
             $per_lun_hash->{$key}->{per_path_hash}->{$path_id}->{target_port} = $target_port;
             $per_lun_hash->{$key}->{per_path_hash}->{$path_id}->{lb_policy} = $policy;

             # calcuate initiator port from portID
             my $init_port = substr($path_id, 0, 2);
             $per_lun_hash->{$key}->{per_path_hash}->{$path_id}->{initiator_port} = $init_port;
          }
       }
    }

    # Dump DSM info when complete getting path info.
    # Valuable debugging path validation issues.
    # Requested by Windows team. Do not check status
    # data is for information reasons only
    $cmd = "$win_client->{cmd_mpc} mpio show summary";
    ($status, @output) = &ClientUtil::execute($conn, $cmd, 180);

    return $per_lun_hash;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $parent, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument \'-parent\'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
    }

    $self->{parent} = $parent;

    # Returning MPIO MP object
    return $self;

}

################################################################################
# Subroutine Name: _mpc_mpio_show_lunpaths
# Objective:       convenience routine to call mpc mpio show lunpaths and
#                  return hash formatted information
# Details:         See POD documentation at the end of this file
################################################################################
sub _mpc_mpio_show_lunpaths {
    my ($self, $conn, $winclient) = @_;

    my ($show_error, $cmd, @output, $status, $line);

    # try command multiple times since exceptions from WMI/MPC
    # are more than possible.
    foreach my $retry (1 .. 3) {
        $cmd = "$winclient->{cmd_mpc} mpio show lunpaths";
        ($status, @output) = &ClientUtil::execute($conn, $cmd, 600);
        if ($status != 0 or (grep(/An error occurred/i, @output) > 0)) {
            $show_error = 1;
            $self->sanlog(-msg => "Error occurred when calling ($cmd), retrying ...");
        } else {
            $show_error = 0;
            last;
        }
    }

    if ($show_error > 0) {
        $self->_set_err("Error: WinMP::MPIO::_mpc_mpio_show_lunpaths() Error while calling mpc mpio show lunpaths");
        return {};
    }

    my $per_lun_hash = {};

    # parse through output generating hash of unique LUNs and their
    # state (up|down) and type (primary|proxy)
    foreach $line (@output) {
        chomp($line);
        $line =~ s/^\s+//;

        # skip lines not important to us
        if ($line =~ /^\s*$/ or $line =~ /BUS/ or
            $line =~ /mpio show lunpaths/i or
            $line =~ /----/ or
            $line !~ /^Disk/i) {
            next;
        }

        my ($device, $path, $state, undef, undef, undef, undef, $protocol) =
                        split (/\s+/, $line);
        $per_lun_hash->{$device}->{$path}->{path_state} = $state;

        if ($path =~ /0x\d+1$/) {
            $per_lun_hash->{$device}->{$path}->{path_type} = "primary";
        } else {
            $per_lun_hash->{$device}->{$path}->{path_type} = "proxy";
        }

        # store the protocol
        $per_lun_hash->{$device}->{protocol} = $protocol;

        # cannot determine DSM from mpio show lunpaths
        # This might be overidden later by data from show luns
        $per_lun_hash->{$device}->{dsm} = "unknown";
    }

    return($per_lun_hash);

}

################################################################################
# Subroutine Name: _mpc_ontapdsm_show_lunpaths
# Objective:       convenience routine to call mpc ontapdsm show lunpaths and
#                  return hash formatted information
# Details:         See POD documentation at the end of this file
################################################################################
sub _mpc_ontapdsm_show_lunpaths {
    my ($self, $conn, $winclient) = @_;

    my ($cmd, @output, $status, $line, $show_error);

    # Example output from mpc ontapdsm show lunpaths:
    # Disk    Policy  DsmID           PathID          State   Proto   Proxy   Primary Reserve Weight  TargetPort
    # ----    ------  --------        --------        -----   -----   -----   ------- ------- ------  ----------
    # Disk19  DLQD    02000106        02000101        Active  FCP     False   False   False   000000ff    0d
    # Disk19  DLQD    01000106        01000101        Active  FCP     False   False   False   000000ff    0b
    # Disk19  DLQD    02000206        02000202        Passive FCP     True    False   False   000000ff    v.0d
    # Disk19  DLQD    01000006        01000002        Passive FCP     True    False   False   000000ff    v.0b

    # try command multiple times since exceptions from WMI/MPC
    # are more than possible.
    foreach my $retry (1 .. 3) {
        $cmd = "$winclient->{cmd_mpc} ontapdsm show lunpaths";
        ($status, @output) = &ClientUtil::execute($conn, $cmd, 600);
        if ($status != 0 or (grep(/An error occurred/i, @output) > 0)) {
            $show_error = 1;
            $self->sanlog(-msg => "Error occurred when calling ($cmd), retrying ...");
        } else {
            $show_error = 0;
            last;
        }
    }

    if ($show_error > 0) {
        $self->_set_err("Error: WinMP::MPIO::_mpc_ontapdsm_show_lunpaths() Error while calling mpc ontapdsm show lunpaths");
        return {};
    }

    my $per_lun_hash = {};

    # parse through output generating hash of unique LUNs and their
    # state (up|down) and type (primary|proxy)
    foreach $line (@output) {
        chomp($line);
        $line =~ s/^\s+//;

        # skip lines not important to us
        if ($line =~ /^\s*$/ or $line =~ /DsmID/ or
            $line =~ /ontapdsm show lunpaths/i or
            $line =~ /----/ or $line !~ /^Disk/i) {
            next;
        }

        my ($device, $policy, $id, $path, $state, $protocol, $proxy, undef, undef, undef, undef) =
                        split (/\s+/, $line);

        # If the path state is not Active or Passive, consider it failed (0).
        if ($state =~ /Active|Passive/i) {
            $per_lun_hash->{$device}->{$path}->{path_state} = 1;
        } else {
            $per_lun_hash->{$device}->{$path}->{path_state} = 0;
        }

        # get proxy/primary information
        if ($proxy =~ /True/i) {
            $per_lun_hash->{$device}->{$path}->{path_type} = "proxy";
        } else {
            $per_lun_hash->{$device}->{$path}->{path_type} = "primary";
        }

        # set protocol
        $per_lun_hash->{$device}->{protocol} = $protocol;
    }

    return($per_lun_hash);

}

1;
