#
# Copyright (c) 2010-2013 NetApp Inc.
# All rights reserved.
#
## @summary Provide an OS generic interface to a SAN Multipathed Device Manager
## @author hrivera@netapp.com
## @status private
## @pod here

=pod

=head1 Package Name

MPDevMgr.pm

=head1 Author

dl-san-automation

=head1 Description

An OS generic multipathed device manager for SanClient

Current Operating System support:
  ESX4.x/ESXi4.x
  ESXi5.x
  ESXi6.x
  Redhat 5u6+
  Redhat 6+
  SLES 10
  Windows 2k3/2k8/2k8r2/2k12/2k12r2

=head1 Inherited Packages

 Error
 San::SanLog

=cut

package San::SanClient::MPDevMgr;

use strict;
use warnings;
use Data::Dumper;
use Clone;
use Scalar::Util qw (weaken);

use Error;
use San::SanLog;
use San::SanUtil;
use San::SanClient::MPDevMgr::MPDev;
use San::SanClient::MPDevMgr::MPDevPath;
use San::SanClient::MPDevMgr::MPDevPathState;

use NACL::CS::Node;
use NACL::CS::Vserver;
use NACL::CS::VserverIscsi;
use NACL::CS::VserverIscsiInitiator;
use NACL::CS::VserverIscsiTpgroup;
use NACL::CS::VserverFcp;
use NACL::CS::VserverFcpInitiator;
use NACL::CS::Volume;
use NACL::CS::StorageAggregate;
use NACL::CS::NetworkInterface;
use NACL::CS::Lun;
use NACL::CS::LunMapped;
use NACL::CS::StorageFailover;

use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NACL::Exceptions::NoElementsFound;

use vars qw (@ISA);

@ISA = qw (Error SanLog);

my $Log = log_global();

###############################################################################
#                               Public Methods
###############################################################################

=pod

=head1 B<Public Methods>

=head2 B<new(%args)>

The constructor for MPDevMgr. Instantiates and returns an MPDevMgr object.
This constructor should not be used directly, use San::SanClient::create_mpdevmgr()
off of a SanClient object instead. The inputs to that method are the same as this.

=over 4

=item Inputs:

=over 2

=item I<-nacl_cluster_objs> => \@array

 Optional: An array reference of NACL::C::Cluster objects to use when
 generating LUN relationships for Cmode LUNs.

=item I<-vservers> => \@array

 Optional: An array reference of vserver names to search through when
 generating LUN relationships.

=item I<-sanfiler7m_objs> => \@array

 Optional: An array reference of San::SanFiler7M objects to use when generating
 LUN relationships for 7mode LUNs.

=item I<-controller_cluster_objs> => \@array

 Optional: (I<deprecated>) An array reference of San::ControllerCluster objects to use when
 generating LUN relationships for Cmode LUNs. It is recommended to use C<-nacl_cluster_objs> instead

=item I<-obj_template> => \%mpdevmgr_obj_template

 Optional: An object template of an MPDevMgr object to use as a template when
 constructing this object.

=back

=item Outputs:

=over 2

A San::SanClient::MPDevMgr object

=back

=item Example:

 # use the create_mpdevmgr() method off the SanClient object to instantiate the object
 # pass in an array of NACL::STask::Cluster/NACL::C::Cluster objects to generate
 # lun relationship data
 $sanclient_obj->create_mpdevmgr(-nacl_cluster_objs => \@nacl_cluster_objs);
 if ($sanclient_obj->error()) {
     $Test->nlog($sanclient_obj->errmsg());
 }

=back

=cut

sub new {
    my ($class, %args) = @_;

    my $self = {};

    bless($self, $class);

    my $sc_obj = delete($args{-sc_obj});
    my $obj_template = delete($args{-obj_template});
    my $hosttype = $sc_obj->get_connection()->{hostp}{hosttype};

    # initialize the way we handle error setting
    $self->{error_dont_overwrite_errors} = 1;

    if ($hosttype =~ /linux-vmware/i)  {
        if ($sc_obj->has_vma_host()) {
            use San::SanClient::MPDevMgr::ESXiMPDevMgr;
            use San::SanClient::MPDevMgr::ESXMPDevMgr;
            @ISA = ("Error", "SanLog",
                    "San::SanClient::MPDevMgr::ESXiMPDevMgr",
                    "San::SanClient::MPDevMgr::ESXMPDevMgr");
        } else {
            use San::SanClient::MPDevMgr::ESXMPDevMgr;
            @ISA = ("Error", "SanLog",
                    "San::SanClient::MPDevMgr::ESXMPDevMgr");
        }
    } elsif ($hosttype =~ /unix-linux.*-citrix/i) {
        use San::SanClient::MPDevMgr::CXSMPDevMgr;
        @ISA = ("Error", "SanLog",
                "San::SanClient::MPDevMgr::CXSMPDevMgr");
    } elsif ($hosttype =~ /linux/) {
        use San::SanClient::MPDevMgr::LinuxMPDevMgr;
        @ISA = ("Error", "SanLog",
                "San::SanClient::MPDevMgr::LinuxMPDevMgr");
    } elsif ($hosttype =~ /windows/) {
        use San::SanClient::MPDevMgr::WinMPDevMgr;
        @ISA = ("Error", "SanLog",
                "San::SanClient::MPDevMgr::WinMPDevMgr");
    } else {
        $self->_set_err("MPDevMgr currently does not have support for hosttype '$hosttype'");
        return $self;
    }

    # initialize a few things
    $self->{devices} = {};
    $self->{os_specific_data} = {};
    my @isa_tmp = @ISA;
    $self->{__ISA__} = \@isa_tmp;

    # if an object template was provided, copy over all the contents
    if (defined($obj_template)) {
        # copy over all the contents
        foreach my $key (keys %{$obj_template}) {
            $self->{$key} = $obj_template->{$key};
        }

        # correct the sc_obj link for all the sub objects
        Scalar::Util::weaken($self->{sc_obj} = $sc_obj);
        foreach my $mpdev ($self->get_device_objs()) {
            Scalar::Util::weaken($mpdev->{sc_obj} = $sc_obj);
        }

        # return now
        return $self;
    }

    # store the san client object with a weak reference
    # to prevent memory leaks due to circular dependencies
    Scalar::Util::weaken($self->{sc_obj} = $sc_obj);

    # call the base class constructor
    $self->SUPER::_new(%args);

    # refresh the devmgr
    # this will create MPDev objects for every NETAPP LUN on the system
    $self->refresh(%args);

    return $self;
}

=pod

=head2 B<is_supported($sanclient_obj)>

Determine if the SanClient object is supported in the MPDevMgr library

=over 4

=item Inputs:

=over 2

=item $sanclient_obj

 Required: SanClient object to check if supported by MPDevMgr.

=back

=item Outputs:

=over 2

1 or 0

=back

=item Example:

 # determine if the sanclient object is supported
 if ($sanclient_obj->San::SanClient::MPDevMgr::is_supported()) {
     ...
 }

=back

=cut

sub is_supported {
    my ($obj) = @_;

    my $hosttype = "";
    my $mptype = "";
    if (ref($obj) =~ /San::SanClient::Host/) {
        $hosttype = $obj->hosttype();
        $mptype = $obj->{mp_type};
    } else {
        $hosttype = $obj->{sc_obj}->hosttype();
        $mptype = $obj->{sc_obj}->{mp_type};
    }

    if ($hosttype =~ /linux-vmware|linux|windows/i and $mptype =~ /dmmp|mpio|native/
       and $hosttype !~ /citrix/i) {
        return 1;
    } else {
        return 0;
    }
}

=pod

=head2 B<find_missing_devices()>

Find any missing devices between mpdevmgr refreshes.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

An array of missing San::SanClient::MPDevMgr::MPDev objects

=back

=item Example:

 # create the mpdevmgr
 $sc_obj->create_mpdevmgr();

 # do things that may or may not cause LUN loss ...

 # refresh the mpdevmgr, and see if any LUNs disappeared.
 $sc_obj->mpdevmgr()->refresh();
 my @missing_mpdev_objs = $sc_obj->mpdevmgr()->find_missing_devices();

=back

=cut

sub find_missing_devices {
    my ($self) = @_;

    my @curr_mpdev_objs = $self->get_device_objs();
    my @old_mpdev_objs = $self->get_old_device_objs();

    my @missing_devices = ();
    if (scalar(@old_mpdev_objs) ne scalar(@curr_mpdev_objs)) {
        foreach my $old_mpdev_obj (@old_mpdev_objs) {
            my $found = 0;
            foreach my $curr_mpdev_obj (@curr_mpdev_objs) {
                if ($old_mpdev_obj->mp_device_name() eq $curr_mpdev_obj->mp_device_name()) {
                    $found = 1;
                    last;
                }
            }
            push(@missing_devices, $old_mpdev_obj) if (not $found);
        }
    }

    return @missing_devices;
}

=pod

=head2 B<refresh(%args)>

This will refresh all San::SanClient::MPDevMgr::MPDev objects that are
associated to the MPDevMgr. It will rediscover the LUN relationships, and
current mount information regarding each device.

=over 4

=item Inputs:

=over 2

=item I<-nacl_cluster_objs> => \@array

 Optional: An array reference of NACL::C::Cluster objects to use when
 generating LUN relationships for Cmode LUNs.

=item I<-vservers> => \@array

 Optional: An array reference of vserver names to search through when
 generating LUN relationships.

=item I<-sanfiler7m_objs> => \@array

 Optional: An array reference of San::SanFiler7M objects to use when generating
 LUN relationships for 7mode LUNs.

=item I<-controller_cluster_objs> => \@array

 Optional: (I<deprecated>) An array reference of San::ControllerCluster objects to use when
 generating LUN relationships for Cmode LUNs. It is recommended to use C<-nacl_cluster_objs>
 instead.

=item I<-rescan> => 0|1

 Optional: If set, will rescan the SCSI bus on the client prior to finding devices. Only Windows implements
 the rescan feature at this time.
 Default: 0 (no rescan)

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # refresh the mpdevmgr object using the specified list of NACL::C::Clsuter/NACL::STask::Cluster objects
 # when generating lun relationship data
 $sanclient_obj->mpdevmgr()->refresh(-nacl_cluster_objs => \@nacl_cluster_objs);
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub refresh {
    my ($self, %args) = @_;

    # default rescan to off
    $args{rescan} = 0 if (not exists($args{rescan}));

    @ISA = @{$self->{__ISA__}};

    # first mark all devices as pending removal
    $self->_mark_all_devices_pending_removal();

    # invoke the refresh method on the base class
    $self->SUPER::_refresh(%args);
    if ($self->error()) {
        return;
    }

    # now, remove all devices that are still pending removal
    # after the refresh, devices that are no longer present
    # should be removed from the active device list and
    # stored in the old devices list with device_state 'old_device_not_present'
    $self->_clear_all_devices_pending_removal();

    # flag devices that don't have any paths to them
    my @dev_objs = $self->get_device_objs();

    # update all these devices
    if (scalar(@dev_objs)) {
        # update device state for devices that may not have any paths
        foreach my $dev_obj (@dev_objs) {
            my $path_state_obj = $dev_obj->get_current_path_state();
            if ($path_state_obj->total_alive_paths() == 0) {
                $dev_obj->device_state("no_paths");
            }
        }

        # load all devices current mount states
        $self->load_all_devices();
        if ($self->error()) {
            $self->_set_err("Failed to load device mount states", $self->errmsg());
            return;
        }

        # generate lun relationships for devices we have
        $self->generate_lun_relationships(%args);
        if ($self->error()) {
            $self->_set_err("Failed to generate lun relationship", $self->errmsg());
            return;
        }
    }

    return;
}

=pod

=head2 B<refresh_paths(%args)>

This will refresh all San::SanClient::MPDevMgr::MPDev objects that are
associated to the MPDevMgr. This is a lighter weight version of San::SanClient::MPDevMgr::refresh()
that only focuses on refreshing the path states for all MPDev objects
currently discovered by the system.

=over 4

=item Inputs:

=over 2

=item I<-controller_cluster_objs> => \@array

 Optional: An array reference of San::ControllerCluster objects to use when
 generating LUN relationships for Cmode LUNs.

=item I<-nacl_cluster_objs> => \@array

 Optional: An array reference of NACL::C::Cluster objects to use when
 generating LUN relationships for Cmode LUNs.

=item I<-vservers> => \@array

 Optional: An array reference of vserver names to search through when
 generating LUN relationships.

=item I<-sanfiler7m_objs> => \@array

 Optional: An array reference of San::SanFiler7M objects to use when generating
 LUN relationships for 7mode LUNs.

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # refresh the mpdevmgr object using the specified list of cc objects
 $sanclient_obj->mpdevmgr()->refresh(-controller_cluster_objs => \@cc_objs);
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub refresh_paths {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    # invoke the refresh method on the inherited class
    $self->SUPER::_refresh_paths(%args);
    if ($self->error()) {
        return;
    }

    # flag devices that don't have any paths to them
    my @dev_objs = $self->get_device_objs();
    foreach my $dev_obj (@dev_objs) {
        my $path_state_obj = $dev_obj->get_current_path_state();
        if ($path_state_obj->total_alive_paths() == 0) {
            $dev_obj->device_state("no_paths");
        }
    }

    return;
}

=pod

=head2 B<generate_lun_relationships(%args)>

This will generate lun relationship data for all MPDev objects
currently discovered.

=over 4

=item Inputs:

=over 2

=item I<-controller_cluster_objs> => \@array

 Optional: An array reference of San::ControllerCluster objects to use when
 generating LUN relationships for Cmode LUNs.

=item I<-nacl_cluster_objs> => \@array

 Optional: An array reference of NACL::C::Cluster objects to use when
 generating LUN relationships for Cmode LUNs.

=item I<-vservers> => \@array

 Optional: An array reference of vserver names to search through when
 generating LUN relationships.

=item I<-sanfiler7m_objs> => \@array

 Optional: An array reference of San::SanFiler7M objects to use when generating
 LUN relationships for 7mode LUNs.

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # generate lun relationships on all the MPDev objects in the MPDevMgr
 $sanclient_obj->mpdevmgr()->generate_lun_relationships(-controller_cluster_objs => \@cc_objs);
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub generate_lun_relationships {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    # obtain all the device objects on the devmgr
    my @mp_device_objs = $self->get_device_objs();

    # return immediately if there are no devices
    if (not scalar(@mp_device_objs)) {
        return;
    }

    my $sc_obj = $self->{sc_obj};
    my $fc_init_wwpn_filter;
    my $iscsi_init_nn_filter;
    my @init_objs = $sc_obj->get_initiators();
    foreach my $init_obj (@init_objs) {
        if ($init_obj->get_protocol() =~ /fcp/i) {
            $fc_init_wwpn_filter .= SanUtil::wwn_add_colons(wwn =>$init_obj->get_portname()) . "|";
        } elsif ($init_obj->get_protocol() =~ /iscsi/i) {
            $iscsi_init_nn_filter .= $init_obj->get_nodename() . "|";
        }
    }
    chop($fc_init_wwpn_filter)  if ($fc_init_wwpn_filter);
    chop($iscsi_init_nn_filter) if ($iscsi_init_nn_filter);

    # get all cluster objects
    my $cluster_objs = $args{-controller_cluster_objs};
    my $vservers = $args{-vservers};
    my $nacl_cluster_objs = $args{-nacl_cluster_objs};
    my $nacl_cluster_content_passthru = $args{-nacl_cluster_content};
    my $sanfiler7m_objs = $args{-sanfiler7m_objs};
    my $cluster_content = {};
    my $nacl_cluster_content = {};

    if (ref($cluster_objs) eq "ARRAY") {
        my %add_args;
        if (ref($vservers) eq "ARRAY") {
            %add_args = (vserver => join(",", @{$vservers}));
        }

        foreach my $cluster_obj (@{$cluster_objs}) {

            # collect data about the cluster upfront to speed up the discovery process
            my $cserver_obj = $cluster_obj->get_cserver_obj();
            if ($cluster_obj->error()) {
                $self->_set_err("Failed to obtain cserver object");
                return;
            }

            my $vserver_iscsi_show_data = $cserver_obj->invoke_cmd_obj(cmd => "vserver_iscsi_show",
                                                                       cmd_timeout => 1800,
                                                                       %add_args,
                                                                   );
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to invoke vserver iscsi show", $cserver_obj->errmsg());
                return;
            }

            my $vserver_fcp_show_data = $cserver_obj->invoke_cmd_obj(cmd => "vserver_fcp_show",
                                                                     cmd_timeout => 1800,
                                                                     %add_args,
                                                                 );
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to invoke vserver fcp show", $cserver_obj->errmsg());
                return;
            }

            my $vserver_iscsi_initiator_show_data = $cserver_obj->invoke_cmd_obj(cmd => "vserver_iscsi_initiator_show",
                                                                                 cmd_timeout => 1800,
                                                                                 %add_args,
                                                                             );
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to invoke vserver iscsi initiator show", $cserver_obj->errmsg());
                return;
            }

            my $vserver_fcp_initiator_show_data = $cserver_obj->invoke_cmd_obj(cmd => "vserver_fcp_initiator_show",
                                                                               cmd_timeout => 1800,
                                                                               %add_args,
                                                                           );
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to invoke vserver fcp initiator show", $cserver_obj->errmsg());
                return;
            }

            my $lun_mapped_show_data = $cserver_obj->invoke_cmd_obj(cmd => "lun_mapped_show",
                                                                    cmd_timeout => 1800,
                                                                    %add_args,);
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to invoke lun mapped show", $cserver_obj->errmsg());
                return;
            }

            my $vserver_iscsi_tpgroup_show_data = $cserver_obj->invoke_cmd_obj(cmd => "vserver_iscsi_tpgroup_show",
                                                                               cmd_timeout => 1800,
                                                                               %add_args,
                                                                           );
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to invoke vserver iscsi tpgroup show", $cserver_obj->errmsg());
                return;
            }

            my @sfo_show_data = $cserver_obj->sfo_show();
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to obtain storage failover information", $cserver_obj->errmsg());
                return;
            }

            my @vserver_objs = $cserver_obj->fetch_vserver_objs(cmd_timeout => 1800,
                                                                %add_args);
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to fetch vserver objects", $cserver_obj->errmsg());
                return;
            }

            my @vol_objs = $cserver_obj->fetch_volume_objs(cmd_timeout => 1800,
                                                           %add_args);
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to fetch volume objects", $cserver_obj->errmsg());
                return;
            }

            my @lun_objs = $cserver_obj->fetch_lun_objs(cmd_timeout => 1800,
                                                        %add_args);
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to fetch LUN objects", $cserver_obj->errmsg());
                return;
            }

            my @lif_objs = $cserver_obj->fetch_lif_objs(cmd_timeout => 1800,
                                                        %add_args);
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to fetch lif objects", $cserver_obj->errmsg());
                return;
            }

            my @aggr_objs = $cserver_obj->get_aggregate_objs();
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to fetch aggregate objects", $cserver_obj->errmsg());
                return;
            }

            # store all this data on cluster content
            push(@{$cluster_content->{vserver_iscsi_show_data}}, @{$vserver_iscsi_show_data});
            push(@{$cluster_content->{vserver_fcp_show_data}}, @{$vserver_fcp_show_data});
            push(@{$cluster_content->{vserver_iscsi_initiator_show_data}},@{$vserver_iscsi_initiator_show_data});
            push(@{$cluster_content->{vserver_fcp_initiator_show_data}}, @{$vserver_fcp_initiator_show_data});
            push(@{$cluster_content->{lun_mapped_show_data}}, @{$lun_mapped_show_data});
            push(@{$cluster_content->{vserver_iscsi_tpgroup_show_data}}, @{$vserver_iscsi_tpgroup_show_data});
            push(@{$cluster_content->{sfo_show_data}}, @sfo_show_data);
            push(@{$cluster_content->{vserver_objs}}, @vserver_objs);
            push(@{$cluster_content->{vol_objs}}, @vol_objs);
            push(@{$cluster_content->{lun_objs}}, @lun_objs);
            push(@{$cluster_content->{lif_objs}}, @lif_objs);
            push(@{$cluster_content->{aggr_objs}}, @aggr_objs);
        }
    } elsif (ref($nacl_cluster_objs) eq "ARRAY") {
        # use NACL objects for discovery
        # prepare the vserver filter array assuming that no vservers passed in
        my @vs_filters = (undef);
        if (ref($vservers) eq "ARRAY" and scalar(@{$vservers})) {
            @vs_filters = ();
            my @vserver_scope_lists = &SanUtil::break_up_array(array => $vservers,
                                                               size => 10);
            foreach my $vs_scope_list (@vserver_scope_lists) {
                my %vs_filter = (filter => { vserver => join(",", @{$vs_scope_list}) });
                push(@vs_filters, \%vs_filter);
            }
        }

        # initialize the nacl cluster content fields
        $nacl_cluster_content->{vserver_objs} = [];
        $nacl_cluster_content->{vserver_iscsi_objs} = [];
        $nacl_cluster_content->{vserver_iscsi_init_objs} = [];
        $nacl_cluster_content->{vserver_fcp_objs} = [];
        $nacl_cluster_content->{vserver_fcp_init_objs} = [];
        $nacl_cluster_content->{vol_objs} = [];
        $nacl_cluster_content->{lun_objs} = [];
        $nacl_cluster_content->{lun_mapped_objs} = [];
        $nacl_cluster_content->{lif_objs} = [];
        $nacl_cluster_content->{aggr_objs} = [];
        $nacl_cluster_content->{sfo_show_objs} = [];

        # traverse each cluster
        foreach my $nacl_cluster_obj (@{$nacl_cluster_objs}) {

            # gather all the necessary data up front for this cluster
            my @global_vs_objs;
            my $max_retries = 5;
            my $retries = 0;
            my $success = 0;
            while (1) {
                try {
                    foreach my $vs_filter (@vs_filters) {
                        my @vs_objs = NACL::CS::Vserver->fetch(command_interface => $nacl_cluster_obj,
                                                               allow_empty => 1,
                                                               %{$vs_filter});
                        push(@global_vs_objs, @vs_objs);
                        push(@{$nacl_cluster_content->{vserver_objs}}, @vs_objs);
                        $success = 1;
                    }
                } otherwise {
                    my $e = shift;
                    if ($e->text() =~ /Could Not Connect/i and $retries < $max_retries) {
                        $retries++;
                        $Log->comment("Unable to connect to cluster, will retry in 20 seconds (retry ${retries} of ${max_retries})");
                        sleep 20;
                    } else {
                        $self->_set_err("Exception caught: " . $e->text());
                    }
                };
                if ($self->error()) {
                    return;
                }
                last if $success;
            }

            # if there are no vservers that match the filter, then there really is no point in continuing with this cluster
            next if (not scalar(@global_vs_objs));

            try {
                # only collect iscsi info if this client is doing iscsi
                if ($iscsi_init_nn_filter) {
                    foreach my $vs_filter (@vs_filters) {
                        my @vs_iscsi_objs = NACL::CS::VserverIscsi->fetch(command_interface => $nacl_cluster_obj,
                                                                          allow_empty => 1,
                                                                          %{$vs_filter});
                        push(@{$nacl_cluster_content->{vserver_iscsi_objs}}, @vs_iscsi_objs);
                    }
                }
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return;
            }

            try {
                # only collect iscsi info if this client is doing iscsi
                if ($iscsi_init_nn_filter) {
                    foreach my $vs_filter (@vs_filters) {
                        my $vservers = $vs_filter->{filter}->{vserver};
                        my %filter = (filter => { vserver => $vservers });
                        $filter{filter}->{'initiator-name'} = $iscsi_init_nn_filter;
                        my @vs_iscsi_init_objs = NACL::CS::VserverIscsiInitiator->fetch(command_interface => $nacl_cluster_obj,
                                                                                        allow_empty => 1,
                                                                                        %filter);
                        push(@{$nacl_cluster_content->{vserver_iscsi_init_objs}}, @vs_iscsi_init_objs);
                    }
                }
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return;
            }

            try {
                # only collect iscsi info if this client is doing iscsi
                if ($iscsi_init_nn_filter) {
                    foreach my $vs_filter (@vs_filters) {
                        my @vs_iscsi_tpgroup_objs = NACL::CS::VserverIscsiTpgroup->fetch(command_interface => $nacl_cluster_obj,
                                                                                         allow_empty => 1,
                                                                                         %{$vs_filter});
                        push(@{$nacl_cluster_content->{vserver_iscsi_tpgroup_objs}}, @vs_iscsi_tpgroup_objs);
                    }
                }
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return;
            }

            try {
                # only collect vserver fcp objects if this client is doing FCP
                if ($fc_init_wwpn_filter) {
                    foreach my $vs_filter (@vs_filters) {
                        my @vs_fcp_objs = NACL::CS::VserverFcp->fetch(command_interface => $nacl_cluster_obj,
                                                                      allow_empty => 1,
                                                                      %{$vs_filter});
                        push(@{$nacl_cluster_content->{vserver_fcp_objs}}, @vs_fcp_objs);
                    }
                }
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return;
            }

            try {
                # only collect vserver fcp initiator objects if this client is doing FCP
                if ($fc_init_wwpn_filter) {
                    foreach my $vs_filter (@vs_filters) {
                        my $vservers = $vs_filter->{filter}->{vserver};
                        my %filter = (filter => { vserver => $vservers });
                        $filter{filter}->{wwpn} = $fc_init_wwpn_filter;
                        my @vs_fcp_init_objs = NACL::CS::VserverFcpInitiator->fetch(command_interface => $nacl_cluster_obj,
                                                                                    allow_empty => 1,
                                                                                    %filter);
                        push(@{$nacl_cluster_content->{vserver_fcp_init_objs}}, @vs_fcp_init_objs);
                    }
                }
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return;
            }

            # if there exists volume nacl cluster content for this cluster, use that instead
            if (exists $nacl_cluster_content_passthru->{$nacl_cluster_obj->name()} and
                exists $nacl_cluster_content_passthru->{$nacl_cluster_obj->name()}->{vol_objs} and
                scalar(@{$nacl_cluster_content_passthru->{$nacl_cluster_obj->name()}->{vol_objs}})) {
                push(@{$nacl_cluster_content->{vol_objs}},
                     @{$nacl_cluster_content_passthru->{$nacl_cluster_obj->name()}->{vol_objs}});
            } else {
                try {
                    foreach my $vs_filter (@vs_filters) {
                        my @vol_objs = NACL::CS::Volume->fetch(command_interface => $nacl_cluster_obj,
                                                               allow_empty => 1,
                                                               requested_fields => ["vserver","volume","aggregate","size"],
                                                               %{$vs_filter});
                        push(@{$nacl_cluster_content->{vol_objs}}, @vol_objs);
                    }
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Exception caught: " . $e->text());
                };
                if ($self->error()) {
                    return;
                }
            }

            # if there exists lun nacl cluster content for this cluster, use that instead
            if (exists $nacl_cluster_content_passthru->{$nacl_cluster_obj->name()} and
                exists $nacl_cluster_content_passthru->{$nacl_cluster_obj->name()}->{lun_objs} and
                scalar(@{$nacl_cluster_content_passthru->{$nacl_cluster_obj->name()}->{lun_objs}})) {
                push(@{$nacl_cluster_content->{lun_objs}},
                     @{$nacl_cluster_content_passthru->{$nacl_cluster_obj->name()}->{lun_objs}});
            } else {
                try {
                    foreach my $vs_filter (@vs_filters) {
                        my @lun_objs = NACL::CS::Lun->fetch(command_interface => $nacl_cluster_obj,
                                                            allow_empty => 1,
                                                            requested_fields => ["vserver","path","volume","qtree","lun","size","serial","ostype"],
                                                            %{$vs_filter});
                        push(@{$nacl_cluster_content->{lun_objs}}, @lun_objs);
                    }
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Exception caught: " . $e->text());
                };
                if ($self->error()) {
                    return;
                }
            }

            # if there exists lun_mapped nacl cluster content for this cluster, use that instead
            if (exists $nacl_cluster_content_passthru->{$nacl_cluster_obj->name()} and
                exists $nacl_cluster_content_passthru->{$nacl_cluster_obj->name()}->{lun_mapped_objs} and
                scalar(@{$nacl_cluster_content_passthru->{$nacl_cluster_obj->name()}->{lun_mapped_objs}})) {
                push(@{$nacl_cluster_content->{lun_mapped_objs}},
                     @{$nacl_cluster_content_passthru->{$nacl_cluster_obj->name()}->{lun_mapped_objs}});
            } else {
                try {
                    foreach my $vs_filter (@vs_filters) {
                        my @lun_mapped_objs = NACL::CS::LunMapped->fetch(command_interface => $nacl_cluster_obj,
                                                                         allow_empty => 1,
                                                                         requested_fields => ["vserver","igroup","lun-id","path","initiators"],
                                                                         %{$vs_filter});
                        push(@{$nacl_cluster_content->{lun_mapped_objs}}, @lun_mapped_objs);
                    }
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Exception caught: " . $e->text());
                };
                if ($self->error()) {
                    return;
                }
            }

            try {
                foreach my $vs_filter (@vs_filters) {
                    my @lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $nacl_cluster_obj,
                                                                     allow_empty => 1,
                                                                     requested_fields => ["vserver","lif",
                                                                                          "wwpn","address","netmask", "netmask-length",
                                                                                          "home-node","home-port","curr-node","curr-port",
                                                                                          "data-protocol","status-admin"],
                                                                     %{$vs_filter});
                    push(@{$nacl_cluster_content->{lif_objs}}, @lif_objs);
                }
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return;
            }

            try {
                my @aggr_objs = NACL::CS::StorageAggregate->fetch(command_interface => $nacl_cluster_obj,
                                                                  allow_empty => 1);
                push(@{$nacl_cluster_content->{aggr_objs}}, @aggr_objs);
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return;
            }

            try {
                my @sfo_show_objs = NACL::CS::StorageFailover->fetch(command_interface => $nacl_cluster_obj,
                                                                     allow_empty => 1);
                push(@{$nacl_cluster_content->{sfo_show_objs}}, @sfo_show_objs);
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return;
            }
        }
    } elsif (ref($sanfiler7m_objs) ne "ARRAY" or not scalar(@{$sanfiler7m_objs})) {
        $self->sanlog(-msg => "No cluster/controller objects were passed in to generate lun relationships against");
    }

    # for each device, determine the lun relationship,
    # invoke the generate_lun_relationship() method off each mp device
    my @relationships;
    my @non_relationships;
    foreach my $mpdev_obj (@mp_device_objs) {
        $mpdev_obj->generate_lun_relationship(%args,
                                              -cluster_content => $cluster_content,
                                              -nacl_cluster_content => $nacl_cluster_content);
        if ($mpdev_obj->error()) {
            $self->_set_err("Failed to generate lun relationship", $mpdev_obj->errmsg());
            return;
        }

        # create the node id to home node mapping for this mp device
        # if an ontap home node was discovered for it
        if (defined($mpdev_obj->ontap_home_node()) and length($mpdev_obj->ontap_home_node())) {
            $mpdev_obj->ontap_node_id($self->_convert_to_node_id(-node_name => $mpdev_obj->ontap_home_node()));
        }

        if ($mpdev_obj->has_lun_relationship()) {
            push(@relationships, "\t" . $mpdev_obj->mp_device_name() . " (" . $mpdev_obj->ontap_controller() . ": " . $mpdev_obj->ontap_lun_path() . ")");
        } else {
            push(@non_relationships, "\t" . $mpdev_obj->mp_device_name());
        }
    }

    if (scalar(@relationships)) {
        $self->sanlog(-msg => "LUN Relationships:\n" . join("\n", @relationships));
    }
    if (scalar(@non_relationships)) {
        $self->sanlog(-msg => "LUN Relationships not made:\n" . join("\n", @non_relationships));
    }

    return;
}

=pod

=head2 B<get_device_objs()>

Obtain all MPDev objects registered with the MPDevMgr

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # get the array of MPDev objects that are associated with this MPDevMgr
 my @mpdev_objs = $sanclient_obj->mpdevmgr()->get_device_objs();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub get_device_objs {
    my ($self) = @_;

    @ISA = @{$self->{__ISA__}};

    return values %{$self->get_devices_hash()};
}

=pod

=head2 B<get_old_device_objs()>

Obtain all (old) MPDev objects registered with the MPDevMgr. These represent the previous state of all
MPDev objects

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # get the array of old MPDev objects that are associated with this MPDevMgr
 my @mpdev_objs = $sanclient_obj->mpdevmgr()->get_old_device_objs();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub get_old_device_objs {
    my ($self) = @_;

    @ISA = @{$self->{__ISA__}};

    return values %{$self->get_old_devices_hash()};
}

=pod

=head2 B<get_device_names()>

Obtain all MPDev names registered with the MPDevMgr.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # obtain the names of all the MPDev objects
 my @mpdev_names = $sanclient_obj->mpdevmgr()->get_device_names();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub get_device_names {
    my ($self) = @_;

    @ISA = @{$self->{__ISA__}};

    return keys %{$self->get_devices_hash()};
}

=pod

=head2 B<get_old_device_names()>

Obtain all old MPDev names registered with the MPDevMgr.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # obtain the names of all the old MPDev objects
 my @mpdev_names = $sanclient_obj->mpdevmgr()->get_old_device_names();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub get_old_device_names {
    my ($self) = @_;

    @ISA = @{$self->{__ISA__}};

    return keys %{$self->get_old_devices_hash()};
}

=head2 B<get_devices_hash()>

Obtain a hash that maps MPDev name to MPDev object for all MPDev
objects in the MPDevMgr.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # get the $hash = { <name1> => <obj1>, <name2> => <obj2>, ... }
 my $hash = $sanclient_obj->mpdevmgr()->get_devices_hash();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub get_devices_hash {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    return $self->{devices};
}


=pod

=head2 B<get_old_devices_hash()>

Obtain a hash that maps MPDev name to MPDev object for all old MPDev
objects in the MPDevMgr.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # get the $hash = { <name1> => <obj1>, <name2> => <obj2>, ... }
 my $hash = $sanclient_obj->mpdevmgr()->get_old_devices_hash();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub get_old_devices_hash {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    return $self->{old_devices};
}

=pod

=head2 B<get_device()>

Obtain the MPDev object specified by name

=over 4

=item Inputs:

=over 2

=item I<-mp_device_name> => scalar

Required: The MPDev object to retrieve.

=item I<-old_devices> => scalar

Optional: search through old devices

Default: 0 (off)

=back

=item Outputs:

=over 2

An MPDev object if found, otherwise an error will be set.

=back

=item Example:

 # Get the MPDev object associated this this device name
 my $mpdev_obj = $sanclient_obj->mpdevmgr()->get_device(-mp_device_name => "naa.60098129391fe3832abc834df");
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub get_device {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    my $device_hash;
    if ($args{-old_devices}) {
        $device_hash = $self->get_old_devices_hash();
    } else {
        $device_hash = $self->get_devices_hash();
    }

    if (exists($device_hash->{$args{-mp_device_name}})) {
        return $device_hash->{$args{-mp_device_name}};
    } else {
        $self->_set_err("Device '" . $args{-mp_device_name} . "' does not exist in the Device Manager");
        return undef;
    }
}

=pod

=head2 B<reset_all_device_states()>

Reset the device state for all active MPDev objects to 'free'. Does not affect devices
that have device_state: 'no_paths'.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Mark all devices free
 $sanclient_obj->mpdevmgr()->reset_all_device_states();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub reset_all_device_states {
    my ($self) = @_;

    @ISA = @{$self->{__ISA__}};

    foreach my $mp_dev_obj ($self->get_device_objs()) {
        # skip devices that have no_paths
        next if ($mp_dev_obj->device_state() =~ /no_paths/i);

        $mp_dev_obj->device_state("free");
    }

    return;
}

=pod

=head2 B<load_all_devices()>

Load in the current mount/partitioning state for all MPDev objects.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Load in the current state for all devices
 $sanclient_obj->mpdevmgr()->load_all_devices();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub load_all_devices {
    my ($self) = @_;

    @ISA = @{$self->{__ISA__}};

    $self->SUPER::_load_all_devices();
    if ($self->error()) {
        $self->_set_err("Failed to load device states");
        return;
    }

    return;
}

=head2 B<offline_devices()>

Offline the devices specified.

=over 4

=item Inputs:

=over 2

=item I<-mpdev_objs> => \@array

 Required: An array reference of MPDev objects to offline

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # offline the specified list of MPDev objects
 $sanclient_obj->mpdevmgr()->offline_devices(-mpdev_objs => \@mpdev_objs);
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub offline_devices {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    $self->SUPER::_offline_devices(%args);
    if ($self->error()) {
        $self->_set_err("Failed to offline the specified device list");
        return;
    }

    return;
}

=pod

=head2 B<generate_device_reports()>

Report all devices registered with the MPDevMgr object in a readable fashion
so that it can be displayed in the logs.

=over 4

=item Inputs:

=over 2

=item I<-mpdev_objs> => \@array

 Required: An array reference of MPDev objects to generate reports

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # return back a pretty report of the devices
 my $reports = $sanclient_obj->mpdevmgr()->generate_device_reports();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }
 # print it to the screen
 use NATE::Log;
 NATE::Log::log_global()->comment(@{$reports});

=back

=cut

sub generate_device_reports {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    my $user_objs = delete($args{-mpdev_objs});
    my @mpdev_objs = ();
    if (!$user_objs) {
        @mpdev_objs = $self->get_device_objs();
        if ($self->error()) {
            $self->_set_err("Failed to get device hash from the MPDevMgr");
            return [];
        }
    } else {
        # use the specified user device list
        if (ref($user_objs) eq "ARRAY") {
            @mpdev_objs = @{$user_objs};
        }
    }

    my @reports = ();

    # generate reports for each device
    foreach my $mpdev_obj (@mpdev_objs) {
        if (ref($mpdev_obj) !~ /MPDev$/) {
            $self->_set_err("Invalid object specified into generate_device_reports(), " .
                            "objects must be derived from class 'San::SanClient::MPDevMgr::MPDev'.");
            return [];
        }
        push(@reports, @{$mpdev_obj->generate_device_report()});
        push(@reports, "");
    }

    return \@reports;
}

=head2 B<os_specific_validation()>

Perform OS Specific validation routines.

=over 4

=item Inputs:

=over 2

=item I<-mpdev_objs> => \@array

 Optional: An array reference of MPDev objects to validate

 Default: All MPDev objects discovered on the system

=back

=item Outputs:

=over 2

Returns two ARRAY references, ($errors, $warnings)

=back

=item Example:

 # validate all mpdev objects discovered
 $sanclient_obj->mpdevmgr()->os_specific_validation();
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub os_specific_validation {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    my ($errors, $warnings) = $self->SUPER::_os_specific_validation(%args);
    if ($self->error()) {
        $self->_set_err("OS Specific Validation failed");
        return ([], []);
    }

    return ($errors, $warnings);
}

=pod

=head2 B<get_filtered_devices>

Obtain MPDev objects that match the specified set of filters. If
a filter is not specified, then it will not be applied. Invoking this
method without any parameters is the equivalent of invoking
San::SanClient::MPDevMgr::get_device_objs()

=over 4

=item Inputs:

=over 2

=item I<-vservers> => \@array

 Optional: An array reference of C-Mode ONTAP vserver names. Only MPDev
 objects that belong to the vservers specified will pass this filter.
 This is mutually exclusive with the -controllers_7m filter.

=item I<-controllers_7m> => \@array

 Optional: An array reference of 7-Mode ONTAP controller names. Only MPDev
 objects that belong to the controller specified will pass this filter.
 This is mutually exclusive with the -vservers filter.

=item I<-ontap_home_nodes> => \@array

 Optional: An array reference of ONTAP controller names.
 Objects that have LUN backings that physically reside on any aggregate that belongs to the
 controllers specified will pass this filter.

=item I<-protocols> => \@array

 Optional: An array reference protocol types to match. Objects that have LUNs that match
 any of these protols will pass this filter.

=item I<-ontap_lun_ostype_regex> => regex

 Optional: A perl regular expression that specifies which ONTAP LUN Ostypes to filter by.

=item I<-ontap_lun_path_regex> => regex

 Optional: A perl regular expression that specifies which ONTAP LUN paths to filter by.

=item I<-ontap_lun_revision_regex> => regex

 Optional: A perl regular expression that specifies which ONTAP LUN revision to filter by.
 This provides the ability to filter LUNs by ONTAP revision, IE: 8200 vs 8100.

=item I<-ontap_mode> => scalar

 Optional: Specifies whether to filter by cmode or 7mode LUNs.
 Possible values: "any" | "cmode" | "7mode"
 Default: "any"

=item I<-device_vendor> => $scalar

 Optional: Find devices that match this SCSI Vendor field

=item I<-device_state> => scalar

 Optional: Specifies which device states to look for. By default, all devices are
 initialized to "free", unless the device is discovered to have no paths available.
 Possible values: "free" | "no_paths"
 Default: "" (all device states)

=item I<-device_count> => scalar

 Optional: Specifies how many devices to return after all the filters have been applied.
 All devices that match the filters will be returned if no count specified.

=item I<-device_index> => scalar

 Optional: Specifies what numerical index to seek the returned device list to.
 Default: 0

=item I<-reserve_devices> => scalar

 Optional: Specifies whether to set the device_state of all returned devices to 'not_free'
 Default: 0 (dont reserve)

=item I<-os_specific_data_filter> => \%hash

 Optional: A hash reference specification that describes which OS specific fields to match for.
 OS Specific fields are special fields that are not generic enough to be placed on the MPDev object.
 Each implementation of an OSMPDevMgr object defines its own set of os specific data. See the POD
 For each OS MPDevMgr object for details on which os_specific_data fields are available for that platform.
 Example: -os_specific_data_filter => {-datastore_label => "somedatastore_label",
                                       -datastore_uuid => "3492349234-45949354adfec"}

=back

=item Outputs:

=over 2

An array of San::SanClient::MPDevMgr::MPDev objects that matched the specified filters

=back

=item Example:

 # Find devices that match the following vservers
 my @mpdev_objs = $sanclient_obj->mpdevmgr()->get_filtered_devices(-vservers => ["myvserver1", "myvserver2"]);
 if ($sanclient_obj->mpdevmgr()->error()) {
     $Test->nlog($sanclient_obj->mpdevmgr()->errmsg());
 }

=back

=cut

sub get_filtered_devices {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    # parse the arguments
    my @vservers = (exists($args{-vservers}) and ref($args{-vservers}) eq "ARRAY") ?
                   @{$args{-vservers}} : ();
    my @controllers_7m = (exists($args{-controllers_7m}) and ref($args{-controllers_7m}) eq "ARRAY") ?
                         @{$args{-controllers_7m}} : ();
    my @ontap_home_nodes = (exists($args{-ontap_home_nodes}) and ref($args{-ontap_home_nodes}) eq "ARRAY") ?
                           @{$args{-ontap_home_nodes}} : ();
    my @protocols = (exists($args{-protocols}) and ref($args{-protocols}) eq "ARRAY") ?
                    @{$args{-protocols}} : ();
    my $ontap_lun_ostype_regex = exists($args{-ontap_lun_ostype_regex}) ? $args{-ontap_lun_ostype_regex} : undef;
    my $ontap_lun_path_regex = exists($args{-ontap_lun_path_regex}) ? $args{-ontap_lun_path_regex} : undef;
    my $ontap_lun_revision_regex = exists($args{-ontap_lun_revision_regex}) ? $args{-ontap_lun_revision_regex} : undef;
    my $ontap_mode = exists($args{-ontap_mode}) ? $args{-ontap_mode} : "any";
    my $reserve_devices = exists($args{-reserve_devices}) ? $args{-reserve_devices} : 0;
    my $device_state = exists($args{-device_state}) ? $args{-device_state} : "";
    my $device_vendor = exists($args{-device_vendor}) ? $args{-device_vendor} : undef;
    my $device_count = exists($args{-device_count}) ? $args{-device_count} : undef;
    my $device_index = exists($args{-device_index}) ? $args{-device_index} : 0;
    my $os_specific_data_filter = exists($args{-os_specific_data_filter}) ? $args{-os_specific_data_filter} : undef;

    # get all devices
    my @all_devices = $self->get_device_objs();
    my @filtered_device_list = @all_devices;

    # (1) filter by vservers if they are specified
    if (scalar(@vservers)) {
        @filtered_device_list = grep {
            my $vserver_name = $_->ontap_vserver();
            if (not defined($vserver_name)) {
                # this device is not attached to a vserver
                # filter it out
                0;
            } else {
                # if this is in the vserver list
                grep(/^$vserver_name$/, @vservers);
            }
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match vservers : '@vservers'");
            return (wantarray()) ? () : [];
        }
    }

    # (2) filter by controllers_7m if they are specified
    if (scalar(@controllers_7m)) {
        @filtered_device_list = grep {
            my $controller_7m = $_->ontap_controller_7m();
            if (not defined($controller_7m)) {
                # this device is not attached to a 7m controller
                # filter it out
                0;
            } else {
                # if this is in the 7m controller list
                grep(/^$controller_7m$/, @controllers_7m);
            }
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match controllers_7m : '@controllers_7m'");
            return (wantarray()) ? () : [];
        }
    }

    # (2a) filter by ontap home nodes if they are specified
    if (scalar(@ontap_home_nodes)) {
        @filtered_device_list = grep {
            my $ontap_home_node = $_->ontap_home_node();
            if (not defined($ontap_home_node)) {
                # this device does not have an ontap home node assigned
                0;
            } else {
                # check if this device lives on one of the ontap home nodes passed in
                grep(/^$ontap_home_node$/, @ontap_home_nodes);
            }
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match ontap_home_nodes : '@ontap_home_nodes'");
            return (wantarray()) ? () : [];
        }
    }

    # (3) filter by protocols if they are specified
    if (scalar(@protocols)) {
        @filtered_device_list = grep {
            my $protocol = $_->protocol();
            if (not defined($protocol)) {
                # this device does not have an assigned protocol
                0;
            } else {
                grep(/^$protocol$/i, @protocols);
            }
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match protocols : '@protocols'");
            return (wantarray()) ? () : [];
        }
    }

    # (4) filter by LUN OS Type if specified
    if (defined($ontap_lun_ostype_regex)) {
        @filtered_device_list = grep {
            my $lun_ostype = $_->ontap_lun_ostype();
            if (not defined($lun_ostype)) {
                # no os type defined on this device, skip
                0;
            } else {
                $lun_ostype =~ $ontap_lun_ostype_regex;
            }
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match ontap lun ostype: $ontap_lun_ostype_regex");
            return (wantarray()) ? () : [];
        }
    }

    # (5) filter by LUN Path if specified
    if (defined($ontap_lun_path_regex)) {
        @filtered_device_list = grep {
            my $lun_path = $_->ontap_lun_path();
            if (not defined($lun_path)) {
                # no lun path defined on this device, skip
                0;
            } else {
                $lun_path =~ $ontap_lun_path_regex;
            }
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match ontap lun path: $ontap_lun_path_regex");
            return (wantarray()) ? () : [];
        }
    }

    # (6) filter by LUN Revision if specified
    if (defined($ontap_lun_revision_regex)) {
        @filtered_device_list = grep {
            my $device_revision = $_->device_revision();
            if (not defined($device_revision)) {
                # no device revision defined on this device, skip
                0;
            } else {
                $device_revision =~ $ontap_lun_revision_regex;
            }
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match ontap lun revision: $ontap_lun_revision_regex");
            return (wantarray()) ? () : [];
        }
    }

    # (7) filter by ontap mode
    if (defined($ontap_mode) and $ontap_mode !~ /^any$/i) {
        @filtered_device_list = grep {
            $_->ontap_mode() =~ /^$ontap_mode$/i;
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match ontap mode: $ontap_mode");
            return (wantarray()) ? () : [];
        }
    }

    # (8) filter by device state
    if (defined($device_state) and $device_state ne "") {
        @filtered_device_list = grep {
            $_->device_state() =~ /^$device_state$/i;
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match device_state: $device_state");
            return (wantarray()) ? () : [];
        }
    }

    # (9) filter by devices that match the os specific data filters
    if (defined($os_specific_data_filter)) {
        foreach my $field (keys %{$os_specific_data_filter}) {
            @filtered_device_list = grep {
                if (defined($_->get_os_specific_data($field)) and
                    ($_->get_os_specific_data($field) eq $os_specific_data_filter->{$field})) {
                    1;
                } else {
                    0;
                }
            } @filtered_device_list;
        }
    }

    # (10) filter by device vendor
    if (defined($device_vendor) and $device_vendor ne "") {
        @filtered_device_list = grep {
            $_->device_vendor() =~ /^$device_vendor$/i;
        } @filtered_device_list;

        if (not scalar(@filtered_device_list)) {
            $self->_set_err("No devices match device_vendor: $device_vendor");
            return (wantarray()) ? () : [];
        }
    }

    # the next task is to make all the devices alternate from different
    # controllers so that all ontap nodes are equally represented
    @filtered_device_list = $self->_balance_by_ontap_node(-device_list => \@filtered_device_list,
                                                          -device_index => $device_index);

    # (final) only obtain exactly device count devices from the list
    if (defined($device_count)) {
        # determine how many devices need to be removed.
        my $devices_to_pop = scalar(@filtered_device_list) - $device_count;
        if ($devices_to_pop < 0) {
            $self->_set_err("There are not enough devices available " .
                            "to satisfy the requested device count " .
                            "of '$device_count'. There are only " . scalar(@filtered_device_list) .
                            " devices to choose from.");
            return (wantarray()) ? () : [];
        }
        # pop off $devices_to_pop devices from the list
        for (1..$devices_to_pop) {
            pop(@filtered_device_list);
        }
    }

    # if the user requested to reserve the devices,
    # mark all the filtered devices as not free
    if ($reserve_devices) {
        foreach my $mpdev_obj (@filtered_device_list) {
            # skip devices that have no paths
            next if ($mpdev_obj->device_state() eq "no_paths");

            $mpdev_obj->device_state("not_free");
        }
    }

    # return the final filtered list
    return (wantarray()) ? @filtered_device_list : \@filtered_device_list;
}

=pod

=head1 B<Public Accessor/Setter Methods>

=head2 B<set_os_specific_data(%args)>

Set OS specific data on the device manager.

=over 4

=item Inputs:

=over 2

=item I<key> => $data

 Generic {key => data} fields to store on the object.

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # set some generic key => value data on the object
 $mpdevmgr_obj->set_os_specific_data(key1 => "value1",
                                     key2 => "value2");
 if ($mpdevmgr_obj->error()) {
    # ...
 }

=back

=cut

sub set_os_specific_data {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    %args = &SanUtil::strip_dash(%args);

    foreach (keys %args) {
        $self->{os_specific_data}->{$_} = $args{$_};
    }
}

=pod

=head2 B<get_os_specific_data()>

Get OS specific data from this device manager.

=over 4

=item Inputs:

=over 2

=item I<$key>

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # get some generic key => value data on the object
 my $value = $mpdevmgr_obj->get_os_specific_data("some_key");
 if ($mpdev_obj->error()) {
    # ...
 }

=back

=cut

sub get_os_specific_data {
    my ($self, $key) = @_;

    @ISA = @{$self->{__ISA__}};

    if (defined($key)) {
        return $self->{os_specific_data}->{$key};
    } else {
        return $self->{os_specific_data};
    }
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _balance_by_ontap_node()
# Objective: Balance the given list of devices by ontap node
###############################################################################
sub _balance_by_ontap_node {
    my ($self, %args) = @_;

    my @device_list = (exists($args{-device_list}) and ref($args{-device_list}) eq "ARRAY") ?
                      @{$args{-device_list}} : ();
    my $device_index = exists($args{-device_index}) ? $args{-device_index} : 0;

    my %node_hash = ();

    foreach my $device (@device_list) {
        my $ontap_home_node = $device->ontap_home_node();
        if ($ontap_home_node ne "") {
            # add this device into that node's key array
            push(@{$node_hash{$device->ontap_home_node()}}, $device);
        } else {
            # add this device to the uncategorized array
            push(@{$node_hash{_unknown_}}, $device);
        }
    }

    # now balance out the device list so that the final list we return
    # to the caller has alternating devices from different nodes/controllers.
    my @balanced_device_list = ();
    while(scalar(keys %node_hash) > 0) {
        # for every node key, shift off one of the devices and push it into the final array
        foreach my $node_array (sort keys %node_hash) {
            push(@balanced_device_list, shift(@{$node_hash{$node_array}}));
            if (scalar(@{$node_hash{$node_array}}) == 0) {
                delete $node_hash{$node_array};
            }
        }
    }

    # now seek the list to the index defined by the device index parameter
    # this allows for the caller to balance the array on demand
    for (my $i = 0; $i < $device_index; $i++) {
        push(@balanced_device_list, shift(@balanced_device_list));
    }

    return @balanced_device_list;
}

###############################################################################
# Method:    _add_device()
# Objective: Register the specified device with the MPDevMgr
###############################################################################
sub _add_device {
    my ($self, %args) = @_;

    my $new_device = $args{-mp_device_obj};
    my $path_objs = exists($args{-mp_device_path_objs}) ? $args{-mp_device_path_objs} : undef;

    if (ref($new_device) ne "San::SanClient::MPDevMgr::MPDev") {
        $self->_set_err("Invalid device object -mp_device_obj, must " .
                        "be of type San::SanClient::MPDevMgr::MPDev");
        return;
    }

    # if a previous version of this mpdev object exists in the devices hash,
    if (exists($self->{devices}->{$new_device->mp_device_name()})) {
        my $old_device = $self->{devices}->{$new_device->mp_device_name()};

        # copy all lun relationship information from the old device into the new device
        $self->_copy_lun_relationship_data(-src_mpdev_obj  => $old_device,
                                           -dest_mpdev_obj => $new_device);

        # we need to obtain the MPDevPathStates from the old devices
        # and add them one by one into the new object
        # reverse the list so that we have the oldest path state first
        my @old_path_state_objs = reverse $old_device->get_path_states();
        foreach my $old_path_state_obj (@old_path_state_objs) {
            $new_device->_add_path_state(-mp_device_path_objs => [$old_path_state_obj->get_path_objs()]);
        }

        # change the device state of the old device to 'old_device_still_present'
        $old_device->device_state("old_device_still_present");

        # store the old device in the old devices hash
        $self->{old_devices}->{$old_device->mp_device_name()} = $old_device;
    }

    # register the device in mpdevmgr
    $self->{devices}->{$new_device->mp_device_name()} = $new_device;

    # if there is a new path state to add for this device, add it now
    if (defined($path_objs)) {
        $new_device->_add_path_state(-mp_device_path_objs => $path_objs);
    }

    return;
}

###############################################################################
# Method:    _remove_device()
# Objective: Unregister the specified device with the MPDevMgr
###############################################################################
sub _remove_device {
    my ($self, %args) = @_;

    if (ref($args{-mp_device_obj}) ne "San::SanClient::MPDevMgr::MPDev") {
        $self->_set_err("Invalid device object -mp_device_obj, must " .
                        "be of type San::SanClient::MPDevMgr::MPDev");
        return;
    }

    delete($self->{devices}->{$args{-mp_device_obj}->mp_device_name()});

    return;
}

###############################################################################
# Method:    _clear_devices()
# Objective: Unregister all devices with the MPDevMgr
###############################################################################
sub _clear_devices {
    my ($self) = @_;

    # copy the device hash to the old device hash
    $self->{old_devices} = $self->{devices};

    # delete the devices attribute
    delete($self->{devices});

    # initialize the devices hash
    $self->{devices} = {};

    return;
}

###############################################################################
# Method:    _mark_all_devices_pending_removal()
# Objective: Mark all devices in the MPDevMgr for pending_removal
###############################################################################
sub _mark_all_devices_pending_removal {
    my ($self) = @_;

    foreach my $mpdev_obj ($self->get_device_objs()) {
        $mpdev_obj->device_state("pending_removal");
    }

    # Clone all the devices into the old device hash
    $self->{old_devices} = Clone::clone($self->{devices});

    return;
}

###############################################################################
# Method:    _clear_all_devices_pending_removal()
# Objective: Clear all devices in the MPDevMgr for pending_removal
###############################################################################
sub _clear_all_devices_pending_removal {
    my ($self) = @_;

    # find all existing mpdev objs that are pending removal
    foreach my $mpdev_obj ($self->get_device_objs()) {
        if ($mpdev_obj->device_state() eq "pending_removal") {

            # if this device is present in the old devices hash, change
            # the old device state to "old_device_not_present
            my $old_dev_hash = $self->get_old_devices_hash();
            if (exists($old_dev_hash->{$mpdev_obj->mp_device_name()})) {
                my $old_mpdev_obj = $old_dev_hash->{$mpdev_obj->mp_device_name()};
                $old_mpdev_obj->device_state("old_device_not_present");
            }

            # remove this device from the MPDevMgr since it is no longer
            # present on the system
            $self->_remove_device(-mp_device_obj => $mpdev_obj);
        }
    }

    return;
}

###############################################################################
# Method:    _copy_lun_relationship_data()
# Objective: Copy lun relationship information from one device to another
###############################################################################
sub _copy_lun_relationship_data {
    my ($self, %args) = @_;

    my $src_mpdev_obj = $args{-src_mpdev_obj};
    my $dest_mpdev_obj = $args{-dest_mpdev_obj};

    # if the source does not have a lun relationship, abort now
    return if (not $src_mpdev_obj->has_lun_relationship());

    # these are the fields we need to clone
    my @fields = (
        'ontap_lun_name',
        'ontap_vol_name',
        'lun_relationship',
        'lun_obj',
        'ontap_vserver',
        'ontap_controller_7m',
        'ontap_home_node',
        'ontap_lun_path',
        'ontap_lun_ostype',
        'ontap_home_node_hapartner',
        'ontap_node_id',
        'ontap_qtree_name',
        'ontap_aggregate_name',
        'ontap_igroups'
    );

    # clone all the fields over
    foreach my $field (@fields) {
        $dest_mpdev_obj->{$field} = Clone::clone($src_mpdev_obj->{$field});
    }

    return;
}

################################################################################
# Subroutine Name: _convert_to_node_id()
# Objective:       Given a node name, convert to a node id
################################################################################
sub _convert_to_node_id {
    my ($self, %args) = @_;

    my $node_name = $args{-node_name};

    if (not exists($self->{_node_id_hash})) {
        # initialize the node id hash
        $self->{_node_id_hash}->{id_count} = 0;
    }

    # if this node does not have an id assigned to it yet,
    # generate one and  increment the id count
    if (not exists($self->{_node_id_hash}->{nodes}->{$node_name})) {
        $self->{_node_id_hash}->{nodes}->{$node_name} =
            $self->{_node_id_hash}->{id_count};
        $self->{_node_id_hash}->{id_count} += 1;
    }

    # return the id
    return $self->{_node_id_hash}->{nodes}->{$node_name};
}

1;
