use strict;
use warnings;

use File::Basename;
use San::SanClient::MPDevMgr::MPDevPath;

package San::SanClient::MPDevMgr::CXSMPDevMgr::CXSMPDev;

sub _new {
    my ($self, %args) = @_;

    return $self;
}

sub _refresh {
    my ($self, %args) = @_;

    my $sc_obj = $self->{sc_obj};

    if ($self->mp_type() =~ /dmmp/i) {
        $self->_refresh_dmmp_device(%args);
        return $self;
    } else {
        $self->_set_err("Invalid mp_type");
        return $self;
    }
}

sub _refresh_dmmp_device {
    my ($self) = @_;

    my $mp_device_name = $self->mp_device_name();
    my $sc_obj = $self->{sc_obj};

    # invoke the DMMP discovery script
    my $cmd = "perl /sbin/netapp_cxs_dmmp_discovery";
    my ($status, @output) = &ClientUtil::execute($sc_obj->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return;
    }

    # convert the output into device blocks
    @output = split(/\n\n/, join("\n", @output));

    my $found  = 0;

    # parse each device block
    foreach my $device_block (@output) {
        my @dev_block_split = split(/\n/, $device_block);

        my $device_vendor;
        my $device_model;
        my $lun_id;
        my $device_revision;
        my $device_scsi_level;
        my $device_size;
        my $protocol;
        my $protocol_path_counts;
        my $lun_paths;
        my $scsi_id;
        my $alua;

        foreach my $line (@dev_block_split) {
            # parse out the mp device name, device vendor, device model, device revision, scsi level
            if ($line =~ /^(360\S+)\s+LUN=(\S+)\s+REV=(\S+)\s+SCSI=(\S+)\s+size=(\S+)/) {
                if ($mp_device_name ne $1) {
                    next;
                }

                $found = 1;
                $mp_device_name = $1;
                $scsi_id = $1;
                $device_vendor = "NETAPP";
                $device_model = "LUN";
                $lun_id = $2;
                $device_revision = $3;
                $device_scsi_level = $4;
                $device_size = $5;
                next;
            }


            # parse the information for this path
            if (($line =~ /\d+:\d+:\d+:\d+\s+(\S+)\s+(\S+)\s+\[proto=(iscsi|fcp)\]\s+\[alua=(\S+),(\S+)\]\s+(\S+)\s+\<=\>\s+(\S+)/)
               and $found == 1) {
                my $raw_scsi_device_name = "/dev/" . $1;
                my $path_state = $2;
                my $this_paths_protocol = $3;
                $alua = $4;
                my $alua_state = uc($5);
                my $target_data = $6;
                my $initiator_data = $7;

                my $mp_dev_path_obj;

                # handle per protocol parsing
                if ($this_paths_protocol =~ /iscsi/i) {
                    # increment this lun's path count for this protocol
                    $protocol_path_counts->{iscsi}++;

                    # parse the iscsi target information from the data string
                    my $target_iscsi_nodename;
                    my $target_ip_address;
                    my $target_ip_port;
                    my $target_tpgt;
                    if ($target_data =~ /^(\S+?),(\S+?):(\d+?),(\d+)/) {
                        $target_iscsi_nodename = $1;
                        $target_ip_address = $2;
                        $target_ip_port = $3;
                        $target_tpgt = $4;
                    }

                    # the initiator data should be just the nodename
                    my $initiator_iscsi_nodename = $initiator_data;

                    # create the MPDevPath object
                    $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $raw_scsi_device_name,
                                                                               -lun_id => $lun_id,
                                                                               -path_state => $path_state,
                                                                               -initiator_iqn => $initiator_iscsi_nodename,
                                                                               -target_iqn => $target_iscsi_nodename,
                                                                               -target_iscsi_alias => "",
                                                                               -target_iscsi_session => "",
                                                                               -target_ip_address => $target_ip_address,
                                                                               -target_ip_port => $target_ip_port,
                                                                               -tpgt => $target_tpgt,
                                                                               -alua_state => $alua_state,
                                                                               -protocol => "iscsi");
                } elsif ($this_paths_protocol =~ /fcp/i) {
                    # increment this lun's path count for this protocol
                    $protocol_path_counts->{fcp}++;

                    # parse the target and initiator wwnns and wwpns
                    my $target_fcp_nodename;
                    my $target_fcp_portname;
                    if ($target_data =~ /^(\S+?),(\S+)$/) {
                        $target_fcp_nodename = $1;
                        $target_fcp_portname = $2;
                    }
                    my $initiator_fcp_nodename;
                    my $initiator_fcp_portname;
                    if ($initiator_data =~ /^(\S+?),(\S+)$/) {
                        $initiator_fcp_nodename = $1;
                        $initiator_fcp_portname = $2;
                    }

                    # create the MPDevPath object
                    $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $raw_scsi_device_name,
                                                                               -lun_id => $lun_id,
                                                                               -path_state => $path_state,
                                                                               -initiator_wwnn => $initiator_fcp_nodename,
                                                                               -initiator_wwpn => $initiator_fcp_portname,
                                                                               -target_wwnn => $target_fcp_nodename,
                                                                               -target_wwpn => $target_fcp_portname,
                                                                               -alua_state => $alua_state,
                                                                               -protocol => "fcp");

                }

                $protocol = $this_paths_protocol;
                $lun_paths->{$raw_scsi_device_name} = $mp_dev_path_obj;
                last;
            }
        }

        # note that the protocol for this mp device is mixed if it contains more
        # than one protocol
        $protocol = "mixed" if (scalar(keys %{$protocol_path_counts}) > 1);

        if ($alua =~ /0/) {
            $alua = "disabled";
        } elsif ($alua =~ /1/) {
            $alua = "enabled";
        } else {
            $alua = "unknown";
        }

        # assign all the devices
        $self->device_size($device_size);
        $self->protocol($protocol);
        $self->mp_device_path("/dev/mapper/" . $mp_device_name);
        $self->device_vendor($device_vendor);
        $self->device_model($device_model);
        $self->device_revision($device_revision);
        $self->device_scsi_level($device_scsi_level);
        $self->lun_id($lun_id);
        $self->alua($alua);

        # add the newly discovered path state for this device
        # convert hash to array of path objects first
        my @path_objs = values %{$lun_paths};
        $self->_add_path_state(-mp_device_path_objs => \@path_objs);

        if ($found == 1) {
            last;
        }
    }

    return $self;
}

sub _refresh_paths {
    my ($self, %args) = @_;

    # obtain the current path state, for each path in the current path state
    # clone each path object, and refresh the alua state and path state of
    # the path. after all paths have been refreshed, create a new path state.

    my $curr_path_state_obj = $self->get_current_path_state();
    my @curr_path_objs = $curr_path_state_obj->get_path_objs();
    my @new_path_objs;
    my $mp_device_name = $self->mp_device_name();
    my $sc_obj = $self->{sc_obj};

    require San::ScsiTool;
    require San::ScsiTool::ScsiUtil;

    my $scsi_tool_obj = ScsiTool->create_scsi_tool(-client => $sc_obj,
                                                   -tool   => "spt");
    if ($scsi_tool_obj->error()) {
        $self->_set_err($scsi_tool_obj->errmsg());
        return $self;
    }

    my $scsi_util_obj = San::ScsiTool::ScsiUtil->new(-scsi_tool_obj => $scsi_tool_obj);
    if ($scsi_util_obj->error()) {
        $self->_set_err($scsi_util_obj->errmsg());
        return $self;
    }

    foreach my $path_obj (@curr_path_objs) {
        # build a new path object with current path info
        my $new_path_obj = new San::SanClient::MPDevMgr::MPDevPath(%{$path_obj});

        # refresh the alua state and path state for this path
        my $raw_scsi_device_name = $new_path_obj->path_name();
        require File::Basename;
        my $raw_scsi_device_basename = &File::Basename::basename($raw_scsi_device_name);

        my $cmd = "cat /sys/block/$raw_scsi_device_basename/device/state";
        my ($status, @output) = &ClientUtil::execute($sc_obj->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to execute the command ($cmd)");
            return $self;
        }

        my ($path_state, $alua_state);

        if ($output[0] =~ /running/i) {
            $path_state = "active";
        } else {
            $path_state = "dead";
        }

        # verify that this file exists
        # /x/eng/localtest/arch/i386-redhat-linux8/bin/
        &ClientUtil::execute($sc_obj->{conn}, "touch /x/eng/localtest/arch/");

        # determine the alua state of this path
        my $ret_val = $scsi_tool_obj->test_scsi_cmd(-device => $raw_scsi_device_name,
                                                    -cdb_name => "inquiry_page_C1h");
        if (ref($ret_val) eq "HASH" and exists($ret_val->{SCSI_DATA})) {
            my $s_data = $scsi_util_obj->scsi_bytes_to_strings(-scsi_data_ref => $ret_val->{SCSI_DATA},
                                                               -scsi_cmd_name => "inquiry_page_C1h");
            if (ref($s_data) eq "HASH") {
                if ($s_data->{PROXY} == 1) {
                    $alua_state = "ANO";
                } elsif ($s_data->{PROXY} == 0) {
                    $alua_state = "AO";
                }
            } else {
                $alua_state = "DOWN";
            }
        } else {
            $alua_state = "UNKNOWN";
        }

        # update the states
        $new_path_obj->path_state($path_state);
        $new_path_obj->alua_state($alua_state);

        push(@new_path_objs, $new_path_obj);
    }

    $self->_add_path_state(-mp_device_path_objs => \@new_path_objs);

    return;
}

sub _load {
    my ($self) = @_;

    # update the current state of device into MPDev
    # determine the device state whether it is free or already in use

    my ($cmd, $status, @output);
    my $sc_obj = $self->{sc_obj};

    my $mp_device = $self->mp_device_name();

    # get list of SRs and determine whether SR exist for this device
    $cmd = "xe sr-list params=uuid,name-label,sm-config";
    ($status, @output) = &ClientUtil::execute($sc_obj->{conn}, $cmd);
    if ($status != 0) {
        return;
    }

    if (not grep(/$mp_device/, @output)) {
        # device is definitely free since SR doesn't exist
        $self->device_state("free");
        $self->set_os_specific_data(-sr_label => "",
                                    -sr_uuid  => "");
    } else {
        my $sr_uuid = "";
        my $sr_name = "";
        foreach my $line (@output) {
            if ($line =~ /uuid.*:\s*(\S+)/i) {
                $sr_uuid = $1;
            } elsif ($line =~ /name-label.*:(.+)/i) {
                $sr_name = $1;
                $sr_name =~ s/^\s*//;
            } elsif ($line =~ /sm-config.*:(.+)/i) {
                if ($1 =~ /scsi-$mp_device/i) {
                    # mark this device state as used
                    $self->device_state("used");

                    # store os specific data about this datastore
                    $self->set_os_specific_data(-sr_label => $sr_name,
                                                -sr_uuid  => $sr_uuid);
                    return;
                }
                $sr_uuid = "";
                $sr_name = "";
            }
        }
    }

    return;
}

sub _mount {
    my ($self, %args) = @_;

    # NOOP

    return;
}

sub _unmount {
    my ($self) = @_;

    # NOOP

    return;
}

sub _mkfs {
    my ($self, %args) = @_;

    # NOOP

    return;
}

sub _destroy_partition {
    my ($self) = @_;

    # NOOP

    return;
}

sub _create_partition {
    my ($self) = @_;

    # NOOP

    return;
}

sub _online_device {
    my ($self, %args) = @_;

    # NOOP

    return;
}

sub _offline_device {
    my ($self, %args) = @_;

    # NOOP

    return;
}

1;
