#
# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary MPDevMgr base class for ESX
## @author hrivera@netapp.com
## @status private
## @pod ../MPDevMgr.pm

package San::SanClient::MPDevMgr::ESXMPDevMgr;

use strict;
use warnings;
use Data::Dumper;
use San::SanClient::MPDevMgr::MPDev;
use San::SanClient::MPDevMgr::MPDevPath;
use San::SanUtil;

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: Create an MPDevMgr instance
###############################################################################
sub _new {
    my ($self, %args) = @_;

    return $self;
}

###############################################################################
# Method:    load_all_devices()
# Objective: Load in the current state of all device's filesystem/partition
###############################################################################
sub _load_all_devices {
    my ($self) = @_;

    my $sc_obj = $self->{sc_obj};

    my @mp_dev_objs = $self->get_device_objs();
    if (not scalar(@mp_dev_objs)) {
        return;
    }

    # load os specific information about this device
    my $cmd = "esxcfg-scsidevs --vmfs";
    my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to load VMFS filesystem states");
        return;
    }

    foreach my $mpdev_obj (@mp_dev_objs) {
        my $mp_device_name = $mpdev_obj->mp_device_name();

        my ($line) = grep(/^\s*${mp_device_name}:/, @output);
        if (not defined($line)) {
            # there is no data to load for this device, set it to null and go to the next
            # device in the list
            $mpdev_obj->set_os_specific_data(-datastore_label => "",
                                             -datastore_uuid => "",
                                             -datastore_path => "",
                                             -datastore_partition_path => "");
            # update some fields on this object, stating that it is not mounted,
            # and doesnt have a filesystem
            $mpdev_obj->is_partitioned(0);
            $mpdev_obj->is_mounted(0);
            $mpdev_obj->has_filesystem(0);
            $mpdev_obj->fs_type("");
            $mpdev_obj->mount_point("");
            $mpdev_obj->mp_device_partition_list([]);

            next;
        }

        # parse out the data for this device
        if ($line =~ /^\s*(${mp_device_name}:\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.+)$/) {
            my $mp_partition_name = $1;
            my $mp_partition_path = $2;
            my $datastore_uuid = $3;
            my $unknown = $4;
            my $datastore_label = $5;

            # store os specific data about this datastore
            $mpdev_obj->set_os_specific_data(-datastore_label => $datastore_label,
                                             -datastore_uuid => $datastore_uuid,
                                             -datastore_path => "/vmfs/volumes/${datastore_label}",
                                             -datastore_partition_path => $mp_partition_path);

            $mpdev_obj->is_partitioned(1);
            $mpdev_obj->is_mounted(1);
            $mpdev_obj->has_filesystem(1);
            $mpdev_obj->fs_type("vmfs");
            $mpdev_obj->mount_point("/vmfs/volumes/${datastore_label}");
            $mpdev_obj->mp_device_partition_list([$mp_partition_path]);
        } else {
            $self->_set_err("Failed to parse output properly during load of $mp_device_name");
            return;
        }
    }

    return;
}

###############################################################################
# Method:    offline_devices()
# Objective: Offline a group of devices specified
###############################################################################
sub _offline_devices {
    my ($self, %args) = @_;

    my @mpdev_objs = exists($args{-mpdev_objs}) ? @{$args{-mpdev_objs}}: ();
    if (not scalar(@mpdev_objs)) {
        $self->_set_err("No MPDev objects were specified");
        return;
    }

    my $sc_obj = $self->{sc_obj};

    if ($sc_obj->get_os_version() =~ /^4\./) {
        my $corestorage_ns = $sc_obj->{esxcli_corestorage_namespace};

        # list out the claim rules
        my $cmd = "esxcli ${corestorage_ns} claimrule list";
        my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                          -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Failed to list out claimrules");
            return;
        }

        # parse out the claim rules, we need to generate a list of rule numbers
        # that we can use (between 102 and 199)
        my @used_rules = map {
            if (/^\S+\s+(\d+)/) {
                $1;
            }
        } @output;

        # find all MASK_PATH location rules and add them to @rules_to_remove
        # these are rules we may have created in the past but never successfully removed
        my @rules_to_remove;
        foreach my $line (@output) {
            if ($line =~ /^MP\s+(\d+)\s+file\s+location\s+MASK_PATH/i) {
                push(@rules_to_remove, $1);
            }
        }

        my $starting_rule = 1000;
        my $last_rule = 30000;

        # there is a caveat with VxDMP, we have to find the NETAPP rule and move it to a higher rule (50000)
        # check the first mpdev obj in the list, if it is VxDMP perform this logic
        if ($mpdev_objs[0]->mp_type() =~ /VxDMP/) {
            foreach my $line (@output) {
                if ($line =~ /^MP\s+(\d+)\s+file\s+vendor\s+VxDMP\s+vendor=NETAPP/) {
                    my $netapp_rule = $1;
                    if ($netapp_rule ne "50000") {
                        # perform the move
                        $cmd = "esxcli ${corestorage_ns} claimrule move --rule $netapp_rule --new-rule 50000";
                        ($status) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                              -include_host_name => 1);
                        if ($status > 0) {
                            $self->_set_err("Failed to move the NETAPP VxDMP claimrule from $netapp_rule to 50000");
                            return;
                        }
                        # load in the new rules
                        $cmd = "esxcli ${corestorage_ns} claimrule load";
                        ($status) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                              -include_host_name => 1);
                        if ($status > 0) {
                            $self->_set_err("Could not load claim rules");
                            return;
                        }
                        last;
                    }
                }
            }
        }

        # foreach device, create a claim rule using the next available rule in the unused_rules list
        foreach my $mpdev_obj (@mpdev_objs) {
            # keep incrementing the starting rule until we find one that is unused
            while (grep(/^$starting_rule$/, @used_rules)) {
                $starting_rule++;
            }
            # if there are no rules left to use, bail out
            if ($starting_rule > $last_rule) {
                # issue a warning
                &SanUtil::nate_log('-comment', "There are no more claimrules left to use.");
                last;
            }

            my $mp_device_name = $mpdev_obj->mp_device_name();

            # generate a claimrule to mask away this device
            my @path_objs = $mpdev_obj->get_current_path_state()->get_path_objs();
            if (not scalar(@path_objs)) {
                $self->_set_err("There are no paths to $mp_device_name");
                return;
            }

            my %seen_adapters;

            foreach my $path_obj (@path_objs) {
                my $path_name = $path_obj->path_name();
                if ($path_name =~ /^(vmhba\d+):C(\d+):T(\d+):L(\d+)/) {
                    my $adapter = $1;
                    my $controller = $2;
                    my $target = $3;
                    my $lun = $4;

                    # we only need one rule per lun from each adapter that is added to the seen_adapters hash
                    next if (exists($seen_adapters{$adapter}));

                    # keep incrementing the starting rule until we find one that is unused
                    while (grep(/^$starting_rule$/, @used_rules)) {
                        $starting_rule++;
                    }
                    # if there are no rules left to use, bail out
                    if ($starting_rule > $last_rule) {
                        last;
                    }
                    push(@rules_to_remove, $starting_rule);

                    # mask out the adapter target and lun, controller represents multipath paths to this iscsi device
                    if ($path_obj->protocol() =~ /iscsi/i) {
                        # we need one claimrule per LUN for iSCSI
                        $seen_adapters{$adapter}++;
                        $cmd = "esxcli ${corestorage_ns} claimrule add --rule $starting_rule -P MASK_PATH -t location -A $adapter -T $target -L $lun";
                    } elsif ($path_obj->protocol() =~ /fcp/i) {
                        # we need one claimrule per path for FCP
                        $cmd = "esxcli ${corestorage_ns} claimrule add --rule $starting_rule -P MASK_PATH -t location -A $adapter -C $controller -T $target -L $lun";
                    } else {
                        # error out
                        $self->_set_err("Could not create claimrule for path '$path_name', it has an invalid protocol '" . $path_obj->protocol() . "'");
                        return;
                    }

                    ($status) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                          -include_host_name => 1);
                    if ($status > 0) {
                        $self->_set_err("Could not add claim rule for $mp_device_name");
                        return;
                    }

                    # increment the starting rule
                    $starting_rule++;
                }
            }
        }

        # load the claimrule list
        $cmd = "esxcli ${corestorage_ns} claimrule load";
        ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                       -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Could not load claim rules");
            return;
        }

        # foreach device, offline them
        foreach my $mpdev_obj (@mpdev_objs) {
            $mpdev_obj->offline_device();
            if ($mpdev_obj->error()) {
                $self->_set_err("Failed to offline device", $mpdev_obj->errmsg());
                return;
            }
        }

        # now, remove all the rules we added
        foreach my $rule (@rules_to_remove) {
            $cmd = "esxcli ${corestorage_ns} claimrule delete --rule $rule";
            ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                           -include_host_name => 1);
            if ($status > 0) {
                $self->_set_err("Could not remove claim rule $rule");
                return;
            }
        }
    }

    return;
}

###############################################################################
# Method:    refresh()
# Objective: Rediscover all Multipathed LUNs on the system and populate the
#            Device Manager
###############################################################################
sub _refresh {
    my ($self, %args) = @_;

    my $sc_obj = $self->{sc_obj};

    my ($cmd, $status, @scsidevs_list, @scsidevs_verbose_list,
        @esxcfg_mpath_list, @esxcli_nmp_device_list, @esxcli_nmp_path_list);
    my $corestorage_ns = $sc_obj->{esxcli_corestorage_namespace};

    # collect list of NetApp LUNs
    # esxcli storage core device list (ESXi5 and later)
    # esxcli corestorage device list (ESX4)

    # list out all devices
    # esxcli storage core device list (ESXi5 and later)
    # esxcli corestorage device list (ESX4)
    if ($sc_obj->get_os_version() =~ /^4\./) {
        $cmd = "esxcfg-scsidevs -l";
        ($status, @scsidevs_verbose_list) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                                      -include_host_name => 1);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $self;
        }
            # separate all the "chunks" to make parsing easier
        @scsidevs_verbose_list = map {
            my $line = $_;
            if ($line =~ /^\S/) {
                "\n$line";
            } else {
                "$line";
            }
        } @scsidevs_verbose_list;
        @scsidevs_verbose_list = split(/\n\n/, join("\n", @scsidevs_verbose_list));

        # we may have injected a newline at the beginning, remove it
        foreach (@scsidevs_verbose_list) {
            s/^\s*\n+//;
        }
    } else {
        $cmd = "esxcli ${corestorage_ns} device list";
        ($status, @scsidevs_verbose_list) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                                      -include_host_name => 1);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $self;
        }
        @scsidevs_verbose_list = split(/\n\n/, join("\n", @scsidevs_verbose_list));
    }


    # list out all paths to each device
    # esxcli storage core path list (ESXi5 and later)
    # esxcfg-mpath -b (ESX4)
    if ($sc_obj->get_os_version() =~ /^4\./) {
        $cmd = "esxcfg-mpath -b";
    } else {
        $cmd = "esxcli storage core path list";
    }
    ($status, @esxcfg_mpath_list) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                              -include_host_name => 1);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return $self;
    }
    @esxcfg_mpath_list = split(/\n\n/, join("\n", @esxcfg_mpath_list));

    # list out all device data
    my $nmp_namespace = $sc_obj->{esxcli_nmp_namespace};
    $cmd = "esxcli ${nmp_namespace} device list";
    ($status, @esxcli_nmp_device_list) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                                   -include_host_name => 1);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return $self;
    }
    @esxcli_nmp_device_list = split(/\n\n/, join("\n", @esxcli_nmp_device_list));

    # list out detailed path information for each path
    $cmd = "esxcli ${nmp_namespace} path list";
    ($status, @esxcli_nmp_path_list) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                                 -include_host_name => 1);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return $self;
    }
    @esxcli_nmp_path_list = split(/\n\n/, join("\n", @esxcli_nmp_path_list));

    my @devices;

    # iterate through each device
    foreach my $scsidev_block (@scsidevs_verbose_list) {
        if ($scsidev_block !~ /NETAPP/) {
            next;
        }

        my $naa_device_name;
        my $raw_scsi_device_name;
        my $device_size;
        my $mp_type;
        my $protocol;
        my $protocol_path_counts;
        my $naa_device_path;
        my $device_vendor;
        my $device_model;
        my $device_revision = "";
        my $scsi_level = "";
        my $lun_id;
        my $is_offline = 0;
        my $vaai_status = "";

        # discovery inquiry information for this device
        foreach my $scsidev_line (split(/\n/, $scsidev_block)) {
            if ($scsidev_line =~ /^(\S+)/) {
                $naa_device_name = $1;
            } elsif ($scsidev_line =~ /Display Name:.+iSCSI Disk/) {
                $protocol = "iscsi";
            } elsif ($scsidev_line =~ /Display Name:.+(Fibre Channel Disk|Fibre Channel Over Ethernet Disk)/) {
                $protocol = "fcp";
            } elsif ($scsidev_line =~ /Devfs Path:\s+(\S+)/) {
                $naa_device_path = $1;
                $raw_scsi_device_name = $naa_device_path;
            } elsif ($scsidev_line =~ /Multipath Plugin:\s+(.+)/) {
                $mp_type = $1;
            } elsif ($scsidev_line =~ /Plugin:\s+(.+)/) {
                $mp_type = $1;
            } elsif ($scsidev_line =~ /^\s*Size:\s+(\S+)/) {
                $device_size = $1 . "MB";
            } elsif ($scsidev_line =~ /Vendor:\s+(.+?)\s+Model:\s+(.+?)\s+Revis:\s+(\S+)/) {
                $device_vendor = $1;
                $device_model = $2;
                $device_revision = $3;
            } elsif ($scsidev_line =~ /Vendor:\s+(.+)/) {
                $device_vendor = $1;
                # remove trailing whitespace
                $device_vendor =~ s/\s+$//;
            } elsif ($scsidev_line =~ /Model:\s+(.+)/) {
                $device_model = $1;
                # remove trailing whitespace
                $device_model =~ s/\s+$//;
            } elsif ($scsidev_line =~ /Revision:\s+(.+)/) {
                $device_revision = $1;
            } elsif ($scsidev_line =~ /SCSI Level:\s+(\S+)/) {
                $scsi_level = $1;
            } elsif ($scsidev_line =~ /VAAI Status:\s+(.+)/) {
                $vaai_status = $1;
            }
        }

        # if the naa_device_path is undefined, then this device is offline "detached"
        if (not defined($naa_device_path)) {
            # build the naa device path manually
            $naa_device_path = "/vmfs/devices/disks/${naa_device_name}";
            $raw_scsi_device_name = $naa_device_path;
            $is_offline = 1;
        }

        # discover the lun paths to this device
        my $lun_paths;

        #naa.60a980004334652f4f5a4d7a78527734 : NETAPP Fibre Channel Disk (naa.60a980004334652f4f5a4d7a78527734)
        # vmhba1:C0:T0:L7 LUN:7 state:active fc Adapter: WWNN: 20:00:00:e0:8b:9c:eb:c1 WWPN: 21:00:00:e0:8b:9c:eb:c1  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:82:96:97:d7:94
        # vmhba1:C0:T1:L7 LUN:7 state:active fc Adapter: WWNN: 20:00:00:e0:8b:9c:eb:c1 WWPN: 21:00:00:e0:8b:9c:eb:c1  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:82:86:97:d7:94
        # vmhba2:C0:T0:L7 LUN:7 state:active fc Adapter: WWNN: 20:01:00:e0:8b:bc:eb:c1 WWPN: 21:01:00:e0:8b:bc:eb:c1  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:81:96:97:d7:94
        # vmhba2:C0:T1:L7 LUN:7 state:active fc Adapter: WWNN: 20:01:00:e0:8b:bc:eb:c1 WWPN: 21:01:00:e0:8b:bc:eb:c1  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:81:86:97:d7:94

        #naa.60a980004334652f4f5a4e31414b716a : NETAPP Fibre Channel Disk (naa.60a980004334652f4f5a4e31414b716a)
        # vmhba1:C0:T1:L14 LUN:14 state:active fc Adapter: WWNN: 20:00:00:e0:8b:9c:eb:c1 WWPN: 21:00:00:e0:8b:9c:eb:c1  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:82:86:97:d7:94
        # vmhba1:C0:T0:L14 LUN:14 state:active fc Adapter: WWNN: 20:00:00:e0:8b:9c:eb:c1 WWPN: 21:00:00:e0:8b:9c:eb:c1  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:82:96:97:d7:94
        # vmhba2:C0:T1:L14 LUN:14 state:active fc Adapter: WWNN: 20:01:00:e0:8b:bc:eb:c1 WWPN: 21:01:00:e0:8b:bc:eb:c1  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:81:86:97:d7:94
        # vmhba2:C0:T0:L14 LUN:14 state:active fc Adapter: WWNN: 20:01:00:e0:8b:bc:eb:c1 WWPN: 21:01:00:e0:8b:bc:eb:c1  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:81:96:97:d7:94

        foreach my $esxcfg_mpath_block (@esxcfg_mpath_list) {
            if ($sc_obj->get_os_version() =~ /^4\./) {
                # ESX4
                next if ($esxcfg_mpath_block !~ /^\s*${naa_device_name}\s+:\s+NETAPP/);

                foreach my $esxcfg_mpath_block (split(/\n/, $esxcfg_mpath_block)) {
                    # parse an FCP path
                    if ($esxcfg_mpath_block =~ /^\s*(vmhba\S+)\s+LUN:(\d+)\s+state:(\S+)\s+fco?e?\s+Adapter:?\s+(WWNN:\s+\S+\s+WWPN:\s+\S+|Unavailable)\s+Target:?\s+(WWNN:\s+\S+\s+WWPN:\s+\S+|Unavailable)/) {
                        my $path_name = $1;
                        $lun_id = $2;
                        my $path_state = $3;
                        my $adapter = $4;
                        my $target = $5;

                        my ($path_initiator_wwnn) = ($adapter =~ /WWNN:\s+(\S+)/);
                        $path_initiator_wwnn = "Unavailable" if (not defined($path_initiator_wwnn));
                        my ($path_initiator_wwpn) = ($adapter =~ /WWPN:\s+(\S+)/);
                        $path_initiator_wwpn = "Unavailable" if (not defined($path_initiator_wwpn));
                        my ($path_target_wwnn)    = ($target  =~ /WWNN:\s+(\S+)/);
                        $path_target_wwnn    = "Unavailable" if (not defined($path_target_wwnn));
                        my ($path_target_wwpn)    = ($target  =~ /WWPN:\s+(\S+)/);
                        $path_target_wwpn    = "Unavailable" if (not defined($path_target_wwpn));


                        # create a path object
                        my $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $path_name,
                                                                                      -lun_id => $lun_id,
                                                                                      -path_state => $path_state,
                                                                                      -initiator_wwnn => $path_initiator_wwnn,
                                                                                      -initiator_wwpn => $path_initiator_wwpn,
                                                                                      -target_wwnn => $path_target_wwnn,
                                                                                      -target_wwpn => $path_target_wwpn,
                                                                                      -protocol => "fcp");
                        $protocol_path_counts->{fcp}++;
                        $lun_paths->{$path_name} = $mp_dev_path_obj;

                    } elsif ($esxcfg_mpath_block =~ /^\s*(vmhba\S+)\s+LUN:(\d+)\s+state:(\S+).+?iscsi\s+Adapter:\s+(\S+)\s+Target:\s+(?:IQN=)?(\S+)/) {
                        # parse an iSCSI path
                        my $path_name = $1;
                        $lun_id = $2;
                        my $path_state = $3;
                        my $path_initiator_iqn = $4;
                        my $path_target_iqn = $5;
                        my $iscsi_target_alias;
                        my $iscsi_session;
                        my $tpgt;

                        # if the target iqn is unavailable, this path is probably dead
                        if ($path_target_iqn =~ /Unavailable/i) {
                            $path_target_iqn = "";
                            $path_initiator_iqn = "";
                            $iscsi_target_alias = "";
                            $iscsi_session = "";
                            $tpgt = "";
                        } elsif ($esxcfg_mpath_block =~ /Alias=(\S+)?\s+Session=(\S+)\s+PortalTag=(\S+)/) {
                            $iscsi_target_alias = $1;
                            $iscsi_session = $2;
                            $tpgt = $3;
                        }

                        # create a path object
                        my $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $path_name,
                                                                                      -lun_id => $lun_id,
                                                                                      -path_state => $path_state,
                                                                                      -initiator_iqn => $path_initiator_iqn,
                                                                                      -target_iqn => $path_target_iqn,
                                                                                      -target_iscsi_alias => $iscsi_target_alias,
                                                                                      -target_iscsi_session => $iscsi_session,
                                                                                      -target_ip_address => "",
                                                                                      -target_ip_port => "",
                                                                                      -tpgt => $tpgt,
                                                                                      -protocol => "iscsi");

                        $protocol_path_counts->{iscsi}++;
                        $lun_paths->{$path_name} = $mp_dev_path_obj;
                    }
                }
            } else {
                # ESXi5 or later

                # skip device blocks that don't match the current device
                next if ($esxcfg_mpath_block !~ /^\s*Device:\s+${naa_device_name}/m);

                # handle fcp paths
                if ($esxcfg_mpath_block =~ /Transport: fc/i) {

                    my $path_name;
                    $lun_id = "";
                    my $path_state;
                    my $path_initiator_wwnn;
                    my $path_initiator_wwpn;
                    my $path_target_wwnn;
                    my $path_target_wwpn;

                    foreach my $line (split("\n", $esxcfg_mpath_block)) {
                        if ($line =~ /Runtime Name:\s*(\S+)/) {
                            $path_name = $1;
                        } elsif ($line =~ /LUN:\s*(\S+)/) {
                            $lun_id = $1;
                        } elsif ($line =~ /State:\s*(\S+)/) {
                            $path_state = $1;
                        } elsif ($line =~ /Adapter Transport Details: WWNN: (\S+) WWPN: (\S+)/) {
                            $path_initiator_wwnn = $1;
                            $path_initiator_wwpn = $2;
                        } elsif ($line =~ /Target Transport Details: WWNN: (\S+) WWPN: (\S+)/) {
                            $path_target_wwnn = $1;
                            $path_target_wwpn = $2;
                        }
                    }

                    $path_initiator_wwnn = "Unavailable" if (not defined($path_initiator_wwnn));
                    $path_initiator_wwpn = "Unavailable" if (not defined($path_initiator_wwpn));
                    $path_target_wwnn    = "Unavailable" if (not defined($path_target_wwnn));
                    $path_target_wwpn    = "Unavailable" if (not defined($path_target_wwpn));

                    # create a path object
                    my $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $path_name,
                                                                                  -lun_id => $lun_id,
                                                                                  -path_state => $path_state,
                                                                                  -initiator_wwnn => $path_initiator_wwnn,
                                                                                  -initiator_wwpn => $path_initiator_wwpn,
                                                                                  -target_wwnn => $path_target_wwnn,
                                                                                  -target_wwpn => $path_target_wwpn,
                                                                                  -protocol => "fcp");
                    $protocol_path_counts->{fcp}++;
                    $lun_paths->{$path_name} = $mp_dev_path_obj;
                } elsif ($esxcfg_mpath_block =~ /Transport: iscsi/i) {
                    # parse an iSCSI path
                    my $path_name;
                    $lun_id = "";
                    my $path_state;
                    my $path_initiator_iqn;
                    my $path_target_iqn;
                    my $iscsi_target_alias;
                    my $iscsi_session;
                    my $tpgt;

                    foreach my $line (split("\n", $esxcfg_mpath_block)) {
                        if ($line =~ /Runtime Name:\s*(\S+)/) {
                            $path_name = $1;
                        } elsif ($line =~ /LUN:\s*(\S+)/) {
                            $lun_id = $1;
                        } elsif ($line =~ /State:\s*(\S+)/) {
                            $path_state = $1;
                        } elsif ($line =~ /Adapter Transport Details: Unavailable/i) {
                            $path_initiator_iqn = "";
                        } elsif ($line =~ /Adapter Transport Details: (\S+)/) {
                            $path_initiator_iqn = $1;
                        } elsif ($line =~ /Target Transport Details: Unavailable/i) {
                            $path_target_iqn = "";
                            $iscsi_target_alias = "";
                            $iscsi_session = "";
                            $tpgt = "";
                        } elsif ($line =~ /Target Transport Details:\s+IQN=(\S+)\s+Alias=(\S+)?\s+Session=(\S+)\s+PortalTag=(\S+)/) {
                            $path_target_iqn = $1;
                            $iscsi_target_alias = $2;
                            $iscsi_session = $3;
                            $tpgt = $4;
                        }
                    }

                    # create a path object
                    my $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $path_name,
                                                                                  -lun_id => $lun_id,
                                                                                  -path_state => $path_state,
                                                                                  -initiator_iqn => $path_initiator_iqn,
                                                                                  -target_iqn => $path_target_iqn,
                                                                                  -target_iscsi_alias => $iscsi_target_alias,
                                                                                  -target_iscsi_session => $iscsi_session,
                                                                                  -target_ip_address => "",
                                                                                  -target_ip_port => "",
                                                                                  -tpgt => $tpgt,
                                                                                  -protocol => "iscsi");
                    $protocol_path_counts->{iscsi}++;
                    $lun_paths->{$path_name} = $mp_dev_path_obj;
                } else {
                    $self->_set_err("Invalid transport type for naa device $naa_device_name");
                    return $self;
                }
            }
        }

        # determine if this is a multi protocol device
        if (scalar(keys %{$protocol_path_counts}) > 1) {
            $protocol = "mixed";
        }

        # sometimes devices don't have the display name containing the protocol
        # so dynamically determine the protocol here:
        if (!$protocol) {
            if ($protocol_path_counts->{iscsi}) {
                $protocol = "iscsi";
            } elsif ($protocol_path_counts->{fcp}) {
                $protocol = "fcp";
            } else {
                # probably should never happen but whatever
                $protocol = "unknown";
            }
        }

        # discover alua information for this device and working path information
        my $storage_array_type = "";
        my $path_selection_policy;
        my @working_paths;
        my $fcoe_disk = 0;

        #naa.60a98000687047624334614d514b7673
        #  Device Display Name: NETAPP iSCSI Disk (naa.60a98000687047624334614d514b7673)
        #  Storage Array Type: VMW_SATP_DEFAULT_AA
        #  Storage Array Type Device Config: SATP VMW_SATP_DEFAULT_AA does not support device configuration.
        #  Path Selection Policy: VMW_PSP_FIXED
        #  Path Selection Policy Device Config: {preferred=vmhba33:C0:T1:L3;current=vmhba33:C0:T1:L3}
        #  Working Paths: vmhba33:C0:T1:L3

        #naa.60a980004334643442345547376a4355
        #  Device Display Name: NETAPP Fibre Channel Disk (naa.60a980004334643442345547376a4355)
        #  Storage Array Type: VMW_SATP_ALUA
        #  Storage Array Type Device Config: {implicit_support=on;explicit_support=off; explicit_allow=on;alua_followover=on;{TPG_id=1,TPG_state=ANO}{TPG_id=0,TPG_state=AO}}
        #  Path Selection Policy: VMW_PSP_RR
        #  Path Selection Policy Device Config: {policy=rr,iops=1000,bytes=10485760,useANO=0;lastPathIndex=0: NumIOsPending=0,numBytesPending=0}
        #  Working Paths: vmhba2:C0:T1:L23, vmhba3:C0:T0:L23

        foreach my $esxcli_nmp_device_block (@esxcli_nmp_device_list) {
            next if ($esxcli_nmp_device_block !~ /^\s*$naa_device_name/);

            foreach my $esxcli_nmp_device_line (split(/\n/, $esxcli_nmp_device_block)) {
                if ($esxcli_nmp_device_line =~ /Storage Array Type: (\S+)/) {
                    $storage_array_type = $1;
                } elsif ($esxcli_nmp_device_line =~ /Path Selection Policy: (\S+)/) {
                    $path_selection_policy = $1;
                } elsif ($esxcli_nmp_device_line =~ /Working Paths: (.+)/) {
                    my $working_paths = $1;
                    $working_paths =~ s/\s+//g;
                    @working_paths = split(/,/, $working_paths);
                } elsif ($esxcli_nmp_device_line =~ /Fibre Channel Over Ethernet Disk/i) {
                    $fcoe_disk = 1;
                }
            }
            last;
        }

        # check if alua is enabled or not enabled on this device
        my $alua;
        if (defined($storage_array_type)) {
            if ($storage_array_type =~ /ALUA/i) {
                $alua = "enabled";
            } else {
                $alua = "disabled";
            }
        } else {
            $alua = "unknown";
        }

        # iterate through all the lun paths, and indicate which are active optimized and which are active non optimized
        # also note down which are working paths/not working paths
        foreach my $path_name (keys %{$lun_paths}) {
            if (grep(/^$path_name$/, @working_paths)) {
                # this is a working path
                $lun_paths->{$path_name}->{working_path} = "yes";
            } else {
                if ($mp_type =~ /VxDMP/i) {
                    # primary/secondary path states are determined differently with vxdmp
                    $lun_paths->{$path_name}->{working_path} = "no"  if ($lun_paths->{$path_name}->{path_state} =~ /standby|dead|off/i);
                    $lun_paths->{$path_name}->{working_path} = "yes" if ($lun_paths->{$path_name}->{path_state} =~ /active/i);
                } else {
                    # this is not a working path
                    $lun_paths->{$path_name}->{working_path} = "no";
                }
            }

            my $path_identifier;

            # generate the proper path identifier for this path
            if ($lun_paths->{$path_name}->{protocol} eq "iscsi") {
                # iscsi path identifier
                # iqn.1998-01.com.vmware:ibm3550-rtp01-460ff1e2-00023d000001,iqn.1992-08.com.netapp:sn.da83862daa2d11dfb70f123478563412:vs.4,t,1033-naa.60a98000316b2f7770243873382d476e
                $path_identifier = $lun_paths->{$path_name}->{initiator_iqn} . "-" .
                    $lun_paths->{$path_name}->{target_iscsi_session} . "," .
                        $lun_paths->{$path_name}->{target_iqn} . ",t," .
                            $lun_paths->{$path_name}->{tpgt} . "-" . $naa_device_name;
            } elsif ($lun_paths->{$path_name}->{protocol} eq "fcp") {
                # fc path identifier:
                # fc.200100e08bbcebc1:210100e08bbcebc1-fc.500a09808697d794:500a09818697d794-naa.60a980004334652f4f5a506f6e307967
                # fcoe path identifier:
                # fcoe.1000001b21531dd3:2000001b21531dd3-fcoe.500a0980892b9344:500a0986892b9344-naa.60a98000572d57635234612f73534d50
                my $init_wwnn = $lun_paths->{$path_name}->{initiator_wwnn};
                my $init_wwpn = $lun_paths->{$path_name}->{initiator_wwpn};
                my $target_wwnn = $lun_paths->{$path_name}->{target_wwnn};
                my $target_wwpn = $lun_paths->{$path_name}->{target_wwpn};
                $init_wwnn =~ s/\://g;
                $init_wwpn =~ s/\://g;
                $target_wwnn =~ s/\://g;
                $target_wwpn =~ s/\://g;

                $path_identifier = "fco?e?." . $init_wwnn . ":" .
                    $init_wwpn . "-fco?e?." .
                        $target_wwnn . ":" . $target_wwpn .
                            "-" . $naa_device_name;
            } else {
                $self->_set_err("Failed to determine protocol of path '$path_name'");
                return $self;
            }

            #iqn.1998-01.com.vmware:auto-ibm3550-03-6e1004b8-00023d000001,iqn.1992-08.com.netapp:sn.84209003,t,1001-naa.60a9800068704533426f614d514b6f38
            #   Runtime Name: vmhba33:C0:T0:L3
            #   Device: naa.60a9800068704533426f614d514b6f38
            #   Device Display Name: NETAPP iSCSI Disk (naa.60a9800068704533426f614d514b6f38)
            #   Group State: active
            #   Array Priority: 0
            #   Storage Array Type Path Config: SATP VMW_SATP_DEFAULT_AA does not support path configuration.
            #   Path Selection Policy Path Config: {current: yes; preferred: yes}

            #fc.20000000c96e4dd7:10000000c96e4dd7-fc.500a09808697d794:500a09818697d794-naa.60a98000433464344234534b6b747642
            #   Runtime Name: vmhba3:C0:T1:L20
            #   Device: naa.60a98000433464344234534b6b747642
            #   Device Display Name: NETAPP Fibre Channel Disk (naa.60a98000433464344234534b6b747642)
            #   Group State: active unoptimized
            #   Array Priority: 0
            #   Storage Array Type Path Config: {TPG_id=1,TPG_state=ANO,RTP_id=1,RTP_health=UP}
            #   Path Selection Policy Path Config: PSP VMW_PSP_RR does not support path configuration.

            #fcoe.1000001b21531dd3:2000001b21531dd3-fcoe.500a0980892b9344:500a0986892b9344-naa.60a98000572d57635234612f73534d50
            #   Runtime Name: vmhba35:C0:T1:L3
            #   Device: naa.60a98000572d57635234612f73534d50
            #   Device Display Name: NETAPP Fibre Channel Over Ethernet Disk (naa.60a98000572d57635234612f73534d50)
            #   Group State: active
            #   Array Priority: 0
            #   Storage Array Type Path Config: {TPG_id=2,TPG_state=AO,RTP_id=6,RTP_health=UP}
            #   Path Selection Policy Path Config: PSP VMW_PSP_RR does not support path configuration.

            # parse out the reported alua state for each path
            if ($mp_type =~ /NMP/i) {
                foreach my $esxcli_nmp_path_block (@esxcli_nmp_path_list) {
                    next if ($esxcli_nmp_path_block !~ /^$path_identifier/);
                    foreach my $esxcli_nmp_path_line (split(/\n/, $esxcli_nmp_path_block)) {
                        if ($esxcli_nmp_path_line =~ /Group State: (.+)/) {
                            my $alua_state = $1;
                            $lun_paths->{$path_name}->{alua_state} = "AO"   if ($alua_state =~ /active/i);
                            $lun_paths->{$path_name}->{alua_state} = "ANO"  if ($alua_state =~ /active unoptimized/i);
                            $lun_paths->{$path_name}->{alua_state} = "DOWN" if ($alua_state =~ /dead/i);
                            last;
                        }
                    }
                }
            } elsif ($mp_type =~ /VxDMP/i) {
                # alua state is determined in a different way with Veritas DMP multipathing
                $lun_paths->{$path_name}->{alua_state} = "AO"   if ($lun_paths->{$path_name}->{path_state} =~ /active/);
                $lun_paths->{$path_name}->{alua_state} = "ANO"  if ($lun_paths->{$path_name}->{path_state} =~ /standby/);
                $lun_paths->{$path_name}->{alua_state} = "DOWN" if ($lun_paths->{$path_name}->{path_state} =~ /dead|off/);
            }

                # if no alua state was parsed, mark it UNKKNOWN
            if (not defined($lun_paths->{$path_name}->{alua_state})) {
                $lun_paths->{$path_name}->{alua_state} = "UNKNOWN";
            }
        }

        # create the MPDev object for this device
        my $mp_device = new San::SanClient::MPDevMgr::MPDev(-sc_obj => $sc_obj,
                                                            -mp_device_name => $naa_device_name,
                                                            -device_size => $device_size,
                                                            -mp_type => $mp_type,
                                                            -protocol => $protocol,
                                                            -mp_device_path => $naa_device_path,
                                                            -device_vendor => $device_vendor,
                                                            -device_model => $device_model,
                                                            -device_revision => $device_revision,
                                                            -scsi_level => $scsi_level,
                                                            -lun_id => $lun_id,
                                                            -alua => $alua,
                                                            -is_offline => $is_offline);

        # store some os specific data about this device
        $mp_device->set_os_specific_data(-path_selection_policy => $path_selection_policy);
        $mp_device->set_os_specific_data(-raw_scsi_device_name => $raw_scsi_device_name);
        $mp_device->set_os_specific_data(-vaai_status => $vaai_status);
        $mp_device->set_os_specific_data(-storage_array_type => $storage_array_type);

        # add the new device to the device manager
        # will overwrite itself if it always exists in the devmgr
        $self->_add_device(-mp_device_obj => $mp_device,
                           -mp_device_path_objs => [values %{$lun_paths}]);
    }

    return $self;
}

###############################################################################
# Method:    refresh_paths()
# Objective: Rediscover all Multipathed LUNs on the system and populate the
#            Device Manager
###############################################################################
sub _refresh_paths {
    my ($self, %args) = @_;

    # refresh paths is just an alias for refresh on ESX
    $self->refresh();
    if ($self->error()) {
        return;
    }

    return;
}

###############################################################################
# Method:    os_specific_validation()
# Objective: Perform OS Specific validation on all devices passed in
###############################################################################
sub _os_specific_validation {
    my ($self) = @_;

    my $dev_hash = $self->get_devices_hash();
    my $old_dev_hash = $self->get_old_devices_hash();

    # check if there is anything to validate
    if (not scalar(keys %{$dev_hash}) and not scalar(keys %{$old_dev_hash})) {
        return ([], []);
    }

    my @warnings = ();
    my @errors = ();
    my @bad_cmode_satp_luns = ();

    # perform some validation for ESX
    foreach my $mp_device_name (keys %{$dev_hash}) {
        my $dev_obj = $dev_hash->{$mp_device_name};

        # perform some validation of NMP devices
        if ($dev_obj->mp_type =~ /NMP/i) {
            # validate that satp hasn't changed since the last time we discovered this devic
            my $satp_current = $dev_obj->get_os_specific_data("storage_array_type");
            if (exists($old_dev_hash->{$mp_device_name})) {
                # compare the satp values
                my $satp_old = $old_dev_hash->{$mp_device_name}->get_os_specific_data("storage_array_type");
                if ($satp_current ne $satp_old) {
                    # satp type changed for this device
                    push(@warnings, "\t$mp_device_name: SATP type has changed, Old value: ($satp_old),\tCurrent Value: ($satp_current)");
                }
            }

            # validate that the SATP for this device is VMW_SATP_ALUA if it is a cmode LUN
            if ($dev_obj->ontap_mode() =~ /cmode/i and $satp_current ne "VMW_SATP_ALUA") {
                push(@bad_cmode_satp_luns, $mp_device_name . " : " . $satp_current);
            }
        }
    }

    # push on warnings that we find
    if (scalar(@bad_cmode_satp_luns)) {
        push(@warnings, "\tDetected non-ALUA C-Mode LUNs. These LUNs should have 'VMW_SATP_ALUA'. " .
                 "You may need to verify that the VMW_SATP_ALUA plugin module is loaded on the ESX host.");
        push(@warnings, "\t\t" . join("\n\t\t", @bad_cmode_satp_luns));
    }

    return (\@errors, \@warnings);
}

1;
