#
# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary MPDev base class for ESX
## @author hrivera@netapp.com
## @status private
## @pod ../MPDev.pm

package San::SanClient::MPDevMgr::ESXMPDevMgr::ESXMPDev;

use strict;
use warnings;
use San::SanClient::MPDevMgr::MPDevPath;

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: Create an MPDev instance
###############################################################################
sub _new {
    my ($self, %args) = @_;

    return $self;
}

###############################################################################
# Method:    refresh()
# Objective: Rediscover the state of this device
###############################################################################
sub _refresh {
    my ($self, %args) = @_;

    my $sc_obj = $self->{sc_obj};
    my $naa_device_name = $self->{mp_device_name};

    my $naa_device_path;
    my $device_vendor;
    my $device_model;
    my $device_revision;
    my $scsi_level;
    my $raw_scsi_device_name;
    my $device_size;
    my $mp_type;
    my $vaai_status = "";

    # discovery inquiry information for this device
    my $cmd;
    my $status2;
    my @output2;

    my $corestorage_ns = $sc_obj->{esxcli_corestorage_namespace};
    if ($sc_obj->get_os_version() =~ /^4\./) {
        $cmd = "esxcfg-scsidevs -c -d $naa_device_name";
        ($status2, @output2) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
        if ($status2 != 0) {
            $self->_set_err("Failed");
            return $self;
        }
    } else {
        $cmd = "esxcli ${corestorage_ns} device list --device=${naa_device_name}";
        ($status2, @output2) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                         -include_host_name => 1);
        if ($status2 != 0) {
            $self->_set_err("Failed");
            return $self;
        }
    }

    foreach my $scsidev_line (split(/\n/, @output2)) {
        if ($scsidev_line =~ /Display Name:.+iSCSI Disk/) {
            $self->protocol("iscsi");
        } elsif ($scsidev_line =~ /Display Name:.+(Fibre Channel Disk|Fibre Channel Over Ethernet Disk)/) {
            $self->protocol("fcp");
        } elsif ($scsidev_line =~ /Devfs Path:\s+(\S+)/) {
            $naa_device_path = $1;
            $raw_scsi_device_name = $naa_device_path;
        } elsif ($scsidev_line =~ /Multipath Plugin:\s+(.+)/) {
            $mp_type = $1;
        } elsif ($scsidev_line =~ /Plugin:\s+(.+)/) {
            $mp_type = $1;
        } elsif ($scsidev_line =~ /Size:\s+(\S+)/) {
            $device_size = $1 . "MB";
        } elsif ($scsidev_line =~ /Vendor:\s+(.+?)\s+Model:\s+(.+?)\s+Revis:\s+(\S+)/) {
            $device_vendor = $1;
            $device_model = $2;
            $device_revision = $3;
        } elsif ($scsidev_line =~ /Vendor:\s+(.+)/) {
            $device_vendor = $1;
            # remove trailing whitespace
            $device_vendor =~ s/\s+$//;
        } elsif ($scsidev_line =~ /Model:\s+(.+)/) {
            $device_model = $1;
            # remove trailing whitespace
            $device_model =~ s/\s+$//;
        } elsif ($scsidev_line =~ /Revision:\s+(.+)/) {
            $device_revision = $1;
        } elsif ($scsidev_line =~ /SCSI Level:\s+(\S+)/) {
            $scsi_level = $1;
        } elsif ($scsidev_line =~ /VAAI Status:\s+(.+)/) {
            $vaai_status = $1;
        }
    }

    # if the naa_device_path is undefined, then this device is offline "detached"
    if (not defined($naa_device_path)) {
        # build the naa device path manually
        $naa_device_path = "/vmfs/devices/disks/${naa_device_name}";
        $self->{is_offline} = 1;
    }

    # assign all the updated attributes
    $self->mp_device_path($naa_device_path);
    $self->device_vendor($device_vendor);
    $self->device_model($device_model);
    $self->device_revision($device_revision);
    $self->device_scsi_level($scsi_level);
    $self->set_os_specific_data(-raw_scsi_device_name => $raw_scsi_device_name);
    $self->set_os_specific_data(-vaai_status => $vaai_status);
    $self->device_size($device_size);
    $self->mp_type($mp_type);

    # refresh paths for this multipathed device
    $self->refresh_paths();

    return;
}

###############################################################################
# Method:    refresh_paths()
# Objective: Rediscover the state of this device's paths
###############################################################################
sub _refresh_paths {
    my ($self, %args) = @_;

    my $sc_obj = $self->{sc_obj};
    my $naa_device_name = $self->{mp_device_name};

    # discover the lun paths to this device
    my $cmd;
    if ($sc_obj->get_os_version() =~ /^4\./) {
        $cmd = "esxcfg-mpath -b -d $naa_device_name";
    } else {
        $cmd = "esxcli storage core path list --device=$naa_device_name";
    }
    my ($status2, @output2) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
    if ($status2 != 0) {
        $self->_set_err("Failed");
        return;
    }

    my $lun_paths = {};
    my $protocol_path_counts = {};
    my $nmp_namespace = $sc_obj->{esxcli_nmp_namespace};

    if ($sc_obj->get_os_version() =~ /^4\./) {
        foreach my $line2 (@output2) {
            next if ($line2 =~ /NETAPP/);

            if ($line2 =~ /^\s*(vmhba\S+)\s+LUN:(\d+)\s+state:(\S+).+?WWNN:\s+(\S+)\s+WWPN:\s+(\S+)\s+Target:?\s+WWNN:\s+(\S+)\s+WWPN:\s+(\S+)/) {

                my $path_name = $1;
                my $lun_id = $2;
                $self->lun_id($lun_id);
                my $path_state = $3;
                my $path_initiator_wwnn = $4;
                my $path_initiator_wwpn = $5;
                my $path_target_wwnn = $6;
                my $path_target_wwpn = $7;

                my $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $path_name,
                                                                              -lun_id => $lun_id,
                                                                              -path_state => $path_state,
                                                                              -initiator_wwnn => $path_initiator_wwnn,
                                                                              -initiator_wwpn => $path_initiator_wwpn,
                                                                              -target_wwnn => $path_target_wwnn,
                                                                              -target_wwpn => $path_target_wwpn,
                                                                              -protocol => "fcp");
                $protocol_path_counts->{fcp}++;

                $lun_paths->{$path_name} = $mp_dev_path_obj;

            } elsif ($line2 =~ /^\s*(vmhba\S+)\s+LUN:(\d+)\s+state:(\S+).+?iscsi\s+Adapter:\s+(\S+)\s+Target:\s+(?:IQN=)?(\S+)/) {
                my $path_name = $1;
                my $lun_id = $2;
                $self->lun_id($lun_id);
                my $path_state = $3;
                my $path_initiator_iqn = $4;
                my $path_target_iqn = $5;
                my $iscsi_target_alias;
                my $iscsi_session;
                my $tpgt;

                if ($path_target_iqn =~ /Unavailable/i) {
                    $iscsi_target_alias = "";
                    $iscsi_session = "";
                    $tpgt = "";
                } elsif ($line2 =~ /Alias=(\S+)?\s+Session=(\S+)\s+PortalTag=(\S+)/) {
                    $iscsi_target_alias = $1;
                    $iscsi_session = $2;
                    $tpgt = $3;
                }

                my $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $path_name,
                                                                              -lun_id => $lun_id,
                                                                              -path_state => $path_state,
                                                                              -initiator_iqn => $path_initiator_iqn,
                                                                              -target_iqn => $path_target_iqn,
                                                                              -target_iscsi_alias => $iscsi_target_alias,
                                                                              -target_iscsi_session => $iscsi_session,
                                                                              -target_ip_address => "",
                                                                              -target_ip_port => "",
                                                                              -tpgt => $tpgt,
                                                                              -protocol => "iscsi");


                $protocol_path_counts->{iscsi}++;

                $lun_paths->{$path_name} = $mp_dev_path_obj;
            }
        }
    } else {
        # esx5 or later
        # separate the paths into path blocks
        my @path_list_blocks = split("\n\n", join("\n",@output2));
        foreach my $esxcfg_mpath_block (@path_list_blocks) {
            if ($esxcfg_mpath_block =~ /Transport: fc/i) {
                my $path_name;
                my $lun_id = "";
                my $path_state;
                my $path_initiator_wwnn;
                my $path_initiator_wwpn;
                my $path_target_wwnn;
                my $path_target_wwpn;

                foreach my $line (split("\n", $esxcfg_mpath_block)) {
                    if ($line =~ /Runtime Name:\s*(\S+)/) {
                        $path_name = $1;
                    } elsif ($line =~ /LUN:\s*(\S+)/) {
                        $lun_id = $1;
                        $self->lun_id($lun_id);
                    } elsif ($line =~ /State:\s*(\S+)/) {
                        $path_state = $1;
                    } elsif ($line =~ /Adapter Transport Details: WWNN: (\S+) WWPN: (\S+)/) {
                        $path_initiator_wwnn = $1;
                        $path_initiator_wwpn = $2;
                    } elsif ($line =~ /Target Transport Details: WWNN: (\S+) WWPN: (\S+)/) {
                        $path_target_wwnn = $1;
                        $path_target_wwpn = $2;
                    }
                }

                $path_initiator_wwnn = "Unavailable" if (not defined($path_initiator_wwnn));
                $path_initiator_wwpn = "Unavailable" if (not defined($path_initiator_wwpn));
                $path_target_wwnn    = "Unavailable" if (not defined($path_target_wwnn));
                $path_target_wwpn    = "Unavailable" if (not defined($path_target_wwpn));

                # create a path object
                my $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $path_name,
                                                                              -lun_id => $lun_id,
                                                                              -path_state => $path_state,
                                                                              -initiator_wwnn => $path_initiator_wwnn,
                                                                              -initiator_wwpn => $path_initiator_wwpn,
                                                                              -target_wwnn => $path_target_wwnn,
                                                                              -target_wwpn => $path_target_wwpn,
                                                                              -protocol => "fcp");
                $protocol_path_counts->{fcp}++;
                $lun_paths->{$path_name} = $mp_dev_path_obj;
            } elsif ($esxcfg_mpath_block =~ /Transport: iscsi/i) {
                # parse an iSCSI path
                my $path_name;
                my $lun_id = "";
                my $path_state;
                my $path_initiator_iqn;
                my $path_target_iqn;
                my $iscsi_target_alias;
                my $iscsi_session;
                my $tpgt;

                foreach my $line (split("\n", $esxcfg_mpath_block)) {
                    if ($line =~ /Runtime Name:\s*(\S+)/) {
                        $path_name = $1;
                    } elsif ($line =~ /LUN:\s*(\S+)/) {
                        $lun_id = $1;
                        $self->lun_id($lun_id);
                    } elsif ($line =~ /State:\s*(\S+)/) {
                        $path_state = $1;
                    } elsif ($line =~ /Adapter Transport Details: Unavailable/i) {
                        $path_initiator_iqn = "";
                    } elsif ($line =~ /Adapter Transport Details: (\S+)/) {
                        $path_initiator_iqn = $1;
                    } elsif ($line =~ /Target Transport Details: Unavailable/i) {
                        $path_target_iqn = "";
                        $iscsi_target_alias = "";
                        $iscsi_session = "";
                        $tpgt = "";
                    } elsif ($line =~ /Target Transport Details:\s+IQN=(\S+)\s+Alias=(\S+)?\s+Session=(\S+)\s+PortalTag=(\S+)/) {
                        $path_target_iqn = $1;
                        $iscsi_target_alias = $2;
                        $iscsi_session = $3;
                        $tpgt = $4;
                    }
                }

                # create a path object
                my $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $path_name,
                                                                              -lun_id => $lun_id,
                                                                              -path_state => $path_state,
                                                                              -initiator_iqn => $path_initiator_iqn,
                                                                              -target_iqn => $path_target_iqn,
                                                                              -target_iscsi_alias => $iscsi_target_alias,
                                                                              -target_iscsi_session => $iscsi_session,
                                                                              -target_ip_address => "",
                                                                              -target_ip_port => "",
                                                                              -tpgt => $tpgt,
                                                                              -protocol => "iscsi");
                $protocol_path_counts->{iscsi}++;
                $lun_paths->{$path_name} = $mp_dev_path_obj;
            } else {
                $self->_set_err("Invalid transport type for naa device $naa_device_name");
                return;
            }
        }
    }

    # determine if this is a multi protocol device
    if (scalar(keys %{$protocol_path_counts}) > 1) {
        $self->protocol("mixed");
    }

    # discovery alua information for this device and working path information
    $cmd = "esxcli ${nmp_namespace} device list -d $naa_device_name";
    ($status2, @output2) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                     -include_host_name => 1);
    if ($status2 != 0) {
        $self->_set_err("Failed");
        return;
    }
    my $storage_array_type;
    my $path_selection_policy = "";
    my @working_paths;
    my $fcoe_disk = 0;

    foreach my $line2 (@output2) {
        if ($line2 =~ /Storage Array Type: (\S+)/) {
            $storage_array_type = $1;
        } elsif ($line2 =~ /Path Selection Policy: (\S+)/) {
            $path_selection_policy = $1;
        } elsif ($line2 =~ /Working Paths: (.+)/) {
            my $working_paths = $1;
            $working_paths =~ s/\s+//g;
            @working_paths = split(/,/, $working_paths);
        } elsif ($line2 =~ /Fibre Channel Over Ethernet Disk/i) {
            $fcoe_disk = 1;
        }
    }

    # store the path selection policy
    $self->set_os_specific_data(-path_selection_policy => $path_selection_policy);

    # check if alua is enabled or not enabled on this device
    if ($storage_array_type =~ /ALUA/i) {
        $self->alua("enabled");
    } else {
        $self->alua("disabled");
    }

    # iterate through all the lun paths, and indicate which are active optimized and which are active non optimized
    # also note down which are working paths/not working paths
    foreach my $path_name (keys %{$lun_paths}) {
        if (grep(/^$path_name$/, @working_paths)) {
            # this is a working path
            $lun_paths->{$path_name}->{working_path} = "yes";
        } else {
            # this is not a working path
            $lun_paths->{$path_name}->{working_path} = "no";
        }

        @output2 = ();
        if ($lun_paths->{$path_name}->{protocol} eq "iscsi") {
            # iscsi path identifier
            # iqn.1998-01.com.vmware:ibm3550-rtp01-460ff1e2-00023d000001,
            # iqn.1992-08.com.netapp:sn.da83862daa2d11dfb70f123478563412:vs.4,t,1033-naa.60a98000316b2f7770243873382d476e
            my $path_identifier = $lun_paths->{$path_name}->{initiator_iqn} . "-" .
                                  $lun_paths->{$path_name}->{target_iscsi_session} . "," .
                                  $lun_paths->{$path_name}->{target_iqn} . ",t," .
                                  $lun_paths->{$path_name}->{tpgt} . "-" .
                                  $naa_device_name;
            $cmd = "esxcli ${nmp_namespace} path list -p $path_identifier";
            ($status2, @output2) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                             -include_host_name => 1);
            if ($status2 != 0) {
                $self->_set_err("Failed");
                return $self;
            }
        } elsif ($lun_paths->{$path_name}->{protocol} eq "fcp") {
            # fc path identifier
            # fc.200100e08bbcebc1:210100e08bbcebc1-fc.500a09808697d794:500a09818697d794-naa.60a980004334652f4f5a506f6e307967
            # fcoe path identifier
            # fcoe.200100e08bbcebc1:210100e08bbcebc1-fcoe.500a09808697d794:500a09818697d794-naa.60a980004334652f4f5a506f6e307967
            my $init_wwnn = $lun_paths->{$path_name}->{initiator_wwnn};
            my $init_wwpn = $lun_paths->{$path_name}->{initiator_wwpn};
            my $target_wwnn = $lun_paths->{$path_name}->{target_wwnn};
            my $target_wwpn = $lun_paths->{$path_name}->{target_wwpn};
            $init_wwnn =~ s/\://g;
            $init_wwpn =~ s/\://g;
            $target_wwnn =~ s/\://g;
            $target_wwpn =~ s/\://g;

            # determine whether its an fcoe or fc lun
            my $label;
            if ($fcoe_disk) {
                $label = "fcoe";
            } else {
                $label = "fc";
            }

            my $path_identifier = $label . "." . $init_wwnn . ":" . $init_wwpn . "-${label}." . $target_wwnn . ":" . $target_wwpn . "-" . $naa_device_name;
            $cmd = "esxcli ${nmp_namespace} path list -p $path_identifier";
            ($status2, @output2) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                             -include_host_name => 1);
            if ($status2 != 0) {
                $self->_set_err("Failed");
                return;
            }
        } else {
            $self->_set_err("Failed");
            return;
        }

        # parse out the reported alua state from the OS
        foreach my $line (@output2) {
            if ($line =~ /Group State: (.+)/) {
                my $alua_state = $1;
                $lun_paths->{$path_name}->{alua_state} = "AO"   if ($alua_state =~ /active/i);
                $lun_paths->{$path_name}->{alua_state} = "ANO"  if ($alua_state =~ /active unoptimized/i);
                $lun_paths->{$path_name}->{alua_state} = "DOWN" if ($alua_state =~ /dead/i);
                last;
            }
        }
        # if no alua state was parsed, mark UNK
        if (not defined($lun_paths->{$path_name}->{alua_state})) {
            $lun_paths->{$path_name}->{alua_state} = "UNKNOWN";
        }
    }

    # add the newly discovered path state for this device
    # convert hash to array of path objects first
    my @path_objs = values %{$lun_paths};
    $self->_add_path_state(-mp_device_path_objs => \@path_objs);

    return;
}

###############################################################################
# Method:    create_partition()
# Objective: Create a partition on this device
###############################################################################
sub _create_partition {
    my ($self) = @_;

    # get the path to this disk
    my $raw_disk_path = $self->mp_device_path();

    my $sc_obj = $self->{sc_obj};

    # fdisk this device
    my $cmd = "echo 'n\np\n1\n\n\nt\nfb\nx\nb\n1\n128\np\nw\n' " .
              "| fdisk ${raw_disk_path}:0";
    my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Could not create partition on $raw_disk_path");
        return [];
    }

    return ["${raw_disk_path}:1"];
}

###############################################################################
# Method:    destroy_partition()
# Objective: Destroy the partition on this device
###############################################################################
sub _destroy_partition {
    my ($self) = @_;

    # get the path to this disk
    my $raw_disk_path = $self->mp_device_path();

    my $sc_obj = $self->{sc_obj};

    # fdisk this device
    my $cmd = "echo 'd\n1\nw\n' " .
              "| fdisk $raw_disk_path";
    my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Could not destroy partition on $raw_disk_path");
        return;
    }

    return;
}

###############################################################################
# Method:    mkfs()
# Objective: Create a filesystem on this device's main partition
###############################################################################
sub _mkfs {
    my ($self, %args) = @_;

    # establish default values
    my $mp_device_path = $self->mp_device_path();
    my $mp_device_name = $self->mp_device_name();
    my $partition_path = $mp_device_path . ":1";
    my $default_fs_label = "nate_datastore_" . $mp_device_name;

    # parse args
    my $fs_type = exists($args{-fs_type}) ? $args{-fs_type} : "vmfs3";
    my $fs_label = exists($args{-fs_label}) ? $args{-fs_label} : $default_fs_label;
    my $block_size = exists($args{-block_size}) ? $args{-block_size} : "1m";

    my $sc_obj = $self->{sc_obj};

    # make a file system on the device (in other words, create a datastore)
    my $retry_count = 2;
    while (1) {
        my $cmd = "vmkfstools -C $fs_type -b $block_size -S $fs_label $partition_path";
        my ($status, @output) = $sc_obj->_vmkfstools_cmd($cmd);
        if ($status > 0) {
            if (grep(/Unable to create vmfs/, @output) and $retry_count > 0) {
                # retry the create vmkfstools call if this failed...
                $retry_count--;
                redo;
            }
            $self->_set_err("Failed to create '$fs_type' filesystem on device $partition_path");
            return;
        } else {
            last;
        }
    }

    # set the mount point to be the path to the datastore name
    $self->mount_point("/vmfs/volumes/${fs_label}");

    return;
}

###############################################################################
# Method:    mount()
# Objective: Mount this device's filesystem
###############################################################################
sub _mount {
    my ($self, %args) = @_;

    # NOOP, since making a vmfs datastore has the same effect as mounting

    return $self->mount_point();
}

###############################################################################
# Method:    unmount()
# Objective: Unmount this device's filesystem
###############################################################################
sub _unmount {
    my ($self, %args) = @_;

    # NOOP, since destroying a vmfs datastore has the same effect as unmounting

    return;
}

###############################################################################
# Method:    online_device()
# Objective: online this device
###############################################################################
sub _online_device {
    my ($self, %args) = @_;

    # NOOP for Console based ESX versions

    return;
}

###############################################################################
# Method:    offline_device()
# Objective: offline this device
###############################################################################
sub _offline_device {
    my ($self, %args) = @_;

    my $sc_obj = $self->{sc_obj};
    my $corestorage_ns = $sc_obj->{esxcli_corestorage_namespace};

    if ($sc_obj->get_os_version() =~ /^4\./) {
        # reclaim the device (esxcli corestorage claiming reclaim -d $mp_device_name)
        my $mp_device_name = $self->mp_device_name();
        my $cmd = "esxcli ${corestorage_ns} claiming reclaim -d ${mp_device_name}";
        my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                          -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Could not reclaim paths to device $mp_device_name");
            return;
        }
    }

    return;
}

###############################################################################
# Method:    list_contents
# Objective: return back a list of files inside this device
###############################################################################
sub _list_contents {
    my ($self, %args) = @_;

    my $subdir = exists($args{-subdir}) ? $args{-subdir} : undef;

    my $mp_device_name = $self->mp_device_name();
    my $datastore_label = $self->get_os_specific_data("datastore_label");
    if (not length($datastore_label)) {
        $self->_set_err("Cannot list contents of Device $mp_device_name because it is not a VMFS datastore");
        return (wantarray()) ? () : [];
    }

    my $path = "/vmfs/volumes/${datastore_label}";
    if (defined($subdir)) {
        $path .= "/${subdir}";
    }

    my $cmd = "ls '${path}' | cat";
    my ($status, @output) = $self->{sc_obj}->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to list contents of $datastore_label");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    load()
# Objective: Load in the current state of this device's filesystem/partition
###############################################################################
sub _load {
    my ($self) = @_;

    my $mp_device_name = $self->mp_device_name();
    my $sc_obj = $self->{sc_obj};

    # load os specific information about this device
    my $cmd = "esxcfg-scsidevs --vmfs | grep \"${mp_device_name}:\"";
    my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        # there is no data for this device to load
        # un set any of the os specific data fields for this device
        $self->set_os_specific_data(-datastore_label => "",
                                    -datastore_uuid => "",
                                    -datastore_path => "",
                                    -datastore_partition_path => "");

        # update some fields on this object, stating that it is not mounted,
        # and doesnt have a filesystem
        $self->is_partitioned(0);
        $self->is_mounted(0);
        $self->has_filesystem(0);
        $self->fs_type("");
        $self->mount_point("");
        $self->mp_device_partition_list([]);

        return;
    }

    # parse out the data for this device
    if ($output[0] =~ /^\s*(${mp_device_name}:\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/) {
        my $mp_partition_name = $1;
        my $mp_partition_path = $2;
        my $datastore_uuid = $3;
        my $unknown = $4;
        my $datastore_label = $5;

        # store os specific data about this datastore
        $self->set_os_specific_data(-datastore_label => $datastore_label,
                                    -datastore_uuid => $datastore_uuid,
                                    -datastore_path => "/vmfs/volumes/${datastore_label}",
                                    -datastore_partition_path => $mp_partition_path);

        $self->is_partitioned(1);
        $self->is_mounted(1);
        $self->has_filesystem(1);
        $self->fs_type("vmfs");
        $self->mount_point("/vmfs/volumes/${datastore_label}");
        $self->mp_device_partition_list([$mp_partition_path]);

        return;
    } else {
        $self->_set_err("Failed to parse output properly during load of $mp_device_name");
        return;
    }
}

###############################################################################
# Method:    mask_lun_paths()
# Objective: To mask all the paths for the device
###############################################################################
sub _mask_lun_paths {
    my ($self,%args) = @_;

    my (@rule_result);
    my $sc_obj = $self->{sc_obj};
    my $naa_id = $self->{mp_device_name};
    my $path;
    my $vmhba;
    my $channel;
    my $target;
    my $lun;
    my $rule_id;
    my @rule_ids;
    my $cmd;
    my $status;

    my $corestorage_ns = $sc_obj->{esxcli_corestorage_namespace};
    $rule_id = 100;

    my $hash = $self->get_current_path_state()->get_path_objs_hash();
    if (!defined($hash)) {
        $self->_set_err("No lun with naa id $naa_id is mapped to this host");
        return;
    } else {
        foreach $path (keys %{$hash}) {
            #Getting the vmhba, channel,target and lun fields required for lun masking
            $path =~ m/(vmhba\d+):C(\d+):T(\d+):L(\d+)\s+\.*/;
            $vmhba = $1;
            $channel = $2;
            $target = $3;
            $lun = $4;
            #Running the esx commands according to the version
            
            do {
                $rule_id++;
                $cmd = "esxcli $corestorage_ns claimrule add --rule $rule_id -t location -A $vmhba -C $channel -T $target -L $lun -P MASK_PATH";
                ($status, my @rule_result) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
                
            } while ($rule_result[0] =~ /already exists/);
            #pushing the rule_id in the list of rule_ids 
            push(@rule_ids, $rule_id);
            
            #Unclaiming path to a device 
            $cmd = "esxcli $corestorage_ns claiming unclaim -t location -A $vmhba -C $channel -T $target -L $lun";
            ($status) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to unclaim the paths");
                return;
            }
        }
        
        $self->sanlog(-msg => "loading all the newly added rules");
        $cmd = "esxcli $corestorage_ns claimrule load";
        ($status) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to load rules");
            return;
        }	
        $cmd = "esxcli $corestorage_ns claimrule run";
        ($status) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to run rules");
            return ;
        }
    }
    $self->set_os_specific_data(-rule_ids => \@rule_ids);
    return;
}
###############################################################################
# Method:    unmask_lun_paths()
# Objective: Unmask the paths for this device
###############################################################################
sub _unmask_lun_paths {
    my ($self, %args) = @_;

    my ($line,$rule);
    my ($vmhba_field,$channel_field,$target_field,$lun_field,@line_fields,@vmhba_fields,@channel_fields,@target_fields,@lun_fields);
    my ($vmhba,$lun,$target,$channel,$rule_id,$cmd,$status,@output,@rule_ids);

    my $sc_obj = $self->{sc_obj};
    my $corestorage_ns = $sc_obj->{esxcli_corestorage_namespace};
    my $rule_ids = $self->get_os_specific_data(-key => "rule_ids");
    if (ref($rule_ids) ne "ARRAY") {
        $self->_set_err("No rules array found");
        return;
    }
    @rule_ids = @{$self->get_os_specific_data(-key => "rule_ids")};
    #getting the parameters set for the given rule is
    
    $cmd="esxcli $corestorage_ns claimrule list";
    ($status,@output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
    
    foreach $rule_id (@rule_ids) {
        foreach $line (@output) {
            @line_fields = split(/\s+/,$line);
            $rule = $line_fields[1];

            if ($rule eq $rule_id) {
                $vmhba_field = $line_fields[5];
                $channel_field = $line_fields[6];
                $target_field = $line_fields[7];
                $lun_field = $line_fields[8];
                @vmhba_fields = split(/=/,$vmhba_field);
                @channel_fields = split(/=/,$channel_field);
                @target_fields = split(/=/,$target_field);
                @lun_fields = split(/=/,$lun_field);
                $vmhba = $vmhba_fields[1];
                $channel = $channel_fields[1];
                $target = $target_fields[1];
                $lun = $lun_fields[1];
                last;
            }
        }	
        $cmd ="esxcli $corestorage_ns claimrule delete -r $rule_id";
        ($status) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
        if($status > 0){
            $self->_set_err("Failed to remove the rule with $rule_id");
            return;
        }
        #Unclaim all paths to a device and then run the loaded claimrules on each of the paths to reclaim them	
        $cmd = "esxcli $corestorage_ns claiming unclaim -t location -A $vmhba -C $channel -T $target -L $lun";
        ($status) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to unclaim the paths");
            return;
        }	
        #loading the rules
        $cmd = "esxcli $corestorage_ns claimrule load";
        ($status) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to load rules");
            return;
        }
        #"Running with all the newly added/deletetd rules"	
        $cmd= "esxcli $corestorage_ns claimrule run";
        ($status)= $sc_obj->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to run the updated rules");
            return;
        }
    }
    
    return;
}

1;
