#
# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary MPDevMgr base class for ESXi
## @author hrivera@netapp.com
## @status private
## @pod ../MPDevMgr.pm

package San::SanClient::MPDevMgr::ESXiMPDevMgr;

use strict;
use warnings;
use Data::Dumper;
use San::SanClient::MPDevMgr::MPDev;
use San::SanClient::MPDevMgr::MPDevPath;
use San::SanClient::MPDevMgr::ESXMPDevMgr;

use vars qw (@ISA);

@ISA = qw (Error SanLog San::SanClient::MPDevMgr::ESXMPDevMgr);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: Create an MPDevMgr instance
###############################################################################
sub _new {
    my ($self, %args) = @_;

    # invoke the constructor of ESXMPDevMgr
    $self->SUPER::_new(%args);

    return $self;
}

###############################################################################
# Method:    refresh()
# Objective: Rediscover all multipathed LUNs on the system
###############################################################################
sub _refresh {
    my ($self, %args) = @_;

    # invoke the refresh method of ESXMPDevMgr
    $self->SUPER::_refresh(%args);
    if ($self->error()) {
        $self->_set_err("Failed to refresh ESXi devices");
        return;
    }

    # if any of the devices discovered are vxdmp devices, then discover some
    # extra information for them too
    $self->_refresh_vxdmp_devices();
    if ($self->error()) {
        $self->_set_err("Failed to refresh VxDMP devices");
        return;
    }

    return;
}

###############################################################################
# Method:    _refresh_vxdmp_devices()
# Objective: Refresh any vxdmp devices
###############################################################################
sub _refresh_vxdmp_devices {
    my ($self) = @_;

    # search all the mpdev objects and flag if we find one owned by VxDMP
    my $found = 0;
    my @mpdev_objs = $self->get_device_objs();
    foreach my $mpdev_obj (@mpdev_objs) {
        if ($mpdev_obj->mp_type() =~ /VxDMP/i) {
            $found = 1;
            last;
        }
    }

    # if we found a vxdmp device, then gather dmpnode enclosure information
    # to store locally on each mpdev object
    if ($found) {
        my $sc_obj = $self->{sc_obj};
        my ($status, @output) = $sc_obj->_vxdmpadm_execute_cmd(-args => "getdmpnode");

        foreach my $mpdev_obj (@mpdev_objs) {
            if ($mpdev_obj->mp_type() =~ /VxDMP/i) {
                my $mp_device_name = $mpdev_obj->mp_device_name();
                # parse out the enclosure name for this mpdevice
                foreach my $line (@output) {
                    if ($line =~ /^\s*$mp_device_name\s+(\S+)\s+(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)$/) {
                        my $state = $1;
                        my $enclosure_type = $2;
                        my $num_total_paths = $3;
                        my $num_enabled_paths = $4;
                        my $num_disabled_paths = $5;
                        my $enclosure_name = $6;

                        # store it on the object
                        $mpdev_obj->set_os_specific_data(-vxdmp_enclosure_name => $enclosure_name,
                                                         -vxdmp_state => $state,
                                                         -vxdmp_enclosure_type => $enclosure_type,
                                                         -vxdmp_total_paths => $num_total_paths,
                                                         -vxdmp_num_enabled_paths => $num_enabled_paths,
                                                         -vxdmp_num_disabled_paths => $num_disabled_paths);
                        if ($mpdev_obj->error()) {
                            $self->_set_err("Error setting VxDMP os specific data on object", $mpdev_obj->errmsg());
                            return;
                        }
                        last;
                    }
                }
            }
        }
    }

    return;
}


###############################################################################
# Method:    offline_devices()
# Objective: Offline a group of devices specified
###############################################################################
sub _offline_devices {
    my ($self, %args) = @_;

    my @mpdev_objs = exists($args{-mpdev_objs}) ? @{$args{-mpdev_objs}}: ();
    if (not scalar(@mpdev_objs)) {
        $self->_set_err("No MPDev objects were specified");
        return;
    }

    my $sc_obj = $self->{sc_obj};

    # if os version is ESXi5, just offline each device one by one
    if ($sc_obj->get_os_version() =~ /^(5|6)\./) {
        foreach my $mpdev_obj (@mpdev_objs) {
            $mpdev_obj->offline_device();
            if ($mpdev_obj->error()) {
                $self->_set_err("Failed to offline device", $mpdev_obj->errmsg());
                return;
            }
        }
    } else {
        # invoke the offline_devices() method in ESXMPDevMgr, it will handle both ESXi4 and ESX4
        $self->SUPER::_offline_devices(%args);
        if ($self->error()) {
            $self->_set_err("Failed to offline devices");
            return;
        }
    }

    return;
}

1;
