#
# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary MPDev base class for ESXi
## @author hrivera@netapp.com
## @status private
## @pod ../MPDev.pm

package San::SanClient::MPDevMgr::ESXiMPDevMgr::ESXiMPDev;

use strict;
use warnings;
use San::SanClient::MPDevMgr::MPDevPath;

# inherit the methods from ESXMPDev
use San::SanClient::MPDevMgr::ESXMPDevMgr::ESXMPDev;

use vars qw (@ISA);

@ISA = qw (Error SanLog San::SanClient::MPDevMgr::ESXMPDevMgr::ESXMPDev);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: Create an MPDev instance
###############################################################################
sub _new {
    my ($self, %args) = @_;

    # invoke the constructor of ESXMPdev
    $self->SUPER::_new(%args);

    return $self;
}

###############################################################################
# Method:    refresh()
# Objective: Rediscover this device
###############################################################################
sub _refresh {
    my ($self, %args) = @_;

    # invoke the refresh method of ESXMPDev
    $self->SUPER::_refresh(%args);
    if ($self->error()) {
        $self->_set_err("Failed to refresh ESXi MP device");
        return;
    }

    # if this device is a vxdmp device, then discover some
    # extra information for it
    $self->_refresh_vxdmp_device();
    if ($self->error()) {
        $self->_set_err("Failed to refresh VxDMP MP device");
        return;
    }

    return;
}

###############################################################################
# Method:    _refresh_vxdmp_device()
# Objective: Refresh this vxdmp device
###############################################################################
sub _refresh_vxdmp_device {
    my ($self) = @_;

    # skip non VxDMP devices
    if ($self->mp_type() !~ /VxDMP/i) {
        return;
    }

    my $mp_device_name = $self->mp_device_name();
    my $sc_obj = $self->{sc_obj};

    # get dmpnode information
    my ($status, @output) = $sc_obj->_vxdmpadm_execute_cmd(-args => "getdmpnode dmpnodename=${mp_device_name}");
    if ($status > 0) {
        $self->_set_err("Failed to get dmp node information for $mp_device_name");
        return;
    }

    # parse out the vxdmp enclosure information
    foreach my $line (@output) {
        if ($line =~ /^\s*$mp_device_name\s+(\S+)\s+(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)$/) {
            my $state = $1;
            my $enclosure_type = $2;
            my $num_total_paths = $3;
            my $num_enabled_paths = $4;
            my $num_disabled_paths = $5;
            my $enclosure_name = $6;

            # store it on the object
            $self->set_os_specific_data(-vxdmp_enclosure_name => $enclosure_name,
                                        -vxdmp_state => $state,
                                        -vxdmp_enclosure_type => $enclosure_type,
                                        -vxdmp_total_paths => $num_total_paths,
                                        -vxdmp_num_enabled_paths => $num_enabled_paths,
                                        -vxdmp_num_disabled_paths => $num_disabled_paths);
            if ($self->error()) {
                $self->_set_err("Error setting VxDMP os specific data on object");
                return;
            }
            last;
        }
    }

    return;
}

###############################################################################
# Method:    create_partition()
# Objective: Create a partition on this device
###############################################################################
sub _create_partition {
    my ($self) = @_;

    my $sc_obj = $self->{sc_obj};

    my $netapp_vma_bin_path = $sc_obj->{NETAPP_VMA_BIN};
    my $mp_device_path = $self->mp_device_path();

    my $cmd = $netapp_vma_bin_path . "esx_remote_fdisk.pl --device_path $mp_device_path --action create";

    my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Could not fdisk $mp_device_path");
        return [];
    }

    return ["${mp_device_path}:1"];
}

###############################################################################
# Method:    unmount()
# Objective: Unmount a datastore on this device
###############################################################################
sub _unmount {
    my ($self) = @_;

    my $sc_obj = $self->{sc_obj};

    # unmount the datastore
    my $datastore_uuid = $self->get_os_specific_data("datastore_uuid");

    if (defined($datastore_uuid)) {
        # if we are running ESXi5, unmount the datastore
        if ($sc_obj->get_os_version() =~ /^(5|6)\./) {
            my $cmd = "esxcli storage filesystem unmount --volume-uuid=\"$datastore_uuid\"";
            my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                              -include_host_name => 1);
            if ($status > 0) {
                $self->_set_err("Could not unmount datastore '$datastore_uuid'");
                return;
            }
        }
    }

    return;
}

###############################################################################
# Method:    offline_device()
# Objective: Offline this device from this host.
###############################################################################
sub _offline_device {
    my ($self, %args) = @_;

    my $sc_obj = $self->{sc_obj};

    # if on ESXi5, change the state of the naa device to "off" so that
    # the ESX host handles a PDL (Permanent Device Loss) condition with this LUN
    if ($sc_obj->get_os_version() =~ /^(5|6)\./) {
        my $mp_device_name = $self->mp_device_name();
        my $cmd = "esxcli storage core device set --device=\"$mp_device_name\" --state=\"off\"";
        my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                          -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Could not offline device $mp_device_name");
            return;
        }
    } elsif ($sc_obj->get_os_version() =~ /^4\./) {
        # invoke the ESXMPDev's offline device method for esx4 since it is the same for ESX and ESXi
        $self->SUPER::_offline_device(%args);
        if ($self->error()) {
            $self->_set_err("Failed to offline device");
            return;
        }
    }

    return;
}

###############################################################################
# Method:    online_device()
# Objective: Online this device from this host.
###############################################################################
sub _online_device {
    my ($self) = @_;

    my $sc_obj = $self->{sc_obj};

    # if on ESXi5, change the state of the naa device to "on"
    if ($sc_obj->get_os_version() =~ /^(5|6)\./) {
        my $mp_device_name = $self->mp_device_name();
        my $cmd = "esxcli storage core device set --device=\"$mp_device_name\" --state=\"on\"";
        my ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd,
                                                          -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Could not online device $mp_device_name");
            return;
        }
    }

    return;
}

###############################################################################
# Method:    destroy_partition()
# Objective: Destroy the partition on this device
###############################################################################
sub _destroy_partition {
    my ($self) = @_;

    my $sc_obj = $self->{sc_obj};

    my $netapp_vma_bin_path = $sc_obj->{NETAPP_VMA_BIN};
    my $mp_device_path = $self->mp_device_path();
    my $mp_device_name = $self->mp_device_name();

    my ($status, @output, $cmd);

    # remove the datastore if it exist
    my $datastore_uuid = $self->get_os_specific_data("datastore_uuid");
    if (defined($datastore_uuid)) {
        # remove the datastore
        $cmd = $netapp_vma_bin_path . "esx_remove_datastore.pl --datastore_uuid $datastore_uuid";
        ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Could not remove the datastore '$datastore_uuid' on device '$mp_device_path'");
            return;
        }
    }

    # destroy the partition
    $cmd =  $netapp_vma_bin_path . "esx_remote_fdisk.pl --device_path $mp_device_path --action destroy";
    ($status, @output) = $sc_obj->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Could not fdisk $mp_device_path");
        return;
    }

    return;
}

###############################################################################
# Method:    list_contents
# Objective: return back a list of files inside this device
###############################################################################
sub _list_contents {
    my ($self, %args) = @_;

    my $subdir = exists($args{-subdir}) ? $args{-subdir} : undef;

    my $mp_device_name = $self->mp_device_name();
    my $datastore_label = $self->get_os_specific_data("datastore_label");
    if (not length($datastore_label)) {
        $self->_set_err("Cannot list contents of Device $mp_device_name because it is not a VMFS datastore");
        return (wantarray()) ? () : [];
    }

    my $path = "[${datastore_label}]";
    if (defined($subdir)) {
        $path .= " ${subdir}";
    }

    my $cmd = "vifs --dir '${path}'";
    my ($status, @output) = $self->{sc_obj}->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to list contents of $datastore_label");
        return (wantarray()) ? () : [];
    }

    # filter out lines that do not represent files or directories
    @output = grep {
        $_ !~ /^(\s+|Content Listing|\-+)$/
    } @output;

    return (wantarray()) ? @output : \@output;
}

1;
