#
# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary MPDevMgr base class for Linux
## @author hrivera@netapp.com
## @status private
## @pod ../MPDevMgr.pm

package San::SanClient::MPDevMgr::LinuxMPDevMgr;

use strict;
use warnings;

use NATE::Log qw(log_global);
use San::SanClient::MPDevMgr::MPDev;
use San::SanClient::MPDevMgr::MPDevPath;
use San::SanCIT;

use vars qw( $Log );

$Log = log_global();

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: Create an MPDevMgr instance
###############################################################################
sub _new {
    my ($self, %args) = @_;

    # if this client has DMMP multipathing, load the dmmp discovery script onto the system
    if ($self->{sc_obj}->{mp_type} =~ /dmmp/i) {
        $self->_load_netapp_dmmp_discovery_script();
        if ($self->error()) {
            $self->_set_err("Failed to load NetApp DMMP Discovery script", $self->errmsg());
            return $self;
        }
    }

    my $build_sg_util_map_flag;
    if (    exists($args{'-build_sg_util_map'})
        and length($build_sg_util_map_flag = $args{'-build_sg_util_map'})) {
        # set flag to build sg_util_map on $self to be used later
        $self->set_os_specific_data(-build_sg_util_map => $build_sg_util_map_flag);
        if ($self->error()) {
            $self->_set_err("Failed to set os specific " .
                            "data on $self for sg_util flag.",
                            $self->errmsg());
            return $self;
        }
    }

    return $self;
}

###############################################################################
# Method:    load_all_devices()
# Objective: Load in the current state of all device's filesystem/partition
###############################################################################
sub _load_all_devices {
    my ($self) = @_;

    # Not implemented

    return;
}

###############################################################################
# Method:    refresh()
# Objective: Rediscover all Multipathed LUNs on the system and populate the
#            Device Manager
###############################################################################
sub _refresh {
    my ($self, %args) = @_;

    my $sc_obj = $self->{sc_obj};

    # based on mp type, perform discovery
    if ($sc_obj->{mp_type} =~ /dmmp/i) {
        $self->_refresh_dmmp_devices();
    } else {

    }

    return;
}

###############################################################################
# Method:    refresh_paths()
# Objective: Rediscover all Multipathed LUNs on the system and populate the
#            Device Manager
###############################################################################
sub _refresh_paths {
    my ($self, %args) = @_;

    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh paths");
        return;
    }

    return;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _load_netapp_dmmp_discovery_script()
# Objective: Install the NetApp DMMP Discovery script on the host
###############################################################################
sub _load_netapp_dmmp_discovery_script {
    my ($self) = @_;

    # obtain the path to the p4 dir being used for this test
    my $p4_dir = &SanCIT::get_p4_dir();
    my $local_file = "${p4_dir}/SANInterop/tools/linux/netapp_dmmp_discovery";
    my $remote_file = "/sbin/netapp_dmmp_discovery";

    # transfer the dmmp discovery script to the remote host.
    if ($self->{sc_obj}->ClientUtil::scp_file(-local_file => $local_file,
                                              -remote_file => $remote_file) == $ClientUtil::FALSE
        or $self->error()) {
        $self->_set_err("SCP transfer failed", $self->errmsg());
        return;
    }

    return;
}

###############################################################################
# Method:    _refresh_dmmp_devices()
# Objective: Discover the DMMP Devices on the system
###############################################################################
sub _refresh_dmmp_devices {
    my ($self) = @_;

    # invoke the DMMP discovery script
    my $sc_obj = $self->{sc_obj};
    my $cmd = "perl /sbin/netapp_dmmp_discovery";
    my ($status, @output) = &ClientUtil::execute($sc_obj->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return;
    }

    # convert the output into device blocks
    @output = split(/\n\n/, join("\n", @output));

    # check if sg_util is installed before running sg_map
    my $build_sg_util_map_flag = $self->get_os_specific_data('build_sg_util_map');
    my ($sg_map_status, @sg_map_output);
    if ($build_sg_util_map_flag) {
        if ($sc_obj->_is_rpm_installed(-rpm => "sg3_utils")) {
            # run sg_map command to gather list of mappings between sg
            # devices and SCSI devices
            # FORMAT: /dev/sg0  /dev/sda
            $cmd = "sg_map";
            ($sg_map_status, @sg_map_output) = &ClientUtil::execute($sc_obj->{conn}, $cmd);
            if ($sg_map_status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($sg_map_status)");
                return;
            }
        } else {
            $Log->warn("Was unable to build the sg device map because sg3_utils is not installed.");
        }
    }

    # parse each device block
    foreach my $device_block (@output) {
        my @dev_block_split = split(/\n/, $device_block);

        my $mp_device_name;
        my $device_vendor;
        my $device_model;
        my $lun_id;
        my $device_revision;
        my $device_scsi_level;
        my $device_size;
        my $protocol;
        my $protocol_path_counts;
        my $lun_paths;

        foreach my $line (@dev_block_split) {
            # parse out the mp device name, device vendor, device model, device revision, scsi level
            # this will handle both user friendly and non user friendly device names
            if ($line =~ /^(\S+)\s+dm\-\d+\s+(NETAPP)\s*,\s*(LUN(?: C\-Mode)?\b)\s+LUN=(\S+)\s+REV=(\S+)\s+SCSI=(\S+)/ or
                $line =~ /^(\S+)\s+\(360\S+?\)\s+dm\-\d+\s+(NETAPP)\s*,\s*(LUN(?: C\-Mode)?\b)\s+LUN=(\S+)\s+REV=(\S+)\s+SCSI=(\S+)/) {
                $mp_device_name = $1;
                $device_vendor = $2;
                $device_model = $3;
                $lun_id = $4;
                $device_revision = $5;
                $device_scsi_level = $6;
                next;
            }

            # parse the size of the LUN
            if ($line =~ /\[?size=(\S+?)]?/) {
                $device_size = $1;
                next;
            }

            # parse the information for this path
            if ($line =~ /(?:_|\-)\s+\d+:\d+:\d+:\d+\s+(\S+)\s+\S+\s+\[?(active|ready|faulty|failed)\]?\s*\[?(active|ready|faulty|failed)\]?\s+(?:running)?\s*\[proto=(iscsi|fcp)\]\[alua=(\S+?)\]\s+(\S+)\s+\<=\>\s+(\S+)/) {
                my $raw_scsi_device_name = "/dev/" . $1;
                my $path_state = "[$2][$3]";
                my $this_paths_protocol = $4;
                my $alua_state = uc($5);
                my $target_data = $6;
                my $initiator_data = $7;

                $path_state = "dead"   if ($path_state =~ /faulty|failed/i);
                $path_state = "active" if ($path_state =~ /active|ready/i);

                my $mp_dev_path_obj;

                # handle per protocol parsing
                if ($this_paths_protocol =~ /iscsi/i) {
                    # increment this lun's path count for this protocol
                    $protocol_path_counts->{iscsi}++;

                    # parse the iscsi target information from the data string
                    my $target_iscsi_nodename = $1;
                    my $target_ip_address = $1;
                    my $target_ip_port = $1;
                    my $target_tpgt = $1;
                    if ($target_data =~ /^(\S+?),(\S+?):(\d+?),(\d+)/) {
                        $target_iscsi_nodename = $1;
                        $target_ip_address = $2;
                        $target_ip_port = $3;
                        $target_tpgt = $4;
                    }

                    # the initiator data should be just the nodename
                    my $initiator_iscsi_nodename = $initiator_data;

                    # create the MPDevPath object
                    $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $raw_scsi_device_name,
                                                                               -lun_id => $lun_id,
                                                                               -path_state => $path_state,
                                                                               -initiator_iqn => $initiator_iscsi_nodename,
                                                                               -target_iqn => $target_iscsi_nodename,
                                                                               -target_iscsi_alias => "",
                                                                               -target_iscsi_session => "",
                                                                               -target_ip_address => $target_ip_address,
                                                                               -target_ip_port => $target_ip_port,
                                                                               -tpgt => $target_tpgt,
                                                                               -alua_state => $alua_state,
                                                                               -protocol => "iscsi");
                } elsif ($this_paths_protocol =~ /fcp/i) {
                    # increment this lun's path count for this protocol
                    $protocol_path_counts->{fcp}++;

                    # parse the target and initiator wwnns and wwpns
                    my $target_fcp_nodename;
                    my $target_fcp_portname;
                    if ($target_data =~ /^(\S+?),(\S+)$/) {
                        $target_fcp_nodename = $1;
                        $target_fcp_portname = $2;
                    }
                    my $initiator_fcp_nodename;
                    my $initiator_fcp_portname;
                    if ($initiator_data =~ /^(\S+?),(\S+)$/) {
                        $initiator_fcp_nodename = $1;
                        $initiator_fcp_portname = $2;
                    }

                    # create the MPDevPath object
                    $mp_dev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $raw_scsi_device_name,
                                                                               -lun_id => $lun_id,
                                                                               -path_state => $path_state,
                                                                               -initiator_wwnn => $initiator_fcp_nodename,
                                                                               -initiator_wwpn => $initiator_fcp_portname,
                                                                               -target_wwnn => $target_fcp_nodename,
                                                                               -target_wwpn => $target_fcp_portname,
                                                                               -alua_state => $alua_state,
                                                                               -protocol => "fcp");

                }

                $protocol = $this_paths_protocol;
                $lun_paths->{$raw_scsi_device_name} = $mp_dev_path_obj;

                if (scalar(@sg_map_output)) {
                    # parse sg_map array to find line that matches
                    # $raw_scsi_device_name and extract it.
                    my @sg_device_name_array
                        = grep {/\/dev\/sg\d+\s+$raw_scsi_device_name\s*$/} @sg_map_output;
                    my $sg_device_name = '';
                    if (scalar(@sg_device_name_array)
                        and $sg_device_name_array[0] =~ /(\/dev\/sg\d+)/) {
                        $sg_device_name = $1;
                    }
                    # set_os_specific_data on the MPDevPath obj
                    $mp_dev_path_obj->set_os_specific_data(-sg_map => $sg_device_name);
                    if ($mp_dev_path_obj->error()) {
                        $self->_set_err("Failed to set os specific " .
                                        "data for MPDevPath object " .
                                        "$mp_dev_path_obj",
                                        $mp_dev_path_obj->errmsg());
                        return $self;
                    }
                }
            }
        }

        $protocol = "mixed" if (scalar(keys %{$protocol_path_counts}) > 1);

        # create the MPDev object
        my $mpdev_obj = new San::SanClient::MPDevMgr::MPDev(-sc_obj => $sc_obj,
                                                            -mp_device_name => $mp_device_name,
                                                            -mp_type => $sc_obj->{mp_type},
                                                            -device_size => $device_size,
                                                            -protocol => $protocol,
                                                            -mp_device_path => "/dev/mapper/" . $mp_device_name,
                                                            -device_vendor => $device_vendor,
                                                            -device_model => $device_model,
                                                            -device_revision => $device_revision,
                                                            -scsi_level => $device_scsi_level,
                                                            -lun_id => $lun_id,
                                                            -alua => "enabled");

        # load this devices current mount paritioned states
        $mpdev_obj->load();
        if ($mpdev_obj->error()) {
            $self->_set_err("Failed to load MPDev object for $mp_device_name", $mpdev_obj->errmsg());
            return;
        }

        # add this device to the devmgr
        $self->_add_device(-mp_device_obj => $mpdev_obj,
                           -mp_device_path_objs => [values %{$lun_paths}]);
    }
}

###############################################################################
# Method:    os_specific_validation()
# Objective: Perform OS Specific validation on all devices passed in
###############################################################################
sub _os_specific_validation {
    my ($self) = @_;

    my $dev_hash = $self->get_devices_hash();
    my $old_dev_hash = $self->get_old_devices_hash();

    # check if there is anything to validate
    if (not scalar(keys %{$dev_hash}) and scalar(keys %{$old_dev_hash})) {
        return ([], []);
    }

    my @warnings = ();
    my @errors = ();

    return (\@errors, \@warnings);
}

1;
