#
# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an OS generic interface to a SAN Multipathed Device
## @author hrivera@netapp.com
## @status private
## @pod here

=pod

=head1 Package Name

MPDev.pm

=head1 Author

dl-san-automation

=head1 Description

An OS generic multipathed device object

Current Operating System support:
 ESX4.x/ESXi4.x
 ESXi5.x
 ESXi6.x
 Redhat 5.[6-8]
 Redhat 6.x
 SLES 10

=head1 Inherited Packages

 Error
 San::SanLog

=cut

package San::SanClient::MPDevMgr::MPDev;

use strict;
use warnings;
use Scalar::Util qw(weaken);
use Data::Dumper;

use Error;
use San::SanCache;
use San::SanUtil;
use San::SanClient::MPDevMgr::MPDevPath;
use San::SanClient::MPDevMgr::MPDevPathState;
use San::SanLog;

use vars qw (@ISA);

@ISA = qw (Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

=pod

=head1 B<Public Methods>

=head2 B<new(%args)>

The constructor for MPDev. Instantiates and returns an MPDev object based
on the OS of the SanClient object.

=over 4

=item Inputs:

=over 2

=item I<-sc_obj> => San::SanClient

 Required: A San::SanClient object. This is the host to associate the object with

=item I<-mp_device_name> => scalar

 Required: The name of the multipathed device.

=item I<-mp_device_path> => scalar

 Required: The full path to the multipathed device,

=item I<-os_specific_data> => \%hash

 Optional: A hash reference of OS specific data {$key => $value, $keyN => $valueN} pairs.

=item I<-mp_device_partition_list> => \@array

 Optional: An array reference of partition paths for this device.

=item I<-device_size> => scalar

 Optional: The size of the device.

=item I<-protocol> => scalar

 Optional: The protocol of the device.

=item I<-device_vendor> => scalar

 Optional: The SCSI vendor of the device.

=item I<-device_model> => scalar

 Optional: The SCSI model of the device.

=item I<-device_revision> => scalar

 Optional: The SCSI revision of the device.

=item I<-scsi_level> => scalar

 Optional: The SCSI level of the device.

=item I<-lun_id> => scalar

 Optional: The LUN id of the device.

=item I<-alua> => scalar

 Optional: Specifies whether ALUA is enabled or disabled.
 Valid Values: "enabled" | "disabled"

=item I<-device_state> => scalar

 Optional: Specifies the device state.
 Default: "free"
 Valid Values: "free" | "used" | "no_paths"

=item I<-path_states> => \@array

 Optional: an array reference of San::SanClient::MPDevMgr::MPDevPathState objects

=item I<-lun_relationship> => scalar

 Optional: Specifies whether a LUN relationship has been created for this device
 Default: 0 (no relationship)
 Valid Values: 0 | 1

=item I<-lun_obj> => San::ControllerCluster::Filer10M::Vserver::Volume::Lun / NACL::C::Lun

 Optional: A San::ControllerCluster::Filer10M::Vserver::Volume::Lun / NACL::C::Lun object that
 is maps to this device object.
 Default: {}

=item I<-ontap_mode> => scalar

 Optional: Specifies the ONTAP mode of this device.
 Default: ""
 Valid Values: "cmode" | "7mode"

=item I<-ontap_vserver> => scalar

 Optional: Specifies the ONTAP vserver of this device.

=item I<-ontap_controller_7m> => scalar

 Optional: Specifies the ONTAP 7mode controller name of this device.

=item I<-ontap_home_node> => scalar

 Optional: Specifies ONTAP controller name of this device.

=item I<-ontap_lun_path> => scalar

 Optional: Specifies the ONTAP LUN path of this device

=item I<-ontap_lun_name> => scalar

 Optional: Specifies the ONTAP LUN name of this device

=item I<-ontap_vol_name> => scalar

 Optional: Specifies the ONTAP vol name of this device

=item I<-ontap_lun_ostype> => scalar

 Optional: Specifies the ONTAP LUN OS type of this device

=item I<-is_partitioned> => scalar

 Optional: Specifies whether this device is partitioned or not.

=item I<-is_mounted> => scalar

 Optional: Specifies whether this device is mounted or not.

=item I<-is_offline> => scalar

 Optional: Specifies whether this device is offline or not.

=item I<-has_filesystem> => scalar

 Optional: Specifies whether this device has a filesystem or not.

=item I<-fs_type> => scalar

 Optional: Specifies the filesystem type for this device.

=item I<-mount_point> => scalar

 Optional: Specifies the mount point this device is mounted on.

=back

=item Outputs:

=over 2

A San::SanClient::MPDev object

=back

=item Example:

 my $mpdev_obj = new San::SanClient::MPDevMgr::MPDev(-sc_obj => $sanclient_obj,
                                                     -mp_device_name => "mpath2",
                                                     -mp_device_path => "/dev/mapper/mpath2");

=back

=cut

sub new {
    my ($class, %args) = @_;

    my $self = {};

    bless($self, $class);

    # strip all passed in arguments
    %args = &SanUtil::strip_dash(%args);

    my $sc_obj = delete($args{sc_obj});
    my $hosttype = $sc_obj->get_connection()->{hostp}{hosttype};

    # initialize the way we handle error setting
    $self->{error_dont_overwrite_errors} = 1;

    if ($hosttype =~ /linux-vmware/i) {
        if ($sc_obj->has_vma_host()) {
            use San::SanClient::MPDevMgr::ESXiMPDevMgr::ESXiMPDev;
            @ISA = ("Error", "SanLog",
                    "San::SanClient::MPDevMgr::ESXiMPDevMgr::ESXiMPDev",
                    "San::SanClient::MPDevMgr::ESXMPDevMgr::ESXMPDev");
        } else {
            use San::SanClient::MPDevMgr::ESXMPDevMgr::ESXMPDev;
            @ISA = ("Error", "SanLog", "San::SanClient::MPDevMgr::ESXMPDevMgr::ESXMPDev");
        }
    } elsif ($hosttype =~ /unix-linux.*-citrix/i) {
        use San::SanClient::MPDevMgr::CXSMPDevMgr::CXSMPDev;
        @ISA = ("Error", "SanLog", "San::SanClient::MPDevMgr::CXSMPDevMgr::CXSMPDev");
    } elsif ($hosttype =~ /unix-linux/) {
        use San::SanClient::MPDevMgr::LinuxMPDevMgr::LinuxMPDev;
        @ISA = ("Error", "SanLog", "San::SanClient::MPDevMgr::LinuxMPDevMgr::LinuxMPDev");
    } elsif ($hosttype =~ /windows/) {
        use San::SanClient::MPDevMgr::WinMPDevMgr::WinMPDev;
        @ISA = ("Error", "SanLog", "San::SanClient::MPDevMgr::WinMPDevMgr::WinMPDev");
    } else {
        $self->_set_err("MPDev currently does not have support for hosttype $hosttype");
        return $self;
    }

    my @isa_tmp = @ISA;
    $self->{__ISA__} = \@isa_tmp;

    # store the sc_obj attribute after we weaken it
    # this prevents memory leaks with circular dependencies
    Scalar::Util::weaken($self->{sc_obj} = $sc_obj);

    # initialize attributes
    $self->{os_specific_data} = {};
    $self->{mp_device_name} = "";
    $self->{mp_device_path} = "";
    $self->{mp_device_partition_list} = [];
    $self->{device_size} = "";
    $self->{protocol} = "";
    $self->{device_vendor} = "";
    $self->{device_model} = "";
    $self->{device_revision} = "";
    $self->{scsi_level} = "";
    $self->{lun_id} = "";
    $self->{alua} = "";
    $self->{device_state} = "free";
    $self->{path_states} = [];
    $self->{lun_relationship} = 0;
    $self->{lun_obj} = {};
    $self->{ontap_mode} = "";
    $self->{ontap_vserver} = "";
    $self->{ontap_controller_7m} = "";
    $self->{ontap_home_node} = "";
    $self->{ontap_node_id} = "";
    $self->{ontap_lun_path} = "";
    $self->{ontap_lun_name} = "";
    $self->{ontap_lun_ostype} = "";
    $self->{ontap_vol_name} = "";
    $self->{is_partitioned} = 0;
    $self->{is_mounted} = 0;
    $self->{is_offline} = 0;
    $self->{has_filesystem} = 0;
    $self->{fs_type} = "";
    $self->{mount_point} = "";

    # store all the passed in arguments
    foreach (keys %args) {
        $self->{$_} = $args{$_};
    }

    $self->SUPER::_new(%args);

    return $self;
}

=pod

=head2 B<create_partition()>

Create a partition on this multipathed device

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # create a partition on this device
 $mpdev_obj->create_partition();
 if ($mpdev_obj->error()) {
    # ...
 }

=back

=cut

sub create_partition {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    if ($self->is_partitioned()) {
        my $mp_device_path = $self->mp_device_path();
        $self->_set_err("Cannot create partition on ${mp_device_path}, " .
                        "it is already partitioned.");
        return;
    }

    my $partitions = $self->SUPER::_create_partition(%args);

    if (not $self->error()) {
        # then a partition was probably created,
        # indicate that this device is partitioned
        $self->is_partitioned(1);

        # store the partition list on the object
        $self->mp_device_partition_list($partitions);
    }

    return;
}

=pod

=head2 B<destroy_partition()>

Destroy all partitions on this multipathed device

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # destroy partitions on this device
 $mpdev_obj->destroy_partition();
 if ($mpdev_obj->error()) {
    # ...
 }

=back

=cut

sub destroy_partition {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    # verify that this device is partioned before trying to destroy it
    if (not $self->is_partitioned()) {
        my $mp_device_path = $self->mp_device_path();
        $self->_set_err("Cannot destroy partition on ${mp_device_path}, " .
                        "it is not partitioned.");
        return;
    }

    $self->SUPER::_destroy_partition(%args);

    if (not $self->error()) {
        # then a partition was probably created,
        # indicate that this device is
        # no longer partitioned and does not have a fs
        $self->is_partitioned(0);
        $self->mp_device_partition_list([]);
        $self->has_filesystem(0);
        $self->is_mounted(0);
        $self->mount_point("");
    }

    return;
}

=pod

=head2 B<mkfs(%args)>

Create a filesystem on the device.

=over 4

=item Inputs:

=over 2

=item I<-fs_type> => scalar

 Optional: Specifies the file system type to create.
 Accepted values depend on the OS type.

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # create an ext3 filesystem on this device
 $mp_dev_obj->mkfs(-fs_type => "ext3");
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub mkfs {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    if ($self->has_filesystem()) {
        my $mp_device_path = $self->mp_device_path();
        $self->_set_err("Cannot create filesystem on ${mp_device_path}, " .
                        "it already has a filesystem");
        return;
    }

    $self->SUPER::_mkfs(%args);

    if (not $self->error()) {
        # then a filesystem has been laid down
        # indicate that this device has a filesystem
        $self->has_filesystem(1);
    }

    return;
}

=pod

=head2 B<mount(%args)>

Mount the filesystem on this device to a mount point..

=over 4

=item Inputs:

=over 2

=item I<-mount_point> => scalar

 Required: Specifies the mount point path to mount the device to

=item I<-create_dir> => scalar

 Optional: If set to 1, will create the mount point directory if it does not exist.
 Default: 1

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # mount this device to a mount point
 $mp_dev_obj->mount(-mount_point => "/mnt/mymountpoint");
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub mount {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    my $mount_point = $self->SUPER::_mount(%args);

    if (not $self->error()) {
        # then this device should be mounted
        # indicate that this device is now mounted
        $self->is_mounted(1);
        $self->mount_point($mount_point);
    }

    return;
}

=pod

=head2 B<unmount()>

Unmount the filesystem on this device.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # unmount this device
 $mp_dev_obj->unmount();
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub unmount {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    $self->SUPER::_unmount(%args);

    if (not $self->error()) {
        # then this device should be unmounted
        # indicate that this device is now unmounted
        $self->is_mounted(0);
        $self->mount_point("");
    }

    return;
}

=pod

=head2 B<offline_device()>

Offline this device.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Offline this device
 $mp_dev_obj->offline_device();
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub offline_device {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    $self->SUPER::_offline_device(%args);

    if (not $self->error()) {
        # then this device should be unmounted
        # indicate that this device is now unmounted
        $self->is_offlined(1);
    }

    return;
}

=pod

=head2 B<online_device()>

Online this device.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # online this device
 $mp_dev_obj->online_device();
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub online_device {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    $self->SUPER::_online_device(%args);

    if (not $self->error()) {
        # then this device should be unmounted
        # indicate that this device is now unmounted
        $self->is_offlined(0);
    }

    return;
}

=pod

=head2 B<load()>

Load the current state of this dev (mount point/partition/fs status)

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # load the state of this device into the object
 $mp_dev_obj->load();
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub load {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    $self->SUPER::_load(%args);

    return;
}

=pod

=head2 B<refresh()>

Refresh the current device attributes into the object

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # refresh all object attributes for this device
 $mp_dev_obj->refresh();
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub refresh {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    $self->SUPER::_refresh(%args);

    return;
}

=pod

=head2 B<refresh_paths()>

Similar to San::SanClient::MPDevMgr::MPDev::refresh(), but only refresh
pathing data to the object.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # refresh path data for this device
 $mp_dev_obj->refresh_paths();
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub refresh_paths {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    $self->SUPER::_refresh_paths(%args);

    return;
}

=pod

=head2 B<list_contents()>

List all files that are stored in this device's filesystem

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # list all files in this device's filesystem
 my @files = $mp_dev_obj->list_contents();
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub list_contents {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    return $self->SUPER::_list_contents(%args);
}


=pod

=head2 B<generate_lun_relationship()>

Discover the lun relationship between this device and the specified 7mode/Cmode nodes

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Generate a lun relationship for this device using the given nodes to search through
 $mp_dev_obj->generate_lun_relationship();
 if ($mp_dev_obj->error()) {
     # ...
 }

=back

=cut

sub generate_lun_relationship {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    if ($self->ontap_mode() =~ /^cmode$/i) {
        my $controller_cluster_objs = $args{-controller_cluster_objs};
        my $nacl_cluster_objs = $args{-nacl_cluster_objs};

        # try building a lun relationship using San::ControllerCluster if it exists
        if (ref($nacl_cluster_objs) eq "ARRAY" and scalar(@{$nacl_cluster_objs})) {
            $self->_generate_lun_relationship_cmode_NACL(%args);
            if ($self->error()) {
                $self->_set_err("Error encountered while generating lun relationship for MP " .
                                "device '$self->{mp_device_name}'");
                return;
            }
        }

        # if we found a relationship, then return
        if ($self->has_lun_relationship()) {
            return;
        }

        # fall back to building a lun relationship using San::ControllerCluster if it exists
        if (ref($controller_cluster_objs) eq "ARRAY" and scalar(@{$controller_cluster_objs})) {
            $self->_generate_lun_relationship_cmode_CC(%args);
            if ($self->error()) {
                $self->_set_err("Error encountered while generating lun relationship for MP " .
                                "device '$self->{mp_device_name}'");
                return;
            }
        }
    } elsif ($self->ontap_mode() =~ /^7mode$/i) {
        $self->_generate_lun_relationship_7mode(%args);
        if ($self->error()) {
            $self->_set_err("Error encountered while generating lun relationship for MP " .
                            "device '$self->{mp_device_name}'");
            return;
        }
    } else {
        $self->_set_err("Invalid ONTAP Mode : '" . $self->ontap_mode() . "'");
        return;
    }

    return;
}

###############################################################################
#                          Public Accessor/Setter Methods
###############################################################################

=pod

=head1 B<Public Accessor/Setter Methods>

=head2 B<set_os_specific_data(%args)>

Set OS specific data on the device.

=over 4

=item Inputs:

=over 2

=item I<-key> => $data

 Generic {-key => data} fields to store on the object.

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # set some generic key => value data on the object
 $mpdev_obj->set_os_specific_data(-key1 => "value1",
                                  -key2 => "value2");
 if ($mpdev_obj->error()) {
    # ...
 }

=back

=cut

sub set_os_specific_data {
    my ($self, %args) = @_;

    @ISA = @{$self->{__ISA__}};

    %args = &SanUtil::strip_dash(%args);

    foreach (keys %args) {
        $self->{os_specific_data}->{$_} = $args{$_};
    }
}

=pod

=head2 B<get_os_specific_data()>

Get OS specific data from this device.

=over 4

=item Inputs:

=over 2

=item I<$key>

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # get some generic key => value data on the object
 my $value = $mpdev_obj->get_os_specific_data("some_key");
 if ($mpdev_obj->error()) {
    # ...
 }

=back

=cut

sub get_os_specific_data {
    my ($self, $key) = @_;

    @ISA = @{$self->{__ISA__}};

    if (defined($key)) {
        return $self->{os_specific_data}->{$key};
    } else {
        return $self->{os_specific_data};
    }
}

=head2 B<mp_device_name($)>

Get/Set the 'mp_device_name' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $mp_device_name = $mpdev_obj->mp_device_name();

 # Set the value
 $mpdev_obj->mp_device_name($new_value);

=back

=cut

sub mp_device_name {
    my ($self) = @_;

    return $self->_get_or_set("mp_device_name");
}

=pod

=head2 B<mp_device_path($)>

Get/Set the 'mp_device_path' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $mp_device_path = $mpdev_obj->mp_device_path();

 # Set the value
 $mpdev_obj->mp_device_path($new_value);

=back

=cut

sub mp_device_path {
    my ($self, $val) = @_;

    return $self->_get_or_set("mp_device_path", $val);
}

=pod

=head2 B<is_partitioned($)>

Get/Set the 'is_partitioned' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $is_partitioned = $mpdev_obj->is_partitioned();

 # Set the value
 $mpdev_obj->is_partitioned($new_value);

=back

=cut

sub is_partitioned {
    my ($self, $val) = @_;

    return $self->_get_or_set("is_partitioned", $val);
}

=pod

=head2 B<is_offlined($)>

Get/Set the 'is_offlined' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $is_offlined = $mpdev_obj->is_offlined();

 # Set the value
 $mpdev_obj->is_offlined($new_value);

=back

=cut

sub is_offlined {
    my ($self, $val) = @_;

    return $self->_get_or_set("is_offlined", $val);
}

=pod

=head2 B<is_mounted($)>

Get/Set the 'is_mounted' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $is_mounted = $mpdev_obj->is_mounted();

 # Set the value
 $mpdev_obj->is_mounted($new_value);

=back

=cut

sub is_mounted {
    my ($self, $val) = @_;

    return $self->_get_or_set("is_mounted", $val);
}

=pod

=head2 B<mount_point($)>

Get/Set the 'mount_point' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $mount_point = $mpdev_obj->mount_point();

 # Set the value
 $mpdev_obj->mount_point($new_value);

=back

=cut

sub mount_point {
    my ($self, $val) = @_;

    return $self->_get_or_set("mount_point", $val);
}

=pod

=head2 B<fs_type($)>

Get/Set the 'fs_type' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $fs_type = $mpdev_obj->fs_type();

 # Set the value
 $mpdev_obj->fs_type($new_value);

=back

=cut

sub fs_type {
    my ($self, $val) = @_;

    return $self->_get_or_set("fs_type", $val);
}

=pod

=head2 B<has_filesystem($)>

Get/Set the 'has_filesystem' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $has_filesystem = $mpdev_obj->has_filesystem();

 # Set the value
 $mpdev_obj->has_filesystem($new_value);

=back

=cut

sub has_filesystem {
    my ($self, $val) = @_;

    return $self->_get_or_set("has_filesystem", $val);
}

=pod

=head2 B<mp_device_partition_list($)>

Get/Set the 'mp_device_partition_list' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $mp_device_partition_list = $mpdev_obj->mp_device_partition_list();

 # Set the value
 $mpdev_obj->mp_device_partition_list($new_value);

=back

=cut

sub mp_device_partition_list {
    my ($self, $val) = @_;

    return $self->_get_or_set("mp_device_partition_list", $val);
}

=pod

=head2 B<device_size($)>

Get/Set the 'device_size' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $device_size = $mpdev_obj->device_size();

 # Set the value
 $mpdev_obj->device_size($new_value);

=back

=cut

sub device_size {
    my ($self, $val) = @_;

    return $self->_get_or_set("device_size", $val);
}

=pod

=head2 B<protocol($)>

Get/Set the 'protocol' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $protocol = $mpdev_obj->protocol();

 # Set the value
 $mpdev_obj->protocol($new_value);

=back

=cut

sub protocol {
    my ($self, $val) = @_;

    return $self->_get_or_set("protocol", $val);
}

=pod

=head2 B<device_vendor($)>

Get/Set the 'device_vendor' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $device_vendor = $mpdev_obj->device_vendor();

 # Set the value
 $mpdev_obj->device_vendor($new_value);

=back

=cut

sub device_vendor {
    my ($self, $val) = @_;

    return $self->_get_or_set("device_vendor", $val);
}

=pod

=head2 B<device_model($)>

Get/Set the 'device_model' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $device_model = $mpdev_obj->device_model();

 # Set the value
 $mpdev_obj->device_model($new_value);

=back

=cut

sub device_model {
    my ($self, $val) = @_;

    return $self->_get_or_set("device_model", $val);
}

=pod

=head2 B<device_revision($)>

Get/Set the 'device_revision' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $device_revision = $mpdev_obj->device_revision();

 # Set the value
 $mpdev_obj->device_revision($new_value);

=back

=cut

sub device_revision {
    my ($self, $val) = @_;

    return $self->_get_or_set("device_revision", $val);
}

=pod

=head2 B<device_scsi_level($)>

Get/Set the 'device_scsi_level' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $device_scsi_level = $mpdev_obj->device_scsi_level();

 # Set the value
 $mpdev_obj->device_scsi_level($new_value);

=back

=cut

sub device_scsi_level {
    my ($self, $val) = @_;

    return $self->_get_or_set("scsi_level", $val);
}

=pod

=head2 B<lun_id($)>

Get/Set the 'lun_id' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $lun_id = $mpdev_obj->lun_id();

 # Set the value
 $mpdev_obj->lun_id($new_value);

=back

=cut

sub lun_id {
    my ($self, $val) = @_;

    return $self->_get_or_set("lun_id", $val);
}

=pod

=head2 B<alua($)>

Get/Set the 'alua' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $alua = $mpdev_obj->alua();

 # Set the value
 $mpdev_obj->alua($new_value);

=back

=cut

sub alua {
    my ($self, $val) = @_;

    return $self->_get_or_set("alua", $val);
}

=pod

=head2 B<mp_type($)>

Get/Set the 'mp_type' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $mp_type = $mpdev_obj->mp_type();

 # Set the value
 $mpdev_obj->mp_type($new_value);

=back

=cut

sub mp_type {
    my ($self, $val) = @_;

    return $self->_get_or_set("mp_type", $val);
}

=pod

=head2 B<ontap_mode($)>

Get/Set the 'ontap_mode' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_mode = $mpdev_obj->ontap_mode();

 # Set the value
 $mpdev_obj->ontap_mode($new_value);

=back

=cut

sub ontap_mode {
    my ($self) = @_;

    my $val;

    # dynamically determine the ontap mode based on device model
    if ($self->device_model() =~ /LUN C\-mode/i) {
        $val = "cmode";
    } elsif ($self->device_model() =~ /LUN/i) {
        $val = "7mode";
    }

    return $self->_get_or_set("ontap_mode", $val);
}

=pod

=head2 B<ontap_lun_name($)>

Get/Set the 'ontap_lun_name' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_lun_name = $mpdev_obj->ontap_lun_name();

 # Set the value
 $mpdev_obj->ontap_lun_name($new_value);

=back

=cut

sub ontap_lun_name {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_lun_name", $val);
}

=pod

=head2 B<ontap_vol_name($)>

Get/Set the 'ontap_vol_name' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_vol_name = $mpdev_obj->ontap_vol_name();

 # Set the value
 $mpdev_obj->ontap_vol_name($new_value);

=back

=cut

sub ontap_vol_name {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_vol_name", $val);
}

=head2 B<ontap_aggregate_name($)>

Get/Set the 'ontap_aggregate_name' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_aggregate_name = $mpdev_obj->ontap_aggregate_name();

 # Set the value
 $mpdev_obj->ontap_aggregate_name($new_value);

=back

=cut

sub ontap_aggregate_name {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_aggregate_name", $val);
}

=pod

=head2 B<ontap_lun_path($)>

Get/Set the 'ontap_lun_path' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_lun_path = $mpdev_obj->ontap_lun_path();

 # Set the value
 $mpdev_obj->ontap_lun_path($new_value);

=back

=cut

sub ontap_lun_path {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_lun_path", $val);
}

=pod

=head2 B<ontap_lun_ostype($)>

Get/Set the 'ontap_lun_ostype' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_lun_ostype = $mpdev_obj->ontap_lun_ostype();

 # Set the value
 $mpdev_obj->ontap_lun_ostype($new_value);

=back

=cut

sub ontap_lun_ostype {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_lun_ostype", $val);
}

=pod

=head2 B<ontap_vserver($)>

Get/Set the 'ontap_vserver' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_vserver = $mpdev_obj->ontap_vserver();

 # Set the value
 $mpdev_obj->ontap_vserver($new_value);

=back

=cut

sub ontap_vserver {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_vserver", $val);
}

=pod

=head2 B<ontap_controller_7m($)>

Get/Set the 'ontap_controller_7m' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_controller_7m = $mpdev_obj->ontap_controller_7m();

 # Set the value
 $mpdev_obj->ontap_controller_7m($new_value);

=back

=cut

sub ontap_controller_7m {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_controller_7m", $val);
}

=pod

=head2 B<ontap_controller($)>

Get/Set the 'ontap_controller' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_controller = $mpdev_obj->ontap_controller();

 # Set the value
 $mpdev_obj->ontap_controller($new_value);

=back

=cut

sub ontap_controller {
    my ($self) = @_;

    if (defined($self->ontap_vserver()) and
        $self->ontap_vserver() ne "") {
        return $self->ontap_vserver();
    } elsif (defined($self->ontap_controller_7m()) and
             $self->ontap_controller_7m() ne "") {
        return $self->ontap_controller_7m();
    } else {
        return "";
    }
}

=pod

=head2 B<ontap_home_node($)>

Get/Set the 'ontap_home_node' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_home_node = $mpdev_obj->ontap_home_node();

 # Set the value
 $mpdev_obj->ontap_home_node($new_value);

=back

=cut

sub ontap_home_node {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_home_node", $val);
}

=pod

=head2 B<ontap_home_node_hapartner($)>

Get/Set the 'ontap_home_node_hapartner' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_home_node_hapartner = $mpdev_obj->ontap_home_node_hapartner();

 # Set the value
 $mpdev_obj->ontap_home_node_hapartner($new_value);

=back

=cut

sub ontap_home_node_hapartner {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_home_node_hapartner", $val);
}

=pod

=head2 B<ontap_node_id($)>

Get/Set the 'ontap_node_id' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_node_id = $mpdev_obj->ontap_node_id();

 # Set the value
 $mpdev_obj->ontap_node_id($new_value);

=back

=cut

sub ontap_node_id {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_node_id", $val);
}

=pod

=head2 B<device_state($)>

Get/Set the 'device_state' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $device_state = $mpdev_obj->device_state();

 # Set the value
 $mpdev_obj->device_state($new_value);

=back

=cut

sub device_state {
    my ($self, $val) = @_;

    return $self->_get_or_set("device_state", $val);
}

=pod

=head2 B<lun_obj($)>

Get/Set the 'lun_obj' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $lun_obj = $mpdev_obj->lun_obj();

 # Set the value
 $mpdev_obj->lun_obj($new_value);

=back

=cut

sub lun_obj {
    my ($self, $val) = @_;

    # if this is an assignment operation, clean up the object
    if (defined($val)) {
        # if the value is a nacl object and it has a blessed command interface
        # remove the command interface
        if (ref($val) =~ /NACL/i and ref($val->{command_interface})) {
            # reassign the command interface to be the name of the command interface so that it can be reassigned at a later time if needed
            my $ci = delete($val->{command_interface});
            $val->{_command_interface_name} = $ci->name();
        } elsif (ref($val) =~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/i) {
            # this might be a controller cluster Lun object, clean up all the unneeded fields
            delete($val->{parent_obj});
            delete($val->{conn_obj});
            delete($val->{cmd_obj});
            delete($val->{fields});
            delete($val->{build_info});
        }
    }

    return $self->_get_or_set("lun_obj", $val);
}


=pod

=head2 B<set_lun_relationship(%args)>

Set the lun relationship attribute

=over 4

=item Inputs:

=over 2

=item I<-lun_obj> => San::ControllerCluster::Filer10M::Vserver::Volume::Lun

 Optional: The lun object to associate the lun with

=item I<-state> => scalar

 Optional: The lun relationship state  (0 or 1)
 Default: 0 (no relationship)

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # set the lun relationship
 $mpdev_obj->set_lun_relationship(-lun_obj => $lun_obj,
                                  -state   => 1);

=back

=cut

sub set_lun_relationship {
    my ($self, %args) = @_;

    my $lun_obj = exists($args{-lun_obj}) ? $args{-lun_obj} : undef;
    my $state = exists($args{-state}) ? $args{-state} : 0;

    # update the lun object
    $self->lun_obj($lun_obj);

    # set the lun relationship state
    $self->_get_or_set("lun_relationship", $state);

    return;
}

=pod

=head2 B<has_lun_relationship()>

Check if this device has a LUN relationship.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

0 (does not have relationship) or 1 (has relationship)

=back

=item Example:

 # check if this device has a lun relationship
 if ($mpdev_obj->has_lun_relationship() == 1) {
     ...
 } else {
     ...
 }

=back

=cut

sub has_lun_relationship {
    my ($self) = @_;

    return $self->{lun_relationship};
}

=pod

=head2 B<get_current_path_state()>

Get the most recently discovered MPDevPathState object for this device

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

San::SanClient::MPDevMgr::MPDevPathState object

=back

=item Example:

 # get the current path state
 my $mpdevpathstate_obj = $mpdev_obj->get_current_path_state();
 if ($mpdev_obj->error()) {
     ...
 }

=back

=cut

sub get_current_path_state {
    my ($self) = @_;

    if (ref($self->{path_states}) eq "ARRAY" and defined($self->{path_states}->[0])) {
        return $self->{path_states}->[0];
    } else {
        $self->_set_err("There are no path states available for device " . $self->mp_device_name());
        return undef;
    }
}

=pod

=head2 B<get_previous_path_state()>

Get the previously discovered MPDevPathState object for this device

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

San::SanClient::MPDevMgr::MPDevPathState object

=back

=item Example:

 # get the previous path state
 my $mpdevpathstate_obj = $mpdev_obj->get_previous_path_state();
 if ($mpdev_obj->error()) {
     ...
 }

=back

=cut

sub get_previous_path_state {
    my ($self) = @_;

    if (scalar(@{$self->{path_states}}) > 1) {
        return $self->{path_states}->[1];
    } else {
        return $self->{path_states}->[0];
    }
}

=pod

=head2 B<get_path_states()>

Get all path state objects.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Array reference of San::SanClient::MPDevMgr::MPDevPathState objects

=back

=item Example:

 # get all path state objects (most recent first)
 my @mpdevpathstate_objs = $mpdev_obj->get_path_states();
 if ($mpdev_obj->error()) {
     ...
 }

=back

=cut

sub get_path_states {
    my ($self) = @_;

    return (wantarray() ? @{$self->{path_states}} : $self->{path_states});
}

=pod

=head2 B<compare_path_states()>

Compare the current and previous path state object.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Returns 'undef' if there is no difference, returns a hash of San::SanClient::MPDevMgr::MPDevPath objects
keyed by path name if there is a difference. The objects in the hash represent the paths that are different.

=back

=item Example:

 # compare
 my $hash = $mpdev_obj->compare_path_states();
 if (defined($hash)) {
     # difference
 } else {
     # no difference
 }

=back

=cut

sub compare_path_states {
    my ($self) = @_;

    my $curr_path_state = $self->get_current_path_state();
    my $prev_path_state = $self->get_previous_path_state();

    my $diff_path_objs_hash = undef;

    my $curr_path_objs_hash = {%{$curr_path_state->get_path_objs_hash()}};
    my $prev_path_objs_hash = {%{$prev_path_state->get_path_objs_hash()}};

    require Data::Compare;

    foreach my $path_key (keys %{$curr_path_objs_hash}) {
        my $curr_path_obj = delete $curr_path_objs_hash->{$path_key};
        my $prev_path_obj = delete $prev_path_objs_hash->{$path_key};

        if (not Compare($curr_path_obj, $prev_path_obj)) {
            $diff_path_objs_hash->{$path_key} = $curr_path_obj;
        }
    }

    return $diff_path_objs_hash;
}

=pod

=head2 B<find_missing_or_dead_paths()>

Compare the current and previous path state object and return an array
of errors found.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Returns an array of any error strings found during the comparison.

=back

=item Example:

 # find all missing or dead paths
 if (my @errors = $mpdev_obj->find_missing_or_dead_paths()) {
    # oh man something went wrong with this device
 } else {
    # all is good in the world
 }

=back

=cut

sub find_missing_or_dead_paths {
    my ($self) = @_;

    my $curr_path_state = $self->get_current_path_state();
    my $prev_path_state = $self->get_previous_path_state();
    if (not defined($curr_path_state) or not defined($prev_path_state)) {
        return ();
    }

    my @errors = ();

    # find missing paths
    if ($curr_path_state->total_paths() ne $prev_path_state->total_paths()) {
        push(@errors, "Total path count is " . $curr_path_state->total_paths() .
                      ", it should be " . $prev_path_state->total_paths());
    }

    # make sure number of alive paths is the same
    if ($curr_path_state->total_alive_paths() ne $prev_path_state->total_alive_paths()) {
        push(@errors, "Total alive path count is " . $curr_path_state->total_alive_paths() .
                      ", it should be " . $prev_path_state->total_alive_paths());
    }

    # make sure number of dead paths is the same
    if ($curr_path_state->total_dead_paths() ne $prev_path_state->total_dead_paths()) {
        push(@errors, "Total dead path count is " . $curr_path_state->total_dead_paths() .
                      ", it should be " . $prev_path_state->total_dead_paths());
    }

    return @errors;
}

=pod

=head2 B<mask_lun_paths>

Mask the paths for this MPDev object

=over 4

=item Inputs:

=over 2

None

=back

item Outputs:

=over 2

None

=back

=item Example:

 #Mask all paths to this device
 $mpdev_obj->mask_lun_paths();

=back

=cut

sub mask_lun_paths {
    my ($self) = @_;
    return $self->SUPER::_mask_lun_paths();
}

=head2 B<unmask_lun_paths>

Unmasks all paths to this device.

=over 4

=item Inputs:

=over 2

None

=back

item Outputs:

=over 2

None

=back

=item Example:

 #unmask all paths to this device
 $mpdev_obj->unmask_lun_paths();

=back

=cut

sub unmask_lun_paths {
    my ($self) = @_;
    return $self->SUPER::_unmask_lun_paths();
}

=head2 B<generate_device_reports>

Generate a device report for this MPDev object contained path information
and LUN information.

=over 4

=item Inputs:

=over 2

None

=back

item Outputs:

=over 2

None

=back

=item Example:

 # return back a pretty device report and print it
 my $report = $mpdev_obj->generate_device_report();
 use NATE::Log;
 NATE::Log::log_global()->comment(@{$report});

=back

=cut

sub generate_device_report {
    my ($self) = @_;

    my @report;

    if ($self->has_lun_relationship()) {
        if ($self->ontap_mode() =~ /cmode/i) {
            push(@report, "MP Device: " . $self->mp_device_name() . " (protocol=" . $self->protocol() . ",ontap_mode=C) [" .
                          $self->ontap_vserver() . ": " . $self->ontap_lun_path() . "]");
        } elsif ($self->ontap_mode() =~ /7mode/i) {
            push(@report, "MP Device: " . $self->mp_device_name() . " (protocol=" . $self->protocol() . ",ontap_mode=7) [" .
                          $self->ontap_controller_7m() . ": " . $self->ontap_lun_path() . "]");
        }
    } else {
        push(@report, "MP Device: " . $self->mp_device_name() . " (protocol=" . $self->protocol() . ")");
    }

    # display its Path info
    my $path_state = $self->get_current_path_state();
    push(@report, "\tTotal Paths       : " . $path_state->total_paths() .
                  " (iscsi=" . $path_state->total_iscsi_paths() .
                  ",fcp=" . $path_state->total_fcp_paths() . ")");
    push(@report, "\tTotal Alive Paths : " . $path_state->total_alive_paths() .
                  " (iscsi=" . $path_state->total_iscsi_alive_paths() .
                  ",fcp=" . $path_state->total_fcp_alive_paths() . ")");
    push(@report, "\tTotal Dead Paths  : " . $path_state->total_dead_paths() .
                  " (iscsi=" . $path_state->total_iscsi_dead_paths() .
                  ",fcp=" . $path_state->total_fcp_dead_paths() . ")");
    push(@report, "\tTotal AO Paths    : " . scalar(@{$path_state->get_filtered_path_objs(-alua_state => "AO")}));
    push(@report, "\tTotal ANO Paths   : " . scalar(@{$path_state->get_filtered_path_objs(-alua_state => "ANO")}));

    return \@report;
}

###############################################################################
#                                Private Methods
###############################################################################

###############################################################################
# Method:    _get_or_set()
# Objective: Generic getter/setting method
###############################################################################
sub _get_or_set {
    my ($self, $attr, $val) = @_;

    if (defined($val)) {
        $self->{$attr} = $val;
    }

    return exists($self->{$attr}) ? $self->{$attr} : "";
}

###############################################################################
# Method:    _generate_lun_relationship_cmode_NACL()
# Objective: Generate a lun relationship using NACL APIs
###############################################################################
sub _generate_lun_relationship_cmode_NACL {
    my ($self, %args) = @_;

    my $nacl_cluster_objs = $args{-nacl_cluster_objs};
    my $cluster_content = $args{-nacl_cluster_content};

    my $dev_discovered = 0;

    foreach my $mpdevpath_obj ($self->get_current_path_state()->get_path_objs()) {
        # discover the lun path/vserver/igroup name for this path
        # if it's iSCSI or FCP

        # skip discovery on this path if it is dead
        #next if ($mpdevpath_obj->path_state() =~ /dead/);

        if ($mpdevpath_obj->protocol() eq "iscsi") {
            my $target_iqn = $mpdevpath_obj->target_iqn();
            my $init_iqn = $mpdevpath_obj->initiator_iqn();
            my $lun_id = $self->lun_id();
            my $vserver_name = $self->ontap_vserver(); #xxx

            # obtain the vserver name that owns this target iqn
            my @vserver_iscsi_show_data = @{$cluster_content->{vserver_iscsi_objs}};

            # skip this path if we don't know the nodename for this path's target or we don't have any info about
            # the vserver name
            next if ((not $target_iqn or $target_iqn =~ /Unavailable/i)
                      and (not $vserver_name)); #xxx

            my ($vserver_iscsi_show) = grep {
                if ($target_iqn and $_->target_name() eq $target_iqn) {
                    1;
                } elsif ($vserver_name and $_->vserver() eq $vserver_name) { #xxx
                    1;
                } else {
                    0;
                }
            } @vserver_iscsi_show_data;

            # skip this path if there is no vserver that owns this target iqn nodename
            next if (not defined($vserver_iscsi_show));

            # we now know the vserver name that owns this iqn
            $vserver_name = $vserver_iscsi_show->vserver();

            # we also are certain of the target_iqn field, update it just in case we did't have it before
            $mpdevpath_obj->target_iqn($vserver_iscsi_show->target_name()); #xxx

            # find the igroup that this path is talking to
            my @vserver_iscsi_initiator_show_data = @{$cluster_content->{vserver_iscsi_init_objs}};

            my ($vserver_iscsi_initiator_show) = grep {
                if ($_->initiator_name() eq $init_iqn and
                    $_->vserver()        eq $vserver_name) {
                    1;
                } else {
                    0;
                }
            } @vserver_iscsi_initiator_show_data;

            # skip this path if we don't have any iscsi initiator show data
            next if (not defined($vserver_iscsi_initiator_show));

            # we now know the name of the igroups that this initiator is connected to.
            my @igroup_names = $vserver_iscsi_initiator_show->igroup();

            # we now need to find the lun that is mapped to these igroups
            # search for the lun that is mapped to this igroup at this lun id
            my @lun_mapped_show_data = @{$cluster_content->{lun_mapped_objs}};

            my $igroup_name;

            my @igroups_final;
            my ($lun_mapped_show) = grep {
                my $lun_mapped_show_elm = $_;

                # if this lun map entry matches our lun id/vserver and initiator wwpn...
                if ($lun_mapped_show_elm->lun_id()   eq $lun_id and
                    $lun_mapped_show_elm->vserver()  eq $vserver_name and
                    grep(/$init_iqn/, $lun_mapped_show_elm->initiators())) {
                    # filter the list of igroups this lun is mapped to
                    @igroups_final = grep {
                        if ($_ eq $lun_mapped_show_elm->igroup()) {
                            1;
                        } else {
                            0;
                        }
                    } @igroup_names;

                    # if there are no igroups, then this lun does not belong to this MPDev
                    if (not scalar(@igroups_final)) {
                        0;
                    } else {
                        1;
                    }
                } else {
                    0;
                }
            } @lun_mapped_show_data;

            # skip this path if we don't have any lun mapped show data
            next if (not defined($lun_mapped_show));

            # we now know the lun path to the path to this LUN
            my $lun_path = $lun_mapped_show->path();

            $self->{ontap_lun_path} = $lun_path;
            $self->{ontap_vserver} = $vserver_name;
            foreach my $igroup_name (@igroups_final) {
                $self->{ontap_igroups}->{$igroup_name}++;
            }

            # discover the lif object for this path
            my $lif_name = $mpdevpath_obj->ontap_target_port_name();
            my $lif_obj;
            if (not $lif_name) {
                # discover the lif for this path using the tpgt() from the path
                # if we dont know the lif name yet
                my @vserver_iscsi_tpgroup_show_data = @{$cluster_content->{vserver_iscsi_tpgroup_objs}};

                # find the lif that owns the tpgt for this path
                my $tpgt = $mpdevpath_obj->tpgt();
                my ($vserver_iscsi_tpgroup_show) = grep {
                    if ($_->tpgroup_tag() eq $tpgt) {
                        1;
                    } else {
                        0;
                    }
                } @vserver_iscsi_tpgroup_show_data;

                # if we found a tpgroup show match, extract the lif name
                if (defined($vserver_iscsi_tpgroup_show)) {
                    ($lif_name) = $vserver_iscsi_tpgroup_show->lif(); # possible that this array could contain multiples
                }
            }

            # find the lif object that is owned by this vserver and that shares
            # the name we just discovered for this path
            my @lif_objs = @{$cluster_content->{lif_objs}};
            ($lif_obj) = grep {
                if ($_->lif()     eq $lif_name and
                    $_->vserver() eq $vserver_name) {
                    1;
                } else {
                    0;
                }
            } @lif_objs;

            # store the lif object on this path if it's defined
            if (defined($lif_obj)) {
                $mpdevpath_obj->set_lif_relationship(-state => $SanUtil::TRUE,
                                                     -lif_obj => $lif_obj);
            }
            $dev_discovered = 1;
        } elsif ($mpdevpath_obj->protocol() eq "fcp") {
            my $target_wwnn  = $mpdevpath_obj->target_wwnn();
            my $target_wwpn  = $mpdevpath_obj->target_wwpn();
            my $init_wwpn    = $mpdevpath_obj->initiator_wwpn();
            my $lif_name     = $mpdevpath_obj->ontap_target_port_name();
            my $lun_id       = $self->lun_id();
            my $vserver_name = $self->ontap_vserver();

            # skip this path if we don't know the wwnn for this path's target
            next if ((not $target_wwnn or $target_wwnn =~ /Unavailable/i) and
                    (not $vserver_name));

            # obtain the vserver name that owns this target wwnn
            my @vserver_fcp_show_data = @{$cluster_content->{vserver_fcp_objs}};

            # normalize the target_wwnn
            if ($target_wwnn) {
                $target_wwnn = $self->_normalize_fc_name($target_wwnn);
            }

            my ($vserver_fcp_show) = grep {
                # normalize the wwnn in the ontap output
                my $t_wwnn = $_->target_name();
                $t_wwnn = $self->_normalize_fc_name($t_wwnn);
                if ($t_wwnn eq $target_wwnn) {
                    1;
                } elsif ($_->vserver() eq $vserver_name) {
                    1;
                } else {
                    0;
                }
            } @vserver_fcp_show_data;

            # if there is no vserver that owns this target wwnn, skip this path
            next if (not defined($vserver_fcp_show));

            # we now know the vserver name that owns this wwpn
            $vserver_name = $vserver_fcp_show->vserver();
            $self->ontap_vserver($vserver_name);

            # reassign the target_wwnn into the path that we just found
            $mpdevpath_obj->target_wwnn($self->_normalize_fc_name($vserver_fcp_show->target_name()));

            # if we dont have the lun path already, determine it via the initiator wwpn
            # and the vserver name
            if (not $self->ontap_lun_path()) {
                # find the igroup that this path is talking to
                my @vserver_fcp_initiator_show_data = @{$cluster_content->{vserver_fcp_init_objs}};

                # normalize the wwpns
                $init_wwpn = $self->_normalize_fc_name($init_wwpn);

                my ($vserver_fcp_initiator_show) = grep {
                    # normalize the wwpn in the ontap output
                    my $i_wwpn = $_->wwpn();
                    my $vserver = $_->vserver();

                    $i_wwpn = $self->_normalize_fc_name($i_wwpn);

                    # match the values
                    if ($i_wwpn  eq $init_wwpn and
                        $vserver eq $vserver_name) {
                        1;
                    } else {
                        0;
                    }
                } @vserver_fcp_initiator_show_data;

                # if there is no vserver fcp init data, skip
                next if (not defined($vserver_fcp_initiator_show));

                # we now know the name of the igroups that this initiator is connected to.
                my @igroup_names = $vserver_fcp_initiator_show->igroup();

                # we now need to find the lun that is mapped to these igroups
                # search for the lun that is mapped to this igroup at this lun id
                my @lun_mapped_show_data = @{$cluster_content->{lun_mapped_objs}};

                my $igroup_name;

                my @igroups_final;
                my ($lun_mapped_show) = grep {
                    my $lun_mapped_show_elm = $_;

                    my @inits = $lun_mapped_show_elm->initiators();
                    @inits = map { $self->_normalize_fc_name($_); } @inits;

                    # if this lun map entry matches our lun id/vserver and initiator wwpn...
                    if ($lun_mapped_show_elm->lun_id()  eq $lun_id and
                        $lun_mapped_show_elm->vserver() eq $vserver_name and
                        grep(/$init_wwpn/, @inits)) {
                        # filter the list of igroups this lun is mapped to
                        @igroups_final = grep {
                            if ($_ eq $lun_mapped_show_elm->igroup()) {
                                1;
                            } else {
                                0;
                            }
                        } @igroup_names;

                        # if there are no igroups, then this lun does not belong to this MPDev
                        if (not scalar(@igroups_final)) {
                            0;
                        } else {
                            1;
                        }
                    } else {
                        0;
                    }
                } @lun_mapped_show_data;

                # skip this path if we don't have any lun mapped show data
                next if (not defined($lun_mapped_show));

                # store the igroups hash on the mpdev obj
                foreach my $igroup_name (@igroups_final) {
                    $self->{ontap_igroups}->{$igroup_name}++;
                }

                # we now know the lun path to the path to this LUN
                my $lun_path = $lun_mapped_show->path();

                # store the lun path and vserver on the mpdev obj
                $self->ontap_lun_path($lun_path);
            }

            # discover the lif object that is tied to this path
            my @lif_objs = @{$cluster_content->{lif_objs}};
            my ($lif_obj) = grep {
                my $vs_name = $_->vserver();
                my $wwpn = $_->wwpn();
                my $lif = $_->lif();

                # first check if we know the lif name, if we do
                # then just search for that
                if ($vs_name eq $vserver_name and
                    $lif eq $lif_name) {
                    1;
                } else {
                    # otherwise try to find a wwpn match
                    # normalize the wwpns first
                    $wwpn = $self->_normalize_fc_name($wwpn);
                    $target_wwpn = $self->_normalize_fc_name($target_wwpn);

                    if ($wwpn eq $target_wwpn and
                            $vs_name eq $vserver_name) {
                        1;
                    } else {
                        0;
                    }
                }
            } @lif_objs;

            # store the lif object on this path if it's defined
            if (defined($lif_obj)) {
                $mpdevpath_obj->set_lif_relationship(-state => $SanUtil::TRUE,
                                                     -lif_obj => $lif_obj);
            }

            $dev_discovered = 1;
        }
    }

    # if this device had at least one path discovered, then we have
    # enough data to discover the lun information
    if ($dev_discovered == 1) {

        # if we do not know the lun path at this point then return
        if (not defined($self->{ontap_lun_path})) {
            return;
        }

        # parse the lun, vol, qtree fields
        if ($self->{ontap_lun_path} =~ /^\/vol\/(.+?)\/(.+?)\/(.+?)$/) {
            $self->{ontap_vol_name} = $1;
            $self->{ontap_qtree_name} = $2;
            $self->{ontap_lun_name} = $3;
        } elsif ($self->{ontap_lun_path} =~ /^\/vol\/(.+?)\/(.+?)$/) {
            $self->{ontap_vol_name} = $1;
            $self->{ontap_qtree_name} = "\"\"";
            $self->{ontap_lun_name} = $2;
        } else {
            $self->_set_err("Could not parse ONTAP LUN path '" . $self->{ontap_lun_path} . "'");
            return;
        }

        my @vol_objs = @{$cluster_content->{vol_objs}};
        my @aggr_objs = @{$cluster_content->{aggr_objs}};

        # grab the vol object that this device lives in
        my ($vol_obj) = grep {
            if ($_->volume()  eq $self->{ontap_vol_name} and
                $_->vserver() eq $self->{ontap_vserver}) {
                1;
            } else {
                0;
            }
        } @vol_objs;

        # fail if we can't find the vol object for this device
        if (not defined($vol_obj)) {
            $self->_set_err("Vol object does not exist for " . $self->{ontap_vserver} .
                            ":/vol/" . $self->{ontap_vol_name});
            return;
        }

        # grab the aggregate object that this vol lives in
        my $aggr_name = $vol_obj->aggregate();

        # store the aggr name on the object
        $self->{ontap_aggregate_name} = $aggr_name;
        my ($aggr_obj) = grep {
            if ($_->aggregate() eq $aggr_name) {
                1;
            } else {
                0;
            }
        } @aggr_objs;

        # fail if we can't find the aggr object for this volume
        if (not defined($aggr_obj)) {
            $self->_set_err("Aggregate object does not exist for aggr '$aggr_name'");
            return;
        }

        # determine which node this aggr lives on
        # aggrs dont necessarily span across nodes (yet)
        my @node_list = $aggr_obj->nodes();

        # store the node list on the object
        $self->ontap_home_node($node_list[0]);

        # determine the hapartner for this node
        my @sfo_show_data = @{$cluster_content->{sfo_show_objs}};
        my ($sfo_show) = grep {
            if ($_->node() eq $node_list[0]) {
                1;
            } else {
                0;
            }
        } @sfo_show_data;

        if (defined($sfo_show)) {
            # store the ha partner on the object
            $self->ontap_home_node_hapartner($sfo_show->partner_name());
        } else {
            $self->ontap_home_node_hapartner("unknown");
        }

        # and finally, attach the lun object to the device
        my @lun_objs = @{$cluster_content->{lun_objs}};

        # parse out the lun object that matches the filters
        my ($lun_obj) = grep {
            if ($_->lun() eq $self->{ontap_lun_name} and
                $_->volume() eq $self->{ontap_vol_name} and
                $_->qtree() eq $self->{ontap_qtree_name} and
                $_->vserver() eq $self->{ontap_vserver}) {
                1;
            } else {
                0;
            }
        } @lun_objs;

        if (not defined($lun_obj)) {
            $self->_set_err("Lun object does not exist for " .
                            "vserver : '" . $self->{ontap_vserver} . "' " .
                            "volume : '" . $self->{ontap_vol_name} . "' " .
                            "qtree : '" . $self->{ontap_qtree_name} . "' " .
                            "lun : '" . $self->{ontap_lun_name} . "'");
            return;
        }

        # establish the lun relationship to this device
        $self->set_lun_relationship(-state => $SanUtil::TRUE,
                                    -lun_obj => $lun_obj);

        # store the ostype of the lun on this device
        $self->ontap_lun_ostype($lun_obj->ostype());
    }

    return;
}

###############################################################################
# Method:    _generate_lun_relationship_cmode_CC()
# Objective: Generate the LUN relationship for cmode devices using CC APIs
###############################################################################
sub _generate_lun_relationship_cmode_CC {
    my ($self, %args) = @_;

    my $controller_cluster_objs = $args{-controller_cluster_objs};
    my $cluster_content = $args{-cluster_content};

    my $dev_discovered = 0;

    foreach my $mpdevpath_obj ($self->get_current_path_state()->get_path_objs()) {
        # discover the lun path/vserver/igroup name for this path
        # if it's iSCSI or FCP

        # skip discovery on this path if it is dead
        #next if ($mpdevpath_obj->path_state() =~ /dead/);

        if ($mpdevpath_obj->protocol() eq "iscsi") {
            my $target_iqn = $mpdevpath_obj->target_iqn();
            my $init_iqn = $mpdevpath_obj->initiator_iqn();
            my $lun_id = $mpdevpath_obj->lun_id();

            # skip this path if we don't know the nodename for this path's target
            next if ($target_iqn eq "" or $target_iqn =~ /Unavailable/i);

            # obtain the vserver name that owns this target iqn
            my @vserver_iscsi_show_data = @{$cluster_content->{vserver_iscsi_show_data}};

            my ($vserver_iscsi_show) = grep {
                if ($_->{target_name} eq $target_iqn) {
                    1;
                } else {
                    0;
                }
            } @vserver_iscsi_show_data;

            # skip this path if there is no vserver that owns this target iqn nodename
            next if (not defined($vserver_iscsi_show));

            # we now know the vserver name that owns this iqn
            my $vserver_name = $vserver_iscsi_show->{vserver};

            # find the igroup that this path is talking to
            my @vserver_iscsi_initiator_show_data = @{$cluster_content->{vserver_iscsi_initiator_show_data}};

            my ($vserver_iscsi_initiator_show) = grep {
                if ($_->{initiator_name} eq $init_iqn and
                    $_->{vserver}        eq $vserver_name) {
                    1;
                } else {
                    0;
                }
            } @vserver_iscsi_initiator_show_data;

            # we now know the name of the igroups that this initiator is connected to.
            my $igroup_name_set = $vserver_iscsi_initiator_show->{igroup};
            my @igroup_names = split(/,/, $igroup_name_set);

            # we now need to find the lun that is mapped to these igroups
            # search for the lun that is mapped to this igroup at this lun id
            my @lun_mapped_show_data = @{$cluster_content->{lun_mapped_show_data}};

            my $igroup_name;

            my @igroups_final;
            my ($lun_mapped_show) = grep {
                my $lun_mapped_show_elm = $_;

                # if this lun map entry matches our lun id/vserver and initiator wwpn...
                if ($lun_mapped_show_elm->{lun_id}  eq $lun_id and
                    $lun_mapped_show_elm->{vserver} eq $vserver_name and
                    $lun_mapped_show_elm->{initiators} =~ /$init_iqn/) {
                    # filter the list of igroups this lun is mapped to
                    @igroups_final = grep {
                        if ($_ eq $lun_mapped_show_elm->{igroup}) {
                            1;
                        } else {
                            0;
                        }
                    } @igroup_names;

                    # if there are no igroups, then this lun does not belong to this MPDev
                    if (not scalar(@igroups_final)) {
                        0;
                    } else {
                        1;
                    }
                } else {
                    0;
                }
            } @lun_mapped_show_data;

            # we now know the lun path to the path to this LUN
            my $lun_path = $lun_mapped_show->{path};

            $self->{ontap_lun_path} = $lun_path;
            $self->{ontap_vserver} = $vserver_name;
            foreach my $igroup_name (@igroups_final) {
                $self->{ontap_igroups}->{$igroup_name}++;
            }

            # discover the lif for this path
            my @vserver_iscsi_tpgroup_show_data = @{$cluster_content->{vserver_iscsi_tpgroup_show_data}};

            # find the lif that owns the tpgt for this path
            my $tpgt = $mpdevpath_obj->tpgt();
            my ($vserver_iscsi_tpgroup_show) = grep {
                if ($_->{tpgroup_tag} eq $tpgt) {
                    1;
                } else {
                    0;
                }
            } @vserver_iscsi_tpgroup_show_data;

            # find the lif object that is owned by this vserver and that shares
            # the name we just discovered for this path
            my $lif_name = $vserver_iscsi_tpgroup_show->{lif};
            my @lif_objs = @{$cluster_content->{lif_objs}};
            my ($lif_obj) = grep {
                if ($_->get_name(-use_cache => 1) eq $lif_name and
                    $_->get_vserver(-use_cache => 1) eq $vserver_name) {
                    1;
                } else {
                    0;
                }
            } @lif_objs;

            # store the lif object on this path if it's defined
            if (defined($lif_obj)) {
                $mpdevpath_obj->set_lif_relationship(-state => $SanUtil::TRUE,
                                                     -lif_obj => $lif_obj);
            }
            $dev_discovered = 1;
        } elsif ($mpdevpath_obj->protocol() eq "fcp") {
            my $target_wwnn = $mpdevpath_obj->target_wwnn();
            my $target_wwpn = $mpdevpath_obj->target_wwpn();
            my $init_wwpn   = $mpdevpath_obj->initiator_wwpn();
            my $lun_id      = $mpdevpath_obj->lun_id();

            # skip this path if we don't know the wwnn for this path's target
            next if ($target_wwnn eq "" or $target_wwnn =~ /Unavailable/i);

            # obtain the vserver name that owns this target wwnn
            my @vserver_fcp_show_data = @{$cluster_content->{vserver_fcp_show_data}};

            # normalize the target_wwnn
            $target_wwnn = $self->_normalize_fc_name($target_wwnn);

            my ($vserver_fcp_show) = grep {
                # normalize the wwnn in the ontap output
                my $t_wwnn = $_->{target_name};

                $t_wwnn = $self->_normalize_fc_name($t_wwnn);
                if ($t_wwnn eq $target_wwnn) {
                    1;
                } else {
                    0;
                }
            } @vserver_fcp_show_data;

            # if there is no vserver that owns this target wwnn, skip this path
            next if (not defined($vserver_fcp_show));

            # we now know the vserver name that owns this wwpn
            my $vserver_name = $vserver_fcp_show->{vserver};

            # find the igroup that this path is talking to
            my @vserver_fcp_initiator_show_data = @{$cluster_content->{vserver_fcp_initiator_show_data}};

            # normalize the wwpns
            $init_wwpn = $self->_normalize_fc_name($init_wwpn);

            my ($vserver_fcp_initiator_show) = grep {
                # normalize the wwpn in the ontap output
                my $i_wwpn = $_->{wwpn};
                my $vserver = $_->{vserver};

                $i_wwpn = $self->_normalize_fc_name($i_wwpn);

                # match the values
                if ($i_wwpn  eq $init_wwpn and
                    $vserver eq $vserver_name) {
                    1;
                } else {
                    0;
                }
            } @vserver_fcp_initiator_show_data;

            # if there is no vserver fcp initiator match, skip this path
            next if (not defined($vserver_fcp_initiator_show));

            # we now know the name of the igroups that this initiator is connected to.
            my $igroup_name_set = $vserver_fcp_initiator_show->{igroup};
            my @igroup_names = split(/,/, $igroup_name_set);

            # we now need to find the lun that is mapped to these igroups
            # search for the lun that is mapped to this igroup at this lun id
            my @lun_mapped_show_data = @{$cluster_content->{lun_mapped_show_data}};

            my $igroup_name;

            my @igroups_final;
            my ($lun_mapped_show) = grep {
                my $lun_mapped_show_elm = $_;
                # if this lun map entry matches our lun id/vserver and initiator wwpn...
                if ($lun_mapped_show_elm->{lun_id}  eq $lun_id and
                    $lun_mapped_show_elm->{vserver} eq $vserver_name and
                    $self->_normalize_fc_name($lun_mapped_show_elm->{initiators}) =~ /$init_wwpn/) {
                    # filter the list of igroups this lun is mapped to
                    @igroups_final = grep {
                        if ($_ eq $lun_mapped_show_elm->{igroup}) {
                            1;
                        } else {
                            0;
                        }
                    } @igroup_names;

                    # if there are no igroups, then this lun does not belong to this MPDev
                    if (not scalar(@igroups_final)) {
                        0;
                    } else {
                        1;
                    }
                } else {
                    0;
                }
            } @lun_mapped_show_data;

            # we now know the lun path to the path to this LUN
            my $lun_path = $lun_mapped_show->{path};

            $self->{ontap_lun_path} = $lun_path;
            $self->{ontap_vserver} = $vserver_name;
            foreach my $igroup_name (@igroups_final) {
                $self->{ontap_igroups}->{$igroup_name}++;
            }

            # discover the lif object that is tied to this path
            my @lif_objs = @{$cluster_content->{lif_objs}};
            my ($lif_obj) = grep {
                my $vs_name = $_->get_vserver(-use_cache => 1);
                my $wwpn = $_->get_wwpn(-use_cache => 1);

                # normalize the wwpns
                $wwpn = $self->_normalize_fc_name($wwpn);
                $target_wwpn = $self->_normalize_fc_name($target_wwpn);

                if ($wwpn eq $target_wwpn and
                    $vs_name eq $vserver_name) {
                    1;
                } else {
                    0;
                }
            } @lif_objs;

            # store the lif object on this path if it's defined
            if (defined($lif_obj)) {
                $mpdevpath_obj->set_lif_relationship(-state => $SanUtil::TRUE,
                                                     -lif_obj => $lif_obj);
            }

            $dev_discovered = 1;
        }
    }

    # if this device had at least one path discovered, then we have
    # enough data to discover the lun information
    if ($dev_discovered == 1) {

        # if we do not know the lun path at this point then return
        if (not defined($self->{ontap_lun_path})) {
            return;
        }

        # discover the lun object for this device
        # no qtree support here yet, need to use basename better
        ($self->{ontap_vol_name}, $self->{ontap_lun_name}) = ($self->{ontap_lun_path} =~ /^\/vol\/(.+?)\/(.+?)$/);

        my @vol_objs = @{$cluster_content->{vol_objs}};
        my @aggr_objs = @{$cluster_content->{aggr_objs}};

        # grab the vol object that this device lives in
        my ($vol_obj) = grep {
            if ($_->get_name(-use_cache => 1) eq $self->{ontap_vol_name} and
                $_->get_vserver(-use_cache => 1) eq $self->{ontap_vserver}) {
                1;
            } else {
                0;
            }
        } @vol_objs;

        # grab the aggregate object that this vol lives in
        my $aggr_name = $vol_obj->get_aggregate(-use_cache => 1);
        # store the aggr name on the object
        $self->{ontap_aggregate_name} = $aggr_name;
        my ($aggr_obj) = grep {
            if ($_->get_name(-use_cache => 1) eq $aggr_name) {
                1;
            } else {
                0;
            }
        } @aggr_objs;

        # determine which node this aggr lives on
        my $node_list = $aggr_obj->get_nodes(-use_cache => 1);

        # store the node list on the object
        $self->ontap_home_node($node_list);

        # determine the hapartner for this node
        my @sfo_show_data = @{$cluster_content->{sfo_show_data}};
        my ($sfo_show) = grep {
            if ($_->{node} eq $node_list) {
                1;
            } else {
                0;
            }
        } @sfo_show_data;

        # store the ha partner on the object
        $self->ontap_home_node_hapartner($sfo_show->{partner_name});

        # and finally, attach the lun object to the device
        my @lun_objs = @{$cluster_content->{lun_objs}};

        # parse out the lun object that matches the filters
        my ($lun_obj) = grep {
            if ($_->get_name(-use_cache => 1) eq $self->{ontap_lun_name} and
                $_->get_volume(-use_cache => 1) eq $self->{ontap_vol_name} and
                $_->get_vserver(-use_cache => 1) eq $self->{ontap_vserver}) {
                1;
            } else {
                0;
            }
        } @lun_objs;

        # establish the lun relationship to this device
        $self->set_lun_relationship(-state => $SanUtil::TRUE,
                                    -lun_obj => $lun_obj);

        # store the ostype of the lun on this device
        $self->ontap_lun_ostype($self->lun_obj()->{ostype});
    }

    return;
}

###############################################################################
# Method:    _generate_lun_relationship_7mode()
# Objective: Generate the LUN relationship for 7mode devices using SanFiler7M
###############################################################################
sub _generate_lun_relationship_7mode {
    my ($self, %args) = @_;

    my $sanfiler7m_objs = $args{-sanfiler7m_objs};

    if (ref($sanfiler7m_objs) ne "ARRAY" or not scalar(@{$sanfiler7m_objs})) {
        return;
    }

    my $dev_discovered = 0;

    foreach my $mpdevpath_obj ($self->get_current_path_state()->get_path_objs()) {

        # discover the lun for this device if it hasnt been done already
        if ($mpdevpath_obj->protocol() eq "iscsi" and not $self->has_lun_relationship()) {
            my $initiator_nodename = $mpdevpath_obj->initiator_iqn();
            my $target_nodename = $mpdevpath_obj->target_iqn();
            my $tpgt = $mpdevpath_obj->tpgt();
            my $lun_id = $mpdevpath_obj->lun_id();

            foreach my $controller_obj (@{$sanfiler7m_objs}) {
                my $item = "$controller_obj get_iscsi_nodename()";
                my $nodename;
                if ($self->{sc_obj}->is_cache_data_valid($item)) {
                    ($nodename) = $self->{sc_obj}->get_cache_data($item);
                } else {
                    $nodename = $controller_obj->get_iscsi_nodename();
                    if ($controller_obj->error()) {
                        $self->_set_err("Failed to get iscsi nodename for controller", $controller_obj->errmsg());
                        return;
                    }
                    $self->{sc_obj}->store_cache_data($item, $nodename);
                }

                # skip this controller if it doesnt match the nodename for this path
                next if ($nodename ne $target_nodename);

                # determine if this lun lives on this controller by tracing init nodename and target portal group tag
                # to map the igroup that is being used
                # iscsi initiator show (use tpgroup tag/init nodename) to map, to igroup, then do
                my $cmd = "iscsi initiator show";
                my @output;
                $item = "$controller_obj $cmd";
                if ($self->{sc_obj}->is_cache_data_valid($item)) {
                    @output = $self->{sc_obj}->get_cache_data($item);
                } else {
                    @output = $controller_obj->execute($cmd);
                    if ($controller_obj->error()) {
                        $self->_set_err("Failed to get iscsi initiator list for controller", $controller_obj->errmsg());
                        return;
                    }
                    $self->{sc_obj}->store_cache_data($item, @output);
                }

                #fas270-rtp07> iscsi initiator show
                #Initiators connected:
                #TSIH  TPGroup  Initiator/ISID/IGroup
                #   1    1001   sles10u232iscsidev (iqn.1996-04.de.suse:01:3c6fb9f34f5 / 00:02:3d:01:00:00 / nate_sles10u232iscsidev-i)
                # or
                #   1    1001   iqn.1996-04.de.suse:01:3c6fb9f34f5 / 00:02:3d:01:00:00 / nate_sles10u232iscsidev-i ; another igroup ; another igroup
                my $igroup;
                my @igroups;
                chomp(@output);
                foreach my $line (@output) {
                    if ($line =~ /^\s*\d+\s+${tpgt}.+?${initiator_nodename}\s+\/\s+.+?\/\s+(.+)/) {
                        $igroup = $1;
                        $igroup =~ s/\)\s*$//g; # get rid of trailing parens
                        $igroup =~ s/\r|\n//g;  # get rid of newlines/returns
                        @igroups = split(/\s*;\s*/, $igroup); # split on semi colons surrounded by spaces
                        chomp(@igroups); # make sure we clean up the output
                        last;
                    }
                }

                # go to the next controller if we could not parse out an igroup for this nodename
                next if (not scalar(@igroups));

                # scan over every igroup to find where this LUN is
                foreach my $igroup (@igroups) {
                    # parse the lunpath with the proper lun id of this lun
                    # that is the lun path.
                    my $cmd = "lun show -m -g $igroup";
                    my @output;
                    my $item = "$controller_obj $cmd";
                    if ($self->{sc_obj}->is_cache_data_valid($item)) {
                        @output = $self->{sc_obj}->get_cache_data($item);
                    } else {
                        @output = $controller_obj->execute($cmd);
                        if ($controller_obj->error()) {
                            $self->_set_err("Failed to get lun list for igroup $igroup", $controller_obj->errmsg());
                            return;
                        }
                        $self->{sc_obj}->store_cache_data($item, @output);
                    }

                    #fas270-rtp07> lun show -m -g nate_sles10u232iscsidev-i
                    #LUN path                            Mapped to          LUN ID  Protocol
                    #-----------------------------------------------------------------------
                    #/vol/nate_1_1_sles10u232iscsidev/nate_sles10u232iscsidev_linux_0 nate_sles10u232iscsidev-i        0     iSCSI

                    # finally, then parse the lunpath with the proper lun id of this lun
                    # that is the lun path.

                    foreach my $line (@output) {
                        if ($line =~ /^\/vol\/(\S+)\/(\S+)\s+${igroup}\s+${lun_id}\s+iSCSI/) {
                            # then we found the lun, save the controller, volume and lun
                            $self->{ontap_vol_name} = $1;
                            $self->{ontap_lun_name} = $2;
                            $self->{ontap_lun_path} = "/vol/$1/$2";
                            $self->{ontap_controller_7m} = $controller_obj->get_name();
                            $self->{ontap_igroups}->{$igroup}++;

                            # determine the ontap lun ostype
                            my $cmd = "lun show -v " . $self->{ontap_lun_path};
                            my @lun_show_output;
                            my $item = "$controller_obj $cmd";
                            if ($self->{sc_obj}->is_cache_data_valid($item)) {
                                @lun_show_output = $self->{sc_obj}->get_cache_data($item);
                            } else {
                                @lun_show_output = $controller_obj->execute($cmd);
                                if ($controller_obj->error()) {
                                    $self->_set_err("Failed to get OS type for LUN " . $self->{ontap_lun_path}, $controller_obj->errmsg());
                                    return;
                                }
                                $self->{sc_obj}->store_cache_data($item, @lun_show_output);
                            }

                            foreach my $lun_show_line (@lun_show_output) {
                                if ($lun_show_line =~ /^\s*Multiprotocol Type:\s*(\S+)/i) {
                                    $self->{ontap_lun_ostype} = $1;
                                    last;
                                }
                            }

                            # if we were unable to parse the lun os type error out
                            if ($self->{ontap_lun_ostype} eq "") {
                                $self->_set_err("Failed to parse OS type for LUN " . $self->{ontap_lun_path});
                                return;
                            }

                            # mark that we have discovered this device and break out
                            $dev_discovered = 1;
                            last;
                        }
                    }

                    # stop scanning the igroups if we found it
                    last if ($dev_discovered == 1);
                }
            }
        } elsif ($mpdevpath_obj->protocol() eq "fcp" and not $self->has_lun_relationship()) {
            # discover the details for this LUN if its a fibrechannel path
            my $target_nodename = $mpdevpath_obj->target_wwnn();
            my $initiator_portname = $mpdevpath_obj->initiator_wwpn();
            my $lun_id = $mpdevpath_obj->lun_id();

            foreach my $controller_obj (@{$sanfiler7m_objs}) {
                my $item = "$controller_obj get_fcp_nodename()";
                my $nodename;
                if ($self->{sc_obj}->is_cache_data_valid($item)) {
                    ($nodename) = $self->{sc_obj}->get_cache_data($item);
                } else {
                    $nodename = $controller_obj->get_fcp_nodename();
                    if ($controller_obj->error()) {
                        $self->_set_err("Failed to get fcp nodename for controller", $controller_obj->errmsg());
                        return;
                    }
                    $self->{sc_obj}->store_cache_data($item, $nodename);
                }

                # normalize the wwnns and wwpns
                $nodename = $self->_normalize_fc_name($nodename);
                $target_nodename = $self->_normalize_fc_name($target_nodename);
                $initiator_portname = $self->_normalize_fc_name($initiator_portname);

                # skip this controller if it doesnt match the nodename for this path
                next if ($nodename ne $target_nodename);

                my $cmd = "fcp show initiator";
                my @output;
                $item = "$controller_obj $cmd";
                if ($self->{sc_obj}->is_cache_data_valid($item)) {
                    @output = $self->{sc_obj}->get_cache_data($item);
                } else {
                    @output = $controller_obj->execute($cmd);
                    if ($controller_obj->error()) {
                        $self->_set_err("Failed to get fcp initiator list for controller", $controller_obj->errmsg());
                        return;
                    }
                    $self->{sc_obj}->store_cache_data($item, @output);
                }

                #fas270-rtp07> fcp show initiator
                #Initiators connected on adapter 0c:
                #Portname                  Group
                #--------                  -----
                #21:01:00:e0:8b:bc:eb:c1   nate_ibm3550-rtp01
                #21:00:00:e0:8b:9c:eb:c1   nate_ibm3550-rtp01
                #10:00:00:00:c9:6e:4d:d7   nate_ibm3550-rtp04
                #10:00:00:00:c9:6e:4d:d6   nate_ibm3550-rtp04
                #21:01:00:1b:32:20:d0:37
                #21:00:00:1b:32:00:d0:37
                #10:00:00:00:c9:5f:69:da
                #10:00:00:00:c9:5f:69:d9
                #10:00:00:00:c9:6e:a2:a3
                #10:00:00:00:c9:6e:a2:a2

                my $igroup;
                my @igroups;
                chomp(@output);
                foreach my $line (@output) {
                    next if ($line =~ /Initiators connected|Portname|\-\-\-\-|^\s*$/);
                    # remove leading whitespaces if any
                    $line =~ s/^\s*//;
                    my @split_array = split(/\s+/, $line);
                    my $igroup_wwpn = $split_array[0];
                    # normalize wwpn listed in the output before matching it
                    $igroup_wwpn = $self->_normalize_fc_name($igroup_wwpn);

                    if ($igroup_wwpn eq $initiator_portname) {
                        if ($line =~ /^\S+\s+(.+)/i) {
                            my $igroup_list = $1;
                            $igroup_list =~ s/\s+//g;
                            @igroups = split(/;/, $igroup_list);
                        }
                    }
                }

                # go to the next controller if we could not parse out any igroups for this portname
                next if (not scalar(@igroups));

                foreach my $igroup (@igroups) {
                    my $cmd = "lun show -m -g $igroup";
                    my @output;
                    my $item = "$controller_obj $cmd";
                    if ($self->{sc_obj}->is_cache_data_valid($item)) {
                        @output = $self->{sc_obj}->get_cache_data($item);
                    } else {
                        @output = $controller_obj->execute($cmd);
                        if ($controller_obj->error()) {
                            $self->_set_err("Failed to get lun list for igroup $igroup", $controller_obj->errmsg());
                            return;
                        }
                        $self->{sc_obj}->store_cache_data($item, @output);
                    }

                    #fas270-rtp07> lun show -m -g nate_ibm3550-rtp04
                    #LUN path                            Mapped to          LUN ID  Protocol
                    #-----------------------------------------------------------------------
                    #/vol/nate_1_1_ibm3550_rtp04/nate_ibm3550-rtp04_vmware_0 nate_ibm3550-rtp04        4       FCP
                    #/vol/nate_1_2_ibm3550_rtp04/nate_ibm3550-rtp04_vmware_1 nate_ibm3550-rtp04        5       FCP
                    #/vol/nate_1_3_ibm3550_rtp04/nate_ibm3550-rtp04_vmware_2 nate_ibm3550-rtp04        6       FCP
                    #/vol/nate_1_4_ibm3550_rtp04/nate_ibm3550-rtp04_vmware_3 nate_ibm3550-rtp04        7       FCP

                    # finally, then parse the lunpath with the proper lun id of this lun
                    # that is the lun path.
                    foreach my $line (@output) {
                        if ($line =~ /^\/vol\/(\S+)\/(\S+)\s+${igroup}\s+${lun_id}\s+FCP/) {
                            # then we found the lun, save the controller, volume and lun
                            $self->{ontap_vol_name} = $1;
                            $self->{ontap_lun_name} = $2;
                            $self->{ontap_lun_path} = "/vol/$1/$2";
                            $self->{ontap_controller_7m} = $controller_obj->get_name();
                            $self->{ontap_igroups}->{$igroup}++;

                            # determine the ontap lun ostype
                            my $cmd = "lun show -v " . $self->{ontap_lun_path};
                            my @lun_show_output;
                            my $item = "$controller_obj $cmd";
                            if ($self->{sc_obj}->is_cache_data_valid($item)) {
                                @lun_show_output = $self->{sc_obj}->get_cache_data($item);
                            } else {
                                @lun_show_output = $controller_obj->execute($cmd);
                                if ($controller_obj->error()) {
                                    $self->_set_err("Failed to get OS type for LUN " . $self->{ontap_lun_path}, $controller_obj->errmsg());
                                    return;
                                }
                                $self->{sc_obj}->store_cache_data($item, @lun_show_output);
                            }

                            foreach my $lun_show_line (@lun_show_output) {
                                if ($lun_show_line =~ /Multiprotocol Type:\s*(\S+)/i) {
                                    $self->{ontap_lun_ostype} = $1;
                                    last;
                                }
                            }

                            # if we were unable to parse the lun os type error out
                            if ($self->{ontap_lun_ostype} eq "") {
                                $self->_set_err("Failed to parse OS type for LUN " . $self->{ontap_lun_path});
                                return;
                            }

                            # mark that we have discovered this device and break out
                            $dev_discovered = 1;
                            last;
                        }
                    }

                    # stop searching igroups if we discovered it already
                    last if ($dev_discovered);
                }

                # stop searching through nodes if we discovered already
                last if ($dev_discovered);
            }
        }

        # stop searching the Path objects if we discovered from one path
        last if ($dev_discovered);
    }

    # if we've discovered the lun relationship, mark it here
    if ($dev_discovered == 1) {
        $self->set_lun_relationship(-state => $dev_discovered);

        # the home node for this device is the node that the LUN
        # currently lives on
        $self->ontap_home_node($self->ontap_controller_7m());
    }

    return;
}

###############################################################################
# Method:    _add_path_state()
# Objective: Add a new path state for this device
###############################################################################
sub _add_path_state {
    my ($self, %args) = @_;

    my $path_objs = $args{-mp_device_path_objs};

    # for each new path, walk down the path states
    # until we find a lif that was discovered for it
    # and copy it over to the new path
    foreach my $path_obj (@{$path_objs}) {

        # skip this path if it already has a lif relationship
        next if ($path_obj->has_lif_relationship());

        # scan older path states for this path
        # if we find a match (initiator <-> target), then copy the lif object over
        foreach my $old_path_state_obj ($self->get_path_states()) {

            # break out of the loop if we have obtained a lif relationship
            last if ($path_obj->has_lif_relationship());

            foreach my $old_path_obj ($old_path_state_obj->get_path_objs()) {

                # skip this path if it doesnt have a lif relationship
                next if (not $old_path_obj->has_lif_relationship());

                # if we matched all the unique fields of both paths
                # or if we have an ontap target port name match (windows with ontapdsm)
                # then copy over the lif object
                if (($path_obj->protocol()          =~ /iscsi/i and
                     $path_obj->target_iqn()        eq $old_path_obj->target_iqn()        and
                     $path_obj->tpgt()              eq $old_path_obj->tpgt()              and
                     $path_obj->target_ip_address() eq $old_path_obj->target_ip_address() and
                     $path_obj->initiator_iqn()     eq $old_path_obj->initiator_iqn())
                   or
                    ($path_obj->protocol()          =~ /fcp/i and
                     $path_obj->target_wwnn()       eq $old_path_obj->target_wwnn()    and
                     $path_obj->target_wwpn()       eq $old_path_obj->target_wwpn()    and
                     $path_obj->initiator_wwnn()    eq $old_path_obj->initiator_wwnn() and
                     $path_obj->initiator_wwpn()    eq $old_path_obj->initiator_wwpn())
                   or
                    ($path_obj->ontap_target_port_name() eq $old_path_obj->ontap_target_port_name())) {

                    # copy the lif obj from the old path to this path
                    $path_obj->set_lif_relationship(-lif_obj => $old_path_obj->lif_obj(),
                                                    -state => 1);
                    last;
                }

                # Some OSes do not cache this level of information for paths that go away,
                # For ESX, we will have to key off the path name since it shouldnt change
                # For windows, we can also key off the path name of the previous path
                # Rhel6 FCP has a caveat that the path name gets recycled when the path dies and comes back
                # so we can't assume this will work for rhel6 FCP.
                # Rhel5u6 FCP has a caveat where the target_wwnn() field of the dead path will be 'ffffffffffffffff'
                # so we have to key off the path name for that.
                # For these OSes, we have to key off the path name not changing
                my $hosttype = $self->{sc_obj}->get_connection()->{hostp}{hosttype};
                if ($hosttype =~ /linux\-vmware|windows/i or 
                    ($hosttype =~ /unix\-linux\-redhat/i  and $self->{sc_obj}->{major_revision} !~ /rhel6|ol6/i)) {
                    if ($path_obj->path_name() eq $old_path_obj->path_name()) {
                        # copy the lif obj from the old path to this path
                        $path_obj->set_lif_relationship(-lif_obj => $old_path_obj->lif_obj(),
                                                        -state => 1);
                        last;
                    }
                }
            }
        }
    }

    # unshift the new lun path state on the path states array,
    # (newest state in the front of the list)
    my $path_state_hash = new San::SanClient::MPDevMgr::MPDevPathState(-path_objs => $path_objs);

    # store the new path state hash on the path state array
    # unshifting so the latest path state is always
    # at the beginning of the array
    unshift(@{$self->{path_states}}, $path_state_hash);

    return;
}

###############################################################################
# Method:    _normalize_fc_name()
# Objective: Given an fc name, normalize it by removing colons and lowercase it
###############################################################################
sub _normalize_fc_name {
    my ($self, $wwn) = @_;

    $wwn =~ s/://g;
    return lc($wwn);
}

1;
