#
# Copyright (c) 2010-2013 NetApp Inc.
# All rights reserved.
#
## @summary Provide an OS generic interface to a SAN Multipathed Device Path
## @author hrivera@netapp.com
## @status private
## @pod here

=pod

=head1 Package Name

MPDev.pm

=head1 Author

dl-san-automation

=head1 Description

An OS generic object representation of a single path to a multipathed device.

Current Operating System support:
 ESX4.x/ESXi4.x
 ESXi5
 Redhat 5u6
 Redhat 6
 SLES 10
 Windows

=head1 Inherited Packages

 Error
 San::SanLog

=cut

package San::SanClient::MPDevMgr::MPDevPath;

use strict;
use warnings;
use Data::Dumper;

use Error;
use San::SanLog;

use vars qw (@ISA);

@ISA = qw (Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

=pod

=head1 B<Public Methods>

=head2 B<new(%args)>

The constructor for MPDev. Instantiates and returns an MPDev object based
on the OS of the SanClient object.

=over 4

=item Inputs:

=over 2

=item I<-os_specific_data> => \%hash

 Optional: A hash reference of OS specific data {$key => $value, $keyN => $valueN} pairs.

=item I<-protocol> => scalar

 Optional: Protocol of this path.
 Valid Values: 'fcp' or 'iscsi'

=item I<-path_state> => scalar

 Optional: The state of the path "dead", "active"

=item I<-alua_state> => scalar

 Optional: The ALUA state of the path "AO", "ANO", "UNKNOWN", "DOWN"

=item I<-lif_obj> => San::ControllerCluster::Filer10M::Vserver::Lif / NACL::C::NetworkInterface

 Optional: A San::ControllerCluster::Filer10M::Vserver::Lif / NACL::C::NetworkInterface object that
 is maps to this path object.
 Default: {}

=item I<-ignore_lif_relationship> => scalar

 Optional: Specifies if the caller cares that this path has or does not have a lif relationship.

=item I<-lun_id> => scalar

 Optional: Specifies the lun id of the device that owns this path

=item I<-initiator_iqn> => scalar

 Optional: Specifies the iqn of the initiator of this path, if this is iSCSI

=item I<-target_iqn> => scalar

 Optional: Specifies the iqn of the target of this path, if this is iSCSI

=item I<-target_iscsi_alias> => scalar

 Optional: Specifies the iscsi alias of the target for this path

=item I<-target_iscsi_session> => scalar

 Optional: Specifies the iscsi session of the target for this path

=item I<-target_ip_address> => scalar

 Optional: Specifies the iscsi ip address of the target for this path

=item I<-target_ip_port> => scalar

 Optional: Specifies the iscsi ip port of the target for this path

=item I<-tpgt> => scalar

 Optional: Specifies the tpgt of the iscsi path

=item I<-target_wwpn> => scalar

 Optional: Specifies the target WWPN fcp of the path

=item I<-target_wwnn> => scalar

 Optional: Specifies the target WWNN fcp of that path

=item I<-initiator_wwpn> => scalar

 Optional: Specifies the initiator WWPN fcp of the path

=item I<-initiator_wwnn> => scalar

 Optional: Specifies the initiator WWNN fcp of that path

=item I<-ontap_target_port_name> => scalar

 Optional: Specifies the name of the ONTAP target port for this path

=back

=item Outputs:

=over 2

A San::SanClient::MPDevPath object

=back

=item Example:

 my $mpdevpath_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_name => $path_name,
                                                             -lun_id => $lun_id,
                                                             -path_state => $path_state,
                                                             -initiator_iqn => $path_initiator_iqn,
                                                             -target_iqn => $path_target_iqn,
                                                             -target_iscsi_alias => $iscsi_target_alias,
                                                             -target_iscsi_session => $iscsi_session,
                                                             -target_ip_address => "192.168.1.1",
                                                             -target_ip_port => "3260",
                                                             -ontap_target_port_name => "iscsilif1",
                                                             -tpgt => $tpgt,
                                                             -protocol => "iscsi"););

=back

=cut

sub new {

    my ($class, %args) = @_;

    my $self = {};

    bless($self, $class);

    %args = &SanUtil::strip_dash(%args);

    $self->{ignore_lif_relationship} = 0;

    my $lif_obj = delete $args{lif_obj};
    my $lif_relationship = exists($args{lif_relationship}) ? delete($args{lif_relationship}) : 1;
    if ($lif_obj) {
        $self->set_lif_relationship(-lif_obj => $lif_obj,
                                    -state => $lif_relationship);
    } else {
        $self->set_lif_relationship();
    }

    foreach (keys %args) {
        $self->{$_} = $args{$_};
    }

    return $self;
}

###############################################################################
#                          Public Accessor/Setter Methods
###############################################################################

=pod

=head1 B<Public Accessor/Setter Methods>

=head2 B<protocol($)>

Get/Set the 'protocol' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $protocol = $mpdevpath_obj->protocol();

 # Set the value
 $mpdevpath_obj->protocol($new_value);

=back

=cut

sub protocol {
    my ($self, $val) = @_;

    return $self->_get_or_set("protocol", $val);
}

=pod

=head2 B<path_state($)>

Get/Set the 'path_state' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $path_state = $mpdevpath_obj->path_state();

 # Set the value
 $mpdevpath_obj->path_state($new_value);

=back

=cut

sub path_state {
    my ($self, $val) = @_;

    return $self->_get_or_set("path_state", $val);
}

=pod

=head2 B<lun_id($)>

Get/Set the 'lun_id' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $lun_id = $mpdevpath_obj->lun_id();

 # Set the value
 $mpdevpath_obj->lun_id($new_value);

=back

=cut

sub lun_id {
    my ($self, $val) = @_;

    return $self->_get_or_set("lun_id", $val);
}

=pod

=head2 B<alua_state($)>

Get/Set the 'alua_state' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $alua_state = $mpdevpath_obj->alua_state();

 # Set the value
 $mpdevpath_obj->alua_state($new_value);

=back

=cut

sub alua_state {
    my ($self, $val) = @_;

    return $self->_get_or_set("alua_state", $val);
}

=pod

=head2 B<path_name($)>

Get/Set the 'path_name' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $path_name = $mpdevpath_obj->path_name();

 # Set the value
 $mpdevpath_obj->path_name($new_value);

=back

=cut

sub path_name {
    my ($self, $val) = @_;

    return $self->_get_or_set("path_name", $val);
}

=pod

=head2 B<target_iqn($)>

Get/Set the 'target_iqn' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $target_iqn = $mpdevpath_obj->target_iqn();

 # Set the value
 $mpdevpath_obj->target_iqn($new_value);

=back

=cut

sub target_iqn {
    my ($self, $val) = @_;

    return $self->_get_or_set("target_iqn", $val);
}

=pod

=head2 B<target_ip_address($)>

Get/Set the 'target_ip_address' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $target_ip_address = $mpdevpath_obj->target_ip_address();

 # Set the value
 $mpdevpath_obj->target_ip_address($new_value);

=back

=cut

sub target_ip_address {
    my ($self, $val) = @_;

    return $self->_get_or_set("target_ip_address", $val);
}

=pod

=head2 B<target_ip_port($)>

Get/Set the 'target_ip_port' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $target_ip_port = $mpdevpath_obj->target_ip_port();

 # Set the value
 $mpdevpath_obj->target_ip_port($new_value);

=back

=cut

sub target_ip_port {
    my ($self, $val) = @_;

    return $self->_get_or_set("target_ip_port", $val);
}

=pod

=head2 B<target_iscsi_alias($)>

Get/Set the 'target_iscsi_alias' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $target_iscsi_alias = $mpdevpath_obj->target_iscsi_alias();

 # Set the value
 $mpdevpath_obj->target_iscsi_alias($new_value);

=back

=cut

sub target_iscsi_alias {
    my ($self, $val) = @_;

    return $self->_get_or_set("target_iscsi_alias", $val);
}

=pod

=head2 B<target_iscsi_session($)>

Get/Set the 'target_iscsi_session' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $target_iscsi_session = $mpdevpath_obj->target_iscsi_session();

 # Set the value
 $mpdevpath_obj->target_iscsi_session($new_value);

=back

=cut

sub target_iscsi_session {
    my ($self, $val) = @_;

    return $self->_get_or_set("target_iscsi_session", $val);
}

=pod

=head2 B<tpgt($)>

Get/Set the 'tpgt' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $tpgt = $mpdevpath_obj->tpgt();

 # Set the value
 $mpdevpath_obj->tpgt($new_value);

=back

=cut

sub tpgt {
    my ($self, $val) = @_;

    return $self->_get_or_set("tpgt", $val);
}

=pod

=head2 B<initiator_iqn($)>

Get/Set the 'initiator_iqn' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $initiator_iqn = $mpdevpath_obj->initiator_iqn();

 # Set the value
 $mpdevpath_obj->initiator_iqn($new_value);

=back

=cut

sub initiator_iqn {
    my ($self, $val) = @_;

    return $self->_get_or_set("initiator_iqn", $val);
}

=pod

=head2 B<target_wwnn($)>

Get/Set the 'target_wwnn' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $target_wwnn = $mpdevpath_obj->target_wwnn();

 # Set the value
 $mpdevpath_obj->target_wwnn($new_value);

=back

=cut

sub target_wwnn {
    my ($self, $val) = @_;

    return $self->_get_or_set("target_wwnn", $val);
}

=pod

=head2 B<target_wwpn($)>

Get/Set the 'target_wwpn' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $target_wwpn = $mpdevpath_obj->target_wwpn();

 # Set the value
 $mpdevpath_obj->target_wwpn($new_value);

=back

=cut

sub target_wwpn {
    my ($self, $val) = @_;

    return $self->_get_or_set("target_wwpn", $val);
}

=pod

=head2 B<initiator_wwnn($)>

Get/Set the 'initiator_wwnn' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $initiator_wwnn = $mpdevpath_obj->initiator_wwnn();

 # Set the value
 $mpdevpath_obj->initiator_wwnn($new_value);

=back

=cut

sub initiator_wwnn {
    my ($self, $val) = @_;

    return $self->_get_or_set("initiator_wwnn", $val);
}

=pod

=head2 B<initiator_wwpn($)>

Get/Set the 'initiator_wwpn' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $initiator_wwpn = $mpdevpath_obj->initiator_wwpn();

 # Set the value
 $mpdevpath_obj->initiator_wwpn($new_value);

=back

=cut

sub initiator_wwpn {
    my ($self, $val) = @_;

    return $self->_get_or_set("initiator_wwpn", $val);
}

=head2 B<ontap_target_port_name($)>

Get/Set the 'ontap_target_port_name' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $ontap_target_port_name = $mpdevpath_obj->ontap_target_port_name();

 # Set the value
 $mpdevpath_obj->ontap_target_port_name($new_value);

=back

=cut

sub ontap_target_port_name {
    my ($self, $val) = @_;

    return $self->_get_or_set("ontap_target_port_name", $val);
}

=pod

=head2 B<lif_obj($)>

Get/Set the 'lif_obj' attribute.

=over 4

=item Inputs:

=over 2

=item Scalar

 Optional: The new value of the attribute

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Get the value
 my $lif_obj = $mpdevpath_obj->lif_obj();

 # Set the value
 $mpdevpath_obj->lif_obj($new_value);

=back

=cut

sub lif_obj {
    my ($self, $val) = @_;

    # if this is an assignment operation, clean up the object
    if (defined($val)) {
        # if the value is a nacl object and it has a blessed command interface
        # remove the command interface
        if (ref($val) =~ /NACL/i) {
            if (ref($val->{command_interface})) {
                # reassign the command interface to be the name of the command interface so that it can be reassigned at a later time if needed
                my $ci = delete($val->{command_interface});
                $val->{_command_interface_name} = $ci->name();
            }
            # store the lif name as the ontap target port name
            $self->ontap_target_port_name($val->lif());
        } elsif (ref($val) =~ /San::ControllerCluster::Filer10M::Vserver::Lif/i) {
            # this might be a controller cluster Lif object, clean up all the unneeded fields
            delete($val->{parent_obj});
            delete($val->{conn_obj});
            delete($val->{cmd_obj});
            delete($val->{fields});
        }
    }

    return $self->_get_or_set("lif_obj", $val);
}

=pod

=head2 B<set_lif_relationship($)>

Set the 'lif_relationship' attribute.

=over 4

=item Inputs:

=over 2

=item I<-lun_obj> => San::ControllerCluster::Filer10M::Vserver::Lif or NACL::CS::NetworkInterface 

 Optional: The lif object to associate the path to

=item I<-state> => scalar

 Optional: The lif relationship state  (0 or 1)
 Default: 0 (no relationship)

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # Set a lif relationship
 $mpdevpath_obj->set_lif_relationship(-lif_obj => $lif_obj,
                                      -state => 1);

 # Reset a lif relationship
 $mpdevpath_obj->set_lif_relationship();

=back

=cut

sub set_lif_relationship {
    my ($self, %args) = @_;

    my $lif_obj = exists($args{-lif_obj}) ? $args{-lif_obj} : undef;
    my $state = exists($args{-state}) ? $args{-state} : 0;

    # set the lun object if one was specified
    $self->lif_obj($lif_obj);

    # set the lun relationship state
    $self->_get_or_set("lif_relationship", $state);

    return;
}

=pod

=head2 B<has_lif_relationship()>

Check if this device has a lif relationship.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

0 (does not have relationship) or 1 (has relationship)

=back

=item Example:

 # check if this device has a lif relationship
 if ($mpdev_obj->has_lif_relationship() == 1) {
     ...
 } else {
     ...
 }

=back

=cut

sub has_lif_relationship {
    my ($self) = @_;

    return $self->{lif_relationship};
}

=head2 B<ignore_lif_relationship()>

Check if this device path is ignoring lif relationships

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

0 (does not want to ignore lif relationships) or 1 (ignores lif relationships)

=back

=item Example:

 # check if this path object cares about lif relationships or not
 if ($mpdevpath_obj->ignore_lif_relationship() == 1) {
     ...
 } else {
     ...
 }

=back

=cut

sub ignore_lif_relationship {
    my ($self) = @_;

    return $self->{ignore_lif_relationship};
}

=head2 B<set_os_specific_data(%args)>

Set OS specific data on the device path.

=over 4

=item Inputs:

=over 2

=item I<-key> => $data

 Generic {-key => data} fields to store on the object.

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # set some generic key => value data on the object
 $mpdevpath_obj->set_os_specific_data(-key1 => "value1",
                                      -key2 => "value2");
 if ($mpdev_obj->error()) {
    # ...
 }

=back

=cut

sub set_os_specific_data {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    foreach (keys %args) {
        $self->{os_specific_data}->{$_} = $args{$_};
    }
}

=pod

=head2 B<get_os_specific_data()>

Get OS specific data from this device path.

=over 4

=item Inputs:

=over 2

=item I<$key>

=back

=item Outputs:

=over 2

None

=back

=item Example:

 # get some generic key => value data on the object
 my $value = $mpdevpath_obj->get_os_specific_data("some_key");
 if ($mpdev_obj->error()) {
    # ...
 }

=back

=cut

sub get_os_specific_data {
    my ($self, $key) = @_;

    if (defined($key)) {
        return $self->{os_specific_data}->{$key};
    } else {
        return $self->{os_specific_data};
    }
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _get_or_set()
# Objective: Generic getter/setting method
###############################################################################
sub _get_or_set {
    my ($self, $attr, $val) = @_;

    if (defined($val)) {
        $self->{$attr} = $val;
    }

    return exists($self->{$attr}) ? $self->{$attr} : "";
}

1;
