#
# Copyright (c) 2010-2013 NetApp Inc.
# All rights reserved.
#
## @summary Provide an OS generic interface to a state collection of SAN Multipathed Device Paths
## @author hrivera@netapp.com
## @status private
## @pod here

=pod

=head1 Package Name

MPDev.pm

=head1 Author

dl-san-automation

=head1 Description

An OS generic object representation of a state collection of paths (MPDevPath) to a SAN multipathed
device.

Current Operating System support:
 ESX4.x/ESXi4.x
 ESXi5
 Redhat 5u6
 Redhat 6
 SLES 10
 Windows

=head1 Inherited Packages

 Error
 San::SanLog

=cut

package San::SanClient::MPDevMgr::MPDevPathState;

use strict;
use warnings;
use Error;
use San::SanLog;

use vars qw (@ISA);

@ISA = qw (Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

=pod

=head1 B<Public Methods>

=head2 B<new(%args)>

The constructor for MPDev. Instantiates and returns an MPDev object based
on the OS of the SanClient object.

=over 4

=item Inputs:

=over 2

=item I<-path_objs> => \@array

 Optional: A hash reference of OS specific data {$key => $value, $keyN => $valueN} pairs.

=back

=item Outputs:

=over 2

A San::SanClient::MPDevPathState object

=back

=item Example:

 my $mpdevpath_state_obj = new San::SanClient::MPDevMgr::MPDevPath(-path_objs => \@mpdevpath_objs);

=back

=cut

sub new {
    my ($class, %args) = @_;

    my $self = {};

    bless($self, $class);

    my $path_objs = $args{-path_objs};
    my $path_objs_hash = {};

    # initialize the path counters
    $self->{total_paths} = 0;
    $self->{total_alive_paths} = 0;
    $self->{total_dead_paths} = 0;

    $self->{total_iscsi_paths} = 0;
    $self->{total_iscsi_alive_paths} = 0;
    $self->{total_iscsi_dead_paths} = 0;

    $self->{total_fcp_paths} = 0;
    $self->{total_fcp_alive_paths} = 0;
    $self->{total_fcp_dead_paths} = 0;

    # count the protocols (alive/dead) for this new path state
    foreach my $path_obj (@{$path_objs}) {
        # incrememnt the total # of paths for this path state
        $self->{total_paths}++;

        my $path_protocol = $path_obj->protocol();
        my $path_name = $path_obj->path_name();
        my $path_state = $path_obj->path_state();

        if ($path_protocol =~ /iscsi/i) {
            $self->{total_iscsi_paths}++;
            if ($path_state =~ /active|alive|standby|passive/i) {
                $self->{total_iscsi_alive_paths}++;
                $self->{total_alive_paths}++;
            } else {
                $self->{total_iscsi_dead_paths}++;
                $self->{total_dead_paths}++;
            }
        } elsif ($path_protocol =~ /fcp/i) {
            $self->{total_fcp_paths}++;
            if ($path_state =~ /active|alive|standby|passive/i) {
                $self->{total_fcp_alive_paths}++;
                $self->{total_alive_paths}++;
            } else {
                $self->{total_fcp_dead_paths}++;
                $self->{total_dead_paths}++;
            }
        }

        # store the path obj in a lookup table hash
        $path_objs_hash->{$path_name} = $path_obj;
    }

    # store the lookup table on self
    $self->{path_objs_hash} = $path_objs_hash;

    # store the timestamp that this path state was created
    $self->{path_collection_time} = time();

    return $self;
}

=pod

=head2 B<get_path_objs_hash()>

Obtain the hash map for all MPDevPath objects.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

A hash of MPDevPath objects, keyed by path name.

=back

=item Example:

 my $path_objs_hash = $mpdevpathstate_obj->get_path_objs_hash();

=back

=cut

sub get_path_objs_hash {
    my ($self) = @_;

    return $self->{path_objs_hash};
}

=pod

=head2 B<get_path_objs()>

Obtain an array or array reference of MPDevPath objects in this path state.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

An array or array reference of MPDevPath objects

=back

=item Example:

 my $path_objs = $mpdevpathstate_obj->get_path_objs();

 my @path_objs = $mpdevpathstate_obj->get_path_objs();

=back

=cut

sub get_path_objs {
    my ($self) = @_;

    my @return = values %{$self->{path_objs_hash}};

    return wantarray() ? @return : \@return;
}

=pod

=head2 B<get_filtered_path_objs()>

Obtain an array or array reference of MPDevPath objects in this path state that 
meet the specified filters.

=over 4

=item Inputs:

=over 2

=item I<-alua_state> => scalar

 Optional: "ANO", "AO", "DOWN", or "UNKNOWN"

=back

=item Outputs:

=over 2

An array or array reference of MPDevPath objects that match the filters.

=back

=item Example:

 # obtain all unoptimized MPDevPath objects in this path state object
 my @path_objs = $mpdevpathstate_obj->get_filtered_path_objs(-alua_state => "ANO");

=back

=cut

sub get_filtered_path_objs {
    my ($self, %args) = @_;

    my $alua_state;

    foreach (keys %args) {
        if (/-?alua_state/) {
            # alua_state must be "AO" or "ANO"
            if (length($alua_state = $args{$_}) and $alua_state !~ /^(AO|ANO)$/) {
                $self->_set_err("Invalid value: $alua_state for argument '-alua_state'");
                return (wantarray()) ? () : [];
            }
            last;
        }
    }

    my @path_objs = $self->get_path_objs();
    my @return = undef;

    if (defined($alua_state)) {
        @return = grep {$_->alua_state() eq $alua_state} @path_objs;
    } else {
        @return = @path_objs;
    }

    return wantarray() ? @return : \@return;
}

=pod

=head2 B<get_path_names()>

Obtain an array or array reference of path names in this path state.

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

An array or array reference of path names.

=back

=item Example:

 my $path_names = $mpdevpathstate_obj->get_path_names();

 my @path_names = $mpdevpathstate_obj->get_path_names();

=back

=cut

sub get_path_names {
    my ($self) = @_;

    my @return = keys %{$self->{path_objs_hash}};

    return wantarray() ? @return : \@return;
}

=pod

=head2 B<total_paths()>

Obtain the total number of paths in this path state

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Number of paths

=back

=item Example:

 if ($mpdevpathstate_obj->total_paths() > 1) {
     ...
 }

=back

=cut

sub total_paths {
    my ($self) = @_;

    return $self->{total_paths};
}

=pod

=head2 B<total_alive_paths()>

Obtain the total number of alive paths in this path state

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Number of alive paths

=back

=item Example:

 if ($mpdevpathstate_obj->total_alive_paths() > 1) {
     ...
 }

=back

=cut

sub total_alive_paths {
    my ($self) = @_;

    return $self->{total_alive_paths};
}

=pod

=head2 B<total_dead_paths()>

Obtain the total number of dead paths in this path state

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Number of dead paths

=back

=item Example:

 if ($mpdevpathstate_obj->total_dead_paths() > 1) {
     ...
 }

=back

=cut

sub total_dead_paths {
    my ($self) = @_;

    return $self->{total_dead_paths};
}

=pod

=head2 B<total_iscsi_paths()>

Obtain the total number of iSCSI paths in this path state

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Number of iSCSI paths

=back

=item Example:

 if ($mpdevpathstate_obj->total_iscsi_paths() > 1) {
     ...
 }

=back

=cut

sub total_iscsi_paths {
    my ($self) = @_;

    return $self->{total_iscsi_paths};
}

=pod

=head2 B<total_iscsi_alive_paths()>

Obtain the total number of alive iSCSI paths in this path state

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Number of alive iSCSI paths

=back

=item Example:

 if ($mpdevpathstate_obj->total_iscsi_alive_paths() > 1) {
     ...
 }

=back

=cut

sub total_iscsi_alive_paths {
    my ($self) = @_;

    return $self->{total_iscsi_alive_paths};
}

=pod

=head2 B<total_iscsi_dead_paths()>

Obtain the total number of dead iSCSI paths in this path state

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Number of dead iSCSI paths

=back

=item Example:

 if ($mpdevpathstate_obj->total_iscsi_dead_paths() > 1) {
     ...
 }

=back

=cut

sub total_iscsi_dead_paths {
    my ($self) = @_;

    return $self->{total_iscsi_dead_paths};
}

=pod

=head2 B<total_fcp_paths()>

Obtain the total number of FCP paths in this path state

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Number of FCP paths

=back

=item Example:

 if ($mpdevpathstate_obj->total_fcp_paths() > 1) {
     ...
 }

=back

=cut

sub total_fcp_paths {
    my ($self) = @_;

    return $self->{total_fcp_paths};
}

=pod

=head2 B<total_fcp_alive_paths()>

Obtain the total number of alive FCP paths in this path state

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Number of alive FCP paths

=back

=item Example:

 if ($mpdevpathstate_obj->total_fcp_alive_paths() > 1) {
     ...
 }

=back

=cut

sub total_fcp_alive_paths {
    my ($self) = @_;

    return $self->{total_fcp_alive_paths};
}

=pod

=head2 B<total_fcp_dead_paths()>

Obtain the total number of dead FCP paths in this path state

=over 4

=item Inputs:

=over 2

None

=back

=item Outputs:

=over 2

Number of dead FCP paths

=back

=item Example:

 if ($mpdevpathstate_obj->total_fcp_dead_paths() > 1) {
     ...
 }

=back

=cut

sub total_fcp_dead_paths {
    my ($self) = @_;

    return $self->{total_fcp_dead_paths};
}

1;
