#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary MPDevMgr base class for Windows
## @author hrivera@netapp.com
## @status private
## @pod ../MPDevMgr.pm

package San::SanClient::MPDevMgr::WinMPDevMgr;

use strict;
use warnings;
use Data::Dumper;

use San::SanClient::MPDevMgr::MPDev;
use San::SanClient::MPDevMgr::MPDevPath;
use San::SanCIT;
use NATE::Log qw(log_global);

my $Log;

$Log = log_global();

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: Create an MPDevMgr instance
###############################################################################
sub _new {
    my ($self, %args) = @_;

    return $self;
}

###############################################################################
# Method:    load_all_devices()
# Objective: Load in the current state of all device's filesystem/partition
###############################################################################
sub _load_all_devices {
    my ($self) = @_;

    # Not implemented

    return;
}

###############################################################################
# Method:    os_specific_validation()
# Objective: Perform OS Specific validation on all devices passed in
###############################################################################
sub _os_specific_validation {
    my ($self) = @_;

    # Not implemented

    return;
}

###############################################################################
# Method:    refresh()
# Objective: Rediscover all Multipathed LUNs on the system and populate the
#            Device Manager
###############################################################################
sub _refresh {
    my ($self, %args) = @_;

    # determine what DSM is installed, if MSDSM, then refresh the msdsm method

    my $sc_obj = $self->{sc_obj};
    my $mpc_bin = $sc_obj->{WinClient}->{cmd_mpc};

    my ($status, @mpc_show_disks, @mpc_show_luns);
    my $error = 0;

    # rescan the pnp bus before we refresh if requested
    if ($args{rescan}) {
        &ClientUtil::execute($sc_obj->{conn}, $sc_obj->{WinClient}->{cmd_dev_rescan});
    }

    # mpc show disks
    for (1..5) {
        ($status, @mpc_show_disks) = &ClientUtil::execute($sc_obj->{conn}, "$mpc_bin show disks", 1200);
        if (grep(/An error occurred/i, @mpc_show_disks)) {
            $error = 1;
        } else {
            $error = 0;
            last;
        }
    }
    if ($error) {
        # could not validate show disks on this host, not much we can do here:
        $self->_set_err("'mpc show disks' failed after 5 retries, giving up.");
        return;
    }

    foreach my $line (@mpc_show_disks) {
        next if ($line !~ /^Disk\d+/);

        my @line = split(/\s+/, $line);
        my $mp_device_name = $line[0];
        my $lun_id = $line[1];
        my $size = $line[2];
        my $size_units = $line[3];
        my $state = $line[4];
        my $is_boot = $line[5];
        my $vendor = $line[6];
        my $partition_style = $line[7];
        my $disk_type = $line[8];
        my $drive_letter = $line[9];

        # the label may have spaces, so regex match from here on
        my $label = "";
        my $is_quorum = 0;
        my $cluster_name = "";
        my $cluster_node_name = "";
        if ($line =~ /\[(.+?)\]\s+(\S+)\s+(\S+)\s+(\S+)/) {
            $label = $1;
            $is_quorum = $2;
            $cluster_name = $3;
            $cluster_node_name = $4;
        }

        my $mp_device_path = $mp_device_name;
        $mp_device_path =~ s/Disk/\\\\.\\PHYSICALDRIVE/;

        # create an MPDev object for this disk
        my $mpdev_obj = new San::SanClient::MPDevMgr::MPDev(sc_obj => $sc_obj,
                                                            mp_device_name => $mp_device_name,
                                                            device_size => "$size $size_units",
                                                            mp_device_path => $mp_device_path,
                                                            device_vendor => $vendor,
                                                            device_model => "LUN C-mode", # XXX assuming cmode only
                                                            lun_id => $lun_id);

        # store os specific data fields for a few things
        $mpdev_obj->set_os_specific_data(cluster_quorum => $is_quorum,
                                         cluster_name => $cluster_name,
                                         cluster_node_name_owner => $cluster_node_name,
                                         volume_label => $label,
                                         partition_style => $partition_style,
                                         is_boot => $is_boot,
                                         disk_type => $disk_type,
                                         drive_letter => $drive_letter,
                                         state => $state,
                                         size => $size,
                                         size_units => $size_units);

        if ($drive_letter !~ /None/i) {
            $mpdev_obj->is_partitioned(1);
            $mpdev_obj->is_mounted(1);
            $mpdev_obj->has_filesystem(1);
            $mpdev_obj->fs_type("NTFS"); # hack need to fix this
            $mpdev_obj->mount_point($drive_letter);
            $mpdev_obj->mp_device_partition_list([$drive_letter]);
        } else {
            $mpdev_obj->is_partitioned(0);
            $mpdev_obj->is_mounted(0);
            $mpdev_obj->has_filesystem(0);
            $mpdev_obj->fs_type("");
            $mpdev_obj->mount_point("");
            $mpdev_obj->mp_device_partition_list([]);
        }

        # add this device to the devmgr
        $self->_add_device(-mp_device_obj => $mpdev_obj);
    }

    for (1..5) {
        ($status, @mpc_show_luns) = &ClientUtil::execute($sc_obj->{conn}, "$mpc_bin show luns", 1200);
        if (grep(/An error occurred|Error\:/i, @mpc_show_luns)) {
            $error = 1;
        } else {
            $error = 0;
            last;
        }
    }
    if ($error) {
        # could not validate show luns on this host, not much we can do here:
        $self->_set_err("'mpc show luns' failed after 5 retries, giving up.");
        return;
    }

    my $global_dsm;

    # corrolate each line of mpc show luns with the mpdev obj we registered earlier
    foreach my $line (@mpc_show_luns) {
        next if ($line !~ /^Disk\d+/);

        my @split = split(/\s+/, $line);
        # abort if this line does not have the right amount of fields
        if (scalar(@split) != 11) {
            $self->_set_err("Unable to parse 'mpc show lun output' line '$line'");
            return;
        }
        my $mp_device_name = $split[0];
        my $lun_id = $split[1];
        my $dsm = $split[2];
        my $lbpolicy = $split[3];
        my $num_paths = $split[4];
        my $serial = $split[5];
        my $cfmode = $split[6];
        my $filer = $split[7];
        my $proto = lc($split[8]);
        my $luntype = $split[9];
        my $lunpath = $split[10];

        # set the global dsm if its not defined
        if (!$global_dsm) {
            $global_dsm = $dsm;
        } else {
            # overwrite the dsm if we find a device that doesnt match the global dsm
            # and the global dsm is currently set to UNK
            if (($global_dsm eq 'UNK') and ($dsm ne $global_dsm)) {
                $global_dsm = $dsm;
            }
        }

        # find the mpdevobj that matches this one
        my $mpdev_obj = $self->get_device(-mp_device_name => $mp_device_name);
        if ($self->error()) {
            $self->_set_err("Unable to find a device that matches $mp_device_name");
            return;
        }

        $mpdev_obj->ontap_lun_path($lunpath);
        $mpdev_obj->ontap_lun_ostype($luntype);
        $mpdev_obj->protocol($proto);
        $mpdev_obj->ontap_vserver($filer); # XXX assuming cmode
        $mpdev_obj->alua("enabled"); # XXX assuming cmode

        # implicitly set the ontap mode
        $mpdev_obj->ontap_mode();

        $mpdev_obj->set_os_specific_data(load_balancing_policy => $lbpolicy,
                                         num_paths => $num_paths,
                                         dsm => $dsm,
                                         serial => $serial,
                                         cfmode => $cfmode);
    }

    # update the DSM field on the mpdevmgr object
    $self->set_os_specific_data(installed_dsm => $global_dsm);

    # refresh lun paths for all MPDev objects
    $self->refresh_paths();
    if ($self->error()) {
        $self->_set_err("Unable to refresh paths");
    }

    return;
}

###############################################################################
# Method:    refresh_paths()
# Objective: Discover all Lun Paths for all devices on the system
###############################################################################
sub _refresh_paths {
    my ($self, %args) = @_;

    my $global_dsm = $self->get_os_specific_data("installed_dsm");

    # depending on the DSM that is loaded, find paths to devices
    if (defined($global_dsm)) {
        my $all_lun_paths = {};
        if ($global_dsm =~ /ontap/) {
            $all_lun_paths = $self->_refresh_paths_ontapdsm();
        } elsif ($global_dsm =~ /msdsm|msisc/) {
            $all_lun_paths = $self->_refresh_paths_msdsm();
        } elsif ($global_dsm =~ /none|UNK/i) {
            # no dsm is installed (could be a virtual machine)
            # not necessarily an error
        } else {
            $self->_set_err("Detected an unsupported DSM '$global_dsm'. ");
            return;
        }

        foreach my $mp_device_name (keys %{$all_lun_paths}) {
            # find the mpdev object for this set of paths
            my $mpdev_obj = $self->get_device(-mp_device_name => $mp_device_name);
            if ($self->error()) {
                $self->_set_err("Unable to find a device that matches $mp_device_name");
                return;
            }

            my $lun_paths = $all_lun_paths->{$mp_device_name}->{lun_paths};
            my $protocol_path_counts = $all_lun_paths->{$mp_device_name}->{protocol_path_counts};

            if (scalar(keys %{$protocol_path_counts}) > 1) {
                $mpdev_obj->protocol("mixed");
            }

            my @path_objs = values %{$lun_paths};
            $mpdev_obj->_add_path_state(-mp_device_path_objs => \@path_objs);
        }
    } else {
        # no DSM luns found, not an error
    }

    # find all:
    #    (1) non NETAPP devices
    #    (2) non-dsm devices
    #    (3) devices with no path states
    # and add an empty path state to them
    foreach my $mpdev_obj ($self->get_device_objs()) {
        my @path_states = $mpdev_obj->get_path_states();
        if ($mpdev_obj->device_vendor !~ /NETAPP/i or
            $mpdev_obj->get_os_specific_data("dsm") =~ /none/i or
            not scalar(@path_states)) {
            $mpdev_obj->_add_path_state(-mp_device_path_objs => []);
        }
    }

    return;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _refresh_paths_msdsm()
# Objective: Discover the paths to the MS DSM Devices on the system
###############################################################################
sub _refresh_paths_msdsm {
    my ($self) = @_;

    my $sc_obj = $self->{sc_obj};
    my $mpc_bin = $sc_obj->{WinClient}->{cmd_mpc};

    my @init_objs = $sc_obj->get_initiators();
    @init_objs = grep { $_->get_protocol() =~ /iscsi/i } @init_objs;

    # pull out all the nodename for this initiator
    my ($iscsi_nodename) = map { $_->get_nodename() } @init_objs;

    my $all_lun_paths;

    my ($status, @mpc_show_mpio_lunpaths);

    ($status, @mpc_show_mpio_lunpaths) = &ClientUtil::execute($sc_obj->{conn}, "${mpc_bin} mpio show lunpaths");

    foreach my $line (@mpc_show_mpio_lunpaths) {
        $line =~ s/^\s+//;
        next if ($line !~ /^Disk\d+/);
        my @line = split(/\s+/, $line);
        my $disk_name = $line[0];
        my $path_id = $line[1];
        my $path_state = $line[2];
        my $path_port = $line[3];
        my $path_bus = $line[4];
        my $path_target = $line[5];
        my $path_lun_id = $line[6];
        my $path_protocol = lc($line[7]);

        # we do not know alua states in msdsm from this output
        my $alua_state = "UNKNOWN";
        $path_state = "active" if ($path_state eq "1");
        $path_state = "dead"   if ($path_state eq "0");

        my $mpdev_obj = $self->get_device(-mp_device_name => $disk_name);
        if ($self->error()) {
            $self->_set_err("Unable to find a device that matches $disk_name");
            return;
        }

        # create the mpdev path object
        # since we know we will never be able to create lif relationships between MPDevPaths and NACL::CS::NetworkInterface
        # objects, we will explicitly set the 'ignore_lif_relationship_warnings' flag
        my $mpdev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(path_name => $path_id,
                                                                     path_state => $path_state,
                                                                     protocol => $path_protocol,
                                                                     alua_state => $alua_state,
                                                                     ignore_lif_relationship => 1);

        if ($path_protocol =~ /fcp/i) {
            $mpdev_path_obj->target_wwpn("");
            $mpdev_path_obj->target_wwnn("");
            $mpdev_path_obj->initiator_wwpn("");
            $mpdev_path_obj->initiator_wwnn("");
            $mpdev_path_obj->lun_id($mpdev_obj->lun_id());
        } elsif ($path_protocol =~ /iscsi/i) {
            $mpdev_path_obj->initiator_iqn($iscsi_nodename);
            $mpdev_path_obj->target_iqn("");
            $mpdev_path_obj->target_iscsi_alias("");
            $mpdev_path_obj->target_iscsi_session("");
            $mpdev_path_obj->target_ip_address("");
            $mpdev_path_obj->target_ip_port("");
            $mpdev_path_obj->tpgt("");
        }

        $mpdev_path_obj->set_os_specific_data(disk_name => $disk_name,
                                              dsm => "msdsm",
                                              path_bus => $path_bus,
                                              path_port => $path_port,
                                              path_target => $path_target,
                                              path_protocol => $path_protocol);

        # store all the path objects in the all lun paths hash for later assignment
        $all_lun_paths->{$disk_name}->{lun_paths}->{$path_id} = $mpdev_path_obj;
        $all_lun_paths->{$disk_name}->{protocol_path_counts}->{$path_protocol}++;
    }

    return $all_lun_paths;
}

###############################################################################
# Method:    _refresh_paths_ontapdsm()
# Objective: Discover the paths to the ONTAP DSM Devices on the system
###############################################################################
sub _refresh_paths_ontapdsm {
    my ($self) = @_;

    my $sc_obj = $self->{sc_obj};
    my $mpc_bin = $sc_obj->{WinClient}->{cmd_mpc};

    my @init_objs = $sc_obj->get_initiators();
    @init_objs = grep { $_->get_protocol() =~ /iscsi/i } @init_objs;

    # pull out all the nodename for this initiator
    my ($iscsi_nodename) = map { $_->get_nodename() } @init_objs;

    #my $iscsi_info = $self->{sc_obj}->get_iscsi_info());

    my $all_lun_paths;

    my ($status, @mpc_show_ontapdsm_lunpaths);

    ($status, @mpc_show_ontapdsm_lunpaths) = &ClientUtil::execute($sc_obj->{conn}, "${mpc_bin} ontapdsm show lunpaths");

    foreach my $line (@mpc_show_ontapdsm_lunpaths) {
        next if ($line !~ /^Disk\d+/);
        my @line = split(/\s+/, $line);
        my $disk_name = $line[0];
        my $lbpolicy = $line[1];
        my $dsm_id = $line[2];
        my $path_id = $line[3];
        my $path_state = $line[4];
        my $path_protocol = lc($line[5]);
        my $is_proxy = $line[6];
        my $is_primary = $line[7];
        my $reserve = $line[8];
        my $path_weight = $line[9];
        my $target_port_name = $line[10];

        my $alua_state;
        $alua_state = "AO" if ($is_primary =~ /True/i);
        $alua_state = "ANO" if ($is_proxy =~ /True/i);
        $alua_state = "DOWN" if ($path_state !~ /Active|Passive/i);

        my $mpdev_obj = $self->get_device(-mp_device_name => $disk_name);
        if ($self->error()) {
            $self->_set_err("Unable to find a device that matches $disk_name");
            return;
        }

        my $mpdev_path_obj = new San::SanClient::MPDevMgr::MPDevPath(path_name => $path_id,
                                                                     path_state => $path_state,
                                                                     protocol => $path_protocol,
                                                                     alua_state => $alua_state);

        if ($path_protocol =~ /fcp/i) {
            $mpdev_path_obj->target_wwpn("");
            $mpdev_path_obj->target_wwnn("");
            $mpdev_path_obj->initiator_wwpn("");
            $mpdev_path_obj->initiator_wwnn("");
            $mpdev_path_obj->lun_id($mpdev_obj->lun_id());
        } elsif ($path_protocol =~ /iscsi/i) {
            $mpdev_path_obj->initiator_iqn($iscsi_nodename);
            $mpdev_path_obj->target_iqn("");
            $mpdev_path_obj->target_iscsi_alias("");
            $mpdev_path_obj->target_iscsi_session("");
            $mpdev_path_obj->target_ip_address("");
            $mpdev_path_obj->target_ip_port("");
            $mpdev_path_obj->tpgt("");
        }

        $mpdev_path_obj->set_os_specific_data(disk_name => $disk_name,
                                              dsm_id => $dsm_id,
                                              dsm => "ontap",
                                              load_balancing_policy => $lbpolicy,
                                              is_proxy => $is_proxy,
                                              is_primary => $is_primary,
                                              reserve => $reserve,
                                              path_weight => $path_weight);

        # store the ontap target port name value on the path
        $mpdev_path_obj->ontap_target_port_name($target_port_name);

        # store all the path objects in the all lun paths hash for later assignment
        $all_lun_paths->{$disk_name}->{lun_paths}->{$path_id} = $mpdev_path_obj;
        $all_lun_paths->{$disk_name}->{protocol_path_counts}->{$path_protocol}++;
    }

    return $all_lun_paths;
}

1;
