# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface I/O tools
## @author muthug@netapp.com
## @status public
## @pod here

package SanIo;

use strict;

use San::SanClient;
use San::SanClient::ClientUtil;
use Error;
use San::SanClient::SanIo::Sio;
use San::SanClient::Dt;

use vars qw(
	    $_CLIENT_CNT    
	    @ISA
	    $JOB_CNT
	    );

@ISA=qw(Error);


use constant IO_TOOL_SIO                  => 1;
use constant IO_TOOL_DT                   => 2;
use constant IO_TOOL_FILER_SIO            => 3;

################################################################################
# Subroutine Name: new ()
# Objective:       Create a new SanIo object 
# Details:         See POD documentation at the end of this file
################################################################################
sub new 
{
   my ($class, %args) = @_;
   

   #
   # Create a hash to store attributes and bless the hash to create an object.
   #
   my $self = bless ( {}, $class);

   my $io_tool_flag       = 0;
   my $client_conn_flag   = 0;
   my $client_io_dev_flag = 0;
   my $wait_flag          = 1;
   my %io_tool_args;
   my $io_tool;
   my $san_client_obj;
   my $client_io_dev;

   $self->_clear_err();

   # Verify there are no invalid input parameters 
   foreach my $param (keys %args) {
      if ($param =~ /-io_tool$|io_tool$/) {
         $io_tool = $args{$param};
         if (not length($io_tool))  {
            $self->_set_err("$param must be one of "    .
                         "('sio' or '1', 'dt' or '2' or 'filersio' or '3'), " .
                         "not: '$io_tool'");
            return $self;
         }
         if( $io_tool =~ m/^sio$/i || $io_tool == IO_TOOL_SIO ){
            $self->{io_tool} = IO_TOOL_SIO;
         } elsif ($io_tool =~ m/^dt$/i || $io_tool == IO_TOOL_DT) {
            $self->{io_tool} = IO_TOOL_DT;
         } elsif ($io_tool =~ m/^filersio$/i || $io_tool == IO_TOOL_FILER_SIO) {
            $self->{io_tool} = IO_TOOL_FILER_SIO;
         } else { 
            $self->_set_err("$param must be one of "    .
                         "('sio' or '1', 'dt' or '2' or 'filersio' or '3'), " .
                         "not: '$io_tool'");
            return $self;
         }
         $io_tool_flag = 1;
      }
      elsif ($param =~ /-san_client_obj$|san_client_obj$/) {
         $san_client_obj = $args{$param};
         $client_conn_flag = 1;
      }
      elsif ($param =~ /-wait_flag$|wait_flag$/) {
         $wait_flag = $args{$param};
      }
      elsif ($param =~ /-client_io_dev$|client_io_dev$/) {
         $client_io_dev = $args{$param};
         $client_io_dev_flag = 1;
      } 
      else {
         #
         # Pass all other args down to the IO tool.
         #
         $io_tool_args{$param} = $args{$param};
      }
   }

   #
   # Check for required arguments
   #
   if( $io_tool_flag == 0 ) {
      $self->_set_err("'-io_tool' parameter must be specified");
      return $self;
   }
   if( $client_conn_flag == 0 ) {
      $self->_set_err("'-san_client_obj' parameter must be specified");
      return $self;
   }
   if( $client_io_dev_flag == 0 ) {
      $self->_set_err("'-client_io_dev' parameter must be specified");
      return $self;
   }

   #
   # Initialize attributes
   #
   $self->{io_tool_obj}    = "";
   $self->{san_client_obj} = $san_client_obj;
   $self->{client_conn}    = $self->{san_client_obj}->get_connection();
   $self->{wait_flag}      = $wait_flag;

   #
   # Complete the io_tools hash
   #
   $io_tool_args{client_conn}   = $self->{client_conn};
   $io_tool_args{wait_flag}     = $wait_flag;
   $io_tool_args{client_io_dev} = $client_io_dev;

   #
   # Create appropriate IO tool (sio, dt or filersio) object 
   #
   if( $self->{io_tool} == IO_TOOL_SIO) {
      $self->{io_tool_obj} = new Sio(%io_tool_args);
      if ($self->{io_tool_obj}->error()) {
         $self->_clear_err();
         $self->_set_err($self->{io_tool_obj}->errmsg());
         $self->{io_tool_obj} = "";
         return $self;
      }
   } elsif ($self->{io_tool} == IO_TOOL_DT){
      $self->_set_err("dt is not currently supported");
      return $self;
   } else {
      $self->_set_err("$io_tool is not currently supported");
      return $self;
   }
   return $self;
} # end new()

################################################################################
# Subroutine Name: san_io_run ()
# Objective:       To start the IO on the given device list on the given client 
# Details:         See POD documentation at the end of this file
################################################################################
sub san_io_run ()
{
   my $self          = shift;


   $self->_clear_err();
   my $status;
   my %results;

   if( $self->{io_tool_obj} eq "" ) {
      $self->_set_err("Sio Object is not initialized");
      return $ClientUtil::FALSE;      
   }

   ($status, %results) = $self->{io_tool_obj}->start_io();
   if( $status == $ClientUtil::FALSE) {
      $self->_set_err($self->{io_tool_obj}->errmsg());
      return $ClientUtil::FALSE, %results;      
   }
   return $ClientUtil::TRUE, %results;      
}

################################################################################
# Subroutine Name: san_io_stop ()
# Objective:       To stop the IO on the client 
# Details:         See POD documentation at the end of this file
################################################################################
sub san_io_stop ()
{
   my $self          = shift;
   my $status;

   $self->_clear_err();
   $status = $self->{io_tool_obj}->stop_io();
   if( $status == $ClientUtil::FALSE) {
      $self->_set_err($self->{io_tool_obj}->errmsg());
      return $ClientUtil::FALSE
   }
   return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: san_io_verify_data ()
# Objective:       To verify the data on the device list 
# Details:         See POD documentation at the end of this file
################################################################################
sub san_io_verify_data ()
{
   my $self          = shift;

   $self->_clear_err();
   return $self->{io_tool_obj}->verify_data();
}

################################################################################
# Subroutine Name: san_io_is_alive ()
# Objective:       To verify the io is still running on the background
# Details:         See POD documentation at the end of this file
################################################################################
sub san_io_is_alive ()
{
   my $self          = shift;

   $self->_clear_err();
   return $self->{io_tool_obj}->is_alive();
}


################################################################################
# Subroutine Name: san_io_wait ()
# Objective:       To wait until io completes
# Details:         See POD documentation at the end of this file
################################################################################
sub san_io_wait ()
{
   my $self          = shift;

   my $status;
   my %results;

   $self->_clear_err();
   if( $self->{io_tool_obj} eq "" ) {
      $self->_set_err("Sio Object is not initialized");
      return $ClientUtil::FALSE;      
   }

   ($status, %results) = $self->{io_tool_obj}->wait_io();
   if( $status == $ClientUtil::FALSE) {
      $self->_set_err($self->{io_tool_obj}->errmsg());
      return $ClientUtil::FALSE, %results;      
   }
   return $ClientUtil::TRUE, %results;      
}

1;


#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

SanIo.pm

=head1 Authors

=begin html 

David Pressley
Muthukumaran Govindarajan

=end html

=head1 Description

SanIo provides an API interface to Run IO to the IO device using specify IO tool

=head1 See Also

Package inherits from the following modules. See modules for interface detail

B<Sio.pm>
B<Dt.pm>

=head1 Attributes

=head2 B<io_tool>

scalar = To determine which IO tool to use for IO operation

=head2 B<san_client_obj>

reference = SanClient object refernce

=head2 I<wait_flag> (default = 1)

scalar = Determine the IO operation to wait for completion or run in the background

=head2 B<client_io_devs>

reference = holds the client io device list

=head2 I<read_pct> for SIO (default = 50)

scalar = read percentage for the io operation on the given device list. range [0-100]

=head2 I<random_pct> for SIO (default = 50)

scalar = random percentage for the io operation on the given device list. range [0-100]

=head2 I<block_size> for SIO (default = 1024)

scalar = Io block size

=head2 B<first_byte> for SIO 

scalar = Starting location of the IO operation in the given io device

=head2 B<last_byte> for SIO

scalar = Ending location of the IO operation in the given io device

=head2 I<run_time> for SIO (default = 3600)

scalar = Specifiy the the amount of time to run the io in seconds

=head2 I<num_threads> for SIO (default = 10)

scalar = Numbers of threads to use by sio

=head2 I<pass_thru> 

scalar = Other valid parameters necessary for the IO tools to execute

=head1 Public Methods

=head2 B<new(%args)>

Create a SanIo object to run IO to the given list of IO devices on given client

=over 4

=item Inputs

=item * -io_tool

IO tool to use for running Io. 
Valid arguments are  "sio" or 1 for SIO , 
                     "dt"  or 2 for DT
                     "filersio" or 3 for FILER_SIO.

=item * -wait_flag (default=1)

To determine that san_io_run function to wait or return immediately after starting Io.
Valid arguments are  1 to start the io and wait to complete
                     0 to start the operation and function will return immediately

=item * for other parameters check the Attributes

=item Outputs

return the SanIo objects

=item Examples

   my $san_obj = new San(\*ntest::NTEST_STDOUT);
   if ($san_obj->error())
   {
       logresult('FATAL', "Failed to instantiate San object" . $san_obj->errmsg());
   }
   
   #
   ## @step <u>Objective: Setup the San.</u>
   #
   $san_obj->setup(-filers      =>$FILERS,
                   -clients     =>$CLIENT,
                   -protocol    =>$PROTOCOL,
                   -lun_list    =>$LUN_LIST,
                   -aggr        =>$AGGR,
                   -num_disks   =>$NUM_DISKS,
                   -vol         =>$LUN_PREFIX,
                   -vol_size    =>$VOL_SIZE,
                   -device_type =>"raw",
                   -skip_setup  =>$SKIP_SETUP,
                   -skip_config =>$SKIP_CONFIG,
                  );
   
   if ($san_obj->error())
   {
       my $error_msg = join "\n", $san_obj->errmsg();
       $error_count++;
       $error_string .= "Failed to setup San object: $error_msg\n";
       goto CLEANUP;
   }
   
   #
   ## @step Objective: Get the devices for the client.
   #
   if ($SKIP_SETUP == 0)
   {
       @device_array = $san_obj->get_devices(-client      =>$CLIENT,
                                             -vol         =>$LUN_PREFIX,
                                             -lun_list    =>$LUN_LIST,
                                             -device_type =>"raw",
                                             -protocol    =>$PROTOCOL);
       if ($san_obj->error())
       {
           my $error_msg = join "\n", $san_obj->errmsg();
           $error_count++;
           $error_string .= "Failed to find devices for " .
                            "San object: $error_msg\n";
           goto CLEANUP;
       }
       else
       {
           $client_io_devs .= join " ", @device_array;
       }
   }
   else
   {
       logcomment("Skipping setup and using parameter CLIENT_IO_DEVS");
       unless (length($CLIENT_IO_DEVS))
       {
           $error_count++;
           $error_string .= "No devices in CLIENT_IO_DEVS " .
                            "with SKIP_SETUP set\n";
           goto CLEANUP;
       }
       $client_io_devs = $CLIENT_IO_DEVS;
   }
   logcomment("Using devices $client_io_devs");
   
   #
   # Find the client object to send to SanIo
   #
   my $client_obj = $san_obj->get_client($CLIENT);
   if ($san_obj->error())
   {
       my $error_msg = join "\n", $san_obj->errmsg();
       $error_count++;
       $error_string .= "Failed to find the " .
                        "Client object: $error_msg\n";
       goto CLEANUP;
   }
   
   #
   ## @step <u>Objective: Instantiate a SanIo Object.</u>
   #
   my $san_io_obj = new SanIo(
           -io_tool => "sio",
           -san_client_obj => $client_obj,                  # Client to run I/O 
           -wait_flag => 1,                              # wait for io to complete
           -client_io_dev => $client_io_devs,              # client_io_dev 
           -read_pct   => $CLIENT_IO_READ_PCT,          # read_pct      
           -random_pct => $CLIENT_IO_RANDOM_PCT,        # random_pct    
           -num_threads => $CLIENT_IO_NUM_THREADS,      # num_threads   
           -block_size => $CLIENT_IO_BLOCK_SIZE,       # block_size    
           -first_byte  => $CLIENT_IO_DEV_OFFSET_BEGIN, # first_byte    
           -last_byte   => $CLIENT_IO_DEV_OFFSET_END,   # last_byte     
           -run_time     => $CLIENT_IO_RUN_TIME,         # time in seconds     
           -pass_thru   => $CLIENT_IO_PASS_THRU        # sio_switches
           ); 
   if ($san_io_obj->error())
   {
       my $error_msg = join "\n", $san_io_obj->errmsg();
       $error_count++;
       $error_string .= "Failed to instantiate " .
                        "SanIo object: $error_msg\n";
       goto CLEANUP;
   }

=back

=head2 B<san_io_run()>

Run IO to the given list of IO devices using specific IO tool

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful and the hash value.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Examples

   my($status, %results) = $san_io_obj->san_io_run();
   if ($san_io_obj->error())
   {
       my $error_msg = join "\n", $san_obj->errmsg();
       $error_count++;
       $error_string .= "Failed to execute I/O: $error_msg\n";
       goto CLEANUP;
   }
   # If SIO, on successful run it will give following values in the hash
   print "error_count = $results{'error_count'}\n";
   print "iops = $results{'iops'}\n";
   print "tput = $results{'tput'}\n";
   print "latency_calculated = $results{'latency_calculated'}\n";
   print "latency_actual = $results{'latency_actual'}\n";
   print "read = $results{'read'}\n";
   print "random = $results{'random'}\n";
   print "ios = $results{'ios'}\n";
   print "seconds = $results{'seconds'}\n";
   print "threads = $results{'threads'}\n";
   print "blksize = $results{'blksize'}\n";
   print "csw_io = $results{'csw_io'}\n";
   print "int_io = $results{'int_io'}\n";
   print "sys = $results{'sys'}\n";
   print "usr = $results{'usr'}\n";
   print "wt = $results{'wt'}\n";
   print "idl = $results{'idl'}\n";
   print "uio = $results{'uio'}\n";
   print "ukb = $results{'ukb'}\n";
   print "uuio = $results{'uuio'}\n";
   print "tuio = $results{'tuio'}\n";

=back

=head2 B<san_io_stop()>

Stop the IO operation if it is running on the background

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Examples

   my($status) = $san_io_obj->san_io_stop();
   if ($san_io_obj->error())
   {
       my $error_msg = join "\n", $san_obj->errmsg();
       $error_count++;
       $error_string .= "Failed to execute I/O: $error_msg\n";
       goto CLEANUP;
   }

=back

=head2 B<san_io_is_alive()>

Verify that the IO operation is still running on the background. This function will be useful if you run IO on the background

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if IO is still going on else 0.

=item Examples

    # Check if DataTest I/O has completed on the volume 
    # every 5 seconds.
    sleep 5 while($san_io_obj->san_io_is_alive());

=back

=head2 B<san_io_verify_data()>

Verify the data pattern written on the device list. 

=over 4 

=item Inputs

data_pattern : optional. If this parameter is not passed it will be check against 
the pattern what it worte when you created this object. 

=item Outputs

If no pattern is passed when creating object this function will return error.
Returns a scalar value of 1 if the pattern matched. If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.
Returns 1 if an error occured in the previous call, 0 otherwise.

=item Examples

    $SanIoObj->san_io_verify_data("A");
       or
    $SanIoObj->san_io_verify_data();
      
=back


=head2 B<san_io_wait()>

Wait for IO to complete and return the output from io tool. This function will be useful if you run IO in the background

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful and the hash value.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Examples

   my($status, %results) = $san_io_obj->san_io_wait();
   if ($san_io_obj->error())
   {
       my $error_msg = join "\n", $san_obj->errmsg();
       goto CLEANUP;
   }

=back


=cut

