# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to the Sio I/O tool
## @author muthug@netapp.com
## @status public
## @pod here

package Sio;

use strict;

use Tharn;
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use vars qw(
            $_CLIENT_CNT    
            @ISA
            $JOB_CNT
            );

@ISA=qw(Error SanLog);


################################################################################
# Subroutine Name: new ()
# Objective:       Create a new SIO object 
# Details:         See POD documentation at the end of this file
################################################################################
sub new 
{
   my ($class, %args) = @_;

   my $client_conn      = undef;
   my $client_io_dev    = undef;
   my $wait_flag        = 1;
   my $read_pct         = 50;
   my $random_pct       = 50;
   my $block_size       = 1024;
   my $first_byte       = 0;
   my $last_byte        = "1g";
   my $run_time         = 3600;  # 1 hour
   my $num_threads      = 10;
   my $other_sio_params = "";
   my $sio_subtest_id   = "run_sio";
   
   #
   # Create a hash to store attributes and bless the hash to create an object.
   #
   my $self = bless ( {}, $class);
   
   #
   # Parse the argument hash and overwrite the default for any known argument.
   #
   foreach my $param (keys %args) {
      if ($param =~ /-client_conn|client_conn/) {
         $client_conn = $args{$param};
      }
      elsif ($param =~ /-wait_flag|wait_flag/) {
         $wait_flag = $args{$param};
      }
      elsif ($param =~ /-client_io_dev|client_io_dev/) {
         $client_io_dev = $args{$param};
      } 
      elsif ($param =~ /-read_pct|read_pct/) {
         $read_pct = $args{$param};
      } 
      elsif ($param =~ /-random_pct$|random_pct$/) {
         $random_pct = $args{$param};
      } 
      elsif ($param =~ /-block_size$|block_size$/) {
         $block_size = $args{$param};
      } 
      elsif ($param =~ /-first_byte$|first_byte$/) {
         $first_byte = $args{$param};
      } 
      elsif ($param =~ /-last_byte$|last_byte$/) {
         $last_byte = $args{$param};
      } 
      elsif ($param =~ /-run_time$|run_time$/) {
         $run_time = $args{$param};
      } 
      elsif ($param =~ /-num_threads$|num_threads$/) {
         $num_threads = $args{$param};
      } 
      elsif ($param =~ /-sio_subtest_id$|sio_subtest_id$/) {
         $sio_subtest_id = $args{$param};
      } 
      elsif ($param =~ /-pass_thru$|pass_thru$/) {
         $other_sio_params = $args{$param};
      } 
      else {
         #
         # Ignore
         #
         $self->sanlog(-msg => "Sio::new() Ignoring argument $param");
      }
   } # end foreach arg

   #
   # Initialize attributes
   #
   $self->{client_conn}      = $client_conn;
   $self->{wait_flag}        = $wait_flag;
   $self->{client_io_devs}   = $client_io_dev;
   $self->{read_pct}         = $read_pct;
   $self->{random_pct}       = $random_pct;
   $self->{block_size}       = $block_size;
   $self->{first_byte}       = $first_byte;
   $self->{last_byte}        = $last_byte;
   $self->{run_time}         = $run_time;
   $self->{num_threads}      = $num_threads;
   $self->{other_sio_params} = $other_sio_params;
   $self->{sio_subtest_id}   = $sio_subtest_id;

   $self->{sub_test_obj}     = "";
   $self->{sub_test_pid}     = "";

   my ($status, $errstr) = $self->_sio_params_validation();
   if($status == $ClientUtil::FALSE) {
      $self->_set_err( "$errstr");
      return $self;
   }
   return $self;
} # end new()

################################################################################
# Subroutine Name: start_io ()
# Objective:       To start the IO on the given device list on the given client 
# Details:         See POD documentation at the end of this file
################################################################################
sub start_io ()
{
   my ($self) = @_;

   $self->_clear_err();

   my %results = 0;
   my ($status, $errstr) = $self->_sio_params_validation();

   if($status == $ClientUtil::FALSE) {
      $results{'error_count'} = 1;
      $results{'error_msg'} = $errstr;
      $self->_set_err( $errstr);
      return ($ClientUtil::FALSE, %results);      
   }

   ($status, %results) = $self->_invoke_sio();
   if($status == $ClientUtil::FALSE) {
      $self->_set_err( $results{'error_msg'});
      return ($ClientUtil::FALSE, %results);
   }
   return ($ClientUtil::TRUE, %results);
}
################################################################################
# Subroutine Name: stop_io ()
# Objective:       To stop the IO on the client 
# Details:         See POD documentation at the end of this file
################################################################################
sub stop_io ()
{
   my ($self) = @_;
   
   $self->_clear_err();

   if($self->{sub_test_obj} eq "") {
      $self->_set_err("Sio subtest is not running or not initialized");
      return $ClientUtil::FALSE;
   }
   if ($self->{sub_test_obj}->is_alive() == 1) {
      $self->{sub_test_obj}->kill();
      $self->{sub_test_obj} = "";
      $self->{sub_test_pid} = "";
   }
   return $ClientUtil::TRUE;
}
################################################################################
# Subroutine Name: verify_data ()
# Objective:       To verify the data on the given device list 
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_data ()
{
   my ($self) = @_;

   $self->_clear_err();
   
   $self->sanlog(-msg => "NOT YET IMPLEMENTED:");
   return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: is_alive ()
# Objective:       To verify the IO operation is still running or not
# Details:         See POD documentation at the end of this file
################################################################################
sub is_alive ()
{
   my ($self) = @_;
   
   $self->_clear_err();

   if($self->{sub_test_obj} eq "") {
      $self->_set_err("Sio subtest is not running or not initialized");
      return $ClientUtil::FALSE;
   }
   return $self->{sub_test_obj}->is_alive(); # return 1 if alive, 0 otherwise
}


################################################################################
# Subroutine Name: wait_io ()
# Objective:       To wait until IO completes
# Details:         See POD documentation at the end of this file
################################################################################
sub wait_io ()
{
   my ($self) = @_;
   
   $self->_clear_err();
   
   if($self->{sub_test_obj} eq "") {
      $self->_set_err("Sio subtest is not running or not initialized");
      return $ClientUtil::FALSE;
   }
   # wait for io to complete, if it is running...........
   if ($self->{sub_test_obj}->is_alive() == 1) {
      $self->sanlog(-msg => "Waiting for IO to complete.....");
      $self->{sub_test_obj}->wait_finish();
   }   
   #
   # check out the result return from the subtest
   #
   my $status = 0;
   my %results = 0;
   ($status, %results) = $self->get_io_result();
   if($status == $ClientUtil::FALSE) { 
      $self->_set_err($self->errmsg());
      return $ClientUtil::FALSE, %results;      
   }
   return ($ClientUtil::TRUE, %results);
}

################################################################################
# Subroutine Name: get_io_result ()
# Objective:       To get the IO result after IO completes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_io_result ()
{
   my ($self) = @_;
   
   $self->_clear_err();
   
   if($self->{sub_test_obj} eq "") {
      $self->_set_err("Sio subtest is not running or not initialized");
      return $ClientUtil::FALSE;
   }
   if ($self->{sub_test_obj}->is_alive() == 1) {
      $self->_set_err("Sio subtest is still running");
      return $ClientUtil::FALSE;
   }
   
   #
   # check out the result return from the subtest
   #
   my %results = 0;
   $results{'error_cont'} = 0;
   $results{'error_msg'} = "";
   my @ret_results = $self->{sub_test_obj}->rslts();
   my @subtest_messages = $self->{sub_test_obj}->results();
   my @temp_result = split("\n", $subtest_messages[0]);

   foreach my $result (@ret_results) {
      my $result_type = $result->type();
      my $no_msg = $result->messages();
      $self->sanlog(-msg => "RESULT TYPE = $result_type", -msg_level => 7);
      $self->sanlog(-msg => "No of Messages = $no_msg", -msg_level => 7);
      if($result_type =~ /PASS/){ 
         #
         # if subtest passes, subtest prints the hash value thorugh logcomment
         # output will be: first line <key> and next line <value>
         # 
         for(my $i=0; $i < @temp_result; $i += 2){
            $results{$temp_result[$i]} = $temp_result[$i+1];
         }
      }else {
         if( $result_type =~ /FAIL/){
            #
            # index 0 will have key = 'error_count'
            # index 1 will have value for the key='error_count'
            # index 2 will have key = 'error_msg';
            # rest will have the error message;
            $results{$temp_result[0]} = $temp_result[1];
            $results{'error_msg'} = $temp_result[3];
            for(my $i=4; $i < $no_msg; $i++){
               $results{'error_msg'} .= "\n" . $temp_result[$i];
            }
         } else{
            $results{'error_cont'}++;
            $results{'error_msg'} .= $result . "\n";
         }
      }
   } # end foreach my $result (@ret_results)
   
   $self->sanlog(-msg => "Sio::get_io_result(): "           .
    "**********RESULT =\n"                                  .
    "error_count        = $results{'error_count'}\n"        .
    "error message      = $results{'error_msg'}\n"          .
    "iops               = $results{'iops'}\n"               .
    "tput               = $results{'tput'}\n"               .
    "latency_calculated = $results{'latency_calculated'}\n" .
    "latency_actual     = $results{'latency_actual'}\n"     .
    "read               = $results{'read'}\n"               .
    "random             = $results{'random'}\n"             .
    "ios                = $results{'ios'}\n"                .
    "seconds            = $results{'seconds'}\n"            .
    "threads            = $results{'threads'}\n"            .
    "blksize            = $results{'blksize'}\n"            .
    "csw_io             = $results{'csw_io'}\n"             .
    "int_io             = $results{'int_io'}\n"             .
    "sys                = $results{'sys'}\n"                .
    "usr                = $results{'usr'}\n"                .
    "wt                 = $results{'wt'}\n"                 .
    "idl                = $results{'idl'}\n"                .
    "uio                = $results{'uio'}\n"                .
    "ukb                = $results{'ukb'}\n"                .
    "uuio               = $results{'uuio'}\n"               .
    "tuio               = $results{'tuio'}\n",
    -msg_level => 7);

   # interpret the result and set the success flag
   if( $results{'error_count'} > 0 ) {
      $self->_set_err( $results{'error_msg'});
      return ($ClientUtil::FALSE, %results);
   }
   return ($ClientUtil::TRUE, %results);
}

################################################################################
# Subroutine Name: _sio_params_validation ()
# Objective:       To verify the sio parameters sent to new()
# Details:         See POD documentation at the end of this file
################################################################################
sub _sio_params_validation() 
{
   my ($self) = @_;

   if( $self->{client_conn} eq "" ) {
      return $ClientUtil::FALSE, "Need Client connection";
   }

   if( $self->{wait_flag} >= 1) {
      $self->{wait_flag} = 1;
   } elsif ($self->{wait_flag} <= 0){
      $self->{wait_flag} = 0;
   }else {
      $self->{wait_flag} = 1;
   }

   if( $self->{client_io_devs} eq "" ) {
      return $ClientUtil::FALSE, "Need Client IO device list";
   }
   
   if ($self->{read_pct} !~ m/^\d+$/ || $self->{read_pct} < 0 || $self->{read_pct} > 100)
   {
      return $ClientUtil::FALSE, "Read Percent must be (0-100), not: $self->{read_pct}"; 
   }
   
   if ($self->{random_pct} !~ m/^\d+$/ || $self->{random_pct} < 0 || $self->{random_pct} > 100)
   {
      return $ClientUtil::FALSE, "Random Percent must be (0-100), not: $self->{random_pct}"; 
   }


   if ($self->{block_size} !~ m/^\d+[k]?$/ ) {
      return $ClientUtil::FALSE, "block_size must be 512, 1k, 4k, etc. not: $self->{block_size}"; 
   } else {
      my $blksize =  $self->{block_size};
      $blksize =~ s/(\d+).*/$1/;
      if ($self->{block_size} =~ /k/) {
         $blksize = $blksize*1024;
      }
      if ($blksize < 512 || $blksize > 262144 ) {
         return $ClientUtil::FALSE, "block_size must be 512 to 262144, not: $blksize"; 
      }
   }
   
   if ($self->{first_byte} !~ m/^\d+[kgmt]?$/ ) {
      return $ClientUtil::FALSE, "first_byte must be numeric, not: $self->{first_byte}"; 
   }

   if ($self->{last_byte} !~ m/^\d+[kgmt]?$/ ) {
      return $ClientUtil::FALSE, "last_byte must be numeric, not: $self->{last_byte}"; 
   }

   if ($self->{run_time} == 0 && ($self->{other_sio_params} !~ /[fillonce|numops]/)) {
      return $ClientUtil::FALSE, "If run_time = 0, you must use -fillonce or -numops option"; 
   } elsif ($self->{run_time} !~ m/^\d+$/ ||  ($self->{run_time} > 0 && $self->{run_time} < 10)) {
      return $ClientUtil::FALSE, "run_time must be a 10 seconds or greater, not: $self->{run_time}"; 
   }

   if ($self->{num_threads} !~ m/^\d+$/ || $self->{num_threads} < 1 || $self->{num_threads} > 256) {
       return $ClientUtil::FALSE, "threads must be 1 to 256, not: $self->{num_threads}"; 
   }

   # Verify the host is alive
   my $host = host $self->{client_conn}->{hostp}->{hostname};
   my $hostid = $host->id();
   my $hostip = $host->default_ip();
#   if (!host_wait($hostip)) {
#       return $ClientUtil::FALSE, "Client $hostid at $hostip does not appear to be alive";
#   }
   
   return $ClientUtil::TRUE, "OK";
}

################################################################################
# Subroutine Name: _invoke_sio ()
# Objective:       To start the IO on the given device list on the given client 
# Details:         
################################################################################
sub _invoke_sio ()
{
   my ($self) = @_;
   my %results = 0;

   # need to handle the CLIENT_IO_DEVS_PER_SIO

   $self->{sub_test_obj} = subtest(\&_run_sio, 
                                   '-bg'   ,
                                   '-runid'   , $self->{sio_subtest_id},
                                   '-onexit' , \&_subtest_exit_handler,
                                   '--',
                                   $self, $self->{client_conn},$self->{client_io_devs},
                                   $self->{read_pct}, $self->{random_pct}, $self->{block_size},
                                   $self->{first_byte}, $self->{last_byte}, $self->{run_time}, 
                                   $self->{num_threads}, $self->{other_sio_params});

   if( $self->{sub_test_obj} eq '') {
      $results{'error_count'} = 1;
      $results{'error_msg'} = "Unable to launch the subtest";
      return ($ClientUtil::FALSE, %results);
   }

   $self->{sub_test_pid} = $self->{sub_test_obj}->pid();

   if( $self->{wait_flag} == 1) {
      if ($self->{sub_test_obj}->is_alive() == 1) {
         $self->{sub_test_obj}->wait_finish();
      }
      #
      # check out the result return from the subtest
      #
      my $status = 0;
      my %results = 0;
      ($status, %results) = $self->get_io_result();
      if($status == $ClientUtil::FALSE) { 
         $self->_set_err($self->errmsg());
         return ($ClientUtil::FALSE, %results);      
      }
   } else {
      $results{'error_count'} = 0;
      $results{'error_msg'} = "I/O Started Successfully........";
   }

   return ($ClientUtil::TRUE, %results);
} # end invoke_io()

################################################################################
# Subroutine Name: _run_sio ()
# Objective:       Function to invoke the sio tool with specified arguments 
## All input arguments are supported as hash keyword-value pairs.
##
## It returns information from sio in a hash.
## The following hash keys may be present:
## version     - the version of sio
## error_count - the number of errors seen
## iops        - I/O operations per second
## tput        - KB per second
## latency_actual     - latency in milliseconds
## latency_calculated - minimum latency in milliseconds
## read        - read percentage
## random      - random percentage
## ios         - total I/O operations
## seconds     - runlength in seconds
## threads     - threads
## blksize     - block size
## csw_io 
## int_io 
## sys 
## usr 
## wt 
## idl 
## uio 
## ukb 
## uuio 
## tuio 
## 
## Most sio input options are supported by this function.
## The following are NOT supported:
##   -noheader 
##   -help:        Display the help message.
##   -version:     Display detailed version log. 
##                 (But the current version is returned on all calls)
##   -debug:       Output detailed debug info.  Be prepared for a lot of info.
##   -niceoutput:  In addition to the standard single line output,
##                 print single column, more human-readable format.
##
## This function terminates the current subtest with a SCRIPT error
## if it detects an undefined input keyword,
## or an invalid value for an input parameter.
##
## This function terminates the current subtest with a CONFIG error
## if it detects the client is not configured for command line input.
##
## This function announces non-FATAL error meessages when
## data does not verify,
## devices cannot be opened (and 'openfailok' was not specified),
## devices do not exist,
## or an unknown output message is seen from sio.
##
## @arg client_conn 
##      Required: the client connection object
## @arg device
##      Required: The device name as recognized by the client.
##
## @arg block_size
##      Optional: the block size of I/O operations.
##      Default is 1024 which is compatible with all platforms.
##      Maximum is 262144
## @arg conn_type
##      Optional: #rsh, #telnet or #console
##      It is highly recommended to let take the default.
## @arg last_byte
##      Optional: the last (0-origin) byte offset into
##      the device for I/O operations. Default is 1g.
##      Must be 512 to 262144
## @arg runlength
##      Optional: the number of seconds to run I/O.
##      Default is 180. 
##      This script does not allow values less than 10 seconds.
## @example
## 
## To run a default I/O mix for 120 seconds:
## 
##   _run_sio($cc,$CLIENT_IO_DEV, 
################################################################################
sub _run_sio() 
{
   my ($self, $client_conn, $client_io_dev, $read_pct, $random_pct, 
       $block_size, $first_byte, $last_byte, $run_time, $num_threads, $other_sio_params) = @_;

   my %results;
   $results{'error_count'} = 0;
   
   # Format the list of devices.
   $client_io_dev = _format_sio_device("$client_io_dev");
   
   # Create the system independent program invocation.
   my $cmd  = "sio $read_pct $random_pct"
                  . " $block_size $first_byte $last_byte" 
                  . " $run_time" 
                  . " $num_threads" 
                  . " $client_io_dev"
                  . " $other_sio_params";

   my $byte_count;
   $byte_count = $last_byte;
   $byte_count =~ s/(\d+).*/$1/;
   if ($last_byte =~ m/k/i) {
      $byte_count = $byte_count * 1024;
   } elsif ($last_byte =~ m/m/i) {
      $byte_count = $byte_count * 1024 * 1024;
   } elsif ($last_byte =~ m/g/i) {
      $byte_count = $byte_count * 1024 * 1024 * 1024;
   } elsif ($last_byte =~ m/t/i) {
      $byte_count = $byte_count * 1024 * 1024 * 1024 * 1024;
   }
   
   # Send the command to the client and get sio output.
   #
   # Allow some additional time for sio to return
   # after it completes the specified $runlength
   # yet, timeout in case there is a hang condition
   my $timeout = $run_time + 1800;
   my $one_meg = 1024 * 1024;
   # because fillonce, use last_byte and blksize to determine timeout.
   if ($run_time == 0) {
      $timeout = (($byte_count / $one_meg) + 1800);
   }
   
   $self->sanlog( -msg => "Sio::_run_sio:\n"            .
                          "Command     = $cmd\n"        .
                          "byte_count  = $byte_count\n" .
                          "timeout     = $timeout\n",
       -msg_level => 3);
   
   my $client = $client_conn->{hostp}->{hostname};
   
   my ($status, @output) = &ClientUtil::execute($client_conn, $cmd, $timeout);


   if ($status gt 0) {
      $results{'error_count'} = 1;
      $results{'error_msg'} = "Error executing command on device \'$client\' \nCommand = " . $cmd . "\nOutput = " . join "\n", @output;
      logresult("FAIL", %results);
      return $ClientUtil::FALSE, %results;      
   } 
   
   my @unexpected_output;
   my $error_message = "Unrecognized sio output on device \'$client\'";
   my $last_label = '';

   foreach my $line (@output) {
      #
      # When sio finishes correctly, it reports stats.
      #
      # Version: 5.54.1
      # Read: 100 Rand: 0 BlkSz: 512 BegnBlk: 0 EndBlk: 10240 Secs: 5 Threads: 2 Devs: 1  ./tempfile
      # IOPS,   TPUT(KB/s),     LAT(ms),        READ,   RAND,   IOS,    SEC,    THDS    BLKSZ   CSW_IO, INT_IO, SYS,    USR,    WT,     IDL,    uIO,    uKB,    uuIO,   tuIO
      # 4784,   2392,           0.418[0.372],   100,    0,      24159,  5,      2,      512,    2.2,    1.7,    32,     68,     0,      0,      134,    267.6,  284,    418
      # Maximum latency observed:214.655000
      #
      # Terminating threads ...lash<133>
      
      #Version: 5.61
      #Read: 50 Rand: 50 BlkSz: 1024 BegnBlk: 0 EndBlk: 100 Secs: 120 Threads: 1 Devs: 1  \\\\.\\PHYSICALDRIVE2
      #IOPS,TPUT(KB/s),LAT(ms)Calc'd,LAT(ms)Actual,READ,RAND,IOS,SEC,THDS,BLKSZ,CSW_IO,INT_IO,SYS,USR,WT,IDL,uIO,uKB,uuIO,tuIO
      #504,504,1.984,[1.957],50,50,60496,120,1,1024,2.4,1.1,79,1,0,20,1567,1567.0,20,1587
      
      #504,    504,            2.0[2.0],       50,     50,     60485,  120,    1,      1024,   2.4,    1.1,    78,     1,      1,      20,     1547,   1547.5
      
      #openRead: 50 Rand: 50 BlkSz: 1024 BegnBlk: 0 EndBlk: 100 Secs: 90 Threads: 1 Devs: 1  \\\\.\\PHYSICALDRIVE1
      #: No error
      
      #Version: 6.15\r\r\n
      #Read: 100 Rand: 0 BlkSz: 1024 BegnBlk: 0 EndBlk: 4 Secs: 1073741823 Threads: 1 Devs: 1  \\\\.\\PhysicalDrive1\r\r\n
      #Execution less than expected (runtime 1073741823, actual 1)\r\r\n
      #IOPS,TPUT(KB/s),LAT(ms)Calc'd,LAT(ms)Actual,READ,RAND,IOS,SEC,THDS,BLKSZ\r\r\n
      #4790,4790,0.209,[0.264],100,0,4790,1,1,1024\r\r\n
      #Maximum latency observed:1000.000000\r\r\n
      #\r\r\n
      #Terminating threads ...
      # Normal output lines with no data to return
      # But it may be intermixed with STDERR text
      #$line =~ s/Read:\s+\d+\s*//;
      #$line =~ s/Rand:\s+\d+\s*//;
      #$line =~ s/BlkSz:\s+\d+\s*//;
      #$line =~ s/BegnBlk:\s+\d+\s*//;
      #$line =~ s/EndBlk:\s+\d+\s*//;
      #$line =~ s/Secs:\s+\d+\s*//;
      #$line =~ s/Threads:\s+\d+\s*//;
      #$line =~ s/Devs:\s+\d+\s+\S+//;
      # the above lines are trying to remove:
      # Read: 0 Rand: 0 BlkSz: 65536 BegnBlk: 0 EndBlk: 1600 Secs: 1073741823 Threads: 10 Devs: 1  /dev/rdsk/c3t1d1s6
      # Let's ignore the line
      if ($line =~ /^Read:\s+\d+\s+Rand:\s+/) {
         # this is the display of params info
      } elsif  ($line =~ /^(\d+),\s*(\d+),\s*([\.\d]+),?\[([\.\d]+)\],\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*([\.\d]+),\s*([\.\d]+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*([\.\d]+),\s*(\d+),\s*(\d+)/) {
         $results{'iops'} = $1;    # I/O operations per second
         $results{'tput'} = $2;    # KB per second
         $results{'latency_calculated'} = $3;     # latency in milliseconds
         $results{'latency_actual'} = $4;     # latency in milliseconds
         $results{'read'} = $5;    # read percentage
         $results{'random'} = $6;  # random percentage
         $results{'ios'} = $7;     # total I/O operations
         $results{'seconds'} = $8; # runlength in seconds
         $results{'threads'} = $9; # threads
         $results{'blksize'} = $10;   # block size
         $results{'csw_io'} = $11;   # 
         $results{'int_io'} = $12;  # 
         $results{'sys'} = $13;  # 
         $results{'usr'} = $14;  # 
         $results{'wt'} = $15;  # 
         $results{'idl'} = $16;  # 
         $results{'uio'} = $17;  # 
         $results{'ukb'} = $18;  # 
         $results{'uuio'} = $19;  # 
         $results{'tuio'} = $20;  # 
      } elsif ($line =~ /^(\d+),\s*(\d+),\s*([\.\d]+),?\[([\.\d]+)\],\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*([\.\d]+),\s*([\.\d]+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*([\.\d]+)/) {
         # Older versions only sent back 18 values
         $results{'iops'} = $1;    # I/O operations per second
         $results{'tput'} = $2;    # KB per second
         $results{'latency_calculated'} = $3;     # latency in milliseconds
         $results{'latency_actual'} = $4;     # latency in milliseconds
         $results{'read'} = $5;    # read percentage
         $results{'random'} = $6;  # random percentage
         $results{'ios'} = $7;     # total I/O operations
         $results{'seconds'} = $8; # runlength in seconds
         $results{'threads'} = $9; # threads
         $results{'blksize'} = $10;   # block size
         $results{'csw_io'} = $11;   # 
         $results{'int_io'} = $12;  # 
         $results{'sys'} = $13;  # 
         $results{'usr'} = $14;  # 
         $results{'wt'} = $15;  # 
         $results{'idl'} = $16;  # 
         $results{'uio'} = $17;  # 
         $results{'ukb'} = $18;  # 
      } elsif ($line =~ /^(\d+),\s*(\d+),\s*([\.\d]+),?\[([\.\d]+)\],\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+)/) {
         # Older versions only sent back 18 values
         $results{'iops'} = $1;    # I/O operations per second
         $results{'tput'} = $2;    # KB per second
         $results{'latency_calculated'} = $3;     # latency in milliseconds
         $results{'latency_actual'} = $4;     # latency in milliseconds
         $results{'read'} = $5;    # read percentage
         $results{'random'} = $6;  # random percentage
         $results{'ios'} = $7;     # total I/O operations
         $results{'seconds'} = $8; # runlength in seconds
         $results{'threads'} = $9; # threads
         $results{'blksize'} = $10;   # block size
      } elsif ($line =~ /^0,0,/) {
         # No I/O was performed!
         $results{'iops'} = 0;    # I/O operations per second
         $results{'tput'} = 0;    # KB per second
         logresult('WARN',"No I/O performed by sio");
      } elsif ($line =~ /^\s*$|^Terminating threads .../) {
         # Normal output lines with no data to return
      } elsif ($line =~ /^sio\s/) {
         # echo of sio input command
      } elsif ($line =~ /^IOPS,\s+TPUT|^IOPS,\s*TPUT\(KB\/s\)/) {
         # Normal output lines with no data to return
      } elsif ($line =~ /I\/O completed successfully|PASS/) {
         # If random options are used, this string is printed in
         # lieu of stats.
      } elsif ($line =~ /Version:\s*(\S+)/) {
         $results{'version'} = $1;   
      } elsif ($line =~ /Maximum latency observed:\s*(\d+)/) {
         $results{'max_latency'} = $1;  
      } elsif ($line =~ m/Execution less than expected/ &&
             $other_sio_params =~ m/-numops/ ) {
             # because runlength=0, sec is default to 1073741823
             # use -numops to read
             # we should ignore
             next;
      # ignore extra output messages from verion 6.22
      } elsif ($line =~ /^(global_|Computed)/) {
         next;     
      # Configuration Errors
      } elsif ($line =~ /host or user not listed in rhosts/) {
         $results{'error_count'}++;
         $results{'error_msg'} = "CLIENT $client needs to enable rsh access";
         logresult("FAIL", %results);
         return $ClientUtil::FALSE, %results;      
      #  ExitConfigError("CLIENT $client needs to enable rsh access");
      } elsif ($line =~ / sio:  not found/) {
         $results{'error_count'}++;
         $results{'error_msg'} = "CLIENT $client needs access to the sio program";
         logresult("FAIL", %results);
         return $ClientUtil::FALSE, %results;      
      #  ExitConfigError("CLIENT $client needs access to the sio program");
      # Script Errors
      } elsif ($line =~ /Unknown size type /) {
         $results{'error_count'} = 1;
         $results{'error_msg'} = "Invalid block size parameter";
         logresult("FAIL", %results);
         return $ClientUtil::FALSE, %results;      
      #  ExitScriptError("Invalid block size parameter");
      # Other Errors    
      } elsif ($line =~ /File open error|CreateFile fails with The system cannot find the file specified/) {
         $error_message = "SIO CLIENT $client must be configured with: $client_io_dev";
         $results{'error_count'}++;
      } elsif ($line =~ m/sio:\s+(.*)/) {
         $error_message = "SIO reported an error on CLIENT $client: $1";
         $results{'error_count'}++;
      #Version: 6.06\r\n
      #Read: 50 Rand: 50 BlkSz: 1024 BegnBlk: 0 EndBlk: 100 Secs: 240 Threads: 1 Devs: 1  /dev/rnate_lvlog
      #open: No such file or directory
      #File: /dev/rnate_lvlog
      #
      #sio: File open error
      } elsif ($line =~ m/(open: No such file or directory)/) {
         $error_message = "SIO reported an error on CLIENT $client: $line";  
         $results{'error_count'}++;
      } elsif ($line =~ m/File: (\S+)/) {
         if (0 == $results{'error_count'}) {
             # Previous message should have been seen first,
             # don't count this unless it is first
             $error_message = "SIO reported an error on CLIENT $client with file: $1";
             $results{'error_count'}++;
         }
      # write fails 
      } elsif ($line =~ m/error accessing file|WriteFile fails|Filename:/) {
         $error_message = "SIO reported write error on CLIENT $client: $line";  
         $results{'error_count'}++;
      } elsif ($line =~ m/error reading:/) {
         $error_message = "SIO reported write error on CLIENT $client: $line";  
         $results{'error_count'}++;
      } elsif ($line =~ m/connect to address|Connection refused|Trying\s+\S+\s+rsh/) {  
         # connect to address 10.72.237.39 port 544: Connection refused\r\n
         # Trying krb4 rsh...\r\n
         # connect to address 10.72.237.39 port 544: Connection refused\r\n
         # trying normal rsh (/usr/bin/rsh)\r\n
         # just ignore these lines....
         next;
      }
      else {
         push(@unexpected_output,$line);
         $results{'error_count'}++;
      }   #
    } # end of main foreach loop

    if (@unexpected_output or $results{'error_count'} > 0 ) {
       $results{'error_msg'} = "$error_message" . join "\n", @unexpected_output;
       logresult("FAIL", %results);
       return $ClientUtil::FALSE, %results;      
    }
    logresult("PASS", %results);
    return ($ClientUtil::TRUE, %results);

} # end of _run_sio

# ######################################################################
## @name ::_format_sio_device
## @summary Return a formatted device name from something like "e:"
##
## @arg in_device Required: Device string to format.
##
# ######################################################################
sub _format_sio_device($) 
{
   my $in_device = shift;
   my $out_device = $in_device;
   
   # Match any number of preceeding characters that are "\" or "."
   # followed by any letter and ":" such as "e:"
   # followed by anything.
   if ($out_device =~ m/[\S]*[a-zA-Z]\:.*/i)
   {
      # Remove anything that preceeds any letter followed by a colon.
      $out_device =~ s/\S+([a-zA-Z]\:\S*)/$1/g;
      
      # Replace "e:" with "\\\\.\\e:".
      $out_device =~ s/([a-zA-Z]\:)/\\\\\.\\$1/g;
   }
   
   return $out_device;
} # end _format_sio_device

# ######################################################################
## @name ::_subtest_exit_handler
## @summary 
##
## @arg subtest_obj Required: subtest object
##
# ######################################################################
sub _subtest_exit_handler($) {

    my $subtest_obj = shift;
    my @results = $subtest_obj->rslts();

    foreach my $result (@results) {
        if ($result->severity() >= 0) {
            if ($result->type() =~ /FAIL|SCRIPT|CONFIG/i) {
                logresult ('FAIL',"subtest: $subtest_obj failed\n");
                return $ClientUtil::FALSE;
            }

        }
    }
    return $ClientUtil::TRUE;
}

1;


#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Sio.pm

=head1 Author

=begin html 

David Pressley
Muthukumaran Govindarajan

=end html

=head1 Description

Sio provides an API interface to a run IO using Sio tools on the client 

=head1 Attributes

=head2 B<client_conn>

reference = client connection refernce

=head2 I<wait_flag>

scalar = Determine the start_io function to return immediately or wait to finsh io

=head2 B<client_io_devs>

reference = holds the client io device list

=head2 I<read_pct>

scalar = read percentage for the io operation on the given device list. range [0-100]

=head2 I<random_pct>

scalar = random percentage for the io operation on the given device list. range [0-100]

=head2 I<block_size>

scalar = Io block size

=head2 B<first_byte>

scalar = Starting location of the IO operation in the given io device

=head2 B<last_byte>

scalar = Ending location of the IO operation in the given io device

=head2 I<run_time>

scalar = Specifiy the the amount of time to run the io

=head2 I<num_threads>

scalar = Numbers of threads to use by sio

=head2 I<sio_subtest_id>

scalar = name of the run id to use for sio subtest, default "run_sio"

=head2 I<other_sio_params>

scalar = Execute with other valid sio parameters 

=head2 B<sub_test_obj>

reference = Reference to the subtest object 

=head2 B<sub_test_pid>

scalar = PID value of the subtest

=head1 Public Methods

=head2 B<new(%args)>

Create the Sio objects 

=over 4

=item Inputs

=item * parent 

parent class 

=item * for other parameters refer the attributes section

=item Outputs

None

=item Examples

      my $sio_obj = new Sio(-client_conn   => $client_conn,
                           -wait_flag     => $wait_flag,
                           -client_io_dev => $client_io_dev, 
                           -read_pct      => $read_pct,
                           -random_pct    => $random_pct,
                           -block_size    => $block_size,
                           -first_byte    => $first_byte,
                           -last_byte     => $last_byte,
                           -run_time      => $run_time,
                           -num_threads   => $num_threads,
                           -pass_thru     => $pass_thru);

      if (sio_obj->error()) {
         print $sio_obj->errmsg();
         return 0;
      }
   
=back

=head2 B<start_io()>

To start the IO operation on the given list of devices

=over 4

=item Inputs

None

=item Outputs

Returns an scalar value 1 if io started or completed successfully 
If an error occurred the error flag is set, errmsg is set to the failure reason, 
and 0 is passed back as the return value.


=item Examples

      my $sio_obj = new Sio(-client_conn   => $client_conn,
                           -wait_flag     => $wait_flag,
                           -client_io_dev => $client_io_dev, 
                           -read_pct      => $read_pct,
                           -random_pct    => $random_pct,
                           -block_size    => $block_size,
                           -first_byte    => $first_byte,
                           -last_byte     => $last_byte,
                           -run_time      => $run_time,
                           -num_threads   => $num_threads,
                           -pass_thru     => $pass_thru);

      if (sio_obj->error()) {
         print $sio_obj->errmsg();
         return 0;
      }

      my ($status, %results) = $sio_obj->start_io();
      if( $status == $ClientUtil::FALSE) {
         my $error_msg = "Error: SanIo::san_io_run(): ". $sio_obj->errmsg();
         $self->_set_err($error_msg);
         return $ClientUtil::FALSE, %results;      
      }

=back

=head2 B<stop_io()>

To stop the IO operation if the IO is running on the background

=over 4 

=item Inputs

None

=item Outputs

Returns an scalar value 1 if io stopped successfully 
If an error occurred the error flag is set, errmsg is set to the failure reason, 
and 0 is passed back as the return value.

=item Examples

   $sio_obj->stop_io();

=back

=head2 B<verify_data()>

Verify that filecontent is in tact and not corrupted after a write operation.

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if the file content matches generated pattern, otherwise 0.  If an error occured 
the error flag is set, errmsg is set to the failure reason.

=item Examples

   $sio_obj->verify_data();

=back

=head2 B<is_alive()>

Determine if i/o operation is still running in the background.

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if i/o operation is still running, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Examples

    # Check if DataTest I/O has completed on the volume 
    # every 5 seconds.
    sleep 5 while($sio_obj->is_alive());

=back

=head2 B<wait_io()>

Wait for IO to complete, if IO still running in the background. After IO completes or not running it iwll return 
the result in hash.

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value 1 and the results hash on success.
If an error occurred the error flag is set, errmsg is set to the failure reason, 
and 0 is passed back as the return value.


=item Examples

   my ($status, %results) = $sio_obj->wait_io();
   if( $status == $ClientUtil::FALSE) {
      my $errmsg = join "\n", $self->{io_tool_obj}->errmsg();
      $self->_set_err("Error: SanIO::san_io_wait(): $errmsg");
      return $ClientUtil::FALSE, %results;      
   }

=back

=head2 B<get_io_result()>

To get the result for the IO completes. If this functional called while IO still running, it will set an error.
Maksue sure call this function after IO completes.

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value 1 and the results hash on success, if io completed successfully without error.
If an error occurred the error flag is set, errmsg is set to the failure reason, 
and 0 is passed back as the return value.


=item Examples

   my ($status, %results) = $sio_obj->get_io_result();
   if( $status == $ClientUtil::FALSE) {
      my $errmsg = join "\n", $sio_obj->errmsg();
      return $ClientUtil::FALSE;      
   }

=back



=cut

