# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Sanlun
## @author latif@netapp.com
## @status private
## @pod here

package Sanlun;

use strict;

use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C    M E T H O D S
################################################################################
###############################################################################
# Method:    new
# Objective: Sanlun Object constructor.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;
    my ($self, $hosttype, $obj_template);

    $self = {};

    ## bless the object upfront in case an error occurs.
    bless ($self, $class);

    # check for existance of required -parent argument

    if (defined($args{-parent})) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    if (defined($args{-obj_template})) {
        $obj_template = $args{-obj_template};
    }

    $hosttype = $args{-parent}->{conn}->{hostp}{hosttype};

    # determine if this is a virtual machine
    my $hypervisor_type;
    if ($args{-parent}->{is_virtual} == 1) {
        $hypervisor_type = $args{-parent}->{hypervisor_type};
    }

    ## load the appropriate Sanlun for the hosttype
    if ($hosttype =~ /unix-solaris/i) {
        if ($hypervisor_type =~ /VMware|HyperV|Microsoft|Citrix|KVM|OVM/i) {
            require San::SanClient::Sanlun::VSolarisSanlun;
            $self = San::SanClient::Sanlun::VSolarisSanlun->_new(%args);
        } else {
            require San::SanClient::Sanlun::SolarisSanlun;
            $self = San::SanClient::Sanlun::SolarisSanlun->_new(%args);
        }
    } elsif ($hosttype =~ /unix-hpux/i) {
        require San::SanClient::Sanlun::HPUXSanlun;
        $self = San::SanClient::Sanlun::HPUXSanlun->_new(%args);
    } elsif ($hosttype =~ /unix-aix/i) {
        require San::SanClient::Sanlun::AixSanlun;
        $self = San::SanClient::Sanlun::AixSanlun->_new(%args);
    } elsif ($hosttype =~ /unix-linux-sanblaze/i) {
        require San::SanClient::Sanlun::SanBlazeSanlun;
        $self = San::SanClient::Sanlun::SanBlazeSanlun->_new(%args);
    } elsif ($hosttype =~ /unix-linux/i) {
        if ($hypervisor_type =~ /VMware|HyperV|Microsoft|Citrix|KVM|OVM/i) {
            require San::SanClient::Sanlun::VLinuxSanlun;
            $self = San::SanClient::Sanlun::VLinuxSanlun->_new(%args);
        } else {
            require San::SanClient::Sanlun::LinuxSanlun;
            $self = San::SanClient::Sanlun::LinuxSanlun->_new(%args);
        }
    } elsif ($hosttype =~ /windows/i) {
        if ($hypervisor_type =~ /VMware|HyperV|Microsoft|Citrix|KVM|OVM/i) {
            require San::SanClient::Sanlun::VWinSanlun;
            $self = San::SanClient::Sanlun::VWinSanlun->_new(%args);
        } else {
            require San::SanClient::Sanlun::WinSanlun;
            $self = San::SanClient::Sanlun::WinSanlun->_new(%args);
        }
    } elsif ($hosttype =~ /linux-vmware/i) {
        if ($args{-parent}->has_vma_host()) {
            $hypervisor_type = "VMware";
            require San::SanClient::Sanlun::ESXiSanlun;
            $self = San::SanClient::Sanlun::ESXiSanlun->_new(%args);
        } else {
            require San::SanClient::Sanlun::ESXSanlun;
            $self = San::SanClient::Sanlun::ESXSanlun->_new(%args);
        }
    } elsif ($hosttype =~ /sanblaze/i) {
        require San::SanClient::Sanlun::SanBlazeSanlun;
        $self = San::SanClient::Sanlun::SanBlazeSanlun->_new(%args);
    } else {
        $self->_set_err("Platform '$hosttype' not currently supported by Sanlun object");
        return $self;
    }

    # caller specified a template to copy
    if (defined($obj_template)) {
        # copy over all components of the object template
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        # fix the parent link
        $self->{parent} = $args{-parent};

        # if a winclient object was specified, update the link for it as well
        if (exists($args{WinClient})) {
            $self->{WinClient} = $args{WinClient};
        }

        # return now, dont do anything extra
        return $self;
    }

    # only load the version if we are not windows/vmware/sanblaze and we are not a virtual machine
    if ($hosttype !~ /windows|sanblaze|vmware|citrix/i and $hypervisor_type !~ /VMware|HyperV|Microsoft|Citrix|KVM|OVM/i and
            not $args{-parent}->has_vma_host()) {
        # get sanlun version and build number
        ($self->{version}, $self->{build_number}) = $self->_version();
    } else {
        ($self->{version}, $self->{build_number}) = (undef, undef);
    }

    #
    # By returning $self here we are returning a parent class such
    # as SolarisSanlun, not a Sanlun object.  This is OK because
    # the OS specific Sanlun::XXXSanlun classes add Sanlun in @ISA
    # so Sanlun is the base class for the parent
    #
    return ($self);

}


###############################################################################
# Method:    get_sanlun_version
# Objective: Returns the Sanlun version to the user.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanlun_version {
    my ($self) = @_;

    return $self->{version};
}

###############################################################################
# Method:    get_sanlun_build_number
# Objective: Returns the Sanlun version to the user.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanlun_build_number {
    my ($self) = @_;

    return $self->{build_number};
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

###############################################################################
# Method:    _version
# Objective: Queries the Sanlun version and build number on this client.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _version {
    my ($self) = @_;
    my ($version, $build_number) = (undef, undef);

    # execute a sanlun version command
    my $path = $self->{parent}->get_hu_bin_path();
    my $cmd = $path . "/sanlun version";
    my ($status, @output) = &ClientUtil::execute($self->{conn},$cmd);

    if (grep(/operable program or batch file/, @output) != 0) {
        $self->sanlog(-msg => "Sanlun is not installed on this system",
                      -msg_level => 3);
        return (undef, undef);
    } elsif ($status > 0) {
        $self->_set_err("Error: Sanlun::_version()",
                        "Command ($cmd) returned a bad status ($status)");
        return (undef, undef);
    }

    # parse the version output (versionnum -> $1, buildnumber -> $2)
    foreach my $line (@output) {
        if ($line =~ /version\s+(\S+)\s+\((\S+)\)/i) {
            $version = $1;
            $build_number = $2;
        } elsif ($line =~ /(\d*\.\d*)\.(\d+\.\d+.*)/i) {
            $version = $1;
            $build_number = $2;
        } elsif ($line =~ /(\d*\.\S+?)\.(\d+\.\d+.*)/i) {
            $version = $1;
            $build_number = $2;
        }
    }

    # if either version or buildnumber are undefined, set error.
    if ($version == undef or $build_number == undef) {
        $self->_set_err("Error: Sanlun::_version()",
                        "No Sanlun version or build number.");
        return (undef, undef);
    }

    return ($version, $build_number);
}

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Sanlun.pm

=head1 Authors

=begin html

Mike Latif

=end html

=head1 Description

Sanlun provides an API interface to the Sanlun utility.  This package should
be instantiated through the SanClient package create_sanlun() method.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>

=head1 Attributes

=head2 B<version>

Sanlun Version (float e.g. "3.0")

=head2 B<build_number>

Sanlun Build Number (int)

=head1 Public Methods

=head2 B<display_adapters(%args)>

Display adapter information for a particular protocol.

=over 4

=item Inputs

=item * -protocol (default: fcp-(solaris, hpux, aix, windows)  iscsi-(hpux, aix, linux, windows))

Device protocol

=item Outputs

Returns an array or array reference if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving adapter  information");

    $Test->log($sanlunObj->display_adapters(-protocol=>'iscsi'));
    if ($sanlunObj->error){
        $Test->log($sanlunObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<get_lun_path(%args)>

Retrieves the filer lun path for a host device.

=over 4

=item Inputs

=item * -device

Device on the host

=item Outputs

Returns a scalar value containing the full lun path for a host device.
If an error occurred, the error flag is set, errmsg is set to the failure reason, and
an empty string "" is passed back as the return value.

=item Examples

 #Get lun path on solaris host
 $path=$Sanlun->get_lun_path(-device=>'c3t1d5');
 if ($Sanlun->error){
    $Test->nlog($Sanlun->errmsg());
    return 0;
 }

 print "Device 'c3t1d5' maps to lun path '$path'";


=back

=head2 B<lun_show()>

Retrieve a list of all lun information

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of lun information if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and an empty array is returned.


=item Examples

    $Test->log("-- Retrieving lun information");
    @output=$Sanlun->lun_show();
    if ($Sanlun->error){
        $Test->log($Sanlun->errmsg());
        return $TCase::FAIL;
    }

    $Test->nlog(@output);


=back

=cut

1;

