# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to sanlun utility
## @author knopp@netapp.com
## @status Private
## @pod ../Sanlun.pm

# Package Name
##############
package San::SanClient::Sanlun::AixSanlun;

use strict;

# Package Requirements
######################
use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error Sanlun SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return the path to a lun given a device
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self, %args) = @_;

    my ($cmd, $device, $status, @output);
    my ($filer_name, $lun_name, $lun_path);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device= $args{$_})) {
                $self->_set_err("Error: AixSanlun::get_lun_path() Invalid value for argument '-device'");
                return ""; 
            }        
            next;
        } else {
            $self->_set_err("Error: AixSanlun::get_lun_path() Invalid parameter specified \'$_\'");
            return "";                                 
        }
    }
    
    # Make sure device value format is valid
    unless (defined $device and $device =~ /hdisk/) {
        $self->_set_err("Error: AixSanlun::get_lun_path() Invalid parameter specified \'$_\'");
        return "";   
    }

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show -d $device | grep $device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixSanlun::get_lun_path() ",
                        "Command ($cmd) returned a bad status ($status)");
        return "";
    }

    if (defined $output[0]) {
        # Remove leading white spaces if there
        $output[0] =~ s/^\s*//;

        ($filer_name, $lun_name) = split(/\s+/, $output[0]);

        # Append colon in the end of filer name if it is not present
        if ($filer_name !~ /:$/) {
            $filer_name .= ":";
        }

        $lun_path = $filer_name . $lun_name;
    }

    return $lun_path;
}

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;

    my ($cmd, $device, $status, @output);
    my ($model, $version, $description, $initiator);
    my ($location_code, $dev, $sw_init);
    my (@init_output);

    $self->_clear_err();

    # parse arguments
    unless (defined $protocol and $protocol =~ /fcp|iscsi/i){
        $self->_set_err("Error: AixSanlun::display_adapters() Missing or invalid protocol type ($protocol)");
        return (wantarray()) ? () : []; 
    }
    
    if ($protocol =~ /fcp/i) {
        my $path = $self->{parent}->get_hu_bin_path();
        $cmd = $path . "/sanlun fcp show adapter -v";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: AixSanlun::display_adapters()",
                            "Command ($cmd) returned a bad status ($status)");

            return (wantarray()) ? () : []; 
        }
    } elsif ($protocol =~ /iscsi/i) {
        # The command 'Sanlun iscsi show adapter' does not work.  Will need
        # to retrieve the information elsewhere and return it in the Sanlun 
        # output format 

        if ($self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /software/) {
            # Lets get software information. The expected output looks like 
            # devices.iscsi_sw.rte 5.2.0.50 C  F  iSCSI Software Device Driver
            $cmd = "lslpp -L devices.iscsi_sw.rte | grep devices ";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: AixSanlun::display_adapters() ",
                                "Command ($cmd) returned a bad status ($status)");
                return (wantarray()) ? () : []; 
            }
            ($model, $version, $description) = 
                         $output[0] =~ /\s+(\S+)\s+(\S+)\s+\S+\s+\S+\s+(.*)/;
        
            # Lets get the initiator name from the iscsi protocol device
            # for the SW initiator. 
            $cmd = "lsdev | grep \'iSCSI Protocol Device\'";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status gt 0) {
                $self->_set_err("Error: AixSanlun::display_adapters() No SW Initiator Protocol Device found in the system");
                return ($ClientUtil::FALSE);
            } else {
                foreach $dev (@output) {
                    my $location_code = (split(/\s+/, $dev))[2];
                    if ($location_code =~ /iSCSI/) {
                        $sw_init = (split(/\s+/, $dev))[0];
                    }
                }
            }

            $cmd = "lsattr -El $sw_init | grep initiator_name | awk '{print \$2}'";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: AixSanlun::get_adapters() ",
                                "Command ($cmd) returned a bad status ($status)");
                return (wantarray()) ? () : []; 
            }
            $initiator = $output[0];

            # Let construct the format to mimic how Sanlun returns the data
            @output = ();
            push @output, "aix-iscsi-sw initiator name: " . $initiator,
                          "model:   " . $model,
                          "model description: " . $description,
                          "driver version: " . $version,
                          "WWPN: " . $initiator;

        } elsif ($self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /hardware/) {
            my (@lsattr_output);
            # get model, version, and description from lslpp for HW TOE
            $cmd = "lslpp -L devices.pci.1410d002.rte | grep devices";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: AixSanlun::display_adapters() ",
                                "Command ($cmd) returned a bad status ($status)");
                return (wantarray()) ? () : [];
            }
            ($model, $version, $description) =
                         $output[0] =~ /\s+(\S+)\s+(\S+)\s+\S+\s+\S+\s+(.*)/;

            # get the list of iSCSI adapters in the system
            $cmd =  "lsdev -Cc adapter | grep \'iSCSI Adapter\' | awk '{print \$1}'";
            ($status, @init_output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: AixSanlun::display_adapters() Could not determine iSCSI HW adapter list (@output)");
                return (wantarray()) ? () : []; 
            }
 
            # get the IQN for each iSCSI adapter and generate the iSCSI
            # adapter information arrary.
            @output = ();
            my $init_count = 0;
            foreach my $init (@init_output) {
                $cmd = "lsattr -El $init | grep initiator_name | awk '{print \$2}'";
                ($status, @lsattr_output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Error: AixSanlun::get_adapters() ",
                                    "Command ($cmd) returned a bad status ($status)");
                    return (wantarray()) ? () : []; 
                } else {
                    $initiator = $lsattr_output[0];
                }

                # Let construct the format to mimic how Sanlun returns the data
                push(@output, "aix-hw-toe" . $init_count . " initiator name: " . $initiator,
                              "model:   " . $model,
                              "model description: " . $description,
                              "driver version: " . $version,
                              "WWPN: " . $initiator);
                $init_count++;
            }
        } else {
             $self->_set_err("Error: AixSanlun::display_adapters() iSCSI initiator type ($self->{iscsi}->{CONFIG_ISCSI_TYPE}) is not valid");
             return (wantarray()) ? () : []; 
        }
    }

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_protocol
# Objective:       return protocol  
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;

    my ($protocol);
    
    $self->_clear_err();

    if (defined $self->{protocol}) {
        $protocol = $self->{protocol};
    } else {
        $self->_set_err("Error: AixSanlun::get_protocol() Sanlun object has no protocol specified");
        $protocol = "";     
    }

    return $protocol;
    
} 

################################################################################
# Subroutine Name: get_filer()
# Objective:       No-op
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filer {
    my ($self) = @_;
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns. 
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self, %args) = @_;

    my ($bPrimary, $cmd, $device, $protocol, $status, @output);

    $self->_clear_err();

    # flag to retrieve only primary
    # device paths
    $bPrimary = 0;

    # parse arguments
    foreach (keys %args) {
      if (/-?_primary_paths_only/) {
          if (not length($bPrimary = $args{$_})) {
            $self->_set_err("Error: AixSanlun::lun_show() Invalid value for argument '-_primary_paths_only'");
            return ($ClientUtil::FALSE); 
          }        
          next;
      } else  {
        $self->_set_err("Error: AixSanlun::lun_show() Invalid parameter specified \'$_\'");
        return ""; 
      }
    }

    # Currently AIX Sanlun doesn't support the hidden '-s' option to allow
    # only the primary paths to for lun to display.  The best we can 
    # do for now is to only apply the '-_primary_paths_only' option when
    # the iscsi protocol is specified. Otherwise return all sanlun output 
    # for the selected protocol.  
    $protocol = $self->get_protocol();

    my $path = $self->{parent}->get_hu_bin_path();
    if ($bPrimary == 1 && $protocol =~ /iscsi|fcp/i) {
       $cmd = $path . "/sanlun lun show all | grep -i $protocol | awk '{print \$3}'";
    } elsif ($protocol =~ /iscsi|fcp/i) {
       # Only the iscsi version of sanlun returns the protocol column.   
       $cmd = $path . "/sanlun lun show all | grep -i $protocol";
    } elsif ($protocol =~ /fcp/i){
       # The fcp version of sanlun only return fcp connected luns
       $cmd = $path . "/sanlun lun show all ";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixSanlun::lun_show()",
                        "Command ($cmd) returned a bad status ($status)");

        return (wantarray()) ? () : []; 
    }
    
    return (wantarray()) ? @output : \@output; 

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($parent, $self);

    $self = {};

    bless ($self, $class);        

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                return ($ClientUtil::FALSE);
            }
            next;
        } 
    }     

    # initilize the object members
    $self->{parent} = $parent;
    $self->{conn} = $parent->get_connection();
    $self->{protocol} = $parent->get_protocol();

    if ($self->{protocol} =~ /iscsi/i) {
        # Default to SW initiator if not specified
        if (not defined($self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE})) {
            $self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE} = "software";
        }

        $self->sanlog(-msg => "iSCSI Initiator type is $self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE}");
    }

    return $self;
}

1;
