# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to sanlun utility
## @author hrivera@netapp.com
## @status Private
## @pod ../Sanlun.pm

# Package Name
##############
package San::SanClient::Sanlun::ESXSanlun;

use strict;

# Package Requirements
######################
use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use vars qw( @ISA );

@ISA = qw(Error SanLog Sanlun);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self, %args) = @_;

    my ($cmd, $device, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Error: ESXSanlun::get_lun_path() " .
                                "Invalid value for argument '-device'");
                return "";
            }
            next;
        }
    }

    $cmd = "/usr/sbin/sanlun lun show | egrep $device | awk '{print \$1\$2}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXSanlun::get_lun_path()",
                        "Command ($cmd) returned a bad status ($status)");
        return "";
    }

    return (defined $output[0]) ? $output[0] : "";
}

################################################################################
# Subroutine Name: get_filer()
# Objective:       No-op
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filer {
    my ($self) = @_;
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_protocol
# Objective:       return protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;

    my ($protocol);

    $self->_clear_err();

    if (defined $self->{protocol}) {
        $protocol = $self->{protocol};
    } else {
        $self->_set_err("Error: ESXSanlun::get_protocol() " .
                        "Sanlun object has no protocol specified");
        $protocol = "";
    }

    return $protocol;
}

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;

    my ($cmd, $device, $status, @output, @vmhba_adapters, @return);

    $self->_clear_err();

    # parse arguments
    unless (defined $protocol and $protocol =~ /fcp|iscsi/i){
        $self->_set_err("Missing or invalid protocol type ($protocol)");
        return (wantarray()) ? () : [];
    }

    # FCP
    if ($protocol =~ /fcp/i) {
        # not using sanlun fcp show adapter -v due to a bug in the current release that
        # displays duplicate adapter names.

        @vmhba_adapters = @{$self->{parent}->{vmhba_fcp_adapters}};

        foreach my $adapter (@vmhba_adapters) {
            if ($self->{parent}->get_os_version() =~ /^4\./) {
                # esx4 Output looks like:
                # vmhba2  qla2xxx           link-up   fc.2001001b3220d037:2101001b3220d037    (7:0.1) QLogic Corp ISP2432-based 4Gb Fibre Channel to PCI Express HBA
                $cmd = "/usr/sbin/esxcfg-scsidevs -a | grep -P \"^$adapter\\s\" | awk {\'print \$4\'}";
            } else {
                # esx3 Output looks like:
                # vmhba1 10000000c96e4e9c 19:0.0
                # vmhba2 10000000c96e4e9d 19:0.1
                # vmhba32 iqn.1998-01.com.vmware:ibm3550-rtp05-72308013 sw
                $cmd = "/usr/sbin/esxcfg-mpath -a | grep -P \"^$adapter\\s\" | awk {\'print \$2\'}";
            }

            my ($status, @wwpn) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0 or scalar(@wwpn) == 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return (wantarray()) ? () : [];
            }

            $wwpn[0] =~ s/\s//g;

            # if on ESX4, parse the wwpn differently.
            if ($self->{parent}->get_os_version() =~ /^4\./) {
                $wwpn[0] =~ s/.+\:(\S+)/$1/;
            }

            push(@return, "adapter name: " . $adapter);
            push(@return, "WWPN: " . $wwpn[0]);
        }
    }

    # iSCSI
    if ($protocol =~ /iscsi/i) {

        # Software iSCSI
        if ($self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /^software$/i) {
            @vmhba_adapters = @{$self->{parent}->{vmhba_swiscsi_adapters}};

            # check if swiscsi is running,
            $cmd = "/usr/sbin/esxcfg-swiscsi --query";
            ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Could not query the Software iSCSI service");
                return (wantarray()) ? () : [];
            }

            # if its not enabled, enable it.
            if (grep {/Software iSCSI is not enabled/i} @output) {
                $cmd = "/usr/sbin/esxcfg-swiscsi --enable";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Could not enable to Software iSCSI service");
                    return (wantarray()) ? () : [];
                }

                # we may need to reboot the esx machine after enabling the iscsi initiator                
                if (grep(/reboot/i, @output)) {
                    $self->{parent}->reboot();
                    if ($self->{parent}->error()) {
                        $self->_set_err($self->{parent}->errmsg());
                        return (wantarray()) ? () : [];
                    }
                }
            }
        }

        # Hardware iSCSI
        elsif ($self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /^hardware$/i) {
            @vmhba_adapters = @{$self->{parent}->{vmhba_hwiscsi_adapters}};
        }

        # determine the iqn name for each adapter
        foreach my $vmhba_adapter (@vmhba_adapters) {

            # display the info for this vmhba_adapter
            $cmd = "/usr/sbin/vmkiscsi-tool $vmhba_adapter -I";
            my ($status, @initiator_names) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Could not obtain the initiator name for $vmhba_adapter");
                return (wantarray()) ? () : [];
            }

            # now substitute "iSCSI Node Name:" for "Initiator node name:"
            @initiator_names =
                map {
                    if (s/^iSCSI Node Name\s*:/Initiator node name:/) {
                        $_;
                    }
                } @initiator_names;

            # push the results onto the final output
            push(@return, @initiator_names);
        }
    }

    return (wantarray()) ? @return : \@return;
}

################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self, %args ) = @_;

    $self->_clear_err();

    my $cmd = "/usr/sbin/sanlun lun show";

    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 or grep(/Segmentation fault/i, @output)) {
        $self->_set_err("Error: ESXSanlun::lun_show()",
                        "sanlun lun show failed");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}


################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self,
        $parent
       );

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: ESXSanlun::_new() Invalid value for ".
                                "argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for missing -parent argument
    if (not defined($parent)) {
        $self->_set_err("Error: ESXSanlun::_new()",
                        "Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{protocol} = $parent->get_protocol();
    $self->{parent} = $parent;

    return $self;
}

1;

