# Copyright (c) 2010,2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to an ESXi sanlun object
## @author hrivera@netapp.com, arrance@netapp.com
## @status Private
## @pod none

# Package Name
##############
package San::SanClient::Sanlun::ESXiSanlun;

use strict;

# Package Requirements
######################
use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;
use Data::Dumper;

use vars qw( @ISA );

@ISA = qw(Error SanLog Sanlun San::SanClient::Sanlun::ESXSanlun);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_protocol
# Objective:       return protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;

    my ($cmd, $status, @output, $device, @vmhba_adapters, @return);

    $self->_clear_err();

    # parse arguments
    unless (defined $protocol and $protocol =~ /fcp|iscsi/i){
        $self->_set_err("Missing or invalid protocol type ($protocol)");
        return (wantarray()) ? () : [];
    }

    # FCP
    if ($protocol =~ /fcp/i) {
        @vmhba_adapters = @{$self->{parent}->{vmhba_fcp_adapters}};

        my $core_storage_namespace = $self->{parent}->{esxcli_corestorage_namespace};

        foreach my $adapter (@vmhba_adapters) {
            my ($status, @wwpn);
            if ($self->{parent}->get_os_version() =~ /^(5|6)\./) {
                if ($self->{parent}->is_cache_data_valid("vmhba_raw_output")) {
                    @wwpn = $self->{parent}->get_cache_data("vmhba_raw_output");
                    @wwpn = grep(/^$adapter\s/, @wwpn);
                    @wwpn = map { s/^.+(fco?e?\.\S+\:\S+).+$/$1/; $_ } @wwpn;
                } else {
                    $cmd = "esxcli $core_storage_namespace adapter list | grep -P \"$adapter\\s\" | awk {\'print \$4\'}";
                    ($status, @wwpn) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd,
                                                                         -include_host_name => 1);
                }
            } else {
                $cmd = "/usr/bin/esxcfg-scsidevs -a | grep -P \"$adapter\\s\" | awk {\'print \$3\'}";
                ($status, @wwpn) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
            }
            #esxi output looks like
            #vmhba5      lpfc820        fc.20000000c972201a:10000000c972201a             (6:0.0)  LPe11000 4Gb Fibre Channel Host Adapter
            #vmhba33      fcoe      link-up     fcoe.1000001b21531dd2:2000001b21531dd2  () Software FCoE
            if ($status > 0 or scalar(@wwpn) == 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return (wantarray()) ? () : [];
            }

            $wwpn[0] =~ s/fco?e?\.\S+\:(\S+)/$1/;

            push(@return, "adapter name: " . $adapter);
            push(@return, "WWPN: " . $wwpn[0]);
        }
    }

    # iSCSI
    if ($protocol =~ /iscsi/i) {

        # Software iSCSI
        if ($self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /^software$/i) {
            @vmhba_adapters = @{$self->{parent}->{vmhba_swiscsi_adapters}};

            # check if swiscsi is running,
            if ($self->{parent}->get_os_version() =~ /^(5|6)\./) {
                $cmd = "esxcli iscsi software get";
            } else {
                $cmd = "/usr/bin/esxcfg-iscsi --swiscsi --list";
            }
            ($status, @output) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd,
                                                                   -include_host_name => 1);
            if ($status > 0) {
                $self->_set_err("Could not query the Software iSCSI service");
                return (wantarray()) ? () : [];
            }

            # if its not enabled, enable it.
            if (grep {/Software iSCSI is not enabled|false/i} @output) {
                if ($self->{parent}->get_os_version() =~ /^(5|6)\./) {
                    $cmd = "esxcli iscsi software set --enabled=\"true\"";
                } else {
                    $cmd = "/usr/bin/esxcfg-iscsi --swiscsi --enable";
                }
                ($status, @output) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd,
                                                                       -include_host_name => 1);
                if ($status > 0) {
                    $self->_set_err("Could not enable to Software iSCSI service");
                    return (wantarray()) ? () : [];
                }

                # have to update vmhba adapters if we dont have a sw iscsi vmhba here already...
                if (not scalar(@vmhba_adapters)) {
                    $self->{parent}->_update_vmhba_adapters();
                    if ($self->{parent}->error()) {
                        $self->_set_err("Failed to update vmhba adapters after starting the swiscsi service",
                                        $self->{parent}->errmsg());
                        return (wantarray()) ? () : [];
                    }
                    # reload the array
                    @vmhba_adapters = @{$self->{parent}->{vmhba_swiscsi_adapters}};
                }
            }
        }

        # Hardware iSCSI
        elsif ($self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /^hardware$/i) {
            @vmhba_adapters = @{$self->{parent}->{vmhba_hwiscsi_adapters}};
        }

        # determine the iqn name for each adapter
        foreach my $vmhba_adapter (@vmhba_adapters) {

            # display the info for this vmhba_adapter
            if ($self->{parent}->get_os_version() =~ /^(5|6)\./) {
                $cmd = "esxcli iscsi adapter get --adapter $vmhba_adapter";
                my ($status, @initiator_names) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd,
                                                                                   -include_host_name => 1);
                if ($status > 0) {
                    $self->_set_err("Could not obtain the initiator name for $vmhba_adapter");
                    return (wantarray()) ? () : [];
                }

                # now substitute "Name:" with "Initiator node name:"
                #need to make sure not to pick up the alias line in the output in the vma.
                @initiator_names = grep(/Name:\s+iqn/, @initiator_names);
                @initiator_names =
                    map {
                        if (s/^\s*Name:/Initiator node name:/) {
                            $_;
                        }
                    } @initiator_names;
                # push the results onto the final output
                push(@return, @initiator_names);
            } else {
                $cmd = "/usr/bin/esxcfg-iscsi --iscsiname --list $vmhba_adapter";
                my ($status, @initiator_names) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
                if ($status > 0) {
                    $self->_set_err("Could not obtain the initiator name for $vmhba_adapter");
                    return (wantarray()) ? () : [];
                }

                # now substitute "iSCSI Node Name:" with "Initiator node name:"
                #need to make sure not to pick up the alias line in the output in the vma.
                @initiator_names = grep(/iSCSI Node Name/, @initiator_names);
                @initiator_names =
                    map {
                        if (s/^\s*iSCSI Node Name\s*:/Initiator node name:/) {
                            $_;
                        }
                    } @initiator_names;
                # push the results onto the final output
                push(@return, @initiator_names);
            }
        }
    }

    return (wantarray()) ? @return : \@return;
}
################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self,
        $parent
       );

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for missing -parent argument
    if (not defined($parent)) {
        $self->_set_err("Error: Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{protocol} = $parent->get_protocol();
    $self->{parent} = $parent;

    return $self;
}

1;
