# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to sanlun utility
## @author knopp@netapp.com
## @status Private
## @pod ../Sanlun.pm

# Package Name
##############
package San::SanClient::Sanlun::HPUXSanlun;

use strict;

# Package Requirements
######################
use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog Sanlun);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return the path to a lun given a device
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self, %args) = @_;

    my ($cmd, $device, $status, @output);
    my ($filer_name, $lun_name, $lun_path);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Error: HPUXSanlun::get_lun_path() Invalid value for argument '-device'");
                return "";
            }
            next;
        } else {
            $self->_set_err("Error: HPUXSanlun::get_lun_path() Invalid parameter specified \'$_\'");
            return "";
        }
    }

    # Make sure device value format is valid
    unless (defined $device and $device =~ /c\d+t\d+d\d+$/) {
        $self->_set_err("Error: HPUXSanlun::get_lun_path() Invalid parameter specified \'$_\'");
        return "";
    }

    # make sure this is a raw device (and contains the proper prefix)
    if ($device !~ /dsk/) {
        $device = "/dev/rdsk/$device";
    } elsif ($device !~ /rdsk/) {
        $device =~ s/dsk/rdsk/;
    }

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = "export PATH=\$PATH:/usr/sbin; " . $path . "/sanlun lun show -d $device | grep $device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXSanlun::get_lun_path()", "Command ($cmd) returned a bad status ($status)");
        return "";
    }

    if (defined $output[0]) {
        # Remove leading white spaces if there
        $output[0] =~ s/^\s*//;

        ($filer_name, $lun_name) = split(/\s+/, $output[0]);

        # Append colon in the end of filer name if it is not present
        if ($filer_name !~ /:$/) {
            $filer_name .= ":";
        }

        $lun_path = $filer_name . $lun_name;
    }

    return $lun_path;
}

################################################################################
# Subroutine Name: get_filer()
# Objective:       No-op
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filer {
    my ($self) = @_;
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;
    my ($cmd, $device, $status, @output);

    $self->_clear_err();

    # parse arguments
    unless (defined $protocol and $protocol =~ /fcp|iscsi/i) {
        $self->_set_err("Error: HPUXSanlun::get_adapters() Missing or invalid protocol type");
        return (wantarray()) ? () : [];
    }

    if ($protocol =~ /fcp/i) {
        my $path = $self->{parent}->get_hu_bin_path();
        $cmd = $path . "/sanlun fcp show adapter -v";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUXSanlun::get_adapters()", "Command ($cmd) returned a bad status ($status)");

            return (wantarray()) ? () : [];
        }
    } elsif ($protocol =~ /iscsi/i) {

        # The command 'Sanlun iscsi show adapter' does not work.  Will need
        # to retrieve the information elsewhere and return it in the Sanlun
        # output format

        # Lets get software information.  The expected output looks like
        # iSCSI-00              B.11.11.03d    HP-UX iSCSI Software Initiator
        $cmd = "/usr/sbin/swlist iSCSI-00 | grep '# iSCSI-00'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUXSanlun::get_adapters()", "Command ($cmd) returned a bad status ($status)");

            return (wantarray()) ? () : [];
        }
        my ($model, $version, $description) = $output[0] =~ /\#\s+(\S+)\s+(\S+)\s+(.*)/;

        # Lets get the initiator name from the iscsi software initiator tool
        $cmd = "/opt/iscsi/bin/iscsiutil -l | grep 'Initiator Name' | awk '{print \$4}'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUXSanlun::get_adapters()", "Command ($cmd) returned a bad status ($status)");

            return (wantarray()) ? () : [];
        }

        my $initiator = $output[0];

        # Let construct the format to mimic how Sanlun returns the data
        @output = ();
        push @output, "iscsi-hp initiator name: " . $initiator, "model:   " . $model,
          "model description: " . $description, "driver version: " . $version, "WWPN: " . $initiator;

    } else {

        $self->_set_err("Error: HPUXSanlun::get_adapters() No iscsi support for this platform");
        return (wantarray()) ? () : [];

    }

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_protocol
# Objective:       return protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;
    my ($protocol);

    $self->_clear_err();

    if (defined $self->{protocol}) {
        $protocol = $self->{protocol};
    } else {
        $self->_set_err("Error: HPUXsanlun::get_protocol() Sanlun object has no protocol specified");
        $protocol = "";
    }

    return $protocol;

}

################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self, %args) = @_;
    my ($cmd, $device, $bPrimary, $status, $primary_paths, @output);

    $self->_clear_err();

    ## init variables

    # flag to retrieve only primary
    # device paths
    $bPrimary = 0;
    $primary_paths = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?_primary_paths_only/) {
            if (not length($bPrimary = $args{$_})) {
                $self->_set_err("Error: HPUXSanlun::lun_show() Invalid value for argument '-_primary_paths_only'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?primary_paths$/) {
            if (not length($primary_paths = $args{$_})) {
                $self->_set_err("Error: HPUXSanlun::lun_show() Invalid value for argument '-primary_paths'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    # Can't allow both _primary_paths_only and primary_paths to both be set
    # They are mutually exclusive.
    if ($bPrimary == 1 and $primary_paths == 1) {
        $self->_set_err("Error: HPUXSanlun::lun_show() Both arguments '_primary_paths_only' and 'primary_paths' " .
                        "have been provided, they are mutually exclusive, only one of the arguments is allowed");
        return ($ClientUtil::FALSE);
    }

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd =
      ($bPrimary == 1)
      ? "export PATH=\$PATH:/usr/sbin; " . $path . "/sanlun lun show all -s"
      : "export PATH=\$PATH:/usr/sbin; " . $path . "/sanlun lun show all | egrep -i '" . $self->get_protocol() . " '";

    # Check to see if the user wants to get primary paths
    # This does override the _primary_paths_only argument.
    if ($primary_paths == 1) {
        $cmd = "export PATH=\$PATH:/usr/sbin; " . $path . "/sanlun lun show all -p";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXSanlun::lun_show()", "Command ($cmd) returned a bad status ($status)");

        return (wantarray()) ? () : [];
    }

    # If primary filter chosen pass back
    # only the luns with the specified
    # protocol
    if ($bPrimary == 1) {
        my @array =
          ($self->get_protocol() =~ /fcp/)
          ? grep (/\@primary/, @output)
          : grep (/\@iscsi/,   @output);

        @output = @array;
    }

    return (wantarray()) ? @output : \@output;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($build, $parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: HPUXSanlun::_new() Invalid value for argument '-parent'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    # initilize the object members
    $self->{parent}   = $parent;
    $self->{conn}     = $parent->get_connection();
    $self->{protocol} = $parent->get_protocol();

    return $self;
}

1;

