# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to sanlun utility
## @author knopp@netapp.com, rawat@netapp.com
## @status Private
## @pod ../Sanlun.pm

# Package Name
##############
package San::SanClient::Sanlun::LinuxSanlun;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanLog;
use San::SanClient::Sanlun;
use San::SanUtil;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog Sanlun);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;
    my ($cmd, $device, $status, @output);

    $self->_clear_err();

    # parse arguments
    unless (defined $protocol and $protocol =~ /fcp|iscsi/i) {
        $self->_set_err("Missing or invalid protocol type");
        return (wantarray()) ? () : [];
    }

    if ($protocol =~ /iscsi/i) {
        # The sanlun iscsi show adapter command is no longer supported so
        # lets find the information needed and send it back in the format
        # that "sanlun iscsi show adapter -v" would have used.
        my $initiator   = "";
        my $model       = "iSCSI driver";
        my $description = "";
        my $version     = "";
        my $firmware    = "software";
        my $type        = "linux-iscsi-sw";

        # Lets get software information.  The part of the expected output
        # that we care about looks like:
        #   Cisco iSCSI Driver Version ... 4.0.188.18 ( 21-May-2004 )

        #before calling this cmd make sure iscsid is running on the Host.

        if ($self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /hardware/) {
            $cmd = "ps -ef | grep iscsid | grep -v grep";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status).",
                                "Note: iscsid is probably not running.");
                return (wantarray()) ? () : [];
            }
        }

        if ($self->{parent}->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /hardware/) {
            if ($self->{parent}->{major_revision} =~ /rhel4|sles9/i) {
                $cmd = "/sbin/iscsi-ls | grep \"Driver Version\"";
                my ($status, @iscsi_output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Command ($cmd) returned a bad status ($status)");
                    return (wantarray()) ? () : [];
                }
    
                # So far we have seen the the following in output[0]:
                #     SFNet iSCSI Driver Version ...4:0.1.11(12-Jan-2005)
                #     Cisco iSCSI Driver Version ... 4:0.1.11 (12-Jan-2005)
                #             SFNet iSCSI Driver Version ... 3.6.3 (27-Jun-2005 )
                #
                # Note that sometimes there's a space right before the version
                # and sometimes there isn't.  If we replace "... " with "..."
                # and " (" with "(", that'll get rid of that space and put the
                # version in a predictable location.
                #
                $iscsi_output[0] =~ s/^\s+//;         # strip leading whitespace
                $iscsi_output[0] =~ s/\.\.\.\s*//;    # strip "..." or "... "
                $iscsi_output[0] =~ s/\s*\(.*//;      # strip everything after the version
                my @init_arr = split(' ',$iscsi_output[0]);

                # after the split output[0] = SFNet & output[4] = 4:0.1.11
                $description = $init_arr[0];
                $version     = $init_arr[4];
                $cmd         = "grep \"^InitiatorName=\" /etc/initiatorname.iscsi";
            } else {
                # SLES10/SLES11 & above and RHEL5/RHEL6 & above there is no iscsi-ls command and it uses open-iscsi
                # so assigning "open-iscsi" to $description and assigning iscsid version to $version
                $cmd = "/sbin/iscsid -v";
                my ($status, @iscsi_output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("iscsi packages are not installed");
                    return (wantarray()) ? () : [];
                }

                my @tmparr = split(/\s+/, $iscsi_output[0]);
                $version     = $tmparr[2];
                $description = "open-iscsi";
                $cmd         = "grep \"^InitiatorName=\" /etc/iscsi/initiatorname.iscsi";
            }
        } else {
            if ($self->{parent}->{iscsi}->{CONFIG_ISCSI_HARDWARE_VENDOR} =~ /serverengines/i) {
                $description = "Emulex/ServerEngines Hardware iSCSI Initiator";
                $model       = $description;
                $type        = "linux-iscsi-hw";
                ($initiator, $version, $firmware) = $self->_get_adapter_info_serverengines();
                if ($initiator eq "" or $self->error()) {
                    my $errmsg_str = join("\n", $self->errmsg());
                    $self->_set_err("Could not gather IQN,Driver,FW from the Emulex/ServerEninges ISCSI adapter\n" .
                                    "$errmsg_str\n");
                    return (wantarray()) ? () : [];
                }
                $cmd = "";
            } elsif  ($self->{parent}->{iscsi}->{CONFIG_ISCSI_HARDWARE_VENDOR} =~ /emulex/i) {
                $description = "Emulex Hardware iSCSI Initiator";
                $model       = $description;
                $type        = "linux-iscsi-hw";
                ($initiator, $version, $firmware) = $self->_get_adapter_info_emulex();
                if ($initiator eq "" or $self->error()) {
                    my $errmsg_str = join("\n", $self->errmsg());
                    $self->_set_err("Could not gather IQN,Driver,FW from the Emulex iSCSI adapter\n" .
                                    "$errmsg_str\n");
                    return (wantarray()) ? () : [];
                }
                $cmd = "";
            } elsif  ($self->{parent}->{iscsi}->{CONFIG_ISCSI_HARDWARE_VENDOR} =~ /chelsio/i) {
                $description = "Chelsio Hardware iSCSI Initiator";
                $model       = $description;
                $type        = "linux-iscsi-hw";
                ($initiator, $version, $firmware) = $self->_get_adapter_info_chelsio();
                if ($initiator eq "" or $self->error()) {
                    my $errmsg_str = join("\n", $self->errmsg());
                    $self->_set_err("Could not gather IQN,Driver,FW from the Chelsio iSCSI adapter\n" .
                                    "$errmsg_str\n");
                    return (wantarray()) ? () : [];
                }
                $cmd = "";
            } else {
                # this is case of HW iSCSI adapter has been specified, but no
                # specific vendor. In this case grab from a self populated file.
                # Fail safe case.
                $description = "Unknown Hardware iSCSI";
                $model       = $description;
                $type        = "linux-iscsi-hw";
                $version     = "Unknown";
                $firmware    = "Unknown";
                $cmd         = "grep \"^InitiatorName=\" /var/tmp/initiatorname.iscsi.hw";
            }
        }

        # run proper command to get the initiator information
        if (defined($cmd) and $cmd ne "") {
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return (wantarray()) ? () : [];
            }

            if (grep(/InitiatorName/, @output)) {
                $initiator = $output[0];
                $initiator =~ s/InitiatorName=//s;
                $initiator =~ s/\s.*$//;             # strip trailing whitespace
            }
        }

        # Lets construct the format to mimic how Sanlun returns the data
        @output = ();
        push(@output, "$type initiator name: " . $initiator,
                      "model: " . $model,
                      "model description: " . $description,
                      "driver version: " . $version,
                      "firmware version: " . $firmware,
                      "WWPN: " . $initiator);
    }

    # If the protocol is Fibrechannel
    if ($protocol =~ /fcp/i) {
        # There are three ways of getting the WWPNs.
        # Method 1:
        #-----------------------------
        # use "sanlun fcp show adapter -v" to get the wwpns
        #-----------------------------
        # Method 2:
        #-----------------------------
        # User has to copy the WWPNs to a temp location(say /var/tmp/nate_wwpns) manually and 
        # we will read the wwpns from this file
        # user has to copy the write WWPNs in the following format
        # host<n> WWPN1
        # host<n+1> WWPN2
        # and so on 
        #-----------------------------
        # Method 3:
        #-----------------------------
        # Get the WWPNs from the file /sys/class/scsi_host/<host_num>/port_name for rhel4
        # Get the WWPNs from the file /sys/class/fc_host/<host_num>/port_name for rhel5 
        #-----------------------------
        
        # We use three different method because method 3 doesnt work for older OSs.
        # Method 1 doesnt work for non-GAed OSs(only for emulex card type).
        # If both the methods are not working then user has to make sure that method 2 works

        my $path = $self->{parent}->get_hu_bin_path();
        $cmd = $path . "/sanlun fcp show adapter -v";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0 or grep(/Unable\s+to\s+locate\s+\/usr\/lib\d*\/libHBAAPI\.so\s+library|^error$/i, @output)
                        or grep(/\/usr\/lib\d*\/bfahbaapi.so\s+\=/i, @output)
                        or scalar(@output) == 0) {
            # Method 1 has failed. Trying method 2
            $cmd = "cat /var/tmp/nate_wwpns";
            my ($status, @port_name) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0 or grep(/No\s+such\s+file\s+or\s+directory/i, @port_name)) {
                # Method 2 has failed. Trying method 3
                $cmd = "ls /sys/class/fc_host";
                my ($status, @fc_hosts) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0 or grep(/No\s+such\s+file\s+or\s+directory/i, @fc_hosts)) {
                    # All three methods failed. Error out.
                    $self->_set_err("Not able to get the wwpns of the host" .
                                    "Copy the WWPNs in file /var/tmp/nate_wwpns in the following format\n" . 
                                    "host<n> WWPN1\nhost<n+1> WWPN2 and so on....");
                    return (wantarray()) ? () : [];
                } else {
                    my @hosts = split(/\s+/, $fc_hosts[0]);
                    foreach my $fc_host (@hosts) {
                        # No need to test for minor_revision because we have already done it
                        # For RHEL4.7 and later the port_name file is not in /sys/class/fc_host/ 
                        # its inside /sys/class/scsi_host/ dir
                        if ($self->{parent}->{major_revision} =~ /rhel4/i) {
                            $cmd = "cat /sys/class/scsi_host/$fc_host/port_name";
                        } else {
                            $cmd = "cat /sys/class/fc_host/$fc_host/port_name";
                        }
                        my ($status, @port_name) = &ClientUtil::execute($self->{conn}, $cmd);
                        if ($status > 0) {
                            $self->_set_err("Couldn't get the WWPN of the port $fc_host" .
                                            "Copy the WWPNs in file /var/tmp/nate_wwpns in the following format\n" . 
                                            "host<n> WWPN1\nhost<n+1> WWPN2 and so on....");
                            return (wantarray()) ? () : [];
                        }
    
                        # In this output the port name starts with 0x to show that it is 
                        # represented in hexadecimal format. We need to remove 0x from its name 
                        # Here is the example 
                        ##########################################################################
                        #cat /sys/class/fc_host/host7/port_name
                        #0x10000000c95518f9
                        ##########################################################################
                        my @tmp = split(//, $port_name[0]);
                        shift(@tmp);
                        shift(@tmp);
                        my $portname = join ("", @tmp); 
                            
                        my @tmp_entry = ();
                        $tmp_entry[0] = "adapter name:      $fc_host";
                        $tmp_entry[1] = "WWPN:              $portname";
                        $tmp_entry[2] = "WWNN:              \"\"";
                        $tmp_entry[3] = "driver name:       \"\"";
                        $tmp_entry[4] = "model:             \"\"";
                        $tmp_entry[5] = "model description: \"\"";
                        $tmp_entry[6] = "serial number:     \"\"";
                        $tmp_entry[7] = "hardware version:  \"\"";
                        $tmp_entry[8] = "driver version:    \"\"";
                        $tmp_entry[9] = "firmware version:  \"\"";
                        $tmp_entry[10] = "Number of ports:  \"\"";
                        $tmp_entry[11] = "port type:        \"\"";
                        $tmp_entry[12] = "port state:       \"\"";
                        $tmp_entry[13] = "supported speed:  \"\"";
                        $tmp_entry[14] = "negotiated speed: \"\"";
                        $tmp_entry[15] = "OS device name:   \"\"";
                        push (@output, @tmp_entry);  
                    }
                }    
            } else {
                @output = ();
                foreach my $line (@port_name) {
                    next if ($line =~ /^$/i);
                    my @tmp = split(/\s+/, $line);
                    
                    my @tmp_entry = ();
                    $tmp_entry[0] = "adapter name:      $tmp[-2]";
                    $tmp_entry[1] = "WWPN:              $tmp[-1]";
                    $tmp_entry[2] = "WWNN:              \"\"";
                    $tmp_entry[3] = "driver name:       \"\"";
                    $tmp_entry[4] = "model:             \"\"";
                    $tmp_entry[5] = "model description: \"\"";
                    $tmp_entry[6] = "serial number:     \"\"";
                    $tmp_entry[7] = "hardware version:  \"\"";
                    $tmp_entry[8] = "driver version:    \"\"";
                    $tmp_entry[9] = "firmware version:  \"\"";
                    $tmp_entry[10] = "Number of ports:  \"\"";
                    $tmp_entry[11] = "port type:        \"\"";
                    $tmp_entry[12] = "port state:       \"\"";
                    $tmp_entry[13] = "supported speed:  \"\"";
                    $tmp_entry[14] = "negotiated speed: \"\"";
                    $tmp_entry[15] = "OS device name:   \"\"";
                    
                    push (@output, @tmp_entry);  
                }
            }
        }
    } 

    # Unsupported protocol
    if ($protocol !~ /iscsi|fcp/) {
        $self->_set_err("No $protocol support for this platform");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: get_protocol
# Objective:       return protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;
    my ($protocol);

    $self->_clear_err();

    if (defined $self->{protocol}) {
        $protocol = $self->{protocol};
    } else {
        $self->_set_err("Sanlun  object has no protocol specified");
        $protocol = "";
    }
    return $protocol;
}

################################################################################
# Subroutine Name: get_filer()
# Objective:       Will return filer of the device(LUN)
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filer {
    my ($self, %args) = @_;

    my ($cmd, $device, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
    if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Invalid value for argument '-device'");
                return "";
            }
            next;
        }
    }

    my $path = $self->{parent}->get_hu_bin_path();

    $cmd = $path . "/sanlun lun show  | grep $device | awk '{print \$1}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return "";
    }

    $output[0] =~ s/:$//g;

    return (defined $output[0]) ? $output[0] : "";
}

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return the path to a lun given a device
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self, %args) = @_;

    my ($cmd, $device, $status, @output);
    my ($filer_name, $lun_name, $lun_path);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Invalid value for argument '-device'");
                return "";
            }
            next;
        }
    }

    # Make sure device value format is valid
    unless (defined $device and $device =~ /sd/) {
        $self->_set_err("No device specified \'$_\'");
        return "";
    }

    # Add /dev/ in the begining of device name if only device short name(sd\w+) is passed
    if ($device =~ /^sd/i and $device !~ /\/dev\//i) {
        $device = "/dev/$device";
    }

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show -d $device | grep $device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return "";
    }

    if (defined $output[0]) {
        # Remove leading white spaces if there
        $output[0] =~ s/^\s*//;

        ($filer_name, $lun_name) = split(/\s+/, $output[0]);

        # Append colon in the end of filer name if it is not present
        if ($filer_name !~ /:$/) {
            $filer_name .= ":";
        }

        $lun_path = $filer_name . $lun_name;
    }

    return $lun_path;
}

################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self, %args) = @_;
    my ($cmd, $device, $status, @output);

    $self->_clear_err();

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show all";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    # Sanlun return code 203 if no LUNs are available on the host. It also
    # prints out error_msg "no controller LUNs available" on same failure
    if (($status == 203) or (grep(/no controller LUNs available/i, @output))) {
        # No need to _set_err because "LUNs are not available" is a valid
        # scenario if PRE_CLEANUP is running. The test will fail later
        # if it is not PRE_CLEANUP
        $self->sanlog(-msg => "No LUNs are available on host");
        return (wantarray()) ? () : [];
    } elsif ($status > 0) {
        # Sometimes (due to the asynchronous nature of iscsi restart,
        # need to do the sanlun lun show twice to see LUNs
        $self->_clear_err();

        &SanUtil::nate_sleep(5);

        # No need to create cmd again.
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return (wantarray()) ? () : [];
        }
    }

    return (wantarray()) ? @output : \@output;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments                                                                                           
    foreach (keys %args) {                                                                                      
        if (/-?parent/) {                                                                                       
            if (not length($parent = $args{$_})) {                                                              
                $self->_set_err("Invalid value for argument '-parent'");           
                return $self;
            }                                                                                                   
            next;                                                                                               
        }                                                                                                       
    }                                                                                                           

    # initilize the object members                                                                              
    $self->{parent} = $parent;                                                                                  
    $self->{conn} = $parent->get_connection();                                                                  
    $self->{protocol} = $parent->get_protocol();                                                                

    return $self;
}

################################################################################
# Subroutine Name: _get_adapter_info_serverengines()
# Objective:       extract adapter information from the ServerEngines iSCSI Initiator
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_adapter_info_serverengines {
    my ($self) = @_;

    my ($cmd, $client_tool, $iqn, $status, @output);
    my ($driver, $firmware);

    # make sure that the client tool is installed properly
    $client_tool = "/opt/ServerEngines/BladeEngine2/smclp/client/client";
    $cmd = "ls $client_tool";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not locate client tool ($client_tool)\n" .
                        "Please make sure the ServerEngines client RPMs are installed");
        return (undef, undef, undef);
    }

    # execute client tool and extract info on the card - iSCSI HBA info
    $cmd = "$client_tool -x \"show /BladeEngine0\"";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not extract info on ServerEngines card");
        return (undef, undef, undef);
    }

    # grab the FW revision
    $firmware = (grep(/ActiveFirmwareVersion/i, @output))[0];
    $firmware =~ s/\s+//g;
    $firmware =~ s/ActiveFirmwareVersion\s*=\s*//;

    # execute client tool and extract info on the ISSI HBA settings
    $cmd = "$client_tool -x \"show /BladeEngine0/iSCSIPhysicalHBA0\"";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not extract iSCSI HBA info on ServerEngines card");
        return (undef, undef, undef);
    }

    # execute client tool and extract info on the card ISSI Driver info
    $cmd = "$client_tool -x \"show /BladeEngine0/iSCSIPhysicalHBA0/Driver\"";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not extract driver info on ServerEngines card");
        return (undef, undef, undef);
    }

    # grab the Driver revision
    $driver = (grep(/DriverVersion/i, @output))[0];
    $driver =~ s/\s+//g;
    $driver =~ s/DriverVersion\s*=\s*//;

    # execute client tool and extract the IQN name
    $cmd = "$client_tool -x \"show /BladeEngine0/iSCSIPhysicalHBA1/LogicalHBA0/InitiatoriSCSIName\" | grep InitiatoriSCSIName";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not extract IQN name using ServerEngines client tool");
        return (undef, undef, undef);
    }

    # will produce output similar to:
    #   InitiatoriSCSIName = iqn.1990-07.com.Emulex.00-00-c9-93-bf-1c
    $iqn = (grep(/^\s*InitiatoriSCSIName\s*\=\s*/, @output))[0];
    $iqn =~ s/^\s*InitiatoriSCSIName\s*\=\s*//;

    # verify IQN format
    if ($iqn !~ /^iqn\./i) {
        $self->_set_err("IQN name extracted is not properly formatted ($iqn)");
        return (undef, undef, undef);
    } else {
       return ($iqn, $driver, $firmware);
    }

}

################################################################################
# Subroutine Name: _get_adapter_info_chelsio()
# Objective:       extract adapter information from the Chelsio iSCSI Initiator
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_adapter_info_chelsio {
    my ($self) = @_;

    my ($cmd, $client_tool, $iqn, $status, @output);
    my ($driver, $firmware);

    return ($iqn, $driver, $firmware);
}

################################################################################
# Subroutine Name: _get_adapter_info_emulex()
# Objective:       extract adapter information from the Emulex iSCSI Initiator
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_adapter_info_emulex {
    my ($self) = @_;

    my ($cmd, $client_tool, $iqn, $status, @cnas, @output);
    my ($potential_mac, $driver, $firmware);

    # make sure that the Emulex hbacmd tool is installed properly
    $client_tool = "/usr/sbin/hbacmd";
    $cmd = "ls $client_tool";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        # if hbacmd tool is not found in /usr/sbin/, try in /usr/sbin/hbanyware/ directory
        $client_tool = "/usr/sbin/hbanyware/hbacmd";
        $cmd = "ls $client_tool";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Could not locate client tool ($client_tool).\nPlease make sure " .
                            "Emulex HBAnywhere/OCmanager RPMs for iSCSI are installed");
            return (undef, undef, undef);
        }
    }

    # get the list of CNAs on the system
    $cmd = "$client_tool listHBAs";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could execute ($cmd) to get list of CNAs on the system");
        return (undef, undef, undef);
    }

    # get list of cna MAC that are iSCSI (and not NIC)
    foreach my $line (@output) {
        if ($line =~ /^(Port MAC|Permanent MAC)/i) {
            $potential_mac = $line;
        }
        if ($line =~ /Port\s+Type\s+:\s+iSCSI/i) {
            push(@cnas, $potential_mac);
        }
    }

    # get first CNA and grab MAC to use to get IQN information
    my $first_cna = $cnas[0];
    $first_cna =~ s/(Port MAC|Permanent MAC)\s*://;
    $first_cna =~ s/\s+//g;

    # get the iqn name
    $cmd = "$client_tool GetInitiatorProperties $first_cna";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could execute ($cmd) to CNA properties for CNA MAC ($first_cna)");
        return (undef, undef, undef);
    }
    $iqn = (grep(/^Initiator iSCSI Name:/, @output))[0];
    $iqn =~ s/Initiator iSCSI Name\s*://;
    $iqn =~ s/\s+//g;

    # get the CNA driver / firmware
    $cmd = "$client_tool HbaAttributes $first_cna";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could execute ($cmd) to CNA properties for CNA MAC ($first_cna)");
        return (undef, undef, undef);
    }
    $driver = (grep(/^Driver Version\s*:/, @output))[0];
    $driver =~ s/Driver Version\s*://;
    $driver =~ s/\s+//g;

    $firmware = (grep(/^Operational FW\s*:/, @output))[0];
    $firmware =~ s/Operational FW\s*://;
    $firmware =~ s/\s+//g;

    return ($iqn, $driver, $firmware);
}

1;

