# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to sanlun utility
## @author latif@netapp.com
## @status Private
## @pod ../Sanlun.pm

# Package Name
##############
package San::SanClient::Sanlun::SanBlazeSanlun;

use strict;

# Package Requirements
######################
use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;
use San::FIT::Misfit::Services::IO::Sanblaze;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog Sanlun);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;
    my ($cmd, $device, $status, @output);

    $self->_clear_err();

    # parse arguments
    unless (defined $protocol and $protocol =~ /fcp/i) {
        $self->_set_err("Missing or invalid protocol type");
        return (wantarray()) ? () : [];
    }

    # Check to see if the user specified a list of initiators to use, otherwise discover the list
    if (not defined($self->{parent}->{'user_defined_initiators'})) {
        # Retrieve the entire port list available on the SANBlaze host
        my $port_list;
        eval {
            $port_list = $self->{sanblaze_object}->sanblaze_port_list();
        };

        if ($@) {
            $self->_set_err("Unable to retrieve port list from the SANBlaze host: " . $@);
            return (wantarray()) ? () : [];
        }

        # Generate list of ports to filter on if user defines them
        if (defined($self->{parent}->{'user_defined_initiators'}->{filter_ports}->{list})) {
            foreach my $filter_port (split(/\,/, $self->{parent}->{'user_defined_initiators'}->{filter_ports}->{list})) {
                $self->{parent}->{'user_defined_initiators'}->{filter_ports}->{$filter_port} = 1;
            }
        } else {
            # No user defined port filters exist, include all of the existing ports into the filter list
            foreach my $port (keys(%{$port_list->{ports}->{port}})) {
                $self->{parent}->{'user_defined_initiators'}->{filter_ports}->{$port} = 1;
            }
        }

        # Go through each port and grab the port status to gather WWPN's
        foreach my $port (keys(%{$port_list->{ports}->{port}})) {
            # Check to see if the port is present in the filtered list
            if (not defined($self->{parent}->{'user_defined_initiators'}->{filter_ports}->{$port})) {
                # Not defined, skip to next port
                next;
            }
            my $port_status;

            eval {
                $port_status = $self->{sanblaze_object}->sanblaze_port_status(-port => $port);
            };

            if ($@) {
                $self->_set_err("Unable to retrieve port ($port) status from the SANBlaze host: " . $@);
                return (wantarray()) ? () : [];
            }

            # Gather all initiators and save them for later processing
            foreach my $init (keys(%{$port_status->{status}})) {
                if ($init =~ /initiator(\d+)_wwpn/) {
                    my $init_name = $1;
                    $self->{parent}->{'user_defined_initiators'}->{"WWPN$port$init_name"}->{VALUE} = $port_status->{status}->{$init};
                }
            }
        }
    }

    foreach my $init (keys(%{$self->{parent}->{'user_defined_initiators'}})) {
        if ($init =~ /WWPN(\d+)/) {
            my $adpt_name = "FC" . $1;
            my $wwpn = $self->{parent}->{'user_defined_initiators'}->{$init}->{VALUE};
            my @tmp_entry = ();
            $tmp_entry[0] = "adapter name:      $adpt_name";
            $tmp_entry[1] = "WWPN:              $wwpn";
            $tmp_entry[2] = "WWNN:              \"\"";
            $tmp_entry[3] = "driver name:       \"\"";
            $tmp_entry[4] = "model:             \"\"";
            $tmp_entry[5] = "model description: \"\"";
            $tmp_entry[6] = "serial number:     \"\"";
            $tmp_entry[7] = "hardware version:  \"\"";
            $tmp_entry[8] = "driver version:    \"\"";
            $tmp_entry[9] = "firmware version:  \"\"";
            $tmp_entry[10] = "Number of ports:  \"\"";
            $tmp_entry[11] = "port type:        \"\"";
            $tmp_entry[12] = "port state:       \"\"";
            $tmp_entry[13] = "supported speed:  \"\"";
            $tmp_entry[14] = "negotiated speed: \"\"";
            $tmp_entry[15] = "OS device name:   \"\"";

            push (@output, @tmp_entry);
        }
    }

    # If the list if initiators is empty, set error
    if (scalar(@output) == 0) {
        $self->_set_err("No initiators located");
        return return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: get_protocol
# Objective:       return protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;
    my ($protocol);

    $self->_clear_err();

    if (defined $self->{protocol}) {
        $protocol = $self->{protocol};
    } else {
        $self->_set_err("Sanlun  object has no protocol specified");
        $protocol = "";
    }
    return $protocol;
}

################################################################################
# Subroutine Name: get_filer()
# Objective:       Will return filer of the device(LUN)
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filer {
    my ($self, %args) = @_;

    $self->_set_err("Unsupported call to get_filer()");
    return "";
}

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return the path to a lun given a device
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self, %args) = @_;

    $self->_set_err("Unsupported call to get_lun_path");
    return "";
}

################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self, %args) = @_;

    $self->_set_err("Unsupported call to lun_show()");
    return (wantarray()) ? () : [];
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # initilize the object members
    $self->{parent} = $parent;
    $self->{conn} = $parent->get_connection();
    $self->{protocol} = $parent->get_protocol();

    # Create SANBlaze object to be used later to discover ports and wwpn's
    eval {
        $self->{sanblaze_object} = San::FIT::Misfit::Services::IO::Sanblaze->new(-host => $self->{parent}->{conn}->{connhost});
    };

    if ($@) {
        $self->_set_err("Unable to create SANBlaze object: " . $@);
        return $self;
    }

    return $self;
}

1;

