# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to sanlun utility
## @author knopp@netapp.com
## @status Private
## @pod ../Sanlun.pm

# Package Name
##############
package San::SanClient::Sanlun::SolarisSanlun;

use strict;

# Package Requirements
######################
use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog Sanlun);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return the path to a lun given a device
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self, %args) = @_;

    my ($cmd, $device, $status, @output);
    my ($filer_name, $lun_name, $lun_path);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
      if (/-?device/) {
          if (not length($device= $args{$_})) {
            $self->_set_err("Error: SolarisSanlun::get_lun_path() Invalid value for argument '-device'");
            return ""; 
          }        
          next;
      }
    }
    
    ### Make sure device value format is valid
    unless (defined $device and $device =~ /^c\d+t\w+d\d+/){
        $self->_set_err("Error: SolarisSanlun::get_lun_path() Invalid parameter specified \'$_\'");
        return "";   
    }

    ### IMPORTANT: Original device target may have changed due to failover.  So 
    ### make sure the search takes this into account.
    my($ctrl,$targ,$dskid) = $device =~ /(c\d+)(t\w+)(d\d+)/i;
    
    $device = "/dev/rdsk/".$ctrl.$targ.$dskid."s2";
    
    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show -d  $device | egrep $device";
    ($status, @output) = &ClientUtil::execute($self->{conn},$cmd);
    if ($status > 0) {
        $self->_set_err("Error: SolarisSanlun::get_lun_path()",
                "Command ($cmd) returned a bad status ($status)");
        return "";
    }

    if (defined $output[0]) {
        # Remove leading white spaces if there
        $output[0] =~ s/^\s*//;

        ($filer_name, $lun_name) = split(/\s+/, $output[0]);

        # Append colon in the end of filer name if it is not present
        if ($filer_name !~ /:$/) {
            $filer_name .= ":";
        }

        $lun_path = $filer_name . $lun_name;
    }

    return $lun_path;
}

################################################################################
# Subroutine Name: get_protocol
# Objective:       return protocol  
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;

    my ($protocol);
    
    $self->_clear_err();

    if (defined $self->{protocol}) {
       $protocol = $self->{protocol};
    } else {
        $self->_set_err("Error: SolarisSanlun::get_protocol() Sanlun object has no protocol specified");
        $protocol = "";     
    }

    return $protocol;
         
} 

################################################################################
# Subroutine Name: get_filer()
# Objective:       No-op
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filer {
    my ($self) = @_;
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;

    my ($cmd, $device, $status, @output);

    $self->_clear_err();

    # parse arguments
    unless (defined $protocol and $protocol =~ /fcp|iscsi|nfs/i){
        $self->_set_err("Error: SolarisSanlun::display_adapters() Missing or invalid protocol type");
        return (wantarray()) ? () : []; 
    }
    
    if ($protocol =~ /fcp/i) {
        my $path = $self->{parent}->get_hu_bin_path();
        $cmd = $path . "/sanlun fcp show adapter -v";
        ($status, @output) = &ClientUtil::execute($self->{conn},$cmd);
        if ($status > 0) {
            # if sanlun status is 208 that means that no adapters are
            # present. Not necessarily an error (just saying no adapters).
            # Don't error out, just return empty array.
            if ($status != 208) {
                $self->_set_err("Error: SolarisSanlun::display_adapters()",
                                "Command ($cmd) returned a bad status ($status)");
            } 
            return (wantarray()) ? () : []; 
        }
    } elsif ($protocol =~ /iscsi/i) {
        $cmd = "iscsiadm list initiator-node";
        ($status, @output) = &ClientUtil::execute($self->{conn},$cmd);
        if ($status > 0) {
            $self->_set_err("Error: SolarisSanlun::display_adapters()",
                            "Command ($cmd) returned a bad status ($status)");
            return (wantarray()) ? () : [];
        }
    } elsif ($protocol =~ /nfs/i) {
        return (wantarray()) ? () : []; 
    } else {
        $self->_set_err("Error: SolarisSanlun::display_adapters() No protocol ($protocol) support for this platform");
        return (wantarray()) ? () : []; 
    }

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns. 
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self, %args) = @_;

    my ($bPrimary, $cmd, $device, $status, @output, $mp_type);
    
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) {     
        if (/-?_primary_paths_only/) {
            if (not length($bPrimary = $args{$_})) {
                $self->_set_err("Error: SolarisSanlun::lun_show() Invalid value for argument '-_primary_paths_only'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        } elsif (/-?mp_type/) {
            if (not length($mp_type = $args{$_})) {
                $self->_set_err("Error: SolarisSanlun::lun_show() Invalid value for argument '-mp_type'");
                return ($ClientUtil::FALSE);
            }
            next;
        } 
    }
    
    # Special command for MPXIO
    my $path = $self->{parent}->get_hu_bin_path();
    if ($bPrimary == 1 and $mp_type !~ /mpxio/i) {
        $cmd = $path . "/sanlun lun show -p";
    } else {
        $cmd = $path . "/sanlun lun show all";
    }
    
    ($status, @output) = &ClientUtil::execute($self->{conn},$cmd);
    if ($status > 0) {
        $self->_set_err("Error: SolarisSanlun::lun_show()",
                        "Command ($cmd) returned a bad status ($status)",
                        @output);
        return (wantarray()) ? () : []; 
    }

    if ($bPrimary == 1 and $mp_type !~ /mpxio/i) {
        @output = grep(/ \(lun\ \d+\)/i, @output);
    }
    
    return (wantarray()) ? @output : \@output; 

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($parent, $self);

    $self = {};

    bless $self, $class;    

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: SolarisSanlun::_new() Invalid value for argument '-parent'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }     

    # initilize the object members
    $self->{parent} = $parent;
    $self->{conn} = $parent->get_connection();
    $self->{protocol} = $parent->get_protocol();

    return $self;
}
