# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to sanlun utility
## @author hrivera@netapp.com
## @status Private
## @pod ../Sanlun.pm


# Package Name
##############
package San::SanClient::Sanlun::VLinuxSanlun;

use strict;

# Package Requirements
######################
use File::Basename;

use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;
use San::SanClient::Sanlun::LinuxSanlun;
use Error;
use San::SanLog;

use vars qw( @ISA );

@ISA = qw(San::SanClient::Sanlun::LinuxSanlun Error SanLog Sanlun);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;

    $self->_clear_err();

    return (wantarray()) ? () : [];
}

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self, %args) = @_;

    $self->_clear_err();

    return "/vol/virtual/virtual_lun";
}

################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self,%args) = @_;

    my ($cmd, $status, @output, @vendor);

    $self->_clear_err();

    # pull out all(ide, virtio, scsi) devices
    $cmd = "ls /sys/block | grep -P \'^x*[vhs]d[a-z]*\$\'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    # not checking for error because status could be 0 or 1 depending
    # on if devices were found or not.

    # filter out partitions IE: /dev/sdb1, /dev/sdc2
    @output = grep {
        if ($_ !~ /[0-9]$/) {
            1;
        } else {
            0;
        }
    } @output;

    my @lun_show_output = ();

    # build the header
    push(@lun_show_output, "\n filer        lun-pathname            ".
                           "device   vendor protocol    ".
                           "lun-size        lun-state\n");

    my $filer_name = "VIRTUAL:";
    my $lun_path = "/vol/virtual/virtual_lun";
    my $vendor = "";
    my $protocol = "VirtualDisk";
    my $lun_size = "";
    my $lun_state = "GOOD";

    # Get multipathed root device
    my @mp_root_devices = $self->{parent}->get_root_dev();
    if ($self->{parent}->error) {
        $self->_set_err($self->{parent}->errmsg());
        return (wantarray()) ? () : [];
    }

    # The root will be on multipathed device and devices that we have
    # discovered are not multipathed devices. So we have to get
    # all the sd devices belongs to multipathed root(/) device
    my @sd_root_devices;
    my @other_root_devices;
    foreach my $root_dev (@mp_root_devices) {
        if ($root_dev =~ /^hd|^vd|^xvd/) {
            # if this is an IDE/vd/xvd root device, add it to the other_root_devices list
            push(@other_root_devices, "/dev/${root_dev}");
        } elsif ($root_dev =~ /^sd/) {
            # if this is a SCSI device (non-multipathed), then add it to the sd_root_devices array
            push(@sd_root_devices, "/dev/${root_dev}");
        } else {
            $cmd = "multipath -ll $root_dev";
            my ($status, @multipath_output) = &ClientUtil::execute($self->{conn}, $cmd);
            my @sd_dev_info = grep (/\s+sd\w+\s+/i, @multipath_output);
            foreach my $line (@sd_dev_info) {
                my ($sd_dev) = ($line =~ /\s+(sd\w+)\s+/i);
                # add the /dev/ at the beginning if it doesnt exist
                $sd_dev = "/dev/" . $sd_dev;
                push (@sd_root_devices, $sd_dev);
            }
        }
    }

    # get cdrom devices and exclude them
    my @cdrom_devices;
    my @cdrom_output;
    $cmd = "ls -l /dev/cdrom";
    ($status, @cdrom_output) = &ClientUtil::execute($self->{conn}, $cmd);
    # don't fail if the cmd returns errors status
    if ($status == 0) {
        my @fields = split(/\s+/, $cdrom_output[0]);
        push (@cdrom_devices, $fields[-1]);
    }

    # get the devices mounted as XenServer Tools and exclude them
    my @disk_lable_output;
    my @xs_tools_devices;
    $cmd = "ls -l /dev/disk/by-label";
    ($status, @disk_lable_output) = &ClientUtil::execute($self->{conn}, $cmd);
    # don't fail if the cmd returns errors status
    # @disk_lable_output would look like this
    # total 0
    # lrwxrwxrwx. 1 root root 10 Jun 20 13:36 XenServer\x20Tools -> ../../xvdd
    @disk_lable_output = grep(/XenServer.*Tools/, @disk_lable_output);
    foreach my $line (@disk_lable_output) {
        my @split_fields = split(/\s+/, $line);
        if ($split_fields[-1] =~ /(xvd\w+)/) {
            push(@xs_tools_devices, "/dev/" . $1);
        }
    }

    # add each device to the output
    foreach my $device (@output) {
        # skip any invalid lines or if its the boot device
        # being very paranoid here.
        next if ($device !~ /x*[vhs]d/);
        next if ($device =~ /\/dev\/x*[vhs]d$/);

        # add the /dev/ at the beginning if it doesnt exist
        if ($device !~ /^\/dev\//) {
            $device = "/dev/" . $device;
        }

        # Exclude the ide/vd/xvd root devices
        if (grep(/^$device$/, @other_root_devices) > 0){
            next;
        }

        # Exclude the scsi root devices
        if (grep(/^$device$/, @sd_root_devices) > 0){
            next;
        }

        # Exclude the xenserver tools devices
        if (grep(/^$device$/, @xs_tools_devices) > 0){
            next;
        }

        # obtain the size of this device
        my ($short_name) = basename($device);

        # Exclude cdrom devices
        if (grep(/^$short_name$/, @cdrom_devices) > 0){
            next;
        }

        # exclude ide devices that are cdroms
        if ($short_name =~ /^hd/) {
            $cmd = "cat /sys/block/${short_name}/device/media | grep cdrom";
            ($status) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status == 0) {
                # the grep matched, this device is a cdrom skip it
                next;
            }
        }

        $cmd = "fdisk -l $device | grep -i ^Disk";
        my ($status, @output_array) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            # if we cannot determine the size, then this device is not valid
            # For SLES9, this may happen because there exist placeholders for every
            # [vhs]d* device in /dev/ even if the devices do not exist.
            next;
        }
        # @output_array could look like this:
        # Disk /dev/sdc doesn't contain a valid partition table
        # Disk /dev/sdc: 5368 MB, 5368709120 bytes
        my ($line) = grep(/\d+\s+bytes/i, @output_array);
        my ($size_bytes) = ($line =~ /(\d+)\s+bytes/i);
        my $size_gigabytes = &ClientUtil::convert_to_bytes($size_bytes, 'g');
        if ($size_gigabytes < 1) {
            # If lun size is less than 1 giga byte, we have to report the size in mega bytes.
            my $size_megabytes = &ClientUtil::convert_to_bytes($size_bytes, 'm');
            $lun_size = $size_megabytes . "m (" . $size_bytes . ")";
        } else {
            $lun_size = $size_gigabytes . "g (" . $size_bytes . ")";
        }

        # obtain the vendor of this device (only if it's a SCSI disk)
        if ($short_name =~ /^sd/) {
            $cmd = "cat /sys/block/$short_name/device/vendor";
            ($status, @vendor) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $vendor = "Unknown";
            } else {
                $vendor = $vendor[0];
            }
        } elsif ($short_name =~ /^hd/) {
            # otherwise this is an IDE disk
            $vendor = "IDE";
        } elsif ($short_name =~ /^xvd/) {
            $vendor = "XEN";
        }

        # build this line for lun show
        my $lun_show_line = $filer_name . "  " . $lun_path . " " . $device .
                            "  " . $vendor . "  " . $protocol . " " .
                            $lun_size . " " . $lun_state . "\n";
        push(@lun_show_output, $lun_show_line);
    }

    $self->sanlog(-msg => "@lun_show_output");

    return (wantarray()) ? @lun_show_output : \@lun_show_output;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my (
        $self,
        $parent
       );

    $self = {};

    bless($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: VLinuxSanlun::_new() Invalid value for".
                                " argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for missing -parent argument
    if (not defined($parent)) {
        $self->_set_err("Error: VLinuxSanlun::_new()",
                        "Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{protocol} = $parent->get_protocol();
    $self->{parent} = $parent;

    return $self;
}


1;

