# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to sanlun utility
## @author hrivera@netapp.com
## @status Private
## @pod ../Sanlun.pm

# Package Name
##############
package San::SanClient::Sanlun::VSolarisSanlun;

use strict;

# Package Requirements
######################
use File::Basename;

use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;
use San::SanClient::Sanlun::SolarisSanlun;
use Error;
use San::SanLog;

use vars qw( @ISA );

@ISA = qw(San::SanClient::Sanlun::SolarisSanlun Error SanLog Sanlun);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self, %args) = @_;
    
    $self->_clear_err();
    
    return "/vol/virtual/virtual_lun";
}

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;
    
    $self->_clear_err();
        
    return (wantarray()) ? () : [];
}

################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns. 
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self, %args) = @_;
    
    my ($cmd,
        $status,
        @output,
        @lun_show_output,
       );
    
    $self->_clear_err();

    # call /usr/sbin/format to display the available devices
    $cmd = "echo -e \"disk\n0\nquit\" | /usr/sbin/format";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: VSolarisSanlun::lun_show()",
                        "Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }
    
    my @devices = ();

    # parse the output, and pull out each device name found.
    # ignore the the disk if the device number is 0.
    foreach my $line (@output) {
        if ($line =~ /\s*(\d+)\.\s*(c\d+t\d+d\d+)\s*/) {
            my $device_number = $1;
            my $device_name = $2;

            # we will temporarily store the device info in this hash
            my $dev_hsh = {};
            
            # ignore the device that is mounted on /
            # df -k / | grep $device_name
            $cmd = "df -k | grep $device_name";
            my ($df_status, @df_output) = &ClientUtil::execute($self->{conn}, $cmd);
            if (grep(/\%\s*\/\s*$/, @df_output)) {
                # then $device_name is the boot disk. Skip it.
                next;
            }

            # get the vendor name of this device
            $cmd = "echo -e \"disk\n" . $device_number ."\ninquiry\nquit\" | /usr/sbin/format | grep Vendor";
            my ($vendor_status, @vendor_output) = &ClientUtil::execute($self->{conn}, $cmd);
            foreach my $vendor_line (@vendor_output) {
                ($dev_hsh->{vendor}) = ($vendor_line =~ /Vendor\:\s*(\S+)\s*/);
            }
            
            # store the device path
            $dev_hsh->{device_path} = "/dev/rdsk/" . $device_name . "s2";
            $dev_hsh->{partition_name} = "/dev/rdsk/" . $device_name . "p0";
            $dev_hsh->{device_number} = $device_number;
            push(@devices, $dev_hsh);
        }
    }
    
    # Emulate "sanlun lun show"
    # build the header
    push(@lun_show_output, "   filer:        lun-pathname  ".
                           "device filename  vendor  protocol  ".
                           "lun size  lun state  device_number  partition_name\n");     
    my $i = 0;

    # add each device to the output
    foreach my $device (@devices) { 
        my $filer_name = "VIRTUAL:";
        my $lun_path = "/vol/virtual/virtual_lun";
        my $vendor = $device->{vendor};
        my $protocol = "VirtualDisk";
        my $lun_size = "0g (0)";
        my $lun_state = "GOOD";
        my $device_path = $device->{device_path};
        my $device_number = $device->{device_number};
        my $partition_name = $device->{partition_name};

        # build this line for lun show
        my $lun_show_line = $filer_name . "  " . $lun_path . "_$i" . " " . $device_path . 
            "  " . $vendor . "  " . $protocol . " " . $lun_size . " " . $lun_state . " " . $device_number . " " . $partition_name  . "\n";
        $i++;
        push(@lun_show_output, $lun_show_line);
    }

    $self->sanlog(-msg => "@lun_show_output");
    chomp(@lun_show_output);

    return (wantarray()) ? @lun_show_output : \@lun_show_output; 
}


################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ( 
        $parent,
        $self,
       );

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: VSolarisSanlun::_new() Invalid value for argument '-parent'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }     

    # initilize the object members
    $self->{parent} = $parent;
    $self->{conn} = $parent->get_connection();
    $self->{protocol} = $parent->get_protocol();

    return $self;
}


1;
