# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to sanlun utility
## @author pressley@netapp.com
## @status public
## @pod ../Sanlun.pm


# Package Name
##############
package San::SanClient::Sanlun::WinSanlun;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;

use Error;
use San::SanLog;

use vars qw(@ISA $Sanlun);

@ISA = qw(Error SanLog Sanlun);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_lun_path()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_path {
    my ($self, %args) = @_;

    my $device;
    my $lun_path = "";

    $self->_clear_err();

    #
    # parse arguments
    #
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device= $args{$_})) {
                $self->_set_err("Error: WinSanlun::get_lun_path() " .
                                "Invalid value for argument '-device'");
                return "";
            }
            next;
        } else {
            $self->_set_err("Error: WinSanlun::get_lun_path() " .
                            "Invalid parameter specified \'$_\'");
            return "";
        }
    }

    #
    # Make sure device value format is valid
    #
    unless (defined $device  and length($device)) {
        $self->_set_err("Error: WinSanlun::get_lun_path() " .
                        "Invalid parameter specified \'$_\'");
        return "";
    }

    #
    # Update WinClients device (disk) list
    #
    $self->{parent}->_WinClient_getDevices(-refresh => 0);
    if ($self->{WinClient}->error) {
        $self->_set_err("Error: WinSanlun::get_lun_path() " .
                        "WinClient failed to get devices");
        return "";
    }

    #
    # When we called getDevices, it filled the WinClient disks element with a hash
    # of disk objects keyed by disk number.
    #
    foreach my $key (keys %{$self->{WinClient}->{disks}}) {
        my $diskObj = $self->{WinClient}->{disks}{$key};

        #
        # Are we looking at the device we want?
        #
        if (($device eq $diskObj->{rawName}) or
            ($device eq $diskObj->{device} )) {
            unless (defined $diskObj->{lunPath}) {
                $self->_set_err("Error: WinSanlun::get_lun_path() " .
                                "No lunPath defined for device "    .
                                $device);
                return "";
            }

            unless (defined $diskObj->{filer}) {
                $self->_set_err("Error: WinSanlun::get_lun_path() " .
                                "No filer defined for device "    .
                                $device);
                return "";
            }

            #
            # Save the lun path as filer:path
            #
            $lun_path = $diskObj->{filer} . ":" . $diskObj->{lunPath};
            last; # exit loop, we have our device
        }
    }

    return $lun_path;
}

################################################################################
# Subroutine Name: get_filer()
# Objective:       No-op
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filer {
    my ($self) = @_;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display_adapters()
# Objective:       return adapter info for protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub display_adapters {
    my ($self, $protocol) = @_;

    my @output;
    my @getoutput;

    $self->_clear_err();

    #
    # parse arguments
    #
    unless (defined $protocol and $protocol =~ /fcp|iscsi/i) {
        $self->_set_err("Error: WinSanlun::get_adapters() " .
                        "Missing or invalid protocol type");
        return (wantarray()) ? () : [];
    }

    if ($protocol =~ /fcp/i) {
        @getoutput = $self->{WinClient}->getFcpInitiators();
        if ($self->{WinClient}->error()) {
            my @tmp_err = $self->{WinClient}->errmsg();
            $self->_set_err("Error: WinSanlun::get_adapters()"    .
                            " could not get FCP initiator date: " .
                            "@tmp_err");
            return (wantarray()) ? () : [];
        }

        #
        # We will create a hash of hash where the top level hash is
        # keyed by the AdapterName.
        #
        my %fcpHash;
        my $name = "_error_";
        foreach my $line (@getoutput) {
            next if $line =~ /^\s*$/;       # skip empty lines
            $line         =~ s/^\s+//;      # strip leading whitespce

            $self->sanlog(-msg_level => 9,
                          -msg => "WinSanlun::display_adapters getInitiators " .
                                  "Line is $line");
            if ($line =~ /^AdapterName:/) {
                (undef, $name) = split(/\s+/, $line);
                # burt684200 convert + signs to dashes
                $name =~ s/\+/-/g;

                $fcpHash{$name}{AdapterName} = $name;
            } elsif ($line =~ /^PortWWN:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{WWPN} = $value;
            } elsif ($line =~ /^NodeWWN:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{WWNN} = $value;
            } elsif ($line =~ /^DriverName:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{DriverName} = $value;
            } elsif ($line =~ /^Model:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{Model} = $value;
            } elsif ($line =~ /^ModelDescription:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{ModelDescription} = $value;
            } elsif ($line =~ /^SerialNumber:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{SerialNumber} = $value;
            } elsif ($line =~ /^HardwareVersion:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{HardwareVersion} = $value;
            } elsif ($line =~ /^DriverVersion:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{DriverVersion} = $value;
            } elsif ($line =~ /^FirmwareVersion:/) {
                my (undef, $value) = split(/\s+/, $line);
                if (not defined($value) or $value eq "") {
                    $value = "NA";
                }
                $fcpHash{$name}{FirmwareVersion} = $value;
            } elsif ($line =~ /^NumberOfPorts:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{NumberOfPorts} = $value;
            } elsif ($line =~ /^PortType:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{PortType} = $value;
            } elsif ($line =~ /^PortState:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{PortState} = $value;
            } elsif ($line =~ /^PortSupportedSpeed:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{PortSupportedSpeed} = $value;
            } elsif ($line =~ /^PortSpeed:/) {
                my (undef, $value) = split(/\s+/, $line);
                $fcpHash{$name}{PortSpeed} = $value;
            }
        }

        #
        # We now have our hash holding all our information, let's put it
        # in sanlun fcp show adapter format:
        #
        foreach my $name (keys %fcpHash) {
            my $line;
            my $adapterHash = $fcpHash{$name};
            $line = "$name              WWPN:" . $adapterHash->{WWPN};
            push (@output, $line);
            $line = "WWNN:              " . $adapterHash->{WWNN};
            push (@output, $line);
            $line = "driver name:       " . $adapterHash->{DriverName};
            push (@output, $line);
            $line = "model:             " . $adapterHash->{Model};
            push (@output, $line);
            $line = "model description: " . $adapterHash->{ModelDescription};
            push (@output, $line);
            $line = "serial number:     " . $adapterHash->{SerialNumber};
            push (@output, $line);
            $line = "hardware version:  " . $adapterHash->{HardwareVersion};
            push (@output, $line);
            $line = "driver version:    " . $adapterHash->{DriverVersion};
            push (@output, $line);
            $line = "firmware version:  " . $adapterHash->{FirmwareVersion};
            push (@output, $line);
            $line = "Number of ports:   " . $adapterHash->{NumberOfPorts};
            push (@output, $line);
            $line = "port type:         " . $adapterHash->{PortType};
            push (@output, $line);
            $line = "port state:        " . $adapterHash->{PortState};
            push (@output, $line);
            $line = "supported speed:   " . $adapterHash->{PortSupportedSpeed};
            push (@output, $line);
            $line = "negotiated speed:  " . $adapterHash->{PortSpeed};
            push (@output, $line);
            $line = "OS device name:    N/A";  # not applicable for windows
            push (@output, $line);
        }
    }
    if ($protocol =~ /iscsi/i) {
        # make sure the sw iscsi service is enabled
        my ($status, @start_output, @iscsi_output);
        if ($self->{WinClient}->{hostver} =~ /2008|windows server 7/) {
            ($status, @start_output) = &ClientUtil::execute($self->{parent}->{conn},
                                                            "net start \"Microsoft iSCSI Initiator Service\"");
            if ($status != 0 or not grep(/was started successfully|has already been started/i, @start_output)) {
                $self->_set_err("Failed to start the iSCSI service on the client");
                return (wantarray()) ? () : [];
            }
        }

        #
        # Display the iscsi adapters
        #
        ($status, @iscsi_output) = $self->{WinClient}->iscsiShowAdapters();
        if ($self->{WinClient}->error()) {
            my (@errmsg) = $self->{WinClient}->errmsg();
            $self->_set_err("Error: WinSanlun::display_adapters() ".
                            "iscsiShowAdapters() failed:\n @errmsg");
            return (wantarray()) ? () : [];
        }

        #
        # We are expecting the format of @output to look like this:
        #
        # iSCSINodeName:: iqn.xxxxxx...
        #
        @iscsi_output = grep(/iSCSINodeName:/i, @iscsi_output);

        if (scalar(@iscsi_output) != 1) {
            $self->_set_err("Error: WinSanlun::display_adapters() Multiple ".
                            "iscsi nodenames found in output: @iscsi_output");
            return (wantarray()) ? () : [];
        }

        my ($initiator) = ($iscsi_output[0] =~ /iSCSINodeName:\s*(iqn.\S+)/i);

        @output = ();

        my $model       = "TBD";
        my $description = "TBD";
        my $version     = "TBD";

        push(@output, "iscsi-win initiator name: $initiator");
        push(@output, "model: $model");
        push(@output, "model description: $description");
        push(@output, "driver version: $version");
        push(@output, "WWNN: $initiator");
    }

    if ($protocol !~ /iscsi|fcp/) {
        $self->_set_err("Error: WinSanlun::display_adapters() No $protocol " .
                        "support for this platform");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: get_protocol
# Objective:       return protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;

    my $protocol;

    $self->_clear_err();

    if (defined $self->{protocol}) {
        $protocol=$self->{protocol};
    } else {
        $self->_set_err("Error: WinSanlun::get_protocol() " .
                        "Sanlun object has no protocol specified");
        $protocol = "";
    }

    return $protocol;
}

################################################################################
# Subroutine Name: lun_show()
# Objective:       display information about luns.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_show {
    my ($self, %args) = @_;

    my $primary = 0;   # get primary path only
    my @output;
    my $header  = "    filer:         lun-pathname         device filename" .
                  "          adapter        protocol        lun size        " .
                  "lun state";

    $self->_clear_err();

    #
    # parse arguments
    #
    foreach (keys %args) {
        if (/-?_primary_paths_only/) {
            if (not length($primary = $args{$_})) {
                $self->_set_err("Error: WinSanlun::lun_show() " .
                                "Invalid value for argument "   .
                                "'-_primary_paths_only'");
                return ($ClientUtil::FALSE);
            }
            next;
        } else {
            $self->_set_err("Error: WinSanlun::lun_show() " .
                            "Invalid parameter specified \'$_\'");
            return "";
        }
    }

    #
    # If the primary only switch isn't set put the header information
    # in the output array.
    #
    push (@output, $header) unless ($primary);

    #
    # Update WinClients device (disk) list
    #
    $self->{parent}->_WinClient_getDevices(-refresh => 0);
    if ($self->{WinClient}->error) {
        $self->_set_err("Error: WinSanlun::get_lun_path() " .
                        "WinClient failed to get devices");
        return "";
    }

    #
    # When we called getDevices, it filled the WinClient disks element with
    # a hash of disk objects keyed by disk number.
    # I believe that we will return an entry for every disk in this hash
    # with our without the primary paths only switch because the way windows
    # handles multipathing we only see one device for multiple paths (this is
    # quite different than what we see with UNIX platforms and sanlun).
    #
    foreach my $key (keys %{$self->{WinClient}->{disks}}) {
        my $diskObj = $self->{WinClient}->{disks}{$key};
        my $current_line = "";

        #
        # Don't mess with (return) the boot disk
        #
        next if ($diskObj->{boot});

        #
        # Don't mess with (return) the quorum disk (Q: drive)
        #
        next if ($diskObj->{isQuorum});

        #
        # If primary filter chosen pass back only the luns with the
        # specified protocol
        #
        if ($primary and ($self->get_protocol() ne $diskObj->{protocol})) {
            next;
        }

        if ($primary) {
            #
            # For the primary path only, we emulate
            # rtpqa-fas3070-3:/vol/vol0/my_lun (LUN 0)  /dev/rdsk/c5t2d0s2 (c5t2d0s2)
            #
            $current_line = $diskObj->{filer}   . ":"      .
                            $diskObj->{lunPath} . " (LUN " .
                            $diskObj->{lun}     . ")  "    .
                            $diskObj->{rawName} . " ("     .
                            $diskObj->{name}    . ")";
        } else {
            my $lunSize = $diskObj->{lunSize}; # lun size looks like this "50 g"
            $lunSize =~ s/\s+//g; # remove the space
            my $disk_bytes = $self->_convert_to_bytes($lunSize);

            my $filer_tmp = $diskObj->{filer};
            $filer_tmp =~ s/\s+//g; # remove any spaces
            #
            # Normal sanlun lun show line to emulate
            # rtpqa-fas3070-3: /vol/vol0/my_lun   /dev/rdsk/c5t2d0s2          lpfc2       FCP         1g (1073741824)     GOOD
            #
            # $current_line = $diskObj->{filer}    . ": "        .
            $current_line = $filer_tmp           . ": "        .
                            $diskObj->{lunPath}  . "   "       .
                            $diskObj->{rawName}  . "         " .
                            "TBD"                . "       "   .
                            $diskObj->{protocol} . "         " .
                            $lunSize             . " ("        .
                            $disk_bytes          . ")     "    .
                            $diskObj->{lunState};
        }

        #
        # Add this disk object to the output
        #
        push (@output, $current_line);
    }

    return (wantarray()) ? @output : \@output;
}


################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my $build;
    my $winclient;
    my $self;
    my $parent;

    $self = {};

    bless $self, $class;

    #
    # parse arguments
    #
    foreach (keys %args) {
        if (/-?WinClient/) {
            $winclient = $args{$_};
            if (not (ref($winclient) eq "WinClient::Host")) {
                $self->_set_err("Error: WinSanlun::_new() Invalid value for argument '-WinClient'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: WinSanlun::_new() Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    #
    # check for existance of required -WinClient argument
    #
    unless (defined $winclient and length $winclient) {
        $self->_set_err("Error: WinSanlun::_new() ".
                        "Missing required argument '-WinClient'");
        # TBD - this error comes back as "3" not the actual string.
        return $self;
    }

    #
    # initilize the object members
    #
    $self->{WinClient} = $winclient;
    $self->{conn} = $parent->get_connection();
    $self->{protocol} = $parent->get_protocol();
    $self->{parent} = $parent;

    return $self;

}

################################################################################
# Subroutine Name: _convert_to_bytes()
# Objective:       convert size strings to bytes.
# Details:         See POD documentation at the end of this file
################################################################################
sub _convert_to_bytes {
    my ($self, $size) = @_;

    my $bytes;
    my $nbr;
    my $unit;

    #
    # If no units, then use 'm'
    #
    $size .= "m" if ($size =~ /^\d+$/);

    ($nbr, $unit) = $size =~ /(\d+)(\w+)/;

    if ($unit =~/k/i) {
        $bytes = $nbr * 1024;
    } elsif ($unit =~ /m/i){
        $bytes = $nbr * 1024 * 1024;
    } elsif ($unit =~ /g/i){
        $bytes = $nbr * 1024 * 1024 * 1024;
    } elsif ($unit =~ /t/i){
        $bytes = $nbr * 1024 * 1024 * 1024 * 1024;
    }

   return $bytes;
}

1;
