# Copyright (c) 2007-2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author latif@netapp.com
## @status private
## @pod here

package SanClientCluster;

# Compiler directives.
use strict;

# Module import.
use Tharn;
use San::SanLog;
use San::SanClientCluster::ClientClusterUtil;

use Utils qw(make_host make_conn);
use Error;
use Exporter;

use vars qw(@EXPORT 
            @ISA);
 
@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new Client Cluster object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, $cluster_type);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    foreach (keys %args) {
        if (/-?cluster_type/) {
            if (not length($cluster_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-cluster_type'");
                return $self;
            }
            next;
        }
    }

    # load the appropriate Cluster libraries
    if ($cluster_type =~ /rhcs|ocfs|vcs/i) {
        require San::SanClientCluster::Linux;
        $self = San::SanClientCluster::Linux->_new(%args);
    } elsif ($cluster_type =~ /hpux-mcsg/i){
        require San::SanClientCluster::HPUX;
        $self = San::SanClientCluster::HPUX->_new(%args);
    } elsif ($cluster_type =~ /windows/i) {
        require San::SanClientCluster::Windows;
        $self = San::SanClientCluster::Windows->_new(%args);
    } else {
        $self->_set_err("Cluster type '" . $cluster_type . "' not currently supported with this library");
        return $self;
    }

    return $self if ($self->error);

    $self->{parent} = $self->{-parent};

    return $self;
}

################################################################################
# Subroutine Name: execute()
# Objective:       will execute the cmd on target node
# Details:         See POD documentation at the end of this file
################################################################################
sub execute {
    my ($self, %args) = @_;

    my ($cmd, $target_node, $timeout, $status, @output);

    foreach (keys %args) {
        if (/-?cmd/) {
            if (not length($cmd = $args{$_})) {
                $self->_set_err("Invalid value for flag -cmd");
                return (wantarray()) ? () : [];
            }
        } elsif (/-?target_node/) {
            if (not length($target_node = $args{$_})) {
                $self->_set_err("Invalid value for flag -target_node");
                return (wantarray()) ? () : [];
            }
        } elsif (/-?timeout/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Invalid value for flag -timeout");
                return (wantarray()) ? () : [];
            }
        }
    }

    # fail out if cmd is not specified
    if (!defined $cmd) {
        $self->_set_err("-cmd is a required parameter");
        return (wantarray()) ? () : [];
    }

    # Execute the cmd on primary node if target is not passed
    if (!defined $target_node) {
        $target_node = $self->get_primary_node();
    }

    if (!defined $timeout) {
        $timeout = 600;
    }

    ($status, @output) = &ClientUtil::execute($target_node->{conn}, $cmd, $timeout);
    if ($status > 0) {
        $self->_set_err("failed to execute the cmd $cmd");
        return (wantarray()) ? @output : \@output;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
#                               Private Methods
###############################################################################

1;
