# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary
## Creates HPUX - MC Service Guard Cluster Object.
##
## @author  gowtham@netapp.com
## @status
## /* $Id$ */

package San::SanClientCluster::HPUX::MCSG;

use strict;

use Error;
use San::SanLog;
use San::SanClient;
use San::SanClientCluster::HPUX::MCSG::Node;
use San::SanClientCluster::HPUX::MCSG::SharedDiskGroup;
use San::SanClientCluster::HPUX::MCSG::SharedDiskGroup::Resource;

# Define required variables
use vars qw(
      @EXPORT
      @ISA
      );

@ISA = qw( SanLog Error );

###########################################################################
# Subroutine Name: discover()
# Input: -self, $self (called when and existing cluster is found)
# Output: True or False
# Tasks: * Discovers attributes of a working cluster
#        * Creates objects for discovered devices
#        * Populates the self with available data
###########################################################################
sub discover {
    my ($self, %args) = @_;

    my ($cmd, $nodeObj, $primary_nodeObj, $sdgObj, $resourceObj, $key, $value, @pkg_host, %mapping);

    # Initialize primary node object variable for later use.
    $primary_nodeObj = $self->get_primary_node();

    # Populate nodelist.
    foreach my $client (@{$args{-client_list}}) {
        if ($client eq $primary_nodeObj->{hostname}) {
            next;
        }
        $self->add_node(-host           =>  $client,
                        -protocol       =>  $args{-protocol},
                        -mp_type        =>  $args{-mp_type},
                        -alua_support   =>  $args{-alua_support},
                        -conn_type      =>  "THARNHOST",
                        -cluster_subnet =>  $args{-cluster_subnet},
                        -parent         =>  $self,
                        -primary        =>  "0");

	# Check if there was any error adding objects to node list.
        if ($self->error()) {
            $self->_set_err("Unable to create node object for client $client." . $self->errmsg());
            return $self;
        }
    }

    # Check if the cluster is up.
    if ($self->cluster_status() ne "up") {
        $self->_set_err("Cluster $self->{cluster_name} is not up." . $self->errmsg());
        return $self;
    }

    # Find NATE created packages and draw the data structures for cleanup
    foreach my $client (@{$args{-client_list}}) {
        $cmd = "/usr/sbin/cmviewcl | grep $client | awk {'print \$1'}";

        @pkg_host = $primary_nodeObj->{sanclientObj}->execute($cmd, -hide=>"1");

        if ($primary_nodeObj->{sanclientObj}->error) {
            $self->_set_err($primary_nodeObj->{sanclientObj}->errmsg());
            return $self;
        }

        if ($pkg_host[1] =~ /^nate_/) {
            # Package is NATE created (automated)
            $pkg_host[1] =~ s/nate_//;
            $mapping{$pkg_host[0]} = $pkg_host[1];
        }
    }

    # Populating the SDG and Resources.
    while(($key, $value) = each (%mapping)) {
      $sdgObj = San::SanClientCluster::HPUX::MCSG::SharedDiskGroup->_new(-name       => "n_vg_$key",
                                                                         -clusterObj => $self,
                                                                         -parent     => $self,
                                                                         -type       => $self->{cluster_type});
      if ($sdgObj->error()) {
          $self->_set_err("Failed to create an SDG object n_vg_$key ." . $sdgObj->errmsg());
          return $self;
      }

      push (@{$self->{sdglist}}, $sdgObj);
      $resourceObj = San::SanClientCluster::HPUX::MCSG::SharedDiskGroup::Resource->_new(-name   => $value,
                                                                                        -parent => $sdgObj,
                                                                                        -load   => "1",
                                                                                        -vgname => $sdgObj->{svgname});

      if ($resourceObj->error()) {
          $self->_set_err("Failed to create an Resource object  ." . $resourceObj->errmsg());
          return $self;
      }

      push (@{$sdgObj->{resources}}, $resourceObj);

    }
    return $self;
}

###########################################################################
# Subroutine Name: get_client_dev_lists()
# Input: -self
# Return: Client_Dev_List
###########################################################################
sub get_client_dev_lists {
    my ($self) = @_;

    my (%mapping, $client_dev_list, $primary_nodeObj, $key, $value, $num_lvs, @pkg_host, $cmd);

    $num_lvs = $self->{num_lvs};
    $client_dev_list = "";
    $primary_nodeObj = $self->get_primary_node();

    foreach my $node (values %{$self->{nodelist}}) {
        $cmd = "/usr/sbin/cmviewcl | grep $node->{hostname} | awk {'print \$1'}";

        @pkg_host = $primary_nodeObj->{sanclientObj}->execute($cmd, -hide=>"1");

        if ($primary_nodeObj->{sanclientObj}->error) {
            $self->_set_err($primary_nodeObj->{sanclientObj}->errmsg());
            return $self;
        }

        if ($pkg_host[1] =~ /^nate_/) {
            # Package is NATE created (automated)
            $mapping{$pkg_host[0]} = $pkg_host[1];
        }
    }

    # Creating the client_dev_list
    while (($key, $value) = each (%mapping)) {
      $client_dev_list .= $key;

      foreach my $node (values %{$self->{nodelist}}) {
          if ($key eq $node->{hostname}) {
              foreach my $index (1 .. $num_lvs) {
                  $client_dev_list .= ",/" . $value . "0" . $index;
              }
          } else {
              foreach my $index (1 .. $num_lvs) {
                  $client_dev_list .= ",/" . $mapping{$node->{hostname}} . "0" . $index . "nfs";
              }
          }
      }

      $client_dev_list .= ";";
    }

    return $client_dev_list;
}

###########################################################################
# Subroutine Name: create_node()
# Input: -self(SanClientCluster Object), -host
# Output: -Node Object
# Task: Creates a Node Object
###########################################################################
sub create_node {
    my ($self,%args) = @_;

    my $nodeObj = San::SanClientCluster::HPUX::MCSG::Node->_new(%args);

    if ($nodeObj->error()) {
        $self->_set_err("Unable to Create New Node Object for $args{-host} for cluster $self->{cluster_name}" . $nodeObj->errmsg());
        return $self;
    }

    return $nodeObj;
}

###########################################################################
# Subroutine Name: add_node()
# Input: -self(SanClientCluster Object), -host
# Tasks: Creates a node by calling create_node()
#        Adds the node to the cluster
###########################################################################
sub add_node {
    my ($self,%args) = @_;

    my $nodeObj = $self->create_node(%args);

    if ($nodeObj->error()) {
        $self->_set_err("Unable to Add Node object" . $nodeObj->errmsg());
        return $ClientUtil::FALSE;
    }

    my $hostname = $args{-host};
    $self->{nodelist}->{$hostname} = $nodeObj;

    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: get_nodes()
# Input: -self
# Return: Array/Reference of the list of nodes
###########################################################################
sub get_nodes {
    my ($self,%args) = @_;

    return (wantarray()) ? (values %{$self->{nodelist}}) : $self->{nodelist};
}

###########################################################################
# Subroutine Name: get_node()
# Input: -nodename
# Return: Node Object for <nodename>
###########################################################################
sub get_node {
    my ($self,%args) = @_;

    my $hostname = $args{-host};

    return $self->{nodelist}->{$hostname};
}

###########################################################################
# Subroutine Name: get_primary_node()
# Input: SanClientCluster Object
# Task: Determines which node is set as primary
###########################################################################
sub get_primary_node {
    my ($self,%args) = @_;

    my ($nodeobj, $nodelist);

    $nodelist = $self->get_nodes();

    foreach $nodeobj (values %$nodelist) {
        if ($nodeobj->isprimary()) {
            return $nodeobj;
        }
    }

    return undef;
}

###########################################################################
# Subroutine Name: set_primary_node()
# Input: -nodename
# Task : Get the primary node unset it and set the given node as primary
#        node
############################################################################
sub set_primary_node {
    my ($self,%args) = @_;

    unless(grep (/^-host$/, (keys %args))) {
      $self->_set_err("Missing host name");
      return $ClientUtil::FALSE;
    }

    my $currprim = $self->get_primary_node();

    if (defined $currprim) {
        $currprim->unset_primary_node();
    }

    $currprim = $self->get_node(%args);

    if (defined $currprim) {
        $currprim->set_primary_node();
        return $ClientUtil::TRUE
    } else {
        $self->_set_err("Node $args{-host} not found in Cluster. Unable to set primary");
        return $ClientUtil::FALSE;
    }
}

###########################################################################
# Subroutine Name: get_resource_groups()
# Input: -SanClientCluster Object
# Task : Traverses SDG object(s) and collects resource group object(s)
# Return: Resource Object(s)
############################################################################
sub get_resource_groups {
    my ($self) = @_;

    my @resource_groups = ();

    foreach my $sdgObj (@{$self->{sdglist}}) {
        foreach my $resourceObj (@{$sdgObj->{resources}}) {
            push (@resource_groups, $resourceObj);
        }
    }

    return wantarray() ? @resource_groups : \@resource_groups;
}

###########################################################################
# Subroutine Name: get_disk_groups()
# Input: -self
# Return: SharedDiskGroup Object(s)
############################################################################
sub get_disk_groups {
    my ($self) = @_;

    return $self->{sdglist};
}

###########################################################################
# Subroutine Name: add_device()
# Input: -self
# Task : Populates the self with a hash containing <node_name> as KEY and
#        list of devices as VALUE
############################################################################
sub add_device {
    my ($self, %args) = @_;

    my ($primary_node_obj, $client_mp_type, @allClients, $cmd);

    $primary_node_obj = $args{-primary};
    $client_mp_type   = $self->{mp_type};
    @allClients       = @{$args{-allClients}};
    my $client_obj = $primary_node_obj->get_sanclient_obj();

    if ($primary_node_obj->error()) {
        $self->_set_err($primary_node_obj->errmsg() . "\n");
        return $self;
    }

    my $devmgrobj = $client_obj->create_devmgr();
    if ($client_obj->error()) {
        $self->_set_err($client_obj->errmsg() . "\n");
        return $self;
    }

    my @free_devices = $devmgrobj->get_free_devices();
    if ($devmgrobj->error()) {
        $self->_set_err($devmgrobj->errmsg() . "\n");
         return $self;
    }

    # Creating a hash that contains <Hostname> as key and <List of free devices in that host for VG creation> as value.
    my %devices_hostname = ();

    foreach my $client (@allClients) {
        my $nodeObj = $self->get_node(-host => $client);

        $cmd = "export PATH=\$PATH:/usr/sbin; $nodeObj->{sanclientObj}->{huobj}->{hu_bin_path}/sanlun lun show -b | grep nate_$client | awk {'print \$4'}";

        my @output = $client_obj->execute($cmd, -hide=>"1");
        if ($client_obj->error) {
            $self->_set_err("Error while executing sanlun lun show -b." . $client_obj->errmsg());
            return $self;
        }

        foreach my $device (@output){
            $device =~ s/\/dev\/dsk\///;
        }

        if ($client_obj->error()) {
            $self->_set_err($client_obj->errmsg() . "\n");
            return $self;
        }

        foreach my $index (0 .. scalar(@output)-1) {
            my $flag = 0;

            foreach my $device (@free_devices) {
                if ($device eq $output[$index]) {
                    $flag = 1;
                }
            }

            if ($flag == 0) {
                delete  $output[$index];
            }
        }

        $devices_hostname{$client} = \@output;
    }

    # Replace legacy names with contemporary disk names
    foreach (keys %devices_hostname) {
        if($client_mp_type =~ /mass/i) {
           my @actualdevs = ();
           # For HPUX 11iv3
           foreach my $device (@{$devices_hostname{$_}}) {
               $cmd = "/usr/sbin/ioscan -m dsf /dev/dsk/$device | grep disk | awk {'print \$1'}";
               my @output = $client_obj->execute($cmd, -hide => "1");
               if ($client_obj->error) {
                   $self->_set_err($client_obj->errmsg() . "\n");
                   return $self;
               }

               $output[0] =~ s/\/dev\/disk\///;
               push (@actualdevs, @output);
           }

           $devices_hostname{$_} = \@actualdevs;
        }
    }

    $self->{devlist} = \%devices_hostname;
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: cluster_create()
# Input: -self
# Task : Create the cluster with the list of nodes given
# Commands used: -cmquerycl
############################################################################
sub cluster_create {
    my ($self) = @_;

    my ($cmd, $nodes, $pnode, @output);

    $nodes = "";

    foreach my $node (values %{$self->{nodelist}}) {
        $pnode = $node if ($node->isprimary());
        $nodes = $nodes . " -n $node->{hostname}";
    }

    # create the cluster config file in primary node

    # CLUSTER_NAME                 cluster1
    # HOSTNAME_ADDRESS_FAMILY      IPV4
    # FIRST_CLUSTER_LOCK_VG        /dev/vg01
    # NODE_NAME                    hpux_11
    # NETWORK_INTERFACE            lan0
    # HEARTBEAT_IP                 10.72.210.11
    # NETWORK_INTERFACE            lan1
    # HEARTBEAT_IP                 192.168.1.11
    # FIRST_CLUSTER_LOCK_PV        /dev/disk/disk48
    # NODE_NAME                    hpux_23
    # NETWORK_INTERFACE            lan0
    # HEARTBEAT_IP                 10.72.210.23
    # NETWORK_INTERFACE            lan1
    # HEARTBEAT_IP                 192.168.1.23
    # FIRST_CLUSTER_LOCK_PV        /dev/disk/disk31
    # MEMBER_TIMEOUT               14000000
    # AUTO_START_TIMEOUT           600000000
    # NETWORK_POLLING_INTERVAL     2000000
    # NETWORK_FAILURE_DETECTION    INOUT
    # NETWORK_AUTO_FAILBACK        YES
    # SUBNET                       10.72.210.0
    # IP_MONITOR                   ON
    # POLLING_TARGET               10.72.210.1
    # SUBNET                       192.168.1.0
    # IP_MONITOR                   OFF
    # MAX_CONFIGURED_PACKAGES      300
    # VOLUME_GROUP                 /dev/vg01
    # VOLUME_GROUP                 /dev/vg02

    $cmd = "/usr/sbin/cmquerycl -v -C /etc/cmcluster/n_$self->{cluster_name}.config $nodes";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $self;
    }

    # Change the cluster name alone
    $cmd = "cat /etc/cmcluster/n_$self->{cluster_name}.config | grep \"CLUSTER_NAME\" | awk {'print \$2'}";
    @output = $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $self;
    }

    $cmd = "cat /etc/cmcluster/n_$self->{cluster_name}.config | sed 's/$output[0]/$self->{cluster_name}/' > /etc/cmcluster/n_$self->{cluster_name}.config.new";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $self;
    }

    $cmd = "mv /etc/cmcluster/n_$self->{cluster_name}.config.new /etc/cmcluster/n_$self->{cluster_name}.config";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $self;
    }

    # Propogate changes to all other nodes in the cluster
    foreach my $node (values %{$self->{nodelist}}) {
        if ($pnode->{hostname} ne $node->{hostname}) {
            $cmd = "rcp /etc/cmcluster/n_$self->{cluster_name}.config $node->{hostname}:/etc/cmcluster/";
            $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($pnode->{sanclientObj}->error()) {
                $self->_set_err($pnode->{sanclientObj}->errmsg());
                return $self;
            }
        }
    }

    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: cluster_enable()
# Input: -self
# Task : Check and apply the current configuration for the cluster
#        and run the cluster
# Commands used: -cmcheckconf
#                -cmapplyconf
#                -cmruncl
############################################################################
sub cluster_enable {
    my ($self) = @_;

    my ($all_nodes, $pnode, $cmd);

    foreach my $node (values %{$self->{nodelist}}) {
        $all_nodes .= " -n $node->{hostname}";
    }

    foreach my $node (values %{$self->{nodelist}}) {
        $pnode = $node if ($node->{isprimary} == 1);
    }

    # Check the cluster configuration
    $cmd = "/usr/sbin/cmcheckconf -v -C /etc/cmcluster/n_$self->{cluster_name}.config";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }
    &SanUtil::nate_log('-comment', "Cluster configuration checked successfully.");

    # Apply the cluster configuration
    $cmd = "/usr/sbin/cmapplyconf -f -v -C /etc/cmcluster/n_$self->{cluster_name}.config";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }
    &SanUtil::nate_log('-comment', "Cluster configuration applied successfully.");

    # Run the cluster
    $cmd = "/usr/sbin/cmruncl $all_nodes";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }
    &SanUtil::nate_log('-comment', "Cluster $self->{clustername} is up and running successfully.");

    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: cluster_status()
# Input: -self
# Task : Parse cmviewcl to check the status of the cluster
# Commands used: -cmviewcl
############################################################################
sub cluster_status {
    my ($self) = @_;

    my ($pnode, @ret_val, $cmd);

    # Return a string of the format
    # <clusterstatus>

    foreach my $node (values %{$self->{nodelist}}) {
        $pnode = $node if ($node->{isprimary} == 1);
    }

    $cmd = "/usr/sbin/cmviewcl | grep $self->{cluster_name} | awk '{print \$2}'";
    @ret_val = $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ret_val[0];
}

###########################################################################
# Subroutine Name: cluster_destroy()
# Input: -self
# Task : Halt the cluster and delete the config files of the cluster
# Commands used: -cmhaltcl
#                -cmdeleteconf
############################################################################
sub cluster_destroy {
    my ($self) = @_;

    my ($cmd, $primary_nodeObj);

    $primary_nodeObj = $self->get_primary_node();

    # If cluster is up, halt it and then destroy else directly destroy cluster configuration.
    if ($self->cluster_status() eq "up") {
        # Halt the cluster
        $cmd = "/usr/sbin/cmhaltcl";
        $primary_nodeObj->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($primary_nodeObj->{sanclientObj}->error()) {
            $self->_set_err("Unable to halt cluster $self->{cluster_name} " . $primary_nodeObj->{sanclientObj}->errmsg());
            return $self;
        }
    }

    # Destroy the cluster configuration.
    $cmd = "/usr/sbin/cmdeleteconf -f -c $self->{cluster_name}";
    $primary_nodeObj->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($primary_nodeObj->{sanclientObj}->error()) {
        $self->_set_err("Unable to delete cluster $self->{cluster_name} " . $primary_nodeObj->{sanclientObj}->errmsg());
        return $self;
    }

    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: cluster_validate()
# Input: -self
# Task : Validate each node in the cluster
###########################################################################
sub cluster_validate {
    my ($self) = @_;

    foreach my $node (values %{$self->{nodelist}}) {
        $node->validate();
        if ($node->error()) {
            $self->_set_err($node->errmsg());
            return $ClietnUtil::FALSE;
        }
        next;
    }
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: validate_interconnect()
# Input: -self
# Task : Validate cluster interconnect
###########################################################################
sub validate_interconnect {
    my ($self) = @_;

    # No-op for HPUX
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: disable_interconnect()
# Input: -self
# Task : Disable cluster interconnect
###########################################################################
sub disable_interconnect {
    my ($self) = @_;

    # No-op for HPUX
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: create_shared_disk_group()
# Input: -self
# Task : Create a shared disk group object
###########################################################################
sub create_shared_disk_group {
    my ($self, %args)=@_;

    my ($sdgObj);

    # Create a shared disk group object
    $sdgObj=San::SanClientCluster::HPUX::MCSG::SharedDiskGroup->_new(%args, -clusterObj => $self);
    
    push (@{$self->{sdg}}, $sdgObj);

    return $sdgObj;
}


###########################################################################
# Subroutine Name: get_interconnect()
# Input: -self
# Task : Get cluster interconnect
###########################################################################
sub get_interconnect {
    my ($self) = @_;

    # No-op for HPUX
    return $ClientUtil::TRUE;
}

##########################################################################
############################# PRIVATE METHODS ############################
##########################################################################

###########################################################################
# Name: _new
# Input: -host, -cluster_name
# Output: returns SanCluster object
# Tasks: * Creates empty SanCluster Object
#       * Creates Node object for -host
#       * Sets attributes of cluster on SanCluster object (-cluster_name)
#       * Sets attribute on Node object that -host is the primary node
###########################################################################
sub _new {
    my ($class, %args) = @_;

    my $self = {};

    # Make self an object of passed class
    bless ($self, $class);

    my $load_inits;
    $self->{cluster_type}   = undef;
    $self->{cluster_subnet} = undef;
    $self->{cluster_name}   = undef;
    $self->{protocol}       = undef;
    $self->{mp_type}        = undef;
    $self->{domainname}     = undef;
    $self->{sdglist}        = ();
    $self->{devlist}        = {};
    $self->{nodelist}       = {};
    $self->{quorum_type}    = undef;

    # Check if Hostname is provided
    unless(grep (/^-host$/, (keys %args))) {
      $self->_set_err("Error: San::SanClientCluster::HPUX::_new() Missing host name");
      return $self;
    }

    # Check if Clustername is provided
    unless(grep (/^-cluster_name$/, (keys %args))) {
      $self->_set_err("Error: San::SanClientCluster::HPUX::_new() Missing Cluster name");
      return $self;
    }

    # Validate protocol is provided and is FCP(supported)
    if (grep (/^-protocol$/, (keys %args))) {
        unless ($args{-protocol} =~ /^fcp$/i) {
          $self->_set_err("Error: San::SanClientCluster::HPUX::_new() Unsupported protocol type '-protocol' \(" . $args{-protocol} . "\)");
          return $self;
        }
    } else {
        $self->_set_err("Error: San::SanClientCluster::HPUX::_new() Missing protocol type.");
        return $self;
    }

    # Validate mp_type is provided and is PV_LINKS
    if (grep (/^-mp_type$/, (keys %args))) {
        unless ($args{-mp_type} =~ /^mass$|^pvlinks$/i) {
          $self->_set_err("Error: San::SanClientCluster::HPUX::_new() Unsupported mp_type '-mp_type' \(" . $args{-mp_type} . "\)");
          return $self;
        }
    } else {
        $self->_set_err("Error: San::SanClientCluster::HPUX::_new() Missing mp_type.");
        return $self;
    }

    # Validate per client cluster quoram type is provided and is VG
    if (grep (/^-quorum_type$/, (keys %args))) {
        unless ($args{-quorum_type} =~ /^lun$|^vg$/i) {
          $self->_set_err("Error: San::SanClientCluster::HPUX::_new() Unsupported quorum_type '-quorum_type' \(" . $args{-quorum_type} . "\)");
          return $self;
        }
    } else {
        $self->_set_err("Error: San::SanClientCluster::HPUX::_new() Missing quorum_type.");
        return $self;
    }

    # Initialize parameters
    $self->{cluster_type}   = $args{-cluster_type};
    $self->{cluster_name}   = $args{-cluster_name};
    $self->{cluster_subnet} = $args{-cluster_subnet};
    $self->{protocol}       = $args{-protocol};
    $self->{alua_support}   = $args{-alua_support};
    $self->{mp_type}        = $args{-mp_type};
    $self->{num_lvs}        = $args{-num_lvs};
    $self->{quorum_type}    = $args{-quorum_type};

    if ($args{-protocol} =~ /nfs/i) {
        $args{-mp_type} = "none";
        $load_inits = 0;
    } else {
        $load_inits = 1;
    }

    # Create the primary node object as a part of cluster creation
    if (not $self->add_node(-host                => $args{-host},
                            -parent              => $self,
                            -primary             => "1",
                            -protocol            => $args{-protocol},
                            -cluster_subnet      => $args{-cluster_subnet},
                            -mp_type             => $args{-mp_type},
                            -load_inits          => $load_inits,
                            -alua_support        => $args{-alua_support},
                            -vm_hpux_revision    => $args{-vm_hpux_revision},
                            -vm_hpux_extent_size => $args{-vm_hpux_extent_size})) {
        $self->_set_err("Unable to add the initial host." . $self->errmsg());
        return $self;
    }

    return $self;
}

1;
