# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary
## Creates HPUX - MC Service Guard Node Object.
##
## @author  gowtham@netapp.com
## @status
## /* $Id$ */

package San::SanClientCluster::HPUX::MCSG::Node;

use strict;
use Error;
use San::SanLog;
use San::SanClient;
use Net::Ping;

use vars qw(
      @EXPORT
      @ISA
      );

@ISA = qw( SanLog Error );

###########################################################################
# Subroutine Name: add()
# Input: -self
# Tasks: Adds a node to the cluster
###########################################################################
sub add {
    my ($self) = @_;

    # No-op for HPUX
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: set_primary_node()
# Input: -self
# Tasks: Removes a node from the cluster
###########################################################################
sub remove {
    my ($self) = @_;

    # No-op for HPUX
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: set_primary_node()
# Input: -self
# Tasks: Sets the node as Primary node in the cluster
###########################################################################
sub set_primary_node {
    my ($self, %args) = @_;

    $self->{isprimary} = 1;
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: unset_primary_node()
# Input: -self
# Tasks: Unsets the node as Primary node in the cluster
###########################################################################
sub unset_primary_node {
    my ($self, %args) = @_;

    $self->{isprimary} = 0;
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: is_primary()
# Input: -self
# Tasks: Checks if the node is primary node or not
# Return: TRUE/FALSE
###########################################################################
sub isprimary {
    my ($self, %args) = @_;

    if ($self->{isprimary} == 1) {
        return $ClientUtil::TRUE;
    }

    return $ClientUtil::FALSE;
}

###########################################################################
# Subroutine Name: validate()
# Input: -nodeObject
# Tasks: Validates the heartbeat interface and check on the list of
#      Software(s) installed
# Return: TRUE/FALSE
###########################################################################
sub validate {
    my ($self,@args) = @_;
    return $self->validate_heartbeat() && $self->validate_software() && $self->validate_nfs();
}

###########################################################################
# Subroutine Name: validate_nfs()
# Input: -nodeObject
# Tasks: Checks for parameters NFS_SERVER=1 and START_MOUNTD=1 in
#      "/etc/rc.config.d/nfsconf"
# Return: TRUE/FALSE
###########################################################################
sub validate_nfs {
    my ($self,@args) = @_;

    my($cmd, $clientObj, @buf);

    $clientObj = $self->{sanclientObj};

    # Check for NFS_CORE = 1
    $cmd = "grep  -e ^NFS_ -e ^START_MOUNTD /etc/rc.config.d/nfsconf";
    @buf = $clientObj->execute($cmd);
    if ($clientObj->error()) {
        $self->_set_err("Error during NFS Configuration check:" . $clientObj->errmsg());
        return $ClientUtil::FALSE;
    }

    my $errmsg = "";

    # NFS error if NFS_SERVER=0 or START_MOUNTD=0 in /etc/rc.config.d/nfsconf of host
    foreach my $line (@buf) {
        if ($line =~ /0/) {
            $errmsg = $errmsg."NFS Config error $line in /etc/rc.config.d/nfsconf\n";
        }
    }

    if ($errmsg eq "") {
        return $ClientUtil::TRUE;
    }

    $self->_set_err($errmsg);
    return $ClientUtil::FALSE;
}

###########################################################################
# Subroutine Name: validate_heartbeat()
# Input: -nodeObject
# Tasks: Validates the heartbeat connectivity by pinging to primarynode
#        HB IP Address
# Return: TRUE/FALSE
###########################################################################
sub validate_heartbeat {
    my ($self,@args) = @_;

    my($cmd, $clientObj, $primNodeObj, $primHBIP, @buf, $loss);

    # Obtain the primary node's Heartbeat IP address
    $primNodeObj = $self->{parent}->get_primary_node();
    $clientObj = $self->{sanclientObj};
    my @hs = split(/:/, $self->{cluster_subnet});

    foreach my  $hsubnet (@hs) {
        my $hbip = $self->{subnets}->{$hsubnet};
        $cmd = "/usr/sbin/ping $hbip -n 2";
        @buf = $clientObj->execute($cmd, -hide=>"1");
        if ($clientObj->error()) {
            $self->_set_err("Error during ping: " . $clientObj->errmsg());
            return $ClientUtil::FALSE;
        }

        # Check for the word "packet loss" and "0% packet loss"
        $loss = 0;
        foreach my $line(@buf) {
            if ($line=~ /loss/ && $line=~ / 0\% packet loss/) {
                $loss = 1;
            }
        }

        if ($loss == 0) {
            $self->_set_err("Ping to $hbip failed from $self->{hostname}" . $clientObj->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: validate_software()
# Input: -nodeObject
# Tasks: Verify if the node is having required software
# Return: TRUE/FALSE
###########################################################################
sub validate_software {
    my ($self,@args) = @_;

    my ($cmd, $clientObj, @buf);

    # Check for serviceguard software
    $cmd = "/usr/sbin/swlist |grep T1905";
    $clientObj = $self->{sanclientObj};
    @buf = $clientObj->execute($cmd);

    if ($clientObj->error()) {
        $self->_set_err("MC ServiceGuard not found." . $clientObj->errmsg());
        return $ClientUtil::FALSE;
    } else {
        &SanUtil::nate_log('-comment',"MC ServiceGuard Installed");
    }

    # Check for serviceguard NFS toolkit
    $cmd = "/usr/sbin/swlist |grep B5140";
    $clientObj = $self->{sanclientObj};
    @buf = $clientObj->execute($cmd, -hide=>"1");

    if ($clientObj->error()) {
        $self->_set_err("NFS Toolkit not found.");
        return $ClientUtil::FALSE;
    } else {
        &SanUtil::nate_log('-comment',"Service Guard NFS Toolkit Installed");
    }

    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: get_sanclient_obj()
# Input: -self
# Return: SanClient Object of -host
###########################################################################
sub get_sanclient_obj {
    my ($self) = @_;

    return $self->{sanclientObj};
}

###########################################################################
# Subroutine Name: enable_cluster_on_boot()
# Input: -nodeObject
# Tasks: Sets AUTOSTART_CMCLD=1 in "/etc/rc.config.d/cmcluster"
#        so that the cluster is automatically started at boot time
###########################################################################
sub enable_cluster_on_boot {
    my ($self) = @_;

    my ($cmd, $clientObj, @buf);

    # Validate MCSG installation
    if (!$self->validate_software()) {
        $self->_set_err("Cluster Software is not installed.");
        return $ClientUtil::FALSE;
    }

    # Make changes to cmcluster file to enable cluster autostart (AUTOSTART_CMCLD=1)
    $cmd = "cat /etc/rc.config.d/cmcluster |sed 's/AUTOSTART_CMCLD=[01]/AUTOSTART_CMCLD=1/' > /etc/rc.config.d/cmcluster.temp";
    $clientObj = $self->{sanclientObj};
    @buf = $clientObj->execute($cmd, -hide=>"1");
    if ($clientObj->error()) {
        $self->_set_err("Unable to read cmcluster config file");
        return $ClientUtil::FALSE;
    }

    $cmd = "mv /etc/rc.config.d/cmcluster.temp /etc/rc.config.d/cmcluster";
    $clientObj->execute($cmd ,-hide=>"1");
    if ($clientObj->error) {
        $self->_set_err("Unable to write cmcluster config file");
        return $ClientUtil::FALSE;
    }

    # Cluster autostart successfully enabled
    &SanUtil::nate_log('-comment',"Cluster Enabled at boot");
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: disable_cluster_on_boot()
# Input: -nodeObject
# Tasks: Sets AUTOSTART_CMCLD=0 in "/etc/rc.config.d/cmcluster"
#        so that cluster autostart is disabled at boot
###########################################################################
sub disable_cluster_on_boot {
    my ($self) = @_;

    my ($cmd, $clientObj, @buf);

    # Validate MCSG installation
    if (!$self->validate_software()) {
        $self->_set_err("Cluster Software is not installed.");
        return $ClientUtil::FALSE;
    }

    # Make changes to cmcluster file to disable cluster autostart(AUTOSTART_CMCLD=0)
    $cmd = "cat /etc/rc.config.d/cmcluster |sed 's/AUTOSTART_CMCLD=[01]/AUTOSTART_CMCLD=0/' > /etc/rc.config.d/cmcluster.temp";
    $clientObj = $self->{sanclientObj};
    @buf = $clientObj->execute($cmd,-hide=>"1");
    if ($clientObj->error) {
        $self->_set_err("Unable to read cmcluster config file");
        return $ClientUtil::FALSE;
    }

    $cmd = "mv /etc/rc.config.d/cmcluster.temp /etc/rc.config.d/cmcluster";
    $clientObj->execute($cmd ,-hide => "1");
    if ($clientObj->error) {
        $self->_set_err("Unable to write cmcluster config file");
        return $ClientUtil::FALSE;
    }

    # Cluster autostart successfully disabled
    &SanUtil::nate_log('-comment',"Cluster Disabled at boot");
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: halt()
# Input: -nodeObject
# Tasks: Checks if the packages on the node are halted
#        Halts the node
# Commands used: -cmhaltnode -t
#                -cmhaltnode
###########################################################################
sub halt {
    my ($self) = @_;

    my (@output, $cmd, $flag);


    @output = ();
    $flag = 0;

    # Tests if the node is running any packages/dependencies
    # cmhaltnode -t makes an assessment of the package placement without affecting the current state of the nodes or packages
    $cmd = "/usr/sbin/cmhaltnode -t $self->{hostname}";
    @output = $self->{sanclientObj}->execute($cmd, -hide => "1");
    if ($self->{sanclientObj}->error) {
        $self->_set_err($self->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Check cmhaltnode -t output to validate that package(s) are running on the node
    foreach my $string (@output) {
        if ($string =~ /package/i) {
            $flag = 1;
            last;
        }
    }

    if ($flag == 1) {
        $self->_set_err("Cannot halt node $self->{hostname}, package(s) are running on it.");
        return $ClientUtil::FALSE;
    }

    # No packages/dependencies, hence the node can be stopped.
    $cmd = "/usr/sbin/cmhaltnode $self->{hostname}";
    $self->{sanclientObj}->execute($cmd, -hide => "1");
    if ($self->{sanclientObj}->error) {
        $self->_set_err($self->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }
    &SanUtil::nate_log('-comment',"Node $self->{hostname} halted successfully.");
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: reboot()
# Input: -self
# Tasks: Reboots the node
# Commands used: -reboot
###########################################################################
sub reboot {
    my ($self) = @_;

    my $cmd = "reboot";
    $self->{sanclientObj}->execute($cmd, -hide=>"1");
    &SanUtil::nate_log('-comment',"Node $self->{hostname} has been Warm booted.");
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: node_status()
# Input: -node Object
# Tasks: Parse the output of cmviewcl to check about the status
# Commands used: -cmviewcl
# Return: Array of the format "Status, <pkg1>, <pkg2>, <pkg3>, ..."
###########################################################################
sub node_status {
    my ($self) = @_;

    my (@output, $cmd, $ctr);

    @output = ();

    $cmd = "/usr/sbin/cmviewcl | grep $self->{hostname}";
    my @tmp_arr = $self->{sanclientObj}->execute($cmd, -hide => "1");
    if ($self->{sanclientObj}->error()) {
        $self->_set_err($self->{sanclientObj}->errmsg());
        return @output;
    }

    @tmp_arr = grep (/\S/, @tmp_arr);
    push @output, $tmp_arr[2];
    # Each of the running packages in node $self->{hostname}
    for ($ctr = 3; $ctr < $#tmp_arr; $ctr += 5) {
         push @output, $tmp_arr[$ctr];
    }

    # Return list or scalar depending on context
    return wantarray() ? @output : \@output;
}

#######################################################################
######################### PRIVATE METHODS #############################
#######################################################################

###########################################################################
# Subroutine Name: _new()
# Input: -parent(SanClientCluster Object), -host
# Output: Returns SanCluster object
# Tasks: Creates new empty Node Object
#        Creates SanClient object for -host
#        Sets attributes of cluster on SanCluster object (-cluster_name)
#        Sets attribute on Node object that -host is the primary node
###########################################################################
sub _new {
    my ($class, %args) = @_;

    my( $clientObj,$protocol,$sanClusterObj);

    my $returnhash = {};
    my $self = {};

    # Make self an object of the passed class
    bless ($self, $class);

    $self->{hostname}       = undef;
    $self->{sanclientObj}   = {};
    $self->{isprimary}      = undef;
    $self->{parent}         = {};
    $self->{cluster_subnet} = undef;
    $self->{mp_type}        = undef;
    $self->{alua_support}   = undef;
    $self->{subnets}        =  {};

    # Check if required parameters are provided for the host
    foreach (keys %args) {
        if (/-?host$/) {
            if ( not length($args{$_})) {
                $self->_set_err("Invalid value for -host: $args{$_}");
                return $self;
            } else {
                $self->{hostname} = $args{$_};
            }
            next;
        } elsif (/^-?protocol$/) {
            if (not length($args{$_})) {
                $self->_set_err("Invalid value for -protocol: $args{$_}");
                return $self;
            } else {
                $protocol = $args{$_};
            }
            next;
        } elsif (/^-?parent$/) {
            if (not length($args{$_})) {
                $self->_set_err("Invalid value for -parent: $args{$_}");
                return $self;
            } else {
                $self->{parent} = $args{$_};
            }
            next;
        } elsif (/^-?primary$/) {
            if ($args{$_} =~ /[01]/) {
                $self->{isprimary} = $args{$_};
            } else {
                $self->_set_err("Invalid value for -primary: $args{$_}, Given ");
                return $self;
            }
            next;
        }
    }

    # Create SanClient object
    $clientObj = SanClient->new(-host         => $self->{hostname},
                                -protocol     => $protocol,
                                -conn_type    => "THARNHOST",
                                -timeout      => 300,
                                -debug_level  => 0,
                                -mp_type      => $args{-mp_type},
                                -alua_support => $args{-alua_support});

    if ($clientObj->error()) {
        $self->_set_err($clientObj->errmsg);
        return $self;
    }

    # Now that we have created the clientObj pass the required options
    if (defined $args{-vm_hpux_revision} && defined $args{-vm_hpux_extent_size}) {
        $clientObj->options(-vm_hpux_revision    => $args{-vm_hpux_revision},
                            -vm_hpux_extent_size => $args{-vm_hpux_extent_size});
    }

    # Discover
    $clientObj->discover();

    # Assignment
    $self->{sanclientObj} = $clientObj;
    $self->{cluster_subnet} = $args{-cluster_subnet};
    $returnhash = $self->_parse_cmquerycl($self);
    my @hs = split (/:/, $self->{cluster_subnet});
    my $cnt = 0;
    foreach my $ip (@hs) {
        if (defined $returnhash->{$returnhash->{$ip} . "_HBIP"}) {
            $cnt+=1;
        }
        $self->{subnets}->{$ip} = $returnhash->{$returnhash->{$ip} . "_HBIP"};
    }

    # Check if enough (atleast 2) LAN networks are available
    if(scalar(@hs) != $cnt && $cnt < 2){
       $self->_set_err("System $self->{hostname} should be connected to atleast two diferent subnets.");
       return $self;
    }

    # Validate the node
    if (not $self->validate()) {
        $self->_set_err("Failed to validate node $self->{hostname}. " . $self->errmsg());
        return $self;
    }

    # Start the NFS server in the node.
    my $cmd = "/sbin/init.d/nfs.server start";
    $self->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($self->{sanclientObj}->error) {
        $self->_set_err("Unable to start nfs server on node $self->{hostname}." . $self->{sanclientObj}->errmsg());
        return $self;
    }
    return $self;
}

###########################################################################
# Subroutine Name: _parse_cmquerycl()
# Input: -nodeObject
# Tasks: Parse the output of cmquerycl to obtain info about the node
# Commands used: -cmquerycl -l net -n <nodename>
# Return: Hash containing the info about the host
###########################################################################
sub _parse_cmquerycl {
    my ($self) = @_;

    my $cmd = "";
    my @hs = split (/:/, $self->{cluster_subnet});
    my $returnhash = {};

    # Run cmquerycl on the passed node object
    $cmd = "/usr/sbin/cmquerycl -f line -l net -n $self->{hostname}";
    my @output = $self->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($self->{sanclientObj}->error) {
        $self->_set_err($self->{sanclientObj}->errmsg());
        return {};
    }

    # Parse the output to get node specific information
    foreach my $line (@output) {
        if ($line=~ /hostname_address_family/) {
            my @arr = split (/=/, $line);
            $arr[1] =~ s/\n//g;
            $returnhash->{HOSTNAME_ADDRESS_FAMILY} = $arr[1];
        }
        if ($line=~ /auto_start_timeout/) {
            my @arr = split (/=/, $line);
            $arr[1] =~ s/\n//g;
            $returnhash->{AUTO_START_TIMEOUT} = $arr[1];
        }
        if ($line=~ /node_timeout/) {
            my @arr = split (/=/, $line);
            $arr[1] =~ s/\n//g;
            $returnhash->{MEMBER_TIMEOUT} = $arr[1];
        }
        if ($line=~ /network_polling_interval/) {
            my @arr = split (/=/, $line);
            $arr[1] =~ s/\n//g;
            $returnhash->{NETWORK_POLLING_INTERVAL} = $arr[1];
        }
        if ($line=~ /network_failure_detection/) {
            my @arr = split (/=/, $line);
            $arr[1] =~ s/\n//g;
            $returnhash->{NETWORK_FAILURE_DETECTION} = $arr[1];
        }
        if ($line=~ /network_auto_failback/) {
            my @arr = split (/=/, $line);
            $arr[1] =~ s/\n//g;
            $returnhash->{NETWORK_AUTO_FAILBACK} = $arr[1];
        }

        # More information: LAN1 IP, LAN2 IP, LAN1 IPM, LAN2 IPM, LAN1 PTGT, LAN2 PTGT
        foreach my $ctr (0 .. scalar(@hs)-1) {
            if (grep (/subnet=$hs[$ctr]/, $line)) {
                my @arr = split(/\W|:/, $line);
                $returnhash->{$hs[$ctr]} = $arr[3];
                my $hbip = $arr[5];
                for my $i (6 .. 8) {
                    $hbip .= ".".$arr[$i];
                }
                $returnhash->{$arr[3]."_HBIP"} = $hbip;
            }

            if (defined($returnhash->{$hs[$ctr]}) && grep (/subnet:$hs[$ctr]/ && /ip_monitor=on/, $line)) {
                $returnhash->{$returnhash->{$hs[$ctr]}."_IPM"} = "ON";
            }

            if (defined($returnhash->{$hs[$ctr]}) && grep (/subnet:$hs[$ctr]/ && /polling_target=/, $line)) {
                my @arr = split (/=/, $line);
                $returnhash->{$returnhash->{$hs[$ctr]}."_PTGT"} = $arr[1];
            }
        }
    }

    return $returnhash;
}

1;
