# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Creates HPUX - MC Service Guard SD Object.
##
## @author  gowtham@netapp.com
## @status
##/* $Id$ */

package San::SanClientCluster::HPUX::MCSG::SharedDiskGroup;

use strict;
use Error;
use Data::Dumper;
use San::SanLog;
use San::SanClient;
use San::SanClient::Lvm;
use San::SanClient::ClientUtil;

# Define global variables
use vars qw(
      @EXPORT
      @ISA
      );

@ISA = qw( SanLog Error );

###########################################################################
# Subroutine Name: add_device()
# Input: Hash containing host_name as key and list of devices as value
###########################################################################
sub add_device {
    my ($self, %args) = @_;

    # No-Op for HP-UX
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: create_dg()
# Input: -devices_list,-num_lvs, -lv_layout, -vg_version, -lv_size
# Tasks: Create a LVM Object
#        Create a SanClient Object
#        Create a VG using SanClient Object
###########################################################################
sub create_dg {
    my ($self, %args) = @_;

    my ($cmd, $major, $devs, $num_lvs, $lv_layout, $lv_size);

    $major = $args{-hpux_revision};
    $devs = $args{-devices};

    # Use default LV size if not provided
    if (defined $args{-lv_size}) {
        $lv_size = $args{-lv_size};
    } else {
        $lv_size = '100m';
    }

    $lv_layout = $args{-lv_layout};
    $num_lvs = $self->{parent}->{num_lvs};
    &SanUtil::nate_log('-comment', "Value of num_lvs is : " . $num_lvs);

    # Try to get the primary node object
    my $primary_node = $self->{parent}->get_primary_node();
    unless (defined $primary_node) {
      $self->_set_err("Error in Obtaining Primary node.");
      return $self;
    }

    my $clientObj = $primary_node->get_sanclient_obj();
    if ($primary_node->error) {
        $self->_set_err($primary_node->errmsg());
        return $self;
    }

    my $lvm_default = $clientObj->get_lvm_default();
    # Create a primary_lvm object
    my $lvm_obj = $clientObj->create_lvm(lvm => $lvm_default);
    if ($clientObj->error) {
        $self->_set_err($clientObj->errmsg());
        return $self;
    }

    # Initialize devices
    $lvm_obj->initialize(-devices => $devs);
    if ($lvm_obj->error) {
        $self->_set_err("Error: Failed to init devices". $lvm_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Create a VG in Primary node
    my $dgObj = $lvm_obj->create_vg(-name=>$self->{svgname},-devices=>$devs);
    if ($lvm_obj->error) {
        $self->_set_err($lvm_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Create "num_lvs" logical volumes on the VG just created and layout filesystem
    foreach my $index (1 .. $num_lvs) {
        my $lvol_Obj = $dgObj->create_vol(-name   => "lvol$index",
                                          -size   => $lv_size,
                                          -layout => $lv_layout);
        if ($dgObj->error) {
            $self->_set_err("Failed to create a logical volume on $self->{svgname} : ". $dgObj->errmsg());
            return $self;
        }
        push @{$self->{lvollist}}, $lvol_Obj ;
        # Creating filesystem on created logical volume(s)
        if (not $lvol_Obj->create_fs(-type=>"vxfs")) {
            $self->_set_err("Failed to create a File System :" . $lvol_Obj->errmsg());
            return $self;
        }
    }

    # Create a mapfile using export and disable the vg
    $cmd = "/usr/sbin/vgchange -a n $self->{svgname}";
    $primary_node->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($primary_node->{sanclientObj}->error) {
        $self->_set_err($primary_node->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/sbin/vgexport -p -s -m /tmp/mapfile.$self->{svgname} $self->{svgname}";
    $primary_node->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($primary_node->{sanclientObj}->error) {
        $self->_set_err($primary_node->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    if ($major =~ /2\./) {
        $major = 128;
    } else {
        $major = 64;
    }

    # Call _propagate at last.
    if ($self->_propagate($major, $primary_node)) {
        return $ClientUtil::TRUE;
    }

    return $ClientUtil::FALSE;
}

###########################################################################
# Subroutine Name: get_devices()
# Input: -node_name
# Tasks: Provides the list of devices associated with the node
###########################################################################
sub get_devices {
    my ($self, $nodename)  = @_;

    return (wantarray()) ? @{$self->{parent}->{devlist}->{$nodename}} : $self->{parent}->{devlist}->{$nodename};
}

###########################################################################
# Subroutine Name: get_lvm_object()
# Input: -self
# Tasks: Collects the LVM object associated with the node
# Return: LVM Object
###########################################################################
sub get_lvm_object {
    my ($self) = @_;

    # No-Op for HP-UX
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: create_rg()
# Input: -node Object
# Tasks: Destroy devices associated with the node
# Commands used: None
###########################################################################
sub create_rg {
    my ($self, %args) = @_;

    my ($rgObj);

    # Create resource group object
    $rgObj = SanClientCluster::HPUX::MCSG::SharedDiskGroup::Resource->_new(-parent => $self, %args);

    if ($rgObj->error) {
        $self->_set_err($rgObj->errmsg());
	return $rgObj;
    }

    push (@{$self->{rgs}}, $rgObj);

    return $rgObj;
}

###########################################################################
# Subroutine Name: destroy()
# Input: -node Object
# Tasks: Destroy devices associated with the node
# Commands used: -vgexport
###########################################################################
sub destroy {
    my ($self) = @_;

    my ($node_name, $nodeObj, $clientObj, $lvm_default, $lvm_obj, $cmd);

    # Remove the VG(s) from all nodes using vgexport.
    foreach my $node (values %{$self->{parent}->{nodelist}}) {
        $cmd = "/usr/sbin/vgexport /dev/$self->{svgname}";
        $node->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($node->{sanclientObj}->error) {
            $self->_set_err($node->{sanclientObj}->errmsg());
            return $ClientUtil::FALSE;
        }
    }
    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: remove_devices()
# Tasks: Removes the associated DG(s) from the datastructure
###########################################################################
sub remove_devices {
    my ($self) = @_;

    # No-op for HPUX
    return $ClientUtil::TRUE;
}

###########################################################################
############################ PRIVATE METHODS ##############################
###########################################################################

###########################################################################
# Subroutine Name: _new()
# Input: -SanClientCluster Object
# Tasks: Creates an empty SharedDiskGroup Object
###########################################################################
sub _new {
    my ($class, %args) = @_;

    my $self = {};

    # Make self an object of the passed class
    bless ($self, $class);

    my ($cmd, $status, $parent, $type, $primary_node, @output);
    $self->{svgname}    = undef;
    $self->{parent}     = {};
    $self->{sandevlist} = ();
    $self->{resources}  = ();
    $self->{lvollist}   = ();

    # Validate input arguments
    foreach (keys %args) {
        if (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return $self;
            }
            next;
        } elsif (/-?clusterObj/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clusterObj'");
                return $self;
            }
            next;
        }
    }

    unless (defined $type and length $type) {
      $self->_set_err("-type is a required argument");
      return $self;
    }

    unless (defined $parent and length $parent) {
      $self->_set_err("-clusterObj is a required argument");
      return $self;
    }

    # Create and return empty SharedDiskGroup object
    $self->{svgname} = $args{-name};
    $self->{parent}  = $args{-parent};
    return $self;
}

###########################################################################
# Subroutine Name: _propagate()
# Tasks: Performs a vg export on primary node
#        Performs impoprting of vg on all other nodes
# Commands used: -vgexport
#                -vgimport
#                -vgchange
#                -vgcfgbackup
#                -mknod
#                -mkdir
###########################################################################
sub _propagate() {
    my ($self, $major, $primary_node) = @_;

    my ($cmd, @output, $status, $highest, $highest_num, $entry, $minor, $len);

    foreach my $node (values %{$self->{parent}->{nodelist}}) {
        if ($primary_node->{hostname} ne $node->{hostname}) {
            $cmd = "/usr/bin/rcp /tmp/mapfile.$self->{svgname} $node->{hostname}:/tmp";
            $primary_node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($primary_node->{sanclientObj}->error) {
                $self->_set_err($primary_node->{sanclientObj}->errmsg());
                return $ClientUtil::FALSE;
            }

            $cmd = "/usr/bin/mkdir /dev/$self->{svgname}";
            $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error) {
                $self->_set_err($node->{sanclientObj}->errmsg());
                return $ClientUtil::FALSE;
            }

            # Obtain the next highest minor number with respect to the given major number
            $cmd = "ls -l /dev/*/group";
            @output = $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error) {
                $self->_set_err($node->{sanclientObj}->errmsg());
                return $ClientUtil::FALSE;
            }

            if (grep(/not found/i, @output)) {
                # No VG groups are created on this client, start with the first minor number
                $highest = "000000";
            }

            foreach $entry (@output) {
                $minor = (split(/\s+/, $entry))[5];
                $minor =~ s/0x//;
                if (not defined($highest) or $minor > $highest) {
                    $highest = $minor;
                }
            }

            # Get the highest minor number and increment.
            $highest_num = substr($highest, 1, 1);
            $highest_num++;

            # Get the size of the string representing the new value
            $len = length($highest_num);

            # Run the mknod command with the newly aquired minor device number
            if ($len > 1) {
                $cmd = "/usr/sbin/mknod /dev/$self->{svgname}/group c $major 0x0" . $highest_num . "000";
            } else {
                $cmd = "/usr/sbin/mknod /dev/$self->{svgname}/group c $major 0x0" . $highest_num . "0000";
            }
            @output = $node->{sanclientObj}->execute($cmd, -hide=>"1");;
            if ($node->{sanclientObj}->error) {
                $self->_set_err($node->{sanclientObj}->errmsg());
                 return $ClientUtil::FALSE;
            }

            # Import the VG
            my $version = $node->{sanclientObj}->execute("uname -r", -hide => 1, -_hostcmd => 1)->[0];
            # vgimport with HPUX 11iv2 / HPUX 11iv3
            if ($version =~ /B.11.23/) {
                $cmd = "/usr/sbin/vgimport -s -m /tmp/mapfile.$self->{svgname} /dev/$self->{svgname}";
            } else {
                $cmd = "/usr/sbin/vgimport -N -s -m /tmp/mapfile.$self->{svgname} /dev/$self->{svgname}";
            }
            $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error) {
                $self->_set_err($node->{sanclientObj}->errmsg());
                return $ClientUtil::FALSE;
            }

            # For HPUX 11iv3, add the required paths
            if ($version =~ /B.11.23/) {
                $cmd = "export PATH=\$PATH:/usr/sbin; $node->{sanclientObj}->{huobj}->{hu_bin_path}/ntap_config_paths";
                $node->{sanclientObj}->execute($cmd, -hide=>"1");
                if ($node->{sanclientObj}->error) {
                    $self->_set_err($node->{sanclientObj}->errmsg());
                    return $ClientUtil::FALSE;
                }
            }

            # Activate the VG, do vgcfgbackup and then deactivate VG
            $cmd = "/usr/sbin/vgchange -a y $self->{svgname}";
            $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error) {
                $self->_set_err($node->{sanclientObj}->errmsg());
                return $ClientUtil::FALSE;
            }

            $cmd = "/usr/sbin/vgcfgbackup /dev/$self->{svgname}";
            $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($primary_node->{sanclientObj}->error) {
                $self->_set_err($primary_node->{sanclientObj}->errmsg());
                return $ClientUtil::FALSE;
            }

            $cmd = "/usr/sbin/vgchange -a n $self->{svgname}";
            $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error) {
                $self->_set_err($node->{sanclientObj}->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    }

    # Make the VG active
    $cmd = "/usr/sbin/vgchange -a y $self->{svgname}";
    $primary_node->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($primary_node->{sanclientObj}->error) {
        $self->_set_err($primary_node->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

1;
