# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Creates HPUX - MC Service Guard Cluster Object.
##
## @author  gowtham@netapp.com
## @status
##
#

package San::SanClientCluster::HPUX::MCSG::SharedDiskGroup::Resource;

use strict;
use Error;
use San::SanLog;
use San::SanClient;
use San::SanClient::ClientUtil;

use vars qw(
      @EXPORT
      @ISA
      );

@ISA = qw( SanLog Error SanClient);

############################################################################
# Subroutine Function: create()
# Input: VG, Primary node and IP address of the nodes.
# Tasks: Calls the appropriate create function based on HPUX 11i v2/v3
############################################################################
sub create {
    my ($self, %args) = @_;

    my ($pnode, $cmd, $num_lvs);

    $num_lvs = $self->{parent}->{parent}->{num_lvs};
    $pnode = $self->{parent}->{parent}->get_primary_node();

    # Create Modular/Legacy configuration files for package
    # depending on HPUX version
    my $version = $pnode->{sanclientObj}->execute("uname -r", -hide => 1, -_hostcmd => 1)->[0];
    if ($version =~ /B.11.31/){
        if (not $self->create_modular(%args)) {
            $self->_set_err("Error while creating Legacy configuration files for package nate_$self->{pkgname}." . $self->errmsg());
            return $self;
        }
    } else {
        if (not $self->create_legacy(%args)) {
            $self->_set_err("Error while creating Modular configuration files for package nate_$self->{pkgname}." . $self->errmsg());
            return $self;
        }
    }

    # Now that the file(s) are written onto primary node, propogate changes to all other nodes,
    # Change permissions and create the mount directories

    # Copying the folder to all other hosts
    foreach my $node (values %{$self->{parent}->{parent}->{nodelist}}) {
        if ($pnode->{hostname} eq $node->{hostname}) {
            next;
        }

        $cmd = "rcp -r /etc/cmcluster/nate_$self->{pkgname} $node->{hostname}:/etc/cmcluster";

        $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($pnode->{sanclientObj}->error()) {
            $self->_set_err($pnode->{sanclientObj}->errmsg());
            return $self;
        }
    }

    foreach my $node (values %{$self->{parent}->{parent}->{nodelist}}) {
        # Change the permissions of the package directory
        $cmd  = "chmod -R 744 /etc/cmcluster/nate_$self->{pkgname}";
        $node->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($node->{sanclientObj}->error()) {
            $self->_set_err($node->{sanclientObj}->errmsg());
            return $self;
        }

        # Create the mount directories for packages
        foreach my $index ( 1 .. $num_lvs) {
            $cmd = "mkdir /nate_$self->{pkgname}0$index /nate_$self->{pkgname}0$index" . "nfs";
            $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error()) {
                $self->_set_err($node->{sanclientObj}->errmsg());
                return $self;
            }
        }
    }

    # Deactivate the VG, since it'll be activated on running the packages.
    $cmd = "/usr/sbin/vgchange -a n /dev/$self->{parent}->{svgname}";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $self;
    }
    $self->sanlog( msg => "A new package nate_$self->{pkgname} was successsfully created." );

    return $ClientUtil::TRUE;
}

############################################################################
# Subroutine Function: create_legacy()
# Input: VG, primary node and IP address of the nodes.
# Tasks: Generates package conf and cntl files and uses the input
#        parameters to edit the files.(11i v2)
############################################################################
sub create_legacy {
    my ($self, %args) = @_;

    my ($num, $num_lvs, @hs, $cmd, $pnode);

    $num = $args{-num};
    $num_lvs = $self->{parent}->{parent}->{num_lvs};
    @hs = split (/:/, $self->{parent}->{parent}->{cluster_subnet});
    $pnode = $self->{parent}->{parent}->get_primary_node();

    # Open files in  primary node
    $cmd = "/usr/bin/mkdir /etc/cmcluster/nate_$self->{pkgname}";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/bin/touch /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "> /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/bin/touch /etc/cmcluster/nate_$self->{pkgname}/nfs.cntl";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "> /etc/cmcluster/nate_$self->{pkgname}/nfs.cntl";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/bin/touch /etc/cmcluster/nate_$self->{pkgname}/hanfs.nate_$self->{pkgname}.sh";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "> /etc/cmcluster/nate_$self->{pkgname}/hanfs.nate_$self->{pkgname}.sh";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Copy files from nfs folder under /opt and populate in /etc/cmcluster/<package_name>
    $cmd = "cp /opt/cmcluster/nfs/hanfs.sh /etc/cmcluster/nate_$self->{pkgname}/temp_hanfs.sh";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "cp /opt/cmcluster/nfs/nfs.mon /etc/cmcluster/nate_$self->{pkgname}/nfs.mon";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "cp /opt/cmcluster/nfs/nfs.flm /etc/cmcluster/nate_$self->{pkgname}/nfs.flm";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Create file 1 (nfs.conf)
    my @pkg = ();
    push (@pkg, 'PACKAGE_NAME        nate_' . $self->{pkgname});
    push (@pkg, 'PACKAGE_TYPE        FAILOVER');
    push (@pkg, 'FAILOVER_POLICY        CONFIGURED_NODE');
    push (@pkg, 'FAILBACK_POLICY        MANUAL');
    my $first_node_name = $self->{parent}->{svgname};
    $first_node_name =~  s/n_vg_//;
    push (@pkg, 'NODE_NAME        ' . $first_node_name);
    foreach my $node (values %{$self->{parent}->{parent}->{nodelist}}) {
        if ($node->{hostname} eq $first_node_name) {
            next;
        }
        push (@pkg, 'NODE_NAME        ' . $node->{hostname});
    }
    push (@pkg, 'AUTO_RUN        YES');
    push (@pkg, 'LOCAL_LAN_FAILOVER_ALLOWED        YES');
    push (@pkg, 'NODE_FAIL_FAST_ENABLED        NO');
    push (@pkg, 'RUN_SCRIPT    /etc/cmcluster/nate_' . $self->{pkgname} . '/nfs.cntl');
    push (@pkg, 'RUN_SCRIPT_TIMEOUT        NO_TIMEOUT');
    push (@pkg, 'HALT_SCRIPT     /etc/cmcluster/nate_' . $self->{pkgname} . '/nfs.cntl');
    push (@pkg, 'HALT_SCRIPT_TIMEOUT             NO_TIMEOUT');
    push (@pkg, 'SERVICE_NAME  nfs' . $num . '.monitor');
    push (@pkg, 'SUBNET        ' . $hs[0]);

    # Write file 1 (nfs.conf) into the primary node
    foreach my $val (@pkg) {
        $cmd = "echo $val >> /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
        $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($pnode->{sanclientObj}->error) {
            $self->_set_err("Error while writing into file nfs.conf ." . $pnode->{sanclientObj}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    # Create file 2 (nfs.cntl)
    my @cntl = ();
    push (@cntl, '. \${SGCONFFILE:=/etc/cmcluster.conf}');
    push (@cntl, 'PATH=\$SGSBIN:/usr/bin:/usr/sbin:/etc:/bin');
    push (@cntl, 'VGCHANGE=\"vgchange -a e\"');
    push (@cntl, 'VG[0]=\"' . $self->{parent}->{svgname} . '\"');
    push (@cntl, 'DEACTIVATION_RETRY_COUNT=0');
    push (@cntl, 'KILL_PROCESSES_ACCESSING_RAW_DEVICES=\"NO\"');
    foreach my $index (0 .. $num_lvs-1) {
        push (@cntl, 'LV[' . $index . ']=/dev/' . $self->{parent}->{svgname} . '/lvol' . ($index+1) );
        push (@cntl, 'FS[' . $index . ']=/nate_' . $self->{pkgname} . '0' . ($index+1) );
        push (@cntl, 'FS_MOUNT_OPT[' . $index . ']=\"-o rw\"');
        push (@cntl, 'FS_UMOUNT_OPT[' . $index . ']=\"\"');
        push (@cntl, 'FS_FSCK_OPT[' . $index . ']=\"\"');
        push (@cntl, 'FS_TYPE[' . $index . ']=\"vxfs\"');
    }
    push (@cntl, 'VXVOL=\"vxvol -g \\$DiskGroup startall\"');
    push (@cntl, 'FS_UMOUNT_COUNT=1');
    push (@cntl, 'FS_MOUNT_RETRY_COUNT=0');
    push (@cntl, 'CONCURRENT_VGCHANGE_OPERATIONS=1');
    push (@cntl, 'ENABLE_THREADED_VGCHANGE=1');
    push (@cntl, 'CONCURRENT_FSCK_OPERATIONS=1');
    push (@cntl, 'CONCURRENT_MOUNT_AND_UMOUNT_OPERATIONS=1');
    push (@cntl, 'IP[0]=\"' . $self->{pkgip} . '\"');
    push (@cntl, 'SUBNET[0]=\"' . $hs[0] . '\"');
    push (@cntl, 'HA_NFS_SCRIPT_EXTENSION=\"nate_' . $self->{pkgname} . '.sh\"');
    push (@cntl, 'log_file=\${SG_SCRIPT_LOG_FILE:-\$0.log}');
    $cmd = "/usr/sbin/cmmakepkg -s /etc/cmcluster/nate_$self->{pkgname}/temp.cntl";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Write file 2 (nfs.cntl) into primary node
    foreach my $val (@cntl) {
        $cmd = "echo $val >> /etc/cmcluster/nate_$self->{pkgname}/nfs.cntl";
        $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($pnode->{sanclientObj}->error) {
            $self->_set_err("Error while writing into file nfs.cntl . " . $pnode->{sanclientObj}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    my $endcmd = "\"# END asdfghjk\"";
    $cmd = "echo $endcmd >> /etc/cmcluster/nate_$self->{pkgname}/temp.cntl";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/bin/awk \'/START OF CUSTOMER DEFINED FUNCTIONS/,/END asdfghjk/{print}\' /etc/cmcluster/nate_$self->{pkgname}/temp.cntl >> /etc/cmcluster/nate_$self->{pkgname}/nfs.cntl";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "rm /etc/cmcluster/nate_$self->{pkgname}/temp.cntl";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Create file 3 (hanfs.sh)
    my @hanfs = ();
    push (@hanfs, 'unset UNIX95');
    push (@hanfs, 'PRE_U95=true');
    push (@hanfs, 'export PRE_U95');
    push (@hanfs, 'PATH=/sbin:/usr/bin:/usr/sbin:/etc:/bin');
    push (@hanfs, 'export PATH');
    my $all_clients = $pnode->{hostname};

    foreach my $node (values %{$self->{parent}->{parent}->{nodelist}}) {
        if ($node->{hostname} eq $pnode->{hostname}) {
            next;
        }
        $all_clients .= ":" . $node->{hostname};
    }

    foreach my $indx ( 0 .. $num_lvs-1) {
        push (@hanfs, 'XFS[' . $indx . ']=\"-o root=' . $all_clients  . ' /nate_' . $self->{pkgname} . '0' . ($indx+1) . '\"');
    }

    push (@hanfs, 'NFS_SERVICE_NAME[0]=\"nfs' . $num . '.monitor\"');
    push (@hanfs, 'NFS_SERVICE_CMD[0]=\"/etc/cmcluster/nate_' . $self->{pkgname} . '/nfs.mon\"');

    # Change NFS_FILE_LOCK_MIGRATION to 0
    push (@hanfs, 'NFS_FILE_LOCK_MIGRATION=0');
    push (@hanfs, 'NFS_FLM_SCRIPT=\"\${0%/*}/nfs.flm\"');
    push (@hanfs, 'FLM_LOCK=\"/tmp/hanfs_flm_lockdir*\"');
    push (@hanfs, 'FLM_LOCKDIR=\"/tmp/hanfs_flm_lockdir_\$BOOT_TIME\"');
    push (@hanfs, 'FLM_LOCKFILE=\"\$FLM_LOCKDIR/lock.\$\$\"');

    $cmd = "echo $endcmd >> /etc/cmcluster/nate_$self->{pkgname}/temp_hanfs.sh";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Writing file 3 (hanfs.sh) into the primary node
    foreach my $val (@hanfs) {
        $cmd = "echo $val >> /etc/cmcluster/nate_$self->{pkgname}/hanfs.nate_$self->{pkgname}.sh";
        $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($pnode->{sanclientObj}->error) {
            $self->_set_err("Error while writing into file hanfs.nate_$self->{pkgname}.sh ." . $pnode->{sanclientObj}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    $cmd = "/usr/bin/awk \'/function get_flm_lock/,/END asdfghjk/{print}\' /etc/cmcluster/nate_$self->{pkgname}/temp_hanfs.sh >> /etc/cmcluster/nate_$self->{pkgname}/hanfs.nate_$self->{pkgname}.sh";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Remove temporary file temp_hanfs.sh
    $cmd = "rm /etc/cmcluster/nate_$self->{pkgname}/temp_hanfs.sh";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

############################################################################
# Subroutine Function: create_modular()
# Input: vg, primary node and IP address of the nodes.
# Tasks: Generates package conf file (HPUX 11iv3)
############################################################################
sub create_modular {
    my($self, %args) = @_;

    my ($cmd, $num, $num_lvs, $pnode, $all_clients);

    $num = $args{-num};
    $num_lvs = $self->{parent}->{parent}->{num_lvs};
    $pnode = $self->{parent}->{parent}->get_primary_node();

    # Open files in primary node
    $cmd = "/usr/bin/mkdir /etc/cmcluster/nate_$self->{pkgname}";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/bin/touch /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "> /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # All parameters needed for creating nfs.conf are hard-coded and written
    my @pkg = ();
    push (@pkg, 'package_name        nate_' . $self->{pkgname});
    push (@pkg, 'package_description        \"Serviceguard Package\"');
    push (@pkg, 'module_name        sg/basic');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/all');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/failover');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/priority');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/dependency');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/weight');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/monitor_subnet');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/package_ip');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/resource');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/volume_group');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/filesystem');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/pev');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/external_pre');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/external');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/acp');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        nfs/hanfs');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        nfs/hanfs_export');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        sg/service');
    push (@pkg, 'module_version        1');
    push (@pkg, 'module_name        nfs/hanfs_flm');
    push (@pkg, 'module_version        1');
    push (@pkg, 'package_type        failover');

    my @nodes = ();
    my @var = ();
    my $subnet;
    my $first_node_name = $self->{parent}->{svgname};
    $first_node_name =~  s/n_vg_//;
    my $ip_subnet = (split (/:/, $self->{parent}->{parent}->{cluster_subnet}))[0];

    push (@var, 'ip_subnet        ' . $ip_subnet);
    push (@nodes, 'node_name        ' . $first_node_name);
    push (@var, 'ip_subnet_node        ' . $first_node_name);

    foreach my $node (values %{$self->{parent}->{parent}->{nodelist}}) {
        if ($node->{hostname} ne $first_node_name) {
            push (@nodes, 'node_name        ' . $node->{hostname});
            push (@var, 'ip_subnet_node        ' . $node->{hostname});
        }
    }

    push (@var, 'ip_address        ' . $self->{pkgip});
    push (@nodes, 'auto_run        yes');
    push (@nodes, 'node_fail_fast_enabled        no');
    push (@nodes, 'run_script_timeout        no_timeout');
    push (@nodes, 'halt_script_timeout        no_timeout');
    push (@nodes, 'successor_halt_timeout        no_timeout');
    push (@nodes, 'script_log_file        \$SGRUN/log/\$SG_PACKAGE.log');
    push (@nodes, 'operation_sequence        \$SGCONF/scripts/sg/external_pre.sh');
    push (@nodes, 'operation_sequence        \$SGCONF/scripts/sg/volume_group.sh');
    push (@nodes, 'operation_sequence        \$SGCONF/scripts/sg/filesystem.sh');
    push (@nodes, 'operation_sequence        \$SGCONF/scripts/nfs/hanfs_export.sh');
    push (@nodes, 'operation_sequence        \$SGCONF/scripts/sg/package_ip.sh');
    push (@nodes, 'operation_sequence        \$SGCONF/scripts/nfs/hanfs_flm.sh');
    push (@nodes, 'operation_sequence        \$SGCONF/scripts/sg/external.sh');
    push (@nodes, 'operation_sequence        \$SGCONF/scripts/sg/service.sh');
    push (@nodes, 'operation_sequence        \$SGCONF/scripts/sg/resource.sh');
    push (@nodes, 'failover_policy        configured_node');
    push (@nodes, 'failback_policy        automatic');
    push (@nodes, 'priority        no_priority');
    push (@nodes, 'local_lan_failover_allowed        yes');
    push (@var, 'service_name        nfs' . $num . '.monitor');
    push (@var, 'service_cmd        \"\$SGCONF/scripts/nfs/nfs_upcc.mon\"');
    push (@var, 'service_restart        none');
    push (@var, 'service_fail_fast_enabled        no');
    push (@var, 'service_halt_timeout        0');
    push (@var, 'concurrent_vgchange_operations        1');
    push (@var, 'enable_threaded_vgchange        1');
    push (@var, 'vgchange_cmd        \"vgchange -a e\"');
    push (@var, 'vg        ' . $self->{parent}->{svgname});
    push (@var, 'deactivation_retry_count        2');
    push (@var, 'kill_processes_accessing_raw_devices        no');
    push (@var, 'concurrent_fsck_operations        1');
    push (@var, 'concurrent_mount_and_umount_operations        1');
    push (@var, 'fs_mount_retry_count        0');
    push (@var, 'fs_umount_retry_count        1');

    foreach my $index (1 .. $num_lvs) {
        push (@var, 'fs_name        /dev/' . $self->{parent}->{svgname} . '/lvol' . $index);
        push (@var, 'fs_directory        /nate_' . $self->{pkgname} . '0' . $index);
        push (@var, 'fs_type        \"vxfs\"');
        push (@var, 'fs_mount_opt        \"-o rw\"');
        push (@var, 'fs_umount_opt        \"\"');
        push (@var, 'fs_fsck_opt        \"\"');
    }

    # Take from /etc/cmcluster/examples
    push (@var, 'external_pre_script        /etc/cmcluster/nate_' . $self->{pkgname} . '/external_script.template');
    push (@var, 'external_script        /etc/cmcluster/nate_' . $self->{pkgname} . '/external_script.template');

    $all_clients = $pnode->{hostname};
    foreach my $node (values %{$self->{parent}->{parent}->{nodelist}}) {
        if ($node->{hostname} eq $pnode->{hostname}) {
            next;
        }
        $all_clients .= ":" . $node->{hostname};
    }

    foreach my $index (1 .. $num_lvs) {
        push (@var, 'nfs/hanfs_export/XFS        \"-o root='. $all_clients .' /nate_' . $self->{pkgname} . '0' . $index .'\"');
    }

    push (@var, 'nfs/hanfs_export/FILE_LOCK_MIGRATION        0');

    foreach my $index (1 .. $num_lvs) {
        push (@var, 'nfs/hanfs_export/SNFS        ' . $self->{pkgip} . ':/nate_' . $self->{pkgname} . '0' . $index);
        push (@var, 'nfs/hanfs_export/CNFS        /nate_' . $self->{pkgname} .  '0' . $index .'nfs');
    }

    push (@var, 'nfs/hanfs_export/MONITOR_INTERVAL        10');
    push (@var, 'nfs/hanfs_export/MONITOR_LOCKD_RETRY        4');
    push (@var, 'nfs/hanfs_export/MONITOR_DAEMONS_RETRY        4');
    push (@var, 'nfs/hanfs_export/PORTMAP_RETRY        4');
    push (@var, 'nfs/hanfs_flm/FLM_HOLDING_DIR        \"/export/sm\"');
    push (@var, 'nfs/hanfs_flm/NFSV4_FLM_HOLDING_DIR        \"/export/nfsv4\"');
    push (@var, 'nfs/hanfs_flm/PROPAGATE_INTERVAL        5');
    push (@var, 'nfs/hanfs_flm/STATMON_WAITTIME        90');

    # Write the conf file in primary node first
    foreach my $val (@pkg) {
        $cmd = "echo $val >> /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
        $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($pnode->{sanclientObj}->error) {
            $self->_set_err("Error while writing into file nfs.conf ." . $pnode->{sanclientObj}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    foreach my $val (@nodes) {
        $cmd = "echo $val >> /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
        $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($pnode->{sanclientObj}->error) {
            $self->_set_err("Error while writing into file nfs.conf ." . $pnode->{sanclientObj}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    foreach my $val (@var) {
        $cmd = "echo $val >> /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
        $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($pnode->{sanclientObj}->error) {
            $self->_set_err("Error while writing into file nfs.conf ." . $pnode->{sanclientObj}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    $cmd = "cp /etc/cmcluster/examples/external_script.template /etc/cmcluster/nate_$self->{pkgname}/";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "cp /etc/cmcluster/scripts/nfs/nfs_xmnt /etc/cmcluster/nate_$self->{pkgname}/";
    $pnode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($pnode->{sanclientObj}->error()) {
        $self->_set_err($pnode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

############################################################################
# Subroutine Function: destroy()
# Input: -package name
# Tasks: Check if the package is still running. Disable if running, use
#        disable()
#        Delete the configuration files for the package
#        Delete the configuration files from the host FS
# Commands used: -cmdeleteconf
#                -rm
############################################################################
sub destroy {
    my ($self) = @_;

    my $pkgname = $self->{pkgname};
    my ($return_val, $cmd);

    if (defined $self->_isrunning($pkgname)) {
        $self->disable($pkgname);
    }

    my $preferrednode = $self->{parent}->{parent}->get_primary_node();

    # Now the package is disabled, cluster configuration can be deleted
    $cmd = "/usr/sbin/cmdeleteconf -f -c $self->{parent}->{parent}->{cluster_name} -p nate_$pkgname";
    $return_val = $preferrednode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($preferrednode->{sanclientObj}->error) {
        $self->_set_err("Error in deleting configuration files of of package nate_$pkgname." . $preferrednode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    } else {
        $self->sanlog( msg => "Package configuration files for package: nate_$pkgname were successsfully deleted." );
    }

    # Config file(s) have been deleted successfully,
    # so delete the /etc/cmcluster/nate_<package> folder as well in all the nodes of the cluster.
    foreach my $node(values %{$self->{parent}->{parent}->{nodelist}}) {
        $cmd = "rm -r /etc/cmcluster/nate_$self->{pkgname}";
        $return_val = $node->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($node->{sanclientObj}->error()) {
            $self->_set_err($node->{sanclientObj}->errmsg());
            return $self;
        }
    }

    # Call the parent VG's destroy method so as to remove the VG(s) from all nodes.
    if (not $self->{parent}->destroy()) {
        $self->_set_err("Error in Cleaning up vg $self->{parent}->{svgname} " . $self->{parent}->errmsg());
        return $self;
    }
    $self->sanlog(msg => "Vg $self->{parent}->{svgname} Cleaned up Successfully. ");

    return $ClientUtil::TRUE;
}

############################################################################
# Subroutine Function: enable()
# Input: -SanClientCluster Object.
# Tasks: Checks the generated configuration and control files
#        Run the package on the respective node
#        Enables the AUTORUN parameter on the package
# Commands used: -cmcheckconf
#                -cmapplyconf
#                -cmrunpkg
#                -cmmodpkg
############################################################################
sub enable {
    my ($self, $node_name)= @_;

    my ($cmd, $node, $num_lvs);

    $num_lvs = $self->{parent}->{parent}->{num_lvs};
    $node = $self->{parent}->{parent}->get_node(-host=>$node_name);

    # Check configuration for the created package in all the available nodes of the cluster
    $cmd = "/usr/sbin/cmcheckconf -k -v -C /etc/cmcluster/n_$self->{parent}->{parent}->{cluster_name}.config -P /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
    $node->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($node->{sanclientObj}->error()) {
        $self->_set_err("Error while Checking configuration for package nate_$self->{pkgname}" . $node->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Apply configuration to the created package
    $cmd = "/usr/sbin/cmapplyconf -k -v -C /etc/cmcluster/n_$self->{parent}->{parent}->{cluster_name}.config -f -P /etc/cmcluster/nate_$self->{pkgname}/nfs.conf";
    $node->{sanclientObj}->execute($cmd, -hide=>"1");
    $self->sanlog( msg => "Package nate_$self->{pkgname} has been successsfully configured in all the nodes in the cluster." );
    my $preferrednode = $self->{parent}->{parent}->get_primary_node();

    # Run the package
    $cmd = "/usr/sbin/cmrunpkg -v nate_$self->{pkgname}";
    $node->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($node->{sanclientObj}->error()) {
        $self->_set_err("Error while trying to run the package nate_$self->{pkgname}" . $node->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Enable autorun for the package
    $cmd = "/usr/sbin/cmmodpkg -e nate_$self->{pkgname}";
    $node->{sanclientObj}->execute($cmd, -hide=>"1");
    $self->sanlog( msg => "Package nate_$self->{pkgname} is up and running from node $node_name." );

    # Enable cross mounting of package
    foreach my $node (values %{$self->{parent}->{parent}->{nodelist}}) {
        foreach my $index ( 1 .. $num_lvs) {
            $cmd = "/sbin/mount $self->{pkgip}:/nate_$self->{pkgname}0$index /nate_$self->{pkgname}0$index" . "nfs";
            $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error()) {
                $self->_set_err($node->{sanclientObj}->errmsg());
                return $self;
            }
        }
    }

    $self->sanlog(msg => "Cross-Mounting of package nate_$self->{pkgname} done successfully.");

    return $ClientUtil::TRUE;
}

############################################################################
# Subroutine Function: disable()
# Input: -package name
# Output: returns Resource object
# Tasks: Creates new Resource Object
############################################################################
sub disable {
    my ($self, %hash) = @_;

    my $pkgname = %hash->{-pkgname};
    my ($cmd, @output);
    my $preferrednode = $self->{parent}->{parent}->get_primary_node();
    my $num_lvs = $self->{parent}->{parent}->{num_lvs};

    # Remove the cross-mount(s) first
    foreach my $node (values %{$self->{parent}->{parent}->{nodelist}}) {
        foreach my $index ( 1 .. $num_lvs) {
            $cmd = "/usr/sbin/umount /nate_$self->{pkgname}0$index" . "nfs";
            @output = $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error() or $output[0] =~ /busy/) {
                $self->_set_err("Unable to unmount the File system /nate_$self->{pkgname}0$index" . "nfs in node $node->{hostname}. " . $node->{sanclientObj}->errmsg());
                return $self;
            }

            # Delete the cross mount directories
            $cmd = "rm -r /nate_$self->{pkgname}0$index" . "nfs";
            $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error()) {
                $self->_set_err("Error while deleting the mount directory of package nate_$self->{pkgname} " . $node->{sanclientObj}->errmsg());
                return $self;
            }
        }
    }

    # Halt the package(s)
    $cmd = "/usr/sbin/cmmodpkg -d nate_$self->{pkgname}";
    $preferrednode->{sanclientObj}->execute($cmd, -hide=>"1");
    $cmd = "/usr/sbin/cmhaltpkg nate_$self->{pkgname}";
    $preferrednode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($preferrednode->{sanclientObj}->error()) {
        $self->_set_err("Error while halting/disabling autorun for package nate_$self->{pkgname}" . $preferrednode->{sanclientObj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Delete the mount directories created.
    foreach my $node (values %{$self->{parent}->{parent}->{nodelist}}) {
        foreach my $index ( 1 .. $num_lvs) {
            $cmd = "rm -r /nate_$self->{pkgname}0$index";
            $node->{sanclientObj}->execute($cmd, -hide=>"1");
            if ($node->{sanclientObj}->error()) {
                $self->_set_err("Error while deleting the mount directory of package nate_$self->{pkgname}" . $node->{sanclientObj}->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

############################################################################
# Subroutine Function: move()
# Input: -node on which the package has to run
# Tasks: Retrieve the node using get_node()
#        Halt the package
#        Run package on the specified node
#        Enable AUTORUN
# Commands used:  -cmhaltpkg
#                 -cmrunpkg -n
#                 -cmmodpkg -e
############################################################################
sub move {
    my ($self, $hostname) = @_;

    my ($hostname, $nodeObj, @return_val, $cmd);
    $nodeObj = $self->{parent}->{parent}->get_node(-host => "$hostname");

    # Disable the package on the node it is currently running
    if ($self->disable($self->{pkgname}) eq $ClientUtil::FALSE) {
        return $ClientUtil::FALSE;
    } else {
        # Package has been halted
        # Now, run on the specified node and enable autorun
        $cmd = "/usr/sbin/cmrunpkg -n $hostname nate_$self->{pkgname}";
        @return_val = $nodeObj->{sanclientObj}->execute($cmd, -hide=>"1");
        if ($nodeObj->{sanclientObj}->error()) {
            $self->_set_err("Package could not be ran on the specified node $hostname." . $nodeObj->{sanclientObj}->errmsg());
            return $ClientUtil::FALSE;
        }

        # Enable package switching for the specified package
	$cmd = "/usr/sbin/cmmodpkg -e nate_$self->{pkgname}";
        $nodeObj->{sanclientObj}->execute($cmd, -hide=>"1");
    }

    return $ClientUtil::TRUE;
}

############################################################################
# Subroutine Function: retrieve_node()
# Input: -package name
# Tasks: Returns node on which specified package is running
# Return: nodeObject if found/ $ClientUtil::FALSE otherwise
############################################################################
sub retrieve_node {
    my ($self, %hash) = @_;

    my $pkgname = %hash->{-pkgname};
    my $hostname = $self->_isrunning($pkgname);

    if (defined $hostname) {
        return $self->{parent}->{parent}->get_node(-host => $hostname);
    }
    $self->_set_err("Package $pkgname is not running."  . $self->errmsg());

    return $ClientUtil::FALSE;
}

############################################################################
########################## PRIVATE METHODS #################################
############################################################################

############################################################################
# Subroutine Function: _new()
# Input: -parent(shared disk group), -package name, -package IP
# Output: returns Resource object
# Tasks: Creates new Resource Object
############################################################################
sub _new {
    my ($class, %args) = @_;

    my $self = {};

    # Make self an object of the passed class
    bless ($self, $class);

    my ($invalid);
    $self->{pkgname}  = undef;
    $self->{pkgip}    = undef;
    $self->{vgname}   = undef;
    $self->{lvollist} = ();
    $self->{parent}   = {};

    # Package IP validation
    if (defined %args->{-pkgip}) {
        if (%args->{-pkgip} =~ /([0-9]*)\.([0-9]*)\.([0-9]*)\.([0-9]*)/) {
            my @eachone= split (/\./, %args->{-pkgip});
            $invalid = 0;
            foreach my $str(@eachone) {
                if ($str < 0 || $str > 255) {
                    $invalid = 1;
                }
                next;
            }
            if ($invalid eq 1) {
                $self->_set_err("The required parameter -packageIP is not an IP address.");
                return $self;
            } else {
                $self->{pkgip} = %args->{-pkgip};
            }
        } else {
            $self->_set_err("The required parameter -packageIP was not in correct format.");
            return $self;
        }
    } elsif (not defined %args->{-load}) {
        $self->_set_err("The required parameter -packageIP was not defined.");
        return $self;
    }

    # Populate resource object with contents from hash
    $self->{pkgname} = %args->{-name};
    $self->{vgname}  = %args->{-vgname};
    $self->{parent}  = %args->{-parent};
    &SanUtil::nate_log('-comment', "A new resource object was successsfully created." );
    $self->sanlog( msg => "A new resource object was successsfully created." );
    return $self;
}

############################################################################
# Subroutine Function: _error_check()
# Input: Output of the execute command
# Tasks:  Checks for /error/ in input
# Return: $ClientUtil::TRUE/FALSE
############################################################################
sub _error_check {
    my ($self, @return_val) = @_;

    foreach my $val (@return_val) {
        if ($val =~ /error/i ) {
            return $ClientUtil::FALSE;
        }
    }
    return $ClientUtil::TRUE;
}

############################################################################
# Subroutine Function: _isrunning()
# Input: -package name, -ARGS[0 1]
# Tasks: Checks the packages runnning on the node
# Commands used: -cmviewcl
# Return: Hostname which runs the package if ARGS is 1
# Return: TRUE/FALSE if ARGS is 0
############################################################################
sub _isrunning {
    my ($self, $pkgname) = @_;

    my ($cluster, @nodearr, $node, $nodeattr, @packagearr, $package, $packageattr, @field, $key, $val, $name_, $type, $ele_in_arr, $cmd);

    $cluster={};
    $node={};
    $nodeattr=0;
    $package={};
    $packageattr=0;

    $cmd = "/usr/sbin/cmviewcl -f line";
    my $preferrednode = $self->{parent}->{parent}->get_primary_node();
    my @output = $preferrednode->{sanclientObj}->execute($cmd, -hide=>"1");
    if ($preferrednode->{sanclientObj}->error) {
        $self->_set_err("Error while executing cmviewcl command." . $preferrednode->{sanclientObj}->errmsg());
        return $self;
    }

    foreach my $line (@output) {
        @field=split (/[\:\=\|]/, $line);
        if ($#field == 1) {
            # Its a cluster information.
            ($key,$val)=@field;
             $cluster->{$key}=$val;
        } elsif ($#field == 3) {
            # Its a package or node information
            if ($field[0] =~ /package/) {
                $packageattr++;
                ($type,$name_,$key,$val)=@field;
                $package->{$key}=$val;
                # Push into package array
                if ($packageattr == 7) {
                    push @packagearr,$package;
                    $package={};
                    $packageattr=0;
                }
            } else {
                # Its a node information
                $nodeattr++;
                ($type,$name_,$key,$val)=@field;
                $node->{$key}=$val;
                # Push into node array
                if ($nodeattr == 5) {
                    push @nodearr,$node;
                    $node={};
                    $nodeattr=0;
                }
            }
        }
    }

    foreach $ele_in_arr (@packagearr) {
        if ($ele_in_arr->{name} eq "nate_$pkgname") {
            if ($ele_in_arr->{state} eq "running") {
                return $ele_in_arr->{owner};
            }
            next;
        }
    }

    $self->_set_err("Package $pkgname is not up and running.");
}

1;
