# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Cluster
## @author  rawat@netapp.com
## @status  Private
## @pod     here

package San::SanClientCluster::Linux::RHCS;

use San::SanClientCluster::Linux::RHCS::SharedDiskGroup;
use San::SanClientCluster::Linux::RHCS::Node;
use San::SanClient::ClientUtil;
use San::SanUtil;
use San::SanLog;
use San::SanClientCluster;
use Tharn;

use XML::Simple;
use strict;
use Error;
use Data::Dumper;

use vars qw(@ISA);
@ISA = qw(Error SanLog);

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new cluster object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {    
    my ($class, %args) = @_; 

    my ($conn, $parent, $hosttype, $noDiscovery, $nodeObj);
    my ($self, $cluster_ip, $cluster_subnet, $cluster_name, $host, $lvm_type, $exists);

    $self = {}; 

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?host/) {
            if (not length($host = $args{$_})) {
                $self->_set_err("Invalid value for argument '-host'");
                return $self;
            }   
            next;
        } elsif (/-?noDiscovery/) {
            if (not length($noDiscovery = $args{$_})) {
                $self->_set_err("Invalid value for argument '-noDiscovery'");
                return $self;
            }   
            next;
        } elsif (/-?lvm_type/) {
            if (not length($lvm_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lvm_type'");
                return $self;
            }   
            next;
        } elsif (/-?cluster_ip/) {
            if (not length($cluster_ip = $args{$_})) {
                $self->_set_err("Invalid value for argument '-cluster_ip'");
                return $self;
            }   
            next;
        } elsif (/-?cluster_subnet/) {
            if (not length($cluster_subnet = $args{$_})) {
                $self->_set_err("Invalid value for argument '-cluster_subnet'");
                return $self;
            }   
            next;
        } elsif (/-?cluster_name/) {
            if (not length($cluster_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-cluster_name'");
                return $self;
            }   
            next;
        }
    }

    unless (defined $host and length $host) {
        $self->_set_err("-host is a required argument");
        return $self;
    }

    unless (defined $cluster_name and length $cluster_name) {
        $cluster_name = "nate_rhcs";
    }

    $self->{cluster_ip} = $cluster_ip;
    $self->{cluster_subnet} = $cluster_subnet;
    $self->{cluster_name} = $cluster_name;

    # both these fields are for internal use
    $self->{unique_node_id} = 1;
    $self->{config_version} = 1;

    # Save lvm type on SanClientCluster object
    $self->{lvm_type} = $lvm_type;
    $self->{noDiscovery} = $noDiscovery;

    if ($noDiscovery == $ClientUtil::FALSE) {
        # Will create SanClient object for the node
        $nodeObj = $self->create_node(%args);
        if ($self->error()) {
            return $self;
        }
        
        # Validate the node before adding it to the cluster
        $nodeObj->validate();
        if ($nodeObj->error()) {
            $self->_set_err($nodeObj->errmsg);
            return $self;
        }

        # increment the nodeID
        $self->{unique_node_id} = $self->{unique_node_id} + 1;

        push (@{$self->{hosts}}, $nodeObj);

        # Set a node to primary
        $self->set_primary_node($host);
        if ($self->error) {
            return $ClientUtil::TRUE;    
        }
        return $self;
    }

    # The /etc/cluster/cluster.conf is in XML format. We need a 
    # data structure to manipulate cluster.conf. It will create 
    # a data structure to manipulate cluster.conf. 
    $self->_build_cluster_config_data_structure();
    if ($self->error) {
        return $self;
    }

    # Will create SanClient before adding this node to the cluster
    $self->add_node(%args);
    if ($self->error) {
        return $self;
    }

    # Set the node as primary node 
    $self->set_primary_node($host);
    if ($self->error) {
        return $self;    
    }
    return $self;
}

################################################################################
# Subroutine Name: get_client_dev_lists()
# Objective:       Will return the client device list 
# Details:         See POD documentation at the end of this file
################################################################################
sub get_client_dev_lists {
    my ($self) = @_;
    my ($cmd, $target_node, $timeout, $status, @output, $vol_name, $vol_num, $resource_type);
    my $client_dev_list = "";

    my $primary_node = $self->get_primary_node();
    if ($self->error()) {
        return "";
    }

    # Delete already existing client_dev_list
    delete $self->{CDL};

    foreach my $sdgObj (@{$self->{sdg}}) {
        foreach my $rgObj (@{$sdgObj->{rgs}}) {
            $resource_type = $rgObj->{type};
            foreach my $vol_obj (@{$rgObj->{volObjs}}) {
                $vol_name = $vol_obj->get_name();
                ($vol_num) = ($vol_name =~ /nvol-(\d+)/);
                # Get the mount point and service from Cluster config data structure
                my $service = $self->{ccds}->{cluster}->[0]->{rm}->[0]->{service}->[$vol_num]->{name};
                my $mnt_pnt = $self->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type}->[$vol_num]->{mountpoint};

                # Check if the service $service is running or not on one of the nodes
                # Here is the output of the clustat
                ##-----------------------------------------------------------
                #Cluster Status for rhel5u4 @ Wed Dec  9 15:52:02 2009
                #Member Status: Quorate

                #Member Name                    ID          Status
                #------ ----                    ----        ------
                #IBMx336-200-189                1       Online, Local, rgmanager
                #IBMx336-200-190                2       Online, rgmanager

                #Service Name                           Owner (Last)                State
                #------- ----                           ----- ------                -----
                #service:service_nate_gfs_nvol-1        IBMx336-200-189             started
                #service:service_nate_gfs_nvol-2        IBMx336-200-189             started
                ##-----------------------------------------------------------
            
                sleep(5);
                $cmd = "clustat -s $service | grep service";
                @output = $self->execute(-cmd => $cmd, -target_node => $primary_node, -timeout => 30);
                my ($nodename, $service_status) = ($output[0] =~ /\s*service:\S+\s+\(*(\S+)\)*\s+(\S+)/i);
                if ($service_status =~ /disable|stopped/) {
                    if ($nodename =~ /none/) {
                        $nodename = $primary_node->{name};
                    }
                    # Enable the service on node if its disabled/stopped
                    $cmd = "clusvcadm -e $service -m $nodename";
                    $self->execute(-cmd => $cmd, -target_node => $primary_node, -timeout => 30);
                    if ($self->error) {
                        return "";
                    }
                }
                # Save the mount points for node $nodename
                push (@{$self->{CDL}->{$nodename}}, $mnt_pnt);
            }
        }
    }

    # Create the client_dev_list
    foreach my $node (@{$self->{hosts}}) {
        my $nodename = $node->{name};
        next if ($self->{CDL}->{$nodename} eq undef);
        # Add the node to client_dev_list
        $client_dev_list .= "$nodename,";
        foreach my $mnt_pnt (@{$self->{CDL}->{$nodename}}) {
             # Add each mount point to client_dev_list
             $client_dev_list .= "$mnt_pnt,";
        }
        # Turn the last , into a ; to complete CLIENT_DEV_LIST for this client
        $client_dev_list =~ s/,$//;
    }

    $self->sanlog(-msg => "Client Device List is $client_dev_list");
    return $client_dev_list;
}

################################################################################
# Subroutine Name: discover()
# Objective:       Will discover the already existing Cluster along with its
#                  nodes, shared disk groups and resources
# Details:         See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self, %args) = @_;
    my ($cmd, $target_node, $timeout, $status, @output, $cluster_file, $ccds);
    my ($nodename);
    
    # We need to set options in this way so that we get exactly what we want
    my $xs = new XML::Simple(keeproot => 1,suppressempty => 1,forcearray => 1,keyattr =>"");
    my $name = $self->{cluster_name};

    # Its not possible to read xml file directly into xml data structure. Populate xml DS with
    # some initial data
    my $initial_ccds = "<ccds alias=\"$name\" config_version=\"0\" name=\"$name\"> </ccds>";
    $self->{ccds} = $xs->XMLin("$initial_ccds");

    $cmd = "ls /etc/cluster/cluster.conf";
    @output = $self->execute(-cmd => $cmd, -timeout => 30);
    if ($self->error) {
        $self->_set_err("Cluster configuration file doesnt exists");
        return $ClientUtil::FALSE;
    } else {
        $cmd = "cat /etc/cluster/cluster.conf";
        @output = $self->execute(-cmd => $cmd, -timeout => 30);
        foreach my $line (@output) {
            $cluster_file .= $line . " ";
        }
    }

    # Read the cluster.conf file into data structure
    $self->{ccds} = $xs->XMLin("$cluster_file"); 

    # Get all the nodes 
    foreach my $node (@{$self->{ccds}->{cluster}->[0]->{clusternodes}->[0]->{clusternode}}) {
        $nodename = $node->{name};

        my $primary_node = $self->get_primary_node();
        next if ($nodename eq $primary_node->{name});

        my $nodeObj = $self->create_node(-host => $nodename, %args);
        if ($self->error) {
            return $ClientUtil::FALSE;    
        }

        # Save nodobj on Cluster Object
        push (@{$self->{hosts}}, $nodeObj);
    }

    # Create shared disk group object
    my $sdgObj = $self->create_shared_disk_group(-type => $self->{lvm_type}, -exists => 1);
    if ($self->error) {
        return $ClientUtil::FALSE;    
    }

    # Get all the resources
    my $rgObj = $sdgObj->create_rg(-parent => $sdgObj, -resource_type => "clusterfs", -exists => 1);
    if ($self->error) {
        return $ClientUtil::FALSE;    
    }

    return $ClientUtil::TRUE; 
}
#for

################################################################################
# Subroutine Name: execute()
# Objective:       will execute the cmd on target node
# Details:         See POD documentation at the end of this file
################################################################################
sub execute {
    my ($self, %args) = @_;
    my ($cmd, $target_node, $timeout, $status, @output);

    foreach (keys %args) {
        if (/-?cmd/) {
            if (not length($cmd = $args{$_})) {
                $self->_set_err("Invalid value for flag -cmd");
                return (wantarray()) ? () : [];
            }
        } elsif (/-?target_node/) {
            if (not length($target_node = $args{$_})) {
                $self->_set_err("Invalid value for flag -target_node");
                return (wantarray()) ? () : [];
            }
        } elsif (/-?timeout/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Invalid value for flag -timeout");
                return (wantarray()) ? () : [];
            }
        }
    }

    # fail out if cmd is not specified
    if (!defined $cmd) {
        $self->_set_err("-cmd is a required parameter");
        return (wantarray()) ? () : [];
    }

    # Execute the cmd on primary node if target is not passed
    if (!defined $target_node) {
        $target_node = $self->get_primary_node();
    }

    if (!defined $timeout) {
        $timeout = 600;
    }

    ($status, @output) = &ClientUtil::execute($target_node->{conn}, $cmd, $timeout);
    if ($status > 0) {
        $self->_set_err("failed to execute the cmd $cmd");
        return (wantarray()) ? @output : \@output;
    }

    return (wantarray()) ? @output : \@output;
}
#=cut
################################################################################
# Subroutine Name: _add_failoverdomain_data_structure()
# Objective:       Add failover domain info to the cluster config data structure
# Details:         See POD documentation at the end of this file
################################################################################
sub _add_failoverdomain_data_structure {
    my ($self) = @_;

    my ($cmd, $status, @output);

    # For Linux, We use single failover domain.
    # all the nodes will be added to same failover domain
    $self->{ccds}->{cluster}->[0]->{rm}->[0]->{failoverdomains}->[0]->{failoverdomain}->[0]->{name} = "nate_fd";
    $self->{ccds}->{cluster}->[0]->{rm}->[0]->{failoverdomains}->[0]->{failoverdomain}->[0]->{ordered} = "0";
    $self->{ccds}->{cluster}->[0]->{rm}->[0]->{failoverdomains}->[0]->{failoverdomain}->[0]->{restricted} = "1";
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _write_cluster_ds()
# Objective:       write the cluster data structure into the file 
#                  /etc/cluster/cluster.conf
# Details:         See POD documentation at the end of this file
################################################################################
sub _write_cluster_ds {
    my ($self) = @_;

    my ($cluster_conf, $conn, $cmd, $status, @output);

    my @nodes = $self->get_nodes();
    foreach my $node (@nodes) {
        $node->_create_cluster_conf();
        if ($node->error()) {
            $self->_set_err($node->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _build_cluster_config_data_structure()
# Objective:       Will populate the cluster data structure with basic cluster info
# Details:         See POD documentation at the end of this file
################################################################################
sub _build_cluster_config_data_structure {
    my ($self) = @_;

    # We need to set options in this way so that we get exactly what we want
    my $xs = new XML::Simple(keeproot => 1, suppressempty => 1, forcearray => 1, keyattr =>"");
    my $name = $self->{cluster_name};

    # $self->{config_version} is currently not utilize by libs anywhere
    my $config_version = $self->{config_version};
    my $initial_ccds = "<cluster alias=\"$name\" config_version=\"$config_version\" name=\"$name\"> </cluster>";

    $self->{config_version} = $self->{config_version} + 1;
    $self->{ccds} = $xs->XMLin("$initial_ccds");

    if (!defined ($self->{ccds})) {
        $self->_set_err("ccds is not initialized properly");
        return $ClientUtil::FALSE;
    }

    # add post_join_delay and post_fail_delay. These are the default values
    $self->{ccds}->{cluster}->[0]->{fence_daemon}->[0]->{post_join_delay} = 3;
    $self->{ccds}->{cluster}->[0]->{fence_daemon}->[0]->{post_fail_delay} = 0;

    $self->{ccds}->{cluster}->[0]->{cman}->[0]->{expected_votes} = 1;
    $self->{ccds}->{cluster}->[0]->{cman}->[0]->{two_node} = 1;

    # Add fence device to the cluster
    $self->{ccds}->{cluster}->[0]->{fencedevices}->[0]->{fencedevice}->[0]->{name} = "mf";
    $self->{ccds}->{cluster}->[0]->{fencedevices}->[0]->{fencedevice}->[0]->{agent} = "fence_manual";
    
    # add failover domain info to the clsuter data structure
    $self->_add_failoverdomain_data_structure();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: add_node()
# Objective:       Create a new node object
# Details:         See POD documentation at the end of this file
################################################################################
sub add_node {
    my ($self, %args) = @_;

    my ($nodeObj, $host);

    foreach (keys %args) {
        if (/-?host/) {
            if (not length($host = $args{$_})) {
                $self->_set_err("Invalid value for argument '-host'");
                return $ClientUtil::FALSE;
            }
        }
    }

    unless (defined $host and length $host) {
        $self->_set_err("-host is a required argument");
        return $ClientUtil::FALSE;
    }

    # Will create SanClient object for the node
    $nodeObj = $self->create_node(%args);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }
    
    # Validate the node before adding it to the cluster
    $nodeObj->validate();
    if ($nodeObj->error()) {
        $self->_set_err($nodeObj->errmsg);
        return $ClientUtil::FALSE;
    }

    # Add the node to the Cluster physically. Will populate the Data
    # structure for the node 
    $nodeObj->add();
    if ($nodeObj->error()) {
        $self->_set_err($nodeObj->errmsg);
        return $ClientUtil::FALSE;
    }

    # increment the nodeID
    $self->{unique_node_id} = $self->{unique_node_id} + 1;

    push (@{$self->{hosts}}, $nodeObj);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_node()
# Objective:       Return a node Obj depending upon the input
# Details:         See POD documentation at the end of this file
################################################################################
sub get_node {
    my ($self, %args) = @_;

    my ($use_host, $host, $primary, $random, $line, $node, @nodes);

    # Select a node, argument is to select a single node, either primary or a node at random
    # OR user can passin the host name through -host
    foreach (keys %args) {
        # host name is passed
        if (/-?host/) {
            if (not length($host = $args{$_})) {
                $self->_set_err("Invalid value for flag -host");
                return {};
            }
            $use_host = 1;
            last;
        # Primary node is being asked
        } elsif (/-?primary/) {
            $primary = 1;
            last;
        # Random node is being asked
        } elsif (/-?random/) {
            $random = 1;
            last;
        }
    }

    if ($use_host != 1 and $random != 1 and $primary != 1){
        $self->_set_err("One of the 3 flags need to be pass, '-host' " . 
                        "'-primary' or '-random'");
        return {};
    }

    @nodes = $self->get_nodes();
    if ($self->error()) {
        $self->_set_err("failed to get the nodes list");
        return {};
    }
    
    if ($use_host == 1) {
        foreach $node (@nodes) {
            if ($node->{name} == $host) {
                return $node;
            }
        }
    } elsif ($primary == 1) {
        return $self->get_primary_node();
    } else {
        # else return first nodeObj in the list 
        return $nodes[0];
    }

    $self->_set_err("failed to get the node");
    return {};
}

################################################################################
# Subroutine Name: get_nodes()
# Objective:       Retunrs the list of Node Objects 
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nodes {
    my ($self) = @_;

    my (@nodeObjs);

    if (defined $self->{hosts}) {
        @nodeObjs = @{$self->{hosts}};
    } else {
        return (wantarray())? () : [];
    }

    return (wantarray())? @nodeObjs : \@nodeObjs;
}

################################################################################
# Subroutine Name: get_primary_node()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_primary_node {
    my ($self) = @_;

    my ($nodeObj, @nodeObjs);

    @nodeObjs = $self->get_nodes();
    if ($self->error()) {
        return {};    
    }

    foreach $nodeObj (@nodeObjs){
        if ($nodeObj->{primary_node} == 1) {
            return $nodeObj;
        }
    }
    
    $self->_set_err("Could not find the primary node. Probably there are no" .
                    "nodes in the cluster");
    return {};
}

################################################################################
# Subroutine Name: set_primary_node()
# Objective:       Will set the passed node as primary node
# Details:         See POD documentation at the end of this file
################################################################################
sub set_primary_node {
    my ($self, $host) = @_;

    my ($new_primary_node, $old_primary_node);

    $old_primary_node = $self->get_primary_node();
    if ($self->error()) {
        $self->sanlog("No nodes are added to the cluster. Could not return any node");
        $self->_clear_err();
    }

    $new_primary_node = $self->get_node(-host => $host);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $new_primary_node->{primary_node} = 1;
    $self->{primary_node} = $new_primary_node;
    
    # Even if above call fails then there will still be a primary node
    $old_primary_node->{primary_node} = 0;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: cluster_create()
# Objective:       Will create the physical cluster
# Details:         See POD documentation at the end of this file
################################################################################
sub cluster_create {
    my ($self) = @_;

    # It will check if primary node has all the required cluster packages installed
    $self->cluster_validate();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }
    
    # will create /etc/cluster/cluster.conf file
    $self->_write_cluster_ds();    
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: cluster_enable()
# Objective:       Start all the service required for the cluster to run
# Details:         See POD documentation at the end of this file
################################################################################
sub cluster_enable {
    my ($self) = @_;

    my ($node, @nodes, $first_node);

    @nodes = $self->get_nodes();

    $first_node = shift(@nodes);

    # In case of fence device as manual fencing it is not possible
    # to start cluster services on all the nodes one by one. If we try
    # to start the service on one node it will wait for the other
    # node to join the cluster while trying to start fenced. Thats why
    # we fork a subtest to start the cluster on primary node and wait
    # and start the services on the rest of the nodes
    my $runid = "start_cluster" . $first_node->{name}; 
    my $subtest = &Tharn::subtest(sub {$first_node->start_cluster_services();}, -runid, $runid, -bg);

    # start the service on rest of the nodes
    foreach my $node (@nodes) {
        $node->start_cluster_services();
        if ($node->error()) {
            $self->_set_err($node->errmsg());

            # wait for the subtest to finish before erroring out
            &SanUtil::nate_wait_finish($subtest);
            return $ClientUtil::FALSE;
        }
    }

    &SanUtil::nate_wait_finish($subtest);
    if (not $subtest->is_alive()) {
        my $exit_status = $subtest->exit_status();
        if ($exit_status > 0) {
            $self->_set_err("Failed in the subtest $runid");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: cluster_disable()
# Objective:       Will stop the cluster
# Details:         See POD documentation at the end of this file
################################################################################
sub cluster_disable {
    my ($self) = @_;

    my ($node, @nodes);

    # Go and disable the cluster on each node. Cluster 
    # can be disabled on each node one after the other 
    # so we dont need to fork a subtest here
    @nodes = $self->get_nodes();
    foreach $node (@nodes) {
        $node->stop_cluster_services();
        if ($node->error()) {
            $self->_set_err($node->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: cluster_validate()
# Objective:       is to verify that all the cluster related packages are available 
#                  On the primary Node
# Details:         See POD documentation at the end of this file
################################################################################
sub cluster_validate {
    my ($self) = @_;

    my ($node, @nodes);

    # This will validate if primary node is pingabl from each node
    # and all required cluster rpms are there on each node
    @nodes = $self->get_nodes();
    foreach my $node (@nodes) {
        $node->validate();
        if ($node->error()) {
            $self->_set_err($node->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog("Cluster Validation test passed");
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: cluster_destroy()
# Objective:       Will disable the cluster and then will delete 
#                  /etc/cluster/cluster.conf
# Details:         See POD documentation at the end of this file
################################################################################
sub cluster_destroy {
    my ($self) = @_;

    my ($node, @nodes);

    # It will disable the cluster on each node and delete 
    # /etc/cluster/cluster.conf from each node
    @nodes = $self->get_nodes();
    foreach $node (@nodes){
        $node->_destroy_cluster();
        if ($node->error()) {
            $self->_set_err($node->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;;
}

################################################################################
# Subroutine Name: cluster_status()
# Objective:       Return TRUE or FALSE based on whether cluster is running or not
# Details:         See POD documentation at the end of this file
################################################################################
sub cluster_status {
    my ($self, $node) = @_;

    my ($cmd);

    # check on the primary node if node is not passed
    if (!defined $node) {
        $node = $self->get_primary_node();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    # We will be fine with TRUE or FALSE
    # We dont need actual status (onile/offline)
    $cmd ="/etc/init.d/cman status";
    $self->execute(-cmd => $cmd, -target_node => $node, -timeout => 30);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: set_arbitration_policy()
# Objective:       set policy on the cluster if there is any
# Details:         See POD documentation at the end of this file
################################################################################
sub set_arbitration_policy {
    my ($self) = @_;

    # No-op for Linux
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: validate_interconnect()
# Objective:       Validate if the primary node is able to ping the other node
# Details:         See POD documentation at the end of this file
################################################################################
sub validate_interconnect {
    my ($self, $node) = @_;

    my ($primary_node, $cmd, $status, @output);

    $primary_node = $self->get_primary_node();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Currently we support only two Node So validation of 
    # one other from primary node will be sufficient
    $node->_validate_interconnect($primary_node);
    if ($node->error()) {
        $self->_set_err($primary_node->errmsg());
        return $ClientUtil::FALSE;
    }
    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: disable_interconnect()
# Objective:       Disable the interconnect
# Details:         See POD documentation at the end of this file
################################################################################
sub disable_interconnect {
    my ($self) = @_;

    # No-op for time being
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_shared_disk_group()
# Objective:       will create shared disk group object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_shared_disk_group {
    my ($self, %args) = @_;

    my ($sdgObj);

    # Create share disk group object
    $sdgObj = San::SanClientCluster::Linux::RHCS::SharedDiskGroup->_new(%args, -clusterObj => $self);
    
    push (@{$self->{sdg}}, $sdgObj);
    return $sdgObj;
}

################################################################################
# Subroutine Name: create_quorum()
# Objective:       create the quorum
# Details:         See POD documentation at the end of this file
################################################################################
sub create_quorum {
    my ($self) = @_;
    
    # Noop for Linux
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_resource_groups()
# Objective:       get the list of resource groups objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_resource_groups {
    my ($self) = @_;

    my (@resource_groups);

    my @disk_groups = $self->get_disk_groups();
    foreach my $sdgObj (@disk_groups) {
        if (defined $sdgObj) {
            push (@resource_groups, @{$sdgObj->{rgs}});
        }
    }

    return (wantarray())? @resource_groups : \@resource_groups;
}

################################################################################
# Subroutine Name: get_disk_groups()
# Objective:       get the list of shared disk groups objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_disk_groups {
    my ($self) = @_;

    my (@disk_groups);
    
    if (defined $self->{sdg}) {
        @disk_groups = @{$self->{sdg}};
    } else {
        return (wantarray())? () : []; 
    }
    
    return (wantarray())? @disk_groups : \@disk_groups;
}

################################################################################
# Subroutine Name: add_device()
# Objective:       Add the a device to the cluster
# Details:         See POD documentation at the end of this file
################################################################################
sub add_device {
    my ($self, %args) = @_;

    my ($device);

    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Invalid value for argument '-device'");
                return $ClientUtil::FALSE;
            }   
            next;
        }
    }

    unless (defined $device and length $device) {
        $self->_set_err("-device is a required argument");
        return $ClientUtil::FALSE;
    }

    # Initialize various values for the device
    $self->{shared_cluster_devices}->{$device}->{mnt_pt} = "";
    $self->{shared_cluster_devices}->{$device}->{storage_group} = "";
    $self->{shared_cluster_devices}->{$device}->{device_name} = $device;
    $self->{shared_cluster_devices}->{$device}->{sig_of_device} = $device;
    $self->{shared_cluster_devices}->{$device}->{quorum} = "no";

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_interconnect()
# Objective:       get the interconnect
# Details:         See POD documentation at the end of this file
################################################################################
sub get_interconnect {
    my ($self) = @_;

    # No-op for time being
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_node()
# Objective:       Create the node object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_node {
    my ($self, %args) = @_;

    my ($nodeObj);

    # This will create the node object followed by creating the 
    # SanClient object for the node
    $nodeObj = San::SanClientCluster::Linux::RHCS::Node->_new(-clusterObj => $self, -noDiscovery => $self->{noDiscovery}, %args);
    if ($nodeObj->error()) {
        $self->_set_err($nodeObj->errmsg);
    }

    return $nodeObj;
}

################################################################################
# Subroutine Name: cluster_refresh()
# Objective:       Refresh cluster configuration, for RHCS, offline cluster
#                  write new config file and renable cluster.
# Details:         See POD documentation at the end of this file
################################################################################
sub cluster_refresh {
    my ($self) = @_;

    # This is necessary for Linux to "hup" its cluster service
    # disabling the cluster on every node
    $self->cluster_disable();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # This will dump the cluster's XML data structure
    # into /etc/cluster/cluster.conf file
    $self->_write_cluster_ds();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Bring back online the cluster
    $self->cluster_enable();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

1;
