# Copyright (c) 2009 NetApp Inc.
# All rights reserved.

## @summary Provide an interface to Cluster
## @author  rawat@netapp.com
## @status  Private
## @pod     here

package San::SanClientCluster::Linux::RHCS::Node;

use Error;
use Data::Dumper;
use strict;

use San::SanClient;
use San::SanLog;
use San::SanClient::ClientUtil;

use vars qw(@ISA);
@ISA = qw(Error SanLog);

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create Node object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $clientObj, $status, $host, $protocol, $clusterObj, @fcp_targets, @iscsi_targets, @output);
    my ($noDiscovery);

    my $self = {};
    bless $self, $class;

    foreach (keys %args) {
        if (/-?host/) {
            if (not length($host = $args{$_})) {
                $self->_set_err("invalid for argument -host"); 
                return $self;
            }
        } elsif (/-?noDiscovery/) {
            if (not length($noDiscovery = $args{$_})) {
                $self->_set_err("Invalid value for argument '-noDiscovery'");
                return $self;
            }   
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("invalid for argument -protocol"); 
                return $self;
            }
        } elsif (/-?clusterObj/) {
            if (not length($clusterObj = $args{$_})) {
                $self->_set_err("invalid for argument -clusterObj"); 
                return $self;
            }
        } elsif (/-?fcp_targets/) {
            if (not length(@fcp_targets = @{$args{$_}})) {
                $self->_set_err("invalid for argument -fcp_targets"); 
                return $self;
            }
        } elsif (/-?iscsi_targets/) {
            if (not length(@iscsi_targets = @{$args{$_}})) {
                $self->_set_err("invalid for argument -iscsi_targets"); 
                return $self;
            }
        }
    }

    unless (defined $host and length $host) {
        $self->_set_err("-host is a required argument");
        return $self;
    }

    # create SanClient object
    $clientObj = SanClient->new(%args);
    if ($clientObj->error()) {
        $self->_set_err($clientObj->errmsg);
        return $self;
    }

    if ($noDiscovery == $ClientUtil::TRUE) {
        # Configure the node. It will start iscsi/fcp on the host
        if ($protocol =~ /fcp/i) {
            $clientObj->configure(-targets => \@fcp_targets);
        } else {
            $clientObj->configure(-targets => \@iscsi_targets);
        }
        if ($clientObj->error()) {
            $self->_set_err($clientObj->errmsg);
            return $self;
        }
 
        # Will discover the luns on the host
        $clientObj->discover();
        if ($clientObj->error) {
            $self->_set_err($clientObj->errmsg);
            return $self;
        }
 
        # Create device manager object
        my $devmgrObj = $clientObj->create_devmgr();
        if ($clientObj->error) {
            $self->_set_err($clientObj->errmsg);
            return $self;
        }
 
        # Get the list of free devices
        my @devices = $devmgrObj->get_free_devices();
        if ($devmgrObj->error) {
            $self->_set_err($devmgrObj->errmsg);
            return $self;
        }
 
        # Add rest of the devices to the cluster
        foreach my $device (@devices) {
            $clusterObj->{shared_cluster_devices}->{$device}->{storage_group} = "";
            $clusterObj->{shared_cluster_devices}->{$device}->{device_name} = $device;
        }
    }
    
    $self->{clientObj} = $clientObj;
    $self->{conn} = $clientObj->{conn};
    $self->{parent} = $clusterObj;
    $self->{name} = $host;
    
    return $self;
}

################################################################################
# Subroutine Name: _create_cluster_conf()
# Objective:       write the cluster data structure into the file 
#                  /etc/cluster/cluster.conf
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_cluster_conf {
    my ($self) = @_;

    my ($cluster_conf, $conn, $cmd, $status, @output);

    my $xs = new XML::Simple(keeproot => 1, suppressempty => 1, forcearray => 1, keyattr =>"");

    # Read cluster config data structure in xml format
    if (! defined $self->{parent}->{ccds}) {
        $self->_set_err("ccds is undefined");
        return $ClientUtil::FALSE;
    }
    $cluster_conf = $xs->XMLout($self->{parent}->{ccds});
    if (! defined $cluster_conf) {
        $self->_set_err("unable to define cluster_conf");
        return $ClientUtil::FALSE;
    }

    # This line the first line in cluster.conf file which is not part of data structure
    my $xml_format = "<?xml version=\"1.0\" ?>";

    # This file need to be copied to all the nodes the cluster
    $cmd = "echo '$xml_format' > /etc/cluster/cluster.conf";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "echo '$cluster_conf' >> /etc/cluster/cluster.conf";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _destroy_cluster()
# Objective:       Will disable the cluster and then will delete 
#                  /etc/cluster/cluster.conf
# Details:         See POD documentation at the end of this file
################################################################################
sub _destroy_cluster {
    my ($self) = @_;

    my ($cmd);

    # Check if the cluster is running
    if ($self->_cluster_service_status() == $ClientUtil::TRUE) {
        $self->stop_cluster_services();
        if($self->error()){
             $self->_set_err("failed to disable the cluster for node $self->{name}");
             return $ClientUtil::FALSE;
        }
    } elsif ($self->error()) {
        return $ClientUtil::FALSE;
    }
    
    # Take the backup of cluster.conf
    $cmd = "yes | cp -f --backup /etc/cluster/cluster.conf /etc/cluster/cluster.conf.org";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Remove the cluster.conf file
    $cmd = "rm -f /etc/cluster/cluster.conf";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;;

}

################################################################################
# Subroutine Name: _cluster_service_status()
# Objective:       Return TRUE or FALSE based on whether cluster is running or not
# Details:         See POD documentation at the end of this file
################################################################################
sub _cluster_service_status {
    my ($self) = @_;

    my ($cmd);

    $cmd = "/etc/init.d/cman status";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error) {
        if ($self->{parent}->errmsg() =~ /pid\s+file\s+exists/i) {
            # clear error from $self because we dont care about it
            $self->_clear_err();
            return $ClientUtil::FALSE;
        } else {
            $self->_set_err($self->{parent}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _verify_lvm_rpms()
# Objective:       Verify rpm packages for LVM 
# Details:         See POD documentation at the end of this file
################################################################################
sub _verify_lvm_rpms {
    my ($self, $node) = @_;

    my ($cmd);

    $cmd = "/bin/rpm -q lvm2";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # This rpm is not necessary but require if you want to use gui to configure lvm/clvm
    $cmd = "/bin/rpm -q system-config-lvm";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->{parent}->_clear_err(); 
        $self->sanlog("Warning:SanCluster::Linux::RHCS::Node::_verify_lvm_rpms " . 
                      "rpm system-config-lvm is not installed");
    }

    $cmd = "/bin/rpm -q lvm2-cluster";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _validate_interconnect()
# Objective:       Validate the node. Validation includes pinging to the primary node. If it
#                  is not able to ping primary node then validation will fail
# Details:         See POD documentation at the end of this file
################################################################################
sub _validate_interconnect {
    my ($self, $node) = @_;

    my ($cmd, $status, @output);
    
    # ping primary node from this node
    $self->ping($node);
    if ($self->error()) {
        $self->_set_err("Failed to ping $node->{name} node by its nodename. " . 
                        "Will not be able to join the cluster. Make sure you have added an entry " .
                        "for each node in cluster to the /etc/hosts file of this client" );
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: add()
# Objective:       Preparing the internal DS for cluster.conf file
# Details:         See POD documentation at the end of this file
################################################################################
sub add {
    my ($self) = @_;

    my ($cmd, $node_id);

    $node_id = $self->{parent}->{unique_node_id};

    my $ccds = $self->{parent}->{cluster};
    
    ############################################################################
    # /etc/cluster/cluster.conf file consists of three parts 
    # clusternodes, fencedevics and managed resources
    ############################################################################
    # <?xml version="1.0" ?>
    #<cluster alias="nate_rhcs" config_version="12" name="nate_rhcs">
    #   <clusternodes>
    #   </clusternodes>
    #   <fencedevices>
    #   </fencedevices>
    #   <rm>
    #   </rm>
    #</cluster> 
    ############################################################################
    
    ############################################################################
    # Here we are going to prepare DS for clusternodes. The following
    # part of the cluster.conf file is going to be added by folloing code
    ############################################################################
    #   <clusternodes>
    #       <clusternode name="<nodename>" nodeid="<nodeid>" votes="1">
    #           <fence>
    #                   <method name="1">
    #                       <device name="mf" nodename="<nodename>"/>
    #                   </method>
    #           </fence>
    #       </clusternode>
    #   </clusternodes>
    #   <rm>
    #       <failoverdomains>
    #           <failoverdomain name="nate_fd" ordered="0" restricted="0">
    #               <failoverdomainnode name="<nodename>" priority="1"/>
    #           </failoverdomain>
    #       </failoverdomains>
    #   </rm>
    #############################################################################

    my $clusternodes = $self->{parent}->{ccds}->{cluster}->[0]->{clusternodes};
    # Adding node info to the cluster config data structure
    $self->{parent}->{ccds}->{cluster}->[0]->{clusternodes}->[0]->{clusternode}->[$node_id]->{name} = $self->{name};
    $self->{parent}->{ccds}->{cluster}->[0]->{clusternodes}->[0]->{clusternode}->[$node_id]->{nodeid} = $node_id;
    $self->{parent}->{ccds}->{cluster}->[0]->{clusternodes}->[0]->{clusternode}->[$node_id]->{votes} = 1;
    $self->{parent}->{ccds}->{cluster}->[0]->{clusternodes}->[0]->{clusternode}->[$node_id]->{fence}->[0]->{method}->[0]->{name} = 1;
    $self->{parent}->{ccds}->{cluster}->[0]->{clusternodes}->[0]->{clusternode}->[$node_id]->{fence}->[0]->{method}->[0]->{device}->[0]->{nodename} = $self->{name};
    $self->{parent}->{ccds}->{cluster}->[0]->{clusternodes}->[0]->{clusternode}->[$node_id]->{fence}->[0]->{method}->[0]->{device}->[0]->{name} = "mf";

    my $clusterFD = $self->{parent}->{ccds}->{cluster}->[0]->{rm}->[0]->{failoverdomains};
    $self->{parent}->{ccds}->{cluster}->[0]->{rm}->[0]->{failoverdomains}->[0]->{failoverdomain}->[0]->{failoverdomainnode}->[$node_id]->{priority} = "1";
    $self->{parent}->{ccds}->{cluster}->[0]->{rm}->[0]->{failoverdomains}->[0]->{failoverdomain}->[0]->{failoverdomainnode}->[$node_id]->{name} = $self->{name};

    $self->{node_id} = $node_id;
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: disable_cluster_at_bootup()
# Objective:       Disable the cluster to start on bootup
# Details:         See POD documentation at the end of this file
################################################################################
sub disable_cluster_at_bootup {
    my ($self) = @_;

    my ($cmd, $status, @output);
    
    # Disable resource group manager to start the bootup
    $cmd = "chkconfig --level 2345 rgmanager off";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Disable global filesystem daemon to start the bootup
    $cmd = "chkconfig --level 2345 gfs off";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Disable cluster lvm daemon to start the bootup
    $cmd = "chkconfig --level 2345 clvmd off";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Disable cluster manager daemon to start the bootup
    $cmd = "chkconfig --level 2345 cman off";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: enable_cluster_at_bootup()
# Objective:       Enable the cluster to start on bootup 
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_cluster_at_bootup {
    my ($self) = @_;
    my ($cmd, $status, @output);
    
    # Enable resource group manager to start the bootup
    $cmd = "chkconfig --level 2345 rgmanager on";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Enable global filesystem daemon to start the bootup
    $cmd = "chkconfig --level 2345 gfs on";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Enable cluster lvm daemon to start the bootup
    $cmd = "chkconfig --level 2345 clvmd on";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Enable cluster manager daemon to start the bootup
    $cmd = "chkconfig --level 2345 cman on";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: start_cluster_services()
# Objective:       start the cluster services on the node
# Details:         See POD documentation at the end of this file
################################################################################
sub start_cluster_services {
    my ($self) = @_;

    my ($cmd, $status, @output);
    
    # Start cluster manager daemon
    $cmd = "/etc/init.d/cman start";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }
    
    # Start clustered lvm daemon
    $cmd = "/etc/init.d/clvmd start";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Start gfs
    $cmd = "/etc/init.d/gfs start";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Start resource group manager daemon
    $cmd = "/etc/init.d/rgmanager start";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: stop_cluster_services()
# Objective:       Stop the cluster services on the node
# Details:         See POD documentation at the end of this file
################################################################################
sub stop_cluster_services {
    my ($self) = @_;

    my ($conn, $cmd, $status, @output);
    
    # Stop resource group manager daemon
    $cmd = "/etc/init.d/rgmanager stop";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Stop gfs
    $cmd = "/etc/init.d/gfs stop";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Stop clustered lvm daemon
    $cmd = "/etc/init.d/clvmd stop";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Stop cluster manager daemon
    $cmd = "/etc/init.d/cman stop";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: join()
# Objective:       join the node to the cluster while cluster is online
# Details:         See POD documentation at the end of this file
################################################################################
sub join {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $cmd = "cman_tool join";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: leave()
# Objective:       leave the node from the cluster
# Details:         See POD documentation at the end of this file
################################################################################
sub leave {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $cmd = "cman_tool leave";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove()
# Objective:       Remove the node from the cluster
# Details:         See POD documentation at the end of this file
################################################################################
sub remove {
    my ($self) = @_;

    my ($cmd, $status, @output);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: set_primary_node()
# Objective:       set the node as primary node
# Details:         See POD documentation at the end of this file
################################################################################
sub set_primary_node {
    my ($self) = @_;

    $self->{primary_node} = 1;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: node_status()
# Objective:       check if the cluster is running or not on this node 
# Details:         See POD documentation at the end of this file
################################################################################
sub node_status {
    my ($self) = @_;

    my ($cmd, $status, @output);

    ######################################################################################### 
    # here is the sample output of the clustat cmd
    ##########################################################################################
    # Cluster Status for nate_rhcs @ Wed Feb 11 22:25:44 2009
    # Member Status: Quorate
    # Member Name                                                     ID   Status
    # ------ ----                                                     ---- ------
    # lnx236-103                                                        1 Online, Local
    # lnx236-104                                                        2 Offline
    ##########################################################################################
    $cmd = "clustat | grep $self->{name}";
    @output = $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return "";
    }

    my @tmp_arr = split (/\s+/, $output[0]);
    $status = $tmp_arr[3];

    if($status =~ /online/) {
        $status = "online";
    } else {
        $status = "offline";
    }

    return $status;
}

################################################################################
# Subroutine Name: get_sanclient_obj()
# Objective:       return SanClient object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_sanclient_obj {
    my ($self) = @_;

    if (!defined $self->{clientObj}) {
        $self->_set_err("ClientObj is not initialized for this Node");
        return {};
    }
    return $self->{clientObj};
}

################################################################################
# Subroutine Name: validate()
# Objective:       Validate the node. Validation includes pinging to the primary node. If it
#                  is not able to ping primary node then validation will fail
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self) = @_;

    my ($cmd, $status, @output);
    
    # The node should be able to ping the primary node through interface which is 
    # going to be used for heartbeat 
    my $primary_node = $self->{parent}->get_primary_node();

    if ($self->{parent}->error) {
        # If primary node is not defined it means that this is the
        # first node added and it is the primary node
        $primary_node = $self;
        $self->{parent}->_clear_err();
    }

    # ping primary node from this node
    $self->ping($primary_node);
    if ($self->error()) {
        # Wanted to set different error message
        $self->_set_err("Failed to ping primary node by node_name. " . 
                        "Will not be able to join the cluster. Make sure you have added an entry " .
                        "for each node in cluster to the /etc/hosts file of this client" );
        return $ClientUtil::FALSE;
    }

    # Verify if all the rpms are installed on the node
    $cmd = "/bin/rpm -q rgmanager";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "/bin/rpm -q cluster-snmp";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "/bin/rpm -q cluster-cim";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "/bin/rpm -q modcluster";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # This rpm is not necessary but require if you want to use gui to configure cluster
    $cmd = "/bin/rpm -q system-config-cluster";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->{parent}->_clear_err();
        $self->sanlog("Warning: SanCluster::Linux::RHCS::Node::validate rpm system-config-cluster is not installed");
    }

    # It will verify lvm/clvm related rpms
    $self->_verify_lvm_rpms();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: ping()
# Objective:       Will ping the node passed
# Details:         See POD documentation at the end of this file
################################################################################
sub ping {
    my ($self, $node) = @_;

    my ($cmd);

    # Ping $node by its node name
    use Data::Dumper;
    print "self" . Dumper($self) . "\n";
    print "parent" . Dumper($self->{parent}) . "\n";
    $cmd = "/bin/ping -c 2 $node->{name}";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err("Unable to ping the node $node->{name}");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reboot()
# Objective:       Reboot the node
# Details:         See POD documentation at the end of this file
################################################################################
sub reboot {
    my ($self) = @_;
    
    my ($client_obj);

    $client_obj = $self->{clientObj};

    # Will call reboot on SanClient to reboot the Node
    $client_obj->reboot();
    if ($client_obj->error()) {
        $self->_set_err($client_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: halt()
# Objective:       Halt the node
# Details:         See POD documentation at the end of this file
################################################################################
sub halt {
    my ($self) = @_;

    my ($cmd);

    # Halt the node
    $cmd = "/sbin/halt $self->{name}";
    $self->{parent}->execute(-cmd => $cmd, -target_node => $self);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

1;
