# Copyright (c) 2009 NetApp Inc.
# All rights reserved.

## @summary Provide an interface to SharedDiskGroup object 
## @author  rawat@netapp.com
## @status  Private
## @pod     here

package San::SanClientCluster::Linux::RHCS::SharedDiskGroup;

use Error;
use Data::Dumper;
use strict;

use San::SanClientCluster::Linux::RHCS::SharedDiskGroup::Resource;
use San::SanClient;
use San::SanLog;
use San::SanClient::ClientUtil;

use vars qw(@ISA);
@ISA = qw(Error SanLog);

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create the shared disk group disk object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($parent, $type, $primary_node, $exists);
    my $self = {};
    
    bless $self, $class;

    foreach (keys %args) {
        if (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return $self;
            }   
            next;
        } elsif (/-?exists/) {
            if (not length($exists = $args{$_})) {
                $self->_set_err("Invalid value for argument '-exists'");
                return $self;
            }   
            next;
        } elsif (/-?clusterObj/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clusterObj'");
                return $self;
            }   
            next;
        }
    }

    # Make sure that the required arguments are passed
    unless (defined $type and length $type) {
        $self->_set_err("-type is a required argument");
        return $self;
    }

    unless (defined $parent and length $parent) {
        $self->_set_err("-clusterObj is a required argument");
        return $self;
    }

    $self->{parent} = $parent;

    # Get the primary node
    $primary_node = $self->{parent}->get_node(-primary => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $self;
    }

    # Get the SanClient object
    my $clientObj = $primary_node->get_sanclient_obj();
    if ($primary_node->error) {
        $self->_set_err($primary_node->errmsg());
        return $self;
    }

    # create lvm object
    my $lvmObj = $clientObj->create_lvm(-lvm => $type);
    if ($clientObj->error) {
        $self->_set_err($clientObj->errmsg());
        return $self;
    }

    $self->{parent}->{lvm} = $lvmObj;

    # We are not going to call initialize because vg already
    # exists. Saving vg object on $self here instead of inside initialize()
    if ($exists) {
        my @vgObj = $lvmObj->get_vgs();
        # There will be only one VG 
        $self->{vgObj} = $vgObj[0];
    }

    return $self;
}

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: create_rg()
# Objective:       Create resource group object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_rg {
    my ($self, %args) = @_;

    my ($rgObj);
    
    # Create resource group object
    $rgObj = SanCluster::Linux::RHCS::SharedDiskGroup::Resource->_new(-parent => $self, %args);
    if ($rgObj->error) {
        $self->_set_err($rgObj->errmsg());
        return $rgObj;
    }

    push (@{$self->{rgs}}, $rgObj);

    return $rgObj;
}

################################################################################
# Subroutine Name: destroy()
# Objective:       Destroy all the resource group objects. Will also destroy SDG object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;

    my ($rgObj, $cluster_obj);
    
    my @devices = $self->get_devices();

    # Destroy all the resource group types
    foreach $rgObj (@{$self->{rgs}}) {
        $rgObj->destroy();
        if ($rgObj->error()) {
            $self->_set_err($rgObj->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    foreach my $device (@devices) {
        $cluster_obj->{shared_cluster_devices}->{$device}->{free} = 1;
    }

    # Remove each device 
    foreach my $device (@devices) {
        $self->remove_device($device);
        if ($self->error()) {
            return $ClientUtil::TRUE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: initialize()
# Objective:       Initialize the devices
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize {
    my ($self, %args) = @_;

    my ($lvmObj, $volObj, @devices);

    # get the list of the devices to initailize
    @devices = $self->get_devices();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    $lvmObj = $self->{parent}->{lvm};

    # initialize the disks
    $lvmObj->initialize(-devices => \@devices);
    if ($lvmObj->error) {
        $self->_set_err($lvmObj->errmsg());
        return $ClientUtil::FALSE;
    }

    # create vol group object
    my $vgObj = $lvmObj->create_vg(-devices => \@devices, -name => 'n_shared_vg');
    if ($vgObj->error) {
        $self->_set_err($vgObj->errmsg());
        return $ClientUtil::FALSE;
    }

    my $cluster_obj = $self->{parent};
    foreach my $device (@devices) {
        $cluster_obj->{shared_cluster_devices}->{$device}->{free} = 0;
    }

    $self->{vgObj} = $vgObj;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove_device()
# Objective:       remove the device from cluster
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_device {
    my ($self, $device) = @_;

    my ($cluster_obj);

    $cluster_obj = $self->{parent};

    # Remove the device passed to it
    foreach my $dev_serial (keys %{$cluster_obj->{shared_cluster_devices}}) {
        if ($device eq $cluster_obj->{shared_cluster_devices}->{$dev_serial}->{device_name}){
            if ($cluster_obj->{shared_cluster_devices}->{$dev_serial}->{free} == 0) {
                $self->_set_err("Device is not free");
                return $ClientUtil::FALSE;
            } else {
                delete $cluster_obj->{shared_cluster_devices}->{$dev_serial};
                return $ClientUtil::TRUE;
            }
        }
    }
    
    $self->_set_err("Device is not present in device list");
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_devices()
# Objective:       return the devices list
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;

    my ($cluster_obj, @devices);

    $cluster_obj = $self->{parent};
    
    foreach my $dev_serial (keys %{$cluster_obj->{shared_cluster_devices}}) {
        push (@devices, $dev_serial);
    }

    return wantarray() ? @devices : \@devices;

}

################################################################################
# Subroutine Name: get_lvm_object()
# Objective:       return lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lvm_object {
    my ($self) = @_;

    if (defined $self->{parent}->{lvm} and $self->{parent}->{lvm} ne {}) {
        return $self->{parent}->{lvm};
    } else {
        $self->_set_err("failed to get lvm obj");
        return {};
    }
}

1;
