# Copyright (c) 2009 NetApp Inc.
# All rights reserved.

## @summary Provide an interface to Cluster Resource
## @author  rawat@netapp.com
## @status  Private
## @pod     here

package SanCluster::Linux::RHCS::SharedDiskGroup::Resource;

use Error;
use Data::Dumper;
use strict;

use San::SanLog;
use San::SanClient::ClientUtil;

use vars qw(@ISA);
@ISA = qw(Error SanLog);

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create resource group object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($resource_type, $parent, $exists);
    my $self = {};
    
    bless $self, $class;

    foreach (keys %args) {
        if (/-?resource_type/) {
            if (not length($resource_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-resource_type'");
                return $self;
            }   
            next;
        } elsif (/-?exists/) {
            if (not length($exists = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }   
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }   
            next;
        }
    }

    # Make sure that the required arguments are passed
    unless (defined $resource_type and length $resource_type) {
        $self->_set_err("-resource_type is a required argument");
        return $self;
    }

    unless (defined $parent and length $parent) {
        $self->_set_err("-parent is a required argument");
        return $self;
    }

    $self->{parent} = $parent;
    $self->{type} = $resource_type;
    
    # If exists is set then no need to create lvs and also
    # no need to populate cluster config data structure
    # because its already there
    if ($exists) {
        if (defined $parent->{vgObj}) {
            @{$self->{volObjs}} = $parent->{vgObj}->get_vols();
        } else {
            $self->sanlog(-msg => "No Vg found on the host");
        }
        return $self;
    }

    # Create clusterfs resource type
    if ($resource_type =~ /clusterfs/) {
        $self->_create_clusterfs_resource_type(%args);    
        if ($self->error) {
            return $self;
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: _build_clusterfs_resource_ds()
# Objective:       will add all the lvs to the cluster config data structure
# Details:         See POD documentation at the end of this file
################################################################################
sub _build_clusterfs_resource_ds {
    my ($self) = @_;

    my ($cmd, $status, @output, $clusterObj, $vgObj, $vg_name, $mnt_pnt);
    my ($vol_name, $fstype, $resource_type, $vol_num, $resource_name);
    
    $clusterObj = $self->{parent}->{parent};
    $vgObj = $self->{parent}->{vgObj};
    $vg_name = $vgObj->get_name();
    
    foreach my $vol_obj (@{$self->{volObjs}}) {
        $vol_name = $vol_obj->get_name();
        ($vol_num) = ($vol_name =~ /nvol-(\d+)/);

        $mnt_pnt = $vol_obj->get_mount();
        if (! $mnt_pnt) {
            $mnt_pnt = "/mnt/qa/$vg_name/$vol_name";
        }

        $fstype = $vol_obj->get_fs();
 
        $resource_type = $self->{type};

        $resource_name = "nate_" . $fstype . "_" . $vol_name;

        # Creating clusterfs resource type data structure
        # This is the part of the /etc/cluster/cluster.conf that is going to be 
        # created by the following data structure. This is the One instance of
        # resource type clusterfs. This will also create the script resource
        # type for clusterfs. Since creation of script resource
        # type will highly depend on the resource type for which it is going 
        # to be created. So its better if we create script with the same 
        # resource type that is going to use it
        #########################################################################
        #<cluster alias="nate_rhcs" config_version="3" name="nate_rhcs">
        #   <rm>
        #       <resources>
        #           <clusterfs device="<vol_name>" force_unmount="1" fsid="<fsid>" fstype="<fstype>" mountpoint="<mnt_pnt>" name="<resource_name>" options=""/>
        #           <script file="<script_path>" name="<script_name>" />
        #       </resources>
        #       <service autostart="0" domain="nate_fd" name"<service_name>" recovery="relocate">
        #           <clusterfs ref="<vol_name>"/>
        #           <script ref="<script_name>"/>
        #       </service>
        #   </rm>
        #########################################################################
        #my $resource = $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type};
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type}->[$vol_num]->{options} = "";
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type}->[$vol_num]->{name} = $resource_name;
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type}->[$vol_num]->{mountpoint} = "$mnt_pnt";
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type}->[$vol_num]->{device} = "/dev/$vg_name/$vol_name";
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type}->[$vol_num]->{fstype} = "$fstype";
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type}->[$vol_num]->{force_unmount} = "1";
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type}->[$vol_num]->{fsid} = 10000 + $vol_num;

        # For each resource we will create a service 
        #my $service = $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{service};
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{service}->[$vol_num]->{clusterfs}->[0]->{ref} = $resource_name;
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{service}->[$vol_num]->{name} = "service_" . $resource_name;
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{service}->[$vol_num]->{domain} = "nate_fd";
        $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{service}->[$vol_num]->{autostart} = 0;

        # Push the resources in $self->{resources}
        push (@{$self->{resources}}, $resource_name);
    }
    
    return $ClientUtils::TRUE;
}

################################################################################
# Subroutine Name: _create_clusterfs_resource_type()
# Objective:       will create gfs resource type
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_clusterfs_resource_type {
    my ($self, %args) = @_;

    my ($rgObj, $size, $vgObj, $lvmObj, $layout, $num_of_lvs, $iotype);
    my ($cluster_fstype, $pass_thru, $volObj, $parent, $cmd, $status, @output);

    foreach (keys %args) {
        if (/-?vol_layout/) {
            if (not length($layout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vol_layout'");
                return $ClientUtils::FALSE;
            }   
            next;
        } elsif (/-?cluster_fstype/) {
            if (not length($cluster_fstype = $args{$_})) {
                $self->_set_err("Invalid value for argument '-cluster_fstype'");
                return $ClientUtils::FALSE;
            }   
            next;
        } elsif (/-?num_of_lvs/) {
            if (not length($num_of_lvs = $args{$_})) {
                $self->_set_err("Invalid value for argument '-num_of_lvs'");
                return $ClientUtils::FALSE;
            }   
            next;
        } elsif (/-?client_io_type/) {
            if (not length($iotype = $args{$_})) {
                $self->_set_err("Invalid value for argument '-client_io_type'");
                return $ClientUtils::FALSE;
            }   
            next;
        } elsif (/-?vol_size/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vol_size'");
                return $ClientUtils::FALSE;
            }   
            next;
        }
    }

    my $clusterObj = $self->{parent}->{parent};

    # Make sure required arguments are passed
    unless (defined $layout and length $layout) {
        $self->_set_err("-layout is a required argument");
        return $ClientUtils::FALSE;
    }

    unless (defined $size and length $size) {
        $self->_set_err("-size is a required argument");
        return $ClientUtils::FALSE;
    }

    unless (defined $num_of_lvs and length $num_of_lvs) {
        $self->_set_err("-num_of_lvs is a required argument");
        return $ClientUtils::FALSE;
    }

    # Assign default values to these parameter if not passed
    unless (defined $iotype and length $iotype) {
        $iotype = "fs";
    }

    unless (defined $cluster_fstype and length $cluster_fstype) {
        $cluster_fstype = "gfs";
    }

    $vgObj = $self->{parent}->{vgObj};

    # create LVs 
    foreach my $vol (1 .. $num_of_lvs) {
        my $name = "nvol-$vol";
        $volObj = $vgObj->create_vol(-layout => $layout , -size => $size, -name => $name, -cluster => $clusterObj->{cluster_name});
        if ($vgObj->error) {
            $self->_set_err($vgObj->errmsg);
            return $ClientUtils::FALSE;
        }

        $volObj->create_fs(-type => $cluster_fstype);
        if ($volObj->error) {
            $self->_set_err($volObj->errmsg);
            return $ClientUtils::FALSE;
        }

        push (@{$self->{volObjs}}, $volObj);
    }

    # add clusterfs resources info to the cluster config data structure
    $self->_build_clusterfs_resource_ds();
    if ($self->error) {
        return $ClientUtils::FALSE;
    }

    return $ClientUtils::TRUE;
}

################################################################################
# Subroutine Name: _destroy_clusterfs_resource_type()
# Objective:       will destroy gfs resource type
# Details:         See POD documentation at the end of this file
################################################################################
sub _destroy_clusterfs_resource_type {
    my ($self) = @_;

    my ($vgObj, $iotype, $clusterObj, $volObj, @volObjs);

    # Get lvm object
    my $lvmObj = $self->{parent}->get_lvm_object();
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Get the list of VG objects
    my @vgObjs = $lvmObj->get_vgs();
    if ($lvmObj->error()) {
        $self->_set_err($lvmObj->errmsg());
        return $ClientUtil::FALSE;
    }
    my $resource_type = $self->{type};
    
    foreach $vgObj (@vgObjs) {
        # Get the list of vol objects
        @volObjs = $vgObj->get_vols();
        if ($vgObj->error()) {
            $self->_set_err($vgObj->errmsg());
            return $ClientUtil::FALSE;
        }
        foreach $volObj (@volObjs) {
            # Get the name before destroying the volObj
            my $fstype = $volObj->get_fs();
            my $volname = $volObj->get_name();
            my ($vol_num) = ($volname =~ /vol-(\d+)/);
            my $resource_name = "nate_" . $fstype . "_" . $volname;
            my $service = $self->_get_service_name($resource_name);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
            
            # Disable the service before deleting resource type entry from ccds
            $self->disable(-service => $service);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
            
            # destroy the volume object and physical volume
            $volObj->destroy();
            if ($volObj->error()) {
                $self->_set_err($volObj->errmsg());
                return $ClientUtil::FALSE;
            }

            # delete the resource type clusterfs from cluster config data structure
            delete $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{resources}->[0]->{$resource_type}->[$vol_num];

            # also delete the service corresponding to this resource
            delete $clusterObj->{ccds}->{cluster}->[0]->{rm}->[0]->{service}->[$vol_num];
        }

        # Destroy the VG object and physical VG
        $vgObj->destroy();
        if ($vgObj->error()) {
            $self->_set_err($vgObj->errmsg());
            return $ClientUtil::FALSE;
        }
    }
    
    return $ClientUtils::TRUE;
}

################################################################################
# Subroutine Name: _get_cluster_obj()
# Objective:       will return SanCluster object
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_cluster_obj {
    my ($self) = @_;
    
    if (defined $self->{parent}->{parent}) {
        return $self->{parent}->{parent};
    } else {
        $self->_set_err("Cluster object is not defined");
        return {};
    }
}

################################################################################
# Subroutine Name: _get_service_name()
# Objective:       return the service for the resource 
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_service_name {
    my ($self, $resource_name) = @_;
    
    if (defined $resource_name) {
        return "service_". $resource_name;
    } else {
        $self->_set_err("Resource type $resource_name doesnt exists");
        return "";
    }
}

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: destroy()
# Objective:       destroy the resource
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;

    # $self->disable() will be called from each one's separate 
    # destroy subroutines
    if ($self->{type} =~ /clusterfs/) {
        $self->_destroy_clusterfs_resource_type();
    } 

    if ($self->error) {
        return $ClientUtils::FALSE;
    }

    return $ClientUtils::TRUE;
}


################################################################################
# Subroutine Name: move()
# Objective:       reallocate the resource(service) to the target node
# Details:         See POD documentation at the end of this file
################################################################################
sub move {
    my ($self, %args) = @_;

    my ($service, $to_node, $cmd, $clusterObj);
    
    foreach (keys %args) {
        if (/-?service/) {
            if (not length($service = $args{$_})) {
                $self->_set_err("Invalid value for argument '-service'");
                return $ClientUtils::FALSE;
            }   
            next;
        } elsif (/-?node/) {
            if (not length($to_node = $args{$_})) {
                $self->_set_err("Invalid value for argument '-node'");
                return $ClientUtils::FALSE;
            }   
            next;
        }
    }

    unless (defined $service and length $service) {
        $self->_set_err("-service is a required argument");
        return $ClientUtils::FALSE;
    }

    $clusterObj = $self->_get_cluster_obj();
    if ($self->error()) {
        return $ClientUtils::FALSE;
    }

    my @nodes = $clusterObj->get_nodes();
    if (scalar(@nodes) < 2) {
        $self->_set_err("To move the service Cluster needs to have atleast 2 nodes");
        return $ClientUtils::FALSE;
    }
    
    # Relocationg the service to passed node
    # if the target node is not specified then the service will
    # reallocate to other node (in case of 2-clusterd node)
    if (defined $to_node) {
        $cmd = "clusvcadm -r $service -m $to_node->{name}";
    } else {
        $cmd = "clusvcadm -r $service";
    }
    $clusterObj->execute(-cmd => $cmd);
    if ($clusterObj->error) {
        $self->_set_err($clusterObj->errmsg());
        return $ClientUtils::FALSE;
    }
    
    return $ClientUtils::TRUE;
}

################################################################################
# Subroutine Name: enable()
# Objective:       enable the resource (service) on the target node
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args)= @_;

    my ($service, $on_node, $cmd);
    
    foreach (keys %args) {
        if (/-?service/) {
            if (not length($service = $args{$_})) {
                $self->_set_err("Invalid value for argument '-service'");
                return $ClientUtils::FALSE;
            }   
            next;
        } elsif (/-?node/) {
            if (not length($on_node = $args{$_})) {
                $self->_set_err("Invalid value for argument '-node'");
                return $ClientUtils::FALSE;
            }   
            next;
        }
    }

    unless (defined $service and length $service) {
        $self->_set_err("-service is a required argument");
        return $ClientUtils::FALSE;
    }

    my $clusterObj = $self->_get_cluster_obj();
    if ($self->error()) {
        return $ClientUtils::FALSE;
    }

    # enabling the service on the node
    # if the target node is not specified then the service will
    # be enabled on the primary node
    if (defined $on_node) {
        $cmd = "clusvcadm -e $service -m $on_node->{name}";
    } else {
        $cmd = "clusvcadm -e $service";
    }

    $clusterObj->execute(-cmd => $cmd);
    if ($clusterObj->error) {
        $self->_set_err($clusterObj->errmsg());
        return $ClientUtils::FALSE;
    }
    
    return $ClientUtils::TRUE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       Disable the resource type. Exact service name need to be 
#                  passed
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    my ($service, $cmd);

    foreach (keys %args) {
        if (/-?service/) {
            if (not length($service = $args{$_})) {
                $self->_set_err("Invalid value for argument '-service'");
                return $ClientUtils::FALSE;
            }   
            next;
        }
    }

    unless (defined $service and length $service) {
        $self->_set_err("-service is a required argument");
        return $ClientUtils::FALSE;
    }

    my $clusterObj = $self->_get_cluster_obj();
    if ($self->error()) {
        return $ClientUtils::FALSE;
    }

    # Disabling the service in the cluster
    # It is not possible to disable the service on node other than self 
    $cmd = "clusvcadm -d $service";
    $clusterObj->execute(-cmd => $cmd);
    if ($clusterObj->error) {
        $self->_set_err($clusterObj->errmsg());
        return $ClientUtils::FALSE;
    }
    
    return $ClientUtils::TRUE;
}

1;
