##
## This library is used to populate the Windows Cluster object created by
## WinCluster.pm
##
##
## @summary
##  Creates Windows Cluter object and imports associated libraries.  All architecture
##  specific properties are defined here and extrapolated elsewhere.
##
## @author  Jonathan.Dascenzo@Netapp.com
## @status  public
##

package San::SanClientCluster::Windows::MSCS;

use strict;
use WinClient;
use Error;
use San::SanLog;
use San::SanClient;
use San::SanClientCluster::Windows::MSCS::Node;
use San::SanClientCluster::Windows::MSCS::SharedDiskGroup;

use vars qw(
    @EXPORT
    @ISA
    );

@ISA = qw( SanLog Error );

# The major version number of windows 2008 is 6.
# This helps us differentiate between Windows releases.
my $WIN2K8BUILD = 6;

###########################################################################
# Name: _new
# Input: -host, -cluster_ip, -cluster_subnet, -cluster_name
# Output: returns SanCluster object
# Tasks: * Creates empty SanCluster Object
#       * Creates Node object for -host
#       * Sets attributes of cluster on SanCluster object (-cluster_ip,
#       *  -cluster_subnet, -cluster_name)
#       * Sets attribute on Node object that -host is the primary node
############################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $nodeObj, $msg, $host, $noDiscovery);
    $self = {};

    $self->{cluster_ip}             = undef;
    $self->{cluster_subnet}         = undef;
    $self->{cluster_name}           = undef;
    $self->{domain_name}            = undef;
    $self->{domain_admin}           = undef;
    $self->{domain_adminpass}       = undef;
    $self->{domain_controller_name} = undef;
    $self->{build}                  = undef;
    $self->{protocol}               = undef;
    $self->{skip_device_format}     = 0;
    $self->{diskGroups}             = ( );
    $self->{hyperVgroups}           = ( );
    $self->{clusterPrivData}        = ( );
    $self->{clusterPrivDataAge}     = "0";

    # need to bless this object
    bless $self, $class;

    # clear the errors first
    $self->_clear_err();

    foreach (keys %args){
        if (/^-?host/i) {
            if (not length($host = $args{$_}) ) {
                $self->_set_err("Invalid value for -host: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?protocol/i) {
            if (not length($self->{protocol} = $args{$_}) ) {
                $self->_set_err("Invalid value for -protocol: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?cluster_ip/i) {
            if (not length($self->{cluster_ip} = $args{$_})) {
                $self->_set_err("Invalid value for -cluster_ip: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?cluster_subnet/i) {
            if (not length($self->{cluster_subnet} = $args{$_})) {
                $self->_set_err("Invalid value for -cluster_subnet: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?cluster_name/i) {
            if (not length($self->{cluster_name} = $args{$_})) {
                $self->_set_err("Invalid value for -cluster_name: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?domain_name/i) {
            if (not length($self->{domain_name} = $args{$_})) {
                $self->_set_err("Invalid value for -domain_name: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?domain_admin$/i) {
            if (not length($self->{domain_admin} = $args{$_})) {
                $msg = "Invalid value for -domain_admin: $args{$_}";
                $self->_set_err($msg);
                return $self;
            }
            next;
        } elsif (/^-?domain_adminpass/i) {
            if (not length($self->{domain_adminpass} = $args{$_})) {
                $self->_set_err("Invalid value for -domain_adminpass: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?noDiscovery/i) {
            if (not length($noDiscovery = $args{$_})) {
                $self->_set_err("Invalid value for -noDiscovery: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?domain_controller_name/i) {
            if (not length($self->{domain_controller_name} = $args{$_})) {
                $self->_set_err("Invalid value for -domain_controller_name: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?skip_device_format/i) {
            if (not length($self->{skip_device_format} = $args{$_})) {
                $self->_set_err("Invalid value for -skip_device_format: $args{$_}");
                return $self;
            }
            next;
        }
    }

    if (not length($host)){
        $self->_set_err(-msg => "The required parameter -host was not supplied");
        return $ClientUtil::FALSE;
    }

    # Create Node object for the primary node
    $nodeObj = $self->create_node(-host => $host);
    if ($self->error()) {
        return $self;
    }

    # Put this nodeObj onto self
    push (@{$self->{nodes}}, $nodeObj);

    # Set Primary Node Attribute on the node object
    $self->set_primary_node(-host => $nodeObj->{hostname});
    if ($self->error()) {
        return $self;
    }

    # the build of the primary node will translate to the build of the cluster
    $self->{build} = $nodeObj->{sanClientObj}->{WinClient}->{build};

    # Now that we have a Node Object, lets see if the cluster exists already.
    # If the cluster already exists, collect the information about it.
    my $clusterStatus = $nodeObj->node_status();

    if ($nodeObj->error()) {
        $self->_set_err(@{$nodeObj->{errmsg}});
        return $self;
    }

    if (($clusterStatus == $ClientUtil::TRUE) and ($noDiscovery != $ClientUtil::TRUE)) {
        $self->sanlog(-msg => "A working cluster was detected.");

        # Discover cluster attributes
        $self->_load();
        if ($self->error()) {
            return $self;
        }

        $self->sanlog(-msg => "sanClusterObj creation completed. Cluster discovery completed.");
        return $self;
    }

    # Account for the two possible reasons to skip cluster discovery.
    if ($noDiscovery == $ClientUtil::TRUE) {
        $self->sanlog("The parameter -noDiscovery was supplied. If a working cluster exists " .
          "it has been ignored.");
    } else {
        $self->sanlog(-msg => "self creation completed successfully. No existing cluster " .
          "was detected. End SanCluster::_new.");
    }

    # The structure should be (mostly) full now, check for omissions on self.
    foreach my $key (keys %{$self}) {
        # Skip the non-mandatory items. These items may not exist within the cluster
        # and they are not needed to do the creation
        next if ($key =~ /diskGroups/i);
        next if ($key =~ /errmsg/i);
        next if ($key =~ /hyperVgroups/i);
        next if ($key =~ /clusterPrivData/i);

        if (not length($self->{$key})) {
            $self->_set_err("The required value $key is not defined");
        }
    }
    if ($self->error()) {
        $self->sanlog(@{$self->{errmsg}});
        return $self;
    }

    return $self;
}

###########################################################################
# Name: _load
# Input: -self, $self (called when and existing cluster is found)
# Output: True or False
# Tasks: * discovers attributes of a working cluster
#          - Does not recreate node objects.
#        * Creates objects for discovered devices
#        * Populates the self with available data
############################################################################
sub _load {
    my ($self, %args) = @_;
    my $headerLines = 4;
    my $nodeObj;

    $self->sanlog(-msg => "Loading cluster details on to sanClusterObj...");

    # cmd to determine cluster nodes
    my $cmd = 'cluster . node';
    my (@output) = $self->execute(-cmd => $cmd);

    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # The output looks like this:
    # Listing status for all available nodes:
    #
    # Node           Node ID Status
    # -------------- ------- ---------------------
    # CLUSTERTEST1         1 Up
    # CLUSTERTEST2         2 Up
    #
    # Go through the output but don't start collecting info
    # until we get past the header stuff. Its four lines.
    foreach (1 .. $headerLines) {
        shift (@output);
    }

    foreach my $line (@output) {
        # Finish when the return 0 is found
        last if ($line =~ /^0$/);

        # start splitting up the info
        my ($nodeName, undef, $nodeStatus) = split(/\s+/, $line);
        $nodeName = lc($nodeName);

        # The node objects don't change, so we don't need to recreate them, skip  creating
        # it if we have one already. All we need to do is make sure the deviceId is up to date.
        $nodeObj = $self->get_node(-host => $nodeName);

        # Check to see if our specified node was found. An error would indicate the node didn't exist
        if ($self->error()) {
            # We don't have an existing node object, and the node is up, we will create
            # a new object.
            $self->sanlog(-msg => "Found new cluster node $nodeName. Attempting to create node object.");
            $self->_clear_err;

            #Create Node object for node we found
            $nodeObj = $self->create_node(-host => $nodeName);
            if ( $self->error() ) {
                return $ClientUtil::FALSE;
            }

            # add this newly created object to the self.
            push(@{$self->{nodes}}, $nodeObj);

        } elsif ( $nodeStatus =~ /down/i) {
                $self->sanlog(-msg => "Cluster node $nodeName is down. Will not attempt to create " .
                  " a node object. Will not attempt to update deviceId maps");
        } else {
            $self->sanlog(-msg => "Using the existing nodeObj for $nodeName");
        }
    }

    # Update the self->deviceId_map. There is an existing cluster
    my $result = $self->_create_deviceId_map();

    if ($self->error()) {
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }
    $self->sanlog(-msg => "Completed cluster node Discovery. Starting cluster group discovery...");

    # Now discover all the groups.
    my @clusterGroupNames = $self->discover_groups();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # This loop creates and stores cluster group objects for each name found in discover_groups.
    foreach my $name (@clusterGroupNames) {
        # Get a new, empty diskGroupObject for each name.
        my $groupObj = $self->create_shared_disk_group();
        if ($self->error()) {
           return $ClientUtil::FALSE;
        }

        # Fill the object with pertinent data
        $groupObj->_load(-name => $name);
        if ($groupObj->error()) {
            $self->_set_err(@{$groupObj->{errmsg}});
            return $ClientUtil::FALSE;
        }

        # Get the resources associated with the group
        $groupObj->discover_resources();
        if ($groupObj->error()) {
            $self->_set_err(@{$groupObj->{errmsg}});
            return $ClientUtil::FALSE;
        }

        # Add the diskGroupObj to ourselves. We have two different storage bins for groups
        # at this time. We are not prepared to handle hyperV machine groups, so we are just seperating
        # HyperV groups into their own storage bin on self
        if ($groupObj->{groupType} =~ /HyperV/i) {
           push (@{$self->{hyperVgroups}}, $groupObj);
        } else {
           push (@{$self->{diskGroups}}, $groupObj);
        }
    }

    # discover a quorum disk
    $self->discover_quorum();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Cluster discovery completed");
    return $ClientUtil::TRUE;
}

#############################################################
# Name: create_node
# Input: -self (SanCluster object), -host
# Output: Node object
# Tasks: * Creates Node Object
#############################################################
sub create_node {
    my ($self, %args) = @_;
    my $host;

    foreach (keys %args){
        if (/^-?host/i) {
            if (not length($host = $args{$_}) ) {
                $self->_set_err("Invalid value for -host: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return {};
            }
            next;
       }
    }
    if (not length($host)){
        $self->_set_err(-msg => "The required parameter -host was not supplied");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "About to create node object for $host.");

    #Create node object
    my $nodeObj = San::SanClientCluster::Windows::MSCS::Node->_new(-parent => $self, -host => $host, -protocol => $self->{protocol});
    if ($nodeObj->error()) {
        $self->_set_err(@{$nodeObj->{errmsg}});
        return {};
    }

    $self->sanlog(-msg => "A new node object \"$host\" was created successfully");

    return $nodeObj;
}

#############################################################
# Name: _create_node_dc
# Input: -self (SanCluster object)
# Output: windows domain controller object
# Tasks: * check the SanClient object for a DC Node object
#        * Create a DC object if one does not exist
#        * Add DC object to SanCluster Object
#############################################################
sub _create_node_dc {
     my ($self, %args) = @_;
     my $dcName;
     # allow an alternate dc to be specified. Default to the
     # one on the self
     if (length($args{-dc})) {
        $dcName = $args{-dc};
     } else {
        $dcName = $self->{domain_controller_name};
     }

     # if the domain controller node object already exists for the name
     # that we specified, we will just return it.
     if ($self->{domain_controller}->{nodeObj}->{sanClientObj}->{WinClient}->{hostname} =~ /$dcName/) {
        return $self->{domain_controller};
     }

     my $nodeObj = $self->create_node(-host => $dcName);

     if ($self->error()) {
         # Ignore the message about no disks
         if (grep /No\s+cluster\s+disks\s+were\s+found/i, @{$self->{errmsg}}) {
             $self->sanlog(-msg => "Clearing and ignoring \"no disk\" error from node object creation. " .
                                   "This is because this is a Domain Controller and it is not expected " .
                                   "to have storage attached.");
             $self->_clear_err();
         } else {
             return $nodeObj;
         }
     }

     # add the Domain Controller node object to self
     # if the domain controller object already exists, overwrite it.
     # There will only be one.
     $self->{domain_controller} = $nodeObj;

     return $nodeObj;
}

#############################################################
# Name: _cluster_account_delete
# Input: -self (SanCluster object), -host
# Output: True or False
# Tasks: * Create/get the DC object with _create_node_dc
#        * Run the delete command on the DC
#############################################################
sub _cluster_account_delete {
    my ($self) = @_;

    # the number of seconds to wait for domain replication
    my $sleepTime = 45;

    # Create/get the SanClient Object for the domain controller.
    my $dcnodeObj = $self->_create_node_dc();
    if ( $self->error() ) {
        return $ClientUtil::FALSE;
    }

    my $host = $dcnodeObj->{sanClientObj}->{WinClient}->{hostname};

    # the account must be deleted from a domain controller. The domain controller is not a
    # cluster node. Cannot execute through $self->execute
    $self->sanlog(-msg => "Executing Deletion of $self->{cluster_name} on $host...");

    my $deleteCmd = "net computer \\\\$self->{cluster_name} /DEL";

    # The computer account may not be found. We call this routine from cluster_create and add_node to make sure
    # the account is deleted before starting, otherwise the create command will fail. In that case,
    # the account should not exist.
    my (@output) = $self->execute(-host => $host, -cmd => $deleteCmd);

    if ( $self->error()) {
        return $ClientUtil::FALSE;
    } elsif (grep /no\ssuch\scomputer:/, @output) {
        $self->sanlog(-msg => "The computer account $self->{cluster_name} did not exist");
        return $ClientUtil::TRUE
    } elsif (not grep /successfully/i, @output) {
       $self->_set_err("Deletion of the computer account $self->{cluster_name} " .
         "failed.");
       $self->sanlog(-msg => @{$self->{errmsg}});
       return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "The computer account \"$self->{cluster_name}\" was successfully deleted");
    # Need some time for this to replicate among the domain controllers...
    &SanUtil::nate_sleep($sleepTime);

    return $ClientUtil::TRUE;
}

################################################################
# Name: add_node
# Input: -self (SanCluster object), -host
# Output: returns True or False
# Tasks: * Creates node object by calling create_node()
#       TODO * Runs Validate_node() on new node object
#       * Adds node to cluster by running create_node()
#       * call create_shared_disk_group to create a new diskgroup for the node
#       * update the deviceIDmap on $self
################################################################
sub add_node {
    my ($self, %args) = @_;
    my $host;

    foreach (keys %args){
        if (/^-?host/i) {
            if (not length($host = $args{$_}) ) {
                $self->_set_err("Invalid value for -host: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (not length($host)){
        $self->_set_err(-msg => "The required parameter -host was not supplied");
        return $ClientUtil::FALSE;
    }

    # call create_node to create the nodeObj
    my $nodeObj = $self->create_node(-host => $host);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Clean the cluster node, just in case. Existing cluster data will prevent success.
    # make sure to specify the -saveObj param. If you don't it will try to delete the node
    # from the cluster object, but error out because the node is not yet part of the cluster.
    $nodeObj->remove(-saveObj => $ClientUtil::TRUE);
    if ($nodeObj->error()) {
        $self->_set_err(@{$nodeObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Call the commands to add to the cluster
    $nodeObj->add();
    if ($nodeObj->error()) {
        $self->_set_err(@{$nodeObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Put the newly created node Object on the self
    push (@{$self->{nodes}}, $nodeObj);

    # Create a new, empty disk group object
    my $groupObj = $self->create_shared_disk_group();
    if ($self->error()) {
       return $ClientUtil::FALSE;
    }

    # Create a shared disk Group on the cluster using the object above
    $groupObj->create_group(-name => $nodeObj->{hostname}, -owner => $nodeObj->{hostname});
    if ($groupObj->error()) {
       $self->_set_err(@{$groupObj->{errmsg}});
       return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "A corresponding cluster group for node $host was created successfully");
    return $ClientUtil::TRUE;

}

#################################################################
# Name: get_nodes
# Input: -self (SanCluster object)
# Output: array of node objects
# Tasks: * Reads node attribute on SanCluster Object
#################################################################
sub get_nodes {
    my ($self) = @_;

    if (not scalar(@{$self->{nodes}}) > 0) {
        $self->_set_err("There are no existing node objects on the san cluster object.");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return ();
    }

    return @{$self->{nodes}};
}

#################################################################
# Name get_node
# Input: -self (SanCluster object), -host
# Output: returns single node object
# Tasks: * Reads node attribute on SanCluster Object
#################################################################
sub get_node {
    my ($self, %args) = @_;
    my $host;
    my $obj;

    foreach (keys %args) {
        if (/^-?host/i) {
            if (not length($host = $args{$_}) ) {
                $self->_set_err("Invalid value for -host: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return {};
            }
            next;
        }
    }
    if (not length($host)){
        $self->_set_err(-msg => "The required parameter -host was not supplied");
        return $ClientUtil::FALSE;
    }

    # Get all the cluster nodes and then check for errors.
    my @nodes = $self->get_nodes();
    if ($self->error()) {
         return {};
    }

    foreach $obj (@nodes){
        if ($obj->{hostname} =~ /$host/i){
            return $obj;
        }
    }

    $self->_set_err("The object could not be found for the supplied node name \"$host\"");
    $self->sanlog(-msg => @{$self->{errmsg}});

    return {};
}

################################################################
# Name: get_primary_node
# Input: -self (SanCluster object)
# Output: returns single node object for primary_node
# Tasks: * Reads node attribute on SanCluster object
#        * Determines which node is set as primary
################################################################
sub get_primary_node {
    my ($self) = @_;

    my @nodes = $self->get_nodes;
    if ($self->error()) {
         return {};
    }

    foreach my $obj (@nodes){
       if ( $obj->{primary_cluster_node} ==  $ClientUtil::TRUE ){
           return $obj;
       }
    }
    $self->_set_err("A primary node could not be located.");
    $self->sanlog(-msg => @{$self->{errmsg}});

    return {};
}

################################################################
# Name: get_disk_groups
# Input: -self (SanCluster object)
# Output: returns an array of cluster group objects
# Tasks: * Find and return the array of diskGroup objects
################################################################
sub get_disk_groups {
    my ($self) = @_;

    if (not $self->{diskGroups}) {
        return ();
    }

    return @{$self->{diskGroups}};
}

################################################################
# Name: get_disk_group
# Input: -self (SanCluster object), -name
# Output: returns a single disk group object given the groups name
# Tasks: * Find and return diskGroup object matching $diskGroupObj->{name} = -name
################################################################
sub get_disk_group {
    my ($self, %args) = @_;
    my $groupName;
    my $diskGroupObj;

    foreach (keys %args){
        if (/^-?name/i) {
            if (not length($groupName = $args{$_}) ) {
                $self->_set_err("Invalid value for -name: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return {};
            }
            next;
        }
    }
    if (not length($groupName)){
        $self->_set_err(-msg => "The required parameter -name was not supplied");
        return $ClientUtil::FALSE;
    }

    my @diskGroups = $self->get_disk_groups();

    foreach $diskGroupObj (@diskGroups) {
        if ($diskGroupObj->{name} =~ /$groupName/i) {
            return $diskGroupObj;
        }
    }
    $self->_set_err("Could not find an object for the specified group \"$groupName\".");
    $self->sanlog(-msg => @{$self->{errmsg}});
    return {};
}

################################################################
# Name: set_primary_node
# Input: -self (SanCluster object), -host
# Output: returns True or False
# Tasks: * Calls $nodeObj = get_node(-node)
#        * Calls get_primary_node() and unsets primary node attribute
################################################################
sub set_primary_node {
    my ($self, %args) = @_;
    my $host;

    foreach (keys %args){
        if (/^-?host/i) {
            if (not length($host = $args{$_}) ) {
                $self->_set_err("Invalid value for -host: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return $ClientUtil::FALSE;
            }
            next;
        }
    }
    if (not length($host)){
        $self->_set_err(-msg => "The required parameter -host was not supplied");
        return $ClientUtil::FALSE;
    }

    # Get the node object for the -host option
    my $nodeObj = $self->get_node(-host => $host);

    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Determine the current primary node
    # Don't worry too much about errors here, because if this is the first node to be created,
    # there will be no primary node.
    my $currentPrimaryNode = $self->get_primary_node();

    if ($self->error()) {
        $self->sanlog(-msg => "There was an error in get_primary node. This is likely returned because as the cluster " .
          "is being created or discovered, the first node will not yet be set as the primary node. The first node in a " .
            "newly created or discovered cluster will be given the \"Primary Node\" designation.\n" .
              "Ignoring error message: @{$self->{errmsg}}");
        # Clear the error so it doesn't interfere later
        $self->_clear_err();
    }
    #Unset Current primary node, if there was a previous primary
    if (length($currentPrimaryNode->{primary_cluster_node})) {
        $currentPrimaryNode->{primary_cluster_node} = $ClientUtil::FALSE;
    }

    #Set Primary Node Attribute on the new primary node object
    my $status = $nodeObj->set_primary_node();

    if ($nodeObj->error()) {
        $self->_set_err(@{$nodeObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "The primary cluster node has been set to $nodeObj->{hostname}");
    return $ClientUtil::TRUE;
}

################################################################
# Name: cluster_create
# Input: -self (SanCluster object)
# Output: returns True or False
# Tasks: * Act as a wrapper to _cluster_create
#        * makes sure that the rsh daemon is running the appropriate context
#        * changes the rsh daemon as necessary and resets when complete
################################################################
sub cluster_create {
    my ($self, %args) = @_;

    my $nodeObj = $self->get_primary_node();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }
    my $WinClientObj = $nodeObj->{sanClientObj}->{WinClient};
    my $systemAccount = "system";
    my $validUserAccount = "administrator";
    my @errorCollector;

    # Make sure that commands on the system are being executed as any user. i.e. not system
    my $commandContext = $WinClientObj->wrshdnt_check_context();

    if (not length($commandContext)) {
        $self->_set_err("There was a problem determining the current rsh context on $WinClientObj->{hostname}");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "The context of the rsh daemon is found to be $commandContext");

    if ($commandContext =~ /$systemAccount/i) {
        # Modify the wrshdnt service
        $WinClientObj->wrshdnt_change_context( $validUserAccount );
        # Check for errors, do not even bother calling _cluster_create if we cannot guarantee the context of rshd
        if ($WinClientObj->error()) {
            $self->_set_err(@{$WinClientObj->{errmsg}});
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
    }
    # Now that the rsh service is correct, create the cluster
    $self->_cluster_create();

    # Check the errors, but do not return when an error is found, just note that we found one.
    # We may need to reset rshd before exiting this method.
    if ($self->error()) {
        $self->sanlog(-msg => @{$self->{errmsg}});
        push(@errorCollector, "The cluster was not created successfully");
        $self->_clear_err();
    }
    # change the rsh service context back to system
    if ($commandContext =~ /"system"/) {
        my $status = $WinClientObj->wrshdnt_change_context($systemAccount);
        if ($WinClientObj->error()) {
            $self->sanlog(-msg => @{$self->{errmsg}});
            push(@errorCollector, "There was a problem resetting the RSH service context to \"system\".\n " .
                                  "It may be necessary to manually reset the service context.");
            $self->_set_err(@errorCollector);
            return $ClientUtil::FALSE;
        }
    } elsif (scalar(@errorCollector) >= 1) {
            $self->_set_err(@errorCollector);
            return $ClientUtil::FALSE;
   }

    # Add the build to the cluster object.
    $self->{build} = $nodeObj->{sanClientObj}->{WinClient}->{build};

    # Discover cluster attributes to fill the cluster object with the newly created cluster.
    $self->_load();
    if ($self->error()) {
        # This error is normal if the cluster was just created
        if (grep /No\s+cluster\sdisks\s+were\s+found/i, @{$self->{errmsg}}) {
            $self->_clear_err();
        } else {
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "The cluster_create process completed successfully");

    return $ClientUtil::TRUE;
}

################################################################
# Name: _cluster_create
# Input: -self (SanCluster object)
# Output: returns True or False
# Tasks: * Check for a previously existing cumputer account for the cluster name.
#        * Calls cluster_validate
#        * Calls system commands to create cluster
#        * call create_shared_disk_group to create a new diskgroup for the node
################################################################
sub _cluster_create {
    my ($self) = @_;

    my $nodeObj = $self->get_primary_node();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }
    my $WinClientObj = $nodeObj->{sanClientObj}->{WinClient};

    # Delete an existing cluster name account on the domain controller.
    # Then clean the current node to get rid of any old cluster data
    # Failures from this method are not necessarily fatal. We will just sanlog
    # and rely and the output of the cluster create command to tell us if there were
    # really problems creating the cluster.
    $self->_cluster_account_delete();
    if ($self->error()) {
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # cluster node cleanup
    $nodeObj->remove(-saveObj => $ClientUtil::TRUE);
    if ($nodeObj->error()) {
        $self->_set_err(@{$nodeObj->{errmsg}});
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Hear ye, hear ye....
    $self->sanlog(-msg => "Attempting creation of cluster $self->{cluster_name} " .
                          "on node $nodeObj->{hostname}");

    # Line up a command to create the cluster its a little different for 2k8 and up
    my $clusterCreateCommand;
    if ($WinClientObj->{build} >= $WIN2K8BUILD) {
        # 2k8
        $nodeObj->sanlog(-msg => "Creating a Windows 2k8 \(or higher\) cluster");

        $clusterCreateCommand = 'cluster /cluster:' . $self->{cluster_name} . ' /create' .
                                ' /ipaddr:' .  $self->{cluster_ip} . '/' . $self->{cluster_subnet};
    } else {
        # 2k3
        $nodeObj->sanlog(-msg => "Creating a Windows 2k3 cluster");

        $clusterCreateCommand = 'cluster /cluster:' . $self->{cluster_name} . ' /create' .
                                ' /node:' .  $nodeObj->{hostname} .
                                ' /user:' . $self->{domain_name} . '\\' . $self->{domain_admin} .
                                ' /pass:' . $self->{domain_adminpass} .
                                ' /ipaddr:' . $self->{cluster_ip} .
                                ' /min /unattend';
    }

    # Execute the command to create the cluster on the primary cluster node
    my @output = $self->execute(-cmd => $clusterCreateCommand);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }
    # Check for errors in the output
    if (grep /error/i, @output) {
       $self->_set_err("The cluster was not created successfully.");
       return $ClientUtil::FALSE;
    }

    # Get a blank disk group object
    my $groupObj = $self->create_shared_disk_group();
    if ($self->error()) {
       return $ClientUtil::FALSE;
    }

    # Create a shared disk Group on the cluster using the object above
    $groupObj->create_group(-name => $nodeObj->{hostname}, -owner => $nodeObj->{hostname});
    if ($groupObj->error()) {
       $self->_set_err(@{$groupObj->{errmsg}});
       return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "A disk group was created for the first cluster node.");

    return $ClientUtil::TRUE;
}

################################################################
# Name: cluster_enable
# Input: -self (SanCluster object)
# Output: returns True or False
# Tasks: * calls system commands to enable cluster
################################################################
sub cluster_enable {
    my ($self) = @_;

    my $startCommand = 'sc start "ClusSvc"';
    my $queryCommand = 'sc query "ClusSvc"';
    my $correctState = "RUNNING";
    my $queryRetries = 3;
    my $queryRetriesCount = 0;
    my $startRetries = 3;
    my $startRetriesCount = 0;
    my $serviceStartTimeout = 20;
    my @output;
    my $errorCondition = $ClientUtil::FALSE;
    my @errorCollection;

    my @nodes = $self->get_nodes();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    foreach my $nodeObj (@nodes){

       my $host = $nodeObj->{sanClientObj}->{WinClient}->{hostname};

       $nodeObj->sanlog(-msg => "Starting the cluster service on $host");

       # We attempt to start the service a total of $startRetriescount times. For each attempt to start the service
       # a second until loop is entered to query the service $queryRetriesCount times, until the service successfully
       # enters the running state. The idea here is to cut down on the amount of time required to wait to ensure the
       # service is in the proper state. Since this is intended to operate on every cluster node in turn, the errors are
       # all saved up until the routine has executed on every node in the cluster.

       # Query the initial state of the service
       (@output) = $self->execute(-cmd => $queryCommand, -host => $host);
        if (grep(/$correctState/i, @output)) {
            $self->sanlog(-msg => "The service is already in the $correctState state. There is nothing to do on $host");
            next;
        }

        # This loop controls the attempts to actually start the service. This loop will be repeated $startRetries times
        until (grep(/$correctState/i, @output)) {
            $startRetriesCount++;

            # Reset the query count to 0 for each start attempt. Each time we try to start the service,
            # we query the states $queryRetriesCount
            $queryRetriesCount = 0;
            # Exit the loop when the retry count expires
            if ($startRetriesCount > $startRetries){
                $self->sanlog(-msg => "The cluster service failed to start on $host.");
                push (@errorCollection, "The cluster service failed to start on $host.\n");
                $errorCondition = $ClientUtil::TRUE;
                next;
            }

            $self->sanlog(-msg => "Attempt $startRetriesCount to start Cluster Service on $host");

            (@output) = $self->execute(-cmd => $startCommand, -host => $host);

            # Not concerned with errors returning from execute, because we are going to retry, we will also
            # deal with this condition by searching the output for the expected text $correctState
            if ($self->error()) {
                $self->_clear_err();
            }

            # For each attempt at starting, this loop will attempt to query the service and find a "running" state
            # if the retry count is exhausted control is returned to the outer loop to attempt the service start again
            # until the retry count is exhausted
            until (grep(/$correctState/i, @output)){

                $queryRetriesCount++;
                # Give the Service some seconds to start
                &SanUtil::nate_sleep($serviceStartTimeout);

                (@output) = $self->execute(-cmd => $queryCommand, -host => $host);

                if ($self->error()) {
                    # we are just going to retry and look again. Clear the error
                    $self->_clear_err();
                }

                # Exit the until query loop when the query retry count expires or the service reports STARTED
                last if (($queryRetriesCount == $queryRetries) or (grep(/$correctState/i, @output)));
            }
            # Exit the while loop when the retry count is expired and the service is not reported as started
            if (($startRetriesCount >= $startRetries) and (not grep (/$correctState/i, @output))) {
                $self->sanlog(-msg => "The cluster service failed to start on $host after $startRetriesCount tries.");
                push (@errorCollection, "The cluster service failed to start on $host after $startRetriesCount tries.");
                $errorCondition = $ClientUtil::TRUE;
                next;
            }
        }
        if ($errorCondition == $ClientUtil::FALSE) {
            $self->sanlog(-msg => "Successfully enabled cluster service on all hosts");
            return $ClientUtil::TRUE;
        } else {
            $self->_set_err("Did NOT Successfully enable cluster service on all hosts\n@errorCollection");
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::FALSE;
       }
    }
}

################################################################
# Name: cluster_disable
# Input: -self (SanCluster object)
# Output: returns True or False
# Tasks: * calls system commands to disable cluster
################################################################
sub cluster_disable {
    my ($self) = @_;

    my $stopCommand = 'sc stop "ClusSvc"';
    my $queryCommand = 'sc query "ClusSvc"';
    my $correctState = "STOPPED";
    my $queryRetries = 3;
    my $queryRetriesCount = 0;
    my $stopRetries = 3;
    my $stopRetriesCount = 0;
    my $serviceStopTimeout = 5;
    my @output;
    my $errorCondition = $ClientUtil::FALSE;
    my @errorCollection;

    my @nodes = $self->get_nodes();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    foreach my $nodeObj (@nodes){

        my $host = $nodeObj->{sanClientObj}->{WinClient}->{hostname};

       # Query the initial state of the service
       (@output) = $self->execute(-cmd => $queryCommand, -host => $host);
        if (grep(/$correctState/i, @output)) {
            $self->sanlog(-msg => "The service is already in the $correctState state. There is nothing to do on $host");
            next;
        }

        # We attempt to start the service a total of $startRetriescount times. For each attempt to start the service
        # a second until loop is entered to query the service $queryRetriesCount times, until the service successfully
        # enters the running state. The idea here is to cut down on the amount of time required to wait to ensure the
        # service is in the proper state. Since this is intended to operate on every cluster node in turn, the errors are
        # all saved up until the routine has executed on every node in the cluster.

        # This loop controls the attempts to actually stop the service.
        until (grep /$correctState/, @output) {
            $stopRetriesCount++;

            # Reset the query count to 0 for each stop attempt. Each time we try to stop the service,
            # we query the states $queryRetriesCount
            $queryRetriesCount = 0;
            # Exit the loop when the retry count expires
            if ($stopRetriesCount > $stopRetries) {
                $self->sanlog(-msg => "The cluster service failed to stop on $host after $stopRetriesCount tries.");
                push (@errorCollection, "The cluster service failed to stop on $host after $stopRetriesCount tries.");
                $errorCondition = $ClientUtil::TRUE;
                next;
            }

            $nodeObj->sanlog(-msg => "Attempt $stopRetriesCount to start Cluster Service on $host");

            (@output) = $self->execute(-cmd => $stopCommand, -host => $host);

            # We only check the return status, because we are looking for certain keywords in the output
            if ($self->error()){
                $self->_clear_err();
            }

            # For each attempt at stopping, this loop will attempt to query the service and find a "stopped" state
            # if the retry count is exhausted control is returned to the outer loop to attempt to stop the service again.
            # Reset the query count to 3 for each start attempt. Each time we try to start the service,
            # we query the states $queryRetriesCount
            until (grep /$correctState/, @output){

                $queryRetriesCount++;

                # Give the Service some seconds to stop
                &SanUtil::nate_sleep($serviceStopTimeout);

                (@output) = $self->execute(-cmd => $queryCommand, -host => $host);

                # Not concerned with errors returning from execute, because we are going to retry, we will also
                # deal with this condition by searching the output for the expected text $correctState
                if ($self->error()){
                    $self->_clear_err();
                }
                # Exit the while loop when the retry count expires
                last if (($queryRetriesCount == $queryRetries)or(grep /$correctState/, @output));
            }
            # Exit the while loop when the retry count is expired and the service is not reported as stopped
            if (($stopRetriesCount >= $stopRetries)&&(not grep /$correctState/, @output)){
                $self->sanlog(-msg => "The cluster service failed to stop on $host after $stopRetriesCount tries." .
                                       " Lets just brutally kill it now.");
                my $killstatus = &ClientUtil::kill_process_instances($host, "clussvc.exe");

                if ($killstatus != 0) {
                    push (@errorCollection, "The cluster service failed to stop on $host after $stopRetriesCount tries." .
                                            " Furthermore, clussvc.exe could not be killed with taskkill.exe");
                    $errorCondition = $ClientUtil::TRUE;
                }
                next;
            }
        }
        if ($errorCondition == $ClientUtil::FALSE) {
            $self->sanlog(-msg => "Successfully disabled cluster service on all hosts");
            return $ClientUtil::TRUE;
        } else {
            $self->_set_err("Did NOT Successfully disable cluster service on all hosts\n@errorCollection");
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::FALSE;
       }
    }
}

#########################################
# Name: cluster_destroy
# Input: -self (SanCluster object)
# Output: returns True or False
# Tasks: * calls system commands to destroy cluster on each node
#        * Remove the cluster account from the windows domain controller
#########################################
sub cluster_destroy {
    my ($self) = @_;

    my @nodes = $self->get_nodes();
    my $globalError = $ClientUtil::FALSE;
    my $nodeError = $ClientUtil::FALSE;

    foreach my $nodeObj (@nodes){

        my $host = $nodeObj->{sanClientObj}->{WinClient}->{hostname};

        # Clean the node using cluster.exe
        $nodeObj->remove();

        # We are going to try to clean all the nodes, before we return anything
        # just keep track of the errors
        if ($nodeObj->error()) {
            $nodeObj->sanlog(-msg => @{$nodeObj->{errmsg}});
            $nodeError = $ClientUtil::TRUE;
            $globalError = $ClientUtil::TRUE;
        }

        if ($nodeError == $ClientUtil::FALSE) {
            # We completed properly on this node
            $nodeObj->sanlog(-msg => "Cluster_destroy succeeded on $host.");
        }
    }

    if ($globalError == $ClientUtil::FALSE) {
        $self->sanlog(-msg => "cluster_destroy completed on all nodes. Removing $self->{cluster_name} " .
        " domain account...");

    } elsif ($nodeError|$globalError != $ClientUtil::TRUE) {
        $self->_set_err("Cluster_destroy was not entirely successful in cleaning all cluster nodes.");
        $self->sanlog(-msg => @{$self->{errmsg}});
    }

    # Delete an existing cluster name account on the domain controller.
    my $status = $self->_cluster_account_delete();

    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "cluster_destroy completed successfully");
    return $ClientUtil::TRUE;
}

###########################################################################
# Name: add_device
# Input:  -self, cluster object
# Output: returns True or False
# Tasks: * Works on primary node
#        * init each device
#        * format each device
#        * assign drive letter or mount point (if necessary)
#        * add device to storage group
#        * add WinClient disk object to self
############################################################################
sub add_device {
    my ($self, %args) = @_;
    my $nodeName;

    foreach (keys %args){
        if (/^-?node/i) {
            if (not length($nodeName = $args{$_}) ) {
                $self->_set_err("Invalid value for -node: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return ();
            }
            next;
        }
    }

    # If we don't get a client name to use, use the primary node
    my $nodeObj;
    if (not length($nodeName)) {
        $nodeObj = $self->get_primary_node();
    } else {
        $nodeObj = $self->get_node(-host => $nodeName);
    }

    my $sanClientObj = $nodeObj->get_sanclient_object();
    my $WinClientObj = $nodeObj->_get_winClient();

    # get the attached devices on current node
    my @diskArray = $nodeObj->get_devices();
    if ($nodeObj->error()) {
        $self->_set_err(@{$nodeObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    my @dev_names = ();
    foreach my $devObj (@diskArray) {
        push (@dev_names, $devObj->{name});
    }
    my $msg = join("\n", @dev_names);
    $self->sanlog(-msg => "List of discovered raw devices includes:\n$msg");

    # format and mount the devices within the @dev_names array
    # If the devices already exist formatted and mounted, one could supply
    # the parameter SKIP_DEVICE_FORMAT to SanClientCluster.thpl
    if (not $self->{skip_device_format}) {
        $sanClientObj->format(-devices => \@dev_names);
    }

    # We just added a whole bunch of new devices to the WinClient object of
    # the current node. Now, we need to update the deviceId maps. If this is 2008,
    # The cluster doesn't exist yet, so we need to specify the no -nocluster option
    # so that we look at Win32_DiskPartition class, as if we were Win2k3.
    my $result;
    if ($nodeObj->{sanClientObj}->{WinClient}->{build} >= $WIN2K8BUILD) {
         # 2k8 and up. There is no cluster yet, cannot use MSCluster_Disk yet.
        $result = $self->_create_deviceId_map(-nocluster => $ClientUtil::TRUE);
    } else {
        # 2k3
        $result = $self->_create_deviceId_map();
    }
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Next, add all the devices to the group that belongs to this node
    # We need to determine which group to add the devices to. $nodeObj->get_preferred_groups
    # will return all the group objects with this node as the preferred owner.
    my $groupName;
    my @groupObjs = $nodeObj->get_preferred_groups();
    if ($nodeObj->error()) {
        $self->_set_err(@{$nodeObj->{errmsg}});
        return $ClientUtil::FALSE;
    }
    # If more then one group is found, do some decision making.
    # If we get more then one group, start with the first group in the list.
    # If we find one with ^NATE_ in the name, take that one. If more then one
    # group fits that description, the first one to meet said criteria will be chosen.
    my $tempObj = shift(@groupObjs);
    $groupName = $tempObj->{name};
    if (scalar(@groupObjs > 1)) {
        foreach my $groupObj (@groupObjs) {
            if ($groupObj->{name} =~ /^NATE_/i) {
                $groupName = $groupObj->{name};
                last;
            }
        }
    }

    # Now that we have a name for the group, get the group object for the group we are going to populate
    $self->sanlog(-msg => "Populating the group \"$groupName\" with disks discovered on $nodeObj->{hostname}");

    foreach my $diskNum (@dev_names) {
        my $diskObj = $nodeObj->{sanClientObj}->{WinClient}->{disks}{$diskNum};
        # Check for the disk signature
        if (not length($diskObj->{sig})) {
            $self->sanlog("Attempting to acquire the disk signature for disk $diskObj->{mount}.");
            # Attempt to gather the signature with getDiskInfo if we still don't have it.
            $diskObj->getDiskInfo();
            if ($diskObj->error()) {
                $self->_set_err(@{$diskObj->{errmsg}});
                $self->sanlog(-msg => @{$self->{errmsg}});
                return $ClientUtil::FALSE;
            # Thats bad if we still don't know the disk signature
            } elsif (not length($diskObj->{sig})) {
                $self->_set_err("Unable to determine the disk signature for the disk $diskObj->{mount}.");
                $self->sanlog(-msg => @{$self->{errmsg}});
                return $ClientUtil::FALSE;
            }
        }
        # Make sure the disk is not offline. If it is offline skip the disk.
        if ($diskObj->{status} =~ /offline/i) {
            $self->_set_err("The disk number $diskNum is offline. Disk number $diskNum has not been added " .
                            "to the cluster.");
            $self->sanlog(-msg => @{$self->{errmsg}});
            next;
        }
        # make sure that WinClient has the disk mount.
        if (not length($diskObj->{mount})) {
            $self->_set_err("The WinClient object does not have the mount point data for disk $diskNum\n" .
                            "$diskObj->{mount}. Check the status of the disks on host $WinClientObj->{hostname}");
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }

        # add this disk to the storage group
        $self->_add_disk_to_storage_group(-diskId    => $diskObj->{sig},
                                          -groupName => $groupName,
                                          -diskName  => $diskObj->{filer} . "_Disk" . $diskObj->{mount});
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # make sure that all the resource objects we just created have the $diskObj->{mnt_pt}
    # populated. This routine will try to populate the data if it doesn't already exist.
    $self->_find_cluster_disk_mounts();
    if ($self->error()) {
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   create_shared_disk_group
# Input:  -self self, -name, -owner
# Output: shared group object
# Tasks:  call SharedDiskGroup->_new to create the object
###########################################################################
sub create_shared_disk_group {
    my ($self) = @_;

    # Get a blank $diskGroupObj
    my $diskGroupObj = San::SanClientCluster::Windows::MSCS::SharedDiskGroup->_new(-parent => $self);
    if ($diskGroupObj->error()) {
        $self->_set_err(@{$diskGroupObj->{errmsg}});
        return $diskGroupObj;
    }

    return $diskGroupObj;
}

###########################################################################
# Name:   _add_disk_to_storage_group
#         This function should ONLY BE CALLED from add_device
# Input:  -self self,
#         -diskId the unique ID of the disk, can be signature or GUID
#         -diskName the name to assign the disk
#         -group group to add the disk to
# Output: True or False
# Tasks:  * Create the cluster disk res
#         * Map the physical disk to the cluster resource
#         * Update $self->{shared_device_resources
###########################################################################
sub _add_disk_to_storage_group {
    my ($self, %args) = @_;
    my $diskId;
    my $groupName;
    my $diskName;

    foreach (keys %args){
        if (/^-?diskId/i) {
            $diskId = $args{$_};
            next;
        } elsif (/^-?groupName/i) {
            $groupName = $args{$_};
            next;
        } elsif (/^-?diskName/i) {
            $diskName = $args{$_};
            next;
        }
    }

    if ((not length($diskId)) or (not length($groupName)) or (not length($diskName))) {
        $self->_set_err("The function did not receive all the required parameters.\n" .
                        "The following values were received:\n" .
                        "-diskId=$diskId\n-groupName=$groupName\n-diskName=$diskName");
        $self->sanlog(@{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Get the group object that the resource will belong to
    my $diskGroupObj = $self->get_disk_group(-name => $groupName);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Call the create method to get a blank resource object
    my $resourceObj = $diskGroupObj->create_resource(-name => $diskName, -id => $diskId );
    if ($diskGroupObj->error()) {
        $self->_set_err(@{$diskGroupObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Now call the create create_disk method on the resource object, to actually create resource on the cluster
    my $result = $resourceObj->create_disk(-name => $diskName, -id => $diskId );
    if ($resourceObj->error()) {
        $self->_set_err(@{$resourceObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # put the new resourceObj on the corresponding GroupObj
    push (@{$diskGroupObj->{resources}}, $resourceObj);

    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   execute
# Input:  -self self, -cmd The command to execute
# Output: an array of command output
# Tasks:  * Get the primary cluster node
#         * execute a command on the primary cluster node through $ClientUtil
###########################################################################
sub execute {
    my ($self, %args) = @_;

    my $host;
    my $cmd;
    my $timeout;
    my ($status, @output);

    foreach (keys %args){
        if (/^-?cmd/i) {
            if (not length($cmd = $args{$_}) ) {
                $self->_set_err("Invalid value for -cmd: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return ();
            }
            next;
        } elsif (/^-?host/i) {
            # -host is an optional parameter that allows execution on nodes
            # other then the default primary node.
           if (not length($host = $args{$_}) ) {
                $self->_set_err("Invalid value for -host: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return ();
            }
            next;
        } elsif (/^-?timeout/i) {
            # -timeout is an optional parameter that allows execution on nodes
            # other then the default primary node.
           if (not length($timeout = $args{$_}) ) {
                $self->_set_err("Invalid value for -timeout: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return ();
            }
            next;
        }
    }
    if (not length($cmd)){
        $self->_set_err(-msg => "The required parameter -cmd was not supplied");
        return $ClientUtil::FALSE;
    }

    # Get the node object. Default to the primary node
    my $nodeObj;
    # First check to see if this is the defined domain controller for the cluster
    if ($host =~ /$self->{domain_controller_name}/i) {
        # The domain controller object is not stored with the cluster nodes
        $nodeObj = $self->{domain_controller};
    } elsif (length($host)) {
        # Get the requested node, if defined.
        $nodeObj = $self->get_node(-host => $host);
    } else {
        # Get the primary node, if no other host is specified
        $nodeObj = $self->get_primary_node();
    }

    # Execute command with clientUtil, and report status
    ($status, @output) = &ClientUtil::execute($nodeObj->{sanClientObj}->{conn}, $cmd, $timeout);
    if ($status != 0) {
        $self->_set_err("A bad status was returned from ClientUtil::execute:\n@output");
    }
    return @output;
}

###########################################################################
# Name:   create_quroum
# Input:  -self
# Output: True or False
# Tasks:  * Find a disk to use as the quorum and add to Cluster Group
#         * Update the cluster group with the new resources
#         * Set the quorum attribute
###########################################################################
sub create_quorum {
    my ($self) = @_;

    my $diskSerial;
    my $lastDiskSerial;
    my $quorumSerial;

    # Check and make sure we don't already have a quorum disk
    # This check does not look for a local quorum
    if (length($self->{quorum})) {
        $self->_set_err("A quorum disk \"$self->{quorum}\" is already defined " .
                        "for this cluster");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Get the current cluster disks
    my @resourceObjs = $self->get_disks();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    my $resourceObj;
    # Configure a quorum disk
    # If a disk is mounted at Q, that will be the quorum...
    # Otherwise we just take the last one in the loop
    foreach my $Obj (@resourceObjs) {
        if ($Obj->{mount_pt} =~ /Q/i) {
            $resourceObj = $Obj;
            last;
        }
        if ($Obj->{name} =~ /$resourceObjs[-1]->{name}/) {
            $resourceObj = $Obj;
        }
    }

    # Bail out now if we still don't have a quorum.
    if (not length($resourceObj->{name})) {
        $self->_set_err("A suitable quorum disk was not found.");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Name the disk
    my $diskName = "Quorum-Disk" . $resourceObj->{mount_pt};

    $self->sanlog(-msg => "The disk mounted at $resourceObj->{mount_pt} will be " .
                          " the cluster quorum. The disk will be added to the group " .
                          "\"Cluster Group\" as $diskName");

    my $status = $resourceObj->rename(-name => $diskName);
    if ($resourceObj->error()) {
        $self->_set_err(@{$resourceObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # The resource has to be offline before it can be moved.
    $status = $resourceObj->offline(-name => $diskName);
        if ($resourceObj->error()) {
        $self->_set_err(@{$resourceObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $status = $resourceObj->move(-group => "Cluster Group");
    if ($resourceObj->error()) {
        $self->_set_err(@{$resourceObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $status = $resourceObj->online(-name => $diskName);
        if ($resourceObj->error()) {
        $self->_set_err(@{$resourceObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Cluster Quorum $diskName added to \"Cluster Group\".");

    # Tell the cluster that this disk is now the quorum
    my $cmd = 'cluster . /quorum:"' . $diskName . '" && cluster . /quorum';
    my (@output) = $self->execute(-cmd => $cmd);

    if ((grep /error/i, @output) or (not grep /^$diskName/i, @output)) {
        $self->_set_err("The quorum resource does not appear to have been created correctly.");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    } elsif ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "The cluster quorum is now set to \"$diskName\".");

    $self->discover_quorum();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Create_quorum completed successfully.");
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   destroy_quorum
# Input:  -self
# Output: True or False
# Tasks:  * Set the quorum attribute to Node Majority
#         * Delete the disk.
###########################################################################
sub destroy_quorum {
    my ($self) = @_;

    my $diskSerial;
    my $lastDiskSerial;
    my $quorumSerial;

    # Get the quorum disk, if there is one, otherwise bail.
    my $qdisk = $self->get_quorum();
    if ($self->error()) {
        if ($self->errmsg() =~ /A quorum disk was not found/) {
            $self->_clear_err();
            $self->sanlog(
                msg => "No quorum disk was found skipping cleanup step."
            );
        } else {
            return $ClientUtil::FALSE;
        }
    }

    # Set the quorum attribute to node majority:
    my $cmd = 'cluster . /quorum /NODE';
    my (@output) = $self->execute(-cmd => $cmd);
    if ((grep /error/i, @output) or (not grep /Node Majority/i, @output)) {
        $self->_set_err("The quorum resource does not appear to have been set correctly.");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    } elsif ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Finish removing the disk from the cluster.
    $qdisk->offline();
    if ($qdisk->error()) {
        return $ClientUtil::FALSE;
    }
    $qdisk->destroy();
    if ($qdisk->error()) {
        return $ClientUtil::FALSE;
    }

    # Remove the disk object from the Cluster Group object
    my $group = $self->get_disk_group(name => 'Cluster Group');
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }
    my $index = 0;
    foreach my $resource (@{$group->{resources}}) {
        if ($resource->{name} eq $qdisk->{name}) {
            $self->sanlog(
                -msg => "Removing " . $qdisk->{name} . " from Cluster Group."
            );
            splice(@{$group->{resources}},$index,1);
            last;
        }
        $index++;
    }

    $self->sanlog(-msg => "destroy_quorum completed successfully.");
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   get_disks
# Input:  -self self
# Output: An array of physical disk objects
# Tasks:  * Runs get_disk_groups
#         * Compiles an array of physical disk resourceObj's
###########################################################################
sub get_disks {
    my ($self) = @_;

    my @diskObjs = ();
    # Get an array of group objects
    my @groupObjs = $self->get_disk_groups();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    foreach my $groupObj (@groupObjs) {
        # Inside the group oject is an array of resource objects called "resources"
        foreach my $resourceObj (@{$groupObj->{resources}}) {
           # only interested in objects with type "physical disk"
           if ($resourceObj->{type} =~ /Physical\s+Disk/i) {
               # add it to the collection array
               push (@diskObjs, $resourceObj);
           }
       }
    }
    if (scalar(@diskObjs) == 0) {
       $self->_set_err("No cluster disks were found");
       $self->sanlog(-msg => @{$self->{errmsg}});
    }
    return @diskObjs;
}

###########################################################################
# Name:   get_disk
# Input:  -self self, -name name of disk
# Output: A single physical disk object
# Tasks:  * Runs get_disks_
#         * returns a single disk resource object
###########################################################################
sub get_disk {
    my ($self, %args) = @_;
    my $diskName;

    foreach (keys %args){
        if (/^-?name/i) {
            if (not length($diskName = $args{$_}) ) {
                $self->_set_err("Invalid value for -name: $args{$_}");
                $self->sanlog(-msg => @{$self->{errmsg}});
                return $ClientUtil::FALSE;
            }
            next;
        }
    }
    if (not length($diskName)){
        $self->_set_err(-msg => "The required parameter -name was not supplied");
        return $ClientUtil::FALSE;
    }

    # use get_disks to obtain all available disks in cluster
    my @diskObjs = $self->get_disks();
    if ($self->error()) {
        return {};
    }
    my $diskObj;
    foreach $diskObj (@diskObjs) {
        if ($diskObj->{name} =~ /$diskName/i) {
            return $diskObj;
        }
   }

   # The disk was not found
   $self->_set_err("The requested disk \"$diskName\", was not found.");
   $self->sanlog(-msg => @{$self->{errmsg}});
   return {};
}

###########################################################################
# Name:   get_quorum
# Input:  -self self,
# Output: quroum disk object
# Tasks:  * Runs get_disks_
#         * returns a single quorum disk resource object
###########################################################################
sub get_quorum {
    my ($self)= @_;

    # use get_disks to obtain all available disks in cluster
    my @diskObjs = $self->get_disks();

    # There were no disks
    if ($self->error()) {
        return {};
    }

    foreach my $diskObj (@diskObjs) {
        if ($diskObj->{quorum} == $ClientUtil::TRUE) {
            return $diskObj;
        }
    }

    # We didn't find a quorum disk.
    $self->_set_err("A quorum disk was not found");
    $self->sanlog(-msg => @{$self->{errmsg}});
    return {};
}

###########################################################################
# Name:   discover_quorum
# Input:  -self self
# Output: True or False
# Tasks:  * Find the quorum disk
#         * Update the information on the corresponding disk object
###########################################################################
sub discover_quorum {
    my ($self) = @_;

    my $headerLines = 2;
    my $quorumDiskName;

    my @diskObjs = $self->get_disks();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Searching for an existing cluster quorum disk");

    my $cmd = 'cluster . /quorum';
    my (@output) = $self->execute(-cmd => $cmd);
    # Quorum Resource Name Device                                   Max Log Size
    # -------------------- ---------------------------------------- ------------
    # Disk E                E:\MSCS\                                 4194304

    if ($self->error()) {
        return $ClientUtil::FALSE;
    }
    # Lose the header lines
    foreach (1 .. $headerLines) {
        shift (@output);
    }

    # in 2k8, the default configuration, will be a node majority
    # in this case, there will be no quorum object. just return TRUE
    # Witness Resource Name Path                                          Type
    # --------------------- --------------------------------------------- --------
    # (Node Majority)                                                     Majority
    if (grep (/Node\s+Majority/i, @output)) {
        $self->sanlog(-msg => "This cluster is a type \"Node Majority\", " .
                              " there is no Quroum device");
        return $ClientUtil::TRUE;
    }

    # The quorum disk name can have spaces in it.
    foreach my $line (@output) {
        my @quorumData = split(/\s+/, $line);
        # The last two elements of the array cannot have spaces.
        # We don't need them. Get rid of them.
        pop(@quorumData);
        pop(@quorumData);
        # quorum disk name is all thats left.
        $quorumDiskName = join(' ', @quorumData);
        if (not length($quorumDiskName)) {
            # If we get here, there is no cluster quorum defined. This is a
            # possible scenario, and will still return true.
            $self->_set_err("Unable to find the current quorum disk");
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::TRUE;
        }
    }

    # If its a local quorum device, it is no further use to us. return true now because
    # there are no errors and we are complete
    if ($quorumDiskName =~ /local\s+quorum/i) {
        return $ClientUtil::TRUE;
    }

    $self->sanlog(-msg => "Discovered the cluster quorum \"$quorumDiskName\".");
    # Go through the disks array and find the corresponding disk object
    # so that we can update the quorum data.
    foreach my $diskObj (@diskObjs) {
        if ($diskObj->{name} =~ /$quorumDiskName/i) {
            $diskObj->{quorum} = $ClientUtil::TRUE;
            $self->{quorum} = $quorumDiskName;
            return $ClientUtil::TRUE;
        }
    }

    # This is an error, this means that there is no existing object for the
    # the quorum disk. That means something probably went wrong in discovery.
    $self->_set_err("Unable to find an existing resource object for the quorum disk \"$quorumDiskName\"");
    $self->sanlog(-msg => @{$self->{errmsg}});
    return $ClientUtil::FALSE;
}

#############################################################
# Name: discover_groups
# Input: -self (sanClusterObj)
# Output: array of names representing existing cluster groups
# Tasks: * Interrogate the cluster for group names
#############################################################
sub discover_groups {
    my ($self) = @_;

    my @clusterGroupNames =();
    # cmd to determine cluster groups
    my $cmd = 'cluster . group';

    # Blank out the current groups array on self so we don't duplicate anything
    @{$self->{diskGroups}} = ();

    my @clusterGroupOut = $self->execute(-cmd => $cmd);

    if ($self->error()) {
        return $ClientUtil::FALSE;
    }
    # The output looks like this:
    # Listing status for all available resource groups:
    #
    # Group                Node            Status
    # -------------------- --------------- ------
    # Cluster Group        CLUSTERTEST1    Online
    # NATE_clustertest1    CLUSTERTEST1    Online
    # NATE_clustertest2    CLUSTERTEST1    Online
    #

    # Get rid of the first four lines, then go through the rest
    # and pull out all the cluster group names
    foreach ( 1 .. 4 ) {
        shift( @clusterGroupOut );
    }
    foreach my $line ( @clusterGroupOut ) {
        # First gotta deal with spaces in the Status ("Partially Online")
        $line =~ s/Partially Online/PartiallyOnline/;
        # Gotta deal with freaking spaces in the group name....grumble...
        my @dataDump = split (/\s+/, $line);
        # The last two objects will not have spaces, whatever is left
        # is the group name
        my ($status, $currentOwner, $groupName);
        $status = pop(@dataDump);
        $currentOwner = pop(@dataDump);
        $groupName = join(' ', @dataDump);

       if (( length($status) and (length($currentOwner)) and (length($groupName)))) {
           $self->sanlog(-msg => "Discovered cluster group $groupName");
           push (@clusterGroupNames, $groupName);
       } else {
           # This is in no way fatal, only informational...Just means we encountered some
           # output that we didn't understand.
           $self->sanlog(-msg => "Did not properly discover a cluster group from \"$line\"");
       }
    }
    return @clusterGroupNames;
}

###########################################################################
# Name:  _create_deviceId_map {
# Input:  -self self Object
# Output: True or False
# Tasks: * Run system commands to retrieve a list of deviceID's that
#          that is paired with the related drive Letter
#        * Populates a hash on the SanCluster object with the DeviceID as the
#          the key, and the Letter as the value. We end up with a list of
#          all the cluster devices, owned by any node in this hash.
###########################################################################
sub _create_deviceId_map {
    my ($self, %args) = @_;
    # The -NodeOnly parameter is only supplied in situations where windows 2008
    # has a cluster created, but no storage added to the cluster. In such a case
    # The MSCluster WMI class is not populated yet. So, we need to get the volume
    # information from the node only. Win32_Volume class will be used instead.
    my $noClusterDisks;

    foreach (keys %args){
        if (/^-?nocluster/i) {
            if (not length($noClusterDisks = $args{$_}) ) {
                $self->_set_err("Invalid value for -nocluster: $args{$_}");
                $self->sanlog(-msg => @{$self->{errmsg}});
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($noClusterDisks == $ClientUtil::TRUE) {
        $self->sanlog(-msg => "The -nocluster parameter was supplied and will " .
                              "cause Win2k3 type discovery methods to be used.");
    }

    # I tried putting this in %SystemDrive%\temp...but it complains "access denied"
#    my $tempFile = "%SystemRoot%\\get_device_mnt.vbs";
     my $tempFile = "%SystemDrive%\\get_device_mnt.vbs";

    # Get the primary node and find out what type of cluster this is.
    my $primaryNodeObj = $self->get_primary_node();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    my $build = $primaryNodeObj->{sanClientObj}->{WinClient}->{build};
    my $Win2k8 = $ClientUtil::FALSE;

    my @clusterNodes;

    if ((int($build) >= $WIN2K8BUILD) && (not length($noClusterDisks))) {
        $Win2k8 = $ClientUtil::TRUE;
        $self->sanlog(-msg => "Using Win2k8 and up discovery method");
        # On windows 2008 and up, the WMI features have been improved and we can
        # get all the data about clustered disks from the primary node alone.
        my $nodeObj = $self->get_primary_node();
        push (@clusterNodes, $nodeObj);
    } else {
        # On 2k3, we need to query each node individually to figure out what disks
        # It currently owns
        $self->sanlog(-msg => "Using Win2k3 discovery method");
        @clusterNodes = $self->get_nodes();
    }

    foreach my $nodeObj (@clusterNodes) {
        my $host = $nodeObj->{hostname};
        $self->sanlog(-msg => "Updating deviceId_map hash on self with " .
         "data from $host");

        # The idea here is to get a hash on self that maps the device drive
        # letter to a unique identifier that is known by the cluster service. This is harder
        # then it sounds. Especially on 2k8 and up. There isn't an easily accessible direct corelation
        # and cluster.exe is of no help. We have an easier time on 2k3 because we do not support GPT
        # disks in the cluster. 2k8 will support GPT and MBR, so the identifiers are different. In
        # either case we can use WMI.
        my ($lineOne, @theRest);
        if (($Win2k8 != $ClientUtil::TRUE) or (length($noClusterDisks))) {
            # This is the windows 2003 script. This will give you a device ID that is also avaailable
            # through cluster.exe res /priv. In this case the string we will use as deviceId  and the
            # corresponding driver letter are both available from the same WMI class.
            $lineOne =  'Set objWMIService = GetObject("winmgmts:' . "\\\\" . $host . '\root\cimv2")',
            @theRest = ( 'Set colItems = objWMIService.ExecQuery("Select * from Win32_Volume")',
                         'For Each objItem In colItems',
                         'WScript.Echo "ID: " ^& objItem.DeviceID ^& " Letter: " ^& objItem.DriveLetter',
                         'Next');
        } else {
            # The 2008 script. The wizarding is more intense to do the matching on 2008. We need to do this
            # differently because the MSCluster class does not exist in 2k3, and the disks can be either mbr or
            # gpt.

            $lineOne = 'Set objWMIServiceCluster = GetObject("winmgmts:' . "\\\\" . $host . '\root\MSCluster")';
            @theRest = ( 'Set colItemsRes = objWMIServiceCluster.ExecQuery("Select * from MSCluster_Resource")',
                         'For Each objItemRes In colItemsRes',
                         'strResName = objItemRes.Name',
                         "Set colDisks = objWMIServiceCluster.ExecQuery(\"ASSOCIATORS OF {MSCluster_Resource.Name=\"\"\" ^& strResName ^& \"\"\"} WHERE AssocClass = \" ^& \" MSCluster_ResourceToDisk\")",
                         'for each objDisk in colDisks',
                         'strDiskId = objDisk.ID',
                         "set colParts = objWMIServiceCluster.ExecQuery(\"ASSOCIATORS OF {MSCluster_Disk.ID=\"\"\" ^& strDiskId ^& \"\"\"} WHERE AssocClass = \" ^& \" MSCluster_DiskToDiskPartition\")",
                         'for each objPart in colParts',
                         'Wscript.Echo "Name: " ^& objDisk.ID ^& " Path: " ^& objPart.Path',
                         'Next',
                         'Next',
                         'Next');
        }

        # Start with lineone to create the file
        my $cmd = "echo $lineOne > $tempFile";
        my (@output) = $self->execute(-cmd => $cmd, -host => $host);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # append @theRest to the file on the client machine
        foreach my $line (@theRest) {
            my $cmd = "echo $line >> $tempFile";
            my (@output) = $self->execute(-cmd => $cmd, -host => $host);
            # There should be no output from this command.
            if ($self->error()) {
                 return $ClientUtil::FALSE;
            }
        }

        # execute the vbscript using cscript
        $cmd = "cscript $tempFile";
        (@output) = $self->execute(-cmd => $cmd, -host => $host);
        # I am not checking the output here, because if no disks are owned by this node, there
        # may not be any data returned. The output would look like this:

        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
        # Make sure that the output array is not empty. That would be bad.
        if (scalar(@output) == 0 ) {
            $self->_set_err(msg => "No Devices were found while trying to generate the device ID map");
            return $ClientUtil::FALSE;
        }

        if ($Win2k8 != $ClientUtil::TRUE) {
            # If we are lower then 2k8, this is it. Wrap it up here, we are done once the
            # output is collected. Output will look like this:
            #
            # ID: \\\\?\\Volume{348d7ff1-c76f-11dd-8857-806e6f6e6963}\\ Letter: C:\r\r\n
            # ID: \\\\?\\Volume{348d7fef-c76f-11dd-8857-806e6f6e6963}\\ Letter: D:\r\r\n
            foreach my $line (@output) {
                if ( $line =~ /^ID:\s+.+Volume{(.+)}.+\s+Letter:\s+(.+):$/ ) {
                    # Extract the device ID and letter from the match, and add to the hash.
                    # This is going on the self so that we can easily look at all
                    # disk letters in use, on the top level
                    $self->{deviceId_map}->{$1} = $2;
                }
            }
            $self->sanlog(-msg => "Completed updating the deviceId_map on $host");

        } else {
            # On 2k8 the output is slighty different. This output is an example of
            # 2k8 with GPT disks. Disks using mbr will have a slightly different output
            # The regex is designed to be loose, to account for this as well as the possible
            # usage of junction points
            # Output when GPT disks are in use:
            # Name: {c672d715-8f98-4482-b448-33593e216d15} Path: E:
            # Name: {57d06cbb-1c9f-4440-b311-1b9d73c501d5} Path: O:
            #
            # An example of MBR disk output
            # Name: 441922277 Path K:
            # Name: 441922284 Path L:
            foreach my $line (@output) {
                if ( $line =~ m/Name:\s+(.+)\s+Path:\s+(.):$/ ) {
                    $self->{deviceId_map}->{$1} = $2;
                }
            }
        }
    }
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   balance_disks_among_groups
# Input:  -self self
# Output: True or False
# Tasks:  * Get an array of disk objects
#         * Determine the number of configured physical disk objects
#         * Dissiminate the available disks evenly among the groups
#           - Groups configured with preferred owners will receive disks
###########################################################################
sub balance_disks_among_groups {
    my ($self) = @_;

    $self->sanlog(-msg => "Begin rebalancing resources among the available groups");

    # To avoid problems with MPC and clusters on 2k3, move all the groups to the
    # primary node. First we need to know what the groups are. Doing things this way
    # keeps us from having to run mpc on each cluster node to discover information related to
    # the disks, such as filer name...
    my @groupObjs = $self->get_disk_groups();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Get the primary node object
    my $primaryNodeObj = $self->get_primary_node();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Make sure each group is moved to the primary node.
    foreach my $groupObj (@groupObjs) {

        # Don't worry about a move if the group is already on the primary node
        next if ($groupObj->{currentOwner} =~ /$primaryNodeObj->{hostname}/i);

        $groupObj->move(-owner => $primaryNodeObj->{hostname});
        if ($groupObj->error()) {
            $self->_set_err(@{$groupObj->{errmsg}});
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
    }

    # Get an array of disk resource Objects
    my @diskObjs = $self->get_disks();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Complain if there are no disks
    if (not(@diskObjs)) {
        $self->_set_err("There are no disks found. The disks cannot be balanced.");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Check for the filer attribute on each object
    my $noFilerData = $ClientUtil::FALSE;
    my $counter = 0;
    foreach my $diskObj (@diskObjs) {
        if (not length($diskObj->{filer})) {
            $noFilerData = $ClientUtil::TRUE;
        }
        $counter++;
    }
    $self->sanlog(-msg => "There are " .  scalar(@diskObjs) . " cluster disks available in this cluster.");

    # If the filer data does not exist, run _discover_disk_filer to make an attempt
    # to gather the data.
    if ($noFilerData != $ClientUtil::FALSE) {
        $primaryNodeObj->_discover_disk_filer();
        if ($primaryNodeObj->error()) {
            $self->_set_err(@{$primaryNodeObj->{errmsg}});
            return $ClientUtil::FALSE;
        }
    }

    my @filerNames;
    # Go through the list of disks again, just in case we had to rediscover the filer
    # attributes above. Create an array of filer names.
    foreach my $diskObj (@diskObjs) {
        if (not grep(/$diskObj->{filer}/i, @filerNames)) {
            push (@filerNames, $diskObj->{filer});
        }
    }

    # No filers would be bad
    if (not scalar(@filerNames) >= 1) {
        $self->_set_err("No filer name could be determined from any resource objects");
        $self->sanlog(@{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Fill out the groupObj array by getting the preferred groups
    # for every node in the cluster.
    my @nodeObjs = $self->get_nodes();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Not sure the goal of this loop here, but removing the 'my' broke the
    # original behavior, so for now I'll just empty the array, so that the same
    # behavior is maintained.
    @groupObjs = ();
    foreach my $nodeObj (@nodeObjs) {
        my @prefGroups = $nodeObj->get_preferred_groups();
        if ($self->error()) {
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
        foreach my $group (@prefGroups) {
            push(@groupObjs, $group);
        }
    }
    if (not scalar(@groupObjs) >= 1) {
        $self->_set_err("A preferred owner is required for at least one group in the " .
                        "cluster. If a preferred owner is not defined for a group, we cannot tell which " .
                        "cluster node it represents, and are therefore unable to determine how to " .
                        "balance the available disks among the current cluster nodes. Please define a " .
                        "preferred owner for each group that should receive a share of the available disks");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    my $numberOfNodes = scalar(@nodeObjs);
    my $numberOfDisks = scalar(@diskObjs);

    $self->sanlog(-msg => "There are " . scalar(@nodeObjs) . " cluster members. There are " .
      scalar(@groupObjs). " preferred groups found.");

    # Lets make sure that there are enough disks to go around
    if ($numberOfDisks < $numberOfNodes) {
        $self->_set_err("There are not enough disks to assign at least one disk to " .
          "each cluster node");
        $self->sanlog(@{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # The idea here is to be able to ensure that each node/group that has a preferred
    # owner will receive disks from each controller head that exists in the cluster.
    # In order to achieve this we do the following until we are out of disks

    # |--------| - Disk1             Group/Node 1
    # | FilerA |                     ------------
    # |--------| - Disk2   ___|\     FilerA - Disk1
    #                     |     \    FilerB - Diks1
    # |--------| - Disk1  |___  /
    # | FilerB |              |/     Group/Node 2
    # |--------| - Disk2             ------------
    #                                FilerA - Disk2
    #                                FilerB - Disk2

    # 1. Select a (or the next) filer
    # 2. Select a (or the next) group
    # 3. Go through the list of disks
    # 4. Find a disk that belongs to the filer that we have selected
    # 5. Move the resource to the group, if isn't already there
    # 6. GOTO 1, until there are no disks left to allocate

    # Do the disk allocating...
    # While we have disks remaining
    while (@diskObjs) {
        # Alternate the filer on each pass.
        foreach my $filer (@filerNames)  {
            # Select each resource group once per filer name
            foreach my $groupObj (@groupObjs) {
                my $groupName = $groupObj->{name};
                # Cycle through the disk objects until the filer name
                # of one disk matches the filer we are working on. Add
                # to the current group
                # The counter keeps track of which disk we are looking at in the array.
                my $counter = 0;
                foreach my $diskObj (@diskObjs) {
                    # Lose the quorum disk when we find it, it shouldn't move,
                    # but it will show up as a disk.
                    if ($diskObj->{quorum}) {
                       splice (@diskObjs, $counter, 1);
                       next;
                    }
                    if ($diskObj->{filer} =~ /$filer/i) {
                        # Do not try to move to specified group if it is already there
                        if ($diskObj->{groupName} !~ /$groupName/i) {

                            # Do the move of the disk to the new group
                            $diskObj->move(-group => $groupName);
                            if ($diskObj->error()) {
                                $self->_set_err(@{$diskObj->{errmsg}});
                                return $ClientUtil::FALSE;
                            }
                            # update the groupName field on the diskObj
                            $diskObj->{groupName} = $groupName;
                        } else {
                            # The disk matches our filer need, and is already a member of the group
                            $self->sanlog(-msg => "Disk \"$diskObj->{name}\" on filer " .
                              "$filer already belongs to the group \"$groupName\".");
                        }
                        # remove the disk from the pool whether we moved it or not
                        splice (@diskObjs, $counter, 1);
                        # Do not look at any more disks until we move to the next group
                        last;
                    }
                    # keep track of the position in the diskObj array
                    $counter++;
                }
            }
        }
    }
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   get_client_dev_list
# Input:  -self self
# Output: Array of clientDevList values for each cluster node
# Tasks:  * Find the disk mount points for each of the disks
#         * Present a summary of mount points by owner.
#           - This means that mounts belong to a group that "should" be owned by the listed node
#           - It DOES NOT mean the current group is owned by the preferred owner at this time.
#           - see self->move_to_preferred_owner()
#         * Return CLIENT_DEV_LIST
###########################################################################
sub get_client_dev_lists {
    my ($self) = @_;

    my @ownerNames;
    my @diskData;
    my $clientDevList;

    # Get a list of disks to determine groups
    my @diskObjs = $self->get_disks();
    if ($self->error()) {
        return ();
    }

    # Start the msg to be sent to the log via sanlog
    my @msg = ("The cluster disks are configured as follows:\n");

    # Need to gather all the owner names, so that we can work by name of the node while laying
    # out the data for the log file and CLIENT_DEV_LIST
    foreach my $diskObj (@diskObjs) {
        # Skip the quorum disk. We don't mess with that.
        next if ($diskObj->{quorum} == $ClientUtil::TRUE);
        # Do not need to save the ownerName if we already know about it
        next if (grep (/$diskObj->{diskGroupObj}->{preferredOwner}/i, @ownerNames));
        # Keep track of the possible owners we find
        push(@ownerNames, $diskObj->{diskGroupObj}->{preferredOwner});
    }
    # The log data will be laid out relative to the node name
    foreach my $ownerName (@ownerNames) {
        # Start the clientdevlist with the owner node name
        $clientDevList .= lc($ownerName) . ",";
        # the msg array goes to the log file, add the client name as a heading
        # For each pass we will draw out only lines relevant to our current owner
        # Repeat this for every preferred owner found in the cluster object.
        push(@msg, "Node $ownerName:\n");
        # Get the last object in the array, so that we can check it later.
        foreach my $diskObj (@diskObjs) {
            # If this resource does not match the owner we are working on, move on
            next if ($diskObj->{diskGroupObj}->{preferredOwner} !~ /$ownerName/i);
            # Skip the quorum resource
            next if ($diskObj->{quorum} == $ClientUtil::TRUE);
            # Put the disk info into the @msg to display in the log
            push (@msg, "Resource: $diskObj->{name}, Group: $diskObj->{groupName}, Mount: $diskObj->{mount_pt}\n");
            # add the disk letter to the client devlist
            $clientDevList .= $diskObj->{mount_pt} . ",";
        }
        # Turn the last , into a ; to complete CLIENT_DEV_LIST for this client
        $clientDevList =~ s/,$/;/;
    }
    # Throw this clients dev list on the @msg to save it.
    push (@msg, $clientDevList);
    # put a carriage return on the msg array to seperate the different groups
    push (@msg, "\n");

    $self->sanlog(-msg => \@msg);
    return $clientDevList;
}

###########################################################################
# Name:   move_to_preferred_owner
# Input:  -self self
# Output: True or FALSE
# Tasks:  * Find all groups with preferred owners listed
#         * Move the group to the listed preferred owner
###########################################################################
sub move_to_preferred_owner {
    my ($self) = @_;

    $self->sanlog(-msg => "Moving each group to its first designated preferred owner");

    # Get all the groups
    my @groupObjs = $self->get_disk_groups();

    foreach my $groupObj (@groupObjs) {
        # Skip this group if there is no listed preferred owner
        if (not length($groupObj->{preferredOwner})) {
            $self->sanlog(-msg => "A preferred object is not defined for the " .
                                  "group named \"$groupObj->{name}\". The group will not be moved.");
            next;
        }
        # Do nothing if the group is in the right spot already
        if ($groupObj->{currentOwner} !~ /$groupObj->{preferredOwner}/) {

            $self->sanlog(-msg => "The cluster group \"$groupObj->{name}\" is being moved " .
                                  "from node $groupObj->{currentOwner} to node " .
                                  "$groupObj->{preferredOwner}...");

            $groupObj->move(-owner => $groupObj->{preferredOwner});
            if ($groupObj->error()) {
                $self->_set_err(@{$groupObj->{errmsg}});
                $self->sanlog(-msg => @{$self->{errmsg}});
                return $ClientUtil::FALSE;
            }

            $self->sanlog(-msg => "The group \"$groupObj->{name}\" was successfully moved " .
                                  "from $groupObj->{currentOwner} to node $groupObj->{preferredOwner}...");
        } else {
            $self->sanlog(-msg => "The group \"$groupObj->{name}\" is already owned by the " .
                                  "preferred owner node.");
        }
    }
    $self->sanlog(-msg => "All cluster groups are now owned by the appropriate owners");
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   _find_cluster_disk_mounts
# Input:  -self self
# Output: True or False
# Tasks:  * Collect all the groups in the cluster
#         * Make sure that the deviceId map is up to date
#         * Add the mountpoint to the resource object
#         * This should only be run WHEN A CLUSTER EXISTS!
###########################################################################
sub _find_cluster_disk_mounts {
    my ($self) = @_;

    my $diskId;
    my $needToRun = $ClientUTIL::TRUE;
    my @diskObjs = $self->get_disks();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # This method exists to update the mount_pts on all the resource objs
    # that live in the cluster. We do this through this method because we have
    # to write a vbscript to the host in order to get the deviceId map information
    # doing this once per resource is VERY time consuming. Using this routine,
    # we can write/run the VB code one time and find all the mount_pt values for
    # every resource object in one shot.

    # See if we need to do anything. If the mount_pts are filled out, we
    # just return true.
    foreach my $diskObj (@diskObjs) {
        if (not length($diskObj->{mount_pt})) {
            $needToRun = $ClientUtil::FALSE;
            last;
        }
        # if we arrive at the last disk in the array, and we have
        # all the mount_pt values, we are done
        if ($diskObj->{name} =~ /$diskObjs[-1]->{name}/) {
           $self->sanlog(-msg => "All resources have mount_pt values. There is " .
                                 "nothing to do.");
           return $ClientUtil::TRUE;
        }
    }

    # Start by updating the deviceId map.
    my $result = $self->_create_deviceId_map();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    foreach my $diskObj (@diskObjs) {
        # Check for the deviceId first. On 2k3 this is always the correct value to use because the disks are
        # only relevant to the nodes via WMI, and GPT disks are not supported . In 2k8, the disks are relevant
        # to the cluster in WMI. In 2k8 if the device is GPT, we use the deviceId field to identify the disk.
        # Alternatively if the disk is an MBR disk, we need use the signature to identify the device in the cluster.
        # In this case, the GUID field will be blank. Because of this, we can check for the GUID first.
        # If its blank, use the signature.
        if (length($diskObj->{deviceId})) {
            $diskId = $diskObj->{deviceId};
        } else {
            $diskId = $diskObj->{diskSignature};
        }

        # If the deviceId_map doesn't have a letter for this diskId, complain about it.
        if (not length($self->{deviceId_map}->{$diskId})) {
            $self->_set_err("Could not correlate a disk mount point for the disk named " .
                            "\"$diskObj->{name}\" with  disk ID: $diskId");
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
        # We found the mount, update the object
        $diskObj->{mount_pt} = $self->{deviceId_map}->{$diskId};
    }

    $self->sanlog(-msg => "Successully added mountpoints to disk objects");
    return $ClientUtil::TRUE;
}

1;
