#
# This library is used to populate the Windows Cluster object created by
# WinCluster.pm
#
#
## @summary
##  Creates Windows Cluter object and imports associated libraries.  All architecture
##  specific properties are defined here and extrapolated elsewhere.
##
## @author  Jonathan.Dascenzo@Netapp.com
## @status  public
##

package San::SanClientCluster::Windows::MSCS::Node;

use strict;
use Error;
use WinClient;
use San::SanClient; 
use San::SanLog;

use vars qw(
    @ISA
    @EXPORT
    );

@ISA = qw( SanLog Error );
 
#####################################################
#Input: -parent (sanClusterObj)
#Output: a blank node object
#Tasks: * create an empty node object
####################################################
sub _new {
    my ($class, %args) = @_;
    my $host;
    my $protocol;
    my $self = {};

    bless $self, $class;

    foreach (keys %args){
        if (/^-?parent/) {
            if (not length($self->{sanClusterObj} = $args{$_})) {
                $self->_set_err("Invalid value for -parent: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return $self;
            }
            next;
        } elsif (/^-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for -protocol: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return $self;
            }
            next;
        } elsif (/^-?host/i) {
            if (not length($host = lc($args{$_}))) {
                $self->_set_err("Invalid value for -host: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return $self;
            }
            next;
        }
    }

    if (not length($self->{sanClusterObj})){
        $self->_set_err(-msg => "The required parameter -parent was not supplied");
        return $self;
    }
    if (not length($protocol)){
        $self->_set_err(-msg => "The required parameter -protocol was not supplied");
        return $self;
    }
    if (not length($host)){
        $self->_set_err(-msg => "The required parameter -host was not supplied");
        return $self;
    }


    # Tack on the parent object
    my $sanClusterObj = $self->{sanClusterObj};

    $self->sanlog(-msg => "Creating a new nodeObj for host $host");
 
    my $sanClientObj = SanClient->new(-host => $host,
                                      -protocol => $protocol,
                                      -conn_type => "THARNHOST",
                                      -load_inits => 0,
                                      -timeout => 300,
                                      -debug_level => 0);

   # If there is an error on sanClient, move it up to the nodeObj
   # so that the caller can see it. 
   if ($sanClientObj->error()) {
        $self->_set_err(@{$sanClientObj->{errmsg}});
   }
 
   $self->{sanClientObj} = $sanClientObj;
   
   #Add some other useful info to the top of the node Obj
   $self->{hostname}             = $host;
   $self->{primary_cluster_node} = $ClientUtil::FALSE;
   $self->{cluster_name}         = $sanClusterObj->{cluster_name};
   $self->{cluster_ip}           = $sanClusterObj->{cluster_ip};
 
   return $self;
}

#####################################################
#Input: -self (Node object)
#Output: return True or False
#Tasks: * calls system commands to add node to cluster
####################################################
sub add {
    my ($self) = @_;

    my $sanClusterObj = $self->{sanClusterObj};
    my $WinClientObj = $self->_get_winClient();
    my $adminAccount = "administrator";
    my $error = $ClientUtil::FALSE;

    # Make sure that commands on the system are being executed as any user. i.e. not system
    my $commandContext = $WinClientObj->wrshdnt_check_context();

    $sanClusterObj->sanlog(-msg => "The context of the rsh daemon is found to be $commandContext");

    if (not length($commandContext)) {
        $sanClusterObj->_set_err("There was a problem determining the current rsh context on " .
                                 $WinClientObj->{hostname});
        $sanClusterObj->sanlog(@{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    if ( $commandContext !~ /$adminAccount/i ) {
        $sanClusterObj->sanlog(-msg => "Changing the context of the rsh daemon to be $adminAccount");
        # Modify the wrshdnt service
        my $status = $WinClientObj->wrshdnt_change_context($adminAccount);
        # Check for errors, do not bother with the add command if we cannot guarantee the context of rshd, 
        # it will fail.
        if ($WinClientObj->error()) {
            $self->_set_err(@{$WinClientObj->{errmsg}});
            $sanClusterObj->sanlog(@{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
        if ($status != $ClientUtil::TRUE) {
            $self->_set_err("A bad status was returned from wrshdnt_change_context changing the " .
                            "command context to $adminAccount");
            $sanClusterObj->sanlog(@{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
    }

    # The cluster add command
    # The commands are just slightly differnt between 2k3 and 2k8.
    my $cmd;
    if ($WinClientObj->{build} >= 6) {
        # This is 2k8+
        $sanClusterObj->sanlog(-msg => "Node $self->{hostname} is Windows 2008 or higher");
        $cmd = 'cluster /cluster:' . $sanClusterObj->{cluster_ip} . ' /add /Node:' . $self->{hostname};
    } else {
        # this is 2k3
        $sanClusterObj->sanlog(-msg => "Node $self->{hostname} is Windows 2003.");
        $cmd = 'cluster /cluster:' . $sanClusterObj->{cluster_ip} . 
               ' /add:' . $self->{hostname} .
               ' /min /unattended /password:' . $sanClusterObj->{domain_adminpass};
    }

    # Execute the command on the primary cluster node.
    my (@output) = &ClientUtil::execute($self->{sanClientObj}->{conn}, $cmd);
    # Check the status, if things are bad, just report so that we can set the rsh context back. 
    # we will decide what to return later. 
    if (grep (/error/i, @output)) {
        $self->_set_err("There was a problem adding $WinClientObj->{hostname} to the cluster " . 
                        "called $sanClusterObj->{cluster_name}.");
         $self->sanlog(-msg => @{$self->{errmsg}});
         $error = $ClientUtil::TRUE;
     }

    $sanClusterObj->sanlog(-msg => "Successfully added $WinClientObj->{hostname} to the cluster"); 

    # If the previous context was system, we changed it to administrator
    # so put it back. 
    if ($commandContext =~ /system/i ) {
        $sanClusterObj->sanlog(-msg => "Changing the context of the rsh daemon back to system");
        # Modify the wrshdnt service
        my $status = $WinClientObj->wrshdnt_change_context("system");
        # Check for errors, we don't want to overwrite any errors from the add commands, so I just print 
        # we also don't want to error out of this routine, let it complete
        if ($WinClientObj->error()) {
            $sanClusterObj->sanlog(@{$WinClientObj->{errmsg}});
            $error = $ClientUtil::TRUE;
        }
        if ($status != $ClientUtil::TRUE) {
            $self->sanlog(-msg => "A bad status was returned from wrshdnt_change_context changing the " .
                                 " command context to system");
            $error = $ClientUtil::TRUE;
        }
    }
    if ($error == $ClientUtil::TRUE) {
        $sanClusterObj->sanlog(-msg => "nodeObj->add did not complete successfully");
        return $ClientUtil::FALSE;
    }
    $sanClusterObj->sanlog(-msg => "nodeObj->add completed successfully");
    $sanClusterObj->sanlog(-msg => "$WinClientObj->{hostname} was added to the cluster");
    return $ClientUtil::TRUE;
}

#####################################################
#Input: -self (Node object)
#Output: returns WinClientObj 
#Tasks: Return WinClient
####################################################
sub _get_winClient() {
    my ($self) = @_;
    
    return $self->{sanClientObj}->{WinClient};
}

#####################################################
#Input: -self (Node object)
#Output: return True or False
#Tasks: * calls system commands to remove node from cluster
####################################################
sub remove {
     my ($self,%args) = @_;
     my $host = $self->{sanClientObj}->{WinClient}->{hostname};
     my $clusterDestroyCmd = 'cluster node /force';
     my $retries = 3;
     my $counter = 1;
     my $saveObj = $ClientUtil::FALSE;
     my (@output);
     my $sanClusterObj = $self->{sanClusterObj};

    # Sometimes you need to execute the remove function, without actually removing the node object 
    # from the sanClusterObj. Such as the safety net of removing the cluster account from the domain 
    # before attempting to create the cluster. 
    foreach (keys %args){
        if (/^-?SaveObj/i) {
            if (not length($saveObj = $args{$_})) {
                $self->_set_err("Invalid value for -SaveObj: $args{$_}");
                $self->sanlog(@{$self->{errmsg}});
                return ();
            }
        }
     }

     while ($counter <= $retries) {

         $self->sanlog(-msg => "Attempting to clean the cluster service on $host. This is " . 
                              "attempt $counter of $retries.");
         # Execute the command
         (@output) = &ClientUtil::execute($self->{sanClientObj}->{conn}, $clusterDestroyCmd);

         if (not grep /successful/i, @output) {
             $self->sanlog(-msg => "Cluster clean failed on $host. This is attempt $counter of $retries.");
             if ($counter != $retries) {
                 &SanUtil::nate_sleep(10);
             }
         } else {
             # If we succeeded, exit the loop. 
             last;
         }
         $counter++;
     }
     if (not grep /successful/i, @output) {
         $self->_set_err("Cluster clean failed on $host. $retries retries were exhausted. ");
         $self->sanlog(-msg => @{$self->{errmsg}});
         return $ClientUtil::FALSE;
     }
    
     if ($saveObj == $ClientUtil::FALSE) {
         # Delete the node object from the sanClusterObj...If this doesn't work, or the node object isn't found, 
         # the object is inconsistent and we should error. 
         my $index = 0;
         my $deleted = $ClientUtil::FALSE;
         foreach my $nodeObj (@{$sanClusterObj->{nodes}}) {
            if ($nodeObj->{sanClientObj}->{WinClient}->{hostname} =~ /$host/i) {
                splice(@{$sanClusterObj->{nodes}}, $index, 1);
                $deleted = $ClientUtil::TRUE;
                last;
            }
         }
         if ($ClientUtil::TRUE != $deleted) { 
             $self->_set_err("Unable to remove the node object for \"$host\" from sanClusterObj.");
             $self->sanlog(-msg => @{$self->{errmsg}});
             return $ClientUtil::FALSE;
         }
     }
     
     $self->sanlog(-msg => "Successfully cleaned cluster service on $host");

     return $ClientUtil::TRUE;
}

#####################################################
#Input: -self (Node object)
#Output: return True or False
#Tasks: * Sets primary node attribute on -self 
####################################################
sub set_primary_node {
    my ($self) = @_;
    
   #Set Primary Node Attribute on the primary node object
   $self->{primary_cluster_node} = $ClientUtil::TRUE;

   return $ClientUtil::TRUE;
} 

###########################################################################
# Name:  get_devices
# Input:  -self Node Object
# Output: Array of disks
# Tasks: * Find all the attached devices 
#        * Populate the disk hash on the WinClient object in the NodeObj 
###########################################################################
sub get_devices {
    my ($self) = @_;

    my $sanClientObj = $self->get_sanclient_object();

    return $sanClientObj->get_devices();
}

###########################################################################
# Name:  node_status
# Input:  -self Node Object
# Output: True or False
# Tasks: * Call system commands to determine status of the node
#        * Returns true if the node is part of the cluster 
###########################################################################
sub node_status {
    my ($self) = @_;
    
    my $retryCount = 3;
    my $retrySleep = 30; 
    my $retry = 0;
    my $host = $self->{sanClientObj}->{WinClient}->{hostname};
    my $cmd = 'cluster . node';
    my $sanClusterObj = $self->{sanClusterObj};

    while ( $retry <= $retryCount ) {
        $retry++;

        my (@output) = $sanClusterObj->execute(-cmd => $cmd, -host => $self->{hostname});

        if ($sanClusterObj->error()) {
            $self->_set_err(@{$sanClusterObj->{errmsg}});
            $self->sanlog(@{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
        # Just bailout if the cluster node isn't a member of the cluster
        # If that is the case the error message looks like this:
        # System error 1753 has occurred (0x000006d9).
        # There are no more endpoints available from the endpoint mapper.
        if (grep(/0x000006d9/i, @output)) {
           $self->sanlog(-msg => "$host is not part of a working cluster");
           return $ClientUtil::FALSE;
        }
        # The output looks like this:
        # Listing status for all available nodes:
        #
        # Node           Node ID Status
        # -------------- ------- ---------------------
        # CLUSTERTEST1         1 Up
        # CLUSTERTEST2         2 Up

        # Check to verify that the node is in the UP condition
        foreach my $line (@output) {
            if ($line =~ /$host/i) {
                if ($line =~ /up/i) { 
                    $self->sanlog(-msg => "The cluster node $host is UP, and part of the cluster");
                    return $ClientUtil::TRUE;
                } elsif ($line =~ /down/i) {
                    $self->sanlog(-msg => "the cluster node $host is DOWN, and is not a participating " . 
                                          "cluster member");
                    return $ClientUtil::FALSE;
                # If the status is "joining" give it $retrySleep seconds and look again.
                } elsif ($line =~ /joining/i) {
                    &SanUtil::nate_sleep($retrySleep);
                    last;
                }
            }
        }
    }
    # the retry count was exhausted if we got here
    $self->sanlog(-msg => "$host does not appear to be part of the cluster");
    return $ClientUtil::FALSE;
} 

###########################################################################
# Name:  get_sanclient_object
# Input:  -self Node Object
# Output: sanClientObj 
# Tasks: * return the sanClientObj from the $self
###########################################################################
sub get_sanclient_object {
    my ($self) = @_;
    
    return $self->{sanClientObj};
}

###########################################################################
# Name:  validate 
# Input:  -self Node Object
# Output: True or False
# Tasks: * Run tests on the client node to determine if the client 
#           is valid to be part of the cluster
###########################################################################
sub validate {
  my ($self) = @_;

  my $WinClientObj = $self->_get_winClient();
  my $sanClusterObj = $self->{sanClusterObj};

  # On windows, this happens as part of the cluster add command
  # There is currently no "test run" way to run this command
  # So we just check to see that the node version matches the cluster version
  if ($self->{sanClusterObj}->{build} != $WinClientObj->{build}) {
        $sanClusterObj->_set_err("The cluster build version \"$self->{sanClusterObj}->{build}\" does " . 
                                  "not match the node \"$self->{hostname}\" version " .
                                   "\"$WinClientObj->{build}\"");
        $sanClusterObj->sanlog(-msg => @{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
  }
  return $ClientUtil::TRUE;
}

###########################################################################
# Name:  halt 
# Input:  -self Node Object
# Output: True or False
# Tasks: * Run system commands to halt the node
###########################################################################
sub halt {
    # Well sure, this would be really nice to do...save us a lot of trouble 
    # configuring the cluster...But, this machine is really more like a coloring
    # book then a computer utilizing a powerful operating system. 
    # So, we just suck it up, long for a real OS, and try to code around it.
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:  get_preferred_groups
# Input: -self ($self) 
# Output: returns an array of diskGroup objects that have this node listed as the 
#           preferred owner.
# Tasks: * Determine groups that belong to this node
###########################################################################
sub get_preferred_groups {
    my ($self) = @_;

    my $sanClusterObj = $self->{sanClusterObj};
    my @collector = ();
    my @allGroups = $sanClusterObj->get_disk_groups();

    foreach my $groupObj (@allGroups) {
        if ($groupObj->{preferredOwner} =~ /$self->{hostname}/i) {
           # This group has a preferred owner listed as the nodeObj we received 
           push (@collector, $groupObj);
        }
    }
    return @collector;
}

###########################################################################
# Name:  get_groups
# Input: -self ($self) 
# Output: returns an array of diskGroup objects that are currently owned by this node.
# Tasks: * Find groups owned by this node
###########################################################################
sub get_groups {
    my ($self) = @_;

    my $sanClusterObj = $self->{sanClusterObj};
    my @collector = ();
 
    my @allGroups = $sanClusterObj->get_disk_groups();
   
    if ($sanClusterObj->error()) { 
        $self->_set_err(@{$sanClusterObj->{errmsg}});
    }

    foreach my $groupObj (@allGroups) {
        if ($groupObj->{currentOwner} =~ /$self->{hostname}/i) {
           # This group belongs to the nodeObj we received 
           push (@collector, $groupObj);
        }
    }
    return @collector;
}

###########################################################################
# Name:  get_resources
# Input: -self ($self) 
# Output: an array of resource objects that are members of the groups that  
#           are currently owned by this node 
# Tasks: * collect resources owned by this node 
###########################################################################
sub get_resources {
    my ($self) = @_;
    my @groups = $self->get_groups();
    my @collector = ();
  
    foreach my $groupObj (@groups) {
        # There is no error to check, this routine just returns the variable from the object
        # There is no oppurtunity to introduce errors. There may not be any resources in this group
        my @deviceObjs = $groupObj->get_devices();
        
        foreach my $devObj (@deviceObjs) {
            push (@collector, $devObj);
        }
    }
   return @collector;
}
   
#####################################################
# Name: discover_disk_filer
# Input: -self (Node object)
# Output: return True or False
# Tasks: * Find the storage controller name for each disk owned by this node 
#        * populate the data $resourceObj->{filer}
####################################################
sub _discover_disk_filer() {
    my ($self) = @_;

    my $sanClusterObj = $self->{sanClusterObj};
    my $mpcShowDisksCommand = "mpc show disks";
    my $mpcShowLunsCommand = "mpc show luns";
    my @mpcCommands = ("$mpcShowDisksCommand", "$mpcShowLunsCommand");
    my @mpcDisksOutput;
    my @mpcLunsOutput;
    my $mpcRetries = 0;

    my @resourceObjs = $self->get_resources();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $sanClusterObj->sanlog(-msg => "Finding the Storage Controller associated with each physical " . 
                                   "disk resource on cluster member $self->{hostname}");
    # Run both MPC commands. The only data we would have at this stage, is the 
    # disk letter. We cannot derive the filer directly from the letter. We need to 
    # match the letter to disk number (i.e. Disk2) and then the disk number to the
    # filer. On win2003 we COULD go directly from the serial number to the filer, 
    # but on 2k8 I can no longer get the serial number right from cluster.exe...So 
    # we just go the long way for any OS...
    foreach my $cmd (@mpcCommands) { 
        my $mpcMaxRetries = 3;
      
        if ($mpcRetries >= $mpcMaxRetries) {
            $sanClusterObj->_set_err("The maximum retry count of $mpcRetries was exhausted by $cmd");
            $sanClusterObj->sanlog(-msg => @{$sanClusterObj->{errmsg}});
            # Return TRUE here, because MPC will throw an exception on 2k3, if there are no disks
            # or reserved disks on the node being interrogated. FALSE Would be appropriate, but until
            # mpc gets fixed...just live with it. 
            return $ClientUtil::FALSE;
        }
   
        my (@output) = $sanClusterObj->execute(-cmd => $cmd, -host => $self->{hostname} );

        if ($sanClusterObj->error()) {
            $self->_set_err(@{$sanClusterObj->{errmsg}});
            $self->sanlog(@{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
        if (grep /error/i, @output) {
            $sanClusterObj->_set_err("An error occurred running $cmd. This is the $mpcRetries try of" . 
                                     " $mpcMaxRetries tries.");
            $sanClusterObj->sanlog(@{$sanClusterObj->{errmsg}});
            $mpcRetries++;
            &SanUtil::nate_sleep(10);
            redo;
        }
 
        $sanClusterObj->sanlog(-msg => "The command $cmd completed successfully on $self->{hostname}");
        $mpcRetries = 0; 

        # Get rid of the header data so we don't match it up later.
        for (1 .. 3) {
            shift (@output); 
        }
        # Decide where to keep the output
        if ($cmd =~ /lun/i) {
            @mpcLunsOutput = @output;
        } elsif ($cmd =~ /disk/i) {
            @mpcDisksOutput = @output;
        }
    }

    foreach my $resourceObj (@resourceObjs) {
        my $disk; 
        # Skip resources that are not physical disks and don't mess with the quorum 
        next if (($resourceObj->{type} !~ /Physical\s+Disk/i) or ($resourceObj->{quorum} == $ClientUtil::TRUE));
        # Get the line that describes our disk by drive mount point
        # was using a simple grep here, but the "M" in the size field screws that up.
        foreach my $line (@mpcDisksOutput) {
            my @temp = split(/\s+/, $line);
            # Disk13  2       5245 M  OK      False   NETAPP  MBR     Basic   M       NTFS    [NATE_F_13_M:]  False   None    None
            # make sure you match whitespace around the mount_pt or you get the header line. 
            if ($temp[9] =~ /$resourceObj->{mount_pt}/i) {
               $disk = shift(@temp);
               last;
            }
        }
        if (not length($disk)) {
            $self->_set_err("Unable to match disk \"$resourceObj->{mount_pt}\" with output from mpc show disks."); 
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
               
        # Match the line containing our disk number
        my @temp = grep(/^$disk\s+/, @mpcLunsOutput);
        # move on if this node doesn't own the disk 
        # purposely not including i in the match so I don't match "unknown"
        next if ($temp[0] =~ /\sUNK\s/);
        my @words = split(/\s+/, $temp[0]);
        # Disk13  2       ontap   DLQD    4       HnXKVoLpqFed    Unkown  wshe3070c-10    /vol/nate_2_1/nate_wshedl380-05_windows_gpt_1   windows_gpt
        # Update the resource object with the filer
        # The filer name is in postion 7...
        $resourceObj->{filer} = $words[7];
        
        if (not $resourceObj->{filer}) {
            # If we still don't know the filer, complain about it. 
            $self->_set_err("The Storage Controller could not be determined for " .
                            "physical disk resource \"$resourceObj->{name}\".");
            $sanClusterObj->sanlog(-msg => @{$self->{errmsg}});
        }
    }
    return $ClientUtil::TRUE;
}

1;
