#
# This library is used to populate the Windows Cluster object created by
# WinCluster.pm
#
#
## @summary
##  Creates Windows Cluter shared disk libraries and imports associated libraries.  All architecture
##  specific properties are defined here and extrapolated elsewhere.
##
## @author  Jonathan.Dascenzo
## @status  public
##

package San::SanClientCluster::Windows::MSCS::SharedDiskGroup;

use strict;
use Error;
use WinClient;
use San::SanClient;
use San::SanClientCluster::Windows::MSCS::SharedDiskGroup::Resource;
use San::SanLog;

use vars qw( @ISA @EXPORT );
@ISA = qw( SanLog Error );

###########################################################################
# Name:   _new
# Input:  -parent the parent object. Usually sanClusterObj
# Output: a blank shareddiskgroup object
# Tasks:  Create an empty shared disk group objectt
###########################################################################
sub _new {
    my ($class, %args) = @_;
    my $self = {};

    bless $self, $class;

    # clear the errors first
    $self->_clear_err();

    foreach (keys %args) {
        if (/^-?parent/i) {
            if (not length($self->{sanClusterObj} = $args{$_})) {
                $self->_set_err("Invalid value for -parent: $args{$_}");
                $self->sanlog(-msg => @{$self->{errmsg}});
                return $self;
            }
        }
    }
    if (not length($self->{sanClusterObj})) {
        $self->_set_err("The required parameter -parent was not supplied");
        $self->sanlog(@{$self->{errmsg}});
        return $self;
    }

    # Define the object structure here. We will populate it later.
    $self->{preferredOwner} = undef;
    $self->{currentOwner}   = undef;
    $self->{status}         = undef;
    $self->{name}           = undef;
    $self->{group_type}     = undef;
    $self->{resources}      = ();

    $self->sanlog(-msg => "A new diskGroupObj was created");

   return $self;
}

###########################################################################
# Name:   _load
# Input:  -self an empty diskGroup object, -name name of the group
# Output: True or False
# Tasks:  associate any discovered data related to the named diskgroup with
#         the object that we received
###########################################################################
sub _load {
    my ($self, %args) = @_;

    my $sanClusterObj = $self->{sanClusterObj};
    my $groupName;

    foreach (keys %args) {
        if (/^-?name/i) {
            if (!length($groupName = $args{$_})) {
                $self->_set_err("Invalid value for -name: $args{$_}");
                $self->sanlog(-msg => @{$self->{errmsg}});
                return $self;
            }
        }
    }
    if (not length($groupName)) {
        $self->_set_err("The required paramter -name was not supplied");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Announce our intentions
    $self->sanlog(-msg => "Loading group object with $groupName data");

    # cmd to determine cluster groups
    my $cmd = 'cluster . group';
    my (@clusterGroupOut) = $sanClusterObj->execute(-cmd => $cmd);
    if ($sanClusterObj->error()) {
        $self->_set_err(@{$sanClusterObj->{errmsg}});
        $self->sanlog(-msg => @{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # The output looks like this:
    #Listing status for all available resource groups:
    #
    # Group                Node            Status
    # -------------------- --------------- ------
    # Cluster Group        CLUSTERTEST1    Online
    # NATE_clustertest1    CLUSTERTEST1    Online
    # NATE_clustertest2    CLUSTERTEST1    Online

    # Go through the output and match the groupname that
    # we were called with. Need to match at the beginning of the line because
    # the node name can/will match the group name
    my @found = grep (/^$groupName\s/, @clusterGroupOut);
    if (not scalar(@found) > 0) {
        $self->_set_err("The supplied group name \"$groupName\" was not found on this cluster." .
                        "The existing groups are:\n@clusterGroupOut");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # the above grep should only return a single line, so index 0
    my @dataDump = split (/\s+/, $found[0]);
    # The last two objects will not have spaces
    my $status = pop (@dataDump);
    my $currentOwner = pop (@dataDump);

    #Update the object with the data we found, if we got all the data
    if ((length($status)) && (length($currentOwner))) {

        $self->sanlog(-msg => "Found the requested group $groupName. Gathering data..." );

        $self->{currentOwner}   =  $currentOwner;
        $self->{status}         =  $status;
        $self->{name}           =  $groupName;

        my $cacheTimeout = "60";
        my $cmd = 'cluster . res /priv';
        # Check the age of the cluster priv data -
        # If it exceeds $cacheTimeout, the refresh it. If not just use the cache data
        if (($sanClusterObj->{clusterPrivDataAge} + $cacheTimeout) < time() ) {

            $sanClusterObj->{clusterPrivData} = $sanClusterObj->execute(-cmd => $cmd);
            if ($sanClusterObj->error()) {
                $self->_set_err(@{$sanClusterObj->{errmsg}});
                $self->sanlog(-msg => @{$sanClusterObj->{errmsg}});
                return $ClientUtil::FALSE;
            }
            $sanClusterObj->{clusterPrivDataAge} = time();
        }

        # Need to find out if this is a HyperV group. It takes a while to run this command for each group that is
        # found, so lets cache the output on the self sancluster object
        if ((grep(/Virtual\s+Machine/, $sanClusterObj->{clusterPrivData})) and
            (grep(/$groupName/, $sanClusterObj->{clusterPrivData}))) {
            $self->{group_type} = "hyperv";
        } else {
            $self->{group_type} = "disk";
        }

        # Get the preferred owner. Some groups will not have preferred owners.
        $self->{preferredOwner} = $self->get_preferred_owner(-group => $groupName);
        if ($self->error()) {
            return $ClientUtil::TRUE;
        }

        # We have success
        $self->sanlog(-msg => "The received object now represents the group \"$groupName\".");
        return $ClientUtil::TRUE;
    }

    $self->_set_err("The information for $groupName was unable to be discovered");
    $self->sanlog(-msg => @{$self->{errmsg}});
    return $ClientUtil::FALSE;
}

###########################################################################
# Name:   create_resource
# Input:  -self diskGroup object
# Output: An empty resource object
# Tasks:  Contact the Resource.pm _new method and get a new resource object
###########################################################################
sub create_resource {
    my ($self) = @_;

    my $resourceObj = San::SanClientCluster::Windows::MSCS::SharedDiskGroup::Resource->_new(-parent => $self);
    if ($resourceObj->error()) {
        $self->_set_err(@{$resourceObj->{errmsg}});
        return {};
    }

    return $resourceObj;
}

###########################################################################
# Name:   discover_resources
# Input:  -self diskGroup object
# Output: True of False
# Tasks:  Find all the resources that are associated with group name
#         on the object that we received
###########################################################################
sub discover_resources {
    my ($self, %args) = @_;

    my $groupName = $self->{name};
    my $sanClusterObj = $self->{sanClusterObj};

    $self->sanlog(-msg => "Discovering resources in group $groupName...");

    # Erase the $self->{resources} array so that nothing is duplicated
    $self->{resources} = ();

    # Now we gather all the resource data on the cluster
    my $cmd = 'cluster . res';
    my (@clusterRes) = $sanClusterObj->execute(-cmd => $cmd);
    if ($sanClusterObj->error()) {
        $self->_set_err(@{$sanClusterObj->{errmsg}});
        $self->sanlog(-msg => @{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # The output we expect looks like this:
    # Resource             Group                Node            Status
    # ------------         ---------            -------------   -------
    # Diske                04_Disks             WSHEDL380-05    Online
    # Diskf                04_Disks             WSHEDL380-05    Online
    # DiskG                wshedl380-05         WSHEDL380-05    Online

    # Look at every cluster resource. When we find one in the current group
    # send it to resource.pm to create an object for the resource.
    foreach my $line (@clusterRes) {
         # use some Perl wizarding to grab the names of the resources that reside in the group
         # that we are working on. See the output above.

         # Need to also match the offline/online status. This is in case the groupname and the node name are
         # the same. The \S in this expression is meant to catch the nodename, which cannot include spaces.
         if ( $line =~ m/^(.+)\s+$groupName\s+\S+\s+(Online|Offline)/i) {
             my $resourceName = $1;
             # Get rid of trailing spaces
             $resourceName =~ s/\s+$//;
             $self->sanlog(-msg => "Found resource \"$resourceName\" to be a member of group $groupName" );

             # Create an empty object for each resource that is found.
             my $resourceObj = $self->create_resource();
             if ($self->error()) {
                 return $ClientUtil::FALSE;
             }

             # Associate the emtpy object with an existing cluster resource
             $resourceObj->_load(-name => $resourceName);
             if ($resourceObj->error()) {
                 $self->_set_err(@{$resourceObj->{errmsg}});
                 return $ClientUtil::FALSE;
             }

             # Put the resource object on the diskGroup (-parent) object
             push (@{$self->{resources}}, $resourceObj);
        }
    }
    $self->sanlog(-msg => "Discovery of resources belonging to group \"$groupName\" completed.");
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   create_group
# Input:  -self diskGroupObj, -name, -owner
# Output: True or False
# Tasks:  Create a cluster storage group
###########################################################################
sub create_group {
    my ($self, %args) = @_;

    my $preferredOwner;
    my $groupName;
    my $currentOwner;
    my $sanClusterObj = $self->{sanClusterObj};

    # Check for some required parameters
    foreach (keys %args) {
        if (/^-?name/i) {
            if (not length($groupName = $args{$_})) {
                $self->_set_err("Invalid value for -name: $args{$_}");
                return $self;
            }
            next;
        } elsif (/^-?owner/i) {
            if (not length($preferredOwner = $args{$_})) {
                $self->_set_err("Invalid value for -owner: $args{$_}");
                return $self;
            }
        }
    }
    if (not length($groupName)) {
       $self->_set_err("The required parameter -name was not supplied");
       $self->sanlog(-msg => @{$self->{errmsg}});
       return $ClientUtil::FALSE;
    }
    if (not length($preferredOwner)) {
       $self->_set_err("The required parameter -owner was not supplied");
       $self->sanlog(-msg => @{$self->{errmsg}});
       return $ClientUtil::FALSE;
    }

    # cmd to determine cluster groups
    my $cmd = 'cluster . group';
    my @output = $sanClusterObj->execute(-cmd => $cmd);
    if ($sanClusterObj->error()) {
        $self->_set_err(@{$sanClusterObj->{errmsg}});
        $self->sanlog(-msg => @{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # The output looks like this:
    # Listing status for all available resource groups:
    #
    # Group                Node            Status
    # -------------------- --------------- ------
    # Cluster Group        CLUSTERTEST1    Online
    # NATE_clustertest1    CLUSTERTEST1    Online
    # NATE_clustertest2    CLUSTERTEST1    Online

    # Just need to verify that the group doesn't exist before
    # creating a new one. If we don't match at the beginning of the line,
    # we will match the nodename in the node column. Don't change that.
    if (grep(/^$groupName\s/, @output)) {
        $self->sanlog(-msg => "The group $groupName already exists.");
        $self->_set_err("$groupName already exists.");
        return $ClientUtil::FALSE;
    }

    # Here are the commands to create and online the group
    my $cmd = 'cluster . group "' . $groupName . '" /create && ' .
              # Set the owner
              'cluster . group "' . $groupName .  '" /SETOWNERS:' . $preferredOwner . ' && ' .
              # Online the group
              'cluster . group "' . $groupName . '" /on';
    my (@output) = $sanClusterObj->execute(-cmd => $cmd);
    # Error checks..
    if ($sanClusterObj->error()) {
        $self->_set_err(@{$sanClusterObj->{errmsg}});
        $self->sanlog(-msg => @{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
    }
    if (grep (/error/i, @output)) {
        $self->_set_err("Detected an error while attempting to create shared disk group \"$groupName\".");
        $self->sanlog(-msg => @{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Associate the data of the new group with the diskGroupObj
    $self->_load(-name => $groupName);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Put this newly created disk group object on the sanClusterObj
    push(@{$sanClusterObj->{diskGroups}}, $self);

    $self->sanlog(-msg => "The diskgroupObj \"$groupName\" was created successfully");
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   get_preferred_owner
# Input:  -self diskGroupObj, -group name of group
# Output: The nodeName of the groups preferred owner
# Tasks:  Find and return the preferred owner
###########################################################################
sub get_preferred_owner {
    my ($self, %args) = @_;

    my $groupName;
    my $preferredOwner = undef;
    my $numberOfHeadingLines = 5;
    my $sanClusterObj = $self->{sanClusterObj};

    foreach (keys %args) {
        if (/^-?group/i) {
            if (not length($groupName = $args{$_})) {
                $self->_set_err("Invalid value for -group: $args{$_}");
                return $self;
            }
            next;
        }
    }

    if (not length($groupName)) {
            $self->_set_err("The required parameter -group was not supplied");
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $preferredOwner;
    }
    # Get the groups preferred owner
    my $cmd = 'cluster . group "' . $groupName .  '" /LISTOWNERS';
    my (@output) = $sanClusterObj->execute( -cmd => $cmd );
    if ($sanClusterObj->error()) {
        $self->_set_err(@{$sanClusterObj->{errmsg}});
        $self->sanlog(-msg => @{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Sample of the output. There may be more than one preferred owner,
    # if so, we are just going to get and use the first one.
    #
    # Listing preferred owners for resource group 'nate_clustertest1':
    #
    # Preferred Owner Nodes
    # ---------------------
    # CLUSTERTEST1

    # Skip all the headings up to the ----- line.
    foreach ( 1 .. $numberOfHeadingLines ) {
        shift (@output);
    }
    my $line = @output[0];

    # The owner node does not allow spaces.
    if ($line =~ /^(.+)$/) {
        $preferredOwner = $1;
    }
    if (not length($preferredOwner)) {
        # This is not fatal, we just don't know where to put the group...
        # It stays where it lays.
        $self->sanlog(-msg => "A preferred owner was not found for $groupName");
        return $preferredOwner;
    }
    $self->sanlog(-msg => "Discovered $preferredOwner as the preferred owner " .
                          "of $groupName");
    return $preferredOwner;
}

###########################################################################
# Name:   destroy
# Input:  -self diskGroupObj
# Output: True or False
# Tasks:  destroy the disk group
###########################################################################
sub destroy {
    my ($self) = @_;
    my $sanClusterObj = $self->{sanClusterObj};
    my $groupName = $self->{name};

    my $cmd = 'cluster . group "' . $groupName . '" /DEL';
    my (@output) = $sanClusterObj->execute(-cmd => $cmd);
    if ($sanClusterObj->error()) {
        $self->_set_err(@{$sanClusterObj->{errmsg}});
        $self->sanlog(-msg => @{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
    }
    # We get a system 145 error if the group is not empty
    # System error 145 has occurred (0x00000091)
    if (grep(/System\s+error\s+145\s+occurred/, @output)) {
        $self->_set_err("The group cannot be deleted until all resources have been removed");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # A successful output looks like this:
    # Deleting resource group 'test'
    if ((grep(/error/i, @output)) or (not grep(/Deleting\s+resource\s+group\s+/i, @output))) {
        $self->_set_err("A problem occurred removing $groupName");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }
    $self->sanlog(-msg => "$groupName was successfully destroyed");
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   move
# Input:  -self diskGroupObj, -owner where to move the group
# Output: True or False
# Tasks:  Move the disk group and update the object
###########################################################################
sub move {
    my ($self, %args) = @_;

    my $sanClusterObj = $self->{sanClusterObj};
    my $groupName = $self->{name};
    my $originalOwner = $self->{currentOwner};
    my $originalStatus = $self->{status};

    my $moveToNode;
    foreach (keys %args) {
        if (/^-?owner/i) {
            if (not length($moveToNode = $args{$_})) {
                $self->_set_err("Invalid value for -owner: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # If the group is already on the requested node, just return
    if ($moveToNode =~ /$self->{currentOwner}/i) {
        $self->sanlog(-msg => "The group $groupName is already owned by the member node \"$moveToNode\".");
        return $ClientUtil::TRUE;
    }

    my $cmd = 'cluster . group "' . $groupName . '" /move';
    # If an owner is not supplied, it is not fatal. when the owner is not supplied
    # then the move command is issued without the argument, and the disk group is moved
    # off the current owner node, and onto ANY other cluster node.
    if (not length($moveToNode)) {
            $self->sanlog(cmd => "The -owner parameter was not supplied. The group " .
                                 "move command will be executed without a destination node.");
    } else {
        $cmd .= ':' . $moveToNode;
    }

    my (@output) = $sanClusterObj->execute(-cmd => $cmd);
    if ($sanClusterObj->error()) {
        $self->_set_err(@{$sanClusterObj->{errmsg}});
        $self->sanlog(-msg => @{$sanClusterObj->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Output example:
    # Moving resource group '04_disks'...

    # Group                Node            Status
    # -------------------- --------------- ------
    # 04_disks             WSHEDL380-05    Online

    my $status;
    my $newOwner;
    foreach my $line (@output) {
        if ($line =~ /^$groupName\s+(\S+)\s+(\w+)/) {
            $newOwner = $1;
            $status   = $2;
            last;
        }
    }

    # Checking the status and the output of the command
    if ((not length($newOwner)) or (not length($status))) {
        $self->_set_err("The new owner of $groupName could not be determined from the " .
                        "output of $cmd");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }
    # Now check to make sure the group indeed moved to another node, and
    # matches the status before the move (it could already be offline.)
    if ($newOwner =~ /$originalOwner/i) {
        $self->_set_err("The disk group $groupName failed to move. The starting owner " .
                        "$originalOwner and the resulting owner $newOwner are the same.");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }
    if ($status !~ /$originalStatus/i) {
        $self->_set_err(
            "The group $groupName status:$status does not match the status " .
            "before the move $originalStatus."
        );
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # We succeeded. Update the object
    $self->{currentOwner} = $newOwner;

    $self->sanlog(-msg => "The group $groupName was successfully move from $originalOwner " .
                         " to $newOwner.");
    return $ClientUtil::TRUE;
}

###########################################################################
# Name:   get_devices
# Input:  -self diskGroupObj
# Output: an array of resource objects associated with the diskGroupObj
# Tasks:  Find and return the array of disk group resources
###########################################################################
sub get_devices {
    my ($self) = @_;

    # It is entirely possible that a group does not contain any resources.
    return @{$self->{resources}};
}

1;
