#
# This library is used to populate the Windows Cluster object created by
# WinCluster.pm
#
#
## @summary
##  Creates Windows Cluter shared disk libraries and imports associated libraries.  All architecture
##  specific properties are defined here and extrapolated elsewhere.
##
## @author  Jonathan.Dascenzo@Netapp.com
## @status  public
##

package San::SanClientCluster::Windows::MSCS::SharedDiskGroup::Resource;

use strict;
use Error;
use WinClient;
use San::SanClient;
use San::SanLog;
use Tharn;

use vars qw( @ISA @EXPORT );
@ISA = qw( SanLog Error Tharn );

#############################################################
# Name: _new
# Input: -parent (diskGroupObj)
# Output: an empty diskGroupObj
# Tasks: * Build and return an empty resourceObj
#############################################################
sub _new {
    my ($class, %args) = @_;
    my $self = {};

    bless $self, $class;

    # clear the errors first
    $self->_clear_err();

    foreach (keys %args){
        if (/^-?parent/i) {
            if (not length($self->{diskGroupObj} = $args{$_}) ) {
                $self->_set_err("Invalid value for -parent: $args{$_}");
                return $self;
            }
            next;
         }
    }

    if (not length($self->{diskGroupObj})) {
        $self->_set_err("The required parameter -parent was not supplied");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $self;
    }

    # Define the structure of the disks data hash. Will populate later.
    $self->{name}           = undef;
    $self->{diskSignature}  = undef;
    $self->{groupName}      = undef;
    $self->{status}         = undef;
    $self->{type}           = undef;
    $self->{deviceId}       = undef;
    $self->{mount_pt}       = undef;
    $self->{quorum}         = undef;
    $self->{filer}          = undef;

    $self->sanlog( msg => "A new resource object was successsfully created." );

    return $self;
}

#############################################################
# Name: _load
# Input: -self (resourceObj), -name name of resource
# Output: True or False
# Tasks: * Discover attributes of resource (-name) and populate the object
#############################################################
sub _load() {
    my ($self, %args) = @_;
    my $sanClusterObj = $self->{diskGroupObj}->{sanClusterObj};
    my $resourceName;

    foreach (keys %args){
        if (/^-?name/i) {
            if (not length($resourceName = $args{$_}) ) {
                $self->_set_err("Invalid value for -name: $args{$_}");
                 return $ClientUtil::FALSE;
            }
            next;
        }
    }
    if (not length($resourceName)) {
        $self->_set_err("The required parameter -name was not supplied");
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $self->sanlog( msg => "Begin loading this object with resource $resourceName");

    # Now we gather all the relevant priv information about the resource $resourceName
    my $cmd = 'cluster . res "' . $resourceName .
              '"  /priv && ' .
              'cluster . res "' . $resourceName .
              '" /prop';
    my (@clusterPriv) = $sanClusterObj->execute(-cmd => $cmd);
    if ($sanClusterObj->error()) {
        return $ClientUtil::FALSE;
    }
    # The output we expect looks like this:
    # D  diskz                Signature                      4108365212 (0xf4e0ad9c)
    # D  diskz                SkipChkdsk                     0 (0x0)
    # D  diskz                ConditionalMount               1 (0x1)
    # S  diskz                SerialNumber                   HnX7S4LpqEmc
    # B  diskz                DiskUniqueIds                  10 00 00 00 ... (220 bytes)
    # B  diskz                MountVolumeInfo                9C AD E0 F4 ... (32 bytes)

    # Run the res command to get the group information
    $cmd = 'cluster . res "' . $resourceName . '"';
    my (@clusterRes) = $sanClusterObj->execute(-cmd => $cmd);
    if ($sanClusterObj->error()) {
        return $ClientUtil::FALSE;
    }

    # Next we can get the status, and the group name
    # This is taken from the cluster . res command
    my $status;
    my $groupName;

    # hunt through the resource array and return the line that
    # matches the disk name.
    my @resLine =  grep (/^$resourceName/i, @clusterRes);

    # The data looks like this
    # Listing status for resource 'diskz':\r\r\r\n
    #
    # Resource             Group                Node            Status\r\r\r\n
    # -------------------- -------------------- --------------- ------\r\r\r\n
    # diskz                05_Disks             WSHEDL380-05    Online\r\r\r\n

    # Split up all the data. We need to be able to handle spaces in the
    # Resource name field and the Group name field.
    # The grep above will return two lines of data, it will match
    # the header AND the data we want. This makes the line we are
    # interested in index 1.
    my @resLineWords = split(/\s+/, $resLine[0] );

    # Count the words in the resource name, this tells us
    # how many array members to shift off the beginning of the array
    my @resourceNameWords = split(/\s+/, $resourceName);

    # We can get the status value here, its the last value
    # We know that the last two values cannot contain spaces.
    $status = pop (@resLineWords);
    # We don't need the node argument here, so lose that
    pop (@resLineWords);
    # on the other end of the array, shift off the words that make the disk name
    foreach ( 1 .. scalar(@resourceNameWords) ) {
       shift (@resLineWords);
    }
    # The group name is all thats left
    $groupName = join (' ', @resLineWords);

    # Define the data we expect to be able to find in this array
    my $signature;
    my $type;
    my $deviceId;
    # on 2k8 and up the SerialNumber is no longer available from the output of
    # cluster.exe /priv or /prop. Use this value to note it is >=2k8 and we need to
    # do some extra magic to get this value on those systems.
    my $thisIs2k8 = $ClientUtil::FALSE;
    foreach my $line (@clusterPriv) {
        # Skip lines until we find one with the disk name in it
        next unless ($line =~ /$resourceName/);

      # On everything ->
            # S  DiskE                Type                           Physical Disk
        if ($line =~ /^S.+Type\s+(.+)$/) {
            $type = $1;
      # <- end everything

      # On 2k3 only  ->
            # M  diskE                MPVolGuids                     7e00 \??\Volume{60970921-ba31-11dd-b629-001b2106a255}
        } elsif ($line =~ /^M.+MPVolGuids\s+\w+.+Volume{(.+)}$/) {
            $deviceId = $1;

            # D  DiskE                Signature                      1809637504 (0x6bdce080)
        }elsif ( $line =~ m/^D\s+$resourceName\s+Signature\s+\d+\s+\(0x(\w+)\)$/ ) {
            $signature = $1;
      # <- end 2k3 only

      # On 2k8 and up ->
            # S  Cluster Disk 21      DiskIdGuid                     {e95d2e2b-620b-4eaa-aee6-694f0f6c61f5}
        } elsif ($line =~ /^S\s+$resourceName\s+DiskIdGuid\s+({.+})$/) {
            $deviceId = $1;
            $thisIs2k8 = $ClientUtil::TRUE;

            # D  Cluster Disk 10      DiskSignature                  441922291 (0x1a5732f3)
        } elsif ($line =~ /^D\s+$resourceName\s+DiskSignature\s+(\d+)\s+\(0x\w+\)$/) {
            $signature = $1;
        }
     # <- end 2k8 and up
    }
    # This mount point is the last thing to get
    # This data structure on the sanClusterObj is populated from
    # $nodeObj->_get_deviceId_map(). First check the signature,
    # if the signature is blank, use the guid. If this is 2k3 or
    # 2k8 mbr the signature will be defined. 2k8 GPT uses the GUID.
    my $mount_pt;
    if (length($sanClusterObj->{deviceId_map}->{$signature})) {
       $mount_pt = $sanClusterObj->{deviceId_map}->{$signature};
    } else {
       $mount_pt = $sanClusterObj->{deviceId_map}->{$deviceId};
    }

    # Fill out the object with all our cool new data
    $self->{name} = $resourceName;
    $self->{diskSignature} = $signature;
    $self->{groupName} = $groupName;
    $self->{status} = $status;
    $self->{type} = $type;
    $self->{deviceId} = $deviceId;
    $self->{mount_pt} = $mount_pt;
    $self->{quorum} = $ClientUtil::FALSE;

    # Lets check all the variables to make sure we didn't miss anything.
    # We will only check for objects of type "Physical Disk". Other resource
    # types will not be fully populated, but may be further utilized in the future
    if ($self->{type} =~ /Physical\s+Disk/i) {
        foreach my $key (keys %{$self}) {
            # Skip the errmsg field
            next if ($key =~ /errmsg/i);
            # The filer may not yet be filled out
            next if ($key =~ /filer/i);
            # skip the mount_pt field. This may not be configured yet. Especially on 2k8. This is becasue
            # the initial call from cluster_create to _create_deviceId_map is called with the -NoCluster option
            # because the cluster does not exist yet. In this case the MSCluster tree in WMI is empty. This means we
            # have to call the Win32_DiskPartition tree to find the disk identifiers for the physical disks
            # which belong to the node. We expect to have the cluster_create code path call _find_cluster_disk_mounts
            # at some point if the disk object mount_pt values are not populated from information that we are able to
            # to get from cluster.exe, such as when using 2k3.
            next if ($key =~ /mount_pt/i);
            # the deviceId will not be populated if the disk is 2k8 MBR
            next if (($key =~ /deviceId/) and (! $sanClusterObj->{build} >= $sanClusterObj::WIN2K8BUILD));

            if (! length($self->{$key})) {
                $self->_set_err("The resource object field $key for resource $resourceName was not populated.");
                $self->sanlog(-msg => @{$self->{errmsg}});
                return $ClientUtil::FALSE;
            }
        }
    }
    $self->sanlog(-msg => "The resourceObj now contains data related to $resourceName");
    return $ClientUtil::TRUE;
}

#############################################################
# Name: create_disk
# Input: -self (resourceObj)
#        -id the signature or guid of the disk
#        -name name of the disk to create
# Output: True or False
# Tasks: * create a new shared disk resource
#############################################################
sub create_disk {
    my ($self, %args) = @_;

    my $diskGroupObj = $self->{diskGroupObj};
    my $groupName = $diskGroupObj->{name};
    my $sanClusterObj = $diskGroupObj->{sanClusterObj};
    my $primaryNodeObj = $sanClusterObj->get_primary_node();
    my $build = $sanClusterObj->{build};
    my $diskId;
    my $diskName;

    foreach (keys %args) {
        if (/^-?name/i) {
            if (not length($args{$_}) ) {
                $self->_set_err("Invalid value for -name: $args{$_}");
                 return $ClientUtil::FALSE;
            } else {
                 $diskName = $args{$_};
            }
            next;
        } elsif (/^-?id/i) {
            if (!length($args{$_}) ) {
                $self->_set_err("Invalid value for -id: $args{$_}");
                 return $ClientUtil::FALSE;
            } else {
                 $diskId = $args{$_};
            }
            next;
       }
    }
    if (not length($diskName)) {
        $self->_set_err("The required parameter -name was not supplied");
        $self->sanlog(@{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }
    if (not length($diskId)) {
        $self->_set_err("The required parameter -id was not supplied. This " .
          "option can be either the disk signature for mbr disks, or a GUID for GPT type" .
           "devices. GPT devices are only supported by windows 2k8 and up.");
        $self->sanlog(@{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # Check the format of the diskId/Signature. If the string is encased in curly
    # brackets, then it is a disk GUID (GPT disk) and a different command must be
    # supplied to create the cluster disk. Bail out if we are not at least windows 2k8.
    # GPT type disks are only supported for clusters in 2k8+
    $build = $primaryNodeObj->{sanClientObj}->{WinClient}->{build};
    my $diskCreate;
    if ($diskId =~ /{.+}/) {
        # The disk is a GPT type disk
        if (not int($build) >= 6 ) {
            # This is NOT at least 2k8
            $sanClusterObj->_set_err("An attempt was made to create a GPT cluster disk on " .
             "windows 2003 build: $build. This configuration is not supported");
            $sanClusterObj->sanlog(-msg => @{$sanClusterObj->{errmsg}});
            return $ClientUtil::FALSE;
        }
        $sanClusterObj->sanlog(-msg => "Creating a GPT type cluster disk");
        $diskCreate = "DiskIdGuid=$diskId";

    } else {
        # The disk is using an old style signature and is MBR
        $sanClusterObj->sanlog(-msg => "Creating a MBR type cluster disk");
        if (int($build) >= 6) {
            # 2k8 and up
            $diskCreate = "diskSignature=0x$diskId";
        } else {
            # 2k3
            $diskCreate = "signature=0x$diskId";
        }
    }
    # Command to create the cluster
    my $cmd = 'cluster . res "' . $diskName . '" ' .
              '/create /group:"' . $groupName . '" ' .
              '/type:"physical disk" && cluster . res "' . $diskName . '" ' .
              '/priv '. $diskCreate .
              ' && cluster . res "' . $diskName . '" /On';

    # You may fire when ready.
    my (@output,$status) = $sanClusterObj->execute(-cmd => $cmd);

    if (grep /error/i, @output) {
        $self->_set_err("An error was received creating disk \"$diskName\"." );
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "The disk \"$diskName\" was created successfully");

    # Associate the object we received with this newly created disk
    $self->_load(-name => $diskName);

    if ($self->error()) {
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

#############################################################
# Name: move
# Input: -self (resourceObj)
#        -group name of the group to move the resource into
# Output: True or False
# Tasks: * Move resource to a new group
#        * Destroy the group we received, and create a new one to put on the new group
#############################################################
sub move {
    my ($self, %args) = @_;

    my $diskGroupObj = $self->{diskGroupObj};
    my $sanClusterObj = $diskGroupObj->{sanClusterObj};
    my $resName = $self->{name};

    my $moveTo;
    foreach (keys %args) {
        if (/^-?group/i) {
            if (not length($moveTo = $args{$_}) ) {
                $self->_set_err("Invalid value for -group: $args{$_}");
                 return $ClientUtil::FALSE;
            }
        }
    }
    if (not length($moveTo)) {
        $self->_set_err("The required parameter -group was not supplied.");
        $self->sanlog(@{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $sanClusterObj->sanlog(-msg => "Moving resource \"$resName\" to group \"$moveTo\".");

    my $cmd = 'cluster . res "' . $resName . '" /move:"' . $moveTo . '"';

    my @output = $sanClusterObj->execute(-cmd => $cmd);

    if (grep /error/i, @output) {
        $self->_set_err("An error was received moving \"$resName\" to group " .
          "\"$moveTo\"." );
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # get an object for the destination group
    my $destinationGroupObj = $sanClusterObj->get_disk_group(-name => $moveTo);

    # Update the resourceObj with the new group name and point
    # point to the new parent group object
    $self->{groupName} = $moveTo;
    $self->{diskGroupObj} = $destinationGroupObj;

    # push the resource onto the object for the destination group
    push (@{$destinationGroupObj->{resources}}, $self);

    # remove the resourceGroupObj from the originating group
    my $counter = 0;
    foreach my $res (@{$diskGroupObj->{resources}}) {
        if ($res->{name} =~ /$resName/) {
            splice (@{$diskGroupObj->{resources}}, $counter, 1);
            last;
        }
        $counter++;
        # If we get to the last array element, there was a problem removing the
        # disk from the old group.
        if ($res->{name} =~ /$diskGroupObj->{resources}[-1]->{name}/) {
            $self->_set_err("There was a problem removing $res->{name} from the " .
              "originating group \"$diskGroupObj->{name}\".");
            $self->sanlog(-msg => @{$self->{errmsg}});
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "Resource \"$self->{name}\" was successfully moved to " .
                          "group \"$moveTo\".");

    return $ClientUtil::TRUE;
}

#############################################################
# Name: rename
# Input: -self (resourceObj)
#        -name The new name for the the resource
# Output: True or False
# Tasks: * assign a new name to resource
#############################################################
sub rename {
    my ($self, %args) = @_;

    my $name;
    my $diskGroupObj = $self->{diskGroupObj};
    my $sanClusterObj = $diskGroupObj->{sanClusterObj};
    my $originalName = $self->{name};

    foreach (keys %args) {
        if (/^-?name/i) {
            if (not length($name = $args{$_}) ) {
                $self->_set_err("Invalid value for -name: $args{$_}");
                 return $ClientUtil::FALSE;
            }
        }
    }
    if (not length($name)) {
        $self->_set_err("The required argument -name was not supplied.");
        $self->sanlog(@{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $sanClusterObj->sanlog(-msg => "\"Renaming $self->{name}\" to \"$name\".");

    my $cmd = 'cluster . res "' . $self->{name} . '" /REN:"' . $name . '"';

    my @output = $sanClusterObj->execute(-cmd => $cmd);

    if (grep /error/i, @output) {
        $self->_set_err("An error was received renaming $self->{name} to " .
                        "\"$name\"." );
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    # No rediscovery here, this is just a single attribute change
    $self->{name} = $name;

    $self->sanlog(-msg => "Resource \"$originalName\" was successfully renamed as \"$name\".");
}

#############################################################
# Name: offline
# Input: -self (resourceObj)
#        -name The name for the the resource to offline
# Output: True or False
# Tasks: * offline a resource
#############################################################
sub offline {
    my ($self, %args) = @_;

    my $name;
    my $diskGroupObj = $self->{diskGroupObj};
    my $sanClusterObj = $diskGroupObj->{sanClusterObj};

    $sanClusterObj->sanlog(-msg => "Offlining " . $self->{name});

    my $cmd = 'cluster . res "' . $self->{name} . '" /OFF';

    my @output = $sanClusterObj->execute(-cmd => $cmd);

    if (grep /error/i, @output) {
        $self->_set_err("An error was received offlining " . $self->{name});
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Resource " . $self->{name} . "was successfully offlined.");
    return $ClientUtil::TRUE;
}

#############################################################
# Name: online
# Input: -self (resourceObj)
#        -name The name for the the resource to offline
# Output: True or False
# Tasks: * online a resource
#############################################################
sub online {
    my ($self, %args) = @_;

    my $name;
    my $diskGroupObj = $self->{diskGroupObj};
    my $sanClusterObj = $diskGroupObj->{sanClusterObj};

    $sanClusterObj->sanlog(-msg => "Onlining " . $self->{name});

    my $cmd = 'cluster . res "' . $self->{name} . '" /ON';

    my @output = $sanClusterObj->execute(-cmd => $cmd);

    if (grep /error/i, @output) {
        $self->_set_err("An error was received onlining " . $self->{name});
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Resource " . $self->{name} . "was successfully onlined.");
    return $ClientUtil::TRUE;
}

#############################################################
# Name: online
# Input: -self (resourceObj)
#        -name The name for the the resource to offline
# Output: True or False
# Tasks: * online a resource
#############################################################
sub destroy {
    my ($self, %args) = @_;

    my $name;
    my $diskGroupObj = $self->{diskGroupObj};
    my $sanClusterObj = $diskGroupObj->{sanClusterObj};

    $sanClusterObj->sanlog(-msg => "Destroying " . $self->{name});

    my $cmd = 'cluster . res "' . $self->{name} . '" /DEL';

    my @output = $sanClusterObj->execute(-cmd => $cmd);

    if (grep /error/i, @output) {
        $self->_set_err("An error was received destroying " . $self->{name});
        $self->sanlog(-msg => @{$self->{errmsg}});
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Resource " . $self->{name} . "was successfully destroyed.");
    return $ClientUtil::TRUE;
}
1;
