# Copyright (c) 2007-20010 NetApp Inc.
# All rights reserved.
#
## @summary Abstraction layer to instantiate either SanFiler7M or Filer10M object
## @author korry@netapp.com dl-san-automation
## @status shared

package SanFiler;

# Compiler directives.
use strict;

# Module import.
use Tharn;
use San::SanLog;
use Utils qw( make_host );
use Error;
use Exporter;
use Data::Dumper;

use vars qw( @EXPORT @ISA );

@ISA = qw( Error SanLog );

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new controller object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($proto, %args) = @_;
    my $class = ref($proto) || $proto;
    my ($filer, $cserver, $host_obj);

    # init variables
    my $self = {};
    bless $self,$class;

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();
    # Process only mandatory arguments. Allow other arguments to pass thru.
    # Pass thru arguments will be processed by SanFiler7M or Filer10M.
    # If an error occurs during the arg processing,
    # The error flag and errmsg values are set and the object is returned.
    # It is up to the caller to check for errors after the new call.
    foreach (keys %args) {
        if (/-?filer$/) {
            length($args{$_}) or
                do {
                    $self->_set_err("invalid value for \"-filer\": $args{$_}");
                    return $self;
                };
        }
        if (/-?cserver$/) {
            length($args{filer} = delete $args{$_}) or
                do {
                    $self->_set_err("invalid value for \"-cserver\": $args{$_}");
                    return $self;
                };
            $args{is_cserver} = $SanUtil::TRUE;
        }
    }

    # Verify mandatory parameters.
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" or \"-cserver\" not specified");
        return $self;
    }

    if ($args{is_cserver}) {

        # create host object for cserver dynamically via first node
        $host_obj = &SanUtil::create_host_obj(-cserver => $args{filer},
                                              -cluster_mgmt_ip => $args{cluster_mgmt_ip},
                                              -nodes_in_cluster => $args{nodes_in_cluster});
    } else {

        # create host object for filer
        $host_obj = &SanUtil::create_host_obj(-filer => $args{filer});
    }

    if ($host_obj->hosttype() =~ /^filer-ng|^vsim-ng/) {
        require San::ControllerCluster::Filer10M;
        $self = San::ControllerCluster::Filer10M->new(%args, node => $host_obj->{node});
        # check to make sure Node object is created without error
        if ($self->error()) {
            $self->_set_err("can not create Filer10M object ", $self->errmsg());
            return $self;
        }
    } elsif ($host_obj->hosttype() =~ /^filer-cfe|^vsim-cfe|^filer-vsim|^filer-vsa|^filer$/i) {
        require San::SanFiler7M;
        $self = San::SanFiler7M->_new(%args);
        # check to make sure SanFiler7M object is created without error
        if ($self->error()) {
            $self->_set_err("can not create SanFiler7M object ", $self->errmsg());
            return $self;
        }
    } else {
        $self->sanlog(-msg => "Filer is not of recognized type, making" .
                              "a scsit simulator faux filer instead.");
        require San::ScsitSimulator;
        $self = San::ScsitSimulator->_new(%args);
    }

    return $self;
} # end new()



###############################################################################
#                               Private Methods
###############################################################################


#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

SanFiler.pm

=head1 Author

Korry Nguyen

=head1 Description

SanFiler.pm provides an abstraction interface to either SanFiler7M or Filer10M object based on filer type.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<San::SanFiler7M | San::ControllerCluster::Filer10M>

B<Error.pm>

B<SanLog.pm>

=head1 Attributes

=head2 B<filer>

string = Controller name

=head2 B<NOTE:>

For additional attributes, please see San:SanFiler7M or San::ControllerCluster::Filer10M module to learn more about attributes that can be passed in.

=cut

1;
