# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/San/SanFiler/Cmds/Portset.pm#1 $

# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
# 
# test/lib/San/SanFiler/Cmds/Portset.pm
##  
## @summary Utility for testing portset functionality
## @description
##    This utility to provide varity of function to test the 
##    portset functionality. This library uses the CmdsDriver.pm 
##    to execute the portset commands for both CLI and ZAPI.
##
## @description Objective:
## To provide library functions for the testing porset. All the 
## porset testing srcipt will use this library
## @author Muthukumaran Govindaraj (muthug@netapp.com)
## @status public
## @change 2007-06-07 muthug:      Original template created
## @change 2007-06-25 muthug:      Changes made to use CmdsDriver.pm
## @change 2008-01-15 pressley:    Integrated with SanFiler
## @change 2008-03-26 muthug:     changes made in igroup create - ostype to os-type to support AS and TT

package Portset;

#
# Compiler directives.
#
use strict;

#
# Module import.
#
use San::SanLog;
use San::SanUtil;
use Error;
use CmdsDriver;
use Exporter;

use vars qw(
            @EXPORT
            @ISA
           );

#
# Inherit from the Error, SanLog, and SanUtil classes
#
@ISA = qw(Error SanLog SanUtil) ;


###############################################################################
# Subroutine Name:  new_portset()
# Objective:        Create a new portset object 
# Details:          Class Constructor, see POD below for full details
###############################################################################
sub new_portset 
{
    my ($class, %args) = @_;
    my (
        $filer_conn,
        $cmd_obj,
        $filer,
        $debug_level
       );

    my $self = bless ( {}, $class);
    
    $self->sanlog(-msg => "portset new");
    $self->_clear_err();

    my $debug_level   = 2;
    my $log           = \*ntest::NTEST_STDOUT;
    my $filer_version = "";
    my $interface     = "cli"; # interface is set in here and never reset
    
    #
    # parse arguments
    #
    foreach (keys %args) 
    {
        if (/^-?filer_conn$/) 
        {
            if (not length($filer_conn = $args{$_})) 
            {
                my $errmsg = "Error: Portset::new_portset() Invalid value ".
                             "for argument '-filer_conn'";
                $self->_set_err($errmsg);
                return(0, $errmsg); 
            }          
            next;
        }
        elsif (/^-?interface$/) 
        {
            if (not length($interface = $args{$_})) 
            {
                my $errmsg = "Error: Portset::new_portset() Invalid value ".
                             "for argument '-interface'";
                $self->_set_err($errmsg);
                return(0, $errmsg); 
            }          
            next;
        }
        elsif (/^-?cmd_obj$/)
        {
            if (not length($cmd_obj = $args{$_})) 
            {
                my $errmsg = "Error: Portset::new_portset() Invalid value ".
                             "for argument '-cmd_obj'";
                $self->_set_err($errmsg);
                return(0, $errmsg); 
            }          
            next;
        }
        elsif (/^-?filer_version$/)
        {
            if (not length($filer_version = $args{$_}) or $args{$_} !~ /\d+/)
            {
                $self->_set_err("Error: Portset::new_portset() Invalid value ".
                                "for argument '-filer_version'");
                return $self; 
            }          
            next;
        }
        elsif (/^-?debug_level$/)
        {
            if (not length($debug_level = $args{$_})) 
            {
                $self->_set_err("Error: Portset::new_portset() Invalid value ".
                                "for argument '-debug_level'");
                return $self; 
            }          
            next;
        }
        elsif (/^-?log$/)
        {
            if (not length($log = $args{$_})) 
            {
                $self->_set_err("Error: new_portset() Invalid value ".
                                "for argument '-log'");
                return $self;
            }
            next;
       }
        elsif (/^-?filer$/)
        {
            if (not length($filer = $args{$_})) 
            {
                $self->_set_err("Error: Portset::new_portset() Invalid value ".
                                "for argument '-filer'");
                return $self; 
            }          
            next;
        }
        else 
        {
            $self->_set_err("Error: Portset::new_portset: Invalid parameter " .
                            "specified: $_");
            return $self;
        }
    }

    $self->{connection}     = $filer_conn;
    $self->{filer_version}  = $filer_version;
    $self->{debug_level}    = $debug_level;
    $self->{interface}      = $interface;
    $self->{cmd_obj}        = $cmd_obj;
    $self->{log}            = $log;

    #
    # Check for required parameters -filer or -filer_conn
    #
    unless (defined($filer) or defined($filer_conn))
    {
        $self->_set_err("Error: Portset::new_portset:  User must supply " .
                        "either '-filer' or '-filer_conn' ");
        return $self; 
    }

    #
    # Get and/or set the filer name
    #
    if(!defined($filer))
    {
      $self->{name} = $self->{connection}->hostp->hostname;
    }
    else
    {
        $self->{name} = $filer;
    }

    #
    # Make a connection if filer_conn is not passed in
    # Note, filer_connect comes from SanUtil.pm
    #
    unless (defined($self->{connection}))
    {
        $self->{connection} = $self->filer_connect();
        return $self if ($self->error);
    }

    #
    # Create a CmdsDriver object if we not given one
    #
    if(!defined($self->{cmd_obj})) 
    { 
        $self->{cmd_obj} = new CmdsDriver(
                                -filer          => $self->{name},
                                -log            => $log,
                                -debug_level    => $debug_level);

        if ($self->{cmd_obj}->error())
        {
            my $errmsg = "Error: Portset::new_portset: "            .
                         "Failed to instantiate CmdsDriver object " .
                         join("\n", $self->{cmd_obj}->errmsg());
            $self->{cmd_obj} = undef;
            $self->_set_err($errmsg);
            return $self, $errmsg;
        }
    }

    $self->{cmd_obj}->filer($self->{name});
    $self->{cmd_obj}->connection($self->{connection});
    $self->{cmd_obj}->interface($self->{interface});

    #
    # If filer_version wasn't given,
    # discover version otherwise set filer version
    #
    $self->sanlog(-msg => ref($self)                        . 
                          "::Portset_new() filer_verion = " .
                          $self->{filer_version},
                  -msg_level => 2);


    if ($self->{filer_version} eq "")
    {
        $self->{cmd_obj}->discover_filer_version($self->{name});
        $self->{filer_version} = $self->{cmd_obj}->filer_version();
    } else {
        $self->{cmd_obj}->filer_version($self->{filer_version});
    }

    return $self;
} # end new_portset()

###############################################################################
# Subroutine Name:  set_cli_partner_mode()
# Objective:        Set partner mode for the cmdsDriver object
# Details:          See POD documentation
###############################################################################
sub set_cli_partner_mode() 
{
    my ($self, $mode) = @_;
    
    $self->sanlog(-msg => "set cli partner mode");
    $self->_clear_err();
    
    if($mode != 1)
    {
        $mode = 0;
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::set_cli_partner_mode CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    $self->{cmd_obj}->partner_mode($mode);
} # end set_cli_partner_mode()

###############################################################################
# Subroutine Name:  get_filer_ontap_version()
# Objective:        access filer-version attribute
# Details:          See POD documentation
###############################################################################
sub get_filer_ontap_version {
my ($self) = @_;
    return $self->{filer_version};
}

###############################################################################
# Subroutine Name:  portset_create()
# Objective:        create a portset on the filer 
# Details:          See POD documentation
###############################################################################
sub portset_create()
{
    my ($self, %args) = @_;
    my (
        $portset_name,
        $ports_ref,
        $portset_type,
       );
    
    $self->sanlog(-msg => "portset create");
    $self->_clear_err();

    #
    # $ports_ref (argument should be a list reference of ports info) and it is 
    # defualt for cli to create portset without port, ZAPI only creates portsets
    # without ports.
    #
    # parse arguments
    #
    foreach (keys %args) 
    {
        if (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_create() " .
                             "Invalid value for argument " .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?ports$/)
        {
            if (not length($ports_ref = $args{$_}))
            {
                my $errmsg = "Error: Portset::portset_create() " .
                             "Invalid value for argument " .
                             "'-ports'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }
            next;
        }
        elsif (/^-?portset_type$/)
        {
            if (not length($portset_type = $args{$_}))
            {
                my $errmsg = "Error: Portset::portset_create() " .
                             "Invalid value for argument " .
                             "'-portset_type'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_create() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_create: CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    my $interface = $self->{cmd_obj}->interface();


    #
    # Verify mandatory parameters
    #
    if (!defined ($portset_name)) 
    {
        $self->_set_err("Error: Portset::portset_create(): " .
                        "-portset_name is a required argument");
        return (0, "Portset name(portset_name) is a required argument");
    }
    if (!defined ($portset_type)) 
    {
        $self->_set_err("Error: Portset::portset_create(): " .
                        "Portset type(portset_type) is a required argument");
        return (0, "Portset type(portset_type) is a required argument");
    } 
    elsif($interface =~ m/cli/i)
    {
        if( $portset_type =~ /fcp/i) 
        {
            $portset_type = "-f";
        } 
        elsif ($portset_type =~ /iscsi/i) 
        {
            $portset_type = "-i";
        }
    }

    my $ports = join(' ', @$ports_ref) if (defined($ports_ref));

    my @cmds_args = ("-cmd", "portset-create");
    push (@cmds_args, "-portset-name", $portset_name);
    push (@cmds_args, "-portset-type", $portset_type);
    push (@cmds_args, "-portset-port-name", $ports) if (defined($ports_ref) && 
                                                        $interface =~ m/cli/i);
    push (@cmds_args, "-check_output", "1");
    push (@cmds_args, "-check_result", "1");

    my $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
     
    if ($self->{cmd_obj}->error()) 
    {
        my $errmsg = "Portset::portset_create: "   .
                     "Failed calling execute_cmd " .
                     "using $interface:\n" .
                     "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    #
    # Check that the result and output aren't wrong
    #
    if (!$resultsHash_ref->{result_check} || !$resultsHash_ref->{output_check}) 
    {
        my $errmsg = "Portset::portset_create: "     .
                     "Result or output check failed" .
                     "Output is "                    .
                     $resultsHash_ref->{output};
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    if(defined($ports_ref) && $interface =~ m/zapi/i) 
    {
        my ($result, $text) = $self->portset_add(
                                        portset_name    => $portset_name, 
                                        ports           => \@$ports_ref);
        if(!$result)
        {
            $self->_set_err("Error: Portset::portset_create() adding ports: $text");
            return (0, $text);
        }
    }

    return (1, $resultsHash_ref->{output});
} # end portset_create()

###############################################################################
# Subroutine Name:  portset_add()
# Objective:        Add ports to a portset 
# Details:          See POD documentation
###############################################################################
sub portset_add()            
{
    my ($self, %args) = @_;
    my (
        $portset_name,
        $ports_ref,
       );

    $self->sanlog(-msg => "portset add");
    $self->_clear_err();
    
    # $ports_ref (argument should be a list reference of ports info)
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_add() " .
                             "Invalid value for argument "    .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?ports$/)
        {
            if (not length($ports_ref = $args{$_}))
            {
                my $errmsg = "Error: Portset::portset_add() " .
                             "Invalid value for argument "    .
                             "'-ports'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_add() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Verify mandatory parameters
    #
    if (!defined ($portset_name)) 
    {
        my $errmsg = "Error: Portset::portset_add() " .
                     "Portset name(portset_name) is required argument";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }
    if (!defined ($ports_ref)) 
    {
        my $errmsg = "Error: Portset::portset_add() " .
                     "Portset ports(ports) is required argument";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_add CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    my $resultsHash_ref;
    my $interface = $self->{cmd_obj}->interface();

    #
    # ZAPI and CLI need different formats for parameters
    #
    if( $interface =~ m/zapi/i) 
    {
        foreach my $port (@$ports_ref) 
        {
            my @cmds_args = ("-cmd", "portset-add");
            push (@cmds_args, "-portset-name", $portset_name);
            push (@cmds_args, "-portset-port-name", $port);
            push (@cmds_args, "-check_output", "1");
            push (@cmds_args, "-check_result", "1");

            $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
            
            if ($self->{cmd_obj}->error()) 
            {
                my $errmsg = "Failed calling execute_cmd for 'portset-add' " .
                             "using $interface:\n" .
                             "Reason: " . 
                             join("\n", $self->{cmd_obj}->errmsg());
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }

            #
            # Check that the result and output aren't wrong
            #
            if (!$resultsHash_ref->{result_check} || 
                !$resultsHash_ref->{output_check}) 
            {
                my $errmsg = "Portset::portset_add "     .
                             "Result or output check failed" .
                             "Output is "                    .
                             $resultsHash_ref->{output};
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }
        }
    } 
    else  
    {
        my $ports = join(' ', @$ports_ref) if (defined($ports_ref));
        my @cmds_args = ("-cmd", "portset-add");
        push (@cmds_args, "-portset-name", $portset_name);
        push (@cmds_args, "-portset-port-name", $ports);
        push (@cmds_args, "-check_output", "1");
        push (@cmds_args, "-check_result", "1");

        $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
        if ($self->{cmd_obj}->error()) 
        {
            my $errmsg = "Failed calling execute_cmd for 'portset-add' " .
                         "using $interface:\n" .
                         "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }

        #
        # Check that the result and output aren't wrong
        #
        if (!$resultsHash_ref->{result_check} || 
            !$resultsHash_ref->{output_check}) 
        {
                my $errmsg = "Portset::portset_add "     .
                             "Result or output check failed" .
                             "Output is "                    .
                             $resultsHash_ref->{output};
                $self->_set_err($errmsg);
                return (0, $errmsg);
        }
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  portset_destroy()
# Objective:        Destroy a portset 
# Details:          See POD documentation
###############################################################################
sub portset_destroy()
{
    my ($self, %args) = @_;
    my (
        $portset_name,
        $force,
       );
    
    $self->sanlog(-msg => "portset destroy");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_destroy() " .
                             "Invalid value for argument "        .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?force$/)
        {
            if (not length($force = $args{$_}))
            {
                my $errmsg = "Error: Portset::portset_destroy() " .
                             "Invalid value for argument "        .
                             "'-force'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_destroy() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Verify mandatory parameters
    #
    if (!defined ($portset_name)) 
    {
        $self->_set_err("Portset name(portset_name) is required argument");
        return (0, "Portset name(portset_name) is required argument");
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_destroy CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    my $interface = $self->{cmd_obj}->interface();

    my @cmds_args = ("-cmd", "portset-destroy");
    push (@cmds_args, "-portset-name", $portset_name);
    push (@cmds_args, "-check_output", "1");
    push (@cmds_args, "-check_result", "1");

    #
    # CLI and ZAPI need different parameters for force
    #
    if( defined($force) && $force =~ /(yes|true|1)/i ) 
    {
        push (@cmds_args, "-force", "") if($interface =~ m/cli/i);
        push (@cmds_args, "-force", $force) if($interface =~ m/zapi/i);
    } 
    elsif(defined($force))
    {  # useful for negatine test cases
        push (@cmds_args, "-force", $force);
    }

    my $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
    if ($self->{cmd_obj}->error()) 
    {
        my $errmsg = "Failed calling execute_cmd for 'portset-destroy' " .
                     "using $interface:\n".
                     "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    #
    # Check that the result and output aren't wrong
    #
    if (!$resultsHash_ref->{result_check} || !$resultsHash_ref->{output_check}) 
    {
        my $errmsg = "Portset::portset_destroy "     .
                     "Result or output check failed" .
                     "Output is "                    .
                     $resultsHash_ref->{output};
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  portset_remove()
# Objective:        Remove ports from a portset 
# Details:          See POD documentation
###############################################################################
sub portset_remove()
{
    my ($self, %args) = @_;
    my (
        $portset_name,
        $ports_ref,
       );
    
    $self->sanlog(-msg => "portset remove");
    $self->_clear_err();
    
    # $ports_ref (argument should be a list reference of ports info)
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_remove() " .
                             "Invalid value for argument "       .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?ports$/)
        {
            if (not length($ports_ref = $args{$_}))
            {
                my $errmsg = "Error: Portset::portset_remove() " .
                             "Invalid value for argument "       .
                             "'-ports'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_remove() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Verify required parameters
    #
    if (!defined ($portset_name)) 
    {
        $self->_set_err("Portset name(portset_name) is required argument");
        return (0, "Portset name(portset_name) is required argument");
    }
    if (!defined ($ports_ref)) 
    {
        $self->_set_err("Portset ports(ports) is/are required argument");
        return (0, "Portset ports(ports) is/are required argument");
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_remove CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    my $resultsHash_ref;
    my $interface = $self->{cmd_obj}->interface();
    
    #
    # ZAPI and CLI have different command formats
    #
    if($interface =~ m/zapi/i) 
    {
        foreach my $port (@$ports_ref) 
        {
            my @cmds_args = ("-cmd", "portset-remove");
            push (@cmds_args, "-portset-name", $portset_name);
            push (@cmds_args, "-portset-port-name", $port);
            push (@cmds_args, "-check_output", "1");
            push (@cmds_args, "-check_result", "1");

            $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
            
            if ($self->{cmd_obj}->error()) 
            {
                my $errmsg = "Failed calling execute_cmd for " .
                             "'portset-remove' using $interface:\n".
                             "Reason: " . 
                             join("\n", $self->{cmd_obj}->errmsg());
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }

            #
            # Check that the result and output aren't wrong
            #
            if (!$resultsHash_ref->{result_check} || 
                !$resultsHash_ref->{output_check}) 
            {
                my $errmsg = "Portset::portset_remove "      .
                             "Result or output check failed" .
                             "Output is "                    .
                             $resultsHash_ref->{output};
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }
        }
    } 
    else 
    {
        my $ports = join(' ', @$ports_ref) if (defined($ports_ref));
        my @cmds_args = ("-cmd", "portset-remove");
        push (@cmds_args, "-portset-name", $portset_name);
        push (@cmds_args, "-portset-port-name", $ports);
        push (@cmds_args, "-check_output", "1");
        push (@cmds_args, "-check_result", "1");
        
        $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
        if ($self->{cmd_obj}->error()) 
        {
            my $errmsg = "Failed calling execute_cmd for " .
                         "'portset-remove' using $interface:\n".
                         "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }

        #
        # Check that the result and output aren't wrong
        #
        if (!$resultsHash_ref->{result_check} || 
            !$resultsHash_ref->{output_check}) 
        {
            my $errmsg = "Portset::portset_remove "      .
                         "Result or output check failed" .
                         "Output is "                    .
                         $resultsHash_ref->{output};
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  portset_show()
# Objective:        Display information on a portset 
# Details:          See POD documentation
###############################################################################
sub portset_show()
{
    #
    # TBD: Currently the ZAPI returns a hash for output... this may
    # eventually need to be translated into text
    #
    my ($self, %args) = @_;
    my (
        $portset_name,
       );
    
    $self->sanlog(-msg => "portset show");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?portset_name$/i) 
        {
            $portset_name = $args{$_};      
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_show() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_show CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }
    my $interface = $self->{cmd_obj}->interface();

    my @cmds_args = ("-cmd", "portset-list-info");
    push (@cmds_args, "-portset-name", $portset_name) if(defined($portset_name));
    push (@cmds_args, "-check_output", "1");
    push (@cmds_args, "-check_result", "1");

    my $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);

    if ($self->{cmd_obj}->error()) 
    {
        my $errmsg = "Failed calling execute_cmd for 'portset-list-info' " .
                     "using $interface:\n".
                     "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
        $self->_set_err($errmsg);  
        return (0, $errmsg);
    }

    #
    # Check that the result and output aren't wrong
    #
    if (!$resultsHash_ref->{result_check} || !$resultsHash_ref->{output_check}) 
    {
        my $errmsg = "Portset::portset_show "      .
                     "Result or output check failed" .
                     "Output is "                    .
                     $resultsHash_ref->{output};
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  get_portset_info()
# Objective:        Handle information for a portset 
# Details:          See POD documentation
###############################################################################
sub get_portset_info() 
{
    my ($self, %args) = @_;
    my (
        $portset_name,
       );

    $self->sanlog(-msg => "get portset info");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::get_portset_info() " .
                             "Invalid value for argument "         .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::get_portset_info() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    my %portset = ();

    my ($result, $output) = $self->portset_show(portset_name=>$portset_name);
    if(!$result)
    {
        my $errmsg = "Error: Portset::get_portset_info: $output";
        $self->_set_err($errmsg);
        return ($result, $errmsg, %portset);
    }

    #
    # The output needs to be handled differently for both
    # interfaces
    #
    my $cmd_interface = $self->{cmd_obj}->interface();
    if($cmd_interface =~ m/cli/i) 
    {
        my $dummy_delim = "DUMMY__DELIMITER";
        $output =~s/(\S+ \(\w+\):)/$dummy_delim $1/g;
        my @output = grep{/\S+/} split($dummy_delim, $output);
        my $port;

        foreach my $line(@output) 
        {
            if ($line =~/^\s+(\S+) \((\w+)\):/) 
            {        
                $port = $1;
                $portset{$port}{type}=$2;            
            }

            #
            # Get the ports and igroups
            #
            if ($line =~/ports:(.*?)igroups:\s+(.*)/s || $line =~/ports:(.*)/s)
            {
                my ($filer_w_port, $igroup) = ($1, $2);

                my @filer_port = $filer_w_port =~ /(\S+.*)/g;
                $filer_w_port = "@filer_port";
                $portset{$port}{filername1}= $filer_port[0];
                $portset{$port}{filername2}= $filer_port[1];
                $portset{$port}{igroups} = $igroup;    
            }    
        }
    } 
    elsif ($cmd_interface =~ m/zapi/i) 
    {
        my @output = $output =~m#<portset-info>(.*?)</portset-info>#gs;
        my $port;
        foreach my $line(@output) 
        {
        #   print "LINE = $line\n"; getc;
            if ($line =~m#<portset-name>(\S+)</portset-name>#) 
            {
                $port = $1;
            } 
            if ($line =~m#<portset-type>(\S+)</portset-type>#) 
            {
                 $portset{$port}{type}=$1;        
            } 

            my @filer_port;
            if (@filer_port = $line =~m#<portset-port-name>(.*?)</portset-port-name>#g) 
            {
                my $filer1="";
                my $filer2="";            
                my $filer1_interface;
                my $filer2_interface;
                foreach my $filer_port(@filer_port) 
                {
                #   print "filer_port = $filer_port\n"; getc;
                    my ($filername, $interface) = split ":", $filer_port;
                    
                    if ($filer1 eq "" || $filername eq $filer1)
                    {
                        $filer1 = $filername;
                        $filer1_interface .= " $interface";
                    } 
                    elsif(($filer1 ne "" && $filername ne $filer1)) 
                    {
                        $filer2 = $filername;
                        $filer2_interface .= " $interface";
                    }
                }

                $portset{$port}{filername1}= "$filer1$filer1_interface";
                $portset{$port}{filername2}= "$filer2$filer2_interface";                     
            }

            my @initiator;
            if (@initiator = $line =~m#<initiator-group-name>(.*?)</initiator-group-name>#g)
            {
                $portset{$port}{igroups} = "@initiator";                    
            }
        }
    }
    # for debug purpose only
    #   foreach my $key (keys %portset) 
    #   {
    #       print "Portset Name: $key\n";
    #       print "\tType = $portset{$key}{type}\n";
    #       print "\tFiler1 = $portset{$key}{filername1}\n";
    #       print "\tFiler2 = $portset{$key}{filername2}\n";
    #       print "\tIgroups = $portset{$key}{igroups}\n";
    #   }
    #   getc;
    
    return (1, $output, %portset);
}

###############################################################################
# Subroutine Name:  destroy_all_portsets()
# Objective:        Destroy all portsets 
# Details:          See POD documentation
###############################################################################
sub destroy_all_portsets()
{
    my $self = shift;
    
    $self->sanlog(-msg => "destroy all portsets");
    $self->_clear_err();

    my ($result, $output, %hport) = $self->get_portset_info();
    my $force = "true";

    foreach my $key (keys %hport) 
    {
        $self->portset_destroy(portset_name => $key, force => $force);
    }
}

###############################################################################
# Subroutine Name:  any_portset_exist()
# Objective:        Check for portset existance 
# Details:          See POD documentation
###############################################################################
sub any_portset_exist()
{
    my $self = shift;

    $self->sanlog(-msg => "any portset exist");
    $self->_clear_err(); 

    my ($result, $output, %hport) = $self->get_portset_info();
    
    #
    # Check the result
    #
    if( $result == 0 )
    {
        $self->_set_err($output);
        return (0, $output);
    }

    my $cnt=0;
    my $ps_names="";

    foreach my $key (keys %hport) 
    {
        $ps_names = $ps_names."'$key' ";
        $cnt++;
    }
    
    if($cnt)
    { 
        return (1, $ps_names);
    }

    $self->_set_err("No portset exists");
    return (0, "No portset exists");
}

###############################################################################
# Subroutine Name:  portset_exist()
# Objective:        Check if a specific portset exists 
# Details:          See POD documentation
###############################################################################
sub portset_exist()
{
    my ($self, %args) = @_;
    my (
        $portset_name,
       );
    
    $self->sanlog(-msg => "portset exist");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_exist() "    .
                             "Invalid value for argument "         .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_exist() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Verify required parameters
    #
    if (!defined ($portset_name)) 
    {
        $self->_set_err("Portset name(portset_name) is required argument");
        return (0, "Portset name(portset_name) is required argument");
    }

    my ($result, $output, %hport) = $self->get_portset_info(portset_name => $portset_name);
    
    #
    # Check the result
    #
    if( $result == 0 )
    {
        $self->_set_err($output);
        return (0, $output);
    }

    my $key = each(%hport);
    if($key !~ /$portset_name/ && !exists($hport{$portset_name}{type}) )
    { 
        $self->_set_err("Portset::portset_exist: $portset_name not found");
        return (0, "Portset::portset_exist: $portset_name not found");
    }

    return 1;
}

###############################################################################
# Subroutine Name:  ports_exist_in_portset()
# Objective:        Check if certain ports are in a portset 
# Details:          See POD documentation
###############################################################################
sub ports_exist_in_portset()
{
    my ($self, %args) = @_;
    my (
        $portset_name,
        $ports_ref,
        $filer,
       );

    $self->sanlog(-msg => "ports exist in portset");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?filer$/) 
        {
            if (not length($filer = $args{$_})) 
            {
                my $errmsg = "Error: Portset::ports_exist_in_portset() " .
                             "Invalid value for argument "               .
                             "'-filer'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?ports$/) 
        {
            if (not length($ports_ref = $args{$_})) 
            {
                my $errmsg = "Error: Portset::ports_exist_in_portset() " .
                             "Invalid value for argument "               .
                             "'-ports'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portse_name() " .
                             "Invalid value for argument "               .
                             "'-ports'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::ports_exist_in_portset() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Verify required parameters
    #
    if (!defined ($portset_name)) 
    {
        $self->_set_err("Portset name(portset_name) is required argument");
        return (0, "Portset name(portset_name) is required argument");
    }
    if (!defined ($ports_ref)) 
    {
        $self->_set_err("Portset ports(ports) is required argument");
        return (0, "Portset ports(ports) is required argument");
    }

    #   if (!defined ($filer)) 
    #   {
    #      $self->_set_err("Filer name (filer) is required argument"); 
    #        return (0, "Filer name (filer) is required argument");
    #   }

    my ($result, $output, %hport) = $self->get_portset_info(portset_name => $portset_name);
    
    #
    # Check the result
    #
    if( $result == 0 )
    {
        $self->_set_err("Error: Portset::ports_exist_in_portset $output");
        return (0, "Error: Portset::ports_exist_in_portset $output");
    }

    my $key = each(%hport);
    
    if($key !~ /$portset_name/ && !exists($hport{$portset_name}{type}) )
    { 
        $self->_set_err("Error: Portset::ports_exist_in_portset $portset_name not found");
        return (0, "Error: Portset::ports_exist_in_portset $portset_name not found");
    }

    # port can be in the following format
    # 1. filer:port - check for specfic filer with the port
    # 2. port       - check port present in both the filers 
    # 3. if filer specified, that takes priorty and check only that filer:port
    my @portList = @$ports_ref;

    foreach my $port (@portList) 
    {
    #   print "PORT TO VERIFY = $port\n";
        my $portsetfiler = "";
        
        #
        # Handle each format
        #
        if (!defined ($filer)) 
        {
            #
            # Handle filer:port case
            #
            if($port =~ /:/) 
            {
                ($portsetfiler, $port) = split(":", $port);
#               print "filer NOT DEFINED and PORT HAS ($portsetfiler:$port)\n";
                if( $hport{$key}{filername1} =~ /$portsetfiler/ )
                {
                    $portsetfiler = $hport{$key}{filername1};    
                } 
                elsif ($hport{$key}{filername2} =~ /$portsetfiler/) 
                { 
                    $portsetfiler = $hport{$key}{filername2};    
                }
                else
                {
                    $self->_set_err("Filer($filer) not found in the " .
                                   "portset($portset_name)");
                    return (0, "Filer($filer) not found in the " .
                               "portset($portset_name)");
                }

                if( $portsetfiler !~ /$port/ )
                {
                    $self->_set_err("Port($portsetfiler:$port) is not found in " .
                                   "the portset($portset_name)");
                    return (0, "Port($portsetfiler:$port) is not found in " .
                               "the portset($portset_name)");
                }
            }
            else
            { # Handle error checking
                my $err_str="";
                my $err_flag = 0;
#               print "filer NOT DEFINED and PORT WITH ONLY port($port), so verify on both the filer\n";
                if( $hport{$key}{filername1} !~ /$port/ ) {
                    $err_str .= "port($hport{$key}{filername1}:$port) is not present in the portset($portset_name)";
                    $err_flag = 1;
                }
                if( $hport{$key}{filername2} !~ /$port/ ) {
                    $err_str .= "port($hport{$key}{filername2}:$port) is not present in the portset($portset_name)";
                    $err_flag = 1;
                }
                if($err_flag) { 
                    $self->_set_err($err_str);
                    return (0, $err_str);
                }
            }
        }
        else 
        { # Handle port case
#           print "filer($filer) 'DEFINED' and PORT($port)\n";
            if( $hport{$key}{filername1} =~ /$filer/ )
            {
                $portsetfiler = $hport{$key}{filername1};    
            } 
            elsif ($hport{$key}{filername2} =~ /$filer/) 
            { 
                $portsetfiler = $hport{$key}{filername2};    
            }
            else
            {
                $self->_set_err("Filer($filer) not found in the " .
                               "portset($portset_name)");
                return (0, "Filer($filer) not found in the " .
                           "portset($portset_name)");
            }

            #
            # Even though the filer is specified still port_ref may 
            # contain filer:port so split it
            #
            if($port =~ /:/) 
            {
                ($portsetfiler, $port) = split(":", $port) if($port =~ /:/);
            }
            
            if($portsetfiler !~ /$port/ ) 
            {
                $self->_set_err("($filer:$port) is not found in the " .
                               "portset($portset_name)");
                return (0, "($filer:$port) is not found in the " .
                           "portset($portset_name)");
            }
        }
#       print"**** iteration complete\n"; getc;
    }

    return 1;
}

###############################################################################
# Subroutine Name:  port_exist_in_portset()
# Objective:        Check if a certain port is in a portset 
# Details:          See POD documentation
###############################################################################
sub port_exist_in_portset()
{
    my ($self, %args) = @_;
    my (
        $portset_name,
        $port,
        $filer,
       );

    $self->sanlog(-msg => "port exist in portset");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?filer$/) 
        {
            if (not length($filer = $args{$_})) 
            {
                my $errmsg = "Error: Portset::port_exist_in_portset() " .
                             "Invalid value for argument "               .
                             "'-filer'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?port$/) 
        {
            if (not length($port = $args{$_})) 
            {
                my $errmsg = "Error: Portset::port_exist_in_portset() " .
                             "Invalid value for argument "               .
                             "'-port'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::port_exist_in_portset() " .
                             "Invalid value for argument "               .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::port_exist_in_portset() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }
    
    #
    # Verify required parameters
    #
    if (!defined ($portset_name)) 
    {
        $self->_set_err("Portset name(portset_name) is required argument");
        return (0, "Portset name(portset_name) is required argument");
    }
    if (!defined ($port)) 
    {     
        $self->_set_err("Portset port(PORT) is required argument");
        return (0, "Portset port(PORT) is required argument");
    }
    if (!defined ($filer) || $port =~ /:/) 
    {
        ($filer, $port) = split(":", $port) if($port =~ /:/);
    }

    my ($result, $output, %hport) = $self->get_portset_info(portset_name => $portset_name);
    
    #
    # Check result
    #
    if( $result == 0 )
    {
        $self->_set_err("Error: Portset::port_exist_in_portset $output");
        return (0, "Error: Portset::port_exist_in_portset $output");
    }

    my $key = each(%hport);
    if($key !~ /$portset_name/ && !exists($hport{$portset_name}{type}) )
    { 
        $self->_set_err("$portset_name not found");
        return (0, "$portset_name not found");
    }
    
    if( length($filer) && defined ($filer)) 
    { # Check for filer
        my $portsetfiler = "";
        if( $hport{$key}{filername1} =~ /$filer/ )
        {
            $portsetfiler = $hport{$key}{filername1};    
        } 
        elsif ($hport{$key}{filername2} =~ /$filer/) 
        { 
            $portsetfiler = $hport{$key}{filername2};    
        }
        else
        {
            $self->_set_err("Filer($filer) not found in the " .
                           "portset($portset_name)");
            return (0, "Filer($filer) not found in the " .
                       "portset($portset_name)");
        }
        
        if($portsetfiler !~ /$port/ ) 
        {
            $self->_set_err("($filer:$port) is not found in the " .
                           "portset($portset_name)");
            return (0, "($filer:$port) is not found in the " .
                       "portset($portset_name)");
        }
    } 
    else 
    { # Check for port
        if( $hport{$key}{filername1} !~ /$port/ &&  
            $hport{$key}{filername2} !~ /$port/ )
        {
            $self->_set_err("Port($port) is not found in the " .
                           "portset($portset_name)");
            return (0, "Port($port) is not found in the " .
                       "portset($portset_name)");
        }
    }

    return 1;
}

###############################################################################
# Subroutine Name:  portset_bound_to_igroup()
# Objective:        Check portset binding
# Details:          See POD documentation
###############################################################################
sub portset_bound_to_igroup()
{
    my ($self, %args) = @_;
    my (
        $portset_name,
        $igroup_name,
       );
    
    $self->sanlog(-msg => "portset bound to igroup");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?igroup_name$/) 
        {
            if (not length($igroup_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_bound_to_igroup() " .
                             "Invalid value for argument "                .
                             "'-igroup_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_bound_to_igroup() " .
                             "Invalid value for argument "                .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_bound_to_igroup() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }
    
    #
    # Verify required parameters
    #
    if (!defined ($portset_name)) 
    {
        $self->_set_err("Portset name(portset_name) is required argument");
        return (0, "Portset name(portset_name) is required argument");
    }
    if (!defined ($igroup_name)) 
    {
        $self->_set_err("Igroup name (igroup_name) is required argument");
        return (0, "Igroup name (igroup_name) is required argument");
    }

    my ($result, $output, %hport) = $self->get_portset_info(portset_name => $portset_name);
    
    #
    # Check result
    #
    if( $result == 0 )
    {
        $self->_set_err("Error: Portset::portset_bound_to_igroup $output");
        return (0, "Error: Portset::portset_bound_to_igroup $output");
    }
    
    my $key = each(%hport);
    if($key !~ /$portset_name/ && !exists($hport{$portset_name}{igroups}) )
    { 
        $self->_set_err("$portset_name not found");
        return (0, "$portset_name not found");
    }
    if($hport{$key}{igroups} =~ /$igroup_name/ )
    {
        return 1;
    }
    $self->_set_err("The portset($portset_name) is not bound to igroup($igroup_name)");
    return (0, "The portset($portset_name) is not bound to igroup($igroup_name)");
}

###############################################################################
# Subroutine Name:  portset_igroup_create()
# Objective:        Create an igroup 
# Details:          See POD documentation
###############################################################################
sub portset_igroup_create()
{
    my ($self, %args) = @_;
    my (
        $igroup_type,
        $igroup_name,
        $igroup_ostype,
        $igroup_nodes,
        $portset_name,
       );
    
    $self->sanlog(-msg => "portset igroup create");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?igroup_name$/i) 
        {
            if (not length($igroup_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_create() " .
                             "Invalid value for argument "              .
                             "'-igroup_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?igroup_type$/i) 
        {
            if (not length($igroup_type = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_create() " .
                             "Invalid value for argument "              .
                             "'-igroup_type'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?igroup_ostype$/i) 
        {
            if (not length($igroup_ostype = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_create() " .
                             "Invalid value for argument "              .
                             "'-igroup_ostype'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?igroup_nodes$/i) 
        {
            if (not length($igroup_nodes = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_create() " .
                             "Invalid value for argument "              .
                             "'-igroup_nodes'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?portset_name$/i) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_create() " .
                             "Invalid value for argument "              .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_igroup_create() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_igroup_create CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }
    my $interface = $self->{cmd_obj}->interface();

    #
    # Verify required parameters
    #
    if (!defined ($igroup_name)) 
    {
        $self->_set_err("Igroup name(igroup_name) is required argument");
        return (0, "Igroup name(igroup_name) is required argument");
    }
    if (!defined ($igroup_type)) 
    {
        $self->_set_err("Igroup type(igroup_type) is required argument");
        return (0, "Igroup type(igroup_type) is required argument");
    } 
    elsif($interface =~ m/cli/i)
    {
        if( $igroup_type =~ /fcp/i) 
        {
            $igroup_type = "-f";
        } 
        elsif ($igroup_type =~ /iscsi/i) 
        {
            $igroup_type = "-i";
        }
    }
    if (!defined ($igroup_ostype)) 
    {
        $self->_set_err("Igroup os-type(igroup_ostype) is required argument");
        return (0, "Igroup os-type(igroup_ostype) is required argument");
    }

    my $nodes = join(' ', @$igroup_nodes) if (defined($igroup_nodes));

    my @cmds_args = ("-cmd", "igroup-create");
    push (@cmds_args, "-initiator-group-name", $igroup_name);
    push (@cmds_args, "-initiator-group-type", $igroup_type);
    # for tt it will accept as -os-type, for all other relese -ostype or -os-type - 3/26/08
    push (@cmds_args, "-os-type", $igroup_ostype);
    push (@cmds_args, "-bind-portset", $portset_name) if (defined($portset_name));
    push (@cmds_args, "-check_output", "1");
    push (@cmds_args, "-check_result", "1");

    if($interface =~ m/cli/i) 
    {
        push (@cmds_args, "-node", $nodes) if (defined($igroup_nodes));
    }
    
    my $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
    if ($self->{cmd_obj}->error()) 
    {
        my $errmsg = "Failed calling execute_cmd for 'igroup-create' " .
                     "using $interface:\n".
                     "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    #
    # Check that the result and the output aren't wrong
    #
    if (!$resultsHash_ref->{result_check} || !$resultsHash_ref->{output_check}) 
    {
        my $errmsg = "Portset::portset_igroup_create " .
                     "Result or output check failed"   .
                     "Output is "                      .
                     $resultsHash_ref->{output};
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    if(defined($igroup_nodes) && $interface =~ m/zapi/i) 
    {
        my ($result, $text) = $self->portset_igroup_add(
                                    igroup_name     => $igroup_name,
                                    igroup_nodes    => \@$igroup_nodes);
        if(!$result)
        {
            $self->_set_err($text);
        }
        return (0, $text) if(!$result);
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  portset_igroup_add()
# Objective:        Add nodes to an igroup
# Details:          See POD documentation
###############################################################################
sub portset_igroup_add()            
{
    my ($self, %args) = @_;
    my (
        $igroup_nodename_ref,
        $igroup_name,
        $force,
       );
    
    $self->sanlog(-msg => "portset igroup add");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?igroup_name$/) 
        {
            if (not length($igroup_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_add() " .
                             "Invalid value for argument "              .
                             "'-igroup_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?igroup_nodes$/) 
        {
            if (not length($igroup_nodename_ref = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_add() " .
                             "Invalid value for argument "              .
                             "'-igroup_nodes'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?force$/) 
        {
            if (not length($force = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_add() " .
                             "Invalid value for argument "              .
                             "'-force'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_igroup_add() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Verify mandatory parameters
    #
    if (!defined ($igroup_name)) 
    {
        $self->_set_err("Igroup name(igroup_name) is required argument");
        return (0, "Igroup name(igroup_name) is required argument");
    }
    if (!defined ($igroup_nodename_ref)) 
    {
        $self->_set_err("Igroup nodes(igroup_nodes) is required argument");
        return (0, "Igroup nodes(igroup_nodes) is required argument");
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_igroup_add CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    my $resultsHash_ref;
    my $interface = $self->{cmd_obj}->interface();
    
    #
    # Commands must be handled differently for each interface
    #
    if($interface =~ m/zapi/i) 
    {
        foreach my $node (@$igroup_nodename_ref) 
        {
            my @cmds_args = ("-cmd", "igroup-add");
            push (@cmds_args, "-initiator-group-name", $igroup_name);
            push (@cmds_args, "-initiator", $node);
            push (@cmds_args, "-force", $force) if (defined($force));
            push (@cmds_args, "-check_output", "1");
            push (@cmds_args, "-check_result", "1");
            
            $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
            
            if ($self->{cmd_obj}->error()) 
            {
                my $errmsg = "Failed calling execute_cmd for 'igroup-add' " .
                             "using $interface:\n".
                             "Reason: " . 
                             join("\n", $self->{cmd_obj}->errmsg());
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }

            #
            # Check that the result and output aren't wrong
            #
            if (!$resultsHash_ref->{result_check} || 
                !$resultsHash_ref->{output_check}) 
            {
                my $errmsg = "Portset::portset_igroup_add "    .
                             "Result or output check failed"   .
                             "Output is "                      .
                             $resultsHash_ref->{output};
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }
        }
    } 
    elsif($interface =~ m/cli/i) 
    {
        my $nodes = join(' ', @$igroup_nodename_ref) if (defined($igroup_nodename_ref));
        my @cmds_args = ("-cmd", "igroup-add");
        push (@cmds_args, "-initiator-group-name", $igroup_name);
        push (@cmds_args, "-initiator", $nodes);

        if( defined($force) && $force =~ /(yes|true|1)/i ) 
        {
            push (@cmds_args, "-force", "");
        }

        push (@cmds_args, "-check_output", "1");
        push (@cmds_args, "-check_result", "1");

        $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
        
        if ($self->{cmd_obj}->error()) 
        {
            my $errmsg = "Failed calling execute_cmd for 'igroup-add' " .
                         "using $interface:\n".
                         "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }

        #
        # Check that the result and output aren't wrong
        # 
        if (!$resultsHash_ref->{result_check} || 
            !$resultsHash_ref->{output_check}) 
        {
            my $errmsg = "Portset::portset_igroup_add "    .
                         "Result or output check failed"   .
                         "Output is "                      .
                         $resultsHash_ref->{output};
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  portset_igroup_destroy()
# Objective:        Destroy an igroup
# Details:          See POD documentation
###############################################################################
sub portset_igroup_destroy()
{
    my ($self, %args) = @_;
    my (
        $force,
        $igroup_name,
       );
    
    $self->sanlog(-msg => "portset igroup destroy");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?igroup_name$/) 
        {
            if (not length($igroup_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_destroy() " .
                             "Invalid value for argument "               .
                             "'-igroup_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?force$/) 
        {
            if (not length($force = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_destroy() " .
                             "Invalid value for argument "               .
                             "'-force'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_igroup_destroy() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }

    #
    # Verify mandatory parameters
    #
    if (!defined ($igroup_name)) 
    {
        $self->_set_err("Igroup name(igroup_name) is required argument");
        return (0, "Igroup name(igroup_name) is required argument");
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_igroup_destroy CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    my $interface = $self->{cmd_obj}->interface();
    my @cmds_args = ("-cmd", "igroup-destroy");
    push (@cmds_args, "-initiator-group-name", $igroup_name);
    
    #
    # Define force
    #
    if( defined($force) && $force =~ /(yes|true|1)/i ) 
    {
        push (@cmds_args, "-force", "") if($interface =~ m/cli/i);
        push (@cmds_args, "-force", $force) if($interface =~ m/zapi/i);
    } 
    elsif(defined($force))
    {  # useful for negatine test cases
        push (@cmds_args, "-force", $force);
    }

    push (@cmds_args, "-check_output", "1");
    push (@cmds_args, "-check_result", "1");
    
    my $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
    
    if ($self->{cmd_obj}->error()) 
    {
        my $errmsg = "Failed calling execute_cmd for 'igroup-destroy' " .
                     "using $interface:\n".
                     "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    #
    # Check that the result and output aren't wrong
    #
    if (!$resultsHash_ref->{result_check} || !$resultsHash_ref->{output_check}) 
    {
        my $errmsg = "Portset::portset_igroup_destroy " .
                     "Result or output check failed"    .
                     "Output is "                       .
                     $resultsHash_ref->{output};
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  portset_igroup_bind()
# Objective:        Bind an igroup 
# Details:          See POD documentation
###############################################################################
sub portset_igroup_bind()
{
    my ($self, %args) = @_;
    my (
        $portset_name,
        $igroup_name,
       );
    
    $self->sanlog(-msg => "portset igroup bind");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?igroup_name$/) 
        {
            if (not length($igroup_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_bind() " .
                             "Invalid value for argument "            .
                             "'-igroup_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_bind() " .
                             "Invalid value for argument "            .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_igroup_bind() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }
    
    #
    # Verify mandatory parameters
    #
    if (!defined ($igroup_name)) 
    {
        $self->_set_err("Igroup name(igroup_name) is required argument");
        return (0, "Igroup name(igroup_name) is required argument");
    }
    if (!defined ($portset_name)) 
    {
        $self->_set_err("Portset name(portset_name) is required argument");
        return (0, "Portset name(portset_name) is required argument");
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_igroup_bind CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    my $interface = $self->{cmd_obj}->interface();
    my @cmds_args = ("-cmd", "igroup-bind-portset");
    push (@cmds_args, "-initiator-group-name", $igroup_name);
    push (@cmds_args, "-portset-name", $portset_name);
    push (@cmds_args, "-check_output", "1");
    push (@cmds_args, "-check_result", "1");

    my $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
    
    if ($self->{cmd_obj}->error()) 
    {
        my $errmsg = "Failed calling execute_cmd for 'igroup-bind-portset' " .
                     "using $interface:\n".
                     "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    #
    # Check that the result and output aren't wrong
    #
    if (!$resultsHash_ref->{result_check} || !$resultsHash_ref->{output_check}) 
    {
        my $errmsg = "Portset::portset_igroup_bind "    .
                     "Result or output check failed"    .
                     "Output is "                       .
                     $resultsHash_ref->{output};
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  portset_igroup_unbind()
# Objective:        Unbind an igroup 
# Details:          See POD documentation
###############################################################################
sub portset_igroup_unbind()
{
    my ($self, %args) = @_;
    my (
        $igroup_name,
       );
    
    $self->sanlog(-msg => "portset igroup unbind");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?igroup_name$/) 
        {
            if (not length($igroup_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_unbind() " .
                             "Invalid value for argument "              .
                             "'-igroup_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_igroup_unbind() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }
    
    #
    # Verify mandatory parameters
    #
    if (!defined ($igroup_name)) 
    {
        $self->_set_err("Igroup name(igroup_name) is required argument");
        return (0, "Igroup name(igroup_name) is required argument");
    }

    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_igroup_unbind CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    my $interface = $self->{cmd_obj}->interface();
    my @cmds_args = ("-cmd", "igroup-unbind-portset");
    push (@cmds_args, "-initiator-group-name", $igroup_name);
    push (@cmds_args, "-check_output", "1");
    push (@cmds_args, "-check_result", "1");

    my $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
    
    if ($self->{cmd_obj}->error()) 
    {
        my $errmsg = "Failed calling execute_cmd for 'igroup-unbind-portset'" .
                     " using $interface:\n".
                     "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    #
    # Check that the result and output aren't wrong
    #
    if (!$resultsHash_ref->{result_check} || !$resultsHash_ref->{output_check}) 
    {
        my $errmsg = "Portset::portset_igroup_unbind "  .
                     "Result or output check failed"    .
                     "Output is "                       .
                     $resultsHash_ref->{output};
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  portset_igroup_show()
# Objective:        Display information on an igroup
# Details:          See POD documentation
###############################################################################
sub portset_igroup_show()
{
    #
    # TBD: Currently the ZAPI returns a hash for output... this may
    # eventually need to be translated into text
    #
    my ($self, %args) = @_;
    my (
        $verbose,
        $igroup_name,
       );
    
    $self->sanlog(-msg => "portset igroup show");
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?igroup_name$/) 
        {
            if (not length($igroup_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_show() " .
                             "Invalid value for argument "              .
                             "'-igroup_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?verbose$/) 
        {
            if (not length($verbose = $args{$_})) 
            {
                my $errmsg = "Error: Portset::portset_igroup_show() " .
                             "Invalid value for argument "              .
                             "'-verbose'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::portset_igroup_show() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }
    
    #
    # Make sure we have a valid CmdsDriver object in $self->{cmd_obj}
    #
    unless ($self->cmds_driver_exist())
    {
        my $errmsg = "Portset::portset_igroup_show CmdsDriver object does not exist";
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    my $interface = $self->{cmd_obj}->interface();
    my @cmds_args = ("-cmd", "igroup-list-info");
    push (@cmds_args, "-initiator-group-name", $igroup_name) if( defined($igroup_name));
    push (@cmds_args, "-verbose", "") if(defined($verbose) && 
        $interface =~ m/cli/i);
    push (@cmds_args, "-check_output", "1");
    push (@cmds_args, "-check_result", "1");

    my $resultsHash_ref = $self->{cmd_obj}->execute_cmd(@cmds_args);
    
    if ($self->{cmd_obj}->error()) 
    {
        my $errmsg = "Failed calling execute_cmd for 'igroup-list-info' " .
                     "using $interface:\n".
                     "Reason: " . join("\n", $self->{cmd_obj}->errmsg());
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    #
    # Check that the result and output aren't wrong
    #
    if (!$resultsHash_ref->{result_check} || !$resultsHash_ref->{output_check}) 
    {
        my $errmsg = "Portset::portset_igroup_show "    .
                     "Result or output check failed"    .
                     "Output is "                       .
                     $resultsHash_ref->{output};
        $self->_set_err($errmsg);
        return (0, $errmsg);
    }

    return (1, $resultsHash_ref->{output});
}

###############################################################################
# Subroutine Name:  igroup_bound_to_portset()
# Objective:        Check igroup binding 
# Details:          See POD documentation
###############################################################################
sub igroup_bound_to_portset()
{
    my ($self, %args) = @_;
    my (
        $portset_name,
        $igroup_name,
       );

    $self->sanlog(-msg => "igroup bound to portset");
    $self->_clear_err();

    # parse arguments
    foreach (keys %args) 
    {
        if (/^-?igroup_name$/) 
        {
            if (not length($igroup_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::igroup_bound_to_portset() " .
                             "Invalid value for argument "              .
                             "'-igroup_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        elsif (/^-?portset_name$/) 
        {
            if (not length($portset_name = $args{$_})) 
            {
                my $errmsg = "Error: Portset::igroup_bound_to_portset() " .
                             "Invalid value for argument "              .
                             "'-portset_name'";
                $self->_set_err($errmsg);
                return (0, $errmsg);
            }        
            next;
        }
        else 
        {
            my $errmsg = "Error: Portset::igorup_bound_to_portset() " .
                         "Invalid parameter specified \'$_\'";
            $self->_set_err($errmsg);
            return (0, $errmsg);
        }
    }
    
    #
    # Verify mandatory parameters
    #
    if (!defined ($igroup_name)) 
    {
        $self->_set_err("Igroup name (igroup_name) is required argument");
        return (0, "Igroup name (igroup_name) is required argument");
    }
    if (!defined ($portset_name)) 
    {
        $self->_set_err("Portset name(portset_name) is required argument");
        return (0, "Portset name(portset_name) is required argument");
    }

    my $exp_str = "";

    #
    # Handle interfaces differently
    #
    my $interface = $self->{cmd_obj}->interface();
    if( $interface =~ m/cli/i )
    {
        $exp_str = "Pset:  $portset_name";
    } 
    else 
    {
        $exp_str = "<initiator-group-portset-name>$portset_name" .
                   "<\/initiator-group-portset-name>";
    }

    my ($result, $text) = $self->portset_igroup_show(igroup_name => $igroup_name);
    if(!$result)
    {
        $self->_set_err($text);
    }
    return (0, $text) if (!$result); 

    if ( ($text =~ /$igroup_name/ && $text =~ /$portset_name/ && 
          $text =~ /$exp_str/) ) 
    {
        return 1;
    }
    $self->_set_err("The igroup($igroup_name) is not bound to " .
                   "portset($portset_name)");
    return (0, "The igroup($igroup_name) is not bound to " .
               "portset($portset_name)");
}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Portset.pm

=head1 Author

Muthukumaran Govindaraj (muthug@netapp.com)

=head1 Description

This utility to provide varity of function to test the 
portset functionality. This library uses the CmdsDriver.pm 
to execute the portset commands for both CLI and ZAPI.

=head1 See Also

Package inherits from the following modules. See these modules for further 
interface detail.

B<San::Sanlog>

B<San::SanUtil>

B<Error>

B<CmdsDriver>

B<Exporter>

=head1 Attributes

=head2 B<interface>

string = the interface to use (cli or zapi)

=head2 B<name>

string = filer name

=head2 B<connection>

object = filer connection

=head2 B<cmd_obj>

object = commands driver object

=head1 Public Methods

=head2 B<new_portset()>

Creates a new portset object.  Give default values to all supported attributes.

=over 4

=item Inputs

=item * I<-interface> (default is cli)

string = cli or zapi

=item * I<-filer> (required if filer_conn if not passed in)

string = filer name

=item * I<-filer_conn> (optional)

object = filer connection object

=item * I<-cmd_obj> (optional)

object = commands driver object

=item Outputs

Returns a portset object

=item Examples

    #
    # Create a portset object
    #
    $portset_obj = new_portset Portset(
                       -filer          => "rtpqa-fas2020-3");
    if ($portset_obj->error())
    {
       my $errmsg = join("\n", $portset_obj->errmsg());
       $Test->nlog($errmsg);
       return $TCD::FATAL;
    }


=back

=head2 B<any_portset_exist()>

Find out if any portsets exist

=over 4

=item Outputs

Returns 1 if a portset exists, 0 if no portsets exist and any output

=item Examples

    my ($result, $output) = $portset_obj->any_portset_exist();    

=back

=head2 B<destroy_all_portsets()>

Destroys all portsets

=over 4

=item Examples

    $portset_obj->destroy_all_portsets();    

=back

=head2 B<get_filer_ontap_version()>

Gets the filer version

=over 4

=item Outputs

Returns the filer version

=item Examples

    $version = $portset_obj->get_filer_ontap_version();    

=back

=head2 B<get_portset_info(%args)>

Get information on a specific portset

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item Outputs

Returns 1 if command was successful, 0 if command unsuccessful, any output, a hash
of what the command returned

=item Examples

    my ($result, $output, %info) = $portset_obj->get_portset_info(
                                            -portset_name   => $portset_name);    

=back

=head2 B<igroup_bound_to_portset(%args)>

Determine whether or not an igroup is bound to a portset

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item * I<-igroup_name>

string = name of the igroup

=item Outputs

Returns 1 if command successful, 0 if command unsuccessful, any output

=item Examples

    my ($result, $output) = $portset_obj->igroup_bound_to_portset(
                                            -portset_name   => $portset_name,
                                            -igroup_name    => $igroup_name);    

=back

=head2 B<port_exist_in_portset(%args)>

Determine whether or not a port exists in a portset

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item * I<-port>

string = port value to check

=item * I<-filer>

string = the filer you are running on

=item Outputs

Returns 1 if port exists, 0 if port doesn't exist, any output

=item Examples

    my ($result, $output) = $portset_obj->port_exist_in_portset(
                                            -portset_name   => $portset_name,
                                            -port           => $port,
                                            -filer          => $filer_name);    

=back

=head2 B<ports_exist_in_portset(%args)>

Determine whether or not a set of ports exist in a portset

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item * I<-ports_ref>

string = reference to the list of ports to check

=item * I<-filer>

string = the filer you are running on

=item Outputs

Returns 1 if ports exist, 0 if ports don't exist, any output

=item Examples

    my ($result, $output) = $portset_obj->ports_exist_in_portset(
                                            -portset_name   => $portset_name,
                                            -ports_ref      => \@ports,
                                            -filer          => $filer_name);    

=back

=head2 B<portset_add(%args)>

Add ports to a portset

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item * I<-ports_ref>

string = a reference to the list of ports

=item Outputs

Returns 1 if command was successful, 0 if command unsuccessful, any output, a hash
of what the command returned

=item Examples

    my ($result, $output, %info) = $portset_obj->portset_add(
                                            -portset_name   => $portset_name,
                                            -ports_ref      => \@ports);    

=back

=head2 B<portset_bound_to_igroup(%args)>

Determine whether or not a portset is bound to an igroup

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item * I<-igroup_name>

string = name of the igroup

=item Outputs

Returns 1 if command successful, 0 if command unsuccessful, any output

=item Examples

    my ($result, $output) = $portset_obj->portset_bound_to_igroup(
                                            -portset_name   => $portset_name,
                                            -igroup_name    => $igroup_name);    

=back

=head2 B<portset_create(%args)>

Create a portset

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item * I<-portset_type>

string = iscsi or fcp

=item * I<-ports_ref>

string = a reference to the list of ports

=item Outputs

Returns 1 if command was successful, 0 if command unsuccessful, any output, a hash
of what the command returned

=item Examples

    my ($result, $output, %info) = $portset_obj->portset_create(
                                            -portset_name   => $portset_name,
                                            -ports_ref      => \@ports,
                                            -portset_type   => "fcp");    

=back

=head2 B<portset_destroy(%args)>

Destroy a portset

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item * I<-force>

string = yes | no, 1 | 0, true | false, etc.

=item Outputs

Returns 1 if command was successful, 0 if command unsuccessful, any output, a hash
of what the command returned

=item Examples

    my ($result, $output, %info) = $portset_obj->portset_create(
                                            -portset_name   => $portset_name,
                                            -force          => "yes");    

=back

=head2 B<portset_exist(%args)>

Determine whether or not a portset exists

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item Outputs

Returns 1 if portset exists, 0 if portset doesn't exist, any output

=item Examples

    my ($result, $output) = $portset_obj->portset_bound_to_igroup(
                                            -portset_name   => $portset_name);    

=back

=head2 B<portset_remove(%args)>

Removes ports from portset

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item * I<-ports_ref>

string = a reference to the port list

=item Outputs

Returns 1 if command was successful, 0 if command unsuccessful, any output, a hash
of what the command returned

=item Examples

    my ($result, $output, %info) = $portset_obj->portset_create(
                                            -portset_name   => $portset_name,
                                            -ports_ref      => \@ports);    

=back

=head2 B<portset_show(%args)>

List information on a specific portset

=over 4

=item Inputs

=item * I<-portset_name>

string = name of the portset

=item Outputs

Returns 1 if command was successful, 0 if command unsuccessful, any output, a hash
of what the command returned

=item Examples

    my ($result, $output, %info) = $portset_obj->portset_show(
                                            -portset_name   => $portset_name);    

=back

=head2 B<set_cli_partner_mode()>

Sets the partner mode in cmdsDriver

=over 4

=item Examples

    $portset_obj->set_cli_partner_mode();    

=back

=cut

1;
