# Copyright (c) 2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Filer ISCSI targets
## @author thomasm@netapp.com
## @status private
## @pod here

# Package Name
##############
package ISCSITarget;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use San::SanUtil;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: display()
# Objective:       Display ISCSITarget information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: get_initiator_iqns
# Objective:       Return initiator iqns
# Details:         See POD documentation at the end of this file
################################################################################
sub get_initiator_iqns {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{inits}) {
        return (wantarray()) ? @{$self->{inits}} : $self->{inits};
    } else {
        $self->_set_err("Error: ISCSITarget::get_initiator_wwpns() ISCSITarget object is missing initiator port name(s)");
        return  (wantarray()) ? () : [] ;
    }
}

################################################################################
# Subroutine Name: get_initcnt
# Objective:       Return initiators cnt logged into target
# Details:         See POD documentation at the end of this file
################################################################################
sub get_initcnt {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{initcnt}){
        return $self->{initcnt};
    } else {
        return  0;
    }
}

################################################################################
# Subroutine Name: is_mgmt_port
# Objective:       Returns 1 if this port is the management port, 0 if not
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mgmt_port {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{is_mgmt_port};
}

################################################################################
# Subroutine Name: get_name
# Objective:       Return ISCSITarget object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: ISCSITarget::get_name() ISCSITarget object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_nodename
# Objective:       Return ISCSITarget nodename
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nodename {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{nodename}) {
        return $self->{nodename};
    } else {
        $self->_set_err("Error: ISCSITarget::get_nodename() ISCSITarget object is missing nodename");
        return "";
    }
}

################################################################################
# Subroutine Name: get_mac
# Objective:       Return mac address of the interface
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mac {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{mac}) {
        return $self->{mac};
    } else {
        $self->_set_err("Error: ISCSITarget::get_mac() ISCSITarget object is missing mac address");
        return "";
    }
}

################################################################################
# Subroutine Name: get_mode
# Objective:       Return mode the iscsi target
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mode {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{mode}) {
        return $self->{mode};
    } else {
        $self->_set_err("Error: ISCSITarget::get_mode() ISCSITarget object is missing mode (ip ver)");
        return "";
    }
}

################################################################################
# Subroutine Name: get_type
# Objective:       Return ISCSITarget adapter type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_type {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{type}) {
        return $self->{type};
    } else {
        $self->_set_err("Error: ISCSITarget::get_type() ISCSITarget object is missing adapter type");
        return "";
    }
}

################################################################################
# Subroutine Name: get_ip
# Objective:       Return ISCSITarget ip address (default ipv4)
# Details:         See POD documentation at the end of this file
################################################################################
sub get_ip {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{ip} and $self->{ip} ne "not_set") {
        return $self->{ip};
    } else {
        $self->_set_err("Error: ISCSITarget::get_ip() ISCSITarget object is missing ip address");
        return "";
    }
}

################################################################################
# Subroutine Name: get_ipv4
# Objective:       Return ISCSITarget ipv4 address
# Details:         See POD documentation at the end of this file
################################################################################
sub get_ipv4 {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{ipv4} and $self->{ipv4} ne "not_set") {
        return $self->{ipv4};
    } else {
        $self->_set_err("Error: ISCSITarget::get_ipv4() ISCSITarget object is missing ipv4 address");
        return "";
    }
}

################################################################################
# Subroutine Name: get_ipv4_netmask
# Objective:       Return ISCSITarget ipv4 netmask
# Details:         See POD documentation at the end of this file
################################################################################
sub get_ipv4_netmask {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{ipv4_netmask}) {
        return $self->{ipv4_netmask};
    } else {
        $self->_set_err("ISCSITarget object is missing ipv4 netmask");
        return "";
    }
}

################################################################################
# Subroutine Name: get_ipv6
# Objective:       Return ISCSITarget ip address (default ipv4)
# Details:         See POD documentation at the end of this file
################################################################################
sub get_ipv6 {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{ipv6} and $self->{ipv6} ne "not_set") {
        return $self->{ipv6};
    } else {
        $self->_set_err("Error: ISCSITarget::get_ipv6() ISCSITarget object is missing ipv6 address");
        return "";
    }
}

################################################################################
# Subroutine Name: get_ipv6_prefix_len
# Objective:       Return ISCSITarget ipv6 prefix len
# Details:         See POD documentation at the end of this file
################################################################################
sub get_ipv6_prefix_len {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{ipv6_prefix_len}) {
        return $self->{ipv6_prefix_len};
    } else {
        $self->_set_err("ISCSITarget object is missing ipv6 prefix_len");
        return "";
    }
}

################################################################################
# Subroutine Name: get_subnet
# Objective:       Return ISCSITarget working subnet
# Details:         See POD documentation at the end of this file
################################################################################
sub get_subnet {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{subnet}) {
        return $self->{subnet};
    } else {
        $self->_set_err("ISCSITarget object is missing ipv6 prefix_len");
        return "";
    }
}

################################################################################
# Subroutine Name: get_ipv4_subnet
# Objective:       Return ISCSITarget ipv4 subnet
# Details:         See POD documentation at the end of this file
################################################################################
sub get_ipv4_subnet {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{ipv4_subnet}) {
        return $self->{ipv4_subnet};
    } else {
        $self->_set_err("ISCSITarget object is missing ipv4 subnet");
        return "";
    }
}

################################################################################
# Subroutine Name: get_ipv6_subnet
# Objective:       Return ISCSITarget ipv6 subnet
# Details:         See POD documentation at the end of this file
################################################################################
sub get_ipv6_subnet {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{ipv6_subnet}) {
        return $self->{ipv6_subnet};
    } else {
        $self->_set_err("ISCSITarget object is missing ipv6 subnet");
        return "";
    }
}

################################################################################
# Subroutine Name: get_tpgt
# Objective:       Return ISCSITarget target portal group tag
# Details:         See POD documentation at the end of this file
################################################################################
sub get_tpgt {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{tpgt}) {
        return $self->{tpgt};
    } else {
        $self->_set_err("Error: ISCSITarget::get_tpgt() ISCSITarget object is missing target port group tag (tpgt)");
        return "";
    }
}

################################################################################
# Subroutine Name: get_tpgn
# Objective:       Return ISCSITarget target portal group name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_tpgn {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{tpgn}) {
        return $self->{tpgn};
    } else {
        $self->_set_err("Error: ISCSITarget::get_tpgn() ISCSITarget object is missing target port group name (tpgn)");
        return "";
    }
}

################################################################################
# Subroutine Name: get_state
# Objective:       Return ISCSITarget state (enabled, disabled, unknown)
# Details:         See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{state}) {
        return $self->{state};
    } else {
        $self->_set_err("Error: ISCSITarget::get_state() ISCSITarget object is missing state");
        return "";
    }
}

################################################################################
# Subroutine Name: is_vif
# Objective:       Return if the interface is a represented as VIF/IGRP
# Details:         See POD documentation at the end of this file
################################################################################
sub is_vif {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{is_vif}) {
        return $self->{is_vif};
    } else {
        $self->_set_err("Error: ISCSITarget::is_vif() ISCSITarget object is missing is_vif");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vif_ints
# Objective:       Return the VIF/IGRP interfaces that make up that VIF
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vif_ints {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{vif_ints}) {
        return $self->{vif_ints};
    } else {
        $self->_set_err("Error: ISCSITarget::is_vif() ISCSITarget object is missing vif_ints");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vif_type
# Objective:       Return the VIF/IGRP type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vif_type {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{vif_type}) {
        return $self->{vif_type};
    } else {
        $self->_set_err("Error: ISCSITarget::is_vif() ISCSITarget object is missing vif_type");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vlan_tag
# Objective:       Return the tag of the VLAN is applicable
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vlan_tag {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{vlan_tag}) {
        return $self->{vlan_tag};
    } else {
        $self->_set_err("Error: ISCSITarget::is_vif() ISCSITarget object is missing vlan_tag");
        return "";
    }
}

################################################################################
# Subroutine Name: set_ip_mode
# Objective:       Set the working ip for this target to be the ipv4|ipv6 ip
# Details:         See POD documentation at the end of this file
################################################################################
sub set_ip_mode {
    my ($self, $mode) = @_;

    $self->_clear_err();

    if ($mode =~ /ipv4/i) {
        $self->{ip} = $self->{ipv4};
        $self->{subnet} = $self->{ipv4_subnet};
        $self->{mode} = "ipv4";
    } elsif ($mode =~ /ipv6/i) {
        $self->{ip} = $self->{ipv6};
        $self->{subnet} = $self->{ipv6_subnet};
        $self->{mode} = "ipv6";
    } else {
        $self->_set_err("Invalid IP mode: '$mode', must be ipv4 or ipv6");
        return $SanUtil::FALSE;
    }

    # if the working ip is empty then flag this as an error
    if ($self->{ip} eq "") {
        $self->_set_err("Invalid IP mode: '$mode', this target does not have an $mode ip address");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Subroutine Name: iscsi_interface_enable()
# Objective:       Enable iSCSI on this target interface
# Details:         See POD documentation at the end of this file
################################################################################
sub iscsi_interface_enable {
    my ($self) = @_;

    $self->_clear_err();

    my $interface_name = $self->get_name();
    my $resultsHash_ref;

    # invoke the iscsi_interface_disable zapi on the filer object
    my $filer_obj = $self->{parent};
    my $cmd_obj = $filer_obj->cmd_obj();

    $filer_obj->sanlog(-msg => "ISCSITarget::iscsi_interface_enable() Enabling " .
                               "iSCSI on interface '$interface_name'",
                       -msg_level => 1);

    if ($filer_obj->interface() eq "zapi")
    {
        # issue the interface enable zapi
        $resultsHash_ref =
            $cmd_obj->execute_cmd(-cmd                       => "iscsi-interface-enable",
                                  -cmd_filer                 => $filer_obj->get_name(),
                                  -check_result              => 1,
                                  '-interface-name'          => $interface_name);
    } else {
        $self->_set_err("Unsupported interface. Use interface 'zapi'");
        return $ClientUtil::FALSE;
    }

    if ($cmd_obj->error())
    {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to enable iSCSI on interface $interface_name: $errmsg");
        return $ClientUtil::FALSE;
    }

    $filer_obj->log_cmd_results($resultsHash_ref);

    # Since we checked the result, lets see if our check was OK
    if ($resultsHash_ref->{result_check})
    {
        return $ClientUtil::TRUE;
    }
    else
    {
        $self->_set_err("Results check failed");
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: iscsi_interface_disable()
# Objective:       Disable iSCSI on this target interface
# Details:         See POD documentation at the end of this file
################################################################################
sub iscsi_interface_disable {
    my ($self) = @_;

    $self->_clear_err();

    my $interface_name = $self->get_name();
    my $resultsHash_ref;

    # invoke the iscsi_interface_disable zapi on the filer object
    my $filer_obj = $self->{parent};
    my $cmd_obj = $filer_obj->cmd_obj();

    $filer_obj->sanlog(-msg => "ISCSITarget::iscsi_interface_disable() Disabling " .
                               "iSCSI on interface '$interface_name'",
                       -msg_level => 1);

    if ($filer_obj->interface() eq "zapi")
    {
        # issue the interface disable zapi
        $resultsHash_ref =
            $cmd_obj->execute_cmd(-cmd                       => "iscsi-interface-disable",
                                  -cmd_filer                 => $filer_obj->get_name(),
                                  -check_result              => 1,
                                  '-interface-name'          => $interface_name);
    } else {
        $self->_set_err("Unsupported interface. Use interface 'zapi'");
        return $ClientUtil::FALSE;
    }

    if ($cmd_obj->error())
    {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to disable iSCSI on interface $interface_name: $errmsg");
        return $ClientUtil::FALSE;
    }

    $filer_obj->log_cmd_results($resultsHash_ref);

    # Since we checked the result, lets see if our check was OK
    if ($resultsHash_ref->{result_check})
    {
        return $ClientUtil::TRUE;
    }
    else
    {
        $self->_set_err("Results check failed");
        return $ClientUtil::FALSE;
    }
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new ISCSITarget object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($name, $self, $type, $state, $nodename, $ipv4, $ipv6, $ip, $is_mgmt_port);
    my ($parent, @inits, $mac, $tpgt, $tpgn, $ipv4_netmask, $ipv6_prefix_len);
    my ($is_vif, $vif_type, $vif_ints, $vlan_tag, $vfiler);

    $self = {};

    bless($self, $class);

    foreach (keys %args) {
        if (/-?nodename$/) {
            if (not length($nodename = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?state$/) {
            if (not length($state = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?vif_type$/) {
            $vif_type = $args{$_};
            next;
        } elsif (/-?type$/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?ipv4$/) {
            if (not length($ipv4 = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?ipv4_netmask$/) {
            if (not length($ipv4_netmask = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?ipv6$/) {
            if (not length($ipv6 = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?ipv6_prefix_len$/) {
            if (not length($ipv6_prefix_len = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?tpgt$/) {
            if (not length($tpgt = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?tpgn$/) {
            if (not length($tpgn = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        } elsif (/-?mac$/) {
            if (not length($mac = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
        } elsif (/-?is_mgmt_port$/) {
            if (not length($is_mgmt_port = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
        } elsif (/-?is_vif$/) {
            if (not length($is_vif = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
        } elsif (/-?vif_ints$/) {
            $vif_ints = $args{$_};
            next;
        } elsif (/-?vlan_tag$/) {
            $vlan_tag = $args{$_};
            next;
         } elsif (/-?vfiler/) {
             $vfiler = $args{$_};
             next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: ISCSITarget::_new() Invalid value for argument '$_'");
                return $self;
            }
            next;
        }
    }

    # check for filer parent object.
    unless (defined $parent and length $parent) {
        $self->_set_err("Error: ISCSITarget::_new()",
                        "Required parameter filer object reference is missing '-parent'");
        return $self;
    }

    # initiialize members
    $self->{name}            = $name;
    $self->{nodename}        = $nodename;
    $self->{state}           = $state;
    $self->{type}            = $type;
    $self->{parent}          = $parent;
    $self->{is_mgmt_port}    = $is_mgmt_port;

    # initialize ipv4 attributes
    $self->{ipv4}            = $ipv4;
    $self->{ipv4_netmask}    = $ipv4_netmask;
    if ($ipv4 ne "not_set") {
        $self->{ipv4_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv4,
                                                       -ip_version => "ipv4",
                                                       -ipv4_netmask => $ipv4_netmask);
    } else {
        $self->{ipv4_subnet} = "";
    }

    # initialize ipv6 attributes
    $self->{ipv6}            = $ipv6;
    $self->{ipv6_prefix_len} = $ipv6_prefix_len;
    if ($ipv6 ne "not_set") {
        $self->{ipv6_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv6,
                                                       -ip_version => "ipv6",
                                                       -ipv6_prefix_len => $ipv6_prefix_len);
    } else {
        $self->{ipv6_subnet} = "";
    }

    $ip = "not_set";
    if (($ipv4 ne "") and ($ipv4 ne "not_set")) {
        $ip = $ipv4;
        $self->{mode} = "ipv4";
    } elsif (($ipv6 ne "") and ($ipv6 ne "not_set")){
        $ip = $ipv6;
        $self->{mode} = "ipv6";
    }

    $self->{ip}              = $ip;
    $self->{subnet}          = $self->{ipv4_subnet};
    $self->{tpgt}            = $tpgt;
    $self->{tpgn}            = $tpgn;
    $self->{mac}             = $mac;
    $self->{filer_name}      = $parent->get_name();

    # set VIF/IGRP information
    $self->{is_vif}          = $is_vif;
    if ($is_vif == 1) {
        $self->{vif_ints}    = $vif_ints;
        $self->{vif_type}    = $vif_type;
    } else {
        $self->{vif_ints}    = undef;
        $self->{vif_type}    = undef;
    }
    $self->{vlan_tag}        = $vlan_tag;

    @inits = $self->_get_initiators();
    $self->{inits} = \@inits;
    $self->{initcnt} = scalar(@inits);

     # set vfiler information
     if ($vfiler) {
         $self->{vfiler} = $vfiler;
     } else {
         $self->{vfiler} = "";
     }

    return $self;

}

################################################################################
# Subroutine Name: _get_initiators()
# Objective:       load initiators hooked to target
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_initiators {
    my ($self)= @_;

    my ($cmd, $output, $status, $tpgt, $init, @inits, @output);
    my ($version);

    @inits = ();

    # get target portal group tag to match to for this target
    $tpgt = $self->get_tpgt();
    return (wantarray()) ? () : [] if $self->error();

    # if tag is not set, return empty list
    if ($tpgt =~ /not_set/i) {
        return (wantarray()) ? () : [];
    }

    # get iscsi initiators logged in
    $version = $self->{parent}->get_version_major();
    if ($self->{parent}->error()) {
        my $errmsg = join "\n", $self->{parent}->errmsg();
        $self->_set_err("could not get ONTAP major revision: $errmsg");
        return (wantarray()) ? () : [];
    }

    if ($version =~ /7\.0/i) {
        $cmd = "iscsi show initiator";
    } else {
        $cmd = "iscsi initiator show";
    }
    if ($self->{parent}->{interface} eq "cli") {
        @output = $self->{parent}->execute($cmd, -hide => 1);
        if ($self->{parent}->error()) {
            $self->_set_err("command ($cmd) failed over CLI");
            return (wantarray()) ? () : [];
        }
    } else {
        ($output, $status, @output) = &SanUtil::system_zapi($self->{parent}->get_name(), $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("command ($cmd) failed over system-cli ZAPI");
            return (wantarray()) ? () : [];
        }
    }

    foreach my $line (@output) {
        # skip lines w/o IQNs
        next if ($line !~ /iqn/i);

        $line =~ s/^\s+//;
        my $tag = (split(/\s+/, $line))[1];

        # if the tag is the target portal group
        # tag for the target, push into the list
        # of initiators.
        if ($tag == $tpgt) {
            if ($line =~ /.*(iqn\.\S+)/) {
                $init = $1;
                push(@inits, $init);
            }
        }
    }

    return (wantarray()) ? @inits : \@inits;

}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=cut
