# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Filer targets
## @author knopp@netapp.com
## @status private
## @pod here

# Package Name
##############
package Target;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanLog;

use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: display()
# Objective:       Display Target information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self)=@_;
    my ($cnt, $name, $header, $summary, @output, @buffer);

    $self->_clear_err();

    $name = $self->{name};

    $cnt = 0;

    $header = 
        sprintf("Target Detail:\n".
                "%-10.10s %-4.4s %-8.8s %-16.16s %-16.16s %-8.8s %-10.10s %-8.8s\n".
                "%10.10s %4.4s %8.8s %16.16s %16.16s %-8.8s %10.10s %-8.8s",
                "FILER","NAME", "TYPE", "NODENAME","PORTNAME","STATUS","VERSION","INIT CNT",
                "="x10,"="x4,"="x8,"="x16,"="x16,"="x8,"="x10,"="x8);

    push @buffer,$header;

    
    $summary = 
        sprintf("%-10.10s %-4.4s %-8.8s %-16.16s %-16.16s %-8.8s %-10.10s    %d\n%87.87s",
                $self->{filer}->get_name(),
                $self->get_name(),
                $self->get_type(),
                $self->get_nodename(),
                $self->get_portname(),
                $self->get_status(),
                $self->get_version(),
                $self->get_initcnt(),
                "="x87
                );

    push @buffer, $summary, "Target Initiators:";

    foreach my $initiator ($self->get_initiator_wwpns()) {
        $summary = sprintf(" %d) %-24.24s", ++$cnt, $initiator);
        push @buffer, $summary;
    }


    push @buffer, "\n\n";

    return (wantarray()) ? @buffer : \@buffer;

}

################################################################################
# Subroutine Name: get_initiator_wwpns
# Objective:       Return initiator wwpns
# Details:         See POD documentation at the end of this file
################################################################################
sub get_initiator_wwpns {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{inits}) {
        return (wantarray()) ? @{$self->{inits}} : $self->{inits};
    } else {
        $self->_set_err("Error: Target::get_initiator_wwpns() Target object is missing initiator port name(s)");
        return  (wantarray()) ? () : [] ;
    }

    
}
################################################################################
# Subroutine Name: get_initcnt
# Objective:       Return initiators cnt logged into target
# Details:         See POD documentation at the end of this file
################################################################################
sub get_initcnt {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{initcnt}) {
        return $self->{initcnt};
    } else {
        return  0;
    }
    
}
################################################################################
# Subroutine Name: get_name
# Objective:       Return Target object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: Target::get_name() Target object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_nodename
# Objective:       Return Target nodename
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nodename {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{nodename}) {
        return $self->{nodename};
    } else {
        $self->_set_err("Error: Target::get_nodename() Target object is missing nodename");
        return "";
    }
}
################################################################################
# Subroutine Name: get_type
# Objective:       Return Target adapter type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_type {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{type}) {
        return $self->{type};
    } else {
        $self->_set_err("Error: Target::get_type() Target object is missing adapter type");
        return "";
    }
}

################################################################################
# Subroutine Name: get_portname
# Objective:       Return Target portname
# Details:         See POD documentation at the end of this file
################################################################################
sub get_portname {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{portname}) {
        return $self->{portname};
    } else {
        $self->_set_err("Error: Target::get_portname() Target object is missing portname");
        return "";
    }
}


################################################################################
# Subroutine Name: get_version
# Objective:       Return Target version
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{version}) {
        return $self->{version};
    } else {
        $self->_set_err("Error: Target::get_version() Target object is missing version");
        return "";
    }
}


################################################################################
# Subroutine Name: get_description
# Objective:       Return Target description
# Details:         See POD documentation at the end of this file
################################################################################
sub get_description {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{description}) {
        return $self->{description};
    } else {
        $self->_set_err("Error: Target::get_description() Target object is missing description");
        return "";
    }
}

################################################################################
# Subroutine Name: get_status
# Objective:       Return Target status
# Details:         See POD documentation at the end of this file
################################################################################
sub get_status {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{status}){
        return $self->{status};
    }
    else {
        $self->_set_err("Error: Target::get_status() Target object is missing status");
        return "";
    }
}


################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new Target object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, $args) = @_;

    my ($name, $self, $status, $version, $nodename, $portname, $description);
    my ($parent, $type, $log, @inits);

    $self = {};

    bless($self, $class);  

    foreach (keys %$args) {
        if (/^Slot$|^Name$/i) {
            if (not length($name = $args->{$_})) {
                $self->_set_err("Error: Target::_new() Invalid value for argument '$_'");
                return $self; 
            }        
            next;
        } elsif (/Status/i) {
            if (not length($status = $args->{$_})) {
                $self->_set_err("Error: Target::_new() Invalid value for argument 'Status'");
                return $self; 
            }        
            next;
        } elsif (/Rev/i) {
            if (not length($version = $args->{$_})) {
                $self->_set_err("Error: Target::_new() Invalid value for argument 'Rev'");
                return $self;
            }        
            next;
        } elsif (/Nodename/i) {
            if (not length(($nodename) = $args->{$_}=~/(\S+)/)) {
                $self->_set_err("Error: Target::_new() Invalid value for argument 'Nodename'");
                return $self; 
            }        
            next;
        } elsif (/Portname/i) {
            if (not length(($portname) = $args->{$_}=~/(\S+)/)) {
                $self->_set_err("Error: Target::_new() Invalid value for argument 'Portname'");
                return $self;
            }        
            next;
        } elsif (/Description/i) {
            if (not length($description = $args->{$_})) {
                $self->_set_err("Error: Target::_new() Invalid value for argument 'Description'");
                return $self; 
            }        
            next;
        } elsif (/Type/i) {
            if (not length($type = $args->{$_})) {
                $self->_set_err("Error: Target::_new() Invalid value for argument 'Adapter Type'");
                return $self; 
            }        
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args->{$_})) {
                $self->_set_err("Error: Target::_new() Invalid value for argument '-parent'");
                return $self; 
            }        
            next;
        } elsif (/-?loghdl/) {
            if (not length($log = $args->{$_})) {
                $self->_set_err("Error: Target::_new() Invalid value for argument '-loghdl'");
                return $self; 
            }        
            next;
        } else {
            $self->_set_err("Error: Target::_new() Invalid parameter specified \'$_\'");
            return $self;                                 
        }
    }     
    
    $self->{log} = $log if (defined $log);
    
    # check for filer parent object.  
    unless (defined $parent and length $parent) {
        $self->_set_err("Error: Target::_new()",
                        "Required parameter filer object reference is missing '-parent'");
        return $self; 
    }
    $portname =~ s/\://g;
    $nodename =~ s/\://g;

    $self->{name}        = $name;
    $self->{description} = $description;
    $self->{portname}    = $portname;
    $self->{nodename}    = $nodename;
    $self->{status}      = $status;
    $self->{type}        = $type;
    $self->{version}     = $version;
    $self->{filer}       = $parent;
    $self->{filer_name}  = $parent->get_name();

    @inits = $self->_get_initiators();
    $self->{inits} = \@inits;

    $self->{initcnt} = scalar(@inits);

    return $self;

}

################################################################################
# Subroutine Name: _get_initiators()
# Objective:       load initiators hooked to target 
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_initiators {
    my ($self) = @_;
    my ($cmd, $name, $init, @inits, @output);

    $name = $self->get_name();

    $cmd = "fcp show initiator " . $name;

    @output = $self->{filer}->execute($cmd, -hide => 1);
    if ($self->{filer}->error()) {
        $self->_set_err($self->{filer}->errmsg());
        return (wantarray()) ? () : [];
    }
    
    foreach my $line (@output){
        next if ($line !~ /^(\d+):/);
        ($init) = $line=~/^(\S+)/;
        $init =~ s/\://g;
        push(@inits, $init);
    }

    return (wantarray()) ? @inits : \@inits;

}

1;


#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Target.pm

=head1 Author

=begin html 

Gary Knopp


=end html

=head1 Description

Target provides an API interface to a filers target adapters.  The package should  
be instantiated through either of the two SanFiler2 package methods; get_target and get_targets(). 


=head1 Attributes

=head2 B<name>

string = Name of the target initiator object name

=head2 B<description>

string = Description of the target initiator

=head2 B<portname>

string = Name of the target portname

=head2 B<nodename>

string = Name of the target nodename

=head2 B<inits>

list = list of the target initiator wwpns

=head2 B<initcnt>

value = Number of target initiators count logged into target

=head2 B<status>

string = Status of the target initiator

=head2 B<type>

string = Type of the target adapter

=head2 B<version>

string = Version of the target initiator

=head1 Public Methods    

=head2 B<display()>

Display detailed target information

=over 4

=item Inputs

None

=item Outputs

Returns an array or array reference if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and an empty array or array reference
is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving target information");

    $Test->log($targObj->display());
    if ($targObj->error){
        $Test->log($targObj->errmsg());
        return $TCase::FAIL;
    }


=back

=head2 B<error()>

This method allows you to determine if any errors occured in the
previous method call.  Note that the error flag is reset at the
beginning of each public method, so it should be checked after B<every>
major method call.  If you do not check for errors, your calling
program may fail at some point and you will not know why.

=over 4 

=item Inputs

None

=item Outputs

Returns 1 if an error occured in the previous call, 0 otherwise.

=item Examples

    # Determine if error occurred...
    if ($targObj->error) {
        $Test->nlog("Error encountered in previous method call",$targObj->errmsg);
    }


=back

=head2 B<errmsg()>

This method allows you to get the text of the error message if an error
occurred on the previous method call.

=over 4

=item Inputs

None

=item Outputs

Returns the text of the last error message set.  If no error message is
set, an empty string will be returned.

=item Examples

    # Check for errors.
    if ($targObj->error) {
        $Test->nlog($targObj->errmsg);
    }


=back

=head2 B<get_description()>

Description of target adapter

=over 4

=item Inputs

none

=item Outputs

Returns the description of the target adapter.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The target description is '".$targObj->get_description()."'");


=back

=head2 B<get_initcnt()>

Total number of initiators connected to target

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containining the number of connected initiators.  If an error
occurs the error flag is set, errmsg is set to the failure reason, and 0 is passed
back as the return value;

=item Examples

    my $cnt=$targObj->get_initcnt();

    $Test->nlog($targObj->get_name()." has a total of $cnt initiators.");

=back

=head2 B<get_initiator_wwpns()>

List of initiator world wide portnames connected to target

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of initiator world wide portnames.  If an error
occurs the error flag is set, errmsg is set to the failure reason, and an empty array
or array referecedis returned.        

=item Examples

    my @wwpns=$targObj->get_initiator_wwpns();

    if ($targObj->error){
        $Test->nlog($targObj->errmsg());
    }
    else {
        $Test->nlog("Initiator WWPNS:",@wwpns);
    }


=back


=head2 B<get_name()>

Name of target adapter

=over 4

=item Inputs

none

=item Outputs

Returns the name of the target.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The target name is '".$targObj->get_name()."'");

=back

=head2 B<get_nodename(I<%args>)>

Retrieve nodename of target

=over 4

=item Inputs

=item * I<-format> (default: 0);

A boolean flag to indicate whether the nodename should be formatted 
(ex 10:a9:80:01:02:00:8e:1c), otherwise it is returned without the 
':' seperators.


=item Outputs

Returns a scalar containing the target nodename.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The target nodename is ".$targObj->get_nodename(-format=>1));

=back

=head2 B<get_portname(I<%args>)>

Retrieve portname of target

=over 4

=item Inputs

=item * I<-format> (default: 0)

A boolean flag to indicate whether the portname should be formatted 
(ex 10:a9:80:01:02:00:8e:1c), otherwise it is returned without the 
':' seperators.

=item Outputs

Returns a scalar containing the target portname.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The target portname is ".$targObj->get_portname(-format=>1));

=back

=head2 B<get_status()>

Retrieve the status of the target.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the target status.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The target status is ".$targObj->get_status());


=back

=head2 B<get_type()>

Retrieve the target type

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the target type.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    $Test->nlog("The target type is ".$targObj->get_type());


=back

=head2 B<get_version()>

Retrieve the fimrware version of the target adapter

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the targets adapters firmware version.  If 
an error occurs the error flag is set, errmsg is set to the failure 
reason, and "" is returned.        


=item Examples

    $Test->nlog("The target firmware version is ".$targObj->get_version());

=back

=cut
