# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Filer Qtree functions
## @author latif@netapp.com
## @status private
## @pod here

# Package Name
##############
package Qtree;

use strict;

# Package Requirements
######################
use Error;

use San::SanLog;
use San::SanUtil;

use vars qw( @ISA );

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: stats
# Objective:       get the stats for a qtree
# Details:         See POD documentation at the end of this file
################################################################################
sub stats() {
    my ($self) = @_;
    my ( $interface, $volume, $cmd_obj, $resultsHash_ref, $output );

    # there is no zapi for qtree stats
    $interface = "cli";
    $volume    = $self->get_volname();
    $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
	my $parent = $self->{parent};

    my %cmd_args = (
        -cmd           => "qtree-stats",
        -cmd_filer     => $parent->{filer},
        -cmd_interface => $interface,
        -volume        => $volume,
		-check_result     => 1,
        -cmd_privilege => "diag",
    );

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ( $cmd_obj->error() ) {
        $self->_set_err( "CmdsDriver failed: ", $cmd_obj->errmsg() );
        $self->sanlog(
            -msg       => join( "\n", $cmd_obj->errmsg() ),
            -msg_level => 1
        );
        return $ClientUtil::FALSE;
    }

    # ensure result_check passed
    unless ( $resultsHash_ref->{result_check} ) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};
    return $output;
}
################################################################################
# Subroutine Name: security
# Objective:       set security for  a qtree
# Details:         See POD documentation at the end of this file
################################################################################
sub security() {
    my ( $self, %args ) = @_;
    my ( $interface, $path, $cmd_obj, $resultsHash_ref, $output,
        $security_style );

    # there is no zapi support for security
    $interface = "cli";
    $path      = $self->{path};
    $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
	my $parent = $self->{parent};

    #
    # Parse the arguments
    #
    foreach ( keys %args ) {
        if (/-?security_style/) {
            if ( not length( $security_style = $args{$_} ) ) {
                $self->_set_err("Invalid value for argument '-security_style'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }

    my %cmd_args = (
        -cmd           => "qtree-security",
        -cmd_filer     => $parent->{filer},
        -cmd_interface => $interface,
        -path          => $path,
		-check_result     => 1,
        -cmd_privilege => "diag",
    );
    $cmd_args{'security-style'} = $security_style if ($security_style);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ( $cmd_obj->error() ) {
        $self->_set_err( "CmdsDriver failed: ", $cmd_obj->errmsg() );
        $self->sanlog(
            -msg       => join( "\n", $cmd_obj->errmsg() ),
            -msg_level => 1
        );
        return $ClientUtil::FALSE;
    }

    # ensure result_check passed
    unless ( $resultsHash_ref->{result_check} ) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};
    if ($security_style) {
        return $ClientUtil::TRUE;
    }
    else {
        return $output;
    }
}
################################################################################
# Subroutine Name: oplocks
# Objective:       enable or disable oplocks on a qtree
# Details:         See POD documentation at the end of this file
################################################################################
sub oplocks() {
    my ( $self, %args ) = @_;
    my ( $interface, $path, $cmd_obj, $resultsHash_ref, $output, $option );

    # there is no zapi is support for oplocks
    $interface = "cli";
    $path      = $self->{path};
    $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
	my $parent = $self->{parent};
    #
    # Parse the arguments
    #
    foreach ( keys %args ) {
        if (/-?option/) {
            if ( not length( $option = $args{$_} ) ) {
                $self->_set_err("Invalid value for argument '-option'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }

    my %cmd_args = (
        -cmd           => "qtree-oplocks",
        -cmd_filer     => $parent->{filer},
        -cmd_interface => $interface,
        -path          => $path,
		-check_result     => 1,
        -cmd_privilege => "diag",
    );

    $cmd_args{'option'} = $option if ($option);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ( $cmd_obj->error() ) {
        $self->_set_err( "CmdsDriver failed: ", $cmd_obj->errmsg() );
        $self->sanlog(
            -msg       => join( "\n", $cmd_obj->errmsg() ),
            -msg_level => 1
        );
        return $ClientUtil::FALSE;
    }

    # ensure result_check passed
    unless ( $resultsHash_ref->{result_check} ) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};
    if ($option) {
        return $ClientUtil::TRUE;
    }
    else {
        return $output;
    }
}
################################################################################
# Subroutine Name: rename
# Objective:       rename a qtree
# Details:         See POD documentation at the end of this file
################################################################################
sub rename() {
    my ( $self, %args ) = @_;
    my ( $new_qtree_name, $interface,       $volume, $new_path,    $path );
    my ( $cmd_obj,        $resultsHash_ref, $output, $output_hash, $status );

    #
    # Parse the arguments
    #
	my $parent = $self->{parent};
    foreach ( keys %args ) {
        if (/-?new_qtree_name/) {
            if ( not length( $new_qtree_name = $args{$_} ) ) {
                $self->_set_err("Invalid value for argument '-new_qtree_name'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }

    if ( !defined($new_qtree_name) ) {
        $self->_set_err("new_qtree_name is required argument");
        return $ClientUtil::FALSE;
    }
    
    # only implemented for zapi since it is not a san function
    $interface = "zapi";
    $volume    = $self->get_volname();
    $new_path  = "/vol/$volume/$new_qtree_name";
    $path      = $self->{path};
    $cmd_obj   = $self->{parent}->{parent}->cmd_obj();

    my %cmd_args = (
        -cmd              => "qtree-rename",
        -cmd_filer        => $parent->{filer},
        -cmd_interface    => $interface,
        "-new-qtree-name" => $new_path,
        -qtree            => $path,
        -check_result     => 1,
        -cmd_privilege    => "diag",
    );
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ( $cmd_obj->error() ) {
        $self->_set_err( "CmdsDriver failed: ", $cmd_obj->errmsg() );
        $self->sanlog(
            -msg       => join( "\n", $cmd_obj->errmsg() ),
            -msg_level => 1
        );
        return $ClientUtil::FALSE;
    }

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ( $resultsHash_ref->{result_check} ) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    if ( $interface eq "zapi" ) {
        # If interface is zapi, convert the output from xml to a hash.
        $output_hash = $self->{parent}->{parent}->xml2hash($output);
        $status      = $output_hash->{status};
        if ( $status eq "failed" ) {
            $self->_set_err( "ZAPI failed: ", $output_hash->{reason} );
            return $ClientUtil::FALSE;
        }
    }
    $self->{name} = $new_qtree_name;
    $self->{path} = "/vol/$volume/$self->{name}";
    return $ClientUtil::TRUE;
}
################################################################################
# Subroutine Name: destroy
# Objective:       destroy a qtree
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;

    my $force     = 0;
    my $parent = $self->{parent};
    my $path      = $self->{path};
    my $name      = $self->get_name();
    my $volume = $parent->get_name();

    #my $filer     = $self->get_filername();
    my $interface = $self->{parent}->{parent}->interface();
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my @qtrees;

    $self->_clear_err();

    #
    # Parse arguments
    #
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            if (defined $force) {
                $force = 1;
            }
            next;
        }
    }     
	
    # Determine if the full path to the qtree is provided
    unless (defined $path) {
        $self->_set_err("qtree path is missing from object");
        return $ClientUtil::FALSE;
    }
    
    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"              => "qtree-delete",
                    "-cmd_filer"        => $parent->{filer},
                    "-check_result"     => 1,
                    "-check_output"     => 1,
                    "-cmd_interface"    => $interface,
                    "-path"             => "/vol/$volume/$name",
					-cmd_privilege    => "diag",
                   );

    $cmd_args{"-force"} = "1" if ($force and ($interface eq "zapi"));
    $cmd_args{"-force"} = ""  if ($force and ($interface eq "cli"));
    
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()) );
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};
    
    # Ensure result_check and output_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status = $output_hash->{status};
    
        if ($status eq "failed") {
            $self->_set_err("Unable to delete qtree", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } else {
        my @output = split("\n", $output);
    
        # Message of success is not always returned depending on qtree quota.
        if (defined $output[0] and length $output[0]) {     
            if ($output[0] =~ /qtree destroy:/i) {         
                $self->_set_err("Unable to delete qtree named '$name'", @output);
                return $ClientUtil::FALSE;
            }
        }
    }

    # Remove the qtreeObj from the filer volume object
    foreach my $qtreeObj (@{$self->{parent}->{qtrees}}) { 
        if ($qtreeObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @qtrees, $qtreeObj;
        }
    }

    $self->{parent}->{qtrees} = \@qtrees;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_filername
# Objective:       Return qtree object filername
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filername {
    my ($self) = @_;

    $self->_clear_err();
                
    if (defined $self->{filer}){
        return $self->{filer}->get_name();
    } else {
        $self->_set_err("qtree object is missing filer name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_name
# Objective:       Return qtree object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();
                
    if (defined $self->{name}){
       return $self->{name};
    } else {
        $self->_set_err("qtree object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_path
# Objective:       Return qtree object path
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path {
    my ($self) = @_;

    $self->_clear_err();
                
    if (defined $self->{path}){
        return $self->{path};
    } else {
        $self->_set_err("qtree object is missing path");
        return "";
    }
}

################################################################################
# Subroutine Name: get_volname
# Objective:       Return qtree's volume name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_volname {
    my ($self) = @_;

    $self->_clear_err();
                
    if (defined $self->{parent}){
        return $self->{parent}->get_name();
    } else {
        $self->_set_err("qtree object is missing volume name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_status
# Objective:       Return qtree status
# Details:         See POD documentation at the end of this file
################################################################################
sub get_status {
    my ($self) = @_;

    $self->_clear_err();
                
    if (defined $self->{status}){
        return $self->{status};
    } else {
        $self->_set_err("qtree object is missing status");
        return "";
    }
}

################################################################################
# Subroutine Name: get_exports
# Objective:       QTREE NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_exports {

    return (wantarray()) ? () : \();
}

################################################################################
# Subroutine Name: get_quota
# Objective:       Return qtree quota
# Details:         See POD documentation at the end of this file
################################################################################
sub get_quota {
    my ($self) = @_;

    $self->_clear_err();
                
    if (defined $self->{quota}){
        return $self->{quota};
    } else {
        $self->_set_err("qtree object is missing quota");
        return 0;
    }
}

################################################################################
# Subroutine Name: get_parent
# Objective:       return parent (Volume) object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_parent {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined $self->{parent}) {
        return $self->{parent};
    } 

    $self->_set_err("qtree object has no parent");
    return {};
}

################################################################################
# Subroutine Name: is_exported
# Objective:       Return true if exported to a host or a specific IP
# Details:         See POD documentation at the end of this file
################################################################################
sub is_exported {
    my ($self, %args) = @_;
    
    my $host_ip;
    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?host_ip/) {
            if (not length($host_ip = $args{$_})) {
                $self->_set_err("Invalid value for argument '-host_ip'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (defined $self->{exported}){
        # if the user passed in an host_ip, see if 
        # this qtree is exported to that host_ip 
        if (defined $host_ip) {
            foreach my $host_exp (keys %{$self->{exports}}) {
                if ($host_ip eq $host_exp) {
                    return $ClientUtil::TRUE;
                }
            }

            # At this point we should have returned if the qtree host was found
            # return false since we went through the exports and didn't find it.
            return $ClientUtil::FALSE;
        } else {
            # otherwise just return if the qtree is exported
            return $ClientUtil::TRUE;
        } 
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: export()
# Objective:       Export a qtree to host/ip
# Details:         See POD documentation at the end of this file
################################################################################
sub export {
    my ($self, %args) = @_;

    my ($cmd, @output, $host_ip, $permissions);
    my $path = $self->get_path();
    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?host_ip/) {
            if (not length($host_ip = $args{$_})) {
                $self->_set_err("Invalid value for argument '-host_ip'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?permissions/) {
            if (not length($permissions = $args{$_})) {
                $self->_set_err("Invalid value for argument '-permissions'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Determine if the host_ip is provided
    unless (defined $host_ip and length $host_ip) {
        $self->_set_err("host_ip is a required argument");
        return $ClientUtil::FALSE;
    }

    # if permissions are not defined default them to read/write
    unless (defined $permissions) {
        $permissions = "rw";
    }

    $self->sanlog(-msg => "Exporting qtree $path to host IP $host_ip");

    # Example export command: exportfs -io rw=192.168.0.33 /vol/nate_1_1_hp2620_rtp11
    $cmd = "exportfs -io $permissions=" . $host_ip . " " . $path;

    @output = $self->{parent}->{parent}->execute($cmd, -hide => 1);

    if ($self->{parent}->{parent}->error() or grep(/exportfs\:/, @output)) {
        $self->_set_err($self->{parent}->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{exported} = $ClientUtil::TRUE;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unexport()
# Objective:       Unexport a qtree from a host/ip
# Details:         See POD documentation at the end of this file
################################################################################
sub unexport {
    my ($self) = @_;

    my ($cmd, @output);
    my $path = $self->get_path();

    $self->_clear_err();

    $self->sanlog(-msg => "Unexporting qtree $path");

    # Example unexport command: exportfs -u /vol/nate_1_1_hp2620_rtp11
    $cmd = "exportfs -u " . $path;

    @output = $self->{parent}->{parent}->execute($cmd, -hide => 1);

    if ($self->{parent}->{parent}->error() or grep(/exportfs\:/, @output)) {
        $self->_set_err($self->{parent}->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{exported} = $ClientUtil::FALSE;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: resize_quota()
# Objective:       QTREE NO-OP - for now
# Details:         See POD documentation at the end of this file
################################################################################
sub resize_quota {

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: set_quota()
# Objective:       Define quota for Qtree
# Details:         See POD documentation at the end of this file
################################################################################
sub set_quota {
    my ($self, %args) = @_;

    my ($path, $quota, $name, $volume, $cmd, $output, @output, $errmsg, $parent);

    $parent = $self->{parent};
    $path = $self->{path};
    $name = $self->get_name();
    $volume = $parent->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?quota/) {
            if (not length($quota = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->_clear_err();

    # Check to see if quota was provided
    if (not defined($quota)) {
        $self->_set_err("No quota provided");
        return $ClientUtil::FALSE;
    }

    # Check to see if a quota is defined, if so set a quota on the qtree
    # This MUST be done with the CLI because ZAPI support in 7.1, 7.2 and even 7.3
    # is not 100%, there are some missing ZAPI calls.
    if (defined($quota) and $quota != 0) {
        # Gather the quota information to set limits on qtree
        $cmd = "priv set -q diag;rdfile /etc/quotas";
        my @quotas = $parent->{parent}->execute($cmd, -hide => 1);

        # Output the quotas file before modification just in case the original is needed
        $self->sanlog(-msg => "Quotas file /etc/quotas before modification:\n" . join("\n", @quotas) . "\n");

        # Remove the qtree from the list and recreate the quotas string
        @quotas = grep(!/$volume\/$name/i, @quotas);

        # Create quotas string
        # Example: /vol/nate_1_1/nate_qtree_hp2620-rtp11_hpux_0 tree 100M - - - -
        push(@quotas, $path . " tree " . $quota . " - - - -");

        # Clear out the quotas file
        $cmd = "priv set -q diag;rm /etc/quotas";
        $parent->{parent}->execute($cmd, -hide => 1);
        if ($parent->error()) {
            $self->_set_err($parent->errmsg());
            return $ClientUtil::FALSE;
        }

        # Write the new quotas file line by line
        foreach my $line (@quotas) {
            $cmd = "priv set -q diag;wrfile -a /etc/quotas \"$line\"";
            @output = $parent->{parent}->execute($cmd, -hide => 1);
            if ($parent->error()) {
                $self->_set_err($parent->errmsg());
                return $ClientUtil::FALSE;
            }
        }

        # To make these changes effect, turn off and on the quota
        $cmd = "quota off $volume";
        @output = $parent->{parent}->execute($cmd, -hide => 1);

        # Turn on quotas on the parent volume
        $cmd = "quota on -w $volume";
        @output = $parent->{parent}->execute($cmd, -hide => 1);
        if ($parent->error() or grep(/contains no entries for volume|could not read quotas file/, @output)) {
            $self->_set_err("Unable to properly set quota for qtree $path");
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Quota limit is invalid ($quota)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new qtree object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($exist, $exported, $name, $parent, $self, $quota, $quota_kb, $status); 
    $status = "online";

    $self={};

    bless $self,$class; 

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-name'");
                return $self; 
            }        
            next;
        } elsif (/-?quota/) {
            if (not length($quota = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota'");
                return $self; 
            }        
            next;
        } elsif (/-?exported/) {
            if (not length($exported = $args{$_})) {
                $self->_set_err("Invalid value for argument '-exported'");
                return $self; 
            }        
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self; 
            }        
            next;
        } elsif (/-?exist/) {
            if (not length($exist = $args{$_})) {
                $self->_set_err("Invalid value for argument '-exist'");
                return $self; 
            }        
            next;
        }
    }
   
    # check for required parameter qtree name.  It should not 
    # contain path information
    unless (defined $name and length $name and $name !~ /\//) {
        $self->_set_err("Missing or invalid value for required parameter '-name'");
        return $ClientUtil::FALSE;
    }

    # Check for Volume, parent object.  
    unless (defined $parent and length $parent) {
        $self->_set_err("Required parameter parent volume object is missing '-parent'");
        return $ClientUtil::FALSE;
    }

    # Check for parameter qtree quota. It should contain units
    unless (defined $quota and length $quota and $quota =~ /\d+[b|k|m|g|t]?/i) {
        $quota_kb = 0;
    } else {
        $quota_kb = &ClientUtil::convert_to_bytes($quota, "k");
    }

    # inititialize object variables
    $self->{filer}  = $self->{parent}->{filer};
    $self->{parent} = $parent;
    $self->{quota}   = $quota_kb;
    $self->{name}   = $name;

    #
    # If we are adding an existing qtree then we are just
    # storing qtree attributes
    #
    if (defined $exist and $exist == 1) {
        $self->{path} = "/vol/" . $self->{parent}->get_name() . "/" . $name;
        $self->{status}  = $status;
        $self->{exported} = $exported;
    } else {
        $self->{path} = "/vol/" . $self->{parent}->get_name() . "/" . $name;
        $self->{status}   = "online";
        $self->{exported}  = 0;

        $self->_create();
        return $self if ($self->error);
    }
    return $self;
}

################################################################################
# Subroutine Name: _create()
# Objective:       create a new qtree
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self) = @_;

    my ($path, $quota, $name, $volume, $cmd, $output, @output,
        $errmsg, $resultsHash_ref, $cmd_obj, $interface);

    my $parent = $self->{parent};

    $path = $self->{path};
    $quota = $self->{quota};
    $name = $self->get_name();
    $volume = $parent->get_name();
    $cmd_obj = $parent->{parent}->cmd_obj();
    $interface = $parent->{parent}->interface();

    $self->_clear_err();

    # create qtree
    if (lc($interface) eq "cli") {
        $resultsHash_ref = $cmd_obj->execute_cmd(
                                              -cmd           => "qtree-create",
                                              -cmd_filer     => $parent->{filer},
                                              -cmd_interface => $interface,
                                              -check_result  => 1,
                                              -qtree         => "/vol/$volume/$name");
    } else {
        $resultsHash_ref = $cmd_obj->execute_cmd(
                                               -cmd           => "qtree-create",
                                               -cmd_filer     => $parent->{filer},
                                               -cmd_interface => $interface,
                                               -check_result  => 1,
                                               -volume        => "$volume",
                                               -qtree         => $name);
    }

    # log results
    $parent->{parent}->log_cmd_results($resultsHash_ref);

    # check for CmdsDriver failure
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # get the output
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # check for command failure
    if (lc($interface) eq "zapi") {
        # If interface is zapi, convert the output from xml to a string.
        my $output_hash = $parent->{parent}->xml2hash($output);
        my $status = $output_hash->{status};

        if ($status eq "failed") {
            $self->_set_err("ZAPI failed:", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } else {
        my @output = split("\n", $output);
        if (defined $output[0] and length $output[0]) {
            if ($output[0] =~ /qtree:/i ) {
                $self->_set_err("CLI failed: @output");
                return $ClientUtil::FALSE;
            }
        }
    }

    # Check to see if a quota is defined, if so set a quota on the qtree
    # This MUST be done with the CLI because ZAPI support in 7.1, 7.2 and even 7.3
    # are not 100%, there are some missing ZAPI calls.
    if (defined($quota) and $quota != 0) {
        $self->set_quota(-quota => $quota);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

qtree.pm

=head1 Authors

=begin html 

Michael Latif

=end html

=head1 Description

Qtree provides an API interface to a filer qtree throu the use of rsh.
Instantiating this object is quite different to Lun.pm, don't use SanFiler2
as Lun.pm  recommends, but use the create method provided in Volume.pm
create_qtree.

=head1 Attributes

=head2 B<name>

string = Qtree name

=head2 B<quota>

string = Quota size

=head2 B<exported>

boolean = Exported or not

=head2 B<parent>

object = Parent object (Volume object)

=head2 B<destroy(%args)>

Destroy a filer qtree

=over 4

=item Inputs

=item * I<-force> (default: 0)

The option will automatically unexport the qtree so that it can be destroyed.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occurred the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Destroy qtree");

    $qtreeObj->destroy();
    if ($qtreeObj->error()){
        $Test->log($qtreeObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<set_quota(%args)>

Set the requested quota on the qtree.

=over 4

=item Inputs

=item * <-quota>

This option will set the provided quota on the qtree.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occurred the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Set qtree quota");

    $qtreeObj->set_quota(-quota => "2gb");
    if ($qtreeObj->error()){
        $Test->log($qtreeObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<is_exported(%args)>

Check to see if a qtree is alredy exported

=over 4

=item Inputs

=item * <-host_ip>

Host ip to look for, used to determine if the qtree is exported to this host ip.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occurred the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Check to see if qtree is exports");

    $qtreeObj->set_quota(-host_ip => "10.60.99.99");
    if ($qtreeObj->error()){
        $Test->log($qtreeObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<export(%args)>

Export a qtree to a particular host_ip with specific permissions.

=over 4

=item Inputs

=item * <-host_ip>

Provided to export a qtree to.

=item * <-permissions> (Default: "rw")

Provide permissions to be used when exporting a qtree to a host_ip.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occurred the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Export qtree");

    $qtreeObj->export(-host_ip => "10.60.99.99", -permissions => "rw");
    if ($qtreeObj->error()){
        $Test->log($qtreeObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<unexport(%args)>

Unexport a qtree.

=over 4

=item Inputs

- NONE -

=item Outputs

Returns a scalar value of 1 if successful.  If an error occurred the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- UNExport qtree");

    $qtreeObj->unexport();
    if ($qtreeObj->error()){
        $Test->log($qtreeObj->errmsg());
        return $TCase::FAIL;
    }

=back

=end html
