#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/San/SanLog.pm#1 $

# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides log handling methods that
## should be inherited by other object oriented
## packages
##
## @author pressley@netapp.com
## @status Shared
## @pod here

package SanLog;

use lib();

#
# Compiler directives.
#
use strict;

use San::SanUtil;
use File::Basename;

#
# Log level constants
#
use constant LOG_QUIET        => 0;
use constant LOG_VERBOSE_LOW  => 1;
use constant LOG_VERBOSE_HIGH => 3;
use constant LOG_DEBUG_LOW    => 5;
use constant LOG_DEBUG_HIGH   => 7;
use constant LOG_DEBUG_ALL    => 101;

###############################################################################
# Method:    sanlog
# Objective: Log a message to the standard NATE log and to the loghdl.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sanlog
{
    my($self,%args) = @_;

    my @msg;
    my $msg_level      = 0;
    my $use_loghdl     = 1;
    my $indent         = 0;
    my $add_caller     = 1;
    my $add_call_stack = 0;
    my $debug_level    = 1;

    #
    # If the parent object has a debug_level we will use it and compare it to
    # the msg_level.
    #
    if (defined($self->{debug_level}))
    {
        $debug_level = $self->{debug_level};
    }
    foreach (keys %args)
    {
        if (/-?msg$/)
        {
            if (length($args{$_}))
            {
                #
                # If we are given an array, push the arg as an array, otherwise
                # treat it like a string and push that string.
                #
                push @msg, (ref($args{$_}) eq "ARRAY") ? @{$args{$_}} : $args{$_};
            } else
            {
                push @msg , "NO MESSAGE in sanlog";
            }
            next;
        }
        elsif (/-?msg_level$/)
        {
            $msg_level = $args{$_};
            unless (length($msg_level) and ($msg_level >= 0))
            {
                #
                # Ignore bogus message levels
                #
                $msg_level  = 1;
            }
            next;
        }
        elsif (/-?indent$/)
        {
            $indent = $args{$_};
            unless (length($indent) and ($indent >= 0))
            {
                #
                # Ignore bogus values
                #
                $indent  = 0;
            }
            next;
        }
        elsif (/-?add_caller$/)
        {
            $add_caller = $args{$_};
            unless (length($add_caller) and ($add_caller >= 0))
            {
                #
                # Ignore bogus values
                #
                $add_caller = 0;
            }
            next;
        }
        elsif (/-?add_call_stack$/)
        {
            $add_call_stack = $args{$_};
            unless (length($add_call_stack) and ($add_call_stack >= 0))
            {
                #
                # Ignore bogus values
                #
                $add_call_stack = 0;
            }
            next;
        }
        elsif (/-?use_loghdl$/)
        {
            $use_loghdl = $args{$_};
            if (not length($use_loghdl = $args{$_}) or
                           $use_loghdl !~ /no|yes|1|0|off|on/i)
            {
                #
                # Ignore bogus values
                #
                $use_loghdl  = 1;
            }
            else 
            {
                #
                # Set it to 1 or 0
                #
                if ($use_loghdl =~ /yes|1|on/i)
                {
                    $use_loghdl = 1;
                }
                else
                {
                    $use_loghdl = 0;
                }
            }
            next;
        } # end if -use_loghdl
    } # end foreach arg

    return if ($msg_level > $debug_level);
    
    #
    # Create a buffer with indent spaces
    #
    my $buffer = "";
    for (my $i = 0 ; $i < $indent ; $i++)
    {
        $buffer .= " ";
    }
    $buffer .= "- ";

    if (($use_loghdl) and (defined $self->{log}))
    {
        foreach my $line (@msg)
        {
            #
            # If a line contains a newline replace it with a
            # newline followed by the buffer so indentation is
            # correct.
            #
            $line =~ s/\n/\n   $buffer/g;
            print {${$self->{log}}} "   " . $buffer . $line."\n";
        }
    } # end if log defined

    #
    # Check to see if we should add the call stack or the caller information.
    #
    if (($debug_level > 100) || ($add_call_stack))
    {
        my @call_stack = _print_call_stack();
        push(@msg, @call_stack);
    }

    if ($add_caller) {
        #
        # If add_caller was set, prefix the line with filename::method()
        #
        my $caller_prefix = _caller_prefix();
        my $first_line = $msg[0];
        $msg[0] = "$caller_prefix $first_line";
    }

    #
    # Log handle or not, we always write to the NATE log.
    #
    &SanUtil::nate_log('-comment', join("\n", @msg));
} # end sanlog()

###############################################################################
# Method:    get_debug_level, set_debug_level
# Objective: Return or set the debug level attribute
# Details:   Get simply returns an integer representing the current
#            debug level.  Set will take an integer and set the
#            current debug level.
###############################################################################
sub get_debug_level
{
    my ($self) = @_;
    
    return $self->{debug_level};
} # end of get_debug_level method

sub set_debug_level
{
    my ($self, $level) = @_;
    
    $self->{debug_level} = $level;
    my $self_ref = ref($self);
    $self->sanlog(-msg => "$self_ref set its San debug level to $level",
                  -msg_level => LOG_VERBOSE_HIGH);

    #
    # When this level gets greater than 100, sanlog will print out the
    # call stack.  I'd like Error.pm to do the same.
    #
    if ($level > 100)
    {
        $self->{call_stack_on_error} = 1;
    }
    else
    {
        $self->{call_stack_on_error} = 0;
    }
    return 1;
} # end of set_debug_level method

###############################################################################
# Method:    entry
# Objective: Log an entry point message to the standard NATE log 
#            and to the loghdl.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub entry {
    my ($self, %args) = @_;
 
    %args = &SanUtil::strip_dash(%args);

    my $level = 0;
    my $msg = "-> Enter " . _parent_caller() . "():" . _parent_lineno();
   
    if (exists $args{msg_level}) {
        $level = $args{msg_level};
    }
    # if the entry point has a msg, append to the message
    if (exists $args{msg}) {
        $msg .= " - $args{msg}";
    }
    # let sanlog logs the message
    $self->sanlog(-msg => $msg,
                  -msg_level => $level);

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    exit
# Objective: Log an exit point message to the standard NATE log 
#            and to the loghdl.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub exit {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my $level = 0;
    my $msg = "<- Exit " . _parent_caller() . "():" . _parent_lineno();
   
    if (exists $args{msg_level}) {
        $level = $args{msg_level};
    }
    # if the entry point has a msg, append to the message
    if (exists $args{msg}) {
        $msg .= " - $args{msg}";
    }
    # let sanlog logs the message
    $self->sanlog(-msg => $msg,
                  -msg_level => $level);

    return $SanUtil::TRUE;
}

################################################################################
##                               Private Methods
################################################################################

################################################################################
## Method:    _print_call_stack
## Objective: Use caller to get and return the call stack
################################################################################
sub _print_call_stack
{
    my $depth  = 2;
    push my @output,
         "--------------------------------------------------------------------";
    push @output, "Call Stack:";
    push @output,
         "--------------------------------------------------------------------";
    while ( (my ($package,$filename,$line,$subroutine) = caller($depth)) &&
            ($depth < 7))
    {
        push @output,
             $depth . " " . $filename . "::" . $subroutine . ": $line";
        $depth++;
    } # end while
    push @output,
         "--------------------------------------------------------------------";
    return @output;
} # end _print_call_stack()

################################################################################
## Method:    _msg_level_string
## Objective: Translate the message level into a string to print
################################################################################
sub _msg_level_string
{
    my ($level) = @_;

    if ($level le LOG_QUIET)
    {
        return "LOG_QUIET";
    }
    elsif ($level le LOG_VERBOSE_LOW)
    {
        return "LOG_VERBOSE_LOW";
    }
    elsif ($level le LOG_VERBOSE_HIGH)
    {
        return "LOG_VERBOSE_HIGH";
    }
    elsif ($level le LOG_DEBUG_LOW)
    {
        return "LOG_DEBUG_LOW";
    }
    elsif ($level le LOG_DEBUG_HIGH)
    {
        return "LOG_DEBUG_HIGH";
    }
    else
    {
        return "LOG_DEBUG_ALL";
    }
} # end _msg_level_string

# _parent_caller returns the subroutine of caller of depth 2
sub _parent_caller { (caller(2))[3] }

# _caller_prefix returns filename::subroutine of caller of depth 2
sub _caller_prefix {
    my ($package,$filename,$line,$subroutine) = caller(2);
    my $basename = basename($filename);
    my ($file, $extension) = split /\./, $basename;

    return $file . " -- "  . $subroutine . "()";
}

# _parent_lineno returns the line number of caller of depth 1.
# this will return the line number within the above subroutine
sub _parent_lineno { (caller(1))[2] }

1;

=pod

=head1 Package Name

SanLog.pm

=head1 Author

David Pressley

=head1 Description

SanLog provides an API for performing logging and managing logging levels.
Use @ISA to inherit these methods into another package.  Use the method sanlog 
in your package to log a comment to the log file and/or a log handle.  The
following constants are provided help provide consistent logging levels.

=over

=item LOG_QUIET

=item LOG_VERBOSE_LOW

=item LOG_VERBOSE_HIGH

=item LOG_DEBUG_LOW

=item LOG_DEBUG_HIGH

=item LOG_DEBUG_ALL

=back

Note that LOG_DEBUG_ALL will cause a partial call stack to be logged with every
log message.  Two attributes will be added to your object when you inherit
this module: debug_level, call_stack_on_error.  The debug_level is the
current message level that is used to determine whether or not to log a
given message.  The call_stack_on_error attribute is set by set_debug_level()
when the debug level is set to LOG_DEBUG_ALL and is used by the Error.pm package. 

=over 4

=item Examples

use vars qw( @EXPORT @ISA);

@ISA = qw(SanLog);

=back

=head1 Public Methods

=head2 B<sanlog()>

This method allows you to log a message to the log file and optionally to a log
handle.  The API allows for 3 optional fields -use_loghdl, -msg_level, -indent.
If use_loghdl is given the message will be sent $self->loghdl in addition to
the standard NATE log.  If the msg_level is given, it must be greater than or
equal to $self->debug_level or it will be ignored.  If the debug_level is greater
than 100, we will print the call stack.

=over 4

=item Inputs

=item * -msg

An array of messages.

=item * I<-use_loghdl> (default to 1)

Don't print to $self->loghdl if this flag is 0 or "no".

=item * I<-msg_level> (default to 1)

Ignore this message if the msg_level is less than $self->debug_level.

=item * I<-indent> (default to 0)

Number of spaces to print before each line in the msg.

=item Outputs

None.

=item Examples

 $self->sanlog(-msg        => "Aggr test_aggr created",
               -msg_level  => LOG_VERBOSE_LOW,
               -use_loghdl => 1,
               -indent     => 2)

=back

=head2 B<get_debug_level()>

Find the current debug level.

=over 4

=item Outputs

Returns a scalar that's the current debug level.

=item Examples

    #
    # Instantiate San Object and get debug level.
    #
    my $sanObj = San::new();
    my $debug = $sanObj->get_debug_level();

=back

=head2 B<set_debug_level(level)>

Set the current debug level.

=over 4

=item Inputs

=item * level

New debug level.

=item Outputs

Returns a scalar 1.

=item Examples

    #
    # Instantiate San Object and set debug level.
    #
    my $sanObj = San::new();
    my $debug = $sanObj->set_debug_level(LOG_DEBUG_LOW);

=back

=cut
