# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary
## Provides utility functions to collect information and
## validate controllers and switches in a Fabric Metrocluster
##
## @author archanam, mc-dev, blocks-host automation, kaaron
##
## @status shared
##

package SanMCUtil;

# Compiler directives
use strict;

# Module import
use Data::Dumper;
use San::SanLog;
use San::SanUtil;
use Error;
use Tharn;
use San::SanSwitch;
use San::SanClient::ClientUtil;
use Armadillo;

use constant DO_NOTHING => 0;
use constant MERGE_DATA => 1;
use constant COPY_DATA  => 2;
use constant HOURS_12   => 720;

use constant DEBUG => 0;

use constant MAX_DEVICE_PATHS => 2;
use vars qw(
            @EXPORT
            @ISA
            $FAULT_SWITCH_PORT_RETRY_COUNT
           );

# Inherit from the Error, SanLog classes
@ISA = qw(Error SanLog);

&param("FAULT_SWITCH_PORT_RETRY_COUNT", "-default", 20);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new SanMCUtil object.
# Details:   Accepts an array of controllers and switches
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {};
    my (@controllers, @fabric1_switches, @fabric2_switches,
        @fabric1_switches_raw, @fabric2_switches_raw,
        @fab1_fcvi_isl, @fab2_fcvi_isl, @bridges);
    my ($debug_level, $error_level, $resync_timeout,
        $drives_to_skip, $aggrs_to_skip, $drive_types_to_skip,
        $hw_based_own);

    bless $self, $class;

    #Initialize some params
    $debug_level = $error_level = $resync_timeout = 0;
    $drives_to_skip = $aggrs_to_skip = "";
    $hw_based_own = $ClientUtil::FALSE;

    # parse arguments
    foreach (keys %args) {
        if (/-?controllers/) {
            @controllers = @{$args{$_}};
        } elsif (/-?fabric1_switches/) {
            @fabric1_switches_raw = @{$args{$_}};
        } elsif (/-?fabric2_switches/) {
            @fabric2_switches_raw = @{$args{$_}};
        } elsif (/-?fabric1_fcvi_isl/) {
            @fab1_fcvi_isl = @{$args{$_}};
        } elsif (/-?fabric2_fcvi_isl/) {
            @fab2_fcvi_isl = @{$args{$_}};
        } elsif (/-?debug_level/) {
            $debug_level = $args{$_};
        } elsif (/-?error_level/) {
            $error_level = $args{$_};
        } elsif (/-?resync_timeout/) {
            $resync_timeout = $args{$_};
        } elsif (/-?drives_to_skip/) {
            $drives_to_skip = $args{$_};
        } elsif (/-?drive_types_to_skip/) {
            $drive_types_to_skip = $args{$_};
        } elsif (/-?aggrs_to_skip/) {
            $aggrs_to_skip = $args{$_};
        } elsif (/-?hw_based_own/) {
            $hw_based_own = $args{$_};
        } elsif (/-?bridges/) {
            @bridges = @{$args{$_}};
        }
    }

    # check for required parameters
    if (not scalar(@controllers)) {
        $self->_set_err("Must specify parameter -controllers");
        return $self;
    }

    if (scalar(@fabric1_switches_raw) < 2) {
        $self->_set_err("Must specify parameter fabric1_switches");
        return $self;
    }

    if (scalar(@fabric2_switches_raw) < 2) {
        $self->_set_err("Must specify parameter fabric2_switches");
        return $self;
    }

    #Initialize some paramters
    $self->{warnings_set} = $ClientUtil::FALSE;
    @{$self->{warnings}} = ();
    $self->set_debug_level($debug_level);
    $self->{ERROR_LEVEL} = $error_level;
    $self->{RESYNC_TIMEOUT} = $resync_timeout;

    # Save the arrays in the object
    $self->{controllers} = \@controllers;
    $self->{HW_BASED_OWNERSHIP} = $hw_based_own;
    $self->{DRIVES_TO_SKIP} = $drives_to_skip;
    $self->{DRIVE_TYPES_TO_SKIP} = $drive_types_to_skip;
    $self->{AGGRS_TO_SKIP} = $aggrs_to_skip;
    $self->{bridges} = \@bridges;

    # Drives to skip format:
    # <drive>,<drive>...<drive>
    # Drives are serial numbers
    my @drives = split (',', $drives_to_skip);
    foreach my $drive (@drives) {
        # Format drive_name(drive_serial)
        if ($drive =~ /(\S+)\((\S+)\)/) {
           push (@{$self->{ignoring_vals}->{drive_names}},$1);
           push (@{$self->{ignoring_vals}->{drive_serial}},$2);
        }
    }

    # Drive types to skip format:
    # <drivetype>,<drivetype>,...<drivetype>
    # Drives are serial numbers
    my @drive_types = split (',', $drive_types_to_skip);
    if (scalar (@drive_types)) {
            $self->{ignoring_vals}->{drive_types} = \@drive_types;
    }

    # aggrs to skip format:
    # fas6070-rtp09;<aggr>,<aggr>...<aggr>::fas6070-rtp10;<aggr>,...<aggr>
    my @ignr_controllers = split ('::', $aggrs_to_skip);
    foreach my $controller (@ignr_controllers) {
        my @controller_aggrs = split (';', $controller);
        my @aggrs = split (',', $controller_aggrs[1]);
        $self->{ignoring_vals}->{aggrs}->{$controller_aggrs[0]} = \@aggrs;
    }

    # Parse out the VF/VSAN information and the module information
    # Possible combinations:
    #  brcd300-1(5,7)
    #  brcd300-2(5,9)[2]
    #  brcd300-3
    #  brcd300-4(3,4)
    #  cisco9710-fcs40[1]
    #  cisco9710-fcs40(10,20)[1]
    foreach my $switch (@fabric1_switches_raw) {
        if ($switch =~ /([\w-]+)(\(([\d,]*)\))?\s*(\[(\d)\])?/) {
            $self->sanlog(-msg => "Found in match: |$1|$2|$3|$4|$5|");
            my $switch_name = $1;
            push (@fabric1_switches, $switch_name);
            @{$self->{vfs}->{$switch_name}} = split(/,/, $3) if (not $3 eq "");
            $self->{modules}->{fab1}->{$switch_name} = $5 if (not $5 eq "");
        } else {
            &SanUtil::nate_log('warn', "Unable to match. Assuming you entered switches without VSANs/VFs or Modules");
            push (@fabric1_switches, $switch);
        }
    }
    # Do the same thing for fabric 2
    foreach my $switch (@fabric2_switches_raw) {
        if ($switch =~ /([\w-]+)(\(([\d,]*)\))?\s*(\[(\d)\])?/) {
            $self->sanlog(-msg => "Found in match: |$1|$2|$3|$4|$5|");
            my $switch_name = $1;
            push (@fabric2_switches, $switch_name);
            push(@{$self->{vfs}->{$switch_name}}, split(/,/, $3)) if (not $3 eq "");
            $self->{modules}->{fab2}->{$switch_name} = $5 if (not $5 eq "");
        } else {
            &SanUtil::nate_log('warn', "Unable to match. Assuming you entered switches without VSANs/VFs or Modules");
            push (@fabric2_switches, $switch);
        }
    }

    $self->{fabric1_switches_raw} = \@fabric1_switches_raw;
    $self->{fabric2_switches_raw} = \@fabric2_switches_raw;
    $self->{fabric1_switches} = \@fabric1_switches;
    $self->{fabric2_switches} = \@fabric2_switches;
    $self->{fab1_fcvi_isl} = \@fab1_fcvi_isl;
    $self->{fab2_fcvi_isl} = \@fab2_fcvi_isl;

    # Copy the Metrocluster fabrics into the global obj
    # Each fabric consists of two switches
    @{$self->{fabrics}->{fab1}->{switches}} = @fabric1_switches;
    @{$self->{fabrics}->{fab2}->{switches}} = @fabric2_switches;

    @{$self->{fabrics}->{fab1}->{fcvi_isl}} = @fab1_fcvi_isl;
    @{$self->{fabrics}->{fab2}->{fcvi_isl}} = @fab2_fcvi_isl;

    $self->{controller_hash} = {};
    $self->{switch_hash} = {};
    $self->{storage_hash} = {};

    return $self;
}

################################################################################
# Method:    get_controller_info
# Objective: Collect information about controllers and save the information in
#            a hash
# Details:   Parse every line in the output of the command sysconfig -v. If the
#            line matches FC Host Adapter or FCVI Host Adapter then push it into
#            an array of hashes under a key, which is the adapter number. For
#            example, this is the line you're looking for:
#            slot 3: FC Host Adapter 3a (QLogic 2432 rev. 3, N-port, <UP>)
#            Extract 3a(adapter) and store the remaining lines under this key,
#            till you encounter another line which has FC Host Adapter.
#            This loop populates the data structure $hash->{$key}->{raw} which
#            might look like:
#            $hash -> 3a ->   Board name:        QLE2464
#                     |   |   Serial Number:     GFC0719P68276
#                     |   |   Firmware rev:      4.0.24
#                     |   |   Host Port Id:      0x10e00
#                     |   |   FC Node Name:      2:100:001b32:1714ee
#                     |   |   SFF Vendor:        FINISAR CORP.
#                     |   |   SFF Part Number:   FTLF8524E2KNL
#                     |   |   SFF Serial Number: PBB08V4
#                     |   |   SFF Capabilities:  1, 2 or 4 Gbit
#                     |   |   Link Data Rate:    4 Gbit
#                     |   |   Switch Port:       brcd5000-rtp03:14
#                               ....
#            Also,searches for any line having Model name and stores the model
#            name of the controller in an array
#
################################################################################
sub get_controller_info {
    my ($self) = @_;
    my (@controllers, @cmd_output);
    my ($key, $hash, $controller, $line, $controller_found, @storage_show_output,
        @disk_show_output, @cf_monitor_all, @cf_monitor, @disk_list_info);
    my ($output, $status, $cmd);

    foreach $controller (@{$self->{controllers}}) {
        # Store the command that needs to be executed
        $cmd = "sysconfig -v";

        # Use system_zapi to get the output of the command and store it in @cmd_output
        ($output, $status, @cmd_output) = &SanUtil::system_zapi($controller, $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
        }

        my $searching_for_card = 0;
        my $found_disks = 0;

        ## Initialize all values for FC and FCVI cards.
        my ($card, $card_type, $node_name, $port_name, $switch_port) = "";
        my $adapter = "";

        foreach $line (@cmd_output) {

            # If we found a FC or FCVI Card we need to get the data for it.
            if ($searching_for_card) {
                # FC Node Name:      2:000:000e1e:10f111
                if ($line =~ /\s*FC Node Name:?\s+:?\s?([\w:]+)/) {
                    $key = $1;
                    ## Remove the colon from the string.
                    $key =~ tr/://d;
                    $node_name = $key;
                    $hash->{$controller}->{adapters}->{$card_type}->{$adapter}->{wwnn} = $node_name;
                    ## FC Port Name:      2:100:000e1e:10f111
                } elsif ($line =~ /\s*FC Port Name:?\s+:?\s?([\w:]+)/) {
                    $key = $1;
                    ## Remove the colon from the string.
                    $key =~ tr/://d;
                    $port_name = $key;
                    $hash->{$controller}->{adapters}->{$card_type}->{$adapter}->{wwpn} = $port_name;
                    ## Switch Port:       cisco9710-fcs40:1-3
                } elsif ($line =~ /\s*Switch Port:?\s+:?\s?([\w\-:]+)/) {
                    $switch_port = $1;
                    $hash->{$controller}->{adapters}->{$card_type}->{$adapter}->{switch_port} = $switch_port;
                ##      cisco9710-fcs41:1-7.173L15     : NETAPP   X410_HVIPC288A15 NA02 272.0GB (562884296 520B/sect)
                } elsif ($line =~ /^\s*([-\w:\.]+)\s+:\s*[\w]+\s+([\w]+)\s+[\w]+\s+[\.\w]+\s+\S+\s+\S+/) {
                    $hash->{$controller}->{adapters}->{$card_type}->{$adapter}->{disks}->{models}->{$2}->{firmwares}->{$3} += 1;
                    $hash->{$controller}->{adapters}->{$card_type}->{$adapter}->{num_disks} += 1;
                } elsif ($line =~ /FC-to-SAS Bridge:/) {
                    $found_disks = 1;
                }

                ## If we found all the information that we're looking for then we can add it to the hash
                ##  and reset all the values back to an empty string for the next card that we find.
                if ($node_name ne "" and $port_name ne "" and $switch_port ne "" and $card_type eq 'fcvi') {
                    ($card, $card_type, $node_name, $port_name, $switch_port) = "";
                    $searching_for_card = 0;
                    next;
                ## If we haven't found everything then we'll move on to the next line.
                } elsif ($node_name ne "" and $port_name ne "" and $switch_port ne "" and $card_type eq 'fc' and $found_disks) {
                    ($card, $card_type, $node_name, $port_name, $switch_port) = "";
                    $found_disks = 0;
                    $searching_for_card = 0;
                    next;
                } else {
                    next;
                }
            }

            # look for ontap version in this line
            # NetApp Release 7.3RC2: Wed Jun 25 13:18:42 PDT 2008
            # Eg: Ontap version is 7.3RC2
            if ($line =~ /^\s*\S+\s+Release\s+(\S+):.+/) {
                $self->sanlog(-msg => "Found Release");
                $hash->{$controller}->{ontap_version} = $1;
            }

            # look for system id, system name, partner id, partner name on this line
            # System ID: 0118051496 (fas3070-rtp12); partner ID: 0118051524 (fas3070-rtp11)
            elsif ($line =~ /^\s*System ID:\s+(\S+)\s+\((\S+)\)[\S+\s+]\s*partner ID:\s+(\S+)\s+\((\S+)\)/) {
                $self->sanlog(-msg => "Found System ID");
                $hash->{$controller}->{system_id}    = $1;
                $hash->{$controller}->{system_name}  = $2;
                $hash->{$controller}->{partner_id}   = $3;
                $hash->{$controller}->{partner_name} = $4;
            }

            #  slot 3: FC Host Adapter 3a (QLogic 2432 rev. 3, N-port, <UP>)
            elsif ($line =~ /slot\s+\d+:\s+FC Host Adapter\s+([\w]+)\s+\([\w\s\.\,-]+<(\w+)/) {
                # update the key with the FC adapter Eg: 3a
                $adapter = $1;
                $status = $2;
                if ($status =~ /up|online/i) {
                    $card_type = 'fc';
                    $hash->{$controller}->{adapters}->{$card_type}->{$adapter}->{raw} = $line;
                    $searching_for_card = 1;
                    next;
                } else {
                    &SanUtil::nate_log('warn', "FC Host Adapter $adapter skipped. Status=$status");
                }
            }
            ##  slot 6: FCVI Host Adapter 6a (QLogic 8324(2672) rev. 2, F-port, <UP>)
            elsif ($line =~ /slot\s+\d+:\s+FCVI Host Adapter\s+([\w]+)\s+\([\(\)\w\s\.\,-]+<(\w+)/) {
                $adapter = $1;
                $status = $2;
                if ($status =~ /up|online/i) {
                    $card_type = 'fcvi';
                    $hash->{$controller}->{adapters}->{$card_type}->{$adapter}->{raw} = $line;
                    $searching_for_card = 1;
                    next;
                } else {
                    &SanUtil::nate_log('warn', "FCVI Host Adapter $adapter skipped. Status=$status");
                }
            }

            elsif ($line =~ /bypassed/ or $line =~ /Failed/i or $line =~ /\s0\.0GB/) {
                push(@{$hash->{$controller}->{adapters}->{$adapter}->{warnings}}, $line);
                if ($line =~ /\s*(\S+)\s*:\s*\S+\s+\S+\s+\S+/) {
                    push(@{$hash->{$controller}->{adapters}->{$adapter}->{bad_disk}}, $1);
                }
            }
        }

        # call storage aggr status -r and store the output
        $cmd = "aggr status -r";
        my ($output, $status, @aggr_status_output) = &SanUtil::system_zapi($controller, $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
        }
        $hash->{$controller}->{aggr_status_r} = \@aggr_status_output;


        # call storage show disk -p and store the output
        $cmd = "storage show disk -p";
        ($output, $status, @storage_show_output) = &SanUtil::system_zapi($controller, $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
        }
        $hash->{$controller}->{storage_show_disks} = \@storage_show_output;

        # call disk show -v and store the output array
        $cmd = "disk show -v";
        ($output, $status, @disk_show_output) = &SanUtil::system_zapi($controller, $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
        }
        $hash->{$controller}->{disks_show} = \@disk_show_output;
        $hash->{$controller}->{total_num_disks} = scalar(grep(/$controller/, @disk_show_output));

        # call priv set test and store the output array
        $cmd = "priv set diag";
        ($output, $status, @cmd_output) = &SanUtil::system_zapi($controller, $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
        }

        # call cf monitor all and store the output array
        $cmd = "cf monitor all";
        ($output, $status, @cf_monitor_all) = &SanUtil::system_zapi($controller, $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
        }
        $hash->{$controller}->{cf_monitor_all} = \@cf_monitor_all;

        # call cf monitor and store the output array
        $cmd = "cf monitor";
        ($output, $status, @cf_monitor) = &SanUtil::system_zapi($controller, $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
        }
        $hash->{$controller}->{cf_monitor} = \@cf_monitor;


    }
    return $hash;
}

################################################################################
# Method:    get_disk_info
# Objective: Collect information about all the disks, aggrs, and volumes and save
#            the information in a hash.
# Details:   Create a list of unique disks collected from various disk ZAPIs.
#            The purpose of this function is just to collect data and no other
#            checks will be done during this. The info we collect here is from
#            disk-list-info, disk-sanown-list-info, and aggr-list-info. The two
#            disk ZAPIs are referenced against each other to make sure we have all
#            the disks the systems sees. The disk-list-info ZAPI contains only the
#            disks that are acknowledged by RAID so might now include all the disks
#            that a system sees. disk-sanown-list-info ZAPI should include all the
#            disks that a system sees.
#
################################################################################
sub get_disk_info {
    my ($self) = @_;

    my ($info, $temp_disk_hash);
    my $keep = 0;

    # collect disk related info.
    foreach my $controller (keys %{$self->{controller_hash}}) {
        # Get the locally attached switches
        my $cmd = "fcadmin link_state";
        my ($output, $status, @fcadmin_out) = &SanUtil::system_zapi($controller, $cmd, "diag");
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err(-msg => "SanMCUtil:get_disk_info() : Failed to execute" .
                                    " system-cli ZAPI ($cmd)");
            return {};
        }

        # Sample output:
        # Link State for adapter channel 0a: UP - N-port --> Port 1 on switch 'metro8-fab2' (100000051e0f502f)
        # Link State for adapter channel 0b: OFFLINE (physical)
        # Link State for adapter channel 0c: UP - N-port --> Port 1 on switch 'metro8-fab1' (100000051e56582a)
        # Link State for adapter channel 0d: OFFLINE (physical)
        foreach my $channel (@fcadmin_out) {
            if ($channel =~ /\s*Link State for adapter channel (\S+): UP - \S+ --> Port (\S+) on switch '(\S+)'/) {
                $info->{adapters}->{$controller}->{$1}->{port} = $2;
                $info->{adapters}->{$controller}->{$1}->{switch} = $3;
                if (not exists $info->{local_switches}->{$controller}->{$3}) {
                    $info->{local_switches}->{$controller}->{$3}->{$1} = $2;
                }
            }
        }

        foreach my $controller (keys %{$info->{local_switches}}) {
            $self->sanlog( -msg => "Controller $controller local switches: " .
                                    join (',', keys %{$info->{local_switches}->{$controller}}));
        }

        # Get disk information from the zapi.
        $cmd = "storage-shelf-list-info";
        # Use zapi to get the output of the command and store it in $output
        ($output, $status) = &SanUtil::cmds_driver_zapi($controller,
                                                           $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute ZAPI ($cmd)");
            return {};
        }

        foreach my $shelf (@{$output->{'shelf-list'}->{'shelf-info'}}) {
            my $module = $shelf->{'module'};
            # Get the data for DS14 disks
            foreach my $port (@{$shelf->{'port-hub-list'}->{'port-hub-info'}}) {
                if ($port->{'port-id'} !~ /in|out/ and
                    $port->{'port-state'} !~ /empty/i) {
                    my $disk = $shelf->{'channel-name'}. "." . $port->{'port-id'};
                    if (exists $info->{data_collection}->{shelf_info}->{$disk}) {
                        foreach my $key (keys %{$port}) {
                            # Make sure that this value matches in both places others don't matter
                            if ($key =~ /port-state/i and
                                $port->{$key} ne $info->{data_collection}->{shelf_info}->{$disk}->{$key}) {
                                $self->_set_err("$key for $disk seen differently " .
                                    " on different controllers ($port->{$key}) " .
                                    " ($info->{data_collection}->{shelf_info}->{$disk}->{$key}).");
                                return {};
                            }
                        }
                    } else {
                        %{$info->{data_collection}->{shelf_info}->{$disk}} = %{$port};
                    }

                    my $ses_hash = $info->{data_collection}->{shelf_info}->{$disk};
                    $ses_hash->{'disk-name'} = $disk;
                    $ses_hash->{'shelf-name'} = $shelf->{'shelf-name'};
                    $ses_hash->{'shelf-state'} = $shelf->{'shelf-state'};
                    $ses_hash->{'module-info'}->{module} = $module;
                    $ses_hash->{'module-info'}->{state} = $shelf->{'module-state'};
                    $ses_hash->{'module-info'}->{'firmware-rev-A'} = $shelf->{'firmware-rev-A'};
                    $ses_hash->{'module-info'}->{'firmware-rev-B'} = $shelf->{'firmware-rev-B'};

                    # Initialize the value of see if this disk will be seen by the disk
                    $ses_hash->{acked} = $ClientUtil::FALSE;
                }
            }

            # Get the data for FibreBridge attached disks
            # Differences are the 'phy-expander-list', 'phy_expander_info' and the way $disk is calculated.
            foreach my $port (@{$shelf->{'phy-expander-list'}->{'phy-expander-info'}}) {
                if ($port->{'port-id'} !~ /in|out/ and
                    $port->{'port-state'} !~ /empty/i and
                    $port->{'port-id'} !~ /sqr\d|cir\d|sil\d/){
                    my $disk = $shelf->{'channel-name'}. "." . $port->{'port-id'};
                    if (exists $info->{data_collection}->{shelf_info}->{$disk}) {
                        foreach my $key (keys %{$port}) {
                            ## Make sure that this value matches in both places others don't matter
                            if ($key =~ /port-state/i and
                                $port->{$key} ne $info->{data_collection}->{shelf_info}->{$disk}->{$key}) {
                                $self->_set_err("$key for $disk seen differently " .
                                    " on different controllers ($port->{$key}) " .
                                    " ($info->{data_collection}->{shelf_info}->{$disk}->{$key}).");
                                return {};
                            }
                        }
                    } else {
                        %{$info->{data_collection}->{shelf_info}->{$disk}} = %{$port};
                    }

                    my $ses_hash = $info->{data_collection}->{shelf_info}->{$disk};
                    $ses_hash->{'disk-name'} = $disk;
                    $ses_hash->{'shelf-name'} = $shelf->{'shelf-name'};
                    $ses_hash->{'shelf-state'} = $shelf->{'shelf-state'};
                    $ses_hash->{'module-info'}->{module} = $module;
                    $ses_hash->{'module-info'}->{state} = $shelf->{'module-state'};
                    $ses_hash->{'module-info'}->{'firmware-rev-A'} = $shelf->{'firmware-rev-A'};
                    $ses_hash->{'module-info'}->{'firmware-rev-B'} = $shelf->{'firmware-rev-B'};

                    ## Initialize the value of see if this disk will be seen by the disk
                    $ses_hash->{acked} = $ClientUtil::FALSE;
                }
            }
        }

        # Get disk information from the zapi.
        $cmd = "disk-list-info";

        # Use zapi to get the output of the command and store it in $output
        ($output, $status) = &SanUtil::cmds_driver_zapi($controller,
                                                           $cmd,
                                                           "diag");
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute ZAPI ($cmd)");
            return {};
        }

        # We'll never get just one disk in the system so let's just put the whole hash in.
        %{$info->{data_collection}->{$controller}->{disk_list_info}} =
                %{$output->{'disk-details'}->{'disk-detail-info'}};

        # Get the sanown disk info from zapi
        if ($self->{HW_BASED_OWNERSHIP} == $ClientUtil::FALSE) {
            $cmd = "disk-sanown-list-info";
            # Use zapi to get the output of the command and store it in $output
            ($output, $status) = &SanUtil::cmds_driver_zapi($controller,
                                                            $cmd,
                                                            "diag");
            if ($status == SanUtil::ZAPI_FAILURE) {
                $self->_set_err("Failed to execute ZAPI ($cmd)");
                return {};
            }

            %{$info->{data_collection}->{$controller}->{disk_owner_detail_info}} =
                %{$output->{'disk-sanown-details'}->{'disk-sanown-detail-info'}};
        }

        # Get the aggregate info from zapi
        $cmd = "aggr-list-info";
        # Use zapi to get the output of the command and store it in $output
        ($output, $status) = &SanUtil::cmds_driver_zapi($controller,
                                                        $cmd,
                                                        "diag");
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute ZAPI ($cmd)");
            return {};
        }

        # We need to find out see if we have just aggr or more that one.
        if (exists($output->{'aggregates'}->{'aggr-info'}->{'name'})) {
            my $aggr_name = $output->{'aggregates'}->{'aggr-info'}->{'name'};
            %{$info->{aggrs}->{$controller}->{$aggr_name}} =
                    %{$output->{'aggregates'}->{'aggr-info'}};
        } else {
            %{$info->{aggrs}->{$controller}} = %{$output->{'aggregates'}->{'aggr-info'}};
        }

        # collect volume info for all vols inside aggrs.
        foreach my $aggr (keys %{$info->{aggrs}->{$controller}}) {
            my $aggr_hash = $info->{aggrs}->{$controller}->{$aggr};
            # if it is an aggregate and not a trad vol move forward
            if ($aggr_hash->{'type'} =~ /aggr/) {

                my @volumes = ();
                # If there is only one vol then the format will be name=> volname
                # otherwise, each volume will be it's own key.
                if ($aggr_hash->{'volume-count'} == 0) {
                    # No volumes nothing to do
                } elsif ($aggr_hash->{'volume-count'} == 1) {
                    push (@volumes,
                          $aggr_hash->{'volumes'}->{'contained-volume-info'}->{'name'});
                } else {
                    foreach my $vol (keys %{$aggr_hash->{'volumes'}->{'contained-volume-info'}}) {
                        push (@volumes,$vol);
                    }
                }

                # Clear out the current volumes hash
                delete $aggr_hash->{'volumes'}->{'contained-volume-info'};

                # In cases where we have large number of volumes we need
                # to sleep after a volume
                my $volumes_called = 0;
                foreach my $vol (@volumes) {
                    # Call the zapi and collect the volume info
                    $cmd = "volume-list-info";
                    my $args;
                    $args->{-volume} = $vol;
                    ($output, $status) = &SanUtil::cmds_driver_zapi($controller,
                                                                    $cmd,
                                                                    "diag","",
                                                                    $args);
                    if ($status == SanUtil::ZAPI_FAILURE) {
                        $self->_set_err("Failed to execute ZAPI ($cmd)");
                        return {};
                    }
                    # Save the data in the hash
                    %{$aggr_hash->{'volumes'}->{$vol}} =
                        %{$output->{'volumes'}->{'volume-info'}};

                    $volumes_called++;

                    if (($volumes_called % 10) == 0) {
                        &SanUtil::nate_sleep(1);
                    }
                }
            }
        }
    }

    # Lets put all the disks in a common hash
    foreach my $controller (keys %{$self->{controller_hash}}) {

        # Get ownership data and add it to this $info->{disks}
        my $owner_detail_hash;
        if ($self->{HW_BASED_OWNERSHIP} == $ClientUtil::TRUE) {
            $owner_detail_hash = {};
        } else {
            $owner_detail_hash =
                $info->{data_collection}->{$controller}->{disk_owner_detail_info};
        }

        foreach my $sown_disk (keys %{$owner_detail_hash}) {

            # Keys the disks by serial
            my $disk_serial = $owner_detail_hash->{$sown_disk}->{'serial-no'};

            if (exists $info->{disks}->{$disk_serial}){
                # We already have it. Let's move on.
            } else {

                # We didn't have it so let's store it.
                %{$info->{disks}->{$disk_serial}} = %{$owner_detail_hash->{$sown_disk}};

                # Since we a storing by serial we need to store it's primary name.
                $info->{disks}->{$disk_serial}->{'primary-name'} = $sown_disk;

                # Raid hasn't seen this disk yet.
                $info->{disks}->{$disk_serial}->{'acknowledged-by-raid'}->{$controller} = $ClientUtil::FALSE;
                $info->{disks}->{$disk_serial}->{'percieved-owner'} = $controller;

            }

            # Acknowledge that sanown sees it.
            $info->{disks}->{$disk_serial}->{'acknowledged-by-sanown'}->{$controller} = $ClientUtil::TRUE;
        }

        # Now look through raid data.
        my $disk_detail_hash = $info->{data_collection}->{$controller}->{disk_list_info};
        foreach my $disk_name (keys %{$disk_detail_hash}) {

            my $disk_serial = $disk_detail_hash->{$disk_name}->{'serial-number'};
            my $data_action = DO_NOTHING;

            if ($disk_detail_hash->{$disk_name}->{'raid-state'} !~ /partner/) {

                # Making serial the key so let's store the primary port name as well
                $info->{disks}->{$disk_serial}->{'primary-name'} = $disk_name;

                #HW based system need some additional infor added
                if ($self->{HW_BASED_OWNERSHIP} == $ClientUtil::TRUE) {
                    $info->{disks}->{$disk_serial}->{'owner'} = $controller;
                }

                $info->{disks}->{$disk_serial}->{'percieved-owner'} = $controller;
                # Failed disk so let's mark it that way.
                if ($disk_detail_hash->{$disk_name}->{'raid-state'} =~ /broken/) {
                    $info->{disks}->{$disk_serial}->{'is-failed'} = $ClientUtil::TRUE;
                }

                # What are we going to do with the data
                if (exists $info->{disks}->{$disk_serial}) {
                    $data_action = MERGE_DATA;
                } else {
                    $data_action = COPY_DATA;
                }

            } else {
                $data_action = DO_NOTHING;
                # Don't add partner disks now, we'll get them from partner
            }

            # This makes sure that the disks is acknowledged by raid and not just sanown.
            $info->{disks}->{$disk_serial}->{'acknowledged-by-raid'}->{$controller} = $ClientUtil::TRUE;

            # Actually copy or merge the data
            if ($data_action == DO_NOTHING) {
                # Do what it says
            } elsif ($data_action == MERGE_DATA) {

                # We are good save this info. If a key of same type exists we'll
                # overwrite it.
                while (my ($key, $val) = each (%{$disk_detail_hash->{$disk_name}})) {
                    $info->{disks}->{$disk_serial}->{$key} = $val;
                }
            } else {
                # If we get here, the data doesn't exists so let plop the whole
                # hash in here.
                %{$info->{disks}->{$disk_serial}} = %{$disk_detail_hash->{$disk_name}};
            }
        }
    }

    # Go through the list and get the secondary name and disk UID
    # if we can from storage show
    foreach my $controller (keys %{$self->{controller_hash}}) {
        # Get ownership data and add it to this $info->{disks}
        my $owner_detail_hash;
        if ($self->{HW_BASED_OWNERSHIP} == $ClientUtil::TRUE) {
            $owner_detail_hash = {};
        } else {
            $owner_detail_hash =
                $info->{data_collection}->{$controller}->{disk_owner_detail_info};
        }

        foreach my $sown_disk (keys %{$owner_detail_hash}) {
            # Keys the disks by serial
            my $disk_serial = $owner_detail_hash->{$sown_disk}->{'serial-no'};

            # If it not owned by one of the two controllers, skip it
            if (!(grep /$info->{disks}->{$disk_serial}->{'owner'}/, (keys %{$self->{controller_hash}}))) {
                next;
            }

            if ($info->{disks}->{$disk_serial}->{'is-failed'} =~ /true/i or
                   $info->{disks}->{$disk_serial}->{'acknowledged-by-raid'}->{$controller} != $ClientUtil::TRUE) {
                $info->{disks}->{$disk_serial}->{'is-failed'} = $ClientUtil::TRUE;

                # If it is failed by sanown we may need to collect more data.
                my @keys = ("secondary-name", "disk-uid");
                my $disk_info_hash = $self->_get_storage_show_disk_info($controller,
                                                                        $sown_disk,
                                                                        @keys);
                foreach my $key (@keys) {
                    $info->{disks}->{$disk_serial}->{$key} = $disk_info_hash->{$key};
                }
            }
        }
    }

    # Looking at the shelf info and merge it with the disk data that we have
    foreach my $disk (keys %{$info->{disks}}) {

        # If the disk is connected to Armadillo, check to make sure
        # it is logged into the storage layer.
        if ($info->{disks}->{$disk}->{type} !~ /^FC/){

            $info->{disks}->{$disk}->{'acknowledged-by-storage'} = 0;
            my $primary_name = $info->{disks}->{$disk}->{'primary-name'};
            my $secondary_name = "";
            my $sec_name = "";
            my $bay = $info->{disks}->{$disk}->{'bay'};
            my @channel_name = split (/\./, $primary_name);
            my $pri_name = $channel_name[0] . "." . $bay;

            # Make sure the primary name exists in the shelf info.
            if (exists $info->{data_collection}->{shelf_info}->{$pri_name}) {
                my $ses_hash = $info->{data_collection}->{shelf_info}->{$pri_name};
                my $shelf = $ses_hash->{'shelf-name'};
                $ses_hash->{acked} = $ClientUtil::TRUE;
                $info->{disks}->{$disk}->{'acknowledged-by-storage'} += 1;
                $info->{disks}->{$disk}->{ses_info}->{$ses_hash->{'module-info'}->{module}} = $ses_hash;
            }

            if (defined $info->{disks}->{$disk}->{'secondary-name'}) {
                $secondary_name = $info->{disks}->{$disk}->{'secondary-name'};
                my @channel_name = split (/\./, $secondary_name);
                $sec_name = $channel_name[0] . "." . $bay;
            }

            # Make sure the secondary name exists in the shelf info.
            if ($secondary_name ne "" and
                exists $info->{data_collection}->{shelf_info}->{$sec_name}) {
                my $ses_hash = $info->{data_collection}->{shelf_info}->{$sec_name};
                my $shelf = $ses_hash->{'shelf-name'};
                $ses_hash->{acked} = $ClientUtil::TRUE;
                $info->{disks}->{$disk}->{'acknowledged-by-storage'} += 1;
                $info->{disks}->{$disk}->{ses_info}->{$ses_hash->{'module-info'}->{module}} = $ses_hash;
            }
        }

        # The disk is a FC disk.
        else {
            my $pri_name = $info->{disks}->{$disk}->{'primary-name'};
            my $sec_name = "";
            $info->{disks}->{$disk}->{'acknowledged-by-storage'} = 0;

            #check the secondary name
            if (defined $info->{disks}->{$disk}->{'secondary-name'}) {
                $sec_name = $info->{disks}->{$disk}->{'secondary-name'};
            }

            # Lets see if this disk exists in the disks names generated from the list.
            if (exists $info->{data_collection}->{shelf_info}->{$pri_name}) {
                my $ses_hash = $info->{data_collection}->{shelf_info}->{$pri_name};
                my $shelf = $ses_hash->{'shelf-name'};
                $ses_hash->{acked} = $ClientUtil::TRUE;
                $info->{disks}->{$disk}->{'acknowledged-by-storage'} += 1;
                $info->{disks}->{$disk}->{ses_info}->{$ses_hash->{'module-info'}->{module}} =
                    $ses_hash;
            }

            # Does the secondary path of the disk also exists?
            if ($sec_name ne "" and
                exists $info->{data_collection}->{shelf_info}->{$sec_name}) {
                my $ses_hash = $info->{data_collection}->{shelf_info}->{$sec_name};
                my $shelf = $ses_hash->{'shelf-name'};
                $ses_hash->{acked} = $ClientUtil::TRUE;
                $info->{disks}->{$disk}->{'acknowledged-by-storage'} += 1;
                $info->{disks}->{$disk}->{ses_info}->{$ses_hash->{'module-info'}->{module}} =
                    $ses_hash;
            }
        }
    }

    # Make sure all the disks from shelf are acked. If they aren't acked, it means that
    # the disk is present on the shelf but is not seen by ONTAP.
    foreach my $ses_disk (keys %{$info->{data_collection}->{shelf_info}}) {
        if ($info->{data_collection}->{shelf_info}->{$ses_disk}->{acked} != $ClientUtil::TRUE) {
            # See if we are supposed to skip this drive, otherwise add to list
            if (grep (/$ses_disk/i, @{$self->{ignoring_vals}->{drive_names}}) != 0) {
                # Let us ignore this drive since we are told to do so
                &SanUtil::nate_log('-warn', "Ignoring disk $ses_disk.");
            } else {
                %{$info->{unacked_disks}->{$ses_disk}} =
                        %{$info->{data_collection}->{shelf_info}->{$ses_disk}};
            }
        }
    }

    # Convert all the WWNN in a format that match the switch ns output
    # Disk UUID output:
    #   UID: 20000000:8773661D:00000000:00000000:00000000:00000000:00000000:00000000:00000000:0000000
    # nsshow output:
    #   NL   0109cd;      3;21:00:00:00:87:73:66:1d;20:00:00:00:87:73:66:1d; na
    #   FC4s: FCP [NETAPP  X272_HJURE073F10NA14]
    #   Fabric Port Name: 20:09:00:05:1e:56:5a:52
    #   Permanent Port Name: 21:00:00:00:87:73:66:1d
    #
    # Setting the common format to 200000008773661d

    foreach my $disk (keys %{$info->{disks}}) {
        if ($info->{disks}->{$disk}->{'disk-uid'}) {
            my @uids = split (/:/, $info->{disks}->{$disk}->{'disk-uid'});
            $info->{disks}->{$disk}->{'wwnn'} = "$uids[0]" . "$uids[1]";
        }
    }

    # Collect the pool info and seperates it into switch and ports it's connected to
    # Format: $info->{pool_info}->{<controller>}->{<pool>}->{<switch>}->{ports}
    # This will only work for fabric metrocluster
    foreach my $disk (keys %{$info->{disks}}) {
        if (exists ($info->{disks}->{$disk}->{'owner'}) and
            exists ($info->{disks}->{$disk}->{'pool'})) {
            my $owner = $info->{disks}->{$disk}->{'owner'};
            my $pool = $info->{disks}->{$disk}->{'pool'};

            # chomp out the pool info it is in format Pool0
            if ($pool =~ /^Pool/) {
                $pool =~ s/^Pool//;
                $info->{disks}->{$disk}->{'pool'} = $pool;
            }

            if (exists ($info->{disks}->{$disk}->{'primary-name'})) {
                my @switch_name = split (/:/, $info->{disks}->{$disk}->{'primary-name'});
                my @port_num = split (/\./, $switch_name[1]);
                my $switch = $switch_name[0];
                my $port = $port_num[0];

                #See if the port already exists in the array
                if ((exists $info->{pool_info}->{$owner}->{$pool}->{$switch}->{$port}) and
                    (grep { $_ eq $disk} @{$info->{pool_info}->{$owner}->{$pool}->{$switch}->{$port}})) {
                    # We already have that disk/port combination
                } else {
                    push (@{$info->{pool_info}->{$owner}->{$pool}->{$switch}->{$port}}, $disk);
                }
            }

            if (exists ($info->{disks}->{$disk}->{'secondary-name'})) {
                my @switch_name = split (/:/, $info->{disks}->{$disk}->{'secondary-name'});
                my @port_num = split (/\./, $switch_name[1]);
                my $switch = $switch_name[0];
                my $port = $port_num[0];

                #See if the port already exists in the array
                if ((exists $info->{pool_info}->{$owner}->{$pool}->{$switch}->{$port}) and
                    (grep { $_ eq $disk} @{$info->{pool_info}->{$owner}->{$pool}->{$switch}->{$port}})) {
                    # We already have that disk/port combination
                } else {
                    push (@{$info->{pool_info}->{$owner}->{$pool}->{$switch}->{$port}}, $disk);
                }
            }
        }
    }

    $self->sanlog(-msg_level => 6,
                  -msg => "Disk info:\n" . Dumper $info->{disks});
    return $info;
}

###############################################################################
# Method:       collect_controller_info
# Objective:    Collects the switch information and saves it in a hash
# Details:      calls get_controller_info() and saves the hash that is returned
#               from the function back into the object.
###############################################################################
sub collect_controller_info {
    my ($self) = @_;

    # call get_controller_info() method to get the hash and store it in the object
    my $hash = {};
    $hash = $self->get_controller_info();
    if ($self->error){
        return $ClientUtil::FALSE;
    }
    $self->{controller_hash} = $hash;

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:       collect_disk_info
# Objective:    Collects the disk information and saves it in a hash
# Details:      calls get_disk_info() and saves the hash that is returned
#               from the function back into the object. Must be run after
#               collect_controller_info because the disk names are collected
#               there and must be the same object that was passed into
#
#               collect_controller_info.
###############################################################################
sub collect_disk_info {
    my ($self) = @_;

    # call get_disk_info() method to get the hash and store it in the object
    my $hash = {};
    $hash = $self->get_disk_info();
    if ($self->error){
        return $ClientUtil::FALSE;
    }
    $self->{storage_hash} = $hash;

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:       collect_bridge_info
# Objective:    Collects the bridge information and saves it in a hash
# Details:      calls get_bridge_info() and saves the hash that is returned
#               from the function back into the object.
#
###############################################################################
sub collect_bridge_info {
    my ($self) = @_;

    # call get_disk_info() method to get the hash and store it in the object
    my $hash = {};
    $hash = $self->get_bridge_info();
    if ($self->error){
        return $ClientUtil::FALSE;
    }

    $self->{bridge_hash} = $hash;

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:       get_bridge_info
# Objective:    Collects the bridge information as a hash and returns it.
# Details:      Creates bridge objects for bridges listed in the @bridges
#               array. Gets necessary information about them.
#
###############################################################################
sub get_bridge_info {
    my ($self) = @_;

    my (@bridge_array, $bridge, $arm_obj, $ref_disk_logins, $output);
    my $bridge_hash = {};

    @bridge_array = @{$self->{bridges}};

    foreach $bridge (@bridge_array) {

        # Create the armadillo object
        $arm_obj = Armadillo->new(-bridge => $bridge,
                                  -conn_type => "THARNHOST");
        if ($arm_obj->error) {
            my $msg = join($arm_obj->errmsg(), "\n");
            $self->_set_err("Failed to create Armadillo object for $bridge " .
                            "Reason: $msg");
            return {};
        }
        $self->sanlog(-msg => "Succesfully created Armadillo object $bridge");

        $ref_disk_logins = $arm_obj->get_disk_logins();

        $self->sanlog(-msg => "Retrieved disks from $bridge");

        $bridge_hash->{$bridge}->{disks} = $ref_disk_logins;

        $output = $arm_obj->display_info();
        $bridge_hash->{$bridge}->{info} = $output;
    }

    return $bridge_hash;
}

################################################################################
# Method:       get_switch_info
# Objective:    Collects the switch information and saves it in a hash
# Details:      Creates a SanSwitch object and a Port manager object. E- ports
#               are obtained by scanning through the 'type' field of
#               get_nameserver().Version, model and sfp information of the
#               switches is stored in a hash.
################################################################################
sub get_switch_info {

    my ($self) = @_;
    # define global variables
    my (@switches, $switch, @switch_firmwares, @switch_models, $models,
        $firmware, $ns, $audit, $port_type_from_sysconfig, $ports_discovered);

    # Check to see if we are using modules. If we are then we only need
    #  one of the fabrics.
    my $modules_found = $ClientUtil::FALSE;
    push (@switches, @{$self->{fabric1_switches}});
    if (exists $self->{modules}) {
        $modules_found = $ClientUtil::TRUE;
        # We don't need to push the second fabric on the array.
    } else {
        push (@switches, @{$self->{fabric2_switches}});
    }

    ## Build a hash of the wwpns that we know of in the test bed.
    my $sysconfig_hash = $self->get_sysconfig_data();
    if ($self->error()) {
        return {};
    }

    foreach $switch (@switches) {
        my $s_obj;

        if (ref ($self->{switch_hash}->{$switch}->{switch_obj}) =~ /switch/i) {
            $s_obj = $self->{switch_hash}->{$switch}->{switch_obj};
            $s_obj->enable_sancache(-invalidate_cache => $SanUtil::TRUE);
        } else {
            my $vfs = undef;

            if (defined $self->{vfs}->{$switch}) {
                $vfs = join (',', @{$self->{vfs}->{$switch}});
            }

            $s_obj = SanSwitch->new(-host           => $switch,
                                    -persistent     => $SwitchUtil::TRUE,
                                    -enable_caching => $SwitchUtil::TRUE,
                                    -defined_vf_ids => $vfs,
                                    -port_retries   => 20
                                   );
            if ($s_obj->error) {
                my $msg = join($s_obj->errmsg(),"\n");
                $self->_set_err("Failed to create SanSwitch object for $switch ".
                                "Reason: $msg");
                return {};
            }
        }

        # Create a Port Manager object for the switch
        my $port_mgr = $s_obj->create_portmgr();
        if ($s_obj->error) {
            my $msg = join("\n", $s_obj->errmsg());
            $self->_set_err("Could not create a PortMgr object ($msg)");
            return {};
        }

        # get the name server information about the switch to obtain type and wwp
        # and iterate over the type to find E-ports.
        $s_obj->invalidate_items_cache_data("nameserver");
        my $all_virtual_fabrics = 1;
        $ns = $port_mgr->get_nameserver($all_virtual_fabrics);

        $self->sanlog(-msg_level => 6,
                      -msg => "Nameserver for $switch:\n" . Dumper ($ns));

        foreach my $ns_key (sort keys(%{$ns})) {

            #Ignore ports that we don't need
            next if ($ns->{$ns_key}->{type} =~ /unknown|offline/i or
                     $ns->{$ns_key}->{kind} eq "disk" or
                     not exists $ns->{$ns_key}->{wwpn});

            if ($s_obj->{vf_status} == 1) {
                $s_obj->set_vf_context($ns->{$ns_key}->{'fid'});
                if ($s_obj->error) {
                    my @msg = $s_obj->errmsg();
                    &SanUtil::nate_log('-fatal',
                            "inject_fault.thpl: could not set virtual fabric context\n" .
                            "to ($ns->{$ns_key}->{'fid'}) on switch ($switch)\n" .
                            "Reason: @msg");
                }
            }

            ## If we got this wwpn from sysconfig then set the port_type here.
            my $wwpn = $ns->{$ns_key}->{wwpn};
            if (exists $sysconfig_hash->{$wwpn}) {
                $port_type_from_sysconfig = $sysconfig_hash->{$wwpn};
            }

            my @port_nums = $port_mgr->get_port_nums(-wwpn => $wwpn);

            if ($port_mgr->error) {
                my $msg = join($port_mgr->errmsg(),"\n");
                $self->_set_err("Could not get port number for WWPN $wwpn ($msg)");
                return {};
            }

            foreach my $port_num (@port_nums) {

                # Call get_port_info to get the port information
                my $port_info = $port_mgr->get_port_info($port_num);
                if ($port_mgr->error) {
                    $self->_set_err("Information unavailable on port $port_num");
                    return {};
                } else {

                    # If we are using VFs/VSANs, we need to see if this
                    # port is in one of the VFs/VSANs that the user listed.
                    # If it is not then we will skip it.
                    # E_Ports are always on the default VSAN so we'll allow those too.
                    my $skip = $ClientUtil::TRUE;
                    foreach my $vf (@{$self->{vfs}->{$switch}}) {
                        if (($ns->{$ns_key}->{'fid'} =~ /^$vf|1$/i) or ($port_info->{'vsan'} =~ /^$vf|1$/i)) {
                            $skip = $ClientUtil::FALSE;
                        }
                    }
                    next if ($skip == $ClientUtil::TRUE);

                    ## If we got this wwpn from sysconfig for this port then we can go
                    ##  ahead and hardcode the port_type for the port.
                    if ($port_type_from_sysconfig ne "") {
                        $port_info->{'type'} = $port_type_from_sysconfig;
                    }
                    $port_type_from_sysconfig = "";

                    $port_info->{fid} = $ns->{$ns_key}->{'fid'} if ($s_obj->{vf_status} == 1);

                    $self->sanlog(-msg =>
                            "Port $port_num Info\n"
                            . "Type                   = $port_info->{type}\n"
                            . "Speed                  = $port_info->{speed}\n"
                            . "Min Speed              = $port_info->{min_speed}\n"
                            . "Max Speed              = $port_info->{max_speed}\n"
                            . "State                  = $port_info->{state}\n"
                            . "State Reason           = $port_info->{state_reason}\n"
                            . "LX Mode                = $port_info->{lx_mode}\n"
                            . "Max/Resv Buffers       = $port_info->{max_resv_buffers}\n"
                            . "Link Distance          = $port_info->{link_distance}\n"
                            . "SFP Speed Capabilities = $port_info->{sfp_speed_capabilities}\n"
                            . "SFP Max 9u             = $port_info->{sfp_max_9u_meters}\n"
                            . "SFP Max 50u            = $port_info->{sfp_max_50u_meters}\n"
                            . "SFP Max 62.5u          = $port_info->{sfp_max_625u_meters}\n"
                            . "SFP Vendor Name        = $port_info->{sfp_vendor_name}\n"
                            . "SFP Part Number        = $port_info->{sfp_part_number}\n"
                            . "SFP Wavelength         = $port_info->{sfp_wavelength}\n"
                            . "FCID                   = $port_info->{fcid}\n");

                    # Store the port information in the hash 'audit'
                    if ($port_info->{'type'} =~ /E_Port/i) {
                        $audit->{$switch}->{eport}->{$port_num} = {%{$port_info}};
                        push(@{$ports_discovered->{$switch}->{eport}}, $port_num);
                    } elsif ($port_info->{'type'} =~ /FL_Port/i) {
                        $audit->{$switch}->{lport}->{$port_num} = {%{$port_info}};

                        # We may get called again in case of L-ports for each
                        # device on the loop, we just need one port from that
                        # list to create this has, so if we already have it
                        # ignore the rest
                        last;
                    } elsif ($port_info->{'type'} =~ /fport/) {
                        $audit->{$switch}->{fport}->{$port_num} = {%{$port_info}};
                        push(@{$ports_discovered->{$switch}->{fport}}, $port_num);
                    } elsif ($port_info->{'type'} =~ /fcvi/) {
                        $audit->{$switch}->{fcvi}->{$port_num} = {%{$port_info}};
                        push(@{$ports_discovered->{$switch}->{fcvi}}, $port_num);
                    } else { ## Assume ATTO Port
                        $audit->{$switch}->{aport}->{$port_num} = {%{$port_info}};
                        push(@{$ports_discovered->{$switch}->{aport}}, $port_num);
                    }
                }
            }
        }

        # Store the version information in an array
        my $version = $s_obj->get_firmware_version();
        if ($s_obj->error) {
            $self->_set_err("Could not get switch firmware");
            return {};
        } else {
            $self->sanlog(-msg => "FW revision = $version");
        }
        push(@{$audit->{$switch}->{fwversion}}, $version);

        # Store the model information in an array
        my $model = $s_obj->get_model();
        if ($s_obj->error) {
            $self->_set_err("Could not get switch model");
            return {};
        } else {
            $self->sanlog(-msg => "Switch Model = $model");
        }
        push(@{$audit->{$switch}->{switchmodel}}, $model);

        # logout so that we don't hold this connection open
        $s_obj->disconnect();

        $s_obj->disable_sancache(-invalidate_cache => $SanUtil::TRUE);

        # Copy the switch object onto the switch hash
        $audit->{$switch}->{switch_obj} = $s_obj;
        $audit->{$switch}->{port_obj} = $port_mgr;
        $audit->{$switch}->{ns_obj} = $ns;

    }

    ## Sort out which ports are in which module.
    if ($modules_found) {
        $self->_get_module_ports($audit);
        if ($self->error()) {
            return {};
        }
    }

    &SanUtil::nate_log('comment', "Ports Discovered: \n" . Dumper($ports_discovered));

    $self->sanlog(-msg_level => 6,
                  -msg => Dumper ($audit));

    return $audit;
}

################################################################################
# Method:       collect_switch_info
# Objective:    Collects the switch information and saves it in a hash
# Details:      calls get_switch_info() and saves the hash that is returned
#               from the function back into the object.
################################################################################
sub collect_switch_info {
    my ($self) = @_;

    # call get_switch_info() method to get the hash and store it in the object
    my $switch_hash = {};
    $switch_hash = $self->get_switch_info();

    if ($self->error){
        return $ClientUtil::FALSE;
    }

    # Save the hash in the object
    $self->{switch_hash} = $switch_hash;

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      merge_collected_data()
# Objective:   Merges all the data that has cross dependencies on all the calls
#              collect calls being made. The user make sure that the object
#              is populated properly and all the necessary collect functions
#              have been called.
################################################################################
sub merge_collected_data {
    my ($self) = @_;

    # Merge collected data about disk login in the switch nameserver and
    # from ontap.
    # Disks connected via FC:
    foreach my $switch (keys %{$self->{switch_hash}}) {
        foreach my $ns_ent (keys %{$self->{switch_hash}->{$switch}->{ns_obj}}) {
            my $ns_hash_ent = $self->{switch_hash}->{$switch}->{ns_obj}->{$ns_ent};
            my $disk_hash = $self->{storage_hash}->{disks};
            if ($ns_hash_ent->{kind} =~ /disk/ or
                (&SwitchUtil::normalize_type($ns_hash_ent->{type}) =~ /FL_Port/i and
                $ns_hash_ent->{count} > 1)) {

                my $entry_found = 0;
                # Check through and find the matching disk in the storage hash
                foreach my $disk (keys %{$disk_hash}) {
                    if ($disk_hash->{$disk}->{wwnn} =~ /$ns_hash_ent->{wwnn}/i) {
                        $disk_hash->{$disk}->{'acknowledged-by-switch'}->{$switch} = $ns_ent;
                        $disk_hash->{$disk}->{'switch-alpa'} = $ns_hash_ent->{address};
                        $entry_found++;
                    }
                }
                if ($entry_found == 0) {
                    push (@{$self->{switch_hash}->{$switch}->{unmatched_ns_entries}}, $ns_ent);
                }
            }
        }
        $self->sanlog(-msg_level => 7,
                      -msg => Dumper ($self->{switch_hash}->{$switch}->{ns_obj}));
        # TODO this has to be tunred into an ERROR in one of the later
        # checks. We have found that we don't have entried on the filer that
        # match the disks
        if (defined $self->{switch_hash}->{$switch}->{unmatched_ns_entries} and
            scalar @{$self->{switch_hash}->{$switch}->{unmatched_ns_entries}}){
           &SanUtil::nate_log('-warn', "Found unmatched nameserver entries " .
                                       "for $switch. Talk to MC dev about:\n" .
                                       Dumper ($self->{switch_hash}->{$switch}->{unmatched_ns_entries}));
        }
    }

    # Disks connected via SAS/SATA
    my $disk_hash = $self->{storage_hash}->{disks};
    foreach my $bridge (keys %{$self->{bridge_hash}}) {
        my @disk_list = @{$self->{bridge_hash}->{$bridge}->{disks}};
        foreach my $disk (@disk_list) {
            # If the disk that the bridge sees is already in our disk hash:
            if (exists $disk_hash->{$disk} and $disk_hash->{$disk}->{disk_type} !~ /^FC/) {
                # Add an entry in the disk_hash to show that the disk is seen by
                # the bridge.
                $disk_hash->{$disk}->{'acknowledged-by-switch'}->{$bridge} = $bridge;
            }
        }
    }

    # Clear any aggregates and disks that we need to ignore.
    foreach my $controller (@{$self->{controllers}}) {
        foreach my $ignr_aggr (@{$self->{ignoring_vals}->{aggrs}->{$controller}}) {
            if (exists $self->{storage_hash}->{aggrs}->{$controller}->{$ignr_aggr}) {
                delete $self->{storage_hash}->{aggrs}->{$controller}->{$ignr_aggr};
                &SanUtil::nate_log('-warn',
                                   "Deleting aggr $ignr_aggr from being checked.");
            }
        }
    }

    # Skips drives
    foreach my $ignr_disk (@{$self->{ignoring_vals}->{drive_serial}}) {
        if (exists $self->{storage_hash}->{disks}->{$ignr_disk}) {
            delete $self->{storage_hash}->{disks}->{$ignr_disk};
            &SanUtil::nate_log('-warn',
                               "Deleting disk $ignr_disk from being checked.");
        }
    }

    # Skip drives of a give type
    foreach my $type (@{$self->{ignoring_vals}->{drive_types}}) {
        foreach my $disk (keys %{$self->{storage_hash}->{disks}}) {
            if ($self->{storage_hash}->{disks}->{$disk}->{'disk-model'} =~ /$type/i) {
                delete $self->{storage_hash}->{disks}->{$disk};
                &SanUtil::nate_log('-warn',
                                   "Deleting disk $disk of type $type.");
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      display_info()
# Objective:   Prints out the controller and switch information
# Details:     Loops through controller_hash and switch_hash, retrieves the
#              information and prints it
################################################################################
sub display_info {
    my ($self) = @_;

    my ($count, $fnumber, $controller, $fmware, @per_adap_models, @count_of_disks,
        @per_adap_firmwares, @controllers, @display_info);
    my (@switches, $switch, @switch_firmwares, @switch_models, $models, $firmware,);
    my (@per_shelf_models, @per_shelf_firmware, @count_of_shelves, $sh_fmware,
        $sh_count, $per_shelf_model, $sh_model);

    my $hash  = $self->{controller_hash};
    my $shash = $self->{switch_hash};
    my $dhash = $self->{storage_hash};
    my $bhash = $self->{bridge_hash};
    my $comment =       "*************************************************************\n";
    my $comment_minor = "-------------------------------------------------------------\n";

    # Loops through controller information and prints it
    push (@display_info, "\n$comment" . "* System Summary:\n" . $comment);
    foreach $controller (@{$self->{controllers}}) {
        my $display_line = '';
        push (@display_info, "\n$comment_minor| $controller \n". $comment_minor);
        push (@display_info,  sprintf ("%18s \t %s \n","Controller Model:", $hash->{$controller}->{controller_model}));
        push (@display_info,  sprintf ("%18s \t %s \n","ONTAP version:", $hash->{$controller}->{ontap_version}));
        push (@display_info,  sprintf ("%18s \t %s \n","System Name:", $hash->{$controller}->{system_name}));
        push (@display_info,  sprintf ("%18s \t %s \n","System Id:", $hash->{$controller}->{system_id}));
        push (@display_info,  sprintf ("%18s \t %s \n","Partner Name:", $hash->{$controller}->{partner_name}));
        push (@display_info,  sprintf ("%18s \t %s \n","Partner Id:", $hash->{$controller}->{partner_id}));


        # Prints the number of disks, shelves, firmware versions and model names
        # of evey adapter
        foreach my $adapter (sort keys(%{$hash->{$controller}->{adapters}})) {

            my $adapter_hash = $hash->{$controller}->{adapters}->{$adapter};
            #TODO TODO Ignore OFFLINE adapters - Link State for adapter channel 0b: OFFLINE (physical)

            push (@display_info, "\n\tInformation for FC adapter \"$adapter\"\n");
            push (@display_info, "\t===================================\n");
            # push FCVI chipset information into display information array
            if ($adapter_hash->{FCVI_chipset} ne "") {
                push (@display_info,"\t\t". sprintf ("%-16s\t%s\n", "Adapter Type:" , "FCVI"));
                push (@display_info,"\t\t". sprintf("%-16s\t%s\n","Chipset:", $adapter_hash->{FCVI_chipset}));
            }

            if ($adapter_hash->{card_chipset} ne "") {
                push (@display_info,"\t\t". sprintf ("%-16s\t%s\n", "Adapter Type:" , "HBA"));
                push (@display_info,"\t\t". sprintf("%-16s\t%s\n","Chipset:", $adapter_hash->{card_chipset}));
            }

            if ($adapter_hash->{link_data_rate} ne "") {
                push (@display_info,"\t\t". sprintf("%-16s\t%s\n","Link Data Rate:", $adapter_hash->{link_data_rate}));
            }

            if ($adapter_hash->{switch_port} ne "") {
                push (@display_info,"\t\t". sprintf("%-16s\t%s\n","Switch Port:", $adapter_hash->{switch_port}));
            }

            if ($adapter_hash->{firmware} ne "") {
                push (@display_info,"\t\t". sprintf("%-16s\t%s\n","Firmware Ver.:", $adapter_hash->{firmware}));
            }

            if (keys(%{$adapter_hash->{disks}}) != "" and keys(%{$adapter_hash->{shelves}}) != "") {
                my $header = "\t\t". sprintf("%-23s %8s %10s\n", "Model", "Firmware",
                                             "Disk Count");
                $header .= "\t\t". sprintf("%-23s %8s %10s\n", "-----", "--------",
                                           "----------");
                push(@display_info, $header);
            }
            @per_adap_models = keys(%{$adapter_hash->{disks}->{models}});
            @per_shelf_models = keys(%{$adapter_hash->{shelves}->{models}});

            # Prints all unique models of every disk TODO use the disk HASH
            foreach my $per_adap_model (@per_adap_models) {
                @per_adap_firmwares = keys(%{$adapter_hash->{disks}->{models}->{$per_adap_model}->{firmwares}});
                # Prints all unique firmware versions of each disk
                foreach $fnumber (@per_adap_firmwares) {
                    push(@count_of_disks,
                         $adapter_hash->{disks}->{models}->{$per_adap_model}->{firmwares}->{$fnumber});
                }
                $fmware = pop(@per_adap_firmwares);
                $count  = pop(@count_of_disks);

                my $display_line = sprintf("\t\t%-23s   %-8s      %-6s\n",
                                   $per_adap_model, $fmware, $count);
                push (@display_info, $display_line);
            }

            # Prints all unique models for every shelf TODO get using ZAPIs add scalar
            if (keys(%{$adapter_hash->{disks}}) != "" and scalar (keys(%{$adapter_hash->{shelves}->{models}})) != 0) {
                push (@display_info, sprintf("\n\t\t%-11s %-20s %11s\n", "Shelf model", "Shelf FW",
                                             "Shelf Count"));
                push (@display_info, sprintf("\t\t%-11s %-20s %11s\n", "-----------", "----------",
                                             "-----------"));
                foreach $per_shelf_model (@per_shelf_models) {
                    @per_shelf_firmware = keys( %{$adapter_hash->{shelves}->{models}->{$per_shelf_model}->{firmwares}});
                    # Prints all unique firmware versions of every shelf
                    foreach my $sh_fmnumber (@per_shelf_firmware) {
                        $sh_fmware = $sh_fmnumber;
                        $sh_count  = $adapter_hash->{shelves}->{models}->{$per_shelf_model}->{firmwares}->{$sh_fmnumber};
                        $sh_model  = $per_shelf_model;
                        push (@display_info, sprintf("\t\t%-11s %-20s %11s\n", $sh_model, $sh_fmware, $sh_count));
                    }

                }
            }

            # push the warning line into display info array
            if ($#{$adapter_hash->{warnings}} != -1) {
                push (@display_info, "\n\t\t". "Failed/Bypassed Disks:\n");
                push (@display_info, "\t\t". "-----------------------\n");
                foreach my $warning_line (@{$adapter_hash->{warnings}}) {
                    push (@display_info, "\t$warning_line");
                }
                push (@display_info, "\n");
            }

        }
    }

    # Printing switch information
    push (@display_info, "\n$comment_minor| Switch Information \n". $comment_minor);
    my $heading = sprintf("\t%-15s %8s %-15s\n", "Switch", "Firmware", "Model");
    $heading .=   sprintf("\t%-15s %8s %-15s\n", "------", "--------", "-----");
    push (@display_info, $heading);
    push (@switches, @{$self->{fabric1_switches}});
    push (@switches, @{$self->{fabric2_switches}});
    foreach $switch (@switches) {
        $firmware = shift(@{$shash->{$switch}->{fwversion}});
        $models   = shift(@{$shash->{$switch}->{switchmodel}});
        my $display_line = sprintf("\t%-15s %-9s %-15s\n",$switch,$firmware,$models);
        push(@display_info, $display_line);
    }

    # Prints the information of the E-Port of a switch
    foreach $switch (@switches) {
        if ($shash->{$switch}->{eport}->{type} =~ /^E-Port$/) {
            my $display_line = sprintf("Switch Information:" . $switch . "\n"
                                . "Type                     =" . $shash->{$switch}->{eport}->{type} . "\n"
                                . "Speed                    =" . $shash->{$switch}->{eport}->{speed} . "\n"
                                . "Min Speed                =" . $shash->{$switch}->{eport}->{min_speed} . "\n"
                                . "Max Speed                =" . $shash->{$switch}->{eport}->{max_speed} . "\n"
                                . "State                    =" . $shash->{$switch}->{eport}->{state} . "\n"
                                . "State Reason             =" . $shash->{$switch}->{eport}->{state_reason} . "\n"
                                . "LX Mode                  =" . $shash->{$switch}->{eport}->{lx_mode} . "\n"
                                . "Max/Resv Buffers         =" . $shash->{$switch}->{eport}->{max_resv_buffers} . "\n"
                                . "Link Distance            =" . $shash->{$switch}->{eport}->{link_distance} . "\n"
                                . "SFP Speed Capabilities   =" . $shash->{$switch}->{eport}->{sfp_speed_capabilities} . "\n"
                                . "SFP Max 9u               =" . $shash->{$switch}->{eport}->{sfp_max_9u_meters} . "\n"
                                . "SFP Max 50u              =" . $shash->{$switch}->{eport}->{sfp_max_50u_meters} . "\n"
                                . "SFP Max 62.5u            =" . $shash->{$switch}->{eport}->{sfp_max_625u_meters} . "\n"
                                . "SFP Vendor Name          =" . $shash->{$switch}->{eport}->{sfp_vendor_name} . "\n"
                                . "SFP Part Number          =" . $shash->{$switch}->{eport}->{sfp_part_number} . "\n"
                                . "SFP Wavelength           =" . $shash->{$switch}->{eport}->{sfp_wavelength} . "\n"
                                . "FCID                     =" . $shash->{$switch}->{eport}->{fcid} . "\n");
           push (@display_info, $display_line);
        }
    }

    # Print the disk info.
    push (@display_info, "\n$comment_minor| Disk Information \n". $comment_minor);
    $heading = sprintf("\t%-20s %-25s %-22s %-10s %-6s %-9s\n", "Owner", "Disk", "Serial", "Aggr",
                          "Pool", "RAID");
    $heading .=   sprintf("\t%-20s %-25s %-22s %-10s %-6s %-9s\n", "------", "------", "--------",
                          "-----","-----","------");

    push (@display_info, $heading);

    foreach my $disk (keys %{$dhash->{disks}}) {
        push (@display_info, sprintf("\t%-20s %-25s %-22s %-10s %-6s %-9s\n",
                    $dhash->{disks}->{$disk}->{'owner'},
                    $dhash->{disks}->{$disk}->{'primary-name'},
                    $dhash->{disks}->{$disk}->{'serial-number'},
                    $dhash->{disks}->{$disk}->{'aggregate'},
                    $dhash->{disks}->{$disk}->{'pool'},
                    $dhash->{disks}->{$disk}->{'raid-state'}));
    }

    foreach my $bridge (keys %{$bhash}) {
        push (@display_info, @{$bhash->{$bridge}->{'info'}});
    }

    $self->sanlog(-msg => join('', @display_info));

    # TODO - Maybe print this to file so we can save it in some way and use for burts, logs, etc.
    print join ('', @display_info);

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      validate_initial_environment()
# Objective:   Validates the metrocluster environment
# Details:     This functions validates to see if the metrocluster configuration
#              is valid, before injecting any faults
################################################################################
sub validate_initial_environment {
    my ($self) = @_;
    my (@errors, @warnings);
    my $return_val = $ClientUtil::TRUE;

    @warnings = @errors = ();

    if ($self->validate_initial_environment_cf_monitor() == $ClientUtil::FALSE or
        $self->error) {
        $return_val = $ClientUtil::FALSE;

        if ($self->error) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }

        if ($self->{warnings_set} == $ClientUtil::TRUE) {

            # Collect any warnings we have and clear flags
            push (@warnings, @{$self->{warnings}});
            $self->{warnings_set} = $ClientUtil::FALSE;
            @{$self->{warnings}} = ();
        }
    }

    # Verify the HA interconnect status
    if ($self->verify_ic_status() == $ClientUtil::FALSE or
        $self->error) {
        # Returning false because regardles of warn or error we return false.
        $return_val = $ClientUtil::FALSE;

        if ($self->error) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }

        if ($self->{warnings_set} == $ClientUtil::TRUE) {

            # Collect any warnings we have and clear flags
            push (@warnings, @{$self->{warnings}});
            $self->{warnings_set} = $ClientUtil::FALSE;
            @{$self->{warnings}} = ();
        }
    }

    # Make sure we didn't get a takeover.
    # Passing in false to ensure we are not detecting takeover condition.
    if ($self->verify_cfo_status($ClientUtil::FALSE) == $ClientUtil::FALSE or
        $self->error) {
        # Returning false because regardles of warn or error we return false.
        $return_val = $ClientUtil::FALSE;

        if ($self->error) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }

        if ($self->{warnings_set} == $ClientUtil::TRUE) {

            # Collect any warnings we have and clear flags
            push (@warnings, @{$self->{warnings}});
            $self->{warnings_set} = $ClientUtil::FALSE;
            @{$self->{warnings}} = ();
        }
    }

    if ($self->validate_initial_environment_disks() == $ClientUtil::FALSE or
        $self->error) {
        $return_val = $ClientUtil::FALSE;

        if ($self->error) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }

        if ($self->{warnings_set} == $ClientUtil::TRUE) {

            # Collect any warnings we have and clear flags
            push (@warnings, @{$self->{warnings}});
            $self->{warnings_set} = $ClientUtil::FALSE;
            @{$self->{warnings}} = ();
        }
    }

    # Call validate_license to confirm that the appropriate licenses are installed.
    if ($self->validate_initial_environment_licenses() == $ClientUtil::FALSE or
        $self->error) {
        $return_val = $ClientUtil::FALSE;
        if ($self->error) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }

        if ($self->{warnings_set} == $ClientUtil::TRUE) {

            # Collect any warnings we have and clear flags
            push (@warnings, @{$self->{warnings}});
            $self->{warnings_set} = $ClientUtil::FALSE;
            @{$self->{warnings}} = ();
        }
    }

    if (scalar(@errors)) {
        $self->_set_err("\n". join("", @errors));
        $return_val = $ClientUtil::FALSE;
    }

    if (scalar(@warnings)) {
       $self->{warnings_set} = $ClientUtil::TRUE;
       @{$self->{warnings}} = @warnings;
       $return_val = $ClientUtil::FALSE;
    }

    return $return_val;

}

################################################################################
# Method:      validate_initial_environment()
# Objective:   Validates the metrocluster environment
# Details:     This functions validates to see if the metrocluster configuration
#              is valid, before injecting any faults
################################################################################
sub validate_initial_environment_sysconfig {
    my ($self) = @_;

    my ($error_msg, @errors);

}

################################################################################
# Method:      validate_initial_environment_disks()
# Objective:   Validates disks in the metrocluster environment
# Details:     This functions validates to see if the metrocluster configuration
#              is valid, before injecting any faults
################################################################################
sub validate_initial_environment_disks {
    my ($self) = @_;

    my ($error_msg, @errors, @warnings);

    my $return_val = $ClientUtil::TRUE;
    my $error_header_set = 0;

    # Check if aggregate state is online, if so check it is mirrored warn otherwise
    # and note unmirrored aggrs for future tests. If it is not online check it is
    # offline and not is any other state. And log initial offline states for future
    # tests.
    foreach my $controller (keys %{$self->{controller_hash}}) {

        # Check each of the aggregates
        my $aggr_hash = $self->{storage_hash}->{aggrs}->{$controller};

        foreach my $aggr (keys %{$aggr_hash}) {

            my (@aggr_error, $error_lvl);
            $error_lvl = 0;

            # Set intial vals
            $aggr_hash->{$aggr}->{'is-offline-at-start'} = $ClientUtil::FALSE;
            $aggr_hash->{$aggr}->{'is-failed-at-start'} = $ClientUtil::FALSE;
            $aggr_hash->{$aggr}->{'is-partial-at-start'} = $ClientUtil::FALSE;

            $aggr_hash->{$aggr}->{'start-state'} = $aggr_hash->{$aggr}->{'state'};
            $aggr_hash->{$aggr}->{'start-mirror-status'} =
                    $aggr_hash->{$aggr}->{'mirror-status'};

            if ($aggr_hash->{$aggr}->{'state'} =~ /offline|online/) {

                if ($aggr_hash->{$aggr}->{'state'} eq 'offline') {
                    # If it is offline mark it so. Checking offline first
                    # because it could be mirrored and offline.
                    $aggr_hash->{$aggr}->{'is-offline-at-start'} = $ClientUtil::TRUE;
                    $error_lvl = 1;
                    push (@aggr_error, "\t\tAggr \"$aggr\" on $controller ".
                          "is offline.\n");

                } elsif ($aggr_hash->{$aggr}->{'mirror-status'} eq 'mirrored') {

                    # Everything as expected keep going

                } elsif ($aggr_hash->{$aggr}->{'mirror-status'} eq 'unmirrored') {

                    # We might be able to allow unmirrored aggrs as long as not root
                    if ($aggr_hash->{$aggr}->{'has-local-root'} eq 'true') {
                       push (@errors,"\t\tFATAL: Root aggregate \"$aggr\" unmirrored on $controller.");
                    } else {
                       $aggr_hash->{$aggr}->{'is-unmirrored-at-start'} = $ClientUtil::TRUE;
                       $error_lvl = 2;
                       push (@aggr_error, "\t\tAggr \"$aggr\" on $controller ".
                             "is unmirrored.\n");
                    }

                } elsif ($aggr_hash->{$aggr}->{'mirror-status'} =~ /resync/) {

                    $error_lvl = 2;
                    push (@aggr_error, "\t\tAggr \"$aggr\" on $controller ".
                          "is resyncing wait for it to finish.\n");

                } else {

                    # It must be one of the states mirror degraded,
                    # failed, limbo, and . It should be fixed before
                    # we go forward.
                    push(@errors, "\t\tOn $controller, aggr " .
                         "\"$aggr\" has mirror-status \"" .
                         $aggr_hash->{$aggr}->{'mirror-status'} .
                         "\" is not in one of the acceptable states.\n");

                }

            } elsif ($aggr_hash->{$aggr}->{'state'} =~ /failed/) {

                $error_lvl = 2;
                $aggr_hash->{$aggr}->{'is-failed-at-start'} = $ClientUtil::TRUE;
                push (@aggr_error, "\t\tAggr \"$aggr\" on $controller is " .
                      $aggr_hash->{$aggr}->{'state'} . ".\n");

            } elsif ($aggr_hash->{$aggr}->{'state'} =~ /partial/) {

                $error_lvl = 3;
                $aggr_hash->{$aggr}->{'is-partial-at-start'} = $ClientUtil::TRUE;
                push (@aggr_error, "\t\tAggr \"$aggr\" on $controller is " .
                      $aggr_hash->{$aggr}->{'state'} . ".\n");

            } else {

                push(@errors, "\t\tAggr \"$aggr\" on $controller ".
                     "is in an unacceptable state ".
                     $aggr_hash->{$aggr}->{'state'} . ".\n");

            }

            # Mark it error or warning based on the error.
            if (scalar (@aggr_error)) {
                my $header = "";
                if ($error_header_set == 0) {
                    $header = "Found following aggregate related issues:\n";
                    $error_header_set = 1;
                    if ($self->{ERROR_LEVEL} >= $error_lvl) {
                        push (@warnings, "\n\tWARNING: $header");
                    } else {
                        push (@errors, "\n\tERROR: $header");
                    }
                }

                #is it a warning or error?
                if ($self->{ERROR_LEVEL} >= $error_lvl) {
                   @warnings = (@warnings, @aggr_error);
                } else {
                    @errors = (@errors, @aggr_error);
                }
            }
        }
    }

    my $disk_hash = $self->{storage_hash}->{disks};
    my $storage_hash = $self->{storage_hash};
    my (@no_sec_port_disks, @unowned_disks, @owned_by_others);
    my (@system_falied_disks, @raid_falied_disks);

    foreach my $disk (keys %{$disk_hash}) {
        my $owner = "";
        if (exists ($disk_hash->{$disk}->{'owner'})) {
            $owner = $disk_hash->{$disk}->{'owner'};

            # Find out if the disks are owned by some other controller
            my $found_on_controller = $ClientUtil::FALSE;
            foreach my $controller (keys %{$self->{controller_hash}}) {
                if ($owner =~ /^\s*$controller\s*$/i) {
                    $found_on_controller = $ClientUtil::TRUE;
                }
            }

            if ($found_on_controller == $ClientUtil::FALSE) {
                push (@owned_by_others, $disk);
                # Ignore all other errors if it is not owned by us.
                next;
            }
        }

        # Check the disks to see if the disk if failed, unowned or if the secondary port is
        # connected. They could have multiple conditions but only one will be acknowledged
        # with failed disks taking predence over unowned and disks with no secondary ports.
        if ($disk_hash->{$disk}->{'is-failed'} == $ClientUtil::TRUE) {
            # Look at the failed disks first since failed disks might lead more invalid data
            if ($disk_hash->{$disk}->{'acknowledged-by-raid'}->{$owner} == $ClientUtil::TRUE) {
                # Acked by RAID and failed
                push(@raid_falied_disks, $disk);
            } else {
                # Here we will have only system failed disks
                push(@system_falied_disks, $disk);
            }
        } elsif (!exists $disk_hash->{$disk}->{'owner'}) {
            # If it is not failed then let's see if it unowned
            push (@unowned_disks, $disk);
        } elsif (($disk_hash->{$disk}->{'secondary-port'} ne 'A' and
                  $disk_hash->{$disk}->{'secondary-port'} ne 'B')
                 or
                 ($disk_hash->{$disk}->{'port'} eq
                  $disk_hash->{$disk}->{'secondary-port'})
                 or
                 ($disk_hash->{$disk}->{'primary-name'} eq
                  $disk_hash->{$disk}->{'secondary-name'})) {
            # It is not failed or unowned, lets check for secondary ports.
            push(@no_sec_port_disks, $disk);
        }
    }

    if (scalar(@no_sec_port_disks)) {
        my @sec_errors;
        # We found some disks we no secondary port.
        push (@sec_errors, sprintf ("Found %d disks with no secondary ports:\n",
                    scalar (@no_sec_port_disks)));

        foreach my $disk (@no_sec_port_disks) {

            my $owner = $disk_hash->{$disk}->{'owner'};
            my $type = $disk_hash->{$disk}->{'raid-state'};
            my $name = $disk_hash->{$disk}->{'primary-name'};

            if (exists $disk_hash->{$disk}->{'aggregate'}) {
                push (@sec_errors, sprintf("\t\t $name($disk) owned by \"$owner\"" .
                      " for \"aggregate\" %s does not have a secondary port.\n",
                      $disk_hash->{$disk}->{'aggregate'}));
            } else {
                push (@sec_errors, sprintf("\t\t $name($disk) owned by \"$owner\"" .
                      " of type \"$type\" does not have a secondary port.\n"));
            }
        }

        if ($self->{ERROR_LEVEL} == 3) {
            # If error level is set to 3 we can put these in as warnings
            push (@warnings, "\n\tWARNING: ");
            @warnings = (@warnings, @sec_errors);
        } else {
            push (@errors, "\n\tERROR: ");
            @errors = (@errors, @sec_errors);
        }
    }

    if (scalar (@unowned_disks)) {
        my @unown_errors;

        push (@unown_errors, sprintf ("Found %d unowned disks:\n",
              scalar(@unowned_disks)));

        foreach my $disk (@unowned_disks) {
            my $name = $disk_hash->{$disk}->{'primary-name'};
            push (@unown_errors, sprintf("\t\t$name($disk) is not owned.\n"));
        }

        if ($self->{ERROR_LEVEL} >= 2) {
            # If error level is set to 2 we can put these in as warnings
            push (@warnings, "\n\tWARNING: ");
            @warnings = (@warnings, @unown_errors);
        } else {
            push (@errors, "\n\tERROR: ");
            @errors = (@errors, @unown_errors);
        }
    }

    if (scalar (@owned_by_others)) {
        my @own_errors;

        push (@own_errors, sprintf ("Found %d disks owned by other nodes:\n",
              scalar(@owned_by_others)));

        foreach my $disk (@owned_by_others) {
            my $name = $disk_hash->{$disk}->{'primary-name'};
            my $owner = $disk_hash->{$disk}->{'owner'};
            push (@own_errors, sprintf("\t\t$name($disk) is owned by " .
                                         " $owner, which is not one of " .
                                         " the controllers defined on " .
                                         " this setup.\n"));
        }

        if ($self->{ERROR_LEVEL} >= 2) {
            # If error level is set to 2 we can put these in as warnings
            push (@warnings, "\n\tWARNING: ");
            @warnings = (@warnings, @own_errors);
        } else {
            push (@errors, "\n\tERROR: ");
            @errors = (@errors, @own_errors);
        }
    }

    if (scalar (@system_falied_disks)) {
        my @fail_errors;

        push (@fail_errors, sprintf("Found %d system failed disks:\n",
              scalar(@system_falied_disks)));

        foreach my $disk (@system_falied_disks) {
            my $name = $disk_hash->{$disk}->{'primary-name'};
            push (@fail_errors, sprintf("\t\t$name($disk) is failed.\n"));
        }

        if ($self->{ERROR_LEVEL} >= 1) {
            # If error level is set to 1 we can put these in as warnings
            push (@warnings, "\n\tWARNING: ");
            @warnings = (@warnings, @fail_errors);
        } else {
            push (@errors, "\n\tERROR: ");
            @errors = (@errors, @fail_errors);
        }
    }

    if (scalar(@raid_falied_disks)) {
        my @fail_errors;

        push (@fail_errors, sprintf("Found %d RAID-failed disks:\n",
              scalar(@raid_falied_disks)));

        foreach my $disk (@raid_falied_disks) {
            my $name = $disk_hash->{$disk}->{'primary-name'};
            push (@fail_errors, sprintf("\t\t$name($disk) is failed.\n"));
        }

        if ($self->{ERROR_LEVEL} >= 1) {
            # If error level is set to 1 we can put these in as warnings
            push (@warnings, "\n\tWARNING: ");
            @warnings = (@warnings, @fail_errors);
        } else {
            push (@errors, "\n\tERROR: ");
            @errors = (@errors, @fail_errors);
        }
    }

    # Check if the disks are seen by all levels
    $self->_validate_disk_acknowledgements_at_all_layers($disk_hash, undef);
    # Merge errors, warnings will automatically get merged
    if ($self->error) {
        push (@errors, "\n\t" . $self->errmsg());
        $self->_clear_err();
        $return_val = $ClientUtil::FALSE;
    }

    # Make sure there aren't any unacked disks
    if (scalar (keys %{$self->{storage_hash}->{unacked_disks}})) {

        my $str = "\n\t";
        $str .= ($self->{ERROR_LEVEL} >= 4) ? "WARNING: ": "ERROR: ";
        $str .= "Found unacked disks that are seen by ses and ".
                  " not by any other layer (sanown, raid).\n";

        foreach my $disk (keys %{$self->{storage_hash}->{unacked_disks}}) {
            $str .= "\t\t Disk ($disk) of shelf ".
                    "$self->{storage_hash}->{unacked_disks}->{$disk}->{'shelf-name'}.\n";
            if ($self->{ERROR_LEVEL} >= 4) {
                push (@warnings,$str);
            } else {
                push (@errors, $str);
            }
            $str = "";
        }
        $return_val = $ClientUtil::FALSE;
    }

    # Checking the storage related states
    $self->_validate_storage_states($disk_hash);
    if ($self->error) {
        my @obj_error = $self->errmsg();
        my $msg = join("\n\t", @obj_error);
        push (@errors, "\n\t" . $msg);
        $self->_clear_err();
        $return_val = $ClientUtil::FALSE;
    }

    $self->_validate_pool_mappings($storage_hash);
    # Merge errors, warnings will automatically get merged
    if ($self->error) {
        my @obj_error = $self->errmsg();
        my $msg = join("\n\t", @obj_error);
        push (@errors, "\n\t" . $msg);
        $self->_clear_err();
        $return_val = $ClientUtil::FALSE;
    }

    if (scalar(@warnings)) {
        push (@{$self->{warnings}}, @warnings);
        $self->{warnings_set} = $ClientUtil::TRUE;
        $return_val = $ClientUtil::FALSE;
    }

    if (scalar(@errors)) {
        $self->_set_err("Disk validation failed correct following errors:\n".
                        join ("", @errors));
        $return_val = $ClientUtil::FALSE;
    }

    return $return_val;
}

################################################################################
# Method:      validate_initial_environment_cf_monitor()
# Objective:   Validates cf monitor data for MC env
# Details:     This functions validates to see if the cf monitor data
#              is valid, before injecting any faults
################################################################################
sub validate_initial_environment_cf_monitor {
    my ($self) = @_;

    my ($error_msg, @errors);
    my ($mailbox, $local, $partner, $fabric1, $fabric2,
        $mailbox_found, $count_mailbox, $link_found);

    # Check if the controller hash is empty
    if (scalar(keys %{$self->{controller_hash}}) == 0) {
        $self->_set_err("controller_hash is empty");
        return $ClientUtil::FALSE;
    }

    # Store the fabric 1 and 2 switches
    my @fab1_switches = @{$self->{fabric1_switches}};
    my @fab2_switches = @{$self->{fabric2_switches}};

    # Loop through each controller and parse the cf monitor all output
    foreach my $controller (keys %{$self->{controller_hash}}) {
        # Reset flags
        $link_found = 0;
        $mailbox_found = 0;

        # Save the output per controller locally
        my @cfmonitor_output = @{$self->{controller_hash}->{$controller}->{cf_monitor_all}};

        # Check for the following strings in the cf monitor all output:
        # 1. VIA Interconnect is up (link 0 up, link 1 up)
        # 2. takeover capability on-line
        foreach my $line (@cfmonitor_output) {
            if( $line =~ /VIA Interconnect is up \(link 0 up, link 1 up\)/i
                or $line =~ /takeover capability on-line/i ) {
                # add warning if one link is down
                # set flag. Indicates link is up.
                $link_found = 1;
            } elsif ($line =~ /\s*Disk\s.*mailbox disk/) {
                # If line looks like this:
                # Disk rtpmet-browho-1:2.33 is a local mailbox disk
                # Store them in the hash

                # Set flag. Indicates mailboxes found.
                $mailbox_found = 1;
                push (@{$self->{controller_hash}->{$controller}->{mailbox_disks}}, $line);
            }
        }

        # If flag is not set, error.
        if($link_found != 1) {
            $error_msg = sprintf("\n\t" . $controller . " Link is not up" .
                                 " or takeover capability not online\n");
            push (@errors, $error_msg);
        }

        if($mailbox_found != 1) {
            $error_msg = sprintf("\n\t" . $controller . " Mailboxes not found");
            push (@errors, $error_msg);
        }
    }

    # Loop through each controller and count the number of local, partner, and total number of mailbox disks
    foreach my $controller (keys %{$self->{controller_hash}}) {
        # Reset variables
        $local= 0;
        $partner = 0;
        $count_mailbox = 0;
        $fabric1 = 0;
        $fabric2 = 0;

        # Loop through each mailbox disk and grep for the switch name
        foreach $mailbox (@{$self->{controller_hash}->{$controller}->{mailbox_disks}}) {
            # total number of mailboxes per controller
            $count_mailbox += 1;

            if( $mailbox =~ /local/i or $mailbox =~ /primary/i) {
                # Total number of local mailbox disks per controller
                $local += 1;
            } elsif ( $mailbox =~ /partner/i or $mailbox =~ /backup/i) {
                # Total number of local mailbox disks per controller
                $partner += 1;
            }

            # Compare switch name with fabric1 and fabric2 switch names and increment count accordingly
            if( $mailbox =~ /\s*Disk\s+(\S+):\d+.\d+\s+.*/ ) {
                my $temp = $1;
                if ( grep(/$temp/, @fab1_switches) != 0 ) {
                    $fabric1 += 1;
                } elsif (grep(/$temp/, @fab2_switches) != 0 ) {
                    $fabric2 += 1;
                }
            }
        }

        # Perform following checks:
        # There are 8 mail boxes; 4 local and 4 on the partner
        # There are 8 mail boxes; 4 on one fabric and 4 on the 2nd fabric
        if( $count_mailbox != 8) {
            $error_msg = sprintf("\n\t" . $controller . " should have 8 mailboxes.");
            push (@errors, $error_msg);
        }
        if ($local != 4) {
            $error_msg = sprintf("\n\t" . $controller . " should have 4 local mailboxes.");
            push (@errors, $error_msg);
        }
        if ($partner != 4) {
            $error_msg = sprintf("\n\t" . $controller . " should have 4 partner mailboxes.");
            push (@errors, $error_msg);
        }

        # Ignore this for the time being as there is no basis for this test.
        #if ($fabric1 + $fabric2 != 8) {
        #    $error_msg = sprintf("\n\t" . $controller .
        #                         " There should be 4 mailboxes on fabric 1 and 4 mailboxes on fabric2.");
        #    push (@errors, $error_msg);
        #}
    }

    # If errors present, print and set_err
    if (scalar(@errors) != 0 and $self->{ERROR_LEVEL} >= 3) {
        push (@{$self->{warnings}}, @errors);
        $self->{warnings_set} = $ClientUtil::TRUE;
        return $ClientUtil::FALSE;
    } elsif (scalar(@errors) != 0) {
        $self->_set_err(join ("", @errors) ."\nValidation failed correct above errors.\n\n");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      validate_initial_environment_licenses()
# Objective:   Validates that the correct licenses are installed
# Details:     This function checks if the cluster_remote &
#              syncmirror_local licenses are installed on both filer heads
#
################################################################################
sub validate_initial_environment_licenses {
    my ($self) = @_;

    my ($error_msg, @errors);
    my ($mailbox, $local, $partner, $fabric1, $fabric2,
        $mailbox_found, $count_mailbox, $link_found);

    # Initializing the license detection variables
    my $cluster_remote_found = 0;
    my $syncmirror_local_found = 0;

    # Check if the controller hash is empty
    foreach my $controller (keys %{$self->{controller_hash}}) {

        ## Check to see if options are available to enable MC
        my $cf_cmd = "options cf.mode";
        my ($cf_output, $cf_status) = &SanUtil::cmds_driver_zapi($controller,
                                                           "system-cli",
                                                           "diag",
                                                           $cf_cmd);
        if ($cf_status == SanUtil::ZAPI_FAILURE) {
                $self->_set_err("Failed to execute system-cli ZAPI ($cf_cmd)");
                return {};
            }
        my $sync_cmd = "options cf.remote_syncmirror.enable";
        my ($sync_output, $sync_status) = &SanUtil::cmds_driver_zapi($controller,
                                                               "system-cli",
                                                               "diag",
                                                               $sync_cmd);
        if ($sync_status == SanUtil::ZAPI_FAILURE) {
                $self->_set_err("Failed to execute system-cli ZAPI ($sync_cmd)");
                return {};
            }
        if ($cf_output !~ /No such option/) {
            $cluster_remote_found = 1 if $cf_output !~ /non_ha/;
            $syncmirror_local_found = 1 if $sync_output !~ /off/;
        } else { ## SN.0 is not installed. We'll ensure the license is present.

            my $cmd = "license-list-info";
            # Use license info zapi to get the output of the command and store it in @output
            my ($output, $status) = &SanUtil::cmds_driver_zapi($controller,
                                                               $cmd,
                                                               "diag");
            if ($status == SanUtil::ZAPI_FAILURE) {
                $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
                return {};
            }

            $self->sanlog(-msg_level => 8,
                          -msg => "Printing licenses for $controller");
            $self->sanlog(-msg_level => 8,
                          -msg => Dumper ($output));

            # For each license found in the controller, check to see if the 2 target licenses are enabled
            foreach my $license (@{$output->{'licenses'}->{'license-info'}}) {
                my $service = $license->{'service'};
                if (($license->{'service'} =~ /cluster_remote|cf_remote/)
                    and ($license->{'is-licensed'} eq 'true')) {
                    $cluster_remote_found = 1;
                } elsif (($license->{'service'} =~ /syncmirror_local/)
                    and ($license->{'is-licensed'} eq 'true')) {
                    $syncmirror_local_found = 1;
                }
            }
        }

        # If one or more of the licenses are not found, push the error messages into the error string.
        if ($cluster_remote_found == 0) {
            $error_msg = sprintf("\n" . $controller . " does not have cluster_remote licensed or option enabled");
            push (@errors, $error_msg);
        } elsif ($syncmirror_local_found == 0) {
            $error_msg = sprintf("\n" . $controller . " does not have syncmirror_local licensed or option enabled");
            push (@errors, $error_msg);
        }
    }

    # If errors present, print and set_err
    if (scalar(@errors) != 0) {
        $self->_set_err(join("", @errors) . "Validation failed correct above errors\n");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      validate_metrocluster_fault()
# Objective:   Validates the metrocluster environment after a fault is injected
# Details:     This functions validates to see if the metrocluster configuration
#              is valid and consistent state is maintained, after injecting
#              each fault.
################################################################################
sub validate_metrocluster_fault {
    my ($self) = @_;

    my (@controllers, @fabric1_switches, @fabric2_switches);
    my (@errors, @warnings);

    my $return_val = $ClientUtil::TRUE;

    # Clear any warnings that we might have
    $self->{warnings_set} = $ClientUtil::FALSE;
    @{$self->{warnings}} = ();

    @warnings = ();
    @errors = ();

    $self->create_updated_metrocluter_object();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg_level => 7,
                  -msg => "The updated metrocluster object.");
    $self->sanlog(-msg_level => 7,
                  -msg => Dumper ($self->{current_obj}));

    # Verify the HA interconnect status
    if ($self->verify_ic_status() == $ClientUtil::FALSE or
        $self->error) {
        # Returning false because regardles of warn or error we return false.
        $return_val = $ClientUtil::FALSE;

        if ($self->error) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }

        if ($self->{warnings_set} == $ClientUtil::TRUE) {

            # Collect any warnings we have and clear flags
            push (@warnings, @{$self->{warnings}});
            $self->{warnings_set} = $ClientUtil::FALSE;
            @{$self->{warnings}} = ();
        }
    }

    # Make sure we didn't get a takeover.
    # Passing in false to ensure we are not detecting takeover condition.
    if ($self->verify_cfo_status($ClientUtil::FALSE) == $ClientUtil::FALSE or
        $self->error) {

        # Returning false because regardles of warn or error we return false.
        $return_val = $ClientUtil::FALSE;

        if ($self->error) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }

        if ($self->{warnings_set} == $ClientUtil::TRUE) {

            # Collect any warnings we have and clear flags
            push (@warnings, @{$self->{warnings}});
            $self->{warnings_set} = $ClientUtil::FALSE;
            @{$self->{warnings}} = ();
        }
    }

    # Verify the status of the disks
    if ($self->validate_metrocluster_fault_disks() == $ClientUtil::FALSE or
        $self->error) {

        # Returning false because regardles of warn or error we return false.
        $return_val = $ClientUtil::FALSE;

        if ($self->error) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }

        if ($self->{warnings_set} == $ClientUtil::TRUE) {

            # Collect any warnings we have and clear flags
            push (@warnings, @{$self->{warnings}});
            $self->{warnings_set} = $ClientUtil::FALSE;
            @{$self->{warnings}} = ();
        }
    }

    # Verify the status of the aggrs and nothing should be degraded
    if ($self->validate_metrocluster_fault_data($ClientUtil::FALSE) == $ClientUtil::FALSE or
        $self->error) {

        # Returning false because regardles of warn or error we return false.
        $return_val = $ClientUtil::FALSE;

        if ($self->error) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }

        if ($self->{warnings_set} == $ClientUtil::TRUE) {

            # Collect any warnings we have and clear flags
            push (@warnings, @{$self->{warnings}});
            $self->{warnings_set} = $ClientUtil::FALSE;
            @{$self->{warnings}} = ();
        }
    }

    if (scalar(@warnings)) {
        @{$self->{warnings}} = @warnings;
        $self->{warnings_set} = $ClientUtil::TRUE;
    }

    if (scalar(@errors)) {
        $self->_set_err("Found following errors while checking configuration" .
                        " after fault:\n" . join ("", @errors));
    }

    return $return_val;
}

################################################################################
# Method:      create_updated_metrocluter_object
# Objective:   Collect all the info to create an updated object.
# Details:     Call all the collect info methods to see the system status and
#              store it in current_obj.
################################################################################
sub create_updated_metrocluter_object {
    my ($self) = @_;
    my (@error);

    # Create a temp object that will contain the current data.
    $self->{current_obj} = SanMCUtil->new(
                                          -controllers => \@{$self->{controllers}},
                                          -fabric1_switches => \@{$self->{fabric1_switches_raw}},
                                          -fabric2_switches => \@{$self->{fabric2_switches_raw}},
                                          -bridges => \@{$self->{bridges}},
                                          -debug_level => $self->get_debug_level(),
                                          -hw_based_own => $self->{HW_BASED_OWNERSHIP},
                                          -error_level => $self->{ERROR_LEVEL},
                                          -drives_to_skip => $self->{DRIVES_TO_SKIP},
                                          -aggrs_to_skip => $self->{AGGRS_TO_SKIP}
                                         );

    if ($self->{current_obj}->error) {
        $self->_set_err("Could not create fault verification mc object: (".
                $self->{current_obj}->errmsg() . ")");

        return $ClientUtil::FALSE;
    }

    # call collect_bridge_info to collect the needed information from
    # the bridges. This populates the object with bridge information
    $self->{current_obj}->collect_bridge_info();
    if ($self->{current_obj}->error) {
        @error = $self->{current_obj}->errmsg();
        $self->_set_err("Error calling collect_bridge_info(): @error");
    }

    # call collect_controller_info to collect the needed information from
    # the controllers. This populates the object with controller information
    $self->{current_obj}->collect_controller_info();
    if ($self->{current_obj}->error) {
        @error = $self->{current_obj}->errmsg();
        $self->_set_err("Error calling collect_controller_info(): @error");
    }

    # call collect_disk_info to collect the needed information from
    # the controllers. This populates the object with controller information
    $self->{current_obj}->collect_disk_info();
    if ($self->{current_obj}->error) {
        @error = $self->{current_obj}->errmsg();
        $self->_set_err("Error calling collect_disk_info(): @error");
    }

    # call collect_switch_info to collect the needed information from
    # the controllers. This populates the object with controller information
    $self->{current_obj}->collect_switch_info();
    if ($self->{current_obj}->error) {
        @error = $self->{current_obj}->errmsg();
        $self->_set_err("Error calling collect_switch_info(): @error");
    }

    # call collect_switch_info merge the collected data that needs multiple
    # hash data privously collected
    $self->{current_obj}->merge_collected_data();
    if ($self->{current_obj}->error) {
        @error = $self->{current_obj}->errmsg();
        $self->_set_err("Error calling collect_switch_info(): @error");
    }

    if (scalar (@error)) {
        $self->_set_err("Unable to create current data object successfully:\n". join("",@error));
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      validate_metrocluster_fault_disks()
# Objective:   Validates the metrocluster disk env. after a fault is injected
# Details:     This functions validates to see if the metrocluster configuration
#              is valid and consistent state of the disks is maintained, after
#              injecting each fault. Make sure we didn't lose any disks, no new
#              disks were found, no disks were failed, no missing secondary
#              ports.
################################################################################
sub validate_metrocluster_fault_disks {
    my ($self) = @_;

    my (@errors, @warnings);
    my (@errored_disks);
    my $return_val = $ClientUtil::TRUE;

    my $current_obj = $self->{current_obj};

    # Disk hashes before the fault and after
    my $dhash_b = $self->{storage_hash}->{disks};
    my $dhash_a = $current_obj->{storage_hash}->{disks};

    foreach my $disk (keys %{$dhash_b}) {
        # Make sure the disk exists before and after
        my $current_errors = scalar (@errors);
        my $owner = $dhash_b->{$disk}->{'percieved-owner'};

        if (!exists ($dhash_a->{$disk})) {
            push (@errors, "\t\t" . $dhash_b->{$disk}->{'primary-name'} .
                    "($disk) doesn't exists after the fault.\n");
        } else {
            # If the disks exists after the test. Now check the attributes.
            # Is the disk failed now?
            if (($dhash_a->{$disk}->{'is-failed'} == $ClientUtil::TRUE) and
                 ($dhash_b->{$disk}->{'is-failed'} != $ClientUtil::TRUE)) {
                # If the disk was not failed before but is after then we have an issue.
                push (@errors, "\t\t" . $dhash_b->{$disk}->{'primary-name'} .
                        " failed during the test.\n");
            }

            if (($dhash_a->{$disk}->{'secondary-port'} !~ /A|B/ and
                 $dhash_b->{$disk}->{'secondary-port'} =~ /A|B/)
                or
                ($dhash_a->{$disk}->{'port'} eq
                 $dhash_a->{$disk}->{'secondary-port'})
                or
                ($dhash_a->{$disk}->{'primary-name'} eq
                 $dhash_a->{$disk}->{'secondary-name'})) {

                # Is the disk missing it's secondary port?
                # if the status was the same before the make it a warning.
                if (($dhash_b->{$disk}->{'port'} eq
                     $dhash_b->{$disk}->{'secondary-port'})
                    or
                    ($dhash_b->{$disk}->{'primary-name'} eq
                     $dhash_b->{$disk}->{'secondary-name'})) {
                    my $str = "\t\t" . $dhash_b->{$disk}->{'primary-name'} .
                              "($disk) still missing its secondary path.\n";
                    &SanUtil::nate_log('-warn' , "$str");
                } else {
                    my $str =
                          "\t\t" . $dhash_b->{$disk}->{'primary-name'} .
                          "($disk) lost its redundancy after the fault.\n " .
                          "\t\t\tThe disk status before:\n".
                          "\t\t\t\tPrimary: " .
                          $dhash_b->{$disk}->{'primary-name'} .
                          "\n\t\t\t\tESH: " .
                          $dhash_b->{$disk}->{'port'} .
                          "\n\t\t\t\tSecondary: " .
                          $dhash_b->{$disk}->{'secondary-name'} .
                          "\n\t\t\t\tESH: " .
                          $dhash_b->{$disk}->{'secondary-port'} .
                          "\n\t\t\tThe disk status after:\n".
                          "\t\t\t\tPrimary: " .
                          $dhash_a->{$disk}->{'primary-name'} .
                          "\n\t\t\t\tESH: " .
                          $dhash_a->{$disk}->{'port'} .
                          "\n\t\t\t\tSecondary: NOT FOUND\n";
                    push (@errors, $str);
                }
            }
        }
        if ($current_errors != scalar (@errors)) {
            push (@errored_disks, $disk);
        }
    }

    $self->_validate_disk_acknowledgements_at_all_layers($dhash_a, $dhash_b);
    # Merge errors, warnings will automatically get merged
    if ($self->error) {
        push (@errors, "\n\t");
        push (@errors, $self->errmsg());
        $self->_clear_err();
    }

    # Checking the storage related states
    $self->_validate_storage_states($dhash_a, $dhash_b);
    if ($self->error) {
        push (@errors, "\n\t" . $self->errmsg());
        $self->_clear_err();
        $return_val = $ClientUtil::FALSE;
    }

    my $storage_hash = $current_obj->{storage_hash};
    $self->_validate_pool_mappings($storage_hash);
    # Merge errors, warnings will automatically get merged
    if ($self->error) {
        push (@errors, "\n\t" . $self->errmsg());
        $self->_clear_err();
    }

    # Are there minions in lab inserting new disks into the setup?
    foreach my $disk (keys %{$dhash_a}) {
        if (!exists ($dhash_a->{$disk})) {
            push (@errors, "\t\t" . $dhash_a->{$disk}->{'primary-name'} .
                  "($disk) didn't exists in the system before the fault" .
                  " but was found in the system after the fault.\n");
            push (@errored_disks, $disk);
        }
    }

    if (scalar @errors) {
        $self->_set_err("Found errors validating disks:\n", join ("",@errors));
        $return_val = $ClientUtil::FALSE;
    }
    return $return_val;
}

################################################################################
# Method:      validate_metrocluster_fault_data()
# Objective:   Validates the metrocluster environment after a fault is injected
# Details:     This functions validates to see if the aggrs, vols, luns, etc in
#              the config are in acceptable state on par with what was seen at
#              start of the fault. The input bool parameter is to identify the
#              cases in which we would expected to see degraged mirror.
################################################################################
sub validate_metrocluster_fault_data {
    my ($self, %args) = @_;
    my ($expecting_degraded, $node_in_disaster, $fsid_option);
    my (@errors, @warnings);
    my $return_val = $ClientUtil::TRUE;
    my $current_obj = $self->{current_obj};

    $expecting_degraded = $ClientUtil::FALSE;
    $node_in_disaster = "";
    $fsid_option = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?expecting_degraded/) {
            $expecting_degraded = $args{$_};
        } elsif (/-?node_in_disaster/) {
            $node_in_disaster = $args{$_};
        } elsif (/-?fsid_option/) {
            $fsid_option = $args{$_};
        }
    }

    foreach my $controller (keys %{$self->{controller_hash}}) {

        my $in_cfod = $ClientUtil::FALSE;

        # We we were passed in a node that is in cfod then we need
        # handle it slightly differently
        if ($controller eq $node_in_disaster) {
            $in_cfod = $ClientUtil::TRUE;
        }

        # Check each of the aggregates hashes before and after
        my $aggr_hash_b = $self->{storage_hash}->{aggrs}->{$controller};
        my $aggr_hash_a = $current_obj->{storage_hash}->{aggrs}->{$controller};

        # Walk through the current aggr has and look for issues
        foreach my $aggr (keys %{$aggr_hash_b}) {

            # Looking at the mirror status at the start of the test and
            # comparing with the current status. And if they are the same
            # status as the start make sure the 'state' condition are the
            # same also.
            if ($aggr_hash_b->{$aggr}->{'start-mirror-status'} eq
                $aggr_hash_a->{$aggr}->{'mirror-status'}) {

                # If we are in CFOD then we shouldn't have our aggregates
                # mirrored.
                if ($in_cfod == $ClientUtil::TRUE) {
                    if ($aggr_hash_a->{$aggr}->{'mirror-status'} =~ /mirrored/i) {
                        push (@errors, "During CFOD, we shouldn't have any" .
                              " aggregates mirrored. The Aggr \'$aggr\' was" .
                              " found mirrored");
                    }
                }

                my $errors_found = 0;
                # Make sure aggr state matches what is was before
                if (($aggr_hash_b->{$aggr}->{'is-offline-at-start'} == $ClientUtil::TRUE) and
                    !($aggr_hash_a->{$aggr}->{'state'} =~ /offline/)) {
                    $errors_found = 1;
                } elsif (($aggr_hash_b->{$aggr}->{'is-failed-at-start'} == $ClientUtil::TRUE) and
                         !($aggr_hash_a->{$aggr}->{'state'} =~ /failed/)) {
                    $errors_found = 1;
                } elsif (($aggr_hash_b->{$aggr}->{'is-partial-at-start'} == $ClientUtil::TRUE) and
                         !($aggr_hash_a->{$aggr}->{'state'} =~ /partial/)) {
                    $errors_found = 1;
                } else {
                    # It must have been online. Other states shouldn't be here
                    if (!$aggr_hash_a->{$aggr}->{'state'} =~ /online/) {
                        $return_val = $ClientUtil::FALSE;
                        push (@errors, "Was expecting aggr online but found" .
                            " in state \"" . $aggr_hash_a->{$aggr}->{'state'} .
                            "\"\n");
                    }
                }

                if ($errors_found) {
                    $return_val = $ClientUtil::FALSE;
                    push(@errors, "Aggr \'$aggr\'\'s state changed before" .
                        "\"" . $aggr_hash_b->{$aggr}->{'state'} . "\"" .
                        " and after \"" . $aggr_hash_b->{$aggr}->{'state'} .
                        "\"\n");
                }

            } else {

                # Were we expecting degraded aggrs? CFOD, etc.
                if ($in_cfod or
                   (($expecting_degraded) and
                    ($aggr_hash_a->{$aggr}->{'mirror-status'} eq 'mirror degraded'))) {

                    # We are expecting a different state from the intial
                    if ($in_cfod) {
                        # TODO

                    }
                } else {
                    $return_val = $ClientUtil::FALSE;
                    push (@errors, "Unexpected change in mirror status, the " .
                                    "status for $aggr on $controller before" .
                                    " was \"" .
                                    $aggr_hash_b->{$aggr}->{'mirror-status'} .
                                    "\" and after \"" .
                                    $aggr_hash_a->{$aggr}->{'mirror-status'} .
                                    "\".\n");
                }
            }
        }
    }

    if (scalar @errors) {
        $self->_set_err("Found following issues with verifying data(vols,aggrs,etc):\n",
                        join ("",@errors));
    }
    return $return_val;
}

################################################################################
# Method:      verify_ic_status()
# Objective:   Validates that the ic_status are RVs are correctly configured
# Details:     This functions validates to see if the interconnect status is
#              in the normal state.
################################################################################
sub verify_ic_status {
    my ($self) = @_;
    my (@errors, @warnings);

    foreach my $controller (keys %{$self->{controller_hash}}) {
        my $cmd = "ic status";

        # Use system_zapi to get the output of the command and store it in @cmd_output
        my ($output, $status, @cmd_output) = &SanUtil::system_zapi($controller, $cmd, "diag");
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
            return $ClientUtil::FALSE;
        }

        my $link0_up = 0;
        my $link1_up = 0;
        my $rv_nic_link = "";
        my $rv1_nic_link = "";
        my $rv2_nic_link = "";
        my $cfo_rv_up = "";
        my $cfo_rv1_up = "";
        my $cfo_rv2_up = "";
        my $cfo_rv_role = "";
        my $cfo_rv1_role = "";
        my $cfo_rv2_role = "";

        # Sample data that we will be parsing through
        #
        # Link 0: up
        # Link 1: down
        # cfo_rv connection state :           CONNECTED
        # cfo_rv nic used :               0
        # cfo_rv2 connection state :          NOT CONNECTED
        # cfo_rv2 nic used :              1

        # Sample data that we will be parsing through
        # Link 0: up
        # Link 1: up
        # cfo_rv connection state :           CONNECTED
        # cfo_rv nic used :               0
        # cfo_rv connection role :            source_rv
        # cfo_rv1 connection state :          CONNECTED
        # cfo_rv1 nic used :              1
        # cfo_rv1 connection role :           dest_rv
        # cfo_rv2 connection state :          CONNECTED
        # cfo_rv2 nic used :              1
        # cfo_rv2 connection role :           Not Applicable

        foreach my $line (@cmd_output) {
            my @nic_status = split (/:/,$line);
            $nic_status[1] =~ s/^\s+//;
            $nic_status[1] =~ s/\s+$//;

            # We have to check for each line of the ic status command
            if ($nic_status[0] =~ /Link 0/ and $nic_status[1] =~ /up/) {
                # link 0 is up
                $link0_up = 1;
            } elsif ($nic_status[0] =~ /Link 1/ and $nic_status[1] =~ /up/) {
                # link 1 is up
                $link1_up = 1;
            } elsif ($nic_status[0] =~ /cfo_rv\snic/ and $nic_status[1] !~ /0/) {
                # cfo_rv nic is not connected to it's default nic 0
                $rv_nic_link = $nic_status[1];
            } elsif ($nic_status[0] =~ /cfo_rv1\snic/ and $nic_status[1] !~ /1/) {
                # cfo_rc nic is not connected to it's default nic 1
                $rv1_nic_link = $nic_status[1];
            } elsif ($nic_status[0] =~ /cfo_rv2\snic/ and $nic_status[1] !~ /1/) {
                # cfo_rv2 nic is not connected to it's default nic 1
                $rv2_nic_link = $nic_status[1];
            } elsif ($line =~/cfo_rv\sconnection\sstate/ and $nic_status[1] !~ /^CONNECTED$/) {
                # cfo_rv connection state is not as expected.
                $cfo_rv_up = $nic_status[1];
            } elsif ($line =~/cfo_rv1\sconnection\sstate/ and $nic_status[1] !~ /^CONNECTED$/) {
                # cfo_rv1 connection state is not as expected.
                $cfo_rv_up = $nic_status[1];
            } elsif ($line =~/cfo_rv2\sconnection\sstate/ and $nic_status[1] !~ /^CONNECTED$/) {
                # cfo_rv2 connection state is not as expected.
                $cfo_rv2_up = $nic_status[1];
            }
        }

        if ($cfo_rv_up ne "") {
            push (@errors, "On $controller, cfo_rv connection is not up, " .
                  "current state ($cfo_rv_up).\n");
        }

        if ($rv_nic_link ne "") {
            push (@warnings, "On $controller, cfo_rv is connected through nic " .
                  "$rv_nic_link, rather than the default nic 0.\n");
        }

        if ($rv1_nic_link ne "") {
            push (@warnings, "On $controller, cfo_rv is connected through nic " .
                  "$rv_nic_link, rather than the default nic 1.\n");
        }

        if ($link0_up == 0) {
            push (@errors, "On $controller, IC nic 0 is down.\n");
        }

        if ($cfo_rv1_up ne "") {
            push (@errors, "On $controller, cfo_rv1 connection is not up, " .
                  "current state ($cfo_rv1_up).\n");
        }

        if ($cfo_rv2_up ne "") {
            push (@errors, "On $controller, cfo_rv2 connection is not up, " .
                  "current state ($cfo_rv2_up).\n");
        }

        if ($rv2_nic_link ne "") {
            push (@warnings, "On $controller, cfo_rv2 is connected through nic ".
                  "$rv2_nic_link, rather than the default nic 1.\n");
        }

        if ($link1_up == 0) {
            push (@errors, "On $controller, IC nic 1 is down.\n");
        }
    }

    if ($self->{ERROR_LEVEL} >= 1) {
        @warnings = (@warnings, @errors);
        @errors = ();
    }

    if (scalar (@warnings)) {
        push (@{$self->{warnings}}, @warnings);
        $self->{warnings_set} = $ClientUtil::TRUE;
    }

    if (scalar (@errors)) {
        $self->_set_err("Found following errors with \"ic status\":\n\t" .
                        join ("\t",@errors));
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      verify_cfo_status()
# Objective:   Validates the cfo status
# Param:       $takeover_occured - bool - idenify if we are checking
#              to see if a takeover has happened
# Details:     This functions validates to see if cf status is in an acceptable
#              state based on the input parameter. If it is true then we will
#              check to see in takeover mode other not in takeover and cf
#              enabled.
################################################################################
sub verify_cfo_status {

    my ($self, $takeover_occured) = @_;

    foreach my $controller (keys %{$self->{controller_hash}}) {

        # Issue a cf status to system to see what it looks like for logs
        my $cmd = "cf monitor all";

        # Use system_zapi to get the output of the command
        my ($output, $status);
        ($output, $status, undef) = &SanUtil::system_zapi($controller, $cmd, "diag");
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
        }

        $cmd = "cf status";

        # Use system_zapi to get the output of the command
        ($output, $status, undef) = &SanUtil::system_zapi($controller, $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
        }

        # Get cf-status information from the zapi
        $cmd = "cf-status";

        # Use zapi to get the output of the command and store it in $output
        ($output, $status) = &SanUtil::cmds_driver_zapi($controller,
                                                           $cmd,
                                                           "diag");
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute ZAPI ($cmd)");
            return $ClientUtil::FALSE;
        }

        # Are we checking to see if takeover has happened or not.
        if ($takeover_occured == $ClientUtil::TRUE) {

            # If cf is enabled then make sure it is in takeover.
            if (($output->{'is-enabled'} eq 'true') and
                ($output->{'state'} !~ /TAKEN OVER|TAKEOVER/i)) {

                $self->_set_err("The system is not in the expected takeover" .
                                " state, current state (" .
                                $output->{'state'} . ")");
                return $ClientUtil::FALSE;
            } else {
                # We are in Takeover or taken over. So we're good.
                return $ClientUtil::TRUE;
            }

        } else {

            # Check if clustering is enabled and not in takeover or
            # intermediate states (TAKEOVER, TAKEOVER STARTED, TAKEN OVER...)
            if (($output->{'is-enabled'} !~ /true/i) or
                ($output->{'state'} ne 'CONNECTED')) {

                $self->_set_err("CFO not enabled or in unexpected state (".
                                $output->{'state'} . ").");
                return $ClientUtil::FALSE;
            } else {
                # Everything is connected and cf is enabled
                return $ClientUtil::TRUE;
            }
        }
    }

    return $ClientUtil::FALSE;
}

################################################################################
# Method:      remirror_unsync_aggrs()
# Objective:   Remirror aggregates split up during CFOD.
# Details:     This functions resync the aggregates that were split during CFOD.
################################################################################
sub remirror_unsync_aggrs {
    my ($self, $fault_name, $disaster_node, $takeover_node) = @_;

    # Get the updated disk and aggr hash.
    my $storage_hash_cur = $self->get_disk_info();
    my $mirrored_aggr;
    if ($self->error){
        return $ClientUtil::FALSE;
    }
    my $storage_hash_old = $self->{storage_hash};


    $self->sanlog(-msg_level => 3,
                  -msg =>  Dumper ($storage_hash_old->{aggrs}));

    # Find out matching aggregates pairs on each node and ignore if already
    # resyncing.

    foreach my $controller (keys %{$self->{controller_hash}}) {
        # The takeover should should have it's aggrs resyncing by now. If not
        # warn and continue resyncing
        my $aggr_hash_a = $storage_hash_cur->{aggrs}->{$controller};
        my $aggr_hash_b = $storage_hash_old->{aggrs}->{$controller};

        foreach my $aggr (keys %{$aggr_hash_a}) {

            if ($controller eq $takeover_node) {
                # See if the state is the same as before
                if ($aggr_hash_b->{$aggr}->{'start-mirror-status'} eq
                    $aggr_hash_a->{$aggr}->{'mirror-status'}) {
                   # Aggregate was unmirrored ???
                } elsif ($aggr_hash_a->{$aggr}->{'mirror-status'} =~ /resynchronizing/) {
                    # We are resyncing already good
                    $self->{cur_data}->{resync_aggrs}->{$controller}->{$aggr} = "resync started";
                    next;
                } elsif ($aggr_hash_a->{$aggr}->{'mirror-status'} =~ /degraded/) {
                    $mirrored_aggr = $self->_mirror_aggrs($aggr, $aggr_hash_a, $controller);
                    if ($self->error) {
                        return $ClientUtil::FALSE;
                    } else {
                        # We were successful in resyncing now add to list
                        $self->{cur_data}->{resync_aggrs}->{$controller}->{$mirrored_aggr} = "resync started";
                    }
                }
            } else {
                # This is the TAKEN over node
                if ($aggr_hash_b->{$aggr}->{'start-mirror-status'} eq
                    $aggr_hash_a->{$aggr}->{'mirror-status'}) {
                    # We have unmirrored data????
                } else {
#TODO check for unmirrored /failed aggrs
                    # Only look at aggregates that existed before, new aggregates will mostly
                    # be victims
                    if (exists $aggr_hash_b->{$aggr}->{'state'}) {
                        &SanUtil::nate_log("-comment", "We'll need to mirror aggregate $aggr.");

                        # Could it be already resyncing?
                        if (exists $aggr_hash_a->{$aggr}->{'mirror-status'} and
                            $aggr_hash_a->{$aggr}->{'mirror-status'} =~ /resynchronizing/) {
                            $mirrored_aggr = $aggr;
                        } else {
                            $mirrored_aggr = $self->_mirror_aggrs($aggr, $aggr_hash_a, $controller);
                        }

                        if ($self->error) {
                            return $ClientUtil::FALSE;
                        } else {
                            # We were successful in resyncing now add to list
                            $self->{cur_data}->{resync_aggrs}->{$controller}->{$mirrored_aggr} =
                                "resync started";
                        }
                    }
                }
            }
        }
    }

    $self->_wait_resyncing_to_finish();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      wait_resyncing_aggr
# Objective:   Wait for all resyncing aggrs on all nodes to finish.
################################################################################
sub wait_resyncing_aggr {
    my ($self) = @_;

    # Get the updated disk and aggr hash.
    my $storage_hash_cur = $self->get_disk_info();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    foreach my $controller (keys %{$self->{controller_hash}}) {
        # The takeover should should have it's aggrs resyncing by now. If not
        # warn and continue resyncing
        my $aggr_hash = $storage_hash_cur->{aggrs}->{$controller};

        foreach my $aggr (keys %{$aggr_hash}) {
            if ($aggr_hash->{$aggr}->{'mirror-status'} =~ /resynchronizing/i) {
                $self->{cur_data}->{resync_aggrs}->{$controller}->{$aggr} =
                    "resyncing";
            }
        }
    }

    $self->_wait_resyncing_to_finish();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      cf_giveback()
# Objective:   Do the giveback, if normal giveback fails do a forced giveback.
################################################################################
sub cf_giveback() {
    my ($self, $fault_name, $disaster_node, $takeover_node) = @_;

    &SanUtil::nate_log('-comment' , "Trying normal giveback for $disaster_node on $takeover_node.");
    my $cmd = 'cf-giveback';
    my $args;
    my ($output, $status) = &SanUtil::cmds_driver_zapi($takeover_node,
                                                       $cmd,
                                                       "diag");
    if ($status == SanUtil::ZAPI_FAILURE) {

        $args->{-force} = 'true';
        &SanUtil::nate_log('-warn' , "Normal giveback failed for test $fault_name. " .
                           "Trying forced giveback for $disaster_node on $takeover_node.");
        ($output, $status) = &SanUtil::cmds_driver_zapi($takeover_node,
                                                        $cmd,
                                                        "diag", "",
                                                        $args);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute giveback successfully.");
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      verify_data()
# Objective:   Make sure all the volumes that we expected to stay online are.
# TODO need to add code that make sure the volumes are mounted and accessible
# and luns stay online if it is not a FSID cfod etc. %args the inputs.
################################################################################
sub verify_data {
    my ($self, $after_cfod, $fsid_option) = @_;

    # Get the updated disk and aggr hash.
    my $storage_hash_cur = $self->get_disk_info();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

   return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: check_if_disaster_takeover_conditions_met
# Objective: Check if we are reading to send "cf forcetakeover -d".
###############################################################################
sub check_if_disaster_takeover_conditions_met {
    my ($self, %args) = @_;

    my $timeout = "";
    my $takeover_node = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?timeout/) {
            $timeout = $args{$_};
            next;
        } elsif (/-?takeover_node/) {
            $takeover_node = $args{$_};
            next;
        }
    }

    $timeout = 300 if ($timeout eq "");

    if ($takeover_node eq "") {
        $self->_set_err("Invalid takeover node.");
        return $ClientUtil::FALSE;
    }

    # Detect when it is safe to do a CFOD. We we are cable of doing CFOD,
    # 'cf status' command will say something about 'cf forcetakeover -d'
    my $time = 0;

    while ($time < $timeout) {
        my $cmd = "cf status";
        my ($output, $status) = &SanUtil::system_zapi($takeover_node, $cmd);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute system-cli ZAPI ($cmd)");
            return $ClientUtil::FALSE;
        }

        if ($output =~ /cf forcetakeover -d/) {
            &SanUtil::nate_log('-comment' , "\'cf status\' says we are ready" .
                    " for takeover:\n $output.");
            last;
        }

        &SanUtil::nate_sleep(20);
        $time += 20;

        # We've run out of time fail the test now.
        if ($time >= $timeout) {
            $self->_set_err("Waited $timeout " .
                    "seconds for \'cf status\' to come back and " .
                    "say we are ready for \'cf forcetakeover -d\'" .
                    " last output from \'cf status\':\n $output");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: disaster_takeover
# Objective: Send the equivalent of "cf forcetakeover -d" via zapi.
###############################################################################
sub disaster_takeover {
    my ($self, %args) = @_;

    my $takeover_node = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?takeover_node/) {
            $takeover_node = $args{$_};
            next;
        }
    }

    if ($takeover_node eq "") {
        $self->_set_err("Invalid takeover node.");
        return $ClientUtil::FALSE;
    }

    &SanUtil::nate_log('-comment', "Issuing CFOD on $takeover_node.");
    my $cmd = 'cf-force-takeover';
    my $args;
    $args->{'-disaster'} = 'true';
    my ($output, $status) = &SanUtil::cmds_driver_zapi($takeover_node,
                                                       $cmd,
                                                       "diag", "",
                                                       $args);

    if ($status == SanUtil::ZAPI_FAILURE) {
        $self->_set_err("Failed to execute ZAPI ($cmd)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: get_sysconfig_data()
# Objective: Take the data retrieved when we ran sysconfig -v and create a
#            hash that can be used to discover port types.
###############################################################################
sub get_sysconfig_data {
    my ($self) = @_;

    my ($port_hash, $wwpn);

    ## Check to see if the controller_hash is already populated.
    ## Error out if it hasn't because we need that info here.
    if (not exists $self->{controller_hash}) {
        $self->_set_err("get_switch_info run before controller_hash exists");
        return $ClientUtil::FALSE;
    }

    foreach my $controller (keys %{$self->{controller_hash}}) {
        foreach my $adapter (keys %{$self->{controller_hash}->{$controller}->{adapters}->{fcvi}}) {
            next if ($adapter =~ /warnings|disks|shelves/i);
            $wwpn = $self->{controller_hash}->{$controller}->{adapters}->{fcvi}->{$adapter}->{wwpn};
            $port_hash->{$wwpn} = 'fcvi';
        }
        foreach my $adapter (keys %{$self->{controller_hash}->{$controller}->{adapters}->{fc}}) {
            next if ($adapter =~ /warnings|disks|shelves/i);
            $wwpn = $self->{controller_hash}->{$controller}->{adapters}->{fc}->{$adapter}->{wwpn};
            $port_hash->{$wwpn} = 'fport';
        }
    }

    &SanUtil::nate_log('comment', "Sysconfig Port Hash: " . Dumper($port_hash));
    return $port_hash;
}

################################################################################
############################ Private Functions #################################
################################################################################

################################################################################
# Method:      _get_module_ports
# Objective:   Wait for all resyncing aggrs on all nodes to finish.
################################################################################
sub _get_module_ports {
    my ($self, $audit) = @_;

    my @port_types = ('aport', 'fport', 'fcvi', 'eport');
    my ($module);

    foreach my $switch (keys %{$audit}) {
        foreach my $port_type (@port_types) {
            foreach my $port (keys %{$audit->{$switch}->{$port_type}}) {
                if ($port =~ /^fc(\d+)\//) {
                    $module = $1;
                } else {
                    $self->_set_err("Unsupported switch using modules.\nPort:$port - Matching for fc2/1");
                    return $self;
                }

                $self->{modules}->{$switch}->{$module}->{$port_type}->{$port} =
                    {%{$audit->{$switch}->{$port_type}->{$port}}};
            }
        }
    }
}

################################################################################
# Method:      _wait_resyncing_to_finish
# Objective:   Wait for all resyncing aggrs on all nodes to finish.
################################################################################
sub _wait_resyncing_to_finish {
    my ($self) = @_;

    my $min_counter = 0;
    # Wait for aggrs to finish resyncing
    &SanUtil::nate_log("-comment", "Waiting for all the aggr resynchronizing to finish.");

    while ($self->_aggrs_done_resyncing() != $ClientUtil::TRUE) {
        if ($self->error) {
            return $ClientUtil::FALSE;
        }

        if ($self->{RESYNC_TIMEOUT} > 0) {
            if ($min_counter >= $self->{RESYNC_TIMEOUT}) {
                # We were unable to complete resycing aggregates
                # in the alloted time. Aborting.
                $self->_set_err("Aggregates are still not" .
                                " mirrored after $min_counter minutes" .
                                ". Aborting wait for resyncing aggrs.\n");
                return $ClientUtil::FALSE;
            }
        } elsif ($min_counter > HOURS_12) {
            $self->_set_err("Aggregates are still not" .
                            " mirrored after $min_counter minutes" .
                            ". Aborting wait for resyncing aggrs.\n");
            return $ClientUtil::FALSE;
        }

        &SanUtil::nate_log("-comment", "Sleeping 60 sec and waiting for it to finish resyncing");
        &SanUtil::nate_sleep(60);
        $min_counter++;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      _aggrs_done_resyncing()
# Objective:   Return false until all the aggregates are done resycing.
# Details:     Keep checking $self->{cur_data}->{resync_aggrs} to see if all
#              aggregates there are syncing.
################################################################################
sub _aggrs_done_resyncing {
    my ($self) = @_;

    my $aggrs = $self->{cur_data}->{resync_aggrs};
    foreach my $controller (keys %{$aggrs}) {
        foreach my $aggr (keys %{$aggrs->{$controller}}) {
            my $cmd = "aggr-list-info";
            my $args;
            $args->{-aggregate} = $aggr;
            my ($output, $status) = &SanUtil::cmds_driver_zapi($controller,
                                                               $cmd,
                                                               "diag","",
                                                               $args);
            if ($status == SanUtil::ZAPI_FAILURE) {
                $self->_set_err("Failed to execute ZAPI ($cmd)");
                return $ClientUtil::FALSE;
            }

            # We need to find out see if we have just aggr or more that one.
            if ($output->{'aggregates'}->{'aggr-info'}->{'mirror-status'} ne 'mirrored') {
                &SanUtil::nate_log("-comment", "$aggr on $controller has not finished " .
                                   "resynchronizing, current mirror status: ".
                                   $output->{'aggregates'}->{'aggr-info'}->{'mirror-status'});

                return $ClientUtil::FALSE;
            } elsif ($output->{'aggregates'}->{'aggr-info'}->{'mirror-status'} eq 'mirrored') {
                # If we it is mirrored then remove it from the list to check
                &SanUtil::nate_log("-comment", "$aggr on $controller has finished " .
                                   "resynchronizing, current mirror status: ".
                                   $output->{'aggregates'}->{'aggr-info'}->{'mirror-status'});

                delete $aggrs->{$controller}->{$aggr};
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:      _mirror_aggrs()
# Objective:   Actually do the mirroring after finding victim aggr.
# Details:     Call the "aggr-mirror" zapi with the victim to mirror.
################################################################################
sub _mirror_aggrs {
    my ($self, $aggr_to_mirror, $aggr_hash, $controller) = @_;

    my $mirror_partner = "";

    # Find the matching pair for it to mirror
    foreach my $aggr (keys %{$aggr_hash}) {
        # trying and match by name first, usally we have format
        # aggr0 and it's mirror partner aggr0(1)
        if ($aggr ne $aggr_to_mirror and
            $aggr =~ /$aggr_to_mirror\((\S+)\)/) {
            $mirror_partner = $aggr;
            last;
        }
    }

    if ($mirror_partner ne "") {
        &SanUtil::nate_log("-comment", "Mirroring $aggr_to_mirror to $mirror_partner.");

        my $cmd = "aggr-mirror";
        my $args;
        $args->{"aggregate"} = $aggr_to_mirror;
        $args->{"victim-aggregate"} = $mirror_partner;
        my ($output, $status) = &SanUtil::cmds_driver_zapi($controller,
                                                        $cmd,
                                                        "diag","",
                                                        $args);
        if ($status == SanUtil::ZAPI_FAILURE) {
            $self->_set_err("Failed to execute ZAPI ($cmd)");
            return "";
        }

    } else {
        $self->_set_err("Unable to find victim aggregate for $aggr_to_mirror.");
    }

    return $aggr_to_mirror;
}

################################################################################
# Method:      _validate_storage_states
# Objective:   Validates the shelf states and port states seen from the shelf.
################################################################################
sub _validate_storage_states {
    my ($self, $disk_hash, $disk_hash_old) = @_;

    my (@errors,@warnings);
    my $return_val = $ClientUtil::TRUE;
    my $compare_disk_hashes =
        (not defined $disk_hash_old) ? $ClientUtil::FALSE : $ClientUtil::TRUE;
    my $error_shelves = {};

    foreach my $disk (keys %{$disk_hash}) {
        # Look through each of the shelf info and check the shelf state.
        foreach my $module (keys %{$disk_hash->{$disk}->{ses_info}}) {
            my $ses_info = $disk_hash->{$disk}->{ses_info}->{$module};
            my $ses_info_old = $disk_hash_old->{$disk}->{ses_info}->{$module};
            if ($ses_info->{'shelf-state'} !~ /online/i) {
                if ($compare_disk_hashes == $ClientUtil::TRUE and
                    ($ses_info->{'shelf-state'} ==
                     $ses_info_old->{'shelf-state'})) {
                    # Ignore we are in the same state as before
                } else {
                    if (exists $error_shelves->{$ses_info->{'shelf-name'}}) {
                        # Let's not spam the logs for each of the disk attached
                        # to the shelf;
                        next;
                    }

                    my $str = "Shelf $ses_info->{'shelf-name'} is in an " .
                              "unexpected state $ses_info->{'shelf-state'}.\n";

                    if ($self->{ERROR_LEVEL} >= 4) {
                        push (@warnings, $str);
                    } else {
                        push (@errors, $str);
                    }
                    $error_shelves->{$ses_info->{'shelf-name'}} = $ClientUtil::TRUE;
                }
            }

            # Check the port states on the shelf to see if anything is bad...
            if ($ses_info->{'port-state'} !~ /ok/i) {

                if ($compare_disk_hashes == $ClientUtil::TRUE and
                    ($ses_info->{'port-state'} ==
                     $ses_info_old->{'port-state'})) {
                    # Ignore we are in the same state as before
                } else {
                    my $str = "SES reports an invalid state (".
                                $ses_info_old->{'port-state'} .
                                ") for disk ($disk).\n" ;
                    if ($self->{ERROR_LEVEL} >= 4) {
                        push (@warnings, $str);
                    } else {
                        push (@errors, $str);
                    }
                    # Ignore all other ports for this disk
                    last;
                }
            }

             #TODO add crc errors etc.
        }
    }

    if (scalar(@warnings)) {
        push (@{$self->{warnings}}, @warnings);
        $self->{warnings_set} = $ClientUtil::TRUE;
        $return_val = $ClientUtil::FALSE;
    }

    if (scalar(@errors)) {
        $self->_set_err("Found acknowledgements errors:\n" . join ("\t", @errors));
        $return_val = $ClientUtil::FALSE;
    }
    return $return_val;
}

################################################################################
# Method:      _validate_disk_acknowledgements_at_all_layers
# Objective:   Validates if the disk is see at all the layers: RAID, storage,
#              sanown, and the switches.
################################################################################
sub _validate_disk_acknowledgements_at_all_layers {
    my ($self, $disk_hash, $disk_hash_old) = @_;

    my (@errors,@warnings);
    my $return_val = $ClientUtil::TRUE;
    # Compare the hashes if two of them are given
    my $compare_disk_hashes =
        (not defined $disk_hash_old) ? $ClientUtil::FALSE : $ClientUtil::TRUE;

    foreach my $disk (keys %{$disk_hash}) {
        my $found_on_controller_num = 0;

        foreach my $controller (keys %{$self->{controller_hash}}) {

            my ($ack_by_raid, $ack_by_sanown, $ack_by_storage, $ack_by_switches);
            $ack_by_raid = $ack_by_sanown = $ack_by_storage = $ack_by_switches = $ClientUtil::TRUE;

            # Check the RAID if it is seen and acknowledged by the controller.
            if ((not defined $disk_hash->{$disk}->{'acknowledged-by-raid'}->{$controller}) or
                ($disk_hash->{$disk}->{'acknowledged-by-raid'}->{$controller} != $ClientUtil::TRUE)) {

                # Are we comparing? then check the previous state.
                if ($compare_disk_hashes == $ClientUtil::TRUE and
                    ((not defined $disk_hash_old->{$disk}->{'acknowledged-by-raid'}->{$controller}) or
                     ($disk_hash_old->{$disk}->{'acknowledged-by-raid'}->{$controller} != $ClientUtil::TRUE))) {
                    #Do nothing as it was in the same state in the old hash.
                } elsif ($disk_hash->{$disk}->{'acknowledged-by-raid'}->{$controller} != $ClientUtil::TRUE and
                         $disk_hash->{$disk}->{'is-failed'} == $ClientUtil::TRUE) {
                    # If it is failed by sanown and then we may not see in RAID and will already
                    # show up as a system failed disk
                } else {
                    $ack_by_raid = $ClientUtil::FALSE;
                }
            }

            # Check if sanown has seen this disk and ignore in case of hw-based ownership.
            if ($self->{HW_BASED_OWNERSHIP} == $ClientUtil::FALSE and
                ((not defined $disk_hash->{$disk}->{'acknowledged-by-sanown'}->{$controller}) or
                 ($disk_hash->{$disk}->{'acknowledged-by-sanown'}->{$controller} != $ClientUtil::TRUE))) {
                if ($compare_disk_hashes == $ClientUtil::TRUE and
                    ((not defined $disk_hash_old->{$disk}->{'acknowledged-by-sanown'}->{$controller}) or
                     ($disk_hash_old->{$disk}->{'acknowledged-by-sanown'}->{$controller} != $ClientUtil::TRUE))) {
                    #Do nothing as it was in the same state in the old hash.
                } else {
                    $ack_by_sanown = $ClientUtil::FALSE;
                }
            }

            # Make sure that we've been acknowledged the switch/bridge that we are connected to.
            # We are just checking the number since it must be seen by two switches/bridges.
            my $disk_type = $disk_hash->{$disk}->{'type'};
            if ($disk_type !~ /^FC/){ # This disk is connected via FC-to-SAS bridge...
                if (scalar (keys %{$disk_hash->{$disk}->{'acknowledged-by-switch'}} == MAX_DEVICE_PATHS)) {
                    $ack_by_switches = $ClientUtil::TRUE;
                } else {
                    $ack_by_switches = $ClientUtil::FALSE;
                }
            }
            else { # This disk is connected via DS14...
                if (scalar (keys %{$disk_hash->{$disk}->{'acknowledged-by-switch'}} != MAX_DEVICE_PATHS)) {
                    if ($compare_disk_hashes == $ClientUtil::TRUE and
                        (scalar (keys %{$disk_hash_old->{$disk}->{'acknowledged-by-switch'}}) ==
                            scalar (keys %{$disk_hash->{$disk}->{'acknowledged-by-switch'}}))) {
                        #Do nothing as it was in the same state in the old hash.
                    } else {
                        $ack_by_switches = $ClientUtil::FALSE;
                    }
                }
            }

            # In Storage is it seen in two places, one for each port?
            if ($disk_hash->{$disk}->{'acknowledged-by-storage'} != MAX_DEVICE_PATHS) {
                if ($compare_disk_hashes == $ClientUtil::TRUE and
                    ($disk_hash_old->{$disk}->{'acknowledged-by-storage'} ==
                     $disk_hash->{$disk}->{'acknowledged-by-storage'})) {
                    #Do nothing as it was in the same state in the old hash.
                }
               # elsif ($disk_hash->{$disk}->{'acknowledged-by-storage'} != 0 and
               #          $disk_hash->{$disk}->{'is-failed'} != $ClientUtil::TRUE) {
                    # Failed drives act weirdly if it is seen atleast once we are ok.
               # }
                else {
                    $ack_by_storage = $ClientUtil::FALSE;
                }
            }

            if ($ack_by_raid == $ClientUtil::FALSE or
                $ack_by_sanown == $ClientUtil::FALSE or
                $ack_by_switches == $ClientUtil::FALSE or
                $ack_by_storage == $ClientUtil::FALSE) {

                # Print the output the disk not found at each layer.
                my $str = "\t\tOn controller \"$controller\", found disk " .
                          $disk_hash->{$disk}->{'primary-name'} .
                          " ($disk) not seen at all layers on $controller: " .
                          " Storage layer (" .
                          &_convert_bool_to_str($ack_by_storage, "found", "not found",
                                               $disk_hash->{$disk}->{'acknowledged-by-storage'},
                                               MAX_DEVICE_PATHS) .
                          "), Sanown layer (" .
                          # Ignore in case of HW based ownership
                          (($self->{HW_BASED_OWNERSHIP} == $ClientUtil::TRUE)?
                           "N/A": &_convert_bool_to_str($ack_by_sanown, "found", "not found")) .
                          "), Switch/Bridge Nameservers (" .
                          &_convert_bool_to_str($ack_by_switches, "found", "not found",
                                               scalar (keys %{$disk_hash->{$disk}->{'acknowledged-by-switch'}}),
                                               MAX_DEVICE_PATHS) .
                          "), and RAID layer (" .
                          &_convert_bool_to_str($ack_by_raid, "found", "not found") .
                          ").\n";

                if ($self->{ERROR_LEVEL} >= 3) {
                    push (@warnings, $str);
                } else {
                    push (@errors, $str);
                }
            }
        }
    }
    my $header = "Found acknowledgements errors:\n
                  \tSample Output: XYZ layer (found), XYZ layer (not found [(actual acks/expected acks)])\n";
    if (scalar(@warnings)) {
        push (@{$self->{warnings}}, $header);
        push (@{$self->{warnings}}, @warnings);
        $self->{warnings_set} = $ClientUtil::TRUE;
        $return_val = $ClientUtil::FALSE;
    }

    if (scalar(@errors)) {
        $self->_set_err($header . join ("", @errors));
        $return_val = $ClientUtil::FALSE;
    }
    return $return_val;
}

################################################################################
# Method:      _convert_bool_to_str
# Objective:   Used to return a string version of ture or false depending on
#              the input.
################################################################################
sub _convert_bool_to_str {
    my ($val, $true_str, $false_str, $actual_val, $expected_val) = @_;

    my $append_vals = "";

    if (not defined $true_str) {
        $true_str = "true";
    }

    if (not defined $false_str) {
        $false_str = "false";
    }

    if ((defined $actual_val) and
        (defined $expected_val)) {
        $append_vals = "($actual_val/$expected_val)";
    }

    if ($val == $ClientUtil::TRUE) {
        return $true_str;
    } else {
        return $false_str. $append_vals;
    }
}

################################################################################
# Method:      _str_disk_error
# Objective:   Return a string with all the info regarding the disk.
################################################################################
sub _str_disk_error {
    my ($self, $disk, $storage_hash, $storage_hash_after) = @_;
    my $str = "";

    if (not defined $storage_hash) {
        return "";
    }
    $str = "\t\t". $self->{storage_hash}->{disks}->{$disk}->{'primary-name'} . "($disk)\n";

    # TODO more info needs to be added for the disk
    return $str;
}

################################################################################
# Method:      _validate_pool_mappings
# Objective:   Validates if disks are seen in local and remote switches and
#              it has the correct pool depending to the site.
################################################################################
sub _validate_pool_mappings {
    my ($self, $storage_hash) = @_;
    my (@errors,@warnings);
    my $return_val = $ClientUtil::TRUE;

    # Hash structure @{$info->{pool_info}->{$owner}->{$pool}->{$switch}->{$port}}, $disk
    foreach my $controller (keys %{$storage_hash->{pool_info}}) {
        # Check Pool 0 switch matching the local switches
        foreach my $switch (keys %{$storage_hash->{pool_info}->{$controller}->{0}}) {
            my $found_in_local = 0;
            foreach my $lcl_switch (keys %{$storage_hash->{local_switches}->{$controller}}) {
                if ($lcl_switch =~ /$switch/i) {
                    $found_in_local = 1;
                    last;
                }
            }
            if ($found_in_local == 0) {
                my $str = "Found disk(s) from pool 0 connected to a switch " .
                          "($switch) that is not one of the local switches. " .
                          "The following disk(s) are on the offending list:\n" .
                          "These disks could be assigned to the wrong pool.\n";
                foreach my $err_port (keys %{$storage_hash->{pool_info}->{$controller}->{0}->{$switch}}) {
                    foreach my $err_disk (@{$storage_hash->{pool_info}->{$controller}->{0}->{$switch}->{$err_port}}) {
                        $str .= $self->_str_disk_error($err_disk, $storage_hash);
                    }
                }

                if ($self->{ERROR_LEVEL} >= 5) {
                    push (@warnings, $str);
                } else {
                    push (@errors, $str);
                }
            }
        }

        # Now check pool 1. We'll be looking at remote switches. other_controllers local
        foreach my $switch (keys %{$storage_hash->{pool_info}->{$controller}->{1}}) {
            my $found_in_remote = 0;
            my $other_controller = "";

            # Get the other controller
            foreach my $rmt_ctlr (keys %{$storage_hash->{pool_info}}) {
                if ($controller !~ /$rmt_ctlr/) {
                    $other_controller = $rmt_ctlr;
                    last;
                }
            }

            # other controller not found.
            if ($other_controller eq "") {
                $self->_set_err("Unable to find the other controller\n.");
                return $ClientUtil::FALSE;
            }

            foreach my $rmt_switch (keys %{$storage_hash->{local_switches}->{$other_controller}}) {
                if ($rmt_switch =~ /$switch/i) {
                    $found_in_remote = 1;
                    last;
                }
            }

            if ($found_in_remote == 0) {
                my $str = "Found disk(s) from pool 1 connected to a switch " .
                          "($switch) that is not one of the remote switches. " .
                          "The following disk(s) are on the offending list:\n";
                foreach my $err_port (keys %{$storage_hash->{pool_info}->{$controller}->{1}->{$switch}}) {
                    foreach my $err_disk (@{$storage_hash->{pool_info}->{$controller}->{1}->{$switch}->{$err_port}}) {
                        $str .= $self->_str_disk_error($err_disk, $storage_hash);
                    }
                }

                if ($self->{ERROR_LEVEL} >= 5) {
                    push (@warnings, $str);
                } else {
                    push (@errors, $str);
                }
            }
        }
    }

    if (scalar(@warnings)) {
        push (@{$self->{warnings}}, @warnings);
        $self->{warnings_set} = $ClientUtil::TRUE;
        $return_val = $ClientUtil::FALSE;
    }

    if (scalar(@errors)) {
        $self->_set_err("Found pool errors:\n\t" . join ("\t", @errors));
        $return_val = $ClientUtil::FALSE;
    }
    return $return_val;
}


###############################################################################
# Method:       _get_storage_show_disk_info
# Objective:    Collects the disk information from storage show disk -p
# Details:      The user passes in the controller, disk_name and the keys to
#               find through the command.
#               Currently following keys are supported:
#                   + disk-uid, secondary-name.
###############################################################################
sub _get_storage_show_disk_info {
    my ($self, $controller, $disk_name, @keys) = @_;

    my $disk_info_hash = {};
    foreach my $key (@keys) {
        $disk_info_hash->{$key} = undef;
    }

    my $cmd = "storage show disk -p $disk_name";
    # Use system_zapi to get the output of the command and store it in @cmd_output
    my ($output, $status, @cmd_output) = &SanUtil::system_zapi($controller, $cmd);
    if ($status == SanUtil::ZAPI_FAILURE) {
        $self->_set_err(-msg => "Failed to execute system-cli ZAPI ($cmd)");
        return {};
    }

    foreach my $line (@cmd_output) {
        my ($name, $value) = split (/:/, $line, 2);
        # make sure this is just at the start. remove extra spaces. in{ 'disk_name' }
        $value =~ s/^\s+|\s+$//g; #remove leading and trailing spaces
        if (($name =~ /Disk:/) or
            ($name =~ /Sec Name:/)) {
            if (grep (/secondary-name/, @keys) != 0 and
                $disk_name !~ /$value/i) {
                $disk_info_hash->{'secondary-name'} = $value;
            }
        } elsif ($name =~ /UID:/ and
                 grep (/disk-uid/, @keys) != 0) {
            $disk_info_hash->{'disk-uid'} = $value;
        } else {
            # Ignore
        }
    }

    return $disk_info_hash;
}

=cut
# Code below can be used to see populate the _get_storage_show_disk_info above
# for more output values from storage show disk -p
# Also the code below contains the parsing code for aggr status -r if it needs
# to be used in the future.
            $temp_disk_hash = ();
            foreach my $line (@cmd_output) {
                my ($name, $value) = split (/:/, $line, 2);
# make sure this is just at the start. remove extra spaces. in{ 'disk_name' }
                $value =~ s/^\s+|\s+$//g; #remove leading and trailing spaces
                if ($name =~ /Disk/) {
                        if ($disk_name eq $value){
                            # We have disk that is primary and active
                            $keep = 1;
                        } else {
                            # We don't care about saving other values
                            last;
                        }
                    $temp_disk_hash->{ 'disk_name' } = $value;
                } elsif ($name =~ /Shelf/) {
                    $temp_disk_hash->{ 'shelf' } = $value;
                } elsif ($name =~ /Bay/) {
                    $temp_disk_hash->{ 'bay' } = $value;
                } elsif ($name =~ /RPM/) {
                    $temp_disk_hash->{ 'rpm' } = $value;
                } elsif ($name =~ /Serial/) {
                    $temp_disk_hash->{ 'serial' } = $value;
                } elsif ($name =~ /Vendor/) {
                    $temp_disk_hash->{ 'vendor' } = $value;
                } elsif ($name =~ /Model/) {
                    $temp_disk_hash->{ 'model' } = $value;
                } elsif ($name =~ /Rev/) {
                    $temp_disk_hash->{ 'rev' } = $value;
                } elsif ($name =~ /WWN/) {
                    $temp_disk_hash->{ 'wwn' } = $value;
                } elsif ($name =~ /UID/) {
                    $temp_disk_hash->{ 'uid' } = $value;
                } elsif ($name =~ /Downrev/) {
                    $temp_disk_hash->{ 'downrev' } = $value;
                } elsif ($name =~ /Pri Port/) {
                    $temp_disk_hash->{ 'primary_port' } = $value;
                } elsif ($name =~ /Sec Name/) {
                    $temp_disk_hash->{ 'secondary_port_name' } = $value;
                } elsif ($name =~ /Sec Port/) {
                    $temp_disk_hash->{ 'secondary_port' } = $value;
                } elsif ($name =~ /Power-on Hours/) {
                    $temp_disk_hash->{ 'power_on_hours' } = $value;
                } elsif ($name =~ /Blocks read/) {
                    $temp_disk_hash->{ 'blocks_read' } = $value;
                } elsif ($name =~ /Blocks written/) {
                    $temp_disk_hash->{ 'blocks_written' } = $value;
                } elsif ($name =~ /Time interval/) {
                    $temp_disk_hash->{ 'time_interval' } = $value;
                } elsif ($name =~ /Glist count/) {
                    $temp_disk_hash->{ 'glist_count' } = $value;
                } elsif ($name =~ /Scrub last done/) {
                    $temp_disk_hash->{ 'scrub_last_done' } = $value;
                } elsif ($name =~ /Scrub count/) {
                    $temp_disk_hash->{ 'scrub_count' } = $value;
                } elsif ($name =~ /LIP count/) {
                    $temp_disk_hash->{ 'lip_count' } = $value;
                } elsif ($name =~ /Dynamically qualified/) {
                    $temp_disk_hash->{ 'dynamically_qualified' } = $value;
                } else {
                    # Ignore
                }
            }
            # Parse 'aggr status -r' to collect more information. If we are not keeping,
            # then there is not need to go through this.
            if ($keep) {
                my @aggr_status = @{$self->{controller_hash}->{$controller}->{aggr_status_r}};

                my ($aggr, $plex, $rg, $pool, $spare, $partner);
                $aggr = $plex = $rg = $pool = '';
                $spare = $partner = 0;
                foreach my $line (@aggr_status) {
                    if ($line =~ /\s*Aggregate\s*(\S+)\s*\((\S+)\)?/) {
                        $aggr = $1;
                        $spare = 0;
                        $partner = 0;
                    } elsif ( $line =~ /\s*Plex\s*(\S+)\s*?/) {
                        $plex = $1;
                        my @temp = split(',',$line);
                        foreach my $item (@temp) {
                            if ($item =~ /pool(\S)/) {
                                $pool =  "Pool$1";
                            }
                        }
                        $spare = 0;
                        $partner = 0;
                    } elsif ( $line =~ /\s*RAID\sgroup\s*(\S+)\s*?/) {
                        $rg = $1;
                        $spare = 0;
                        $partner = 0;
                    } elsif ($line =~ /\s*(\S+)\s*spare\sdisks\s*/) {
                        $aggr = $plex = $rg = '';
                        $spare = 1;
                        $partner = 0;
                        $pool = $1;
                    } elsif ($line =~ /\s*Partner\sdisks\s*/) {
                        $aggr = $plex = $rg = '';
                        $partner = 1;
                        $spare = 0;
                    } elsif ($line =~ /\s*(\S+)\s*(\S+)\s*(\S+)\s*\S+\s*\S+\s*(\S+)\s*?/) {
                        if (($1 =~ /RAID/) or ($1 =~ /------/) or ($1 =~ /Spare/)) {
                            # Ignore these
                        } elsif (!$partner) {
                            if ($disk_name eq $2) {
                                if (!$spare) {
                                    $temp_disk_hash->{ 'owner' } = $controller;
                                    $temp_disk_hash->{ 'aggregate' } = $aggr;
                                    $temp_disk_hash->{ 'disk_type' } = $1;
                                    $temp_disk_hash->{ 'raid_group' } = $rg;
                                    $temp_disk_hash->{ 'plex' } = $plex;
                                    $temp_disk_hash->{ 'pool' } = $pool;
                                } else {
                                    $temp_disk_hash->{ 'owner' } = $controller;
                                    $temp_disk_hash->{ 'aggregate' } = '';
                                    $temp_disk_hash->{ 'disk_type' } = 'spare';
                                    $temp_disk_hash->{ 'raid_group' } = '';
                                    $temp_disk_hash->{ 'plex' } = '';
                                    $temp_disk_hash->{ 'pool' } = $pool;
                                }
                            }
                        }
                    }
                }
            }

            if ($keep) {
                $keep = 0;
                while (my($key, $val) = each %$temp_disk_hash) {
                   $hash->{$controller}->{active_disks}->{$disk_name}->{$key} = $val;
                }
            }
        }
    }



    # Now need to compare both controllers
    foreach my $controller (@{$self->{controllers}}) {
        foreach my $disk (keys %{$hash->{$controller}->{active_disks}}) {
            # Now look through each controller other than $controller
            # and see if the disks exists using serial number
            my $serial =
                    $hash->{$controller}->{active_disks}->{$disk}->{ 'serial' };
            my $found = 0;
            foreach my $other_controller (@{$self->{controllers}}) {
                if (($controller ne $other_controller)) {
                    # not looking on the same controller
                    foreach my $other_disk (keys %{$hash->{$other_controller}->{active_disks}}) {
                        if ($serial eq $hash->{$other_controller}->{active_disks}->{$other_disk}->{ 'serial' }) {
                            $found = 1;
                            last;
                        }
                    }
                    if (!$found) {
                        # Disk doesn't exists here
                        $hash->{$controller}->{not_seen_disks}->{$disk}->{other_controller} =
                                $other_controller;

                    }
                }
            }
        }
    }
    ## put errors in array and the print at the end.
    # Did we find any disks not seen by all controllers?
    foreach my $controller (@{$self->{controllers}}) {
        if (exists $hash->{$controller}->{not_seen_disks}) {
            foreach my $disk (keys %{$hash->{$controller}->{not_seen_disks}}) {
# Print info about the disk.
                $self->_set_err(-msg => "Disk $disk exists on $controller but not on  ".
                        $hash->{$controller}->{not_seen_disks}->{$disk}->{other_controller});
            }
        }
        # link hash into main hash. Both are the same so doesn't matter what we copy.
        $hash->{active_disks} = {%{$hash->{$controller}->{active_disks}}};
    }
}
=cut

1;

