#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers, clients, switches, etc...
## @author latif@netapp.com
## @status private
## @pod here

package San::SanMgr;

# Compiler directives.
use strict;

# Load resource specific modules
use San::SanMgr::Controller;
use San::SanMgr::Client;
use San::SanMgr::Fabric;
use San::SanMgr::HyperVisor;
use San::SanMgr::Switch;
use San::SanMgr::EventMan;
use San::SanMgr::Util;
use ParamsMgr;
use SubIPC;

# Load San specific modules
use San::SanClient;
use San::SanSwitch;
use San::ControllerCluster;

# Module import.
use NATE::ProcManager;
use NATE::Events;
use NATE::Inc qw(find_file);
use Tharn;
use Error;
use San::SanLog;
use Data::Dumper;
use File::Basename;
use Cwd;

use vars qw(@ISA @DTD_SCHEMA_FILES $SANMGR_DTD_PATH);

@ISA = qw(San::SanMgr::Util Error SanLog);

# Exported global variables
###########################
$SanMgr::TRUE  = $SanUtil::TRUE;
$SanMgr::FALSE = $SanUtil::FALSE;

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new SanMgr object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($proto, %args) = @_;

    my $self = {};
    my $class = ref($proto) || $proto;
    bless ($self, $class);

    %args = &SanUtil::strip_dash(%args);

    if($args{private_paramset}) {
        # Create a nate paramset just for this obj.
        $self->{nate_paramset}        = &ParamsMgr::clone();
    }

    # Obtain the SanMgr DTD Path
    $SANMGR_DTD_PATH = $self->_get_dtd_path();

    # Internal DTD XML Schema files for SanMgr
    @DTD_SCHEMA_FILES = ($SANMGR_DTD_PATH . "/scenario.dtd",
                         $SANMGR_DTD_PATH . "/clients.dtd",
                         $SANMGR_DTD_PATH . "/clusters.dtd",
                         $SANMGR_DTD_PATH . "/hypervisors.dtd",
                         $SANMGR_DTD_PATH . "/fabrics.dtd",
                         $SANMGR_DTD_PATH . "/eventman.dtd");

    $self->sanlog(-msg => "Using SANMGR_DTD_PATH : " . &ParamsMgr::get("SANMGR_DTD_PATH", paramset => $self->EnGenX::Util::get_paramset()));

    # if nodes is passed in, save for load_resources reuse
    $self->{nodes} = delete $args{nodes} if exists $args{nodes};

    # if clusters is passed in, save for load_resources reuse
    $self->{clusters} = delete $args{clusters} if exists $args{clusters};

    # replace -file with -xml_file
    if (exists $args{file}) {
        $args{xml_file} = delete $args{file};
    }

    # get the filepath of the xml_file
    $args{xml_file} = $self->_get_xml_path($args{xml_file});
    if ($self->error()) {
        return $self;
    }

    &ParamsMgr::process_args(%args);
    my $xml_file             = &ParamsMgr::get_arg(-arg => "xml_file",              -required =>  1, -default  => "");
    my @schema_files         = &ParamsMgr::get_arg(-arg => "schema_files",          -default  => [], -ref_type => "ARRAY");
    my $load_resources       = &ParamsMgr::get_arg(-arg => "load_resources",        -required =>  0, -default  => "1");
    my $run_id_prefix        = &ParamsMgr::get_arg(-arg => "run_id_prefix",         -required =>  0, -default  => "");
    my $enable_subtest_count = &ParamsMgr::get_arg(-arg => "enable_subtest_count",  -required =>  0, -default  => "1");
    my $obj_creation_timeout = &ParamsMgr::get_arg(-arg => "obj_creation_timeout",  -required =>  0, -default  => "300");
    my $display_xml          = &ParamsMgr::get_arg(-arg => "display_xml",           -default => "1");
    my $display_xml_runid    = &ParamsMgr::get_arg(-arg => "display_xml_runid",     -default => "XML-environment");
    my $clear_paramsmgr      = &ParamsMgr::get_arg(-arg => "clear_paramsmgr",       -default => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Set the run_id_prefix
    $self->{run_id_prefix} = $run_id_prefix;
    $self->{subtest_count} = 1;
    $self->{enable_subtest_count} = $enable_subtest_count;

    # store the obj creation timeout
    $self->{obj_creation_timeout} = $obj_creation_timeout;

    # allow for error stack build ups
    $self->{error_dont_overwrite_errors} = 1;
    $self->{error_stack_style}           = "normal";

    $self->sanlog(-msg => "Loading xml file $xml_file",
                  -msg_level => 1);

    # Before we load the new XML files, check to see if we need to clear out the ParamSet on ParamsMgr
    if ($clear_paramsmgr == 1) {
        &ParamsMgr::clear_global_paramset();
        if (&ParamsMgr::error()) {
            $self->_set_err("Failed to clear ParamsMgr: ", &ParamsMgr::get_error());
            return $self;
        }
    }

    # Process and Validate the xml file and any override schema files
    $self->load_xml(-xml_files    => [$xml_file],
                    -schema_files => \@schema_files,
                    -display_xml => $display_xml,
                    -display_xml_runid => $display_xml_runid);
    if ($self->error()) {
        return $self;
    }

    # Get the set of environment parameters from ParamsMgr
    my $environment = &ParamsMgr::search(search_path => [], key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Set the objects debug level
    $self->debug_level(-debug_level => $environment->{DEBUG_LEVEL});
    if ($self->error()) {
        return $self;
    }

    # Set the debug level on ParamsMgr and SubIPC
    &ParamsMgr::debug_level(-debug_level => $environment->{DEBUG_LEVEL});
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    &SubIPC::debug_level(-debug_level => $environment->{DEBUG_LEVEL});
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $self;
    }

    # Create local objects for each of the resource types
    # these are to be used later for config, setup, validate and cleanup work.
    $self->{sm_controller_object} = San::SanMgr::Controller->_new(-parent => $self);
    $self->{sm_client_object}     = San::SanMgr::Client->_new(-parent     => $self);
    $self->{sm_hypervisor_object} = San::SanMgr::HyperVisor->_new(-parent => $self);
    $self->{sm_switch_object}     = San::SanMgr::Switch->_new(-parent     => $self);
    $self->{sm_fabric_object}     = San::SanMgr::Fabric->_new(-parent     => $self);
    $self->{sm_eventman_object}   = San::SanMgr::EventMan->_new(-parent   => $self);

    # If load_resources is set, load up objects for the resources
    if ($load_resources == 1) {
        $self->load_resources(%args);
        if ($self->error()) {
            return $self;
        }

        # start the vs_client_relationship_svc
        $self->_init_vs_client_relationship_svc();

        # load all fabric resources
        $self->{sm_fabric_object}->load_resource();
        if ($self->{sm_fabric_object}->error()) {
            $self->_set_err($self->{sm_fabric_object}->get_error());
            return $SanMgr::FALSE;
        }
    } else {
        $self->sanlog(-msg => "No resource objects loaded load_resources => 0",
                      -msg_level => 1);

        # start the vs_client_relationship_svc
        $self->_init_vs_client_relationship_svc();
    }

    return $self;
}

###############################################################################
# Method:    controller
# Objective: Access the Controller Module
# Details:   See POD documentation at the end of this file.
###############################################################################
sub controller {
    my ($self) = @_;

    return $self->{sm_controller_object};
}

###############################################################################
# Method:    client
# Objective: Access the Client Module
# Details:   See POD documentation at the end of this file.
###############################################################################
sub client {
    my ($self) = @_;

    return $self->{sm_client_object};
}

###############################################################################
# Method:    switch
# Objective: Access the Switch Module
# Details:   See POD documentation at the end of this file.
###############################################################################
sub switch {
    my ($self) = @_;

    return $self->{sm_switch_object};
}

###############################################################################
# Method:    hypervisor
# Objective: Access the Hypervisor Module
# Details:   See POD documentation at the end of this file.
###############################################################################
sub hypervisor {
    my ($self) = @_;

    return $self->{sm_hypervisor_object};
}

###############################################################################
# Method:    fabric
# Objective: Access the Fabric Module
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fabric {
    my ($self) = @_;

    return $self->{sm_fabric_object};
}

###############################################################################
# Method:    eventman
# Objective: Access the Eventman Module
# Details:   See POD documentation at the end of this file.
###############################################################################
sub eventman {
    my ($self) = @_;

    return $self->{sm_eventman_object};
}

###############################################################################
# Method:    setup
# Objective: Execute setup routine on each resource type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup {
    my ($self, %args) = @_;

    $self->clear_error();

    &SubIPC::remove_subtest();

    my ($st_status, $subtest, $controller_objs);
    my (@messages) = ();

    # Get the set of actions parameters from ParamsMgr
    my $actions = &ParamsMgr::search(search_path => [], key => "ACTIONS",
                                     paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Check to see if user wants to run setup at all
    if ($actions->{SETUP} == 0) {
        $self->sanlog(-msg => "Skipping setup due to actions flag set to \"0\"",
                      -msg_level => 1);
        return $SanMgr::TRUE;
    }

    $self->sanlog(-msg => "Starting SanMgr setup",
                  -msg_level => 1);

    ##############################################
    # Switch setup
    ##############################################
    $self->{sm_switch_object}->setup(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_switch_object}->error()) {
        $self->_set_err("Switch Setup failed", $self->{sm_switch_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Controller setup
    ##############################################
    $self->{sm_controller_object}->setup(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_controller_object}->error()) {
        $self->_set_err("Controller Setup failed",
                        $self->{sm_controller_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ###########################################################################
    # TODO: Remove CONFIG code once controller config is properly implemented #
    ###########################################################################

    ##############################################
    # Hypervisor config
    ##############################################
    $self->{sm_hypervisor_object}->config(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_hypervisor_object}->error()) {
        $self->_set_err("HyperVisor Config failed",
                        $self->{sm_hypervisor_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Client config
    ##############################################
    $self->{sm_client_object}->config(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_client_object}->error()) {
        $self->_set_err("Client Config failed",
                        $self->{sm_client_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Fabric config
    ##############################################
    $self->{sm_fabric_object}->config(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_fabric_object}->error()) {
        $self->_set_err("Fabric Config failed",
                        $self->{sm_fabric_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # HyperVisor setup
    ##############################################
    $self->{sm_hypervisor_object}->setup(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_hypervisor_object}->error()) {
        $self->_set_err("HyperVisor Setup failed", $self->{sm_hypervisor_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Client setup
    ##############################################
    $self->{sm_client_object}->setup(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_client_object}->error()) {
        $self->_set_err("Client Setup failed", $self->{sm_client_object}->errmsg());
        return $SanMgr::FALSE;
    }

    return $SanMgr::TRUE;
} ## end sub setup

###############################################################################
# Method:    config
# Objective: Execute config routine on each resource type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config {
    my ($self) = @_;

    $self->clear_error();

    &SubIPC::remove_subtest();

    my ($st_status, $subtest);

    # Get the set of actions parameters from ParamsMgr
    my $actions = &ParamsMgr::search(search_path => [], key => "ACTIONS",
                                     paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Check to see if user wants to run config at all
    if ($actions->{CONFIG} == 0) {
        $self->sanlog(-msg => "Skipping config due to actions flag set to \"0\"",
                      -msg_level => 1);
        return $SanMgr::TRUE;
    }

    ##############################################
    # Cluster config
    ##############################################
    $self->{sm_controller_object}->config(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_controller_object}->error()) {
        $self->_set_err("Controller Config failed",
                        $self->{sm_client_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Hypervisor config
    ##############################################
    $self->{sm_hypervisor_object}->config(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_hypervisor_object}->error()) {
        $self->_set_err("HyperVisor Config failed",
                        $self->{sm_hypervisor_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Client config
    ##############################################
    $self->{sm_client_object}->config(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_client_object}->error()) {
        $self->_set_err("Client Config failed",
                        $self->{sm_client_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Fabric config
    ##############################################
    $self->{sm_fabric_object}->config(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_fabric_object}->error()) {
        $self->_set_err("Fabric Config failed",
                        $self->{sm_fabric_object}->errmsg());
        return $SanMgr::FALSE;
    }

    return $SanMgr::TRUE;
} ## end sub config

###############################################################################
# Method:    cleanup
# Objective: Execute cleanup routine on each resource type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ($self) = @_;

    $self->clear_error();

    &SubIPC::remove_subtest();

    my ($st_status, $subtest);

    # Get the set of actions parameters from ParamsMgr
    my $actions = &ParamsMgr::search(search_path => [], key => "ACTIONS",
                                     paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Check to see if user wants to run cleanup at all
    if ($actions->{CLEANUP} == 0) {
        $self->sanlog(-msg => "Skipping cleanup due to actions flag set to \"0\"",
                      -msg_level => 1);
        return $SanMgr::TRUE;
    }

    ##############################################
    # Client cleanup
    ##############################################
    $self->{sm_client_object}->cleanup(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_client_object}->error()) {
        $self->_set_err("Client Cleanup failed", $self->{sm_client_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # HyperVisor cleanup
    ##############################################
    $self->{sm_hypervisor_object}->cleanup(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_hypervisor_object}->error()) {
        $self->_set_err("HyperVisor Cleanup failed", $self->{sm_hypervisor_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Fabric deconfig (hacked for now)
    ##############################################
    $self->{sm_fabric_object}->deconfig(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_fabric_object}->error()) {
        $self->_set_err("Fabric Deconfig failed", $self->{sm_fabric_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ######################################################################
    # TODO: Remove DECONFIG code once controller deconfig is implemented #
    ######################################################################

    ##############################################
    # Client deconfig
    ##############################################
    $self->{sm_client_object}->deconfig(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_client_object}->error()) {
        $self->_set_err("Client Deconfig failed", $self->{sm_client_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # HyperVisor deconfig
    ##############################################
    $self->{sm_hypervisor_object}->deconfig(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_hypervisor_object}->error()) {
        $self->_set_err("HyperVisor Deconfig failed", $self->{sm_hypervisor_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Controller cleanup
    ##############################################
    $self->{sm_controller_object}->cleanup(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_controller_object}->error()) {
        $self->_set_err("Controller Cleanup failed", $self->{sm_controller_object}->errmsg());
        return $SanMgr::FALSE;
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    deconfig
# Objective: Execute deconfig routine on each resource type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub deconfig {
    my ($self) = @_;

    $self->clear_error();

    # Get the set of actions parameters from ParamsMgr
    my $actions = &ParamsMgr::search(search_path => [], key => "ACTIONS",
                                     paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Check to see if user wants to run deconfig at all
    if ($actions->{DECONFIG} == 0) {
        $self->sanlog(-msg => "Skipping deconfig due to actions flag set to \"0\"",
                      -msg_level => 1);
        return $SanMgr::TRUE;
    }

    ##############################################
    # Fabric deconfig
    ##############################################
    $self->{sm_fabric_object}->deconfig(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_fabric_object}->error()) {
        $self->_set_err("Fabric Deconfig failed", $self->{sm_fabric_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Client deconfig
    ##############################################
    $self->{sm_client_object}->deconfig(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_client_object}->error()) {
        $self->_set_err("Client Deconfig failed", $self->{sm_client_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # HyperVisor deconfig
    ##############################################
    $self->{sm_hypervisor_object}->deconfig(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_hypervisor_object}->error()) {
        $self->_set_err("HyperVisor Deconfig failed", $self->{sm_hypervisor_object}->errmsg());
        return $SanMgr::FALSE;
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    validate
# Objective: Execute validate routine on each resource type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self) = @_;

    $self->clear_error();

    my ($st_status, $subtest);

    # Get the set of actions parameters from ParamsMgr
    my $actions = &ParamsMgr::search(search_path => [], key => "ACTIONS",
                                     paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Check to see if user wants to run validation at all
    if ($actions->{VALIDATE} == 0) {
        $self->sanlog(-msg => "Skipping validation due to actions flag set to \"0\"",
                      -msg_level => 1);
        return $SanMgr::TRUE;
    }

    ##############################################
    # Switch validate
    ##############################################
    $self->{sm_switch_object}->validate(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_switch_object}->error()) {
        $self->_set_err("Switch Validation failed", $self->{sm_switch_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Controller validate
    ##############################################
    $subtest = &SubIPC::create(-id  => $self->{run_id_prefix} . "Controller_Validate",
                               -esc => $self->{enable_subtest_count},
                               -run => sub { $self->{sm_controller_object}->validate(); },
                               -ipc => 1,
                               -bg  => 1);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err("Unable to wait for Controller Validate subtest "
                      . "properly\n\t" . &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: "
                       . $st_status->{failed}
                       . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "Controller validate have passed\n\t\tFailed: "
                             . $st_status->{failed}
                             . "\n\t\tPassed: " . $st_status->{passed});
    }

    ##############################################
    # Client validate
    ##############################################
    $self->{sm_client_object}->validate(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_client_object}->error()) {
        $self->_set_err("Client Validation failed", $self->{sm_client_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # HyperVisor validate
    ##############################################
    $self->{sm_hypervisor_object}->validate(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_hypervisor_object}->error()) {
        $self->_set_err("HyperVisor Validation failed", $self->{sm_hypervisor_object}->errmsg());
        return $SanMgr::FALSE;
    }

    ##############################################
    # Fabric validate
    ##############################################
    $self->{sm_fabric_object}->validate(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{sm_fabric_object}->error()) {
        $self->_set_err("Fabric Validate failed", $self->{sm_fabric_object}->errmsg());
        return $SanMgr::FALSE;
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    load_resources
# Objective: Go through all resources provided by caller (XML file) and create
#            necessary objects for each to be stored on SanMgr for later use
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_resources {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $reconnect_clients = &ParamsMgr::get_arg(-arg => "reconnect_client_objects_after_load",
                                                -default  => $SanMgr::TRUE);

    my $reconnect_hypervisors = &ParamsMgr::get_arg(-arg => "reconnect_hypervisor_objects_after_load",
                                                    -default  => $SanMgr::FALSE);
    my $reconnect_switches = &ParamsMgr::get_arg(-arg => "reconnect_switch_objects_after_load",
                                                 -default  => $SanMgr::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to parse args", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Go through all resources provided by caller (XML file) and create
    # necessary objects for each to be stored on SanMgr for later use

    # load client/hypervisor/switch resources in the background
    my ($client_listener, $client_flistener, @client_procs) = $self->{sm_client_object}->load_resource(%args);
    if ($self->{sm_client_object}->error()) {
        $self->_set_err($self->{sm_client_object}->get_error());
        return $SanMgr::FALSE;
    }

    my ($hypervisor_listener, $hypervisor_flistener, @hypervisor_procs) = $self->{sm_hypervisor_object}->load_resource(%args);
    if ($self->{sm_hypervisor_object}->error()) {
        $self->_set_err($self->{sm_hypervisor_object}->get_error());
        return $SanMgr::FALSE;
    }

    my ($switch_listener, $switch_flistener, @switch_procs) = $self->{sm_switch_object}->load_resource(%args);
    if ($self->{sm_switch_object}->error()) {
        $self->_set_err($self->{sm_switch_object}->get_error());
        return $SanMgr::FALSE;
    }

    # initialize the listeners complete count to 0
    $self->{listeners_complete_count} = 0;

    # install the listeners for each resource type
    NATE::Events::listener_add($client_listener);
    NATE::Events::listener_add($client_flistener);

    NATE::Events::listener_add($hypervisor_listener);
    NATE::Events::listener_add($hypervisor_flistener);

    NATE::Events::listener_add($switch_listener);
    NATE::Events::listener_add($switch_flistener);

    # set up the proc manager to spawn off parallel processes to load each individual module
    my $subtest_prefix = &SubIPC::_get_next_subtest_count();
    my $pm;
    my @all_procs = (@client_procs, @hypervisor_procs, @switch_procs);
    if (scalar(@all_procs)) {
        $pm = NATE::ProcManager->new(proc_info => \@all_procs,
                                     runid => "${subtest_prefix}_load_objects");
        # start the processes
        $pm->start();
    }

    # reap and cleanup the processes
    if (scalar(@all_procs)) {
        $pm->wait();
        $pm->destroy();
    }

    # load controller object resources in the foreground
    $self->{sm_controller_object}->load_resource();
    if ($self->{sm_controller_object}->error()) {
        $self->_set_err("Encountered an error while loading controller objects", $self->{sm_controller_object}->get_error());
        # dont bail yet, let all the other errors propagate as well
    }

    # burt656995: keep processing listeners until all the listeners have completed
    # it might happen that we remove the listeners prematurely before all the resources are loaded.
    my $total_proc_count = scalar(@all_procs);
    my $object_creation_timeout_time = time() + $self->{obj_creation_timeout};
    while ($total_proc_count > $self->{listeners_complete_count}) {
        NATE::Events::eventloop_drain(timeout => 1);
        if (time > $object_creation_timeout_time) {
            $self->_set_err("The Load Objects task exceeded the timeout '$self->{obj_creation_timeout}'");
            last;
        }
    }

    # clean up the listeners
    NATE::Events::listener_remove($client_listener);
    NATE::Events::listener_remove($client_flistener);

    NATE::Events::listener_remove($hypervisor_listener);
    NATE::Events::listener_remove($hypervisor_flistener);

    NATE::Events::listener_remove($switch_listener);
    NATE::Events::listener_remove($switch_flistener);

    # check if any of the individual modules hit any errors
    my $fail_count = 0;

    # if the number of registered clients and procs dont match something bad happened inside a load objects
    # process
    if ($self->{sm_client_object}->error()) {
        $self->_set_err("Encountered an error while loading client objects", $self->{sm_client_object}->errmsg());
    }
    my $num_client_procs = @client_procs;
    my $num_client_objs = scalar(keys %{$self->{_objects}->{client}});
    $fail_count = $num_client_procs - $num_client_objs;
    if ($fail_count > 0) {
        $self->_set_err("Detected that $fail_count out of $num_client_procs client load object processes failed to initialize.");
    }

    # if the number of registered hypervisors and procs dont match something bad happened inside a load objects
    # process
    if ($self->{sm_hypervisor_object}->error()) {
        $self->_set_err("Encountered an error while loading hypervisor objects", $self->{sm_hypervisor_object}->errmsg());
    }
    my $num_hypervisor_procs = @hypervisor_procs;
    my $num_hypervisor_objs = scalar(keys %{$self->{_objects}->{hypervisor}});
    $fail_count = $num_hypervisor_procs - $num_hypervisor_objs;
    if ($fail_count > 0) {
        $self->_set_err("Detected that $fail_count out of $num_hypervisor_procs hypervisor load object processes failed to initialize.");
    }

    # if the number of registered switchs and procs dont match something bad happened inside a load objects
    # process
    if ($self->{sm_switch_object}->error()) {
        $self->_set_err("Encountered an error while switch objects", $self->{sm_switch_object}->errmsg());
    }
    my $num_switch_procs = @switch_procs;
    my $num_switch_objs = scalar(keys %{$self->{_objects}->{switch}});
    $fail_count = $num_switch_procs - $num_switch_objs;
    if ($fail_count > 0) {
        $self->_set_err("Detected that $fail_count out of $num_switch_procs switch load object processes failed to initialize.");
    }

    # catch all errors here and report them
    if ($self->error()) {
        $self->_set_err("The load_objects task failed, please investigate.");
        return $SanMgr::FALSE;
    }

    # reconnect all client/hypervisor/switch objects if requested
    if ($reconnect_clients) {
        my @objs = $self->get_objects(-type => "client");
        foreach my $client_obj (@objs) {
            $client_obj->reconnect();
            if ($client_obj->error()) {
                $self->_set_err("Failed to reconnect to client " . $client_obj->hostname(),
                                $client_obj->errmsg());
                return $SanMgr::FALSE;
            }
        }
    }
    if ($reconnect_hypervisors) {
        my @objs = $self->get_objects(-type => "hypervisors");
        foreach my $hypervisor_obj (@objs) {
            $hypervisor_obj->reconnect();
            if ($hypervisor_obj->error()) {
                $self->_set_err("Failed to reconnect to hypervisor " . $hypervisor_obj->hostname(),
                                $hypervisor_obj->errmsg());
                return $SanMgr::FALSE;
            }
        }
    }
    if ($reconnect_switches) {
        my @objs = $self->get_objects(-type => "switches");
        foreach my $switch_obj (@objs) {
            $switch_obj->reconnect();
            if ($switch_obj->error()) {
                $self->_set_err("Failed to reconnect to switch " . $switch_obj->hostname(),
                                $switch_obj->errmsg());
                return $SanMgr::FALSE;
            }
        }
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method: initialize_eventman
# Objective: Initialize the eventman object via the XML file/schema specified
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initialize_eventman {
    my ($self, %args) = @_;

    $self->clear_error();

    # initialize the sanmgr eventman module
    # this will dynamically build the eventman object
    # using elements from the ParamSet that will
    # define the api calls to invoke
    $self->{sm_eventman_object}->initialize();
    if ($self->{sm_eventman_object}->error()) {
        $self->_set_err("Failed to initialize eventman", $self->{sm_eventman_object}->errmsg());
        return $SanMgr::FALSE;
    }

    # unset the eventman tags from the paramset
    &ParamsMgr::unset("EVENTMAN");
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to remove EVENTMAN from the NATE ParamSet", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    load_xml
# Objective: Load XML file(s) into ParamSet via ParamsMgr
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_xml {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my @xml_files = &ParamsMgr::get_arg(-arg => "xml_files",
                                        -required => 1,
                                        -ref_type => "ARRAY");
    my @override_schema_files = &ParamsMgr::get_arg(-arg => "schema_files",
                                                    -ref_type => "ARRAY",
                                                    -default => []);
    my $display_xml = &ParamsMgr::get_arg(-arg => "display_xml",
                                          -default => 1);
    my $runid = &ParamsMgr::get_arg(-arg => "display_xml_runid",
                                    -default => "XML-environment");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Use the SanMgr internal schema files for each module
    # if requested to (default)
    my @schema_files = @DTD_SCHEMA_FILES;

    # if any override schema files were specified, push them on the list
    if (scalar(@override_schema_files)) {
        push(@schema_files, @override_schema_files);
    }

    # validate & load each xml_file into ParamsMgr
    foreach my $xml_file (@xml_files) {

        # Validate the XML file against the provided schema file list
        my $xml_string = &ParamsMgr::validate_and_process_xml(-schema_files => \@schema_files,
                                                              -xml_file => $xml_file);
        if (&ParamsMgr::error()) {
            $self->_set_err("Failed to validate XML file: $xml_file", &ParamsMgr::get_error());
            return $SanMgr::FALSE;
        }

        # load the xml string that was generated by validate_xml into paramsmgr
        &ParamsMgr::load_xml(-xml_string => $xml_string,
                             -display_xml => $display_xml,
                             -display_xml_runid => $runid,
                             paramset           => $self->EnGenX::Util::get_paramset()
                            );
        if (&ParamsMgr::error()) {
            $self->_set_err("Failed to load XML String", &ParamsMgr::get_error());
            return $ParamsMgr::FALSE;
        }
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _store_controller_cluster_objs
# Objective: Helper method that stores controller cluster content as objects
# Details:   None. Private method.
###############################################################################
sub _store_controller_cluster_objs {
    my ($self, %args) = @_;

    $self->clear_error();

    # Locals
    my ($data, $sorted_data, $enable_cluster_refresh) = undef;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # tells us whether to refresh the cluster object or not
    $enable_cluster_refresh = &ParamsMgr::get_arg(-arg => "enable_cluster_refresh",
                                                  -default  => "0");

    # Main data structure containing data created by SanMgr's setup
    $data = &ParamsMgr::get_arg(-arg => "data", -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Log the entry we are adding to the cleanup stack
    my $dumpout = Dumper($data);
    $self->sanlog(-msg => "Dumper of data getting added to cleanup stack:\n"
                        . $dumpout,
                  -msg_level  => 7,
                  -add_caller => 1);

    # Store the data on a cleanup stack. This will be used by SanMgr's cleanup
    $self->{_objects}->{cleanup_stack} = $data;

    foreach my $key (keys(%{$data})) {
        $self->sanlog(-msg => "Storing Cleanup Data for $key object",
                      -msg_level  => 7,
                      -add_caller => 1);
        $self->{_objects}->{$key} = $data->{$key};
    }

    # Refresh the cserver objects for all clusters if requested to do it
    if ($enable_cluster_refresh == 1 and defined($self->{_objects}->{cluster})) {
        foreach my $cluster (keys(%{$self->{_objects}->{cluster}})) {
            my $cserver_obj = $self->{_objects}->{cluster}->{$cluster}->get_cserver_obj();
            if ($self->{_objects}->{cluster}->{$cluster}->error()) {
                $self->_set_err("Failed to obtain cserver object for cluster '$cluster'",
                                $self->{_objects}->{cluster}->{$cluster}->errmsg());
                return $SanMgr::FALSE;
            }
            my $cserver_obj_name = $cserver_obj->get_name();
            $cserver_obj->refresh();
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to refresh cserver '$cserver_obj_name'",
                                $cserver_obj->errmsg());
                return $SanMgr::FALSE;
            }

            my @vservers = $cserver_obj->get_vserver_objs();
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to get vserver objects for cserver '$cserver_obj_name'",
                                $cserver_obj->errmsg());
                return $SanMgr::FALSE;
            }
            foreach my $vserver (@vservers) {
                my $vserver_name = $vserver->get_name();
                $vserver->refresh();
                if ($vserver->error()) {
                    $self->_set_err("Failed to refresh vserver '$vserver_name'",
                                    $vserver->errmsg());
                    return $SanMgr::FALSE;
                }
            }
        }
    }

    return $SanMgr::TRUE;
} ## end _store_controller_cluster_objs()

###############################################################################
# Method:    _get_dtd_path
# Objective: Helper method that generates the DTD File path based on
#            environment variables.
# Details:   None. Private method.
###############################################################################
sub _get_dtd_path {
    my ($self) = @_;

    # (1) default to the non null value of SANMGR_DTD_PATH
    my $dtd_path = &Tharn::param("SANMGR_DTD_PATH", "-default", "");
    &ParamsMgr::set(key => "SANMGR_DTD_PATH", value => $dtd_path, paramset => $self->EnGenX::Util::get_paramset());

    if (defined($dtd_path) and $dtd_path !~ /^\s*$/) {
        return $dtd_path;
    }

    # (2) try to determine from NATE_LIB/NATE_LIB_UNIX
    my @natelib;
    if (exists($ENV{"NATE_LIB"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB"});
    }
    if (not scalar(@natelib) and exists($ENV{"NATE_LIB_UNIX"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB_UNIX"});
    }

    # walk through all the natelib paths until we find one that works
    foreach my $natelib_path (@natelib) {
        $dtd_path = $natelib_path . "/lib/San/SanMgr/DTDFiles";
        if (-d $dtd_path) {
            &ParamsMgr::set(key => "SANMGR_DTD_PATH", value => $dtd_path, paramset => $self->EnGenX::Util::get_paramset());
            return $dtd_path;
        }
    }

    # (3) try to determine the path based on the TESTNAME variable
    my $testname = $ENV{"TESTNAME"};
    if ($testname =~ /^(\/.+?\/p4\/.+?\/)/) {
        $dtd_path = $1 . "lib/San/SanMgr/DTDFiles";

        # make sure this path exists, otherwise dont use it
        if (-d $dtd_path) {
            &ParamsMgr::set(key => "SANMGR_DTD_PATH", value => $dtd_path, paramset => $self->EnGenX::Util::get_paramset());
            return $dtd_path;
        }
    }

    # (4) otherwise, default to sanrtest's version
    $dtd_path = "/u/sanrtest/p4/test/lib/San/SanMgr/DTDFiles";
    &ParamsMgr::set(key => "SANMGR_DTD_PATH", value => $dtd_path, paramset => $self->EnGenX::Util::get_paramset());

    return $dtd_path;
}

###############################################################################
# Method:    _get_xml_path
# Objective: to get the full path name with extension (xml)
# Parameter:  $xml_file (Optional)   -> user specified xml name
# Expected Output if script location is in
# "/u/muthug/p4/test/san/lun_mgmt/portset/basic_func.thpl";
# and all xmls exist.
#
#  _get_xml_filename() =
#     "/u/muthug/p4/test/san/lun_mgmt/portset/basic_func.xml";
#  _get_xml_filename("") =
#     "/u/muthug/p4/test/san/lun_mgmt/portset/basic_func.xml";
#  _get_xml_filename("gmk1.xml.xml") =
#     "/u/muthug/p4/test/san/lun_mgmt/portset/gmk1.xml.xml";
#  _get_xml_filename("justfile") =
#     "/u/muthug/p4/test/san/lun_mgmt/portset/justfile.xml";
#  _get_xml_filename("p4/test/san/lun_mgmt/portset/fullpath.xml") =
#     "p4/test/san/lun_mgmt/portset/fullpath.xml";
#  _get_xml_filename("/san/lun_mgmt/portset/fullpathwitoutext") =
#     "/san/lun_mgmt/portset/fullpathwitoutext.xml";
#  _get_xml_filename("/u/any/junk/folder/name/only/portset/") =
#     "/u/any/junk/folder/name/only/portset/basic_func.xml";
#
# Return:  returns a string with full path with filename and extension(.xml)
###############################################################################
sub _get_xml_path {
    my ($self, $xml_file) = @_;
    my $old_xml_file = $xml_file;
    my ($sc_file, $sc_dir, $sc_ext);
    my ($file,    $dir,    $ext);

    # if user passed nothing or empty ("") check the script location with name
    # the same as the script
    unless (length $xml_file) {
        $xml_file = $Tharn::Scriptspec;

        # get the last .thpl(or any ext) name and replace with .xml
        $xml_file =~ s/\.[^.]*/.xml/;
        unless ( -f $xml_file) {
            $self->_set_err("No Xml File was passed in and $xml_file does not exist");
            return "";
        }
        return $xml_file;
    }

    # always ensure that a file ends with .xml
    if ($xml_file !~ /\.xml$/) {
        $xml_file .= ".xml";
    }

    # -f checks to see if the file exists and is a plain file
    if (-f $xml_file) {
        return $xml_file;
    } else {
        # second get the script location: dir:filename:ext
        # this regex will split something like
        # /u/mcclella/p4/test/san/sanity/testXml_path.xml
        # into
        # $sc_dir = /u/mcclella/p4/test/san/sanity
        # $sc_file = testXml_path
        # $sc_ext = xml
        ($sc_dir, $sc_file, $sc_ext) = $Tharn::Scriptspec =~ m#(\S+)/(\S+)\.(\S+)$#;

        # Get file location info
        ($file, $dir, $ext) = &File::Basename::fileparse($xml_file, qr/\.[^.]*/);

        # If dir is not passed, grab it from script location
        if ($dir =~ /^.\/$|\.$|^\/$|""/) {
            $xml_file = "$sc_dir/";
        } else {
            $xml_file = $dir;
        }

        # If file is not passed, grab it from script name
        unless (length $file) {
            $xml_file .= $sc_file;
        } else {
            $xml_file .= $file;
        }

        # Add the extension back
        if (length $ext) {
            $xml_file .= $ext;
        }

        # -f checks to see if the file exists and is a plain file
        # If the file does not exist, this was the final location to check
        # and an error is returned.
        return $xml_file if (-f $xml_file);

        # Still have not found the file.  Seems like we hit this when
        # $sc_dir is a relative path (smokes).
        # Next, call NATE::Inc::find_file() with the current
        # $xml_file and see if it can find it in the nate lib path.
        my @natelib;
        if (exists($ENV{"NATE_LIB"})) {
            @natelib = split(/;/, $ENV{"NATE_LIB"});
        }
        if (not scalar(@natelib) and exists($ENV{"NATE_LIB_UNIX"})) {
            @natelib = split(/;/, $ENV{"NATE_LIB_UNIX"});
        }
        $xml_file = find_file($xml_file, search => [@natelib]);

        return $xml_file if (-f $xml_file);

        # If we get here, we didn't find the file anywhere we know to look

        # Get the cwd for the error message and set the error
        my $cwd = &Cwd::getcwd();

        $self->_set_err("Could not locate your XML file "
                       . "$old_xml_file does not exist as a direct path "
                       . "and $file"."$ext"." can not be found in the following:\n"
                       . "  - $cwd\n"
                       . "  - $sc_dir\n"
                       . "NATELIB below\n"
                       . "  - " . join("\n  - ",@natelib)
                       );
        return "";
    } # end else
}    # end of sub _get_xml_path

1;

=pod

=head1 Package Name

SanMgr.pm

=head1 Author

dl-san-automation

=head1 Description

A C-Mode SAN Provisioning Library. SanMgr relies on an XML Configuration file
to describe the inputs for provisioning. Please refer to the SanMgr Guide Wiki
for details on how to create an XML configuration file.
http://wikid.netapp.com/w/SAN_Automation/SanMgr_Guide

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>

B<San/SanLog.pm>

B<SanMgr/Util.pm>

=head1 B<Public Methods>

=head2 B<new(%args)>

Creates a SanMgr object

=over 4

=item Inputs

=item *B<-xml_file | -file> Required

The XML Configuration file path to use during setup() and cleanup()

=item *I<-schema_files> Default: []

An array refrence of DTD file paths to use when validating -xml_file.
Precendence of the DTD files is greater for those files at the end of the list.

=item *I<-run_id_prefix> Default: ""

A prefix to prepend to all subtest runids created by this SanMgr object

=item *I<-enable_subtest_count> Default: 1

Specifies whether to enable a three-digit padded subtest count prefix at
the beginning of every subtest runid.

=item *I<-display_xml> Default: 1

Specifies whether to display the Post processed XML and DTD rules in a subtest once
the XML file is validated.

=item *I<-display_xml_runid> Default: "XML-environment"

Indicates the subtest runid to use when creating the XML environment subtest.

=item *I<-load_resources> Default: 1

Specifies whether SanMgr new() should instantiate objects for all resources
specified in the XML file. Disabling this is not recommended unless you know what you
are doing.

=item *I<-clear_paramsmgr> Default: 1

If set, will clear out all parameters in ParamsMgr prior to loading the XML file.
This option is useful when dealing with multiple SanMgr objects in the same script.

=item *I<-reconnect_client_objects_after_load> Default: 1

If set, will reconnect all client objects after they get loaded so that the process
that invoked this function has client objects with working connections.

=item *I<-reconnect_hypervisor_objects_after_load> Default: 0

If set, will reconnect all hypervisor objects after they get loaded so that the process
that invoked this function has hypervisor objects with working connections.

=item *I<-reconnect_switch_objects_after_load> Default: 0

If set, will reconnect all switch objects after they get loaded so that the process
that invoked this function has switch objects with working connections.

=item Outputs

Returns a SanMgr object

=item Examples

 use San::SanMgr;

 my $sanmgr_obj  = new San::SanMgr(-xml_file => "/path/to/my/xml.file");
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

=back

=head2 B<setup()>

Invokes the setup phase for SanMgr (which also invokes the config phase)

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file");
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 $sanmgr_obj->setup()
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

=back

=head2 B<config()>

Invokes the config phase for SanMgr. The method setup() invokes config(), so
dont invoke both in the same script unless you know what you are doing

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file");
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 $sanmgr_obj->config()
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

=back

=head2 B<cleanup()>

Invokes the cleanup phase for SanMgr (Which also invokes the deconfig phase).

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file");
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 $sanmgr_obj->cleanup()
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

=back

=head2 B<deconfig()>

Invokes the deconfig phase for SanMgr. The method cleanup() invokes deconfig(), so
dont invoke both in the same script unless you know what you are doing

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file");
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 $sanmgr_obj->deconfig()
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

=back

=head2 B<validate()>

Invokes the validate phase for SanMgr.

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file");
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 $sanmgr_obj->validate()
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

=back

=head2 B<load_resources()>

Loads all resources from the xml file into the SanMgr object. This
will instantiate SanClient/ControllerCluster/SanSwitch objects and is
proportional in time to the number of resources in the xml file.

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 use San::SanMgr;

 # don't load resources in the constructor
 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file",
                                  -load_resources => 0);
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # ...
 # do stuff before loading the resources
 # ...

 $sanmgr_obj->load_resources()
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

=back

=head2 B<load_xml()>

Load additional XML files into SanMgr.

=over 4

=item Inputs

=item *B<-xml_file | -file> Required

The XML Configuration file path to load.

=item *I<-schema_files> Default: []

An array refrence of DTD file paths to use when validating -xml_file.
Precendence of the DTD files is greater for those files at the end of the list.

=item *I<-display_xml> Default: 1

Specifies whether to display the Post processed XML and DTD rules in a subtest once
the XML file is validated.

=item *I<-display_xml_runid> Default: "XML-environment"

Indicates the subtest runid to use when creating the XML environment subtest.

=item Outputs

None

=item Examples

 use San::SanMgr;

 # don't load resources in the constructor
 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file",
                                  -load_resources => 0);
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # ...
 # do stuff before loading the second xml file
 # ...

 $sanmgr_obj->load_xml(-xml_file => "/path/to/my/second/xml.file")
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

=back

=head2 B<initialize_eventman()>

Initialize an EventMan object based on the <eventman> tags specified in the
XML configuration file.

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file")
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # initialize the eventman object
 $sanmgr_obj->initialize_eventman()
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # invoke setup
 $sanmgr_obj->setup();
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # grab the eventman object that was created
 my ($eventman_obj) = $sanmgr_obj->get_objects(-type => "eventman");
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # run events (refresh resources is needed if setup() is invoked after initialize_eventman()
 # see the POD for EventMan.pm for more details on how to use EventMan APIs
 $eventman_obj->run_events(-refresh_resources => 1);
 if ($eventman_obj->error()) {
     $Test->nlog($eventman_obj->errmsg());
     return;
 }

=back

=head2 B<controller()>

Access the Controller module

=over 4

=item Inputs

None

=item Outputs

A San::SanMgr::Controller object

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file")
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # invoke setup only on the controller module
 $sanmgr_obj->controller()->setup();
 if ($sanmgr_obj->controller()->error()) {
     $Test->nlog($sanmgr_obj->controller()->errmsg());
     return;
 }

=back

=head2 B<client()>

Access the Client module

=over 4

=item Inputs

None

=item Outputs

A San::SanMgr::Client object

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file")
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # invoke setup only on the client module
 $sanmgr_obj->client()->setup();
 if ($sanmgr_obj->client()->error()) {
     $Test->nlog($sanmgr_obj->client()->errmsg());
     return;
 }

=back

=head2 B<hypervisor()>

Access the Hypervisor module

=over 4

=item Inputs

None

=item Outputs

A San::SanMgr::Hypervisor object

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file")
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # invoke setup only on the hypervisor module
 $sanmgr_obj->hypervisor()->setup();
 if ($sanmgr_obj->hypervisor()->error()) {
     $Test->nlog($sanmgr_obj->hypervisor()->errmsg());
     return;
 }

=back

=head2 B<fabric()>

Access the Fabric module

=over 4

=item Inputs

None

=item Outputs

A San::SanMgr::Fabric object

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file")
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # invoke setup only on the fabric module
 $sanmgr_obj->fabric()->setup();
 if ($sanmgr_obj->fabric()->error()) {
     $Test->nlog($sanmgr_obj->fabric()->errmsg());
     return;
 }

=back

=head2 B<switch()>

Access the Switch module

=over 4

=item Inputs

None

=item Outputs

A San::SanMgr::Switch object

=item Examples

 use San::SanMgr;

 my $sanmgr_obj = new San::SanMgr(-xml_file => "/path/to/my/xml.file")
 if ($sanmgr_obj->error()) {
     $Test->nlog($sanmgr_obj->errmsg());
     return;
 }

 # invoke setup only on the switch module
 $sanmgr_obj->switch()->setup();
 if ($sanmgr_obj->switch()->error()) {
     $Test->nlog($sanmgr_obj->switch()->errmsg());
     return;
 }

=back

=cut
