#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on clients
## @author latif@netapp.com
## @status private
## @pod here

package San::SanMgr::Client;

# Compiler directives.
use strict;

# Module import.
use Tharn;
use Error;
use San::SanLog;
use San::SanMgr::Util;
use EnGenX::Util;
use San::SanClient;
use Data::Dumper;
use Data::Dump;
use ParamsMgr;

use NATE::Process;
use NATE::ProcManager;
use NATE::STAF::Peer;
use NATE::STAF qw(staf_global);
use NATE::Util::Sockets;
use NATE::BaseException qw(:try);
use Sys::Hostname;

# Load resource specific modules
use San::SanMgr::Client::setup;
use San::SanMgr::Client::config;
use San::SanMgr::Client::cleanup;
use San::SanMgr::Client::deconfig;
use San::SanMgr::Client::validate;

use vars qw(@ISA);

@ISA = qw(San::SanMgr::Util Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new resource object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($proto, %args) = @_;

    my $self = {};

    my $class = ref($proto) || $proto;
    bless ($self, $class);

    # allow for error stack build ups
    $self->{error_dont_overwrite_errors} = 1;
    $self->{error_stack_style} = "normal";

    &ParamsMgr::process_args(%args);
    my $parent = &ParamsMgr::get_arg(-arg => "parent", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Save parent
    $self->{parent} = $parent;

    # Get the set of environment parameters from ParamsMgr
    my $environment = &ParamsMgr::search(search_path => [], key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Set the objects debug level
    $self->debug_level(-debug_level => $environment->{DEBUG_LEVEL});
    if ($self->error()) {
        return $self;
    }

    return $self;
}

###############################################################################
# Method:    setup
# Objective: Launch subtest for each client to perform setup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup {
    my ($self, %args) = @_;

    $self->clear_error();

    my $multi_clients = EnGenX::Util::_get_all_clients_paramsets($self);

    my (@messages, @subtests);
    my $subtest;
    my $device_list = undef;

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }
    foreach my $clients (@{$multi_clients}){
        # perform setup if it is enabled across all clients
        if ($clients->{SETUP} == 1) {
            my $client_arg;
            my $max_setup_failure_count = $clients->{MAX_SETUP_FAILURE_COUNT};
            my $required_clients        = $clients->{REQUIRED_CLIENTS};

            # go through the members of the CLIENTS tag looking
            # for the CLIENT\d*(s) to work on; this
            # separates them from the attributes
            foreach my $key (keys(%{$clients})){
                if ($key =~ /^CLIENT\d*$/){
                    $client_arg->{$key} = $clients->{$key};
                }
            }
            $subtest = &SubIPC::create(-id => "$run_id_prefix" . "Client-setup",
                                   -esc => $self->{parent}->{enable_subtest_count},
                                   -run => sub { $self->_setup_all_clients(-max_setup_failure_count => $max_setup_failure_count,
                                                                           -required_clients => $required_clients,
                                                                           -clients => $client_arg);},
                                   -ipc => 1,
                                   -bg => 1);
            if (&SubIPC::error()) {
                $self->_set_err(&SubIPC::get_error());
                return $SanMgr::FALSE;
            }
            push(@subtests, $subtest);
        }
    }

    # reap the setup subtests if they exist
    if (scalar(@subtests)) {
        # Wait for subtests to finish
        &SubIPC::wait_finish(-subtest => \@subtests);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }

        # Get messages to process
        @messages = &SubIPC::get_messages(-subtest => \@subtests, -type => "client_setup_complete");

        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }

        # Go through the messages and pull all of the device_lists together and notify the parent
        foreach my $message (@messages) {
            # We are interested in only the message (device_list).
            foreach my $client (keys(%{$message->{extra}->{device_list}})) {
                push(@{$device_list->{$client}}, @{$message->{extra}->{device_list}->{$client}->{DEV_LIST}});
            }
        }
    }

    my $client_paramsets = EnGenX::Util::_get_all_client_paramsets($self);
    foreach my $client (@{$client_paramsets}) {
        my $dl = $client->{DEV_LIST};
        my $name = $client->{NAME};
        # skip this client if it doesnt have a dev list
        next if (not defined($dl) or $dl eq "");
        # create the dev list for this host and it to the device_list hash
        my @devs = split(",", $dl);
        @devs = map { {path => $_, io_type => "fs"} } @devs;
        $device_list->{$name} = \@devs;
    }

    # store device_list on $self's parent if it is defined
    if (defined($device_list)) {
        # store device_list on $self's parent
        $self->{parent}->store_client_dev_list(-device_list => $device_list);
        if ($self->{parent}->error()) {
            $self->_set_err("Failed to store client device list: " . $self->{parent}->get_error());
            return $SanMgr::FALSE;
        }
    }

    # check to see if the client setup subtests failed
    if (scalar(@subtests)) {
        my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
        if ($st_status->{failed} != 0) {
            $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
            return $SanMgr::FALSE;
        } else {
            $self->sanlog(-msg => "Client setup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        }
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _setup_all_clients
# Objective: Launch subtest for each client to perform setup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _setup_all_clients {

    my ($self, %args) = @_;

    $self->clear_error();
    my ($st_status, @subtests);
    my ($device_list);
    &ParamsMgr::process_args(%args);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my $max_setup_failure_count = &ParamsMgr::get_arg(-arg => "max_setup_failure_count",  -required => 0, -default => "1");
    my $required_clients = &ParamsMgr::get_arg(-arg => "required_clients",  -required => 0, -default => "");
    my $clients = &ParamsMgr::get_arg(-arg => "clients",  -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument getting failed", &ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    if (not scalar(keys %{$clients})) {
        &SubIPC::pass(-msg => "There are no Clients to perform setup for");
        return $SanMgr::TRUE;
    }

    # Go through client list and start setup
    foreach my $client (keys(%{$clients})) {
        my $current_client = $clients->{$client};
        my $client_name = $current_client->{NAME};
        $current_client->{SETUP} = 1 if (not defined($current_client->{SETUP}));

        # skip setup on this client if it is not running setup or if it has a dev list
        if ($current_client->{SETUP} == 0 or $current_client->{DEV_LIST} ne "") {
            $self->sanlog(-msg => "Skipping client " . $client_name . " because setup tag set to \"0\" or a dev list is defined",
                          -msg_level => 0);
            next;
        }

        my $subtest = &SubIPC::create(-id => $client_name,
                                      -run => sub { &San::SanMgr::Client::setup::setup_single($self, -resource => $current_client); },
                                      -parent => $self,
                                      -ipc => 1, -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => &SubIPC::get_error());
            return $SanMgr::FALSE;
        }

        push(@subtests, $subtest);
    }

    # return if there were no client setup subtests
    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "No Clients have setup enabled");
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # Get messages to process
    my @messages = &SubIPC::get_messages(-subtest => \@subtests, -type => "setup_complete");

    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # Go through the messages and pull all of the device_lists together and notify the parent
    foreach my $message (@messages) {
        # We are interested in only the message (device_list).
        foreach my $device (@{$message->{extra}->{device_list}}) {
            push(@{$device_list->{$message->{extra}->{name}}->{DEV_LIST}}, $device);
        }
    }

    # Write back to parent the device_list generated
    &SubIPC::write(-subtest => [], -msg => "Setup completed", -type => "client_setup_complete", -device_list => $device_list);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => "Unable to write (config_complete) message to parent");
        return $SanMgr::FALSE;
    }

    $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);

    if ($st_status->{failed} > $max_setup_failure_count) {
            $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
    }
    # Go through the subtests and see if any clients in the required client list
    # have failed.
    foreach my $subtest (@subtests){
        my $client_name = $subtest->{id};
        if (($subtest->{exit_status} =~ /fail/i) and ($required_clients =~ /$client_name/)) {
            $self->_set_err("A subtest failure has been reported:\n\t\tThe required client " . $subtest->{id} . "has failed setup.\n" .
                "Failed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
    }

    if($st_status->{failed} > 0){
        $self->sanlog(-msg => "Setup passed on enough clients to continue " .
        "running the test.\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }
    &SubIPC::pass(-msg => "Clients have been setup");
    return $SanMgr::TRUE;
}

###############################################################################
# Method:    config
# Objective: Launch subtest for each client to perform config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config {
    my ($self, %args) = @_;

    $self->clear_error();
    my @subtests;
    my $multi_clients = EnGenX::Util::_get_all_clients_paramsets($self);

    foreach my $clients (@{$multi_clients}){

        my $client_arg;
        foreach my $key (keys(%{$clients})){
            if ($key =~ /^CLIENT\d*$/){
                $client_arg->{$key} = $clients->{$key};
            }
        }

        if ($clients->{CONFIG} == 0) {
            $self->sanlog(-msg => "Skipping Client Config");
            next;
        }

        my $max_setup_failure_count = $clients->{MAX_SETUP_FAILURE_COUNT};
        my $required_clients        = $clients->{REQUIRED_CLIENTS};
        &ParamsMgr::process_args(%args);

        my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
        if (&ParamsMgr::error()) {
            $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
            return $SanMgr::FALSE;
        }

        my $subtest = &SubIPC::create(-id  => "$run_id_prefix" . "Client-config",
                                    -esc => $self->{parent}->{enable_subtest_count},
                                    -run => sub { $self->_config_all_clients(-max_setup_failure_count => $max_setup_failure_count,
                                                                             -required_clients => $required_clients,
                                                                             -clients => $client_arg);},
                                    -ipc => 1,
                                    -bg  => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);

    }

    # return if there are no subtests running client config
    if (not scalar(@subtests)) {
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed}) {
        $self->_set_err("A config subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    }
    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _config_all_clients
# Objective: Launch subtest for each client to configure
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _config_all_clients {

    my ($self, %args) = @_;

    $self->clear_error();
    my ($st_status, @subtests);
    my ($device_list);

    &ParamsMgr::process_args(%args);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my $max_setup_failure_count = &ParamsMgr::get_arg(-arg => "max_setup_failure_count",  -required => 0, -default => "1");
    my $required_clients = &ParamsMgr::get_arg(-arg => "required_clients",  -required => 0, -default => "");
    my $clients = &ParamsMgr::get_arg(-arg => "clients",  -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    if (not scalar(keys %{$clients})) {
        &SubIPC::pass(-msg => "There are no Clients in this clients group to setup");
        next;
    }

    # Go through client list and start config
    foreach my $client (keys(%{$clients})) {
        my $current_client = $clients->{$client};
        my $client_name = $current_client->{NAME};
        $current_client->{CONFIG} = 1 if (not defined($current_client->{CONFIG}));

        # Check to see if resource is to be skipped
        if ($current_client->{CONFIG} == 0) {
            $self->sanlog(-msg => "Skipping client " . $client_name . " because config tag set to \"0\"",
                          -msg_level => 0);
            next;
        }

        my $subtest = &SubIPC::create(-id => $client_name,
                                      -run => sub { &San::SanMgr::Client::config::config_single($self, -resource => $current_client); },
                                      -parent => $self,
                                      -ipc => 1, -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => &SubIPC::get_error());
            return $SanMgr::FALSE;
        }

        push(@subtests, $subtest);
    }

    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "No Clients have config enabled");
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => &SubIPC::get_error());
       return $SanMgr::FALSE;
    }

    $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);

    if ($st_status->{failed} > $max_setup_failure_count) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    foreach my $subtest (@subtests){
        my $client_name = $subtest->{id};
        if (($subtest->{exit_status} =~ /fail/i) and ($required_clients =~ /$client_name/)) {
            $self->_set_err("A subtest failure has been reported:\n\t\tThe required client " . $subtest->{id} . "has failed config.\n" .
            "Failed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE
        }
    }

    if($st_status->{failed} > 0){
        $self->sanlog(-msg => "Config passed on enough clients to continue " .
        "running the test.\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }
    &SubIPC::pass(-msg => "Clients have been configured");
    return $SanMgr::TRUE;
}

###############################################################################
# Method:    cleanup
# Objective: Launch subtest for each client to perform cleanup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ($self, %args) = @_;

    my @subtests;

    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $multi_clients = EnGenX::Util::_get_all_clients_paramsets($self);

    foreach my $clients (@{$multi_clients}) {
        if ($clients->{CLEANUP} == 0) {
            $self->sanlog(-msg => "Skipping Client Cleanup.");
            next;
        }

        # go through the members of the CLIENTS tag looking
        # for the CLIENT\d*(s) to work on; this
        # separates them from the attributes
        my $client_arg = {};
        foreach my $key (keys(%{$clients})){
            if ($key =~ /^CLIENT\d*$/){
                $client_arg->{$key} = $clients->{$key};
            }
        }
        
        my $subtest = &SubIPC::create(-id  => "$run_id_prefix" . "Client-cleanup",
                                  -esc => $self->{parent}->{enable_subtest_count},
                                  -run => sub { $self->_cleanup_all_clients(-clients => $client_arg); },
                                  -ipc => 1,
                                  -bg  => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }
        push (@subtests, $subtest);
    }

    # return if there are no subtests running client cleanup
    if (not scalar(@subtests)) {
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed}) {
        $self->_set_err("A cleanup subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    }
    $self->sanlog(-msg => "Client cleanup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _cleanup_all_clients
# Objective: Launch subtest for each client to perform cleanup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cleanup_all_clients {
    my ($self, %args) = @_;

    $self->clear_error();

    my ($st_status, @subtests);
    my ($device_list);

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(-arg => "clients",  -required => 1, -default => {});
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    foreach my $client (keys(%{$clients})) {
        my $current_client = $clients->{$client};
        my $client_name = $current_client->{NAME};
        $current_client->{CLEANUP} = 1 if (not defined($current_client->{CLEANUP}));

        # Check to see if resource is to be skipped
        if ($current_client->{CLEANUP} == 0) {
            $self->sanlog(-msg => "Skipping client " . $client_name . " because cleanup tag set to \"0\"",
                          -msg_level => 0);
            next;
        }

        my $subtest = &SubIPC::create(-id => $client_name,
                                      -run => sub { &San::SanMgr::Client::cleanup::cleanup_single($self, -resource => $current_client); },
                                      -parent => $self,
                                      -ipc => 1, -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => &SubIPC::get_error());
            return $SanMgr::FALSE;
        }
        
        push(@subtests, $subtest);
    }
    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "No Clients have cleanup enabled");
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # Get messages to process
    my @messages = &SubIPC::get_messages(-subtest => \@subtests, -type => "cleanup_complete");
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # Write back to parent the device_list generated
    &SubIPC::write(-subtest => [], -msg => "Cleanup completed", -type => "cleanup_complete", -device_list => $device_list);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => "Unable to write (config_complete) message to parent");
        return $SanMgr::FALSE;
    }
    $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);

    if ($st_status->{failed} != 0) {
        &SubIPC::fail(-msg => "A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        &SubIPC::pass(-msg => "All subtests have passed:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::TRUE;
    }
}

###############################################################################
# Method:    deconfig
# Objective: Launch subtest for each client to perform deconfig
# Details:   See POD documentation at the end of this file.
###############################################################################
sub deconfig {
    my ($self, %args) = @_;
    
    my @subtests;
    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $multi_clients = EnGenX::Util::_get_all_clients_paramsets($self);

    foreach my $clients (@{$multi_clients}) {
        if ($clients->{DECONFIG} == 0) {
            next;
        }

        # go through the members of the CLIENTS tag looking
        # for the CLIENT\d*(s) to work on; this
        # separates them from the attributes
        my $client_arg = {};
        foreach my $key (keys(%{$clients})){
            if ($key =~ /^CLIENT\d*$/){
                $client_arg->{$key} = $clients->{$key};
            }
        }

        my $subtest = &SubIPC::create(-id  => "$run_id_prefix" . "Client-deconfig",
                                      -esc => $self->{parent}->{enable_subtest_count},
                                      -run => sub { $self->_deconfig_all_clients(-clients => $client_arg); },
                                      -ipc => 1,
                                      -bg  => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }
        push (@subtests, $subtest);
    }

    # return if there are no subtests running client deconfig
    if (not scalar(@subtests)) {
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);

    if ($st_status->{failed}) {
        $self->_set_err("A deconfig subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _deconfig_all_clients
# Objective: Launch subtest for each client to deconfigure
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _deconfig_all_clients {
    my ($self, %args) = @_;
    $self->clear_error();

    my ($st_status, @subtests);
    my ($device_list);

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(-arg => "clients",  -required => 1, -default => {});
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Go through client list and start deconfig
    foreach my $client (keys(%{$clients})) {
        my $current_client = $clients->{$client};
        my $client_name = $current_client->{NAME};
        $current_client->{DECONFIG} = 1 if (not defined($current_client->{DECONFIG}));
        
        # Check to see if resource is to be skipped
        if ($current_client->{DECONFIG} == 0) {
            $self->sanlog(-msg => "Skipping client " . $client_name . " because deconfig tag set to \"0\"",
                          -msg_level => 0);
            next;
        }
        
        my $subtest = &SubIPC::create(-id => $client_name,
                                      -run => sub { &San::SanMgr::Client::deconfig::deconfig_single($self, -resource => $current_client); },
                                      -parent => $self,
                                      -ipc => 1, -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => &SubIPC::get_error());
            return $SanMgr::FALSE;
        }
        
        push(@subtests, $subtest);
    }
    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "No Clients have deconfig enabled");
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "Client Deconfig has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "Clients have been deconfigured");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    validate
# Objective: Launch subtest for all clients to perform validation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self, %args) = @_;
    my @subtests;
    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $multi_clients = EnGenX::Util::_get_all_clients_paramsets($self);

    foreach my $clients (@{$multi_clients}) {
        if ($clients->{VALIDATE} == 0) {
            $self->sanlog(-msg => "Skipping Client Validation for this client group");
            next;
        }

        # go through the members of the CLIENTS tag looking
        # for the CLIENT\d*(s) to work on; this
        # separates them from the attributes
        my $client_arg = {};
        foreach my $key (keys(%{$clients})){
            if ($key =~ /^CLIENT\d*$/){
                $client_arg->{$key} = $clients->{$key};
            }
        }
        
        my $subtest = &SubIPC::create(-id  => "$run_id_prefix" . "Client-validate",
                                  -esc => $self->{parent}->{enable_subtest_count},
                                  -run => sub { $self->_validate_all_clients(-clients => $client_arg); },
                                  -ipc => 1,
                                  -bg  => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "Client validation has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::remove_subtest();

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _validate_all_clients
# Objective: Launch subtest for each client to perform validation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _validate_all_clients {
    my ($self, %args) = @_;

    $self->clear_error();

    my ($st_status, @subtests);

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(-arg => "clients",  -required => 1, -default => {});
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    foreach my $client (keys(%{$clients})) {
        my $current_client = $clients->{$client};
        $current_client->{VALIDATE} = 1 if (not defined($current_client->{VALIDATE}));
        
        # Check to see if resource is to be skipped
        if ($current_client->{VALIDATE} == 0) {
            $self->sanlog(-msg => "Skipping client " . $current_client->{NAME} . " because validate tag set to \"0\"",
                          -msg_level => 0);
            next;
        }
        
        my $subtest = &SubIPC::create(-id => $current_client->{NAME},
                                      -run => sub { &San::SanMgr::Client::validate::validate_single($self, -resource => $current_client); },
                                      -parent => $self,
                                      -ipc => 1, -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => &SubIPC::get_error());
            return $SanMgr::FALSE;
        }
        
        push(@subtests, $subtest);
    }
    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "No Clients have validate enabled");
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);

    $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        &SubIPC::fail(-msg => "A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        &SubIPC::pass(-msg => "All subtests have passed:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::TRUE;
    }
}

###############################################################################
# Method:    load_resource
# Objective: Gather list of clients and create objects for each one
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_resource {
    my ($self, %args) = @_;

    $self->clear_error();

    my $staf = staf_global();

    # Get the set of environment parameters from ParamsMgr
    my $environment = &ParamsMgr::search(search_path => [], key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }
    my $global_set = $self->EnGenX::Util::get_paramset();

    my $all_clients = $self->EnGenX::Util::_get_all_client_paramsets();

    # make sure there are no duplicate clients
    my %dup_hash;
    foreach my $client_elm (@{$all_clients}) {
        $dup_hash{$client_elm->{NAME}}++;
        if ($dup_hash{$client_elm->{NAME}} > 1) {
            $self->_set_err("Detected duplicate client entries for the client " . $client_elm->{NAME});
            return $SanMgr::FALSE;
        }
    }

    my @procs = ();

    foreach my $client_elm (@{$all_clients}){
        # Spawn a new process to load this client
        my $proc = NATE::Process->new(codespec => '{ #EnGenX load ' . $client_elm->{NAME} . '
                                                     use San::SanMgr::Client;
                                                     San::SanMgr::Client::_load_client(); }',
                                      runid    => "client__" . $client_elm->{NAME},
                                      onexit   => \&NATE::Process::on_exit_propagate_worst_result,
                                      client   => 'localhost',
                                      params   => { "client_ds" => $client_elm,
                                                    "environment" => $environment,
                                                    "staf_parent_handle" => $staf->handle(),
                                                    "staf_parent_endpoint" => NATE::Util::Sockets::get_address(Sys::Hostname::hostname())});
        push(@procs, $proc);
    }

    # set up the Event listener and callback methods
    my $filter = sub {
        my $msg = shift;
        return ($msg->{type} =~ /SanMgr\/Client\/SanClient\/\S+\/Load_Complete/);
    };
    my $callback = sub {
        my ($msg, $listener) = @_;

        # parse out the message information
        my ($client_name) = ($msg->{type} =~ /SanMgr\/Client\/SanClient\/(\S+?)\/Load_Complete/);
        my $sm_client_obj = $listener->{sm_client_obj};

        # rebuilt the serialized object
        my $obj = eval($msg->{message}->{obj_serialized});
        if (not defined($obj)) {
            $sm_client_obj->_set_err("The SanClient object for client $client_name did not get transfered properly");
            $sm_client_obj->{parent}->{listeners_complete_count}++;
            return $SanMgr::FALSE;
        }

        # recreate the  SanClient object using the object template
        $obj = SanClient->new(-obj_template => $obj);
        if ($obj->error()) {
            $sm_client_obj->_set_err("Failed to instantiate SanClient object from template of client $client_name",
                                     $obj->errmsg());
            $sm_client_obj->{parent}->{listeners_complete_count}++;
            return $SanMgr::FALSE;
        }

        # store the object on the sanmgr resource hash
        $sm_client_obj->{parent}->{_objects}->{client}->{$obj->hostname()} = $obj;

        # validate the md5 for the obj
        if (not $sm_client_obj->_validate_obj_md5_hash(obj => $obj,
                                                       obj_md5_hex_before => $msg->{message}->{obj_md5_hex})) {
            $sm_client_obj->{parent}->{listeners_complete_count}++;
            return $SanMgr::FALSE;
        }

        $sm_client_obj->{parent}->{listeners_complete_count}++;
        return $SanMgr::TRUE;
    };
    my $listener = {filter => $filter, callback => $callback, sm_client_obj => $self};

    # build the fail listener
    my $fail_filter = sub {
        my $msg = shift;
        return ($msg->{type} =~ /SanMgr\/Client\/SanClient\/\S+\/Load_Failed/);
    };
    my $fail_callback = sub {
        my ($msg, $flistener) = @_;

        my $sm_client_obj = $flistener->{sm_client_obj};

        # parse out the message information
        my ($client_name) = ($msg->{type} =~ /SanMgr\/Client\/SanClient\/(\S+?)\/Load_Failed/);
        $sm_client_obj->_set_err("The SanClient object for client $client_name failed to instantiate.");
        $sm_client_obj->{parent}->{listeners_complete_count}++;
        return $SanMgr::FALSE;
    };

    my $fail_listener = {filter => $fail_filter, callback => $fail_callback, sm_client_obj => $self};

    return ($listener, $fail_listener, @procs);
}

###############################################################################
# Method:    _load_client
# Objective: Load the client object and send it back to the parent
###############################################################################
sub _load_client {
    my ($class) = @_;

    # fetch params passed into the process
    my $global_params = NATE::ParamSet->get_global();
    my $client_ds = $global_params->get("client_ds");
    my $environment = $global_params->get("environment");
    my $staf_parent_handle = $global_params->get("staf_parent_handle");
    my $staf_parent_endpoint = $global_params->get("staf_parent_endpoint");

    my $parent = NATE::STAF::Peer->new(handle   => $staf_parent_handle,
                                       endpoint => $staf_parent_endpoint);

    #############################################
    ### General client parameters
    #############################################
    my @errors = ();

    my $conn_type       = exists($client_ds->{CONN_TYPE})    ? $client_ds->{CONN_TYPE}    : 'THARNHOST';
    my $protocol        = exists($client_ds->{PROTOCOL})     ? $client_ds->{PROTOCOL}     :
        push(@errors, "Missing required_field for client 'protocol'");
    my $debug_level     = exists($client_ds->{DEBUG_LEVEL})  ? $client_ds->{DEBUG_LEVEL}  : 0;
    my $load_inits      = exists($client_ds->{LOAD_INITS})   ? $client_ds->{LOAD_INITS}   : 1;
    my $alua_support    = exists($client_ds->{ALUA_SUPPORT}) ? $client_ds->{ALUA_SUPPORT} : 1;
    my $mp_type         = exists($client_ds->{MP_TYPE})      ? $client_ds->{MP_TYPE}      : "";
    my $lvm_type        = exists($client_ds->{LVM_TYPE})     ? $client_ds->{LVM_TYPE}     : "";
    my $hypervisor_type = exists($client_ds->{HYPERVISOR})   ? $client_ds->{HYPERVISOR}   : "";

    # check for argument errors
    if (scalar(@errors)) {
        require NATE::Exceptions::Argument;
        $parent->message_put(type => "SanMgr/Client/SanClient/" . $client_ds->{NAME} . "/Load_Failed",
                             message => {});
        NATE::Exceptions::Argument->throw(join("\n", @errors));
    }

    # initialize the pnate attribute if it doesnt exist
    my $pnate = 0;
    if (defined($environment) and exists($environment->{PNATE})) {
        $pnate = $environment->{PNATE};
    }

    # Create a fake self so that we can get the paramset from the "EnGenX" object
    # using the util method.
    my $fake_self = {};
    bless($fake_self, "EnGenX");

    # generate a fake object with all the options built into it, so we
    # can pass in the iscsi type value into the sanclient constructor
    # needed for Linux & ESX
    my $obj = {};
    $fake_self->San::SanMgr::Util::assign_sanclient_options(-client_obj => $obj,
                                                            -datastruct => $client_ds,
                                                            -env_datastruct => $environment);

    # Create client object
    my $cl_obj = SanClient->new(-host                     => $client_ds->{NAME},
                                -conn_type                => $conn_type,
                                -debug_level              => $debug_level,
                                -load_inits               => $load_inits,
                                -mp_type                  => $mp_type,
                                -lvm_type                 => $lvm_type,
                                -protocol                 => $protocol,
                                -pnate_environment        => $pnate,
                                -hypervisor_type          => $hypervisor_type,
                                -alua_support             => $alua_support,
                                -iscsi_type               => $obj->{iscsi}->{CONFIG_ISCSI_TYPE}
                            );

    if ($cl_obj->error()) {
        $parent->message_put(type => "SanMgr/Client/SanClient/" . $client_ds->{NAME} . "/Load_Failed",
                             message => {});
        NATE::BaseException->throw("Failed to create SanClient object for " . $client_ds->{NAME} . " : " .
                                   $cl_obj->errmsg());
        return;
    }

    # assign all iscsi_options & fcp_options to this sanclient object
    $fake_self->San::SanMgr::Util::assign_sanclient_options(-client_obj => $cl_obj,
                                                            -datastruct => $client_ds,
                                                            -env_datastruct => $environment);

    # destroy the connection to the sanclient object,
    # we have to do this or else the object we send back to the parent gets messed up
    $cl_obj->disconnect_connection();

    # calculate the md5 for the hypervisor obj
    my $obj_md5_hex = San::SanMgr::Util->_calculate_obj_md5_hash($cl_obj);

    $parent->message_put(type => "SanMgr/Client/SanClient/" . $client_ds->{NAME} ."/Load_Complete",
                         message => {obj_md5_hex => $obj_md5_hex,
                                     obj_serialized => Data::Dump::dump($cl_obj)});

    return;
}

################################################################################
# Subroutine Name: _kill_io_instances_subtest
# Objective:       Kill all instances of the client program on any client
################################################################################
sub _kill_io_instances_subtest {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $client = &ParamsMgr::get_arg(-arg => "client", -required => 1, -default => "");
    my $resource = &ParamsMgr::get_arg(-arg => "resource", -required => 1);
    my $client_io_program = &ParamsMgr::get_arg(-arg => "client_io_program", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my ($exe_to_kill, @exes_to_kill);
    my (@subtests);

    # skip kill io processes if the flag is disabled
    if ($resource->{KILL_IO_PROCESSES} == 0) {
        $self->sanlog(-msg => "Skipping kill_io_processes because it " .
                              "is disabled for client '$client'",
                      -msg_level => "0");
        return $SanMgr::TRUE;
    }

    # determine the process application to kill (all instances)
    if ($client_io_program =~ /dt/ or $client_io_program =~ /all/i) {
        push (@exes_to_kill, "dt");
    }

    if ($client_io_program =~ /sio/ or $client_io_program =~ /all/i) {
        push (@exes_to_kill, "sio");
    }

    if ($client_io_program =~ /blast/ or $client_io_program =~ /all/i) {
        push (@exes_to_kill, "blast");
    }

    if ($client_io_program !~ /sio/ and
        $client_io_program !~ /dt/ and
        $client_io_program !~ /blast/ and
        $client_io_program !~ /all/i) {
        $self->_set_err("No valid client_io_program passed in ($client_io_program)");
        return $SanMgr::FALSE;
    }

    my @procs;
    foreach $exe_to_kill (@exes_to_kill) {
        # kill the I/O process still on the host (all platforms)
        my $proc = NATE::Process->new(codespec => '{ use San::SanClient::ClientUtil; ' .
                                                  '  &ClientUtil::kill_process_instances("' . $client . '","' . $exe_to_kill . '"); }',
                                      client => 'localhost',
                                      runid => "io_task_kill_" . $exe_to_kill . "_" . $client,
                                      onexit => \&NATE::Process::on_exit_die_worst_result);
        push(@procs, $proc);
    }

    my $pm = NATE::ProcManager->new(proc_info => \@procs, runid => "io_task_kill");
    $pm->start();
    $pm->wait();
    $pm->destroy();

    return $SanMgr::TRUE;
}

1;
