#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on resource cleanup
## @author latif@netapp.com
## @status private
## @pod here

package San::SanMgr::Client::cleanup;

# Compiler directives.
use strict;
use San::SanLog;
use NACL::C::Client;
use NACL::STask::Client::Mount;


use Data::Dumper;
use NATE::BaseException qw(:try);

use vars qw(@ISA);
@ISA = qw(Error SanLog);

###############################################################################
# Method:    cleanup_single
# Objective: Cleanup run against a single resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # $self is a Client object
    # $resource is a hash of parameters for a particular resource to work on
    $self->sanlog(-msg => "Resource parameters/data:\n" . Dumper($resource),
                  -msg_level => 3);

    # Load up the resource information and get them via get_arg
    my $caller_id = $resource->{NAME} . "|" . time();
    &ParamsMgr::process_args(%{$resource}, -CALLER => $caller_id);

    #############################################
    ### General client parameters
    #############################################
    my $client_name =               &ParamsMgr::get_arg(-arg => "NAME", -required => 1, -default_to_env => 1,
                                                        -CALLER => $caller_id);
    my $protocol =                  &ParamsMgr::get_arg(-arg => "PROTOCOL", -required => 1, -default_to_env => 1,
                                                        -CALLER => $caller_id);
    my $io_type =                   &ParamsMgr::get_arg(-arg => "IO_TYPE", -required => 1, -default_to_env => 1,
                                                        -CALLER => $caller_id);
    my $debug_level =               &ParamsMgr::get_arg(-arg => "DEBUG_LEVEL", -required => 0, -default_to_env => 1,
                                                        -default => 0, -CALLER => $caller_id);
    my $lvm_type =                  &ParamsMgr::get_arg(-arg => "LVM_TYPE", -required => 0, -default_to_env => 1,
                                                        -default => "", -CALLER => $caller_id);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Get client object
    my ($hostobj) = $self->get_objects(-type => "client", -names => [$client_name]);
    if ($self->error()) {
        $self->_set_err("Failed to retrieve object for client '$client_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to the client
    $hostobj->reconnect(-disconnect_existing_conn => 0);
    if ($hostobj->error()) {
        $self->_set_err("Failed to reconnect to client '$client_name'", $hostobj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # kill off any running I/O program on the client systems.
    $self->_kill_io_instances_subtest(-client => $client_name,
                                      -resource => $resource,
                                      -client_io_program => "all");
    if ($self->error()) {
        $self->_set_err("Failed to kill IO instances");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    if ($protocol =~ /FCP|iSCSI/i) {
        if ($lvm_type ne "" and $lvm_type ne "none") {
            my (@vgObjs, $vgObj, @volObjs, $volObj, $lvmObj);

            # create a lvm object
            $lvmObj = $hostobj->create_lvm(-lvm => $lvm_type);
            if ($lvmObj->error()) {
                $self->_set_err ("Failed to create the LVM object", $hostobj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            # cycle through the volume groups / logical voulumes and destroy
            # all the objects. Make sure on filesystems to unmount before
            # trying to destroy
            @vgObjs = $lvmObj->get_vgs();
            foreach $vgObj (@vgObjs) {
                @volObjs = $vgObj->get_vols();
                foreach $volObj (@volObjs) {
                    if ($io_type =~ /fs/) {
                        if ($volObj->is_mounted() == 1) {
                            $volObj->unmount();
                            if ($volObj->error()) {
                                $self->_set_err("Failed to unmount the filesystem", $volObj->errmsg());
                                &SubIPC::fail(-object => $self);
                                return $SanMgr::FALSE;
                            }
                        }
                    }
                    $volObj->destroy();
                    if ($volObj->error()) {
                        $self->_set_err("Failed to destroy volume object", $volObj->errmsg());
                        &SubIPC::fail(-object => $self);
                        return $SanMgr::FALSE;
                    }
                }
                $vgObj->destroy();
                if ($vgObj->error()) {
                    $self->_set_err("Failed to destroy VG object", $vgObj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }
            }

            # cleanup any orphaned devices, for now just run on AIX, HPUX, and Solaris.
            # Linux and Windows are implemented with reboot and we want to avoid this.
            if ($hostobj->hosttype() =~ /(hpux|solaris|unix-aix)/i) {
                $hostobj->orphaned_device_cleanup();
                if ($hostobj->error()) {
                    $self->_set_err("Unable to cleanup orphaned devices", $hostobj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }
            }
        } else {
            # Create a device manager object
            my $devmgrobj = $hostobj->create_devmgr();
            if ($hostobj->error) {
                $self->_set_err("Failed to create device manager object", $hostobj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            if ($io_type =~ /fs/) {
                foreach my $device ($devmgrobj->get_devices()) {
                    $device->unmount_fs();
                }
            }
            # cleanup any orphaned devices, for now just run on AIX and HP
            if ($hostobj->hosttype() =~ /unix-aix/i or $hostobj->hosttype() =~ /hpux/i) {
                $hostobj->orphaned_device_cleanup();
                if ($hostobj->error()) {
                    $self->_set_err("Unable to cleanup orphaned devices", $hostobj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }
            }
        }
    }

    # clean up NFS for unix hosts
    if ($protocol =~ /nfs/i) {
        # create the Command interface for this client
        my $client_ci;
        try {
            $client_ci = new NACL::C::Client(name => $client_name,
                                             apiset_must => {conntype => 'ssh-persist',
                                                             connid => 'ssh'});
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # clean up all mount points related to the nfs target hash
        my @mount_objs = NACL::STask::Client::Mount->find(command_interface => $client_ci,
                                                          apiset_must => {conntype => 'ssh-persist',
                                                                          connid => 'ssh'});

        # clean up all mount points that are owned by EnGenX
        foreach my $mount_obj (@mount_objs) {
            try {
                # only purge engenx mount points
                if ($mount_obj->mount_point() =~ /engenx/i) {
                    $mount_obj->purge(force => 1);
                }
            } otherwise {
                $self->_set_err("Exception caught while unmounting mount point: " . shift->text());
            };
            if ($self->error()) {
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }
    }

    # Write back to parent the device_list generated
    &SubIPC::write(-msg => "Completed client ($client_name) cleanup", -device_list => {},
                   -type => "cleanup_complete", -name => $client_name);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg    => "Unable to write (config_complete) message to parent",
                      -object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "Client cleanup complete");

    return $SanMgr::TRUE;
}

1;
