#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on resource config
## @author latif@netapp.com
## @status private
## @pod here

package San::SanMgr::Client::config;

# Compiler directives.
use strict;
use San::SanLog;
use ParamsMgr;
use San::SanUtil;

use Data::Dumper;

use vars qw(@ISA);
@ISA = qw(Error SanLog);

###############################################################################
# Method:    config_single
# Objective: Config run against a single resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # $self is a Client object
    # $resource is a hash of parameters for a particular resource to work on
    $self->sanlog(-msg => "Resource parameters/data:\n" . Dumper($resource),
                  -msg_level => 3);

    # Load up the resource information and get them via get_arg
    my $caller_id = $resource->{NAME} . "|" . time();
    #&ParamsMgr::process_args(%{$resource});
    &ParamsMgr::process_args(%{$resource}, -CALLER => $caller_id);

    #############################################
    ### General client parameters
    #############################################
    my $client_name =               &ParamsMgr::get_arg(-arg => "NAME", -required => 1, -default_to_env => 1,
                                                        -CALLER => $caller_id);
    my $protocol =                  &ParamsMgr::get_arg(-arg => "PROTOCOL", -required => 1, -default_to_env => 1,
                                                        -CALLER => $caller_id);
    my $hypervisor =                &ParamsMgr::get_arg(-arg => "HYPERVISOR", -default => "none", -default_to_env => 1,
                                                        -CALLER => $caller_id);
    my $debug_level =               &ParamsMgr::get_arg(-arg => "DEBUG_LEVEL", -required => 0, -default_to_env => 1,
                                                        -default => "0", -CALLER => $caller_id);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Setup some basic default
    my $cf_mode = "ssi";

    # Get client object
    my ($hostobj) = $self->get_objects(-type => "client", -names => [$client_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get client ($client_name) object");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to the client
    $hostobj->reconnect(-disconnect_existing_conn => 0);
    if ($hostobj->error()) {
        $self->_set_err("Failed to reconnect to client '$client_name'", $hostobj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # kill off any running I/O program on the client systems.
    $self->_kill_io_instances_subtest(-client => $client_name,
                                      -resource => $resource,
                                      -client_io_program => "all");
    if ($self->error()) {
        $self->_set_err("Error killing IO instances");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # skip config for virtual clients that are under hypervisor control
    if ($hypervisor !~ /^none$/i) {
        &SubIPC::pass(-msg => "Skipping Client config for client '$client_name' because it is a virtual machine");
        return $SanMgr::TRUE;
    }

    my $data;
    foreach my $i (1..5) {
        ($data) = $self->{parent}->get_target_list(-client_name => $client_name,
                                                   -filter_iscsi_targets_by_portset => $hostobj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                   -filter_fcp_targets_by_portset   => $hostobj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
        if ($self->{parent}->error()) {
            if ($i >= 5) {
                # error state
                last;
            } else {
                # retry
                $self->{parent}->_clear_err();
                sleep 20;
                next;
            }
        } else {
            # success
            last;
        }
    }

    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get iSCSI/FCP target lists");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    # Run the configure routine within the client library
    my %targets;
    if ($protocol =~ /iscsi/i) {
        $targets{-iscsi_targets} = \@iSCSITargets;
    }
    if ($protocol =~ /fcp/i) {
        $targets{-fcp_targets} = \@FcpTargets;
    }
    $hostobj->configure(-cfo_mode => $cf_mode, %targets);

    # Check to make sure that configure has properly worked
    if ($hostobj->error()) {
        $self->_set_err("Failed to configure $protocol for client ($client_name)", $hostobj->errmsg);
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Write back to parent the device_list generated
    &SubIPC::write(-msg => "Completed client ($client_name) config",
                   -type => "config_complete", -name => $client_name);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg    => "Unable to write (config_complete) message to parent",
                      -object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "Client config complete");

    return $SanMgr::TRUE;
}

=pod
=head1 Package Name
San::SanMgr::Client::config
=head1 Author
Michael M. Latif
=head1 Description
Provide a common set of libraries to work on resource config
=head1 See Also
Package inherits from the following modules. See these modules for further
interface detail.
B<Error.pm>
B<San/SanLog.pm>
B<San/SanUtil>
B<ParamsMgr.pm>

=head1 Public Methods
=head2 B<config_single($self, %args)>
This function provisions a client with the provided resource data.
It is intended to be used within a subtest. The function will issue a SubIPC::pass or fail
to inform the caller of final result.
=over 4
=item Inputs
=item * I<-resource>
Hash containing resource data to be used when provisioning the resource
=item Outputs
By using SubIPC::write this function can return the device list provisioned. A message with
the extra data of 'device_list' is sent to the caller. Its the responsibility of the caller
to retrieve the message and pull out the device list for future use.
=item Examples
&San::SanMgr::Client::config::config_single($self, -resource => {-name => "client1", -protocol => "fcp" ..});
=back

=cut

1;
