#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on resource deconfig
## @author hrivera@netapp.com
## @status private
## @pod here

package San::SanMgr::Client::deconfig;

# Compiler directives.
use strict;
use San::SanLog;
use ParamsMgr;
use San::SanUtil;

use Data::Dumper;

use vars qw(@ISA);
@ISA = qw(Error SanLog);

###############################################################################
# Method:    deconfig_single
# Objective: Deconfig run against a single resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub deconfig_single {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource",
                                       -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $client_name = $resource->{NAME};

    # skip deconfig for virtual clients.
    if ($resource->{HYPERVISOR} !~ /^none$/i) {
        &SubIPC::pass(-msg => "Skipping Client deconfig for client '$client_name' because it is a virtual machine");
        return $SanMgr::TRUE;
    }

    # get the sanclient object for this client
    my ($client_obj) = $self->get_objects(-type  => "client",
                                          -names => [$client_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get client ($client_name) object");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to this host
    $client_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($client_obj->error()) {
        $self->_set_err("Failed to reconnect to client ($client_name)", $client_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # kill off any running I/O program on the client systems.
    $self->_kill_io_instances_subtest(-client => $client_name,
                                      -resource => $resource,
                                      -client_io_program => "all");
    if ($self->error()) {
        $self->_set_err("Failed to kill IO instances");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $data;
    foreach my $i (1..5) {
        # obtain the vserver targets that are associated with this host
        ($data) = $self->{parent}->get_target_list(-client_name => $client_name,
                                                   -filter_iscsi_targets_by_portset => $client_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                   -filter_fcp_targets_by_portset   => $client_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
        if ($self->{parent}->error()) {
            if ($i >= 5) {
                # error state
                last;
            } else {
                # retry
                $self->{parent}->_clear_err();
                sleep 20;
                next;
            }
        } else {
            # success
            last;
        }
    }

    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get iSCSI/FCP target lists");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    my $protocol = $client_obj->get_protocol();

    # deconfigure iSCSI/FCP
    if ($protocol =~ /iscsi|fcp/i and scalar(@iSCSITargets)) {
        $client_obj->deconfigure(-targets => \@iSCSITargets,
                                 -clean_level => $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($client_obj->error()) {
            $self->_set_err("Failed to deconfigure $protocol for the client ($client_name)", $client_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # clean up any orphaned devices
    $client_obj->orphaned_device_cleanup();
    if ($client_obj->error()) {
        $self->_set_err("Failed cleanup orphaned devices", $client_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # deconfigure NFS
    if ($protocol =~ /nfs/i) {
        # NFS currently has no deconfiguration steps to perform
    }

    &SubIPC::pass(-msg => "Client '$client_name' has been deconfigured.");

    return $SanMgr::TRUE;
}

=pod

=head1 Package Name

San::SanMgr::Client::deconfig

=head1 Author

Robert Hunter

=head1 Description

Provide a common set of libraries to work on resource deconfig

=head1 See Also

Package inherits from the following modules. See these modules for further

interface detail.

B<Error.pm>

B<San/SanLog.pm>

B<San/SanUtil>

B<ParamsMgr.pm>

=head1 Public Methods

=head2 B<deconfig_single($self, %args)>

This function provisions a client with the provided resource data.

It is intended to be used within a subtest. The function will issue a SubIPC::pass or fail

to inform the caller of final result.

=over 4

=item Inputs

=item * I<-resource>

Hash containing resource data to be used when provisioning the resource

=item Outputs

By using SubIPC::write this function can return the device list provisioned. A message with

the extra data of 'device_list' is sent to the caller. Its the responsibility of the caller

to retrieve the message and pull out the device list for future use.

=item Examples

&San::SanMgr::Client::deconfig::deconfig_single($self, -resource => {-name => "client1", -protocol => "fcp" ..});

=back

=cut

1;
