#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on resource setup
## @author latif@netapp.com
## @status private
## @pod hereN

package San::SanMgr::Client::setup;

# Compiler directives.
use strict;
use San::SanLog;
use ParamsMgr;
use San::SanUtil;
use NACL::Exceptions::NoElementsFound;
use NACL::CS::Volume;
use NACL::C::Client;
use NACL::C::Client::Mount;
use NACL::C::Client::Directory;
use NATE::BaseException qw(:try);
use EnGenX::Util;
use Digest::MD5 qw(md5_hex);
use Data::Dumper;

use vars qw(@ISA);
@ISA = qw(Error SanLog);

###############################################################################
# Method:    setup_single
# Objective: Setup run against a single resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # $self is a Client object
    # $resource is a hash of parameters for a particular resource to work on
    $self->sanlog(-msg => "Resource parameters/data:\n" . Dumper($resource),
                  -msg_level => 3);

    # Load up the resource information and get them via get_arg
    my $caller_id = $resource->{NAME} . "|" . time();
    #&ParamsMgr::process_args(%{$resource});
    &ParamsMgr::process_args(%{$resource}, -CALLER => $caller_id);

    #############################################
    ### General client parameters
    #############################################
    my $client_name =               &ParamsMgr::get_arg(-arg => "NAME", -required => 1, -default_to_env => 1,
                                                        -CALLER => $caller_id);
    my $protocol =                  &ParamsMgr::get_arg(-arg => "PROTOCOL", -required => 1, -default_to_env => 1,
                                                        -CALLER => $caller_id);
    my $io_type =                   &ParamsMgr::get_arg(-arg => "IO_TYPE", -required => 1, -default_to_env => 1,
                                                        -CALLER => $caller_id);
    my $fs_type =                   &ParamsMgr::get_arg(-arg => "FS_TYPE", -required => 0, -default_to_env => 1,
                                                        -CALLER => $caller_id, -default => "");
    my $fs_rotate =                 &ParamsMgr::get_arg(-arg => "FS_ROTATE", -required => 0, -default_to_env => 1,
                                                        -CALLER => $caller_id, -default => "0");
    my $lun_count =                 &ParamsMgr::get_arg(-arg => "LUN_COUNT", -required => 0, -default_to_env => 1,
                                                        -default => "", -CALLER => $caller_id);
    my $lun_path_regex =            &ParamsMgr::get_arg(-arg => "LUN_PATH_REGEX", -required => 0, -default_to_env => 1,
                                                        -default => undef, -CALLER => $caller_id);
    $lun_path_regex = eval($lun_path_regex);
    my $mount_options_pass_thru =   &ParamsMgr::get_arg(-arg => "MOUNT_OPTIONS_PASS_THRU", -required => 0, -default_to_env => 1,
                                                        -default => "", -CALLER => $caller_id);
    my $mount_options_discard =     &ParamsMgr::get_arg(-arg => "MOUNT_OPTIONS_DISCARD", -required => 0, -default_to_env => 1,
                                                        -default => 0, -CALLER => $caller_id);
    my $debug_level =               &ParamsMgr::get_arg(-arg => "DEBUG_LEVEL", -required => 0, -default_to_env => 1,
                                                        -default => 0, -CALLER => $caller_id);

    #############################################
    ### NTP Sync parameters
    #############################################
    my $ntp_sync =                  &ParamsMgr::get_arg(-arg => "NTP_SYNC", -required => 0, -default_to_env => 1,
                                                        -default => 0, -CALLER => $caller_id);
    my $ntp_time_server =           &ParamsMgr::get_arg(-arg => "NTP_TIME_SERVER", -required => 1, -default_to_env => 1,
                                                        -default => "ntp0.eng.netapp.com", dependant_on => ["NTP_SYNC"],
                                                        -CALLER => $caller_id);
    my $ntp_time_zone =             &ParamsMgr::get_arg(-arg => "NTP_TIME_ZONE", -required => 1, -default_to_env => 1,
                                                        -default => "EST5EDT", -dependant_on => ["NTP_SYNC"],
                                                        -CALLER => $caller_id);

    #############################################
    ### VG parameters
    #############################################
    my $mp_type =                   &ParamsMgr::get_arg(-arg => "MP_TYPE", -required => 0, -default_to_env => 1,
                                                        -default => "", -CALLER => $caller_id);
    my $lvm_type =                  &ParamsMgr::get_arg(-arg => "LVM_TYPE", -required => 0, -default_to_env => 1,
                                                        -default => "", -CALLER => $caller_id);
    my $vm_vol_size =                &ParamsMgr::get_arg(-arg => "VM_VOL_SIZE", -required => 1, -default_to_env => 1,
                                                        -default => "1g", -dependant_on => ["lvm_type"],
                                                        -CALLER => $caller_id);
    my $vm_num_lvs =                &ParamsMgr::get_arg(-arg => "VM_NUM_LVS", -required => 1, -default_to_env => 1,
                                                        -default => "1", -dependant_on => ["lvm_type"],
                                                        -CALLER => $caller_id);
    my $vm_num_fcp_lvs =            &ParamsMgr::get_arg(-arg => "VM_NUM_FCP_LVS", -required => 1, -default_to_env => 1,
                                                        -default => "0", -dependant_on => ["lvm_type"],
                                                        -CALLER => $caller_id);
    my $vm_num_iscsi_lvs =          &ParamsMgr::get_arg(-arg => "VM_NUM_ISCSI_LVS", -required => 1, -default_to_env => 1,
                                                        -default => "0", -dependant_on => ["lvm_type"],
                                                        -CALLER => $caller_id);
    my $vm_lv_type =                &ParamsMgr::get_arg(-arg => "VM_LV_TYPE", -required => 1, -default_to_env => 1,
                                                        -default => "fs", -dependant_on => ["lvm_type"],
                                                        -CALLER => $caller_id);
    my $vm_vg_layout =              &ParamsMgr::get_arg(-arg => "VM_VG_LAYOUT", -required => 1, -default_to_env => 1,
                                                        -default => "concat", -dependant_on => ["lvm_type"],
                                                        -CALLER => $caller_id);
    my $vm_stripe_size =          &ParamsMgr::get_arg(-arg => "VM_STRIPE_SIZE", -required => 1, -default_to_env => 1,
                                                        -default => "64k", -dependant_on => ["lvm_type"],
                                                        -CALLER => $caller_id);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Setup some basic default
    my $vm_vg_name = "n_vg";
    my $cf_mode = "ssi";
    my $device_list;
    my $device_data;

    # Get client object
    my ($hostobj) = $self->get_objects(-type => "client", -names => [$client_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get client ($client_name) object: ", $self->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to the client
    $hostobj->reconnect(-disconnect_existing_conn => 0);
    if ($hostobj->error()) {
        $self->_set_err("Failed to reconnect to client '$client_name'", $hostobj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # kill off any running I/O program on the client systems.
    $self->_kill_io_instances_subtest(-client => $client_name,
                                      -resource => $resource,
                                      -client_io_program => "all");
    if ($self->error()) {
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # NTP time sync the host
    if ($ntp_sync == 1) {
        if ($hostobj->can("ntp_sync")) {
            $hostobj->ntp_sync(-ntp_server => $ntp_time_server,
                               -ntp_time_zone => $ntp_time_zone);
            if ($hostobj->error()) {
                my @error = $hostobj->errmsg();
                $self->sanlog(-msg => "Failed to NTP time sync the $client_name (@error))",
                              -msg_type => "warn");
                $hostobj->_clear_err();
            }
        } else {
            $self->sanlog("No method defined for SanClient::ntp_sync() Skipping NTP setup on $client_name");
        }
    }

    if ($protocol =~ /fcp|iscsi/i) {

        # Setup the protocols array with the protocols specified by caller
        my @protocols = ();
        if ($protocol =~ /fcp/i) {
            push(@protocols, "fcp");
        }
        if ($protocol =~ /iscsi/i) {
            push(@protocols, "iscsi");
        }

        my ($vserver_list_ref, $luns_ref) = $self->{parent}->get_luns_and_vservers_associated_to_client(-client_name => $client_name);
        if ($self->{parent}->error()) {
            $self->{parent}->_set_err("Failed to get vservers associated to client (" . $client_name . ")");
            &SubIPC::fail(-object => $self->{parent});
            return $SanMgr::FALSE;
        }
        my @vserver_list = @$vserver_list_ref;
        my @luns = @$luns_ref;

        # default num expected luns to at least 1 lun
        my $num_expected_luns = scalar(@luns) ? scalar(@luns) : 1;

        # discover resources on the host. This includes configuration
        # information (configuration files) and sanlun information.
        # It wil also run devfsadm to discover LUNs
        $hostobj->discover(-num_expected_luns => $num_expected_luns);
        if ($hostobj->error) {
            $self->_set_err("Host discover failed", $hostobj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # Create a device manager object
        my $devmgrobj = $hostobj->create_devmgr();
        if ($hostobj->error) {
            $self->_set_err("Failed to create device manager obj", $hostobj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # get the list of devices pathnames to use
        # on disk and filesystem operations. This is
        # extracted from the object which got this information
        # from sanlun
        my @devices = $devmgrobj->get_free_devices();
        if ($devmgrobj->error()) {
            $self->_set_err("Unable to run get_free_devices", $devmgrobj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        if (scalar(@devices) > 0) {
            $self->sanlog(-msg => "List of discovered raw devices includes:\n" . Dumper(@devices));
        } else {
            $self->_set_err("No available devices found");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        if ($lvm_type ne "" and $lvm_type ne "none") {
            # Make sure the volume size is set
            if ($vm_vol_size eq "") {
                $self->_set_err("VM volume size is not specified");
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            # create a lvm object
            my $lvm_obj = $hostobj->create_lvm(-lvm => $lvm_type);
            if ($hostobj->error) {
                $self->_set_err("Failed to run create lvm object", $hostobj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            # after we have configured (and created LVM object), validate it's config
            # to make sure everything is setup properly. Currently only warning
            # later we may want to make this more severe
            $hostobj->validate(-cfo_mode => $cf_mode);
            if ($hostobj->error) {
                $self->sanlog(-msg => "Validation of host ($client_name) failed",
                              -msg_type => "warn");
                $hostobj->_clear_err();
            }

            # Default the vm_num_fcp_lvs and vm_num_iscsi_lvs to vm_num_lvs if they are not defined
            if ($vm_num_fcp_lvs eq "" or $vm_num_fcp_lvs == 0) {
                $vm_num_fcp_lvs = $vm_num_lvs;
            }

            if ($vm_num_iscsi_lvs eq "" or $vm_num_iscsi_lvs == 0) {
                $vm_num_iscsi_lvs = $vm_num_lvs;
            }

            $self->sanlog(-msg => "Getting list of free devices\n\tvservers: @vserver_list\n\tprotocols: @protocols");

            # Collect, init and prepare all devices of protocol type fcp or iscsi
            foreach my $protocol (@protocols) {
                # Get the init states "online", "online invalid", "error", "online thinrclm", or "online thin".
                # "online thinrclm" and "online thin" are the same thing.  Some operating systems use thinrcl
                # others only use thin. "online invalid" and "online invalid thinrclm" are the same depending on vsersion of Veritas.
                my @devices_tmp = $lvm_obj->get_free_devices(-inited_states => ["online invalid", "online invalid thinrclm", "error", "online", "online thinrclm", "online thin"],
                                                             -protocol => $protocol,
                                                             -filers => \@vserver_list);
                if ($lvm_obj->error) {
                    $self->_set_err("Failed to get list of free $protocol devices from LVM", $lvm_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                if (scalar(@devices_tmp) == 0) {
                    $self->_set_err("Failed to get list of free $protocol devices from LVM (device list was empty)");
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                $self->sanlog(-msg => "Initializing (" . scalar(@devices_tmp) . ") $protocol devices");
                # Init list of protocol devices
                $lvm_obj->initialize(-devices => \@devices_tmp);
                if ($lvm_obj->error) {
                    my $device_list = join("\n", @devices_tmp);
                    $self->_set_err("Failed to initialize $protocol devices\n:${device_list}", $lvm_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                $self->sanlog(-msg => "Creating volume group for $protocol with the following devices:" .
                              join("\n\t", @devices_tmp));
                # Now create a volume group using the list of protocol devices that we discovered.
                $device_data->{$protocol}->{'vg_obj'} = $lvm_obj->create_vg(-devices => \@devices_tmp,
                                                                            -name => $vm_vg_name . "_" . substr($protocol, 0, 1));

                if ($lvm_obj->error) {
                    $self->_set_err("Failed to create vg for $protocol devices", $lvm_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                if ($protocol =~ /fcp/i) {
                    $device_data->{$protocol}->{'num_lvs'} = $vm_num_fcp_lvs;
                } else {
                    $device_data->{$protocol}->{'num_lvs'} = $vm_num_iscsi_lvs;
                }

                $self->sanlog(-msg => "Creating " . $device_data->{$protocol}->{'num_lvs'} . " logical volumes on " . $vm_vg_name . "_" . substr($protocol, 0, 1));
            }

            # Go through devices found for each protocol
            foreach my $protocol (keys(%{$device_data})) {
                my $vg_obj     = $device_data->{$protocol}->{'vg_obj'};
                my $num_lvs    = $device_data->{$protocol}->{'num_lvs'};
                my @vg_devices = $vg_obj->get_devices();

                # Special case data for SVM lvm type
                my $num_dev_per_vol;
                my @svm_devices;
                if ($lvm_type =~ /svm/i) {
                    foreach my $tmp_device (@vg_devices) {
                        if ($tmp_device =~ /s\d+$/) {
                            $tmp_device =~ s/s\d+$/s0/g;
                        } else {
                            $tmp_device .= "s0";
                        }
                        push @svm_devices, $tmp_device;
                    }

                    # Calculate the number of devices per volume for SVM
                    $num_dev_per_vol = scalar(@svm_devices)/ $num_lvs;
                    $num_dev_per_vol = int($num_dev_per_vol);
                }

                $self->sanlog(-msg => "Using vg object for protocol $protocol, creating " .
                                      "$num_lvs logical volumes on vg " .
                                      $vg_obj->get_name());

                my $fs_type_index = 0;
                my $mixed_fs_type_verified = 0;
                my @fs_type_mixed;

                for (my $vol = 1; $vol <= $num_lvs; $vol++) {
                    my $volume_name = "nvol-" . $vol . "-" . substr($protocol, 0, 1);

                    $self->sanlog(-msg => "\nCreating Logical Volume ($volume_name) with parameters:\n" .
                                          "\tLayout: $vm_vg_layout\n" .
                                          "\tVolume Size: $vm_vol_size\n" .
                                          "\tProtocol: $protocol\n");

                    my $vol_obj;
                    if ($lvm_type =~ /svm/i) {
                        # Special case for SVM lvm type.  We need to tell it what devices to
                        # use for each volume.
                        my $device_list_per_vol;
                        for my $i (1 .. $num_dev_per_vol) {
                            $device_list_per_vol = $device_list_per_vol . shift(@svm_devices) . " ";
                        }

                        $vol_obj = $vg_obj->create_vol(-layout           => $vm_vg_layout,
                                                       -size             => $vm_vol_size,
                                                       -name             => $volume_name,
                                                       -num_dev_per_vol  => $num_dev_per_vol,
                                                       -num_of_vols      => $num_lvs,
                                                       -dev_per_vol_list => $device_list_per_vol);

                        if ($vg_obj->error) {
                            $self->_set_err("Failed to create volume object", $vg_obj->errmsg());
                            &SubIPC::fail(-object => $self);
                            return $SanMgr::FALSE;
                        }
                    } else {
                        $vol_obj = $vg_obj->create_vol(-layout       => $vm_vg_layout,
                                                        -size        => $vm_vol_size,
                                                        -stripe_size => $vm_stripe_size,
                                                        -name        => $volume_name);

                        if ($vg_obj->error) {
                            $self->_set_err("Failed to create volume object", $vg_obj->errmsg());
                            &SubIPC::fail(-object => $self);
                            return $SanMgr::FALSE;
                        }
                    }

                    my $device_info;

                    # Save some attributes of the device onto the device_info hash
                    ($device_info->{'vserver'}, $device_info->{'volume'}, $device_info->{'lun'},
                     $device_info->{'protocol'}, $device_info->{'serial'}) =
                        ((), (), (), (), $protocol);

                    # Determine if the caller wants filesystems created on the
                    # newly created volumes
                    if ($io_type =~ /fs/) {
                        if ($vm_lv_type !~ /:|mixed/i) {
                            # Same file system will be created on all the volumes.
                            # Create a fs based on the vm_lv_type requested.
                            $vol_obj->create_fs(-type => $vm_lv_type);
                            if ($vol_obj->error) {
                                $self->_set_err("Failed to create filesystem", $vol_obj->errmsg());
                                &SubIPC::fail(-object => $self);
                                return $SanMgr::FALSE;
                            }
                        } else {
                            # Mixed file systems has been requested by user.
                            # There are 2 options,
                            # 1. When user directly provides colon separated fs list in vm_lv_type.
                            # 2. When vm_lv_type is set to "mixed", create mixed fs from the template provided in host libraries.
                            #    It various from one OS to another. (Currently this option is supported only in Linux).
                            if ($mixed_fs_type_verified == 0) {
                                # Initial setup required for mixed fs. Needs to be done only once in the first iteration.
                                if ($vm_lv_type =~ /:/) {
                                    # vm_lv_type is set to colon separated fs by user
                                    &SanUtil::nate_log('-comment', "Mixed FS requested by user: $vm_lv_type");
                                    @fs_type_mixed = split(/:/, $vm_lv_type);
                                } elsif ($vm_lv_type =~ /mixed/i) {
                                    # vm_lv_type is set to "mixed".
                                    # Get mixed fs type template from host library.
                                    if ($hostobj->can("get_mixed_fs_template")) {
                                        # Currently it is supported only on Linux
                                        my @fs_array = $hostobj->get_mixed_fs_template();
                                        my $vm_lv_type_temp = join(":", @fs_array);
                                        &SanUtil::nate_log('-comment', "Mixed FS template: $vm_lv_type_temp");
                                        @fs_type_mixed = split(/:/, $vm_lv_type_temp);
                                    } else {
                                        my $hostname = $hostobj->{host};
                                        $self->_set_err("Mixed file systems are not supported yet for $hostname");
                                        &SubIPC::fail(-object => $self);
                                        return $SanMgr::FALSE;
                                    }
                                }
                                $mixed_fs_type_verified = 1;
                            }

                            $vol_obj->create_fs(-type => $fs_type_mixed[$fs_type_index]);
                            if ($vol_obj->error) {
                                $self->_set_err("Failed to create filesystem", $vol_obj->errmsg());
                                &SubIPC::fail(-object => $self);
                                return $SanMgr::FALSE;
                            }

                            # increment or reset counter to determine which fs type to use on this loop.
                            if ($fs_type_index == (scalar(@fs_type_mixed) - 1)) {
                                $fs_type_index = 0;
                            } else {
                                $fs_type_index++;
                            }
                        }

                        # Mount the newly created fs
                        if ($vol_obj->is_mounted() == 0) {
                            $vol_obj->mount(-pass_thru => $mount_options_pass_thru,
                                            -discard => $mount_options_discard);
                            if ($vol_obj->error) {
                                $self->_set_err("Failed to mount filesystem", $vol_obj->errmsg());
                                &SubIPC::fail(-object => $self);
                                return $SanMgr::FALSE;
                            }
                        } else {
                            $self->sanlog(-msg => "file-system (" . $vol_obj->{mount} . ") on $volume_name " .
                                                  "is already mounted",
                                          -msg_type => "warn");
                        }

                        $device_info->{path} = $vol_obj->{mount};
                        $device_info->{io_type} = "fs";
                    } else {
                        $device_info->{path} = $vol_obj->get_raw_device();
                        $device_info->{io_type} = "raw";
                    }

                    push(@{$device_list}, $device_info);
                }
            }
        } else {
            # after we have configured validate it's config (non LVM case)
            # to make sure everything is setup properly. Currently only warning
            # later we may want to make this more severe
            $hostobj->validate(-cfo_mode => $cf_mode);
            if ($hostobj->error) {
                $self->_set_err("Validation of host ($client_name) failed", $hostobj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            $self->sanlog(-msg => "Getting list of free devices");
            my @devices = ();

            # If the host is a virtual client, only list a single item in the protocols array to gather all
            # devices found for provisioning.
            if ($hostobj->{is_virtual} == 1) {
                $self->sanlog(-msg => "Client ($client_name) is a virtual client, gathering list of all available devices");
                @protocols = ("all");
            }

            # Go through each protocol defined and get the list of devices to use for provisioning
            foreach my $protocol (@protocols) {
                my @device_names;

                # If the protocol is all, use get_free_devices without filters
                if ($protocol eq "all") {
                    my @devices_tmp = $devmgrobj->get_devices();
                    foreach my $device (@devices_tmp) {
                        push(@device_names, $device->{name});
                    }
                    $protocol = "";
                } else {
                    # Get the init states "online", "online invalid thinrclm", "online invalid", "error", "online thinrclm", or "online thin".
                    # "online thinrclm" and "online thin" are the same thing.  Some operating systems use thinrcl
                    # others only use thin. "online invalid" and "online invalid thinrclm" are the same depending on vsersion of Veritas.
                    @device_names = $devmgrobj->get_free_devices(-inited_states => ["online invalid", "online invalid thinrclm", "error", "online", "online thin", "online thinrclm"],
                                                                 -protocol => $protocol,
                                                                 -filers => \@vserver_list);
                }
                if ($devmgrobj->error) {
                    $self->_set_err("Failed to get list of free $protocol devices from DeviceMgr",
                                    $devmgrobj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                if (scalar(@device_names) == 0) {
                    $self->_set_err("Failed to get list of free $protocol devices from DeviceMgr " .
                                    "(device list was empty)");
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                # perform lun path regex filters on the discovered devices
                if (defined($lun_path_regex)) {
                    my @filtered_out_luns;
                    my @filtered_in_luns;
                    $self->sanlog(-msg => "Filtering list of devices with lun path regex filter '$lun_path_regex'");
                    @device_names = grep {
                        my $dev_name = $_;
                        my $dev_obj = $devmgrobj->get_device($dev_name);
                        my ($vserver, $lun_path) = ($dev_obj->{'lun'} =~ /^\s*(\S+)\s*\:\s*(\S+)\s*$/);
                        if ($lun_path !~ $lun_path_regex) {
                            # filter out luns that dont match the filter
                            push(@filtered_out_luns, "$dev_name\t: $vserver:$lun_path");
                            0;
                        } else {
                            # this lun matches
                            push(@filtered_in_luns, "$dev_name\t: $vserver:$lun_path");
                            1;
                        }
                    } @device_names;

                    $self->sanlog(-msg => "Device list (protocol:$protocol) :\n" .
                                      "  The following LUNs will be used:\n\t"     . join("\n\t", @filtered_in_luns) . "\n\n" .
                                      "  The following LUNs will not be used:\n\t" . join("\n\t", @filtered_out_luns));
                }

                # Filter the device list if the user specifies a lun count that they want to use
                if ($lun_count ne "") {
                    $self->sanlog(-msg => "Filtering list of devices by the specified lun count: $lun_count");
                    @device_names = $hostobj->client_lun_filter($lun_count, @device_names);
                    if ($hostobj->error) {
                        $self->_set_err("Failed to apply LUN count ($lun_count) for $protocol client ($client_name)",
                                       $hostobj->errmsg());
                        &SubIPC::fail(-object => $self);
                        return $SanMgr::FALSE;
                    }
                    $self->sanlog(-msg => "Device list (protocol:$protocol) has been filtered down to:\n" . Dumper(@device_names));
                }

                # If io_type is filesystem 'fs', format the list of protocol devices found
                if ($io_type =~ /fs/) {
                    # format the devices within the @device_names array
                    if ($hostobj->isa("San::SanClient::Host::WinHost")) {
                        $hostobj->format(-devices => \@device_names, -type => $fs_type, -fs_rotate => $fs_rotate);
                    } else {
                        $hostobj->format(-devices => \@device_names);
                    }
                    if ($hostobj->error()) {
                        $self->_set_err("Format failed on $protocol devices", $hostobj->errmsg());
                        &SubIPC::fail(-object => $self);
                        return $SanMgr::FALSE;
                    }
                }

                # Go through each device and set the protocol attribute on the device object
                foreach my $device_name (@device_names) {
                    my $device = $devmgrobj->get_device($device_name);
                    if ($devmgrobj->error) {
                        $self->_set_err("Failed to get device object for device ($device_name)",
                                        $devmgrobj->errmsg());
                        &SubIPC::fail(-object => $self);
                        return $SanMgr::FALSE;
                    }

                    $device->{protocol} = $protocol;
                }

                push(@devices, @device_names);
            }

            my $fs_type_index = 0;
            my $mixed_fs_type_verified = 0;
            my @fs_type_mixed;

            foreach my $device_name (@devices) {
                my $device = $devmgrobj->get_device($device_name);
                my $device_info;

                if ($devmgrobj->error) {
                    $self->_set_err("Failed to get device object for device ($device_name)",
                                    $devmgrobj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                # Save some attributes of the device onto the device_info hash
                my ($vserver, $volume, $lun) = ($device->{'lun'} =~ /^(\S+)\:\/vol\/(\S+)\/(\S+)$/);
                push(@{$device_info->{'vserver'}}, $vserver);
                push(@{$device_info->{'volume'}}, $volume);
                push(@{$device_info->{'lun'}}, $lun);
                $device_info->{'protocol'} = $device->{protocol};

                # Push the device serial number onto the device info hash if it is defined
                if (defined($device->{serial}) and $device->{serial} ne "") {
                    push(@{$device_info->{'serial'}}, $device->{serial});
                }

                # Running NON CLIENT_VMS
                if ($io_type =~ /fs/) {
                    # For windows, format already does the format() call,
                    # so there is no need to do again.
                    if ($hostobj->hosttype() !~ /windows/i) {
                        if ($fs_type !~ /:|mixed/i) {
                            # Same file system will be created on all the volumes.
                            # Create a fs based on the vm_lv_type requested.
                            $device->create_fs(-type => $fs_type);
                            if ($device->error()) {
                                $self->_set_err("Create FS failed on device ($device->{name})",
                                                $device->errmsg());
                                &SubIPC::fail(-object => $self);
                                return $SanMgr::FALSE;
                            }
                        } else {
                            # Mixed file systems has been requested by user.
                            # There are 2 options,
                            # 1. When user directly provides colon separated fs list in fs_type.
                            # 2. When fs_type is set to "mixed", create mixed fs from the template provided in host libraries.
                            #    It various from one OS to another. (Currently this option is supported only in Linux).
                            if ($mixed_fs_type_verified == 0) {
                                # Initial setup required for mixed fs. Needs to be done only once in the first iteration.
                                if ($fs_type =~ /:/) {
                                    # fs_type is set to colon separated fs by user
                                    &SanUtil::nate_log('-comment', "Mixed FS requested by user: $fs_type");
                                    @fs_type_mixed = split(/:/, $fs_type);
                                } elsif ($fs_type =~ /mixed/i) {
                                    # fs_type is set to "mixed".
                                    # Get mixed fs type template from host library.
                                    if ($hostobj->can("get_mixed_fs_template")) {
                                        # Currently it is supported only on Linux
                                        my @fs_array = $hostobj->get_mixed_fs_template();
                                        my $fs_type_temp = join(":", @fs_array);
                                        &SanUtil::nate_log('-comment', "Mixed FS template: $fs_type_temp");
                                        @fs_type_mixed = split(/:/, $fs_type_temp);
                                    } else {
                                        my $hostname = $hostobj->{host};
                                        $self->_set_err("Mixed file systems are not supported yet for $hostname");
                                        &SubIPC::fail(-object => $self);
                                        return $SanMgr::FALSE;
                                    }
                                }
                                $mixed_fs_type_verified = 1;
                            }

                            $device->create_fs(-type => $fs_type_mixed[$fs_type_index]);
                            if ($device->error()) {
                                $self->_set_err("Create FS failed on device ($device->{name})",
                                                $device->errmsg());
                                &SubIPC::fail(-object => $self);
                                return $SanMgr::FALSE;
                            }

                            # increment or reset counter to determine which fs type to use on this loop.
                            if ($fs_type_index == (scalar(@fs_type_mixed) - 1)) {
                                $fs_type_index = 0;
                            } else {
                                $fs_type_index++;
                            }
                        }

                     }

                    $device->mount_fs(-discard => $mount_options_discard);
                    if ($device->error()) {
                        $self->_set_err("Mount FS failed on device ($device->{name})", $device->errmsg());
                        &SubIPC::fail(-object => $self);
                        return $SanMgr::FALSE;
                    }

                    $device->load();
                    if ($device->error()) {
                        $self->_set_err("Load of information failed for device ($device->{name})",
                                        $device->errmsg());
                        &SubIPC::fail(-object => $self);
                        return $SanMgr::FALSE;
                    }

                    $device_info->{path} = $device->get_mount();
                    $device_info->{io_type} = "fs";
                } else {
                    # For each device in the device list, add the raw
                    # name to the fs_return variable.
                    $device_info->{path} = $device->get_raw_device();
                    $device_info->{io_type} = "raw";
                }

                push(@{$device_list}, $device_info);
            }
        }
    }

    # setup NFS for unix hosts hosts
    if ($protocol =~ /nfs/i) {
        my $nfs_target_hash;
        try {
            $nfs_target_hash = $self->{parent}->EnGenX::Util::get_nfs_target_relationship_hash(client_name => $client_name);
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # create the Command interface for this client
        my $client_ci;
        try {
            $client_ci = new NACL::C::Client(name => $client_name,
                                             apiset_must => {conntype => 'ssh-persist',
                                                             connid => 'ssh'});
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # build out the nfs options datastructure up front
        # this will contain a sweet lookup table that we can search through when trying to
        # find matches for nfs_options
        my $all_nfs_options;
        try {
            $all_nfs_options = $self->EnGenX::Util::generate_nfs_options(resource => $resource);
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        my $match = 0;
        # we should now have a list of volumes to mount and a list of lifs to mount each volume
        foreach my $cluster_name (keys %{$nfs_target_hash}) {
            # skip if this cluster does not match the filter
            next if (!$self->EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                  cluster_name => $cluster_name));
            foreach my $vserver_name (keys %{$nfs_target_hash->{$cluster_name}->{vservers}}) {
                # skip if this vserver does not match the filter
                next if (!$self->EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                      cluster_name => $cluster_name,
                                                                      vserver_name => $vserver_name));

                foreach my $lif_obj (@{$nfs_target_hash->{$cluster_name}->{vservers}->{$vserver_name}->{lif_objs}}) {
                    my $lif_address = $lif_obj->address();
                    my $lif_name = $lif_obj->lif();
                    my $lif_curr_node = $lif_obj->curr_node();

                    # skip if the lif address / lif name don't match
                    next if (!$self->EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                          cluster_name => $cluster_name,
                                                                          vserver_name => $vserver_name,
                                                                          lif_address => $lif_address,
                                                                          lif_name => $lif_name));

                    foreach my $volume_name (keys %{$nfs_target_hash->{$cluster_name}->{vservers}->{$vserver_name}->{volumes}}) {
                        my $vol_obj = $nfs_target_hash->{$cluster_name}->{vservers}->{$vserver_name}->{volumes}->{$volume_name}->{vol_obj};
                        my $junction_path = $vol_obj->junction_path();
                        my $aggr_obj = $nfs_target_hash->{$cluster_name}->{aggr_objs}->{$vol_obj->aggregate()};
                        my $vol_curr_node = $aggr_obj->nodes()->[0];

                        # discover the affinity between the volume and the lif (local|remote) based
                        # on the node
                        my $actual_affinity;
                        if ($lif_curr_node eq $vol_curr_node) {
                            $actual_affinity = "local";
                        } else {
                            $actual_affinity = "remote";
                        }

                        # now that we have all our fields, lets do one final lookup
                        # to find the nfs options that we can use for this volume
                        my $nfs_options;
                        try {
                            $nfs_options = $self->EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                                       cluster_name => $cluster_name,
                                                                                       vserver_name => $vserver_name,
                                                                                       lif_address => $lif_address,
                                                                                       lif_name => $lif_name,
                                                                                       volume_name => $volume_name,
                                                                                       junction_path => $junction_path,
                                                                                       volume_lif_affinity => $actual_affinity);
                        } otherwise {
                            my $e = shift;
                            $self->_set_err("Exception caught: " . $e->text());
                        };
                        if ($self->error()) {
                            &SubIPC::fail(-object => $self);
                            return $SanMgr::FALSE;
                        }

                        # skip this junction path if we could not find an nfs options match
                        next if (!$nfs_options);

                        # translate the nfs_options into parameters that NACL::C::Client::Mount->create() understands
                        my $mount_options = $self->EnGenX::Util::_translate_nfs_options_for_mount(nfs_options => $nfs_options);

                        # construct the mount point path so that it is unique to the cluster/vserver/lif/junction_path
                        # make sure there are no dashes in the mountpoint
                        # note: $junction_path already starts with a /
                        my $mount_point = "/mnt/engenx/${cluster_name}/${vserver_name}/${lif_address}${junction_path}";

                        # mount this volume on this lif
                        my $mount;
                        try {
                            $mount = NACL::C::Client::Mount->create(command_interface => $client_ci,
                                                                    server => $lif_address,
                                                                    path => $junction_path,
                                                                    type => "nfs",
                                                                    mount_point => $mount_point,
                                                                    apiset_must => {connid => 'ssh',
                                                                                    conntype => 'ssh-persist'},
                                                                    %{$mount_options});
                        } otherwise {
                            my $e = shift;
                            $self->_set_err("Exception caught: " . $e->text());
                        };
                        if ($self->error()) {
                            &SubIPC::fail(-object => $self);
                            return $SanMgr::FALSE;
                        }
                        $mount_point = $mount->mount_point();

                        # create a unique sub directory in this mount point

                        my $uuid = md5_hex($mount_point);
                        $uuid = lc($uuid);
                        # chop off the first 24 chars
                        $uuid =~ s/^\S{24}//g;

                        if ($nfs_options->{create_unique_subdir}) {
                            $mount_point = $mount->mount_point() . "/${uuid}";
                            my (@output) = $hostobj->execute("mkdir $mount_point");
                            if ($hostobj->error() and not grep(/File exists/i, @output)) {
                                $self->_set_err($hostobj->errmsg());
                                &SubIPC::fail(-object => $self);
                                return $SanMgr::FALSE;
                            }
                            $hostobj->_clear_err();
                        }

                        # create the client dev list entry for this mount point
                        my $device_info;
                        $device_info->{path} = $mount_point;
                        $device_info->{io_type} = "fs";
                        $device_info->{protocol} = "nfs";
                        $device_info->{vserver_name} = $vserver_name;
                        $device_info->{cluster_name} = $cluster_name;
                        $device_info->{volume_name} = $volume_name;
                        $device_info->{qtree_name} = undef;
                        $device_info->{is_qtree} = 0;
                        $device_info->{junction_path} = $junction_path;
                        $device_info->{unique_subdir} = $nfs_options->{create_unique_subdir};
                        $device_info->{lif_name} = $lif_name;
                        $device_info->{lif_address} = $lif_address;
                        $device_info->{lif_volume_affinity} = $actual_affinity;
                        push(@{$device_list}, $device_info);

                        # skip qtrees if this volume explicitly specified to not mount qtrees
                        next if (!$nfs_options->{mount_qtrees});

                        # iterate through all qtrees in this volume and attempt to mount each one as well
                        foreach my $qtree_obj (@{$nfs_target_hash->{$cluster_name}->{vservers}->{$vserver_name}->{volumes}->{$volume_name}->{qtree_objs}}) {
                            my $qtree_name = $qtree_obj->qtree();
                            try {
                                $nfs_options = undef;
                                $nfs_options = $self->EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                                           cluster_name => $cluster_name,
                                                                                           vserver_name => $vserver_name,
                                                                                           lif_address => $lif_address,
                                                                                           lif_name => $lif_name,
                                                                                           volume_name => $volume_name,
                                                                                           qtree_name => $qtree_name,
                                                                                           junction_path => $junction_path,
                                                                                           volume_lif_affinity => $actual_affinity);
                            } otherwise {
                                my $e = shift;
                                $self->_set_err("Exception caught: " . $e->text());
                            };
                            if ($self->error()) {
                                &SubIPC::fail(-object => $self);
                                return $SanMgr::FALSE;
                            }

                            # skip this qtree if we could not find an nfs options match
                            next if (!$nfs_options);

                            # construct the mount point path so that it is unique to the cluster/vserver/lif/junction_path
                            # make sure there are no dashes in the mountpoint
                            # note: $junction_path already starts with a /
                            $mount_point = "/mnt/engenx/${cluster_name}/${vserver_name}/${lif_address}${junction_path}__qtree__${qtree_name}";

                            # mount this qtree
                            try {
                                $mount = NACL::C::Client::Mount->create(command_interface => $client_ci,
                                                                        server => $lif_address,
                                                                        path => $junction_path . '/' . $qtree_name,
                                                                        type => "nfs",
                                                                        mount_point => $mount_point,
                                                                        apiset_must => {connid => 'ssh',
                                                                                        conntype => 'ssh-persist'},
                                                                        %{$mount_options});
                            } otherwise {
                                my $e = shift;
                                $self->_set_err("Exception caught: " . $e->text());
                            };
                            if ($self->error()) {
                                &SubIPC::fail(-object => $self);
                                return $SanMgr::FALSE;
                            }
                            $mount_point = $mount->mount_point();

                            $uuid = md5_hex($mount_point);
                            $uuid = lc($uuid);
                            # chop off the first 24 chars
                            $uuid =~ s/^\S{24}//g;

                            # check if we need to create a unique mount point for this qtree
                            if ($nfs_options->{create_unique_subdir}) {
                                $mount_point = $mount->mount_point() . "/${uuid}";
                                my (@output) = $hostobj->execute("mkdir $mount_point");
                                if ($hostobj->error() and not grep(/File exists/i, @output)) {
                                    $self->_set_err($hostobj->errmsg());
                                    &SubIPC::fail(-object => $self);
                                    return $SanMgr::FALSE;
                                }
                                $hostobj->_clear_err();
                            }

                            # create the client dev list entry for this mount point
                            my $device_info;
                            $device_info->{path} = $mount_point;
                            $device_info->{io_type} = "fs";
                            $device_info->{protocol} = "nfs";
                            $device_info->{vserver_name} = $vserver_name;
                            $device_info->{cluster_name} = $cluster_name;
                            $device_info->{volume_name} = $volume_name;
                            $device_info->{qtree_name} = $qtree_name;
                            $device_info->{is_qtree} = 1;
                            $device_info->{junction_path} = $junction_path;
                            $device_info->{unique_subdir} = $nfs_options->{create_unique_subdir};
                            $device_info->{lif_name} = $lif_name;
                            $device_info->{lif_address} = $lif_address;
                            $device_info->{lif_volume_affinity} = $actual_affinity;
                            push(@{$device_list}, $device_info);
                        }
                    }
                }
            }
        }
    }

    # Write back to parent the device_list generated
    &SubIPC::write(-msg => "Completed client ($client_name) setup", -device_list => $device_list,
                   -type => "setup_complete", -name => $client_name);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg    => "Unable to write (setup_complete) message to parent",
                      -object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "Client setup complete");

    return $SanMgr::TRUE;
}

1;
