#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author latif@netapp.com
## @status private
## @pod here

package San::SanMgr::Controller;

# Compiler directives.
use strict;

# Module import.
use Tharn;
use Error;
use San::SanLog;
use San::SanMgr::Util;

use Data::Dumper;

# NACL Libs
use NACL::C::Node;
use NACL::STask::Cluster;
use NACL::C::SystemNodeDate;

# Load resource specific modules
use San::SanMgr::Controller::setup;
use San::SanMgr::Controller::config;
use San::SanMgr::Controller::cleanup;
use San::SanMgr::Controller::validate;
use San::SanMgr::Controller::create_cleanup_obj;
use San::SanMgr::Controller::ControllerHelper;

use vars qw(@ISA);

@ISA = qw(San::SanMgr::Util Error SanLog
          San::SanMgr::Controller::setup
          San::SanMgr::Controller::cleanup
          San::SanMgr::Controller::create_cleanup_obj
          San::SanMgr::Controller::ControllerHelper);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new resource object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($proto, %args) = @_;

    my $self = {};

    my $class = ref($proto) || $proto;
    bless ($self, $class);

    # Clear out the errors
    @{$self->{errors}} = ();

    # allow for error stack build ups
    $self->{error_dont_overwrite_errors} = 1;
    $self->{error_stack_style} = "normal";

    &ParamsMgr::process_args(%args);
    my $parent = &ParamsMgr::get_arg(-arg => "parent",
                                     -required => 1,
                                     -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Save parent
    $self->{parent} = $parent;

    # Get the set of environment parameters from ParamsMgr
    my $environment = &ParamsMgr::search(search_path => [],
                                         key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Set the objects debug level
    $self->debug_level(-debug_level => $environment->{DEBUG_LEVEL});
    if ($self->error()) {
        return $self;
    }

    return $self;
}

###############################################################################
# Method:    setup
# Objective: Launch subtest to setup all clusters
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup {
    my ($self, %args) = @_;

    $self->clear_error();
    # check to see if setup is disabled for all clusters
    my $clusters = &ParamsMgr::search(-search_path => [],
                                     -key => "CLUSTERS",
                                     paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my @messages;
    my $subtest;

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # perform setup if it is enabled across all clusters
    if ($clusters->{SETUP} == 1) {
        $subtest = &SubIPC::create(-id => "$run_id_prefix" . "Controller-setup",
                                   -esc => $self->{parent}->{enable_subtest_count},
                                   -run => sub { $self->_setup_all_clusters(); },
                                   -ipc => 1,
                                   -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }

        &SubIPC::wait_finish(-subtest => [$subtest]);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }

        # Get messages to process
        @messages = &SubIPC::get_messages(-subtest => [$subtest], -type => "setup_cluster_complete");
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }

        # if any messages were sent to us, process them
        if (scalar(@messages)) {
            my $controller_objs;
            # Go through the messages, pull the data out
            foreach my $message (@messages) {
                if (defined($message->{extra}->{obj_data})) {
                    $controller_objs = $message->{extra}->{obj_data};
                }
            }

            if (defined($controller_objs)) {
                # store this data on the sanmgr object if it exists
                unless ($self->{parent}->_store_controller_cluster_objs(%args,
                                                                        data => $controller_objs)) {
                    $self->_set_err("Failed to store controller objects in SanMgr");
                    return $SanMgr::FALSE;
                }
            }
        }

        my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
        if ($st_status->{failed} != 0) {
            $self->_set_err("A subtest failure has been reported:" .
                            "\n\t\tFailed: " . $st_status->{failed} .
                            "\n\t\tPassed: " . $st_status->{passed});
            return $SanMgr::FALSE;
        } else {
            $self->sanlog(-msg => "Controller setup has passed\n\t\tFailed: " .
                                  $st_status->{failed} . "\n\t\tPassed: " .
                                  $st_status->{passed});
        }

        &SubIPC::remove_subtest();
    } else {
        $self->sanlog(-msg => "Skipping Controller Setup");
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _setup_all_clusters
# Objective: Launch subtest for each cluster to perform setup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _setup_all_clusters {
    my ($self) = @_;

    $self->clear_error();

    # Local Variables
    my ($st_status);
    my ($reuse_policy) = "";
    my (@messages, @data) = ();
    my ($cluster_data) = undef;
    my @subtests = ();
    my %interclus_params = ();

    # Get list of controllers to run validation against
    my $clusters = &ParamsMgr::search(-search_path => [], -key => "CLUSTERS",
                                      -search_type => "",
                                      -key_filter => qr/^CLUSTER\d*$/,
                                      paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Get global reuse policy which determines if existing objects can be reused
    $reuse_policy = &ParamsMgr::get("REUSE_POLICY.SETUP", paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Default setting is "reuse". This means that SanMgr will match objects
    # by name and ignore other attributes. This policy can be overridden on an
    # element basis.
    unless (defined($reuse_policy)) {
         $reuse_policy = "reuse";
    }

    # Go through the list and identify unique clusters
    foreach my $cluster (keys(%{$clusters})) {

        # Adding reuse policy to the resource data structure
        $clusters->{$cluster}->{REUSE_POLICY} = $reuse_policy;

        # Check to see if resource is to be skipped
        if ($clusters->{$cluster}->{SETUP} == 0) {
            $self->sanlog(-msg => "Skipping cluster " .
                                  $clusters->{$cluster}->{NAME} .
                                  " because setup tag set to \"0\"",
                          -msg_level  => 1);
            next;
        }
        $self->sanlog(-msg => "Setting up cluster " .
                               $clusters->{$cluster}->{NAME},
                      -msg_level  => 1);

        #
        # Intercluster
        # Verify if the cluster will be connecting to a peer (intercluster requested)
        #
        if ( $self->_is_intercluster_requested('cluster' => $clusters->{$cluster}) ) {
            # Create the intercluster parameters, the hash that is required for the
            # calling ControllerCluster::Intercluster_Obj_Create
            # FIRST: Gather the arguments for that call.
            my %create_args=('cluster' => $clusters->{$cluster},
                             'interclus_params' => \%interclus_params
                            );
            # SECOND: Make the call
            unless ( $self->_create_intercluster_params(%create_args)){
                $self->_set_err("ERROR while creating the intercluster parameters. ");
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }# end intercluster create parameters

        my $subtest =
            &SubIPC::create(-id  => "setup_controller_cluster__" .
                                    $clusters->{$cluster}->{NAME},
                            -run => sub {
                                         $self->setup_controller_cluster(
                                            -resource => $clusters->{$cluster}
                                        );
                                    },
                            -parent => $self,
                            -ipc    => 1,
                            -bg     => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        push(@subtests, $subtest);
    }

    #
    # Intercluster
    #
    # If user has defined that provided clusters need to connect, the intercluster
    # parameters have been put together, so now setup the intercluster connection
    if ( defined ($interclus_params{'CLUSTERS'})){
        $self->_setup_intercluster('cluster'=>$clusters->{(keys %{$clusters})[-1]},
                                   'interclus_params'=> \%interclus_params);
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);

    @messages = &SubIPC::get_messages(-subtest => \@subtests, -type => "setup_cluster_content_complete");
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Only pass the data we need
    foreach my $message (@messages) {
        if (defined($message->{extra}->{obj_data})) {
            foreach my $obj_data (@{$message->{extra}->{obj_data}}) {
                push(@data, $obj_data);
            }
        }
    }

    $cluster_data = $self->sort_data_by_type(data => \@data);

    if ($self->error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Now bubble up all data to SanMgr (parent object)
    &SubIPC::write(-msg => "Controller Cluster Ready",
                   -obj_data => $cluster_data,
                   -type => "setup_cluster_complete",
                   -name => "cluster_setup");
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Set the appropriate status corresponding to cluster setup
    $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        &SubIPC::fail(-msg => "A subtest failure has been reported:\n\t\tFailed: " .
                              $st_status->{failed} . "\n\t\tPassed: " .
                              $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        &SubIPC::pass(-msg => "All subtests have passed:\n\t\tFailed: " .
                              $st_status->{failed} . "\n\t\tPassed: " .
                              $st_status->{passed});
        return $SanMgr::TRUE;
    }
}

###############################################################################
# Method:    config
# Objective: Launch subtest for each controller to perform config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg      => "run_id_prefix",
                                            -required => 0,
                                            -default  => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "TODO");
    return $SanMgr::TRUE;
}

###############################################################################
# Method:    cleanup
# Objective: Launch subtest for all clusters to cleanup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $clusters = &ParamsMgr::search(search_path => [],
                                      key => "CLUSTERS",
                                      paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # only perform cluster cleanup if cleanup is enabled on all clusters
    if ($clusters->{CLEANUP} == 1) {
        my $subtest = &SubIPC::create(-esc => $self->{parent}->{enable_subtest_count},
                                      -id  => $run_id_prefix . "Controller-cleanup",
                                      -run => sub { $self->_cleanup_all_clusters(-subtested => $SanMgr::TRUE); },
                                      -ipc => 1,
                                      -bg  => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            return $SanMgr::FALSE;
        }

        &SubIPC::wait_finish(-subtest => [$subtest]);
        if (&SubIPC::error()) {
            $self->_set_err("Unable to wait for Controller Cleanup subtest properly\n\t" . &SubIPC::get_error());
            return $SanMgr::FALSE;
        }

        my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
        if ($st_status->{failed} != 0) {
            $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
            return $SanMgr::FALSE;
        } else {
            $self->sanlog(-msg => "Controller Cleanup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        }

        &SubIPC::remove_subtest();
    } else {
        $self->sanlog(-msg => "Skipping Controller Cleanup");
    }

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _cleanup_all_clusters
# Objective: Launch subtest for each controller to perform cleanup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cleanup_all_clusters {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($cleanup_stack) = undef;

    # Process arguments.
    &ParamsMgr::process_args(%args);

    # Get the list of things that have to be cleaned up
    $cleanup_stack = $self->{parent}->{_objects}->{cleanup_stack};

    if ( !(defined($cleanup_stack) )) {

        # Get list of controllers to run cleanup against
        my $clusters = &ParamsMgr::search(-search_path => [],
                                          -key => "CLUSTERS",
                                          -search_type => "",
                                          -key_filter => qr/^CLUSTER\d*$/,
                                          paramset => $self->EnGenX::Util::get_paramset());

        if (&ParamsMgr::error()) {
            $self->_set_err(&ParamsMgr::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # Go through the list and identify unique clusters
        foreach my $cluster (keys(%{$clusters})) {

            # Check to see if resource is to be skipped
            if ($clusters->{$cluster}->{CLEANUP} == 0) {
                $self->sanlog(-msg => "Skipping cluster " .
                                  $clusters->{$cluster}->{NAME} .
                                  " because cleanup tag set to \"0\"",
                          -msg_level  => 1);
                next;
            }
            $self->sanlog(-msg => "Creating cluster cleanup objs" .
                               $clusters->{$cluster}->{NAME},
                          -msg_level  => 1);
            my $cluster_data =
                $self->create_cleanup_obj_controller_cluster(-resource => $clusters->{$cluster});
            if ($self->error()) {
                $self->_set_err("Failed to create cleanup objects on cluster " . $clusters->{$cluster}->{NAME});
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
            my $sorted_cluster_data = $self->sort_data_by_type(data => $cluster_data);
            if ($self->error()) {
                $self->_set_err("Failed to sort data by type on cluster " . $clusters->{$cluster}->{NAME});
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            # if there is no sorted cluster data, then we found nothing to cleanup
            if (not defined($sorted_cluster_data)) {
                &SubIPC::pass(-msg => "Did not find anything on the cluster to cleanup based on the XML input");
                return $SanMgr::TRUE;
            }

            unless ($self->{parent}->_store_controller_cluster_objs(%args, data => $sorted_cluster_data)) {
                $self->_set_err("Failed to store controller objects in SanMgr");
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            $cleanup_stack = $sorted_cluster_data;
        }
    }

    # Log the cleanup stack.  Changing the Dumper depth temporarily to see the full
    # cleanup stack
    my $dumper_depth = $Data::Dumper::Maxdepth;
    $Data::Dumper::Maxdepth = 3;
    my $dumpout = Dumper($cleanup_stack);
    $Data::Dumper::Maxdepth = $dumper_depth;
    $self->sanlog(-msg => "Dumper of all Controller cleanup data:\n"
                        . $dumpout,
                  -msg_level  => 1);

    my $subtest =
        &SubIPC::create(-id     => "cleanup_sanmgr",
                        -run    => sub { $self->cleanup_controller_cluster(-data => $cleanup_stack); },
                        -parent => $self,
                        -ipc    => 1,
                        -bg     => 1);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => [$subtest]);

    # Set the appropriate status corresponding to cluster cleanup
    my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        &SubIPC::fail(-msg => "A subtest failure has been reported:\n\t\tFailed: " .
                              $st_status->{failed} . "\n\t\tPassed: " .
                              $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        &SubIPC::pass(-msg => "All subtests have passed:\n\t\tFailed: " .
                              $st_status->{failed} . "\n\t\tPassed: " .
                              $st_status->{passed});
        return $SanMgr::TRUE;
    }
}

###############################################################################
# Method:    validate
# Objective: Runs set of validation methods to verify that controllers are
#            in expected state.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self) = @_;

    $self->clear_error();

    # Get the set of validation parameters from ParamsMgr
    my $validate = &ParamsMgr::search(search_path => [],
                                      key => "VALIDATE_POLICY",
                                      paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Get the set of environment parameters from ParamsMgr
    my $environment = &ParamsMgr::search(search_path => [],
                                         key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Get list of controller clusters to run validation against
    my $controllers = &ParamsMgr::search(search_path => [],
                                         key => "",
                                         search_type => "CLUSTER",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # TODO: Add controller cluster validation
    &SubIPC::pass(-msg => "TODO");
    return $SanMgr::TRUE;
}

##############################################################################
# Method:    load_resource
# Objective: Gather list of clusters and create objects for each one
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_resource {
    my ($self) = @_;

    $self->clear_error();

    # Get list of clusters
    my $clusters = &ParamsMgr::search(-search_path => [], -key => "CLUSTERS",
                                      -search_type => "",
                                      -key_filter  => qr/^CLUSTER\d*$/,
                                      paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $environment = &ParamsMgr::search(search_path => [], key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }
    my $timezone = $environment->{NTP_TIME_ZONE};

    # obtain the clusters element (so we can parse attributes off of it)
    my $clusters_element = &ParamsMgr::search(-search_path => [], -key => "CLUSTERS",
                                              paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # set clusters cmd interface
    my $cmd_interface = $clusters_element->{CMD_INTERFACE};

    my @nacl_nodes;
    if (exists $self->{parent}->{nodes}) {
        @nacl_nodes = @{delete $self->{parent}->{nodes}};
    } else {
        try {
            # Get all the Nodes that are defined.
            @nacl_nodes = NACL::C::Node->find();
        } otherwise {
            my $exception = shift;
            $self->_set_err("Failed to find nodes:" . $exception->text());
        };
        if ($self->error()) {
            return $SanMgr::FALSE;
        }
    }
    my @nacl_cluster_objs_array;
    if (exists $self->{parent}->{clusters}) {
        @nacl_cluster_objs_array = @{delete $self->{parent}->{clusters}};
    } else {
        try {
            # Get all the unique cluster objects from the Nodes that are defined.
            @nacl_cluster_objs_array =
                NACL::STask::Cluster->find_unique_clusters(nodes => \@nacl_nodes);
        } otherwise {
            my $exception = shift;
            $self->_set_err("Failed to find_unique_clusters:" . $exception->text());
        };
        if ($self->error()) {
            return $SanMgr::FALSE;
        }
    }
    my %nacl_cluster_objs_hash = ();
    # Sort the cluster object because we only want to call name() once for
    # each cluster object and we need to use this info in a loop.
    foreach my $nacl_cluster_obj (@nacl_cluster_objs_array) {
        $nacl_cluster_objs_hash{$nacl_cluster_obj->name()} = $nacl_cluster_obj;
    }
    foreach my $cluster (keys(%{$clusters})) {
        # Check to see if we don't already have a cluster object
        if (defined($self->{parent}->{_objects}->{cluster}->{$clusters->{$cluster}->{NAME}})) {
            next;
        }

        # override cmd_interface param if defined
        if (defined($clusters->{$cluster}->{CMD_INTERFACE})) {
            $cmd_interface = $clusters->{$cluster}->{CMD_INTERFACE};
        }

        # get one of the node members from the datastructure
        my $clus_name = $clusters->{$cluster}->{NAME};
        my $node_name;
        my @node_members = split(/,/, $clusters->{$cluster}->{NODES});
        if (scalar(@node_members)) {
            # parse out the clus name from one of the nodes.
            $node_name = $node_members[0];
        } else {
            # otherwise, assume that the cluster name is one of the nodes
            $node_name = $clus_name;
        }

        # Create cluster object
        my $cluster_obj = San::ControllerCluster->new(-filer         => $node_name,
                                                      -cmd_interface => $cmd_interface,
                                                      -debug_level   => $self->get_debug_level(),
                                                      -new_cmds_obj  => $SanMgr::TRUE
                                                      );
        if ($cluster_obj->error()) {
            $self->_set_err($cluster_obj->errmsg());
            return $SanMgr::FALSE;
        }

        my @node_objs = $cluster_obj->get_node_objs();
        if ($cluster_obj->error()) {
            $self->_set_err($cluster_obj->errmsg());
            return $SanMgr::FALSE;
        }

        foreach my $node (@node_objs) {
            my $name = $node->get_name();
            $self->{parent}->{_objects}->{node}->{$name} = $node;
        }

        # extract the cserver object and store it on the sanmgr object
        my $cserver_obj = $cluster_obj->get_cserver_obj();
        if ($cluster_obj->error()) {
            $self->_set_err("Failed to obtain the cserver object from cluster '" . $clus_name . "'",
                            $cluster_obj->errmsg());
            return $SanMgr::FALSE;
        }
        $self->{parent}->{_objects}->{cserver}->{$cserver_obj->get_name()} = $cserver_obj;
        $self->{parent}->{_objects}->{nacl_cluster}->{$cserver_obj->get_name()}
                            = $nacl_cluster_objs_hash{$cserver_obj->get_name()};

        # set the timezone if one was specified
        if ($timezone) {
            # If you have multiple processes running and you modify the timezone multiple
            # times too quickly, you will get failures.  If we get a failure here, sleep and
            # retry a few times.
            my $timezone_sleep = 5;
            my $timezone_count = 1;
            while ($timezone_count <= 5) {
                if ($timezone_count > 1) {
                    $self->sanlog(-msg => "Retrying timezone ($timezone_count of 5 retries)");
                }
                try {
                    # for good measure, set the timezone for the cluster object
                    NACL::C::SystemNodeDate->modify(command_interface => $nacl_cluster_objs_hash{$cserver_obj->get_name()},
                                                    timezone => $timezone,
                                                    node => "*",
                                                    apiset_must => {interface => "CLI"});
                    $self->clear_error();
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Failed to set timezone on cluster "
                                   . $cserver_obj->get_name() . " to $timezone", $e->text());
                };

                # If the command passed, self->error() will be cleared
                # and we are done with retries
                last if (!($self->error()));

                # Sleep and increment the sleep time by 5 seconds
                &SanUtil::nate_sleep($timezone_sleep);
                $timezone_sleep = $timezone_sleep + 5;
                $timezone_count++;
            } # end while

            if ($self->error()) {
                return $SanMgr::FALSE;
            }
        }

        # Need to store the cluster objects in nacl_node because QOS support
        # needs to not worry about whether the cserver passed into vserver setup
        # is a cserver or a node for load balancing.
        $self->{parent}->{_objects}->{nacl_node}->{$cserver_obj->get_name()}
                            = $nacl_cluster_objs_hash{$cserver_obj->get_name()};

        # Save the object for future use
        $self->{parent}->{_objects}->{cluster}->{$clus_name} = $cluster_obj;
    }

    # Now store all the nodes that we found earlier in nacl_node. These node
    # object should already be version by this point.
    foreach my $nacl_node_obj (@nacl_nodes) {
        $self->{parent}->{_objects}->{nacl_node}->{$nacl_node_obj->name()}
                            = $nacl_node_obj;
    }

    return $SanMgr::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

1;
