#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of helper methods to work on controllers
## @author jose.porrua@netapp.com
## @status private
## @pod here

package San::SanMgr::Controller::ControllerHelper;

use Clone qw(clone);
use POSIX;
use Data::Dumper;
use Tie::IxHash;

# compiler directives
use strict;

###############################################################################
# Method:    sort_data_by_type
# Objective: Organizes controller data based on the object's type.
###############################################################################
sub sort_data_by_type {
    my ($self, %args) = @_;

    $self->clear_error();

    # Locals
    my (@vservers, @aggregates, @volumes, @qtrees, @luns, @files) = ();
    my (@lifs, @igroups, @portsets, @tpgroups, @qos_policy_groups) = ();
    my (@protocols, @iscsi_security_settings, @isns_services, @lun_binds) = ();
    my (@ifgrps, @vlans) = ();
    my ($data, $sorted_data) = undef;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing data created by SanMgr's setup
    $data = &ParamsMgr::get_arg(-arg => "data",
                                -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # Extract data and put it on different "buckets"
    foreach my $obj (@{$data}) {
        while (my ($key, $value) = each (%{$obj})) {
            if ($key =~ /type/i and $value =~ /^vserver$/) {
                push(@vservers, $obj);
            } elsif ($key =~ /type/i and $value =~ /^volume$/) {
                push(@volumes, $obj);
            } elsif ($key =~ /type/i and $value =~ /^qtree$/) {
                push(@qtrees, $obj);
            } elsif ($key =~ /type/i and $value =~ /^lun$/) {
                push(@luns, $obj);
            }  elsif ($key =~ /type/i and $value =~ /^file$/) {
                push(@files, $obj);
            } elsif ($key =~ /type/i and $value =~ /^lun_bind$/) {
                push(@lun_binds, $obj);
            } elsif ($key =~ /type/i and $value =~ /^igroup$/) {
                push(@igroups, $obj);
            } elsif ($key =~ /type/i and $value =~ /^lif$/) {
                push(@lifs, $obj);
            } elsif ($key =~ /type/i and $value =~ /^tpgroup$/) {
                push(@tpgroups, $obj);
            } elsif ($key =~ /type/i and $value =~ /^portset$/) {
                push(@portsets, $obj);
            } elsif ($key =~ /type/i and $value =~ /^aggregate$/) {
                push(@aggregates, $obj);
            }  elsif ($key =~ /type/i and $value =~ /^(fcp|iscsi)_service$/) {
                push(@protocols, $obj);
            } elsif ($key =~ /type/i and $value =~ /^iscsi_security_setting$/) {
                push(@iscsi_security_settings, $obj);
            } elsif ($key =~ /type/i and $value =~ /^isns_service$/) {
                push(@isns_services, $obj);
            } elsif ($key =~ /type/i and $value =~ /^qos_policy_group$/) {
                push(@qos_policy_groups, $obj);
            } elsif ($key =~ /type/i and $value =~ /^ifgrp$/) {
                push(@ifgrps, $obj);
            } elsif ($key =~ /type/i and $value =~ /^vlan$/) {
                push(@vlans, $obj);
            }
        }
    }

    # Build a hash based on each bucket
    $sorted_data->{vserver}                = \@vservers                 if (scalar(@vservers));
    $sorted_data->{volume}                 = \@volumes                  if (scalar(@volumes));
    $sorted_data->{qtree}                  = \@qtrees                   if (scalar(@qtrees));
    $sorted_data->{lun}                    = \@luns                     if (scalar(@luns));
    $sorted_data->{file}                   = \@files                    if (scalar(@files));
    $sorted_data->{igroup}                 = \@igroups                  if (scalar(@igroups));
    $sorted_data->{lif}                    = \@lifs                     if (scalar(@lifs));
    $sorted_data->{tpgroup}                = \@tpgroups                 if (scalar(@tpgroups));
    $sorted_data->{portset}                = \@portsets                 if (scalar(@portsets));
    $sorted_data->{aggregate}              = \@aggregates               if (scalar(@aggregates));
    $sorted_data->{protocol_service}       = \@protocols                if (scalar(@protocols));
    $sorted_data->{iscsi_security_setting} = \@iscsi_security_settings  if (scalar(@iscsi_security_settings));
    $sorted_data->{isns_service}           = \@isns_services            if (scalar(@isns_services));
    $sorted_data->{lun_bind}               = \@lun_binds                if (scalar(@lun_binds));
    $sorted_data->{qos_policy_group}       = \@qos_policy_groups        if (scalar(@qos_policy_groups));
    $sorted_data->{ifgrp}                  = \@ifgrps                   if (scalar(@ifgrps));
    $sorted_data->{vlan}                   = \@vlans                    if (scalar(@vlans));

    return $sorted_data;
} # End of sort_data_by_type

###############################################################################
# Method:    _wait_setup
# Objective: A generic "wait" method. Just a wrapper around wait_finish
#            to avoid repeating code and simplifying methods.
# Inputs:    A hash:
#            %args (subtest => an array ref with SubIPC subtests)
# Returns:   TRUE if nothing failed FALSE otherwise. This method calls
#            &SubIPC::pass or &SubIPC::fail accordingly.
###############################################################################
sub _wait_setup {
    my ($self, %args) = @_;

    $self->clear_error();

    # Locals
    my ($st_status, $subtest) = undef;

    # Parse args
    &ParamsMgr::process_args(%args);

    $subtest = &ParamsMgr::get_arg(-arg => "subtest",
                                   -required => $SanMgr::FALSE,
                                   -default => &SubIPC::get_subtests());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => $subtest);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my @messages = &SubIPC::get_messages(-type => "setup_object_complete");
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # Now bubble up all data to SanMgr (parent object)
    &SubIPC::write(-msg => "Controller Cluster Ready", -obj_data => \@messages,
                   -type => "setup_object_complete", -name => "object_setup");
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # Get the status of a subtest and report it
    $st_status = &SubIPC::get_subtest_status(-subtest => $subtest);
    if ($st_status->{failed} != $SanMgr::FALSE) {
        $self->sanlog(-msg => "A subtest failure has been reported:\n\t\tFailed: " .
                              $st_status->{failed} . "\n\t\tPassed: " .
                              $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "All subtests have passed:\n\t\tFailed: " .
                              $st_status->{failed} . "\n\t\tPassed: " .
                              $st_status->{passed});
        return $SanMgr::TRUE;
    }
} # End _wait_setup

###############################################################################
# Method:     _format_log
# Objective:  Adds a simple box around a message for logging purposes.
# Parameters: None.
# Returns:   The message
###############################################################################
sub _format_log {
    my ($self, $msg) = @_;

    my $caller_prefix = $self->_caller_prefix(); # inherited from SanLog

    my $header  = "\n" x 2 . "#-" x 39 . "#\n";
    my $body    = "#  " . "SAN MGR: " . $caller_prefix . " " . $msg . "\n";
    my $footer  = "\#-" x 39 . "#\n";

    # Prints something like (easier to identify SanMgr entries on the logs):
    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
    #  SAN MGR: setup::method(): Some message
    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
    return ($header . $body . $footer);
} # End _format_log

###############################################################################
# Method:     _return_if_subtest
# Objective:  A helper method that's responsible for either returning a value
#             or send a SubIPC message to a parent. Pretty handy for methods
#             that can be both subtested or not.
# Parameters: A hash:
#             %args(subtested => 1|0 indicating if this is a subtest or not,
#                   status => pass|fail
#                   msg => an string to be appended to an IPC msg,
#                   write_parent => tells this method to write data to parent
#                   write_msg_type => ID of the message
#                   ret_value => the return value for this method.
# Returns:   A value or a subIPC message.
###############################################################################
sub _return_if_subtest {
    my ($self, %args) = @_;

    # Locals
    my ($ret_value) = undef;
    my ($subtested, $status, $msg, $write_msg_type) = "";
    my ($write_parent) = $SanMgr::FALSE;

    # Parse args
    &ParamsMgr::process_args(%args);

    # Determines wether this is a subtest and a message has to be sent
    $subtested = &ParamsMgr::get_arg(-arg => "subtested",
                                     -required => $SanMgr::TRUE);

    # Did it pass or fail?
    $status = &ParamsMgr::get_arg(-arg => "status",
                                  -required => $SanMgr::TRUE);

    # Message to return
    $msg = &ParamsMgr::get_arg(-arg => "msg",
                               -required => $SanMgr::TRUE);

    # Do we want to send data to the parent?
    $write_parent = &ParamsMgr::get_arg(-arg => "write_parent",
                                        -required => $SanMgr::FALSE,
                                        -default => $SanMgr::FALSE);

    # ID corresponding to the message we'll send to the parent
    $write_msg_type = &ParamsMgr::get_arg(-arg => "write_msg_type",
                                          -required => $SanMgr::FALSE,
                                          -default => $SanMgr::FALSE);

    # Data to return
    $ret_value = &ParamsMgr::get_arg(-arg => "ret_value",
                                     -required => $SanMgr::FALSE,
                                     -default => $SanMgr::FALSE);

    # By default, if it's a subtest and something fails, return a fail status.
    # If users specify a return value, return it, otherwise return false.

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument Parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self) if ($subtested);
        if ($ret_value != $SanMgr::FALSE) {
            return $ret_value;
        } else {
            return $SanMgr::FALSE;
        }
    }

    # If this is marked as pass we'll have two options: return a success value
    # or write a message to the parent returning a given value.

    if ($status =~ /^pass$/i) {
        if ($subtested) {
            if ($write_parent) {
                # Write data to the parent of this caller
                &SubIPC::write(-msg => "Data from child process",
                               -obj_data => $ret_value,
                               -type => $write_msg_type,
                               -name => "sanmgr_return_if_subtest");
                if (&SubIPC::error()) {
                    $self->_set_err("Unable to write to parent", &SubIPC::get_error());
                    &SubIPC::fail(-object => $self);
                }
            }
            &SubIPC::pass(-msg => $msg);
        } else {
            if ($ret_value != $SanMgr::FALSE) {
                return $ret_value;
            } else {
                return $SanMgr::TRUE;
            }
        }
    } elsif ($status =~ /^fail$/i) {
        if ($subtested) {
            if ($write_parent) {
                # Something failed. No need to write this data since
                # this data contains info about things that were built
                # successfully.
            }
            &SubIPC::fail(-object => $self);
        } else {
            if ($ret_value != $SanMgr::FALSE) {
                return $ret_value;
            } else {
                return $SanMgr::FALSE;
            }
        }
    }

} # End _return_if_subtest

###############################################################################
# Method:    _generate_copies
# Objective: Generates "n" copies of an object defined in the params. It's
#            responsible for renaming and appending indexes to the names
#            if needed. It also assigns IPs to copies of iSCSI lifs
# Input:     A reference to a hash containing object attributes
# Returns:   An array or a reference to an array containing all hashes
#            with all the attributes corresponding to a certain object.
###############################################################################
sub _generate_copies {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local variables
    my ($prefix, $name, $data, $element, $parent_name) = undef;  # Refs
    my (@objs, %ips) = ();                         # Lists
    my ($type) = "";                               # Scalars
    my ($num_copies, $range) = 1;                  # By default create 1 object.
    my $offset = 0;                                # start counting at 0

    &ParamsMgr::process_args(%args);

    $element = &ParamsMgr::get_arg(-arg => "element",
                                   -required => $SanMgr::TRUE);

    $parent_name = &ParamsMgr::get_arg(-arg => "parent_name",
                                       -required => $SanMgr::FALSE,
                                       -default => "");

    $type = &ParamsMgr::get_arg(-arg => "type",
                                -required => $SanMgr::FALSE,
                                -default => "none");

    # Make sure we don't mess up with the original data structure.
    $data = Clone::clone($element);

    # Check for count or loop (Compatibility with FitMgr's syntax)
    if (defined($data->{COUNT})) {
        $num_copies = delete $data->{COUNT};
    } elsif (defined($data->{LOOP})) {
        $num_copies = delete $data->{LOOP};
    }

    # Check for an index or offset. i.e. an offset of 50 results in ob_50, ob_51
    # instead of ob_0, ob_1, etc.

    if (defined($data->{INDEX})) {
        $offset = delete $data->{INDEX};
    } elsif (defined($data->{OFFSET})) {
        $offset = delete $data->{OFFSET};
    }

    if ($type =~ /^VSERVER/i) {
        # Collect all IP addresses and lifs
        foreach my $key (keys %{$data}) {
            if ($key =~ /^LIF/i) {
                if ( List::MoreUtils::any { /^ADDRESS_RANGE/i } keys(%{$data->{$key}}) ) {
                    # Store the complete range of IPs for each lif in the vserver
                    # object. They will be used when setup creates iscsi lifs
                    my $lif_name = $data->{$key}->{NAME};
                    my @ips_range = ();
                    # There could be more than one tag of ADDRESS_RANGEs so we must verify
                    # that the tag name is either plain ADDRESS_RANGE or ADDRESS_RANGE with
                    # a number that is actually the index of the tag. The regex only
                    # supports 9 address range tags
                    foreach my $lif_key (sort keys %{$data->{$key}} ){
                        if ($lif_key =~ /^ADDRESS_RANGE($|[1-9]$)/i) {
                            my @ips = $self->_get_ip_addresses( $data->{$key}->{$lif_key} );
                            push (@ips_range, @ips);
                        }
                    }
                    # Randomize the list of ip addresses
                    $ips{$lif_name}->{ips} = _randomize_ip_addresses(@ips_range);
                    $ips{$lif_name}->{range} = floor( scalar(@ips_range) / $num_copies );
                }
            }
        }
    }

    # If there's only one copy, leave the name of the resource alone
    if ($num_copies <= 1) {
        if ($type =~ /^VSERVER/i) {
            # Verify there is a range of ips available
            if ( keys(%ips) ) {
                # Assign the ip range values to the node object, with the
                # LIF name as a key excluding the range value.
                foreach my $key (keys(%ips)) {
                    $self->{_objects}->{cluster}->{$data->{NAME}}->{$key}->{ips} =
                        $ips{$key}->{ips};
                }
            }
        }

        # Log that we are making copies and what they are
        my $dumpout = Dumper($data);
        $self->sanlog(-msg => "Making copies of type $type. "
                            . "Dumper of data to return (single "
                            . "copy case):\n$dumpout",
                            -msg_level  => 7);

        push(@objs, $data);
    } else {
        # Go through each element, tweak the name
        for (my $i = 0; $i < $num_copies; $i++) {
            my %tmp_copy;

            # keep the insertion order of the hash so that
            # everything is kept in the proper order
            tie(%tmp_copy, "Tie::IxHash");

            %tmp_copy = %{$data};
            # Checking for volumes names of the form <volume_name>_### and padding with a
            # 0 so that the name is <volume_name>_0###  because the former format is
            # reserved for striped volumes and will cause volume creation to fail.
            if ($type =~ /^VOL/i and (($i + $offset) >= 100) and (($i + $offset) < 1000)) {
                $tmp_copy{NAME} = $data->{NAME} . "_0" . ($i + $offset);
            } else {
                $tmp_copy{NAME} = $data->{NAME} . "_" . ($i + $offset);
            }

            # Store a range of IPs for each vserver in SanMgr's object. This
            # will be used when setup creates iscsi lifs
            if ($type =~ /^VSERVER/i and keys(%ips)) {
                foreach my $key (keys(%ips)){
                    my @ips = @{$ips{$key}->{ips}};
                    my @assigned_ips = splice( @ips, ($ips{$key}->{range})*$i, $ips{$key}->{range} );
                    $self->{_objects}->{cluster}->{$tmp_copy{NAME}}->{$key}->{ips} = \@assigned_ips;
                    $self->{_objects}->{cluster}->{$tmp_copy{NAME}}->{$key}->{unassigned_ips} = \@ips;
                }
                $tmp_copy{PROTOCOL_COUNT} = $i;
            } elsif ($type =~ /^LIF/i and $data->{'DATA-PROTOCOL'} =~ /^iscsi$/i) {
                unless (defined $parent_name and length $parent_name) {
                    $self->_set_err("Required parent_name");
                    return (wantarray()) ? () : [];
                }

                if (defined($self->{_objects}->{cluster}->{$parent_name}->{$data->{NAME}}->{ips})) {
                    $self->{_objects}->{cluster}->{$parent_name}->{$tmp_copy{NAME}} =
                       $self->{_objects}->{cluster}->{$parent_name}->{$data->{NAME}};
                } else {
                    # set error because xml was not sepcify correctly
                    $self->_set_err("Invalid XML for LIF tag");
                    return (wantarray()) ? () : [];
                }
            }

            # Log that we are making copies and what they are
            my $dumpout = Dumper(%tmp_copy);
            $self->sanlog(-msg => "Making copies of type $type. "
                                . "Dumper of data to return (multiple "
                                . "copy case):\n$dumpout",
                                -msg_level  => 7);

            push(@objs, \%tmp_copy);
        }
    }

    return (wantarray()) ? @objs : \@objs;
} # End _generate_copies

###############################################################################
# Method:    _process_common_args
# Objective: Get policies and normalize attributes
# Details:   Won't see POD documentation at the end of this file.
###############################################################################
sub _process_common_args {
    my ($self, %args) = @_;

    $self->clear_error();

    # Locals
    my ($reuse_policy, $cleanup_policy) = "";
    my (%params) = ();

    # Check if there's a local "reuse policy" for this object.
    if (defined($args{REUSE_POLICY})) {
        $reuse_policy = delete $args{REUSE_POLICY};
    } else {
        # There isn't one so get the global reuse policy for setup.
        $reuse_policy = &ParamsMgr::get("REUSE_POLICY.SETUP", paramset => $self->EnGenX::Util::get_paramset());
        # Should it fail if the policy is not there? NO! Defaulting to "reuse"
        if (&ParamsMgr::error()) {
            $reuse_policy = "reuse";
        }
    }

    # Check the cleanup policy for this object.
    if (defined($args{CLEANUP_POLICY})) {
        $cleanup_policy = delete $args{CLEANUP_POLICY};
    } else {
        # There isn't one so get the global reuse policy for cleanup.
        $cleanup_policy = &ParamsMgr::get("REUSE_POLICY.CLEANUP", paramset => $self->EnGenX::Util::get_paramset());
        # Use "new" as the default (only cleanup new objects)
        if (&ParamsMgr::error()) {
            $cleanup_policy = "new";
        }
    }

    # XXX: There's a weird ParamsMgr behavior adding _attr to tags
    # remove them or else the lib will fail!
    foreach my $key (keys (%args)) {
        if ($key =~ /^_/) {
            delete $args{$key};
        }
    }

    # Convert all args to lower case. Our libraries do not like upper cases!
    %args = map { lc $_ => $args{$_} } keys %args;

    return (%args,
            reuse_policy => $reuse_policy,
            cleanup_policy => $cleanup_policy);
} # End _process_common_args

###############################################################################
# Method:    _get_ip_addresses
# Objective: Generate a list of IP addresses based on an address range.
# Inputs:    A reference to a hash:
#            ip_data => contains address range for iSCSI lifs
# Returns:   A list or reference to a list containing IP addresses.
###############################################################################
sub _get_ip_addresses {
    my ($self, $ip_data) = @_;

    # Locals
    my ($ip_range) = "";
    my (@ips) = ();

    # If there's an ip range get it, otherwise return an empty list
    if (defined($ip_data->{IP})) {
        $ip_range = $ip_data->{IP};
    } else {
        return (wantarray()) ? () : [];
    }

    # if the ip range contains a colon, treat it as an ipv6 range
    if ($ip_range =~ /:/) {
        my @chunks = split(/:/, $ip_range);
        my $last_chunk = $chunks[-1];
        my $hex_regex = "[A-Fa-f0-9]";
        if ($last_chunk =~ /\[($hex_regex+)-($hex_regex+)\]/) {
            my $range_base = hex($1);
            my $range_max  = hex($2);

            for (my $ip_segment = $range_base; $ip_segment <= $range_max; $ip_segment++){
                my $stringified_chunk = sprintf("%x", $ip_segment);
                $chunks[-1] = $stringified_chunk;
                my $addr = join(":" , @chunks);
                push(@ips, $addr);
            }
        }
    }else{
        # 10.10.10.[start-end]
        # 10.10.[start-end].[start-end]
        # 10.10.[start-end].10
        # split ip_range into four bytes
        my ($first, $second, $third, $fourth) = split(/\./, $ip_range);

        # find the range in the third and fourth byte
        if ($third =~ /\[(\d+)-(\d+)\]/) {
            for (my $i = $1; $i <= $2; $i++) {
                # do we have the range in the forth byte as well?
                if ($fourth =~ /\[(\d+)-(\d+)\]/) {
                    for (my $j = $1; $j <= $2; $j++) {
                       push(@ips, "$first.$second.$i.$j");
                    }
                } else {
                    # range in third byte only
                    push(@ips, "$first.$second.$i.$fourth");
                }
            }
        } elsif ($fourth =~ /\[(\d+)-(\d+)\]/) {
            # range in fourth byte only
            for (my $i = $1; $i <= $2; $i++) {
                push(@ips, "$first.$second.$third.$i");
            }
        }
    }

    return (wantarray()) ? @ips : \@ips;
} # End _get_ip_addresses

###############################################################################
# Method:    _randomize_ip_addresses
# Objective: Randomize the order of the element of the provided list of IP
#            addresses to reduce the chances of fetching a duplicate IP.
# Inputs:    A list containing IP addresses.
# Returns:   A list or reference to a list containing IP addresses.
###############################################################################
sub _randomize_ip_addresses {
    my (@ips) = @_;

    # Unless the user passed in more than 1 ip then we should not randomize.
    unless (scalar(@ips > 1)) {
        return (wantarray()) ? @ips : \@ips;
    }

    #  Swapping each element in the array with another randomly selected element.
    #  Fisher-Yates algorithm
    for (my $i = scalar(@ips); --$i; ) {
        my $j = int rand ($i + 1);
        next if $i == $j;
        @ips[$i,$j] = @ips[$j,$i];
    }

    return (wantarray()) ? @ips : \@ips;
} # End _randomize_ip_addresses

###############################################################################
# Method:    _get_next_available_ip
# Objective: Generate a list of IP addresses based on an address range.
# Inputs:    A reference to a hash:
#            ip_data => contains address range for iSCSI lifs
# Returns:   A list or reference to a list containing IP addresses.
###############################################################################
sub _get_next_available_ip {
    my ($self, %args) = @_;

    # Locals
    my ($ips, $unassigned_ips) = [];
    my $lif_args = undef;

    # whoever is calling this method, must define ips and unassigned_ips.
    &ParamsMgr::process_args(%args);

    $ips = &ParamsMgr::get_arg(-arg => "ips",
                               -required => $SanMgr::TRUE);

    $unassigned_ips = &ParamsMgr::get_arg(-arg => "unassigned_ips",
                                          -required => $SanMgr::TRUE);

    # If there's an address available in the IP array, pop that address and return it.
    if (ref($ips) =~ /ARRAY/) {
        if( scalar(@{$ips}) ) {
            return pop(@{$ips});
        }
    } else {
        $self->_set_err("Could not retrieve a valid IP addrres, the provided " .
                        "address range (IPs array reference) is BLANK");
        return undef;
    }

    # If there's an address available in the UNASSIGNED IP array and IP array is empty,
    # pop that address and return it.
    if (ref($unassigned_ips) =~ /ARRAY/) {
        if( scalar(@{$unassigned_ips}) ) {
            return pop(@{$unassigned_ips});
        }
    } else {
        $self->_set_err("Could not retrieve a valid IP addrres, the provided " .
                        "address range (UNASSIGNED_IPs array reference) is BLANK");
        return undef;
    }

    # Both IPS and UNASSIGNED arrays are empty, throw an error.
    $self->_set_err("Could not retrieve a valid IP addrres, specify another " .
                    "address or another address range");
    return undef;

} # End _get_next_available_ip

###############################################################################
# Method:    _is_intercluster_requested
# Objective: Quickly parse through the cluster's keys to see if there is any
#            cluster LIF information.
# Inputs:    The reference to the cluster hash.
# Returns:   True if it found two or more cluster LIFs. False if otherwise.
###############################################################################
sub _is_intercluster_requested{
    my ($self, %args) = @_;
    # Locals
    my @lifs_found = ();

    # Need the cluster Object to verify the keys.
    &ParamsMgr::process_args(%args);
    my $cluster_ref = &ParamsMgr::get_arg(-arg => "cluster",
                                          -required => $SanMgr::TRUE);

    # The cluster should have at least two LIFs to use the intercluster
    # connections, so, let's verify how many LIFs where defined.
    foreach my $key (keys %{$cluster_ref}){
        if ($key =~ /^LIF($|[1-9]$)/i){
            push (@lifs_found, $cluster_ref->{$key}->{'NAME'});
        }
    }

    # If no lifs were defined, no cluster peer connection is needed
    if ( scalar(@lifs_found) < 1){
        return $SanMgr::FALSE;
    }
    # If only one lif was defined, throw an error, at least two lifs are needed to
    # create the intercluster
    elsif ( scalar(@lifs_found) == 1){
        $self->_set_err("For cluster ".$cluster_ref->{'NAME'}." only ONE lif '".$lifs_found[0].
                        "' was defined. Two or more lifs are required to create an intercluster connection");
        return $SanMgr::FALSE;
    }

    # ELSE, more than one LIF was defined, we can proceed.
    return $SanMgr::TRUE;

} # End _is_intercluster_requested

###############################################################################
# Method:    _create_intercluster_params
# Objective: Generate and porpulate the hash that Intercluster.pm is expecting
#           to be able to create the cluster peer conections. This is the
#           structure of the hash:
# 'CLUSTERS' =>{
#                'cluster_1'=>{
#                              'node_1'=>{'INTERCLUSTER_IP1'  =>  "9.9.9.90",
#                                         'INTERCLUSTER_PORT1'=>  "e0c",
#                                         'INTERCLUSTER_MTU1' =>  "1500",
#                                         'INTERCLUSTER_MASK1'=>  "255.254.0.0",
#                                         'INTERCLUSTER_GATEWAY1'=> "9.0.0.10"
#                                        }
#                              'node_n'=>{'INTERCLUSTER_IP1' => ...,
#                                         ...
#                                        }
#                             },
#                'cluster_2'=>{
#                              'node_1'=>{
#                                        'INTERCLUSTER_IP1' => ... ,
#                                        ...
#                                       }
#                              'node_n'=>{
#                                         'INTERCLUSTER_IP1'  =>  ...,
#                                         ...
#                                        }
#                             }
# Inputs:    The reference to cluster hash.
#            The reference to the intercluster_parameters hash
# Returns:   True if the reference of the intercluster_parameters hash was porperly
#            modifiec. False if not.
###############################################################################
sub _create_intercluster_params {
    my ($self, %args) = @_;
    # Locals
    my $cluster_ref = {};
    my $interclus_params_ref = {};
    my $cserver_obj;
    my @clusters=();
    my $cluster='';


    # Must have the source of the info (cluster) and the hash that will be
    # modified (intercluster_params).
    &ParamsMgr::process_args(%args);
    $cluster_ref = &ParamsMgr::get_arg(-arg => "cluster",
                                       -required => $SanMgr::TRUE);
    $interclus_params_ref = &ParamsMgr::get_arg(-arg => "interclus_params",
                                                -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Required argument missing", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # FIRST, Create a cserver object.
    $cserver_obj = $self->_create_cserver(resource => $cluster_ref);
    if ($self->error()) {
        $self->_set_err("Failed to create cserver object");
        return $SanMgr::FALSE;
    }

    # SECOND, Get the Cluster Name (identity)
    @clusters = $cserver_obj->execute(-cmd => "cluster identity show");
    $cluster = $clusters[0]->{'name'};
    if ($cluster eq '') {
        $self->_set_err("Could not get the Cluster Name on the filer '', Cluster must ".
                        "be set up before trying to connect with a peer");
        return $SanMgr::FALSE;
    }

    # THIRD, Use the cluster name as a key in the hash that will be formed using
    # the cluster information
    foreach my $key (keys %{$cluster_ref}){
        if ($key =~ /^LIF($|[1-9]$)/i){
            if ( ($cluster_ref->{$key}->{'ROLE'}) eq 'intercluster') {
                # TO DO - If mask and gateway are not provided, could go to the
                # tharnhost file and get them.
                my $node = $cluster_ref->{$key}->{'HOME-NODE'};
                my $mtu = exists ($cluster_ref->{$key}->{'MTU'})?($cluster_ref->{$key}->{'MTU'}):"1500";

                $interclus_params_ref->{'CLUSTERS'}->{$cluster}->{$node}->{'INTERCLUSTER_NAME1'} = $cluster_ref->{$key}->{'NAME'};
                $interclus_params_ref->{'CLUSTERS'}->{$cluster}->{$node}->{'INTERCLUSTER_IP1'} = $cluster_ref->{$key}->{'ADDRESS'};
                $interclus_params_ref->{'CLUSTERS'}->{$cluster}->{$node}->{'INTERCLUSTER_PORT1'} = $cluster_ref->{$key}->{'HOME-PORT'};
                $interclus_params_ref->{'CLUSTERS'}->{$cluster}->{$node}->{'INTERCLUSTER_MTU1'} = $mtu;
                $interclus_params_ref->{'CLUSTERS'}->{$cluster}->{$node}->{'INTERCLUSTER_MASK1'} = $cluster_ref->{$key}->{'NETMASK'};
                $interclus_params_ref->{'CLUSTERS'}->{$cluster}->{$node}->{'INTERCLUSTER_GATEWAY1'} = $cluster_ref->{$key}->{'GATEWAY'};
            } else {
                $self->_set_err("LIF role is '".$cluster_ref->{$key}->{'ROLE'}."' and should be 'intercluster' ".
                                "Verify that the ROLE and PLACEMENT of this LIF is correct. LIFs that are not ".
                                "'intercluster should be placed inside of a VSERVER Tag"
                                );
                return $SanMgr::FALSE;
            }
        }
    }

    return $SanMgr::TRUE;

} # End _create_intercluster_params

1;
