#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to cleanup controller clusters
## @author latif@netapp.com, porrua@netapp.com
## @status private
## @pod here

package San::SanMgr::Controller::cleanup;

# Compiler directives.
use strict;
use San::SanLog;
use ParamsMgr;
use San::SanUtil;
use San::SanMgr::Controller::ControllerHelper;

use NACL::C::LunBind;
use NACL::CS::LunBind;
use NACL::C::QosPolicyGroup;
use NACL::C::NetworkPortIfgrp;
use NACL::C::NetworkPortVlan;
use NACL::C::VolumeFile;
use NATE::BaseException qw (:try);

use Data::Dumper;

use vars qw(@ISA);
@ISA = qw(Error SanLog San::SanMgr::Controller::ControllerHelper);

###############################################################################
# Method:    cleanup_controller_cluster
# Objective: Setup run against a single controller cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup_controller_cluster {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($data, $cserver, $vservers, $cleanup_stack, $aggregates, $ifgrps, $vlans) = undef;
    my ($vserver_name) = "";
    my (@subtests) = ();

    $self->sanlog(-msg => $self->_format_log("Start cleaning up cluster"),
                  -add_caller => 1);

    # Process arguments. A reasource is nothing more than a controller cluster
    # element which may come from a config file (i.e. XML)

    &ParamsMgr::process_args(%args);

    # The main cleanup stack
    $data = &ParamsMgr::get_arg(-arg => "data", -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to parse args", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # first reconnect to all controllers since we don't know what the state
    # of our existing connections may be in
    $self->reconnect_all_controllers();
    if ($self->error()) {
        $self->_set_err("Failed to reconnect to all nodes/cservers");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # run wafliron on aggrs that have it enabled
    $self->_run_wafliron_on_aggregates(-data => $data);
    if ($self->error()) {
        $self->_set_err("Failed to run wafliron on aggregates");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # The cleanup stack is basically a hash whose keys correspond to
    # each object type. Get the aggregates, vservers and group objects
    # based on vservers. This will allow us to delete stuff in parallel
    $aggregates = delete $data->{aggregate} if (defined($data->{aggregate}));
    $vservers   = delete $data->{vserver}   if (defined($data->{vserver}));


    foreach my $vserver (@{$vservers}) {
        $vserver_name = $vserver->{name};
        my $do_cleanup  = $vserver->{cleanup};
        if(not $do_cleanup){
            next;
        }
        my $cserver_name = $vserver->{cserver};
        my @vserver_data = ();
        foreach my $key (keys(%{$data})) {

            my @entries = @{$data->{$key}};
            foreach my $entry (@entries) {
                if (defined($entry->{vserver})) {
                    if ($entry->{vserver} eq "$vserver_name") {
                        push(@vserver_data, $entry);
                    }
                }
            }
        }

        # if the current vserver doesnt have parallel enabled, then reap the current
        # set of subtests.
        my $parallel = $vserver->{parallel};
        if ($parallel eq 0 and scalar(@subtests)) {
            $self->_reap_vserver_cleanup_subtests(-subtests => \@subtests);
            if ($self->error()) {
                $self->_set_err("A vserver cleanup subtest failed");
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
            @subtests = ();
        }

        $self->sanlog(-msg => $self->_format_log("Cleaning up vserver "
                                                     . $vserver_name),
                      -add_caller => 1);
        my $dumpout = Dumper(@vserver_data);
        $self->sanlog(-msg => "Dumper of cleanup stack for vserver "
                          . "$vserver_name\n"
                              . $dumpout,
                      -msg_level  => 3,
                      -add_caller => 1);

        # Destroying vservers in their own subtests
        my $subtest = &SubIPC::create(
            -id => "cleanup_vserver_" . $vserver->{name},
            -run => sub {
                $self->_cleanup_vserver(vserver => $vserver_name,
                                        cserver => $cserver_name,
                                        cleanup => $vserver->{cleanup},
                                        reused => $vserver->{reused},
                                        data => \@vserver_data);
                if ($self->error()) {
                    $self->_set_err("Failed to cleanup vserver $vserver_name");
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                &SubIPC::pass(-msg => "Vserver $vserver_name has been cleaned up");
                return $SanMgr::FALSE;
            },
            -ipc => $SanMgr::TRUE,
            -bg => $SanMgr::TRUE
        );
        if (&SubIPC::error()) {
            $self->_set_err("Failed to spawn cleanup subtest for $vserver_name",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        push(@subtests, $subtest);

        # reap the subtest list
        # if the user doesnt want max prallelism and
        #### the user doesnt want any parallelism on this vserver
        #### or
        #### we have hit a max vserver parallel setup count
        if ($parallel !~ /^max$/i and ($parallel eq 0 or scalar(@subtests) == $parallel)) {
            $self->_reap_vserver_cleanup_subtests(-subtests => \@subtests);
            if ($self->error()) {
                $self->_set_err("A vserver cleanup subtest failed");
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE
            }
            @subtests = ();
        }
    } # end foreach vserver

    # if there are still subtests that haven't been reaped, reap them all here:
    if (scalar(@subtests)) {
        $self->_reap_vserver_cleanup_subtests(-subtests => \@subtests);
        if ($self->error()) {
            $self->_set_err("A vserver cleanup subtest failed");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    $ifgrps = $data->{ifgrp};
    $vlans = $data->{vlan};

    # Clean up vlans
    if (defined $vlans) {
        $self->_cleanup_obj(data => $vlans,
                            type => "vlan");
        if ($self->error()) {
            $self->_set_err("Failed to cleanup vlans");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # Clean up ifgrps
    if (defined $ifgrps) {
        $self->_cleanup_obj(data => $ifgrps,
                            type => "ifgrp");

        if ($self->error()) {
            $self->_set_err("Failed to cleanup ifgrps");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # Clean up aggregates
    if (defined $aggregates) {
        $self->_cleanup_obj(data => $aggregates,
                            type => "aggregate");
        if ($self->error()) {
            $self->_set_err("Failed to cleanup aggregates");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    &SubIPC::pass(-msg => "Finished cleaning up a cluster.");

    return $SanMgr::TRUE;
} # End of cleanup_controller_cluster

###############################################################################
# Method:    _reap_vserver_cleanup_subtests
# Objective: Method responsible for Reaping vserver cleanup subtests
# Inputs:    A hash containing:
#            %args = (subtest => an array of SubIPC subtests)
# Returns:   True or False
###############################################################################
sub _reap_vserver_cleanup_subtests {
    my ($self, %args) = @_;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    my @subtests = &ParamsMgr::get_arg(-arg => "subtests",
                                       -ref_type => "ARRAY",
                                       -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to parse arguments", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Need to wait for ALL vservers to complete.
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("Unable to wait on subtests to finish", &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # Get the status of a subtest and report it
    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != $SanMgr::FALSE) {
        $self->sanlog(-msg => "Found a vserver cleanup subtest failure:" .
                              "\n\t\tFailed: " . $st_status->{failed} .
                              "\n\t\tPassed: " . $st_status->{passed});
        $self->_set_err("Failed to cleanup vservers");
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "All vserver cleanup subtests have passed:" .
                              "\n\t\tFailed: " . $st_status->{failed} .
                              "\n\t\tPassed: " . $st_status->{passed});
        $self->sanlog(-msg => $self->_format_log("Finished vserver cleanup"));
    }

    # Do some house keeping and remove all vserver subtests.
    &SubIPC::remove_subtest(subtest => \@subtests) if (scalar(@subtests));

    return $SanMgr::TRUE;
} ## end sub _reap_vserver_cleanup_subtests

###############################################################################
# Method:    _cleanup_vserver
# Objective: Cleanup vservers created or reused by SanMgr
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cleanup_vserver {
    my ($self, %args) = @_;

    # Locals
    my ($data, $vserver_obj) = undef;
    my ($vserver_name, $reused, $cserver_name) = "";
    my ($cleanup) = $SanMgr::FALSE;

    $self->clear_error();

    $self->sanlog(-msg => $self->_format_log("Start cleaning up vserver"),
                  -add_caller => 1);

    # Process arguments.
    &ParamsMgr::process_args(%args);

    # Vserver data and name
    $data        = &ParamsMgr::get_arg(-arg => "data",
                                       -required => $SanMgr::TRUE);
    $cleanup      = &ParamsMgr::get_arg(-arg => "cleanup",
                                        -required => $SanMgr::FALSE,
                                        -default => "new");
    $reused       = &ParamsMgr::get_arg(-arg => "reused",
                                        -required => $SanMgr::FALSE,
                                        -default => $SanMgr::FALSE);
    $vserver_name = &ParamsMgr::get_arg(-arg => "vserver",
                                        -required => $SanMgr::TRUE);
    $cserver_name = &ParamsMgr::get_arg(-arg => "cserver",
                                        -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());

        return $SanMgr::FALSE;
    }

    $self->sanlog(-msg => $self->_format_log("Start cleaning up vserver:\n"
                                             . "name      is $vserver_name\n"
                                             . "policy    is $cleanup\n"
                                             . "reused    is $reused"));

    # Need to use a new connection on each vserver
    my ($cserver_obj) = $self->get_objects(types => ['node', 'cserver'],
                                           names => [$cserver_name]);
    if ($self->error()) {
        $self->_set_err("Unable to get cserver object for $cserver_name");
        return $SanMgr::FALSE;
    }

    $cserver_obj->reconnect(-disconnect_existing_conn => $SanUtil::FALSE);
    if ($cserver_obj->error()) {
        $self->_set_err("Failed to reconnect to $cserver_name", $cserver_obj->errmsg());
        return $SanMgr::FALSE;
    }

    my (@nacl_cluster_objs) = $self->get_objects('types' => ["node", "cluster"],
                                                 'names' => [$cserver_name],
                                                 'obj_class' => "nacl");
    if ($self->error()) {
        $self->_set_err("Unable to obtain nacl cluster objects for $cserver_name");
        return $SanMgr::FALSE;
    }

    # refresh the command interface for the nacl cluster object
    foreach my $nacl_cluster_obj (@nacl_cluster_objs) {
        $nacl_cluster_obj->refresh_command_interface();
    }

    # Get the vserver object
    ($vserver_obj) = $cserver_obj->get_vserver_obj(-name => $vserver_name);
    if ($self->error()) {
        $self->_set_err("Failed to get object for vserver '$vserver_name'", $cserver_obj->errmsg());
        return $SanMgr::FALSE;
    }

    # Delete portsets
    $self->sanlog(-msg => $self->_format_log("Start cleaning up portsets "
                                           . "on vserver $vserver_name"),
                  -add_caller => 1);

    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "portset");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup portsets for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Delete tpgroups
    $self->sanlog(-msg => $self->_format_log("Start cleaning up tpgroups "
                                           . "on vserver $vserver_name"),
                  -add_caller => 1);
    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "tpgroup");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup tpgroups for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Delete Lun Binds
    $self->sanlog(-msg => $self->_format_log("Start cleaning up luns binds "
                                           . "on vserver $vserver_name"),
                  -add_caller => 1);
    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "lun_bind");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup LUN binds for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Delete luns
    $self->sanlog(-msg => $self->_format_log("Start cleaning up luns "
                                           . "on vserver $vserver_name"),
                  -add_caller => 1);
    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "lun");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup LUNs for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Delete files
    $self->sanlog(-msg => $self->_format_log("Start cleaning up files "
                                           . "on vserver $vserver_name"),
                  -add_caller => 1);
    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "file");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup FILEs for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Delete igroups
    $self->sanlog(-msg => $self->_format_log("Start cleaning up igroups "
                                           . "on vserver $vserver_name"),
                  -add_caller => 1);
    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "igroup");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup igroups for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Delete qtrees
    $self->sanlog(-msg => $self->_format_log("Start cleaning up qtrees "
                                           . "on vserver $vserver_name"),
                  -add_caller => 1);
    $self->_cleanup_obj(data        => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver     => $vserver_name,
                        type        => "qtree");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup qtrees for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Delete volumes
    $self->sanlog(-msg => $self->_format_log("Start cleaning up volumes "
                                           . "on vserver $vserver_name"),
                  -add_caller => 1);
    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "volume");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup volumes for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Delete lifs
    $self->sanlog(-msg => $self->_format_log("Start cleaning up lifs "
                                           . "on vserver $vserver_name"),
                  -add_caller => 1);
    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "lif");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup lifs for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Cleanup iSNS Settings
    $self->sanlog(-msg => $self->_format_log("Start cleaning up iSNS "
                                           . "settings on vserver "
                                           . $vserver_name),
                  -add_caller => 1);

    $self->_cleanup_obj(data        => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver     => $vserver_name,
                        type        => "isns_service");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup the iSNS settings for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Cleanup iSCSI Security Settings
    $self->sanlog(-msg => $self->_format_log("Start cleaning up iscsi "
                                           . "security settings on vserver "
                                           . $vserver_name),
                  -add_caller => 1);

    $self->_cleanup_obj(data        => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver     => $vserver_name,
                        type        => "iscsi_security_setting");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup the iSCSI security settings for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Delete protocol services
    $self->sanlog(-msg => $self->_format_log("Start cleaning up iscsi "
                                           . "service on vserver "
                                           . $vserver_name),
                  -add_caller => 1);

    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "iscsi_service");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup the iSCSI service for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # FCP service cleanup
    $self->sanlog(-msg => $self->_format_log("Start cleaning up fcp "
                                           . "service on vserver "
                                           . $vserver_name),
                  -add_caller => 1);
    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "fcp_service");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup the FCP service for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Qos policy group cleanup
    $self->sanlog(-msg => $self->_format_log("Start cleaning up qos policy "
                                           . "group on vserver "
                                           . $vserver_name),
                  -add_caller => 1);
    $self->_cleanup_obj(data    => $data,
                        cserver_obj => $cserver_obj,
                        vserver_obj => $vserver_obj,
                        vserver => $vserver_name,
                        type    => "qos_policy_group");
    if ($self->error()) {
        $self->_set_err("Failed to cleanup the Qos policy groups for vserver '$vserver_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Check if this vserver was reused or if it was created by sanmgr and
    # use the policy to determine whether it needs to be destroyed or not
    if (($reused) and ($cleanup =~ /new/i)) {
        $self->sanlog(-msg => $self->_format_log(
                      "Skipping cleaning up "
                    . "for reused vserver $vserver_name"));
    } else {
        $self->sanlog(-msg => $self->_format_log(
                          "Clean up is deleting vserver $vserver_name"));

        # Delete the vserver.  We should only get here if the cleanup policy
        # and reused flag tell us it's OK to delete the vserver.  Note that
        # the force flag will cause the vserver object to stop services,
        # delete routing groups, kill any remaining igroups, lifs, luns, etc.
        $vserver_obj->delete(force => $SanMgr::TRUE);
        if ($vserver_obj->error()) {
            $self->_set_err("Unable to delete vserver $vserver_name" ,
                            $vserver_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # disconnect from the cserver
    $cserver_obj->disconnect_conn();

    &SubIPC::pass(-msg => "Finished cleaning up vserver $vserver_name");

    return $SanMgr::TRUE;
} # End of _cleanup_vserver

###############################################################################
# Method:    _cleanup_obj
# Objective: Cleanup objects contained on a vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cleanup_obj {
    my ($self, %args) = @_;

    # Locals
    my ($vserver_name, $type, $vvol_path) = "";
    my ($data, $cserver_obj, $vserver_obj, $iscsi_obj, $nacl_cluster_obj) = undef;
    my (@lun_objs, @vol_objs, @qtree_objs, @igroup_objs, @iscsi_security_show)  = ();

    $self->clear_error();

    # Process arguments.
    &ParamsMgr::process_args(%args);

    $data = &ParamsMgr::get_arg(-arg      => "data",
                                -required => $SanMgr::TRUE);

    $cserver_obj = &ParamsMgr::get_arg(-arg     => "cserver_obj",
                                       -default => undef);

    $vserver_obj = &ParamsMgr::get_arg(-arg     => "vserver_obj",
                                       -default => undef);

    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver",
                                        -required => $SanMgr::FALSE,
                                        -default  => "_");

    $type = &ParamsMgr::get_arg(-arg      => "type",
                                -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Log the entry we are adding to the cleanup stack
    my $dumpout = Dumper($data);
    $self->sanlog(-msg => "Cleaning up $type objects on vserver $vserver_name "
                        . "from this cleanup stack:\n"
                        . $dumpout,
                  -msg_level  => 3,
                  -add_caller => 1);

    # We only need the nacl cluster object for lun binds or qos policy groups
    if($type eq "qos_policy_group" or
       $type eq "lun_bind" or
       $type eq "file") {
        # We need the cserver name to get the nacl cluster objects from SanMgr
        my $cluster_name = $cserver_obj->get_name();
        if ($cserver_obj->error()) {
            $self->_set_err("Unable to parse cserver name", $cserver_obj->errmsg());
            return $SanMgr::FALSE;
        }

        # Get nacl cluster object for use with cleanup for NACL created Objects.
        ($nacl_cluster_obj) = $self->get_objects('type' => "node",
                                                 'names' => [$cluster_name],
                                                 'obj_class' => 'nacl');
        if ($self->error()) {
            # Error is already set on $self, just return.
            $self->_set_err("Unable to obtain nacl cluster object for " . $cluster_name);
            return $SanMgr::FALSE;
        }
    }

    # initialize a few things before looping through the items to cleanup
    if ($type eq "lun") {
        # fetch all lun objects right away if type is lun
        @lun_objs = $cserver_obj->fetch_lun_objs(-vserver => $vserver_name);
        if ($cserver_obj->error()) {
            $self->_set_err("Failed to fetch all LUN objects in vserver '$vserver_name'",
                            $cserver_obj->errmsg());
            return $SanMgr::FALSE;
        }
    } elsif ($type eq "volume") {
        # fetch all the vol objects right away if type is volume
        @vol_objs = $vserver_obj->get_volume_objs();
        if ($vserver_obj->error()) {
            $self->_set_err("Failed to fetch all Volume objects in vserver '$vserver_name'",
                            $cserver_obj->errmsg());
            return $SanMgr::FALSE;
        }
    } elsif ($type eq "qtree") {
        # fetch all qtree objects right away if type is qtree
        @qtree_objs = $cserver_obj->fetch_qtree_objs(-vserver => $vserver_name);
        if ($cserver_obj->error()) {
            $self->_set_err("Failed to fetch all qtree objects in vserver '$vserver_name'",
                            $cserver_obj->errmsg());
            return $SanMgr::FALSE;
        }
    } elsif ($type eq "igroup") {
        # fetch all igroup objects right away if type is igroup
        @igroup_objs = $cserver_obj->fetch_igroup_objs(-vserver => $vserver_name);
        if ($cserver_obj->error()) {
            $self->_set_err("Failed to fetch all igroup objects in vserver '$vserver_name'",
                            $cserver_obj->errmsg());
            return $SanMgr::FALSE;
        }
    } elsif ($type eq "iscsi_security_setting") {
        # fetch all iscsi security settings
        if ($vserver_obj->iscsi_exist()) {
            $iscsi_obj = $vserver_obj->get_iscsi_obj();
            @iscsi_security_show = $iscsi_obj->initiator_auth_list_info();
            if ($iscsi_obj->error()) {
                $self->_set_err("Failed to obtain current iSCSI security settings");
                return [];
            }
        }
    }

    # Go through all the clean up stack and delete objects according to the
    # cleanup policy
    foreach my $item (@{$data}) {
        if ($item->{reused}) {
            if ($item->{cleanup} =~ /new/i) {
                my $itemdump = Dumper($item);
                $self->sanlog(-msg => "Skipping cleanup for reused item:\n"
                                      . $itemdump,
                              -add_call_stack => 1);

                next;
            }
        }

        # Check to see if the type of the current item is the same as
        # the type sent to the method call.  If it isn't skip to the next item.
        if ($item->{type} eq $type) {
            my $itemdump = Dumper($item);
            $self->sanlog(-msg => "Cleaning up $type item:\n $itemdump");
        } else {
            my $itemdump = Dumper($item);
            $self->sanlog(-msg => "Not Cleaning up item "
                          . "because item is not a $type:\n $itemdump",
                          -msg_level      => 3,
                          -add_call_stack => 1);
            next;
        }

        #if there are lun binds we must remove them first before
        #anything else is done.
        if ($item->{type} =~ /^lun_bind$/i){
            #build up the vvol path
            if (defined($item->{qtree}) and length($item->{qtree})) {
                $vvol_path = "/vol/$item->{volume}/$item->{qtree}/$item->{name}";
            } else {
                $vvol_path = "/vol/$item->{volume}/$item->{name}";
            }

            foreach my $pep (@{$item->{lun_binds}}){
                try {
                    #do lun unbinds on a per vvol basis
                    NACL::C::LunBind->destroy( command_interface => $nacl_cluster_obj,
                                               vserver => $vserver_name,
                                               'vvol-path' => $vvol_path,
                                               'protocol-endpoint-path' => $pep,
                                               'force' => 'TRUE',
                                              );
                } otherwise {
                    #catch the NACL error
                    my $e = shift;

                    # set an error only if it wasn't a "does not exist' error
                    if ($e->text() !~ /entry doesn\'t exist|does not exist/i) {
                        $self->_set_err("NACL Exception " . ref($e) . " caught: ". $e->text());
                    } else {
                        $self->sanlog(-msg => "Cannot cleanup lun bind $vvol_path (bound to $pep). It doesn't exist, or the protocol endpoint $pep is missing.",
                                      -msg_type => "warn");
                    }
                };
                if ($self->error()){
                    return $SanUtil::FALSE;
                }
            }
        }

        if ($item->{type} =~ /^lun$/i) {
            # Obtain the lun object that matches this lun item
            my $lun_obj;
            foreach my $obj (@lun_objs) {
                if ($obj->get_name() eq $item->{name}) {
                    # If two volumes in this vserver have luns with the same name,
                    # we could cleanup the wrong lun.  If we know the names match,
                    # check the paths.
                    my $item_path = "/vol/" . $item->{volume};
                    if (defined($item->{qtree}) and length($item->{qtree}) and $item->{qtree} ne "-") {
                        $item_path .= "/" . $item->{qtree};
                    }
                    $item_path .= "/" . $item->{name};
                    my $lun_path = $obj->get_path(-use_cache => $SanMgr::TRUE);

                    # If the paths match, we found the right lun
                    # and we get out of the loop
                    if ($item_path eq $lun_path) {
                        $lun_obj = $obj;
                        last;
                    }
                }
            }

            # if the object doesnt exist, warn
            if (not defined($lun_obj)) {
                $self->sanlog(-msg => "Cannot cleanup LUN " . $item->{name} . ". It does not exist in the cluster.",
                              -msg_type => "warn");
                next;
            }

            # Start by checking if it's mapped or not and unmap it if needed
            if ($lun_obj->is_mapped()) {
                foreach my $lun_map (@{$item->{lun_maps}}) {
                    $lun_obj->unmap(igroup => $lun_map);
                    if ($lun_obj->error()) {
                        $self->_set_err("Unable to unmap LUN $item->{name} from $lun_map",
                                        $lun_obj->errmsg());
                        return $SanMgr::FALSE;
                    }
                }
            }

            # Delete the lun
            $lun_obj->delete();
            if ($lun_obj->error()) {
                $self->_set_err("Unable to delete LUN $item->{name}",
                                $lun_obj->errmsg());
                return $SanMgr::FALSE;
            }
        } elsif ($item->{type} =~ /^igroup$/i) {
            # find the igroup object in the list
            my $igroup_obj;
            foreach my $obj (@igroup_objs) {
                if ($obj->get_name(-use_cache => $SanMgr::TRUE) eq $item->{name}) {
                    $igroup_obj = $obj;
                    last;
                }
            }

            # if the object doesnt exist, warn
            if (not defined($igroup_obj)) {
                $self->sanlog(-msg => "Cannot cleanup Igroup " . $item->{name} . ". It does not exist in the vserver $vserver_name.",
                              -msg_type => "warn");
                next;
            }

            # Delete the igroup (force it)
            $igroup_obj->delete(force => "true");
            if ($igroup_obj->error()) {
                $self->_set_err("Failed to delete igroup $item->{name}",
                                $igroup_obj->errmsg());
                return $SanMgr::FALSE;
            }
        } elsif ($item->{type} =~ /^qtree$/i) {
            # Obtain the qtree object that matches this qtree item
            # TODO: Check the case where we have a qtree with the same
            # name in two different volumes (same for lun)
            my $qtree_obj;
            foreach my $obj (@qtree_objs) {
                if ($obj->get_name() eq $item->{name}) {
                    # If two volumes in this vserver have qtrees with the same name,
                    # we could cleanup the wrong qtree.  If we know the names match,
                    # check the paths.
                    my $item_path  = "/vol/" . $item->{volume} . "/" . $item->{name};
                    my $qtree_path = $obj->get_path(-use_cache => $SanMgr::TRUE);

                    # If the paths match, we found the right qtree
                    # and we get out of the loop
                    if ($item_path eq $qtree_path) {
                        $qtree_obj = $obj;
                        last;
                    }
                }
            }

            # if the object doesnt exist, warn
            if (not defined($qtree_obj)) {
                $self->sanlog(-msg => "Cannot cleanup qtree " . $item->{name} . ". It does not exist in the cluster.",
                              -msg_type => "warn");
                next;
            }

            # Delete the qtree
            $qtree_obj->delete();
            if ($qtree_obj->error()) {
                $self->_set_err("Unable to delete qtree $item->{name}",
                                $qtree_obj->errmsg());
                return $SanMgr::FALSE;
            }

        } elsif ($item->{type} =~ /^volume$/i) {
            # obtain the vol object that matches this volume item
            my $vol_obj;
            foreach my $obj (@vol_objs) {
                if (($obj->get_name() eq $item->{name})  and
                    ($vserver_name    eq $item->{vserver})) {
                    $vol_obj = $obj;
                    last;
                }
            }

            # if the object doesnt exist, warn
            if (not defined($vol_obj)) {
                $self->sanlog(-msg => "Cannot cleanup volume " . $item->{name} . ". It does not exist in the cluster.",
                              -msg_type => "warn");
                next;
            }

            $vol_obj->delete(force => "true");
            if ($vol_obj->error()) {
                $self->_set_err("Failed to delete volume $item->{name}",
                                $vol_obj->errmsg());
                return $SanMgr::FALSE;
            }
        } elsif ($item->{type} =~ /^portset$/i) {
            # Get portsets created or reused by sanmgr
            my $portset_obj = $vserver_obj->get_portset_obj(portset => $item->{name});
            if ($vserver_obj->error()) {
                if ($vserver_obj->errmsg() =~ /failed to get portset object|can not find portset/i) {
                    # This is a quick fix to patch an error in building the
                    # portset and igroup bind cleanup objects. When there are
                    # multiple igroup binds then there are multiple portset
                    # cleanup items created when there should only be one
                    # cleanup item for that portset.
                    next;
                } else {
                    $self->_set_err("Failed to get portset $item->{name} object for vserver $vserver_name",
                                    $vserver_obj);
                    return $SanMgr::FALSE;
                }
            }

            # if the object doesnt exist, warn
            if (not defined($portset_obj)) {
                $self->sanlog(-msg => "Cannot cleanup portset " . $item->{name} . ". It does not exist in the cluster.",
                              -msg_type => "warn");
                next;
            }

            # destroy the portset (force it)
            $portset_obj->destroy(force => "true");
            if ($portset_obj->error()) {
                $self->_set_err("Failed to delete portset $item->{name}",
                                $portset_obj->errmsg());
                return $SanMgr::FALSE;
            }
        } elsif ($item->{type} =~ /^tpgroup$/i) {
            # first we need the iscsi object on the vserver
            my $iscsi_obj = $vserver_obj->get_iscsi_obj();
            $vserver_obj->_clear_err();

            # Get the tpgroup from the iscsi object
            my $tpgroup_obj = $iscsi_obj->get_tpgroup_obj(tpgroup => $item->{name});
            if ($iscsi_obj->error()) {
                $self->_set_err("Failed to get tpgroup $item->{name} object for vserver $vserver_name",
                                $iscsi_obj->errmsg());
            }

            # if the object doesnt exist, warn
            if (not defined($tpgroup_obj)) {
                $self->sanlog(-msg => "Cannot cleanup tpgroup " . $item->{name} . ". It does not exist in the cluster.",
                              -msg_type => "warn");
                next;
            }

            # destroy the tpgroup
            $tpgroup_obj->destroy(force => "true");
            if ($tpgroup_obj->error()) {
                $self->_set_err("Failed to delete tpgroup $item->{name}",
                                $tpgroup_obj->errmsg());
                return $SanMgr::FALSE;
            }
        } elsif ($item->{type} =~ /^lif$/i) {
            # Get lifs created or reused by sanmgr
            my $lif_obj = $vserver_obj->get_lif_obj(lif => $item->{name});
            if ($vserver_obj->error()) {
                $self->_set_err("Failed to get LIF $item->{name} object for vserver $vserver_name",
                                $vserver_obj->errmsg());
                return $SanMgr::FALSE;
            }

            # if the object doesnt exist, warn
            if (not defined($lif_obj)) {
                $self->sanlog(-msg => "Cannot cleanup lif " . $item->{name} . ". It does not exist in the cluster.",
                              -msg_type => "warn");
                next;
            }

            # destroy the lif
            $lif_obj->delete();
            if ($lif_obj->error()) {
                $self->_set_err("Failed to delete lif $item->{name}",
                                $lif_obj->errmsg());
                return $SanMgr::FALSE;
            }
        } elsif ($item->{type} =~ /^qos_policy_group$/i) {
            try {
                NACL::C::QosPolicyGroup->delete(
                                            command_interface => $nacl_cluster_obj,
                                            'policy-group'      => $item->{name}
                                            );
            } catch NACL::APISet::Exceptions::ResponseException with {
                my $exception = shift;
                if ($exception->text() =~ /entry doesn't exist/) {
                    # if the object doesnt exist, warn
                    $self->sanlog(-msg => "Cannot cleanup qos policy group " .
                                          $item->{name} . ". It does not" .
                                          " exist in the cluster.",
                                  -msg_type => "warn");
                } elsif ($exception->text() =~
                        /Policy group could not be removed because there are still workloads associated with it/i) {
                    # If the qos policy group still has workloads associated with it
                    # then warn and leave the policy group.  This could be because the
                    # user associated the policy group with the containing vserver or
                    # something else that the user created but did not clean up.
                    $self->sanlog(-msg =>
                                "Cannot cleanup qos policy group " .
                                $item->{name} . ". It is still associated" .
                                " with something in vserver: " . $item->{name} .
                                ". If the vserer is cleaned up then this policy" .
                                " group will also be removed with the deletion" .
                                " of the vserver. Otherwise this policy group" .
                                " will remain.",
                                -msg_type => "warn");
                } else {
                    $self->_set_err("Failed to delete qos policy group $item->{name}",
                                    $exception->text());
                }
            }; # end of delete try/catch
            if ($self->error()) {
                return $SanMgr::FALSE;
            }
        } elsif ($item->{type} =~ /^file$/i) {
            try {
                NACL::C::VolumeFile->delete_file(command_interface => $nacl_cluster_obj,
                                                 file => $item->{name},
                                            );
            } catch NACL::APISet::Exceptions::ResponseException with {
                my $exception = shift;
                if ($exception->text() =~ /No such file/) {
                    # if the object doesnt exist, warn
                    $self->sanlog(-msg => "Cannot cleanup file" .
                                          $item->{name} . ". It does not" .
                                          " exist in the cluster.",
                                          -msg_type => "warn");
                } else {
                    $self->_set_err("Failed to delete file $item->{name}",
                                    $exception->text());
                }
            }; # end of delete try/catch
            if ($self->error()) {
                return $SanMgr::FALSE;
            }
        }elsif ($item->{type} =~ /^aggregate/i) {
            # skip aggr cleanup if the told to not cleanup aggrs
            if ($item->{cleanup} =~ /noaggr/i) {
                $self->sanlog(
                    -msg => $self->_format_log("Skipping aggregate cleanup "
                                             . "because the cleanup policy "
                                             . "is 'noaggr'"));
                next;
            }

            # get the cserver object that was used to create this aggregate
            my $cserver_name = $item->{cserver};
            ($cserver_obj) = $self->get_objects(types => ["node", "cserver"],
                                                names => [$cserver_name]);
            if ($self->error()) {
                $self->_set_err("Unable to get node/cserver object for '$item->{cserver}'");
                return $SanMgr::FALSE;
            }

            # obtain the aggr
            my $aggr_obj = $cserver_obj->get_aggregate(aggregate => $item->{name});
            if ($cserver_obj->error()) {
                $self->_set_err("Failed to get aggregate $item->{name} object for vserver $vserver_name",
                                $cserver_obj->errmsg());
                return $SanMgr::FALSE;
            }

            # if the object doesnt exist, warn
            if (not defined($aggr_obj)) {
                $self->sanlog(-msg => "Cannot cleanup aggregate " . $item->{name} . ". It does not exist in the cluster.",
                              -msg_type => "warn");
                next;
            }

            # Delete the aggr
            $aggr_obj->delete();
            if ($aggr_obj->error()) {
                $self->_set_err("Failed to delete aggregate $item->{name}",
                                $aggr_obj->errmsg());
                return $SanMgr::FALSE;
            }
        } elsif ($item->{type} =~ /^iscsi_service$/i) {

            # Get the vserver object
            my ($vserver_obj) = $self->get_objects(type    => 'vserver',
                                                   names   => [$vserver_name]);
            if ($self->error()) {
                $self->_set_err("Failed to get object for vserver $vserver_name");
                return $SanMgr::FALSE;
            }

            # Check to see if the iscsi_service is running.  If it is,
            # we clean it up.  If it's not, cleanup is a noop. We do
            # this because get_iscsi_obj() will create it if it's
            # not there.
            if ($vserver_obj->iscsi_exist()) {

                # Get the iscsi object
                my $iscsi_obj = $vserver_obj->get_iscsi_obj();
                if ($vserver_obj->error()) {
                    $self->_set_err("Could not get iscsi object", $vserver_obj->errmsg());
                    return $SanMgr::FALSE;
                }

                # Stop the iscsi service
                if (defined ($item->{status_admin}) and
                    ($item->{status_admin} =~ /down/i)) {
                   $self->sanlog(-msg => "Clean up for iscsi service is "
                                    . "skipping the iscsi stop because "
                                    . "the user started the service with "
                                    . "status-admin down");
                } else {
                    $iscsi_obj->stop();
                    if ($iscsi_obj->error()) {
                        my $errmsg = join "\n", $iscsi_obj->errmsg();
                        if ($errmsg =~ /Service.*not started/) {
                           $self->sanlog(-msg => "Iscsi stop failed because the"
                                               . " service wasn't started. "
                                               . "Ignoring error.");
                        } else {
                           $self->_set_err("Could not stop iscsi", $iscsi_obj->errmsg());
                           return $SanMgr::FALSE;
                       }
                    }
                }

                # Delete the iscsi service
                $iscsi_obj->delete();
                if ($iscsi_obj->error()) {
                    $self->_set_err("Could not delete iscsi", $iscsi_obj->errmsg());
                    return $SanMgr::FALSE;
                }
            } else {
                $self->sanlog(-msg => "Clean up for iscsi service is a noop"
                                    . " because iscsi does not exist");
            }
        } elsif ($item->{type} =~ /^fcp_service$/i) {
            # Get the vserver object
            my ($vserver_obj) = $self->get_objects(type    => 'vserver',
                                                   names   => [$vserver_name]);
            if ($self->error()) {
                $self->_set_err("Failed to get object for vserver $vserver_name");
                return $SanMgr::FALSE;
            }

            # Check to see if the fcp service is running.  If it is,
            # we clean it up.  If it's not, cleanup is a noop.
            # We do this because get_fcp_obj() will create it if
            # it's not there.
            if ($vserver_obj->fcp_exist()) {

                # Get the fcp object
                my $fcp_obj = $vserver_obj->get_fcp_obj();
                if ($vserver_obj->error()) {
                    $self->_set_err("Could not get fcp object", $vserver_obj->errmsg());
                    return $SanMgr::FALSE;
                }

                # Stop the fcp service
                if (defined ($item->{status_admin}) and
                    ($item->{status_admin} =~ /down/i)) {
                    $self->sanlog(-msg => "Clean up for fcp service is "
                                    . "skipping the fcp stop because "
                                    . "the user started the service with "
                                    . "status-admin down");
                } else {
                    $fcp_obj->stop();
                    if ($fcp_obj->error()) {
                        my $errmsg = join "\n", $fcp_obj->errmsg();
                        if ($errmsg =~ /Service.*not started/) {
                           $self->sanlog(-msg => "FCP stop failed because the"
                                               . " service wasn't started. "
                                               . "Ignoring error.");
                        } else {
                           $self->_set_err("Could not stop fcp", $fcp_obj->errmsg());
                           return $SanMgr::FALSE;
                        }
                    }
                }

                # Delete the fcp service
                $fcp_obj->delete();
                if ($fcp_obj->error()) {
                    $self->_set_err("Could not delete fcp", $fcp_obj->errmsg());
                    return $SanMgr::FALSE;
                }
            } else {
                $self->sanlog(-msg => "Clean up for fcp service is a noop"
                                    . " because fcp does not exist");
            }
        } elsif ($item->{type} =~ /^iscsi_security_setting$/) {

            # if we have the iscsi object defined, clean up this iscsi security setting
            if (defined($iscsi_obj)) {

                # check to see if this initiator nodename has an iscsi security setting
                my $initiator_nodename = $item->{initiator};

                # only delete this security setting if it is in the iscsi security show output
                # or
                # if the initiator nodename is a wildcard (eui*|iqn*) or (*)
                if (grep { $_->{initiator_name} =~ /^${initiator_nodename}$/ } @iscsi_security_show or
                    $initiator_nodename eq "eui*|iqn*" or $initiator_nodename eq "*") {
                    # delete the iscsi security setting if it exists
                    $iscsi_obj->initiator_delete_auth('initiator-name' => $initiator_nodename);
                    if ($iscsi_obj->error()) {
                        # If the error is the message about not being able
                        # to delete the default ISCSI security setting or that there are no entries, ignore
                        # it and move on.
                        if ($iscsi_obj->errmsg() !~ /Cannot delete default|There are no entries/) {
                            $self->_set_err("Failed to remove iSCSI security setting for initiator '$initiator_nodename'");
                            return $SanMgr::FALSE;
                        }
                    }
                }
            } else {
                $self->sanlog(-msg => "Cannot cleanup iscsi security setting " . $item->{initiator} .
                                      ". There is no iSCSI service for this vserver.",
                              -msg_type => "warn");
                next;
            }
        } elsif ($item->{type} =~ /^isns_service$/i) {

            # Get the vserver object
            my ($vserver_obj) = $self->get_objects(type    => 'vserver',
                                                   names   => [$vserver_name]);
            if ($self->error()) {
                $self->_set_err("Failed to get object for vserver $vserver_name");
                return $SanMgr::FALSE;
            }

            # Check to see if the iscsi_service is running.  If it is,
            # we clean it up.  If it's not, cleanup is a noop. We do
            # this because get_iscsi_obj() will create it if it's
            # not there.

            if ($vserver_obj->iscsi_exist()) {
                # Get the iscsi object
                my $iscsi_obj = $vserver_obj->get_iscsi_obj();
                if ($vserver_obj->error()) {
                    $self->_set_err("Could not get iscsi object", $vserver_obj->errmsg());
                    return $SanMgr::FALSE;
                }

                my $isns_cleanup_args = {};
                $isns_cleanup_args->{"-vserver"} = $vserver_name;
                $iscsi_obj->isns_delete(%{$isns_cleanup_args});

                if ($iscsi_obj->error()) {
                    my $errmsg = join "\n", $iscsi_obj->errmsg();
                    if ($errmsg =~ /Cannot find/) {
                       $self->sanlog(-msg => "Isns delete failed because the"
                                               . " service isn't running. "
                                               . "Ignoring error.");
                    } else {
                       $self->_set_err("Could not stop isns", $iscsi_obj->errmsg());
                       return $SanMgr::FALSE;
                   }
                }
            }
        } elsif ($item->{type} =~ /^vlan$/i) {

            my $cluster_name = $item->{cluster_name};

            # Get nacl cluster object for use with cleanup for NACL created Objects.
            ($nacl_cluster_obj) = $self->get_objects('type' => "node",
                                                     'names' => [$cluster_name],
                                                     'obj_class' => 'nacl');
            if ($self->error()) {
                # Error is already set on $self, just return.
                $self->_set_err("Unable to obtain nacl cluster object for " . $cluster_name);
                return $SanMgr::FALSE;
            }

            try {
                NACL::C::NetworkPortVlan->delete(
                                    command_interface => $nacl_cluster_obj,
                                    node              => $item->{node},
                                    'vlan-id'         => $item->{'vlan-id'},
                                    port              => $item->{port});

            } otherwise {
                my $exception = shift;
                if ($exception->text() =~ /no entries/i) {
                    $self->sanlog(-msg => "Cannot cleanup vlan " . $item->{'vlan-id'} . ". It does not exist in the cluster.",
                                  -msg_type => "warn");
                } else {
                    $self->_set_err("Unable to delete vlan " . $item->{'vlan-id'} . " "
                        . $exception->text());
                }
            };
            if ($self->error) {
                return $SanMgr::FALSE;
            }

        } elsif ($item->{type} =~ /^ifgrp$/i) {

            my $cluster_name = $item->{cluster_name};

            # Get nacl cluster object for use with cleanup for NACL created Object
            ($nacl_cluster_obj) = $self->get_objects('type' => "node",
                                                     'names' => [$cluster_name],
                                                     'obj_class' => 'nacl');
            if ($self->error()) {
                # Error is already set on $self, just return.
                $self->_set_err("Unable to obtain nacl cluster object for " . $cluster_name);
                return $SanMgr::FALSE;
            }

            try {
                NACL::C::NetworkPortIfgrp->delete(
                                    command_interface => $nacl_cluster_obj,
                                    node              => $item->{node},
                                    'ifgrp'           => $item->{name});
            } otherwise {
                my $exception = shift;
                if ($exception->text() =~ /entry doesn't exist/i) {
                    $self->sanlog(-msg => "Cannot cleanup ifgrp " . $item->{name} . ". It does not exist in the cluster.",
                                  -msg_type => "warn");
                } else {
                    $self->_set_err("Unable to delete ifgrp " . $item->{name} . " "
                        . $exception->text());
                }
            };
            if ($self->error) {
                return $SanMgr::FALSE;
            }
        }
    } # end foreach item in data stack

    return $SanMgr::TRUE,
}

###############################################################################
# Method:    _run_wafliron_on_aggregates
# Objective: Run wafliron on aggregates that have it enabled
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _run_wafliron_on_aggregates {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # The main cleanup stack
    my $data = &ParamsMgr::get_arg(-arg => "data", -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to parse args", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # filter the aggregate data down to only the aggrs that have "wafliron" set to 1
    my @aggr_data;
    foreach my $key (keys %{$data}) {
        if ($key eq "aggregate") {
            foreach my $item (@{$data->{$key}}) {
                if ($item->{wafliron} == 1) {
                    push(@aggr_data, $item);
                }
            }
        }
    }

    # return immediately if there are no aggrs to run wafliron
    if (not scalar(@aggr_data)) {
        $self->sanlog(-msg => "No aggregates have been specified to run wafliron");
        return $SanMgr::TRUE;
    }

    # run wafliron on all the aggrs with it enabled
    my @aggr_objs;
    my $cserver_obj;
    foreach my $item (@aggr_data) {
        $self->sanlog(-msg => $self->_format_log("Start wafliron on aggregate " . $item->{name}),
                      -add_caller => 1);

        # get the cserver object that was used to create this aggregate
        my $cserver_name = $item->{cserver};
        ($cserver_obj) = $self->get_objects(types => ["node", "cserver"],
                                            names => [$cserver_name]);
        if ($self->error()) {
            $self->_set_err("Unable to get node/cserver object for '$item->{cserver}'");
            return $SanMgr::FALSE;
        }

        # obtain the aggr object
        my $aggr_obj = $cserver_obj->get_aggregate(aggregate => $item->{name});
        if ($cserver_obj->error()) {
            $self->_set_err("Failed to get aggregate $item->{name} object",
                            $cserver_obj->errmsg());
            return $SanMgr::FALSE;
        }

        # start wafliron on this aggr
        $aggr_obj->wafliron_start();
        if ($aggr_obj->error()) {
            $self->_set_err("Failed to start wafliron on aggr " . $item->{name} .
                            $aggr_obj->errmsg());
            return $SanMgr::FALSE;
        }

        # add this object to the list
        push(@aggr_objs, $aggr_obj);
    }

    # wait for wafliron to finish on all the aggrs that we started wafliron
    $cserver_obj->wait_for_wafliron_to_complete(-aggr_objs => \@aggr_objs);
    if ($cserver_obj->error()) {
        $self->_set_err("Failed while waiting for wafliron to complete on aggregates",
                        $cserver_obj->errmsg());
        return $SanMgr::FALSE;
    }

    return $SanMgr::TRUE;
}

1;
