#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to create_cleanup_obj controller clusters
## @author dalmond@netapp.com
## @status private
## @pod here

package San::SanMgr::Controller::create_cleanup_obj;

# Compiler directives.
use strict;
use San::SanLog;
use ParamsMgr;
use San::SanUtil;
use San::SanMgr::Controller::ControllerHelper;
use NACL::CS::StorageAggregate;
use NACL::CS::Vserver;
use NACL::CS::VserverFcp;
use NACL::CS::VserverIscsi;
use NACL::CS::VserverIscsiTpgroup;
use NACL::CS::Volume;
use NACL::CS::Lun;
use NACL::CS::LunIgroup;
use NACL::CS::LunPortset;
use NACL::CS::VolumeQtree;
use NACL::CS::NetworkPortIfgrp;
use NACL::CS::NetworkPortVlan;
use NACL::CS::NetworkInterface;
use NACL::Exceptions::NoElementsFound;
use NATE::BaseException qw(:try);

use Data::Dumper;

use vars qw(@ISA);
@ISA = qw(Error SanLog San::SanMgr::Controller::ControllerHelper);

###############################################################################
# Method:    create_cleanup_obj_controller_cluster
# Objective: create_cleanup_obj run against a single controller cluster
###############################################################################
sub create_cleanup_obj_controller_cluster {
    my ($self, %args) = @_;
    my @cluster_cleanup_objs = ();

    $self->clear_error();

    # Local Variables
    my ($resource, $cserver, $aggr_data) = undef;
    my ($vlan_data, $ifgrp_data, $vserver_data) = undef;

    # Process arguments. A resource is nothing more than a controller cluster
    # element which may come from a config file (i.e. XML)
    &ParamsMgr::process_args(%args);

    # The main controller cluster data structure
    $resource = &ParamsMgr::get_arg(-arg      => "resource",
                                    -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to parse arguments", &ParamsMgr::get_error());
        return [];
    }

    # get the nacl cluster object for this cluster, we'll use this to send fetch commands
    my ($nacl_cluster_obj) = $self->get_objects(types => ["node", "cluster"],
                                                names => [$resource->{NAME}],
                                                obj_class => "nacl");
    if ($self->error()) {
        # Error is already set on $self, just return.
        $self->_set_err("Unable to obtain NACL cluster object for " . $resource->{NAME});
        return [];
    }

    # Create aggregate cleanup data first
    $aggr_data = $self->_create_cleanup_obj_aggregates(resource => $resource,
                                                       cserver => $nacl_cluster_obj);
    if ($self->error()) {
        return [];
    }
    push(@cluster_cleanup_objs, @{$aggr_data});

    # load ifgrp cleanup data
    $ifgrp_data = $self->_create_cleanup_obj_ifgrps(resource => $resource,
                                                    cserver => $nacl_cluster_obj);
    if ($self->error()) {
        return [];
    }
    push(@cluster_cleanup_objs, @{$ifgrp_data});

    # load vlan cleanup data
    $vlan_data = $self->_create_cleanup_obj_vlans(resource => $resource,
                                                  cserver => $nacl_cluster_obj);
    if ($self->error()) {
        return [];
    }
    push(@cluster_cleanup_objs, @{$vlan_data});

    # load vserver cleanup data
    $vserver_data = $self->_create_cleanup_obj_vservers(resource  => $resource,
                                                        cserver   => $nacl_cluster_obj);
    if ($self->error()) {
        return [];
    }

    push(@cluster_cleanup_objs, @{$vserver_data});

    return \@cluster_cleanup_objs;
} ## end sub create_cleanup_obj_controller_cluster

###############################################################################
# PRIVATE METHODS
#
###############################################################################

###############################################################################
# Method:    _create_cleanup_obj_ifgrps
# Objective: Method responsible for gathering all data corresponding to the
#            ifgrps and creating cleanup data for them.
# Inputs:    A hash containing:
#            %args = (cserver => filer object
#                     resource => hash containing all xml data
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_ifgrps {

    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($ifgrps, $cserver, $data) = undef;
    my (@ifgrp_tag_data, @ifgrp_data) = ();

    # Process ifgrp and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing ifgrp details
    $ifgrps = &ParamsMgr::get_arg(-arg => "resource",
                                  -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    # Get ifgrp data from the resource data structure
    foreach my $key (keys (%{$ifgrps})) {
        if ($key =~ /^IFGRP\d*$/i) {
            # NOTE: Currently Params Mgr does not expand "looped" or "cloned"
            # elements. So we need to generate a list of copies. generate_copies
            # takes care of renaming things, indexes, prefixes, etc.

            push (@ifgrp_tag_data, $self->_generate_copies(element => $ifgrps->{$key},
                                                           type    => "ifgrp"));
        }
    }

    # if there are any ifgrps in the XML, grab the ifgrp objects from the controller
    my @ifgrp_objs;
    my @current_ifgrp_list;
    if (scalar(@ifgrp_tag_data)) {
        # fetch all component states
        try {
            @ifgrp_objs = NACL::CS::NetworkPortIfgrp->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @ifgrp_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }
        # convert the objs to an array of names
        @current_ifgrp_list = map {
            $_->ifgrp();
        } @ifgrp_objs;
    }

    foreach my $ifgrp (@ifgrp_tag_data) {
        # skip this ifgrp if it doesn't exist on the cluster
        my $ifgrp_name = $ifgrp->{NAME};
        if (not grep(/^${ifgrp_name}$/, @current_ifgrp_list)) {
            next;
        }

        my $data = {
            name         => $ifgrp->{NAME},
            type         => "ifgrp",
            node         => $ifgrp->{NODE},
            cluster_name => $cserver->name()
        };

        # Collect all ifgrp data. This will be written to the parent process
        push(@ifgrp_data, $data);
    }

    return \@ifgrp_data;
}

###############################################################################
# Method:    _create_cleanup_obj_vlans
# Objective: Method responsible for gathering all data corresponding to the
#            vlans and creating cleanup data for them.
# Inputs:    A hash containing:
#            %args = (cserver => filer object
#                     resource => hash containing all xml data
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_vlans {

    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($vlans, $cserver, $data) = undef;
    my (@vlan_tag_data, @vlan_data) = ();

    # Process vlan and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing vlan details
    $vlans = &ParamsMgr::get_arg(-arg => "resource",
                                      -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    # Get vlan data from the resource data structure
    foreach my $key (keys (%{$vlans})) {
        if ($key =~ /^VLAN\d*$/i) {
            # NOTE: Currently Params Mgr does not expand "looped" or "cloned"
            # elements. So we need to generate a list of copies. generate_copies
            # takes care of renaming things, indexes, prefixes, etc.

            push (@vlan_tag_data, $self->_generate_copies(element => $vlans->{$key},
                                                          type    => "vlan"));
        }
    }

    # if there are any vlans in the XML, grab the vlan objects from the controller
    my @vlan_objs;
    my @current_vlan_list;
    if (scalar(@vlan_tag_data)) {
        # fetch all component states
        try {
            @vlan_objs = NACL::CS::NetworkPortVlan->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @vlan_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }
        # convert the objs to an array of vlan ids
        @current_vlan_list = map {
            $_->vlan_id();
        } @vlan_objs;
    }

    foreach my $vlan (@vlan_tag_data) {
        # skip this vlan if it doesn't exist on the cluster
        my $vlan_id = $vlan->{'VLAN-ID'};
        if (not grep(/^${vlan_id}$/, @current_vlan_list)) {
            next;
        }

        my $data = {
                      'vlan-id'    => $vlan->{'VLAN-ID'},
                      port         => $vlan->{PORT},
                      node         => $vlan->{NODE},
                      type         => "vlan",
                      cluster_name => $cserver->name()
                   };

        # Collect all vlan data. This will be written to the parent process
        push(@vlan_data, $data);
    }

    return \@vlan_data;
}

###############################################################################
# Method:    _create_cleanup_obj_aggregates
# Objective: Method responsible for gathering all data corresponding to the
#            aggregates and calling the method responsible for creating them.
# Inputs:    A hash containing:
#            %args = (cserver => filer object
#                     aggregates => hash containing all aggr data
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_aggregates {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($aggregates, $cserver, $data) = undef;
    my (@aggrs, @aggr_data, @aggr_objs) = ();

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing aggregate details
    $aggregates = &ParamsMgr::get_arg(-arg => "resource",
                                      -required => $SanMgr::TRUE);

    # cserver object (NACL Cluster object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    # Get aggregate data from the resource data structure
    foreach my $key (keys (%{$aggregates})) {
        if ($key =~ /^AGGR\d*$/i) {
            push (@aggrs, $self->_generate_copies(element => $aggregates->{$key},
                                                  type    => "aggr"));
        }
    }

    my @current_aggr_list;
    # if there are Aggrs in the XML, then fetch the aggr component states
    if (scalar(@aggrs)) {
        # fetch all aggregates component states
        try {
            @aggr_objs = NACL::CS::StorageAggregate->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @aggr_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }
        # convert the objs to an array of names
        @current_aggr_list = map {
            $_->aggregate();
        } @aggr_objs;
    }

    foreach my $aggr (@aggrs) {
        # if an aggr listed in the xml does not currently exist on the cluster,
        # dont make a cleanup object for it
        my $aggr_name = $aggr->{NAME};
        if (not grep(/^${aggr_name}$/, @current_aggr_list)) {
            next;
        }

        my $data = $self->_create_cleanup_obj_aggregate(cserver   => $cserver,
                                                        subtested => $SanMgr::FALSE,
                                                        aggr      => $aggr);


        push(@aggr_data, $data);
    }

    return \@aggr_data;
} ## end sub _create_cleanup_obj_aggregates

###############################################################################
# Method:    _create_cleanup_obj_aggregate
# Objective: Method responsible for creating or reusing aggregate objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     aggr_data => hash containing all args for an aggr
#                     subtested => 1|0 indicating if this is a subtest.)
# Returns:   A hash containing name, reuse flag, cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_aggregate {
    my ($self, %args) = @_;

    $self->clear_error();

    # Locals
    my ($aggr, $cserver, $data) = undef;
    my ($name) = "";
    my ($cleanup_policy) = "";
    my (%aggr_args) = ();
    my ($wafliron) =  $SanMgr::FALSE;

    # Process args and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Hash containing objects params
    $aggr = &ParamsMgr::get_arg(-arg => "aggr",
                                -required => $SanMgr::TRUE);

    # A cserver object (NACL Cluster object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    # Start by getting the "name" of the object. If this is not defined
    # everything else will fail.
    if (defined($aggr->{NAME})) {
        $name = $aggr->{NAME};
    } else {
        $self->_set_err("Missing aggregate name. Check your XML or params.");
        return $SanMgr::FALSE;
    }

    # Get common things like policies and normalize params. NOTE, this does not
    # set an error flag.
    %aggr_args = $self->_process_common_args(%$aggr);
    $cleanup_policy = delete $aggr_args{cleanup_policy};
    $wafliron       = delete $aggr_args{wafliron};

    $data = {
        reused => $SanMgr::TRUE,
        name => $name,
        type => "aggregate",
        wafliron => $wafliron,
        cleanup => $cleanup_policy,
        cserver => $cserver->name(),
    };

    return $data;

} ## end sub _create_cleanup_obj_aggregate

###############################################################################
#                       V S E R V E R    M E T H O D S
#
#   _create_vserver creates vserver objects first, then protocol services
#     and then sets up its contents. Then it gets data from each internal object
#     and bubbles it up to Controller::create_cleanup_obj.
#
###############################################################################

###############################################################################
# Method:    _create_cleanup_obj_vservers
# Objective: Method responsible for creating or reusing vserver objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     vserver_data => hash containing all args for a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_vservers {

    my ($self, %args) = @_;
    my @return_data   = ();

    $self->clear_error();
    # Local Variables
    my ($cserver, $resource) = undef;
    my ($cserver_name) = "";
    my (@vservers, @vserver_data) = ();

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing vserver details
    $resource = &ParamsMgr::get_arg(-arg => "resource",
                                    -required => $SanMgr::TRUE);

    # cserver object (NACL Cluster object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Get the cserver name which is needed to create a vserver
    $cserver_name = $cserver->name();

    # Get all vservers and append their corresponding aggr name.
    # This will be needed in order to create them
    foreach my $vs_key (keys (%{$resource})) {
        if ($vs_key =~ /^VSERVER\d*$/i) {

            $resource->{$vs_key}->{FILER} = $cserver_name;
            push(@vservers, $self->_generate_copies(element => $resource->{$vs_key},
                                                    type    => "vserver"));
        }
    }

    # if there are any vservers in the XML, fetch the component state objects for
    # all vservers
    my (@vserver_objs, @volume_objs, @lun_objs, @qtree_objs,
        @current_vserver_list, @vserver_fcp_objs, @vserver_iscsi_objs) = ();
    my (@igroup_objs, @lif_objs, @tpgroup_objs, @portset_objs) = ();

    if (scalar(@vservers)) {
        # fetch all component states
        try {
            @vserver_objs = NACL::CS::Vserver->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @vserver_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }
        # convert the objs to an array of vserver names
        @current_vserver_list = map {
            $_->vserver();
        } @vserver_objs;

        # fetch all component states
        try {
            @vserver_fcp_objs = NACL::CS::VserverFcp->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @vserver_fcp_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }

        # fetch all component states
        try {
            @vserver_iscsi_objs = NACL::CS::VserverIscsi->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @vserver_iscsi_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }

        # fetch all volumes
        try {
            @volume_objs = NACL::CS::Volume->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @volume_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }

        # fetch all qtree objects
        try {
            @qtree_objs = NACL::CS::VolumeQtree->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @qtree_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }
        # remove qtree objects that represent volumes
        @qtree_objs = grep {
            if ($_->qtree() eq "\"\"" or $_->qtree() eq "") {
                0;
            } else {
                1;
            }
        } @qtree_objs;


        # fetch all luns
        try {
            @lun_objs = NACL::CS::Lun->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @lun_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }

        try {
            @igroup_objs = NACL::CS::LunIgroup->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @igroup_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }

        try {
            @tpgroup_objs = NACL::CS::VserverIscsiTpgroup->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @tpgroup_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }

        try {
            @portset_objs = NACL::CS::LunPortset->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @portset_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }

        try {
            @lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $cserver);
        } catch NACL::Exceptions::NoElementsFound with {
            @lif_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
        };
        if ($self->error()) {
            return [];
        }

    }

    # create all the vservers sequentially
    foreach my $vserver (@vservers) {
        my $vs_name = $vserver->{NAME};

        # skip this vserver if it does not exist on the cluster
        if (not grep(/^$vs_name$/, @current_vserver_list)) {
            next;
        }

        # filter lun objs that belong to this vserver
        my @lun_objs_for_vs = grep {
            my $obj = $_;
            $obj->vserver() =~ /^$vs_name$/;
        } @lun_objs;

        # filter vol objs that belong to this vserver
        my @vol_objs_for_vs = grep {
            my $obj = $_;
            $obj->vserver() =~ /^$vs_name$/;
        } @volume_objs;

        # filter qtree objs that belong to this vserver
        my @qtree_objs_for_vs = grep {
            my $obj = $_;
            $obj->vserver() =~ /^$vs_name$/;
        } @qtree_objs;

        # filter igroup objs that belong to this vserver
        my @igroup_objs_for_vs = grep {
            my $obj = $_;
            $obj->vserver() =~ /^$vs_name$/;
        } @igroup_objs;

        # filter lif objs that belong to this vserver
        my @lif_objs_for_vs = grep {
            my $obj = $_;
            $obj->vserver() =~ /^$vs_name$/;
        } @lif_objs;

        # filter tpgroup objs that belong to this vserver
        my @tpgroup_objs_for_vs = grep {
            my $obj = $_;
            $obj->vserver() =~ /^$vs_name$/;
        } @tpgroup_objs;

        # filter portset objs that belong to this vserver
        my @portset_objs_for_vs = grep {
            my $obj = $_;
            $obj->vserver() =~ /^$vs_name$/;
        } @portset_objs;

        # determine if this vserver has iscsi / fcp services to clean up
        my @iscsi_obj = grep {
            my $obj = $_;
            $obj->vserver() =~ /^$vs_name$/;
        } @vserver_iscsi_objs;
        my @fcp_obj = grep {
            my $obj = $_;
            $obj->vserver() =~ /^$vs_name$/;
        } @vserver_fcp_objs;

        # create this vserver and push its data onto the vserver data
        # array
        my $vserver_cleanup_objs = $self->_create_cleanup_vserver(-vserver => $vserver,
                                                                  -cserver => $cserver,
                                                                  -iscsi_enabled => scalar(@iscsi_obj) ? 1 : 0,
                                                                  -fcp_enabled => scalar(@fcp_obj) ? 1 : 0,
                                                                  -volume_objs => \@vol_objs_for_vs,
                                                                  -qtree_objs => \@qtree_objs_for_vs,
                                                                  -lun_objs => \@lun_objs_for_vs,
                                                                  -igroup_objs => \@igroup_objs_for_vs,
                                                                  -portset_objs => \@portset_objs_for_vs,
                                                                  -tpgroup_objs => \@tpgroup_objs_for_vs,
                                                                  -lif_objs => \@lif_objs_for_vs);
        if ($self->error()) {
            $self->_set_err("Failed to create vserver cleanup content for $vserver");
            return [];
        }

        push(@return_data, @{$vserver_cleanup_objs});

        my $iscsi_security_cleanup_hash = {
            vserver      => $vs_name,
            reused       => $SanMgr::TRUE,
            initiator    => "eui*|iqn*",
            type         => "iscsi_security_setting",
            cleanup      => "noaggr",
        };

        my $cluster_name = $cserver->name();
        my @iscsi_clients = $self->{parent}->get_clients_associated_to_vserver(-cluster_name => $cluster_name,
                                                                               -vserver_names => [$vs_name],
                                                                               -protocols => ["iscsi", "mixed"]);

        if ($self->{parent}->error()) {
            $self->_set_err("Failed to get clients associated to vserver $vs_name", $self->{parent}->errmsg());
            $self->{parent}->_clear_err();
            return [];
        }

        foreach my $client (@iscsi_clients) {

            # if this client matches an iSCSI nodename or a FCP WWPN, then we won't be able to create a specific iSCSI setting for it
            # because we can't accurately map it to a real client. Assume that it will use the default vserver iSCSI security settings
            if ($client =~ /^(iqn\.\d+\-\d+\.\S*|[A-Fa-f0-9]{2}(?::?[A-Fa-f0-9]{2}){7})$/i) {
                next;
            }

            # set up iscsi security for this client if it is doing CHAP
            # get the SanClient object for this client
            my ($client_obj) = $self->get_objects(-types => ["client", "hypervisor"],
                                                  -names => [$client]);
            if ($self->error()) {
                $self->_set_err("Failed to get client object for $client");
                return [];
            }

            # fetch the initiator objects for this client, filter out non-iscsi initiators
            my @initiator_objs = $client_obj->get_initiators();
            @initiator_objs = grep { $_->get_protocol() =~ /iscsi/ } @initiator_objs;

            if (not scalar(@initiator_objs)) {
                $self->_set_err("There are no iSCSI initiators for the client '$client'");
                return [];
            }

            # pull out all the nodenames for this initiator
            my @nodenames = map { $_->get_nodename() } @initiator_objs;

            # remove dups from the nodename array
            my %nodename_hash;
            foreach (@nodenames) {
                $nodename_hash{$_}++;
            }
            @nodenames = keys %nodename_hash;

            foreach my $nodename (@nodenames) {
                my $iscsi_security_cleanup_hash = {
                    vserver      => $vs_name,
                    reused       => $SanMgr::TRUE,
                    initiator    => $nodename,
                    type         => "iscsi_security_setting",
                    cleanup      => "noaggr",
                };

                push(@return_data, $iscsi_security_cleanup_hash);
            }
        }

        my $global_iscsi_options = &ParamsMgr::search(search_path => [],
                                                      key => "ENVIRONMENT",
                                                      key_filter => qr/^GLOBAL_ISCSI_OPTIONS$/,
                                                      paramset => $self->EnGenX::Util::get_paramset());

        if (&ParamsMgr::error()) {
            $self->_set_err(&ParamsMgr::get_error());
            return [];
        }

        # parse out the parameter in the global iscsi options tag
        my $global_isns_address = $global_iscsi_options->{GLOBAL_ISCSI_OPTIONS}->{ISNS_SERVER};

        if (defined ($vserver->{ISNS_SERVICE}) or (defined ($global_isns_address)
            and ($global_isns_address ne ""))){

            my $isns_cleanup_hash = {
                vserver      => $vs_name,
                reused       => $SanMgr::TRUE,
                type         => "isns_service"
            };

            push(@return_data, $isns_cleanup_hash);
        }

        if ($self->error()) {
            $self->_set_err("Unable to create vserver '$vs_name'");
            return [];
        }
    }

    return \@return_data;
} ## end sub _create_cleanup_obj_vservers


###############################################################################
# Method:    _create_cleanup_vserver
# Objective: Creates cleanup obj for vserver
# Input:     A hash containing a vserver object, vserver args, cserver obj,
#            reuse policy and cleanup policy.
###############################################################################
sub _create_cleanup_vserver {
    my ($self, %args) = @_;

    my @return_data   = ();
    $self->clear_error();

    my ($name);
    my ($reused) = $SanMgr::FALSE;
    my @vserver_data = ();
    # Process args and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Hash containing objects params
    my $vserver = &ParamsMgr::get_arg(-arg      => "vserver",
                                      -required => $SanMgr::TRUE);
    # A cserver object (filer obj)
    my $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                      -required => $SanMgr::TRUE);
    my $volume_objs = &ParamsMgr::get_arg(-arg      => "volume_objs",
                                          -required => $SanMgr::TRUE);
    my $qtree_objs = &ParamsMgr::get_arg(-arg      => "qtree_objs",
                                         -required => $SanMgr::TRUE);
    my $lun_objs = &ParamsMgr::get_arg(-arg      => "lun_objs",
                                       -required => $SanMgr::TRUE);
    my $igroup_objs = &ParamsMgr::get_arg(-arg      => "igroup_objs",
                                          -required => $SanMgr::TRUE);
    my $tpgroup_objs = &ParamsMgr::get_arg(-arg      => "tpgroup_objs",
                                           -required => $SanMgr::TRUE);
    my $portset_objs = &ParamsMgr::get_arg(-arg      => "portset_objs",
                                           -required => $SanMgr::TRUE);
    my $lif_objs = &ParamsMgr::get_arg(-arg      => "lif_objs",
                                       -required => $SanMgr::TRUE);
    my $iscsi_enabled = &ParamsMgr::get_arg(-arg      => "iscsi_enabled",
                                            -required => $SanMgr::TRUE);
    my $fcp_enabled = &ParamsMgr::get_arg(-arg      => "fcp_enabled",
                                          -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return {};
    }

    my %vserver_hash = %{$vserver};

    # Start by getting the "name" of the object. If this is not defined
    # everything else will fail.
    if (defined($vserver_hash{NAME})) {
        $name = $vserver_hash{NAME};
        $vserver_hash{VSERVER} = $name;
    } else {
        $self->_set_err("Missing vserver name. Check your XML or params.");
        return [];
    }

    my $aggregate;

    if (defined($vserver_hash{AGGREGATE})) {
        $aggregate = $vserver_hash{AGGREGATE};
    } elsif ($vserver_hash{AGGR}){
        $vserver_hash{AGGREGATE} = delete $vserver_hash{AGGR};
        $aggregate = $vserver_hash{AGGREGATE};
    } else {
        $self->_set_err("Missing aggregate name. Check your XML or params.");
        return [];
    }

    # Normalize params and get local policies
    my %vserver_args = $self->_process_common_args(%vserver_hash);

    my $cleanup_policy = delete $vserver_args{cleanup_policy};
    my $reuse_policy   = delete $vserver_args{reuse_policy};

    # build the vserver cleanup create_cleanup_obj
    my $data = {
        reused    => $SanMgr::TRUE,
        name      => $name,
        aggregate => $aggregate,
        type      => "vserver",
        parallel  => $vserver->{PARALLEL},
        cleanup   => $cleanup_policy,
        cserver   => $cserver->name()
    };

    push(@return_data, $data);

    my $protocol_content = $self->_create_cleanup_obj_protocol_services(
        vserver_name   => $name,
        vserver_args   => $vserver,
        cserver        => $cserver,
        iscsi_enabled  => $iscsi_enabled,
        fcp_enabled    => $fcp_enabled,
        reuse_policy   => $reuse_policy,
        cleanup_policy => $cleanup_policy);

    push(@return_data, @{$protocol_content});

    # Setup vserver content (vols, luns, lifs, etc).
    my $vserver_content = $self->_create_cleanup_obj_vserver_content(
        vserver_name => $name,
        vserver_args => $vserver,
        aggregate    => $aggregate,
        vserver_data => $data,
        volume_objs  => $volume_objs,
        qtree_objs   => $qtree_objs,
        lun_objs     => $lun_objs,
        igroup_objs  => $igroup_objs,
        lif_objs     => $lif_objs,
        portset_objs => $portset_objs,
        tpgroup_objs => $tpgroup_objs,
        cserver      => $cserver);

    push(@return_data, @{$vserver_content});

    return \@return_data;

} ## end _create_vserver

###############################################################################
# Method:    _create_cleanup_obj_protocol_services
# Objective: Creates fcp or iscsi services on a given vserver.
# Input:     A hash containing a vserver object, vserver args, cserver obj,
#            reuse policy and cleanup policy.
# Returns:   Cleanup data.  Will _set_err flag on error.
###############################################################################
sub _create_cleanup_obj_protocol_services {
    my ($self, %args) = @_;
    my @return_data   = ();

    # Local variables
    my ($vs_obj, $vs_args, $iscsi_obj, $fcp_obj) = undef;  # Refs
    my ($vs_name)      = "";                               # Scalars
    my (@cleanup_data) = ();                               # Array of Hash
    my ($fcp_enabled, $iscsi_enabled) = $SanMgr::FALSE;    # Booleans
    my ($iscsi_cleanup_hash, $fcp_cleanup_hash) = {};      # Cleanup Data
    my ($cserver)              = undef;
    my ($cleanup_policy, $reuse_policy);
    my ($iscsi_status_admin, $fcp_status_admin) = undef;

    # Same story, process args
    &ParamsMgr::process_args(%args);

    # vserver params/attributes
    $vs_args = &ParamsMgr::get_arg(-arg      => "vserver_args",
                                   -required => $SanMgr::TRUE);

    # A cserver (filer object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Iscsi/FCP enabled
    $iscsi_enabled = &ParamsMgr::get_arg(-arg       => "iscsi_enabled",
                                         -required  => $SanMgr::FALSE,
                                         -default   => $SanMgr::FALSE);

    $fcp_enabled   = &ParamsMgr::get_arg(-arg       => "fcp_enabled",
                                         -required  => $SanMgr::FALSE,
                                         -default   => $SanMgr::FALSE);

    # Cleanup Policy
    $cleanup_policy = &ParamsMgr::get_arg(-arg      => "cleanup_policy",
                                          -required => $SanMgr::TRUE);

    # Reuse Policy
    $reuse_policy   = &ParamsMgr::get_arg(-arg      => "reuse_policy",
                                          -required => $SanMgr::TRUE);

    # Vserver name
    $vs_name        = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                          -required => $SanMgr::TRUE);

    # Check to make sure all the arguments were properly processed
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    if ($iscsi_enabled){

        # Create the cleanup stack entry
        $iscsi_cleanup_hash = {
            reused    => $SanMgr::TRUE,
            vserver   => $vs_name,
            type      => "iscsi_service",
            cleanup   => $cleanup_policy,
            cserver   => $cserver->name()
        };

        # Push the iscsi create_cleanup_obj data on the cleanup stack then
        # create_cleanup_obj FCP
        # and push the fcp cleanup data on that same stack and return it.
        push(@return_data, $iscsi_cleanup_hash);
    }

    # Do the same thing for FCP that we did for ISCSI
    if ($fcp_enabled) {

        # Create the cleanup stack entry
        $fcp_cleanup_hash = {
            reused    => $SanMgr::TRUE,
            vserver   => $vs_name,
            type      => "fcp_service",
            cleanup   => $cleanup_policy,
            cserver   => $cserver->name()
        };

        # Push the fcp create_cleanup_obj data on the cleanup stack
        push(@return_data, $fcp_cleanup_hash);
    }

    return \@return_data;

} ## end sub _create_cleanup_obj_protocol_services

###############################################################################
# Method:    _create_cleanup_obj_vserver_content
# Objective: Responsible for calling methods to create the contents of a
#            vserver such as vols, igroups, lifs, etc.
# Input:     A hash containing a vserver object, a cserver object, vserver data
# Returns:   Array reference containing hashes with vserver content data
###############################################################################
sub _create_cleanup_obj_vserver_content {
    my ($self, %args) = @_;
    my @return_data   = ();

    $self->clear_error();

    # Locals
    my ($cserver, $vserver_obj, $data) = undef;
    my ($vserver_name, $aggregate) = "";
    my (@vserver_content, @volumes, @luns, @lifs, @portsets) = ();
    my ($vserver_data, $lif_data, $igroup_data, $volume_data, $tpgroup_objs, $portset_objs) = undef;
    my ($portset_data, $tpgroup_data, $volume_objs, $lun_objs, $qtree_objs, $lif_objs, $igroup_objs) = undef;
    my (@tpgroups, @igroups) = ();

    # Process and validate args
    &ParamsMgr::process_args(%args);

    # The vserver where all objects will be creates
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                       -required => $SanMgr::TRUE);

    # A data structure containing Vserver params and content
    $vserver_data = &ParamsMgr::get_arg(-arg      => "vserver_args",
                                        -required => $SanMgr::TRUE);

    # A cserver (filer object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # A hash containing results from vserver create_cleanup_obj
    $data = &ParamsMgr::get_arg(-arg      => "vserver_data",
                                -required => $SanMgr::TRUE);

    # Name of the aggregate
    $aggregate = &ParamsMgr::get_arg(-arg      => "aggregate",
                                     -required => $SanMgr::TRUE);

    # vol objs
    $volume_objs = &ParamsMgr::get_arg(-arg      => "volume_objs",
                                       -required => $SanMgr::TRUE);

    # qtree objs
    $qtree_objs = &ParamsMgr::get_arg(-arg      => "qtree_objs",
                                      -required => $SanMgr::TRUE);

    # lun objs
    $lun_objs = &ParamsMgr::get_arg(-arg      => "lun_objs",
                                    -required => $SanMgr::TRUE);

    $tpgroup_objs = &ParamsMgr::get_arg(-arg      => "tpgroup_objs",
                                        -required => $SanMgr::TRUE);

    $portset_objs = &ParamsMgr::get_arg(-arg      => "portset_objs",
                                        -required => $SanMgr::TRUE);

    # igroup objs
    $igroup_objs = &ParamsMgr::get_arg(-arg      => "igroup_objs",
                                    -required => $SanMgr::TRUE);

    # lif objs
    $lif_objs = &ParamsMgr::get_arg(-arg      => "lif_objs",
                                    -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Look for vservers inside each aggregate and put them on separate lists.
    foreach my $key (keys (%{$vserver_data})) {
        if ($key =~ /^VOL\d*$/i) {
            push(@volumes,
                 $self->_generate_copies(element => $vserver_data->{$key},
                                         type    => "vol"));
        } elsif ($key =~ /^LIF\d*$/i) {
            # NOTE: specifying type to tell generate copies to
            # assign different IPs to each element (iSCSI)
            push(@lifs,
                 $self->_generate_copies('element'     => $vserver_data->{$key},
                                         'parent_name' => $vserver_name,
                                         'type'        => "LIF",
                                        )
                 );
            if ($self->error()) {
                $self->_set_err("Failed to generate copy for lif elements");
                return [];
            }
        } elsif ($key =~ /^IGROUP\d*$/i) {
            push(@igroups,
                 $self->_generate_copies(element => $vserver_data->{$key},
                                         type    => "igroup"));
        } elsif ($key =~ /^PORTSET\d*$/i) {
            push(@portsets,
                 $self->_generate_copies(element => $vserver_data->{$key},
                                         type    => "portset"));
        } elsif ($key =~ /^TPGROUP\d*$/i) {
            push(@tpgroups,
                 $self->_generate_copies(element => $vserver_data->{$key},
                                         type    => "tpgroup"));
        }
    }

    # Start with lifs
    if (scalar(@lifs) and scalar(@{$lif_objs})) {
        $lif_data = $self->_create_cleanup_obj_lifs(vserver_name => $vserver_name,
                                                    lifs => \@lifs,
                                                    lif_objs => $lif_objs,
                                                    cserver => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to create_cleanup_obj lifs");
            return [];
        }
    }

    # Then iGroups
    if (scalar(@igroups) and scalar(@{$igroup_objs})) {
       $igroup_data = $self->_create_cleanup_obj_igroups(vserver_name => $vserver_name,
                                                         igroups => \@igroups,
                                                         igroup_objs => $igroup_objs,
                                                         cserver => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to create_cleanup_obj igroups");
            return [];
        }
    }

    # Then portsets
    if (scalar(@portsets) and scalar(@{$portset_objs})) {
        $portset_data = $self->_create_cleanup_obj_portsets(vserver_name => $vserver_name,
                                                            portsets => \@portsets,
                                                            portset_objs => $portset_objs,
                                                            cserver => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to create_cleanup_obj portsets");
            return [];
        }
    }

    # Then tpgroups
    if (scalar(@tpgroups) and scalar(@{$tpgroup_objs})) {
        $tpgroup_data = $self->_create_cleanup_obj_tpgroups(vserver_name => $vserver_name,
                                                            tpgroups => \@tpgroups,
                                                            tpgroup_objs => $tpgroup_objs,
                                                            cserver => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to create_cleanup_obj tpgroups");
            return [];
        }
    }

    # Create volumes (if there are volumes to cleanup for this vs)
    if (scalar(@volumes) and scalar(@{$volume_objs})) {
        $volume_data = $self->_create_cleanup_obj_volumes(vserver_name => $vserver_name,
                                                          aggregate    => $aggregate,
                                                          volumes      => \@volumes,
                                                          volume_objs  => $volume_objs,
                                                          qtree_objs   => $qtree_objs,
                                                          lun_objs     => $lun_objs,
                                                          cserver      => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to create_cleanup_obj volumes");
            return [];
        }
    }

    if (defined ($lif_data)) {
        push(@return_data, @{$lif_data});
    }
    if (defined ($igroup_data)) {
        push(@return_data, @{$igroup_data});
    }
    if (defined ($portset_data)) {
        push(@return_data, @{$portset_data});
    }
    if (defined ($tpgroup_data)) {
        push(@return_data, @{$tpgroup_data});
    }
    if (defined ($volume_data)) {
        push(@return_data, @{$volume_data});
    }

    return \@return_data;
} ## end sub _create_cleanup_obj_vserver_content

###############################################################################
#                           L I F   M E T H O D S
#
#  Things to consider:
#  1. Lifs are created serially. So there's no real need for having plural,
#     singular methods.
#  2. Currently, our way of assigning IPs is not clever at all!
#     Need a better way to handle duplicate ip addresses for ISCSI Lifs
#
###############################################################################

###############################################################################
# Method:    _create_cleanup_obj_lifs
# Objective: Method responsible for creating or reusing lif objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     lif_data => hash containing all args for a lif
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_lifs {
    my ($self, %args) = @_;
    my @return_data    = ();

    $self->clear_error();

    # Local Variables
    my ($cserver, $lifs, $data, $vserver_obj, $lif_obj, $lif_objs) = undef;
    my ($vserver_name, $name, $gateway, $reuse_policy, $cleanup_policy,
        $protocol) = "";
    my ($lif_exists, $reused) =  $SanMgr::FALSE;
    my (@lif_data) = ();

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing lif details
    $lifs = &ParamsMgr::get_arg(-arg      => "lifs",
                                -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);
    $lif_objs = &ParamsMgr::get_arg(-arg      => "lif_objs",
                                    -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # convert lif objs to lif names
    my @lif_names = map {
        $_->lif();
    } @{$lif_objs};

    # Get lif data from the resource data structure
    foreach my $lif (@{$lifs}) {

        # Normalize params and get local policies
        my %lif_args    = $self->_process_common_args(%$lif);
        $cleanup_policy = delete $lif_args{cleanup_policy};
        $reuse_policy   = delete $lif_args{reuse_policy};
        $gateway        = delete $lif_args{gateway};

        # Get lif name
        if (defined($lif_args{name})) {
            $name          = $lif_args{name};
            $lif_args{lif} = delete $lif_args{name};
        } else {
            $self->_set_err("Lif name is missing!");
            return [];
        }

        # Get protocol name
        if (defined($lif_args{'data-protocol'})) {
            $protocol = $lif_args{'data-protocol'};
        }

        # skip if this lif does not exist on the cluster
        next if (not grep(/^$name$/, @lif_names));

        foreach my $key (keys(%lif_args)) {
            if ($key =~ /^ADDRESS_RANGE($|[1-9]$)/i) {
                delete $lif_args{$key};
            }
        }

        # Store the LIF attributes, either FCP or iSCSI, in the vserver object
        # This is needed in order to load fcp or iscsi targets using the right
        # data.

        %lif_args = (%lif_args, vserver => $vserver_name);
        my (%fcp_targets, %iscsi_targets);

        if ($protocol =~ /^fcp$/i) {
            # force copy the hash
            %fcp_targets = %lif_args;
            push (@{$vserver_obj->{fcp_lif_data}}, \%fcp_targets);
        } elsif ($protocol =~ /^iscsi$/i) {
            %iscsi_targets = %lif_args;
            push (@{$vserver_obj->{iscsi_lif_data}}, \%iscsi_targets);
        }

        $data = {
            reused        => $SanMgr::TRUE,
            fcp_targets   => \%fcp_targets,
            iscsi_targets => \%iscsi_targets,
            name          => $name,
            vserver       => $vserver_name,
            type          => "lif",
            cleanup       => $cleanup_policy
        };
        push(@return_data, $data);
    }

    return \@return_data;
} ## end sub _create_cleanup_obj_lifs

###############################################################################
#                         I G R O U P   M E T H O D S
#
#  Things to consider:
#  1. Igroups are created serially. So there's no real need for having plural,
#     singular methods. They are created after lifs.
#  2. TODO: Handle igroup options.
#
###############################################################################

###############################################################################
# Method:    _create_cleanup_obj_igroups
# Objective: Method responsible for creating or reusing igroup objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     igroup_data => hash containing all args for an igroup
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_igroups {
    my ($self, %args) = @_;
    my @return_data   = ();
    $self->clear_error();

    # Local Variables
    my ($cserver, $igroups, $data, $igroup_obj, $igroup_objs) = undef;
    my ($vserver_name, $name, $reuse_policy, $cleanup_policy) = "";
    my ($reused) =  $SanMgr::FALSE;
    my (@igroup_data, %igroup_args) = ();
    my ($alua_state) = ();

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing igroup details
    $igroups = &ParamsMgr::get_arg(-arg      => "igroups",
                                   -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    $igroup_objs = &ParamsMgr::get_arg(-arg      => "igroup_objs",
                                       -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # convert igroup objs to igroup names
    my @igroup_names = map {
        $_->igroup();
    } @{$igroup_objs};

    # Get igroup data from the resource data structure
    foreach my $igroup (@{$igroups}) {

        # Normalize params and get local policies
        %igroup_args    = $self->_process_common_args(%$igroup);
        $cleanup_policy = delete $igroup_args{cleanup_policy};
        $reuse_policy   = delete $igroup_args{reuse_policy};
        $alua_state     = delete $igroup_args{alua};

        my (@inits, @options) = ();

        # Get igroup name
        if (defined($igroup_args{name})) {
            $name = $igroup_args{name};
            $igroup_args{igroup} = delete $igroup_args{name};
        } else {
            $self->_set_err("Igroup name is missing!");
            return [];
        }

        # skip this igroup if it doesnt exist
        next if (not grep(/^$name$/, @igroup_names));

        $data = {
            reused    => $SanMgr::TRUE,
            name      => $name,
            initiator => \@inits,
            options   => \@options,
            vserver   => $vserver_name,
            type      => "igroup",
            cleanup   => $cleanup_policy
        };

        push (@return_data, $data);
    } # end foreach igroup

    return \@return_data;

} ## end sub _create_cleanup_obj_igroups

###############################################################################
#               P O R T S E T  &  T P G R O U P S  M E T H O D S
#
#  Things to consider:
#  1. Portsets are created serially, once all igroups have been created.
#  2. Binding to igroups is done after creation
#  3. For tpgroups, iSCSI service has to be started.
#
###############################################################################

###############################################################################
# Method:    _create_cleanup_obj_portsets
# Objective: Method responsible for creating or reusing portset objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     portset_data => hash containing all args for a portset
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_portsets {
    my ($self, %args) = @_;
    my @return_data   = ();

    $self->clear_error();

    # Local Variables
    my ($cserver, $portsets, $data, $vserver_obj, $portset_obj) = undef;
    my ($vserver_name, $name, $reuse_policy, $cleanup_policy, $igroup_name) = "";
    my ($portset_exists, $reused) =  $SanMgr::FALSE;
    my (@portset_data, %portset_args) = ();
    my ($igroup_obj, $igroup) = undef;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing portset details
    $portsets = &ParamsMgr::get_arg(-arg     => "portsets",
                                   -required => $SanMgr::TRUE);

    my $portset_objs = &ParamsMgr::get_arg(-arg     => "portset_objs",
                                           -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed",
                        &ParamsMgr::get_error());
        return [];
    }

    # convert portset objs to portset names
    my @portset_names = map {
        $_->portset();
    } @{$portset_objs};

    # Get portset data from the resource data structure
    foreach my $portset (@{$portsets}) {
        # Normalize params and get local policies
        %portset_args   = $self->_process_common_args(%$portset);
        $cleanup_policy = delete $portset_args{cleanup_policy};
        $reuse_policy   = delete $portset_args{reuse_policy};

        # Get portset name
        if (defined($portset_args{name})) {
            $name = $portset_args{name};
            $portset_args{portset} = delete $portset_args{name};
        } else {
            $self->_set_err("Portset name is missing!");
            return [];
        }

        # skip this portset if it doesnt exist
        next if (not grep(/^$name$/, @portset_names));

        # Remove igroups from hash. Binding is done later.
        foreach my $key (keys(%portset_args)) {
            if ($key =~ /igroup_bind/i) {
                $igroup = delete $portset_args{igroup_bind};
                $igroup_name = $igroup->{NAME};
            }
        }

        $data = {
            reused  => $SanMgr::TRUE,
            name    => $name,
            igroup  => $igroup_name,
            vserver => $vserver_name,
            type    => "portset",
            cleanup => $cleanup_policy
        };

        push(@return_data, $data);
    }

    return \@return_data;
} ## end sub _create_cleanup_obj_portsets

###############################################################################
# Method:    _create_cleanup_obj_tpgroups
# Objective: Method responsible for creating or reusing tpgroup objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     tpgroupdata => hash containing all args for a tpgroup
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_tpgroups { #TODO
    my ($self, %args) = @_;
    my @return_data   = ();

    $self->clear_error();

    # Local Variables
    my ($cserver, $tpgroups, $data, $vserver_obj, $tpgroup_obj) = undef;
    my ($vserver_name, $name, $reuse_policy, $cleanup_policy) = "";
    my ($tpgroup_exists, $reused) =  $SanMgr::FALSE;
    my (@tpgroup_data, %tpgroup_args) = ();
    my ($iscsi_obj) = undef;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing tpgroup details
    $tpgroups = &ParamsMgr::get_arg(-arg      => "tpgroups",
                                    -required => $SanMgr::TRUE);

    my $tpgroup_objs = &ParamsMgr::get_arg(-arg     => "tpgroup_objs",
                                           -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed",
                        &ParamsMgr::get_error());
        return [];
    }

    # convert tpgroup objs to tpgroup names
    my @tpgroup_names = map {
        $_->tpgroup();
    } @{$tpgroup_objs};

    # Get tpgroup data from the resource data structure
    foreach my $tpgroup (@{$tpgroups}) {
        # Normalize params and get local policies
        %tpgroup_args = $self->_process_common_args(%$tpgroup);
        $cleanup_policy = delete $tpgroup_args{cleanup_policy};
        $reuse_policy = delete $tpgroup_args{reuse_policy};

        # Get tpgroup name
        if (defined($tpgroup_args{name})) {
            $name = $tpgroup_args{name};
            $tpgroup_args{tpgroup} = delete $tpgroup_args{name};
        } else {
            $self->_set_err("Tpgroup name is missing!");
            return [];
        }

        # skip this tpgroup if it doesnt exist
        next if (not grep(/^$name$/, @tpgroup_names));

        $data = {
            reused  => $SanMgr::TRUE,
            name    => $name,
            vserver => $vserver_name,
            type    => "tpgroup",
            cleanup => $cleanup_policy
        };

        push(@return_data, $data);

    } # end foreach tpgroup

    return \@return_data;
} ## end sub _create_cleanup_obj_tpgroups

###############################################################################
#                    VO L U M E S  &  L U N S  M E T H O D S
#
#  Things to consider:
#  1. Volumes are created in parallel by default, unless users disable this
#     mode in the XML. Each vol subtest creates luns serially.
#  2. LUNs will be created serially to avoid creating a large number or logs
#
###############################################################################

###############################################################################
# Method:    _create_cleanup_obj_volumes
# Objective: Method responsible for creating or reusing volume objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     volume_data => hash containing all args for a volume
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _create_cleanup_obj_volumes {
    my ($self, %args) = @_;
    my @return_data   = ();

    $self->clear_error();

    # Local Variables
    my ($vserver, $cserver, $volumes, $vol_obj, $volume_objs, $lun_objs, $qtree_objs) = undef;
    my ($cserver_name, $vserver_name, $aggregate) = "";
    my (@volume_data, @subtests, @messages) = ();
    my ($parallel, $st_status) = $SanMgr::TRUE;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing volume details
    $volumes = &ParamsMgr::get_arg(-arg      => "volumes",
                                   -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    # Name of the aggregate
    $aggregate = &ParamsMgr::get_arg(-arg      => "aggregate",
                                     -required => $SanMgr::TRUE);

    # vol objs
    $volume_objs = &ParamsMgr::get_arg(-arg      => "volume_objs",
                                       -required => $SanMgr::TRUE);

    # lun objs
    $lun_objs = &ParamsMgr::get_arg(-arg      => "lun_objs",
                                    -required => $SanMgr::TRUE);
    # qtree objs
    $qtree_objs = &ParamsMgr::get_arg(-arg      => "qtree_objs",
                                      -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Get the cserver name which is needed to create a volume
    $cserver_name = $cserver->name();

    # convert the volume objects into an array of vol names
    my @current_volumes = map {
        $_->volume();
    } @{$volume_objs};

    # Go through all volumes and create them. By default, SanMgr will create
    # them in parallel.
    foreach my $volume (@{$volumes}) {
        my $volume_name = $volume->{NAME};

        # skip this volume if it does not currently exist on the system
        if (not grep(/^$volume_name$/, @current_volumes)) {
            next;
        }

        # Get all the luns, qtrees and options contained on each volume
        my @luns    = ();
        my @qtrees  = ();
        my @options = ();
        my @efficiency_options = ();

        foreach my $key (keys(%{$volume})) {
            # We'll need a volume name later, piggy back it here
            if ($key =~ /^LUN\d*$/i) {
                $volume->{$key}->{VOLUME} = $volume->{NAME};
                push(@luns,
                     $self->_generate_copies(element => $volume->{$key},
                                             type    => "lun"));
            } elsif ($key =~ /^QTREE\d*$/i) {
                $volume->{$key}->{VOLUME} = $volume->{NAME};
                push(@qtrees,
                     $self->_generate_copies(element => $volume->{$key},
                                             type    => "qtree"));
            } elsif ($key =~ /^OPTIONS$/i) {
                push(@options,
                     $self->_generate_copies(element => $volume->{$key},
                                             type    => "options"));
            } elsif ($key =~ /^EFFICIENCY_OPTIONS$/i) {
                push(@efficiency_options,
                     $self->_generate_copies(element => $volume->{$key},
                                             type    => "efficiency_options"));
            }
        }

        # find all the lun objects that are currently in this volume
        my @lun_objs_currently_in_vol = grep {
            $volume_name eq $_->volume();
        } @{$lun_objs};

        # find all the qtree objects that are currently in this volume
        my @qtree_objs_currently_in_vol = grep {
            $volume_name eq $_->volume();
        } @{$qtree_objs};

        # Append the aggregate name needed to create the object
        $volume->{AGGREGATE} = $aggregate;

        my $volume_data = $self->_create_cleanup_obj_volume(
                                     cserver              => $cserver,
                                     vserver_name         => $vserver_name,
                                     volume               => $volume,
                                     vol_name             => $volume->{NAME},
                                     aggregate            => $aggregate,
                                     lun_objs             => \@lun_objs_currently_in_vol,
                                     qtree_objs           => \@qtree_objs_currently_in_vol,
                                     luns                 => \@luns,
                                     qtrees               => \@qtrees,
                                     efficiency_options   => \@efficiency_options,
                                     options              => \@options);


        push (@return_data, @{$volume_data});
    } # end foreach volume

    return \@return_data;

} ## end sub _create_cleanup_obj_volumes

###############################################################################
# Method:    _create_cleanup_obj_volume
# Objective: Method responsible for creating or reusing volume objects.
# Inputs:    A hash containing cserver name, volume data, aggregate name,
#            vserver name, subtested flag, luns and volume options.
# Returns:   An array of hashes containing volume data.
###############################################################################
sub _create_cleanup_obj_volume {

    my ($self, %args) = @_;
    my (@return_data) = ();

    $self->clear_error();

    # Locals
    my ($cserver, $vserver_obj, $data, $volume_name, $volume, $lun_objs, $qtree_objs) = undef;
    my ($name, $vserver_name, $aggregate, $percent_snapshot_space) = "";
    my ($cleanup_policy, $reuse_policy) = "";
    my (%volume_args, %create_args, @volume_data) = ();
    my ($volume_exists, $reused) =  $SanMgr::FALSE;
    my ($snapshot_policy) = "";
    my ($luns, $options, $lun_data, $efficiency_options, $qtrees, $qtree_data) = undef;
    my ($autosize_percent, $autosize) = undef;
    my ($enable_efficiency, $size, $max_autosize,$autosize_percent_used) = 0;

    # Process args and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Name of vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    # A cserver object (filer obj)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Volume
    $volume = &ParamsMgr::get_arg(-arg      => "volume",
                                  -required => $SanMgr::TRUE);

    # Volume name
    $volume_name = &ParamsMgr::get_arg(-arg      => "vol_name",
                                  -required => $SanMgr::TRUE);

    # Luns on this volume
    $luns = &ParamsMgr::get_arg(-arg      => "luns",
                                -required => $SanMgr::FALSE,
                                -default  => []);

    # lun objects that currently exist in this volume
    $lun_objs = &ParamsMgr::get_arg(-arg      => "lun_objs",
                                    -required => $SanMgr::TRUE);

    # Qtrees on this volume
    $qtrees = &ParamsMgr::get_arg(-arg      => "qtrees",
                                -required => $SanMgr::FALSE,
                                -default  => []);

    # qtree objects that currently exist in this volume
    $qtree_objs = &ParamsMgr::get_arg(-arg      => "qtree_objs",
                                      -required => $SanMgr::TRUE);

    # efficiency_options on this volume
    $efficiency_options = &ParamsMgr::get_arg(
                                -arg      => "efficiency_options",
                                -required => $SanMgr::FALSE,
                                -default  => []);

    # Volume Options
    $options = &ParamsMgr::get_arg(-arg      => "options",
                                   -required => $SanMgr::FALSE,
                                   -default  => []);

    # Name of the aggregate
    $aggregate = &ParamsMgr::get_arg(-arg      => "aggregate",
                                     -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }


    # Start by getting the "name" of the object. If this is not defined
    # everything else will fail.

    if (defined($volume->{NAME})) {
        $name = delete $volume->{NAME};
        $volume->{VOLUME} = $name;
    } else {
        $self->_set_err("Missing volume name. Check your XML or params.");
        return [];
    }

    # Append the vserver name. Needed by the library to create a volume
    $volume->{VSERVER} = $vserver_name;

    # Normalize params and get local policies
    %volume_args    = $self->_process_common_args(%$volume);
    $cleanup_policy = delete $volume_args{cleanup_policy};
    $reuse_policy   = delete $volume_args{reuse_policy};

    # NOTE: appending aggregate name because it's needed for cleanup purposes
    $data = {
        reused    => $SanMgr::TRUE,
        name      => $name,
        aggregate => $aggregate,
        vserver   => $vserver_name,
        type      => "volume",
        parallel  => $volume->{PARALLEL},
        cleanup   => $cleanup_policy
    };
    push(@return_data, $data);

    # check if there are qtrees inside this volume (and their respective luns) to cleanup
    $qtree_data = $self->_create_cleanup_obj_qtrees(vserver_name => $vserver_name,
                                                    cserver      => $cserver,
                                                    qtrees       => $qtrees,
                                                    qtree_objs   => $qtree_objs,
                                                    lun_objs     => $lun_objs,
                                                    volume_name  => $volume_name);
    if ($self->error()) {
        $self->_set_err("Unable to create_cleanup_obj qtrees for vserver '$vserver_name' for volume '$name'");
        return [];
    }

    # if we found qtrees, push them on the stack
    if (scalar(@{$qtree_data})) {
        push(@return_data, @{$qtree_data});
    }

    # check if there are luns inside this volume (not in qtrees) to cleanup
    $lun_data = $self->_create_cleanup_obj_luns(volume_name  => $volume_name,
                                                luns         => $luns,
                                                lun_objs     => $lun_objs,
                                                vserver_name => $vserver_name,
                                                cserver      => $cserver);
    if ($self->error()) {
        $self->_set_err("Unable to create_cleanup_obj luns for vserver '$vserver_name' for volume '$name'");
        return [];
    }
    if (scalar(@{$lun_data})) {
        push(@return_data, @{$lun_data});
    }

    return \@return_data;
} ## end sub _create_cleanup_obj_volume

###############################################################################
# Method:    _create_cleanup_obj_qtrees
# Objective: Method responsible for creating or reusing qtree objects.
# Inputs:    A hash containing cserver name, volume data, aggregate name,
#            vserver name, luns and volume options.
# Returns:   An array of hashes containing qtree data.
###############################################################################
sub _create_cleanup_obj_qtrees {
    my ($self, %args) = @_;
    my @return_data   = ();

    $self->clear_error();

    # Local Variables
    my ($vserver, $cserver, $volume_name, $qtrees, $qtree_objs, $lun_objs) = undef;
    my ($cserver_name, $vserver_name)         = "";
    my (@qtree_data)    = ();
    my ($parallel, $st_status) = $SanMgr::TRUE;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing volume details
    $qtrees = &ParamsMgr::get_arg(-arg      => "qtrees",
                                  -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    # Volume Object
    $volume_name = &ParamsMgr::get_arg(-arg      => "volume_name",
                                       -required => $SanMgr::TRUE);
    $qtree_objs = &ParamsMgr::get_arg(-arg      => "qtree_objs",
                                      -required => $SanMgr::TRUE);
    $lun_objs = &ParamsMgr::get_arg(-arg      => "lun_objs",
                                    -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Get the cserver name which is needed to create a qtree
    $cserver_name = $cserver->name();

    # convert the qtree objects into an array of vol names
    my @current_qtrees = map {
        $_->qtree();
    } @{$qtree_objs};

    # Go through all qtrees and create them. By default, SanMgr will create
    # them in parallel.
    foreach my $qtree (@{$qtrees}) {
        my $qtree_name = $qtree->{NAME};

        # skip this qtree if it does not currently exist on the system
        if (not grep(/^$qtree_name$/, @current_qtrees)) {
            next;
        }

        # Get all the luns contained in each qtree
        my @luns    = ();
        foreach my $key (keys(%{$qtree})) {
            # We'll need a qtree and volume name later, piggy back it here
            if ($key =~ /^LUN\d*$/i) {
                $qtree->{$key}->{VOLUME} = $volume_name;
                $qtree->{$key}->{QTREE}  = $qtree->{NAME};
                push(@luns,
                     $self->_generate_copies(element => $qtree->{$key},
                                             type    => "lun"));
            }
        }

        # filter out all luns that are not in this qtree
        my @luns_in_this_qtree = grep {
            $qtree_name eq $_->qtree();
        } @{$lun_objs};

        my $qtree_datum = $self->_create_cleanup_obj_qtree(
            cserver              => $cserver,
            vserver_name         => $vserver_name,
            volume_name          => $volume_name,
            qtree                => $qtree,
            lun_objs             => \@luns_in_this_qtree,
            luns                 => \@luns);

        push(@return_data, @{$qtree_datum});
    }

    return \@return_data;

} ## end sub _create_cleanup_obj_qtrees

###############################################################################
# Method:    _create_cleanup_obj_qtree
# Objective: Method responsible for creating or reusing volume objects.
# Inputs:    A hash containing cserver name, volume data, aggregate name,
#            vserver name, subtested flag, luns and volume options.
# Returns:   An array of hashes containing volume data.
###############################################################################
sub _create_cleanup_obj_qtree {
    my ($self, %args) = @_;
    my @return_data   = ();

    $self->clear_error();

    # Locals
    my ($cserver, $vserver_obj, $data, $qtree_obj, $qtree, $volume) = undef;
    my ($name, $vserver_name, $volume_name, $percent_snapshot_space) = "";
    my ($cleanup_policy, $reuse_policy) = "";
    my (%qtree_args, %create_args, @qtree_data, @qtree_objs) = ();
    my ($qtree_exists, $reused) =  $SanMgr::FALSE;
    my ($luns, $options, $lun_objs) = undef;

    # Process args and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Name of vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    # A cserver object (filer obj)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Qtree
    $qtree = &ParamsMgr::get_arg(-arg      => "qtree",
                                 -required => $SanMgr::TRUE);

    # Luns on this qtree
    $luns = &ParamsMgr::get_arg(-arg      => "luns",
                                -required => $SanMgr::FALSE,
                                -default  => []);

    $lun_objs = &ParamsMgr::get_arg(-arg      => "lun_objs",
                                    -required => $SanMgr::TRUE);

    # Volume object
    $volume_name = &ParamsMgr::get_arg(-arg      => "volume_name",
                                       -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Start by getting the "name" of the object. If this is not defined
    # everything else will fail.

    if (defined($qtree->{NAME})) {
        $name = delete $qtree->{NAME};
        $qtree->{QTREE} = $name;
    } else {
        $self->_set_err("Missing qtree name. Check your XML or params.");
        return $SanMgr::FALSE;
    }

    # Append the vserver name. Needed by the library to create a qtree
    $qtree->{VSERVER} = $vserver_name;

    # Normalize params and get local policies
    %qtree_args     = $self->_process_common_args(%$qtree);
    $cleanup_policy = delete $qtree_args{cleanup_policy};
    $reuse_policy   = delete $qtree_args{reuse_policy};

    # NOTE: appending aggregate name because it's needed for cleanup purposes
    $data = {
        reused    => $SanMgr::TRUE,
        name      => $name,
        volume    => $volume_name,
        vserver   => $vserver_name,
        type      => "qtree",
        parallel  => $qtree->{PARALLEL},
        cleanup   => $cleanup_policy
    };

    push(@return_data, $data);

    my $lun_data = $self->_create_cleanup_obj_luns(cserver      => $cserver,
                                                   volume_name  => $volume_name,
                                                   luns         => $luns,
                                                   lun_objs     => $lun_objs,
                                                   qtree_name   => $name,
                                                   vserver_name => $vserver_name);

    if ($self->error()) {
        $self->_set_err("Unable to create_cleanup_obj luns for vserver"
                        . " '$vserver_name' for qtree '$name'");
        return $SanMgr::FALSE;
    }

    if (scalar(@{$lun_data})) {
        push(@return_data, @{$lun_data});
    }

    return \@return_data;

} ## end sub _create_cleanup_obj_qtree

###############################################################################
# Method:    _create_cleanup_obj_luns
# Objective: Method responsible for creating or reusing lun objects.
# Inputs:    A hash containing cserver, volume name, vserver name, lun data.
# Returns:   An array of hashes containing volume data.
###############################################################################
sub _create_cleanup_obj_luns {
    my ($self, %args) = @_;

    my (@return_data) = ();

    $self->clear_error();

    # Local Variables
    my ($cserver, $luns, $data, $lun_obj, $allocation_req, $class, $bind_list) = undef;
    my ($qtree_obj, $parent_obj, $lun_objs) = undef;
    my ($vserver_name, $name, $volume_name, $qtree_name, $current_interface) = "";
    my ($reuse_policy, $cleanup_policy) = "";
    my ($reserve_req, $reserve_state, $allocation_state, $state_req) = "";
    my (@lun_data, %lun_args) = ();
    my ($space_alloc) =  $SanMgr::TRUE;
    my ($lun_exists, $reused) =  $SanMgr::FALSE;
    my ($parent_is_qtree) = $SanMgr::FALSE;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing luns details
    $luns = &ParamsMgr::get_arg(-arg      => "luns",
                                -required => $SanMgr::TRUE);

    # Volume name
    $volume_name = &ParamsMgr::get_arg(-arg     => "volume_name",
                                      -required => $SanMgr::TRUE);

    # Vserver name
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    # Optional qtree name
    $qtree_name = &ParamsMgr::get_arg(-arg      => "qtree_name",
                                      -required => $SanMgr::FALSE,
                                      -default  => "\"\"");

    $lun_objs = &ParamsMgr::get_arg(-arg      => "lun_objs",
                                    -required => $SanMgr::TRUE);

    # Optioinal cserver object (filer obj)
    # Qtree will send this, volume will not
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::FALSE,
                                   -default  => "");

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    # Make sure we have luns to create before we go any further
    if (not scalar(@{$luns})) {
        return [];
    }

    # remove quotes from the qtree name
    $qtree_name =~ s/\"//g;

    # Get lun data from the resource data structure
    foreach my $lun (@{$luns}) {
        my $lun_name = $lun->{NAME};

        # skip this lun if it does not currently exist on the cluster
        my $matched = 0;
        foreach my $lun_obj (@{$lun_objs}) {
            if ($lun_obj->lun() eq $lun_name) {
                if ($lun_obj->qtree() =~ /\"\"/ and $qtree_name eq "" or
                    $lun_obj->qtree() eq $qtree_name) {
                    $matched = 1;
                    last;
                } else {
                    next;
                }
            } else {
                next;
            }
        }

        # Normalize params and get local policies
        %lun_args       = $self->_process_common_args(%$lun);
        $cleanup_policy = delete $lun_args{cleanup_policy};
        $reuse_policy   = delete $lun_args{reuse_policy};
        $state_req      = (defined($lun_args{'state'})) ?
                             delete $lun_args{'state'} : "online";
        $allocation_req = (defined($lun_args{'space-allocation'})) ?
                             delete $lun_args{'space-allocation'} : undef;

        # Get lun name
        if (defined($lun_args{name})) {
            $name = $lun_args{name};
            $lun_args{lun} = delete $lun_args{name};
        } else {
            $self->_set_err("Lun name is missing!");
            return [];
        }

        my @lun_maps = ();

        # Get lun maps and remove them from the params
        foreach my $key (keys(%lun_args)) {
            if ($key =~ /^LUN_MAP\d*$/i) {
                my $lun_map = delete $lun_args{$key};
                if (defined($lun_map->{IGROUP})) {
                    push(@lun_maps, $lun_map->{IGROUP});
                }
            } elsif ($key =~ /^MAP_LIST$/i) {
                my $lun_map_list = delete $lun_args{$key};

                # split the lun map list by commans, and add each entry into the lun_maps array
                my @split_lun_map_list = split(/,/, $lun_map_list);
                push(@lun_maps, @split_lun_map_list);
            } elsif ($key =~ /^BIND_LIST$/i) {
                my $lun_bind_list = delete $lun_args{$key};
                my @lun_binds = split(/,/, $lun_bind_list);

                # convert all the lun binds into their absolute paths
                @lun_binds = map {
                    my $pep = $_;
                    # get rid of spaces
                    $pep =~ s/\s+//g;
                    # remove a starting ./ from the pep
                    $pep =~ s/^\.\///;

                    if ($pep !~ /^\/vol/) {
                        # $pep is a relative path to a lun
                        # we need to add the vol name to the front of the pe lun
                        # check if it has a qtree in it
                        if (not length($qtree_name)) {
                            # the current vvol lun is not in a qtree
                            # this handles implicit and explicit relative qtree pep cases
                            $pep = "/vol/" . $volume_name . "/" . $pep;
                        } else {
                            # the current vvol lun is in a qtree
                            if ($pep =~ /^\.\.\//) {
                                # remove the ../ from the pep
                                $pep =~ s/^\.\.\///;
                                # the user does not want this pep to be in the qtree
                                # prepend just the vol name
                                $pep = "/vol/" . $volume_name . "/"  . $pep;
                            } else {
                                # this pep is a relative path with an implicit qtree name,
                                # ie: it does not contain the qtree name, but the vvol is in a qtree
                                        # prepend the curent volume name with the qtree
                                $pep = "/vol/" . $volume_name . "/" . $qtree_name . "/" . $pep;
                            }
                        }
                    }

                    $pep;
                } @lun_binds;

                # create the data entry for this lun bind cleanup entry
                my $lun_bind_data = {
                    reused => $SanMgr::TRUE,
                    name => $name,
                    lun_binds => \@lun_binds,
                    volume => $volume_name,
                    qtree => $qtree_name,
                    vserver => $vserver_name,
                    type => "lun_bind",
                    cleanup => $cleanup_policy
                };

                push(@return_data, $lun_bind_data);

            }
        }

        $data = {
            reused   => $SanMgr::TRUE,
            name     => $name,
            lun_maps => \@lun_maps,
            volume   => $volume_name,
            qtree    => $qtree_name,
            vserver  => $vserver_name,
            type     => "lun",
            cleanup  => $cleanup_policy
        };

        push(@return_data, $data);
    } # end foreach lun

    return \@return_data;
} ## end sub create_cleanup_obj_luns

1;
