#
# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on fabrics
## @author derricks@netapp.com
## @status private
## @pod here
#
# $Id: //depot/prod/test/Rsierranevada/lib/San/SanMgr/Fabric.pm#5 $

package San::SanMgr::Fabric;

# Compiler directives.
use strict;

# Module import.
use Tharn;
use Error;
use San::SanLog;
use San::SanMgr::Util;
use San::FabricMgr;

use Data::Dumper;

# Load resource specific modules
use San::SanMgr::Fabric::config;
use San::SanMgr::Fabric::deconfig;
use San::SanMgr::Fabric::validate;
use San::SanMgr::Fabric::FabricHelper;

use vars qw(@ISA);

@ISA = qw(San::SanMgr::Util Error SanLog);

@ISA = qw(San::SanMgr::Util Error SanLog
          San::SanMgr::Fabric::config
          San::SanMgr::Fabric::deconfig
          San::SanMgr::Fabric::validate
          San::SanMgr::Fabric::FabricHelper);


###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new resource object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($proto, %args) = @_;

    my $self = {};

    my $class = ref($proto) || $proto;
    bless($self, $class);

    # allow for error stack build ups
    $self->{error_dont_overwrite_errors} = 1;
    $self->{error_stack_style}           = "normal";

    &ParamsMgr::process_args(%args);
    my $parent = &ParamsMgr::get_arg(-arg => "parent", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Save parent
    $self->{parent} = $parent;

    # Get the set of environment parameters from ParamsMgr
    my $environment = &ParamsMgr::search(search_path => [], key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Set the objects debug level
    $self->debug_level(-debug_level => $environment->{DEBUG_LEVEL});
    if ($self->error()) {
        return $self;
    }

    return $self;
} ## end sub _new

###############################################################################
# Method:    config
# Objective: Launch a subtest to config all fabrics
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config {
    my ($self, %args) = @_;

    $self->clear_error();

    # check to see if config is disabled for all fabrics
    my $fabrics = &ParamsMgr::search(
        -search_path => [],
        -key         => "FABRICS",
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }
    $fabrics->{CONFIG} = 0 if (not defined($fabrics->{CONFIG}));

    if ($fabrics->{CONFIG} == 0) {
        $self->sanlog(-msg => "Skipping Fabric Config.");
        return $SanMgr::TRUE;
    }

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $subtest = &SubIPC::create(
        -id  => "$run_id_prefix" . "Fabric-config",
        -esc => $self->{parent}->{enable_subtest_count},
        -run => sub { $self->_config_all_fabrics(); },
        -ipc => 1,
        -bg  => 1
    );
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "Fabric config has passed\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
    } ## end else [ if ($st_status->{failed...

    &SubIPC::remove_subtest();

    return $SanMgr::TRUE;
} ## end sub config

###############################################################################
# Method:    deconfig
# Objective: Launch subtest to deconfig all fabrics
# Details:   See POD documentation at the end of this file.
###############################################################################
sub deconfig {
    my ($self, %args) = @_;

    $self->clear_error();

    # check to see if deconfig is disabled for all fabrics
    my $fabrics = &ParamsMgr::search(
        -search_path => [],
        -key         => "FABRICS",
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }
    $fabrics->{DECONFIG} = 0 if (not defined($fabrics->{DECONFIG}));

    if ($fabrics->{DECONFIG} == 0) {
        $self->sanlog(-msg => "Skipping Fabric Deconfig.");
        return $SanMgr::TRUE;
    }

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $subtest = &SubIPC::create(
        -id  => "$run_id_prefix" . "Fabric-deconfig",
        -esc => $self->{parent}->{enable_subtest_count},
        -run => sub { $self->_deconfig_all_fabrics(); },
        -ipc => 1,
        -bg  => 1
    );
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "Fabric deconfig has passed\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
    } ## end else [ if ($st_status->{failed...

    &SubIPC::remove_subtest();

    return $SanMgr::TRUE;
} ## end sub deconfig

###############################################################################
# Method:    validate
# Objective: Launch subtest to validate all fabrics
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self, %args) = @_;

    $self->clear_error();

    # check to see if config is disabled for all fabrics
    my $fabrics = &ParamsMgr::search(
        -search_path => [],
        -key         => "FABRICS",
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }
    $fabrics->{VALIDATE} = 1 if (not defined($fabrics->{VALIDATE}));

    if ($fabrics->{VALIDATE} == 0) {
        $self->sanlog(-msg => "Skipping Fabric Validation.");
        return $SanMgr::TRUE;
    }

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $subtest = &SubIPC::create(
        -id  => "$run_id_prefix" . "Fabric-validate",
        -esc => $self->{parent}->{enable_subtest_count},
        -run => sub { $self->_validate_all_fabrics(); },
        -ipc => 1,
        -bg  => 1
    );
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "Fabric validation has passed\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
    } ## end else [ if ($st_status->{failed...

    &SubIPC::remove_subtest();

    return $SanMgr::TRUE;
} ## end sub validate

###############################################################################
# Method:    load_resource
# Objective: Gather list of fabrics and create objects for each one
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_resource {
    my ($self) = @_;

    $self->clear_error();

    # Get list of fabrics
    my $fabrics = &ParamsMgr::search(
        -search_path => [],
        -key         => "FABRICS",
        -search_type => "",
        -key_filter  => qr/^FABRIC\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->sanlog(-msg => "No Fabric definitions were found in XML");
        &ParamsMgr::clear_error();
        return $SanMgr::TRUE;
    }

    if (not scalar(keys %{$fabrics})) {
        $self->sanlog(-msg => "No Fabric definitions were found in XML so resource load will be skipped");
        $self->clear_error();
        return $SanMgr::TRUE;
    }

    # Load up the client objects
    my @client_objs = $self->get_objects(-type => "client");
    if ($self->error()) {
        $self->sanlog(-msg => "No client objects found");
        $self->clear_error();
        @client_objs = ();
    }

    my @host_objs = ();

    # Get the client info from the XML
    my $client_xml_hash = &ParamsMgr::search(
        -search_path => [],
        -key         => "CLIENTS",
        -key_filter  => qr/^CLIENT\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->sanlog(-msg => "No client definitions found in XML");
        &ParamsMgr::clear_error();
        $client_xml_hash = {};
    }

    # Process the client objects.  We need to exclude any virtual clients as well
    # as any non-fcp clients.
    foreach my $client_obj (@client_objs) {
        my $client_name = $client_obj->hostname();
        if ($client_obj->error()) {
            $self->_set_err("Failed to get client name from the object");
            return $SanMgr::FALSE;
        }
        unless (defined($client_name) and $client_name ne "") {
            $self->_set_err("Invalid data returned for client name");
            return $SanMgr::FALSE;
        }

        # Look for the client in the XML hash.  We want to make sure that HYPERVISOR is none
        # and PROTOCOL contains fcp.
        foreach my $xml_client (sort keys %{$client_xml_hash}) {
            if ( defined($client_xml_hash->{$xml_client}->{NAME})
                and $client_xml_hash->{$xml_client}->{NAME} eq $client_name) {
                if (defined($client_xml_hash->{$xml_client}->{HYPERVISOR})
                    and defined($client_xml_hash->{$xml_client}->{PROTOCOL})
                    and $client_xml_hash->{$xml_client}->{HYPERVISOR} =~ /^none$/i
                    and $client_xml_hash->{$xml_client}->{PROTOCOL} =~ /fcp/i) {

                    # The client isn't virtual and is fcp so add it to the host_objs array
                    push(@host_objs, $client_obj);
                    last;
                } else {
                    # Looks like this client is either not running fcp or is virtual so
                    # log a message that we are skipping it.
                    $self->sanlog(-msg => "Skipping client '$client_name' because it is either " .
                                          "a virtual client or not configured for fcp.");
                }
            } ## end unless (defined($client_xml_hash...
        } ## end foreach my $xml_client (sort...
    } ## end foreach my $client_obj (@$client_objs)

    # Load up the hypervisor objects
    my @hypervisor_objs = $self->get_objects(-type => "hypervisor");
    if ($self->error()) {
        $self->sanlog(-msg => "No hypervisor objects found");
        $self->clear_error();
        @hypervisor_objs = ();
    }

    # Get the hypervisor info from the XML
    my $hyp_xml_hash = &ParamsMgr::search(
        -search_path => [],
        -key         => "HYPERVISORS",
        -key_filter  => qr/^ESX\d*$|^HYPERV\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->sanlog(-msg => "No hypervisor definitions found in XML");
        &ParamsMgr::clear_error();
        $hyp_xml_hash = {};
    }

    # Process the hypervisor objects.  We need to exclude any non-fcp clients.
    foreach my $hyp_obj (@hypervisor_objs) {
        my $client_name = $hyp_obj->hostname();
        if ($hyp_obj->error()) {
            $self->_set_err("Failed to get client name from the object");
            return $SanMgr::FALSE;
        }
        unless (defined($client_name) and $client_name ne "") {
            $self->_set_err("Invalid data returned for hypervisor name");
            return $SanMgr::FALSE;
        }

        # Look for the client in the XML hash. We want to make sure that
        # PROTOCOL contains fcp.
        foreach my $xml_client (sort keys %{$hyp_xml_hash}) {
            if (defined($hyp_xml_hash->{$xml_client}->{NAME})
                and $hyp_xml_hash->{$xml_client}->{NAME} eq $client_name) {
                if (defined($hyp_xml_hash->{$xml_client}->{PROTOCOL})
                    and $hyp_xml_hash->{$xml_client}->{PROTOCOL} =~ /fcp/i) {

                    # The client isn't fcp so add it to the host_objs array
                    push(@host_objs, $hyp_obj);
                    last;
                } else {
                    # Looks like this client is either not running fcp or is virtual so
                    # log a message that we are skipping it.
                    $self->sanlog(-msg => "Skipping client '$client_name' because it is " .
                                          "not configured for fcp.");
                }
            } ## end unless (defined($hyp_xml_hash...
        } ## end foreach my $xml_client (sort...
    } ## end foreach my $hyp_obj (@$hypervisor_objs)

    # filter out host objs that don't have zoning enabled
    @host_objs = grep {
        my $obj = $_;
        if ($obj->{fcp}->{CONFIG_FCP_PERFORM_ZONING}) {
            1;
        } else {
            $self->sanlog(-msg => "Was requested to skip FCP Zoning for " . $obj->hostname());
            0;
        }
    } @host_objs;

    # Make sure we have at least one client or hypervisor object
    unless (scalar(@host_objs) > 0) {
        $self->_set_err("No FCP clients/hypervisors found that can be used for zoning");
        return $SanMgr::FALSE;
    }

    # Load up the cluster objects
    my @nacl_clusters = $self->get_objects('type' => "cluster",
                                           'obj_class' => 'nacl');
    if ($self->error()) {
        $self->_set_err("Failed to get nacl cluster objects");
        return $SanMgr::FALSE;
    }

    # Grab the vserver names
    my @vservers = $self->get_vserver_names(-type => "regular");
    if ($self->error()) {
        $self->_set_err("Failed to get the vserver names");
        return $SanMgr::FALSE;
    }

    # Get the client vserver map
    my ($client_vserver_map)  = $self->{parent}->get_vserver_client_relationship_hash();
    if ($self->{parent}->error) {
        $self->_set_err("Unable to build the client vserver relationship hash");
        $self->{parent}->_clear_err();
        return $SanMgr::FALSE;
    }

    # Go through fabric list and start setup
    foreach my $fabric (keys(%{$fabrics})) {
        my $current_fabric  = $fabrics->{$fabric};
        my $fabric_name     = $current_fabric->{NAME};
        my $restrict_zoning = $current_fabric->{RESTRICT_ZONING_BASED_ON_EXISTING_IGROUPS};

        # Load up the switch objects for the switches on this fabric
        my $switches = &ParamsMgr::search(
            -search_path => [],
            -exact_key   => 1,
            -key         => "FABRICS.${fabric}",
            -key_filter  => qr/^SWITCH\d*$/,
            paramset => $self->EnGenX::Util::get_paramset()
        );

        unless (scalar(%$switches) > 0) {
            $self->_set_err("Invalid values specified for 'SWITCH' tag in XML");
            return $SanMgr::FALSE;
        }

        # parse through all of the switches in the hash and add them to the fabrics hash
        my @fabric_switches = ();
        foreach my $switch (sort keys %$switches) {
            unless (defined($switches->{$switch}->{NAME}) and $switches->{$switch}->{NAME} ne "") {
                $self->_set_err("Invalid values specified for switch 'name' tag in XML");
                return $SanMgr::FALSE;
            }
            push (@fabric_switches, $switches->{$switch}->{NAME});
        }

        # Check to make sure we actually have switches for this fabric
        unless (scalar(@fabric_switches) > 0) {
            $self->_set_err("No switches defined for fabric '$fabric_name'");
            return $SanMgr::FALSE;
        }

        my @switch_objs = $self->get_objects(-type => "switch", -names => [@fabric_switches]);
        if ($self->error()) {
            my $fabric_switches = join("\n", @fabric_switches);
            $self->_set_err("No switch objects found on fabric '$fabric_name' for switches:\n" .
                            "$fabric_switches");
            return $SanMgr::FALSE;
        }

        # Check to make sure we actually have switch objects for this fabric
        unless (scalar(@switch_objs) > 0) {
            my $fabric_switches = join("\n", @fabric_switches);
            $self->_set_err("No switch objects found on fabric '$fabric_name' for switches:\n" .
                            "$fabric_switches");
            return $SanMgr::FALSE;
        }

        # Create the FabricMgr object
        my $fm_obj = San::FabricMgr->new(
            -parent            => $self,
            -client_objs       => \@host_objs,
            -nacl_cluster_objs => \@nacl_clusters,
            -vservers          => \@vservers,
            -switch_objs       => \@switch_objs,
            -fabric            => $fabric_name,
            -vserver_map       => $client_vserver_map,
            -protocols         => ["fcp"],
            -restrict_zoning_based_on_existing_igroups => $restrict_zoning
        );

        if ($fm_obj->error()) {
            $self->_set_err("Failed to create FabricMgr object for fabric '$fabric_name'", $fm_obj->errmsg());
            return $SanMgr::FALSE;
        }

        # Save the object for future use
        $self->{parent}->{_objects}->{fabric}->{$fabric_name} = $fm_obj;
    }

    return $SanMgr::TRUE;
} ## end sub load_resource

###############################################################################
# Method:    _config_all_fabrics
# Objective: Launch subtest for each fabric to configure
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _config_all_fabrics {
    my ($self) = @_;
    $self->clear_error();

    my ($st_status, @subtests);
    my ($device_list);

    # Get list of fabrics
    my $fabrics = &ParamsMgr::search(
        -search_path => [],
        -key         => "FABRICS",
        -search_type => "",
        -key_filter  => qr/^FABRIC\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    if (not scalar(keys %{$fabrics})) {
        &SubIPC::pass(-msg => "There are no fabrics to perform config for");
        return $SanMgr::TRUE;
    }

    # set up the Event listener and callback methods
    my $filter = sub {
        my $msg = shift;
        return ($msg->{type} =~ /config_complete/);
    };
    my $callback = sub {
        my ($msg, $listener) = @_;

        # parse out the message information
        my ($fabric_name) = $msg->{message}->{name};
        my $sm_fabric_obj = $listener->{sm_fabric_obj};

        $sm_fabric_obj->{summary_info}->{$fabric_name} = $msg->{message}->{summary_info};
        $sm_fabric_obj->{listeners_complete_count}++;
        return $SanMgr::TRUE;
    };
    my $listener = {filter => $filter, callback => $callback, sm_fabric_obj => $self};

    # build the fail listener
    my $fail_filter = sub {
        my $msg = shift;
        return ($msg->{type} =~ /config_failed/);
    };
    my $fail_callback = sub {
        my ($msg, $flistener) = @_;

        my $sm_fabric_obj = $flistener->{sm_fabric_obj};

        # parse out the message information
        my ($fabric_name) = $msg->{message}->{name};
        $sm_fabric_obj->_set_err($msg->{message}->{msg});
        $sm_fabric_obj->{listeners_complete_count}++;
        return $SanMgr::FALSE;
    };

    my $fail_listener = {filter => $fail_filter, callback => $fail_callback, sm_fabric_obj => $self};

    # initialize the listeners complete count to 0
    $self->{listeners_complete_count} = 0;

    # Adding listeners for messages.
    # The messages will go through all the listeners until it find one that is matches
    # the filter for. On match if will then execute the callback.
    NATE::Events::listener_add($listener);
    NATE::Events::listener_add($fail_listener);

    # Go through fabric list and start config
    foreach my $fabric (keys(%{$fabrics})) {
        my $current_fabric = $fabrics->{$fabric};
        my $fabric_name    = $current_fabric->{NAME};
        $current_fabric->{CONFIG} = 1 if (not defined($current_fabric->{CONFIG}));

        # Check to see if resource is to be skipped
        if ($current_fabric->{CONFIG} == 0) {
            $self->sanlog(
                -msg       => "Skipping fabric " . $fabric_name . " because config tag set to \"0\"",
                -msg_level => 0
            );
            next;
        } ## end if ($current_fabric->{...

        # Reconnect the clients
        $self->_reconnect_all_clients(-fabric => $fabric_name);
        if ($self->error) {
            $self->_set_err("Failed to reconnect all clients on fabric '$fabric_name'");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        my $subtest = &SubIPC::create(
            -id     => $fabric_name,
            -run    => sub { $self->config_single(-resource => $current_fabric); },
            -parent => $self,
            -ipc    => 1,
            -bg     => 1
        );
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        push(@subtests, $subtest);
    } ## end foreach my $fabric (keys(%{...

    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "No fabrics have config enabled");
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $total_proc_count = scalar(@subtests);
    my $object_creation_timeout_time = time + $self->{parent}->{obj_creation_timeout};
    # listeners_complete_count gets updated as the $listener and $fail_listener match msgs.
    while ($total_proc_count > $self->{listeners_complete_count}) {
        # eventloop_drain will clear the message queue and apply listeners to the messages.
        NATE::Events::eventloop_drain(timeout => 1);
        if (time > $object_creation_timeout_time) {
            $self->_set_err("The Load Objects task exceeded the timeout '$self->{obj_creation_timeout}'");
            last;
        }
    }

    # Remove the listeners that we added since all the subtest have completed.
    NATE::Events::listener_remove($listener);
    NATE::Events::listener_remove($fail_listener);

    $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "Fabric Config has passed\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
    } ## end else [ if ($st_status->{failed...
    $self->sanlog(-msg => "Aliases and Zones Created Hash:\n" .
                      "Description: This contains data on all aliases and zones" .
                      "created by All Fabrics Configured.\n" .
                      Dumper($self->{summary_info}));
    $self->_create_summary();
    &SubIPC::pass(-msg => "Fabrics have been configured");

    return $SanMgr::TRUE;
} ## end sub _config_all_fabrics

###############################################################################
# Method:    _create_summary
# Objective: Launch subtest for each fabric to deconfigure
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _create_summary {
    my ($self) = @_;
    $self->clear_error();
    my %deleted_hash = ();
    my %created_hash = ();
    my %created_summary = ();
    foreach my $fabric_key (keys %{$self->{summary_info}}) {
        # Sum initiator aliases
        $created_summary{initiator_aliases_created} +=
            $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{totals_info}->{initiator_aliases_created};
        $created_summary{initiator_aliases_reused} +=
            $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{totals_info}->{initiator_aliases_reused};

        # Sum target aliases
        $created_summary{target_aliases_created} +=
            $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{totals_info}->{target_aliases_created};
        $created_summary{target_aliases_reused} +=
            $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{totals_info}->{target_aliases_reused};

        # Sum zones
        $created_summary{zones_created} +=
            $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{totals_info}->{zones_created};
        $created_summary{zones_reused} +=
            $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{totals_info}->{zones_reused};

        # Add the deleted targets to the merged hash for deleted_wwpns_info
        foreach my $wwpn_key ( keys %{$self->{summary_info}->{$fabric_key}->{deleted_wwpns_info_hash}->{vservers}}) {
            $deleted_hash{vservers}{$wwpn_key} =
                $self->{summary_info}->{$fabric_key}->{deleted_wwpns_info_hash}->{vservers}->{$wwpn_key}
                unless exists ($deleted_hash{vservers}{$wwpn_key});
            $deleted_hash{vservers}{$wwpn_key}{count}++;
        }
        # Add the deleted initiators to the merged hash for deleted_wwpns_info
        foreach my $wwpn_key ( keys %{$self->{summary_info}->{$fabric_key}->{deleted_wwpns_info_hash}->{sanclients}}) {
            $deleted_hash{clients}{$wwpn_key} =
                $self->{summary_info}->{$fabric_key}->{deleted_wwpns_info_hash}->{sanclients}->{$wwpn_key}
                unless exists ($deleted_hash{clients}{$wwpn_key});
            $deleted_hash{clients}{$wwpn_key}{count}++;
        }
        # Add the created targets info to the merged hash for aliases and zoned created
        foreach my $cluster ( keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{aliases}->{vservers}}) {
            foreach my $vserver ( keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{aliases}->{vservers}->{$cluster}}) {
                foreach my $alias (@{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{aliases}->{vservers}->{$cluster}->{$vserver}->{$fabric_key}->{new}}) {
                    push(@{$created_summary{vservers}{$cluster}{$vserver}{$fabric_key}{new}},$alias);
                    $created_hash{vservers}{$alias->{wwpn}} = $alias
                        unless exists ($created_hash{vservers}{$alias->{wwpn}});
                    $created_hash{vservers}{$alias->{wwpn}}{count}++;
                }
                foreach my $alias (@{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{aliases}->{vservers}->{$cluster}->{$vserver}->{$fabric_key}->{reused}}) {
                    push(@{$created_summary{vservers}{$cluster}{$vserver}{$fabric_key}{reused}},$alias);
                    $created_hash{vservers}{$alias->{wwpn}} = $alias
                        unless exists ($created_hash{vservers}{$alias->{wwpn}});
                    $created_hash{vservers}{$alias->{wwpn}}{count}++;
                }
            }
        }
        # Add the created initiator info to the merged hash for aliases and zoned created
        foreach my $client ( keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{aliases}->{clients}}) {
            foreach my $alias (@{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{aliases}->{clients}->{$client}->{$fabric_key}->{new}}) {
                push(@{$created_summary{clients}{$client}{$fabric_key}{new}},$alias);
                $created_hash{clients}{$alias->{wwpn}} = $alias
                    unless exists ($created_hash{clients}{$alias->{wwpn}});
                $created_hash{clients}{$alias->{wwpn}}{count}++;
            }
            foreach my $alias (@{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{aliases}->{clients}->{$client}->{$fabric_key}->{reused}}) {
                push(@{$created_summary{clients}{$client}{$fabric_key}{reused}},$alias);
                $created_hash{clients}{$alias->{wwpn}} = $alias
                    unless exists ($created_hash{clients}{$alias->{wwpn}});
                $created_hash{clients}{$alias->{wwpn}}{count}++;
            }
        }
        # Add the created target zones info to the merged hash for aliases and zoned created
        foreach my $cluster ( keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{vservers}}) {
            foreach my $vserver ( keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{vservers}->{$cluster}}) {
                foreach my $zone (keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{vservers}->{$cluster}->{$vserver}->{$fabric_key}->{new}}) {
                    $created_summary{zones}{vservers}{$cluster}{$vserver}{$fabric_key}{new}{$zone} =
                        $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{vservers}->{$cluster}->{$vserver}->{$fabric_key}->{new}->{$zone};
                }
                foreach my $zone (keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{vservers}->{$cluster}->{$vserver}->{$fabric_key}->{reused}}) {
                    $created_summary{zones}{vservers}{$cluster}{$vserver}{$fabric_key}{reused}{$zone} =
                        $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{vservers}->{$cluster}->{$vserver}->{$fabric_key}->{reused}->{$zone};
                }
            }
        }
        # Add the created initiator zones info to the merged hash for aliases and zoned created
        foreach my $client ( keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{clients}}) {
            foreach my $zone (keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{clients}->{$client}->{$fabric_key}->{new}}) {
                $created_summary{zones}{clients}{$client}{$fabric_key}{new}{$zone} =
                    $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{clients}->{$client}->{$fabric_key}->{new}->{$zone};
            }
            foreach my $zone (keys %{$self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{clients}->{$client}->{$fabric_key}->{reused}}) {
                $created_summary{zones}{clients}{$client}{$fabric_key}{reused}{$zone} =
                    $self->{summary_info}->{$fabric_key}->{aliases_and_zones_info_hash}->{zones}->{clients}->{$client}->{$fabric_key}->{reused}->{$zone};
            }
        }
    }
    my $fabric_count = scalar (keys %{$self->{summary_info}});
    foreach my $deleted_wwpn (keys %{$deleted_hash{vservers}}) {
        if ($deleted_hash{vservers}{$deleted_wwpn}{count} >= $fabric_count) {
            next;
        }
        foreach my $created_wwpn (keys %{$deleted_hash{vservers}}) {
            if ($created_wwpn eq $deleted_wwpn) {
                delete $deleted_hash{vservers}{$deleted_wwpn};
            }
        }
    }
    foreach my $deleted_wwpn (keys %{$deleted_hash{clients}}) {
        if ($deleted_hash{clients}{$deleted_wwpn}{count} >= $fabric_count) {
            next;
        }
        foreach my $created_wwpn (keys %{$deleted_hash{clients}}) {
            if ($created_wwpn eq $deleted_wwpn) {
                delete $deleted_hash{clients}{$deleted_wwpn};
            }
        }
    }

    my $jobid  = Tharn::param('JOBID');
    my $logdir = Tharn::param('LOGDIR');
    # replace everything after the job id because it is included in runid
    $logdir =~ s/$jobid\S*/$jobid/;
    my $runid  = Tharn::param('RUNID');

    # Initiator_Aliases_Created logging
    $Log->redirect(path=>"$logdir/$runid/Initiator_Aliases_Created.log");
    $Log->comment($created_summary{initiator_aliases_created}." new initiator aliases");
    $Log->comment($created_summary{initiator_aliases_reused}." reused initiator aliases");
    $Log->comment($created_summary{initiator_aliases_created} +
                  $created_summary{initiator_aliases_reused}." total initiator aliases");
    $Log->comment("This contains all the initiator aliases create by " .
                  "Fabric_Config.\n" .
                  Dumper($created_summary{clients}));
    #Logging the following so that the log will PASS
    $Log->comment("END OF TEST \/ RESULTS SUMMARY\nNo results reported\n");
    # Direct $Log calls to write back to the original log file
    $Log->reset_path();

    # Target_Aliases_Created logging
    $Log->redirect(path=>"$logdir/$runid/Target_Aliases_Created.log");
    $Log->comment($created_summary{target_aliases_created}." new target aliases");
    $Log->comment($created_summary{target_aliases_reused}." reused target aliases");
    $Log->comment($created_summary{target_aliases_created} +
                  $created_summary{target_aliases_reused}." total target aliases");
    $Log->comment("This contains all the target aliases create by " .
                  "Fabric_Config.\n" .
                  Dumper($created_summary{vservers}));
    #Logging the following so that the log will PASS
    $Log->comment("END OF TEST \/ RESULTS SUMMARY\nNo results reported\n");
    # Direct $Log calls to write back to the original log file
    $Log->reset_path();

    # Zones_Created logging
    $Log->redirect(path=>"$logdir/$runid/Zones_Created.log");
    $Log->comment($created_summary{zones_created}." new zones");
    $Log->comment($created_summary{zones_reused}." reused zones");
    $Log->comment($created_summary{zones_created} +
                  $created_summary{zones_reused}." total zones");
    $Log->comment("This contains all the zones create by " .
                  "Fabric_Config.\n" .
                  Dumper($created_summary{zones}));
    #Logging the following so that the log will PASS
    $Log->comment("END OF TEST \/ RESULTS SUMMARY\nNo results reported\n");
    # Direct $Log calls to write back to the original log file
    $Log->reset_path();

    # Target_Aliases_Not_Created logging
    if(scalar (keys %{$deleted_hash{vservers}})) {
        $Log->warn("There are Target WWPNs that were never used to create aliases");
        $Log->redirect(path=>"$logdir/$runid/Target_Aliases_Not_Created.log");
        $Log->comment("This contains all the Target wwpns not used in " .
                      "Fabric_Config.\n" .
                      Dumper($deleted_hash{vservers}));
        #Logging the following so that the log will PASS
        $Log->comment("END OF TEST \/ RESULTS SUMMARY\nNo results reported\n");
        # Direct $Log calls to write back to the original log file
        $Log->reset_path();
    }

    # Initiator_Aliases_Not_Created logging
    if(scalar (keys %{$deleted_hash{clients}})){
        $Log->warn("There are Initiator WWPNs that were never used to create aliases");
        $Log->redirect(path=>"$logdir/$runid/Initiator_Aliases_Not_Created.log");
        $Log->comment("This contains all the Initiator wwpns not used in " .
                      "Fabric_Config.\n" .
                      Dumper($deleted_hash{clients}));
        #Logging the following so that the log will PASS
        $Log->comment("END OF TEST \/ RESULTS SUMMARY\nNo results reported\n");
        # Direct $Log calls to write back to the original log file
        $Log->reset_path();
    }

    $self->sanlog(-msg => "There are $fabric_count fabrics");
    $self->sanlog(-msg => "LOGDIR:" . $logdir);
    $self->sanlog(-msg => "RUNID:" . $runid);
    $self->sanlog(-msg => "JOBID:" . $jobid);
    $self->sanlog(-msg => "Deleted Hash:\n" .
                  "Description: This contains data on all wwpns " .
                  "not used by all the fabrics.\n" .
                  Dumper(\%deleted_hash));
    $self->sanlog(-msg => "Created Summary:\n" .
              "Description: This contains data on all wwpns " .
              "used by all the fabrics.\n" .
              Dumper(\%created_summary));

    return $SanMgr::TRUE;
} ## end sub _create_summary

###############################################################################
# Method:    _deconfig_all_fabrics
# Objective: Launch subtest for each fabric to deconfigure
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _deconfig_all_fabrics {
    my ($self) = @_;
    $self->clear_error();

    my ($st_status, @subtests);
    my ($device_list);

    # Get list of fabrics
    my $fabrics = &ParamsMgr::search(
        -search_path => [],
        -key         => "FABRICS",
        -search_type => "",
        -key_filter  => qr/^FABRIC\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    if (not scalar(keys %{$fabrics})) {
        &SubIPC::pass(-msg => "There are no fabrics to perform config for");
        return $SanMgr::TRUE;
    }

    # Go through fabric list and start deconfig
    foreach my $fabric (keys(%{$fabrics})) {
        my $current_fabric = $fabrics->{$fabric};
        my $fabric_name    = $current_fabric->{NAME};
        $current_fabric->{DECONFIG} = 1 if (not defined($current_fabric->{DECONFIG}));

        # Check to see if resource is to be skipped
        if ($current_fabric->{DECONFIG} == 0) {
            $self->sanlog(
                -msg       => "Skipping fabric " . $fabric_name . " because deconfig tag set to \"0\"",
                -msg_level => 0
            );
            next;
        } ## end if ($current_fabric->{...

        # Reconnect the clients
        $self->_reconnect_all_clients(-fabric => $fabric_name);
        if ($self->error) {
            $self->_set_err("Failed to reconnect all clients on fabric '$fabric_name'");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        my $subtest = &SubIPC::create(
            -id     => $fabric_name,
            -run    => sub { $self->deconfig_single(-resource => $current_fabric); },
            -parent => $self,
            -ipc    => 1,
            -bg     => 1
        );
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        push(@subtests, $subtest);
    } ## end foreach my $fabric (keys(%{...

    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "No fabrics have deconfig enabled");
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "Fabric Deconfig has passed\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
    } ## end else [ if ($st_status->{failed...

    &SubIPC::pass(-msg => "Fabrics have been deconfigured");

    return $SanMgr::TRUE;
} ## end sub _deconfig_all_fabrics

###############################################################################
# Method:    _validate_all_fabrics
# Objective: Launch subtest for each fabric to validate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _validate_all_fabrics {
    my ($self) = @_;
    $self->clear_error();

    my ($st_status, @subtests);
    my ($device_list);

    # Get list of fabrics
    my $fabrics = &ParamsMgr::search(
        -search_path => [],
        -key         => "FABRICS",
        -search_type => "",
        -key_filter  => qr/^FABRIC\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    if (not scalar(keys %{$fabrics})) {
        &SubIPC::pass(-msg => "There are no fabrics to perform validate for");
        return $SanMgr::TRUE;
    }

    # Go through fabric list and start validate
    foreach my $fabric (keys(%{$fabrics})) {
        my $current_fabric = $fabrics->{$fabric};
        my $fabric_name    = $current_fabric->{NAME};
        $current_fabric->{VALIDATE} = 1 if (not defined($current_fabric->{VALIDATE}));

        # Check to see if resource is to be skipped
        if ($current_fabric->{VALIDATE} == 0) {
            $self->sanlog(
                -msg       => "Skipping fabric " . $fabric_name . " because validate tag set to \"0\"",
                -msg_level => 0
            );
            next;
        } ## end if ($current_fabric->{...

        # Reconnect the clients
        $self->_reconnect_all_clients(-fabric => $fabric_name);
        if ($self->error) {
            $self->_set_err("Failed to reconnect all clients on fabric '$fabric_name'");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        my $subtest = &SubIPC::create(
            -id     => $fabric_name,
            -run    => sub { $self->validate_single(-resource => $current_fabric); },
            -parent => $self,
            -ipc    => 1,
            -bg     => 1
        );
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        push(@subtests, $subtest);
    } ## end foreach my $fabric (keys(%{...

    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "No fabrics have validate enabled");
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "Fabric validate has passed\n\t\tFailed: "
              . $st_status->{failed}
              . "\n\t\tPassed: "
              . $st_status->{passed});
    } ## end else [ if ($st_status->{failed...

    &SubIPC::pass(-msg => "Fabrics have been validated");

    return $SanMgr::TRUE;
} ## end sub _validate_all_fabrics

###############################################################################
# Method:    _reconnect_all_clients
# Objective: Reconnect to the clients
# Inputs:    -fabric   Name of a fabric to reconnect the clients on
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _reconnect_all_clients {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $fabric = &ParamsMgr::get_arg(-arg => "fabric", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Get fabric object
    my ($fabric_obj) = $self->get_objects(-type => "fabric", -names => [$fabric]);
    if ($self->error()) {
        $self->_set_err("Failed to get object for fabric '$fabric'");
        return $SanMgr::FALSE;
    } ## end if ($self->error())

    # Get all of the client objects off of the fabric and reconnect them.
    my $client_objs = $fabric_obj->get_client_hash();
    if ($fabric_obj->error) {
        $self->_set_err("Failed to retrieve the client hash for fabric '$fabric'",
                        $fabric_obj);
        return $SanMgr::FALSE;
    } ## end if ($fabric_obj->error)

    # Iterate over each client objects and reconnect them.
    foreach my $client (keys %$client_objs) {
        my $client_obj = $client_objs->{$client};
        unless (ref($client_obj) =~ /sanclient/i) {
            $self->_set_err("Invalid data type value for client '$client'");
            return $SanMgr::FALSE;
        }

        $client_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
        if ($client_obj->error()) {
            $self->_set_err("Unable to reconnect client '$client'", $client_obj->errmsg);
            return $SanMgr::FALSE;
        }

        $self->sanlog(-msg => "Successfully reconnected to client '$client' on fabric '$fabric'");
    }

    return $SanMgr::TRUE;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

San::SanMgr::Fabric

=head1 Author

Derrick Shultz

=head1 Description

Package that provides simple accessor routines to perform config, deconfig, and validation
of a fabric resource. This package can NOT be called outside of SanMgr.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm, SanLog.pm, San::SanMgr::Util.pm>

=head1 Public Methods

=head2 B<config()>

This function will subtest out a call to configure all fabrics.

=over 4

=item Inputs

None

=item Outputs

This function uses SanMgr::TRUE and SanMgr::FALSE to return status.

=item Examples

Self is a SanMgr object containing the sm_fabric_object.

$self->{sm_fabric_object}->config();

=back

=head2 B<deconfig()>

This function will subtest out a call to deconfigure all fabrics.

=over 4

=item Inputs

None

=item Outputs

This function uses SanMgr::TRUE and SanMgr::FALSE to return status.

=item Examples

Self is a SanMgr object containing the sm_fabric_object.

$self->{sm_fabric_object}->deconfig();

=back

=head2 B<load_resource()>

This function will create and store a FabricMgr object for each fabric.

=over 4

=item Inputs

None

=item Outputs

This function uses SanMgr::TRUE and SanMgr::FALSE to return status.

=item Examples

Self is a SanMgr object containing the sm_fabric_object.

$self->{sm_fabric_object}->load_resource();

=back

=head2 B<validate()>

This function will subtest out a call to validate all fabrics.

=over 4

=item Inputs

None

=item Outputs

This function uses SanMgr::TRUE and SanMgr::FALSE to return status.

=item Examples

Self is a SanMgr object containing the sm_fabric_object.

$self->{sm_fabric_object}->validate();

=back

=cut
