# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of helper methods to work on fabrics
## @author derricks@netapp.com
## @status private
## @pod here

package San::SanMgr::Fabric::FabricHelper;

use POSIX;
use Data::Dumper;
use Tie::IxHash;

# compiler directives
use strict;

###############################################################################
# Method:     dump_all_hashes
# Objective:  Dump all of the FabricMgr hashes in a subtest
# Parameters: None
# Returns:    True on success.  Fail on error.
###############################################################################
sub dump_all_hashes {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $fabric_name = &ParamsMgr::get_arg(-arg => "fabric", -required => 1);
    my $subtested   = &ParamsMgr::get_arg(-arg => "subtested", -required => 0, -default => 0);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        if ($subtested) {
            &SubIPC::fail(-object => $self);
        } else {
            return $SanMgr::FALSE;
        }
    }

    my $dump_id = $fabric_name . "_dump_hashes";
    my @subtests = ();
    my $subtest = &SubIPC::create(
                                    -id     => $dump_id, 
                                    -run    => sub { $self->_dump_all_hashes_subtest(-fabric => $fabric_name); },
                                    -parent => $self,
                                    -ipc    => 1,
                                    -bg     => 1
                                 );
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        if ($subtested) {
            &SubIPC::fail(-object => $self);
        } else {
            return $SanMgr::FALSE;
        }
    }
    push(@subtests, $subtest);

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        if ($subtested) {
            &SubIPC::fail(-object => $self);
        } else {
            return $SanMgr::FALSE;
        }
    }

    if ($subtested) {
        &SubIPC::pass(-msg => "Fabric Hash Dump Complete");
    } else {
        return $SanMgr::TRUE;
    }
}

###############################################################################
#               PRIVATE FUNCTIONS
###############################################################################

###############################################################################
# Method:     _dump_all_hashes_subtest
# Objective:  Dump all of the FabricMgr hashes
# Parameters: None
# Returns:    True on success.  Fail on error.
###############################################################################
sub _dump_all_hashes_subtest {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $fabric_name = &ParamsMgr::get_arg(-arg => "fabric", -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Get fabric object
    my ($fabric_obj) = $self->get_objects(-type => "fabric", -names => [$fabric_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get fabric ($fabric_name) object");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($self->error())

    $fabric_obj->dump_hashes();
    if ($fabric_obj->error) {
        $self->_set_err("Failed to dump the hashes for fabric ($fabric_name)", $fabric_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($fabric_obj->error)

    &SubIPC::pass(-msg => "Fabric Hash Dump Complete");

    return $SanMgr::TRUE;
}

1;
