#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on resource config
## @author derricks@netapp.com
## @status private
## @pod here
#
# $Id: //depot/prod/test/Rsierranevada/lib/San/SanMgr/Fabric/config.pm#2 $

package San::SanMgr::Fabric::config;

# Compiler directives.
use strict;
use San::SanLog;
use ParamsMgr;
use San::SanUtil;

use Data::Dumper;

use vars qw(@ISA);
@ISA = qw(Error SanLog);

###############################################################################
# Method:    config_single
# Objective: Config run against a single resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_single {
    use NATE::STAF::Peer qw($Parent);
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Parent->message_put(type => "config_failed",
                             message => {msg => "Failed to validate parameter for the config_single",
                                         name => "Fabric wasn't defined yet"});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # $self is a Fabric object
    # $resource is a hash of parameters for a particular resource to work on
    $self->sanlog(
        -msg       => "Resource parameters/data:\n" . Dumper($resource),
        -msg_level => 3
    );

    # Load up the resource information and get them via get_arg
    my $caller_id = $resource->{NAME} . "|" . time();
    &ParamsMgr::process_args(%{$resource}, -CALLER => $caller_id);

    #############################################
    ### General fabric parameters
    #############################################
    my $fabric_name = &ParamsMgr::get_arg(
        -arg            => "NAME",
        -required       => 1,
        -default_to_env => 1,
        -CALLER         => $caller_id
    );
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Parent->message_put(type => "config_failed",
                             message => {msg => "Failed to validate parameter for the resource",
                                         name => $fabric_name});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Get fabric object
    my ($fabric_obj) = $self->get_objects(-type => "fabric", -names => [$fabric_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get fabric ($fabric_name) object");
        $Parent->message_put(type => "config_failed",
                             message => {msg => "Failed to get fabric ($fabric_name) object",
                                         name => $fabric_name});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($self->error())

    # Grab the switch objects and reconnect them
    my $switches_hash = $fabric_obj->get_switch_hash();
    if ($fabric_obj->error()) {
        $self->_set_err("Failed to get switch hash for fabric ($fabric_name)",
                        $fabric_obj->errmsg());
        $Parent->message_put(type => "config_failed",
                             message => {msg => "Failed to get switch hash for fabric ($fabric_name)",
                                         name => $fabric_name});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($self->error())

    foreach my $switch (sort keys %$switches_hash) {
        # Reconnect the switch object in case we are using the same switch
        # in multiple fabrics
        my $obj = $switches_hash->{$switch};
        $obj->reconnect(-disconnect_existing_conn => $SanUtil::FALSE);
        if ($obj->error()) {
            $obj->_set_err("Unable to reconnect switch '$switch'", $obj->errmsg());
            $Parent->message_put(type => "config_failed",
                     message => {msg => "Unable to reconnect switch '$switch'",
                                 name => $fabric_name});
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # Build out all of the hashes. Normally you don't have to do this as they are
    # build on demand, however, we want to subtest out the dump of the hashes and
    # we don't want the dump to go throught the build out phase.
    # Dump out the hashes
    $fabric_obj->build_all_hashes();
    if ($fabric_obj->error) {
        # Subtest off the dump of the fabric hashes
        $self->dump_all_hashes(-fabric => $fabric_name);
        $self->_set_err("Failed to build the hashes for fabric ($fabric_name)", $fabric_obj->errmsg());
        $Parent->message_put(type => "config_failed",
                     message => {msg => "Failed to build the hashes for fabric ($fabric_name)",
                         name => $fabric_name});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($fabric_obj->error)

    # Subtest off the dump of the fabric hashes
    $self->dump_all_hashes(-fabric => $fabric_name);
    if ($self->error()) {
        $Parent->message_put(type => "config_failed",
                             message => {msg => "Failed to dump all hashes for fabric ($fabric_name)",
                                           name => $fabric_name});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Run the "create" routines within the fabric library
    # Create the aliases
    $fabric_obj->create_aliases();
    if ($fabric_obj->error) {
        $self->_set_err("Failed to create the aliases for fabric ($fabric_name)", $fabric_obj->errmsg());
        $Parent->message_put(type => "config_failed",
                             message => {msg => "Failed to create the aliases for fabric ($fabric_name)",
                                         name => $fabric_name});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($fabric_obj->error)

    # Create the zones
    $fabric_obj->create_zones();
    if ($fabric_obj->error) {
        $self->_set_err("Failed to create the zones for fabric ($fabric_name)", $fabric_obj->errmsg());
        $Parent->message_put(type => "config_failed",
                             message => {msg => "Failed to create the zones for fabric ($fabric_name)",
                                         name => $fabric_name});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($fabric_obj->error)

    # Activate the changes
    $fabric_obj->activate_fabric_changes();
    if ($fabric_obj->error) {
        $self->_set_err("Failed to activate the changes for fabric ($fabric_name)", $fabric_obj->errmsg());
        $Parent->message_put(type => "config_failed",
                             message => {msg => "Failed to activate the changes for fabric ($fabric_name)",
                                         name => $fabric_name});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($fabric_obj->error)

    # Write back to parent the device_list generated
    $Parent->message_put(type => "config_complete",
                         message =>
                            {msg => "Completed fabric ($fabric_name) config",
                             name => $fabric_name,
                             summary_info =>
                                {aliases_and_zones_info_hash =>
                                    $fabric_obj->get_aliases_and_zones_created_hash(),
                                 deleted_wwpns_info_hash     =>
                                    $fabric_obj->get_deleted_wwpns_info_hash}
                            }
                        );

    $fabric_obj->dump_aliases_and_zones_created_hash();
    &SubIPC::pass(-msg => "Fabric config complete");

    return $SanMgr::TRUE;
} ## end sub config_single

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod
=head1 Package Name

San::SanMgr::Fabric::config

=head1 Author

Derrick Shultz

=head1 Description

Provide a common set of libraries to work on resource config

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm, San::SanLog.pm>

=head1 Public Methods

=head2 B<config_single()>

This function will configure a single fabric.

=over 4

=item Inputs

=item * I<-resource>

XML resource hash for the fabric being configured.

=item Outputs

This function uses &SubIPC::pass and &SubIPC::fail to return status

=item Examples

&San::SanMgr::Fabric::config::config_single($self, -resource => $current_fabric);

=back

=cut
