#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on resource deconfig
## @author derricks@netapp.com
## @status private
## @pod here
#
# $Id: //depot/prod/test/Rsierranevada/lib/San/SanMgr/Fabric/deconfig.pm#1 $

package San::SanMgr::Fabric::deconfig;

# Compiler directives.
use strict;
use San::SanLog;
use ParamsMgr;
use San::SanUtil;

use Data::Dumper;

use vars qw(@ISA);
@ISA = qw(Error SanLog);

###############################################################################
# Method:    deconfig_single
# Objective: Deconfig run against a single resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub deconfig_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # $resource is a hash of parameters for a particular resource to work on
    $self->sanlog(
        -msg       => "Resource parameters/data:\n" . Dumper($resource),
        -msg_level => 3
    );

    # Load up the resource information and get them via get_arg
    my $caller_id = $resource->{NAME} . "|" . time();
    &ParamsMgr::process_args(%{$resource}, -CALLER => $caller_id);

    #############################################
    ### General fabric parameters
    #############################################
    my $fabric_name = &ParamsMgr::get_arg(
        -arg            => "NAME",
        -required       => 1,
        -default_to_env => 1,
        -CALLER         => $caller_id
    );
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Get fabric object
    my ($fabric_obj) = $self->get_objects(-type => "fabric", -names => [$fabric_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get fabric ($fabric_name) object");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($self->error())

    # Grab the switch objects and reconnect them
    my $switches_hash = $fabric_obj->get_switch_hash();
    if ($fabric_obj->error()) {
        $self->_set_err("Failed to get switch hash for fabric ($fabric_name)",
                        $fabric_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($self->error())

    foreach my $switch (sort keys %$switches_hash) {
        # Reconnect the switch object in case we are using the same switch
        # in multiple fabrics
        my $obj = $switches_hash->{$switch};
        $obj->reconnect(-disconnect_existing_conn => $SanUtil::FALSE);
        if ($obj->error()) {
            $obj->_set_err("Unable to reconnect switch '$switch'", $obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # Build out all of the hashes. Normally you don't have to do this as they are
    # build on demand, however, we want to subtest out the dump of the hashes and
    # we don't want the dump to go throught the build out phase.
    # Dump out the hashes
    $fabric_obj->build_all_hashes();
    if ($fabric_obj->error) {
        # Check to see what kind of error we received.  If we receved one because
        # we don't have enough resources, then just return TRUE and skip
        # deconfig.
        my @errmsg = $fabric_obj->errmsg();
        if (grep(/Not enough targets on node/, @errmsg) > 0
            or grep(/Invalid number of resources for fabric/, @errmsg) > 0) {
            my $errmsg = join("\n", @errmsg);
            $self->sanlog(-msg => "Skipping fabric deconfig due to lack of resources:\n" . $errmsg);
            &SubIPC::pass(-msg => "Skipping fabric deconfig due to lack of resources:\n" . $errmsg);
            return $SanMgr::TRUE;
        } else {
            $self->_set_err("Failed to build the hashes for fabric ($fabric_name)", @errmsg);
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    } ## end if ($fabric_obj->error)

    # Subtest off the dump of the fabric hashes
    $self->dump_all_hashes(-fabric => $fabric_name);
    if ($self->error()) {
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Run the "delete" routines within the fabric library
    # Delete the zones
    $fabric_obj->delete_zones();
    if ($fabric_obj->error) {
        $self->_set_err("Failed to delete the zones for fabric ($fabric_name)", $fabric_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($fabric_obj->error)

    # Delete the aliases
    $fabric_obj->delete_aliases();
    if ($fabric_obj->error) {
        $self->_set_err("Failed to delete the aliases for fabric ($fabric_name)", $fabric_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($fabric_obj->error)

    # Activate the fabric changes
    $fabric_obj->activate_fabric_changes();
    if ($fabric_obj->error) {
        $self->_set_err("Failed to activate the fabric changes for fabric ($fabric_name)", $fabric_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } ## end if ($fabric_obj->error)

    # Write back to parent the device_list generated
    &SubIPC::write(
        -msg  => "Completed fabric ($fabric_name) deconfig",
        -type => "deconfig_complete",
        -name => $fabric_name
    );
    if (&SubIPC::error()) {
        $self->_set_err("Unable to write (deconfig_complete) message to parent", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "Fabric deconfig complete");

    return $SanMgr::TRUE;
} ## end sub deconfig_single

1;

#############################################################################
#                             POD Documentation
#############################################################################
=pod

=head1 Package Name

San::SanMgr::Fabric::deconfig

=head1 Author

Derrick Shultz

=head1 Description

Provide a common set of libraries to work on resource deconfig

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm, San::SanLog.pm>

=head1 Public Methods

=head2 B<deconfig_single()>

This function will deconfigure a single fabric

=over 4

=item Inputs

=item * I<-resource>

XML resource hash for the fabric being deconfigured.

=item Outputs

This function uses &SubIPC::pass and &SubIPC::fail to return status

=item Examples

&San::SanMgr::Fabric::deconfig::deconfig_single($self, -resource => $current_fabric);

=back

=cut
