#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to configure/setup hypervisors
## @author hrivera@netapp.com
## @status private
## @pod here

package San::SanMgr::HyperVisor;

# Compiler directives.
use warnings;
use strict;

# Module import.
use Tharn;
use Error;
use San::SanLog;
use San::SanMgr::Util;
use San::SanClient;
use Data::Dumper;
use Data::Dump;
use ParamsMgr;

use NATE::Process;
use NATE::STAF::Peer;
use NATE::STAF qw(staf_global);
use NATE::Util::Sockets;
use NATE::BaseException qw(:try);
use Sys::Hostname;

# Load resource specific modules
use San::SanMgr::HyperVisor::ESX::config;
use San::SanMgr::HyperVisor::ESX::setup;
use San::SanMgr::HyperVisor::ESX::cleanup;
use San::SanMgr::HyperVisor::ESX::deconfig;
use San::SanMgr::HyperVisor::HyperV::config;
use San::SanMgr::HyperVisor::HyperV::setup;
use San::SanMgr::HyperVisor::HyperV::cleanup;
use San::SanMgr::HyperVisor::HyperV::deconfig;
use San::SanMgr::HyperVisor::CXS::config;
use San::SanMgr::HyperVisor::CXS::setup;
use San::SanMgr::HyperVisor::CXS::cleanup;
use San::SanMgr::HyperVisor::CXS::deconfig;

use vars qw(@ISA);

@ISA = qw(San::SanMgr::Util Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new resource object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($class, %args) = @_;

    my $self = {};

    bless ($self, $class);

    # allow for error stack build ups
    $self->{error_dont_overwrite_errors} = 1;
    $self->{error_stack_style} = "normal";

    &ParamsMgr::process_args(%args);
    my $parent = &ParamsMgr::get_arg(-arg => "parent",
                                     -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Save parent
    $self->{parent} = $parent;

    # Get the set of environment parameters from ParamsMgr
    my $environment = &ParamsMgr::search(search_path => [],
                                         key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Set the objects debug level
    $self->debug_level(-debug_level => $environment->{DEBUG_LEVEL});
    if ($self->error()) {
        return $self;
    }

    return $self;
}

###############################################################################
# Method:    setup
# Objective: Launch subtest for all HyperVisors to perform setup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup {
    my ($self, %args) = @_;

    $self->clear_error();

    # check to see if setup is disabled for all hypervisors
    my $hypervisors = &ParamsMgr::search(-search_path => [],
                                         -key => "HYPERVISORS",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    if (not defined($hypervisors) or not defined($hypervisors->{SETUP}) or $hypervisors->{SETUP} == 0) {
        $self->sanlog(-msg => "Skipping HyperVisor Setup.");
        return $SanMgr::TRUE;
    }

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $subtest = &SubIPC::create(-id => "$run_id_prefix" . "HyperVisor-setup",
                                  -esc => $self->{parent}->{enable_subtest_count},
                                  -run => sub { $self->_setup_all_hypervisors(); },
                                  -ipc => 1,
                                  -bg => 1);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # parse out Hyperv CDL
    my @messages = &SubIPC::get_messages(-subtest => [$subtest],
                                         -type => "hyperv_client_dev_list");
    if (&SubIPC::error()) {
        $self->_set_err("Failed to parse hyperv client dev list messages", &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    # prepare the device list for all hyper v hosts
    my $device_list;
    foreach my $message (@messages) {
        foreach my $client (keys %{$message->{extra}->{device_list}}) {
            push(@{$device_list->{$client}}, @{$message->{extra}->{device_list}->{$client}->{DEV_LIST}});
        }
    }
    if (defined($device_list)) {
        $self->{parent}->store_client_dev_list(-device_list => $device_list);
        if ($self->{parent}->error()) {
            $self->_set_err("Failed to store client dev list", $self->{parent}->errmsg());
            return $SanMgr::FALSE;
        }
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor Setup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::remove_subtest();

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _setup_all_hypervisors
# Objective: Launch setup subtests for each HyperVisor type
###############################################################################
sub _setup_all_hypervisors {
    my ($self) = @_;

    # Get list of ESX hypervisors to run setup against
    my $esx_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^ESX\d*$/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Get list of HyperV hypervisors to run setup against
    my $hyperv_hypervisors = &ParamsMgr::search(-search_path => [],
                                                -key => "HYPERVISORS",
                                                -key_filter => qr/^HYPERV\d*$/,
                                                paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Get list of Citrix XenServer hypervisors to run setup against
    my $cxs_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^CXS\d*$/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @subtests = ();

    # run setup on ESX
    if (scalar(keys %{$esx_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "ESX-setup",
                                      -run => sub { &San::SanMgr::HyperVisor::ESX::setup::setup_all($self,
                                                                                                    -resources => $esx_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run setup on HyperV
    if (scalar(keys %{$hyperv_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "HyperV-setup",
                                      -run => sub { &San::SanMgr::HyperVisor::HyperV::setup::setup_all($self,
                                                                                                       -resources => $hyperv_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run setup on CXS
    if (scalar(keys %{$cxs_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "CXS-setup",
                                      -run => sub { &San::SanMgr::HyperVisor::CXS::setup::setup_all($self,
                                                                                                    -resources => $cxs_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "Skipping Hypervisors setup because there are no hypervisors");
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("HyperVisor Setup failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # parse out Hyperv CDL
    my @messages = &SubIPC::get_messages(-subtest => \@subtests,
                                         -type => "hyperv_client_dev_list");
    if (&SubIPC::error()) {
        $self->_set_err("Failed to parse hyperv client dev list messages", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # prepare the device list for all hyper v hosts
    my $device_list;
    foreach my $message (@messages) {
        foreach my $client (keys %{$message->{extra}->{device_list}}) {
            $device_list->{$client} = $message->{extra}->{device_list}->{$client};
        }
    }
    if (defined($device_list)) {
        $self->{parent}->store_client_dev_list(-device_list => $device_list);
        if ($self->{parent}->error()) {
            $self->_set_err("Failed to store client dev list", $self->{parent}->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # send the final device list up to the parent
    &SubIPC::write(-subtest => [],
                   -msg => "Hyper-V CDL",
                   -type => "hyperv_client_dev_list",
                   -device_list => $device_list);
    if (&SubIPC::error()) {
        $self->_set_err("Failed to write Hyperv CDL to parents", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor Setup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "HyperVisors have been setup");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    config
# Objective: Launch subtest for all HyperVisors to perform config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config {
    my ($self, %args) = @_;

    $self->clear_error();

    # check to see if config is disabled for all hypervisors
    my $hypervisors = &ParamsMgr::search(-search_path => [],
                                         -key => "HYPERVISORS",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    if (not defined($hypervisors) or not defined($hypervisors->{CONFIG}) or $hypervisors->{CONFIG} == 0) {
        $self->sanlog(-msg => "Skipping HyperVisor Config.");
        return $SanMgr::TRUE;
    }

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $subtest = &SubIPC::create(-id => "$run_id_prefix" . "HyperVisor-config",
                                  -esc => $self->{parent}->{enable_subtest_count},
                                  -run => sub { $self->_config_all_hypervisors(); },
                                  -ipc => 1,
                                  -bg => 1);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor config has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::remove_subtest();

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _config_all_hypervisors
# Objective: Launch subtest for each HyperVisor type to perform config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _config_all_hypervisors {
    my ($self) = @_;

    $self->clear_error();

    # Get list of ESX hypervisors to run config against
    my $esx_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^ESX\d*$/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # Get list of HyperV hypervisors to run config against
    my $hyperv_hypervisors = &ParamsMgr::search(-search_path => [],
                                                -key => "HYPERVISORS",
                                                -key_filter => qr/^HYPERV\d*$/,
                                                paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # Get list of Citrix XS hypervisors to run config against
    my $cxs_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^CXS\d*$/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my @subtests = ();

    # run config on ESX
    if (scalar(keys %{$esx_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "ESX-config",
                                      -run => sub { &San::SanMgr::HyperVisor::ESX::config::config_all($self,
                                                                                                      -resources => $esx_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run config on HyperV
    if (scalar(keys %{$hyperv_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "HyperV-config",
                                      -run => sub { &San::SanMgr::HyperVisor::HyperV::config::config_all($self,
                                                                                                         -resources => $hyperv_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run config on CXS
    if (scalar(keys %{$cxs_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "CXS-config",
                                      -run => sub { &San::SanMgr::HyperVisor::CXS::config::config_all($self,
                                                                                                      -resources => $cxs_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "Skipping Hypervisors config because there are no hypervisors");
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor Config has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "HyperVisors have been configured");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    cleanup
# Objective: Launch subtest for each client to perform cleanup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ($self, %args) = @_;

    $self->clear_error();

    # check to see if cleanup is disabled for all hypervisors
    my $hypervisors = &ParamsMgr::search(-search_path => [],
                                         -key => "HYPERVISORS",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    if (not defined($hypervisors) or not defined($hypervisors->{CLEANUP}) or $hypervisors->{CLEANUP} == 0) {
        $self->sanlog(-msg => "Skipping HyperVisor Cleanup.");
        return $SanMgr::TRUE;
    }

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $subtest = &SubIPC::create(-id => "$run_id_prefix" . "HyperVisor-cleanup",
                                  -esc => $self->{parent}->{enable_subtest_count},
                                  -run => sub { $self->_cleanup_all_hypervisors(); },
                                  -ipc => 1,
                                  -bg => 1);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor cleanup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::remove_subtest();

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _cleanup_all_hypervisors
# Objective: Launch subtest for each hypervisor type to perform cleanup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cleanup_all_hypervisors {
    my ($self) = @_;

    $self->clear_error();

    # Get list of ESX hypervisors to run cleanup against
    my $esx_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^ESX\d*/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # Get list of HyperV hypervisors to run cleanup against
    my $hyperv_hypervisors = &ParamsMgr::search(-search_path => [],
                                                -key => "HYPERVISORS",
                                                -key_filter => qr/^HYPERV\d*/,
                                                paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # Get list of Citrix XS hypervisors to run cleanup against
    my $cxs_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^CXS\d*/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my @subtests = ();

    # run cleanup on ESX
    if (scalar(keys %{$esx_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "ESX-cleanup",
                                      -run => sub { &San::SanMgr::HyperVisor::ESX::cleanup::cleanup_all($self,
                                                                                                        -resources => $esx_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run cleanup on HyperV
    if (scalar(keys %{$hyperv_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "HyperV-cleanup",
                                      -run => sub { &San::SanMgr::HyperVisor::HyperV::cleanup::cleanup_all($self,
                                                                                                           -resources => $hyperv_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run cleanup on CXS
    if (scalar(keys %{$cxs_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "CXS-cleanup",
                                      -run => sub { &San::SanMgr::HyperVisor::CXS::cleanup::cleanup_all($self,
                                                                                                        -resources => $cxs_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "Skipping Hypervisors cleanup because there are no hypervisors");
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor Cleanup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "HyperVisors have been cleaned up");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    deconfig
# Objective: Launch subtest for hypervisor deconfig
# Details:   See POD documentation at the end of this file.
###############################################################################
sub deconfig {
    my ($self, %args) = @_;

    $self->clear_error();

    # check to see if deconfig is disabled for all hypervisors
    my $hypervisors = &ParamsMgr::search(-search_path => [],
                                         -key => "HYPERVISORS",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    if (not defined($hypervisors) or not defined($hypervisors->{DECONFIG}) or $hypervisors->{DECONFIG} == 0) {
        $self->sanlog(-msg => "Skipping HyperVisor Deconfig");
        return $SanMgr::TRUE;
    }

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $subtest = &SubIPC::create(-id => "$run_id_prefix" . "HyperVisor-deconfig",
                                  -esc => $self->{parent}->{enable_subtest_count},
                                  -run => sub { $self->_deconfig_all_hypervisors(); },
                                  -ipc => 1,
                                  -bg  => 1);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor deconfig has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::remove_subtest();

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _deconfig_all_hypervisors
# Objective: Launch subtest for each hypervisor to perform deconfiguration
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _deconfig_all_hypervisors {
    my ($self) = @_;

    $self->clear_error();

    # Get list of ESX hypervisors to run deconfig against
    my $esx_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^ESX\d*/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # Get list of HyperV hypervisors to run deconfig against
    my $hyperv_hypervisors = &ParamsMgr::search(-search_path => [],
                                                -key => "HYPERVISORS",
                                                -key_filter => qr/^HYPERV\d*/,
                                                paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # Get list of CXS hypervisors to run deconfig against
    my $cxs_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^CXS\d*/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my @subtests = ();

    # run deconfig on ESX
    if (scalar(keys %{$esx_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "ESX-deconfig",
                                      -run => sub { &San::SanMgr::HyperVisor::ESX::deconfig::deconfig_all($self,
                                                                                                          -resources => $esx_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run deconfig on HyperV
    if (scalar(keys %{$hyperv_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "HyperV-deconfig",
                                      -run => sub { &San::SanMgr::HyperVisor::HyperV::deconfig::deconfig_all($self,
                                                                                                             -resources => $hyperv_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run deconfig on CXS
    if (scalar(keys %{$cxs_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "CXS-deconfig",
                                      -run => sub { &San::SanMgr::HyperVisor::CXS::deconfig::deconfig_all($self,
                                                                                                          -resources => $cxs_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "Skipping Hypervisors deconfig because there are no hypervisors");
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor Deconfig has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "HyperVisors have been deconfigured");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    validate
# Objective: Launch subtest for hypervisor validate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self, %args) = @_;

    $self->clear_error();

    # check to see if validate is disabled for all hypervisors
    my $hypervisors = &ParamsMgr::search(-search_path => [],
                                         -key => "HYPERVISORS",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    if (not defined($hypervisors) or not defined($hypervisors->{VALIDATE}) or $hypervisors->{VALIDATE} == 0) {
        $self->sanlog(-msg => "Skipping HyperVisor Validation");
        return $SanMgr::TRUE;
    }

    &ParamsMgr::process_args(%args);
    my $run_id_prefix = &ParamsMgr::get_arg(-arg => "run_id_prefix",  -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    my $subtest = &SubIPC::create(-id => "$run_id_prefix" . "HyperVisor-validate",
                                  -esc => $self->{parent}->{enable_subtest_count},
                                  -run => sub { $self->_validate_all_hypervisors(); },
                                  -ipc => 1,
                                  -bg => 1);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => [$subtest]);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor validation has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::remove_subtest();

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _validate_all_hypervisors
# Objective: Launch subtest for each hypervisor to perform validation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _validate_all_hypervisors {
    my ($self) = @_;

    $self->clear_error();

    # Get list of ESX hypervisors to run validate against
    my $esx_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^ESX\d*/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # Get list of HyperV hypervisors to run validate against
    my $hyperv_hypervisors = &ParamsMgr::search(-search_path => [],
                                                -key => "HYPERVISORS",
                                                -key_filter => qr/^HYPERV\d*/,
                                                paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # Get list of CXS hypervisors to run validate against
    my $cxs_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^CXS\d*/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my @subtests = ();

    # run validate on ESX
    if (scalar(keys %{$esx_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "ESX-validate",
                                      -run => sub { &San::SanMgr::HyperVisor::ESX::validate::validate_all($self,
                                                                                                          -resources => $esx_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run validate on HyperV
    if (scalar(keys %{$hyperv_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "HyperV-validate",
                                      -run => sub { &San::SanMgr::HyperVisor::HyperV::validate::validate_all($self,
                                                                                                             -resources => $hyperv_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    # run validate on CXS
    if (scalar(keys %{$cxs_hypervisors})) {
        my $subtest = &SubIPC::create(-id => "CXS-validate",
                                      -run => sub { &San::SanMgr::HyperVisor::CXS::validate::validate_all($self,
                                                                                                          -resources => $cxs_hypervisors); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err(&SubIPC::get_error());
            &SubIPC::fail(-msg => $self->errmsg());
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (not scalar(@subtests)) {
        &SubIPC::pass(-msg => "Skipping Hypervisors validate because there are no hypervisors");
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperVisor Validation has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "HyperVisors have been validated");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    load_resource
# Objective: Gather list of hypervisors and create objects for each one
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_resource {
    my ($self, %args) = @_;

    $self->clear_error();

    my $staf = staf_global();

    my $environment = &ParamsMgr::search(search_path => [], key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # Get list of all hypervisors to create object for them
    my $all_hypervisors = &ParamsMgr::search(-search_path => [],
                                             -key => "HYPERVISORS",
                                             -key_filter => qr/^ESX\d*|^HYPERV\d*|^CXS\d*/,
                                             paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # make sure there are no duplicate hypervisors
    my %dup_hash;
    foreach my $hypervisor (keys(%{$all_hypervisors})) {
        $dup_hash{$all_hypervisors->{$hypervisor}->{NAME}}++;
        if ($dup_hash{$all_hypervisors->{$hypervisor}->{NAME}} > 1) {
            $self->_set_err("Detected duplicate hypervisor entries for the client " . $all_hypervisors->{$hypervisor}->{NAME});
            return $SanMgr::FALSE;
        }
    }

    # Go through the list of hypervisors and create an object for each one
    my @procs;
    foreach my $hypervisor_tag (keys(%{$all_hypervisors})) {
        # Spawn a new process to load this client
        my $proc = NATE::Process->new(codespec => '{ #EnGenX load' . $all_hypervisors->{$hypervisor_tag}->{NAME} . '
                                                     use San::SanMgr::HyperVisor; 
                                                     San::SanMgr::HyperVisor::_load_hypervisor(); }',
                                      runid => "hypervisor__" . $all_hypervisors->{$hypervisor_tag}->{NAME},
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result,
                                      client => 'localhost',
                                      params => { "hypervisors_ds" => $all_hypervisors,
                                                  "hypervisor" => $hypervisor_tag,
                                                  "environment" => $environment,
                                                  "staf_parent_handle" => $staf->handle(),
                                                  "staf_parent_endpoint" => NATE::Util::Sockets::get_address(Sys::Hostname::hostname())});
        push(@procs, $proc);
    }

    # set up the Event listener and callback methods
    my $filter = sub {
        my $msg = shift;
        return ($msg->{type} =~ /SanMgr\/HyperVisor\/SanClient\/\S+\/Load_Complete/);
    };
    my $callback = sub {
        my ($msg, $listener) = @_;

        # parse out the message information
        my ($hypervisor_name) = ($msg->{type} =~ /SanMgr\/HyperVisor\/SanClient\/(\S+?)\/Load_Complete/);
        my $sm_hypervisor_obj = $listener->{sm_hypervisor_obj};

        # rebuild the serialized object
        my $obj = eval($msg->{message}->{obj_serialized});
        if (not defined($obj)) {
            $sm_hypervisor_obj->_set_err("The SanClient object for hypervisor $hypervisor_name did not get transfered properly");
            $sm_hypervisor_obj->{parent}->{listeners_complete_count}++;
            return $SanMgr::FALSE;
        }
        $obj = SanClient->new(-obj_template => $obj);
        if ($obj->error()) {
            $sm_hypervisor_obj->_set_err("Failed to instantiate SanClient object from template of hypervisor $hypervisor_name",
                                         $obj->errmsg());
            $sm_hypervisor_obj->{parent}->{listeners_complete_count}++;
            return $SanMgr::FALSE;
        }

        # store the object on the sanmgr resource hash
        $sm_hypervisor_obj->{parent}->{_objects}->{hypervisor}->{$obj->hostname()} = $obj;

        # validate the md5 for the obj
        if (not $sm_hypervisor_obj->_validate_obj_md5_hash(obj => $obj,
                                                           obj_md5_hex_before => $msg->{message}->{obj_md5_hex})) {
            $sm_hypervisor_obj->{parent}->{listeners_complete_count}++;
            return $SanMgr::FALSE;
        }

        $sm_hypervisor_obj->{parent}->{listeners_complete_count}++;
        return $SanMgr::TRUE;
    };
    my $listener = {filter => $filter, callback => $callback, sm_hypervisor_obj => $self};

    # build the fail listener
    my $fail_filter = sub {
        my $msg = shift;
        return ($msg->{type} =~ /SanMgr\/HyperVisor\/SanClient\/\S+\/Load_Failed/);
    };
    my $fail_callback = sub {
        my ($msg, $flistener) = @_;
        my $sm_hypervisor_obj = $flistener->{sm_hypervisor_obj};

        # parse out the message information
        my ($hypervisor_name) = ($msg->{type} =~ /SanMgr\/HyperVisor\/SanClient\/(\S+?)\/Load_Failed/);
        $sm_hypervisor_obj->_set_err("The SanClient object for hypervisor $hypervisor_name failed to instantiate.");
        $sm_hypervisor_obj->{parent}->{listeners_complete_count}++;
        return $SanMgr::FALSE;
    };

    my $fail_listener = {filter => $fail_filter, callback => $fail_callback, sm_hypervisor_obj => $self};

    return ($listener, $fail_listener, @procs);
}

sub _load_hypervisor {
    # fetch params passed into the process
    my $global_params = NATE::ParamSet->get_global();
    my $all_hypervisors = $global_params->get("hypervisors_ds");
    my $hypervisor_tag = $global_params->get("hypervisor");
    my $environment = $global_params->get("environment");
    my $staf_parent_handle = $global_params->get("staf_parent_handle");
    my $staf_parent_endpoint = $global_params->get("staf_parent_endpoint");

    # Load up the resource information and get them via get_arg
    my $caller_id = $all_hypervisors->{$hypervisor_tag}->{NAME} . "|" . time();
    &ParamsMgr::process_args(%{$all_hypervisors->{$hypervisor_tag}}, -CALLER => $caller_id);

    #############################################
    ### General hypervisor parameters
    #############################################
    my $hypervisor_name = &ParamsMgr::get_arg(-arg => "NAME", -required => 1, -default_to_env => 1,
                                              -CALLER => $caller_id);
    my $conn_type =       &ParamsMgr::get_arg(-arg => "CONN_TYPE", -default => "THARNHOST", -default_to_env => 1,
                                              -CALLER => $caller_id);
    my $protocol =        &ParamsMgr::get_arg(-arg => "PROTOCOL", -required => 1, -default_to_env => 1,
                                              -CALLER => $caller_id);
    my $debug_level =     &ParamsMgr::get_arg(-arg => "DEBUG_LEVEL", -required => 0, -default_to_env => 1,
                                              -default => 0, -CALLER => $caller_id);
    my $vma_host =        &ParamsMgr::get_arg(-arg => "VMA_HOST", -required => 0, -default_to_env => 1,
                                              -default => 1, -CALLER => $caller_id);
    my $alua_support =    &ParamsMgr::get_arg(-arg => "ALUA_SUPPORT", -required => 0, -default_to_env => 1,
                                              -default => 1, -CALLER => $caller_id);

    #############################################
    ### VG parameters
    #############################################
    my $mp_type =         &ParamsMgr::get_arg(-arg => "MP_TYPE", -required => 0, -default_to_env => 1,
                                              -default => "", -CALLER => $caller_id);
    my $lvm_type =        &ParamsMgr::get_arg(-arg => "LVM_TYPE", -required => 0, -default_to_env => 1,
                                              -default => "", -CALLER => $caller_id);

    my $parent = NATE::STAF::Peer->new(handle => $staf_parent_handle,
                                       endpoint => $staf_parent_endpoint);

    if (&ParamsMgr::error()) {
        $parent->message_put(type => "SanMgr/HyperVisor/SanClient/" . $hypervisor_name . "/Load_Failed",
                             message => {});
        NATE::BaseException->throw("Failed to parse parameters for " . $all_hypervisors->{$hypervisor_tag}->{NAME} . " : " .
                                   &ParamsMgr::get_error());
        return;
    }

    # initialize the pnate attribute if it doesnt exist
    my $pnate = 0;
    if (defined($environment) and exists($environment->{PNATE})) {
        $pnate = $environment->{PNATE};
    }

    my $fake_self = {};
    bless($fake_self, "EnGenX");
	
    # generate a fake object with all the options built into it, so we
    # can pass in the iscsi type value into the sanclient constructor
    # needed for Linux & ESX
    my $obj = {};
    $fake_self->San::SanMgr::Util::assign_sanclient_options(-client_obj => $obj,
                                                            -datastruct => $all_hypervisors->{$hypervisor_tag},
                                                            -env_datastruct => $environment);
    # Create hypervisor SanClient object
    my $hypervisor_obj = SanClient->new(-host              => $hypervisor_name,
                                        -conn_type         => $conn_type,
                                        -debug_level       => $debug_level,
                                        -load_inits        => 1,
                                        -mp_type           => $mp_type,
                                        -lvm_type          => $lvm_type,
                                        -protocol          => $protocol,
                                        -pnate_environment => $pnate,
                                        -vma_host          => $vma_host,
                                        -alua_support      => $alua_support,
                                        -iscsi_type        => $obj->{iscsi}->{CONFIG_ISCSI_TYPE});
    if ($hypervisor_obj->error()) {
        $parent->message_put(type => "SanMgr/HyperVisor/SanClient/" . $hypervisor_name . "/Load_Failed",
                             message => {});
        NATE::BaseException->throw("Failed to create SanClient object for " . $all_hypervisors->{$hypervisor_tag}->{NAME} . " : " .
                                       $hypervisor_obj->errmsg());
        return;
    }

    # assign all iscsi_options & fcp_options to this sanclient object
    $fake_self->San::SanMgr::Util::assign_sanclient_options(-client_obj => $hypervisor_obj,
                                                            -datastruct => $all_hypervisors->{$hypervisor_tag},
                                                            -env_datastruct => $environment);

    # destroy the connection to the sanclient object,
    # we have to do this or else the object we send back to the parent gets messed up
    $hypervisor_obj->disconnect_connection();

    # calculate the md5 for the hypervisor obj
    my $obj_md5_hex = San::SanMgr::Util->_calculate_obj_md5_hash($hypervisor_obj);

    $parent->message_put(type => "SanMgr/HyperVisor/SanClient/" . $hypervisor_name . "/Load_Complete",
                         message => {obj_md5_hex => $obj_md5_hex,
                                     obj_serialized => Data::Dump::dump($hypervisor_obj)});

    return;
}

=pod

=cut

1;
