#
# Copyright (c) 2013 NetApp Inc.
# All rights reserved.
#
## @summary Citrix XS Cleanup module
## @author mathanku@netapp.com
## @status private
## @pod here

package San::SanMgr::HyperVisor::CXS::cleanup;

use San::SanMgr::Util;
use Data::Dumper;
use strict;
use warnings;
use Tharn;

###############################################################################
# Method:    cleanup_all
# Objective: Cleanup every Citrix XS host
###############################################################################
sub cleanup_all {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);

    my $cxs_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                        -ref_type => "HASH",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @subtests;

    # iterate through all CXS Hosts and issue cleanup_single
    foreach my $cxs_host (keys %{$cxs_hosts}) {
        my $cxs_host_name = $cxs_hosts->{$cxs_host}->{NAME};

        # skip cleanup on this host if told to skip
        $cxs_hosts->{$cxs_host}->{CLEANUP} = 1 if (not defined($cxs_hosts->{$cxs_host}->{CLEANUP}));
        if ($cxs_hosts->{$cxs_host}->{CLEANUP} == 0) {
            $self->sanlog(-msg => "Skipping CXS Cleanup on host '$cxs_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id => $cxs_host_name . '-cleanup',
                                      -run => sub { &cleanup_single($self,
                                                                    -resource => $cxs_hosts->{$cxs_host}); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to cleanup ${cxs_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No CXS HyperVisors to cleanup");
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests,
                         -sync_enable => 1);
    if (&SubIPC::error()) {
        $self->_set_err("CXS-cleanup failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "CXS-cleanup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All CXS HyperVisors have been cleanup.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    cleanup_single
# Objective: Cleanup a single CXS-V host
###############################################################################
sub cleanup_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $cxs_host = &ParamsMgr::get_arg(-arg => "resource",
                                       -ref_type => "HASH",
                                       -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $cxs_host_name = $cxs_host->{NAME};
    my $protocol = $cxs_host->{PROTOCOL};
    my $index = $cxs_host->{_INDEX};
    $index = "" if (not defined($index));

    my $virtual_machines = &ParamsMgr::search(-search_path => ["HYPERVISORS"],
                                              -key => "CXS" . $index,
                                              -key_filter => qr/^VM\d*$/,
                                              paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # obtain an array of all the virtual machine names
    my @vm_names = map { $virtual_machines->{$_}->{NAME} } (keys %{$virtual_machines});

    my ($cxs_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$cxs_host_name]);
    if ($self->error()) {
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # we need to obtain the list of cluster objects that this client is related to as well
    my $hash = $self->get_vserver_client_relationship_hash();
    if ($self->error()) {
        $self->_set_err("Failed to get cluster/client relationship hash for '$cxs_host_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }
    my @clusters = keys %{$hash->{$cxs_host_name}};

    # get list of vservers that this client will talk to
    my @vserver_list = $self->get_vservers_associated_to_client(-client_name => $cxs_host_name);
    if ($self->error()) {
        $self->_set_err("Failed to get vservers associated to '$cxs_host_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to this host
    $cxs_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($cxs_host_obj->error()) {
        $self->_set_err($cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # fetch the NACL cluster objects that this client is related to
    my @nacl_cluster_objs = $self->get_objects('type' => "node",
                                               'names' => \@clusters,
                                               'obj_class' => 'nacl');
    if ($self->error()) {
        $self->_set_err("Failed to fetch NACL Cluster objects");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # create the mpdevmgr (this will discover luns on the esx host)
    my $mpdevmgr_obj = $cxs_host_obj->create_mpdevmgr(-nacl_cluster_objs => \@nacl_cluster_objs,
                                                      -vservers => \@vserver_list);
    if ($cxs_host_obj->error()) {
        $self->_set_err($cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    foreach my $virtual_machine (values %{$virtual_machines}) {
        # get all the devices attached to this VM
        my @vdi_devices = $cxs_host_obj->get_devices_from_vm(-vm_name => $virtual_machine->{NAME});
        if ($cxs_host_obj->error()) {
            $self->_set_err("Unable to obtain the devices attched to the VM", $cxs_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        if (scalar(@vdi_devices) > 0) {
            $cxs_host_obj->remove_devices_from_vm(-vm_name => $virtual_machine->{NAME},
                                                  -devices => \@vdi_devices);
            if ($cxs_host_obj->error()) {
                $self->_set_err("Failed to remove the devices from VM", $cxs_host_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }
    }

    # remove empty storage repositories
    $cxs_host_obj->destroy_storage_repos();
    if ($cxs_host_obj->error()) {
        $self->_set_err("Failed to destroy SRs", $cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "Completed cleanup on CXS Host");

    return $SanMgr::TRUE;
}

1;
