#
# Copyright (c) 2013 NetApp Inc.
# All rights reserved.
#
## @summary CXS Config module
## @author mathanku@netapp.com
## @status private
## @pod here

package San::SanMgr::HyperVisor::CXS::config;

use San::SanMgr::Util;
use Data::Dumper;

###############################################################################
# Method:    config_all
# Objective: Config all Citrix XS Hosts
###############################################################################
sub config_all {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);

    my $cxs_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                        -ref_type => "HASH",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # iterate through all CXS Hosts and issue config_single
    foreach my $cxs_host (keys %{$cxs_hosts}) {
        my $cxs_host_name = $cxs_hosts->{$cxs_host}->{NAME};

        # skip config on this host if told to skip
        $cxs_hosts->{$cxs_host}->{CONFIG} = 1 if (not defined($cxs_hosts->{$cxs_host}->{CONFIG}));
        if ($cxs_hosts->{$cxs_host}->{CONFIG} == 0) {
            $self->sanlog(-msg => "Skipping CXS Config on host '$cxs_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id =>  $cxs_host_name . '-config',
                                      -run => sub { &config_single($self,
                                                                   -resource => $cxs_hosts->{$cxs_host}); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to config ${cxs_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No CXS HyperVisors to configure");
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("CXS-config failed", &SubIPC::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "CXS-config has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All CXS HyperVisors have been Configured.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    config_single
# Objective: Config a single CXS Host
###############################################################################
sub config_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource",
                                       -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $client_name = $resource->{NAME};

    # obtain the sanclient object
    my ($cxs_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$client_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get client ($client_name) object");
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # get the list of vserver targets that this initiator is associated with
    my ($data) = $self->get_target_list(-client_name => $client_name,
                                        -filter_iscsi_targets_by_portset => $cxs_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                        -filter_fcp_targets_by_portset   => $cxs_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->error()) {
        $self->_set_err("Failed to get iSCSI/FCP target lists");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    # reconnect to this host
    $cxs_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($cxs_host_obj->error()) {
        $self->_set_err($cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $protocol = $cxs_host_obj->get_protocol();

    # configure iSCSI/FCP
    if ($protocol =~ /iscsi|fcp/i) {
        $cxs_host_obj->configure(-targets => \@iSCSITargets);
        if ($cxs_host_obj->error()) {
            $self->_set_err("Failed to configure $protocol for the client ($client_name)", $cxs_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # configure NFS
    if ($protocol =~ /nfs/i) {
        # NFS currently has no configuration steps to perform
    }

    # Load up the resource information and get them via get_arg
    my $caller_id = $resource->{NAME} . "|" . time();
    &ParamsMgr::process_args(%{$resource}, -CALLER => $caller_id);

    # NTP sync the hypervisors
    my $ntp_sync =        &ParamsMgr::get_arg(-arg => "NTP_SYNC", -required => 0, -default_to_env => 1,
                                              -default => 0, -CALLER => $caller_id);
    my $ntp_time_server = &ParamsMgr::get_arg(-arg => "NTP_TIME_SERVER", -required => 1, -default_to_env => 1,
                                              -default => "ntp0.eng.netapp.com", dependant_on => ["NTP_SYNC"],
                                              -CALLER => $caller_id);
    my $ntp_time_zone =   &ParamsMgr::get_arg(-arg => "NTP_TIME_ZONE", -required => 1, -default_to_env => 1,
                                              -default => "EST5EDT", -dependant_on => ["NTP_SYNC"],
                                              -CALLER => $caller_id);
    if ($ntp_sync == 1) {
        $cxs_host_obj->ntp_sync(-ntp_server => $ntp_time_server,
                                -ntp_time_zone => $ntp_time_zone);
        if ($cxs_host_obj->error()) {
            my @error = $cxs_host_obj->errmsg();
            $self->sanlog(-msg => "Failed to NTP time sync the $client_name (@error))",
                          -msg_type => "warn");
            $cxs_host_obj->_clear_err();
        }
    }

    &SubIPC::pass(-msg => "CXS Host has been Configured.");

    return $SanMgr::TRUE;
}

1;
