#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Citrix XS Deconfig Module
## @author mathanku@netapp.com
## @status private
## @pod here

package San::SanMgr::HyperVisor::CXS::deconfig;

use San::SanMgr::Util;
use Data::Dumper;

###############################################################################
# Method:    deconfig_all
# Objective: Deconfig every CXS host
###############################################################################
sub deconfig_all {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);

    my $cxs_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                        -ref_type => "HASH",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # iterate through all CXS Hosts and issue deconfig_single
    foreach my $cxs_host (keys %{$cxs_hosts}) {
        my $cxs_host_name = $cxs_hosts->{$cxs_host}->{NAME};

        # skip deconfig on this host if told to skip
        $cxs_hosts->{$cxs_host}->{DECONFIG} = 1 if (not defined($cxs_hosts->{$cxs_host}->{DECONFIG}));
        if ($cxs_hosts->{$cxs_host}->{DECONFIG} == 0) {
            $self->sanlog(-msg => "Skipping CXS Deconfig on host '$cxs_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id =>  $cxs_host_name . '-deconfig',
                                      -run => sub { &deconfig_single($self,
                                                                     -resource => $cxs_hosts->{$cxs_host}); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to deconfig ${cxs_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No CXS HyperVisors to deconfigure");
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("CXS-deconfig failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "CXS-deconfig has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All CXS HyperVisors have been Deconfigured.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    deconfig_single
# Objective: Deconfig a single CXS host
###############################################################################
sub deconfig_single {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource",
                                       -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $client_name = $resource->{NAME};

    # get the sanclient object for this CXS host
    my ($cxs_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$client_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get client ($client_name) object");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to this host
    $cxs_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($cxs_host_obj->error()) {
        $self->_set_err($cxs_host_obj->errmsg());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # obtain the vserver targets that are associated with this host
    my ($data) = $self->get_target_list(-client_name => $client_name,
                                        -filter_iscsi_targets_by_portset => $cxs_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                        -filter_fcp_targets_by_portset   => $cxs_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->error()) {
        $self->_set_err("Failed to get iSCSI/FCP target lists");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    my $protocol = $cxs_host_obj->get_protocol();

    # deconfigure iSCSI/FCP
    if ($protocol =~ /iscsi|fcp/i) {
        $cxs_host_obj->deconfigure(-targets => \@iSCSITargets);
        if ($cxs_host_obj->error()) {
            $self->_set_err("Failed to deconfigure $protocol for the client ($client_name)", $cxs_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # deconfigure NFS
    if ($protocol =~ /nfs/i) {
        # NFS currently has no deconfiguration steps to perform
    }

    # clean up any orphaned devices
    if ($protocol =~ /iscsi|fcp/i) {
        $cxs_host_obj->orphaned_device_cleanup(-targets => \@iSCSITargets);
        if ($cxs_host_obj->error()) {
            $self->_set_err("Failed cleanup orphaned devices", $cxs_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    &SubIPC::pass(-msg => "CXS Host '$client_name' has been deconfigured.");

    return $SanMgr::TRUE;
}

1;
