#
# Copyright (c) 2013 NetApp Inc.
# All rights reserved.
#
## @summary Citrix XS Setup module
## @author mathanku@netapp.com
## @status private
## @pod here

package San::SanMgr::HyperVisor::CXS::setup;

use San::SanMgr::Util;
use Data::Dumper;
use Tharn;

###############################################################################
# Method:    setup_all
# Objective: Setup all CXS Hosts
###############################################################################
sub setup_all {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);

    my $cxs_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                        -ref_type => "HASH",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @subtests;
    my $balance_count = 0;

    # iterate through all CXS Hosts and issue setup_single
    foreach my $cxs_host (keys %{$cxs_hosts}) {
        my $cxs_host_name = $cxs_hosts->{$cxs_host}->{NAME};

        # skip setup on this host if told to skip
        $cxs_hosts->{$cxs_host}->{SETUP} = 1 if (not defined($cxs_hosts->{$cxs_host}->{SETUP}));
        if ($cxs_hosts->{$cxs_host}->{SETUP} == 0) {
            $self->sanlog(-msg => "Skipping CXS Setup on host '$cxs_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id => $cxs_host_name . '-setup',
                                      -run => sub { &setup_single($self,
                                                                  -resource => $cxs_hosts->{$cxs_host},
                                                                  -balance => $balance_count); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to setup ${cxs_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        $balance_count += 1;
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No CXS HyperVisors to setup");
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests,
                         -sync_enable => 1);
    if (&SubIPC::error()) {
        $self->_set_err("CXS-setup failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # parse the CXS CientDevList
    my @messages = &SubIPC::get_messages(-subtest => \@subtests,
                                         -type => "cxs_client_dev_list");
    if (&SubIPC::error()) {
        $self->_set_err("Failed to parse cxs client dev list messages", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # prepare the device list for all cxs hosts
    my $device_list;
    foreach my $message (@messages) {
        foreach my $device (@{$message->{extra}->{device_list}}) {
            push(@{$device_list->{$message->{extra}->{name}}->{DEV_LIST}}, $device);
        }
    }

    # send the final device list up to the parent
    &SubIPC::write(-subtest => [],
                   -msg => "CXS CDL",
                   -type => "cxs_client_dev_list",
                   -device_list => $device_list);
    if (&SubIPC::error()) {
        $self->_set_err("Failed to write CXS CDL to parents", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "CXS-setup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All CXS HyperVisors have been setup.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    setup_single
# Objective: Setup a single CXS Host
###############################################################################
sub setup_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $cxs_host = &ParamsMgr::get_arg(-arg => "resource",
                                       -ref_type => "HASH",
                                       -required => 1);
    my $balance = &ParamsMgr::get_arg(-arg => "balance",
                                      -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $cxs_host_name = $cxs_host->{NAME};

    my $protocol = $cxs_host->{PROTOCOL};
    my @protocols;
    if ($protocol =~ /iscsi/i) {
        push(@protocols, "iscsi");
    }
    if ($protocol =~ /fcp/i) {
        push(@protocols, "fcp");
    }

    my $vdi_lun_path_regex_filter = $cxs_host->{VDI_LUN_PATH_REGEX};
    if (defined($vdi_lun_path_regex_filter)) {
        $vdi_lun_path_regex_filter = eval($vdi_lun_path_regex_filter);
    } else {
        $vdi_lun_path_regex_filter = undef;
    }

    # get list of vservers that this client will talk to
    my @vserver_list = $self->get_vservers_associated_to_client(-client_name => $cxs_host_name);
    if ($self->error()) {
        $self->_set_err("Failed to get vservers associated to '$cxs_host_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $virtual_machines = &ParamsMgr::search(-search_path => ["HYPERVISORS"],
                                              -key => "CXS" . $index,
                                              -key_filter => qr/^VM\d*$/,
                                              paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err("Unable to search for VM keys", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @vm_names = map { $virtual_machines->{$_}->{NAME} } (keys %{$virtual_machines});

    my ($cxs_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$cxs_host_name]);
    if ($self->error()) {
        $self->_set_err("No SanClient objects found for CXS Hypervisor '$cxs_host_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to this host
    $cxs_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($cxs_host_obj->error()) {
        $self->_set_err("Unable to reconnect to CXS host '$cxs_host_name'", $cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # cleanly shutdown all virtual machines
    $cxs_host_obj->domain_management(-action => "shutdown",
                                     -vm_list => \@vm_names);
    if ($cxs_host_obj->error()) {
        $self->_set_err("Unable to power off VMs", $cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # get the list of vserver targets that this initiator is associated with
    my ($data) = $self->get_target_list(-client_name => $cxs_host_name,
                                        -filter_iscsi_targets_by_portset => $cxs_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                        -filter_fcp_targets_by_portset   => $cxs_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->error()) {
        $self->_set_err("Failed to get iSCSI/FCP target lists");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    # discover new luns
    $cxs_host_obj->discover(-targets => \@iSCSITargets);
    if ($cxs_host_obj->error()) {
        $self->_set_err("Unable to discover devices", $cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # we also need to obtain the list of cluster objects that this client is related to as well
    my $hash = $self->get_vserver_client_relationship_hash();
    if ($self->error()) {
        $self->_set_err("Failed to get cluster/client relationship hash for '$cxs_host_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @clusters = keys %{$hash->{$cxs_host_name}};

    # fetch the NACL cluster objects
    my @nacl_cluster_objs = $self->get_objects('type' => "node",
                                               'names' => \@clusters,
                                               'obj_class' => 'nacl');
    if ($self->error() or not scalar(@nacl_cluster_objs)) {
        $self->_set_err("Failed to fetch NACL Cluster objects for clusters (@clusters)");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # create mpdevmgr object
    my $cxs_dev_mgr = $cxs_host_obj->create_mpdevmgr(-nacl_cluster_objs => \@nacl_cluster_objs,
                                                     -vservers => \@vserver_list);
    if ($cxs_host_obj->error()) {
        $self->_set_err("Unable to create MPDev manager object", $cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @vdi_devices = $cxs_dev_mgr->get_filtered_devices(-vservers => \@vserver_list,
                                                         -device_index => $balance++,
                                                         -ontap_lun_ostype_regex => qr/linux/i,
                                                         -ontap_lun_path_regex => $vdi_lun_path_regex_filter,
                                                         -protocols => \@protocols,
                                                         -ontap_mode => "cmode",
                                                         -device_state => "free");
    if ($cxs_dev_mgr->error()) {
        $self->_set_err("Failed to obtain VDI devices", $cxs_dev_mgr->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $devices_list = "";
    foreach my $vdi_device (@vdi_devices) {
        $devices_list .= "  " . $vdi_device->{mp_device_name} . "\n";
    }

    $self->sanlog(-msg => "List of filtered devices:\n$devices_list");

    my $num_vdi_per_lun = defined($cxs_host->{VDIS_PER_LUN}) ? $cxs_host->{VDIS_PER_LUN} : 1;
    my ($total_requested_vdi_count) = &_verify_vdi_count($self,
                                                         $virtual_machines,
                                                         \@vdi_devices,
                                                         $num_vdi_per_lun);
    if ($self->error()) {
        $self->_set_err("Invalid VDI Counts");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @storage_repos = $cxs_host_obj->create_storage_repos(-devices => \@vdi_devices);
    if ($cxs_host_obj->error()) {
        $self->_set_err("Failed to create storage repositories", $cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @all_vdi_disks = $cxs_host_obj->create_vdi_disks(-storage_repos => \@storage_repos,
                                                        -num_vdi_per_lun => $num_vdi_per_lun);
    if ($cxs_host_obj->error()) {
        $self->_set_err("Failed to create VDIs", $cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    foreach my $virtual_machine (values %{$virtual_machines}) {
        my $virtual_machine_name = $virtual_machine->{NAME};
        my $vdi_count = defined($virtual_machine->{VIRTUAL_DISKS}) ? $virtual_machine->{VIRTUAL_DISKS} : 0;
        my @vdi_devices = ();

        # figure out the OStype of this virtual machine
        my $tharnhost_hsh = host($virtual_machine->{NAME});
        my $ostype = $tharnhost_hsh->{hosttype};
        $ostype = "linux" if ($ostype =~ /linux/i);
        $ostype = "windows" if ($ostype =~ /windows/i);
        if (not length($ostype) or $ostype !~ /windows|linux/i) {
            $self->_set_err("The hosttype '$ostype' of virtual machine '$virtual_machine_name' is invalid, only linux and windows hosttypes are supported ");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        for (my $i = 0; $i < $vdi_count; $i++) {
             push (@vdi_devices, shift(@all_vdi_disks));
        }

        # attach all devices to this guest
        if (scalar(@vdi_devices) > 0) {
            $self->sanlog(-msg => "Attaching " . @vdi_devices . " devices to ($virtual_machine_name)");

            $cxs_host_obj->add_devices_to_vm(-devices => \@vdi_devices,
                                             -vm_name => $virtual_machine_name);
            if ($cxs_host_obj->error()) {
                $self->_set_err("Failed to attach devices to VM ($virtual_machine_name)");
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        } else {
            $self->sanlog(-msg => "No devices requested to be attached to VM ($virtual_machine_name)");
        }
    }

    # turn all VMs back on
    $cxs_host_obj->domain_management(-action => "power_on",
                                     -vm_list => \@vm_names);
    if ($cxs_host_obj->error()) {
        $self->_set_err("Unable to power on VMs", $cxs_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "CXS HyperVisor has been setup.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _verify_vdi_count
# Objective: Calculate the total number of virtual disks (VDI) requested
###############################################################################
sub _verify_vdi_count {
    my ($self, $virtual_machines, $vdi_devices_lun_list, $num_vdi_per_lun) = @_;

    my $total_requested_vdi_count = 0;

    # calculate the total vdi disk request counts
    foreach my $vm (values %{$virtual_machines}) {
        $total_requested_vdi_count += $vm->{VIRTUAL_DISKS};
    }

    # calculate available vdi count
    my $actual_vdi_count = scalar(@$vdi_devices_lun_list) * $num_vdi_per_lun;

    if ($actual_vdi_count < $total_requested_vdi_count) {
        $self->_set_err("Requested '$total_requested_vdi_count' VDIs when only '$actual_vdi_count' VDIs are available");
    }

    return ($total_requested_vdi_count);
}

1;
