package San::SanMgr::HyperVisor::ESX::cleanup;

use San::SanMgr::Util;
use Data::Dumper;
use strict;
use warnings;
use Tharn;

sub cleanup_all {
    my ($self, %args) = @_;

    $self->clear_error();

    print Dumper(%args);

    &ParamsMgr::process_args(%args);

    my $esx_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                        -ref_type => "HASH",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @subtests;

    # iterate through all ESX Hosts and issue cleanup_single
    foreach my $esx_host (keys %{$esx_hosts}) {
        my $esx_host_name = $esx_hosts->{$esx_host}->{NAME};

        # skip cleanup on this host if told to skip
        $esx_hosts->{$esx_host}->{CLEANUP} = 1 if (not defined($esx_hosts->{$esx_host}->{CLEANUP}));
        if ($esx_hosts->{$esx_host}->{CLEANUP} == 0) {
            $self->sanlog(-msg => "Skipping ESX Cleanup on host '$esx_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id => $esx_host_name . '-cleanup',
                                      -run => sub { &cleanup_single($self,
                                                                  -resource => $esx_hosts->{$esx_host}); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to cleanup ${esx_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No ESX HyperVisors to cleanup");

        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests,
                         -sync_enable => 1);
    if (&SubIPC::error()) {
        $self->_set_err("ESX-cleanup failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("\&SubIPC::getsubtest_status failed.", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "ESX-cleanup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All ESX HyperVisors have been cleanup.");

    return $SanMgr::TRUE;
}

sub cleanup_single {
    my ($self, %args) = @_;

    print Dumper(%args);

    &ParamsMgr::process_args(%args);

    my $esx_host = &ParamsMgr::get_arg(-arg => "resource",
                                       -ref_type => "HASH",
                                       -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $esx_host_name = $esx_host->{NAME};
    my $protocol = $esx_host->{PROTOCOL};
    my $index = $esx_host->{_INDEX};
    $index = "" if (not defined($index));

    my $virtual_machines = &ParamsMgr::search(-search_path => ["HYPERVISORS"],
                                              -key => "ESX" . $index,
                                              -key_filter => qr/^VM\d*$/,
                                              paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # we need to obtain the list of cluster objects that this client is related to as well
    # TODO: consider getting rid of this
    my $hash = $self->{parent}->get_vserver_client_relationship_hash();
    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get cluster/client relationship hash for '$esx_host_name'");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }
    my @clusters = keys %{$hash->{$esx_host_name}};

    # get list of vservers that this client will talk to
    my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $esx_host_name);
    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get vservers associated to '$esx_host_name'");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    # obtain an array of all the virtual machine names
    my @vm_names = map { $virtual_machines->{$_}->{NAME} } (keys %{$virtual_machines});

    my ($esx_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$esx_host_name]);
    if ($self->error()) {
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to this host
    $esx_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($esx_host_obj->error()) {
        $self->_set_err($esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    if (scalar(@vm_names)) {
        # power off all the virtual machines on this host
        $esx_host_obj->power_management(-action => "power_off",
                                        -vm_list => \@vm_names);
        if ($esx_host_obj->error()) {
            $self->_set_err($esx_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    my $vmfs_lvm_obj = $esx_host_obj->create_lvm(-lvm => 'vmfs');
    if ($esx_host_obj->error()) {
        $self->_set_err($esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $rdm_lvm_obj = $esx_host_obj->create_lvm(-lvm => 'rdm');
    if ($esx_host_obj->error()) {
        $self->_set_err($esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # fetch the NACL cluster objects that this client is related to
    my @nacl_cluster_objs = $self->get_objects('type' => "node",
                                              'names' => \@clusters,
                                              'obj_class' => 'nacl');
    if ($self->error()) {
        $self->_set_err("Failed to fetch NACL Cluster objects");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # create the mpdevmgr (this will discover luns on the esx host)
    my $mpdevmgr_obj = $esx_host_obj->create_mpdevmgr(-nacl_cluster_objs => \@nacl_cluster_objs,
                                                      -vservers => \@vserver_list);
    if ($esx_host_obj->error()) {
        $self->_set_err($esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    if (scalar(@vm_names)) {
        my (@subtests, @all_subtests);

        # spawn off subtests to add all vmdks to each guest
        my $parallel_limit = $esx_host->{VMDK_ADD_REMOVE_PARALLEL_LIMIT};

        foreach my $virtual_machine (values %{$virtual_machines}) {
            my $virtual_machine_name = $virtual_machine->{NAME};

            my $cleanup_vm_subroutine = sub {
                # open a new connection to the esx client or else they would interfere with eachother
                $esx_host_obj->reconnect(-disconnect_existing_conn => 0);
                if ($esx_host_obj->error()) {
                    $self->_set_err($esx_host_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                # destroy all the vmdks (rdms/vmfs) associated to this virtual machine
                $vmfs_lvm_obj->destroy_vmdks(-vm_name => $virtual_machine_name);
                if ($vmfs_lvm_obj->error()) {
                    $self->_set_err($vmfs_lvm_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                $rdm_lvm_obj->destroy_vmdks(-vm_name => $virtual_machine_name);
                if ($rdm_lvm_obj->error()) {
                    $self->_set_err($rdm_lvm_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                # cleanup this vm's config file
                $esx_host_obj->cleanup_vm_config_file(-vm_list => [$virtual_machine_name],
                                                      -vmdk_filepath_regex_filter => qr/nate_rdm_|nate_datastore_/i);
                if ($esx_host_obj->error()) {
                    $self->_set_err($esx_host_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                &Tharn::disconnect($esx_host_obj->{conn}, '-ok');

                &SubIPC::pass(-msg => "Removed VMDKs from $virtual_machine_name");
            };

            # wait for subtests to finish before we spawn off the next batch of subtests
            if ($parallel_limit !~ /unlimited/i and
                scalar(@subtests) >= $parallel_limit and
                $esx_host_obj->has_vma_host()) {
                # wait for these subtests to finish
                &SubIPC::wait_finish(-subtest => \@subtests);
                if (&SubIPC::error()) {
                    $self->_set_err("Failed to remove VMDKs to virtual machines", &SubIPC::get_error());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                # reset the subtest list
                @subtests = ();
            }

            my $subtest = &SubIPC::create(-id => $virtual_machine_name . '-remove_vmdks',
                                          -run => $cleanup_vm_subroutine,
                                          -parent => $self,
                                          -ipc => 1,
                                          -bg => 1);
            if (&SubIPC::error()) {
                $self->_set_err("Failed to instantiate a subtest to add vmdks to $virtual_machine_name",
                                &SubIPC::get_error());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            # push onto the working subtest list
            push(@subtests, $subtest);

            # keep track of all the subtests as well
            push(@all_subtests, $subtest);
        }

        # if there are any subtests left in the list, reap them
        if (scalar(@subtests)) {
            &SubIPC::wait_finish(-subtest => \@subtests);
            if (&SubIPC::error()) {
                $self->_set_err("Failed to remove vmdks from virtual machines", &SubIPC::get_error());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }

        my $st_status = &SubIPC::get_subtest_status(-subtest => \@all_subtests);
        if (&SubIPC::error()) {
            $self->_set_err("\&SubIPC::getsubtest_status failed.", &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        if ($st_status->{failed} != 0) {
            $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # turn on all virtual machines
        $esx_host_obj->power_management(-action => "power_on",
                                        -vm_list => \@vm_names);

        if ($esx_host_obj->error()) {
            $self->_set_err($esx_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # destroy datastores on this esx host
    $vmfs_lvm_obj->destroy_datastores(-hostname => $esx_host_name,
                                      -delete_non_empty_datastores => $esx_host->{DELETE_NON_EMPTY_DATASTORES});
    if ($vmfs_lvm_obj->error()) {
        $self->_set_err($vmfs_lvm_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # sync
    &SubIPC::sync();

    # offline devices that we are cleaning up
    my @mpdev_objs = $mpdevmgr_obj->get_filtered_devices(-vservers => \@vserver_list);
    if ($mpdevmgr_obj->error()) {
        # ignore error if there are no devices to offline
        $mpdevmgr_obj->_clear_err();
    } else {
        # offline the devices that were returned
        $mpdevmgr_obj->offline_devices(-mpdev_objs => \@mpdev_objs);
        if ($mpdevmgr_obj->error()) {
            $self->_set_err("Failed to offline devices", $mpdevmgr_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # sync
    &SubIPC::sync();

    # refresh each esx host
    $esx_host_obj->restart_mgmt_vmware_service();
    if ($esx_host_obj->error()) {
        $self->_set_err($esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # sync
    &SubIPC::sync();

    sleep 30;

    $esx_host_obj->vmfs_rescan();
    if ($esx_host_obj->error()) {
        $self->_set_err($esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # done

    &SubIPC::pass(-msg => "ESX Host has been cleaned up.");

    return $SanMgr::TRUE;
}

1;
