package San::SanMgr::HyperVisor::ESX::deconfig;


use San::SanMgr::Util;
use Data::Dumper;

sub deconfig_all {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);

    my $esx_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                        -ref_type => "HASH",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # iterate through all ESX Hosts and issue deconfig_single
    foreach my $esx_host (keys %{$esx_hosts}) {
        my $esx_host_name = $esx_hosts->{$esx_host}->{NAME};

        # skip deconfig on this host if told to skip
        $esx_hosts->{$esx_host}->{DECONFIG} = 1 if (not defined($esx_hosts->{$esx_host}->{DECONFIG}));
        if ($esx_hosts->{$esx_host}->{DECONFIG} == 0) {
            $self->sanlog(-msg => "Skipping ESX Deconfig on host '$esx_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id =>  $esx_host_name . '-deconfig',
                                      -run => sub { &deconfig_single($self,
                                                                   -resource => $esx_hosts->{$esx_host}); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to deconfig ${esx_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No ESX HyperVisors to deconfigure");

        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("ESX-deconfig failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "ESX-deconfig has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All ESX HyperVisors have been Deconfigured.");

    return $SanMgr::TRUE;
}

sub deconfig_single {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource",
                                       -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $client_name = $resource->{NAME};

    # get the sanclient object for this esx host
    my ($esx_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$client_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get client ($client_name) object");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to this host
    $esx_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($esx_host_obj->error()) {
        $self->_set_err($esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # obtain the vserver targets that are associated with this host
    my ($data) = $self->{parent}->get_target_list(-client_name => $client_name,
                                                  -filter_iscsi_targets_by_portset => $esx_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                  -filter_fcp_targets_by_portset   => $esx_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get iSCSI/FCP target lists");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    my $protocol = $esx_host_obj->get_protocol();

    # deconfigure iSCSI/FCP
    if ($protocol =~ /iscsi|fcp/i and scalar(@iSCSITargets)) {
        $esx_host_obj->deconfigure(-iscsi_targets => \@iSCSITargets,
                                   -clean_level => $esx_host_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($esx_host_obj->error()) {
            $self->_set_err("Failed to deconfigure $protocol for the client ($client_name)", $esx_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # deconfigure NFS
    if ($protocol =~ /nfs/i) {
        # NFS currently has no deconfiguration steps to perform
    }

    # clean up any orphaned devices
    $esx_host_obj->orphaned_device_cleanup();
    if ($esx_host_obj->error()) {
        $self->_set_err("Failed cleanup orphaned devices", $esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # restart the host agent
    $esx_host_obj->restart_mgmt_vmware_service();
    if ($esx_host_obj->error()) {
        $self->_set_err("Failed to restart mgmt-vmware service", $esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "ESX Host '$client_name' has been deconfigured.");

    return $SanMgr::TRUE;
}

1;
