package San::SanMgr::HyperVisor::ESX::setup;

use San::SanMgr::Util;
use Data::Dumper;
use strict;
use warnings;
use Tharn;

###############################################################################
# Method:    setup_all
# Objective: Setup all ESX Hosts
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup_all {
    my ($self, %args) = @_;

    $self->clear_error();

    print Dumper(%args);

    &ParamsMgr::process_args(%args);

    my $esx_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                        -ref_type => "HASH",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @subtests;
    my $balance_count = 0;

    # iterate through all ESX Hosts and issue setup_single
    foreach my $esx_host (keys %{$esx_hosts}) {
        my $esx_host_name = $esx_hosts->{$esx_host}->{NAME};

        # skip setup on this host if told to skip
        $esx_hosts->{$esx_host}->{SETUP} = 1 if (not defined($esx_hosts->{$esx_host}->{SETUP}));
        if ($esx_hosts->{$esx_host}->{SETUP} == 0) {
            $self->sanlog(-msg => "Skipping ESX Setup on host '$esx_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id => $esx_host_name . '-setup',
                                      -run => sub { &setup_single($self,
                                                                  -resource => $esx_hosts->{$esx_host},
                                                                  -balance => $balance_count); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to setup ${esx_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        $balance_count++;
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No ESX HyperVisors to setup");

        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests,
                         -sync_enable => 1);
    if (&SubIPC::error()) {
        $self->_set_err("ESX-setup failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "ESX-setup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All ESX HyperVisors have been setup.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    setup_single
# Objective: Setup a single ESX Host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup_single {
    my ($self, %args) = @_;

    print Dumper(%args);

    # given a single hypervisor, setup this hypervisor

    &ParamsMgr::process_args(%args);

    my $esx_host = &ParamsMgr::get_arg(-arg => "resource",
                                       -ref_type => "HASH",
                                       -required => 1);
    my $balance = &ParamsMgr::get_arg(-arg => "balance",
                                      -default => 0);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $esx_host_name = $esx_host->{NAME};
    my $protocol = $esx_host->{PROTOCOL};
    my @protocols;
    if ($protocol =~ /iscsi/i) {
        push(@protocols, "iscsi");
    }
    if ($protocol =~ /fcp/i) {
        push(@protocols, "fcp");
    }

    my $index = $esx_host->{_INDEX};
    $index = "" if (not defined($index));
    my $num_vmdks_per_datastore = $esx_host->{VMDKS_PER_DATASTORE};
    $num_vmdks_per_datastore = 1 if (not defined($num_vmdks_per_datastore));

    my $scsi_bus = $esx_host->{SCSI_BUS};
    $scsi_bus = "lsilogic" if (not defined($scsi_bus));

    my $vmfs_datastore_percentage = $esx_host->{VMFS_DATASTORE_PERCENTAGE};
    $vmfs_datastore_percentage = "100" if (not defined($vmfs_datastore_percentage));

    my $vmfs_lun_path_regex_filter = $esx_host->{VMFS_LUN_PATH_REGEX};
    if (defined($vmfs_lun_path_regex_filter)) {
        $vmfs_lun_path_regex_filter = eval($vmfs_lun_path_regex_filter);
    } else {
        $vmfs_lun_path_regex_filter = undef;
    }

    # This filter controls which VNFS Datastores have VMDKs created in them and which dont
    my $vmfs_vmdk_lun_path_regex_filter = $esx_host->{VMFS_VMDK_LUN_PATH_REGEX};
    if (defined($vmfs_vmdk_lun_path_regex_filter)) {
        $vmfs_vmdk_lun_path_regex_filter = eval($vmfs_vmdk_lun_path_regex_filter);
    } else {
        $vmfs_vmdk_lun_path_regex_filter = undef;
    }

    # get list of vservers that this client will talk to
    my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $esx_host_name);
    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get vservers associated to '$esx_host_name'");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    # we also need to obtain the list of cluster objects that this client is related to as well
    my $hash = $self->{parent}->get_vserver_client_relationship_hash();
    if ($self->{parent}->error()) {
        $self->{parent}->_clear_err();
        $self->_set_err("Failed to get cluster/client relationship hash for '$esx_host_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @clusters = keys %{$hash->{$esx_host_name}};

    my $virtual_machines = &ParamsMgr::search(-search_path => ["HYPERVISORS"],
                                              -key => "ESX" . $index,
                                              -key_filter => qr/^VM\d*$/,
                                              paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err("Unable to search for VM keys", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @vm_names = map { $virtual_machines->{$_}->{NAME} } (keys %{$virtual_machines});

    my ($esx_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$esx_host_name]);
    if ($self->error()) {
        $self->_set_err("No SanClient objects foun for ESX Hypervisor '$esx_host_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to this host
    $esx_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($esx_host_obj->error()) {
        $self->_set_err("Unable to reconnect to ESX host '$esx_host_name'", $esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # default VMFS type to vmfs3 (ESX4.*) or vmfs5 (ESXi5.* later)
    my $vmfs_type = $esx_host->{VMFS_TYPE};
    if (not length($vmfs_type)) {
        if ($esx_host_obj->get_os_version() =~ /^(5|6)\./) {
            $vmfs_type = "vmfs5";
        } else {
            $vmfs_type = "vmfs3";
        }
    }

    if (scalar(@vm_names)) {
        # power off all the virtual machines on this host
        $esx_host_obj->power_management(-action => "power_off",
                                        -vm_list => \@vm_names);

        if ($esx_host_obj->error()) {
            $self->_set_err("Unable to power off VMs '@vm_names'", $esx_host_obj->errmsg());
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # rotate syslogs
    $esx_host_obj->rotate_logs();
    if ($esx_host_obj->error()) {
        $self->_set_err("Unable to rotate logs for ESX host '$esx_host_name'", $esx_host_obj->errmsg());
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # issue discover()
    $esx_host_obj->discover(-num_expected_luns => 1);
    if ($esx_host_obj->error()) {
        $self->_set_err("Unable to discover LUNs for ESX host '$esx_host_name'", $esx_host_obj->errmsg());
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # fetch the NACL cluster objects
    my @nacl_cluster_objs = $self->get_objects('type' => "node",
                                              'names' => \@clusters,
                                              'obj_class' => 'nacl');
    if ($self->error() or not scalar(@nacl_cluster_objs)) {
        $self->_set_err("Failed to fetch NACL Cluster objects for clusters (@clusters)");
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # create an MPDevMgr (this will find all LUNs on the system)
    my $esx_dev_mgr = $esx_host_obj->create_mpdevmgr(-nacl_cluster_objs => \@nacl_cluster_objs,
                                                     -vservers => \@vserver_list);
    if ($esx_host_obj->error()) {
        $self->_set_err("Failed to create MPDevMgr", $esx_host_obj->errmsg());
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # perform os specific validation on the discovered luns
    my ($os_specific_validation_errors, $os_specific_validation_warnings) = $esx_dev_mgr->os_specific_validation();
    if ($esx_dev_mgr->error()) {
        $self->_set_err("Failed to perform os specific validation", $esx_dev_mgr->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }
    if (scalar(@{$os_specific_validation_warnings})) {
        &SanUtil::nate_log('-warn', "Detected SAN LUN Configuration warnings on $esx_host_name:\n\t" .
                               join("\n", @{$os_specific_validation_warnings}));
    }
    if (scalar(@{$os_specific_validation_errors})) {
        $self->_set_err("Detected SAN LUN Configuration errors on $esx_host_name:\n\t" .
                            join("\n", @{$os_specific_validation_errors}));
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # verify that there are enough vmfs/rdm luns to go around
    my ($rdm_vmdk_requested_total, $vmfs_vmdk_requested_total) =
        &_verify_vmfs_rdm_counts($self,
                                 $esx_dev_mgr,
                                 $vmfs_lun_path_regex_filter,
                                 $virtual_machines,
                                 $num_vmdks_per_datastore,
                                 \@vserver_list);
    if ($self->error()) {
        $self->_set_err("Invalid VMFS/RDM Counts");
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # setup esx psp policies
    if (defined($esx_host->{PSP_LIST}) and $esx_host->{PSP_LIST} ne "") {
        &_set_psp_policies($self, $esx_host_obj, $esx_host, $balance);
        if ($self->error()) {
            $self->_set_err("Failed to set PSP policies on '$esx_host_name'");
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # setup VxDMP IO policies if the object has a vx_io_policy established
    if ($esx_host_obj->{vx_io_policy}) {
        &_set_vxdmp_io_policies($self, $esx_host_obj, $esx_host);
        if ($self->error()) {
            $self->_set_err("Failed to set VxDMP IO policies on '$esx_host_name'");
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # configure qfull settings after all luns have been discovered
    if ($esx_host->{CONFIGURE_QFULL}) {
        $esx_host_obj->qfull_configure();
        if ($esx_host_obj->error()) {
            $self->_set_err("Failed to configure qfull for host $esx_host_name", $esx_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    my @all_vmfs_vmdks;

    my $vmfs_lvm_obj = $esx_host_obj->create_lvm(-lvm => 'vmfs');
    if ($esx_host_obj->error()) {
        $self->_set_err("Failed to create VMFS LVM object for host '$esx_host_name'", $esx_host_obj->errmsg());
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $rdm_lvm_obj = $esx_host_obj->create_lvm(-lvm => "rdm");
    if ($esx_host_obj->error()) {
        $self->_set_err("Failed to create RDM LVM object for host '$esx_host_name'", $esx_host_obj->errmsg());
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    if (scalar(@vm_names) and $vmfs_vmdk_requested_total > 0) {
        # create datastores & vmfs vmdks
        # determine number of datastores to create

        my @vmfs_devices = $esx_dev_mgr->get_filtered_devices(-vservers => \@vserver_list,
                                                              -device_index => $balance++,
                                                              -ontap_lun_ostype_regex => qr/vmware/i,
                                                              -ontap_lun_path_regex => $vmfs_lun_path_regex_filter,
                                                              -protocols => \@protocols,
                                                              -ontap_mode => "cmode",
                                                              -device_state => "free");
        if ($esx_dev_mgr->error()) {
            $self->_set_err("Failed to obtain VMFS devices", $esx_dev_mgr->errmsg());
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        my @datastores = $vmfs_lvm_obj->create_datastores(-vmfs_devices => \@vmfs_devices,
                                                          -num_luns_per_dstore => 1,
                                                          -datastore_type => "one-to-one",
                                                          -vmfs_type => $vmfs_type,
                                                          -block_size => "auto",
                                                          -hostname => $esx_host_name);
        if ($vmfs_lvm_obj->error()) {
            $self->_set_err("Failed to create VMFS datastores", $vmfs_lvm_obj->errmsg());
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # filter out the vmfs_devices that we dont want to create VMDKs on
        # using $vmfs_vmdk_lun_path_regex_filter
        foreach my $vmfs_device (@vmfs_devices) {
            # load the current state of this vmfs device
            $vmfs_device->load();

            # get the name of the datastore on this device
            my $datastore_label = $vmfs_device->get_os_specific_data("datastore_label");

            # determine the lun path
            my $lun_path = $vmfs_device->ontap_lun_path();
            if ($lun_path !~ $vmfs_vmdk_lun_path_regex_filter) {
                # remove this datastore_label from @datastores
                @datastores = grep {
                    $_ !~ /^$datastore_label$/;
                } @datastores;
            }
        }

        my $num_vmfs_vmdks_total;
        if ($num_vmdks_per_datastore <= 0) {
            $num_vmfs_vmdks_total = scalar(@datastores);
        } else {
            $num_vmfs_vmdks_total = scalar(@datastores) * $num_vmdks_per_datastore;
        }

        my $vmfs_vmdk_disk_type = $esx_host->{VMFS_VMDK_DISK_TYPE};

        @all_vmfs_vmdks = $vmfs_lvm_obj->create_vmdks(-datastores => \@datastores,
                                                      -controller => $scsi_bus,
                                                      -datastore_type => "one-to-one",
                                                      -num_vmfs_vmdks => $num_vmfs_vmdks_total,
                                                      -datastore_usage_percentage => $vmfs_datastore_percentage,
                                                      -vmfs_vmdk_disk_type => $vmfs_vmdk_disk_type,
                                                      -reuse => 1 # reuse vmdks that may already exist in datastores
                                                  );
        if ($vmfs_lvm_obj->error()) {
            $self->_set_err("Failed to create VMFS VMDKs", $vmfs_lvm_obj->errmsg());
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    my (@subtests, @all_subtests);

    # create rdms vmdks for each guest,
    # spawn off subtests to add all vmdks to each guest
    my $parallel_limit = $esx_host->{VMDK_ADD_REMOVE_PARALLEL_LIMIT};

    foreach my $virtual_machine (values %{$virtual_machines}) {
        my $virtual_machine_name = $virtual_machine->{NAME};
        my $vmfs_count = defined($virtual_machine->{VIRTUAL_DISKS}) ? $virtual_machine->{VIRTUAL_DISKS} : 0;
        my $rdm_count = defined($virtual_machine->{RAW_DISKS}) ? $virtual_machine->{RAW_DISKS} : 0;
        my $vm_scsi_bus = $scsi_bus;
        if (defined($virtual_machine->{SCSI_BUS}) and $virtual_machine->{SCSI_BUS} !~ /^\s*$/) {
            $vm_scsi_bus = $virtual_machine->{SCSI_BUS};
        }

        my $rdm_lun_path_regex_filter = $virtual_machine->{RAW_LUN_PATH_REGEX};
        if (defined($rdm_lun_path_regex_filter)) {
            $rdm_lun_path_regex_filter = eval($rdm_lun_path_regex_filter);
        } else {
            $rdm_lun_path_regex_filter = undef;
        }

        my @rdm_vmdks_for_this_guest;
        my @vmfs_vmdks_for_this_guest;

        for (my $i = 0; $i < $vmfs_count; $i++) {
            push(@vmfs_vmdks_for_this_guest, shift(@all_vmfs_vmdks));
        }

        my @rdm_devices;

        if ($rdm_count > 0) {
            # we need to parse out the hosttype for this virtual machine
            my $hosttype_hsh = host($virtual_machine_name);
            my $hosttype_hostrec = $hosttype_hsh->{hosttype};
            my $ostype;
            # need to convert linux host type to 'linux'
            if ($hosttype_hostrec =~ /linux/) {
                $ostype = "linux";
            } elsif ($hosttype_hostrec =~ /solaris/i) {
                $ostype = "solaris";
            } else {
                $ostype = $hosttype_hostrec;
            }

            # get a list of RDM devices that match these specifications
            @rdm_devices = $esx_dev_mgr->get_filtered_devices(-vservers => \@vserver_list,
                                                              -device_index => $balance++,
                                                              -protocols => \@protocols,
                                                              -ontap_mode => "cmode",
                                                              -ontap_lun_path_regex => $rdm_lun_path_regex_filter,
                                                              -device_count => $rdm_count,
                                                              -ontap_lun_ostype_regex => qr/$ostype/i,
                                                              -device_state => "free");
            if ($esx_dev_mgr->error()) {
                $self->_set_err("Failed to get RDM devices for $virtual_machine_name", $esx_dev_mgr->errmsg());
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }

        my $add_vmdks_to_vm_subroutine = sub {
            # open a new connection to the esx client or else they would interfere with eachother
            $esx_host_obj->reconnect(-disconnect_existing_conn => 0);
            if ($esx_host_obj->error()) {
                $self->_set_err($esx_host_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            # create the RDM vmdks if there are any devices in the array
            if (scalar(@rdm_devices) > 0) {
                @rdm_vmdks_for_this_guest = $rdm_lvm_obj->create_vmdks(-rdm_devices => \@rdm_devices,
                                                                       -controller => $vm_scsi_bus =~ /pvscsi|lsisas1068/ ? "lsilogic" : $vm_scsi_bus,
                                                                       -rdm_type => "physical",
                                                                       -vm_name => $virtual_machine_name);
                if ($rdm_lvm_obj->error()) {
                    $self->_set_err("Failed to create RDM VMDKs for $virtual_machine_name", $rdm_lvm_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }
            }

            # apply all vmdks to this guest's config file
            $esx_host_obj->add_vmdks_to_vm(-vm_name => $virtual_machine_name,
                                           -vmfs_vmdk_files => \@vmfs_vmdks_for_this_guest,
                                           -rdm_vmdk_files => \@rdm_vmdks_for_this_guest,
                                           -controller => $vm_scsi_bus,
                                           -preclean => 1);
            if ($esx_host_obj->error()) {
                $self->_set_err("Failed to add VMDKs to $virtual_machine_name", $esx_host_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            &Tharn::disconnect($esx_host_obj->{conn}, '-ok');

            &SubIPC::pass(-msg => "Added VMDKs to $virtual_machine_name");
        };

        # wait for subtests to finish before we spawn off the next batch of subtests
        if ($parallel_limit !~ /unlimited/i and
            scalar(@subtests) >= $parallel_limit and
            $esx_host_obj->has_vma_host()) {
            # wait for these subtests to finish
            &SubIPC::wait_finish(-subtest => \@subtests);
            if (&SubIPC::error()) {
                $self->_set_err("Failed to add VMDKs to virtual machines", &SubIPC::get_error());
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            # reset the subtest list
            @subtests = ();
        }

        my $subtest = &SubIPC::create(-id => $virtual_machine_name . '-add_vmdks',
                                      -run => $add_vmdks_to_vm_subroutine,
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to add vmdks to $virtual_machine_name",
                            &SubIPC::get_error());
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # push onto the working subtest list
        push(@subtests, $subtest);

        # keep track of all the subtests as well
        push(@all_subtests, $subtest);
    }

    # remove spare VMFS vmdks not in use, ie: all elements left in @all_vmfs_vmdks
    if (defined($vmfs_lvm_obj) and $vmfs_vmdk_requested_total > 0) {
        $vmfs_lvm_obj->destroy_spare_vmdks(-vmfs_vmdk_files => \@all_vmfs_vmdks);
        if ($vmfs_lvm_obj->error()) {
            $self->_set_err("Failed to destroy spare VMFS VMDKs", $vmfs_lvm_obj->errmsg());
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    if (scalar(@vm_names)) {
        # if there are any subtests still in the list, take care of them now
        if (scalar(@subtests)) {
            &SubIPC::wait_finish(-subtest => \@subtests);
            if (&SubIPC::error()) {
                $self->_set_err("Failed to add VMDKs to virtual machines", &SubIPC::get_error());
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }

        # see which ones failed
        my $st_status = &SubIPC::get_subtest_status(-subtest => \@all_subtests);
        if ($st_status->{failed} != 0) {
            $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # turn on all virtual machines
        $esx_host_obj->power_management(-action => "power_on",
                                        -vm_list => \@vm_names);
        if ($esx_host_obj->error()) {
            $self->_set_err("Failed to power on VMs '@vm_names'", $esx_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # sync
    &SubIPC::sync();

    # refresh each esx host
    $esx_host_obj->restart_mgmt_vmware_service();
    if ($esx_host_obj->error()) {
        $self->_set_err("Failed to restart mgmt-vmware service", $esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # sync
    &SubIPC::sync();

    sleep 30;

    # rescan vmfs datastores
    $esx_host_obj->vmfs_rescan();
    if ($esx_host_obj->error()) {
        $self->_set_err("Failed to rescan VMFS datastores", $esx_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # done
    &SubIPC::pass(-msg => "ESX Host has been Setup.");

    return $SanMgr::TRUE;
}

sub _set_psp_policies {
    my ($self, $esx_client_obj, $esx_host_hash, $balance) = @_;

    # only operate on 4.0/5.0 systems
    if ($esx_client_obj->get_os_version() !~ /^(4|5|6)\./) {
        return;
    }

    # obtain the devmgr object
    my $esx_dev_mgr = $esx_client_obj->{obj}->{mpdevmgr};

    # get every device
    my @all_devices = $esx_dev_mgr->get_device_objs();

    # initialize the VM_PSP_LIST_RANDOM_SEED
    # set a rseed if one isnt specified
    if (not defined($esx_host_hash->{PSP_LIST_RANDOM_SEED}) or $esx_host_hash->{PSP_LIST_RANDOM_SEED} =~ /^\s*$/) {
        # generate one using the balance value for this esx host and the current time
        $esx_host_hash->{PSP_LIST_RANDOM_SEED} = time() + ($balance * $balance);
    }

    # set the seed for all rand() calls
    srand($esx_host_hash->{PSP_LIST_RANDOM_SEED});

    my @psp_list = split(/,/, $esx_host_hash->{PSP_LIST});
    my @tmp_psp_list = @psp_list;
    my $print_msg = 1;

    # shuffle the array
    &ClientUtil::randomize_array(\@tmp_psp_list);

    # foreach device, apply a randomized psp setting
    foreach my $device_obj (@all_devices) {
        # skip this lun if it is not controller by NMP
        next if ($device_obj->mp_type() ne "NMP");

        if ($print_msg) {
            $self->sanlog(-msg => "Setting PSP Policies on all LUNs with the following properties.\n" .
                                  "\tPSP_LIST = " . $esx_host_hash->{PSP_LIST} . "\n".
                                  "\tPSP_LIST_RANDOM_SEED = " . $esx_host_hash->{PSP_LIST_RANDOM_SEED});
        }
        $print_msg = 0;

        # if the temp psp list is empty, populate it again
        if (scalar(@tmp_psp_list) == 0) {
            # update the empty temp list to be fully populated again
            @tmp_psp_list = @psp_list;

            # shuffle the array
            &ClientUtil::randomize_array(\@tmp_psp_list);
        }

        # shift off a psp policy to use for this lun
        my $psp_policy = shift(@tmp_psp_list);

        # obtain the naa handle for this lun
        my $naa_device = $device_obj->mp_device_name();

        # set the policy
        $esx_client_obj->{mp}->set_psp_policy(-naa_device => $naa_device,
                                              -psp_policy => $psp_policy,
                                              -native_cmd => 1);
        if ($esx_client_obj->{mp}->error()) {
            $self->_set_err("Failed to set PSP Policy on $naa_device", $esx_client_obj->{mp}->error());
            return;
        }
    }

    return;
}

sub _set_vxdmp_io_policies {
    my ($self, $esx_client_obj, $esx_host) = @_;

    # only operate on ESXi systems
    if (not $esx_client_obj->has_vma_host()) {
        return;
    }

    # get every device rdm/vmfs/other
    # get devices from the mpdevmgr object
    my $esx_dev_mgr = $esx_client_obj->{obj}->{mpdevmgr};
    my @all_devices = $esx_dev_mgr->get_device_objs();

    my $io_policy = $esx_client_obj->{vx_io_policy};

    my $enclosures = {};
    # foreach device, apply a randomized psp setting
    foreach my $device_obj (@all_devices) {
        # skip this lun if it is not controlled by VxDMP
        next if ($device_obj->mp_type() !~ /VxDMP/i);

        # get the enclosure name from the object
        my $mp_device_name = $device_obj->mp_device_name();
        my $enclosure_name = $device_obj->get_os_specific_data("vxdmp_enclosure_name");
        if (not $enclosure_name) {
            &SanUtil::nate_log('-warn', "Could not determine VxDMP enclosure name for " .
                               "$mp_device_name, skipping setting IO policy");
            next;
        }

        # store the enclosures in a hash to get rid of dups
        $enclosures->{$enclosure_name}++;
    }

    # if we found any enclosures, then apply the io policy on each one we found
    # this eliminates the possibility of duplicate calls to setting the io policy
    if (keys %{$enclosures}) {
        $self->sanlog(-msg => "Setting VxDMP I/O Policy $io_policy on all enclosures");

        foreach my $enclosure_name (keys %{$enclosures}) {
            # set the vxdmp io policy
            $esx_client_obj->{mp}->set_vxdmp_io_policy(-enclosure => $enclosure_name,
                                                       -io_policy => $io_policy);
            if ($esx_client_obj->{mp}->error()) {
                $self->_set_err($esx_client_obj->{mp}->errmsg());
                return;
            }
        }
    }

    return;
}

sub _verify_vmfs_rdm_counts {
    my ($self, $esx_dev_mgr, $vmfs_lun_path_regex_filter, $virtual_machines, $num_vmdks_per_datastore, $vserver_list) = @_;

    my ($vmfs_vmdk_requested_total, $rdm_vmdk_requested_total);
    $vmfs_vmdk_requested_total = 0;
    $rdm_vmdk_requested_total = 0;

    my @all_rdm_devices;

    foreach my $virtual_machine (values %{$virtual_machines}) {
        my $vmfs_vmdk_count = defined($virtual_machine->{VIRTUAL_DISKS}) ? $virtual_machine->{VIRTUAL_DISKS} : 0;
        my $rdm_vmdk_count = defined($virtual_machine->{RAW_DISKS}) ? $virtual_machine->{RAW_DISKS} : 0;

        $vmfs_vmdk_requested_total += $vmfs_vmdk_count;
        $rdm_vmdk_requested_total  += $rdm_vmdk_count;

        if ($rdm_vmdk_count > 0) {
            my $hosttype_hsh = host($virtual_machine->{NAME});
            my $hosttype_hostrec = $hosttype_hsh->{hosttype};
            my $ostype;
            if ($hosttype_hostrec =~ /linux/i) {
                $ostype = "linux";
            } elsif ($hosttype_hostrec =~ /solaris/i) {
                $ostype = "solaris";
            } else {
                $ostype = $hosttype_hostrec;
            }

            my $rdm_lun_path_regex_filter = $virtual_machine->{RAW_LUN_PATH_REGEX};
            if (defined($rdm_lun_path_regex_filter)) {
                $rdm_lun_path_regex_filter = eval($rdm_lun_path_regex_filter);
            } else {
                $rdm_lun_path_regex_filter = undef;
            }

            my @rdm_devices = $esx_dev_mgr->get_filtered_devices(-ontap_mode => "cmode",
                                                                 -device_count => $rdm_vmdk_count,
                                                                 -ontap_lun_ostype_regex => qr/$ostype/,
                                                                 -ontap_lun_path_regex => $rdm_lun_path_regex_filter,
                                                                 -vservers => $vserver_list);
            if ($esx_dev_mgr->error()) {
                $self->_set_err("Failed to get RDM devices", $esx_dev_mgr->errmsg());
                return ();
            }
            push(@all_rdm_devices, @rdm_devices);
        }
    }
    my @all_vmfs_devices;

    # if any vmfs vmdks were requested, then obtain them
    if ($vmfs_vmdk_requested_total > 0) {
        # obtain all vmfs devices on the devmgr
        @all_vmfs_devices = $esx_dev_mgr->get_filtered_devices(-ontap_mode => "cmode",
                                                               -ontap_lun_ostype_regex => qr/vmware/i,
                                                               -ontap_lun_path_regex => $vmfs_lun_path_regex_filter,
                                                               -vservers => $vserver_list);
        if ($esx_dev_mgr->error()) {
            $self->_set_err("Failed to get VMFS devices", $esx_dev_mgr->errmsg());
            return ();
        }
    }

    # mark all the devices we obtained as free
    foreach my $device_obj (@all_rdm_devices, @all_vmfs_devices) {
        $device_obj->device_state("free");
    }

    my $vmfs_vmdk_actual_count = scalar(@all_vmfs_devices) * $num_vmdks_per_datastore;
    my $rdm_vmdk_actual_count = scalar(@all_rdm_devices);

    if ($vmfs_vmdk_actual_count < $vmfs_vmdk_requested_total) {
        $self->_set_err("Requested '$vmfs_vmdk_requested_total' VMFS VMDKs when only '$vmfs_vmdk_actual_count' VMFS VMDKs are available");
    }

    if ($rdm_vmdk_actual_count < $rdm_vmdk_requested_total) {
        $self->_set_err("Requested '$rdm_vmdk_requested_total' RDM VMDKs when only '$rdm_vmdk_actual_count' RDM VMDKs are available");
    }

    return ($rdm_vmdk_requested_total, $vmfs_vmdk_requested_total);
}

1;
