#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Hyper-V Cleanup module
## @author hrivera@netapp.com
## @status private
## @pod here

package San::SanMgr::HyperVisor::HyperV::cleanup;

use San::SanMgr::Util;
use Data::Dumper;
use strict;
use warnings;
use Tharn;

###############################################################################
# Method:    cleanup_all
# Objective: Cleanup every Hyper-V host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup_all {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);

    my $hyperv_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                        -ref_type => "HASH",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @subtests;

    # iterate through all HYPERV Hosts and issue cleanup_single
    foreach my $hyperv_host (keys %{$hyperv_hosts}) {
        my $hyperv_host_name = $hyperv_hosts->{$hyperv_host}->{NAME};

        # skip cleanup on this host if told to skip
        $hyperv_hosts->{$hyperv_host}->{CLEANUP} = 1 if (not defined($hyperv_hosts->{$hyperv_host}->{CLEANUP}));
        if ($hyperv_hosts->{$hyperv_host}->{CLEANUP} == 0) {
            $self->sanlog(-msg => "Skipping HYPERV Cleanup on host '$hyperv_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id => $hyperv_host_name . '-cleanup',
                                      -run => sub { &cleanup_single($self,
                                                                  -resource => $hyperv_hosts->{$hyperv_host}); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to cleanup ${hyperv_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No HYPERV HyperVisors to cleanup");

        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests,
                         -sync_enable => 1);
    if (&SubIPC::error()) {
        $self->_set_err("HYPERV-cleanup failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HYPERV-cleanup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All HYPERV HyperVisors have been cleanup.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    cleanup_single
# Objective: Cleanup a single Hyper-V host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $hyperv_host = &ParamsMgr::get_arg(-arg => "resource",
                                          -ref_type => "HASH",
                                          -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $hyperv_host_name = $hyperv_host->{NAME};
    my $protocol = $hyperv_host->{PROTOCOL};
    my $index = $hyperv_host->{_INDEX};
    $index = "" if (not defined($index));

    my $virtual_machines = &ParamsMgr::search(-search_path => ["HYPERVISORS"],
                                              -key => "HYPERV" . $index,
                                              -key_filter => qr/^VM\d*$/,
                                              paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # obtain an array of all the virtual machine names
    my @vm_names = map { $virtual_machines->{$_}->{NAME} } (keys %{$virtual_machines});

    my ($hyperv_host_obj) = $self->get_objects(-type => "hypervisor",
                                               -names => [$hyperv_host_name]);
    if ($self->error()) {
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # get list of vservers that this client will talk to
    my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $hyperv_host_name);
    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get vservers associated to '$hyperv_host_name'");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    # reconnect to this host
    $hyperv_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($hyperv_host_obj->error()) {
        $self->_set_err($hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # power off all the virtual machines on this host
    $hyperv_host_obj->power_management(-action => "shutdown",
                                       -vm_list => \@vm_names);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable to power off vms", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # need to rescan for devices just in case
    $hyperv_host_obj->discover();
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable discover devices", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # make a devmgr (scope it to only the vservers that are a part of this test)
    my $devmgr_obj = $hyperv_host_obj->create_devmgr(-target_list => \@vserver_list,
                                                     -lun_path_regex => undef,
                                                     -vhd_lun_path_regex => undef);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable to create device manager object", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # foreach vm
    ## get passthrough disks from vms
    ## remove passthrough disks from vms
    ## get vhd devices from vm
    ## remove vhd files from vm
    ## for each vhd lun, cleanup vhds, and destroy partition

    foreach my $virtual_machine (values %{$virtual_machines}) {
        # get passthrough disks from the this VM
        my @pt_devices = $hyperv_host_obj->get_passthrough_devices_from_vm(-vm_name => $virtual_machine->{NAME});
        if ($hyperv_host_obj->error()) {
            $self->_set_err("Unable to obtain passthrough devices from VM", $hyperv_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        if (scalar(@pt_devices) > 0) {
            $hyperv_host_obj->remove_passthrough_devices_from_vm(-vm_name => $virtual_machine->{NAME},
                                                                 -passthrough_devices => \@pt_devices);
            if ($hyperv_host_obj->error()) {
                $self->_set_err("Failed to remove passthrough devices from VM", $hyperv_host_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }

        my $vhd_relationship_hash = $hyperv_host_obj->get_vhd_devices_from_vm(-vm_name => $virtual_machine->{NAME});
        if ($hyperv_host_obj->error()) {
            $self->_set_err("Failed to obtain VHD devices from VM", $hyperv_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        my @vhd_files = keys %{$vhd_relationship_hash};
        if (scalar(@vhd_files) > 0) {
            $hyperv_host_obj->remove_vhd_files_from_vm(-vm_name => $virtual_machine->{NAME},
                                                       -vhd_files => \@vhd_files);
            if ($hyperv_host_obj->error()) {
                $self->_set_err("Failed to remove VHD devices from VM", $hyperv_host_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            foreach my $vhd_device (values %{$vhd_relationship_hash}) {
                # skip vhd devices that dont have any more vhds
                if ($vhd_device->{vhd_count} == 0) {
                    next;
                }

                # destroy all vhds on this device
                $vhd_device->destroy_all_vhds();
                if ($vhd_device->error()) {
                    $self->_set_err("Failed to destroy VHD files from VHD device", $vhd_device->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                # destroy the partition
                $vhd_device->destroy_partition();
                if ($vhd_device->error()) {
                    $self->_set_err("Failed to destroy partition on VHD device", $hyperv_host_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }
            }
        }
    }

    # remove empty drive slots
    $hyperv_host_obj->remove_empty_drive_slots_from_vms(-vm_list => \@vm_names);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Failed to remove empty drive slots", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # power on all the virtual machines that we shut down
    $hyperv_host_obj->power_management(-action => "power_on",
                                       -vm_list => \@vm_names);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable to power on VMs", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "Completed cleanup on Hyper-V Host");

    return $SanMgr::TRUE;
}

1;
