#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Hyper-V Config module
## @author hrivera@netapp.com
## @status private
## @pod here

package San::SanMgr::HyperVisor::HyperV::config;

use San::SanMgr::Util;
use Data::Dumper;

###############################################################################
# Method:    config_all
# Objective: Config all Hyper-V Hosts
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_all {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);

    my $hyperv_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                           -ref_type => "HASH",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # iterate through all HyperV Hosts and issue config_single
    foreach my $hyperv_host (keys %{$hyperv_hosts}) {
        my $hyperv_host_name = $hyperv_hosts->{$hyperv_host}->{NAME};

        # skip config on this host if told to skip
        $hyperv_hosts->{$hyperv_host}->{CONFIG} = 1 if (not defined($hyperv_hosts->{$hyperv_host}->{CONFIG}));
        if ($hyperv_hosts->{$hyperv_host}->{CONFIG} == 0) {
            $self->sanlog(-msg => "Skipping HyperV Config on host '$hyperv_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id =>  $hyperv_host_name . '-config',
                                      -run => sub { &config_single($self,
                                                                   -resource => $hyperv_hosts->{$hyperv_host}); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to config ${hyperv_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No HyperV HyperVisors to configure");

        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("HyperV-config failed", &SubIPC::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperV-config has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All HyperV HyperVisors have been Configured.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    config_single
# Objective: Config a single HYPERV Host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource",
                                       -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $client_name = $resource->{NAME};

    # given a single hypervisor, configure this hypervisor
    # By the time this function is done, this client should be
    # ready for fc zoning, or iscsi logins

    # obtain the sanclient object
    my ($hyperv_host_obj) = $self->get_objects(-type => "hypervisor",
                                               -names => [$client_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get client ($client_name) object");
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # get the list of vserver targets that this initiator is associated with
    my ($data) = $self->{parent}->get_target_list(-client_name => $client_name,
                                                  -filter_iscsi_targets_by_portset => $hyperv_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                  -filter_fcp_targets_by_portset   => $hyperv_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get iSCSI/FCP target lists");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    my $protocol = $hyperv_host_obj->get_protocol();

    # configure iSCSI/FCP
    if ($protocol =~ /iscsi|fcp/i) {
        $hyperv_host_obj->configure(-iscsi_targets => \@iSCSITargets,
                                    -fcp_targets => \@FcpTargets);
        if ($hyperv_host_obj->error()) {
            $self->_set_err("Failed to configure $protocol for the client ($client_name)", $hyperv_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # configure NFS
    if ($protocol =~ /nfs/i) {
        # NFS currently has no configuration steps to perform
    }

    # Load up the resource information and get them via get_arg
    my $caller_id = $resource->{NAME} . "|" . time();
    &ParamsMgr::process_args(%{$resource}, -CALLER => $caller_id);

    # NTP sync the hypervisors
    my $ntp_sync =        &ParamsMgr::get_arg(-arg => "NTP_SYNC", -required => 0, -default_to_env => 1,
                                              -default => 0, -CALLER => $caller_id);
    my $ntp_time_server = &ParamsMgr::get_arg(-arg => "NTP_TIME_SERVER", -required => 1, -default_to_env => 1,
                                              -default => "ntp0.eng.netapp.com", dependant_on => ["NTP_SYNC"],
                                              -CALLER => $caller_id);
    my $ntp_time_zone =   &ParamsMgr::get_arg(-arg => "NTP_TIME_ZONE", -required => 1, -default_to_env => 1,
                                              -default => "EST5EDT", -dependant_on => ["NTP_SYNC"],
                                              -CALLER => $caller_id);
    if ($ntp_sync == 1) {
        $hyperv_host_obj->ntp_sync(-ntp_server => $ntp_time_server,
                                   -ntp_time_zone => $ntp_time_zone);
        if ($hyperv_host_obj->error()) {
            my @error = $hyperv_host_obj->errmsg();
            $self->sanlog(-msg => "Failed to NTP time sync the $client_name (@error))",
                          -msg_type => "warn");
            $hyperv_host_obj->_clear_err();
        }
    }

    &SubIPC::pass(-msg => "HyperV Host has been Configured.");

    return $SanMgr::TRUE;
}

1;
