#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Hyper-V Deconfig Module
## @author hrivera@netapp.com
## @status private
## @pod here

package San::SanMgr::HyperVisor::HyperV::deconfig;

use San::SanMgr::Util;
use Data::Dumper;

###############################################################################
# Method:    deconfig_all
# Objective: Deconfig every Hyper-V host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub deconfig_all {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);

    my $hyperv_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                        -ref_type => "HASH",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # iterate through all HyperV Hosts and issue deconfig_single
    foreach my $hyperv_host (keys %{$hyperv_hosts}) {
        my $hyperv_host_name = $hyperv_hosts->{$hyperv_host}->{NAME};

        # skip deconfig on this host if told to skip
        $hyperv_hosts->{$hyperv_host}->{DECONFIG} = 1 if (not defined($hyperv_hosts->{$hyperv_host}->{DECONFIG}));
        if ($hyperv_hosts->{$hyperv_host}->{DECONFIG} == 0) {
            $self->sanlog(-msg => "Skipping HyperV Deconfig on host '$hyperv_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id =>  $hyperv_host_name . '-deconfig',
                                      -run => sub { &deconfig_single($self,
                                                                   -resource => $hyperv_hosts->{$hyperv_host}); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to deconfig ${hyperv_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No HyperV HyperVisors to deconfigure");

        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("HyperV-deconfig failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperV-deconfig has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All HyperV HyperVisors have been Deconfigured.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    deconfig_single
# Objective: Deconfig a single Hyper-V host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub deconfig_single {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);
    my $resource = &ParamsMgr::get_arg(-arg => "resource",
                                       -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $client_name = $resource->{NAME};

    # get the sanclient object for this hyperv host
    my ($hyperv_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$client_name]);
    if ($self->error()) {
        $self->_set_err("Failed to get client ($client_name) object");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # reconnect to this host
    $hyperv_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($hyperv_host_obj->error()) {
        $self->_set_err($hyperv_host_obj->errmsg());
        &SubIPC::fail(-msg => $self->errmsg());
        return $SanMgr::FALSE;
    }

    # obtain the vserver targets that are associated with this host
    my ($data) = $self->{parent}->get_target_list(-client_name => $client_name,
                                                  -filter_iscsi_targets_by_portset => $hyperv_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                  -filter_fcp_targets_by_portset   => $hyperv_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get iSCSI/FCP target lists");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    my $protocol = $hyperv_host_obj->get_protocol();

    # deconfigure iSCSI/FCP
    if ($protocol =~ /iscsi|fcp/i and scalar(@iSCSITargets)) {
        $hyperv_host_obj->deconfigure(-iscsi_targets => \@iSCSITargets,
                                      -clean_level => $hyperv_host_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($hyperv_host_obj->error()) {
            $self->_set_err("Failed to deconfigure $protocol for the client ($client_name)", $hyperv_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # deconfigure NFS
    if ($protocol =~ /nfs/i) {
        # NFS currently has no deconfiguration steps to perform
    }

    # clean up any orphaned devices
    $hyperv_host_obj->orphaned_device_cleanup();
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Failed cleanup orphaned devices", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "HyperV Host '$client_name' has been deconfigured.");

    return $SanMgr::TRUE;
}

1;
