#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Hyper-V Setup module
## @author hrivera@netapp.com
## @status private
## @pod here

package San::SanMgr::HyperVisor::HyperV::setup;

use San::SanMgr::Util;
use Data::Dumper;
use Tharn;

###############################################################################
# Method:    setup_all
# Objective: Setup all Hyper-V Hosts
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup_all {
    my ($self, %args) = @_;

    $self->clear_error();

    &ParamsMgr::process_args(%args);

    my $hyperv_hosts = &ParamsMgr::get_arg(-arg => "resources",
                                           -ref_type => "HASH",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @subtests;
    my $balance_count = 0;

    # iterate through all HyperV Hosts and issue setup_single
    foreach my $hyperv_host (keys %{$hyperv_hosts}) {
        my $hyperv_host_name = $hyperv_hosts->{$hyperv_host}->{NAME};

        # skip setup on this host if told to skip
        $hyperv_hosts->{$hyperv_host}->{SETUP} = 1 if (not defined($hyperv_hosts->{$hyperv_host}->{SETUP}));
        if ($hyperv_hosts->{$hyperv_host}->{SETUP} == 0) {
            $self->sanlog(-msg => "Skipping HyperV Setup on host '$hyperv_host_name'.");
            next;
        }

        my $subtest = &SubIPC::create(-id => $hyperv_host_name . '-setup',
                                      -run => sub { &setup_single($self,
                                                                  -resource => $hyperv_hosts->{$hyperv_host},
                                                                  -balance => $balance_count); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to setup ${hyperv_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        $balance_count += 1;
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No HyperV HyperVisors to setup");

        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests,
                         -sync_enable => 1);
    if (&SubIPC::error()) {
        $self->_set_err("HyperV-setup failed", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # parse the hyperv CDL
    my @messages = &SubIPC::get_messages(-subtest => \@subtests,
                                         -type => "hyperv_client_dev_list");
    if (&SubIPC::error()) {
        $self->_set_err("Failed to parse hyperv client dev list messages", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # prepare the device list for all hyper v hosts
    my $device_list;
    foreach my $message (@messages) {
        foreach my $device (@{$message->{extra}->{device_list}}) {
            push(@{$device_list->{$message->{extra}->{name}}->{DEV_LIST}}, $device);
        }
    }

    # send the final device list up to the parent
    &SubIPC::write(-subtest => [],
                   -msg => "Hyper-V CDL",
                   -type => "hyperv_client_dev_list",
                   -device_list => $device_list);
    if (&SubIPC::error()) {
        $self->_set_err("Failed to write Hyperv CDL to parents", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "HyperV-setup has passed\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    &SubIPC::pass(-msg => "All HyperV HyperVisors have been setup.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    setup_single
# Objective: Setup a single HYPERV Host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup_single {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $hyperv_host = &ParamsMgr::get_arg(-arg => "resource",
                                       -ref_type => "HASH",
                                       -required => 1);
    my $balance = &ParamsMgr::get_arg(-arg => "balance",
                                      -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $hyperv_host_name = $hyperv_host->{NAME};

    my $protocol = $hyperv_host->{PROTOCOL};
    my @protocols;
    if ($protocol =~ /iscsi/i) {
        push(@protocols, "iscsi");
    }
    if ($protocol =~ /fcp/i) {
        push(@protocols, "fcp");
    }

    my $vhd_lun_path_regex_filter = $hyperv_host->{VHD_LUN_PATH_REGEX};
    if (defined($vhd_lun_path_regex_filter)) {
        $vhd_lun_path_regex_filter = eval($vhd_lun_path_regex_filter);
    } else {
        $vhd_lun_path_regex_filter = undef;
    }

    # see if requested to allocate luns for the physical host
    my $physical_host_lun_path_regex_filter = $hyperv_host->{PHYSICAL_HOST_LUN_PATH_REGEX};
    if (defined($physical_host_lun_path_regex_filter)) {
        $physical_host_lun_path_regex_filter = eval($physical_host_lun_path_regex_filter);
    } else {
        $physical_host_lun_path_regex_filter = undef;
    }

    # assume the io type for the hyperv host will be fs
    my $io_type = defined($hyperv_host->{IO_TYPE}) ? $hyperv_host->{IO_TYPE} : "fs";

    # get list of vservers that this client will talk to
    my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $hyperv_host_name);
    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get vservers associated to '$hyperv_host_name'");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    # we also need to obtain the list of cluster objects that this client is related to as well
    # TODO consider getting rid of this at some point
    my $hash = $self->{parent}->get_vserver_client_relationship_hash();
    if ($self->{parent}->error()) {
        $self->{parent}->_set_err("Failed to get cluster/client relationship hash for '$hyperv_host_name'");
        &SubIPC::fail(-object => $self->{parent});
        return $SanMgr::FALSE;
    }

    my $virtual_machines = &ParamsMgr::search(-search_path => ["HYPERVISORS"],
                                              -key => "HYPERV" . $index,
                                              -key_filter => qr/^VM\d*$/,
                                              paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err("Unable to search for VM keys", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my @vm_names = map { $virtual_machines->{$_}->{NAME} } (keys %{$virtual_machines});

    my ($hyperv_host_obj) = $self->get_objects(-type => "hypervisor",
                                               -names => [$hyperv_host_name]);
    if ($self->error()) {
        $self->_set_err("No SanClient objects foun for HYPERV Hypervisor '$hyperv_host_name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # obtain VHD and PT total requested counts
    my $total_vhd_lun_count = 0;
    my $total_ptd_lun_count = 0;

    ($total_vhd_lun_count, $total_ptd_lun_count) = &_get_vhd_ptd_counts($virtual_machines);

    # reconnect to this host
    $hyperv_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable to reconnect to HYPERV host '$hyperv_host_name'", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # cleanly shutdown all virtual machines
    $hyperv_host_obj->power_management(-action => "shutdown",
                                       -vm_list => \@vm_names);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable to power off VMs", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # discover new luns
    $hyperv_host_obj->discover(-num_expected_luns => $total_vhd_lun_count + $total_ptd_lun_count);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable to discover devices", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # clean up empty drive slots on all guests
    $hyperv_host_obj->remove_empty_drive_slots_from_vms(-vm_list => \@vm_names);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable to remove empty drive slots from VMs", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # create devmgr object
    my $devmgr_obj = $hyperv_host_obj->create_devmgr(-balance => $balance,
                                                     -vhd_lun_path_regex => undef,
                                                     -lun_path_regex => undef);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable to create device manager object", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # if specified to provision LUNs to the hyper-v host, then grab devices that match
    # the regular expression specified
    if (defined($physical_host_lun_path_regex_filter)) {
        my @all_physical_host_luns = $devmgr_obj->get_free_devices(-filters => \@vserver_list,
                                                                   -device_type => "passthrough",
                                                                   -lun_path_regex => $physical_host_lun_path_regex_filter,
                                                                   -protocol => join("|", @protocols));
        if ($devmgr_obj->error()) {
            $self->_set_err("Unable to get free physical host devices", $devmgr_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        # format these disks and mount them up, and send up the client dev list data
        my @device_info_list;
        foreach my $dev_obj (@all_physical_host_luns) {
            my $device_info = {};

            # Save some attributes of the device onto the device_info hash
            if ($dev_obj->get_lun() =~ /^\/vol\/(\S+)\/(\S+)$/) {
                push(@{$device_info->{'volume'}}, $1);
                push(@{$device_info->{'lun'}}, $2);
            }
            push(@{$device_info->{'vserver'}}, $dev_obj->get_filer_name());
            $device_info->{'protocol'} = lc($dev_obj->get_protocol());

            # Push the device serial number onto the device info hash if it is defined
            if (defined($dev_obj->{disk_obj}->{serial}) and $dev_obj->{disk_obj}->{serial} ne "") {
                push(@{$device_info->{'serial'}}, $dev_obj->{disk_obj}->{serial});
            }

            if ($io_type =~ /fs/) {
                $dev_obj->create_fs();
                if ($dev_obj->error()) {
                    $self->_set_err("Failed to format device", $dev_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                $dev_obj->mount_fs();
                if ($dev_obj->error()) {
                    $self->_set_err("Failed to assign driveletter to device", $dev_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }

                # parse out the drive letter
                $device_info->{path} = $dev_obj->get_mount();
                $device_info->{io_type} = "fs";
            } else {
                $device_info->{path} = $dev_obj->get_raw_device();
                $device_info->{io_type} = "raw";
            }
            push(@device_info_list, $device_info);
        }

        if (scalar(@device_info_list)) {
            &SubIPC::write(-msg => "Client Device List for Hyper-V host ($hyperv_host_name)",
                           -device_list => \@device_info_list,
                           -type => "hyperv_client_dev_list", -name => $hyperv_host_name);
            if (&SubIPC::error()) {
                $self->_set_err(&SubIPC::get_error());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }
    }

    # get all vhd devices and store in vhd pool
    my @all_vhd_devices = ();
    # obtain all the VHD devices by using get_free_devices()
    if ($total_vhd_lun_count > 0) {
        @all_vhd_devices = $devmgr_obj->get_free_devices(-filers => \@vserver_list,
                                                         -device_type => "vhd",
                                                         -num_devices => $total_vhd_lun_count,
                                                         -lun_path_regex => $vhd_lun_path_regex_filter,
                                                         -protocol => join("|", @protocols));
        if ($devmgr_obj->error()) {
            $self->_set_err("Unable to get free VHD devices", $devmgr_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # foreach vm
    ## if vm wants vhds
    ### take some vhds from the pool
    ### store vhds in datastruct
    ## if vm wants PTDs
    ### take some ptds from the pool
    ### store ptds in datastruct
    foreach my $virtual_machine (values %{$virtual_machines}) {
        # allocate some vhd devices for this VM
        my @vhd_devs_for_this_vm = ();
        for (my $i = 0; $i < $virtual_machine->{VIRTUAL_DISKS}; $i++) {
            push(@vhd_devs_for_this_vm, shift(@all_vhd_devices));
            @all_vhd_devices = reverse(@all_vhd_devices);
        }
        $virtual_machine->{vhd_dev_objs} = \@vhd_devs_for_this_vm;

        # allocate some ptd devices for this VM
        my @ptd_devs_for_this_vm = ();
        my $raw_lun_path_regex = $virtual_machine->{RAW_LUN_PATH_REGEX};
        if (defined($raw_lun_path_regex)) {
            $raw_lun_path_regex = eval($raw_lun_path_regex);
        } else {
            $raw_lun_path_regex = undef;
        }

        # figure out the OStype of this virtual machine
        my $tharnhost_hsh = host($virtual_machine->{NAME});
        my $ostype = $tharnhost_hsh->{hosttype};
        $ostype = "linux" if ($ostype =~ /linux/i);
        $ostype = "windows" if ($ostype =~ /windows/i);
        if (not length($ostype) or $ostype !~ /windows|linux/i) {
            $self->_set_err("The hosttype '$ostype' of virtual machine '$virtual_machine->{NAME}' is invalid, only linux and windows hosttypes are supported ");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        @ptd_devs_for_this_vm = $devmgr_obj->get_free_devices(-filers => \@vserver_list,
                                                              -device_type => "passthrough",
                                                              -num_devices => $virtual_machine->{RAW_DISKS},
                                                              -lun_ostype_regex => qr/$ostype/i,
                                                              -lun_path_regex => $raw_lun_path_regex,
                                                              -protocol => join("|", @protocols));
        if ($devmgr_obj->error()) {
            $self->_set_err("Unable to get free PassThrough devices", $devmgr_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        $virtual_machine->{ptd_dev_objs} = \@ptd_devs_for_this_vm;
    }

    foreach my $virtual_machine (values %{$virtual_machines}) {
        my (@vhds_for_this_guest);
        if (scalar(@{$virtual_machine->{vhd_dev_objs}})) {
            (@vhds_for_this_guest) = $devmgr_obj->create_vhd_files(-vhd_devices => $virtual_machine->{vhd_dev_objs},
                                                                   -num_vhds_per_device => 1,
                                                                   -mount_type => "letter",
                                                                   -vhd_types => [split(/,/, $hyperv_host->{VHD_DISK_TYPE})]);
            if ($devmgr_obj->error()) {
                $self->_set_err("Failed to create VHD files", $devmgr_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }
        $virtual_machine->{vhd_files} = \@vhds_for_this_guest;
    }

    # foreach vm,  attach the devices in a parallel subtest
    foreach my $virtual_machine (values %{$virtual_machines}) {
        my $subtest = &SubIPC::create(-id => $virtual_machine->{NAME} . '-attach_devices',
                                      -run => sub { &_attach_devices_to_vm($self,
                                                                           -balance => $balance,
                                                                           -hyperv_host_obj => $hyperv_host_obj,
                                                                           -hyperv_host_resource => $hyperv_host,
                                                                           -virtual_machine_resource => $virtual_machine); },
                                      -parent => $self,
                                      -ipc => 1,
                                      -bg => 1);
        if (&SubIPC::error()) {
            $self->_set_err("Failed to instantiate a subtest to setup ${hyperv_host_name}",
                            &SubIPC::get_error());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
        $balance += 1;
        push(@subtests, $subtest);
    }

    if (scalar(@subtests) == 0) {
        &SubIPC::pass(-msg => "No HyperV HyperVisors to setup");

        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish(-subtest => \@subtests,
                         -sync_enable => 1);
    if (&SubIPC::error()) {
        $self->_set_err("Failed to attach devices to VMs", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->_set_err("A subtest failure has been reported:\n\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    } else {
        $self->sanlog(-msg => "VHD/PTD Devices have been created and attachedn\t\tFailed: " . $st_status->{failed} . "\n\t\tPassed: " . $st_status->{passed});
    }

    # turn all VMs back on, or turn each vm on inside it's own subtest
    $hyperv_host_obj->power_management(-action => "power_on",
                                       -vm_list => \@vm_names);
    if ($hyperv_host_obj->error()) {
        $self->_set_err("Unable to create device manager object", $hyperv_host_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    &SubIPC::pass(-msg => "All HyperV HyperVisors have been setup.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _attach_devices_to_vm
# Objective: Given a virtual machine, attach virtual devices to it
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _attach_devices_to_vm {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $hyperv_host = &ParamsMgr::get_arg(-arg => "hyperv_host_resource",
                                          -ref_type => "HASH",
                                          -required => 1);
    my $virtual_machine = &ParamsMgr::get_arg(-arg => "virtual_machine_resource",
                                              -ref_type => "HASH",
                                              -required => 1);
    my $hyperv_host_obj = &ParamsMgr::get_arg(-arg => "hyperv_host_obj",
                                              -required => 1);
    my $balance = &ParamsMgr::get_arg(-arg => "balance",
                                      -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # establish a new random seed for this subtest to allow concurrent access
    # to diskpart
    srand($balance);

    # set up the Load balancing policies for all LUNs
    foreach my $device (@{$virtual_machine->{vhd_dev_objs}}, @{$virtual_machine->{ptd_dev_objs}}) {
        $device->set_lbpolicy();
        if ($device->error()) {
            $self->_set_err("Failed to set Load Balance policy on device", $device->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    my $devmgr_obj = $hyperv_host_obj->{obj}{devmgr};

    # select the PTD and VHD controller types, default to the Hyperv Host value
    # if no value is specified at the virtual machine level
    my $controller_type;
    if (defined($virtual_machine->{CONTROLLER_TYPE})) {
        $controller_type = $virtual_machine->{CONTROLLER_TYPE};
    } else {
        $controller_type = $hyperv_host->{CONTROLLER_TYPE};
    }

    # configure any PTD devices first
    if (scalar(@{$virtual_machine->{ptd_dev_objs}})) {
        $devmgr_obj->configure_passthrough_disks(-passthrough_devices => $virtual_machine->{ptd_dev_objs},
                                                 -controller_types => [split(/,/, $controller_type)]);
        if ($devmgr_obj->error()) {
            $self->_set_err("Failed to configure passthrough devices", $devmgr_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # attach VHD files
    if (scalar(@{$virtual_machine->{vhd_files}})) {
        $hyperv_host_obj->add_vhd_files_to_vm(-vhd_files => $virtual_machine->{vhd_files},
                                              -vm_name => $virtual_machine->{NAME},
                                              -controller_types => [split(/,/, $controller_type)]);
        if ($hyperv_host_obj->error()) {
            $self->_set_err("Unable to attach VHD files to vm", $hyperv_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # attach PTD handles
    if (scalar(@{$virtual_machine->{ptd_dev_objs}})) {
        $hyperv_host_obj->add_passthrough_disks_to_vm(-passthrough_devices => $virtual_machine->{ptd_dev_objs},
                                                      -vm_name => $virtual_machine->{NAME});
        if ($hyperv_host_obj->error()) {
            $self->_set_err("Unable to attach PTD files to vm", $hyperv_host_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    &SubIPC::pass(-msg => "VHDs/PTDs have been attached.");

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _get_vhd_ptd_counts
# Objective: Calculate the total number of virtual/passthrough disks requested
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_vhd_ptd_counts {
    my ($virtual_machines) = @_;

    my $total_vhd = 0;
    my $total_ptd = 0;

    # calculate the total vhd and total ptd disk request counts
    foreach my $vm (values %{$virtual_machines}) {
        $total_vhd += $vm->{VIRTUAL_DISKS};
        $total_ptd += $vm->{RAW_DISKS};
    }

    return ($total_vhd, $total_ptd);
}

1;
