#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on switch
## @author latif@netapp.com
## @status private
## @pod here

package San::SanMgr::Switch;

# Compiler directives.
use strict;

# Module import.
use Tharn;
use Error;
use San::SanLog;
use San::SanMgr::Util;
use San::SanSwitch;
use Data::Dumper;
use Data::Dump;

use NATE::Process;
use NATE::STAF::Peer;
use NATE::STAF qw(staf_global);
use NATE::Util::Sockets;
use Sys::Hostname;

# Load resource specific modules
use San::SanMgr::Switch::setup;
use San::SanMgr::Switch::config;
use San::SanMgr::Switch::cleanup;
use San::SanMgr::Switch::validate;

use vars qw(@ISA);

@ISA = qw(San::SanMgr::Util Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new resource object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($proto, %args) = @_;

    my $self = {};

    my $class = ref($proto) || $proto;
    bless ($self, $class);

    # allow for error stack build ups
    $self->{error_dont_overwrite_errors} = 1;
    $self->{error_stack_style} = "normal";

    &ParamsMgr::process_args(%args);
    my $parent = &ParamsMgr::get_arg(-arg => "parent", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Save parent
    $self->{parent} = $parent;

    # Get the set of environment parameters from ParamsMgr
    my $environment = &ParamsMgr::search(search_path => [], key => "ENVIRONMENT",
                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Set the objects debug level
    $self->debug_level(-debug_level => $environment->{DEBUG_LEVEL});
    if ($self->error()) {
        return $self;
    }

    return $self;
}

###############################################################################
# Method:    setup
# Objective: Launch subtest for all switches to perform setup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup {
    my ($self, %args) = @_;

    $self->clear_error();

    # noop

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    config
# Objective: Launch subtest for each switch to perform config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config {
    my ($self) = @_;

    $self->clear_error();

    # TODO

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    cleanup
# Objective: Launch subtest for each switch to perform cleanup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ($self) = @_;

    $self->clear_error();

    # TODO

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    deconfig
# Objective: Launch subtest for each switch to perform deconfig
# Details:   See POD documentation at the end of this file.
###############################################################################
sub deconfig {
    my ($self) = @_;

    $self->clear_error();

    # TODO

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    validate
# Objective: Launch subtest for each switch to perform validation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self) = @_;

    $self->clear_error();

    # TODO

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    load_resource
# Objective: Gather list of switches and create objects for each one
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_resource {
    my ($self, %args) = @_;

    $self->clear_error();

    my $staf = staf_global();

    # Get list of switches
    my $switches = &ParamsMgr::search(-search_path => [],
                                      -key => "SWITCHES",
                                      -key_filter => qr/^SWITCH\d*$/,
                                      paramset => $self->EnGenX::Util::get_paramset());

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    # make sure there are no duplicate switches
    my %dup_hash;
    foreach my $switch (keys(%{$switches})) {
        $dup_hash{$switches->{$switch}->{NAME}}++;
        if ($dup_hash{$switches->{$switch}->{NAME}} > 1) {
            $self->_set_err("Detected duplicate switch entries for the switch " . $switches->{$switch}->{NAME});
            return $SanMgr::FALSE;
        }
    }

    # Go through the list and create an object for each one
    my @procs;
    foreach my $switch (keys(%{$switches})) {
        # Spawn a new process to load this client
        my $proc = NATE::Process->new(codespec => '{ #EnGenX load ' . $switches->{$switch}->{NAME} . '
                                                     use San::SanMgr::Switch;
                                                     San::SanMgr::Switch::_load_switch(); }',
                                      runid => "switch__" . $switches->{$switch}->{NAME},
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result,
                                      client => 'localhost',
                                      params => { "switches_ds" => $switches,
                                                  "switch" => $switch,
                                                  "debug_level" => $self->{debug_level},
                                                  "staf_parent_handle" => $staf->handle(),
                                                  "staf_parent_endpoint" => NATE::Util::Sockets::get_address(Sys::Hostname::hostname())});
        push(@procs, $proc);
    }

    # set up the Event listener and callback methods
    my $filter = sub {
        my $msg = shift;
        return ($msg->{type} =~ /SanMgr\/Switch\/SanSwitch\/\S+?\/Load_Complete/);
    };
    my $callback = sub {
        my ($msg, $listener) = @_;

        my ($switch_name) = ($msg->{type} =~ /SanMgr\/Switch\/SanSwitch\/(\S+?)\/Load_Complete/);
        my $sm_switch_obj = $listener->{sm_switch_obj};

        # rebuild the SanSwitch object
        my $obj = eval($msg->{message}->{obj_serialized});
        $obj = SanSwitch->new(-obj_template => $obj);
        if ($obj->error()) {
            $sm_switch_obj->_set_err("Failed to instantiate SanSwitch object from template",
                                     $obj->errmsg());
            $sm_switch_obj->{parent}->{listeners_complete_count}++;
            return $SanMgr::FALSE;
        }

        # store the object on the sanmgr resource hash
        $sm_switch_obj->{parent}->{_objects}->{switch}->{$obj->hostname()} = $obj;

        # validate the md5 for the obj
        if (not $sm_switch_obj->_validate_obj_md5_hash(obj => $obj,
                                                       obj_md5_hex_before => $msg->{message}->{obj_md5_hex})) {
            $sm_switch_obj->{parent}->{listeners_complete_count}++;
            return $SanMgr::FALSE;
        }

        $sm_switch_obj->{parent}->{listeners_complete_count}++;
        return $SanMgr::TRUE;
    };
    my $listener = {filter => $filter, callback => $callback, sm_switch_obj => $self};

    # build the fail listener
    my $fail_filter = sub {
        my $msg = shift;
        return ($msg->{type} =~ /SanMgr\/Switch\/SanSwitch\/\S+\/Load_Failed/);
    };
    my $fail_callback = sub {
        my ($msg, $flistener) = @_;

        my $sm_switch_obj = $flistener->{sm_switch_obj};

        # parse out the message information
        my ($switch_name) = ($msg->{type} =~ /SanMgr\/Switch\/SanSwitch\/(\S+?)\/Load_Failed/);
        $sm_switch_obj->_set_err("The SanSwitch object for switch $switch_name failed to instantiate.");
        $sm_switch_obj->{parent}->{listeners_complete_count}++;
        return $SanMgr::FALSE;
    };

    my $fail_listener = {filter => $fail_filter, callback => $fail_callback, sm_switch_obj => $self};

    return ($listener, $fail_listener, @procs);
}

sub _load_switch {
    # fetch params passed into the process
    my $global_params = NATE::ParamSet->get_global();
    my $switches = $global_params->get("switches_ds");
    my $switch = $global_params->get("switch");
    my $debug_level = $global_params->get("debug_level");
    my $staf_parent_handle = $global_params->get("staf_parent_handle");
    my $staf_parent_endpoint = $global_params->get("staf_parent_endpoint");

    my $parent = NATE::STAF::Peer->new(handle => $staf_parent_handle,
                                       endpoint => $staf_parent_endpoint);
    # Create switch object.
    my $sw_obj = SanSwitch->new(-host             => $switches->{$switch}->{NAME},
                                -conn_type        => $switches->{$switch}->{CONN_TYPE},
                                -admin_domain     => $switches->{$switch}->{ADMIN_DOMAIN},
                                -port_retries     => $switches->{$switch}->{PORT_RETRIES},
                                -verify_on_enable => $switches->{$switch}->{VERIFY_ON_ENABLE},
                                -persistent       => 1,
                                -debug_level      => $debug_level
                            );
    if ($sw_obj->error()) {
        $parent->message_put(type => "SanMgr/Switch/SanSwitch/" . $switches->{$switch}->{NAME} . "/Load_Failed",
                             message => {});
        NATE::BaseException->throw("Failed to create SanSwitch object for " . $switches->{$switch}->{NAME} . " : " .
                                   $sw_obj->errmsg());
        return $SanMgr::FALSE;
    }

    # destroy the connection to the object,
    # we have to do this or else the object we send back to the parent gets messed up
    $sw_obj->disconnect();

    # calculate the md5 for the switch object
    my $obj_md5_hex = San::SanMgr::Util->_calculate_obj_md5_hash($sw_obj);

    $parent->message_put(type => "SanMgr/Switch/SanSwitch/" . $switches->{$switch}->{NAME} . "/Load_Complete",
                         message => {obj_md5_hex => $obj_md5_hex,
                                     obj_serialized => Data::Dump::dump($sw_obj)});

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    _subtest_event_handler
# Objective: Example subtest event handler
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _subtest_event_handler {
    my ($subtest, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $message = &ParamsMgr::get_arg(-arg => "message",  -required => 1, -default => "");
    my $parent = &ParamsMgr::get_arg(-arg => "parent",  -required => 0, -default => {});
    if (&ParamsMgr::error()) {
        $subtest->_set_err(&ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    $subtest->sanlog(-msg => "Read from subtest: \n" . Dumper($message) . "\n",
                     -msg_level => 3);

    return $SanMgr::TRUE;
}

1;
