#
# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide a utility package for SanMgr
## @author latif@netapp.com
## @status private
## @pod here

package San::SanMgr::Util;

# Compiler directives.
use strict;
use warnings;

# Module import.
use Tharn;
use Error;
use Digest::MD5;
use Data::Dumper;

# NACL Libs
use NACL::STask::Cluster;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# San specific modules
use San::ControllerCluster;
use San::SanLog;

# EnGenX libs
use EnGenX::Util;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    get_objects
# Objective: Returns objects to caller matching type and name (if requested)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_objects {
    my ($self, %args) = @_;

    return $self->_get_objs_sorter('obj_class' => "controller_cluster",
                                   %args);
}

###############################################################################
# Method:    get_nacl_objects
# Objective: Returns supported NACL objects to the caller matching type and
#            name (if requested)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nacl_objects {
    my ($self, %args) = @_;

    return $self->_get_objs_sorter('obj_class' => "nacl",
                                   %args);
}

###############################################################################
# Method:    filter_objects
# Objective: filters a get_objects call by a filter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_clients_with_valid_device_list {

    my ($self) = @_;
    my @filtered_array;
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return {};
    }
    my @client_objs = $self->get_objects( -type => "client");
    my $client_dev_list = $self->get_client_dev_list();
    foreach my $client (@client_objs){
        foreach my $key (keys (%{$client_dev_list})){
            if ($key eq $client->{host}){
                push (@filtered_array, $client);
            }
        }
    }
    return (wantarray()) ? @filtered_array : \@filtered_array;
}
###############################################################################
# Method:    get_target_list
# Objective: Returns hash of iscsi/fcp targets and details about VServers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_target_list {
    my ($self, %args) = @_;
    return $self->EnGenX::Util::get_target_list(%args);
}

###############################################################################
# Method:    debug_level
# Objective: Set SanMgr debug level and return it to caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub debug_level {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $debug_level = &ParamsMgr::get_arg(-arg => "debug_level", -required => 0, -default => 0);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return 0;
    }

    $self->{debug_level} = $debug_level;

    return $debug_level;
}

###############################################################################
# Method:    get_error
# Objective: Return the errors string on the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_error {
    my ($self) = @_;

    return $self->errmsg();
}

###############################################################################
# Method:    clear_error
# Objective: Clear the error string(s) set on the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clear_error {
    my ($self) = @_;

    $self->_clear_err();

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    sanlog
# Objective: Wrapper to SanLog::sanlog() function
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sanlog {
    my ($self, %args) = @_;

    my ($msg_type, $msg, $msg_level, $enter, $exit, $add_call_stack);

    &ParamsMgr::process_args(%args);
    $enter = &ParamsMgr::get_arg(-arg => "enter", -required => 0, -default => 0);
    $exit = &ParamsMgr::get_arg(-arg => "exit", -required => 0, -default => 0);
    if ($enter == 1 or $exit == 1) {
        # If enter or exit is either specified, msg is not required, if not defined
        # default it to an empty string
        $msg = &ParamsMgr::get_arg(-arg => "msg", -required => 0, -default => "");
    } else {
        $msg = &ParamsMgr::get_arg(-arg => "msg", -required => 1, -default => "");
    }

    $msg_type = &ParamsMgr::get_arg(-arg => "msg_type", -required => 0, -default => "");
    $msg_level = &ParamsMgr::get_arg(-arg => "msg_level", -required => 0, -default => 0);
    $add_call_stack = &ParamsMgr::get_arg(-arg => "add_call_stack", -required => 0, -default => 0);
    if (&ParamsMgr::error()) {
        return $SanMgr::FALSE;
    }

    # If the message is empty that could mean that the user specified an enter or exit flag
    # default it to _parent_caller
    if ($msg eq "") {
        $msg = $self->_parent_caller() . "()";
    } else {
        # Prepend the caller name/description to the msg
        $msg = $self->_parent_caller() . "(): " . $msg;
    }

    # Determine if the caller called with an enter or exit request
    # Also set the msg_level to 3 if this is an enter/exit request
    if ($enter == 1) {
        $msg = "[Function Enter] " . $msg;
        $msg_level = 3;
    } elsif ($exit == 1) {
        $msg = "[Function Exit] " . $msg;
        $msg_level = 3;
    }

    # If the message type is warn then call SanUtil::nate_log with
    # warning type and message
    if ($msg_type =~ /warn/i) {
        &SanUtil::nate_log('-warn', $msg);
        return $SanMgr::TRUE;
    }

    # Call sanlog with the msg and msg_level
    $self->SanLog::sanlog(-msg            => $msg,
                          -msg_level      => $msg_level,
                          -add_call_stack => $add_call_stack);

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    store_client_dev_list
# Objective: Store client device list on object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub store_client_dev_list {
    my ($self, %args) = @_;

    $self->sanlog(-enter => 1);

    &ParamsMgr::process_args(%args);
    my $device_list = &ParamsMgr::get_arg(-arg => "device_list",
                                          -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SanMgr::FALSE;
    }

    if (keys(%{$self->{_client_dev_list}})){
        foreach my $client (keys(%{$device_list})) {
            $self->{_client_dev_list}->{$client} = $device_list->{$client};
        }
    }else{

        $self->{_client_dev_list} = $device_list;
    }
    $self->sanlog(-exit => 1);

    return $SanMgr::TRUE;
}

###############################################################################
# Method:    get_client_dev_list
# Objective: Get client device list from internal object.
#            Also allows caller to filter on attributes of client device list.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_client_dev_list {
    my ($self, %args) = @_;

    $self->sanlog(-enter => 1);

    $self->clear_error();

    my ($cdl, $ncdl);

    &ParamsMgr::process_args(%args);

    my $filter_client   = &ParamsMgr::get_arg(-arg => "client",     -required => 0, -default => undef);
    my $filter_vserver  = &ParamsMgr::get_arg(-arg => "vserver",    -required => 0, -default => undef);
    my $filter_volume   = &ParamsMgr::get_arg(-arg => "volume",     -required => 0, -default => undef);
    my $filter_lun      = &ParamsMgr::get_arg(-arg => "lun",        -required => 0, -default => undef);
    my $filter_protocol = &ParamsMgr::get_arg(-arg => "protocol",   -required => 0, -default => undef);
    my $filter_serial   = &ParamsMgr::get_arg(-arg => "serial",     -required => 0, -default => undef);
    my $cdl_format      = &ParamsMgr::get_arg(-arg => "cdl_format", -required => 0, -default => $SanMgr::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return undef;
    }

    if (not defined($self->{_client_dev_list})) {
        $self->_set_err("Client device list is not defined!");
        return undef;
    }

    $cdl = $self->{_client_dev_list};

    # Go through each client in the client device list and apply a client filter if
    # requested by caller
    foreach my $client (keys(%{$cdl})) {
        # Check filter_client, if its not empty, process the filter
        if (defined($filter_client) and $filter_client ne "") {
            # If the client matches the filter client, save for returning
            if ($filter_client =~ $client) {
                $ncdl->{$client} = $cdl->{$client};
            }
        } else {
            # No filter specified, save all clients for returning
            $ncdl->{$client} = $cdl->{$client};
        }
    }

    # Filter out the client dev list based on the lun attributes
    $ncdl = $self->_filter_cdl(-filter_key => "vserver",
                               -filter_str=> $filter_vserver,
                               -filter_cdl => $ncdl);
    $ncdl = $self->_filter_cdl(-filter_key => "volume",
                               -filter_str=> $filter_volume,
                               -filter_cdl => $ncdl);
    $ncdl = $self->_filter_cdl(-filter_key => "lun",
                               -filter_str=> $filter_lun,
                               -filter_cdl => $ncdl);
    $ncdl = $self->_filter_cdl(-filter_key => "protocol",
                               -filter_str=> $filter_protocol,
                               -filter_cdl => $ncdl);
    $ncdl = $self->_filter_cdl(-filter_key => "serial",
                               -filter_str=> $filter_serial,
                               -filter_cdl => $ncdl);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to filter client device list:", &ParamsMgr::get_error());
        return undef;
    }

    # Check to see if anything is getting returned, if not set an error
    if (scalar(keys(%{$ncdl})) == 0) {
        $self->_set_err("Client device list is empty or filter was not found");
        return undef;
    }

    $self->sanlog(-exit => 1);

    if ($cdl_format == $SanMgr::TRUE) {
        my $final_client_dev_list;

        foreach my $client (keys %{$ncdl}) {
            $final_client_dev_list .= $client . ",";
            foreach my $cdl_entry (@{$ncdl->{$client}}) {
                my $path = $cdl_entry->{path};
                $final_client_dev_list .= $path . ",";
            }
            chop($final_client_dev_list);
            $final_client_dev_list .= ";";
        }
        chop($final_client_dev_list);
        return $final_client_dev_list;
    }

    return $ncdl;
}

###############################################################################
# Method:    _filter_cdl
# Objective: _filter_cdl returns a filter client dev list based on attributes
#            on each device within each client
###############################################################################
sub _filter_cdl {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $filter_key = &ParamsMgr::get_arg(-arg => "filter_key", -required => 1, -default => "");
    my $filter_str = &ParamsMgr::get_arg(-arg => "filter_str", -required => 0, -default => undef);
    my $filter_cdl = &ParamsMgr::get_arg(-arg => "filter_cdl", -required => 0, -default => {});

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return {};
    }

    my $fcdl;

    # If the filter string is undefined, just return the cdl
    if (not defined($filter_str)) {
        return $filter_cdl;
    }

    # Go through each client in the client device list and apply a vserver filter if
    # requested by caller
    foreach my $client (keys(%{$filter_cdl})) {
        foreach my $lun (@{$filter_cdl->{$client}}) {
            # Determine if the filter_key contains any valid data
            if (defined($lun->{$filter_key})) {
                if (ref($lun->{$filter_key}) eq "ARRAY") {
                    # The data is an array, process it
                    my $found = 0;
                    foreach my $item (@{$lun->{$filter_key}}) {
                        if ($item =~ $filter_str) {
                            $found = 1;
                            last;
                        }
                    }
                    if ($found == 1) {
                        push(@{$fcdl->{$client}}, $lun);
                    }
                } else {
                    # The data is a string
                    if ($lun->{$filter_key} =~ $filter_str) {
                        push(@{$fcdl->{$client}}, $lun);
                    }
                }
            }
        }
    }

    return $fcdl;
}

###############################################################################
# Method:    _parent_caller
# Objective: _parent_caller returns the subroutine of caller of depth 2
# Details:   format for the caller:
#      0         1          2      3            4
#      ($package, $filename, $line, $subroutine, $hasargs,
#      5          6          7            8       9         10
#      $wantarray, $evaltext, $is_require, $hints, $bitmask, $hinthash) = caller($i);
#      $i is the caller depth
###############################################################################
sub parent_caller { (caller(2))[3] }

#==============================================================================
# Method:    _get_controller_objects
# Objective: Get controller objects created by SanMgr querying the cluster
# Details:   None.
#==============================================================================
sub _get_controller_objects {
    my ($self, %args) = @_;

    $self->clear_error();
    $self->sanlog(-enter => 1);

    my ($return)  = {};
    my (@objects) = ();
    my $cserver_obj = undef;
    my ($obj_exists);

    &ParamsMgr::process_args(%args);
    my $type  = &ParamsMgr::get_arg(-arg => "type", -required => 1);
    my $data  = &ParamsMgr::get_arg(-arg => "data", -required => 1);
    my $obj_class= &ParamsMgr::get_arg(-arg => "obj_class",-required => 0,
                                       -default => "controller_cluster");
    my @names = &ParamsMgr::get_arg(-arg      => "names",
                                    -ref_type => "ARRAY",
                                    -default  => []);
    my $vserver = &ParamsMgr::get_arg(-arg      => "vserver",
                                      -required => 0,
                                      -default  => "");
    my $volume_filter = &ParamsMgr::get_arg(-arg      => "volume_filter",
                                            -default  => "");
    my $qtree_filter = &ParamsMgr::get_arg(-arg      => "qtree_filter",
                                            -default  => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    if (ref($data->{$type}) =~ /ARRAY/i) {
        foreach my $obj (@{$data->{$type}}) {
            # Check obj vserver name and skip this entry if it doesn't
            # match a passed in vserver_name (if one's passed in)
            next if ((length($vserver)) and ($obj->{vserver} ne $vserver));

            # skip this object if they passed in a volume/qtree filter and the
            # object has a volume/qtree associated to it and the filter doesnt match
            next if (length($volume_filter) and exists($obj->{volume}) and $obj->{volume} ne $volume_filter);
            next if (length($qtree_filter)  and exists($obj->{qtree})  and $obj->{qtree}  ne $qtree_filter);

            # Check if users want all objects or only those matching name
            my $obj_name = $obj->{name};
            if (scalar(@names) == 0 or grep(/^$obj_name$/i, @names)) {
                if ($obj_class =~/nacl/i){
                     my $nacl_obj = $self->_retrieve_nacl_obj(object=>$obj,
                                                              %args);
                     push(@objects, $nacl_obj);

                } else {
                    my $cserver = $obj->{cserver};
                    ($cserver_obj) = $self->get_objects(types => ["node", "cserver"],
                                                        names => [$cserver]);
                    if ($self->error()) {
                        $self->_set_err("Unable to get node/cserver object for '$cserver'");
                        return (wantarray()) ? () : [];
                    }
                    if ($type =~ /aggr|aggregate/i) {
                        $obj_exists = $cserver_obj->aggr_exist($obj->{name});
                        $cserver_obj->_clear_err();
                        if ($obj_exists) {
                            my $aggr_obj =
                            $cserver_obj->get_aggregate(aggregate => $obj->{name});
                            if ($cserver_obj->error()) {
                                $self->_set_err("Failed to get aggregate object for $obj->{name}", $cserver_obj->errmsg());
                                return (wantarray()) ? () : [];
                            }
                            push(@objects, $aggr_obj);
                        }
                    } elsif ($type =~ /vserver/i) {
                        $obj_exists = $cserver_obj->vserver_exist(vserver => $obj->{name});
                        $cserver_obj->_clear_err();
                        if ($obj_exists) {
                            my $vserver_obj =
                            $cserver_obj->get_vserver_obj(vserver => $obj->{name});
                            if ($cserver_obj->error()) {
                                $self->_set_err("Failed to get vserver object for $obj->{name}", $cserver_obj->errmsg());
                                return (wantarray()) ? () : [];
                            }
                            # NOTE: iscsi_targets and fcp_targets have to be
                            # appended to every vserver object
                            if (defined($obj->{iscsi_lif_data})) {
                                $vserver_obj->{iscsi_lif_data} = $obj->{iscsi_lif_data};
                            }

                            if (defined($obj->{fcp_lif_data})) {
                                $vserver_obj->{fcp_lif_data} = $obj->{fcp_lif_data};
                            }
                            push(@objects, $vserver_obj);
                        }
                    } elsif ($type =~ /volume|vol/i) {
                        $obj_exists =
                        $cserver_obj->volume_exist(name => $obj->{name},
                                                   vserver => $obj->{vserver});
                        $cserver_obj->_clear_err();
                        if ($obj_exists) {
                            my $vol_obj =
                            $cserver_obj->get_vol(vserver => $obj->{vserver},
                                                  volume => $obj->{name});
                            if ($cserver_obj->error()) {
                                $self->_set_err("Failed to get volume object for $obj->{name}", $cserver_obj->errmsg());
                                return (wantarray()) ? () : [];
                            }
                            push(@objects, $vol_obj);
                        }
                    } elsif ($type =~ /igroup/i) {
                        $obj_exists = $cserver_obj->igroup_exist(name    => $obj->{name},
                                                                 vserver => $obj->{vserver});
                        $cserver_obj->_clear_err();
                        if ($obj_exists) {
                            my $igroup_obj =
                            $cserver_obj->get_igroup(vserver => $obj->{vserver},
                                                     igroup  => $obj->{name});
                            if ($cserver_obj->error()) {
                                $self->_set_err("Failed to get igroup object for $obj->{name}", $cserver_obj->errmsg());
                                return (wantarray()) ? () : [];
                            }
                            push(@objects, $igroup_obj);
                        }
                    } elsif ($type =~ /lun/i) {
                        $obj_exists =
                        $cserver_obj->volume_exist(name => $obj->{volume},
                                                   vserver => $obj->{vserver});
                        $cserver_obj->_clear_err();
                        if ($obj_exists) {
                            my $vol_obj =
                            $cserver_obj->get_vol(vserver => $obj->{vserver},
                                                  volume => $obj->{volume});
                            if ($cserver_obj->error()) {
                                $self->_set_err("Failed to get volume object for $obj->{volume}", $cserver_obj->errmsg());
                                return (wantarray()) ? () : [];
                            }

                            my $container_obj;
                            if(exists($obj->{qtree}) and $obj->{qtree}) {
                                $obj_exists =
                                  $vol_obj->lun_exist(vserver => $obj->{vserver},
                                                      qtree   => $obj->{qtree},
                                                      lun     => $obj->{name});
                                $vol_obj->_clear_err();
                                $container_obj = $vol_obj->get_qtree_obj( name => $obj->{qtree});
                            } else {
                                $obj_exists =
                                  $vol_obj->lun_exist(vserver => $obj->{vserver},
                                                      lun     => $obj->{name});
                                $vol_obj->_clear_err();
                                $container_obj = $vol_obj;
                            }

                            if ($obj_exists) {
                                my $lun_obj = $container_obj->get_lun(lun => $obj->{name},
                                                                      vserver => $obj->{vserver});
                                if ($container_obj->error()) {
                                    $self->_set_err("Failed to get LUN object for $obj->{name}", $container_obj->errmsg());
                                    return (wantarray()) ? () : [];
                                }
                                push(@objects, $lun_obj);
                            }
                        }
                    } elsif ($type =~ /qtree/i) {
                        $obj_exists =
                        $cserver_obj->volume_exist(name => $obj->{volume},
                                                   vserver => $obj->{vserver});
                        $cserver_obj->_clear_err();
                        if ($obj_exists) {
                            my $vol_obj =
                            $cserver_obj->get_vol(vserver => $obj->{vserver},
                                                  volume => $obj->{volume});
                            if ($cserver_obj->error()) {
                                $self->_set_err("Failed to get volume object for $obj->{volume}", $cserver_obj->errmsg());
                                return (wantarray()) ? () : [];
                            }
                            $obj_exists =
                            $vol_obj->qtree_exist(qtree => $obj->{name});
                            $vol_obj->_clear_err();
                            if ($obj_exists) {
                                my $qtree_obj = $vol_obj->get_qtree_obj(name => $obj->{name});
                                if ($vol_obj->error()) {
                                    $self->_set_err("Failed to get Qtree object for $obj->{name}", $vol_obj->errmsg());
                                    return (wantarray()) ? () : [];
                                }
                                push(@objects, $qtree_obj);
                            }
                        }
                    } elsif ($type =~ /portset/i) {
                        my $vserver_obj =
                        $cserver_obj->get_vserver_obj(name => $obj->{vserver});
                        if ($cserver_obj->error()) {
                            $self->_set_err("Failed to get vserver object for $obj->{vserver}", $cserver_obj->errmsg());
                            return (wantarray()) ? () : [];
                        }
                        $obj_exists =
                        $vserver_obj->portset_exist(portset => $obj->{name});
                        $vserver_obj->_clear_err();
                        if ($obj_exists) {
                            my $portset_obj =
                            $vserver_obj->get_portset_obj(portset => $obj->{name});
                            if ($vserver_obj->error()) {
                                $self->_set_err("Failed to get porset object for $obj->{name}", $vserver_obj->errmsg());
                                return (wantarray()) ? () : [];
                            }
                            push(@objects, $portset_obj);
                        }
                    } elsif ($type =~ /tpgroup/i) {
                        my $vserver_obj =
                        $cserver_obj->get_vserver_obj(name => $obj->{vserver});
                        if ($cserver_obj->error()) {
                            $self->_set_err("Failed to get vserver object for $obj->{vserver}", $cserver_obj->errmsg());
                            return (wantarray()) ? () : [];
                        }
                        my $iscsi_obj = $vserver_obj->get_iscsi_obj();
                        $vserver_obj->_clear_err();
                        $obj_exists =
                        $iscsi_obj->tpgroup_exist(tpgroup => $obj->{name});
                        if ($obj_exists) {
                            my $tpgroup_obj =
                            $iscsi_obj->get_tpgroup_obj(tpgroup => $obj->{name});
                            push(@objects, $tpgroup_obj);
                        }
                    } elsif ($type =~ /lif/i) {
                        my $vserver_obj =
                        $cserver_obj->get_vserver_obj(name => $obj->{vserver});
                        if ($cserver_obj->error()) {
                            $self->_set_err("Failed to get vserver object for $obj->{vserver}", $cserver_obj->errmsg());
                            return (wantarray()) ? () : [];
                        }
                        $obj_exists =
                        $vserver_obj->lif_exist(lif => $obj->{name});
                        $vserver_obj->_clear_err();
                        if ($obj_exists) {
                            my $lif_obj = $vserver_obj->get_lif_obj(lif => $obj->{name});
                            if ($vserver_obj->error()) {
                                $self->_set_err("Failed to get lif object for $obj->{name}", $vserver_obj->errmsg());
                                return (wantarray()) ? () : [];
                            }
                            push(@objects, $lif_obj);
                        }
                    } #if type
                }# if obj_class
            }# if names
        }#foreach
    } else {
        $self->_set_err("Provided Data->{TYPE} is not a reference to an ARRAY.".
                        "The reference type is: " . ref($data->{$type}));
        return (wantarray()) ? () : [];
    }#if array

    return (wantarray()) ? @objects : \@objects;
}

###############################################################################
# Method:    get_vms_associated_to_hypervisor
# Objective: Get the list of VMs that are associated to a specific Hypervisor
###############################################################################
sub get_vms_associated_to_hypervisor {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $hypervisor_name = &ParamsMgr::get_arg(-arg => "hypervisor_name",
                                              -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    my @vm_names;

    my $esx_hypervisors_xml_hash = &ParamsMgr::search(-search_path => [],
                                                      -key => "HYPERVISORS",
                                                      -key_filter => qr/^ESX\d*$/,
                                                      paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    my $hyperv_hypervisors_xml_hash = &ParamsMgr::search(-search_path => [],
                                                         -key => "HYPERVISORS",
                                                         -key_filter => qr/^HYPERV\d*$/,
                                                         paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    my $cxs_hypervisors_xml_hash = &ParamsMgr::search(-search_path => [],
                                                      -key => "HYPERVISORS",
                                                      -key_filter => qr/^CXS\d*$/,
                                                      paramset => $self->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    foreach my $hypervisor_hash (values %{$esx_hypervisors_xml_hash}, values %{$hyperv_hypervisors_xml_hash},
                                 values %{$cxs_hypervisors_xml_hash}) {
        if ($hypervisor_name eq $hypervisor_hash->{NAME}) {
            # extract the VM names inside of this object
            # and store on the vm_names array
            foreach my $key (keys %{$hypervisor_hash}) {
                if ($key =~ /^VM\d*$/) {
                    push(@vm_names, $hypervisor_hash->{$key}->{NAME});
                }
            }
            last;
        }
    }

    return (wantarray()) ? @vm_names : \@vm_names;
}

###############################################################################
# Method:    get_luns_associated_to_client
# Objective: Get the list of luns that are associated to
#            a specified client
###############################################################################
sub get_luns_associated_to_client {
    my ($self, %args) = @_;
    return $self->EnGenX::Util::get_luns_associated_to_client(%args);
}

###############################################################################
# Method:    get_luns_and_vserver_associated_to_client
# Objective: Get the list of luns that are associated to
#            a specified client
###############################################################################
sub get_luns_and_vservers_associated_to_client {
    my ($self, %args) = @_;
    return $self->EnGenX::Util::get_luns_and_vservers_associated_to_client(%args);
}

###############################################################################
# Method:    get_vservers_associated_to_client
# Objective: Get the list of vservers that are associated to a specified client
###############################################################################
sub get_vservers_associated_to_client {
    my ($self, %args) = @_;
    return $self->EnGenX::Util::get_vservers_associated_to_client(%args);
}

###############################################################################
# Method:    get_clusters_associated_to_client
# Objective: Get the list of cluster names that are associated to
#            a specified client
###############################################################################
sub get_clusters_associated_to_client {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # generate the relationship hash
    my ($hash) = $self->get_vserver_client_relationship_hash();
    if ($self->error()) {
        $self->_set_err("Failed to generate vserver client relationship hash", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    my @clusters = keys %{$hash->{$client_name}};

    # pull out all the clusters that are related to this client
    return (wantarray()) ? @clusters : \@clusters;
}

###############################################################################
# Method:    get_clients_associated_to_vserver
# Objective: Get the clients that are associated to a vserver/cluster
###############################################################################
sub get_clients_associated_to_vserver {
    my ($self, %args) = @_;
    return $self->EnGenX::Util::get_clients_associated_to_vserver(%args);
}

###############################################################################
# Method:    get_client_relationship_hash
# Objective: get the client relationship hash for a particular client
###############################################################################
sub get_client_relationship_hash {
    my ($self, %args) = @_;
    return $self->EnGenX::Util::get_client_relationship_hash(%args);
}

###############################################################################
# Method:    get_vserver_names
# Objective: return a filtered list of all vservers on a sanmgr object
###############################################################################
sub get_vserver_names {
    my ($self, %args) = @_;
    return $self->EnGenX::Util::get_vserver_names(%args);
}


###############################################################################
# Method:    get_vserver_client_relationship_hash
# Objective: Build a hash mapping clients with vservers they will be connected to.
###############################################################################
sub get_vserver_client_relationship_hash {
    my ($self, %args) = @_;
    return $self->EnGenX::Util::get_vserver_client_relationship_hash(%args);
}

###############################################################################
# Method:    normalize_fcp_wwn
# Objective: Normalize an FCP WWN (add colons, lowercase it)
###############################################################################
sub normalize_fcp_wwn {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $wwn = exists($args{"wwn"}) ? delete($args{"wwn"}) : delete($args{"-wwn"});
    $self->_set_err("Missing required_parameter") and return if (not defined($wwn));

    return lc(&SanUtil::wwn_add_colons(-wwn => $wwn));
}

###############################################################################
# Method:    assign_sanclient_options
# Objective: Given a sanclient object and it's paramset, assign all sub options
###############################################################################
sub assign_sanclient_options {
    my ($self, %args) = @_;

    my $ret = $SanMgr::FALSE;
    try {
        $ret = $self->EnGenX::Util::assign_sanclient_options(%args);
    } otherwise {
        my $e = shift;
        if (ref($self)) {
            $self->_set_err("Exception caught: " . $e->text());
        } else {
            $e->throw();
        }
    };

    return $ret;
}

###############################################################################
# Method:    reconnect_all_controllers
# Objective: Reconnects to controller objects in this test
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect_all_controllers {
    my ($self, %args) = @_;

    $self->sanlog(-enter => 1);

    # reconnect to all node and cserver objects
    my (@node_objs) = $self->get_objects(type => "node");
    if ($self->error()) {
        $self->_set_err("Unable to obtain node objects");
        return $SanMgr::FALSE;
    }
    my (@cserver_objs) = $self->get_objects(type => "cserver");
    if ($self->error()) {
        $self->_set_err("Unable to obtain cserver objects");
        return $SanMgr::FALSE;
    }

    foreach my $cserver_obj (@cserver_objs, @node_objs) {
        # reconnect to this filer object
        $cserver_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
        if ($cserver_obj->error()) {
            $self->_set_err("Failed to reconnect to node", $cserver_obj->errmsg());
            return $SanMgr::FALSE;
        }
    }

    $self->sanlog(-exit => 1);

    return $SanMgr::TRUE;
}

sub _init_vs_client_relationship_svc {
    return shift->EnGenX::Util::_init_vs_client_relationship_svc(@_);
}

sub _get_all_client_paramsets {
    return shift->EnGenX::Util::_get_all_client_paramsets(@_);
}

sub _get_all_clients_paramsets {
    return shift->EnGenX::Util::_get_all_clients_paramsets(@_);
}

#==============================================================================
# Method:    _get_objs_sorter
# Objective: Common code between ControllerCluster get_objects method and a new
#            NACL get_nacl_objs method.
# Details:   This is a private method to avoid duplication of code.
#==============================================================================
sub _get_objs_sorter {
    my ($self, %args) = @_;

    $self->_clear_err();
    $self->sanlog(-enter => 1);

    # Locals
    my @return= ();
    my %params=();
    my @errors= ();

    #Input Parameters
    &ParamsMgr::process_args(%args);
    # Need to know what class is the object the user is requesting. This could be
    # either a ControllerCluster instance or a NACL intance.
    my $type_arg = &ParamsMgr::get_arg(-arg => "type",     -required => 0, -default => undef);
    my @types    = &ParamsMgr::get_arg(-arg => "types",    -required => 0, -default => []);
    my $name     = &ParamsMgr::get_arg(-arg => "name",     -required => 0, -default => undef);
    my @names    = &ParamsMgr::get_arg(-arg => "names",    -ref_type => "ARRAY",
                                       -default => []);
    my $obj_class= &ParamsMgr::get_arg(-arg => "obj_class",-required => 0,
                                       -default => "controller_cluster");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $self->sanlog(-exit => 1);
        return (wantarray()) ? () : [];
    }

    # Either -type or -types must be specified.
    # Will consolidate both in the array (@types) and work from there, but we need to
    # verify that 'type' has been defined before pushing it to the array. Failing to do so,
    # we could be adding an 'undef' entry in the array that will cause the method to fail
    if (defined($type_arg)){
        push(@types, $type_arg);
    } else {
        if(not scalar(@types)) {
            $self->_set_err("-type or -types must be specified");
            $self->sanlog(-exit => 1);
            return (wantarray()) ? () : [];
        }
    }

    # If -name is passed in then push it onto the @names array.
    if (defined($name)){
        push(@names, $name);
    }

    # Get together all the data needed
    # If method is called from setup or cleanup, the path to objects
    # is slightly different
    if (not defined $self->{parent} or not defined($self->{parent}->{_objects})) {
        if (defined($self->{_objects})) {
            $params{data} = $self->{_objects}
        } else {
            $self->_set_err("Objects not defined in SanMgr.");
            $self->sanlog(-exit => 1);
            return (wantarray()) ? () : [];
        }
    } else {
        $params{data} = $self->{parent}->{_objects};
    }

    # Add names to params
    $params{names} = \@names;

    # Get and Validate the data to work with.
    # Validate the Objects_class request.
    if ($obj_class =~ /controller_cluster|nacl/i) {
        $params{obj_class} = $obj_class;
    } else {
        $self->_set_err("Input parameter 'obj_class' has a not recognized value. ".
                        "Supported classes are ControllerCluster (controller_cluster) ".
                        ".or NACL (nacl)");
        $self->sanlog(-exit => 1);
        return (wantarray()) ? () : [];
    }

    # Use the gathered data to get each TYPE of object
    foreach my $type (@types) {
        # Check $self to see if it has the list type defined on it
        if (defined($params{data}->{$type})) {
            $params{type}=$type;
        } else {
            my $err_msg = ("\tObject type ($type) not defined on cached data");
            push(@errors, $err_msg);
            next;
        }

        # Differentiate between ControllerCluster and NACL objects.
        # TODO: When NACL supports Clients, Switches, Hypervisors, etc. then this code
        #       could be merged into one single method.
        #
        # Since NACL does not support this classes, discard the user's Object_Class choice
        # and force to return ControllerCluster classes
        if ($type =~ /client|switch|hypervisor|fabric|eventman/i) {
            $params{obj_class} = "controller_cluster";
            # Private Method call
            push(@return, $self->_get_cached_objs(%params) );
        }
        # The following objects are supported in both NACL and ControllerCluster. Pass the user's
        # argument that could hold a choice for Object_Class
        elsif($type =~ /cserver|node|cluster/i) {
            # Private Method call
            push(@return, $self->_get_cached_objs(%params) );
        }
        # Not all objects are cached, therefore the information to create them either for
        # ControllerCluster or NACL, need to get additional information.
        else {
            push(@return, $self->_get_controller_objects(%params,%args) );
        }
        # keep track of all errors that had happened
        if ($self->error()) {
            push(@errors, $self->errmsg());
            $self->_clear_err();
        }
    }


    if(!(scalar(@return)) && scalar(@errors)){
        $self->_set_err("Could not find an object with the provided types and names: ",
                        @errors);
        $self->sanlog(-exit => 1);
        return (wantarray()) ? () : [];
    }

    $self->sanlog(-exit => 1);
    return (wantarray()) ? @return : \@return;
}

#==============================================================================
# Method:    _get_cached_objs
# Objective: Common code that is been used by several conditions in _obj_sorter
#           to retrieve objects that are cached. This was created to avoid
#           duplication of code.
#==============================================================================
sub _get_cached_objs {
    my ($self, %args) = @_;

    $self->_clear_err();
    $self->sanlog(-enter => 1);

    # Locals
    my @return= ();

    #Input Parameters
    &ParamsMgr::process_args(%args);
    my $type     = &ParamsMgr::get_arg(-arg => "type",     -required => 1);
    my $data     = &ParamsMgr::get_arg(-arg => "data",     -required => 1);
    my $obj_class= &ParamsMgr::get_arg(-arg => "obj_class",-required => 1);
    my @names    = &ParamsMgr::get_arg(-arg => "names",    -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # Determine if caller requested a particular set of objects from the list
    if (scalar(@names) > 0) {
        foreach my $name (@names) {
            # Check to see if the object/name exists
            if ($obj_class =~ /nacl/i and $type =~/node/i
                and not defined($data->{'nacl_node'}->{$name})){
                $self->_set_err("Invalid name specified ($name) for type ($type)");
                $self->sanlog(-exit => 1);
                return (wantarray()) ? () : [];
            } elsif ($obj_class =~ /nacl/i and $type =~/cluster/i
                and not defined($data->{'nacl_cluster'}->{$name})) {
                $self->_set_err("Invalid name specified ($name) for type ($type)");
                $self->sanlog(-exit => 1);
                return (wantarray()) ? () : [];
            } elsif ($obj_class !~ /nacl/i and not defined($data->{$type}->{$name})) {
                $self->_set_err("Invalid name specified ($name) for type ($type)");
                $self->sanlog(-exit => 1);
                return (wantarray()) ? () : [];
            } elsif ($type !~/node/i and $type !~/cluster/i and not defined($data->{$type}->{$name})) {
                $self->_set_err("Invalid name specified ($name) for type ($type)");
                $self->sanlog(-exit => 1);
                return (wantarray()) ? () : [];
            } else {
                if ($obj_class =~/nacl/i){
                    # NACL::C::Node and NACL::STask::Cluster are stored
                    # in the SanMgr Object.
                    if($type =~/node/i){
                        push(@return, $data->{nacl_node}->{$name});
                    } elsif ($type =~/cluster/i){
                        push(@return, $data->{nacl_cluster}->{$name});
                    } else {
                        my $nacl_obj = $self->_retrieve_nacl_obj(name=>$name,
                                                                type=>$type);
                        push(@return, $nacl_obj);
                    }
                } else {
                    push(@return, $data->{$type}->{$name});
                }
            }
        }
    } else {
        # Return the entire list
        if ($obj_class =~ /nacl/i and $type =~/node/i){
            @names = keys %{$data->{'nacl_node'}};
        } elsif ($obj_class =~ /nacl/i and $type =~/cluster/i) {
            @names = keys %{$data->{'nacl_cluster'}};
        } else {
            @names = keys %{$data->{$type}};
        }
        foreach my $name (@names) {
            if ($obj_class =~/nacl/i){
                # NACL::C::Node and NACL::STask::Cluster are stored
                # in the SanMgr Object.
                if($type =~/node/i){
                    push(@return, $data->{nacl_node}->{$name});
                } elsif ($type =~/cluster/i){
                    push(@return, $data->{nacl_cluster}->{$name});
                } else {
                    my $nacl_obj = $self->_retrieve_nacl_obj(name=>$name,
                                                             type=>$type);
                    push(@return, $nacl_obj);
                }
            } else {
                push(@return, $data->{$type}->{$name});
            }
        }
    }

    $self->sanlog(-exit => 1);
    return (wantarray()) ? @return : \@return;
 }

#==============================================================================
# Method:    _retrieve_nacl_obj
# Objective: Prepares (adds and removes) information that is not needed on the
#            EnGenX NACL call. Also, helps prevent code duplication.
#==============================================================================
sub _retrieve_nacl_obj {
    my ($self, %args) = @_;
    $self->sanlog(-enter => 1);
    # Locals
    my $obj = delete $args{object};
    my $nacl_obj = undef;
    # Delete arguments NACL does not need
    delete $args{data} if (defined $args{data});
    delete $args{obj_class} if (defined $args{obj_class});
    delete $args{names} if (defined $args{names});

    # EnGenX uses UpperCase keys, so add the NAME key.
    $obj->{NAME} = $obj->{name};
    # Add to the resource all the primary keys that were provided by the user
    my %resource = (%{$obj},
                    %args);
    my %params=(type=>$resource{type},
                resource=>\%resource);

    try{
        $nacl_obj = EnGenX::Util->get_nacl_obj(%params);
    }
    otherwise {
        my $exception = shift;
        $self->_set_err("Encountered a problem while creating a NACL object " .
                        "for object '".$obj->{name}."' type(".$obj->{type}.")\n" .
                        $exception
                    );

        return undef;
    };

    $self->sanlog(-exit => 1);
    return $nacl_obj;
}

#==============================================================================
# Method:    _validate_obj_md5_hash
# Objective: Given an object and its previous md5 hash, verify the object
#==============================================================================
sub _validate_obj_md5_hash {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $obj                = &ParamsMgr::get_arg(-arg => "obj",
                                                 -required => 1);
    my $obj_md5_hex_before = &ParamsMgr::get_arg(-arg => "obj_md5_hex_before",
                                                 -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument Parsing failed", &ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # calculate the new md5 value
    my $obj_md5_hex_after = $self->_calculate_obj_md5_hash($obj);

    # verify the md5
    if ($obj_md5_hex_after ne $obj_md5_hex_before) {
        $self->_set_err("MD5 verification for " . $obj->hostname() . " did not pass ($obj_md5_hex_before " .
                        "!= $obj_md5_hex_after). Please file a burt against TESTLIB/san to add support for this OS.");
        return $SanMgr::FALSE;
    }

    return $SanMgr::TRUE;
}

#==============================================================================
# Method:    _calculate_obj_md5_hash
# Objective: Given an object, calculate the md5 hash from it
#==============================================================================
sub _calculate_obj_md5_hash {
    my ($self, $obj) = @_;

    return $self->EnGenX::Util::_calculate_obj_md5_hash($obj);
}

=pod

=head1 Package Name

San::SanMgr::Util

=head1 Author

dl-san-automation

=head1 Description

Provide a common set of functions to be used with SanMgr

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>

B<Tharn.pm>

B<San/SanLog.pm>

=head1 Public Methods

=head2 B<get_objects($self, %args)>

This function is used to retrieve the resource (client, switch, controller,...) objects.
Also can be used to retrieve objects by name.

=over 4

=item Inputs

=item * -self

Self is an SanMgr object

=item * -type

Type can be one of the following:
    client, switch, controller, vserver

=item * I<-name>

Name is optional, this input will filter the type provided to locate only the object with the
name requested.

=item Outputs

Returns back a hash of objects, keyed either by type (if no name provided) OR by name if a
name is provided.

=item Examples

my $client1_object = $self->get_objects(-type => "client", -name => "client1");

OR

my $hash_of_clients = $self->get_objects(-type => "client");

=back


=head2 B<get_nacl_objects($self, %args)>

This function is used to retrieve NACL instances of the resource (client, switch, controller,...)
objects. Also can be used to retrieve objects by name. It has the same API as get_objects()


=over 4

=item Inputs

=item * -self

Self is an SanMgr object

=item * -type

The type could be one of the following, but NACL does not support all of them (as of now):
client, switch, controller, vserver

NOTE: NACL (and so does this method) currently only supports the following types:
        aggr|aggregate
        vserver
        volume|vol
        qtree
        lun
        igroup
        portset
        lif
        tpgroup
        node
        switch
        cserver
        cluster

=item * I<-name>

Name is optional, this input will filter the type provided to locate only the object with the
name requested.

=item Outputs

Returns back a hash of objects, keyed either by type (if no name provided) OR by name if a
name is provided.

=item Examples

my $cserver_object = $self->get_nacl_objects(type => "cserver",
                                             name => $cserver_name);

OR

my $volume_object = $self->get_nacl_objects(type => "volume",
                                            name => $vol_name);

=back

=head2 B<get_target_list($self, %args)>

This function is used to retrieve the target list available from the VServers
defined within the environment/testbed/XML file.

=over 4

=item Inputs

=item * -self

Self is an SanMgr object

=item Outputs

Returns a hash with three keys defined, one is iscsi_targets, this is an array ref of
iscsi targets. The second is fcp_targets, this is an array ref of fcp targets.
The third vserver_objs is an array ref of VServer objects.
    $data->{iscsi_targets} = \@iSCSITargets;
    $data->{fcp_targets} = \@FcpTargets;
    $data->{vserver_objs} = \@VServer_objs;

=item Examples

my $data = $self->get_target_list();
my @iscsi_targets = @{$data->{iscsi_targets}};
my @fcp_targets = @{$data->{fcp_targets}};
my @vserver_objs = @{$data->{vserver_objs}};

=back

=head2 B<get_vserver_names(%args)>

This function is used to retrieve the list of vserver names
defined within the environment/testbed/XML file.

=over 4

=item Inputs

=item * I<-cluster_name>

Cluster is an optional string to narrow the scope of the vserver request

=item * I<-type>

Type of vservers to obtain, valid values are (all|dummy|regular), default: "all"

=item Outputs

Returns an array of vserver names.

=item Examples

my @all_vserver_list = $self->get_vserver_names();
my @cluster_scoped_vserver_list = $self->get_vserver_names(-cluster_name => "cluster1");
my @cluster_scoped_regular_vserver_list = $self->get_vserver_names(-cluster_name => "cluster1",
                                                                   -type         => "regular");
my @cluster_scoped_dummy_vserver_list   = $self->get_vserver_names(-cluster_name => "cluster1",
                                                                   -type         => "dummy");

=back

=head2 B<debug_level($self, %args)>

This function is used to set and get the debug level on SanMgr (or sub objects)

=over 4

=item Inputs

=item * -self

Self is an SanMgr object

=item * I<-debug_level>

Debug level is used to set the internal debug level to what is provided.

=item Outputs

Returns back the set debug level.

=item Examples

Get current debug level
my $debug_level = $self->debug_level();

OR

Set debug level

$self->debug_level(-debug_level => 10);

=back

=head2 B<get_clients_associated_to_vserver(%args)>

Obtain the clients associated to a given vserver/cluster

=over 4

=item Inputs

=item * I<-cluster_name>

Required: The name of the cluster to look up

=item * I<-vserver_names>

Optional: An array reference of vserver names to look up. If not specified or empty,
all vservers will be looked up.
Default: []

=item * I<-protocols>

Optional: An array reference of protocol types to search for.
Default: ["fcp","iscsi"]

=item Outputs

Returns back an array of client names that matched the filters provided,
sets error if an error occurred while looking up the information.

=item Examples

 # obtain all fcp clients associated to cluster "mycluster"
 @clients = $self->get_clients_associated_to_vserver(-cluster_name => "mycluster",
                                                     -protocols => ["fcp"]);
 if ($self->error()) {
    ...
 }

 # obtain all fcp & iscsi clients associated to vservers
 # "vs0" and "vs0" of the cluster "mycluster"
 @clients = $self->get_clients_associated_to_vserver(-cluster_name => "mycluster",
                                                     -vserver_names => ["vs0", "vs1"]);
 if ($self->error()) {
    ...
 }

=back

=head2 B<get_vservers_associated_to_client(%args)>

Obtain the vservers associated to the specified client

=over 4

=item Inputs

=item * I<-client_name>

Required: The name of the client to look up

=item Outputs

Returns back an array of vserver names that are associated to -client_name.
Sets error if an error occurred while looking up the information.

=item Examples

 # obtain all vservers associated to "ibm3550-rtp60"
 @vservers = $self->get_vservers_associated_to_client(-client_name => "ibm3550-rtp60");
 if ($self->error()) {
    ...
 }

=back

=head2 B<get_client_dev_list(%args)>

Get client device list from internal object.
Also allows caller to filter on attributes of client device list.

=over 4

=item Inputs

=item * I<-client>

 Optional: A regular expression to filter by clients.

=item * I<-vserver>

 Optional:  A regular expression to filter by vservers.

=item * I<-volume>

 Optional: A regular expression to filter by volumes.

=item * I<-lun>

 Optional: A regular expression to filter by lun.

=item * I<-protocol>

 Optional: A regular expression to filter by protocol.

=item * I<-serial>

 Optional: A regular expression to filter by LUN serial number.

=item * I<-cdl_format>

 Optional: If set to 1, will return the client dev list in traditional
 client dev list format instead of the hash format.
 Format: client1,/mnt/point1,/mnt/point2;client2,E,F,G,H;client3,etc
 Default: 0

=item Outputs

 Returns back a hash of clients and their associated luns.
 If any filters are specified, this hash will reflect it.
 The format of this hash is:
 {
   client1 => [{ path => "/mnt/mount1", io_type => "fs", vserver => "vs1", lun => "mylun1", volume => "myvol" },
               { path => "/mnt/mount2", io_type => "fs", vserver => "vs1", lun => "mylun2", volume => "myvol" } ],
   client2 => [{ path => "E", io_type => "fs", vserver => "vs2", lun => "mylun1", volume => "myvol" },
               { path => "F", io_type => "fs", vserver => "vs2", lun => "mylun2", volume => "myvol" } ]
 }

=item Examples

  # Obtain all client device lists
  $cdl = $SanMgr_Obj->get_client_dev_list();
  if ($SanMgr_Obj->error()) {
    ...
  }

  # Filtering by client
  $cdl = $SanMgr_Obj->get_client_dev_list(-client => qr/client_01/);
  if ($SanMgr_Obj->error()) {
    ...
  }

  # Filtering by client, vserver, volume, lun and protocol
  $cdl = $SanMgr_Obj->get_client_dev_list(-client => qr/auto-ibm3550-08/,
                                          -vserver => qr/sanmgr_vserv/,
                                          -volume => qr/vol_1/,
                                          -lun => qr/i_3_1/,
                                          -protocol=> qr/iscsi/);
  if ($SanMgr_Obj->error()) {
    ...
  }

=back

=cut

1;
