package San::SanScsiT::DOT_Features::VerifyLun;

use Error;
use Params::Validate qw( :all );
use San::SanClient::ClientUtil;
use San::SanScsiT::XScsiT;
use Data::Compare;
use Data::Dumper;
use Tie::IxHash;

require Exporter;
@ISA = ('Exporter');

@EXPORT = qw(verify_lun
             get_test_paths
            );

use constant {INVALID_LUN => 31,
              BLOCK_SIZE  => 512,};

###############################################################################
# Method:    verify_lun
# Objective: Verify scsi commands for various lun operations
# Details:
###############################################################################

sub verify_lun {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($operation, $pre_lun_attr_hash, $new_size, $lun_obj, $serial_num,
        $dev_id, $new_name, $old_name, $dev_id_type, $lun_path);

    foreach my $param (keys %args) {
        if ($param =~ /^-?operation/i) {
            length($operation = $args{$param})
              or do {
                $self->_set_err(
                              "Invalid value for \"operation\": $args{$param}");
                return 0;
              };
        } elsif ($param =~ /pre_lun_attribute/i) {

            # O/p of the lib routine get_attributes

            length($pre_lun_attr_hash = $args{$param})
              or do {
                $self->_set_err(
                      "Invalid value for \"pre_lun_attr_hash\": $args{$param}");
                return 0;
              };
        } elsif ($param =~ /^-?lun_obj/i) {
            length($lun_obj = $args{$param})
              or do {
                $self->_set_err("Invalid value for \"lun_obj\": $args{$param}");
                return 0;
              };
        } elsif ($param =~ /^-?serial_num/i) {

            # input argument specific to the operation
            length($serial_num = $args{$param})
              or do {
                $self->_set_err(
                             "Invalid value for \"serial_num\": $args{$param}");
                return 0;
              };
        } elsif ($param =~ /^-?device_id_type/i) {

            # input argument specific to the operation - GET DEV ID
            length($dev_id_type = $args{$param})
              or do {
                $self->_set_err(
                              "Invalid value for \"device_id_type\": $args{$param}");
                return 0;
              };
        }elsif ($param =~ /^-?device_id/i) {

            # input argument specific to the operation - GET DEV ID
            length($dev_id = $args{$param})
              or do {
                $self->_set_err(
                              "Invalid value for \"device_id\": $args{$param}");
                return 0;
              };
        } elsif ($param =~ /^-?new_lun_name/i) {

            # input argument specific to the operation - LUN MOVE
            length($new_name = $args{$param})
              or do {
                $self->_set_err(
                               "Invalid value for \"new_lun_name\": $args{$param}");
                return 0;
              };
        } elsif ($param =~ /^-?old_lun_name/i) {

            # input argument specific to the operation - LUN MOVE
            length($old_name = $args{$param})
              or do {
                $self->_set_err(
                               "Invalid value for \"old_lun_name\": $args{$param}");
                return 0;
              };
        } elsif ($param =~ /^-?new_size/i) {

            # input argument specific to the operation - LUN RESIZE
            length($new_size = $args{$param})
              or do {
                $self->_set_err("Invalid value for \"dev_id\": $args{$param}");
                return 0;
              };
        } else {
            $self->_set_err("Error: unhandled argument $param");
            return 0;
        }
    }

    unless (defined $operation) {
        $self->_set_err(
                 "Please specify the lun operation to be verified and lun_obj");
        return $self;
    }

    my $ret;

    my @test_paths = $self->get_test_paths($lun_obj);

    unless($operation eq "move"){
        unless (scalar @test_paths) {
            $self->_set_err("No Paths defined");
            return 0;
        }
    }

    $lun_path = $lun_obj->get_path();

    if ($operation eq "resize") {
        unless (defined $new_size) {
            $self->_set_err("New_size has to be specified for this test");
            return 0;
        }

        $ret = _verify_lun_resize(-xscsit_obj => $self,
                                  -test_paths => \@test_paths,
                                  -new_size   => $new_size);

        if ($ret->{error}) {
            $self->_set_err(  $lun_path. "-" . $ret->{cmd}
                            . " Failed to complete successfully on "
                            . $ret->{path}.". \nAdditional info: ".Dumper($ret->{'err_data'}));
            return 0;
        }
    } elsif ($operation eq "online") {
        $ret = _verify_lun_online(-xscsit_obj => $self,
                                  -test_paths => \@test_paths);

        if ($ret->{error}) {
            $self->_set_err(  $ret->{cmd}
                            . " Failed to complete successfully on "
                            . $ret->{path});
            return 0;
        }
    } elsif ($operation eq "serial") {
        unless (defined $serial_num) {
            $self->_set_err("Serial number has to be specified for this test");
            return 0;
        }

        $ret = _verify_lun_serial(-xscsit_obj => $self,
                                  -test_paths => \@test_paths,
                                  -serial_num => $serial_num);

        if ($ret->{error}) {
            $self->_set_err(  $ret->{cmd}
                            . " Failed to complete successfully on "
                            . $ret->{path});
            return 0;
        }
    } elsif ($operation eq "device_id") {
        unless (defined $dev_id) {
            $self->_set_err("device-id has to be specified for this test");
            return 0;
        }

        $ret = _verify_lun_dev_id(-xscsit_obj => $self,
                                  -test_paths => \@test_paths,
                                  -device_id  => $dev_id,
                                  -device_id_type => $dev_id_type);

        if ($ret->{error}) {
            $self->_set_err(  $ret->{cmd}
                            . " Failed to complete successfully on "
                            . $ret->{path});
            return 0;
        }
    } elsif ($operation eq "offline") {
        $ret = _verify_lun_offline(-xscsit_obj => $self,
                                   -test_paths => \@test_paths);

        if ($ret->{error}) {
            $self->_set_err(  $ret->{cmd}
                            . " Failed to complete successfully on "
                            . $ret->{path});
            return 0;
        }
    } elsif ($operation eq "move") {
        unless (defined $new_name and defined $pre_lun_attr_hash) {
            $self->_set_err(
                "new_name and pre_lun_attr_hash has to be specified for this test"
            );
            return 0;
        }

        $ret = _verify_lun_move(-xscsit_obj        => $self,
                                -test_paths        => \@test_paths,
                                -new_lun_name      => $new_name,
                                -old_lun_name      => $old_name,
                                -pre_lun_attribute => $pre_lun_attr_hash,
                                -lun_obj           => $lun_obj);

        if ($ret->{error}) {
            $self->_set_err(  $ret->{cmd}
                            . " Failed to complete successfully on "
                            . $ret->{path});
            return 0;
        }
    } elsif ($operation eq "create") {
        $ret = _verify_lun_create(-xscsit_obj => $self,
                                  -test_paths => \@test_paths);

        if ($ret->{error}) {
            $self->_set_err(  $ret->{cmd}
                            . " Failed to complete successfully on "
                            . $ret->{path});
            return 0;
        }
    }

    return 1;
}

###############################################################################
# Method:    get_test_paths
# Objective: Filters the paths relevant to a lun obj
# Details:
###############################################################################

sub get_test_paths {
    my ($self, $lun_obj) = @_;

    my $lun_name = $lun_obj->get_path();

    if ($lun_obj->error) {
        $self->_set_err($lun_obj->errmsg());
        return $self;
    }

    my $vserver;

    if ($self->{ontap_mode} =~ /Cmode/i) {
        $vserver = $lun_obj->get_vserver();

        if ($lun_obj->error) {
            $self->_set_err($lun_obj->errmsg());
            return $self;
        }
    } else {
        $vserver = $lun_obj->get_filername();

        if ($lun_obj->error) {
            $self->_set_err($lun_obj->errmsg());
            return $self;
        }
    }

    my $vol_name = $lun_obj->get_volume();

    if ($lun_obj->error) {
        $self->_set_err($lun_obj->errmsg());
        return $self;
    }

    my (@pri_paths, @sec_paths);
    my @test_paths;

    my $path_info =
      $self->{path_hash}->{$vserver}->{$vol_name}->{$lun_name}->{path_info};

    if (defined $path_info->{pri}) {
        @pri_paths = @{$path_info->{pri}};
        push(@test_paths, @pri_paths);
    }

    if (defined $path_info->{sec}) {
        @sec_paths = @{$path_info->{sec}};
        push(@test_paths, @sec_paths);
    }

    return @test_paths;
}

###############################################################################
#                            PRIVATE FUNCTIONS
###############################################################################

###############################################################################
# Method:    _verify_lun_resize
# Objective: Verify scsi commands when the lun is resized
# Details:
###############################################################################

sub _verify_lun_resize {

    # New size entered should be in bytes shown in lun show -v
    my (%args) = @_;

    tie my %err_hash, 'Tie::IxHash';
    tie my %additional_error_data, 'Tie::IxHash';

    my $lun_size;

    $err_hash{error} = 0;
    $err_hash{cmd}   = "";

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {xscsit_obj => {type     => OBJECT,
                            required => 1,
                            isa      => 'San::SanScsiT::XScsiT'},
             test_paths => {type     => ARRAYREF,
                            required => 1},
             new_size   => {type     => SCALAR,
                            required => 1}});
    };

    if ($@) {
        $err_hash{cmd}   = $@;
        $err_hash{path}  = "";
        return \%err_hash;
    }

    my $xscsit_obj = $args{'xscsit_obj'};
    my $paths      = $args{'test_paths'};
    my $new_size   = $args{'new_size'};

    foreach my $client (@{$xscsit_obj->{clients}}) {
        my $scsi_util_obj = $xscsit_obj->{scsit_client_h}->{$client}->{su_ref};
        my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

        foreach my $path (@$paths) {

            # The Inquiry command should complete with a good status

            $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "inquiry",
                                            -buf_len     => BLOCK_SIZE,
                                            -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error()) {
                $err_hash{'error'}              = 1;
                $err_hash{'cmd'}                = "std_inquiry";
                $err_hash{'path'}               = $path;
                $additional_error_data{'expected_status'}    = "0";
                $additional_error_data{'expected_sense_key'} = "0";
                $additional_error_data{'expected_ascnq'}     = "00h/00h";
                $additional_error_data{'actual_status'}      = $return->{SCSI_STATUS};
                $additional_error_data{'actual_sense_key'}   = $return->{SENSE_KEY};
                $additional_error_data{'actual_ascnq'}       = $return->{ASCNQ};
                $err_hash{'err_data'}           = \%additional_error_data;
                return \%err_hash;
            } else {
                $xscsit_obj->sanlog(-msg       => "Success: inquiry",
                                    -msg_level => 3);
            }

            # Any non INQUIRY, REPORT LUNS or REQUEST SENSE command
            # received should get CAPACITY DATA HAS CHANGED unit attention

            $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "read_capacity_16",
                                            -buf_len     => BLOCK_SIZE,
                                            -scsi_status => "CHECK CONDITION",
                                            -sense_key   => "UNIT ATTENTION",
                                            -ascnq       => "2Ah/09h",);

            if ($scsi_tool_obj->error()) {
                $err_hash{'error'}              = 1;
                $err_hash{'cmd'}                = "read_capacity_16";
                $err_hash{'path'}               = $path;
                $additional_error_data{'expected_status'}    = "2";
                $additional_error_data{'expected_sense_key'} = "6";
                $additional_error_data{'expected_ascnq'}     = "2Ah/09h";
                $additional_error_data{'actual_status'}      = $return->{SCSI_STATUS};
                $additional_error_data{'actual_sense_key'}   = $return->{SENSE_KEY};
                $additional_error_data{'actual_ascnq'}       = $return->{ASCNQ};
                $err_hash{'err_data'}           = \%additional_error_data;
                return \%err_hash;
            } else {
                $xscsit_obj->sanlog(-msg       => "Success: read_capacity_16",
                                    -msg_level => 3);
            }

            # READ CAPACITY (16) should reflect the new size

            $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "read_capacity_16",
                                            -buf_len     => BLOCK_SIZE,
                                            -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error()) {
                $err_hash{'error'}              = 1;
                $err_hash{'cmd'}                = "read_capacity_16";
                $err_hash{'path'}               = $path;
                $additional_error_data{'expected_status'}    = "0";
                $additional_error_data{'expected_sense_key'} = "0";
                $additional_error_data{'expected_ascnq'}     = "00h/00h";
                $additional_error_data{'actual_status'}      = $return->{SCSI_STATUS};
                $additional_error_data{'actual_sense_key'}   = $return->{SENSE_KEY};
                $additional_error_data{'actual_ascnq'}       = $return->{ASCNQ};
                $err_hash{'err_data'}           = \%additional_error_data;
                return \%err_hash;
            } else {
                $xscsit_obj->sanlog(-msg       => "Success: read_capacity_16",
                                    -msg_level => 3);
            }

            if ($return->{SCSI_DATA}) {
                $return =
                  $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "read_capacity_16",
                                         -scsi_data_ref => $return->{SCSI_DATA},
                  );

                if ($scsi_util_obj->error()) {
                    $err_hash{'error'}     = 1;
                    $err_hash{'cmd'}       = "read_capacity_16 translation";
                    $err_hash{'path'}      = $path;
                    $additional_error_data{'message'} = $scsi_util_obj->errmsg();
                    $err_hash{'err_data'}  = \%additional_error_data;
                    return \%err_hash;
                } else {
                    $xscsit_obj->sanlog(-msg       => "Success: read_capacity_16",
                                        -msg_level => 3);
                }

                # read capacity 16 returns MAX LBA where the LBA is computed starting from 0
                # Hence adding 1 to max LBA returned to get the size of the LU

                $lun_size =
                  ($return->{RETURNED_LOGICAL_BLOCK_ADDRESS} + 1) * BLOCK_SIZE;
            }

            unless ($new_size == $lun_size) {
               $err_hash{'error'}         = 1;
               $err_hash{'cmd'}           = "read_capacity_16";
               $err_hash{'path'}          = $path;
               $additional_error_data{'message'}       = "Size Mismatch";
               $additional_error_data{'actual_size'} = $lun_size;
               $additional_error_data{'expected_size'}      = $new_size;
               $err_hash{'err_data'}      = \%additional_error_data;
               return \%err_hash;
            }

            # GET VDISK ATTR should reflect the new size
            # Compute new size as the sum of all the streams (prefix,main and suffix)

            $return =
              $scsi_tool_obj->test_scsi_cmd(
                                     -device      => $path,
                                     -cdb_name    => "gva_vdisk_stream_sizes_page",
                                     -buf_len     => BLOCK_SIZE,
                                     -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error()) {
                $err_hash{'error'}              = 1;
                $err_hash{'cmd'}                = "gva_vdisk_stream_sizes_page";
                $err_hash{'path'}               = $path;
                $additional_error_data{'expected_status'}    = "0";
                $additional_error_data{'expected_sense_key'} = "0";
                $additional_error_data{'expected_ascnq'}     = "00h/00h";
                $additional_error_data{'actual_status'}      = $return->{SCSI_STATUS};
                $additional_error_data{'actual_sense_key'}   = $return->{SENSE_KEY};
                $additional_error_data{'actual_ascnq'}       = $return->{ASCNQ};
                $err_hash{'err_data'}           = \%additional_error_data;
                return \%err_hash;
            } else {
                $xscsit_obj->sanlog(
                                 -msg => "Success: gva_vdisk_stream_sizes_page",
                                 -msg_level => 3);
            }

            my $h_ref =
              $scsi_util_obj->scsi_bytes_to_strings(
                                -scsi_cmd_name => "gva_vdisk_stream_sizes_page",
                                -scsi_data_ref => $return->{SCSI_DATA},);

            if ($scsi_util_obj->error()) {
               $err_hash{'error'}         = 1;
               $err_hash{'cmd'}           = "gva_vdisk_stream_sizes_page";
               $err_hash{'path'}          = $path;
               $additional_error_data{'message'}  = $scsi_util_obj->errmsg();
               $err_hash{'err_data'}    = \%additional_error_data;
               return \%err_hash;
            } else {
                $xscsit_obj->sanlog(-msg       => "Success: gva_vdisk_stream_sizes_page",
                                    -msg_level => 3);
            }

            my $stream_size_sum =
              $h_ref->{MAIN_STREAM} +
              $h_ref->{PREFIX_STREAM} +
              $h_ref->{SUFFIX_STREAM};

            unless ($new_size == $stream_size_sum) {
               $err_hash{'error'}    = 1;
               $err_hash{'cmd'}      = "gva_vdisk_stream_sizes_page";
               $err_hash{'path'}     = $path;
               $additional_error_data{'message'}       = "Size Mismatch";
               $additional_error_data{'actual_size'}   = $stream_size_sum;
               $additional_error_data{'expected_size'} = $new_size;
               $err_hash{'err_data'} = \%additional_error_data;
               return \%err_hash;
            }

        }
    }
    return \%err_hash;
}

###############################################################################
# Method:    _verify_lun_online
# Objective: Verify scsi commands when the lun is online
# Details:
###############################################################################

sub _verify_lun_online {
    my (%args) = @_;
    my %err_hash;
    $err_hash{error} = 0;
    $err_hash{cmd}   = "";

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # validate input parameters
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {xscsit_obj => {
                 type     => OBJECT,
                 required => 1,

                 #                                isa      => 'San::SanScsiT::XScsiT'
                           },
             test_paths => {type     => ARRAYREF,
                            required => 1}});
    };

    if ($@) {
        $err_hash{error} = 1;
        $err_hash{cmd}   = $@;
        $err_hash{path}  = "";
        return \%err_hash;
    }

    my $xscsit_obj = $args{'xscsit_obj'};
    my $paths      = $args{'test_paths'};

    foreach my $client (@{$xscsit_obj->{clients}}) {
        my $scsi_util_obj = $xscsit_obj->{scsit_client_h}->{$client}->{su_ref};
        my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

        foreach my $path (@$paths) {

            # Execute READ (10) . The first read should get
            # UNIT ATTENTION

            for (my $ind = 0 ; $ind < 2 ; $ind++) {
                my $return =
                  $scsi_tool_obj->test_scsi_cmd(
                                   -device   => $path,
                                   -cdb_name => "28 00 00 00 00 00 00 00 01 00",
                                   -buf_len  => BLOCK_SIZE,
                                   -scsi_status => "CHECK CONDITION",
                                   -sense_key   => "UNIT ATTENTION",
                                   -ascnq       => "3fh/0eh",);
            }

            # The READ (10) command should complete with GOOD status

            $return =
              $scsi_tool_obj->test_scsi_cmd(
                                   -device   => $path,
                                   -cdb_name => "28 00 00 00 00 00 00 00 01 00",
                                   -buf_len  => BLOCK_SIZE,
                                   -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            # The WRITE command should should complete with GOOD status

            $result =
              $scsi_tool_obj->read_or_write_lba(-dev        => $path,
                                                -lba        => 0,
                                                -read_write => "write",
                                                -alloc_len  => 1,
                                                -pat        => "1",);

            if ($scsi_tool_obj->error) {
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            # The INQUIRY command should complete with a good status

            $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "inquiry",
                                            -buf_len     => 512,
                                            -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "inquiry";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            if ($return->{SCSI_DATA}) {
                $return =
                  $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "inquiry",
                                         -scsi_data_ref => $return->{SCSI_DATA},
                  );

                if ($scsi_util_obj->error) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "inquiry";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                } else {
                    $xscsit_obj->sanlog(-msg       => "Success:",
                                        -msg_level => 3);
                }

                # Check the periferal device type and the peripheral qualifier
                # peripheral qualifier: 000b - Connected, periferal
                # device type: 00h - Direct-access Block Device

                unless (    ($return->{PERIPHERAL_DEVICE_TYPE} == 0)
                        and ($return->{PERIPHERAL_QUALIFIER} == 0)) {
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "Inquiry";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                }
            }

            # The GET VDISK ATTR command should complete with a good status

            $return =
              $scsi_tool_obj->test_scsi_cmd(
                                           -device   => $path,
                                           -cdb_name => "gva_vdisk_status_page",
                                           -buf_len  => 512,
                                           -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "GVA";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            if ($return->{SCSI_DATA}) {
                $return =
                  $scsi_util_obj->scsi_bytes_to_strings(
                                      -scsi_cmd_name => "gva_vdisk_status_page",
                                      -scsi_data_ref => $return->{SCSI_DATA},);

                if ($scsi_util_obj->error) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{cmd}  = "GVA";
                    $err_hash{path} = $path;
                    return \%err_hash;
                } else {
                    $xscsit_obj->sanlog(-msg       => "Success:",
                                        -msg_level => 3);
                }
            }

            unless ($return->{OFFLINE} == 0) {
                $err_hash{error} = 1;
                $err_hash{cmd}   = "GVA";
                $err_hash{path}  = $path;
                return \%err_hash;
            }
        }
    }

    return \%err_hash;
}

###############################################################################
# Method:    _verify_lun_serial
# Objective: Verify the lun serial number
# Details:
###############################################################################

sub _verify_lun_serial {
    my (%args) = @_;
    my %err_hash;

    $err_hash{error} = 0;
    $err_hash{cmd}   = "";

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {xscsit_obj => {
                 type     => OBJECT,
                 required => 1,

                 #                                isa      => 'San::SanScsiT::XScsiT'
                           },
             test_paths => {type     => ARRAYREF,
                            required => 1},
             serial_num => {type     => SCALAR,
                            required => 1}});
    };

    if ($@) {
        $err_hash{error} = 1;
        $err_hash{cmd}   = $@;
        $err_hash{path}  = "";
        return \%err_hash;
    }

    my $xscsit_obj = $args{'xscsit_obj'};
    my $paths      = $args{'test_paths'};
    my $serial_num = $args{'serial_num'};

    foreach my $client (@{$xscsit_obj->{clients}}) {
        my $scsi_util_obj = $xscsit_obj->{scsit_client_h}->{$client}->{su_ref};
        my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

        foreach my $path (@$paths) {
            my $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "inquiry_page_80h",
                                            -buf_len     => 512,
                                            -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "inquiry_page_80h";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            if ($return->{SCSI_DATA}) {
                $return =
                  $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "inquiry_page_80h",
                                         -scsi_data_ref => $return->{SCSI_DATA},
                  );

                if ($scsi_util_obj->error) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "inquiry_page_80h";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                } else {
                    $xscsit_obj->sanlog(-msg => "Success: inquiry_page_80h",
                                        -msg_level => 3);
                }

                # remove null characters
                $return->{PRODUCT_SERIAL_NUMBER} =~ s/[\0]*$//g;

                unless ($serial_num eq $return->{PRODUCT_SERIAL_NUMBER}) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "inquiry_page_80h";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                }
            }

            my $return1 =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "inquiry",
                                            -buf_len     => 512,
                                            -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "inquiry";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            if ($return1->{SCSI_DATA}) {
                $return1 =
                  $scsi_util_obj->scsi_bytes_to_strings(
                                        -scsi_cmd_name => "inquiry",
                                        -scsi_data_ref => $return1->{SCSI_DATA},
                  );

                if ($scsi_util_obj->error) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "inquiry";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                } else {
                    $xscsit_obj->sanlog(-msg       => "Success: inquiry",
                                        -msg_level => 3);
                }

                # remove null characters
                $return1->{VENDOR_SPECIFIC} =~ s/[\0]*$//g;

                unless ($serial_num eq $return1->{VENDOR_SPECIFIC}) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "inquiry";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                }
            }

            $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "inquiry_page_83h",
                                            -buf_len     => 512,
                                            -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "inquiry_page_83h";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            if ($return->{SCSI_DATA}) {
                $return =
                  $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "inquiry_page_83h",
                                         -scsi_data_ref => $return->{SCSI_DATA},
                  );

                if ($scsi_util_obj->error) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "inquiry_page_83h";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                } else {
                    $xscsit_obj->sanlog(-msg => "Success: inquiry_page_83h",
                                        -msg_level => 3);
                }

                unless (
                    ($serial_num eq $return->{ID_LIST}->[0]->{LUN_SERIAL_NUMBER}
                    )
                    and ($return->{ID_LIST}->[1]->{LUN_SERIAL_NUMBER_0_11})
                  ) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "inquiry";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                }
            }
        }
    }

    return \%err_hash;
}

###############################################################################
# Method:    _verify_lun_dev_id
# Objective: Verify the lun device id
# Details:
###############################################################################

sub _verify_lun_dev_id {
    my (%args) = @_;
    my %err_hash;

    $err_hash{error} = 0;
    $err_hash{cmd}   = "";

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {xscsit_obj => {
                 type     => OBJECT,
                 required => 1,

                 #                                isa      => 'San::SanScsiT::XScsiT'
                           },
             test_paths => {type     => ARRAYREF,
                            required => 1},
             device_id => {type     => SCALAR,
                           required => 1},
             device_id_type => {type     => SCALAR,
                                required => 1}});
    };

    if ($@) {
        $err_hash{error} = 1;
        $err_hash{cmd}   = $@;
        $err_hash{path}  = "";
        return \%err_hash;
    }

    my $xscsit_obj     = $args{'xscsit_obj'};
    my $paths          = $args{'test_paths'};
    my $device_id      = $args{'device_id'};
    my $device_id_type = $args{'device_id_type'};
    my $return;

    foreach my $client (@{$xscsit_obj->{clients}}) {
        my $scsi_util_obj = $xscsit_obj->{scsit_client_h}->{$client}->{su_ref};
        my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

        foreach my $path (@$paths) {

            $scsi_util_obj->clear_cc_ua(-paths => $paths);

            if (lc($device_id_type) eq "legacy") {
                $return =
                  $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                -cdb_name    => "get_dev_id",
                                                -buf_len     => 512,
                                                -scsi_status => "GOOD",);

                if ($scsi_tool_obj->error) {
                    $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "get_dev_id";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                }

                if ($return->{SCSI_DATA}) {
                    $return =
                      $scsi_util_obj->scsi_bytes_to_strings(
                                             -scsi_cmd_name => "get_dev_id",
                                             -scsi_data_ref => $return->{SCSI_DATA},
                      );

                    if ($scsi_util_obj->error) {
                        $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                        $err_hash{error} = 1;
                        $err_hash{cmd}   = "get_dev_id_scsi_bytes_to_strings";
                        $err_hash{path}  = $path;
                        return \%err_hash;
                    } else {
                        $xscsit_obj->sanlog(-msg       => "Success: get_dev_id",
                                            -msg_level => 3);
                    }

                    if (    ($return->{ID_VALID_BIT} == 0)
                        and ($return->{ID_LENGTH} == 0)) {
                        $dev_id = "none";
                    } else {
                        $dev_id = $return->{DEV_ID};
                    }

                    unless ($device_id eq $dev_id) {
                        $err_hash{error} = 1;
                        $err_hash{cmd}   = "get_dev_id";
                        $err_hash{path}  = $path;
                        return \%err_hash;
                    }
                }
            }

            $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "report_dev_id",
                                            -buf_len     => 512,
                                            -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "report_dev_id";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            if ($return->{SCSI_DATA}) {
                my $return1 =
                  $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "report_dev_id",
                                         -scsi_data_ref => $return->{SCSI_DATA},
                  );

                if ($scsi_util_obj->error) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "report_dev_id_scsi_bytes_to_strings";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                } else {
                    $xscsit_obj->sanlog(-msg       => "Success: report_dev_id",
                                        -msg_level => 3);
                }

                if (lc($device_id_type) eq "legacy") {
                   my $id = sprintf("%4x",$return1->{DEV_ID});
                    
                   unless ($device_id eq $id) {
                       $err_hash{error} = 1;
                       $err_hash{cmd}   = "Device Id mismatch";
                       $err_hash{path}  = $path;
                       return \%err_hash;
                   }

                } else {
                   my $str = unpack("B32", pack("N", $return1->{DEV_ID}));
                   $str =~ s/^0+(?=\d)//;

                   unless ($device_id eq $str) {
                       $err_hash{error} = 1;
                       $err_hash{cmd}   = "Device Id mismatch";
                       $err_hash{path}  = $path;
                       return \%err_hash;
                   }
                }
            }
        }
    }

    return \%err_hash;
}

###############################################################################
# Method:    _verify_lun_move
# Objective: Verify scsi commands when the lun is offline
# Details:
###############################################################################

sub _verify_lun_move {
    my (%args) = @_;
    my %err_hash;

    $err_hash{error} = 0;
    $err_hash{cmd}   = "";

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {xscsit_obj => {
                 type     => OBJECT,
                 required => 1,
                #isa      => 'San::SanScsiT::XScsiT'
                           },
             test_paths => {type     => ARRAYREF,
                            required => 1},
             new_lun_name => {type     => SCALAR,
                              required => 1},
             old_lun_name => {type     => SCALAR,
                              required => 1},
             pre_lun_attribute => {type     => HASHREF,
                                   required => 1},
             lun_obj => {type     => OBJECT,
                         required => 1}});
    };

    if ($@) {
        $err_hash{error} = 1;
        $err_hash{cmd}   = $@;
        $err_hash{path}  = "";
        return \%err_hash;
    }

    my $xscsit_obj        = $args{'xscsit_obj'};
    my $paths             = $args{'test_paths'};
    my $new_name          = $args{'new_lun_name'};
    my $old_name          = $args{'old_lun_name'};
    my $pre_lun_attr_hash = $args{'pre_lun_attribute'};
    my $lun_obj           = $args{'lun_obj'};

    my $attr;

    foreach my $client (@{$xscsit_obj->{clients}}) {
        my $scsi_util_obj = $xscsit_obj->{scsit_client_h}->{$client}->{su_ref};
        my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

        my $filer;

        if ($xscsit_obj->{ontap_mode} =~ /Cmode/i) {
            $filer = $lun_obj->get_vserver();

        } else {
            $filer = $lun_obj->get_filername();
        }

        if ($lun_obj->error) {
            $xscsit_obj->_set_err($lun_obj->errmsg());
            $err_hash{error} = 1;
            $err_hash{cmd}   = "";
            $err_hash{path}  = "";
            return \%err_hash;
        }

        # by default the type is set to "all" in get_scsi_attributes
        # get all the scsi attributes after the lun move operation
        $xscsit_obj->sanlog(-msg => "Get Scsi Attributes after Lun Move");

        $attr = $xscsit_obj->get_scsi_attributes();

        if ($xscsit_obj->error) {
            $err_hash{error} = 1;
            $err_hash{cmd}   = "";
            $err_hash{path}  = $path;
            return \%err_hash;
        }

        # Verify whether the lun name obtained from get_scsi_attributes
        # is the same as the new name passed as the input

        my $lun_name =
          $attr->{config_data}->{$filer}->{$client}->{$old_name}->{vdisk_info}
          ->{vdisk_path};

        unless ($lun_name eq $new_name) {
            $err_hash{error} = 1;
            $err_hash{cmd}   = "config_data";
            $err_hash{path}  = $path;
            return \%err_hash;
        }

        # Since the lun path changes after the lun move and all the
        # other scsi attributes are supposed to remain the same
        # delete the "vdisk_path" key from the hashes so that it will
        # be convenient to compare the hashes

        delete $pre_lun_attr_hash->{config_data}->{$filer}->{$client}
          ->{$old_name}->{vdisk_info}->{vdisk_path};
        delete $attr->{config_data}->{$filer}->{$client}->{$old_name}
          ->{vdisk_info}->{vdisk_path};

        $xscsit_obj->sanlog(-msg => "Comparing MetaData");
        use Data::Compare;
        my $c = new Data::Compare($pre_lun_attr_hash->{meta_data},
                                  $attr->{meta_data});

        unless ($c->Cmp) {
            $err_hash{error} = 1;
            $err_hash{cmd}   = "meta_data";
            $err_hash{path}  = $path;
            return \%err_hash;
        }

        $xscsit_obj->sanlog(-msg => "Comparing Config Data");
        $c = new Data::Compare($pre_lun_attr_hash->{config_data},
                               $attr->{config_data});

        unless ($c->Cmp) {
            $err_hash{error} = 1;
            $err_hash{cmd}   = "config_data";
            $err_hash{path}  = $path;
            return \%err_hash;
        }

        $xscsit_obj->sanlog(-msg => "Comparing State Data");
        $c = new Data::Compare($pre_lun_attr_hash->{state_data},
                               $attr->{state_data});

        unless ($c->Cmp) {
            $err_hash{error} = 1;
            $err_hash{cmd}   = "state_data";
            $err_hash{path}  = $path;
            return \%err_hash;
        }
    }

    return \%err_hash;
}

###############################################################################
# Method:    _verify_lun_offline
# Objective: Verify scsi commands when the lun is offline
# Details:
###############################################################################

sub _verify_lun_offline {
    my (%args) = @_;
    my %err_hash;

    $err_hash{error} = 0;
    $err_hash{cmd}   = "";

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {xscsit_obj => {
                 type     => OBJECT,
                 required => 1,
               # isa      => 'San::SanScsiT::XScsiT'
                           },
             test_paths => {type     => ARRAYREF,
                            required => 1}});
    };

    if ($@) {
        $err_hash{error} = 1;
        $err_hash{cmd}   = $@;
        $err_hash{path}  = "";
        return \%err_hash;
    }

    my $xscsit_obj = $args{'xscsit_obj'};
    my $paths      = $args{'test_paths'};

    my $attr;

    foreach my $client (@{$xscsit_obj->{clients}}) {
        my $scsi_util_obj = $xscsit_obj->{scsit_client_h}->{$client}->{su_ref};
        my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};
        my $return;

        foreach $path (@$paths) {
            # execute READ (10) . The first read should get
            # UNIT ATTENTION/ REPORTED LUNS DATA HAS CHANGED for each IT nexus

            $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "read_10",
                                            -buf_len     => 512,
                                            -scsi_status => "CHECK CONDITION",
                                            -sense_key   => "UNIT ATTENTION",
                                            -ascnq       => "3fh/0eh",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->sanlog(-msg => $scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            # The read command should fail with
            # CHECK CONDITION (2)/ ILLEGAL REQUEST (5)/ LOGICAL UNIT NOT SUPPORTED (25h/00h)

            $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "read_10",
                                            -buf_len     => 512,
                                            -scsi_status => "CHECK CONDITION",
                                            -sense_key   => "ILLEGAL REQUEST",
                                            -ascnq       => "25h/00h",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            # The Write command should fail with
            # CHECK CONDITION (2)/ ILLEGAL REQUEST (5)/ LOGICAL UNIT NOT SUPPORTED (25h/00h)

            $result =
              $scsi_tool_obj->read_or_write_lba(-dev        => $path,
                                                -lba        => 0,
                                                -read_write => "write",
                                                -alloc_len  => 1,
                                                -pat        => "1",);

            if ($scsi_tool_obj->error) {
                unless (    ($result->{SCSI_STATUS} == 2)
                        and ($result->{SENSE_KEY} == 5)
                        and ($result->{ASCNQ} =~ m#25h/00h#)) {
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "write";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                }
            }

            # The INQUIRY command should complete with a good status
            # when sent to an invalid lun

            $return =
              $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                            -cdb_name    => "inquiry",
                                            -buf_len     => 512,
                                            -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "inquiry";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            if ($return->{SCSI_DATA}) {
                $return =
                  $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "inquiry",
                                         -scsi_data_ref => $return->{SCSI_DATA},
                  );

                if ($scsi_util_obj->error) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "inquiry";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                } else {
                    $xscsit_obj->sanlog(-msg       => "Success: inquiry",
                                        -msg_level => 3);
                }

                # Check the periferal device type. The value should be set to 1Fh
                # (31 in decimal) to indicate an invalid LU

                unless ($return->{PERIPHERAL_DEVICE_TYPE} == INVALID_LUN) {
                    $err_hash{error} = 1;
                    $err_hash{cmd}   = "Inquiry";
                    $err_hash{path}  = $path;
                    return \%err_hash;
                }
            }

            # The GET VDISK ATTR command should complete with a good status
            # when sent to an invalid lun

            $return =
              $scsi_tool_obj->test_scsi_cmd(
                                           -device   => $path,
                                           -cdb_name => "gva_vdisk_status_page",
                                           -buf_len  => 512,
                                           -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $xscsit_obj->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "GVA";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            if ($return->{SCSI_DATA}) {
                $return =
                  $scsi_util_obj->scsi_bytes_to_strings(
                                      -scsi_cmd_name => "gva_vdisk_status_page",
                                      -scsi_data_ref => $return->{SCSI_DATA},);

                if ($scsi_util_obj->error) {
                    $xscsit_obj->_set_err($scsi_util_obj->errmsg());
                    $err_hash{cmd}  = "GVA";
                    $err_hash{path} = $path;
                    return \%err_hash;
                } else {
                    $xscsit_obj->sanlog(
                                       -msg => "Success: gva_vdisk_status_page",
                                       -msg_level => 3);
                }
            }
        }
    }
    return \%err_hash;
}

1;
