package San::SanScsiT::DOT_Features::VerifyPortset;


use Error;
use Data::Dumper;
use San::SanClient::ClientUtil;

####################################################################################
## Method:     verify_portset                                                      #
## Objective:  Parse the args hash, and trickle down to verify a specific funtion  #
## Inputs:     1. (hash) args hash from verify()                                   #
## Output      3. Nothing, or an error with message                                #
####################################################################################
sub verify_portset{
    my ($self, %args) = @_;
    my($feature, $op, $pre_lun_attr_hash); 

    foreach my $param (keys %args) {
        if ($param =~ /^-?feature/i) {
            length($feature = $args{$param})
              or do {
                $self->_set_err("Invalid value for \"feature\": $args{$param}");
                return $self;
              };
              delete $args{$param};
        } elsif ($param =~ /^-?op/i) {
            length($op = $args{$param})
              or do {
                $self->_set_err("Invalid value for \"op\": $args{$param}");
                return $self;
              };
              delete $args{$param};
        } elsif ($param =~ /^-?pre_lun_attr_hash/i) {
            length($pre_lun_attr_hash = $args{$param})
              or do {
                $self->_set_err("Invalid value for \"pre_lun_attr_hash\": $args{$param}");
                return $self;
              };
        } 
    }

    if (($feature eq "portset") and ($op eq "bind")) {
        $self->verify_portset_bind(%args);

    }elsif (($feature eq "portset") and ($op eq "unbind")) {
        $self->verify_portset_unbind(%args);

    }elsif (($feature eq "portset") and ($op eq "add")) {
        $self->verify_portset_add(%args);

    }elsif (($feature eq "portset") and ($op eq "remove")) {
        $self->verify_portset_remove(%args);

    }
}

####################################################################################
## Method:     verify_portset_bind                                                 #
## Objective:  Verify that viewable paths to luns are through ports in the portset #
## Inputs:     1. (array) bound ports                                              #
##             2. (object) pre lun attribute hash (optional)                       #
## Output      3. Nothing, or an error with message                                #
####################################################################################

sub verify_portset_bind{
    my ($self, %args) = @_;
    my (@ports_bound, $pre_lun_attr);

    foreach my $key (%args){
        if ($key =~ /ports_bound/i){
            @ports_bound = split(/,/, $args{$key});
        }
        elsif($key =~ /pre_lun_attribute/i){
            $pre_lun_attr = $args{$key};
        }
        else{
            $self->_set_err("Error: unhandled argument $key\n");
            return $self;
        }
    }

    #
    #Grab the client objects
    #
    my @client_objs;
    if(ref($san_obj) =~ /^San$/i){
        @client_objs = @{$self->{san_obj}->{clients}};
    }elsif(ref($san_obj) =~ /FitMgr|SanMgr/i){
        @client_objs = $self->{san_obj}->get_objects(-type    => "client");
    }else{
        $self->set_err("Error: San object is of unknown type ".ref($san_obj)."\n");
        return $self;
    }

    #
    #Get paths to luns with each client.  Make sure each path
    #is associated with a port in the bound ports list
    #
    foreach my $client (@client_objs){
        #
        #grab the scsi util object
        #
        $scsi_util_obj = $self->{scsit_client_h}->{$client}->{su_ref};

        #
        #test each lun and vserver
        #
        my @lun_objects = $san_obj->get_objects(-type    => "lun");

        foreach my $lun_obj (@lun_objects){
            my $lun_name =  $lun_obj->get_path();

            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $vserver = $lun_obj->get_vserver();
                
            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $vol_name = $lun_obj->get_volume();

            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }
            
            if  ($self->{path_hash}->{$vserver}->{$vol_name}->{$lun_name}->{path_hash}){
                $self->_test_bind($path_hash, $scsi_util_obj->{s_t_obj}, @ports_bound);
            }

        }
    }
}




####################################################################################
## Method:     verify_portset_unbind                                               #
## Objective:  Verify that we can now read on all the paths                        #
## Inputs:     1. (array) Ports that used to be in the portset                     #
##             2. (object) pre lun attribute hash (optional)                       #
## Output      3. Nothing, or an error with message                                #
####################################################################################
sub verify_portset_unbind{
    my ($self, %args) = @_;
    my (@ports_bound, $pre_lun_attr);

    foreach my $key (%args){
        if ($key =~ /ports_bound/i){
            @ports_bound = split(/,/, $args{$key});
        }
        elsif($key =~ /pre_lun_attribute/i){
            $pre_lun_attr = $args{$key};
        }
        else{
            $self->_set_err("Error: unhandled argument $key\n");
            return $self;
        }
    }

    #
    #Grab the client objects
    #
    my @client_objs;
    if(ref($san_obj) =~ /^San$/i){
        @client_objs = @{$self->{san_obj}->{clients}};
    }elsif(ref($san_obj) =~ /FitMgr|SanMgr/i){
        @client_objs = $self->{san_obj}->get_objects(-type    => "client");
    }else{
        $self->set_err("Error: San object is of unknown type ".ref($san_obj)."\n");
        return $self;
    }

    #
    #Get paths to luns with each client.  Make sure each path
    #is associated with a port in the bound ports list
    #
    foreach my $client (@client_objs){
        #
        #grab the scsi util object
        #
        $scsi_util_obj = $self->{scsit_client_h}->{$client}->{su_ref};

        #
        #test each lun and vserver
        #
        my @lun_objects = $san_obj->get_objects(-type    => "lun");

        foreach my $lun_obj (@lun_objects){
            my $lun_name =  $lun_obj->get_path();

            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $vserver = $lun_obj->get_vserver();
                
            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $vol_name = $lun_obj->get_volume();

                if ($lun_obj->error) {
                    $self->_set_err($lun_obj->errmsg());
                    return $self;
                }
            
            if  ($self->{path_hash}->{$vserver}->{$vol_name}->{$lun_name}->{path_hash}){
                $self->_test_unbind($path_hash, $scsi_util_obj->{s_t_obj}, @ports_bound);
            }

        }
    }
}


####################################################################################
## Method:     verify_portset_add                                                  #
## Objective:  Verify that when we add a port, the paths behave correctly          #
## Inputs:     1. (array) Port or ports added to the portest                       #
##             2. (array) Port or ports already in the portset                     #
##             3. (object) pre lun attribute hash (optional)                       #
## Output      4. Nothing, or an error with message                                #
####################################################################################
sub verify_portset_add{
    my ($self, %args) = @_;
    my (@ports_added, @portset, $pre_lun_attr);

    foreach my $key (%args){
        if ($key =~ /ports_added/i){
            @ports_added = split(/,/, $args{$key});
        }
        elsif ($key =~ /portset/){
            @portset = split(/,/, $args{$key});
        }
        elsif ($key =~ /pre_lun_attribute/i){
            $pre_lun_attr = $args{$key};
        }
        else{
            $self->_set_err("Error: unhandled argument $key\n");
            return $self;
        }
    }

    #
    #Grab the client objects
    #
    my @client_objs;
    if(ref($san_obj) =~ /^San$/i){
        @client_objs = @{$self->{san_obj}->{clients}};
    }elsif(ref($san_obj) =~ /FitMgr|SanMgr/i){
        @client_objs = $self->{san_obj}->get_objects(-type    => "client");
    }else{
        $self->set_err("Error: San object is of unknown type ".ref($san_obj)."\n");
        return $self;
    }

    #
    #Get paths to luns with each client.  Make sure each path
    #is associated with a port in the bound ports list
    #
    foreach my $client (@client_objs){
        #
        #grab the scsi tool object
        #
        $scsi_tool_obj = $self->{scsit_client_h}->{$client}->{su_ref}->{s_t_obj};

        #
        #test each lun and vserver
        #
        my @lun_objects = $san_obj->get_objects(-type    => "lun");

        foreach my $lun_obj (@lun_objects){
            my $lun_name =  $lun_obj->get_path();

            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $vserver = $lun_obj->get_vserver();
                
            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $vol_name = $lun_obj->get_volume();

                if ($lun_obj->error) {
                    $self->_set_err($lun_obj->errmsg());
                    return $self;
                }
            
            if  ($self->{path_hash}->{$vserver}->{$vol_name}->{$lun_name}->{path_hash}){
                $self->_test_add_remove($path_hash, $scsi_tool_obj, @portset, @ports_added);
            }

        }
    }
}

####################################################################################
## Method:     verify_portset_remove                                               #
## Objective:  Verify that when we remove a port, the paths behave correctly       #
## Inputs:     1. (array) Port or ports added to the portest                       #
##             2. (array) Port or ports already in the portset                     #
##             3. (object) pre lun attribute hash (optional)                       #
## Output      4. Nothing, or an error with message                                #
####################################################################################
sub verify_portset_remove{
    my ($self, %args) = @_;
    my (@ports_removed, @portset, $pre_lun_attr);

    foreach my $key (%args){
        if ($key =~ /ports_removed/i){
            @ports_removed = split(/,/, $args{$key});
        }
        elsif ($key =~ /portset/){
            @portset = split(/,/, $args{$key});
        }
        elsif ($key =~ /pre_lun_attribute/i){
            $pre_lun_attr = $args{$key};
        }
        else{
            $self->_set_err("Error: unhandled argument $key\n");
            return $self;
        }
    }

    #
    #Grab the client objects
    #
    my @client_objs;
    if(ref($san_obj) =~ /^San$/i){
        @client_objs = @{$self->{san_obj}->{clients}};
    }elsif(ref($san_obj) =~ /FitMgr|SanMgr/i){
        @client_objs = $self->{san_obj}->get_objects(-type    => "client");
    }else{
        $self->set_err("Error: San object is of unknown type ".ref($san_obj)."\n");
        return $self;
    }

    #
    #Get paths to luns with each client.  Make sure each path
    #is associated with a port in the bound ports list
    #
    foreach my $client (@client_objs){
        #
        #grab the scsi tool object
        #
        $scsi_tool_obj = $self->{scsit_client_h}->{$client}->{su_ref}->{s_t_obj};

        #
        #test each lun and vserver
        #
        my @lun_objects = $san_obj->get_objects(-type    => "lun");

        foreach my $lun_obj (@lun_objects){
            my $lun_name =  $lun_obj->get_path();

            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $vserver = $lun_obj->get_vserver();
                
            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $vol_name = $lun_obj->get_volume();

                if ($lun_obj->error) {
                    $self->_set_err($lun_obj->errmsg());
                    return $self;
                }
            
            if  ($self->{path_hash}->{$vserver}->{$vol_name}->{$lun_name}->{path_hash}){
                $self->_test_add_remove($path_hash, $scsi_tool_obj, @portset, @ports_removed);
            }

        }
    }
}

##########################    PRIVATE METHODS     ##############################################
sub _test_bind{
    my ($self, $paths_hash, $scsi_tool_obj, @bound_ports) = @_;

    my @paths = (@$paths_hash->{pri}, @$paths_hash->{sec});

    foreach my $path (@paths){
        my $staus = "GOOD";
        my $sense_key = undef;
        my $ascq = undef;

        unless (grep {$_ eq $path} @bound_ports) {
             #The path is/was not within the portset
             $status = "CHECK CONDITION";
             $sense_key = "UNIT ATTENTION";
             $ascq = "3fh/0eh";                         #lun data has changed
        }

         my $return =
                      $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                    -cdb_name    => "READ_10",
                                                    -buf_len     => 512,
                                                    -scsi_status => $status,
                                                    -sense_key   => $sense_key, 
                                                    -ascnq       => $ascnq,);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            $err_hash{error} = 1;
            $err_hash{cmd}   = "read_10";
            $err_hash{path}  = $path;
            return \%err_hash;
        }

        if ($staus !~ /GOOD/){
            #We need to make sure we get illegal request this time
            my $return =
                          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                        -cdb_name    => "READ_10",
                                                        -buf_len     => 512,
                                                        -scsi_status => "CHECK CONDITION",
                                                        -sense_key   => "ILLEGAL REQUEST",);

            if ($scsi_tool_obj->error) {
                $self->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }
        }
    }
}

sub _test_unbind{
    my ($self, $paths_hash, $scsi_tool_obj, @bound_ports) = @_;

    my @paths = (@$paths_hash->{pri}, @$paths_hash->{sec});

    foreach my $path (@paths){
        my $staus = "GOOD";
        my $sense_key = undef;
        my $ascq = undef;

        unless (grep {$_ eq $path} @bound_ports) {
             #The path is/was not within the portset
             $status = "CHECK CONDITION";
             $sense_key = "UNIT ATTENTION";
             $ascq = "29h/00h";                         #Power on, or reset
        }

         my $return =
                      $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                    -cdb_name    => "READ_10",
                                                    -buf_len     => 512,
                                                    -scsi_status => $status,
                                                    -sense_key   => $sense_key, 
                                                    -ascnq       => $ascnq,);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            $err_hash{error} = 1;
            $err_hash{cmd}   = "read_10";
            $err_hash{path}  = $path;
            return \%err_hash;
        }

        if ($staus !~ /GOOD/){

            #Now we are going to get a lun data changed attention
             $return =
                     $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                   -cdb_name    => "READ_10",
                                                   -buf_len     => 512,
                                                   -scsi_status => "CHECK CONDITION",
                                                   -sense_key   => "UNIT ATTENTION",
                                                   -ascnq       => "3fh/0eh");         #Reported Luns data has changed

            if ($scsi_tool_obj->error) {
                $self->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            #We need to make sure we get illegal request this time
            $return =
                    $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                  -cdb_name    => "READ_10",
                                                  -buf_len     => 512,
                                                  -scsi_status => "CHECK CONDITION",
                                                  -sense_key   => "ILLEGAL REQUEST",);

            if ($scsi_tool_obj->error) {
                $self->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }
        }
    }
}


sub _test_add{
    my ($self, $paths_hash, $scsi_tool_obj, @portset, @ports_added) = @_;

    my @paths = (@$paths_hash->{pri}, @$paths_hash->{sec});

    foreach my $path (@paths){
        my $staus = "GOOD";
        my $sense_key = undef;
        my $ascq = undef;

        unless (grep {$_ eq $path} @bound_ports) {
             #The path is/was not within the portset
             $status = "CHECK CONDITION";
             $sense_key = "ILLEGAL REQUEST";
        }
        
        if (grep {$_ eq $path} @changed_ports) {
             #The path is one just added or removed from the portset
             $status = "CHECK CONDITION";
             $sense_key = "UNIT ATTENTION";
             $ascq = "29h/00h";                         #POR
        }

         my $return =
                      $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                    -cdb_name    => "READ_10",
                                                    -buf_len     => 512,
                                                    -scsi_status => $status,
                                                    -sense_key   => $sense_key, 
                                                    -ascnq       => $ascnq,);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            $err_hash{error} = 1;
            $err_hash{cmd}   = "read_10";
            $err_hash{path}  = $path;
            return \%err_hash;
        }

        if ($ascq =~ /29h\/00h/){
            #We get reported luns data has changed attention
            $return =
                    $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                  -cdb_name    => "READ_10",
                                                  -buf_len     => 512,
                                                  -scsi_status => "CHECK CONDITION",
                                                  -sense_key   => "UNIT ATTENTION"
                                                  -ascnq       => "3fh/0eh",);       #report luns data has changed

            if ($scsi_tool_obj->error) {
                $self->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

            #Get Good this time
            $return =
                    $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                  -cdb_name    => "READ_10",
                                                  -buf_len     => 512,
                                                  -scsi_status => "GOOD",);

            if ($scsi_tool_obj->error) {
                $self->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }
        }
    }
}



sub _test_remove{
    my ($self, $paths_hash, $scsi_tool_obj, @portset, @ports_removed) = @_;

    my @paths = (@$paths_hash->{pri}, @$paths_hash->{sec});

    foreach my $path (@paths){
        my $staus = "GOOD";
        my $sense_key = undef;
        my $ascq = undef;

        unless (grep {$_ eq $path} @bound_ports) {
             #The path is/was not within the portset
             $status = "CHECK CONDITION";
             $sense_key = "ILLEGAL REQUEST";
        }
        
        if (grep {$_ eq $path} @changed_ports) {
             #The path is one just removed from the portset
             $status = "CHECK CONDITION";
             $sense_key = "UNIT ATTENTION";
             $ascq = "3fh/0eh";                         #report luns data has changed
        }

         my $return =
                      $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                    -cdb_name    => "READ_10",
                                                    -buf_len     => 512,
                                                    -scsi_status => $status,
                                                    -sense_key   => $sense_key, 
                                                    -ascnq       => $ascnq,);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            $err_hash{error} = 1;
            $err_hash{cmd}   = "read_10";
            $err_hash{path}  = $path;
            return \%err_hash;
        }

        if ($ascq =~ /3fh\/0eh/){
            #We get  illegal request
            $return =
                    $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                                  -cdb_name    => "READ_10",
                                                  -buf_len     => 512,
                                                  -scsi_status => "CHECK CONDITION",
                                                  -sense_key   => "ILLEGAL REQUEST",);

            if ($scsi_tool_obj->error) {
                $self->_set_err($scsi_tool_obj->errmsg());
                $err_hash{error} = 1;
                $err_hash{cmd}   = "read_10";
                $err_hash{path}  = $path;
                return \%err_hash;
            }

        }
    }
}

1;
