# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary: Provide an interface for retreiving Logical Unit configuration
## information
##
## @author dsharma@netapp.com

package San::SanScsiT::Data::LUConfig;

#
# Compiler directives.
#

use strict;

#
# Module import.
#

use Error;
use San::SanLog;
use San::SanUtil;
use Data::Compare;
use Params::Validate qw( :all );

use vars qw(@ISA);

#
# Inherit from the Error class
#
@ISA = qw(Error SanLog);

###############################################################################
#                            Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new LUConfig object.
# Details:   See POD documentation at the end of this file.
###############################################################################

sub new {
    my %args = @_;

    #
    # Create a hash to store attributes and bless the hash to create an object.
    #
    my $self = {};
    bless $self;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {xscsit_obj => {type     => OBJECT,
                                   required => 1,
                                   isa      => 'San::SanScsiT::XScsiT'}});
    };

    if ($@) {
        $self->_set_err($@);
        return $self;
    }

    my $xscsit_obj = $args{'xscsit_obj'};

    #
    # add the scsi tool obj debug level.
    #

    $self->set_debug_level(3);

    $self->{xscsit_obj} = $xscsit_obj;

    return $self;
}

###############################################################################
# Method:    get_LU_config_data
# Objective: Get
# LU ID - serial number from Std Inq data and other Inq pages as well
# Block size - read capacity, mode sense
# Geometry - Rigid Disk Geometry mode page
# ALUA - enabled/disabled
# Size - read capacity
# Thin Provisioned - read capacity
# Vdisk related info - GVA Cmds
# Read-Only - read/write cmds
# Details:   See POD documentation at the end of this file.
# Notes:     The PR_IN Command will return all the keys which are
#            registered on a Logical Unit. Therefore the hash is built
#            based on one of the primary paths
#
# Hash Dump
#$VAR1 = \{
#   'vs0' => {
#           'ssan-280r-11' => {
#                    '/vol/stib_vol1/lun_offline_sol1' => {
#                                                           'tp_info' => {
#                                                                          'thin_prov_read_zeros' => 1,
#                                                                          'thin_prov_enabled' => 1
#                                                                        },
#                                                           'alua_state' => 0,
#                                                           'lun_size' => 838860800,
#                                                           'vdisk_info' => {
#                                                                             'mlun_count' => 5,
#                                                                             'suffix_stream' => 2097152,
#                                                                             'prefix_stream' => 1048576,
#                                                                             'vdisk_nv_fail' => 0,
#                                                                             'vdisk_no_space' => 0,
#                                                                             'vdisk_read_only' => 0,
#                                                                             'igroup_name' => 'ssan-280r-11_iscsi^@^@',
#                                                                             'vdisk_offline' => 0,
#                                                                             'vdisk_invalid' => 0,
#                                                                             'vdisk_path' => '/vol/stib_vol1/lun_offline_sol1',
#                                                                             'main_stream' => 835715072
#                                                                           },
#                                                           'blk_size' => {
#                                                                           'mode_sense_10_blk_size' => 512,
#                                                                           'read_cap_blk_size' => 512
#                                                                         },
#                                                           'serial_num' => {
#                                                                             'inq_serial_num_std' => '2Fi3S?9Cn7CF',
#                                                                             'inq_serial_num_80' => '2Fi3S?9Cn7CF'
#                                                                           },
#                                                           'geometry_info' => {
#                                                                                'num_heads' => 16,
#                                                                                'num_cylinders' => 800,
#                                                                                'rotation_rate' => 15000
#                                                                              }
#                                                         },
#                    '/vol/stib_vol1/lun_offline_sol0' => {
#                                                           'tp_info' => {
#                                                                          'thin_prov_read_zeros' => 1,
#                                                                          'thin_prov_enabled' => 1
#                                                                        },
#                                                           'alua_state' => 0,
#                                                           'lun_size' => 838860800,
#                                                           'vdisk_info' => {
#                                                                             'mlun_count' => 5,
#                                                                             'suffix_stream' => 2097152,
#                                                                             'prefix_stream' => 1048576,
#                                                                             'vdisk_nv_fail' => 0,
#                                                                             'vdisk_no_space' => 0,
#                                                                             'vdisk_read_only' => 0,
#                                                                             'igroup_name' => 'ssan-280r-11_iscsi^@^@',
#                                                                             'vdisk_offline' => 0,
#                                                                             'vdisk_invalid' => 0,
#                                                                             'vdisk_path' => '/vol/stib_vol1/lun_offline_sol0',
#                                                                             'main_stream' => 835715072
#                                                                           },
#                                                           'blk_size' => {
#                                                                           'mode_sense_10_blk_size' => 512,
#                                                                           'read_cap_blk_size' => 512
#                                                                         },
#                                                           'serial_num' => {
#                                                                             'inq_serial_num_std' => '2Fi3S?9Cn7CE',
#                                                                             'inq_serial_num_80' => '2Fi3S?9Cn7CE'
#                                                                           },
#                                                           'geometry_info' => {
#                                                                                'num_heads' => 16,
#                                                                                'num_cylinders' => 800,
#                                                                                'rotation_rate' => 15000
#                                                                              }
#                                                         }
#                  }
#         }
# };
################################################################################

sub get_LU_config_data {
    my ($self, %args) = @_;
    my %lu_config_data_hash;

    my ($serial_num, $blk_size, $lun_size, $alua_state, $tp_info_ref,
        $vdisk_info_ref, $geometry_info_ref,);

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {client_info_hash => {type     => HASHREF,
                                         required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $client_info_hash = $args{client_info_hash};

    # Iterate through the path_info_hash to get the path info for which PR
    # metadata has to be collected
    # Build the LU config data hash per lun

    foreach my $client (keys %{$client_info_hash}) {
        my $path_info_hash = $client_info_hash->{$client}->{path_hash};
        my $scsi_util_obj  = $client_info_hash->{$client}->{su_ref};

        foreach my $filer (keys %{$path_info_hash}) {
            foreach my $vol_name (keys %{$path_info_hash->{$filer}}) {
                foreach my $lun (keys %{$path_info_hash->{$filer}->{$vol_name}})
                {
                    while (my ($key, $value) =
                         each %{$path_info_hash->{$filer}->{$vol_name}->{$lun}})
                    {

                        my @pri_paths = @{$value->{pri}}
                          if (defined $value->{pri});
                        my @sec_paths = @{$value->{sec}}
                          if (defined $value->{sec});

                        my @path = (@pri_paths, @sec_paths);

                        # Get the Lun Serial Number

                        $serial_num =
                          $self->get_LU_serial_num(-scsi_util_obj => $scsi_util_obj,
                                                   -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return $self;
                        } else {
                            $lu_config_data_hash{$filer}{$client}{$lun}
                              {serial_num} = $serial_num;
                        }

                        $blk_size =
                          $self->get_LU_block_size(-scsi_util_obj => $scsi_util_obj,
                                                   -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return $self;
                        } else {
                            $lu_config_data_hash{$filer}{$client}{$lun}
                              {blk_size} = $blk_size;
                        }

                        $lun_size = $self->get_LU_size(-scsi_util_obj => $scsi_util_obj,
                                                       -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return $self;
                        } else {
                            $lu_config_data_hash{$filer}{$client}{$lun}
                              {lun_size} = $lun_size;
                        }

                        $alua_state =
                          $self->get_ALUA_state(-scsi_util_obj => $scsi_util_obj,
                                                -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return $self;
                        } else {
                            $lu_config_data_hash{$filer}{$client}{$lun}
                              {alua_state} = $alua_state;
                        }

                        $tp_info_ref =
                          $self->get_TP_info(-scsi_util_obj => $scsi_util_obj,
                                             -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return $self;
                        } else {
                            $lu_config_data_hash{$filer}{$client}{$lun}
                              {tp_info} = $tp_info_ref;
                        }

                        $vdisk_info_ref =
                          $self->get_vdisk_info(-scsi_util_obj => $scsi_util_obj,
                                                -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return $self;
                        } else {
                            $lu_config_data_hash{$filer}{$client}{$lun}
                              {vdisk_info} = $vdisk_info_ref;

                            my @vdisk_info;
                            push(@vdisk_info, $vdisk_info_ref);
                        }

                        $geometry_info_ref =
                          $self->get_geometry_info(-scsi_util_obj => $scsi_util_obj,
                                                   -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return $self;
                        } else {
                            $lu_config_data_hash{$filer}{$client}{$lun}
                              {geometry_info} = $geometry_info_ref;
                        }
                    }
                }
            }
        }
    }

    return \%lu_config_data_hash;
}

###############################################################################
# Method:    get_LU_serial_num
# Objective: Get Logical Unit serial number from Inquiry page 80h
#            and GVA
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_LU_serial_num {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths    => {type     => ARRAYREF,
                                      required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

    my @serial_hash;

    foreach my $path (@$paths) {
        my %serial_num_hash;

        my $return =
          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                        -cdb_name    => "inquiry_page_80h",
                                        -buf_len     => 512,
                                        -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return->{SCSI_DATA}) {
            $return =
              $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "inquiry_page_80h",
                                         -scsi_data_ref => $return->{SCSI_DATA},
              );

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success:",
                              -msg_level => 3);
            }

            # remove null characters
            $return->{PRODUCT_SERIAL_NUMBER} =~ s/[\0]*$//g;

            $serial_num_hash{inq_serial_num_80} =
              $return->{PRODUCT_SERIAL_NUMBER};
        }

        my $return1 =
          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                        -cdb_name    => "inquiry",
                                        -buf_len     => 512,
                                        -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return1->{SCSI_DATA}) {
            $return1 =
              $scsi_util_obj->scsi_bytes_to_strings(
                                        -scsi_cmd_name => "inquiry",
                                        -scsi_data_ref => $return1->{SCSI_DATA},
              );

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success:",
                              -msg_level => 3);
            }

            # remove null characters
            $return1->{VENDOR_SPECIFIC} =~ s/[\0]*$//g;

            $serial_num_hash{inq_serial_num_std} = $return1->{VENDOR_SPECIFIC};
        }

        # Xlate function has to be fixed

        #my $return2 =
        #$scsi_tool_obj->test_scsi_cmd(-device      => $path,
        #                               -cdb_name    => "inquiry_page_83h",
        #                               -buf_len     => 512,
        #                               -scsi_status => "GOOD",);
        #if ($scsi_tool_obj->error) {
        #   $self->_set_err($scsi_tool_obj->errmsg());
        #   return 0;
        #}
        #
        #if ($return2->{SCSI_DATA}) {
        #  $return2 =
        #     $scsi_util_obj->scsi_bytes_to_strings(
        #                                      -scsi_cmd_name => "inquiry_page_83h",
        #                                      -scsi_data_ref => $return2->{SCSI_DATA},
        #    );
        #
        #   if ($scsi_util_obj->error) {
        #       $self->_set_err($scsi_util_obj->errmsg());
        #       return 0;
        #   } else {
        #       $self->sanlog(-msg       => "Success:",
        #                     -msg_level => 3);
        #   }
        #
        #   my @id_list = @{$return2->{ID_LIST}};
        #
        #   $serial_num_hash{inq_serial_num_83} = $id_list[0]->{LUN_SERIAL_NUMBER};
        #}

        push(@serial_hash, \%serial_num_hash);
    }

    # Verify that the data received is the same across all the paths
    # return the first element of the array

    for (my $ind = 1 ; $ind < scalar(@serial_hash) ; $ind++) {
        my $c = new Data::Compare($serial_hash[0], $serial_hash[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return 0;
        }
    }

    return $serial_hash[0];
}

###############################################################################
# Method:    get_LU_block_size
# Objective:
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_LU_block_size {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths    => {type     => ARRAYREF,
                                      required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

    my @blk_size;

    foreach my $path (@$paths) {

        my %blk_size_hash;

        my $return =
          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                        -cdb_name    => "read_capacity_16",
                                        -buf_len     => 512,
                                        -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return->{SCSI_DATA}) {
            $return =
              $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "read_capacity_16",
                                         -scsi_data_ref => $return->{SCSI_DATA},
              );

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success:",
                              -msg_level => 3);
            }

            $blk_size_hash{read_cap_blk_size} =
              $return->{BLOCK_LENGTH_IN_BYTES};
        }

        $return =
          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                        -cdb_name    => "mode_sense_10",
                                        -buf_len     => 512,
                                        -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return->{SCSI_DATA}) {
            $return =
              $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "mode_sense_10",
                                         -scsi_data_ref => $return->{SCSI_DATA},
              );

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success:",
                              -msg_level => 3);
            }

            $blk_size_hash{mode_sense_10_blk_size} = $return->{BLOCK_LENGTH};
        }

        push(@blk_size, \%blk_size_hash);
    }

    # Block size from mode sense page is pending

    # Verify that the data received is the same across all the paths
    # return the first element of the array

    for (my $ind = 1 ; $ind < scalar(@blk_size) ; $ind++) {
        my $c = new Data::Compare($blk_size[0], $blk_size[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return 0;
        }
    }

    return $blk_size[0];
}

###############################################################################
# Method:    get_LU_size
# Objective:
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_LU_size {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths    => {type     => ARRAYREF,
                                      required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

    my @lun_size;

    foreach my $path (@$paths) {
        my $return =
          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                        -cdb_name    => "read_capacity_16",
                                        -buf_len     => 512,
                                        -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return->{SCSI_DATA}) {
            $return =
              $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "read_capacity_16",
                                         -scsi_data_ref => $return->{SCSI_DATA},
              );

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success:",
                              -msg_level => 3);
            }

            my $lu_size = ($return->{RETURNED_LOGICAL_BLOCK_ADDRESS} + 1) * 512;
            push(@lun_size, $lu_size);

        }
    }

    # Verify that the data received is the same across all the paths
    # return the first element of the array

    for (my $ind = 1 ; $ind < scalar(@lun_size) ; $ind++) {
        my $c = new Data::Compare($lun_size[0], $lun_size[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return 0;
        }
    }
    return $lun_size[0];
}

###############################################################################
# Method   : get_ALUA_state
# Objective: Return ALUA state (enabled/disabled)
# Details  : See POD documentation at the end of this file.
###############################################################################

sub get_ALUA_state {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths    => {type     => ARRAYREF,
                                      required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

    my @alua_states;

    foreach my $path (@$paths) {
        my $return =
          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                        -cdb_name    => "inquiry",
                                        -buf_len     => 512,
                                        -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return->{SCSI_DATA}) {
            $return =
              $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "inquiry",
                                         -scsi_data_ref => $return->{SCSI_DATA},
              );

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success:",
                              -msg_level => 3);
            }

            push(@alua_states, $return->{TPGS});
        }
    }

    # Verify that the data received is the same across all the paths
    # return the first element of the array

    for (my $ind = 1 ; $ind < scalar(@alua_states) ; $ind++) {
        my $c = new Data::Compare($alua_states[0], $alua_states[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return 0;
        }
    }

    return $alua_states[0];
}

###############################################################################
# Method   : get_TP_info
# Objective: Return thin provisioning information
# return TPE and TPRZ bits from read capacity data
# TPE bit - Thin provisioning enabled bit is set to one indicating that
#           the logical unit implements thin provisioning.
# TPRZ - Thin provisioning read zeroes bit is set to one indicating that
#        a read operation of an unmapped LBA shall retrieve data with
#        all bits set to zero
# Details  : See POD documentation at the end of this file.
###############################################################################

sub get_TP_info {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths    => {type     => ARRAYREF,
                                      required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

    my @tp_info;

    foreach my $path (@$paths) {
        my %tp_info_hash;

        my $return =
          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                        -cdb_name    => "read_capacity_16",
                                        -buf_len     => 512,
                                        -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return->{SCSI_DATA}) {
            $return =
              $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "read_capacity_16",
                                         -scsi_data_ref => $return->{SCSI_DATA},
              );

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success:",
                              -msg_level => 3);
            }

            $tp_info_hash{thin_prov_enabled}    = $return->{TPE};
            $tp_info_hash{thin_prov_read_zeros} = $return->{TPRZ};

            push(@tp_info, \%tp_info_hash);
        }
    }

    # Block size from mode sense page is pending

    # Verify that the data received is the same across all the paths
    # return the first element of the array

    for (my $ind = 1 ; $ind < scalar(@tp_info) ; $ind++) {
        my $c = new Data::Compare($tp_info[0], $tp_info[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return 0;
        }
    }

    return $tp_info[0];
}

###############################################################################
# Method   : get_vdisk_info
# Objective: Return vdisk related info such as path, state etc in a hash
# Details  : See POD documentation at the end of this file.
###############################################################################

sub get_vdisk_info {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths    => {type     => ARRAYREF,
                                      required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

    my @vdisk_info;

    foreach my $path (@$paths) {
        my %vdisk_info_hash;

        my $return =
          $scsi_tool_obj->test_scsi_cmd(
                                      -device   => $path,
                                      -cdb_name => "gva_virtual_disk_path_page",
                                      -buf_len  => 512,
                                      -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return->{SCSI_DATA}) {
            $return =
              $scsi_util_obj->scsi_bytes_to_strings(
                                 -scsi_cmd_name => "gva_virtual_disk_path_page",
                                 -scsi_data_ref => $return->{SCSI_DATA},);

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success:",
                              -msg_level => 3);
            }

            # remove null character
            $return->{VDISK_PATH} =~ s/[\0]*$//g;

            $vdisk_info_hash{vdisk_path} = $return->{VDISK_PATH};

        }

        $return =
          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                        -cdb_name    => "gva_vdisk_status_page",
                                        -buf_len     => 512,
                                        -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return->{SCSI_DATA}) {
            $return =
              $scsi_util_obj->scsi_bytes_to_strings(
                                      -scsi_cmd_name => "gva_vdisk_status_page",
                                      -scsi_data_ref => $return->{SCSI_DATA},);

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success:",
                              -msg_level => 3);
            }

            # remove prepended or trailing non word characters
            $vdisk_info_hash{vdisk_read_only} = $return->{ROFS};
            $vdisk_info_hash{vdisk_no_space}  = $return->{NOSPC};

            # no virtual disk
            $vdisk_info_hash{vdisk_invalid} = $return->{NOENT};

            $vdisk_info_hash{vdisk_nv_fail} = $return->{NVFAIL};
            $vdisk_info_hash{vdisk_offline} = $return->{OFFLINE};

        }

        $return =
          $scsi_tool_obj->test_scsi_cmd(
                                     -device   => $path,
                                     -cdb_name => "gva_vdisk_stream_sizes_page",
                                     -buf_len  => 512,
                                     -scsi_status => "GOOD",);
        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return $TCD::FAIL;
        }

        my $h_ref =
          $scsi_util_obj->scsi_bytes_to_strings(
                                -scsi_cmd_name => "gva_vdisk_stream_sizes_page",
                                -scsi_data_ref => $return->{SCSI_DATA},);

        if ($scsi_util_obj->error) {
            $self->_set_err($scsi_util_obj->errmsg());
            return $TCD::FAIL;
        }

        $vdisk_info_hash{main_stream}   = $h_ref->{MAIN_STREAM};
        $vdisk_info_hash{prefix_stream} = $h_ref->{PREFIX_STREAM};
        $vdisk_info_hash{suffix_stream} = $h_ref->{SUFFIX_STREAM};

        $return =
          $scsi_tool_obj->test_scsi_cmd(
                                      -device   => $path,
                                      -cdb_name => "gva_vdisk_mapped_luns_page",
                                      -buf_len  => 512,
                                      -scsi_status => "GOOD",);
        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return $TCD::FAIL;
        }

        $h_ref =
          $scsi_util_obj->scsi_bytes_to_strings(
                                 -scsi_cmd_name => "gva_vdisk_mapped_luns_page",
                                 -scsi_data_ref => $return->{SCSI_DATA},);

        if ($scsi_util_obj->error) {
            $self->_set_err($scsi_util_obj->errmsg());
            return $TCD::FAIL;
        }

        $vdisk_info_hash{igroup_name} = $h_ref->{IGROUP_NAME};
        $vdisk_info_hash{mlun_count}  = $h_ref->{MLUN_COUNT};

        push(@vdisk_info, \%vdisk_info_hash);
    }

    # Verify that the data received is the same across all the paths
    # return the first element of the array

    for (my $ind = 1 ; $ind < scalar(@vdisk_info) ; $ind++) {
        my $c = new Data::Compare($vdisk_info[0], $vdisk_info[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return 0;
        }
    }

    return $vdisk_info[0];
}

###############################################################################
# Method   : get_geometry_info
# Objective: Return mode_sense geometry page information
# Details  : See POD documentation at the end of this file.
###############################################################################

sub get_geometry_info {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths    => {type     => ARRAYREF,
                                      required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

    my @geometry_info;

    foreach my $path (@$paths) {
        my %geometry_info_hash;

        my $return =
          $scsi_tool_obj->test_scsi_cmd(-device      => $path,
                                        -cdb_name    => "mode_geometry_page",
                                        -buf_len     => 512,
                                        -scsi_status => "GOOD",);

        if ($scsi_tool_obj->error) {
            $self->_set_err($scsi_tool_obj->errmsg());
            return 0;
        }

        if ($return->{SCSI_DATA}) {
            $self->sanlog(-msg       => "Before scsi_bytes_to_strings",
                          -msg_level => 3);

            $return =
              $scsi_util_obj->scsi_bytes_to_strings(
                                         -scsi_cmd_name => "mode_geometry_page",
                                         -scsi_data_ref => $return->{SCSI_DATA},
              );

            if ($scsi_util_obj->error) {
                $self->_set_err($scsi_util_obj->errmsg());
                return 0;
            } else {
                $self->sanlog(-msg       => "Success",
                              -msg_level => 3);
            }

            $geometry_info_hash{num_cylinders} = $return->{NUMBER_OF_CYLINDERS};
            $geometry_info_hash{num_heads}     = $return->{NUMBER_OF_HEADS};
            $geometry_info_hash{rotation_rate} =
              $return->{MEDIUM_ROTATION_RATE};
        }

        push(@geometry_info, \%geometry_info_hash);
    }

    # Verify that the data received is the same across all the paths
    # return the first element of the array

    for (my $ind = 1 ; $ind < scalar(@geometry_info) ; $ind++) {
        my $c = new Data::Compare($geometry_info[0], $geometry_info[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return 0;
        }
    }

    return $geometry_info[0];
}

###############################################################################
#                             POD Documentation
###############################################################################

=head1 Public Methods

=head2 B<new(%args)>

Creates a new LUConfig object

=over 4

=item Inputs

=item * -xscsit_obj

xscsit_obj Object created with the path information for a logical unit

=item Examples

 my $lu_config_data_obj =
      San::SanScsiT::Data::LUConfig::new(-xscsit_obj => $xscsit_obj,);

    if ($lu_config_data_obj->error()) {
        $self->_set_err($lu_config_data_obj->errmsg());
        return $self;
    }

=back

=head2 B<get_LU_config_data($path)>

Gathers LU config information

=over 4

=item Inputs

=item *I<path_info_hash>

hash containing scsit path information to a LU

=item Examples

my $config_data =
      $xscsit_obj->{lu_config_data_obj}->get_LU_config_data($xscsit_obj->{path_hash});

=back

=head2 B<get_LU_serial_num($path)>

Acquires the LU serial number via SCSI inquiry commands (std inquiry,
Inquiry page 83h and 80h)

=over 4

=item Inputs

=item *I<path_info_hash>

hash containing scsit path information to a LU

=item Output

Hash containing the serial number of the LU from
standard inquiry, Inquiry page 83h and 80h

=item Examples

my $config_data =
      $xscsit_obj->{lu_config_data_obj}->get_LU_serial_num($xscsit_obj->{path_hash});

=back

=head2 B<get_LU_block_size($path)>

Acquires the LU block size from SCSI cmd Read Capacity 16 and Mode Sense 10

=over 4

=item Inputs

=item *I<path_info_hash>

hash containing scsit path information to a LU

=item Output

Hash containing the LU block size from SCSI cmd Read Capacity 16 and Mode
Sense 10 commands

=item Examples

my $config_data =
      $xscsit_obj->{lu_config_data_obj}->get_LU_block_size($xscsit_obj->{path_hash});

=back

=cut

1;

