# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary: Provide an interface for retreiving Meta Data info
## Such as the keys used for reservation, reservation type,
## aptpl info
##
## @author dsharma@netapp.com

package San::SanScsiT::Data::MetaData;

#
# Compiler directives.
#

use strict;

#
# Module import.
#

use San::SanLog;
use San::SanUtil;
use Error;
use San::ScsiTool::ScsiPR;
use Data::Compare;
use Params::Validate qw( :all );

use vars qw(@ISA);

#
# Inherit from the Error class
#
@ISA = qw(Error SanLog);

###############################################################################
#                            Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new MetaData object.
# Details:   See POD documentation at the end of this file.
###############################################################################

sub new {
    my %args = @_;

    #
    # Create a hash to store attributes and bless the hash to create an object.
    #
    my $self = {};

    bless $self;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {xscsit_obj => {type     => OBJECT,
                                   required => 1,
                                   isa      => 'San::SanScsiT::XScsiT'}});
    };

    if ($@) {
        $self->_set_err($@);
        return $self;
    }

    my $xscsit_obj = $args{'xscsit_obj'};

    #
    # add the scsi tool obj debug level.
    #

    $self->set_debug_level(3);

    $self->{xscsit_obj} = $xscsit_obj;

    # Create a Persistent Reservation Object to send PR_IN commands
    my $pr_obj = new San::ScsiTool::ScsiPR();

    if ($pr_obj->error()) {
        $self->_set_err("Error :: unable to create PR object");
        return $self;
    }

    $self->{pr_obj} = $pr_obj;

    return $self;
}

###############################################################################
# Method:    get_meta_data
# Objective: Get PR Keys, Reservation, PRGeneration, APTPL info and
#            store it in a hash
# Details:   See POD documentation at the end of this file.
# Notes:     The PR_IN Command will return all the keys which are
#            registered on a Logical Unit. Therefore the hash is built
#            based on one of the primary paths
#
# Hash Dump
# $VAR1 = {
#           'HASH(0x46208c8)' => undef,
#           'ssan-3070-5' => {
#                'ssan-346-18' => {
#                     '/vol/stib_vol_pr1/shared_lun_1' => {
#                           'num_keys' => 1,
#                           'reservation' => {
#                                             'type' => 5,
#                                             'key' => 'aa:bb:cc:dd:44:55:66:77',
#                                             'pr_generation' => 0
#                                            },
#                           'aptpl' => 1,
#                           'key' => [
#                                      'aa:bb:cc:dd:44:55:66:77'
#                                    ]
#                  }
#              }
#          }
#     };
###############################################################################

sub get_meta_data {
    my ($self, %args) = @_;
    my ($keys, $reservation, $aptpl,);

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {client_info_hash => {type     => HASHREF,
                                         required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $client_info_hash = $args{'client_info_hash'};

    # Iterate through the path_info_hash to get the path info for which PR
    # metadata has to be collected
    # Build the metadata hash per lun
    my %meta_data_hash;

    foreach my $client (keys %{$client_info_hash}) {

        my $path_info_hash;

        if (defined $client_info_hash->{$client}->{path_hash}) {
            $path_info_hash = $client_info_hash->{$client}->{path_hash};
        } else {
            $self->_set_err("path_hash has to be defined");
            return {};
        }

        my $scsi_util_obj = $client_info_hash->{$client}->{su_ref};

        foreach my $filer (keys %{$path_info_hash}) {
            foreach my $vol_name (keys %{$path_info_hash->{$filer}}) {
                foreach my $lun (keys %{$path_info_hash->{$filer}->{$vol_name}})
                {
                    while (my ($key, $value) =
                         each %{$path_info_hash->{$filer}->{$vol_name}->{$lun}})
                    {

                        my @pri_paths = @{$value->{pri}}
                          if (defined $value->{pri});
                        my @sec_paths = @{$value->{sec}}
                          if (defined $value->{sec});

                        my @path = (@pri_paths, @sec_paths);

                        # Pr In for read keys will return all the keys
                        # registered on a lun

                        $keys =
                          $self->get_PR_Keys(-scsi_util_obj => $scsi_util_obj,
                                             -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return {};
                        } else {
                            $meta_data_hash{$filer}{$client}{$lun}
                              {registered_keys} = $keys;

                            #            $meta_data_hash{$filer}{$client}{$lun}{num_keys} = scalar(@$keys);
                        }

                        # Get the reservation type that exists on the lun
                        $reservation =
                          $self->get_PR_reservation(
                                               -scsi_util_obj => $scsi_util_obj,
                                               -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return {};
                        } else {
                            $meta_data_hash{$filer}{$client}{$lun}
                              {reservation} = $reservation;
                        }

                        # get - Persist thorugh power loss state
                        # (Activated/non activated)

                        $aptpl =
                          $self->get_PR_APTPL(-scsi_util_obj => $scsi_util_obj,
                                              -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return {};
                        } else {
                            $meta_data_hash{$filer}{$client}{$lun}{aptpl} =
                              $aptpl;
                        }
                    }
                }
            }
        }
    }

    return \%meta_data_hash;
}

###############################################################################
# Method:    get_PR_Keys
# Objective: Reports all the keys that are registered on a Logical Unit
#            and the PR generation value
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_PR_Keys {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths => {type     => ARRAYREF,
                                   required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my @pr_keys;

    foreach my $path (@$paths) {

        my $result =
          $self->{pr_obj}->scsi_pr_in(-util            => $scsi_util_obj,
                                      -device          => $path,
                                      -pr_in_type      => SCSI_PR_IN_READ_KEYS,
                                      -expected_result => "GOOD");

        if ($self->{pr_obj}->error) {
            $self->_set_err($self->{pr_obj}->errmsg());
            return {};
        }

        my @keys1 = @{$result->{KEYS}} if (defined $result->{KEYS});

        push(@pr_keys, \@keys1);
    }

    for (my $ind = 1 ; $ind < scalar(@pr_keys) ; $ind++) {
        my $c = new Data::Compare($pr_keys[0], $pr_keys[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return {};
        }
    }

    return $pr_keys[0];
}

###############################################################################
# Method:    get_PR_reservation
# Objective: Reports the reservation type and the key associated
#            with the reservation if they exist
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_PR_reservation {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths => {type     => ARRAYREF,
                                   required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my @pr_reservations;

    foreach my $path (@$paths) {
        my %reservation;

        my $result =
          $self->{pr_obj}->scsi_pr_in(
                                     -util       => $scsi_util_obj,
                                     -device     => $path,
                                     -pr_in_type => SCSI_PR_IN_READ_RESERVATION,
                                     -expected_result => "GOOD");

        if ($self->{pr_obj}->error) {
            $self->_set_err($self->{pr_obj}->errmsg());
            return {};
        }

        $reservation{pr_generation} = $result->{GENERATION};

        if ($result->{ADD_LENGTH} == 16) {
            $reservation{resv_key} = $result->{KEYS};
            $reservation{type}     = $result->{TYPE};
        } else {
            $reservation{resv_key} = 0;
            $reservation{type}     = 0;
        }

        push(@pr_reservations, \%reservation);
    }

    for (my $ind = 1 ; $ind < scalar(@pr_reservations) ; $ind++) {
        my $c = new Data::Compare($pr_reservations[0], $pr_reservations[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return {};
        }
    }

    return $pr_reservations[0];
}

###############################################################################
# Method:    get_PR_APTPL
# Objective: Reports whether the persist through power loss was
#            activated while the Logical unit was being registered
#            using PR_OUT command
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_PR_APTPL {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths => {type     => ARRAYREF,
                                   required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my @aptpl;

    foreach my $path (@$paths) {
        my $result =
          $self->{pr_obj}->scsi_pr_in(-util            => $scsi_util_obj,
                                      -device          => $path,
                                      -pr_in_type      => SCSI_PR_IN_REPORT_CAP,
                                      -expected_result => "GOOD",);

        if ($self->{pr_obj}->error) {
            $self->_set_err($self->{pr_obj}->errmsg());
            return {};
        }

        my $aptpl = $result->{PTPL_A};

        push(@aptpl, $aptpl);
    }

    for (my $ind = 1 ; $ind < scalar(@aptpl) ; $ind++) {
        my $c = new Data::Compare($aptpl[0], $aptpl[$ind]);

        unless ($c->Cmp) {
            $self->_set_err("Data returned on paths don't match");
            return {};
        }
    }

    return $aptpl[0];
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=head1 Public Methods

=head2 B<new(%args)>

Creates a new MetaData object

=over 4

=item Inputs

=item * -xscsit_obj

XSan Object created with the path information for a logical unit

=item Examples

 my $metadata_obj =
      San::SanScsiT::Data::MetaData::new(-xscsit_obj => $xscsit_obj,);

    if ($metadata_obj->error()) {
        $self->_set_err($metadata_obj->errmsg());
        return $self;
    }

=back

=head2 B<get_meta_data($path)>

Gathers Meta Data information

=over 4

=item Inputs

=item *I<path_info_hash>

hash containing scsit path information to a LU

=item Examples

my $meta_data =
      $xscsit_obj->{meta_data_obj}->get_meta_data($xscsit_obj->{path_hash});

=back

=head2 B<get_PR_Keys($path)>

Returns the peristent reservation keys if they are registered

=over 4

=item Inputs

=item *I<path>

scsit path information to a LU

=item Output

Reference to an array containing the persistent reservation keys

=item Examples

 $keys = $meta_data_obj->get_PR_Keys($path);

=back

=head2 B<get_PR_reservation($path)>

Acquires the persistent reservation information

=over 4

=item Inputs

=item *I<path>

scsit path information to a LU

=item Output

Reference to a Hash containing type of reservation and the reservation key

=item Examples

$reservation = $meta_data_obj->get_PR_reservation($path);

=back

=head2 B<get_PR_APTPL($path)>

Acquires the Persist Through Power Loss Information

=over 4

=item Inputs

=item *I<path>

scsit path information to a LU

=item Output

APTPL information (0/1)

=item Examples

$aptpl = $meta_data_obj->get_PR_APTPL($path);

=back

=cut
