#$Id$
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary: Provide an interface for retreiving Logical Unit state
## information
##
## @author chadwell@netapp.com

package San::SanScsiT::Data::StateData;

#
# Compiler directives.
#

use strict;

#
# Module import.
#

use Error;
use San::SanLog;
use San::SanUtil;
use San::ScsiTool::ScsiPR;
use Params::Validate qw( :all );
use Data::Dumper;

use vars qw(@ISA);

#
# Inherit from the Error class
#
@ISA = qw(Error SanLog);

########################################################################
#                            Public Methods
########################################################################

########################################################################
# Method:    new
# Objective: Create new StateData object.
# Details:   See POD documentation at the end of this file.
########################################################################

sub new {
    my %args = @_;

    #
    # Create a hash to store attributes and bless the hash to create an object.
    #
    my $self = {};
    bless $self;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {xscsit_obj => {type     => OBJECT,
                                   required => 1,
                                   isa      => 'San::SanScsiT::XScsiT'}});
    };

    if ($@) {
        $self->_set_err($@);
        return $self;
    }

    my $xscsit_obj = $args{'xscsit_obj'};

    #
    # add the scsi tool obj debug level.
    #

    $self->set_debug_level(3);

    $self->{xscsit_obj} = $xscsit_obj;

    # Create a Persistent Reservation Object to send PR_IN commands
    my $pr_obj = new San::ScsiTool::ScsiPR();

    if ($pr_obj->error()) {
        $self->_set_err("Error :: unable to create PR object");
        return $self;
    }

    $self->{pr_obj} = $pr_obj;

    return $self;
}

########################################################################
# Method:    get_state_data
# Objective: get:
#            data access state - from std inquiry
#            administrative state
#            reservation state
#            reservation type
#            power condition state
#            operational state, and
#            store it in a hash
# Details:   See POD documentation at the end of this file.
# Notes:
########################################################################

sub get_state_data {
    my ($self, %args) = @_;
    my %state_data_hash;
    my ($access_state, $admin_state, $reservation_state, $reservation_type,
        $power_state, $op_state,);

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {client_info_hash => {type     => HASHREF,
                                         required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return {};
    }

    my $client_info_hash = $args{'client_info_hash'};

    foreach my $client (keys %{$client_info_hash}) {

        my $path_info_hash;

        if (defined $client_info_hash->{$client}->{path_hash}) {
            $path_info_hash = $client_info_hash->{$client}->{path_hash};
        } else {
            $self->_set_err("path_hash has to be defined");
            return {};
        }

        my $scsi_util_obj = $client_info_hash->{$client}->{su_ref};

        # Iterate through the path_info_hash to get the path info for which PR
        # metadata has to be collected
        # Build the metadata hash per lun

        foreach my $filer (keys %{$path_info_hash}) {
            foreach my $vol_name (keys %{$path_info_hash->{$filer}}) {
                foreach my $lun (keys %{$path_info_hash->{$filer}->{$vol_name}})
                {
                    while (my ($key, $value) =
                         each %{$path_info_hash->{$filer}->{$vol_name}->{$lun}})
                    {

                        my @pri_paths = @{$value->{pri}}
                          if (defined $value->{pri});
                        my @sec_paths = @{$value->{sec}}
                          if (defined $value->{sec});

                        my @path = (@pri_paths, @sec_paths);

                        # get the access state
                        $access_state =
                          $self->get_access_state(
                                               -scsi_util_obj => $scsi_util_obj,
                                               -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return {};
                        } else {
                            $state_data_hash{$filer}{$client}{$lun}
                              {access_state} = $access_state;
                        }

                        # get the reservation state
                        $reservation_state =
                          $self->get_reservation_state(
                                               -scsi_util_obj => $scsi_util_obj,
                                               -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return {};
                        } else {
                            $state_data_hash{$filer}{$client}{$lun}
                              {reservation_state} = $reservation_state;
                        }

                        # get reservation type
                        $reservation_type =
                          $self->get_reservation_type(
                                               -scsi_util_obj => $scsi_util_obj,
                                               -test_paths    => \@path);

                        if ($self->error) {
                            $self->_set_err($self->errmsg());
                            return {};
                        } else {
                            $state_data_hash{$filer}{$client}{$lun}
                              {reservation_type} = $reservation_type;
                        }

                        # get the administrative state
                        # @NO-OP currently untestable
                        #$admin_state = $self->get_admin_state(-scsi_util_obj => $scsi_util_obj,
                        #                                       -test_paths    => \@path);
                        #
                        #if ($self->error) {
                        #    $self->sanlog(-msg       => $self->errmsg(),
                        #                   -msg_level => 3);
                        #} else {
                        #    $state_data_hash{$filer}{$client}{$lun}{admin_state} = $admin_state;
                        #}
                        #
                        #get power state
                        #@NO_OP currently untestable
                        #$power_state = $self->get_power_state(-scsi_util_obj => $scsi_util_obj,
                        #                                      -test_paths    => \@path);
                        #
                        #if ($self->error) {
                        #    $self->sanlog(-msg       => $self->errmsg(),
                        #                   -msg_level => 3);
                        #} else {
                        #    $state_data_hash{$filer}{$client}{$lun}{power_state} = $power_state;
                        #}
                        #
                        #get operational state
                        #@NO-OP Currently untestable
                        #$op_state = $self->get_op_state(-scsi_util_obj => $scsi_util_obj,
                        #                                -test_paths    => \@path);
                        #
                        #if ($self->error) {
                        #    $self->sanlog(-msg       => $self->errmsg(),
                        #                   -msg_level => 3);
                        #} else {
                        #     $state_data_hash{$filer}{$client}{$lun}{op_state} = $op_state;
                        #}
                    }
                }
            }
        }
    }

    return \%state_data_hash;
}

########################################################################
# Method:    get_access_state
# Objective: gets the access state: whether or not the device is
#            connected to a valid lun. One byte of information, first
#            three bits are whether or not it's connected. Last five
#            are whether or not the lun is valid.
# Details:   See POD documentation at the end of this file.
########################################################################

sub get_access_state {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths    => {type     => ARRAYREF,
                                      required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return 0;
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};
    my $scsi_tool_obj = $scsi_util_obj->{s_t_obj};

    my @access_states;

    foreach my $path (@$paths) {
        my $std_inq_data =
          $scsi_tool_obj->test_scsi_cmd(-device   => $path,
                                        -buf_len  => 512,
                                        -cdb_name => "inquiry",);

        my $result =
          $scsi_util_obj->xlate_std_inq_bytes($std_inq_data->{SCSI_DATA});

        if ($scsi_util_obj->error) {
            $self->_set_err($scsi_util_obj->errmsg());
            return 0;
        }

        if ($result->{PERIPHERAL_QUALIFIER} == 1) {
            push(@access_states, "Not Connected");
        } elsif ($result->{PERIPHERAL_DEVICE_TYPE} =~ /1F/) {
            push(@access_states, "Invalid Lun");
        } else {
            push(@access_states, "Connected");
        }
    }

    #
    #check to see if the access states all say the same thing
    #regardless of path by making an array whose elements match
    #the last access_state returned from the for loop above. If
    #both arrays are the same size after the grep, then the information
    #is consistent and we can return it
    #
    my @checked_states = grep { $_ eq $access_states[-1] } @access_states;
    if (scalar(@checked_states) == scalar(@access_states)) {
        return $access_states[-1];
    } else {
        $self->set_err("access state information is inconsistent across paths: "
                       . join('\n', @$paths));
        return 0;
    }
}

########################################################################
# Method:    get_reservation_state
# Objective: Reports whether the lun is reserved
# Details:   See POD documentation at the end of this file.
########################################################################

sub get_reservation_state {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths => {type     => ARRAYREF,
                                   required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return 0;
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my @reservation_states;
    foreach my $path (@$paths) {
        my $result =
          $self->{pr_obj}->scsi_pr_in(
                                     -util       => $scsi_util_obj,
                                     -device     => $path,
                                     -pr_in_type => SCSI_PR_IN_READ_RESERVATION,
                                     -expected_result => "GOOD");

        if ($self->{pr_obj}->error) {
            $self->_set_err($self->{pr_obj}->errmsg());
            return 0;
        }

        push(@reservation_states, $result->{RESERVED});
    }

    #
    #check to see if the access states all say the same thing
    #regardless of path by making an array whose elements match
    #the last access_state returned from the for loop above. If
    #both arrays are the same size after the grep, then the information
    #is consistent and we can return it
    #
    my @checked_states =
      grep { $_ eq $reservation_states[-1] } @reservation_states;
    if (scalar(@checked_states) == scalar(@reservation_states)) {
        return $reservation_states[-1];
    } else {
        $self->set_err(
                  "Reservation state information is inconsistent across paths: "
                    . join('\n', @$paths));
        return 0;
    }
}

########################################################################
# Method:    get_reservation_type
# Objective: Reports whether the lun reservation type is spc-2 or pr.
# Details:   See POD documentation at the end of this file.
########################################################################

sub get_reservation_type {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {scsi_util_obj => {type     => OBJECT,
                                      isa      => 'San::ScsiTool::ScsiUtil',
                                      required => 1},
                    test_paths => {type     => ARRAYREF,
                                   required => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return 0;
    }

    my $scsi_util_obj = $args{'scsi_util_obj'};
    my $paths         = $args{'test_paths'};

    my @reservation_types;
    foreach my $path (@$paths) {
        my $result =
          $self->{pr_obj}->scsi_pr_in(
                                     -util       => $scsi_util_obj,
                                     -device     => $path,
                                     -pr_in_type => SCSI_PR_IN_READ_RESERVATION,
                                     -expected_result => "GOOD");

        if ($self->{pr_obj}->error) {
            $self->_set_err($self->{pr_obj}->errmsg());
            return 0;
        }

        my $reservation_type;

        if ($result->{SCSI_STATUS} =~ /RESERVATION CONFLIT|18/i) {
            push(@reservation_types, "scsi2");
        } else {
            push(@reservation_types, "persistent");
        }
    }

    #
    #check to see if the access states all say the same thing
    #regardless of path by making an array whose elements match
    #the last access_state returned from the for loop above. If
    #both arrays are the same size after the grep, then the information
    #is consistent and we can return it
    #
    my @checked_states =
      grep { $_ eq $reservation_types[-1] } @reservation_types;
    if (scalar(@checked_states) == scalar(@reservation_types)) {
        return $reservation_types[-1];
    } else {
        $self->set_err(
                   "Reservation type information is inconsistent across paths: "
                     . join('\n', @$paths));
        return 0;
    }
}

########################################################################
# Method:    get_power_state
# Objective: Reports whether the lun is active, idle, on standby, or stopped
# Details:   See POD documentation at the end of this file.
# NOTES:     NO-OP:  Not testable via scsi
########################################################################

sub get_power_state {
    my ($self, %args) = @_;

    return;
}

########################################################################
# Method:    get_op_state
# Objective: Reports whether the lun paused or unpaused
# Details:   See POD documentation at the end of this file.
# NOTES:     NO-OP:  Not testable via scsi
########################################################################

sub get_op_state {
    my ($self, %args) = @_;

    return;
}

########################################################################
# Method:    get_admin_state
# Objective: Reports the administrative state
# Details:   See POD documentation at the end of this file.
# NOTES:     NO-OP:  Not testable via scsi
########################################################################

sub get_admin_state {
    my ($self, %args) = @_;

    return;
}

1;

