# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary: Provide an interface for retreiving Scsi Attribute
## information
##
## @authors chadwell@netapp.com, dsharma@netapp.com

package San::SanScsiT::XScsiT;

#
#Compiler Directives
#
use strict;

#
#modules to import
#
use San::ScsiTool;
use San::SanUtil;
use San::SanLog;
use San::ScsiTool::ScsiUtil;
use San::SanScsiT::Data::MetaData;
use San::SanScsiT::Data::LUConfig;
use San::SanScsiT::Data::StateData;
use San::SanScsiT::DOT_Features::VerifyLun;
use San::SanScsiT::DOT_Features::VerifyPortset;
use Params::Validate qw( :all );

our @ISA = qw(Error SanUtil);

############################################################################
## Method:    verify
## Objective: Handle calling the appropriate feature_verify method in the
##            DOTFeatures APIs and return success or failure.
## Inputs:    (String) -feature       => [lun | portset] feature type
## Returns:   True on success
##            False on failure
## #########################################################################

sub verify {
    my ($self, %args) = @_;

    my ($feature, $ret);

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    if (!$args{'feature'}) {
        $self->_set_err("-feature and -op are required parameters");
        return 0;
    }

    $feature = $args{'feature'};

    # Delete the argument "feature" as the same hash is going to be passed
    # to $self->verify_lun/ $self->verify_portset

    delete $args{'feature'};

    if ($feature eq "lun") {
        $ret = $self->verify_lun(%args);

    } elsif ($feature eq "portset") {
        $ret = $self->verify_portset(%args);

    } else {
        $self->_set_err("Feature " . $feature . " is not supported");
        return 0;
    }

    if ($self->{error}) {
        $self->_set_err($self->errmsg());
        return 0;
    }

    return $ret;
}

############################################################################
## Method:    get_scsi_attributes
## Objective: create the baseline scsit state used for verification after
##            a feature has been executed.  Contains all pertinent
##            information about the targets current state.
## Input  :   type = type of scsi attribute [metadata|configdata|statedata|all]
## Outputs:   1. Hash of information about the lun.  See POD docs at the end
##               of this file for more information
## #########################################################################

sub get_scsi_attributes {
    my ($self, %args) = @_;

    my ($meta_data, $config_data, $state_data, %attr_hash);

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {type => {
                        type  => SCALAR,
                        regex => qr/^metadata$|^configdata$|^statedata$|^all$/i,
                        default  => "all",
                        optional => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return $self;
    }

    my $type = $args{'type'};

    unless (defined $self->{scsit_client_h}) {
        $self->_set_err("scsit_client_h not defined");
        return $self;
    }

    if ($type =~ /metadata/i) {
        unless (defined $self->{meta_data_obj}) {
            $self->_set_err("meta_data_obj not defined");
            return $self;
        }

        $attr_hash{meta_data} =
          $self->{meta_data_obj}
          ->get_meta_data(-client_info_hash => $self->{scsit_client_h});

        if ($self->{meta_data_obj}->error) {
            $self->_set_err($self->{meta_data_obj}->errmsg());
            return $self;
        }

    } elsif ($type =~ /configdata/i) {

        unless (defined $self->{lu_config_data_obj}) {
            $self->_set_err("lu_config_data_obj not defined");
            return $self;
        }

        $attr_hash{config_data} =
          $self->{lu_config_data_obj}
          ->get_LU_config_data(-client_info_hash => $self->{scsit_client_h});

        if ($self->{lu_config_data_obj}->error) {
            $self->_set_err($self->{lu_config_data_obj}->errmsg());
            return $self;
        }

    } elsif ($type =~ /statedata/i) {

        unless (defined $self->{state_data_obj}) {
            $self->_set_err("state_data_obj not defined");
            return $self;
        }

        $attr_hash{state_data} =
          $self->{state_data_obj}
          ->get_state_data(-client_info_hash => $self->{scsit_client_h});

        if ($self->{state_data_obj}->error) {
            $self->_set_err($self->{state_data_obj}->errmsg());
            return $self;
        }

    } elsif ($type =~ /all/i) {
        unless (    defined $self->{meta_data_obj}
                and defined $self->{lu_config_data_obj}
                and defined $self->{state_data_obj}) {
            $self->_set_err(
                "Objects: state_data_obj, meta_data_obj and  lu_config_data_obj have to be defined"
            );
            return $self;
        }

        $attr_hash{meta_data} =
          $self->{meta_data_obj}
          ->get_meta_data(-client_info_hash => $self->{scsit_client_h});

        if ($self->{meta_data_obj}->error) {
            $self->_set_err($self->{meta_data_obj}->errmsg());
            return $self;
        }

        $attr_hash{config_data} =
          $self->{lu_config_data_obj}
          ->get_LU_config_data(-client_info_hash => $self->{scsit_client_h});

        if ($self->{lu_config_data_obj}->error) {
            $self->_set_err($self->{lu_config_data_obj}->errmsg());
            return $self;
        }

        $attr_hash{state_data} =
          $self->{state_data_obj}
          ->get_state_data(-client_info_hash => $self->{scsit_client_h});

        if ($self->{state_data_obj}->error) {
            $self->_set_err($self->{state_data_obj}->errmsg());
            return $self;
        }

    } else {
        $self->_set_err("$type is not a valid type");
        return $self;
    }

    # return the attribute hash
    return \%attr_hash;
}

############################################################################
## Method:    discover_paths
## Objective: Discover the scsi paths to the luns
## Inputs:    1. (Object) a Lun objects
## #########################################################################

sub discover_paths {
    my ($self, %args) = @_;
    my $lun_obj_ref = {};
    my %lun_path_hash;
    my $force = 0;
    my $lun_type;

    my @client_objs = @{$self->{client_objs}} if (defined $self->{client_objs});
    my @filer_objs  = @{$self->{filer_objs}}  if (defined $self->{filer_objs});
    my @lun_objects = @{$self->{lun_objs}}    if (defined $self->{lun_objs});

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {lun_objs => {type     => ARRAYREF,
                                 optional => 1,},
                    force => {type     => BOOLEAN,
                              optional => 1,
                              default  => 0,},
                    lun_type => {type     => SCALAR,
                                 optional => 1,
                                 default  => "all",},
                    skip_client_disc => {type     => SCALAR,
                                         optional => 1,
                                         default  => 1,}});
    };

    if ($@) {
        $self->_set_err($@);
        return $self;
    }

    $lun_obj_ref = $args{'lun_objs'};
    $force       = $args{'force'};
    $lun_type    = $args{'lun_type'};
    my $skip_client_disc = $args{'skip_client_disc'};

    if (defined $lun_obj_ref) {
        @lun_objects = @{$lun_obj_ref};
    }

    unless (scalar(@lun_objects)) {
        $self->_set_err("No Lun Objects Found");
        return $self;
    }

    use Data::Dumper;

    foreach my $client_obj (@client_objs) {
        my $client_name = $client_obj->hostname();

        if ($client_obj->error) {
            $self->_set_err($client_obj->errmsg());
            return $self;
        }

        unless (
              (ref($self->{scsit_client_h}->{$client_name}->{st_ref}) =~ /any/i)
              or ($skip_client_disc == 1)) {
            $client_obj->discover();
        }

        foreach my $lun_obj (@lun_objects) {
            my (@pri_paths, @sec_paths);
            my $lun_name = $lun_obj->get_path();

            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $filer;

            if ($self->{ontap_mode} eq "CMode") {
                $filer = $lun_obj->get_vserver();

                if ($lun_obj->error) {
                    $self->_set_err($lun_obj->errmsg());
                    return $self;
                }
            } else {
                $filer = $lun_obj->get_filername();

                if ($lun_obj->error) {
                    $self->_set_err($lun_obj->errmsg());
                    return $self;
                }
            }

            my $vol_name = $lun_obj->get_volume();

            if ($lun_obj->error) {
                $self->_set_err($lun_obj->errmsg());
                return $self;
            }

            my $su_obj = $self->{scsit_client_h}->{$client_name}->{su_ref};

            my @devices;

            my $paths_info;

            if (ref($self->{scsit_client_h}->{$client_name}->{st_ref}) =~
                /any/i) {

                # vserver object required to fetcg iscsi portal information
                # while creating any_iscsi device list.
                $paths_info =
                  $su_obj->get_paths_to_lun(
                                       -filer        => $filer,
                                       -lun_obj      => $lun_obj,
                                       -lun          => $lun_name,
                                       -force        => $force,
                                       -vserver_objs => $self->{'vserver_objs'},
                  );
            } else {
                $paths_info =
                  $su_obj->get_paths_to_lun(-filer   => $filer,
                                            -lun_obj => $lun_obj,
                                            -lun     => $lun_name,
                                            -force   => $force,);
            }

            # reset the flag. scanning all luns just once.
            if ($force) {
                $force = 0;
            }

            if ($su_obj->error()) {
                $self->sanlog(-msg       => $su_obj->errmsg(),
                              -msg_level => 0);
            }

            if (defined $paths_info) {
                $lun_path_hash{$filer}{$vol_name}{$lun_name}->{path_info} =
                  $paths_info;
            } else {
                $self->_set_err("No Paths returned for $lun_name on $filer");
                return $self;
            }
            push(@devices, @{$paths_info->{pri}})
              if (defined $paths_info->{pri});
            push(@devices, @{$paths_info->{sec}})
              if (defined $paths_info->{sec});

            $self->{scsit_client_h}->{$client_name}->{path_hash} =
              \%lun_path_hash;

            $su_obj = $self->{scsit_client_h}->{$client_name}->{su_ref};

            $self->{scsit_client_h}->{$client_name}->{$filer}->{$vol_name}
              ->{$lun_name}->{'devices'} = \@devices;

            my @igroup_objs = @{$self->{igroup_objs}}
              if (defined $self->{igroup_objs});

            if ($self->{modify_alua}) {
                for my $igroup (@igroup_objs) {

                    # if alua is set to false set it back to true
                    if ($igroup->get_alua() eq "false") {
                        if ($self->{ontap_mode} eq "7Mode") {
                            $igroup->set(-attribute => "alua",
                                         -value     => "yes",
                                         -force     => "true");
                        } else {
                            $igroup->set(-attribute => "alua",
                                         -value     => "true",
                                         -force     => "true");
                        }

                        if ($igroup->error()) {
                            $self->_set_err($igroup->errmsg());
                        }
                    }
                }
            }

            $su_obj->clear_cc_ua(-paths => \@devices);

            if ($su_obj->error()) {
                $self->_set_err($su_obj->errmsg());
                return $self;
            }
        }
    }

    $self->{path_hash} = \%lun_path_hash;

    return $self->{path_hash};
}

##############           get_scsit_wafl_obj                ####################
# DEscription: Get the Scsi Wafl Object required to fetch LBA/ Block
#              information in the prefix/suffix/lun streams.
#              This is a Cmode only method.
# Input  : Lun Object
#          return 0 - Test Failed (Error message will be set via set_err
#          return scsit_wafl_obj on success
###############################################################################

sub get_scsit_wafl_obj {
    my ($self, $lun_obj) = @_;

    my @filer_objs = @{$self->{'filer_objs'}}
      if (defined $self->{'filer_objs'});

    my $lun = $lun_obj->get_path();

    if ($lun_obj->error) {
        $self->_set_err($lun_obj->errmsg());
        return 0;
    }

    my $volume_obj = $lun_obj->get_parent_obj();

    if ($lun_obj->error) {
        $self->_set_err($lun_obj->errmsg());
        return 0;
    }

    my $vserver_obj = $volume_obj->get_parent_obj();

    if ($volume_obj->error) {
        $self->_set_err($volume_obj->errmsg());
        return 0;
    }

    my $filer_obj = $vserver_obj->get_parent_obj();

    if ($vserver_obj->error) {
        $self->_set_err($vserver_obj->errmsg());
        return 0;
    }

    my $aggr_obj =
      $filer_obj->get_aggregate_obj(name => $volume_obj->get_aggregate());

    if ($filer_obj->error) {
        $self->_set_err($filer_obj->errmsg());
        return 0;
    }

    my $node = $aggr_obj->get_nodes();

    if ($aggr_obj->error) {
        $self->_set_err($aggr_obj->errmsg());
        return 0;
    }

    $self->sanlog(-msg => "Aggregate : " . $aggr_obj->get_name());
    $self->sanlog(-msg => "Node      : " . $node);

    my $filer;

    foreach my $filer_obj (@filer_objs) {
        if ($filer_obj->get_name() =~ /^$node$/) {
            $filer = $filer_obj;
            last;
        }
    }

    $self->sanlog(-msg => "Node Name : " . $filer->get_name());

    my $s_w_obj =
      new san::scsi::scsit_wafl(-filer_obj => $filer,
                                -lun_name  => $lun);

    unless ($s_w_obj) {
        $self->_set_err(  "Failed to create scsit_wafl obj " . "for "
                        . "filer: "
                        . $filer->get_name() . " lun:"
                        . $lun);
        return 0;
    }

    return $s_w_obj;
}

############################################################################
## Method:    new
## Objective: create a new XScsiT object used for feature success verification
##            at the block level.
## Inputs:    1. (Object) a San/SanMgr object
## Outputs:   1. (Object) an XScsiT object
## #########################################################################

sub new {
    my ($class, %args) = @_;

    my ($san_obj,
        $client_obj,
        $debug_level,
        $reboot_required,
        $modify_alua,
        $lun_path,
        $ontap_mode,    # Cmode/7 Mode
        @filer_objs,
        @client_objs,
        @igroup_objs,
        @vserver_objs,
        @lun_objects,
        @filers,
        @clients,
        %clients_h,);

    #
    # Bless the new xscsit object
    # to give it access to its methods
    #
    my $self = {};

    bless $self, $class;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();

    # validate input parameters

    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {san_obj => {
                 type      => OBJECT,
                 required  => 1,
                 callbacks => {
                     'Object Type' => sub {
                         $_[0]      =~ qr/^San/i
                           or $_[0] =~ qr/^San::SanMgr/i
                           or $_[0] =~ qr/^San::FitMgr/i;
                       }
                 },},
             debug_level => {type     => SCALAR,
                             default  => 0,
                             optional => 1},
             reboot_required => {type     => SCALAR,
                                 default  => 1,
                                 optional => 1},
             scsi_tool       => {type     => SCALAR,
                                 optional => 1,
                                 default  => ""},
             scsi_tool_loc   => {type     => SCALAR,
                                 optional => 1,
                                 default  => ""},
             modify_alua     => {type     => SCALAR,
                                 default  => 1,
                                 optional => 1}});
    };

    if ($@) {
        $self->_set_err($@);
        return $self;
    }

    $debug_level     = $args{'debug_level'};
    $reboot_required = $args{'reboot_required'};
    $modify_alua     = $args{'modify_alua'}; 
    my $tool         = $args{'scsi_tool'};
    my $tool_loc     = $args{'scsi_tool_loc'};
    
    #
    # Parse the args list and assign values
    #

    foreach my $param (keys %args) {

        # SanMgr / FitMgr Object
        if ($param =~ /san_obj/i) {
            $san_obj = $args{$param};

            length($san_obj = $args{$param})
              or do {
                $self->_set_err("Invalid value for \"san_obj\": $args{$param}");
                return $self;
              };
        } elsif (/debug_level/) {
            length($debug_level = $args{$param})
              or do {
                $self->_set_err(
                            "Invalid value for \"debug_level\": $args{$param}");
                return $self;
              };
        } elsif (/reboot_required/) {
            $reboot_required = $args{$param};

            length($reboot_required)
              or do {
                $self->_set_err(
                        "Invalid value for \"reboot_required\": $args{$param}");
                return $self;
              };
        }
    }

    $self->{san_obj}     = $san_obj;
    $self->{debug_level} = $debug_level;

    if (ref($san_obj) =~ /^San$/i) {
        if (defined $self->{san_obj}->{filers}) {
            @filer_objs = @{$self->{san_obj}->{filers}};
        } else {
            $self->_set_err("Cant get filer objects from San Obj");
            return $self;
        }

        if (defined $self->{san_obj}->{clients}) {
            @client_objs = @{$self->{san_obj}->{clients}};
        } else {
            $self->_set_err("Cant get client objects from San Obj");
            return $self;
        }

    } elsif (ref($san_obj) =~ /FitMgr|SanMgr/i) {
        @filer_objs = $san_obj->get_objects(-type => "node");

        if ($san_obj->error) {
            $self->_set_err($san_obj->errmsg());
            return $self;
        }

        @vserver_objs = $san_obj->get_objects(-type => "vserver");

        if ($san_obj->error) {
            $self->_set_err($san_obj->errmsg());
            return $self;
        }

        @client_objs = $san_obj->get_objects(-type => "client");

        if ($san_obj->error) {
            $self->_set_err($san_obj->errmsg());
            return $self;
        }

        @igroup_objs = $san_obj->get_objects(-type => "igroup");
        if ($san_obj->error() or not scalar @igroup_objs) {
            $self->_set_err($san_obj->errmsg());
            return $self;
        }
    } else {
        $self->set_err(
                "Error: San object is of unknown type " . ref($san_obj) . "\n");
        return $self;
    }

    if (ref($san_obj) =~ /FitMgr|SanMgr/) {
        @lun_objects = $san_obj->get_objects(-type => "lun");

        if ($san_obj->error) {
            $self->sanlog(-msg       => $san_obj->errmsg(),
                          -msg_level => 1);
        }
    } else {
        if (defined $self->{san_obj}->{config_lun_objs}) {
            @lun_objects = @{$self->{san_obj}->{config_lun_objs}};
        } else {
            $self->sanlog(-msg       => "Can not get lun objects from san_obj",
                          -msg_level => 1);
        }
    }

    if (ref($filer_objs[0]) =~ /7Mode/i) {
        $ontap_mode = "7Mode";
    } else {
        $ontap_mode = "CMode";
    }

    my $meta_data_obj =
      San::SanScsiT::Data::MetaData::new(-xscsit_obj => $self);

    if ($meta_data_obj->error()) {
        $self->_set_err($meta_data_obj->errmsg());
        return $self;
    }

    my $lu_config_data_obj =
      San::SanScsiT::Data::LUConfig::new(-xscsit_obj => $self);

    if ($lu_config_data_obj->error()) {
        $self->_set_err($lu_config_data_obj->errmsg());
        return $self;
    }

    my $state_data_obj =
      San::SanScsiT::Data::StateData::new(-xscsit_obj => $self);

    if ($state_data_obj->error()) {
        $self->_set_err($state_data_obj->errmsg());
        return $self;
    }

    # Get the controller names from the filer object

    foreach my $filer_obj (@filer_objs) {
        my $filer_name = $filer_obj->get_name();

        if ($filer_obj->error) {
            $self->_set_err($filer_obj->errmsg());
            return $self;
        }

        push(@filers, $filer_name);
    }

    foreach my $client_obj (@client_objs) {

        my $client_name = $client_obj->hostname();

        if ($client_obj->error) {
            $self->_set_err($client_obj->errmsg());
            return $self;
        }

        $clients_h{$client_name}{client_obj} = $client_obj;

        push(@clients, $client_name);

        # Step: Instantiate a ScsiTool object

        my %create_scsi_tool_args = (
                                    -client => $clients_h{$client_name}{client_obj},
                                    -debug_level => $debug_level
        );
        
        if ($tool) {
            $create_scsi_tool_args{-tool} = $tool;
        } 
        
        if ($tool_loc) {
            $create_scsi_tool_args{-scsi_tool_loc} = $tool_loc;
        } 

        $clients_h{$client_name}->{st_ref} = ScsiTool->create_scsi_tool( %create_scsi_tool_args );

        if ($clients_h{$client_name}->{st_ref}->error) {
            $self->_set_err("Failed to instantiate Scsi_Tool_Obj object "
                            . $clients_h{$client_name}->{st_ref}->errmsg());
            return $self;
        }

        # Create ScsiUtil Object
        # get the paths to the lun on the client, as an array of
        # devices to the same lun

        $clients_h{$client_name}->{su_ref} = new San::ScsiTool::ScsiUtil(
                          -scsi_tool_obj => $clients_h{$client_name}->{st_ref});

        if ($clients_h{$client_name}->{su_ref}->error()) {
            $self->_set_err($clients_h{$client_name}->{su_ref}->error());
            return $self;
        }

        if (ref($san_obj) =~ /^San$/i) {
            foreach my $filer_obj (@filer_objs) {
                my $href =
                  $clients_h{$client_name}->{su_ref}
                  ->get_resource_names(-san_object => $self->{san_obj});

                if ($clients_h{$client_name}->{su_ref}->error()) {
                    $self->_set_err(
                                   $clients_h{$client_name}->{su_ref}->error());
                    return $self;
                }

                my @igroup_names = @{$href->{igroups}};

                foreach my $igrp (@igroup_names) {
                    my $igroup_obj = $filer_obj->get_igroup(-name => $igrp);

                    unless ($filer_obj->error()) {
                        push(@igroup_objs, $igroup_obj);
                    }
                }
            }
        }

        if ($modify_alua) {
            for my $igroup (@igroup_objs) {

                # if alua is set to true set it back to false
                if ($igroup->get_alua() eq "true") {
                    if ($ontap_mode eq "7Mode") {
                        $igroup->set(-attribute => "alua",
                                     -value     => "no",
                                     -force     => "true");
                    } else {
                        $igroup->set(-attribute => "alua",
                                     -value     => "false",
                                     -force     => "true");
                    }

                    if ($igroup->error()) {
                        $self->_set_err($igroup->errmsg());
                    }
                }
            }
        }

        if ($reboot_required) {

            # routine that will do the appropriate type of
            # reboot (w/ dev cleanup) per hosttype
            $self->sanlog(-msg       => "Client Reboot Requested",
                          -msg_level => 1);

            unless ($client_obj->reboot()) {
                $self->_set_err(
                    "Rebooting the host failed bcos: " . $client_obj->errmsg());
                return $self;
            }
        }
    }

    # Update the object with data

    $self->{meta_data_obj}      = $meta_data_obj;
    $self->{lu_config_data_obj} = $lu_config_data_obj;
    $self->{state_data_obj}     = $state_data_obj;
    $self->{ontap_mode}         = $ontap_mode;
    $self->{filer_objs}         = \@filer_objs;
    $self->{client_objs}        = \@client_objs;
    $self->{igroup_objs}        = \@igroup_objs;
    $self->{lun_objs}           = \@lun_objects;
    $self->{filers}             = \@filers;
    $self->{vserver_objs}       = \@vserver_objs;
    $self->{clients}            = \@clients;
    $self->{scsit_client_h}     = \%clients_h;
    $self->{modify_alua}        = $modify_alua;

    return $self;

}    #end new

#############################################################################
#                             POD Documentation
#############################################################################

=head1 Public Methods

=head2 B<new()>

Creates a new XScsiT object, MetaData object, LU config object and StateData
object.  Gets the paths to the luns created during test setup. ALUA is disabled
on all the igroups created for the test and the client is rebooted to make sure
that the devices are discovered on the host and the alua changes are reflected

=over 4

=item Inputs

=item * I<-san_obj> <required>

San Object created during test set up ( San.pm | FitMgr | SanMgr obj)

=item * I<-debug_level> <optional>

Specify the debug level to be used.
Default is set to 0

=item * I<-reboot_required> <optional>

default = 1

Specify whether a client reboot is required.
This parameter can be disabled if the luns are already discovered on the (solaris) host

=item Outputs

Returns a XScsiT object.

=item Examples

    #
    # Create a XScsiT object
    #

    $san_scsit_obj = new San::SanScsiT::XScsiT(-san_obj => $SanMgr_obj);

    if ($self->error()) {
        $Test->log($self->errmsg());
        return $TCD::CONFIG;
    }

=back

=head2 B<discover_paths(%args)>

Discover the scsi paths to the luns.

=over 4

=item Inputs

=item * I<-lun_objs> <required>

Reference of an array of lun objects for which the scsi paths have to be discovered

=item Outputs

Returns a path hash

=item Examples

    $self->discover_paths(-lun_objs => \@lun_objs);

    if ($self->error()) {
        $Test->log($self->errmsg());
        return $TCD::CONFIG;
    }

=head2 B<verify(%args)>

Handle calling the appropriate feature_verify method in the
DOTFeatures APIs and return success or failure.

=over 4

=item Inputs

=item * I<-feature> <required>

The feature to test
1) Lun
2) Portset

=item * I<-operation> <required>

Operation to be performed on the feature

Allowed Lun operations are
resize
online
serial
device_id
offline
move
create

=item * -I<lun_obj> <required>

Required: Lun Object for which the scsi attributes have to be verified

=item * I<-serial_num> <optional>

The serial num has to be specified if the feature is "Lun" and the operation is "serial"
This should be the value returned by the lun show

ssan-3170-21a1296960328.91798::> lun show -path /vol/stib_vol1/moved_lun_0 -fields serial

vserver path                       serial

------- -------------------------- ------------

vs0     /vol/stib_vol1/moved_lun_0 2Fi3S?9iWzof

=item Examples

    $ret = $self->verify(-feature     => "lun",
                         -operation   => "serial",
                         -lun_obj     => $lun_obj,
                         -serial_num  => $serial_num,
                        );

    if ($self->error()) {
        $Test->nlog($self->errmsg());
        return $TCD::FAIL;
    }

=item * I<-device_id> <optional>

Value of the legacy device ID that has been set on the lun.
This has to be set when the feature is "Lun" and the operation is "device_id"

=item Examples

     $ret = $self->verify(-feature     => "lun",
                          -operation   => "device_id",
                          -lun_obj     => $lun_obj,
                          -device_id   => $dev_id,
                         );

     if ($self->error()) {
         $Test->nlog($self->errmsg());
         return $TCD::FAIL;
     }

=item * I<-new_lun_name> <optional>
=item * I<-old_lun_name> <optional>

These values have to be set when the feature is "Lun" and the operation is "move"
old_lun_name is the name before the lun was renamed
new_lun_name is the name of the lun after it was renamed

=item * I<-pre_lun_attribute> <optional>

Hash containing the o/p of the get_attributes call i.e. the attributes of the
lun before performing the specified operation

=item Examples

    $ret = $self->verify(-feature           => "lun",
                         -operation         => "move",
                         -pre_lun_attribute => $hash_ref,
                         -new_lun_name      => $lun_path,
                         -old_lun_name      => $old_name,
                         -lun_obj           => $lun_obj,
                        );

    if ($self->error()) {
        $Test->nlog($self->errmsg());
        return $TCD::FAIL;
    }

=item * I<-new_size> <optional>
The new size has to be specified if the feature is "Lun" and the operation is "resize"

=item Examples

    $ret = $self->verify(-feature     => "lun",
                         -operation   => "resize",
                         -lun_obj     => $lun_obj,
                         -new_size    => $new_size,
                        );

    if ($self->error()) {
        $Test->nlog($self->errmsg());
        return $TCD::FAIL;
    }

=item Outputs
return 1 on success and error will be set using _set_err in case of a failure


=item Examples

    $ret = $self->verify(-feature   => "lun",
                         -operation => "offline",
                         -lun_obj   => $lun_obj
                        );

    if ($self->error()) {
        $Test->nlog($self->errmsg());
    }

=head2 B<get_scsi_attributes(%args)>

Baseline scsit state used for verification after a feature has been executed.
Contains all pertinent information about the targets current state.

=item * I<-type> <optional>

default: all

Specify the type of scsi attributes to be retrieved
The possible options are (metadata, configdata, statedata or all)

=over 4

=item Outputs
return Hash of information about the lun

=item Examples

    $hash_ref = $self->get_scsi_attributes(-type => "metadata");

    if ($self->error()) {
        $Test->nlog($self->errmsg());
        return $TCD::FAIL;
    }


=back

=head2 B<get_scsit_wafl_obj(%args)>

Get the Scsi Wafl Object required to fetch LBA/ Block 
information in the prefix/suffix/lun streams.
The slist command operates at the node level. 
Hence this method identifies the home node for a 
lun and returns the corresponding scsit_wafl_obj
This is a Cmode only method.                          

=item * I<lun object> <required>

Specify the lun object for which the scsit_wafl object 
has to be returned. 

=over 4

=item Outputs
return scsit_wafl_obj

=item Examples

    $scsit_wafl_obj = $self->get_scsit_wafl_obj($lun_obj);

    if ($self->error()) {
        $Test->nlog($self->errmsg());
        return $TCD::FAIL;
    }


=back

=cut

1;

