# Copyright (c) 2007-2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a SWITCH device
## @author evan@netapp.com
## @status private
## @pod here

package SanSwitch;

# Compiler directives.
use strict;

# Module import.
use Time::HiRes qw(gettimeofday);
use Utils qw(make_host make_conn);
use Tharn;

use San::SanLog;
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::PortMgr;
use San::SanSwitch::ZoneMgr;
use San::SanCache;
use Error;

use vars qw(@ISA $Validation_Errors $Prompts_updated $_CONN_CNT);

@ISA = qw(Error SanLog SanCache);

# Define the default list of validation errors
# Format is as follows:
# $Validateion_Errors->{"string to look for"}->{MSG} = "Message to return";
# $Validateion_Errors->{"string to look for"}->{TYPE} = "Error type of fatal, retryable, or none";
$Validation_Errors = {
        "Operation allowed only if the local switch is part of the current AD" =>
            {
                "MSG" => "This switch's WWN is not currently a member of the specified admin domain. " .
                         "Please add its WWN to the admin domain that you are working with and try again." .
                         "Steps:\n" .
                         "\t1) Run 'switchshow', and obtain the value for 'switchWwn'.\n" .
                         "\t2) Switch to the administrator admin domain '255': 'ad --select 255'.\n" .
                         "\t3) Add the switchWnn obtained from step 1 to admin domain zero: 'ad --add 0 -s <switchWwn>'\n" .
                         "\t4) Apply the change by running 'ad --apply' and answer yes to the question it asks\n" .
                         "\t   it may take a few minutes for this command to complete.",
                "TYPE" => "fatal"
            },
        "Either AD does not exist or user does not have sufficient permissions" =>
            {
                "MSG" => "Invalid AD specified or user has insufficient permissions",
                "TYPE" => "fatal"
            },
        "not found" =>
            {
                "MSG" => "Item was not found on the switch",
                "TYPE" => "fatal"
            },
        "duplicate name" =>
            {
                "MSG" => "Duplicate item was found on the switch",
                "TYPE" => "fatal"
            },
        "invalid alias" =>
            {
                "MSG" => "Alias information is invalid",
                "TYPE" => "fatal"
            },
        "invalid name" =>
            {
                "MSG" => "Name information is invalid",
                "TYPE" => "fatal"
            },
        "invalid zone" =>
            {
                "MSG" => "Zone information is invalid",
                "TYPE" => "fatal"
            },
        "failed\: sfc was rejected" =>
            {
                "MSG" => "Command Failed. SFC was Rejected",
                "TYPE" => "retryable"
            },
        "trans_begin\: there is an outstanding transaction" =>
            {
                "MSG" => "There is an outstanding transaction in progress",
                "TYPE" => "retryable"
            },
        "incomplete command" =>
            {
                "MSG" => "The command specified is incomplete",
                "TYPE" => "fatal"
            },
        "invalid command" =>
            {
                "MSG" => "The command specified is invalid",
                "TYPE" => "fatal"
            },
        "invalid input" =>
            {
                "MSG" => "The input specified is invalid",
                "TYPE" => "fatal"
            },
        "duplicate member" =>
            {
                "MSG" => "The member specified already exists",
                "TYPE" => "fatal"
            },
        "vsan .* is not configured" =>
            {
                "MSG" => "The VSAN specified is not configured",
                "TYPE" => "fatal"
            },
        "one or more zones do not have members" =>
            {
                "MSG" => "One or more zones do not have members",
                "TYPE" => "fatal"
            },
        "name too long" =>
            {
                "MSG" => "Name too long",
                "TYPE" => "fatal"
            }
        };

# initialize so that we do not update prompts on every re-connection
$Prompts_updated = 0;

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new Switch object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($proto, %args) = @_;
    my $class = ref($proto) || $proto;
    my $self = {};
    my ($host, $connection, $conn_type, $passwd, $timeout, $reboot_timeout,
        $port_retries, $user, $persistent, $show_cmd, $execute, $debug_level,
        $log_timeout, $connection_retries, $admin_domain, $verify_on_enable,
        $fcoe_ethernet_block, $use_caching, $defined_vf_ids, $command_retries,
        $command_retry_delay, $validate_offline, $obj_template);

    # init variables
    bless $self, $class;
    $persistent           = 0;
    $debug_level          = 0;
    $timeout              = 180;
    $reboot_timeout       = 600;
    $log_timeout          = 300;
    $port_retries         = 5;
    $verify_on_enable     = 0;
    $fcoe_ethernet_block  = 0;
    $connection_retries   = 30;
    $admin_domain         = undef;
    $use_caching          = $SwitchUtil::FALSE;
    $defined_vf_ids       = undef;
    $command_retries      = 5;
    $command_retry_delay  = 30;
    $validate_offline     = 1;

    # Process arguments.  If an error occurs during the arg processing,
    # The error flag and errmsg values are set and the object is returned.
    # It is up to the caller to check for errors after the new call.
    foreach (keys %args) {
        if (/-?host$/) {
            length($host = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"host\":", "$args{$_}");
                return $self;
              };
        } elsif (/^-?connection$/) {
            length($connection = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"connection\":", "$args{$_}");
                return $self;
              };
        } elsif (/^-?conn_type$/) {
            length($conn_type = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"conn_type\":", "$args{$_}");
                return $self;
              };
        } elsif (/^-?user$/) {
            length($user = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"user\":", "$args{$_}");
                return $self;
              };
        } elsif (/^-?persistent$/) {
            length($persistent = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"persistent\":", "$args{$_}");
                return $self;
              };
        } elsif (/^-?passwd$/) {
            length($passwd = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"passwd\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?timeout$/) {
            length($timeout = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"timeout\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?reboot_timeout$/) {
            length($reboot_timeout = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"reboot_timeout\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?log_timeout$/) {
            length($log_timeout = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"log_timeout\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?port_retries$/) {
            length($port_retries = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"port_retries\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?verify_on_enable$/) {
            length($verify_on_enable = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"verify_on_enable\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?fcoe_ethernet_block$/) {
            length($fcoe_ethernet_block = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"fcoe_ethernet_block\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?connection_retries$/) {
            length($connection_retries = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"connection_retries\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?show_cmd$/) {
            length($show_cmd = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"show_cmd\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?execute$/) {
            length($execute = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"execute\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?debug_level$/) {
            length($debug_level = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"debug_level\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?command_retries$/) {
            length($command_retries = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"command_retries\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?validate_offline$/) {
            length($validate_offline = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"validate_offline\":", "$args{$_}");
                return $self;
            };
        } elsif (/-?command_retry_delay$/) {
            length($command_retry_delay = $args{$_})
              or do {
                $self->_set_err("SanSwitch::new invalid value for \"command_retry_delay\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?defined_vf_ids$/) {
            $defined_vf_ids = $args{$_}
        } elsif (/-?admin_domain$/) {
            $admin_domain = $args{$_};
        } elsif (/-?enable_caching$/) {
            $use_caching = $args{$_};
        } elsif (/^-?obj_template$/) {
            length($obj_template = $args{$_})
                or do {
                    $self->_set_err("Invalid value for \"obj_template\":", "$args{$_}");
                    return $self;
              };
        }
    }

    # if an object template was specified, copy over all the contents
    # and return
    if (defined($obj_template)) {
        return $self->_create_lite_sanswitch_object(-obj_template => $obj_template);
    }

    # Verify mandatory parameters.
    unless (defined $host and length $host) {
        $self->_set_err("SanSwitch::new mandatory parameter \"host\"", "not specified");
        return $self;
    }

    # Set the connection retry count
    $self->{connection_retries} = $connection_retries;

    # Set the validate offline option (checking to see if port goes offline after disable
    # operation on the switch)
    $self->{validate_offline} = $validate_offline;

    # Set the admin domain
    $self->{admin_domain} = $admin_domain;

    # Determine if a connection has been provided
    if (!defined($connection)) {
        # Connect to the host and store the connection obj
        $connection = $self->_connect($host, $user, $passwd,
                                      $timeout, $conn_type);
        if ($self->error) {
            # Error message is already set, return object so caller
            # can get the error.
            return $self;
        }
    }

    $self->{conn} = $connection;

    # determine the right platform and require/import/call the appropriate
    # vendor object's new call.
    my $hosttype = $self->hosttype();
    if ($hosttype =~ /fcp-switch-cisco/i) {
        require San::SanSwitch::FC::Switch::CiscoFCSwitch;
        $self = San::SanSwitch::FC::Switch::CiscoFCSwitch->_new(-conn => $self->get_connection());
    } elsif ($hosttype =~ /ethernet-switch-cisco|switch-ethernet-cisco|ethernet-switch-cisco-ios/i) {
        require San::SanSwitch::Ethernet::Switch::CiscoEthernetSwitch;
        $self = San::SanSwitch::Ethernet::Switch::CiscoEthernetSwitch->_new(-conn => $self->get_connection());
    } elsif ($hosttype =~ /switch-corduroyl/i) {
        require San::SanSwitch::Ethernet::Switch::CorduroyLEthernetSwitch;
        $self = San::SanSwitch::Ethernet::Switch::CorduroyLEthernetSwitch->_new(-conn => $self->get_connection());
    } elsif ($hosttype =~ /switch-brocade/i) {
        # distinquish between FOS Operating System based Brocade switches and
        # ones running the VDX/NOS Operating System. Completely different UIs.
        if ($hosttype =~ /switch-brocade-vdx/i) {
            require San::SanSwitch::FC::Switch::BrocadeVDXFCSwitch;
            $self = San::SanSwitch::FC::Switch::BrocadeVDXFCSwitch->_new(-conn => $self->get_connection());
        } else {
            require San::SanSwitch::FC::Switch::BrocadeFCSwitch;
            $self = San::SanSwitch::FC::Switch::BrocadeFCSwitch->_new(-conn => $self->get_connection());
        }
    } elsif ($hosttype =~ /switch-mcdata/i) {
        require San::SanSwitch::FC::Switch::McDataFCSwitch;
        $self = San::SanSwitch::FC::Switch::McDataFCSwitch->_new(-conn => $self->get_connection());
    } elsif ($hosttype =~ /switch-qlogic/i) {
        require San::SanSwitch::FC::Switch::QlogicFCSwitch;
        $self = San::SanSwitch::FC::Switch::QlogicFCSwitch->_new(-conn => $self->get_connection());
    } elsif ($hosttype =~ /switch-apcon/i) {
        require San::SanSwitch::FC::Switch::ApconFCSwitch;
        $self = San::SanSwitch::FC::Switch::ApconFCSwitch->_new(-conn => $self->get_connection());
    } else {
        $self->_set_err("Error: SanSwitch::new() Switch type $hosttype not", "currently supported by SanSwitch object");
        return $self;
    }

    @ISA = qw(Error SanLog San::SanSwitch SanCache);

    # Store connection parameters in object.
    $self->{host} = $host;
    $self->{user} = $user;
    $self->{pass} = $passwd;

    # save value of conn_type in host object
    $self->{conn_type} = $conn_type;

    # save the connection again since the re-self wipes these out
    $self->{conn}       = $connection;
    $self->{persistent} = $persistent;

    # Set if the switch commands need to be cached
    if ($use_caching == $SwitchUtil::TRUE) {
        $self->enable_sancache();
    } else {
        $self->disable_sancache();
    }

    # Show command mode.  By default when printing, the command
    # issued will be displayed.  This can be turned off by the user.
    $self->{show_cmd} = defined $show_cmd ? $show_cmd : 1;

    $self->{show_data} = 1;

    # Allow CLI commands to be formed, but not executed.
    $self->{execute} = defined $execute ? $execute : 1;

    # Set the timeout if one is given.
    $self->timeout($timeout) if (defined $timeout);

    # Set the reboot timeout
    $self->{reboot_timeout} = $reboot_timeout;

    # Set the log timeout (use for long logging commands)
    # for instance supportShow on Brocade
    $self->{log_timeout} = $log_timeout;

    # Set the port retry count
    $self->{port_retries} = $port_retries;

    # Set the validate offline option (checking to see if port goes offline after disable
    # operation on the switch)
    $self->{validate_offline} = $validate_offline;

    # Set the flag to tell port enable to check characteristics
    # versus state before the port was blocked.
    $self->{verify_on_enable} = $verify_on_enable;

    # Set the flag to tell port blocks for FCoE to
    # block at the Ethernet layer instead of the virtual
    # FC layer.
    $self->{fcoe_ethernet_block} = $fcoe_ethernet_block;

    # Set the connection retry count
    $self->{connection_retries} = $connection_retries;

    # Set the admin domain
    $self->{admin_domain} = $admin_domain;

    # Set the debug_level if one is given.
    $self->set_debug_level($debug_level);

    # Set the command retries and command retry delay
    $self->{command_retry_delay} = $command_retry_delay;
    $self->{command_retries} = $command_retries;

    # Get firmware version and switch model
    $self->{version} = $self->get_firmware_version();
    $self->{model}   = $self->get_model();

    # Get configuration information from the switch
    $self->{config}  = $self->get_config();

    # Get the domain id of the switch
    $self->{domain_id} = $self->_get_domain_id();

    # Get the npv mode of the switch
    if ($self->can("_get_npv_mode")) {
        $self->{npv_mode} = $self->_get_npv_mode();
    } else {
        $self->{npv_mode} = 0;
    }

    # if available get NPIV state of switch
    if ($self->can("_get_npiv")) {
        $self->{npiv} = $self->_get_npiv();
    } else {
        $self->{npiv} = 0;
    }

    # Get FCOE status of the switch
    if ($self->can("_get_fcoe")) {
        $self->{fcoe} = $self->_get_fcoe();
    } else {
        $self->{fcoe} = 0;
    }

    # store whether or not the switch supports admin domains
    if ($self->can("_get_ad_status")) {
        my $ad_status = $self->_get_ad_status();
        if (!defined($ad_status) or $ad_status == 0) {
            $self->{ad_status} = 0;
            $self->sanlog(-msg => "Admin Domain Support is disabled");
        } else {
            $self->{ad_status} = 1;
            $self->sanlog(-msg => "Admin Domain Support is enabled");
        }
    } else {
        $self->{ad_status} = 0;
        $self->sanlog(-msg => "Admin Domain Support is disabled");
    }

    # store if switch supports virtual fabrics.
    # Do not check error, since it could fail on
    # earlier Brocade FOS revisions
    if ($self->can("_get_vf_status")) {
        my $vf_status = $self->_get_vf_status();
        if (!defined($vf_status) or $vf_status =~ /not_supported/i) {
            $self->{vf_status} = 0;
            $self->sanlog(-msg => "Virtual Fabric support is disabled or not installed");
        } else {
            $self->{vf_status} = 1;
            $self->sanlog(-msg => "Virtual Fabric support is enabled");
        }

        # get virtual fabric IDs are store in object
        if ($self->{vf_status} == 1) {
            my @vf_ids = $self->get_vf_ids();
            if ($self->error()) {
                $self->_set_err("Error: SanSwitch::new() " .
                                "Problem getting virtual fabric IDs (FIDS) for switch ($host)\n" .
                                "that is enabled for Virtual Fabric support");
                return $self;
            }

            $self->sanlog(-msg => "Virtual Fabric IDs (FIDS): @vf_ids");

            if (defined $defined_vf_ids) {
                @vf_ids = split (",", $defined_vf_ids);
                $self->sanlog(-msg => "Overriding Virtual Fabric IDs (FIDS)" .
                                      " with predefined values: @vf_ids");
            }
            $self->{vf_ids} = [ @vf_ids ];
        }
    } else {
        # for platforms where "_get_vf_status" does not exist,
        # assume virtual fabric like constructs are not supported (yet)
        $self->{vf_status} = 0;
        $self->{vf_ids} = [];
    }

    return $self;
}

###############################################################################
# Method:    options
# Objective: set user definable options for switch object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub options {
    my ($self, %args) = @_;

    foreach (keys %args) {
        if (/^-?(.+)$/) {
            if (not length($self->{$1} = $args{$_})) {
                $self->_set_err("Error: SanSwitch::options", "Invalid value for argument $1");
                return $SwitchUtil::FALSE;
            }
        }
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    execute
# Objective: Issue arbitrary command to switch and get results.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub execute {
    my ($self, $cmd, $timeout, %args) = @_;

    my ($status, @output, $errmsg, $t0, $t1, $errors_template, $quiet, $use_connectrec);

    # Clear errors.
    $self->_clear_err;

    unless (length $cmd) {
        $self->_set_err("Error: SanSwitch::execute() invalid CLI command");
        return (wantarray()) ? () : [];
    }

    # Default $quiet to 0
    $quiet = 0;
    $use_connectrec = 0;

    # build a params hash for a timeout if the user specified one
    my %timeout_args = ();
    if (defined $timeout) {
        $timeout_args{timeout} = $timeout;
    }

    # Check the %args hash
    foreach (keys %args) {
        if (/-?errors_template/) {
            # Check to make sure that we were actually in hash reference
            if (ref($args{$_}) =~ /hash/i) {
                $errors_template = delete($args{$_});
            } else {
                $self->_set_err("Invalid value for argument '$_'");
                return (wantarray()) ? () : [];
            }
        } elsif (/^-?quiet$/) {
            if (not length($quiet = delete($args{$_}))) {
                $self->_set_err("Invalid value for argument '$_'");
                return (wantarray()) ? () : [];
            }

            # Make sure $quiet is 1 or 0
            unless ($quiet == 1 or $quiet == 0) {
                $self->_set_err("Invalid value '$quiet' for argument '$_'");
                return (wantarray()) ? () : [];
            }
        } elsif (/^-?use_connectrec$/) {
            if (not length($use_connectrec = delete($args{$_}))) {
                $self->_set_err("Invalid value for argument '$_'");
                return (wantarray()) ? () : [];
            }
        }
    }

    # If we have $errors_template defined, make sure it has something in it.
    if (defined($errors_template) and scalar(keys %$errors_template) == 0) {
        $self->_set_err("Invalid format for errors template hash");
        return (wantarray()) ? () : [];
    }

    # Reconnect to switch if we are not connected
    if (not defined $self->{conn}) {
        $self->reconnect();
    }

    if ($self->{show_cmd}) {
        $self->sanlog(-msg => $cmd);
    }

    return (wantarray()) ? () : [] unless ($self->{execute});

    # Run the execute routine inside a loop for $self->{command_retries}
    # iterations if the command fails for a retryable reason.  If it fails
    # for a non-retryable reason or the retry count is exhausted, set an
    # error and bail.
    my $retry_count = 1;
    my $errors_found = $SwitchUtil::FALSE;
    while ($retry_count <= $self->{command_retries}) {
        # Init execution timer
        $t0 = &gettimeofday();

        if ($use_connectrec) {
            ($status, @output) = &SwitchUtil::cr_execute($self->{conn},
                                                     command => $cmd,
                                                     %timeout_args,
                                                     %args);
        } else {
            ($status, @output) = &SwitchUtil::execute($self->{conn},$cmd,
                                                     %timeout_args,
                                                     -quiet   => $quiet);
        }

        # Validate the output for fatal errors
        if (scalar(@output) > 0) {
            my ($error_type, $error_msg);
            if (defined($errors_template)) {
                ($error_type, $error_msg) = $self->_validate_output(-output => \@output,
                                                                    -errors_template => $errors_template);
            } else {
                ($error_type, $error_msg) = $self->_validate_output(-output => \@output);
            }

            if ($self->error()) {
                $self->_set_err("Unable to validate output for command '$cmd':\n" .
                                $self->errmsg());
                return (wantarray()) ? () : [];
            }

            if (defined($error_type)) {
                # We had an error type returned
                if ($error_type eq "fatal" and defined($error_msg) and $error_msg ne "") {
                    # Fatal Errors were found
                    $self->_set_err("Unable to run command ($cmd) on the switch:\n" . $error_msg);
                    $errors_found = $SwitchUtil::TRUE;
                    last;
                } elsif ($error_type eq "retryable" and defined($error_msg) and $error_msg ne "") {
                    # We found a retryable error so increment $retry_count and sleep
                    # for $self->{command_retry_delay};
                    $self->sanlog(-msg => "Attempt $retry_count of $self->{command_retries} -- " . $error_msg);
                    $retry_count++;
                    sleep ($self->{command_retry_delay});
                    next;
                } elsif ($error_type eq "none") {
                    # We didn't find any errors so break out of the while loop.
                    $errors_found = $SwitchUtil::FALSE;
                    last;
                } else {
                    # _validate_output returned an invalid type
                    $self->_set_err("Invalid error type or error message returned from _validate_output");
                    $errors_found = $SwitchUtil::TRUE;
                    last;
                }
            }
        } else {
            # We didn't have any output so break out of the retry loop
            $errors_found = $SwitchUtil::FALSE;
            last;
        }
    }

    # Get the post execution time and save the command time
    $t1 = &gettimeofday();
    $self->{command_time} = $t1 - $t0;

    # Disconnect from switch
    if (!$self->{persistent}) {
        $self->disconnect();
    }

    # Check for standard set of errors returned from the CLI.
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: SanSwitch::execute :", @output);
        return (wantarray()) ? @output : \@output;
    }

    # Check for errors from the validation routine and retry count
    # expiration.
    if ($errors_found == $SwitchUtil::TRUE) {
        if (not $self->error) {
            $self->_set_err("There was an error while running command ($cmd)");
        }
    } elsif ($retry_count > $self->{command_retries}) {
        # We exhaused all retries
        if ($self->error) {
            $self->_set_err("Unable to run command on the switch after retrying $self->{command_retries} time\(s\):\n" .
                            $self->errmsg());
        } else {
            $self->_set_err("Unable to run command on the switch after retrying $self->{command_retries} time\(s\)");
        }
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_connection()
# Objective: Get nate connection object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_connection {
    my ($self) = @_;
    return $self->{conn};
}

###############################################################################
# Method:    fw_version()
# Objective: Get firmware version of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fw_version {
    my ($self) = @_;
    return $self->{version};
}

###############################################################################
# Method:    model()
# Objective: Get model of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub model {
    my ($self) = @_;
    return $self->{model};
}

###############################################################################
# Method:    config()
# Objective: Get configuration information from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config {
    my ($self) = @_;

    return $self->{config};
}

###############################################################################
# Method:    domain_id()
# Objective: Get domain id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub domain_id {
    my ($self) = @_;
    return $self->{domain_id};
}

###############################################################################
# Method:    get_exec_time()
# Objective: Get execution time of last command statement
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_exec_time {
    my ($self) = @_;

    if (length $self->{command_time}) {
        return sprintf "%.4f", $self->{command_time};
    }

    return "";
}

###############################################################################
# Method:    hostname()
# Objective: Get switch host name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub hostname {
    my ($self) = @_;
    return $self->{conn}{hostp}{hostname};
}

###############################################################################
# Method:    hosttype()
# Objective: Get switch host type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub hosttype {
    my ($self) = @_;
    return $self->{conn}{hostp}{hosttype};
}

###############################################################################
# Method:    timeout
# Objective: set or get telnet timeout
# Details:   See POD documentation at the end of this file.
###############################################################################
sub timeout {
    my ($self, $timeout) = @_;

    if (defined $timeout and length $timeout) {
        $self->{timeout} = $timeout;
    } else {
        return $self->{timeout};
    }
}

###############################################################################
# Method:    is_alive
# Objective: Check if the switch is responding on the network
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_alive {
    my ($self) = @_;

    my $hostname = $self->hostname();
    system("ping -c 1 $hostname");

    my $rc = $? >> 8;
    if ($rc == 0) {
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    create_portmgr()
# Objective: Create and pass back a portmgr handle
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_portmgr {
    my ($self, %args) = @_;
    my ($pObj);

    $self->_clear_err();

    $pObj = PortMgr->_new(-parent => $self,
                          %args);

    if ($pObj->error) {
        $self->_set_err($pObj->errmsg());
        return $self;
    }

    # overwrite any portmgr that was created in the past
    $self->{obj}{port} = $pObj;

    # Port object created successfully..returning to THPL
    return $pObj;
}

###############################################################################
# Method:    portmgr()
# Objective: Access the portmgr
# Details:   See POD documentation at the end of this file.
###############################################################################
sub portmgr {
    my ($self) = @_;

    return $self->{obj}{port};
}

###############################################################################
# Method:    create_zonemgr()
# Objective: Create and pass back a zonemgr handle
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_zonemgr {
    my ($self, %args) = @_;

    my ($zObj);

    $self->_clear_err();

    # Check to see if -parent was passed in. If it was, we will just pass
    # it through.  If not, we need to include it when calling _new.
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/) {
            $self->_set_err("Invalid value for argument '-parent'");
            return {};
        }
        # -parent is defined so just pass %args through
        $zObj = ZoneMgr->new(%args);
    } else {
        # -parent is not defined so just pass it to _new in addition to %args
        $zObj = ZoneMgr->new(-parent => $self, %args);
    }

    if ($zObj->error) {
        $self->_set_err($zObj->errmsg());
        return $zObj;
    }

    # Keep pointer to first zonemgr object created
    $self->{obj}{zone} = $zObj if (not defined $self->{obj}{zone});

    # Zone object created successfully.
    return $zObj;
}

###############################################################################
# Method:    is_ad_enabled()
# Objective: Determine if Admin Domains are enabled
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_ad_enabled {
    my ($self) = @_;

    return $self->{ad_status};
}

###############################################################################
# Method:    is_vf_enabled()
# Objective: Determine if Virtual fabrics are enabled
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_vf_enabled {
    my ($self) = @_;

    return $self->{vf_status};
}

###############################################################################
# Method:    is_npv_enabled()
# Objective: Determine if NPV (Access Gateway) mode is enabled
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_npv_enabled {
    my ($self) = @_;

    return $self->{npv_mode};
}

###############################################################################
# Method:    is_npiv_enabled()
# Objective: Determine if NPIV feature is enabled
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_npiv_enabled {
    my ($self) = @_;

    return $self->{npiv};
}

###############################################################################
# Method:    reconnect()
# Objective: reconnect to switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect {
    my ($self, %args) = @_;

    # disconnect the existing connection by default
    my $disconnect_existing_conn = exists($args{-disconnect_existing_conn}) ?
                                   $args{-disconnect_existing_conn} : 1;
    if ($disconnect_existing_conn == 1) {
        $self->disconnect();
    }

    $self->_clear_err();
    $self->sanlog(-msg => "Reconnecting to switch '" . $self->hostname() . "'");
    $self->{conn} = $self->_connect($self->{host},
                                    $self->{user},
                                    $self->{pass},
                                    $self->{timeout},
                                    $self->{conn_type});

    if ($self->error) {
        my $tempmsg = $self->errmsg();
        $self->_set_err("SanSwitch::reconnect() connect failed: $tempmsg");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: disconnect
# Objective:       disconnect a switch object. Ignore warnings (refine)
#                  around this to avoid collisions with Tharn version
#                  (always called package qualified)
# Details:         See POD documentation at the end of this file
################################################################################
no warnings qw(redefine);
sub disconnect {
    my ($self) = @_;

    if (defined $self->{conn}) {
        $self->sanlog(-msg => "Disconnecting Switch object '" .
                              $self->hostname() . "'");
        $self->{conn}->disconnect();
    }

    # reset prompts_updated to make sure modified set of prompts
    # get added back
    $Prompts_updated = 0;

}
use warnings qw(redefine);

###############################################################################
# Method:    set_fcoe_ethernet_block()
# Objective: set the value of fcoe_ethernet_block on switch object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_fcoe_ethernet_block {
    my ($self, $value) = @_;

    $self->{fcoe_ethernet_block} = $value;

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    set_admin_domain()
# Objective: set the value of admin_domain on switch object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_admin_domain {
    my ($self, $value) = @_;

    $self->{admin_domain} = $value;

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    get_admin_domain_database()
# Objective: get the admin domain information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_admin_domain_database {
    my ($self) = @_;

    $self->_clear_err();

    my $ad_hash;

    if ($self->can("_get_admin_domain_database")) {
        $ad_hash = $self->_get_admin_domain_database();
        if ($self->error()) {
            return {};
        }
    } else {
        $self->_set_err("The current switch type does not support this method");
        return {};
    }

    return $ad_hash;
}

###############################################################################
# Method:    get_current_admin_domain()
# Objective: get the current admin domain the switch object is sitting in
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_admin_domain {
    my ($self) = @_;

    $self->_clear_err();

    my $current_ad;
    if ($self->can("_get_current_admin_domain")) {
        $current_ad = $self->_get_current_admin_domain();
        if ($self->error()) {
            return "";
        }
    } else {
        $self->_set_err("The current switch type does not support this method");
        return "";
    }

    return $current_ad;
}

###############################################################################
# Method:    switch_to_admin_domain()
# Objective: switch to the admin domain specified
# Details:   See POD documentation at the end of this file.
###############################################################################
sub switch_to_admin_domain {
    my ($self, $admin_domain) = @_;

    $self->_clear_err();

    if ($self->can("_switch_to_admin_domain")) {
        $self->_switch_to_admin_domain($admin_domain);
        if ($self->error()) {
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("The current switch type does not support this method");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    display_info()
# Objective: Display fo diag purposes information on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub display_info {
    my ($self) = @_;

    my @output = $self->info();
    my $msg = "Switch Information:\n" . join("\n", @output);

    $self->sanlog(-msg => $msg);

}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _connect
# Objective: Create telnet connection to host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _connect {
    my ($self, $hostname, $user, $passwd, $timeout, $conn_type) = @_;
    my ($retries, $connect, $conn_str, $errmsg, $host, $hostInfo, $id);

    $host = Utils::make_host($hostname);
    my $max_retries = $self->{connection_retries};

    if ($self->{persistent} == 1) {
        $id = $SanSwitch::_CONN_CNT++;
    }

    if ($conn_type eq "THARNHOST") {
        $conn_type = "";

        # Check to see if any connection types are defined,
        # if not return the default value
        if (!defined($host->{connection_fields})) {

            # Set the $conn_type to the default of telnet
            $conn_type = "telnet";
        }

        # Go throught each remainder field and set $conn_type
        # IF rsh is defined, set that. This means that if the
        # conn type of rsh is defined in the tharnhost file,
        # we set $conn_type to that. To not use rsh, comment
        # out the conn type of rsh from the tharnhost file.
        # This will then check to see if the next available
        # conn type is defined in the thatnhost file.
        # The order of connection type is:
        # 1) rsh
        # 2) ssh
        # 3) telnet
        # 4) rlogin
        # 5) console
        elsif (defined($host->{connection_fields}->{rsh})) {
            $conn_type = "rsh";
        } elsif (defined($host->{connection_fields}->{ssh})) {
            $conn_type = "ssh";
        } elsif (defined($host->{connection_fields}->{telnet})) {
            $conn_type = "telnet";
        } elsif (defined($host->{connection_fields}->{rlogin})) {
            $conn_type = "rlogin";
        } elsif (defined($host->{connection_fields}->{console})) {
            $conn_type = "console";
        }
    } elsif ($conn_type eq "") {
        $conn_type = "telnet";
    } else {
        # Default $conn_type to the specified value user passed in
    }

    unless (defined($timeout) and length($timeout)) {
        $self->timeout($host->{connection_fields}{$conn_type}{timeout});
    }

    if (defined($user) and length($user)) {
        $host->{connection_fields}{$conn_type}{username} = $user;
    }

    if (defined($passwd) and length($passwd)) {
        $host->{connection_fields}{$conn_type}{password} = $passwd;
    }

    if ($self->{persistent} == 1 and $conn_type eq "ssh") {
        # Reset the Prompts_updated var so the prompts get rebuilt.
        $Prompts_updated = 0;

        $conn_str = $hostname . "#" . $conn_type . "_" . $id . ";"
            . "conntype=libssh-persist;"
            . "username=" . $host->{connection_fields}{$conn_type}{username} . ";"
            . "password=" . "\"" . $host->{connection_fields}{$conn_type}{password} . "\"" . ";"
            . "timeout=" . $self->{timeout}
            . ";inherit=0;";
    } elsif ($self->{persistent} == 1 and $conn_type eq "telnet") {
        # Reset the Prompts_updated var so the prompts get rebuilt.
        $Prompts_updated = 0;

        $conn_str = $hostname . "#" . $conn_type . "_" . $id . ";"
            . "conntype=telnet;"
            . "username=" . $host->{connection_fields}{$conn_type}{username} . ";"
            . "password=" . "\"" . $host->{connection_fields}{$conn_type}{password} . "\"" . ";"
            . "timeout=" . $self->{timeout}
            . ";inherit=0;";
    } else {
        $conn_str = $hostname . "#" . $conn_type . ";"
            . "username=" . $host->{connection_fields}{$conn_type}{username} . ";"
            . "password=" . "\"" . $host->{connection_fields}{$conn_type}{password} . "\"" . ";"
            . "timeout=" . $self->{timeout}
            . ";inherit=0;";
    }

    $self->sanlog(-msg_level => 3,
                  -msg => "Connection string is $conn_str");
    $retries = 0;
    while (!defined($connect) and $retries < $max_retries) {
        $connect = Tharn::connect(
            $conn_str,
            -eval => sub {
                if (@_) {
                    $retries++;
                    $connect = undef;

                    $errmsg = join("", @_);
                    $errmsg =~ s/\n//g;

                    # Display an error in case user doesn't include error
                    # handler on object instantiation.
                    $self->sanlog(-msg => "Warning: SanSwitch::_connect() " .
                                          "unable to connect to switch '" .
                                          $hostname  . "': $errmsg\n" .
                                          "Attempt $retries of $max_retries");

                    sleep(5);
                }
            });
    }

    if ($retries == $max_retries) {
         $self->_set_err($errmsg);
         Tharn::logresult("CONFIG", abort => 0, "Connect retries exceeded ($max_retries)");
         return $self;
    }

    # take into consideration Cisco UCS switch and need to connect to NXOS object
    if ($connect->{hostp}{hosttype} =~ /fcp-switch-cisco-ucs/) {
        # Connect to NXOS
        &SwitchUtil::execute($connect, "connect nxos");
    }

    my (@prompts, @new_prompts);
    if ($connect->{hostp}{hosttype} =~ /cisco/) {
        # add additional prompt for Cisco (conf) mode
        if ($Prompts_updated == 0) {
            my @prompts = $connect->prompts();
            foreach my $prompt (@prompts) {
                my $new_prompt = $prompt;
                $new_prompt =~ s/\#/\(conf\.\*\)\#/;
                push(@new_prompts, $prompt);
                push(@new_prompts, $new_prompt);
            }
            $connect->prompts(@new_prompts);
            $Prompts_updated = 1;
        }

        # Disable paging (fc or ethernet)
        &SwitchUtil::execute($connect, "terminal length 0");

        # if this is an ethernet switch run enable to
        # allow for privledged commands
        my $cisco_os;
        my $is_7k = 0;
        if ($connect->{hostp}{hosttype} =~ /ethernet/i) {
            my @output = &SwitchUtil::execute($connect, "show version");
            if (grep(/Cisco Nexus7000/i, @output) > 0) {
                $cisco_os = "NXOS";
                $is_7k = 1;
            } elsif (grep(/Cisco Nexus/i, @output) > 0) {
                $cisco_os = "NXOS";
            } elsif (grep(/Cisco IOS|Cisco Internetwork Operating System Software/i, @output) > 0) {
                $cisco_os = "IOS";
            } else {
                $self->_set_err("SanSwitch::_connect() switch OS is not NX-OS or IOS and is not supported");
                return $self;
            }

            # If this is IOS, get into enable mode for
            # advanced commands
            if ($cisco_os =~ /IOS/i) {
                &Tharn::say($connect, "enable",
                            '-prompt', "Password:",
                            '-timeout', 30);
                &SwitchUtil::execute($connect, $connect->password());
            }
        }

        # Disable auto-logout
        if ($connect->{hostp}{hosttype} !~ /ethernet/i) {
            # FC switch, this is easy, single command
            &SwitchUtil::execute($connect, "terminal session-timeout 0");
        } else {
            # Ethernet switch requires multiple commands in conf mode
            if ($cisco_os =~ /IOS/i) {
                &SwitchUtil::execute($connect, "conf terminal");
                &SwitchUtil::execute($connect, "no mac-address-table aging-time 0");
                &SwitchUtil::execute($connect, "line vty 0 15");
                &SwitchUtil::execute($connect, "exec-timeout 0 0");
                &SwitchUtil::execute($connect, "end");
            } elsif ($cisco_os =~ /NXOS/i) {
                &SwitchUtil::execute($connect, "conf terminal");
                if ($is_7k == 1) {
                    &SwitchUtil::execute($connect, "no mac address-table aging-time 0");
                } else {
                    &SwitchUtil::execute($connect, "no mac-address-table aging-time 0");
                }
                &SwitchUtil::execute($connect, "line vty");
                &SwitchUtil::execute($connect, "exec-timeout 0");
                &SwitchUtil::execute($connect, "end");
            }
        }
    } elsif ($connect->{hostp}{hosttype} =~ /corduroy/i) {
        # We need to enter enable mode for the corduroys
        # as well
        if ($connect->{hostp}{hosttype} =~ /corduroy/i) {
            &Tharn::say($connect, "enable",
                        '-prompt', "Password:",
                        '-timeout', 30);
            &SwitchUtil::execute($connect, $connect->password());
        }

        # Disable paging
        &SwitchUtil::execute($connect, "terminal length 0");

        # Increase telnet session timeout
        &SwitchUtil::execute($connect, "telnetcon timeout 160");
        &SwitchUtil::execute($connect, "configure");
        &SwitchUtil::execute($connect, "line console");
        &SwitchUtil::execute($connect, "session-timeout 160");
        &SwitchUtil::execute($connect, "exit");
        &SwitchUtil::execute($connect, "exit");
    }

    # Cancel admin mode for QLogic
    if ($connect->{hostp}{hosttype} =~ /qlogic/i) {
        &Tharn::say($connect, "admin cancel",
                    '-prompt', "Please confirm",
                    '-timeout', 30);
        &Tharn::say($connect, "y",
                    '-noprompt', '-prompt', "y",
                    '-timeout', 30);
        # sleep to allow cancel to be processed and send page break command
        &SanUtil::nate_sleep(5);
        &SwitchUtil::execute($connect, "set pagebreak off");
    }

    if ($connect->{hostp}{hosttype} =~ /brocade/) {
        # Disable auto-logout
        if ($connect->{hostp}{hosttype} =~ /vdx/i) {
            &SwitchUtil::execute($connect, "conf terminal");
            &SwitchUtil::execute($connect, "line vty");
            &SwitchUtil::execute($connect, "exec-timeout 0");
            &SwitchUtil::execute($connect, "end");
        } else {
            &SwitchUtil::execute($connect, "timeout 0");
        }

        # Disable paging
        if ($connect->{hostp}{hosttype} =~ /vdx/i) {
            &SwitchUtil::execute($connect, "terminal length 0");
        }

        # take into consideration admin domains and
        # virtual farbics in terms of prompts.
        if ($Prompts_updated == 0 and $connect->{hostp}{hosttype} !~ /vdx/i) {
            @prompts = $connect->prompts();
            foreach my $prompt (@prompts) {
                # only decompose prompts with ":", push in prompt
                if ($prompt !~ /:/ or $prompt =~ /\:\\S\+\:/) {
                    push(@new_prompts, $prompt);
                    next;
                }

                my @fields = split(/:/, $prompt);

                # new prompt for virtual fabrics and Admin domains
                my $new_prompt = $fields[0] . ":\\S+:" . $fields[1];

                # new prompt for Ethernet port block for FCoE and config mode
                # Note, 2 additional prompts, one for standard (derived from
                # existing prompt and another that uses hostname (which could
                # dangerous if the switch prompts have been changed).
                my $new_prompt_config = $fields[0] . "\.\*#";

                # always push in original prompt
                push(@new_prompts, $prompt);

                # push into new prompts only if not already in prompts list
                if (grep(/^$new_prompt$/, @prompts) == 0) {
                    push(@new_prompts, $new_prompt);
                }

                # push into new prompts only if not already in prompts list
                if (grep(/^$new_prompt_config$/, @prompts) == 0) {
                    push(@new_prompts, $new_prompt_config);
                }

                my $hn = $self->hostname();
                if ($hn ne "" and defined ($hn)) {
                    my $new_prompt_hardcode = $self->hostname() . "\.\*#";
                    if (grep(/^$new_prompt_hardcode$/, @prompts) == 0) {
                        push(@new_prompts, $new_prompt_hardcode);
                    }
                }

            }

            if (scalar(@new_prompts) > 0) {
                $connect->prompts(@new_prompts);
            }
        }

        # Select the admin domain if one is defined and greater than 0
        if (defined($self->{admin_domain}) and $self->{admin_domain} > 0) {
            my $admin_domain = $self->{admin_domain};
            &SwitchUtil::execute($connect, "ad --select $admin_domain");

            # Run the ad --show to make sure we made it in. The output should
            # look like the following:
            #
            # brcd200e-shu01:AD50:admin> ad --show
            # Current AD Number: 50  AD Name: AD50
            #
            # Effective configuration:
            # ------------------------
            #
            my ($status, @output) = &SwitchUtil::execute($connect, "ad --show", -quiet => 1);
            unless (grep(/Current AD Number: $admin_domain /, @output) > 0) {
                $self->_set_err("Unable to change to admin domain '$admin_domain'");
                return $connect;
            }
        }

        # check/set FIP Keep Alive (FCoE Specific and FOS specific)
        if ($connect->{hostp}{hosttype} !~ /vdx/i) {
            $self->sanlog(-msg => "Checking FIP KeepAlive");
            my $cmd = "fcoe --fipcfgshow";
            my ($status, @output) = &SwitchUtil::execute($connect, $cmd);
            if (grep(/FIP\s+Unsolicited\s+Advertisement\s+Interval\s+=\s+(\d+)/i, @output) > 0) {
                my $current_fip = (grep(/FIP\s+Unsolicited\s+Advertisement\s+Interval/i, @output))[0];
                $current_fip =~ s/FIP\s+Unsolicited\s+Advertisement\s+Interval\s+=\s+//i;
                if ($current_fip != 8) {
                    $self->sanlog(-msg => "FIP KeepAlive was not 8, setting value to 8");
                    $cmd = "fcoe --fipcfg -advintvl 8";
                    my ($status, @output) = &SwitchUtil::execute($connect, $cmd);
                    if (grep(/is\s+set\s+to\s+=\s+8/i, @output) == 0) {
                        $self->_set_err("Unable to change to set FIP KeepAlive to 8");
                        return $connect;
                    }
                }
            }
        }
    }

    return $connect;
}

###############################################################################
# Method:    _validate_output
# Objective: Validate the output from the execution of a command on the switch
#            to make sure we didn't have any common errors.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _validate_output {
    my ($self, %args) = @_;

    my ($output_ref, $errors_ref);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?output/) {
            # Check to make sure that we were actually in an array reference
            if (ref($args{$_}) =~ /array/i) {
                $output_ref = $args{$_};
            } else {
                $self->_set_err("Invalid value for argument '$_'");
                return ("", "");
            }
        } elsif (/-?errors_template/) {
            # Check to make sure that we were actually in hash reference
            if (ref($args{$_}) =~ /hash/i) {
                $errors_ref = $args{$_};
            } else {
                $self->_set_err("Invalid value for argument '$_'");
                return ("", "");
            }
        }
    }

    # Check to make sure we have the required arguments
    if (defined($output_ref)) {
        # Check to see if there are any messages in @$output_ref.
        # This may or may not be an actual problem since it is
        # entirely possible that a command created no output and thus
        # @$output_ref would be empty.  Don't set an error...just return undef
        if (scalar(@$output_ref) == 0) {
            return ("", "");
        }
    } else {
        $self->_set_err("Missing required parameter '-output'");
        return ("", "");
    }

    # Check to see if $errors_ref is defined.  If not, set it to the global default
    # of $Validation_Errors
    unless (defined($errors_ref) and ref($errors_ref) =~ /hash/i) {
        $errors_ref = $Validation_Errors;
    }

    # Iterate over each error in keys $errors_ref hash and check against @$output_ref array
    # reference.  If a match is found, then set the error to the value of the hash for
    # the key that matched the error
    # The error_ref hash is a multidimensional hash with the following structure:
    # $errors_ref->{search_string}->{MSG}->{return message}
    # $errors_ref->{search_string}->{TYPE}->{error type}
    foreach my $search_msg (sort keys %$errors_ref) {
        if (grep(/$search_msg/i, @$output_ref) > 0) {
            if (defined($errors_ref->{$search_msg}->{TYPE}) and defined($errors_ref->{$search_msg}->{MSG})) {
                return ($errors_ref->{$search_msg}->{TYPE}, $errors_ref->{$search_msg}->{MSG});
            } else {
                $self->_set_err("Unable to decode the Error Message hash");
                return ("","");
            }
        }
    }

    return ("none", "");
}

################################################################################
# Subroutine Name: _create_lite_sanswitch_object()
# Objective:       Given an object template, build a lightweight chatter-less
#                  SanSwitch object
################################################################################
sub _create_lite_sanswitch_object {
    my ($self, %args) = @_;

    my $obj_template = $args{-obj_template};

    # first figure out which constructor to call.
    my $base_class = ref($obj_template);

    # call the base class constructor in a light weight fashion
    if ($base_class =~ /FC::Switch::CiscoFCSwitch/i) {
        require San::SanSwitch::FC::Switch::CiscoFCSwitch;
    } elsif ($base_class =~ /Ethernet::Switch::CiscoEthernetSwitch/) {
        require San::SanSwitch::Ethernet::Switch::CiscoEthernetSwitch;
    } elsif ($base_class =~ /Ethernet::Switch::CorduroyLEthernetSwitch/) {
        require San::SanSwitch::Ethernet::Switch::CorduroyLEthernetSwitch;
    } elsif ($base_class =~ /FC::Switch::BrocadeVDXFCSwitch/) {
        require San::SanSwitch::FC::Switch::BrocadeVDXFCSwitch;
    } elsif ($base_class =~ /FC::Switch::BrocadeFCSwitch/) {
        require San::SanSwitch::FC::Switch::BrocadeFCSwitch;
    } elsif ($base_class =~ /FC::Switch::McDataFCSwitch/) {
        require San::SanSwitch::FC::Switch::McDataFCSwitch;
    } elsif ($base_class =~ /FC::Switch::QlogicFCSwitch/) {
        require San::SanSwitch::FC::Switch::QlogicFCSwitch;
    } elsif ($base_class =~ /FC::Switch::ApconFCSwitch/) {
        require San::SanSwitch::FC::Switch::ApconFCSwitch;
    } else {
        $self->_set_err("SanSwitch Lite constructor does not support the base class '$base_class'");
        return $self;
    }
    $self = $base_class->_new(-conn => $obj_template->{conn});

    # copy all the keys from the template
    foreach my $key (keys %{$obj_template}) {
        $self->{$key} = $obj_template->{$key};
    }

    return $self;
}

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

SanSwitch.pm

=head1 Author

Evan Chiu

=head1 Description

SanSwitch.pm provides an API interface to a switch through the
use of the Telnet library.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<San::SanSwitch::Switch::[I<CiscoFC|CiscoEthernet|BrocadeFC|McDataFC|QlogicFC>]>

B<Error.pm>

=head1 Attributes

=head2 B<show_cmd>

boolean = <1 | 0> (default: 1 (true) By default log commands executed)

=head2 B<show_data>

boolean = <1 | 0>

=head2 B<persistent>

boolean = <1 | 0> (default: 0 (false) By default connection is not connected)

=head2 B<execute>

boolean = <1 | 0> (default: 1)

=head2 B<log>

file handle = File handle used for logging

=head2 B<connect>

connection object = Connection to the host (as returned from Tharn::connect())

=head2 B<conn>

connection object = Connection to the host

=head2 B<host>

string = Host name, connection parameter

=head2 B<user>

string = Username, connection parameter

=head2 B<pass>

string = Password, connection parameter

=head2 B<obj>

hash = Hash of children objects (port)

=head2 B<command_time>

integer = time in seconds of the last command executed

=head2 B<timeout>

integer = timeout to use for say calls

=head2 B<debug_level>

integer = Current debug level

=head1 Public Methods

=head2 B<new(%args)>

Creates a new SanSwitch object.  Uses the specified IP address or machine name
to create a telnet connection. If the connection is successful, the calling
program will be logged into the CLIENT using either the default user/password
specified in the NATE_HOSTS file or a user and password supplied by the caller.

=over 4

=item Inputs

=item * -host

Machine name or IP address of the remote machine.

=item * I<-user> (default: Tharnhost username entry).

Override default username for host login.

=item * I<-passwd> (default: Tharnhost password entry).

Override default password for host login.

=item * I<-timeout> (default: Tharnhost telnet timeout entry).

Telnet timeout in seconds

=item * I<-reboot_timeout> (default: 600)

timeout in seconds for time to reboot switch

=item * I<-log_timeout> (default: 300)

timeout in seconds for time to allow for long logging
commands such as supportShow from Brocade

=item * I<-port_retries> (default: 5)

Override the number of retries to try when getting port information.
A sleep of 3 second per retry is applied.

=item * I<-verify_on_enable> (default: 0)

tell the port_enable() method to verify it' state after the
the port is enabled versus what it was prior to the previous
port block (port_disable()).

=item * I<-fcoe_ethernet_block> (default: 0)

tells the port_enable() and port_disable() methods to
enable/disable ports at the Ethernet layer if working with FCoE ports.
0 = Virtual FC layer, 1 = Ethernet layer

=item * I<-debug_level> (default: 0).

Verbosity level for Sanlog::sanlog()

=item * I<-loghdl>

Filehandle to log output too.

Override the Telnet timeout value.

=item * I<-show_cmd> <1 | 0> (default: 1 (true) ).

By default commands executed will be logged   This can be
turned off by passing the -show_cmd with a value of 0.

=item * I<-connection>

If specified, a active/live connection must be passed in, the &connect function
must be called prior to calling the SanSwitch::new function

=item * I<-persistent>

If specified, the connection will either be maintained or reconnect for each command.

=item * I<-conn_type> <THARNHOST | rsh | telnet | ssh | rlogin | console> (NO DEFAULT).

This parameter will determine what connection type will be used for the client.
THIS CAN NOT BE USED IN CONJUNCTION WITH '-connection'. This parameter will NOT
override '-connection'.

=item * (<-admin_domain>)

If specified, all switch commands will be directed into the Admin Domain specified.

=item * I<-command_retries> (default: 5)

If specified, the switch command will be retried the number of times specified if
the command failed for a reason that is retryable.

=item * I<-command_retry_delay> (default: 30)

If specified, the switch command will be attempted every specified number of seconds until
the number of retries allowed is exhausted.

=item Outputs

Returns a SanSwitch object.  If a successful Telnet connection and login was
achieved, an open Telnet connection is maintained by the object until
the calling program exits or if the persistent option is specified, the connection
may be disconnected and reconnected only when a command is executed.

=item Examples

 # Create a Switch object with a connection to a remote switch.

 $Client = new SanSwitch(
        -host => '192.168.x.y',
        -user => 'smith',
        -passwd  => 'mypassword');

  if ($Client->error) {
     $Test->nlog($Client->errmsg());
     return $TCase::FAIL;
 }

=back

=head2 B<create_portmgr(%args)>

Creates a new PortMgr object

=over 4

=item Inputs

None

=item Outputs

Returns a PortMgr object.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Examples

 # Create a PortMgr object

  $port= $Client=>create_portmgr();
  if ($port->error) {
     $Test->nlog($port->errmsg());
     return $TCase::FAIL;
 }

=back

=head2 B<portmgr(%args)>

Access the portmgr object

=over 4

=item Inputs

None

=item Outputs

Returns the PortMgr object instantiated.

=item Examples

 # Access the PortMgr object
 $switch->portmgr()->some_portmgr_method();
 if ($port->error) {
     $Test->nlog($switch->portmgr()->errmsg());
     return $TCase::FAIL;
 }

=back

=head2 B<create_zonemgr(%args)>

Creates a new ZoneMgr object

=over 4

=item Inputs

None

=item Outputs

Returns a ZoneMgr object.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Examples

 # Create a ZoneMgr object

  $zonemgr= $Switch->create_zonemgr();
  if ($zonemgr->error) {
     $Test->nlog($zonemgr->errmsg());
     return $TCase::FAIL;
 }

=back

=head2 B<execute($cmd,I<%args>)>

Executes command

=over 4

=item Inputs

=item * $cmd

Command to execute on the switch.

=item * I<$hide> default: 0

Hide the command from STDOUT

=item Outputs

Returns an array or an array reference to the data returned by executing the command.
If an error occurred, the error flag is set, errmsg is set to the failure reason, and
an empty array reference is passed back as the return value.

=item Examples

 # Display command output from host
 $buf=$Client->execute("show version");
 if ($Client->error) {
     $Test->nlog($Client->errmsg());
     return $TCase::FAIL;
 }


 foreach $line (@$buf){
     print $line."\n";
 }

=back

=head2 B<get_connection()>

This method returns the nate connection object

=over 4

=item Outputs

Returns a connection object. If an error occured the error flag
is set, errmsg is set to the failure reason

=item Examples

 #Retrieve execution time
 $connObj = $obj->get_connection();

=back

=head2 B<get_exec_time()>

This method returns the execution time in seconds of the last snap command statement executed.  Its the
actual time it took to process the statement and does not include any post processing of the data. The
timer used is a high resolution timer with millisecond accuracy.

=over 4

=item Outputs

Returns a scalar value representing the execution time in seconds.

=item Examples

 # Retrieve execution time

 $time = $obj->get_exec_time();
 print "Execution Time = $time seconds\n";

=back

=head2 B<hostname()>

This method allows you to retrieve the hostname of the switch machine

=over 4

=item Outputs

Returns a scalar value containing the hostname.  If an error occured the error flag
is set, errmsg is set to the failure reason, and 0\'s are passed back as the return values.

=item Examples

 $hostname= $Obj->hostname();
 if ($Obj->error){
     ...
 }

 print "The hostname is $hostname.\n";

=back

=head2 B<hosttype()>

This method allows you to retrieve the hosttype of the switch machine

=over 4

=item Outputs

Returns a scalar value containing the hosttype.  If an error occured the error flag
is set, errmsg is set to the failure reason, and 0\'s are passed back as the return values.

=item Examples

 # Get hosttype
 $hosttype= $Obj->hosttype();
 if ($Obj->error){
     ...
 }

 print "The hosttype is $hosttype.\n";

=back

=head2 B<fw_version()>

This method allows you to retrieve the fw_version  of the switch device

=over 4

=item Outputs

Returns a string value containing the fw_version.  If an error occured the error flag
is set, errmsg is set to the failure reason, and 'unknown' is passed back as the return value.

=item Examples

 # Get fw version
 $fw_ver= $Obj->fw_version();
 if ($Obj->error){
     ...
 }

 print "The version is $fw_ver.\n";

=back

=head2 B<timeout($sec)>

This method allows you to change the Telnet timeout.  The default value
for timeout is the value specified in the Tharnhost file.

=over 4

=item Inputs

=item * $sec

Specify number of seconds for timeout value.

=item Outputs

If successful, the timeout in the Telnet object is set to the specified
value and the previously set value is returned.  If an error occurred,
the error flag is set, errmsg is set to the failure reason returned
from the Telnet module and 0 is passed back as the return value.

=item Examples

 # Set Telnet timeout value to 45 seconds.
 $prev = $obj->timeout(45);
 <check for errors>

=back

=cut

1;
