# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for ports for a Cisco Ethernet switch
## @author thomasm@netapp.com
## @status private
## @pod ./Port.pod

# Package Name
##############
package San::SanSwitch::Ethernet::PortMgr::CiscoEthernetPort;

use strict;

# Package Requirements
######################
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::PortMgr;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog PortMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_port_num()
# Objective:       Retrieve the port number/interface corresponding to a 
#                  WWPN, Return the entire array
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_num {
    my ($self, %args) = @_;

    $self->_clear_err();

    my @ports = $self->get_port_nums(%args);

    # if the caller wants an array, give them the full list, otherwise give them
    # only the first element, burt505898, needed for c-mode 7-mode compatibility
    # TODO: remove once c-mode support is fully implemented
    return (wantarray()) ? @ports : $ports[0];
}

################################################################################
# Subroutine Name: get_port_nums()
# Objective:       Retrieve the all port number/interface corresponding to a 
#                  MAC. 
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_nums {
    my ($self, %args) = @_;

    my ($mac, @found_ports, $found_port);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwpn/) {
            if (not length($mac = $args{$_})) {
                $self->_set_err("Error: CiscoEthernetPort::get_port_nums() " .
                                "Invalid value for argument '-wwpn'");
                return ();
            }
            next;
        }
    }

    # look for the mac address on the switch
    if (defined($mac)) {
        $self->sanlog(-msg => "Searching for MAC $mac");
        $found_port = $self->_get_port_from_mac($mac);
        if (defined($found_port)) {
            $self->sanlog(-msg => "Found match on port $found_port");
            my @fps = split(/\;/, $found_port);
            push(@found_ports, @fps);
        } else {
            $self->_set_err("Error: CiscoEthernetPort::get_port_nums() MAC could not be found on this switch");
            return ();
        }
    } else {
        $self->_set_err("Error: CiscoEthernetPort::get_port_nums() MAC must be specified");
        return ();
    }

    return (@found_ports);
}

################################################################################
# Subroutine Name: port_disable()
# Objective:       Disable the port/interface specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_disable {
    my ($self, $port) = @_;
    my ($status);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    my @cmds = ("conf t", 
                "int $port", 
                "shutdown", 
                "end", 
                "show int $port");

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoEthernetPort::port_disable() ".
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # wait for the port to transition to offline, or report error
    $status = $self->waitfor_port($port, "offline");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: CiscoEthernetPort::port_disable() " .
                        "failed to disable the port, port still appears up");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_enable()
# Objective:       Enable the port specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_enable {
    my ($self, $port) = @_;
    my ($status, @output);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};
    my $count               = 0;

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    my @cmds = ("conf t", 
                "int $port", 
                "no shutdown", 
                "end");

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoEthernetPort::port_enable() " .
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # add explicit sleep into the code here before checking the                                             
    # port to see if it is online (suggested by Dave/Sunil)                                                 
    sleep (3);                                                                                              

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: CiscoEthernetPort::port_enable() " .
                        "failed to enable port, port still appears down");
        return $SwitchUtil::FALSE;
    }

    # compare values now (after enable) versus before the
    # port_disable
    $self->check_port_settings($port);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: set_port_speed()
# Objective:       Set the specified port to the specified speed
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_speed {
    my ($self, %args) = @_;
    my ($status, $port, $speed);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};
    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Error: CiscoEthernetPort::set_port_speed() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?speed/) {
            if ((not length($speed = $args{$_})) && 
                ($args{$_} !~ /^\d$/ or $speed !~ /auto/i)) {
                $self->_set_err("Error: CiscoEthernetPort::set_port_speed() " .
                                "Invalid value for argument '-speed'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # handle auto speed setting
    if ($speed =~ /AUTO/i) {
        $speed = "auto";
    }

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    my @cmds = ("conf t", 
                "int $port", 
                "speed $speed", 
                "end");

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoEthernetPort::set_port_speed() " .
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: CiscoEthernetPort::set_port_speed() " .
                        "Port state did not transtion to online state");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: get_port_info()
# Objective:       Get information about the specified port/interface
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_info {
    my ($self, $port) = @_;
    my ($available_speeds, @output);

    $self->_clear_err();

    # Initialize hash
    my $hash = {};
    $hash->{state}              = "unknown";
    $hash->{speed}              = 0;
    $hash->{media_type}         = "";
    $hash->{mac}                = 0;
    $hash->{min_speed}          = 0;
    $hash->{max_speed}          = 0;
    $hash->{hardware}           = "";
    $hash->{description}        = "";
    $hash->{mtu}                = 0;
    $hash->{is_trunk}           = 0;

    # sometimes Cisco uses Ethernet1/6 and sometimes Eth1/6.
    # This screws up parsing. Get list the port information
    my $stripped_port = $port;
    $stripped_port =~ s/^(\D+)//;

    @output = $self->{parent}->execute("show int $port");
    foreach my $line (@output) {
        if ($line =~ /\, (\d+)\s*(M|G)b\/s\,/) {
            my $speed = $1;
            if ($line =~ /Gb/i) {
                $speed = $speed * 1000;
            } 
            $hash->{speed} = int($speed);
        }
        if ($line =~ /MTU (\d+) bytes/) {
            $hash->{mtu} = int($1);
        }
        if ($line =~ /Hardware(\s+is|:)\s+(.*)\,/i) {
            $hash->{hardware} = $2;
            if ($hash->{hardware} =~ /Port-Channel/i) {
                $hash->{is_portchannel} = 1;
            } else {
                $hash->{is_portchannel} = 0;
            }
        }
        if ($line =~ /$stripped_port is (.*)(\,|)/) {
            my $state = $1;
            $state =~ s/,//;
            $hash->{state} = &SwitchUtil::normalize_state($state);
        }
        if ($line =~ /address is (\S+)/) {
            my $mac = $1;
            $mac =~ s/\://g;
            $mac =~ s/\.//g;
            $hash->{mac} = $mac;
        }
        if ($line =~ /Description: (.*)/) {
            $hash->{description} = $1;
        }
        if ($line =~ /media type is (\S+)/) {
            $hash->{media_type} = $1;
        }
        if ($line =~ /Port mode is (\S+)/) {
            my $is_trunk = $1;
            if ($is_trunk =~ /trunk/i) {
                $hash->{is_trunk} = 1;
            }
        }
    }

    # Load the cached values if we have already looked these values up before
    $hash->{min_speed} = $self->{cache}{min_speed}{$port} 
        if defined($self->{cache}{min_speed}{$port});
    $hash->{max_speed} = $self->{cache}{max_speed}{$port} 
        if defined($self->{cache}{max_speed}{$port});

    # If we have the default or invalid values, we either have not looked up the
    # capabilities of this port before or the previous look up failed.
    # Don't run if this is a trunk since the command will be invalid
    # Don't run if this is a port-channel the command will be invalid
    if (($hash->{min_speed} == 0 or $hash->{max_speed} == 0) and
         $hash->{is_trunk} == 0 and $hash->{is_portchannel} == 0) {
        @output = $self->{parent}->execute("show int $port capabilities");
        foreach my $line (@output) {
            if ($line =~ /Speed:\s+(\S+)/) {
                $available_speeds = $1;
            }
        }

        # speeds will show up as:
        # Speed:                 10,100,1000,auto
        my @values = split(/,/, $available_speeds);
        # take "auto" into consideration and remove from list
        @values = grep(!/auto/, @values);

        # min will be first in list
        $hash->{min_speed} = $values[0];
        # Store this value into memory for later reuse
        $self->{cache}{min_speed}{$port} = $hash->{min_speed};

        # max will be last in list
        $hash->{max_speed} = $values[(scalar(@values) - 1)];
        # Store this value into memory for later reuse
        $self->{cache}{max_speed}{$port} = $hash->{max_speed};
    }

    return $hash;
}

################################################################################
# Subroutine Name: set_port_type()
# Objective:       Set the specified port to the specified port type
#                  no-op for ethernet for now
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_type {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $SwitchUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new port object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: CiscoEthernetPort::_new() " .
                                "Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/i) {
            $self->_set_err("Error: CiscoEthernetPort::_new() " .
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: CiscoEthernetPort::_new() " .
                        "Missing required argument '-parent'");

        return $self;
    }

    # initialize the object members
    $self->{parent} = $parent;

    # for compatibility with FC
    $self->{nameserver} = {};

    # get list of uplink or ISL ports
    my @isl_ports = $self->_get_isl_ports();
    $self->{isl_ports} = \@isl_ports;

    return $self;
}

###############################################################################
# Method:    _get_isl_ports
# Objective: Retrieve list of ports that correspond to ISL ports to other
#            switches
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_isl_ports {
    my ($self) = @_;

    # csc4948-auto1#show cdp neighbors detail
    # -------------------------
    # Device ID: AS12B1L2D8.rtp.netapp.com
    # Entry address(es): 
    #   IP address: 10.60.252.85
    # Platform: cisco WS-C4948,  Capabilities: Switch IGMP 
    # Interface: GigabitEthernet1/48,  Port ID (outgoing port): GigabitEthernet1/40
    # Holdtime : 151 sec

    # Also, need to take into consideration that this output can contain connections
    # to controller ports:
    # Device ID: f3070-229-5
    # Entry address(es): 
    #   IP address: 192.168.10.24
    # Platform: FAS3070,  Capabilities: Host 
    # Interface: GigabitEthernet1/36,  Port ID (outgoing port): e0c
    # Holdtime : 143 sec
    # 
    # Version :
    # NetApp Release 7.3.4: Thu May 27 16:40:57 PDT 2010
    # 
    # So, need to look at Capabilities and only include Switch

    my $capabilities;
    my @isl_ports = ();
    my @output = $self->{parent}->execute("show cdp neighbors detail");
    foreach my $line (@output) {
        if ($line =~ /Capabilities:\s+(.*)/) {
            $capabilities = $1;
        }

        if ($line =~ /^Interface:\s+(\S+)\,/) {
            my $isl_port = $1;
            # only add port to ISL list if connected to is a switch
            if ($capabilities =~ /switch/i) {
                push(@isl_ports, $isl_port);
            }
        }
    }

    my $fw_version = $self->{parent}->get_firmware_version();
    if ($self->{parent}->error) {
        $self->_set_err("Could not get fw version of this switch");
        return ();
    }

    # Determine if there are any port channels built around these isl ports
    # different command set for the Nexus OS versions
    #
    #auto-cisco4948-01>show int summary
    #    *: interface is up
    #      IHQ: pkts in input hold queue     IQD: pkts dropped from input queue
    #          OHQ: pkts in output hold queue    OQD: pkts dropped from output queue
    #              RXBS: rx rate (bits/sec)          RXPS: rx rate (pkts/sec)
    #                  TXBS: tx rate (bits/sec)          TXPS: tx rate (pkts/sec)
    #                      TRTL: throttle count
    #
    #  Interface              IHQ   IQD  OHQ   OQD  RXBS RXPS  TXBS TXPS TRTL
    #------------------------------------------------------------------------
    #* Vlan1                    0     0    0     0     0    0     0    0    0
    #  Vlan2                    0     0    0     0     0    0     0    0    0
    #  Vlan3                    0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/1       0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/2       0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/8       0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/9       0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/10      0     0    0     0     0    0  1000    1    0
    #
    #  Interface              IHQ   IQD  OHQ   OQD  RXBS RXPS  TXBS TXPS TRTL
    #------------------------------------------------------------------------
    #  GigabitEthernet1/11      0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/31      0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/32      0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/33      0     0    0     0     0    0     0    0    0
    #
    #  Interface              IHQ   IQD  OHQ   OQD  RXBS RXPS  TXBS TXPS TRTL
    #------------------------------------------------------------------------
    #  GigabitEthernet1/34      0     0    0     0     0    0     0    0    0
    #  GigabitEthernet1/44      0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/45      0     0    0     0 11000    5  8000   10    0
    #* GigabitEthernet1/46      0     0    0     0  2000    1  2000    2    0
    #  GigabitEthernet1/47      0     0    0     0     0    0     0    0    0
    #* GigabitEthernet1/48      0     0    0     0  2000    3  1000    2    0
    #* Port-channel1            0     0    0     0 10000    6  7000    9    0

    #auto-cisco4948-01>show int port-channel 1
    #Port-channel1 is up, line protocol is up (connected)
    #  Hardware is EtherChannel, address is 001e.4afa.b3ed (bia 001e.4afa.b3ed)
    #    Description: ISL between auto-cisco4948-01 and auto-cisco4948-02
    #  MTU 9000 bytes, BW 2000000 Kbit, DLY 10 usec,
    #     reliability 255/255, txload 1/255, rxload 1/255
    #  Encapsulation ARPA, loopback not set
    #  Keepalive set (10 sec)
    #  Full-duplex, 1000Mb/s, media type is N/A
    #  Media-type configured as RJ45 connector
    #  input flow-control is off, output flow-control is unsupported
    #  Members in this channel: Gi1/45 Gi1/46
    #  ARP type: ARPA, ARP Timeout 04:00:00
    #  Last input 1d12h, output never, output hang never
    #  Last clearing of "show interface" counters never
    #  Input queue: 0/2000/0/0 (size/max/drops/flushes); Total output drops: 0
    #  Queueing strategy: fifo
    #  Output queue: 0/40 (size/max)
    #  5 minute input rate 11000 bits/sec, 6 packets/sec
    #  5 minute output rate 7000 bits/sec, 10 packets/sec
    #     4783830791 packets input, 891870080101 bytes, 0 no buffer
    #     Received 3294733 broadcasts (1471266 multicast)
    #     0 runts, 0 giants, 0 throttles
    #     0 input errors, 0 CRC, 0 frame, 0 overrun, 0 ignored
    #     0 input packets with dribble condition detected
    #     8569414472 packets output, 12352880699389 bytes, 0 underruns
    #     0 output errors, 0 collisions, 0 interface resets
    #     0 babbles, 0 late collision, 0 deferred
    #     0 lost carrier, 0 no carrier
    #     0 output buffer failures, 0 output buffers swapped out

    my $cmd;
    if ($fw_version =~ /NXOS/i) {
        $cmd = "show int brief";
    } else {
        $cmd = "show int summary";
    }
    @output = $self->{parent}->execute($cmd);

    foreach my $line (@output) {
        # if we found a port channel, go to it and determine if any of its
        # members are ISL connections
        if ($line =~ /(Po|Port\-channel)(\d+)/i) {
            my $port_channel_num = $2;

            my @pchan_members = $self->_get_port_channel_members($port_channel_num);
            foreach my $member (@pchan_members) {
                my $port;

                # normalize the name of this port channel member to the format
                # used in @isl_ports
                if ($member =~ /^(GigabitEthernet|Gi)(\d+\/\d+)$/i) {
                    $port = "GigabitEthernet$2";
                } elsif ($member =~ /^(Ethernet|Eth)(\d+\/\d+)$/i) {
                    $port = "Ethernet$2";
                } else {
                    # in this case we couldnt parse the port channel member,
                    # go to the next member in the list
                    next;
                }

                # if a member of this port channel is an ISL port, then add
                # this port channel to the ISL port list
                if (grep(/^\s*${port}\s*$/, @isl_ports)) {
                    # add this port channel to the isl port list and break out
                    push(@isl_ports, "Port-channel" . $port_channel_num);
                    last;
                }
            }
            next;
        }
    }

    return (@isl_ports);
}

###############################################################################
# Method:    _normalize_port_name
# Objective: Normalize the port name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _normalize_port_name {
    my ($self, %args) = @_;

    my ($port, $normalized_port);

    # parse arguments
    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Invalid value for argument '-port'");
                return "";
            }
            next;
        }
    }

    unless (defined($port) and $port ne "") {
        $self->_set_err("Missing required parameter '-port'");
        return "";
    }

    if ($port =~ /^Eth(\d+.*)/i) {
        $normalized_port = "Ethernet" . $1;
    } elsif ($port =~ /^Ethernet(\d+.*)/i) {
        $normalized_port = "Ethernet" . $1;
    } elsif ($port =~ /^GigabitEthernet(\d+.*)/i) {
        $normalized_port = "GigabitEthernet" . $1;
    } elsif ($port =~ /^Gi(\d+.*)/i) {
        $normalized_port = "GigabitEthernet" . $1;
    } elsif ($port =~ /^Port\-Channel(\d+.*)/i) {
        $normalized_port = "Port-Channel" . $1;
    } elsif ($port =~ /^PO(\d+.*)/i) {
        $normalized_port = "Port-Channel" . $1;
    } else {
        $normalized_port = $port;
    }

    return $normalized_port;
}

###############################################################################
# Method:    _get_nameserver_database
# Objective: Retrieve nameserver database information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_nameserver_database {
    # no-op for ethernet
    return {};
}

###############################################################################
# Method:    _get_port_from_mac
# Objective: Given a passed in MAC address, look to see if that MAC is located
#            on the switch. Ignore if the found port is an ISL since we do not
#            want to block the ISL port.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_port_from_mac {
    my ($self, $address) = @_;

    my ($found_line, @interfaces, $interface);
    my ($model, $cmd);

    # get MAC in proper format for Cisco
    $address =~ s/\://g;
    $address =~ s/\-//g;
    $address =~ s/\.//g;
    $address = lc($address);
    
    my $a = substr($address, 0, 4);
    my $b = substr($address, 4, 4);
    my $c = substr($address, 8, 4);
    $address = $a . "." . $b . "." . $c;

    # 4948 output:
    # > show mac-address-table address 000b.cd3f.3cf4        
    # Unicast Entries
    # vlan   mac address     type        protocols               port
    # -------+---------------+--------+---------------------+--------------------
    # 240    000b.cd3f.3cf4   dynamic ip,other              GigabitEthernet1/25   
    #
    # Nexus output:
    # > show mac-address-table address 0050.5673.05d0
    # Legend: 
    # * - primary entry, G - Gateway MAC, (R) - Routed MAC
    # age - seconds since last seen
    # VLAN     MAC Address      Type      age     Secure  NTFY   Ports
    # ---------+-----------------+--------+---------+------+------+----------------
    # * 30       0050.5673.05d0    dynamic   0        False  False Eth3/21

    $model = $self->{parent}->get_model();
    if ($self->{parent}->error) {
        $self->_set_err("Error: CiscoEthernetPort::_get_port_from_mac() ".
                        "error getting model of this switch");
        return undef;
    }

    # different command set for the Nexus 7000 switch
    if ($model =~ /Nexus7000/i) {
        $cmd = "show mac address-table address $address";
    } else {
        $cmd = "show mac-address-table address $address";
    }
    my @output = $self->{parent}->execute($cmd);
    my @matches = grep(/^\s*(\d+|\*\s+\d+)\s+/, @output);

    # no matches found.
    if (scalar(@matches) == 0) {
        return undef;
    }

    # can get multiple ports for same MAC in case of VIF support (w/Port Channel support 
    # on switch)
    if (scalar(@matches) > 1) {
        $self->sanlog(-msg => "NOTE: Found multiple ports that correspond to this MAC ($address), chould be a VIF");
    }

    foreach my $match (@matches) {
        # grab interface from the output line. Interface should be
        # in the format of GigabitEthernet1/25 or Eth3/21 (type/port_num) or
        # 10        02a0.9809.72c7    dynamic 0         Po20
        # or
        # 10        02a0.9809.72c7    dynamic 0         port-channel20
        # for VIF/Port channel configs
        my $line = $match;
        if ($line =~ m/.*?(\S+\/\S+\/\S+|\S+\/\S+|Po\d+|port\-channel\d+)/i) {
            $interface = $1;
        }
        if (not defined($interface) or $interface eq "") {
            $self->_set_err("Unable to parse port (or port-channel) from mac address show output");
            return undef;
        }

        # look to see if the MAC came from a port that is an ISL or uplink port
        # If it is then the MAC is connected on another switch and not this one.
        my $isl_found = 0;
        my $isl_port = "";
        foreach my $isl (@{$self->{isl_ports}}) {
            # Normalize both ports before comparing
            my $norm_isl = $self->_normalize_port_name(-port => $isl);
            if ($self->error) {
                return undef;
            }
            my $norm_int = $self->_normalize_port_name(-port => $interface);
            if ($self->error) {
                return undef;
            }
            if ($norm_isl eq $norm_int) {
                $isl_port = $isl;
                $isl_found = 1;
                last;
            }
        }

        if ($isl_found == 1) {
            $self->sanlog(-msg => "NOTE: MAC ($address) was found on an ISL port ($isl_port), ignoring this\n" .
                                  "since the connection to this MAC is not local on this switch. Make sure that\n" .
                                  "switch connected to port ($isl_port) is in your ETHERNET_SWITCHES list");
            return undef;
        } else {
            # burt505898 (We need to never block entire port channels), in this case,
            # pick apart all the members of the port channel and return them all individually
            if ($interface =~ /Po|port\-channel/i) {
                # get all the individual members of the port channel
                my @pchan_members = $self->_get_port_channel_members($interface);

                # add each port channel member into the interfaces list
                # dont add dups to the list (don't block twice)
                foreach my $pchan_member (@pchan_members) {
                    if (not grep(/^$pchan_member$/, @interfaces)) {
                        push(@interfaces, $pchan_member);
                    }
                }
            } else {
                # non port channel case:
                # push in all interfaces that were found. Look to see if
                # it already in the list (don't block twice).
                if (grep(/^$interface$/, @interfaces) == 0) {
                    push(@interfaces, $interface);
                }
            }
        }
    }

    # remove any dups from the interfaces list
    my %hash;
    foreach my $int (@interfaces) {
        $hash{$int}++;
    }
    @interfaces = keys %hash;

    # join together all the interfaces found and return.
    return (join(';', @interfaces));
}

###############################################################################
# Method:    _get_port_channel_members
# Objective: Obtain the members of a given port channel
###############################################################################
sub _get_port_channel_members {
    my ($self, $port_channel) = @_;

    # if it doesnt have port-channel or Po, append it in
    if ($port_channel !~ /Po|port\-channel/i) {
        $port_channel = "port-channel" . $port_channel;
    }

    my @pchan_members;
    my @pchan_output = $self->{parent}->execute("show int $port_channel");
    foreach my $pchan_line (@pchan_output) {
        if ($pchan_line =~ /Members in this channel:\s+(.+)/i) {
            my $match = $1; 
            $match =~ s/,//g; 
            @pchan_members = split(/\s+/, $match);
            last;
        }
    }

    return (wantarray()) ? @pchan_members : \@pchan_members;
}

###############################################################################
# Method:    _cleanup_config_mode
# Objective: If there was an error in processing a list of commands, this
#            function will attempt to get out of Cisco config mode.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cleanup_config_mode {
    my ($self) = @_;

    my (@cmds, @output);

    # Attempt to get out of config mode, by simply sending newline
    # followed by end command.
    @cmds = (" ", "end");
    foreach my $cmd (@cmds) {
        @output = $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoEthernetPort::_cleanup_config_mode() " .
                            "error executing command ($cmd)");
        }
    }
}

1;

