# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for ports for a Cisco Ethernet switch
## @author thomasm@netapp.com
## @status private
## @pod ./Port.pod

# Package Name
##############
package San::SanSwitch::Ethernet::PortMgr::CorduroyLEthernetPort;

use strict;

# Package Requirements
######################
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::PortMgr;
use San::SanLog;
use Error;
use Data::Dumper;

use vars qw(@ISA);

@ISA = qw(Error SanLog PortMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_port_num()
# Objective:       Retrieve the port number/interface corresponding to a
#                  WWPN, Return the entire array
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_num {
    my ($self, %args) = @_;

    $self->_clear_err();

    my @ports = $self->get_port_nums(%args);

    # if the caller wants an array, give them the full list, otherwise give them
    # only the first element, burt505898, needed for c-mode 7-mode compatibility
    # TODO: remove once c-mode support is fully implemented
    return (wantarray()) ? @ports : $ports[0];
}

################################################################################
# Subroutine Name: get_port_nums()
# Objective:       Retrieve the all port number/interface corresponding to a
#                  MAC.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_nums {
    my ($self, %args) = @_;

    my ($mac, @found_ports, $found_port);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwpn/) {
            if (not length($mac = $args{$_})) {
                $self->_set_err("Error: CorduroyLEthernetPort::get_port_nums() " .
                                "Invalid value for argument '-wwpn'");
                return ();
            }
            next;
        }
    }

    # look for the mac address on the switch
    if (defined($mac)) {
        $self->sanlog(-msg => "Searching for MAC $mac");
        $found_port = $self->_get_port_from_mac($mac);
        if (defined($found_port)) {
            $self->sanlog(-msg => "Found match on port $found_port");
            my @fps = split(/\;/, $found_port);
            push(@found_ports, @fps);
        } else {
            $self->_set_err("Error: CorduroyLEthernetPort::get_port_nums() MAC could not be found on this switch");
            return ();
        }
    } else {
        $self->_set_err("Error: CorduroyLEthernetPort::get_port_nums() MAC must be specified");
        return ();
    }

    return (@found_ports);
}

################################################################################
# Subroutine Name: port_disable()
# Objective:       Disable the port/interface specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_disable {
    my ($self, $port) = @_;
    my ($status);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    my @cmds = ("conf",
                "int $port",
                "shutdown",
                "exit",
                "exit",
                "show port $port");

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CorduroyLEthernetPort::port_disable() ".
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # wait for the port to transition to offline, or report error
    $status = $self->waitfor_port($port, "offline");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: CorduroyLEthernetPort::port_disable() " .
                        "failed to disable the port, port still appears up");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_enable()
# Objective:       Enable the port specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_enable {
    my ($self, $port) = @_;
    my ($status, @output);

    $self->sanlog(-msg => "Passed port is ".$port);
    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};
    my $count               = 0;

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    my @cmds = ("conf",
                "int $port",
                "no shutdown",
                "exit",
                "exit");

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CorduroyLEthernetPort::port_enable() " .
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # add explicit sleep into the code here before checking the
    # port to see if it is online (suggested by Dave/Sunil)
    sleep (3);

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: CorduroyLEthernetPort::port_enable() " .
                        "failed to enable port, port still appears down");
        return $SwitchUtil::FALSE;
    }

    # compare values now (after enable) versus before the
    # port_disable
    $self->check_port_settings($port);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: set_port_speed()
# Objective:       Set the specified port to the specified speed
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_speed {
    my ($self, %args) = @_;
    my ($status, $port, $speed);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};
    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Error: CorduroyLEthernetPort::set_port_speed() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?speed/) {
            if ((not length($speed = $args{$_})) &&
                ($args{$_} !~ /^\d$/ or $speed !~ /auto/i)) {
                $self->_set_err("Error: CorduroyLEthernetPort::set_port_speed() " .
                                "Invalid value for argument '-speed'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # handle auto speed setting
    if ($speed =~ /AUTO/i) {
        $speed = "auto";
    }

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    my @cmds;
    if ($speed =~ /AUTO/i) {
        @cmds = ("conf",
                "int $port",
                "auto-negotiate",
                "exit",
                "exit");
    } else {
        @cmds = ("conf",
                "int $port",
                "speed $speed",
                "exit",
                "exit");
    }

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CorduroyLEthernetPort::set_port_speed() " .
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: CorduroyLEthernetPort::set_port_speed() " .
                        "Port state did not transtion to online state");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: get_port_info()
# Objective:       Get information about the specified port/interface
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_info {
    my ($self, $port) = @_;
    my ($available_speeds, @output);

    $self->_clear_err();

    # Initialize hash
    my $hash = {};
    $hash->{state}              = "unknown";
    $hash->{speed}              = 0;
    $hash->{media_type}         = "";
    $hash->{mac}                = 0;
    $hash->{min_speed}          = 0;
    $hash->{max_speed}          = 0;
    $hash->{hardware}           = "";
    $hash->{description}        = "";
    $hash->{mtu}                = 0;
    $hash->{is_trunk}           = 0;
    $hash->{vlan_id}            = 1;

    # sometimes Cisco uses Ethernet1/6 and sometimes Eth1/6.
    # This screws up parsing. Get list the port information
    my $stripped_port = $port;
    $stripped_port =~ s/^(\D+)//;

    @output = $self->{parent}->execute("show port $port");
    foreach my $line (@output) {
        if ($line =~ /(10|100)(M|G)/) {
            my $speed = $1.$2;
            $hash->{speed} = $speed;
        }
        if ($line =~ /Auto/i) {
            $hash->{speed} = "auto";
        }
        if ($line =~ /MTU (\d+) bytes/) {
            $hash->{mtu} = int($1);
        }
        if ($line =~ /(.+)\s+PC Mbr\s+.*/i) {
            $hash->{hardware} = "Port-Channel";
            $hash->{is_portchannel} = 1;
        }
        if ($line =~ /$port (.*)(Up|Down)/) {
            my $state = $2;
            $state =~ s/,//;
            $hash->{state} = &SwitchUtil::normalize_state($state);
        }
        if ($line =~ /media type is (\S+)/) {
            $hash->{media_type} = $1;
        }
        if ($line =~ /Port mode is (\S+)/) {
            my $is_trunk = $1;
            if ($is_trunk =~ /trunk/i) {
                $hash->{is_trunk} = 1;
            }
        }
    }
    
    # Use 'show port description' command output to obtain the MAC address 
    # and port description.
    @output = $self->{parent}->execute("show port description $port");
    foreach my $line (@output) {
        if ($line =~ /Description(\.+) (.*)/) {
            $hash->{description} = $2;
        }
        if ($line =~ /MAC address(\.+) (.*)/) {
            $hash->{mac} = $2;
        }
    }

    # Use 'show vlan port' command output to obtain the VLAN ID. 
    @output = $self->{parent}->execute("show vlan port $port");
    foreach my $line (@output) {
        if ($line =~ /$port(\s+)\d+\s+(\d+)/) {
            $hash->{vlan_id} = $2;
        }
    }

    # Load the cached values if we have already looked these values up before
    $hash->{min_speed} = $self->{cache}{min_speed}{$port}
        if defined($self->{cache}{min_speed}{$port});
    $hash->{max_speed} = $self->{cache}{max_speed}{$port}
        if defined($self->{cache}{max_speed}{$port});

    # If we have the default or invalid values, we either have not looked up the
    # capabilities of this port before or the previous look up failed.
    # Don't run if this is a trunk since the command will be invalid
    # Don't run if this is a port-channel the command will be invalid
    if (($hash->{min_speed} == 0 or $hash->{max_speed} == 0) and
         $hash->{is_trunk} == 0 and $hash->{is_portchannel} == 0) {
        @output = $self->{parent}->execute("show port $port");
        foreach my $line (@output) {
            if ($line =~ /Speed:\s+(\S+)/) {
                $available_speeds = $1;
            }
        }

        # speeds will show up as:
        # Speed:                 10G,100M,Auto," " (for port-channels)
        my @values = split(/,/, $available_speeds);
        # take "auto" into consideration and remove from list
        @values = grep(!/auto/, @values);

        # min will be first in list
        $hash->{min_speed} = $values[0];
        # Store this value into memory for later reuse
        $self->{cache}{min_speed}{$port} = $hash->{min_speed};

        # max will be last in list
        $hash->{max_speed} = $values[(scalar(@values) - 1)];
        # Store this value into memory for later reuse
        $self->{cache}{max_speed}{$port} = $hash->{max_speed};
    }

    return $hash;
}

################################################################################
# Subroutine Name: set_port_type()
# Objective:       Set the specified port to the specified port type
#                  no-op for ethernet for now
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_type {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $SwitchUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new port object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self);
    my $conn;

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: CorduroyLEthernetPort::_new() " .
                                "Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: CorduroyLEthernetSwitch::_new() " .
                                "Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/i) {
            $self->_set_err("Error: CorduroyLEthernetPort::_new() " .
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: CorduroyLEthernetPort::_new() " .
                        "Missing required argument '-parent'");
        return $self;
    }

    # initialize the object members
    $self->{parent} = $parent;

    # for compatibility with FC
    $self->{nameserver} = {};

    # get list of uplink or ISL ports
    my @isl_ports = $self->_get_isl_ports();
    $self->{isl_ports} = \@isl_ports;

    return $self;
}

###############################################################################
# Method:    _get_isl_ports
# Objective: Retrieve list of ports that correspond to ISL ports to other
#            switches
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_isl_ports {
    my ($self) = @_;
    my @isl_ports = ();

    # STAND ALONE ISL PORTS
    my @output = $self->{parent}->execute("show isdp neighbors");
    #chat-3270-2            0/7          132       H            FAS3270              e1a
    #chat-3270-3            0/8          123       H            FAS3270              e1a
    #chat-cord-cs5          0/13         178       S            CN1610               0/13
    #chat-cord-cs5          0/14         178       S            CN1610               0/14
    #chat-cord-cs5          0/15         178       S            CN1610               0/15
    #chat-cord-cs5          0/16         178       S            CN1610               0/16

    foreach my $line (@output) {
        # skip isdp neighbors that are not listed as switches
        next if ($line !~ /\sS\s/);

        # parse out the port
        if ($line =~ /\s+(\d+\/\d+)\s+/i) {
            my $isl_port = $1;
            push(@isl_ports, $isl_port);
        }
    }

    # PORT-CHANNELS CONTAINING ISL PORTS
    # first check if we have a cached output
    if (!defined($self->{cache}{show_port_all})) {
        @output = $self->{parent}->execute(
            "show port all",
            undef,
            use_connectrec => 1,
            match_table => [
            ("--More--")  => "q",
        ]);
        
        @output = split("\n", $output[0]);
        
        $self->{cache}{show_port_all} = \@output;
    } else {
        @output = @{$self->{cache}{show_port_all}};
    }

    foreach my $line (@output) {
        # check the port channels
        if ($line =~ /^(3\/\d+)\s(.+)Up/i) {
            # we are looking at a port channel that's currently online
            my $pchan_int = $1;

            my @pchan_members = $self->_get_port_channel_members($pchan_int);

            foreach my $member (@pchan_members) {
                # if a member of this port channel is an ISL port, then add
                # this port channel to the ISL port list
                if (grep(/^\s*${member}\s*$/, @isl_ports)) {
                    # add this port channel to the isl port list and break out
                    push(@isl_ports, $pchan_int);
                    last;
                }
            }
        }
    }

    my $fw_version = $self->{parent}->get_firmware_version();
    if ($self->{parent}->error) {
        $self->_set_err("Could not get fw version of this switch");
        return ();
    }

    return (@isl_ports);
}

###############################################################################
# Method:    _normalize_port_name
# Objective: Normalize the port name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _normalize_port_name {
    my ($self, %args) = @_;
    my ($port, $normalized_port);

    # parse arguments
    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Invalid value for argument '-port'");
                return "";
            }
            next;
        }
    }

    unless (defined($port) and $port ne "") {
        $self->_set_err("Missing required parameter '-port'");
        return "";
    }

    if ($port =~ /^Eth(\d+.*)/i) {
        $normalized_port = "Ethernet" . $1;
    } elsif ($port =~ /^Ethernet(\d+.*)/i) {
        $normalized_port = "Ethernet" . $1;
    } elsif ($port =~ /^GigabitEthernet(\d+.*)/i) {
        $normalized_port = "GigabitEthernet" . $1;
    } elsif ($port =~ /^Gi(\d+.*)/i) {
        $normalized_port = "GigabitEthernet" . $1;
    } elsif ($port =~ /^Port\-Channel(\d+.*)/i) {
        $normalized_port = "Port-Channel" . $1;
    } elsif ($port =~ /^PO(\d+.*)/i) {
        $normalized_port = "Port-Channel" . $1;
    } else {
        $normalized_port = $port;
    }

    return $normalized_port;
}

###############################################################################
# Method:    _get_nameserver_database
# Objective: Retrieve nameserver database information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_nameserver_database {
    # no-op for ethernet
    return {};
}

###############################################################################
# Method:    _get_port_from_mac
# Objective: Given a passed in MAC address, look to see if that MAC is located
#            on the switch. Ignore if the found port is an ISL since we do not
#            want to block the ISL port.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_port_from_mac {
    my ($self, $address) = @_;
    my $conn = $self->{parent}->{conn};
    my ($found_line, @interfaces, $interface);
    my ($model, $cmd);

    # CN1610 output:
    # (chat-cord-cs1) #show mac-addr-table all
	#
	#VLAN ID  MAC Address         Interface  IfIndex  Status
	#-------  ------------------  ---------  -------  ------------
	#1        00:07:43:06:43:BC   0/1        1        Learned
	#1        00:07:43:06:43:BD   3/1        18       Learned
	#1        00:07:43:06:44:36   0/3        3        Learned
	#1        00:07:43:06:44:37   3/1        18       Learned
	#1        00:07:43:06:49:20   0/2        2        Learned
	#1        00:07:43:06:49:21   3/1        18       Learned
	#1        00:07:43:06:69:63   0/4        4        Learned
	#1        00:07:43:06:69:64   3/1        18       Learned
	#1        00:07:43:06:F3:82   0/6        6        Learned
	#1        00:07:43:06:F3:83   3/1        18       Learned
	#1        00:07:43:06:FB:0A   0/5        5        Learned
	#1        00:07:43:06:FB:0B   3/1        18       Learned
	#1        00:A0:98:4B:AD:31   3/1        18       Learned
	#1334     00:A0:98:4B:AD:73   5/1        17       Management

    my @output = $self->{parent}->execute("show mac-addr-table");

    # grab only the relevant lines and strip away the header
    my @matches = grep(/^\d+\s+(.*)\s+($address)/i, @output);

    # no matches found.
    if (scalar(@matches) == 0) {
        return undef;
    }

    # can get multiple ports for same MAC in case of VIF support (w/Port Channel support
    # on switch)
    if (scalar(@matches) > 1) {
        $self->sanlog(-msg => "NOTE: Found multiple ports that correspond to this MAC ($address), could be a VIF");
    }

    foreach my $match (@matches) {
        my $line = $match;
        
        # grab interface from the output line
        if ($line =~ /(.*)($address)\s+(\d+\/\d+)\s+/i) {
            $interface = $3;
        }

        # look to see if the MAC came from a port that is an ISL or uplink port
        # If it is then the MAC is connected on another switch and not this one.
        my $isl_found = 0;
        my $isl_port = "";
        foreach my $isl (@{$self->{isl_ports}}) {
            # Normalize both ports before comparing
            my $norm_isl = $self->_normalize_port_name(-port => $isl);
            if ($self->error) {
                return undef;
            }

            my $norm_int = $self->_normalize_port_name(-port => $interface);
            if ($self->error) {
                return undef;
            }

            if ($norm_isl eq $norm_int) {
                $isl_port = $isl;
                $isl_found = 1;
                last;
            }
        }

        # ISL CASE
        # the mac address matched to an ISL port. Bail out because we don't want to block an ISL.
        if ($isl_found == 1) {
            $self->sanlog(-msg => "NOTE: MAC ($address) was found on an ISL port ($isl_port), ignoring this\n" .
                                  "since the connection to this MAC is not local on this switch. Make sure that\n" .
                                  "switch connected to port ($isl_port) is in your ETHERNET_SWITCHES list");
            return undef;
        } else {
            # NON-ISL CASE
            if ($interface =~ /^3\/\d+/i) {

                # Interface is 3/# so it's a port channel.
                my @pchan_members = $self->_get_port_channel_members($interface);

                # add each port channel member into the interfaces list
                # dont add dups to the list (don't block twice)
                foreach my $pchan_member (@pchan_members) {
                    if (not grep(/^$pchan_member$/, @interfaces)) {
                        push(@interfaces, $pchan_member);
                    }
                }
            } else {
                # non port channel case:
                # push in all interfaces that were found. Look to see if
                # it already in the list (don't block twice).
                if (grep(/^$interface$/, @interfaces) == 0) {
                    push(@interfaces, $interface);
                }
            }

        }

    }

    # remove any dups from the interfaces list
    my %hash;
    foreach my $int (@interfaces) {
        $hash{$int}++;
    }
    @interfaces = keys %hash;

    # join together all the interfaces found and return.
    return (join(';', @interfaces));
}

###############################################################################
# Method:    _get_port_channel_members
# Objective: Obtain the members of a given port channel
###############################################################################
sub _get_port_channel_members {
    my ($self, $port_channel) = @_;

    my @pchan_members;
    my @pchan_output = $self->{parent}->execute("show port-channel $port_channel");
    foreach my $pchan_line (@pchan_output) {
        if ($pchan_line =~ /(^\d+\/\d+)/i) {
            my $match = $1;
            $match =~ s/,//g;
            push(@pchan_members, $match);
        }
    }

    return (wantarray()) ? @pchan_members : \@pchan_members;
}

###############################################################################
# Method:    _cleanup_config_mode
# Objective: If there was an error in processing a list of commands, this
#            function will attempt to get out of Cisco config mode.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cleanup_config_mode {
    my ($self) = @_;
    my (@cmds, @output);

    # Attempt to get out of config mode, by simply sending newline
    # followed by end command.
    @cmds = (" ", "exit");
    foreach my $cmd (@cmds) {
        @output = $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoEthernetPort::_cleanup_config_mode() " .
                            "error executing command ($cmd)");
        }
    }
}

1;
