# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a Cisco Ethernet switch
## @author thomasm@netapp.com
## @status private
## @pod ../Switch.pod

use strict;

# Package Name
##############
package San::SanSwitch::Ethernet::Switch::CiscoEthernetSwitch;

# Package Requirements
######################
use San::SanSwitch;
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanSwitch);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

###############################################################################
# Method:    get_firmware_version
# Objective: Get the firmware version of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_firmware_version {
    my ($self) = @_;

    my (@output, $output);
    my $version = "unknown";
    my $type;
    my $is_7k = 0;

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_version})) {
        @output = $self->execute("show version");
        $self->{cache}{show_version} = \@output;
    } else {
        @output = @{$self->{cache}{show_version}};
    }

    # get revision OS type
    if (grep(/Cisco IOS/, @output) > 0) {
        $type = "IOS";
    } elsif (grep(/Cisco Nexus7000/, @output) > 0) {
        $type = "NXOS";
        $is_7k = 1;
    } elsif (grep(/Cisco Nexus/, @output) > 0) {
        $type = "NXOS";
    } else {
        $type = "unsupported";
    }

    # get the firmware revision
    foreach my $line (@output) {
        if ($type eq "IOS") {
            if ($line =~ /.*,\s+Version\s+(.*),.*/i) {
                $version = $1;
                last;
            } 
        } elsif ($type eq "NXOS") {
            # Nexus 7000 has different format (release in 4.2)
            if ($is_7k == 1) {
                if ($line =~ /System\s+version:\s+(\S+)/i) {
                    $version = $1;
                    last;
                } 
            } else {
                if ($line =~ /system:\s+version\s+(\S+)/i) {
                    $version = $1;
                    last;
                } 
            }
        }
    }

    # concat OS and Type
    $version = $version . "-" . $type;

    return $version;
}

###############################################################################
# Method:    get_model
# Objective: Get the model of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_model {
    my ($self) = @_;
    my (@output, $output);
    my $model = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_version})) {
        @output = $self->execute("show version");
        $self->{cache}{show_version} = \@output;
    } else {
        @output = @{$self->{cache}{show_version}};
    }

    if ($self->{version} =~ /IOS/) {
        foreach my $line (@output) {
            # get the model of switch (get from processor)
            if ($line =~ /cisco (.*)\s+\(.*\)\s+processor.*/i) {
                $model = $1;
                last;
            }
        }
    } elsif ($self->{version} =~ /NXOS/) {
        for (my $i = 0; $i < @output; $i++) {
            if ($output[$i] =~ /^Hardware/) {
                $model = $output[$i + 1];
                $model =~ s/\s+//;
                last;
            }
        }
    }

    return $model;
}

###############################################################################
# Method:    dump_topology
# Objective: Dump the fabric topology
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_topology {
    my ($self, %args) = @_;

    $self->_clear_err();

    # No-Op for Cisco Ethernet

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    disable
# Objective: Disable the switch (not currently supported for Cisco)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable {
    my ($self) = @_;

    $self->_clear_err();

    $self->_set_err("Error: CiscoEthernetSwitch::disable() " .
                    "this subroutine is not supported on Cisco switches");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    enable
# Objective: Enable the switch (not currently supported for Cisco)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable {
    my ($self) = @_;

    $self->_clear_err();

    $self->_set_err("Error: CiscoEthernetSwitch::enable() " .
                    "this subroutine is not supported on Cisco switches");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    reboot
# Objective: Reboot the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;
    my $count = 0;

    $self->_clear_err();

    my $original_persistent = $self->{persistent};

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->options(-persistent => 1);

    # execute a show version command to send via execute and it will do the
    # "right thing" on re-connect.
    my @output = $self->execute("show version");

    # Bypass using generic execute subroutine so that the prompt can be
    # controlled
    &Tharn::say($self->{conn}, "reload", '-prompt', 
                'This command will reboot the system.', 
                '-timeout', $self->{timeout});
    eval {
        &Tharn::say($self->{conn}, "y", '-noprompt', '-prompt', "y", 
                    '-timeout', $self->{timeout});
    };

    # Restore persistent setting
    $self->options(-persistent => $original_persistent);

    # sleep to allow boot process to start
    sleep(30);

    # Wait until the switch does not respond
    while ($self->is_alive() == $SwitchUtil::FALSE && 
           $count < $self->{reboot_timeout}) {
        sleep(1);
        $count++;
    }

    if ($count < $self->{reboot_timeout}) {
        # wait some additional time to allow telnet service to be
        # available and responding. Also wait some time to allow all
        # ports to log back in (or nameserver db will be empty).
        sleep(120);
        $self->reconnect();

        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    info
# Objective: Get switch information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    push @output, "Switch name:           " . $self->hostname(), 
                  "Switch type:           " . $self->hosttype(),
                  "Switch FW version:     " . $self->fw_version(), 
                  "Switch Model:          " . $self->model();
    push @output, "\n";

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    ntp_sync
# Objective: Synchronize the clock on the switch with an NTP server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($ntp_server, $ntp_time_zone);
    my ($time_zone, $gmt_offset_hour, $gmt_offset_min);
    my (@output);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: CiscoEthernetSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_server'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: CiscoEthernetSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_time_zone'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    my $original_persistent = $self->{persistent};

    # Set the offset from GMT for the timezone
    if ($ntp_time_zone =~ /EST/i) {
        $gmt_offset_hour = -5;
        $gmt_offset_min  = 0;
        $time_zone  = "EST";
    } elsif ($ntp_time_zone =~ /PST/i) {
        $gmt_offset_hour = -8;
        $gmt_offset_min  = 0;
        $time_zone  = "PST";
    } elsif ($ntp_time_zone =~ /MST/i) {
        $gmt_offset_hour = -7;
        $gmt_offset_min  = 0;
        $time_zone  = "MST";
    } elsif ($ntp_time_zone =~ /IST/i) {
        $gmt_offset_hour = 5;
        $gmt_offset_min  = 30;
        $time_zone  = "IST";
    } else {
        $self->_set_err("Error: CiscoEthernetSwitch::ntp_sync() " .
                        "Time zone $ntp_time_zone not currently supported");
        return $SwitchUtil::FALSE;
    }

    my $gmt_string = sprintf("%02d:%02d", $gmt_offset_hour, $gmt_offset_min);
    $self->sanlog(-msg => "Setting the Time Zone (TZ) to $time_zone " .
                          "(GMT $gmt_string)");

    # get IP address since switch may not have DNS running
    my ($ntp_server_ip);
    my (undef, undef, undef, undef, @addrs) = gethostbyname($ntp_server);
    foreach my $ip (@addrs) {
        my ($a, $b, $c, $d) = unpack('C4', $ip);
        $ntp_server_ip = "$a\.$b\.$c\.$d";
    }

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->options(-persistent => 1);

    my @cmds = ("conf t",
                "ntp server $ntp_server_ip",
                "clock timezone $time_zone $gmt_offset_hour $gmt_offset_min",
                "end");

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        if ($cmd =~ /show clock/) {
            # Wait a few seconds for the updates to take effect for show clock
            sleep(3);
        }
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: CiscoEthernetSwitch::ntp_sync() " .
                            "Could not execute command $cmd");
            # if we have an error, kick out of foreach loop with 
            # error set
            $cmd_error = 1;
            last;
        }
    }

    # handle error, by attempting to get out of config mode
    # and then returning with the proper error set.
    if ($cmd_error == 1) {
        @cmds = (" ", "end");
        foreach my $cmd (@cmds) {
            @output = $self->execute($cmd);
        }
        return $SwitchUtil::FALSE;
    }

    # Restore persistent setting
    $self->options(-persistent => $original_persistent);

    $self->sanlog(-msg => "Local time is: $output[0]");

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    install_firmware
# Objective: Install firmware on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub install_firmware {
    my ($self, %args) = @_;

    $self->_clear_err();
    #TODO: This subroutine to be completed at a later date
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    get_log
# Objective: Retrieve the log file from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_log {
    my ($self) = @_;

    $self->_clear_err();

    my @output = $self->execute("show logging");
    if ($self->error) {
        $self->_set_err("Error: CiscoEthernetSwitch::get_log() " .
                        "Could not retrieve log file");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_config
# Objective: Retrieve the configuration from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_config {
    my ($self) = @_;

    $self->_clear_err();

    my @output = $self->execute("show run");
    if ($self->error) {
        $self->_set_err("Error: CiscoEthernetSwitch::get_config() " .
                        "Could not retrieve configuration");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_state
# Objective: Retrieve the state of the switch (enabled/disabled). Not supported
#            on Cisco, so assume state is online.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();

    return "online";
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new CiscoFC switch object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: CiscoEthernetSwitch::_new() " .
                                "Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } 
    }

    if (!defined($conn) || ref($conn) ne "Connectrec") {
        $self->_set_err("Error: CiscoEthernetSwitch::_new() " .
                        "Invalid value for argument '-conn'");
        return $self;
    }

    # set connection
    $self->{conn} = $conn;

    return $self;
}

###############################################################################
# Method:    _get_npiv
# Objective: Retrieve information to determine if switch is acting NPV mode
#            NA for Ethernet switches
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_npiv {
    my ($self) = @_;

    $self->_clear_err();

    # NA for Ethernet switches
    return(0);
}

###############################################################################
# Method:    _get_domain_id
# Objective: Get the domain id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_domain_id {
    my ($self) = @_;

    # NO-OP

    return "";
}

1;

=pod

=head1 Package Name

CiscoEthernetSwitch.pm

=head1 Author

Tom Masterson

=head1 Description

For description of methods see B<San::SanSwitch::Switch.pod>.

=cut

