# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a Cisco Ethernet switch
## @author thomasm@netapp.com
## @status private
## @pod ../Switch.pod

use strict;

# Package Name
##############
package San::SanSwitch::Ethernet::Switch::CorduroyLEthernetSwitch;

# Package Requirements
######################
use San::SanSwitch;
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanSwitch);


################################################################################
# P U B L I C   M E T H O D S
################################################################################

###############################################################################
# Method:    get_firmware_version
# Objective: Get the firmware version of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_firmware_version {
    my ($self) = @_;

    my (@output, $output);
    my $version = "unknown";
    my $type;

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_version})) {
        @output = $self->execute(
            "show version",
            undef,
            use_connectrec => 1,
            match_table => [
                ("--More--")  => "q",
        ]);

        $self->{cache}{show_version} = \@output;
    } else {
        @output = @{$self->{cache}{show_version}};
    }

    # get the firmware revision
    foreach my $line (@output) {
            if ($line =~ /Software Version\S+?\s+(\S+)/i) {
                $version = $1;
                last;
            }
    }

    # concat OS and Type
    $version = $version;

    return $version;
}

###############################################################################
# Method:    get_model
# Objective: Get the model of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_model {
    my ($self) = @_;
    my (@output, $output);
    my $model = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_version})) {
        @output = $self->execute(
            "show version",
            undef,
            use_connectrec => 1,
            match_table => [
                ("--More--")  => "q",
        ]);

        $self->{cache}{show_version} = \@output;
    } else {
        @output = @{$self->{cache}{show_version}};
    }

    foreach my $line (@output) {
        # get the model of switch (get from processor)
        if ($line =~ /Machine Model\S+?\s+(\S+)/i) {
            $model = $1;
            last;
        }
    }

    return $model;
}

###############################################################################
# Method:    dump_topology
# Objective: Dump the fabric topology
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_topology {
    my ($self, %args) = @_;

    $self->_clear_err();

    # No-Op for Corduroy

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    disable
# Objective: Disable the switch (not currently supported for Cisco)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable {
    my ($self) = @_;

    $self->_clear_err();

    $self->_set_err("Error: CorduroyLEthernetSwitch::disable() " .
                    "this subroutine is not supported on Corduroy-L switches");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    enable
# Objective: Enable the switch (not currently supported for Cisco)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable {
    my ($self) = @_;

    $self->_clear_err();

    $self->_set_err("Error: CorduroyLEthernetSwitch::enable() " .
                    "this subroutine is not supported on Corduroy-L switches");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    reboot
# Objective: Reboot the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;
    my $count = 0;
    my @output;

    $self->_clear_err();

    my $original_persistent = $self->{persistent};

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->options(-persistent => 1);

    # execute the reload command
    @output = $self->execute(
            "reload",
            undef,
            use_connectrec => 1,
            on_eof => sub { return 1; },
            match_table => [
                # Pending changes will be saved rather than discarded
                ("Would you like to save them now?") => "y",
                ("Are you sure you would like to reset the system?")  => "y",
            ]);

    # Restore persistent setting
    $self->options(-persistent => $original_persistent);

    # sleep to allow boot process to start
    sleep(30);

    # Wait until the switch does not respond
    while ($self->is_alive() == $SwitchUtil::FALSE &&
           $count < $self->{reboot_timeout}) {
        sleep(1);
        $count++;
    }

    if ($count < $self->{reboot_timeout}) {
        # wait some additional time to allow telnet service to be
        # available and responding. Also wait some time to allow all
        # ports to log back in (or nameserver db will be empty).
        sleep(60);
        $self->reconnect();

        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    info
# Objective: Get switch information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    push @output, "Switch name:           " . $self->hostname(),
                  "Switch type:           " . $self->hosttype(),
                  "Switch FW version:     " . $self->fw_version(),
                  "Switch Model:          " . $self->model();
    push @output, "\n";

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    ntp_sync
# Objective: Synchronize the clock on the switch with an NTP server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($ntp_server, $ntp_time_zone);
    my (@output);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: CorduroyLEthernetSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_server'");
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: CorduroyLEthernetSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_time_zone'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    my $original_persistent = $self->{persistent};

    $self->sanlog(-msg => "Time zone configuration is not supported for the Corduroy L switch at this time.");

    # get IP address since switch may not have DNS running
    my ($ntp_server_ip);
    my (undef, undef, undef, undef, @addrs) = gethostbyname($ntp_server);
    foreach my $ip (@addrs) {
        my ($a, $b, $c, $d) = unpack('C4', $ip);
        $ntp_server_ip = "$a\.$b\.$c\.$d";
    }

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->options(-persistent => 1);

    # execute sntp commands
    my @cmds = ("config",
                "sntp client mode unicast",
                "sntp server $ntp_server_ip",
                "exit");

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        if ($cmd =~ /show clock/) {
            # Wait a few seconds for the updates to take effect for show clock
            sleep(3);
        }

        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: CorduroyLEthernetSwitch::ntp_sync() " .
                            "Could not execute command $cmd");
            # if we have an error, kick out of foreach loop with
            # error set
            $cmd_error = 1;
            last;
        }
    }

    # handle error, by attempting to get out of config mode
    # and then returning with the proper error set.
    if ($cmd_error == 1) {
        @cmds = (" ", "exit");
        foreach my $cmd (@cmds) {
            @output = $self->execute($cmd);
        }
        return $SwitchUtil::FALSE;
    }

    # Restore persistent setting
    $self->options(-persistent => $original_persistent);

    $self->sanlog(-msg => "Local time is: $output[0]");

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    install_firmware
# Objective: Install firmware on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub install_firmware {
    my ($self, %args) = @_;

    $self->_clear_err();
    #TODO: This subroutine to be completed at a later date
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    get_log
# Objective: Retrieve the log file from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_log {
    my ($self) = @_;
    my @output;

    $self->_clear_err();

    @output = $self->execute(
         "show logging buffered",
         undef,
         use_connectrec => 1,
         match_table => [
                ("--More--")  => "m",
         ]);

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_config
# Objective: Retrieve the configuration from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_config {
    my ($self) = @_;

    $self->_clear_err();

    my @output = $self->execute("show running-config");
    if ($self->error) {
        $self->_set_err("Error: CorduroyLEthernetSwitch::get_config() " .
                        "Could not retrieve configuration");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_state
# Objective: Retrieve the state of the switch (enabled/disabled). Not supported
#            on Cisco, so assume state is online.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();

    return "online";
}

################################################################################qq
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new CorduroyLEthernetSwitch switch object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: CorduroyLEthernetSwitch::_new() " .
                                "Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        }
    }

    if (!defined($conn) || ref($conn) ne "Connectrec") {
        $self->_set_err("Error: CorduroyLEthernetSwitch::_new() " .
                        "Invalid value for argument '-conn'");
        return $self;
    }

    # set connection
    $self->{conn} = $conn;

    return $self;
}

###############################################################################
# Method:    _get_npiv
# Objective: Retrieve information to determine if switch is acting NPV mode
#            NA for Ethernet switches
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_npiv {
    my ($self) = @_;

    $self->_clear_err();

    # NA for Ethernet switches
    return(0);
}

###############################################################################
# Method:    _get_domain_id
# Objective: Get the domain id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_domain_id {
    my ($self) = @_;

    # NO-OP

    return "";
}

1;

=pod

=head1 Package Name

CorduroyLEthernetSwitch.pm

=head1 Author

Dan Baehr

=head1 Description

For description of methods see B<San::SanSwitch::Switch.pod>.

=cut


