# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for ports for a Apcon FC switch
## @author Michael Latif (latif@netapp.com)
## @status private
## @pod ./Port.pod

# Package Name
##############
package San::SanSwitch::FC::PortMgr::ApconFCPort;

use strict;

# Package Requirements
######################
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::PortMgr;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog PortMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: get_port_num()
# Objective:       Retrieve the port number/interface corresponding to a 
#                  WWPN, by returning the first match in the array.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_num {
    my ($self, %args) = @_;
    my (@ports);

    @ports = $self->get_port_nums(%args);
    return "" if ($self->error);

    return $ports[0];

}

################################################################################
# Subroutine Name: get_port_nums()
# Objective:       Retrieve all port number/interface corresponding to a 
#                  WWPN.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_nums {
    my ($self, %args) = @_;
    my ($wwpn, $found_port, @found_ports);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwpn/) {
            if (not length($wwpn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-wwpn'");
                return ();
            }
            next;
        }
    }

    $self->{nameserver} = $self->get_nameserver();
    if (defined($wwpn)) {
        $self->sanlog(-msg => "Searching for WWPN $wwpn",
                      -msg_level => 7);
        $wwpn = lc($wwpn);
        $wwpn =~ s/://g;
        foreach my $interface (keys %{$self->{nameserver}}) {
            if ($self->{nameserver}{$interface}{wwpn} eq $wwpn) {
                $found_port = $interface;
                $self->sanlog(-msg => "Found match on port $found_port",
                              -msg_level => 3);
                push (@found_ports, $found_port);
            }
        }
    } else {
        $self->_set_err("WWPN must be specified");
        return ();
    }

    if (!(@found_ports)) {
        $self->_set_err("no matching port found");
        return ();
    } else {
        return @found_ports;
    }
}

################################################################################
# Subroutine Name: port_disable()
# Objective:       Disable the port/interface specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_disable {
    my ($self, $port) = @_;
    my ($status);

    $self->_clear_err();

    # if requested, get current state to store speed and topology
    if ($self->{parent}->{verify_on_enable} == 1) {
        my $port_state = $self->get_port_info($port);
        if ($self->error) {
            $self->_set_err("error getting port information ($port)");
            return $SwitchUtil::FALSE;
        }
        $self->{$port}->{previous_speed} = $port_state->{speed};
        $self->{$port}->{previous_type} = $port_state->{type};
    }

    # block the specified port
    my $cmd = "disconnect $port";
    $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->{$port}->{previous_speed} = undef;
        $self->{$port}->{previous_type} = undef;
        $self->_set_err("error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    # wait (default) for the port to transition to offline, or report error
    if ($self->{parent}->{validate_offline} == 1) {
        $status = $self->waitfor_port($port, "offline");
        if ($status == $SwitchUtil::FALSE) {
            $self->_set_err("failed to disable port, port still appears to be up");
            return $SwitchUtil::FALSE;
        }
    }

    return $SwitchUtil::TRUE;

}

################################################################################
# Subroutine Name: port_enable()
# Objective:       Enable the port specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_enable {
    my ($self, $port) = @_;
    my ($status, @output);

    $self->_clear_err();

    my $port_info = $self->{initial_active_connections}->{$port};

    if (not defined($port_info)) {
        $self->_set_err("port $port is invalid, the port does not appear " .
                        "to have any connections on the apcon");
        return $SwitchUtil::FALSE;
    }

    my $cmd = "connect duplex $port" . $port_info->{conn_port};
    $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    # add explicit sleep into the code here before checking the
    # port to see if it is online (suggested by Dave/Sunil)
    sleep (3);

    # wait for the port to transition to online, or report error
    # The port connections information will look like this:
    # Connections:        A13 <> A12
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("failed to enable port, port still appears down");
        return $SwitchUtil::FALSE;
    }

    # compare values now (after enable) versus before the
    # port_disable
    $self->check_port_settings($port);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: set_port_speed()
# Objective:       NO-OP for Apcon
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_speed {
    my ($self, %args) = @_;

    $self->_set_err("Invalid action specified for Apcon switch");
    return $SwitchUtil::FALSE;
}

################################################################################
# Subroutine Name: get_port_info()
# Objective:       Get information about the specified port/interface
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_info {
    my ($self, $port) = @_;

    my ($cmd, $lx_mode, @switchShow, @output, @sorted, @speeds);

    $self->_clear_err();

    # Initialize hash
    my $hash = {};

    # general fields
    $hash->{state}            = "unknown";
    $hash->{speed}            = 0;
    $hash->{type}             = "unknown";
    $hash->{min_speed}        = 0;
    $hash->{max_speed}        = 0;
    $hash->{conn_port}        = "";

    # sfpshow fields
    $hash->{sfp_speed_capabilities} = "";
    $hash->{sfp_max_9u_meters}      = 0;
    $hash->{sfp_max_50u_meters}     = 0;
    $hash->{sfp_max_625u_meters}    = 0;
    $hash->{sfp_vendor_name}        = 0;
    $hash->{sfp_part_number}        = 0;
    $hash->{sfp_wavelength}         = 0;

    # call show port info on the specified port
    $cmd = "show port info $port";
    @output = $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("error executing command ($cmd)");
        return {};
    }

    # Example output from show port info
    #apcon_bh_1>> show port info A13
    # A13 Name:               500110a00011b55c
    # A13 Type:               F16-6 1/2/4-Gig Fibre Channel Blade
    # A13 Status:             Active
    # A13 Rate:               00: Repeater Mode, Auto 1/2/4
    # A13 Port Link:          unknown
    # A13 Connections:        A13 <> A12
    # A13 Port Locking:       Disabled
    # A13 Zoning:             Disabled
    # A13 Presets:            -None-
    # A13 Name:               FINISAR CORP.
    # A13 Part:               FTLF8524P2BNV
    # A13 Revision:           A
    # A13 Serial:             PE31LXJ
    # A13 Date:               07/15/08
    # A13 Bitrate:            4200 Mbit/sec
    # A13 Laser:              Shortwave without OFC,
    # A13 Media:              50 micron MMF, 62.5 micron MMF,
    # A13 Protocol:           1.0625 Gbit/sec, 2.125 Gbit/sec, 4.25 Gbit/sec, 1000Base-SX,
    # A13 Diags:              available
    # A13 TX signal:          0.39mW  -4.04dBm
    # A13 RX signal:          0.23mW  -6.32dBm
    # A13 TX bias:            9.53mA
    # A13 Temp:               35.36C
    # A13 Power:              3.20V
    # A13 SFP Alarms:         -None-
    # A13 Sig Change Counter: 1

    foreach my $line (@output) {
        $line =~ s/\t/ /g;
        if ($line =~ /Connections\:\s+\S+\s+\<\>\s+(\S+)/) {
            $hash->{conn_port} = $1;
            $hash->{state} = "online";
        } elsif ($line =~ /Connections\:\s+None/i or $line =~ /Connections\:\s+\-None\-/i) {
            $hash->{state} = "offline";
            $hash->{conn_port} = "";
        } elsif ($line =~ /Part\:\s+(\S+)/) {
            $hash->{sfp_part_number} = $1;
        } elsif ($line =~ /Laser\:\s+(.+)/) {
            $hash->{sfp_wavelength} = $1;
        } elsif ($line =~ /Name\:\s+(.+)\s*/) {
            $hash->{sfp_vendor_name} = $1;
        } elsif ($line =~ /Protocol\:\s+(.+)/) {
            my $speeds = $1;
            my @speeds = ();
            foreach my $speed (split(/\,/, $speeds)) {
                if ($speed =~ /(\d+)\.\d+\s+gbit/i) {
                    push @speeds, $1;
                }
            }

            $hash->{sfp_speed_capabilities} = join(':', @speeds);
        } elsif ($line =~ /Media\:\s+(.+)/) {
            my $media_char = $1;
            foreach my $type (split(/\,/, $media_char)) {
                if ($type =~ /9 micron/) {
                    $hash->{sfp_max_9u_meters} = 1;
                } elsif ($type =~ /50 micron/) {
                    $hash->{sfp_max_50u_meters} = 1;
                } elsif ($type =~ /62\.5 micron/) {
                    $hash->{sfp_max_625u_meters} = 1;
                }
            }
        }

        if ($line =~ /Bitrate:.*(\d+)\s+/i) {
            $hash->{speed} = int($1);
        }
        if ($line =~ /Type:\s+(.+)/) {
            $hash->{type} = $1;
        }
    }

    return $hash;
}

################################################################################
# Subroutine Name: set_port_type()
# Objective:       NO-OP for Apcon
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_type {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_ser_err("Unsupported action requested for Apcon switch");

    return $SwitchUtil::FALSE;
}

################################################################################
# Subroutine Name: port_dump_info()
# Objective:       NO-OP for Apcon
# Details:         See POD documentation at the end of this file
################################################################################
sub port_dump_info {
    my ($self, $port) = @_;

    $self->_clear_err();

    return $SwitchUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new port object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/i) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # initialize the object members
    $self->{parent} = $parent;

    $self->{nameserver} = $self->get_nameserver();

    # The initial_active_connections is a copy of all active connections
    # available on the Apcon switch. This copy is used to determine what
    # connections were available after disconnecting a port, reconnecting
    # to the Apcon switch or any other action that requires disconnect and
    # reconnect to the Apcon switch.
    $self->{initial_active_connections} = $self->{nameserver};

    return $self;
}

###############################################################################
# Method:    _get_nameserver_database
# Objective: Retrieve nameserver database information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_nameserver_database {
    my ($self) = @_;

    my ($cmd, $index, $counter, $num_devices, $output, $line);
    my (@output, $port, $did, $type, $wwpn, $wwnn, $iteration_string);
    my (@lookup_failed_ports);
    my $ns = {};

    $cmd = "show connections";
    @output = $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("error executing command ($cmd)");
        return "";
    }

    # Example show connections output
    # apcon_bh_1>> show connections
    # Locked? Port Name:          Port# Dir Port# Port Name:         Locked?
    # -       50060b0000bdd6ee    D13   <>  D12                      -
    # -       50060b0000bdd6e4    C13   <>  C12                      -
    # -       500110a00011b72c    B13   <>  B12                      -
    #
    # May also look like this
    # vmshe-apcon10g-01>> show connections
    # Locked? Port Name:          Port# Dir Port# Port Name:         Locked?
    # -        10000000c99da4b9 -( A10   <>  A04   10000000c99da4b9 -(-
    # -        10000000c99da4b7 -( A09   <>  A03   10000000c99da4b7 -(-
    # -        10000000c97bc545 -( A08   <>  A02   10000000c97bc545 -(-
    # -        10000000c97bc544 -( A07   <>  A01   10000000c97bc544 -(-

    # Parse through the show connections output and retrieve all connections
    foreach my $line (@output) {
        my $port = "";
        my $wwpn = "";
        my $conn_port = "";

        if ($line =~ /Locked/i) {
            next;
        } elsif ($line =~ /(\S{16})\s+(\S+)\s+\<\>\s+(\S+)\s+/) {
            $wwpn = $1;
            $port = $2;
            $conn_port = $3;
        } elsif ($line =~ /(\S{16})\s+.+\s+(\w\d+)\s+\<\>\s+(\w\d+)\s+/) {
            $wwpn = $1;
            $port = $2;
            $conn_port = $3;
        } else {
            next;
        }

        # Check to make sure the wwpn is valid
        if ($wwpn eq "-") {
            next;
        }

        $wwpn =~ s/\://g;
        $wwpn = lc($wwpn);
        $ns->{$port}->{wwpn} = $wwpn;
        $ns->{$port}->{conn_port} = $conn_port;
    }

    return $ns;

}

1;
