# Copyright (c) 2007-2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for ports for a Brocade FC switch
## @author thomasm@netapp.com
## @status private
## @pod ./Port.pod

# Package Name
##############
package San::SanSwitch::FC::PortMgr::BrocadeFCPort;

use strict;
use Tharn;
# Package Requirements
######################
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::PortMgr;
use San::SanLog;
use Error;
use Data::Dumper;
use File::Path qw(make_path remove_tree);
use NATE::Log qw(log_global);

use vars qw(@ISA);

@ISA = qw(Error SanLog PortMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: get_port_num()
# Objective:       Retrieve the port number/interface corresponding to a
#                  WWPN, by returning the first match in the array.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_num {
    my ($self, %args) = @_;
    my (@ports);

    @ports = $self->get_port_nums(%args);
    return "" if ($self->error);

    return $ports[0];

}

################################################################################
# Subroutine Name: get_port_nums()
# Objective:       Retrieve the all port number/interface corresponding to a
#                  WWPN. Multiple's can be returned when using looking for
#                  e-ports and a trunk is established.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_nums {
    my ($self, %args) = @_;

    my ($wwpn, $found_port, @found_ports);
    my (@possible_fids, $found_on_fid);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwpn/) {
            if (not length($wwpn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-wwpn'");
                return ();
            }
            next;
        }
    }

    # make sure you have a valid WWPN
    if (!defined($wwpn)) {
        $self->_set_err("WWPN must be specified");
        return ();
    }

    # If virtual fabrics are enabled, get the list of
    # possible FIDs to look in. If not, assume there is
    # just the single traditional switch to look at.
    if ($self->{parent}->{vf_status} == 1) {
        @possible_fids = $self->{parent}->get_vf_ids();
    } else {
        @possible_fids = ( "" );
    }

    # Cycle though each FID for possible matches, kick out
    # when match is found.
    foreach my $fid (@possible_fids) {
        # set FID context if using virtual fabrics
        if ($self->{parent}->{vf_status} == 1) {
            $found_on_fid = $fid;
            $self->{parent}->set_vf_context($fid);
            if ($self->{parent}->error) {
                $self->_set_err("Could not set virtual fabric context to $fid");
                return ();
            }
        }

        $self->{nameserver} = $self->get_nameserver();
        if ($self->{parent}->{vf_status} == 1) {
            $self->sanlog(-msg => "Searching for WWPN $wwpn on virtual fabric ID ($fid)",
                          -msg_level => 0);
        } else {
            $self->sanlog(-msg => "Searching for WWPN $wwpn",
                          -msg_level => 0);
        }
        $wwpn = lc($wwpn);
        $wwpn =~ s/://g;
        foreach my $interface (keys %{$self->{nameserver}}) {
            # NPIV - port could be in format <port>_<instance>. Remove instance
            $interface =~ s/\_\d+//;

            my $ns_wwpns = $self->{nameserver}{$interface}{wwpn};
            # in NPIV case you can have multiple WWPNs on an interface
            my @wwpns = split(/;/, $ns_wwpns);
            if (grep(/^\s*$wwpn\s*$/, @wwpns) > 0) {
                $found_port = $interface;
                if ($self->{parent}->{vf_status} == 1) {
                    $self->sanlog(-msg => "Found match on port $found_port in virtual fabric ID ($fid)",
                                  -msg_level => 0);
                } else {
                    $self->sanlog(-msg => "Found match on port $found_port",
                                  -msg_level => 0);
                }
                push (@found_ports, $found_port);
                last;
            }
        }

        # found match on this virtual fabric context, kick out.
        if ($self->{parent}->{vf_status} == 1) {
            last if (@found_ports);
        }
    }

    # looks to see if we found any ports and report accordingly.
    if (!(@found_ports)) {
        $self->_set_err("No matching port found (including any virtual fabrics - if applicable)");
        return ();
    } else {
        return @found_ports;
    }
}

################################################################################
# Subroutine Name: port_disable()
# Objective:       Disable the port/interface specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_disable {
    my ($self, $port) = @_;

    my ($cmd, $status, $ns, $protocol, @login_show);

    $self->_clear_err();

    # check to see if the protocol of the port is FCoE
    # and if fcoe_port_block is set, if it is use the
    # fcoe command on the switch to do the port block
    $cmd = "portdisable $port";
    $protocol = $self->{$port}->{protocol};
    if ($protocol =~ /FCoE/i) {
        # show port login information
        my $login_show = "fcoe --loginshow $port";
        @login_show = $self->{parent}->execute_brocade($login_show);
        if ($self->error) {
            $self->_set_err("Error: BrocadeFCPort::port_disable() ".
                            "error getting FCoE login information ($port)");
            return $SwitchUtil::FALSE;
        }

        # set FCoE command for execution of block (virutal FC)
        $cmd = "fcoe --disable $port";
    } else {
        # show port login information - do not validate call
        # this could be an older FOS w/o this command
        my $login_show = "portloginshow $port";
        @login_show = $self->{parent}->execute_brocade($login_show);
    }

    # if requested, get current state to store speed and topology
    if ($self->{parent}->{verify_on_enable} == 1) {
        my $port_state = $self->get_port_info($port);
        if ($self->error) {
            my @msg = $self->errmsg();
            $self->_set_err("error getting port information ($port)\n" .
                            "Details: @msg");
            return $SwitchUtil::FALSE;
        }
        $self->{$port}->{previous_speed} = $port_state->{speed};
        $self->{$port}->{previous_type} = $port_state->{type};
    }

    # If FCoE choose which type of block to do - virtual FC layer or
    # ethernet layer.
    if (($protocol =~ /FCoE/i and $self->{parent}->{fcoe_ethernet_block} == 0) or $protocol !~ /FCoE/i) {
        # block the specified port (FC)
        $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error) {
            $self->{$port}->{previous_speed} = undef;
            $self->{$port}->{previous_type} = undef;
            $self->_set_err("Error: BrocadeFCPort::port_disable() ".
                            "error executing command ($cmd)");
            return $SwitchUtil::FALSE;
        }
    } elsif ($protocol =~ /FCoE/i and $self->{parent}->{fcoe_ethernet_block} == 1) {
        # block the Ethernet port
        $self->_ethernet_block($port, 'disable', @login_show);
        if ($self->error) {
            my @msg = $self->errmsg();
            $self->{$port}->{previous_speed} = undef;
            $self->{$port}->{previous_type} = undef;
            $self->_set_err("Error: BrocadeFCPort::port_disable() ".
                            "error blocking ethernet port(s)\n" .
                            "Reason: @msg");
            return $SwitchUtil::FALSE;
        }
    }

    # if asked to (default) validate port went offline
    if ($self->{parent}->{validate_offline} == 1) {
        # wait for the port to transition to offline for FC ports, or report error
        if (($protocol =~ /FCoE/i and $self->{parent}->{fcoe_ethernet_block} == 0) or $protocol !~ /FCoE/i) {
            $status = $self->waitfor_port($port, "offline");
            if ($status == $SwitchUtil::FALSE) {
                $self->_set_err("Error: BrocadeFCPort::port_disable() " .
                                "failed to disable port, port still appears up");
                return $SwitchUtil::FALSE;
            }
        }

        # If FCoE (checked in method), look to see if we transition
        # in terms of connected ports.
        if ($protocol =~ /FCoE/i and $self->{parent}->{fcoe_ethernet_block} == 1) {
            $status = $self->_validate_fcoe_port_state($port, "offline");
            if ($status == $SwitchUtil::FALSE) {
                $self->_set_err("Error: BrocadeFCPort::port_disable() " .
                                "failed to disable port from FCoE perspective, port still appears connected");
                return $SwitchUtil::FALSE;
            }
        }
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_enable()
# Objective:       Enable the port specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_enable {
    my ($self, $port) = @_;

    my ($ns, $protocol, $cmd, $status, @output);

    $self->_clear_err();

    # check to see if the protocol of the port is FCoE
    # and if fcoe_port_block is set, if it is use the
    # fcoe command on the switch to do enable the port
    $protocol = $self->{$port}->{protocol};
    if ($protocol =~ /FCoE/i) {
        $cmd = "fcoe --enable $port";
    } else {
        $cmd = "portenable $port";
    }

    # If FCoE choose which type of block to do - virtual FC layer or
    # ethernet layer.
    if (($protocol =~ /FCoE/i and $self->{parent}->{fcoe_ethernet_block} == 0) or
        $protocol !~ /FCoE/i) {
        $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: BrocadeFCPort::port_enable() " .
                            "error executing command ($cmd)");
            return $SwitchUtil::FALSE;
        }
    } elsif ($protocol =~ /FCoE/i and $self->{parent}->{fcoe_ethernet_block} == 1) {
        # enable the Ethernet port
        $self->_ethernet_block($port, 'enable', ());
        if ($self->error) {
            my @msg = $self->errmsg();
            $self->_set_err("Error: BrocadeFCPort::port_enable() ".
                            "error enabling ethernet port(s)\n" .
                            "Reason: @msg");
            return $SwitchUtil::FALSE;
        }
    }

    # add explicit sleep into the code here before checking the
    # port to see if it is online (suggested by Dave/Sunil)
    sleep (3);

    # wait for the port to transition to online, or report error
    if (($protocol =~ /FCoE/i and $self->{parent}->{fcoe_ethernet_block} == 0) or $protocol !~ /FCoE/i) {
        $status = $self->waitfor_port($port, "online");
        if ($status == $SwitchUtil::FALSE) {
            $self->_set_err("Error: BrocadeFCPort::port_enable() " .
                            "failed to enable port, port still appears down");
            return $SwitchUtil::FALSE;
        }
    }

    # If FCoE (checked in method), look to see if we transition
    # in terms of connected ports.
    if ($protocol =~ /FCoE/i and $self->{parent}->{fcoe_ethernet_block} == 1) {
        $status = $self->_validate_fcoe_port_state($port, "online");
        if ($status == $SwitchUtil::FALSE) {
            $self->_set_err("Error: BrocadeFCPort::port_enable() " .
                            "failed to enable port from FCoE perspective, port does not appear connected");
            return $SwitchUtil::FALSE;
        }
    }

    # compare values now (after enable) versus before the
    # port_disable
    $self->check_port_settings($port);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # show port login information
    if ($protocol =~ /FCoE/i) {
        my $login_show = "fcoe --loginshow $port";
        $self->{parent}->execute_brocade($login_show);
        if ($self->error) {
            $self->_set_err("Error: BrocadeFCPort::port_enable() ".
                            "error getting FCoE login information ($port)");
            return $SwitchUtil::FALSE;
        }
    } else {
        # show port login information - do not validate call
        # this could be an older FOS w/o this command
        my $login_show = "portloginshow $port";
        $self->{parent}->execute_brocade($login_show);
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: set_port_speed()
# Objective:       Set the specified port to the specified speed
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_speed {
    my ($self, %args) = @_;
    my ($status, $port, $speed, $cmd);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Error: BrocadeFCPort::set_port_speed() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?speed/) {
            if ((not length($speed = $args{$_})) &&
                ($args{$_} !~ /^\d$/ || $speed !~ /auto/i)) {
                $self->_set_err("Error: BrocadeFCPort::set_port_speed() " .
                                "Invalid value for argument '-speed'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # handle auto speed setting
    if ($speed =~ /AUTO/i) {
        $speed = 0;
    }

    if ($self->{parent}->get_firmware_version() =~ /^v3/) {
        $cmd = "portCfgSpeed $port, $speed";
    } else {
        $cmd = "portCfgSpeed $port $speed";
    }
    $self->{parent}->execute_brocade($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error: BrocadeFCPort::set_port_speed() " .
                        "error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: BrocadeFCPort::set_port_speed() " .
                        "Port state did not transtion to online state");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: get_port_info()
# Objective:       Get information about the specified port/interface
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_info {
    my ($self, $port) = @_;

    my ($cmd, $lx_mode, @switchShow, @output, @sorted, @speeds);

    $self->_clear_err();

    # Initialize hash
    my $hash = {};

    # NPIV case, remove instance from port
    # could be in <port>_<instance> format
    $port =~ s/\_\d+//;

    # general fields
    $hash->{state}            = "unknown";
    $hash->{state_reason}     = "";
    $hash->{speed}            = 0;
    $hash->{type}             = "unknown";
    $hash->{fcid}             = 0;
    $hash->{min_speed}        = 0;
    $hash->{max_speed}        = 0;

    # buffershow fields
    $hash->{lx_mode}          = "L0";
    $hash->{max_resv_buffers} = 0;
    $hash->{link_distance}    = 0;

    # sfpshow fields
    $hash->{sfp_speed_capabilities} = "";
    $hash->{sfp_max_9u_meters}      = 0;
    $hash->{sfp_max_50u_meters}     = 0;
    $hash->{sfp_max_625u_meters}    = 0;
    $hash->{sfp_vendor_name}        = 0;
    $hash->{sfp_part_number}        = 0;
    $hash->{sfp_wavelength}         = 0;

    # call portshow on the specified port
    $cmd = "portShow $port";
    @output = $self->{parent}->execute_brocade($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error: BrocadeFCPort::get_port_info() " .
                        "error executing command ($cmd)");
        return {};
    }

    foreach my $line (@output) {
        if ($line =~ /Speed:\s+(\S+)Gbps/i) {
            # remove fact that port could have negotiated to this speed
            # i.e N8Gbps
            my $raw_speed = $1;
            $raw_speed =~ s/^N//;
            $hash->{speed} = int($raw_speed);
        }
        if ($line =~ /portDisableReason:\s+(\S+)/i) {
            $hash->{state_reason} = $1;
        }
        if ($line =~ /portState:\s+\d+\s+(\S+)/) {
            $hash->{state} = &SwitchUtil::normalize_state($1);
        }
        if ($line =~ /portId:\s+(\S+)/i) {
            $hash->{fcid} = hex($1);
        }
    }

    # type from port show is not reliable (does not show
    # FL_port for instance). So, get this from the nameserver.
    my $ns = $self->get_nameserver();
    return {} if ($self->error);
    my $type = $ns->{$port}->{type};
    # SwitchShow uses -Port and other Brocade operations _Port
    # convert to _Port naming
    $type =~ s/\-Port/\_Port/i;
    $hash->{type} = &SwitchUtil::normalize_type($type);
    return {} if ($self->error);

    # Load the cached values if we have already looked these values up before
    $hash->{min_speed} = $self->{cache}{min_speed}{$port}
        if defined($self->{cache}{min_speed}{$port});
    $hash->{max_speed} = $self->{cache}{max_speed}{$port}
        if defined($self->{cache}{max_speed}{$port});

    # If we have the default or invalid values, we either have not looked up the
    # capabilities of this port before or the previous look up failed
    if ($hash->{min_speed} == 0 or $hash->{max_speed} == 0) {
        # Look up supported speed (to determine min/max) based on the
        # model of the switch.
        my ($min, $max);
        if ($ns->{$port}->{protocol} !~ /FCoE/i) {
            my ($min, $max) = $self->_get_supported_speeds($port);
            if ($self->error) {
                my @msg = $self->errmsg();
                $self->_set_err("Error: BrocadeFCPort::get_port_info() " .
                                "error getting supported min/max speed (@msg)");
                return {};
            }
        } else {
            ($min, $max) = (10, 10);
        }

        # update values
        $hash->{min_speed} = $min;
        $hash->{max_speed} = $max;

        # update cache
        $self->{cache}{min_speed}{$port} = $min;
        $self->{cache}{max_speed}{$port} = $max;
    }

    # get lx_mode, max_resv_buffers, link_distance from
    # portbuffershow - note, this is 4x FW and greater (not on 3.x)
    if ($self->{parent}->get_firmware_version() !~ /^v3/) {
        if (defined($self->{parent}->{cache}{buffer_show}{$port})) {
            my @output = @{$self->{parent}->{cache}{buffer_show}{$port}};
        } else {
            $cmd = "portbuffershow";
            @output = $self->{parent}->execute_brocade($cmd);
            if ($self->{parent}->error) {
                $self->_set_err("Error: BrocadeFCPort::get_port_info() " .
                                "error executing command ($cmd)");
                return {};
            }

            # Cache values to minimize repeat lookups
            $self->{parent}->{cache}{buffer_show}{$port} = [ @output ];

            foreach my $line (@output) {
                $line =~ s/^\s+//;
                my ($b_port, $p_type, $lx_mode, $max_resv,
                    $buf_usage, $needed_buf, $link_distance,
                    $remaining) = split(/\s+/, $line);
                if ($port eq $b_port) {
                    if ($lx_mode !~ /-/) {
                        $hash->{lx_mode} = $lx_mode;
                    }
                    if ($max_resv !~ /-/) {
                        $hash->{max_resv_buffers} = $max_resv;
                    }
                    if ($link_distance !~ /-/) {
                        $hash->{link_distance}    = $link_distance;
                    }
                    last;
                }
            }
        }
    } else {
        # portbuffershow does not exist on v3x FW, so recommended
        # way was to get from portShow output
        if (defined($self->{parent}->{cache}{switch_show})) {
            my @switchShow = @{$self->{parent}->{cache}{switch_show}};
        } else {
            $cmd = "switchshow";
            my @switchShow = $self->{parent}->execute_brocade($cmd);
            if ($self->{parent}->error) {
                $self->_set_err("Error: BrocadeFCPort::get_port_info() " .
                                "error executing command ($cmd)");
                return {};
            }

            # Cache values to minimize repeat lookups
            $self->{parent}->{cache}{switch_show} = [ @switchShow ];
        }

        # traverse switchshow output to look for distance on
        # E-port and return correct distance characteristic
        foreach my $line (@switchShow) {
            if ($line =~ /port\s+$port\:/i) {
                if ($line =~ /L0/) {
                    $hash->{lx_mode} = "L0";
                } elsif ($line =~ /LE/) {
                    $hash->{lx_mode} = "LE";
                } elsif ($line =~ /L1/) {
                    $hash->{lx_mode} = "L1";
                } elsif ($line =~ /L2/) {
                    $hash->{lx_mode} = "L2";
                } else {
                    $hash->{lx_mode} = "L0";
                }
                last;
            }
        }
    }

    # Load the cached sfp information if we have already looked these
    # values up before
    $hash->{sfp_speed_capabilities} = $self->{cache}{sfp_speed_capabilities}{$port}
        if defined($self->{cache}{sfp_speed_capabilities}{$port});
    $hash->{sfp_max_9u_meters} = $self->{cache}{sfp_max_9u_meters}{$port}
        if defined($self->{cache}{sfp_max_9u_meters}{$port});
    $hash->{sfp_max_50u_meters} = $self->{cache}{sfp_max_50u_meters}{$port}
        if defined($self->{cache}{sfp_max_50u_meters}{$port});
    $hash->{sfp_max_625u_meters} = $self->{cache}{sfp_max_625u_meters}{$port}
        if defined($self->{cache}{sfp_max_625u_meters}{$port});
    $hash->{sfp_vendor_name} = $self->{cache}{sfp_vendor_name}{$port}
        if defined($self->{cache}{sfp_vendor_name}{$port});
    $hash->{sfp_part_number} = $self->{cache}{sfp_part_number}{$port}
        if defined($self->{cache}{sfp_part_number}{$port});
    $hash->{sfp_wavelength} = $self->{cache}{sfp_wavelength}{$port}
        if defined($self->{cache}{sfp_wavelength}{$port});

    # if not cached, call sfpshow and fill in data.
    if (not defined($self->{cache}{sfp_speed_capabilities}{$port}) or
        not defined($self->{cache}{sfp_max_9u_meters}{$port}) or
        not defined($self->{cache}{sfp_max_50u_meters}{$port}) or
        not defined($self->{cache}{sfp_max_625u_meters}{$port}) or
        not defined($self->{cache}{sfp_vendor_name}{$port}) or
        not defined($self->{cache}{sfp_part_number}{$port}) or
        not defined($self->{cache}{sfp_wavelength}{$port})) {

        # gather SFP information for the specified port
        $cmd = "sfpshow $port";
        @output = $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: BrocadeFCPort::get_port_info() " .
                            "error executing command ($cmd)");
            return {};
        }

        # take info consideration FCoE and no SFP
        if (grep(/does\s+not\s+use\s+an\s+SFP/, @output) == 0) {
            my ($sfp_speed_capabilities) = (grep(/Transceiver:/, @output))[0];
            $sfp_speed_capabilities =~ s/Transceiver:\s*//;

            # get speeds. In format:
            # Transceiver: 150c402001000000 100,200,400_MB/s M5,M6 sw Inter_dist
            $sfp_speed_capabilities = (split(/_MB/, $sfp_speed_capabilities))[0];
            $sfp_speed_capabilities =~ s/\s*\S+\s+//;

            # get speeds and divide by 100 to store (1,2,4,8) formatted values
            my @tmp = split(/\,/, $sfp_speed_capabilities);
            foreach my $val (@tmp) {
                $val = int($val / 100);
            }

            # generate ':' separated list of speeds
            $sfp_speed_capabilities = join(':', @tmp);
            $hash->{sfp_speed_capabilities} = $sfp_speed_capabilities;
            $self->{cache}{sfp_speed_capabilities}{$port} = $sfp_speed_capabilities;

            my ($sfp_max_9u_meters) = (grep(/Length 9u:.*meters/, @output))[0];
            $sfp_max_9u_meters =~ s/Length 9u:\s*//;
            $sfp_max_9u_meters =~ s/\s+\(.*\)//;
            $hash->{sfp_max_9u_meters} = $sfp_max_9u_meters;
            $self->{cache}{sfp_max_9u_meters}{$port} = $sfp_max_9u_meters;

            my ($sfp_max_50u_meters) = (grep(/Length 50u:.*meters/, @output))[0];
            $sfp_max_50u_meters =~ s/Length 50u:\s*//;
            $sfp_max_50u_meters =~ s/\s+\(.*\)//;
            $hash->{sfp_max_50u_meters} = $sfp_max_50u_meters;
            $self->{cache}{sfp_max_50u_meters}{$port} = $sfp_max_50u_meters;

            my ($sfp_max_625u_meters) = (grep(/Length 62\.5u:.*meters/, @output))[0];
            $sfp_max_625u_meters =~ s/Length 62\.5u:\s*//;
            $sfp_max_625u_meters =~ s/\s+\(.*\)//;
            $hash->{sfp_max_625u_meters} = $sfp_max_625u_meters;
            $self->{cache}{sfp_max_625u_meters}{$port} = $sfp_max_625u_meters;

            my ($sfp_vendor_name) = (grep(/Vendor Name:/, @output))[0];
            $sfp_vendor_name =~ s/Vendor Name:\s*//;
            $hash->{sfp_vendor_name}= $sfp_vendor_name;
            $self->{cache}{sfp_vendor_name}{$port} = $sfp_vendor_name;

            my ($sfp_part_number) = (grep(/Vendor PN:/, @output))[0];
            $sfp_part_number =~ s/Vendor PN:\s*//;
            $hash->{sfp_part_number} = $sfp_part_number;
            $self->{cache}{sfp_part_number}{$port} = $sfp_part_number;

            my ($sfp_wavelength) = (grep(/Wavelength:/, @output))[0];
            $sfp_wavelength =~ s/Wavelength:\s*//;
            $sfp_wavelength =~ s/\s+\(.*\)//;
            $hash->{sfp_wavelength} = $sfp_wavelength;
            $self->{cache}{sfp_wavelength}{$port} = $sfp_wavelength;
        }
    }

    return $hash;
}

################################################################################
# Subroutine Name: set_port_type()
# Objective:       Set the specified port to the specified port type
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_type {
    my ($self, %args) = @_;
    my ($status, $cmd, $port, $type, @types);

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Error: BrocadeFCPort::set_port_type() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: BrocadeFCPort::set_port_type() " .
                                "Invalid value for argument '-type'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # receive a normalized port and here we convert to Brocade specific
    $type = $self->_get_port_type_from_normalized($type);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # make sure requested port type to change to is valid
    @types = $self->_get_supported_port_types($port);
    unless (grep(/$type/i, @types)) {
        $self->_set_err("Error: BrocadeFCPort::set_port_type() " .
                        "Port type specified is not supported on this switch");
        return $SwitchUtil::FALSE;
    }

    # if current and asked for type are same, just return
    my $prev_type = $self->get_port_info($port);
    if ($prev_type->{type} eq $type) {
        $self->sanlog(-msg => "BrocadeFCPort::set_port_type() Port type is already set to $prev_type->{type}, no action needed");
        return $SwitchUtil::TRUE;
    }

    # run twice through, first time to unlock the
    # port from the specific type (portcfgdefault), and
    # next time to set/lock to the appropriate type
    my ($cmd_part);
    if ($self->{parent}->get_firmware_version() =~ /^v3/) {
        $cmd_part = "$port, ";
    } else {
        $cmd_part = "$port ";
    }

    # determine the command based on type
    if ($type =~ /E_Port/i) {
        $cmd = "portcfgeport $cmd_part 1";
    } elsif ($type =~ /EX_Port/i) {
        $cmd = "portcfgexport $cmd_part 1";
    } elsif ($type =~ /G_Port/i) {
        $cmd = "portcfggport $cmd_part 1";
    } elsif ($type =~ /L_Port/i) {
        $cmd = "portcfglport $cmd_part 1";
    } elsif ($type =~ /N_Port/i) {
        $cmd = "portcfgnport $cmd_part 1";
    } elsif ($type =~ /VEX_Port/i) {
        $cmd = "portcfgvexport $cmd_part 1";
    } elsif ($type =~ /F_Port/i or $type =~ /auto/i) {
        $cmd = undef;
    }

    # changing to F_Port is just the single call
    # to portcfgdefault.
    my (@cmds);
    if (defined $cmd) {
        @cmds = ( "portcfgdefault $port", "$cmd" );
    } else {
        @cmds = ( "portcfgdefault $port" );
    }

    foreach my $cmd (@cmds) {
        my @output = $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: BrocadeFCPort::set_port_type() " .
                            "error executing command ($cmd)");
            return $SwitchUtil::FALSE;
        }
    }

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: BrocadeFCPort::set_port_type() " .
                        "Port state did not transtion to online state");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_dump_info()
# Objective:       Dump the info for the provided port.
# Details:         See POD documentation at the end of this file
################################################################################
sub port_dump_info {
    my ($self, $port, $debug, %dump_info) = @_;

    my ($protocol , $cmd);

    $self->_clear_err();

    $self->sanlog(-msg => "BrocadeFCPort::port_dump_info() Dumping log for port $port");

    # If we are FCoE, show the loginshow output for the port.
    $protocol = $self->{$port}->{protocol};
    $self->sanlog(-msg => "BrocadeFCPort::port_dump_info() Dumping log for port $port");

    if ($protocol =~ /FCoE/i) {
        # show port login information
        my $login_show = "fcoe --loginshow $port";
        $self->{parent}->execute_brocade($login_show);
        if ($self->error) {
            $self->_set_err("Error: BrocadeFCPort::port_dump_info() " .
                            "error getting FCoE login information ($port)");
            return $SwitchUtil::FALSE;
        }
    } else {
        # show port login information - do not validate call
        # this could be an older FOS w/o this command
        my $login_show = "portloginshow $port";
        $self->{parent}->execute_brocade($login_show);
    }

    # Get the nameserver information
    my $ns = $self->get_nameserver();
    if ($self->error) {
        $self->_set_err("Error: BrocadeFCPort::port_dump_info() " .
                        "Error getting nameserver information");
        return $SwitchUtil::FALSE;
    }

    # Both FCoE Virtual and FC Physical use the same dump command.  It only
    # differs by port or index depending on if you are using a director or not.
    if ($ns->{$port}->{index} eq "NA") {
        # Run "portlogdumpport $port" on the switch port
        $cmd = "portlogdumpport $port";
    } else {
        # Run "portlogdumpport $index" on the director index
        $cmd = "portlogdumpport $ns->{$port}->{index}";
    }

    # We only dump the port if it is FC or Virtual FCoE.  Per Brocade, there
    # is currently no equivalent cmsh command to dump 10gig port information.
    if (($protocol =~ /FCoE/i and $self->{parent}->{fcoe_ethernet_block} == 0) or
        $protocol !~ /FCoE/i) {
        $self->sanlog(-msg => "BrocadeFCPort::port_dump_info() Dumping log for port $port");
        $self->{parent}->execute_brocade($cmd, 120);
        if ($self->{parent}->error) {
            $self->_set_err("Error: BrocadeFCPort::port_dump_info() " .
                            "error executing command ($cmd)");
            return $SwitchUtil::FALSE;
        }
    }

    # Run the supportsave command on the switch if the flag is set.
    if ($debug == 1) {

        # Save the supportsave logs with the nate logs
        $dump_info{protocol} = "scp";
        my $switch_name = $self->{parent}->hostname();
        unless(defined($switch_name) and $switch_name ne "") {
              $self->_set_err("Error: BrocadeFCPort::port_dump_info() " .
                              "Unable to retrieve the switch name from the switch object");
              return $SwitchUtil::FALSE;
        }
        my $logdir_temp = &Tharn::param("LOGDIR");
        $dump_info{remote_dir} = $logdir_temp . "/Supportsave_logs/$switch_name-$port/";
        $self->sanlog(-msg => "BrocadeFCPort::port_dump_info() remote_dir = $dump_info{remote_dir}");
        if ((-d $dump_info{remote_dir}) < 1) {
            if ((make_path(($dump_info{remote_dir})) < 1)) {
                $self->_set_err("Error: BrocadeFCPort::port_dump_info() " .
                                 "Error creating directory $dump_info{remote_dir}");
                return $SwitchUtil::FALSE;
            }
       }

       my $output = $self->{parent}->supportsave(%dump_info);
       if ($self->{parent}->error) {
            $self->_set_err("Error: BrocadeFCPort::port_dump_info() " .
                             "Supportsave failed");
            return $SwitchUtil::FALSE;
        }
    }

    # additional request to run nsshow -r
    # do not validate call this could be an older FOS w/o this command
    $cmd = "nsshow -r";
    $self->{parent}->execute_brocade($cmd, 120);

    return $SwitchUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new port object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: BrocadeFCPort::_new() " .
                                "Invalid value for argument '-parent'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/i) {
            $self->_set_err("Error: BrocadeFCPort::_new() " .
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: BrocadeFCPort::_new() " .
                        "Missing required argument '-parent'");
        return $self;
    }

    # initialize the object members
    $self->{parent} = $parent;

    $self->{nameserver} = $self->get_nameserver();

    return $self;
}

###############################################################################
# Method:    _get_nameserver_database
# Objective: Retrieve nameserver database information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_nameserver_database {
    my ($self, $all_vf) = @_;

    my ($cmd, $index, $counter, $num_devices, $output, $line);
    my (@output, $port, $did, $type, $wwpn, $wwnn, $iteration_string);
    my ($protocol, $tmp_output, @lookup_failed_ports, $symb, $sym_type);
    my (@fids, $npiv_port, $check_all_vfs, $port_index, $ns_port_index);

    # check to see if we want to get nameserver entries from all
    # virtual fabrics, or only the current context
    if (defined($all_vf) and $all_vf == 1 and
        $self->{parent}->{vf_status} == 1) {
        $check_all_vfs = 1;
    } else {
        $check_all_vfs = 0;
    }

    # invalidate the switchShow cache, since we want to do
    # an initial lookup with fresh information. We can use the
    # cache in the context of the function, but when leaving,
    # we also want to invalidate.
    $self->{parent}->{cache}{switch_show} = undef;

    # Get the NPV mode
    my $switch_ad_hash = {};
    unless ($self->{parent}->is_npv_enabled()) {
        # Grab the switch domain since we are not in access gateway (npv) mode.
        my $switch_domain = $self->{parent}->domain_id();
        if ($switch_domain eq "") {
            $self->_set_err("Unable to retrieve the domain id for the switch");
            return {};
        }

        # If virtual fabrics are not enabled, load up the admin domain hash
        # so we can document what AD(s) a port belongs to.
        my $ad_hash = {};
        unless ($self->{parent}->is_vf_enabled()) {
            # Make sure admin domains are enabled
            if ($self->{parent}->is_ad_enabled()) {
                # Looks like we are good to go
                $ad_hash = $self->{parent}->get_admin_domain_database();
                if ($self->{parent}->error) {
                    $self->_set_err("Unable to retrieve the admin domain database hash:\n",
                                    $self->{parent}->errmsg());
                    return {};
                }

                # Pull out the admin domain sub hash for the switch DID.  We should end up
                # with a hash that looks like this:
                #
                # switch_ad_hash = { AD50 => {
                #                               1 => 1,
                #                               2 => 1,
                #                               6 => 1,
                #                               10 => 1,
                #                             },
                #                     AD51 => {
                #                               3 => 1,
                #                               4 => 1,
                #                             },
                #                     AD52 => {
                #                               1 => 1,
                #                               4 => 1,
                #                             },
                #                   }
                if (defined($ad_hash->{by_switch}->{$switch_domain})) {
                    $switch_ad_hash = $ad_hash->{by_switch}->{$switch_domain};
                }
            }
        }
    }

    # If we have admin domains enabled, we want to switch to AD255
    # prior to getting the rest of the information.  We will use the
    # existence and non-empty $orig_ad later in the method to decide
    # if we are operating on ADs.
    my $orig_ad;
    if ($self->{parent}->is_ad_enabled() and $self->{parent}->{vf_status} != 1) {
        $orig_ad = $self->{parent}->get_current_admin_domain();
        if ($self->{parent}->error) {
            $self->_set_err("Unable to retrieve the current admin domain for the switch.\n",
                            $self->{parent}->errmsg());
            return {};
        }

        unless($orig_ad == 255) {
            # Switch to AD255 since we weren't already there
            $self->{parent}->switch_to_admin_domain("255");
            if ($self->{parent}->error) {
                $self->_set_err("Unable to switch to AD255 on the switch.\n",
                                $self->{parent}->errmsg());
                return {};
            }
        }
    }

    my $ns = {};

    # If we have been requested to get from all FIDS (virtual fabrics)
    # get that list. If not, just have one entry for the physical switch
    # or the current context.
    if ($check_all_vfs == 1 and $self->{parent}->{vf_status} == 1) {
        @fids = $self->{parent}->get_vf_ids();
    } else {
        @fids = ( "" );
    }

    # foreach FID go and get the NS information
    foreach my $fid (@fids) {
        # set FID context if requested
        if ($check_all_vfs == 1) {
            $self->{parent}->set_vf_context($fid);
            if ($self->{parent}->error) {
                $self->_set_err("Error: BrocadeFCPort::_get_nameserver_database() could not set virtual fabric context to $fid");

                # Try to switch back to the original AD if we are using ADs before
                # bailing out.
                if (defined($orig_ad) and $orig_ad ne "") {
                    $self->{parent}->switch_to_admin_domain($orig_ad);
                    if ($self->{parent}->error()) {
                        my $errmsg = join("\n", $self->{parent}->errmsg());
                        $self->sanlog(-msg => "Unable to switch to AD${orig_ad} on the switch:\n" . $errmsg);
                    }
                }
                return {};
            }
        }

        # do this process in 2 phases, first use nsshow to get information
        # from the nameserver, and then use switchShow to get information
        # from the switch about e-ports and include that in the ns hash.
        #
        # Example nsshow output:
        # > nsshow
        #  Type Pid    COS     PortName                NodeName           TTL(sec)
        #  N    010000;    2,3;10:00:00:00:c9:6e:a2:a3;20:00:00:00:c9:6e:a2:a3; na
        # FC4s: FCP
        # NodeSymb: [41] "Emulex LPe11002-M4 FV2.70A5 DV7.3.2_vmw10"
        # Fabric Port Name: 20:00:00:05:1e:03:5d:c3
        #
        # for a disk:
        #  NL   020ab2;      3;50:00:00:e0:1c:46:c3:a1;50:00:00:e0:1c:46:c3:a0; na
        # FC4s: FCP [NETAPP  X276_FAL9E288F10NA05]
        # Fabric Port Name: 20:0a:00:05:1e:59:e8:0a
        # Permanent Port Name: 50:00:00:e0:1c:46:c3:a1

        my $cnt = 0;
        my $num_retries = 5;
        while ($cnt < $num_retries) {
             $cmd = "nsshow";
             @output = $self->{parent}->execute_brocade($cmd);
            if ($self->{parent}->error) {
                $self->_set_err("Error: BrocadeFCPort::_get_nameserver_database() " .
                                "error executing command ($cmd)");

                # Try to switch back to the original AD if we are using ADs before
                # bailing out.
                if (defined($orig_ad) and $orig_ad ne "") {
                    $self->{parent}->switch_to_admin_domain($orig_ad);
                    if ($self->{parent}->error()) {
                        my $errmsg = join("\n", $self->{parent}->errmsg());
                        $self->sanlog(-msg => "Unable to switch to AD${orig_ad} on the switch:\n" . $errmsg);
                    }
                }
                return {};
            }

            # account for the case where the switch it up after a reboot,
            # it can be logged into, but the fabric is not up yet.
            if (grep(/not up yet/i, @output) > 0) {
                sleep(60);
                $cnt++;
            } else {
                $cnt = $num_retries;
            }
        }

        # get rid of uneeded lines of data
        @output = grep(!/^\s+FC4s\:/, @output);
        # scalar format of output array
        $output = join("\n", @output);

        $iteration_string = "The Local Name Server has ";
        (undef, undef, undef, $num_devices) =
                              ($output =~ /((.|\n)*)($iteration_string)(\d+)/);

        # loop for specified number of devices, looking for ports and wwpns
        $counter = 0;
        while ($counter < $num_devices) {
            $counter++;

            # Look through the output for lines the have either N or NL, these
            # will be device listings
            ($line, $output) = ($output =~ /(NL?\s+.*\:\S+\:\S+\:.*)((.|\n)*)/);

            # get PortSymb/NodeSymb for the device (if available)
            $symb = "";
            if ($output =~ /^\s+(PortSymb|NodeSymb)\:.*\n/) {
                ($sym_type, $symb, $output) = ($output =~ /\s+(PortSymb|NodeSymb)\:\s+\[\d+\]\s+\"(.*)\".*((.|\n)*)/);
            }

            # get Port Index (if available - FOS 6.1+)
            $ns_port_index = "";
            if ($output =~ /\s+Port\s+Index\:\s+(\d+).*\n/) {
                ($ns_port_index, $output) = ($output =~ /\s+Port\s+Index\:\s+(\d+).*((.|\n)*)/);
            }

            # Look through the line found and snag some non-whitespace followed
            # by a ;, this is the DID
            ($type, $did, $line) = ($line =~ /(\S+)\s+(\S*\;)((.|\n)*)/);
            $did =~ s/\;//;

            # get the wwpn by using the fact that all our devices are
            # at least class 3
            (undef, $wwpn, $wwnn) = ($line =~ /(3\;)(\S+)\;(\S+)\;/);

            # on directors, you need slot/port to distinquish port. nsshow
            # does not give that to use and we need to covert by using
            # switchShow data.
            $port = $self->_convert_to_port($did, $ns_port_index);
            if ($self->error) {
                $self->_clear_err();
                push (@lookup_failed_ports, $did);
            } else {
                # Convert to lowercase since that's what the other switch vendors do
                $wwpn = lc $wwpn;
                $wwpn =~ s/://g;

                $wwnn = lc $wwnn;
                $wwnn =~ s/://g;

                # this could be a disk on loop where you have multiple
                # entries in NS for each disk. If we see a count > 1, add
                # an index to the port
                $ns->{$port}->{count} = ($ns->{$port}->{count} + 1);
                if ($ns->{$port}->{count} > 1) {
                    $port = $port . "_" . $ns->{$port}->{count};
                    $ns->{$port}->{kind} = "disk";
                }

                # store information in hash
                my $base_port = $port;
                $base_port =~ s/\_\d+//;
                $ns->{$port}->{port} = $base_port;
                $ns->{$port}->{wwpn} = $wwpn;
                $ns->{$port}->{wwnn} = $wwnn;
                $ns->{$port}->{address} = $did;

                # special case setting of NL -> L-Port in
                # case of MC and multiple devices logged
                # since later code is physical port based
                # and will not set type
                if ($ns->{$port}->{kind} eq "disk" and $type =~ /NL/) {
                    $ns->{$port}->{type} = "L-Port";
                } else {
                    $ns->{$port}->{type} = $type;
                }

                $ns->{$port}->{n_type} = $type;
                $ns->{$port}->{symb} = $symb;
                $ns->{$port}->{fid} = $fid;
            }
        }

        # get switch show to get e-ports to round out ns hash
        if (defined($self->{parent}->{cache}{switch_show}) and $check_all_vfs == 0) {
            @output = @{$self->{parent}->{cache}{switch_show}};
        } else {
            $cmd = "switchshow";
            @output = $self->{parent}->execute_brocade($cmd);
            if ($self->{parent}->error) {
                $self->_set_err("Error: BrocadeFCPort::_get_nameserver_database() " .
                                "error executing command ($cmd)");
                # Try to switch back to the original AD if we are using ADs before
                # bailing out.
                if (defined($orig_ad) and $orig_ad ne "") {
                    $self->{parent}->switch_to_admin_domain($orig_ad);
                    if ($self->{parent}->error()) {
                        my $errmsg = join("\n", $self->{parent}->errmsg());
                        $self->sanlog(-msg => "Unable to switch to AD${orig_ad} on the switch:\n" . $errmsg);
                    }
                }
                return {};
            }

            # Cache values to minimize repeat lookups
            if ($check_all_vfs == 0) {
                $self->{parent}->{cache}{switch_show} = [ @output ];
            }
        }

        # Example switchShow (non director):
        # Area Port Media Speed State     Proto
        # =====================================
        #   0   0   id    N4   Online     F-Port  10:00:00:00:c9:6e:a2:a3
        #   1   1   id    N4   Online     F-Port  10:00:00:00:c9:6e:4d:d7
        #   2   2   id    N4   Online     F-Port  10:00:00:00:c9:6e:4e:9d
        #   3   3   id    N4   No_Light
        #
        # FOS 6.3.0+ format (non-director):
        # Index Port Address Media Speed State     Proto
        # ==============================================
        #   0   0   730000   id    N4   Online      FC  F-Port  21:01:00:e0:8b:a6:c2:42
        #   1   1   730100   id    N4   Online      FC  F-Port  21:01:00:e0:8b:a3:03:24
        #   2   2   730200   id    N4   Online      FC  F-Port  21:03:00:e0:8b:e3:03:24
        #  23  23   731700   id    N4   Online      FC  LD E-Port  10:00:00:05:1e:57:b4:86 "sw062113" (upstream)
        #
        # Example switchShow (director):
        # Index Slot Port Address Media Speed State Proto
        # ===================================================
        # 0    1    0   ed0000   id    N2   Online  E-Port  10:00:00:05:1e:44:32:00 "brcddcx_rtp02" (upstream)(Trunk master)
        # 1    1    1   ed0100   id    N2   Online  E-Port  (Trunk port, master is Slot  1 Port  0 )
        #
        # Example switchShow (director (new format - 6.3+)):
        # Index Slot Port Address Media Speed State     Proto
        # ===================================================
        # 152    2  24   1a9840   id    N8  Online      FC  E-Port  10:00:00:05:1e:07:90:a0 "sw062037_92" (upstream)(Trunk master)
        # 153    2  25   1a9940   id    N8  Online      FC  E-Port  (Trunk port, master is Slot  2 Port 24 )
        # 154    2  26   1a9a40   id    N4  Online      FC  E-Port  (Trunk port, master is Slot  2 Port 27 )

        # determine if this is a director
        my $director = 0;
        if (grep(/Index\s+Slot\s+Port/i, @output) > 0) {
            $director = 1;
        }

        # determine if this new FOS6.3.0+ format for non-director
        my $new_edge_format = 0;
        if ((grep(/Index\s+Port/i, @output) > 0) and $director == 0) {
            $new_edge_format = 1;
        }

        foreach my $line (@output) {
            # skip lines that are not port representations
            if ($line !~ /^\s*\d+/ and $line !~ /^\s*port/) {
                next;
            }

            # skip embedded port license issues
            if ($line =~ /No POD License/i) {
                next;
            }

            # remove initial whitepsace.
            $line =~ s/^\s*//;

            # get fields
            my ($e_port_to, $slot, $port, $address, $type, $wwpn);

            # take info consideration long distance eports
            # 12  12   id    4G   Online           LD E-Port  10:00:00:05:1e:0f:50:bf "brcd300-rtp04" (downstream)
            my $e_port_type = "";
            if ($line =~ /\s+(L\w+)\s+E\-Port/i) {
                $e_port_type = $1;
                $line =~ s/L\w+\s+E\-Port/E\-Port/;
            } else {
                $e_port_type = "STANDARD";
            }

            # Grab protocol (if in output) to use later since
            # it effects the split.
            $line =~ m/\s+(FC|FCoE|iSCSI)\s+/i;
            $protocol = $1;
            if ($protocol !~ /FC|FCoE|iSCSI/i) {
                $protocol = "NotAvailable";
            }

            # Determine if the port is NPIV (will need to
            # get connected WWPNs a different way (from
            # portShow)
            $line =~ m/\d+\s+(NPIV)\s+/;
            $npiv_port = $1;

            # or check to see if we have multiple VN-Port logins
            # Or using NPIV on FCoE
            $line =~ m/(\d+)\s+VN-Port\(s\)/;
            my $vn_port_count = $1;
            if ($protocol =~ /FCOE/i and $vn_port_count > 1) {
                $npiv_port = "NPIV";
            }

            # split to grab fields, conditionalize on
            # director/edge.
            my @fields = split(/\s+/, $line);
            if ($director == 1) {
                $port_index = $fields[0];
                $slot = $fields[1];
                $port = $fields[2];
                $address = $fields[3];

                # account for later output where protocol is being printed
                if (defined($protocol) and $protocol =~ /FC|FCoE|iSCSI/i) {
                    $type = $fields[8];
                    $wwpn = $fields[9];
                    $e_port_to = $fields[10];
                } else {
                    $type = $fields[7];
                    $wwpn = $fields[8];
                    $e_port_to = $fields[9];
                }
            } else {
                $port = $fields[1];
                $port =~ s/://;
                $address = "NA";
                $port_index = "NA";

                # account for later output where protocol is being printed
                if (defined($protocol) and $protocol =~ /FC|FCoE|iSCSI/i) {
                    if ($new_edge_format == 0) {
                        $type = $fields[6];
                        $wwpn = $fields[7];
                        $e_port_to = $fields[8];
                    } else {
                        $address = $fields[2];
                        $type = $fields[7];
                        $wwpn = $fields[8];
                        $e_port_to = $fields[9];
                    }
                } else {
                    $type = $fields[5];
                    $wwpn = $fields[6];
                    $e_port_to = $fields[7];
                }
            }

            # if this is a director switch the port
            # should be a representation of slot/port
            if ($director == 1) {
                $port = $slot . "/" . $port;
            }

            # if this is a NPIV port, get the connected
            # WWPNs from output of portShow
            if ($npiv_port =~ /NPIV/i) {
                $ns->{$port}->{npiv} = 1;
                $port =~ s/\_\d+//;
                $cmd = "portShow $port";
                my @port_info = $self->{parent}->execute_brocade($cmd);
                if ($self->{parent}->error) {
                    $self->_set_err("Error: BrocadeFCPort::_get_nameserver_database() " .
                                    "error executing command ($cmd)");

                    # Try to switch back to the original AD if we are using ADs before
                    # bailing out.
                    if (defined($orig_ad) and $orig_ad ne "") {
                        $self->{parent}->switch_to_admin_domain($orig_ad);
                        if ($self->{parent}->error()) {
                            my $errmsg = join("\n", $self->{parent}->errmsg());
                            $self->sanlog(-msg => "Unable to switch to AD${orig_ad} on the switch:\n" . $errmsg);
                        }
                    }

                    return {};
                }

                # from output, grab WWPNs like:
                # portWwn of device(s) connected:
                # \t10:00:00:05:1e:8c:d1:f5
                # \t20:0a:00:05:1e:76:95:80
                my $start = 0;
                my @npiv_wwpns = ();
                foreach my $line (@port_info) {
                    if ($line =~ /^\s*portWwn\s+of\s+device\(s\)\s+connected/i) {
                        $start = 1;
                        next;
                    } elsif ($line =~ /^\s*\w+\:\s/i) {
                        $start = 0;
                    }
                    if ($start == 1) {
                        $line =~ s/\s+//g;
                        # only push if line contains data (WWPN format)
                        push(@npiv_wwpns, $line) if ($line =~ /^\w+\:\w+\:/);
                    }
                }

                # format WWPN output
                if (scalar(@npiv_wwpns) > 0) {
                    $wwpn = join(';', @npiv_wwpns);
                } else {
                    $wwpn = "NONE";
                }

                # log for debug reasons
                $self->sanlog(-msg =>
                              "BrocadeFCPort::_get_nameserver_database() $port is NPIV enable/connected and has the following\n" .
                              "devices connected: $wwpn");
            } else {
                $ns->{$port}->{npiv} = 0;
            }

            # condence wwpn
            $wwpn = lc $wwpn;
            $wwpn =~ s/://g;

            # cleanup e_port_to (remove "")
            $e_port_to =~ s/\"//g;

            # If FCoE port (determined by protocol), add to hash
            # only if NPIV has been determined to be on for that port,
            # Otherwise, it should ne in NS
            if ($protocol =~ /FCoE/i and $npiv_port =~ /NPIV/i) {
                $ns->{$port}->{wwpn} = $wwpn;
                $ns->{$port}->{wwnn} = "NA";
                $ns->{$port}->{address} = $address;
                $ns->{$port}->{type} = $type;
                $ns->{$port}->{protocol} = $protocol;
                $ns->{$port}->{symb} = "";
                $ns->{$port}->{fid} = $fid;
            } elsif ($npiv_port =~ /NPIV/i) {
                # case where this is not specifically FCOE, do not
                # add everything, just update the wwpn field since it
                # now will contain multiple devices
                $ns->{$port}->{wwpn} = $wwpn;
            }

            # if e-port add to hash, for other ports, update
            # hash with the type which is more valid in switchShow
            # output.
            if ($type =~ /E\-Port/) {
                # store information in hash

                # can contain either:
                # 0 0 id  N8   Online  FC  E-Port  10:00:00:05:1e:36:01:62 "brcd48000_rtp1" (upstream)(Trunk master)
                # 1 1 id  N8   Online  FC  E-Port  (Trunk port, master is Port  0 )
                # if no TRUNK port, set WWPN to NA
                if ($wwpn !~ /trunk/i) {
                    $ns->{$port}->{wwpn} = $wwpn;
                    $ns->{$port}->{e_port_to} = $e_port_to;
                } else {
                    $ns->{$port}->{wwpn} = 'NA';
                    $ns->{$port}->{e_port_to} = 'NA';
                }

                $ns->{$port}->{wwnn} = "NA";
                $ns->{$port}->{address} = "NA";
                $ns->{$port}->{type} = $type;
                $ns->{$port}->{symb} = "";
                $ns->{$port}->{protocol} = $protocol;
                $ns->{$port}->{e_port_type} = $e_port_type;
                $ns->{$port}->{fid} = $fid;
            } else {
                if (defined $type) {
                    # special case FC-VI based on portsymb
                    if ($ns->{$port}->{symb} =~ /FCVI/) {
                        $ns->{$port}->{type} = "F_Port-FC_VI";
                    } else {
                        $ns->{$port}->{type} = $type;
                        $ns->{$port}->{protocol} = $protocol;
                    }
                } else {
                    $ns->{$port}->{type} = "offline";
                    $ns->{$port}->{fid} = $fid;
                }
            }

            # Set protocol on object since it won't change
            # we can use w/o grabbing NS all the time
            $self->{$port}->{protocol} = $protocol;

            # Define the index attribute on the port.
            $ns->{$port}->{index} = $port_index;

            # See if we have an admin domain assigned to the port.
            $ns->{$port}->{admin_domain} = [];
            foreach my $ad (sort keys %$switch_ad_hash) {
                # We want to skip AD0 since we never want
                # any library code operating on that AD.
                next if ($ad eq "AD0");

                # Strip off "AD";
                my $ad_number = $ad;
                $ad_number =~ s/^AD//g;

                if (defined($switch_ad_hash->{$ad}->{$port})) {
                    push (@{$ns->{$port}->{admin_domain}}, $ad_number);
                }
            }
        } ##END foreach my $line (@output) {
    } ##END foreach my $fid (@fids) {

    if (@lookup_failed_ports) {
        $self->sanlog(-msg =>
                      "NOTE: following adresses could not be looked up in switchShow output:\n" .
                      "@lookup_failed_ports\n" .
                      "This could be because they are NPIV connected ports\n");
    }

    # Switch back to the original AD if we are using admin domains
    if (defined($orig_ad) and $orig_ad ne "") {
        $self->{parent}->switch_to_admin_domain($orig_ad);
        if ($self->{parent}->error()) {
            $self->_set_err("Unable to switch to AD${orig_ad} on the switch:\n", $self->{parent}->errmsg());
            return {};
        }
    }

    return $ns;
}

###############################################################################
# Method:    _convert_to_port
# Objective: Converts the did (from nsshow data) to a slot/port designation
#            if a director, if not converts address to port number.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _convert_to_port {
    my ($self, $did, $port_index) = @_;

    my ($cmd, @types, @output);

    # can cache switchShow since we are only looking for
    # slot/port and that should not change in test.
    if (defined($self->{parent}->{cache}{switch_show})) {
        @output = @{$self->{parent}->{cache}{switch_show}};
    } else {
        $cmd = "switchshow";
        @output = $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: BrocadeFCPort::_convert_to_slot() " .
                            "error executing command ($cmd)");
            return "";
        }

        # Cache values to minimize repeat lookups
        $self->{parent}->{cache}{switch_show} = [ @output ];
    }

    # Looks to see if switch is a slotted architecture
    # If it is not, just return the port (from the did)
    if (grep(/Index\s+Slot\s+Port/i, @output) == 0) {
        # If this a later release of FOS that stores
        # the port index, simply return it. If it does not
        # it is the older NS output that does not contain
        # and the DID/Address can be decoded.
        if (defined($port_index) and $port_index ne "") {
            return $port_index;
        }

        # decode the address
        my $port = substr($did, 2, 2);
        $port = hex($port);

        return $port;
    }

    # Look through the data to find the did address and return slot/port
    foreach my $line (@output) {
        # skip lines that are not port representations
        if ($line !~ /^\s*\d+/ or
            $line =~ /No POD License/i) {
            next;
        }

        # remove initial whitepsace.
        $line =~ s/^\s*//;

        # split to get fields.
        my @fields = split(/\s+/, $line);
        my $index = $fields[0];
        my $slot = $fields[1];
        my $port = $fields[2];
        my $address = $fields[3];

        # If this a later release of FOS that stores
        # the port index in NS, use it in search.
        if (defined($port_index) and $port_index ne "") {
            if ($port_index == $index) {
                return "$slot/$port";
            }
        } else {
            # check to see if passed in address matches did and then
            # return slot/port. In case of FCoE compare only first
            # 16bit since last 8 address NPIV
            my $addr_short = substr($address, 0, (length($address) - 2));
            my $did_short = substr($did, 0, (length($did) - 2));
            if ($addr_short eq $did_short) {
                return "$slot/$port";
            }
        }
    }

    # If we come out of the loop w/o finding the slot/port and returning
    # data, we have an error.
    $self->_set_err("Error: BrocadeFCPort::_convert_to_slot() " .
                    "Could not find slot/port for address ($did)");
    return "";

}

###############################################################################
# Method:    _get_supported_port_types
# Objective: Retrieve the supported port types for this switch - for Brocade
#            hardcode since no command exists to get the capabilities
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_supported_port_types {
    my ($self, $port) = @_;
    my (@types);

    if (defined($self->{cache}{port_types})) {
        @types = @{$self->{cache}{port_types}};
    } else {
        @types = ("F_Port", "E_Port", "G_Port",
                  "FL_Port", "N_Port", "L_Port",
                  "EX_Port", "VEX_Port", "VF_Port");

        # Cache values to minimize repeat lookups
        $self->{cache}{port_types} = [ @types ];
    }

    return (wantarray()) ? @types : \@types;
}

###############################################################################
# Method:    _get_port_type_from_normalized
# Objective: Take in a normalized port name and return one specific to Brocade
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_port_type_from_normalized {
    my ($self, $type) = @_;

    if ($type eq "F_Port") {
        return $type;
    }

    if ($type eq "FL_Port") {
        return $type;
    }

    if ($type eq "E_Port") {
        return $type;
    }

    if ($type eq "G_Port") {
        return $type;
    }

    if ($type eq "N_Port") {
        return $type;
    }

    if ($type =~ /auto/i) {
        return $type;
    }

    # Brocade only
    if ($type eq "EX_Port") {
        return $type;
    }

    # Brocade only
    if ($type eq "VEX_Port") {
        return $type;
    }

    # for FCoE
    if ($type eq "VF_Port") {
        return "F_Port";
    }

    # if nothing matched set error
    $self->_set_err("Error: BrocadeFCPort::_get_port_type_from_normalized() " .
                    "($type) cannot be mapped to a valid Brocade port type");
    return "";

}

###############################################################################
# Method:    _get_supported_speeds
# Objective: Based on the model of the switch and the port, determine the
#            speed capabilities of the port.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_supported_speeds {
    my ($self, $port) = @_;

    # get the switch model
    my $model = $self->{parent}->get_model();
    if ($self->{parent}->error) {
        my @msg = $self->{parent}->errmsg();
        $self->_set_err("Error: BrocadeFCPort::_get_supported_speeds() " .
                        "Could not determine mode of switch (@msg)");
        return ("", "");
    }

    # start looking at each model to determine speed capabilities
    if ($model =~ /3200|3800|3250|3850|3900|12000|24000/) {
        # 2G switches
        return ("1", "2");
    } elsif ($model =~ /4016|4020|4012|4024|4424/) {
        # 4G embedded blade switches
        return ("1", "4");
    } elsif ($model =~ /200e|4100|5000|4900|7500|7600|sabre/i) {
        # 4G switches and routers
        return ("1", "4");
    } elsif ($model =~ /6520|6510|DCX_PLUS|DCX_4S_PLUS/i) {
        # 16G switches and DCX w/16G blades
        return ("4", "16");
    } elsif ($model =~ /5424|300|5100|5300|5450|5470|5480|DCX|BES|7800|^8000$|VA40FC/i) {
        # 8G switches, DCX does not support older blades
        # Per Sunil, "1G" is documented, but does not work
        return ("2", "8");
    } elsif ($model =~ /48000/i) {
        # case by case based on blade and port
        # value = SLOT/PORT, get the slot to determine the type
        my ($slot, $iport) = split(/\//, $port);
        my $cmd = "slotshow";
        my @output = $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: BrocadeFCPort::_get_supported_speeds() " .
                            "error executing command ($cmd)");
            return ("", "");
        }

        # from the slot, get the ID of the blade in that slot
        # ID = 43|37|51|52|55|46 = 8G blades
        # ID = 24|36|17|33|31|18 = 4G blades
        # ID = 39 = 10G blade
        my $slot_line = (grep(/\s+$slot\s+/, @output))[0];
        $slot_line =~ s/^\s+//;
        my $slot_id   = (split(/\s+/, $slot_line))[3];

        if ($slot_id =~ /37|51|52|55|43|46|75|77/) {
            return ("2", "8");
        } elsif ($slot_id =~ /24|17|36|33|31|18/) {
            return ("1", "4");
        } elsif ($slot_id =~ /39/) {
            return ("10", "10");
        } else {
            $self->_set_err("Error: BrocadeFCPort::_get_supported_speeds() " .
                            "Could not determine min/max speed on 48K based on slot ID ($slot_id)");
            return ("", "");
        }
    } else {
        $self->_set_err("Error: BrocadeFCPort::_get_supported_speeds() " .
                        "Could not determine min/max speed based on switch type ($model)");
        return ("", "");
    }

}

###############################################################################
# Method:    _validate_fcoe_port_state
# Objective: Validate the port specified logged in you logged out based on
#            port enable/disable operations.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _validate_fcoe_port_state {
    my ($self, $port, $state) = @_;

    my ($cmd, $fcoe_count, @output);

    # if FCoE, verify via fcoe --loginshow
    $fcoe_count = 0;
    if (defined($self->{$port}->{protocol}) and
        $self->{$port}->{protocol} =~ /FCoE/i) {
        while ($fcoe_count < $self->{parent}{port_retries}) {
            # get login information
            $cmd = "fcoe --loginshow $port";
            @output = $self->{parent}->execute_brocade($cmd);
            if ($self->{parent}->error) {
                $self->_set_err("Error: BrocadeFCPort::_validate_fcoe_port_state() " .
                                "error executing command ($cmd)");
                return $SwitchUtil::FALSE;
            }

            # validate that the port went offline from FCoE perspective
            # - nothing connected
            # later validate that port came online from FCoE perspectve
            # Example output:
            # fcs-br8000-01:admin> fcoe --loginshow 10
            # Number of connected devices: 1
            # ==================================================================================================================
            #   Peer Type Connect Info         Device WWN         Device MAC      Session MAC FCoE Port MAC      Te port
            # ==================================================================================================================
            # FCOE_DEVICE   Direct    10:00:00:05:1e:8c:d1:f5  00:05:1e:8c:d1:f5  0e:fc:00:51:0a:04  00:05:1e:76:95:08   Te 0/8
            # fcs-br8000-01:admin> fcoe --loginshow 11
            # Number of connected devices: 0
            if ($state =~ /offline/i) {
                if (grep(/Number\s+of\s+connected\s+devices:\s+0/i, @output) > 0) {
                    last;
                }
            } elsif ($state =~ /online/i) {
                if (grep(/Number\s+of\s+connected\s+devices:\s+0/i, @output) == 0) {
                    last;
                }
            }
            sleep(3);
            $fcoe_count++;
        }
    } else {
        # simply return if not FCoE
        return $SwitchUtil::TRUE;
    }

    # check to make sure FCoE retries were not
    # exhausted.
    if ($fcoe_count < $self->{parent}{port_retries}) {
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }

}

###############################################################################
# Method:    _get_admin_domain_database
# Objective: Log into the switch and retrieve all of the admin domain
#            information found.  This will be accomplished through AD255.
#            The data will be returned in a hash that has the data saved from
#            the perspective of the admin domain and from the perspective of the
#            switch.
#
#            $hash = { by_ad     => {AD#}  => {fcid} => {port1} = 1,
#                      by_switch => {fcid} => {AD#}  => {port1} = 1,
#                    }
#
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_admin_domain_database {
    my ($self) = @_;

    $self->_clear_err();

    # Grab the switch object from the portmgr object
    my $switch_obj = $self->{parent};

    unless(ref($switch_obj) =~ /sanswitch/i) {
        $self->_set_err("Unable to retrieve the switch object from the portmgr object");
        return {};
    }

    my $switch_name = $switch_obj->hostname();
    unless(defined($switch_name) and $switch_name ne "") {
        $self->_set_err("Unable to retrieve the switch name from the switch object");
        return {};
    }

    # Make sure the switch isn't in access gateway mode
    if ($switch_obj->is_npv_enabled()) {
        # Looks like NPV is enabled so it can't support admin domains.  Set an
        # error and return false;
        $self->_set_err("Switch '$switch_name' is in Access Gateway Mode. " .
                        "Admin domains are not supported");
        return {};
    }

    # Make sure Virtual Fabrics are NOT enabled on the switch
    if ($switch_obj->is_vf_enabled()) {
        # Looks like VF is enabled so it can't support admin domains.  Set an
        # error and return false;
        $self->_set_err("Switch '$switch_name' has virtual fabrics enabled so admin " .
                        "domains are not supported");
        return {};
    }

    # Verify that we are running a version that supports admin domains
    unless ($switch_obj->is_ad_enabled()) {
        $self->_set_err("Admin Domains are not supported on this switch");
        return {};
    }

    # Now that we know we have a switch that "should" be able to support admin domains,
    # go ahead and retrieve the current admin domain context of the current connection
    # so we can switch back to it when we are done.  We will need to switch to context
    # 255 to dump the database.
    my $orig_ad = $switch_obj->get_current_admin_domain();
    if ($switch_obj->error) {
        $self->_set_err("Unable to retrieve the current admin domain for switch '$switch_name':\n",
                        $switch_obj->errmsg());
        return {};
    }

    unless($orig_ad == 255) {
        # Switch to AD255 since we weren't already there
        $switch_obj->switch_to_admin_domain("255");
        if ($switch_obj->error) {
            $self->_set_err("Unable to switch to AD255 on switch '$switch_name':\n", $switch_obj->errmsg());
            return {};
        }
    }

    # Do an ad --show to dump the AD config
    my $cmd = "ad --show";
    $self->sanlog(-msg => "Executing command ($cmd)");

    my @ad_output;
    my $log = log_global();
    if ($log->may_debug()) {
        @ad_output = $switch_obj->execute_brocade($cmd, "", "", -quiet => 0, -retry_if_fabric_busy => 1);
    } else {
        @ad_output = $switch_obj->execute_brocade($cmd, "", "", -quiet => 1, -retry_if_fabric_busy => 1);
    }
    if ($switch_obj->error()) {
        $self->_set_err("Unable to run command ($cmd) on switch '$switch_name':\n",
                        $switch_obj->errmsg());

        # Try to switch back to the original AD before bailing
        $switch_obj->switch_to_admin_domain($orig_ad);
        if ($switch_obj->error) {
            $self->_set_err("Unable to switch to AD${orig_ad} on switch " .
                            "'$switch_name':\n", $switch_obj->errmsg());
        }
        return {};
    }

    # Parse through the output
    # The output is going to look like this
    #
    # Current AD Number: 255  AD Name: AD255
    #
    # Transaction buffer configuration:
    # ---------------------------------
    # no configuration

    # Defined configuration:
    # ----------------------
    #
    # AD Number:   0  AD Name: AD0    State: Active
    #
    #         Device WWN members:             50:0a:09:83:87:f9:3e:b2;
    #                                         50:0a:09:82:87:f9:3e:b2;
    #                                         50:0a:09:81:87:f9:3e:b2;
    #         Switch port members:            31,31; 30,30; 21,21; 20,20;
    #                                         13,13; 12,12; 10,10; 11,11;
    #                                         14,14; 15,15; 31,30; 30,31;
    #                                         21,20; 20,21; 13,12; 12,13;
    #                                         10,11; 11,10; 14,15; 15,14;
    #                                         31,29; 30,28; 21,23; 20,22;
    #
    # AD Number:  10  AD Name: AD10   State: Active
    #
    #         Switch port members:            21,11; 20,10; 19,13; 11,21;
    #                                         10,20; 14,16; 15,17; 16,14;
    #                                         17,15; 18,12; 21,10; 20,11;
    #                                         19,12; 11,20; 10,21; 14,17;
    #                                         15,16; 16,15; 17,14; 18,13;
    #         Switch WWN members:             10:00:00:05:1e:03:3b:4d;
    #                                         10:00:00:05:1e:07:89:67;
    #                                         10:00:00:05:1e:02:35:f7;
    #                                         10:00:00:05:1e:04:db:1b;

    my $ad_hash = {};
    my $ad_number = "";
    foreach my $line (@ad_output) {
        if ($line =~ /AD Number:\s+(\d+)\s+/) {
            $ad_number = $1;
        } elsif ($line =~ /(\d+\,\d+.*\;)/) {
            # We have a line that has port numbers in it.
            my $members = $1;
            $members =~ s/\s+//g;
            my @pairs = split(/\;/, $members);

            # Loop over each fcid,port pair and split it out
            # and add it to the hash for the AD
            foreach my $pair (@pairs) {
                my ($fcid, $port) = split(/\,/, $pair);

                # Add the info to the AD hash.  If $ad_number is
                # not currently set, then we have a data error.
                unless ($ad_number ne "") {
                    $self->_set_err("Unexpected data returned from (ad --show) for switch '$switch_name'");

                    # Try to switch back to the original AD before bailing
                    $switch_obj->switch_to_admin_domain($orig_ad);
                    if ($switch_obj->error) {
                        $self->_set_err("Unable to switch to AD${orig_ad} on switch '$switch_name':\n", $switch_obj->errmsg());
                    }
                    return {};
                }

                my $ad_name = "AD" . $ad_number;

                $ad_hash->{by_ad}->{$ad_name}->{$fcid}->{$port} = 1;
                $ad_hash->{by_switch}->{$fcid}->{$ad_name}->{$port} = 1;
            }
        }
    } ##END foreach my $line (@ad_output) {

    # Switch back to original AD
    $switch_obj->switch_to_admin_domain($orig_ad);
    if ($switch_obj->error) {
        $self->_set_err("Unable to switch to AD${orig_ad} on switch '$switch_name':\n", $switch_obj->errmsg());
        return {};
    }

    return $ad_hash;
}

###############################################################################
# Method:    _ethernet_block
# Objective: Either shutdown (disable) or no shutdown (enable) the
#            ports retrieved from fcoe --loginshow <port> command for
#            disable and the saved list of ports (from disable)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _ethernet_block {
    my ($self, $port, $operation, @login_show) = @_;

    my (@ports_to_use, $command);

    # determine if we are going enble or disable the port
    if ($operation =~ /disable/i) {
        # Example output from fcoe --loginshow <port>
        # fcs-br8000-01:admin> fcoe --loginshow 10
        # Number of connected devices: 2
        # ==========================================================================================================================
        #   Peer Type Connect Info         Device WWN         Device MAC      Session MAC FCoE Port MAC      Te port
        # ==========================================================================================================================
        # FCOE_DEVICE   Direct    10:00:00:05:1e:8c:d1:f5  00:05:1e:8c:d1:f5  0e:fc:00:51:0a:04  00:05:1e:76:d2:08   Te 0/8
        # FCOE_DEVICE   Direct    10:00:00:05:1e:8c:c4:0d  00:05:1e:8c:c4:0d  0e:fc:00:51:0a:01  00:05:1e:76:d2:09   Te 0/9

        # get list of ports from an ethernet perspective to block from
        # the loginshow output.
        foreach my $line (@login_show) {
            # skip non-device lines
            if ($line !~ /FCOE_DEVICE/i) {
                next;
            }

            # split line to grab 10gig ports
            my $ether_port = (split(/\s+/, $line))[7];
            push (@ports_to_use, $ether_port);
        }

        if (scalar(@ports_to_use) == 0) {
            $self->_set_err("Error: BrocadeFCPort::_ethernet_block() no ethernet ports were found to block from FC port ($port)");
            return $SwitchUtil::FALSE;
        }

        # set command to use on ethernet command
        $command = "shutdown";
    } else {
        # extract the list of ports that were blocked with the previous
        # disable call
        if (defined($self->{$port}->{ethernet_ports})) {
            @ports_to_use = split(/\;/, $self->{$port}->{ethernet_ports});
        } else {
            $self->_set_err("Error: BrocadeFCPort::_ethernet_block() no ethernet ports were found to enable from FC port ($port)\n" .
                            "Meaning no ethernet ports were saved from the last block on this port");
            return $SwitchUtil::FALSE;
        }

        # set command to use on ethernet command
        $command = "no shutdown";
    }

    # get into cmsh to execute the ethernet commands
    $self->{parent}->execute_brocade("cmsh");
    if ($self->{parent}->error) {
        $self->_set_err("Error: BrocadeFCPort::_ethernet_block() failed to execute command (cmsh)");
        return $SwitchUtil::FALSE;
    }

    # block that port from an ethernet perspective
    foreach my $port (@ports_to_use) {
        # enter configure terminal
        $self->{parent}->_execute_configure_terminal();
        if ($self->{parent}->error()) {
            $self->_set_err("The command 'configure terminal' did not work properly", $self->{parent}->errmsg());
            return $SwitchUtil::FALSE;
        }

        # defined the commands to do the disable/enable
        my @cmds = ( "interface tengigabitethernet $port", $command, "end" );

        # execute each command (note, do not call brocade_execute, since this
        # portion is not virtual fabric aware and will interact with the cmsh
        # already started).
        foreach my $cmd (@cmds) {
            $self->{parent}->execute($cmd);
            if ($self->{parent}->error) {
                $self->_set_err("Error: BrocadeFCPort::_ethernet_block() failed to execute command ($cmd)");
                return $SwitchUtil::FALSE;
            }
        }
    }

    # get out of cmsh - do outside loop in case multiple ports are being blocked
    # Use execute and not execute_brocade since we will still be in cmsh mode and
    # execute_brocade tries to manipulate the contect for virtual fabrics.
    $self->{parent}->execute("exit");
    if ($self->{parent}->error) {
        $self->_set_err("Error: BrocadeFCPort::_ethernet_block() failed to exit from cmsh");
        return $SwitchUtil::FALSE;
    }

    # if the operation was disable, add the list of ports that were
    # disabled to self relative to the port (to unblock later).
    # if enable operation clear the port list.
    if ($operation =~ /disable/i) {
        my $ports_blocked = join(';', @ports_to_use);
        $self->{$port}->{ethernet_ports} = $ports_blocked;
    } else {
        $self->{$port}->{ethernet_ports} = "";
    }

    return $SwitchUtil::TRUE;
}

1;
