# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for ports for a Brocade VDX switch
## running NOS Operating System
## @author thomasm@netapp.com
## @status private
## @pod ./Port.pod

# Package Name
##############
package San::SanSwitch::FC::PortMgr::BrocadeVDXFCPort;

use strict;

# Package Requirements
######################
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::PortMgr;
use San::SanLog;
use Error;
use Data::Dumper;

use vars qw(@ISA);

@ISA = qw(Error SanLog PortMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: get_port_num()
# Objective:       Retrieve the port number/interface corresponding to a
#                  WWPN, by returning the first match in the array.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_num {
    my ($self, %args) = @_;
    my (@ports);

    @ports = $self->get_port_nums(%args);
    return "" if ($self->error);

    return $ports[0];

}

################################################################################
# Subroutine Name: get_port_nums()
# Objective:       Retrieve the all port number/interface corresponding to a
#                  WWPN. Multiple's can be returned when using looking for
#                  e-ports and a trunk is established.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_nums {
    my ($self, %args) = @_;

    my ($wwpn, $found_port, @found_ports);
    my (@possible_fids, $found_on_fid);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwpn/) {
            if (not length($wwpn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-wwpn'");
                return ();
            }
            next;
        }
    }

    # make sure you have a valid WWPN
    if (!defined($wwpn)) {
        $self->_set_err("WWPN must be specified");
        return ();
    }

    $self->{nameserver} = $self->get_nameserver();
    $self->sanlog(-msg => "Searching for WWPN $wwpn",
                  -msg_level => 7);

    $wwpn = lc($wwpn);
    $wwpn =~ s/://g;
    foreach my $interface (keys %{$self->{nameserver}}) {
        my $ns_wwpn = $self->{nameserver}{$interface}{wwpn};
        if ($wwpn eq $ns_wwpn) {
            if ($self->{nameserver}{$interface}{protocol} =~ /FCOE/i) {
                $found_port = $self->{nameserver}{$interface}{physical_interface}
            } else {
                $found_port = $interface;
            }
            # strip off instance of this port on physical port (NPIV)
            $found_port =~ s/_\d+$//;

            $self->sanlog(-msg => "Found match on port $found_port",
                          -msg_level => 3);
            push (@found_ports, $found_port);
        }
    }

    # looks to see if we found any ports and report accordingly.
    if (!(@found_ports)) {
        $self->_set_err("No matching port found)");
        return ();
    } else {
        return @found_ports;
    }
}

################################################################################
# Subroutine Name: port_disable()
# Objective:       Disable the port/interface specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_disable {
    my ($self, $port) = @_;

    my ($cmd, $status, $fcoe_port);

    $self->_clear_err();

    # get port info
    my $port_state = $self->get_port_info($port);
    if ($self->error) {
            $self->_set_err("Error getting port state ($port)");
            return $SwitchUtil::FALSE;
    }
    
    # Choose which type of block to do - virtual FC layer or ethernet layer.
    if (($self->{parent}->{fcoe_ethernet_block} == 0) and (not $port_state->{is_isl} eq "1")) {
        $fcoe_port = $self->lookup_fcoe_port($port);
        # get port info
        my $fcoe_port_state = $self->get_port_info($fcoe_port);

        if ($self->error) {
            $self->_set_err("Error: BrocadeVDXFCPort::port_disable() ".
                            "error getting port information ($port)");
            return $SwitchUtil::FALSE;
        }

        # if requested, get current state to store speed and topology
        if ($self->{parent}->{verify_on_enable} == 1) {
            $self->{$port}->{previous_speed} = $fcoe_port_state->{speed};
            $self->{$port}->{previous_type} = $fcoe_port_state->{type};
        }

        my @cmds = ( "conf t", "interface fcoe $fcoe_port", "shutdown", "end" );
        foreach $cmd (@cmds) {
            my @output = $self->{parent}->execute_brocade($cmd);
            if ($self->error) {
                $self->{$fcoe_port}->{previous_speed} = undef;
                $self->{$fcoe_port}->{previous_type} = undef;
                $self->_set_err("Error executing command ($cmd)");
                return $SwitchUtil::FALSE;
            }
        }
    } else {
        $status = $self->_port_disable_ethernet($port);
        return ($status);
    }

    # if asked to (default) validate port went offline
    # fcoe port case
    if ($self->{parent}->{fcoe_ethernet_block} == 0) {
        $status = $self->waitfor_port($fcoe_port, "offline");
        if ($status == $SwitchUtil::FALSE) {
            $self->_set_err("Failed to disable port, port still appears up");
            return $SwitchUtil::FALSE;
        }
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_enable()
# Objective:       Enable the port specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_enable {
    my ($self, $port) = @_;

    my ($cmd, $status, @output, $fcoe_port);

    $self->_clear_err();

    # get port info
    my $port_state = $self->get_port_info($port);
    if ($self->error) {
            $self->_set_err("Error getting port state ($port)");
            return $SwitchUtil::FALSE;
    }

    # If FCoE choose which type of block to do - virtual FC layer or
    # ethernet layer.
    if (($self->{parent}->{fcoe_ethernet_block} == 0) and (not $port_state->{is_isl} eq "1")) {
        # reload and lookup fcoe port again as this may have changed
        # in NOS 3.0.0
        $self->{nameserver} = $self->get_nameserver();
        if ($self->error) {
            return $SwitchUtil::FALSE;
        }
        $fcoe_port = $self->lookup_fcoe_port($port);
        if ($self->error) {
            return $SwitchUtil::FALSE;
        }

        my @cmds = ( "conf t", "interface fcoe $fcoe_port", "no shutdown", "end" );
        foreach $cmd (@cmds) {
            @output = $self->{parent}->execute_brocade($cmd);
            if ($self->error) {
                $self->_set_err("Error executing command ($cmd)");
                return $SwitchUtil::FALSE;
            }
        }

        # get port info
        $self->get_port_info($port);
    } else {
        # enable the Ethernet port from the fcoe port
        $status = $self->_port_enable_ethernet($port);
        return ($status);
    }

    # add explicit sleep into the code here before checking the
    # port to see if it is online
    sleep (3);

    # wait for the port to transition to online, or report error
    # case of checking for fcoe port
    if ($self->{parent}->{fcoe_ethernet_block} == 0) {
        $status = $self->waitfor_port($fcoe_port, "online");
        if ($status == $SwitchUtil::FALSE) {
             $self->_set_err("Failed to enable port, port still appears down");
             return $SwitchUtil::FALSE;
        }
        # compare values now (after enable) versus before the
        # port_disable
        $self->check_port_settings($fcoe_port);
        if ($self->error) {
            return $SwitchUtil::FALSE;
        }
    }

    # reload and lookup fcoe port again as this may have changed
    # in NOS 3.0.0
    $self->{nameserver} = $self->get_nameserver();
    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: set_port_speed()
# Objective:       Set the specified port to the specified speed
#                  Currently this is not supported
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_speed {
    my ($self, %args) = @_;
    my ($status, $port, $speed, $cmd);

    $self->_clear_err();

    $self->_set_err("Setting port speed is not supported on VDX/NOS");

    return $SwitchUtil::FALSE;
}

################################################################################
# Subroutine Name: get_port_info()
# Objective:       Get information about the specified port/interface
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_info {
    my ($self, $port) = @_;

    my ($cmd, @output);

    $self->_clear_err();

    # Initialize hash
    my $hash = {};

    # NPIV case, remove instance from port
    # could be in <port>_<instance> format
    $port =~ s/\_\d+//;

    # general fields
    $hash->{state}            = "unknown";
    $hash->{ethernet_port}    = "unknown";
    $hash->{speed}            = 10;
    $hash->{type}             = "unknown";
    $hash->{fcid}             = "NA";
    $hash->{min_speed}        = 0;
    $hash->{max_speed}        = 10;

    # intialize port $protocol to ethernet 
    # call to "show interface fcoe" will set $protcol to FCoE if appropriate (port is FCoE port).
    my $protocol = "Eth";

    # if there is no stored protocol, find it.
    $cmd = "show interface fcoe";
    @output = $self->{parent}->execute_brocade($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error executing command ($cmd)");
        return {};
    }

    foreach my $line (@output) {
        if ($line =~ /FCOE $port\s+is/i) {
            $protocol = "FCOE";
        }
    } 

    if (not ($protocol =~ /FCOE/i)) {
        # call portshow on the specified ethernet port
        # Example output (of lines parsed):
        # TenGigabitEthernet 1/0/5 is up, line protocol is up (connected)
        # Hardware is Ethernet, address is 0005.338b.8504
        #     Current address is 0005.338b.8504
        # Pluggable media present
        # Interface index (ifindex) is 4697948164
        # MTU 2500 bytes
        # LineSpeed Actual     : 10000 Mbit
        # LineSpeed Configured : Auto, Duplex: Full
        # Flowcontrol rx: off, tx: off
        # Last clearing of show interface counters: 1w1d05h
        # Queueing strategy: fifo
        $cmd = "show interface tengigabitethernet $port";
        @output = $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error and $self->{parent}->{fcoe_ethernet_block} == 0) {
            # if this is a FCOE (virtual block) and we are in this block and
            # working on ethernet port and the call to show the ethernet
            # interface fails, set the state to offline (assuming that
            # the we wanted to get fcoe port info and that port is down maybe
            # to us disabling the port).
            $hash->{state} = &SwitchUtil::normalize_state("offline");
            $self->{parent}->_clear_err();
            return $hash;
        } elsif ($self->{parent}->error) {
            $self->_set_err("Error executing command ($cmd)");
            return {};
        }

        foreach my $line (@output) {
            if ($line =~ /line\s+protocol\s+is\s+(\S+)/i) {
                $hash->{state} = &SwitchUtil::normalize_state($1);
            }
        }

        # make some assumptions since ethernet port coming in
        $hash->{type} = "E-Port";
        $hash->{ethernet_port} = $port;
    } else {
        # $cmd is set to show login here for logging purposes
        $cmd = "show fcoe login";
        $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error executing command ($cmd)");
            return {};
        }
        # call portshow on the specified fcoe port
        # Example output (of lines parsed):
        # Interface Fcoe 1/1/1
        # ---------------------------------------------------------
        # Fcoe 1/1/1 is Up, Line protocol is Down
        # Ethernet port is TenGigabitEthernet  1/0/1
        # Interface index (ifindex) is 402718720
        # Config Mode is VF, Current Mode is VF
        # Last clearing of show interface counters: 2011-09-20 16:28:34.665831
        $cmd = "show interface fcoe $port";
        @output = $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error executing command ($cmd)");
            return {};
        } 
    
        foreach my $line (@output) {
            if ($line =~ /Line\s+protocol\s+is\s+(\S+)/i) {
                $hash->{state} = &SwitchUtil::normalize_state($1);
            }
            if ($line =~ /Current\s+Mode\s+is\s+(\S+)/i) {
                $hash->{type} = &SwitchUtil::normalize_type($1);
            }
            if ($line =~ /Ethernet\s+port\s+is\s+TenGigabitEthernet\s+(\S+)/i) {
                $hash->{ethernet_port} = $1;
            }
        }
    }

    return $hash;
}

################################################################################
# Subroutine Name: set_port_type()
# Objective:       Set the specified port to the specified port type
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_type {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_dump_info()
# Objective:       Dump the info for the provided port
# Details:         See POD documentation at the end of this file
################################################################################
sub port_dump_info {
    my ($self, $port) = @_;

    my ($protocol, $cmd);
    my ($ethernet_port);

    $self->_clear_err();

    # Dump the status of the fcoe interface if applicable
    $protocol = $self->{nameserver}{$port}{protocol};
    if ($protocol =~ /FCOE/i) {
        # show port fcoe login information
        $cmd = "show interface fcoe $port";
        $self->{parent}->execute_brocade($cmd);
        if ($self->error) {
            $self->_set_err("Error getting FCoE information ($port)");
            return $SwitchUtil::FALSE;
        }
    }

    # if this is FCOE then get the physical port stored, else
    # the port passed in should be e-port/Ethernet port
    if ($protocol =~ /FCOE/i) {
        $ethernet_port = $self->{$port}->{ethernet_port};
    } else {
        $ethernet_port = $port;
    }

    # show ethernet port information
    $cmd = "show interface tengigabitethernet $ethernet_port";
    $self->{parent}->execute_brocade($cmd);
    if ($self->error) {
        $self->_set_err("Error getting Ethernet Port information ($ethernet_port)");
        return $SwitchUtil::FALSE;
    }

    # show RAS trace info
    $cmd = "show logging raslog";
    $self->{parent}->execute_brocade($cmd);
    if ($self->error) {
        $self->_set_err("Error getting RAS trace information");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new port object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/i) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # initialize the object members
    $self->{parent} = $parent;

    $self->{nameserver} = $self->get_nameserver();

    return $self;
}

###############################################################################
# Method:    _get_nameserver_database
# Objective: Retrieve nameserver database information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_nameserver_database {
    my ($self, $all_vf) = @_;

    # invalidate the switchShow cache, since we want to do
    # an initial lookup with fresh information. We can use the
    # cache in the context of the function, but when leaving,
    # we also want to invalidate.
    $self->{parent}->{cache}{show_name_server} = undef;

    my $ns = {};

    # Example show name-server detail rbridge <x> output:
    # PID: 011d19
    #   Port Name: 20:13:00:A0:98:25:14:A8
    #   Node Name: 21:06:00:A0:98:25:14:A8
    #   FC4s: FCP
    #  PortSymb: [51] "NetApp FC Target Port (8112) vs3_ibmx3650_fcs01:fc3"
    #  NodeSymb: [33] "NetApp Vserver vs3_ibmx3650_fcs01"
    #  Fabric Port Name: 20:1D:00:05:33:8B:84:DB
    #  Permanent Port Name: 20:13:00:A0:98:25:14:A8
    #  Device type: NPIV Unknown(initiator/target)
    #  Interface: Fcoe 1/1/5
    #  Physical Interface: Te 1/0/5
    #  Share Area: No

    # get nameserver information relative to the local switch (rbridge-id)
    my $cmd = "show name-server detail rbridge-id $self->{parent}->{domain_id}";
    my @output = $self->{parent}->execute_brocade($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error executing command ($cmd)");
        return {};
    }

    # get number of name server entries
    my $num_entries = (grep(/total number of/i, @output))[0];
    $num_entries =~ m/total\s+number\s+of\s+(\d+)\s+entries/;
    $num_entries = $1;
    $self->sanlog(-msg => "Total number of name server entries = $num_entries");

    # parse out each name-server entry and push into array (or arrays)
    my $found = $SwitchUtil::FALSE;
    my @ns_entry = ();
    my @all_ns_entries = ();
    foreach my $line (@output) {
        # ignore newlines and total number of NS entries line
        if ($line =~ /^\s*$/) {
            next;
        }

        # PID: denotes new name-server entry. If this is new/first entry
        # push PID into ns entry and mark as found
        if ($line =~ /PID:\s+\S+/i and $found == $SwitchUtil::FALSE) {
            $found = $SwitchUtil::TRUE;
        }

        if (($line =~ /PID:\s+\S+/i or $line =~ /total number of/i) and
             $found == $SwitchUtil::TRUE) {
            # PID: denotes new name-server entry. If this is another
            # entry push the existing array into total list and
            # start a new ns entry. Also, take into consideration the
            # last entry in the NS (where do more PID: are found)
            if (scalar(@ns_entry) > 0) {
                push(@all_ns_entries, [ @ns_entry ]);
            }

            @ns_entry = ();
        }

        push(@ns_entry, $line);

    }

    # traverse each ns array and extract details
    foreach my $one_ns_entry (@all_ns_entries) {
        my @per_ns_entry = @{$one_ns_entry};

        # get port context
        my $port = (grep(/\s+Interface:/, @per_ns_entry))[0];
        $port =~ s/\s+Interface:\s+\S+\s+//i;

        # could be multiple entries on same port (NPIV for instance)
        $ns->{$port}->{count} = ($ns->{$port}->{count} + 1);
        if ($ns->{$port}->{count} > 1) {
            $port = $port . "_" . $ns->{$port}->{count};
            $ns->{$port}->{kind} = "npiv";
        }

        # pull out ns attributes
        foreach my $line (@per_ns_entry) {
            if (grep(/Device\s+type\:\s+NPIV/i, @per_ns_entry) > 0) {
                $ns->{$port}->{npiv} = 1;
            } else {
                $ns->{$port}->{npiv} = 0;
            }

            if ($line =~ /^\s+Port\s+Name\:/i) {
                $ns->{$port}->{wwpn} = $line;
                $ns->{$port}->{wwpn} =~ s/\s+Port\s+Name\:\s+//i;
                $ns->{$port}->{wwpn} =~ s/://g;
                $ns->{$port}->{wwpn} = lc($ns->{$port}->{wwpn});
            }

            if ($line =~ /^\s+Node\s+Name\:/i) {
                $ns->{$port}->{wwnn} = $line;
                $ns->{$port}->{wwnn} =~ s/\s+Node\s+Name\:\s+//i;
                $ns->{$port}->{wwnn} =~ s/://g;
                $ns->{$port}->{wwnn} = lc($ns->{$port}->{wwnn});
            }

            if ($line =~ /^\s+PortSymb\:/i) {
                $ns->{$port}->{p_symb} = $line;
                $ns->{$port}->{p_symb} =~ s/\s+PortSymb\:\s+\[\d+\]\s+//i;
                $ns->{$port}->{p_symb} =~ s/\"//g;
            }

            if ($line =~ /^\s+NodeSymb\:/i) {
                $ns->{$port}->{n_symb} = $line;
                $ns->{$port}->{n_symb} =~ s/\s+NodeSymb\:\s+\[\d+\]\s+//i;
                $ns->{$port}->{n_symb} =~ s/\"//g;
            }

            if ($line =~ /PID\:/i) {
                $ns->{$port}->{address} = $line;
                $ns->{$port}->{address} =~ s/PID:\s+//i;
            }

            if ($line =~ /^\s+Interface\:\s+(\S+)\s+(\S+)/i) {
                $ns->{$port}->{protocol} = $1;
                $ns->{$port}->{port} = $2;
            }

            if ($line =~ /^\s+Physical\s+Interface\:\s+(\S+)\s+(\S+)/i) {
                $ns->{$port}->{type} = $1;
                $ns->{$port}->{physical_interface} = $2;
            }
        }

        # for now, all ports (init, target) in NS will be FCOE
        $ns->{$port}->{protocol} = "FCOE";
    }

    # get ISL information and add to NS hash
    $cmd = "show fabric islports";
    @output = $self->{parent}->execute_brocade($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error executing command ($cmd)");
        return {};
    }

    # get available ISL ports 
    # format of output (NOS 2.1):
    #   1 2/0/1           Up      ISL (Trunk port, Primary is 2/0/3 )
    #   2 2/0/2           Up      ISL (Trunk port, Primary is 2/0/3 )
    #   3 2/0/3           Up      ISL 10:00:00:05:33:8b:84:db "vdx6720-fcs01" (upstream)(Trunk Primary)
    #   4 2/0/4           Up      ISL (Trunk port, Primary is 2/0/3 )
    #   5 2/0/5           Down    Down (Admin)
    #   6 2/0/6           Down    Down (Admin)
    # format of output (NOS 2.1.1):
    # 19 Te 120/0/19    Up  ISL 10:00:00:05:33:6e:80:05 "vcs3-113" (Trunk Primary)
    # 20 Te 120/0/20    Up  ISL 10:00:00:05:33:6f:71:33 "vcs3-112" (upstream)(Trunk Primary)

    my @isl_ports = grep(/(\d+\/\d+\/\d+)\s+Up/i, @output);
    foreach my $isl_line (@isl_ports) {
        $isl_line =~ m/(\d+\/\d+\/\d+)\s+Up\s+ISL\s+(.*)/i;
        my $isl_port = $1;
        my $isl_data = $2;

        # fill in appropriate field to ns entry
        $ns->{$isl_port}->{type} = "E-Port";
        $ns->{$isl_port}->{protocol} = "ETHERNET";
        $ns->{$isl_port}->{port} = $isl_port;
        $ns->{$isl_port}->{ethernet_port} = $isl_port;
        $ns->{$isl_port}->{e_port_type} = "STANDARD";
        $ns->{$isl_port}->{wwnn} = 'NA';
        $ns->{$isl_port}->{address} = 'NA';
        $ns->{$isl_port}->{n_symb} = 'NA';
        $ns->{$isl_port}->{p_symb} = 'NA';
        $ns->{$isl_port}->{is_isl} = '1';

        # determine e-port type (TE or FI)
        # 123 Fi 113/0/3        Up  ISL 50:00:51:e4:30:a0:1e:71 
        # 19  Te 120/0/19       Up  ISL 10:00:00:05:33:6e:80:05
        if ($isl_line =~ /\s+FI\s+/i) {
            $ns->{$isl_port}->{etype} = "FI";
        } else {
            $ns->{$isl_port}->{etype} = "TE";
        }

        # pull apart the primary trunk port
        if ($isl_data =~ /Trunk Primary/i) {
            $isl_data =~ m/(\S+)\s+(\S+)\s+(.*)/;
            $ns->{$isl_port}->{wwpn} = $1;
            $ns->{$isl_port}->{e_port_to} = $2;
            $ns->{$isl_port}->{e_port_to} =~ s/\"//g;
        } else {
            $ns->{$isl_port}->{wwpn} = 'NA';
            $ns->{$isl_port}->{e_port_to} = 'NA';
        }

        # get the primary trunk port
        if ($isl_data =~ /Trunk port/i) {
            $isl_data =~ m/Primary\s+is\s+(\d+\/\d+\/\d+)/;
            $ns->{$isl_port}->{primary_trunk_port} = $1;
        }
    }

    return $ns;
}

###############################################################################
# Method:    _validate_ethernet_port_state
# Objective: Validate the port specified logged in you logged out based on
#            port enable/disable operations.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _validate_ethernet_port_state {
    my ($self, $port, $state) = @_;

    # loop for the number of specified retries
    foreach my $retry (1 .. $self->{parent}{port_retries}) {
        # call portshow on the specified port
        # Example of output being parsed:
        # TenGigabitEthernet 1/0/1 is up, line protocol is up (connected)
        # Hardware is Ethernet, address is 0005.338b.8500
        #     Current address is 0005.338b.8500
        # Pluggable media present
        # Interface index (ifindex) is 4697686016
        # MTU 9216 bytes
        # LineSpeed Actual     : 10000 Mbit
        # LineSpeed Configured : Auto, Duplex: Full
        # Flowcontrol rx: off, tx: off
        # Last clearing of show interface counters: 6d14h20m
        # Queueing strategy: fifo
        my $cmd = "show interface TenGigabitEthernet $port";
        my @output = $self->{parent}->execute_brocade($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error executing command ($cmd)");
            return {};
        }
    
        # find current port state
        my $port_state;
        foreach my $line (@output) {
            if ($line =~ /line\s+protocol\s+is\s+(\S+)/i) {
                $port_state = $1;
                last;
            }
        }

        # if looking for online and up, return true
        if ($state =~ /online/i and $port_state =~ /up/i) {
            return $SwitchUtil::TRUE;
        }

        # if looking for offline and down, return true
        if ($state =~ /offline/i and $port_state =~ /down/i) {
            return $SwitchUtil::TRUE;
        }
        &SanUtil::nate_sleep(1);
    }

    # $cmd is set to show login here for both Ethernet and FCoE
    my $cmd = "show fcoe login";
    $self->{parent}->execute_brocade($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error executing command ($cmd)");
        return {};
    }

    # otherwise return false (did not get proper state and retries exhausted)
    return $SwitchUtil::FALSE;
}

################################################################################
# Subroutine Name: _port_enable_ethernet()
# Objective:       Enable the port specified for ethernet only
# Details:         See POD documentation at the end of this file
################################################################################
sub _port_enable_ethernet {
    my ($self, $port) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    # enable the Ethernet port
    my @cmds = ( "conf t", "interface TenGigabitEthernet $port", "no shutdown", "end" );
    foreach $cmd (@cmds) {
        @output = $self->{parent}->execute_brocade($cmd);
        if ($self->error) {
            my @msg = $self->errmsg();
            $self->_set_err("Error enabling ethernet port(s)\n" .
                            "Reason: @msg");
            return $SwitchUtil::FALSE;
        }
    }

    # check to make sure port comes online
    $status = $self->_validate_ethernet_port_state($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Failed to enable port, port still appears down");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: _port_disable_ethernet()
# Objective:       Enable the port specified for ethernet only
# Details:         See POD documentation at the end of this file
################################################################################
sub _port_disable_ethernet {
    my ($self, $port) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    # enable the Ethernet port
    my @cmds = ( "conf t", "interface TenGigabitEthernet $port", "shutdown", "end" );
    foreach $cmd (@cmds) {
        @output = $self->{parent}->execute_brocade($cmd);
        if ($self->error) {
            my @msg = $self->errmsg();
            $self->_set_err("Error disabling ethernet port(s)\n" .
                            "Reason: @msg");
            return $SwitchUtil::FALSE;
        }
    }

    # add explicit sleep into the code here before checking the
    # port to see if it is online
    sleep (3);

    # check to make sure port goes offline
    $status = $self->_validate_ethernet_port_state($port, "offline");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Failed to disable ethernet port, port still appears up");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: lookup_fcoe_port()
# Objective:       Return the fcoe port associated with this eth port
################################################################################
sub lookup_fcoe_port {
    my ($self, $eth_port) = @_;

    # set term length to 0 for non paginated command output
    my $term_cmd = "term length 0";
    my @output = $self->{parent}->execute_brocade($term_cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error executing command ($term_cmd)");
        return {};
    }

    my $cmd = "show interface fcoe";
    @output = $self->{parent}->execute_brocade($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error executing command ($cmd)");
        return {};
    }

    # find dynamic fcoe port from interface command, not nameserver
    my $port;
    while (defined (my $line = shift(@output))) {
        if ($line =~ /Fcoe\s+(\S+)\s+is/i) {
            my $next_line = shift(@output);
            my $fcoe_port = $1;
            if ($next_line =~ /TenGigabitEthernet\s+(\S+)/i) {
                if ($eth_port eq $1) {
                    return $fcoe_port;
                }
            }elsif($next_line =~ /TenGigabitEthernet\s+/i){
                return $fcoe_port;
            }
        }
    }

    # if we cannot find the fcoe port, set an error and return false
    $self->_set_err("Failed to find fcoe port on physical port $eth_port");
    return "";
}

1;
