# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for ports for a Cisco FC switch
## @author evan@netapp.com
## @status private
## @pod ./Port.pod

# Package Name
##############
package San::SanSwitch::FC::PortMgr::CiscoFCPort;

use strict;
use Tharn;

# Package Requirements
######################
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::PortMgr;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog PortMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_port_num()
# Objective:       Retrieve the port number/interface corresponding to a
#                  WWPN, by returning the first match in the array.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_num {
    my ($self, %args) = @_;
    my (@ports);

    $self->_clear_err();

    @ports = $self->get_port_nums(%args);
    return "" if ($self->error);

    return $ports[0];

}

################################################################################
# Subroutine Name: get_port_nums()
# Objective:       Retrieve the all port number/interface corresponding to a
#                  WWPN. Multiple's can be returned when using looking for
#                  e-ports and a trunk is established.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_nums {
    my ($self, %args) = @_;
    my ($wwpn, $found_port, @found_ports);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwpn/) {
            if (not length($wwpn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-wwpn'");
                return ();
            }
            next;
        }
    }

    $self->{nameserver} = $self->get_nameserver();
    if (defined($wwpn)) {
        $self->sanlog(-msg => "Searching for WWPN $wwpn",
                      -msg_level => 7);
        $wwpn = lc($wwpn);
        $wwpn =~ s/://g;
        foreach my $interface (keys %{$self->{nameserver}}) {
            my $ns_wwpns = $self->{nameserver}{$interface}{wwpn};
            # in NPV case you can have multiple WWPNs on an interface
            my @wwpns = split(/;/, $ns_wwpns);
            if (grep(/$wwpn/, @wwpns) > 0) {
                $found_port = $interface;
                $self->sanlog(-msg => "Found match on port $found_port",
                              -msg_level => 3);
                push (@found_ports, $found_port);
            }
        }
    } else {
        $self->_set_err("WWPN must be specified");
        return ();
    }

    if (!(@found_ports)) {
        $self->_set_err("No matching port found");
        return ();
    } else {
        return @found_ports;
    }
}

################################################################################
# Subroutine Name: port_disable()
# Objective:       Disable the port/interface specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_disable {
    my ($self, $port) = @_;

    my (@cmds, $have_port_state, $port_state, $status);

    $have_port_state = 0;

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    # check to see if port contain's "fc" and if not add. It might
    # not if coming from the tharnhost file.
    if ($port !~ /^\s*fc/ and
        $port !~ /^\s*vfc/ and
        $port !~ /bay|ext/ and
        $self->{parent}->get_model() !~ /nexus/i) {
        $port = "fc" . $port;
    }

    # if requested, get current state to store speed and topology
    if ($self->{parent}->{verify_on_enable} == 1) {
        $port_state = $self->get_port_info($port);
        if ($self->error) {
            $self->_set_err("Error: CiscoFCPort::port_disable() ".
                            "error getting port state ($port)");
            return $SwitchUtil::FALSE;
        }
        $self->{$port}->{previous_speed} = $port_state->{speed};
        $self->{$port}->{previous_type} = $port_state->{type};
        $have_port_state = 1;
    }

    # if this is FCoE (vfc port) and we want to block at the
    # ethernet port level, grab the ethernet port, and block at the
    # the ethernet level
    if ($self->{parent}->{fcoe_ethernet_block} == 1 and $port =~ /^vfc/i) {
        # if we already called get_port_info (from verify), don't call again.
        if ($have_port_state == 0) {
           $port_state = $self->get_port_info($port);
           if ($self->error) {
               $self->_set_err("Error: CiscoFCPort::port_disable() ".
                               "error getting port state ($port)");
               return $SwitchUtil::FALSE;
           }
        }

        # get ethernet port from get_port_info (port state)
        my $ether_port = $port_state->{ethernet_port};
        if (not defined($ether_port)) {
            $self->nate_log("-warn",
                            "Warning: CiscoFCPort::port_disable() ".
                                "error getting ethernet port from the the vfc port ($port)");
            return $SwitchUtil::FALSE;
        }

        # save ports blocked so that we have for enable operation
        $self->{$port}->{ethernet_ports} = $ether_port;

        # set commands relative to ethernet
        @cmds = ("conf t", "int $ether_port", "shutdown", "end", "show int $port");
    } else {
        # set commands relative to FC
        @cmds = ("conf t", "int $port", "shutdown", "end", "show int $port");
    }

    # execute the prepared commands
    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->{$port}->{previous_speed} = undef;
            $self->{$port}->{previous_type} = undef;
            $self->_set_err("Error: CiscoFCPort::port_disable() ".
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # if asked to (default) validate port went offline
    if ($self->{parent}->{validate_offline} == 1) {
        # wait for the port to transition to offline, or report error
        $status = $self->waitfor_port($port, "offline");
        if ($status == $SwitchUtil::FALSE) {
            $self->_set_err("Error: CiscoFCPort::port_disable() " .
                            "failed to enable port, port still appears up");
            return $SwitchUtil::FALSE;
        }
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_enable()
# Objective:       Enable the port specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_enable {
    my ($self, $port) = @_;

    my (@cmds, $status, @output);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};
    my $count               = 0;

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    # check to see if port contain's "fc" and if not add. It might
    # not if coming from the tharnhost file.
    if ($port !~ /^\s*fc/ and
        $port !~ /^\s*vfc/ and
        $port !~ /bay|ext/ and
        $self->{parent}->get_model() !~ /nexus/i) {
        $port = "fc" . $port;
    }

    # if we are FCoE and enabling a port, just the list of
    # ethernet ports block from the previous disable operation
    if ($self->{parent}->{fcoe_ethernet_block} == 1 and $port =~ /^vfc/) {
        my $ether_port = $self->{$port}->{ethernet_ports};
        if (not defined($ether_port) or $ether_port eq "") {
            $self->_set_err("Error: BrocadeFCPort::_ethernet_block() no ethernet ports were found to enable from FC port ($port)\n" .
                            "Meaning no ethernet ports were saved from the last block on this port");
            return $SwitchUtil::FALSE;
        }
        # Ethernet port for enable
        @cmds = ("conf t", "int $ether_port", "no shutdown", "end");
    } else {
        # FC port for enable
        @cmds = ("conf t", "int $port", "no shutdown", "end");
    }

    # execute the prepared commands
    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoFCPort::port_enable() " .
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # add explicit sleep into the code here before checking the
    # port to see if it is online (suggested by Dave/Sunil)
    sleep (3);

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: CiscoFCPort::port_enable() " .
                        "failed to enable port, port still appears down");
        return $SwitchUtil::FALSE;
    }

    # compare values now (after enable) versus before the
    # port_disable
    $self->check_port_settings($port);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # reset so that next disable will populate properly
    $self->{$port}->{ethernet_ports} = "";

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: set_port_speed()
# Objective:       Set the specified port to the specified speed
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_speed {
    my ($self, %args) = @_;
    my ($status, $port, $speed);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};
    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Error: CiscoFCPort::set_port_speed() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?speed/) {
            if ((not length($speed = $args{$_})) &&
                ($args{$_} !~ /^\d$/ or $speed !~ /auto/i)) {
                $self->_set_err("Error: CiscoFCPort::set_port_speed() " .
                                "Invalid value for argument '-speed'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # handle auto speed setting
    if ($speed =~ /AUTO/i) {
        $speed = "auto";
    }

    # Cisco requires this setting to be 1000, 2000, 4000, or auto
    $speed *= 1000 if ($speed =~ /\d/);

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    # check to see if port contain's "fc" and if not add. It might
    # not if coming from the tharnhost file.
    if ($port !~ /^\s*fc/ and
        $port !~ /^\s*vfc/ and
        $port !~ /bay|ext/ and
        $self->{parent}->get_model() !~ /nexus/i) {
        $port = "fc" . $port;
    }

    my @cmds = ("conf t",
                "int $port",
                "switchport speed $speed",
                "end");

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoFCPort::set_port_speed() " .
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: CiscoFCPort::set_port_speed() " .
                        "Port state did not transtion to online state");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: get_port_info()
# Objective:       Get information about the specified port/interface
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_info {
    my ($self, $port) = @_;
    my (@output);

    $self->_clear_err();

    # Initialize hash
    my $hash = {};
    $hash->{state}         = "unknown";
    $hash->{state_reason}  = "";
    $hash->{speed}         = 0;
    $hash->{type}          = "unknown";
    $hash->{fcid}          = 0;
    $hash->{min_speed}     = 0;
    $hash->{max_speed}     = 0;
    $hash->{hardware}      = "";
    $hash->{vsan}          = 0;
    $hash->{ethernet_port} = undef;

    # get port information differently for nexus switches (really Ethernet)
    if (($self->{parent}->get_firmware_version() =~ /NX-OS/i and
        $self->{parent}->get_model() =~ /nexus/i) or $port =~ /vfc/i) {
        $hash = $self->_get_port_info_nexus($port, $hash);
        return $hash;
    }

    # check to see if port contain's "fc" and if not add. It might
    # not if coming from the tharnhost file.
    if ($port !~ /^\s*fc/ and
        $port !~ /^\s*vfc/ and
        $port !~ /bay|ext/) {
        $port = "fc" . $port;
    }

    @output = $self->{parent}->execute("show int $port");
    foreach my $line (@output) {
        if ($line =~ /Speed is (\d+)/) {
            $hash->{speed} = int($1);
        }
        if ($line =~ /Hardware is (\w+)/i) {
            $hash->{hardware} = $1;
        }
        if ($line =~ /$port is \S+\s+\((.*)\)/) {
            $hash->{state_reason} = $1;
        }
        if ($line =~ /$port is (\S+)/) {
            $hash->{state} = &SwitchUtil::normalize_state($1);
        }
        if ($line =~ /Port mode is (\S+)/) {
            my $type = $1;
            $type =~ s/,//;
            $hash->{type} = &SwitchUtil::normalize_type($type);
        }
        if ($line =~ /FCID is (\S+)/) {
            $hash->{fcid} = hex($1);
        }
        if ($line =~ /port vsan is (\d+)/i) {
            $hash->{vsan} = $1;
        }
    }

    # if this is a vfc port (on MDS), then set the speed to 10G for now
    if ($port =~ /vfc/i) {
        $hash->{min_speed} = 10;
        $hash->{max_speed} = 10;
    }

    # Load the cached values if we have already looked these values up before
    $hash->{min_speed} = $self->{cache}{min_speed}{$port}
        if defined($self->{cache}{min_speed}{$port});
    $hash->{max_speed} = $self->{cache}{max_speed}{$port}
        if defined($self->{cache}{max_speed}{$port});

    # If we have the default or invalid values, we either have not looked up the
    # capabilities of this port before or the previous look up failed
    if ($hash->{min_speed} == 0 or $hash->{max_speed} == 0) {
        @output = $self->{parent}->execute("show int $port capabilities");
        foreach (@output) {
            if ($_ =~ /Min Speed is (\d+)/) {
                $hash->{min_speed} = $1;

                # Store this value into memory for later reuse
                $self->{cache}{min_speed}{$port} = $hash->{min_speed};
            } elsif ($_ =~ /Max Speed is (\d+)/) {
                $hash->{max_speed} = $1;

                # Store this value into memory for later reuse
                $self->{cache}{max_speed}{$port} = $hash->{max_speed};
            }
        }
    }

    return $hash;
}

################################################################################
# Subroutine Name: _get_port_info_nexus()
# Objective:       Get information about the specified port/interface specific
#                  to Nexus FCOE/Ethernet
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_port_info_nexus {
    my ($self, $port, $hash) = @_;
    my (@output);

    @output = $self->{parent}->execute("show int $port");
    foreach my $line (@output) {
        if ($line =~ /$port is (\S+)/) {
            $hash->{state} = &SwitchUtil::normalize_state($1);
        }
        if ($line =~ /Hardware is (\w+)/i) {
            $hash->{hardware} = $1;
        }
        if ($line =~ /port mode is (\S+)/i) {
            my $type = $1;
            $type =~ s/,//;
            $hash->{type} = &SwitchUtil::normalize_type($type);
        }
        if ($line =~ /$port is \S+\s+\((.*)\)/) {
            $hash->{state_reason} = $1;
        }
        if ($line =~ /FCID is (\S+)/) {
            $hash->{fcid} = hex($1);
        }
        if ($line =~ /port vsan is (\d+)/i) {
            $hash->{vsan} = $1;
        }
        if ($line =~ /Bound interface is (\S+)/i) {
            $hash->{ethernet_port} = $1;
        }
    }

    # only supported speed is 10G
    $hash->{speed} = "10";

    # Load the cached values if we have already looked these values up before
    $hash->{min_speed} = $self->{cache}{min_speed}{$port}
        if defined($self->{cache}{min_speed}{$port});
    $hash->{max_speed} = $self->{cache}{max_speed}{$port}
        if defined($self->{cache}{max_speed}{$port});

    # hard-code to 10G
    $hash->{min_speed} = "10";
    $hash->{max_speed} = "10";

    $self->{cache}{min_speed}{$port} = $hash->{min_speed};
    $self->{cache}{max_speed}{$port} = $hash->{max_speed};

    return $hash;
}

################################################################################
# Subroutine Name: set_port_type()
# Objective:       Set the specified port to the specified port type
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_type {
    my ($self, %args) = @_;
    my ($status, $port, $type);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};

    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Error: CiscoFCPort::set_port_type() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: CiscoFCPort::set_port_type() " .
                                "Invalid value for argument '-type'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # receive a normalized port and here we convert to Cisco specific
    $type = $self->_get_port_type_from_normalized($type);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    my @types = $self->_get_supported_port_types($port);
    unless (grep(/$type/i, @types)) {
        $self->_set_err("Error: CiscoFCPort::set_port_type() " .
                        "Port type specified is not supported on this switch");
        return $SwitchUtil::FALSE;
    }

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    # check to see if port contain's "fc" and if not add. It might
    # not if coming from the tharnhost file.
    if ($port !~ /^\s*fc/ and
        $port !~ /^\s*vfc/ and
        $port !~ /bay|ext/ and
        $self->{parent}->get_model() !~ /nexus/i) {
        $port = "fc" . $port;
    }

    my @cmds = ("conf t",
                "int $port",
                "switchport mode $type",
                "end");

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoFCPort::set_port_type() " .
                            "error executing command ($cmd)");
            $cmd_error = 1;
            last;
        }
    }

    # handle error from command loop
    if ($cmd_error == 1) {
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: CiscoFCPort::set_port_type() " .
                        "Port state did not transtion to online state");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_dump_info()
# Objective:       Enable the port specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_dump_info {
    my ($self, $port, $debug, %dump_info) = @_;

    my ($cmd, $ip, $username, $password);

    $self->_clear_err();

    $self->sanlog(-msg => "CiscoFCPort::port_dump_info() Dumping log for port $port");

    # check to see if port contain's "fc" and if not add. It might
    # not if coming from the tharnhost file.
    if ($port !~ /^\s*fc/ and
        $port !~ /^\s*vfc/ and
        $port !~ /bay|ext/ and
        $self->{parent}->get_model() !~ /nexus/i) {
        $port = "fc" . $port;
    }

    $self->sanlog(-msg => "CiscoFCPort::port_dump_info() Dumping log for port $port");
    $cmd = "show port internal all interface $port";
    $self->{parent}->execute($cmd, 180);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to run command ($cmd)");

        # handle error from command loop
        $self->_cleanup_config_mode();
        return $SwitchUtil::FALSE;
    }

    $self->sanlog(-msg => "CiscoFCPort::port_dump_info() Dumping logging information of the ports ");
    $cmd = "show logging last 50";
    $self->{parent}->execute($cmd, 60);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to run command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    if ($debug == 1) {
        # execute the prepared command if the flag FAULT_SWITCH_PORT_DEBUG_INFO_ON_FAILURE is set
        $self->sanlog(-msg => "CiscoFCPort::port_dump_info() Dumping switch information on the switch for debug purposes ");
        $cmd = "show tech-support";
        my $subtest_rc = subtest( sub {$self->{parent}->execute($cmd, 1200)},
                                           -runid, "tech_support_information");
        if ($self->{parent}->error) {
            $self->_set_err("Unable to run command ($cmd)");
            return $SwitchUtil::FALSE;
        }
   }

   return $SwitchUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new port object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: CiscoFCPort::_new() " .
                                "Invalid value for argument '-parent'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/i) {
            $self->_set_err("Error: CiscoFCPort::_new() " .
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: CiscoFCPort::_new() " .
                        "Missing required argument '-parent'");

        return $self;
    }

    # initialize the object members
    $self->{parent} = $parent;

    $self->{nameserver} = $self->get_nameserver();

    return ($self);
}

###############################################################################
# Method:    _get_nameserver_database
# Objective: Retrieve nameserver database information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_nameserver_database {
    my ($self, %args) = @_;
    my $index;
    my $ns     = {};

    # If NPV, show fcs database command does not exist and lookup
    # not possible, simply return with empty NS
    if ($self->{parent}->is_npv_enabled()) {
        return $ns;
    }

    my $cmd = "show fcs database";
    my @output = $self->{parent}->execute("$cmd", 600);
    if ($self->{parent}->error) {
        $self->_set_err("Error: CiscoFCPort::_get_nameserver_database() " .
                        "error executing command ($cmd)");
    }

    # We will collect flogi output later if needed in case of L-ports
    $cmd = "show flogi database";
    my @flogi_output = $self->{parent}->execute("show flogi database", 600);
    if ($self->{parent}->error) {
        $self->_set_err("Error: CiscoFCPort::_get_nameserver_database() " .
                        "error executing command ($cmd)");
    }

    my @temp_output = ();

    # Search for start of interface information and weed it out
    #-------------------------------------------------------------------
    #Interface  fWWN                     Type     Attached-pWWNs
    #-------------------------------------------------------------------
    foreach my $line (@output) {
        if ($line !~ /^Interface|^--------/) {
            push (@temp_output, $line);
        }
    }
    @output = @temp_output;
    # Clear out temp stuff for use later
    @temp_output = ();

    # Search for start of interface information and truncate the output
    #---------------------------------------------------------------------------
    #INTERFACE  VSAN    FCID            PORT NAME               NODE NAME
    #---------------------------------------------------------------------------
    foreach my $line (@flogi_output) {
        if ($line !~ /^INTERFACE|^--------/) {
            push (@temp_output, $line);
        }
    }
    @flogi_output = @temp_output;

    # Clear out any caching we may care about
    $self->{cache}{fcns_database} = undef;

    # Each line of the output should look like the following:
    # fc1/5      20:05:00:0d:ec:00:22:80  F        20:00:00:01:73:81:37:75
    # 9124e (HP/Cisco):
    # bay2       20:10:00:0d:ec:41:a9:00  F        21:01:00:1b:32:20:5b:76
    my ($last_interface, $flogi_count);
    my $vsan_id = "";
    foreach my $line (@output) {
        if ($line =~ /FCS Local Database in VSAN:\s+(\d+)/i) {
            # Save the vsan number to store
            $vsan_id = $1;
        } elsif ($line !~ /Interface/ and
            ($line =~ /\d+\/\d+/ or $line =~ /bay|ext/i or $line =~ /vfc\d+/) and
            $line =~ /(\S+)\s+\S+\s+(\S+)\s+(\S+)/) {
            my $hash      = {};
            my $interface = $1;
            my $raw_wwpn = $3;

            $hash->{type} = $2;
            $hash->{wwpn} = lc($3);
            $hash->{wwpn} =~ s/://g;

            # Save the vsan in the fid attribute
            $hash->{fid} = $vsan_id;

            # Set the protocol
            if ($interface =~ /^vfc/i) {
                $hash->{protocol} = "FCoE";
            } elsif ($interface =~ /^fc/i) {
                $hash->{protocol} = "FC";
            } else {
                $hash->{protocol} = "NotAvailable";
            }

            # If it the port is of type F, we also need to check if it FCVI
            if ($hash->{type} =~ /^F$/i) {
                if (($self->_is_fport_wwpn_FCVI($raw_wwpn)) == $SwitchUtil::TRUE) {
                    $hash->{type} = "F_Port-FC_VI";
                }
            }

            # Check if a port is a member of a port channel , and change the port type to E
            # fc6/1          21:41:00:05:30:01:be:b2  TF       Port Channel Member
            if ($hash->{type} =~ /^TF/i and $interface =~ /^\s*fc\d+\/\d+/) {
                $hash->{type} = "E";
            }

            # We may need to do collect some extra info if it is a disk port
            # attached to the switch
            if ($hash->{type} =~ /^FL$/i) {
                $hash->{count} = 1;
                ($hash->{wwnn}, $hash->{address}) =
                    $self->_get_wwnn_and_address_from_flogi(\@flogi_output,
                                                            $hash->{wwpn});
            }
            $ns->{$interface} = $hash;
            $last_interface = $interface;
        } elsif ($line =~ /^\s+([\d|\w+|\:]+)$/) {
            # handle case for NPV where you have multiple WWPNs on the
            # same physical port. Example:
            # fc9/43     22:2b:00:0c:85:67:b1:80  F        20:18:00:0d:ec:3d:95:40
            #                                              10:00:00:00:c9:51:27:67
            # add the new WWPN to the list of WWPNs already existing for
            # that interface.
            my $last_wwpn = $ns->{$last_interface}->{wwpn};
            my $add_wwpn = $1;
            $add_wwpn = lc($add_wwpn);
            $add_wwpn =~ s/://g;

            if ($ns->{$last_interface}->{type} =~ /^FL$/i) {
                # this could be a disk on loop where you have multiple
                # entries in NS for each disk. If we see a count > 1, add
                # an index to the port
                $ns->{$last_interface}->{count} =
                    ($ns->{$last_interface}->{count} + 1);
                if ($ns->{$last_interface}->{count} > 1) {
                    my $int_FL = $last_interface . "_" .
                                 $ns->{$last_interface}->{count};
                    $ns->{$int_FL}->{kind} = "disk";
                    ($ns->{$int_FL}->{wwnn}, $ns->{$int_FL}->{address}) =
                        $self->_get_wwnn_and_address_from_flogi(\@flogi_output,
                                                                $add_wwpn);
                    $ns->{$int_FL}->{type} = $ns->{$last_interface}->{type};
                }
            } else {
                $ns->{$last_interface}->{wwpn} = $last_wwpn . ";" . $add_wwpn;
            }
        }
    }

    return $ns;
}

##############################################################################
# Method:    _is_fport_wwpn_FCVI
# Objective: Check if a given fport's wwpn is used for FCVI.
# Details:   See POD documentation at the end of this file.
##############################################################################
sub _is_fport_wwpn_FCVI {
    my ($self, $wwpn) = @_;

    my $is_fcvi = $SwitchUtil::FALSE;
    my @output = ();
    if (defined $self->{cache}{fcns_database}) {
        @output = @{$self->{cache}{fcns_database}};
    } else {
        my $cmd = "show fcns database local";
        @output = $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoFCPort::_is_fport_wwpn_FCVI " .
                            "error executing command ($cmd)");
            return $SwitchUtil::FALSE;
        }
        @{$self->{cache}{fcns_database}} = \@output;
    }

    # The format would look something like:
    # 0x021200    N     21:01:00:e0:8b:a7:03:9a (Qlogic)        scsi-fcp:init fc-vi
    foreach my $line (@output) {
        if ($line =~ /fc-vi/i and
            $line =~ /$wwpn/i) {
            $is_fcvi = $SwitchUtil::TRUE;
        }
    }

    return $is_fcvi;
}

##############################################################################
# Method:    _get_wwnn_and_address_from_flogi
# Objective: Retrieve a wwnn and FCID from given flogi output and wwpn
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_wwnn_and_address_from_flogi {
    my ($self, $flogi_output, $wwpn) = @_;
    my ($wwnn, $addr);
    $wwnn = $addr = "";

    foreach my $line (@{$flogi_output}) {
        # Parse this format
        #---------------------------------------------------------------------------
        #INTERFACE  VSAN    FCID            PORT NAME               NODE NAME
        #---------------------------------------------------------------------------
        #fc1/1      2     0x011300  21:01:00:e0:8b:a7:8c:98  20:01:00:e0:8b:a7:8c:98
        #fc1/2      2     0x011400  50:0a:09:80:00:02:54:99  50:0a:09:80:00:02:54:99
        #fc1/5      2     0x0111b5  22:00:00:0c:50:fe:ec:9e  20:00:00:0c:50:fe:ec:9e
        #fc1/5      2     0x0111b6  22:00:00:00:87:44:00:28  20:00:00:00:87:44:00:28
        if ($line !~ /^Total/i and
            $line =~ /\S+\s+\S+\s+(\S+)\s+(\S+)\s+(\S+)/) {
            my $add_wwpn = $2;
            $wwnn = lc($3);
            $addr = lc($1);
            $add_wwpn = lc($add_wwpn);
            $add_wwpn =~ s/://g;
            $wwnn =~ s/://g;

            if ($add_wwpn =~ /$wwpn/i) {
                return ($wwnn, $addr);
            }
        }
    }

    # Return empty (wwnn and addr)
    return ($wwnn, $addr);
}

###############################################################################
# Method:    _get_supported_port_types
# Objective: Retrieve the supported port types for this switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_supported_port_types {
    my ($self, $port) = @_;
    my (@output, @types);

    if (defined($self->{cache}{port_types})) {
        @types = @{$self->{cache}{port_types}};
    } else {
        my $original_persistent = $self->{parent}->{persistent};

        # Make connection persistent so commands can be sent through one
        # connection session
        $self->{parent}->options(-persistent => 1);

        # check to see if port contain's "fc" and if not add. It might
        # not if coming from the tharnhost file.
        if ($port !~ /^\s*fc/ and
            $port !~ /^\s*vfc/ and
            $port !~ /bay|ext/ and
            $self->{parent}->get_model() !~ /nexus/i) {
            $port = "fc" . $port;
        }

        my @cmds = ("conf t",
                    "int $port", "switchport mode ?");

        my $cmd_error = 0;
        foreach my $cmd (@cmds) {
            @output = $self->{parent}->execute($cmd);
            if ($self->{parent}->error) {
                $self->_set_err("Error: CiscoFCPort::_get_supported_port_types() " .
                                "error executing command ($cmd)");
                $cmd_error = 1;
                last;
            }
        }

        # handle error from command loop
        if ($cmd_error == 1) {
            $self->_cleanup_config_mode();
            return $SwitchUtil::FALSE;
        }

        # Get out of config mode to ensure next commmands sent to switch
        # work correctly. This doesn't really send 'end' but appends to the
        # previous unfinished command. While it results in an invalid command
        # it achieves the same result of getting us out of config mode
        $self->{parent}->execute("end");

        # Restore persistent setting
        $self->{parent}->options(-persistent => $original_persistent);

        foreach my $line (@output) {
            push(@types, (split(/\s+/, $line))[1]);
        }

        # Cache values to minimize repeat lookups
        $self->{cache}{port_types} = \@types;
    }

    return (wantarray()) ? @types : \@types;
}

###############################################################################
# Method:    _cleanup_config_mode
# Objective: If there was an error in processing a list of commands, this
#            function will attempt to get out of Cisco config mode.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cleanup_config_mode {
    my ($self) = @_;

    my (@cmds, @output);

    # Attempt to get out of config mode, by simply sending newline
    # followed by end command.
    @cmds = (" ", "end");
    foreach my $cmd (@cmds) {
        @output = $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: CiscoFCPort::_cleanup_config_mode() " .
                            "error executing command ($cmd)");
        }
    }
}

###############################################################################
# Method:    _get_port_type_from_normalized
# Objective: Take in a normalized port name and return one specific to Cisco
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_port_type_from_normalized {
    my ($self, $type) = @_;

    if ($type eq "F_Port") {
        return "F";
    }

    if ($type eq "FL_Port") {
        return "FL";
    }

    if ($type eq "E_Port") {
        return "E";
    }

    if ($type eq "G_Port") {
        return "G";
    }

    if ($type eq "N_Port") {
        return "NP";
    }

    if ($type =~ /auto/i) {
        return "auto";
    }

    # if nothing matched set error
    $self->_set_err("Error: CiscoFCPort::_get_port_type_from_normalized() " .
                    "($type) cannot be mapped to a valid Cisco port type");
    return "";

}

1;

