# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for ports for a McData FC switch
## @author thomasm@netapp.com
## @status private
## @pod ./Port.pod

# Package Name
##############
package San::SanSwitch::FC::PortMgr::McDataFCPort;

use strict;

# Package Requirements
######################
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::PortMgr;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog PortMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_port_num()
# Objective:       Retrieve the port number/interface corresponding to a 
#                  WWPN, by returning the first match in the array.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_num {
    my ($self, %args) = @_;
    my (@ports);

    @ports = $self->get_port_nums(%args);
    return "" if ($self->error);

    return $ports[0];

}

################################################################################
# Subroutine Name: get_port_nums()
# Objective:       Retrieve the all port number/interface corresponding to a 
#                  WWPN. Multiple's can be returned when using looking for
#                  e-ports and a trunk is established.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_nums {
    my ($self, %args) = @_;
    my ($wwpn, $found_port, @found_ports);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwpn/) {
            if (not length($wwpn = $args{$_})) {
                $self->_set_err("Error: McDataFCPort::get_port_nums() " .
                                "Invalid value for argument '-wwpn'");
                return ();
            }
            next;
        }
    }

    $self->{nameserver} = $self->get_nameserver();
    if (defined($wwpn)) {
        $self->sanlog(-msg => "Searching for WWPN $wwpn");
        $wwpn = lc($wwpn);
        $wwpn =~ s/://g;
        foreach my $interface (keys %{$self->{nameserver}}) {
            if ($self->{nameserver}{$interface}{wwpn} eq $wwpn) {
                $found_port = $interface;
                $self->sanlog(-msg => "Found match on port $found_port");
                push (@found_ports, $found_port);
            }
        }
    } else {
        $self->_set_err("Error: McDataFCPort::get_port_nums() WWPN must be specified");
        return ();
    }

    if (!(@found_ports)) {
        $self->_set_err("Error: McDataFCPort::get_port_nums() no matching port found");
        return ();
    } else {
        return @found_ports;
    }
}

################################################################################
# Subroutine Name: port_disable()
# Objective:       Disable the port/interface specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_disable {
    my ($self, $port) = @_;
    my ($status, $output);

    $self->_clear_err();

    # if requested, get current state to store speed and topology
    if ($self->{parent}->{verify_on_enable} == 1) {
        my $port_state = $self->get_port_info($port);
        if ($self->error) {
            $self->_set_err("Error: McDataFCPort::port_disable() ".
                            "error getting port state ($port)");
            return $SwitchUtil::FALSE;
        }
        $self->{previous_speed} = $port_state->{speed};
        $self->{previous_type} = $port_state->{type};
    }

    # block the specified port
    my $cmd = "config port blocked $port true";
    $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->{previous_speed} = undef;
        $self->{previous_type} = undef;
        $self->_set_err("Error: McDataFCPort::port_disable() ".
                        "error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    # wait for the port to transition to offline, or report error
    $status = $self->waitfor_port($port, "offline");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: McDataFCPort::port_disable() " .
                        "failed to enable port, port still appears up");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_enable()
# Objective:       Enable the port specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_enable {
    my ($self, $port) = @_;
    my ($status);

    $self->_clear_err();

    my $cmd = "config port blocked $port false";
    $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error: McDataFCPort::port_enable() " .
                        "error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    # add explicit sleep into the code here before checking the                                             
    # port to see if it is online (suggested by Dave/Sunil)                                                 
    sleep (3);                                                                                              

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: McDataFCPort::port_enable() " .
                        "failed to enable port, port still appears down");
        return $SwitchUtil::FALSE;
    }

    # compare values now (after enable) versus before the
    # port_disable
    $self->check_port_settings($port);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: set_port_speed()
# Objective:       Set the specified port to the specified speed
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_speed {
    my ($self, %args) = @_;
    my ($status, $port, $speed, $cmd);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Error: McDataFCPort::set_port_speed() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?speed/) {
            if ((not length($speed = $args{$_})) && 
                ($args{$_} !~ /^\d$/ || $speed !~ /auto/i)) {
                $self->_set_err("Error: McDataFCPort::set_port_speed() " .
                                "Invalid value for argument '-speed'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # handle auto speed setting
    if ($speed =~ /AUTO/i) {
        $speed = "negotiate";
    }

    # McDATA requires G at end of speed, check and set if not there.
    if ($speed !~ /G\s*$/i and $speed !~ /negotiate/i) {
        $speed = $speed . "G";
    }

    $cmd = "config port speed $port $speed";
    $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error: McDataFCPort::set_port_speed() " .
                        "error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: McDataFCPort::set_port_speed() " .
                        "Port state did not transtion to online state");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: get_port_info()
# Objective:       Get information about the specified port/interface
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_info {
    my ($self, $port) = @_;
    my (@output, @sorted, @speeds, $speed_capabilities);

    $self->_clear_err();

    # Initialize hash
    my $hash = {};
    $hash->{state}        = "unknown";
    $hash->{state_reason} = "";
    $hash->{speed}        = 0;
    $hash->{type}         = "unknown";
    $hash->{fcid}         = 0;
    $hash->{min_speed}    = 0;
    $hash->{max_speed}    = 0;

    my $cmd = "show port profile $port";
    @output = $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error: McDataFCPort::get_port_info() " .
                        "error executing command ($cmd)");
        return {};
    }

    foreach my $line (@output) {
        if ($line =~ /Operating Speed:\s+(.*)/i) {
            my $speed = $1;
            if ($speed !~ /Not Established/i) {
                $speed =~ s/G//;
                $hash->{speed} = int($speed);
            }
            next;
        }
        if ($line =~ /Blocked Reason:\s+(\S+)/i) {
            $hash->{state_reason} = $1;
            next;
        }
        if ($line =~ /Operational State:\s+(\S+)/) {
            $hash->{state} = &SwitchUtil::normalize_state($1);
            next;
        }
        if ($line =~ /FC Address:\s+(\S+)/i) {
            my $addr = $1;
            if ($addr !~ /N\/A/i) {
                $hash->{fcid} = hex($addr);
            } else {
                $hash->{fcid} = $addr;
            }
            next;
        }
        if ($line =~ /Configured Type:\s+(\S+)/i) {
            $hash->{type} = &SwitchUtil::normalize_type($1);
            next;
        }
        if ($line =~ /Speed Capability:\s+(.*)/i) {
            $speed_capabilities = $1;
            next;
        }
    }

    # Load the cached values if we have already looked these values up before
    $hash->{min_speed} = $self->{cache}{min_speed}{$port} 
        if defined($self->{cache}{min_speed}{$port});
    $hash->{max_speed} = $self->{cache}{max_speed}{$port} 
        if defined($self->{cache}{max_speed}{$port});

    if ($hash->{min_speed} == 0 || $hash->{max_speed} == 0) {
        $speed_capabilities =~ s/G//g;
        @speeds = split(/\s+/, $speed_capabilities);

        # sort numerically
        @sorted = sort {$a <=> $b} @speeds;
                
        $hash->{min_speed} = $sorted[0];
        # Store this value into memory for later reuse
        $self->{cache}{min_speed}{$port} = $hash->{min_speed};

        $hash->{max_speed} = $sorted[(scalar(@sorted) - 1)];
        # Store this value into memory for later reuse
        $self->{cache}{max_speed}{$port} = $hash->{max_speed};
    }

    return $hash;
}

################################################################################
# Subroutine Name: set_port_type()
# Objective:       Set the specified port to the specified port type
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_type {
    my ($self, %args) = @_;
    my ($status, @output, $cmd, $port, $type, @types);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Error: McDataFCPort::set_port_type() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: McDataFCPort::set_port_type() " .
                                "Invalid value for argument '-type'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # receive a normalized port and here we convert to McData specific
    $type = $self->_get_port_type_from_normalized($type);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # make sure requested port type to change to is valid
    @types = $self->_get_supported_port_types($port);
    unless (grep(/$type/i, @types)) {
        $self->_set_err("Error: McDataFCPort::set_port_type() " .
                        "Port type specified is not supported on this switch");
        return $SwitchUtil::FALSE;
    }

    # if current and asked for type are same, just return
    my $prev_type = $self->get_port_info($port);
    if ($prev_type->{type} eq $type) {
        $self->sanlog(-msg => "McDataFCPort::set_port_type() Port type is already set to $prev_type->{type}, no action needed");
        return $SwitchUtil::TRUE;
    }

    $cmd = "config port type $port $type";
    @output = $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error: McDataFCPort::set_port_type() " .
                        "error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: McDataFCPort::set_port_type() " .
                        "Port state did not transtion to online state");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_dump_info()
# Objective:       NO-OP for McData
# Details:         See POD documentation at the end of this file
################################################################################
sub port_dump_info {
    my ($self, $port) = @_;

    $self->_clear_err();

    return $SwitchUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new port object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: McDataFCPort::_new() " .
                                "Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/i) {
            $self->_set_err("Error: McDataFCPort::_new() " .
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: McDataFCPort::_new() " .
                        "Missing required argument '-parent'");
        return $self;
    }

    # initialize the object members
    $self->{parent} = $parent;

    $self->{nameserver} = $self->get_nameserver();

    return $self;
}

###############################################################################
# Method:    _get_nameserver_database
# Objective: Retrieve nameserver database information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_nameserver_database {
    my ($self) = @_;

    my ($output, $line);
    my ($port, $type, $wwpn, $wwnn);

    my $ns = {};

    my $cmd = "show port status";
    my @output = $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error: McDataFCPort::_get_nameserver_database() " .
                        "error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    # Example output (4500)
    # Root> show port status
    # Port  State             Type    Attached WWN             Beaconing  Reason
    # ----  ----------------  ------  -----------------------  ---------  ------
    # 0     Online            ePort   10:00:08:00:88:E3:03:72  Disabled
    # 1     No Light          gxPort  None                     Disabled

    foreach $line (@output) {

        # all lines we care about start with digit
        if ($line !~ /^\d+/) {
            next;
        }

        # remove error conditions
        if ($line =~ /Module not installed/ or 
            $line =~ /Optics not installed/ or
            $line =~ /Unaddressable/) {
            next;
        }

        # put line in a format that we can split deterministically
        # "Not Installed" and "No Light" can cause problems.
        $line =~ s/Not Installed/Not-Installed/;
        $line =~ s/No Light/No-Light/;

        # split line into valid fields
        my @fields = split(/\s+/, $line);

        # Convert to lowercase since that's what the other switch vendors do
        my $wwpn = lc $fields[3];
        $wwpn =~ s/://g;

        # get type
        my $type = $fields[2];

        # get port
        my $port = $fields[0];

        # get state
        my $state = $fields[1];

        # store information in hash
        $ns->{$port}->{wwpn} = $wwpn;
        $ns->{$port}->{type} = $type;
        $ns->{$port}->{state} = $state;
    }

    return $ns;

}

###############################################################################
# Method:    _get_supported_port_types
# Objective: Retrieve the supported port types for this switch - for McData
#            hardcode since no command exists to get the capabilities
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_supported_port_types {
    my ($self, $port) = @_;
    my (@types);

    if (defined($self->{cache}{port_types})) {
        @types = @{$self->{cache}{port_types}};
    } else {
        @types = ("fPort", "ePort", "gxPort", 
                  "nPort", "flPort");

        # Cache values to minimize repeat lookups
        $self->{cache}{port_types} = [ @types ];
    }

    return (wantarray()) ? @types : \@types;
}

###############################################################################
# Method:    _get_port_type_from_normalized
# Objective: Take in a normalized port name and return one specific to McData
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_port_type_from_normalized {
    my ($self, $type) = @_;

    if ($type eq "F_Port") {
        return "fPort";
    }

    if ($type eq "FL_Port") {
        return "flPort";
    }

    if ($type eq "E_Port") {
        return "ePort";
    }
 
    if ($type eq "G_Port") {
        return "gxPort";
    }

    if ($type eq "N_Port") {
        return "nPort";
    }

    # closest to auto would be gPort
    if ($type =~ /auto/i) {
        return "gPort";
    }

    # if nothing matched set error
    $self->_set_err("Error: McDataFCPort::_get_port_type_from_normalized() " .
                    "($type) cannot be mapped to a valid McData port type");
    return "";

}

1;
