# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for ports for a Qlogic FC switch
## @author thomasm@netapp.com
## @status private
## @pod ./Port.pod

# Package Name
##############
package San::SanSwitch::FC::PortMgr::QlogicFCPort;

use strict;

# Package Requirements
######################
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::PortMgr;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog PortMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_port_num()
# Objective:       Retrieve the port number/interface corresponding to a 
#                  WWPN, by returning the first match in the array.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_num {
    my ($self, %args) = @_;
    my (@ports);

    $self->_clear_err();

    @ports = $self->get_port_nums(%args);
    return "" if ($self->error);

    return $ports[0];

}

################################################################################
# Subroutine Name: get_port_nums()
# Objective:       Retrieve the all port number/interface corresponding to a 
#                  WWPN. Multiple's can be returned when using looking for
#                  e-ports and a trunk is established.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_nums {
    my ($self, %args) = @_;
    my ($wwpn, $found_port, @found_ports);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwpn/) {
            if (not length($wwpn = $args{$_})) {
                $self->_set_err("Error: QlogicFCPort::get_port_nums() " .
                                "Invalid value for argument '-wwpn'");
                return ();
            }
            next;
        }
    }

    $self->{nameserver} = $self->get_nameserver();
    if (defined($wwpn)) {
        $self->sanlog(-msg => "Searching for WWPN $wwpn");
        $wwpn = lc($wwpn);
        $wwpn =~ s/://g;
        foreach my $interface (keys %{$self->{nameserver}}) {
            if ($self->{nameserver}{$interface}{wwpn} eq $wwpn) {
                $found_port = $interface;
                $self->sanlog(-msg => "Found match on port $found_port");
                push (@found_ports, $found_port);
            }
        }
    } else {
        $self->_set_err("Error: QlogicFCPort::get_port_nums() WWPN must be specified");
        return ();
    }

    if (!(@found_ports)) {
        $self->_set_err("Error: QlogicFCPort::get_port_nums() no matching port found");
        return ();
    } else {
        return @found_ports;
    }
}

################################################################################
# Subroutine Name: port_disable()
# Objective:       Disable the port/interface specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_disable {
    my ($self, $port) = @_;
    my ($status);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};

    # if requested, get current state to store speed and topology
    if ($self->{parent}->{verify_on_enable} == 1) {
        my $port_state = $self->get_port_info($port);
        if ($self->error) {
            $self->_set_err("Error: QlogicFCPort::port_disable() ".
                            "error getting port state ($port)");
            return $SwitchUtil::FALSE;
        }
        $self->{$port}->{previous_speed} = $port_state->{speed};
        $self->{$port}->{previous_type} = $port_state->{type};
    }

    # block the specified port

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    my @cmds = ( "admin stop", "admin start", 
                 "set port $port state offline",
                 "admin stop" );

    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->{$port}->{previous_speed} = undef;
            $self->{$port}->{previous_type} = undef;
            $self->_set_err("Error: QlogicFCPort::port_disable() ".
                            "error executing command ($cmd)");
            return $SwitchUtil::FALSE;
        }
    }

    $self->{parent}->options(-persistent => $original_persistent);

    # wait for the port to transition to offline, or report error
    $status = $self->waitfor_port($port, "offline");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: QlogicFCPort::port_disable() " .
                        "failed to enable port, port still appears up");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: port_enable()
# Objective:       Enable the port specified
# Details:         See POD documentation at the end of this file
################################################################################
sub port_enable {
    my ($self, $port) = @_;
    my (@output, $status);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};
    my $count = 0;

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    my @cmds = ( "admin stop", "admin start", 
                 "set port $port state online",
                 "admin stop" );

    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: QlogicFCPort::port_enable() " .
                            "error executing command ($cmd)");
            return $SwitchUtil::FALSE;
        }
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    # add explicit sleep into the code here before checking the                                             
    # port to see if it is online (suggested by Dave/Sunil)                                                 
    sleep (3);                                                                                              

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: QlogicFCPort::port_enable() " .
                        "failed to enable port, port still appears down");
        return $SwitchUtil::FALSE;
    }

    # compare values now (after enable) versus before the
    # port_disable
    $self->check_port_settings($port);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: set_port_speed()
# Objective:       Set the specified port to the specified speed
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_speed {
    my ($self, %args) = @_;
    my ($port, $speed, $status);

    $self->_clear_err();

    my $original_persistent = $self->{parent}->{persistent};
    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("Error: QlogicFCPort::set_port_speed() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?speed/) {
            if ((not length($speed = $args{$_})) && 
                ($args{$_} !~ /^\d$/ || $speed !~ /auto/i)) {
                $self->_set_err("Error: QlogicFCPort::set_port_speed() " .
                                "Invalid value for argument '-speed'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # handle auto speed setting
    if ($speed =~ /AUTO/i) {
        $speed = "auto";
    }

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->{parent}->options(-persistent => 1);

    my @cmds = ( "admin stop", "admin start", 
                 "set port $port speed $speed",
                 "admin stop" );

    foreach my $cmd (@cmds) {
        $self->{parent}->execute($cmd);
        if ($self->{parent}->error) {
            $self->_set_err("Error: QlogicFCPort::set_port_speed() " .
                            "error executing command ($cmd)");
            return $SwitchUtil::FALSE;
        }
    }

    # wait for the port to transition to online, or report error
    $status = $self->waitfor_port($port, "online");
    if ($status == $SwitchUtil::FALSE) {
        $self->_set_err("Error: QlogicFCPort::set_port_speed() " .
                        "Port state did not transtion to online state");
        return $SwitchUtil::FALSE;
    }

    # Restore persistent setting
    $self->{parent}->options(-persistent => $original_persistent);

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: get_port_info()
# Objective:       Get information about the specified port/interface
# Details:         See POD documentation at the end of this file
################################################################################
sub get_port_info {
    my ($self, $port) = @_;

    my (@output, $speed_capabilities);
    my (@sorted, @speeds);

    $self->_clear_err();

    # Initialize hash
    my $hash = {};
    $hash->{state}        = "unknown";
    $hash->{state_reason} = "";
    $hash->{speed}        = 0;
    $hash->{type}         = "unknown";
    $hash->{fcid}         = 0;
    $hash->{min_speed}    = 0;
    $hash->{max_speed}    = 0;

    @output = $self->{parent}->execute("show port $port");
    foreach my $line (@output) {
        if ($line =~ /LinkSpeed\s+(\d+)Gb\/s/) {
            $hash->{speed} = int($1);
        }
        if ($line =~ /EpIsoReason\s+(\S+)/) {
            $hash->{state_reason} = $1;
        }
        if ($line =~ /OperationalState\s+(\S+)/) {
            $hash->{state} = &SwitchUtil::normalize_state($1);
        }
        if ($line =~ /PortID\s+(\S+)/) {
            $hash->{fcid} = hex($1);
        }
        if ($line =~ /RunningType\s+(\S+)/) {
            $hash->{type} = &SwitchUtil::normalize_type($1);
        }
        if ($line =~ /MediaSpeeds/) {
            $speed_capabilities = $line;
        }
    }

    # Load the cached values if we have already looked these values up before
    $hash->{min_speed} = $self->{cache}{min_speed}{$port} 
        if defined($self->{cache}{min_speed}{$port});
    $hash->{max_speed} = $self->{cache}{max_speed}{$port} 
        if defined($self->{cache}{max_speed}{$port});

    if ($hash->{min_speed} == 0 || $hash->{max_speed} == 0) {
        $speed_capabilities =~ s/Gb\/s//g;
        $speed_capabilities =~ s/^\s+MediaSpeeds\s+//;
        $speed_capabilities =~ s/\s+XmitterEnabled.*//;
        @speeds = split(/,/, $speed_capabilities);

        # sort numerically
        @sorted = sort {$a <=> $b} @speeds;
                
        $hash->{min_speed} = $sorted[0];
        # Store this value into memory for later reuse
        $self->{cache}{min_speed}{$port} = $hash->{min_speed};

        $hash->{max_speed} = $sorted[(scalar(@sorted) - 1)];
        # Store this value into memory for later reuse
        $self->{cache}{max_speed}{$port} = $hash->{max_speed};
    }

    return $hash;
}

################################################################################
# Subroutine Name: set_port_type()
# Objective:       Set the specified port to the specified port type
# Details:         See POD documentation at the end of this file
################################################################################
sub set_port_type {
    my ($self, %args) = @_;
    my ($port, $type);
    my ($match, $before, $after);
    my ($start, @cmds);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?port/) {
            if (not length($port = $args{$_})) {
                $self->_set_err("$start, @cmds, Error: QlogicFCPort::set_port_type() " .
                                "Invalid value for argument '-port'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: QlogicFCPort::set_port_type() " .
                                "Invalid value for argument '-type'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # receive a normalized port and here we convert to QLogic specific
    $type = $self->_get_port_type_from_normalized($type);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # make sure requested port type to change to is valid
    my @types = $self->_get_supported_port_types($port);
    unless (grep(/$type/i, @types)) {
        $self->_set_err("Error: QlogicFCPort::set_port_type() " .
                        "Port type specified is not supported on this switch");
        return $SwitchUtil::FALSE;
    }


    # if current and asked for type are same, just return
    my $prev_type = $self->get_port_info($port);
    if ($prev_type->{type} eq $type) {
        $self->sanlog(-msg => "QlogicFCPort::set_port_type() Port type is already set to $prev_type->{type}, no action needed");
        return $SwitchUtil::TRUE;
    }

    # get into admin mode
    @cmds = ( "admin stop", "admin start", "config edit" );
    foreach my $cmd (@cmds) {
        my @output = $self->{parent}->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: QlogicFCPort::set_port_type() " .
                            "Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
    }

    # to do the set of port type, we need to get down to using say
    # since we are being asked questions.
    my $connection = $self->{parent}->get_connection();
    $start = 1;
    ($match,$before,$after) = &Tharn::say($connection, "set config port $port",
                                          '-prompt', "]",
                                          '-timeout', 30);
    my $error = 0;
    while ($start == 1) {
        if ($after =~ /PortType/i) {
            ($match,$before,$after) = &Tharn::say($connection, $type,
                                                  '-prompt', "]",
                                                  '-timeout', 30);
        } elsif ($after =~ /To discard this configuration/i) {
            @cmds = ( "config save", "config activate", "admin stop" );
            foreach my $cmd (@cmds) {
                my @output = $self->{parent}->execute($cmd);
                if ($self->{parent}->error) {
                    $self->_set_err("Error: QlogicFCPort::set_port_type() " .
                                    "error executing command ($cmd)");
                    $error = 1;
                    last;
                }
            }
            $start = 0;
        } elsif ($after =~ /Illegal or out of range value/i) {
             $self->_set_err("Error: QlogicFCPort::set_port_type() " .
                             "Illegal setting of port type");
             $error = 1;
             last;
        } else {
            ($match,$before,$after) = &Tharn::say($connection, " ",
                                                  '-prompt', "]",
                                                  '-timeout', 30);
        }
    }

    # error could get from config loop
    if ($error == 1) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }
}

################################################################################
# Subroutine Name: port_dump_info()
# Objective:       NO-OP for Qlogic
# Details:         See POD documentation at the end of this file
################################################################################
sub port_dump_info {
    my ($self, $port) = @_;

    $self->_clear_err();

    return $SwitchUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new port object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: QlogicFCPort::_new() " .
                                "Invalid value for argument '-parent'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/i) {
            $self->_set_err("Error: QlogicFCPort::_new() " .
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: QlogicFCPort::_new() " .
                        "Missing required argument '-parent'");

        return $self;
    }

    # initialize the object members
    $self->{parent} = $parent;

    $self->{nameserver} = $self->get_nameserver();

    return ($self);
}

###############################################################################
# Method:    _get_nameserver_database
# Objective: Retrieve nameserver database information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_nameserver_database {
    my ($self) = @_;

    my $ns = {};

    # 2 phase. Use show ns to get standard (non e-port(ISL))
    # and use show topology to get e-port information.
    my $cmd = "show ns";
    my @output = $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error: QlogicFCPort::_get_nameserver_database() " .
                        "error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    foreach my $line (@output) {
        # skip non-port lines
        if ($line !~ /^\s+\d+/) {
            next;
        } else {
            $line =~ s/^\s+//;
        }

        # SANbox #> show ns
        # Seq Domain    Port   Port
        # No  ID        ID     Type COS PortWWN                 NodeWWN
        # --- ------    ------ ---- --- -------                 -------
        # 1   4 (0x4)   040000 N    3   50:0a:09:82:96:a7:a7:66 50:0a:09:80:...
        # 2   4 (0x4)   040100 N    3   50:0a:09:84:96:a7:a7:66 50:0a:09:80:...
        my @fields = split(/\s+/, $line);
        my $port_id = $fields[3];
        my $type = $fields[4];
        my $wwpn = $fields[6];
        my $wwnn = $fields[7];

        my $port = substr $port_id, 2, 2;
        $port = hex $port;             

        # Convert to lowercase since that's what the other switch vendors do
        $wwpn = lc $wwpn;
        $wwpn =~ s/://g;

        $wwnn = lc $wwnn;
        $wwnn =~ s/://g;

        # store information in hash
        $ns->{$port}->{wwpn} = $wwpn;
        $ns->{$port}->{wwnn} = $wwnn;
        $ns->{$port}->{address} = $port_id;
        $ns->{$port}->{type} = $type;
    }

    # get e-port information and add to ns hash
    $cmd = "show topology";
    @output = $self->{parent}->execute($cmd);
    if ($self->{parent}->error) {
        $self->_set_err("Error: QlogicFCPort::_get_nameserver_database() " .
                        "error executing command ($cmd)");
        return $SwitchUtil::FALSE;
    }

    foreach my $line (@output) {
        # skip non-port lines
        if ($line !~ /^\s+\d+/) {
            next;
        } else {
            $line =~ s/^\s+//;
        }

        # SANbox #> show topology
        # Port    Loc   Local                   Rem  Remote                  Unique
        # Number  Type  PortWWN                 Type NodeWWN                 ID
        # ------  ----  -------                 ---- -------                 ------
        # 97      E     20:61:00:c0:dd:0c:ad:b6 E    10:00:00:c0:dd:0c:ca:7d 3(0x3) D
        # 99      E     20:63:00:c0:dd:0c:ad:b6 E    10:00:00:c0:dd:0c:ca:7d 3(0x3) D
        my @fields = split(/\s+/, $line);
        my $port_id = $fields[5];

        my $port = $fields[0];
        # strip of Ext and Bay (embedded blade switch)
           $port =~ s/Bay//gi;
           $port =~ s/Ext//gi;

        my $type = $fields[1];
        my $wwpn = $fields[4];
        my $wwnn = "NA";

        # Convert to lowercase since that's what the other switch vendors do
        $wwpn = lc $wwpn;
        $wwpn =~ s/://g;

        # only add if e-port
        if ($type eq "E") {
            # store information in hash
            $ns->{$port}->{wwpn} = $wwpn;
            $ns->{$port}->{wwnn} = $wwnn;
            $ns->{$port}->{address}  = $port_id;
            $ns->{$port}->{type} = $type;
        } else {
            # return true port type versus nameserver
            $ns->{$port}->{type} = $type;
        }
    }

    return $ns;

}

###############################################################################
# Method:    _get_supported_port_types
# Objective: Retrieve the supported port types for this switch - for Qlogic
#            hardcode since no command exists to get the capabilities
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_supported_port_types {
    my ($self, $port) = @_;
    my (@types);

    if (defined($self->{cache}{port_types})) {
        @types = @{$self->{cache}{port_types}};
    } else {
        @types = ("F", "FL", "G", "N", "E", "GL");

        # Cache values to minimize repeat lookups
        $self->{cache}{port_types} = [ @types ];
    }

    return (wantarray()) ? @types : \@types;
}

###############################################################################
# Method:    _get_port_type_from_normalized
# Objective: Take in a normalized port name and return one specific to Qlogic
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_port_type_from_normalized {
    my ($self, $type) = @_;

    if ($type eq "F_Port") {
        return "F";
    }

    if ($type eq "FL_Port") {
        return "FL";
    }

    if ($type eq "E_Port") {
        return "E";
    }
 
    if ($type eq "G_Port") {
        return "G";
    }

    if ($type eq "N_Port") {
        return "N";
    }

    # if nothing matched set error
    $self->_set_err("Error: QlogicFCPort::_get_port_type_from_normalized() " .
                    "($type) cannot be mapped to a valid Cisco port type");
    return "";

}

1;
