# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a Apcon switch
## @author Michael M. Latif (latif@netapp.com)
## @status private
## @pod ../Switch.pod

use strict;

# Package Name
##############
package San::SanSwitch::FC::Switch::ApconFCSwitch;

# Package Requirements
######################
use San::SanSwitch;
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanSwitch);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

###############################################################################
# Method:    get_firmware_version
# Objective: Get the firmware version of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_firmware_version {
    my ($self) = @_;
    my (@output, $output);
    my $version = "unknown";

    $self->_clear_err();

    # Example output of show switch command
    # apcon_bh_1>> show switch
    # Switch Name:                    apcon_bh_1
    # 
    # Switch Model:                   ACI-2058
    # Switch Manufacture Date:        2008-02-27
    # Switch Serial Number:           5801239
    # Motherboard Serial Number:      63M1367
    # 
    # Firmware Version:               214 build 0044
    # 
    # Primary IP Address
    # IP Address:                     10.60.242.26
    # Subnet Mask:                    255.255.252.0
    # Gateway:                        10.60.240.1

    # Power Supply 1:                 No Alarm
    # Power Supply 2:                 No Alarm
    # Power Supply 3:                 No Alarm

    # Temperature:                    Within Limit
    # Current:                        22.7 degrees Celsius
    # Limit:                          50.0 degrees Celsius

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_switch})) {
        @output = $self->execute("show switch");
        $self->{cache}{show_switch} = [ @output ];
    } else {
        @output = @{$self->{cache}{show_switch}};
    }

    if (!$self->error) {
        $version = (grep(/Firmware Version\:/i, @output))[0];
        $version =~ s/Firmware Version\:\s+//i;
    } else {
        $self->_set_err("Unable to execute show switch command");
    }

    return $version;
}

###############################################################################
# Method:    get_model
# Objective: Get the model of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_model {
    my ($self) = @_;
    my (@output, $output);
    my $model = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_switch})) {
        @output = $self->execute("show switch");
        $self->{cache}{show_switch} = [ @output ];
    } else {
        @output = @{$self->{cache}{show_switch}};
    }

    if (!$self->error) {
       $model = (grep(/Switch Model\:/i, @output))[0];
       $model =~ s/Switch Model\:\s+//i;
    } else {
        $self->_set_err("Unable to execute show switch command");
    }

    return $model;
}

###############################################################################
# Method:    dump_topology
# Objective: Dump the fabric topology
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_topology {
    my ($self, %args) = @_;

    $self->_clear_err();

    # No-Op for Apcon

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    disable
# Objective: NO-OP for Apcon
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable {
    my ($self) = @_;

    $self->_set_err("Disable of Apcon switch is not supported");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    enable
# Objective: NO-OP for Apcon
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable {
    my ($self) = @_;

    $self->_set_err("Enable of Apcon switch is not supported");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    shutdown
# Objective: NO-OP for Apcon
# Details:   See POD documentation at the end of this file.
###############################################################################
sub shutdown {
    my ($self) = @_;

    $self->_set_err("Shutdown of Apcon switch is NOT supported. " .
                    "This is not a recommended action because the apcon is " .
                    "an infrastructure device shared by multiple groups.");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    supportshow
# Objective: Run supportshow on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub supportshow {
    my ($self) = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    reboot
# Objective: NO-OP for Apcon
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;

    $self->_set_err("Reset of Apcon switch is NOT supported. " .
                    "This is not a recommended action because the apcon is " .
                    "an infrastructure device shared by multiple groups.");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    info
# Objective: Get switch information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my (@output);

    push @output, "Switch name:           " . $self->hostname(), 
                  "Switch type:           " . $self->hosttype(),
                  "Switch FW version:     " . $self->fw_version(), 
                  "Switch Model:          " . $self->model();
    push @output, "\n";

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    ntp_sync
# Objective: NO-OP for Apcon
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    # This function is a NO-OP for Apcon because it is a disruptive action.
    # When the NTP server is set all active telnet sessions are reset. This
    # could lead to automation failures.

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    install_firmware
# Objective: NO-OP for Apcon
# Details:   See POD documentation at the end of this file.
###############################################################################
sub install_firmware {
    my ($self, %args) = @_;

    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    get_log
# Objective: NO-OP for Apcon
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_log {
    my ($self) = @_;

    $self->_clear_err();

    # This feature is not implemented because Apcon needs a secondary
    # server to write its logs to via a log service. There is no way to
    # directly access the switch logs.

    return (wantarray()) ? () : [];
}

###############################################################################
# Method:    get_config
# Objective: Retrieve the configuration from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_config {
    my ($self) = @_;

    $self->_clear_err();
    my (@output);

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_switch})) {
        @output = $self->execute("show switch");
        $self->{cache}{show_switch} = [ @output ];
    } else {
        @output = @{$self->{cache}{show_switch}};
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_state
# Objective: NO-OP for Apcon
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state {
    my ($self) = @_;
    my $state = "unknown";

    $self->_clear_err();

    return $state;
}

###############################################################################
# Method:    get_current_vf_context_id
# Objective: get the current vf context id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_vf_context_id {
    my ($self) = @_;

    # NOOP for Apcons

    return "";
}

###############################################################################
# Method:    module_reboot
# Objective: Reboot the module of a director switch.
# Details:   Unimplemented.
###############################################################################
sub module_reboot {
    my $self = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type.");
    return $SwitchUtil::FALSE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new Apcon switch object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } 
    }

    if (!defined($conn) || ref($conn) ne "Connectrec") {
        $self->_set_err("Invalid value for argument '-conn'");
        return $self;
    }

    $self->{conn} = $conn;
    return $self;
}

###############################################################################
# Method:    _get_domain_id
# Objective: Get the domain id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_domain_id {
    my ($self) = @_;

    # NO-OP

    return "";
}


1;

=pod

=head1 Package Name

ApconFCSwitch.pm

=head1 Author

Michael M. Latif

=head1 Description

For description of methods see B<San::SanSwitch::Switch.pod>.

=cut

