# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a Brocade FC switch
## @author thomasm@netapp.com
## @status private
## @pod ../Switch.pod

use strict;

# Package Name
##############
package San::SanSwitch::FC::Switch::BrocadeFCSwitch;

# Package Requirements
######################
use San::SanSwitch;
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use San::SanUtil;
use Error;
use NATE::Log qw(log_global);
use vars qw(@EXPORT @ISA);
use Data::Dumper;
@ISA = qw(Error SanLog SanSwitch);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

###############################################################################
# Method:    execute_brocade
# Objective: Execute a command to the switch, if virtual fabric support is
#            enabled, this will execute to the defined (store in object)
#            virtual fabric ID.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub execute_brocade {
    my ($self, $cmd, $timeout, $cmd_prefix, %args) = @_;

    my (@output, $switch_vf_context_id, $quiet, $retry_if_fabric_busy);

    $self->_clear_err();

    # Default $quiet to 0
    $quiet = 0;
    $retry_if_fabric_busy = 0;
    # Process the args
    foreach (keys %args) {
        if (/^-?quiet$/) {
            if (not length($quiet = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return ();
            }

            # Make sure we received an acceptable value
            unless ($quiet == 0 or $quiet == 1) {
                $self->_set_err("Invalid value '$quiet' for argument '$_'");
                return ();
            }
            next;
        }
        if (/^-?retry_if_fabric_busy$/) {
            if (not length($retry_if_fabric_busy = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return ();
            }

            # Make sure we received an acceptable value
            unless ($retry_if_fabric_busy == 0 or $retry_if_fabric_busy == 1) {
                $self->_set_err("Invalid value '$retry_if_fabric_busy' for argument '$_'");
                return ();
            }
            next;
        }
    }

    # If we are in a virtual fabric context, use fosexec to execute
    # the command to the appropriate virtual fabric (FID).  If we are not
    # in a virtual fabric context but instead have an admin domain defined,
    # we just run the command because we should have selected the ad during
    # the connection phase
    if ($self->{vf_status} == 1) {
        if (!defined($self->{vf_context_id})) {
            $self->_set_err("Error: BrocadeFCSwitch::execute_brocade() " .
                            "Could not determine virtual fabric context to execute command");
            return ();
        }

        # Get the current FID on the switch
        ($switch_vf_context_id, undef, undef) = $self->_get_vf_attributes();

        # Check to see if the current vf_context_id matches the
        # FID context that is set on the switch.  If they differ
        # we need to switch to the vf_context_id to run the commands.
        if ($self->{vf_context_id} != $switch_vf_context_id) {
            # If we have timeout defined, pass it to _change_switch_vf_context
            # so that it can be passed to the execute routine.
            my $status;
            if (defined($timeout) and $timeout ne "") {
                $status = $self->_change_switch_vf_context(-vf => $self->{vf_context_id}, -timeout => $timeout);
            } else {
                $status = $self->_change_switch_vf_context(-vf => $self->{vf_context_id});
            }

            if ($status eq $SwitchUtil::FALSE) {
                if ($self->error) {
                    return ();
                } else {
                    $self->_set_err("Unable to change the switch virtual fabric context to $self->{vf_context_id}");
                    return ();
                }
            }
        }
        # Add an arbitrary 1 second sleep so we don't overload the switch
        # with rapid commands and confuse the output
        &SanUtil::nate_sleep(1);
    }

    # If we have a command prefix defined ($cmd_prefix), put it in front of
    # $cmd.  A good example of how this might be used is prepending "echo y|" to
    # the front of a command so that the echo y is piped into an ad --exec.

    if (defined($cmd_prefix) and $cmd_prefix ne "") {
        $cmd = $cmd_prefix . " " . $cmd;
    }

    # Execute the command and return the output.  Pass the timeout flag
    # if $timeout is defined., try up to 3 times if fabric is busy

    if($retry_if_fabric_busy) {
        my $retry_count = 0;
        while ((grep(/try again later/i, @output) and $retry_count < 3) or $retry_count == 0) {
            if (defined($timeout) and $timeout ne "") {
                @output = $self->execute($cmd, $timeout, -quiet => $quiet);
            } else {
                @output = $self->execute($cmd, "", -quiet => $quiet);
            }
            if ($retry_count > 0 and $retry_count < 3) {
                &SanUtil::nate_sleep(30);
            }
            $retry_count++;
        }
    } else {
        if (defined($timeout) and $timeout ne "") {
            @output = $self->execute($cmd, $timeout, -quiet => $quiet);
        } else {
            @output = $self->execute($cmd, "", -quiet => $quiet);
        }
    }

    # Set the virtual fabric context back to the original value if it changed.
    # Check to see if the current vf_context_id matches the
    # FID context that is set on the switch.  If they differ
    # we need to switch to the origina FID value
    #
    # Any errors that my be raised by _set_err from this point should not stop
    # us from returning @output since we should have valid output from the
    # original requested command.  We will set any necessary errors and then
    # return @output instead of ().
    if ($self->{vf_status} == 1) {
        # Add an arbitrary 1 second sleep so we don't overload the switch
        # with rapid commands and confuse the output
        &SanUtil::nate_sleep(1);
        if (not defined($self->{vf_context_id})) {
            $self->_set_err("vf_context_id is not defined on the switch object");
            return(@output);
        }
        if (not defined($switch_vf_context_id)) {
            $self->_set_err("Unable to determine the original virtual fabric id of the switch");
            return(@output);
        }

        if ($self->{vf_context_id} != $switch_vf_context_id) {
            # If we have timeout defined, pass it to _change_switch_vf_context
            # so that it can be passed to the execute routine.
            my $status;
            if (defined($timeout) and $timeout ne "") {
                    $status = $self->_change_switch_vf_context(-vf => $switch_vf_context_id, -timeout => $timeout);
            } else {
                $status = $self->_change_switch_vf_context(-vf => $switch_vf_context_id);
            }

            if ($status eq $SwitchUtil::FALSE) {
                if ($self->error) {
                    return(@output);
                } else {
                    $self->_set_err("Unable to change the switch virtual fabric context to $self->{vf_context_id}");
                    return(@output);
                }
            }
        }
    }

    return(@output);
}

###############################################################################
# Method:    get_firmware_version
# Objective: Get the firmware version of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_firmware_version {
    my ($self) = @_;
    my (@output, $output);
    my $version = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{version})) {
        @output = $self->execute("version");
        $self->{cache}{version} = [ @output ];
    } else {
        @output = @{$self->{cache}{version}};
    }

    if (!$self->error) {
        $version = (grep(/Fabric OS\:/i, @output))[0];
        $version =~ s/Fabric OS\:\s+//i;
    }

    return $version;
}

###############################################################################
# Method:    supportshow
# Objective: Run supportshow on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub supportshow {
    my ($self, %args) = @_;
    my ($cmd, @output);

    # Clear an pre-existing errors
    $self->_clear_err();

    $cmd = "supportshow";

    # Use the say method in order to override the prompt
    $self->sanlog(-msg => "Sending \"supportshow\"");
    @output = &Tharn::say($self->{conn}, $cmd, -timeout, 60, '-blind');
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return undef;
    }
    return 0;
}

###############################################################################
# Method:    get_ad255_show
# Objective: Get the ad --show output from a switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_ad255_show {
    my ($self) = @_;

    $self->_clear_err();

    # Get the switch name
    my $switch_name = $self->hostname();
    unless (defined($switch_name) and $switch_name ne "") {
        $self->_set_err("Unable to retrieve the switch name");
        return (wantarray()) ? () : [];
    }

    # Check if we have a cached output already
    my (@output, $status);
    if (defined($self->{cache}{ad_show})) {
        @output = @{$self->{cache}{ad_show}};
    } else {
        # Make sure Virtual Fabrics are NOT enabled on the switch
        if ($self->is_vf_enabled()) {
            # Looks like VF is enabled so it can't support admin domains.  Set an
            # error and return false;
            $self->_set_err("Switch '$switch_name' has virtual fabrics enabled so admin " .
                            "domains are not supported");
            return (wantarray()) ? () : [];
        }

        # Now that we know we have a switch that "should" be able to support admin domains,
        # go ahead and retrieve the current admin domain context of the current connection
        # so we can switch back to it when we are done.  We will need to switch to context
        # 255 to dump the database.
        my $orig_ad = $self->get_current_admin_domain();
        if ($self->error) {
            $self->_set_err("Unable to retrieve the current admin domain for switch '$switch_name':\n",
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }

        unless($orig_ad == 255) {
            # Switch to AD255 since we weren't already there
            $self->switch_to_admin_domain("255");
            if ($self->error) {
                # Check to see if we had a permission problem. If so, we want
                # to clear the error and return empty.  If it wasn't a permission
                # error, then set error and bail.
                my @errmsg = $self->errmsg();
                if (grep(/insufficient permissions/i, @errmsg) > 0) {
                    # Looks like it was a permissions error so clear
                    # the error and print out a message
                    $self->sanlog(-msg => "\n" .
                                          "  Unable to switch to admin domain 255 on switch '$switch_name'.\n" .
                                          "  Either it doesn't exist or the user has insufficient permissions.\n" .
                                          "  If you are using FabricMgr in C-Mode or 7-Mode or any switch events\n" .
                                          "  in C-Mode, you may experience issues since the libraries will not be\n" .
                                          "  able to determine what admin domain each port belongs to.");
                    $self->_clear_err();
                } else {
                    # This was some other error so don't clear the error
                    $self->_set_err("Unable to switch to AD255 on switch '$switch_name':\n", $self->errmsg());
                }

                # Return an empty array regardless of the circumstances
                return (wantarray()) ? () : [];
            }
        }

        # Do an ad --show to dump the AD config
        my $cmd = "ad --show";
        $self->sanlog(-msg => "Executing command ($cmd)");

        my @output;
        my $log = log_global();
        if ($log->may_debug()) {
            @output = $self->execute_brocade($cmd, "", "", -quiet => 0, -retry_if_fabric_busy => 1);
        } else {
            @output = $self->execute_brocade($cmd, "", "", -quiet => 1, -retry_if_fabric_busy => 1);
        }

        if ($self->error()) {
            $self->_set_err("Unable to run command ($cmd) on switch '$switch_name':\n",
                            $self->errmsg());

            # Try to switch back to the original AD before bailing
            $self->switch_to_admin_domain($orig_ad);
            if ($self->error) {
                $self->_set_err("Unable to switch to AD${orig_ad} on switch " .
                                "'$switch_name':\n", $self->errmsg());
            }
            return (wantarray()) ? () : [];
        }

        # Cache the output for later
        $self->{cache}{ad_show} = [ @output ];

        # Switch back to the original AD
        $self->switch_to_admin_domain($orig_ad);
        if ($self->error) {
            $self->_set_err("Unable to switch to AD${orig_ad} on switch " .
                            "'$switch_name':\n", $self->errmsg());
            return (wantarray()) ? () : [];
        }
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_model
# Objective: Get the model of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_model {
    my ($self) = @_;
    my (@output, $output);
    my $model = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{switch_show})) {
        @output = $self->execute("switchShow");
        $self->{cache}{switch_show} = [ @output ];
    } else {
        @output = @{$self->{cache}{switch_show}};
    }

    if (!$self->error) {
       $model = (grep(/SwitchType\:/i, @output))[0];
       $model =~ s/SwitchType\:\s+//i;
    }

    # Determine English name of the switch model - embedded switches do not
    # show up in chassisname so we need to use the SwitchType from SwitchShow
    if ($model =~ /45\.0/) {
       $model = "4016";
    } elsif ($model =~ /61\.0/) {
       $model = "4424";
    } elsif ($model =~ /37\.1/) {
       $model = "4020";
    } elsif ($model =~ /29\.3/) {
       $model = "4012";
    } elsif ($model =~ /43\.2/) {
       $model = "4024";
    } elsif ($model =~ /9\.2/) {
       $model = "3800";
    } elsif ($model =~ /16\.2/) {
       $model = "3200";
    } elsif ($model =~ /10\./) {
       $model = "12000";
    } elsif ($model =~ /12\./) {
       $model = "3900";
    } elsif ($model =~ /17\./) {
       $model = "3800";
    } elsif ($model =~ /18\./) {
       $model = "3000";
    } elsif ($model =~ /^21\./) {
       $model = "24000";
    } elsif ($model =~ /22\./) {
       $model = "3016";
    } elsif ($model =~ /26\./) {
       $model = "3850";
    } elsif ($model =~ /27\./) {
       $model = "3250";
    } elsif ($model =~ /32\./) {
       $model = "4100";
    } elsif ($model =~ /32\./) {
       $model = "4100";
    } elsif ($model =~ /^33\./) {
       $model = "3014";
    } elsif ($model =~ /34\./) {
       $model = "200e";
    } elsif ($model =~ /38\./) {
       $model = "7420";
    } elsif ($model =~ /42\./) {
       $model = "48000";
    } elsif ($model =~ /44\./) {
       $model = "4900";
    } elsif ($model =~ /46\./) {
       $model = "7500";
    } elsif ($model =~ /51\./) {
       $model = "4018";
    } elsif ($model =~ /55\.2/) {
       $model = "7600";
    } elsif ($model =~ /58\./) {
       $model = "5000";
    } elsif ($model =~ /73\./) {
       $model = "5470";
    } elsif ($model =~ /62\./) {
       $model = "DCX";
    } elsif ($model =~ /77\./) {
       $model = "DCX_4S";
    } elsif ($model =~ /64\./) {
       $model = "5300";
    } elsif ($model =~ /66\./) {
       $model = "5100";
    } elsif ($model =~ /67\./) {
       $model = "BES";
    } elsif ($model =~ /67\./) {
       $model = "BES";
    } elsif ($model =~ /71\./) {
       $model = "300";
    } elsif ($model =~ /72\./) {
       $model = "5480";
    } elsif ($model =~ /75\./) {
       $model = "M5424";
    } elsif ($model =~ /76\.6|76\.7/) {
       $model = "8000";
    } elsif ($model =~ /83\.3/) {
       $model = "7800";
    } elsif ($model =~ /86\./) {
       $model = "5450";
    } elsif ($model =~ /120\./) {
       $model = "DCX_PLUS";
    } elsif ($model =~ /^121\./) {
       $model = "DCX_4S_PLUS";
    } elsif ($model =~ /109\.1/) {
       $model = "6510";
    } elsif ($model =~ /^133\./) {
       $model = "6520";
    } elsif ($model =~ /92\.1/) {
       $model = "VA40FC";
    } else {
        @output = $self->execute("chassisname");
        $model = $output[0];
    }

    return $model;
}

###############################################################################
# Method:    disable
# Objective: Disable the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable {
    my ($self, %args) = @_;
    my (@output);

    $self->_clear_err();

    my $vf = "";
    my $cmd = "switchdisable";

    foreach (keys %args) {
        if (/-?vf/) {
            if (not length($vf = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vf'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # If vf to disable is given then go ahead and disable
    if ($vf ne "") {
        $self->_execute_vf_command($vf, $cmd);
        if ($self->error) {
            #Just use the error that was set.
            return $SwitchUtil::FALSE;
        } else {
            # Return true since we disable successfully.
            return $SwitchUtil::TRUE;
        }
    } else {
        # Let's just run the disable on the switch
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: BrocadeFCSwitch::disable() " .
                            "Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    enable
# Objective: Enable the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable {
    my ($self, %args) = @_;
    my (@output);

    $self->_clear_err();

    my $vf = "";
    my $cmd = "switchenable";

    foreach (keys %args) {
        if (/-?vf/) {
            if (not length($vf = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vf'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # If vf to enable is given then go ahead and enable that vf
    if ($vf ne "") {
        $self->_execute_vf_command($vf, $cmd);
        if ($self->error) {
            #Just use the error that was set.
            return $SwitchUtil::FALSE;
        } else {
            # Return true since we enable successfully.
            return $SwitchUtil::TRUE;
        }
    } else {
        # Let's just enable the physical switch
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: BrocadeFCSwitch::enable() " .
                            "Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: shutdown()
# Objective:       shutdown Brocade switch
# Details:         See POD documentation at the end of this file
################################################################################
sub shutdown {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "sysshutdown";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # sleep 1 minute to make sure shutdown occurs
    &SanUtil::nate_sleep(60);

    # Invalidate the connection on the object
    &Tharn::disconnect($self->{conn}, '-ok');

    # Ping the host every 10 seconds until it has completely shutdown
    my $alive = 1;

    for (my $cnt = 0; $cnt < 36; $cnt++) {
        my $status = &SanUtil::ping_host($self->hostname());
        if ($status == $SwitchUtil::FALSE) {
            $self->sanlog(-msg => "Switch OS is shutdown");
            $alive = 0;
            last;
        } else {
            $self->sanlog(-msg => "Ping was successful",
                          -msg_level => 7);
            &SanUtil::nate_sleep(10);
        }
    }

    # Machine is still alive, shutdown did not work
    if ($alive) {
        $self->_set_err("Switch OS never shutdown");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    supportsave
# Objective: Saves support information for RASLOG, TRACE, supportShow, core file,
#            FFDC data, and other support information to a remote FTP location.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub supportsave {
    my ($self, %args) = @_;
    my ($host_username, $host_password, $host_ip, $remote_dir);
    my ($timeout, $protocol);

    # Clear any pre-existing errors
    $self->_clear_err();

    # Check to see if the supplied arguments are valid
    foreach (keys %args) {
    if(/^-?host_username$/){
        if(not length($host_username = $args{$_})) {
        $self->_set_err("Invalid value for argument '-host_username', must not be blank");
                return $SwitchUtil::FALSE;
        }
     next;
    } elsif (/^-?host_password$/) {
            if (not length($host_password = $args{$_})) {
                $self->_set_err("Invalid value for argument '-host_password', must not be blank");
                return $SwitchUtil::FALSE;
            }
            next;
    } elsif (/^-?host_ip$/) {
            if (not length($host_ip = $args{$_})) {
                $self->_set_err("Invalid value for argument '-host_ip', must not be blank");
                return $SwitchUtil::FALSE;
            }
            next;
    } elsif  (/^-?remote_dir$/) {
            if (not length($remote_dir = $args{$_})) {
                $self->_set_err("Invalid value for argument '-remote_dir', must not be blank");
                return $SwitchUtil::FALSE;
            }
            next;
    } elsif  (/^-?timeout$/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-timeout', must not be blank");
                return $SwitchUtil::FALSE;
            }
            next;
    } elsif  (/^-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-protocol', must not be blank");
                return $SwitchUtil::FALSE;
            }
            next;
    }
    }

# Check to see if all mandatory argument were passed
    if (not defined $host_username) {
    $self->_set_err("Mandatory argument '-host_username' was not specified");
    return $SwitchUtil::FALSE;
    }
    if (not defined $host_password) {
    $self->_set_err("Mandatory argument '-host_password' was not specified");
    return $SwitchUtil::FALSE;
    }
    if (not defined $host_ip) {
    $self->_set_err("Mandatory argument '-host_ip' was not specified");
    return $SwitchUtil::FALSE;
    }
    if (not defined $remote_dir) {
    $self->_set_err("Mandatory argument '-remote_dir' was not specified");
    return $SwitchUtil::FALSE;
    }

    #defualt to ftp
    $protocol = (exists $args{protocol} ? $args{protocol} : 'ftp');
    #default to 30 mins
    $timeout = (exists $args{timeout} ? $args{timeout} : 1800);

    my $cmd = "supportsave -u $host_username -p $host_password -h $host_ip -d $remote_dir -l $protocol -n";
    my @output = $self->execute($cmd, $timeout);
    if($self->error) {
    $self->_set_err("Error: BrocadeFCSwitch::supportsave() " .
            "Could not execute command supportsave");
    return $SwitchUtil::FALSE;
    }

    my $check = $output[$#output];
    if($check =~/SupportSave completed/i){
    return $SwitchUtil::TRUE;
    } else {
    my $host_name = $self->hostname();
    $self->_set_err("Could not complete the transfer please check the switch: $host_name for more info..\n");
    return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    reboot
# Objective: Reboot the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    # force reconnect to switch
    $self->reconnect();
    if ($self->error) {
        $self->_set_err("Error: BrocadeFCSwitch::reboot() " .
                        "Could not properly reconnect to the switch before reboot");
        return $SwitchUtil::FALSE;
    }

    &SanUtil::nate_sleep(10);

    $self->execute(
            "reboot",
            '5',
            use_connectrec => 1,
            match_table    => [
                                ('\[y\/n\]')  => "y"
                              ],
            on_timeout     => sub {return},
        );

    $self->sanlog(-msg =>
                  "Waiting 30 seconds to allow reboot process to start");

    # sleep to allow boot process to start
    &SanUtil::nate_sleep(30);

    # Wait until the switch does not respond
    my $count = 0;
    while ($self->is_alive() == $SwitchUtil::FALSE &&
           $count < $self->{reboot_timeout}) {
        &SanUtil::nate_sleep(1);
        $count++;
    }

    if ($count < $self->{reboot_timeout}) {
        # wait some additional time to allow telnet service to be
        # available and responding. Also wait some time to allow all
        # ports to log back in (or nameserver db will be empty).
        &SanUtil::nate_sleep(120);
        $self->reconnect();
        if ($self->error) {
            $self->_set_err("Error: BrocadeFCSwitch::reboot() " .
                            "Could not properly reconnect to the switch after reboot");
            return $SwitchUtil::FALSE;
        }

        # invalidate the switchShow cache, since we want to do
        # an lookup after reboot with fresh information.
        $self->{cache}{switch_show} = undef;

        return $SwitchUtil::TRUE;
    } else {
        $self->_set_err("Error: BrocadeFCSwitch::reboot() " .
                        "Could not properly reboot the switch (exceeded allowable reboot/ping time)");
        return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    info
# Objective: Get switch information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my (@output);
    my $Config = $self->config();
    my $Debug = join("\n",@$Config);

    push @output, "Switch name:           " . $self->hostname(),
                  "Switch type:           " . $self->hosttype(),
                  "Switch FW version:     " . $self->fw_version(),
                  "Switch Model:          " . $self->model(),
                  "Switch Domain ID:      " . $self->domain_id(),
                  "Switch Config:         " . $Debug;
    push @output, "\n";

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    ntp_sync
# Objective: Synchronize the clock on the switch with an NTP server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($ntp_server, $ntp_time_zone);
    my ($gmt_offset_hour, $gmt_offset_min, $original_persistent);
    my (@output, $ntp_server_ip);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: BrocadeFCSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_server'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: BrocadeFCSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_time_zone'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # Set the offset from GMT for the timezone
    if ($ntp_time_zone =~ /EST/i) {
        $gmt_offset_hour = -5;
        $gmt_offset_min  = 0;
    } elsif ($ntp_time_zone =~ /CST/i) {
        $gmt_offset_hour = -6;
        $gmt_offset_min  = 0;
    } elsif ($ntp_time_zone =~ /PST/i) {
        $gmt_offset_hour = -8;
        $gmt_offset_min  = 0;
    } elsif ($ntp_time_zone =~ /IST/i) {
        $gmt_offset_hour = 5;
        $gmt_offset_min  = 30;
    } elsif ($ntp_time_zone =~ /MST/i) {
        $gmt_offset_hour = -7;
        $gmt_offset_min  = 0;
    } else {
        $self->_set_err("Error: BrocadeFCSwitch::ntp_sync() " .
                        "Time zone $ntp_time_zone not currently supported");
        return $SwitchUtil::FALSE;
    }

    my $gmt_string = sprintf("%02d:%02d", $gmt_offset_hour, $gmt_offset_min);
    $self->sanlog(-msg => "Setting the Time Zone (TZ) to $ntp_time_zone " .
                          "(GMT $gmt_string)");

    # setup the proper timezone
    my $cmd = "tsTimeZone $gmt_offset_hour, $gmt_offset_min";
    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Error: BrocadeFCSwitch::ntp_sync() " .
                        "Could not execute command $cmd");
        return $SwitchUtil::FALSE;
    }

    # get IP address since switch may not have DNS running
    my (undef, undef, undef, undef, @addrs) = gethostbyname($ntp_server);
    foreach my $ip (@addrs) {
        my ($a, $b, $c, $d) = unpack('C4', $ip);
        $ntp_server_ip = "$a\.$b\.$c\.$d";
    }

    # setup relationship with NTP time server
    $cmd = "tsclockserver \"$ntp_server_ip\"";
    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Error: BrocadeFCSwitch::ntp_sync() " .
                        "Could not execute command $cmd");
        return $SwitchUtil::FALSE;
    }

    # get date to have in log
    $cmd = "date ; tsclockserver";
    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Error: BrocadeFCSwitch::ntp_sync() " .
                        "Could not execute command $cmd");
        return $SwitchUtil::FALSE;
    }

    $self->sanlog(-msg =>
                  "Local time is: $output[0]\n" .
                  "Note: switch might have to be rebooted to have TZ change take place");

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    dump_topology
# Objective: Dump the fabric topology
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_topology {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $switch_name = $self->hostname();
    if ($self->error()) {
        return $SwitchUtil::FALSE;
    }

    # Make sure this isn't an NPV/AG switch
    if ($self->is_npv_enabled()) {
        $self->sanlog(-msg => "Skipping switch '$switch_name' since it is in NPV or AG mode.");
        return $SwitchUtil::TRUE;
    }

    my ($cmd, @output, $received_prompt, $expected_prompt);

    # Run a "topologyshow".  This command will go interactive due to paging so we need to use say
    $cmd = "topologyshow";
    # the expected paging prompt for both commands
    $expected_prompt = "Type \<CR\> to continue, Q\<CR\> to stop:";
    $self->sanlog(-msg => "Executing command ($cmd)");
    (undef, undef, undef, $received_prompt) = &Tharn::say($self->{conn}, $cmd,
                                                          '-prompt', $expected_prompt,
                                                          '-timeout', "30",
                                                          '-comment', "Attempted to send $cmd to the switch");

    # Now we need to loop until we stop getting prompted
    # to continue
    while ($received_prompt =~ /$expected_prompt/) {
        (undef, undef, undef, $received_prompt) = &Tharn::say($self->{conn}, ' ',
                                                              '-prompt', $expected_prompt,
                                                              '-timeout', "30",
                                                              '-comment', "Attempted to send <CR> to the switch");
    }

    # Run a "fabricshow" This command can go interactive due to paging so we need to use say
    $cmd = "fabricshow";
    $self->sanlog(-msg => "Executing command ($cmd)");
    (undef, undef, undef, $received_prompt) = &Tharn::say($self->{conn}, $cmd,
                                                          '-prompt', $expected_prompt,
                                                          '-timeout', "30",
                                                          '-comment', "Attempted to send $cmd to the switch");

    # Now we need to loop until we stop getting prompted
    # to continue
    while ($received_prompt =~ /$expected_prompt/) {
        (undef, undef, undef, $received_prompt) = &Tharn::say($self->{conn}, ' ',
                                                              '-prompt', $expected_prompt,
                                                              '-timeout', "30",
                                                              '-comment', "Attempted to send <CR> to the switch");
    }

    if ($self->error()) {
        $self->_set_err("Unable to run command ($cmd) on switch '$switch_name':\n", $self->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    install_firmware
# Objective: Install firmware on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub install_firmware {
    my ($self, %args) = @_;
    my ($cmd, $count, $still_flashing, $counter, $num_tries, $debug_output);
    my ($host_ip, $host_username, $host_password, $firmware_path);
    my $firmware_install_timeout = 600;
    my (@output, @verify_output);

    # Clear any pre-existing errors
    $self->_clear_err();

    # Check to see if the supplied arguments are valid
    foreach (keys %args) {
        if (/^-?host_ip$/) {
            if (not length($host_ip = $args{$_})) {
                $self->_set_err("Invalid value for argument '-host_ip', must not be blank");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/^-?host_username$/) {
            if (not length($host_username = $args{$_})) {
                $self->_set_err("Invalid value for argument '-host_username', must not be blank");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/^-?host_password$/) {
            # Set the value blindly incase the password is intentionally blank
            $host_password = $args{$_};
            next;
        } elsif (/^-?firmware_path$/) {
            if (not length($firmware_path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-firmware_path', must not be blank");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/^-?firmware_install_timeout$/) {
            if ($args{$_} =~ /\d+/) {
                $firmware_install_timeout = $args{$_};
            } else {
                $self->_set_err("Invalid value for argument '-firmware_install_timeout': "
                            .   "\"$args{$_}\", numeric value expected");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # Check to see if all mandatory argument were passed
    if (not defined $host_ip) {
        $self->_set_err("Mandatory argument '-host_ip' was not specified");
        return $SwitchUtil::FALSE;
    }
    if (not defined $host_username) {
        $self->_set_err("Mandatory argument '-host_username' was not specified");
        return $SwitchUtil::FALSE;
    }
    if (not defined $host_password) {
        $self->_set_err("Mandatory argument '-host_password' was not specified");
        return $SwitchUtil::FALSE;
    }
    if (not defined $firmware_path) {
        $self->_set_err("Mandatory argument '-firmware_path' was not specified");
        return $SwitchUtil::FALSE;
    }

    # Print the passed arguments
    my $message = sprintf("\n%-25s %s\n%-25s %s\n%-25s %s\n%-25s %s\n%-25s %s\n",
                      "host_ip:", $host_ip,
                      "host_username:", $host_username,
                      "host_password:", $host_password,
                      "firmware_path:", $firmware_path,
                      "firmware_install_path:", $firmware_install_timeout);

    $self->sanlog(-msg => $message);

    # Assemble the command with all of the passed params
    $cmd = "firmwaredownload -p scp $host_ip,$host_username,$firmware_path,$host_password";

    # Use the say method in order to override the prompt
    $self->sanlog(-msg => "Sending \"firmwaredownload\" command:\t$cmd");
    @output = &Tharn::say($self->{conn}, $cmd, '-prompt', 'Do you want to continue');
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return undef;
    }

    $self->sanlog(-msg => "Answering \"Y\" to confirm... ");
    $cmd = "Y";

    # Use the say method to remove the prompt and manually set other values
    # Installation process may take up to 10 minutes so the default value of
    # $firmware_install_timeout is 10 minutes
    @output = &Tharn::say($self->{conn}, $cmd, "-wait", '.*HA Rebooting.*',
                          -noprompt, -timeout, $firmware_install_timeout, -maxidle, 180);
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return undef;
    }

    $self->sanlog(-msg => "Firmwaredownload command completed... Waiting 30 "
                        . "seconds to allow reboot process to start...");

    # sleep to allow boot process to start
    &SanUtil::nate_sleep(30);

    # Try to ping once every second until the switch responds
    $self->sanlog(-msg => "Pinging the switch until it is found to be alive "
                        . "(maximum: $self->{reboot_timeout} seconds)...");
    $count = 0;
    while (($self->is_alive() == $SwitchUtil::FALSE) && ($count < $self->{reboot_timeout})) {
        &SanUtil::nate_sleep(1);
        $count++;
    }

    # If the switch is back up, wait another 120 seconds and then reconnect
    if ($count < $self->{reboot_timeout}) {
        $self->sanlog(-msg => "Switch appears to be back online...");
        $self->sanlog(-msg => "Waiting another 120 seconds to allow telnet, "
                            . "nameserver, etc. to come back up...");
        # wait some additional time to allow telnet service to be
        # available and responding. Also wait some time to allow all
        # ports to log back in (or nameserver db will be empty).
        &SanUtil::nate_sleep(120);
        $self->reconnect();
        if ($self->error) {
            $self->_set_err("Could not properly reconnect to the switch after installing firmware...");
            return $SwitchUtil::FALSE;
        }
    # If the switch is not awake after the configured timeout, error out
    } else {
        $self->_set_err("Switch did not properly reboot (exceeded allowable reboot/ping time)...");
        return $SwitchUtil::FALSE;
    }

    # After successful firmwaredownload, the switch will commit, etc.
    # The firmwaredownload is fully complete when the firmwaredownloadstatus
    # command returns "Firmwaredownload command has completed successfully."
    # Check the output for the above every 30 seconds for 10 minutes before failing
    $self->sanlog(-msg => "Preparing to run \"firmwaredownloadstatus\" every 30"
                        . " seconds for 10 minutes or until the output contains: \"completed successfully\"...");
    $still_flashing = $SwitchUtil::TRUE;
    $counter        = 0;
    $num_tries      = 20;
    $cmd            = "firmwaredownloadstatus";
    while ($still_flashing && $counter < $num_tries) {
        $self->sanlog(-msg => "ATTEMPT " . ($counter + 1) . " of $num_tries:\t"
                            . "Waiting 30 seconds before running \"firmwaredownloadstatus\"...");
        &SanUtil::nate_sleep(30);
        @verify_output = $self->execute($cmd, $self->{log_timeout});
        if ($self->error) {
            $self->_set_err("Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
        if (scalar(@verify_output)) {
            foreach (@verify_output){
                if($_ =~ m/Firmwaredownload command has completed successfully/i) {
                    $still_flashing = $SwitchUtil::FALSE;
                    $self->sanlog(-msg => "The output of \"firmwaredownloadstatus\""
                                        . " indicates the download completed successfully...");
                }
            }
        }
        if ($still_flashing) {
            $self->sanlog(-msg => "Could not find \"completed successfully\" in the output...");
        }
        $counter++;
    }

    # If the success string was not found after the allotted time, fail
    # else (it was successful) print the output of the "firmwareshow" and
    # the "version" commands
    if ($still_flashing) {
        $self->_set_err("Could not verify that firmwaredownload completed successfully...");
        return $SwitchUtil::FALSE;
    } else {
        $cmd = "firmwareshow";
        $self->sanlog(-msg => "Executing \"$cmd\"...");
        @output = $self->execute($cmd, $self->{log_timeout});
        if ($self->error) {
            $self->_set_err("Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }

        $cmd = "version";
        $self->sanlog(-msg => "Executing \"$cmd\"...");
        @output = $self->execute($cmd, $self->{log_timeout});
        if ($self->error) {
            $self->_set_err("Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
        return $SwitchUtil::TRUE;
    }
    $self->_set_err("Could not complete firmwaredownload");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    get_log
# Objective: Retrieve the log file from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_log {
    my ($self) = @_;

    $self->_clear_err();

    my $cmd = "supportShow";
    my @output = $self->execute($cmd, $self->{log_timeout});
    if ($self->error) {
        $self->_set_err("Error: BrocadeFCSwitch::get_log() " .
                        "Could not execute command $cmd");
        return undef;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_config
# Objective: Retrieve the configuration from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_config {
    my ($self) = @_;

    my ($cmd, $counter, @output);

    $self->_clear_err();

    my $expected_prompt = qr/Type <CR> or <SPACE BAR> to continue, <q> to stop/;

    # Use SanSwitch::execute to invoke SwithUtil::cr_execute and match_table as
    # an argument to react to prompt with a space character
    $cmd = "configshow";
    @output = $self->execute($cmd,
                             60, #timeout
                             use_connectrec => 1,
                             match_table  => [
                                $expected_prompt => sub{
                                                        $self->{conn}->send(" ");
                                                        undef;
                                                       }
                                             ],
                             check_status => 1);

    # If the switch supports FCOE get into cmsh and get the output of  "show running-config"
    $self->sanlog(-msg => "Checking if the switch supports FCOE");
    if (($self->{fcoe})) {
        $self->execute("cmsh", $self->{log_timeout});
        if ($self->error) {
            $self->_set_err("Could not execute command $cmd");
            return undef;
        }

        $cmd = "show running-config";
        @output = $self->execute($cmd, $self->{log_timeout});
        if ($self->error) {
            $self->_set_err("Could not execute command $cmd");
            return undef;
        }

        $cmd = "exit";
        @output = $self->execute($cmd, $self->{log_timeout});
        if ($self->error) {
            $self->_set_err("Could not execute command $cmd");
            return undef;
        }
    } else {
        $self->sanlog(-msg => "FCOE is not supported on this switch");
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_state
# Objective: Retrieve the state of the switch (enabled/disabled)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state {
    my ($self, %args) = @_;
    my $state = "unknown";

    $self->_clear_err();

    my $vf = "";

    foreach (keys %args) {
        if (/-?vf/) {
            if (not length($vf = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vf'");
                return $state;
            }
            next;
        }
    }

    my $cmd = "switchShow";
    my @output = ();

    if ($vf ne "") {
        @output = $self->_execute_vf_command($vf, $cmd);
    } else {
        @output = $self->execute($cmd);
    }

    if ($self->error) {
        $self->_set_err("Error: BrocadeFCSwitch::get_state() " .
                        "Could not execute command $cmd");
        return $state;
    }

    $state = (grep(/switchState\:/i, @output))[0];
    $state =~ s/switchState:\s+//i;

    return &SwitchUtil::normalize_state($state);
}

###############################################################################
# Method:    get_vf_ids
# Objective: Retrieve the virtual farbic IDs on this switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vf_ids {
    my ($self) = @_;

    my @fids = ();

    $self->_clear_err();

    # If we already have stored, return value.
    if (defined($self->{vf_ids})) {
        return(@{$self->{vf_ids}});
    }

    # output from lscfg --show in FOS <7.2:
    # Created switches:  128(ds)  5(bs)  6
    #
    # Port      0     1     2     3     4     5     6     7     8     9
    # -------------------------------------------------------------------
    # FID     128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 |

    # output from lscfg --show in FOS >=7.2:
    # Created switches FIDs(Domain IDs):  128(ds)  5(bs)  6
    #
    # Port      0     1     2     3     4     5     6     7     8     9
    # -------------------------------------------------------------------
    # FID     128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 |
    my $cmd = "lscfg --show";
    my @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Error: BrocadeFCSwitch::get_vf_ids() " .
                        "Could not execute command $cmd");
        return undef;
    }

    # extract FIDs (fabric IDs)
    my $fids_line = (grep(/Created switches.*:/i, @output))[0];
    $fids_line =~ s/\s*Created switches.*:\s+//;
    $fids_line =~ s/\(\S+\)//g;
    $fids_line =~ s/\s+$//g;
    @fids = split(/\s+/, $fids_line);

    return(@fids);

}

###############################################################################
# Method:    set_vf_context
# Objective: set the context of the switch you want commands to be execute
#            to in a virtual fabric context
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_vf_context {
    my ($self, $fid) = @_;

    my (@fids);

    @fids = $self->get_vf_ids();
    if (grep(/$fid/, @fids) == 0) {
        $self->_set_err("Error: BrocadeFCSwitch::set_vf_context() " .
                        "Virtual fabric ID (FID) passed ($fid), is not a valid FID\n" .
                        "on this switch (FID list = @fids)");
        return $SwitchUtil::FALSE;
    } else {
        $self->{vf_context_id} = $fid;
    }

    return $SwitchUtil::TRUE;

}

###############################################################################
# Method:    get_current_vf_context_id
# Objective: get the current vf context id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_vf_context_id {
    my ($self) = @_;

    return $self->{vf_context_id};
}

###############################################################################
# Method:    module_reboot
# Objective: Reboot the module of a director switch.
# Details:   Unimplemented.
###############################################################################
sub module_reboot {
    my $self = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type.");
    return $SwitchUtil::FALSE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new BrocadeFC switch object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: BrocadeFCSwitch::_new() " .
                                "Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        }
    }

    if (!defined($conn) || ref($conn) ne "Connectrec") {
        $self->_set_err("Error: BrocadeFCSwitch::_new() " .
                        "Invalid value for argument '-conn'");
        return $self;
    }

    # store connection
    $self->{conn} = $conn;

    return $self;
}

###############################################################################
# Method:    _get_ad_status
# Objective: Check if Admin Domains are supported
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_ad_status {
    my ($self) = @_;

    my $status;

    $self->_clear_err();
    my $cmd = "ad --show";

    my @output;
    my $log = log_global();
    if ($log->may_debug()) {
        @output = $self->execute_brocade($cmd, "", "", -quiet => 0, -retry_if_fabric_busy => 1);
    } else {
        @output = $self->execute_brocade($cmd, "", "", -quiet => 1, -retry_if_fabric_busy => 1);
    }

    if ($self->error) {
        # do not set error, rather clear since this can
        # be run to a switch that is not 6.x and does
        # not support fosconfig (and we don't want to die)
        $self->_clear_err();
        return $SwitchUtil::FALSE;
    }

    if (grep(/not supported|command not found/i, @output)) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }
}

###############################################################################
# Method:    _get_vf_status
# Objective: Retrieve the status virtual fabric support on this switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_vf_status {
    my ($self) = @_;

    my $status;

    $self->_clear_err();

    # output from fosconfig --show
    # FC Routing service:     disabled
    # iSCSI service:          Service not supported on this Platform
    # iSNS client service:        Service not supported on this Platform
    # Virtual Fabric:         enabled
    my $cmd = "fosconfig --show";
    my @output = $self->execute($cmd);
    if ($self->error) {
        # do not set error, rather clear since this can
        # be run to a switch that is not 6.x and does
        # not support fosconfig (and we don't want to die)
        $self->_clear_err();
        return undef;
    }

    # If we do not find Virtual Fabric in the output, then this
    # means this switch does not support. Otherwise, we just grab
    # the state enables|disabled
    if (grep(/Virtual Fabric\:\s+enabled/, @output) == 0) {
        $status = "not_supported";
    } else {
        $status = (grep(/Virtual Fabric/, @output))[0];
        $status =~ s/\s*Virtual Fabric\:\s+//;
        $status =~ s/\s+$//;

        # Now, if it is supported, get the current FID and store
        my ($fid, undef, undef) = $self->_get_vf_attributes();
        if ($self->error or !defined($fid) or $fid eq "") {
            $self->_set_err("Error: BrocadeFCSwitch::get_vf_status() " .
                            "Could not determine current Virtual fabric ID of the switch");
            return undef;
        }
        $self->{vf_context_id} = $fid;
    }

    return($status);

}


###############################################################################
# Method:    _get_vf_attributes
# Objective: Retrieve the list of attributes about the current
#            Virtual fabric context. Will return the Fabric ID (FID),
#            if this is the Base Switch, and if this is Default Switch.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_vf_attributes {
    my ($self) = @_;

    $self->_clear_err();

    my $cmd = "switchShow";
    my @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Error: BrocadeFCSwitch::get_vf_current_fid() " .
                        "Could not execute command $cmd");
        return undef;
    }

    # From switchShow:
    # LS Attributes:  [FID: 5, Base Switch: Yes, Default Switch: No]
    my $attr_line = (grep(/LS Attributes\:/i, @output))[0];
    $attr_line =~ m/.*FID\:\s+(\d+)\,\s+Base Switch\:\s+(\S+)\,\s+Default Switch\:\s+(\S+).*/;
    my $fid = $1;
    my $base_switch = $2;
    my $default_switch = $3;

    return($fid, $base_switch, $default_switch);

}

###############################################################################
# Method:    _change_switch_vf_context
# Objective: Change the active virtual fabric context on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _change_switch_vf_context {
    my ($self, %args) = @_;

    my (@output, $cmd, $switch_vf_context_id, $vf, $timeout);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vf/) {
            if (not length($vf = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vf'");
                return $SwitchUtil::FALSE;
            }
        } elsif (/-?timeout/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-timeout'");
                return $SwitchUtil::FALSE;
            }
        }
    }

    # Check for required parameters
    unless (defined($vf) and $vf ne "") {
        $self->_set_err("Invalid value for argument '-vf'");
        return $SwitchUtil::FALSE;
    }

    if ($self->{vf_status} == 1) {
        $cmd = "setcontext $vf";

        # Execute the command and return the output.  Pass the timeout flag
        # if $timeout is defined.
        if (defined($timeout) and $timeout ne "") {
            @output = $self->execute($cmd, $timeout);
        } else {
            @output = $self->execute($cmd);
        }
    } else {
        $self->_set_err("Switch object is not configured for Virtual fabrics");
        return $SwitchUtil::FALSE;
    }
    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    _execute_vf_command
# Objective: Execute the given command on a specified vf.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _execute_vf_command {
    my ($self, $vf, $cmd) = @_;
    my ($old_vf_status, $old_vf_context, $reset_status, $reset_context,
        @output);

    $reset_status = $reset_context = $SwitchUtil::FALSE;

    $self->_clear_err();

    # If vf context is not set set it
    if ($vf != $self->{vf_context_id}) {
        # Save old state
        $old_vf_context = $self->{vf_context_id};
        $reset_context = $SwitchUtil::TRUE;

        # Set to the context we are going to run with
        $self->set_vf_context($vf);
        if ($self->error) {
            #Just use the error that was set.
            return ();
        }
    }

    if ($self->{vf_status} != 1) {
        $self->{vf_status} = 1;
        $reset_status = $SwitchUtil::TRUE;
    }

    # Run the command
    @output = $self->execute_brocade($cmd);
    # Ignore errors at this point reset old status before checking.

    # Reset to the original state
    $self->{vf_status} = $old_vf_status if ($reset_status);
    $self->{vf_context_id} = $old_vf_context if ($reset_context);

    # Check if the execute_brocade ran successfully.
    if ($self->error) {
        #Just use the error that was set.
        return ();
    } else {
        return @output;
    }
}

###############################################################################
# Method:    _get_admin_domain_database
# Objective: Log into the switch and retrieve all of the admin domain
#            information found.  This will be accomplished through AD255.
#            The data will be returned in a hash that has the data saved from
#            the perspective of the admin domain and from the perspective of the
#            switch.
#
#            $hash = { by_ad     => {AD#}  => {fcid} => {port1} = 1,
#                      by_switch => {fcid} => {AD#}  => {port1} = 1,
#                    }
#
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_admin_domain_database {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_name = $self->hostname();
    unless(defined($switch_name) and $switch_name ne "") {
        $self->_set_err("Unable to retrieve the switch name from the switch object");
        return {};
    }

    # Make sure Access Gateway Mode is NOT enabled on the switch
    if ($self->is_npv_enabled()) {
        # Looks like AG is enabled so it can't support admin domains.  Set an
        # error and return false;
        $self->_set_err("Switch '$switch_name' is in access gateway mode so admin " .
                        "domains are not supported");
        return {};
    }

    # Make sure Virtual Fabrics are NOT enabled on the switch
    if ($self->is_vf_enabled()) {
        # Looks like VF is enabled so it can't support admin domains.  Set an
        # error and return false;
        $self->_set_err("Switch '$switch_name' has virtual fabrics enabled so admin " .
                        "domains are not supported");
        return {};
    }

    # Make sure Admin Domains are enabled on the switch
    unless ($self->is_ad_enabled()) {
        # Looks like AD is not enabled. Set an error and return false;
        $self->_set_err("Switch '$switch_name' does not support admin domains");
        return {};
    }

    my @ad_output = $self->get_ad255_show();
    if ($self->error) {
        $self->_set_err("Unable to retrieve the 'ad --show' output for switch '$switch_name':\n",
                        $self->errmsg()) ;
        return {};
    }

    # Parse through the output
    # The output is going to look like this
    #
    # Current AD Number: 255  AD Name: AD255
    #
    # Transaction buffer configuration:
    # ---------------------------------
    # no configuration
    #
    # Defined configuration:
    # ----------------------
    #
    # AD Number:   0  AD Name: AD0    State: Active
    #
    #         Device WWN members:             50:0a:09:83:87:f9:3e:b2;
    #                                         50:0a:09:82:87:f9:3e:b2;
    #                                         50:0a:09:81:87:f9:3e:b2;
    #         Switch port members:            31,31; 30,30; 21,21; 20,20;
    #                                         13,13; 12,12; 10,10; 11,11;
    #                                         14,14; 15,15; 31,30; 30,31;
    #                                         21,20; 20,21; 13,12; 12,13;
    #                                         10,11; 11,10; 14,15; 15,14;
    #                                         31,29; 30,28; 21,23; 20,22;
    #
    # AD Number:  10  AD Name: AD10   State: Active
    #
    #         Switch port members:            21,11; 20,10; 19,13; 11,21;
    #                                         10,20; 14,16; 15,17; 16,14;
    #                                         17,15; 18,12; 21,10; 20,11;
    #                                         19,12; 11,20; 10,21; 14,17;
    #                                         15,16; 16,15; 17,14; 18,13;
    #         Switch WWN members:             10:00:00:05:1e:03:3b:4d;
    #                                         10:00:00:05:1e:07:89:67;
    #                                         10:00:00:05:1e:02:35:f7;
    #                                         10:00:00:05:1e:04:db:1b;

    my $ad_hash = {};
    my $ad_number = "";
    foreach my $line (@ad_output) {
        if ($line =~ /AD Number:\s+(\d+)\s+/) {
            $ad_number = $1;
        } elsif ($line =~ /(\d+\,\d+.*\;)/) {
            # We want to outright skip AD0 since we don't want the libraries
            # to ever operate on the AD.
            next if ($ad_number == 0);

            # We have a line that has port numbers in it.
            my $members = $1;
            $members =~ s/\s+//g;
            my @pairs = split(/\;/, $members);

            # Loop over each fcid,port pair and split it out
            # and add it to the hash for the AD
            foreach my $pair (@pairs) {
                my ($fcid, $port) = split(/\,/, $pair);

                # Add the info to the AD hash.  If $ad_number is
                # not currently set, then we have a data error.
                unless ($ad_number ne "") {
                    $self->_set_err("Unexpected data returned from (ad --show) for switch '$switch_name'");
                    return {};
                }

                my $ad_name = "AD" . $ad_number;

                $ad_hash->{by_ad}->{$ad_name}->{$fcid}->{$port} = 1;
                $ad_hash->{by_switch}->{$fcid}->{$ad_name}->{$port} = 1;
            }
        }
    } ##END foreach my $line (@ad_output) {

    return $ad_hash;
}

###############################################################################
# Method:    _get_current_admin_domain()
# Objective: get the current admin domain the switch object is sitting in
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_current_admin_domain {
    my ($self) = @_;

    my ($current_ad);

    $self->_clear_err();

    # Get the switch name
    my $switch_name = $self->hostname();
    unless (defined($switch_name) and $switch_name ne "") {
        $self->_set_err("Unable to retrieve the switch name");
        return "";
    }

    # Make sure Access Gateway Mode is NOT enabled on the switch
    if ($self->is_npv_enabled()) {
        # Looks like AG is enabled so it can't support admin domains.  Set an
        # error and return false;
        $self->_set_err("Switch '$switch_name' is in access gateway mode so admin " .
                        "domains are not supported");
        return {};
    }

    # Make sure Virtual Fabrics are NOT enabled on the switch
    if ($self->is_vf_enabled()) {
        # Looks like VF is enabled so it can't support admin domains.  Set an
        # error and return false;
        $self->_set_err("Switch '$switch_name' has virtual fabrics enabled so admin " .
                        "domains are not supported");
        return {};
    }

    # Make sure Admin Domains are enabled on the switch
    unless ($self->is_ad_enabled()) {
        # Looks like AD is not enabled. Set an error and return false;
        $self->_set_err("Switch '$switch_name' does not support admin domains");
        return {};
    }

    # Run the ad --show to make sure we made it in. The output should
    # look like the following:
    #
    # brcd200e-shu01:AD50:admin> ad --show
    # Current AD Number: 50  AD Name: AD50
    #
    # Effective configuration:
    # ------------------------
    #
    # Do an ad --show to dump the AD config
    my $cmd = "ad --show";
    $self->sanlog(-msg => "Executing command ($cmd)");

    my @output;
    my $log = log_global();
    if ($log->may_debug()) {
        @output = $self->execute_brocade($cmd, "", "", -quiet => 0, -retry_if_fabric_busy => 1);
    } else {
        @output = $self->execute_brocade($cmd, "", "", -quiet => 1, -retry_if_fabric_busy => 1);
    }

    if ($self->error()) {
        $self->_set_err("Unable to run command ($cmd) on switch '$switch_name':\n",
                        $self->errmsg());

        return "";
    }

    foreach my $line (@output) {
        if ($line =~ /Current AD Number:\s+(\d+)\s+/i) {
            $current_ad = $1;
            return $current_ad;
        }
    }

    # we didn't find the current ad
    $self->_set_err("Unable to determine the current admin domain");

    return "";
}

###############################################################################
# Method:    _get_npv_mode
# Objective: Get the NPV (AG) mode for the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_npv_mode {
    my ($self) = @_;

    $self->_clear_err();

    # Check if we have a cached output already
    my @output;
    if (!defined($self->{cache}{switch_show})) {
        @output = $self->execute("switchShow");
        if ($self->error) {
            $self->_set_err("Unable to get the 'switchshow' output",
                            $self->errmsg());
            return $SwitchUtil::FALSE;
        }
        $self->{cache}{switch_show} = [ @output ];
    } else {
        @output = @{$self->{cache}{switch_show}};
    }

    # check for AG (Access Gateway Mode)
    if (grep(/Access Gateway Mode/i, @output) > 0) {
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
}


###############################################################################
# Method:    _get_fcoe
# Objective: Check if the switch has FCOE enabled
# Details:   See POD documentation at the end of this file.
# Author: Ameet Gadvi
###############################################################################
sub _get_fcoe {
    my ($self) = @_;

    $self->_clear_err();

    # Check if we have a cached output already
    my @output;
    my $cmd = "fcoe --fipcfgshow";
    @output = $self->execute($cmd);
    if ($self->error) {
        # do not set error, rather clear since this can
        # be run to a switch that does not support FCOE
        # (and we don't want to die)
        $self->_clear_err();
        return $SwitchUtil::FALSE;
    }

    # Check if the switch supports FCOE
    if (grep(/FIP\s+Unsolicited\s+Advertisement\s+Interval\s+=\s+(\d+)/i, @output) > 0) {
        $self->sanlog(-msg => "FCOE is supported on the switch");
    } else {
        $self->sanlog(-msg => "FCOE is not supported on the switch");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    _get_npiv
# Objective: Get the npiv capability of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_npiv {
    my ($self) = @_;

    $self->_clear_err();

    # Check if we have a cached output already
    my @output;
    if (!defined($self->{cache}{portcfg_show})) {
        @output = $self->execute("portcfgshow");
        if ($self->error) {
            $self->_set_err("Unable to get the 'portcfgshow' output",
                            $self->errmsg());
            return $SwitchUtil::FALSE;
        }
        $self->{cache}{portcfg_show} = [ @output ];
    } else {
        @output = @{$self->{cache}{portcfg_show}};
    }

    # Check for NPIV enabled on any port
    #
    # NPIV capability   ON ON ON ON   ON ON ON ON   ON ON ON ON   ON ON ON ON
    #
    if (grep(/^NPIV capability\s+.*ON.*/i, @output)) {
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    _get_domain_id
# Objective: Get the domain id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_domain_id {
    my ($self) = @_;

    $self->_clear_err();

    my $domain_id = "";

    # Check to see if the switch is in Access Gateway Mode.
    if ($self->is_npv_enabled()) {
        # Looks like we are in AG mode so just return ""
        return "";
    } else {
        # Since we aren't in AG mode and didn't receive an error during the
        # check, we can try to get the domain id.
        # Check if we have a cached switchshow output already
        my @output;
        if (!defined($self->{cache}{switch_show})) {
            @output = $self->execute("switchShow");
            if ($self->error) {
                $self->_set_err("Unable to get the 'switchshow' output",
                            $self->errmsg());
                return "";
            }
            $self->{cache}{switch_show} = [ @output ];
        } else {
            @output = @{$self->{cache}{switch_show}};
        }

        $domain_id = (grep(/switchDomain\:/i, @output))[0];
        $domain_id =~ s/switchDomain\:\s+//i;
    }

    return $domain_id;
}

###############################################################################
# Method:    _switch_to_admin_domain()
# Objective: switch to the admin domain specified
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _switch_to_admin_domain {
    my ($self, $admin_domain) = @_;

    my ($current_ad);

    $self->_clear_err();

    # Make sure we were passed in an admin domain
    unless (defined($admin_domain) and $admin_domain >= 0 and $admin_domain <= 255) {
        $self->_set_err("Invalid value for admin domain.  Value must be in the range of 0 to 255");
        return $SwitchUtil::FALSE;
    }

    # Get the current admin domain
    $current_ad = $self->get_current_admin_domain();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # See if we even need to switch
    if ($admin_domain == $current_ad) {
        $self->sanlog(-msg => "Current AD '$current_ad' matches requested AD '$admin_domain'");
        return $SwitchUtil::TRUE;
    } else {
        # We don't match, so determine what to do
        if ($current_ad == 0) {
            # Since we are in AD 0, we can just switch
            my $cmd = "ad --select " . $admin_domain;
            my @output = $self->execute($cmd, "", -quiet => 0);
            if ($self->error()) {
                $self->_set_err("Unable to switch to AD${admin_domain}:\n", $self->errmsg());
                return (wantarray()) ? () : [];
            }

            # Make sure we made it
            $current_ad = $self->get_current_admin_domain();
            if ($self->error) {
                return $SwitchUtil::FALSE;
            }

            unless ($admin_domain == $current_ad) {
                $self->_set_err("Failed to switch to admin domain '$admin_domain'");
                return $SwitchUtil::FALSE;
            }
        } else {
            # We are not in AD 0 so we need to first logout, then switch
            my $cmd = "logout";
            my @output = $self->execute($cmd, "", -quiet => 0);
            if ($self->error()) {
                $self->_set_err("Unable to log out of current AD on switch:\n", $self->errmsg());
                return (wantarray()) ? () : [];
            }

            # If they wanted to go to AD0, then we don't do a select.
            unless ($admin_domain == 0) {
                # Since we are in AD 0, we can just switch
                my $cmd = "ad --select " . $admin_domain;
                my @output = $self->execute($cmd, "", -quiet => 0);
                if ($self->error()) {
                    $self->_set_err("Unable to switch to AD${admin_domain} on switch:\n", $self->errmsg());
                    return (wantarray()) ? () : [];
                }
            }

            # Make sure we made it
            $current_ad = $self->get_current_admin_domain();
            if ($self->error) {
                return $SwitchUtil::FALSE;
            }

            unless ($admin_domain == $current_ad) {
                $self->_set_err("Failed to switch to admin domain '$admin_domain'");
                return $SwitchUtil::FALSE;
            }
        } ## END if ($current_ad == 0) {
    } ## END if ($admin_domain == $current_ad) {

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    _execute_configure_terminal
# Objective: Helper function to handle the execution of the 'configure terminal'
# command
###############################################################################
sub _execute_configure_terminal {
    my ($self) = @_;

    my $cmd = "configure terminal";
    my @output = $self->execute($cmd);
    if ($self->error()) {
        $self->_set_err("Command $cmd failed", $self->errmsg());
        return $SwitchUtil::FALSE;
    }

    my $retry = 1;

    # if the output does not contain "Enter configuration commands" then we may not have properly
    # entering the configure terminal CLI
    while (not grep(/Enter configuration commands/i, @output) and $retry <= 5) {
        $self->sanlog(-msg => "'configure terminal' did not properly execute will retry (count: $retry / 5)");
        &SanUtil::nate_sleep(1);

        @output = $self->execute($cmd);
        if ($self->error()) {
            $self->_set_err("Failed to run 'configure terminal'", $self->errmsg());
            return $SwitchUtil::FALSE;
        }
        $retry++;
    }

    return $SwitchUtil::TRUE;
}

1;

=pod

=head1 Package Name

BrocadeFCSwitch.pm

=head1 Author

Tom Masterson

=head1 Description

For description of methods see B<San::SanSwitch::Switch.pod>.

=head2 METHODS

=item C<install_firmware>

=over 4

This method will start the firmwaredownload command on a Brocade switch, verify that
the command completed successfully, and then diplay the current firmware level.

=over 4

=item Returns

Returns SwitchUtil::FALSE if failure or SwitchUtil::TRUE if succesful

=item Mandatory Arguments

=over 4

=item -host_ip

IP address of a host to use for obtaining the firmware

=item -firmware_path

Full path to firmware

=item -host_username

Username to log in to the host for obtaining the firmware

=item -host_password

Password to log in to the host for obtaining the firmware

=back

=item Optional Arguments

=over 4

=item -firmware_install_timeout

Timeout to use when issuing firmwaredownload command. If this value is not
provided as an argument, the default value will be used.

=back

=item Outputs

Returns a true or false if success or failure respectively. If an error occurred
the errmsg is set to the reason

=item Examples

=back

    my %brocade_fw_hash = (
        -host_ip       => $MC_BROCADE_FW_DOWNLOAD_HOST,
        -firmware_path => $MC_BROCADE_FW_DOWNLOAD_SOURCEPATH,
        -host_username => $MC_BROCADE_FW_DOWNLOAD_USER,
        -host_password => $MC_BROCADE_FW_DOWNLOAD_PASSWORD,
    );
    $switch_obj->install_firmware(%brocade_fw_hash);
        if ($switch_obj->error) {
            @msg = $switch_obj->errmsg();
            &SanUtil::nate_log('-fatal', "inject_fault.thpl: Failed to install destination firmware on ($switch)\n" .
                               "Reason: @msg");
            return $TCD::FATAL;
        }

=back

=cut

=item C<supportsave>

=over 4

This method  Saves support information for RASLOG, TRACE, supportShow, core file,
FFDC data, and other support information to a remote FTP location.

=over 4

=item Returns

None

=over 4

=item Mandatory Arguments

=over 4

=item -host_ip

IP address of a host

=item -host_username

Username to log in to the host

=item -host_password

Password to log in to the host

=item -remote_dir

Remote directory path where you want the support information
to be saved

=back

=item Optional Arguments

=over 4

=item -timeout (sec)

timeout is used to wait for the command to be finished.
Defualt value will be used if not provided.
defualt value is set to 1800 secs | 30 mins

=back

=item protocol

Given protocol will be used to transfer support information.
Defualt is set to ftp. Choices are: ftp | scp | sftp.

=item Examples

=back

    $san_switch->supportsave(host_username=>$username, host_password=>$password
                 ,host_ip=>$ip,remote_dir=>$path);

=back

=cut
