# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a Brocade FC switch running
## NOS/VDX operating system
## @author thomasm@netapp.com
## @status private
## @pod ../Switch.pod

use strict;

# Package Name
##############
package San::SanSwitch::FC::Switch::BrocadeVDXFCSwitch;

# Package Requirements
######################
use San::SanSwitch;
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanSwitch);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

###############################################################################
# Method:    execute_brocade
# Objective: Execute a command to the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub execute_brocade {
    my ($self, $cmd, $timeout, $cmd_prefix, %args) = @_;

    my (@output, $quiet);

    $self->_clear_err();

    # Default $quiet to 0
    $quiet = 0;

    # Process the args
    foreach (keys %args) {
        if (/^-?quiet$/) {
            if (not length($quiet = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return ();
            }

            # Make sure we received an acceptable value
            unless ($quiet == 0 or $quiet == 1) {
                $self->_set_err("Invalid value '$quiet' for argument '$_'");
                return ();
            }
            next;
        }
    }

    # If we have a command prefix defined ($cmd_prefix), put it in front of
    # $cmd.  A good example of how this might be used is prepending "echo y|" to
    # the front of a command so that the echo y is piped into an ad --exec.

    if (defined($cmd_prefix) and $cmd_prefix ne "") {
        $cmd = $cmd_prefix . " " . $cmd;
    }

    # Execute the command and return the output.  Pass the timeout flag
    # if $timeout is defined.
    if (defined($timeout) and $timeout ne "") {
        @output = $self->execute($cmd, $timeout, -quiet => $quiet);
    } else {
        @output = $self->execute($cmd, "", -quiet => $quiet);
    }

    return(@output);
}

###############################################################################
# Method:    get_firmware_version
# Objective: Get the firmware version of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_firmware_version {
    my ($self) = @_;
    my (@output, $output);
    my $version = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    # Example output:
    # Network Operating System Software
    # Network Operating System Version: 2.1.0
    # Copyright (c) 1995-2010 Brocade Communications Systems, Inc.
    # Build Time: 18:13:52 Sep 14, 2011
    # Firmware name: v2.1.0_bld32
    # Control Processor: Freescale Semiconductor 8548E with 2000 MB of memory
    # 
    # Appl     Primary/Secondary Versions 
    # ------------------------------------------
    # NOS      v2.1.0_bld32
    #          v2.1.0_bld32
    @output = $self->execute("show version");
    if (!$self->error) {
        $version = (grep(/Firmware\s+name\:\s+/i, @output))[0];
        $version =~ s/Firmware\s+name\:\s+//i;
    }

    return $version;
}

###############################################################################
# Method:    get_ad255_show
# Objective: Get the ad --show output from a switch
#            Not applicable on NOS/VDX - No Admin Domain support
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_ad255_show {
    my ($self) = @_;

    my (@output);

    $self->_clear_err();

    $self->_set_err("Admin Domains are not supported on NOS/VDX Brocade Switch");
    return (@output);
}

###############################################################################
# Method:    get_model
# Objective: Get the model of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_model {
    my ($self) = @_;
    my (@output, $output);
    my $model = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_system})) {
        @output = $self->execute("show system");
        $self->{cache}{show_system} = [ @output ];
    } else {
        @output = @{$self->{cache}{show_system}};
    }

    if (!$self->error) {
       $model = (grep(/Hardware Rev\s+\:\s+/i, @output))[0];
       $model =~ s/Hardware Rev\s+\:\s+//i;
    }

    # Determine English name of the switch model
    if ($model =~ /95\.2/) {
       $model = "6720";
    }

    return $model;
}

###############################################################################
# Method:    disable
# Objective: Disable the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable {
    my ($self, %args) = @_;
    my (@output);

    $self->_clear_err();

    # Bypass using generic execute subroutine so that the prompt can be
    # controlled.
    &Tharn::say($self->{conn}, "chassis disable", '-prompt',
                'Are you sure you want to disable all chassis ports now',
                '-timeout', $self->{timeout});
    &Tharn::say($self->{conn}, "y", '-prompt', "Chassis has been disabled",
                '-timeout', $self->{timeout});

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    enable
# Objective: Enable the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable {
    my ($self, %args) = @_;
    my (@output);

    $self->_clear_err();

    # Let's just run the disable on the switch
    my $cmd = "chassis enable";
    @output = $self->execute("chassis enable");
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: shutdown()
# Objective:       shutdown Brocade switch
#                  Not supported on VDX/NOS Switches
# Details:         See POD documentation at the end of this file
################################################################################
sub shutdown {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_set_err("Shutdown is not supported on Brocade VDX switches");

    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    supportsave
# Objective: NO-OP
# Details:   None
###############################################################################
sub supportsave {
    my ($self) = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    supportshow
# Objective: Run supportshow on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub supportshow {
    my ($self) = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    reboot
# Objective: Reboot the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    # force reconnect to switch
    $self->reconnect();
    if ($self->error) {
        $self->_set_err("Could not properly reconnect to the switch before reboot");
        return $SwitchUtil::FALSE;
    }

    &SanUtil::nate_sleep(10);

    # to do the reset, we need to get down to using say
    # since we going to execute command blind
    my $connection = $self->get_connection();
    &Tharn::say($self->{conn}, "reload", '-prompt',
                'Are you sure you want to reload the switch',
                '-timeout', $self->{timeout});
    eval {
        @output = &Tharn::say($self->{conn}, "y", '-blind');
    };

    $self->sanlog(-msg =>
                  "Waiting 60 seconds to allow reboot process to start");

    # sleep to allow boot process to start
    &SanUtil::nate_sleep(60);

    # Wait until the switch does not respond
    my $count = 0;
    while ($self->is_alive() == $SwitchUtil::FALSE and
           $count < $self->{reboot_timeout}) {
        &SanUtil::nate_sleep(1);
        $count++;
    }

    if ($count < $self->{reboot_timeout}) {
        # wait some additional time to allow telnet service to be
        # available and responding. Also wait some time to allow all
        # ports to log back in (or nameserver db will be empty).
        &SanUtil::nate_sleep(120);
        $self->reconnect();
        if ($self->error) {
            $self->_set_err("Could not properly reconnect to the switch after reboot");
            return $SwitchUtil::FALSE;
        }

        # invalidate the switchShow cache, since we want to do
        # an lookup after reboot with fresh information.
        $self->{cache}{show_system} = undef;

        return $SwitchUtil::TRUE;
    } else {
        $self->_set_err("Could not properly reboot the switch (exceeded allowable reboot/ping time)");
        return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    info
# Objective: Get switch information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;

    my (@output, $cmd);

    # execute some commands just to get data into logs
    # Don't validate data
    $cmd = "show process memory summary";
    $self->execute($cmd);
    $cmd = "show process cpu summary";
    $self->execute($cmd);
    $cmd = "show system monitor";
    $self->execute($cmd);
    $cmd = "show vcs";
    $self->execute($cmd);

    push (@output, "Switch name:           " . $self->hostname(),
                   "Switch type:           " . $self->hosttype(),
                   "Switch FW version:     " . $self->fw_version(),
                   "Switch Model:          " . $self->model());

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    ntp_sync
# Objective: Synchronize the clock on the switch with an NTP server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($ntp_server, $ntp_time_zone);
    my ($gmt_offset_hour, $gmt_offset_min, $original_persistent);
    my (@output, $ntp_server_ip);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Invalid value for argument '-ntp_server'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '-ntp_time_zone'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # Set the offset from GMT for the timezone
    if ($ntp_time_zone =~ /EST/i) {
        $ntp_time_zone = "America/New_York";
    } elsif ($ntp_time_zone =~ /CST/i) {
        $ntp_time_zone = "America/Chicago";
    } elsif ($ntp_time_zone =~ /PST/i) {
        $ntp_time_zone = "America/Los_Angeles";
    } elsif ($ntp_time_zone =~ /IST/i) {
        $ntp_time_zone = "Asia/Kolkata";
    } elsif ($ntp_time_zone =~ /MST/i) {
        $ntp_time_zone = "America/Denver";
    } else {
        $self->_set_err("Time zone $ntp_time_zone not currently supported");
        return $SwitchUtil::FALSE;
    }

    $self->sanlog(-msg => "Setting the Time Zone (TZ) to $ntp_time_zone");

    # setup the proper timezone
    my $cmd = "clock timezone $ntp_time_zone";
    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return $SwitchUtil::FALSE;
    }

    # get IP address since switch may not have DNS running
    my (undef, undef, undef, undef, @addrs) = gethostbyname($ntp_server);
    foreach my $ip (@addrs) {
        my ($a, $b, $c, $d) = unpack('C4', $ip);
        $ntp_server_ip = "$a\.$b\.$c\.$d";
    }

    # execute set of time server
    my @cmds = ( "conf t", "ntp server $ntp_server_ip", "end" );
    foreach $cmd (@cmds) {
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
    }

    # get date to have in log
    $cmd = "show clock";
    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return $SwitchUtil::FALSE;
    }

    $self->sanlog(-msg => "Local time is: $output[0]");

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    dump_topology
# Objective: Dump the fabric topology
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_topology {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $switch_name = $self->hostname();
    if ($self->error()) {
        return $SwitchUtil::FALSE;
    }

    my ($cmd, @output);

    # Run a "show fabric route topology"
    $cmd = "show fabric route topology";
    $self->sanlog(-msg => "Executing command ($cmd)");
    @output = $self->execute_brocade($cmd, "", "", -quiet => 0);
    if ($self->error()) {
        $self->_set_err("Unable to run command ($cmd) on switch '$switch_name':\n", $self->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Run a "show fabric all"
    $cmd = "show fabric all";
    $self->sanlog(-msg => "Executing command ($cmd)");
    @output = $self->execute_brocade($cmd, "", "", -quiet => 0);
    if ($self->error()) {
        $self->_set_err("Unable to run command ($cmd) on switch '$switch_name':\n", $self->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    install_firmware
# Objective: Install firmware on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub install_firmware {
    my ($self, %args) = @_;

    #TODO: This subroutine to be completed at a later date
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    get_log
# Objective: Retrieve the log file from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_log {
    my ($self) = @_;

    $self->_clear_err();

    my $cmd = "show logging raslog";
    my @output = $self->execute($cmd, $self->{log_timeout});
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return undef;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_config
# Objective: Retrieve the configuration from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_config {
    my ($self) = @_;

    $self->_clear_err();

    my $cmd = "show running-config";
    my @output = $self->execute($cmd, $self->{log_timeout});
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return undef;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_state
# Objective: Retrieve the state of the switch (enabled/disabled)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state {
    my ($self, %args) = @_;
    my $state = "unknown";

    $self->_clear_err();

    my $cmd = "show system";
    my @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return $state;
    }

    $state = (grep(/Status\s+\:\s+/i, @output))[0];
    $state =~ s/s\s+\:\s+//i;

    return &SwitchUtil::normalize_state($state);
}

###############################################################################
# Method:    get_vf_ids
# Objective: Retrieve the virtual farbic IDs on this switch
#            Virtual fabrics are not supported on VDX/NOS at this time,
#            simply return empty array
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vf_ids {
    my ($self) = @_;

    my @output = ();

    $self->_clear_err();

    return(@output);

}

###############################################################################
# Method:    set_vf_context
# Objective: set the context of the switch you want commands to be execute
#            to in a virtual fabric context. No-OP on VDX switches.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_vf_context {
    my ($self, $fid) = @_;

    return $SwitchUtil::TRUE;

}

###############################################################################
# Method:    get_current_vf_context_id
# Objective: get the current vf context id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_vf_context_id {
    my ($self) = @_;

    return $self->{vf_context_id};
}

###############################################################################
# Method:    module_reboot
# Objective: Reboot the module of a director switch.
# Details:   Unimplemented.
###############################################################################
sub module_reboot {
    my $self = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type.");
    return $SwitchUtil::FALSE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new BrocadeFC switch object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $self, $conn);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        }
    }

    if (!defined($conn) || ref($conn) ne "Connectrec") {
        $self->_set_err("Invalid value for argument '-conn'");
        return $self;
    }

    # store connection
    $self->{conn} = $conn;

    return $self;
}

###############################################################################
# Method:    _get_domain_id
# Objective: Get the domain id of the switch (in VDX case rbridge_id)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_domain_id {
    my ($self) = @_;

    $self->_clear_err();

    my $domain_id = "";

    my $cmd = "show license id";
    my @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Unable to execute command ($cmd)\n" .
                        $self->errmsg());
        return "";
    }

    $domain_id = (grep(/\s+(\d+)\s+(\s+)/i, @output))[0];
    $domain_id =~ m/\s+(\d+)\s+(\s+)/i;
    $domain_id = $1;

    return $domain_id;
}

1;

=pod

=head1 Package Name

BrocadeVDXFCSwitch.pm

=head1 Author

Tom Masterson

=head1 Description

For description of methods see B<San::SanSwitch::Switch.pod>.

=cut

