# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a Cisco FC switch
## @author evan@netapp.com
## @status private
## @pod ../Switch.pod

use strict;

# Package Name
##############
package San::SanSwitch::FC::Switch::CiscoFCSwitch;

# Package Requirements
######################
use San::SanSwitch;
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanSwitch);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

###############################################################################
# Method:    get_firmware_version
# Objective: Get the firmware version of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_firmware_version {
    my ($self) = @_;
    my (@output, $output);
    my $version = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_version})) {
        @output = $self->execute("show version");
        $self->{cache}{show_version} = \@output;
    } else {
        @output = @{$self->{cache}{show_version}};
    }
    $output = join('', @output);

    if (!$self->error) {
        if ($output =~ /SAN-OS/ && $output =~ /system:\s+version\s+(\S+)/) {
            $version = "SAN-OS " . $1;
        } elsif ($output =~ /FabricWare/ && $output =~ /system:\s+(\S+)/) {
            $version = "FabricWare " . $1;
        } elsif ($output =~ /NX-OS|Cisco UCS/ && $output =~ /system:\s+version\s+(\S+)/) {
            $version = "NX-OS " . $1;
        }
    }

    return $version;
}

###############################################################################
# Method:    get_model
# Objective: Get the model of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_model {
    my ($self) = @_;
    my (@output, $output);
    my $model = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_version})) {
        @output = $self->execute("show version");
        $self->{cache}{show_version} = \@output;
    } else {
        @output = @{$self->{cache}{show_version}};
    }

    if (!$self->error) {
        for (my $i = 0; $i < @output; $i++) {
            if ($output[$i] =~ /^Hardware/) {
                $model = $output[$i + 1];
                $model =~ s/\s+//;
                last;
            }
        }
    }

    return $model;
}

###############################################################################
# Method:    disable
# Objective: Disable the switch (not currently supported for Cisco)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable {
    my ($self, %args) = @_;

    my $vf = "";

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?vf/) {
            if (not length($vf = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vf'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # If vf to disable are given then go ahead and disable, otherwise error out
    # since there is not disable.
    if ($vf ne "") {
        $self->_vf_disable($vf);
        if ($self->error) {
            #Just use the error that was set.
            return $SwitchUtil::FALSE;
        }

        # Return true since we disable successfully.
        return $SwitchUtil::TRUE;
    } else {
        $self->_set_err("Error: CiscoFCSwitch::disable() " .
                        "this subroutine is not supported on Cisco switches");
    }

    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    enable
# Objective: Enable the switch (not currently supported for Cisco)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable {
    my ($self, %args) = @_;
    my $vf = "";

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?vf/) {
            if (not length($vf = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vf'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # If vf to enable are given then go ahead and disable, otherwise error out
    # since there is not disable.
    if ($vf ne "") {
        $self->_vf_enable($vf);
        if ($self->error) {
            #Just use the error that was set.
            return $SwitchUtil::FALSE;
        }

        # Return true since we disable successfully.
        return $SwitchUtil::TRUE;
    } else {
        $self->_set_err("Error: CiscoFCSwitch::enable() " .
                        "this subroutine is not supported on Cisco switches");
    }

    return $SwitchUtil::FALSE;
}

################################################################################
# Subroutine Name: shutdown()
# Objective:       shutdown cisco switch
# Details:         See POD documentation at the end of this file
################################################################################
sub shutdown {
    my ($self) = @_;

    $self->_clear_err();

    $self->_set_err("Shutdown is not supported on Cisco switches");

    return $ClientUtil::FALSE;
}

###############################################################################
# Method:    supportsave
# Objective: NO-OP
# Details:   None
###############################################################################
sub supportsave {
    my ($self) = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    supportshow
# Objective: Run supportshow on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub supportshow {
    my ($self) = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    reboot
# Objective: Reboot the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;
    my $count = 0;

    $self->_clear_err();

    my $original_persistent = $self->{persistent};

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->options(-persistent => 1);

    # execute a show version command to send via execute and it will do the
    # "right thing" on re-connect.
    my @output = $self->execute("show version");

    # Bypass using generic execute subroutine so that the prompt can be
    # controlled
    &Tharn::say($self->{conn}, "reload", '-prompt',
                'This command will reboot the system.',
                '-timeout', $self->{timeout});
    eval {
        &Tharn::say($self->{conn}, "y", '-noprompt', '-prompt', "y",
                    '-timeout', $self->{timeout});
    };

    # Restore persistent setting
    $self->options(-persistent => $original_persistent);

    # sleep to allow boot process to start
    sleep(30);

    # Wait until the switch does not respond
    while ($self->is_alive() == $SwitchUtil::FALSE &&
           $count < $self->{reboot_timeout}) {
        sleep(1);
        $count++;
    }

    if ($count < $self->{reboot_timeout}) {
        # wait some additional time to allow telnet service to be
        # available and responding. Also wait some time to allow all
        # ports to log back in (or nameserver db will be empty).
        sleep(120);
        $self->reconnect();
        if ($self->error) {
            $self->_set_err("Error: CiscoFCSwitch::reboot() " .
                            "Could not properly reconnect to the switch after reboot");
            return $SwitchUtil::FALSE;
        }

        return $SwitchUtil::TRUE;
    } else {
        $self->_set_err("Error: CiscoFCSwitch::reboot() " .
                        "Could not properly reboot the switch (exceeded allowable reboot/ping time)");
        return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    module_reboot
# Objective: Reboot a module of a Cisco director switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub module_reboot {
    my ($self, %args) = @_;
    my $module = "";

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?module/) {
            if (not length($module = $args{$_})) {
                $self->_set_err("Invalid value for argument '-module'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # Check to make sure the user gave us a module to reboot.
    if ($module ne "") {
        my @output = $self->execute("reload module $module");
        my $output = join('', @output);
        if (not $output =~ /reloading module $module/i){
            $self->_set_err("Unable to reboot module $module.\nOutput: $output");
            return $SwitchUtil::FALSE;
        }

        ## What if we are disconnected from a switch here???

        &SanUtil::nate_log('comment', "Waiting for module to come up...");

        # Wait for module to come back up.
        my $count = 0;
        do {
            sleep(30);
            @output = $self->execute("show module $module | grep \'^$module\'");
            $output = join('', @output);
            $count++;
            if ($count > 20) {
                $self->_set_err("Module failed to come up after 10 minutes.");
                return $SwitchUtil::FALSE;
            }
        } while ($output !~ /ok/i);

        &SanUtil::nate_log('pass', "Module reload successful!");

    } else {
        $self->_set_err("Error: CiscoFCSwitch::module_reboot() " .
                            "this subroutine needs to have a module passed in.");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    info
# Objective: Get switch information
# Details:   See POD documentation at the end of this file.
###############################################################################

sub info {
    my ($self) = @_;
    my (@output);
    my $Config = $self->config();
    my $Debug = join("\n",@$Config);
    $self->_clear_err();

    push @output, "Switch name:           " . $self->hostname(), 
                  "Switch type:           " . $self->hosttype(),
                  "Switch FW version:     " . $self->fw_version(), 
                  "Switch Model:          " . $self->model(),
                  "Switch Domain ID:      " . $self->domain_id(),
		  "Switch Config:         " . $Debug;
    push @output, "\n";

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    ntp_sync
# Objective: Synchronize the clock on the switch with an NTP server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($ntp_server, $ntp_time_zone);
    my ($time_zone, $gmt_offset_hour, $gmt_offset_min);
    my (@output);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: CiscoFCSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_server'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: CiscoFCSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_time_zone'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    my $original_persistent = $self->{persistent};

    # Set the offset from GMT for the timezone
    if ($ntp_time_zone =~ /EST/i) {
        $gmt_offset_hour = -5;
        $gmt_offset_min  = 0;
        $time_zone  = "EST";
    } elsif ($ntp_time_zone =~ /PST/i) {
        $gmt_offset_hour = -8;
        $gmt_offset_min  = 0;
        $time_zone  = "PST";
    } elsif ($ntp_time_zone =~ /MST/i) {
        $gmt_offset_hour = -7;
        $gmt_offset_min  = 0;
        $time_zone  = "MST";
    } elsif ($ntp_time_zone =~ /IST/i) {
        $gmt_offset_hour = 5;
        $gmt_offset_min  = 30;
        $time_zone  = "IST";
    } else {
        $self->_set_err("Error: CiscoFCSwitch::ntp_sync() " .
                        "Time zone $ntp_time_zone not currently supported");
        return $SwitchUtil::FALSE;
    }

    my $gmt_string = sprintf("%02d:%02d", $gmt_offset_hour, $gmt_offset_min);
    $self->sanlog(-msg => "Setting the Time Zone (TZ) to $time_zone " .
                          "(GMT $gmt_string)");

    # get IP address since switch may not have DNS running
    my ($ntp_server_ip);
    my (undef, undef, undef, undef, @addrs) = gethostbyname($ntp_server);
    foreach my $ip (@addrs) {
        my ($a, $b, $c, $d) = unpack('C4', $ip);
        $ntp_server_ip = "$a\.$b\.$c\.$d";
    }

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->options(-persistent => 1);

    my @cmds = ("conf t",
                "ntp server $ntp_server_ip",
                "clock timezone $time_zone $gmt_offset_hour $gmt_offset_min",
                "end");

    # Check to see if the switch is Nexus switch and set the proper command array
    if ($self->fw_version() =~ /NX-OS/i and
        $self->model() =~ /nexus/i) {
        @cmds = ("conf t",
                 "ntp server $ntp_server_ip use-vrf management",
                 "clock timezone $time_zone $gmt_offset_hour $gmt_offset_min",
                 "ntp distribute",
                 "ntp commit",
                 "end");
    }

    my $cmd_error = 0;
    foreach my $cmd (@cmds) {
        if ($cmd =~ /show clock/) {
            # Wait a few seconds for the updates to take effect for show clock
            sleep(3);
        }
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: CiscoFCSwitch::ntp_sync() " .
                            "Could not execute command $cmd");
            # if we have an error, kick out of foreach loop with 
            # error set
            $cmd_error = 1;
            last;
        }
    }

    # handle error, by attempting to get out of config mode
    # and then returning with the proper error set.
    if ($cmd_error == 1) {
        @cmds = (" ", "end");
        foreach my $cmd (@cmds) {
            @output = $self->execute($cmd);
        }
        return $SwitchUtil::FALSE;
    }

    # skip for FabricWare (9020) switches - command
    # does not exist.
    my $fw_revsion = $self->get_firmware_version();
    if ($fw_revsion !~ /FabricWare/) {
        @cmds = ("ntp sync-retry",
                 "show clock");
        foreach my $cmd (@cmds) {
            @output = $self->execute($cmd);
            if ($self->error) {
                $self->_set_err("Error: CiscoFCSwitch::ntp_sync() " .
                                "Could not execute command $cmd");
                return $SwitchUtil::FALSE;
            }
        }
    }

    # Restore persistent setting
    $self->options(-persistent => $original_persistent);

    $self->sanlog(-msg => "Local time is: $output[0]");

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    dump_topology
# Objective: Dump the fabric topology
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_topology {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $switch_name = $self->hostname();
    if ($self->error()) {
        return $SwitchUtil::FALSE;
    }

    # Make sure this isn't an NPV/AG switch
    if ($self->is_npv_enabled()) {
        $self->sanlog(-msg => "Skipping switch '$switch_name' since it is in NPV or AG mode.");
        return $SwitchUtil::FALSE;
    }

    my ($cmd, @output);

    # Run a "show fcdomain domain-list"
    $cmd = "show fcdomain domain-list";
    $self->sanlog(-msg => "Executing command ($cmd)");
    @output = $self->execute($cmd);
    if ($self->error()) {
        $self->_set_err("Unable to run command ($cmd) on switch '$switch_name':\n", $self->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Run a "show topology"
    $cmd = "show topology";
    $self->sanlog(-msg => "Executing command ($cmd)");
    @output = $self->execute($cmd);
    if ($self->error()) {
        $self->_set_err("Unable to run command ($cmd) on switch '$switch_name':\n", $self->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Run a "show fcs ie"
    $cmd = "show fcs ie";
    $self->sanlog(-msg => "Executing command ($cmd)");
    @output = $self->execute($cmd);
    if ($self->error()) {
        $self->_set_err("Unable to run command ($cmd) on switch '$switch_name':\n", $self->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    install_firmware
# Objective: Install firmware on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub install_firmware {
    my ($self, %args) = @_;

    $self->_clear_err();
    #TODO: This subroutine to be completed at a later date
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    get_log
# Objective: Retrieve the log file from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_log {
    my ($self) = @_;

    $self->_clear_err();

    my @output = $self->execute("show logging logfile");
    if ($self->error) {
        $self->_set_err("Error: CiscoFCSwitch::get_log() " .
                        "Could not retrieve log file");
        return undef;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_config
# Objective: Retrieve the configuration from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_config {
    my ($self) = @_;

    $self->_clear_err();

    my @output = $self->execute("show run");
    if ($self->error) {
        $self->_set_err("Error: CiscoFCSwitch::get_config() " .
                        "Could not retrieve configuration");
        return undef;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_state
# Objective: Retrieve the state of the switch (online/offline). Will return
#            online or offline if all VSANs are online or offline respectively.
#            Returns unknown when there are mixed.
#            If a vf argument is passed it will operate on that vf only.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state {
    my ($self, %args) = @_;
    my $state = "unknown";

    $self->_clear_err();

    my $vf = "";

    foreach (keys %args) {
        if (/-?vf/) {
            if (not length($vf = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vf'");
                return $state;
            }
            next;
        }
    }

    # If the state was requested for a specific vsan then return that value
    if ($vf ne "") {
        return &SwitchUtil::normalize_state($self->_get_vf_state($vf));
    }

    # Otherwise check if all configured vsans are enabled/disabled

    # Get a list of the configured vsans
    my $cmd = "show vsan | include \"vsan \" | include information";
    my @output = ();

    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return $state;
    }

    # Parse and store the vsan #'s
    my @configured_vsans = ();
    foreach my $line (@output) {
        if ($line =~ /^vsan\s+(\d+).*$/) {
            push(@configured_vsans, $1);
        }
    }

    # By default, we assume there are no mixed states
    my $has_mixed_state = 0;

    if (scalar(@configured_vsans)) {

        # Use the first element's status as the baseline
        my $common_state = &SwitchUtil::normalize_state($self->_get_vf_state(@configured_vsans[0]));

        # Check to see if the rest of the configured VSANs match the state
        foreach (@configured_vsans) {
            my $vsan_state = &SwitchUtil::normalize_state($self->_get_vf_state($_));

            # If the current state matches the common state then move on, otherwise we have mixed states
            if ($vsan_state eq $common_state) {
                next;
            } else {
                $has_mixed_state = 1;
                last;
            }
        }

        # If every VSAN shared the same state, prepare to return this value
        if (!$has_mixed_state) {
            $state = $common_state;
        }
    }

    return $state;
}

###############################################################################
# Method:    get_current_vf_context_id
# Objective: get the current vf context id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_vf_context_id {
    my ($self) = @_;

    # NOOP for Cisco

    return "";
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new CiscoFC switch object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: CiscoFCSwitch::_new() " .
                                "Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } 
    }

    if (!defined($conn) || ref($conn) ne "Connectrec") {
        $self->_set_err("Error: CiscoFCSwitch::_new() " .
                        "Invalid value for argument '-conn'");
        return $self;
    }

    # set connection
    $self->{conn} = $conn;

    return $self;
}

###############################################################################
# Method:    _get_domain_id
# Objective: Get the domain id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_domain_id {
    my ($self) = @_;

    # NO-OP

    return "";
}

###############################################################################
# Method:    _get_npiv
# Objective: Retrieve information to determine if switch is acting NPV mode
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_npiv {
    my ($self) = @_;

    $self->_clear_err();

    # determine if this is a NPIV enabled switch, ignore test if
    # FabricWare (9020)
    my $fw_revsion = $self->get_firmware_version();
    if ($fw_revsion !~ /FabricWare/) {
        my @output = $self->execute("show running-config \| include npv");
        if (grep(/feature npiv|npiv enable/i, @output) >= 1) {
            return(1);
        } else {
            return(0);
        }
    } else {
        return(0);
    }
}

###############################################################################
# Method:    _get_npv_mode
# Objective: Get the npv mode of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_npv_mode {
    my ($self) = @_;

    $self->_clear_err();

    # determine if this is a NPV enabled switch, ignore test if
    # FabricWare (9020)
    my $fw_revsion = $self->get_firmware_version();
    if ($fw_revsion !~ /FabricWare/) {
        my @output = $self->execute("show running-config \| include npv");
        if (grep(/feature npv|feature fcoe-npv|npv enable/i, @output) >= 1) {
            return(1);
        } else {
            return(0);
        }
    } else {
        return(0);
    }
}

################################################################################
# Subroutine Name: _vf_enable
# Objective:       enable a vf (vsan on Cisco)
# Details:         See POD documentation at the end of this file
################################################################################
sub _vf_enable {
    my ($self, $vf) = @_;

    my ($cmd_error, @output);
    my $original_persistent = $self->{persistent};

    $self->_clear_err();

    my $vf_state = $self->_get_vf_state($vf);
    if ($self->error) {
        # Error is already set
        return $SwitchUtil::FALSE;
    }

    # Make sure the vf state is not already active
    if ($vf_state =~ /online/i) {
        $self->_set_err("The vf state is already enabled.");
        return $SwitchUtil::FALSE;
    }

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->options(-persistent => 1);

    my @cmds = ("conf t",
                "vsan database",
                "no vsan $vf suspend",
                "end");

    $cmd_error = 0;
    foreach my $cmd (@cmds) {
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Could not execute command $cmd");
            # if we have an error, kick out of foreach loop with 
            # error set
            $cmd_error = 1;
            last;
        }
    }

    # handle error, by attempting to get out of config mode
    # and then returning with the proper error set.
    if ($cmd_error == 1) {
        @cmds = (" ", "end");
        foreach my $cmd (@cmds) {
            @output = $self->execute($cmd);
        }
    }

    # Restore persistent setting
    $self->options(-persistent => $original_persistent);

    # If we had an error we need to return false
    if ($cmd_error == 1) {
        return $SwitchUtil::FALSE;
    }

    # Wait for vsan to be suspended and then check the status
    sleep(15);
    $vf_state = $self->_get_vf_state($vf);
    if ($self->error) {
        # Error is already set
        return $SwitchUtil::FALSE;
    }

    # Make sure the vf state is active
    if ($vf_state !~ /online/i) {
        $self->_set_err("Unable to enable the vf successfully.");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

################################################################################
# Subroutine Name: _vf_disable
# Objective:       Disable a vf (vsan on Cisco)
# Details:         See POD documentation at the end of this file
################################################################################
sub _vf_disable {
    my ($self, $vf) = @_;

    my ($cmd_error, @output);
    my $original_persistent = $self->{persistent};

    $self->_clear_err();

    my $vf_state = $self->_get_vf_state($vf);
    if ($self->error) {
        # Error is already set
        return $SwitchUtil::FALSE;
    }

    # Make sure the vf state is not already suspended
    if ($vf_state =~ /offline/i) {
        $self->_set_err("The vf state is already disabled.");
        return $SwitchUtil::FALSE;
    }

    # Make connection persistent so commands can be sent through one connection
    # session
    $self->options(-persistent => 1);

    my @cmds = ("conf t",
                "vsan database",
                "vsan $vf suspend",
                "end");

    $cmd_error = 0;
    foreach my $cmd (@cmds) {
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Could not execute command $cmd");
            # if we have an error, kick out of foreach loop with 
            # error set
            $cmd_error = 1;
            last;
        }
    }

    # handle error, by attempting to get out of config mode
    # and then returning with the proper error set.
    if ($cmd_error == 1) {
        @cmds = (" ", "end");
        foreach my $cmd (@cmds) {
            @output = $self->execute($cmd);
        }
    }

    # Restore persistent setting
    $self->options(-persistent => $original_persistent);

    # If we had an error we need to return false
    if ($cmd_error == 1) {
        return $SwitchUtil::FALSE;
    }

    # Wait for vsan to be suspended and then check the status
    sleep(15); 
    $vf_state = $self->_get_vf_state($vf);
    if ($self->error) {
        # Error is already set
        return $SwitchUtil::FALSE;
    }

    # Make sure the vf state is suspended
    if ($vf_state !~ /offline/i) {
        $self->_set_err("Unable to disable the vf successfully.");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

sub _get_vf_state {
    my ($self, $vf) = @_;

    my $state = "unknown";

    # Normal output
    # metro4-fab2# show vsan 2
    # vsan 2 information
    #     name:VSAN2  state:active
    #     interoperability mode:default
    #     loadbalancing:src-id/dst-id/oxid  
    #     operational state:up

    # output with the include statements
    # metro4-fab2# show vsan 2 | include state: | include name:
    #     name:VSAN2  state:active 

    my $cmd = "show vsan $vf | include state: | include name:";
    my @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Could not execute command $cmd");
        return $state;
    } else {
        my @parsed_state = split(/state:/i,$output[0]);
        $state = $parsed_state[1];
    }

    return &SwitchUtil::normalize_state($state);
}

1;

=pod

=head1 Package Name

CiscoFCSwitch.pm

=head1 Author

Evan Chiu

=head1 Description

For description of methods see B<San::SanSwitch::Switch.pod>.

=cut

