# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a McData FC switch
## @author thomasm@netapp.com
## @status private
## @pod ../Switch.pod

use strict;

# Package Name
##############
package San::SanSwitch::FC::Switch::McDataFCSwitch;

# Package Requirements
######################
use San::SanSwitch;
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanSwitch);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

###############################################################################
# Method:    get_firmware_version
# Objective: Get the firmware version of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_firmware_version {
    my ($self) = @_;
    my (@output, $output);
    my $version = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_system})) {
        @output = $self->execute("show system");
        $self->{cache}{show_system} = [ @output ];
    } else {
        @output = @{$self->{cache}{show_system}};
    }

    if (!$self->error) {
        $version = (grep(/Firmware Version\:/i, @output))[0];
        $version =~ s/Firmware Version\:\s+//i;
    }

    return $version;
}

###############################################################################
# Method:    get_model
# Objective: Get the model of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_model {
    my ($self) = @_;
    my (@output, $output);
    my $model = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_system})) {
        @output = $self->execute("show system");
        $self->{cache}{show_system} = [ @output ];
    } else {
        @output = @{$self->{cache}{show_system}};
    }

    if (!$self->error) {
       $model = (grep(/Model Name\:/i, @output))[0];
       $model =~ s/Model Name\:\s+//i;
    }

    return $model;
}

###############################################################################
# Method:    dump_topology
# Objective: Dump the fabric topology
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_topology {
    my ($self, %args) = @_;

    $self->_clear_err();

    # No-Op for McData

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    disable
# Objective: Disable the switch 
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    my $cmd = "maint system setOnlineState 0";
    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Error: McDataFCSwitch::disable() " .
                        "Could not execute command $cmd");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    enable
# Objective: Enable the switch 
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    my $cmd = "maint system setOnlineState 1";
    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Error: McDataFCSwitch::enable() " .
                        "Could not execute command $cmd");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    shutdown
# Objective: NO-OP for McData
# Details:   See POD documentation at the end of this file.
###############################################################################
sub shutdown {
    my ($self) = @_;

    # LATIF add set error
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    supportsave
# Objective: NO-OP
# Details:   None
###############################################################################
sub supportsave {
    my ($self) = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    supportshow
# Objective: Run supportshow on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub supportshow {
    my ($self) = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    reboot
# Objective: Reboot the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    my $cmd = "maint system ipl";
    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Error: McDataFCSwitch::reboot() " .
                        "Could not execute command $cmd");
        return $SwitchUtil::FALSE;
    }

    # Wait until the switch does not respond
    my $count = 0;
    while ($self->is_alive() == $SwitchUtil::FALSE && 
           $count < $self->{reboot_timeout}) {
        sleep(1);
        $count++;
    }

    if ($count < $self->{reboot_timeout}) {
        # McData switch really does not reboot, but resets
        # at the SW level, so this operation should be very
        # quick. Still want to reconnect and start connection
        # from a fresh point of view.
        sleep(120);
        $self->reconnect();

        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    info
# Objective: Get switch information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my (@output);

    push @output, "Switch name:           " . $self->hostname(), 
                  "Switch type:           " . $self->hosttype(),
                  "Switch FW version:     " . $self->fw_version(), 
                  "Switch Model:          " . $self->model();
    push @output, "\n";

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    ntp_sync
# Objective: Synchronize the clock on the switch with an NTP server.
#            No-op on McData, since the firmware does not support NTP.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($ntp_server, $ntp_time_zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: McDataFCSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_server'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: McDataFCSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_time_zone'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => 
                  "McDataFCSwitch::ntp_sync() McData firmware does not support NTP");
    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    install_firmware
# Objective: Install firmware on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub install_firmware {
    my ($self, %args) = @_;

    #TODO: This subroutine to be completed at a later date
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    get_log
# Objective: Retrieve the log file from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_log {
    my ($self) = @_;

    $self->_clear_err();

    my @output = $self->execute("show eventlog", $self->{log_timeout});
    if ($self->error) {
        $self->_set_err("Error: McDataFCSwitch::get_log() " .
                        "Could not retrieve log file");
        return undef;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_config
# Objective: Retrieve the configuration from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_config {
    my ($self) = @_;

    $self->_clear_err();

    my @output = $self->execute("show port config");
    if ($self->error) {
        $self->_set_err("Error: McDataFCSwitch::get_config() " .
                        "Could not retrieve configuration");
        return undef;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_state
# Objective: Retrieve the state of the switch (enabled/disabled)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state {
    my ($self) = @_;
    my (@output, $output);
    my $state = "unknown";

    $self->_clear_err();

    # Don't want to get from cache since state need to reflect
    # reality.
    my $cmd = "show switch";
    @output = $self->execute($cmd);
    if ($self->error) {
        $self->_set_err("Error: McDataFCSwitch::get_state() " .
                        "Could not execute command $cmd");
        return undef;
    } else {
       $state = (grep(/^State\:/i, @output))[0];
       $state =~ s/^State:\s+//i;
    }

    return $state;
}

###############################################################################
# Method:    get_current_vf_context_id
# Objective: get the current vf context id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_vf_context_id {
    my ($self) = @_;

    # NOOP for McData

    return "";
}

###############################################################################
# Method:    module_reboot
# Objective: Reboot the module of a director switch.
# Details:   Unimplemented.
###############################################################################
sub module_reboot {
    my $self = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type.");
    return $SwitchUtil::FALSE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new McDataFC switch object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: McDataFCSwitch::_new() " .
                                "Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } 
    }

    if (!defined($conn) || ref($conn) ne "Connectrec") {
        $self->_set_err("Error: McDataFCSwitch::_new() " .
                        "Invalid value for argument '-conn'");
        return $self;
    }

    $self->{conn} = $conn;
    return $self;
}

###############################################################################
# Method:    _get_domain_id
# Objective: Get the domain id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_domain_id {
    my ($self) = @_;

    # NO-OP

    return "";
}


1;

=pod

=head1 Package Name

McDataFCSwitch.pm

=head1 Author

Tom Masterson

=head1 Description

For description of methods see B<San::SanSwitch::Switch.pod>.

=cut

