# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a Qlogic FC switch
## @author thomasm@netapp.com
## @status private
## @pod ../Switch.pod

use strict;

# Package Name
##############
package San::SanSwitch::FC::Switch::QlogicFCSwitch;

# Package Requirements
######################
use San::SanSwitch;
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanSwitch);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

###############################################################################
# Method:    get_firmware_version
# Objective: Get the firmware version of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_firmware_version {
    my ($self) = @_;
    my (@output);
    my $version = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_about})) {
        @output = $self->execute("show about");
        $self->{cache}{show_about} = [ @output ];
    } else {
        @output = @{$self->{cache}{show_about}};
    }

    if (!$self->error) {
        $version = (grep(/\s+ActiveSWVersion/i, @output))[0];
        $version =~ s/\s+ActiveSWVersion\s+//i;
    }

    return $version;
}

###############################################################################
# Method:    get_model
# Objective: Get the model of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_model {
    my ($self) = @_;
    my (@output);
    my $model = "unknown";

    $self->_clear_err();

    # Check if we have a cached output already
    if (!defined($self->{cache}{show_about})) {
        @output = $self->execute("show about");
        $self->{cache}{show_about} = [ @output ];
    } else {
        @output = @{$self->{cache}{show_about}};
    }

    if (!$self->error) {
       $model = (grep(/\sSystemDescription/i, @output))[0];
       $model =~ s/\s+SystemDescription\s+//i;
    }

    return $model;
}

###############################################################################
# Method:    dump_topology
# Objective: Dump the fabric topology
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_topology {
    my ($self, %args) = @_;

    $self->_clear_err();

    # No-Op for Qlogic

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    disable
# Objective: Disable the switch 
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    my $original_persistent = $self->{persistent};

    my @cmds = ( "admin stop", "admin start", 
                 "set switch state offline" , "admin stop" );

    # Make connection persistent so commands can be sent through
    # one connection session 
    $self->options(-persistent => 1); 

    foreach my $cmd (@cmds) {
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: QlogicFCSwitch::disable() " .
                            "Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
    }
   
    $self->options(-persistent => $original_persistent);

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    enable
# Objective: Enable the switch 
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    my $original_persistent = $self->{persistent};

    my @cmds = ( "admin stop", "admin start", 
                 "set switch state online" , "admin stop" );

    # Make connection persistent so commands can be sent through
    # one connection session 
    $self->options(-persistent => 1); 

    foreach my $cmd (@cmds) {
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: QlogicFCSwitch::enable() " .
                            "Could not execute command switchenable");
            return $SwitchUtil::FALSE;
        }
    }

    $self->options(-persistent => $original_persistent);

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    shutdown
# Objective: NO-OP for Qlogic
# Details:   See POD documentation at the end of this file.
###############################################################################
sub shutdown {
    my ($self) = @_;

    # LATIF set error
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    supportsave
# Objective: NO-OP
# Details:   None
###############################################################################
sub supportsave {
    my ($self) = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    supportshow
# Objective: Run supportshow on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub supportshow {
    my ($self) = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type");
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    reboot
# Objective: Reboot the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;
    my $count = 0;
    my (@output);

    $self->_clear_err();

    my $original_persistent = $self->{persistent};

    my @cmds = ( "admin stop", "admin start" );

    # Make connection persistent so commands can be sent through
    # one connection session 
    $self->options(-persistent => 1); 

    foreach my $cmd (@cmds) {
        @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: QlogicFCSwitch::reboot() " .
                            "Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
    }
  
    # to do the reset, we need to get down to using say
    # since we are being asked questions.
    my $connection = $self->get_connection();
    &Tharn::say($connection, "reset",
                '-prompt', "Please confirm",
                '-timeout', 30);
    eval {
        &Tharn::say($connection, "y", 
                    '-prompt', "Good bye",
                    '-timeout', 60);
    };

    $self->options(-persistent => $original_persistent);

    $self->sanlog(-msg => 
                  "Waiting 30 seconds to allow reboot process to start");

    # sleep to allow boot process to start
    sleep(30);
 
    # Wait until the switch does not respond
    while ($self->is_alive == $SwitchUtil::FALSE && 
           $count < $self->{reboot_timeout}) {
        sleep(1);
        $count++;
    }

    if ($count < $self->{reboot_timeout}) {
        # wait some additional time to allow telnet service to be
        # available and responding. Also wait some time to allow all
        # ports to log back in (or nameserver db will be empty).
        sleep(120);
        $self->reconnect(); 
        if ($self->error) {
            $self->_set_err("Error: BrocadeFCSwitch::reboot() " .
                            "Could not properly reconnect to the switch after reboot");
            return $SwitchUtil::FALSE;
        }

        return $SwitchUtil::TRUE;
    } else {
        $self->_set_err("Error: QlogicFCSwitch::reboot() " .
                        "Could not properly reboot the switch (exceeded allowable reboot/ping time)");
        return $SwitchUtil::FALSE;
    }
}

###############################################################################
# Method:    info
# Objective: Get switch information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my (@output);

    push @output, "Switch name:           " . $self->hostname(), 
                  "Switch type:           " . $self->hosttype(),
                  "Switch FW version:     " . $self->fw_version(), 
                  "Switch Model:          " . $self->model();
    push @output, "\n";

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    ntp_sync
# Objective: Synchronize the clock on the switch with an NTP server.
#            No-op on Qlogic, since the firmware does not support NTP.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($ntp_server_ip, $ntp_server, $ntp_time_zone);
    my ($start, $match, $before, $after);

    $self->_clear_err();

    my $original_persistent = $self->{persistent};

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: QlogicFCSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_server'");
                return $SwitchUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: QlogicFCSwitch::ntp_sync() " .
                                "Invalid value for argument '-ntp_time_zone'");
                return $SwitchUtil::FALSE;
            }
            next;
        }
    }

    # set timezone
    my @cmds = ( "admin stop", "admin start", 
                 "config edit", 
                 "set timezone $ntp_time_zone", 
                 "config save", "admin stop" );

    # Make connection persistent so commands can be sent through
    # one connection session 
    $self->options(-persistent => 1); 

    foreach my $cmd (@cmds) {
        my @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: QlogicFCSwitch::ntp_sync() " .
                            "Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
    }

    # Make connection persistent so commands can be sent through
    # one connection session 
    $self->options(-persistent => 1); 

    # get into admin mode
    @cmds = ( "admin stop", "admin start" );
    foreach my $cmd (@cmds) {
        my @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: QlogicFCSwitch::ntp_sync() " .
                            "Could not execute command $cmd");
            return $SwitchUtil::FALSE;
        }
    }

    # get IP address since switch may not have DNS running
    my (undef, undef, undef, undef, @addrs) = gethostbyname($ntp_server);
    foreach my $ip (@addrs) {
        my ($a, $b, $c, $d) = unpack('C4', $ip);
        $ntp_server_ip = "$a\.$b\.$c\.$d";
    }

    # different interfaces for 7.4+ FW and < 7.4 FW.
    # account for this in command invocation. Convert
    # FW string to integer representation to do 
    # greater than comparisons on.
    my $firmware = $self->get_firmware_version();
    $firmware =~ m/V(\d+)\.(\d+)/i;
    my $major = $1;
    my $minor = $2;
    if (length($minor) == 1) {
        $minor = "0" . $minor;
    }
    $firmware = $major . $minor;
    if ($firmware >= 704) {
        # we need to get down to using say since we are being asked questions.
        my $connection = $self->get_connection();
        $start = 1;
        ($match,$before,$after) = &Tharn::say($connection, "set setup system ntp",
                                              '-prompt', " : ",
                                              '-timeout', 30);
        while ($start == 1) {
            if ($after =~ /NTPClientEnabled/i) {
                ($match,$before,$after) = &Tharn::say($connection, "True",
                                                      '-prompt', " : ",
                                                      '-timeout', 30);
            } elsif ($after =~ /NTPServerDiscovery/i) {
                ($match,$before,$after) = &Tharn::say($connection, "1",
                                                      '-prompt', " : ",
                                                      '-timeout', 30);
            } elsif ($after =~ /NTPServerAddress/i) {
                ($match,$before,$after) = &Tharn::say($connection, $ntp_server_ip,
                                                      '-prompt', "]",
                                                      '-timeout', 30);
            } elsif ($after =~ /activate this system setup/i) {
                ($match,$before,$after) = &Tharn::say($connection, "y",
                                                      '-prompt', "admin",
                                                      '-timeout', 30);
            } elsif ($after =~ /setup saved and activated/i) {
                my @output = $self->execute("admin stop");
                $start = 0;
            } elsif ($after =~ /Illegal or out of range value/i) {
                 $self->_set_err("Error: QlogicFCSwitch::ntp_sync() " .
                                 "Illegal setting of ntp parameters");
                 return $SwitchUtil::FALSE;
            } 
        }
    } else {
        # we need to get down to using say since we are being asked questions.
        my $connection = $self->get_connection();
        $start = 1;
        ($match,$before,$after) = &Tharn::say($connection, "set setup system",
                                              '-prompt', "]",
                                              '-timeout', 30);
        while ($start == 1) {
            if ($after =~ /NTPClientEnabled/i) {
                ($match,$before,$after) = &Tharn::say($connection, "True",
                                                      '-prompt', "]",
                                                      '-timeout', 30);
            } elsif ($after =~ /NTPServerAddress/i) {
                ($match,$before,$after) = &Tharn::say($connection, $ntp_server_ip,
                                                      '-prompt', "]",
                                                      '-timeout', 30);
            } elsif ($after =~ /activate this system setup/i) {
                ($match,$before,$after) = &Tharn::say($connection, "y",
                                                      '-prompt', "admin",
                                                      '-timeout', 30);
            } elsif ($after =~ /setup saved and activated/i) {
                my @output = $self->execute("admin stop");
                $start = 0;
            } elsif ($after =~ /Illegal or out of range value/i) {
                 $self->_set_err("Error: QlogicFCSwitch::ntp_sync() " .
                                 "Illegal setting of ntp parameters");
                 return $SwitchUtil::FALSE;
            } else {
                ($match,$before,$after) = &Tharn::say($connection, " ",
                                                      '-prompt', "]",
                                                      '-timeout', 30);
            }
        }
    }

    $self->options(-persistent => $original_persistent);

    return $SwitchUtil::TRUE;
}

###############################################################################
# Method:    install_firmware
# Objective: Install firmware on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub install_firmware {
    my ($self, %args) = @_;

    #TODO: This subroutine to be completed at a later date
    return $SwitchUtil::FALSE;
}

###############################################################################
# Method:    get_log
# Objective: Retrieve the log file from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_log {
    my ($self) = @_;

    $self->_clear_err();

    my $cmd = "show log";
    my @output = $self->execute($cmd, $self->{log_timeout});
    if ($self->error) {
        $self->_set_err("Error: QlogicFCSwitch::get_log() " .
                        "Could not execute command $cmd");
        return undef;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_config
# Objective: Retrieve the configuration from the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_config {
    my ($self) = @_;
    my (@return);

    $self->_clear_err();

    my $original_persistent = $self->{persistent};

    my @cmds = ( "show config switch", "show config zoning" );

    # Make connection persistent so commands can be sent through
    # one connection session 
    $self->options(-persistent => 1); 

    foreach my $cmd (@cmds) {
        my @output = $self->execute($cmd);
        if ($self->error) {
            $self->_set_err("Error: QlogicFCSwitch::get_config() " .
                            "Could not execute command $cmd");
            return undef;
        }
        push (@return, @output);
    }

    $self->options(-persistent => $original_persistent);

    return (wantarray()) ? @return : \@return;
}

###############################################################################
# Method:    get_state
# Objective: Retrieve the state of the switch (enabled/disabled)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state {
    my ($self) = @_;
    my (@output);
    my $state = "unknown";

    $self->_clear_err();

    # Do not want to grab from cache.
    my $cmd = "show switch";
    @output = $self->execute("show switch");
    if ($self->error) {
        $self->_set_err("Error: QlogicFCSwitch::get_state() " .
                        "Could not execute command $cmd");
        return undef;
    } else {
       $state = (grep(/\s+OperationalState/i, @output))[0];
       $state =~ s/\s+OperationalState\s+//i;
    }

    return $state;
}

###############################################################################
# Method:    get_current_vf_context_id
# Objective: get the current vf context id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_vf_context_id {
    my ($self) = @_;

    # NOOP for Qlogic

    return "";
}

###############################################################################
# Method:    module_reboot
# Objective: Reboot the module of a director switch.
# Details:   Unimplemented.
###############################################################################
sub module_reboot {
    my $self = @_;

    $self->_set_err("This feature is not supported or not implemented for this switch type.");
    return $SwitchUtil::FALSE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new QlogicFC switch object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: QlogicFCSwitch::_new() " .
                                "Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } 
    }

    if (!defined($conn) || ref($conn) ne "Connectrec") {
        $self->_set_err("Error: QlogicFCSwitch::_new() " .
                        "Invalid value for argument '-conn'");
        return $self;
    }

    $self->{conn} = $conn;
    return $self;
}

###############################################################################
# Method:    _get_domain_id
# Objective: Get the domain id of the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_domain_id {
    my ($self) = @_;

    # NO-OP

    return "";
}

1;

=pod

=head1 Package Name

QlogicFCSwitch.pm

=head1 Author

Tom Masterson

=head1 Description

For description of methods see B<San::SanSwitch::Switch.pod>.

=cut

