# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to zone a switch
## @author derricks@netapp.com
## @status private
## @pod here
## @unit_test thpl p4/test/lib/UnitTest/test_San_SanSwitch_ZoneMgr.thpl

package San::SanSwitch::FC::ZoneMgr::BrocadeFCZone;

# Compiler directives.
use strict;

# Module import.
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Alias;
use San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Zone;
use San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::ZoneSet;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog ZoneMgr);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new_alias
# Objective: Create a new alias object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new_alias {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $alias_object = San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Alias->new(-parent => $self, %args);
    if ($alias_object->error) {
        $self->_set_err($alias_object->errmsg());
        return $alias_object;
    }

    # Add the object to the cache
    if (not $self->_alias_cache_exists(-alias => $alias_object)) {
        if (not $self->error) {

            # Object doesn't exist in the cache and there weren't
            # any errors so go ahead and add it to the cache.
            if (not $self->_alias_cache_add(-alias => $alias_object)) {

                # Failed to add it to the cache
                if (not $self->error) {
                    $self->_set_err("Unable to add alias to alias Manager alias cache");
                }
                return $alias_object;
            } ## end if (not $self->_alias_cache_add...
        } ## end if (not $self->error)
    } else {

        # We found an entry in the cache for this object so it must be a duplicate.
        # Set an error and bail.
        $self->_set_err("Duplicate entry found in cache for alias object");
        return $alias_object;
    } ## end else [ if (not $self->_alias_cache_exists...

    # Return the alias object
    return $alias_object;
} ## end sub new_alias

###############################################################################
# Method:    new_zone
# Objective: Create a new zone object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new_zone {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $zone_object = San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Zone->new(-parent => $self, %args);
    if ($zone_object->error) {
        $self->_set_err($zone_object->errmsg());
        return $zone_object;
    }

    # Add the object to the cache
    if (not $self->_zone_cache_exists(-zone => $zone_object)) {
        if (not $self->error) {

            # Object doesn't exist in the cache and there weren't
            # any errors so go ahead and add it to the cache.
            if (not $self->_zone_cache_add(-zone => $zone_object)) {

                # Failed to add it to the cache
                if (not $self->error) {
                    $self->_set_err("Unable to add zone to Zone Manager zone cache");
                }
                return $zone_object;
            } ## end if (not $self->_zone_cache_add...
        } ## end if (not $self->error)
    } else {

        # We found an entry in the cache for this object so it must be a duplicate.
        # Set an error and bail.
        $self->_set_err("Duplicate entry found in cache for zone object");
        return $zone_object;
    } ## end else [ if (not $self->_zone_cache_exists...

    # Return the zone object
    return $zone_object;
} ## end sub new_zone

###############################################################################
# Method:    new_zoneset
# Objective: Create a new zoneset object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new_zoneset {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $zoneset_object = San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::ZoneSet->new(-parent => $self, %args);
    if ($zoneset_object->error) {
        $self->_set_err($zoneset_object->errmsg());
        return $zoneset_object;
    }

    # Add the object to the cache
    if (not $self->_zoneset_cache_exists(-zoneset => $zoneset_object)) {
        if (not $self->error) {

            # Object doesn't exist in the cache and there weren't
            # any errors so go ahead and add it to the cache.
            if (not $self->_zoneset_cache_add(-zoneset => $zoneset_object)) {

                # Failed to add it to the cache
                if (not $self->error) {
                    $self->_set_err("Unable to add zoneset to Zone Manager zoneset cache");
                }
                return $zoneset_object;
            } ## end if (not $self->_zoneset_cache_add...
        } ## end if (not $self->error)
    } else {

        # We found an entry in the cache for this object so it must be a duplicate.
        # Set an error and bail.
        $self->_set_err("Duplicate entry found in cache for zone object");
        return $zoneset_object;
    } ## end else [ if (not $self->_zoneset_cache_exists...

    # Return the zoneset object
    return $zoneset_object;
} ## end sub new_zoneset

###############################################################################
# Method:    alias_discover
# Objective: Discover the aliases on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub alias_discover {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Zero out the aliases reference in the ZoneMgr object
    $self->{aliases} = {};

    # Run alishow on the switch to get a list of all aliases.
    # Format of alishow command is "alishow *,0"
    my $cmd = "alishow *,0";
    my @output = $switch_obj->execute_brocade($cmd, $self->{timeout}, "", -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "does not exist" error
    # If we don't find anything just go ahead and return TRUE
    # since this isn't an error condition.
    if (grep(/\"\S+\" does not exist/, @output) > 0) {
        $self->sanlog(-msg => "No aliases were found on the switch");
        return $SwitchUtil::TRUE;
    }

    # Decode the alias information into a hash reference
    my $aliases_ref = $self->_decode_aliases(@output);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    foreach my $alias (sort keys %${aliases_ref}) {
        if (not defined($aliases_ref->{$alias}->{member}) or $aliases_ref->{$alias}->{member} eq "") {
            $self->_set_err("Alias $alias does not have a member defined");
            return $SwitchUtil::FALSE;
        }
        my $t_aliasobj = $self->new_alias(-alias => $alias, -member => $aliases_ref->{$alias}->{member});
        unless (defined($t_aliasobj) and ref($t_aliasobj) =~ /Alias/i) {
            if (not $self->error) {
                $self->_set_err("Unable to create alias object for alias $alias");
            }
            return $SwitchUtil::FALSE;
        } ## end unless (defined($t_aliasobj...

        # Update the alias_object "on_switch" parameter.
        $t_aliasobj->{on_switch} = $SwitchUtil::TRUE;

        # Add the alias object to the alias cache
        if (not $self->_alias_cache_add(-alias => $t_aliasobj)) {
            if (not $self->error) {
                $self->_set_err("Unable to alias $alias to ZoneMgr alias cache");
            }
            return $SwitchUtil::FALSE;
        } ## end if (not $self->_alias_cache_add...
    } ## end foreach my $alias (sort keys...

    return $SwitchUtil::TRUE;
} ## end sub alias_discover

###############################################################################
# Method:    zone_discover
# Objective: Discover the zones on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_discover {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Zero out the zones reference in the ZoneMgr object
    $self->{zones} = {};

    # Run zoneshow on the switch to get a list of all zones.
    # Format of zoneshow command is "zoneshow *,0"
    my $cmd = "zoneshow *,0";
    my @output = $switch_obj->execute_brocade($cmd, $self->{timeout}, "", -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "does not exist" error
    # If we don't find anything just go ahead and return TRUE
    # since this isn't an error condition.
    if (grep(/\"\S+\" does not exist/, @output) > 0) {
        $self->sanlog(-msg => "No zones were found on the switch");
        return $SwitchUtil::TRUE;
    }

    # Decode the zone information into a hash reference
    my $zones_ref = $self->_decode_zones(@output);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    foreach my $zone (sort keys %${zones_ref}) {

        # Get all of the alias objects for any aliases on the zone
        my @aliases = ();
        foreach my $alias (keys %{$zones_ref->{$zone}}) {
            my $aliasobj = $self->get_aliases(-alias => $alias);
            unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {

                # If we got here then the zone was likely not created
                # with aliases but instead with wwpn or port.  Create
                # a generic alias object based the wwpn or port.
                if ($self->error) {
                    return $SwitchUtil::FALSE;
                } else {
                    unless ($alias =~ /\S+:\S+:\S+:\S+:\S+:\S+:\S+:\S+|\d+,\d+/) {
                        $self->_set_err("Unable to determine member and alias $alias information for zone $zone");
                        return $SwitchUtil::FALSE;
                    }

                    # Create a new alias object
                    $aliasobj = $self->new_alias(-alias => $alias, -member => $alias);
                    unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {
                        if (not $self->error) {
                            $self->_set_err("Unable to create alias object for alias $alias");
                        }
                        return $SwitchUtil::FALSE;
                    } ## end unless (defined($aliasobj)...

                    # Update the on_switch parameter of the alias object
                    $aliasobj->{on_switch} = $SwitchUtil::TRUE;
                } ## end else [ if ($self->error)
            } ## end unless (defined($aliasobj)...
            push(@aliases, $aliasobj);
        } ## end foreach my $alias (keys %{$zones_ref...

        # Create the zone objects
        my $t_zoneobj;
        if (scalar(@aliases) > 0) {
            $t_zoneobj = $self->new_zone(-zone => $zone, -aliases => \@aliases);
        } else {
            $t_zoneobj = $self->new_zone(-zone => $zone);
        }
        unless (defined($t_zoneobj) and ref($t_zoneobj) =~ /Zone/i) {
            if (not $self->error) {
                $self->_set_err("Unable to create zone object for zone $zone");
            }
            return $SwitchUtil::FALSE;
        } ## end unless (defined($t_zoneobj...

        # Update the on_switch parameter of the zone object
        $t_zoneobj->{on_switch} = $SwitchUtil::TRUE;

        # Add the zone object to the zone cache
        if (not $self->_zone_cache_add(-zone => $t_zoneobj)) {
            if (not $self->error) {
                $self->_set_err("Unable to zone $zone to ZoneMgr zone cache");
            }
            return $SwitchUtil::FALSE;
        } ## end if (not $self->_zone_cache_add...
    } ## end foreach my $zone (sort keys...

    return $SwitchUtil::TRUE;
} ## end sub zone_discover

###############################################################################
# Method:    zoneset_discover
# Objective: Discover the zone sets on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zoneset_discover {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Zero out the zones reference in the ZoneMgr object
    $self->{zonesets} = {};

    # Run cfgshow on the switch to get a list of all zonesets.
    # Format of cfgshow command is "cfgshow *,0"
    my $cmd = "cfgshow *,0";
    my @output = $switch_obj->execute_brocade($cmd, $self->{timeout}, "", -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "does not exist" error
    # If we don't find anything just go ahead and return TRUE
    # since this isn't an error condition.
    if (grep(/\"\S+\" does not exist/, @output) > 0) {
        $self->sanlog(-msg => "No zones were found on the switch");
        return $SwitchUtil::TRUE;
    }

    # Decode the zoneset information into a hash reference
    my $zonesets_ref = $self->_decode_zonesets(@output);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    foreach my $zoneset (sort keys %${zonesets_ref}) {

        # Get all of the zone objects for any zones on the zoneset
        my @zones = ();
        foreach my $zone (keys %{$zonesets_ref->{$zoneset}}) {
            my $zoneobj = $self->get_zones(-zone => $zone);
            if (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
                push(@zones, $zoneobj);
            } else {
                if (not $self->error) {
                    $self->_set_err("No zone object was found for zone $zone in the ZoneMgr Zone cache");
                }
                return $SwitchUtil::FALSE;
            } ## end else [ if (defined($zoneobj) ...
        } ## end foreach my $zone (keys %{$zonesets_ref...

        # Create the zoneset objects
        my $t_zonesetobj;
        if (scalar(@zones) > 0) {
            $t_zonesetobj = $self->new_zoneset(-zoneset => $zoneset, -zones => \@zones);
        } else {
            $t_zonesetobj = $self->new_zoneset(-zoneset => $zoneset);
        }

        unless (defined($t_zonesetobj) and ref($t_zonesetobj) =~ /ZoneSet/i) {
            if (not $self->error) {
                $self->_set_err("Unable to create zoneset object for zoneset $zoneset");
            }
            return $SwitchUtil::FALSE;
        } ## end unless (defined($t_zonesetobj...

        # Update the on_switch parameter of the zoneset object
        $t_zonesetobj->{on_switch} = $SwitchUtil::TRUE;

        # Add the zoneset object to the zoneset cache
        if (not $self->_zoneset_cache_add(-zoneset => $t_zonesetobj)) {
            if (not $self->error) {
                $self->_set_err("Unable to zone $zoneset to ZoneMgr zoneset cache");
            }
            return $SwitchUtil::FALSE;
        } ## end if (not $self->_zoneset_cache_add...
    } ## end foreach my $zoneset (sort keys...

    return $SwitchUtil::TRUE;
} ## end sub zoneset_discover

###############################################################################
# Method:    active_zoneset_discover
# Objective: Discover the active zoneset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub active_zoneset_discover {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Run cfgactvshow on the switch to get a list of all zoneshow
    # Format of zoneshow command is "zoneshow *,0"
    my $cmd = "cfgactvshow";
    my @output = $switch_obj->execute_brocade($cmd, $self->{timeout}, "", -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to see if we are in "No Access" mode which is the
    # default mode when you don't have any active zoneset defined.
    # If we are in this mode, just return TRUE since there isn't
    # any objects to add and it isn't an error condition.
    if (grep(/No Access/, @output) > 0) {
        return $SwitchUtil::TRUE;
    }

    foreach (@output) {

        # Remove any leading and trailing whitespace
        s/^\s//g;
        s/\s$//g;

        # Grab the line that starts with "cfg:" and get the zoneset name
        if (/^cfg:\s+(\S+)/) {

            my $zoneset = $1;

            # Get the object for the active zoneset
            my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);
            if ($self->error) {
                return $SwitchUtil::FALSE;
            }
            if (defined($zonesetobj)) {
                if (ref($zonesetobj) =~ /ZoneSet/i) {

                    # Set the active_zoneset on the zonemgr object to $zonsetobj
                    $self->{active_zoneset} = $zonesetobj;
                    return $SwitchUtil::TRUE;
                } elsif (scalar(keys %$zonesetobj) == 0) {

                    # Looks like there isn't an object for this
                    # zoneset so we need to create one.
                    my $t_zonesetobj = $self->new_zoneset(-zoneset => $zoneset);
                    if (not defined($t_zonesetobj) and ref($t_zonesetobj) !~ /ZoneSet/i) {
                        if (not $self->error) {
                            $self->_set_err("Unable to create zoneset object for zoneset $zoneset");
                        }
                        return $SwitchUtil::FALSE;
                    } else {
                        $self->{active_zoneset} = $t_zonesetobj;
                        return $SwitchUtil::TRUE;
                    }
                } else {

                    # We got back something totally unexpected
                    $self->_set_err("Unable to determine the object for zoneset $zoneset");
                    return $SwitchUtil::FALSE;
                } ## end else [ if (ref($zonesetobj) =~...
            } else {
                $self->_set_err("Unable to determine the object for zoneset $zoneset");
                return $SwitchUtil::FALSE;
            }

        } ## end if (/^cfg:\s+(\S+)/)
    } ## end foreach (@output)

    # if we made it this far then we didn't find the line with "cfg:" so set
    # an error and return false.
    $self->_set_err("Unable to find active zoneset information in the output of commend ($cmd)");
    return $SwitchUtil::FALSE;
} ## end sub active_zoneset_discover

###############################################################################
# Method:    config_enable
# Objective: enable the zoneset config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_enable {
    my ($self, %args) = @_;

    my ($zoneset);

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed in something
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    # Get the zoneset object
    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);

    unless (defined($zonesetobj) and ref($zonesetobj) =~ /ZoneSet/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($zonesetobj...

    # Run cfgenable on the switch to enable the zoneset
    # Format of cfgenable command is "cfgenable zoneset";
    my $cmd        = "cfgenable $zoneset";
    my $cmd_prefix = "echo y\|";
    my @output     = $switch_obj->execute_brocade($cmd, $self->{timeout}, $cmd_prefix);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Update the zonemgr object's active zoneset definition.
    $self->{active_zoneset} = $zonesetobj;

    return $SwitchUtil::TRUE;
} ## end sub config_enable

###############################################################################
# Method:    config_disable
# Objective: disable the active config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_disable {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Run cfgdisable on the switch to disable the active config
    # Format of cfgdisable command is "cfgdisable";
    my $cmd        = "cfgdisable";
    my $cmd_prefix = "echo y\|";
    my @output     = $switch_obj->execute_brocade($cmd, $self->{timeout}, $cmd_prefix);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub config_disable

###############################################################################
# Method:    config_save
# Objective: save the switch config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_save {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Run cfgsave on the switch to save switch config
    # Format of cfgsave command is "cfgsave";
    my $cmd        = "cfgsave";
    my $cmd_prefix = "echo y\|";
    my @output     = $switch_obj->execute_brocade($cmd, $self->{timeout}, $cmd_prefix);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub config_save

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $timeout, $vsan);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?timeout/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-timeout'");
                return $self;
            }
            next;
        } elsif (/-?vsan/) {
            if (not length($vsan = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vsan'");
                return $self;
            }
            next;
        } ## end elsif (/-?vsan/)
    } ## end foreach (keys %args)

    # check for existance of required -parent argument
    if (defined($parent)) {
        unless (ref($parent) =~ /SanSwitch/i) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # Check for existance of -timeout arguement.  Default to 120 if not found.
    unless (defined($timeout) and $timeout ne "") {
        $timeout = 240;
    }

    # Check for existance of -vsan arguement.  Default to undef if not found.
    unless (defined($vsan) and $vsan ne "") {
        $vsan = undef;
    }

    # initialize the object members
    $self->{parent}         = $parent;
    $self->{aliases}        = {};
    $self->{zones}          = {};
    $self->{zonesets}       = {};
    $self->{timeout}        = $timeout;
    $self->{vsan}           = $vsan;
    $self->{active_zoneset} = undef;

    return $self;
} ## end sub _new

###############################################################################
# Method:    _decode_aliases
# Objective: Decode switch alias output into a hash reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _decode_aliases {
    my ($self, @input) = @_;

    $self->_clear_err();

    # Initialize the hash reference;
    my $alias_ref = {};

    my $alias;
    foreach (@input) {
        chomp();

        # Skip any blank lines or lines that say "alishow"
        next if (/^\s*$|alishow/);

        # Remove all leading and trailing whitespace
        s/^\s+//g;
        s/\s+$//g;

        # Once we have removed the whitespaces, we should be left with something
        # like the following:
        #
        # alias: sunv215_shu02_qlc0
        # 21:00:00:e0:8b:88:01:8d
        #
        # OR
        #
        # alias: sunv215_shu02_qlc0
        # 21:00:00:e0:8b:88:01:8d
        #
        # OR
        # 
        # alias: fas3070_shu02_1a 105,6
        # 
        # OR
        #
        # alias: fas3070_shu02_1a
        # 105,6

        if (/^alias:\s+(\S+)\s+(\d+\,\d+|\S+\:\S+\:\S+\:\S+\:\S+\:\S+\:\S+\:\S+)/) {
            # We found both the alias and member on the same line
            $alias_ref->{$1}->{member} = $2;
        } elsif (/^alias:\s+(\S+)/) {
            # Found the alias name
            $alias = $1;
        } elsif (/^(\d+\,\d+|\S+\:\S+\:\S+\:\S+\:\S+\:\S+\:\S+\:\S+)/) {
            # Found the member information
            my $member = $1;

            # Check to make sure that $alias is defined.  if it isn't, then we have
            # an issue with the data provided so set an error and bail out.
            unless (defined($alias) and $alias ne "") {
                $self->_set_err("Unable to determine alias name from line:\n" .
                                $_ . "\nInput data being parsed contained:\n",
                                @input);
                return {};
            }
            # If we found either a did,port or wwpn then we assume that is the
            # member for the alias that was sent.
            $alias_ref->{$alias}->{member} = $member;
        } ## end elsif (/^(\d+\,\d+|\S+\:\S+\:\S+\:\S+\:\S+\:\S+\:\S+\:\S+)/)
    } ## end foreach (@input)

    return $alias_ref;
} ## end sub _decode_aliases

###############################################################################
# Method:    _decode_zones
# Objective: Decode switch zoneshow output into a hash reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _decode_zones {
    my ($self, @input) = @_;

    $self->_clear_err();

    # Initialize the hash reference;
    my $zone_ref = {};

    my $zone;
    foreach (@input) {
        chomp();

        # Skip any blank lines
        next if (/^\s*$/);

        # Skip any line that says "zoneshow"
        next if (/zoneshow/);

        # Remove any leading and trailing whitespace
        s/^\s+//g;
        s/\s+$//g;

        # Once we have removed the whitespaces, we should be left with something
        # like the following:
        #
        # zone: zone_name
        #
        # OR
        #
        # zone: zone_name
        # alias1; alias2; alias3;
        # alias4; alias5; alias6
        #
        # OR
        #
        # zone: zone_name
        # alias1; alias2;
        #
        # OR
        #
        # zone: zone_name
        # alias1
        #
        if (/^zone:\s+(\S+)/) {
            $zone = $1;
            $zone_ref->{$zone} = {};
        } else {
            # Check to make sure that $zone and $zone_ref->{$zone} are defined.  If they
            # aren't, then we have an issue with the data provided so set an error and bail out.
            unless (defined($zone) and defined($zone_ref->{$zone})) {
                $self->_set_err("Unable to determine zone name from line:\n" .
                                $_ . "\nInput data being parsed contained:\n",
                                @input);
                return {};
            }

            # Remove all whitespace, then split on the semicolon.
            s/\s+//g;
            my @aliases = split(/\;/);

            # Loop through @aliases and create a hash entry in the form of $zone_ref->{$zone}->{$alias} = $SwitchUtil::TRUE;
            foreach my $alias (@aliases) {
                $zone_ref->{$zone}->{$alias} = $SwitchUtil::TRUE;
            }
        } ## end else [ if (/^zone:\s+(\S+)/)
    } ## end foreach (@input)

    return $zone_ref;
} ## end sub _decode_zones

###############################################################################
# Method:    _decode_zonesets
# Objective: Decode switch cfgshow output into a hash reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _decode_zonesets {
    my ($self, @input) = @_;

    $self->_clear_err();

    # Initialize the hash reference;
    my $zoneset_ref = {};

    my $zoneset;
    foreach (@input) {
        chomp();

        # Skip any blank lines
        next if (/^\s*$/);

        # Skip any line that says "cfgshow"
        next if (/cfgshow/);

        # Remove any leading and trailing whitespace
        s/^\s+//g;
        s/\s+$//g;

        # Once we have removed the whitespaces, we should be left with something
        # like the following:
        #
        # cfg: shared zone1; zone2; zone3;
        # zone4; zone5; zone6;
        # zone7; zone8
        #
        # OR
        #
        # cfg: shared
        # zone1;
        # zone2;
        # zone3
        #
        if (/^cfg:\s+(\S+)\s+(\S.*)/) {

            # We have both the zoneset name and zones on the same line
            # as in cfg: zoneset zone1; zone2
            $zoneset = $1;
            my $zonelist = $2;

            # Remove any whitespace from $zonelist and then split it out.
            $zonelist =~ s/\s+//g;
            my @zones = split(/\;/, $zonelist);

            # Set the  value for the $zone_ref->{$zoneset} hash
            foreach my $zone (@zones) {
                $zoneset_ref->{$zoneset}->{$zone} = $SwitchUtil::TRUE;
            }
        } elsif (/^cfg:\s+(\S+)$/) {

            # We have only the zoneset on the line so we will initialize
            # the zoneset_ref->{$zoneset} hash with an anonymous hash.
            $zoneset = $1;
            $zoneset_ref->{$zoneset} = {};
        } else {

            # Everything else should be zones belonging to the previous
            # zoneset that was found.  We need to make sure we have $zoneset
            # and $zoneset_ref->{$zoneset} defined.  If we don't, something
            # went wrong so set an error and bail.
            unless ($zoneset and $zoneset_ref->{$zoneset}) {
                $self->_set_err("Unable to determine zoneset name");
                return {};
            }

            # Remove all whitespace, then split on the semicolon.
            s/\s+//g;
            my @zones = split(/\;/);

            # Set the  value for the $zone_ref->{$zoneset} hash
            foreach my $zone (@zones) {
                $zoneset_ref->{$zoneset}->{$zone} = $SwitchUtil::TRUE;
            }
        } ## end else [ if (/^cfg:\s+(\S+)\s+(\S.*)/)
    } ## end foreach (@input)

    # Return the zoneset hash reference.  Format of the hash reference will be
    # $zoneset_ref->{$zoneset}->{$zone} = $SwitchUtil::TRUE;
    return $zoneset_ref;
} ## end sub _decode_zonesets

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

San::SanSwitch::FC::ZoneMgr::BrocadeFCZone.pm

=head1 Author

Derrick Shultz

=head1 Description

Provides an abstract API interface to Alias, Zone, and ZoneSet operations on Brocade switches.

=head1 See Also

Package inherits from the following modules. See modules for interface detail

=over 4

=item * B<San::SanSwitch::ZoneMgr.pm>

=item * B<San::SanLog.pm>

=item * B<Error.pm>

=back

=head1 Attributes

=head2 B<active_zoneset>

Object of the active zoneset.  Defaults to undef.

=head2 B<aliases>

Reference to a hash of alias objects with a key of the alias name and the value of the alias object.  Defaults to {}.

=head2 B<parent>

ZoneMgr object = Reference to a parent SanSwitch object

=head2 B<timeout>

String containing an optional command timeout.  Default is 120.

=head2 B<vsan>

String containing an optional vsan id.

=head2 B<zones>

Reference to a hash of zone objects with a key of the zone name and the value of the zone object.  Defaults to {}.

=head2 B<zonesets>

Reference to a hash of zoneset objects with a key of the zoneset name and the value of the zoneset object.  Defaults to {}.

=head1 Public Methods

=head2 B<active_zoneset_discover()>

Performs discovery of the active zoneset on the switch and populates the $self->{active_zoneset} attribute
with the associated zoneset object.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->active_zoneset_discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to discover the active zoneset");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<alias_discover()>

Performs alias discovery of all aliases on the switch and creates objects for each discovered alias.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->alias_discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to perform alias discovery");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<config_disable()>

Disables active Config/Zoneset on the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->config_disable();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to disable the active zoneset");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<config_enable()>

Enables a config/zoneset on the switch

=head3 Inputs

=over 4

=item B<-zoneset>

String containing the name of the zoneset to enable

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->config_enable(-zoneset => $zoneset);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to enable zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<config_save()>

Saves the current config

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->config_save();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to save the configuration");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<new_alias()>

Creates a new alias object by calling San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Alias->new() and adds
it to the ZoneMgr object alias cache.  Blindly passes all args through to
San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Alias->new()

=head3 Inputs

See San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Alias->new()

=head3 Outputs

Returns an alias object. If an error occurs, the error flag is set on $self and errmsg is
set to the failure prior to returning the object.

=head3 Examples

    my $aliasobj = $self->new_alias(-alias => $alias_name, -member => $alias_member);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<new_zone()>

Creates a new zone object by calling San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Zone->new() and adds
it to the ZoneMgr object zone cache.  Blindly passes all args through to
San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Zone->new()

=head3 Inputs

See San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Zone->new()

=head3 Outputs

Returns a zone object. If an error occurs, the error flag is set on $self and errmsg is
set to the failure prior to returning the object.

=head3 Examples

    my $zoneobj = $self->new_zone(-zone => $zone_name, -aliases => \@aliases);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<new_zoneset()>

Creates a new zoneset object by calling San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::ZoneSet->new() and adds
it to the ZoneMgr object zone cache.  Blindly passes all args through to
San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::ZoneSet->new()

=head3 Inputs

See San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::ZoneSet->new()

=head3 Outputs

Returns a zoneset object. If an error occurs, the error flag is set on $self and errmsg is
set to the failure prior to returning the object.

=head3 Examples

    my $zonesetobj = $self->new_zoneset(-zoneset => $zoneset_name, -zones => \@zones);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zone_discover()>

Performs zone discovery of all zones on the switch and creates objects for each discovered zone.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zone_discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to perform zone discovery");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zoneset_discover()>

Performs zoneset discovery of all zonesets on the switch and creates objects for each discovered zoneset.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zoneset_discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to perform zoneset discovery");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head1 Private Methods

=head2 B<_decode_aliases()>

Take output from alishow and decode it into a hash of aliases keyed by the alias name with a value of
$Switchutil::TRUE.

=head3 Inputs

An array containing output from the alishow command.

=head3 Outputs

On success, returns a hash reference of aliases keyed by the alias name with a value of
$Switchutil::TRUE.  If an error occurs, the error flag is set, errmsg is
set to the failure reason and {} is returned.  If there isn't an error but there aren't
any aliases found in the output, then {} is returned.

=head3 Examples

    my $cmd = "alishow *,0";
    my @output = $switch_obj->brocade_execute($cmd);
    my $alias_ref = $self->_decode_aliases(@output);
    if ($switch_obj->error) {
        $self->_set_err("There was an issue trying to decode the aliases:\n" . $switch_obj->errmsg);
        return $SwitchUtil::FALSE;
    }

=head2 B<_decode_zones()>

Take output from zoneshow and decode it into a hash of zones keyed by the zone name with a value of
a reference to a hash of aliases with a key of the alias name and the value of $Switchutil::TRUE.

=head3 Inputs

An array containing output from the zoneshow command.

=head3 Outputs

On success, returns a reference to a hash of zones keyed by the zone name with a value of
a reference to a hash of aliases with a key of the alias name and the value of $Switchutil::TRUE.
If an error occurs, the error flag is set, errmsg is set to the failure reason and {} is returned.
If there isn't an error but there aren't any zones found in the output, then {} is returned.

=head3 Examples

    my $cmd = "zoneshow *,0";
    my @output = $switch_obj->brocade_execute($cmd);
    my $zone_ref = $self->_decode_zones(@output);
    if ($switch_obj->error) {
        $self->_set_err("There was an issue trying to decode the zones:\n" . $switch_obj->errmsg);
        return $SwitchUtil::FALSE;
    }

=head2 B<_decode_zonesets()>

Take output from cfgshow and decode it into a hash of zonesets keyed by the zoneset name with a value of
a reference to a hash of zones with a key of the zone name and the value of $Switchutil::TRUE.

=head3 Inputs

An array containing output from the cfgshow command.

=head3 Outputs

On success, returns a reference to a hash of zonesets keyed by the zoneset name with a value of
a reference to a hash of zones with a key of the zone name and the value of $Switchutil::TRUE.
If an error occurs, the error flag is set, errmsg is set to the failure reason and {} is returned.
If there isn't an error but there aren't any zonesets found in the output, then {} is returned.

=head3 Examples

    my $cmd = "cfgshow *,0";
    my @output = $switch_obj->brocade_execute($cmd);
    my $zoneset_ref = $self->_decode_zonesets(@output);
    if ($switch_obj->error) {
        $self->_set_err("There was an issue trying to decode the zonesets:\n" . $switch_obj->errmsg);
        return $SwitchUtil::FALSE;
    }

=head2 B<_new()>

Instantiates the Brocade ZoneMgr object.

=head3 Inputs

=over 4

=item B<-parent>

Parent Switch object.

=item B<-timeout (optional)>

String containing the timeout to use for commands.  Defaults to 120.

=item B<-vsan (optional)>

String containing the vsan ID.  Defaults to undef.

=back

=head3 Outputs

Returns a ZoneMgr object. If an error occurs, the error flag is set and errmsg is
set to the failure reason before the object is returned.

=head3 Examples

    my $object = San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::_new(-parent => $self,
                                                                  -timeout => "60");
    if ($object->error) {
        $self->_set_err($object->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=cut
