# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to alias related switch commands
## @author derricks@netapp.com
## @status private
## @pod here
## @unit_test thpl p4/test/lib/UnitTest/test_San_SanSwitch_ZoneMgr.thpl

package San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Alias;

# Compiler directives.
use strict;

# Module import.
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    create
# Objective: create the alias on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the alias object has a member defined since we can't
    # create an alias on the switch without that information.
    my $member = $self->get_member();
    if (not defined($member) or $member eq "") {
        $self->_set_err("Alias object does not currently have a member defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Run alicreate on the switch with the alias name and member.
    # Format of alicreate is as follows:
    # alicreate "alias", "member"
    my $cmd = "alicreate \"" . ${alias} . "\", \"" . $member . "\"";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub create

###############################################################################
# Method:    destroy
# Objective: destroy the alias on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Run alidelete on the switch with the alias name and member.
    # Format of alidelete is as follows:
    # alidelete "alias"
    my $cmd = "alidelete \"" . ${alias} . "\"";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub destroy

###############################################################################
# Method:    remove_member
# Objective: remove an alias member from the alias
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_member {
    my ($self, $member) = @_;

    $self->_clear_err();

    if (!$member) {
        $self->_set_err("Missing required parameter member");
        return $SwitchUtil::FALSE;
    }

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Run aliremove on the switch with the alias name and member.
    # Format of alidelete is as follows:
    # aliremove "alias","wwpn"
    my $cmd = "aliremove \"" . ${alias} . "\",\"" . $member . "\"";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # clear the member field (XXX we really need to maintain an array of members here)
    if ($self->{member} and $self->{member} eq $member) {
        delete($self->{member});
    }

    return $SwitchUtil::TRUE;
} ## end sub remove_member

###############################################################################
# Method:    add_member
# Objective: add an alias member to the alias
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_member {
    my ($self, $member) = @_;

    $self->_clear_err();

    if (!$member) {
        $self->_set_err("Missing required parameter member");
        return $SwitchUtil::FALSE;
    }

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Run aliadd on the switch with the alias name and member.
    # Format of alidelete is as follows:
    # aliadd "alias","wwpn"
    my $cmd = "aliadd \"" . ${alias} . "\",\"" . $member . "\"";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # update the member field
    # XXX we really need to maintain an array of members here
    $self->{member} = $member;

    return $SwitchUtil::TRUE;
} ## end sub add_member

###############################################################################
# Method:    validate
# Objective: validate the alias on the switch against the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the current alias object actually has a member defined
    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the current alias object actually has a member defined
    my $member = $self->get_member();
    if (not defined($member) or $member eq "") {
        $self->_set_err("Alias object does not currently have a member defined");
        return $SwitchUtil::FALSE;
    }

    # Get the alias information from the switch
    my @output = $self->_query_alias();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "doesn't exist" error
    if (grep(/\"\S+\" does not exist/, @output) > 0) {
        $self->_set_err("Alias $alias does not exist on the switch");
        return $SwitchUtil::FALSE;
    }

    # Decode the alias information into a hash reference
    my $alias_ref = $self->{parent}->_decode_aliases(@output);
    if ($self->{parent}->error()) {
        $self->_set_err("Failed to decode alias output", $self->{parent}->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we actually got back a hash
    if (ref($alias_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return $SwitchUtil::FALSE;
    }

    # Check to see if the switch member is the same as the defined member
    if (defined($alias_ref->{$alias})) {
        if (defined($alias_ref->{$alias}->{member}) and $alias_ref->{$alias}->{member} ne "") {

            # The alias we are looking for was found in the hash so check
            # the member against the one on the object
            if ($member eq $alias_ref->{$alias}->{member}) {
                return $SwitchUtil::TRUE;
            } else {
                return $SwitchUtil::FALSE;
            }
        } else {
            $self->_set_err("Member information for alias $alias was blank");
            return $SwitchUtil::FALSE;
        }
    } else {
        # the alias wasn't found on the switch hash so return false
        return $SwitchUtil::FALSE;
    }
} ## end sub validate

###############################################################################
# Method:    exists
# Objective: Check whether the alias exists on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub exists {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Get the alias information from the switch
    my @output = $self->_query_alias();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "doesn't exist" error
    if (grep(/\"\S+\" does not exist/, @output) > 0) {
        return $SwitchUtil::FALSE;
    }

    # Decode the alias information into a hash reference
    my $alias_ref = $self->{parent}->_decode_aliases(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Failed to decode aliases:\n", $self->{parent}->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we actually got back a hash
    if (ref($alias_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return $SwitchUtil::FALSE;
    }

    # Double check that the alias you are looking for is found as a
    # key to the hash reference
    if (defined($alias_ref->{$alias})) {
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
} ## end sub exists

###############################################################################
# Method:    get_member
# Objective: Return the member information from the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_member {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{member})) {
        $self->_set_err("Unable to retrieve the 'member' key off of the object");
        return "";
    }

    return $self->{member};
}

###############################################################################
# Method:    get_type
# Objective: Return the type information from the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_type {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{type})) {
        $self->_set_err("Unable to retrieve the 'type' key off of the object");
        return "";
    }

    return $self->{type};
}

###############################################################################
# Method:    get_name
# Objective: Return the alias name from the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{alias})) {
        $self->_set_err("Unable to retrieve the 'alias' key off of the object");
        return "";
    }
    return $self->{alias};
}

###############################################################################
# Method:    discover
# Objective: Discover the member information for the alias
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return (undef, undef);
    }

    # Get the alias information from the switch
    my @output = $self->_query_alias();
    if ($self->error) {
        return (undef, undef);
    }

    # Check to make sure we didn't get a "doesn't exist" error
    if (grep(/\"\S+\" does not exist/, @output) > 0) {
        $self->_set_err("Alias $alias does not exist on the switch");
        return (undef, undef);
    }

    # Decode the alias information into a hash reference
    my $alias_ref = $self->{parent}->_decode_aliases(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Failed to decode aliases:\n", $self->{parent}->errmsg());
        return (undef, undef);
    }

    # Check to make sure we actually got back a hash
    if (ref($alias_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return (undef, undef);
    }

    # Check to see if the alias exists and the member info is populated
    if (defined($alias_ref->{$alias})) {
        if (not defined($alias_ref->{$alias}->{member}) or $alias_ref->{$alias}->{member} eq "") {
            $self->_set_err("Error discovering alias.  Member information is blank");
            return (undef, undef);
        }
    } else {
        $self->_set_err("Alias $alias was not found on the switch");
        return (undef, undef);
    }

    return ($alias_ref->{$alias}->{member}, undef);
} ## end sub discover

###############################################################################
# Method:    is_on_switch
# Objective: returns the "on_switch" value off of the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_on_switch {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{on_switch}) and $self->{on_switch} ne "") {
        $self->_set_err("Invalid value for 'on_switch' key on object");
        return $SwitchUtil::FALSE;
    }

    return $self->{on_switch};
}

###############################################################################
# Method:    new
# Objective: Create new alias object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, $parent, $alias, $member);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?alias/) {
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '-alias'");
                return $self;
            }
            next;
        } elsif (/-?member/) {
            if (not length($member = $args{$_})) {
                $self->_set_err("Invalid value for argument '-member'");
                return $self;
            }
            next;
        } ## end elsif (/-?member/)
    } ## end foreach (keys %args)

    # Verify mandatory parameters.
    if (defined($parent) and $parent ne "") {
        unless (ref($parent) =~ /BrocadeFCZone/i) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Mandatory parameter \"alias\"", "not specified");
        return $self;
    }

    # Verify the optional parameters
    if (not defined($member)) {
        $member = "";
    }

    # initialize the object members
    $self->{parent}    = $parent;
    $self->{alias}     = $alias;
    $self->{member}    = $member;
    $self->{type}      = undef;
    $self->{on_switch} = $SwitchUtil::FALSE;

    return $self;
} ## end sub new

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _query_alias
# Objective: Query the switch and return the alias information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _query_alias {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return ();
    }

    my $switch_obj = $self->{parent}->{parent};

    # Run "alishow" on the switch with the alias name
    # Format of command is "alishow alias_name,0"
    my $cmd = "alishow " . ${alias} . ",0";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return ();
    }

    return @output;
} ## end sub _query_alias

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Alias.pm

=head1 Author

Derrick Shultz

=head1 Description

Provides an abstract API interface to Alias operations on Brocade switches.

=head1 See Also

Package inherits from the following modules. See modules for interface detail

=over 4

=item * B<San::SanLog.pm>

=item * B<Error.pm>

=back

=head1 Attributes

=head2 B<alias>

String containing name of alias

=head2 B<member>

String containing alias member information

=head2 B<parent>

ZoneMgr object = Reference to a parent ZoneMgr object

=head2 B<on_switch>

Boolean of $SwitchUtil::TRUE or $SwitchUtil::FALSE representing whether or not the alias is currently found on
the switch.

=head2 B<type>

String containing alias type information.  For Brocade switches, this defaults to undef since there is no concept
of an alias type on Brocade.

=head1 Public Methods

=head2 B<create()>

Creates the alias on the switch.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->create();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<destroy()>

Removes the alias from the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->destroy();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<remove_member()>

Remove a member from this alias

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->remove_member($wwpn);
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<add_member()>

Add a member to this alias

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->add_member($wwpn);
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<discover()>

Discovers the member and type information for the alias on the switch

=head3 Inputs

None

=head3 Outputs

Returns ($member, undef) on success.  If an error occurs, the error flag is set and errmsg is
set to the failure reason and (undef, undef) is returned.  Since Brocade switches do not have
any concept of alias type, "undef" is always returned in the type field.

=head3 Examples

    my ($member, $type) = $aliasobj->discover();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    }

=head2 B<exists()>

Checks the existance of the alias on the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned. If there wasn't an error, but the alias
doesn't actually exist it returns $SwitchUtil::FALSE without setting an error.

=head3 Examples

    my $status = $aliasobj->exists();
    if ($status == $SwitchUtil::FALSE) {
        if ($aliasobj->error) {
            $self->_set_err($aliasobj->errmsg());
            return $SwitchUtil::FALSE;
        } else {
            return $SwitchUtil::FALSE;
        }
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<get_member()>

Returns the member string from the alias object.

=head3 Inputs

None

=head3 Outputs

Returns the member string from the alias object.

=head3 Examples

    my $member = $aliasobj->get_member();

=head2 B<get_name()>

Returns the name string from the alias object.

=head3 Inputs

None

=head3 Outputs

Returns the name string from the alias object.

=head3 Examples

    my $name = $aliasobj->get_name();

=head2 B<get_type()>

Returns the type string from the alias object.

=head3 Inputs

None

=head3 Outputs

Returns the type string from the alias object.

=head3 Examples

    my $type = $aliasobj->get_type();

=head2 B<is_on_switch()>

Method to return the on_switch key from the object

=head3 Inputs

None

=head3 Outputs

Returns a boolean based on the value of the on_switch key. If the key doesn't exist or is empty, it sets an error.

=head3 Examples

    if ($obj->is_on_switch()) {
        do stuff here;
    }

=head2 B<new()>

Instantiates the alias object.

=head3 Inputs

=over 4

=item B<-alias>

String containing the name of the alias.

=item B<-member (optional)>

String containing the member information.

=item B<-parent>

Parent ZoneMgr object

=back

=head3 Outputs

Returns an alias object. If an error occurs, the error flag is set and errmsg is
set to the failure reason before the object is returned.

=head3 Examples

    my $aliasobj = San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::Alias::new(-parent => $self,
                                                                          -alias => $alias_name,
                                                                          -member => $member);
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<validate()>

Validates the alias on the switch against the object

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->validate();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head1 Private Methods

=head2 B<_query_alias()>

Runs "alishow <alias>,0" on the switch and returns the output from the command.

=head3 Inputs

None

=head3 Outputs

On success, returns an array containing the "alishow" output. If an error occurs, the error is
set to the failure reason and () is returned.

=head3 Examples

    my @query_output = $aliasobj->_query_zone();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    }

=cut
