# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to zoneset related switch commands
## @author derricks@netapp.com
## @status private
## @pod here
## @unit_test thpl p4/test/lib/UnitTest/test_San_SanSwitch_ZoneMgr.thpl

package San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::ZoneSet;

# Compiler directives.
use strict;

# Module import.
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    create
# Objective: create the zoneset on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zoneset object has a defined name
    my $zoneset = $self->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Zoneset object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Check to make sure the object has zones defined
    my $zones_ref = $self->get_zones();
    unless (scalar(keys %$zones_ref) > 0) {
        $self->_set_err("Zoneset object does not currently have any zones defined");
        return $SwitchUtil::FALSE;
    }

    # We are going to use a temp array to pull the zone names out of the
    # zone objects.
    my @zones = ();
    foreach my $zone (keys %$zones_ref) {
        push(@zones, $zone);
    }

    # Join @zones into a string separated by ";";
    my $zones = join("\;", @zones);

    # Run cfgcreate on the switch with the zoneset name and zones.
    # Format of cfgcreate is as follows:
    # cfgcreate "zoneset", "zone[;zone;zone]"
    my $cmd = "cfgcreate \"" . ${zoneset} . "\", \"" . $zones . "\"";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub create

###############################################################################
# Method:    destroy
# Objective: destroy the zoneset on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zoneset object has a defined name
    my $zoneset = $self->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Zoneset object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Run cfgdelete on the switch with the zoneset name.
    # Format of cfgdelete is as follows:
    # cfgdelete "zoneset"
    my $cmd = "cfgdelete \"" . ${zoneset} . "\"";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub destroy

###############################################################################
# Method:    validate
# Objective: validate the zoneset on the switch against the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zoneset object has a defined name
    my $zoneset = $self->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Zoneset object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Get the zoneset information from the switch
    my @output = $self->_query_zoneset();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "does not exist" error
    if (grep(/does not exist/, @output) > 0) {
        $self->_set_err("Zoneset $zoneset does not exist on the switch");
        return $SwitchUtil::FALSE;
    }

    # Decode the switch output into a hash reference
    my $switch_zoneset_ref = $self->{parent}->_decode_zonesets(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to decode zonesets:\n", $self->{parent}->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Double check that the zoneset we are looking for is found as a
    # key to the hash reference
    if (not defined($switch_zoneset_ref->{$zoneset})) {
        $self->_set_err("Zoneset information returned from the switch does not match the object Zoneset name");
        return $SwitchUtil::FALSE;
    }

    my $zonesetobj_ref = $self->get_zones();

    # compare the keys in the $zonesetobj_ref and $switch_zoneset_ref->{$zoneset} hashes and store
    # any missing zones in @missing_zones array.
    my @missing_zones = ();
    foreach my $zone (sort keys %$zonesetobj_ref) {
        my $found = undef;
        foreach my $switch_zone (sort keys %{$switch_zoneset_ref->{$zoneset}}) {
            if ($switch_zone eq $zone) {
                $found = $SwitchUtil::TRUE;
                last;
            }
        } ## end foreach my $switch_zone (sort...
        if (not defined($found)) {

            # We didn't find $zone so save that info.
            push(@missing_zones, $zone);
        }
    } ## end foreach my $zone (sort keys...
    if (scalar(@missing_zones) > 0) {
        $self->_set_err("Unable to validate zoneset $zoneset. The switch zoneset is missing zones @missing_zones.");
        return $SwitchUtil::FALSE;
    }

    # Do a check in the other direction.  Check to make sure we don't have extra zones
    # defined in the zoneset on the switch versus what is defined on the object
    my @extra_zones = ();
    foreach my $switch_zone (sort keys %{$switch_zoneset_ref->{$zoneset}}) {
        my $found = undef;
        foreach my $zone (sort keys %$zonesetobj_ref) {
            if ($switch_zone eq $zone) {
                $found = $SwitchUtil::TRUE;
                last;
            }
        } ## end foreach my $zone (sort keys...
        if (not defined($found)) {

            # We didn't find $zone so save that info.
            push(@extra_zones, $switch_zone);
        }
    } ## end foreach my $switch_zone (sort...
    if (scalar(@extra_zones) > 0) {
        $self->_set_err(
            "Unable to validate zoneset $zoneset. The switch zoneset contains the extra zones @extra_zones.");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub validate

###############################################################################
# Method:    exists
# Objective: Check whether the zoneset exists on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub exists {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zoneset object has a defined name
    my $zoneset = $self->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Zoneset object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Get the zoneset information from the switch
    my @output = $self->_query_zoneset();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "does not exist" error
    if (grep(/does not exist/, @output) > 0) {
        return $SwitchUtil::FALSE;
    }

    # Decode the zone information into a hash reference
    my $zoneset_ref = $self->{parent}->_decode_zonesets(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to decode zonesets:\n", $self->{parent}->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Double check that the zone you are looking for is found as a
    # key to the hash reference
    unless (defined($zoneset_ref->{$zoneset})) {
        $self->_set_err("Zoneset information returned from the switch does not match the object Zoneset name");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub exists

###############################################################################
# Method:    get_zones
# Objective: Return the zone information from the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_zones {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{zones})) {
        $self->_set_err("Unable to retrieve the 'zones' key off of the object");
        return {};
    }

    return $self->{zones};
}

###############################################################################
# Method:    get_name
# Objective: Return the zoneset name from the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{zoneset})) {
        $self->_set_err("Unable to retrieve the 'zoneset' key off of the object");
        return "";
    }

    return $self->{zoneset};
}

###############################################################################
# Method:    add_zone
# Objective: Add a zone to the zoneset on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_zone {
    my ($self, %args) = @_;

    my ($zoneobj);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed a zone object
            if (not length($zoneobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($zoneobj)) {
        unless (ref($zoneobj) =~ /Zone/) {
            $self->_set_err("Invalid value for argument '-zone'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    # Verify that the zone object has a defined name
    my $zone = $zoneobj->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object provided does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the zoneset object has a defined name
    my $zoneset = $self->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Zoneset object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Run cfgadd on the switch with the zoneset name and zone.
    # Format of cfgadd is as follows:
    # cfgadd "zoneset", "zone"
    my $cmd = "cfgadd \"" . ${zoneset} . "\", \"" . $zone . "\"";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check for an "already contains" error
    if (grep(/already contains/, @output) > 0) {

        # Zoneset already contains the zone
        $self->_set_err("Zoneset $zoneset already contains zone $zone");
        return $SwitchUtil::FALSE;
    } ## end if (grep(/already contains/...

    # Update the zoneset object with the new zone
    $self->{zones}->{$zone} = $zoneobj;

    return $SwitchUtil::TRUE;
} ## end sub add_zone

###############################################################################
# Method:    remove_zone
# Objective: Remove a zone from the zoneset on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_zone {
    my ($self, %args) = @_;

    my ($zoneobj);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed a zone object
            if (not length($zoneobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($zoneobj)) {
        unless (ref($zoneobj) =~ /Zone/) {
            $self->_set_err("Invalid value for argument '-zone'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    # Verify that the zone object has a defined name
    my $zone = $zoneobj->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object provided does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the zoneset object has a defined name
    my $zoneset = $self->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Zoneset object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Run cfgremove on the switch with the zoneset name and zone.
    # Format of cfgremove is as follows:
    # cfgremove "zoneset", "zone"
    my $cmd = "cfgremove \"" . ${zoneset} . "\", \"" . $zone . "\"";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to see if we got a "is not in" error on the remove
    if (grep(/is not in/, @output) > 0) {

        # Unable to find zone in zoneset
        $self->_set_err("Zone $zone was not found in zoneset $zoneset");
        return $SwitchUtil::FALSE;
    } ## end if (grep(/is not in/, ...

    # check to see if $self->{zones}->{$zone} is defined.  If it is, delete
    # the hash entry.
    if (defined($self->{zones}->{$zone})) {
        unless (delete($self->{zones}->{$zone})) {
            $self->_set_er("Unable to delete zone $zone from zone object hash");
            return $SwitchUtil::FALSE;
        }
    } ## end if (defined($self->{zones...

    # Adding 2 second sleep.  For some reason, brocade is returning some weird
    # control characters which is causing some buffering issues with the prompt.
    # Adding the sleep keeps this from causing issues with future commands.
    &SanUtil::nate_sleep(2);

    return $SwitchUtil::TRUE;
} ## end sub remove_zone

###############################################################################
# Method:    verify_zone
# Objective: Check that a zone exists in the zoneset on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_zone {
    my ($self, %args) = @_;

    my ($zoneobj);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed a zone object
            if (not length($zoneobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($zoneobj)) {
        unless (ref($zoneobj) =~ /Zone/) {
            $self->_set_err("Invalid value for argument '-zone'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    # Verify that the zone object has a defined name
    my $zone = $zoneobj->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object provided does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Discover the zones in the zoneset
    my $zone_ref = $self->discover();
    if ($self->error) {
        $self->_set_error("Error discovering zone information for zoneset:\n" . $self->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we got back zones
    unless (scalar(keys %{$zone_ref}) > 0) {
        $self->_set_err("No zones were returned from the switch for zoneset " . $self->get_name());
        return $SwitchUtil::FALSE;
    }

    # Check the zone reference to make sure $zone is one of the keys
    if (defined($zone_ref->{$zone})) {
        return $SwitchUtil::TRUE;
    } else {

        # We couldn't find the alias in the zone on the switch
        return $SwitchUtil::FALSE;
    }
} ## end sub verify_zone

###############################################################################
# Method:    discover
# Objective: Discover the alias information for the zoneset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zoneset object has a defined name
    my $zoneset = $self->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Zoneset object does not have a name defined");
        return {};
    }

    # Get the zoneset information from the switch
    my @output = $self->_query_zoneset();
    if ($self->error) {
        return {};
    }

    # Check to make sure we didn't get a "does not exist" error
    if (grep(/does not exist/, @output) > 0) {
        $self->_set_err("Zoneset $zoneset does not exist on the switch");
        return {};
    }

    # Decode the zoneset information into a hash reference
    my $zoneset_ref = $self->{parent}->_decode_zonesets(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to decode zonesets:\n", $self->{parent}->errmsg());
        return {};
    }

    # Double check that the zoneset you are looking for is found as a
    # key to the hash reference
    unless (defined($zoneset_ref->{$zoneset})) {
        $self->_set_err("Zoneset information returned from the switch does not match");
        return {};
    }

    # Return the zone information
    return $zoneset_ref->{$zoneset};
} ## end sub discover

###############################################################################
# Method:    is_on_switch
# Objective: returns the "on_switch" value off of the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_on_switch {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{on_switch}) and $self->{on_switch} ne "") {
        $self->_set_err("Invalid value for 'on_switch' key on object");
        return $SwitchUtil::FALSE;
    }

    return $self->{on_switch};
}

###############################################################################
# Method:    new
# Objective: Create new zoneset object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, $parent, $zoneset, @zones);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
        } elsif (/-?zoneset/) {
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '-zoneset'");
                return $self;
            }
        } elsif (/-?zones/) {
            unless (ref($args{$_}) =~ /ARRAY/ and scalar(@zones = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-zones'");
                return $self;
            }
        } ## end elsif (/-?zones/)
    } ## end foreach (keys %args)

    # Verify mandatory parameters.
    if (defined($parent) and $parent ne "") {
        unless (ref($parent) =~ /BrocadeFCZone/i) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Mandatory parameter \"zoneset\"", "not specified");
        return $self;
    }

    # Verify the optional parameters
    # See if we were given zones.  If we were, then get the name of
    # each zone and use it as a key in $zones_ref with the value
    # being the zone object.
    my $zones_ref = {};
    if (scalar(@zones > 0)) {
        foreach my $zoneobj (@zones) {
            my $zone = $zoneobj->get_name();
            unless (defined($zone) and $zone ne "") {
                $self->_set_err("Unable to determine zone name");
                return $self;
            }
            $zones_ref->{$zone} = $zoneobj;
        } ## end foreach my $zoneobj (@zones)
    } ## end if (scalar(@zones > 0))

    # initialize the object aliases
    $self->{parent}    = $parent;
    $self->{zoneset}   = $zoneset;
    $self->{zones}     = $zones_ref;
    $self->{on_switch} = $SwitchUtil::FALSE;

    return $self;
} ## end sub new

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _query_zoneset
# Objective: Query the switch and return the zoneset information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _query_zoneset {
    my ($self) = @_;

    $self->_clear_err();

    my $zoneset    = $self->get_name();
    my $switch_obj = $self->{parent}->{parent};

    # Run cfgshow on the switch with the zoneset name
    # Format of zonesetshow command is "zonesetshow <zoneset>,0"
    my $cmd = "cfgshow " . ${zoneset} . ",0";
    my @output = $switch_obj->execute_brocade($cmd, $self->{parent}->{timeout}, "", -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return ();
    }

    return @output;
} ## end sub _query_zoneset

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::ZoneSet.pm

=head1 Author

Derrick Shultz

=head1 Description

Provides an abstract API interface to ZoneSet operations on Brocade switches.

=head1 See Also

Package inherits from the following modules. See modules for interface detail

=over 4

=item * B<San::SanLog.pm>

=item * B<Error.pm>

=back

=head1 Attributes

=head2 B<parent>

ZoneMgr object = Reference to a parent ZoneMgr object

=head2 B<zoneset>

String containing name of the zoneset

=head2 B<zones>

Reference to a hash containing zone objects with a key of the zone name and a value of the zone object.

=head2 B<on_switch>

Boolean of $SwitchUtil::TRUE or $SwitchUtil::FALSE representing whether or not the zoneset is currently found on
the switch.

=head1 Public Methods

=head2 B<add_zone()>

Adds a zone to an already existing zoneset on a switch

=head3 Inputs

=over 4

=item B<-zone>

Zone object to add to the zoneset.

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zonesetobj->add_zone(-zone => $zoneobj);
    if ($zonesetobj->error) {
        $self->_set_err($zonesetobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<create()>

Creates the zoneset on the switch.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zonesetobj->create();
    if ($zonesetobj->error) {
        $self->_set_err($zonesetobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<destroy()>

Removes the zoneset from the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zonesetobj->destroy();
    if ($zonesetobj->error) {
        $self->_set_err($zonesetobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<discover()>

Discovers the zone information for the zoneset on the switch

=head3 Inputs

None

=head3 Outputs

Returns a reference to a hash containing zone names as the key and $SwitchUtil::TRUE as the value like
$zoneset_ref->{"zone1" => $SwitchUtil::TRUE, "zone2" => $SwitchUtil::TRUE}. If an error occurs, the error is
set to the failure reason and {} is returned.

=head3 Examples

    my $zone_ref = $zonesetobj->discover();
    if ($zonesetobj->error) {
        $self->_set_err($zonesetobj->errmsg());
        return $SwitchUtil::FALSE;
    }

=head2 B<exists()>

Checks for the existance of the zoneset on the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error message is
set to the failure reason and $SwitchUtil::FALSE is returned. If there wasn't an error,
but the zone doesn't actually exist it returns $SwitchUtil::FALSE without setting an error.

=head3 Examples

    my $status = $zonesetobj->exists();
    if ($status == $SwitchUtil::FALSE) {
        if ($zonesetobj->error) {
            # We had a problem checking for the existance of the zoneset
            $self->_set_err($zonesetobj->errmsg());
            return $SwitchUtil::FALSE;
        } else {
            # No problems were found. The Zone doesn't exist.
            return $SwitchUtil::FALSE;
        }
    } else {
        # Zone Exists
        return $SwitchUtil::TRUE;
    }

=head2 B<get_zones()>

Method to get all of the zone objects in the zoneset

=head3 Inputs

None

=head3 Outputs

Returns a reference to a hash of zone objects keyed on the zone name with a value of the zone object
Returns the member string from the zone object.

=head3 Examples

    my $zones_ref = $zonesetobj->get_zones();

=head2 B<get_name()>

Method for retrieving the name of the zoneset

=head3 Inputs

None

=head3 Outputs

Returns the name string from the zoneset object.

=head3 Examples

    my $name = $zonesetobj->get_name();

=head2 B<is_on_switch()>

Method to return the on_switch key from the object

=head3 Inputs

None

=head3 Outputs

Returns a boolean based on the value of the on_switch key. If the key doesn't exist or is empty, it sets an error.

=head3 Examples

    if ($obj->is_on_switch()) {
        do stuff here;
    }

=head2 B<new()>

Instantiates the ZoneSet object.

=head3 Inputs

=over 4

=item B<-parent>

ZoneMgr object

=item B<-zoneset>

String containing the zoneset name

=item B<-zones (Optional)>

Reference to an array of zones to add to the zoneset

=back

=head3 Outputs

Returns a zoneset object. If an error occurs, the error flag is set and errmsg is
set to the failure reason before the object is returned.

=head3 Examples

    my $zonesetobj = San::SanSwitch::FC::ZoneMgr::BrocadeFCZone::ZoneSet::new(-parent => $self,
                                                                              -zoneset => $zoneset_name,
                                                                              -zones => \@zones);
    if ($zonesetobj->error) {
        $self->_set_err($zonesetobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<remove_zone()>

Removes a zone from an existing zoneset on a switch

=head3 Inputs

=over 4

=item B<-zone>

Zone object to remove from the zoneset

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zonesetobj->remove_zone(-zone => $zoneobj);
    if ($zonesetobj->error) {
        $self->_set_err($zonesetobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<validate()>

Validates the zoneset on the switch against the object.  This validation includes making sure that the zoneset actually
has zones defined on it and then makes sure that the zoneset matches exactly.  This means that all zones defined
in the zoneset object match the zones in the switch zoneset.  Any extra zones in either the object or the switch will
cause a failure.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zonesetobj->validate();
    if ($zonesetobj->error) {
        $self->_set_err($zonesetobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<verify_zone()>

Checks the zoneset on the switch to make sure that it currently contains the zone specified.

=head3 Inputs

=over 4

=item B<-zone>

Zone object to verify in the zoneset.

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zonesetobj->verify_zone(-zone => $zoneobj);
    if ($zonesetobj->error) {
        $self->_set_err($zonesetobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head1 Private Methods

=head2 B<_query_zoneset()>

Runs "cfgshow <zoneset>,0" on the switch and returns the output from the command.

=head3 Inputs

None

=head3 Outputs

On success, returns an array containing the "cfgshow" output. If an error occurs, the error is
set to the failure reason and () is returned.

=head3 Examples

    my @query_output = $zonesetobj->_query_zoneset();
    if ($zonesetobj->error) {
        $self->_set_err($zonesetobj->errmsg());
        return $SwitchUtil::FALSE;
    }

=cut
