# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to zone a switch
## @author derricks@netapp.com
## @status private
## @pod here
## @unit_test thpl p4/test/lib/UnitTest/test_San_SanSwitch_ZoneMgr.thpl

package San::SanSwitch::FC::ZoneMgr::CiscoFCZone;

# Compiler directives.
use strict;

# Module import.
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Alias;
use San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Zone;
use San::SanSwitch::FC::ZoneMgr::CiscoFCZone::ZoneSet;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog ZoneMgr);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new_alias
# Objective: Create a new alias object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new_alias {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $alias_object = San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Alias->new(-parent => $self, %args);
    if ($alias_object->error) {
        $self->_set_err($alias_object->errmsg());
        return $alias_object;
    }

    # Add the object to the cache
    if (not $self->_alias_cache_exists(-alias => $alias_object)) {
        if (not $self->error) {

            # Object doesn't exist in the cache and there weren't
            # any errors so go ahead and add it to the cache.
            if (not $self->_alias_cache_add(-alias => $alias_object)) {

                # Failed to add it to the cache
                if (not $self->error) {
                    $self->_set_err("Unable to add alias to alias Manager alias cache");
                }
                return $alias_object;
            } ## end if (not $self->_alias_cache_add...
        } ## end if (not $self->error)
    } else {

        # We found an entry in the cache for this object so it must be a duplicate.
        # Set an error and bail.
        $self->_set_err("Duplicate entry found in cache for alias object");
        return $alias_object;
    } ## end else [ if (not $self->_alias_cache_exists...

    # Return the alias object
    return $alias_object;
} ## end sub new_alias

###############################################################################
# Method:    new_zone
# Objective: Create a new zone object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new_zone {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $zone_object = San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Zone->new(-parent => $self, %args);
    if ($zone_object->error) {
        $self->_set_err($zone_object->errmsg());
        return $zone_object;
    }

    # Add the object to the cache
    if (not $self->_zone_cache_exists(-zone => $zone_object)) {
        if (not $self->error) {

            # Object doesn't exist in the cache and there weren't
            # any errors so go ahead and add it to the cache.
            if (not $self->_zone_cache_add(-zone => $zone_object)) {

                # Failed to add it to the cache
                if (not $self->error) {
                    $self->_set_err("Unable to add zone to Zone Manager zone cache");
                }
                return $zone_object;
            } ## end if (not $self->_zone_cache_add...
        } ## end if (not $self->error)
    } else {

        # We found an entry in the cache for this object so it must be a duplicate.
        # Set an error and bail.
        $self->_set_err("Duplicate entry found in cache for zone object");
        return $zone_object;
    } ## end else [ if (not $self->_zone_cache_exists...

    # Return the zone object
    return $zone_object;
} ## end sub new_zone

###############################################################################
# Method:    new_zoneset
# Objective: Create a new zoneset object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new_zoneset {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $zoneset_object = San::SanSwitch::FC::ZoneMgr::CiscoFCZone::ZoneSet->new(-parent => $self, %args);
    if ($zoneset_object->error) {
        $self->_set_err($zoneset_object->errmsg());
        return $zoneset_object;
    }

    # Add the object to the cache
    if (not $self->_zoneset_cache_exists(-zoneset => $zoneset_object)) {
        if (not $self->error) {

            # Object doesn't exist in the cache and there weren't
            # any errors so go ahead and add it to the cache.
            if (not $self->_zoneset_cache_add(-zoneset => $zoneset_object)) {

                # Failed to add it to the cache
                if (not $self->error) {
                    $self->_set_err("Unable to add zoneset to Zone Manager zoneset cache");
                }
                return $zoneset_object;
            } ## end if (not $self->_zoneset_cache_add...
        } ## end if (not $self->error)
    } else {

        # We found an entry in the cache for this object so it must be a duplicate.
        # Set an error and bail.
        $self->_set_err("Duplicate entry found in cache for zone object");
        return $zoneset_object;
    } ## end else [ if (not $self->_zoneset_cache_exists...

    # Return the zoneset object
    return $zoneset_object;
} ## end sub new_zoneset

###############################################################################
# Method:    alias_discover
# Objective: Discover the aliases on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub alias_discover {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{vsan}) and $self->{vsan} ne "") {
        $vsan = $self->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Zero out the aliases reference in the ZoneMgr object
    $self->{aliases} = {};

    # Run "show fcalias" on the switch to get a list of all aliases in vsan $vsan
    # Format of command is "show fcalias vsan $vsan"
    my $cmd = "show fcalias vsan $vsan";
    my @output = $switch_obj->execute($cmd, $self->{timeout}, -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)\n" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "Alias not present" error
    # If we don't find anything just go ahead and return TRUE
    # since this isn't an error condition.
    if (grep(/alias not present/i, @output) > 0) {
        $self->sanlog(-msg => "No aliases were found on the switch for vsan $vsan");
        return $SwitchUtil::TRUE;
    }

    # Decode the alias information into a hash reference
    my $aliases_ref = $self->_decode_aliases(@output);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Get the device aliases
    $cmd = "show device-alias database";
    @output = $switch_obj->execute($cmd, $self->{timeout}, -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)\n" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "Alias not present" error
    # If we don't find anything just go ahead and return TRUE
    # since this isn't an error condition.
    if (grep(/alias not present/i, @output) > 0) {
        $self->sanlog(-msg => "No aliases were found on the switch for vsan $vsan");
        return $SwitchUtil::TRUE;
    }

    # Decode the device aliases information into a hash reference
    my $devaliases_ref = $self->_decode_aliases(@output);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Add the device aliases to the $alaises_ref hash
    map {$aliases_ref->{$_} = $devaliases_ref->{$_}} keys %$devaliases_ref;

    # Parse through the aliases_ref hash and create the new alias objects
    foreach my $alias (sort keys %${aliases_ref}) {

        # Check to make sure that the member and type are defined
        if (not defined($aliases_ref->{$alias}->{member})
            or not defined($aliases_ref->{$alias}->{type})
            or $aliases_ref->{$alias}->{member} eq ""
            or $aliases_ref->{$alias}->{type}   eq "") {
            $self->_set_err("Alias member or type parameters not defined");
            return $SwitchUtil::FALSE;
        } ## end if (not defined($aliases_ref...

        my $t_aliasobj = $self->new_alias(
            -alias    => $alias,
            -member   => $aliases_ref->{$alias}->{member},
            -type     => $aliases_ref->{$alias}->{type}
        );

        if (not defined($t_aliasobj) or ref($t_aliasobj) !~ /alias/i) {
            if (not $self->error) {
                $self->_set_err("Unable to create alias object for alias $alias");
            }
            return $SwitchUtil::FALSE;
        } ## end if (not defined($t_aliasobj...

        # Update the alias_object "on_switch" parameter.
        $t_aliasobj->{on_switch} = $SwitchUtil::TRUE;

        # Add the alias object to the alias cache
        if (not $self->_alias_cache_add(-alias => $t_aliasobj)) {
            if (not $self->error) {
                $self->_set_err("Unable to alias $alias to ZoneMgr alias cache");
            }
            return $SwitchUtil::FALSE;
        } ## end if (not $self->_alias_cache_add...
    } ## end foreach my $alias (sort keys...


    return $SwitchUtil::TRUE;
} ## end sub alias_discover

###############################################################################
# Method:    zone_discover
# Objective: Discover the zones on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_discover {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{vsan}) and $self->{vsan} ne "") {
        $vsan = $self->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Zero out the zones reference in the ZoneMgr object
    $self->{zones} = {};

    # Run "show zone" on the switch to get a list of all zones.
    # Format of the command is "show zone vsan $vsan";
    my $cmd = "show zone vsan $vsan";
    my @output = $switch_obj->execute($cmd, $self->{timeout}, -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)\n" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "does not exist" error
    # If we don't find anything just go ahead and return TRUE
    # since this isn't an error condition.
    if (grep(/zone not present/i, @output) > 0) {
        $self->sanlog(-msg => "No zones were found on the switch");
        return $SwitchUtil::TRUE;
    }

    # Decode the zone information into a hash reference
    my $zones_ref = $self->_decode_zones(@output);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    foreach my $zone (sort keys %${zones_ref}) {

        # Get all of the alias objects for any aliases on the zone
        my @aliases = ();
        foreach my $alias (keys %{$zones_ref->{$zone}}) {
            my $aliasobj = $self->get_aliases(-alias => $alias);
            unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {

                # If we got here then the zone was likely not created
                # with aliases but instead with wwpn or port.  Create
                # a generic alias object based the wwpn or port.
                if ($self->error) {
                    return $SwitchUtil::FALSE;
                } else {
                    # Create a new alias object.  Since we are only creating
                    # the object and are not actually using it for any real work,
                    # it doesn't really matter how we create it.
                    if ($alias =~ /\S+:\S+:\S+:\S+:\S+:\S+:\S+:\S+/) {
                        # We have a wwpn.  This could either be a fwwn or a pwwn.
                        $aliasobj = $self->new_alias(-alias => $alias, -member => $alias, -type => "pwwn");
                    } elsif ($alias =~ /\d+\/\d+/) {
                        # We have an interface
                        $aliasobj = $self->new_alias(-alias => $alias, -member => $alias, -type => "interface");
                    } elsif ($alias =~ /\S+/) {
                        # We have an Device Alias that isn't an interface
                        $aliasobj = $self->new_alias(-alias => $alias, -member => $alias, -type => "pwwn");
                    } else {
                        $self->_set_err("Unable to determine member and alias $alias information for zone $zone");
                        return $SwitchUtil::FALSE;
                    }

                    unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {
                        if (not $self->error) {
                            $self->_set_err("Unable to create alias object for alias $alias");
                        }
                        return $SwitchUtil::FALSE;
                    } ## end unless (defined($aliasobj)...

                    # Update the on_switch parameter of the alias object
                    $aliasobj->{on_switch} = $SwitchUtil::TRUE;
                } ## end else [ if ($self->error)
            } ## end unless (defined($aliasobj)...
            push(@aliases, $aliasobj);
        } ## end foreach my $alias (keys %{$zones_ref...

        # Create the zone objects
        my $t_zoneobj;
        if (scalar(@aliases) > 0) {
            $t_zoneobj = $self->new_zone(-zone => $zone, -aliases => \@aliases);
        } else {
            $t_zoneobj = $self->new_zone(-zone => $zone);
        }
        unless (defined($t_zoneobj) and ref($t_zoneobj) =~ /Zone/i) {
            if (not $self->error) {
                $self->_set_err("Unable to create zone object for zone $zone");
            }
            return $SwitchUtil::FALSE;
        } ## end unless (defined($t_zoneobj...

        # Update the on_switch parameter of the zone object
        $t_zoneobj->{on_switch} = $SwitchUtil::TRUE;

        # Add the zone object to the zone cache
        if (not $self->_zone_cache_add(-zone => $t_zoneobj)) {
            if (not $self->error) {
                $self->_set_err("Unable to zone $zone to ZoneMgr zone cache");
            }
            return $SwitchUtil::FALSE;
        } ## end if (not $self->_zone_cache_add...
    } ## end foreach my $zone (sort keys...

    return $SwitchUtil::TRUE;
} ## end sub zone_discover

###############################################################################
# Method:    zoneset_discover
# Objective: Discover the zone sets on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zoneset_discover {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{vsan}) and $self->{vsan} ne "") {
        $vsan = $self->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Zero out the zones reference in the ZoneMgr object
    $self->{zonesets} = {};

    # Run "show zoneset" on the switch to get a list of all zonesets.
    # Format of then command is "show zoneset vsan $vsan"
    my $cmd = "show zoneset vsan $vsan";
    my @output = $switch_obj->execute($cmd, $self->{timeout}, -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)\n" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "zoneset not present" error
    # If we don't find anything just go ahead and return TRUE
    # since this isn't an error condition.
    if (grep(/zoneset not present/i, @output) > 0) {
        $self->sanlog(-msg => "No zonesets were found on the switch for vsan $vsan");
        return $SwitchUtil::TRUE;
    }

    # Decode the zoneset information into a hash reference
    my $zonesets_ref = $self->_decode_zonesets(@output);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    foreach my $zoneset (sort keys %${zonesets_ref}) {

        # Get all of the zone objects for any zones on the zoneset
        my @zones = ();
        foreach my $zone (keys %{$zonesets_ref->{$zoneset}}) {
            my $zoneobj = $self->get_zones(-zone => $zone);
            if (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
                push(@zones, $zoneobj);
            } else {
                if (not $self->error) {
                    $self->_set_err("No zone object was found for zone $zone in the ZoneMgr Zone cache");
                }
                return $SwitchUtil::FALSE;
            } ## end else [ if (defined($zoneobj) ...
        } ## end foreach my $zone (keys %{$zonesets_ref...

        # Create the zoneset objects
        my $t_zonesetobj;
        if (scalar(@zones) > 0) {
            $t_zonesetobj = $self->new_zoneset(-zoneset => $zoneset, -zones => \@zones);
        } else {
            $t_zonesetobj = $self->new_zoneset(-zoneset => $zoneset);
        }

        unless (defined($t_zonesetobj) and ref($t_zonesetobj) =~ /ZoneSet/i) {
            if (not $self->error) {
                $self->_set_err("Unable to create zoneset object for zoneset $zoneset");
            }
            return $SwitchUtil::FALSE;
        } ## end unless (defined($t_zonesetobj...

        # Update the on_switch parameter of the zoneset object
        $t_zonesetobj->{on_switch} = $SwitchUtil::TRUE;

        # Add the zoneset object to the zoneset cache
        if (not $self->_zoneset_cache_add(-zoneset => $t_zonesetobj)) {
            if (not $self->error) {
                $self->_set_err("Unable to zone $zoneset to ZoneMgr zoneset cache");
            }
            return $SwitchUtil::FALSE;
        } ## end if (not $self->_zoneset_cache_add...
    } ## end foreach my $zoneset (sort keys...

    return $SwitchUtil::TRUE;
} ## end sub zoneset_discover

###############################################################################
# Method:    active_zoneset_discover
# Objective: Discover the active zoneset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub active_zoneset_discover {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{vsan}) and $self->{vsan} ne "") {
        $vsan = $self->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Run "show zoneset active vsan $vsan" and look for the "zoneset name" line.
    my $cmd = "show zoneset active vsan $vsan";
    my @output = $switch_obj->execute($cmd, $self->{timeout}, -quiet => 1);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)\n" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check for "zoneset not present" error.
    if (grep(/zoneset not present/i, @output) > 0) {
        $self->{active_zoneset} = undef;
        $self->sanlog(-msg => "No active zonesets were found on the switch for vsan $vsan");
        return $SwitchUtil::TRUE;
    }

    # Loop throught the output and look for "zoneset name ZONESET_NAME vsan $vsan"
    foreach (@output) {
        if (/zoneset name (\S+) vsan $vsan/) {
            my $zoneset = $1;

            # Get the object for the active zoneset
            my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);
            if ($self->error) {
                return $SwitchUtil::FALSE;
            }
            if (defined($zonesetobj)) {
                if (ref($zonesetobj) =~ /ZoneSet/i) {

                    # Set the active_zoneset on the zonemgr object to $zonsetobj
                    $self->{active_zoneset} = $zonesetobj;
                    return $SwitchUtil::TRUE;
                } elsif (scalar(keys %$zonesetobj) == 0) {

                    # Looks like there isn't an object for this
                    # zoneset so we need to create one.
                    my $t_zonesetobj = $self->new_zoneset(-zoneset => $zoneset);
                    if (not defined($t_zonesetobj) and ref($t_zonesetobj) !~ /ZoneSet/i) {
                        if (not $self->error) {
                            $self->_set_err("Unable to create zoneset object for zoneset $zoneset");
                        }
                        return $SwitchUtil::FALSE;
                    } else {
                        $self->{active_zoneset} = $t_zonesetobj;
                        return $SwitchUtil::TRUE;
                    }
                } else {

                    # We got back something totally unexpected
                    $self->_set_err("Unable to determine the object for zoneset $zoneset");
                    return $SwitchUtil::FALSE;
                } ## end else [ if (ref($zonesetobj) =~...
            } else {
                $self->_set_err("Unable to determine the object for zoneset $zoneset");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/zoneset name (\S+) vsan $vsan/)
    } ## end foreach (@output)

    # if we made it this far then we didn't find the line with "cfg:" so set
    # an error and return false.
    $self->_set_err("Unable to find active zoneset information in the output of command ($cmd)");
    return $SwitchUtil::FALSE;
} ## end sub active_zoneset_discover

###############################################################################
# Method:    config_enable
# Objective: enable the zoneset config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_enable {
    my ($self, %args) = @_;

    my ($zoneset);

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{vsan}) and $self->{vsan} ne "") {
        $vsan = $self->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset object
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    # Get the zoneset object
    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);

    unless (defined($zonesetobj) and ref($zonesetobj) =~ /ZoneSet/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($zonesetobj...

    my $cmd = "conf t";
    my @output = $switch_obj->execute($cmd);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)\n" . $switch_obj->errmsg());
        # handle error, by attempting to get out of config mode
        # and then returning with the proper error set.
        my @cmds = (" ", "end");
        foreach my $cmd (@cmds) {
            @output = $switch_obj->execute($cmd);
        }
        return $SwitchUtil::FALSE;
    }

    # use say for the next set of commands
    my ($match, $before, $after);
    my $connection = $switch_obj->get_connection();
    if (not defined($connection) or $connection eq "") {
        $self->_set_err("Unable to get connection info for switch");
        return $SwitchUtil::FALSE;
    }

    ($match,$before,$after) = &Tharn::say($connection, "zoneset activate name $zoneset vsan $vsan",
                                          '-prompt', "]",
                                          '-timeout', 30);

    if ($after =~ /Do you want to continue/i) {
        ($match,$before,$after) = &Tharn::say($connection, "y",
                                              '-prompt', "#",
                                              '-timeout', 30);
    }

    my @cmds = ("zoneset distribute full vsan $vsan", "end");
    foreach my $cmd (@cmds) {
        my @output = $switch_obj->execute($cmd);
        if ($switch_obj->error) {
            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            $self->_set_err("Error executing command ($cmd)\n" . $switch_obj->errmsg());
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $switch_obj->execute($cmd);
            }
            $SwitchUtil::FALSE;
        }
    }

    # Update the zonemgr object's active zoneset definition.
    $self->{active_zoneset} = $zonesetobj;

    return $SwitchUtil::TRUE;
} ## end sub config_enable

###############################################################################
# Method:    config_disable
# Objective: disable the active config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_disable {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{vsan}) and $self->{vsan} ne "") {
        $vsan = $self->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Get the name of the active zoneset
    my $zoneset;
    if (defined($self->{active_zoneset}) and ref($self->{active_zoneset}) =~ /zoneset/i) {
        $zoneset = $self->{active_zoneset}->get_name();
        unless (defined($zoneset) and $zoneset ne "") {
            $self->_set_err("Zoneset object does not have a name defined");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Unable to determine the active zoneset for vsan $vsan");
        return $SwitchUtil::FALSE;
    }

    # Run the commands to enable the zoneset for the specified vsan
    my @cmds = ("conf t", "no zoneset activate name $zoneset vsan $vsan", "zoneset distribute full vsan $vsan", "end");

    foreach my $cmd (@cmds) {
        my @output = $switch_obj->execute($cmd);
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)\n" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $switch_obj->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($self->error)
    } ## end foreach my $cmd (@cmds)

    # Update the zonemgr object's active zoneset definition.
    $self->{active_zoneset} = undef;

    return $SwitchUtil::TRUE;
} ## end sub config_disable

###############################################################################
# Method:    config_save
# Objective: save the switch config
# Details:   See POD documentation at the end of this file.
###############################################################################
sub config_save {
    my ($self) = @_;

    $self->_clear_err();

    my $switch_obj = $self->{parent};

    # Run the commands to enable the zoneset for the specified vsan
    my $cmd    = "copy run start";
    my @output = $switch_obj->execute($cmd);
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)\n" . $switch_obj->errmsg());
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub config_save

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $timeout, $vsan);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?timeout/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-timeout'");
                return $self;
            }
            next;
        } elsif (/-?vsan/) {
            if (not length($vsan = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vsan'");
                return $self;
            }
            next;
        } ## end elsif (/-?vsan/)
    } ## end foreach (keys %args)

    # check for existance of required -parent argument
    if (defined($parent)) {
        unless (ref($parent) =~ /SanSwitch/i) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # Check for existance of -timeout arguement.  Default to 120 if not found.
    unless (defined($timeout) and $timeout ne "") {
        $timeout = 120;
    }

    # Check for existance of -vsan arguement.  Default to 1 if not found.
    unless (defined($vsan) and $vsan ne "") {
        $timeout = 1;
    }

    # initialize the object members
    $self->{parent}         = $parent;
    $self->{aliases}        = {};
    $self->{zones}          = {};
    $self->{zonesets}       = {};
    $self->{timeout}        = $timeout;
    $self->{vsan}           = $vsan;
    $self->{active_zoneset} = undef;

    return $self;
} ## end sub _new

###############################################################################
# Method:    _decode_aliases
# Objective: Decode switch alias output into a hash reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _decode_aliases {
    my ($self, @input) = @_;

    $self->_clear_err();

    # Initialize the hash reference;
    my $alias_ref = {};

    my ($alias, $member, $type, $vsan);
    foreach (@input) {
        chomp();

        # Skip any blank lines or lines that say "show fcalias"
        next if (/^\s*$|show fcalias/);

        # Remove all leading and trailing whitespace
        s/^\s+//g;
        s/\s+$//g;

        # Once we have removed the whitespaces, we should be left with something
        # like the following:
        #
        # fcalias name FAS3050-RTP01-0D-2-MIXED vsan 1
        # pwwn 50:0a:09:89:86:97:ab:2d
        #
        # OR
        #
        # fcalias name FAS3050-RTP01-0D-2-MIXED vsan 1
        # interface fc1/30
        #
        # OR
        #
        # fcalias name FAS3050-RTP01-0D-2-MIXED vsan 1
        # fwwn 20:22:00:0d:ec:a3:06:bf
        #
        # OR
        #
        # device-alias name fas3070-rtp02-SM-1B pwwn 21:01:00:e0:8b:a6:7a:9d
        # device-alias name fas6070-rtp01-SM-8B pwwn 21:01:00:e0:8b:a6:95:9a
        # device-alias name fas6070-rtp02-SM-8B pwwn 21:01:00:e0:8b:a6:43:45

        if (/^fcalias name (\S+) vsan/) {
            $alias = $1;
        } elsif (/^device-alias name (\S+) (\S+) (\S+)/) {
            $alias  = $1;
            $type   = $2;
            $member = $3;

            # Create an entry in the $alias_ref hash
            $alias_ref->{$alias}->{member} = $member;
            $alias_ref->{$alias}->{type}   = $type;

            # Set $alias, $type, and $member to undef
            $alias  = undef;
            $member = undef;
            $type   = undef;
        } elsif (/^(interface|fwwn|pwwn) (\S+)/) {
            $type   = $1;
            $member = $2;

            # Check to make sure that $alias is defined.  if it isn't, then we have
            # an issue with the data provided so set an error and bail out.
            unless (defined($alias) and $alias ne "") {
                $self->_set_err("Unable to determine alias name from line:\n" .
                                $_ . "\nInput data being parsed contained:\n",
                                @input);

                return {};
            }

            # If we found either a did,port or wwpn then we assume that is the
            # member for the alias that was sent.
            $alias_ref->{$alias}->{member} = $member;
            $alias_ref->{$alias}->{type}   = $type;
        } ## end elsif (/^(interface|fwwn|pwwn) (\S+)/)
    } ## end foreach (@input)

    return $alias_ref;
} ## end sub _decode_aliases

###############################################################################
# Method:    _decode_zones
# Objective: Decode switch zone output into a hash reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _decode_zones {
    my ($self, @input) = @_;

    $self->_clear_err();

    # Initialize the hash reference;
    my $zone_ref = {};

    my ($zone, $alias, $vsan, $type, $fcalias);

    # We need to keep track of whether we found an fcalias line or
    # not since if we do, it will then be followed by the information
    # for that fcalias (pwwn).  We need to skip that next line.
    # However, if we didn't find an fcalias, then we want to use that pwwn line
    foreach (@input) {
        chomp();

        # Skip any blank lines or "show zone"
        next if (/^\s*$|show zone/);

        # Remove any trailing whitespace
        s/\s+$//g;

        # Once we have removed the whitespaces, we should be left with something
        # like the following:
        #
        # zone name V3070-01-0A-HDS-CTRL0-0A vsan 1
        #   pwwn 50:06:0e:80:10:49:d5:70
        #   pwwn 50:0a:09:82:00:01:dd:ec [V3070-rtp01-0a]
        #
        # OR
        #
        # zone name derrick_test_zone1 vsan 19
        #   fcalias name derrick_test_alias1 vsan 19
        #     pwwn 10:00:00:00:c9:43:21:a2
        #   fcalias name derrick_test_alias2 vsan 19
        #     pwwn 10:00:00:00:c9:43:21:a3
        #
        # OR
        #
        # zone name derrick_test_zone2 vsan 19
        #   pwwn 10:00:00:00:c9:66:1c:74 [hpdl380g5-rtp09]
        #
        if (/^zone name (\S+) vsan/) {
            $zone = $1;
            $zone_ref->{$zone} = {};
        } elsif (/^\s+fcalias name (\S+) vsan/) {

            # We have an fcalias
            $alias = $1;
            $fcalias = 1;
        } elsif (/^\s+pwwn (\S+) \[(\S+)\]/) {
            # Check to see if this line is following an fcalias
            if ($fcalias) {
                # It is so just skip it.  Also set fcalias back to 0;
                $fcalias = 0;
                next;
            } else {
                # We have an device-alias
                $alias = $2;
            }
        } elsif (/^\s+(pwwn|fwwn|interface) (\S+)/) {
            if ($fcalias) {
                # It is so just skip it.  Also set fcalias back to 0;
                $fcalias = 0;
                next;
            } else {
                # We don't have an fcalias nor a device alias
                $alias = $2;
            }
        }

        # if we have an alias and we have a zone, add it to the zone_ref hash.
        if (defined($alias) and $alias ne "") {
            if (defined($zone) and defined($zone_ref->{$zone})) {
                # Add the alias to the zone
                $zone_ref->{$zone}->{$alias} = $SwitchUtil::TRUE;
                $alias = undef;
            } else {
                $self->_set_err("Unable to determine zone name from line:\n" .
                                $_ . "\nInput data being parsed contained:\n",
                                @input);
                return {};
            }
        }
    } ## end foreach (@input)

    return $zone_ref;
} ## end sub _decode_zones

###############################################################################
# Method:    _decode_zonesets
# Objective: Decode switch zoneset output into a hash reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _decode_zonesets {
    my ($self, @input) = @_;

    $self->_clear_err();

    # Initialize the hash reference;
    my $zoneset_ref = {};

    my $zoneset;
    foreach (@input) {
        chomp();

        # Skip any blank lines or "show zoneset"
        next if (/^\s*$|show zoneset/);

        # Remove any leading and trailing whitespace
        s/^\s+//g;
        s/\s+$//g;

        # Once we have removed the whitespaces, we should be left with something
        # like the following:
        #
        # zoneset name Fabric1-FCP-Switch-Qual-IAQ-Zset vsan 1
        # zone name V3070-01-0A-HDS-CTRL0-0A vsan 1
        # pwwn 50:06:0e:80:10:49:d5:70
        # pwwn 50:0a:09:82:00:01:dd:ec [V3070-rtp01-0a]
        # zone name V3070-02-0A-HDS-CTRL0-0A vsan 1
        # pwwn 50:0a:09:82:00:01:91:2a [V3070-rtp02-0a]
        # pwwn 50:06:0e:80:10:49:d5:71
        # zone name Evan-FCoE vsan 1
        # fwwn 20:22:00:0d:ec:a3:06:bf
        # fwwn 20:23:00:0d:ec:a3:06:bf
        # fwwn 20:26:00:0d:ec:a3:06:bf
        # fwwn 20:27:00:0d:ec:a3:06:bf
        # interface fc3/2 swwn 20:00:00:0d:ec:a3:06:80
        # interface fc3/3 swwn 20:00:00:0d:ec:a3:06:80
        # interface fc3/4 swwn 20:00:00:0d:ec:a3:06:80
        # pwwn 10:00:00:00:c9:66:26:7c [hpdl380g5-rtp05]
        # pwwn 21:01:00:1b:32:20:5a:0c [hpdl380g5-rtp06]
        # pwwn 10:00:00:00:c9:66:24:cc [hpdl380g5-rtp07]
        # pwwn 10:00:00:00:c9:66:1c:58
        # pwwn 10:00:00:00:c9:76:ec:09 [hpdl380g5-rtp08]
        # pwwn 10:00:00:00:c9:76:ec:08
        # pwwn 10:00:00:00:c9:66:25:cc
        # pwwn 21:00:00:1b:32:00:5a:0c
        # fwwn 20:1c:00:0d:ec:a3:06:bf
        # fwwn 20:1d:00:0d:ec:a3:06:bf
        # pwwn 10:00:00:00:c9:46:d5:38
        # pwwn 10:00:00:00:c9:43:21:a2
        if (/^zoneset name (\S+) vsan/) {

            # We have both the zoneset name and zones on the same line
            # as in cfg: zoneset zone1; zone2
            $zoneset = $1;
            $zoneset_ref->{$zoneset} = {};
        } elsif (/^zone name (\S+) vsan/) {

            # We have a zone
            my $zone = $1;

            # If $zoneset and $zoneset_ref->{$zoneset} are both defined,
            # then we can create an entry in $zoneset_ref for $zone.  If
            # not, then something is wrong so set an error and bail.
            if (defined($zoneset) and $zoneset ne "" and defined($zoneset_ref->{$zoneset})) {
                $zoneset_ref->{$zoneset}->{$zone} = $SwitchUtil::TRUE;
            } else {
                $self->_set_err("Unable to determine zoneset name");
                return {};
            }
        } ## end elsif (/^zone name (\S+) vsan/)
    } ## end foreach (@input)

    # Return the zoneset hash reference.  Format of the hash reference will be
    # $zoneset_ref->{$zoneset}->{$zone} = $SwitchUtil::TRUE;
    return $zoneset_ref;
} ## end sub _decode_zonesets

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

San::SanSwitch::FC::ZoneMgr::CiscoFCZone.pm

=head1 Author

Derrick Shultz

=head1 Description

Provides an abstract API interface to Alias, Zone, and ZoneSet operations on Cisco switches.

=head1 See Also

Package inherits from the following modules. See modules for interface detail

=over 4

=item * B<San::SanSwitch::ZoneMgr.pm>

=item * B<San::SanLog.pm>

=item * B<Error.pm>

=back

=head1 Attributes

=head2 B<active_zoneset>

Object of the active zoneset.  Defaults to undef.

=head2 B<aliases>

Reference to a hash of alias objects with a key of the alias name and the value of the alias object.  Defaults to {}.

=head2 B<parent>

ZoneMgr object = Reference to a parent SanSwitch object

=head2 B<timeout>

String containing an optional command timeout.  Default is 120.

=head2 B<vsan>

String containing an optional vsan id.

=head2 B<zones>

Reference to a hash of zone objects with a key of the zone name and the value of the zone object.  Defaults to {}.

=head2 B<zonesets>

Reference to a hash of zoneset objects with a key of the zoneset name and the value of the zoneset object.  Defaults to {}.

=head1 Public Methods

=head2 B<active_zoneset_discover()>

Performs discovery of the active zoneset on the switch and populates the $self->{active_zoneset} attribute
with the associated zoneset object.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->active_zoneset_discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to discover the active zoneset");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<alias_discover()>

Performs alias discovery of all aliases on the switch and creates objects for each discovered alias.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->alias_discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to perform alias discovery");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<config_disable()>

Disables active Config/Zoneset on the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->config_disable();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to disable the active zoneset");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<config_enable()>

Enables a config/zoneset on the switch

=head3 Inputs

=over 4

=item B<-zoneset>

String containing the name of the zoneset to enable

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->config_enable(-zoneset => $zoneset);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to enable zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<config_save()>

Saves the current config

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->config_save();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to save the configuration");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<new_alias()>

Creates a new alias object by calling San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Alias->new() and adds
it to the ZoneMgr object alias cache.  Blindly passes all args through to
San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Alias->new()

=head3 Inputs

See San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Alias->new()

=head3 Outputs

Returns an alias object. If an error occurs, the error flag is set on $self and errmsg is
set to the failure prior to returning the object.

=head3 Examples

    my $aliasobj = $self->new_alias(-alias => $alias_name, -member => $alias_member);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<new_zone()>

Creates a new zone object by calling San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Zone->new() and adds
it to the ZoneMgr object zone cache.  Blindly passes all args through to
San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Zone->new()

=head3 Inputs

See San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Zone->new()

=head3 Outputs

Returns a zone object. If an error occurs, the error flag is set on $self and errmsg is
set to the failure prior to returning the object.

=head3 Examples

    my $zoneobj = $self->new_zone(-zone => $zone_name, -aliases => \@aliases);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<new_zoneset()>

Creates a new zoneset object by calling San::SanSwitch::FC::ZoneMgr::CiscoFCZone::ZoneSet->new() and adds
it to the ZoneMgr object zone cache.  Blindly passes all args through to
San::SanSwitch::FC::ZoneMgr::CiscoFCZone::ZoneSet->new()

=head3 Inputs

See San::SanSwitch::FC::ZoneMgr::CiscoFCZone::ZoneSet->new()

=head3 Outputs

Returns a zoneset object. If an error occurs, the error flag is set on $self and errmsg is
set to the failure prior to returning the object.

=head3 Examples

    my $zonesetobj = $self->new_zoneset(-zoneset => $zoneset_name, -zones => \@zones);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zone_discover()>

Performs zone discovery of all zones on the switch and creates objects for each discovered zone.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zone_discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to perform zone discovery");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zoneset_discover()>

Performs zoneset discovery of all zonesets on the switch and creates objects for each discovered zoneset.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zoneset_discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to perform zoneset discovery");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head1 Private Methods

=head2 B<_decode_aliases()>

Take output from "show fcalias vsan $vsan" and decode it into a hash of aliases keyed by the alias name with a value of
$Switchutil::TRUE.

=head3 Inputs

An array containing output from the "show fcalias vsan X" command.

=head3 Outputs

On success, returns a hash reference of aliases keyed by the alias name with a value of
$Switchutil::TRUE.  If an error occurs, the error flag is set, errmsg is
set to the failure reason and {} is returned.  If there isn't an error but there aren't
any aliases found in the output, then {} is returned.

=head3 Examples

    my $cmd = "show fcalias vsan 2"
    my @output = $switch_obj->brocade_execute($cmd);
    my $alias_ref = $self->_decode_aliases(@output);
    if ($switch_obj->error) {
        $self->_set_err("There was an issue trying to decode the aliases:\n" . $switch_obj->errmsg);
        return $SwitchUtil::FALSE;
    }

=head2 B<_decode_zones()>

Take output from zoneshow and decode it into a hash of zones keyed by the zone name with a value of
a reference to a hash of aliases with a key of the alias name and the value of $Switchutil::TRUE.

=head3 Inputs

An array containing output from the zoneshow command.

=head3 Outputs

On success, returns a reference to a hash of zones keyed by the zone name with a value of
a reference to a hash of aliases with a key of the alias name and the value of $Switchutil::TRUE.
If an error occurs, the error flag is set, errmsg is set to the failure reason and {} is returned.
If there isn't an error but there aren't any zones found in the output, then {} is returned.

=head3 Examples

    my $cmd = "zoneshow *,0";
    my @output = $switch_obj->brocade_execute($cmd);
    my $zone_ref = $self->_decode_zones(@output);
    if ($switch_obj->error) {
        $self->_set_err("There was an issue trying to decode the zones:\n" . $switch_obj->errmsg);
        return $SwitchUtil::FALSE;
    }

=head2 B<_decode_zonesets()>

Take output from cfgshow and decode it into a hash of zonesets keyed by the zoneset name with a value of
a reference to a hash of zones with a key of the zone name and the value of $Switchutil::TRUE.

=head3 Inputs

An array containing output from the cfgshow command.

=head3 Outputs

On success, returns a reference to a hash of zonesets keyed by the zoneset name with a value of
a reference to a hash of zones with a key of the zone name and the value of $Switchutil::TRUE.
If an error occurs, the error flag is set, errmsg is set to the failure reason and {} is returned.
If there isn't an error but there aren't any zonesets found in the output, then {} is returned.

=head3 Examples

    my $cmd = "cfgshow *,0";
    my @output = $switch_obj->brocade_execute($cmd);
    my $zoneset_ref = $self->_decode_zonesets(@output);
    if ($switch_obj->error) {
        $self->_set_err("There was an issue trying to decode the zonesets:\n" . $switch_obj->errmsg);
        return $SwitchUtil::FALSE;
    }

=head2 B<_new()>

Instantiates the Cisco ZoneMgr object.

=head3 Inputs

=over 4

=item B<-parent>

Parent Switch object.

=item B<-timeout (optional)>

String containing the timeout to use for commands.  Defaults to 120.

=item B<-vsan (optional)>

String containing the vsan ID.  Defaults to undef.

=back

=head3 Outputs

Returns a ZoneMgr object. If an error occurs, the error flag is set and errmsg is
set to the failure reason before the object is returned.

=head3 Examples

    my $object = San::SanSwitch::FC::ZoneMgr::CiscoFCZone::_new(-parent => $self,
                                                                -timeout => "60");
    if ($object->error) {
        $self->_set_err($object->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=cut
