# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to alias related switch commands
## @author derricks@netapp.com
## @status private
## @pod here
## @unit_test thpl p4/test/lib/UnitTest/test_San_SanSwitch_ZoneMgr.thpl

package San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Alias;

# Compiler directives.
use strict;

# Module import.
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    create
# Objective: create the alias on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the alias object has a member defined since we can't
    # create an alias on the switch without that information.
    my $member = $self->get_member();
    if (not defined($member) or $member eq "") {
        $self->_set_err("Alias object does not currently have a member defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the alias object has a type defined since we can't
    # create an alias on the switch without that information.
    my $type = $self->get_type();
    if (not defined($type) or $type eq "") {
        $self->_set_err("Alias object does not currently have a type defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Run the commands to delete an alias
    my @cmds = ("conf t", "fcalias name $alias vsan $vsan", "member $type $member", "end");

    foreach my $cmd (@cmds) {
        my @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $switch_obj->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    } ## end foreach my $cmd (@cmds)

    return $SwitchUtil::TRUE;
} ## end sub create

###############################################################################
# Method:    destroy
# Objective: destroy the alias on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Run the commands to delete an alias
    my @cmds = ("conf t", "no fcalias name $alias vsan $vsan", "end");

    foreach my $cmd (@cmds) {
        my @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $self->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    } ## end foreach my $cmd (@cmds)

    return $SwitchUtil::TRUE;
} ## end sub destroy

###############################################################################
# Method:    remove_member
# Objective: remove a member from the alias on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_member {
    my ($self, $member) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the alias object has a type defined since we can't
    # create an alias on the switch without that information.
    my $type = $self->get_type();
    if (not defined($type) or $type eq "") {
        $self->_set_err("Alias object does not currently have a type defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Run the commands to remove the member from the alias
    my @cmds = ("conf t", "fcalias name $alias vsan $vsan", "no member $type $member", "end");

    foreach my $cmd (@cmds) {
        my @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $switch_obj->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    } ## end foreach my $cmd (@cmds)

    # clear the member field (XXX we really need to maintain an array of members here)
    if ($self->{member} and $self->{member} eq $member) {
        delete($self->{member});
    }

    return $SwitchUtil::TRUE;
} ## end sub remove_member

###############################################################################
# Method:    add_member
# Objective: add a member from the alias on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_member {
    my ($self, $member) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the alias object has a type defined since we can't
    # create an alias on the switch without that information.
    my $type = $self->get_type();
    if (not defined($type) or $type eq "") {
        $self->_set_err("Alias object does not currently have a type defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Run the commands to add the member to the alias
    my @cmds = ("conf t", "fcalias name $alias vsan $vsan", "member $type $member", "end");

    foreach my $cmd (@cmds) {
        my @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $switch_obj->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    } ## end foreach my $cmd (@cmds)

    # update the member field
    # XXX we really need to maintain an array of members here
    $self->{member} = $member;

    return $SwitchUtil::TRUE;
} ## end sub add_member

###############################################################################
# Method:    validate
# Objective: validate the alias on the switch against the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the current alias object actually has a member defined
    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the current alias object actually has a member defined
    my $member = $self->get_member();
    if (not defined($member) or $member eq "") {
        $self->_set_err("Alias object does not currently have a member defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the current alias object actually has a type defined
    my $type = $self->get_type();
    if (not defined($type) or $type eq "") {
        $self->_set_err("Alias object does not currently have a type defined");
        return $SwitchUtil::FALSE;
    }

    # Get the alias information from the switch
    my @output = $self->_query_alias();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "not present" error
    if (grep(/not present/, @output) > 0) {
        $self->_set_err("Alias $alias does not exist on the switch");
        return $SwitchUtil::FALSE;
    }

    # Decode the alias information into a hash reference
    my $alias_ref = $self->{parent}->_decode_aliases(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to decode aliases:\n", $self->{parent}->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we actually got back a hash
    if (ref($alias_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return $SwitchUtil::FALSE;
    }

    # Check to see if the switch member is the same as the defined member
    if (defined($alias_ref->{$alias})) {

        # Check the member on the object against the switch member
        if (defined($alias_ref->{$alias}->{member}) and $alias_ref->{$alias}->{member} ne "") {
            if ($member ne $alias_ref->{$alias}->{member}) {
                return $SwitchUtil::FALSE;
            }
        } else {
            $self->_set_err("Member information for alias $alias was blank");
            return $SwitchUtil::FALSE;
        }

        # Check the type on the object against the switch type
        if (defined($alias_ref->{$alias}->{type}) and $alias_ref->{$alias}->{type} ne "") {
            if ($type ne $alias_ref->{$alias}->{type}) {
                return $SwitchUtil::FALSE;
            }
        } else {
            $self->_set_err("Type information for alias $alias was blank");
            return $SwitchUtil::FALSE;
        }
    } else {

        # the alias wasn't found on the switch hash so return false
        return $SwitchUtil::FALSE;
    }

    # Everything looks good so return TRUE
    return $SwitchUtil::TRUE;
} ## end sub validate

###############################################################################
# Method:    exists
# Objective: Check whether the alias exists on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub exists {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Get the alias information from the switch
    my @output = $self->_query_alias();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "not present" error
    if (grep(/not present/, @output) > 0) {
        return $SwitchUtil::FALSE;
    }

    # Decode the alias information into a hash reference
    my $alias_ref = $self->{parent}->_decode_aliases(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to decode aliases:\n", $self->{parent}->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we actually got back a hash
    if (ref($alias_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return $SwitchUtil::FALSE;
    }

    # Double check that the alias you are looking for is found as a
    # key to the hash reference
    if (defined($alias_ref->{$alias})) {
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
} ## end sub exists

###############################################################################
# Method:    get_member
# Objective: Return the member information from the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_member {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{member})) {
        $self->_set_err("Unable to retrieve the 'member' key off of the object");
        return "";
    }

    return $self->{member};
}

###############################################################################
# Method:    get_type
# Objective: Return the type information from the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_type {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{type})) {
        $self->_set_err("Unable to retrieve the 'type' key off of the object");
        return "";
    }

    return $self->{type};
}

###############################################################################
# Method:    get_name
# Objective: Return the alias name from the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{alias})) {
        $self->_set_err("Unable to retrieve the 'alias' key off of the object");
        return "";
    }
    return $self->{alias};
}

###############################################################################
# Method:    discover
# Objective: Discover the member information for the alias
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return (undef, undef);
    }

    # Get the alias information from the switch
    my @output = $self->_query_alias();
    if ($self->error) {
        return (undef, undef);
    }

    # Check to make sure we didn't get a "not present" error
    if (grep(/not present/, @output) > 0) {
        $self->_set_err("Alias $alias does not exist on the switch");
        return (undef, undef);
    }

    # Decode the alias information into a hash reference
    my $alias_ref = $self->{parent}->_decode_aliases(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to decode aliases:\n", $self->{parent}->errmsg());
        return (undef, undef);
    }

    # Check to make sure we actually got back a hash
    if (ref($alias_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return (undef, undef);
    }

    # Check to see if the alias exists and the member info is populated
    if (defined($alias_ref->{$alias})) {

        # Check for the member info
        if (not defined($alias_ref->{$alias}->{member}) or $alias_ref->{$alias}->{member} eq "") {
            $self->_set_err("Error discovering alias.  Member information is blank");
            return (undef, undef);
        }

        # Check for the type info
        if (not defined($alias_ref->{$alias}->{type}) or $alias_ref->{$alias}->{type} eq "") {
            $self->_set_err("Error discovering alias.  Type information is blank");
            return (undef, undef);
        }
    } else {
        $self->_set_err("Alias $alias was not found on the switch");
        return (undef, undef);
    }

    return ($alias_ref->{$alias}->{member}, $alias_ref->{$alias}->{type});
} ## end sub discover

###############################################################################
# Method:    is_on_switch
# Objective: returns the "on_switch" value off of the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_on_switch {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{on_switch}) and $self->{on_switch} ne "") {
        $self->_set_err("Invalid value for 'on_switch' key on object");
        return $SwitchUtil::FALSE;
    }

    return $self->{on_switch};
}

###############################################################################
# Method:    new
# Objective: Create new Alias object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, $parent, $alias, $member, $type);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?alias/) {
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '-alias'");
                return $self;
            }
            next;
        } elsif (/-?member/) {
            if (not length($member = $args{$_})) {
                $self->_set_err("Invalid value for argument '-member'");
                return $self;
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return $self;
            }
            next;
        } ## end elsif (/-?type/)
    } ## end foreach (keys %args)

    # Verify mandatory parameters.
    if (defined($parent) and $parent ne "") {
        unless (ref($parent) =~ /CiscoFCZone/i) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Mandatory parameter \"alias\"", "not specified");
        return $self;
    }

    # Verify the optional parameters
    if (not defined($member)) {
        $member = "";
    }

    if (not defined($type)) {
        $type = "";
    }

    # initialize the object members
    $self->{parent}    = $parent;
    $self->{alias}     = $alias;
    $self->{member}    = $member;
    $self->{type}      = $type;
    $self->{on_switch} = $SwitchUtil::FALSE;

    return $self;
} ## end sub new

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _query_alias
# Objective: Query the switch and return the alias information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _query_alias {
    my ($self) = @_;

    $self->_clear_err();

    my $alias = $self->get_name();
    if (not defined($alias) or $alias eq "") {
        $self->_set_err("Alias object does not currently have a name defined");
        return ();
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Run "show fcalias" on the switch with the alias name
    # Format of command is "show fcalias"
    my $cmd = "show fcalias name $alias vsan $vsan";
    my @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return ();
    }

    return @output;
} ## end sub _query_alias

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Alias.pm

=head1 Author

Derrick Shultz

=head1 Description

Provides an abstract API interface to Alias operations on Cisco switches.

=head1 See Also

Package inherits from the following modules. See modules for interface detail

=over 4

=item * B<San::SanLog.pm>

=item * B<Error.pm>

=back

=head1 Attributes

=head2 B<alias>

String containing name of alias

=head2 B<member>

String containing alias member information

=head2 B<parent>

ZoneMgr object = Reference to a parent ZoneMgr object

=head2 B<on_switch>

Boolean of $SwitchUtil::TRUE or $SwitchUtil::FALSE representing whether or not the alias is currently found on
the switch.

=head2 B<type>

String containing alias type information

=head1 Public Methods

=head2 B<create()>

Creates the alias on the switch.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->create();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<destroy()>

Removes the alias from the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->destroy();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<remove_member()>

Remove a member from this alias

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->remove_member($wwpn);
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<add_member()>

Add a member to this alias

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->add_member($wwpn);
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<discover()>

Discovers the member and type information for the alias on the switch

=head3 Inputs

None

=head3 Outputs

Returns ($member, $type) on success.  If an error occurs, the error flag is set and errmsg is
set to the failure reason and (undef, undef) is returned.

=head3 Examples

    my ($member, $type) = $aliasobj->discover();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    }

=head2 B<exists()>

Checks the existance of the alias on the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned. If there wasn't an error, but the alias
doesn't actually exist it returns $SwitchUtil::FALSE without setting an error.

=head3 Examples

    my $status = $aliasobj->exists();
    if ($status == $SwitchUtil::FALSE) {
        if ($aliasobj->error) {
            $self->_set_err($aliasobj->errmsg());
            return $SwitchUtil::FALSE;
        } else {
            return $SwitchUtil::FALSE;
        }
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<get_member()>

Returns the member string from the alias object.

=head3 Inputs

None

=head3 Outputs

Returns the member string from the alias object.

=head3 Examples

    my $member = $aliasobj->get_member();

=head2 B<get_name()>

Returns the name string from the alias object.

=head3 Inputs

None

=head3 Outputs

Returns the name string from the alias object.

=head3 Examples

    my $name = $aliasobj->get_name();

=head2 B<get_type()>

Returns the type string from the alias object.

=head3 Inputs

None

=head3 Outputs

Returns the type string from the alias object.

=head3 Examples

    my $type = $aliasobj->get_type();

=head2 B<is_on_switch()>

Method to return the on_switch key from the object

=head3 Inputs

None

=head3 Outputs

Returns a boolean based on the value of the on_switch key. If the key doesn't exist or is empty, it sets an error.

=head3 Examples

    if ($obj->is_on_switch()) {
        do stuff here;
    }

=head2 B<new()>

Instantiates the alias object.

=head3 Inputs

=over 4

=item B<-alias>

String containing the name of the alias.

=item B<-member (optional)>

String containing the member information.

=item B<-parent>

Parent ZoneMgr object

=item B<-type (optional)>

String containing the type information.

=back

=head3 Outputs

Returns an alias object. If an error occurs, the error flag is set and errmsg is
set to the failure reason before the object is returned.

=head3 Examples

    my $aliasobj = San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Alias::new(-parent => $self,
                                                                        -alias => $alias_name,
                                                                        -member => $member,
                                                                        -type => $type);
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<validate()>

Validates the alias on the switch against the object

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $aliasobj->validate();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head1 Private Methods

=head2 B<_query_alias()>

Runs "show fcalias <alias>" on the switch and returns the output from the command.

=head3 Inputs

None

=head3 Outputs

On success, returns an array containing the "show fcalias" output. If an error occurs, the error is
set to the failure reason and () is returned.

=head3 Examples

    my @query_output = $aliasobj->_query_zone();
    if ($aliasobj->error) {
        $self->_set_err($aliasobj->errmsg());
        return $SwitchUtil::FALSE;
    }

=cut
