# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to zone related switch commands
## @author derricks@netapp.com
## @status private
## @pod here
## @unit_test thpl p4/test/lib/UnitTest/test_San_SanSwitch_ZoneMgr.thpl

package San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Zone;

# Compiler directives.
use strict;
use warnings;

# Module import.
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    create
# Objective: create the zone on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zone object has a defined name
    my $zone = $self->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Check to make sure the object has alias defined
    my $aliases_ref = $self->get_aliases();
    my $wwpns_ref = $self->get_wwpns();
    if (!scalar(keys %{$aliases_ref}) and !scalar(@{$wwpns_ref})) {
        $self->_set_err("Cannot create zone, because the Zone object " .
                        "'$zone' does not currently have any " .
                        "aliases or wwpns associate to it");
        return $SwitchUtil::FALSE;
    }

    # Define the commands to run
    my @cmds = ("conf t",
                "zone name $zone vsan $vsan");

    # Add the alias commands to the @cmds array
    foreach my $alias (sort keys %${aliases_ref}) {
        push(@cmds, "member fcalias $alias");
    }

    # add all wwpn commands to te array
    foreach my $wwpn (@{$wwpns_ref}) {
        push(@cmds, "member pwwn $wwpn");
    }

    # Add the "end" command to @cmds
    push(@cmds, "end");

    # Execute the commands on the switch
    foreach my $cmd (@cmds) {
        my @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $switch_obj->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    } ## end foreach my $cmd (@cmds)

    return $SwitchUtil::TRUE;
} ## end sub create

###############################################################################
# Method:    destroy
# Objective: destroy the zone on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zone object has a defined name
    my $zone = $self->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Define the commands to run
    my @cmds = ("conf t",
                "no zone name $zone vsan $vsan",
                "end");

    # Execute the commands
    foreach my $cmd (@cmds) {
        my @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $self->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    } ## end foreach my $cmd (@cmds)

    return $SwitchUtil::TRUE;
} ## end sub destroy

###############################################################################
# Method:    validate
# Objective: validate the zone on the switch against the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zone object has a defined name
    my $zone = $self->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the current zone object actually has an alias defined
    my $aliasobj_ref = $self->get_aliases();
    unless (scalar(keys %{$aliasobj_ref}) > 0) {
        $self->_set_err("Zone object does not currently have any aliases defined");
        return $SwitchUtil::FALSE;
    }

    # Since get_aliases returns an hash reference of alias objects keyed by the
    # alias name, we can run keys on the reference then push that into the @aliases array.
    my @aliases = ();
    foreach my $t_alias (sort keys %{$aliasobj_ref}) {
        push(@aliases, $t_alias);
    }

    # Get the zone information from the switch
    my @output = $self->_query_zone();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "not present" error
    if (grep(/not present/, @output) > 0) {
        $self->_set_err("Zone $zone not present on the switch");
        return $SwitchUtil::FALSE;
    }

    # Decode the switch output into a hash reference
    my $switch_zone_ref = $self->{parent}->_decode_zones(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to decode zones:\n", $self->{parent}->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we actually got back a hash
    if (ref($switch_zone_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return $SwitchUtil::FALSE;
    }

    # Double check that the zone we are looking for is found as a
    # key to the hash reference
    my @switch_aliases = ();
    if (defined($switch_zone_ref->{$zone})) {

        # Since the value of the zone hash reference key $zone is
        # actually a reference to an hash of alias names, we need to iterate over
        # the array reference and then push that into the @switch_aliases array.
        foreach my $t_alias (sort keys %{$switch_zone_ref->{$zone}}) {
            push(@switch_aliases, $t_alias);
        }
    } else {
        $self->_set_err("Zone information returned from the switch does not match the object Zone name");
        return $SwitchUtil::FALSE;
    }

    # compare the value defined for @aliases with the alias list that
    # was returned and stored in @switch_aliases.
    # We will store any missing aliases in the array @missing_aliases;
    my @missing_aliases = ();
    foreach my $alias (sort @aliases) {
        my $found = undef;
        foreach my $switch_alias (sort @switch_aliases) {
            if ($switch_alias eq $alias) {
                $found = $SwitchUtil::TRUE;
                last;
            }
        } ## end foreach my $switch_alias (sort...
        if (not defined($found)) {

            # We didn't find $alias so save that info.
            push(@missing_aliases, $alias);
        }
    } ## end foreach my $alias (sort @aliases)
    if (scalar(@missing_aliases) > 0) {
        $self->_set_err("Unable to validate zone $zone. The switch zone is missing aliases @missing_aliases.");
        return $SwitchUtil::FALSE;
    }

    # Do a check in the other direction.  Check to make sure we don't have extra aliases
    # defined in the zone on the switch versus what is defined on the object
    my @extra_aliases = ();
    foreach my $switch_alias (sort @switch_aliases) {
        my $found = undef;
        foreach my $alias (sort @aliases) {
            if ($switch_alias eq $alias) {
                $found = $SwitchUtil::TRUE;
                last;
            }
        } ## end foreach my $alias (sort @aliases)
        if (not defined($found)) {

            # We didn't find $alias so save that info.
            push(@extra_aliases, $switch_alias);
        }
    } ## end foreach my $switch_alias (sort...
    if (scalar(@extra_aliases) > 0) {
        $self->_set_err("Unable to validate zone $zone. The switch zone contains the extra aliases @extra_aliases.");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub validate

###############################################################################
# Method:    exists
# Objective: Check whether the zone exists on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub exists {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zone object has a defined name
    my $zone = $self->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Get the zone information from the switch
    my @output = $self->_query_zone();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we didn't get a "not present" error
    if (grep(/not present/, @output) > 0) {
        return $SwitchUtil::FALSE;
    }

    # Decode the zone information into a hash reference
    my $zone_ref = $self->{parent}->_decode_zones(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to decode zones:\n", $self->{parent}->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we actually got back a hash
    if (ref($zone_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return $SwitchUtil::FALSE;
    }

    # Double check that the zone you are looking for is found as a
    # key to the hash reference.
    if (not defined($zone_ref->{$zone})) {
        # Zone wasn't found on the switch
        return $SwitchUtil::FALSE;
    }

    # If we made it this far then we are good and found the zone on the switch
    return $SwitchUtil::TRUE;
} ## end sub exists

###############################################################################
# Method:    is_member
# Objective: Check whether the zone contains a specified member alias/wwpn
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_member {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $alias = delete($args{alias});
    my $wwpn = delete($args{wwpn});

    if ($wwpn) {
        my $wwpns = $self->get_wwpns();
        if (scalar(@{$wwpns})) {
            return $SwitchUtil::TRUE if (grep(/^$wwpn$/i, @{$wwpns}));
        }
    } elsif ($alias) {
        my $aliases = $self->get_aliases();
        if (scalar(keys %{$aliases})) {
            return $SwitchUtil::TRUE if (CORE::exists($aliases->{$alias}));
        }
    } else {
        $self->_set_err("Missing required parameters 'alias' or 'wwpn'");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::FALSE;
} ## end sub is_member

##############################################################################
# Method:    get_aliases
# Objective: Return the alias information from the object as a hash reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_aliases {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{aliases})) {
        $self->_set_err("Unable to retrieve the aliases key off of the object");
        return {};
    }
    return $self->{aliases};
}

##############################################################################
# Method:    get_wwpns
# Objective: Return the wwpn information from the object as a hash reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_wwpns {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{wwpns})) {
        $self->_set_err("Unable to retrieve the wwpns key off of the object");
        return {};
    }
    return $self->{wwpns};
}

###############################################################################
# Method:    get_name
# Objective: Return the zone name from the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{zone})) {
        $self->_set_err("Unable to retrieve the 'zone' key off of the object");
        return "";
    }
    return $self->{zone};
}

###############################################################################
# Method:    add_alias
# Objective: Add an alias to the zone on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_alias {
    my ($self, %args) = @_;

    my ($aliasobj);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias object
            if (not length($aliasobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($aliasobj)) {
        unless (ref($aliasobj) =~ /Alias/) {
            $self->_set_err("Invalid value for argument '-alias'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    # Verify that the alias object has a defined name
    my $alias = $aliasobj->get_name();
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Alias object provided does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the zone object has a defined name
    my $zone = $self->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Define the commands to run
    my @cmds = ("conf t",
                "zone name $zone vsan $vsan",
                "member fcalias $alias",
                "end");

    # Execute the commands
    my @output;
    foreach my $cmd (@cmds) {
        @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $self->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    } ## end foreach my $cmd (@cmds)

    # Check for an "Duplicate member" error
    if (grep(/duplicate member/i, @output) > 0) {

        # Zone already contains the alias
        $self->_set_err("Zone $zone already contains alias $alias");
        return $SwitchUtil::FALSE;
    } ## end if (grep(/already contains/...

    # Add the new alias object to the alias objects on the zone
    $self->{aliases}->{$alias} = $aliasobj;

    return $SwitchUtil::TRUE;
} ## end sub add_alias

###############################################################################
# Method:    remove_alias
# Objective: Remove an alias from the zone on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_alias {
    my ($self, %args) = @_;

    my ($aliasobj);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias object
            if (not length($aliasobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($aliasobj)) {
        unless (ref($aliasobj) =~ /Alias/) {
            $self->_set_err("Invalid value for argument '-alias'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    # Verify that the alias object has a defined name
    my $alias = $aliasobj->get_name();
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Alias object provided does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Verify that the zone object has a defined name
    my $zone = $self->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Define the commands to run
    my @cmds = ("conf t",
                "zone name $zone vsan $vsan",
                "no member fcalias $alias",
                "end");

    # Execute the commands
    my @output;
    foreach my $cmd (@cmds) {
        @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $self->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    }

    # Check to see if we got a "not present" error on the remove
    if (grep(/not present/i, @output) > 0) {

        # Unable to find alias in zone
        $self->_set_err("Alias $alias was not found in zone $zone");
        return $SwitchUtil::FALSE;
    } ## end if (grep(/is not in/, ...

    # Remove the alias from the zone object
    if (defined($self->{aliases}->{$alias})) {
        unless (delete($self->{aliases}->{$alias})) {
            $self->_set_err("Unable to remove alias $alias from zone object for zone $zone");
            return $SwitchUtil::FALSE;
        }
    } ## end if (defined($self->{aliases...

    return $SwitchUtil::TRUE;
} ## end sub remove_alias

###############################################################################
# Method:    add_wwpn
# Objective: Add a wwpn to this zone
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_wwpn {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $wwpn = delete($args{wwpn}) // $self->_set_err("Missing required parameter 'wwpn'");
    if ($self->error()) {
        return $SwitchUtil::FALSE;
    }

    # Verify that the zone object has a defined name
    my $zone = $self->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Define the commands to run
    my @cmds = ("conf t",
                "zone name $zone vsan $vsan",
                "member pwwn $wwpn",
                "end");

    # Execute the commands
    my @output;
    foreach my $cmd (@cmds) {
        @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $self->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    } ## end foreach my $cmd (@cmds)

    # Check for an "Duplicate member" error
    if (grep(/duplicate member/i, @output) > 0) {
        # Zone already contains the wwpn
        $self->_set_err("Zone $zone already contains wwpn $wwpn");
        return $SwitchUtil::FALSE;
    }

    # store the wwpn on the zone objects wwpn array
    push(@{$self->{wwpns}}, $wwpn);

    return $SwitchUtil::TRUE;
} ## end sub add_wwpn

###############################################################################
# Method:    remove_wwpn
# Objective: Remove a wwpn from this zone
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_wwpn {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $wwpn = delete($args{wwpn}) // $self->_set_err("Missing required parameter 'wwpn'");
    if ($self->error()) {
        return $SwitchUtil::FALSE;
    }

    # Verify that the zone object has a defined name
    my $zone = $self->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Define the commands to run
    my @cmds = ("conf t",
                "zone name $zone vsan $vsan",
                "no member pwwn $wwpn",
                "end");

    # Execute the commands
    my @output;
    foreach my $cmd (@cmds) {
        @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
        if ($switch_obj->error) {
            $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());

            # handle error, by attempting to get out of config mode
            # and then returning with the proper error set.
            @cmds = (" ", "end");
            foreach my $cmd (@cmds) {
                @output = $self->execute($cmd);
            }
            return $SwitchUtil::FALSE;
        } ## end if ($switch_obj->error)
    } ## end foreach my $cmd (@cmds)

    # Check to see if we got a "not present" error on the remove
    if (grep(/not present/i, @output) > 0) {
        # Unable to find alias in zone
        $self->_set_err("Alias $wwpn was not found in zone $zone");
        return $SwitchUtil::FALSE;
    } ## end if (grep(/is not in/, ...

    # get rid of this wwpn from the wwpn array
    @{$self->{wwpns}} = grep { !/^$wwpn$/i } @{$self->{wwpns}};

    return $SwitchUtil::TRUE;
} ## end sub remove_wwpn

###############################################################################
# Method:    verify_alias
# Objective: Check that an alias exists in the zone on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_alias {
    my ($self, %args) = @_;

    my ($aliasobj);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias object
            if (not length($aliasobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($aliasobj)) {
        unless (ref($aliasobj) =~ /Alias/) {
            $self->_set_err("Invalid value for argument '-alias'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    # Verify that the alias object has a defined name
    my $alias = $aliasobj->get_name();
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Alias object provided does not have a name defined");
        return $SwitchUtil::FALSE;
    }

    # Discover the aliases in the zone.  These are names only, not objects
    my $alias_ref = $self->discover();
    if ($self->error) {
        $self->_set_error("Error discovering zone information:\n" . $self->errmsg());
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we actually got back a hash
    if (ref($alias_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we got back aliases
    unless (scalar(keys %{$alias_ref}) > 0) {
        $self->_set_err("No aliases were returned from the switch for zone " . $self->get_name());
        return $SwitchUtil::FALSE;
    }

    # Check the alias hash reference to see if the $alias is one of the keys
    if (defined($alias_ref->{$alias})) {
        return $SwitchUtil::TRUE;
    } else {

        # We couldn't find the alias in the zone on the switch
        return $SwitchUtil::FALSE;
    }
} ## end sub verify_alias

###############################################################################
# Method:    discover
# Objective: Discover the alias information for the zone
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover {
    my ($self) = @_;

    $self->_clear_err();

    # Verify that the zone object has a defined name
    my $zone = $self->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Zone object does not have a name defined");
        return {};
    }

    # Get the zone information from the switch
    my @output = $self->_query_zone();
    if ($self->error) {
        return {};
    }

    # Check to make sure we didn't get a "not present" error
    if (grep(/not present/, @output) > 0) {
        $self->_set_err("Zone $zone not present on the switch");
        return {};
    }

    # Decode the alias information into a hash reference
    my $zone_ref = $self->{parent}->_decode_zones(@output);
    if ($self->{parent}->error) {
        $self->_set_err("Unable to decode zones:\n", $self->{parent}->errmsg());
        return {};
    }

    # Check to make sure we actually got back a hash
    if (ref($zone_ref) !~ /hash/i) {
        $self->_set_err("Invalid data type returned. Expected data type is HASH");
        return $SwitchUtil::FALSE;
    }

    # Make sure that $zone_ref->{$zone}  actually contains something.  We
    # should never end up in a situation where a zone isn't
    # defined with an alias, but check just in case.
    if (defined($zone_ref->{$zone}) and scalar(keys %{$zone_ref->{$zone}}) == 0) {
        $self->_set_err("Error discovering zone.  Alias information is blank");
        return {};
    }

    # Return the alias information
    return $zone_ref->{$zone};
} ## end sub discover

###############################################################################
# Method:    is_on_switch
# Objective: returns the "on_switch" value off of the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_on_switch {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{on_switch}) and $self->{on_switch} ne "") {
        $self->_set_err("Invalid value for 'on_switch' key on object");
        return $SwitchUtil::FALSE;
    }

    return $self->{on_switch};
}

###############################################################################
# Method:    new
# Objective: Create new zone object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, $parent, $zone, @aliases, @wwpns);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?zone/) {
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '-zone'");
                return $self;
            }
            next;
        } elsif (/-?aliases/) {
            unless (ref($args{$_}) =~ /ARRAY/ and scalar(@aliases = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-aliases'");
                return $self;
            }
            next;
        } elsif (/-?wwpns/) {
            unless (ref($args{$_}) =~ /ARRAY/) {
                $self->_set_err("Invalid value for argument '-wwpns'");
                return $self;
            }
            @wwpns = @{$args{$_}};
            next;
        }
    } ## end foreach (keys %args)

    # Verify mandatory parameters.
    if (defined($parent) and $parent ne "") {
        unless (ref($parent) =~ /CiscoFCZone/i) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Mandatory parameter \"zone\"", "not specified");
        return $self;
    }

    my $aliases_ref = {};

    # Verify the optional parameters

    # See if we were given aliases.  If we were, then get the name of
    # each alias and use it as a key in $aliases_ref with the value
    # being the alias object.
    if (scalar(@aliases > 0)) {
        foreach my $aliasobj (@aliases) {
            my $alias = $aliasobj->get_name();
            unless (defined($alias) and $alias ne "") {
                $self->_set_err("Unable to determine alias name");
                return $self;
            }
            $aliases_ref->{$alias} = $aliasobj;
        } ## end foreach my $aliasobj (@aliases)
    } ## end if (scalar(@aliases > ...

    # initialize the object aliases
    $self->{parent}    = $parent;
    $self->{zone}      = $zone;
    $self->{aliases}   = $aliases_ref;
    $self->{wwpns}     = \@wwpns;
    $self->{on_switch} = $SwitchUtil::FALSE;

    return $self;
} ## end sub new

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _query_zone
# Objective: Query the switch and return the zone information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _query_zone {
    my ($self) = @_;

    $self->_clear_err();

    my $zone       = $self->get_name();
    my $switch_obj = $self->{parent}->{parent};

    # Get the vsan id
    my $vsan;
    if (defined($self->{parent}->{vsan}) and $self->{parent}->{vsan} ne "") {
        $vsan = $self->{parent}->{vsan};
    } else {
        $self->_set_err("VSAN Id is not set on the ZoneMgr object");
        return $SwitchUtil::FALSE;
    }

    # Run "show zone" on the switch with the zone name
    # Format of "show zone" command is "show zone name <ZONE> vsan <VSAN>"
    my $cmd = "show zone name $zone vsan $vsan";
    my @output = $switch_obj->execute($cmd, $self->{parent}->{timeout});
    if ($switch_obj->error) {
        $self->_set_err("Error executing command ($cmd)" . $switch_obj->errmsg());
        return ();
    }

    return @output;
} ## end sub _query_zone

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Zone.pm

=head1 Author

Derrick Shultz

=head1 Description

Provides an abstract API interface to Zone operations on Cisco switches.

=head1 See Also

Package inherits from the following modules. See modules for interface detail

=over 4

=item * B<San::SanLog.pm>

=item * B<Error.pm>

=back

=head1 Attributes

=head2 B<parent>

ZoneMgr object = Reference to a parent ZoneMgr object

=head2 B<zone>

String containing name of the zone

=head2 B<aliases>

Reference to a hash containing alias objects with a key of the alias name and a value of the alias object.

=head2 B<on_switch>

Boolean of $SwitchUtil::TRUE or $SwitchUtil::FALSE representing whether or not the zone is currently found on
the switch.

=head1 Public Methods

=head2 B<is_member()>

Check if an alias or wwpn is a member of this zone.

=head3 Inputs

=over 4

=item B<-alias> xor B<-wwpn>

Alias string or WWPN string to check membership status

=back

=head3 Outputs

returns 0 or 1 (sets an error if something bad happened)

=head3 Examples

    if ($zoneobj->is_member(alias => $alias_string)) {
        # is a member
    } else {
        if ($zoneobj->error) {
            # handle error
        } else {
            # not a member
        }
    }

=head2 B<add_alias()>

Adds an alias to an already existing zone on a switch

=head3 Inputs

=over 4

=item B<-alias>

Alias object to add to the zone.

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zoneobj->add_alias(-alias => $aliasobj);
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<add_wwpn()>

Adds an wwpn to an already existing zone on a switch

=head3 Inputs

=over 4

=item B<-wwpn>

WWPN string to add to the zone.

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zoneobj->add_wwpn(-wwpn => $wwpn);
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<create()>

Creates the zone on the switch.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zoneobj->create();
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<destroy()>

Removes the zone from the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zoneobj->destroy();
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<discover()>

Discovers the alias information for the zone on the switch

=head3 Inputs

None

=head3 Outputs

Returns a reference to a hash containing alias names as the key and $SwitchUtil::TRUE as the value like
$zone_ref->{"alias1" => $SwitchUtil::TRUE, "alias2" => $SwitchUtil::TRUE}. If an error occurs, the error is
set to the failure reason and {} is returned.

=head3 Examples

    my $alias_ref = $zoneobj->discover();
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    }

=head2 B<exists()>

Checks for the existance of the zone on the switch

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error message is
set to the failure reason and $SwitchUtil::FALSE is returned. If there wasn't an error,
but the alias doesn't actually exist it returns $SwitchUtil::FALSE without setting an error.

=head3 Examples

    my $status = $zoneobj->exists();
    if ($status == $SwitchUtil::FALSE) {
        if ($zoneobj->error) {
            # We had a problem checking for the existance of the zone
            $self->_set_err($zoneobj->errmsg());
            return $SwitchUtil::FALSE;
        } else {
            # No problems were found. The Zone doesn't exist.
            return $SwitchUtil::FALSE;
        }
    } else {
        # Zone Exists
        return $SwitchUtil::TRUE;
    }

=head2 B<get_aliases()>

Method to get all of the alias objects in the zone

=head3 Inputs

None

=head3 Outputs

Returns a reference to a hash of alias objects keyed on the alias name with a value of the alias object
Returns the member string from the alias object.

=head3 Examples

    my $aliases_ref = $zoneobj->get_aliases();

=head2 B<get_name()>

Method for retrieving the name of the zone

=head3 Inputs

None

=head3 Outputs

Returns the name string from the zone object.

=head3 Examples

    my $name = $zoneobj->get_name();

=head2 B<is_on_switch()>

Method to return the on_switch key from the object

=head3 Inputs

None

=head3 Outputs

Returns a boolean based on the value of the on_switch key. If the key doesn't exist or is empty, it sets an error.

=head3 Examples

    if ($obj->is_on_switch()) {
        do stuff here;
    }

=head2 B<new()>

Instantiates the zone object.

=head3 Inputs

=over 4

=item B<-zone>

String containing the name of the zone.

=item B<-aliases (optional)>

Reference to an array containing alias objects

=item B<-parent>

Parent ZoneMgr object.

=back

=head3 Outputs

Returns a zone object. If an error occurs, the error flag is set and errmsg is
set to the failure reason before the object is returned.

=head3 Examples

    my $zoneobj = San::SanSwitch::FC::ZoneMgr::CiscoFCZone::Zone::new(-parent => $self,
                                                                      -zone => $zone_name,
                                                                      -aliases => \@aliases);
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<remove_alias()>

Removes an alias from an existing zone on a switch

=head3 Inputs

=over 4

=item B<-alias>

Alias object remove from the zone

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zoneobj->remove_alias(-alias => $aliasobj);
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<remove_wwpn()>

Removes an wwpn from an existing zone on a switch

=head3 Inputs

=over 4

=item B<-wwpn>

WWPN to remove from the zone

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zoneobj->remove_wwpn(-wwpn => $wwpn_string);
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<validate()>

Validates the zone on the switch against the object.  This validation includes making sure that the zone actually
has aliases defined on it and then makes sure that the zone matches exactly.  This means that all aliases defined
in the zone object match the aliases in the switch zone.  Any extra aliases in either the object or the switch will
cause a failure.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zoneobj->validate();
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<verify_alias()>

Checks the zone on the switch to make sure that is currently contains the alias specified.

=head3 Inputs

=over 4

=item B<-alias>

Alias object to verify in the zone.

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set, errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $zoneobj->verify_alias(-alias => $aliasobj);
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head1 Private Methods

=head2 B<_query_zone()>

Runs "show zone <zone>" on the switch and returns the output from the command.

=head3 Inputs

None

=head3 Outputs

On success, returns an array containing the "show zone" output. If an error occurs, the error is
set to the failure reason and () is returned.

=head3 Examples

    my @query_output = $zoneobj->_query_zone();
    if ($zoneobj->error) {
        $self->_set_err($zoneobj->errmsg());
        return $SwitchUtil::FALSE;
    }

=cut
