# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide a Port Manager interface to switches
## @author evan@netapp.com
## @status private
## @pod here

package PortMgr;

use strict;

use Error;
use San::SanCache;

use vars qw(@ISA);

@ISA = qw(Error SanLog SanCache);

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new port manager object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $hosttype);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # by default, dont load the nameserver
    my $load_nameserver = $SwitchUtil::FALSE;

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ "SanSwitch") {
            $self->_set_err("Error: PortMgr::_new() " . 
                            "Invalid value for argument '-parent'");
            return $self;
        }
        $hosttype = $args{-parent}->get_connection()->{hostp}{hosttype};
    } elsif (grep (/^-?load_nameserver$/, (keys %args))) {
        $load_nameserver = $args{-load_nameserver};
    } else {
        $self->_set_err("Error: PortMgr::_new() " .
                        "Missing required argument '-parent'");
        return $self;
    }

    # load the appropriate Port for the hosttype
    if ($hosttype =~ /fcp-switch-cisco/i) {
        require San::SanSwitch::FC::PortMgr::CiscoFCPort;
        $self = San::SanSwitch::FC::PortMgr::CiscoFCPort->_new(%args);
    } elsif ($hosttype =~ /ethernet-switch-cisco|switch-ethernet-cisco|ethernet-switch-cisco-ios/i) {
        require San::SanSwitch::Ethernet::PortMgr::CiscoEthernetPort;
        $self = San::SanSwitch::Ethernet::PortMgr::CiscoEthernetPort->_new(%args);
    } elsif ($hosttype =~ /switch-corduroyl/i) {
        require San::SanSwitch::Ethernet::PortMgr::CorduroyLEthernetPort;
        $self = San::SanSwitch::Ethernet::PortMgr::CorduroyLEthernetPort->_new(%args);
    } elsif ($hosttype =~ /switch-brocade/i) {
        if ($hosttype =~ /switch-brocade-vdx/i) {
            require San::SanSwitch::FC::PortMgr::BrocadeVDXFCPort;
            $self = San::SanSwitch::FC::PortMgr::BrocadeVDXFCPort->_new(%args);
        } else {
            require San::SanSwitch::FC::PortMgr::BrocadeFCPort;
            $self = San::SanSwitch::FC::PortMgr::BrocadeFCPort->_new(%args);
        }
    } elsif ($hosttype =~ /switch-mcdata/i) {
        require San::SanSwitch::FC::PortMgr::McDataFCPort;
        $self = San::SanSwitch::FC::PortMgr::McDataFCPort->_new(%args);
    } elsif ($hosttype =~ /switch-qlogic/i) {
        require San::SanSwitch::FC::PortMgr::QlogicFCPort;
        $self = San::SanSwitch::FC::PortMgr::QlogicFCPort->_new(%args);
    } elsif ($hosttype =~ /switch-apcon/i) {
        require San::SanSwitch::FC::PortMgr::ApconFCPort;
        $self = San::SanSwitch::FC::PortMgr::ApconFCPort->_new(%args);
    } else {
        $self->_set_err("Error: PortMgr::_new() Switch type '" . $hosttype . "' " .
                        "not currently supported with this library");
        return $self;
    }

    $self->{parent} = $args{-parent};

    # load the name server if requested
    if ($load_nameserver == $SwitchUtil::TRUE) {
        $self->load_nameserver();
        if ($self->error()) {
            $self->_set_err("Failed to load the nameserver", $self->errmsg());
            return $self;
        }
    }

    # By returning $self here we are returning a parent class such
    # as CiscoFCPort, not a PortMgr object.  This is OK because
    # the switch specific PortMgr::XXXPort classes add PortMgr to @ISA
    # so PortMgr is the base class for the parent
    return ($self);
}

################################################################################
# Subroutine Name: waitfor_port()
# Objective:       to wait for the state of the port to meet a passed in
#                  codition (i.e offline or online)
# Details:         See POD documentation at the end of this file
################################################################################
sub waitfor_port {
    my ($self, $port, $state) = @_;

    my ($pinfo);

    # Check status of the port to make sure it match'es passed state
    # Will try at most $self->{parent}{port_retries} times with 3 second sleeps
    my $count = 0;
    while ($count < $self->{parent}{port_retries} and
           ($pinfo = $self->get_port_info($port))->{state} !~ /$state/i) {
        sleep(3);
        $count++;
    }

    if ($count < $self->{parent}{port_retries}) {
        # Update the nameserver information
        $self->{nameserver} = $self->get_nameserver();
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: check_port_settings()
# Objective:       get's information about the passed in port and then
#                  compares that state versus state stored in object
#                  (which gets saved before the port disable). Idea is to
#                  catch cases where port is blocked and after re-enable it logs
#                  in with a different speed and/or type (topology)
# Details:         See POD documentation at the end of this file
################################################################################
sub check_port_settings {
    my ($self, $port) = @_;

    # If we were not requested to check, bail out
    if ($self->{parent}->{verify_on_enable} != 1) {
        return $SwitchUtil::TRUE;
    }

    # See if we have previous information for this port. If we don't, then
    # just bail.
    unless (defined($self->{$port}->{previous_type}) or defined($self->{$port}->{previous_speed})) {
        # We don't have any previous information so just log a message and return.
        $self->sanlog(-msg => "Skipping type/speed validation for port '$port' since " .
                              "no previous information found");
        return $SwitchUtil::TRUE;
    }

    # Check speed and type of the port to make sure it matches passed state
    # Will try at most $self->{parent}{port_retries} times with 3 second sleeps
    my $count = 1;

    while ($count <= $self->{parent}{port_retries}) {
        $self->sanlog(-msg => "Beginning type/speed validation attempt '$count' of '$self->{parent}{port_retries}'");

        # Initialize speed and type variables
        my $speed_is_valid = $SwitchUtil::FALSE;
        my $type_is_valid  = $SwitchUtil::FALSE;

        # Sleep 3 seconds before checking
        sleep(3);

        # if defined, compare values now (after re-enable) versus before
        # the start of the block
        my $port_state = $self->get_port_info($port);
        if ($self->error) {
            $self->_set_err("Error getting port state information for port ($port)");
            return $SwitchUtil::FALSE;
        }

        # check current type versus one store before port disable
        if (defined($self->{$port}->{previous_type})) {
            if ($port_state->{type} ne $self->{$port}->{previous_type}) {
                # Port type mismatch
                $self->sanlog(-msg => "Miscompare on port ($port) for port type: " .
                                      "Current ($port_state->{type}), before block " .
                                      "($self->{$port}->{previous_type})");
            } else {
                # Port type looks good.
                $self->sanlog(-msg => "Comparison of port ($port) for port type (before and after block), passed");
                $type_is_valid = $SwitchUtil::TRUE;
            }
        }

        # check current speed versus one stored before port disable
        if (defined($self->{$port}->{previous_speed})) {
            if ($port_state->{speed} != $self->{$port}->{previous_speed}) {
                # Port speed mismatch
                $self->sanlog(-msg => "Miscompare on port ($port) for port speed: " .
                                      "Current ($port_state->{speed}), before block " .
                                      "($self->{$port}->{previous_speed})");
            } else {
                # Port speed lookds good
                $self->sanlog(-msg => "Comparison of port ($port) for port speed (before and after block), passed");
                $speed_is_valid = $SwitchUtil::TRUE;
            }
        }

        # Check to see if both speed and type are valid
        if ($speed_is_valid == $SwitchUtil::TRUE and $type_is_valid == $SwitchUtil::TRUE) {
            # Looks good.  Reset the "previous" values and return true
            $self->{$port}->{previous_speed} = undef;
            $self->{$port}->{previous_type} = undef;

            return $SwitchUtil::TRUE;
        }

        # Looks like something wasn't right so increment
        # the count so we can possibly retry.
        $count++;
    }

    # If we made it this far then we ran out of retries before validating
    # that the type and speed were correct.  Reset the "previous" values,
    # set error and return false.
    $self->{$port}->{previous_speed} = undef;
    $self->{$port}->{previous_type} = undef;

    $self->_set_err("Miscompare on port ($port) for port speed or type");
    return $SwitchUtil::FALSE;
}

################################################################################
# Subroutine Name: get_nameserver()
# Objective:       returns the hash representing the namsserver. Key to the
#                  hash is the interface id of the port instance. Will not
#                  refresh the data, but just return the currect representation
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nameserver {
    my ($self, $all_vf) = @_;
    my @ns_cache = ();

    if ($self->{parent}->is_cache_data_valid("nameserver")) {
        (@ns_cache) = $self->{parent}->get_cache_data("nameserver");
        return $ns_cache[0];
    }
    
    $self->{nameserver} = $self->_get_nameserver_database($all_vf);
    return {} if ($self->error);
    
    if ($self->{parent}->is_sancache_enabled()) {
        $self->{parent}->store_cache_data("nameserver", $self->{nameserver});
    }

    return($self->{nameserver});

}

################################################################################
# Subroutine Name: load_nameserver()
# Objective:       loads the nameserver into memory
################################################################################
sub load_nameserver {
    my ($self) = @_;

    # get the name server information about the switch to obtain type and wwpn
    # and iterate over the type to find port types. Tell the call to do over all
    # virtual fabrics if applicable.
    my $all_virtual_fabrics = 1;
    my $ns = $self->get_nameserver($all_virtual_fabrics);
    foreach my $ns_key (sort keys(%{$ns})) {
        my $ns_port = $ns_key;
        my $type = &SwitchUtil::normalize_type($ns->{$ns_key}->{type});

        # if the WWPN is not defined or NULL which could be in FCoE case,
        # move to next WWPN
        my $wwpn = $ns->{$ns_key}->{wwpn};
        if (not defined($wwpn) or $wwpn eq "") {
            next;
        }

        # look at e/f/l ports and bucket the ports
        if ($type =~ /^E-Port$|E_Port/i) {
            # For e-port, just key (or $ns_port) since this will
            # be the e-port and no lookup is needed.
            my $port_info = $self->get_port_info($ns_port);
            if ($self->error) {
                $self->_set_err("Failed to get information on port $ns_port", $self->errmsg());
                return $SwitchUtil::FALSE;
            } else {
                # Store the port information on the portMgr object
                $self->{eport}->{$ns_port}->{port_info} = {%{$port_info}};
                # Store virtual fabric information if applicable
                if ($self->{parent}->{vf_status} == 1) {
                    $self->{eport}->{$ns_port}->{vf_context_id} = $ns->{$ns_key}->{fid};
                }
            }
        } elsif ($type =~ /^L-Port$|L_Port/i or
                 $type =~ /^FL-Port$|FL_Port/i) {
            # if the wwpn contains ";" then we know it's from an NPIV
            # port (multiple WWPNs). This case this these will not be an lport
            # so, skip this port
            if ($wwpn =~ /\;/) {
                next;
            }
            my $port_num = $self->get_port_num(-wwpn => $wwpn);
            if ($self->error) {
                $self->_set_err("Failed to get port number for WWPN $wwpn", $self->errmsg());
                return $SwitchUtil::FALSE;
            }

            # Call get_port_info to get the port information
            my $port_info = $self->get_port_info($port_num);
            if ($self->error) {
                $self->_set_err("Failed to get information on port $port_num", $self->errmsg());
                return $SwitchUtil::FALSE;
            } else {
                # Store the port information on the portMgr object
                $self->{lport}->{$port_num}->{port_info} = {%{$port_info}};
                # Store virtual fabric information if applicable
                if ($self->{parent}->{vf_status} == 1) {
                    $self->{lport}->{$port_num}{vf_context_id} = $self->{parent}->{vf_context_id};
                }
            }
        } elsif ($type =~ /^F-Port$|F_Port/i) {
            # if the wwpn contains ";" then we know it's from an NPIV
            # port (multiple WWPNs). This case this these will be an fport,
            # but we only need to look up the first wwpn since they will
            # all point to the same port
            if ($wwpn =~ /\;/) {
                my @wwpns = split(/\;/, $wwpn);
                $wwpn = $wwpns[0];
            }
            my $port_num = $self->get_port_num(-wwpn => $wwpn);
            if ($self->error) {
                $self->_set_err("Failed to get port number for WWPN $wwpn", $self->errmsg());
                return $SwitchUtil::FALSE;
            }

            # Call get_port_info to get the port information
            my $port_info = $self->get_port_info($port_num);
            if ($self->error) {
                $self->_set_err("Failed to get information on port $port_num", $self->errmsg());
                return $SwitchUtil::FALSE;
            } else {
                # Store the port information on the portMgr object
                $self->{fport}->{$port_num}->{port_info} = {%{$port_info}};
                # Store virtual fabric information if applicable
                if ($self->{parent}->{vf_status} == 1) {
                    $self->{fport}->{$port_num}{vf_context_id} = $self->{parent}->{vf_context_id};
                }
            }
        }
    }

    return $SwitchUtil::TRUE;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Port.pm

=head1 Author

Evan Chiu, Tom Masterson

=head1 Description

Provides an abstract API interface to allow operations on switch ports.

=head1 See Also

Package inherits from the following modules. See modules for interface detail

B<Error.pm>

=head1 Attributes

=head2 B<parent>

SanSwitch object = Reference to a parent SanSwitch object

=head1 Public Methods

=head2 B<get_port($name)>

Returns the port number correspnding to the FCID, WWPN, or WWNN provided.
Returns error if none exists.

=over 4

=item Inputs

=item * I<-wwpn>

The WWPN of the target or initiator

=item Outputs

Returns the port number. If an error occurs, the error flag is set and errmsg is
set to the failure reason.  

=item Examples

    # Return a port interface.
    $Test->log("-- Get port interface");
    my $port = $portObj->get_port_num(-wwpn => '50:0a:09:82:8a:f2:ad:c2');
    if ($portObj->error){
        $Test->log($portObj->errmsg());    
        return 0;
    }

=back

=head2 B<port_disable()>

Disables the specified port.

=over 4

=item Inputs

=item * -port

The port number/interface to disable

=item Outputs

Disables the port. If an error occurs, the error flag is set and errmsg is
set to the failure reason.  

=item Examples

    # Disable port fc1/5
    $Test->log("-- Disabling port fc1/5");
    $portObj->port_disable('fc1/5');
    if ($portObj->error){
        $Test->log($portObj->errmsg());    
        return 0;
    }

=back

=head2 B<port_enable()>

Enables the specified port.

=over 4

=item Inputs

=item * -port

The port number/interface to enable

=item Outputs

Enables the port. If an error occurs, the error flag is set and errmsg is
set to the failure reason.  

=item Examples

    # Enable port 5
    $Test->log("-- Enabling port fc1/5");
    $portObj->port_enable('fc1/5');
    if ($portObj->error){
        $Test->log($portObj->errmsg());    
        return 0;
    }

=back

=head2 B<set_port_speed()>

Sets the speed on a specified port

=over 4

=item Inputs

=item * I<-port>

The port number to change the speed

=item * I<-speed>

The speed to set it to. This can be 1, 2, 4, or auto.

=item Outputs

Sets the port speed. If an error occurs, the error flag is set and errmsg is
set to the failure reason.  

=item Examples

    # Set port fc1/5 to 4Gb
    $Test->log("-- Setting port fc1/5 to 4Gbs");
    $portObj->set_port_speed(-port => fc1/5, -speed => 4);
    if ($portObj->error){
        $Test->log($portObj->errmsg());    
        return 0;
    }

=back

=head2 B<set_port_type()>

Sets the type on a specified port. Brocade style port types will be used
so that we can have portable scripts at the thpl layer. For example F_Port, 
FL_Port, E_Port, ... types will be used. For cases where a specific vendor
supports a type is not analogous to Brcoade, that specific switch vendor type
can be used (but will not be portable to run in all switch environments).
Examples of specifc types are Cisco's SD, ST, TL types.

=over 4

=item Inputs

=item * I<-port>

The port number to change the type

=item * I<-type>

The type to set it to. For example: E, F, FL, or auto.

=item Outputs

Sets the port type. If an error occurs, the error flag is set and errmsg is
set to the failure reason.  

=item Examples

    # Set port fc1/5 to be an E-port
    $Test->log("-- Setting port fc1/5 to type E-port");
    $portObj->set_port_speed(-port => fc1/5, -type => 'E_Port');
    if ($portObj->error){
        $Test->log($portObj->errmsg());    
        return 0;
    }

=back

1;

