# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides common switch methods
## @author evan@netapp.com
## @status private
## @pod here

use strict;

# Package Name
##############
package SwitchUtil;

use Tharn;
use San::SanUtil;

# Exported global variables
###########################
$SwitchUtil::TRUE  = 1;
$SwitchUtil::FALSE = 0;

###############################################################################
# Subroutine Name: execute
# Objective: Execute command on switch and return output
###############################################################################
sub execute {
    my ($connection, $cmd, %args) = @_;
    my ($env_output, $output, @output, $status);
    my ($match,      $before, $after,  $default_timeout);
    my ($timeout,    $cmdlog, $quiet);

    $default_timeout = "180";
    $cmdlog          = 1;
    $status          = $SwitchUtil::FALSE;
    $quiet           = 0;

    foreach (keys %args) {
        if (/^-?cmdlog$/) {
            $cmdlog = $args{$_};
        } elsif (/^-?timeout$/) {
            $timeout = $args{$_};
        } elsif (/^-?quiet$/) {
            $quiet = $args{$_};
        }
    }

    # clear buffers
    $connection->discardbuff();

    # execute the passed command
    if ($quiet == 0) {
        ($match, $before, $after) = &Tharn::say($connection, $cmd, '-timeout',
                (defined($timeout) and $timeout ne "") ? $timeout : $default_timeout);
    } else {
        ($match, $before, $after) = &Tharn::say($connection, $cmd, '-timeout',
                (defined($timeout) and $timeout ne "") ? $timeout : $default_timeout, -noverbose);
    }

    # strip output and remove CR/LF
    chomp($after);
    $after =~ s/\r//g;

    # strip any telnet ctrl characters that get
    # added to the output (this problem seems
    # to only occur with certain platforms.
    $after =~ s/ \cH//g;

    # create an array from the output
    @output = split(/\n/, $after);

    # strip original command out of the output
    if ($output[0] =~ /$cmd/) {
        shift @output;
    }

    # Check output for known error strings to determine status
    unless (   $after =~ /command not found/i
            || $after =~ /invalid command/i
            || $after =~ /requested config change not allowed/i
            || $after =~ /Cannot resolve/i
            || $after =~ /command not found/i
            || $after =~ /Invalid input parameters\./i
            || $after =~ /syntax error/i
            || $after =~ /Command not found/i
            || $after =~ /error\s+\d+\:/i
            || $after =~ /^error\:/i) {
        $status = $SwitchUtil::TRUE;
    }

    return ($status, @output);
}


###############################################################################
# Subroutine Name: cr_execute
# Objective: Execute command on switch and return output
###############################################################################
sub cr_execute {
    my ($connection, %args) = @_;
    %args = &SanUtil::strip_dash(%args);
    my $status = 0;

    if (not exists($args{timeout})) {
        # default timeout of 180 is used if not user defined
        $args{timeout} = 180;
    }

    # clear buffers
    $connection->discardbuff();

    my @output = $connection->execute(%args);

    # Check output for known error strings to determine status
    if (grep(/command not found/i, @output)) {
        $status = $SwitchUtil::TRUE;
    }

    return ($status, @output);
}


###############################################################################
# Subroutine Name: normalize_state
# Objective: Convert online/offline variants into a standard value
###############################################################################
sub normalize_state {
    my ($state) = @_;
    my $str;

    if (   $state =~ /enable/i
        || $state =~ /up/i
        || $state =~ /trunking/i
        || $state =~ /1/i
        || $state =~ /active/i
        || $state =~ /online/i) {
        $str = "online";
    } elsif (   $state =~ /disable/i
             || $state =~ /down/i
             || $state =~ /0/i
             || $state =~ /No Light/i
             || $state =~ /offline/i
             || $state =~ /suspended/i
             || $state =~ /unknown/i) {
        $str = "offline";
    } else {
        $str = $state;
    }

    return $str;
}

###############################################################################
# Subroutine Name: normalize_type
# Objective: Convert various port types to normalized value (use Brocade as
#            standard representatio - unless specified)
###############################################################################
sub normalize_type {
    my ($type) = @_;
    my $str;

    if ($type eq "F_Port-FC_VI") {
        return $type;
    } 

    if ($type eq "F_Port" or
        $type eq "F-Port" or
        $type eq "fPort" or
        $type eq "Fx" or
        $type eq "F" or
        $type eq "VF") {
        return "F_Port";
    } 

    if ($type eq "FL_Port" or
        $type eq "FL-Port" or
        $type eq "flPort" or
        $type eq "L_Port" or
        $type eq "L-Port" or
        $type eq "L" or
        $type eq "FL" or
        $type eq "GL") {
        return "FL_Port";
    } 

    if ($type eq "E_Port" or
        $type eq "E-Port" or
        $type eq "ePort" or
        $type eq "TE" or
        $type eq "E" or
        $type eq "VE") {
        return "E_Port";
    }
 
    if ($type eq "G_Port" or
        $type eq "G-Port" or
        $type eq "gxPort" or
        $type eq "G") {
        return "G_Port";
    }

    if ($type eq "N_Port" or
        $type eq "N-Port" or
        $type eq "nPort" or
        $type eq "NP" or
        $type eq "N") {
        return "N_Port";
    }

    # Brocade only port type
    if ($type eq "EX_Port") {
        return "EX_Port";
    }

    # Brocade only port type
    if ($type eq "VEX_Port") {
        return "VEX_Port";
    }

    # Cisco only port type
    if ($type eq "SD") {
        return "SD";
    }

    # Cisco only port type
    if ($type eq "ST") {
        return "ST";
    }

    # Cisco only port type
    if ($type eq "TL") {
        return "TL";
    }

    # allow auto to passed through
    if ($type =~ /auto/i) {
        return "auto";
    }

    # offline is a passthrough
    if ($type =~ /offline/i) {
        return $type
    }

    # If you reach this location in code, we could
    # not determine the port type, and returning null
    # string.
    return "";

}

1;

=pod

=head1 Package Name

    SwitchUtil.pm

=cut

