# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to zone a switch
## @author derricks@netapp.com
## @status private
## @pod here
## @unit_test thpl p4/test/lib/UnitTest/test_San_SanSwitch_ZoneMgr.thpl

package ZoneMgr;

# Compiler directives.
use strict;

# Module import.
use San::SanSwitch::SwitchUtil;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    discover
# Objective: Discover the aliases, zones, and zonesets
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover {
    my ($self) = @_;

    $self->_clear_err();

    # Call $self->alias_discover to build out the alias objects
    unless ($self->alias_discover()) {
        if (not $self->error) {
            $self->_set_err("Unable to discover switch aliases");
        }
        return $SwitchUtil::FALSE;
    } ## end unless ($self->alias_discover...

    # Call $self->zone_discover to build out the zone objects
    unless ($self->zone_discover()) {
        if (not $self->error) {
            $self->_set_err("Unable to discover switch zones");
        }
        return $SwitchUtil::FALSE;
    } ## end unless ($self->zone_discover...

    # Call $self->zoneset_discover to build out the zoneset objects
    unless ($self->zoneset_discover()) {
        if (not $self->error) {
            $self->_set_err("Unable to discover switch zonesets");
        }
        return $SwitchUtil::FALSE;
    } ## end unless ($self->zoneset_discover...

    # Call $self->active_zoneset_discover to discover the active zoneset
    unless ($self->active_zoneset_discover()) {
        if (not $self->error) {
            $self->_set_err("Unable to discover active zoneset");
        }
        return $SwitchUtil::FALSE;
    } ## end unless ($self->active_zoneset_discover...

    return $SwitchUtil::TRUE;
} ## end sub discover

###############################################################################
# Method:    member_get_aliases
# Objective: Get a list of aliases that a wwpn or port belongs to
# Details:   See POD documentation at the end of this file.
###############################################################################
sub member_get_aliases {
    my ($self, %args) = @_;

    my ($member);

    $self->_clear_err();

    # Initialize some variables
    my $type = "cache";

    foreach (keys %args) {
        if (/-?member/) {
            # Check to make sure that we were actually passed in something
            if (not length($member = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }
        } elsif (/-?type/) {
            # Check to make sure that we were actually passed an alias
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }

            # Make sure that type is either switch or cache
            unless ($type eq "switch" or $type eq "cache") {
                $self->_set_err("Invalid value for argument '$_'.  Must be 'cache' or 'switch'");
                return {};
            }
        }
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($member) and $member ne "") {
        $self->_set_err("Missing required argument '-member'");
        return {};
    }

    # Get a hash of all of the current aliases in the cache
    my $cache_aliases_ref = $self->get_aliases();
    if ($self->error) {
        return {};
    }

    # Loop through the cache aliases hash and compare each member to the one
    # we are looking for.  If we find it, add the alias object to the hash
    # reference member_aliases_ref;
    my $member_aliases_ref = {};

    foreach my $alias (sort keys %$cache_aliases_ref) {
        my $aliasobj = $cache_aliases_ref->{$alias};
        unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {
            $self->_set_err("Unable to retrieve object for alias $alias");
            return {};
        }

        # Get the member for the alias
        my $t_member = $aliasobj->get_member();
        if (defined($member)) {
            if ($member eq $t_member) {
                # Looks like we found what we were looking.  See if it is on the switch
                # if the type is switch.
                if ($type eq "switch") {
                    # Looks like they only want switch stuff
                    unless ($aliasobj->is_on_switch()) {
                        # is_on_switch failed.  check for an error
                        if ($aliasobj->error()) {
                            # Yep...found an error
                            my $self->_set_err("Unable to get the on_switch key for the object:\n" .
                                               $aliasobj->errmsg());
                            return {};
                        } else {
                            # no errors...just wasn't on the switch so we will skip it.
                            next;
                        }
                    }
                }
                $member_aliases_ref->{$alias} = $aliasobj;
            } ## end if ($member eq $t_member)
        } else {

            # The alias doesn't have a member defined.  This should never
            # happen so if it does, set an error and bail.
            $self->_set_err("Alias $alias does not have a member defined");
            return {};
        } ## end else [ if (defined($member))
    } ## end foreach my $alias (sort keys...

    # Return the hash reference
    return $member_aliases_ref;
} ## end sub member_get_aliases

###############################################################################
# Method:    alias_create
# Objective: Create an alias on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub alias_create {
    my ($self, %args) = @_;

    my ($alias);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $aliasobj = $self->get_aliases(-alias => $alias);
    unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for alias $alias");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($aliasobj)...

    # Make sure the alias doesn't already exist on the switch.
    if ($aliasobj->exists()) {
        # if there is an error on the aliasobject, then bail out
        if ($aliasobj->error()) {
            $self->_set_err("Encountered an error trying to determine if $alias exists", $aliasobj->errmsg());
            return $SwitchUtil::FALSE;
        } else {
            $self->_set_err("Alias $alias already exists on the switch", $aliasobj->errmsg());
            return $SwitchUtil::FALSE;
        }
    } else {

        # if there is an error on the aliasobject, then bail out
        if ($aliasobj->error()) {
            $self->_set_err("Encountered an error trying to determine if $alias exists", $aliasobj->errmsg());
            return $SwitchUtil::FALSE;
        }

        # Alias doesn't exist on the switch so go ahead and create it
        if ($aliasobj->create()) {
            $self->sanlog(-msg => "Alias $alias has been created on the switch");
            $aliasobj->{on_switch} = $SwitchUtil::TRUE;
        } else {
            if ($aliasobj->error) {
                $self->_set_err($aliasobj->errmsg());
            } else {
                $self->_set_err("Unable to create alias $alias on the switch");
            }
            return $SwitchUtil::FALSE;
        } ## end else [ if ($aliasobj->create(...
    } ## end else [ if ($aliasobj->exists(...

    return $SwitchUtil::TRUE;
} ## end sub alias_create

###############################################################################
# Method:    alias_destroy
# Objective: destroy an alias on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub alias_destroy {
    my ($self, %args) = @_;

    my ($alias);
    my $force = 0;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } elsif (/-?force/) {
            $force = $args{$_};
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $aliasobj = $self->get_aliases(-alias => $alias);
    unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for alias $alias");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($aliasobj)...

    # Make sure the alias exists on exists switch and in the alias cache
    # by calling $self->alias_verify.  If it isn't found in either
    # location, print a message and return true.  If it is found, but
    # is not validated, then set an error.
    if (not $force and not $self->alias_verify(-alias => $alias)) {
        if (not $self->error) {
            $self->_set_err("Unable to verify alias status in the cache and on the switch");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $self->alias_verify...

    # Check to make sure the alias isn't presently used in any zones.  If
    # it is, then set an error and bail.
    my $zones_ref = $self->alias_get_zones(-alias => $alias, -type => "switch");
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to see if any of the zone objects returned have the on_switch parameter set
    # to $SwitchUtil::TRUE;
    my @zones = ();
    foreach my $zone (keys %$zones_ref) {
        my $t_zoneobj = $zones_ref->{$zone};
        if ($t_zoneobj->{on_switch} == $SwitchUtil::TRUE) {

            # Store the zone name for later use;
            push(@zones, $zone);
        }
    } ## end foreach my $zone (keys %$zones_ref)

    # Check the size of @zones to see if any zones were found that the alias
    # belongs that exist on the switch
    if (scalar(@zones) > 0) {
        $self->_set_err("Alias $alias currently belongs to one or more zones (@zones)");
        return $SwitchUtil::FALSE;
    }

    # We should now know we have a valid alias to delete
    if ($aliasobj->destroy()) {
        $self->sanlog(-msg => "Alias $alias has been removed from the switch");
        $aliasobj->{on_switch} = $SwitchUtil::FALSE;
    } else {
        if ($aliasobj->error) {
            $self->_set_err("Failed to remove alias from the switch:\n" . $aliasobj->errmsg());
        } else {
            $self->_set_err("Failed to remove alias from the switch");
        }
        return $SwitchUtil::FALSE;
    } ## end else [ if ($aliasobj->destroy...

    return $SwitchUtil::TRUE;
} ## end sub alias_destroy

###############################################################################
# Method:    alias_exists
# Objective: Verify an alias exists on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub alias_exists {
    my ($self, %args) = @_;

    my ($alias);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $aliasobj = $self->get_aliases(-alias => $alias);
    unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for alias $alias");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($aliasobj)...

    # check existence of the object with $aliasobj->exists()
    if (not $aliasobj->exists()) {
        if ($aliasobj->error) {
            $self->_set_err("Failed to check if alias $alias exists:\n" . $aliasobj->errmsg());
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $aliasobj->exists...

    # Everything looks good so return TRUE
    return $SwitchUtil::TRUE;
} ## end sub alias_exists

###############################################################################
# Method:    alias_get_zones
# Objective: Get a list of zones an alias belongs to
# Details:   See POD documentation at the end of this file.
###############################################################################
sub alias_get_zones {
    my ($self, %args) = @_;

    my ($alias);

    $self->_clear_err();

    # Initialize some variables
    my $type = "cache";

    foreach (keys %args) {
        if (/-?alias/) {
            # Check to make sure that we were actually passed an alias
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }
        } elsif (/-?type/) {
            # Check to make sure that we were actually passed an alias
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }

            # Make sure that type is either switch or cache
            unless ($type eq "switch" or $type eq "cache") {
                $self->_set_err("Invalid value for argument '$_'.  Must be 'cache' or 'switch'");
                return {};
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Missing required argument '-alias'");
        return {};
    }

    my $aliasobj = $self->get_aliases(-alias => $alias);
    unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for alias $alias");
        }
        return {};
    } ## end unless (defined($aliasobj)...

    # Get a hash of all of the current zones in the cache
    my $cache_zones_ref = $self->get_zones();
    if ($self->error) {
        return {};
    }

    # Loop through the zone hash reference  and
    # see if the alias belongs to any of the zones.  We will use
    # the temp hash reference $zones_ref to store the zones with
    # the format $zones_ref->{$zone} = zone_obj
    my $zones_ref = {};
    foreach my $t_zone (sort keys %$cache_zones_ref) {
        # Check to see if the zone object has the alias in it we are looking for.
        my $t_zoneobj = $cache_zones_ref->{$t_zone};

        # Get the aliases from the object
        my $aliases_ref = $t_zoneobj->get_aliases();
        if ($t_zoneobj->error()) {
            $self->_set_error("Unable to get the aliases for zone '$t_zone':\n" .
                              $t_zoneobj->errmsg());
            return {};
        }

        # Look for the alias in the aliases ref.
        if (defined($aliases_ref->{$alias}) and $aliases_ref->{$alias} ne "") {
            # looks like we have a good match.  See if they want to only return stuff
            # that is on the switch and add to the zones_ref accordingly.
            if ($type eq "switch") {
                unless ($t_zoneobj->is_on_switch()) {
                    if ($t_zoneobj->error()) {
                        # Had an error on the method
                        $self->_set_error("Unable to retrieve the 'on_switch' key from the object:\n" .
                                          $t_zoneobj->errmsg());
                    } else {
                        # No error...just not on the switch, so skip this one.
                        next;
                    }
                }
            }
            $zones_ref->{$t_zone} = $t_zoneobj;
        } ## end if (defined($t_zoneobj...

    } ## end foreach my $t_zone (sort keys...
    return $zones_ref;
} ## end sub alias_get_zones

###############################################################################
# Method:    alias_verify
# Objective: Verify an alias object against the switch cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub alias_verify {
    my ($self, %args) = @_;

    my ($alias);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $aliasobj = $self->get_aliases(-alias => $alias);
    unless (defined($aliasobj) and ref($aliasobj) =~ /Alias/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for alias $alias");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($aliasobj)...

    # validate the object with $aliasobj->validate()
    if (not $aliasobj->validate()) {
        if ($aliasobj->error) {
            $self->_set_err("Failed to validate alias $alias:\n" . $aliasobj->errmsg());
        } else {
            $self->_set_err("Failed to validate alias $alias");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $aliasobj->validate...

    # Everything looks good so return TRUE
    return $SwitchUtil::TRUE;
} ## end sub alias_verify

###############################################################################
# Method:    zone_create
# Objective: Create a zone on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_create {
    my ($self, %args) = @_;

    my ($zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed a zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zone $zone");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($zoneobj) ...

    # Make sure the zone doesn't already exist on the switch.
    if ($zoneobj->exists()) {
        $self->_set_err("Zone $zone already exists on the switch");
        return $SwitchUtil::FALSE;
    } else {

        # Zone doesn't exist on the switch so go ahead and create it.
        # First verify the alias objects since we don't want to create a zone
        # from aliases that don't already exist.  Return falase if we can't
        # validate the aliases as good.

        my $aliases_ref = $zoneobj->get_aliases();
        if (scalar(keys %$aliases_ref) == 0) {
            $self->_set_err("Unable to determine the aliases for zone $zone");
            return $SwitchUtil::FALSE;
        }

        foreach my $t_alias (sort keys %$aliases_ref) {

            # Verify the alias exists
            if (not $self->alias_exists(-alias => $t_alias)) {
                if (not $self->error()) {
                    $self->_set_err("Unable to verify alias '$t_alias' from zone '$zone' exists");
                }
                return $SwitchUtil::FALSE;
            } ## end if (not $self->alias_exists...
        } ## end foreach my $t_alias (sort keys...

        # Looks like all of the aliases defined on the zone object are good so
        # we can continue to create the zone on the switch
        if ($zoneobj->create()) {
            $self->sanlog(-msg => "Zone $zone has been created on the switch");
            $zoneobj->{on_switch} = $SwitchUtil::TRUE;
        } else {
            if (not $self->error) {
                $self->_set_err("Unable to create zone $zone on the switch");
            }
            return $SwitchUtil::FALSE;
        } ## end else [ if ($zoneobj->create())
    } ## end else [ if ($zoneobj->exists())

    return $SwitchUtil::TRUE;
} ## end sub zone_create

###############################################################################
# Method:    zone_destroy
# Objective: Destroy a zone on the switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_destroy {
    my ($self, %args) = @_;

    my ($zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed a zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zone $zone");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($zoneobj) ...

    # Make sure the zone doesn't belong to any zonesets. If it does,
    # then set an error and bail.
    my $zonesets_ref = $self->zone_get_zonesets(-zone => $zone, -type => "switch");
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Check to see if any of the zoneset objects returned have the on_switch parameter set
    # to $SwitchUtil::TRUE;
    my @zonesets = ();
    foreach my $zoneset (keys %$zonesets_ref) {
        my $t_zonesetobj = $zonesets_ref->{$zoneset};
        if ($t_zonesetobj->{on_switch} == $SwitchUtil::TRUE) {

            # Store the zoneset name for later use;
            push(@zonesets, $zoneset);
        }
    } ## end foreach my $zoneset (keys %$zonesets_ref)

    # Check the size of @zonesets to see if any zonesets were found that the zones
    # belongs that exist on the switch
    if (scalar(@zonesets) > 0) {
        $self->_set_err("Zone $zone currently belongs to one or more zonesets (@zonesets)");
        return $SwitchUtil::FALSE;
    }

    # We should now know we have a valid zone to delete
    if ($zoneobj->destroy()) {
        $self->sanlog(-msg => "Zone $zone has been removed from the switch");
        $zoneobj->{on_switch} = $SwitchUtil::FALSE;
    } else {
        if ($zoneobj->error) {
            $self->_set_err("Failed to remove zone $zone from the switch:\n" . $zoneobj->errmsg());
        } else {
            $self->_set_err("Failed to remove zone $zone from the switch");
        }
        return $SwitchUtil::FALSE;
    } ## end else [ if ($zoneobj->destroy(...

    return $SwitchUtil::TRUE;
} ## end sub zone_destroy

###############################################################################
# Method:    zone_exists
# Objective: Verify that a zone exists
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_exists {
    my ($self, %args) = @_;

    my ($zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed a zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zone $zone");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($zoneobj) ...

    # Check if the zone exists by calling $zoneobj-exists();
    if (not $zoneobj->exists()) {
        if ($zoneobj->error) {
            $self->_set_err("Unable to check if zone $zone exists on the switch:\n" . $zoneobj->errmsg());
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $zoneobj->exists...

    # Everything looks good so return TRUE
    return $SwitchUtil::TRUE;
} ## end sub zone_exists

###############################################################################
# Method:    zone_verify
# Objective: Verify a zone object against the switch cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_verify {
    my ($self, %args) = @_;

    my ($zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed a zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zone $zone");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($zoneobj) ...

    # validate the object with $zoneobj->validate()
    if (not $zoneobj->validate()) {
        if ($zoneobj->error) {
            $self->_set_err("Failed to validate zone $zone:\n" . $zoneobj->errmsg());
        } else {
            $self->_set_err("Failed to validate zone $zone");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $zoneobj->validate...

    # Everything looks good so return TRUE
    return $SwitchUtil::TRUE;
} ## end sub zone_verify

###############################################################################
# Method:    zone_add_alias
# Objective: Add an alias to a zone
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_add_alias {
    my ($self, %args) = @_;

    my ($zone, $alias);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were passed an zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } elsif (/-?alias/) {

            # Check to make sure that we were passed an alias
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end elsif (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zone $zone");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($zoneobj) ...

    my $aliasobj = $self->get_aliases(-alias => $alias);
    unless (defined($alias) and $alias ne "") {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for alias $alias");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($alias) and...

    # Check to make sure the zone exists on the switch
    if (not $self->zone_exists(-zone => $zone)) {
        if (not $self->error) {
            $self->_set_err("Unable to verify that zone '$zone' exists");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $self->zone_exists...

    # Check to make sure the alias exists
    if (not $self->alias_exists(-alias => $alias)) {
        if (not $self->error) {
            $self->_set_err("Unable to verify that alias '$alias' exists");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $self->alias_exists...

    # Check to make sure the alias is not already a member of the zone
    my $aliases_ref = $zoneobj->get_aliases();

    if (defined($aliases_ref->{$alias})) {

        # We have found that the alias is already in the zone
        # so print a message and return false.
        $self->_set_err("Alias $alias is already in zone $zone");
        return $SwitchUtil::FALSE;
    } ## end if (defined($aliases_ref...

    # If we made it this far then we didn't find the alias in the zone
    # so proceed with adding it to the switch.
    if (not $zoneobj->add_alias(-alias => $aliasobj)) {
        if (not $self->error) {
            $self->_set_err("Unable to add alias $alias to zone $zone");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $zoneobj->add_alias...

    return $SwitchUtil::TRUE;
} ## end sub zone_add_alias

###############################################################################
# Method:    zone_remove_alias
# Objective: Remove an alias from a zone
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_remove_alias {
    my ($self, %args) = @_;

    my ($zone, $alias);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were passed an zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } elsif (/-?alias/) {

            # Check to make sure that we were passed an alias
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end elsif (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zone $zone");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($zoneobj) ...

    my $aliasobj = $self->get_aliases(-alias => $alias);
    unless (defined($alias) and $alias ne "") {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for alias $alias");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($alias) and...

    # Check to make sure the zone exists on the switch and matches the current
    # zone object definition.
    if (not $self->zone_exists(-zone => $zone)) {
        if (not $self->error) {
            $self->_set_err("Unable to verify that zone '$zone' exists");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $self->zone_exists...

    # Check to make sure the alias is a member of the zone
    my $aliases_ref = $zoneobj->get_aliases();

    if (not defined($aliases_ref->{$alias})) {

        # Alias was not found to exist in the zone object so
        # there is no way we can delete it.
        $self->_set_err("Alias $alias was not found in zone $zone");
        return $SwitchUtil::FALSE;
    }

    # If we made it this far then we found the alias in the zone
    # so proceed with removing it from the zone on the switch
    if (not $zoneobj->remove_alias(-alias => $aliasobj)) {
        if (not $self->error) {
            $self->_set_err("Unable to remove alias $alias from zone $zone");
        }
        return $SwitchUtil::FALSE;
    }

    # Get the new list of aliases on the zone object
    $aliases_ref = $zoneobj->get_aliases();

    # Check to make sure the zone still exists on the switch since removing
    # the last alias may remove the zone.  If it is gone then we need to remove
    # it from any zonesets. If the zone doesn't have any more members, then we
    # will need to delete it from the switch.
    if ($self->zone_exists(-zone => $zone ) and scalar(keys %$aliases_ref) == 0) {
        # The zone still exists on the switch but the alias list is
        # empty so we need remove the zone from any zonesets and then delete the
        # zone from the switch.
        my $zonesets_ref = $self->zone_get_zonesets(-zone => $zone, -type => "switch");
        if (not $self->error) {
            if (defined($zonesets_ref) and scalar(keys %$zonesets_ref) > 0) {
                # looks like this zone belogs to zonesets so we need
                # to remove it from them as well.
                foreach my $zoneset (sort keys %$zonesets_ref) {
                    if (not $self->zoneset_remove_zone(-zoneset => $zoneset, -zone => $zone)) {
                        if (not $self->error) {
                            $self->_set_err("Unable to remove zone $zone from zoneset $zoneset");
                        }
                        return $SwitchUtil::FALSE;
                    }
                }
            }
        } else {
            return $SwitchUtil::FALSE;
        }

        if (not $self->zone_destroy(-zone => $zone)) {
            if (not $self->error) {
                $self->_set_err("Unable to remove zone from the switch");
            }
            return $SwitchUtil::FALSE;
        }
    } elsif (not $self->zone_exists(-zone => $zone)) {
        # The zone no longer exists so we need to clean up the zoneset
        my $zonesets_ref = $self->zone_get_zonesets(-zone => $zone, -type => "switch");
        if (not $self->error) {
            if (defined($zonesets_ref) and scalar(keys %$zonesets_ref) > 0) {
                # looks like this zone belogs to zonesets so we need
                # to remove it from them as well.
                foreach my $zoneset (sort keys %$zonesets_ref) {
                    if (not $self->zoneset_remove_zone(-zoneset => $zoneset, -zone => $zone)) {
                        if (not $self->error) {
                            $self->_set_err("Unable to remove zone $zone from zoneset $zoneset");
                        }
                        return $SwitchUtil::FALSE;
                    }
                }
            }
        } else {
            return $SwitchUtil::FALSE;
        }
    }

    return $SwitchUtil::TRUE;
} ## end sub zone_remove_alias

###############################################################################
# Method:    zone_verify_alias
# Objective: Verify an alias is in a zone
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_verify_alias {
    my ($self, %args) = @_;

    my ($zone, $alias);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were passed an zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } elsif (/-?alias/) {

            # Check to make sure that we were passed an alias
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end elsif (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zone $zone");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($zoneobj) ...

    my $aliasobj = $self->get_aliases(-alias => $alias);
    unless (defined($alias) and $alias ne "") {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for alias $alias");
        }
        return $SwitchUtil::FALSE;
    } ## end unless (defined($alias) and...

    # Check to make sure the alias is a member of the zone
    my $aliases_ref = $zoneobj->get_aliases();

    if (defined($aliases_ref->{$alias})) {
        return $SwitchUtil::TRUE;
    } else {

        # Alias was not found to exist in the zone object so
        return $SwitchUtil::FALSE;
    }
} ## end sub zone_verify_alias

###############################################################################
# Method:    zone_get_aliases
# Objective: Get a list of aliases that the zone contains
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_get_aliases {
    my ($self, %args) = @_;

    my ($zone);

    $self->_clear_err();

    # Initialize some variables
    my $type = "cache";

    foreach (keys %args) {
        if (/-?zone/) {
            # Check to make sure that we were actually passed a zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }
        }
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return {};
    }

    # Grab the object for the zone
    my $zoneobj = $self->get_zones(-zone => $zone);
    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zone $zone");
        }
        return {};
    } ## end unless (defined($zoneobj) ...

    # Grab the aliases for the zone
    my $aliases_ref = $zoneobj->get_aliases();
    if ($zoneobj->error()) {
        $self->_set_err("Unable to retrieve the aliases hash for zone '$zone':\n",
                        $zoneobj->errmsg());
        return {};
    }

    return $aliases_ref;
} ##END  sub zone_get_aliases {

###############################################################################
# Method:    zone_get_zonesets
# Objective: Get a list of zonesets that the zone belongs to
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zone_get_zonesets {
    my ($self, %args) = @_;

    my ($zone);

    $self->_clear_err();

    # Initialize some variables
    my $type = "cache";

    foreach (keys %args) {
        if (/-?zone/) {
            # Check to make sure that we were actually passed a zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }
        } elsif (/-?type/) {
            # Check to make sure that we were actually passed an alias
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }

            # Make sure that type is either switch or cache
            unless ($type eq "switch" or $type eq "cache") {
                $self->_set_err("Invalid value for argument '$_'.  Must be 'cache' or 'switch'");
                return {};
            }
        }
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return {};
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        if (not $self->error) {
            $self->_set_err("Unable to retrieve the object for zone $zone");
        }
        return {};
    } ## end unless (defined($zoneobj) ...

    # Get a hash of all of the zonesets in the cache
    my $cache_zonesets_ref = $self->get_zonesets();
    if ($self->error) {
        return {};
    }

    # Loop through the zoneset hash reference and
    # see if the zone belongs to any of the zonesets.  We will use
    # the temp hash reference $zonesets_ref to store the zonesets with
    # the format $zonesets_ref->{$zoneset} = zoneset_obj
    my $zonesets_ref = {};
    foreach my $cache_zoneset (sort keys %$cache_zonesets_ref) {

        # Check to see if the zoneset object has the zone in it we are looking for.
        my $cache_zonesetobj = $cache_zonesets_ref->{$cache_zoneset};

        # Get the zones for the zoneset
        my $zones_ref = $cache_zonesetobj->get_zones();
        if ($cache_zonesetobj->error()) {
            $self->_set_error("Unable to get the zones for zoneset '$cache_zoneset':\n" .
                              $cache_zonesetobj->errmsg());
        }

        # See if the zone is found in the zoneset reference
        if (defined($zones_ref->{$zone})) {
            if ($type eq "switch") {
                unless ($cache_zonesetobj->is_on_switch()) {
                    if ($cache_zonesetobj->error()) {
                        # Had an error on the method
                        $self->_set_error("Unable to retrieve the 'on_switch' key from the object:\n" .
                                          $cache_zonesetobj->errmsg());
                    } else {
                        # No error...just not on the switch, so skip this one.
                        next;
                    }
                }
            }
            # looks like we have a good match so store the zoneset object in
            # the hash
            $zonesets_ref->{$cache_zoneset} = $cache_zonesetobj;
        } ## end if (defined($cache_zonesetobj...
    } ## end foreach my $cache_zoneset (...
    return $zonesets_ref;
} ## end sub zone_get_zonesets

###############################################################################
# Method:    zoneset_create
# Details:   See POD documentation at the end of this file.
# Objective: Create a zoneset on a switch
###############################################################################
sub zoneset_create {
    my ($self, %args) = @_;

    my ($zoneset);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zonesetobj) and ref($zonesetobj) =~ /ZoneSet/i) {
        $self->_set_err("Unable to retrieve object for zoneset $zoneset");
        return $SwitchUtil::FALSE;
    }

    # Make sure the zoneset doesn't already exist on the switch.
    if ($zonesetobj->exists()) {
        $self->_set_err("Zoneset $zoneset already exists on the switch");
        return $SwitchUtil::FALSE;
    } else {
        if ($zonesetobj->error) {
            $self->_set_err("Failed to check if Zoneset $zoneset exists on the switch: \n" . $zonesetobj->errmsg());
            return $SwitchUtil::FALSE;
        }
    } ## end else [ if ($zonesetobj->exists...

    # Zoneset doesn't exist on the switch so go ahead and create it.
    # First verify the zone objects since we don't want to create a zoneset
    # from zones that don't already exist.  Return false if we can't
    # validate the zones as good.

    my $zones_ref = $zonesetobj->get_zones();
    if (scalar(keys %$zones_ref) == 0) {
        $self->_set_err("Unable to determine the zones for zoneset $zoneset");
        return $SwitchUtil::FALSE;
    }

    # Verify each zone object exists
    foreach my $t_zone (sort keys %$zones_ref) {
        my $t_zoneobj = $zones_ref->{$t_zone};
        if (not $self->zone_exists(-zone => $t_zone)) {
            if (not $self->error()) {
                $self->_set_err("Unable to verify that zone '$t_zone' from zoneset '$zoneset' exists");
            }
            return $SwitchUtil::FALSE;
        } ## end if (not $self->zone_exists...
    } ## end foreach my $t_zone (sort keys...

    # Looks like all of the zones defined on the zoneset object are good so
    # we can continue to create the zoneset on the switch
    if ($zonesetobj->create()) {
        $self->sanlog(-msg => "Zoneset $zoneset has been created on the switch");
        $zonesetobj->{on_switch} = $SwitchUtil::TRUE;
    } else {
        if (not $self->error) {
            $self->_set_err("Unable to create zoneset $zoneset on the switch");
        }
        return $SwitchUtil::FALSE;
    } ## end else [ if ($zonesetobj->create...

    return $SwitchUtil::TRUE;
} ## end sub zoneset_create

###############################################################################
# Method:    zoneset_destroy
# Objective: Destroy a zoneset on a switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zoneset_destroy {
    my ($self, %args) = @_;

    my ($zoneset);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zonesetobj) and ref($zonesetobj) =~ /ZoneSet/i) {
        $self->_set_err("Unable to retrieve object for zoneset $zoneset");
        return $SwitchUtil::FALSE;
    }

    # Make sure the zoneset exists on the switch.
    if (not $zonesetobj->exists()) {
        if ($zonesetobj->error) {
            $self->_set_err("Zoneset $zoneset does not exist on the switch:\n" . $zonesetobj->errmsg());
        } else {
            $self->_set_err("Zoneset $zoneset does not exist on the switch");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $zonesetobj->exists...

    # Go ahead and remove it
    if ($zonesetobj->destroy()) {
        $self->sanlog(-msg => "Zoneset $zoneset has been removed from the switch");
        $zonesetobj->{on_switch} = $SwitchUtil::FALSE;
    } else {
        if ($zonesetobj->error()) {
            $self->_set_err("Failed to delete zoneset $zoneset from the switch:\n" . $zonesetobj->errmsg());
        } else {
            $self->_set_err("Failed to delete zoneset $zoneset from the switch");
        }
        return $SwitchUtil::FALSE;
    } ## end else [ if ($zonesetobj->destroy...

    return $SwitchUtil::TRUE;
} ## end sub zoneset_destroy

###############################################################################
# Method:    zoneset_exists
# Objective: Check the switch to see if the zoneset exists
# Details:   See POD documentation at the end of this file.
###############################s###############################################
sub zoneset_exists {
    my ($self, %args) = @_;

    my ($zoneset);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zonesetobj) and ref($zonesetobj) =~ /ZoneSet/i) {
        $self->_set_err("Unable to retrieve object for zoneset $zoneset");
        return $SwitchUtil::FALSE;
    }

    # Check the switch for the zoneset by calling $zonesetobj->exists();
    if (not $zonesetobj->exists()) {
        if ($zonesetobj->error) {
            $self->_set_err("Unable to verify that zoneset $zoneset exists on the switvh:\n" . $zonesetobj->errmsg());
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $zonesetobj->exists...

    # Everything looks good so return TRUE
    return $SwitchUtil::TRUE;
} ## end sub zoneset_exists

###############################################################################
# Method:    zoneset_verify
# Objective: Verify the contents of a zoneset
# Details:   See POD documentation at the end of this file.
###############################s###############################################
sub zoneset_verify {
    my ($self, %args) = @_;

    my ($zoneset);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zonesetobj) and ref($zonesetobj) =~ /ZoneSet/i) {
        $self->_set_err("Unable to retrieve object for zoneset $zoneset");
        return $SwitchUtil::FALSE;
    }

    # Validate the object with $zonesetobj->validate()
    if (not $zonesetobj->validate()) {
        if ($zonesetobj->error) {
            $self->_set_err("Failed to validate zoneset $zoneset:\n" . $zonesetobj->errmsg());
        } else {
            $self->_set_err("Failed to validate zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $zonesetobj->validate...

    # Everything looks good so return TRUE
    return $SwitchUtil::TRUE;
} ## end sub zoneset_verify

###############################################################################
# Method:    zoneset_add_zone
# Objective: Add a zone to a zoneset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zoneset_add_zone {
    my ($self, %args) = @_;

    my ($zoneset, $zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } elsif (/-?zone/) {

            # Check to make sure that we were passed an zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end elsif (/-?zone/)

    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zonesetobj) and ref($zonesetobj) =~ /ZoneSet/i) {
        $self->_set_err("Unable to retrieve object for zoneset $zoneset");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        $self->_set_err("Unable to retrieve object for zone $zone");
        return $SwitchUtil::FALSE;
    }

    # Check to make sure the zoneset exists on the switch
    if (not $self->zoneset_exists(-zoneset => $zoneset)) {
        if (not $self->error) {
            $self->_set_err("Unable to verify that zoneset '$zoneset' exists");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $self->zoneset_exists...

    # Check to make sure the zone exists
    if (not $self->zone_exists(-zone => $zone)) {
        if (not $self->error) {
            $self->_set_err("Unable to verify that zone $zone exists");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $self->zone_exists...

    # Check to make sure the zone is not already a member of the zoneset
    my $zones_ref = $zonesetobj->get_zones();

    if (defined($zones_ref->{$zone})) {

        # We have found that the zone is already in the zoneset
        # so print a message and bail
        $self->_set_err("Zone $zone was found in zoneset $zoneset");
        return $SwitchUtil::FALSE;
    } ## end if (defined($zones_ref...

    # If we made it this far then we didn't find the zone in the zoneset
    # so proceed with adding it to the switch.
    if (not $zonesetobj->add_zone(-zone => $zoneobj)) {
        if (not $self->error) {
            $self->_set_err("Unable to add zone $zone to zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    } ## end if (not $zonesetobj->add_zone...

    return $SwitchUtil::TRUE;
} ## end sub zoneset_add_zone

###############################################################################
# Method:    zoneset_remove_zone
# Objective: Remove a zone from a zoneset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zoneset_remove_zone {
    my ($self, %args) = @_;

    my ($zoneset, $zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } elsif (/-?zone/) {

            # Check to make sure that we were passed an zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end elsif (/-?zone/)

    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zonesetobj) and ref($zonesetobj) =~ /ZoneSet/i) {
        $self->_set_err("Unable to retrieve object for zoneset $zoneset");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        $self->_set_err("Unable to retrieve object for zone $zone");
        return $SwitchUtil::FALSE;
    }

    # Check to make sure the zone is a member of the zoneset
    my $zones_ref = $zonesetobj->get_zones();

    if (not defined($zones_ref->{$zone})) {

        # We have found that the zone does not exist in the zoneset so
        # so print a message and bail
        $self->_set_err("Zone $zone was not found in zoneset $zoneset");
        return $SwitchUtil::FALSE;
    } ## end if (not defined($zones_ref...

    # If we made it this far then we found the zone in the zoneset
    # so proceed with deleting it from the switch.
    if ($zonesetobj->remove_zone(-zone => $zoneobj)) {
        $self->sanlog(-msg => "Zone $zone was removed from zoneset $zoneset");
    } else {
        if (not $self->error) {
            $self->_set_err("Unable to remove zone $zone from zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    } ## end else [ if ($zonesetobj->remove_zone...

    return $SwitchUtil::TRUE;
} ## end sub zoneset_remove_zone

###############################################################################
# Method:    zoneset_verify_zone
# Objective: Verify that a zone is in the zoneset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub zoneset_verify_zone {
    my ($self, %args) = @_;

    my ($zoneset, $zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } elsif (/-?zone/) {

            # Check to make sure that we were passed an zone
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end elsif (/-?zone/)

    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    # Check to make sure we have the required arguments
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zonesetobj) and ref($zonesetobj) =~ /ZoneSet/i) {
        $self->_set_err("Unable to retrieve object for zoneset $zoneset");
        return $SwitchUtil::FALSE;
    }

    my $zoneobj = $self->get_zones(-zone => $zone);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    unless (defined($zoneobj) and ref($zoneobj) =~ /Zone/i) {
        $self->_set_err("Unable to retrieve object for zone $zone");
        return $SwitchUtil::FALSE;
    }

    # Check to make sure the zone is a member of the zoneset
    my $zones_ref = $zonesetobj->get_zones();
    if (defined($zones_ref->{$zone})) {
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
} ## end sub zoneset_verify_zone

###############################################################################
# Method:    get_active_zoneset
# Objective: Return the active zoneset object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_active_zoneset {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{active_zoneset}) and ref($self->{active_zoneset}) =~ /zoneset/i) {
        # Call $self->active_zoneset_discover to discover the active zoneset
        $self->active_zoneset_discover();
        if ($self->error) {
            $self->_set_err("Unable to discover active zoneset");
            return {};
        } ## end unless ($self->active_zoneset_discover...
    }
    return $self->{active_zoneset};
}

###############################################################################
# Method:    get_aliases
# Objective: Return a hash reference of all alias objects or a single object
#            if an alias is passed as a parameter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_aliases {
    my ($self, %args) = @_;

    my ($alias);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    if (defined($alias) and $alias ne "") {

        # An alias name was passed as a parameter so we are only going
        # to return that object if it exists;
        if (defined($self->{aliases}->{$alias})
            and $self->{aliases}->{$alias} ne "") {
            return $self->{aliases}->{$alias};
        } else {
            return {};
        }
    } else {

        # No alias parameter was passed in so simply return all alias objects
        return $self->{aliases};
    }
} ## end sub get_aliases

###############################################################################
# Method:    get_zones
# Objective: Return a hash reference of all zone objects or a single object if a
#            zone is passed as a parameter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_zones {
    my ($self, %args) = @_;

    my ($zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed a zone name
            if (not length($zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    if (defined($zone) and $zone ne "") {

        # A zone name was passed as a parameter so we are only going
        # to return that object if it exists;
        if (defined($self->{zones}->{$zone})
            and $self->{zones}->{$zone} ne "") {
            return $self->{zones}->{$zone};
        } else {
            return {};
        }
    } else {

        # No zone parameter was passed in so simply return all zone objects
        return $self->{zones};
    }
} ## end sub get_zones

###############################################################################
# Method:    get_zonesets
# Objective: Return a hash reference of all zoneset objects or a single object if a
#            zoneset is passed as a parameter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_zonesets {
    my ($self, %args) = @_;

    my ($zoneset);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed a zoneset
            if (not length($zoneset = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return {};
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    if (defined($zoneset) and $zoneset ne "") {

        # A zoneset name was passed as a parameter so we are only going
        # to return that object if it exists;
        if (defined($self->{zonesets}->{$zoneset}) and $self->{zonesets}->{$zoneset} ne "") {
            return $self->{zonesets}->{$zoneset};
        } else {
            return {};
        }
    } else {

        # No zoneset parameter was passed in so simply return all zoneset objects
        return $self->{zonesets};
    }
} ## end sub get_zonesets

###############################################################################
# Method:    invalidate_cache
# Objective: Invalidate the discovered zone, alias, and zoneset cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub invalidate_cache {
    my ($self) = @_;

    # initialize the object members back to their starting default values
    $self->{aliases}        = {};
    $self->{zones}          = {};
    $self->{zonesets}       = {};
    $self->{active_zoneset} = undef;

    return $SwitchUtil::TRUE;
} ## end sub invalidate_cache

###############################################################################
# Method:    new
# Objective: Create new
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, $hosttype);
    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # check for existence of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanSwitch/) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
        $hosttype = $args{-parent}->get_connection()->{hostp}{hosttype};
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # determine the right platform and require/import/call the appropriate
    # vendor object's new call. We are only supporting Brocade and Cisco.
    # For now, on Brocade, only support FOS based switches, not VDX/NOS based ones.
    if ($hosttype =~ /fcp-switch-cisco/i) {
        require San::SanSwitch::FC::ZoneMgr::CiscoFCZone;
        $self = San::SanSwitch::FC::ZoneMgr::CiscoFCZone->_new(%args);
    } elsif ($hosttype =~ /switch-brocade/i) {
        if ($hosttype =~ /switch-brocade-vdx/i) {
            $self->_set_err("Switch type $hosttype not currently supported by ZoneMgr object");
            return $self;
        } else {
            require San::SanSwitch::FC::ZoneMgr::BrocadeFCZone;
            $self = San::SanSwitch::FC::ZoneMgr::BrocadeFCZone->_new(%args);
        }
    } else {
        $self->_set_err("Switch type $hosttype not currently supported by ZoneMgr object");
        return $self;
    }

    @ISA = qw(Error SanLog San::SanSwitch::ZoneMgr);

    # Call discover to populate the ZoneMgr object with switch data
    unless ($self->discover()) {
        if (not $self->error) {
            $self->_set_err("Unable to discover switch aliases");
        }
        return $self;
    } ## end unless ($self->discover())
    return $self;
} ## end sub new

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _alias_cache_add
# Objective: Add the alias object to the cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _alias_cache_add {
    my ($self, %args) = @_;

    my ($aliasobj);

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias object
            if (not length($aliasobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($aliasobj)) {
        unless (ref($aliasobj) =~ /Alias/i) {
            $self->_set_err("Invalid value for argument '-alias'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $alias = $aliasobj->get_name();
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Unable to determine the name of the alias");
        return $SwitchUtil::FALSE;
    }

    # Just add the alias object onto the hash reference in $self->{aliases}
    $self->{aliases}->{$alias} = $aliasobj;

    return $SwitchUtil::TRUE;
} ## end sub _alias_cache_add

###############################################################################
# Method:    _alias_cache_delete
# Objective: Remove the alias object from the cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _alias_cache_delete {
    my ($self, %args) = @_;

    my ($aliasobj);

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias object
            if (not length($aliasobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($aliasobj)) {
        unless (ref($aliasobj) =~ /Alias/i) {
            $self->_set_err("Invalid value for argument '-alias'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $alias = $aliasobj->get_name();
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Unable to determine the name of the alias");
        return $SwitchUtil::FALSE;
    }

    # Check to see if the alias exists.  If it does, just delete it.
    if ($self->_alias_cache_exists(-alias => $aliasobj)) {
        if (not delete($self->{aliases}->{$alias})) {
            $self->_set_err("Unable to delete alias $alias from the ZoneMgr alias cache");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Alias $alias does not exist in the ZoneMgr alias cache");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub _alias_cache_delete

###############################################################################
# Method:    _alias_cache_exists
# Objective: Checks to see if an alias exists in the ZoneManager cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _alias_cache_exists {
    my ($self, %args) = @_;

    my ($aliasobj);

    foreach (keys %args) {
        if (/-?alias/) {

            # Check to make sure that we were actually passed an alias object
            if (not length($aliasobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?alias/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($aliasobj)) {
        unless (ref($aliasobj) =~ /Alias/i) {
            $self->_set_err("Invalid value for argument '-alias'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-alias'");
        return $SwitchUtil::FALSE;
    }

    my $alias = $aliasobj->get_name();
    unless (defined($alias) and $alias ne "") {
        $self->_set_err("Unable to determine the name of the alias");
        return $SwitchUtil::FALSE;
    }

    # check to see if there is an alias key by the name of $alias
    # in the cache.
    if (defined($self->{aliases}->{$alias})
        and $self->{aliases}->{$alias} ne "") {
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
} ## end sub _alias_cache_exists

###############################################################################
# Method:    _zone_cache_add
# Objective: Add the zone object to the cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _zone_cache_add {
    my ($self, %args) = @_;

    my ($zoneobj);

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed an zone object
            if (not length($zoneobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($zoneobj)) {
        unless (ref($zoneobj) =~ /Zone/i) {
            $self->_set_err("Invalid value for argument '-zone'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zone = $zoneobj->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Unable to determine the name of the zone");
        return $SwitchUtil::FALSE;
    }

    # Just add the zone object onto the hash reference in $self->{zones}
    $self->{zones}->{$zone} = $zoneobj;

    return $SwitchUtil::TRUE;
} ## end sub _zone_cache_add

###############################################################################
# Method:    _zone_cache_delete
# Objective: Remove the zone object from the cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _zone_cache_delete {
    my ($self, %args) = @_;

    my ($zoneobj);

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed an zone object
            if (not length($zoneobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($zoneobj)) {
        unless (ref($zoneobj) =~ /Zone/i) {
            $self->_set_err("Invalid value for argument '-zone'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zone = $zoneobj->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Unable to determine the name of the zone");
        return $SwitchUtil::FALSE;
    }

    # Check to see if there is a hash key for the zone name in question
    # and if there is, just delete it.
    if ($self->_zone_cache_exists(-zone => $zoneobj)) {
        if (not delete($self->{zones}->{$zone})) {
            $self->_set_err("Unable to delete zone $zone from the ZoneMgr zone cache");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Zone $zone does not appear to exist in the ZoneMgr zone cache");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub _zone_cache_delete

###############################################################################
# Method:    _zone_cache_exists
# Objective: Checks to see if a zone exists in the ZoneManager cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _zone_cache_exists {
    my ($self, %args) = @_;

    my ($zoneobj);

    foreach (keys %args) {
        if (/-?zone/) {

            # Check to make sure that we were actually passed an zone object
            if (not length($zoneobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zone/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($zoneobj)) {
        unless (ref($zoneobj) =~ /Zone/i) {
            $self->_set_err("Invalid value for argument '-zone'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-zone'");
        return $SwitchUtil::FALSE;
    }

    my $zone = $zoneobj->get_name();
    unless (defined($zone) and $zone ne "") {
        $self->_set_err("Unable to determine the name of the zone");
        return $SwitchUtil::FALSE;
    }

    # check to see if there is a zone key by the name of $zone
    # in the cache.
    if (defined($self->{zones}->{$zone}) and $self->{zones}->{$zone} ne "") {
        return $SwitchUtil::TRUE;
    } else {
        return $SwitchUtil::FALSE;
    }
} ## end sub _zone_cache_exists

###############################################################################
# Method:    _zoneset_cache_add
# Objective: Add the zoneset object to the cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _zoneset_cache_add {
    my ($self, %args) = @_;

    my ($zonesetobj);

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset object
            if (not length($zonesetobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($zonesetobj)) {
        unless (ref($zonesetobj) =~ /ZoneSet/i) {
            $self->_set_err("Invalid value for argument '-zoneset'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    my $zoneset = $zonesetobj->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Unable to determine the name of the zoneset");
        return $SwitchUtil::FALSE;
    }

    # Just add a new key in the hash for the zoneset with the object as the value
    $self->{zonesets}->{$zoneset} = $zonesetobj;

    return $SwitchUtil::TRUE;
} ## end sub _zoneset_cache_add

###############################################################################
# Method:    _zoneset_cache_delete
# Objective: Remove the zoneset object from the cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _zoneset_cache_delete {
    my ($self, %args) = @_;

    my ($zonesetobj);

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset object
            if (not length($zonesetobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($zonesetobj)) {
        unless (ref($zonesetobj) =~ /ZoneSet/i) {
            $self->_set_err("Invalid value for argument '-zoneset'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    my $zoneset = $zonesetobj->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Unable to determine the name of the zoneset");
        return $SwitchUtil::FALSE;
    }

    # Check to see if there is a hash key for the zoneset name in question
    # and if there is, just delete it.
    if ($self->_zoneset_cache_exists(-zoneset => $zonesetobj)) {
        if (not delete($self->{zonesets}->{$zoneset})) {
            $self->_set_err("Unable to delete zoneset $zoneset from the ZoneMgr zoneset cache");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Zoneset $zoneset does not appear to exist in the ZoneMgr zoneset cache");
        return $SwitchUtil::FALSE;
    }

    return $SwitchUtil::TRUE;
} ## end sub _zoneset_cache_delete

###############################################################################
# Method:    _zoneset_cache_exists
# Objective: Checks to see if a zone exists in the ZoneManager cache
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _zoneset_cache_exists {
    my ($self, %args) = @_;

    my ($zonesetobj);

    foreach (keys %args) {
        if (/-?zoneset/) {

            # Check to make sure that we were actually passed an zoneset object
            if (not length($zonesetobj = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $SwitchUtil::FALSE;
            }
        } ## end if (/-?zoneset/)
    } ## end foreach (keys %args)

    # Check to make sure we have the required arguments
    if (defined($zonesetobj)) {
        unless (ref($zonesetobj) =~ /ZoneSet/i) {
            $self->_set_err("Invalid value for argument '-zoneset'");
            return $SwitchUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-zoneset'");
        return $SwitchUtil::FALSE;
    }

    my $zoneset = $zonesetobj->get_name();
    unless (defined($zoneset) and $zoneset ne "") {
        $self->_set_err("Unable to determine the name of the zoneset");
        return $SwitchUtil::FALSE;
    }

    # check to see if there is a zoneset key by the name of $zoneset
    # in the cache.
    my $cache_zonesets_ref = $self->get_zonesets();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    # Make sure we have the zoneset in the cache and that it isn't empty
    if (defined($cache_zonesets_ref->{$zoneset})) {
        if (scalar(keys %{$cache_zonesets_ref->{$zoneset}}) > 0) {
            return $SwitchUtil::TRUE;
        } else {

            # We should ever end up in this situation since a zoneset must
            # have at leasrt one zone
            $self->_set_err("Cache for zoneset $zoneset is empty");
            return $SwitchUtil::FALSE;
        } ## end else [ if (scalar(keys %{$cache_zonesets_ref...
    } else {
        return $SwitchUtil::FALSE;
    }
} ## end sub _zoneset_cache_exists

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

San::SanSwitch::ZoneMgr.pm

=head1 Author

Derrick Shultz

=head1 Description

Provides an abstract API interface to Alias, Zone, and ZoneSet operations on Brocade and Cisco switches.

=head1 See Also

Package inherits from the following modules. See modules for interface detail

=over 4

=item * B<San::SanLog.pm>

=item * B<Error.pm>

=back

=head1 Attributes

=head2 B<active_zoneset>

Object of the active zoneset.  Defaults to undef.

=head2 B<aliases>

Reference to a hash of alias objects with a key of the alias name and the value of the alias object.  Defaults to {}.

=head2 B<parent>

ZoneMgr object = Reference to a parent SanSwitch object

=head2 B<timeout>

String containing an optional command timeout.  Default is 120.

=head2 B<vsan>

String containing an optional vsan id.

=head2 B<zones>

Reference to a hash of zone objects with a key of the zone name and the value of the zone object.  Defaults to {}.

=head2 B<zonesets>

Reference to a hash of zoneset objects with a key of the zoneset name and the value of the zoneset object.  Defaults to {}.

=head1 Public Methods

=head2 B<alias_create()>

Creates the alias on the switch by calling the appropriate method on the alias object.

=head3 Inputs

=over 4

=item * B<-alias>

String containing the alias to create

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->alias_create(-alias => $alias_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to create alias $alias");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<alias_destroy()>

Removes the alias from the switch by calling the appropriate method on the alias object.  The alias should not
currently be a member of any zones.

=head3 Inputs

=over 4

=item * B<-alias>

String containing the alias to remove

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->alias_destroy(-alias => $alias_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to destroy alias $alias");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<alias_exists()>

Checks for the existence of the alias on the switch by calling the appropriate method on the alias object.

=head3 Inputs

=over 4

=item * B<-alias>

String containing the alias to check for existence

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->alias_exists(-alias => $alias_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to check if alias $alias exists");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<alias_get_zones()>

Returns a list of zones that the alias currently belongs to.

=head3 Inputs

=over 4

=item * B<-alias>

String containing the alias look up in the zone cache

=back

=head3 Outputs

On Success, returns a reference to a hash containing the zones that the alias belongs to and their associated
objects.  The key of the hash is the zone name and the value is the zone object.  On failure, the error flag is set,
the errmsg is set to the failure reason, and an anonymous hash is returned ({}).  If there is no error
and the member is not part of an alias, an anonymous hash is returned ({}).

=head3 Examples

    my $zones_ref = $self->alias_get_zones(-alias => $alias_name);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    foreach my $zone (keys %$zones_ref) {
        my $zoneobj = $zones_ref->{$zone};
        # Do stuff with the object
    }

=head2 B<alias_verify()>

Checks for the existence of the alias on the switch and verifies that it's members matches the alias object members.

=head3 Inputs

=over 4

=item * B<-alias>

String containing the alias to verify

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->alias_verify(-alias => $alias_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to verify alias $alias");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<discover()>

Performs discovery of the aliases, zones, and zonesets associated with the ZoneMgr object by making calls to the
appropriate switch specific discovery routines of alias_discover(), zone_discover(), zoneset_discover(), and
active_zoneset_discover()

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE on success. If an error occurs, the error flag is set and errmsg is
set to the failure reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error attempting to discover the alias, zones, and zoneset information");
        }
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<get_active_zoneset()>

Returns the object of the active zoneset

=head3 Inputs

None

=head3 Outputs

Returns the object of the active zoneset. No error checking is done and no error flags are set.

=head3 Examples

    my $active_zoneset_obj = $self->get_active_zoneset();
    if (scalar(keys %$active_zoneset_obj) > 1) {
        if (ref($active_zoneset_obj) !~ /zoneset/i) {
            $self->_set_err("There was an issue with the active zoneset object");
            return $SwitchUtil::FALSE;
        }
    } else {
        # There wasn't an active zoneset object defined on the ZoneMgr object.
        # This isn't an error.
    }


=head2 B<get_aliases()>

Returns a reference to an anonymous hash containing a single alias object or all
alias objects defined on the ZoneMgr object.

=head3 Inputs

=over 4

=item * B<-alias (optional)>

String containing the name of the alias that you wish to retrieve the object for.

=back

=head3 Outputs

If the -alias parameter is specified, the object for that alias will be returned or an empty anonmymous hash
will be returned if no object exists for the alias.  If the -alias parameter is not specified, then all alias
objects will be returned in an anonymous hash with a key of the alias name and the value of the alias object.
If an error occurs, the error flag is set, the errmsg is set to the failure reason and an empty anonymous
hash is returned.

=head3 Examples

    # Get a single object
    my $aliasobj = $self->get_aliases(-alias => $alias_name);
    if (scalar(keys %$aliasobj) == 0) {
        if ($self->error) {
            # We had an error so return false
            return $SwitchUtil::FALSE;
        } else {
            # This isn't an error.  There just isn't an object for this alias.
            # Code should decide what to do in this case.
        }
    }

    # Get all alias objects
    my $aliasobj_ref = $self->get_aliases();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    if (scalar(keys %$aliasobj_ref) > 0) {
        # We found alias objects
    } else {
        # No alias objects were found but there was not
        # an error.
    }

=head2 B<get_zones()>

Returns a reference to an anonymous hash containing a single zone object or all
zone objects defined on the ZoneMgr object.

=head3 Inputs

=over 4

=item * B<-zone (optional)>

String containing the name of the zone that you wish to retrieve the object for.

=back

=head3 Outputs

If the -zone parameter is specified, the object for that zone will be returned or an empty anonmymous hash
will be returned if no object exists for the zone.  If the -zone parameter is not specified, then all zone
objects will be returned in an anonymous hash with a key of the zone name and the value of the zone object.
If an error occurs, the error flag is set, the errmsg is set to the failure reason and an empty anonymous
hash is returned.

=head3 Examples

    # Get a single object
    my $zoneobj = $self->get_zones(-zone => $zone_name);
    if (scalar(keys %$zoneobj) == 0) {
        if ($self->error) {
            # We had an error so return false
            return $SwitchUtil::FALSE;
        } else {
            # This isn't an error.  There just isn't an object for this zone.
            # Code should decide what to do in this case.
        }
    }

    # Get all zone objects
    my $zonesobj_ref = $self->get_zones();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    if (scalar(keys %$zonesobj_ref) > 0) {
        # We found zone objects
    } else {
        # No zone objects were found but there was not
        # an error.
    }

=head2 B<get_zonesets()>

Returns a reference to an anonymous hash containing a single zoneset object or all
zoneset objects defined on the ZoneMgr object.

=head3 Inputs

=over 4

=item * B<-zoneset (optional)>

String containing the name of the zoneset that you wish to retrieve the object for.

=back

=head3 Outputs

If the -zoneset parameter is specified, the object for that zoneset will be returned or an empty anonmymous hash
will be returned if no object exists for the zoneset.  If the -zoneset parameter is not specified, then all zoneset
objects will be returned in an anonymous hash with a key of the zoneset name and the value of the zoneset object.
If an error occurs, the error flag is set, the errmsg is set to the failure reason and an empty anonymous
hash is returned.

=head3 Examples

    # Get a single object
    my $zonesetobj = $self->get_zonesets(-zoneset => $zoneset_name);
    if (scalar(keys %$zonesetobj) == 0) {
        if ($self->error) {
            # We had an error so return false
            return $SwitchUtil::FALSE;
        } else {
            # This isn't an error.  There just isn't an object for this zoneset.
            # Code should decide what to do in this case.
        }
    }

    # Get all zoneset objects
    my $zonesetsobj_ref = $self->get_zonesets();
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    if (scalar(keys %$zonesetsobj_ref) > 0) {
        # We found zoneset objects
    } else {
        # No zoneset objects were found but there was not
        # an error.
    }

=head2 B<invalidate_cache()>

Invalidates the ZoneMgr cache.

=head3 Inputs

None

=head3 Outputs

Returns $SwitchUtil::TRUE.  No error flags will be set.

=head3 Examples

    my $status;
    $status = $self->invalidate_cache();

    # Re-run discover
    $status = $self->discover();
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an error while attempting to run ZoneMgr discovery");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<member_get_aliases()>

Returns the alias information for the provided member.

=head3 Inputs

=over 4

=item * B<-member>

String containing member to look up.

=back

=head3 Outputs

On Success, returns a reference to a hash containing the aliases that the member belongs to and their associated
objects.  The key of the hash is the alias name and the value is the alias object.  On failure, the error flag is set,
the errmsg is set to the failure reason, and an anonymous hash is returned ({}).  If there is no error
and the member is not part of an alias, an anonymous hash is returned ({}).

=head3 Examples

    my $alias_ref = $self->member_get_aliases(-member => $member);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    foreach my $alias (keys %$alias_ref) {
        my $aliasobj = $alias_ref->{$alias};
        # Do stuff with the object
    }

=head2 B<new()>

Instantiates the ZoneMgr object.

=head3 Inputs

=over 4

=item * B<-parent>

Parent Switch object.

=item * B<-timeout (optional)>

String containing the timeout to use for commands.  Defaults to 120.

=item * B<-vsan (optional)>

String containing the vsan ID.  Defaults to undef.

=back

=head3 Outputs

Returns a ZoneMgr object. If an error occurs, the error flag is set and errmsg is
set to the failure reason before the object is returned.

=head3 Examples

    my $object = San::SanSwitch::ZoneMgr::new(-parent => $self,
                                              -timeout => "60");
    if ($object->error) {
        $self->_set_err($object->errmsg());
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zone_add_alias()>

Adds an alias to a zone

=head3 Inputs

=over 4

=item * B<-zone>

String containing the zone to add the alias to.

=item * B<-alias>

String containing the alias to add to the zone.

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zone_add_alias(-zone => $zone_name, -alias => $alias_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to add alias $alias to zone $zone");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zone_create()>

Creates the zone on the switch by calling the appropriate method on the zone object.

=head3 Inputs

=over 4

=item * B<-zone>

String containing the zone to create

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zone_create(-zone => $zone_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to create zone $zone");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zone_destroy()>

Removes the zone from the switch by calling the appropriate method on the zone object.  The zone should not
currently be a member of any zonesets.

=head3 Inputs

=over 4

=item * B<-zone>

String containing the zone to remove

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zone_destroy(-zone => $zone_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to destroy zone $zone");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zone_exists()>

Checks for the existence of the zone on the switch by calling the appropriate method on the zone object.

=head3 Inputs

=over 4

=item * B<-zone>

String containing the zone to check for existence

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zone_exists(-zone => $zone_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to check if zone $zone exists");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zone_get_aliases()>

Returns a hash reference of aliases that the zone contains

=head3 Inputs

=over 4

=item * B<-zone>

String containing the zone look up in the zoneset cache

=back

=head3 Outputs

On Success, returns a reference to a hash containing the aliases that the zone contains and their associated
objects.  The key of the hash is the alias name and the value is the alias object.  On failure, the error flag
is set, the errmsg is set to the failure reason, and an anonymous hash is returned ({}).  If there is no error
and the zone is not part of a zoneset, an anonymous hash is returned ({}).

=head3 Examples

    my $alias_ref = $self->zone_get_aliases(-zone => $zone_name);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    foreach my $alias (keys %$aliases_ref) {
        my $aliasobj = $aliases_ref->{$alias};
        # Do stuff with the object
    }

=head2 B<zone_get_zonesets()>

Returns a list of zonesets that the zone currently belongs to.

=head3 Inputs

=over 4

=item * B<-zone>

String containing the zone look up in the zoneset cache

=back

=head3 Outputs

On Success, returns a reference to a hash containing the zonesets that the zone belongs to and their associated
objects.  The key of the hash is the zoneset name and the value is the zoneset object.  On failure, the error flag
is set, the errmsg is set to the failure reason, and an anonymous hash is returned ({}).  If there is no error
and the zone is not part of a zoneset, an anonymous hash is returned ({}).

=head3 Examples

    my $zonesets_ref = $self->zone_get_zonesets(-zone => $zone_name);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    }

    foreach my $zoneset (keys %$zonesets_ref) {
        my $zonesetobj = $zonesets_ref->{$zoneset};
        # Do stuff with the object
    }

=head2 B<zone_remove_alias()>

Removes an alias from a zone

=head3 Inputs

=over 4

=item * B<-zone>

String containing the zone to remove the alias from.

=item * B<-alias>

String containing the alias to remove from the zone.

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zone_remove_alias(-zone => $zone_name, -alias => $alias_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to remove alias $alias from zone $zone");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zone_verify()>

Checks for the existence of the zone on the switch and verifies that it's aliases match the zone object aliases.

=head3 Inputs

=over 4

=item * B<-zone>

String containing the zone to verify

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zone_verify(-zone => $zone_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to verify zone $zone");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zone_verify_alias()>

Verifies that an alias exists in a zone and that all information matches the cache.

=head3 Inputs

=over 4

=item * B<-zone>

String containing the zone to verify the alias against

=item * B<-alias>

String containing the alias verify against the zone

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zone_verify_alias(-zone => $zone_name, -alias => $alias_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to verify alias $alias against zone $zone");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zoneset_add_zone()>

Adds a zone to a zoneset

=head3 Inputs

=over 4

=item * B<-zoneset>

String containing the zoneset to add the zone to.

=item * B<-zone>

String containing the zone to add to the zoneset.

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zoneset_add_zone(-zoneset => $zoneset_name, -zone => $zone_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to add zone $zone to zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zoneset_create()>

Creates the zoneset on the switch by calling the appropriate method on the zoneset object.

=head3 Inputs

=over 4

=item * B<-zoneset>

String containing the zoneset to create

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zoneset_create(-zoneset => $zoneset_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to create zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zoneset_destroy()>

Removes the zoneset from the switch by calling the appropriate method on the zoneset object.

=head3 Inputs

=over 4

=item * B<-zoneset>

String containing the zoneset to remove

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zoneset_destroy(-zoneset => $zoneset_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to destroy zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zoneset_exists()>

Checks for the existence of the zoneset on the switch by calling the appropriate method on the zoneset object.

=head3 Inputs

=over 4

=item * B<-zoneset>

String containing the zoneset to check for existence

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zoneset_exists(-zoneset => $zoneset_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to check if zoneset $zoneset exists");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zoneset_remove_zone()>

Removes a zone from a zoneset

=head3 Inputs

=over 4

=item * B<-zoneset>

String containing the zoneset to remove the zone from.

=item * B<-zone>

String containing the zone to remove from the zoneset.

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zoneset_remove_zone(-zoneset => $zoneset_name, -zone => $zone_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to remove zone $zone from zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zoneset_verify()>

Checks for the existence of the zoneset on the switch and verifies that it's zones match the zoneset object zones.

=head3 Inputs

=over 4

=item * B<-zoneset>

String containing the zoneset to verify

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zoneset_verify(-zoneset => $zoneset_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to verify zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    }

=head2 B<zoneset_verify_zone()>

Verifies that a zone exists in a zoneset and that all information matches the cache.

=head3 Inputs

=over 4

=item * B<-zoneset>

String containing the zoneset to verify the zone against

=item * B<-zone>

String containing the zone verify against the zoneset

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->zoneset_verify_zone(-zoneset => $zoneset_name, -zone => $zone_name);
    if ($status == $SwitchUtil::FALSE) {
        if (not $self->error) {
            $self->_set_err("There was an issue while attempting to verify zone $zone against zoneset $zoneset");
        }
        return $SwitchUtil::FALSE;
    }

=head1 Private Methods

=head2 B<_alias_cache_add()>

Adds an alias object to the ZoneMgr cache

=head3 Inputs

=over 4

=item * B<-alias>

Alias object

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->_alias_cache_add(-alias => $aliasobj);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<_alias_cache_delete()>

Deletes an alias object from the ZoneMgr cache

=head3 Inputs

=over 4

=item * B<-alias>

Alias object

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->_alias_cache_delete(-alias => $aliasobj);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<_alias_cache_exists()>

Checks to see if an alias exists in the ZoneMgr cache

=head3 Inputs

=over 4

=item * B<-alias>

Alias object

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->_alias_cache_exists(-alias => $aliasobj);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zone_cache_add()>

Adds a zone object to the ZoneMgr cache

=head3 Inputs

=over 4

=item * B<-zone>

zone object

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->_zone_cache_add(-zone => $zoneobj);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zone_cache_delete()>

Deletes a zone object from the ZoneMgr cache

=head3 Inputs

=over 4

=item * B<-zone>

zone object

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->_zone_cache_delete(-zone => $zoneobj);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zone_cache_exists()>

Checks to see if a zone exists in the ZoneMgr cache

=head3 Inputs

=over 4

=item * B<-zone>

zone object

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->_zone_cache_exists(-zone => $zoneobj);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zoneset_cache_add()>

Adds a zoneset object to the ZoneMgr cache

=head3 Inputs

=over 4

=item * B<-zoneset>

zoneset object

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->_zoneset_cache_add(-zoneset => $zonesetobj);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zoneset_cache_delete()>

Deletes a zoneset object from the ZoneMgr cache

=head3 Inputs

=over 4

=item * B<-zoneset>

zoneset object

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->_zoneset_cache_delete(-zoneset => $zonesetobj);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=head2 B<zoneset_cache_exists()>

Checks to see if a zoneset exists in the ZoneMgr cache

=head3 Inputs

=over 4

=item * B<-zoneset>

zoneset object

=back

=head3 Outputs

Returns $SwitchUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $SwitchUtil::FALSE is returned.

=head3 Examples

    my $status = $self->_zoneset_cache_exists(-zoneset => $zonesetobj);
    if ($self->error) {
        return $SwitchUtil::FALSE;
    } else {
        return $SwitchUtil::TRUE;
    }

=cut
