# Copyright (c) 2007-2010 NetApp Inc.
# All rights reserved.
#
## @summary Provides common methods that
## should be inherited by other object oriented
## packages
##
## @author korry@netapp.com
## @status Shared
## @pod here

package SanUtil;

use lib();

#
# Compiler directives.
#
use strict;

# module import
#
use Data::Dumper;
use Filer::Version qw(get_ontap_version);
use San::SanLog;
use CmdsDriver;
use Error;
use Exporter;
use Tharn qw($Log);
use Tharn;
use Net::Ping;
use Net::IP;

use vars qw(
  @EXPORT
  @ISA
  );

no warnings qw(redefine);

# constant for blocks host ZAPI calls
use constant ZAPI_TIMEOUT     => 600;
use constant ZAPI_FAILURE     => -1;
use constant PROMPT_NORM      => '(::[^>#%]*)?\*?[>#%]';
use constant DEFAULT_USERNAME => 'diag';
use constant DEFAULT_PASSWD   => 'netapp1!';
use constant DEFAULT_CONNTYPE => 'ssh-persist';

#
# Inherit from the Error, SanLog classes
#
@ISA = qw(Error SanLog);

# experted variables
$SanUtil::TRUE  = 1;
$SanUtil::FALSE = 0;
$SanUtil::PASS  = 1;
$SanUtil::FAIL  = 0;

# Define global results hash for nate_* functions
my $global_results;

# Define running NATE version
my $NATE_version = &NATE::version_string();

###############################################################################
# Method:    filer_connect
# Objective: Make a connection to a storage controller
# Input:     connection type (rsh, telnet, console, etc) default = rsh
# Return: connection object
###############################################################################
sub filer_connect {
    my ($self, %args) = @_;
    my ($connect, $conn_str, $conn_type, $filer,);

    # Variable init.
    $self->{connected} = 0;

    $filer = Tharn::host($self->{name});

    $self->{default_ip} = $filer->{default_ip};

    $conn_str = "#rsh";

    $self->_clear_err();

    #  check for overrides args
    #
    foreach (keys %args) {
        if (/-?conn_type$/) {
            if (not length($self->{conn_type} = $args{$_})
                or $args{$_} !~ /(rsh|telnet|console)/) {
                $self->_set_err(
                       ref($self) . "::new() invalid value for \"-conn_type\":",
                       "$args{$_}");
                return $self;
            }
            $conn_str = "#" . $self->{conn_type};
        } else {
            $self->_set_err(
                        ref($self) . "::new() Invalid parameter specified: $_");
            return $self;
        }
    }    # end foreach arg

    $connect = Tharn::connect(
        $conn_str . ";timeout=$self->{timeout}",
        -eval => sub {
            if (@_) {
                my $errmsg = join("", @_);
                $errmsg =~ s/\n//g;
                $self->_set_err($errmsg);
                $Tharn::InEND = 1;    #Don't abort
                Tharn::logresult("CONFIG", "\n@_");
                return $self;
            }
        });

    return $connect;
}    # end of filer_conn method

###############################################################################
# Method:    get_filer_conn
# Objective: Get filer connection
# Input:     None
# Return:    filer connection object
###############################################################################
sub get_filer_conn {
    my ($self) = @_;
    $self->sanlog(-msg => ref($self) . "::" . "get filer connection object",
                  -msg_level => 3);
    return (defined($self->{connection})) ? $self->{connection} : undef;
}

##############################################################################
# Method:    filer_ontap_version_get
# Objective: Get ONTAP version
# Input:     None
# Return:    ONTAP version number (x.y, 7.2, 7.3, etc.)
###############################################################################
sub filer_ontap_version_get {
    my ($self, %args) = @_;
    my $filer = $self->{name};
    my $version;
    my $tharnhost;

    $self->_clear_err();

    #
    #  check for overrides args
    #
    foreach (keys %args) {
        if (/-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err(
                    ref($self)
                      . "::filer_ontap_version_get() invalid value for \"-filer\":",
                    "$args{$_}");
                return $SanUtil::FAIL;
            }
        } else {
            $self->_set_err(
                ref($self)
                  . "::filer_ontap_version_get() Invalid parameter specified: $_"
                  );
            return $SanUtil::FAIL;
        }
    }    # end foreach arg

    unless (defined($filer)) {
        $self->_set_err(
                  ref($self)
                    . "::filer_ontap_version_get() Undefined -filer parameter");
        return $SanUtil::FAIL;
    }

    # get host object
    $tharnhost = Tharn::host($filer);

    $version = (get_ontap_version($tharnhost));
    $self->sanlog(-msg => ref($self) . "::" . "filer_version = " . $version,
                  -msg_level => 2);

    return $version;

}    # end of filer_ontap_version_get method

###############################################################################
# Method:    strip_dash
# Objective: strip the dash '-' from argument hash
# Input:     hash argument
# Return:    hash argument
###############################################################################
sub strip_dash {
    my (@args) = @_;

    # we want to maintain compatibility by accepting '-' in argument hash
    # however, most if not all of attribute/value pairs hash arguments
    # from 10-mode SAN library will be used to create the 'get' accessor
    # method on fly. for this reason, we have to strip out the dash from
    # the attribute of the hash.
    # We use a +2 in the for loop because if you use a hash as an array
    # then the array is of the form [key,value,key,value] and we only
    # want to update the keys. By updating the passed in hash this way
    # we dont have to have the extra delete(key) call that we would need
    # if using the hash as a hash.
    for (my $i=0; $i<(scalar @args); $i=$i+2) {
        $args[$i] =~  s/^-//;
    }

    my %args = @args;
    return %args;
}    # end of strip_dash method

###############################################################################
# Method:    xml2hash
# Objective: Convert XML output string in nested hash
# Input:     XML out string
# Return:    Nested hash based on XML tags
###############################################################################
sub xml2hash {
    my ($self, $xml_str) = @_;

    my $XML_obj = new XML::Simple;
    my $data    = $XML_obj->XMLin($xml_str);

    return $data;

}    # end of xml2hash method

###############################################################################
# Method:    dumper_get
# Objective: Get output from object
# Input:     Any hash or array object
# Return:    string from the dump
###############################################################################
sub dumper_get {
    my ($self, $obj, $debug_level) = @_;

    # return empty if debug level < 7
    return "" if ($debug_level < 7);

    # lets' create new Data::Dumper object
    my $d_obj = Data::Dumper->new([$obj], [ref($obj)]);

    # set depth to 1 if debug_level 7 >=< 9
    # set depth to 2 if debug_level 9 >=< 11
    # otherwise dump everything. extreme case
    if ($debug_level < 9) {
        $d_obj->Maxdepth(1);
    } elsif ($debug_level < 11) {
        $d_obj->Maxdepth(2);
    }

    return $d_obj->Dump();
}    # end of dumper_get method

###############################################################################
# Method:    cmds_driver_get
# Objective: Get commands driver object
# Input:     None
# Return:    command driver object
###############################################################################
sub cmds_driver_get {
    my ($self) = @_;
    $self->sanlog(-msg       => ref($self) . "::" . "get cmds driver object",
                  -msg_level => 3);
    return (defined($self->{cmd_obj})) ? $self->{cmd_obj} : undef;
}

###############################################################################
# Method:    cmds_driver_exist
# Objective: check if commands driver does exist
# Input:     None
# Return:    TRUE if commands driver exists. If not FALSE
###############################################################################
sub cmds_driver_exist {
    my ($self) = @_;
    $self->sanlog(
                -msg => ref($self) . "::" . "check for existing of cmds driver",
                -msg_level => 3);
    return (defined($self->{cmd_obj}) and ref($self->{cmd_obj}) =~ /CmdsDriver/)
      ? $SanUtil::TRUE
      : $SanUtil::FALSE;
}

###############################################################################
# Method:    log_cmd_results
# Objective: Prints out the results hash of a cmdsDriver object command call.
# Input:     A reference to the results hash of the command. (This is what is
#            returned by CmdsDriver::execute_cmd().
# Return:    None
###############################################################################
sub log_cmd_results {
    my ($self, $resultsHash_ref) = @_;
    my $dumpOut;

    my $print_string = "\n  ------------------------------------------\n";
    $print_string .= "              Results Hash\n";
    $print_string .= "  ------------------------------------------\n";
    foreach my $key (keys %{$resultsHash_ref}) {
        my $value = $resultsHash_ref->{$key};
        if (ref($value) =~ "HASH") {
            $dumpOut = Dumper($value);
            $value   = $dumpOut;
        }
        $print_string .= "    $key => $value\n";
    }
    $print_string .= "  ------------------------------------------";

    $self->sanlog(-msg       => $print_string,
                  -msg_level => 7);
}    # end log_cmd_results

###############################################################################
# Subroutine Name: system_zapi
# Objective: Execute commands to the filer via ZAPI and SanFiler.pm's CmdDriver
#            interface.
# Returns:   1) output of command execution in scalar context
#            2) status of execute on filer (0=FAIL, 1=PASS)
#            3) output of command in list context
#
# Input:     1) Filer Name
#            2) Cmd to be executed
#            3) Priv level on filer to execute Cmd in
###############################################################################
sub system_zapi {
    my ($filer, $cmd, $priv_level, $timeout) = @_;

    my ($output, $status, $args, @l_output);

    # determine timeout
    if (not defined $timeout or $timeout eq "") {
        $timeout = ZAPI_TIMEOUT;
    }
    $args->{-cmd_timeout} = $timeout;

    # Execute the system-cli ZAPI
    # Execute in a retry loop of 3 attempts.
    foreach my $retry (1 .. 3) {
        ($output, $status, @l_output) =
          &cmds_driver_zapi($filer, "system-cli", $priv_level, $cmd, $args);

        # If we did not fail, kick out of the retry loop
        if ($status != ZAPI_TIMEOUT) {
            last;
        } else {
            my $retry_num = ($retry + 1);
            &SanUtil::nate_log(
                '-comment',
                "SanUtil::system_zapi() system-cli ZAPI command ($cmd) failed, retry will be attempted ($retry_num of 3)"
                );
            &SanUtil::nate_sleep(10);
        }
    } ## end foreach my $retry (1 .. 3)

    return ($output, $status, @l_output);

} ## end sub system_zapi

###############################################################################
# Subroutine Name: cmds_driver_zapi
# Objective: Execute commands to the filer via ZAPI and SanFiler.pm's CmdDriver
#            interface. Command can either be via 'system-cli' ZPI by sending
#            type as 'system-cli' or it can execute a specific ZAPI.
#
# Returns:   1) output of command execution in scalar context
#            2) status of execute on filer (0=FAIL, 1=PASS, -1=FAIL in ZAPI
#            3) output of command in list context.
#
# Inputs:    1) Filer Name
#            2) Type of the ZAPI to execute
#            3) Priv level on filer to execute Cmd in,
#            4) Cmd to be executed,
#            5) Any additional params you need to pass to ZAPI
###############################################################################
sub cmds_driver_zapi {
    my ($filer, $type, $priv_level, $cmd, $add_args) = @_;

    my ($cmd_obj, $filer_obj, $version, $args, @args);
    my ($output, @list_output);

    # default to priv level diag if not specified
    if (not defined($priv_level) or $priv_level eq "") {
        $priv_level = "diag";
    }

    # split command to get in a format ZAPI accepts
    if (defined($cmd) and $cmd ne "") {
        @args = split(/\s+/, $cmd);
    } else {
        $cmd = $type;
    }

    # get from cache if available
    if (defined($SanUtil::g_filer_objs->{$filer})) {
        $cmd_obj = $SanUtil::g_filer_objs->{$filer}->{cmd_obj};
    } else {

        # create new CmdsDriver object
        $cmd_obj = new CmdsDriver();

        if ($cmd_obj->error()) {
            my $errmsg = join "\n", $cmd_obj->errmsg();
            &SanUtil::nate_log(
                '-comment',
                "SanUtil::cmds_driver_zapi() Failed to instantiate CmdsDriver object ($errmsg)"
                );

            # return ZAPI_FAILURE for status and let the caller check.
            return ("", ZAPI_FAILURE, ());
        } ## end if ($cmd_obj->error())

        $SanUtil::g_filer_objs->{$filer}->{cmd_obj} = $cmd_obj;
    } ## end else [ if (defined($SanUtil::g_filer_objs...

    # recommendation from Dave Pressley to set to 7.3
    $version = "7.3";

    # check type of call since we need to set the format we want to
    # see the data come back in (scalar (system-cli) for hash (specific ZAPI)
    if ($type =~ /system-cli/) {
        &SanUtil::nate_log(
            '-comment',
            "$filer >> Executing command ($cmd) via ZAPI (system-cli:priv $priv_level)"
            );
        $add_args->{-cmd_output_format} = "system";
    } else {
        &SanUtil::nate_log('-comment',
                           "$filer >> Executing specific ZAPI command ($type)");
        $add_args->{-cmd_output_format} = "hash";
    }

    # make call to ZAPI via CmdsDriver
    my $resultsHash_ref =
      $cmd_obj->execute_cmd(-cmd               => $type,
                            -cmd_filer         => $filer,
                            -cmd_interface     => "zapi",
                            -cmd_filer_version => $version,
                            -priv              => $priv_level,
                            -args              => \@args,
                            %{$add_args});

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        &SanUtil::nate_log('-comment',
            "SanUtil::cmds_driver_zapi() Unable to execute to ZAPI ($cmd) to $filer"
              . "Error Message = $errmsg");

        # return -1 for status and let the caller check.
        return ("", ZAPI_FAILURE, ());
    }

    # Get results from the CmdsDriver call
    $resultsHash_ref = $cmd_obj->resultsHash();
    my $zapi_result = $resultsHash_ref->{result};
    if (not defined $zapi_result or $zapi_result != 0) {
        my $zapi_reason = $resultsHash_ref->{reason};
        &SanUtil::nate_log('-comment',
                           "SanUtil::cmds_driver_zapi() ZAPI ($cmd) failed.\n"
                             . "Reason = $zapi_reason\n"
                             . "Result Code = $zapi_result");

        # return -1 for status and let the caller check.
        return ("", ZAPI_FAILURE, ());
    } ## end if (not defined $zapi_result...

    # Assuming good ZAPI call, get output. For system-cli this will
    # be a scalar string, for true ZAPI call, this will be as Perl
    # hash (no more need to convert to XML here).
    $output = $resultsHash_ref->{output};

    # if system-cli, get the output and return the output string
    if ($type =~ /system-cli/) {
        if (defined $output and $output ne "") {
            &SanUtil::nate_log('-comment',
                "$filer << Output from command ($cmd) via ZAPI (system-cli:priv $priv_level):\n"
                  . $output);
        } else {
            &SanUtil::nate_log('-comment',
                "$filer << Output from command ($cmd) via ZAPI (system-cli:priv $priv_level):\n"
                  . "No output was generated by this command");

            # always return a string to caller
            $output = "";
        }

        # return output also in list context
        @list_output = split(/\n/, $output);
        return ($output, $zapi_result, @list_output);
    } else {
        &SanUtil::nate_log('-comment',
                           "$filer >> Executed specific ZAPI command ($type)");
        return ($output, $zapi_result);
    }

} ## end sub cmds_driver_zapi

###############################################################################
# Subroutine Name: ping_host
# Objective : Ping the specified ip/device name,
# Returns   : SanUtil::TRUE (can ping) / SanUtil::FALSE (cannot ping)
###############################################################################
sub ping_host {
    my ($device_ip, $device_name) = @_;

    my $ping = Net::Ping->new();

    # if net ping does not work
    if (not $ping->ping($device_ip, 1)) {

        # try system ping (Solaris)
        my @ping_sys    = `/usr/sbin/ping $device_ip 2 2>&1`;
        my $ping_status = $? >> 8;
        if (grep(/is alive/i, @ping_sys) or $ping_status == 0) {
            return $SanUtil::TRUE;
        } elsif (not grep(/no answer from/i, @ping_sys)) {

            # try system ping (Linux)
            @ping_sys    = `/bin/ping $device_ip -c 2 2>&1`;
            $ping_status = $? >> 8;
            if (grep(/\s0\% packet loss/i, @ping_sys) or $ping_status == 0) {
                return $SanUtil::TRUE;
            }
        }

        # try using the device name.
        if ($device_name ne "") {
            return &ping_host($device_name, "");
        } else {

            # give up, host is not pingable
            return $SanUtil::FALSE;
        }
    } ## end if (not $ping->ping($device_ip...

    # host is pingable
    return $SanUtil::TRUE;
} ## end sub ping_host

###############################################################################
# Subroutine Name: nate_process_results
#
# Objective:    This subroutine will take all subtests that the caller created
#               and grab all results from them and create similar results on the
#               callers results object.
#               The point of this subroutine is to bubble up results from children
#               because NATE does not do this automatically.
#
# Returns:      True/False
#
# Input:        None
#
###############################################################################
sub nate_process_results {

    # This next chunk of code allows nate_process_results to self regulate
    # the amount of times it is run. This is necessary for NATE 3 becuase
    # this code will thrash the STAF queue and cause premature failures.
    if (not defined($global_results->{execute_count})) {

        # Determine if the attribute execute_count exists, if not define it
        $global_results->{execute_count} = 0;
    }

    # If the execute_count is not mod 4, then skip this cycle
    if ($global_results->{execute_count} % 4 != 0) {
        $global_results->{execute_count}++;
        return $SanUtil::TRUE;
    } else {
        $global_results->{execute_count}++;
    }

    my ($test_id, @subtests, @failed_subtests);

    &Tharn::param("THPL_BUBBLE_UP_FATAL_ON_ERROR", "-default", 1);

    if ($NATE_version =~ /(\d+)\.(\d+)\.(\d+)/) {
        my ($major, $minor_1, $minor_2) = ($1, $2, $3);
        if ($major >= 3) {

            # If saved result_id is empty, set it to -1, this is necessary because
            # results can have id starting at 0.
            if ($global_results->{since_id} eq "") {
                $global_results->{since_id} = -1;
            }

            # Grab warning results from own_results. By adding a since_id the function only returns results
            # with id greater then the requested.
            my @own_results =
              &Tharn::own_results(result_type => "WARN",
                                  since_id    => $global_results->{since_id});

            if (scalar(@own_results) == 0) {

                # If there are no results to process, return
                return $SanUtil::TRUE;
            } elsif (
                $own_results[-1]->{'result_id'} > $global_results->{since_id}) {

                # If the last result_id is greater then saved result_id, update saved result_id
                $global_results->{since_id} = $own_results[-1]->{'result_id'};
            } else {

                # If the last result_id is less or equal to saved result_id, return
                return $SanUtil::TRUE;
            }

            # Go through each result and process it
            foreach my $result (@own_results) {

                # Grab the misc data associated with this result
                my $misc_data = $result->misc();

                # Update result data retrieved from misc data
                $result->{unique_result_id} = $misc_data->{unique_result_id};
                $result->{child_runid}      = $misc_data->{runid};

                # Check to see if we already have proceed this result in the past
                # If not, process it, if already processed, skip it
                if (not defined($global_results->{$result->{unique_result_id}}))
                {
                    $global_results->{$result->{unique_result_id}} = $result;

                    # Validate that the child runid is not the same as the caller
                    # If it is skip this result, no need to process it because
                    # this result has already been processed
                    if ($result->{child_runid} eq $result->{runid}) {
                        next;
                    }

                    # Format the message to include the child_runid
                    my $message =
                      $result->{child_runid} . " "
                      . join(" ", @{$result->{messages}});
                    @{$result->{messages}} = ("$message");

                    # Log the result. This method will print the result but NOT
                    # actually log a new result.
                    $Log->log(join("\n", $result->format_for_tharn()));
                } ## end if (not defined($global_results...
            } ## end foreach my $result (@own_results)
        } else {

            # Gather a list of subtest launched by caller
            @subtests = Subtest::get_all();

            # If there are no subtests available on the caller, return.
            if (scalar(@subtests) == 0) {
                return $SanUtil::TRUE;
            }

            foreach my $subtest (@subtests) {

                # Check to see if the subtest test_id is less then current value
                # If so, set $test_id to subtest value
                if ($test_id < $subtest->test_id()) {
                    $test_id = $subtest->test_id();
                }
            }

            foreach my $subtest (@subtests) {
                my @results = $subtest->own_results();

                foreach my $result (@results) {

                    # Grab the object reference to determine message type
                    # Split reference in :: and then grab the last part
                    my @parts = split(/::/, ref($result));
                    my $type = lc($parts[$#parts]);

                    # If type is info or pass, ignore this result
                    if ($type =~ /info|pass|na/i) {
                        next;
                    }

                    # Check to see if we already have proceed this result in the past
                    # If not, process it, if already processed, skip it
                    if (not
                        defined($global_results->{$result->{unique_result_id}}))
                    {
                        $global_results->{$result->{unique_result_id}} =
                          $result;

                        if ($type =~ /fail|fatal/i) {

                            # If type is fail or fatal, push subtest that holds this result
                            # on failed subtest list.
                            push(@failed_subtests, $subtest);
                        }

                        # Build out message for result to be created on caller.
                        my $message = join(" ", @{$result->{messages}});

                        # If the runid is not part of the message string, prepend it.
                        if (($message =~ /^\S+\//) ne
                            ($result->{runid} =~ /^\S+\//)) {
                            $message = $result->{runid} . " " . $message;
                        }

                        # Log subtest result on caller
                        &SanUtil::nate_log($type, $message);
                    } ## end if (not defined($global_results...
                } ## end foreach my $result (@results)
            } ## end foreach my $subtest (@subtests)
        } ## end else [ if ($major >= 3)
    } ## end if ($NATE_version =~ /(\d+)\.(\d+)\.(\d+)/)

    # If THPL_BUBBLE_UP_FATAL_ON_ERROR is set to 0 kill off subtests that are still alive.
    # If the failed subtest list is not empty, and subtest is still alive, kill the subtest
    if (&Tharn::param("THPL_BUBBLE_UP_FATAL_ON_ERROR") == 0
        && scalar(@failed_subtests) != 0) {
        &SanUtil::nate_print_results(-results => $global_results);
        foreach my $subtest (@failed_subtests) {
            if ($subtest->isAlive()) {
                $subtest->kill(
                    message => [
                        "comment",
                        "Subtest has been killed by parent, see parent logs for details"
                      ]);
            }
        }

        # If failed subtest list is not empty
        if (scalar(@failed_subtests) != 0 && $test_id == 2) {

            # If this caller is the highest level parent and there was a failure logged
            # set an abortive failure to end the entire test.
            &SanUtil::nate_log(
                              "-fatal",
                              "Fatal error has occurred in one of the subtests",
                              -abortive => 1);
        } elsif (scalar(@failed_subtests) != 0) {

            # If the caller is a child subtest and there was a failure logged
            # Sleep until parent kills this subtest.
            &SanUtil::nate_sleep(1000);
        }
    } elsif (&Tharn::param("THPL_BUBBLE_UP_FATAL_ON_ERROR") == 1
             && scalar(@failed_subtests) != 0) {
        &SanUtil::nate_log("-fatal",
                           "Fatal error has occurred in one of the subtests",
                           -abortive => 1);
    }

    return $SanUtil::TRUE;
} ## end sub nate_process_results

###############################################################################
# Subroutine Name: nate_log
#
# Objective:    This subroutine will take inputs similar to &Tharn::tstlog
#               and create the apporiate result object and log the result
#
# Returns:      True/False
#
# Input:        1) type - Type of message (similar to Tharn::tstlog)
#               2) message - Message to print
#               3) -abortive - (Optional) Set to 1 to make message types of
#                              fail or fatal to be abortive
#               4) -print_result - (Optional) Provide 1 or 0 to print result
#                                  or not print result and just log result
#
###############################################################################
sub nate_log {
    my ($type, $message, %args) = @_;

    &Tharn::param("THPL_BUBBLE_UP_FATAL_ON_ERROR", "-default", 1);

    my $result;
    my $abortive     = 0;
    my $print_result = 1;

    # If THPL_BUBBLE_UP_FATAL_ON_ERROR is set to 1, set the abortive flag to 1
    if (&Tharn::param("THPL_BUBBLE_UP_FATAL_ON_ERROR") == 1) {
        $abortive = 1;
    }

    foreach (keys %args) {
        if (/-?abortive$/) {
            length($abortive = $args{$_})
              or do {
                &SanUtil::nate_log(
                              "-fatal",
                              "SanUtil::nate_log invalid value for 'abortive': "
                                . $args{$_},
                              abortive => 1);
                return $SanUtil::FALSE;
              };
            delete($args{$_});
        } elsif (/-?print_result$/) {
            length($print_result = $args{$_})
              or do {
                &SanUtil::nate_log(
                          "-fatal",
                          "SanUtil::nate_log invalid value for 'print_result': "
                            . $args{$_},
                          abortive => 1);
                return $SanUtil::FALSE;
              };
            delete($args{$_});
        } ## end elsif (/-?print_result$/)
    } ## end foreach (keys %args)

    # Check to see what the message type is and create result object
    if ($type =~ /-?comment|info/i) {

        # We don't need to process this type (comment or info) like the
        # rest of the results. This will remove the creation of a results
        # object, saving process time on NATE. Just logcomment the message.
        # This change is for NATE3, there are some serious performance issues
        # that will overwhelm the NATE db used in NATE3 and above.
        &Tharn::logcomment($message);
        return $SanUtil::TRUE;
    } elsif ($type =~ /-?na/i) {
        $type   = "na";
        $result = na_result($message);
    } elsif ($type =~ /-?pass/i) {
        $type   = "pass";
        $result = pass_result($message);
    } elsif ($type =~ /-?warn/i) {
        $type   = "warn";
        $result = warn_result($message);

        # Tells NATE 3 that this result needs to be bubbled up to parents.
        $result->{bubble} = 1;
    } elsif ($type =~ /-?fail/i) {
        $type = "fail";
        $result = fail_result($message, abortive => $abortive);
    } elsif ($type =~ /-?fatal/i) {
        $type = "fatal";
        $result = fatal_result($message, abortive => $abortive);
    } elsif ($type =~ /-?script/i) {
        $type = "script";
        $result = script_result($message, abortive => "1");
    } elsif ($type =~ /-?config/i) {
        $type = "config";
        $result = config_result($message, abortive => "1");
    } else {
        $type = "script";

        # Type is not valid, set message type of "fatal" with abortive flag set
        $result = script_result($message, abortive => "1");
    }

    # If extra key/value pairs are provided with args hash, set values on results hash
    foreach (keys %args) {
        $result->{$_} = $args{$_};
    }

    # Set a unique result id for each result
    # Result id example: 584892927410.1910.warn
    # This id is derived from the result time stamp,
    # and math with rand() and result type appended
    $result->{unique_result_id} =
      (($result->{timestamp} + time() + rand()) / rand()) . "." . $type;

    # If this is NATE 3 we need to save the unique_result_id under the misc field of the result.
    # This is done because any attribute that is added to the result that is NOT put into the
    # misc field, will be discarded.
    if ($NATE_version =~ /(\d+)\.(\d+)\.(\d+)/) {
        my ($major, $minor_1, $minor_2) = ($1, $2, $3);
        if ($major >= 3) {
            $result->misc({unique_result_id => $result->{unique_result_id},
                           runid            => $result->{runid}});

            # Since this result is going to be printed to the logs,
            # there will be no need to process the result in the future.
            # We do this by defining the result in the global_results hash.
            $global_results->{$result->{unique_result_id}} = $result;
        }
    } ## end if ($NATE_version =~ /(\d+)\.(\d+)\.(\d+)/)

    # Log the result
    if ($print_result == 1) {
        logresult($result);
    } else {
        addresult($result);
    }

    return $SanUtil::TRUE;
} ## end sub nate_log

###############################################################################
# Subroutine Name: nate_print_results
#
# Objective:    This subroutine will take the results hash provided and
#               traverse through its keys and create a message string to
#               print.
#
# Returns:      True/False
#
# Input:        1) -results - (Required) Results hash to use to format message
#               2) -type - (Optional) Message type to use
#
###############################################################################
sub nate_print_results {
    my (%args) = @_;

    my $message = "\n";
    my $type    = "-comment";
    my $results = {};

    foreach (keys %args) {
        if (/-?results$/) {
            length($results = $args{$_})
              or do {
                &SanUtil::nate_log(
                     "-fatal",
                     "SanUtil::nate_print_results invalid value for 'results': "
                       . $args{$_},
                     abortive => 1);
                return $SanUtil::FALSE;
              };
        } elsif (/-?type$/) {
            length($type = $args{$_})
              or do {
                &SanUtil::nate_log(
                        "-fatal",
                        "SanUtil::nate_print_results invalid value for 'type': "
                          . $args{$_},
                        abortive => 1);
                return $SanUtil::FALSE;
              };
        } ## end elsif (/-?type$/)
    } ## end foreach (keys %args)

    # Validate that the results hash has been provided by caller,
    # if not set fatal abortive error
    if (not defined($results)) {
        &SanUtil::nate_log(
                   "-fatal",
                   "SanUtil::nate_print_results No results defined" . $args{$_},
                   abortive => 1);
    }

    # Build message from results hash
    foreach my $key (sort(keys(%{$results}))) {
        my $result = $results->{$key};
        $message .=
          $result->{runid} . ":\n\t" . join(" ", @{$result->{messages}}) . "\n";
    }

    # Log message
    &SanUtil::nate_log($type, $message);

    return $SanUtil::TRUE;
} ## end sub nate_print_results

###############################################################################
# Subroutine Name: nate_sleep
#
# Objective:    This subroutine provides support for NATE 3.0 and NATE 2.0
#               within thpl's and libraries.
#               Will sleep the provided time using the apporiate NATE 3/2 call
#
# Returns:      True/False
#
# Input:        1) sleep time - (Required) Time in seconds to sleep
#
###############################################################################
sub nate_sleep {
    my ($sleep_time, %args) = @_;
    my $process_results = 1;

    # if requested to not process results, dont process them.
    if (exists($args{-process_results})) {
        $process_results = $args{-process_results};
    }

    # Grab NATE version and run necessary sleep function
    if ($NATE_version =~ /(\d+)\.(\d+)\.(\d+)/) {
        my ($major, $minor_1, $minor_2) = ($1, $2, $3);
        if ($major <= 2) {

            # Since we are in NATE 2.x, process results before and after sleep
            if ($process_results == 1 and $sleep_time > 5) {
                &SanUtil::nate_process_results();
            }

            # If NATE version is 2.x.x or less, use standard perl sleep()
            sleep($sleep_time);

            if ($process_results == 1 and $sleep_time > 5) {
                &SanUtil::nate_process_results();
            }
        } elsif ($major >= 3 and $minor_1 >= 2 and $minor_2 >= 16) {
            &Tharn::snooze($sleep_time);
        } else {

            # Version of NATE is less then 3.2.16, use standard perl sleep()
            sleep($sleep_time);
        }
    } else {

        # Unable to determine NATE version, fall back to standard sleep
        sleep($sleep_time);
    }

    return $SanUtil::TRUE;
} ## end sub nate_sleep

###############################################################################
# Subroutine Name: nate_wait_finish
#
# Objective:    This subroutine will take a list of subtests and wait for them
#               to finish. While it waits it will call the process results
#               subroutine.
#
# Returns:      True/False
#
# Input:        1) subtests - (Required) List of subtests to wait for them to
#                             finish.
#
###############################################################################
sub nate_wait_finish {
    my @subtests = @_;

    # Once all subtests are not alive, run wait_finish on each subtest and
    # process results.
    foreach my $subtest (@subtests) {
        if ($subtest->isAlive()) {
            &SanUtil::nate_process_results();
        }
        $subtest->wait_finish();
        &SanUtil::nate_process_results();
    }

    return $SanUtil::TRUE;
} ## end sub nate_wait_finish

###############################################################################
# Subroutine Name: nate_subtest
#
# Objective:    Provide single point to kick off subtests.
#               This function will determine if NATE 3 is being used and add
#               the apporiate flag to support NATE 3 subtesting.
#
#               ** WARNING ** This function will enable bubble up in NATE 3,
#               this function should only be called by the top level parents
#               that require all warnings to be bubbled up to them and no
#               further. Otherwise use standard Tharn::subtest() to create a
#               subtest handle.
#
# Returns:      Subtest handle, see Tharn::subtest() for details
#
# Input:        Standard subtest parameters, see Tharn::subtest() for details
#
###############################################################################
sub nate_subtest {
    my @subtest_args = @_;

    # Grab NATE version and run Tharn::subtest with the associated options
    if ($NATE_version =~ /(\d+)\.(\d+)\.(\d+)/) {
        my ($major, $minor_1, $minor_2) = ($1, $2, $3);
        if ($major >= 3) {
            push(@subtest_args, "-bubble-results");
        }
    }

    my $subtest = &Tharn::subtest(@subtest_args);

    return $subtest;
} ## end sub nate_subtest

###############################################################################
# Subroutine Name: get_avail_controller_licenses
#
# Objective:    Retrieves available controller licenses
# Returns:      Hash containing service name (key) and license (value) pair
# Input:        None
#
###############################################################################
sub get_avail_controller_licenses {

    my ($version) = @_;

    # Define a hash to hold licenses. This is being done to NOT hard code sensitive
    # NetApp license information within automation
    my ($lics);
    my $lic_filev2 =
      "/x/eng/rtpdist/web/eng-web/html/iaq/automation/controller_licensev2.txt";
    my $lic_file =
      "/x/eng/rtpdist/web/eng-web/html/iaq/automation/controller_license.txt";
    if ($version ge "8.2") {
        if (-e "$lic_filev2") {
            unless (open(LIC, "<$lic_filev2")) {
                &SanUtil::nate_log(
                    "-warn",
                    "FilerConfig.thpl: Unable to open controller license file ($lic_file) ($!) ($?)"
                    );
                return {};
            }
            my @output = <LIC>;
            close(LIC);

            foreach my $line (@output) {
                chomp($line);
                $line =~ s/\cM//g;
                my ($service, $type, $license, undef) = split(/:/, $line);
                $lics->{$service}{$type} = $license;
            }
        } else {
            &SanUtil::nate_log("-warn",
                "SanUtil::get_avail_controller_licenses() Unable to locate apporiate license file\n"
                  . "This could happen if you are running pNATE, the license feature is not supported\n"
                  . "under pNATE");
        }
    } else {

        # Check to see if the license file is available, if so, parse it and pull the information in
        if (-e "$lic_file") {
            unless (open(LIC, "<$lic_file")) {
                &SanUtil::nate_log(
                    "-warn",
                    "FilerConfig.thpl: Unable to open controller license file ($lic_file) ($!) ($?)"
                    );
                return {};
            }

            # Pull the content of the file into an array
            my @output = <LIC>;
            close(LIC);

            # Go through each line and parse out license information
            foreach my $line (@output) {

                # Cleanup the line output
                chomp($line);
                $line =~ s/\cM//g;

                # Check to make sure the line is of valid format.
                # remove text in parens since it causes problems with
                # recognition and split later
                $line =~ s/\(.*\)//g;
                if ($line !~ /^\S+\s+\S+/) {
                    next;
                }

                # Split the line on 4 spaces, this is the format that is captured directly from the license wiki page
                # First part is the service, next is the license itself, then the description of the license
                my ($service, $license, undef) = split(/\s+/, $line);

                # Store the license in the lics hash
                $lics->{$service} = $license;
            } ## end foreach my $line (@output)
        } else {
            &SanUtil::nate_log("-warn",
                "SanUtil::get_avail_controller_licenses() Unable to locate apporiate license file\n"
                  . "This could happen if you are running pNATE, the license feature is not supported\n"
                  . "under pNATE");
        }
    } ## end else [ if ($version ge "8.2")
    return $lics;
} ## end sub get_avail_controller_licenses

###############################################################################
# Subroutine Name: get_ontap_flexvol_limits
#
# Objective:    Obtain the flexvol limit given ONTAP revision, and controller
#               model.
#
# Input:        1) -ontap_version - (Required) ONTAP version (ie: 7.3.1)
#               2) -controller_model - (Required) FAS Controller Model
#                                      (ie: FAS3070)
#               3) -type - (Required) 'Minor' or 'Major', determines which
#                          FlexVol look up table to use.
#
###############################################################################
sub get_ontap_flexvol_limits {
    my (%args) = @_;

    my ($ontap_version, $controller_model, $type);

    foreach (keys %args) {
        if (/-?ontap_version$/) {
            length($ontap_version = $args{$_})
              or do {
                &SanUtil::nate_log("-warn",
                    "SanUtil::get_ontap_flexvol_limits() invalid value for 'ontap_version': "
                      . $args{$_});
                return ($SanUtil::FALSE, 0);
              };
        } elsif (/-?controller_model$/) {
            length($controller_model = $args{$_})
              or do {
                &SanUtil::nate_log("-warn",
                    "SanUtil::get_ontap_flexvol_limits() invalid value for 'controller_model': "
                      . $args{$_});
                return ($SanUtil::FALSE, 0);
              };
        } elsif (/-?type$/) {
            length($type = $args{$_})
              or do {
                &SanUtil::nate_log("-warn",
                    "SanUtil::get_ontap_flexvol_limits() invalid value for 'type': "
                      . $args{$_});
                return ($SanUtil::FALSE, 0);
              };
        }
    } ## end foreach (keys %args)

    # check for required parameters
    if (not defined($ontap_version)) {
        &SanUtil::nate_log("-warn",
                           "SanUtil::get_ontap_flexvol_limits() "
                             . "Missing required parameter 'ontap_version': ");
        return ($SanUtil::FALSE, 0);
    }
    if (not defined($controller_model)) {
        &SanUtil::nate_log("-warn",
                           "SanUtil::get_ontap_flexvol_limits() "
                             . "Missing required parameter 'controller_model': "
                             );
        return ($SanUtil::FALSE, 0);
    }
    if (not defined($type) or $type !~ /major|minor/i) {
        &SanUtil::nate_log("-warn",
                           "SanUtil::get_ontap_flexvol_limits() "
                             . "Missing or invalid parameter 'type': ");
        return ($SanUtil::FALSE, 0);
    }

    # lower case the type
    $type = lc($type);

    # Build the ONTAP Destination table for both major and minor NDUs
    my $ontap_table = {};
    my $rel723_724  = qr/7\.2\.3|7\.2\.4/;
    my $rel725_726  = qr/7\.2\.5.*|7\.2\.6.*|gordon/;
    my $rel730_733  = qr/7\.3|7\.3\.1.*|7\.3\.2.*|7\.3\.3.*|ironcity/;

    my $rel70 = qr/7\.0|anchor/;
    my $rel71 = qr/7\.1|tsing/;
    my $rel72 = qr/7\.2|gordon/;
    my $rel73 = qr/7\.3|ironcity/;
    my $rel80 = qr/8\.0|boiler/;
    my $rel81 = qr/8\.1|rolling/;
    my $rel82 = qr/8\.2|sierra|main/;

    # Example (major NDU):
    #     $ontap_table->{"FAS270|N3700|GF270"}->{"major"}->{$rel723_724} = 25;
    # If performing a major ndu to a 7.2.3 or 7.2.4 release on a FAS270,
    # the flexvol limit will be 25.

    # Example (minor NDU):
    #     $ontap_table->{"FAS3170|V3170|N6070"}->{"minor"}->{$rel80} = 300;
    # If performing a minor ndu within the 8.0 release on a FAS3170,
    # the flexvol limit will be 300.

    # not clustered ONTAP capable
    $ontap_table->{"FAS270|N3700|GF270"}->{"major"}->{$rel723_724} = 25;
    $ontap_table->{"FAS270|N3700|GF270"}->{"major"}->{$rel725_726} = 50;
    $ontap_table->{"FAS270|N3700|GF270"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS270|N3700|GF270"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS270|N3700|GF270"}->{"minor"}->{$rel70}      = 25;
    $ontap_table->{"FAS270|N3700|GF270"}->{"minor"}->{$rel71}      = 25;
    $ontap_table->{"FAS270|N3700|GF270"}->{"minor"}->{$rel72}      = 100;
    $ontap_table->{"FAS270|N3700|GF270"}->{"minor"}->{$rel73}      = 150;

    $ontap_table->{"FAS920|V920|GF920"}->{"major"}->{$rel723_724} = 25;
    $ontap_table->{"FAS920|V920|GF920"}->{"major"}->{$rel725_726} = 75;
    $ontap_table->{"FAS920|V920|GF920"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS920|V920|GF920"}->{"minor"}->{$rel70}      = 25;
    $ontap_table->{"FAS920|V920|GF920"}->{"minor"}->{$rel71}      = 25;
    $ontap_table->{"FAS920|V920|GF920"}->{"minor"}->{$rel72}      = 100;
    $ontap_table->{"FAS920|V920|GF920"}->{"minor"}->{$rel73}      = 150;

    $ontap_table->{"FAS940|V940|GF940"}->{"major"}->{$rel723_724} = 25;
    $ontap_table->{"FAS940|V940|GF940"}->{"major"}->{$rel725_726} = 75;
    $ontap_table->{"FAS940|V940|GF940"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS940|V940|GF940"}->{"minor"}->{$rel70}      = 25;
    $ontap_table->{"FAS940|V940|GF940"}->{"minor"}->{$rel71}      = 25;
    $ontap_table->{"FAS940|V940|GF940"}->{"minor"}->{$rel72}      = 100;
    $ontap_table->{"FAS940|V940|GF940"}->{"minor"}->{$rel73}      = 150;

    $ontap_table->{"FAS960|V960|GF960"}->{"major"}->{$rel723_724} = 50;
    $ontap_table->{"FAS960|V960|GF960"}->{"major"}->{$rel725_726} = 150;
    $ontap_table->{"FAS960|V960|GF960"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS960|V960|GF960"}->{"minor"}->{$rel70}      = 50;
    $ontap_table->{"FAS960|V960|GF960"}->{"minor"}->{$rel71}      = 50;
    $ontap_table->{"FAS960|V960|GF960"}->{"minor"}->{$rel72}      = 150;
    $ontap_table->{"FAS960|V960|GF960"}->{"minor"}->{$rel73}      = 150;

    $ontap_table->{"FAS980|V980|GF980"}->{"major"}->{$rel723_724} = 100;
    $ontap_table->{"FAS980|V980|GF980"}->{"major"}->{$rel725_726} = 150;
    $ontap_table->{"FAS980|V980|GF980"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS980|V980|GF980"}->{"minor"}->{$rel70}      = 100;
    $ontap_table->{"FAS980|V980|GF980"}->{"minor"}->{$rel71}      = 100;
    $ontap_table->{"FAS980|V980|GF980"}->{"minor"}->{$rel72}      = 250;
    $ontap_table->{"FAS980|V980|GF980"}->{"minor"}->{$rel73}      = 250;

    $ontap_table->{"FAS2010|V2010"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS2010|V2010"}->{"minor"}->{$rel72}      = 100;
    $ontap_table->{"FAS2010|V2010"}->{"minor"}->{$rel73}      = 150;

    $ontap_table->{"FAS2020|V2020|N3300"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS2020|V2020|N3300"}->{"minor"}->{$rel72}      = 100;
    $ontap_table->{"FAS2020|V2020|N3300"}->{"minor"}->{$rel73}      = 150;

    $ontap_table->{"FAS2050|V2050|N3600"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS2050|V2050|N3600"}->{"minor"}->{$rel72}      = 100;
    $ontap_table->{"FAS2050|V2050|N3600"}->{"minor"}->{$rel73}      = 150;

    $ontap_table->{"FAS3020|V3020|N5200"}->{"major"}->{$rel723_724} = 25;
    $ontap_table->{"FAS3020|V3020|N5200"}->{"major"}->{$rel725_726} = 75;
    $ontap_table->{"FAS3020|V3020|N5200"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS3020|V3020|N5200"}->{"minor"}->{$rel70}      = 25;
    $ontap_table->{"FAS3020|V3020|N5200"}->{"minor"}->{$rel71}      = 25;
    $ontap_table->{"FAS3020|V3020|N5200"}->{"minor"}->{$rel72}      = 100;
    $ontap_table->{"FAS3020|V3020|N5200"}->{"minor"}->{$rel73}      = 150;

    # clustered ONTAP capable

    $ontap_table->{"FAS2040|V2040|N3400"}->{"major"}->{$rel730_733} = 200;
    $ontap_table->{"FAS2040|V2040|N3400"}->{"major"}->{$rel80}      = 200;
    $ontap_table->{"FAS2040|V2040|N3400"}->{"major"}->{$rel81}      = 200;
    $ontap_table->{"FAS2040|V2040|N3400"}->{"major"}->{$rel82}      = 200;

    $ontap_table->{"FAS2040|V2040|N3400"}->{"minor"}->{$rel73}      = 200;
    $ontap_table->{"FAS2040|V2040|N3400"}->{"minor"}->{$rel80}      = 200;
    $ontap_table->{"FAS2040|V2040|N3400"}->{"minor"}->{$rel81}      = 200;
    $ontap_table->{"FAS2040|V2040|N3400"}->{"minor"}->{$rel82}      = 200;

    $ontap_table->{"FAS3040|V3040|N5300"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS3040|V3040|N5300"}->{"major"}->{$rel80}      = 500;
    $ontap_table->{"FAS3040|V3040|N5300"}->{"major"}->{$rel81}      = 500;
    $ontap_table->{"FAS3040|V3040|N5300"}->{"major"}->{$rel82}      = 500;

    $ontap_table->{"FAS3040|V3040|N5300"}->{"minor"}->{$rel72}      = 150;
    $ontap_table->{"FAS3040|V3040|N5300"}->{"minor"}->{$rel73}      = 150;
    $ontap_table->{"FAS3040|V3040|N5300"}->{"minor"}->{$rel80}      = 500;
    $ontap_table->{"FAS3040|V3040|N5300"}->{"minor"}->{$rel81}      = 500;
    $ontap_table->{"FAS3040|V3040|N5300"}->{"minor"}->{$rel82}      = 500;

    $ontap_table->{"FAS3140|V3140|N6040"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS3140|V3140|N6040"}->{"major"}->{$rel80}      = 500;
    $ontap_table->{"FAS3140|V3140|N6040"}->{"major"}->{$rel81}      = 500;
    $ontap_table->{"FAS3140|V3140|N6040"}->{"major"}->{$rel82}      = 500;

    $ontap_table->{"FAS3140|V3140|N6040"}->{"minor"}->{$rel72}      = 150;
    $ontap_table->{"FAS3140|V3140|N6040"}->{"minor"}->{$rel73}      = 150;
    $ontap_table->{"FAS3140|V3140|N6040"}->{"minor"}->{$rel80}      = 500;
    $ontap_table->{"FAS3140|V3140|N6040"}->{"minor"}->{$rel81}      = 500;
    $ontap_table->{"FAS3140|V3140|N6040"}->{"minor"}->{$rel82}      = 500;

    $ontap_table->{"FAS3050|V3050|N5500"}->{"major"}->{$rel723_724} = 50;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"major"}->{$rel725_726} = 150;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"major"}->{$rel730_733} = 150;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"major"}->{$rel80}      = 200;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"major"}->{$rel81}      = 200;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"major"}->{$rel82}      = 200;

    $ontap_table->{"FAS3050|V3050|N5500"}->{"minor"}->{$rel70}      = 50;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"minor"}->{$rel71}      = 50;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"minor"}->{$rel72}      = 150;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"minor"}->{$rel73}      = 150;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"minor"}->{$rel80}      = 200;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"minor"}->{$rel81}      = 200;
    $ontap_table->{"FAS3050|V3050|N5500"}->{"minor"}->{$rel82}      = 200;

    $ontap_table->{"FAS3160|V3160|N6060"}->{"major"}->{$rel730_733} = 200;
    $ontap_table->{"FAS3160|V3160|N6060"}->{"major"}->{$rel80}      = 500;
    $ontap_table->{"FAS3160|V3160|N6060"}->{"major"}->{$rel81}      = 500;
    $ontap_table->{"FAS3160|V3160|N6060"}->{"major"}->{$rel82}      = 500;

    $ontap_table->{"FAS3160|V3160|N6060"}->{"minor"}->{$rel72}      = 200;
    $ontap_table->{"FAS3160|V3160|N6060"}->{"minor"}->{$rel73}      = 300;
    $ontap_table->{"FAS3160|V3160|N6060"}->{"minor"}->{$rel80}      = 500;
    $ontap_table->{"FAS3160|V3160|N6060"}->{"minor"}->{$rel81}      = 500;
    $ontap_table->{"FAS3160|V3160|N6060"}->{"minor"}->{$rel82}      = 500;

    $ontap_table->{"FAS3070|V3070|N5600"}->{"major"}->{$rel730_733} = 300;
    $ontap_table->{"FAS3070|V3070|N5600"}->{"major"}->{$rel80}      = 500;
    $ontap_table->{"FAS3070|V3070|N5600"}->{"major"}->{$rel81}      = 500;
    $ontap_table->{"FAS3070|V3070|N5600"}->{"major"}->{$rel82}      = 500;

    $ontap_table->{"FAS3070|V3070|N5600"}->{"minor"}->{$rel72}      = 250;
    $ontap_table->{"FAS3070|V3070|N5600"}->{"minor"}->{$rel73}      = 300;
    $ontap_table->{"FAS3070|V3070|N5600"}->{"minor"}->{$rel80}      = 500;
    $ontap_table->{"FAS3070|V3070|N5600"}->{"minor"}->{$rel81}      = 500;
    $ontap_table->{"FAS3070|V3070|N5600"}->{"minor"}->{$rel82}      = 500;

    $ontap_table->{"FAS3170|V3170|N6070"}->{"major"}->{$rel730_733} = 300;
    $ontap_table->{"FAS3170|V3170|N6070"}->{"major"}->{$rel80}      = 500;
    $ontap_table->{"FAS3170|V3170|N6070"}->{"major"}->{$rel81}      = 500;
    $ontap_table->{"FAS3170|V3170|N6070"}->{"major"}->{$rel82}      = 500;

    $ontap_table->{"FAS3170|V3170|N6070"}->{"minor"}->{$rel72}      = 250;
    $ontap_table->{"FAS3170|V3170|N6070"}->{"minor"}->{$rel73}      = 300;
    $ontap_table->{"FAS3170|V3170|N6070"}->{"minor"}->{$rel80}      = 500;
    $ontap_table->{"FAS3170|V3170|N6070"}->{"minor"}->{$rel81}      = 500;
    $ontap_table->{"FAS3170|V3170|N6070"}->{"minor"}->{$rel82}      = 500;

    $ontap_table->{"FAS6030|V6030|N7600"}->{"major"}->{$rel730_733} = 300;
    $ontap_table->{"FAS6030|V6030|N7600"}->{"major"}->{$rel80}      = 500;
    $ontap_table->{"FAS6030|V6030|N7600"}->{"major"}->{$rel81}      = 500;
    $ontap_table->{"FAS6030|V6030|N7600"}->{"major"}->{$rel82}      = 500;

    $ontap_table->{"FAS6030|V6030|N7600"}->{"minor"}->{$rel72}      = 250;
    $ontap_table->{"FAS6030|V6030|N7600"}->{"minor"}->{$rel73}      = 300;
    $ontap_table->{"FAS6030|V6030|N7600"}->{"minor"}->{$rel80}      = 500;
    $ontap_table->{"FAS6030|V6030|N7600"}->{"minor"}->{$rel81}      = 500;
    $ontap_table->{"FAS6030|V6030|N7600"}->{"minor"}->{$rel82}      = 500;

    $ontap_table->{"FAS6040|V6040|N7700"}->{"major"}->{$rel730_733} = 300;
    $ontap_table->{"FAS6040|V6040|N7700"}->{"major"}->{$rel80}      = 500;
    $ontap_table->{"FAS6040|V6040|N7700"}->{"major"}->{$rel81}      = 500;
    $ontap_table->{"FAS6040|V6040|N7700"}->{"major"}->{$rel82}      = 500;

    $ontap_table->{"FAS6040|V6040|N7700"}->{"minor"}->{$rel72}      = 250;
    $ontap_table->{"FAS6040|V6040|N7700"}->{"minor"}->{$rel73}      = 300;
    $ontap_table->{"FAS6040|V6040|N7700"}->{"minor"}->{$rel80}      = 500;
    $ontap_table->{"FAS6040|V6040|N7700"}->{"minor"}->{$rel81}      = 500;
    $ontap_table->{"FAS6040|V6040|N7700"}->{"minor"}->{$rel82}      = 500;

    $ontap_table->{"FAS6070|V6070|N7800"}->{"major"}->{$rel730_733} = 300;
    $ontap_table->{"FAS6070|V6070|N7800"}->{"major"}->{$rel80}      = 500;
    $ontap_table->{"FAS6070|V6070|N7800"}->{"major"}->{$rel81}      = 500;
    $ontap_table->{"FAS6070|V6070|N7800"}->{"major"}->{$rel82}      = 500;

    $ontap_table->{"FAS6070|V6070|N7800"}->{"minor"}->{$rel72}      = 250;
    $ontap_table->{"FAS6070|V6070|N7800"}->{"minor"}->{$rel73}      = 300;
    $ontap_table->{"FAS6070|V6070|N7800"}->{"minor"}->{$rel80}      = 500;
    $ontap_table->{"FAS6070|V6070|N7800"}->{"minor"}->{$rel81}      = 500;
    $ontap_table->{"FAS6070|V6070|N7800"}->{"minor"}->{$rel82}      = 500;

    $ontap_table->{"FAS6080|V6080|N7900"}->{"major"}->{$rel730_733} = 300;
    $ontap_table->{"FAS6080|V6080|N7900"}->{"major"}->{$rel80}      = 500;
    $ontap_table->{"FAS6080|V6080|N7900"}->{"major"}->{$rel81}      = 500;
    $ontap_table->{"FAS6080|V6080|N7900"}->{"major"}->{$rel82}      = 500;

    $ontap_table->{"FAS6080|V6080|N7900"}->{"minor"}->{$rel72}      = 250;
    $ontap_table->{"FAS6080|V6080|N7900"}->{"minor"}->{$rel73}      = 300;
    $ontap_table->{"FAS6080|V6080|N7900"}->{"minor"}->{$rel80}      = 500;
    $ontap_table->{"FAS6080|V6080|N7900"}->{"minor"}->{$rel81}      = 500;
    $ontap_table->{"FAS6080|V6080|N7900"}->{"minor"}->{$rel82}      = 500;

    # Carnegie (l/m/h) (minor NDUs within IC & BR, major NDUs to BR)
    $ontap_table->{"FAS3210|V3210|N6210"}->{"major"}->{$rel80} = 500;
    $ontap_table->{"FAS3210|V3210|N6210"}->{"major"}->{$rel81} = 500;
    $ontap_table->{"FAS3210|V3210|N6210"}->{"major"}->{$rel82} = 500;

    $ontap_table->{"FAS3210|V3210|N6210"}->{"minor"}->{$rel73} = 300;
    $ontap_table->{"FAS3210|V3210|N6210"}->{"minor"}->{$rel80} = 500;
    $ontap_table->{"FAS3210|V3210|N6210"}->{"minor"}->{$rel81} = 500;
    $ontap_table->{"FAS3210|V3210|N6210"}->{"minor"}->{$rel82} = 500;

    $ontap_table->{"FAS3240|V3240|N6240"}->{"major"}->{$rel80} = 500;
    $ontap_table->{"FAS3240|V3240|N6240"}->{"major"}->{$rel81} = 500;
    $ontap_table->{"FAS3240|V3240|N6240"}->{"major"}->{$rel82} = 500;

    $ontap_table->{"FAS3240|V3240|N6240"}->{"minor"}->{$rel73} = 300;
    $ontap_table->{"FAS3240|V3240|N6240"}->{"minor"}->{$rel80} = 500;
    $ontap_table->{"FAS3240|V3240|N6240"}->{"minor"}->{$rel81} = 500;
    $ontap_table->{"FAS3240|V3240|N6240"}->{"minor"}->{$rel82} = 500;

    $ontap_table->{"FAS3270|V3270|N6270"}->{"major"}->{$rel80} = 500;
    $ontap_table->{"FAS3270|V3270|N6270"}->{"major"}->{$rel81} = 500;
    $ontap_table->{"FAS3270|V3270|N6270"}->{"major"}->{$rel82} = 500;

    $ontap_table->{"FAS3270|V3270|N6270"}->{"minor"}->{$rel73} = 300;
    $ontap_table->{"FAS3270|V3270|N6270"}->{"minor"}->{$rel80} = 500;
    $ontap_table->{"FAS3270|V3270|N6270"}->{"minor"}->{$rel81} = 500;
    $ontap_table->{"FAS3270|V3270|N6270"}->{"minor"}->{$rel82} = 500;

    # Absolut (v/l/m) (minor NDUs within IC & BR, major NDUs to BR)
    $ontap_table->{"FAS6210|V6210|N7550T"}->{"major"}->{$rel80} = 500;
    $ontap_table->{"FAS6210|V6210|N7550T"}->{"major"}->{$rel81} = 500;
    $ontap_table->{"FAS6210|V6210|N7550T"}->{"major"}->{$rel82} = 500;

    $ontap_table->{"FAS6210|V6210|N7550T"}->{"minor"}->{$rel73} = 300;
    $ontap_table->{"FAS6210|V6210|N7550T"}->{"minor"}->{$rel80} = 500;
    $ontap_table->{"FAS6210|V6210|N7550T"}->{"minor"}->{$rel81} = 500;
    $ontap_table->{"FAS6210|V6210|N7550T"}->{"minor"}->{$rel82} = 500;

    $ontap_table->{"FAS6240|V6240|N7750T"}->{"major"}->{$rel80} = 500;
    $ontap_table->{"FAS6240|V6240|N7750T"}->{"major"}->{$rel81} = 500;
    $ontap_table->{"FAS6240|V6240|N7750T"}->{"major"}->{$rel82} = 500;

    $ontap_table->{"FAS6240|V6240|N7750T"}->{"minor"}->{$rel73} = 300;
    $ontap_table->{"FAS6240|V6240|N7750T"}->{"minor"}->{$rel80} = 500;
    $ontap_table->{"FAS6240|V6240|N7750T"}->{"minor"}->{$rel81} = 500;
    $ontap_table->{"FAS6240|V6240|N7750T"}->{"minor"}->{$rel82} = 500;

    $ontap_table->{"FAS6280|V6280|N7950T"}->{"major"}->{$rel80} = 500;
    $ontap_table->{"FAS6280|V6280|N7950T"}->{"major"}->{$rel81} = 500;
    $ontap_table->{"FAS6280|V6280|N7950T"}->{"major"}->{$rel82} = 500;

    $ontap_table->{"FAS6280|V6280|N7950T"}->{"minor"}->{$rel73} = 300;
    $ontap_table->{"FAS6280|V6280|N7950T"}->{"minor"}->{$rel80} = 500;
    $ontap_table->{"FAS6280|V6280|N7950T"}->{"minor"}->{$rel81} = 500;
    $ontap_table->{"FAS6280|V6280|N7950T"}->{"minor"}->{$rel82} = 500;

    # Delphi (l/m) (minor NDUs within RR and SN)
    $ontap_table->{"FAS3220|V3220|N6220"}->{"minor"}->{$rel81} = 500;
    $ontap_table->{"FAS3220|V3220|N6220"}->{"minor"}->{$rel82} = 500;

    $ontap_table->{"FAS3250|V3250|N6250"}->{"minor"}->{$rel81} = 500;
    $ontap_table->{"FAS3250|V3250|N6250"}->{"minor"}->{$rel82} = 500;

    $ontap_table->{"FAS3220|V3220|N6220"}->{"major"}->{$rel82} = 500;

    $ontap_table->{"FAS3250|V3250|N6250"}->{"major"}->{$rel82} = 500;

    # Buell and Bimota (minor NDUs within SN)
    $ontap_table->{"FAS8020|V8020|N8020"}->{"minor"}->{$rel82} = 1000;
    $ontap_table->{"FAS8040|V8040|N8040"}->{"minor"}->{$rel82} = 1000;
    $ontap_table->{"FAS8060|V8060|N8060"}->{"minor"}->{$rel82} = 1000;
    $ontap_table->{"FAS8080|V8080|N8080"}->{"minor"}->{$rel82} = 1000;
    
    # VSA models
    $ontap_table->{"VSA1000"}->{"minor"}->{$rel80}         = 200;
    $ontap_table->{"VSA1000"}->{"minor"}->{$rel81}         = 200;
    $ontap_table->{"VSA1000"}->{"minor"}->{$rel82}         = 200;

    $ontap_table->{"VSA2000"}->{"minor"}->{$rel80}         = 200;
    $ontap_table->{"VSA2000"}->{"minor"}->{$rel81}         = 200;
    $ontap_table->{"VSA2000"}->{"minor"}->{$rel82}         = 200;

    $ontap_table->{"DataONTAP-vM100"}->{"minor"}->{$rel80} = 200;
    $ontap_table->{"DataONTAP-vM100"}->{"minor"}->{$rel81} = 200;
    $ontap_table->{"DataONTAP-vM100"}->{"minor"}->{$rel82} = 200;

    # hardcode the specific parameters for this controller model
    foreach my $model (keys %{$ontap_table}) {

        # if we matched the model of controller
        if ($controller_model =~ /$model/) {

            # find the ontap release that we are going to
            foreach my $ontap_release (keys %{$ontap_table->{$model}->{$type}})
            {
                if ($ontap_version =~ /$ontap_release/i) {
                    my $flex_vol_limit =
                      $ontap_table->{$model}->{$type}->{$ontap_release};
                    return ($SanUtil::TRUE, $flex_vol_limit);
                }
            }
        } ## end if ($controller_model ...
    } ## end foreach my $model (keys %{$ontap_table...

    return ($SanUtil::FALSE, 0);
} ## end sub get_ontap_flexvol_limits

################################################################################
# Subroutine Name: measure_controller_io_load
# Objective:       This routine will take sysstat usage samples from all
#                  controllers specified in the colon seperate argument string.
#                  It will return the CPU and Disk Uitlization from a 60
#                  second sampling from sysstat
################################################################################
sub measure_cluster_io_load {
    my (%args) = @_;

    my (@clusters, $sample_length);

    foreach (keys %args) {
        if (/^-?clusters$/) {
            length(@clusters = @{$args{$_}})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::measure_cluster_io_load - Invalid value for \"clusters\": $args{$_}"
                    );
                return (0, 0, 0);
              };
        } elsif (/-?sample_length/) {
            length($sample_length = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::measure_cluster_io_load - Invalid value for \"sample_length\": $args{$_}"
                    );
                return (0, 0, 0);
              };
        } ## end elsif (/-?sample_length/)
    } ## end foreach (keys %args)

    if (scalar(@clusters) == 0) {
        &SanUtil::nate_log('-warn',
             "SanUtil::measure_cluster_io_load - No cluster objects provided.");
        return (0, 0, 0);
    }

    # Default $sample_length if not specified
    if ($sample_length eq "" or $sample_length == 0) {
        $sample_length = 60;
    }

    my $io_ops;
    my $total_nodes = 0;

    # Init return parameters for disk and cpu
    my $return_disk = 0;
    my $return_cpu  = 0;

    # nacl or CC
    if (ref($clusters[0]) =~ /NACL/) {
        foreach my $node (@clusters) {
            my $node_name = $node->node();
            my $apiset = $node->apiset(category => "Node",
                                       interface => "CLI",
                                       set => "CMode");
            my $cmd = "run -node $node_name -command \"sysstat -b -c 1 10\"";
            my $output = $apiset->execute_raw_command(command => $cmd);
            
            # Example sysstat output
            #  CPU     FCP   iSCSI Partner  Total     FCP   kB/s   iSCSI   kB/s Partner   kB/s    Disk   kB/s    CP  CP  Disk
            #                                          in    out      in    out      in    out    read  write  time  ty  util
            #  16%       0       0       0      0       0      0       0      0       0      0      94  16729   32%  Zs   23%
            my @header = ("cpu", "fcp", "iscsi", "partner", "total", "fcp_in", "fcp_out", "iscsi_in",
                          "iscsi_out", "partner_in", "partner_out", "disk_read", "disk_write", "cp_time", "cp_ty", "disk_util");
            my @output = split("\n", $output);
            
            # Get rid of any empty lines
            @output = grep(!/^\s*$/, @output);
            
            # Get rid of the command line
            @output = grep(!/priv set/, @output);
            
            # Pull off the header
            @output = grep(!/^\s*CPU/, @output);
            @output = grep(!/\s*in\s+out/, @output);
            
                my $data_line = shift(@output);
            # Strip off white space at the beginning of the data
            $data_line =~ s/^\s*//g;
            my @data = split(/\s+/, $data_line);

            # Build out the stats hash mapping one-to-one header to data
            my $stats;
            for (my $i = 0; $i < @header; $i++) {
                $data[$i] =~ s/\%//g;
                $stats->{$header[$i]} = $data[$i];
            }

            my ($cpu, $disk_util, $cp_time, $cp_ty) = ($stats->{cpu}, $stats->{disk_util}, $stats->{cp_time}, $stats->{cp_ty});
            my $clv_code = ":CLV: filer=$node_name, cpu=$cpu, disk=$disk_util, mem=, cp_time=$cp_time, cp_type=$cp_ty";
            &SanUtil::nate_log('-comment', "$clv_code");            
            $total_nodes++;
            
            $io_ops->{$node_name}->{fc_io_ops}         = $stats->{fcp};
            $io_ops->{$node_name}->{partner_fc_io_ops} = $stats->{partner};
            
            $return_disk += $disk_util;
            $return_cpu  += $cpu;
        }
    } else {
        foreach my $cluster_obj (@clusters) {
            my @node_objs = $cluster_obj->get_node_objs();
            if ($cluster_obj->error()) {
                &SanUtil::nate_log('-warn',
                                   "SanUtil::measure_cluster_io_load - Error when retrieving node objects:\n"
                                   . $cluster_obj->errmsg());
                return (0, 0, 0);
            }
            
            foreach my $node (@node_objs) {
                my $node_name = $node->get_name();
                my $disk_util = 0;
                my $cpu       = 0;
                my ($cp_time, $cp_type) = (0, "");
                my ($fc_io_ops, $partner_fc_io_ops);
                
                my $sysstats = $node->sysstat(-interval => $sample_length);
                if ($node->error()) {
                    &SanUtil::nate_log('-warn',
                                       "SanUtil::measure_cluster_io_load - Error when retrieving node ("
                                       . $node->get_name()
                                       . ") stats:\n"
                                       . $node->errmsg());
                    return (0, 0, 0);
                }
                
                # Get the following values
                #    cpu, fcp, partner, disk
                ($cpu, $fc_io_ops, $partner_fc_io_ops, $disk_util, $cp_time,
                 $cp_type)
                    = ($sysstats->{cpu},     $sysstats->{fcp},
                       $sysstats->{partner}, $sysstats->{disk_util},
                       $sysstats->{cp_time}, $sysstats->{cp_ty});
                
                $io_ops->{$node_name}->{fc_io_ops}         = $fc_io_ops;
                $io_ops->{$node_name}->{partner_fc_io_ops} = $partner_fc_io_ops;
                
                $return_disk += $disk_util;
                $return_cpu  += $cpu;
                
                my $clv_code =
                    ":CLV: filer=$node_name, cpu=$cpu, disk=$disk_util, mem=, cp_time=$cp_time, cp_type=$cp_type";
                &SanUtil::nate_log('-comment', "$clv_code");
                $total_nodes++;
            } ## end foreach my $node (@node_objs)
        } ## end foreach my $cluster_obj (@clusters)        
    }

    # Return the average Disk load and CPU between the controllers
    return (int($return_cpu / $total_nodes),
            int($return_disk / $total_nodes), $io_ops);

} ## end sub measure_cluster_io_load

################################################################################
# Subroutine Name: measure_controller_io_load
# Objective:       This routine will take sysstat usage samples from all
#                  controllers specified in the colon seperate argument string.
#                  It will return the CPU and Disk Uitlization from a 60
#                  second sampling from sysstat
################################################################################
sub measure_controller_io_load {
    my ($filers, $sample_length) = @_;

    # Default $sample_length if not specified
    if ($sample_length eq "" or $sample_length == 0) {
        $sample_length = 60;
    }

    my $io_ops;

    # Init return parameters for disk and cpu
    my $return_disk = 0;
    my $return_cpu  = 0;

    # Split colon seperated list of filers
    my @filers = split(/:/, $filers);

    foreach my $filer (@filers) {
        my $disk_util = 0;
        my $cpu       = 0;
        my ($fc_io_ops, $partner_fc_io_ops);

        # Run sysstat
        my $cmd = "sysstat -b -c 1 $sample_length";
        my ($output, $status) = &SanUtil::system_zapi($filer, $cmd, "diag");
        if ($status == SanUtil::ZAPI_FAILURE) {
            &SanUtil::nate_log('-fail',
                      "host_io.thpl: Failed to execute system-cli ZAPI ($cmd)");
        }

        # Grab output and split, looking for cpu and disk load numbers
        # Example output:
        # CPU   FCP iSCSI Partner  Total    FCP   kB/s   iSCSI   kB/s Partner  kB/s   Disk  kB/s   CP   CP Disk
        #                                    in    out      in    out      in   out   read write  time  ty util
        #  1%    19     0       4     23      0     23       0      0      11     6      3    11    0%  -    1%
        my @output = split(/\n/, $output);
        foreach my $line (@output) {
            $line =~ s/^\s+//;
            if ($line !~ /^\d+\%/) {
                next;
            } else {
                my @parts = split(/\s+/, $line);
                $disk_util = $parts[15];
                $disk_util =~ s/\%//g;
                $cpu = $parts[0];
                $cpu =~ s/\%//g;

                $fc_io_ops         = $parts[1];
                $partner_fc_io_ops = $parts[3];
            } ## end else [ if ($line !~ /^\d+\%/)
        } ## end foreach my $line (@output)

        $io_ops->{$filer}->{fc_io_ops}         = $fc_io_ops;
        $io_ops->{$filer}->{partner_fc_io_ops} = $partner_fc_io_ops;

        $return_disk += $disk_util;
        $return_cpu  += $cpu;

        my $clv_code = ":CLV: filer=$filer, cpu=$cpu, disk=$disk_util, mem=,";
        &SanUtil::nate_log('-comment', "$clv_code");
    } ## end foreach my $filer (@filers)

    # Return the average Disk load and CPU between the controllers
    return (int($return_cpu / scalar(@filers)),
            int($return_disk / scalar(@filers)), $io_ops);

} ## end sub measure_controller_io_load

################################################################################
# Subroutine Name: get_controller_disk_space
# Objective:       This routine will get the disk space on controllers
# Returns:         ($SanUtil::TRUE | $SanUtil::FALSE] , <error|success message>)
################################################################################
sub get_controller_disk_space {
   my (%args) = @_;
   my ($status, @output, $cmd, $filers);
   foreach (keys %args) {
      if (/-?filers/) {
          length($filers = $args{$_})
             or do {
               &SanUtil::nate_log(
                   '-warn',
                   "SanUtil::get_controller_disk_space() - Invalid value for \"filers\": $args{$_}"
                   );
               return $SanUtil::FALSE;
             };
      }
    }
    my $error = 0;
    if (not defined($filers)) {
        &SanUtil::nate_log("-warn",
                           "SanUtil::get_controller_disk_space() "
                           . "Missing required parameter 'filers': ");
        return $SanUtil::FALSE;
    }

    my @filers = split(/:/, $filers);
    foreach my $filer (@filers) {
        &SanUtil::nate_log('-comment',
                           "SanUtil::get_controller_disk_space() - getting disk space on filer ($filer)");
        $cmd = "df -h";
        (@output, $status) = &SanUtil::system_zapi($filer, $cmd, "diag");
        if ($status == SanUtil::ZAPI_FAILURE) {
            $error++;
            &SanUtil::nate_log('-warn',
                               "SanUtil::get_controller_disk_space() - Failed to execute system-cli ZAPI ($cmd)");
            next;
        }
     }

     if ($error > 0) {
         return $SanUtil::FALSE;
     }
     return $SanUtil::TRUE;
}

################################################################################
# Subroutine Name: get_cluster_disk_space
# Objective:       This routine will get the disk space on cluster nodes
# Returns:         ($SanUtil::TRUE | $SanUtil::FALSE] , <error|success message>)
################################################################################
sub get_cluster_disk_space {
    my (%args) = @_;
    my (@clusters, @node_objs);
    foreach (keys %args) {
       if (/^-?clusters$/) {
           scalar(@clusters = @{$args{$_}})
            or do {
              &SanUtil::nate_log(
                  '-warn',
                  "SanUtil::get_cluster_disk_space() - Invalid value for \"clusters\": $args{$_}"
                  );
              return $SanUtil::FALSE;
            };
        }
    }
    my $error = 0;
    if (scalar(@clusters) == 0) {
        &SanUtil::nate_log('-warn',
                           "SanUtil::get_cluster_disk_space() - No cluster objects provided.");
        return $SanUtil::FALSE;
    }

    foreach my $cluster_obj (@clusters) {
       try {
           @node_objs = $cluster_obj->node_objs();
       } otherwise {
           # Looks like we caught an error.  Don't fatal but simply warn and continue
           my $exception_object = shift;

           $error++;
           &SanUtil::nate_log('-warn',
                              "SanUtil::get_cluster_disk_space() - when retrieving node objects:\n"
                              . $cluster_obj->errmsg(). "The exception caught is" . $exception_object->text());
           next;
       };
       foreach my $node (@node_objs) {
          my $node_name = $node->name();
          &SanUtil::nate_log('-comment',
                             "SanUtil::get_cluster_disk_space() - Executing df -h on node ($node_name)");
          try {
              my $apiset = $node->apiset(category => "Node",
                                        interface => "CLI",
                                        set => "CMode");
              my $cmd = "df -h";
              my $output = $apiset->execute_raw_command(command => $cmd);
          } otherwise {
              # Looks like we caught an error.  Don't fatal but simply warn and continue
              my $exception_object = shift;

              $error++;
              &SanUtil::nate_log('-warn',
                                 "SanUtil::get_cluster_disk_space() - Failed to execute df -h"
                                 . "The exception caught is" . $exception_object->text());
          };
       }
    }
    if ($error > 0) {
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;

}

################################################################################
# Subroutine Name: validate_filer_lun_serial_numbers
# Objective:       This routine will validate that both controllers have the
#                  same LUN serial numbers.
# Returns:         ($SanUtil::TRUE | $SanUtil::FALSE] , <error|success message>)
################################################################################
sub validate_filer_lun_serial_numbers {
    my ($filer_objs) = @_;

    my ($version, $status, $filer, $cfmode, @output, $output);
    my (@filers, $filer_obj, @filer_serial_nums, $cmd);

    foreach my $filer (keys %{$filer_objs}) {

        # get filer object for that filer name
        my $filer_obj = $filer_objs->{$filer};

        # get the controller version
        $version = $filer_obj->get_version_major();
        if ($filer_obj->error()) {
            my @error = $filer_obj->errmsg();
            return ($SanUtil::FALSE,
                "SanUtil::validate_filer_lun_serial_numbers Could not the ONTAP major revision\n"
                  . "Error: "
                  . join('\n', @error));
        }

        # get cfmode of the controller
        ($cfmode, $status) = &SanUtil::system_zapi($filer, "fcp show cfmode");
        if ($status == SanUtil::ZAPI_FAILURE) {
            return (
                $SanUtil::FALSE,
                "Filer ($filer) failed to execute system-cli ZAPI (fcp show cfmode)"
                );
        }

        # get the lun config output
        if ($version !~ /7\.0|7\.1/i) {
            eval {
                &SanUtil::nate_sleep(1);
                $cmd = "lun config";
                ($output, $status, @output) =
                  &SanUtil::system_zapi($filer, $cmd);
                if ($status == SanUtil::ZAPI_FAILURE) {
                    return (
                        $SanUtil::FALSE,
                        "Filer ($filer) failed to execute system-cli ZAPI ($cmd)"
                        );
                }
            };
            if ($@) {
                return ($SanUtil::FALSE,
                     "Filer ($filer) failed to execute system-cli ZAPI ($cmd) "
                       . "while attempting to obtain filer lun serial numbers");
            }

            # get the filer lun serial number from the lun config output to verify later
            my $filer_serial = (grep(/^filer\.serialnum\s+=\s+/, @output))[0];
            $filer_serial =~ s/^filer\.serialnum\s+=\s+//;
            $filer_serial =~ s/\"//g;
            push(@filer_serial_nums, $filer_serial);
            push(@filers,            $filer);
        } ## end if ($version !~ /7\.0|7\.1/i)
    } ## end foreach my $filer (keys %{$filer_objs...

    # Make sure all filer_serial numbers are same in SSI mode
    if ($cfmode =~ /single_image/i and scalar(@filers) > 1) {
        if ($filer_serial_nums[0] ne $filer_serial_nums[1]) {
            return ($SanUtil::FALSE,
                "Filer LUN serial number (filer\.serialnum in lun_config output)\n"
                  . "does not match on both heads. In 'single_image cfmode' it must match.\n"
                  . "($filers[0] = $filer_serial_nums[0] and $filers[1] = $filer_serial_nums[1])\n"
                  . "Please update burt319787 with this error\n"
                  . "You can use the command \'lun config filer.serialnum <matching_number>\n"
                  . "to correct this (priv level diag)\n");
        } else {
            return ($SanUtil::TRUE,
                "Filer LUN serial numbers (filer\.serialnum in lun_config output) on both heads match\n"
                  . "($filers[0] = $filer_serial_nums[0] and $filers[1] = $filer_serial_nums[1])"
                  );
        }
    } ## end if ($cfmode =~ /single_image/i...

    # otherwise we are either not in single_image or there are not enough controllers, this is not a failure
    return ($SanUtil::TRUE,
        "Filer LUN serial number validation bypassed. Either not in cfmode 'single_image',"
          . "or in a single head cluster configuration.");
} ## end sub validate_filer_lun_serial_numbers

###############################################################################
# Method:    power_control
# Objective: control the power state of the outlets.
#            NOTE: This will only work if the device is in rpower
# Details:   See POD documentation at the end of this file.
###############################################################################
sub power_control {
    my (%args) = @_;

    my ($status, @output, $cmd);

    my $state      = "";
    my $device     = "";
    my $rpower_cmd = "/usr/software/utils/bin/rpower";

    foreach (keys %args) {
        if (/-?state/) {
            length($state = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::power_control - Invalid value for \"state\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        } elsif (/-?rpower_cmd/) {
            length($rpower_cmd = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::power_control - Invalid value for \"rpower_cmd\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        } elsif (/-?device/) {
            length($device = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::power_control - Invalid value for \"device\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        } ## end elsif (/-?device/)
    } ## end foreach (keys %args)

    # Check to see if the called provided device name
    if ($device eq "" or not defined($device)) {
        &SanUtil::nate_log(
                 '-warn',
                 "SanUtil::power_control - Invalid input value for \"-device\":"
                   . $device);
        return $SanUtil::FALSE;
    }

    # Check to see if caller sent any state
    if ($state eq "" or not defined($state)) {
        &SanUtil::nate_log('-warn',
             "SanUtil::power_control - Invalid value for \"-state\":" . $state);
        return $SanUtil::FALSE;
    }

    # Validate that the caller sent one of three possible states
    if ($state !~ /^(on|off|reboot)$/) {
        &SanUtil::nate_log('-warn',
             "SanUtil::power_control - invalid value for \"-state\":" . $state);
        return $SanUtil::FALSE;
    }

    # Determine if the device has RLM's enabled, look at the tharnhost file for the device and grab RLM_ENABLED
    my $dth_info = &Tharn::host($device);

    if ($dth_info->{'RLM_ENABLED'} == 1) {
        &SanUtil::nate_log('-comment',
                   "SanUtil::power_control - device $device has RLM's enabled");

        # Create a connection to the device, set conn_str to #rlm
        my $conn = Tharn::connect(
            $device
              . "#rlm;conntype=ssh-persist;timeout=180;prompts='RLM.*|SP.*\>'",
            -timeout => 180,
            -eval    => sub {
                if (@_) {
                    $Tharn::InEND = 1;    #Don't abort
                    &SanUtil::nate_log('-warn',
                        "SanUtil::power_control - error when attempting to connect to device ($device) RLM\n"
                          . join("\n", @_));
                    return $SanUtil::FALSE;
                }
            });

        if (not defined($conn) or $conn eq "") {
            return $SanUtil::FALSE;
        }

        # If the action is to reboot, then we need to issue a 'system power cycle'
        # If the action is to power off, then we need to issue a 'system power off'
        # Both of these states require an answer of 'y' when prompted
        if ($state =~ /reboot|off/) {
            if ($state =~ /reboot/) {
                $cmd = "system power cycle";
            } else {
                $cmd = "system power off";
            }
            my (undef, undef, undef, $received_prompt) =
              &Tharn::say(
                        $conn,       "$cmd",
                        '-noprompt', '-prompt',
                        'This will cause a dirty shutdown of your appliance',
                        '-timeout', 120,
                        '-ignore',  "echo ignore",
                        '-comment', "Attempted $cmd");

            # Verify that we received the correct prompt, if not we had a failure
            if ($received_prompt =~
                /This will cause a dirty shutdown of your appliance/) {

                # Issue a 'y' to the prompt and wait for expected prompt 'RLM.*\>'
                (undef, undef, undef, $received_prompt) =
                  &Tharn::say($conn,          "y",
                              '-noprompt',    '-prompt',
                              'RLM.*|SP.*\>', '-timeout',
                              120,            '-ignore',
                              "echo ignore",  '-comment',
                              "Attempted $cmd");

                # Verify that we received the correct prompt back
                if ($received_prompt =~ /RLM.*|SP.*\>/) {
                    my $msg;
                    if ($state =~ /reboot/) {
                        $msg = "power cycled";
                    } else {
                        $msg = "powered off";
                    }
                    &SanUtil::nate_log(
                        '-comment',
                        "SanUtil::power_control - device $device has been $msg"
                        );
                    return $SanUtil::TRUE;
                }
            } ## end if ($received_prompt =~...

            &SanUtil::nate_log('-warn',
                           "SanUtil::power_control - failed to execute '$cmd'");
            return $SanUtil::FALSE;
        } else {
            $cmd = "system power on";

            # State is on, just issue a 'system power on', no extra prompts are given
            my (undef, undef, undef, $received_prompt) =
              &Tharn::say($conn,          "$cmd",
                          '-noprompt',    '-prompt',
                          'RLM.*|SP.*\>', '-timeout',
                          120,            '-ignore',
                          "echo ignore",  '-comment',
                          "Attempted $cmd");
            if ($received_prompt =~ /RLM.*|SP.*\>/) {
                &SanUtil::nate_log(
                    '-comment',
                    "SanUtil::power_control - device $device has been powered on"
                    );
                return $SanUtil::TRUE;
            }

            &SanUtil::nate_log('-warn',
                           "SanUtil::power_control - failed to execute '$cmd'");
            return $SanUtil::FALSE;
        } ## end else [ if ($state =~ /reboot|off/)

        return $SanUtil::TRUE;
    } else {

        # if the this machine has a powercmd, use that instead
        if (defined($dth_info->{powercmd})) {
            $rpower_cmd = $dth_info->{powercmd};
        } else {

            # default to the regular rpower cmd
            $rpower_cmd = "${rpower_cmd} ${device}";
        }

        $cmd = "${rpower_cmd} -v -n -f";
        @output = split(/\n/, `$cmd`);

        # Check to see if we have any entry for this device
        if (scalar(@output) == 0) {
            &SanUtil::nate_log('-warn',
                "SanUtil::power_control - Could not find rpower entry for device: "
                  . $device . "\n"
                  . join("\n", @output));
            return $SanUtil::FALSE;
        }

        # Check to see if we only have ONE entry for this device
        if (grep(/Type:/, @output) > 1) {
            &SanUtil::nate_log('-warn',
                "SanUtil::power_control - Found more then one entry in rpower for device: "
                  . $device . "\n"
                  . join("\n", @output));
            return $SanUtil::FALSE;
        }

        # Check to see if we have any power ports associated with the rpower record
        if (grep(/Power Outlet|ps\d+:|\s+Power:\s+/, @output) < 1) {
            &SanUtil::nate_log('-warn',
                "SanUtil::power_control - Did not find any power outlets in rpower for device: "
                  . $device . "\n"
                  . join("\n", @output));
            return $SanUtil::FALSE;
        }

        # Before changing state, capture old state to validate state change worked
        $cmd = "${rpower_cmd} status";

        &SanUtil::nate_log('-comment',
               "SanUtil::power_control - issuing the following command ($cmd)");

        ($status, @output) =
          &SanUtil::nate_server_execute("-cmd"         => $cmd,
                                        "-timeout"     => 30,
                                        "-retry"       => 3,
                                        "-retry_delay" => 5);
        if ($status != 0 or grep(/FATAL/, @output)) {
            &SanUtil::nate_log('-warn',
                "SanUtil::power_control - failed to collect initial state for device "
                  . $device . "\n"
                  . join("\n", @output));
            return $SanUtil::FALSE;
        }

        # Expected output too look like this:
        # > rpower hp2620-rtp11 status
        # found entry hp2620-rtp11(=>hp2620-rtp11.rtp.netapp.com) - status requested for 2 power outlets
        # sending 'status' request to sentry CDU at pd055b1l2d9.rtp.netapp.com(5/5).
        # Outlet: 5(.aa5) State: Off - Off
        # sending 'status' request to sentry CDU at pd055b1l2d9.rtp.netapp.com(29/29).
        # Outlet: 29(.ba5) State: Off - Off
        my $not_matched = 0;

        # Go through each line in the output and pull the states for each plug
        foreach my $line (@output) {
            chomp($line);
            $line =~ s/\cM//g;
            if ($line =~ /Outlet:\s+(\S+)\s+State:\s+(\S+)\s+-/) {
                my ($plug, $p_state) = ($1, lc($2));

                # Check to see if the plug state does not match the requested state
                # this will tell the rest of the code to either continue to change the state
                # or return true.
                if ($state ne $p_state) {
                    $not_matched = 1;
                }
            } ## end if ($line =~ /Outlet:\s+(\S+)\s+State:\s+(\S+)\s+-/)
        } ## end foreach my $line (@output)

        # If found a plug state not matched requested state, continue to change state
        # Otherwise return true because all plugs match requested state
        if ($not_matched == 0) {
            &SanUtil::nate_log('-comment',
                "SanUtil::power_control - plug state alredy match requested state\n"
                  . join("\n", @output));
            return $SanUtil::TRUE;
        }

        $cmd = "$rpower_cmd $state";
        &SanUtil::nate_log('-comment',
               "SanUtil::power_control - issuing the following command ($cmd)");

        ($status, @output) =
          &SanUtil::nate_server_execute("-cmd"         => $cmd,
                                        "-timeout"     => 60,
                                        "-retry"       => 3,
                                        "-retry_delay" => 5);
        if ($status != 0 or grep(/FATAL/, @output)) {
            &SanUtil::nate_log('-warn',
                "SanUtil::power_control - failed to change the power state for device "
                  . $device
                  . " to ($state)\n"
                  . join("\n", @output));
            return $SanUtil::FALSE;
        }

        # Before changing state, capture old state to validate state change worked
        $cmd = "$rpower_cmd status";

        &SanUtil::nate_log('-comment',
               "SanUtil::power_control - issuing the following command ($cmd)");

        ($status, @output) =
          &SanUtil::nate_server_execute("-cmd"         => $cmd,
                                        "-timeout"     => 30,
                                        "-retry"       => 3,
                                        "-retry_delay" => 5);
        if ($status != 0 or grep(/FATAL/, @output)) {
            &SanUtil::nate_log('-warn',
                "SanUtil::power_control - failed to collect initial state for device "
                  . $device
                  . " to ($state)\n"
                  . join("\n", @output));
            return $SanUtil::FALSE;
        }

        # Grab state again and compare the previous state and current state
        $not_matched = 0;

        # Go through each line in the output and pull the states for each plug
        foreach my $line (@output) {
            chomp($line);
            $line =~ s/\cM//g;
            if ($line =~ /Outlet:\s+(\S+)\s+State:\s+(\S+)\s+-/) {
                my ($plug, $p_state) = ($1, lc($2));
                if ($p_state ne $state) {
                    $not_matched = 1;
                }
            }
        } ## end foreach my $line (@output)

        # Found that all plugs have changed state
        if ($not_matched == 0 or $state eq "reboot") {
            &SanUtil::nate_log('-comment',
                               "Power state for device "
                                 . $device
                                 . " has been changed to ($state)");
        } else {
            &SanUtil::nate_log('-comment',
                    "Power state for device "
                  . $device
                  . " has not been changed, failed to change plug states to ($state)\n"
                  . join("\n", @output));
            return $SanUtil::FALSE;
        }
        return $SanUtil::TRUE;
    } ## end else [ if ($dth_info->{'RLM_ENABLED'...
} ## end sub power_control

###############################################################################
# Method:    parse_client_vms
# Objective: Provides an a common function to parse CLIENT_VMS parameter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub parse_client_vms {
    my (%args) = @_;

    my ($client_vms, $mp_type, $lvm_type);

    foreach (keys %args) {
        if (/-?client_vms/) {
            length($client_vms = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::parse_client_vms - Invalid value for \"client_vms\": $args{$_}"
                    );
                return (wantarray()) ? () : [];
              };
        } ## end if (/-?client_vms/)
    } ## end foreach (keys %args)

    if (not defined($client_vms) or $client_vms =~ /^\s*$/) {
        &SanUtil::nate_log('-warn',
                         "SanUtil::parse_client_vms: CLIENT_VMS list is empty");
        return (wantarray()) ? () : [];
    }

    ($mp_type, $lvm_type) = $client_vms =~ /^(\S+)\-(\S+)$/;

    return ($mp_type, $lvm_type);
} ## end sub parse_client_vms

###############################################################################
# Method:    collect_target_list
# Objective: Provided controllers, this function will collect iSCSI and FCP
#            targets available for use and returns back a data structure.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub collect_target_list {
    my (%args) = @_;

    my ($filers,      $debug_output);
    my (@fcp_targets, @iscsi_targets);
    my (@FcpTargets,  @iSCSITargets, @FilerObjs, @CFOModes) = ();
    my $data;

    foreach (keys %args) {
        if (/-?filers/) {
            length($filers = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::collect_target_list - Invalid value for \"filers\": $args{$_}"
                    );
                return {};
              };
        } elsif (/-?debug_output/) {
            length($debug_output = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::collect_target_list - Invalid value for \"debug_output\": $args{$_}"
                    );
                return {};
              };
        } ## end elsif (/-?debug_output/)
    } ## end foreach (keys %args)

    # If debug_ouput is not defined, default the value to 0
    if (not defined($debug_output)) {
        $debug_output = 0;
    }

    # get controller target information via SanFiler
    foreach my $filer_name (split(/\:/, $filers)) {
        my $filer = SanFiler->new(-filer           => $filer_name,
                                  -debug_level     => ($debug_output + 1),
                                  -interface       => "zapi",
                                  -conn_type       => "#rsh",
                                  -enable_sancache => 1);
        if ($filer->error()) {
            my @error = $filer->errmsg();
            &SanUtil::nate_log('-warn',
                "SanUtil::collect_target_list Could not create filer object ($filer_name)\n"
                  . "Error: "
                  . join('\n', @error));
            return {};
        }

        # get the controller target objects
        @iscsi_targets =
          $filer->get_iscsi_targets(
              -state                      => "enabled",
              -valid_ip                   => 1,
              -filer_accesslist_inclusion => $args{-filer_accesslist_inclusion},
              -filer_accesslist_exclusion => $args{-filer_accesslist_exclusion}
              );
        if ($filer->error()) {
            my @error = $filer->errmsg();
            &SanUtil::nate_log('-warn',
                "SanUtil::collect_target_list Could not get iSCSI target objects from controller ($filer_name)\n"
                  . "Error: "
                  . join('\n', @error));
            return {};
        }

        @fcp_targets = $filer->get_targets();
        if ($filer->error()) {
            my @error = $filer->errmsg();
            &SanUtil::nate_log('-fatal',
                "SanUtil::collect_target_list Could not get FCP target objects from controller ($filer_name)\n"
                  . "Error: "
                  . join('\n', @error));
            return {};
        }

        # get cfmode
        my $Cfmode = $filer->get_cfmode();
        if ($filer->error()) {
            my @error = $filer->errmsg();
            &SanUtil::nate_log(
                     '-fatal',
                     "ClientConfig.thpl: Could not get cfmode from controller\n"
                       . "Error: "
                       . join('\n', @error));
        }

        # push CFO modes into array for each controller
        push(@CFOModes, $Cfmode);

        # push targets into array for each controller
        push(@iSCSITargets, @iscsi_targets);
        push(@FcpTargets,   @fcp_targets);

        # push targets into array for each controller
        push(@FilerObjs, $filer);
    } ## end foreach my $filer_name (split...

    my (@target_ips) = map { $_->{filer_name} . ":" . $_->{ip}; } @iSCSITargets;
    &SanUtil::nate_log('-comment',
                       "SanUtil::collect_target_list Using iSCSI TARGETS |"
                         . join(", ", @target_ips) . "|\n");

    $data->{iscsi_targets} = \@iSCSITargets;
    $data->{fcp_targets}   = \@FcpTargets;
    $data->{filer_objs}    = \@FilerObjs;
    $data->{cfo_modes}     = \@CFOModes;

    return $data;
} ## end sub collect_target_list

###############################################################################
# Method:    get_ip_subnet()
# Objective: Given an ip address and netmask/prefixlen, determine the subnet
###############################################################################
sub get_ip_subnet {
    my (%args) = @_;

    my ($ip_address, $ip_version, $ipv4_netmask, $ipv6_prefix_len);

    # parse args
    foreach (keys %args) {
        if (/-?ip_address/) {
            length($ip_address = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::get_ip_subnet - Invalid value for \"ip_address\": $args{$_}"
                    );
                return "";
              };
        } elsif (/-?ip_version/) {
            length($ip_version = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::get_ip_subnet - Invalid value for \"ip_version\": $args{$_}"
                    );
                return "";
              };
        } elsif (/-?ipv4_netmask/) {
            length($ipv4_netmask = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::get_ip_subnet - Invalid value for \"ipv4_netmask\": $args{$_}"
                    );
                return "";
              };
        } elsif (/-?ipv6_prefix_len/) {
            length($ipv6_prefix_len = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::get_ip_subnet - Invalid value for \"ipv6_prefix_len\": $args{$_}"
                    );
                return "";
              };
        } ## end elsif (/-?ipv6_prefix_len/)
    } ## end foreach (keys %args)

    # check for required options
    if (not defined($ip_address)) {
        &SanUtil::nate_log(
              '-warn',
              "SanUtil::get_ip_address - Missing required parameter -ip_address"
              );
        return "";
    }
    if (not defined($ip_version)) {
        &SanUtil::nate_log(
              '-warn',
              "SanUtil::get_ip_address - Missing required parameter -ip_version"
              );
        return "";
    }

    # handle IPv4 Addresses
    if ($ip_version =~ /ipv4|4/i) {
        if (not defined($ipv4_netmask)) {
            &SanUtil::nate_log(
                '-warn',
                "SanUtil::get_ip_address - Missing required parameter -ipv4_netmask"
                );
            return "";
        }

        # if the ipv4_netmask is not in IP form, then its probably in hex, convert the hex to ipv4 format
        if ($ipv4_netmask !~ /\d+\.\d+\.\d+\.\d+/) {
            $ipv4_netmask = &hex2ipv4($ipv4_netmask);
        }

        # convert the ip address and netmask to decimal form and do a bitwise and to generate the subnet address.
        my $ipv4_obj       = new Net::IP($ip_address);
        my $netmask_obj    = new Net::IP($ipv4_netmask);
        my $subnet_decimal = $ipv4_obj->intip() & $netmask_obj->intip();
        my $subnet_ip = &Net::IP::ip_bintoip(&dec2bin32($subnet_decimal), 4);

        # get rid of trailing zeros on the ip address for the subnet since we only
        # want the network portion of the ip
        $subnet_ip =~ s/\.0.*//;

        return $subnet_ip;

    } elsif ($ip_version =~ /ipv6|6/i) {

        # handle IPv6 Addresses
        if (not defined($ipv6_prefix_len)) {
            &SanUtil::nate_log(
                '-warn',
                "SanUtil::get_ip_address - Missing required parameter -ipv6_prefix_len"
                );
            return "";
        }

        my $ipv6_obj = new Net::IP($ip_address);

        # convert the ipv6 address to binary form and perform a
        # bitwise and with the binary mask of the prefix length
        my $subnet_bin =
          $ipv6_obj->binip() & &Net::IP::ip_get_mask($ipv6_prefix_len, 6);

        # convert the binary to hex since ipv6 is in hex format
        my $subnet_ip = &bin2hex($subnet_bin);

        # add a colon after every four hex characters
        $subnet_ip =~ s/(....)/$1:/g;
        chop($subnet_ip);

        # convert the subnet ip to the short consolidated version
        my $subnet_obj = new Net::IP($subnet_ip);
        return $subnet_obj->short();
    } else {
        &SanUtil::nate_log('-warn',
                  "SanUtil::get_ip_address - Invalid ip_version '$ip_version'");
        return "";
    }
} ## end sub get_ip_subnet

###############################################################################
# Method:    get_ip_broadcast()
# Objective: Given an ip address and netmask/prefixlen, determine the broadcast
###############################################################################
sub get_ip_broadcast {
    my (%args) = @_;

    my ($ip_address, $ipv4_netmask);

    # parse args
    foreach (keys %args) {
        if (/-?ip_address/) {
            length($ip_address = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::get_ip_broadcast - Invalid value for \"ip_address\": $args{$_}"
                    );
                return "";
              };
        } elsif (/-?ipv4_netmask/) {
            length($ipv4_netmask = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::get_ip_broadcast - Invalid value for \"ipv4_netmask\": $args{$_}"
                    );
                return "";
              }
        } ## end elsif (/-?ipv4_netmask/)
    } ## end foreach (keys %args)

    # check for required options
    if (not defined($ip_address)) {
        &SanUtil::nate_log(
            '-warn',
            "SanUtil::get_ip_broadcast - Missing required parameter -ip_address"
            );
        return "";
    }
    if (not defined($ipv4_netmask)) {
        &SanUtil::nate_log(
            '-warn',
            "SanUtil::get_ip_broadcast - Missing required parameter -ipv4_netmask"
            );
        return "";
    }

    # if the ipv4_netmask is not in IP form, then its probably in hex, convert the hex to ipv4 format
    if ($ipv4_netmask !~ /\d+\.\d+\.\d+\.\d+/) {
        $ipv4_netmask = &hex2ipv4($ipv4_netmask);
    }

    # convert the ip address and netmask to decimal form and do a bitwise and to generate the subnet address.
    my $ipv4_subnet =
      &SanUtil::get_ip_subnet(-ip_address   => $ip_address,
                              -ip_version   => "ipv4",
                              -ipv4_netmask => $ipv4_netmask);

    my $ipv4_subnet_obj = new Net::IP($ipv4_subnet);
    my $netmask_obj     = new Net::IP($ipv4_netmask);

    # calcualte the broadcast by ORing the subnet with the complement of the netmask
    my $broadcast_decimal =
      $ipv4_subnet_obj->intip() | (~$netmask_obj->intip());

    # convert the broadcast to binary and then to IP
    my $broadcast_ip = &Net::IP::ip_bintoip(&dec2bin32($broadcast_decimal), 4);

    return $broadcast_ip;
} ## end sub get_ip_broadcast

###############################################################################
# Method:    bin2hex
# Objective: Convert binary to hex
###############################################################################
sub bin2hex {
    my ($bin) = @_;

    my %conv = map { sprintf("%04b", $_), sprintf "%x", $_ } 0 .. 15;
    $bin =~ s/([01]{4})/$conv{ $1 }/g;

    return $bin;
}

###############################################################################
# Method:    hex2ipv4
# Objective: Convert hex to ipv4 format
###############################################################################
sub hex2ipv4 {
    my ($hex) = @_;

    $hex =~ s/^0x//;
    my @array = split("", $hex);
    my $val;
    for (my $i = 0 ; $i < scalar(@array) ; $i += 2) {
        my $byte = $array[$i] . $array[$i + 1];
        $val .= hex($byte) . ".";
    }
    chop($val);

    return $val;
} ## end sub hex2ipv4

###############################################################################
# Method:    dec2bin32
# Objective: Convert decimal to 32 bit binary
###############################################################################
sub dec2bin32 {
    my ($dec) = @_;

    my $str = unpack("B32", pack("N", $dec));
    $str =~ s/^0+(?=\d)//;    # otherwise you'll get leading zeros
    return $str;
}

###############################################################################
# Method:    bin2dec32
# Objective: 32 bit binary to decimal
###############################################################################
sub bin2dec32 {
    my ($bin) = @_;

    return unpack("N", pack("B32", substr("0" x 32 . $bin, -32)));
}

###############################################################################
# Method:    collect_controller_rastrace_dump
# Objective: collects rastrace dump on controllers (7mode)
###############################################################################
sub collect_controller_rastrace_dump {
    my (%args) = @_;

    my ($status, @output, $output, $cmd, $filers);
    my $modules_hash;

    foreach (keys %args) {
        if (/-?filers/) {
            length($filers = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::collect_controller_rastrace_dump() - Invalid value for \"filers\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        }
    }

    my $error = 0;

    # Split colon seperated list of filers
    my @filers = split(/:/, $filers);

    foreach my $filer (@filers) {

        &SanUtil::nate_log('-comment',
                       "SanUtil::collect_controller_rastrace_dump() - Collecting rastrace dump on filer ($filer)");

        # Get list of rastrace modules
        # module id 0 is to get all  modules
        $cmd = "rastrace list -m 0";
        ($output, $status, @output) = &SanUtil::system_zapi($filer, $cmd, "diag");
        if ($status == SanUtil::ZAPI_FAILURE) {
            $error++;
            &SanUtil::nate_log('-warn',
                      "SanUtil::collect_controller_rastrace_dump() - Failed to execute system-cli ZAPI ($cmd)");
            next;
        }

        # grep only active modules
        my @grep_output = grep(/ACTIVE/, @output);

        foreach my $line (@grep_output) {
            if ($line =~ /id\s*\=\s*(\d+)\,\s*inst\s*\=\s*(\d+)/) {
                my $module_id = $1;
                my $instance_no = $2;

                # Get the dump
                $cmd = "rastrace dump -m $module_id -i $instance_no";
                ($output, $status, @output) = &SanUtil::system_zapi($filer, $cmd, "diag");
                if ($status == SanUtil::ZAPI_FAILURE) {
                    $error++;
                    &SanUtil::nate_log('-warn',
                              "SanUtil::collect_controller_rastrace_dump() - Failed to execute system-cli ZAPI ($cmd)");
                    next;
                }
            } else {
                &SanUtil::nate_log('-warn',
                        "SanUtil::collect_controller_rastrace_dump() - Error when parsing rastrace list output");
                return $SanUtil::FALSE;
            }
        }
    }

    if ($error > 0) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    collect_cluster_rastrace_dump
# Objective: Collects rastrace on controller clusters (cmode)
###############################################################################
sub collect_cluster_rastrace_dump {
    my (%args) = @_;

    my (@clusters, @node_objs);

    foreach (keys %args) {
        if (/^-?clusters$/) {
            length(@clusters = @{$args{$_}})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::collect_cluster_rastrace_dump() - Invalid value for \"clusters\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        }
    }

    if (scalar(@clusters) == 0) {
        &SanUtil::nate_log('-warn',
             "SanUtil::collect_cluster_rastrace_dump() - No cluster objects provided.");
        return $SanUtil::FALSE;
    }

    my $error = 0;
    foreach my $cluster_obj (@clusters) {
        try {
            @node_objs = $cluster_obj->node_objs();
        } otherwise {
            $error++;
            &SanUtil::nate_log('-warn',
                "SanUtil::collect_cluster_rastrace_dump() - when retrieving node objects:\n"
                  . $cluster_obj->errmsg());
            next;
        };

        foreach my $node (@node_objs) {
            my $node_name = $node->name();

            &SanUtil::nate_log('-comment',
                "SanUtil::collect_cluster_rastrace_dump() - Collecting rastrace dump on node ($node_name)");

            try {
                my $apiset = $node->apiset(category => "Node",
                                           interface => "CLI",
                                           set => "CMode");
                my $cmd = "run -node $node_name -command \"priv set diag; rastrace list -m 0\"";
                my $output = $apiset->execute_raw_command(command => $cmd);

                # Example "rastrace list -m 0" output
                #     SCSI, id =  3, inst = 0,     ACTIVE
                #     VDISK, id =  5, inst = 0,     ACTIVE
                #     VDISK, id =  5, inst = 1,     ACTIVE
                #     VDISK, id =  5, inst = 2,     ACTIVE
                #     VDISK, id =  5, inst = 3,     ACTIVE

                my @output = split("\n", $output);

                # grep only active modules
                my @grep_output = grep(/ACTIVE/, @output);

                foreach my $line (@grep_output) {
                    if ($line =~ /id\s*\=\s*(\d+)\,\s*inst\s*\=\s*(\d+)/) {
                        my $module_id = $1;
                        my $instance_no = $2;

                        try {
                            $cmd = "run -node $node_name -command \"priv set diag; rastrace dump -m $module_id -i $instance_no\"";
                            $output = $apiset->execute_raw_command(command => $cmd);
                        } otherwise {
                            # rastrace dump has failed, issue a warning
                            $error++;
                            &SanUtil::nate_log('-warn',
                                "SanUtil::collect_controller_rastrace_dump() - Failed to dump rastarce " .
                                "for module ($module_id) and instance ($instance_no) on $node_name");
                        };

                    } else {
                        $error++;
                        &SanUtil::nate_log('-warn',
                            "SanUtil::collect_controller_rastrace_dump() - Error when parsing rastrace list output");
                        return $SanUtil::FALSE;
                    }
                }
            } otherwise {
                # rastrace dump has failed, issue a warning
                $error++;
                &SanUtil::nate_log('-warn',
                    "SanUtil::collect_controller_rastrace_dump() - Error when getting list of rastrace modules on $node_name");
            };
        }
    }

    if ($error > 0) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    nate_server_execute
# Objective: Provides an execute function for NATE control server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nate_server_execute {
    my (%args) = @_;

    my ($output, @output, $status, $retry,    $retry_delay, $cmd);
    my ($match,  $before, $after,  $sentinel, $timeout,     $show_output);

    $timeout     = "300";
    $retry       = 0;
    $retry_delay = 0;
    @output      = ();
    $show_output = 1;

    foreach (keys %args) {
        if (/-?timeout/) {
            length($timeout = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::nate_server_execute - Invalid value for \"timeout\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        } elsif (/-?cmd/) {
            length($cmd = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::nate_server_execute - Invalid value for \"cmd\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        } elsif (/-?retry$/) {
            length($retry = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::nate_server_execute - Invalid value for \"retry\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        } elsif (/-?retry_delay$/) {
            length($retry_delay = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::nate_server_execute - Invalid value for \"retry_delay\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        } elsif (/-?show_output$/) {
            length($show_output = $args{$_})
              or do {
                &SanUtil::nate_log(
                    '-warn',
                    "SanUtil::nate_server_execute - Invalid value for \"show_output\": $args{$_}"
                    );
                return $SanUtil::FALSE;
              };
        } ## end elsif (/-?show_output$/)
    } ## end foreach (keys %args)

    if (!defined($cmd) or $cmd =~ /^\s*$/) {
        &SanUtil::nate_log('-warn',
                    "SanUtil::nate_server_execute: Command passed in is empty");
        $status = 1;
        return ($status, @output);
    }

    # add to the command the echo of status to get that information from
    # the unix shell. Also add the sentinel string DoNe to command so that
    # we can match on that string.
    $sentinel = 'DoNe';
    $cmd = "(" . $cmd . " 2>&1;echo \"rstatus=\$\?\" ;echo $sentinel) 2>&1";

    if ($show_output) {
        &SanUtil::nate_log('-comment',
                           "NATE_CONTROL_SERVER >> Executing command ($cmd)");
    }

    my $OLD_SIG_ALRM = $SIG{ALRM};
    $SIG{ALRM} = sub { die "TIMEOUT\n" };
    eval {
        alarm($timeout);
        $output = `$cmd`;
        alarm(0);
    };
    $SIG{ALRM} = $OLD_SIG_ALRM;

    # Failed to execute command
    if ($@ =~ /TIMEOUT/) {
        &SanUtil::nate_log('-warn',
             "SanUtil::nate_server_execute: Unable to execute command: ($cmd)");
        $status = 1;
    } else {

        # strip output and remove CR/LF
        chomp($output);
        $output =~ s/\r//g;

        # remove first line
        $cmd    =~ s/\;echo "rstatus=\$\?" ;echo $sentinel//g;
        $output =~ s/"rstatus=\$\?" ;//g;

        # remove the outer shell around the command
        $cmd =~ s/^\(//;
        $cmd =~ s/\) 2\>\&1$//;

        # strip sentinel and then status from output
        $output =~ s/\n$sentinel.*//;

        # strip any telnet ctrl characters that get
        # added to the output (this problem seems
        # to only occur with certain platforms).
        $output =~ s/ \cH//g;

        # locate status.  Remember that commands
        # echo themselves so make sure you don't
        # get the command line
        ($status) = $output =~ /rstatus=(\d+)/;

        # handle case where rstatus is not in the output
        # stream. One case is when NATE times out the command
        # In case of filer, do not do this since rstatus only
        # exists for clients.
        if (not defined $status) {
            print "Unable to determine status |@output|\n";
            $status = 1;
        }
    } ## end else [ if ($@ =~ /TIMEOUT/)

    $output =~ s/rstatus=\d+//;

    # create an array from the output
    @output = split(/\n/, $output);

    ## We need to determine if command is multiline
    ## and exists in the output.  IMPORTANT: Not all
    ## commands echo back their input so don't forget to
    ## check before modifying the output.
    my @cmds = split(/\n/, $cmd);

    ## If output contains second line of cmd
    ## then cmd was echoed to output and is
    ## multiline
    if (defined $cmds[1] and length $cmds[1]) {
        ## Remove continuation character and parens (from group commands
        ## added recently)
        $cmds[1] =~ s/\\//g;
        $cmds[1] =~ s/\)//g;
        $cmds[1] =~ s/\(//g;
        if ($output[0] =~ /$cmds[1]/) {
            ## determine size of command and strip
            ## from there.
            @output = @output[@cmds - 1 .. @output - 1];

        }
    } ## end if (defined $cmds[1] and...

    ## Go through the output and remove any line that has the command in it
    ## We do this, because none of the function that call execute expect
    ## to have to filter out their cmd from the returned output.
    my @output_sorted;
    foreach my $line (@output) {
        my $match = 0;
        foreach my $cmd (@cmds) {
            if ($cmd eq $line) {
                $match = 1;
                last;
            }
        }

        if ($match == 0) {
            push(@output_sorted, $line);
        }
    } ## end foreach my $line (@output)

    @output = @output_sorted;

    # Check to see if the caller wants us to retry the command if it failed
    if ($status != 0 and $retry != 0) {

        # Delete the argument -retry, no need to send it to the retry calls
        delete($args{-retry});

        # While status is not 0 (failed) and retries still remain
        # execute the specified commands
        while ($status != 0) {
            &SanUtil::nate_log(
                '-warn',
                "SanUtil::nate_server_execute: Retry execution of cmd ($cmd) ($retry) more times"
                );

            # Sleep for the -retry_delay specified
            &SanUtil::nate_sleep($retry_delay);

            # disable show output because it will look messy through the recursive calls
            ($status, @output) =
              &SanUtil::nate_server_execute(%args, -show_output => 0);
            $retry--;

            # Check to see if we have run out of retries
            if ($retry <= 0) {
                last;
            }
        } ## end while ($status != 0)
    } ## end if ($status != 0 and $retry...

    if ($show_output) {
        &SanUtil::nate_log(
            '-comment',
            "NATE_CONTROL_SERVER << Output from command ($cmd) (status: $status)\n@output"
            );
    }

    return ($status, @output);
} ## end sub nate_server_execute

################################################################################
# Method:    cmds_driver_call
# Objective: Calls execute_cmd with the cmd, and options that are passed in.
# Inputs  -cmd     - required - cmd to execute
#         -options - optional - hash of options to use
# Returns $SanUtil::FALSE and _set_err - help failed.
#         $SanUtil::TRUE on success
################################################################################
sub cmds_driver_call {
    my ($self, %args) = @_;
    my $cmd;        # arg - required - cmd to execute
    my $options;    # arg - optional - hash of options to use

    $self->sanlog(-msg       => "Calling SanUtil::cmds_driver_call",
                  -msg_level => 3);

    my $err_prefix =
      $self->_parent_caller . "():" . $self->_parent_lineno . " -";

    # This function must be called from within a SanFiler context.  If not
    # called from SanFiler, exit.
    unless (ref($self) =~ /SanFiler/) {
        $self->_set_err(
            "$err_prefix cmds_driver_call must be called from SanFiler object");
        return $SanUtil::FALSE;
    }

    my $cmd_obj = $self->cmd_obj();

    # Debugging
    my $dumpOut = Dumper(%args);
    $self->sanlog(
               -msg => "SanUtil::cmds_driver_call(): Args passed in:\n$dumpOut",
               -msg_level => 3);

    # Parse arguments
    foreach (keys %args) {
        if (/-?cmd/) {
            $cmd = $args{$_};
            if (not length($cmd)) {
                $self->_set_err(
                               "$err_prefix Invalid value for argument '-cmd'");
                return $SanUtil::FALSE;
            }
            next;
        } elsif (/-?options/) {
            $options = $args{$_};

            # Don't check for length.  We parse the keys below.
            next;
        } else {
            $self->_set_err("$err_prefix Invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    }    # end foreach arg

    # check for existance of required arguments
    unless (defined $cmd) {
        $self->_set_err("$err_prefix Missing required argument '-cmd'");
        return $SanUtil::FALSE;
    }

    # Build the command line
    my %cmd_args = (-cmd_filer     => $self->{filer},
                    -check_output  => 1,
                    -check_result  => 1,
                    -cmd_interface => $self->interface(),
                    -cmd           => "$cmd");

    # If the user passed in some options, add them to the cmd_args.
    while (my ($key, $value) = each(%{$options})) {
        $cmd_args{"$key"} = $value;
    }

    # Execute the cmd.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err($cmd_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->log_cmd_results($resultsHash_ref);

    # If check_result flag is on AND the result check failed, return error.
    if ($cmd_args{-check_result} && !$resultsHash_ref->{result_check}) {
        $self->_set_err(  "$err_prefix results check failed:\n"
                        . "$cmd: result=$resultsHash_ref->{result}");
        return $SanUtil::FALSE;
    }

    # If check_output flag is on AND the output check failed, return error.
    if ($cmd_args{-check_output} && !$resultsHash_ref->{output_check}) {
        $self->_set_err(  "$err_prefix output check failed:\n"
                        . "$cmd: output=$resultsHash_ref->{output}");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;

}    # end cmds_driver_call

################################################################################
# Method:    execute_cmd_array
# Objective: Loops to execute tests which are contained in an array of hashes.
# Inputs:    -tests (required) => array of hashes
#            Each array index contains:
#               (required) cmd => "xml test name to execute"
#               (optional) options => A hash of options to append.
#            -stop_on_error (optional)
#                1 = exit when first error is detected
#                0 = (default) = do not exit on error, execute all cmds
# Sample format:
#      @tests = ({cmd  => "radius_neg-add"},
#                {cmd  => "radius_neg-add-p"},
#                {cmd     => "radius_neg-add-ip-invalid",
#                 options => {"-radius-ip-addr" => "1.1.1.1.1"}},
#                {cmd     => "radius_neg-add-ip-invalid",
#                 options => {"-radius-ip-addr" => "1.1.1.1.1",
#                             "-port"           => 100}});
#       $Filer_Obj->execute_cmd_array("-tests" => \@tests));
# Returns $SanUtil::FALSE - failures detected and set_err.
#         $SanUtil::TRUE  - all tests passed
################################################################################
sub execute_cmd_array {
    my ($self, %args) = @_;
    my $tests;              # arg - required - tests and options to execute
    my $options;            # arg - optional - hash of options to use
    my @failed_subtests;    # Array listing the failed sub-tests.
    my $stop_on_error = 0;  # See description above.

    $self->sanlog(-msg       => "Calling SanUtil::execute_cmd_array",
                  -msg_level => 3);

    # Debugging
    my $dumpOut = Dumper(%args);
    $self->sanlog(
              -msg => "SanUtil::execute_cmd_array(): Args passed in:\n$dumpOut",
              -msg_level => 3);

    # Parse arguments
    foreach (keys %args) {
        if (/-?tests/) {
            $tests = $args{$_};
            if (not length($tests)) {
                $self->_set_err("Invalid value for argument '-tests'");
                return $SanUtil::FALSE;
            }
            next;
        } elsif (/-?stop_on_error/) {
            $stop_on_error = $args{$_};
            if (not length($stop_on_error) or $stop_on_error !~ /^0|1$/) {
                $self->_set_err("Invalid value for argument '-stop_on_error'");
                return $SanUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    }    # end foreach arg

    # check for existance of required arguments
    unless (defined $tests) {
        $self->_set_err("Missing required argument '-tests'");
        return $SanUtil::FALSE;
    }

    # Loop on all of the tests in the array.
    foreach my $current_hash (@{$tests}) {

        # Insert pause to avoid sentinal 'DoNe' out of order issues
        if ($self->interface() eq "cli") {
            &SanUtil::nate_sleep(2);
        }

        # Execute the cmd.  If it fails, add the test name and options to the
        # list of failures.
        if (!$self->cmds_driver_call(%{$current_hash})) {
            $self->sanlog(-msg       => $self->errmsg(),
                          -msg_level => 1);

            # Build error string with test case name and options.
            my $err_str = $current_hash->{"cmd"} . " with options: ";
            while (my ($key, $value) = each(%{$current_hash->{"options"}})) {
                $err_str .= "$key => $value, ";
            }

            # Save the error string.
            push(@failed_subtests, $err_str . "\n");

            # If user wants to stop on error, then break out of for loop.
            if ($stop_on_error) {
                last;
            }

        } ## end if (!$self->cmds_driver_call...

    } ## end foreach my $current_hash (@...

    # If there are items in @failed_subtests, return and indicate failure.
    if (scalar @failed_subtests) {
        my $total = scalar @failed_subtests;
        $self->_set_err("$total sub-tests failed:\n@failed_subtests");
        return $SanUtil::FALSE;
    }

    # Fail count is zero, everything passed.
    return $SanUtil::TRUE;

}    # end execute_cmd_array

################################################################################
# Subroutine Name: get_device_connection
# Objective:       Given a device, this function can return the proper
#                  connection type.
################################################################################
sub get_device_connection {
    my (%args) = @_;

    my ($resource_name);

    # Parse the args passed in
    foreach (keys %args) {
        if (/-?resource_name/) {

            # Check to make sure that we were actually passed in something
            if (not length($resource_name = $args{$_})) {
                &SanUtil::nate_log('-warn', "Invalid value for arguement '$_'");
                return undef;
            }
        }
    } ## end foreach (keys %args)

    # Make sure we have the required parameters
    if (not defined($resource_name)) {
        &SanUtil::nate_log('-warn',
                           "Missing required parameter '-resource_name'");
        return undef;
    }

    # Get the device data handle from the host function
    my $ddh = &Tharn::host($resource_name);
    if (not defined($ddh) or ref($ddh) !~ /hostrec/i) {
        &SanUtil::nate_log('-warn',
               "Unable to retrieve Hostrec object for device ($resource_name)");
        return undef;
    }

    # Get the host type so we can set the default connection
    # type if it wasn't passed in.
    my $host_type = $ddh->{hosttype};
    if (not defined($host_type) or $host_type eq "") {
        &SanUtil::nate_log('-warn',
                   "Unable to determine host type for device ($resource_name)");
        return undef;
    }

    if (not defined($ddh->{connection_fields})) {

        # Define the default connection unless it is already defined
        if ($host_type =~ /windows/i) {
            return "#rsh";
        } else {
            &SanUtil::nate_log(
                '-warn',
                "Unable to find any connection entries for device ($resource_name)"
                );
            return undef;
        }
    } elsif (defined($ddh->{connection_fields}->{ssh})) {
        return "#ssh";
    } elsif (defined($ddh->{connection_fields}->{rsh})) {
        return "#rsh";
    } elsif (defined($ddh->{connection_fields}->{"libssh-persay"})) {
        return "#libssh-persay";
    } elsif (defined($ddh->{connection_fields}->{telnet})) {
        return "#telnet";
    } elsif (defined($ddh->{connection_fields}->{rlogin})) {
        return "#rlogin";
    } elsif (defined($ddh->{connection_fields}->{"rsh-socket"})) {
        return "#rsh-socket";
    } elsif (defined($ddh->{connection_fields}->{console})) {
        return "#console";
    } else {

        # If we didn't find any of the connection fields we were
        # looking for, return #rsh for windows and undef for
        # everything else. We will set an error if the host isn't
        # windows.
        if ($host_type =~ /windows/i) {
            return "#rsh";
        } else {
            &SanUtil::nate_log(
                '-warn',
                "Unable to find any connection entries for device ($resource_name)"
                );
            return undef;
        }
    } ## end else [ if (not defined($ddh->...
} ## end sub get_device_connection

################################################################################
# Subroutine Name: get_login_credentials
# Objective:       Pass in  a resource name and connection type and
#                  return a username and password.
################################################################################
sub get_login_credentials {
    my (%args) = @_;

    my ($resource_name, $connection_type, $username, $password);

    # Parse the args passed in
    foreach (keys %args) {
        if (/-?resource_name/) {

            # Check to make sure that we were actually passed in something
            if (not length($resource_name = $args{$_})) {
                &SanUtil::nate_log('-warn', "Invalid value for argument '$_'");
                return (undef, undef);
            }
        } elsif (/-?connection_type/) {

            # Check to make sure that we were actually passed in something
            if (not length($connection_type = $args{$_})) {
                &SanUtil::nate_log('-warn', "Invalid value for argument '$_'");
                return (undef, undef);
            }

            # Remove the "#" from the front of the connection_type.
            $connection_type =~ s/^#//g;
        } ## end elsif (/-?connection_type/)
    } ## end foreach (keys %args)

    # Make sure we have the required parameters
    if (not defined($resource_name)) {
        &SanUtil::nate_log('-warn',
                           "Missing required parameter '-resource_name'");
        return (undef, undef);
    }

    if (not defined($connection_type)) {
        &SanUtil::nate_log('-warn',
                           "Missing required parameter '-connection_type'");
        return (undef, undef);
    }

    # Get the device data handle from the host function
    my $ddh = &Tharn::host($resource_name);
    if (not defined($ddh) or ref($ddh) !~ /hostrec/i) {
        &SanUtil::nate_log('-warn',
               "Unable to retrieve Hostrec object for device ($resource_name)");
        return (undef, undef);
    }

    # Check to make sure we have a connection entry for type connection type passed in.
    if (   not defined($ddh->{connection_fields})
        or not defined($ddh->{connection_fields}->{$connection_type})) {
        &SanUtil::nate_log(
            '-warn',
            "Unable to get any connection entries for connection type ($connection_type)"
            );
        return (undef, undef);
    }

    # Get the username for the connection type
    if (defined($ddh->{connection_fields}->{$connection_type}->{username})) {
        $username = $ddh->{connection_fields}->{$connection_type}->{username};
    } else {
        $username = undef;
    }

    # Get the password for the connection type
    if (defined($ddh->{connection_fields}->{$connection_type}->{password})) {
        $password = $ddh->{connection_fields}->{$connection_type}->{password};
    } else {
        $password = undef;
    }

    return ($username, $password);
} ## end sub get_login_credentials

################################################################################
# Subroutine Name: parse_ndvm_vol_names()
# Objective:       Parse the ndvm vol list and return a hash of filers with arrays of volumes
################################################################################
sub parse_ndvm_vol_names {
    my (%args) = @_;

    my (@filer_objs);

    # parse arguments
    foreach (%args) {
        if (/-?filers/) {
            @filer_objs = @{$args{$_}};
            if (scalar(@filer_objs) == 0) {
                &SanUtil::nate_log(
                    '-fatal',
                    "Error: SanUtil::parse_ndvm_vol_names() '-filers' value was empty"
                    );
            }
        }
    } ## end foreach (%args)

    # Make sure we received filer objects
    unless (scalar(@filer_objs) > 0) {
        &SanUtil::nate_log(
            '-fatal',
            "Error: SanUtil::parse_ndvm_vol_names() Missing parameter '-filers'"
            );
    }

    # Get some tharn parameters
    my $ndvm_vol_names = &get_param("NDVM_VOL_NAMES");

    # Make sure that NDVM_VOL_NAMES is populated
    if (not defined($ndvm_vol_names) or $ndvm_vol_names eq "") {
        &SanUtil::nate_log('-comment', "Invalid value for NDVM_VOL_NAMES");
        &SanUtil::nate_log('-script',
                           "SanUtil::parse_ndvm_vol_names(): "
                             . "Tharnhost NDVM_VOL_NAMES parameter is invalid.\n"
                             );
    }

    # Strip /vol/ from the names of the volumes passed in
    $ndvm_vol_names =~ s/\/vol\///g;

    my $ndvm_vol_hash = {};

    # If $ndvm_vol_names is in the form of filer1,vol1[,vol2,vol3:filer2,vol1,vol2], we
    # will break it apart and create the ndvm vol hash.  If it is a regex, we will use the
    # regex, get a list of volumes for the controller, eliminate any dummy vols and then
    # create teh ndvm vol hash.  Regex should look like this:
    # /reg.*ex/
    # /reg.*ex/i
    if ($ndvm_vol_names =~ /\S+\,\S+/) {

        # We should have a list that looks like the following:
        # filer1,vol1,vol2,vol3
        #
        # Or
        #
        # filer1,vol1,vol2,vol3:filer2,vol1,vol2:[etc]
        #
        # Split out the string
        my @filer_groups = split(/\:/, $ndvm_vol_names);

        # Iterate over the filer groups and split them into volume groups
        foreach my $filer_group (@filer_groups) {
            my @volume_group = split(/\,/, $filer_group);

            # We should now have an array with index 0 being the filer and the rest of the
            # array being volumes.  Make sure we have at least 2 items in the array (filer and 1 or more
            # volumes).
            unless (scalar(@volume_group) >= 2) {
                &SanUtil::nate_log('-fatal',
                    "Error: SanUtil::parse_ndvm_vol_names() Invalid value found "
                      . "for NDVM_VOL_NAMES");
            }

            # Shift off the filer name from @volume_group which will leave volumes
            my $filer = shift(@volume_group);

            # create the ndvm_vol_hash entry for the filer
            $ndvm_vol_hash->{$filer} = \@volume_group;
        } ## end foreach my $filer_group (@filer_groups)
    } elsif ($ndvm_vol_names =~ /^\s*\/(.+)\/(\S*)/) {
        my $vol_regex = qr/(?$2)$1/;

        # Use the filer objects passed in to generate retrieve all of the volumes and store it in
        # a hash with the following format.
        #
        # $filer_hash->
        #           {<filer_name>}->
        #                   'object'     => <filer object>
        #                   'volumes'    => <filer volumes>
        #
        my $filer_hash = {};
        foreach my $object (@filer_objs) {
            if (ref($object) =~ /SanFiler/i) {

                # Get the filer name
                my $filer_name = $object->get_name();
                if ($object->error()) {
                    &SanUtil::nate_log('-fatal',
                        "Error: SanUtil::parse_ndvm_vol_names() Unable to retrieve "
                          . "filer name:\n"
                          . $object->errmsg());
                }

                # Get the filer volumes
                my $volumes = $object->get_volnames();
                if ($object->error()) {
                    &SanUtil::nate_log('-fatal',
                        "Error: SanUtil::parse_ndvm_vol_names() Unable to retrieve "
                          . "filer volume names:\n"
                          . $object->errmsg());
                }

                # Initialize the hash
                $filer_hash->{$filer_name}->{object}  = $object;
                $filer_hash->{$filer_name}->{volumes} = $volumes;

            } else {

                # Looks like we didn't get a filer object so blow up.
                &SanUtil::nate_log('-fatal',
                    "Error: SanUtil::parse_ndvm_vol_names() Invalid value for "
                      . "'-filers' parameter.  Must be an array of filer objects."
                      );
            }
        } ## end foreach my $object (@filer_objs)

        # Iterate over the filers in the filer hash and create the ndvm_vol_hash with the contents.
        foreach my $filer (sort keys %{$filer_hash}) {
            my @volumes;
            foreach my $volume (@{$filer_hash->{$filer}->{volumes}}) {

                # Skip any dummy volumes
                if ($volume =~ /_dummy_/) {
                    next;
                }

                # Now try to match the regex. Add it to the @volumes array if matched.
                if ($volume =~ $vol_regex) {
                    push(@{$ndvm_vol_hash->{$filer}}, $volume);
                }
            } ## end foreach my $volume (@{$filer_hash...
        } ## end foreach my $filer (sort keys...

        # Check to make sure we found some volumes.  If we didn't, then fatal out and return.
        # If we did find volumes, use the list to build out the $ndvm_vol_hash hash
        unless (scalar(keys %$filer_hash) > 0) {
            &SanUtil::nate_log(
                 '-fatal',
                 "Error: SanUtil::parse_ndvm_vol_names() Unable to find any "
                   . "volumes that match the regular expression '$ndvm_vol_names'"
                   );
        }
    } else {
        &SanUtil::nate_log('-fatal',
                "Error: SanUtil::parse_ndvm_vol_names() Invalid value for "
              . "NDVM_VOL_NAMES parameter. Parameter must be as string in the form of\n"
              . "filer1,vol1,vol2,vol3:filer2,vol1,vol2,vol3 or a regular expression in the"
              . "form of /reg.*ex/ or /reg.*ex/i");
    }

    return $ndvm_vol_hash;
} ## end sub parse_ndvm_vol_names

################################################################################
# Subroutine Name: set_param()
# Objective:       Change the value of a parameter in the environment using
#                  Tharn.
################################################################################
sub set_param {
    my ($param, $value) = @_;

    # if the parameter doesnt exist, pull it into the environment.
    if (not &param_exists($param)) {
        &param("$param", "-default", "$value");
    }

    # Given a parameter $param, this will update its value with $value
    &Tharn::param("$param", -set, $value);

    return;
} ## end sub set_param

################################################################################
# Subroutine Name: get_param()
# Objective:       Obtain the value of a parameter in the environment using
#                  Tharn. Ignore warnings (refine) around this to avoid
#                  collisions with Tharn version (always called package
#                  qualified)
################################################################################
sub get_param {
    my ($param) = @_;

    # verify the param exists
    if (not &is_param_defined($param)) {
        return "";
    }

    # get the value from the environment
    return &Tharn::param("$param");
} ## end sub get_param

################################################################################
# Subroutine Name: param_exists()
# Objective:       Determine if the specified parameter exists in Tharn
################################################################################
sub param_exists {
    my ($param) = @_;

    if (not defined(&Tharn::param("$param"))) {
        return 0;
    } else {
        return 1;
    }
}

################################################################################
# Subroutine Name: is_param_defined()
# Objective:       Determine if the specified parameter is defined already.
################################################################################
sub is_param_defined {
    my ($param) = @_;

    if (not &param_exists($param)) {
        # param is not in environment, is not defined
        return 0;
    } elsif (not defined(&Tharn::param("$param")) or &Tharn::param("$param") eq "") {
        # param exists but is not defined or param is empty
        return 0;
    } else {
        # param exists and is defined
        return 1;
    }
}

################################################################################
# Subroutine Name: break_up_array()
# Inputs: array => \@arrayref
#         size  => $size_of_broken_up_arrays
# Output: this will take an input array and break it up into smaller arrays
# of size 'size'
################################################################################
sub break_up_array {
    my (%args) = @_;

    my $array = delete($args{array}) || die "Missing required parameter 'array'";
    my $size  = delete($args{size})  || die "Missing required parameter 'size'";

    my $hash = {};
    my $i = 0;
    my $j = 0;
    foreach my $element (@{$array}) {
        push(@{$hash->{$i}}, $element);
        $j++;
        if ($j >= $size) {
            $j = 0;
            $i++;
        }
    }

    return values %{$hash};
}

################################################################################
# Subroutine Name: wwn_add_colons()
# Inputs:    String containing a wwpn or wwnn as "-wwn => $wwpn"
# Ouptut:    String containing a wwpn or wwnn with colons added
# Objective: Take a wwpn or wwnn without colons and return one with colons.
################################################################################
sub wwn_add_colons {
    my (%args) = @_;

    my $wwn;

    # parse arguments
    foreach (%args) {
        if (/-?wwn/) {
            unless (length($wwn = $args{$_}) > 0) {
                &SanUtil::nate_log('-fatal',
                     "Error: SanUtil::wwn_add_colons() '-wwn' value was empty");
            }
        }
    }

    # make sure we were passed a wwn
    unless (defined($wwn)) {

        # Looks like we didn't get a wwn
        &SanUtil::nate_log(
            '-fatal',
            "Error: SanUtil::wwn_add_colons() Missing required parameter '-wwn'"
            );
    }

    # Just in case they passed in one with colons, go ahead and strip out any that exist.
    $wwn =~ s/\://g;

    my $count   = 0;
    my $new_wwn = "";
    while ($count < 14) {
        $new_wwn = $new_wwn . substr($wwn, $count, 2) . ":";
        $count += 2;
    }
    $new_wwn = $new_wwn . substr($wwn, $count, 2);

    return $new_wwn;
} ## end sub wwn_add_colons

################################################################################
# Subroutine Name  configure_remote_syslog_on_NCS
# Objective:
################################################################################
sub configure_remote_syslog_on_NCS {
    my (%args) = @_;

    my (@output, $status, $cmd);
    my (@syslog_conf_content, @syslog_conf_new_content, $remote_syslog_dir, $line_escaped);

    my $syslog_conf_file = "/etc/rsyslog.conf";

    # Parse arguments
    foreach (keys %args) {
        if (/-?remote_syslog_dir/) {
            if (not length($remote_syslog_dir = $args{$_})) {
                &SanUtil::nate_log(
                              '-fatal',
                              "San::SanUtil::configure_remote_syslog_on_NCS "
                                . "Missing value for argument '-remote_syslog_dir': $args{$_}"
                               );
            }
             next;
        }
    }

    if ($remote_syslog_dir =~ /^\s*$/) {
        &SanUtil::nate_log('-fatal',
                          "San::SanUtil::configure_remote_syslog_on_NCS "
                            . "'-remote_syslog_dir' is a required argument");
    }

    # Currently only Linux NCS are supported as remote syslog server
    if (&SanUtil::get_NATE_server_OS_type() !~ /linux/i) {
        &SanUtil::nate_log('-warn', "Currently NATE server running on Linux is " .
                                    "supported as remote syslog server. Make sure " .
                                    "you start your test on NATE server which is running on Linux");
        return $SanUtil::FALSE;
    }

    if (&SanUtil::is_rpm_installed_on_NATE_server(-rpm => "rsyslog") == $SanUtil::FALSE) {
       &SanUtil::nate_log('-warn', "rsyslog service is required on NATE server " .
                                    "to make it as remote syslog server. Request " .
                                    "NATE server admin for rsyslog service");
        return $SanUtil::FALSE;
    }

    # Check whether rsyslog service is running
    $cmd = "sudo service rsyslog status";
    ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
    if ($status != 0) {
        &SanUtil::nate_log('-warn', "rsyslog service is not running on NATE server. " .
                                    "Currently this is the only service supported for remote syslogging. " .
                                    "Request NATE server admin for rsyslog service");
        return $SanUtil::FALSE;
    }

    push (@syslog_conf_new_content, "\$ModLoad imudp\n");
    push (@syslog_conf_new_content, "\$UDPServerRun 514\n");
    push (@syslog_conf_new_content, "\$ModLoad imtcp\n");
    push (@syslog_conf_new_content, "\$InputTCPServerRun 514\n");
    push (@syslog_conf_new_content, "\$template DynFile\,\"$remote_syslog_dir/%fromhost-ip%_syslog\"\n");
    push (@syslog_conf_new_content, "*.* ?DynFile\n");
    push (@syslog_conf_new_content, "& ~\n");

    foreach my $line (@syslog_conf_new_content) {
       $line_escaped = $line;
       $line_escaped =~ s/(\/)/\\$1/g;
       $line_escaped =~ s/(\$)/\\$1/g;
       if($line_escaped =~ /\*\.\*/i) {
          $line_escaped =~ s/([*@.])/\\$1/g;
          $line_escaped =~ s/(\?)/\\$1/g;
       }

       my $cmd =  "perl -p -i.bak -e 'BEGIN{\$x=0}; END{exit \$x}; if (\$x==0) { if (s/^\\s*\#*\\s*($line_escaped)\\s*/\\1/) {\$x=1}}' $syslog_conf_file";
       ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
       if ($status == 1) {
           # then we successfully made the change, show the diff of the changes we made.
           $cmd = "diff $syslog_conf_file.bak $syslog_conf_file";
           &SanUtil::nate_server_execute(-cmd => $cmd);
       } else {
           # otherwise no instances of $param were modified, append the $param and value to the end of the file
           $cmd = "echo \'$line\' >> $syslog_conf_file";
           ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
           if ($status != 0) {
               &SanUtil::nate_log('-warn', "Command ($cmd) returned status ($status");
               return $SanUtil::FALSE;
           }
       }
    }

    $cmd = "cat $syslog_conf_file";
    ($status, @syslog_conf_content) = &SanUtil::nate_server_execute(-cmd => $cmd);
    if ($status != 0) {
        &SanUtil::nate_log('-warn', "Failed to display content of file ($syslog_conf_file) on NATE server");
        return $ClientUtil::FALSE;
    }
    $cmd = "sudo service rsyslog reload";
   ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
   if ($status != 0) {
       &SanUtil::nate_log('-warn', "Failed to reload syslog service on NATE server");
       return $ClientUtil::FALSE;
   }
   return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  get_NATE_server_OS_type
# Objective:       Will return OS type on NATE server. Will return Linux if OS is
#                  any flavour of linux else it will return Solaris because we
#                  only support RedHat(Linux) and Solaris OSes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_NATE_server_OS_type {
    my (@output, $status, $cmd);

    # Check for the OS type
    $cmd = "uname";
    ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
    if ($status > 0) {
        &SanUtil::nate_log('-fatal',
                           "San::SanUtil::get_NATE_server_OS_type "
                             . "Failed to run command \'$cmd\' on NATE Server");
    }
    if ($output[0] =~ /Linux/i) {
        return "linux";
    } else {
        return "solaris";
    }
} ## end sub get_NATE_server_OS_type

################################################################################
# Subroutine Name: is_rpm_installed_on_NATE_server()
# Objective: Will check whether rpm is installed or not on NATE Server
# Details: See POD documentation at the end of this file
################################################################################
sub is_rpm_installed_on_NATE_server {
    my (%args) = @_;

    my ($rpm, $cmd, $status, @output);

    # Parse arguments
    foreach (keys %args) {
        if (/-?rpm/) {
            if (not length($rpm = $args{$_})) {
                &SanUtil::nate_log(
                              '-fatal',
                              "San::SanUtil::is_rpm_installed_on_NATE_server "
                                . "Missing value for argument '-rpm': $args{$_}"
                                );
            }
            next;
        } ## end if (/-?rpm/)
    } ## end foreach (keys %args)

    if (not defined($rpm) or not length($rpm)) {
        &SanUtil::nate_log('-fatal',
                           "San::SanUtil::is_rpm_installed_on_NATE_server "
                             . "'-rpm' is a required argument");
    }

    $cmd = "rpm -q $rpm";
    ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
    if ($status > 0) {

        # Not setting the error because we will not fail out even if
        # this command fails
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub is_rpm_installed_on_NATE_server

###############################################################################
# Method Name: create_host_obj
# Objective:   Create a host object for cserver/node using NodeA nate_host
#              and dynamically set hostname, default_ip, prompt_norm
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_host_obj {
    my (%args) = @_;

    my (@nodes, $first_node, $prompt, $host_obj, $node_host_obj);
    my ($cluster_mgmt_ip);
    my ($username, $password, $conntype) = undef;
    my $is_cserver = $SanUtil::FALSE;

    %args = &SanUtil::strip_dash(%args);

    # if filer parameter, just create filer host object
    if (exists($args{filer})) {

        # assume it's filer. so we create filer host obj
        return &_create_filer_host_obj($args{filer});
    } elsif (exists($args{cserver})) {

        # we expected cluster mgmt IP and nodes to pass in with cserver
        # if not, we have to resolve by using nate host for backward compatibility
        if (exists($args{cluster_mgmt_ip})) {

            # set cluster mgmt ip, node list, and first node
            $cluster_mgmt_ip = $args{cluster_mgmt_ip};
            @nodes           = @{$args{nodes_in_cluster}};
            $first_node      = $nodes[0];

            # lets extract the connection information used for the first node's ssh entry
            # create host object using first_node, we will need this
            # when creating a connection into the cserver's cluster mgmt ip,
            # otherwise it would default to the DEFAULT_USER and DEFAULT_PASSWD constants
            $node_host_obj = &Tharn::host($first_node);
            unless (ref($node_host_obj) =~ /HostRec/i) {
                &SanUtil::nate_log('-fatal',
                                "failed to create host object for $first_node");
            }

            # save ssh conn info from node host object
            if (exists($node_host_obj->{connection_fields}->{ssh})) {
                $username =
                  $node_host_obj->{connection_fields}->{ssh}->{username};
                $password =
                  $node_host_obj->{connection_fields}->{ssh}->{password};
                $conntype =
                  $node_host_obj->{connection_fields}->{ssh}->{conntype};
            }
        } else {

            # get default cserver from nate host
            my $cserver = &Tharn::param('CLUSTER_NAME');

            # get a list of nodes from nate host
            my $node_list = &Tharn::param('NODES') || &Tharn::param('FILERS');
            @nodes = split /,|:/, $node_list;

            # determine the passed in cserver is actually cserver or node
            if ($args{cserver} =~ /$cserver/) {
                $is_cserver = $SanUtil::TRUE;
            } else {

                # if NODES or FILERS is not defined, we assume the arg is filer_name
                if (scalar(@nodes)) {
                    unless (
                        List::MoreUtils::any { $_ =~ /$args{cserver}/ }
                        @nodes
                      ) {
                        $is_cserver = $SanUtil::TRUE;
                    }
                }
            } ## end else [ if ($args{cserver} =~ ...

            # we can assume it's cluster server
            if ($is_cserver == $SanUtil::TRUE) {

                # create host object for cserver dynamically via first node
                # use first node because it contained cluster info
                $first_node = $nodes[0];

                # create host object using first_node
                $node_host_obj = &Tharn::host($first_node);
                unless (ref($node_host_obj) =~ /HostRec/i) {
                    &SanUtil::nate_log('-fatal',
                                "failed to create host object for $first_node");
                }

                # set cluster mgmt ip
                $cluster_mgmt_ip = $node_host_obj->{CLUSTER_MGMT_IP};

                # save ssh conn info from node host object
                if (exists($node_host_obj->{connection_fields}->{ssh})) {
                    $username =
                      $node_host_obj->{connection_fields}->{ssh}->{username};
                    $password =
                      $node_host_obj->{connection_fields}->{ssh}->{password};
                    $conntype =
                      $node_host_obj->{connection_fields}->{ssh}->{conntype};
                }
            } else {

                # assume it's filer. so we create filer host obj
                return &_create_filer_host_obj($args{cserver});
            }
        } ## end else [ if (exists($args{cluster_mgmt_ip...

        # create cserver host object
        my $cserver_hostspec =
            $args{cserver}
          . ";hosttype=filer-ng-cserver"
          . ";default_ip="
          . $cluster_mgmt_ip;
        $host_obj = &Tharn::host($cserver_hostspec);
        unless (ref($host_obj) =~ /HostRec/i) {
            &SanUtil::nate_log('-fatal',
                     "failed to create cserver host object for $args{cserver}");
        }

        # set ssh conn for cserver host object
        $host_obj->{connection_fields}->{ssh} = {};
        $host_obj->{connection_fields}->{ssh}->{username} =
          (defined $username) ? $username : DEFAULT_USERNAME;
        $host_obj->{connection_fields}->{ssh}->{password} =
          (defined $password) ? $password : DEFAULT_PASSWD;
        $host_obj->{connection_fields}->{ssh}->{conntype} =
          (defined $conntype) ? $conntype : DEFAULT_CONNTYPE;
        $host_obj->{connection_fields}->{ssh}->{connhost} = $cluster_mgmt_ip;
        $host_obj->{node} = $first_node;

        # set prompt norm for nodes
        foreach my $node (@nodes) {

            # set prompt for each node
            $prompt = $node . PROMPT_NORM;

            # only set prompt if it does not exist
            unless (
                List::MoreUtils::any { $_ =~ /$prompt/ }
                @{$host_obj->{prompt_norm}}
              ) {
                push @{$host_obj->{prompt_norm}}, $prompt;
            }
        } ## end foreach my $node (@nodes)

        # set prompt for cserver
        my $cserver_prompt = $args{cserver} . PROMPT_NORM;

        # only set prompt if it does not exist
        unless (
            List::MoreUtils::any { $_ =~ /$cserver_prompt/ }
            @{$host_obj->{prompt_norm}}
          ) {
            push @{$host_obj->{prompt_norm}}, $cserver_prompt;
        }

        return $host_obj;
    } else {
        &SanUtil::nate_log('-fatal',
                           "required -cserver or -filer parameter specified");
    }
} ## end sub create_host_obj

###############################################################################
# Method Name: array_get_duplicates
# Objective:   Retrieve a list of duplicate elements inside of the provided array
# Details:     See POD documentation at the end of this file.
###############################################################################
sub array_get_duplicates {
    my (@input_array) = @_;

    # Locals
    my %seen       = {};
    my @duplicates = ();

    foreach my $elem (@input_array) {
        next unless $seen{$elem}++;
        push(@duplicates, $elem);
    }

    return (wantarray()) ? @duplicates : \@duplicates;
} ## end sub array_get_duplicates

###############################################################################
# Method Name: hash_to_string
# Objective:   Convert a hash to a string
#  IE: $hash{key1} => "val1"
#      $hash{-key2} => "val2"
#    will become:
#      "-key1 val1 -key2 val2"
# Details:     See POD documentation at the end of this file.
###############################################################################
sub hash_to_string {
    my (%hash) = @_;

    # strip the leading dashes
    %hash = &SanUtil::strip_dash(%hash);

    # remove dups in the hash
    my %new_hash;
    foreach my $key (keys %hash) {
        $new_hash{$key} = $hash{$key};
    }

    # build the string
    my $string;
    foreach my $key (keys %new_hash) {
        $string .= "-" . $key . " " . $new_hash{$key} . " ";
    }
    chop($string);

    return $string;
} ## end sub hash_to_string

###############################################################################
##                               Private Methods
################################################################################

sub _create_filer_host_obj {
    my ($filer) = @_;
    my ($host_obj, $prompt);

    # create host object for filer
    $host_obj = &Tharn::host($filer);

    # set node attribute
    $host_obj->{node} = $host_obj->hostname();

    $prompt = $filer . PROMPT_NORM;

    # only set prompt if it does not exist
    unless (
        List::MoreUtils::any { $_ =~ /$prompt/ }
        @{$host_obj->{prompt_norm}}
      ) {
        push @{$host_obj->{prompt_norm}}, $prompt;
    }

    return $host_obj;
} ## end sub _create_filer_host_obj

use warnings qw(redefine);

1;

=pod

=head1 Package Name

SanUtil.pm

=head1 Author

Korry Nguyen

=head1 Description

SanUtils provides a common set of methods.  Use @ISA to
inherit these methods into another package.

=over 4

=item Examples

use vars qw( @EXPORT @ISA);

@ISA = qw(SanUtil);

=back

=head1 Public Methods

=head2 B<filer_connect()>

Make connection to the storage controller

=over 4

=item Inputs

=item * I<-conn_type> (default to rsh)

Type of connection to connect to storage controller (rsh,telnet,console,etc.)

=item Outputs

Connection object

=item Examples

 my $self->{connection} = $self->filer_connect(-conn_type => "telnet");

=back

=head2 B<xml2hash(xml_str)>

Convert XML format output into nested hash structure

=over 4

=item Inputs

=item * xml_str

XML string output

=item Outputs

Nested hash ref based on XML tags

=item Examples

    # execute ZAPI commands
    my $ret_ref = $cmdsDriver->execute_cmd(@api_args);
    my $out = $self->xml2hash($ret_ref->{output});

=back

=head2 B<cmds_driver_exist()>

Check if commands driver exist

=over 4

=item Outputs

TRUE if exist. otherwise return FALSE

=item Examples

    # execute ZAPI commands
    if ($self->cmds_driver_exist()) {
        $cmdDriver = $self->cmds_driver_get();
    }

=back

=head2 B<cmds_driver_get()>

Get command driver object

=over 4

=item Outputs

Commands driver object

=item Examples

    # execute ZAPI commands
    if ($self->cmds_driver_exist()) {
        $cmdDriver = $self->cmds_driver_get();
    }

=back

=head2 B<log_cmd_results()>

Prints out the results hash of a cmdsDriver object command call.

=over 4

=item Inputs

=item * resultsHashRef

 A reference to the results hash of a command. This is what is
 returned by CmdsDriver::execute_cmd().

=item Outputs

None

=item Examples

    # Print command results hash
    $self->log_cmd_results($resultsHash_ref);

=back

=head2 B<dumper_get(data_obj)>

Get data structure from object

=over 4

=item Inputs

=item * data object

Object with data structure

=item Outputs

String of data structure from the object

=item Examples

    #
    my $out_str = $self->dumper($data_obj);

=back

=head2 B<cmds_driver_call(%args)>

Calls a single execute_cmd with the cmd and options that are passed in.

=over 4

=item Inputs

=item * -cmd

(Required) Command to execute.

=item * -options

(Optional) Hash of options to use.

=item Outputs

On success, returns: $SanUtil::TRUE.
On failure, returns: $SanUtil::FALSE and _set_err.

=item Examples

    # Example where a cmd is stored in a hash:
    $self->cmds_driver_call(%{$current_hash});

    # Example without hash:
    $self->cmds_driver_call(-cmd     => "radius_neg-add-port-bogus-text",
                            -options => {"-radius-ip-addr"    => "hostname",
                                         "-port"              => "bogus22",
                                         "-is-default-server" => 1});

=back

=head2 B<get_device_connection(%args)>

Uses Tharn::host to determine and return the connection type for the device provided.

=over 4

=item Inputs

=item * -resource_name

(Required) Resource name to look up.

=item Outputs

On success, returns: string containing the connection type for the resource
On fail, returns: undef

=item Example:

    $conn_type = &SanUtil::get_device_connection(-resource_name => "name_of_resource");

=back

=head2 B<get_login_credentials(%args)>

Uses Tharn::host to determine and return the username and password for the resource name
and connection type provided.

=over 4

=item Inputs

=item * -resource_name

(Required) Resource name to look up.

=item * -connection_type

(Required) Connection type to look up.

=item Outputs

On success, returns: username and password strings ($username, $password).
On fail, returns: undef,undef

=item Example:

    ($username, $password) = &SanUtil::get_device_connection(-resource_name => "name_of_resource",
                                                             -connection_type => "connection_type");

=back

=head2 B<execute_cmd_array(%args)>

Executes multiple commands with their options that are stored in an array of hashes.
This function must be called from within a SanFiler context.

=over 4

=item Inputs

=item * -tests

(Required) Reference to an array of hashes which contains tests & options

  # Each array index contains:
  (required) cmd => xml test name to execute
  (optional) options => A hash of options to append.
  # Sample format:
  @tests = ({cmd => "radius_neg-add"},
            {cmd => "radius_neg-add-p"},
            {cmd     => "radius_neg-add-ip-invalid",
             options => {"-radius-ip-addr" => "1.1.1.1.1"}},
            {cmd     => "radius_neg-add-ip-invalid",
             options => {"-radius-ip-addr" => "1.1.1.1.1",
                         "-port"           => 100}});

=item * -stop_on_error

(Optional) 0 = (default) Do NOT stop on error, execute all cmds.
           1 = Stop executing commands on first error.

=item Outputs

On success, returns: $SanUtil::TRUE. (all tests passed)
On failure, returns: $SanUtil::FALSE and _set_err.

=item Examples

    # Using @tests input from sample format above:
    $Filer_Obj->execute_cmd_array("-tests" => \@tests);

    # Using @tests input from sample format above:
    $Filer_Obj->execute_cmd_array("-tests"         => \@tests,
                                  "-stop_on_error" => 1);

=back

=head2 B<create_host_obj(%args)>

Create a host record object for cserver or node

=over 4

=item Inputs

=item * -cserver | -filer

(Required) cserver or node name

=item * -cluster_mgmt_ip

(Optional) cluster mgmt ip address

=item * -nodes_in_cluster

(Optional) list of nodes in a cluster

=item Outputs

On success, returns: HostRec object
On fail, fatal exit

=item Example:

    $filer_host_obj = &SanUtil::create_host_obj(-filer => $filer_name);

    or

    $cserver_host_obj = &SanUtil::create_host_obj(-cserver => $cserver_name);

    or

    $cserver_host_obj = &SanUtil::create_host_obj(-cserver => $cserver_name,
                                                  -cluster_mgmt_ip => $clus_mgmt_ip,
                                                  -nodes_in_cluster => \@nodes);

=back

=head2 B<array_get_duplicates(%args)>

Returns an array containing the duplicated values (or refs) found in the provided
array.

=over 4

=item Inputs

=item * array

(Required) The array to work with.

=item Outputs

Another array with the duplicated values.

=item Example:
    @array_ips = (10.200.300.40, 10.200.300.41, 10.200.300.42, 10.200.300.43,
                  10.200.300.44, 10.200.300.40, 10.200.300.45, 10.200.300.42);
    @duplicated_ips = array_get_duplicates(@array_ips);

# Example Output
    @duplicated_ips = 10.200.300.40, 10.200.300.42

=back

=head2 B<hash_to_string(%args)>

Converts a key/value pair hash into a string.

=over 4

=item Inputs

=item * hash

(Required) The hash to work with

=item Outputs

A string representation of the hash.

=item Example:

    $hash{key1}  = "val1"
    $hash{-key2} = "val2"

    $string = &SanUtil::hash_to_string(%hash);
    # $string is now "-key1 val1 -key2 val2"

=back

=cut
