# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/San/SanXfun/SanSnapMirror.pm#1 $
#
# Copyright (c) 2009 NetApp, Inc.
# All rights reserved.
#
# ~/test/lib/San/SanXfun/SanSnapMirror.pm
##
## @summary SnapMirror Utility to test with SanXfun suite
## @description
##   This package store the SnapMirror environmental variables and functions to do SnapMirror operations
##
## @description Objective:
##   This package store the SnapMirror environmental variables and functions to do SnapMirror operations
##   like initalize, update, break, resync and status.
##
## @author Muthukumaran Govindarajan (muthug@netapp.com)
##
## @status public
##
## @pod here
##
## @change 2008-10-30 muthu: Original template created

package SanSnapMirror;

use strict;

use Tharn;    # for host
use Exporter;
use San::SanLog;
use San::SanUtil;
use Error;

use SnapMirror
  qw(snapmirror_status snapmirror_update snapmirror_quiesce snapmirror_break snapmirror_resync snapmirror_destinations_list);
use SnapMirror::Helper
  qw(sm_init rsm_init sm_wait_sync sm_cleanup sm_del_config);

use vars qw( @EXPORT @ISA );

#
# Inherit from the Error, SanLog, and SanUtil classes
#
@ISA = qw(Error SanLog SanUtil);

#
# constant values
#
use constant SM_TIMEOUT => 7200;    # 7200 secs. ie 2 hrs.

##############################################################################
###   Function:                  - Class Constructor
###   Parameters:                - See below
###   Returns:
###      on success              - this class object
##############################################################################
sub new {
    my ( $class, %args ) = @_;
    my $self = bless( {}, $class );

    $self->_clear_err();

    $self->{log} = \*ntest::NTEST_STDOUT;

    foreach ( keys %args ) {
        if (/-log/) {
            length( $self->{log} = $args{$_} )
              or do {
                $self->_set_err("Invalid value for \"-log\": $args{$_}");
                return $self;
              };
        }
        else {
            $self->_set_err("Invalid parameter specified: $_");
            return $self;
        }
    }
    $self->{'san_sm_setup'} = 0;

    return ($self);
}
##############################################################################
###   Function:     - setup
###                   Setup will initialize required parameters for SnapMirror
###                   functionality. 
###
###   Parameters:   - see below or the Pod Documents.
###
###   Returns:
###      on success  - 1
###      on failure  - 0
##############################################################################
sub setup() {
    my ( $self, %args ) = @_;

    $self->_clear_err();

    #
    # SnapMirror replication type (vsm or qsm)
    #
    $self->{'sm_replicate_type'} = $args{'-sm_replicate_type'};
    if ( !defined( $self->{'sm_replicate_type'} ) ) {
        $self->_set_err(
            "Mandatory parameter '-sm_replicate_type' not specified");
        return $SanUtil::FALSE;
    }
    elsif ( !length( $self->{'sm_replicate_type'} ) ) {
        $self->_set_err("Invalid value for '-sm_replicate_type' : " .
                                 $self->{sm_replicate_type} );
        return $SanUtil::FALSE;
    }
    $self->{'sm_replicate_type'} =
      ( $self->{'sm_replicate_type'} =~ /qsm/i ) ? "qsm" : "vsm";

    #
    # SnapMirror Sync type (sync or async)
    #

    $self->{'sm_sync_type'} = $args{'-sm_sync_type'};
    if ( !defined( $self->{'sm_sync_type'} ) ) {
        $self->_set_err("Mandatory parameter '-sm_sync_type' not specified");
        return $SanUtil::FALSE;
    }
    elsif ( !length( $self->{'sm_sync_type'} ) ) {
        $self->_set_err(
            "Invalid value for '-sm_sync_type' : $self->{sm_sync_type}");
        return $SanUtil::FALSE;
    }
    $self->{'sm_sync_type'} =
      ( lc( $self->{'sm_sync_type'} ) eq "sync" ) ? "sync" : "async";

    #
    # Source filer Name
    #
    $self->{'src_filer_name'} = $args{'-src_filer_name'};
    if ( !defined( $self->{'src_filer_name'} ) ) {
        $self->_set_err("Mandatory parameter '-src_filer_name' not specified");
        return $SanUtil::FALSE;
    }
    elsif ( !length( $self->{'src_filer_name'} ) ) {
        $self->_set_err(
            "Invalid value for '-src_filer_name' : $self->{src_filer_name}");
        return $SanUtil::FALSE;
    }

    #
    # Destiantion filer Name
    #
    $self->{'dst_filer_name'} = $args{'-dst_filer_name'};
    if ( !defined( $self->{'dst_filer_name'} ) ) {
        $self->_set_err("Mandatory parameter '-dst_filer_name' not specified");
        return $SanUtil::FALSE;
    }
    elsif ( !length( $self->{'dst_filer_name'} ) ) {
        $self->_set_err(
            "Invalid value for '-dst_filer_name' : $self->{dst_filer_name}");
        return $SanUtil::FALSE;
    }

    #
    # Source SnapMirror path
    #
    $self->{'src_snapmirror_path'} = $args{'-src_snapmirror_path'};
    if ( !defined( $self->{'src_snapmirror_path'} ) ) {
        $self->_set_err(
            "Mandatory parameter '-src_snapmirror_path' not specified");
        return $SanUtil::FALSE;
    }
    elsif ( !length( $self->{'src_snapmirror_path'} ) ) {
        $self->_set_err("Invalid value for '-src_snapmirror_path' : " . 
                                $self->{src_snapmirror_path});
        return $SanUtil::FALSE;
    }

    #
    # Destination SnapMirror path
    #
    $self->{'dst_snapmirror_path'} = $args{'-dst_snapmirror_path'};
    if ( !defined( $self->{'dst_snapmirror_path'} ) ) {
        $self->_set_err(
            "Mandatory parameter '-dst_snapmirror_path' not specified");
        return $SanUtil::FALSE;
    }
    elsif ( !length( $self->{'dst_snapmirror_path'} ) ) {
        $self->_set_err("Invalid value for '-dst_snapmirror_path' : " . 
                                        $self->{dst_snapmirror_path} );
        return $SanUtil::FALSE;
    }

    #
    # Initalize varaibale set it to 1.
    #
    $self->{'san_sm_setup'} = 1;
    return $SanUtil::TRUE;
}
##############################################################################
###   Function:     - san_sm_initialize
###                   Establish the Snap mirror connection between source and 
###                   destination volume and wait for the initial 
###                   update to complete
###
###   Parameters:   - None
###
###   Returns:
###      on success  - 1
###      on failure  - 0
##############################################################################
sub san_sm_initialize() {
    my ($self) = @_;

    $self->_clear_err();

    if ( !$self->{'san_sm_setup'} ) {
        $self->_set_err(
            "setup() should be called, before accessing this function");
        return $SanUtil::FALSE;
    }

    my $subtest_obj;

    my $src = "$self->{src_filer_name}:$self->{src_snapmirror_path}";
    my $dst = "$self->{dst_filer_name}:$self->{dst_snapmirror_path}";

    my $src_host = host( $self->{src_filer_name} );
    my $dst_host = host( $self->{dst_filer_name} );

    # if snampmirror sync type is *ASYNC*
    if ( $self->{sm_sync_type} =~ /async/i ) {
        $subtest_obj = sm_init(
            srchost => $src_host,
            dsthost => $dst_host,
            src     => $src,
            dst     => $dst,
            timeout => SM_TIMEOUT
        );
    }    # if snampmirror sync type is *SYNC*
    elsif ( $self->{sm_sync_type} =~ /sync/i ) {
        $subtest_obj = rsm_init(
            srchost => $src_host,
            dsthost => $dst_host,
            src     => $src,
            dst     => $dst,
            timeout => SM_TIMEOUT
        );
    }
    else {
        $self->_set_err("Unknown snapmirror sync type '$self->{sm_sync_type}'");
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}
##############################################################################
###   Function:         - san_sm_state
###                       Check the Snap mirror connection status between 
###                       source and destination volume. Check it is 
###                       Snap mirrored or not
###
###   Parameters:       - None
###
###   Returns:
###      on success     - 1, and state and status of the snapmirror
###      on failure     - 0
##############################################################################
sub san_sm_state() {
    my ($self) = @_;

    $self->_clear_err();

    if ( !$self->{'san_sm_setup'} ) {
        $self->_set_err(
            "setup() should be called, before accessing this function");
        return $SanUtil::FALSE;
    }

    my $src = "$self->{src_filer_name}:$self->{src_snapmirror_path}";
    my $dst = "$self->{dst_filer_name}:$self->{dst_snapmirror_path}";

    my $dst_host = host( $self->{dst_filer_name} );

    my $s;
    eval {
        $s = snapmirror_status( host => $dst_host, src => $src, dst => $dst );
    };

    my $text = $@;
    if ( $text =~ /SnapMirror status for non-existent relationship:/ ) {
        $self->_set_err($text);
        return $SanUtil::FALSE;
    }
    return ( $SanUtil::TRUE, $s->{'state'}, $s->{'status'} );
}
##############################################################################
###   Function:         - san_sm_update
###                       Update the Snap mirror connection between 
###                       source and destination volume and wait to complete
###
###   Parameters:       - None
###
###   Returns:
###      on success     - 1
###      on failure     - 0
##############################################################################
sub san_sm_update() {
    my ($self) = @_;

    $self->_clear_err();

    if ( !$self->{'san_sm_setup'} ) {
        $self->_set_err(
            "setup() should be called, before accessing this function");
        return $SanUtil::FALSE;
    }

    my $src      = "$self->{src_filer_name}:$self->{src_snapmirror_path}";
    my $dst      = "$self->{dst_filer_name}:$self->{dst_snapmirror_path}";
    my $dst_host = host( $self->{dst_filer_name} );

    my $subtest_obj;

    # if snampmirror sync type is *ASYNC*
    if ( lc( $self->{sm_sync_type} ) eq "async" ) {
        $subtest_obj = snapmirror_update(
            host    => $dst_host,
            src     => $src,
            dst     => $dst,
            timeout => SM_TIMEOUT
        );
    }    # if snampmirror sync type is *SYNC*
    elsif ( lc( $self->{sm_sync_type} ) eq "sync" ) {
        sm_wait_sync(
            host    => $dst_host,
            src     => $src,
            dst     => $dst,
            timeout => SM_TIMEOUT
        );
    }
    else {
        $self->_set_err("Unknown snapmirror sync type '$self->{sm_sync_type}'");
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}
##############################################################################
###   Function:         - san_sm_quiesce
###                       Quiesce the Snap mirror connection between 
###                       source and destination volume and wait to complete
###
###   Parameters:       - None
###
###   Returns:
###      on success     - 1
###      on failure     - 0
##############################################################################
sub san_sm_quiesce() {
    my ($self) = @_;

    $self->_clear_err();

    if ( !$self->{'san_sm_setup'} ) {
        $self->_set_err(
            "setup() should be called, before accessing this function");
        return $SanUtil::FALSE;
    }

    my $dst_host    = host( $self->{dst_filer_name} );
    my $subtest_obj = snapmirror_quiesce(
        host    => $dst_host,
        dstpath => $self->{dst_snapmirror_path},
        timeout => SM_TIMEOUT
    );
    return $SanUtil::TRUE;
}
##############################################################################
###   Function:         - san_sm_break
###                       Break the Snap mirror connection between 
###                       source and destination volume
###
###   Parameters:       - None
###
###   Returns:
###      on success     - 1
###      on failure     - 0
##############################################################################
sub san_sm_break() {
    my ($self) = @_;

    $self->_clear_err();

    if ( !$self->{'san_sm_setup'} ) {
        $self->_set_err(
            "setup() should be called, before accessing this function");
        return $SanUtil::FALSE;
    }

    my $dst_host = host( $self->{dst_filer_name} );

    # if QSM quise and then break
    if ( $self->{sm_replicate_type} =~ /qsm/i ) {
        snapmirror_quiesce(
            host    => $dst_host,
            dstpath => $self->{dst_snapmirror_path},
            timeout => SM_TIMEOUT
        );
    }
    snapmirror_break(
        host    => $dst_host,
        dstpath => $self->{dst_snapmirror_path}
    );
    return $SanUtil::TRUE;
}
##############################################################################
###   Function:         - san_sm_resync
###                       Re-Sync the Snap mirror connection between 
###                       source and destination volume and wait to complete
###
###   Parameters:       - None
###
###   Returns:
###      on success     - 1
###      on failure     - 0
##############################################################################
sub san_sm_resync() {
    my ($self) = @_;

    $self->_clear_err();

    if ( !$self->{'san_sm_setup'} ) {
        $self->_set_err(
            "setup() should be called, before accessing this function");
        return $SanUtil::FALSE;
    }

    my $src      = "$self->{src_filer_name}:$self->{src_snapmirror_path}";
    my $dst      = "$self->{dst_filer_name}:$self->{dst_snapmirror_path}";
    my $dst_host = host( $self->{dst_filer_name} );

    my $subtest_obj = snapmirror_resync(
        host    => $dst_host,
        src     => $src,
        dst     => $dst,
        timeout => SM_TIMEOUT
    );
    return $SanUtil::TRUE;
}

##############################################################################
###   Function:         - san_sm_snapshot_get
###                       return the current snaphot used by snampmirror
###   Parameters:       - None
###
###   Returns:
###      on success     - 1
###      on failure     - 0
##############################################################################
sub san_sm_snapshot_get() {
    my ($self) = @_;

    $self->_clear_err();

    if ( !$self->{'san_sm_setup'} ) {
        $self->_set_err(
            "setup() should be called, before accessing this function");
        return $SanUtil::FALSE;
    }

    my $src_host = host( $self->{src_filer_name} );

    my $s_ref;
    eval { $s_ref = snapmirror_destinations_list( host => $src_host ); };
    my $text = $@;
    if ($text) {
        $self->_set_err($text);
        return $SanUtil::FALSE;
    }

    #
    # TODO: Check this snapshot exist still in the filer,
    # or caller should check it out.....
    #
    return $s_ref->{'source-snapshot'};
}
##############################################################################
###   Function:         - san_sm_cleanup
###                       Release and Break the Snap mirror connection 
###                       between source and destination volume
###
###   Parameters:       - None
###
###   Returns:
###      on success     - 1
###      on failure     - 0
##############################################################################
sub san_sm_cleanup() {
    my ($self) = @_;

    $self->_clear_err();

    if ( !$self->{'san_sm_setup'} ) {
        $self->_set_err(
            "setup() should be called, before accessing this function");
        return $SanUtil::FALSE;
    }

    my $src = "$self->{src_filer_name}:$self->{src_snapmirror_path}";
    my $dst = "$self->{dst_filer_name}:$self->{dst_snapmirror_path}";

    my $src_host = host( $self->{src_filer_name} );
    my $dst_host = host( $self->{dst_filer_name} );

    #
    # require Quiesce only for QSM, but for safe case
    # this will be called for both VSM and QSM
    #
    snapmirror_quiesce(
        host    => $dst_host,
        dstpath => $self->{dst_snapmirror_path},
        timeout => SM_TIMEOUT
    );
    sleep(10);    # wait for quiesce to complete.
                  #
                  # Break the snapmirror relationship
                  #
    snapmirror_break(
        host    => $dst_host,
        dstpath => $self->{dst_snapmirror_path}
    );

    #
    # Cleanp the register entries
    #
    sm_cleanup(
        srchost => $src_host,
        dsthost => $dst_host,
        src     => $src,
        dst     => $dst
    );
    return $SanUtil::TRUE;
}

##############################################################################
#	                   P R I V A T E   M E T H O D S                         #
##############################################################################

#############################################################################
#                          POD Documentation                                #
#############################################################################

=pod

=head1 Package Name

SanSnapMirror.pm

=head1 Author

Muthukumaran Govindarajan (muthug@netapp.com)

=head1 Description

SanSnapMirror.pm provides an API interface to SnapMirror functioanlity.

=head1 See Also

Package uses the following modules. See these modules for further 
interface detail.

B<SnapMirror.pm>
   
   snapmirror_status 
   snapmirror_update 
   snapmirror_quiesce 
   snapmirror_break 
   snapmirror_resync 
   snapmirror_destinations_list

B<SnapMirror::Helper.pm>
  
  sm_init 
  rsm_init 
  sm_wait_sync 
  sm_cleanup 
  sm_del_config

=head1 Attributes

=head2 B<-sm_sync_type>

Defines the SnapMirror type, either sync or async

=head2 B<-sm_replicate_type>

Defines the SnapMirror replication type, either volume or qtree level replciation

=head2 B<-src_filer_name>

Name of the filer which will have the source volume

=head2 B<-dst_filer_name>

Name of the filer which will have the destination volume, for local transfer this name can be same as source filer name.

=head2 B<-src_snapmirror_path>
  
Defines the SnapMirror source volume path.
   if it is volume SnapMirror, just source volume name.
      src_snapmirror_path = src_volume_name
   
   If it is qtree SnapMirror, it should have full volume path with qtree
      src_snapmirror_path = /vol/src_volume_name/qtree_name

=head2 B<-dst_snapmirror_path>
  
Defines the SnapMirror destination volume path.
   if it is volume SnapMirror, just destination volume name.
      dst_snapmirror_path = dst_volume_name
   
   If it is qtree SnapMirror, it should have full volume path with qtree
      dst_snapmirror_path = /vol/dst_volume_name/qtree_name

=head1 Public Methods

Also refer SanXfunIoUtil.pm and SanSnapMirror.pm for addtional Public methods, which can 
be access throguh SanXfun object.

=head2 B<new()>

Constructor

=over 4

=item Inputs

=item * log (defaults to \*ntest::NTEST_STDOUT)

File descriptor for a log file.

=item Outputs

Returns a bless object of this class.

=item Examples

   #
   # Create a SanSnapMirror object
   #
   my $san_sm_obj = new SanSnapMirror(\*ntest::NTEST_STDOUT);
   if ($san_sm_obj->error) {
      my $err_mag = $san_sm_obj->errmsg();
      $Test->nlog("Failed to instantiate SanSnapMirror object:\nReason: $err_mag");
      return $TCD::FAIL;
   }

=back

=head2 B<setup(%args)>

Setup will initialize required parameters for SnapMirror functionality.

=over 4

=item Inputs

=item * I<-sm_replicate_type> <vsm | qsm> (default: vsm)

Defines the SnapMirror replication type, either volume or qtree level replciation

=item * I<-sm_sync_type> <sync | async> (default: async)

Defines the SnapMirror type, either sync or async

=item * I<-src_filer_name> 

Name of the filer which will have the source volume

=item * I<-dst_filer_name> 

Name of the filer which will have the destination volume, for local transfer this name can be same as source filer name.

=item * I<-src_snapmirror_path> 

Defines the SnapMirror source volume path.
   if it is volume SnapMirror, just source volume name.
      src_snapmirror_path = src_volume_name
   
   If it is qtree SnapMirror, it should have full volume path with qtree
      src_snapmirror_path = /vol/src_volume_name/qtree_name

=item * I<-dst_snapmirror_path> 

Defines the SnapMirror destination volume path.
   if it is volume SnapMirror, just destination volume name.
      dst_snapmirror_path = dst_volume_name
   
   If it is qtree SnapMirror, it should have full volume path with qtree
      dst_snapmirror_path = /vol/dst_volume_name/qtree_name

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

   #
   # Instantiate SanSnapMirror Object
   #
   my $san_sm_obj = new SanSnapMirror(\*ntest::NTEST_STDOUT);
   if ($san_sm_obj->error) {
      my $err_mag = $san_sm_obj->errmsg();
      $Test->nlog("Failed to instantiate SanSnapMirror object:\nReason: $err_mag");
      return $TCD::FAIL;
   }
   #
   # Call SanSnapMirror setup to create environment to testing
   #
   my @sm_agr;
   push( @sm_agr, "-sm_replicate_type",   $sm_replicate_type);
   push( @sm_agr, "-sm_sync_type",        $sm_sync_type);
   push( @sm_agr, "-src_filer_name",      $src_filer_name);
   push( @sm_agr, "-dst_filer_name",      $dst_filer_name);
   push( @sm_agr, "-src_snapmirror_path", $src_snapmirror_path);
   push( @sm_agr, "-dst_snapmirror_path", $dst_snapmirror_path);
         
   $san_sm_obj->setup(@sm_agr);
   if($san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $san_sm_obj->_set_err("Failed to setup SnapMirror object: $errmsg");
      return $SanUtil::FALSE;
   }

=back

=head2 B<san_sm_initialize()>

Establish the Snap mirror connection between source and destination volume 
and wait for the initial update to complete.

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

   #
   # Instantiate SanSnapMirror Object
   #
   my $san_sm_obj = new SanSnapMirror(\*ntest::NTEST_STDOUT);
   if ($san_sm_obj->error) {
      my $err_mag = $san_sm_obj->errmsg();
      $Test->nlog("Failed to instantiate SanSnapMirror object:\nReason: $err_mag");
      return $TCD::FAIL;
   }
   #
   # Call SanSnapMirror setup to create environment to testing
   #
   my @sm_agr;
   push( @sm_agr, "-sm_replicate_type",   $sm_replicate_type);
   push( @sm_agr, "-sm_sync_type",        $sm_sync_type);
   push( @sm_agr, "-src_filer_name",      $src_filer_name);
   push( @sm_agr, "-dst_filer_name",      $dst_filer_name);
   push( @sm_agr, "-src_snapmirror_path", $src_snapmirror_path);
   push( @sm_agr, "-dst_snapmirror_path", $dst_snapmirror_path);
         
   $san_sm_obj->setup(@sm_agr);
   if($san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $san_sm_obj->_set_err("Failed to setup SnapMirror object: $errmsg");
      return $SanUtil::FALSE;
   }
   #
   # create snapmirror realtionship
   #
   $san_sm_obj->san_sm_initialize();
   if ( $san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $self->_set_err("Failed to establish SnapMirror relationship: $errmsg");
      return $SanUtil::FALSE;
   }

=back

=head2 B<san_sm_state()>

Check the Snap mirror connection status between source and destination volume.
Check it is snap mirrored or not.

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful, along with State and status of snapMIrror.
If an error occured the error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

   # Instantiate SanSnapMirror Object
   my $san_sm_obj = new SanSnapMirror(\*ntest::NTEST_STDOUT);
   ...
   ...         
   $san_sm_obj->setup(@sm_agr);
   ...
   ...
   $san_sm_obj->san_sm_initialize();
   ...
   ...
   my ($ret, $state, $status) = $san_sm_obj->san_sm_state();
   if ( $san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $self->_set_err($errmsg);
      return $SanUtil::FALSE;
   }
   
=back

=head2 B<san_sm_update()>

Update the Snap mirror connection between source and destination volume and wait to complete

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

   # Instantiate SanSnapMirror Object
   my $san_sm_obj = new SanSnapMirror(\*ntest::NTEST_STDOUT);
   ...
   ...         
   $san_sm_obj->setup(@sm_agr);
   ...
   ...
   $san_sm_obj->san_sm_initialize();
   ...
   ...
   my ($ret, $state, $status) = $san_sm_obj->san_sm_state();
   if ( $san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $self->_set_err($errmsg);
      return $SanUtil::FALSE;
   }
   ....
   ....
   #
   # do snapmirror update 
   #
   $san_sm_obj->san_sm_update();
   if ( $san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $self->_set_err("Failed to Update SnapMirror destination: $errmsg");
      return $SanUtil::FALSE;
   }

=back

=head2 B<san_sm_quiesce()>

Quiesce the Snap mirror connection between source and destination volume and wait to complete

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

   # Instantiate SanSnapMirror Object, do setup and initialize 
   ...
   ...
   ...
   #
   # do quiesce 
   #
   $san_sm_obj->san_sm_quiesce();
   if ( $san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $self->_set_err("Failed to quiesce: $errmsg");
      return $SanUtil::FALSE;
   }

=back

=head2 B<san_sm_break()>

Break the Snap mirror connection between source and destination volume

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

   # Instantiate SanSnapMirror Object, do setup and initialize 
   ...
   ...
   ...
   #
   # do quiesce and break the snapmirror relationship
   #
   $san_sm_obj->san_sm_quiesce();
   ...
   $san_sm_obj->san_sm_break();
   if ( $san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $self->_set_err("Failed to break the SnapMirror relationship: $errmsg");
      return $SanUtil::FALSE;
   }

=back

=head2 B<san_sm_resync()>

Re-Sync the Snap mirror connection between source and destination volume and wait to complete.
This functiona can be used after the snapmirror break.

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

   # Instantiate SanSnapMirror Object, do setup and initialize 
   ...
   ...
   # do break the snapmirror relationship
   $san_sm_obj->san_sm_break();
   ...
   $san_sm_obj->san_sm_resync();
   if ( $san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $self->_set_err("Failed to re-eync the SnapMirror relationship: $errmsg");
      return $SanUtil::FALSE;
   }

=back

=head2 B<san_sm_snapshot_get()>

To get snapmirror snaphot from the snapmirror destiantion volume.

=over 4

=item Inputs

None

=item Outputs

Returns a the Snapshot name of snapmirror destination if successful returns. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

   # Instantiate SanSnapMirror Object, do setup and initialize 
   ...
   ...
   ...
   my $snapshotName = $san_sm_obj->san_sm_snapshot_get();
   if ( $san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $self->_set_err($errmsg);
      return $SanUtil::FALSE;
   }

=back


=head2 B<san_sm_cleanup()>

Release and Break the Snap mirror connection between source and destination volume.
And cleanup all the register entries from the controller.

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

   # Instantiate SanSnapMirror Object, do setup and initialize 
   ...
   ...
   # do all the other operation
   ...
   ...
   # once snapmirror no longer required, call this function
   $san_sm_obj->san_sm_cleanup();
   if ( $san_sm_obj->error() ) {
      my $errmsg = $san_sm_obj->errmsg();
      $self->_set_err($errmsg);
      return $SanUtil::FALSE;
   }

=back

=cut

1;

