# Hey emacs, this is basically -*- perl -*-
# $Id:  $
#
# Copyright (c) 2009 NetApp, Inc.
# All rights reserved.
#
# lib/San/SanXfun/UsecaseScsiCmds.pm
#
# @summary Usecase for SCSI commands using San::ScsiTool to use with
#  SanXfun environment.
#
# @description
#   Usecase for SCSI commands to test cross functionality using SanXfun
#
# @author Muthukumaran Govindarajan (muthug@netapp.com)
# @status public
# @change 2009-02-24 muthu: Original template created
# @change 2009-02-25 Ranjith Ayyandurai (ranjitha@netapp.com): POD documentation
# @change 2009-05-25 Jaikumar (jaikumar@netapp.com): Changes to functions.
#
# @pod here

# Class : UsecaseScsiCmds

package UsecaseScsiCmds;

use strict;

use San::ScsiTool;

use Tharn;    # for subtest, logcomment

use Exporter;
use San::SanLog;
use San::SanUtil;
use Error;

use vars qw( @EXPORT @ISA );

#
# Inherit from the Error, SanLog, and SanUtil classes
#
@ISA = qw(Exporter Error SanLog SanUtil);

use constant SCSI_READ6_BLK_SIZE => 255;

###############################################################################
# Method:    new (constructor)
# Objective: To instantiate this object to use with SanXfun.
# Inputs:    -sanxfun (must)    - refernce to the SanXfun Object
#            -log (optional)    - default to \*ntest::NTEST_STDOUT
#            -client (optional) - require for ScsiTool (see San::ScsiTool pod)
#                                 if not specified it will fill from SanXfun
#            -tool (optional)   - require for ScsiTool (see San::ScsiTool pod)
#                                 default to "spt"
#            -debug_level       - require for ScsiTool (see San::ScsiTool pod)
#
# Outputs:   bless object of this class, if error, error message can be retrived
#            thorugh errmsg().
###############################################################################
sub new {
    my ($class, %args) = @_;
    my $self = bless({}, $class);

    $self->{log}                    = \*ntest::NTEST_STDOUT;
    $self->{'sanxfun'}              = undef;
    $self->{'scsi_cmd_subtest_obj'} = undef;
    $self->{'scsi_tool_obj'}        = undef;

    $self->sanlog(-msg => "UsecaseScsiCmds new");
    $self->_clear_err();

    my $scsi_tool;

    foreach (keys %args) {
        if (/^-?sanxfun$/) {
            $self->{'sanxfun'} = $args{$_};
            if (ref($self->{'sanxfun'}) !~ /SanXfun/) {
                $self->_set_err(  "Invalid value for \"-sanxfun\": "
                                . "$args{$_}, need a SanXfun reference");
                $self->{'sanxfun'} = undef;
                return $self;
            }

            #
            # remove this entry, ScsiTool do not need this
            #
            delete $args{$_};
        } elsif (/^-?log$/) {
            $self->{log} = $args{$_};
            if (!length($self->{log})) {
                $self->_set_err("Invalid value for \"-log\": $args{$_}");
                $self->{'sanxfun'} = undef;
                return $self;
            }

            #
            # remove this entry, ScsiTool do not need this
            #
            delete $args{$_};
        } elsif (/^-?client$/) {
            my $sanClient = $args{$_};
            if (ref($sanClient) !~ /SanClient/) {
                $self->_set_err(  "Invalid value for \"-client\": "
                                . "$args{$_}, need a SanClient reference");
                $self->{'sanxfun'} = undef;
                return $self;
            }
            $self->{'sanclient'} = $sanClient;
        } elsif (/^-?tool$/) {
            $scsi_tool = $args{$_};
            if (!length($scsi_tool)) {
                $self->_set_err("Invalid value for \"-tool\": " . "$args{$_}");
                $self->{'sanxfun'} = undef;
                return $self;
            }
        }

        #
        # reset of the parameters might need for ScsiTool, just pass those
        #
    } # end foreach (keys %args)

    #
    # Check required paramters
    #
    if (ref($self->{'sanxfun'}) !~ /SanXfun/) {
        $self->_set_err("Need a Xfun reference for '-sanxfun'");
        $self->{'sanxfun'} = undef;
        return $self;
    }

    #
    # if -tool param not specified, default to spt....
    #
    if (!defined($scsi_tool) or !length($scsi_tool)) {
        $args{-tool} = "spt";
    }

    #
    # if -client param not specified, default to SanXfun Client....
    #
    if (  !defined($self->{'sanclient'})
        or ref($self->{'sanclient'}) !~ /SanClient/) {
        my $sanClient = $self->{'sanxfun'}->get_san_client();
        if ($self->{'sanxfun'}->error) {
            $self->_set_err("Unable to get Sanclinet Object from SsnXfun; "
                            . $self->{'sanxfun'}->errmsg());
            $self->{'sanxfun'} = undef;
            return $self;
        }
        $args{-client} = $sanClient;
        $self->{'sanclient'} = $sanClient;
    } # end if (!defined($self->{'sanclient'...

    #
    # create SCSI Tool Object
    #
    my $sctool_obj = ScsiTool->create_scsi_tool(%args);
    if ($sctool_obj->error) {
        $self->_set_err(
             "Failed to instantiate ScsiTool object: " . $sctool_obj->errmsg());
        $self->{'sanxfun'} = undef;
        return $self;
    }
    $self->{'scsi_tool_obj'} = $sctool_obj;
    return $self;
} # end sub new
###############################################################################
# Method:    test_scsi_cmds
# Objective: To send SCSI commands continuosly while testin other feature
# Inputs: -devices   - array reference to device list, this is optional if
#                      '-lun_objs' is passed.
#         -lun_objs  - array reference to Lun Objects, this is optional if
#                      '-devices' is passed.
#         -scsi_test_count - if this is zero scsi command will run infinte time
#           (default: 0)     until user calls "stop_scsi_cmds_test" or any
#                            failure occur during the test. If scsi_test_count
#                            non-zero then we loop that many times.
#         -scsi_test_intervals - in Seconds, wait between iterations
#             (default 5 sec)
# Outputs:   TRUE(1) if success, otherwise FALSE(0) error message can be
#            retrived thorugh errmsg().
###############################################################################
sub test_scsi_cmds() {
    my ($self, %args) = @_;

    $self->sanlog(-msg => "test_scsi_cmds");
    $self->_clear_err();

    $self->sanlog(-msg => "Inside the Subroutine");

    if (ref($self->{'sanxfun'}) !~ /SanXfun/) {
        $self->_set_err(  "-sanxfun is not initialize, "
                        . "create this object before calling this");
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Before the for loop");
    my $devices;
    my $lun_objs;

    my $scsi_test_count     = 0;
    my $scsi_test_intervals = 10;    # defualt to 10 secs.

    my $dev_objs_found = $SanUtil::FALSE;
    my $lun_objs_found = $SanUtil::FALSE;

    foreach (keys %args) {

        $self->sanlog(-msg => "Inside the loop");
        if (/-devices/) {
            $devices = $args{$_};
            if (ref($devices) !~ /ARRAY/) {
                $self->_set_err("Invalid value for \"-devices\": $devices"
                          . " Need to pass ARRAY Refernce of lun device names");
                return $SanUtil::FALSE;
            }
            $dev_objs_found = $SanUtil::TRUE;
        } elsif (/-lun_objs/) {
            $lun_objs = $args{$_};
            if (ref($lun_objs) !~ /ARRAY/) {
                $self->_set_err( "Invalid value for \"-lun_objs\": $lun_objs"
                               . " Need to pass ARRAY refernce of lun Objects");
                return $SanUtil::FALSE;
            }
            $lun_objs_found = $SanUtil::TRUE;
        } elsif (/-scsi_test_count/) {
            $scsi_test_count = $args{$_};
            if (!length($scsi_test_count)) {
                $self->_set_err(
                           "Invalid value for \"-scsi_test_count\": $args{$_}");
                return $SanUtil::FALSE;
            }
        } elsif (/-scsi_test_intervals/) {
            $scsi_test_intervals = $args{$_};
            if (!length($scsi_test_count)) {
                $self->_set_err(
                       "Invalid value for \"-scsi_test_intervals\": $args{$_}");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameter specified: $_");
            return $SanUtil::FALSE;
        }
    } # end foreach (keys %args)
    $self->sanlog(-msg => "Completed the validation part");

    #
    # check for devices list or lun list
    #
    if ($dev_objs_found != $SanUtil::TRUE and $lun_objs_found != $SanUtil::TRUE)
    {
        $self->_set_err("Need to sepcifiy either -lun_objs or -devices");
        return $SanUtil::FALSE;
    }

    #
    # if device is not specified and go get the device list based on lunObjs.
    #
    if ($dev_objs_found == $SanUtil::FALSE) {
        @{$devices} = $self->_get_lun_device_list(-lun_objs => $lun_objs);
        if ($self->error) {
            $self->_set_err($self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    #
    # check for backgroung, if not run it in forground.
    #
    $self->sanlog(-msg => "Starting scsi cmd test in background "
                  . " for device: @$devices");
    $self->{scsi_cmd_subtest_obj} =
      subtest(\&_scsi_cmds_test_subtest, '-bg',
              '-runid',                  "scsi_cmds_test",
              '-onexit',                 \&_scsi_test_exit_handler,
              '--',                      $self,
              $devices,                  $scsi_test_count,
              $scsi_test_intervals);

    #
    # wait for subtest to start
    #
    $self->sanlog(-msg => "Waiting for scsi cmd test to start....");

    SanUtil::nate_sleep(30);
    if (!$self->{scsi_cmd_subtest_obj}->is_alive()) {
        if (!$self->_get_scsi_cmds_test_result()) {
            return $SanUtil::FALSE;
        }
    }

    #
    #  if $scsi_test_count is 0, return true, mean running in the background for infintine time
    #
    if ($scsi_test_count != 0) {
        if ($self->{scsi_cmd_subtest_obj}->is_alive() == 1) {
            $self->{scsi_cmd_subtest_obj}->wait_finish();
        }
        if (!$self->_get_scsi_cmds_test_result()) {
            return $SanUtil::FALSE;
        }
    }
    return $SanUtil::TRUE;
} # end sub test_scsi_cmds()
###############################################################################
# Method:    stop_scsi_cmds_test
# Objective: To stop the SCSI command test (test_scsi_cmds) which is running
#            in the background
# Inputs:    None
# Outputs:   TRUE(1) if success, otherwise FALSE(0) error message can be
#            retrived thorugh errmsg().
###############################################################################
sub stop_scsi_cmds_test() {
    my ($self) = @_;

    if (!$self->{scsi_cmd_subtest_obj}) {
        $self->_set_err("scsi cmds test is not started.");
        return $SanUtil::FALSE;
    }

    if ($self->{scsi_cmd_subtest_obj}->is_alive() == 1) {
        $self->{scsi_cmd_subtest_obj}->kill(message =>
                ["Pass", "stop_scsi_cmds_test::SanXfun SCSI cmds test Passed"]);
    }
    $self->{scsi_cmd_subtest_obj}->wait_finish();
    $self->{scsi_cmd_subtest_obj} = undef;
    return $SanUtil::TRUE;
} # end sub stop_scsi_cmds_test()
###############################################################################
# Method:    scsi_read6
# Objective: To read from specified lba and no. of blocks using SCSI read6
#            command.
# Inputs: -device   - device name of the lun, this is optional if
#                      '-lun_obj' is passed.
#         -lun_obj  - lun Object, this is optional if '-device' is passed.
#                     if this is passed function will try to find the device
#                     name.
#         -lba      - Starting lba addres in decimal
#         -no_of_blocks - No of blocks to read from specified lba address.
#                         since it read6, maximum blocks can be read it 255.
# Outputs:   FALSE(0) if error, error message can be retrived thorugh errmsg().
#            if success, data return from the scsi command.
###############################################################################
sub scsi_read6() {
    my ($self, %args) = @_;

    $self->sanlog(-msg => "scsi_read6");
    $self->_clear_err();

    my ($device, $lun_obj, $lba, $no_of_blocks);

    my $dev_found     = $SanUtil::FALSE;
    my $lun_obj_found = $SanUtil::FALSE;

    foreach (keys %args) {
        if (/-device/) {
            $device = $args{$_};
            if (!length($device)) {
                $self->_set_err("Invalid value for \"-device\": $device");
                return $SanUtil::FALSE;
            }
            $dev_found = $SanUtil::TRUE;
        } elsif (/-lun_obj/) {
            $lun_obj = $args{$_};
            if (ref($lun_obj) !~ /Lun/) {
                $self->_set_err(  "Invalid value for \"-lun_obj\": $lun_obj"
                                . " Need to pass a refernce to lun Object");
                return $SanUtil::FALSE;
            }
            $lun_obj_found = $SanUtil::TRUE;
        } elsif (/-lba/) {
            $lba = $args{$_};
            if (!length($lba)) {
                $self->_set_err("Invalid value for \"-lba\": $args{$_}");
                return $SanUtil::FALSE;
            }
        } elsif (/-no_of_blocks/) {
            $no_of_blocks = $args{$_};
            if (!length($no_of_blocks)) {
                $self->_set_err(
                              "Invalid value for \"-no_of_blocks\": $args{$_}");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameter specified: $_");
            return $SanUtil::FALSE;
        }
    } # end foreach (keys %args)

    #
    # check for devices list or lun list
    #
    if ($dev_found != $SanUtil::TRUE and $lun_obj_found != $SanUtil::TRUE) {
        $self->_set_err("Need to sepcifiy either -lun_obj or -device");
        return $SanUtil::FALSE;
    }
    unless (defined($lba) or length($lba)) {
        $self->_set_err("Need to sepcifiy \"-lba\"");
        return $SanUtil::FALSE;
    }

    unless (defined($no_of_blocks) or length($no_of_blocks)) {
        $self->_set_err("Need to sepcifiy \"-no_of_blocks\"");
        return $SanUtil::FALSE;
    }

    if ($no_of_blocks > SCSI_READ6_BLK_SIZE) {
        $self->_set_err(
               "SCSI read6 do not support more than 255 blocks: $no_of_blocks");
        return $SanUtil::FALSE;
    }

    #
    # if device is not specified and go get the device list based on lunObjs.
    #
    if ($dev_found == $SanUtil::FALSE) {
        my @devices_list = $self->_get_lun_device_list(-lun_objs => [$lun_obj]);
        if ($self->error) {
            $self->_set_err($self->errmsg());
            return $SanUtil::FALSE;
        }
        $device = $devices_list[0];
    }

    my @lba_hex_str = unpack("A2 A2 A2", sprintf("%06x", $lba));
    my $no_of_blocks_in_hex = sprintf("%02x", $no_of_blocks);
    my $data_length = $no_of_blocks * 512;

    #
    # Read(6) 1 block: (lba 2097151)
    # cdb struture : 08-cmd for read 6, next 3 hex for lba address, next 1 hex for no of blocks, last must be 00
    # spt cdb='08 1f ff ff 01 00' dir=read length=512
    #
    my $cdb = "08 @lba_hex_str $no_of_blocks_in_hex 00";
    $self->sanlog(  -msg => "SCSI Read6 for\nDevice = $device\n"
                  . "LBA Address = $lba\n"
                  . "No Of blocks = $no_of_blocks\n"
                  . "cdb = $cdb");

    my $ret =
      $self->{'scsi_tool_obj'}->test_scsi_cmd(-device   => $device,
                                              -cdb_name => $cdb,
                                              -buf_len  => $data_length);

    if ($self->{'scsi_tool_obj'}->error) {
        my $errmsg =
          "Scsi Read 6 failed : " . $self->{'scsi_tool_obj'}->errmsg();
        $self->_set_err($errmsg);
        return $SanUtil::FALSE;
    }
    return $ret;
} # end sub scsi_read6()
###############################################################################
#                  P R I V A T E   M E T H O D S                        #
###############################################################################

###############################################################################
# Method:    _scsi_cmds_test_subtest
# Objective: Subtest function run in the back groud. This will test all the
#            critical SCSI commands. See San::ScsiTool::critical_cmds for the
#            details of SCSI commands.
# Inputs: see function for more details
# Outputs:   FALSE(0) with error message, if error occured.
###############################################################################
sub _scsi_cmds_test_subtest() {
    my ($self, $devices, $scsi_test_count, $scsi_test_interval) = @_;

    my $scsi_tool_obj = $self->{'scsi_tool_obj'};

    my $test_counter = 0;
    while (1) {
        ++$test_counter;
        last if $scsi_test_count != 0 && $test_counter > $scsi_test_count;
        foreach my $device (@{$devices}) {
            my $ret_code = $scsi_tool_obj->critical_cmds(-device => $device);
            if (!$ret_code) {
                my $err_msg =
                    "SanXfun SCSI cmds test failed for "
                  . "device ($device): "
                  . $self->errmsg();
                logresult("FAIL", $err_msg);
                $self->_set_err($err_msg);
                return $SanUtil::FALSE;
            }
        } # end foreach my $device (@{$devices...
        SanUtil::nate_sleep($scsi_test_interval);
    } # end while (1)
    logresult("PASS", "SanXfun SCSI cmds test Passed");
    return $SanUtil::TRUE;
} # end sub _scsi_cmds_test_subtest()
###############################################################################
# Method:    _get_lun_device_list
# Objective:  To get the device name for the given lun Object(s).
# Inputs:    -lun_objs - array refernce to Lun objects.
# Outputs:   FALSE(0) with error message, if error occured. or array of
#            device name
###############################################################################
sub _get_lun_device_list() {
    my ($self, %args) = @_;

    if ($args{'-lun_objs'} !~ /ARRAY/) {
        $self->_set_err("Need to pass ARRAY refernce of lun Objects");
        return $SanUtil::FALSE;
    }

    my @lunObjRef = @{$args{'-lun_objs'}};

    my $client_obj = $self->{'sanclient'};
    my $devmgr_obj = $client_obj->create_devmgr();
    if ($client_obj->error) {
        $self->_set_err(
                     "Failed to createDevMgr object: " . $client_obj->errmsg());
        return $SanUtil::FALSE;
    }

    #
    # Get the host name
    #
    my $client_name = $client_obj->hostname();
    if ($client_obj->error) {
        my $errmsg = $client_obj->errmsg();
        $self->_set_err("Call to hostname failed: ", "$errmsg");
        return $SanUtil::FALSE;
    }

    #
    # Get the host type
    #
    my $client_type = $client_obj->hosttype();
    if ($client_obj->error) {
        my $errmsg = $client_obj->errmsg();
        $self->_set_err("Call to hosttype failed: ", "$errmsg");
        return $SanUtil::FALSE;
    }
    $client_type =~ s/unix-solaris/solaris/;    # strip off unix-

    my %devarg = ();
    $devarg{-refresh} = 1;

    my @device_array = ();

    foreach my $lunObj (@lunObjRef) {
        if (ref($lunObj) !~ /Lun/) {
            $self->_set_err("Is not Lun Object refernce");
            return $SanUtil::FALSE;
        }

        #
        # Get full path for matching for the San::get_devices()
        #
        my $lun_name_path = $lunObj->get_path();
        if ($lunObj->error()) {
            my $errmsg = $lunObj->errmsg();
            $self->_set_err(  "Unable to get the lun path from "
                            . "the lun object for reason $errmsg");
            return $SanUtil::FALSE;
        }

        #
        # Stripe out the /vol/vol_name from the path,
        # get <qtree_name>/lun name
        #
        $lun_name_path =~ s#^/vol/\w+/(.*)#$1#;

        # get the client device list for the lun
        my $filer_name = $lunObj->get_filername();
        if ($lunObj->error()) {
            my $errmsg = join "\n", $lunObj->errmsg();
            $self->_set_err(  "Unable to get the filer name from "
                            . "the lun object for reason $errmsg");
            return $SanUtil::FALSE;
        }

        # get the volume name associated with the lun object
        my $vol_name = $lunObj->get_volname();
        if ($lunObj->error()) {
            my $errmsg = join "\n", $lunObj->errmsg();
            $self->_set_err(  "Unable to get the volume name from "
                            . "the lun object for reason $errmsg");
            return $SanUtil::FALSE;
        }

        $devarg{-filer} = $filer_name;
        $devarg{-vol}   = $vol_name;
        $devarg{-lun}   = "^$lun_name_path\$";    # match exact lun name

        $self->sanlog(-msg =>
            "fetching device name for $filer_name:$vol_name:$lun_name_path on client $client_name"
        );

        my @devices = $devmgr_obj->get_free_devices(%devarg);
        if ($devmgr_obj->error()) {
            my $errmsg = $devmgr_obj->errmsg();
            $self->_set_err(
                      "Get_devices failed for device($lun_name_path): $errmsg");
            return $SanUtil::FALSE;
        }
        if (@devices <= 0) {
            $self->_set_err("Device ('$lun_name_path') not found");
            return $SanUtil::FALSE;
        }

        #
        # Now that we have the devices, lets push them on our array.
        #
        foreach my $device (@devices) {
            $device =
              $self->_format_device_string($device, "raw", $client_type);
            push @device_array, $device;
        }    # end foreach device
             #
             # We only want to refresh (call sanlun one time)
             #
        $devarg{-refresh} = 0;
    } # end foreach my $lunObj (@lunObjRef)
    if (@device_array <= 0) {
        $self->_set_err("None of the devices found");
        return $SanUtil::FALSE;
    }

    return @device_array;
} # end sub _get_lun_device_list()
###############################################################################
# Method:    _format_device_string
# Objective: Format a devices so user can run I/O immediately.
# Inputs:    device       - the device in question
#            device_type  - raw|vm|fs
#            host_type    - type of client (solaris|hpux|aix|...)
#
# Outputs:   updated device name
###############################################################################
sub _format_device_string {
    my ($self, $device, $device_type, $host_type) = @_;

    if (($host_type =~ /solaris/) and ($device_type =~ /raw/)) {

        #
        # If we have a raw device, that's been formatted, lets
        # make sure that it writes to slice 6 instead of slice 2.
        #
        $device =~ s/s2/s6/g;
        unless ($device =~ /^\/dev\/rdsk\//) {
            $device = "/dev/rdsk/$device";
        }
    } elsif (($host_type =~ /linux/) and ($device_type =~ /raw/)) {

        #
        # Add the "/dev" if it's not there.
        #
        unless ($device =~ /^\/dev\//) {
            $device = "/dev/$device";
        }
    } elsif (($host_type =~ /hpux/) and ($device_type =~ /raw/)) {

        #
        # Add the "/dev/rdsk/" if it's not there.
        #
        unless ($device =~ /^\/dev\//) {
            $device = "/dev/rdsk/$device";
        }
    } elsif (($host_type =~ /aix/) and ($device_type =~ /raw/)) {

        #
        # Add the "/dev/rdsk/" if it's not there.
        #
        unless ($device =~ /^\/dev\//) {
            $device = "/dev/r" . $device;
        }
    }

    return $device;

} # end sub _format_device_string
###############################################################################
# Method:    _get_scsi_cmds_test_result
# Objective: To get a result from the Subtest
# Inputs:    None
# Outputs:   TRUE(1) if success, otherwise FALSE(0) error message can be
#            retrived thorugh errmsg().
###############################################################################
sub _get_scsi_cmds_test_result() {
    my ($self) = @_;

    if ($self->{scsi_cmd_subtest_obj}->is_alive() == 1) {
        $self->{scsi_cmd_subtest_obj}->wait_finish();
    }
    my @results       = $self->{scsi_cmd_subtest_obj}->results();
    my $scsi_test_err = "";
    if ("@results" !~ /SanXfun SCSI cmds test Passed/) {

        #
        # get the error from subtest run
        #
        foreach my $line (@results) {
            if ($line =~
                qr/SanXfun SCSI cmds test failed for device \(\s+\): \s+/) {
                if ($scsi_test_err eq "") {
                    $scsi_test_err = $line;
                } else {
                    $scsi_test_err .= "\n" . $line;
                }
                last;
            }
        } # end foreach my $line (@results)
    } # end if ("@results" !~ /SanXfun SCSI cmds test Passed/)
    if ($scsi_test_err ne "") {
        $self->_set_err($scsi_test_err);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} # end sub _get_scsi_cmds_test_result()
###############################################################################
# Method:    _scsi_test_exit_handler
# Objective: Exit handler for subtest, call automatically when Subtest exit.
# Inputs:    None
# Outputs:   None
###############################################################################
sub _scsi_test_exit_handler($) {
    my ($subtestobj) = @_;

    # Pass the child's results to the parent.
    my @child_results = $subtestobj->results();
    NATE::Result::store(@child_results);

    # Handle the worst result.
    my $worst = NATE::Result::worst(@child_results);
    my $type  = $worst->type();
    if ($type =~ /^FAIL|FATAL$/) {
        logresult("FAIL", "SanXfun SCSI cmds test failed");
    } elsif ($type eq 'SCRIPT') {
        logresult("-script", "SanXfun SCSI cmds test failed");
    }
} # end sub _scsi_test_exit_handler($)

1;

