#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/San/SanXgig.pm#1 $

# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Control a Finisar Xgig Chassis
## @author evan@netapp.com
## @status private
## @pod here

=pod

=head1 NAME

SanXgig.pm

=head1 DESCRIPTION

This package can be used to create a Finisar::TC, Finisar::XJ, or
Finisar::Frunner object to control a Finisar Xgig Analyzer, Jammer,
or Frunner product using the Frunner agent. Frunner must be running 
on a Windows machine. This package is also dependent on Finisar 
modules being available. Please refer to Finisar's help pages in
lib/Finisar/frunner.chm for available methods.

=head1 METHODS

=cut

package SanXgig;

use strict;

use Error;
use Finisar::XJ;
use Finisar::TC;
use Finisar::Frunner;

use vars qw(@ISA);
@ISA = qw(Error);

# Glocal variables
$SanXgig::TRUE  = 0;
$SanXgig::FALSE = 1;
$SanXgig::WARN  = 2;

# State strings. If more are added, please update the state_string
# subroutine as well
$SanXgig::STATE_CONFIG_ONLY = 0;
$SanXgig::STATE_CONNECTING  = 1;
$SanXgig::STATE_STARTING    = 2;
$SanXgig::SATE_RUNNING      = 3;
$SanXgig::STATE_STOPPING    = 4;
$SanXgig::STATE_STOPPED     = 5;
$SanXgig::STATE_SKIPPING    = 6;
$SanXgig::STATE_SKIPPED     = 7;
$SanXgig::STATE_PAUSING     = 8;
$SanXgig::STATE_PAUSED      = 9;

##########################################################################

=head2 C<new()>

=over 4

=item * Description

Creates a new SanXgig object.

=item * Arguments Hash

=over 4

=item 1 -frunner_host
 (Required)

The hostname or IP address of the Windows machine running the Frunner
application.

=item 2 -frunner_port
 (Optional)

The port that Frunner is listening on. Default is 1001.

=item 3 -frunner_dir
 (Optional)

The directory on the Frunner host that contains configuration information.
This is usually only used for the Jammer for the .tsk files. By default it
will use c:\temp.

=item 4 -chassis_host
 (Required)

The Finisar Xgig chassis hostname or IP address.

=item 5 -speed
 (Required)

The speed to set the Xgig ports to. This should be a numerical value
of 1, 2, 4, or 8.

=item 6 -obp
 (Required)

The ordinal, blade, port of the chassis to use. This should be an array
value of [ o, b, p] where o is the ordinal number, b is the blade number
and p is the port number. For example, a valid value would be [ 1, 2, 1].

=item 7 -ports
 (Optional)

The number of ports to lock from the ordinal, blade, port value. By
default, this will be 2. If the ordinal, blade, port value was [1, 2, 3],
and the ports value was 2, then it would lock [1, 2, 3] and [1, 2, 4].

=item 8 -type
 (Required)

The type of Xgig you are connecting to. This can be 'jammer', 'analyzer',
or 'frunner'.

=item 9 -personality
 (Optional)

The port personality type to use. This value can be "FC" for the Fibre
Channel Protocol or "GE" for the Gigabit Ethernet Protocol. Default
value is FC.

=back

=item * Returns

Finisar::XJ or Finisar::TC or Finisar::Frunner object with this object
stored in the 'SanXgig' key for reference.

=back

=cut

##########################################################################
sub new {
    my ($class, %args) = @_;
    my $xgig;
    my $self = {};
    bless ($self,$class);

    foreach (keys %args) {
        if (/-?frunner_host/) {
            length($self->{frunner_host} = $args{$_}) or
                do {
                    $self->_set_err("SanXgig::new() invalid value for \"-frunner_host\":",
                            "$args{$_}");
                    return $self;
                };
        }
        elsif (/-?frunner_port/) {
            length($self->{frunner_port} = $args{$_}) or
                do {
                    $self->_set_err("SanXgig::new() invalid value for \"-frunner_port\":",
                            "$args{$_}");
                    return $self;
                };
        }
        elsif (/-?frunner_dir/) {
            length($self->{frunner_dir} = $args{$_}) or
                do {
                    $self->_set_err("SanXgig::new() invalid value for \"-frunner_dir\":",
                            "$args{$_}");
                    return $self;
                };
        }
        elsif (/-?chassis_host/) {
            length($self->{chassis_host} = $args{$_}) or
                do {
                    $self->_set_err("SanXgig::new() invalid value for \"-chassis_host\":",
                            "$args{$_}");
                    return $self;
                };
        }
        elsif (/-?speed/) {
            if ( (not length($self->{speed} = $args{$_})) && $args{$_} !~ /^\d$/ ) {
                $self->_set_err("SanXgig::new() invalid value for \"-speed\":",
                        "$args{$_}");
                return $self;
            };
        }
        elsif (/-?obp/) {
            if ( ref($self->{obp} = $args{$_}) ne "ARRAY" ) {
                $self->_set_err("SanXgig::new() invalid value for \"-obp\":",
                        "$args{$_}");
                return $self;
            };
        }
        elsif (/-?ports/) {
            length($self->{ports} = $args{$_}) or
                do {
                    $self->_set_err("SanXgig::new() invalid value for \"-ports\":",
                            "$args{$_}");
                    return $self;
                };
        }
        elsif (/-?type/) {
            length($self->{type} = $args{$_}) or
                do {
                    $self->_set_err("SanXgig::new() invalid value for \"-type\":",
                            "$args{$_}");
                    return $self;
                };
        }
        elsif (/-?personality/) {
            length($self->{personality} = $args{$_}) or
                do {
                    $self->_set_err("SanXgig::new() invalid value for \"-personality\":",
                            "$args{$_}");
                    return $self;
                };
        }
    }

    # Check mandatory parameters
    if (!defined($self->{frunner_host})) {
        $self->_set_err("SanXgig::new() mandatory parameter \"-frunner_host\" not specified");
        return $self;
    }

    if (!defined($self->{chassis_host})) {
        $self->_set_err("SanXgig::new() mandatory parameter \"-chassis_host\" not specified");
        return $self;
    }

    if (!defined($self->{obp})) {
        $self->_set_err("SanXgig::new() mandatory parameter \"-obp\" not specified");
        return $self;
    }

    if (!defined($self->{type})) {
        $self->_set_err("SanXgig::new() mandatory parameter \"-type\" not specified");
        return $self;
    }

    # Set Frunner object identifier
    $self->{name} = "NATE-$self->{type}-" . $$;

    # Set the default Frunner port if not provided
    $self->{frunner_port} = 1001 if (!defined($self->{frunner_port}));

    # Set the default number of ports to use if not provided
    $self->{ports} = 2 if (!defined($self->{ports}));

    # Set the default personlity if not provided
    $self->{personality} = "FC" if (!defined($self->{personality}));

    # Convert speed value to string value
    if ($self->{speed} == 1) {
        $self->{speed} = "1.0625";
    } elsif ($self->{speed} == 2) {
        $self->{speed} = "2.1250";
    } elsif ($self->{speed} == 4) {
        $self->{speed} = "4.2500";
    } elsif ($self->{speed} == 8) {
        $self->{speed} = "8.5000";
    } else {
        $self->_set_err("SanXgig::new() unable to translate speed value of $self->{speed}");
        return $self;
    }

    if ($self->{type} =~ /jammer/i) {
        $xgig = Finisar::XJ::New($self->{name}, $self->{frunner_host}, $self->{frunner_port});
    } elsif ($self->{type} =~ /analyzer/i) {
        $xgig = Finisar::TC::New($self->{name}, $self->{frunner_host}, $self->{frunner_port});
    } elsif ($self->{type} =~ /frunner/i) {
        $xgig = Finisar::Frunner::New($self->{name}, $self->{frunner_host}, $self->{frunner_port});
    } else {
        $self->_set_err("SanXgig::new() invalid type parameter specified");
        return $self;
    }

    $xgig->Configdir( $self->{frunner_dir}? $self->{frunner_dir} : $xgig->{configDir} );
    $xgig->Personality($self->{personality});
    $xgig->Speed($self->{speed});
    $xgig->Port($self->{chassis_host}, @{$self->{obp}}, $self->{ports});
    
    # Save this objects settings into Finisar's object and return Finisar's object.
    $xgig->{SanXgig} = $self;
    return $xgig;
}

##########################################################################

=head2 C<state_string()>

=over 4

=item * Description

Converts the state code into a descriptive string.

=item * Arguments

=over 4

=item 1 State
 (Required)

The state code to be converted.

=back

=item * Returns

A descriptive string.

=back

=cut

##########################################################################
sub state_string {
    my ($state) = @_;

    my @state_array = qw (STATE_CONFIG_ONLY
                          STATE_CONNECTING
                          STATE_STARTING
                          STATE_RUNNING
                          STATE_STOPPING
                          STATE_STOPPED
                          STATE_SKIPPING
                          STATE_SKIPPED
                          STATE_PAUSING
                          STATE_PAUSED);

    return ($state_array[int($state)]);
}

=pod

=head1 AUTHOR

Evan Chiu (evan@netapp.com)

=head1 COPYRIGHT

Copyright (c) 2008 Network Appliance, Inc. All rights reserved.

=cut

