# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides a mechanism to load, retrieve, and translate schema of
## command definitions into actual command for execution
##
## @author korry@netapp.com, dl-ssan-test
## @status public

package Schema;

# compiler directives
use strict;

# imported modules
use San::Schema::Cdef;
use San::SanLog;
use San::SanUtil;
use Error;

use vars qw( @EXPORT @ISA @Known_cdef_locations );

@ISA = qw( Error SanLog SanUtil );

use NATE::Log qw(log_global);

my $Log = log_global();

# global variables
my $Schema = undef;

###############################################################################
# Method:    new()
# Objective: Create schema object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {};

    # global Schema object only instantiated once throughout an application
    # the reason is cdef schema object only needed to create once
    # this enables us to reuse Schema object throughout once it's created
    if (defined $Schema) {
        return $Schema;
    } else {

        # bless instance into class and save as a singleton
        $Schema = bless $self, $class;

        # initialize required field check flag
        # the intent is to enable req_field_chk flag so cdef object can validate
        # the passed in required parameters. However, there are discrepancy
        # with required params in cdefs data structure which forces us to
        # temporary disable this checking functionality and validate parameters
        # for all cdef calls.
        # Hopefully, they (Mhost team) will fix it in the future :-(.
        $Schema->{req_field_chk} = $SanUtil::FALSE;

        return $Schema;
    }
} ## end sub new
###############################################################################
# Method:    get_cdef_objs()
# Objective: return array of command definition (cdef) objects
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_cdef_objs {
    my ($self) = @_;

    return (wantarray()) ? @{$self->{cdef_objs}} : $self->{cdef_objs};
}

###############################################################################
# Method:    is_loaded()
# Objective: Determine if cdefs for a particular branch are loaded.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_loaded {
    my ($self, %args) = @_;

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    return (exists $self->{$args{filer}}) ? $SanUtil::TRUE : $SanUtil::FALSE;
}

###############################################################################
# Method:    load()
# Objective: Load command definitions
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load {
    my ($self, %args) = @_;

    my ($branch, $filer, $path, @cdef_objs, @files);
    my ($required_params, $optional_params, $alias_params, $alias_map);
    my @cdef_dirs = ();

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # process parameters
    while (my ($key, $value) =  each(%args)) {
        if ($key =~ /-?branch$/) {
            if (! length($branch = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /-?filer$/) {
            if (! length($filer = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /-?path$/) {
            if (! length($path = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $SanUtil::FALSE;
            }
        }
    }

    # verify mandatory parameters
    unless (defined $branch and length $branch) {
        $self->_set_err("mandatory parameter \"-branch\" not specified ");
        return $SanUtil::FALSE;
    }
    unless (defined $filer and length $filer) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $SanUtil::FALSE;
    }
    unless (defined $path and length $path) {
        $self->_set_err("mandatory parameter \"-path\" not specified ");
        return $SanUtil::FALSE;
    }

    # keep track of what branch each filer reflects
    $self->{$filer} = $branch;

    # verify if cdef schema of the specified branch has been loaded
    if ($self->_cdef_is_loaded(-branch => $branch)) {
        $self->sanlog(-msg => "cdefs for $branch are already loaded");
        return $SanUtil::TRUE;
    }

    # store the current branch
    push @{$self->{branches}}, $branch;

    # if cdef path string contained "/cdefs", it's valid
    # so load cdef files from there
    if ($path =~ /\/cdefs|\/NACL\/APISet/) {
        push @cdef_dirs, $path;
    } else {
        # path is buildroot path and cdef files are scatter in various directories
        # we need to load cdef files from a known cdef locations
        foreach my $location (@Known_cdef_locations) {
            my $cdef_path = $path . $location;
            push @cdef_dirs, $cdef_path;
        }
    }

    # load cdef files from the list of directories
    foreach my $dir (@cdef_dirs) {

        # check for valid cdef path
        if (-d $dir) {

            # retrieve cdef files
            # non-windows path
            opendir(DIR, $dir) or do {
                $self->_set_err("can not open dir: " . $dir);
                return $SanUtil::FALSE;
            };
            foreach my $file (readdir(DIR)) {
                # skip files that does not have cdefs extension
                next if ($file !~ /\.cdefs$/);
                push @files, "$dir\/$file";
            }
            closedir(DIR);
        } else {
            $self->_set_err("invalid cdef path: " . $dir);
            return $SanUtil::FALSE;
        }
    }

    # if files list is empty, cdef files have not been consolidated
    if (scalar(@files) == 0) {
        $self->_set_err("no cdef file found in: " . $path);
        return $SanUtil::FALSE;
    }

    # get command definition from each file and create cdef object per command
    foreach my $file (sort @files) {
        $self->sanlog(-msg_level =>  7,
                      -msg       => "cdef: " . $file);

        # initialize command definition hash per cdef file
        my %CommandDefinitions = {};

        # get command definitions info from cdef file in order to update
        # attribute/value pairs in CommandDefinition hash
        eval NATE::Inc::get_file("$file");

        # there are some issues with certain cdef files that don't
        # generate %Attributes variable causing exception in eval.
        # when the cdef generation is fixed, uncomment the below lines
        # to catch exception
         if ($@) {
             $Log->warn("CDEF file did not compile properly: $file");
             next;
         }

        # create cdef object for each command
        # and build command with its params
        foreach my $cdef (keys %{CommandDefinitions}) {

            # skip over definitions that are incomplete
            # the options attribute is part of the cdef file
            next if (not defined $CommandDefinitions{$cdef}{options});

            # get params of each type (required, optionals, and aliases)
            $required_params =
                (defined $CommandDefinitions{$cdef}{required_options})
              ? [@{$CommandDefinitions{$cdef}{required_options}}]
              : [];

            $optional_params =
                (defined $CommandDefinitions{$cdef}{optional_options})
              ? [@{$CommandDefinitions{$cdef}{optional_options}}]
              : [];

            $alias_params =
              (defined $CommandDefinitions{$cdef}{additional_options})
              ? $CommandDefinitions{$cdef}{additional_options}
              : {};

            $alias_map =
              (defined $CommandDefinitions{$cdef}{alias_map})
              ? $CommandDefinitions{$cdef}{alias_map}
              : {};

            # store alias keys in options attribute per command definition
            foreach my $key (keys %$alias_params) {
                $CommandDefinitions{$cdef}{options}{$key} = $alias_params->{$key};
            }

            # instantiate cdef object
            my $cdef_obj = Cdef->new(-action          => $CommandDefinitions{$cdef}{command},
                                     -required_params => $required_params,
                                     -optional_params => $optional_params,
                                     -alias_params    => $CommandDefinitions{$cdef}{options},
                                     -alias_map       => $alias_map,
                                     -file            => $file);
            if ($cdef_obj->error()) {
                $self->_set_err("failed to instantiate cdef object", $cdef_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # add to cdef array
            push @cdef_objs, $cdef_obj;

            # per cdef object, store it in schema object under branch attribute
            $self->{$branch}{$cdef} = $cdef_obj;
        }
    }
    # store array of cdef objects in schema object
    push @{$self->{cdef_objs}}, @cdef_objs;

    return $SanUtil::TRUE;
} ## end sub load

###############################################################################
# Method:    translate()
# Objective: return cmd syntax for requested system operation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub translate {
    my ($self, %args) = @_;

    my ($branch, $cmd, $cdef_obj, $filer);

    $self->_clear_err();

    # remove the dash '-' from input args because attributes could be used
    # to build the 'get' accessor routines on the fly
    %args = &SanUtil::strip_dash(%args);

    # pre-set returned cmd to empty string
    $cmd = "";

    # ensure the 'filer' and 'cdef' parameters are passed in
    foreach my $param ('filer', 'cdef') {
        unless (defined($args{$param})) {
            $self->_set_err("missing required parameter '" . $param . "'");
            return $cmd;
        }
    }

    # retrieve the branch from schema object
    $branch = $self->{$args{filer}};

    # remove filer arg that is not needed for translation
    delete $args{filer};

    # retrieve the requested cdef object
    $cdef_obj = $self->{$branch}{$args{cdef}};
    unless (ref($cdef_obj) =~ /Cdef/i) {
        $self->_set_err("can not find cdef object for ". $args{cdef},
                        $self->dumper_get(\%args, $self->get_debug_level()));
        return "";
    }

    # This is where we get_req_field_check and pass to cdef_obj->translate().
    # However, there are discrepancy with required params in data structure
    # of cdef files which forced us to temporary disable this checking
    # functionality. Hopefully, they (Mhost team) will fix it in the future.

    # clear/disable the required field check flag and pass into cdef_obj
    $args{req_field_chk} = $self->disable_req_field_check();
    # translate definition
    $cmd = $cdef_obj->translate(%args);
    if ($cdef_obj->error()) {
        $self->_set_err("cdef translation failed",
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $cdef_obj->errmsg());
        return "";
    }

    return $cmd;
} ## end sub translate

###############################################################################
# Method:    get_req_field_check()
# Objective: retrieve state of cmd's input requirement check
###############################################################################
sub get_req_field_check {
    my ($self) = @_;

    return $self->{req_field_chk};
}

###############################################################################
# Method:    enable_req_field_check()
# Objective: set/enable the required field check flag
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable_req_field_check {
    my ($self) = @_;

    # set/enable the required field check flag
    $self->{req_field_chk} = $SanUtil::TRUE;

    # return the current state of required field flag
    return $self->{req_field_chk};
}

###############################################################################
# Method:   disable_req_field_check()
# Objective: clear/disable the required field check flag
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable_req_field_check {
    my ($self) = @_;

    # clear/disable the required field check flag
    $self->{req_field_chk} = $SanUtil::FALSE;

    # return the current state of required field flag
    return $self->{req_field_chk};
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _cdef_is_loaded()
# Objective: determine if cdefs for a particular branch have been loaded
###############################################################################
sub _cdef_is_loaded {
    my ($self, %args) = @_;


    # remove the dash '-' from input args because attributes could be used
    # to build the 'get' accessor routines on the fly
    %args = &SanUtil::strip_dash(%args);

    # return TRUE if a particular branch has been loaded and
    # stored in the schema object
    return grep { /$args{branch}/ } @{$self->{branches}};
}

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Schema.pm

=head1 Author

Korry Nguyen (dl-ssan-test)

=head1 Description

Schema.pm provides a mechanism to load, retrieve, and translate schema of command definitions (cdefs) into actual commands for execution on the filer.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm, San::Schema::Cdef.pm>

=head1 Attributes

=head2 B<req_field_chk>

boolean = <0 | 1> (default is 0) flag to check required fields of command definition (cdef)

=head1 B<Public Methods>

=head2 B<new()>

Create a schema object. If schema object already exist, it will be reused.

=over 4

=item Inputs

None

=item Outputs

Return a schema object

=item Examples

 use San::Schema;

 # Create schema object
 #
 my $schema_obj = Schema->new();
 if (! defined($schema_obj) or ref($schema_obj) !~ /Schema/) {
     $self->_set_err("can not create schema object");
     return $SanUtil::FALSE;
 }

=back

=head2 B<disable_req_field_check()>

Disable the required field check flag

=over 4

=item Inputs

None

=item Outputs

Return FALSE = 0

=item Examples

 # Disable the required field check flag when translating cdef into
 # command. This ignores the checking of cdef required fields.
 #
 $schema_obj->disable_req_field_check();
 my $cmd = $schema_obj->translate(-filer => $filer_name,
                                  -cdef  => $cdef,
                                  %args);
 if ($schema_obj->error()) {
     $self->_set_err("failed to translate cdef schema into command",
                     $schema_obj->errmsg());
     return $SanUtil::FALSE;
 }

=back

=head2 B<enable_req_field_check()>

Enable the required field check flag

=over 4

=item Inputs

None

=item Outputs

Return TRUE = 1

=item Examples

 # Check required fields when translating cdef request.
 # If any mismatch during the validation, translate will fail
 #
 $schema_obj->enable_req_field_check();
 my $cmd = $schema_obj->translate(-filer => $filer_name,
                                  -cdef  => $cdef,
                                  %args);
 if ($schema_obj->error()) {
     $self->_set_err("failed to translate cdef schema into command",
                     $schema_obj->errmsg());
     return $SanUtil::FALSE;
 }

=back

=head2 B<get_req_field_check()>

Get the required field check flag

=over 4

=item Inputs

None

=item Outputs

Return the value of required field flag (TRUE | FALSE)

=item Examples

 # Get the current value of required field check
 #
 my $current_req_field = $schema_obj->get_req_field_check();

=back

=head2 B<get_cdef_objs()>

Get command definition objects

=over 4

=item Inputs

None

=item Outputs

Return array or arrayref of command definition objects

=item Examples

 # Get all available cdef objects
 #
 my @cdef_objs = $schema_obj->get_cdef_objs();
 if (scalar(@cdef_objs) == 0) {
     $self->_set_err("could not find cdef object");
     return $SanUtil::FALSE;
 }

=back

=head2 B<is_loaded(%args)>

Check whether cdef schema is loaded or not

=over 4

=item Inputs

=item * I<-filer> => <filer_name>

Name of filer or controller

=item Outputs

Return TRUE = 1 if schema is loaded; otherwise FALSE = 0

=item Examples

 # load the cdef schema if it hasn't been loaded
 #
 unless ($schema_obj->is_loaded(-filer => $filer_name)) {
     $schema_obj->load(-path   => $self->{cdef_path},
                       -filer  => $self->get_filer(),
                       -branch => $self->get_branch());
     if ($schema_obj->error()) {
         $self->_set_err("failed to load cdef schema", $schema_obj->errmsg());
         return $SanUtil::FALSE;
     }
 }

=back

=head2 B<load(%args)>

Load cdef schema

=over 4

=item Inputs

=item * I<-filer> => <$filer_name>

Name of filer or controller

=item * I<-path> => <cdef_path>

Path that contains cdef files

=item * I<-branch> => <build branch>

Name of the build branch (ei. mainng, smokejumper ...)

=item Outputs

Return TRUE = 1 if schema is loaded; otherwise FALSE = 0

=item Examples

 # load the cdef schema
 #
 $schema_obj->load(-path   => $self->{cdef_path},
                   -filer  => $self->get_filer(),
                   -branch => $self->get_branch());
 if ($schema_obj->error()) {
     $self->_set_err("failed to load cdef schema", $schema_obj->errmsg());
     return $SanUtil::FALSE;
 }

=back

=head2 B<translate(%args)>

Convert command definition into actual command

=over 4

=item Inputs

=item * I<-filer> => <$filer_name>

Name of filer or controller

=item * I<-cdef> => <command_definition>

Command definition (ei. volume_transition, lun_create, ...)

=item * I<%args>

Other attributes based on each command definition.
(See cdef files for additional attributes per command definition)

=item Outputs

Return formated command that is ready to be executed on the filer

=item Examples

 # Convert cdef into formated command that can be executed on a filer
 #
 my $cmd = $schema_obj->translate(-filer  => $self->get_filer(),
                                  -cdef   => $cdef,
                                  %args);
 if ($schema_obj->error()) {
     $self->_set_err("failed to translate cdef schema: " . $cdef,
                     $schema_obj->errmsg());
     return $SanUtil::FALSE;
 }

=cut

1;
