# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to get required and optional fields
##          from command definitions
##
## @author korry@netapp.com, dl-ssan-test
## @status Shared

package Cdef;

# compiler directives
use strict;

# imported modules
use Module::Common qw(validate_mixed);
use Params::Validate qw( SCALAR HASHREF ARRAYREF );
use Exporter;
use Data::Dumper;
use Error;
use San::SanLog;
use San::SanUtil;
use Tie::IxHash;

# constant variables
use constant REQ => 'required';
use constant OPT => 'optional';
use constant INV => 'invalid';

use vars qw(@EXPORT @ISA);

our @EXPORT = qw(REQ OPT INV);

@ISA = qw( Exporter Error SanLog SanUtil );

###############################################################################
# Method:    new()
# Objective: Create cdef object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {
        (%args) =
          Module::Common::validate_mixed(@input_params,
                         [],
                         {action          => {type => SCALAR,   required => 1},
                          required_params => {type => ARRAYREF, required => 1},
                          optional_params => {type => ARRAYREF, required => 1},
                          alias_params    => {type => HASHREF,  required => 1},
                          alias_map       => {type => HASHREF,  default => {}},
                          file            => {type => SCALAR,   required => 1}});
    };
    if ($@) {
        $self->_set_err("fail to validate parameters: " . $@);
        return $self;
    }

    # create 'get' accessor method for each attribute
    # set 'no strict' due to dereference which will be used to create
    # get accessor methods on the fly
    no strict 'refs';
    foreach my $field (keys %args) {

        # set public metadata fields (action, requireds, optionals, etc.)
        $self->{$field} = $args{$field};

        # create get accessor methods fore each field
        *{"get_" . lc $field} = sub {
            my ($self) = @_;
            return $self->{$field};
        };
    }
    use strict 'refs';

    return $self;

} ## end sub new

###############################################################################
# Method:    is_param_mapped()
# Objective: Determine if a passed in cmd input parameter (alias) is mapped to
#            SUT parameter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_param_mapped {
    my ($self, %args) = @_;

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    %args = &SanUtil::strip_dash(%args);

    return (exists $self->get_alias_params()->{$args{param}}) ? $SanUtil::TRUE : $SanUtil::FALSE;
}

###############################################################################
# Method:    _get_alias_root()
# Objective: Return system cmd parameter corresponding to common alias
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_alias_root {
    my ($self, $from) = @_;

    my ($alias);

    # cdef parameters has '-' so convert underscore '_' to dash '-'
    $from =~ s/_/-/g;

    # get alias from cdef
    $alias = $self->get_alias_params()->{$from};

    # in SN builds there is an alias map, if there is an alias map for this key,
    # obtain the translation here
    if (defined($self->get_alias_map()->{$from})) {
        $alias = "-" . $self->get_alias_map()->{$from};
    }

    # ensure alias is exist and matched in either the required or
    # optional fields
    unless (defined $alias
            and (grep { -$_ =~ /^$alias$/ } @{$self->get_required_params()}, @{$self->get_optional_params()})) {
        $self->_set_err("invalid command parameter '$from'");
        return "";
    }
    return $alias;
} ## end sub get_alias_root

###############################################################################
# Method:    _get_alias_type()
# Objective: Determine what type of input parameter alias is; required,
#            optional, or invalid
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_alias_type {
    my ($self, $alias) = @_;

    my ($param);

    # get the alias parameter
    $param = $self->_get_alias_root($alias);

    # determine the parameter type
    return $self->_get_param_type($param);
}

###############################################################################
# Method:    _get_param_type()
# Objective: Determine what type of input parameter; required,
#            optional, or invalid
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_param_type {
    my ($self, $param) = @_;

    # return parameter type
    return REQ if (grep { -$_ =~ /^$param$/ } @{$self->get_required_params()});
    return OPT if (grep { -$_ =~ /^$param$/ } @{$self->get_optional_params()});
    return INV;
}

###############################################################################
# Method:    translate()
# Objective: return cmd syntax for requested system operation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub translate {
    my ($self, %args);
    # preserve the order of input fields
    tie %args, 'Tie::IxHash';
    ($self, %args) = @_;

    my ($attribute, $cmd, $param, $value, @cmd_inputs);
    my (@invalids, @missings, @optionals, @params, @requireds, %type);
    my (@required_params, @optional_params);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse command attributes/parameters
    foreach my $key (keys %args) {

        # skip cdef and req_field_chk args
        next if ($key =~ /-?(cdef|req_field_chk)$/i);

        # if key is "fields", push onto @cmd_inputs
        if ($key =~ /fields/i) {
            # replace '_' with '-'
            $args{fields} =~ s/_/-/g;

            push @cmd_inputs, "-" . $key . " " . $args{$key};
            next;
        }

        # resolve param alias to proper param name. Store aliases that
        # aren't mapped to a proper param name.
        # (ex. volume will become -volume)
        $param = $self->_get_alias_root($key);
        if ($self->error() or $param eq "") {
            # store invalid key in an array
            push @invalids, $key;
        }

        if ($param) {

            # filter param name by types (required,optional,invalid)
            push @{$type{$self->_get_param_type($param)}}, $param;

            # rebuild the cmd line string with proper param names
            # add quote "" if value has space.
            $value =
              ($args{$key} =~ /\s/)
              ? "\"" . $args{$key} . "\""
              : $args{$key};

            push @cmd_inputs, $param . " " . $value;
        }
    }

    # return an error if an invalid field was provided
    if (scalar(@invalids)) {
        $self->_set_err("invalid command parameter ("
                        . join(', ', @invalids)
                        . ") for action '"
                        . $self->get_action()
                        . "'", $self->errmsg());
        return "";
    }

    # if required field check flag is set,
    # check to ensure all required fields were provided
    if ($args{req_field_chk} == $SanUtil::TRUE) {

        # ensure the provided required fields were provided
        # if not, push to missings array
        @required_params = @{$self->get_required_params()};
        foreach my $required (@required_params) {
            if (not scalar(grep { /^\-?$required$/i } @{$type{required}})) {
                push @missings, $required;
            }
        }

        # Return an error if a required field is missing
        if (@missings) {
            $self->_set_err("missing required parameter ("
                            . join(', ', @missings)
                            . ") for action '"
                            . $self->get_action()
                            . "'", $self->errmsg());
            return "";
        }

    }

    # return command and its parameters
    return $self->get_action() . " " . join (" ", @cmd_inputs);
} ## end sub translate

###############################################################################
#  P R I V A T E
###############################################################################

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Cdef.pm

=head1 Author

Korry Nguyen (dl-ssan-test)

=head1 Description

Cdef.pm provides an interface to get command definition and its required and optional fields.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm>

=head1 Attributes

=head2 B<action>

string = command definition (cdef) action

=head2 B<required_params>

arrayref = array of required parameters that pertain to the cdef action

=head2 B<optional_params>

arrayref = array of optional parameters that pertain to the cdef action

=head2 B<alias_params>

hashref = hash of alias parameters that mapped to additional options of the cdef action

=head2 B<file>

string = cdef file that contains data structures for cdef command

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a cdef object per cdef action.

=over 4

=item Inputs

=item * I<-action> => <command action>

Command action (ei. volume create, volume rename, lun create, ...)

=item * I<-required_params> => <arrayref of required options>

List of required options per cdef command

=item * I<-optional_params> => <arrayref of optional options>

List of optional options per cdef command

=item * I<-alias_params> => <hashref of options>

List of aliases that mapped to attributes of a particular cdef file

=item * I<-file> => <cdef file>

Cdef file that contained data structures which pertain to cdef action

=item Outputs

Return cdef object

=item Examples

 use San::Schema::Cdef;

 # Instantiate cdef object
 # NOTE: %CommandDefinitions was defined in the cdef $file
 #
 my $cdef_obj = Cdef->new(-action => $CommandDefinitions{$cdef}{command},
                          -required_params => $required_params,
                          -optional_params => $optional_params,
                          -alias_params => $CommandDefinitions{$cdef}{options},
                          -file => $file);
 if ($cdef_obj->error()) {
     $self->_set_err("fail to instantiate cdef object");
     return $SanUtil::FALSE;
 }

=back

=head2 B<get_action()>

Get command action

=over 4

=item Inputs

None

=item Outputs

Return the formated command
(ex. volume create)

=item Examples

 # get cdef command action
 #
 my $cmd = $cdef_obj->get_action();

=back

=head2 B<get_alias_params()>

Get an alias list of attributes from a particular cdef file

=over 4

=item Inputs

None

=item Outputs

Return a hashref list of attributes
(ex. dsid => "-dsid", volume-type => "-volume-type")

=item Examples

 # get alias params
 #
 my $alias_href = $cdef_obj->get_alias_params();

=back

=head2 B<get_optional_params()>

Get optional parameters

=over 4

=item Inputs

None

=item Outputs

Return an arrayref of optional parameters
(ex. [qw( size state type )])

=item Examples

 # get optional parameters
 #
 my $opt_aref = $cdef_obj->get_optional_params();

=back

=head2 B<get_required_params()>

Get required parameters

=over 4

=item Inputs

None

=item Outputs

Return an arrayref of required parameters
(ex. [qw( vserver aggregate volume )])

=item Examples

 # get required parameters
 #
 my $req_aref = $cdef_obj->get_required_params();

=back

=head2 B<get_file()>

Get cdef file name

=over 4

=item Inputs

None

=item Outputs

Return a string
(ex. volume)

=item Examples

 # get cdef file name
 #
 my $cdef_file = $cdef_obj->get_file();

=back

=head2 B<is_param_mapped(%args)>

Verify input parameter is mapped to ONTAP command parameter

=over 4

=item Inputs

=item * I<-param> => <optional | required parameter>

Command parameter

=item Outputs

Return TRUE | FALSE (1 | 0)

=item Examples

 # verify the required param (newname) for "volume rename" is mapped to ONTAP
 # command parameter
 #
 if (! $cdef_obj->is_param_mapped(-param => "newname") {
     $self->_set_err("\"newname\" param is not mapped to command in cdef file");
     return $SanUtil::FALSE;
 }

=back

=head2 B<translate(%args)>

Convert command definition into actual command

=over 4

=item Inputs

=item * I<-filer> => <$filer_name>

Name of filer or controller

=item * I<-cdef> => <command_definition>

Command definition (ei. volume_transition, lun_create, ...)

=item * I<%args>

Other attributes based on each command definition.
(See cdef files for additional attributes per command definition)

=item Outputs

Return formated command that is ready to be executed on the filer

=item Examples

 # Convert cdef into formated command that can be executed on a filer
 #
 my $cmd = $cdef_obj->translate(-cdef   => "volume_create",
                                -vserver => $vserver_name,
                                -aggregate => $aggr_name,
                                -volume => $volume_name,
                                -size => "100m");
 if ($cdef_obj->error()) {
     $self->_set_err("failed to translate cdef: volume_create",
                     $cdef_obj->errmsg());
     return $SanUtil::FALSE;
 }

=cut

