#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/San/ScsiTool.pm#5 $

## @summary Provide an interface to several client side SCSI tools
## @author bhanup@netapp.com
## @status public
## @pod here

package ScsiTool;

use strict;

# better to use integer so that bitmask manipulations work
# as expected in a two completement system.
use integer;

use Error;
use San::ScsiTool::CDB;
use San::SanLog;
use Data::Dumper;
use San::ScsiTool::Xcopy;

use vars qw( @ISA );

@ISA = qw(Error SanLog Xcopy);

#
# bit mask to indicate what the send_cmd api should
# return. Call expected_responses with these flags
# to set them for a particular transaction
#
use constant {

    # bitmasks to be used with the expected_responses api
    TOOLNAME      => 0x000001,    # Name of the tool used
    IOCTL_STATUS  => 0x000002,    # the Cmd Status
    SCSI_DATA     => 0x000004,    # Data returned for the Cmd in a hex array
    SCSI_STATUS   => 0x000008,    # SCSI Status code and string
    HOST_STATUS   => 0x000010,    # Host Status
    DRIVER_STATUS => 0x000020,    # Driver Status
    TOOL_LIB_STATUS =>
      0x000040,    # optional status to indicate tool capabilities/restrictions.
    RESIDUAL_BYTES => 0x000080,     # Residual Bytes
    BYTES_XFERD    => 0x000100,     # Bytes transferred
    SENSE_KEY      => 0x000200,     # SCSI Sense Key and string
    ASCNQ          => 0x000400,     # asc/ascq pair and string rep
    SENSE_DATA     => 0x000800,     # all sense data in a hex array
    RSP_REBUILD    => 0x001000,     # flag to retain these bit masks for
                                    # the life of the obj.
    RSP_ADD        => 0x002000,     # add this mask to the current mask
    RSP_DEL        => 0x004000,     # remove this mask from the current mask
    RESID_UNDER    => 0x008000,     # resid under flag value
    RESID_OVER     => 0x010000,     # resid over flag value
    MAX_RSP_INDEX  => 0x020000,     # if needed add more indexes into
                                    # the resp_hsh and set
    REQ_RESPONSE   => "required",

    # values of the 'op' field in the send_cmd api
    SCSI_CDB            => 0x000001,     # default value.
    ABORT_TASK          => 0x000002,
    ABORT_TASK_SET      => 0x000004,
    LUN_RESET           => 0x000008,
    CLEAR_ACA           => 0x000010,
    CLEAR_TASK_SET      => 0x000020,
    QUERY_TASK          => 0x000040,
    BUS_RESET           => 0x000080,     # invalid in SAM3r14
    TARGET_RESET        => 0x000100,     # invalid in SAM3r14
    NACA_BIT            => 0x000200,     # should set the NACA bit in the
                                         # cdb control byte.
    LINK_BIT            => 0x000400,     # should set the LINK bit in the
                                         # cdb control byte.
    LOGOUT              => 0x000800,     #this block is new TMF's
    LOGIN               => 0x001000,     #supported by any-iscsi
    TARGET_WARM_RESET   => 0x002000,
    TARGET_COLD_RESET   => 0x004000,
    TASK_REASSIGN       => 0x008000,
    UNREGISTER          => 0x010000,
    REGISTER            => 0x020000,
    QUERY_TASK_SET      => 0x040000,
    QUERY_ASYNC_EVENT   => 0x080000,

    # values of the 'qtag' field in the send_cmd api
    HEAD_OF_Q   => 0x000001,
    ORDERED_CMD => 0x000002,
    SIMPLE_CMD  => 0x000004,
    ACA_ATTR    => 0x000008,

    # values of the DIR field in the send_cmd api
    NONE  => "none",
    READ  => "read",
    WRITE => "write",

    # values of ON_ERROR field in the send_cmd api
    STOP     => 0x000000,
    CONTINUE => 0x000001,

    # values for the TOOL_LIB_STATUS
    TOOL_LIB_OK => 0x000000,    # if no issue encountered.
    OP_UNSUP    => 0x000001,    # if the tool does not support a opcode
    QTAG_UNSUP  => 0x000002,    # if the tool does not support a qtag
    DEV_UNSUP   => 0x000004,    # if the tool does not support a device name
    PAT_UNSUP   => 0x000008,    # if the tool does not support setting a pattern

    # values for response codes
    RESP_CODE_CURR_INFO_FIXED => 0x70,

    # values for sense keys
    SK_ILLEGAL_REQUEST => 0x05,
    SK_COPY_ABORTED    => 0x0a,

    # values for ASC(Q)
    ASC_INVALID_COMMAND_OP_CODE                   => 0x20,
    ASCQ_INVALID_COMMAND_OP_CODE                  => 0x00,
    ASC_3RD_PARTY_DEV_FAIL                        => 0x0D,
    ASCQ_3RD_PARTY_DEV_FAIL                       => 0x01,
    ASC_LBA_OUT_OF_RANGE                          => 0x21,
    ASCQ_LBA_OUT_OF_RANGE                         => 0x00,
    ASC_INVALID_FIELD_IN_PARAM                    => 0x26,
    ASCQ_INVALID_FIELD_IN_PARAM                   => 0x00,
    ASCQ_TOO_MANY_TARGET_DESCRIPTORS              => 0x06,
    ASCQ_UNSUPPORTED_TARGET_DESC_TYPE_CODE        => 0x07,
    ASCQ_TOO_MANY_SEGMENT_DESCRIPTORS             => 0x08,
    ASCQ_UNSUPPORTED_SEGMENT_DESC_TYPE_CODE       => 0x09,
    ASC_UNREACHABLE_COPY_TARGET                   => 0x08,
    ASCQ_UNREACHABLE_COPY_TARGET                  => 0x04,
    ASC_COPY_TARGET_DEVICE_NOT_REACHABLE          => 0x0D,
    ASCQ_COPY_TARGET_DEVICE_NOT_REACHABLE         => 0x02,
        
    CHECK_CONDITION => 2,
    HARDWARE_ERROR  => 4,
    DATA_PROTECT    => 7,
    UNIT_ATTENTION  => 6,
    NOT_READY       => 2,
    };

use constant {

    # default values of the other params used by send_cmd api
    DEF_DEV         => REQ_RESPONSE,
    DEF_PATH        => 0,
    DEF_LEN         => 512,
    DEF_CDB         => "00 00 00 00 00 00",
    DEF_CDB_PARAM   => 0,
    DEF_OP          => 1,                     # SCSI_CDB default value.
    DEF_OP_NAME     => 0,
    DEF_QTAG        => SIMPLE_CMD,            #SIMPLE_CMD default value
    DEF_CNT         => 1,                     # by default execute once
    DEF_RUNTIME     => 0,                     # if 0 then execute DEF_CNT times
    DEF_SLEEP       => 0,                     # sleep time in secs
    DEF_TMO         => 5,                     # timeout in secs
    DEF_DEBUG_LEVEL => 0,                     # info only on failures
    DEF_ONERR       => 0,                     # STOP default value
    DEF_DIR         => "read",                # READ default value
    DEF_DATA_IN     => "undef hash",
    DEF_DATA_OUT    => "undef hash",
    DEF_DATA_PAT    => "no pattern",
    DEF_DISABLED    => "disabled",
    DEF_EXPECT      => "undef hash",
    DEF_EXPECT_ASQ    => "undef hash",
    DEF_EXPECT_ASC    => "undef hash",
    DEF_EXPECT_SKEY   => "undef hash",
    DEF_EXPECT_STATUS => "undef hash",
    DEF_EXPECT_RESID  => "undef hash",
    };

my %scsiCdbs = (
    inquiry          => "12 00 00 00 80 00",
    inquiry_page_00h => "12 01 00 00 80 00",
    inquiry_page_80h => "12 01 80 00 80 00",
    inquiry_page_83h => "12 01 83 00 9c 00",
    inquiry_page_85h => "12 01 85 00 80 00",
    inquiry_page_87h => "12 01 87 10 00 00",
    inquiry_page_B0h => "12 01 B0 00 80 00",
    inquiry_page_B2h => "12 01 B2 00 80 00",
    inquiry_page_C0h => "12 01 C0 00 80 00",
    inquiry_page_C1h => "12 01 C1 00 80 00",
    inquiry_page_C2h => "12 01 C2 00 80 00",
    report_luns      => "A0 00 00 00 00 00 00 00 00 32 00 00",
    report_luns_1    => "A0 00 01 00 00 00 00 00 00 32 00 00",
    report_luns_2    => "A0 00 02 00 00 00 00 00 00 32 00 00",
    test_unit_ready  => "00 00 00 00 00 00",
    report_dev_id    => "A3 05 00 00 00 00 00 00 00 32 00 00",
    read_capacity_10 => "25 00 00 00 00 00 00 00 00 00",
    read_capacity_16 => "9E 10 00 00 00 00 00 00 00 00 00 00 00 32 00 00",

    # netapp cmd to get device id to support openVMS OS.
    get_dev_id                => "EC 00 00 00 00 00 00 00 32 00",
    report_target_port_groups => "A3 0A 00 00 00 00 00 00 00 64 00 00",
    report_supported_opcodes  => "A3 0C 00 00 00 00 00 00 00 64 00 00",
    receive_copy_results      => "84 03 00 00 00 00 00 00 00 00 00 00 00 80 00 00",
    read_6                    => "08 00 00 00 01 00",
    read_10                   => "28 00 00 00 00 00 00 00 64 00",
    mode_sense_6_alua         => "1A 00 0a 01 64 00",
    read_keys_8               => "5e 00 00 00 00 00 00 00 08 00",
    read_keys                 => "5e 00 00 00 00 00 00 00 80 00",
    request_sense             => "03 00 00 00 80 00",
    stop                      => "1B 00 00 00 00 00",
    start                     => "1B 00 00 00 01 00",
    write6                    => "0A 00 00 00 01 00",
    write_10                  => "2A 00 00 00 00 00 00 00 01 00",
    write_16                  => "8A 00 00 00 00 00 00 00 00 00 00 00 00 01 00 00",
    write_same_10             => "41 00 00 00 00 00 00 00 01 00",
    write_same_16             => "93 00 00 00 00 00 00 00 00 00 00 00 00 01 00 00",
    verify_10                 => "2F 00 00 00 00 00 00 00 01 00",
    verify_16                 => "8F 00 00 00 00 00 00 00 00 00 00 00 00 01 00 00",
    unmap_no_op               => "42 00 00 00 00 00 00 00 00 00",
    xcopy_no_op               => "83 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00",
    reserve_10                => "56 00 00 00 00 00 00 00 00 00",
    reserve_6                 => "16 00 00 00 00 00",
    release_10                => "57 00 00 00 00 00 00 00 00 00",
    release_6                 => "17 00 00 00 00 00",
    mode_sense_6              => "1A 08 3F 00 32 00",
    mode_sense_10             => "5A 00 00 00 00 00 00 00 32 00",
    mode_geometry_page        => "5A 00 04 00 00 00 00 02 00 00",

    # send diagonistic cdbs
    send_diag_self_test                  => "1D 04 00 00 00 00",
    send_diag_self_test_not_set          => "1D 00 00 00 00 00",
    send_diag_self_test_flags_set         => "1D 07 00 00 00 00",
    send_diag_self_test_not_set_flags_set => "1D 03 00 00 00 00",

    # PR cdbs
    scsi_pr_in_read_keys    => "5E 00 00 00 00 00 00 04 00 00",
    scsi_pr_in_read_resv    => "5E 01 00 00 00 00 00 04 00 00",
    scsi_pr_in_rep_cap      => "5E 02 00 00 00 00 00 04 00 00",
    scsi_pr_in_read_ful_sts => "5E 03 00 00 00 00 00 04 00 00",

    # Command to get the vdisk attributes
    gva_virtual_disk_path_page     => "C0 00 00 0A 98 0A 10 00 64 00",
    gva_vdisk_system_info_page     => "C0 00 00 0A 98 0A 22 00 64 00",
    gva_vdisk_lun_ostype_page      => "C0 00 00 0A 98 0A 18 00 64 00",
    gva_vdisk_vol_page             => "C0 00 00 0A 98 0A 11 00 64 00",
    gva_vdisk_vol_rev_1            => "C0 00 01 0A 98 0A 11 00 64 00",
    gva_vdisk_mapped_luns_page     => "C0 00 00 0A 98 0A 16 00 64 00",
    gva_vdisk_lun_share_page       => "C0 00 00 0A 98 0A 17 00 64 00",
    gva_filername_page             => "C0 00 02 0A 98 0A 00 00 64 00",
    gva_32_bit_ip_addr             => "C0 00 00 0A 98 0A 01 00 64 00",
    gva_64_bit_ip_addr             => "C0 00 00 0A 98 0A 02 00 64 00",
    gva_vdisk_qtree_page           => "C0 00 00 0A 98 0A 12 00 64 00",
    gva_vdisk_snapshot_page        => "C0 00 00 0A 98 0A 13 00 64 00",
    gva_vdisk_status_page          => "C0 00 00 0A 98 0A 14 00 64 00",
    gva_vdisk_stream_sizes_page    => "C0 00 00 0A 98 0A 15 00 28 00",
    gva_vdisk_fcp_config_page      => "C0 00 00 0A 98 0A 20 00 64 00",
    gva_vdisk_fcp_adapter_page     => "C0 00 00 0A 98 0A 21 00 64 00",
    gva_vdisk_asl_info_page        => "C0 00 00 0A 98 0A 23 00 FF 00",
    gva_vdisk_fcp_adap_det_page    => "C0 00 00 0A 98 0A 24 00 FF 00",
    gva_vdisk_logged_init_page       => "C0 00 00 0A 98 0A 25 00 64 00",
    gva_vdisk_igroup_info_page     => "C0 00 00 0A 98 0A 26 00 64 00",
    gva_vdisk_cluster_info_page    => "C0 00 00 0A 98 0A 27 00 64 00",
    gva_vdisk_lun_select_info_page => "C0 00 00 0A 98 0A 32 00 64 00",
    gva_iscsi_config_info          => "C0 00 01 0A 98 0A 40 00 FF 00",
    gva_iscsi_adapter_info         => "C0 00 00 0A 98 0A 41 00 64 00",
    gva_super_page_info            => "C0 00 00 0A 98 0A 00 FF FF 00",
    gva_all_pages                  => "C0 00 00 0A 98 0A FF FF FF 00",
);

my %scsi_status = (0x00 => "GOOD",
                   0x02 => "CHECK CONDITION",
                   0x04 => "CONDITION MET",
                   0x08 => "BUSY",
                   0x10 => "Obsolete",
                   0x14 => "Obsolete",
                   0x18 => "RESERVATION CONFLICT",
                   0x22 => "Obsolete",
                   0x28 => "TASK SET FULL",
                   0x30 => "ACA ACTIVE",
                   0x40 => "TASK ABORTED",);

my %sense_key = (0x00 => "NO SENSE",
                 0x01 => "RECOVERED ERROR",
                 0x02 => "NOT READY",
                 0x03 => "MEDIUM ERROR",
                 0x04 => "HARDWARE ERROR",
                 0x05 => "ILLEGAL REQUEST",
                 0x06 => "UNIT ATTENTION",
                 0x07 => "DATA PROTECT",
                 0x08 => "BLANK CHECK",
                 0x09 => "VENDOR SPECIFIC",
                 0x0A => "COPY ABORTED",
                 0x0B => "ABORTED COMMAND",
                 0x0C => "Obselete",
                 0x0D => "VOLUME OVERFLOW",
                 0x0E => "MISCOMPARE",
                 0x0F => "Reserved",);

my $expected_response = undef;

########################################################################
# P U B L I C    M E T H O D S
########################################################################

########################################################################
# Method:    create_scsi_tool
# Objective: Class method that will return correct type of ScsiTool
#                    based on the host OS and/or tool requested.
# Details:   See POD documentation at the end of this file.
# Notes:     Tools currently supported: SPT, USCSI, PTI, IFC5d, SANBlaze,
#                                       any-iscsi, scsitsim
########################################################################
sub create_scsi_tool {
    my ($caller, %args) = @_;
    my $class = ref($caller) || $caller;

    #
    # HACK: inorder for us to return an error code that is
    # consistent with San libs, we have to instantiate
    # an object of the Class ScsiTool and then set the error
    # string on it and return it if needed.
    # If there are no errors then an object of the sub-class of ScsiTool
    # will be returned.
    # Now what should happen to the st_class_obj? Garbage collection
    # in perl should take care of it.
    # San libs should address this by providing an error obj that can be
    # returned instead.
    #
    my $st_class_obj = {};
    bless $st_class_obj, $class;

    my ($tool, $tool_obj_ref);

    my $client_obj;
    my $debug_level;
    my $scsi_tool_loc;

    #
    # Parse the arguments sent in the args hash
    #
    foreach my $param (keys %args) {
        if ($param =~ /^-?tool$/) {
            $tool = $args{$param};
            if (not length($tool) or $tool !~ /spt|uscsi|sanblaze|scsit_sim|any_iscsi/i) {
                $st_class_obj->_set_err("Error: ScsiTool::create_scsi_tool():  $param must be " .
                                "one of spt, uscsi, pti, ifc5d, sanblaze, scsit_sim, or any_iscsi"     .
                                "not: '$tool'");
                return $st_class_obj;
            }
        } elsif ($param =~ /^-?client$/) {
            $client_obj = $args{$param};

            #
            # Make sure the client we were given is a SanClient object
            #
            unless (ref($client_obj) =~ "SanClient") {
                $st_class_obj->_set_err("Error: ScsiTool::create_scsi_tool(): "
                                        . "Invalid value for '-client', a"
                                        . "SanClient object must be passed");
                return $st_class_obj;
            }
        } elsif ($param =~ /^-?debug_level$/) {
            $debug_level = $args{$param};
            unless (length($debug_level)) {
                $st_class_obj->_set_err("Error: ScsiTool::create_scsi_tool(): "
                                        . "Invalid value for '-debug_level'");
                return $st_class_obj;
            }
        } elsif ($param =~ /^-?scsi_tool_loc/) {
            $scsi_tool_loc = $args{$param};
            unless (length($scsi_tool_loc)) {
                $st_class_obj->_set_err("Error: ScsiTool::create_scsi_tool(): "
                                        . "Invalid value for '-scsi_tool_loc'");
                return $st_class_obj;
            }
        } else {

            #
            # We got an invalid argument, fail
            #
            $st_class_obj->_set_err(  "Error: ScsiTool::create_scsi_tool(): "
                                    . "Invalid argument '$param'");
            return $st_class_obj;
        }
    }    # end foreach argument

    #
    # Check for required arguments
    #
    unless (defined $client_obj) {
        $st_class_obj->_set_err("Error: ScsiTool::create_scsi_tool()",
                                "Missing required argument '-client'");
        return $st_class_obj;
    }

    #
    # If we were not given a debug level, set our debug level
    # to the client's debug level
    #
    $debug_level = $client_obj->get_debug_level() unless (defined $debug_level);

    $args{-debug_level} = $debug_level;

    #
    # If scsi tool location is defined, pass it as an argument to the
    # constructor of the tool.  Currently, only spt will support this arg.
    #
    if (defined $scsi_tool_loc) {
        $args{-scsi_tool_loc} = $scsi_tool_loc;
    }

    #
    # Check to make sure we are on a supported hosttype and set the
    # tool to a proper default if it's not already set
    #
    my $hosttype = $client_obj->hosttype();

    # unless tool is defined dont mess with tool
    unless (defined($tool)) {
        $tool = "spt"      if ($hosttype =~ /windows|[li|u]nix|vmware/);
        $tool = "sanblaze" if ($hosttype =~ /sanblaze/);
        $tool = "any_iscsi" if ($hosttype =~/linux/);
        #
        # update the args param for tool with the tool being used.
        #
        $args{-tool} = $tool;
    }

    unless ($hosttype =~ /[li|u]nix|windows|sanblaze|vmware/) {
        $st_class_obj->_set_err(
                  "Error: ScsiTool::create_scsi_tool() The '$hosttype' platform"
                    . "is not a currently supported by the ScsiTool object");
        return $st_class_obj;
    }

    #
    # load the appropriate ScsiTool module for the given tool
    #
    if ($tool =~ /spt/i) {
        require San::ScsiTool::Spt;
        $tool_obj_ref = San::ScsiTool::Spt->_new(%args);
    } elsif ($tool =~ /uscsi/i) {
        require San::ScsiTool::Uscsi;
        $tool_obj_ref = San::ScsiTool::Uscsi->_new(%args);
    } elsif ($tool =~ /sanblaze/i) {
        require San::ScsiTool::SanBlaze;
        $tool_obj_ref = San::ScsiTool::SanBlaze->_new(%args);
    } elsif ($tool =~ /scsit_sim/i) {
        require San::ScsiTool::ScsitSimCli;
        $tool_obj_ref = San::ScsiTool::ScsitSimCli->_new(%args);
    } elsif ($tool =~ /any_iscsi/i) {
        require San::ScsiTool::AnyIscsi;
        $tool_obj_ref = San::ScsiTool::AnyIscsi->_new(%args);
    } else {
        $st_class_obj->_set_err(
                         "Error: ScsiTool::create_scsi_tool() The '$tool' tool "
                           . "is not currently supported by the ScsiTool Libs");
        return $st_class_obj;
    }

    $tool_obj_ref->sanlog(-msg => "ScsiTool::create_scsi_tool(): \n"
                            . " ScsiTool object created using the "
                            . "$tool tool on a $hosttype machine",
                          -msg_level => 3);

    return $tool_obj_ref;
} ## end sub create_scsi_tool

########################################################################
# Method:    expected_responses
# Objective: Provide a way for the testcases to set the
#         responses that they wish to get after the send_cmd call.
# Details:   See POD documentation at the end of this file.
########################################################################

sub expected_responses {
    my ($self, %args) = @_;
    my ($rsp_bitmask, $bitmask_option);

    $self->_clear_err();
    $expected_response = 1;

    #
    # Parse the arguments sent in the args hash
    #
    foreach my $param (keys %args) {
        if ($param =~ /^-?exp_rsp_bitmask$/) {
            $rsp_bitmask = $args{$param};
            unless (length($rsp_bitmask)) {
                $self->_set_err("Error: "
                            . ref($self)
                            . "::expected_responses(): "
                            . "Invalid value passed for the '-exp_rsp_bitmask' "
                            . "argument: $param");
                return $self;
            }
        } elsif ($param =~ /^-?exp_rsp_option$/) {
            $bitmask_option = $args{$param};
            unless (length($bitmask_option)) {
                $self->_set_err("Error: "
                             . ref($self)
                             . "::expected_responses(): "
                             . "Invalid value passed for the '-exp_rsp_option' "
                             . "argument: $param");
                return $self;
            }
        } else {

            #
            # We got an invalid argument, fail
            #
            $self->_set_err(  "Error: "
                            . ref($self)
                            . "::expected_responses(): "
                            . "Invalid argument '$param'");
            return $self;
        } ## end else [ if ($param =~ /^-?exp_rsp_bitmask$/)
    } ## end foreach my $param (keys %args)

    if ($bitmask_option == RSP_ADD) {
        $self->{EXP_RSP_MASK} |= $rsp_bitmask;
    } elsif ($bitmask_option == RSP_DEL) {
        $self->{EXP_RSP_MASK} &= (~$rsp_bitmask);
    } elsif ($bitmask_option == RSP_REBUILD) {
        $self->{EXP_RSP_MASK} = $rsp_bitmask
          if ($self->{EXP_RSP_MASK} != $rsp_bitmask);
    } else {
        $self->_set_err(  "Error: "
                        . ref($self)
                        . "::expected_responses(): "
                        . "Invalid value for rsp bitmask: $bitmask_option");
        return $self;
    }

    # if here then the just set it to RSP_REBUILD. This way we can reuse code in
    # the build_exp_rsp_hash()
    $self->{RSP_OPTION} = RSP_REBUILD;

    $self->sanlog(  -msg => ref($self)
                  . "::expected_responses() : "
                  . "using option: $self->{RSP_OPTION} and "
                  . "rsp_mask: $self->{EXP_RSP_MASK}");

    #
    # now that the bitmask is set the expected responses hash keys
    # for the sub-classes need to be filled in the send_cmd api
    #

    $self->build_exp_rsp_hash();

    return $self;

} ## end sub expected_responses

########################################################################
# Subroutine Name: send_cmd()
# Objective:       send_cmd method has to be overriden in all tools that subclass
#                  from this base class.
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub send_cmd {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err(  "Error: "
                    . ref($self)
                    . "::send_cmd(): "
                    . "Base class stub was called, when it should "
                    . "have been overridden in the sub class");
    return $self;

} ## end sub send_cmd

########################################################################
# Method:    create_cdb
# Objective: Create a CDB data expected by the send_cmd method
# Details:   See POD documentation at the end of this file.
# Notes:     Example of something you might create that is not
#            tool specific but that a user might call directly
########################################################################
sub create_cdb {
    my ($self, %args) = @_;
    my %cdb         = {};
    my $command     = "scsi_read_6";
    my $buffer_size = 1024;

    $self->_clear_err();

    #
    # Parse the arguments sent in the args hash
    #
    foreach my $param (keys %args) {
        if ($param =~ /^-?command$/) {
            $command = $args{$param};
            unless (length($command)) {
                $self->_set_err(  "Error: ScsiTool::create_cdb(): "
                                . "Invalid value passed for the '-command' "
                                . "argument: $param");
                return $self;
            }
        } elsif ($param =~ /^-?buffer_size$/) {
            $buffer_size = $args{$param};
            unless (length($buffer_size)) {
                $self->_set_err(  "Error: ScsiTool::create_cdb(): "
                                . "Invalid value passed for the '-buffer_size' "
                                . "argument: $param");
                return $self;

            }
        } else {

            #
            # We got an invalid argument, fail
            #
            $self->_set_err(
                      "Error: ScsiTool::new(): " . "Invalid argument '$param'");
            return $self;
        }
    }    # end foreach argument

    #
    # Store the cdb's data
    #
    $cdb{command}     = $command;
    $cdb{buffer_size} = $buffer_size;

    return %cdb;
}    # end create_cdb()

########################################################################
# Subroutine Name: build_exp_rsp_hash()
# Objective:       Take the exp resp bitmask and fill in the keys
#                  of the response hash that sub-class have to address
#                  This should be done when the constructor is executed first.
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub build_exp_rsp_hash {
    my $self        = shift @_;
    my $rsp_bitmask = $self->{EXP_RSP_MASK};
    my $option      = $self->{RSP_OPTION};

    # first get the exp_rsp_hash ref
    my $rsp_ref = $self->get_exp_rsp_params();

    unless ($rsp_ref) {
        $self->_set_err(  "Error: "
                        . ref($self)
                        . "::build_exp_rsp_hash(): responses "
                        . "hash is not defined");
        return $self;
    }

    unless ($rsp_bitmask) {

        $self->_set_err(  "Error: "
                        . ref($self)
                        . "::build_exp_rsp_hash(): expected "
                        . "responses bit mask is not defined");
        return $self;
    }

    #
    # if the response hash is to be rebuilt, then discard the orignal
    # hash and create a new hash. Once the new hash is created,
    #  assign it back to ER_PARAMS
    #
    $rsp_ref = {} if ($option == RSP_REBUILD);

    #
    # (re)contruct the responses hash as set by the
    # responses bitmask
    #
    $rsp_ref->{TOOLNAME}      = REQ_RESPONSE if ($rsp_bitmask & TOOLNAME);
    $rsp_ref->{IOCTL_STATUS}  = REQ_RESPONSE if ($rsp_bitmask & IOCTL_STATUS);
    $rsp_ref->{SCSI_DATA}     = REQ_RESPONSE if ($rsp_bitmask & SCSI_DATA);
    $rsp_ref->{SCSI_STATUS}   = REQ_RESPONSE if ($rsp_bitmask & SCSI_STATUS);
    $rsp_ref->{HOST_STATUS}   = REQ_RESPONSE if ($rsp_bitmask & HOST_STATUS);
    $rsp_ref->{DRIVER_STATUS} = REQ_RESPONSE if ($rsp_bitmask & DRIVER_STATUS);
    $rsp_ref->{TOOL_LIB_STATUS} = REQ_RESPONSE
      if ($rsp_bitmask & TOOL_LIB_STATUS);
    $rsp_ref->{RESIDUAL_BYTES} = REQ_RESPONSE
      if ($rsp_bitmask & RESIDUAL_BYTES);
    $rsp_ref->{RESID_OVER}  = REQ_RESPONSE if ($rsp_bitmask & RESID_OVER);
    $rsp_ref->{RESID_UNDER} = REQ_RESPONSE if ($rsp_bitmask & RESID_UNDER);
    $rsp_ref->{BYTES_XFERD} = REQ_RESPONSE if ($rsp_bitmask & BYTES_XFERD);
    $rsp_ref->{SENSE_KEY}   = REQ_RESPONSE if ($rsp_bitmask & SENSE_KEY);
    $rsp_ref->{ASCNQ}       = REQ_RESPONSE if ($rsp_bitmask & ASCNQ);
    $rsp_ref->{SENSE_DATA}  = REQ_RESPONSE if ($rsp_bitmask & SENSE_DATA);

    #  should we add OTHER?

    $self->{ER_PARAMS} = $rsp_ref if ($option == RSP_REBUILD);

    return $self;
} ## end sub build_exp_rsp_hash

########################################################################
# Subroutine Name: handle_unsup_responses()
# Objective:       Undef all the responses that the testcase wanted
#                  to hear about but the tool cannot support or has choosen
#                  not to update as of now.
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub handle_unsup_responses {
    my $self     = shift @_;
    my $rsp_hash = $self->get_exp_rsp_params();

    while (my ($str, $val) = each(%$rsp_hash)) {
        $rsp_hash->{$str} = undef if ($val eq REQ_RESPONSE);
    }
}

########################################################################
# Subroutine Name: get_send_cmd_params()
# Objective:       Set the defaults for all the params used by HASTA base class
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub get_send_cmd_params {
    my $self = shift @_;
    return $self->{SC_PARAMS};
}

########################################################################
# Subroutine Name: get_exp_rsp_params()
# Objective:       get the params that indicate what particular subclass
#                  is expected to return
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub get_exp_rsp_params {
    my $self = shift @_;
    return $self->{ER_PARAMS};
}

########################################################################
# Subroutine Name: get_def_exp_rsp_bitmask()
# Objective:       get the params that indicate what particular subclass
#                  is expected to return
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub get_def_exp_rsp_bitmask {
    my $self = shift @_;
    return $self->{DEF_EXP_RSP_MASK};
}

########################################################################
# Subroutine Name:  handle_send_cmd_params()
# Objective:        Provide a common interface to handle all the params
#                   that a ScsiTool type object/sub-class should handle.
# Details:          See POD documentation in ScsiTool.pm
# Note:             Each of the sub-class constructors can call this method to
#                   Set defaults and then they can override the required params.
########################################################################
sub handle_send_cmd_params {

    my ($self, %args) = @_;

    # variable for the multiple expect result in spt tool, array index
    my $exp_count=0;

    $self->_clear_err();

    $self->_set_def_params();

    #
    # Parse the arguments sent in the args hash
    # base class will check that the required params are
    # set and that the optional params get set to their
    # defaults
    #
    foreach my $param (keys %args) {
        if ($param =~ /^-?device$/) {
            if (not length($args{$param}) or
            $args{$param} !~ /dev|physicaldrive|target\d+lun\d+|\w+\/\d+\/\d+\:\d+|\d+\.\d+\.\d+\.\d+\:\d+/i) {
                $self->_set_err("Error: ".ref($self)."::handle_send_cmd_params():  $param must be "      .
                                "set to either a system device path, Physicaldrive, ITL, or SanBlaze lun. "   .
                                "not: '$args{$param}'");
                return $self;
            }
            $self->get_send_cmd_params()->{DEVICE} = $args{$param};
        } elsif ($param =~ /^-?mpio_path$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-mpio_path'");
                return $self;
            }
            $self->get_send_cmd_params()->{MPIO_PATH} = $args{$param};

        } elsif ($param =~ /^-?cdb$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-cdb'");
                return $self;
            }

            my $cdb_str = $self->get_cdb_bytes($args{$param});
            if ($self->error) {
                my $err_msg = join "\n", $self->errmsg();
                $self->_set_err(  "Error"
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . $err_msg);
                return $self;
            }

            $self->get_send_cmd_params()->{CDB} = $cdb_str;
        } elsif ($param =~ /^-?cdb_param$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-cdb_param'");
                return $self;
            }
            $self->get_send_cmd_params()->{CDB_PARAM} = $args{$param};
        } elsif ($param =~ /^-?buf_len$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-buf_len'");
                return $self;
            }

            $self->get_send_cmd_params()->{BUF_LEN} = $args{$param};
        } elsif ($param =~ /^-?opcode$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-opcode'");
                return $self;
            }
            $self->get_send_cmd_params()->{OPCODE} = $args{$param};
        } elsif ($param =~ /^-?qtag$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-qtag'");
                return $self;
            }
            $self->get_send_cmd_params()->{QTAG} = $args{$param};
        } elsif ($param =~ /^-?opcode_name$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-opcode_name'");
                return $self;
            }
            $self->get_send_cmd_params()->{OPCODE_NAME} = $args{$param};
        } elsif ($param =~ /^-?repeat_cnt$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-repeat_cnt'");
                return $self;
            }
            $self->get_send_cmd_params()->{REPEAT_CNT} = $args{$param};
        } elsif ($param =~ /^-?runtime$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-runtime'");
                return $self;
            }
            $self->get_send_cmd_params()->{RUNTIME} = $args{$param};
        } elsif ($param =~ /^-?sleep$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-sleep'");
                return $self;
            }
            $self->get_send_cmd_params()->{SLEEP} = $args{$param};
        } elsif ($param =~ /^-?timeout$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-timeout'");
                return $self;
            }
            $self->get_send_cmd_params()->{TIMEOUT} = $args{$param};
        } elsif ($param =~ /^-?on_error$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-on_error'");
                return $self;
            }
            $self->get_send_cmd_params()->{ON_ERROR} = $args{$param};
        } elsif ($param =~ /^-?dir$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-dir'");
                return $self;
            }
            $self->get_send_cmd_params()->{DIR} = $args{$param};
        } elsif ($param =~ /^-?data_in$/) {

            # both data_in and data_out are defined as hash
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-data_in'");
                return $self;
            }
            $self->get_send_cmd_params()->{DATA_IN} = $args{$param};
        } elsif ($param =~ /^-?data_out$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "Invalid value for '-data_out'");
                return $self;
            }
            $self->get_send_cmd_params()->{DATA_OUT} = $args{$param};
        } elsif ($param =~ /^-?data_pat$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "'-data_pat' needs a valid value");
                return $self;
            }
            $self->get_send_cmd_params()->{DATA_PAT} = $args{$param};
        } elsif ($param =~ /^-?enable$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "'-enable' needs a valid value");
                return $self;
            }
            $self->get_send_cmd_params()->{ENABLE} = $args{$param};
        # Spt version 1 and above supports these functions
        } elsif ($param =~ /^-?expect[0-9]$/ or $param =~ /^-?exp[0-9]$/
            or $param =~ /^-?expect$/ or $param =~ /^-?exp$/) { # new changes
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "'-expect' needs a valid value");
                return $self;
            }
            $self->get_send_cmd_params()->{EXPECT}[$exp_count] = $args{$param};
            $exp_count++;
        } elsif ($param =~ /^-?exp_skey$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "'-exp_skey' needs a valid value");
                return $self;
            }
            $self->get_send_cmd_params()->{EXP_SKEY} = $args{$param};
        } elsif ($param =~ /^-?exp_asc$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "'-exp_asc' needs a valid value");
                return $self;
            }
            $self->get_send_cmd_params()->{EXP_ASC} = $args{$param};
        } elsif ($param =~ /^-?exp_asq$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "'-exp_asq' needs a valid value");
                return $self;
            }
            $self->get_send_cmd_params()->{EXP_ASQ} = $args{$param};
        } elsif ($param =~ /^-?exp_resid$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "'-exp_resid' needs a valid value");
                return $self;
            }
            $self->get_send_cmd_params()->{EXP_RESID} = $args{$param};
        } elsif ($param =~ /^-?exp_status$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "'-exp_status' needs a valid value");
                return $self;
            }
            $self->get_send_cmd_params()->{EXP_STATUS} = $args{$param};
        } elsif ($param =~ /^-?dlimit$/) {
            unless (length $args{$param}) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::handle_send_cmd_params(): "
                                . "'-dlimit' needs a valid value");
                return $self;
            }
            $self->get_send_cmd_params()->{DLIMIT} = $args{$param};
        }else {

            #
            # We got an invalid argument, fail
            #
            $self->_set_err(  "Error: "
                            . ref($self)
                            . "::handle_send_cmd_params(): "
                            . "Invalid argument '$param'");
            return $self;
        } ## end else [ if ($param =~ /^-?device$/)
    }    # end foreach argument

    #
    # Check for required arguments
    #
    if ($self->get_send_cmd_params()->{DEVICE} eq DEF_DEV) {
        $self->_set_err("Error: " . ref($self) . "::handle_send_cmd_params()",
                        "Missing required argument '-device'");
        return $self;
    }

    return 1;
}    # end handle_send_cmd_params()


########################################################################
# Subroutine Name: batch_cmd()
# Objective:       send_cmd method has to be overriden in all tools that subclass
#                  from this base class.
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub batch_cmd {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Base class stub was called, when it should "
                    . "have been overridden in the sub class");
    return $self;

} ## end sub batch_cmd

########################################################################
# Start of Canned subs for ease of use
########################################################################

########################################################################
# Method:    critical_cmds
# Objective: Execute SCSI CDBs that will never fail due to target errors.
# Examples:  See POD documentation at the end of this file.
########################################################################

sub critical_cmds {

    my ($self, %args) = @_;

    my ($dev_name, $buf_len);

    $self->_clear_err();

    #
    # Parse the arguments sent in the args hash
    #
    foreach my $param (keys %args) {
        if ($param =~ /^-?device$/) {
            $dev_name = $args{$param};
            unless (length($dev_name)) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::inquiry(): "
                                . "Invalid value passed for the '-device' "
                                . "argument: $param");
                return 0;
            }
        } else {

            #
            # We got an invalid argument, fail
            #
            $self->_set_err(  "Error: "
                            . ref($self)
                            . "::inquiry(): "
                            . "Invalid argument '$param'");
            return 0;
        } ## end else [ if ($param =~ /^-?device$/)
    } ## end foreach my $param (keys %args)

    $buf_len = 255 unless ($buf_len);

    unless ($dev_name) {
        $self->_set_err(  "Error: "
                        . ref($self)
                        . "::inquiry(): "
                        . "Device name has be to specified");
        return 0;
    }

    $self->test_scsi_cmd(-device   => $dev_name,
                         -cdb_name => "inquiry",
                         -buf_len  => $buf_len);

    return 0 if ($self->error());

    $self->test_scsi_cmd(-device   => $dev_name,
                         -cdb_name => "report_luns",
                         -buf_len  => $buf_len);

    # it is assumed that the calling func will check for errors
    return $self->error() ? 0 : 1;
} ## end sub critical_cmds

########################################################################
# Method:    inquiry
# Objective: Execute the INQUIRY SCSI CDB.
# Details:   If the SCSI cmd executes then the return values will be
#            true/false w/ a hash reference to the SCSI data from the
#            execution ($ClientUtil::TRUE/FALSE, $ret_ref).
#            If there was an error encountered by ScsiTool obj then
#            a ($ClientUtil::FALSE, undef, $self) list is returned.
# Examples:  See POD documentation at the end of this file.
########################################################################

sub inquiry {

    my ($self, %args) = @_;

    my ($dev_name, $buf_len);

    $self->_clear_err();

    #
    # Parse the arguments sent in the args hash
    #
    foreach my $param (keys %args) {
        if ($param =~ /^-?device$/) {
            $dev_name = $args{$param};
            unless (length($dev_name)) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::inquiry(): "
                                . "Invalid value passed for the '-device' "
                                . "argument: $param");
                return $self;
            }
        } elsif ($param =~ /^-?buf_len$/) {
            $buf_len = $args{$param};
            unless (length($buf_len)) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::inquiry(): "
                                . "Invalid value passed for the 'buf_len' "
                                . "argument: $param");
                return $self;
            }
        } else {

            #
            # We got an invalid argument, fail
            #
            $self->_set_err(  "Error: "
                            . ref($self)
                            . "::inquiry(): "
                            . "Invalid argument '$param'");
            return $self;
        } ## end else [ if ($param =~ /^-?device$/)
    } ## end foreach my $param (keys %args)

    $buf_len = 128 unless ($buf_len);

    unless ($dev_name) {
        $self->_set_err(  "Error: "
                        . ref($self)
                        . "::inquiry(): "
                        . "Device name has be to specified");
        return;
    }

    return
      $self->test_scsi_cmd(-device   => $dev_name,
                           -cdb_name => "inquiry",
                           -buf_len  => $buf_len);

} ## end sub inquiry

########################################################################
# Sub Name:  get_cdb_bytes()
# Objective: Given a CDB Name find the string of hex bytes that can
#            used in the send_cmd interface.
# Details:   See POD documentation in ScsiTool.pm
#            If the cdb name is a series of 6 hex bytes then just
#            use that one.But else first make sure that there are no
#            3 digits in the string.
########################################################################

sub get_cdb_bytes {

    my ($self, $cmd_name) = @_;

    $self->_clear_err();

    # check if there are any 3 digit decimal values in the cdb string
    # all of the rest will be prefixed with 0x
    if ($cmd_name =~ /\w{3,}?\s/) {
        $self->_set_err(  "Error: "
                        . ref($self)
                        . "::get_cdb_bytes(): "
                        . "Invalid hex byte in: "
                        . $cmd_name);
        return;

    } elsif (   ($cmd_name !~ /^(\w\w\s){5,}?/)
             && ($cmd_name =~ /^\w+$/)) {

        # if we are not dealing with a string of >5 hex bytes, but we are
        # dealing with a name string then look up the name in the cdb
        # hash to figure the cdb bytes.
        my $hash_ref = $self->{cdb_name_hash};
        unless ($hash_ref) {
            $self->_set_err(  "Error: "
                            . ref($self)
                            . "::get_cdb_bytes(): "
                            . "Undefined Hash CDB name to string");
            return;
        }

        foreach my $k (keys %{$hash_ref}) {
            return $hash_ref->{$k} if ($k =~ /\b$cmd_name\b/i);
        }

        $self->_set_err(  "Error: "
                        . ref($self)
                        . "::get_cdb_bytes(): "
                        . "cdb string not found for: "
                        . $cmd_name);
        return;

    } ## end elsif (($cmd_name !~ /^(\w\w\s){5,}?/...

    # if we are here then we are dealing with a properly
    # formatted CDB string (hh hh hh .......). So just
    # return the string so that handle_send_cmd_params
    # can store it internally.

    return $cmd_name;

} ## end sub get_cdb_bytes

########################################################################
# Sub Name:  test_scsi_cmd()
# Objective: A generic interface to test SCSI Cmds by using the name of
#            the SCSI CDB i.e, report_luns, inquiry, read6 etc.
# Details:   See POD documentation in ScsiTool.pm
########################################################################

sub test_scsi_cmd {

    my ($self, %args) = @_;

    my ($dev_name, $cdb_name, $buf_len, $s_status, $s_k);

    my ($cdb_param, $opcode, $flag, $ret_ref, $expected_result, $ascnq, $sense_key, $dir);

    # Assign Default expected result
    $expected_result = "GOOD";
    $cdb_param       = 0;
    $opcode          = ScsiTool::SCSI_CDB;
    $dir             = ScsiTool::READ;
    $ascnq           = 0;
    $sense_key       = 0;

    $self->_clear_err();

    #
    # Parse the arguments sent in the args hash
    #
    foreach my $param (keys %args) {
        if ($param =~ /^-?device$/) {
            $dev_name = $args{$param};
            unless (length($dev_name)) {
                $self->_set_err(  "Invalid value passed for the '-device' "
                                . "argument: $param");
                return 0;
            }
        } elsif ($param =~ /^-?cdb_name$/) {
            $cdb_name = $args{$param};
            unless (length($cdb_name)) {
                $self->_set_err(  "Invalid value passed for the '-cdb_name' "
                                . "argument: $param");
                return 0;
            }
        } elsif ($param =~ /^-?buf_len$/) {
            $buf_len = $args{$param};
            unless (length($buf_len)) {
                $self->_set_err(  "Invalid value passed for the '-buf_len' "
                                . "argument: $param");
                return 0;
            }
        } elsif ($param =~ /^-?scsi_status$/) {
            $s_status = $args{$param};
            unless (length($s_status)) {
                $self->_set_err(  "Invalid value passed for the '-scsi_status' "
                                . "argument: $param");
                return 0;
            }
        } elsif ($param =~ /^-?cdb_param$/) {
            $cdb_param = $args{$param};
            unless (length($cdb_param)) {
                $self->_set_err(  "Invalid value passed for the '-cdb_param' "
                                . "argument: $param");
                return 0;
            }
        } elsif ($param =~ /^-?sense_key$/) {

            $sense_key = $args{$param};

            unless (length($sense_key)) {

                $self->_set_err(  "Invalid value passed for the '-sense_key' "
                                . "argument: $param");
                return 0;

            }

        } elsif ($param =~ /^-?ascnq$/) {

            $ascnq = $args{$param};

            unless (length($ascnq)) {

                $self->_set_err(  "Invalid value passed for the '-ascnq' "
                                . "argument: $param");
                return 0;

            }

        }  elsif ($param =~ /^-?dir$/) {

            $dir = $args{$param};

            unless (length($dir)) {

                $self->_set_err(  "Invalid value passed for the '-dir' "
                                . "argument: $param");
                return 0;

            }

        }elsif($param =~/^-?opcode/){

           $opcode = $args{$param};

            unless (length($opcode)) {

                $self->_set_err(  "Invalid value passed for the '-opcode' "
                                . "argument: $param");
                return 0;

            }

        }else {

            #
            # We got an invalid argument, fail
            #
            $self->_set_err("Invalid argument '$param'");
            return 0;
        }
    } ## end foreach my $param (keys %args)

    unless ($dev_name) {
        $self->_set_err(  "Error: "
                        . ref($self)
                        . "::test_scsi_cmd(): "
                        . "Device name has be to specified");
        return 0;
    }

    $s_status = "GOOD" unless ($s_status);

    my $cdb_str = $self->get_cdb_bytes($cdb_name);

    return 0 if ($self->error);

    $buf_len = 512 unless length ($buf_len);

    # Check if the CDB parameters are specified
    if ($cdb_param) {
        if($dir){
            $ret_ref = $self->send_cmd(-device    => $dev_name,
                                       -cdb       => $cdb_name,
                                       -cdb_param => $cdb_param,
                                       -dir       => $dir,
                                       -buf_len   => $buf_len,);
        }else{
            $ret_ref = $self->send_cmd(-device    => $dev_name,
                                       -cdb       => $cdb_name,
                                       -cdb_param => $cdb_param,
                                       -buf_len   => $buf_len,);
        }
    } elsif($opcode != ScsiTool::SCSI_CDB){
       if($dir){
            $ret_ref = $self->send_cmd(-device    => $dev_name,
                                       -opcode    => $opcode,
                                       -dir       => $dir);
       }else{
            $ret_ref = $self->send_cmd(-device    => $dev_name,
                                       -opcode    => $opcode);

       }
    }else {
        my $cdb_str = $self->get_cdb_bytes($cdb_name);

        return 0 if ($self->error);

        if($dir){
            $ret_ref = $self->send_cmd(-device  => $dev_name,
                                       -cdb     => $cdb_str,
                                       -dir     => $dir,
                                       -buf_len => $buf_len,);
        }else{
            $ret_ref = $self->send_cmd(-device  => $dev_name,
                                       -cdb     => $cdb_str,
                                       -buf_len => $buf_len,);
        }
    }

    return 0 if ($self->error);

    my $res;

    if ($sense_key && $ascnq) {
        $res = $self->check_status(-send_cmd_output => $ret_ref,
                                   -scsi_status     => $s_status,
                                   -sense_key       => $sense_key,
                                   -ascnq           => $ascnq,);
    } else {
        $res = $self->check_status(-send_cmd_output => $ret_ref,
                                   -scsi_status     => $s_status,);
    }

    # If the check status is successful then return the scsi data
    # else return error

    if ($res) {
        return $ret_ref;
    } else {
        $self->_set_err("Check_status failed");
        ## If check_status failed, the caller might still be interested
        ## in parsing through the SCSI_DATA for the SCSI CMD that was
        ## executed. So return an indication that the check failed and
        ## the ref to the hash that contains the SCSI_DATA

        return $ret_ref;
    }
} ## end sub test_scsi_cmd

########################################################################
# Sub Name:  status_string()
# Objective: Convert a hex value to the appropriate SCSI STATUS string
# Details:   a ref to the send_cmd() output and the SCSI Status to check
#            for are required inputs. The sense key, ASC and ASCQ are
#            optional values that can be input to be checked against as
#            well.
########################################################################

sub status_string {

    my ($self, $s_val) = @_;

    $self->_clear_err();

    return $scsi_status{$s_val} if (exists $scsi_status{$s_val});

    # If we are here then return undef;
    $self->_set_err(    ref($self)
                      . "::status_string did not find "
                      . "the STATUS string in hash for: $s_val");
    return;
} ## end sub status_string

########################################################################
# Sub Name:  sense_key_string()
# Objective: Convert a hex value to the appropriate SCSI Sense Key
#            string
# Details:   Looks up the sense_key hash and returns the string associa
#            ted with the particular hex value for SENSE_KEY
########################################################################

sub sense_key_string {

    my ($self, $s_k_val) = @_;

    $self->_clear_err();

    return $sense_key{$s_k_val} if (exists $sense_key{$s_k_val});

    # if we are here then return undef;
    $self->_set_err(    ref($self)
                      . "::sense_key_string did not find "
                      . "the string for: $s_k_val");
    return;

} ## end sub sense_key_string

########################################################################
# Sub Name:  check_status()
# Objective: Check the output from the send_cmd() call for expected values
#            of SCSI Status, Sense key and ASC and ASCQ
# Details:   a ref to the send_cmd() output and the SCSI Status to check
#            for are required inputs. The sense key, ASC and ASCQ are
#            optional values that can be input to be checked against as
#            well.
########################################################################

sub check_status {

    my ($self, %args) = @_;

    my ($ret_ref, $exp_status, $sense_key, $ascnq, $ignore);

    $self->_clear_err();

    #
    # Parse the arguments sent in the args hash
    #
    foreach my $param (keys %args) {

        if ($param =~ /^-?send_cmd_output$/) {

            $ret_ref = $args{$param};

            unless (length($ret_ref)) {

                $self->_set_err(
                               "Invalid value passed for the '-send_cmd_output "
                                 . "argument: $ret_ref");
                return 0;

            }

        } elsif ($param =~ /^-?scsi_status$/) {

            $exp_status = $args{$param};

        } elsif ($param =~ /^-?sense_key$/) {

            $sense_key = $args{$param};

        } elsif ($param =~ /^-?ascnq$/) {

            $ascnq = $args{$param};

        } elsif ($param =~ /^-?ignore$/) {

            $ignore = $args{$param};

        } else {

            #
            # We got an invalid argument, fail
            #
            $self->_set_err("Invalid argument '$param'");
            return 0;

        }
    } ## end foreach my $param (keys %args)

    $exp_status = 0 unless (defined $exp_status);

    $sense_key = 0 unless (defined $sense_key);

    $ascnq = 0 unless (defined $ascnq);

    my $failCnt = 0;

    $self->sanlog(
          -msg => "=== Looking for $exp_status, " . "$sense_key and $ascnq ===",
          -msg_level => 3);

    my ($ret_status, $ret_sk);

    if ($ret_ref->{SCSI_STATUS} =~ /\S{3,}/) {

        #
        # the value passed in has 3 or more non space chars then
        # the tool has already done the string conversion
        #
        $ret_status = $ret_ref->{SCSI_STATUS};

    } else {

        $ret_status = $self->status_string(hex $ret_ref->{SCSI_STATUS});
    }

    if ($ret_ref->{SENSE_KEY} =~ /\S{3,}/) {

        #
        # the value passed in has 3 or more non space chars then
        # the tool has already done the string conversion
        #
        $ret_sk = $ret_ref->{SENSE_KEY};


    } elsif ($sense_key) {

        #
        # only if the caller is interested in checking the sense
        # key, try to find the string for it. Else the tools might
        # be returning garbage value here and that will unnecessarily
        # set the error about not finding a matching sense key string.
        #

        $ret_sk = $self->sense_key_string(hex $ret_ref->{SENSE_KEY});
    } ## end elsif ($sense_key)

    $self->sanlog(-msg => "Target returned, Status: "
                    . $ret_status
                    . " and Sense Key: "
                    . $ret_sk,
                  -msg_level => 3);

    #
    # handling a special case where USCSI IOCTL can fail but still return
    # with a SCSI GOOD status.
    #
    if (uc $ret_status eq "GOOD" && $ret_ref->{IOCTL_STATUS} != 0) {

        $self->sanlog(-msg => "WARN: Target returned SCSI_STATUS: "
                        . "$ret_status. But, Host "
                        . "IOCTL_STATUS: $ret_ref->{IOCTL_STATUS}",
                      -msg_level => 3);

        return 1;

    } ## end if (uc $ret_status eq ...

    if (   $ignore
        && $ret_status eq "CHECK CONDITION"
        && ((   $ret_sk eq "UNIT ATTENTION"
             && $ret_ref->{ASCNQ} =~ m#29h/00h|3fh/0eh|2ah/06h#i)
            || (   $ret_sk eq "ABORTED COMMAND"
                && $ret_sk->{ASCNQ} =~ m#2fh/00h#i))
      ) {

        $self->sanlog(  -msg => "Ignoring status check for "
                      . $ret_status . ", "
                      . $ret_sk . " "
                      . $ret_ref->{ASCNQ});
        return 1;

    } elsif ($ignore) {

        $self->sanlog(-msg => "ignore flag is set but sense data is not "
                      . "expected to be ignored");

    }

    if (uc($exp_status) ne uc($ret_status)) {

        $self->_set_err("ERROR: Target returned SCSI_STATUS: $ret_status");
        $failCnt++;

    }

    if ($sense_key) {

        if (!$failCnt && ($ret_sk =~ /$sense_key/i)) {

            $self->sanlog(-msg => "$sense_key and $ret_sk matched");

            if ($ascnq && $ret_ref->{ASCNQ} =~ m#$ascnq#i) {

                $self->sanlog(
                             -msg => $ret_ref->{ASCNQ} . " and $ascnq matched");

            } elsif ($ascnq) {

                $self->sanlog(
                       -msg => $ret_ref->{ASCNQ} . " and $ascnq did not match");
                $self->_set_err(
                               $ret_ref->{ASCNQ} . " and $ascnq did not match");
                $failCnt++;

            }

        } else {

            $self->sanlog(-msg => "Sense Key: $sense_key and "
                          . "$ret_sk do NOT match for $exp_status");
            $self->_set_err(  "Sense Key: $sense_key and "
                            . "$ret_sk do NOT match for $exp_status");
            $failCnt++;

        }

    } elsif ($ret_status ne "GOOD") {

        $self->sanlog(-msg => "Target returned Sense Key: $ret_sk, "
                        . "ASCNQ is: $ret_ref->{ASCNQ}",
                      -msg_level => 3);

    }

    my $log_str;

    if ($failCnt) {

        # something when wrong let dump out the status values first
        while (my ($str, $val) = each(%$ret_ref)) {
            $log_str .= "$str: $val\n";
        }

        $self->sanlog(-msg       => "ERROR: scsitool returned: \n $log_str",
                      -msg_level => 3);

    } ## end if ($failCnt)

    return $failCnt ? 0 : 1;

} ## end sub check_status

########## read_or_write_lba ################################################
#
# DESCRIPTION:
#   helper routine that will issue the write or read 16 byte cdb using
#   start lba and for the given lba range.
#   buf length (by default) for the cmd, will equal the transfer size is bytes.
# INPUTS:
#   dev         - the device file (I_T_L) to use
#   read_write  - values should be either "read" or "write"
#   lba         - lba offset to start writing, default is 0
#   alloc_len   - number of lbas to write. Default is 8
#   buf_len     - Length of the data out/in buffer to alloc.
#                 Default is alloc_len*512
# RETURN:
#   UNDEF       - if an error is encountered.
#   ret_ref     - the reference to the send_cmd output is returned
#
####################################################################

sub read_or_write_lba {

    my ($self, %args) = @_;

    $self->_clear_err();

    my ($dev, $lba, $alloc_len, $buf_len, $read_write, $pat);

    foreach my $k (keys %args) {

        if ($k =~ /-?dev/) {

            unless (length $args{$k}) {
                $self->_set_err("Invalid value for '-dev': " . $args{$k});
                return;
            }
            $dev = $args{$k};

        } elsif ($k =~ /-?lba/) {

            unless (length $args{$k}) {
                $self->_set_err("Invalid value for '-lba': " . $args{$k});
                return;
            }
            $lba = $args{$k};

        } elsif ($k =~ /-?alloc_len/) {

            unless (length $args{$k}) {
                $self->_set_err("Invalid value for '-alloc_len': " . $args{$k});
                return;
            }
            $alloc_len = $args{$k};

        } elsif ($k =~ /-?read_write/) {

            unless (length $args{$k}) {
                $self->_set_err(
                               "Invalid value for '-read_write': " . $args{$k});
                return;
            }
            $read_write = $args{$k};

        } elsif ($k =~ /-?buf_len/) {

            unless (length $args{$k}) {
                $self->_set_err("Invalid value for '-buf_len': " . $args{$k});
                return;
            }
            $buf_len = $args{$k};

        } elsif ($k =~ /-?pat/) {

            unless (length $args{$k}) {
                $self->_set_err("Invalid value for '-pattern': " . $args{$k});
                return;
            }
            $pat = $args{$k};

        } else {

            $self->_set_err("Unsupported param: $k");
            return;

        }
    } ## end foreach my $k (keys %args)

    unless ($dev) {

        $self->_set_err("Device value has to be specified");
        return;

    }

    $lba = 0 unless ($lba);

    $alloc_len = 8 unless (defined $alloc_len);

    my $xfer_size = $alloc_len * 512;

    $buf_len = $xfer_size unless ($buf_len);

    $pat = ScsiTool::DEF_DATA_PAT unless $pat;

    #
    # use write16 always to be sure that the most common LUN sizes are
    # addressable
    #
    my $c_ref = new CDB(16);

    if ($c_ref->error) {

        $self->_set_err("Failed to create CDB: " . $c_ref->errmsg());
        return;
    }

    my $op_code = $read_write eq "write" ? 0x8A : 0x88;

    $c_ref->set_operation($op_code);

    my $cdb_str;

    if (    $c_ref->set_address($lba)
        and $c_ref->set_transfer_length($xfer_size)) {

        $cdb_str = join ' ', $c_ref->get_cdb_string("with_spaces");

        $self->sanlog(-msg => "The CDB string is: " . $cdb_str);

    } else {

        $self->_set_err("Failed to set CDB address and/or transfer length"
                        . $c_ref->errmsg());
        return;

    }

    my $ret_ref = $self->send_cmd(-device   => $dev,
                                  -cdb      => $cdb_str,
                                  -buf_len  => $buf_len,
                                  -data_pat => $pat,
                                 );

    if ($self->error) {
        $self->_set_err($self->errmsg());
        return;
    }

    $self->check_status(-send_cmd_output => $ret_ref,
                        -scsi_status     => "GOOD");

    if ($self->error) {
        $self->_set_err(
                   "check_status after $read_write failed: " . $self->errmsg());
        return $ret_ref;
    }

    return $ret_ref;

} ## end sub read_or_write_lba

########## check_pattern  ################################################
#
# DESCRIPTION:
#   Make sure the pattern is repeated in the given data buffer.
#   Assuming that the data is represented in hex so the pattern is converted
#   to hex and then repeated+concatenated to be equal in length to the
#   data_buf. Then both the bufs are compared for equality.
# INPUTS:
#   data_buf    - reference to array of hex values
#   pattern     - string to be compared against
# RETURN:
#   undef       - if an error is encountered
#   PASS        - if write same was successful
#
####################################################################

sub check_pattern {

    my ($self, %args) = @_;

    my ($fail_cnt, $data_buf, $pat);

    foreach my $k (keys %args) {

        if ($k =~ /-?data_buf/) {

            unless (length $args{$k}) {
                $self->_set_err("Invalid value for '-data_buf': " . $args{$k});
                return;
            }
            $data_buf = $args{$k};

        } elsif ($k =~ /-?pat/) {

            unless (length $args{$k}) {
                $self->_set_err("Invalid value for '-pat': " . $args{$k});
                return;
            }
            $pat = $args{$k};

        } else {

            $self->_set_err("Unsupported param: $k");
            return;

        }
    } ## end foreach my $k (keys %args)

    unless ($data_buf and $pat) {
        $self->_set_err(  "Error: Invalid Data Buffer: "
                        . $data_buf
                        . "and/or invalid Pattern: "
                        . $pat);
        return;
    }

    # convert the pattern to hex
    my $h_pat = unpack("H*", $pat);

    $self->sanlog(-msg       => "\"$pat\" in hex: " . $h_pat,
                  -msg_level => 5);

    my $chk_buf = $h_pat x ((length join '', @$data_buf) / (length $h_pat));

    # find the modulo and add that substr to the chk_buf

    # using a negative length so that substr returns the first half
    # of the string and not the last part.
    my $mod_pat_l =
      -(length($h_pat) - (length(join '', @$data_buf) % length($h_pat)));

    $self->sanlog(msg => "length of pat that needs to be added: " . $mod_pat_l);

    my $s_str = substr $h_pat, 0, $mod_pat_l;

    $chk_buf .= $s_str if (length $s_str);

    $self->sanlog(-msg       => "pat residual: " . $s_str
                  -msg_level => 5);

    if ($chk_buf eq join '', @$data_buf) {
        $self->sanlog(-msg       => "Pattern matched in the buffer");

        $self->sanlog(-msg       => "Pattern: $pat",
                      -msg_level => 5);
        return 1;
    }

    $self->_set_err(msg => "$pat did not match the data buffer");

    return;

} ## end sub check_pattern

########################################################################
# P R I V A T E   M E T H O D S
########################################################################

########################################################################
# Method:    _base_new
# Objective: ScsiTool Object constructor.
# Details:   See POD documentation at the end of this file.
# Notes:     Tools currently supported: SPT, USCSI, PTI, IFC5d, SANBlaze
########################################################################
sub _base_new {
    my ($caller, %args) = @_;
    my $class = ref($caller) || $caller;

    my $self = {};

    #
    # Bless the object which in a sense creates the class
    #
    bless($self, $class);

    $self->_clear_err();

    $self->set_debug_level($args{-debug_level});

    #
    # By returning $self here we are returning a parent class such
    # as PTI, not a ScsiTool object.  This is OK because
    # the tool specific ScsiTool::toolname classes add SsciTool in @ISA
    # so SsciTool is the base class for this parent
    #
    # set the tool field
    #
    $self->{tool} = $args{-tool};

    #
    # set the client field
    #
    $self->{client} = $args{-client};

    #
    # store the hash ref of the SCSI Cdb Name to String mapping hash
    #
    $self->{cdb_name_hash} = \%scsiCdbs;

    #
    # init eveything to defaults, so that we are sure
    # of the responses from the subclasses
    #
    $self->_set_def_params();

    return $self;
}    # end new()

########################################################################
# Subroutine Name: _set_def_params()
# Objective:       Set the defaults for all the params used by HASTA base class
#                  This should be done when the constructor is executed first.
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub _set_def_params {

    my $self = shift @_;

    #
    # send_cmd api cmd can use any of the following
    # params. These are listed here so that sub-class
    # and testcases know what is supported.
    #
    my %send_cmd_params;

    $send_cmd_params{DEVICE}      = DEF_DEV;
    $send_cmd_params{MPIO_PATH}   = DEF_PATH;
    $send_cmd_params{CDB}         = DEF_CDB;
    $send_cmd_params{CDB_PARAM}   = DEF_CDB_PARAM;
    $send_cmd_params{BUF_LEN}     = DEF_LEN;
    $send_cmd_params{OPCODE}      = DEF_OP;
    $send_cmd_params{QTAG}        = DEF_QTAG;
    $send_cmd_params{OPCODE_NAME} = DEF_OP_NAME;
    $send_cmd_params{REPEAT_CNT}  = DEF_CNT;
    $send_cmd_params{RUNTIME}     = DEF_RUNTIME;
    $send_cmd_params{SLEEP}       = DEF_SLEEP;
    $send_cmd_params{TIMEOUT}     = DEF_TMO;
    $send_cmd_params{ON_ERROR}    = DEF_ONERR;
    $send_cmd_params{DIR}         = DEF_DIR;
    $send_cmd_params{DATA_IN}     = DEF_DATA_IN;
    $send_cmd_params{DATA_OUT}    = DEF_DATA_OUT;
    $send_cmd_params{DATA_PAT}    = DEF_DATA_PAT;
    $send_cmd_params{ENABLE}      = DEF_DISABLED;

    #
    # assign the send_cmd_params to $self->{SC_PARAMS}
    #
    $self->{SC_PARAMS} = \%send_cmd_params;


    #
    # by default all tools must return these values. Unless the
    # exp_rsp_mask is overridden by a call to expected_responses.
    # don't try to break this string with the \ char, perl does not like it
    # and will cause a needle in a haystack type error
    #
    #my $def_exp_rsp_mask = TOOLNAME | IOCTL_STATUS | SCSI_DATA | SCSI_STATUS |\
    #                       HOST_STATUS | DRIVER_STATUS | SENSE_KEY | ASCNQ;
    # 0x63F or 11000111111 or 1599
    my $def_exp_rsp_mask = 1599;

    #
    # Sub classes will need to check that a particular flag is enabled
    # and then use that flag to update the index into the resp_hsh
    # in $self->{response}
    #
    $self->{DEF_EXP_RSP_MASK} = $def_exp_rsp_mask;

    if($expected_response){
        $expected_response = 0;
    }else{
        $self->{EXP_RSP_MASK} = $def_exp_rsp_mask;
        $self->{RSP_OPTION}   = RSP_REBUILD;
    }

    #
    # now that the bit mask is complete make sure to
    # build the response hash such that the sub-class
    # will know what responses to send back to the
    # testcase.
    #
    # the exp_rsp_mask is valid for only
    # the next call to send_cmd;
    my %resp_hash;
    $self->{ER_PARAMS} = \%resp_hash;

    $self->build_exp_rsp_hash();

} ## end sub _set_def_params

########################################################################
#                             POD Documentation
########################################################################

=pod

=head1 Package Name

ScsiTool.pm

=head1 Authors

=begin html

Bhanu Palakodeti

=end html

=head1 Description

ScsiTool provides an API interface to the several Scsi Tools.
This package can be instantiated through the SanClient::create_scsi_tool()
method.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>
B<Sanlog.pm>

=head1 Attributes

=head2 B<tool>

string - SCSI tool this object executes

=head2 B<debug_level>

integer - Current debug level defaults to client's debug level

=head1 Public Methods

=head2 B<create_scsi_tool(%args)>

Create a ScsiTool object

=over 4

=item Inputs

=item * -client

SanClient Object for the client that will execute the SCSI tool

=item * I<-tool>

Tool this object will execute.
Currently supported: SPT, USCSI, PTI, IFC5d, SANBlaze

=item Outputs

Returns a ScsiTool object if successful.  If an error occured the error
flag is set and errmsg will be is set to the failure reason.

=item Examples

    $Test->nlog("Create the ScsiTool Object");
    my $scsi_tool_obj =  ScsiTool->create_scsi_tool(-client      => $client_obj,
                                                    -tool        => $tool);
    if ($scsi_tool_obj->error())
    {
        my $errmsg = $scsi_tool_obj->errmsg();
        logresult('FATAL', "Failed to instantiate ScsiTool object: $errmsg");
    }

=back


=head2 B<send_cmd(%args)>

Use the current tool to send a SCSI command, with a bunch of optional
parameters to be passed into the underlying SCSI Tool. Based on the
individual implementations of the SCSI Tool, some parameters will
not have the desired effect. It is important to consider what tool
will eventually handle the params, before relying heavily on the
params working correctly. For most of SCSI testing, "-device"
"-cdb" and "-cdb_param" seem to be sufficient

=over 4

=item Inputs

=item * -device

SCSI Pass Through path that identifies the ITL to use for the test.
In Native MPIO host stacks the optional "-mpio_path" is required to
uniquely identify an ITL. Another way to get to a unique ITL, in
native MPIO (ALUA) host stacks, is to disable all other available
paths.

=item * -cdb

CDB bytes that represent the test to be executed, in hex without any
leading "0x" or trailing "h". WRITE6 for 1 block is represented as
"0A 00 00 00 01 00"

=item * I<-cdb_param>

Parameter data for the cdb under test.

=item * I<-buf_len>

Num of bytes that the host should allocate for the data being transferred
into or out of the host.

=item * I<-runtime>
=item * I<-timeout>
=item * I<-opcode>
=item * I<-qtag>
=item * I<-opcode_name>
=item * I<-repeat_cnt>
=item * I<-sleep>
=item * I<-dir>
=item * I<-on_error>
=item * I<-data_in>
=item * I<-data_out>
=item * I<-data_pat>

=item Outputs

If the command was executed successfully, TRUE is returned.
If an error occurs the error flag is set, errmsg is set to the failure reason
and the object is returned.

=item Examples

    my $ret_ref = $scsi_tool_obj->send_cmd(-device      => $device,
                                           -cdb         => "08 00 00 00 1 00");

    if ($scsi_tool_obj->error()) {
        my $errmsg = $scsi_tool_obj->errmsg();
        $Test->nlog("Cmd Failed:  $errmsg");
        return $TCD::FAIL;
    }

    my $ret_ref = $scsi_tool_obj->send_cmd(-device      => $device,
                                           -cdb         => "08 00 00 00 1 00",
                                           -cdb_param   => "00 00 00 00",
                                           -mpio_path   => 0x12345,
                                           -buf_len     => 255,
                                           -runtime     => 100,
                                           -timeout     => 5,
                                           -opcode      => ScsiTool::SCSI_CDB|\
                                                           ScsiTool::LUN_RESET|\
                                                           ScsiTool::NACA_BIT,
                                           -qtag        => ScsiTool::HEAD_OF_Q,
                                           -opcode_name => "READ6",
                                           -repeat_cnt  => 10,
                                           -sleep       => 2,
                                           -dir         => ScsiTool::READ,
                                           -on_error    => ScsiTool::STOP,
                                           -data_in     => $any_ref_handled_by_subclass,
                                           -data_out    => $any_ref_handled_by_subclass,
                                           -data_pat    => "AABBCC",
                                           );

=back

=head2 B<check_status(%args)>

Utility to check the status of the SCSI Cmd that just completed. By default
this checks for SCSI GOOD status. SCSI Cmds can fail for several retry-
able error conditions, so it is desirable to be able to check for these
retryable error codes and then retry the same command. This interface
provides the abstraction for checking the error code strings/bytes
without worrying about the SCSI hex layout of the error code frames.

=over 4

=item Inputs

=item * -send_cmd_output

The hash ref returned from the send_cmd() call

=item * -scsi_status

The SCSI STATUS that is expected for the SCSI cmd

sent in the send_cmd() call

=item * -sense_key

Specifies the sense_key value that will be matched against the sense key
value in the "-send_cmd_output" hash ref.
If this matchs and no ascnq param is set, then the fucntion returns true.
If this does not match and the ascnq param is set, then a failure is logged.
If this does not match and the the ascnq param is not set then the function
returns false/zero/undef.

=item * -ascnq

Specifies the asc and ascq values in "xxh/yyh" format, that will be matched
against the ascnq value in the "-send_cmd_output" hash ref.
This value is matched only if "-sense_key" param is also defined.
If this matches the value in the hash, then the function returns true.
Else the function returns false.

=item Outputs

If the check passes a 1 is returned.
If an error occurs the error flag is set, errmsg is set to the failure reason
and a 0 is returned.

=item Examples

    if($self->check_status( -send_cmd_output    => $ret_ref,
                            -scsi_status        => "GOOD"))

=back

=head2 B<expected_responses(%args)>

Interface to tell the underlying Scsi Tool what responses are
expected by the testcases. These responses will be considered as
ideal to fill by the tool implementations

=over 4

=item Inputs

=item * -exp_rsp_bitmask

Use an combination of response bitmasks that are defined in this module.
It is up to the underlying tool to value the bitmask.

=item * -exp_rsp_option

option to let ScsiTool know to either rebuild the required responses hash
or to just this option to the existing hash. This way we can override
the default responses hash with what ever is needed.

=item Outputs

If an error occurs the error flag is set, errmsg is set to the failure reason.
The object is always returned.

=item Examples

    $scsi_tool_obj->expected_responses(-exp_rsp_bitmask => ScsiTool::SCSI_STATUS,
                                       -exp_RSP_OPTION => ScsiTool::RSP_ADD);

=back

=head2 B<read_or_write_lba(%args)>

Helper routine that will issue the write or read 16 byte cdb using
start lba and for the given lba range.
buf length (by default) for the cmd, will equal the transfer size in bytes.

The return from the send_cmd call will be checked for a GOOD status by
default.

=over 4

=item Inputs

=item * -dev

Required param which identifies the path on which to issue the CDB and
check for the response.

=item * -read_write

Use this param to indicate if a read or write operation is requested.
Based on this param a read or write 16 byte CDB will be created.

=item * -lba

The lba offset to be used in the CDB being built. By default the lba will
be assumed to be 0, start of the prefix in a LU.

=item * -alloc_len

This param determines the number of LBAs that will be read or written. The
default value will be 8

=item * -buf_len

This param determines the amount of buffer (in bytes) the host OS is to
allocate that will be used to hold the alloc_len*512 worth of data. Default
will be alloc_len*512 bytes.

=item * -pattern

Optionally this param can be used to write a specific pattern. The value will
be a string that will be repeated to fill in the alloc_len*512 bytes.

=item Outputs

If an error occurs the error flag is set, errmsg is set to the failure reason
and UNDEF is returned.

If there are no errors then the return value from the send_cmd() call is
returned.

=item Examples

    my $ret_ref =
      $scsi_tool_obj->read_or_write_lba(
                                       -dev       => $lc_p,
                                       -start_lba => $start_lba,
                                       -read_write => "read");
    unless ($ret_ref) {
        $Test->nlog(  "Reading from clone failed, w/o Unmap "
                    . "on path $lc_p");
        $fail_cnt++;
    }

    if ($scsi_tool_obj->read_or_write_lba(-dev => $lc_paths[-1],
                                          -start_lba  => $start_lba,
                                          -pattern    => "z",
                                          -read_write => "write"
        ) == FAIL
      ) {
        $Test->nlog("Error:: Write failed");
        return $TCD::FAIL;
    }

=back

=head2 B<check_pattern(%args)>

Make sure the pattern is repeated in the given data buffer.
Assuming that the data is represented in hex so the pattern is converted
to hex and then repeated+concatenated to be equal in length to the
data_buf. Then both the bufs are compared for equality.

=over 4

=item Inputs

=item * -data_buf

Array ref to the hex data that will be checked for a pattern match

=item * -pattern

string that will be converted to hex and compared against the data_buf
contents, repeating as necessary

=item Outputs

If an error occurs the error flag is set, errmsg is set to the failure reason
and UNDEF is returned.

If the pattern does not match, the errmsg is set to indicate miscompare and
UNDEF is returned.

If the pattern matches then 1 is returned

=item Examples

    # verify that the data read matches pattern 'y'
    unless ($ret_ref
            and $scsi_tool_obj-> check_pattern(
                             -data_buf => $ret_ref->{SCSI_DATA},
                             -pattern  => "before snapshot")
      ) {
        $Test->nlog(  "Failed to read back \"before snapshot\" "
                    . "pattern on path $lc_p");
        $fail_cnt++;
    }

=back

=cut

1;
