#$Id: //depot/prod/test/Rsierranevada/lib/San/ScsiTool/AnyIscsi.pm#1 $
## @summary Provide an interface to the any_iscsi tool
## @author chadwell@netapp.com
## @status public
## @pod ../ScsiTool.pm

#
# Package Name
#
package San::ScsiTool::AnyIscsi;

use strict;

#
# Package Requirements
#
use San::ScsiTool;
use San::SanClient::ClientUtil;

use Error;
use San::SanLog;

use vars qw( @ISA $ScsiTool);

@ISA = qw(Error SanLog ScsiTool);
my $pattern_file = "/tmp/Pattern.txt";
my $dt_loc       = "/usr/software/test/bin/dt";
########################################################################
# P U B L I C   M E T H O D S
########################################################################
my $batch_filename = "anyIscsiBatchFile";
my $any_iscsi_loc  = "/u/dsharma/p4/any_iscsi/any_iscsi/any_iscsi ";
########################################################################
# Subroutine Name: send_cmd()
# Objective:       Execute the Spt tool to send a SCSI Command
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub send_cmd {

    my ($self, %args) = @_;
    $self->_clear_err();

    $self->handle_send_cmd_params(%args);

    #
    # Build the string required to run on the target
    #
    $self->build_cmd_string();
    return $self if ($self->error());

    #
    # Add some log messages to keep up
    #
    $self->sanlog(-msg       => "cmd string   => " . $self->{sc_cmd} . "\n",
                  -msg_level => 3);

    #
    # executing on the client
    #
    my $cmd         = $self->{sc_cmd};
    my $client_obj  = $self->{client};
    my $client_conn = $client_obj->get_connection();

    if ($self->{batch_cmd}) {

        #
        #Put the cmd string in the remote file buffer.
        #
        &ClientUtil::create_remote_file($client_conn, $batch_filename, $cmd);
    } else {
        $self->sanlog(-msg => " Execute from AnyIscsi:"
                        . "cmd = $cmd and client_conn=$client_conn",
                      -msg_level => 3);

        if ($self->{pattern}) {
            $cmd = $self->{pattern} . $cmd;
        }

        my ($status, @output) = &ClientUtil::execute($client_conn, $cmd);

        if ($status > 0) {
            $self->sanlog(
                     -msg       => "Command ($cmd) returned a bad status ($status)." 
                     -msg_level => 3);
        }

        #
        # Update the response hash with \n separated
        # lines in an array
        #
        $self->update_rsp_hash(\@output);

        return $self if ($self->error);

        #
        # Set all non-updated responses to undef()
        #
        $self->handle_unsup_responses();

        return $self if ($self->error);

        #
        # Now return the expected parameters back to
        # the calling program.
        #
        return $self->get_exp_rsp_params();
    }
}    # end send_cmd()

########################################################################
# Subroutine Name: update_rsp_hash()
# Objective:       Update response hash variables with information from the
#                  target so the test can access it.
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub update_rsp_hash {
    my ($self, $output_ref) = @_;
    my @output = @$output_ref;

    $self->_clear_err();

    #
    # set response hash to default
    #
    $self->build_exp_rsp_hash();
    return $self if ($self->error);

    #
    # Get the reference to information
    # the test wants
    #
    my $er_ref   = $self->get_exp_rsp_params();
    my $bhs_flag = 0;

    my (@ln_bytes, @data_bytes);
    while (my ($name, $value) = each(%$er_ref)) {
        if ($value == ScsiTool::REQ_RESPONSE) {
            if ($name eq "SCSI_DATA") {

                #
                # take the SCSI bytes off the top
                # one line at a time
                #
                foreach my $ln (@output) {
                    if ($ln =~ /0x.{2}:/) {

                        #
                        # the reference wants an array,
                        # so split the string on spaces
                        #
                        @ln_bytes = split " ", $ln;

                        #
                        # remove the time stamp information and offset info from scsi_data.
                        # we don't want it
                        #
                        my @tmp = @ln_bytes;

                        foreach my $byte (@tmp) {
                            if ($byte =~ /\W/) {
                                shift @ln_bytes;
                            }
                        }

                        @data_bytes = (@data_bytes, @ln_bytes)
                          unless ($bhs_flag);
                        $er_ref->{SCSI_DATA} = \@data_bytes;
                    } elsif ($ln =~ /bhs/) {
                        if (!@data_bytes) {
                            $self->sanlog(-msg =>
                                "bhs data is present but there are no data bytes!"
                                -msg_level => 3);
                        }

                        # Dealing with BHS data
                        $bhs_flag = 1;
                    } elsif ($ln =~ /0 bytes of data/) {
                        $er_ref->{SCSI_DATA} = "0";
                        next;
                    } elsif ($ln =~ /Response data seg length/) {

                        # Done dealing with BHS data
                        $bhs_flag = 0;
                    }
                }

                if ($er_ref->{SCSI_DATA} eq ScsiTool::REQ_RESPONSE) {
                    $self->_set_err(
                                 "SCSI data incorrectly parsed from any_iscsi");
                    return $self;
                }
            } elsif ($name eq "SCSI_STATUS") {
                my $status;
                foreach my $ln (@output) {
                    if ($ln =~ /SCSI Status: (\w+)/) {
                        $status = $1;

                        # require the scsi status values to be
                        # hex digits.  Update the response hash in
                        # the form xxh/yyh
                        #
                        $status =~ /0x(\d+)/;
                        my $scsi_status = $1;
                        if ($scsi_status =~ /^(\d|\w)$/) {
                            $scsi_status = "0" . $scsi_status;
                        }
                        $er_ref->{SCSI_STATUS} = $scsi_status;
                    } elsif ($ln =~ /TMF STATUS \: (\d+)/) {
                        $er_ref->{SCSI_STATUS} = $1;
                    } elsif ($ln =~ /Errorful status returned\: (\w+)/) {
                        $er_ref->{SCSI_STATUS} = $1;
                    }
                }

                if ($er_ref->{SCSI_STATUS} eq ScsiTool::REQ_RESPONSE) {
                    $self->_set_err(
                               "SCSI status incorrectly parsed from any_iscsi");
                    return $self;
                }
            } elsif ($name eq "SENSE_KEY") {

                my $sense;
                foreach my $ln (@output) {
                    if ($ln =~ /Sense Data \: ([0-9A-F][0-9A-F])h\//) {
                        $sense = $1;
                    }
                }

                #
                # It already conforms sense_key to xxh/yyh form
                #
                $er_ref->{SENSE_KEY} = $sense;

                if ($er_ref->{SENSE_KEY} eq ScsiTool::REQ_RESPONSE) {
                    $self->_set_err(
                           "SCSI sense data incorrectly parsed from any_iscsi");
                    return $self;
                }
            } elsif ($name eq "ASCNQ") {
                my ($asc, $ascq);
                foreach my $ln (@output) {
                    if ($ln =~
                        /Sense Data[\s|\t]+\:[\s|\t]+([0-9A-F][0-9A-F]h)\/([0-9A-F][0-9A-F]h)\/([0-9A-F][0-9A-F]h)/
                      ) {
                        $asc  = $2;
                        $ascq = $3;
                    }
                }

                #
                # conform asc and ascq
                # to xxh/yyh form
                #
                $er_ref->{ASCNQ} = $asc . "/" . $ascq;

                if ($er_ref->{ASCNQ} eq ScsiTool::REQ_RESPONSE) {
                    $self->_set_err(
                           "SCSI ASC/Q data incorrectly parsed from any_iscsi");
                    return $self;
                }
            }

            elsif ($name =~ /RESID/) {

                #
                # split the residual information returned
                # and conform it to xxh/yyh
                #
                my $resid;
                foreach my $ln (@output) {
                    if ($ln =~ /Resid\: (0x\d+)/) {
                        $resid = hex $1;
                    }
                }

                $er_ref->{RESID_BYTES} = $resid;

                if ($er_ref->{RESID} eq ScsiTool::REQ_RESPONSE) {
                    $self->_set_err(
                         "SCSI resdiual value incorrectly parsed from any_iscsi"
                    );
                    return $self;
                }
            } elsif ($name eq "BYTES_XFERD") {
                my $xferd;
                foreach my $ln (@output) {
                    if ($ln =~ /Response data seg length\: (\d+)/) {
                        $xferd = $1;
                    }
                }
                $er_ref->{BYTES_XFERD} = $xferd;

                if ($er_ref->{BYTES_XFERD} eq ScsiTool::REQ_RESPONSE) {
                    $self->_set_err(
                        "SCSI bytes transferred value incorrectly parsed from any_iscsi"
                    );
                    return $self;
                }
            }
        }
    }

    #
    # set the toolname
    #
    $er_ref->{TOOLNAME} = "any_iscsi";

    $self->sanlog(-msg       => "Output array string is @data_bytes\n",
                  -msg_level => 3);

    $self->{batch_cmd} ? $er_ref : 1;
}    # end Update_rsp_hash()

########################################################################
# Subroutine Name: build_cmd_string()
# Objective:       Take incoming information from host agnostic test and
#                  build appropriate command string for the Scsi target sim
# Details:         See POD documentation in ScsiTool.pm
########################################################################

sub build_cmd_string {

    my $self = shift @_;

    $self->_clear_err();

    $self->{pattern} = undef;

    my $sc_ref = $self->get_send_cmd_params();

    my $dev        = $sc_ref->{DEVICE};
    my $cdb        = $sc_ref->{CDB};
    my $cdb_param  = $sc_ref->{CDB_PARAM};
    my $dir        = $sc_ref->{DIR};
    my $buf_len    = $sc_ref->{BUF_LEN};
    my $opcode     = $sc_ref->{OPCODE};
    my $qtag       = $sc_ref->{QTAG};
    my $opname     = $sc_ref->{OPCODE_NAME};
    my $repeat_cnt = $sc_ref->{REPEAT_CNT};
    my $runtime    = $sc_ref->{RUNTIME};
    my $timeout    = $sc_ref->{TIMEOUT};
    my $on_error   = $sc_ref->{ON_ERROR};
    my $sleep      = $sc_ref->{SLEEP};
    my $data_in    = $sc_ref->{DATA_IN};
    my $data_out   = $sc_ref->{DATA_OUT};
    my $data_pat   = $sc_ref->{DATA_PAT};
    my $cmd;
    my $caw_buf_length = 1024;
    my $dt_pattern_length = 8;

    # Any_iscsi device: <ip_address> : <lun id>
    my @info;
    my @dev_array;
    my $lun_id = 0;

    if (ref($dev =~ /array/i)) {
        @dev_array = @$dev;
        @info = split(":", $dev_array[0]);
    } else {
        @info = split(":", $dev);
    }

    $dev    = $info[0];
    $lun_id = $info[1];

    my @cdb_param = split " ", $cdb_param;

    $cdb_param = "";

    foreach my $hex_byte (@cdb_param) { $cdb_param .= "0x" . $hex_byte . " "; }

    #
    # every command starts this way if it's going
    # out one command at a time.  If it's being put
    # in a batch file, we only need the lun.
    #

    #
    #If the client is running solaris, the
    #any_iscsi executable in use needs to reflect
    #that
    if ($self->{client}->hosttype() =~ /solaris/) {
        $any_iscsi_loc =~ s/any_iscsi /any_iscsi_solaris /;
    }

    if ($self->{batch_cmd}) {
        $cmd = "0 ";
    } else {
        $cmd = "$any_iscsi_loc" . $dev . " " . $lun_id . " ";
    }

    #
    # this is always the beginning
    # CDB of the command.  It'll be
    # blank if we are sending a TMF,
    # so no harm done.
    #
    $cmd .= "$buf_len " unless ($opcode != ScsiTool::SCSI_CDB);

    #
    #Add the -n command to stop TUR's
    #before batched commands
    #
    $cmd .= "-n " unless (!$self->{batch_cmd});

    my @cdb_bytes = split(" ", $cdb);

    # opcodes 41: WRITE SAME 10
    #         93: WRITE SAME 16
    #         89: COMPARE and WRITE
    #         04: Format Unit
    #         16: Reserve 6
    #         17: Release 6
    #         56: Reserve 10
    #         57: Release 10
    if ($cdb_bytes[0] =~ /41|93|89|04/) {
        $cmd .= " -z ";
    } elsif ($cdb_bytes[0] =~ /16|17|56|57/) {

        # Any iscsi looks at the -Data Out parameters when a non zero
        # buffer length is used.
        # fill the buffer with zeros before use
        # -z option isnt working
        $cmd .= " -F $buf_len 0 ";
    }

    #
    #Add the verbose tag so you can see
    #the return bytes. Also, don't require
    #GOOD status from the initial TUR.
    #
    my $file = $cdb;

    $cmd .= " -b -v ";

    if ($data_pat ne ScsiTool::DEF_DATA_PAT) {
        my $client_obj = $self->{client};
        my $host_type  = $client_obj->hosttype();

        if ($client_obj->error()) {
            $self->_set_err($client_obj->errmsg());
            return $self;
        }

        my $alloc_len = $buf_len / 512;

        my $dt;

        # dt Eg: pattern=0x39c39c39
        # specify this pattern when the number of LBAs to be written is more than 1024

        if ($buf_len > $caw_buf_length) {
            $data_pat = unpack("H*", $data_pat);

            $data_pat x= ($dt_pattern_length / (length $data_pat));

            $dt =
              "$dt_loc of=- bs=512 count=$alloc_len pattern=$data_pat disable=verify,stats | ";

            $self->{pattern} = $dt;
        } else {

            # CAW uses a pattern string which contains 512 bytes of verify pattern
            # and 512 bytes of write pattern. Create a Pattern file with the

            #
            #If there is a remote file there already, remove it
            #

            $client_obj->execute("rm -f " . $pattern_file);

            $client_obj->execute("touch $pattern_file");

            #
            #Number of times the pattern needs repeating (rounded down)
            #
            my $multiplier = $buf_len / length($data_pat);
            $data_pat x= $multiplier;

            #
            #Finds out how many odd characters were left off
            #so they can be added too
            #
            my $chars_left = $buf_len - length($data_pat);

            #
            # data_pat now holds the whole pattern
            #
            $data_pat .= substr($data_pat, 0, $chars_left);

            my $dt =
              "$dt_loc of=$pattern_file bs=512 count=$alloc_len pattern=$data_pat disable=verify,stats;";

            $self->{pattern} = $dt;

            $cmd .= " -d $pattern_file ";
        }
    }

    #
    #
    # If the test is sending a TMF
    # to the SCSIT-sim, then we
    # build something a little different
    #
    my $tmf;
    if ($opcode != ScsiTool::SCSI_CDB) {
        $tmf = "at"    if ($opcode & ScsiTool::ABORT_TASK);
        $tmf = "ats"   if ($opcode & ScsiTool::ABORT_TASK_SET);
        $tmf = "ca"    if ($opcode & ScsiTool::CLEAR_ACA);
        $tmf = "cts"   if ($opcode & ScsiTool::CLEAR_TASK_SET);
        $tmf = "lr"    if ($opcode & ScsiTool::LUN_RESET);
        $tmf = "lo"    if ($opcode & ScsiTool::LOGOUT);
        $tmf = "li"    if ($opcode & ScsiTool::LOGIN);
        $tmf = "twr"   if ($opcode & ScsiTool::TARGET_WARM_RESET);
        $tmf = "tcr"   if ($opcode & ScsiTool::TARGET_COLD_RESET);
        $tmf = "tr"    if ($opcode & ScsiTool::TASK_REASSIGN);
        $tmf = "unreg" if ($opcode & ScsiTool::UNREGISTER);
        $tmf = "reg"   if ($opcode & ScsiTool::REGISTER);
        $tmf = "qt"    if ($opcode & ScsiTool::QUERY_TASK);
        $tmf = "qts"   if ($opcode & ScsiTool::QUERY_TASK_SET);
        $tmf = "qae"   if ($opcode & ScsiTool::QUERY_ASYNC_EVENT);

        if (!$tmf) {
            $self->_set_err("Unsupported OP for any_iscsi TMF: $opcode");
            $self->{TOOL_LIB_STATUS} |= ScsiTool::OP_UNSUP;
            return $self;
        }

        $self->{sc_cmd} = $cmd . " -t $tmf";

        #
        # and return here because we're done
        #
        return;
    } else {

        #
        # add an appropriate qtag if available
        #
        $cmd .= "$qtag " if ($qtag =~ /\-s|\-o|\-h|\-a|\-n|\-v|\-q|\-n/);

        if (length $cdb_param > 5) {
            $cmd .= " -D " . $cdb_param . " -c ";
        }

        #
        # begin the CDB data bytes
        # in hex
        #
        my @cdb = split " ", $cdb;
        foreach my $hex_byte (@cdb) { $cmd .= "0x" . $hex_byte . " "; }
    }

    #
    # put the command string
    # in the object hash
    #
    $self->{sc_cmd} = $cmd;
}    #  end build_cmd_string()

########################################################################
# Subroutine Name: batch_cmd
# Objective:       Set the Scsitool to batch anyiscsi commands in a remote file
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub batch_cmd {
    my ($self, %args) = @_;
    my ($cmd, $device);

    #
    # Parse the arguments sent in the args hash
    #
    foreach my $param (keys %args) {
        if ($param =~ /^-?cmd$/) {
            $cmd = $args{$param};
            if (not length($cmd)) {
                $self->_set_err("invalid value for $param");
                return $self;
            }
        } elsif ($param =~ /^-?device$/) {
            $device = $args{$param};
            if (not length($device)) {
                $self->_set_err("invalid value for $param");
                return $self;
            }
        } else {
            $self->_set_err("unhandled parameter sent to batch_cmd");
            return $self;
        }
    }

    if ($cmd =~ /start/i) {
        $self->{batch_cmd} = 1;

        #
        #set the device
        #
        if (length $device) {
            $self->handle_send_cmd_params(-device => $device);
        }
    } elsif ($cmd =~ /execute/i) {

        #
        #Execute on the client and return...
        #
        my $sc_ref = $self->get_send_cmd_params();
        my $dev    = $sc_ref->{DEVICE};
        my @dev_array;
        if (ref($dev) =~ /array/i) {
            @dev_array = @$dev;
        }

        #
        #If the client is running solaris, the
        #any_iscsi executable in use needs to reflect
        #that
        if ($self->{client}->hosttype() =~ /solaris/) {
            $any_iscsi_loc =~ s/any_iscsi /any_iscsi_solaris /;
        }

        my $batch_cmd =
          "$any_iscsi_loc " . ($dev_array[0] | $dev) . " -f $batch_filename";
        my $client_obj  = $self->{client};
        my $client_conn = $client_obj->get_connection();

        my ($status, @output) = &ClientUtil::execute($client_conn, $batch_cmd);

        if ($status > 0) {
            $self->sanlog(
                      -msg => "Command ($cmd) returned a bad status ($status).",
                      -msg_level => 3);
        }

        #
        #Commands were sent successfully, so
        #remove the remote file
        #
        &ClientUtil::remove_remote_file($client_conn, $batch_filename,
                                        "ignore_err");

        return $self if ($self->error);

        #
        # Update the response hash with \n separated
        # lines in an array
        #
        my $tmp_output = join('\n', @output);
        @output = split(/\\n\\n\\n/, $tmp_output);
        my ($ret_ref, @ret_array, @tmp_output);
        foreach my $ln (@output) {
            @tmp_output = split(/\\n/, $ln);
            $ret_ref = $self->update_rsp_hash(\@tmp_output);
            push @ret_array, $ret_ref;
        }
        return $self if ($self->error);

        #
        #Set batch_cmd back to undefined
        #
        $self->{batch_cmd} = undef();

        #
        # Set all non-updated responses to undef()
        #
        $self->handle_unsup_responses();

        return $self if ($self->error);

        #
        # Now return the array hash of expected
        # parameters back to
        # the calling program.
        #
        return (wantarray()) ? @ret_array : \@ret_array;
    } else {
        $self->_set_err("Batch commands support \"start\" and \"execute\"");
        return $self;
    }
}

########################################################################
# Subroutine Name: _new()
# Objective:       Make a ScsiTool into an any_iscsi
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub _new {

    my ($caller, %args) = @_;
    my $class = ref($caller) || $caller;
    my $self = $class->_base_new(%args);

    return $self;

}    # end _new()

