#$Id: //depot/prod/test/Rsierranevada/lib/San/ScsiTool/CDB.pm#1 $

## @summary Construct the CDB in a SCSI commmand
## @author evan@netapp.com
## @status private
## @pod here

=pod

=head1 NAME

CDB.pm

=head1 DESCRIPTION

This package can be used to create a CDB byte string which includes the SCSI
operation code, logical bus address (LBA), and transfer length.

=head1 METHODS

=cut

package CDB;

use strict;

use bigint;
no warnings 'portable'; 
use San::SanClient::ClientUtil;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error);

use constant{
    BYTE_0 => 0x00000000000000FF,
    BYTE_1 => 0x000000000000FF00,
    BYTE_2 => 0x0000000000FF0000,
    BYTE_3 => 0x00000000FF000000,
    BYTE_4 => 0x000000FF00000000,
    BYTE_5 => 0x0000FF0000000000,
    BYTE_6 => 0x00FF000000000000,
    BYTE_7 => 0xFF00000000000000,
};

# Global definitions of SCSI commands
$CDB::TUR        = 0x00;
$CDB::READ_6     = 0x08;
$CDB::READ_10    = 0x28;
$CDB::READ_12    = 0xA8;
$CDB::READ_16    = 0x88;
$CDB::XDREAD_10  = 0x52;
$CDB::WRITE_6    = 0x0A;
$CDB::WRITE_10   = 0x2A;
$CDB::WRITE_12   = 0xAA;
$CDB::WRITE_16   = 0x8A;
$CDB::XDWRITE_10 = 0x50;

# The following codes have the service action appended to make them unique
$CDB::XDREAD_32  = 0x7F0003;
$CDB::XDWRITE_32 = 0x7F0004;

##########################################################################

=head2 C<new()>

=over 4

=item * Description

Creates a new CDB object.

=item * Arguments

=over 4

=item 1 Command Length
 (Required)

The byte size of the CDB to create. Can be 6, 10, 12, or 16 bytes.
Variable sized CDBs are not currently supported.

=back

=item * Returns

CDB object. Caller should check for errors in the object.

=back

=cut

##########################################################################
sub new {
    my($class, $command_length) = @_;
    my $self = {};
    bless($self, $class);

    $self->{command_length} = $command_length;

    if ($self->{command_length} != 4 &&
        $self->{command_length} != 6 &&
        $self->{command_length} != 10 &&
        $self->{command_length} != 12 &&
        $self->{command_length} != 16 &&
        $self->{command_length} != 32) {
        $self->_set_err("CDB:new parameter \"command_length\" value of $self->{command_length} is not currently supported.");
        return $self;
    }

    $self->{cdb} = [];
    return $self;
}

##########################################################################

=head2 C<get_cdb_string()>

=over 4

=item * Description

Get the ASCII representation of the CDB hex byte string.

=item * Arguments

=over 4

=item 1 Spaces
 (Optional)

A boolean value that specifies whether to put a space between each byte
for readability.

=back

=item * Returns

The CDB hex byte string in ASCII.

=back

=cut

##########################################################################
sub get_cdb_string {
    my($self,$space) = @_;
    my $str;

    if (defined($space)) {
        $space = " ";
    }

    for (my $byte=0; $byte<$self->{command_length}; $byte++) {
        $str .= sprintf("%02x%s", $self->{cdb}[$byte], $space);
    }
    return $str;
}

##########################################################################

=head2 C<set_operation()>

=over 4

=item * Description

Set the SCSI op code.

=item * Arguments

=over 4

=item 1 Operation
 (Required)

The SCSI byte op code. This object has some global definitions to simplify
the specification of the op code. No check is made to the validity of the
op code.

=back

=item * Returns

$ClientUtil::TRUE

=back

=cut

##########################################################################
sub set_operation {
    my($self, $operation) = @_;

    $self->{cdb}[0] = $self->{operation} = $operation;

    # The following commands have Service Action codes appended to them 
    # in the global definition
    if ($self->{operation} == $CDB::XDREAD_32 ||
        $self->{operation} == $CDB::XDWRITE_32 ) {
        # Set the ADDITIONAL CDB LENGTH field
        $self->{cdb}[7] = 0x18;
        
        # Set the SERVICE ACTION field
        $self->{cdb}[8] = &_byte_shift_right( $self->{operation} & 0x00FF00, 2 );
        $self->{cdb}[9] = $self->{operation} & 0x0000FF;

        # Set the OPERATION CODE field
        $self->{cdb}[0] = &_byte_shift_right( $self->{operation} & 0xFF0000, 4 );
    }

    return $ClientUtil::TRUE;
}

##########################################################################

=head2 C<set_bitoffset_in_byte()>

=over 4

=item * Description

Set the bit to 1 in the given byte 

=item * Arguments

=over 4

=item 1 byte_offset 
 (Required)

=item 2 bit_offset 
 (Required)

=item 3 num_of_bits
 (Optional)

=item 3 value
 (Optional)

Set the num of bits (default is 1 bit) starting at bit_offset and counting down, in the byte at byte_offset, to value (default is 1).

=back

=item * Returns

$ClientUtil::TRUE

=back

=cut

##########################################################################
sub set_bitoffset_in_byte {
    my ($self, %args) = @_;

    my ($byte_offset, $bit_offset, $num_of_bits, $cdb_param, $length);

    $cdb_param = "no";

    foreach my $param (keys %args) {
        if ($param =~ /^-?byte_offset/) {
            $byte_offset = $args{$param};
            next;
        } elsif ($param =~ /^-?bit_offset/) {
            $bit_offset = $args{$param};
            next;
        } elsif ($param =~ /^-?num_of_bits/) {
            $num_of_bits = $args{$param};
            next;
        } elsif ($param =~ /^-?cdb_param/) {
            $cdb_param = $args{$param};
            next;
        } else {
            $self->_set_err("Invalid argument passed");
            return $ClientUtil::FALSE;
        }
    }

    if (lc($cdb_param) eq "yes") {
        $length = $self->{command_param_length};
    } else {
        $length = $self->{command_length};
    }

    $num_of_bits = 1 unless ($num_of_bits > 0 and $num_of_bits < 9);
    $byte_offset = 0 unless ($byte_offset < $length);
    $bit_offset  = 0 unless ($bit_offset >= 0 and $bit_offset < 8);

    # create a bit mask so that only the specified number of bits in that
    # byte will be update irrespective of the value. The bits to be updated
    # are assumed to be in the descending offsets.

    my $bit_mask;

    for my $idx (0 .. $num_of_bits - 1) {
        $bit_mask |= 1 << ($bit_offset - $idx);
    }

    if (lc($cdb_param) eq "yes") {
        $self->{cmd_param}[$byte_offset] |= $bit_mask;
    } else {
        $self->{cdb}[$byte_offset] |= $bit_mask;
    }

    return $ClientUtil::TRUE;

}

##########################################################################

=head2 C<set_byte_value()>

=over 4

=item * Description

Set the byte offset to the given value 

=item * Arguments

=over 4

=item 1 byte_offset 
 (Required)

=item 2 required
 (Required)

Set the byte offset in the CDB to the given value. The 
value is checked to be less than 255.

=back

=item * Returns

$ClientUtil::TRUE

=back

=cut

##########################################################################
sub set_byte_value {
    my ($self, %args) = @_;

    my ($byte_offset, $value, $cdb_param,);

    $cdb_param = "no";

    foreach my $param (keys %args) {
        if ($param =~ /^-?byte_offset/) {
            $byte_offset = $args{$param};
            next;
        } elsif ($param =~ /^-?value/) {
            $value = $args{$param};
            next;
        } elsif ($param =~ /^-?cdb_param/) {
            $cdb_param = $args{$param};
            next;
        } else {
            $self->_set_err("Invalid argument passed");
            return $ClientUtil::FALSE;
        }
    }

    $value = 0 unless ($value);
 
	if (lc($cdb_param) eq "yes") {
        $byte_offset = 0 unless ($byte_offset < $self->{command_param_length});
		$self->{cmd_param}[$byte_offset] = $value & 0xFF;
    } else {
        $byte_offset = 0 unless ($byte_offset < $self->{command_length});
        $self->{cdb}[$byte_offset] = $value & 0xFF;
    }

    return $ClientUtil::TRUE;
}

##########################################################################

=head2 C<set_address()>

=over 4

=item * Description

Set the logical block address.

=item * Arguments

=over 4

=item 1 Address
 (Required)

The SCSI Logical Block Address (LBA). The address has a limitation based 
on the byte size of the CDB.

=back

=item * Returns

$ClientUtil::TRUE  If the address was successfully set.

$ClientUtil::FALSE If the address exceeds the CDB's allowable size.

=back

=cut

##########################################################################
sub set_address {
    my($self, $address) = @_;

    $self->{address} = $address;

    if ($self->{command_length} == 6 ) {
        if ($self->{address} > 0x1FFFF) {
            $self->_set_err("CDB:set_address address value exceeds allowable range for operation");
            return $ClientUtil::FALSE;
        }
       $self->{cdb}[1] = &_byte_shift_right( $self->{address} & 0x10000, 4 );
       $self->{cdb}[2] = &_byte_shift_right( $self->{address} & 0x0FF00, 2 );
       $self->{cdb}[3] = $self->{address} & 0x000FF;

    } elsif ($self->{command_length} == 10 || $self->{command_length} == 12) {
        if ($self->{address} > 0xFFFFFFFF) {
            $self->_set_err("CDB:set_address address value exceeds allowable range for operation");
            return $ClientUtil::FALSE;
        }
       $self->{cdb}[2] = &_byte_shift_right( $self->{address} & 0xFF000000, 6 );
       $self->{cdb}[3] = &_byte_shift_right( $self->{address} & 0x00FF0000, 4 );
       $self->{cdb}[4] = &_byte_shift_right( $self->{address} & 0x0000FF00, 2 );
       $self->{cdb}[5] = $self->{address} & 0x000000FF;

    } elsif ($self->{command_length} == 16) {
       if ($self->{address} > 0xFFFFFFFFFFFFFFFF) {
           $self->_set_err("CDB:set_address address value exceeds allowable range for operation");
           return $ClientUtil::FALSE;
       }
       $self->{cdb}[2] = &_byte_shift_right( $self->{address} & 0xFF00000000000000, 14 );
       $self->{cdb}[3] = &_byte_shift_right( $self->{address} & 0x00FF000000000000, 12 );
       $self->{cdb}[4] = &_byte_shift_right( $self->{address} & 0x0000FF0000000000, 10 );
       $self->{cdb}[5] = &_byte_shift_right( $self->{address} & 0x000000FF00000000, 8 );
       $self->{cdb}[6] = &_byte_shift_right( $self->{address} & 0x00000000FF000000, 6 );
       $self->{cdb}[7] = &_byte_shift_right( $self->{address} & 0x0000000000FF0000, 4 );
       $self->{cdb}[8] = &_byte_shift_right( $self->{address} & 0x000000000000FF00, 2 );
       $self->{cdb}[9] = $self->{address} & 0x00000000000000FF;

    } elsif ($self->{command_length} == 32) {
       if ($self->{address} > 0xFFFFFFFFFFFFFFFF) {
           $self->_set_err("CDB:set_address address value exceeds allowable range for operation");
           return $ClientUtil::FALSE;
       }
       $self->{cdb}[12] = &_byte_shift_right( $self->{address} & 0xFF00000000000000, 14 );
       $self->{cdb}[13] = &_byte_shift_right( $self->{address} & 0x00FF000000000000, 12 );
       $self->{cdb}[14] = &_byte_shift_right( $self->{address} & 0x0000FF0000000000, 10 );
       $self->{cdb}[15] = &_byte_shift_right( $self->{address} & 0x000000FF00000000, 8 );
       $self->{cdb}[16] = &_byte_shift_right( $self->{address} & 0x00000000FF000000, 6 );
       $self->{cdb}[17] = &_byte_shift_right( $self->{address} & 0x0000000000FF0000, 4 );
       $self->{cdb}[18] = &_byte_shift_right( $self->{address} & 0x000000000000FF00, 2 );
       $self->{cdb}[19] = $self->{address} & 0x00000000000000FF;
    } else {
        $self->_set_err("CDB:set_address unknown command length");
        return $ClientUtil::FALSE;

    }
    return $ClientUtil::TRUE;
}

##########################################################################

=head2 C<set_transfer_length()>

=over 4

=item * Description

Set the transfer byte length.

=item * Arguments

=over 4

=item 1 Length
 (Required)

The SCSI transfer length in bytes. The transfer length has a limitation 
based on the byte size of the CDB.

=back

=item * Returns

$ClientUtil::TRUE  If the transfer length was successfully set.

$ClientUtil::FALSE If the transfer length exceeds the CDB's allowable
size.

=back

=cut

##########################################################################
sub set_transfer_length {
    my($self, $transfer_length) = @_;

    $self->{transfer_length} = $transfer_length;

    if ($self->{transfer_length} % 512) {
        $self->_set_err("CDB::set_transfer_length value must be a multiple of 512");
        return $ClientUtil::FALSE;
    } else {
        $self->{transfer_length} = $self->{transfer_length} / 512;
    }

    if ($self->{command_length} == 6 ) {
        if ($self->{transfer_length} > 0xFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[4] = $self->{transfer_length};

    } elsif ($self->{command_length} == 10 ) {
        if ($self->{transfer_length} > 0xFFFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[7] = &_byte_shift_right( $self->{transfer_length} & 0xFF00, 2 );
        $self->{cdb}[8] = $self->{transfer_length} & 0x00FF;

    } elsif ($self->{command_length} == 12 ) {
        if ($self->{transfer_length} > 0xFFFFFFFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[6] = &_byte_shift_right( $self->{transfer_length} & 0xFF000000, 6 );
        $self->{cdb}[7] = &_byte_shift_right( $self->{transfer_length} & 0x00FF0000, 4 );
        $self->{cdb}[8] = &_byte_shift_right( $self->{transfer_length} & 0x0000FF00, 2 );
        $self->{cdb}[9] = $self->{transfer_length} & 0x000000FF;

    } elsif ($self->{command_length} == 16) {
        if ($self->{transfer_length} > 0xFFFFFFFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[10] = &_byte_shift_right( $self->{transfer_length} & 0xFF000000, 6 );
        $self->{cdb}[11] = &_byte_shift_right( $self->{transfer_length} & 0x00FF0000, 4 );
        $self->{cdb}[12] = &_byte_shift_right( $self->{transfer_length} & 0x0000FF00, 2 );
        $self->{cdb}[13] = $self->{transfer_length} & 0x000000FF;

    } elsif ($self->{command_length} == 32) {
        if ($self->{transfer_length} > 0xFFFFFFFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[28] = &_byte_shift_right( $self->{transfer_length} & 0xFF000000, 6 );
        $self->{cdb}[29] = &_byte_shift_right( $self->{transfer_length} & 0x00FF0000, 4 );
        $self->{cdb}[30] = &_byte_shift_right( $self->{transfer_length} & 0x0000FF00, 2 );
        $self->{cdb}[31] = $self->{transfer_length} & 0x000000FF;

    } else {
        $self->_set_err("CDB:set_address unknown command length");
        return $ClientUtil::FALSE;

    }
    return $ClientUtil::TRUE;

}

##########################################################################

=head2 C<set_pr_out_cdb()>

=over 4

=item * Description

Creates the CDB for persistent reservation command pr_out

=item * Arguments

=over 4

=item 1 Service Action
 (Required)

Service requested by the initiator

=item 2 Persistent Reservation Type
 (Required)

Reservation type requested by the initiator

=item 3 Parameter List Length
 (Optional)

The value is set to 0x18 by default. This parameter
should be set to specify the length if additional
parameter data is present

=item 4 Control Byte
 (Optional)

Default: set to zero

=back

=item * Returns

None

=back

=cut

##########################################################################

sub set_pr_out_cdb {
    my ($self, %args) = @_;
    my ($opcode, $service_action, $pr_type, $param_lst_len, $control);

    # set the default values
    $opcode         = 0x5F;
    $service_action = 0;
    $pr_type        = 0;
    $param_lst_len  = 0x18;    # without additional parameter data

    foreach my $param (keys %args) {
        if ($param =~ /^-?service_action/) {
            $service_action = $args{$param};
            next;
        } elsif ($param =~ /^-?pr_type/) {
            $pr_type = $args{$param};
            next;
        } elsif ($param =~ /^-?param_lst_len/) {
            $param_lst_len = $args{$param};
            next;
        } elsif ($param =~ /^-?control/) {
            $control = $args{$param};
            next;
        } else {
            $self->_set_err("Invalid argument passed");
            return $ClientUtil::FALSE;
        }
    }

    $self->{command_param_length} = $param_lst_len;

    $self->set_byte_value(-byte_offset => 0,
                          -value       => $opcode,);

    $self->set_byte_value(-byte_offset => 1,
                          -value       => $service_action,);

    $self->set_byte_value(-byte_offset => 2,
                          -value       => $pr_type,);

    $self->set_byte_value(-byte_offset => 3,
                          -value       => 0,);

    $self->set_byte_value(-byte_offset => 4,
                          -value       => 0,);

    $self->set_byte_value(-byte_offset => 5,
                          -value       => 0,);

    $self->set_byte_value(-byte_offset => 6,
                          -value       => 0);

    $self->set_byte_value(-byte_offset => 7,
                          -value       => 0,);

    $self->set_byte_value(-byte_offset => 8,
                          -value       => $param_lst_len,
                         );    # check the offset to be specified

    $self->set_byte_value(-byte_offset => 9,
                          -value       => $control,);

    return $ClientUtil::TRUE;
}

##########################################################################

=head2 C<set_pr_out_cmd_param_list()>

=over 4

=item * Description

Creates the command parameter list for persistent reservation command pr_out

=item * Arguments

=over 4

=item 1 Reservation Key
 (Required)

Key used by the intiator to make a reservation ( 8 bytes)
Should be of the form xx:xx:xx:xx:xx:xx:xx:xx 

=item 2 Service Action Key
 (Required)

Key used by the intiator during registration (8 bytes)
Should be of the form xx:xx:xx:xx:xx:xx:xx:xx

=item 3 spec_i_pt
 (optional)

 Default: 0

Specify Initiator Ports

=item 3 ALL_TG_PT
 (optional)

 Default: 0

All target ports

=item 5 APTPL 
 (optional)

=item 6 Transportid parameter data length
 (optional)

 Default: 0

This value has to be specified when spec_i_pt is set

=item 7 Transport ids
 (optional)

This value has to be specified when spec_i_pt is set

=back

=item * Returns

None

=back

=cut

##########################################################################
sub set_pr_out_cmd_param_list {
    my ($self, %args) = @_;
    my ($service_action, $pr_type,         $param_lst_len,
        $control,        $reservation_key, $service_action_key,
        $spec_i_pt,      $all_tg_pt,       $aptpl,
        $tlist_length,   @trans_ids);

    # set the default value
    $tlist_length = 0;
    $spec_i_pt    = 0;
    $all_tg_pt    = 0;
    $aptpl        = 0;

    foreach my $param (keys %args) {
        if ($param =~ /^-?reservation_key/) {
            $reservation_key = $args{$param};
            next;
        } elsif ($param =~ /^-?sa_key/) {
            $service_action_key = $args{$param};
            next;
        } elsif ($param =~ /^-?spec_i_pt/) {
            $spec_i_pt = $args{$param};
            next;
        } elsif ($param =~ /^-?all_tg_pt/) {
            $all_tg_pt = $args{$param};
            next;
        } elsif ($param =~ /^-?aptpl/) {
            $aptpl = $args{$param};
            next;
        } elsif ($param =~ /^-?tlist_length/) {
            $tlist_length = $args{$param};
            next;
        } elsif ($param =~ /^-?tids/) {
            @trans_ids = @{$args{$param}};
            next;
        } else {
            $self->_set_err("Invalid argument passed");
            return $ClientUtil::FALSE;
        }
    }

    if ($spec_i_pt) {
        if ($tlist_length == 0) {
            $self->_set_err("Specify valid transport id parameter data length");
            return $ClientUtil::FALSE;
        }

        if (scalar(@trans_ids) != $tlist_length) {
            $self->_set_err(
                "Number of transport ids specified doesnt match the list length"
            );
            return $ClientUtil::FALSE;
        }
    }

    if ($reservation_key !~ /(\w+):(\w+):(\w+):(\w+):(\w+):(\w+):(\w+):(\w+)/) {
        $self->_set_err(
            "Invalid reservation key. Should be of the type xx:xx:xx:xx:xx:xx:xx:xx"
        );
        return $ClientUtil::FALSE;
    }

    if ($service_action_key !~
        /(\w+):(\w+):(\w+):(\w+):(\w+):(\w+):(\w+):(\w+)/) {
        $self->_set_err(
            "Invalid service action key. Should be of the type xx:xx:xx:xx:xx:xx:xx:xx"
        );
        return $ClientUtil::FALSE;
    }

    my @res_key_ele = split(":", $reservation_key);
    my @sa_key_ele  = split(":", $service_action_key);

    my $ind;

    for ($ind = 0 ; $ind < 8 ; $ind++) {
		my $val1 = "0x".$res_key_ele[$ind];

        $self->set_byte_value(-byte_offset => $ind,
                              -value       => $val1,
                              -cdb_param   => "yes",);

		my $val2 = "0x".$sa_key_ele[$ind];

		$self->set_byte_value(-byte_offset => (8 + $ind),
                              -value       => $val2,
                              -cdb_param   => "yes",);
    }

    for ($ind = 0 ; $ind < 4 ; $ind++) {
        $self->set_byte_value(-byte_offset => (16 + $ind),
                              -value       => 0,
                              -cdb_param   => "yes",);
    }

    $self->set_bitoffset_in_byte(-bit_offset  => 0,
                                 -byte_offset => 20,
                                 -num_of_bits => 1,
                                 -cdb_param   => "yes",) if ($aptpl == 1);

    $self->set_bitoffset_in_byte(-bit_offset  => 2,
                                 -byte_offset => 20,
                                 -num_of_bits => 1) if ($all_tg_pt == 1);

    $self->set_bitoffset_in_byte(-bit_offset  => 3,
                                 -byte_offset => 20,
                                 -num_of_bits => 1) if ($spec_i_pt == 1);

    if (($aptpl == 0) && ($all_tg_pt == 0) && ($spec_i_pt == 0)) {
        $self->set_byte_value(-byte_offset => 20,
                              -value       => 0,
                              -cdb_param   => "yes",);
    }

    $self->set_byte_value(-byte_offset => 21,
                          -value       => 0,
                          -cdb_param   => "yes",);

    $self->set_byte_value(-byte_offset => 22,
                          -value       => 0,
                          -cdb_param   => "yes",);

    $self->set_byte_value(-byte_offset => 23,
                          -value       => 0,
                          -cdb_param   => "yes",);

    if ($spec_i_pt) {
        $self->set_byte_value(-byte_offset => 24,
                              -value       => 0,
                              -cdb_param   => "yes",);

        $self->set_byte_value(-byte_offset => 25,
                              -value       => 0,
                              -cdb_param   => "yes",);

        $self->set_byte_value(-byte_offset => 26,
                              -value       => 0,
                              -cdb_param   => "yes",);

        $self->set_byte_value(-byte_offset => 27,
                              -value       => $tlist_length,
                              -cdb_param   => "yes",);

        for (my $ind = 0 ; $ind < $tlist_length ; $ind++) {
            $self->set_byte_value(-byte_offset => (28 + $ind),
                                  -value       => $trans_ids[$ind],
                                  -cdb_param   => "yes",);
        }
    }

    return $ClientUtil::TRUE;

}

##########################################################################

=head2 C<get_cdb_param_string()>

=over 4

=item * Description

Get the ASCII representation of the CDB hex byte string for 
Persistent Reservation CDB.

=item * Arguments

=over 4

=item 1 Spaces
 (Optional)

A boolean value that specifies whether to put a space between each byte
for readability.

=back

=item * Returns

The CDB hex byte string in ASCII.

=back

=cut

##########################################################################

sub get_cdb_param_string {
    my ($self, $space) = @_;
    my $str;

    if (defined($space)) {
        $space = " ";
    }

    for (my $byte = 0 ; $byte < $self->{command_param_length} ; $byte++) {
        $str .= sprintf("%2x$space", $self->{cmd_param}[$byte], $space);
    }

    return $str;
}

=head2 C<set_unmap_cdb()>

=over 4

=item * Description

Creates the command parameter list for unmap

=item * Arguments

=over 4

=item 1 Parameter List Length
 (Required)

This specifies the length in bytes of the UNMAP parameter data that
shall be sent from the application client to the device server

=item 2 Group Number
 (Required)

Group Number field specifies the group into which attributes
associated with the command should be collected.

=back

=item * Returns

None

=back

=cut

##########################################################################

sub set_unmap_cdb {
    my ($self, %args) = @_;
    my ($opcode,
        $param_lst_len,
        $group_number,
       );

    # set the default value
    $opcode = 0x42;
    $param_lst_len = 0;
    $group_number = 0;

    foreach my $param (keys %args) {
        if($param =~ /^-?param_list_len/){
           $param_lst_len = $args{$param};
           next;
        }elsif($param =~ /^-?group_number/){
           $group_number = $args{$param};
           next;
        }else {
            $self->_set_err("Invalid argument passed ".$param);
            return $ClientUtil::FALSE;
        }
    }

    $self->set_byte_value(-byte_offset => 0,
                          -value       => $opcode,);

    for(my $ind=0; $ind < 5; $ind++){
        $self->set_byte_value(-byte_offset => 1 + $ind,
                              -value       => 0,);
    }

    $self->set_byte_value(-byte_offset => 6,
                          -value       => $group_number,);


    $self->{cdb}[7] = ($param_lst_len & BYTE_1) >> 8;
    $self->{cdb}[8] = ($param_lst_len & BYTE_0);

    $self->set_byte_value(-byte_offset => 9,
                          -value       => 0,);

    return $ClientUtil::TRUE;
}

##########################################################################

=head2 C<set_unmap_cmd_param_list()>

=over 4

=item * Description

Creates the command parameter list for scsi unmap command

=item * Arguments

=over 4

=item 1 Unmap Data Length
 (Required)

The UNMAP DATA LENGTH field specifies the length in bytes of the following
data that is available to be transferred from the data-out buffer.

=item 2 Block Descriptor Data Length
 (Required)

This field specifies the length in bytes of the UNMAP block descriptor
data that is available to be transferred from the data-out buffer

=item 3 Num of Descriptors
 (optional)

 Default: 0

Number of descriptors to unmap

=item 3 Descriptor Hash
(Required)

Contains the lba and the number of blocks to be unmapped for each descriptor

=back

=item * Returns

None

=back

=cut

##########################################################################
sub set_unmap_cmd_param_list {
    my ($self, %args) = @_;
    my ($unmap_data_length,
        $blk_desc_dl,
        $num_of_descriptors,
        $desc_hash_ref,
       );

    # set the default value
    $num_of_descriptors = 0;
    $blk_desc_dl        = 0;
    $unmap_data_length  = 0;

    foreach my $param (keys %args) {
        if($param =~ /^-?unmap_dl/){
            $unmap_data_length = $args{$param};
            next;
        }elsif($param =~ /^-?blk_desc_dl/){
            $blk_desc_dl = $args{$param};
            next;
        }elsif ($param =~ /^-?num_of_descriptors/) {
            $num_of_descriptors = $args{$param};
            next;
        } elsif ($param =~ /^-?descriptor_hash/) {
            $desc_hash_ref = $args{$param};
            next;
        } else {
            $self->_set_err("Invalid argument passed: ". $param);
            return $ClientUtil::FALSE;
        }
    }

    $self->{command_param_length} = ($num_of_descriptors * 16) + 8;

    # UNMAP Parameter list header
    $self->{cmd_param}[0] = ($unmap_data_length & BYTE_1) >> 8 ;
    $self->{cmd_param}[1] = ($unmap_data_length & BYTE_0);

    $self->{cmd_param}[2] = ($blk_desc_dl & BYTE_1) >> 8 ;
    $self->{cmd_param}[3] = ($blk_desc_dl & BYTE_0);

    my $ind;

    for ($ind = 0 ; $ind < 4 ; $ind++) {
        $self->set_byte_value(-byte_offset => (4 + $ind),
                              -value       => 0,
                              -cdb_param   => "yes",);
    }

    # UNMAP block descriptor data

    my $offset = 0;

    for($ind = 1; $ind <= $num_of_descriptors; $ind++) {
        my $lba  = $desc_hash_ref->{$ind}->{lba};
        my $blks = $desc_hash_ref->{$ind}->{no_of_blocks};

        unless(defined $lba){
            $self->_set_err("LBA Value not defined");
            return $ClientUtil::FALSE;
        }

        unless(defined $blks){
            $self->_set_err("Number of blocks value not defined");
            return $ClientUtil::FALSE;
        }

        $self->{cmd_param}[8  + $offset] = ($lba & BYTE_7) >> 56;
        $self->{cmd_param}[9  + $offset] = ($lba & BYTE_6) >> 48;
        $self->{cmd_param}[10 + $offset] = ($lba & BYTE_5) >> 40;
        $self->{cmd_param}[11 + $offset] = ($lba & BYTE_4) >> 32;
        $self->{cmd_param}[12 + $offset] = ($lba & BYTE_3) >> 24;
        $self->{cmd_param}[13 + $offset] = ($lba & BYTE_2) >> 16;
        $self->{cmd_param}[14 + $offset] = ($lba & BYTE_1) >> 8;
        $self->{cmd_param}[15 + $offset] = ($lba & BYTE_0);

        $self->{cmd_param}[16 + $offset] = ($blks & BYTE_3) >> 24;
        $self->{cmd_param}[17 + $offset] = ($blks & BYTE_2) >> 16;
        $self->{cmd_param}[18 + $offset] = ($blks & BYTE_1) >> 8;
        $self->{cmd_param}[19 + $offset] = ($blks & BYTE_0);

        for (my $ind1 = 20 ; $ind1 < 24 ; $ind1++) {
            $self->set_byte_value(-byte_offset => (20 + $offset + $ind1),
                                  -value       => 0,
                                  -cdb_param   => "yes",);
        }

        $offset += 16;
    }

    return $ClientUtil::TRUE;
}


##########################################################################

=head2 C<set_hole_punch_cdb()>

=over 4

=item * Description

Creates the CDB for SCSI PUNCH HOLE command

=item * Arguments

=over 4

=item 1 Number of descriptors
 (Required)

This specifies the number of descriptors to be passed to
the scsi PUNCH HOLE command

=back

=item * Returns

None

=back

=cut

##########################################################################

sub set_hole_punch_cdb {
    my ($self, %args) = @_;
    my ($opcode,$num_of_descriptors);

    # set the default value
    $opcode = 0xC1;

    foreach my $param (keys %args) {
        if($param =~ /^-?num_of_descriptors/){
           $num_of_descriptors = $args{$param};
           next;
        }else {
            $self->_set_err("Invalid argument passed ".$param);
            return $ClientUtil::FALSE;
        }
    }

    $self->set_byte_value(-byte_offset => 0,
                          -value       => $opcode,);

    $self->{cdb}[1] = ($num_of_descriptors & BYTE_1) >> 8;
    $self->{cdb}[2] = ($num_of_descriptors & BYTE_0);

    $self->set_byte_value(-byte_offset => 3,
                          -value       => 0,);

    return $ClientUtil::TRUE;
}

##########################################################################

=head2 C<set_hole_punch_param_list()>

=over 4

=item * Description

Creates the command parameter list for PUNCH HOLE command

=item * Arguments

=over 4

=item 3 Descriptor Hash
(Required)

Contains the lba and the number of blocks to be unmapped for each descriptor

=back

=item * Returns

None

=back

=cut

##########################################################################
sub set_hole_punch_param_list {
    my ($self, %args) = @_;
    my ($num_of_descriptors, $desc_hash_ref,
       );

    foreach my $param (keys %args) {
        if($param =~ /^-?num_of_descriptors/) {
            $num_of_descriptors = $args{$param};
            next;
        } elsif ($param =~ /^-?descriptor_hash/) {
            $desc_hash_ref = $args{$param};
            next;
        } else {
            $self->_set_err("Invalid argument passed: ". $param);
            return $ClientUtil::FALSE;
        }
    }

    $self->{command_param_length} = ($num_of_descriptors * 12);

    # UNMAP block descriptor data

    my $offset = 0;

    for(my $ind = 1; $ind <= $num_of_descriptors; $ind++) {
        my $lba  = $desc_hash_ref->{$ind}->{lba};
        my $blks = $desc_hash_ref->{$ind}->{no_of_blocks};

        unless(defined $lba){
            $self->_set_err("LBA Value not defined");
            return $ClientUtil::FALSE;
        }

        unless(defined $blks){
            $self->_set_err("Number of blocks value not defined");
            return $ClientUtil::FALSE;
        }

        $self->{cmd_param}[0 + $offset] = ($lba & BYTE_7) >> 56;
        $self->{cmd_param}[1 + $offset] = ($lba & BYTE_6) >> 48;
        $self->{cmd_param}[2 + $offset] = ($lba & BYTE_5) >> 40;
        $self->{cmd_param}[3 + $offset] = ($lba & BYTE_4) >> 32;
        $self->{cmd_param}[4 + $offset] = ($lba & BYTE_3) >> 24;
        $self->{cmd_param}[5 + $offset] = ($lba & BYTE_2) >> 16;
        $self->{cmd_param}[6 + $offset] = ($lba & BYTE_1) >> 8;
        $self->{cmd_param}[7 + $offset] = ($lba & BYTE_0);

        $self->{cmd_param}[8  + $offset] = ($blks & BYTE_3) >> 24;
        $self->{cmd_param}[9  + $offset] = ($blks & BYTE_2) >> 16;
        $self->{cmd_param}[10 + $offset] = ($blks & BYTE_1) >> 8;
        $self->{cmd_param}[11 + $offset] = ($blks & BYTE_0);

        $offset += 12;
    }

    return $ClientUtil::TRUE;
}

#### P r i v a t e    M e t h o d s ####

=pod

=head1 PRIVATE METHODS

=cut

##########################################################################

=head2 C<_byte_shift_right()>

=over 4

=item * Description

Shifts an integer to the right by the specified byte count.

=item * Arguments

=over 4

=item 1 Value
 (Required)

The value to shift.

=item 2 Count
 (Required)

The number of bytes to shift to the right.

=back

=item * Returns

The shifted value.

=back

=cut

##########################################################################
sub _byte_shift_right {
    my ($value, $count) = @_;

    return ( $value >> ($count << 2) );
}

=pod

=head1 AUTHOR

Evan Chiu (evan@netapp.com)

=head1 COPYRIGHT

Copyright (c) 2008 Network Appliance, Inc. All rights reserved.

=cut
