#$Id: $
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide helper routines for sending Compare and Write SCSI commands
## @author dsharma@netapp.com

package San::ScsiTool::ScsiCAW;

use strict;
use warnings;
use San::ScsiTool;
use San::ScsiTool::CDB;
use San::SanLog;
use Error;
use Params::Validate qw( :all );
require Exporter;

use vars qw(@ISA);

use constant {BLOCK_SIZE          => 512,
              SCSI_CAW_CDB_LENGTH => 16
             };
#
# Inherit from the Error, SanLog classes
#
our @ISA = qw(Error SanLog Exporter);

sub new {
    my ($caller) = @_;
    my $class = ref($caller) || $caller;

    #
    # Create a hash to store attributes and bless the hash to create an object.
    #
    my $self = {};
    bless $self, $class;

    # add the scsi tool obj debug level.
    #
    $self->set_debug_level(3);

    return $self;
}

####################          set_caw_data                ####################
# Description: The Compare and Write command will send 2 data buffers.
#              The first data buffer will be used to verify the contents on
#              the disk. The contents of the second data buffer will be
#              written to the disk at the specified LBA after successful
#              verification. In this subroutine we will create the data buffers
#              or the pattern to be sent with the CAW scsi command
# Returns: $data_pat - The pattern to be passed to send_cmd while executing CAW
#          TCD::FAIL - Test Failed
###############################################################################

sub set_caw_data {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Validate Params

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                       {blocks           => { type      => SCALAR,
                                              required  => 1},
                        verify_pattern   => { type      => SCALAR,
                                              required  => 1},
                        write_pattern    => { type      => SCALAR,
                                              required  => 1},
                        insert_error     => { type      => SCALAR,
                                              optional  => 1,
                                              default   => "no"},
                        error_offset     => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 1},
                       }
                  );
    };

    if ($@) {
        $self->_set_err($@);
        return $self;
    }

    my $no_of_blocks          = $args{'blocks'};
    my $verification_pattern  = $args{'verify_pattern'};
    my $write_pattern         = $args{'write_pattern'};
    my $error                 = $args{'insert_error'};
    my $error_offset          = $args{'error_offset'};

    my $data_pat = "";

    # Write first block with verification pattern followed by data for write

    for (my $i = 1 ; $i <= ($no_of_blocks * BLOCK_SIZE) ; $i++) {
        if ($i <= ($no_of_blocks / 2) * BLOCK_SIZE) {
            if ($error =~ /yes/i) {
                if ($i == $error_offset) {
                    $data_pat .= "E";
                    next;
                }
            }

            $data_pat .= $verification_pattern;
        } else {
            $data_pat .= $write_pattern;
        }
    }

    return $data_pat;
}

####################             scsi_caw                  ####################
# Description: Execute the CAW (VAW) command
###############################################################################

sub scsi_caw {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Validate Params

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                       {st_obj    => { type     => OBJECT,
                                       required => 1,
                                       callbacks => { 'Object Type' =>
                                                       sub { $_[0] =~ /San::ScsiTool/}
                                                    }
                                     },
                        device           => { type      => SCALAR,
                                              required  => 1},
                        lba              => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 0},
                        no_of_blocks     => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 1},
                        expected_result  => { type      => SCALAR,
                                              optional  => 1,
                                              default   => "GOOD"},
                        sense_key        => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 0},
                        ascnq            => { type      => SCALAR,
                                              optional  => 1,
                                              default   => "0000"},
                        buf_len          => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 512},
                        data_pat         => { type      => SCALAR,
                                              optional  => 1,
                                              default   => ""},
                        vrprotect        => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 0},
                        dpo              => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 0},
                        fua              => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 0},
                        fua_nv           => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 0},
                        set_reserve_bits => { type      => BOOLEAN,
                                              optional  => 1,
                                              default   => 0},
                        group_number     => { type      => SCALAR,
                                              optional  => 1,
                                              default   => 0},
                       }
                  );
    };

    if ($@) {
        $self->_set_err($@);
        return 0;
    }

    my $op_code       = 0x89;
    my $scsi_tool_obj =  $args{'st_obj'};
    my $device        =  $args{'device'};
    my $buf_len       =  $args{'buf_len'};
    my $pattern       =  $args{'data_pat'};
    my $vrprotect     =  $args{'vrprotect'};
    my $dpo           =  $args{'dpo'};
    my $fua           =  $args{'fua'};
    my $fua_nv        =  $args{'fua_nv'};
    my $reserved_bits =  $args{'set_reserve_bits'};
    my $group_number  =  $args{'group_number'};
    my $lba           =  $args{'lba'};
    my $no_of_blocks  =  $args{'no_of_blocks'};
    my $sense_key     =  $args{'sense_key'};
    my $ascnq         =  $args{'ascnq'};
    my $scsi_status   =  $args{'expected_result'};

    my $cdb_ref = new CDB(SCSI_CAW_CDB_LENGTH);

    $cdb_ref->set_operation($op_code);

    $cdb_ref->set_bitoffset_in_byte(-bit_offset  => 1,
                                    -byte_offset => 1,
                                    -num_of_bits => 1,) if ($fua_nv);

    $cdb_ref->set_bitoffset_in_byte(-bit_offset  => 3,
                                    -byte_offset => 1,
                                    -num_of_bits => 1,) if ($fua);

    $cdb_ref->set_bitoffset_in_byte(-bit_offset  => 4,
                                    -byte_offset => 1,
                                    -num_of_bits => 1,) if ($dpo);

    $cdb_ref->set_bitoffset_in_byte(-bit_offset  => 5,
                                    -byte_offset => 1,
                                    -num_of_bits => 3,) if ($vrprotect);

    if (($fua_nv == 0) and ($fua == 0) and ($dpo == 0) and ($vrprotect == 0)) {
        $cdb_ref->set_byte_value(-byte_offset => 1,
                                 -value       => 0,);
    }

    unless ($cdb_ref->set_address($lba)) {
        $self->_set_err("unable to set the address");
        return 0;
    }

    if ($reserved_bits == 1) {
        $cdb_ref->set_bitoffset_in_byte(-bit_offset  => 0,
                                        -byte_offset => 10,
                                        -num_of_bits => 1,);

        $cdb_ref->set_bitoffset_in_byte(-bit_offset  => 5,
                                        -byte_offset => 10,
                                        -num_of_bits => 1,);

        $cdb_ref->set_bitoffset_in_byte(-bit_offset  => 8,
                                        -byte_offset => 10,
                                        -num_of_bits => 1,);

        $cdb_ref->set_bitoffset_in_byte(-bit_offset  => 0,
                                        -byte_offset => 11,
                                        -num_of_bits => 8,);
        
        $cdb_ref->set_byte_value(-byte_offset => 12,
                                 -value       => 1,);
    } else {
        $cdb_ref->set_byte_value(-byte_offset => 10,
                                 -value       => 0,);
        $cdb_ref->set_byte_value(-byte_offset => 11,
                                 -value       => 0,);
        $cdb_ref->set_byte_value(-byte_offset => 12,
                                 -value       => 0,);
    }

    $cdb_ref->set_byte_value(-byte_offset => 13,
                             -value       => $no_of_blocks,);

    if ($group_number) {
        $cdb_ref->set_bitoffset_in_byte(-bit_offset  => 3,
                                        -byte_offset => 14,
                                        -num_of_bits => 4,);
    } else {
        $cdb_ref->set_byte_value(-byte_offset => 14,
                                 -value       => $group_number,);
    }

    $cdb_ref->set_byte_value(-byte_offset => 15,
                             -value       => 0,);

    my $cdb_str = join ' ', $cdb_ref->get_cdb_string("with_spaces");

    my ($result, $data_out);

    if ($pattern) {
        $self->sanlog(-msg       => "Pattern is set and is valid");

        $result =
          $scsi_tool_obj->send_cmd(-device   => $device,
                                   -cdb      => $cdb_str,
                                   -buf_len  => $buf_len,
                                   -data_pat => $pattern,
                                   -dir      => ScsiTool::WRITE,
                                  );
    } else {
        $self->sanlog(-msg       => "No Pattern");

        $result =
          $scsi_tool_obj->send_cmd(-device   => $device,
                                   -cdb      => $cdb_str,
                                   -buf_len  => $buf_len,
                                   -dir      => ScsiTool::WRITE,
                                   );
    }

    if ($scsi_tool_obj->error) {
        $self->_set_err($scsi_tool_obj->errmsg());
        return $self;
    }

    # When ASC and ASQ values are set

    if ($ascnq !~ /0000/) {
        $self->sanlog(-msg       => "ASC and ASQ values are set");

        $scsi_tool_obj->check_status(-send_cmd_output => $result,
                                     -scsi_status     => $scsi_status,
                                     -sense_key       => $sense_key,
                                     -ascnq           => $ascnq);
    } else {

        # When ASC and ASQ values are not set

        $scsi_tool_obj->check_status(-send_cmd_output => $result,
                                     -scsi_status     => $scsi_status,
                                     -sense_key       => $sense_key,);
    }

    if ($scsi_tool_obj->error) {
        $self->_set_err(
                  "check_status after CAW failed: " . $scsi_tool_obj->errmsg());
        return 0;
    }

    return $result;
}

1;
