#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/San/ScsiTool/ScsitSimCli.pm#1 $
#$DateTime:

## @summary Provide an interface to the scsit_sim tool
## @author chadwell@netapp.com
## @status public
## @pod ../ScsiTool.pm


#
# Package Name
#
package San::ScsiTool::ScsitSimCli;

use strict;

#
# Package Requirements
#
use San::ScsiTool;
use San::SanClient::ClientUtil;

use Error;
use San::SanLog;

use vars qw( @ISA $ScsiTool $spt_loc);

@ISA = qw(Error SanLog ScsiTool);

########################################################################
# P U B L I C   M E T H O D S
########################################################################

########################################################################
# Subroutine Name: send_cmd()
# Objective:       Execute the Spt tool to send a SCSI Command
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub send_cmd{
    
    my ($self, %args)=@_;
    $self->_clear_err();
    
    $self->handle_send_cmd_params(%args);

    #
    # Build the string required to run on the target
    #
    $self->build_cmd_string();
    return $self if ($self->error());
    
    #
    # Add some log messages to keep up
    #
    $self->sanlog(-msg => ref($self)."::send_cmd()\n"   .
                          "ScsitSimCli loc     => /usr/local/test/bin/ScsitSimCli"  .
                          "cmd string   => ".$self->{sc_cmd}."\n",
                  -msg_level => 3
                 );

    #
    # executing on the client
    #
    my $cmd = $self->{sc_cmd};
    my $client_obj  = $self->{client};
    my $client_conn = $client_obj->get_connection();

    $self->sanlog(-msg =>" Execute from ScsitSimCli:".
                     "cmd = $cmd and client_conn=$client_conn",
                  -msg_level=> 3
                 ); 
                         

    my ($status, @output) = &ClientUtil::execute($client_conn, $cmd);

    if ($status > 0) {
        $self->sanlog(-msg => "Warning: ".ref($self)."::send_cmd(): " .
                              "Command ($cmd) returned a bad status ($status)."
                      -msg_level=> 3
                     );
     }

    #
    # Update the response hash with \n separated 
    # lines in an array
    #
    $self->update_rsp_hash(\@output);

    return $self if($self->error);

    #
    # Set all non-updated responses to undef()
    #
    $self->handle_unsup_responses();
 
    return $self if($self->error);

    #
    # Now return the expected parameters back to
    # the calling program.
    #
    return $self->get_exp_rsp_params();
}  # end send_cmd()


########################################################################
# Subroutine Name: update_rsp_hash()
# Objective:       Update response hash variables with information from the
#                  target so the test can access it.
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub update_rsp_hash{
    # Data returns from the target in the form-
    # offset: data line for read requests 16 bytes per line
    # ...
    # test status: success or "string err"
    # scsi result: 0x0/0x0/0x0/0x0 is status/sense key/asc/ascq
    # scsi residual: n/0x0 is over|under|none/length 
    # transfer length: 0x0 is data bytes transferred
    
    my ($self, $output_ref)   = @_;
    my @output = @$output_ref;

    #
    # first line of output is always blank
    # so get rid of it
    #
    shift @output;

    $self->_clear_err();
    
    #
    # set response hash to default
    #
    $self->build_exp_rsp_hash();
    return $self if($self->error);
 
    #
    # Get the reference to information
    # the test wants
    #
    my $er_ref = $self->get_exp_rsp_params();
    

    #
    # go through the reference hash to determine 
    # what responses to update.  
    #
    # Here I pre process the statuses line in the output array so I
    # only have to do it once. Note we don't know how many lines of
    # scsi data will be on top, so I reference the array position from
    # the bottom, since we'll always know how far up the array the values line
    # will be.
    #
    my (@statuses, @values);
    @values = split /:/, $output[$#output-2];
    @statuses = split /\//, $values[1];
    

     my (@ln_bytes, @data_bytes);
     while (my($name, $value) = each(%$er_ref)) {
         if ($value == ScsiTool::REQ_RESPONSE){
             if($name eq "SCSI_DATA"){
               #
               # take the SCSI bytes off the top
               # one line at a time
               #
               foreach my $ln (@output){
                    if($ln =~ /0x(\d+):/){
                        #
                        # the reference wants an array, 
                        # so split the string on spaces
                        #
                        @ln_bytes = split " ", $ln;
                        #
                        # the first value is the offset.
                        # we don't want it
                        #
                        shift @ln_bytes;
                        @data_bytes=(@data_bytes, @ln_bytes);
                        
                    }
                    else{
                        $er_ref->{SCSI_DATA}=\@data_bytes if(@data_bytes); 
                        last;
                    }
                 }
             }
             elsif($name eq "SCSI_STATUS"){
                         #
                         # statuses 0 is the status key
                         # remove any extra whitespace
                         #
                         $statuses[0] =~ s/ //g;
                         #
                         # require the ASC and ASCQ values to be
                         # hex digits.  Update the response hash in 
                         # the form xxh/yyh
                         #
                         my $scsi_status= sprintf"%X", $statuses[0];
                         if($scsi_status =~ /^(\d|\w)$/){$scsi_status = "0".$scsi_status;}
                         $er_ref->{SCSI_STATUS}=$self->status_string($statuses[0]);
             }
             elsif($name eq "SENSE_KEY"){
                         #
                         # conform sense_key to xxh/yyh form
                         #
                         my $sense_key= sprintf"%X", $statuses[1];
                         if($sense_key =~ /^(\d|\w)$/){$sense_key="0".$sense_key;}
                         $er_ref->{SENSE_KEY}=$sense_key."h";
             }
             elsif($name eq "ASCNQ"){
                         #
                         # conform asc and ascq 
                         # to xxh/yyh form
                         #
                         my $asc = sprintf "%X", $statuses[2];
                         my $ascq= sprintf "%X", $statuses[3];
                         if($asc =~ /^(\d|\w)$/) {$asc ='0'.$asc;}
                         if($ascq =~ /^(\d|\w)$/) {$ascq ='0'.$ascq;}
                         $er_ref->{ASCNQ}=$asc."h/".$ascq."h";
             }
             elsif($name =~ /RESID/){
                         #
                         # split the residual information returned
                         # and conform it to xxh/yyh
                         # 
                         @values = split /:/, $output[$#output-1];
                         my @residual = split /\//, $values[1];
                         if($residual[0] == "o") {$self->get_exp_rsp_params()->{RESID_OVER}="o";}
                         if($residual[0] == "u") {$self->get_exp_rsp_params()->{RESID_UNDER}="u";}
                         $er_ref->{RESID_BYTES}=$residual[1];
             }
             elsif($name eq "BYTES_XFERD"){
                 @values= split /:/, $output[$#output];
                 $er_ref->{BYTES_XFERD}=$values[1];
             }
         }
     }
    #
    # set the toolname
    #
    $er_ref->{TOOLNAME} = "scsit_sim";
    
    $self->sanlog(-msg=>"Output array string is @data_bytes\n",
                  -msg_level=> 3
                 );

    return 1;
} # end Update_rsp_hash()

########################################################################
# Subroutine Name: build_cmd_string()
# Objective:       Take incoming information from host agnostic test and
#                  build appropriate command string for the Scsi target sim
# Details:         See POD documentation in ScsiTool.pm
########################################################################

sub build_cmd_string{
    #
    # form of necessary output string is   
    # <I-T_Nexus:LUN> <buffer> [flags] <-c CDB data bytes>
    #
     my $self = shift @_;

    $self->_clear_err();

    my $sc_ref          = $self->get_send_cmd_params();

    my $dev             = $sc_ref->{DEVICE};
    my $cdb             = $sc_ref->{CDB};
    my $cdb_param       = $sc_ref->{CDB_PARAM};
    my $dir             = $sc_ref->{DIR};
    my $buf_len         = $sc_ref->{BUF_LEN};
    my $opcode          = $sc_ref->{OPCODE};
    my $qtag            = $sc_ref->{QTAG};
    my $opname          = $sc_ref->{OPCODE_NAME};
    my $repeat_cnt      = $sc_ref->{REPEAT_CNT};
    my $runtime         = $sc_ref->{RUNTIME};
    my $timeout         = $sc_ref->{TIMEOUT};
    my $on_error        = $sc_ref->{ON_ERROR};
    my $sleep           = $sc_ref->{SLEEP};
    my $data_in         = $sc_ref->{DATA_IN};
    my $data_out        = $sc_ref->{DATA_OUT};
    my $data_pat        = $sc_ref->{DATA_PAT};
    my $cmd;
   

    my @dev_array = split("\n",$dev);
    #
    # every command starts this way
    #
    if($self->{client}->hosttype() =~ /unix|vmware/i) {
        $cmd = $self->{client}->{TEST_BIN}."/scsit_test $dev_array[0]:0 ";
    }
    else {
       $self->error("Scsi tartget simulator routines must be run on a linux host.  ABORTING...");
       return $self;
    }
    
    #
    # If the test is sending a TMF
    # to the SCSIT-sim, then we
    # build something a little different
    #
    my $tmf;
    if($opcode != ScsiTool::SCSI_CDB) {
        $tmf = "at" if ($opcode & ScsiTool::ABORT_TASK);
        $tmf = "ats"if ($opcode & ScsiTool::ABORT_TASK_SET);
        $tmf = "ca" if ($opcode & ScsiTool::CLEAR_ACA);
        $tmf = "cts"if ($opcode & ScsiTool::CLEAR_TASK_SET);
        $tmf = "lr" if ($opcode & ScsiTool::LUN_RESET);
        #
        # LOGOUT, LOGIN, and KILL
        # are not implemented so far
        #
        # $tmf = "lo" if ($opcode & ScsiTool::LOGOUT);
        # $tmf = "li" if ($opcode & Scsitool::LOGIN);
        # $tmf = "ki" if ($opcode & ScsiTool::KILL);
        if (!$tmf){
            $self->_set_err("Unsupported OP for ScsitSimCli TMF: $opcode");
            $self->{TOOL_LIB_STATUS} |= ScsiTool::OP_UNSUP;
            return $self;
        }
       
        $self->{sc_cmd}=$cmd." -t $tmf";
        #
        # and return here because we're done
        #
        return;
    }
    
    #
    # this is always the beginning 
    # CDB of the command
    #
    $cmd .="$buf_len ";

    #
    # add an appropriate qtag if available
    #
    $cmd.="$qtag " if($qtag =~ /\-s|\-o|\-h|\-a/);
 
    #
    # read data from a file to send as 
    #  the SCSI data, if available
    #
    if($data_in ne ScsiTool::DEF_DATA_IN){
        $cmd .= "-d $data_in ";   
    }

    #
    # begin the CDB data bytes
    # in hex
    #
    $cmd .="-c ";
    my @cdb = split " ", $cdb;
    foreach my $hex_byte (@cdb)
        {$cmd.="0x".$hex_byte." ";} 
    #
    # put the command string 
    # in the object hash
    #
    $self->{sc_cmd} = $cmd;    
} #  end build_cmd_string()

########################################################################
# Subroutine Name: _new()
# Objective:       Make a ScsiTool into a Scsit_sim
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub _new{
    
    my ($caller, %args) = @_;
    my $class = ref($caller) || $caller;
    my $self = $class->_base_new(%args);

    unless($self->{client}->hosttype() =~ /linux/i) {
        $self->_set_err("ERROR: ScsitSimCli does not support" .
                        "host type $self->{client}" .
                        ", only linux.");
    }
        
    return $self;

} # end _new()
