#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/San/ScsiTool/Spt.pm#4 $
#$DateTime:

## @summary Provide an interface to the SPT tool
## @author bhanup@netapp.com
## @status public
## @pod ../ScsiTool.pm


#
# Package Name
#
package San::ScsiTool::Spt;

use strict;

#
# Package Requirements
#
use San::ScsiTool;
use San::SanClient::ClientUtil;
use Tharn;
use Data::Dumper;

use Error;
use San::SanLog;

use vars qw( @ISA $ScsiTool $spt_loc $dt_loc);

@ISA = qw(Error SanLog ScsiTool);

my $spt_loc = "/usr/local/test/bin/spt";
my $dt_loc  = "/usr/software/test/bin/dt";
my $pattern_file = "/tmp/Pattern.txt";

########################################################################
# P U B L I C   M E T H O D S
########################################################################

########################################################################
# Subroutine Name: send_cmd()
# Objective:       Execute the Spt tool to send a SCSI Command
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub send_cmd
{
    my ($self, %args)   = @_;

    $self->_clear_err();

    $self->handle_send_cmd_params(%args);
    # just return here no need to join, the caller will do that
    return $self if($self->error);

    #
    # build the cmd string that SPT expects for execution
    #
    $self->build_cmd_string();
    # just return here no need to join, the caller will do that
    return $self if($self->error);

    #
    # SPT uses the "emit" param to specify the format of the
    # data to be returned.
    #
    $self->build_emit_string();
    # just return here no need to join, the caller will do that
    return $self if($self->error);

    my $cmd = $self->{sc_cmd};
    my $emit = $self->{er_emit_str};

    $self->sanlog(-msg => ref($self)."::send_cmd()\n"   .
                          "spt loc      => $spt_loc\n"  .
                          "cmd string   => $cmd\n"      .
                          "emit string  => $emit\n",
                  -msg_level    => 3
                 );

    #
    # TBD here's and example of executing on the client
    #
    my $client_obj  = $self->{client};
    my $client_conn = $client_obj->get_connection();

    my $spt_cmd = "$spt_loc $cmd $emit";

    if (defined $self->{pattern}) {
        $spt_cmd = $self->{pattern} . $spt_cmd;
    }

    my ($status, @output) = &ClientUtil::execute($client_conn, $spt_cmd);

    #
    # Deleting the file we made in build_cmd() so that
    # it isn't continually appended to.
    #


    $self->sanlog(-msg => "Client hostname command returned " .
                          join("\n", @output),
                  -msg_level    => 3
                 );

    #
    # Now here comes the power of expected_responses api call.
    #
    $self->update_rsp_hash(\@output);

    if ($status > 0) {
        $self->sanlog(-msg => "Warning: ".ref($self)."::send_cmd(): " .
                              "Command ($cmd) returned a bad status ($status).",
                      -msg_level    => 3);
    }

    #
    # before we send the response hash back to the testcase
    # make sure that all the required params are were not
    # updated are set to undef
    #
    $self->handle_unsup_responses();

    #
    # now we are ready to reply back to the testcase.
    #
    return $self->get_exp_rsp_params();

} # end send_cmd()

########################################################################
# Subroutine Name: send_batch()
# Objective:       Execute the Spt tool to send a batch of SCSI cmds
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub send_batch {
    my ( $self, %args ) = @_;

    $self->_clear_err();

    my $file_path = $args{'file_path'};
    my $timeout =   (defined $args{'timeout'}) ? $args{'timeout'} : 60 ;

    $self->sanlog(
        -msg => ref($self)
            . "::send_cmd()\n"
            . "spt loc      => $spt_loc\n"
            . "file path    => $file_path\n"
        -msg_level => 3
    );

    my $client_obj  = $self->{client};
    my $client_conn = $client_obj->get_connection();

    my $cmd = "$spt_loc script=$file_path";

    my ( $status, @output ) = &ClientUtil::execute( $client_conn, $cmd, $timeout );

    if ( $status > 0 ) {
        $self->sanlog(
            -msg => "Warning: "
                . ref($self)
                . "::send_cmd(): "
                . "Command ($cmd) returned a bad status ($status).",
            -msg_level => 3
        );
    }

    return $status;
}  # end send_batch()

########################################################################
# Method:    expected_responses
# Objective:  Do the same things as the base class expected_responses
#                and in addition set the er_emit_str to undef so that emit_str
#                is reconstructed based on the new responses bitmask.
# Details:   See POD documentation at the end of this file.
########################################################################
sub expected_responses
{
    my ($self, %args) = @_;
    $self->SUPER::expected_responses(%args);
    # just return here no need to join, the caller will do that
    return $self if($self->error);

    #
    # This is reason for overriding the base class.....
    # nil out the er_emit_str so that build_emit_string knows
    # to regenerate the er_emit_str.
    #
    $self->{er_emit_str} = undef;

}

########################################################################
# Subroutine Name: update_rsp_hash()
# Objective:       Take the exp resp bitmask and fill in the keys
#                  of the response hash that sub-class have to address
#                  This should be done after the cmd has completed execution.
# Details:         See POD documentation in ScsiTool.pm
# emit="Status: %status, SCSI Status: %scsi_status, Sense Code: %sense_code, Sense Key: %sense_key, Ascq: %ascq"
########################################################################
sub update_rsp_hash
{
    my ($self, $outp_ref)   = @_;

    $self->_clear_err();

    # since the exp_rsp_params hash is used to convey results to
    # the testcase.....
    # make sure to reset all the expected responses params to defaults.
    # so the next call to update_rsp_hash will always return data
    # that is valid for this execution of send_cmd and not the
    # previous invocation.
    #
    $self->build_exp_rsp_hash();
    return $self if($self->error);

    # first get the exp_rsp_hash ref
    my $er_ref = $self->get_exp_rsp_params();

    my ($srch_str, $spt_emit);

    while (my($name, $value) = each(%$er_ref)) {
        # in the SPT output array we need to find the line
        # that has the string corresponding to the emit string
        # so first find a required value and note its array index
        # then exit
        if($name !~ /TOOL|SCSI_DATA/ && $value eq ScsiTool::REQ_RESPONSE) {
            $srch_str = $name;
            last;
        }
    }

    my $data_found = 0;
    my @scsi_data;
    foreach my $ln (@$outp_ref) {
        # take out any leading or trailing white space
        $ln =~ s/^\s*|\s*$//g;

        if($ln =~ /$srch_str/) {
            $data_found = 0;
            $ln =~ s/^(>\s)*//;
            $spt_emit = $ln;
        } elsif ( $ln =~ /Off(set)?\s+Data/ && !$data_found ) {
            # handle data here.....
            $data_found = 1;
            next;
        }
        if($data_found) {

            my ($hex, $bin) = split /"/, $ln;
            my @ln_data = split /\s+/, $hex;
            # the very first value will be the offset val
            # so take that off
            shift @ln_data;
            push @scsi_data, @ln_data;
        }
    }
    $self->sanlog(-msg => ref($self)."::update_rsp_hash()\n"                .
                          "token used to id the status line: $srch_str\n",
                  -msg_level    => 3
                 );

    $er_ref->{SCSI_DATA}  = \@scsi_data if(@scsi_data);

    unless(length $spt_emit) {
        $self->_set_err("Error: ".ref($self)."::update_rsp_hash(): ".
                        "Could not find emit string SPT output! "    .
                        "using $srch_str as the token"
                       );
        return $self;
    }
    # split the line at identified index, on ',' into tokens
    # Status: 0, SCSI Status: 0, Sense Key: 0, Ascq: 0000, Resid: 138
    my @emit_tokens = split /,/, $spt_emit;

    # then split each token on ':' into string and values
    my ($str, $val);

    foreach my $ln (@emit_tokens) {

        ($str, $val) = split /:/, $ln;

        # take out any leading or trailing white space
        $str =~ s/^\s*|\s*$//g;
        $val =~ s/^\s*|\s*$//g;


        unless (defined $self->get_exp_rsp_params()->{$str}) {
            $self->_set_err("Error: ".ref($self)."::update_rsp_hash(): "        .
                            "Could not find $str token in the exp_rsp_params "  .
                            "hash. The values would have been: $val"
                            );
            return $self;
        }
        # walk the rsp_hash and for each string match
        # assign the returned values.

        if($str eq "ASCNQ") {
            #
            # There is a bug in SPT output which causes ASCNQ field to b set
            # "2020" when SCSI status is GOOD. When SCSI Status is GOOD, ASCNQ
            # values do not matter.
            #
            if($val eq "2020") {

                $val = "0000";

                $self->sanlog(-msg => ref($self)."::update_rsp_hash()".
                                      "ASCNQ set to 2020 resetting it to 0000",
                              -msg_level    => 3
                             );
            }

            $val =~ s#(.{1,2})(.{1,2})#$1h/$2h#;

        }

        $self->get_exp_rsp_params()->{$str} = $val;
    }

    #
    # set the toolname to SPT
    $self->get_exp_rsp_params()->{TOOLNAME} = "spt";

    return 1;
}
########################################################################
# Subroutine Name: build_emit_string()
# Objective:       Take the exp resp bitmask and fill in the keys
#                  of the response hash that sub-class have to address
#                  This should be done when the constructor is executed first.
# Details:         See POD documentation in ScsiTool.pm
# emit="Status: %status, SCSI Status: %scsi_status, Sense Code: %sense_code, Sense Key: %sense_key, Ascq: %ascq"
########################################################################
sub build_emit_string
{
    my $self        = shift @_;

    if($self->{er_emit_str}) {
        $self->sanlog(-msg => ref($self)."::build_emit_string()\n"                .
                              "REUSING emit string: $self->{er_emit_str}\n",
                      -msg_level    => 3
                     );
        return 1;
    }

    my ($ln_sep);

    if($self->{client}->hosttype() =~ /windows/) {
        $ln_sep = "";
    } else {
        $ln_sep ="\\\n";
    }

    my $emit_str    = " emit=\"";

    $self->_clear_err();

    # first get the exp_rsp_hash ref
    my $er_ref = $self->get_exp_rsp_params();

    while (my($name, $value) = each(%$er_ref)) {
        # if the values is required
        # then put that name into the emit string
        if($value eq ScsiTool::REQ_RESPONSE) {
            # then using the name figure out what
            # SPT specific format string needs to go into the emit string.
            $emit_str .= "$name: \%status, $ln_sep" if($name eq "IOCTL_STATUS");
            $emit_str .= "$name: \%scsi_status, $ln_sep" if($name eq "SCSI_STATUS");
            $emit_str .= "$name: \%host_status, $ln_sep" if($name eq "HOST_STATUS");
            $emit_str .= "$name: \%driver_status, $ln_sep" if($name eq "DRIVER_STATUS");
            $emit_str .= "$name: \%resid, $ln_sep" if($name eq "RESIDUAL_BYTES");
            $emit_str .= "$name: \%xfer, $ln_sep" if($name eq "BYTES_XFERD");
            $emit_str .= "$name: \%sense_key, $ln_sep" if($name eq "SENSE_KEY");
            $emit_str .= "$name: \%ascq, $ln_sep" if($name eq "ASCNQ");
            $emit_str .= "$name: \%sense_data, $ln_sep" if($name eq "SENSE_DATA");
        }
    }
    $emit_str .= "\"";

    if($emit_str eq " emit=\"\"") {
        $self->_set_err("Error: ".ref($self)."::build_emit_string(): ".
                        "Could not build SPT emit string!");
        return $self;
    }

    #
    # store the $emit_str in $self.
    #
    $self->{er_emit_str} = $emit_str;

    return 1;
}


########################################################################
# Subroutine Name: build_cmd_string()
# Objective:       For the send_cmd params build the string that
#                  SPT uses to specify which cmd to execute.
#                  This should be done when ever send_cmd is invoked.
# Details:         See POD documentation in ScsiTool.pm
########################################################################
sub build_cmd_string
{
    my $self = shift @_;

    $self->_clear_err();

    $self->{pattern} = undef;

    my $sc_ref          = $self->get_send_cmd_params();

    my $dev             = $sc_ref->{DEVICE};
    my $cdb             = $sc_ref->{CDB};
    my $cdb_param       = $sc_ref->{CDB_PARAM};
    my $dir             = $sc_ref->{DIR};
    my $buf_len         = $sc_ref->{BUF_LEN};
    my $opcode          = $sc_ref->{OPCODE};
    my $qtag            = $sc_ref->{QTAG};
    my $opname          = $sc_ref->{OPCODE_NAME};
    my $repeat_cnt      = $sc_ref->{REPEAT_CNT};
    my $runtime         = $sc_ref->{RUNTIME};
    my $timeout         = $sc_ref->{TIMEOUT};
    my $on_error        = $sc_ref->{ON_ERROR};
    my $sleep           = $sc_ref->{SLEEP};
    my $data_in         = $sc_ref->{DATA_IN};
    my $data_out        = $sc_ref->{DATA_OUT};
    my $data_pat        = $sc_ref->{DATA_PAT};
    my $flag            = $sc_ref->{ENABLE};
    # Below options can be used only on 1.1 and above versions of spt
    my $exp_skey        = $sc_ref->{EXP_SKEY};
    my $exp_status      = $sc_ref->{EXP_STATUS};
    my $exp_asc         = $sc_ref->{EXP_ASC};
    my $exp_asq         = $sc_ref->{EXP_ASQ};
    my $exp_resid       = $sc_ref->{EXP_RESID};
    my $dlimit          = $sc_ref->{DLIMIT};

    my ($cmd, $ln_sep);

    if($self->{client}->hosttype() =~ /windows/) {
        $ln_sep = "";
    } else {
        $ln_sep ="\\\n";
    }

    if($dev eq ScsiTool::REQ_RESPONSE) {
        $self->_set_err("Error: ".ref($self)."::build_cmd_string(): " .
                        "Valid device name is required to send SCSI cmd.");
        return $self;
    }
    $cmd  = " dsf=$dev$ln_sep";

    my @cdb_bytes = split /\s+/, $cdb;

    chomp(@cdb_bytes);

    s/^\s*|\s*$// for(@cdb_bytes);

    my $op_val = hex $cdb_bytes[0];

    if(($op_val == 0x0A) || ($op_val == 0x2A) ||
       ($op_val == 0x8A) ||
       ($op_val == 0x5F) ||
       ($op_val == 0x04) ||
       ($op_val == 0x15) || ($op_val == 0x55) ) {

       $dir = ScsiTool::WRITE;

    }

    #
    #If buf_len = 0, then set bypass flag
    #and change the direction to "none"
    #
    if ($buf_len == 0){
        $cmd .= " enable=bypass$ln_sep";
        $dir = ScsiTool::NONE;
    }

    $cmd .= " dir=".(length ($dir) ?  "$dir$ln_sep" : "read$ln_sep");

    if($opcode != ScsiTool::SCSI_CDB) {
        if($opcode & ScsiTool::ABORT_TASK_SET) {
            $cmd .= " op=abort_task_set$ln_sep";
        }
        elsif($opcode & ScsiTool::BUS_RESET) {
            $cmd .= " op=bus_reset$ln_sep";
        }
        elsif($opcode & ScsiTool::LUN_RESET) {
            $cmd .= " op=lun_reset$ln_sep";
        }
        elsif($opcode & ScsiTool::TARGET_RESET) {
            $cmd .= " op=target_reset$ln_sep";
        }
        elsif($opcode & ScsiTool::NACA_BIT ||
              $opcode & ScsiTool::LINK_BIT) {
            # (REDUNDANT option)
            # the op string is not set here but the cdb that goes into
            # the cmd string is manipulated to enable NACA and/or LINK bits
            # in the control byte of the cdb.
            $cdb_bytes[$#cdb_bytes] |= 0x04 if($opcode & ScsiTool::NACA_BIT);
            $cdb_bytes[$#cdb_bytes] |= 0x01 if($opcode & ScsiTool::LINK_BIT);
            $cdb = join ' ', @cdb_bytes;
        }
        else {
            $self->_set_err("Error: ".ref($self)."::build_cmd_string():" .
                            "Unsupported OP for SPT: $opcode");
            $self->{TOOL_LIB_STATUS} |= ScsiTool::OP_UNSUP;
            return $self;
        }
    }
    #
    # now that we have processed what the opcode is go ahead
    # and update the cdb in the cmd string
    #
    if($ln_sep eq "\\\n") {
        $cmd .= " cdb=\'".(($cdb ne ScsiTool::DEF_CDB)? $cdb : ScsiTool::DEF_CDB)."\'$ln_sep";
        $cmd .= " pout=\'".$cdb_param."\'$ln_sep" if ($cdb_param ne ScsiTool::DEF_CDB_PARAM);
    } else {
        $cmd .= " cdb=\"".(($cdb ne ScsiTool::DEF_CDB)? $cdb : ScsiTool::DEF_CDB)."\"$ln_sep";
        $cmd .= " pout=\"".$cdb_param."\"$ln_sep" if ($cdb_param ne ScsiTool::DEF_CDB_PARAM);
    }

    # length has to come after pout or spt won't use the length specified by
    # the parameter for expected data length.  it will just use whatever
    # the length of the pout is in bytes.  if any scripts were not setting
    # the length correctly, this could cause them to break.
    $cmd .= " length=".(length($buf_len) ? $buf_len : ScsiTool::DEF_LEN)."$ln_sep";

    $cmd .= " enable=$flag$ln_sep" if($flag ne ScsiTool::DEF_DISABLED);
    $cmd .= " sname=$opname$ln_sep" if($opname ne ScsiTool::DEF_OP_NAME);
    $cmd .= " runtime=$runtime$ln_sep" if($runtime != ScsiTool::DEF_RUNTIME);
    $cmd .= " timeout=$timeout$ln_sep" if($timeout != ScsiTool::DEF_TMO);
    #
    # spt will continue by default so make sure to stop if
    # error occurs.
    $cmd .= " onerr=stop$ln_sep" if($on_error eq ScsiTool::DEF_ONERR);

    $cmd .= " sleep=$sleep$ln_sep" if($sleep != ScsiTool::DEF_SLEEP);
    $cmd .= " din=$data_in$ln_sep" if($data_in ne ScsiTool::DEF_DATA_IN);
    $cmd .= " dout=$data_out$ln_sep" if($data_out ne ScsiTool::DEF_DATA_OUT);

    # update the expected value 'the multiple expect string'
    foreach my $exp (@{$sc_ref->{EXPECT}}){
        $cmd .= " expect=$exp$ln_sep";
    }

    # update the expected value 'the exp_status string'
    $cmd .= " status=$exp_status$ln_sep" if(defined $exp_status);

    # update the expected value 'the exp_skey string'
    $cmd .= " skey=$exp_skey$ln_sep" if(defined $exp_skey);

    # update the expected value 'the exp_asc string'
    $cmd .= " asc=$exp_asc$ln_sep" if(defined $exp_asc);

    # update the expected value 'the exp_asq string'
    $cmd .= " asq=$exp_asq$ln_sep" if(defined $exp_asq);

    # update the expected value 'the exp_resid string'
    $cmd .= " resid=$exp_resid$ln_sep" if(defined $exp_resid);

    $cmd .= " dlimit=$dlimit$ln_sep" if(defined $dlimit);

    if($data_pat ne ScsiTool::DEF_DATA_PAT ) {
        my $client_obj  = $self->{client};
        my $host_type   = $client_obj->hosttype();

        if ($client_obj->error()) {
            $self->_set_err($client_obj->errmsg());
            return $self;
        }

        my $alloc_len = $buf_len / 512;

        my $dt;

        # dt Eg: pattern=0x39c39c39
        # specify this pattern when the number of LBAs to be written is more than 1024

        if ($buf_len > 1024) {
             $data_pat = unpack("H*", $data_pat);

             $data_pat x= (8/(length $data_pat));

             $dt = "$dt_loc of=- bs=512 count=$alloc_len pattern=$data_pat disable=verify,stats | ";

             $self->{pattern} = $dt;

             $cmd .= " din=-$ln_sep";
        } else {
            # CAW uses a pattern string which contains 512 bytes of verify pattern
            # and 512 bytes of write pattern. Create a Pattern file with the

            #
            #If there is a remote file there already, remove it
            #

            if($host_type =~ /windows/i){
                $pattern_file = 'c:\\temp\\Pattern.txt';
                $client_obj->execute("del \"$pattern_file\"");
            }else{
                $client_obj->execute("rm -f " .$pattern_file);
            }

            $client_obj->execute("touch $pattern_file");

            #
            #Number of times the pattern needs repeating (rounded down)
            #
            my $multiplier = $buf_len/length($data_pat);
            $data_pat x= $multiplier;

            #
            #Finds out how many odd characters were left off
            #so they can be added too
            #
            my $chars_left = $buf_len - length($data_pat);

            #
            # data_pat now holds the whole pattern
            #
            $data_pat .= substr($data_pat,0,$chars_left);

            my $dt = "$dt_loc of=$pattern_file bs=512 count=$alloc_len pattern=$data_pat disable=verify,stats ;$ln_sep ";

            $self->{pattern} = $dt;

            $cmd .= " din=$pattern_file$ln_sep";
        }
    }

    unless (length $cmd) {
        $self->_set_err("Error: ".ref($self)."::build_cmd_string(): ".
                        "Could not build SPT cmd!");
        return $self;
    }

    #
    # store the $cmd in $self.
    #
    $self->{sc_cmd} = $cmd;

    return 1;
}


########################################################################
# P R I V A T E   M E T H O D S
########################################################################

########################################################################
# Subroutine Name: _create_file()
# Objective:       Create a Pattern File
########################################################################
sub _create_file(){
    # Here block.
    return << 'SCRIPT';

    my ($path,$str) = @_;

    open FILE, ">>$path";

    print FILE $str;

    close FILE;

# This Can NOT be indented
SCRIPT
}

########################################################################
# Subroutine Name: _new()
# Objective:       Create a new sanlun object
# Details:         See POD documentation at the end of this file
########################################################################
sub _new
{
    my ($caller, %args) = @_;
    my $class = ref($caller) || $caller;
    my $self = $class->_base_new(%args);
	# If user supplied the location, just use it.

    if( defined $args{'-tool'}) {
        if($self->{client}->hosttype() =~ /unix|vmware/i) {
			$spt_loc = $self->{client}->{TEST_BIN}."/spt.1.11" ;
			$dt_loc  = $self->{client}->{TEST_BIN}."/dt" ;
		}
		elsif($self->{client}->hosttype() =~ /windows/) {
			$spt_loc = $self->{client}->{TEST_BIN}. "spt.1.11";
			$dt_loc  = $self->{client}->{TEST_BIN}. "dt" ;
		}
		
    } else {
		if($self->{client}->hosttype() =~ /unix|vmware/i) {
			$spt_loc = $self->{client}->{TEST_BIN}."/spt" ;
			$dt_loc  = $self->{client}->{TEST_BIN}."/dt" ;
		}
		elsif($self->{client}->hosttype() =~ /windows/) {
			$spt_loc = $self->{client}->{TEST_BIN}. "spt";
			$dt_loc  = $self->{client}->{TEST_BIN}. "dt" ;
		}
	}
    # If user supplied the location, just use it.
    if( defined $args{'-scsi_tool_loc'}) {
        $spt_loc = $args{'-scsi_tool_loc'};
    }

    return $self;

} # end _new()

