#$Id$
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @author chadwell@netapp.com
## @status public
## @pod here
## @summary This is the script that converts existing
## SCSIT presubmit tests to usable tests by
## SAN QA automation tools (spt or any_iscsi).
## This will enable SNARL runs to test with a lot
## of different CDB's without having to write whole tests
#
## input format is in the form of existing scsitools. this for any-iscsi :
## 
## <lun> <buf_len> [ flags ] -c [bytes]...
## unspecified bytes are 0.
## -s    simple
## -h    head of queue
## -o    ordered
## -x <scsi_status> [<skey> <asc> <ascq>]
## -R [ {{ o | u } <resid_value> } | 0 ]  # that's -R 0 ---> no resid or -R u 5 ---> 5 byte underrun
## 
## or SPT:
##cdb='<cdb>' dir=<read|write|none>  length=<length>  exp_status="<status>" exp_sk="<sense key>" exp_ascnq="<asc/q values example: 0Eh/03h>"
##
## or cmd line:
##cdb;scsiStatus;sensekey;asc;ascq;dataBufLen;<under|over|none>;residLen;[optional 'invalid' cdb indicator]


package San::ScsiTool::TScriptParser;
use strict;
use vars qw(      
          @EXPORT 
          @ISA
          );

our @ISA = qw(Error SanLog);


my (@parsed_cmds, @raw_data);
my($arg, $file, $tool, $ln, $byte, @cdb_cmd, @data_out, $offset, $value, @expected); 

##############################################################################
#Method:    parse_spt
#Objective: parse a file of batched spt commands
#Details:   See POD documentation at the end of this file
##############################################################################
sub parse_spt{
    my ($self, %args) = @_; 

    $self->_clear_err();
    
    foreach my $key (%args){
        if ($key =~ /-?file/){

            unless (length $args{$key}) {
                $self->_set_err("ERROR: -file must contain valid filename\n");
                return 0;
            }

            $file = $args{$key};
        }
    }

    unless (open(INPUTFILE, $file)){
        $self->_set_err("ERROR: could not open file: $file\n");
        return 0;
    }
    
    @parsed_cmds = ();
    while(<INPUTFILE>){
        
        #
        #Instantiate a hash to hold the parsed
        #command
        #
        my $parsed_cmd;

        #
        #make the next line of output the current 
        #working line.
        #
        $ln = $_;
        chomp($ln);

        #
        #If the line starts with a #, ignore it it's a 
        #comment line... like this one
        #
        if($ln =~ /\#|^\n$|^$/){
            next;
        }

        #
        #Otherwise, we parse what we expect to be a
        #CDB scsi command and it's expected return values.
        #Below we strip the parameter name and '' or "" out from
        #around the values we want to save.
        #
        if ($ln =~ /cdb\=('|")(.*?)('|")/){
            $parsed_cmd->{cdb} = $2;
        }
        
        if ($ln =~ /param\=('|")(.*?)('|")/){
            $parsed_cmd->{cdb_param} = $2;
        }

        if ($ln =~ /exp_status\=('|")(.*?)('|")/){
            $parsed_cmd->{scsi_status} = $2;
        }

        if ($ln =~ /exp_sk\=('|")(.*?)('|")/){
            $parsed_cmd->{sense_key} = $2;
        }
    
        #
        #we split up the ASC and ASCQ values, which usually come in pairs
        #in case we need them separately later. this expresion strips the quotes
        #from around the asc/q values, and then matches two groups between 1 
        #and 2 alphanumeric characters followed maybe by an "h" and separated by
        #a slash.  we aren't interested in the "h" though, we only save the 
        #actual value of the asc/q.
        #
        if ($ln =~ /exp_ascnq\=('|")(\w{1,2})h?\/(\w{1,2})h?('|")/){
            $parsed_cmd->{asc} = $2;
            $parsed_cmd->{ascq} = $3;
        }

        if ($ln =~ /dir\=(\w+)/){
            $parsed_cmd->{dir} = $1;
        }

        if ($ln =~ /length\=(\d+)/){
            $parsed_cmd->{buf_len} = $1;
        }

        if ($ln =~ /din\=(\w+)/){
            $parsed_cmd->{data_out} = $1;
        }

        if ($ln =~ /dout\=(\w+)/){
            $parsed_cmd->{response} = $1;
        }
        
        if ($ln =~ /pout\=\'(.*)\'/){
            $parsed_cmd->{param_out} = $1;
        }

        if ($ln =~ /op\=(\w+)/){
            $parsed_cmd->{tmf}=$1;
        }

        push(@parsed_cmds, $parsed_cmd) unless (!$parsed_cmd->{cdb});
    
    }

    close(INPUTFILE);
    $self->sanlog(-msg       => "TScriptParser has parsed " . scalar @parsed_cmds . " scsi commands.", 
                  -msg_level => 1);
    return (wantarray()) ? @parsed_cmds : \@parsed_cmds;

}
    
##############################################################################
#Method:    parse_any_iscsi
#Objective: parse a file of batched any_iscsi commands
#Details:   See POD documentation at the end of this file
##############################################################################
sub parse_any_iscsi {    
    my ($self, %args) = @_; 

    foreach my $key (%args){
        if ($key =~ /-?file/){

            unless (length $args{$key}) {
                $self->_set_err("ERROR: -file must contain valid filename\n");
                return 0;
            }

            $file = $args{$key};
        }
    }

    unless (open(INPUTFILE, $file)){
        $self->_set_err("ERROR: could not open file: $file\n");
        return 0;
    }
    
    @parsed_cmds = undef;
    while(<INPUTFILE>){
        my (@return_expectations_32, @return_expectations_64, @return_expectations_ascii);
        
        #
        #Instantiate a hash to hold the parsed
        #command
        #
        my $parsed_cmd;
        
        #
        #make the next line of output the current 
        #working line.
        #
        $ln = $_;
        chomp($ln);

        #
        #If the line starts with a #, ignore it it's a 
        #comment line... like this one
        #
        if($ln =~ /#|^\n*$|^$/){
            next;
        }

        #
        #Otherwise, we parse what we expect to be a
        #CDB scsi command and it's expected return values.
        #
        @cdb_cmd = split(/ +\t*|\t+/,$ln);
        $parsed_cmd->{lun}  = shift @cdb_cmd;
        $parsed_cmd->{buf_len} = shift @cdb_cmd;
        while (@cdb_cmd){
            $byte = shift(@cdb_cmd);
            if($byte =~ /-s|-h|-o/){ #simple, ordered, or head of queue
                $parsed_cmd->{qtag} = $byte;
            }elsif($byte =~ /-a/){ #ACA task
                $parsed_cmd->{ACA} = "true";
            }elsif($byte =~ /-N/){ #User defined Integer ID for tracking
                $parsed_cmd->{id} = shift(@cdb_cmd) if($cdb_cmd[0] !~ /-\w/);
            }elsif($byte =~ /-A/){ #async mode
                $parsed_cmd->{async} = "true";
            }elsif($byte =~ /-S/){ #Sleep before send
                $parsed_cmd->{sleep} = shift(@cdb_cmd) if($cdb_cmd[0] !~ /-\w/);
            }elsif($byte =~ /-P|-Z|-J|-f/){ #SCSITSIM ONLY
                $parsed_cmd->{sim_only} = "true";
            }elsif($byte =~ /-D/){ #data out bytes
                while(@cdb_cmd){
                    push(@data_out, shift(@cdb_cmd)) if($cdb_cmd[0] !~ /-\w/);
                    if ($cdb_cmd[0] =~ /-\w/){
                        last;
                    }
                }
                $parsed_cmd->{data_out} = \@data_out;
            }elsif ($byte =~ /-x/){    #Expected statuses
                $parsed_cmd->{scsi_status} = shift(@cdb_cmd) if($cdb_cmd[0] !~ /-\w/);
                $parsed_cmd->{scsi_status} =~ s/^0x// ;

                $parsed_cmd->{sense_key} = shift(@cdb_cmd) if($cdb_cmd[0] !~ /-\w/);
                $parsed_cmd->{sense_key} =~ s/^0x// ;

                $parsed_cmd->{asc} = shift(@cdb_cmd) if($cdb_cmd[0] !~ /-\w/);
                $parsed_cmd->{asc} =~ s/^0x// ;
                $parsed_cmd->{asc} = "0".$parsed_cmd->{asc} if ($parsed_cmd->{asc} < 10);

                $parsed_cmd->{ascq} = shift(@cdb_cmd) if($cdb_cmd[0] !~ /-\w/);
                $parsed_cmd->{ascq} =~ s/^0x// ;
                $parsed_cmd->{ascq} = "0".$parsed_cmd->{ascq} if ($parsed_cmd->{ascq} < 10);

            }elsif($byte =~ /-R/){     #residual bytes
                if($cdb_cmd[0] =~ /0/){
                    $parsed_cmd->{resid_value} = shift(@cdb_cmd);
                }else{
                    $parsed_cmd->{resid_ou} = shift(@cdb_cmd) if($cdb_cmd[0] =~ /o|u/);
                    $parsed_cmd->{resid_value} = shift(@cdb_cmd) if($cdb_cmd[0] =~ /\d/);
                }
            }elsif($byte =~ /-I/){     #number of times to repeat
                $parsed_cmd->{count} = shift(@cdb_cmd) if($cdb_cmd[0] !~ /-\w/);
            }elsif($byte =~/-W/){      #expect word at location
                $offset = shift(@cdb_cmd);
                $value = shift(@cdb_cmd);
                push(@return_expectations_32, [$offset, $value]); 
            }elsif($byte =~/-L/){      #expect long at location
                $offset = shift(@cdb_cmd); 
                $value = shift(@cdb_cmd);
                push(@return_expectations_64, [$offset, $value]); 
            }elsif($byte =~/-C/){      #expect ascii value at location
                $offset = shift(@cdb_cmd);
                $value = shift(@cdb_cmd);
                push(@return_expectations_ascii, [$offset, $value]); 
            }elsif($byte =~ /-p/){ #for read expecting special pattern, any_iscsi does this, I don't think we need to worry about it
                $parsed_cmd->{expected_pattern}="true";
            }elsif($byte =~ /-c/){
                #
                #by restraint, the cdb bytes are the last
                #thing in the command line.  Once we reach
                #here, we can just put the rest of the line
                #in the cdb var. But we need to make sure there
                #are at least 5 bytes in the cdb, otherwise
                #we risk failure later.
                #
                while(scalar @cdb_cmd < 5){
                    push(@cdb_cmd, 0);
                }

                #
                #Strip the 0x off the front of any hex bytes
                #because other scsitools may not be able to
                #handle them
                #
                my @bare_bytes;
                foreach my $cdb_byte (@cdb_cmd){
                    if ($cdb_byte =~ /0x(.+)/){
                        push(@bare_bytes, $1);
                    }
                    else{
                        #
                        #make sure the value is in hex
                        #
                        my $hex_byte = sprintf("%x", $cdb_byte);

                        push(@bare_bytes, $hex_byte);
                    }
                }
                
                $parsed_cmd->{cdb} = join(" ",@bare_bytes);

                #
                #If the CDB begins with c3, we need to 
                #mark it as sim-only
                #
                if ($bare_bytes[0] =~ /c3/){
                    $parsed_cmd->{sim_only} = "true";
                }
            }elsif($byte =~ /-t/){ #TMF
                $parsed_cmd->{tmf} = shift(@cdb_cmd) if($cdb_cmd[0] !~ /-\w/);
                $parsed_cmd->{tmf_id} = shift(@cdb_cmd) if($cdb_cmd[0] !~ /-\w/); #required ID for abort tasks, just empty if DNE
            }

        }

        $parsed_cmd->{expected_return_32} = \@return_expectations_32;
        $parsed_cmd->{expected_return_64} = \@return_expectations_64;
        $parsed_cmd->{expected_return_ascii} = \@return_expectations_ascii;

        push(@parsed_cmds, $parsed_cmd) unless($parsed_cmd->{sim_only} or $parsed_cmd->{lun} !~ /\d/ or !$parsed_cmd->{cdb});
    }
    
    close(INPUTFILE);
    $self->sanlog(-msg       => "TScriptParser has parsed " . scalar @parsed_cmds . " scsi commands.", 
                  -msg_level => 1);
    return (wantarray()) ? @parsed_cmds : \@parsed_cmds;
}


##############################################################################
#Method:    parse_cmd_line
#Objective: parse a file of batched command line scsi commands
#Details:   See POD documentation at the end of this file
##############################################################################
sub parse_cmd_line{
    my ($self, %args) = @_;
    # 
    # cdb;scsiStatus;sensekey;asc;ascq;dataBufLen;<under|over|none>;residLen;[optional 'invalid' cdb indicator]
    # 
    # write 16
    #
    # 0x8A 0 0 0 0 0 0 0 0 100 0 0 0 5 0 0;0;0;0;0;1740;under;204
    #

    foreach my $key (%args){
        if ($key =~ /-?file/){
            $file = $args{$key};
        }
    }

    die "must pass in file to parse" unless $file;
    open(INPUTFILE, $file) or die "File $file could not be opened, perhaps you need to pass".
                                    "full qualified path?";
    
    @parsed_cmds = undef;                                
    while(<INPUTFILE>){
        #
        #Instantiate a hash to hold the parsed
        #command
        #
        my $parsed_cmd;

        #
        #make the next line of output the current 
        #working line.
        #
        $ln = $_;
        chomp($ln);

        #
        #If the line starts with a #, ignore it it's a 
        #comment line... like this one
        #
        if($ln =~ /\#|^\n$|^$/){
            next;
        }

        #
        #Otherwise, we parse what we expect to be a
        #CDB scsi command and it's expected return values.
        #
        @cdb_cmd = split(/;/,$ln);
        ($parsed_cmd->{cdb},
         $parsed_cmd->{scsi_status},
         $parsed_cmd->{sense_key},
         $parsed_cmd->{asc},
         $parsed_cmd->{ascq},
         $parsed_cmd->{buf_len},
         $parsed_cmd->{resid_ou},
         $parsed_cmd->{resid},
         $parsed_cmd->{invalid_indicator}) = @cdb_cmd;
        
         push(@parsed_cmds, $parsed_cmd) unless($parsed_cmd->{sim_only});
    }

    close(INPUTFILE);
    $self->sanlog(-msg       => "TScriptParser has parsed " . scalar @parsed_cmds . " scsi commands.", 
                  -msg_level => 1);
    return (wantarray()) ? @parsed_cmds : \@parsed_cmds; 
        
}

sub new{
    my $self  = {};
    bless($self);  
    return $self;
}
1;



####################################################################################################
#                                       POD DOCUMENTATION
####################################################################################################

=pod

=head1 Package Name

San::Task::TScriptParser

=head1 Authors

Derek Chadwell

=head1 Description

TScriptParser provides an API that parses a set of batched scsi commands, usually for use with some tool,
and breaks the information down into a predictable hash for the purposes of using with any tool through
the automated test library.  The module is object oriented, and can be instantiated first with TScriptParser::new().

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>

B<Sanlog.pm>

=head1 Public Methods

=head2 B<parse_spt(%args)>

Parse a set of spt commands from a file. 

=over 4

=item Inputs

=item * -file

Path to the file being parsed.

=item * -output

See output example further down

=head2 B<parse_any_iscsi(%args)>

Parse a set of any_iscsi commands from a file.

=over 4

=item Inputs

=item * -file

Path to the file being parsed.

=item * -output

See output example further down

=head2 B<parse_cmd_line(%args)>

Parse a set of command line scsi calls from a file.

=over 4

=item Inputs

=item * -file

Path to the file being parsed.

=item * -output

See output example further down

=head2 B<new>

Make a new TScriptParser object

=item example
     my $parser_obj = San::ScsiTool::TScriptParser->new();

=head1 Output

If the method was executed successfully, it returns an array of hashes (or a reference to an array of hashes)
where each element contains the hash reference of a parsed command.

=head2 Example Hash

=item example at index N

    $parsed_cmd=  lun                    => <the lun being targeted>
                  buf_len                => <the buffer length if cdb>
                  cdb                    => the scsi CDB to send    
                  qtag                   => <simple, ordered, or head of queue flag>
                  ACA                    => whether or not this is an ACA task
                  id                     => an id for the task (used for stopping it later, etc)
                  async                  => whether the command needs to be sent asynchronously
                  sleep                  => an amount of time to sleep before sending the command
                  data_out               => the file containing data to send out with the command
                  data_in                => the file containing expected read data
                  scsi_status            => the scsi status to expect
                  asc                    => the asc value to expect
                  ascq                   => the ascq value to expect
                  resid_ou               => [o|u] if there are too many or few residual bytes
                  resid_value            => how many residual bytes there are
                  count                  => how many times to run the command
                  expected_pattern       => true if read is expecting special NetApp write pattern
                  tmf                    => the TMF, if it is that instead of a CDB command
                  tmf_id                 => ie the ID telling an abort task TMF which task to abort
                  expected_return_32     => array of doubles [<expected 32 bit word>, <location>]
                  expected_return_64     => array of doubles [<expected 64 bit long>, <location>]
                  expected_return_ascii  => array of doubles [<expected ascii values>, <location>]

=head1 Use

=head2 Example use in a testcase

=item example
    my $parser_obj = San::TScriptParser->new();
    
    my @scsi_cmds = $parser_obj->parse_spt(-file  =>  "/u/chadwell/p4/test/san/scsi/codecov_cdbs_list.spt");
    
    #
    #parse the hashes into an arguments hash, to reduce clutter
    # 
    my %args;
        
    $args{-device}   = $pri_paths[0];
    $args{-cdb_name} = $cmd->{cdb};
    $args{-buf_len}  = $cmd->{buf_len};

    #
    #send the command on the primary path
    #
    my $return = $scsi_tool_obj->test_scsi_cmd(%args);
       


