# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Xcopy Factory Object
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::ScsiTool::XcopyFactory;

# compiler directives
use strict;

# imports
use List::MoreUtils;
use NATE::BaseException qw( :try );
use Params::Validate qw( :all );
use Tharn qw($Log);
use San::SanUtil;
use San::ScsiTool::ScsiUtil;
use San::ScsiTool::Xcopy_T10;
use San::ScsiTool::Xcopy_T10::CDB;
use San::ScsiTool::Xcopy_T10::ParameterList;
use San::ScsiTool::Xcopy_T10::SegmentDescriptor::B2BSegmentDescriptor;
use San::ScsiTool::Xcopy_T10::TargetDescriptor::IdTargetDescriptor;
use
    San::ScsiTool::Xcopy_T10::TargetDescriptor::DeviceTypeSpecificParams::BlockDeviceTypeSpecificParams;

our %DeviceCache;

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   See POD documentation at the end of this file for usage
###############################################################################
sub new {
    $Log->enter();
    my $class = shift;

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) }, );
    my %args = validate(
        @_,
        {   'scsi_util_obj' => {
                type     => OBJECT,
                required => 1,
                isa      => 'San::ScsiTool::ScsiUtil'
            }
        }
    );

    my $self = {};
    $self->{'_scsi_util_obj'} = $args{'scsi_util_obj'};
    $self->{'_scsi_tool_obj'} = $self->{'_scsi_util_obj'}->{'s_t_obj'};
    bless $self, $class;
    $Log->exit();
    return $self;
}

################################################################################
# Method Name: create_xcopy_obj()
# Objective:   Creates an Xcopy object from a hash of copy ops.  An Id Target
#              Descriptor is created (and inqs sent) for each unique device.
#              A Block to Block Segment Descriptor is created for each copy op.
# Parameter:   copy_ops - ARRAYREF containing copy_op HASHREF(s)
#                $copy_op = {
#                  'src_dev' => '/dev/device1',
#                  'dest_dev' => '/dev/device2',
#                  'src_lba' => 8192,
#                  'dest_lba' => 8232,
#                  'num_blocks' => 8
#                }
#              unique_tds - bool whether or not to create a unique TD for dup
#              devices
#              'inq_type' => 0x83
# Return:      San::ScsiTool::Xcopy_T10
# Throws:      NATE::Exceptions::Argument
# Throws:      NATE::BaseException if failures are found finding devices via
#              inquiry or sending the xcopy command.
################################################################################
sub create_xcopy_obj () {

    $Log->enter();
    my $self = shift;

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) }, );
    my %args = validate(
        @_,
        {   'copy_ops' => {
                type     => ARRAYREF,
                required => 1
            },
            'unique_tds' => {
                type     => SCALAR | UNDEF,
                optional => 1
            },
        }
    );

    my @devices;
    my @copy_ops = @{ $args{'copy_ops'} };
    my %device_to_td_map;
    my @td_objs;
    my @sd_objs;

    # Get a list of unique devices from the copy ops
    foreach my $copy_op (@copy_ops) {
        push( @devices, $copy_op->{'src_dev'} );
        push( @devices, $copy_op->{'dest_dev'} );
    }

    unless(defined $args{'unique_tds'} && $args{'unique_tds'} == 0 ) { 
        @devices = List::MoreUtils::uniq(@devices);
    }

    # For each device create an Identification TD object
    foreach my $device (@devices) {
  
        # Create Target Descriptors using the values from the inquiry data
        my $td_obj = $self->create_id_target_descriptor_obj($device);

        $device_to_td_map{$device} = {
            'td_obj' => $td_obj,
            'index'  => scalar(@td_objs)
        };

        push( @td_objs,  $td_obj);
    }

    # For each copy op create a Segment Descriptor Object
    foreach my $copy_op (@copy_ops) {

        # Create a Segment Descriptor
        my $sd_obj
            = San::ScsiTool::Xcopy_T10::SegmentDescriptor::B2BSegmentDescriptor
            ->new(
            'block_device_number_of_blocks' => $copy_op->{'num_blocks'},
            'source_block_device_lba'       => $copy_op->{'src_lba'},
            'destination_block_device_lba'  => $copy_op->{'dest_lba'},
            'source_target_descriptor_index' =>
                $device_to_td_map{ $copy_op->{'src_dev'} }->{'index'},
            'destination_target_descriptor_index' =>
                $device_to_td_map{ $copy_op->{'dest_dev'} }->{'index'}
            );

        push( @sd_objs, $sd_obj );
    }

    my ( $parameter_list_obj, $xcopy_obj );

    # Create a Parameter list Obj
    $parameter_list_obj = San::ScsiTool::Xcopy_T10::ParameterList->new(
        'target_descriptor_objs'  => \@td_objs,
        'segment_descriptor_objs' => \@sd_objs
    );

    # Create the Xcopy CDB obj
    my $xcopy_cdb_obj = San::ScsiTool::Xcopy_T10::CDB->new(
        'parameter_list_length' => $parameter_list_obj->get_length() );

    # Create the Xcopy obj
    $xcopy_obj = San::ScsiTool::Xcopy_T10->new(
        'scsi_tool_obj'      => $self->{'_scsi_tool_obj'},
        'cdb_obj'            => $xcopy_cdb_obj,
        'parameter_list_obj' => $parameter_list_obj
    );

    $self->_randomize_do_no_harm_fields( $xcopy_obj );
    
    $Log->exit();
    return $xcopy_obj;
}

################################################################################
# Method Name: create_xcopy_obj_fixed_segment_length()
# Objective:   Creates an Xcopy object with a given number of segments of
#              a fixed, equal size.
# Parameter:   num_segments
#              Number of segments xcopy object should have
#              Number of blocks (lbas) per segment
#              src_lba - lba offset to start the src copy ops at
#              dest_lba - lba offset to start the dest copy ops at
# Return:      San::ScsiTool::Xcopy_T10
# Throws:      NATE::Exceptions::Argument
# Throws:      NATE::BaseException if failures are found finding devices via
#              inquiry or sending the xcopy command.
################################################################################
sub create_xcopy_obj_fixed_segment_length () {

    $Log->enter();
    my $self = shift;

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) }, );
    my %args = validate(
        @_,
        {   'num_segments' => {
                type     => SCALAR,
                required => 1
            },
            'num_blocks' => {
                type     => SCALAR,
                required => 1
            },
            'src_lba' => {
                type     => SCALAR | UNDEF,
                optional => 1
            },
            'dest_lba' => {
                type     => SCALAR | UNDEF,
                optional => 1
            },
            'src_device' => {
                type     => SCALAR,
                required => 1
            },
            'dest_device' => {
                type     => SCALAR,
                required => 1
            },
        }
    );

    my $curr_src_lba = (defined $args{'src_lba'}) ? $args{'src_lba'} : 0 ;
    my $curr_dest_lba = (defined $args{'dest_lba'}) ? $args{'dest_lba'} : 0 ;
    my @copy_ops;
    
    for ( my $i = 0; $i < $args{'num_segments'}; $i++ ) {
        my $copy_op = {
            'src_dev'    => $args{'src_device'},
            'dest_dev'   => $args{'dest_device'},
            'src_lba'    => $curr_src_lba,
            'dest_lba'   => $curr_dest_lba,
            'num_blocks' => $args{'num_blocks'}
        };
        $curr_src_lba += $args{'num_blocks'};
        $curr_dest_lba += $args{'num_blocks'};
        push( @copy_ops, $copy_op );
    }

    my $xcopy_obj = $self->create_xcopy_obj( 'copy_ops' => \@copy_ops );
    
    $Log->exit();

    return $xcopy_obj;
}

################################################################################
# Method Name: create_id_target_descriptor_obj()
# Objective:   Creates an IdTargetDescriptor with LUN Inquiry information.  It
#              looks in the device cache before sending inquiry's to the LUN.
#
# Parameter:   device - device to create an id target descriptor obj for
# Return:      San::ScsiTool::Xcopy_T10::ParameterList::TargetDescriptor::IdTargetDescriptor
# Throws:      NATE::BaseException if failures are found finding devices via
#              inquiry or sending the xcopy command.
################################################################################
sub create_id_target_descriptor_obj {

    $Log->enter();
    my $self = shift;
    my $device = shift;
    
    my %desc_vals;

    my $ret_ref;
    if ( !exists( $DeviceCache{$device} ) ) {
        print ' Device doesnt exist ' . $device;

        # Send inq 83h
        $ret_ref = $self->{'_scsi_tool_obj'}->send_cmd(
            -device => $device,
            -cdb    => 'inquiry_page_83h'
        );

        if ( $self->{_scsi_tool_obj}->error() ) {
            NATE::BaseException->throw( $self->{'_scsi_tool_obj'}->errmsg() );
        }

        # Parse inq 83h
        my $inq_h = $self->{'_scsi_util_obj'}
            ->xlate_inq_pg_83h_bytes( $ret_ref->{'SCSI_DATA'} );
        if ( $self->{'_scsi_util_obj'}->error() ) {
            NATE::BaseException->throw( $self->{'_scsi_util_obj'}->errmsg() );
        }

        # Get the values needed from the LUN
        # Binary NAA Descriptor (designator, etc)
        %desc_vals = %{ $self->_get_83h_desc_h( 'inq_h' => $inq_h ) };

        $DeviceCache{$device} = \%desc_vals;
    } else {
        %desc_vals = %{ $DeviceCache{$device} };
    }

    # Create Target Descriptors using the values from the inquiry data
    my $td_obj
        = San::ScsiTool::Xcopy_T10::TargetDescriptor::IdTargetDescriptor->new(
        %desc_vals);

    return $td_obj;
}

################################################################################
# Method Name: _find_lun_binary_naa_descriptor()
# Objective:   Searches through inq 83h hash to find the Lun Binary NAA
#              Descriptor
# Parameter:   inq_h - inq 83h hash returned by ScsiUtil xlate method
# Return:      designator descriptor hash ref - format described in ScsiUtil
#              xlate_lun_binary_naa_id documentation.
# Throws:      NATE::BaseException if descriptor is not found
################################################################################
sub _find_lun_binary_naa_descriptor() {

    $Log->enter();
    my ( $self, %args ) = @_;

    my %inq_h = %{ $args{'inq_h'} };

    foreach my $id ( @{ $inq_h{'ID_LIST'} } ) {
        if (   $id->{'CODE_SET'} == 1
            && $id->{'ASSOCIATION'} == 0
            && $id->{'IDENTIFIER_TYPE'} == 3 ) {
            return $id;
        }
    }

    $Log->exit();
    NATE::BaseException->throw('Binary naa descriptor was not be found.');
}

################################################################################
# Method Name: _get_83h_desc_h()
# Objective:   Maps inq 83h binary naa id to params expected by TargetDescriptor
# Parameter:   inq_h - inq 83h hash returned by ScsiUtil xlate method
# Return:      hashref of params for TargetDescriptor
# Throws:      NATE::BaseException if descriptor is not found
################################################################################
sub _get_83h_desc_h() {

    $Log->enter();
    my ( $self, %args ) = @_;

    my $td_values
        = $self->_find_lun_binary_naa_descriptor( 'inq_h' => $args{'inq_h'} );

    my $desg = $td_values->{'NETAPP_MAGIC_NUMBER'}
        . unpack( "H*", $td_values->{'LUN_SERIAL_NUMBER_0_11'} );
    $desg =~ s/(.{2})/$1 /g;
    $desg =~ s/\s+$//;

    my %inq_values = (
        'code_set'          => $td_values->{'CODE_SET'},
        'association'       => $td_values->{'ASSOCIATION'},
        'designator_type'   => $td_values->{'IDENTIFIER_TYPE'},
        'designator_length' => $td_values->{'IDENTIFIER_LENGTH'},
        'designator'        => $desg . ' 00 00 00 00'
    );

    $Log->exit();
    return \%inq_values;
}

################################################################################
# Method Name:  _randomize_do_no_harm_fields
# Objective:    randomly modifies key do no harm fields of the xcopy object
#               PAD, DC, CAT, DISK BLOCK LENGTH 
# Return:       $TCD::PASS
################################################################################
sub _randomize_do_no_harm_fields() {

    $Log->enter();
    my $self = shift;
    my $xcopy_obj = shift;
    
    my $param_obj = $xcopy_obj->get_parameter_list_obj();
    my $sd_objs   = $param_obj->get_segment_descriptor_objs();
    my $td_objs   = $param_obj->get_target_descriptor_objs();
    
    foreach my $sd_obj (@{$sd_objs}) {
        $sd_obj->set_dc( int( rand( 1 ) ) );
        $sd_obj->set_cat( int( rand( 1 ) ) );
    }

    foreach my $td_obj (@{$td_objs}) {
        
        my $disk_block_length = (int( rand( 1 ) ) == 0) ? 0 : 512;
        my $dts_params_obj = $td_obj->get_device_type_specific_parameter_obj();
        $dts_params_obj->set_disk_block_length( $disk_block_length );
        $dts_params_obj->set_pad( int( rand( 1 ) ) );
        
    }      
                   
    $Log->exit();
    return $TCD::PASS;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

San::ScsiTool::XcopyFactory.pm

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

XcopyFactory.pm

High level description of public methods

High level description of parameters

=head1 See Also

None

=head1 Attributes

None

=head1 Revision History

12-22-2011 schroll@netapp.com: Created Interface

=head1 Public Methods

=head2 B<new()>

Class Constructor method. 

=over 4

=item Inputs

=item * I<-scsi_util_obj> <required>

Instance of ScsiUtil

=item Outputs

Returns a XcopyFactory blessed object reference.

=item Examples

   # Create a XcopyFactory object
   try {
       my $factory_obj = San::ScsiTool::XcopyFactory->new( 'scsi_util_obj' => $Scsi_util_obj );
   } catch NATE::BaseException with {
       my $ex = shift;
       $Log->debug( "Error message is " . $ex->text() . "\n" );
   };	

=back

=head2 B<create_xcopy_obj()>

Creates an Xcopy object from a hash of copy ops.  An Id Target
Descriptor is created (and inqs sent) for each unique device.
A Block to Block Segment Descriptor is created for each copy op.

=over 4

=item Inputs

=item * I<copy_ops> <required>

copy_ops - ARRAYREF containing copy_op HASHREF(s)
  $copy_op = {
    'src_dev' => '/dev/device1',
    'dest_dev' => '/dev/device2',
    'src_lba' => 8192,
    'dest_lba' => 8232,
    'num_blocks' => 8
  }

=item Outputs

Returns a Xcopy blessed object reference.

=item Examples

   try {
   	
	  my $factory_obj = San::ScsiTool::XcopyFactory->new( 'scsi_util_obj' => $Scsi_util_obj );

	  my $copy_op = { 'src_dev' => '/dev/rdsk/c4t12d0s6', 
	                  'dest_dev' => '/dev/rdsk/c4t13d0s6', 
	                  'src_lba' => 8192, 
	                  'dest_lba' => 8232,
	                  'num_blocks' => 8
	  };
	
	  my $xcopy_obj = $factory_obj->create_xcopy_obj( 'copy_ops' => [$copy_op] );

      my $return_ref = $xcopy_obj->send_cmd(-device =>'/dev/rdsk/c4t13d0s6'  );
    
   } catch NATE::BaseException with {
       my $ex = shift;
       $Log->debug( "Error message is " . $ex->text() . "\n" );
   };	
    
=back

=cut
