# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Object modeling the Xcopy SCSI command defined in spc4r32.
## Upon creating the object, the user can get the CDB and Parameter
## List of the XCOPY command in a hex string format, or they can send
## the command through the Scsi Tool Object.
##
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::ScsiTool::Xcopy_T10;

# compiler directives
use strict;

# imports
use NATE::Exceptions::Argument;
use Params::Validate qw( :all );
use San::ScsiTool;
use Tharn qw($Log);

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   See POD documentation at the end of this file for usage
###############################################################################
sub new {
    $Log->enter();
    my $class = shift;

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) }, );
    my %args = validate(
        @_,
        {   'scsi_tool_obj' => {
                type     => OBJECT,
                required => 1,
                isa      => 'ScsiTool'
            },
            'cdb_obj' => {
                type     => OBJECT,
                required => 1,
                isa      => 'San::ScsiTool::Xcopy_T10::CDB'
            },
            'parameter_list_obj' => {
                type     => OBJECT | UNDEF,
                optional => 1,
                isa      => 'San::ScsiTool::Xcopy_T10::ParameterList'
            },
        }
    );

    my $self = {
        '_scsi_tool_obj'      => $args{'scsi_tool_obj'},
        '_cdb_obj'            => $args{'cdb_obj'},
        '_parameter_list_obj' => $args{'parameter_list_obj'}
    };

    # bless instance into class
    bless $self, $class;

    # Link the Parameter List Obj to the Xcopy Obj
    if ( defined $self->{'_parameter_list_obj'} ) {
        $self->{'_parameter_list_obj'}->set_xcopy_obj($self);
    }

    $self->set_buf_len( $self->{'_cdb_obj'}->get_parameter_list_length() );
    
    $Log->exit();
    return $self;
}

##############################################################################
# Method:      send_cmd
# Objective:   Sends the Xcopy command through the Scsi Tool Object to the
#              specified destination device.
# Details:     This method converts the CDB and Parameter List Objs
#              into hex strings.  It then sends these hex strings using
#              the Scsi Tool Object to the destination device.
# Parameter:   destination_device - device to send the Xcopy command to
# Return:      Hashref returned by ScsiTool obj send_cmd
# Throws:      NATE::BaseException
###############################################################################
sub send_cmd() {
    $Log->enter();
    my $self = shift;

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) }, );
    my %args = validate(
        @_,
        {   'dir' => {
                type     => SCALAR | UNDEF,
                optional => 1
            },
            'destination_device' => {
                type     => SCALAR ,
                required => 1
            },
        }
    );

    my $return_ref;

    my $dir = ( defined $args{'dir'} ) ? $args{'dir'} : 'write';

    # Ensure SENSE_DATA bytes are returned by spt
    $self->{'_scsi_tool_obj'}->expected_responses(
        -exp_rsp_bitmask => ScsiTool::SENSE_DATA,
        -exp_rsp_option  => ScsiTool::RSP_ADD
    );
    
    if ( defined $self->{'_parameter_list_obj'} ) {
        $return_ref = $self->{'_scsi_tool_obj'}->send_cmd(
            -device    => $args{'destination_device'},
            -cdb       => $self->{'_cdb_obj'}->to_hex_string(),
            -cdb_param => $self->{'_parameter_list_obj'}->to_hex_string(),
            -buf_len   => $self->{'_buf_len'},
            -dir       => $dir
        );
    } else {
        $return_ref = $self->{'_scsi_tool_obj'}->send_cmd(
            -device  => $args{'destination_device'},
            -cdb     => $self->{'_cdb_obj'}->to_hex_string(),
            -buf_len => $self->{'_buf_len'},
            -dir     => $dir
        );
    }

    if ( $self->{'_scsi_tool_obj'}->error() ) {
        NATE::BaseException->throw( $self->{'_scsi_tool_obj'}->errmsg() );
    }

    $self->{'_scsi_tool_obj'}->expected_responses(
        -exp_rsp_bitmask => ScsiTool::SENSE_DATA,
        -exp_rsp_option  => ScsiTool::RSP_DEL,
    );
    
    $Log->exit();
    return $return_ref;
}

################################################################################
# Method Name: get_cdb_obj
# Objective:   Basic Getter (return CDB obj)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_cdb_obj() {
    my $self = shift;
    return $self->{'_cdb_obj'};
}

################################################################################
# Method Name: get_parameter_list_obj
# Objective:   Basic Getter (return ParameterList obj)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_parameter_list_obj() {
    my $self = shift;
    return $self->{'_parameter_list_obj'};
}

################################################################################
# Method Name: get_buf_len
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_buf_len() {
    my $self = shift;
    return $self->{'_buf_len'};
}

################################################################################
# Method Name: set_cdb_obj
# Objective:   Basic setter
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_cdb_obj() {
    my $self = shift;
    $self->{'_cdb_obj'} = shift;
}

################################################################################
# Method Name: set_parameter_list_obj
# Objective:   Basic setter
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_parameter_list_obj() {
    my $self = shift;
    $self->{'_parameter_list_obj'} = shift;
}

################################################################################
# Method Name: set_buf_len
# Objective:   Basic setter
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_buf_len() {
    my $self = shift;
    $self->{'_buf_len'} = shift;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

San::ScsiTool::Xcopy_T10

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

Xcopy_T10.pm

=head1 See Also

None

=head1 Attributes

None

=head1 Revision History

12-22-2011 schroll@netapp.com: Created Interface

=head1 Public Methods

=head2 B<new()>

Class Constructor method.

=over 4

=item Inputs

=item * I<-parameter_list_obj> <required>
=item * I<-cdb_obj> <required>
=item * I<-buf_len> <optional>

Reference to a Xcopy_T10 object.

=item Outputs

Returns a Xcopy_T10 blessed object reference.

=item Examples

    # Create the Xcopy obj
    $xcopy_obj = San::ScsiTool::Xcopy_T10->new(
        'scsi_tool_obj'      => $self->{'scsi_tool_obj'},
        'cdb_obj'            => $xcopy_cdb_obj,
        'parameter_list_obj' => $parameter_list_obj
    );
    
=back

=cut
