# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Xcopy CDB Object
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::ScsiTool::Xcopy_T10::CDB;

# compiler directives
use strict;

# imports
use Tharn qw($Log);
use Tie::IxHash;

# vars
our $XCOPY_OPCODE = 0x83;

# Meta data of CDB fields
our $meta = {
    'opcode' => {
        'length'        => 2**8,
        'field_pointer' => 0
    },
    'reserve_1_b_7_5' => {
        'length'        => 2**3,
        'field_pointer' => 1
    },
    'service_action' => {
        'length'        => 2**5,
        'field_pointer' => 1
    },
    'reserve_2_9' => {
        'length'        => 2**64,
        'field_pointer' => 2
    },
    'parameter_list_length' => {
        'length'        => 2**32,
        'field_pointer' => 10
    },
    'reserve_14' => {
        'length'        => 2**8,
        'field_pointer' => 14
    },
    'control' => {
        'length'        => 2**8,
        'field_pointer' => 15
    }
};

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   See POD documentation at the end of this file for usage
###############################################################################
sub new {
    $Log->enter();
    my ( $class, %args ) = @_;

    my $self = {};
    tie %{$self}, 'Tie::IxHash';

    $self->{'_opcode'}      = $XCOPY_OPCODE;
    $self->{'_service_action'} = 0;
    $self->{'_reserve_1_b_7_5'} = 0;
    $self->{'_reserve_2_9'} = 0;
    $self->{'_reserve_14'}  = 0x00;
    $self->{'_control'}     = 0x00;

    # Bless instance into class
    bless $self, $class;

    # Setter methods called for mandatory parameter validation.
    $self->set_parameter_list_length( $args{'parameter_list_length'} );

    $Log->exit();
    return $self;
}

##############################################################################
# Method:    get_field_length
# Objective: class method that gets length of a field
###############################################################################
sub get_field_length {
    my $class = shift;
    my $field = shift;
    return $meta->{$field}->{'length'};
}

##############################################################################
# Method:    get_field_pointer
# Objective: gets field pointer for a named field
###############################################################################
sub get_field_pointer {
    my $self  = shift;
    my $field = shift;
    return $meta->{$field}->{'field_pointer'};
}

##############################################################################
# Method:    get_bit_pointer
# Objective: gets bit pointer for a named field
###############################################################################
sub get_bit_pointer {
    my $self  = shift;
    my $field = shift;
    return $meta->{$field}->{'bit_pointer'};
}

##############################################################################
# Method:    set_field
# Objective: sets a named field with a value
###############################################################################
sub set_field {
    my $self   = shift;
    my $name   = shift;
    my $val    = shift;
    my $method = "set_" . $name;
    $self->$method($val);
}

################################################################################
# Method Name: to_hex_string
# Objective:   Converts object to a scalar hex string representing the relevant
#              bytes of the Xcopy Target Descriptor.
# Details:     The returned string is formatted as space separated string of
#              sense hexadecimal bytes.  ie.   "00 00 00 00 00 00 00 00"
################################################################################
sub to_hex_string() {
    $Log->enter();
    my $self = shift;

    # Convert reserve_1_b_7_5 to a bit string (32 chars)
    my $reserve_1_b_7_5_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_reserve_1_b_7_5'} ) ),
        29 );

    my $service_action_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_service_action'} ) ),
        27 );

    my $byte_1_bin_str
        = $reserve_1_b_7_5_bin_str
        . $service_action_bin_str;
                
    my $hex_str
        = sprintf( "%02x",  $self->{'_opcode'} )
        . unpack( "H*", pack( "B*", $byte_1_bin_str ) )
        . sprintf( "%016x", $self->{'_reserve_2_9'} )
        . sprintf( "%08x",  $self->{'_parameter_list_length'} )
        . sprintf( "%02x",  $self->{'_reserve_14'} )
        . sprintf( "%02x",  $self->{'_control'} );
        
    # Put a comma after every two characters and trim the trailing whitespace
    $hex_str =~ s/(.{2})/$1 /g;
    $hex_str =~ s/\s+$//;

    $Log->exit();
    return $hex_str;
}

################################################################################
# Method Name: get_opcode
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_opcode() {
    my $self = shift;
    return $self->{'_opcode'};
}

################################################################################
# Method Name: get_reserve_2_9
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_2_9() {
    my $self = shift;
    return $self->{'_reserve_2_9'};
}

################################################################################
# Method Name: get_reserve_list_length
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_parameter_list_length() {
    my $self = shift;
    return $self->{'_parameter_list_length'};
}

################################################################################
# Method Name: get_reserve_14
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_14() {
    my $self = shift;
    return $self->{'_reserve_14'};
}

################################################################################
# Method Name: get_reserve_14
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_control() {
    my $self = shift;
    return $self->{'_control'};
}

################################################################################
# Method Name: set_reserve_1_b_7_5
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_1_b_7_5() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-2^3 - 1')
        if !defined $arg || ( $arg < 0 || $arg > 2 ** 3 - 1 );
    $self->{'_reserve_1_b_7_5'} = $arg;
}

################################################################################
# Method Name: set_service_action
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_service_action() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-2^5 - 1')
        if !defined $arg || ( $arg < 0 || $arg > 2**5 - 1 );
    $self->{'_service_action'} = $arg;
}

################################################################################
# Method Name: set_reserve_2_9
# Objective:   setter that verifies hex string
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_2_9() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 to 2^64 - 1')
        if !defined $arg || ( $arg < 0 || $arg > ( 2**64 - 1 ) );
    $self->{'_reserve_2_9'} = $arg;
}

################################################################################
# Method Name: set_parameter_list_length
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_parameter_list_length() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 to 2^32 - 1')
        if !defined $arg || ( $arg < 0 || $arg > ( 2**32 - 1 ) );
    $self->{'_parameter_list_length'} = $arg;
}

################################################################################
# Method Name: set_reserve_14
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_14() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-255')
        if !defined $arg || ( $arg < 0 || $arg > 255 );
    $self->{'_reserve_14'} = $arg;
}

################################################################################
# Method Name: set_control
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_control() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-255')
        if !defined $arg || ( $arg < 0 || $arg > 255 );
    $self->{'_control'} = $arg;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

San::ScsiTool::Xcopy_T10::CDB

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

CDB.pm

=head1 See Also

None

=head1 Attributes

None

=head1 Revision History

12-22-2011 schroll@netapp.com: Created Interface

=head1 Public Methods

=head2 B<new()>

Class Constructor method. 

=over 4

=item Inputs

=item * I<parameter_list_length> <required>

=item Outputs

Returns a CDB blessed object reference.

=item Examples

   try {
      # Create the Xcopy CDB obj
      my $xcopy_cdb_obj = San::ScsiTool::Xcopy_T10::CDB->new(
          'parameter_list_length' => $parameter_list_obj->get_length() );
 
   } catch NATE::BaseException with {
      my $ex = shift;
      $Log->debug( "Error message is " . $ex->text() . "\n" );
   };
         
=back

=head2 B<to_hex_string()>

Converts object to a scalar hex string representing the relevant
bytes of the Xcopy CDB.

=over 4

=item Inputs

=item Outputs

Scalar hex string representing the relevant bytes of the Xcopy CDB.

=item Examples

    my $hex_str = $cdb_obj->to_hex_string();
    
=back

=cut
