# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary ParameterList Object
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::ScsiTool::Xcopy_T10::ParameterList;

# compiler directives
use strict;

# imports
use NATE::Exceptions::Argument;
use Params::Validate qw( :all );
use San::SanUtil;
use Tharn qw($Log);
use Tie::IxHash;

our $FIXED_LENGTH = 16;

# Meta data of Parameter List fields
our $meta = {
    'list_identifier' => {
        'length'        => 2**8,
        'field_pointer' => 0
    },
    'reserve_1_b_7_6' => {
        'length'        => 4,
        'field_pointer' => 1,
        'bit_pointer'   => 7
    },
    'str' => {
        'length'        => 2,
        'field_pointer' => 1,
        'bit_pointer'   => 5
    },
    'list_id_usage' => {
        'length'        => 4,
        'field_pointer' => 1,
        'bit_pointer'   => 4
    },
    'priority' => {
        'length'        => 4,
        'field_pointer' => 1,
        'bit_pointer'   => 2
    },
    'target_descriptor_list_length' => {
        'length'        => 65536,
        'field_pointer' => 2
    },
    'reserve_4_7' => {
        'length'        => 2**32,
        'field_pointer' => 4
    },
    'segment_descriptor_list_length' => {
        'length'        => 65536,
        'field_pointer' => 8
    },
    'inline_data_length' => {
        'length'        => 2**32,
        'field_pointer' => 12
    }
};

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   See POD documentation at the end of this file for usage
###############################################################################
sub new {
    $Log->enter();
    my $class = shift;

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) }, );
    my %args = validate(
        @_,
        {   'target_descriptor_objs' => {
                type     => ARRAYREF,
                required => 1
            },
            'segment_descriptor_objs' => {
                type     => ARRAYREF,
                required => 1
            }
        }
    );

    my @target_descriptor_objs  = @{ $args{'target_descriptor_objs'} };
    my @segment_descriptor_objs = @{ $args{'segment_descriptor_objs'} };
    my ( $target_descriptor_list_length, $segment_descriptor_list_length );
    my $self = {};
    tie %{$self}, 'Tie::IxHash';

    foreach my $target_descriptor_obj (@target_descriptor_objs) {

        # Calculate the length of the Target Descriptors
        $target_descriptor_list_length
            += $target_descriptor_obj->get_length();

        # Link the Target Descriptor to the Parameter List object
        $target_descriptor_obj->set_parameter_list_obj($self);
    }

    foreach my $segment_descriptor_obj (@segment_descriptor_objs) {

        # Calculate the length of the Segment Descriptors
        $segment_descriptor_list_length
            += $segment_descriptor_obj->get_length();

        # Link the Segment Descriptor to the Parameter List object
        $segment_descriptor_obj->set_parameter_list_obj($self);
    }

    $self->{'_length'}
        = $FIXED_LENGTH 
        + $target_descriptor_list_length
        + $segment_descriptor_list_length;
    $self->{'_list_identifier'} = 0x00;
    $self->{'_reserve_1_b_7_6'} = 0;
    $self->{'_str'}             = 0;
    $self->{'_list_id_usage'}   = 3;      # 11b
    $self->{'_priority'}        = 0;
    $self->{'_target_descriptor_list_length'}
        = $target_descriptor_list_length;
    $self->{'_reserve_4_7'} = 0x00000000;
    $self->{'_segment_descriptor_list_length'}
        = $segment_descriptor_list_length;
    $self->{'_inline_data_length'}      = 0;
    $self->{'_target_descriptor_objs'}  = \@target_descriptor_objs;
    $self->{'_segment_descriptor_objs'} = \@segment_descriptor_objs;
    $self->{'_inline_data'}             = undef;
    $self->{'_xcopy_obj'}               = undef;

    # bless instance into class
    bless $self, $class;
    $Log->exit();
    return $self;
}

##############################################################################
# Method:    get_field_length
# Objective: class method that gets length of a field
###############################################################################
sub get_field_length {
    my $class = shift;
    my $field = shift;
    return $meta->{$field}->{'length'};
}

##############################################################################
# Method:    get_field_pointer
# Objective: gets field pointer for a named field
###############################################################################
sub get_field_pointer {
    my $self  = shift;
    my $field = shift;
    return $meta->{$field}->{'field_pointer'};
}

##############################################################################
# Method:    get_bit_pointer
# Objective: gets bit pointer for a named field
###############################################################################
sub get_bit_pointer {
    my $self  = shift;
    my $field = shift;
    return $meta->{$field}->{'bit_pointer'};
}

##############################################################################
# Method:    set_field
# Objective: sets a named field with a value
###############################################################################
sub set_field {
    my $self   = shift;
    my $name   = shift;
    my $val    = shift;
    my $method = "set_" . $name;
    $self->$method($val);
}

##############################################################################
# Method:    add_target_descriptor_obj
# Objective: Adds a target descriptor object to the ParameterList
# Details:   See POD documentation at the end of this file for usage
###############################################################################
sub add_target_descriptor_obj {
    $Log->enter();
    my $self = shift;

    my %args = validate(
        @_,
        {   'target_descriptor_obj' => {
                type     => OBJECT,
                required => 1,
                isa      => 'San::ScsiTool::Xcopy_T10::TargetDescriptor'
            }
        }
    );

    my $td = $args{'target_descriptor_obj'};

    # Push the target descriptor to the existing array
    push( @{ $self->get_target_descriptor_objs() }, $td );

    # Add the length of the new target descriptor to the existing td list len
    $self->set_target_descriptor_list_length(
        $self->get_target_descriptor_list_length() + $td->get_length() );

   # Add the length of the new target descriptor to the existing ParamList len
    my $param_list_length = $self->get_length() + $td->get_length();
    $self->set_length($param_list_length);

    # Update parameter_list_length of CDB object
    my $cdb_obj = $self->{'_xcopy_obj'}->get_cdb_obj();
    $cdb_obj->set_parameter_list_length($param_list_length);

    $self->{'_xcopy_obj'}->set_buf_len( $param_list_length );
    
    $Log->exit();
}

##############################################################################
# Method:    add_segment_descriptor_obj
# Objective: Adds a segment descriptor object to the ParameterList
# Details:   See POD documentation at the end of this file for usage
###############################################################################
sub add_segment_descriptor_obj {
    $Log->enter();
    my $self = shift;

    my %args = validate(
        @_,
        {   'segment_descriptor_obj' => {
                type     => OBJECT,
                required => 1,
                isa      => 'San::ScsiTool::Xcopy_T10::SegmentDescriptor'
            }
        }
    );

    my $sd = $args{'segment_descriptor_obj'};

    # Push the segment descriptor to the existing array
    push( @{ $self->get_segment_descriptor_objs() }, $sd );

    # Add the length of the new segment descriptor to the existing sd list len
    $self->set_segment_descriptor_list_length(
        $self->get_segment_descriptor_list_length() + $sd->get_length() );

    # Add the length of the new segment descriptor to the existing ParamList len
    my $param_list_length = $self->get_length() + $sd->get_length();
    $self->set_length($param_list_length);

    # Update parameter_list_length of CDB object
    my $cdb_obj = $self->{'_xcopy_obj'}->get_cdb_obj();
    $cdb_obj->set_parameter_list_length($param_list_length);

    $self->{'_xcopy_obj'}->set_buf_len( $param_list_length );
    
    $Log->exit();
}

################################################################################
# Method Name: to_hex_string
# Objective:   Converts object to a scalar hex string representing the relevant
#              bytes of the Xcopy Parameter List.
# Details:     The returned string is formatted as space separated string of
#              sense hexadecimal bytes.  ie.   "00 00 00 00 00 00 00 00"
################################################################################
sub to_hex_string() {
    $Log->enter();
    my ( $self, %args ) = @_;

    # Get the space separated hex string of the Target Descriptor Objs.
    my $target_descriptors_hex_string = '';
    foreach
        my $target_descriptor_obj ( @{ $self->{'_target_descriptor_objs'} } )
    {
        $target_descriptors_hex_string
            .= $target_descriptor_obj->to_hex_string() . ' ';
    }

    # Get the space separated hex string of the Segment Descriptor Objs.
    my $segment_descriptors_hex_string = '';
    foreach my $segment_descriptor_obj (
        @{ $self->{'_segment_descriptor_objs'} } ) {
        $segment_descriptors_hex_string
            .= $segment_descriptor_obj->to_hex_string() . ' ';
    }

    # Convert reserve_1_b_7_6 to a bit string (32 chars) and only keep the
    # last 5 chars
    my $reserve_1_b_7_6_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_reserve_1_b_7_6'} ) ),
        30 );

    my $list_id_usage_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_list_id_usage'} ) ),
        30 );

    my $priority_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_priority'} ) ), 29 );

    my $byte_1_bin_str
        = $reserve_1_b_7_6_bin_str
        . $self->{'_str'}
        . $list_id_usage_bin_str
        . $priority_bin_str;

    my $hex_str
        = sprintf( "%02x", $self->{'_list_identifier'} )
        . unpack( "H*", pack( "B*", $byte_1_bin_str ) )
        . sprintf( "%04x", $self->{'_target_descriptor_list_length'} )
        . sprintf( "%08x", $self->{'_reserve_4_7'} )
        . sprintf( "%08x", $self->{'_segment_descriptor_list_length'} )
        . sprintf( "%08x", $self->{'_inline_data_length'} );

    # Put a comma after every two characters
    $hex_str =~ s/(.{2})/$1 /g;

    return
          $hex_str
        . $target_descriptors_hex_string
        . $segment_descriptors_hex_string
        . $self->{'_inline_data'};

    $Log->exit();
    return $hex_str;
}

################################################################################
# Method Name: get_xcopy_obj
# Objective:   Basic Getter (return Xcopy obj)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_xcopy_obj() {
    my $self = shift;
    return $self->{'_xcopy_obj'};
}

################################################################################
# Method Name: get_length
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_length() {
    my $self = shift;
    return $self->{'_length'};
}

################################################################################
# Method Name: get_list_identifier
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_list_identifier() {
    my $self = shift;
    return $self->{'_list_identifier'};
}

################################################################################
# Method Name: get_reserve_1_b_7_6
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_1_b_7_6() {
    my $self = shift;
    return $self->{'_reserve_1_b_7_6'};
}

################################################################################
# Method Name: get_str
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_str() {
    my $self = shift;
    return $self->{'_str'};
}

################################################################################
# Method Name: get_list_id_usage
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_list_id_usage() {
    my $self = shift;
    return $self->{'_list_id_usage'};
}

################################################################################
# Method Name: get_priority
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_priority() {
    my $self = shift;
    return $self->{'_priority'};
}

################################################################################
# Method Name: get_target_descriptor_list_length
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_target_descriptor_list_length() {
    my $self = shift;
    return $self->{'_target_descriptor_list_length'};
}

################################################################################
# Method Name: get_reserve_4_7
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_4_7() {
    my $self = shift;
    return $self->{'_reserve_4_7'};
}

################################################################################
# Method Name: get_segment_descriptor_list_length
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_segment_descriptor_list_length() {
    my $self = shift;
    return $self->{'_segment_descriptor_list_length'};
}

################################################################################
# Method Name: get_inline_data_length
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_inline_data_length() {
    my $self = shift;
    return $self->{'_inline_data_length'};
}

################################################################################
# Method Name: get_target_descriptor_objs
# Objective:   Basic Getter (return arrayref of target descriptor objs)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_target_descriptor_objs() {
    my $self = shift;

    if ( @_ > 0 ) {
        my $index = shift;
        return $self->{'_target_descriptor_objs'}->[$index];
    }
    return $self->{'_target_descriptor_objs'};
}

################################################################################
# Method Name: get_segment_descriptor_objs
# Objective:   Basic Getter (return arrayref of segment descriptor objs)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_segment_descriptor_objs() {
    my $self = shift;

    if ( @_ > 0 ) {
        my $index = shift;
        return $self->{'_segment_descriptor_objs'}->[$index];
    }
    return $self->{'_segment_descriptor_objs'};
}

################################################################################
# Method Name: get_inline_data
# Objective:   Basic Getter
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_inline_data() {
    my $self = shift;
    return $self->{'_inline_data'};
}

################################################################################
# Method Name: set_xcopy_obj
# Objective:   Basic Setter
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_xcopy_obj() {
    my $self = shift;
    $self->{'_xcopy_obj'} = shift;
}

################################################################################
# Method Name: set_length
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_length() {
    my $self = shift;
    $self->{'_length'} = shift;
}

################################################################################
# Method Name: set_list_identifier
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_list_identifier() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-255')
        if !defined $arg || ( $arg < 0 || $arg > 255 );
    $self->{'_list_identifier'} = $arg;
}

################################################################################
# Method Name: set_reserve_1_b_7_6
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_1_b_7_6() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-3')
        if !defined $arg || ( $arg < 0 || $arg > 3 );
    $self->{'_reserve_1_b_7_6'} = $arg;
}

################################################################################
# Method Name: set_str
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_str() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 or 1')
        if !defined $arg || ( $arg != 0 && $arg != 1 );
    $self->{'_str'} = $arg;
}

################################################################################
# Method Name: set_list_id_usage
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_list_id_usage() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-3')
        if !defined $arg || ( $arg < 0 || $arg > 3 );
    $self->{'_list_id_usage'} = $arg;
}

################################################################################
# Method Name: set_priority
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_priority() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 2^3')
        if !defined $arg || ( $arg < 0 || $arg > ( 2**3 - 1 ) );
    $self->{'_priority'} = $arg;
}

################################################################################
# Method Name: set_target_descriptor_list_length
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_target_descriptor_list_length() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-65535')
        if !defined $arg || ( $arg < 0 || $arg > 65535 );
    $self->{'_target_descriptor_list_length'} = $arg;
}

################################################################################
# Method Name: set_reserve_4_7
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_4_7() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 to 2^32 - 1')
        if !defined $arg || ( $arg < 0 || $arg > ( 2**32 - 1 ) );
    $self->{'_reserve_4_7'} = $arg;
}

################################################################################
# Method Name: set_segment_descriptor_list_length
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_segment_descriptor_list_length() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 to 2^32 - 1')
        if !defined $arg || ( $arg < 0 || $arg > ( 2**32 - 1 ) );
    $self->{'_segment_descriptor_list_length'} = $arg;
}

################################################################################
# Method Name: set_inline_data_length
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_inline_data_length() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 to 2^32 - 1')
        if !defined $arg || ( $arg < 0 || $arg > ( 2**32 - 1 ) );
    $self->{'_inline_data_length'} = $arg;
}

################################################################################
# Method Name: set_target_descriptor_objs
# Objective:   Basic Setter
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_target_descriptor_objs() {
    my $self = shift;
    $self->{'_target_descriptor_objs'} = shift;
}

################################################################################
# Method Name: set_segment_descriptor_objs
# Objective:   Basic Setter
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_segment_descriptor_objs() {
    my $self = shift;
    $self->{'_segment_descriptor_objs'} = shift;
}

################################################################################
# Method Name: set_inline_data
# Objective:   Basic Setter
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_inline_data() {
    my $self = shift;
    $self->{'_inline_data'} = shift;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

San::ScsiTool::Xcopy::ParameterList

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

ParameterList.pm 

=head1 See Also

None

=head1 Attributes

None

=head1 Revision History

12-22-2011 schroll@netapp.com: Created Interface

=head1 Public Methods

=head2 B<new()>

Class Constructor method. 

=over 4

=item Inputs

=item * I<target_descriptor_objs> <required>

Reference to array of TargetDescriptor objects.

=item * I<segment_descriptor_objs> <required>

Reference to array of SegmentDescriptor objects.

=item Outputs

Returns a ParameterList blessed object reference.

=item Examples

   try {
      # Create the Xcopy ParameterList obj
      my $param_list_obj = San::ScsiTool::Xcopy::ParameterList->new(
          'parameter_list_length' => $parameter_list_obj->get_length() );
 
   } catch NATE::BaseException with {
      my $ex = shift;
      $Log->debug( "Error message is " . $ex->text() . "\n" );
   };
         
=back

=head2 B<to_hex_string()>

Converts object to a scalar hex string representing the relevant
bytes of the Xcopy Parameter List.

=over 4

=item Inputs

=item Outputs

Scalar hex string representing the relevant bytes of the Xcopy Parameter List.

=item Examples

    my $hex_str = $parameter_list_obj->to_hex_string();
    
=back

=head2 B<add_segment_descriptor()>

Adds a SegmentDescriptor object to the current list of Segment Descriptors
and adds the length of the descriptor to the segment descriptor length of
the ParameterList.

=over 4

=item Inputs

=item * I<segment_descriptor_obj> <required>

=item Outputs

=item Examples

   try {

      $param_list_obj->add_segment_descriptor(
        'segment_descriptor_obj' => $sd_obj
      );
 
   } catch NATE::BaseException with {
      my $ex = shift;
      $Log->debug( "Error message is " . $ex->text() . "\n" );
   };
         
=back

=head2 B<add_target_descriptor()>

Adds a TargetDescriptor object to the current list of Segment Descriptors
and adds the length of the descriptor to the segment descriptor length of
the ParameterList.

=over 4

=item Inputs

=item * I<target_descriptor_obj> <required>

=item Outputs

=item Examples

   try {

      $param_list_obj->add_target_descriptor(
        'target_descriptor_obj' => $td_obj
      );
 
   } catch NATE::BaseException with {
      my $ex = shift;
      $Log->debug( "Error message is " . $ex->text() . "\n" );
   };
         
=back

=cut
