# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary SegmentDescriptor Object
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::ScsiTool::Xcopy_T10::SegmentDescriptor;

# compiler directives
use strict;

# imports
use Carp;
use NATE::Exceptions::Argument;
use Tharn qw($Log);

##############################################################################
# Method:      _init
# Objective:   Initializes instance variables
# Details:     There is no "new" Class Constructor.  SegmentDescriptor is 
#              called, which in turn, calls the _init method.
# Throws:      NATE::Exceptions::Argument
###############################################################################
sub _init {
    $Log->enter();
    my ( $self, %args ) = @_;

    # Setter methods called for mandatory parameter validation.
    $self->{'_descriptor_type_code'} = 0xe4;
    $self->{'_reserve_1_b_7_2'}      = 0;
    $self->{'_dc'}                   = 0;
    $self->{'_cat'}                  = 0;
    $self->set_source_target_descriptor_index(
        $args{'source_target_descriptor_index'} );    # 0x0000;
    $self->set_destination_target_descriptor_index(
        $args{'destination_target_descriptor_index'} );    # 0x0000;
    $self->{'_parameter_list_obj'} = undef;
    $Log->exit();
}

################################################################################
# Method Name: get_length
# Objective:   abstract - Must be implemented by inheriting class
# Details:
################################################################################
sub get_length() {
    my ($self) = @_;
    my $class = ref($self);
    croak "someone used an (abstract) SegmentDescriptor object"
        if $class eq "San::ScsiTool::Xcopy_T10::SegmentDescriptor";
    croak "call to abstract method ${class}::get_length";
}

################################################################################
# Method Name: to_hex_string
# Objective:   Converts object to a scalar hex string representing the relevant
#              bytes of the Xcopy Segment Descriptor.
# Details:     The returned string is formatted as space separated string of
#              sense hexadecimal bytes.  ie.   "00 00 00 00 00 00 00 00"
################################################################################
sub to_hex_string() {
    $Log->enter();
    my $self = shift;

# Convert reserve_1_b_7_2 to a bit string (32 chars) and only keep the last 6 chars
    my $reserve_1_b_7_2_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_reserve_1_b_7_2'} ) ),
        26 );
    my $byte_1_bin_str
        = $reserve_1_b_7_2_bin_str . $self->{'dc'} . $self->{'_cat'};

    my $hex_str
        = sprintf( "%02x", $self->{'_descriptor_type_code'} )
        . unpack( "H*", pack( "B*", $byte_1_bin_str ) )
        . sprintf( "%04x", $self->{'_descriptor_length'} )
        . sprintf( "%04x", $self->{'_source_target_descriptor_index'} )
        . sprintf( "%04x", $self->{'_destination_target_descriptor_index'} );

    # Put a comma after every two characters and trim the trailing whitespace
    $hex_str =~ s/(.{2})/$1 /g;
    $hex_str =~ s/\s+$//;

    $Log->exit();
    return $hex_str;
}

################################################################################
# Method Name: get_parameter_list_obj
# Objective:   Basic Getter (return ParameterList Obj)
# Details:
################################################################################
sub get_parameter_list_obj() {
    my $self = shift;
    return $self->{'parameter_list_obj'};
}

################################################################################
# Method Name: get_descriptor_type_code
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_descriptor_type_code() {
    my $self = shift;
    return $self->{'_descriptor_type_code'};
}

################################################################################
# Method Name: get_reserve_1_b_7_2
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_reserve_1_b_7_2() {
    my $self = shift;
    return $self->{'_reserve_1_b_7_2'};
}

################################################################################
# Method Name: get_dc
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_dc() {
    my $self = shift;
    return $self->{'dc'};
}

################################################################################
# Method Name: get_cat
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_cat() {
    my $self = shift;
    return $self->{'_cat'};
}

################################################################################
# Method Name: get_descriptor_length
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_descriptor_length() {
    my $self = shift;
    return $self->{'descriptor_length'};
}

################################################################################
# Method Name: get_source_target_descriptor_index
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_source_target_descriptor_index() {
    my $self = shift;
    return $self->{'_source_target_descriptor_index'};
}

################################################################################
# Method Name: get_destination_target_descriptor_index
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_destination_target_descriptor_index() {
    my $self = shift;
    return $self->{'_destination_target_descriptor_index'};
}

################################################################################
# Method Name: set_parameter_list_obj
# Objective:   Basic Setter
# Details:
################################################################################
sub set_parameter_list_obj() {
    my $self = shift;
    $self->{'_parameter_list_obj'} = shift;
}

################################################################################
# Method Name: set_descriptor_type_code
# Objective:   setter that verifies valid decimal value for this field
# Details:
################################################################################
sub set_descriptor_type_code() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-255')
        if !defined $arg || ( $arg < 0 || $arg > 255 );
    $self->{'_descriptor_type_code'} = $arg;
}

################################################################################
# Method Name: set_reserve_1_b_7_2
# Objective:   setter that verifies valid decimal value for this field
# Details:
################################################################################
sub set_reserve_1_b_7_2() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-31')
        if !defined $arg || ( $arg < 0 || $arg > 31 );
    $self->{'_reserve_1_b_7_2'} = $arg;
}

################################################################################
# Method Name: set_dc
# Objective:   setter that verifies valid decimal value for this field
# Details:
################################################################################
sub set_dc() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 or 1')
        if !defined $arg || ( $arg != 0 && $arg != 1 );
    $self->{'dc'} = $arg;
}

################################################################################
# Method Name: set_cat
# Objective:   setter that verifies valid decimal value for this field
# Details:
#              throws NATE::Exceptions::Argument
################################################################################
sub set_cat() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 or 1')
        if !defined $arg || ( $arg != 0 && $arg != 1 );
    $self->{'_cat'} = $arg;
}

################################################################################
# Method Name: set_descriptor_length
# Objective:   setter that verifies valid decimal value for this field
# Details:
#              throws NATE::Exceptions::Argument
################################################################################
sub set_descriptor_length() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-65535')
        if !defined $arg || ( $arg < 0 || $arg > 65535 );
    $self->{'descriptor_length'} = $arg;
}

################################################################################
# Method Name: set_source_target_descriptor_index
# Objective:   Basic Setter
# Details:
#              throws NATE::Exceptions::Argument
################################################################################
sub set_source_target_descriptor_index() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-65535')
        if !defined $arg || ( $arg < 0 || $arg > 65535 );
    $self->{'_source_target_descriptor_index'} = $arg;
}

################################################################################
# Method Name: set_destination_target_descriptor_index
# Objective:   Basic Setter
# Details:
#              throws NATE::Exceptions::Argument
################################################################################
sub set_destination_target_descriptor_index() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-65535')
        if !defined $arg || ( $arg < 0 || $arg > 65535 );
    $self->{'_destination_target_descriptor_index'} = $arg;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

San::ScsiTool::Xcopy_T10::SegmentDescriptor

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

SegmentDescriptor.pm 

=head1 See Also

None

=head1 Attributes

None

=head1 Revision History

12-22-2011 schroll@netapp.com: Created Interface

=head1 Public Methods

=head2 B<get_length()>

Abstract method that must be overriden by child class.  This is because
different types of Segment Descriptors can have different lengths.

=over 4

=item Inputs

=item Outputs

=item Examples

=back

=head2 B<to_hex_string()>

Abstract method that must be overriden by child class.  This is because
different types of Segment Descriptors will be composed of different 
bytes.

=over 4

=item Inputs

=item Outputs

=item Examples

=back

=cut
