# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Block to Block SegmentDescriptor Object
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::ScsiTool::Xcopy_T10::SegmentDescriptor::B2BSegmentDescriptor;

# compiler directives
use strict;

# imports
use NATE::Exceptions::Argument;
use San::ScsiTool::Xcopy_T10::SegmentDescriptor;
use Tharn qw($Log);
use Tie::IxHash;

# inheritance relationship
our @ISA = qw(San::ScsiTool::Xcopy_T10::SegmentDescriptor);

# vars
our $SD_LENGTH = 28;

# Meta data on Block 2 Block Segment Descriptor Fields
our $meta = {

    'descriptor_type_code' => {
        'length'        => 2**8,
        'field_pointer' => 0
    },
    'reserve_1_b_7_2' => {
        'length'        => 2**6,
        'field_pointer' => 1,
        'bit_pointer'   => 0,
    },
    'dc' => {
        'length'        => 2**8,
        'field_pointer' => 1,
        'bit_pointer'   => 0
    },
    'cat' => {
        'length'        => 2**8,
        'field_pointer' => 1,
        'bit_pointer'   => 0
    },
    'descriptor_length' => {
        'length'        => 2**16,
        'field_pointer' => 2
    },
    'source_target_descriptor_index' => {
        'length'        => 2**16,
        'field_pointer' => 4
    },
    'destination_target_descriptor_index' => {
        'length'        => 2**16,
        'field_pointer' => 6
    },
    'reserve_8' => {
        'length'        => 2**8,
        'field_pointer' => 8
    },
    'reserve_9' => {
        'length'        => 2**8,
        'field_pointer' => 9
    },
    'block_device_number_of_blocks' => {
        'length'        => 2**16,
        'field_pointer' => 10
    },
    'source_block_device_lba' => {
        'length'        => 2**64,
        'field_pointer' => 12
    },
    'destination_block_device_lba' => {
        'length'        => 2**64,
        'field_pointer' => 20
    }
};

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   Blesses an empty hash and calls _init
###############################################################################
sub new {
    $Log->enter();
    my ( $class, %args ) = @_;
    my $self = {};
    tie %{$self}, 'Tie::IxHash';
    bless $self, $class;
    $self->_init(%args);
    $Log->exit();
    return $self;
}

##############################################################################
# Method:    _init
# Objective: Class Constructor.
# Details:   There is no "new" Class Constructor.  TargetDescriptor is called,
#            which in turn, calls the _init method.
###############################################################################
sub _init {
    $Log->enter();
    my ( $self, %args ) = @_;

    # Call TargetDescriptor's _init method to initialize its variables.
    $self->San::ScsiTool::Xcopy_T10::SegmentDescriptor::_init(%args);

    # Setter methods called for mandatory parameter validation.
    $self->{'_descriptor_type_code'} = 0x02;
    $self->{'_descriptor_length'}    = 0x18;
    $self->{'_reserve_8'}            = 0x00;
    $self->{'_reserve_9'}            = 0x00;
    $self->set_block_device_number_of_blocks(
        $args{'block_device_number_of_blocks'} );
    $self->set_source_block_device_lba( $args{'source_block_device_lba'} );
    $self->set_destination_block_device_lba(
        $args{'destination_block_device_lba'} );

    $Log->exit();
}

##############################################################################
# Method:    get_field_length
# Objective: class method that gets length of a field
###############################################################################
sub get_field_length {
    my $class = shift;
    my $field = shift;
    return $meta->{$field}->{'length'};
}

##############################################################################
# Method:    get_field_pointer
# Objective: gets field pointer for a named field
###############################################################################
sub get_field_pointer {
    my $self  = shift;
    my $field = shift;
    return $meta->{$field}->{'field_pointer'};
}

##############################################################################
# Method:    get_bit_pointer
# Objective: gets bit pointer for a named field
###############################################################################
sub get_bit_pointer {
    my $self  = shift;
    my $field = shift;
    return $meta->{$field}->{'bit_pointer'};
}

##############################################################################
# Method:    set_field
# Objective: sets a named field with a value
###############################################################################
sub set_field {
    my $self   = shift;
    my $name   = shift;
    my $val    = shift;
    my $method = "set_" . $name;
    $self->$method($val);
}

################################################################################
# Method Name: to_hex_string
# Objective:   Converts object to a scalar hex string representing the relevant
#              bytes of the Xcopy Segment Descriptor.
# Details:     The returned string is formatted as space separated string of
#              sense hexadecimal bytes.  ie.   "00 00 00 00 00 00 00 00"
################################################################################
sub to_hex_string() {
    $Log->enter();
    my $self = shift;

    # Get the hex string for Bytes 0-7 from SegmentDescriptor.pm
    my $hex_str
        = $self->San::ScsiTool::Xcopy_T10::SegmentDescriptor::to_hex_string();

    my $b2b_hex_str
        = sprintf( "%02x",  $self->{'_reserve_8'} )
        . sprintf( "%02x",  $self->{'_reserve_9'} )
        . sprintf( "%04x",  $self->{'_block_device_number_of_blocks'} )
        . sprintf( "%016x", $self->{'_source_block_device_lba'} )
        . sprintf( "%016x", $self->{'_destination_block_device_lba'} );

    # Put a comma after every two characters and trim the trailing whitespace
    $b2b_hex_str =~ s/(.{2})/$1 /g;
    $b2b_hex_str =~ s/\s+$//;

    $Log->exit();
    return $hex_str . ' ' . $b2b_hex_str;
}

################################################################################
# Method Name: get_length
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_length() {
    my $self = shift;
    return $SD_LENGTH;
}

################################################################################
# Method Name: get_reserve_8
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_8() {
    my $self = shift;
    return $self->{'_reserve_8'};
}

################################################################################
# Method Name: get_reserve_9
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_9() {
    my $self = shift;
    return $self->{'_reserve_9'};
}

################################################################################
# Method Name: get_block_device_number_of_blocks
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_block_device_number_of_blocks() {
    my $self = shift;
    return $self->{'_block_device_number_of_blocks'};
}

################################################################################
# Method Name: get_source_block_device_lba
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_source_block_device_lba() {
    my $self = shift;
    return $self->{'_source_block_device_lba'};
}

################################################################################
# Method Name: get_destination_block_device_lba
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_destination_block_device_lba() {
    my $self = shift;
    return $self->{'_destination_block_device_lba'};
}

################################################################################
# Method Name: set_reserve_8
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_8() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-255')
        if !defined $arg || ( $arg < 0 || $arg > 255 );
    $self->{'_reserve_8'} = $arg;
}

################################################################################
# Method Name: set_reserve_9
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_9() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-255')
        if !defined $arg || ( $arg < 0 || $arg > 255 );
    $self->{'_reserve_9'} = $arg;
}

################################################################################
# Method Name: set_block_device_number_of_blocks
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_block_device_number_of_blocks() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-65535')
        if !defined $arg || ( $arg < 0 || $arg > 65535 );
    $self->{'_block_device_number_of_blocks'} = $arg;
}

################################################################################
# Method Name: set_source_block_device_lba
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_source_block_device_lba() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 to 2^64 - 1')
        if !defined $arg || ( $arg < 0 || $arg > ( 2**64 - 1 ) );
    $self->{'_source_block_device_lba'} = $arg;
}

################################################################################
# Method Name: set_destination_block_device_lba
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_destination_block_device_lba() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 to 2^64 - 1')
        if !defined $arg || ( $arg < 0 || $arg > ( 2**64 - 1 ) );
    $self->{'_destination_block_device_lba'} = $arg;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

San::ScsiTool::Xcopy_T10::SegmentDescriptor::B2BSegmentDescriptor

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

B2BSegmentDescriptor.pm

=head1 See Also

None

=head1 Attributes

None

=head1 Revision History

12-22-2011 schroll@netapp.com: Created Interface

=head1 Public Methods

=head2 B<new()>

Class Constructor method. 

=over 4

=item Inputs

=item * I<source_target_descriptor_index> <required>
=item * I<destination_target_descriptor_index> <required>
=item * I<block_device_number_of_blocks> <required>
=item * I<source_block_device_lba> <required>
=item * I<destination_block_device_lba> <required>

=item Outputs

Returns a B2BSegmentDescriptor blessed object reference.

=item Examples

   #
   # Create a B2BSegmentDescriptor object
   #
   try {
       my $sd_obj =
       San::ScsiTool::Xcopy_T10::SegmentDescriptor::B2BSegmentDescriptor->new(
		 'source_target_descriptor_index'       => 0,
		 'destination_target_descriptor_index'  => 1,
	     'block_device_number_of_blocks'        => 8,
		 'source_block_device_lba'              => 8192,
		 'destination_block_device_lba'         => 8232,
       )
   } catch (NATE::Exceptions::Argument) {};	

=back

=head2 B<get_length()>

Returns the length defined for a Block to Block Segment descriptor, as 
defined in spc4r32.  This value is 32.

=over 4

=item Inputs

=item Outputs

The length of a Block to Block Segment Descriptor.

=item Examples

    my $len = $sd_obj->get_length();
    
=back

=head2 B<to_hex_string()>

Converts object to a scalar hex string representing the relevant
bytes of the Xcopy Segment Descriptor.

=over 4

=item Inputs

=item Outputs

Scalar hex string representing the relevant bytes of the Xcopy Segment Descriptor.

=item Examples

    my $hex_str = B2BSegmentDescriptor->to_hex_string();
    
=back

=cut
