# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary TargetDescriptor Object
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::ScsiTool::Xcopy_T10::TargetDescriptor;

# compiler directives
use strict;

# imports
use Carp;
use NATE::Exceptions::Argument;
use
    San::ScsiTool::Xcopy_T10::TargetDescriptor::DeviceTypeSpecificParams::BlockDeviceTypeSpecificParams;
use Tharn qw($Log);

##############################################################################
# Method:      _init
# Objective:   Initializes instance variables
# Details:     There is no "new" Class Constructor.  TargetDescriptor is 
#              called, which in turn, calls the _init method.
# Throws:      NATE::Exceptions::Argument
###############################################################################
sub _init {
    $Log->enter();
    my ( $self, %args ) = @_;

    # If caller doesn't pass in a DeviceTypeSpecificParams obj, create
    # a BlockDeviceTypeSpecificParams object by default..
    my $device_type_specific_params_obj;

    if ( defined $args{'device_type_specific_parameters_obj'} ) {
        $self->{'_device_type_specific_parameter_obj'}
            = $args{'device_type_specific_parameters_obj'};
    } else {
        $self->{'_device_type_specific_parameter_obj'}
            = San::ScsiTool::Xcopy_T10::TargetDescriptor::DeviceTypeSpecificParams::BlockDeviceTypeSpecificParams
            ->new();
    }

    $self->{'_peripheral_device_type'}
        = $self->{'_device_type_specific_parameter_obj'}
        ->get_peripheral_device_type();

    $self->{'_relative_initiator_port'} = 0x0000;
    $self->{'_descriptor_type_code'}    = 0;
    $self->{'_nul'}                     = 0;
    $self->{'_lu_id_type'}              = 0;
    $self->{'_parameter_list_obj'}      = undef;
    $Log->exit();
}

################################################################################
# Method Name: get_length
# Objective:   abstract - Must be implemented by inheriting class
# Details:
################################################################################
sub get_length() {
    my ($self) = @_;
    my $class = ref($self);
    croak "someone used an (abstract) TargetDescriptor object"
        if $class eq "San::ScsiTool::Xcopy_T10::TargetDescriptor";
    croak "call to abstract method ${class}::get_length";
}

################################################################################
# Method Name: to_hex_string
# Objective:   Converts object to a scalar hex string representing the relevant
#              bytes of the Xcopy Target Descriptor.
# Details:     The returned string is formatted as space separated string of
#              sense hexadecimal bytes.  ie.   "00 00 00 00 00 00 00 00"
################################################################################
sub to_hex_string() {
    $Log->enter();
    my ( $self, %args ) = @_;

    my $lu_id_type_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_lu_id_type'} ) ), 30 );

    my $peripheral_device_type_bin_str
        = substr(
        unpack( "B32", pack( "N", $self->{'_peripheral_device_type'} ) ),
        27 );

    my $byte_1_bin_str
        = $lu_id_type_bin_str
        . $self->{'_nul'}
        . $peripheral_device_type_bin_str;

    my $hex_str
        = sprintf( "%02x", $self->{'_descriptor_type_code'} )
        . unpack( "H*", pack( "B*", $byte_1_bin_str ) )
        . sprintf( "%04x", $self->{'_relative_initiator_port'} );

    # Put a comma after every two characters and trim the trailing whitespace
    $hex_str =~ s/(.{2})/$1 /g;
    $hex_str =~ s/\s+$//;

    $Log->exit();
    return $hex_str;
}

################################################################################
# Method Name: get_parameter_list_obj
# Objective:   Basic Getter (return ParameterList obj)
# Details:
################################################################################
sub get_parameter_list_obj() {
    my $self = shift;
    return $self->{'_parameter_list_obj'};
}

################################################################################
# Method Name: get_descriptor_type_code
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_descriptor_type_code() {
    my $self = shift;
    return $self->{'_descriptor_type_code'};
}

################################################################################
# Method Name: get_lu_id_type
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_lu_id_type() {
    my $self = shift;
    return $self->{'_lu_id_type'};
}

################################################################################
# Method Name: get_nul
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_nul() {
    my $self = shift;
    return $self->{'_nul'};
}

################################################################################
# Method Name: get_peripheral_device_type
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_peripheral_device_type() {
    my $self = shift;
    return $self->{'_peripheral_device_type'};
}

################################################################################
# Method Name: get_relative_initiator_port
# Objective:   Basic Getter (return scalar number in decimal)
# Details:
################################################################################
sub get_relative_initiator_port() {
    my $self = shift;
    return $self->{'_relative_initiator_port'};
}

################################################################################
# Method Name: get_device_type_specific_parameter_obj
# Objective:   Basic Getter (return DeviceTypeSpecificParams obj)
# Details:
################################################################################
sub get_device_type_specific_parameter_obj() {
    my $self = shift;
    return $self->{'_device_type_specific_parameter_obj'};
}

################################################################################
# Method Name: set_descriptor_type_code
# Objective:   Basic Setter
# Details:
################################################################################
sub set_parameter_list_obj() {
    my $self = shift;
    $self->{'_parameter_list_obj'} = shift;
}

################################################################################
# Method Name: set_descriptor_type_code
# Objective:   setter that verifies valid decimal value for this field
# Details:
################################################################################
sub set_descriptor_type_code() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-255')
        if !defined $arg || ( $arg < 0 || $arg > 255 );
    $self->{'_descriptor_type_code'} = $arg;
}

################################################################################
# Method Name: set_descriptor_type_code
# Objective:   setter that verifies valid decimal value for this field
# Details:
################################################################################
sub set_lu_id_type() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-3')
        if !defined $arg || ( $arg < 0 || $arg > 3 );
    $self->{'_lu_id_type'} = $arg;
}

################################################################################
# Method Name: set_descriptor_type_code
# Objective:   setter that verifies valid decimal value for this field
# Details:
################################################################################
sub set_nul() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 or 1')
        if !defined $arg || ( $arg != 0 && $arg != 1 );
    $self->{'_nul'} = $arg;
}

################################################################################
# Method Name: set_descriptor_type_code
# Objective:   setter that verifies valid decimal value for this field
# Details:
################################################################################
sub set_peripheral_device_type() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-31')
        if !defined $arg || ( $arg < 0 || $arg > 31 );
    $self->{'_peripheral_device_type'} = $arg;
}

################################################################################
# Method Name: set_descriptor_type_code
# Objective:   setter that verifies valid decimal value for this field
# Details:
################################################################################
sub set_relative_initiator_port() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-65535')
        if !defined $arg || ( $arg < 0 || $arg > 65535 );
    $self->{'_relative_initiator_port'} = $arg;
}

################################################################################
# Method Name: set_descriptor_type_code
# Objective:   Basic Setter
# Details:
################################################################################
sub set_device_type_specific_parameter_obj() {
    my $self = shift;
    $self->{'_device_type_specific_parameter_obj'} = shift;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

San::ScsiTool::Xcopy_T10::TargetDescriptor

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

TargetDescriptor.pm 

=head1 See Also

None

=head1 Attributes

None

=head1 Revision History

12-22-2011 schroll@netapp.com: Created Interface

=head1 Public Methods

=head2 B<get_length()>

Abstract method that must be overriden by child class.  This is because
different types of Segment Descriptors can have different lengths.

=over 4

=item Inputs

=item Outputs

=item Examples

=back

=head2 B<to_hex_string()>

Abstract method that must be overriden by child class.  This is because
different types of Segment Descriptors will be composed of different 
bytes.

=over 4

=item Inputs

=item Outputs

=item Examples

=back

=cut
