# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Block Device Type Specific Params Object
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::ScsiTool::Xcopy_T10::TargetDescriptor::DeviceTypeSpecificParams::BlockDeviceTypeSpecificParams;

# compiler directives
use strict;

# imports
use NATE::Exceptions::Argument;
use Params::Validate qw( :all );
use San::ScsiTool::Xcopy_T10::TargetDescriptor::DeviceTypeSpecificParams;
use Tharn qw($Log);
use Tie::IxHash;

our @ISA
    = qw( San::ScsiTool::Xcopy_T10::TargetDescriptor::DeviceTypeSpecificParams );

# Meta data on Id Target Descriptor Fields
our $meta = {

    'reserve_28_b_7_3' => {
        'length'        => 2**5,
        'field_pointer' => 28,
        'bit_pointer'   => 7
    },
    'pad' => {
        'length'        => 2**1,
        'field_pointer' => 28,
        'bit_pointer'   => 2
    },
    'reserve_28_b_1_0' => {
        'length'        => 2**2,
        'field_pointer' => 28,
        'bit_pointer'   => 1
    },
    'disk_block_length' => {
        'length'        => 2**24,
        'field_pointer' => 29
    }
};

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   Blesses an empty hash and calls _init
###############################################################################
sub new {
    $Log->enter();
    my ( $class, %args ) = @_;
    my $self = {};
    tie %{$self}, 'Tie::IxHash';
    bless $self, $class;
    $self->_init(%args);
    $Log->exit();
    return $self;
}

##############################################################################
# Method:    _init
# Objective: initialize member variables
# Details:   See POD documentation at the end of this file for usage
###############################################################################
sub _init {
    $Log->enter();
    my ( $self, %args ) = @_;

    $self->{'_reserve_28_b_7_3'}  = 0;
    $self->{'_pad'}               = 0;
    $self->{'_reserve_28_b_1_0'}  = 0;
    $self->{'_disk_block_length'} = 0x000200;
    $Log->exit();
}


##############################################################################
# Method:    get_field_length
# Objective: class method that gets length of a field
###############################################################################
sub get_field_length {
    my $class = shift;
    my $field = shift;
    return $meta->{$field}->{'length'};
}

##############################################################################
# Method:    get_field_pointer
# Objective: gets field pointer for a named field
###############################################################################
sub get_field_pointer {
    my $self = shift;
    my $field = shift;
    return $meta->{$field}->{'field_pointer'};
}

##############################################################################
# Method:    get_bit_pointer
# Objective: gets bit pointer for a named field
###############################################################################
sub get_bit_pointer {
    my $self = shift;
    my $field = shift;
    return $meta->{$field}->{'bit_pointer'};
}

##############################################################################
# Method:    set_field
# Objective: sets a named field with a value 
###############################################################################
sub set_field {
        my $self = shift;
        my $name = shift;
        my $val = shift;
        my $method = "set_" . $name;
        $self->$method( $val );
}

################################################################################
# Method Name: to_hex_string
# Objective:   Converts object to a scalar hex string representing the relevant
#              bytes of the Xcopy Block Device Type Specific Params.
# Details:     The returned string is formatted as space separated string of
#              sense hexadecimal bytes.  ie.   "00 00 00 00 00 00 00 00"
################################################################################
sub to_hex_string() {
    $Log->enter();
    my ( $self, %args ) = @_;
    my $reserve_28_b_7_3_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_reserve_28_b_7_3'} ) ),
        27 );

    my $reserve_28_b_1_0_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_reserve_28_b_1_0'} ) ),
        30 );

    my $byte_0_bin_str
        = $reserve_28_b_7_3_bin_str
        . $self->{'_pad'}
        . $reserve_28_b_1_0_bin_str;

    my $hex_str = unpack( "H*", pack( "B*", $byte_0_bin_str ) )
        . sprintf( "%06x", $self->{'_disk_block_length'} );

    # Put a comma after every two characters and trim the trailing whitespace
    $hex_str =~ s/(.{2})/$1 /g;
    $hex_str =~ s/\s+$//;

    $Log->exit();
    return $hex_str;
}

################################################################################
# Method Name: get_reserve_28_b_7_3
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_28_b_7_3() {
    my $self = shift;
    return $self->{'_reserve_28_b_7_3'};
}

################################################################################
# Method Name: get_pad
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_pad() {
    my $self = shift;
    return $self->{'_pad'};
}

################################################################################
# Method Name: get_reserve_28_b_1_0
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_28_b_1_0() {
    my $self = shift;
    return $self->{'_reserve_28_b_1_0'};
}

################################################################################
# Method Name: get_disk_block_length
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_disk_block_length() {
    my $self = shift;
    return $self->{'_disk_block_length'};
}

################################################################################
# Method Name: get_peripheral_device_type
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
#              This is actually a field of the Target Descriptor, not the
#              BlockDeviceTypeSpecificParams.  The user can pass in a
#              BlockDeviceTypeSpecificParams to the Target Descriptor
#              constructor, which will in turn, call this method to get the
#              correct default peripheral device type.
################################################################################
sub get_peripheral_device_type() {

    # Defined in Table 119 spc4r32
    # Block devices are those with peripheral device type codes 0h
    # (i.e, direct access block)
    return 0x00;
}

################################################################################
# Method Name: set_reserve_28_b_7_3
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_28_b_7_3() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-31')
        if !defined $arg || ( $arg < 0 || $arg > 31 );
    $self->{'set_reserve_28_b_7_3'} = $arg;
}

################################################################################
# Method Name: set_pad
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_pad() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 or 1')
        if !defined $arg || ( $arg != 0 && $arg != 1 );
    $self->{'_pad'} = $arg;
}

################################################################################
# Method Name: set_reserve_28_b_1_0
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_28_b_1_0() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-3')
        if !defined $arg || ( $arg < 0 || $arg > 3 );
    $self->{'_reserve_28_b_1_0'} = $arg;
}

################################################################################
# Method Name: set_disk_block_length
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_disk_block_length() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0 to 2^24 - 1')
        if !defined $arg || ( $arg < 0 || $arg > ( 2**24 - 1 ) );
    $self->{'_disk_block_length'} = $arg;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

San::ScsiTool::Xcopy_T10::TargetDescriptor::DeviceTypeSpecificParams::BlockDeviceTypeSpecificParams

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

BlockDeviceTypeSpecificParams.pm 

=head1 See Also

None

=head1 Attributes

None

=head1 Revision History

12-22-2011 schroll@netapp.com: Created Interface

=head1 Public Methods

=head2 B<new()>

Class Constructor method.

=over 4

=item Inputs
    
=item Outputs

Returns a BlockDeviceTypeSpecificParams blessed object reference.

=item Examples

=back

=head2 B<to_hex_string()>

Converts object to a scalar hex string representing the relevant
bytes of the Xcopy Block Device Type Specific Params.

=over 4

=item Inputs

=item Outputs

Scalar hex string representing the relevant bytes of the Xcopy Block Device Type 
Specific Params.


=item Examples

    my $hex_str = $params_obj->to_hex_string();
    
=back

=head2 B<get_peripheral_device_type()>

Returns the peripheral device type associated with block devices, as defined
in Table 119 spc4r32.

=over 4

=item Inputs

=item Outputs

Scalar decimal number representing the peripheral device type for block devices

=item Examples

    my $hex_str = $params_obj->get_peripheral_device_type();
    
=back

=cut
