# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Identification TargetDescriptor Object
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::ScsiTool::Xcopy_T10::TargetDescriptor::IdTargetDescriptor;

# compiler directives
use strict;
use warnings;

# imports
use NATE::Exceptions::Argument;
use San::ScsiTool::Xcopy_T10::TargetDescriptor;
use Tharn qw($Log);
use Tie::IxHash;

# inheritance relationship
our @ISA = qw ( San::ScsiTool::Xcopy_T10::TargetDescriptor );

# vars
our $TD_LENGTH            = 32;
our $DESCRIPTOR_TYPE_CODE = 0xe4;

# Meta data on Id Target Descriptor Fields
our $meta = {

    'descriptor_type_code' => {
        'length'        => 2**8,
        'field_pointer' => 0
    },
    'lu_id_type' => {
        'length'        => 2**2,
        'field_pointer' => 1,
        'bit_pointer'   => 7
    },
    'nul' => {
        'length'        => 2**1,
        'field_pointer' => 1,
        'bit_pointer'   => 5
    },
    'peripheral_device_type' => {
        'length'        => 2**4,
        'field_pointer' => 1,
        'bit_pointer'   => 4
    },
    'relative_initiator_port_identifier' => {
        'length'        => 2**16,
        'field_pointer' => 2    
    },
    'reserved_4_b_7_4' => {
        'length'        => 2**4,
        'field_pointer' => 4,
        'bit_pointer'   => 7
    },
    'code_set' => {
        'length'        => 2**4,
        'field_pointer' => 4,
        'bit_pointer'   => 3
    },
    'reserved_5_b_7_6' => {
        'length'        => 2**2,
        'field_pointer' => 5,
        'bit_pointer'   => 7
    },
    'association' => {
        'length'        => 2**2,
        'field_pointer' => 5,
        'bit_pointer'   => 5
    },
    'designator_type' => {
        'length'        => 2**4,
        'field_pointer' => 5,
        'bit_pointer'   => 3
    },
    'reserve_6' => {
        'length'        => 2**4,
        'field_pointer' => 6
    },
    'designator_length' => {
        'length'        => 2**8,
        'field_pointer' => 7
    },
    'designator' => {
        'length'        => 2**20,
        'field_pointer' => 8
    }
};

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   Blesses an empty hash and calls _init
#            throws NATE::Exceptions::Argument
###############################################################################
sub new {
    $Log->enter();
    my ( $class, %args ) = @_;
    my $self = {};
    tie %{$self}, 'Tie::IxHash';
    bless $self, $class;
    $self->_init(%args);
    $Log->exit();
    return $self;
}

##############################################################################
# Method:    _init
# Objective: Class Constructor.
# Details:
#            throws NATE::Exceptions::Argument
###############################################################################
sub _init {
    $Log->enter();
    my ( $self, %args ) = @_;

    # Call TargetDescriptor's _init method to initialize its variables.
    $self->San::ScsiTool::Xcopy_T10::TargetDescriptor::_init(%args);

    # Setter methods called for mandatory parameter validation.
    $self->set_code_set( $args{'code_set'} );
    $self->set_association( $args{'association'} );
    $self->set_designator_type( $args{'designator_type'} );
    $self->set_designator_length( $args{'designator_length'} );
    $self->set_designator( $args{'designator'} );
    $self->{'_descriptor_type_code'} = $DESCRIPTOR_TYPE_CODE;
    $self->{'_reserve_4_b_7_4'}      = 0;
    $self->{'_reserve_5_b_7_6'}      = 0;
    $self->{'_reserve_6'}            = 0x00;

    $Log->exit();
    return $self;
}

##############################################################################
# Method:    get_field_length
# Objective: class method that gets length of a field
###############################################################################
sub get_field_length {
    my $class = shift;
    my $field = shift;
    return $meta->{$field}->{'length'};
}

##############################################################################
# Method:    get_field_pointer
# Objective: gets field pointer for a named field
###############################################################################
sub get_field_pointer {
    my $self = shift;
    my $field = shift;
    return $meta->{$field}->{'field_pointer'};
}

##############################################################################
# Method:    get_bit_pointer
# Objective: gets bit pointer for a named field
###############################################################################
sub get_bit_pointer {
    my $self = shift;
    my $field = shift;
    return $meta->{$field}->{'bit_pointer'};
}

##############################################################################
# Method:    set_field
# Objective: sets a named field with a value 
###############################################################################
sub set_field {
        my $self = shift;
        my $name = shift;
        my $val = shift;
        my $method = "set_" . $name;
        $self->$method( $val );
}

################################################################################
# Method Name: to_hex_string
# Objective:   Converts object to a scalar hex string representing the relevant
#              bytes of the Xcopy Target Descriptor.
# Details:     The returned string is formatted as space separated string of
#              sense hexadecimal bytes.  ie.   "00 00 00 00 00 00 00 00"
################################################################################
sub to_hex_string() {
    $Log->enter();
    my $self = shift;

    # Get the hex string for Bytes 0-7 from SegmentDescriptor.pm
    my $td_hex_str
        = $self->San::ScsiTool::Xcopy_T10::TargetDescriptor::to_hex_string();

    my $reserve_4_b_7_4_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_reserve_4_b_7_4'} ) ),
        28 );

    my $code_set_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_code_set'} ) ), 28 );

    my $byte_4_bin_str = $reserve_4_b_7_4_bin_str . $code_set_bin_str;

    my $reserve_5_b_7_6_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_reserve_5_b_7_6'} ) ),
        30 );

    my $association_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_association'} ) ), 30 );

    my $designator_type_bin_str
        = substr( unpack( "B32", pack( "N", $self->{'_designator_type'} ) ),
        28 );

    my $byte_5_bin_str
        = $reserve_5_b_7_6_bin_str
        . $association_bin_str
        . $designator_type_bin_str;

    my $designator = $self->{'_designator'};
    $designator =~ s/\s//g;

  # Bytes 8-27 Can hold the designator.  Any bytes not used become reserved
  # bytes.  Since the designtor and reserved field lengths are variable, this
  # class defines only the designator field.  The designator should not exceed
  # 20 bytes
    my $id_td_hex_str
        = unpack( "H*", pack( "B*", $byte_4_bin_str ) )
        . unpack( "H*", pack( "B*", $byte_5_bin_str ) )
        . sprintf( "%02x", $self->{'_reserve_6'} )
        . sprintf( "%02x", $self->{'_designator_length'} )
        . $designator;

    # Put a comma after every two characters and trim the trailing whitespace
    $id_td_hex_str =~ s/(.{2})/$1 /g;
    $id_td_hex_str =~ s/\s+$//;

    $Log->exit();
    return
          $td_hex_str . ' '
        . $id_td_hex_str . ' '
        . $self->get_device_type_specific_parameter_obj()->to_hex_string();
}

################################################################################
# Method Name: get_length
# Objective:   Returns correct length of an Id Target Descriptor
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_length() {
    my $self = shift;
    return $TD_LENGTH;
}

################################################################################
# Method Name: get_lun_serial
# Objective:   Converts the designator bytes representing the lun
#              serial number to ascii
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_lun_serial {
    my $self = shift;
    my @hex_bytes = split( ' ', substr( $self->get_designator(), 12, 35 ) );
    my $hex_str;
    foreach my $hex_byte (@hex_bytes) {
        $hex_str .= chr( hex $hex_byte );
    }

    return $hex_str;
}

################################################################################
# Method Name: get_lun_c
# Objective:   Gets the lun component IdTargetDescriptor represents.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_lun_c {
    my $self          = shift;
    my $cmd_interface = shift;

    my $lun_serial = $self->get_lun_serial();

    my $lun_c = NACL::C::Lun->find(
        command_interface => $cmd_interface,
        filter            => { serial => $lun_serial }
    );

    return $lun_c;
}

################################################################################
# Method Name: get_reserve_4_b_7_4
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_4_b_7_4() {
    my $self = shift;
    return $self->{'_reserve_4_b_7_4'};
}

################################################################################
# Method Name: get_code_set
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_code_set() {
    my $self = shift;
    return $self->{'_code_set'};
}

################################################################################
# Method Name: get_reserve_5_b_7_6
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_5_b_7_6() {
    my $self = shift;
    return $self->{'_reserve_5_b_7_6'};
}

################################################################################
# Method Name: get_association
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_association() {
    my $self = shift;
    return $self->{'_association'};
}

################################################################################
# Method Name: get_designator_type
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_designator_type() {
    my $self = shift;
    return $self->{'_designator_type'};
}

################################################################################
# Method Name: get_reserve_6
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_reserve_6() {
    my $self = shift;
    return $self->{'_reserve_6'};
}

################################################################################
# Method Name: get_designator_length
# Objective:   Basic Getter (return scalar number in decimal)
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_designator_length() {
    my $self = shift;
    return $self->{'_designator_length'};
}

################################################################################
# Method Name: get_designator
# Objective:   Basic Getter
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_designator() {
    my $self = shift;
    return $self->{'_designator'};
}

################################################################################
# Method Name: set_reserve_4_b_7_4
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_4_b_7_4() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-15')
        if !defined $arg || ( $arg < 0 || $arg > 15 );
    $self->{'_reserve_4_b_7_4'} = $arg;
}

################################################################################
# Method Name: set_code_set
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_code_set() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-15')
        if !defined $arg || ( $arg < 0 || $arg > 15 );
    $self->{'_code_set'} = $arg;
}

################################################################################
# Method Name: set_reserve_5_b_7_6
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_reserve_5_b_7_6() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-3')
        if !defined $arg || ( $arg < 0 || $arg > 3 );
    $self->{'_reserve_5_b_7_6'} = $arg;
}

################################################################################
# Method Name: set_association
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
#              throws NATE::Exceptions::Argument
################################################################################
sub set_association() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-3')
        if !defined $arg || ( $arg < 0 || $arg > 3 );
    $self->{'_association'} = $arg;
}

################################################################################
# Method Name: set_designator_type
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
#              throws NATE::Exceptions::Argument
################################################################################
sub set_designator_type() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-15')
        if !defined $arg || ( $arg < 0 || $arg > 15 );
    $self->{'_designator_type'} = $arg;
}

################################################################################
# Method Name: set_reserve_6
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
#              throws NATE::Exceptions::Argument
################################################################################
sub set_reserve_6() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-255')
        if !defined $arg || ( $arg < 0 || $arg > 255 );
    $self->{'_reserve_6'} = $arg;
}

################################################################################
# Method Name: set_designator_length
# Objective:   setter that verifies valid decimal value for this field
# Details:     See POD documentation at the end of this file for usage
#              throws NATE::Exceptions::Argument
################################################################################
sub set_designator_length() {
    my $self = shift;
    my $arg  = shift;
    NATE::Exceptions::Argument->throw('Param must be 0-255')
        if !defined $arg || ( $arg < 0 || $arg > 255 );
    $self->{'_designator_length'} = $arg;
}

################################################################################
# Method Name: set_designator
# Objective:   setter that verifies hex string
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_designator() {
    my $self = shift;
    my $arg  = shift;

    unless ( $arg =~ /^([0-9A-Fa-f]{2}\s){19}[0-9A-Fa-f]{2}$/ ) {
        NATE::Exceptions::Argument->throw(
            'Designator must be a string of 20 space delimited 2 digit hex '
                . 'values.  ie.  00 00 .. 00' );
    }
    $self->{'_designator'} = $arg;
}

1;

# TODO usage create methods and where to put them
# TODO generic constructor and creates ? maybe later ?

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

San::ScsiTool::Xcopy_T10::TargetDescriptor::IdTargetDescriptor

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

IdTargetDescriptor.pm

The IdTargetDescriptor models the Identification Target Descriptor of spc4r32.
to_hex_string can be called to get translate the fields into a string of hex
bytes.  Constructor values come from the id descriptors returned by inq83h
or inq80h. 

=head1 See Also

None

=head1 Attributes

None

=head1 Revision History

12-22-2011 schroll@netapp.com: Created Interface

=head1 Public Methods

=head2 B<new()>

Class Constructor method.

=over 4

=item Inputs

=item * I<-code_set> <required>
=item * I<-association> <required>
=item * I<-designator_type> <required>
=item * I<-designator_length> <required>
=item * I<-designator> <required>
    
=item Outputs

Returns an IdTargetDescriptor blessed object reference.

=item Examples

=back

=head2 B<to_hex_string()>

Converts object to a scalar hex string representing the relevant
bytes of the Xcopy Identification Target Descriptor.

=over 4

=item Inputs

=item Outputs

Scalar hex string representing the relevant bytes of the Xcopy Identification
Target Descriptor

=item Examples

    my $hex_str = $td_obj->to_hex_string();
    
=back

=cut
