# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary provides task methods for San::ControllerCluster class
##
## @author dl-ssan-automation
## @status Shared

package San::Task::ClusterHelper;

# compiler directives
use strict;
use lib();

# imported Perl modules
use Data::Dumper;
use List::MoreUtils qw (any firstidx);

# Nate stuff
use Tharn;

# imported test/lib modules
use ParamsMgr;

# imported SAN modules
use San::SanUtil;
use San::SanConsts;
use San::CommonInterface;
use San::ControllerCluster::Filer10M;

# constants
use constant {
    DEFAULT_EPSILON_MODIFY_SNOOZE => 3,   # in seconds
    DEFAULT_WAIT_PROMPT_AFTER_UNJOIN_TIMEOUT => 600,
    GX_BOOT_SCRIPT => '/x/eng/localtest/noarch/main/os/boot/gx_boot.thpl'
};


###############################################################################
# Method Name: cluster_create
# Objective:   Create a cluster on the specified node.
# Details:     See POD at end of document.
###############################################################################
sub cluster_create {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my $node      = &ParamsMgr::get_arg(-arg      => "node",
                                        -required => $SanUtil::TRUE);
    my $clus_name = &ParamsMgr::get_arg(-arg      => "cluster_name",
                                        -required => $SanUtil::FALSE,
                                        -default  => &Tharn::param("CLUSTER_NAME", "-default", ""));

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        &SanUtil::nate_log("-fail", &ParamsMgr::get_error());
        return undef;
    }

    # Make sure we have a node object
    my $node_obj = &San::Task::ClusterHelper::_get_node_obj(-node => $node);
    unless (defined($node_obj)) {
        # Error is recorded in logs
        return undef;
    }

    # Grab the hostrec for easy access to natehost params.
    my $host = &Tharn::host($node_obj->get_name());

    # Check if cluster is already created
    my @output = $node_obj->invoke(-cdef => "cluster_identity_show");
    if (scalar(@output)) {
        &SanUtil::nate_log(
            "-comment",
            "Cluster already exists..Skipping cluster_create"
        );
        # Store the cluster name so we can check for cserver (should only be 1).
        $clus_name = $output[0]->{name};
    } else {
        # Clear the error if it
        if ($node_obj->errmsg() =~ /"identity" is not a recognized command/){
            $node_obj->_clear_err();
        } else {
            &SanUtil::nate_log( "-fail", "Failed to invoke cdef cluster_identity_show: " .
                               $node_obj->errmsg() );
            return undef;
        }

        # Configure Cluster Lifs
        $node_obj->configure_cluster_lifs();
        if ($node_obj->error()) {
            &SanUtil::nate_log("-fail", "Failed to create Cluster lif: ".$node_obj->errmsg());
            return undef;
        }

        # Verify there is a cluster name if not it must be different than any node
        #    in the cluster.
        unless ( length($clus_name) ) {
            $clus_name = $node_obj->get_name() . time;
        }

        # Build cluster create args
        my %clus_args = (
            clustername => $clus_name,
            license     => &San::Task::ClusterHelper::_get_cluster_license('node_obj'=>$node_obj,
                                                                           %args),
            cmd         => "cluster_create"
        );

        # use common invoke from base class (San::CommonInterface)
        # to translate cdef into UI cmd and execute it
        $node_obj->San::CommonInterface::invoke_cmd_obj(%clus_args);
        if ($node_obj->error()) {
            &SanUtil::nate_log(
                "-fail",
                "failed to invoke_cmd_obj " . $args{cmd} . $node_obj->errmsg()
            );
            return undef;
        }
    }

    # Build lif args from hostrec hash.
    my %lif_args = (
                    vserver   => $clus_name,
                    lif       => "cluster_mgmt",
                    role      => "cluster-mgmt",
                    home_node => $node_obj->get_name(),
                    home_port => $host->{CLUSTER_MGMT_PORT},
                    address   => $host->{CLUSTER_MGMT_IP},
                    netmask   => $host->{CLUSTER_MGMT_MASK},
                    gateway   => $host->{CLUSTER_MGMT_GATEWAY}
                );

    $node_obj->configure_lif(%lif_args);
    if ($node_obj->error()) {
        &SanUtil::nate_log("-fail",
                           "Failed to create cserver lif: ".$node_obj->errmsg()
                        );
        return undef;
    }

    my $clus_obj = San::ControllerCluster->new(
        -cserver => $node_obj->get_name()
    );
    if ($clus_obj->error()) {
        &SanUtil::nate_log(
            "-fail",
            "Failed to create Controller Cluster object: " . $clus_obj->errmsg()
        );
        return undef;
    };

    return $clus_obj;
} # End sub cluster_create

###############################################################################
# Method Name: join_node
# Objective:   Attempt to join node to an already existing cluster.
# Details:     See POD at end of document.
###############################################################################
sub join_node{
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my $clus_obj = &ParamsMgr::get_arg(-arg      => "cluster_obj",
                                       -ref_type => "San::ControllerCluster",
                                       -required => $SanUtil::TRUE);

    my $node = &ParamsMgr::get_arg(-arg      => "node",
                                   -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        &SanUtil::nate_log("-fail", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    $clus_obj->_clear_err();

    my @cluster_members = ();
    my $node_obj = &San::Task::ClusterHelper::_get_node_obj(-node => $node);
    unless (defined($node_obj)) {
        # Error is recorded in logs
        return $SanUtil::FALSE;
    }
    my $node_name = $node_obj->get_name();

    # Get all current cluster memeber names
    @cluster_members = $clus_obj->get_node_members();
    if ($clus_obj->error()) {
        # Error set, bubble up
        return $SanUtil::FALSE;
    }

    # Grab the IP of existing cluster node to be able to join.
    my $cluster_node_idx = firstidx { $_ ne $node_name } @cluster_members;
    my $host = &Tharn::host($cluster_members[$cluster_node_idx]);
    my $clusterjoinip = $host->{CLUSTER_IP1};

    &SanUtil::nate_log("-comment", "Cluster join for filer $node_name");
    $node_obj->cluster_join(-clusteripaddr => $clusterjoinip);
    if ($node_obj->error()) {
        $clus_obj->_set_err("Fail to join $node" . $node_obj->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End join_node

################################################################################
# Method Name: unjoin_node
# Objective:   Attempt to unjoin a node from a cluster.
# Details:     See POD documentation at the end of this file
################################################################################
sub unjoin_node {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my $clus_obj = &ParamsMgr::get_arg(-arg      => "cluster_obj",
                                       -ref_type => "San::ControllerCluster",
                                       -required => $SanUtil::TRUE);

    my $node = &ParamsMgr::get_arg(-arg      => "node",
                                   -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        &SanUtil::nate_log("-fail", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    $clus_obj->_clear_err();

    my ($node_obj, $node_name) = "";
    $node_obj = &San::Task::ClusterHelper::_get_node_obj(-clus_obj => $clus_obj,
                                                         -node     => $node);
    unless (defined($node_obj)) {
        # Error is recorded in logs
        return $SanUtil::FALSE;
    }
    $node_name = $node_obj->get_name();

    # If HA or SFO enabled, disable before attempting unjoin
    if ($node_obj->is_cluster_ha_configured()) {
        &SanUtil::nate_log(
            "-comment",
            "$node_name is HA configured, Attempting to disable."
        );
        $node_obj->ha_modify(-configured => "false");
        if ($node_obj->error()) {
            $clus_obj->_set_err("Failed to modify ha: " . $node_obj->errmsg());
            return $SanUtil::FALSE;
        }
    }
    if ($node_obj->is_sfo_enabled()) {
        &SanUtil::nate_log(
            "-comment",
            "$node_name has SFO enabled, Attempting to disable."
        );
        $node_obj->disable_sfo();
        if ($node_obj->error()) {
            $clus_obj->_set_err("Failed to disable sfo:" . $node_obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # If the unjoin node has epsilon move it to another node.
    if ($node_obj->get_epsilon() eq "true") {
        &SanUtil::nate_log("-comment", "$node_name has epsilon.");

        # Look for a new node to hold epsilon
        my @nodes = $clus_obj->get_node_members();
        # Grab the index of the first node NOT holding espilon
        my $new_node_idx = firstidx { $_ ne $node_name } @nodes;
        my $new_node_obj = $clus_obj->get_node_objs(
            -name => $nodes[$new_node_idx],
            -count => 1
        );
        if ($clus_obj->error()) {
            # Already logged error, return false.
            return $SanUtil::FALSE;
        }

        # Now move epsilon
        $node_obj->modify(-epsilon =>"false");
        if ($node_obj->error()){
            $clus_obj->_set_err(
                "Failed to modify epsilon: " . $node_obj->errmsg()
            );
            return $SanUtil::FALSE;
        }
        &SanUtil::nate_log(
            "-comment",
            "Modified epsilon to \"false\" on $node_name"
        );

        # Apparently it can take some amount of time for epsilon to propagate
        #    its state across the cluster, snooze for (seconds).
        Tharn::snooze(DEFAULT_EPSILON_MODIFY_SNOOZE);
        $new_node_obj->modify(-epsilon =>"true");
        if ($new_node_obj->error()) {
            $clus_obj->_set_err(
                "Failed to modify epsilon: " . $new_node_obj->errmsg()
            );
            return $SanUtil::FALSE;
        }
        &SanUtil::nate_log(
            "-comment",
            "Modified epsilon to \"true\" on " . $nodes[$new_node_idx]
        );
    }

    # If the cserver is on the node we are unjoining we need to look for a
    # new node to host it.
    my $cs_lif_obj = &San::Task::ClusterHelper::get_cserver_lif_obj(
        -cluster_obj => $clus_obj
    );
    if ($clus_obj->error()) {
        # Already logged error, return false
        return $SanUtil::FALSE;
    }
    if ($cs_lif_obj->get_curr_node() eq $node_name or
        $cs_lif_obj->get_home_node() eq $node_name) {
        &SanUtil::nate_log(
            "-comment",
            "$node_name hosts the cserver, it will be moved."
        );
        # Look for a new eligible node to host the cserver
        my @nodes = $clus_obj->get_node_members();
        # Grab the index of the first node which is not the current cserver host
        my $new_node_idx = firstidx { $_ ne $node_name } @nodes;
        &San::Task::ClusterHelper::migrate_cserver(
            -cluster_obj => $clus_obj,
            -to_node     => $nodes[$new_node_idx]
        );
        if ($clus_obj->error()) {
            # Already logged error, return FALSE
            return $SanUtil::FALSE;
        }
        &SanUtil::nate_log(
            "-comment",
            "Migrated the cserver to " . $nodes[$new_node_idx]
        );
    } else {
        &SanUtil::nate_log("-comment", "$node_name does NOT host the cserver.");
    }

    &SanUtil::nate_log("-comment", "Start unjoin of node $node");
    $clus_obj->unjoin(-node => $node);
    if ($clus_obj->error()) {
        # Already logged error, return fail.
        return $SanUtil::FAIL;
    }

    &SanUtil::nate_log(
        "-comment",
        "Unjoin of node $node is complete. Waiting for ontap's "
        . "obligatory wipe_config and reboot."
    );

    return $SanUtil::TRUE;
} # End unjoin_node

################################################################################
# Method Name: detect_unjoin_wipe_config_and_reboot()
# Objective:   detect wipe config and reboot after
#              San::Task::ClusterHelper::unjoin_node or
#              San::ControllerCluster::unjoin().
# Details:     See POD documentation at the end of this file
################################################################################
sub detect_unjoin_wipe_config_and_reboot {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my $clus_obj = &ParamsMgr::get_arg(-arg      => "cluster_obj",
                                       -ref_type => "San::ControllerCluster",
                                       -required => $SanUtil::TRUE);

    my $node = &ParamsMgr::get_arg(-arg      => "node",
                                   -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        &SanUtil::nate_log("-fail", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    $clus_obj->_clear_err();

    my ($node_obj, $node_name) = "";
    $node_obj = &San::Task::ClusterHelper::_get_node_obj(-clus_obj => $clus_obj,
                                                         -node     => $node);
    unless (defined($node_obj)) {
        # Error is recorded in logs
        return $SanUtil::FALSE;
    }
    $node_name = $node_obj->get_name();

    # Before calling unjoin, connect to the console of the unjoin node
    my $conn = $node_obj->console_connect(-login => $SanUtil::FALSE);
    if ($node_obj->error()) {
        $clus_obj->_set_err(
            "Failed to connect to console: " . $node_obj->errmsg()
        );
        return $SanUtil::FALSE;
    }
    # Now we wait for the node to hit the boot menu and return
    my $expected_prompt = "Selection (1-8)?";
    my (undef, undef, undef, $received_prompt) =
        &Tharn::say(
            $conn, " ",
            '-noprompt', '-prompt', $expected_prompt,
            '-timeout', DEFAULT_WAIT_PROMPT_AFTER_UNJOIN_TIMEOUT,
            '-ignore', "echo ignore",
            '-comment', "Attempted to unjoin node, waiting for reboot."
        );
    if ($received_prompt !~ /$expected_prompt/i) {
        $clus_obj->_set_err("Error: Console did not receive expected prompt"
                        . "($expected_prompt)");
        return $SanUtil::FALSE;
    }

    # disconnect from the console
    &Tharn::disconnect($conn);

    return $SanUtil::TRUE;
} # End sub detect_unjoin_wipe_config_and_reboot

################################################################################
# Method Name: get_cserver_lif_obj()
# Objective:   get the lif object for the cserver connection.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_cserver_lif_obj {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my $clus_obj = &ParamsMgr::get_arg(-arg      => "cluster_obj",
                                       -ref_type => "San::ControllerCluster",
                                       -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        &SanUtil::nate_log("-fail", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    $clus_obj->_clear_err();

    # Locals
    my $lif_obj = undef;
    my $node_obj = $clus_obj->get_node_objs(-count => 1);
    my $cserver_obj = $clus_obj->get_cserver_obj();
    # Using the node object to get the cserver lif because we need the
    # parent object to be a connection that is different from the cserver. If
    # any migrations or disruptions take place the connection will fail and the
    # operation will timeout.
    my @lif_objs = $node_obj->fetch_lif_objs(
        -vserver => $cserver_obj->get_name(),
        -role    => "cluster-mgmt"
    );
    if ($node_obj->error()) {
        # Error is set, bubble up
        return $SanUtil::FALSE;
    }

    if (scalar(@lif_objs)) {
        # There can be only one.
        $lif_obj = $lif_objs[0];
    } else {
        # Set the error on the cluster_obj
        $clus_obj->_set_err(
            "There is no cluster-mgmt interface defined in cserver."
        );
    }

    return $lif_obj;
} # End sub get_cserver_lif_obj

###############################################################################
# Method name: gx_boot()
# Objective:   subtest the gx_boot script
# Details:     See POD documentation at the end of this file
################################################################################
sub gx_boot {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my $node  = &ParamsMgr::get_arg(-arg      => "node",
                                    -required => $SanUtil::TRUE);

    my $tasks = &ParamsMgr::get_arg(-arg      => "tasks",
                                    -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        &SanUtil::nate_log("-fail", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    my $subtest = subtest (
        GX_BOOT_SCRIPT,
        'TASKS=' . $tasks,
        'NODES=' . $node,
        '-client'  , 'localhost',
        '-runid'   , "gx_boot:$tasks",
        '-ifscript', "-eval_error",
        '-iffatal' , "-eval_error",
        '-bg'
    );

    $subtest->wait_finish();

    # Verify: Check results from subtest.
    my @subtest_results = $subtest->own_results(result_type => "FAIL");
    if (scalar(@subtest_results) > 0) {
        &SanUtil::nate_log(
            "-fail",
            "gx_boot fake_init task failed: @subtest_results"
        );
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End _fake_init and setup

################################################################################
# Method Name: migrate_cserver()
# Objective:   migrates the cserver lif from current node to specified node
# Details:     See POD documentation at the end of this file
################################################################################
sub migrate_cserver{
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my $clus_obj = &ParamsMgr::get_arg(-arg      => "cluster_obj",
                                       -ref_type => "San::ControllerCluster",
                                       -required => $SanUtil::TRUE);

    my $to_node = &ParamsMgr::get_arg(-arg      => "to_node",
                                      -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        &SanUtil::nate_log("-fail", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    $clus_obj->_clear_err();
    my $lif_obj = undef;
    my ($node_obj, $node_name) = "";
    my ($cserver_obj, $cserver_name) = "";
    my ($cserver_lif_record, $cserver_lif_name) = "";

    $node_obj = &San::Task::ClusterHelper::_get_node_obj(-clus_obj => $clus_obj,
                                                         -node     => $to_node);
    unless (defined($node_obj)) {
        # Error is recorded in logs
        return $SanUtil::FALSE;
    }
    $node_name = $node_obj->get_name();
    if ($node_obj->error()) {
        $clus_obj->_set_error($node_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $cserver_obj = $clus_obj->get_cserver_obj();
    if ($clus_obj->error()) {
        # Error is set, bubble up
        return $SanUtil::FALSE;
    }
    $cserver_name = $cserver_obj->get_name();
    if ($cserver_obj->error()) {
        $clus_obj->_set_error($cserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Fetch the lif object which represents the cserver lif
    $lif_obj = &San::Task::ClusterHelper::get_cserver_lif_obj(
                                                    -cluster_obj => $clus_obj);
    if ($clus_obj->error()) {
        # Error is set, bubble up
        return $SanUtil::FALSE;
    }

    # migrate cserver lif to the other node and make that node its home.
    $lif_obj->migrate(-dest_node => $node_name);
    if ($lif_obj->error()) {
        $clus_obj->_set_err("Failed to migrate lif: " . $lif_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $lif_obj->modify(-home_node => $node_name);
    if ($lif_obj->error()) {
        $clus_obj->_set_err("Failed to modify lif: " . $lif_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Make sure we reconnect after the migration
    $cserver_obj->reconnect();
    if ($cserver_obj->error()) {
        $clus_obj->_set_error(
            "Could not reconnect to cserver after migrate: "
            . $cserver_obj->errmsg()
        );
        return $SanUtil::FALSE;
    }

    # let's verify cserver migrated.
    $lif_obj = &San::Task::ClusterHelper::get_cserver_lif_obj(
                                                    -cluster_obj => $clus_obj);
    if ($clus_obj->error()) {
        # Error is set, bubble up
        return $SanUtil::FALSE;
    }
    if ($lif_obj->get_curr_node() ne $node_name) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End migrate_cserver

################################################################################
##                             Private Functions
################################################################################

################################################################################
# Method Name: _get_cluster_license()
# Objective: Get the BASE License code for the SERIAL NUMBER of the node, it will
#           defualt to the library's default value.
# Required Params:
#    * node_obj = The object that will be used to execute all commands
# Returns:
#    * base cluster license based on node obj.
################################################################################
sub _get_cluster_license {
    my(%args) = @_;
    # Locals
    my $codes=undef;

    &ParamsMgr::process_args(%args);
    my $node_obj = &ParamsMgr::get_arg(-arg      => "node_obj",
                                       -required => $SanUtil::TRUE);
    my $serial_number = &ParamsMgr::get_arg(-arg      => "serial_number",
                                            -required => $SanUtil::FALSE,
                                            -default  => undef);
    if (&ParamsMgr::error()) {
        &SanUtil::nate_log("-fail", &ParamsMgr::get_error());
        return undef;
    }

    # Retrieve the serial number
    unless (defined($serial_number)) {
        $serial_number = $node_obj->_get_serial_number('use_default'=>1);
        if ($node_obj->error()) {
            &SanUtil::nate_log( "-fail", "Failed to retrieve the serial number: \n" .
                               $node_obj->errmsg());
            return undef;
        }
    }

    # TODO
    # Add logic to support the new SierraNevada License V2 strategy.
    # Changes should be similar to the FilerHelper::license_codes.pm that deals
    # with all the RR and SN serials.
    $codes = SanConsts::RollingRock_Licenses->{$serial_number};

    return $codes->{'base'};

}# End _get_cluster_license

################################################################################
# Method Name: _get_node_obj()
# Objective: retrieves the San::ControllerCluster::Filer10M object of node
# Required Params:
#    * node = name of the node for which the Filer10M object is requested.
# Optional Params:
#    * clus_obj = ControllerCluster object, if provided it will be used to get
#                 the requested node object.
# Returns:
#    * a node object or undef with error recorded in logs if errors occur.
################################################################################
sub _get_node_obj {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my $node     = &ParamsMgr::get_arg(-arg      => "node",
                                       -required => $SanUtil::TRUE);

    my $clus_obj = &ParamsMgr::get_arg(-arg      => "clus_obj",
                                       -ref_type => "San::ControllerCluster",
                                       -required => $SanUtil::FALSE,
                                       -default  => undef);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        &SanUtil::nate_log("-fail", &ParamsMgr::get_error());
        return undef;
    }

    my $node_obj = undef;

    if (ref($node) =~ /Filer10M/) {
        $node_obj = $node;
    } elsif (defined($clus_obj)) {
        $node_obj = $clus_obj->get_node_objs(-name => $node,
                                             -count => 1);
        if ($clus_obj->error()) {
            &SanUtil::nate_log(
                "-fail",
                "CONFIG: Failed to retrieve the node object from the clus_obj: "
                . $clus_obj->errmsg()
            );
            return undef;
        }
    } else {
        $node_obj = San::ControllerCluster::Filer10M->new(-filer => $node);
        if ($node_obj->error()) {
            &SanUtil::nate_log(
                "-fail",
                "CONFIG: Failed to create the node object: "
                . $node_obj->errmsg()
            );
            return undef;
        }
    }

    return $node_obj
} # End sub _create_node_obj

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

ControllerClusterHelper.pm

=head1 Author

(dl-ssan-automation)

=head1 Description

ClusterHelper.pm provides helper methods for ControllerCluster.pm.
This package is maintained separately from ControllerCluster.pm to prevent convoluting ControllerCluster object model.

=head1 See Also

See these modules for further interface detail.

B<San::ControllerCluster.pm, San::CommonInterface.pm>

=head1 B<Public Methods>

=head2 B<cluster_create(%args)>

Create cluster on specified node.

=over 4

=item Inputs

=item Required

=item * I<-node>

The name of the node to use or San::ControllerCluster::Filer10M object

=item Optional

=item * I<-cluster_name>

The name of the cluster to use during creation.

=item Outputs

Returns San::ControllerCluster object. if error, logs error and returns undef.

=item Examples

    $Test->step("Create cluster on $FilerA (node1)");
    $Clus_Obj = &San::Task::ClusterHelper::cluster_create(-node => $FilerA);
    if (defined($Clus_Obj)) {
        $Test->nlog("Successfully created the cluster on $FilerA");
    } else {
        $Test->nlog("Error: Failed to setup the cluster on $FilerA");
        return $TCD::FAIL;
    }

=back

=head2 B<join_node(%args)>

Task to join node to a cluster.

NOTE: This function will clear errors on the $Clus_Obj and set failures on
      error.

=over 4

=item Inputs

=item Required

=item * I<-cluster_obj>

The San::ControllerCluster object to join the node to.

=item * I<-node>

The name of the node to join or San::ControllerCluster::Filer10M object

=item Outputs

Returns true. if error, sets error on the Cluster Object and returns false.

=item Examples

   Ex 1: # Join the specified node
   if (&San::Task::ClusterHelper::join_node(-node        => $new_node,
                                            -cluster_obj => $Clus_Obj)) {
       $Test->nlog(" - Successfully joined " . $new_node->get_name()
                   . " to the cluster.");
   } else {
       $Test->nlog(" - Failed to join " . $new_node->get_name()
                   . " to the cluster" . $Clus_Obj->errmsg());
       return $SanUtil::FALSE;
   }

   Ex 2: # Unjoin the specified node
   &San::Task::ClusterHelper::join_node(-node        => $new_node,
                                        -cluster_obj => $Clus_Obj);
   if ($Clus_Obj->error()) {
       $Test->nlog(" - Successfully joined " . $new_node->get_name()
                   . " to the cluster.");
   } else {
       $Test->nlog(" - Failed to join " . $new_node->get_name()
                   . " to the cluster" . $Clus_Obj->errmsg());
       return $SanUtil::FALSE;
   }

=back

=head2 B<unjoin_node(%args)>

Task to unjoin a node from a cluster.

NOTE: This function will clear errors on the $Clus_Obj and set failures on
      error.

=over 4

=item Inputs

=item Required

=item * I<-cluster_obj>

The San::ControllerCluster object to unjoin the node from.

=item * I<-node>

The name of the node to unjoin or San::ControllerCluster::Filer10M object

=item Outputs

Returns true. if error, sets error on the Cluster Object and returns false.

=item Examples

    Ex: 1 # Unjoin the node specified.
    if (&San::Task::ClusterHelper::unjoin_node(-cluster_obj => $Clus_Obj,
                                               -node        => $unjoin_node)){
        $Test->nlog("Completed unjoin process of $unjoin_node");
    } else {
        $Test->nlog("Error: completing unjoin process of $unjoin_node"
                    . $Clus_Obj->errmsg());
        return $SanUtil::FALSE;
    }

    Ex: 2    # Unjoin the node specified.
    &San::Task::ClusterHelper::unjoin_node(-cluster_obj => $Clus_Obj,
                                           -node        => $unjoin_node);
    if ($Clus_Obj->error()){
        $Test->nlog("Completed unjoin process of $unjoin_node");
    } else {
        $Test->nlog("Error: completing unjoin process of $unjoin_node"
                    . $Clus_Obj->errmsg());
        return $SanUtil::FALSE;
    }
    $Test->nlog("Waiting for $unjoin_node to wipeconfig and reboot.");
    # Detect the reboot of the node after the unjoin command.
    &San::Task::ClusterHelper::detect_unjoin_wipe_config_and_reboot(
        -cluster_obj => $Clus_Obj,
        -node        => $unjoin_node
    );
    if ($Clus_Obj->error()) {
        $Test->nlog("failed to detect_unjoin_wipe_config_and_reboot. "
                    . $Clus_Obj->errmsg());
        return $SanUtil::FALSE;
    }

=back

=head2 B<detect_unjoin_wipe_config_and_reboot(%args)>

Task to detect when the unjoined node from a cluster wipeconfigs and reboots
    to the boot menu.

NOTE: This function will clear errors on the $Clus_Obj and set failures on
      error.

=over 4

=item Inputs

=item Required

=item * I<-cluster_obj>

The San::ControllerCluster object.

=item * I<-node>

The name of the node to unjoin or San::ControllerCluster::Filer10M object

=item Outputs

Returns true. if error, sets error on the Cluster Object and returns false.

=item Examples

    # Unjoin the node specified.
    &San::Task::ClusterHelper::unjoin_node(-cluster_obj => $Clus_Obj,
                                           -node        => $unjoin_node);
    if ($Clus_Obj->error()){
        $Test->nlog("Completed unjoin process of $unjoin_node");
    } else {
        $Test->nlog("Error: completing unjoin process of $unjoin_node"
                    . $Clus_Obj->errmsg());
        return $SanUtil::FALSE;
    }

    $Test->nlog("Waiting for $unjoin_node to wipeconfig and reboot.");
    # Detect the reboot of the node after the unjoin command.
    &San::Task::ClusterHelper::detect_unjoin_wipe_config_and_reboot(
        -cluster_obj => $Clus_Obj,
        -node        => $unjoin_node
    );
    if ($Clus_Obj->error()) {
        $Test->nlog("failed to detect_unjoin_wipe_config_and_reboot. "
                    . $Clus_Obj->errmsg());
        return $SanUtil::FALSE;
    }

=back

=head2 B<get_cserver_lif_obj(%args)>

Get get lif_obj for the vserver with the parent set to a Filer10M object
(rather than the cserver object itself) This allows the automation to still
migrate/modify the connection properties of the lif without disrupting the
connection.

=over 4

=item Inputs

=item * I<-cluster_obj>

The San::ControllerCluster object.

=item Outputs

Returns ControllerCluster::Filer10M:Vserver:Lif Object with parent set to a
        ControllerCluster::Filer10M object if successful or undef with an error
        set on the ControllerCluster Object if unsuccessful.

=item Examples

    # ex. 1:
    # Fetch the lif object which represents the cserver lif
    $lif_obj = &San::Task::ClusterHelper::get_cserver_lif_obj(
                                                    -cluster_obj => $clus_obj);
    if ($clus_obj->error()) {
        # Error is set, bubble up
        return $SanUtil::FALSE;
    }

=back

=head2 B<gx_boot(%args)>

Currently a blocking call which will subtest gx_boot.thpl.  It will take a list
of nodename and tasks.

=over 4

=item Inputs

=item * I<-node>

The name or comma seperated list of names of nodes to execute a gx_boot TASK or
    TASKS on.

=item * I<-tasks>

The name or comma seperated list of names of TASKS to execute from gx_boot

=item Outputs

Returns true, or Logs error and returns false otherwise.

=item Examples

    # ex. 1:
    unless (&San::Task::ClusterHelper::gx_boot(-node  => $node,
                                               -tasks => "fake_init")) {
        $Test->nlog("gx_boot:fake_init task failed.");
        return $SanUtil::FALSE;
    }

    # ex. 2:
    unless (&San::Task::ClusterHelper::gx_boot(-node  => "$FilerA,$FilerB",
                                               -tasks => "fake_init")) {
        $Test->nlog("gx_boot:fake_init task failed.");
        return $SanUtil::FALSE;
    }

=back

=head2 B<migrate_cserver(%args)>

Migrates the cserver lif from current node to specified node maintaining the
same home-port.

=over 4

=item Inputs

=item * I<-cluster_obj>

The San::ControllerCluster object.

=item * I<-to_node>

The name of the node or San::ControllerCluster::Filer10M object to migrate the
    cserver lif to.

=item Outputs

Returns true. if error, sets error on the Cluster Object and returns false.

=item Examples

    # ex. 1:
    &San::Task::ClusterHelper::migrate_cserver(
        -cluster_obj => $clus_obj,
        -to_node     => $node_obj
    );
    if ($clus_obj->error()) {
        $Test->nlog("migrage cserver failed: " . $clus_obj->errmsg());
        return $SanUtil::FALSE;
    }

=back

=cut
