# -*- perl -*-
#  $Id: $
#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
#
## @summary Helper methods for testing C-mode Copy Offload
## @author (schroll) dl-ssan-test@netapp.com
## @status Private
## @pod Not here yet.

package San::Task::CopyOffloadHelper;

use strict;
use Exporter;

use Data::Dumper;
use List::Util qw( shuffle );
use NACL::STask::Node;
use Params::Validate qw(:all);
use San::SanLog;
use San::SanUtil;
use Tharn;

#
# Exceptions
#
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

use vars qw( @EXPORT @ISA );

# Command Interface to execute controller queries
our $cmd_interface;

# Default parameters to use for each xcopy command
our %default_xcopy_params = (
    'num_segments'        => 1,
    'segment_size'        => 8192,                 # 4 MB is ESX default
    'segment_size_sample' => [ 8, 8192, 65535 ],
    'timeout'             => 60000,
    'aborts'              => 'random'
);

@ISA = qw( SanLog SanUtil );

###############################################################################
# P U B L I C   M E T H O D S
###############################################################################

###############################################################################
# Method:     new
# Objective:  Creates new CopyOffloadHelper object.
# Parameters: debug_level
#             cmd_interface - NACL::C::Node
# Returns:    A reference to the object.
###############################################################################

sub new {
    my $class = shift;
    my %args  = @_;

    my $self = {};

    bless $self, $class;

    my $debug_level = 2;

    # Validate parameters
    while ( my ( $key, $value ) = each(%args) ) {
        if ( $key =~ /debug_level/i ) {
            unless ( length($value) ) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Missing value for key '-debug_level'");
            }
        } elsif ( $key =~ /cmd_interface/i ) {
            unless ( length($value) ) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Missing value for key '-cmd_interface'");
            }
        }
    }

    $cmd_interface = $args{'cmd_interface'};
    $self->set_debug_level($debug_level);

    $self->sanlog( -msg => "Create CopyOffloadHelper Object with debug level "
            . $debug_level );

    return $self;
}

###############################################################################
# Method:     create_copy_offload_requests
#
# Objective:  Creates an array of spt commands to perform copy offloads.
#             The slices/threads for each command is adjusted to achieve
#             the target_concurrent_ops_count supplied as a parameter.
#             queue_depth and the max_commands_per_session are taken into
#             account when adjusting the slices.
#
#             Note, there are numerous host and protocol variables that go into
#             figuring out how many concurrent SCSI commands are in effect at
#             any given time.  These are difficult to
#             account for, especially across different adapters.  This method
#             does a best effort, but
#             there will be situations where constant concurrency is not
#             obtainable.  For instance, if a takeover happens, and commands
#             are processed over a different path than when the method was
#             first called, the concurrency will be reduced.
#
#             This is only supported on linux.
#             Only one session per lif is supported
#
# Parameters: device_matchings
#             target_concurrent_ops_count
#             runtime
#             xcopy_params {
#               segment_size =>  size in lbas (int),
#                               'random',
#                               'random_sample,
#                                  randomly picks seg size
#                                  from segment_size_sample array
#                               'round_robin_sample'
#                                  alternates block sizes
#                                  from segment_size_sample array
#
#               segment_size_sample => [ 8, 8192 ],
#               num_segments => 'random' or int
#               timeout      => timeout in milliseconds for each cmd
#               aborts       => 1 for aborts, 0 for no aborts, 'random', 'round_robin'
#             }
#             queue_depth
#             max_commands_per_session
#
# Returns:    Array reference where each index holds a different spt command
#             to send xcopies from src lun(s) to a destination lun.
#
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################
sub create_copy_offload_requests {

    my $self = shift;
    my %args = @_;

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) } );
    %args = validate(
        @_,
        {   'device_matchings' => {
                'type'     => ARRAYREF,
                'required' => 1,
            },
            'target_concurrent_ops_count' => {
                'type'     => SCALAR,
                'required' => 1
            },
            'runtime' => {
                'type'     => SCALAR,
                'required' => 1
            },
            'xcopy_params' => {
                'type'     => HASHREF | UNDEF,
                'optional' => 1,
            },
            'path_type' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'default'  => 'MP'
            },
            'queue_depth' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'default'  => 32
            },
            'max_commands_per_session' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'default'  => 128
            }
        }
    );

    my %cmds_per_session;
    my %cmds_per_path;
    my @line_buffer;

    # Use default values for whichever xcopy parameters are not defined.
    my $xcopy_params = {};
    $xcopy_params = $args{'xcopy_params'} if defined $args{'xcopy_params'};
    $self->_populate_xcopy_params($xcopy_params);

    # Assign slices to each device matching pair in an effort to achieve
    # the target_ops at the time this subroutine is called.
    my $i = 0;

    while ( $i < $args{'target_concurrent_ops_count'} ) {

        # Not done slicing until next loop finishes without assigning
        # a slice to a dst device.
        my $is_slices_done = 1;

        foreach my $device_matching ( @{ $args{'device_matchings'} } ) {

            my $dst_device = $device_matching->{'dst_device'};

            # Get eligible paths for dst device based on path type
            # For MP, this will only return one path.
            # For AO and AN, it could return multiple paths.
            my @paths = $self->_get_eligible_paths( $device_matching,
                $args{'path_type'} );

            foreach my $path (@paths) {

                my $path_name  = $path->path_name();
                my $lif_obj    = $path->lif_obj();
                my $session_id = $lif_obj->vserver() . ':' . $lif_obj->lif();

                # If the queue depth is already exceeded for
                # this path, try the next eligible path.
                if ( exists $cmds_per_path{$path_name}
                    && $cmds_per_path{$path_name} >= $args{'queue_depth'} ) {
                    $self->sanlog(
                        -msg => 'Exceeded queue depth for ' . $path_name );
                    next;
                }

                if ( $dst_device->protocol() eq 'iscsi' ) {

                    # If the commands outstanding for this session
                    # is exceeded, try the next eligible path.
                    if ( exists $cmds_per_session{$session_id}
                        && $cmds_per_session{$session_id}
                        >= $args{'max_commands_per_session'} ) {

                        $self->sanlog(
                            -msg => 'Exceeded max cmds per session for '
                                . $path_name );

                        next;
                    }
                }

                # There aren't any queue depth or session limitations for this
                # path, so assign a slice and path to this device matching.
                if ( exists $device_matching->{'slices'} ) {
                    $device_matching->{'slices'} += 1;
                } else {
                    $device_matching->{'slices'} = 1;
                }
                $device_matching->{'path'}
                    = ( $args{'path_type'} eq 'MP' )
                    ? $dst_device->mp_device_path()
                    : $path_name;

                # Update the tracking structures to reflect that the sessions
                # and the paths will have one more command sent over them
                if ( exists $cmds_per_session{$session_id} ) {
                    $cmds_per_session{$session_id} += 1;
                } else {
                    $cmds_per_session{$session_id} = 1;
                }

                if ( exists $cmds_per_path{$path_name} ) {
                    $cmds_per_path{$path_name} += 1;
                } else {
                    $cmds_per_path{$path_name} = 1;
                }

                $is_slices_done = 0;
                $i++;

                last;
            }

            # Break the loop early if the target is met
            if ( $i == $args{'target_concurrent_ops_count'} ) {
                last;
            }
        }
        last if ($is_slices_done);
    }

    # Assign xcopy commands to each device matching now that the paths are
    # known.
    foreach ( my $i = 0; $i < @{ $args{'device_matchings'} }; $i++ ) {

        my $device_matching = $args{'device_matchings'}->[$i];

        if ( !exists $device_matching->{'slices'} ) {
            $self->sanlog( -msg => 'Warning no slices were assigned to '
                    . $self->log_device_matchings( [$device_matching] ) );
            next;
        }

        my @src_devices = @{ $device_matching->{'src_devices'} };
        my $dst_device  = $device_matching->{'dst_device'};
        my $blocks;

        # Assign segment size based on value of segment_size param
        if ( $xcopy_params->{'segment_size'} eq 'random' ) {
            $blocks = ( int( rand(30) ) + 1 ) * 1024 * 1024 / 512;
        } elsif ( $xcopy_params->{'segment_size'} eq 'random_sample' ) {
            my @segment_size_sample
                = @{ $xcopy_params->{'segment_size_sample'} };
            $blocks = $segment_size_sample[ rand @segment_size_sample ];
        } elsif ( $xcopy_params->{'segment_size'} eq 'round_robin_sample' ) {
            my @segment_size_sample
                = @{ $xcopy_params->{'segment_size_sample'} };
            $blocks
                = $segment_size_sample[ $i % scalar(@segment_size_sample) ];
        } else {
            $blocks = $xcopy_params->{'segment_size'};
        }

        # Assign segments based on num_segments parameter
        my $segments
            = ( $xcopy_params->{'num_segments'} eq 'random' )
            ? int( rand(4) ) + 1
            : $xcopy_params->{'num_segments'};

        # Determine whether xcopies will be aborted based on aborts parameter
        my $is_aborted = $xcopy_params->{'aborts'};

        if ( $xcopy_params->{'aborts'} eq 'random' ) {
            $is_aborted = int( rand(2) );
        } elsif ( $xcopy_params->{'aborts'} eq 'round_robin' ) {
            $is_aborted = ( $i % 2 == 0 ) ? 0 : 1;
        }

        # timeout is in milliseconds, not seconds
        my $timeout = ($is_aborted) ? 10 : $xcopy_params->{'timeout'};

        my $line
            = "cdb=0x83 enable=async,recovery onerr=continue runtime="
            . $args{'runtime'}
            . " timeout=$timeout"
            . ' slices='
            . $device_matching->{'slices'}
            . " segments=$segments ";

        foreach my $src_device (@src_devices) {
            $line
                .= 'src='
                . $src_device->mp_device_path()
                . " blocks=$blocks ";
        }

        $line .= 'dst=' . $device_matching->{'path'} . " blocks=$blocks ";
        push( @line_buffer, $line );
    }

    return @line_buffer;
}

###############################################################################
# Method:       exec_simple_copy
#
# Objective:    spt's xcopy enhancements are in the following format
#               cdb=83 <general_params>
#                      src=<dev> <params_for_src_dev>
#                      src=<dev> <params_for_src_dev>
#                      src=<dev> <params_for_src_dev>
#                      dst=<dev> <params_for_dst_dev>
#
#               This subroutine wraps this functionality.  It accepts
#               all of the general parameters, but uses defaults for the
#               src and destination devices.
#
# Parameters:   spt_path - path to spt
#               pass_thru - scalar string to subst in for general_params
#               src_devices - names of src device paths
#               dst_device - name of dst device path
#               blocks - segment size in lba (named after spt parameter)
#
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################
sub exec_simple_copy {

    my $self = shift;
    my %args = @_;

    %args = &SanUtil::strip_dash(%args);

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) } );
    %args = validate(
        @_,
        {   'client_obj' => {
                'type'     => OBJECT,
                'required' => 1,
            },
            'spt_path' => {
                'type'     => SCALAR,
                'required' => 1
            },
            'pass_thru' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1
            },
            'src_devices' => {
                'type'     => ARRAYREF,
                'required' => 1
            },
            'dst_device' => {
                'type'     => SCALAR,
                'required' => 1
            },
            'blocks' => {
                'type'     => SCALAR,
                'optional' => 1
            }, 
            'timeout' => { 
                'type' => SCALAR, 
                'optional' => 1 
            }, 
        }
    );

    my $timeout = (defined($args{'timeout'})) ? $args{'timeout'} : 600;

    my $spt_path = $args{'spt_path'};
    my $cmd      = "$spt_path cdb=0x83 " . $args{'pass_thru'};

    foreach my $device ( @{ $args{'src_devices'} } ) {
        $cmd .= " src=$device";

        if ( defined $args{'blocks'} ) {
            $cmd .= " blocks=" . $args{'blocks'};
        }
    }

    $cmd .= " dst=" . $args{'dst_device'};
    $self->sanlog( -msg => $cmd );

    my $client_conn = $args{'client_obj'}->get_connection();
    my ( $status, @output ) = &ClientUtil::execute( $client_conn, $cmd, $timeout );

    $self->sanlog(
        -msg => "Client hostname command returned " . join( "\n", @output ),
        -msg_level => 1
    );

    if ( $status != 0 ) {
        NATE::BaseException->throw(
            "command ($cmd) returned a bad status ($status).");
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name:  write_same
#
# Objective:    Sends scsi write sames until a range of blocks has been hole
#               punched,  See spt doc for more info.
#
# Params:       starting - number of bytes to start the hole punch at
#               limit - number of bytes to hole punch
#
# Return:       $TCD::PASS
# Throws:       NATE::BaseException
################################################################################
sub write_same {

    my $self = shift;
    my %args = @_;
    %args = &SanUtil::strip_dash(%args);

    my $dev = $args{'dev'};

    my $num_lbas_per_ws_exe = 8192;
    my $length              = 512;

    my $spt_path = $args{'spt_path'};

    my $cmd
        = "$spt_path dsf=$dev cdb='93 08' dir=write length=$length blocks=$num_lbas_per_ws_exe threads=10";

    if ( defined $args{'limit'} ) {
        $cmd .= "limit=" . $args{'limit'};
    }

    if ( defined $args{'starting'} ) {
        $cmd .= "starting=" . $args{'starting'};
    }

    $self->sanlog( -msg => "$cmd" );

    my $client_conn = $args{'client_obj'}->get_connection();
    my ( $status, @output ) = &ClientUtil::execute( $client_conn, $cmd );

    $self->sanlog(
        -msg => "Client hostname command returned " . join( "\n", @output ),
        -msg_level => 1
    );

    if ( $status != 0 ) {
        NATE::BaseException->throw(
            "command ($cmd) returned a bad status ($status).");
    }

    return $TCD::PASS;
}

###############################################################################
# Method:       _get_eligible_paths
#
# Objective:    Gets all the MPDevPath objects based on a path type.
#               If path type is 'random', a randomzied list of path objects
#               will be returned.  If path type is 'MP', only one AO path will
#               be returned.  If path type is 'AO', or 'AN', a list of either
#               'AO' or 'AN' paths will be returned.
#
# Parameters:   device_matching
#               path_type
#
# Returns:      array or arrayref of MPDevPath
###############################################################################
sub _get_eligible_paths {

    my $self            = shift;
    my $device_matching = shift;
    my $path_type       = shift;

    # Get the current path state of the dst device
    my $dst_device = $device_matching->{'dst_device'};
    my $path_state = $dst_device->get_current_path_state();
    my @paths      = $path_state->get_path_objs();

    # Figure out the eligible paths
    # For MP, only the AO path is supported.
    # For AO or AN, any path is supported that matches the path type
    if ( $path_type eq 'MP' ) {

        # Only send over active optimized path
        foreach my $path_obj (@paths) {
            if ( $path_obj->alua_state() eq 'AO' ) {
                @paths = ($path_obj);
                last;
            }
        }
    } elsif ( $path_type eq 'AO' || $path_type eq 'AN' ) {
        @paths = grep { $_->alua_state() eq $path_type; } @paths;
    }

    @paths = shuffle(@paths);

    if ( !@paths ) {
        NATE::BaseException->throw(
            'No eligible paths to LUN.  MP configs require an AO path.');
    }

    return ( wantarray() ) ? @paths : \@paths;
}

###############################################################################
# Method:     get_intranode_intervol_matchings
#
# Objective:  Looks at a client's configuration and selects devices to be
#             used for intra-node, inter-volume, inter-lun copy-offload
#             traffic.  The algorithm is as follows.
#             1.  For each node, pick n volumes to be src volumes.  Only
#                 devices in these volumes are eligible to become src
#                 devices of an xcopy.
#             2.  The remainder of the volumes become dst volumes.
#             3.  devices in the src volumes are matched to devices in the
#                 dst volumes, based on the number of src devices to use
#                 per match, and whether or not the caller wants to match
#                 each set of src devices to exactly one dst device or to
#                 all the dst devices (reuse_dst_devices parameter).
#
# Parameters: client_obj - SanClient.pm
#
#             filter {
#                nodes => ['ssan-3270-47a', 'ssan-3270-48a'],
#                vservers => ['vs1','vs2']
#             }
#
#             num_src_dev_per_match - Number of src devices to match to
#             a dst device
#
#             reuse_dst_devices - see parameter doc of _match_devices
#
# Returns:    ARRAYREF of device matching HASHREFs
#
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################
sub get_intranode_intervol_matchings {

    my $self = shift;
    my %args = @_;
    %args = &SanUtil::strip_dash(%args);

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) } );
    %args = validate(
        @_,
        {   'client_obj' => {
                'type'     => OBJECT,
                'required' => 1,
            },
            'filter' => {
                'type'     => HASHREF | UNDEF,
                'optional' => 1
            },
            'num_src_dev_per_match' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'default'  => 1
            },
            'reuse_dst_devices' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'default'  => 0
            }
        }
    );

    my @device_matchings;

    my $mpdevmgr = $args{'client_obj'}->mpdevmgr();

    if ( $args{'client_obj'}->error() ) {
        NATE::BaseException->throw( $args{'client_obj'}->errmsg() );
    }

    # Create filters for nodes and vservers
    my $node_filter = '*';
    my $vs_filter   = '*';

    if (   exists $args{'filter'}
        && exists $args{'filter'}->{'nodes'} ) {
        $node_filter = join( ',', @{$args{'filter'}->{'nodes'}} );
    }

    my @nodes = NACL::C::Node->find(
        command_interface => $cmd_interface,
        filter            => { node => $node_filter }
    );

    if (   exists $args{'filter'}
        && exists $args{'filter'}->{'vservers'} ) {
        $vs_filter = join( ',', @{ $args{'filter'}->{'vservers'} } );
    }

    foreach my $node (@nodes) {

        # Get the Volume Component States for the node
        my @vol_states = NACL::CS::Volume->fetch(
            command_interface => $cmd_interface,
            filter      => { node => $node->node(), 'vserver' => $vs_filter },
            allow_empty => 1
        );

        # Filter the volumes which have visible luns on the client
        my @filtered_vol_states = $self->_filter_volumes_by_client(
            'mpdevmgr'   => $mpdevmgr,
            'vol_states' => \@vol_states
        );

        # There must be atleast 1 destination volume
        if ( scalar(@filtered_vol_states) < 2 ) {
            $self->sanlog( -msg => 'Not enough volumes to send xcopies on '
                    . $node->node() );
            next;
        }

        # Pick the volumes that will be used to choose the src and dst devs
        my @src_volume_states = splice @filtered_vol_states, 0,
            int( @filtered_vol_states / 2 );
        my @dst_volume_states = @filtered_vol_states;

        # Get the devices eligible to be used as src devices
        my @src_devices = $self->_get_devices_by_volumes(
            'mpdevmgr'      => $mpdevmgr,
            'volume_states' => \@src_volume_states
        );

        # Get the devices eligible to be used as dst devices
        my @dst_devices = $self->_get_devices_by_volumes(
            'mpdevmgr'      => $mpdevmgr,
            'volume_states' => \@dst_volume_states
        );

        $self->_log_devices( \@src_devices );
        $self->_log_devices( \@dst_devices );

        # Match src devices to dst devices
        my $node_device_matchings = $self->_match_devices(
            'num_src_dev_per_match' => $args{'num_src_dev_per_match'},
            'reuse_dst_devices'     => $args{'reuse_dst_devices'},
            'src_devices'           => \@src_devices,
            'dst_devices'           => \@dst_devices,
        );

        @device_matchings = ( @device_matchings, @{$node_device_matchings} );
    }

    return ( wantarray() ) ? @device_matchings : \@device_matchings;
}

###############################################################################
# Method:     get_intralun_matchings
#
# Objective:  Looks at a client's configuration and selects devices to be
#             used for iintralun copy-offload traffic.
#             traffic.
#               1.  All LUNs visible by the client and not filtered by the
#                   vserver_filter parameter are selected for intralun traffic.
#
# Parameters: client_obj - SanClient.pm
#
#             filter {
#                nodes => ['ssan-3270-47a', 'ssan-3270-48a'],
#                vservers => ['vs1','vs2']
#             }
#
# Returns:    ARRAYREF of device matching HASHREFs
#
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################
sub get_intralun_matchings {

    my $self = shift;
    my %args = @_;
    %args = &SanUtil::strip_dash(%args);

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) } );
    %args = validate(
        @_,
        {   'client_obj' => {
                'type'     => OBJECT,
                'required' => 1,
            },
            'filter' => {
                'type'     => HASHREF | UNDEF,
                'optional' => 1
            }
        }
    );

    my @device_matchings;

    my $mpdevmgr = $args{'client_obj'}->mpdevmgr();

    if ( $args{'client_obj'}->error() ) {
        NATE::BaseException->throw( $args{'client_obj'}->errmsg() );
    }

    # Translate filter parameter to what get_filtered_devices is expecting
    my %params;

    $params{-ontap_home_nodes} = [ split( ',', $args{'filter'}->{'nodes'} ) ]
        if defined $args{'filter'}
            && defined $args{'filter'}->{'nodes'};

    if ( defined $args{'filter'} && defined $args{'filter'}->{'vservers'} ) {
        $params{-vservers} = $args{'filter'}->{'vservers'};
    }

    # Get the devices being targeted for the src and dst of xcopy
    my @devices = $mpdevmgr->get_filtered_devices(%params);

    if ( $mpdevmgr->error() ) {
        my $err_msg = $mpdevmgr->errmsg();
        $mpdevmgr->_clear_err();
        NATE::BaseException->throw($err_msg);
    }

    foreach my $device (@devices) {

        my $device_matching = {
            'src_devices' => [$device],
            'dst_device'  => $device
        };

        push( @device_matchings, $device_matching );
    }

    return ( wantarray() ) ? @device_matchings : \@device_matchings;
}

###############################################################################
# Method:     get_interlun_matchings
#
# Objective:  Looks at a client's configuration and selects devices to be
#             used for interlun copy-offload traffic. Algorithm:
#             1.  Loop through each volume that the client has non-filtered
#                 devices on.  Get an array of devices on taht volume.
#             2.  The first half of the device array holds the eligible
#                 src devices.  The second half holes the eligible dst devices.
#             3.  devices in the src volumes are matched to devices in the
#                 dst volumes, based on the number of src devices to use
#                 per match, and whether or not the caller wants to match
#                 each set of src devices to exactly one dst device or to
#                 all the dst devices (reuse_dst_devices parameter).
#
# Parameters: client_obj - SanClient.pm
#
#             filter {
#                nodes => ['ssan-3270-47a', 'ssan-3270-48a'],
#                vservers => ['vs1','vs2']
#             }
#
# Returns:    ARRAYREF of device matching HASHREFs
#
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################
sub get_interlun_matchings {

    my $self = shift;
    my %args = @_;
    %args = &SanUtil::strip_dash(%args);

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) } );
    %args = validate(
        @_,
        {   'client_obj' => {
                'type'     => OBJECT,
                'required' => 1,
            },
            'filter' => {
                'type'     => HASHREF | UNDEF,
                'optional' => 1
            },
            'num_src_dev_per_match' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'default'  => 1
            },
            'reuse_dst_devices' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'default'  => 0
            }
        }
    );

    my @device_matchings;

    my $mpdevmgr = $args{'client_obj'}->mpdevmgr();

    if ( $args{'client_obj'}->error() ) {
        NATE::BaseException->throw( $args{'client_obj'}->errmsg() );
    }

    # Get the Volume Component States
    my @vol_states = NACL::CS::Volume->fetch(
        command_interface => $cmd_interface,
        allow_empty       => 1
    );

    # Translate filter parameters to what get_filtered_devices is expecting
    my %params;

    $params{-ontap_home_nodes} = [ split( ',', $args{'filter'}->{'nodes'} ) ]
        if defined $args{'filter'}
            && defined $args{'filter'}->{'nodes'};

    if ( defined $args{'filter'} && defined $args{'filter'}->{'vservers'} ) {
        $params{-vservers} = $args{'filter'}->{'vservers'};
    }

    # Filter out any devices the caller does not want to target
    my @all_devices = $mpdevmgr->get_filtered_devices(%params);

    foreach my $vol_state (@vol_states) {

        # Find only the devices that are on this volume
        my $volume = $vol_state->volume();

        my @devices = grep {
            my $lun_path = $_->ontap_lun_path();
            if ( not defined($lun_path) ) {
                0;
            } else {
                $lun_path =~ qr/\/vol\/$volume/;
            }
        } @all_devices;

       # If there weren't any devices for this volume or there weren't atleast
       # 2, continue to the next volume
        if ( scalar(@devices) < 2 ) {
            next;
        }

        my @src_devices = @devices;
        my @dst_devices = splice @src_devices, int( scalar(@devices) / 2 );

        # Match src devices to dst devices
        my @vol_device_matchings = $self->_match_devices(
            'num_src_dev_per_match' => $args{'num_src_dev_per_match'},
            'reuse_dst_devices'     => $args{'reuse_dst_devices'},
            'src_devices'           => \@src_devices,
            'dst_devices'           => \@dst_devices,
        );
        @device_matchings = ( @device_matchings, @vol_device_matchings );
    }

    return ( wantarray() ) ? @device_matchings : \@device_matchings;
}

###############################################################################
# Method:     get_internode_matchings
#
# Objective:  Looks at a client's configuration and selects devices to be
#             used for inter-node traffic.  The algorithm is as follows.
#             1.  Look at cross_ha_pair parameter to determine whether the
#                 copies will be across ha pairs or between them.
#             2.  If the value is true, nodes 0, 2, 4, etc will all hold
#                 src devices and nodes 1, 3, 5, etc will all hold dst devices.
#             3.  If the value is false, nodes 1, 3, etc will all hold
#                 src devices and nodes 2, 4, etc will all hold dst devices.
#             4.  devices in the src volumes are matched to devices in the
#                 dst volumes, based on the number of src devices to use
#                 per match, and whether or not the caller wants to match
#                 each set of src devices to exactly one dst device or to
#                 all the dst devices (reuse_dst_devices parameter).
#
# Parameters: client_obj - SanClient.pm
#
#             filter {
#                nodes => ['ssan-3270-47a', 'ssan-3270-48a'],
#                vservers => ['vs1','vs2']
#             }
#             The nodes filter must contain an ordered list of ha pairs or
#             an exception is thrown.
#
#             cross_ha_pair - 1 if copies should be generated between
#             the nodes of an ha pair, 0 if copies should be generated across
#             the nodes of an ha pair.
#
#             num_src_dev_per_match - Number of src devices to match to
#             a dst device
#
#             reuse_dst_devices - see parameter doc of _match_devices
#
# Returns:    ARRAYREF of device matching HASHREFs
#
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################
sub get_internode_matchings {

    my $self = shift;
    my %args = @_;
    %args = &SanUtil::strip_dash(%args);

    # Validate Params
    Params::Validate::validation_options(
        on_fail => sub { NATE::Exceptions::Argument->throw(shift) } );
    %args = validate(
        @_,
        {   'client_obj' => {
                'type'     => OBJECT,
                'required' => 1,
            },
            'filter' => {
                'type'     => HASHREF | UNDEF,
                'optional' => 1
            },
            'cross_ha_pair' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'default'  => 1
            },
            'num_src_dev_per_match' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'default'  => 1
            },
            'reuse_dst_devices' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1,
                'reuse'    => 0
            }
        }
    );

    my @device_matchings;

    my $mpdevmgr = $args{'client_obj'}->mpdevmgr();

    if ( $args{'client_obj'}->error() ) {
        NATE::BaseException->throw( $args{'client_obj'}->errmsg() );
    }

    # Create filters for nodes and vservers
    my $node_filter = '*';
    my $vs_filter   = '*';

    if (   exists $args{'filter'}
        && exists $args{'filter'}->{'nodes'} ) {
        $node_filter = join( ',', @{$args{'filter'}->{'nodes'}} );
    }

    # Verify nodes returned by the node filter are an ordered list of ha pairs
    my @nodes = NACL::STask::Node->find(
        command_interface => $cmd_interface,
        filter            => { node => $node_filter }
    );

    if ( @nodes % 2 != 0 ) {
        NATE::Exceptions::Argument->throw(
            'Node filter must be an ordered list of ha pairs');
    }

    for ( my $i = 0; $i++; $i = $i + 2 ) {
        if ( $nodes[$i]->get_partner_obj()->node() ne $nodes[$i]->node() ) {
            NATE::Exceptions::Argument->throw(
                      'Node filter must be an ordered list of ha pairs '
                    . $nodes[$i]->get_partner_obj()->node() . ':'
                    . $nodes[$i]->node() );
        }
    }

    if (   exists $args{'filter'}
        && exists $args{'filter'}->{'vservers'} ) {
        $vs_filter = join( ',', @{ $args{'filter'}->{'vservers'} } );
    }

    my $starting_node_index = $args{'cross_ha_pair'} ? 0 : 1;

    my $num_nodes = scalar(@nodes);

    for ( my $i = $starting_node_index; $i < $num_nodes; $i = $i + 2 ) {

        my $src_node = $nodes[$i];
        my $dst_node = $nodes[ ( $i + 1 ) % $num_nodes ];

        # Get the src Volume Component States for the node
        my @src_vol_states = NACL::CS::Volume->fetch(
            command_interface => $cmd_interface,
            filter            => {
                node    => $src_node->node(),
                vserver => $vs_filter
            },
            allow_empty => 1
        );

        # Filter the src volumes which have visible luns on the client
        my @src_filtered_vol_states = $self->_filter_volumes_by_client(
            'mpdevmgr'   => $mpdevmgr,
            'vol_states' => \@src_vol_states
        );

        # Get the dst Volume Component States for the node
        my @dst_vol_states = NACL::CS::Volume->fetch(
            command_interface => $cmd_interface,
            filter            => {
                node    => $dst_node->node(),
                vserver => $vs_filter
            },
            allow_empty => 1
        );

        # Filter the dst volumes which have visible luns on the client
        my @dst_filtered_vol_states = $self->_filter_volumes_by_client(
            'mpdevmgr'   => $mpdevmgr,
            'vol_states' => \@dst_vol_states
        );

        # Get the devices eligible to be used as src devices
        my @src_devices = $self->_get_devices_by_volumes(
            'mpdevmgr'      => $mpdevmgr,
            'volume_states' => \@src_filtered_vol_states
        );

        # Get the devices eligible to be used as dst devices
        my @dst_devices = $self->_get_devices_by_volumes(
            'mpdevmgr'      => $mpdevmgr,
            'volume_states' => \@dst_filtered_vol_states
        );

        $self->_log_devices( \@src_devices );
        $self->_log_devices( \@dst_devices );

        # Match src devices to dst devices
        my $node_device_matchings = $self->_match_devices(
            'num_src_dev_per_match' => $args{'num_src_dev_per_match'},
            'reuse_dst_devices'     => $args{'reuse_dst_devices'},
            'src_devices'           => \@src_devices,
            'dst_devices'           => \@dst_devices,
        );

        @device_matchings = ( @device_matchings, @{$node_device_matchings} );
    }

    return ( wantarray() ) ? @device_matchings : \@device_matchings;
}

###############################################################################
# Method:       log_device_matchings
#
# Objective:    logs devices matchings in a readable format
#
# Parameters:   array ref of devices
#
# Exceptions:   None
###############################################################################
sub log_device_matchings {

    my $self             = shift;
    my $device_matchings = shift;

    NATE::BaseException->throw('Devices are not defined')
        unless defined $device_matchings;

    $self->sanlog( -msg => 'Logging device matchings' );

    foreach my $device_matching ( @{$device_matchings} ) {

        my @logging_src_devices = @{ $device_matching->{'src_devices'} };
        foreach my $src_device (@logging_src_devices) {

            $self->sanlog( -msg => 'Src Device:'
                    . $src_device->ontap_vserver() . ':'
                    . $src_device->ontap_lun_path() );

        }

        $self->sanlog( -msg => 'Dst Device:'
                . $device_matching->{'dst_device'}->ontap_vserver() . ':'
                . $device_matching->{'dst_device'}->ontap_lun_path() );
    }

}

################################################################################
# Method Name:  set_wafl_error_fiji
# Objective:    Sets fijis on the nodes designated by the nodes param
# Return:       $SanUtil::TRUE
# Throws:       throws NATE::BaseException
################################################################################
sub set_wafl_error_fiji {

    my $self = shift;
    my %args = @_;
    %args = &SanUtil::strip_dash(%args);

    my @nodes = @{ $args{'nodes'} };

    my $cmd
        = '"priv set test; fiji filter ' . "'("
        . $args{'fiji'}
        . ' (&& (== @ce '
        . $args{'copy_engine'}
        . ')  (>= @offset '
        . $args{'offset'}
        . ')) (return '
        . $args{'wafl_error'} . "))'\"";

    $self->sanlog( -msg => 'setting wafl error fiji:  ' . $cmd );

    foreach my $node (@nodes) {
        $node->run( 'command' => $cmd );
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name:  set_ocm_debug_level
# Objective:    Sets debug level on the nodes designated by the nodes param
# Params:       nodes - nodes to set debug level on
#               value - debug level
# Return:       $SanUtil::TRUE
# Throws:       throws NATE::BaseException
################################################################################
sub set_ocm_debug_level {

    my $self = shift;
    my %args = @_;
    %args = &SanUtil::strip_dash(%args);

    my @nodes = @{ $args{'nodes'} };

    my $cmd = '"priv set test; dbg level ocm=' . $args{'value'} . '"';

    foreach my $node (@nodes) {
        $node->run( 'command' => $cmd );
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:       _filter_volumes_by_client
#
# Objective:    Filters a list of Volume Component States to include only
#               the ones that have devices visible to a client.
#
# Parameters:   mpdevmgr   - San::SanClient::MPDevMgr
#               vol_states - array reference of NACL::CS:Volume
#
# Returns:      array or arrayref of NACL::CS:Volume
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################
sub _filter_volumes_by_client {

    my $self = shift;
    my %args = @_;

    my $mpdevmgr   = $args{'mpdevmgr'};
    my @vol_states = @{ $args{'vol_states'} };

    # Filter the volumes based on which ones are seen by client
    my @filtered_vol_states;

    foreach my $vol_state (@vol_states) {

        my $volume_name = $vol_state->volume();

        $self->sanlog( -msg => 'Filtering volume ' . $volume_name );

        my @devices = $mpdevmgr->get_filtered_devices(
            -vservers             => [ $vol_state->vserver() ],
            -ontap_lun_path_regex => qr/\/vol\/$volume_name/
        );

        if ( $mpdevmgr->error() ) {
            my $err_msg = $mpdevmgr->errmsg();
            $mpdevmgr->_clear_err();
            $self->sanlog( -msg =>
                    'Warning:  mpdevmgr gives off false positives.  The error was'
                    . $err_msg );
        }

        if (@devices) {
            push( @filtered_vol_states, $vol_state );
        } else {
            $self->sanlog( -msg => 'No devices found for ' . $volume_name );
        }
    }

    return ( wantarray() ) ? @filtered_vol_states : \@filtered_vol_states;
}

###############################################################################
# Method:       _get_devices_by_volumes
#
# Objective:    Given an array of NACL::CS::Volume, this returns an array
#               of all the devices found in those volumes.
#
# Parameters:   mpdevmgr   - San::SanClient::MPDevMgr
#               vol_states - array reference of NACL::CS:Volume
#
# Returns:      Arrayref of MPDev.pm
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################
sub _get_devices_by_volumes {

    my $self = shift;
    my %args = @_;

    my $mpdevmgr = $args{'mpdevmgr'};
    my @devices;

    # Get the devices being targeted for the src of an xcopy
    foreach my $volume_state ( @{ $args{'volume_states'} } ) {
        my $volume_name  = $volume_state->volume();
        my @curr_devices = $mpdevmgr->get_filtered_devices(
            -vservers             => [ $volume_state->vserver() ],
            -ontap_lun_path_regex => qr/\/vol\/$volume_name/
        );

        if ( $mpdevmgr->error() ) {
            my $err_msg = $mpdevmgr->errmsg();
            $mpdevmgr->_clear_err();
            NATE::BaseException->throw($err_msg);
        }
        @devices = ( @devices, @curr_devices );
    }

    return ( wantarray() ) ? @devices : \@devices;
}

###############################################################################
# Method:       _match_devices
#
# Objective:    Given arrays of src & dst devices, this function matches n
#               devices from the src array to m devices from the dst array.
#
# Parameters:   num_src_dev_per_match - Number of src devices to match to
#               a dst device
#               src_devices - array ref of MpDev.pm
#               dst_devices - array ref of MpDev.pm
#               reuse_dst_devices -
#               If reuse_dst_devices is set to false, it only matches each
#               set of src devices to exactly one destination device.
#               If reuse_dst_devices is set to true, it matches each
#               set of source devices to all the dst devices.
#
#               Example, 1 src vol (with 3 LUs), 1 dst volume (with 2 LUs)
#
#               If reuse_dst_devices is false,
#               src_lun1 -> dst_lun1
#               src_lun2 -> dst_lun2
#
#               If reuse_dst_devices is true,
#               src_lun1 -> dst_lun1
#               src_lun1 -> dst_lun2
#               src_lun2 -> dst_lun1
#               src_lun2 -> dst_lun2
#               src_lun3 -> dst_lun1
#               src_lun3 -> dst_lun2

# Returns:      arrayref of hashrefs containing src_devices and dst_device
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################
sub _match_devices {

    my $self = shift;
    my %args = @_;

    my @src_devices = @{ $args{'src_devices'} };
    my @dst_devices = @{ $args{'dst_devices'} };
    my @device_matchings;

    for (
        my $i = 0;
        $i < @src_devices;
        $i = $i + $args{'num_src_dev_per_match'}
        ) {

        my @curr_src_devices
            = @src_devices[ $i .. ( $i + $args{'num_src_dev_per_match'} - 1 )
            ];

        if ( $args{'reuse_dst_devices'} ) {

            for ( my $j = 0; $j < @dst_devices; $j++ ) {

                push(
                    @device_matchings,
                    {   'src_devices' => \@curr_src_devices,
                        'dst_device'  => $dst_devices[0]
                    }
                );

            }

        } else {

            # Make sure there are dst devices that can still be matched
            # to src device(s)
            if ( scalar(@dst_devices) >= 1 ) {
                push(
                    @device_matchings,
                    {   'src_devices' => \@curr_src_devices,
                        'dst_device'  => shift(@dst_devices)
                    }
                );
            }
        }
    }

    return ( wantarray() ) ? @device_matchings : \@device_matchings;
}

###############################################################################
# Method:       _log_devices
#
# Objective:    logs an array of devices in a readable format
#
# Parameters:   array ref of devices
#
# Exceptions:   None
###############################################################################
sub _log_devices {

    my $self    = shift;
    my $devices = shift;

    NATE::BaseException->throw('Devices are not defined')
        unless defined $devices;

    $self->sanlog( -msg => 'Logging devices' );
    foreach my $device ( @{$devices} ) {
        $self->sanlog( -msg => $device->ontap_vserver() . ':'
                . $device->ontap_lun_path() );
    }
}

###############################################################################
# Method:       _populate_xcopy_params
#
# Objective:    Use defaults if parameter is not specified
#
# Parameters:   hashref of xcopy params
#
# Exceptions:   None
###############################################################################
sub _populate_xcopy_params {

    my $self         = shift;
    my $xcopy_params = shift;

    foreach my $key ( keys %default_xcopy_params ) {
        unless ( defined $xcopy_params->{$key} ) {
            $xcopy_params->{$key} = $default_xcopy_params{$key};
        }
    }

    return $xcopy_params;
}

1;
