# -*- perl -*-
#  $Id: $
#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
#
## @summary Helper methods for testing C-mode VFit functionality
## @author (anjalit) dl-ssan-test@netapp.com
## @status Private
## @pod Not here yet.

package San::Task::Esx_Vstorage_Helper;

use strict;
use Exporter;
use List::MoreUtils qw(any);
use Error;
use San::ControllerCluster;
use San::SanClient;
use San::SanMgr;
use San::SanLog;
use San::SanUtil;
use Data::Dumper;
use Tharn;
use Params::Validate qw(:all);

#
# Exceptions
#
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

use vars qw( @EXPORT @ISA );

use constant {ESSH_PATH               => "/x/eng/kc-users/samt/bin/essh",
              SCRIPT_PATH             => "/scratch/var/tmp/test.sh",
              VMDK_DEFAULT_SIZE_IN_KB => '4194304',
              VMFS_VMDKS_FOR_CREATE_WL_DEFAULT   => '10',
              MAX_PARALLEL_CONN                  => '8',
              DISK_FORMAT_DEFAULT                => 'eagerzeroedthick',
              USERNAME_DEFAULT                   => 'root',
              PASSWORD_DEFAULT                   => 'netapp1!',
              DEFAULT_CLONES_PER_ITER            => '10',
              SLEEP_BETWEEN_CLONE_ITERS          => '90',
              SLEEP_CONNECTION_RETRY             => '30',
              DEFAULT_ESX_TIMEOUT                => '60000',
              DEFAULT_ESX_CONNECTION_TYPE        => 'ssh-persist',
              DEFAULT_ESX_CONNECTION_RETRY       => '0',
              DEFAULT_ESX_CONNECTION_MAX_RETRIES => '3',
              DEFAULT_CONTROLLER                 => 'lsilogic',
              DEFAULT_DS_TYPE                    => 'one-to-one',
              DEFAULT_SRC_VMDK_PER_DS            => '1',
              DEFAULT_ONTAP_MODE                 => "cmode",
              DEFAULT_DD_SEEK                    => '4',
              DEFAULT_SRC_VMDK_CREATE_SUFFIX     => '600',};

@ISA = qw( Error SanLog SanUtil );

###############################################################################
# P U B L I C   M E T H O D S
###############################################################################

###############################################################################
# Name:     new
# Objective:  Creates new VFitHelper object.
# Parameters: None.
# Returns:    A reference to the object.
###############################################################################

sub new {
    my ($class, %args) = @_;
    my $self = {};

    bless $self, $class;

    my $debug_level = 2;

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /debug_level/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                 "ERROR: Missing value for key '-debug_level'");
            }
        }
    }

    $self->set_debug_level($debug_level);

    $self->sanlog(-msg => "Create Esx_Vstorage_Helper Object with debug level "
                  . $debug_level);

    return $self;
} ## end sub new
################################################################################
# Name     : esx_initialize
# Objective: Perform initialization on the ESX hosts
# Params   : CommandInterface, Esx host object, sanmgr object
# Returns  : Hash reference
#            Hash contains:
#                 Array of vservers,
#                 Array of Source devices,
#                 mpdevmgr object
# Throws   : NATE::ArgumentException on params error, NATE::BaseException
################################################################################
sub esx_initialize {

    my ($self, %args) = @_;
    my ($esx_host_obj, $CmdInterface, $sanmgr_obj);
    $self->sanlog(-msg => "Inside esx_initialize");
    my %esx_initialized_hash;
    my @array_esx_initialize;
    my @src_devices  = ();
    my @vserver_list = ();

    # param checks here..
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /ci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'ci'");
            }
            $CmdInterface = $value;
        } elsif ($key =~ /esx_host_obj/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                               "ERROR: Incorrect value for key 'esx_host_obj'");
            }
            $esx_host_obj = $value;

        } elsif ($key =~ /sanmgr_obj/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                 "ERROR: Incorrect value for key 'sanmgr_obj'");
            }
            $sanmgr_obj = $value;

        }
    } ## end while (my ($key, $value) ...

    $self->sanlog(-msg => "get list of vservers that this client will talk to");

    # get list of vservers that this client will talk to
    @vserver_list = $sanmgr_obj->get_vservers_associated_to_client(
                                     -client_name => $esx_host_obj->hostname());
    if ($self->error()) {
        NATE::BaseException->throw($sanmgr_obj->errmsg());
    }

    unless (scalar(@vserver_list)) {
        $self->sanlog(-msg => "We got empty vserver list on ESX object "
                      . $esx_host_obj->hostname());
        $esx_initialized_hash{'vserver_associated_to_client'} = ();
    }

    # Adding vservers to hash
    $esx_initialized_hash{'vserver_associated_to_client'} = \@vserver_list;

    my @cluster_objs =
      NACL::C::Cluster->find(command_interface => $CmdInterface);

    $self->sanlog(-msg => "Create MPDevMgr");

    # create an MPDevMgr (this will find all LUNs on the system)
    # this will get stored on the esx host object for later use
    my $mpdevmgr_obj =
      $esx_host_obj->create_mpdevmgr(-nacl_cluster_objs => \@cluster_objs,
                                     -vservers          => \@vserver_list);
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    # Adding mpdevmgr to hash
    $esx_initialized_hash{'mpdevmgr'} = $mpdevmgr_obj;

    $self->sanlog(-msg => "get filtered devices");

    # Obtain the device to be used as the datastore
    @src_devices =
      $mpdevmgr_obj->get_filtered_devices(
                                         -vservers => \@vserver_list,
                                         -ontap_lun_ostype_regex => qr/vmware/i,
                                         -ontap_mode => DEFAULT_ONTAP_MODE);

    if ($mpdevmgr_obj->error()) {
        NATE::BaseException->throw($mpdevmgr_obj->errmsg());
    }

    unless (scalar(@src_devices)) {
        $self->sanlog(-msg => "We got empty device list on ESX object "
                      . $esx_host_obj->hostname());
        $esx_initialized_hash{'src_devices'} = ();
    }

    # Adding source devices to hash
    $esx_initialized_hash{'src_devices'} = \@src_devices;

    return \%esx_initialized_hash;

} ## end sub esx_initialize

################################################################################
# Name: subtest_clone
# Objective:   Perform Clone ops by starting parallel connections to single ESX Object
#              Each parallel connection will have clone ops sent using essh
#              Clones will get deleted after they are created.
# Parameters:  ops_info - See get_ops_info
#              max_parallel_conn - max number of parallel subtests to fork off
#              username, password to connect to ESX host.
#              runlength - in seconds
#              DD params - dd_timeout, bs and count.
# Returns:     $SanUtil::TRUE if all subtests pass, $SanUtil::FALSE if subtests fail
# Throws:      NATE::Exceptions::Argument
################################################################################
sub subtest_clone {

    my ($self, %args) = @_;
    my $ops_info;
    my $runlength;
    my $max_parallel_conn;
    my $op_type;
    my $username;
    my $password;
    my $status;
    my $esx_timeout;
    my $dd_timeout;
    my $bs;
    my $count;

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /ops_info/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'ops_info'");
            }
            $ops_info = $value;
        } elsif ($key =~ /runlength/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                  "ERROR: Incorrect value for key 'runlength'");
            }
            $runlength = $value;
            $self->sanlog(-msg => "Runlength is " . $runlength);

        } elsif ($key =~ /max_parallel_conn/i) {
            unless (length($value) and ($value > 0)) {
                NATE::Exceptions::Argument->throw(
                          "ERROR: Incorrect value for key 'max_parallel_conn'");
            }
            $max_parallel_conn = $value;

        } elsif ($key =~ /op_type/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                    "ERROR: Incorrect value for key 'op_type'");
            }
            $op_type = $value;

        } elsif ($key =~ /username/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'username'");
            }
            $username = $value;

        } elsif ($key =~ /password/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'password'");
            }
            $password = $value;

        } elsif ($key =~ /esx_timeout/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                "ERROR: Incorrect value for key 'esx_timeout'");
            }
            $esx_timeout = $value;

        } elsif ($key =~ /dd_timeout/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                 "ERROR: Incorrect value for key 'dd_timeout'");
            }
            $dd_timeout = $value;

        } elsif ($key =~ /bs/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'bs'");
            }
            $bs = $value;

        } elsif ($key =~ /count/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                      "ERROR: Incorrect value for key 'count'");
            }
            $count = $value;

        }
    } ## end while (my ($key, $value) ...

    my @total_clone_ops = @$ops_info;
    my @subtest_hashes;
    my @clone_ops_to_subtest;
    my $num_of_clone_to_subtest =
      int(scalar(@total_clone_ops) / $max_parallel_conn);
    $self->sanlog(
             msg => "Per subtest number of ops is " . $num_of_clone_to_subtest);
    my $remaining_ops = scalar(@total_clone_ops) % $max_parallel_conn;
    for (my $i = 0 ; $i < $max_parallel_conn ; $i++) {
        my @clone_ops_to_subtest =
          @total_clone_ops[$i *
          $num_of_clone_to_subtest .. (
                 $i * $num_of_clone_to_subtest + $num_of_clone_to_subtest - 1)];
        if ($i == $max_parallel_conn - 1) {
            @clone_ops_to_subtest = @total_clone_ops[
              $i *
              $num_of_clone_to_subtest .. ($i * $num_of_clone_to_subtest +
                                           $num_of_clone_to_subtest +
                                           $remaining_ops -
                                           1)];
        }

        my $sub = sub {
            $status =
              $self->_clone_delete_vmdk('ops_info'   => \@clone_ops_to_subtest,
                                        'username'   => $username,
                                        'password'   => $password,
                                        'dd_timeout' => $dd_timeout,
                                        'bs'         => $bs,
                                        'count'      => $count,
                                        'runlength'  => $runlength,);
            unless ($status eq $SanUtil::TRUE) {
                &SubIPC::fail(-msg => "_clone_delete_vmdk failed: ");
            }
            SubIPC::pass(-msg => "Clone operation succeeded");
        };

        push(@subtest_hashes, {'subroutine' => $sub, 'runid' => 'clonessh'});

    } ## end for (my $i = 0 ; $i < $max_parallel_conn...

    return $self->_spawn_subtests('subtest_hashes' => \@subtest_hashes);

} ## end sub subtest_clone

################################################################################
# Name: subtest_create
# Objective:   Perform Create ops by starting parallel connections to single ESX Object
#              Each parallel connection will have create/delete ops sent to ESX
#              Each vmdk created thus, will get deleted after they are created.
# Parameters:  ops_info - See get_ops_info
#              max_parallel_conn - max number of parallel subtests to fork off
#              runlength - in seconds
# Returns:     $SanUtil::TRUE if all subtests pass, $SanUtil::FALSE if subtests fail
# Throws:      NATE::Exceptions::Argument
################################################################################
sub subtest_create {

    my ($self, %args) = @_;
    my $ops_info;
    my $runlength;
    my $max_parallel_conn;
    my $status;

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /ops_info/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'ops_info'");
            }
            $ops_info = $value;
        } elsif ($key =~ /runlength/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                  "ERROR: Incorrect value for key 'runlength'");
            }
            $runlength = $value;
            $self->sanlog(-msg => "Runlength is " . $runlength);

        } elsif ($key =~ /max_parallel_conn/i) {
            unless (length($value) and ($value > 0)) {
                NATE::Exceptions::Argument->throw(
                          "ERROR: Incorrect value for key 'max_parallel_conn'");
            }
            $max_parallel_conn = $value;

        }
    } ## end while (my ($key, $value) ...

    my @total_create_ops = @$ops_info;
    my @subtest_hashes;
    my @create_ops_to_subtest;
    my $num_of_create_to_subtest =
      int(scalar(@total_create_ops) / $max_parallel_conn);
    $self->sanlog(
            msg => "Per subtest number of ops is " . $num_of_create_to_subtest);
    my $remaining_ops = scalar(@total_create_ops) % $max_parallel_conn;
    for (my $i = 0 ; $i < $max_parallel_conn ; $i++) {
        my @create_ops_to_subtest =
          @total_create_ops[$i *
          $num_of_create_to_subtest .. (
               $i * $num_of_create_to_subtest + $num_of_create_to_subtest - 1)];
        if ($i == $max_parallel_conn - 1) {
            @create_ops_to_subtest = @total_create_ops[
              $i *
              $num_of_create_to_subtest .. ($i * $num_of_create_to_subtest +
                                            $num_of_create_to_subtest +
                                            $remaining_ops -
                                            1)];
        }

        my $sub = sub {
            $status =
              $self->_create_delete_vmdk('ops_info'  => \@create_ops_to_subtest,
                                         'runlength' => $runlength,);
            unless ($status eq $SanUtil::TRUE) {
                &SubIPC::fail(-msg => "_create_delete_vmdk failed: ");
            }
            SubIPC::pass(-msg => "_create_delete_vmdk succeeded");
        };

        push(@subtest_hashes,
             {'subroutine' => $sub, 'runid' => 'create_delete'});

    } ## end for (my $i = 0 ; $i < $max_parallel_conn...

    return $self->_spawn_subtests('subtest_hashes' => \@subtest_hashes);
} ## end sub subtest_create

################################################################################
# Name: subtest_clone_use_script
# Objective:   Perform Clone ops by starting parallel connections to single ESX Object
#              Each parallel connection will start clone ops using script on ESX host
#              Clones will get deleted after they are created in each thread.
# Parameters:  ops_info - See get_ops_info
#              max_parallel_conn - max number of parallel subtests to fork off
#              username, password to connect to ESX host.
#              runlength - in seconds
# Returns:     $SanUtil::TRUE if all subtests pass, $SanUtil::FALSE if subtests fail
# Throws:      NATE::Exceptions::Argument
################################################################################
sub subtest_clone_use_script {
    my ($self, %args) = @_;
    my $ops_info;
    my $runlength;
    my $max_parallel_conn;
    my $op_type;
    my $username;
    my $password;
    my $status;
    my $host;
    my $esx_timeout;

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /ops_info/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'ops_info'");
            }
            $ops_info = $value;
        } elsif ($key =~ /runlength/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                  "ERROR: Incorrect value for key 'runlength'");
            }
            $runlength = $value;
            $self->sanlog(-msg => "Runlength is " . $runlength);

        } elsif ($key =~ /max_parallel_conn/i) {
            unless (length($value) and ($value > 0)) {
                NATE::Exceptions::Argument->throw(
                          "ERROR: Incorrect value for key 'max_parallel_conn'");
            }
            $max_parallel_conn = $value;

        } elsif ($key =~ /username/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'username'");
            }
            $username = $value;

        } elsif ($key =~ /password/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'password'");
            }
            $password = $value;

        } elsif ($key =~ /host/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                       "ERROR: Incorrect value for key 'host'");
            }
            $host = $value;
        }
    } ## end while (my ($key, $value) ...

    my @total_clone_ops = @$ops_info;
    my @subtest_hashes;
    my @clone_ops_to_subtest;
    my $clones_per_thread = int(scalar(@total_clone_ops) / $max_parallel_conn);
    my $fail_count        = 0;

    #Create $max_parallel_conn number of threads to run the clone op
    for (my $i = 0 ; $i < $max_parallel_conn ; $i++) {
        my $clone_src_dst_list = "";
        for (my $k = $i * $clones_per_thread ;
             $k < (($i + 1) * $clones_per_thread) ;
             $k++) {
            if ($k < scalar(@total_clone_ops))
            { #if number of clone ops are not exactly divisible by num of threads
                my $src_vmdk = $total_clone_ops[$k]->{'source_vmdk_file'};
                my $dst_vmdk = $total_clone_ops[$k]->{'destination_vmdk_file'};
                $clone_src_dst_list =
                  $clone_src_dst_list . $src_vmdk . " " . $dst_vmdk . " ";
            }
        } ## end for (my $k = $i * $clones_per_thread...
        my $sub = sub {
            my $i          = 0;
            my $start_time = time();
            my $curr_time  = time();
            while ($curr_time - $start_time < $runlength) {
                $self->sanlog(-msg => "Starting batch - " . $i);
                my $clone_all =
                  `ESSH_PATH $host $username $password SCRIPT_PATH $i $clone_src_dst_list`;
                $self->sanlog(-msg => $clone_all);
                if ($clone_all =~ /Connection timed out/) {
                    &SubIPC::fail(
                                 -msg => "SSH connection time out, exiting...");
                    last;
                }
                $i = $i + 1;
                sleep(SLEEP_BETWEEN_CLONE_ITERS);
                $curr_time = time();
            } ## end while ($curr_time - $start_time...
        };

        my $subid = $i . "_clone_delete_using_script";
        push(@subtest_hashes, {'subroutine' => $sub, 'runid' => $subid});

    } ## end for (my $i = 0 ; $i < $max_parallel_conn...

    return $self->_spawn_subtests('subtest_hashes' => \@subtest_hashes);

} ## end sub subtest_clone_use_script

################################################################################
# Method Name: get_ops_info
# Objective:   Creates an array of hashes. Each hash has information that will be used
#              while executing operations on vmdks such as create, clone.
# Parameter:   esx_host_obj - esx host to send clone command on
#              source_vmdks - source vmdks to create cloning operations for
#              disk_format - disk formats used during op, by default it will use
#              all three types of disk formats
# Return:      array of hashes
#              Example of create op below
#              [  $create_op = {
#                  ' esx_host_obj ' => San::SanClient::Host::ESXHost
#                  ' source_datastore ' => '',
#                  ' vmfs_lvm_obj ' => San::SanClient::Lvm,
#                  ' disk_format ' => ' eagerzeroedthick ',
#                }
#              ]
# Throws       NATE::Exceptions::Argument
#              NATE::BaseException
################################################################################
sub get_ops_info {

    my ($self, %args) = @_;
    my $esx_host_obj;
    my $src_datastore;
    my $vmfs_lvm_obj;
    my $disk_format;
    my $op_type;
    my @ops_info;
    my $source_vmdks;
    my $dest_datastore;
    my $vmdk_size_kb = VMDK_DEFAULT_SIZE_IN_KB;
    my $VmdkCount    = 0;

    #    my @DiskFormats = qw(eagerzeroedthick zeroedthick thin);
    my @DiskFormats = qw(DISK_FORMAT_DEFAULT);

    $self->sanlog(-msg => "Start get_ops_info");

    # param checks here..
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /esx_host_obj/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' esx_host_obj
                          '");
            }
            $esx_host_obj = $value;
        } elsif ($key =~ /source_datastore/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' source_datastore
                          '");
            }
            $src_datastore = $value;

        } elsif ($key =~ /dest_datastore/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' dest_datastore
                          '");
            }
            $dest_datastore = $value;

        } elsif ($key =~ /source_vmdks/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' source_vmdks
                          '");
            }
            $source_vmdks = $value;

        } elsif ($key =~ /vmfs_lvm_obj/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' vmfs_lvm_obj
                          '");
            }
            $vmfs_lvm_obj = $value;

        } elsif ($key =~ /disk_format/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' disk_format
                          '");
            }
            $disk_format = $value;

        } elsif ($key =~ /op_type/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' op_type
                          '");
            }
            $op_type = $value;
        } elsif ($key =~ /vmdk_size_kb/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' vmdk_size_kb
                          '");
            }
            $vmdk_size_kb = $value;
        }
    } ## end while (my ($key, $value) ...

    ##TODO: Dump @src_datastore
    $self->sanlog(-msg => "Source datastore obtained in get_ops_info:");
    $self->sanlog(-msg => Dumper(@$src_datastore));

    #    # Create hash of all the information needed for executing operations
    if ($op_type eq 'create') {
        $self->sanlog(-msg => "get_ops_info for op_type create");
        foreach my $source_datastore (@$src_datastore) {
            ##TODO: Log $source_datasource
            $self->sanlog(-msg => "Iterating over Source datastore: "
                          . $source_datastore);
            if (defined($disk_format)) {
                $self->sanlog(
                             -msg => "Operations to be done on vmdks of type : "
                               . $disk_format);
                my $create_op = {'esx_host_obj' => $esx_host_obj,
                                 'vmfs_lvm_obj' => $vmfs_lvm_obj,
                                 'datastore'    => $source_datastore,
                                 'disk_format'  => $disk_format,
                                 'vmdk_size_kb' => $vmdk_size_kb};
                push(@ops_info, $create_op);

            } else {    # use all vmdk disk formats to create hash
                foreach my $format (@DiskFormats) {
                    ##TODO: Log $format
                    $self->sanlog(-msg => "Iterating over format: " . $format);
                    my $create_op = {'esx_host_obj' => $esx_host_obj,
                                     'vmfs_lvm_obj' => $vmfs_lvm_obj,
                                     'datastore'    => $source_datastore,
                                     'disk_format'  => $format,
                                     'vmdk_size_kb' => $vmdk_size_kb};
                    ##TODO: Log $create_op
                    $self->sanlog(-msg => Dumper($create_op));
                    push(@ops_info, $create_op);
                } ## end foreach my $format (@DiskFormats)
            } ## end else [ if (defined($disk_format...
        } ## end foreach my $source_datastore...
    } elsif ($op_type eq 'clone') {
        my $VmdkCount = 0;
        $self->sanlog(-msg => "Dest Datastore is :: ");
        $self->sanlog(Dumper(@$dest_datastore));

        foreach my $dest_datastore (@$dest_datastore) {
            foreach my $format (@DiskFormats) {
                foreach my $source_vmdk_file (@{$source_vmdks}) {

                    # for (my $i = 0 ; $i < 3; $i++){
                    my $clone_op = {'esx_host_obj'     => $esx_host_obj,
                                    'source_vmdk_file' => $source_vmdk_file,
                                    'destination_vmdk_file' =>
                                      "/vmfs/volumes/$dest_datastore/"
                                      . $VmdkCount++ . "_cl_"
                                      . $format,
                                    'disk_format' => $format};
                    push(@ops_info, $clone_op);

                    # }

                } ## end foreach my $source_vmdk_file...
            } ## end foreach my $format (@DiskFormats)
        } ## end foreach my $dest_datastore ...
    } ## end elsif ($op_type eq 'clone')

    $self->sanlog(-msg => "ops array we got from get_ops_info:");
    $self->sanlog(-msg => Dumper(@ops_info));

    $self->sanlog(-msg => "End get ops info");
    return (wantarray()) ? @ops_info : \@ops_info;
} ## end sub get_ops_info

################################################################################
#                               Private Methods
################################################################################

################################################################################
#Name:        _spawn_subtests
#Objective:   Helper function to spawn work in parallel across multiple
#               subtests.
#Params:      Arrayref of hashes containing anonymous subroutines and runids
#               to be used in SubIPC::create.
#
#               [{
#                   subroutine => sub {}
#                   runid      => ' id '
#                }]
# Returns:      $SanUtil::TRUE if there are no exceptions
# Throws:       NATE::BaseException
################################################################################
sub _spawn_subtests {

    my ($self, %args) = @_;
    my $subtest_hashes;
    my @subtests;
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /subtest_hashes/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' subtest_hashes
                          '");
            }
            $subtest_hashes = $value;
        }
    } ## end while (my ($key, $value) ...

    for (my $i = 0 ; $i < @{$subtest_hashes} ; $i++) {

        my $subtest =
          &SubIPC::create(-id  => $subtest_hashes->[$i]->{'runid'},
                          -esc => 1,
                          -run => $subtest_hashes->[$i]->{'subroutine'},
                          -ipc => 1,
                          -bg  => 1);
        if (&SubIPC::error()) {
            $self->sanlog(-msg => &SubIPC::get_error());
            NATE::BaseException->throw(&SubIPC::get_error());
        }
        push(@subtests, $subtest);
    } ## end for (my $i = 0 ; $i < @...

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->sanlog(-msg => &SubIPC::get_error());
        NATE::BaseException->throw(&SubIPC::get_error());
    }

    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != 0) {
        $self->sanlog(
                    -msg => "A subtest failure has been reported:\n\t\tFailed: "
                      . $st_status->{failed}
                      . "\n\t\tPassed: "
                      . $st_status->{passed});
        NATE::BaseException->throw(&SubIPC::get_error());
        &SubIPC::fail(-msg => "spawn subtest failed: " . "\n");

    } ## end if ($st_status->{failed...

    return $SanUtil::TRUE;
} ## end sub _spawn_subtests

################################################################################
# Method Name :   _create_delete_vmdk.
# Objective:   Execute create and delete on datastores from one esx host
# Parameter:   ops_info - Hash of create op - esx_obj, src and destination vmdk
# Return:      $SanUtil::TRUE on success
#              $SanUtil::FALSE on failure
# Exception:   NATE::Exceptions::Argument
################################################################################
sub _create_delete_vmdk {
    my ($self, %args) = @_;
    my $vmdk_size_kb =
      VMDK_DEFAULT_SIZE_IN_KB;  # default size - 4gb, incase user doesnt specify
    my $fail_count = 0;
    my $ops_info;
    my $runlength;

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /ops_info/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'ops_info'");
            }
            $ops_info = $value;
        } elsif ($key =~ /vmdk_size_kb/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                    "ERROR: Incorrect value for key ' op_info
                          '");
            }
            $vmdk_size_kb = $value;
        } elsif ($key =~ /runlength/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                  "ERROR: Incorrect value for key 'runlength'");
            }
            $runlength = $value;
        }
    } ## end while (my ($key, $value) ...

    my $start_time = time();
    my $curr_time  = time();
    $fail_count = 0;
    while ($curr_time - $start_time < $runlength) {

        foreach my $create_op (@$ops_info) {
            my $esx_host_obj     = $create_op->{'esx_host_obj'};
            my $vmfs_lvm_obj     = $create_op->{'vmfs_lvm_obj'};
            my $source_datastore = $create_op->{'datastore'};
            my $disk_format      = $create_op->{'disk_format'};

            $self->sanlog(Dumper($create_op));

            $self->sanlog(  -msg => "::_create_and_delete_vmdk(): Params:"
                          . "\n    disk format is "
                          . $create_op->{'disk_format'}
                          . "\n    ESX Datastore is  "
                          . $source_datastore);

            my $suffix = '_mysfx_' . $disk_format;

            # Create 10 vmdks per datastore in a single iteration
            my (@vmdks) =
              $vmfs_lvm_obj->create_vmdks(
                            -datastores     => [$source_datastore],
                            -controller     => DEFAULT_CONTROLLER,
                            -datastore_type => DEFAULT_DS_TYPE,
                            -num_vmfs_vmdks => VMFS_VMDKS_FOR_CREATE_WL_DEFAULT,
                            -vmfs_vmdk_size_kbytes => $vmdk_size_kb,
                            -vmfs_vmdk_disk_type   => $disk_format,
                            -suffix                => $suffix);
            if ($vmfs_lvm_obj->error()) {
                $self->sanlog(
                      -msg => ' VMDK create failed ' . $vmfs_lvm_obj->errmsg());
                $self->sanlog(
                             -msg => ' VMDK create failed for Source datastore '
                               . $source_datastore);
                $fail_count++;
            }

            # VMDK format returned by create_vmdks needs parsing
            # [datastore] <vmdk_path>
            foreach my $vmdk (@vmdks) {
                my @vmdk_info = split(' ', $vmdk);
                (my $datastore = $vmdk_info[0]) =~ s/[\]\[]//gi;
                my $vmdk_file = ' / vmfs / volumes /' . $datastore . '/
                          ' . $vmdk_info[1];
                $self->sanlog(-msg => "Created vmdk file : " . $vmdk_file);

                $self->sanlog(-msg => "Delete vmdk file : " . $vmdk_file);

                # delete the vmdk file created above
                $esx_host_obj->delete_vmdk(-vmdk_file => $vmdk_file);
                if ($esx_host_obj->error()) {
                    unless ($esx_host_obj->errmsg() =~ /was not found/) {
                        $self->sanlog(
                             -msg => ' VMDK delete failed for Source datastore '
                               . $source_datastore);
                        $fail_count++;
                        $self->sanlog(-msg => ' VMDK create failed '
                                      . $esx_host_obj->errmsg());

                    }
                } ## end if ($esx_host_obj->error...
            } ## end foreach my $vmdk (@vmdks)
        } ## end foreach my $create_op (@$ops_info)
        $curr_time = time();
        $self->sanlog(-msg => "Restarting logic of create/delete ops cycle");

    } ## end while ($curr_time - $start_time...

    if ($fail_count) {
        $self->sanlog(-msg => "Encountered create failures, subtest fails");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;

} ## end sub _create_delete_vmdk

################################################################################
# Method Name: create_source_vmdks
# Description: Helper function to create vmdks of type EZT.
#
# Returns:     array of created vmdk paths
# Throws       NATE::Argument on argument errors
################################################################################
sub create_source_vmdks {

    my ($self, %args) = @_;
    my $srcdatastores;
    my $vmfs_lvm_obj;
    my @vmdks;

    # default values for Optional params
    my $vmdk_sz_kbytes    = VMDK_DEFAULT_SIZE_IN_KB;    # default size is 4g
    my $disk_format       = DISK_FORMAT_DEFAULT;
    my $max_parallel_conn = MAX_PARALLEL_CONN;

    # array of all source vmdks created
    my @src_vmdks;

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /src_datastores/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                             "ERROR: Incorrect value for key 'src_datastores'");
            }
            $srcdatastores = $value;
        } elsif ($key =~ /vmfs_lvm_obj/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                               "ERROR: Incorrect value for key 'vmfs_lvm_obj'");
            }
            $vmfs_lvm_obj = $value;
        } elsif ($key =~ /vmdk_sz_kbytes/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                             "ERROR: Incorrect value for key 'vmdk_sz_kbytes'");
            }
            $vmdk_sz_kbytes = $value;
        } elsif ($key =~ /disk_format/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                "ERROR: Incorrect value for key 'disk_format'");
            }
            $disk_format = $value;
        } elsif ($key =~ /max_parallel_conn/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                          "ERROR: Incorrect value for key 'max_parallel_conn'");
            }
            $max_parallel_conn = $value;
        }
    } ## end while (my ($key, $value) ...
        # Create 9 disks on each data store
    my $suffix = DEFAULT_SRC_VMDK_CREATE_SUFFIX;
    $self->sanlog(
          -msg => "Source datastores before sending to create_vmdks method : ");
    foreach (@$srcdatastores) {
        $self->sanlog(-msg => $_);
    }

    my @src_dss = @$srcdatastores;
    my @subtest_hashes;
    my $num_of_creates_to_subtest =
      (scalar(@src_dss) > $max_parallel_conn)
      ? int(scalar(@src_dss) / $max_parallel_conn)
      : $max_parallel_conn;
    $self->sanlog(msg => "Per subtest number of creates is "
                  . $num_of_creates_to_subtest);
    my $remaining_ops = scalar(@src_dss) % $max_parallel_conn;

    #Spawn max_parallel_conn number of threads to create source vmdks
    for (my $i = 0 ; $i < $max_parallel_conn ; $i++) {

        my @create_to_subtest =
          @src_dss[$i *
          $num_of_creates_to_subtest .. (
               $i * $num_of_creates_to_subtest + $num_of_creates_to_subtest - 1)
          ];

        # if number is not an exact multiple of $max_parallel_conn, add remaining to last thread
        if ($i == $max_parallel_conn - 1) {
            @create_to_subtest = @src_dss[
              $i *
              $num_of_creates_to_subtest .. ($i * $num_of_creates_to_subtest +
                                             $num_of_creates_to_subtest +
                                             $remaining_ops -
                                             1)];
        }

        # Note:: Automation BURT open on checking if vmdk already exists before creating it
        my $sub = sub {
            my @vmdkfiles =
              $vmfs_lvm_obj->create_vmdks(
                                     -datastores     => \@create_to_subtest,
                                     -controller     => DEFAULT_CONTROLLER,
                                     -datastore_type => DEFAULT_DS_TYPE,
                                     -num_vmfs_vmdks => DEFAULT_SRC_VMDK_PER_DS,
                                     -vmfs_vmdk_size_kbytes => $vmdk_sz_kbytes,
                                     -vmfs_vmdk_disk_type   => $disk_format,
                                     -suffix                => $suffix);

            if ($vmfs_lvm_obj->error()) {
                &SubIPC::fail(-msg => " subtest_create_source_vmdk failed: "
                              . $vmfs_lvm_obj->errmsg());
            }
            push(@src_vmdks, @vmdkfiles);
            SubIPC::pass(-msg => "Creating source vmdk succeeded");
        };

        push(@subtest_hashes,
             {'subroutine' => $sub, 'runid' => 'create_source_vmdk'});

    } ## end for (my $i = 0 ; $i < $max_parallel_conn...

    $self->_spawn_subtests('subtest_hashes' => \@subtest_hashes);

    unless (scalar(@src_vmdks)) {
        $self->sanlog(-msg => "No source vmdks found");
        return (wantarray()) ? [] : ();
    }

    # VMDK format returned by create_vmdks is really odd
    # [datastore] <vmdk_path>
    foreach (@src_vmdks) {
        my @vmdk_info = split(' ', $_);
        (my $datastore = $vmdk_info[0]) =~ s/[\]\[]//gi;
        my $vmdk_file = '/vmfs/volumes/' . $datastore . '/' . $vmdk_info[1];
        push(@src_vmdks, $vmdk_file);
    }

    return (wantarray()) ? @src_vmdks : \@src_vmdks;
} ## end sub create_source_vmdks

################################################################################
# Method Name: _subtest_dd
# Objective:   Execute dd operations on an esx host in subtests over different
#              ssh connections.
# Parameter:   esx_host_obj - San::SanClient::Host::ESXHost
#              source_vmdks - vmdks to write to
#              bs - block size to use in random dd writes
#              count - number of random writes of bs to perform
# Return:      $SanUtil::TRUE if all subtests pass, $SanUtil::FALSE if subtests fail
################################################################################
sub subtest_dd {
    my ($self, %args) = @_;

    my @subtest_hashes;
    my $esx_host_obj;
    my $source_vmdks;
    my $bs;
    my $count;
    my $username;
    my $password;
    my $dd_timeout;

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /esx_host_obj/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                               "ERROR: Incorrect value for key 'esx_host_obj'");
            }
            $esx_host_obj = $value;
        } elsif ($key =~ /source_vmdks/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                               "ERROR: Incorrect value for key 'source_vmdks'");
            }
            $source_vmdks = $value;

        } elsif ($key =~ /bs/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'bs'");
            }
            $bs = $value;

        } elsif ($key =~ /count/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                      "ERROR: Incorrect value for key 'count'");
            }
            $count = $value;
        } elsif ($key =~ /username/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'username'");
            }
            $username = $value;
        } elsif ($key =~ /password/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'password'");
            }
            $password = $value;
        } elsif ($key =~ /dd_timeout/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                 "ERROR: Incorrect value for key 'dd_timeout'");
            }
            $dd_timeout = $value;
        }
    } ## end while (my ($key, $value) ...

    unless (scalar(@$source_vmdks)) {
        $self->sanlog(-msg => "No elements in source vmdks array");
    }
    foreach my $source_vmdk_file (@$source_vmdks) {
        my $sub = sub {
            $esx_host_obj->reconnect();
            my $esx_conn =
              $self->esx_connect('hostname' => $esx_host_obj->hostname(),
                                 'username' => $username,
                                 'password' => $password);
            my $status =
              $self->_dd('conn'       => $esx_conn,
                         'file'       => $source_vmdk_file,
                         'bs'         => $bs,
                         'count'      => $count,
                         'dd_timeout' => $dd_timeout);

            Tharn::disconnect($esx_conn, '-ok');

            if ($status > 0) {
                $self->sanlog(-msg => 'dd status greater than zero ' . $status);
                NATE::BaseException->throw("dd failed");
                &SubIPC::fail(-msg => "dd failed");
            }
            &SubIPC::pass(-msg => "dd succeeded");
        };
        push(@subtest_hashes, {'subroutine' => $sub, 'runid' => 'dd'});
    } ## end foreach my $source_vmdk_file...

    unless (scalar(@subtest_hashes)) {
        $self->sanlog(-msg => "No elements in subtest dd hash");
    }

    return $self->_spawn_subtests('subtest_hashes' => \@subtest_hashes);
} ## end sub subtest_dd

################################################################################
# Method Name: _dd
# Objective:   Execute dd on a client with bs and count parameters
# Parameter:   conn - tharn connection
#              bs - block size to use in random dd writes
#              count - number of random writes of bs to perform
#              conn_timeout - time in seconds to wait for the cmd to complete
# Return:      status of the dd command
################################################################################
sub _dd {
    my ($self, %args) = @_;
    my ($conn, $file, $bs, $count, $dd_timeout);

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /conn/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                       "ERROR: Incorrect value for key 'conn'");
            }
            $conn = $value;
        } elsif ($key =~ /file/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                       "ERROR: Incorrect value for key 'file'");
            }
            $file = $value;

        } elsif ($key =~ /bs/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'bs'");
            }
            $bs = $value;

        } elsif ($key =~ /count/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                      "ERROR: Incorrect value for key 'count'");
            }
            $count = $value;

        } elsif ($key =~ /dd_timeout/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                 "ERROR: Incorrect value for key 'dd_timeout'");
            }
            $dd_timeout = $value;

        }
    } ## end while (my ($key, $value) ...
    my $dd_cmd =
      "dd if=/dev/urandom of=$file bs=$bs count=$count conv=notrunc seek=DEFAULT_DD_SEEK";

    my ($status, @output) = &ClientUtil::execute($conn, $dd_cmd, $dd_timeout);

    return $status;
} ## end sub _dd
###############################################################################
# Method:    _esx_connect
# Objective: Create ssh connection to an esx host
# Params:    esx host to connect to
#            user - ssh username
#            password - ssh password
# Returns:   connection object - ssh connection to ESX host
###############################################################################
sub esx_connect {
    my ($self, %args) = @_;
    my $hostname;
    my $username    = USERNAME_DEFAULT;
    my $password    = PASSWORD_DEFAULT;
    my $timeout     = DEFAULT_ESX_TIMEOUT;
    my $conn_type   = DEFAULT_ESX_CONNECTION_TYPE;
    my $max_retries = DEFAULT_ESX_CONNECTION_MAX_RETRIES;
    my $connect;
    my $retries = DEFAULT_ESX_CONNECTION_RETRY;

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /hostname/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'hostname'");
            }
            $hostname = $value;
        } elsif ($key =~ /username/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'username'");
            }
            $username = $value;

        } elsif ($key =~ /password/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'password'");
            }
            $password = $value;

        }
    } ## end while (my ($key, $value) ...
    my $conn_str =
        $hostname
      . "#ssh_esx;"
      . "conntype=$conn_type;username=$username;password=$password;timeout="
      . $timeout
      . ";shell_type=sh;inherit=0;";

    while (!defined($connect) and $retries < $max_retries) {
        $connect = &Tharn::connect(
            $conn_str,
            -eval => sub {
                if (@_) {
                    $retries++;
                    $connect = undef;

                    # Display an error in case user doesn't include error
                    # handler on object instantiation.
                    $self->sanlog(
                                -msg => "WARNING: unable to connect to client '"
                                  . $hostname
                                  . "' Attempt $retries of $max_retries");
                    &Tharn::disconnect($conn_str, '-ok');
                    &SanUtil::nate_sleep(SLEEP_CONNECTION_RETRY);
                } ## end if (@_)
            });
    } ## end while (!defined($connect)...

    # if retry count has been exhausted, set error and return
    if ($retries == $max_retries) {
        return {};
    }

    # return the connection and the conn type
    return $connect;
} ## end sub esx_connect

################################################################################
# Method Name: _clone_delete_vmdk
# Objective:   Execute clone and delete for one esx host using ssh connections
#              DD is performed on the source vmdks before clone.
# Parameter:   ops_info - Hash of clone op - esx_obj, src and destination vmdk
#              username and password of ESX to connect to
#              dd params -
#              dd timeout
#              bs - block size to use in random dd writes
#              count - number of random writes of bs to perform
# Return:      $SanUtil::TRUE on success
#              $SanUtil::FALSE on failure
# Exception:   NATE::Exceptions::Argument
################################################################################
sub _clone_delete_vmdk {
    my ($self, %args) = @_;
    my $fail_count = 0;
    my $ops_info;
    my $runlength;

    # esx conn params --
    my $esx_host_obj;
    my $username = USERNAME_DEFAULT;
    my $password = PASSWORD_DEFAULT;

    # esx conn params --
    # DD params --
    my $dd_timeout;
    my $bs;
    my $count;

    # DD params --

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /ops_info/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'ops_info'");
            }
            $ops_info = $value;
        } elsif ($key =~ /username/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'username'");
            }
            $username = $value;

        } elsif ($key =~ /password/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "ERROR: Incorrect value for key 'password'");
            }
            $password = $value;

        } elsif ($key =~ /bs/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'bs'");
            }
            $bs = $value;

        } elsif ($key =~ /count/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                      "ERROR: Incorrect value for key 'count'");
            }
            $count = $value;
        } elsif ($key =~ /dd_timeout/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                 "ERROR: Incorrect value for key 'dd_timeout'");
            }
            $dd_timeout = $value;
        } elsif ($key =~ /runlength/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                  "ERROR: Incorrect value for key 'runlength'");
            }
            $runlength = $value;
        }
    } ## end while (my ($key, $value) ...

    $self->sanlog(-msg => "Clone workload for Runlength " . $runlength);
    my $start_time = time();
    my $curr_time  = time();

    while ($curr_time - $start_time < $runlength) {
        $self->sanlog("Starting subtest_clone_vmdk");

        foreach my $clone_op (@$ops_info) {
            my $esx_host_obj = $clone_op->{'esx_host_obj'};
            $self->sanlog(-msg => "HOSTNAME is " . $esx_host_obj->hostname());
            $clone_op->{'esx_conn'} =
              $self->esx_connect('hostname' => $esx_host_obj->hostname(),
                                 'username' => $username,
                                 'password' => $password);
            $self->sanlog(-msg => "Starting dd on source");
            $self->_dd('conn'       => $clone_op->{'esx_conn'},
                       'file'       => $clone_op->{'source_vmdk_file'},
                       'bs'         => $bs,
                       'count'      => $count,
                       'dd_timeout' => $dd_timeout);
            Tharn::disconnect($clone_op->{'esx_conn'}, '-ok');

            my $cmd_clone =
                "vmkfstools -i "
              . $clone_op->{'source_vmdk_file'} . " "
              . $clone_op->{'destination_vmdk_file'}
              . " -d eagerzeroedthick";
            my $cmd_delete =
              "vmkfstools -U " . $clone_op->{'destination_vmdk_file'};

            # Do clones/delete 10 times per src-dst pair
          INNER:
            for (my $j = 0 ; $j < DEFAULT_CLONES_PER_ITER ; $j++) {
                $self->sanlog(-msg => "Creating clone");
                my $clone =
                  `ESSH_PATH $esx_host_obj->hostname() $username $password $cmd_clone`;
                $self->sanlog(-msg => $clone);
                unless ($clone =~ /Clone: 100% done/) {
                    $self->sanlog(-msg => "Clone operation did not complete : "
                                  . "For source "
                                  . $clone_op->{'source_vmdk_file'}
                                  . " Dest "
                                  . $clone_op->{'destination_vmdk_file'}
                                  . "iteration "
                                  . $j);
                    last INNER;
                    $fail_count++;
                } ## end unless ($clone =~ /Clone: 100% done/)
                $self->sanlog(-msg => "Deleting clone");
                my $delete =
                  `ESSH_PATH $esx_host_obj->hostname() $username $password $cmd_delete`;
            } ## end for (my $j = 0 ; $j < DEFAULT_CLONES_PER_ITER...

        } ## end foreach my $clone_op (@$ops_info)
            # Increment time
        $curr_time = time();
        $self->sanlog(-msg => "Restarting logic of clone ops");
    } ## end while ($curr_time - $start_time...

    if ($fail_count) {
        $self->sanlog(-msg => "Encountered clone failures, subtest fails");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub _clone_delete_vmdk

1;

