# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/San/Task/IgroupHelper.pm#1 $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Utility for testing igroup functionality
## To provide library functions for the testing igroup.
## All the cmode igroup testing script will use this library
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::Task::IgroupHelper;

# compiler directives
use strict;
use Params::Validate qw( :all );
use San::Application::Igroupconstant;
use San::Application::Portsetconstant;
use San::SanUtil;
use Data::Dumper;

# inherit base class
our @ISA = qw( SanUtil );

# constant variable
use constant NOERRORFOUND => 'NoErrorFound';

###############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class) = @_;
    my $self = {'client_verification' => $SanUtil::FALSE};

    # bless instance into class
    bless $self, $class;
    return $self;
}

###############################################################################
# Method:    create_igroup_in_vserver
# Objective: to create igroup object in given vserver and build the expected
#            igroup attributes (as user requested) for later verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_igroup_in_vserver {
    my ($self, %args) = @_;
    my ($vs_obj, $vs_name, $igrp_attr_expected, $init_names);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(IGRP_VSERVER_OBJ) => {type     => OBJECT,
                                           required => 1,
                                           isa      => 'San::ControllerCluster::Filer10M::Vserver'},
                    (IGRP_ATTR_IGROUP_NAME)  => {type => SCALAR, optional => 1},
                    (IGRP_ATTR_PROTOCOL)     => {type => SCALAR, optional => 1},
                    (IGRP_ATTR_OS_TYPE)      => {type => SCALAR, optional => 1},
                    (IGRP_ATTR_INITIATOR)    => {type => SCALAR, optional => 1},
                    (IGRP_ATTR_PORTSET_NAME) => {type => SCALAR, optional => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return {};
    }

    # remove this arg from args list, since we are using same args list to
    # send it to vs_obj->create_igroup_obj(), which is invalid argument.
    $vs_obj = delete $args{(IGRP_VSERVER_OBJ)};

    # get the name of the vserver
    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get vserver name.\nReason:\n" . $vs_obj->errmsg());
        return {};
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg       => _dump_message(\%args, 1),
                  -msg_level => 5);

    my $cmd_interface = $vs_obj->get_cmd_interface();
    if ($cmd_interface =~ /^zapi$/i) {

        # for zapi portocol type is must param in cmode,
        # so default to mixed, if user does not pass it
        unless (exists($args{(IGRP_ATTR_PROTOCOL)})) {
            $args{(IGRP_ATTR_PROTOCOL)} = 'mixed';
        }

        # zapi does not support to create igroup with initiator
        # so remove it here and add those initiator later(see below code)...
        if (exists($args{(IGRP_ATTR_INITIATOR)})) {
            $init_names = delete $args{(IGRP_ATTR_INITIATOR)};
        }
    }

    # create igroup object
    my $igrp_obj = $vs_obj->create_igroup_obj(%args);
    if ($vs_obj->error()) {
        my $errmsg = $vs_obj->errmsg();

        $self->_set_err(  "Failed to create igroup("
                        . $args{(IGRP_ATTR_IGROUP_NAME)} . ") "
                        . "on vserver($vs_name)\nReason:\n"
                        . $errmsg);
        return {};
    }

    # build the hash - expected igroup attributes for later verification
    # igroup show command will have following attributes for given igroup.
    # vserver|igroup|protocol-type|ostype|portset|uuid|alua|init-details|
    # Sample output
    # igrp_vs_0|igr1|iscsi|solaris|-|d580f63c-cbfd-11df-9a6e-123478563412|true|"iqn.1992-08.com.netapp.0.979.543 (not logged in)","iqn.1992-08.com.netapp.1.121.739 (not logged in)"|

    foreach my $igrp_attr_key (@ExpectedIgroupAttributes) {
        if (exists $args{$igrp_attr_key}) {

            # if key is IGRP_ATTR_INITIATOR or IGRP_ATTR_INITIATOR_DETAILS
            # fill the default, it taken care by below code...
            if ($igrp_attr_key eq IGRP_ATTR_INITIATOR) {
                $igrp_attr_expected->{$igrp_attr_key} = $IgroupAttributesDefault{$igrp_attr_key};
            } else {
                $igrp_attr_expected->{$igrp_attr_key} = $args{$igrp_attr_key};
            }
        } else {
            $igrp_attr_expected->{$igrp_attr_key} = $IgroupAttributesDefault{$igrp_attr_key};
            if ($cmd_interface =~ /^zapi$/i) {
                if ($igrp_attr_key eq IGRP_ATTR_INITIATOR_DETAILS) {
                    $igrp_attr_expected->{$igrp_attr_key} = IGRP_ATTR_ZAPI_NA;
                }
            }
        }
    }

    # fill the vserver name associated with igroup, since user wont pass this
    $igrp_attr_expected->{(IGRP_ATTR_VSERVER_NAME)} = $vs_name;

    # fill up the igroup object also....
    $igrp_attr_expected->{(IGRP_IGROUP_OBJ)} = $igrp_obj;

    # if user pass the initiator as alais name, fill the initiator
    # filed with corresponding wwpn of the alais.

    if (exists $args{(IGRP_ATTR_INITIATOR)}) {
        $self->_update_initiators((IGRP_ATTR_INFO) => $igrp_attr_expected,
                                  '-initiators'    => $args{(IGRP_ATTR_INITIATOR)});
        if ($self->error()) {

            # error already set in the function
            # remove the igroup and just return
            $igrp_obj->delete();
            return {};
        }
    }

    # if cmd_interface is zapi and user requested to add ports,
    # then add it after creating the igroup.
    if ($cmd_interface =~ /^zapi$/i and length($init_names)) {
        $self->add_initiators_to_igroup((IGRP_ATTR_INFO)      => $igrp_attr_expected,
                                        (IGRP_ATTR_INITIATOR) => $init_names);
        if ($self->error()) {
            my $err_str = $self->errmsg();
            $self->_clear_err();
            $self->_set_err(  "Failed to add initiator while creating igroup through zapi("
                            . $args{(IGRP_ATTR_IGROUP_NAME)} . ") "
                            . "on vserver($vs_name)\nReason:\n"
                            . $err_str);

            # since igroup is created, remove the igroup
            # and return the empty string...
            $igrp_obj->delete('-force' => "true");
            return {};
        }
    }

    # get the igroup info from controller, and update the UUID attributes
    # execute with diag privilege since uuid filed is exposed via diag when using zapi
    my $igrp_attr_controller = $igrp_obj->get_list_info(cmd_privilege => "diag");
    if ($igrp_obj->error()) {
        $self->_set_err("Unable to get igroup information." . $igrp_obj->errmsg());

        # since igroup is created, remove the igroup
        # and return the empty string...
        $igrp_obj->delete('-force' => "true");
        return {};
    }

    # update the UUID field
    $igrp_attr_expected->{(IGRP_ATTR_UUID)} = $igrp_attr_controller->{(IGRP_ATTR_UUID)};

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(-msg       => "Expected Igroup Attributes:\n" . _dump_message($igrp_attr_expected, 1),
                  -msg_level => 3);
    return $igrp_attr_expected;
}    # end sub create_igroup_in_vserver

###############################################################################
# Method:    add_initiators_to_igroup
# Objective: to add initiators to existing igroup and update the
#            expected igroup attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_initiators_to_igroup {
    my ($self, %args) = @_;
    my ($igrp_obj, $igrp_name, $igrp_attr_expected);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(IGRP_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Igroup object not present in the Hashref" => sub {
                         $_[0]->{(IGRP_IGROUP_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Igroup/;
                     },
                     "Attributes not qualified" => sub {
                         foreach my $igrp_attr_key (@ExpectedIgroupAttributes) {
                             if (not exists $_[0]->{$igrp_attr_key}) {
                                 return 0;
                             }
                         }
                         return 1;
                       }
                 }},
             (IGRP_ATTR_INITIATOR) => {type => SCALAR, required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => _dump_message(\%args, 1), -msg_level => 5);

    # remove this args((IGRP_ATTR_INFO)) from args list,
    # since we are using same args list to
    # send it to igrp_obj->add(), which is invalid argument.
    $igrp_attr_expected = delete $args{(IGRP_ATTR_INFO)};
    $igrp_obj           = $igrp_attr_expected->{(IGRP_IGROUP_OBJ)};
    $igrp_name          = $igrp_attr_expected->{(IGRP_ATTR_IGROUP_NAME)};

    # add requested ports to portset
    my $cmd_interface = $igrp_obj->get_cmd_interface();

    # through zapi we can add only one port at a time
    if ($cmd_interface =~ /^zapi$/i) {
        my @initiators = split(",", $args{(IGRP_ATTR_INITIATOR)});
        foreach my $initiator (@initiators) {
            $igrp_obj->add((IGRP_ATTR_INITIATOR) => $initiator);
            if ($igrp_obj->error()) {
                $self->_set_err(
                     "Failed to add initiator to " . "igroup($igrp_name) using zapi \nReason:\n" . $igrp_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # update the expected igroup attribute(initiator)
            $self->_update_initiators((IGRP_ATTR_INFO) => $igrp_attr_expected,
                                      '-initiators'    => $initiator);
            if ($self->error()) {

                # error already set in the function, just return
                return $SanUtil::FALSE;
            }
        }    # end of foreach my $port_name (@ports)
    } else {

        # add requested initiators to igroup
        $igrp_obj->add(%args);
        if ($igrp_obj->error()) {
            $self->_set_err("Failed to add initiator(s) to " . "igroup($igrp_name)\nReason:\n" . $igrp_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # update the expected igroup attribute(initiator)

        $self->_update_initiators((IGRP_ATTR_INFO) => $igrp_attr_expected,
                                  '-initiators'    => $args{(IGRP_ATTR_INITIATOR)});
        if ($self->error()) {

            # error already set in the function, just return
            return $SanUtil::FALSE;
        }
    }

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(-msg       => "Expected Igroup Attributes:\n" . _dump_message($igrp_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub add_initiators_to_igroup

###############################################################################
# Method:    remove_initiators_from_igroup
# Objective: to remove initiators from existing igroup and update the
#            expected igroup attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_initiators_from_igroup {
    my ($self, %args) = @_;
    my ($vs_obj, $igrp_obj, $igrp_name, $igrp_attr_expected);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(IGRP_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Igroup object not present in the Hashref" => sub {
                         $_[0]->{(IGRP_IGROUP_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Igroup/;
                     },
                     "Attributes not qualified" => sub {
                         foreach my $igrp_attr_key (@ExpectedIgroupAttributes) {
                             if (not exists $_[0]->{$igrp_attr_key}) {
                                 return 0;
                             }
                         }
                         return 1;
                       }
                 }},
             (IGRP_ATTR_INITIATOR) => {type => SCALAR, required => 1},
             (IGRP_ATTR_FORCE)     => {type => SCALAR, optional => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg       => _dump_message(\%args, 1),
                  -msg_level => 5);

    # remove this args((IGRP_ATTR_INFO)) from args list,
    # since we are using same args list to
    # send it to igrp_obj->remove(), which is invalid argument.
    $igrp_attr_expected = delete $args{(IGRP_ATTR_INFO)};

    $igrp_obj  = $igrp_attr_expected->{(IGRP_IGROUP_OBJ)};
    $igrp_name = $igrp_attr_expected->{(IGRP_ATTR_IGROUP_NAME)};

    $vs_obj = $igrp_obj->get_parent_obj();
    if ($igrp_obj->error()) {
        $self->_set_err("Unable to get vserver name\nReason:\n" . $igrp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    my $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Unable to get vserver name\nReason:\n" . $vs_obj->errmsg());
        return $SanUtil::FALSE;
    }
    my $fcp_obj = "";

    my $cmd_interface = $igrp_obj->get_cmd_interface();

    # through zapi we can add only one port at a time
    if ($cmd_interface =~ /^zapi$/i) {
        my @initiators = split(",", $args{(IGRP_ATTR_INITIATOR)});
        foreach my $initiator (@initiators) {
            my %cmd_args;
            $cmd_args{(IGRP_ATTR_INITIATOR)} = $initiator;
            if (exists $args{(IGRP_ATTR_FORCE)}) {
                $cmd_args{(IGRP_ATTR_FORCE)} = $args{(IGRP_ATTR_FORCE)};
            }
            $igrp_obj->remove(%cmd_args);
            if ($igrp_obj->error()) {
                $self->_set_err(  "Failed to remove initiator from "
                                . "igroup($igrp_name) using zapi \nReason:\n"
                                . $igrp_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # if it is fcp alias name then find out the corresponding wwpn
            if (   $initiator !~ /^(iqn\.\d{4}-\d{2}\.\w+)|^(eui\.[0-9a-fA-F]+)/
                && $initiator !~ /^([0-9a-fA-F]{2}:?){7}[0-9a-fA-F]{2}$/) {

                # get the fcp object to create alias
                if ($fcp_obj eq "") {
                    $fcp_obj = $vs_obj->get_fcp_obj();
                    if ($vs_obj->error()) {
                        $self->_set_err(
                                   "Unable to get fcp object of " . "vserver($vs_name)\nReason:\n" . $vs_obj->errmsg());
                        return $SanUtil::FALSE;
                    }
                }

                my @alias_wwpn = $fcp_obj->wwpnalias_get('alias' => $initiator);
                if ($fcp_obj->error()) {
                    $self->_set_err(
                            "Unable to query alais for wwpn on " . "vserver($vs_name)\nReason:\n" . $fcp_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                my $wwpn;
                if ("@alias_wwpn" =~ /There are no entries matching your query/
                    || scalar(@alias_wwpn) <= 0) {
                    $wwpn = '-';
                } else {
                    my @wwpn_names;
                    foreach my $init (@alias_wwpn) {
                        push(@wwpn_names, $init->{'wwpn'});
                    }

                    # there will be one wwpn with given alias
                    $wwpn = $wwpn_names[0];
                }
                $initiator = $wwpn;
            }
            $igrp_attr_expected->{(IGRP_ATTR_INITIATOR)} =~ s/$initiator,*//;

            # $igrp_attr_expected->{(IGRP_ATTR_INITIATOR_DETAILS)} =~ s/"$initiator .+",*//;
        }    # end of foreach my $port_name (@ports)
    } else {

        # remove requested initiators from igroup
        $igrp_obj->remove(%args);
        if ($igrp_obj->error()) {
            $self->_set_err(
                         "Failed to remove initiator(s) from " . "igroup($igrp_name)\nReason:\n" . $igrp_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # update the expected igroup attribute(initiator)
        my @init_to_remove = split(',', $args{(IGRP_ATTR_INITIATOR)});
        foreach my $initiator (@init_to_remove) {

            # if it is fcp alias name then find out the corresponding wwpn
            if (   $initiator !~ /^(iqn\.\d{4}-\d{2}\.\w+)|^(eui\.[0-9a-fA-F]+)/
                && $initiator !~ /^([0-9a-fA-F]{2}:?){7}[0-9a-fA-F]{2}$/) {

                # get the fcp object to create alias
                if ($fcp_obj eq "") {
                    $fcp_obj = $vs_obj->get_fcp_obj();
                    if ($vs_obj->error()) {
                        $self->_set_err(
                                   "Unable to get fcp object of " . "vserver($vs_name)\nReason:\n" . $vs_obj->errmsg());
                        return $SanUtil::FALSE;
                    }
                }

                my @alias_wwpn = $fcp_obj->wwpnalias_get('alias' => $initiator);
                if ($fcp_obj->error()) {
                    $self->_set_err(
                            "Unable to query alais for wwpn on " . "vserver($vs_name)\nReason:\n" . $fcp_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                my $wwpn;
                if ("@alias_wwpn" =~ /There are no entries matching your query/
                    || scalar(@alias_wwpn) <= 0) {
                    $wwpn = '-';
                } else {
                    my @wwpn_names;
                    foreach my $init (@alias_wwpn) {
                        push(@wwpn_names, $init->{'wwpn'});
                    }

                    # there will be one wwpn with given alias
                    $wwpn = $wwpn_names[0];
                }
                $initiator = $wwpn;
            }
            $igrp_attr_expected->{(IGRP_ATTR_INITIATOR)}         =~ s/$initiator,*//;
            $igrp_attr_expected->{(IGRP_ATTR_INITIATOR_DETAILS)} =~ s/"$initiator .+",*//;
        }
    }

    # if no initiator left fill with default
    unless ($igrp_attr_expected->{(IGRP_ATTR_INITIATOR)}) {
        $igrp_attr_expected->{(IGRP_ATTR_INITIATOR)} = $IgroupAttributesDefault{(IGRP_ATTR_INITIATOR)};
    }
    unless ($igrp_attr_expected->{(IGRP_ATTR_INITIATOR_DETAILS)}) {
        $igrp_attr_expected->{(IGRP_ATTR_INITIATOR_DETAILS)} = $IgroupAttributesDefault{(IGRP_ATTR_INITIATOR_DETAILS)};
    }

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(-msg       => "Expected Igroup Attributes:\n" . _dump_message($igrp_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub remove_initiators_from_igroup

###############################################################################
# Method:    rename_igroup_in_vserver
# Objective: to rename the existing igroup on given vserver and update the
#            expected igroup attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub rename_igroup_in_vserver {
    my ($self, %args) = @_;
    my ($igrp_obj, $igrp_name, $igrp_attr_expected);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(IGRP_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Igroup object not present in the Hashref" => sub {
                         $_[0]->{(IGRP_IGROUP_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Igroup/;
                     },
                     "Attributes not qualified" => sub {
                         foreach my $igrp_attr_key (@ExpectedIgroupAttributes) {
                             if (not exists $_[0]->{$igrp_attr_key}) {
                                 return 0;
                             }
                         }
                         return 1;
                       }
                 }},
             (IGRP_ATTR_RENAME) => {type => SCALAR, required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => _dump_message(\%args, 1), -msg_level => 5);

    # remove this args((IGRP_ATTR_INFO)) from args list,
    # since we are using same args list to
    # send it to igrp_obj->rename(), which is invalid argument.
    $igrp_attr_expected = delete $args{(IGRP_ATTR_INFO)};

    $igrp_obj  = $igrp_attr_expected->{(IGRP_IGROUP_OBJ)};
    $igrp_name = $igrp_attr_expected->{(IGRP_ATTR_IGROUP_NAME)};

    # rename requested igroup to new name
    $igrp_obj->rename(%args);
    if ($igrp_obj->error()) {
        $self->_set_err(  "Failed to rename the igroup($igrp_name) "
                        . "to new name $args{(IGRP_ATTR_RENAME)}\nReason:\n"
                        . $igrp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected igroup attributes
    $igrp_attr_expected->{(IGRP_ATTR_IGROUP_NAME)} = $args{(IGRP_ATTR_RENAME)};

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(-msg       => "Expected Igroup Attributes:\n" . _dump_message($igrp_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub rename_igroup_in_vserver

###############################################################################
# Method:    bind_igroup_in_vserver
# Objective: to bind the portset to the existing igroup on given vserver and
#            update the expected igroup attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub bind_igroup_in_vserver {
    my ($self, %args) = @_;
    my ($igrp_obj, $igrp_name, $igrp_attr_expected, $portset_name);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(IGRP_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Igroup object not present in the Hashref" => sub {
                         $_[0]->{(IGRP_IGROUP_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Igroup/;
                     },
                     "Attributes not qualified" => sub {
                         foreach my $igrp_attr_key (@ExpectedIgroupAttributes) {
                             if (not exists $_[0]->{$igrp_attr_key}) {
                                 return 0;
                             }
                         }
                         return 1;
                       }
                 }},
             (IGRP_ATTR_PORTSET_NAME) => {type => SCALAR, required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => _dump_message(\%args, 1), -msg_level => 5);

    # remove this args((IGRP_ATTR_INFO)) from args list,
    # since we are using same args list to
    # send it to igrp_obj->bind(), which is invalid argument.
    $igrp_attr_expected = delete $args{(IGRP_ATTR_INFO)};

    $igrp_obj     = $igrp_attr_expected->{(IGRP_IGROUP_OBJ)};
    $igrp_name    = $igrp_attr_expected->{(IGRP_ATTR_IGROUP_NAME)};
    $portset_name = $args{(IGRP_ATTR_PORTSET_NAME)};

    # bind requested igroup to the portset
    $igrp_obj->bind(%args);
    if ($igrp_obj->error()) {
        $self->_set_err(
               "Failed to bind the igroup($igrp_name) " . "to portset($portset_name)\nReason:\n" . $igrp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected igroup attributes
    $igrp_attr_expected->{(IGRP_ATTR_PORTSET_NAME)} = $portset_name;

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(-msg       => "Expected Igroup Attributes:\n" . _dump_message($igrp_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub bind_igroup_in_vserver

###############################################################################
# Method:    unbind_igroup_in_vserver
# Objective: to unbind the portset to the existing igroup on given vserver and
#            update the expected igroup attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unbind_igroup_in_vserver {
    my ($self, %args) = @_;
    my ($igrp_obj, $igrp_name, $igrp_attr_expected);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(IGRP_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Igroup object not present in the Hashref" => sub {
                         $_[0]->{(IGRP_IGROUP_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Igroup/;
                     },
                     "Attributes not qualified" => sub {
                         foreach my $igrp_attr_key (@ExpectedIgroupAttributes) {
                             if (not exists $_[0]->{$igrp_attr_key}) {
                                 return 0;
                             }
                         }
                         return 1;
                       }
                 }}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => _dump_message(\%args, 1), -msg_level => 5);

    $igrp_attr_expected = $args{(IGRP_ATTR_INFO)};
    $igrp_obj           = $igrp_attr_expected->{(IGRP_IGROUP_OBJ)};
    $igrp_name          = $igrp_attr_expected->{(IGRP_ATTR_IGROUP_NAME)};

    # unbind requested igroup from portset
    $igrp_obj->unbind();
    if ($igrp_obj->error()) {
        $self->_set_err("Failed to unbind the igroup($igrp_name)" . "\nReason:\n" . $igrp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected igroup attributes
    $igrp_attr_expected->{(IGRP_ATTR_PORTSET_NAME)} = "-";

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(-msg       => "Expected Igroup Attributes\n" . _dump_message($igrp_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub bind_igroup_in_vserver

###############################################################################
# Method:    set_igroup_in_vserver
# Objective: to set the existing igroup attributes on given vserver and update the
#            expected igroup attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_igroup_in_vserver {
    my ($self, %args) = @_;
    my ($igrp_obj, $igrp_name, $igrp_attr_expected);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(IGRP_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Igroup object not present in the Hashref" => sub {
                         $_[0]->{(IGRP_IGROUP_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Igroup/;
                     },
                     "Attributes not qualified" => sub {
                         foreach my $igrp_attr_key (@ExpectedIgroupAttributes) {
                             if (not exists $_[0]->{$igrp_attr_key}) {
                                 return 0;
                             }
                         }
                         return 1;
                       }
                 }},
             (IGRP_ATTR_ATTRIBUTE) => {type => SCALAR, required => 1},
             (IGRP_ATTR_VALUE)     => {type => SCALAR, required => 1},
             (IGRP_ATTR_FORCE)     => {type => SCALAR, optional => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => _dump_message(\%args, 1), -msg_level => 5);

    # remove this args((IGRP_ATTR_INFO)) from args list,
    # since we are using same args list to
    # send it to igrp_obj->rename(), which is invalid argument.
    $igrp_attr_expected = delete $args{(IGRP_ATTR_INFO)};
    $igrp_obj           = $igrp_attr_expected->{(IGRP_IGROUP_OBJ)};
    $igrp_name          = $igrp_attr_expected->{(IGRP_ATTR_IGROUP_NAME)};

    # set requested igroup attribute to new value
    $igrp_obj->set(%args);
    if ($igrp_obj->error()) {
        $self->_set_err(  "Failed to set the igroup($igrp_name) "
                        . "attribute:$args{(IGRP_ATTR_ATTRIBUTE)} "
                        . "to new value $args{(IGRP_ATTR_VALUE)}\nReason:\n"
                        . $igrp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected igroup attributes

    $igrp_attr_expected->{$args{(IGRP_ATTR_ATTRIBUTE)}} = $args{(IGRP_ATTR_VALUE)};

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(-msg       => "Expected Igroup Attributes:\n" . _dump_message($igrp_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub set_igroup_in_vserver

###############################################################################
# Method:    modify_igroup_in_vserver
# Objective: to modify the existing igroup attributes on given vserver and update the
#            expected igroup attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub modify_igroup_in_vserver {
    my ($self, %args) = @_;
    my ($igrp_obj, $igrp_name, $igrp_attr_expected);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(IGRP_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Igroup object not present in the Hashref" => sub {
                         $_[0]->{(IGRP_IGROUP_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Igroup/;
                     },
                     "Attributes not qualified" => sub {
                         foreach my $igrp_attr_key (@ExpectedIgroupAttributes) {
                             if (not exists $_[0]->{$igrp_attr_key}) {
                                 return 0;
                             }
                         }
                         return 1;
                       }
                 }},
             (IGRP_ATTR_OS_TYPE) => {type => SCALAR, optional => 1},
             (IGRP_ATTR_ALUA)    => {type => SCALAR, optional => 1},
             (IGRP_ATTR_FORCE)   => {type => SCALAR, optional => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => _dump_message(\%args, 1), -msg_level => 5);

    # remove this args((IGRP_ATTR_INFO)) from args list,
    # since we are using same args list to
    # send it to igrp_obj->rename(), which is invalid argument.
    $igrp_attr_expected = delete $args{(IGRP_ATTR_INFO)};
    $igrp_obj           = $igrp_attr_expected->{(IGRP_IGROUP_OBJ)};
    $igrp_name          = $igrp_attr_expected->{(IGRP_ATTR_IGROUP_NAME)};

    # modify requested igroup attributtes
    $igrp_obj->modify(%args);
    if ($igrp_obj->error()) {
        $self->_set_err("Failed to modify the igroup($igrp_name) " . "\nReason:\n" . $igrp_obj->errmsg());
        return $SanUtil::FALSE;
    }
    foreach my $attr (keys %args) {

        # since IGRP_ATTR_FORCE is not a igroup attribute of igroup show
        # skip it.
        if ($attr eq IGRP_ATTR_FORCE) {
            next;
        }

        # since it is one value(not list of values like initiators)
        # just assign what user passed
        if (defined $args{$attr}) {
            $igrp_attr_expected->{$attr} = $args{$attr};
        }
    }

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(-msg       => "Expected Igroup Attributes:\n" . _dump_message($igrp_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub modify_igroup_in_vserver

###############################################################################
# Method:    verify_igroup_attributes
# Objective: to verify the igroup attributes are intact.
#            compare the requested attributes with the controller
#            return igroup attributes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_igroup_attributes {
    my ($self, %args) = @_;
    my ($igrp_obj, $igrp_name, %igrp_attr_user, $igrp_attr_controller, @err_str);
    my $err_cnt = 0;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(IGRP_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Igroup object not present in the Hashref" => sub {
                         $_[0]->{(IGRP_IGROUP_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Igroup/;
                     },
                     "Attributes not qualified" => sub {
                         foreach my $igrp_attr_key (@ExpectedIgroupAttributes) {
                             if (not exists $_[0]->{$igrp_attr_key}) {
                                 return 0;
                             }
                         }
                         return 1;
                       }
                 }}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => _dump_message(\%args, 1), -msg_level => 5);

    # copy the attributes, since we need to delete the igroup_obj,
    # do not operate on reference
    %igrp_attr_user = %{$args{(IGRP_ATTR_INFO)}};
    $igrp_obj       = $igrp_attr_user{((IGRP_IGROUP_OBJ))};

    # since igroup_obj is not part of igroup attribute,
    # discard from $igrp_attr_user
    delete $igrp_attr_user{((IGRP_IGROUP_OBJ))};

    # get the name of igroup from "$igrp_attr_user"
    $igrp_name = $igrp_attr_user{(IGRP_ATTR_IGROUP_NAME)};

    # get the igroup info from controller, return reference to hash
    # need to send diag privileage to get the detail initiator info.
    # like alias name, logged in status.
    $igrp_attr_controller = $igrp_obj->get_list_info(cmd_privilege => "diag");
    if ($igrp_obj->error()) {
        $self->_set_err("Unable to get igroup information." . $igrp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # in case, if there is no requested igroup,
    # 'get_list_info' won't return error, so make sure we have valid info.
    # ex:
    # > lun igroup show -vserver igrp_basic0 -igroup _DELETE_
    # > There are no entries matching your query.
    # this will not return error
    unless (ref($igrp_attr_controller) =~ /HASH/
            and exists $igrp_attr_controller->{(IGRP_ATTR_IGROUP_NAME)}) {
        $self->_set_err(
                 "Not a valid igroup information.\n" . "Controller return: " . _dump_message($igrp_attr_controller, 1));
        return $SanUtil::FALSE;
    }

    # display the user request igroup attributes if debug_level is high
    $self->sanlog(-msg       => "Expected:\n" . _dump_message(\%igrp_attr_user, 1),
                  -msg_level => 3);

    # display the controller igroup attributes if debug_level is high
    $self->sanlog(-msg       => "Controller:\n" . _dump_message($igrp_attr_controller, 1),
                  -msg_level => 3);

    my $cmd_interface = $igrp_obj->get_cmd_interface();

    # verify the igroup attributes from user expected list
    # with controller list and return one error message with
    # list of attributes which is/are not matching. User can set
    # _SKIP_CHECK_ value for particular key to skip checking.
    # for key IGRP_ATTR_INITIATOR, the value will be comma separated list.
    # since order of that list will be different, need to check individual
    # values present in the controller list. "_verify_igroup_initiator"
    # does that.
    @err_str = ("Igroup($igrp_name) attributes does not match");
    foreach my $key (keys %igrp_attr_user) {

        # if caller did not care about this particular attribute
        if ($igrp_attr_user{$key} =~ /_SKIP_CHECK_/) {
            next;
        }
        unless (defined $igrp_attr_controller->{$key}) {
            $err_cnt++;
            push(@err_str, "\t$key is not defined in the controller list info");
            next;
        }

        # if the checking key (IGRP_ATTR_INITIATOR_DETAILS) may have
        # comma separated values and
        # it may return in  different order,
        # so check for existence in the list.

        if (   ($key eq IGRP_ATTR_INITIATOR_DETAILS and $cmd_interface =~ /^cli$/)
            || ($key eq IGRP_ATTR_INITIATOR)) {
            my @verify_array = _verify_igroup_initiator($key, $igrp_attr_user{$key}, $igrp_attr_controller->{$key});
            if ($verify_array[0] ne NOERRORFOUND) {
                $err_cnt++;
                push(@err_str,
                         "\t$key => Expected: "
                       . "'$igrp_attr_user{$key}' <-> "
                       . "Controller: '$igrp_attr_controller->{$key}'"
                       . "\nERROR=@verify_array");
            }
        } else {
            if ($igrp_attr_user{$key} ne $igrp_attr_controller->{$key}) {
                $err_cnt++;
                push(@err_str,
                         "\t$key => Expected: "
                       . "'$igrp_attr_user{$key}' <-> "
                       . "Controller: '$igrp_attr_controller->{$key}'");
            }
        }
    }
    if ($err_cnt) {
        my $text = "No. of errors encountered: $err_cnt\n" . join("\n", @err_str);
        $self->_set_err($text);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}    # end sub verify_igroup_attributes

###############################################################################
# Method:    verify_igroup_unique_uuid
# Objective: To verify the igroup uuid uniqueness.
#            verify only test case created igroup
#            verify if igroup created across vserver.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_igroup_unique_uuid {

    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(@input_params, {'igrp_array_ref' => {type => ARRAYREF, required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    my @igrp_objs_array = @{$args{'igrp_array_ref'}};
    my @igroup_uuid     = ();
    foreach my $igrp_attr_expected (@igrp_objs_array) {
        my $igrp_name    = $igrp_attr_expected->{(IGRP_ATTR_IGROUP_NAME)};
        my $vserver_name = $igrp_attr_expected->{(IGRP_ATTR_VSERVER_NAME)};
        my $uuid         = $igrp_attr_expected->{(IGRP_ATTR_UUID)};
        $self->sanlog(-msg => "Unique UUID Verification: " . "Igroup: $igrp_name, Vserver: $vserver_name, UUID: $uuid",
                      -msg_level => 3);
        if (!grep { /^$uuid/ } @igroup_uuid) {
            push(@igroup_uuid, $uuid);
        } else {
            $self->_set_err("Duplicate UUID found. Igroup: $igrp_name, Vserver: $vserver_name, UUID: $uuid");
            return $SanUtil::FALSE;
        }
    }
    return $SanUtil::TRUE;
}    # end of sub verify_igroup_unique_uuid

###############################################################################
# Method:    verify_igroup_portset_bind_unbind
# Objective: to verify the igroup and portset bind/unbind info
#            using igroup show and portset show (cli or zapi)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_igroup_portset_bind_unbind {
    my ($self, %args) = @_;

    my ($ps_obj, $igroup_obj, $igrp_attr_controller, $ps_attr_controller);
    my ($ps_name, $igroup_name, $action);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {'igroup_obj' => {type     => OBJECT,
                                     required => 1,
                                     isa      => 'San::ControllerCluster::Filer10M::Vserver::Igroup'},
                    'portset_obj' => {type     => OBJECT,
                                      required => 1,
                                      isa      => 'San::ControllerCluster::Filer10M::Vserver::Portset'},
                    'action' => {type     => SCALAR,
                                 required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    $action      = $args{'action'};
    $igroup_obj  = $args{'igroup_obj'};
    $igroup_name = $igroup_obj->get_name();
    if ($igroup_obj->error()) {
        $self->_set_err("Unable to get igroup name. Reason:\n" . $igroup_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $ps_obj  = $args{'portset_obj'};
    $ps_name = $ps_obj->get_name();
    if ($ps_obj->error()) {
        $self->_set_err("Unable to get portset name. Reason:\n" . $ps_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # get the igroup info from controller, return reference to hash
    $igrp_attr_controller = $igroup_obj->get_list_info();
    if ($igroup_obj->error()) {
        $self->_set_err("Unable to get igroup information." . $igroup_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # get the portset info from controller, return reference to hash
    $ps_attr_controller = $ps_obj->get_list_info();
    if ($ps_obj->error()) {
        $self->_set_err("Unable to get portset information." . $ps_obj->errmsg());
        return $SanUtil::FALSE;
    }

    if ($action =~ /^bind$/) {

        # verify igroup info has the portset name on it
        if ($igrp_attr_controller->{(IGRP_ATTR_PORTSET_NAME)} ne $ps_name) {
            $self->_set_err(  "Igroup info from controller indicates that "
                            . "Igroup and portset is not matching the binding "
                            . "information:\nExpected portset name: $ps_name\n"
                            . "Controller return: "
                            . $igrp_attr_controller->{(IGRP_ATTR_PORTSET_NAME)});
            return $SanUtil::FALSE;
        }

        # just need to make sure this igroup exist in the list.
        # verify igroup info has the portset name on it
        if ($ps_attr_controller->{(PS_ATTR_IGROUP)} !~ /\b$igroup_name\b/) {
            $self->_set_err(  "Portset info from controller indicates that "
                            . "Igroup and portset is not matching the binding "
                            . "information:\nExpected igroup name: $igroup_name\n"
                            . "Controller return: "
                            . $ps_attr_controller->{(PS_ATTR_IGROUP)});
            return $SanUtil::FALSE;
        }
    } else {

        # verify igroup info do not have the portset name on it
        if ($igrp_attr_controller->{(IGRP_ATTR_PORTSET_NAME)} ne "-") {
            $self->_set_err(  "Igroup info from controller indicates that "
                            . "Portset($ps_name) not unbind from igroup($igroup_name)\n"
                            . "Expected portset name: '-'\n"
                            . "Controller return: "
                            . $igrp_attr_controller->{(IGRP_ATTR_PORTSET_NAME)});
            return $SanUtil::FALSE;
        }

        # make sure this igroup does not exist in the list.
        if ($ps_attr_controller->{(PS_ATTR_IGROUP)} =~ /\b$igroup_name\b/) {
            $self->_set_err(  "Portset info from controller indicates that "
                            . "Igroup($igroup_name) is still bound to "
                            . "portset($ps_name):\nPortset Controller return:"
                            . $ps_attr_controller->{(PS_ATTR_IGROUP)});
            return $SanUtil::FALSE;
        }
    }
    return $SanUtil::TRUE;
}    # end sub verify_portset_igroup_bind

###############################################################################
# Method:    get_igroup_attributes
# Objective: To get the igroup attributes hash for a given a igroup object for
#            using IgroupHelper function. Return igroup expected attribute
#            hash same like create_igroup_in_vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_igroup_attributes {
    my ($self, %args) = @_;

    my ($igroup_obj, $igrp_attr_controller, $igrp_attr_expected);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(IGRP_IGROUP_OBJ) => {type     => OBJECT,
                                          required => 1,
                                          isa      => 'San::ControllerCluster::Filer10M::Vserver::Igroup'}});
    };
    if ($@) {
        $self->_set_err($@);
        return {};
    }
    $igroup_obj = $args{(IGRP_IGROUP_OBJ)};

    # get the portset info from controller, return reference to hash
    $igrp_attr_controller = $igroup_obj->get_list_info(cmd_privilege => "diag");
    if ($igroup_obj->error()) {
        $self->_set_err("Unable to get igroup information." . $igroup_obj->errmsg());
        return {};
    }
    unless (ref($igrp_attr_controller) =~ /HASH/
            and exists $igrp_attr_controller->{(IGRP_ATTR_IGROUP_NAME)}) {
        $self->_set_err(
                 "Not a valid igroup information.\n" . "Controller return: " . _dump_message($igrp_attr_controller, 1));
        return {};
    }

    foreach my $igrp_attr (@ExpectedIgroupAttributes) {
        $igrp_attr_expected->{$igrp_attr} = $igrp_attr_controller->{$igrp_attr};
    }
    $igrp_attr_expected->{(IGRP_IGROUP_OBJ)} = $igroup_obj;

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(-msg       => "Expected Igroup Attributes:\n" . _dump_message($igrp_attr_expected, 1),
                  -msg_level => 3);
    return $igrp_attr_expected;
}    # end sub get_igroup_attributes

###############################################################################
#                        Private Functions
###############################################################################

###############################################################################
# Method:    _verify_igroup_initiator
# Objective: to verify the given igroup initiator attributes are intact.
#            compare the requested attributes with the controller
# Parameter:  all param are required
#   $list_to_check         -> list of initiator name with comma separated
#   $list_from_controller  -> list of initiator name with comma separated
# Return:  array of error string. If nothing found return array with
#          single element NOERRORFOUND.
###############################################################################
sub _verify_igroup_initiator {
    my ($igrp_attr_key, $initiator_expected, $initiator_controller) = @_;
    my (@expected_initiator, @controller_initiator);
    my @err_str = ();

    # if there is no initaitor, then check for defualt("-")
    if (   $initiator_expected eq "-"
        && $initiator_controller eq "-") {
        return (NOERRORFOUND);
    }

    @expected_initiator   = split(',', $initiator_expected);
    @controller_initiator = split(',', $initiator_controller);

    if (scalar(@expected_initiator) != scalar(@controller_initiator)) {
        push(@err_str, "\t\tNumber of entries in '$igrp_attr_key' does not match\n");
        return @err_str;
    }

    # "igroup show" might have attribute (initiator|init_details) with multiple
    # name with comma separated. The list is not same as created. So we need
    # verify all the name is presented as expected(instead simply checking
    # the whole string, since ORDER is different)., so sort the arrays
    @expected_initiator   = sort(@expected_initiator);
    @controller_initiator = sort(@controller_initiator);

    my $no_of_initiators = scalar(@expected_initiator);
    for (my $i = 0; $i < $no_of_initiators; $i++) {
        my $expt_initiator = $expected_initiator[$i];
        my $cont_initiator = $controller_initiator[$i];

        if ($igrp_attr_key eq IGRP_ATTR_INITIATOR_DETAILS) {

            # expected for IGRP_ATTR_INITIATOR_DETAILS
            # epxected initiator(details) foramt (inti_name (if fcp and alis presetn) (intiator logging status)
            # ""iqn.1992-08.com.netapp.0.799.627 (not logged in),"50:0a:09:80:81:02:96:d5 (not logged in)","50:0a:09:80:81:c7:a4:51 (a_1_500a098081c7a451) (not logged in)"

            # remove double quotos(" ") at beginign and ending
            $expt_initiator =~ s/"$//;
            $expt_initiator =~ s/^"//;
            $cont_initiator =~ s/"$//;
            $cont_initiator =~ s/^"//;
            my ($expt_init_name, $expt_alias, $expt_status);
            if ($expt_initiator =~ /(\S+) \((.+)\) \((.+)\)/) {
                $expt_init_name = $1;
                $expt_alias     = $2;
                $expt_status    = $3;
            } elsif ($expt_initiator =~ /(\S+) \((.+)\)/) {
                $expt_init_name = $1;
                $expt_status    = $2;
            } else {
                push(@err_str, "\t\tUnkown Expected Initiator format");
                next;
            }
            my ($cont_init_name, $cont_alias, $cont_status);
            if ($cont_initiator =~ /(\S+) \((.+)\) \((.+)\)/) {
                $cont_init_name = $1;
                $cont_alias     = $2;
                $cont_status    = $3;
            } elsif ($cont_initiator =~ /(\S+) \((.+)\)/) {
                $cont_init_name = $1;
                $cont_status    = $2;
            } else {
                push(@err_str, "\t\tUnkown Controller Initiator format");
                next;
            }
            if ($expt_init_name ne $cont_init_name) {
                push(@err_str, "\t\t$expt_init_name not found");
                next;
            }

            if ($expt_status ne $cont_status) {
                push(@err_str,
                     "\t\t$expt_init_name status not match" . "(Expected:$expt_status, Controller:$cont_status)");
                next;
            }

            # if alias presents, check for exist
            if (defined $expt_alias || defined $cont_alias) {
                my @expected_alias   = split(' ', $expt_alias);
                my @controller_alias = split(' ', $cont_alias);
                if (scalar(@expected_alias) != scalar(@controller_alias)) {
                    push(@err_str, "\t\tNumber of entries in 'alais' does not match");
                }
                @expected_alias   = sort(@expected_alias);
                @controller_alias = sort(@controller_alias);
                my $noofentries = @expected_alias;
                for (my $i = 0; $i < $noofentries; $i++) {
                    if ($expected_alias[$i] ne $controller_alias[$i]) {
                        push(@err_str, "\t\t$expected_alias[$i] not found");
                    }
                }
            }
        } else {

            # expected for IGRP_ATTR_INITIATOR
            # epxected initiator foramt (inti_name,init_name,init_name)
            if ($expt_initiator ne $cont_initiator) {
                push(@err_str, "\t\$expt_initiator not found");
            }
        }
    }
    if (scalar(@err_str)) {
        return @err_str;
    }
    return (NOERRORFOUND);
}    # end sub _verify_igroup_initiator

###############################################################################
# Method:    _update_initiators
# Objective: set the initiator and its associated keys
###############################################################################
sub _update_initiators {

    my ($self, %args) = @_;

    my ($igrp_obj, $vs_obj, $vs_name, $action, $igrp_attr_expected, $initiators);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(IGRP_ATTR_INFO) => {type => HASHREF, required => 1},
                    'initiators'     => {type => SCALAR,  required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => "_update_initiators::\n" . _dump_message(\%args, 1), -msg_level => 5);

    $igrp_attr_expected = $args{(IGRP_ATTR_INFO)};
    unless (exists $igrp_attr_expected->{(IGRP_IGROUP_OBJ)}) {
        $self->_set_err("Unable to get igroup object");
        return $SanUtil::FALSE;
    }

    $igrp_obj = $igrp_attr_expected->{(IGRP_IGROUP_OBJ)};
    unless (ref($igrp_obj) =~ /San::ControllerCluster::Filer10M::Vserver::Igroup/) {
        $self->_set_err("is not a igroup object");
        return $SanUtil::FALSE;
    }
    $initiators = $args{'initiators'};

    # get the fcp object from vserver object
    # using igroup object
    $vs_obj = $igrp_obj->get_parent_obj();
    if ($igrp_obj->error()) {
        $self->_set_err("Unable to get vserver name\nReason:\n" . $igrp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Unable to get vserver name\nReason:\n" . $vs_obj->errmsg());
        return $SanUtil::FALSE;
    }
    my $fcp_obj = "";

    my @list_of_initiators      = split(',', $initiators);
    my @initiators_list         = ();
    my @initiators_details_list = ();

    foreach my $initiator (@list_of_initiators) {
        my ($init_name, $alias, $status);
        push(@initiators_list, $initiator);

        # if $initiator is iscsi do nothing
        if ($initiator =~ /^(iqn\.\d{4}-\d{2}\.\w+)|^(eui\.[0-9a-fA-F]+)/) {
            $init_name = $initiator;

            # TODO - find the logging status
            $status = "(not logged in)";
        } elsif ($initiator =~ /^([0-9a-fA-F]{2}:?){7}[0-9a-fA-F]{2}$/) {
            $init_name = $initiator;

            # if $initiator is fcp wwpn, then find out its alias
            # get the fcp object to create alias
            if ($fcp_obj eq "") {
                $fcp_obj = $vs_obj->get_fcp_obj();
                if ($vs_obj->error()) {
                    $self->_set_err(
                                   "Unable to get fcp object of " . "vserver($vs_name)\nReason:\n" . $vs_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            }
            my @wwpn_alias = $fcp_obj->wwpnalias_get('wwpn' => $initiator);
            if ($fcp_obj->error()) {
                $self->_set_err(
                            "Unable to query alais for wwpn on " . "vserver($vs_name)\nReason:\n" . $fcp_obj->errmsg());
                return $SanUtil::FALSE;
            }
            if ("@wwpn_alias" !~ /There are no entries matching your query/
                && scalar(@wwpn_alias)) {

                # format would be "(a1 a2 a3)"
                $alias = '(';
                foreach my $init (@wwpn_alias) {
                    $alias .= " " . $init->{'alias'};
                }
                $alias .= ')';
            }

            # TODO - find the logging status
            $status = "(not logged in)";
        } else {

            # Here $initiator passed must alais of wwpn, then find out its wwpn
            # get the fcp object to get wwpn of given alias
            if ($fcp_obj eq "") {
                $fcp_obj = $vs_obj->get_fcp_obj();
                if ($vs_obj->error()) {
                    $self->_set_err(
                                   "Unable to get fcp object of " . "vserver($vs_name)\nReason:\n" . $vs_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            }

            # pass alias name to get wwpn(initator name)
            my @alias_wwpn = $fcp_obj->wwpnalias_get('alias' => $initiator);
            if ($fcp_obj->error()) {
                $self->_set_err(
                            "Unable to query alais for wwpn on " . "vserver($vs_name)\nReason:\n" . $fcp_obj->errmsg());
                return $SanUtil::FALSE;
            }
            my $wwpn;
            if ("@alias_wwpn" =~ /There are no entries matching your query/
                || scalar(@alias_wwpn) <= 0) {
                $self->_set_err("Unable to get wwpn of given alais($init_name)");
                return $SanUtil::FALSE;
            } else {
                my @wwpn_names;
                foreach my $init (@alias_wwpn) {
                    push(@wwpn_names, $init->{'wwpn'});
                }

                # there will be one wwpn with given alias
                $wwpn = $wwpn_names[0];
            }
            $alias     = '(' . $initiator . ')';
            $init_name = $wwpn;

            # since we add this in the begging of the forloop
            # pop it add the iniatior name
            pop(@initiators_list);
            push(@initiators_list, $init_name);

            # TODO - find the logging status
            $status = "(not logged in)";
        }
        my $formated_initiators = "\"$init_name ";
        if (defined $alias) {
            $formated_initiators .= "$alias ";
        }
        $formated_initiators .= "$status\"";
        push(@initiators_details_list, $formated_initiators);
    }
    if ($igrp_attr_expected->{(IGRP_ATTR_INITIATOR)} eq "-") {
        $igrp_attr_expected->{(IGRP_ATTR_INITIATOR)} = join(",", @initiators_list);
    } else {
        $igrp_attr_expected->{(IGRP_ATTR_INITIATOR)} .= "," . join(",", @initiators_list);
    }

    my $cmd_interface = $vs_obj->get_cmd_interface();
    if ($cmd_interface =~ /cli/i) {
        if ($igrp_attr_expected->{(IGRP_ATTR_INITIATOR_DETAILS)} eq "-") {
            $igrp_attr_expected->{(IGRP_ATTR_INITIATOR_DETAILS)} = join(",", @initiators_details_list);
        } else {
            $igrp_attr_expected->{(IGRP_ATTR_INITIATOR_DETAILS)} .= "," . join(",", @initiators_details_list);
        }
    } else {
        $igrp_attr_expected->{(IGRP_ATTR_INITIATOR_DETAILS)} = IGRP_ATTR_ZAPI_NA;
    }

    # dump the expected igroup attributes for debug purpose
    $self->sanlog(
               -msg       => "_update_initiators: Modified Igroup Attributes\n" . _dump_message($igrp_attr_expected, 1),
               -msg_level => 3);

    return $SanUtil::TRUE;
}    # end sub _update_initiators

###############################################################################
# Method:    _dump_message
# Objective: Get output from object
# Input:     Any hash or array object
# Return:    string from the dump
###############################################################################
sub _dump_message {
    my ($obj, $hash_level) = @_;

    # create new Data::Dumper object
    my $d_obj = Data::Dumper->new([$obj], [ref($obj)]);

    # Make Data::Dumper pretty, if it is hash
    if (ref($obj) =~ /HASH/) {
        $d_obj->Sortkeys(1);
        $d_obj->Indent(1);
    }
    $d_obj->Maxdepth($hash_level) if (length($hash_level));
    return $d_obj->Dump();
}    # end of _dump_message

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

IgroupHelper.pm

=head1 Author

Muthukumaran Govindarajan (muthug@netapp.com)

=head1 Description

IgroupHelper.pm provides an utility function interface to test igroup functionality
with the main intent to avoid repeated code from thpl level. It is design as Object Oriented model.
Other object(s) can derive this as base class or user can create object from it.

=head1 See Also

IgroupHelper.pm package is used as base class from San::Application::IgroupMgr. See San::Application::IgroupMgr
POD for more details.

B<San::Application::IgroupMgr>

=head1 Attributes

None

=head1 Revision History

Jan-18-2010 muthug@netapp.com: Original template created

=head1 Public Methods

=head2 B<new()>

Class Constructor method. To set the debug level to print debug message at particular level, call
'set_debug_level' method. See Example for more details

=over 4

=item Inputs

None

=item Outputs

Returns a IgroupHelper blessed object reference.

=item Examples

   #
   # Create a IgroupHelper object
   #
   my $igrp_helper_obj = new IgroupHelper();
   if ($igrp_helper_obj->error()) {
      my $err_msg = $igrp_helper_obj->errmsg();
      $Test->nlog("Failed to instantiate IgroupHelper object:\nReason: $err_msg");
      return $TCD::FAIL;
   }
   #
   # set the debug level to print debug message at particular level
   #
   $igrp_helper_obj->set_debug_level($debug_level);

=back

=head2 B<create_igroup_in_vserver(%args)>

Create a igroup in a given vserver and return the hash reference to expected igroup attributes
for later verification.

=over 4

=item Inputs

=item * I<-vserver_obj> <required> 

Reference to vserver object, where igroup will be created.

=item * I<-igroup> <optional>                    

Name of the igroup to create, igroup is must parameter. Since this function can be used
      for negative test case, code would not check for mandatory parameter

=item * I<-protocol_type> <optional> <fcp | iscsi | mixed>

Name of the igroup protocol type. Can be one of these(fcp,iscsi or mixed).
If do not specify it will create "mixed" protocol-type igroup

=item * I<-ostype> <optional> 

Name of the ostype that igroup belongs to. Check San::Application::Igroupconstant OS_TYPE_SUPPORTED definition for the valid ostypes
or check the Build that this script is running for valid ostype.


=item * I<-initiator> <optional> 

Name of the initiator to add it to igroup. Initiator can be iscsi iqn or fcp wwpn or alias of it.

=item * I<-portset> <optional> 

Name of the portset to bind the igroup while creating the igroup

=item Outputs

Returns a reference to hash contains igroup expected attributes info if successful.
If an error occurred the error flag is set, errmsg is set to the failure reason, and {} (empty hash reference)
is passed back as the return value.

Expected attributes to simulate what controller return for igroup attribute.
igroup show command will have following attributes for given igroup.
    vserver|igroup|protocol-type|ostype|portset|initiator|uuid|throttle-resv|throttle-borrow|alua|

if any attribute is not passed fill with "-" as default. Since igroup show output will have "-" , except 
protocol-type = mixed, throttle-resv = 0, throttle-borrow = false, alua = false

NOTE; UUID will be filled after successful igroup creation. 
      in addition to above igroup attributes, it will also store the igroup object for later needs, as part of expected attributes

Sample output from controller.

    igrp_vs|igrp_solaris_default|mixed|solaris|-|-|af3fbed0-0491-11df-a7c3-123478563412|0|false|false|
     

=item Examples

    #
    # Instantiate IgroupHelper Object (see above _new() pod)
    # or use IgroupMgr Object as below
    #
    
    $IgrpMgr_obj = new IgroupMgr();
    if ($IgrpMgr_obj->error) {
     my $err_msg = $IgrpMgr_obj->errmsg();
     $Test->nlog("Failed to instantiate IgroupMgr object:\nReason: $err_msg");
     return $TCD::FAIL;
    }
    #
    # Call setup to create environment to testing
    #
    $IgrpMgr_obj->setup(...);
    if ($IgrpMgr_obj->error) {
     $Test->nlog($IgrpMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    # get vserver object
    $vs_obj = ...;
    
    
    #
    # create default(mixed) windows igroup
    #
    my $igrp_attr_exp = $IgrpMgr_obj->create_igroup_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-igroup'       => "igrp_default,
                                            '-ostype'       => "windows"); 
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to create default igroup: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    # verify igroup created as expected
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    #
    # create default(without initiator) fcp igroup
    #
    my $igrp_attr_exp = $IgrpMgr_obj->create_igroup_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-igroup'       => "igrp_default,
                                            '-ostype'       => "windows",
                                            '-protocol_type' => "fcp"); 
    if ($IgrpMgr_obj->error()) {
     $Test->nlog("Failed to create default fcp igroup: " . $IgrpMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    # verify igroup created as expected
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    
    #
    # create window fcp igroup with initiator
    #
    my $igrp_attr_exp_1 = $IgrpMgr_obj->create_igroup_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-igroup'       => "igrp_default,
                                            '-ostype'       => "windows",
                                            '-protocol_type' => "fcp",
                                            '-initiator'    => "50:10:10:aa:00:00:00::01"); 
    if ($IgrpMgr_obj->error()) {
     $Test->nlog("Failed to create window fcp igroup with initiator: " . $IgrpMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    # verify igroup created as expected
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp_1);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ....
    ....

=back

=head2 B<add_initiators_to_igroup(%args)>

To add fcp(wwpn) or iscsi(iqn) initiator(s) to existing igroup and update the expected igroup attributes for verification if initiator(s) added successfully

=over 4

=item Inputs

=item * I<-igrp_attr_info> <required>

Hash reference to igroup info, basically expects same info which is return from
the function 'create_igroup_in_vserver'. See create_igroup_in_vserver POD for more details.
This hash reference will be updated upon successful addition of initiator(s) to igroup.

=item * I<-initiator> <required> 

Name of the initiator to add it to igroup. Initiator can be iscsi iqn or fcp wwpn or alias of it.
One or more initiators can be specified. If more than one initiators, should specify with comma separated as one string.

=item Outputs

Returns TRUE  if initiator(s) added to igroup successful. If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    # Instantiate IgroupHelper Object (see above _new() pod)
    # or use IgroupMgr Object (see above create_igroup_in_vserver() pod)
    ....
    ....
    # get vserver object
    my $vs_obj = ...;
    
    # get fcp initiators which are logged in
    my $fcp_initiators = ...;
        
    # create igroup with mixed protocol without initiators
    my $igrp_attr_exp = $IgrpMgr_obj->create_igroup_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-igroup'       => "igrp_default,
                                            '-ostype'       => "windows"); 
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to create default igroup: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    # verify igroup created as expected
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ...
    
    # now add initiators to fcp igroup
    $IgrpMgr_obj->add_initiators_to_igroup('-igrp_attr_info'      => $igrp_attr_exp,
                                           '-initiator'          => $fcp_initiators);
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ...
    ...

=back

=head2 B<remove_initiators_from_igroup(%args)>

To remove fcp(wwpn) or iscsi(iqn) initiator(s) from existing igroup and update the expected igroup attributes for verification if initiator(s) removed successfully

=over 4

=item Inputs

=item * I<-igrp_attr_info> <required>

Hash reference to igroup info, basically expects same info which is return from
the function 'create_igroup_in_vserver'. See create_igroup_in_vserver POD for more details.
This hash reference will be updated upon successful removal of initiator(s) from igroup.

=item * I<-initiator> <required> 

Name of the initiator to remove from igroup. Initiator can be iscsi iqn or fcp wwpn or alias of it.
One or more initiators can be specified. If more than one initiators, should specify with comma separated as one string.

=item * I<-force> <optional> (true|false: default: false)

To indicate to use force flag for removal. It will be useful when you want to remove initiators when it is map to lun

=item Outputs

Returns TRUE  if initiator(s) removed from igroup successful. If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    # Instantiate IgroupHelper Object (see above _new() pod)
    # or use IgroupMgr Object (see above create_igroup_in_vserver() pod)
    ....
    ....
    # get vserver object
    my $vs_obj = ...;
    
    # get fcp initiators which are logged in
    my $fcp_initiators = ...;
        
    # create igroup with mixed protocol with initiators
    my $igrp_attr_exp = $IgrpMgr_obj->create_igroup_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-igroup'       => "igrp_default,
                                            '-ostype'       => "windows",
                                            '-initiator'    => $fcp_initiators); 
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to create igroup: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    # verify igroup created as expected
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ...
    
    # now remove initiators from igroup
    $IgrpMgr_obj->remove_initiators_from_igroup('-igrp_attr_info'      => $igrp_attr_exp,
                                                '-initiator'          => "50:10:10:aa:00:00:00::01");
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ...
    ...

=back

=head2 B<rename_igroup_in_vserver(%args)>

To rename the existing igroup on given vserver and update the expected igroup attributes

=over 4

=item Inputs

=item * I<-igrp_attr_info> <required>

Hash reference to igroup info, basically expects same info which is return from
the function 'create_igroup_in_vserver'. See create_igroup_in_vserver POD for more details.
This hash reference will be updated upon successful removal of initiator(s) from igroup.

=item * I<-new_name> <required> 

New igroup name to be renamed with the existing igroup.

=item Outputs

Returns TRUE  if igroup renamed successful. 
If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    # Instantiate IgroupHelper Object (see above _new() pod)
    # or use IgroupMgr Object (see above create_igroup_in_vserver() pod)
    ....
    ....
    # get vserver object
    my $vs_obj = ...;
    
    # get fcp initiators which are logged in
    my $fcp_initiators = ...;
        
    # create igroup with mixed protocol with initiators
    my $igrp_attr_exp = $IgrpMgr_obj->create_igroup_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-igroup'       => "igrp_default,
                                            '-ostype'       => "windows",
                                            '-initiator'    => $fcp_initiators); 
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to create igroup: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    # verify igroup created as expected
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ...
    
    # now rename the igroup with new name
    my $igrp_new_name = "igrp_renamed";
    
    $IgrpMgr_obj->rename_igroup_in_vserver('-igrp_attr_info'      => $igrp_attr_exp,
                                           '-new_name'          => $igrp_new_name);
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ...
    ...

=back

=head2 B<verify_igroup_attributes(%args)>

To verify the requested igroup attributes are intact with controller return attributes

NOTE:

    # verify the igroup attributes from user expected list 
    # with controller list and return one error message with  
    # list of attributes which is/are not matching. User can set 
    # _SKIP_CHECK_ value for particular key to skip checking.

=over 4

=item Inputs

=item * I<-igrp_attr_info> <required>

Hash reference to igroup info, basically expects same info which is return from
the function 'create_igroup_in_vserver'. See create_igroup_in_vserver POD for more details.
This hash reference will be updated upon successful removal of initiator(s) from igroup.

Caller can set value '_SKIP_CHECK_' if does not want to check particular attribute.
Ex: 
   $igrp_attr_expected->{'uuid'} = '_SKIP_CHECK_';

=item Outputs

Returns TRUE  if all expected igroup attributes matches, if successful. 
If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    # Instantiate IgroupHelper Object (see above _new() pod)
    # or use IgroupMgr Object (see above create_igroup_in_vserver() pod)
    ....
    ....
    # get vserver object
    my $vs_obj = ...;
    
    # get fcp initiators which are logged in
    my $fcp_initiators = ...;
        
    # create igroup with mixed protocol with initiators
    my $igrp_attr_exp = $IgrpMgr_obj->create_igroup_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-igroup'       => "igrp_default,
                                            '-ostype'       => "windows",
                                            '-initiator'    => $fcp_initiators); 
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to create igroup: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    
    # skip checking following attributes
    my $igrp_attr_exp_verify =  $igrp_attr_exp;
    $igrp_attr_exp_verify->{'uuid'} = '_SKIP_CHECK_';
    $igrp_attr_exp_verify->{'throttle_borrow'} = '_SKIP_CHECK_';
    
    # verify igroup created as expected
    $IgrpMgr_obj->verify_igroup_attributes('-igrp_attr_info' => $igrp_attr_exp_verify);
    if ($IgrpMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }

=back

=head2 B<verify_igroup_unique_uuid(%args)>

To verify the igroup uuid uniqueness. Verifies if array has igroup info which are created across vserver.

=over 4

=item Inputs

=item * I<-igrp_array_ref> <required>

Array reference to igroup expected info hash.

=item Outputs

Returns TRUE  if all UUID are unique, if successful. 
If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    # Instantiate IgroupHelper Object (see above _new() pod)
    # or use IgroupMgr Object (see above create_igroup_in_vserver() pod)
    ....
    ....
    # get all the vserver objects
    my @Vserver_objs = ...;
    
    my @igrp_objs_array = ();
    my $NoOfIgroups = 50;

    foreach my $vs_obj (@Vserver_objs) {
        foreach my $os_type (@SupportedOS) {
            # create requested number of igroups
            for (my $i = 1; $i <= $NoOfIgroups; $i++) {
                my $igrp_name = "igrp_" . $os_type . "_fcp_" . $i;
                my $igrp_attr_expected =
                  $IgrpMgr_obj->create_igroup_in_vserver('-vserver_obj' => $vs_obj,
                                                         '-igroup'      => $igrp_name,
                                                         '-protocol_type' => "fcp",
                                                         '-ostype'        => $os_type);

                if ($IgrpMgr_obj->error()) {
                    # destroy igroup it would have created in a loop
                    $Test->nlog($IgrpMgr_obj->errmsg());
                    return $TCD::FAIL;
                }
                push(@igrp_objs_array, $igrp_attr_expected);
            }    # end of for (my $i=1; $i <= $NoOfIgroups; $i++)
        }    # end of foreach my $os_type (@SupportedOS)
    }    # end of foreach my $vs_obj (@Vserver_objs)
    
    # verify the igroup uuid uniqueness across vservers
    $IgrpMgr_obj->verify_igroup_unique_uuid('-igrp_array_ref' => \@igrp_info_objs);
    if ($IgrpMgr_obj->error()) {
        # destroy all igroups
        $Test->nlog($IgrpMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ...
    ...
    # destroy all igroups
   
=back

=cut

