#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/San/Task/IscsiHelper.pm#1 $

# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
#
## @summary Just a bunch of helper methods for testing 7-mode iscsi functionality
## @author muralikrishna.attili@netapp.com
## @status public
## @pod here.

package San::Task::IscsiHelper;

use strict;
use Exporter;
use Error;
use San::SanLog;
use San::SanUtil;
use San::SanClient::ClientUtil;

use vars qw( @EXPORT @ISA );

@ISA = qw( Error SanLog SanUtil );
###############################################################################
# P U B L I C   M E T H O D S
###############################################################################

###############################################################################
# Method:     new
# Objective:  Creates new IscsiHelper object.
# Parameters: None.
# Returns:    A reference to the object.
###############################################################################
sub new {
    my ($class, %args) = @_;
    my $self = {};

    # Creating IscsiHelper object.
    bless $self, $class;

    my $debug_level = 1;

    # Validate parameters
    foreach (keys %args) {
        if (/-?debug_level$/) {
            if (not length($debug_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-debug_level'");
                return $self;
            }
        }
    }
    $self->set_debug_level($debug_level);

    return $self;
}    # end of IscsiHelper object creation

###############################################################################
# Method:     reset_isns_config
# Objective:  Used at the start of each test case to make sure the filer has
#	      de-registered from the isns server. Also logs the intiator out of
#	      all targets.
# Parameters: Array of Initiator Object (optional)
#	      Filer Object (required)
#	      iSNS Server Object (required)
# Returns:    $SanUtil::TRUE = Filer not registered on iSNS and initiator has
#                              been logged out of everything.
#	      $SanUtil::FALSE and set_err if Something failed.
# Note:       Set iscsi.isns.rev option with value 22.
#	      options_set (iscsi.isns.rev -> 22)
###############################################################################
sub reset_isns_config {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($filer_obj, $isns_server_obj, @initiators);
    my $iscsi_isns_rev_value = 22;    # set iscsi.isns.rev value to 22

    $self->sanlog(-msg => "Start reset_isns_config");

    # Parse arguments
    foreach (keys %args) {
        if (/-?filer_obj/) {
            $filer_obj = $args{$_};
            if (not length($filer_obj)) {
                $self->_set_err("Invalid value for argument '-filer_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?isns_server_obj/) {
            $isns_server_obj = $args{$_};
            if (not length($isns_server_obj)) {
                $self->_set_err(
                               "Invalid value for argument '-isns_server_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?initiators/) {
            if (not scalar(@initiators = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-initiators'");
                return $SanUtil::FALSE;
            }
        }

    }    # end of parsing arguments

    # check for existance of required arguments
    unless (ref($filer_obj) =~ /^San::SanFiler7M$/i) {
        $self->_set_err("Did not pass in a SanFiler object for '-filer_obj'");
        return $SanUtil::FALSE;
    }
    unless (ref($isns_server_obj) =~ /IsnsServer/i) {
        $self->_set_err(
                 "Did not pass in an IsnsServer object for '-isns_server_obj'");
        return $SanUtil::FALSE;
    }

    # Configure iscsi.isns.rev option to default value 22
    $self->sanlog(-msg       => "Configure iscsi isns rev with value 22",
                  -msg_level => 3);
    $filer_obj->options_set(-option     => "iscsi.isns.rev",
                            -value      => $iscsi_isns_rev_value,
                            -cmd_vfiler => "vfiler0");
    if ($filer_obj->error) {
        $self->_set_err("iscsi.isns.rev option_set failed",
                        $filer_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Start the iscsi service
    $self->sanlog(-msg => "Start iscsi service if the service is not running",
                  -msg_level => 3);
    my $iSCSI_status = $filer_obj->get_iscsi_status();
    if ($filer_obj->error()) {
        $self->_set_err("iscsi get_iscsi_status failed", $filer_obj->errmsg());
        return $SanUtil::FALSE;
    }
    if ($iSCSI_status =~ /iSCSI service is not running|false/i) {
        $filer_obj->iscsi_start();
        if ($filer_obj->error) {
            $self->_set_err("iscsi start failed", $filer_obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # Stop the iscsi isns service
    $self->sanlog(-msg       => "Stop the iscsi isns service",
                  -msg_level => 3);
    $filer_obj->iscsi_isns_stop();
    if ($filer_obj->error) {
        $self->_set_err("iscsi isns stop failed", $filer_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify filer has de-registered from isns server
    $self->sanlog(-msg => "Verify filer has de-registered from isns server",
                  -msg_level => 3);
    $isns_server_obj->verify_isns_deregistration(-filer_obj => $filer_obj);
    if ($isns_server_obj->error) {
        $self->_set_err("Verify filer de-registration is failed",
                        $isns_server_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # For each initiator logout all targets
    $self->sanlog(-msg => "Logout iSCSI initiators", -msg_level => 3);
    foreach my $init (@initiators) {

        # -clean_level=>2 argument logs out of all targets for each initiator
        $init->deconfigure(-clean_level => 2);
        if ($init->error) {
            $self->_set_err("Error:deconfigure failed.", $init->errmsg());
            return $SanUtil::FALSE;
        }
    }
    $self->sanlog(-msg => "reset_isns_config successful");
    return $SanUtil::TRUE;
} ## end sub reset_isns_config

################################################################################
# Subroutine Name: verify_initiator()
# Objective: Verifying initiator can log into target and see proper number of luns.
# Details: See POD documentation at the end of this file
################################################################################
sub verify_initiator {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($filer_obj, $init_obj, @interfaces, $no_of_luns, @target_objs,
        $found_target_obj, @no_of_devices);

    $self->sanlog(
         -msg => "starting verify_initiator: login and verify number of luns.");

    # Verifying arguments.
    foreach (keys %args) {
        if (/-?filer_obj/) {
            if (not length($filer_obj = $args{$_})) {
                $self->_set_err("Invalid value for -filer_obj: $args{$_}");
                return $SanUtil::FALSE;
            }
        } elsif (/-?initiator_obj/) {
            if (not length($init_obj = $args{$_})) {
                $self->_set_err("Invalid value for -initiator_obj: $args{$_}");
                return $SanUtil::FALSE;
            }
        } elsif (/-?interfaces/) {
            if (not scalar(@interfaces = @{$args{$_}})) {
                $self->_set_err("Invalid value for -interfaces: $args{$_}");
                return $SanUtil::FALSE;
            }
        } elsif (/-?luns/) {
            if (not length($no_of_luns = $args{$_})) {
                $self->_set_err("Invalid value for -luns: $args{$_}");
                return $SanUtil::FALSE;
            }
        }
    } ## end foreach (keys %args)

    # Verifying missing arguments.
    unless (ref($filer_obj) =~ /^San::SanFiler7M$/i) {
        $self->_set_err(
            "Invalid value passed in for -filer_obj, value needs to be a SanFiler Object"
        );
        return $SanUtil::FALSE;
    }
    unless (ref($init_obj) =~ /^San::SanClient/i) {
        $self->_set_err(
            "Invalid value passed in for -initiator_obj, value needs to be a SanClient Object"
        );
        return $SanUtil::FALSE;
    }
    unless (scalar(@interfaces)) {
        $self->_set_err("Missing required argument -interfaces.");
        return $SanUtil::FALSE;
    }
    unless (defined $no_of_luns) {
        $self->_set_err("Missing required argument -luns.");
        return $SanUtil::FALSE;
    }

    # Get iscsi target objects array.
    foreach my $int (@interfaces) {
        $found_target_obj = $filer_obj->get_iscsi_target($int);
        if ($filer_obj->error()) {
            $self->_set_err("Unable to get iscsi target object for $int",
                            $filer_obj->errmsg());
            return $SanUtil::FALSE;
        }
        push(@target_objs, $found_target_obj);
    }

    # Perform initiator login.
    $self->sanlog(-msg => "Performing initiator login.");
    $init_obj->configure(-targets => \@target_objs);
    if ($init_obj->error()) {
        $self->_set_err("Unable to login at the given iscsi targets.",
                        $init_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $self->sanlog(-msg => "Login to the given filer Success.", -msg_level => 3);

    # Get the devices(luns).
    @no_of_devices = $init_obj->get_devices(-refresh => 1);
    if ($init_obj->error()) {
        $self->_set_err("Unable to get the iscsi devices.",
                        $init_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $self->sanlog(-msg => "Number of Luns found: " . scalar(@no_of_devices),
                  -msg_level => 3);

    # Verify the correct number of luns were discovered
    if (scalar(@no_of_devices) != $no_of_luns) {
        $self->_set_err(  "Client "
                        . $init_obj->hostname()
                        . " sees "
                        . scalar(@no_of_devices)
                        . " luns."
                        . " Expected to see $no_of_luns.");
        return $SanUtil::FALSE;
    }

    # Client logout.
    $init_obj->deconfigure(-clean_level => 2);
    if ($init_obj->error()) {
        $self->_set_err("Error:deconfigure failed.", $init_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # If Login success and given number of luns found.
    $self->sanlog(-msg => "Login successful and correct number of luns found.");
    return $SanUtil::TRUE;
} ## end sub verify_initiator

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

IscsiHelper.pm

=head1 Author

Muralikrishna Attili

=head1 Description

IscsiHelper.pm is enclosed with a bunch of helper methods for testing 7-mode iscsi functionality

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error SanLog SanUtil>

=head1 Public Methods

=head2 B<new()>

Constructor for IscsiHelper class

=over 4

=item Inputs

=item * I<-debug_level> (optional)

debug_level value for sanlog messages.

default is -debug_level = 1

=item Outputs

IscsiHelper Object

=item Examples

 $iscsiHelper_obj = IscsiHelper->new();

=back

=head2 B<reset_isns_config()>

This function resets the isns configuration back to the starting state for each test case.
The function sets iscsi.isns.rev to 22, starts iscsi, stops isns, verifies the filer is no
longer registered with the isns server, and logs all initiators out of their targets.

=over 4

=item Inputs

=item * -filer_obj (required)

Filer Object

=item * -isns_server_obj (required)

iSNS Server Object

=item * I<-initiators> (optional)

Array of Initiator Objects

=item Outputs

$SanUtil::TRUE - All initiators have been logged out

$SanUtil::FALSE & set_err - something went wrong

=item Examples

 $result =
 $iscsihelperobj->reset_isns_config(-filer_obj=> $SanFilerObj,
                                    -isns_server_obj => $Isnsserverobj,
                                    -initiators =>\@initiator_array);#(optional param)

=back

=head2 B<verify_initiator(%args)>

Verifying initiator can log into target and see proper number of luns.

=over 4

=item Inputs

=item * -filer_obj (required)

The SanFiler7M object.

=item * -initiator_obj (required)

The Initiator object.(i.e SanClient Object)

=item * -interfaces (required)

A reference to an array of interface names on the target.

The client will attempt to login to all of these target interfaces.

=item * -luns (required)

The number of luns to be found.

=item Outputs

Returns $SanUtil::TRUE, if initiator logged in and correct number of luns found.

Returns $SanUtil::FALSE and set_err, if initiator did not log in, or incorrect number of luns found.
If an error occured the error flag and the errmsg is set.

=item Examples

 $result = $self->verify_initiator(-filer_obj     => $filer_obj,
                                   -initiator_obj => $init_obj,
                                   -interfaces    => \@interfaces,
                                   -luns          => $LUNS_COUNT);

=back
=cut
