# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Utility for testing lun clone functionality
## To provide library functions for the testing lun clone.
## All the cmode lun clone testing script will use this library
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::Task::LunCloneHelper;

# compiler directives
use strict;

# imports
use Params::Validate qw( :all );
use San::Application::Lunconstant;
use San::ControllerCluster::Filer10M::Vserver::Volume;
use San::SanUtil;
use San::Task::LunHelper;

our @ISA              = qw( SanUtil );
our $LUN_CLONE_PREFIX = 'cln_';

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:     See POD documentation at the end of this file for usage
###############################################################################
sub new {
    my ( $class, %args ) = @_;

    # bless instance into class
    my $self = {};
    bless $self, $class;

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   ('lun_helper_obj') => {
                    type     => OBJECT,
                    required => 1,
                    isa      => 'San::Task::LunHelper'
                }
            }
        );
    };

    if ($@) {
        $self->_set_err($@);
        return $self;
    }

    # Define instance variables
    $self->{'_lun_helper_obj'} = $args{'lun_helper_obj'};

    return $self;
}

################################################################################
# Method Name: clone_luns()
# Objective:   This function clones multiple full afs or snapshot luns to a
#              specified location (qtree or volume).  It returns an array of the
#              cloned lun attribute hashes, and the indices will correspond to
#              those of the array of parent lun attribute hashes.  By default,
#              it will attempt to cleanup the luns on the filer after all the
#              clones are complete.  In this case, it still returns the array
#              of lun attr hashes, but the lun_obj value will be stale.  If at any
#              point, there is an error in cloning or cleanup, an error is set
#              and an empty array is returned.
# Parameter:   $args{lun_attr_info} - lun_attr_info hash returned by LunHelper's
#                                     create_lun
#              $args{snapshot_obj}  - If present clone lun out of snapshot
#                                     instead of the afs.
#              $args{clone_prefix}  - The prefix of the cloned lun name
#              $args{qtree}         - The destination qtree of the cloned luns
#              $args{no_reserve}    - Space reservation attr of the cloned luns
#              $args{cleanup}       - Whether or not to delete the cloned luns
# Return:      $TCD::PASS if every lun clone success, $TCD::FAIL if not
################################################################################
sub clone_luns() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Initialize variables
    my ( @lun_attrs_array, @cloned_lun_attrs_array, $lun_obj );

    # Validate input parameter (required, correct types)
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'lun_attr_info' => {
                    type     => ARRAYREF,
                    required => 1
                },
                'snapshot_obj' => {
                    type     => OBJECT,
                    optional => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Snapshot'
                },
                'clone_prefix' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $LUN_CLONE_PREFIX
                },
                'qtree' => {
                    type     => SCALAR,
                    optional => 0,
                    default  => $LUN_ATTR_ROOT_QTREE
                },
                'no_reserve' => {
                    type     => SCALAR | UNDEF,
                    optional => 1
                },
                'cleanup' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $SanUtil::TRUE
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return ();
    }

    foreach my $lun_attrs ( @{ $args{lun_attr_info} } ) {

        # Get the lun object out of the lun attributes hash.
        $lun_obj = $lun_attrs->{$LUN_ATTR_OBJ};

        # Clone lun with LunCloneHelper
        my $cloned_lun_attrs;

        if ( defined $args{snapshot_obj} ) {

            $self->sanlog( -msg =>  "# Cloning "
                    . $lun_obj->get_path()
                    . " from snapshot "
                    . $args{snapshot_obj}->get_name()
                    . " to qtree $args{qtree}" );
            $cloned_lun_attrs = $self->clone_lun_snap(
                'src_lun_attr_info' => $lun_attrs,
                'snapshot_obj'      => $args{snapshot_obj},
                'dest_lun'   => $args{clone_prefix} . $lun_obj->get_name(),
                'dest_qtree' => $args{qtree},
                'no_reserve' => $args{no_reserve}
            );
        } else {
            
            $self->sanlog( -msg => "# Cloning "
                    . $lun_obj->get_path()
                    . " to qtree $args{qtree}" );
            $cloned_lun_attrs = $self->clone_lun(
                'src_lun_obj' => $lun_obj,
                'dest_lun'    => $args{clone_prefix} . $lun_obj->get_name(),
                'dest_qtree'  => $args{qtree},
                'no_reserve'  => $args{no_reserve}
            );
        }
        if ( $self->error() ) {
            $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
            return ();
        }
                               
        push( @cloned_lun_attrs_array, $cloned_lun_attrs );
    }

  # If cleanup is set to true, remove the luns from the filer.
  # Note that we will still return the lun attribute hashes of the cloned luns
  # but the lun_obj key will no longer contain a valid lun_obj.
    if ( $args{'cleanup'} == $SanUtil::TRUE ) {
        my $is_passed = $SanUtil::TRUE;
        foreach my $cloned_lun_attrs (@cloned_lun_attrs_array) {
            my $lun_obj = $cloned_lun_attrs->{lun_obj};
            $lun_obj->delete();
            if ( $lun_obj->error() ) {
                $is_passed = $SanUtil::FALSE;
                $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
                $self->_set_err( $lun_obj->errmsg() );
            }
        }
        return () if $is_passed == $SanUtil::FALSE;
    }
    return @cloned_lun_attrs_array;
}

###############################################################################
# Method Name: create_nary_clone_hierarchy()
# Objective:   This function creates a clone hierarchy starting at a single
#              origin lun.  The depth signifies how deep the hierarchy can go.
#              For example, a depth of two will result in cloning of the
#              provided lun and a subsequent cloning of the cloned lun.
#              degree signifies the branching factor of each node.  For example,
#              if degree is 3, the provided lun will have 3 children.  If depth
#              is 2, each of the children will have three children as well.
# Parameter:   $args{lun_attr_info} - lun_attr_info hash returned by LunHelper's
#                                     create_lun
#              $args{snapshot_obj}  - If present clone lun out of snapshot
#                                     instead of the afs.
#              $args{depth}         - The depth is how deep the hierarchy will go.
#              $args{degree}    - The branching factor of each non-leaf node
# Return:      Array of all the cloned luns created in the hierarchy.
#              Note that the mapping to the parent lun is not preserved, which
#              would require a more complex tree or object structure.
###############################################################################
sub create_nary_clone_hierarchy() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Initialize variables
    my ( @lun_attrs_array, @all_cloned_lun_attrs );

    # Validate input parameter (required, correct types)
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'lun_attr_info' => {
                    type     => HASHREF,
                    required => 1
                },
                'snapshot_obj' => {
                    type     => OBJECT,
                    optional => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Snapshot'
                },
                'dest_qtree' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $LUN_ATTR_ROOT_QTREE
                },
                'depth' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => 1
                },
                'degree' => {
                    type     => SCALAR,
                    required => 1
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return ();
    }

    @lun_attrs_array = ( $args{'lun_attr_info'} );

    for ( my $i = 1; $i <= $args{'depth'}; $i++ ) {

        my @curr_cloned_lun_attrs = ();

        foreach my $lun_attrs (@lun_attrs_array) {

            for ( my $j = 0; $j < $args{'degree'}; $j++ ) {

                my $cnt
                    = scalar(@all_cloned_lun_attrs)
                    + scalar(@curr_cloned_lun_attrs)
                    + 1;

                my $cloned_lun_attrs;
                my $lun_obj = $lun_attrs->{lun_obj};

                if ( defined $args{snapshot_obj} && $i == 1 ) {

                    $cloned_lun_attrs = $self->clone_lun_snap(
                        'src_lun_attr_info' => $lun_attrs,
                        'snapshot_obj'      => $args{snapshot_obj},
                        'dest_lun'   => "c${cnt}_" . $lun_obj->get_name(),
                        'dest_qtree' => $args{'dest_qtree'}
                    );
                } else {

                    $cloned_lun_attrs = $self->clone_lun(
                        'src_lun_obj' => $lun_obj,
                        'dest_lun'    => "c${cnt}_" . $lun_obj->get_name(),
                        'dest_qtree'  => $args{'dest_qtree'}
                    );

                }
                if ( $self->error() ) {
                    $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
                    return ();
                }

                push( @curr_cloned_lun_attrs, $cloned_lun_attrs );
            }
        }

        @lun_attrs_array = @curr_cloned_lun_attrs;
        push( @all_cloned_lun_attrs, @curr_cloned_lun_attrs );
    }

    return @all_cloned_lun_attrs;
}

################################################################################
# Method Name: enable_sis_clone_sharing()
# Objective:   This subroutine enables block sharing after a sis clone on the
#              node that contains the volume object.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub enable_sis_clone_sharing() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'node_objs' => {
                    type     => ARRAYREF,
                    required => 1
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # Set the flag to enable sharing on the node that contains the volume
    $args{'sis_clone_sharing_flag'} = $SanUtil::FALSE;

    return $self->_set_sis_clone_sharing(%args);
}

################################################################################
# Method Name: disable_sis_clone_sharing()
# Objective:   This subroutine enables block sharing after a sis clone on the
#              node that contains the volume object.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub disable_sis_clone_sharing() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'node_objs' => {
                    type     => ARRAYREF,
                    required => 1
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # Set the flag to disable sharing on the node that contains the volume
    $args{'sis_clone_sharing_flag'} = $SanUtil::TRUE;

    return $self->_set_sis_clone_sharing(%args);
}

################################################################################
# Method Name: sub_clone_lun()
# Objective:   This subroutine does an afs sub clone from one lun to another, or
#              from part of one lun to a different part of the same lun.
#              It validates the user has passed in the required
#              parameters and returns {} since no new luns are created.
#              San::Task::RangeFactory can help in creation and formatting of
#              the range parameter.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub sub_clone_lun() {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'src_lun_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Lun'
                },
                'dest_lun_obj' => {
                    type     => OBJECT,
                    optional => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Lun'
                },
                'range' => {
                    type     => SCALAR,
                    required => 1
                },
                'verify' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $SanUtil::TRUE
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return {};
    }

    # Pass the original arguments to a helper method
    my $lun_attrs = $self->_clone(%args);

    # If the clone operation fails return {}, otherwise return %lun_attrs
    if ( $self->error() ) {
        $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
    }
    $self->exit( -msg_level => 5 );
    return $lun_attrs;
}

################################################################################
# Method Name: sub_clone_lun_snap()
# Objective:   This subroutine does a snap sub clone from one lun to another.
#              It validates the user has passed in the required
#              parameters and returns {} since no new luns are created.
#              San::Task::RangeFactory can help in creation and formatting of
#              the range parameter.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub sub_clone_lun_snap() {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'snapshot_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Snapshot'
                },
                'src_lun_attr_info' => {
                    type     => HASHREF,
                    required => 1
                },
                'dest_lun_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Lun'
                },
                'range' => {
                    type     => SCALAR,
                    required => 1
                },
                'verify' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $SanUtil::TRUE
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return {};
    }

    # Pass the original arguments to a helper method
    my $lun_attrs = $self->_clone(%args);

    # If the clone operation fails return {}, otherwise return %lun_attrs
    if ( $self->error() ) {
        $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
    }
    $self->exit( -msg_level => 5 );
    return $lun_attrs;
}

################################################################################
# Method Name: sub_clone_file_to_lun()
# Objective:   This subroutine does an afs sub clone from a file to a lun
#              object.  It validates the user has passed in the required
#              parameters and returns {} since no new luns are created.
#              San::Task::RangeFactory can help in creation and formatting of
#              the range parameter.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub sub_clone_file_to_lun() {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'src_file' => {
                    type     => SCALAR,
                    required => 1
                },
                'dest_lun_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Lun'
                },
                'range' => {
                    type     => SCALAR,
                    required => 1
                },
                'verify' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $SanUtil::TRUE
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return {};
    }

    # Pass the original arguments to a helper method
    my $lun_attrs = $self->_clone(%args);

    # If the clone operation fails return {}, otherwise return %lun_attrs
    if ( $self->error() ) {
        $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
    }
    $self->exit( -msg_level => 5 );
    return $lun_attrs;
}

################################################################################
# Method Name: sub_clone_file_to_lun_snap()
# Objective:   This subroutine does a snap sub clone from a file to a lun
#              object.  It validates the user has passed in the required
#              parameters and returns {} since no new luns are created.
#              San::Task::RangeFactory can help in creation and formatting of
#              the range parameter.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub sub_clone_file_to_lun_snap() {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {  'snapshot_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Snapshot'
                },
                'src_file' => {
                    type     => SCALAR,
                    required => 1
                },
                'dest_lun_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Lun'
                },
                'range' => {
                    type     => SCALAR,
                    required => 1
                },
                'verify' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $SanUtil::TRUE
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return {};
    }

    # Pass the original arguments to a helper method
    my $lun_attrs = $self->_clone(%args);

    # If the clone operation fails return {}, otherwise return %lun_attrs
    if ( $self->error() ) {
        $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
    }
    $self->exit( -msg_level => 5 );
    return $lun_attrs;
}

################################################################################
# Method Name: sub_clone_lun_to_file()
# Objective:   This subroutine does an afs sub clone from a lun object to a
#              file.  It validates the user has passed in the required
#              parameters and returns {} since no new luns are created.
#              San::Task::RangeFactory can help in creation and formatting of
#              the range parameter.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub sub_clone_lun_to_file() {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'src_lun_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Lun'
                },
                'dest_file' => {
                    type     => SCALAR,
                    required => 1
                },
                'range' => {
                    type     => SCALAR,
                    required => 1
                },
                'verify' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $SanUtil::TRUE
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return {};
    }

    # Pass the original arguments to a helper method
    my $lun_attrs = $self->_clone(%args);

    # If the clone operation fails return {}, otherwise return %lun_attrs
    if ( $self->error() ) {
        $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
    }
    $self->exit( -msg_level => 5 );
    return $lun_attrs;
}

################################################################################
# Method Name: sub_clone_lun_to_file_snap()
# Objective:   This subroutine does a snapshot sub clone from a lun object to a
#              file.  It validates the user has passed in the required
#              parameters and returns {} since no new luns are created.
#              San::Task::RangeFactory can help in creation and formatting of
#              the range parameter.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub sub_clone_lun_to_file_snap() {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'snapshot_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Snapshot'
                },
                'src_lun_attr_info' => {
                    type     => HASHREF,
                    required => 1
                },
                'dest_file' => {
                    type     => SCALAR,
                    required => 1
                },
                'range' => {
                    type     => SCALAR,
                    required => 1
                },
                'verify' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $SanUtil::TRUE
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return {};
    }

    # Pass the original arguments to a helper method
    my $lun_attrs = $self->_clone(%args);

    # If the clone operation fails return {}, otherwise return %lun_attrs
    if ( $self->error() ) {
        $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
    }
    $self->exit( -msg_level => 5 );
    return $lun_attrs;
}

################################################################################
# Method Name: clone_lun()
# Objective:   This subroutine accepts a lun object, and a desired qtree and
#              lun name.  It validates the user has passed in the required
#              parameters for a full lun clone operation and returns an
#              attribute hash of the cloned lun.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub clone_lun() {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'src_lun_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Lun'
                },
                'dest_qtree' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $LUN_ATTR_ROOT_QTREE
                },
                'dest_lun'   => { type => SCALAR, optional => 0 },
                'no_reserve' => {
                    type     => SCALAR | UNDEF,
                    optional => 1
                },
                'verify' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $SanUtil::TRUE
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return {};
    }

    # Pass the original arguments to a helper method
    my $lun_attrs = $self->_clone(%args);

    # If the clone operation fails return {}, otherwise return %lun_attrs
    if ( $self->error() ) {
        $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
    }
    $self->exit( -msg_level => 5 );
    return $lun_attrs;
}

################################################################################
# Method Name: clone_lun_snap()
# Objective:   Creates a full clone of a lun in a snapshot and returns the
#              hash reference of attributes that the clone is expected to have.
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub clone_lun_snap() {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameter (required, correct types) & set defaults
    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'src_lun_attr_info' => {
                    type     => HASHREF,
                    required => 1
                },
                'snapshot_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Snapshot'
                },
                'dest_qtree' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $LUN_ATTR_ROOT_QTREE
                },
                'dest_lun'   => { type => SCALAR, optional => 0 },
                'no_reserve' => {
                    type     => SCALAR | UNDEF,
                    optional => 1
                },
                'verify' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $SanUtil::TRUE
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return {};
    }

    # Pass the original arguments to a helper method
    my $lun_attrs = $self->_clone(%args);

    # If the clone operation fails return {}, otherwise return %lun_attrs
    if ( $self->error() ) {
        $self->sanlog( -msg => $self->errmsg(), -msg_level => 1 );
    }
    $self->exit( -msg_level => 5 );
    return $lun_attrs;
}

###############################################################################
#                        Private Methods                                      #
###############################################################################

################################################################################
# Method Name: _clone
# Objective:   This subroutine contains common code to handle a request to clone
#              a lun. It contains the logic that determines when to discover
#              and return a cloned lun.  It also determines what luns should be
#              validated.
# Parameter:   $args{'args'} - hash ref to the original arguments of
#                the public api method
# Return:      hash ref to expected attrs array of cloned lun
#              {} if subclone operation
# Details:     POD is not required for local(internal) function.
################################################################################
sub _clone() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();

    # Initialize Variables
    my (%mapped_params,      $lun_helper_obj,    $src_vol_obj,
        $dest_lun_attr_info, $src_lun_attr_info, $cloned_lun_attr_info
    );

    # Get a shorter reference to the LunHelper
    $lun_helper_obj = $self->{_lun_helper_obj};

    # Map all the original public parameters to the parameters that
    # are needed to execute the volume_file_clone_create cdef.
    %mapped_params = %{ $self->_map_params(%args) };
    $src_vol_obj   = $mapped_params{'src_vol_obj'};

    # Perform a clone of the lun object with the supplied parameters
    $self->_file_clone_create(
        src_vol_obj => $src_vol_obj,
        cmd_params  => $mapped_params{'cmd_params'}
    );
    return {} if $self->error();

    # If the src/dest luns already exist, get their attrs before
    # the clone, so that we can verify that they don't change after the clone.
    if ( exists $args{'snapshot_obj'} ) {
        $src_lun_attr_info = $args{'src_lun_attr_info'};
    } elsif ( exists $args{'src_lun_obj'} ) {
        $src_lun_attr_info = $lun_helper_obj->build_lun_attributes(
            -lun_obj => $args{'src_lun_obj'} );
        if ( $lun_helper_obj->error() ) {
            $self->_set_err( $lun_helper_obj->errmsg() );
            return {};
        }
    }
    if ( exists $args{'dest_lun_obj'} ) {
        $dest_lun_attr_info = $lun_helper_obj->build_lun_attributes(
            lun_obj => $args{'dest_lun_obj'} );
        if ( $lun_helper_obj->error() ) {
            $self->_set_err( $lun_helper_obj->errmsg() );
            return {};
        }
    }

    # If the clone operation creates a new lun, discover & validate attrs
    if ( exists $args{'dest_lun'} ) {

        # file_clone_create doesn't return a lun, so discover it
        my $cloned_lun_obj = $self->_discover_lun(
            src_vol_obj => $src_vol_obj,
            dest_lun    => $args{'dest_lun'},
            dest_qtree  => $args{'dest_qtree'},
        );
        return $cloned_lun_obj if $self->error();

        my $dest_path = '/vol/' . $src_vol_obj->get_name();
        if ($args{'dest_qtree'} ne '""' ) {
            $dest_path .= '/' . $args{'dest_qtree'};
        } 
        $dest_path .= '/' . $args{'dest_lun'};
        
        # Build the expected attributes of the cloned lun based on the
        # attributes of the source lun and the cli arguments
        $cloned_lun_attr_info = $self->_build_cloned_lun_attrs_expected(
            lun_attr_info  => $src_lun_attr_info,
            cloned_lun_obj => $cloned_lun_obj,
            dest_lun       => $args{'dest_lun'},
            dest_qtree     => $args{'dest_qtree'},
            dest_path      => $dest_path,
            no_reserve     => $args{'no_reserve'}
        );

        # Validate the cloned lun if verify is set to true
        if ( $SanUtil::TRUE == $args{'verify'} ) {

            # Verify the cloned lun
            $lun_helper_obj->verify_lun_attributes(
                -lun_attr_info => $cloned_lun_attr_info );
            if ( $lun_helper_obj->error() ) {
                $self->_set_err( $lun_helper_obj->errmsg() );
                return {};
            }
        }
    }

    # After the clone is over, verify that none of the attributes in
    # the src lun changed as a result of the clone.  If it was a subclone to
    # another lun, verify that none of the attributes in the destination lun
    # changed as a result of the clone operation.
    if ( $SanUtil::TRUE == $args{'verify'} ) {

        if ( exists $args{'src_lun_obj'} ) {

            $src_lun_attr_info = $lun_helper_obj->verify_lun_attributes(
                lun_attr_info => $src_lun_attr_info );
            if ( $lun_helper_obj->error() ) {
                $self->_set_err( $lun_helper_obj->errmsg() );
                return {};
            }
        }

        if ( exists $args{'dest_lun_obj'} ) {

            $dest_lun_attr_info = $lun_helper_obj->verify_lun_attributes(
                lun_attr_info => $dest_lun_attr_info );
            if ( $lun_helper_obj->error() ) {
                $self->_set_err( $lun_helper_obj->errmsg() );
                return {};
            }
        }
    }

    # If a new lun was created return it, otherwise return {}
    return ( exists $args{'dest_lun'} ) ? $cloned_lun_attr_info : {};
}

################################################################################
# Method Name: _map_params
# Objective:   This subroutine maps the parameters accepted by the public API
#              to those of the file_clone_create cdef.
# Parameter:   $args{'args'} - hash ref to the original arguments of
#                the public api method
# Return:      hash reference containing:
#                src_vol_obj - the source volume object.
#                cmd_params  - hash reference containing cdef params
# Details:     POD is not required for local(internal) function.
################################################################################
sub _map_params() {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();

    # Initialize Variables
    my ($src_vol_obj, $is_snapshot, $is_sub_clone, $src_path,
        $dest_path,   $snapshot,    %cmd_params
    );

    $is_snapshot  = exists $args{'snapshot_obj'};
    $is_sub_clone = exists $args{'range'};

    # If the source resides on the snapshot and not the afs
    if ($is_snapshot) {

        # Get the snapshot name and volume object
        $src_vol_obj = $args{'snapshot_obj'}->get_parent_obj();
        $snapshot    = $args{'snapshot_obj'}->get_name();

        # Get the src_path from either src_lun_attr_info or src_file
        if ( exists $args{'src_lun_attr_info'} ) {
            $args{'src_lun_attr_info'}{$LUN_ATTR_PATH}
                =~ '^/vol/\S+?/(\S+?)$';
            $src_path = $1;
        } else {
            $src_path = $args{'src_file'};
        }

    } else {

        # It's an afs clone, the source has to be a lun or a file
        if ( exists $args{'src_lun_obj'} ) {

            # Get the parent object of the lun
            my $parent_obj = $args{'src_lun_obj'}->get_parent_obj();

            # Get the volume object of the lun
            if ( ref($parent_obj) =~ /Qtree/i ) {
                $src_vol_obj = $parent_obj->get_parent_obj();
            } else {
                $src_vol_obj = $parent_obj;
            }
            $args{'src_lun_obj'}->get_path() =~ '^/vol/\S+?/(\S+?)$';
            $src_path = $1;
        } else {

            # Get the parent object of the destination lun
            my $parent_obj = $args{'dest_lun_obj'}->get_parent_obj();

            # Get the volume object of the lun
            if ( ref($parent_obj) =~ /Qtree/i ) {
                $src_vol_obj = $parent_obj->get_parent_obj();
            } else {
                $src_vol_obj = $parent_obj;
            }
            $src_path = $args{'src_file'};
        }
    }

    # Figure out the destination path
    if ($is_sub_clone) {

        # The destination path could be dest_lun_obj or dest_file.
        if ( exists $args{'dest_lun_obj'} ) {
            $args{'dest_lun_obj'}->get_path() =~ '^/vol/\S+?/(\S+?)$';
            $dest_path = $1;

        } else {
            $dest_path = $args{'dest_file'};
        }
    } else {

        # Full clone only supports lun to lun so dest_path is always lun obj
        $dest_path
            = $args{'dest_qtree'} eq $LUN_ATTR_ROOT_QTREE
            ? $args{'dest_lun'}
            : $args{'dest_qtree'} . '/' . $args{'dest_lun'};
    }

    # Build the hash of command parameters
    %cmd_params = ( '-source-path' => $src_path );
    if ( defined $dest_path ) {
        $cmd_params{'-destination-path'} = $dest_path;
    }
    if ($is_snapshot) {
        $cmd_params{'-snapshot-name'} = $snapshot;
    }
    if ($is_sub_clone) {
        $cmd_params{'-range'} = $args{'range'};
    }
    if ( defined $args{'no_reserve'} ) {
        $cmd_params{'-no-reserve'} = $args{'no_reserve'};
    }

    # Return a hash ref containing the source_vol_obj and the cmd params.
    return {
        src_vol_obj => $src_vol_obj,
        cmd_params  => \%cmd_params
    };
}

################################################################################
# Method Name: _set_sis_clone_sharing
# Objective:   This subroutine sets the setflag wafl_disable_sis_clone_fixup
#              on the node that contains the provided volume.
# Parameter:   $args{'vol_obj'} - volume object
# Parameter:   $args{'sis_clone_sharing_flag'} - value of the flag (0 or 1)
# Return:      $SanUtil::TRUE or $SanUtil::FALSE
# Details:     POD is not required for local(internal) function.
################################################################################
sub _set_sis_clone_sharing() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    my @node_objs = @{ $args{'node_objs'} };
    my $sis_clone_sharing_flag      = $args{'sis_clone_sharing_flag'};

    foreach my $node_obj (@node_objs) {

        my $node_name = $node_obj->get_name();
        if ( $node_obj->error() ) {
            $self->_set_err( $node_obj->errmsg() );
            return $SanUtil::FALSE;
        }

        my %cmd = (
            cmd     => "system_node_run",
            node    => $node_name,
            command => "setflag wafl_disable_sis_clone_fixup $sis_clone_sharing_flag",
            donot_append_vserver => 1
        );

        $node_obj->invoke_cmd_obj(%cmd);
        if ( $node_obj->error() ) {
            $self->_set_err(
                "failed to invoke_cmd_obj "
                    . $node_obj->dumper_get( \%cmd,
                    $self->get_debug_level() ),
                $node_obj->errmsg()
            );
            return $SanUtil::FALSE;
        }
    }

    #TODO Verify that this was actually set by querying the controller
    #Return error if not set correctly.
    return $SanUtil::TRUE;
}

################################################################################
# Method Name: _file_clone_create()
# Objective:   This subroutine calls uses ControllerCluster to clone a lun
#              on the filer and validates whether the clone operation completes
#              without error.
# Parameter:   $args{src_vol_obj} - src volume that contains the lun to clone
#              $args{cmd_params}  - hash ref of params to file_clone_create
# Return:
# Details:     POD is not required for local(internal) function.
################################################################################
sub _file_clone_create {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();

    # Initialize Variables
    my $src_vol_obj;

    # Call file_clone_create with the command parameters
    $src_vol_obj = $args{src_vol_obj};
    $src_vol_obj->file_clone_create( %{ $args{cmd_params} } );
    if ( $src_vol_obj->error() ) {
        $self->_set_err( $src_vol_obj->errmsg() );
        return $SanUtil::FALSE;
    }
    
    # TODO
    # For the most part, the cloned lun can be discovered immediately after
    # calling file_clone_create.  A client can perform IO at this time, but
    # it doesn't mean the sharing process of the clone operation is complete.
    # To find that out, you have to poll for the clone status, which could
    # be implemented here as an option.  ie.  if the is_wait is set to 1,
    # don't return until the status command shows that the clone is complete.
    return $SanUtil::TRUE;
}

################################################################################
# Method Name: _discover_lun()
# Objective:   This operation discovers the cloned lun.  Discovery is necessary
#              because the file_clone_create invokes the clone operation, but
#              it doesn't return a cloned lun.
# Parameter:   $args{'src_vol_obj'} - the src vol obj that contains the lun
#              $args{'dest_lun'}    - the name of the lun
#              $args{'dest_qtree'}  - the qtree of the lun
# Return:      San::Task::ControllerCluster::Vserver::Volume::Lun
# Details:     POD is not required for local(internal) function.
# TODO:        Decide if this is the appropriate place for this function.
#              There might be some LunHelper code that can be tapped into to
#              do the same thing.
################################################################################
sub _discover_lun {
    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();

    # Initialize Variables
    my ( $src_vol_obj, $cloned_lun_obj, $container_obj );
    $src_vol_obj = $args{src_vol_obj};

    # Determine if the container object is a qtree or volume
    if ( $args{'dest_qtree'} eq $LUN_ATTR_ROOT_QTREE ) {
        $container_obj = $src_vol_obj;
    } else {

        # Get the destination qtree object
        $container_obj
            = $src_vol_obj->get_qtree_obj( qtree => $args{'dest_qtree'} );
        if ( $src_vol_obj->error() ) {
            $self->_set_err(
                "Failed to find qtree \nReason:\n" . $src_vol_obj->errmsg() );
            return {};
        }
    }

    # Get the cloned lun that has just been created.
    $cloned_lun_obj = $container_obj->get_lun( lun => $args{'dest_lun'} );
    if ( $container_obj->error() ) {
        $self->_set_err( "Failed to find cloned lun \nReason:\n"
                . $container_obj->errmsg() );
    }

    # Return the cloned lun or {} if there was an error
    return $cloned_lun_obj;
}

################################################################################
# Method Name: _build_cloned_lun_attrs_expected()
# Objective:   This operation builds a hash of the attributes that a cloned
#              lun should have when you query the controller.
# Parameter:   $args{'lun_attr_info'} - hash ref to the attributes of the src
#                lun.
#              $args{'cloned_lun_obj'} - the cloned lun object
#              $args{'dest_lun'}       - the expected name of the cloned lun
#              $args{'dest_qtree'}     - the expected qtree of the cloned lun
#              $args{'no_reserve'}     - the expected space_resv setting of
#                                        the cloned lun
# Return:      hash reference to the hash of expected attrs of the cloned lun
# Details:     POD is not required for local(internal) function.
################################################################################
sub _build_cloned_lun_attrs_expected() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();

    # Initialize variables
    my %src_lun_attrs  = %{ $args{'lun_attr_info'} };
    my $cloned_lun_obj = $args{'cloned_lun_obj'};

    # Get list of attributes that should be inherited from the source lun.
    my @src_lun_attrs_keys = @LunAttrsInherited;

    # Slice the hash of attr/values to get just those that should be inherited
    my %inherited_attrs
        = map { $_ => $src_lun_attrs{$_} } @src_lun_attrs_keys;

    # Merge the inherited attr/values and the default noninherited attr/values
    my %lun_attrs = ( %inherited_attrs, %LunAttrsNonInheritedDefault );

    # The values of the lun & qtree attrs come from the method args.
    $lun_attrs{$LUN_ATTR_OBJ}   = $cloned_lun_obj;
    $lun_attrs{$LUN_ATTR_NAME}  = $args{'dest_lun'};
    $lun_attrs{$LUN_ATTR_QTREE} = $args{'dest_qtree'};
    $lun_attrs{$LUN_ATTR_PATH}  = $args{'dest_path'};
    
    # The value of the space_resv attr comes from the method arg
    # if no_reserve is true.
    if ( 'true' eq $args{'no_reserve'} ) {
        $lun_attrs{$LUN_ATTR_SPACE_RESV} = $LUN_ATTR_SPACE_RESV_DISABLE;
    }

    # TODO - LunHelper verify_lun_attrs is expected to be updated with
    # logic to determine whether lun ids are unique across the entire vserver.
    # Geometry information could be added to LunHelper as well.
    return \%lun_attrs;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

LunCloneHelper.pm

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description


LunCloneHelper.pm provides an utility function interface to test lun clone
functionality with the main intent to avoid repeated code from thpl level. 
It is stateless in that no object state is maintained between cloning method 
calls.  Each public method (with the help of private functions) will validate 
parameters, perform the intended cloning operation, validate that the basic 
functionality of the cloning operation succeeded, and return an attribute 
hash of the cloned lun if appropriate. 

The file_clone_create cdef performs multiple types of cloning operations with
a small number of parameters.

file clone create 
    -vserver <vserver>
    -volume <volume>
    -snapshot-name <snapshot>
    -source-path <source_path>
    -destination-path <destination_path>
    -range <comma delimmited list of tuplets (source_block:dest_block:num_blocks) >

LunCloneHelper is designed to wrap this command with self-documenting public
methods that tell the caller exactly what the operation is intended to do.  

High level description of public methods

clone_lun - Clone a full lun from the afs.

clone_lun_snap - Clone a full lun that exists in a snapshot.

unimplemented

Note - There might not be a public method for every single type of these 
operations.  Some may be combined.

Subclone a portion(s) of a lun in the afs to a different part of the same lun.

Subclone a portion(s) of a lun in the afs to a different lun.

Subclone a portion(s) of a lun in the afs to a different file.

Subclone a portion(s) of a file in the afs to a different lun.

Subclone a portion(s) of a lun in a snapshot to a different part of the same lun.

Subclone a portion(s) of a lun in a snapshot to a different lun.

Subclone a portion(s) of a lun in a snapshot to a different file.

Subclone a portion(s) of a file in a snapshot to a different lun.

High level description of parameters

-src_lun_obj

Cloning methods involving a source lun in the afs will always accept a src_lun_obj.

-snapshot_obj 
-src_lun_attr_info

Cloning methods involving a snapshot will always accept a hash called
src_lun_attr_info instead of accepting a src_lun_obj.  Because the source lun 
resides in the snapshot, there is no guarantee that it shares the attributes 
of the lun in the active file system.  There is also the possibility that the 
lun has been deleted.  In both of these scenarios, it would not be possible 
to accept a src_lun_obj.

-no_reserve

The no_reserve parameter is only specified if the caller explicitly wants the 
cloned lun to be spaced reserved or thin provisioned.  The caller should not specify
this attribute if he wants the cloned lun to inherit the attributes of the parent
lun. 

-verify

The caller can specify a verify parameter if validation should not be performed 
on the cloned lun.

-dest_lun
-dest_qtree
-dest_file

The format of the -destination-path of the file_clone_create cdef is encapsulated
by  parameters such as dest_lun, dest_qtree, and dest_file.
If the format of the destination-path changes in the cdef, the caller should
not have to change any code.   

=head1 See Also

Package relies heavily on LunHelper module. See these San::Task::LunHelper POD
for further interface detail.

B<San::Task::LunHelper>

=head1 Attributes

None

=head1 Revision History

Feb-02-2011 schroll@netapp.com: Added subclone functionality.

Nov-12-2010 schroll@netapp.com: Original template created with clone_lun and 
clone_lun_snap functionality.

=head1 Public Methods

=head2 B<new()>

Class Constructor method. To set the debug level to print debug message at 
particular level, call 'set_debug_level' method. See Example for more details

=over 4

=item Inputs

=item * I<-lun_helper_obj> <required>

Reference to a LunHelper object.

=item Outputs

Returns a LunCloneHelper blessed object reference.

=item Examples

   #
   # Create a LunCloneHelper object
   #
   my $lun_clone_helper = new LunCloneHelper();
   if ($lun_clone_helper->error()) {
      my $err_msg = $lun_clone_helper->errmsg();
      $Test->nlog("Failed to instantiate LunCloneHelper object:\nReason: $err_msg");
      return $TCD::FAIL;
   }
   #
   # set the debug level to print debug message at particular level
   #
   $lun_clone_helper->set_debug_level($debug_level);


=back

=head2 B<clone_lun(%args)>

Create a full clone of a lun in the active file system and return the hash reference
of attributes that the clone is expected to have.

=over 4

=item Inputs

=item * I<-src_lun_obj> <required>

The lun object in the active file system to clone.

=item * I<-dest_lun> <required>

The name of the cloned lun that will be created.

=item * I<-dest_qtree> <optional>

The name of the qtree on which to create the cloned lun.

=item * I<-no_reserve> <optional>

Do not specify this parameter if you want the cloned lun to inherit
the space reservation settings from the src_lun_obj.

'true' if the cloned lun should be created as a thin provisioned lun.
'false' if the cloned lun should be a space reserved lun.

=item Outputs

Returns a reference to a hash reference containing the expected attributes of 
the cloned lun if successful.  The cloned lun object can be retrieved by
accessing the hash with the 'lun_obj' key.

If an error occurred the error flag is set, errmsg is set to the failure reason,
and {} (empty hash reference) is passed back as the return value.

Expected attributes will simulate what the attributes of the cloned lun should 
be.  The names of the hash attributes are consistent with those documented in 
LunHelper.pm, create_lun.

=item Examples

    # Get lun object
    $lun_clone_helper_obj = San::Task::LunCloneHelper->new( lun_helper_obj => $LunMgrObj );

    # Get lun object
    $lun_obj = ...;

    # Clone lun with just the required parameters.
    $cloned_lun_attrs = $lun_clone_helper_obj->clone_lun(
        '-lun_obj' => $lun_obj,
        '-dest_lun' => 'clone_' . $lun_obj->get_name()
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }
    
    # Clone lun with all optional parameters specified.
    $cloned_lun_attrs = $lun_clone_helper_obj->clone_lun(
        '-lun_obj'    => $lun_obj,
        '-dest_lun'   => 'clone_' . $lun_obj->get_name(),
        '-dest_qtree' => 'qtree1',
        '-no_reserve' => 'true',
        '-verify'     => $SanUtil::TRUE
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }


=back

=head2 B<clone_lun_snap(%args)>

Creates a full clone of a lun in a snapshot and returns the hash reference
of attributes that the clone is expected to have.

=over 4

=item Inputs

=item * I<-src_lun_attr_info> <required>

The hash of attributes that the lun had prior to the snapshot being taken. 
This hash must be built from a lun object with LunHelper, build_lun_attributes.

=item * I<-snapshot_obj> <required>

The snapshot object in which the source lun resides.

=item * I<-dest_lun> <required>

The name of the cloned lun that will be created.

=item * I<-dest_qtree> <optional>

The name of the qtree on which to create the cloned lun.

=item * I<-no_reserve> <optional>

Do not specify this parameter if you want the cloned lun to inherit
the space reservation settings from the src_lun_obj.

'true' if the cloned lun should be created as a thin provisioned lun.
'false' if the cloned lun should be a space reserved lun.

=item Outputs

Returns a reference to a hash reference containing the expected attributes of 
the cloned lun if successful.  The cloned lun object can be retrieved by
accessing the hash with the 'lun_obj' key.

If an error occurred the error flag is set, errmsg is set to the failure reason,
and {} (empty hash reference) is passed back as the return value.

Expected attributes will simulate what the attributes of the cloned lun should 
be.  The names of the hash attributes are consistent with those documented in 
LunHelper.pm, create_lun.

=item Examples

    # Get LunCloneHelper object
    $lun_clone_helper_obj = San::Task::LunCloneHelper->new( lun_helper_obj => $LunMgrObj );

    # Get volume object
    $vol_obj = ...;
    
    my $lun_attr_info = $LunMgrObj->create_lun(
        'volume_obj' => $vol_obj,
        'size'   => '50M',
        'ostype' => 'linux',
        'lun_name' => 'lun'
    );
    if ( $LunMgrObj->error() ) {
        $Test->nlog( $LunMgrObj->errmsg() );
        return $TCD::FAIL;
    }

    my $snapshot_obj = $vol_obj->create_snapshot_obj( snapshot => 'snapshot' );
    if ( $snapshot_obj->error() ) {
        $Test->nlog( $snapshot_obj->errmsg() );
        return $TCD::FAIL;
    }

    # Clone lun with just the required parameters.
    $cloned_lun_attrs = $lun_clone_helper_obj->clone_lun_snap(
        '-snapshot_obj' => $snapshot_obj,
        '-src_lun_attr_info' => $lun_attr_info,
        '-dest_lun' => 'lun_clone'
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }
    
    # Clone lun with all the optional parameters.
    $cloned_lun_attrs = $lun_clone_helper_obj->clone_lun_snap(
        '-snapshot_obj'         => $snapshot_obj,
        '-src_lun_attr_info'    => $lun_attr_info,
        '-dest_lun'             => 'lun_clone'
        '-dest_qtree'           => 'qtree1',
        '-no_reserve'           => 'true',
        '-verify'               => $SanUtil::TRUE
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }    

=back

=head2 B<clone_luns(%args)>

Clones multiple luns to a single destination (volume or qtree).  Returns the 
cloned luns as an array of LunAttrInfos. 

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

Array where each index contains a different lun attribute hash to be cloned.

=item * I<-snapshot_obj> <optional>

The name of the cloned lun that will be created.

=item * I<-clone_prefix> <optional>

The prefix of the name of the clone.

=item * I<-qtree> <optional>

The name of the qtree on which to create the cloned lun.

=item * I<-no_reserve> <optional>

Do not specify this parameter if you want the cloned lun to inherit
the space reservation settings from the src_lun_obj.

'true' if the cloned lun should be created as a thin provisioned lun.
'false' if the cloned lun should be a space reserved lun.

=item * I<-cleanup> <optional>

1 if the cloned luns should be deleted before the method returns.
0 if the cloned luns should not be deleted before the method returns.

=item Outputs

Returns an array of hash references containing the expected attributes of 
the cloned luns if successful.  

If an error occurred the error flag is set, errmsg is set to the failure reason,
and {} (empty hash reference) is passed back as the return value.

Expected attributes will simulate what the attributes of the cloned lun should 
be.  The names of the hash attributes are consistent with those documented in 
LunHelper.pm, create_lun.

=item Examples

    # Get lun object
    $lun_clone_helper_obj = San::Task::LunCloneHelper->new( lun_helper_obj => $LunMgrObj );

    # Get array of lun attribute hashes
    $lun_attrs_array = ...;
    
    # Get qtree to clone luns to
    $qtree_obj = ...;

    # Clone lun with just the required parameters.
    @cloned_lun_attrs_array = $CloneHelperObj->clone_luns(
                -lun_attr_info => \@lun_attrs_array,
                -clone_prefix  => $LUN_CLONE_PREFIX,
                -qtree         => $qtree_obj->get_name()
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }
    
=back


=head2 B<sub_clone_lun(%args)>

Create a sub clone of a lun in the active file system.
If a src_lun_obj and dest_lun_obj are specified, the range
of bytes in the src lun is cloned to the destination lun.
If only a src_lun_obj is specified, the range of bytes in the
src lun is copied to a different location of the src lun.

Use San::Task::RangeFactory to simplify the creation and formatting
of the range parameter.

=over 4

=item Inputs

=item * I<-src_lun_obj> <required>

The lun object in the active file system to sub clone.

=item * I<-dest_lun_obj> <optional>

The lun object in the active file system that the range will be cloned to.

=item * I<-verify> <optional>

1 if the lun objects should be verified for correctness after the sub clone 
operation.

=item Outputs

returns {} since no new lun is created

=item Examples

    # Get lun object
    $lun_clone_helper_obj = San::Task::LunCloneHelper->new( lun_helper_obj => $LunMgrObj );

    # Get lun objects
    $src_lun_obj = ...;
    $dest_lun_obj = ...;

    # Sub Clone lun to lun
    $cloned_lun_attrs = $lun_clone_helper_obj->sub_clone_lun(
        '-src_lun_obj'    => $src_lun_obj,
        '-dest_lun_obj'   => $dest_lun_obj,
        '-range       => '0:0:8,8:8:8,16:16:8',
        '-verify'     => $SanUtil::TRUE
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }

    # Sub Clone lun to itself
    # Blocks 0-7 are cloned to Blocks 8-15
    $cloned_lun_attrs = $lun_clone_helper_obj->sub_clone_lun(
        '-src_lun_obj'    => $src_lun_obj,
        '-range       => '0:8:8',
        '-verify'     => $SanUtil::TRUE
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }

=back

=head2 B<sub_clone_lun_snap(%args)>

Create a sub clone of a lun in a snapshot to another lun in the same volume.
The range of bytes in the src lun is cloned to the destination lun.

Use San::Task::RangeFactory to simplify the creation and formatting
of the range parameter.

=over 4

=item Inputs

=item * I<-src_lun_attr_info> <required>

The hash of attributes that the lun had prior to the snapshot being taken. 
This hash must be built from a lun object with LunHelper, build_lun_attributes.

=item * I<-snapshot_obj> <required>

The snapshot object in which the source lun resides.

=item * I<-dest_lun_obj> <required>

The lun object in the active file system that the range will be cloned to.

=item * I<-verify> <optional>

1 if the lun objects should be verified for correctness after the sub clone 
operation.

=item Outputs

returns {} since no new lun is created

=item Examples

    # Get LunCloneHelper object
    $lun_clone_helper_obj = San::Task::LunCloneHelper->new( lun_helper_obj => $LunMgrObj );

    # Get volume object and dest lun obj
    $vol_obj = ...;
    $dest_lun_obj = ...;
    
    my $lun_attr_info = $LunMgrObj->create_lun(
        'volume_obj' => $vol_obj,
        'size'   => '50M',
        'ostype' => 'linux',
        'lun_name' => 'lun'
    );
    if ( $LunMgrObj->error() ) {
        $Test->nlog( $LunMgrObj->errmsg() );
        return $TCD::FAIL;
    }

    my $snapshot_obj = $vol_obj->create_snapshot_obj( snapshot => 'snapshot' );
    if ( $snapshot_obj->error() ) {
        $Test->nlog( $snapshot_obj->errmsg() );
        return $TCD::FAIL;
    }

    # Clone lun with just the optional parameters.
    $cloned_lun_attrs = $lun_clone_helper_obj->sub_clone_lun_snap(
        '-snapshot_obj'         => $snapshot_obj,
        '-src_lun_attr_info'    => $lun_attr_info,
        '-dest_lun_obj'         => $dest_lun_obj,
        '-range               => '0:0:8,8:8:8,16:16:8',
        '-verify'               => $SanUtil::TRUE
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }    

=back

=head2 B<sub_clone_lun_to_file(%args)>

Create a sub clone of a lun in the active file system to a file in the same volume.
The range of bytes in the src lun is cloned to the destination file.

Use San::Task::RangeFactory to simplify the creation and formatting
of the range parameter.

=over 4

=item Inputs

=item * I<-src_lun_obj> <required>

The lun object in the active file system to sub clone.

=item * I<-dest_file> <optional>

The relative path of the destination file.  thefile or qtree1/thefile

=item * I<-verify> <optional>

1 if the lun objects should be verified for correctness after the sub clone 
operation.

=item Outputs

returns {} since no new lun is created

=item Examples

    # Get lun object
    $lun_clone_helper_obj = San::Task::LunCloneHelper->new( lun_helper_obj => $LunMgrObj );

    # Get lun objects
    $src_lun_obj = ...;
    $dest_lun_obj = ...;

    # Sub Clone lun to lun
    $cloned_lun_attrs = $lun_clone_helper_obj->sub_clone_lun_to_file(
        '-src_lun_obj'    => $src_lun_obj,
        '-dest_lun_obj'   => $dest_lun_obj,
        '-range       => '0:0:8,8:8:8,16:16:8',
        '-verify'     => $SanUtil::TRUE
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }

=back

=head2 B<sub_clone_lun_to_file_snap(%args)>

Create a sub clone of a lun in a snapshot to a destination file.
The range of bytes in the src lun is cloned to the destination file.

Use San::Task::RangeFactory to simplify the creation and formatting
of the range parameter.

=over 4

=item Inputs

=item * I<-src_lun_attr_info> <required>

The hash of attributes that the lun had prior to the snapshot being taken. 
This hash must be built from a lun object with LunHelper, build_lun_attributes.

=item * I<-snapshot_obj> <required>

The snapshot object in which the source lun resides.

=item * I<-dest_file> <required>

The relative path of the destination file.  thefile or qtree1/thefile

=item * I<-verify> <optional>

1 if the lun objects should be verified for correctness after the sub clone 
operation.

=item Outputs

returns {} since no new lun is created

=item Examples

    # Get LunCloneHelper object
    $lun_clone_helper_obj = San::Task::LunCloneHelper->new( lun_helper_obj => $LunMgrObj );

    # Get volume object and dest lun obj
    $vol_obj = ...;
    $dest_file = ...;
    
    my $lun_attr_info = $LunMgrObj->create_lun(
        'volume_obj' => $vol_obj,
        'size'   => '50M',
        'ostype' => 'linux',
        'lun_name' => 'lun'
    );
    if ( $LunMgrObj->error() ) {
        $Test->nlog( $LunMgrObj->errmsg() );
        return $TCD::FAIL;
    }

    my $snapshot_obj = $vol_obj->create_snapshot_obj( snapshot => 'snapshot' );
    if ( $snapshot_obj->error() ) {
        $Test->nlog( $snapshot_obj->errmsg() );
        return $TCD::FAIL;
    }

    # Clone lun with just the optional parameters.
    $cloned_lun_attrs = $lun_clone_helper_obj->sub_clone_lun_to_file_snap(
        '-snapshot_obj'         => $snapshot_obj,
        '-src_lun_attr_info'    => $lun_attr_info,
        '-dest_file'            => $dest_file,
        '-range               => '0:0:8,8:8:8,16:16:8',
        '-verify'               => $SanUtil::TRUE
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }    

=back

=head2 B<sub_clone_file_to_lun(%args)>

Create a sub clone of a file in the active file system to a lun in the same volume.
The range of bytes in the src file is cloned to the destination lun.

Use San::Task::RangeFactory to simplify the creation and formatting
of the range parameter.

=over 4

=item Inputs

=item * I<-src_file> <optional>

The relative path of the src file.  thefile or qtree1/thefile

=item * I<-dest_lun_obj> <required>

The lun object in the active file system to clone to.

=item * I<-verify> <optional>

1 if the lun objects should be verified for correctness after the sub clone 
operation.

=item Outputs

returns {} since no new lun is created

=item Examples

    # Get lun object
    $lun_clone_helper_obj = San::Task::LunCloneHelper->new( lun_helper_obj => $LunMgrObj );

    # Get lun objects
    $src_file = ...;
    $dest_lun_obj = ...;

    # Sub Clone lun to lun
    $cloned_lun_attrs = $lun_clone_helper_obj->sub_clone_file_to_lun(
        '-src_file'       => $src_file,
        '-dest_lun_obj'   => $dest_lun_obj,
        '-range           => '0:0:8,8:8:8,16:16:8',
        '-verify'         => $SanUtil::TRUE
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }

=back

=head2 B<sub_clone_file_to_lun_snap(%args)>

Create a sub clone of a file in a snapshot to a lun in the same volume.
The range of bytes in the src file is cloned to the destination lun.

Use San::Task::RangeFactory to simplify the creation and formatting
of the range parameter.

=over 4

=item Inputs

=item * I<-snapshot_obj> <required>

The snapshot object in which the source file resides.

=item * I<-src_file> <optional>

The relative path of the src file.  thefile or qtree1/thefile

=item * I<-dest_lun_obj> <required>

The lun object in the active file system to clone to.

=item * I<-verify> <optional>

1 if the lun objects should be verified for correctness after the sub clone 
operation.

=item Outputs

returns {} since no new lun is created

=item Examples

    # Get LunCloneHelper object
    $lun_clone_helper_obj = San::Task::LunCloneHelper->new( lun_helper_obj => $LunMgrObj );

    # Get volume object and dest lun obj
    $vol_obj = ...;
    $src_file = ...;
    $dest_lun_obj = ...;
    
    my $snapshot_obj = $vol_obj->create_snapshot_obj( snapshot => 'snapshot' );
    if ( $snapshot_obj->error() ) {
        $Test->nlog( $snapshot_obj->errmsg() );
        return $TCD::FAIL;
    }

    # Clone lun with just the optional parameters.
    $cloned_lun_attrs = $lun_clone_helper_obj->sub_clone_file_to_lun_snap(
        '-snapshot_obj'         => $snapshot_obj,
        '-src_file'             => $src_file,
        '-dest_lun_obj'         => $dest_lun_obj,
        '-range               => '0:0:8,8:8:8,16:16:8',
        '-verify'               => $SanUtil::TRUE
    );
    if ( $lun_clone_helper_obj->error() ) {
        $Test->nlog( $lun_clone_helper_obj->errmsg() );
        return $TCD::FAIL;
    }    

=back

=cut
