# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Object to encapsulate behavior of the file clone range parameter
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::Task::LunCloneRange;

# compiler directives
use strict;

# imports
use Params::Validate qw( :all );
use San::SanUtil;

our @ISA = qw( SanUtil );

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:     See POD documentation at the end of this file for usage
###############################################################################
sub new {
    my ( $class, %args ) = @_;

    # bless instance into class
    my $self = {};
    bless $self, $class;

    # Define instance variables
    $self->{'_src_block'}
        = ( defined $args{'src_block'} ) ? $args{'src_block'} : 0;
    $self->{'_dest_block'}
        = ( defined $args{'dest_block'} ) ? $args{'dest_block'} : 0;
    $self->{'_num_blocks'}
        = ( defined $args{'num_blocks'} ) ? $args{'num_blocks'} : 0;

    return $self;
}

################################################################################
# Method Name: get_src_block
# Objective:
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_src_block() {
    my $self = shift;
    return $self->{'_src_block'};
}

################################################################################
# Method Name: get_dest_block
# Objective:
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_dest_block() {
    my $self = shift;
    return $self->{'_dest_block'};
}

################################################################################
# Method Name: get_num_blocks
# Objective:
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub get_num_blocks() {
    my $self = shift;
    return $self->{'_num_blocks'};
}

################################################################################
# Method Name: set_src_block
# Objective:
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_src_block() {
    my $self = shift;
    $self->{'_src_block'} = shift;
}

################################################################################
# Method Name: set_dest_block
# Objective:
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_dest_block() {
    my $self = shift;
    $self->{'_dest_block'} = shift;
}

################################################################################
# Method Name: set_num_blocks
# Objective:
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub set_num_blocks() {
    my $self = shift;
    $self->{'_num_blocks'} = shift;
}

################################################################################
# Method Name: to_string
# Objective:
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub to_string() {
    my $self = shift;
    return
          $self->{'_src_block'} . ':'
        . $self->{'_dest_block'} . ':'
        . $self->{'_num_blocks'};
}

################################################################################
# Method Name: to_string_ranges
# Objective:
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub to_string_ranges() {

    my ( $class, %args ) = @_;

    my $ranges;
    my @range_objs = @{ $args{range_objs} };

    unless ( defined $args{'range_objs'}
        && ref( $args{'range_objs'} ) eq 'ARRAY' )
    {
        return undef;
    }

    for ( my $i = 0; $i < @range_objs; $i++ ) {
        my $range_obj_ref_type = ref( $range_objs[$i] );

        unless ( $range_obj_ref_type =~ /LunCloneRange/i ) {
            return undef;
        }
        $ranges .= $range_objs[$i]->to_string();
        $ranges .= ( $i != @range_objs - 1 ) ? ',' : '';
    }
    return $ranges;
}

################################################################################
# Method Name: to_io_objs_params
# Objective:
# Details:     See POD documentation at the end of this file for usage
################################################################################
sub to_io_objs_params() {

    my ( $self, %args ) = @_;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    my $io_tool = $args{'io_tool'};
    my %io_objs_params;
    my $BYTES_IN_LBA = 512;

    if ( $io_tool == 'dt' ) {

        %io_objs_params = (
            'src-params' => (
                -limit => $self->{'_num_blocks'} * $BYTES_IN_LBA,
                -pass_thru => 'position=' . $self->{'_start_block'}
            ),
            'dest-params' => (
                -limit => $self->{'_num_blocks'} * $BYTES_IN_LBA ,
                -pass_thru => 'position=' . $self->{'_dest_block'}
            )
        );
    }

    return %io_objs_params;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

LunCloneRange.pm

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

LunCloneRange.pm is an object representation of the range parameter in a file clone
create (lun clone).

A range consists of a source block, a number of blocks to clone, and a
destination block.

The file clone create command takes a source-path, destination-path and a range
parameter.  The range parameter accepts ranges in the format 
<src_block:dest_block:num_blocks>.

num_blocks are cloned from the file/lun specified in the -source-path, starting 
at the src_block, to the dest_block of the file/lun specified in the 
-destination-path. 

=head1 See Also

See these San::Task::LunCloneRangeFactory POD for help creating LunCloneRange objects in bulk.

B<San::Task::LunCloneRangeFactory>

=head1 Attributes

None

=head1 Revision History

Feb-02-2010 schroll@netapp.com: LunCloneRange object checked in.  Code Complete 

=head1 Public Methods

=head2 B<new()>

Class Constructor method. 

=over 4

=item Inputs

=item * I<-src_block> <required>

The src block to start cloning blocks from.

=item * I<-dest_block> <required>

The dest block to clone blocks to.

=item * I<-dest_block> <required>

The number of blocks to clone.

=item Outputs

Returns a LunCloneRange blessed object reference.

=item Examples

   # Create a LunCloneRange object
   my $range_obj = San::Task::LunCloneRange->new((
       src_block => 0,
       dest_block => 0,
       num_blocks => 8
   );
   if ($range_obj->error()) {
      my $err_msg = $range_obj->errmsg();
      $Test->nlog("Failed to instantiate LunCloneRange object:\nReason: $err_msg");
      return $TCD::FAIL;
   }

=back

=head2 B<to_string()>

Create a string representation of a LunCloneRange.
<src_block:dest_block:num_blocks>

=over 4

=item Inputs

=item Outputs

scalar -  string representation of a LunCloneRange

=item Examples

   # Create a LunCloneRange object
   my $range_obj = San::Task::LunCloneRange->new((
       src_block => 0,
       dest_block => 0,
       num_blocks => 8
   );
   if ($range_obj->error()) {
      my $err_msg = $range_obj->errmsg();
      $Test->nlog("Failed to instantiate LunCloneRange object:\nReason: $err_msg");
      return $TCD::FAIL;
   }
   
   $range_str = $range_obj->to_string();
   
=back

=head2 B<to_string_ranges()>

Create a string representation of multiple LunCloneRanges.
<src_block:dest_block:num_blocks>, ...

=over 4

=item Inputs

=item * I<-range_objs> <required>

ref to array of LunCloneRange objects.

=item Outputs

Returns a a string representation of multiple LunCloneRanges.

=item Examples

   # Create a LunCloneRange object
   @range_objs = ( ... );
   
   San::Task::LunCloneRange->to_string_ranges( 'range_objs' => \@range_objs );  

=back

=cut
