# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Object to encapsulate behavior of the file clone range parameter
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

###############################################################################
# How to slice up lun or file and how to map the slices?
# Assume Equal Size luns, LunCloneRanges must map up, starting at prefix
# Direct Mapping, Reverse Mapping
# Half and Half Mapping - Copy first half to second half
# Given a lun, slice it up into slivers of x blocks apiece starting at the prefix
# ^^ First and foremost
###############################################################################

package San::Task::LunCloneRangeFactory;

# compiler directives
use strict;

# imports
use List::MoreUtils;
use Params::Validate qw( :all );
use San::SanUtil;
use San::Task::LunCloneRange;
use Scalar::Util;

our @ISA = qw( SanUtil );
our $BYTES_IN_LBA = 512;
our $BYTES_IN_FBN = 4096;

##############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:     See POD documentation at the end of this file for usage
###############################################################################
sub new {
    my ( $class, %args ) = @_;

    # bless instance into class
    my $self = {};
    bless $self, $class;

    return $self;
}

##############################################################################
# Method:    create_ranges_by_num_blocks_per_ranges
# Objective: Creates n number of ranges where each range is of size
#            num_blocks_per_range.  By default, the src block of the LunCloneRanges
#            created will start at 0 and increment by num_blocks_per_range for
#            each range.
#            TODO:  A mapper object could further determine
#            how to map the source blocks to destination blocks.
# Details:     See POD documentation at the end of this file for usage
###############################################################################
sub create_ranges_by_num_blocks_per_range() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'src_size' => {
                    type     => SCALAR,
                    required => 1
                },
                'src_prefix_size' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => 0
                },
                'src_suffix_size' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => 0
                },
                'dest_size' => {
                    type     => SCALAR,
                    optional => 1
                },
                'dest_prefix_size' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => 0
                },
                'dest_suffix_size' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => 0
                },
                'num_blocks_per_range' => {
                    type      => SCALAR,
                    required  => 1,
                    callbacks => {
                        'num_blocks_per_range must be a multiple of 8' =>
                            sub { $_[0] % 8 == 0 },
                    }
                }
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return {};
    }

    my $num_blocks_per_range = $args{'num_blocks_per_range'};
    my $size                 = $args{'src_size'};
    my $prefix_size          = $args{'src_prefix_size'};
    my $suffix_size          = $args{'src_suffix_size'};
    my $dest_size            = $args{'dest_size'};
    my $dest_prefix_size     = $args{'dest_prefix_size'};
    my $dest_suffix_size     = $args{'dest_suffix_size'};

    # Get the size of the src base, prefix and suffix and calculate lba
    my $base_size = $size - $prefix_size - $suffix_size;
    my $num_lba_blocks_in_base
        = $self->_get_lba_block_count_from_bytes( 'num_bytes' => $base_size );
    my $num_lba_blocks_in_prefix = $self->_get_lba_block_count_from_bytes(
        'num_bytes' => $prefix_size );

    # Get the size of the dest base, prefix and suffix and calculate lba
    my ( $dest_base_size, $dest_num_lba_blocks_in_base,
        $dest_num_lba_blocks_in_prefix );
    if ( exists $args{'dest_size'} ) {
        $dest_base_size = $dest_size - $dest_prefix_size - $dest_suffix_size;
        $dest_num_lba_blocks_in_base = $self->_get_lba_block_count_from_bytes(
            'num_bytes' => $dest_base_size );
        $dest_num_lba_blocks_in_prefix
            = $self->_get_lba_block_count_from_bytes(
            'num_bytes' => $dest_prefix_size );
    }

    # Split up the source lun base file into an array of src blocks
    my @src_blocks = $self->_split_by_num_blocks_per_range(
        'num_blocks'           => $num_lba_blocks_in_base,
        'num_blocks_per_range' => $num_blocks_per_range
    );

    my ( @src_blocks_filtered, @dest_blocks );

    # If a destination is another file/lun
    if ( exists $args{'dest_size'} ) {

        # The destination file/lun might not be the same size as the source
        # file/lun, so filter out any blocks that are invalid in
        # the destination.
        foreach my $src_block (@src_blocks) {
            if ( $src_block
                <= $dest_num_lba_blocks_in_base - $num_blocks_per_range )
            {

                # Account for the prefix
                push( @src_blocks_filtered,
                    $src_block + $num_lba_blocks_in_prefix );
                push( @dest_blocks,
                    $src_block + $dest_num_lba_blocks_in_prefix );
            }
        }

        # If the destination is the same file/lun
    } else {

        # Account for the prefix
        foreach my $src_block (@src_blocks) {
            push( @src_blocks_filtered,
                $src_block + $num_lba_blocks_in_prefix );
        }

        # If there are an uneven amount of blocks in the array,
        # pop the last one off.
        if ( ( @src_blocks_filtered % 2 ) != 0 ) {
            pop(@src_blocks_filtered);
        }
        @dest_blocks
            = splice( @src_blocks_filtered, @src_blocks_filtered / 2 );
    }

    # Create an array of LunCloneRange objects for each source block, dest block pair
    # http://search.cpan.org/~adamk/List-MoreUtils-0.30/lib/List/MoreUtils.pm
    return List::MoreUtils::pairwise {
        San::Task::LunCloneRange->new(
            'src_block'  => $a,
            'dest_block' => $b,
            'num_blocks' => $num_blocks_per_range
        );
    }
    @src_blocks_filtered, @dest_blocks;
}

###############################################################################
#                        Private Methods                                      #
###############################################################################

##############################################################################
# Method:    _split_by_num_blocks_per_range
# Objective: Creates an array where each index is the num_blocks_per_range
#            multiplied by the index.  The array ends when the num_blocks has
#            been exceeded.
###############################################################################
sub _split_by_num_blocks_per_range() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    eval {
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   ('num_blocks') => {
                    type     => SCALAR,
                    required => 1
                },
                ('num_blocks_per_range') => {
                    type     => SCALAR,
                    required => 1
                },
            }
        );
    };

    if ($@) {
        $self->sanlog( -msg => 'Validation failed ' . $@, -msg_level => 1 );
        $self->_set_err($@);
        return ();
    }

    my $num_blocks           = $args{'num_blocks'};
    my $num_blocks_per_range = $args{'num_blocks_per_range'};
    my $i                    = 0;
    my $curr_block           = 0;
    my @blocks               = ();

    while ( $curr_block <= $num_blocks - $num_blocks_per_range ) {
        push( @blocks, $curr_block );
        $curr_block = ++$i * $num_blocks_per_range;
    }

    return @blocks;
}

##############################################################################
# Method:    _get_lba_block_count_from_bytes
# Objective: Returns the number of LBA blocks in a number of bytes
###############################################################################
sub _get_lba_block_count_from_bytes() {

    my ( $self, %args ) = @_;

    # Perform method entry functions
    $self->entry( -msg_level => 5 );
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Make sure num_bytes is defined in the array and is not a reference
    unless ( defined $args{'num_bytes'}
        && ref( $args{'num_bytes'} ) eq '' )
    {
        my $err_msg = 'Validation failed - num_bytes is not a SCALAR';
        $self->sanlog( -msg => $err_msg, -msg_level => 1 );
        $self->_set_err($err_msg);
        return undef;
    }

    return $args{'num_bytes'} / $BYTES_IN_LBA;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

LunCloneRange.pm

=head1 Author

Chris Schroll (schroll@netapp.com)

=head1 Description

LunCloneRangeFactory.pm is a simple factory to create LunCloneRange objects in bulk from luns.
These objects can be used to build the -range parameter of the file clone
create (lun clone) command.

The methods of LunCloneRangeFactory typically take size attributes from a lun or file
and divide the lun or file into a number of block ranges, which are returned 
as LunCloneRange objects. 

=head1 See Also

See these San::Task::LunCloneRange POD for help using the created LunCloneRange objects.

B<San::Task::LunCloneRangeFactory>

=head1 Attributes

None

=head1 Revision History

Feb-02-2010 schroll@netapp.com: LunCloneRangeFactory class checked in with a basic
factory method.

=head1 Public Methods

=head2 B<new()>

Class Constructor method. 

=over 4

=item Inputs

=item Outputs

Returns a LunCloneRangeFactory blessed object reference.

=item Examples

   # Create a LunCloneRangeFactory object
   my $range_factory_obj = San::Task::LunCloneRangeFactory->new(;
   if ($range_factory_obj->error()) {
      my $err_msg = $range_factory_obj->errmsg();
      $Test->nlog("Failed to instantiate LunCloneRangeFactory object:\nReason: $err_msg");
      return $TCD::FAIL;
   }

=back

=head2 B<new()>

Class Constructor method. 

=over 4

=item Inputs

=item Outputs

Returns a LunCloneRangeFactory blessed object reference.

=item Examples

   # Create a LunCloneRangeFactory object
   my $range_factory_obj = San::Task::LunCloneRangeFactory->new(;
   if ($range_factory_obj->error()) {
      my $err_msg = $range_factory_obj->errmsg();
      $Test->nlog("Failed to instantiate LunCloneRangeFactory object:\nReason: $err_msg");
      return $TCD::FAIL;
   }

=back

=head2 B<create_ranges_by_num_blocks_per_range(%args)>

This splits up a src lun or file into a number of ranges, starting at the first
block of the lun's base or of the file.  Each subsquent blocks is
num_bl


=over 4

=item Inputs

=item * I<-src_size> <required>

The size of the src.  Either the size of the file or the total size of the lun.

=item * I<-src_prefix_size> <optional>

If the src is a lun, this is the size of the prefix of the lun.

=item * I<-src_suffix_size> <optional>

If the src is a lun, this is the size of the suffix of the lun.

=item * I<-dest_size> <required>

The size of the dest.  Either the size of the file or the total size of the lun.

=item * I<-dest_prefix_size> <optional>

If the dest is a lun, this is the size of the prefix of the lun.

=item * I<-destsuffix_size> <optional>

If the dest is a lun, this is the size of the suffix of the lun.

=item Outputs

returns array of San::Task::LunCloneRange objects. 

=item Examples

    my @ranges  = $LunCloneRangeFactoryObj->create_ranges_by_num_blocks_per_range(
        'src_size' => $lun_attrs->{$LUN_ATTR_OBJ}->get_size(),
        'src_prefix_size' => $lun_attrs->{$LUN_ATTR_OBJ}->get_prefix_size(),
        'dest_size'       => $lun_attrs1->{$LUN_ATTR_OBJ}->get_size(),
        'num_blocks_per_range' => $LUN_SUB_BLOCKS_PER_RANGE
    );
    if ( $LunCloneRangeFactoryObj->error() ) {
         $Test->nlog( $LunCloneRangeFactoryObj->errmsg() );
    }

=back
=cut
