# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/San/Task/LunHelper.pm#1 $
#
# Copyright (c) 2009 NetApp, Inc.
# All rights reserved.
#
##
## @summary Utility for testing lun functionality
## To provide library functions for the testing lun.
## All the cmode lun testing script will use this library
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::Task::LunHelper;

# compiler directives
use strict;
use Exporter;
use San::SanUtil;
use San::SanScsiT::XScsiT;

use vars qw( @EXPORT @ISA );

# inherit base class
@ISA = qw( SanUtil );

# constant variable
use constant NOERRORFOUND => 'NoErrorFound';

# global variable

my @ExpectedLunAttr = qw(vserver path volume qtree lun size prefix_size
                         extent_size suffix_size ostype space_reserve
                         serial comment space_reserve_honored space_allocation
                         state uuid mapped vdiskId block_size device_legacy_id
                         device_binary_id device_text_id read_only
                         inconsistent_filesystem inconsistent_blocks nvfail
                         size_used creation_timestamp);

###############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self;

    # bless instance into class
    bless $self, $class;
    return $self;
}

###############################################################################
# Method:    create_lun
# Objective: to create lun object in given volume or qtree and build the lun
#            attributes that user requested (expected) for later verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_lun {
    my ($self, %args) = @_;

    my ($vol_obj, $vol_name, $lun_attr_expected, $use_fcp_lif_name, %cmds_args,
        $verify, $lun_obj, $qtree_obj, $qtree_name, $container_obj);
    my $verify_serial = 'true';

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /container_obj/) {
            if (ref($value) =~
                   /San::ControllerCluster::Filer10M::Vserver::Volume::Qtree/) {
                $vol_obj = $value->get_parent_obj();
                $qtree_obj = $value;
            } elsif (ref($value) =~
                   /San::ControllerCluster::Filer10M::Vserver::Volume/) {
                $vol_obj = $value;
            } else {
                $self->_set_err("invalid value for \"-$key\": $value");
                return {};    # a reference to an empty hash
            }
            $container_obj = $value;
        } elsif ($key =~ /size|lun_name|ostype|path|option|file-path/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return {};    # a reference to an empty hash
            }
        } elsif ($key eq 'verify') {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key eq 'verify_serial') {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify_serial = $value;
        } else {
            $self->_set_err(
                 "invalid \"-$key\": passed. Valid parameters: " .
                 "container_obj|size|lun_name|ostype|path|option|verify|file-path");
            return {};
        }
    }

    # verify mandatory argument -container_obj
    unless (defined $args{'container_obj'}) {
        $self->_set_err("mandatory parameter \"-container_obj\" is needed");
        return {};
    }

    # verify mandatory cmd argument -lun_name - lun name
    unless (defined $args{'lun_name'} and length $args{'lun_name'}) {
        $self->_set_err("mandatory parameter \"-lun_name\" " .
                        "is need with lun name");
        return {};
    }

    # verify mandatory cmd argument -size
    unless (defined $args{'size'} and length $args{'size'}) {
        $self->_set_err("mandatory parameter \"-size\" " .
                        "is need with lun name");
        return {};
    }

    # verify mandatory cmd argument -ostype
    unless (defined $args{'ostype'} and length $args{'ostype'}) {
        $self->_set_err("mandatory parameter \"-ostype\" " .
                        "is need with lun name");
        return {};
    }
    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    # build the hash - expected lun attributes for later verification
    # lun show command will have following attributes for given lun.
    # vserver|path|volume|qtree|lun|size|ostype|serial|comment|space_resv_enabled|uuid|isOnline|isMapped|block-size|device-id|read-only|alignment|share-state|staging|are-vols-busy|are-vols-onlining|size-used|
    # if any attribute is not passed fill with "-" as default,
    # since lun show output will have "-"
    # Sample output if you have done this - set -showallfields true -showseparator "|" -rows 0
    # vserver|path|volume|qtree|lun|size|ostype|serial|comment|space_resv_enabled|uuid|isOnline|isMapped|block-size|device-id|read-only|alignment|share-state|staging|are-vols-busy|are-vols-onlining|size-used|
    # lun_move_vser0|/vol/lun_move_vol0/lun_move_mapped_aix22|lun_move_vol0|""|lun_move_mapped_aix22|33554432|aix|1BbFb+8SQ4B-|""|false|2af008e0-68f3-11df-acfd-000c297f4d21|true|false|512|-|true|indeterminate|none|false|false|false|10|

    # fill the attribute - lun for both cmd arg and expected info
    $cmds_args{'lun'} = $args{'lun_name'};
    $lun_attr_expected->{'lun'} = $args{'lun_name'};

    # when you create a lun, it should be in state online
    $lun_attr_expected->{'state'} = "online";

    # get the name of the volume
    $vol_name = $vol_obj->get_name();
    if ($vol_obj->error()) {
        $self->_set_err("Failed to get volume name.\nReason:\n" .
                        $vol_obj->errmsg());
        return {};
    }
    if ($qtree_obj) {
        $qtree_name = $qtree_obj->get_name();
        $lun_attr_expected->{'qtree'} = $qtree_name;
    } else {
        $qtree_name = "";
        $lun_attr_expected->{'qtree'} = "\"\"";
    }

    # get the name of the vserver
    my $vs_obj = $vol_obj->get_parent_obj();
    my $vserver = $vs_obj->get_name();
    if ($vol_obj->error()) {
        $self->_set_err("Failed to get vserver name.\nReason:\n" .
                        $vol_obj->errmsg());
        return {};
    }
    $lun_attr_expected->{'vserver'} = $vserver;

    # handle the size and ostype of the LUN
    my $size  = $args{'size'};
    $cmds_args{'size'} = $size;
    $lun_attr_expected->{'size'} = $size;
    my $ostype  = $args{'ostype'};
    $cmds_args{'ostype'} = $ostype;
    $lun_attr_expected->{'ostype'} = $ostype;

    # handle the space reservation setting for the LUN
    if ($args{option}) {
        if ($args{option} =~ /noreserve/) {
            $cmds_args{'space-reserve'} = "disabled";
            $lun_attr_expected->{'space_reserve'} = "disabled";
        } else {
            $lun_attr_expected->{'space_reserve'} = "enabled";
            $cmds_args{'space-reserve'} = "enabled";
        }
        delete $cmds_args{option};
    } else {
        # luns are created space reserved by default
        $lun_attr_expected->{'space_reserve'} = "enabled";
    }

    # fill the attribute - volume/lun_path for both cmd arg and expected info
    $cmds_args{'volume'} = $vol_name;
    $lun_attr_expected->{'volume'} = $vol_name;
    if ($cmds_args{'path'}) {
        $lun_attr_expected->{'path'} = $cmds_args{'path'};
    } else {
        if ($qtree_name) {
            $lun_attr_expected->{'path'} =
                "/vol/$vol_name/$qtree_name/$args{'lun_name'}";
        } else {
            $lun_attr_expected->{'path'} = "/vol/$vol_name/$args{'lun_name'}";
        }
    }

    # if zapi, do some pre-processing cli to zapi argument conversion
    if ($vol_obj->get_cmd_interface() =~ /zapi/i) {
        if ($self->_cli_to_zapi(-cmd_args => \%cmds_args) != $SanUtil::TRUE) {
            return {};
        }
    }

    # create lun object.  hardcoding a large timeout value as mkfile can take
    # a long time; do we need to make this more flexible?
    if (defined($args{'file-path'})) {
        $lun_obj = $container_obj->create_lun_from_file(%cmds_args,
                                       'from_file' => $args{'file-path'},
                                       'size' => $args{'size'},
                                       'timeout' => 7200);
	# Note that the file-path here is just the file name
        $lun_attr_expected->{'file_path'} = $args{'file-path'};

    } else {
        $lun_obj = $container_obj->create_lun_obj(%cmds_args);
    }
    if ($container_obj->error()) {
        my $msg = $container_obj->errmsg();
        $self->sanlog(-msg       => "In create_lun - error from create_lun_obj()");
        $self->_set_err(  "Failed to create lun("
                        . $args{'lun'} . ") in "
                        . $container_obj->get_name()
                        . "\nReason:\n"
                        . $msg);
        return {};
    }

    # verify that the new lun's serial number is unique; any node object
    # should suffice
    if ($verify_serial eq 'true') {
        my $node_obj = $self->{'node_objs'}->[0];
        my @dups;
        if ($node_obj->lun_serials_are_unique(duplicates => \@dups) == $SanUtil::FALSE) {
            if ($node_obj->error()) {
                $self->_set_err($node_obj->errmsg());
                return $SanUtil::FALSE;
            }
            $self->_set_err("Found duplicate serial(s): " . join(',', @dups));
            return $SanUtil::FALSE;
        }
    }

    # fillup the lun object also....
    $lun_attr_expected->{'lun_obj'} = $lun_obj;

    # fill in some more of the attributes with what we expect the default
    # values to be
    $lun_attr_expected->{'mapped'} = "unmapped";
    $lun_attr_expected->{'comment'} = "\"\"";
    $lun_attr_expected->{'read_only'} = "false";
    $lun_attr_expected->{'device_legacy_id'} = "-";
    $lun_attr_expected->{'device_text_id'} = "-";
    $lun_attr_expected->{'device_binary_id'} = "-";
    $lun_attr_expected->{'inconsistent_filesystem'} = "false";
    $lun_attr_expected->{'inconsistent_blocks'} = "false";
    $lun_attr_expected->{'size_used'} = 0;
    $lun_attr_expected->{'space_reserve_honored'} = "true";
    $lun_attr_expected->{'space_allocation'} = "disabled";
    $lun_attr_expected->{'nvfail'} = "false";

    # query the controller for the attributes that we can't make any
    # assumptions about
    my $attr_list = $lun_obj->get_list_info();
    if ($lun_obj->error()) {
        $self->_set_err($lun_obj->errmsg());
        return $SanUtil::FALSE;
    }
    foreach my $key (keys %{$attr_list}) {
        if (!defined($lun_attr_expected->{$key})) {
            $lun_attr_expected->{$key} = $attr_list->{$key};
        }
    }

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # dump the expected lun attributes for debug purpose
    $self->sanlog(-msg       => "Expected lun Attributes after create:\n" .
                                 $self->dumper_get($lun_attr_expected),
                  -msg_level => 3);

    return $lun_attr_expected;
}    # end sub create_lun

###############################################################################
# Method:    offline_lun
# Objective: offline a lun and update the expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub offline_lun {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr_expected, %cmds_args, $verify);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|verify");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }
    $lun_name = $lun_attr_expected->{'lun'};

    # offline the lun object
    $lun_obj->offline();
    if ($lun_obj->error()) {
        $self->_set_err("Failed to offline lun($lun_name) " .
                        "\nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected lun attributes
    $lun_attr_expected->{'state'} = "offline";

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes after add:\n" .
                               $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub offline_lun

###############################################################################
# Method:    test_all_lun_cmds
# Objective: given a lun run all of the possible LUN commands on it and check for failures
# Details:   See POD documentation at the end of this file.
###############################################################################
sub test_all_lun_cmds {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_attr_expected, %cmds_args);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "igroup_obj");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info
    unless (defined $args{'lun_attr_info'}) {
        $self->_set_err("mandatory parameter \"-lun_attr_info\" is needed");
        return $SanUtil::FALSE;
    }

    # This method will do the following operations on a LUN, verifying
    # that the attributes are updated correctly following each operation
    # each of these tests will restore the LUN to it's original state
    # map/unmap
    # get/set/clear/set back comment
    # move/move back to original location
    # offline/online
    # resize/resize back to original location
    # get/set/unset/set back dev_id
    # get/set/set back to original serial number
    # get/reset stats
    # get/set/unset/set back attribute
    # get/set/unset/set back select attribute
    # get/set/set back to original setting -  space reservation
    # get geometry, inquiry info, maxsize, minsize
    # occupied size, PR info, vdisk attributes, scsi reservations
    # initiator list, initiator logged in, restore status

    $lun_attr_expected = $args{'lun_attr_info'};
    $lun_obj           = $lun_attr_expected->{'lun_obj'};

    # test lun map
    $self->test_lun_map(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE;
    }

    # test lun comment
#    $self->test_lun_comment(-lun_attr_info => $lun_attr_expected);
#    if ($self->error()) {
#        return $SanUtil::FALSE;
#    }

    # test lun move
#    $self->test_lun_move(-lun_attr_info => $lun_attr_expected);
#    if ($self->error()) {
#        return $SanUtil::FALSE;
#    }

    # test lun offline/online
#    $self->test_lun_offline_online(-lun_attr_info => $lun_attr_expected);
#    if ($self->error()) {
#        return $SanUtil::FALSE;
#    }

    return $SanUtil::TRUE;
}    # end sub test_all_lun_cmds

###############################################################################
# Method:    online_lun
# Objective: online a lun and update the expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub online_lun {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr_expected, %cmds_args, $verify);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|verify");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }
    $lun_name = $lun_attr_expected->{'lun'};

    # offline the lun object
    $lun_obj->online();
    if ($lun_obj->error()) {
        $self->_set_err("Failed to offline lun($lun_name) " .
                        "\nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected lun attributes
    $lun_attr_expected->{'state'} = "online";

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes after add:\n" .
                               $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub online_lun

###############################################################################
# Method:    move_lun
# Objective: move a lun and update the expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub move_lun {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_src_name, $lun_dest_path, $lun_attr_expected, %cmds_args,
        $verify);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /path/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_dest_path = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|path|lun_obj|verify");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    # verify mandatory cmd argument - lun_path
    unless (defined $args{'path'}) {
        $self->_set_err("mandatory parameter \"-path\" is needed ");
        return $SanUtil::FALSE;
    }

    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }

    $lun_src_name = $lun_attr_expected->{'lun'};

    # move the lun object
    $lun_obj->move(-path => $lun_dest_path);
    if ($lun_obj->error()) {
        $self->_set_err("Failed to move lun($lun_src_name) " .
                        "\nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected lun attributes
    my @fields = split(/\//, $lun_dest_path);
    if (scalar(@fields) == 5) {  # lun is in a qtree
        $lun_attr_expected->{'lun'} = $fields[4];
        $lun_attr_expected->{'qtree'} = $fields[3];
    } else {
        $lun_attr_expected->{'lun'} = $fields[3];
        $lun_attr_expected->{'qtree'} = '""';
    }
    $lun_attr_expected->{'path'} = $lun_dest_path;

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes after add:\n" .
                               $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub move_lun

###############################################################################
# Method:    resize_lun
# Objective: resize a lun and update the expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub resize_lun {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $new_lun_size, $lun_attr_expected,
        %cmds_args, $option, $force, $verify);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /size/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $new_lun_size = $value;
        } elsif ($key =~ /force/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $force = $value;
        } elsif ($key =~ /option/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $option = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|size|force|option|verify");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    # verify mandatory cmd argument - size
    unless (defined $args{'size'}) {
        $self->_set_err("mandatory parameter \"-size\" is needed");
        return $SanUtil::FALSE;
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }
    $lun_name = $lun_attr_expected->{'lun'};

    # resize the lun object
    my %cmd_args;
    %cmd_args->{size} = $new_lun_size;
    if ($option eq "suffix") {
        %cmd_args->{option} = "-S";
    } elsif ($option eq "prefix") {
        %cmd_args->{option} = "-P";
    }
    if ($force eq "true") {
        %cmd_args->{force} = "true";
    }
    $lun_obj->resize(%cmd_args);
    if ($lun_obj->error()) {
        $self->_set_err("Failed to resize lun($lun_name) " .
                        "\nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Resizing lun " . $lun_name . " to " . $new_lun_size);
    # update the expected lun attributes
    $lun_attr_expected->{'size'} = $new_lun_size;

    if ($verify) {
        # the user wants to verify that the resize is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes after add:\n" .
                               $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub resize_lun

###############################################################################
# Method:    set_lun_comment
# Objective: set the lun comment and update the expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_lun_comment {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr_expected, %cmds_args, $comment, $verify);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /comment/) {
            # comment could be a null string if we are trying to clear it
            $comment = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|comment|verify");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }
    $lun_name = $lun_attr_expected->{'lun'};

    # set the comment on the lun object
    $lun_obj->set_comment(-comment => $comment);
    if ($lun_obj->error()) {
        $self->_set_err("Failed to set the comment on lun($lun_name) " .
                        "\nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected lun attributes
    $lun_attr_expected->{'comment'} = $comment;

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes after add:\n" .
                               $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub set_lun_comment

###############################################################################
# Method:    set_lun_dev_id
# Objective: set the lun dev_id and update the expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_lun_dev_id {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr_expected, %cmds_args,
        $dev_id_value, $dev_id_type, $verify);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /dev_id_value/) {
            if (!defined($value)) {  # allow "" to clear device id
                $self->_set_err("Undefined value for \"-$key\"");
                return $SanUtil::FALSE;
            }
            $dev_id_value = $value;
        } elsif ($key =~ /dev_id_type/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $dev_id_type = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|dev_id_binary|dev_id_text|dev_id_none|verify");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    unless (defined $args{'dev_id_type'} and defined $args{'dev_id_value'}) {
        $self->_set_err("dev_id_type and dev_id_value are mandatory parameters");
        return $SanUtil::FALSE;
    }

    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }
    $lun_name = $lun_attr_expected->{'lun'};

    # set or unset the dev_id on the lun object
    if ($dev_id_value eq "" or $dev_id_value eq "-" ) {
        $lun_obj->unset_device_id(-'type' => $dev_id_type);
        if ($lun_obj->error()) {
            $self->_set_err("Failed to unset the device-id on lun($lun_name) " .
                            "\nReason:\n" . $lun_obj->errmsg());
            return $SanUtil::FALSE;
        }
        $dev_id_value = '-';
    } else {
        $lun_obj->set_device_id(-'device-id' => $dev_id_value,
                                -'type'      => $dev_id_type);
        if ($lun_obj->error()) {
            $self->_set_err("Failed to set the device-id on lun($lun_name) " .
                            "\nReason:\n" . $lun_obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # update the expected lun attributes
    if ($dev_id_type =~ /binary/) {
        $lun_attr_expected->{'device_binary_id'} = $dev_id_value;
    }
    if ($dev_id_type =~ /text/) {
        # if the value contains spaces, the CLI display adds double quotes
        if ($lun_obj->get_cmd_interface() =~ /cli/i and $dev_id_value =~ /\s/) {
            $lun_attr_expected->{'device_text_id'} = "\"$dev_id_value\"";
        } else {
            $lun_attr_expected->{'device_text_id'} = $dev_id_value;
        }
    }
    # binary is automatically set too in this case, so we need to convert
    # from decimal to hex (unless we're clearing the dev id)
    if ($dev_id_type =~ /legacy/) {
        $lun_attr_expected->{'device_legacy_id'} = $dev_id_value;
        if ($dev_id_value eq '-') {
            $lun_attr_expected->{'device_binary_id'} = '-';
        } else {
            $lun_attr_expected->{'device_binary_id'} = sprintf( "%08x", $dev_id_value);
        }
    }

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes after add:\n" .
                               $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub set_lun_dev_id

###############################################################################
# Method:    set_lun_reservation
# Objective: set the lun reservation and update the expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_lun_reservation {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr_expected, %cmds_args,
        $space_reserve, $verify);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /space-reserve/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $space_reserve = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|space-reserve|verify");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    # verify mandatory cmd argument - space-reserve
    unless (defined $args{'space-reserve'}) {
        $self->_set_err("mandatory parameter \"-space-reserve\" is needed");
        return $SanUtil::FALSE;
    }

    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }
    $lun_name = $lun_attr_expected->{'lun'};

    # set the reservation on the lun object
    $lun_obj->set_space_resrv('-space-reserve' => $space_reserve);
    if ($lun_obj->error()) {
        $self->_set_err("Failed to set the reservation on lun($lun_name) " .
                        "\nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected lun attributes
    foreach my $lun_attr_name (qw(space_reserve)) {
        $lun_attr_expected->{$lun_attr_name} = $space_reserve;
    }

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes after add:\n" .
                               $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub set_lun_reservation

###############################################################################
# Method:    set_lun_serial
# Objective: set lun serial number and update the expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_lun_serial {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr_expected, %cmds_args, $serial, $verify);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /serial/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $serial = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|serial|verify");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    # verify mandatory cmd argument - serial
    unless (defined $args{'serial'}) {
        $self->_set_err("mandatory parameter \"-serial\" is needed");
        return $SanUtil::FALSE;
    }

    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }
    $lun_name = $lun_attr_expected->{'lun'};

    # set the serial on the lun object
    $lun_obj->set_serial(-serial => $serial);
    if ($lun_obj->error()) {
        $self->_set_err("Failed to set the serial on lun($lun_name) " .
                        "\nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected lun attributes
    $lun_attr_expected->{'serial'} = $serial;

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes after add:\n" .
                               $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub set_lun_comment

###############################################################################
# Method:    get_lun_maxsize
# Objective: get the maximum lun size that can be created as reported by
#            the lun maxsize command.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lun_maxsize {
    my ($self, %args) = @_;

    my ($container_obj, $ostype);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /container_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return {};
            }
            $container_obj = $value;
        } elsif ($key =~ /ostype/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return {};
            }
            $ostype = $value;
        } else {
            $self->_set_err("invalid \"-$key\": passed. Valid parameters: " .
                            "container_obj|ostype");
            return {};
        }
    }

    # verify mandatory cmd argument - container_obj
    unless (defined $args{'container_obj'}) {
        $self->_set_err("mandatory parameter \"-container_obj\" is needed");
        return {};
    }

    # verify mandatory cmd argument - ostype
    unless (defined $args{'ostype'}) {
        $self->_set_err("mandatory parameter \"-ostype\" is needed");
        return {};
    }

    # library returns hashref
    my $lun_maxsize = $container_obj->get_lun_maxsize(-ostype => $ostype);
    if ($container_obj->error()) {
        $self->_set_err($container_obj->errmsg());
        return {};
    }

    return $lun_maxsize;
}

###############################################################################
# Method:    map_lun_to_igroup
# Objective: to map lun to existing igroup and update the
#            expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub map_lun_to_igroup {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $igroup_obj, $igroup_name, $lun_attr_expected,
        $lun_id, $verify, $force);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /igroup_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Igroup/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $igroup_obj = $value;
        } elsif ($key =~ /lun_id/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_id = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key =~ /force/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $force = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|igroup_obj|lun_id|verify");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    # verify mandatory cmd argument - igroup_obj
    unless (defined $args{'igroup_obj'}) {
        $self->_set_err("mandatory parameter \"-igroup_obj\" is needed");
        return $SanUtil::FALSE;
    }

    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }
    $lun_name = $lun_attr_expected->{'lun'};

    $igroup_name = $igroup_obj->get_name();
    if ($igroup_obj->error()) {
        $self->_set_err("Failed to get igroup name \nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # map the lun object
    my %cmd_args = ('-igroup' => $igroup_name);
    $cmd_args{'-lun-id'} = $lun_id if (defined($lun_id));
    $cmd_args{'-force'} = $force if (defined($force));
    $lun_obj->map(%cmd_args);
    if ($lun_obj->error()) {
        $self->_set_err("Failed to map lun($lun_name) to" .
                        "igroup($igroup_name)\nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected lun attributes
    # add mapped -> if user requested to map
    # This isn't yet implemented in lun show
    $lun_attr_expected->{'mapped'} = "mapped";

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes after add:\n" .
                               $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub map_lun_to_igroup

###############################################################################
# Method:    test_lun_map
# Objective: this will test that a given lun can be mapped/unmapped
# Details:   See POD documentation at the end of this file.
###############################################################################
sub test_lun_map {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, @igroup_objs, $igroup_name, $lun_attr_expected,
        %cmds_args);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "igroup_obj");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info
    unless (defined $args{'lun_attr_info'}) {
        $self->_set_err("mandatory parameter \"-lun_attr_info\" is needed");
        return $SanUtil::FALSE;
    }

    $lun_attr_expected = $args{'lun_attr_info'};
    $lun_obj           = $lun_attr_expected->{'lun_obj'};
    $lun_name          = $lun_attr_expected->{'lun'};

    # first get the mapped state of the lun, so that we can return the
    # lun in the state we got it in
    if ($lun_obj->is_mapped()){
        # get the maps and save them, so we can restore them
        @igroup_objs = $lun_obj->get_maps();

        # if the lun was mapped when this test was called, we will
        # unmap the lun, then restore the original maps
        foreach my $igroup_obj (@igroup_objs) {
            $self->unmap_lun_from_igroup(-lun_attr_info => $lun_attr_expected,
                                         -igroup_obj    => $igroup_obj);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
        }   # end of foreach my $igroup_obj (@igroup_objs)

        # now that we have unmapped the lun from each igroup,
        # we will restore the original state
        foreach my $igroup_obj (@igroup_objs) {
            $self->map_lun_to_igroup(-lun_attr_info => $lun_attr_expected,
                                     -igroup_obj    => $igroup_obj);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
        }   # end of foreach my $igroup_obj (@igroup_objs)
    } else {
        # the lun was unmapped when the test was called,
        # we will create an igroup, map the lun to it, then unmap it
        my $vserver_obj = ($lun_obj->get_parent_obj())->get_parent_obj();
        my $igroup_name = "test_lun_map_igroup";
        my $igroup_obj = $vserver_obj->create_igroup_obj(-name            => $igroup_name,
                                                         -"protocol-type" => "mixed",
                                                         -ostype          => "linux");
        if ($vserver_obj->error()) {
            _set_err($vserver_obj->errmsg());
            return $SanUtil::FALSE;
        }
        $self->map_lun_to_igroup(-lun_attr_info => $lun_attr_expected,
                                 -igroup_obj    => $igroup_obj);
        if ($self->error()) {
            return $SanUtil::FALSE;
        }
        $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        $self->unmap_lun_from_igroup(-lun_attr_info => $lun_attr_expected,
                                     -igroup_obj    => $igroup_obj);
        if ($self->error()) {
            return $SanUtil::FALSE;
        }
        $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        $igroup_obj->destroy();
        if ($igroup_obj->error()) {
            _set_err($igroup_obj->errmsg());
            return $SanUtil::FALSE;
        }
   }   # end of if ($lun_obj->is_mapped())

    return $SanUtil::TRUE;
}    # end sub test_lun_map

###############################################################################
# Method:    test_lun_comment
# Objective: this will test that a given lun can have the comment set/unset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub test_lun_comment {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr_expected, %cmds_args, $comment);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "igroup_obj");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info
    unless (defined $args{'lun_attr_info'}) {
        $self->_set_err("mandatory parameter \"-lun_attr_info\" is needed");
        return $SanUtil::FALSE;
    }

    $lun_attr_expected = $args{'lun_attr_info'};
    $lun_obj           = $lun_attr_expected->{'lun_obj'};
    $lun_name          = $lun_attr_expected->{'lun'};

    # first get the comment of the lun, so that we can return the
    # lun in the state we got it in
    $comment = $lun_attr_expected->{'comment'};
    my $lun_obj_comment = $lun_obj->get_comment();

    if ($comment ne $lun_obj_comment) {
        $self->_set_err("Comment not in good starting state.  lun hash has " .
                        $comment . " as the value and lun object shows " .
                        $lun_obj_comment . " as the value");
        return $SanUtil::FALSE;
    }

    # now we want to set the comment to something
    $self->set_lun_comment(-lun_attr_info => $lun_attr_expected,
                           -comment       => "this is a lun comment for test");
    $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE;
    }

    # now we want to clear the comment
    $self->set_lun_comment(-lun_attr_info => $lun_attr_expected,
                           -comment       => "");
    $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE;
    }

    # lastly, set the comment back to it's original value
    $self->set_lun_comment(-lun_attr_info => $lun_attr_expected,
                           -comment       => $comment);
    $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}    # end sub test_lun_comment

###############################################################################
# Method:    test_lun_move
# Objective: this will test that a given lun can be moved/restored
# Details:   See POD documentation at the end of this file.
###############################################################################
sub test_lun_move {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr_expected, %cmds_args, $path);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "igroup_obj");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info
    unless (defined $args{'lun_attr_info'}) {
        $self->_set_err("mandatory parameter \"-lun_attr_info\" is needed");
        return $SanUtil::FALSE;
    }

    $lun_attr_expected = $args{'lun_attr_info'};
    $lun_obj           = $lun_attr_expected->{'lun_obj'};
    $lun_name          = $lun_attr_expected->{'lun'};

    # first get the current lun path, so that we can return the
    # lun in the state we got it in
    $path = $lun_attr_expected->{'path'};
    my $lun_obj_path = $lun_obj->get_lun_path();

    if ($path ne $lun_obj_path) {
        $self->_set_err("Path not in good starting state.  lun hash has " .
                        $path . " as the value and lun object shows " .
                        $lun_obj_path . " as the value");
        return $SanUtil::FALSE;
    }

    $path =~ /(\S+)\/(\S+)/;
    my $lun_dest_path = $1 . "/" . $2 . "_test_lun_move";

    # now we want to move the lun to a new name in the same volume/qtree
    $self->move_lun(-lun_attr_info => $lun_attr_expected,
                    -path          => $lun_dest_path);
    $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE;
    }

    # now, move the lun back to it's original location
    $self->move_lun(-lun_attr_info => $lun_attr_expected,
                    -path          => $path);
    $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}    # end sub test_lun_move

###############################################################################
# Method:    test_lun_offline_online
# Objective: this will test that a given lun can be offlined/onlined
# Details:   See POD documentation at the end of this file.
###############################################################################
sub test_lun_offline_online {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr_expected, %cmds_args, $state);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "igroup_obj");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info
    unless (defined $args{'lun_attr_info'}) {
        $self->_set_err("mandatory parameter \"-lun_attr_info\" is needed");
        return $SanUtil::FALSE;
    }

    $lun_attr_expected = $args{'lun_attr_info'};
    $lun_obj           = $lun_attr_expected->{'lun_obj'};
    $lun_name          = $lun_attr_expected->{'lun'};

    # first get the current lun path, so that we can return the
    # lun in the state we got it in
    $state = $lun_attr_expected->{'state'};
    my $lun_obj_state = $lun_obj->get_state();

    if ($state ne $lun_obj_state) {
        $self->_set_err("State not in good starting state.  lun hash has " .
                        $state . " as the value and lun object shows " .
                        $lun_obj_state . " as the value");
        return $SanUtil::FALSE;
    }

    if ($state =~ /true/) {
        # move the lun to offline state
        $self->offline_lun(-lun_attr_info => $lun_attr_expected);
        $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        # move the lun back to it's original state
        $self->online_lun(-lun_attr_info => $lun_attr_expected);
        $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
        if ($self->error()) {
            return $SanUtil::FALSE;
        }
    } else {
        # move the lun to online state
        $self->online_lun(-lun_attr_info => $lun_attr_expected);
        $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        # move the lun back to it's original state
        $self->offline_lun(-lun_attr_info => $lun_attr_expected);
        $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
        if ($self->error()) {
            return $SanUtil::FALSE;
        }
    }  # end of else if ($state =~ /online/)


    return $SanUtil::TRUE;
}    # end sub test_lun_move

###############################################################################
# Method:    unmap_lun_from_igroup
# Objective: to unmap lun from igroup and update the
#            expected lun attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unmap_lun_from_igroup {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $igroup_obj, $igroup_name, $lun_attr_expected,
        %cmds_args, $verify);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /igroup_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Igroup/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $igroup_obj = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err(
                           "invalid \"-$key\": passed. Valid parameters: " .
                           "lun_attr_info|igroup_obj");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return $SanUtil::FALSE;
    }

    # verify mandatory cmd argument - igroup_obj
    unless (defined $args{'igroup_obj'}) {
        $self->_set_err("mandatory parameter \"-igroup_obj\" is needed");
        return $SanUtil::FALSE;
    }

    if (defined $verify) {
        unless ($verify =~ /true/) {
            $self->_set_err("verify parameter must be set to true");
            return $SanUtil::FALSE;
        }
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
        }
    } else {
        # we were passed in the attr hash, get the lun object from it
        $lun_obj = $lun_attr_expected->{'lun_obj'};
    }
    $lun_name = $lun_attr_expected->{'lun'};

    $igroup_name = $igroup_obj->get_name();
    if ($igroup_obj->error()) {
        $self->_set_err("Failed to get igroup name \nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # unmap lun object
    $lun_obj->unmap($igroup_name);
    if ($lun_obj->error()) {
        $self->_set_err("Failed to unmap " .
                        "lun($lun_name)\nReason:\n" . $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected lun attributes
    # This isn't yet implemented in lun show
    $lun_attr_expected->{'mapped'} = "unmapped";

    if ($verify) {
        # the user wants to verify that the move is reflected successfully
        # on the controller output
            $self->verify_lun_attributes(-lun_attr_info => $lun_attr_expected);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
    }

    # we need to update the LunAttrsArray that this library has stored
    $self->_update_attr_hash_array(-lun_attr_info => $lun_attr_expected);
    if ($self->error()) {
        return $SanUtil::FALSE
    }

    # dump the expected lun attributes for debug purpose
    if ($self->get_debug_level() >= 3) {

        # copy the attributes, since we need to delete the lun_obj,
        # dump will dump whole lun object also, so remove it.
        my %just_lun_attr = %{$lun_attr_expected};
        delete $just_lun_attr{'lun_obj'};
        $self->sanlog(-msg => "Expected lun Attributes " .
                      "after remove:\n" . $self->dumper_get(\%just_lun_attr));
    }
    return $SanUtil::TRUE;
}    # end sub unmap_lun_from_igroup

###############################################################################
# Method:    verify_lun_attributes
# Objective: to verify the lun attributes are intact.
#            compare the requested attributes with the controller
#            return lun attributes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_lun_attributes {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, %lun_attr_user, $lun_attr_controller, @err_str,
        $size_exact);
    my $err_cnt = 0;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /size_exact/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $size_exact = $value;
        } else {
            $self->_set_err("invalid \"-$key\": passed. " .
                            "Valid parameter: lun_attr_info");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info
    unless (defined $args{'lun_attr_info'}) {
        $self->_set_err("mandatory parameter \"-lun_attr_info\" is needed");
        return $SanUtil::FALSE;
    }

    # copy the attributes, since we need to delete the lun_obj,
    # do not operate on reference
    %lun_attr_user = %{$args{'lun_attr_info'}};

    # verify required "lun_obj" present in the "lun_attr_info"
    if (exists($lun_attr_user{'lun_obj'})
        and (ref($lun_attr_user{'lun_obj'})
                     =~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/)) {
        $lun_obj = $lun_attr_user{'lun_obj'};
    } else {
        $self->_set_err("lun_obj is not defined in the lun_attr_info");
        return $SanUtil::FALSE;
    }

    # since lun_obj is not part of lun attribute,
    # discard from $lun_attr_user
    delete $lun_attr_user{'lun_obj'};
    delete $lun_attr_user{'file_path'};

    # get the name of lun from "$lun_attr_user"
    $lun_name = $lun_attr_user{'lun'};

    # get the lun info from controller, return reference to hash
    $lun_attr_controller = $lun_obj->get_list_info();
    if ($lun_obj->error()) {
        $self->_set_err("Unable to get lun information." .
                           $lun_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # if zapi, do some post-processing manipulation of attributes
    if ($lun_obj->get_cmd_interface() =~ /zapi/i) {
        if ($self->_zapi_post_process_attrs(
                    -lun_attr_info => $lun_attr_controller) != $SanUtil::TRUE) {
            return $SanUtil::FALSE;
        }
    }

    # i think dev_id won't be returned from get-list-info, if not
    # need to add that query to this code in order to verify that
    # attribute

    # could also add individual checks in here for commands that return
    # a value when no param is present - i.e. verify that lun comment
    # returns the same value as lun list info

    # definitely should do this for serial numbers

    # display the user request lun attributes if debug_level is high
    $self->sanlog(-msg       => "verify_lun_attributes: Expected:\n" .
                                 $self->dumper_get(\%lun_attr_user),
                  -msg_level => 3);

    # display the controller lun attributes if debug_level is high
    $self->sanlog(-msg       => "verify_lun_attributes: Controller:\n" .
                                 $self->dumper_get($lun_attr_controller),
                  -msg_level => 3);

    # verify the lun attributes from user expected list
    # with controller list and return one error message with
    # list of attributes which is/are not matching. User can set
    # _SKIP_CHECK_ value for particular key to skip checking.
    @err_str = ("lun($lun_name) attributes does not match");
    foreach my $key (keys %lun_attr_user) {

        # if caller did not care about this particular attribute
        if ($lun_attr_user{$key} =~ /_SKIP_CHECK_/) {
            next;
        }
        if ($lun_attr_controller->{$key} =~ /zapi_na/) {
            next;
        }
        unless (defined $lun_attr_controller->{$key}) {
            $err_cnt++;
            push(@err_str, "\t$key is not defined in the controller list info");
            next;
        }

        # need to update this check to verify that the lun size created
        # is within one cylinder size of specified lun size
        # cylinder size depends on type and size of lun (we can get this
        # from lun geometry) - we should not be doing a blind > check here...
        if ($key =~ /size/) {
            if ($size_exact eq "true") {
                # the user wants to do an exact size match
                if ($lun_attr_user{$key} != $lun_attr_controller->{$key}) {
                    $err_cnt++;
                    push(@err_str,
                             "\t$key => Expected: "
                           . "'$lun_attr_user{$key}' <-> "
                           . "Controller: '$lun_attr_controller->{$key}'");
                }
            } else {
                # either size_exact wasn't specified or it was set to
                # something other than true
                if ($lun_attr_user{$key} > $lun_attr_controller->{$key}) {
                    $err_cnt++;
                    push(@err_str,
                             "\t$key => Expected: "
                           . "'$lun_attr_user{$key}' <-> "
                           . "Controller: '$lun_attr_controller->{$key}'");
                }
            }
        } else {
            if ($lun_attr_user{$key} ne $lun_attr_controller->{$key}) {
                $err_cnt++;
                push(@err_str,
                         "\t$key => Expected: "
                       . "'$lun_attr_user{$key}' <-> "
                       . "Controller: '$lun_attr_controller->{$key}'");
            }
        }
    }

    # TODO - verify with VDOM, BCOM KA cache
    if ($err_cnt) {
        my $text = "No. of errors encountered: $err_cnt\n" .
                                             join("\n", @err_str);
        $self->_set_err($text);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}    # end sub verify_lun_attributes

###############################################################################
# Method:    verify_luns_attributes
# Objective: To verify the lun attributes are intact.
# Details:   If verify fails it will destroy the lun and return $TCD::FAIL
###############################################################################
sub verify_luns_attributes {
    my ($self) = @_;
    my $err_str;

    $err_str       = "";
    foreach my $lun_attr_expected (@{$self->{LunAttrsArray}}) {
        $self->verify_lun_attributes('-lun_attr_info' => $lun_attr_expected);
        if ($self->error()) {
            if ($err_str ne "") {
                $err_str .= "\n";
            }
            $err_str .= $self->errmsg();
        }
    }

    if ($err_str ne "") {
        $self->_set_err($err_str);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}    # end sub verify_luns_attributes

###############################################################################
# Method:    build_store_luns_attributes
# Objective: Build the lun attribute hash(es) given a lun object or array of
#            lun objects.  If the attributes are found on the LunAttrsArray,
#            modify those attributes on the hash array, otherwise push the new hash
#            entries onto LunAttrsArray
# Details:   See POD documentation at the end of this file.
###############################################################################
sub build_store_luns_attributes {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr, $lun_attr_controller, @err_str,
        @lun_obj_array, @lun_attr_array);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_obj_array/) {
            if (ref($value) !~ /ARRAY/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "need an array of lun objects");
                return $SanUtil::FALSE;
            }
            @lun_obj_array = @$value;
        } elsif ($key =~ /lun_obj/) {
            if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with lun object");
                return $SanUtil::FALSE;
            }
            $lun_obj = $value;
        } else {
            $self->_set_err("invalid \"-$key\": passed. " .
                            "Valid parameter: lun_obj");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - either lun_obj or lun_obj_array
    unless (defined $args{'lun_obj'} xor defined $args{'lun_obj_array'}) {
        $self->_set_err("mandatory parameter either \"-lun_obj\" or \"-lun_obj_array\" is needed");
        return $SanUtil::FALSE;
    }

    if (@lun_obj_array) {
        # we were passed an array of lun objects to get the attrs for and
        # fill in the local attr hash array
        foreach my $lun_obj (@lun_obj_array) {
            # lun_obj was passed in - build a lun attribute hash from it
            $lun_attr = $self->build_lun_attributes(-lun_obj => $lun_obj);
            if ($self->error()) {
                return $SanUtil::FALSE;
            }
            # now we need to either add this to the array @{$self->{LunAttrsArray}}
            # if it doesn't exist on the array, or we need to replace the existing
            # array entry with this new updated entry
            $self->_update_attr_hash_array(-lun_attr_info => $lun_attr);
            if ($self->error()) {
                return $SanUtil::FALSE
            }
        }    # end of foreach my $lun_obj (@lun_obj_array)
    } else {
        # we were passed in a single lun object
        # lun_obj was passed in - build a lun attribute hash from it
        $lun_attr = $self->build_lun_attributes(-lun_obj => $lun_obj);
        if ($self->error()) {
            return $SanUtil::FALSE;
        }
        # now we need to either add this to the array @{$self->{LunAttrsArray}}
        # if it doesn't exist on the array, or we need to replace the existing
        # array entry with this new updated entry
        $self->_update_attr_hash_array(-lun_attr_info => $lun_attr);
        if ($self->error()) {
            return $SanUtil::FALSE
        }
    }     # end of if (@lun_obj_array) / else
}    # end sub build_store_luns_attributes

###############################################################################
# Method:    build_lun_attributes
# Objective: Build the lun attributes hash for a lun given a lun object
#            Return lun attribute hash
# Details:   See POD documentation at the end of this file.
###############################################################################
sub build_lun_attributes {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr, $lun_attr_controller, @err_str);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_obj/) {
            if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with lun object");
                return {};
            }
        } else {
            $self->_set_err("invalid \"-$key\": passed. " .
                            "Valid parameter: lun_obj");
            return {};
        }
    }

    # verify mandatory cmd argument - lun_obj
    unless (defined $args{'lun_obj'}) {
        $self->_set_err("mandatory parameter \"-lun_obj\" is needed");
        return {};
    }

    # copy the attributes, since we need to delete the lun_obj,
    # do not operate on reference
    $lun_obj = $args{'lun_obj'};

    # get the lun info from controller, return reference to hash
    $lun_attr_controller = $lun_obj->get_list_info();
    if ($lun_obj->error()) {
        $self->_set_err("Unable to get lun information." .
                           $lun_obj->errmsg());
        return {};
    }

    # i think dev_id won't be returned from get-list-info, if not
    # need to add that query to this code in order to add that
    # attribute

    # display the controller lun attributes if debug_level is high
    $self->sanlog(-msg       => "build_lun_attributes: Controller:\n" .
                                 $self->dumper_get($lun_attr_controller),
                  -msg_level => 3);

    # copy the lun attributes from the controller hash to the hash
    # that we will return to the user
    foreach my $key (keys %$lun_attr_controller) {
        ($lun_attr->{$key} = $lun_attr_controller->{$key})
    }
    $lun_attr->{'lun_obj'} = $lun_obj;

    return $lun_attr;
}    # end sub build_lun_attributes

###############################################################################
# Method:    update_lun_attributes
# Objective: Update the expected lun attr hash for a lun.  This takes a lun object
#            or attribute hash and the attribute(s) to be updated.  Return the
#            updated hash to the caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub update_lun_attributes {
    my ($self, %args) = @_;

    my ($lun_obj, $lun_name, $lun_attr, $lun_attr_controller, @err_str,
        $lun_attr_expected);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return {};
            }
            $lun_attr_expected = $value;
        } elsif ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return {};
            }
            $lun_obj = $value;
        } elsif ($key =~ /size|lun_name|ostype|path|space-reserve|state|mapped|serial|volume/) {
            # not yet supporting vserver, uuid, dev_id changes
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return {};
            }
        } else {
            $self->_set_err("invalid \"-$key\": passed. " .
                            "Valid parameter: lun_obj");
            return {};
        }
    }

    # verify mandatory cmd argument - lun_attr_info or lun_obj
    unless ((defined $args{'lun_attr_info'}) xor (defined $args{'lun_obj'})) {
        $self->_set_err("mandatory parameter either \"-lun_attr_info\" " .
                        "or \"-lun_obj\" is required");
        return {};
    }

    if ($lun_obj) {
        # lun_obj was passed in - we either find the existing attr hash or build one
        unless ($lun_attr_expected = $self->_find_attr_hash_array(-lun_obj => $lun_obj)) {
            $lun_attr_expected = $self->build_lun_attributes(-lun_obj => $lun_obj);
            if ($self->error()) {
                return {};
            }
        }
    }

    # copy the lun attributes from the arguments to the hash
    # that we will return to the user
    foreach my $lun_attr_name (@ExpectedLunAttr) {
        ($lun_attr_expected->{$lun_attr_name} = $args{$lun_attr_name})
    }

    return $lun_attr_expected;

}    # end sub update_lun_attributes

###############################################################################
# Method:    _update_attr_hash_array
# Objective: Find the attr hash info in the array @{$self->{LunAttrsArray}}
#            and update it with the attr hash supplied.  if the attr hash
#            is not found in the LunAttrsArray, return false.  if the attr
#            hash is found and updated, return true
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _update_attr_hash_array {
    my ($self, %args) = @_;

    my ($lun_obj, $found, $lun_attr_info);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attr_info = $value;
        } else {
            $self->_set_err("invalid \"-$key\": passed. " .
                            "Valid parameter: lun_attr_info");
            return $SanUtil::FALSE;
        }
    }

    # verify mandatory cmd argument - lun_attr_info
    unless (defined $args{'lun_attr_info'}) {
        $self->_set_err("mandatory parameter \"-lun_attr_info\" is needed");
        return $SanUtil::FALSE;
    }

    # because any of the lun attributes could have changed in this update
    # we are going to search on the lun object in the hash array
    # this will build a temp array in order to replace the array entry if it
    # is found
    $lun_obj = $lun_attr_info->{'lun_obj'};
    my @lun_attr_array_temp;
    foreach my $lun_attr_temp (@{$self->{LunAttrsArray}}) {
        if ($lun_obj eq $lun_attr_temp->{'lun_obj'}) {
            # we need to replace the array entry with the attr info supplied
            push @lun_attr_array_temp, $lun_attr_info;
            $found = 1;
        } else {
            # this array entry didn't match, so push this existing entry onto the temp array
            push @lun_attr_array_temp, $lun_attr_temp;
        }
    }    # end of foreach my $lun_attr_temp (@{$self->{LunAttrsArray}})

    # now replace the global LunAttrsArray with the updated array we just build
    @{$self->{LunAttrsArray}} = ();
    @{$self->{LunAttrsArray}} = @lun_attr_array_temp;

    if ($found) {
        # we found and replaced an entry, return TRUE
        return $SanUtil::TRUE;
    } else {
        # we did not replace an entry, so push the lun_attr_info passed in
        # onto the LunAttrsArray and return FALSE
        push @{$self->{LunAttrsArray}}, $lun_attr_info;
        return $SanUtil::FALSE;
    }
}    # end of sub _update_attr_hash_array

###############################################################################
# Method:    _find_attr_hash_array
# Objective: Find the attr hash info in the array @{$self->{LunAttrsArray}}
#            If the attr hash is not found in the LunAttrsArray, return empty
#            hash.  If the attr hash is found, return it
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _find_attr_hash_array {
    my ($self, %args) = @_;

    my ($lun_obj, $found, $lun_attr_info);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_obj/) {
           if (ref($value) !~ /San::ControllerCluster::Filer10M::Vserver::Volume::Lun/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return {};
            }
            $lun_obj = $value;
        } else {
            $self->_set_err("invalid \"-$key\": passed. " .
                            "Valid parameter: lun_attr_info");
            return {};
        }
    }

    # verify mandatory cmd argument - lun_obj
    unless (defined $args{'lun_obj'}) {
        $self->_set_err("mandatory parameter \"-lun_obj\" is required");
        return {};
    }

    # search on the lun object in the hash array
    foreach my $lun_attr_temp (@{$self->{LunAttrsArray}}) {
        if ($lun_obj eq $lun_attr_temp->{'lun_obj'}) {
            # this hash matches the lun object supplied, return it
            return $lun_attr_temp;
        }
    }    # end of foreach my $lun_attr_temp (@{$self->{LunAttrsArray}})

    # we didn't find the hash above, return an empty hash
    return {};

}    # end of sub _find_attr_hash_array

###############################################################################
# Method:    _cli_to_zapi
# Objective: Pre-process the CLI arguments so that they're ZAPI-like.  The
#            The goal is to isolate these changes here (as much as possible)
#            as ZAPI handling is still evolving and subject to change.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cli_to_zapi {
    my ($self, %args) = @_;
    my $cmd_args;

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /cmd_args/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $cmd_args = $value;
        } else {
            $self->_set_err("invalid \"-$key\": passed. " .
                            "Valid parameter: lun_attr_info");
            return $SanUtil::FALSE;
        }
    }

    if (defined($cmd_args->{'space-reserve'})) {
        if ($cmd_args->{'space-reserve'} eq 'enabled') {
            $cmd_args->{'space-reservation-enabled'} = 'true'
        } else {
            $cmd_args->{'space-reservation-enabled'} = 'false';
        }
        delete $cmd_args->{'space-reserve'};
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    _zapi_post_process_attrs
# Objective: Post-process the attributes returned by ZAPI's lun-get-iter
#            so that they're CLI-like.  The goal is to isolate these changes
#            here (as much as possible) as ZAPI handling is still evolving
#            and subject to change.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _zapi_post_process_attrs {
    my ($self, %args) = @_;
    my $lun_attrs;

    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value " .
                                "Need with HASH reference to lun info");
                return $SanUtil::FALSE;
            }
            $lun_attrs = $value;
        } else {
            $self->_set_err("invalid \"-$key\": passed. " .
                            "Valid parameter: lun_attr_info");
            return $SanUtil::FALSE;
        }
    }

    # extract the vol, qtree, and lun name from the path
    my @fields = split(/\//, $lun_attrs->{path});
    $lun_attrs->{volume} = $fields[2];
    if (scalar(@fields) == 5) {
        $lun_attrs->{qtree} = $fields[3];
        $lun_attrs->{lun} = $fields[4];
    } else {
        $lun_attrs->{qtree} = '""';
        $lun_attrs->{lun} = $fields[3];
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    execute
# Objective: Execute a controller command directly and return the output
# Details:   See POD documentation at the end of this file.
###############################################################################
sub execute {
    my ($self, %args) = @_;
    my $cmd;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /cmd/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return {};    # a reference to an empty hash
            }
            $cmd = $value;
        } else {
            $self->_set_err(
                 "invalid \"-$key\": passed. Valid parameters: cmd");
            return {};
        }
    }

    # verify mandatory argument -cmd
    unless (defined $args{'cmd'}) {
        $self->_set_err("mandatory parameter \"-cmd\" is needed");
        return {};
    }

    return $self->{'cserver_obj'}->execute(-cmd => $cmd, -raw_data => 1);
}

###############################################################################
# Method:    get_size_info
# Objective: To get the interesting size statistics from the system
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_size_info {
    my ($self, %args) = @_;
    my $lun;
    my %size_info;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lun_attr_info/) {
            if (ref($value) !~ /HASH/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return {};    # a reference to an empty hash
            }
            $lun = $value;
        } else {
            $self->_set_err(
                 "invalid \"-$key\": passed. Valid parameters: lun_attr_info");
            return {};
        }
    }

    unless (defined($lun)) {
        $self->_set_err("Parameter -lun_attr_info is mandatory");
        return {};
    }

    my $lun_obj = $lun->{lun_obj};
    my $parent_obj = $lun_obj->get_parent_obj();
    my $vol_obj = (ref($parent_obj) =~ /Qtree/i)
        ? $parent_obj->get_parent_obj()
        : $parent_obj;
    
    # get lun geometry info
    my $lun_geometry = $lun_obj->get_lun_geometry();
    if ($lun_obj->error()) {
        $self->_set_err($lun_obj->errmsg());
        return {};
    }
    $size_info{lun_size} = $lun_geometry->{size};
    $size_info{lun_max_resize} = $lun_geometry->{max_resize_size};

    # CLI reports cylinder size directly; ZAPI doesn't.
    $size_info{lun_cyl_size} = ($vol_obj->get_cmd_interface() =~ /zapi/i)
        ? $lun_geometry->{size} / $lun_geometry->{cylinders}
        : $lun_geometry->{cyl_size};

    # get lun maxsize
    my $lun_maxsize = $self->get_lun_maxsize(-container_obj => $vol_obj,
                                             -ostype => $lun->{ostype});
    if ($self->error()) {
        $self->_set_err($self->errmsg());
        return {};
    }
    # perhaps we should be using the with_ss_reserve value? or both?
    $size_info{lun_maxsize} = $lun_maxsize->{without_ss_reserve};

    # get vol space available
    $size_info{vol_avail} = $vol_obj->get_available();
    if ($vol_obj->error()) {
        $self->_set_err($vol_obj->errmsg());
        return {};
    }

    return \%size_info;
}

###############################################################################
# Method:    host_discover_luns
# Objective: To discover luns from the host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub host_discover_luns {
    my ($self, %args) = @_;
    my $scsit_obj;
    my $luns;
    my $verify = 'false';

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /scsit_obj/) {
            if (ref($value) !~ /San::SanScsiT::XScsiT/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $scsit_obj = $value;
        } elsif ($key =~ /luns/) {
            if (ref($value) !~ /ARRAY/) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $luns = $value;
        } elsif ($key =~ /verify/) {
            if (!length($value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
            $verify = $value;
        } else {
            $self->_set_err(
                 "invalid \"-$key\": passed. Valid parameters: scsit_obj|luns|verify");
            return $SanUtil::FALSE;
        }
    }

    unless (defined($scsit_obj) and defined($luns)) {
        $self->_set_err("Parameters scsit_obj and luns are mandatory");
            return $SanUtil::FALSE;
    }

    my $path_info = $scsit_obj->discover_paths(-lun_objs => $luns);
    if ($scsit_obj->error()) {
        $self->_set_err($scsit_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # XXX Need an API in lieu of this as the data structure format could
    # change underneath us.

    if ($verify eq 'true') {
        # For now, the assumption for correct behavior is that every lun has
        # at least one path to it.  If not, return an error.  At some point,
        # we may want to adjust this thinking.

        foreach my $lun (@{$luns}) {
            my $vol = $lun->get_parent_obj();
            $vol = $vol->get_parent_obj() if (ref($vol) =~ /Qtree/);
            my $vol_name = $vol->get_name();
            my $vs_name = $vol->get_vserver();
            my $lun_path = $lun->get_path();
            my $host_paths = $path_info->{$vs_name}{$vol_name}{$lun_path}{path_info};
            if (scalar(%{$host_paths}) == 0) {
                $self->_set_err("No paths found for $lun_path");
                return $SanUtil::FALSE;
            }
        }
    }

    return $SanUtil::TRUE;
}

1;

##############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

LunHelper.pm

=head1 Author

Ann Harrold (harrold@netapp.com)

=head1 Description

LunHelper.pm provides an utility function interface to test Lun functionality
with the main intent to avoid repeated code from thpl level. It is design as
Object Oriented model. Other object(s) can derive this as base class or user can
create object from it.

=head1 See Also

LunHelper.pm package is used as base class from San::Application::LunMgr. See San::Application::LunMgr
POD for more details.

B<San::Application::LunMgr>

=head1 Attributes

None

=head1 Revision History

Nov-04-2009 harrold@netapp.com: Copied template from Portset
May-10-2010 harrold@netapp.com: modified after initial code review

=head1 Public Methods

=head2 B<new()>

Class Constructor method. To set the debug level to print debug message at particular level, call
'set_debug_level' method. See Example for more details

=over 4

=item Inputs

None

=item Outputs

Returns a LunHelper blessed object reference.

=item Examples

   #
   # Create a LunHelper object
   #
   my $lun_helper_obj = new LunHelper();
   if ($lun_helper_obj->error()) {
      my $err_msg = $lun_helper_obj->errmsg();
      $Test->nlog("Failed to instantiate LunHelper object:\nReason: $err_msg");
      return $TCD::FAIL;
   }
   #
   # set the debug level to print debug message at particular level
   #
   $lun_helper_obj->set_debug_level($debug_level);

=back

=head2 B<create_lun(%args)>

Create a Lun in a given volume and return the hash reference to expected Lun attributes
for later verification.

=over 4

=item Inputs

=item * I<-volume_obj> <required>

The volume object where lun will be created.

=item * I<-lun_name> <required>

Name of the lun to create

=item * I<-size> <required>

Size of the lun to create

=item * I<-os_type> <required>

Name of the lun os type. This should be one of the valid os types, but this
value is not verified, so it can be used to test invalid os types.

=item * I<-path> <optional>

Fully specified path in which to create the lun.  Used for creating luns
in qtrees.

=item * I<-space-reserve> <optional>

Controls whether to create space reserved luns.  Valid values are "noreserve"

=item Outputs

Returns a reference to a hash containing lun expected attributes info if successful.
If an error occurred the error flag is set, errmsg is set to the failure reason,
and {} (empty hash reference) is passed back as the return value.

Expected attributes will simulate what the controller return for lun attribute.
lun show command will have following attributes for given lun.
    vserver|lun-path|volume|lun|size|ostype|space-reserve|state|uuid|serial|

if any attribute is not used fill with "-" as default, since lun show output will have "-"


Sample output
Virtual Server Name|LUN Path|Volume Name|LUN Name|LUN size|OS type of the LUN|Space Reservation|LUN State|UUID of the LUN|LUN Serial Number|
lun_vser|/vol/lun_vs_vol4/lun_create_default_10|lun_vs_vol4|lun_create_default_10|10485760|linux|-|online|df1f01af-ebda-11de-aaf1-123478563412|W-E3GJTk3W-g|

Hash reference looks like following:

  $lun_attr_expected->{'vserver'}
  $lun_attr_expected->{'path'}
  $lun_attr_expected->{'lun'}
  $lun_attr_expected->{'volume'}
  $lun_attr_expected->{'size'}
  $lun_attr_expected->{'ostype'}
  $lun_attr_expected->{'space_resv_enabled'}
  $lun_attr_expected->{'state'}
  $lun_attr_expected->{'mapped'}
  $lun_attr_expected->{'uuid'}
  $lun_attr_expected->{'serial'}
  $lun_attr_expected->{'device_binary_id'}
  $lun_attr_expected->{'device_text_id'}

  additional info, which is not part of lun attributes, but we need this for later verification
  $lun_attr_expected->{'lun_obj'} - reference to lun object

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      # get vserver object
      $vs_obj = ...;

      $cmd_args{'-volume_obj'} = $volume_obj;
      $cmd_args{'-lun_name'}   = $lun_name;
      $cmd_args{'-size'}       = $lun_size;
      $cmd_args{'-ostype'}     = $os_type;
      if ($space_reserve) {
          $cmd_args{"-space-reserve"} = $space_reserve;
      }
      my $lun_attr_expected = $self->create_lun(%cmd_args);

      if ($self->error()) {
          return ();
      }

=back

=head2 B<map_lun_to_igroup(%args)>

Map a lun in a given volume to the igroup specified.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item * I<-igroup_obj> <required>

The igroup object which the lun will be mapped to

=item * I<-lun_id> <optional>

The lun id at which to map the lun.  If this is not specified, the controller
will autoassign a lun id.  This should be a number between 1 and 4095,
although this routine will not check the value.  This allows for the use of
this method to test invalid lun ids.

=item Outputs

Returns TRUE if the lun was successfully mapped, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $self->map_lun_to_igroup(-lun_attr_info => $lun_attr_expected,
                               -igroup_obj    => $igroup_obj);
      if ($self->error()) {
          return ();
      }

=back

=head2 B<move_lun(%args)>

Move a lun in a given volume to the new path specified

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item * I<-path> <required>

The new path of the lun

=item Outputs

Returns TRUE if the lun was successfully moved, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      my $lun_src_path = $lun_info->{'path'};
      $lun_src_path =~ /(\S+)\/(\S+)/;
      my $lun_dest_path = $1 . "/" . $2 . "_moved";
      $LUNMgr_obj->move_lun('-lun_attr_info' => $lun_info,
                            '-path'          => $lun_dest_path);

=back

=head2 B<offline_lun(%args)>

Offline a lun in a given volume

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item Outputs

Returns TRUE if the lun was successfully offlined, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $LUNMgr_obj->offline_lun('-lun_attr_info' => $lun_info);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<online_lun(%args)>

Online a lun in a given volume

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item Outputs

Returns TRUE if the lun was successfully onlined, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $LUNMgr_obj->online_lun('-lun_attr_info' => $lun_info);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<resize_lun(%args)>

Resize a lun in a given volume to the specified size with the option to resize
lun prefix or suffix.  Optionally a force parameter can be used.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item * I<-size> <required>

The new size of the lun to be used

=item * I<-force> <optional>

A flag to control whether to force the resize command.  Valid value is
"true".

=item * I<-option> <optional>

A flag to control whether to resize the prefix or suffix of the lun.  Valid
values are "prefix" and "suffix".

=item Outputs

Returns TRUE if the lun was successfully resized, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      my $new_size = $LunSize + $vol_available + 1024;
      $LUNMgr_obj->resize_lun('-lun_attr_info' => $lun_info,
                              '-size'          => $new_size);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<set_lun_comment(%args)>

Set the comment on a lun in a given volume to the specified string.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item * I<-comment> <optional>

The string to set the comment to.

=item Outputs

Returns TRUE if the lun comment was successfully set, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      my $comment = "This is a new comment";
      $LUNMgr_obj->set_lun_comment('-lun_attr_info' => $lun_info,
                                   '-comment'       => $comment);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<set_lun_dev_id(%args)>

Set the device id on a lun in a given volume to the specified value.  A lun
can have either a text dev id or a binary dev id or both.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item * I<-dev_id_binary> <optional>

The value to set the binary device id to

=item * I<-dev_id_text> <optional>

The value to set the text device id to

=item * I<-dev_id_none> <optional>

Pass in this flag if you wish to clear out the device id.

=item Outputs

Returns TRUE if the lun dev id was successfully set, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $LUNMgr_obj->set_lun_dev_id('-lun_attr_info' => $lun_info,
                                  '-dev_id_binary' => "10");
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<set_lun_reservation(%args)>

Set the space reservation on a lun in a given volume to the specified value.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item * I<-space-reserve> <required>

The value to set the space reservation to.  Valid values are "enable" and "disable"

=item Outputs

Returns TRUE if the lun reservation was successfully set, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $LUNMgr_obj->set_lun_reservation('-lun_attr_info' => $lun_info,
                                       '-space-reserve'    => "enable");
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<set_lun_serial(%args)>

Set the serial number on a lun in a given volume to the specified value.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item * I<-serial> <required>

The value to set the serial number to.

=item Outputs

Returns TRUE if the lun serial number was successfully set, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      my $lun_serial = random_string("CcnncCcnCcCn");

      $LUNMgr_obj->set_lun_serial('-lun_attr_info' => $lun_info,
                                  '-serial'       => $lun_serial);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<test_lun_comment(%args)>

Tests the the comment can be set and unset on a lun.  This method will first
verify that the comment set in the attribute hash matches what is returned from
the controller.  Next it will set the comment to some string and verify it is
set correctly.  After that, it will clear the comment by setting it to "", and
verify it is set correctly.  Lastly, it will set the comment back to what it
was when this method was called.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item Outputs

Returns TRUE if the lun comment tested correctly, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $self->test_lun_comment(-lun_attr_info => $lun_attr_expected);
      if ($self->error()) {
          return $SanUtil::FALSE;
      }

=back

=head2 B<test_lun_map(%args)>

Tests that a lun can be mapped and unmapped. This method will first determine
whether the lun is mapped.  If it is mapped, this method will unmap it from all
tne igroups it is mapped to, verify it is unmapped correctly, then remap it
to all the igroups and verify.  If the lun is unmapped, it will create an igroup,
map the lun to the igroup, verify, then unmap, verify and destroy the igroup.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item Outputs

Returns TRUE if the lun map tested correctly, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $self->test_lun_map(-lun_attr_info => $lun_attr_expected);
      if ($self->error()) {
          return $SanUtil::FALSE;
      }

=back

==head2 B<test_lun_move(%args)>

Tests that a lun can be moved. This method will first determine get the current
path of the lun.  It will then move it to a new name in the current location
and verify the move.  Then, it will restore the lun to it's original location
and verify.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item Outputs

Returns TRUE if the lun move tested correctly, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $self->test_lun_move(-lun_attr_info => $lun_attr_expected);
      if ($self->error()) {
          return $SanUtil::FALSE;
      }

=back

==head2 B<test_lun_offline_online(%args)>

Tests that a lun can be offlined/onlined. This method will first determine get
the current state of the lun.  If it is online, it will offline then online the
lun - verifying the state is correct following each operation.  If the lun is
offline, it will online then offline the lun, verifying the state is correct
following each operation.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item Outputs

Returns TRUE if the lun move tested correctly, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $self->test_lun_offline_online(-lun_attr_info => $lun_attr_expected);
      if ($self->error()) {
          return $SanUtil::FALSE;
      }

=back

head2 B<unmap_lun_from_igroup(%args)>

Unmap a lun in a given volume from the igroup specified.

=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

A hash of lun attributes

=item * I<-igroup_obj> <required>

The igroup object which the lun will be unmapped from

=item Outputs

Returns TRUE if the lun was successfully unmapped, otherwise returns FALSE, error
flag is set and errmsg is set to the failure reason.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object as below
      #

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup(...);
      if ($LUNMgr_obj->error) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      $self->unmap_lun_from_igroup(-lun_attr_info => $lun_attr_expected,
                                   -igroup_obj    => $igroup_obj);
      if ($self->error()) {
          return ();
      }

=back

=head2 B<verify_lun_attributes(%args)>

To verify the expected lun attributes match the attributes returned from the
controller.

NOTE;

    # verify the lun attributes from user expected list
    # with controller list and return one error message with  
    # list of attributes which is/are not matching. User can set 
    # _SKIP_CHECK_ value for particular key to skip checking.


=over 4

=item Inputs

=item * I<-lun_attr_info> <required>

Hash reference to Lun info, basically expects same info which return from
the function 'create_lun_in_vserver'. See create_lun_in_vserver POD for more details

Hash reference looks like following:
  $lun_attr_expected->{'lun_obj'} - reference to Lun object, this attribute is MUST.
  rest are optional or can skip the checking by setting the value of attribute to  _SKIP_CHECK_
  $lun_attr_expected->{'vserver'}
  $lun_attr_expected->{'path'}
  $lun_attr_expected->{'lun'}
  $lun_attr_expected->{'volume'}
  $lun_attr_expected->{'size'}
  $lun_attr_expected->{'ostype'}
  $lun_attr_expected->{'space_resv_enabled'}
  $lun_attr_expected->{'state'}
  $lun_attr_expected->{'mapped'}
  $lun_attr_expected->{'uuid'}
  $lun_attr_expected->{'serial'}
  $lun_attr_expected->{'device_binary_id'}
  $lun_attr_expected->{'device_text_id'}

Caller can set value _SKIP_CHECK_ if does not want to check particular attribute.
Ex:
   $lun_attr_expected->{'uuid'} = '_SKIP_CHECK_';

=item * I<-size_exact> <optional>

Flag to control whether to do an exact size match on the LUN being verified.
If the called knows that the LUN attribute hash is in such a state that the
size is exactly correct, this should be set to "true".  Otherwise, the size
will be verified within the allowable range.  This currently is understood
to be one cylinder size.

=item Outputs

Returns TRUE  if all expected Lun attributes matches, if successful.
If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object (see above create_lun_in_vserver() pod)
      #
      ....
      ....
      ....
      my $lun_name = $lun_name_prefix . "_unmapped_" . $os_type . $lun_count;
      my %cmd_args;
      $cmd_args{'-volume_obj'} = $volume_obj;
      $cmd_args{'-lun_name'}   = $lun_name;
      $cmd_args{'-size'}       = $lun_size;
      $cmd_args{'-ostype'}     = $os_type;
      if ($space_reserve) {
          $cmd_args{"-space-reserve"} = $space_reserve;
      }
      my $lun_attr_expected =
      $self->create_lun(%cmd_args);
      if ($self->error()) {
          # function will destroy the lun(s) and log the failure
          #return _cleanup_tc_luns('-err_msg'       => $self->errmsg());
          return ();
      }

      #
      # verify the lun attributes against controller
      #

      $LUNMgr_obj->verify_lun_attributes('-lun_attr_info' => $lun_attr_expected);
      if ($LUNMgr_obj->error()) {
         $Test->nlog("Failed to verify: " . $LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<verify_luns_attributes(%args)>

To verify the expected luns attributes match the attributes returned from the
controller.  The luns must be created using LunMgr::create_tc_luns. The attribute
hashes from those luns are placed into an array called LunAttrsArray.  This method,
verify_luns_attributes, walks through that array and verifies the attributes
hash for each of those luns.

=over 4

=item Inputs

None

=item Outputs

Returns TRUE  if all expected lun attributes matches, if successful.
If an error occurred the error flag is set, errmsg is set to the failure reason,
and FALSE is passed back as the return value.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object (see above create_lun_in_vserver() pod)
      #
      ....
      ....
      ....
      my $LunsPerOsType = 10;
      $LUNMgr_obj->create_tc_luns(-volumes         => \@Volumes,
                                  -luns_per_ostype => $LunsPerOsType,
                                  -lun_size        => $LunSize,
                                  -lun_name_prefix => "lun_move");

      #
      # verify the lun attributes against controller
      #

      $LUNMgr_obj->verify_luns_attributes();
      if ($LUNMgr_obj->error()) {
         $Test->nlog("Failed to verify: " . $LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<build_lun_attributes(%args)>

This method will build a lun attribute hash given a lun object.  It returns
the lun attributes hash to the caller if it was successful.  Otherwise, it
returns an empty hash and sets errmsg to the error string.  This method does
not update the local lun attribute hash array.

=over 4

=item Inputs

=item * I<-lun_obj> <required>

Lun object which will be used to gather the lun attributes.

=item Outputs

Returns lun attribute hash if successful.  If a failure is detected, returns
an empty hash.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object (see above create_lun_in_vserver() pod)
      #
      ....
      ....
      ....
      $cmds_args{'size'} = $size;
      $cmds_args{'ostype'} = $ostype;
      $cmds_args{'volume'} = $vol_name;

      # create lun object
      my $lun_obj = $vol_obj->create_lun_obj(%cmds_args);
      if ($vol_obj->error()) {
          $Test->nlog("Unable to create lun: " . $vol_obj->error());
          return $TCD::FAIL;
      }

      #
      # build the lun attribute hash from the returned object
      #

      my $lun_attr_hash = $LunMgr_obj->build_lun_attributes(-lun_obj => $lun_obj);
      if ($LunMgr_obj->error()) {
         $Test->nlog("Failed to build lun attributes: " . $LunMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<build_store_luns_attributes(%args)>

This method will build lun attribute hash(es) given a lun object or array of
lun objects.  In addition, it will update the local lun attribute hash array
with the It returns the lun attributes hash to the caller if it was successful.  Otherwise, it
returns an empty hash and sets errmsg to the error string.  This method does
not update the local lun attribute hash array.

=over 4

=item Inputs

=item * I<-lun_obj> <required>

Lun object which will be used to gather the lun attributes.

=item Outputs

Returns lun attribute hash if successful.  If a failure is detected, returns
an empty hash.

=item Examples

      #
      # Instantiate LunHelper Object (see above _new() pod)
      # or use LunMgr Object (see above create_lun_in_vserver() pod)
      #
      ....
      ....
      ....
      $cmds_args{'size'} = $size;
      $cmds_args{'ostype'} = $ostype;
      $cmds_args{'volume'} = $vol_name;

      # create lun object
      my $lun_obj = $vol_obj->create_lun_obj(%cmds_args);
      if ($vol_obj->error()) {
          $Test->nlog("Unable to create lun: " . $vol_obj->error());
          return $TCD::FAIL;
      }

      #
      # build the lun attribute hash from the returned object
      #

      my $lun_attr_hash = $LunMgr_obj->build_lun_attributes(-lun_obj => $lun_obj);
      if ($LunMgr_obj->error()) {
         $Test->nlog("Failed to build lun attributes: " . $LunMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back


=cut

