# -*- perl -*-
#  $Id: $
#
# Copyright (c) 2009 NetApp, Inc.
# All rights reserved.
#
##
## @summary Utility for testing volume snapshot restore functionality
## To provide library functions for the testing the different ways by which a
## lun may be restored, specifically, partial file snapshot restore (pfsr),
## single file snapshot restore (sfsr), and volume restore (vsr),
## both disruptive and non-disruptive.
## All the cmode lun restore testing scripts will use this library
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::Task::LunRestoreHelper;

# set compiler directives
use strict;
require Exporter;
use Readonly;

# load modules
use File::Basename;
use Params::Validate qw( :all );
use San::Application::Lunconstant;
use San::SanUtil;
use San::Task::LunHelper;
use Data::Dumper;

# inherit base class
our @ISA = qw( SanUtil
               San::Task::LunHelper
);
    
our @EXPORT = qw(
    $SNAPSHOT_IN_USE_ERROR
    $SNAPSHOT_BACKED_BY_CLONES_ERROR
);

Readonly our $SNAPSHOT_IN_USE_ERROR  => 'Snapshot(\s+)is(\s+)in(\s+)use';
Readonly our $SNAPSHOT_BACKED_BY_CLONES_ERROR  => 'Volume(\s+)has(\s+)clones(\s+)backed(\s+)by(\s+)snapshots';

# LUN attributes inherited from src LUN for inplace sfsr
my @InplaceLunAttrsInheritedFromSrcLun = (
    $LUN_ATTR_PREFIX_SIZE, $LUN_ATTR_SIZE,       $LUN_ATTR_SUFFIX_SIZE,
    $LUN_ATTR_EXTENT_SIZE, $LUN_ATTR_BLOCK_SIZE, $LUN_ATTR_SIZE_USED,
    $LUN_ATTR_COMMENT,     $LUN_ATTR_SPACE_RESV, $LUN_ATTR_CREATION_TIME,
    $LUN_ATTR_READ_ONLY,   $LUN_ATTR_OSTYPE,     $LUN_ATTR_CLASS,
    $LUN_ATTR_RESTORE_INACCESSIBLE
);

# LUN attributes inherited from dst LUN for inplace sfsr
my @InplaceLunAttrsInheritedFromDstLun = (
    $LUN_ATTR_VSERVER,            $LUN_ATTR_VOLUME,
    $LUN_ATTR_QTREE,              $LUN_ATTR_PATH,
    $LUN_ATTR_NAME,               $LUN_ATTR_SERIAL,
    $LUN_ATTR_UUID,               $LUN_ATTR_VDISK_ID,
    $LUN_ATTR_STATE,              $LUN_ATTR_MAPPED,
    $LUN_ATTR_NVFAIL,             $LUN_ATTR_SPACE_RESV_HONORED,
    $LUN_ATTR_SPACE_ALLOC,        $LUN_ATTR_DEVICE_TEXT_ID,
    $LUN_ATTR_DEVICE_BINARY_ID,   $LUN_ATTR_DEVICE_LEGACY_ID,
    $LUN_ATTR_INCONSIST_FILE_SYS, $LUN_ATTR_INCONSIST_BLOCKS,
    $LUN_ATTR_OBJ,                $LUN_ATTR_MAX_RESIZE_SIZE,
    $LUN_ATTR_IS_CLONE,           $LUN_ATTR_IS_CLONE_AUTODELETE_ENABLED,
    $LUN_ATTR_HAS_METADATA_PROVISIONED, $LUN_ATTR_QOS_POLICY_GROUP,
    $LUN_ATTR_EXTERNAL_CACHE,     $LUN_ATTR_ALIGNMENT
);

# LUN attributes inherited from src LUN for out of place sfsr
my @OutOfPlaceLunAttrsInheritedFromSrcLun = (
    $LUN_ATTR_PREFIX_SIZE, $LUN_ATTR_SIZE,       $LUN_ATTR_SUFFIX_SIZE,
    $LUN_ATTR_EXTENT_SIZE, $LUN_ATTR_BLOCK_SIZE, $LUN_ATTR_SIZE_USED,
    $LUN_ATTR_COMMENT,     $LUN_ATTR_SPACE_RESV, $LUN_ATTR_CREATION_TIME,
    $LUN_ATTR_OSTYPE,      $LUN_ATTR_RESTORE_INACCESSIBLE,
    $LUN_ATTR_CLASS
);

# LUN attributes inherited from dst LUN for out of place sfsr
my @OutOfPlaceLunAttrsInheritedFromDstLun = (
    $LUN_ATTR_VSERVER, $LUN_ATTR_PATH, $LUN_ATTR_VOLUME,
    $LUN_ATTR_QTREE,   $LUN_ATTR_NAME, $LUN_ATTR_MAX_RESIZE_SIZE,
    $LUN_ATTR_HAS_METADATA_PROVISIONED, $LUN_ATTR_QOS_POLICY_GROUP,
    $LUN_ATTR_EXTERNAL_CACHE, $LUN_ATTR_ALIGNMENT
);

# LUN attributes initialized on dst LUN for out of place sfsr
my %OutOfPlaceInitializedLunAttrs = (
    $LUN_ATTR_STATE              => $LUN_ATTR_STATE_ONLINE,
    $LUN_ATTR_MAPPED             => $LUN_ATTR_MAPPED_UNMAPPED,
    $LUN_ATTR_READ_ONLY          => $LUN_ATTR_READ_ONLY_FALSE,
    $LUN_ATTR_SPACE_ALLOC        => $LUN_ATTR_SPACE_ALLOC_DISABLE,
    $LUN_ATTR_SPACE_RESV_HONORED => $LUN_ATTR_SPACE_RESV_HONORED_TRUE,
    $LUN_ATTR_NVFAIL             => $LUN_ATTR_NVFAIL_FALSE,
    $LUN_ATTR_DEVICE_TEXT_ID     => $LUN_ATTR_EMPTY,
    $LUN_ATTR_DEVICE_BINARY_ID   => $LUN_ATTR_EMPTY,
    $LUN_ATTR_DEVICE_LEGACY_ID   => $LUN_ATTR_EMPTY,
    $LUN_ATTR_INCONSIST_FILE_SYS => $LUN_ATTR_INCONSIST_FILE_SYS_FALSE,
    $LUN_ATTR_INCONSIST_BLOCKS   => $LUN_ATTR_INCONSIST_BLOCKS_FALSE,
    $LUN_ATTR_IS_CLONE           => $LUN_ATTR_IS_CLONE_FALSE,
    $LUN_ATTR_IS_CLONE_AUTODELETE_ENABLED => $LUN_ATTR_IS_CLONE_AUTODELETE_ENABLED_FALSE,
);

# LUN attributes
my @VsrLunAttrsInheritedFromSrc = (
    $LUN_ATTR_VSERVER,            $LUN_ATTR_VOLUME,
    $LUN_ATTR_QTREE,              $LUN_ATTR_PATH,
    $LUN_ATTR_NAME,               $LUN_ATTR_SERIAL,
    $LUN_ATTR_UUID,               $LUN_ATTR_VDISK_ID,
    $LUN_ATTR_STATE,              $LUN_ATTR_MAPPED,
    $LUN_ATTR_NVFAIL,             $LUN_ATTR_SPACE_RESV_HONORED,
    $LUN_ATTR_SPACE_ALLOC,        $LUN_ATTR_DEVICE_TEXT_ID,
    $LUN_ATTR_DEVICE_BINARY_ID,   $LUN_ATTR_DEVICE_LEGACY_ID,
    $LUN_ATTR_INCONSIST_FILE_SYS, $LUN_ATTR_INCONSIST_BLOCKS,
    $LUN_ATTR_PREFIX_SIZE,        $LUN_ATTR_SIZE,
    $LUN_ATTR_SUFFIX_SIZE,        $LUN_ATTR_EXTENT_SIZE,
    $LUN_ATTR_BLOCK_SIZE,         $LUN_ATTR_SIZE_USED,
    $LUN_ATTR_COMMENT,            $LUN_ATTR_SPACE_RESV,
    $LUN_ATTR_CREATION_TIME,      $LUN_ATTR_READ_ONLY,
    $LUN_ATTR_OSTYPE
);

#    serial - Override space for the serial number of the LUN
#    UUID - Override space for the UUID of the LUN
#    dev id type - Override space for the dev id type of the LUN
#    dev id - Override space for the dev id of the LUN
#    admin state - Override space for the admin state of the LUN (online/offline)
my @VsrLunAttrsInheritedFromAfs = (
    $LUN_ATTR_SERIAL,
    $LUN_ATTR_UUID,
    $LUN_ATTR_STATE,
    $LUN_ATTR_DEVICE_TEXT_ID,
    $LUN_ATTR_DEVICE_BINARY_ID,
    $LUN_ATTR_DEVICE_LEGACY_ID
);


################################################################################
# Method:     new
# Objective:  Class Constructor.
# Parameters: none
# Details:    See POD documentation at the end of this file.
################################################################################
sub new {
    my ($class) = @_;

    # bless instance into class
    my $self = {};
    bless( $self, $class );

    return $self;
} ## end sub new

################################################################################
# Method:     do_sfsr
# Objective:  to perform a single file snapshot restore of a LUN
# Parameters: $args{'src_lun_attrs_href'} - source LUN attributes from snapshot,
#                                           required
#             $args{'dst_path'}          - destination LUN path, optional
#             $args{'vol_obj'}            - Volume Object, required
#             $args{'snapshot_name'}      - Snapshot Name, required
# Return:     A hash containing the expected LUN attributes
#             'src' - LUN attributes of the src LUN
#             'dst' - LUN attributes of the dst LUN
#             'sfsr_type' - sfsr type, either inplace or out_of_place
# Details:    See POD documentation at the end of this file for examples.
###############################################################################
sub do_sfsr {

    my ( $self, %args ) = @_;

    # Clear any cached error
    $self->_clear_err();

    # Initialize variables
    my ($snapshot_name,     $snapshot_obj,
        %cmd_args,          $destination_path,
        $dst_lun_name,      $container_path,
        @container,         $container_obj,
        $vserver_name,      $src_lun_attrs_href,
        $volume_name,       $vol_obj,
        $src_lun_path,      $dst_lun_path,
        $lun_obj,           $destination_lun_attributes,
        $org_src_lun_attrs, $expected_dst_lun_attrs
    );

    # Initialize LUN Attributes hash for expected attributes return value
    my $sfsr_exp_attrs_hash = {};

    # Remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # Parse and validate arguments
    eval {

        # Validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'src_lun_attrs_href' => {
                    type     => HASHREF,
                    required => 1
                },
                'dst_lun_path' => {
                    type     => SCALAR,
                    regex    => qr/^(.*\/)([^\/]*)$/,
                    optional => 1
                },
             'vol_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa => 'San::ControllerCluster::Filer10M::Vserver::Volume'
                },
                'snapshot_name' => {
                    type     => SCALAR,
                    required => 1
                }
            }
        );
    };
    if ($@) {
        $self->_set_err($@);
        return {};
    }

    # Source LUN attributes hash reference
    $vol_obj            = $args{'vol_obj'};
    $src_lun_attrs_href = $args{'src_lun_attrs_href'};
    $src_lun_path       = $src_lun_attrs_href->{$LUN_ATTR_PATH};

    my %hash_copy = %{$src_lun_attrs_href};

    # Set base dst LUN attributes
    $expected_dst_lun_attrs = \%hash_copy;

    # If the dst_lun_path exists, then pass the
    # 'src_lun_path' and 'dst_lun_path' to the sfsr command
    # otherwise, just pass the src_lun_path to the sfsr command since
    # the destination path for the restore will be the src_lun_path by default
    if ( defined $args{'dst_lun_path'} ) {
        $dst_lun_path = $args{'dst_lun_path'};

        # Build Single File Snapshot Restore command arguments
        %cmd_args = (
            -path           => $src_lun_path,
            '-restore-path' => $dst_lun_path
        );

        # Save the destination path for the restore operation
        $destination_path = $dst_lun_path;
        $self->sanlog( -msg => "cmd_args = " . "@{[%cmd_args]}\n");
        $self->sanlog( -msg => "destination_path = " . $destination_path);
    } else {

        # Build Single File Snapshot Restore command arguments
        %cmd_args = ( -path => $src_lun_path );

        # Save the destination path for the restore operation
        $destination_path = $src_lun_path;
        $self->sanlog( -msg => "cmd_args = " . "@{[%cmd_args]}\n");
        $self->sanlog( -msg => "destination_path = " . $destination_path);
    } ## end else [ if ( defined $args{'dst_lun_path'...

    # Save LUN Name of the destination path
    $dst_lun_name = File::Basename::basename($destination_path);

    # Get the Snapshot Object from the Snapshot Name
    $snapshot_name = $args{'snapshot_name'};
    $snapshot_obj
        = $vol_obj->get_snapshot_obj( -snapshot => $args{'snapshot_name'} );
    if ( $vol_obj->error() ) {
        $self->_set_err( "Failed to get Snapshot Object.\nReason:\n",
            $vol_obj->errmsg() );
        return {};
    }

    # Get the Vserver Name from the Volume Object
    $vserver_name = $vol_obj->get_vserver();
    if ( $vol_obj->error() ) {
        $self->_set_err( "Failed to get vserver.\nReason:\n",
            $vol_obj->errmsg() );
        return {};
    }

    $volume_name = $vol_obj->get_name();
    if ( $vol_obj->error() ) {
        $self->_set_err( "Failed to get Volume Name.\nReason:\n",
            $vol_obj->errmsg() );
        return return {};
    }

    # Get the container for the LUN; need to know if LUN is in vol or qtree
    # Sample regex paths:
    # vol path - /vol/sfsr_29099_vol1/sfsr_src_lun_unmapped_windows_1
    # qtree path - /vol/sfsr_29099_vol1/qtree1/sfsr_src_lun_unmapped_windows_1
    $container_path = File::Basename::dirname($destination_path);
    $container_path =~ s/^\///;    # removing first forward slash in path
    @container = split( /\//, $container_path );

    # If the LUN Path contains a Qtree, then save the QTREE Name
    my $qtree = '""';
    if ( scalar(@container) == 3 ) {
        $qtree = pop(@container);
    }
    $self->sanlog( -msg => "destination lun volume = " . $volume_name);
    $self->sanlog( -msg => "destination lun qtree = " . $qtree);

    # Set container object
    if ( $qtree ne '""' ) {
        $container_obj = $vol_obj->get_qtree_obj( -qtree => $qtree );
        if ( $vol_obj->error() ) {
            $self->_set_err( "Failed to get qtree object.\nReason:\n",
                $vol_obj->errmsg() );
            return {};
        }
    } else {
        $container_obj = $vol_obj;
    }

    my $exists = $container_obj->lun_exist( -lun => $dst_lun_name );
    if ( $container_obj->error() ) {
        $self->_set_err(
            "Unable to determine existence of "
                . "LUN_PATH($destination_path).\nReason:\n",
            $container_obj->errmsg()
        );
        return {};
    } ## end if ( $container_obj->error...
    $self->sanlog( -msg => "lun exists = " . $exists );

    # Check if the destination path exists
    if ($exists) {
        ####################################################################
        # LUN exists, get the LUN Object and build the LUN Attributes before
        # the restore, then restore the LUN by executing the Single File
        # Snapshot Restore of the LUN; after the restore, return the expected
        # lun attributes hash
        ####################################################################
        $self->sanlog( -msg => "exits and inplace " );

        # Save restore type in expected LUN attributes hash
        $sfsr_exp_attrs_hash->{'sfsr_type'} = "inplace";

        # Get the LUN Object of the destination path
        $lun_obj = $container_obj->get_lun( -name => $dst_lun_name );
        if ( $container_obj->error() ) {
            $self->_set_err(
                "Failed to get LUN Object for "
                    . "LUN_PATH($destination_path) before Single File Snapshot"
                    . "Restore.\nReason:\n",
                $container_obj->errmsg()
            );
            return {};
        } ## end if ( $container_obj->error...

        # Build the destination LUN attributes before restore
        $destination_lun_attributes
            = $self->build_lun_attributes( -lun_obj => $lun_obj );
        if ( $self->error() ) {
            $self->_set_err(
                "Failed to build LUN Attributes before "
                    . "Single File Snapshot Restore for "
                    . "LUN_PATH($destination_path)."
                    . "\nReason:\n",
                $self->errmsg()
            );
            return {};
        } ## end if ( $self->error() )

        # Update the src LUN attributes with the LUN attributes inherited
        # from the destination LUN itself and not from the src LUN
        foreach my $key (@InplaceLunAttrsInheritedFromDstLun) {
            $expected_dst_lun_attrs->{$key}
                = $destination_lun_attributes->{$key};
        }
        
        $expected_dst_lun_attrs->{$LUN_ATTR_MAX_RESIZE_SIZE}  = '_SKIP_CHECK_'; 
        $expected_dst_lun_attrs->{$LUN_ATTR_HAS_METADATA_PROVISIONED} = '_SKIP_CHECK_';
        $expected_dst_lun_attrs->{$LUN_ATTR_QOS_POLICY_GROUP}  = '_SKIP_CHECK_';
        $expected_dst_lun_attrs->{$LUN_ATTR_EXTERNAL_CACHE}  = '_SKIP_CHECK_';
        $expected_dst_lun_attrs->{$LUN_ATTR_ALIGNMENT}  = '_SKIP_CHECK_';

    } else {
        ####################################################################
        # LUN does NOT Exist, restore the LUN by executing the
        # Single File Snapshot Restore of LUN, then build and return the
        # expected LUN Attributes hash
        ####################################################################
        $self->sanlog( -msg => "exits and out of place " );

        $sfsr_exp_attrs_hash->{'sfsr_type'} = "out_of_place";

        # Update the expected LUN attributes with the attributes inherited
        # from the destination LUN itself and not from the src LUN
        # LUN Attributes determined by destination path
        $expected_dst_lun_attrs->{$LUN_ATTR_VSERVER} = $vserver_name;
        $expected_dst_lun_attrs->{$LUN_ATTR_PATH}    = $destination_path;
        $expected_dst_lun_attrs->{$LUN_ATTR_VOLUME}  = $volume_name;
        $expected_dst_lun_attrs->{$LUN_ATTR_QTREE}   = $qtree;
        $expected_dst_lun_attrs->{$LUN_ATTR_NAME}    = $dst_lun_name;

        # Update the expected LUN attributes with the attributes that are
        # initialied to default values
        foreach my $key ( keys %OutOfPlaceInitializedLunAttrs ) {
            $expected_dst_lun_attrs->{$key}
                = $OutOfPlaceInitializedLunAttrs{$key};
        }

        # Do not know what these values are since new dst LUN
        $expected_dst_lun_attrs->{$LUN_ATTR_SERIAL}   = "_SKIP_CHECK_";
        $expected_dst_lun_attrs->{$LUN_ATTR_UUID}     = "_SKIP_CHECK_";
        $expected_dst_lun_attrs->{$LUN_ATTR_VDISK_ID} = "_SKIP_CHECK_";
        $expected_dst_lun_attrs->{$LUN_ATTR_OBJ}      = "_SKIP_CHECK_";

        $expected_dst_lun_attrs->{$LUN_ATTR_MAX_RESIZE_SIZE}  = '_SKIP_CHECK_'; 
        $expected_dst_lun_attrs->{$LUN_ATTR_HAS_METADATA_PROVISIONED} = '_SKIP_CHECK_';
        $expected_dst_lun_attrs->{$LUN_ATTR_QOS_POLICY_GROUP}  = '_SKIP_CHECK_';
        $expected_dst_lun_attrs->{$LUN_ATTR_EXTERNAL_CACHE}  = '_SKIP_CHECK_';
        $expected_dst_lun_attrs->{$LUN_ATTR_ALIGNMENT}  = '_SKIP_CHECK_';

    } ## end else [ if ($exists)

        $self->sanlog( -msg => "cmd_args = " . "@{[%cmd_args]}\n");
        $self->sanlog( -msg => "destination_path = " . $destination_path);
    # Execute sfsr command
    $snapshot_obj->restore_file(%cmd_args);
    if ( $snapshot_obj->error() ) {
        $self->_set_err(
            "Failed to execute Single File Snapshot "
                . "Restore of LUN_PATH($destination_path).\nReason:\n",
            $snapshot_obj->errmsg()
        );
        return {};
    } ## end if ( $snapshot_obj->error...

    $sfsr_exp_attrs_hash->{'src'} = { %{$src_lun_attrs_href} };
    $sfsr_exp_attrs_hash->{'dst'} = $expected_dst_lun_attrs;

    return $sfsr_exp_attrs_hash;
} ## end sub do_sfsr

################################################################################
# Method:     verify_sfsr
# Objective:  to perform a single file snapshot restore of a LUN
# Parameters: $args{'sfsr_exp_attrs_href'} - expected sfsr LUN attributes, reqd.
#             $args{'dst_lun_obj'} - destination LUN object, optional because
#             not required for inplace sfsr type, but required for out of place
# Return:     Returns TRUE if successful, FALSE otherwise and sets error.
# Details:    See POD documentation at the end of this file for examples.
###############################################################################
sub verify_sfsr {

    my ( $self, %args ) = @_;

    # Clear any cached error
    $self->_clear_err();

    # Initialize variables
    my ($sfsr_exp_attrs_href, $src_lun_exp_attrs, $dst_lun_exp_attrs,
        $ctl_src_lun_attrs,   $ctl_dst_lun_attrs, $sfsr_type,
        $dst_lun_obj,         $no_src,            $src_afs_exp_attrs_href
    );

    # Remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # Parse and validate arguments
    eval {

        # Validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'sfsr_exp_attrs_href' => {
                    type     => HASHREF,
                    required => 1
                },
                'src_afs_exp_attrs_href' => {
                    type     => HASHREF,
                    optional => 1
                },
                'dst_lun_obj' => {
                    type     => OBJECT,
                    optional => 1,
                    isa =>
                        'San::ControllerCluster::Filer10M::Vserver::Volume::Lun'
                },
                'no_src' => {
                    type     => SCALAR,
                    optional => 1,
                    default => 0
                },
            }
        );
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # Initalize variables from args
    $sfsr_exp_attrs_href = $args{'sfsr_exp_attrs_href'};
    $src_lun_exp_attrs   = $sfsr_exp_attrs_href->{'src'};
    $dst_lun_exp_attrs   = $sfsr_exp_attrs_href->{'dst'};
    $sfsr_type           = $sfsr_exp_attrs_href->{'sfsr_type'};
    $no_src              = $args{'no_src'};
    if ( $sfsr_type eq "out_of_place" ) {
    $self->sanlog( -msg => "out of place = " . $sfsr_type );
        if ( exists $args{'dst_lun_obj'} ) {
            $dst_lun_obj = $args{'dst_lun_obj'};
            $dst_lun_exp_attrs->{$LUN_ATTR_OBJ} = $dst_lun_obj;
        } else {
            $self->_set_err("Need dst lun obj to verify out of place sfsr ");
            return $SanUtil::FALSE;
        }
    } else {
        $dst_lun_obj = $dst_lun_exp_attrs->{$LUN_ATTR_OBJ};
    }
    
    # Array to keep track of errors
    my @errs;

    if ( $no_src == 0 ) {
        # Get src LUN attributes from controller
        # GET SRC LUN ATTR FROM CONTROLLER
        $ctl_src_lun_attrs = $self->build_lun_attributes(
            -lun_obj => $src_lun_exp_attrs->{$LUN_ATTR_OBJ} );
        if ( $self->error() ) {
            $self->_set_err(
                "Failed to get LUN Attributes for "
                    . "src LUN($src_lun_exp_attrs->{$LUN_ATTR_PATH})."
                    . "from controller.\nReason:\n",
                $self->errmsg()
            );
            return $SanUtil::FALSE;
        } ## end if ( $self->error() )

        # Compare expected src LUN attributes against those seen from controller

        my $src_lun_afs_exp_attrs;
        if ( exists $args{'src_afs_exp_attrs_href'}) {
            $src_lun_afs_exp_attrs = $args{'src_afs_exp_attrs_href'};
        } else {
             $src_lun_afs_exp_attrs = $src_lun_exp_attrs;
        }

       $src_lun_afs_exp_attrs->{$LUN_ATTR_MAX_RESIZE_SIZE}  = '_SKIP_CHECK_';
       $src_lun_afs_exp_attrs->{$LUN_ATTR_HAS_METADATA_PROVISIONED} = '_SKIP_CHECK_';
       $src_lun_afs_exp_attrs->{$LUN_ATTR_QOS_POLICY_GROUP}  = '_SKIP_CHECK_';
       $src_lun_afs_exp_attrs->{$LUN_ATTR_EXTERNAL_CACHE}  = '_SKIP_CHECK_';
       $src_lun_afs_exp_attrs->{$LUN_ATTR_ALIGNMENT}  = '_SKIP_CHECK_';


        # COMPARE EXP SRC LUN ATTR AGAINST CONTROLLER
        $self->verify_lun_attributes( -lun_attr_info => $src_lun_afs_exp_attrs);
        if ( $self->error() ) {
            push( @errs,
                      "Expected src LUN attributes did not match those seen "
                    . "on the controller.\nReason:\n",$self->errmsg());
        }
    }
    
    # Get dst LUN attributes from controller
    $ctl_dst_lun_attrs
        = $self->build_lun_attributes( -lun_obj => $dst_lun_obj );
    if ( $self->error() ) {
        $self->_set_err(
            "Failed to get LUN Attributes for "
                . "dst LUN($dst_lun_exp_attrs->{$LUN_ATTR_PATH})."
                . "from controller.\nReason:\n",
            $self->errmsg()
        );
        return $SanUtil::FALSE;
    } ## end if ( $self->error() )

    # Compare dst LUN attributes against those seen from controller
    $self->verify_lun_attributes( -lun_attr_info => $dst_lun_exp_attrs );
    if ( $self->error() ) {
        push( @errs,
                  "Expected dst LUN attributes did not match those seen "
                . "on the controller.\nReason:\n",$self->errmsg() );
    }

    my $exp_value;
    foreach my $attr ( keys %$ctl_dst_lun_attrs ) {
        next if ( $attr eq $LUN_ATTR_OBJ );
        next if ( $no_src == 1 );
        if (( $attr eq $LUN_ATTR_SERIAL )
            and ( $ctl_src_lun_attrs->{$LUN_ATTR_SERIAL} eq
                $ctl_dst_lun_attrs->{$LUN_ATTR_SERIAL} )
            )
        {
            $exp_value = "src and dst different";
        } ## end if ( ( $attr eq $LUN_ATTR_SERIAL...
        if ( $sfsr_type eq "inplace" ) {
            next if ( $attr eq $LUN_ATTR_SERIAL );
            # Inplace sfsr
            if ( grep( /^$attr$/, @InplaceLunAttrsInheritedFromSrcLun ) ) {
                $exp_value = $src_lun_exp_attrs->{$attr};
            } elsif ( grep( /^$attr$/, @InplaceLunAttrsInheritedFromDstLun ) )
            {
                $exp_value = $dst_lun_exp_attrs->{$attr};
            } else {
                $exp_value
                    = "!!!! Did not find expected value for LUN attribute($attr)";
            }
        } else {
            # Out of place sfsr
            next if ( $attr eq $LUN_ATTR_SERIAL );
            if ( $attr eq $LUN_ATTR_UUID ) {
                next if ( $no_src == 1 );
                if ( $ctl_src_lun_attrs->{$LUN_ATTR_UUID} eq
                    $ctl_dst_lun_attrs->{$LUN_ATTR_UUID} )
                {
                    $exp_value = "src and dst different";
                }
            } elsif ( $attr eq $LUN_ATTR_VDISK_ID ) {
                next if ( $no_src == 1 );
                if ( $ctl_src_lun_attrs->{$LUN_ATTR_VDISK_ID} eq
                    $ctl_dst_lun_attrs->{$LUN_ATTR_VDISK_ID} )
                {
                    $exp_value = "src and dst different";
                }
            } elsif (
                grep( /^$attr$/, @OutOfPlaceLunAttrsInheritedFromSrcLun ) )
            {
                $exp_value = $src_lun_exp_attrs->{$attr};
            } elsif (
                grep( /^$attr$/, @OutOfPlaceLunAttrsInheritedFromDstLun ) )
            {
                $exp_value = $dst_lun_exp_attrs->{$attr};
            } elsif ( exists $OutOfPlaceInitializedLunAttrs{$attr} ) {
                $exp_value = $OutOfPlaceInitializedLunAttrs{$attr};
            } else {
                $exp_value
                    = "!!!! Did not find expected value for LUN attribute($attr)";
            }
        } ## end else [ if ( $sfsr_type eq "inplace")

        if ( $no_src == 0 ) {
          if ($exp_value =~ /_SKIP_CHECK_/) {
            next;
          }
          
          if ( $ctl_src_lun_attrs->{$attr} eq $ctl_dst_lun_attrs->{$attr} ) {
            if ( $ctl_dst_lun_attrs->{$attr} ne $exp_value ) {
                push( @errs,
                          "Controller Src and Dst LUN attribute are equal."
                        . "\nReason:\n"
                        . "attr = $attr "
                        . "src = $ctl_src_lun_attrs->{$attr} "
                        . "dst = $ctl_dst_lun_attrs->{$attr} "
                        . "but not equal to expected = $exp_value" );
            } ## end if ( $ctl_dst_lun_attrs...
          }
        } else {
            next if ( $attr eq $LUN_ATTR_SERIAL );
            next if ( $attr eq $LUN_ATTR_UUID );
            next if ( $attr eq $LUN_ATTR_VDISK_ID );
            if ( $ctl_dst_lun_attrs->{$attr} ne $exp_value ) {
                if ( $no_src == 0 ) {
                    push( @errs,
                        "Controller Src and Dst LUN attribute are not equal."
                            . "\nReason:\n"
                            . "attr = $attr "
                            . "src = $ctl_src_lun_attrs->{$attr} "
                            . "dst = $ctl_dst_lun_attrs->{$attr} "
                            . "expected = $exp_value" );
                } else {
                    push( @errs,
                        "Controller Src and Dst LUN attribute are not equal."
                            . "\nReason:\n"
                            . "attr = $attr "
                            . "dst = $ctl_dst_lun_attrs->{$attr} "
                            . "expected = $exp_value" );
                
                }
            } ## end if ( $ctl_dst_lun_attrs...
        } ## end else [ if ( $ctl_src_lun_attrs...
    }    ## foreach

    # Check for any Errors
    if ( scalar(@errs) ) {

        # Set error and bubble up
        $self->_set_err( "SFSR Verify Error:\n" . join "\n", @errs );
        return $SanUtil::FALSE;
    } ## end if ( scalar(@errs) )
    return $SanUtil::TRUE;
} ## end sub verify_sfsr

sub do_vsr {
    my ( $self, %args ) = @_;

    # Clear any cached error
    $self->_clear_err();


    # Initialize variables
    my ($snapshot_name,     $snapshot_obj,
        %cmd_args,          $volume_name,
        $vol_obj,           $exp_lun_attrs_aref,
        $preserve_lun_ids,  $vsr_src_lun_attrs_aref,
        @vsr_src_lun_attrs, @vsr_exp_lun_attrs,
        @vsr_afs_lun_attrs
    );

    # Initialize LUN Attributes hash for expected attributes return value
    my $vsr_exp_attrs_hash = {};

    # Remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # Parse and validate arguments
    eval {

        # Validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'vol_obj' => {
                    type     => OBJECT,
                    required => 1,
                    isa => 'San::ControllerCluster::Filer10M::Vserver::Volume'
                },
                'snapshot_name' => {
                    type     => SCALAR,
                    required => 1
                },
                'preserve_lun_ids' => {
                    type     => SCALAR,
                    optional => 1
                },
                'vsr_src_lun_attrs_aref' => {
                    type     => ARRAYREF,
                    required => 1
                }
            }
        );
    };
    if ($@) {
        $self->_set_err($@);
        return ();
    }

    # Source LUN attributes hash reference
    $vol_obj            = $args{'vol_obj'};
    $vsr_src_lun_attrs_aref = $args{'vsr_src_lun_attrs_aref'};
    @vsr_src_lun_attrs  = @{$vsr_src_lun_attrs_aref};


    # Get the Snapshot Object from the Snapshot Name
    $snapshot_name = $args{'snapshot_name'};
    $snapshot_obj
        = $vol_obj->get_snapshot_obj( -snapshot => $args{'snapshot_name'} );
    if ( $vol_obj->error() ) {
        $self->_set_err( "Failed to get Snapshot Object.\nReason:\n",
            $vol_obj->errmsg() );
        return ();
    }

    $volume_name = $vol_obj->get_name();
    if ( $vol_obj->error() ) {
        $self->_set_err( "Failed to get Volume Name.\nReason:\n",
            $vol_obj->errmsg() );
        return return {};
    }

    # Build Single File Snapshot Restore command arguments
    # vsr, i.e. disruptive snapshot restore, is the default
    my @vsr_afs_lun_objs = $vol_obj->get_luns();
    if ( $vol_obj->error() ) {
        $self->_set_err( "Failed to get vol($volume_name) lun objects.\nReason:\n",
            $vol_obj->errmsg() );
        return return {};
    }

    foreach my $lun_obj (@vsr_afs_lun_objs) {
        my $afs_lun_attrs = $self->build_lun_attributes(
            -lun_obj => $lun_obj );
        if ( $self->error() ) {
            $self->_set_err(
                "Failed to get lun attributes.\nReason:\n",
                $self->errmsg()
            );
            return $SanUtil::FALSE;
        } ## end if ( $self->error() )
        push(@vsr_afs_lun_attrs,$afs_lun_attrs);
    }
    
    # build expected lun attributes, we expect to see all src luns in snapshot
    for (@vsr_src_lun_attrs) {
        my %new_hash = %$_;
        push @vsr_exp_lun_attrs, \%new_hash;
    }

    foreach my $exp_lun_attr_href (@vsr_exp_lun_attrs) {
        my $FOUND_LUN = 0;
        foreach my $afs_lun_attr_href (@vsr_afs_lun_attrs) {
            if ($exp_lun_attr_href->{$LUN_ATTR_UUID} eq $afs_lun_attr_href->{$LUN_ATTR_UUID} ) {
                foreach my $lun_attr ( @VsrLunAttrsInheritedFromAfs ) {
                    $exp_lun_attr_href->{$lun_attr} = $afs_lun_attr_href->{$lun_attr};
                }
                $FOUND_LUN = 1;
            } elsif ($exp_lun_attr_href->{$LUN_ATTR_PATH} eq $afs_lun_attr_href->{$LUN_ATTR_PATH} ) {
                foreach my $lun_attr ( @VsrLunAttrsInheritedFromAfs ) {
                    $exp_lun_attr_href->{$lun_attr} = $afs_lun_attr_href->{$lun_attr};
                }
                $FOUND_LUN = 1;
            }
        }
        if ( $FOUND_LUN == 0) {
            foreach my $lun_attr ( @VsrLunAttrsInheritedFromAfs )  {
                $exp_lun_attr_href->{'OUT_OF_PLACE'} =  "true";
                $exp_lun_attr_href->{$lun_attr} = "_SKIP_CHECK_";
            }
        }
    }

    # execute single file snapshot restore command
    if ( defined $args{'preserve_lun_ids'} ) {
        %cmd_args = ( '-preserve-lun-ids' => $args{'preserve_lun_ids'} );
        # Save restore type in expected LUN attributes hash
        $vsr_exp_attrs_hash->{'vsr_type'} = "nondisruptive";
        $self->sanlog( -msg => "Executing nondisruptive restore");
        $snapshot_obj->restore(%cmd_args);
    } else {
        # Save restore type in expected LUN attributes hash
        $vsr_exp_attrs_hash->{'vsr_type'} = "disruptive";
        $snapshot_obj->restore();
    }
    if ( $snapshot_obj->error() ) {
        $self->_set_err("Failed to restore to lun path." .
                         "\nReason:\n", $snapshot_obj->errmsg() );
    return {};
    }

    $vsr_exp_attrs_hash->{'src'} = \@vsr_src_lun_attrs;
    $vsr_exp_attrs_hash->{'exp'} = \@vsr_exp_lun_attrs;
    
    return $vsr_exp_attrs_hash;
}

################################################################################
# Method:     verify_vsr
# Objective:  to perform a single file snapshot restore of a LUN
# Parameters: $args{'vsr_exp_attrs_aref'} - expected vsr LUN attributes, reqd.
#             $args{'vol_obj'} - volume object, required
# Return:     Returns TRUE if successful, FALSE otherwise and sets error.
# Details:    See POD documentation at the end of this file for examples.
###############################################################################
sub verify_vsr {

    my ( $self, %args ) = @_;

    # Clear any cached error
    $self->_clear_err();

    # Initialize variables
    my (
        $ctl_src_lun_attrs, $vsr_exp_attrs_href,
        $src_lun_attrs,     $exp_lun_attrs,
        @src_lun_attrs,     @exp_lun_attrs,
        @vsr_afs_lun_attrs, $vsr_type,
        $skip_attr
    );

    # Remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # Parse and validate arguments
    eval {

        # Validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {   'vsr_exp_attrs_href' => {
                    type     => HASHREF,
                    required => 1
                },
                'skip_attr' => {
                    type     => SCALAR,
                    optional => 1
                }
            }
        );
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # Initalize variables from args
    $vsr_exp_attrs_href = $args{'vsr_exp_attrs_href'};
    $src_lun_attrs   = $vsr_exp_attrs_href->{'src'};
    $exp_lun_attrs   = $vsr_exp_attrs_href->{'exp'};
    $vsr_type        = $vsr_exp_attrs_href->{'vsr_type'};
    @src_lun_attrs   = @{$src_lun_attrs};
    @exp_lun_attrs   = @{$exp_lun_attrs};
    $skip_attr       = $args{'skip_attr'};
    
    my @errs;
    foreach my $exp_lun_attrs_href (@exp_lun_attrs) {
        my $ctl_lun_attrs_href = $self->build_lun_attributes(
            -lun_obj => $exp_lun_attrs_href->{$LUN_ATTR_OBJ} );
        if ( $self->error() ) {
            $self->_set_err("Failed to build lun attributes\nReason:\n",
                $self->errmsg() );
            return $SanUtil::FALSE;
        }
       
       $Data::Dumper::Maxdepth = 1; 
       my $dumpout = Dumper($ctl_lun_attrs_href);
       $self->sanlog( -msg => "ctl lun attrs = \n" . $dumpout);
        
        $self->sanlog( -msg => "checking ctl and exp attrs for: " . $exp_lun_attrs_href->{$LUN_ATTR_PATH});

        if ($ctl_lun_attrs_href->{$LUN_ATTR_PATH} eq
            $exp_lun_attrs_href->{$LUN_ATTR_PATH} ) {
            foreach my $lun_attr ( keys %$ctl_lun_attrs_href ) {
                    $self->sanlog( -msg => "ctl($lun_attr): " . $ctl_lun_attrs_href->{$lun_attr} );
                    $self->sanlog( -msg => "exp($lun_attr): " . $exp_lun_attrs_href->{$lun_attr} );
                if ( $ctl_lun_attrs_href->{$lun_attr} ne
                     $exp_lun_attrs_href->{$lun_attr} ) {
                    next if ( $lun_attr eq $skip_attr );
                    next if ( (defined $exp_lun_attrs_href->{'OUT_OF_PLACE'} ) &&
                              (grep{/^$lun_attr/} @VsrLunAttrsInheritedFromAfs) );
                     foreach my $src_lun_attrs_href (@src_lun_attrs) {
                        if ($ctl_lun_attrs_href->{$LUN_ATTR_PATH} eq
                            $src_lun_attrs_href->{$LUN_ATTR_PATH} ) {
                            push( @errs,
                                "Controller and Expected LUN attribute are not equal."
                                    . "\nReason:\n"
                                    . "lun path = $exp_lun_attrs_href->{$LUN_ATTR_PATH}:\n"
                                    . "\tvsr type = $vsr_type\n"
                                    . "\tattr = $lun_attr\n"
                                    . "\tctl = $ctl_lun_attrs_href->{$lun_attr}\n"
                                    . "\tsrc = $src_lun_attrs_href->{$lun_attr}\n"
                                    . "\texp = $exp_lun_attrs_href->{$lun_attr}\n");
                        }
                    }
                }
            }
        }
        
        if ( defined $exp_lun_attrs_href->{'OUT_OF_PLACE'} ) {
            foreach my $src_lun_attrs_href (@src_lun_attrs) {
                if ($ctl_lun_attrs_href->{$LUN_ATTR_PATH} eq
                    $src_lun_attrs_href->{$LUN_ATTR_PATH} ) {
                    my $lun_attr = "''";
                    if( $ctl_lun_attrs_href->{$LUN_ATTR_SERIAL} eq
                        $src_lun_attrs_href->{$LUN_ATTR_SERIAL} ) {
                        $lun_attr = $LUN_ATTR_SERIAL;
                    } elsif ($exp_lun_attrs_href->{$LUN_ATTR_UUID} eq
                        $ctl_lun_attrs_href->{$LUN_ATTR_UUID} ) {
                        $lun_attr = $LUN_ATTR_UUID;
                    } elsif ($exp_lun_attrs_href->{$LUN_ATTR_DEVICE_TEXT_ID} eq
                        $ctl_lun_attrs_href->{$LUN_ATTR_DEVICE_TEXT_ID} ) {
                        $lun_attr = $LUN_ATTR_DEVICE_TEXT_ID;
                    } elsif ($exp_lun_attrs_href->{$LUN_ATTR_DEVICE_BINARY_ID} eq
                        $ctl_lun_attrs_href->{$LUN_ATTR_DEVICE_BINARY_ID} ) {
                        $lun_attr = $LUN_ATTR_DEVICE_BINARY_ID;
                    } elsif ($exp_lun_attrs_href->{$LUN_ATTR_DEVICE_LEGACY_ID} eq
                        $ctl_lun_attrs_href->{$LUN_ATTR_DEVICE_LEGACY_ID} ) {
                        $lun_attr = $LUN_ATTR_DEVICE_LEGACY_ID;
                    }
                    if ($lun_attr ne "''") {
                    push( @errs,
                        "Expected and SRC LUN attribute are equal, should be unique."
                            . "\nReason:\n"
                            . "src lun path = $src_lun_attrs_href->{$LUN_ATTR_PATH}:\n"
                            . "ctl lun path = $ctl_lun_attrs_href->{$LUN_ATTR_PATH}:\n"
                            . "\tvsr type = $vsr_type\n"
                            . "\trestore_type = out of place\n"
                            . "\tattr = $lun_attr\n"
                            . "\tctl = $ctl_lun_attrs_href->{$lun_attr}\n"
                            . "\tsrc = $src_lun_attrs_href->{$lun_attr}\n"
                            . "\texp = $exp_lun_attrs_href->{$lun_attr}\n");

                    }
                }
            }
        }

    }

    # Check for any Errors
    if ( scalar(@errs) ) {

        # Set error and bubble up
        $self->_set_err( "VSR Verify Error:\n" . join "\n", @errs );
        #return $SanUtil::FALSE;
    } ## end if ( scalar(@errs) )
    
    return $SanUtil::TRUE;
} ## end sub verify_vsr

################################################################################
# Private Methods
################################################################################
1;
################################################################################
#                             POD Documentation
################################################################################

=pod

=head1 Package Name

LunRestoreHelper.pm

=head1 Author

Liz Haring (liz.haring@netapp.com)

=head1 Revision History

Feb-03-2011 eharing@netapp.com:  Original template creation

=head1 Description

LunRestoreHelper.pm provides an interface to restore a lun or luns in a
given volume restore via the volume snapshot command sub-commands:
partial-restore-file (pfsr), restore (vsr) and restore-file (sfsr).

=head1 Public Methods

=head2 B<new()>

Class Constructor method. To set the debug level to print debug message at
particular level, call 'set_debug_level' method. See Example for more details

=over 4

=item Inputs

none

=item Outputs

Returns a LunRestoreHelper blessed object reference, if the
instantiation of the object was successful, otherwise, returns an empty hash
reference.

=item Examples

    # Instantiate Lun Restore Helper Object
    $LunRestoreHelperObj
        = San::Task::LunRestoreHelper->new( );
    if ( $LunRestoreHelperObj->error() ) {
        $Test->nlog( $LunRestoreHelperObj->errmsg() );
        return $TCD::FAIL;
    }

    $LunRestoreHelperObj->set_debug_level($DEBUG_LEVEL);

=back

=head2 B<do_sfsr(%args)>

Perform a single file snapshot restore operation. The method will return a
hash containing the expected LUN attributes of the restored lun after the
sfsr operation.

=over 4

=item Inputs

=item * I<'src_lun_attrs_href'> <required>

A hash reference containing attributes of the source LUN for the sfsr operation.

=item * I<'dst_lun_path'>

This is the path to the destination LUN for the sfsr operation. If omitted, then
the src_lun_path is also the des_lun_path for the sfsr operation.  Thus, this
input paramter is optional.

=item * I<'vol_obj'> <required>

The volume object.  This volume object must correspond to the volume of the
snapshot from which the sfsr operation will be performed.

=item * I<'snapshot_name'> <required>

The name of the snapshot, with respect to the volume object.

=item Outputs

If successful, returns a hash containing the expected LUN attributes of the
'src' (source) LUN, the 'dst' (destination) LUN and the 'sfsr_type', inplace or
out of place (out_of_place); otherwise, an error message is set to the failure
reason and {} (empty hash reference) is set as the return value.

=item Examples

        # Execute sfsr of a source LUN path
        my $sfsr_exp_attrs_hash =
        $LunRestoreHelperObj->do_sfsr(
            'src_lun_attrs_href' => $src_lun_attrs_info,
            'vol_obj' => $vol_obj,
            'snapshot_name' => "snap1"
        );
        if ( $LunRestoreHelperObj->error() ) {
            $Test->nlog( $LunRestoreHelperObj->errmsg() );
            return $TCD::FAIL;
        }

        # Execute sfsr of a source LUN path to a destination LUN path
        my $sfsr_exp_attrs_hash =
        $LunRestoreHelperObj->do_sfsr(
            'src_lun_attrs_href' => $src_lun_attrs_info,
            'dst_lun_path' => "/vol/test_vol/lun2",
            'vol_obj' => $vol_obj,
            'snapshot_name' => "snap1"
        );
        if ( $LunRestoreHelperObj->error() ) {
            $Test->nlog( $LunRestoreHelperObj->errmsg() );
            return $TCD::FAIL;
        }

=back

=head2 B<verify_sfsr(%args)>

Perform verification of an sfsr restore operation. The method will return TRUE
if the verification was successful and FALSE otherwise.

=over 4

=item Inputs

=item * I<'sfsr_exp_attrs_href'> <required>

A hash reference to the sfsr expected attributes hash returned from do_sfsr().

=item * I<'dst_lun_obj'> <optional>

The destination LUN object.  This LUN object of the destination LUN, required
only when the sfsr type is out of place, not required when the sfsr type is
inplace.

=item Outputs

The method will return TRUE if the verification was successful and FALSE
otherwise. If FALSE, the method will set and return the error reason.

=item Examples

        # Execute inplace sfsr type of a source LUN path, then verify sfsr
        my $sfsr_exp_attrs_hash =
        $LunRestoreHelperObj->do_sfsr(
            'src_lun_attrs_href' => $src_lun_attrs_info,
            'vol_obj' => $vol_obj,
            'snapshot_name' => "snap1"
        );
        if ( $LunRestoreHelperObj->error() ) {
            $Test->nlog( $LunRestoreHelperObj->errmsg() );
            return $TCD::FAIL;
        }

        my $result =
        $LunRestoreHelperObj->verify_sfsr(
            'sfsr_exp_attrs_href' => $sfsr_exp_attrs_hash
        );
        unless ( $result ) {
            $Test->nlog( $LunRestoreHelperObj->errmsg() );
            return $TCD::FAIL;
        }

        # Execute out of place sfsr type of a source LUN path to a destination
        # LUN path, then verify sfsr
        my $sfsr_exp_attrs_hash =
        $LunRestoreHelperObj->do_sfsr(
            'src_lun_attrs_href' => $src_lun_attrs_info,
            'dst_lun_path' => "/vol/test_vol/newlun",
            'vol_obj' => $vol_obj,
            'snapshot_name' => "snap1"
        );
        if ( $LunRestoreHelperObj->error() ) {
            $Test->nlog( $LunRestoreHelperObj->errmsg() );
            return $TCD::FAIL;
        }

        my $result =
        $LunRestoreHelperObj->verify_sfsr(
            'sfsr_exp_attrs_href' => $sfsr_exp_attrs_hash,
            'dst_lun_obj' => $destination_lun_obj,
        );
        unless ( $result ) {
            $Test->nlog( $LunRestoreHelperObj->errmsg() );
            return $TCD::FAIL;
        }
        
=head2 B<do_vsr(%args)>

Perform a volume restore operation. The method will return a
hash to an array reference containing the expected LUN attributes of the
restored LUNs in volume after the vsr operation.

=over 4

=item Inputs

=item * I<'vol_obj'> <required>

The volume object.  This volume object must correspond to the volume of the
snapshot from which the vsr operation will be performed.

=item * I<'snapshot_name'> <required>

The name of the snapshot, with respect to the volume object.


=item * I<'vsr_exp_attrs_href'> <required>

A hash reference to the sfsr expected attributes hash returned from do_sfsr().

=item * I<'preserve_lun_ids'> <required>

The parameter to determins whether executing a disruptive or non-disruptive
volume restore operation.
If set to 1, then operation is non-disruptive volume snapshot restore.

=item Outputs

The method will return TRUE if the verification was successful and FALSE
otherwise. If FALSE, the method will set and return the error reason.
        
=item Examples

        # run and verify vsr on LUNs with
        # afs-online/snap-online and afs-offline/snap-offline
        # Build vsr command arguments
        %vsr_args = (
            'vol_obj'                => $vol_obj,
            'snapshot_name'          => $snapshot_name,
            'preserve_lun_ids'       => "false",
            'vsr_src_lun_attrs_aref' => \@original_src_lun_attrs
        );

        # Execute vsr
        $vsr_exp_attrs_hash = $LunRestoreHelperObj->do_vsr(%vsr_args);
        if ( $LunRestoreHelperObj->error() ) {
            $Test->nlog(
                "Failed to perform vsr.\nReason:\n",
                $LunRestoreHelperObj->errmsg()
            );
            return $TCD::FAIL;
        } ## end if ( $LunRestoreHelperObj...

=back


=head2 B<verify_vsr(%args)>

Perform verification of an vsr restore operation. The method will return TRUE
if the verification was successful and FALSE otherwise.
=over 4

=item Inputs

=item * I<'vol_obj'> <required>

The volume object.  This volume object must correspond to the volume of the
snapshot from which the vsr operation will be performed.

=item * I<'snapshot_name'> <required>

The name of the snapshot, with respect to the volume object.


=item * I<'vsr_exp_attrs_href'> <required>

A hash reference to the sfsr expected attributes hash returned from do_sfsr().

=item * I<'preserve_lun_ids'> <required>

The parameter to determins whether executing a disruptive or non-disruptive
volume restore operation.
If set to 1, then operation is non-disruptive volume snapshot restore.

=item Outputs

The method will return TRUE if the verification was successful and FALSE
otherwise. If FALSE, the method will set and return the error reason.

=item Examples

        # run and verify vsr on LUNs with
        # afs-online/snap-online and afs-offline/snap-offline
        # Build vsr command arguments
        %vsr_args = (
            'vol_obj'                => $vol_obj,
            'snapshot_name'          => $snapshot_name,
            'preserve_lun_ids'       => "false",
            'vsr_src_lun_attrs_aref' => \@original_src_lun_attrs
        );

        # Execute vsr
        $vsr_exp_attrs_hash = $LunRestoreHelperObj->do_vsr(%vsr_args);
        if ( $LunRestoreHelperObj->error() ) {
            $Test->nlog(
                "Failed to perform vsr.\nReason:\n",
                $LunRestoreHelperObj->errmsg()
            );
            return $TCD::FAIL;
        } ## end if ( $LunRestoreHelperObj...

        # Build verify vsr command arguments
        %verify_vsr_args = ( 'vsr_exp_attrs_href' => $vsr_exp_attrs_hash );

        # Verify vsr
        $verify_vsr_result
            = $LunRestoreHelperObj->verify_vsr(%verify_vsr_args);
        if ( $LunRestoreHelperObj->error() ) {
            $Test->nlog(
                "Failed to verify vsr.\nReason:\n",
                $LunRestoreHelperObj->errmsg()
            );
            return $TCD::FAIL;
        } ## end if ( $LunRestoreHelperObj...

=back
