## Copyright (c) 2010 NetApp, Inc.
## All rights reserved.
## $Id: //depot/prod/test/Rsierranevada/lib/San/Task/NdvmHelper.pm#1 $
##
## @summary Helper commands for NDVM functions
## @author Blocks Automation (c-dl-blocks-host-automation@netapp.com)
## @status public
## @unit_test lib/UnitTest/test_San_SanFiler_ndvm.thpl
## @pod here

package San::Task::NdvmHelper;

use strict;
use Error;
use San::SanLog;
use San::SanUtil;
use Data::Dumper;

use vars qw( @ISA );

@ISA = qw( Error SanLog SanUtil );

###############################################################################
# P U B L I C   M E T H O D S
###############################################################################

###############################################################################
# Method:     new
# Objective:  Creates new NdvmHelper object.
# Parameters: None.
# Returns:    A reference to the object.
###############################################################################

sub new {
    my ($class, %args) = @_;

    my ($vol_obj, $filer_obj, $volume, $aggrs, $vol_name);

    my $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # Initialize some variables;
    my $interval = 20;
    my $max_wait = 1200;

    # Parse the arguments
    foreach (keys %args) {
        if (/^-?filer$/) {
            unless (length($filer_obj = $args{$_}) > 0 and ref($filer_obj) =~ /sanfiler/i) {
                $self->_set_err("Invalid value for '-filer': $args{$_}");
                return $self;
            }
        } elsif (/^-?interval$/) {
            unless (length($interval = $args{$_}) > 0 and $interval > 0) {
                $self->_set_err("Invalid value for '-interval': $args{$_}");
                return $self;
            }
        } elsif (/^-?max_wait$/) {
            unless (length($max_wait = $args{$_}) > 0 and $max_wait >= 0) {
                $self->_set_err("Invalid value for '-max_wait': $args{$_}");
                return $self;
            }
        } elsif (/^-?volume$/) {
            unless (length($volume = $args{$_}) > 0) {
                $self->_set_err("Invalid value for '-volume': $args{$_}");
                return $self;
            }
        } elsif (/^-?aggrs$/) {
            unless (length($aggrs = $args{$_}) > 0 and $aggrs =~ /\:/) {
                $self->_set_err("Invalid value for '-aggrs': $args{$_}");
                return;
            }
        } else {
            $self->_set_err("Invalid parameter specified: $_");
            return $self;
        }
    } ## end foreach (keys %args)

    # Verify mandatory parameters.
    unless (defined($filer_obj)) {
        $self->_set_err("Missing required argument '-filer_obj'");
        return $self;
    }
    unless (defined($aggrs)) {
        $self->_set_err("Missing required argument '-aggrs'");
        return $self;
    }
    unless (defined($volume)) {
        $self->_set_err("Missing required argument '-volume'");
        return $self;
    }

    # Get the filer name
    my $filer_name = $filer_obj->get_name();
    if ($filer_obj->error) {
        $self->_set_err("Unable to determine if the name of the filer:\n" . $filer_obj->errmsg());
        return $self;
    } ## end if ($filer_obj->error)

    # Break the aggrs out into an array and make sure that we received only 2
    my @aggrlist = split(/\:/, $aggrs);
    if (scalar(@aggrlist) != 2) {
        $self->_set_err("Invalid number of aggregates specified.  Two must be provided.");
        return $self;
    }

    # Check to see if volume is an object or a string
    if (ref($volume) =~ /volume/i) {

        # Looks like we received a volume object so we will just use that
        $vol_obj  = $volume;
        $vol_name = $vol_obj->get_name();
        if ($vol_obj->error) {
            $self->_set_err("Unable to determine the volume name:\n" . $vol_obj->errmsg());
            return $self;
        }
    } elsif (ref($volume) =~ /HASH|ARRAY/) {

        # We didn't receive a volume object, but we did receive either a hash or an array.
        $self->_set_err("Invalid data type '" . ref($volume) . "' specified for argument '-volume'");
        return $self;
    } else {

        # We didn't get a volume object so go ahead and retrieve it.
        # Make sure the volume exists
        unless ($filer_obj->vol_exist($volume)) {
            if ($filer_obj->error()) {
                $self->_set_err("Unable to determine if volume $volume exists:\n" . $filer_obj->errmsg());
            } else {
                $self->_set_err("Volume $volume does not exist on controller '$filer_name'");
            }
            return $self;
        } ## end unless ($filer_obj->vol_exist...

        # Get the volume object
        $vol_obj = $filer_obj->get_vol($volume);
        if ($filer_obj->error()) {
            $self->_set_err("Unable to retrieve the volume name:\n" . $filer_obj->errmsg());
            return $self;
        }

        $vol_name = $volume;
    } ## end else [ if (ref($volume) =~ /volume/i)

    # Populate the attributes
    $self->{filer_obj}  = $filer_obj;
    $self->{filer_name} = $filer_name;
    $self->{vol_obj}    = $vol_obj;
    $self->{vol_name}   = $vol_name;
    $self->{aggrs}      = \@aggrlist;
    $self->{max_wait}   = $max_wait;
    $self->{interval}   = $interval;

    # Make sure we aren't trying to move the root vol
    my $root_vol = $filer_obj->get_root_vol();
    if ($filer_obj->error) {
        $self->_set_err("Unable to determine the root vol:\n" . $filer_obj->errmsg());
        return $self;
    }
    if ($root_vol eq $vol_name) {
        $self->_set_err("Volume being moved cannot be the controller root volume");
        return $self;
    }

    # Make sure the aggregates are the same bit depth
    my $aggr1_depth = $filer_obj->get_aggr_bit_depth(-aggr => $aggrlist[0]);
    if ($filer_obj->error) {
        $self->_set_err("Unable to determine bit depth for aggr '$aggrlist[0]':\n" . $filer_obj->errmsg());
        return $self;
    } ## end if ($filer_obj->error)
    my $aggr2_depth = $filer_obj->get_aggr_bit_depth(-aggr => $aggrlist[1]);
    if ($filer_obj->error) {
        $self->_set_err("Unable to determine bit depth for aggr '$aggrlist[1]'" . $filer_obj->errmsg());
        return $self;
    } ## end if ($filer_obj->error)

    unless ($aggr1_depth == $aggr2_depth) {
        # aggr NDVMs only work between same aggr bit depth pre RR.1
        if ($filer_obj->get_version() !~ /8.1.[1-9]+|8.[2-9]|Rrollingrock|Rsierranevada|Rfullsteam|Rlongboard|main/) {
            $self->_set_err("Bit depth of $aggrlist[0] ($aggr1_depth) doesn't match $aggrlist[1] ($aggr2_depth)");
            return $self;
        }
    } ## end unless ($aggr1_depth == $aggr2_depth)

    return $self;
} ## end sub new

################################################################################
# Method: ndvm_start
# Objective: Start the ndvm of a volume
################################################################################
sub ndvm_start {
    my ($self, %args) = @_;

    my (%new_args);

    $self->_clear_err();

    # Initialize some variables
    my $unexport          = 0;
    my $filer_obj         = $self->{filer_obj};
    my $filer_name        = $self->{filer_name};
    my $volume            = $self->{vol_name};
    my $pause_retry_count = 5;

    # Parse the required/optional arguments.  Pass through all arguments to vol_obj.
    foreach (keys %args) {
        if (/-?unexport/) {
            if (length($unexport = $args{$_}) > 0 and $unexport =~ /YES|TRUE|1/i) {
                $unexport = 1;
            } else {
                $self->_set_err("Invalid value for argument '-unexport'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?pause_retry_count/) {
            if (not length($pause_retry_count = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-pause_retry_count'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            # Add the key/value pair to %new_args
            $new_args{$_} = $args{$_};
        }
    } ## end foreach (keys %args)

    # Grab the source aggr off of the volume obj
    my $srcaggr = $self->{vol_obj}->get_aggr(-command_retries => 5);
    if ($self->{vol_obj}->error()) {
        $self->_set_err("Unable to determine the source aggr for volume ${volume}:\n" . $self->{vol_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Get the destination aggr
    my $dstaggr = $self->get_dest_aggr();
    if ($self->error) {
        $self->_set_err("Unable to retrieve the destination aggr for volume $volume:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # Verify that the source and destination aggr are not the same
    if ($srcaggr eq $dstaggr) {
        $self->_set_err("Source and Destination aggregates are the same ($srcaggr) for volume $volume");
        return $ClientUtil::FALSE;
    }

    # Grab the list of aggrs on the controller so we can verify that the dstaggr are on the
    # same controller.  This is a requirement of BR.1.
    my @aggrs = $filer_obj->get_aggrs();
    if ($self->error) {
        $self->_set_err("Unable to get the aggrs on the controller:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }
    my $aggr_found = $ClientUtil::FALSE;

    foreach my $aggr (@aggrs) {
        if ($dstaggr eq $aggr) {
            $aggr_found = $ClientUtil::TRUE;
            last;
        }
    } ## end foreach my $aggr (@aggrs)
    if ($aggr_found == $ClientUtil::FALSE) {
        $self->_set_err("Source and Destination aggregates must be on the same controller");
        return $ClientUtil::FALSE;
    }

    # Check to see if the volume is exported.  If we find it exported and the $unexport
    # flag is set 0, then set an error and fail.  If $unexport flag is set to 1, unexport
    # the volume.
    my $export_status = $self->{vol_obj}->is_exported();
    if ($self->{vol_obj}->error) {
        $self->_set_err("Unable to check the export status of volume ${volume}:\n" . $self->{vol_obj}->errmsg());
        return $ClientUtil::FALSE;
    }
    if ($export_status == $ClientUtil::TRUE) {
        if ($unexport == 0) {
            $self->_set_err("Volume $volume is currently exported. Unable to start NDVM");
            return $ClientUtil::FALSE;
        } else {
            $self->{vol_obj}->unexport_fs();
            if ($self->{vol_obj}->error()) {
                $self->_set_err("Unable to unexport volume ${volume}:\n" . $self->{vol_obj}->errmsg());
                return $ClientUtil::FALSE;
            }
        } ## end else [ if ($unexport == 0)
    } ## end if ($export_status == ...

    # Make sure we don't have a migration already running
    my $output_hash = $self->ndvm_status(-verbose => 1);
    if ($self->error()) {
        $self->_set_err("Unable to check the move status of volume ${volume}:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    unless (defined($output_hash->{'migration-state'}) and $output_hash->{'migration-state'} eq "not_running") {
        $self->_set_err("Unable to start the move of volume $volume.  Migration state is currently set to "
              . $output_hash->{'migration-state'});
        return $ClientUtil::FALSE;
    }

    # We should now have a valid configuration so go ahead and move the volume
    $self->{vol_obj}->move_start(
        -aggr              => $dstaggr,
        -interval          => $self->{interval},
        -max_wait          => $self->{max_wait},
        -pause_retry_count => $pause_retry_count,
        %new_args
    );

    if ($self->{vol_obj}->error()) {
        $self->_set_err("Unable to start move of volume ${volume} to aggr ${dstaggr}:\n" . $self->{vol_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub ndvm_start

################################################################################
# Method: ndvm_status
# Objective: Get the status of the vol move
################################################################################
sub ndvm_status {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Initialize some variables
    my $volume  = $self->{vol_name};
    my $verbose = 0;

    # Parse the arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            unless (length($verbose = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-verbose'");
                return {};
            }
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return {};
        }
    }    # end foreach

    my $status_hash = $self->{vol_obj}->move_status(-verbose => $verbose);
    if ($self->{vol_obj}->error()) {
        $self->_set_err("Unable to get the status for the move of volume $volume:\n" . $self->{vol_obj}->errmsg());
        return {};
    }
    # Get the move state
    my $move_state;
    if (defined($status_hash->{'vol-move-status-info'}->{'move-state'})) {
        $move_state = $status_hash->{'vol-move-status-info'}->{'move-state'};
    }

    # Get the migration state
    my $migration_state;
    if (defined($status_hash->{'migration-state'})) {
        $migration_state = $status_hash->{'migration-state'};
    }

    # Print the status to the logs
    if ($move_state ne "" and $migration_state ne "") {
        $self->sanlog(-msg => "NDVM:  Move State is '$move_state' and Migration State is '$migration_state'");
    } elsif ($move_state ne "") {
        $self->sanlog(-msg => "NDVM:  Move State is '$move_state'");
    } elsif ($migration_state ne "") {
        $self->sanlog(-msg => "NDVM:  Migration State is '$migration_state'");
    }

    return $status_hash;
} ## end sub ndvm_status

################################################################################
# Method: ndvm_cutover
# Objective: Perform a manual vol move cutover
################################################################################
sub ndvm_cutover {
    my ($self, %args) = @_;

    my (%new_args, $output_hash);

    $self->_clear_err();

    # Initialize some variables
    my $unexport         = 0;
    my $cutover_attempts = 3;
    my $volume           = $self->{vol_name};

    # Parse the required/optional arguments.  Pass through all arguments to vol_obj.
    foreach (keys %args) {
        if (/-?unexport/) {
            if (length($unexport = $args{$_}) > 0 and $unexport =~ /YES|TRUE|1/i) {
                $unexport = 1;
            } else {
                $self->_set_err("Invalid value for argument '-unexport'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?num_cutover/) {
            unless (length($cutover_attempts = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-num_cutover'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {

            # Add the key/value pair to %new_args
            $new_args{$_} = $args{$_};
        }
    } ## end foreach (keys %args)

    # Make sure we have a migration in the "move" state
    $output_hash = $self->ndvm_status(-verbose => 1);
    if ($self->error()) {
        $self->_set_err("Unable to check the move status of volume ${volume}:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }
    unless (defined($output_hash->{'vol-move-status-info'}->{'move-state'})
            and $output_hash->{'vol-move-status-info'}->{'move-state'} eq "move") {
        $self->_set_err("Unable to start the cutover of volume $volume.  Move state is currently " .
                        $output_hash->{'vol-move-status-info'}->{'move-state'});
        return $ClientUtil::FALSE;
    } ## end unless (defined($output_hash...

    # Check to see if the volume is exported.  If we find it exported and the $unexport
    # flag is set 0, then set an error and fail.  If $unexport flag is set to 1, unexport
    # the volume.
    my $export_status = $self->{vol_obj}->is_exported();
    if ($self->{vol_obj}->error) {
        $self->_set_err("Unable to check the export status of volume ${volume}:\n" . $self->{vol_obj}->errmsg());
        return $ClientUtil::FALSE;
    }
    if ($export_status == $ClientUtil::TRUE) {
        if ($unexport == 0) {
            $self->_set_err("Volume $volume is currently exported. Unable to start NDVM");
            return $ClientUtil::FALSE;
        } else {
            $self->{vol_obj}->unexport_fs();
            if ($self->{vol_obj}->error()) {
                $self->_set_err("Unable to unexport volume ${volume}:\n" . $self->{vol_obj}->errmsg());
                return $ClientUtil::FALSE;
            }
        } ## end else [ if ($unexport == 0)
    } ## end if ($export_status == ...

    # Perform the cutover in a loop.  We will try $cutover_attemps attempts to
    # complete the cutover.
    my $cutovers_attempted = 0;
    my $cutover_start      = time();
    while (1) {

        # Increment the cutovers attempted counter
        $cutovers_attempted++;

        # Log a message
        $self->sanlog(-msg => "Beginning cutover attempt '$cutovers_attempted' of '$cutover_attempts'");

        # Attempt the cutover.
        $self->{vol_obj}->move_cutover(%new_args);

        # We don't care if there is an error because we will retry the command.  We
        # only care that there wasn't one. Due to a zapi bug, we need to check status
        # because the zapi may say it succeeded when it actually failed.  We can tell
        # by the fact that it will be in a running state if it failed.
        if ($self->{vol_obj}->error()) {
            # Log the error with sanlog so we know what is going on.  The next code block will
            # sort out what to do about the failure.
            my $errmsg = join("\n", $self->{vol_obj}->errmsg());
            $self->sanlog(-msg => "$errmsg");
        } else {
            # Make sure the migration is no longer running
            $output_hash = $self->ndvm_status(-verbose => 1);
            if ($self->error()) {
                $self->_set_err("Unable to check the move status of volume ${volume}:\n" . $self->errmsg());
                return $ClientUtil::FALSE;
            }

            # Check if the move is still running.  If it isn't, bail out of the loop because the
            # cutover completed.
            if (defined($output_hash->{'migration-state'})
                and $output_hash->{'migration-state'} eq "not_running") {
                last;
            }
        } ## end unless ($self->{vol_obj}->error...

        # The cutover failed so we need to check the cutover attempts. If we haven't
        # maxed them out then we will need to resume the move (it should be paused) and try again.
        if ($cutovers_attempted >= $cutover_attempts) {
            my $cutover_stop = time();
            my $total_time   = $cutover_stop - $cutover_start;
            $self->_set_err("Cutover of volume $volume failed.  All '$cutover_attempts' attempts have " .
                            "been exhausted. Total time of attempts was '$total_time' seconds.");
            return $ClientUtil::FALSE;
        } ## end if ($cutovers_attempted...

        # Since the cutover failed, sleep for 15 seconds before trying again to allow the system
        # to possibly calm down.
        $self->sanlog(-msg => "Cutover Failed. Sleeping 15 seconds to allow the system to stabilize before retrying");
        &SanUtil::nate_sleep(15);

        # Looks like we have some attempts left so let's make sure it is paused and then resume it.
        $output_hash = $self->ndvm_status(-verbose => 1);
        if ($self->error()) {
            $self->_set_err("Unable to check the move status of volume '$volume':\n" . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        if (defined($output_hash->{'vol-move-status-info'}->{'move-state'})
            and $output_hash->{'vol-move-status-info'}->{'move-state'} =~ /paused/) {

            # Resume the move and set the manual_cutover flag to true.
            $self->ndvm_resume(-manual_cutover => 1, -unexport => 1);
            if ($self->error) {
                $self->_set_err("Unable to resume the move of volume ${volume}:\n" . $self->errmsg());
                return $ClientUtil::FALSE;
            }
        } ## end if (defined($output_hash...
    } ## end while (1)

    my $cutover_stop = time();
    my $total_time   = $cutover_stop - $cutover_start;
    $self->sanlog(-msg => "Cutover completed after '$cutovers_attempted' attempts.  Total time was " .
                          "'$total_time' seconds.");

    # Refresh the volume aggregates
    $self->{vol_obj}->refresh(-aggr => 1, -luns => 0);
    if ($self->{vol_obj}->error) {
        $self->_set_err("Unable to refresh aggregates:\n" . $self->{vol_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub ndvm_cutover

################################################################################
# Method: ndvm_abort
# Objective: Abort the vol move of a volume
################################################################################
sub ndvm_abort {
    my ($self) = @_;

    my ($output_hash);

    $self->_clear_err();

    # Initialize some variables
    my $volume = $self->{vol_name};

    # Make sure we have a migration in the "running" state
    $output_hash = $self->ndvm_status(-verbose => 1);
    if ($self->error()) {
        $self->_set_err("Unable to check the move status of volume ${volume}:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }
    unless (defined($output_hash->{'migration-state'}) and $output_hash->{'migration-state'} eq "running") {
        $self->_set_err("Unable to abort the move of volume $volume.  Migration state is currently set to " .
                        $output_hash->{'migration-state'});
        return $ClientUtil::FALSE;
    }

    # Abort the migration
    $self->{vol_obj}->move_abort();
    if ($self->{vol_obj}->error()) {
        $self->_set_err("Unable to abort the move of volume ${volume}:\n" . $self->{vol_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Make sure the migration is no longer running
    $output_hash = $self->ndvm_status(-verbose => 1);
    if ($self->error()) {
        $self->_set_err("Unable to check the move status of volume ${volume}:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }
    unless (defined($output_hash->{'migration-state'}) and $output_hash->{'migration-state'} eq "not_running") {
        $self->_set_err("Abort of volume $volume failed.  Migration state is currently set to " .
                        $output_hash->{'migration-state'});
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub ndvm_abort

################################################################################
# Method: ndvm_pause
# Objective: Pause a vol move on a volume
################################################################################
sub ndvm_pause {
    my ($self) = @_;

    my ($output_hash);

    $self->_clear_err();

    # Initialize some variables
    my $volume = $self->{vol_name};

    # Make sure we have a migration in the "running" state
    $output_hash = $self->ndvm_status(-verbose => 1);
    if ($self->error()) {
        $self->_set_err("Unable to check the move status of volume ${volume}:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }
    unless (defined($output_hash->{'migration-state'}) and $output_hash->{'migration-state'} eq "running") {
        $self->_set_err("Unable to pause the move of volume $volume.  Migration state is currently set to " .
                        $output_hash->{'migration-state'});
        return $ClientUtil::FALSE;
    }

    # Pause the migration
    $self->{vol_obj}->move_pause();
    if ($self->{vol_obj}->error()) {
        $self->_set_err("Unable to pause the move of volume name:\n" . $self->{vol_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Make sure we have a migration in the "paused" state
    $output_hash = $self->ndvm_status(-verbose => 1);
    if ($self->error()) {
        $self->_set_err("Unable to check the move status of volume ${volume}:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }
    unless (defined($output_hash->{'vol-move-status-info'}->{'move-state'})
        and $output_hash->{'vol-move-status-info'}->{'move-state'} =~ /paused/) {
        $self->_set_err("Pause of volume $volume failed.  Move state is currently " .
                        $output_hash->{'vol-move-status-info'}->{'move-state'});
        return $ClientUtil::FALSE;
    } ## end unless (defined($output_hash...

    return $ClientUtil::TRUE;
} ## end sub ndvm_pause

################################################################################
# Method: ndvm_resume
# Objective: Resume the vol move on a volume
################################################################################
sub ndvm_resume {
    my ($self, %args) = @_;

    my (%new_args);

    $self->_clear_err();

    # Initialize some variables
    my $unexport = 0;
    my $volume   = $self->{vol_name};

    # Parse the required/optional arguments.  Pass through all arguments to vol_obj.
    foreach (keys %args) {
        if (/-?unexport/) {
            if (length($unexport = $args{$_}) > 0 and $unexport =~ /YES|TRUE|1/i) {
                $unexport = 1;
            } else {
                $self->_set_err("Invalid value for argument '-unexport'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {

            # Add the key/value pair to %new_args
            $new_args{$_} = $args{$_};
        }
    } ## end foreach (keys %args)

    # Get the destination aggr
    my $dstaggr = $self->get_dest_aggr();
    if ($self->error) {
        $self->_set_err("Unable to retrieve the destination aggr for volume $volume:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # Make sure we have a migration in the "paused" state
    my $output_hash = $self->ndvm_status(-verbose => 1);
    if ($self->error()) {
        $self->_set_err("Unable to check the move status of volume ${volume}:\n" . $self->errmsg());
        return $ClientUtil::FALSE;
    }
    unless (defined($output_hash->{'vol-move-status-info'}->{'move-state'})
        and $output_hash->{'vol-move-status-info'}->{'move-state'} =~ /paused/) {
        $self->_set_err("Unable to resume the move of volume $volume.  Move state is currently " .
                        $output_hash->{'vol-move-status-info'}->{'move-state'});
        return $ClientUtil::FALSE;
    } ## end unless (defined($output_hash...

    # Check to see if the volume is exported.  If we find it exported and the $unexport
    # flag is set 0, then set an error and fail.  If $unexport flag is set to 1, unexport
    # the volume.
    my $export_status = $self->{vol_obj}->is_exported();
    if ($self->{vol_obj}->error) {
        $self->_set_err("Unable to check the export status of volume ${volume}:\n" . $self->{vol_obj}->errmsg());
        return $ClientUtil::FALSE;
    }
    if ($export_status == $ClientUtil::TRUE) {
        if ($unexport == 0) {
            $self->_set_err("Volume $volume is currently exported. Unable to start NDVM");
            return $ClientUtil::FALSE;
        } else {
            $self->{vol_obj}->unexport_fs();
            if ($self->{vol_obj}->error()) {
                $self->_set_err("Unable to unexport volume ${volume}:\n" . $self->{vol_obj}->errmsg());
                return $ClientUtil::FALSE;
            }
        } ## end else [ if ($unexport == 0)
    } ## end if ($export_status == ...

    # Resume the move
    $self->{vol_obj}->move_resume(
        -aggr     => $dstaggr,
        -interval => $self->{interval},
        -max_wait => $self->{max_wait},
        %new_args,
    );

    if ($self->{vol_obj}->error()) {
        $self->_set_err("Unable to resume the move of volume ${volume}:\n" . $self->{vol_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub ndvm_resume

################################################################################
# Method: get_dest_aggr
# Objective: Gets the destination aggr for the volume.
################################################################################
sub get_dest_aggr {
    my ($self) = @_;

    $self->_clear_err();

    # Initialize some variables
    my $filer_obj = $self->{filer_obj};
    my $aggrs     = $self->{aggrs};

    # Get the volume name
    my $volume = $self->{vol_obj}->get_name();
    if ($self->{vol_obj}->error) {
        $self->_set_err("Unable to retrieve the volume name:\n" . $self->{vol_obj}->errmsg());
        return undef;
    }

    # Get the aggr for the volume
    my $src_aggr = $self->{vol_obj}->get_aggr(-command_retries => 5);
    if ($self->{vol_obj}->error) {
        $self->_set_err("Unable to get the aggregate for volume $volume:\n" . $self->{vol_obj}->errmsg());
        return undef;
    }

    # Define the destination aggr based on where the aggr currently resides and the aggregates
    # specified on $self->{aggrs};  The object should have been defined with 2 aggregates...no
    # more ... no less.
    if (scalar(@{$self->{aggrs}}) != 2) {
        $self->_set_err("Invalid number of aggregates specified on the object");
        return undef;
    }

    if ($src_aggr eq $aggrs->[0]) {
        return $aggrs->[1];
    } elsif ($src_aggr eq $aggrs->[1]) {
        return $aggrs->[0];
    }

    # If we made it this far then something went wrong.
    $self->_set_err(
        "Unable to find source aggregate '$src_aggr' in list of aggregates provided " . "for volume $volume");
    return undef;
} ## end sub get_dest_aggr

###############################################################################
# Accessor Methods
###############################################################################

################################################################################
# Method: get_filer_name
# Objective: Returns the filer name from the ndvm object
################################################################################
sub get_filer_name {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{filer_name}) and $self->{filer_name} ne "") {
        $self->_set_err("Invalid value for filer name on ndvm object");
        return undef;
    }

    return $self->{filer_name};
} ## end sub get_filer_name

################################################################################
# Method: get_filer_object
# Objective: Returns the filer object from the ndvm object
################################################################################
sub get_filer_object {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{filer_obj}) and ref($self->{filer_obj}) =~ /sanfiler/i) {
        $self->_set_err("Invalid value for filer object on ndvm object");
        return {};
    }

    return $self->{filer_obj};
} ## end sub get_filer_object

################################################################################
# Method: get_vol_name
# Objective: Returns the volume name from the ndvm object
################################################################################
sub get_vol_name {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{vol_name}) and $self->{vol_name} ne "") {
        $self->_set_err("Invalid value for volume name on ndvm object");
        return undef;
    }

    return $self->{vol_name};
} ## end sub get_vol_name

################################################################################
# Method: get_vol_object
# Objective: Returns the volume object from the ndvm object
################################################################################
sub get_vol_object {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{vol_obj}) and ref($self->{vol_obj}) =~ /volume/i) {
        $self->_set_err("Invalid value for volume object on ndvm object");
        return {};
    }

    return $self->{vol_obj};
} ## end sub get_vol_object

################################################################################
# Method: get_interval
# Objective: Returns the interval from the ndvm object
################################################################################
sub get_interval {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{interval}) and $self->{interval} ne "") {
        $self->_set_err("Invalid value for interval on ndvm object");
        return undef;
    }

    return $self->{interval};
} ## end sub get_interval

################################################################################
# Method: get_max_wait
# Objective: Returns the max wait time from the ndvm object
################################################################################
sub get_max_wait {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{max_wait}) and $self->{max_wait} ne "") {
        $self->_set_err("Invalid value for max_wait on ndvm object");
        return undef;
    }

    return $self->{max_wait};
} ## end sub get_max_wait

################################################################################
# Method: get_aggrs
# Objective: Returns the aggr list from the ndvm object
################################################################################
sub get_aggrs {
    my ($self) = @_;

    $self->_clear_err();

    unless (scalar(@{$self->{aggrs}}) > 0) {
        $self->_set_err("Invalid value for aggrs array on ndvm object");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @{$self->{aggrs}} : $self->{aggrs};
} ## end sub get_aggrs

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

NdvmHelper.pm

=head1 Author

Derrick Shultz

=head1 Description

NdvmHelper.pm provides an API interface to NDVM commands on a controller.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<San::Sanlog, San::SanUtil, Error>

=head1 Attributes

=head2 B<filer_obj>

object = Filer Object

=head2 B<filer_name>

string = Filer Name

=head2 B<vol_obj>

object = Volume Object

=head2 B<vol_name>

string = Volume Name

=head2 B<aggrs>

Array Reference = Reference to an array of aggrs

=head2 B<max_wait>

String = Maximum wait time in seconds

=head2 B<interval>

String = Sleep interval in seconds

=head1 Public Methods

=head2 B<new()>

Creates a new NdvmHelper object.

=over 4

=item Inputs

=item * -filer

Object of the filer

=item * -volume

Object of a volume or a string containing the volume name.

=item * -aggrs

String conatining a colon separated list of 2 aggrs to be used in the move. One
must be the source and the other must be the destination.  Order is not important.

=item * -interval (Default: 20)

Seconds to sleep between status checks when waiting for a vol move to complete.

=item * -max_wait (Default: 1200);

Maximum seconds to wait to allow a vol move to complete.

=item Outputs

Returns an NdvmHelper object.

=item Supported Interfaces

cli and zapi

=item Examples

 # Create an NdvmHelper object with a volume name

 $ndvm_obj = NdvmHelper->new( -filer    => $filer_obj,
                              -volume   => $vol_name,
                              -aggrs    => "$aggr1:$aggr2",
                              -interval => 10,
                              -max_wait => 600,
                            );

 if ($ndvm_obj->error) {
     $self->_set_err("Unable to create an NdvmHelper object:\n" . $ndvm_obj->errmsg());
     return $ClientUtil::FALSE;
 }

 # Create an NdvmHelper object with a volume object

 $ndvm_obj = NdvmHelper->new( -filer    => $filer_obj,
                              -volume   => $self->{vol_obj},
                              -aggrs    => "$aggr1:$aggr2",
                              -interval => 10,
                              -max_wait => 600,
                            );

 if ($ndvm_obj->error) {
     $self->_set_err("Unable to create an NdvmHelper object:\n" . $ndvm_obj->errmsg());
     return $ClientUtil::FALSE;

=back

=head2 B<ndvm_start(I<%args>)>

Start a non-disruptive volume move

=over 4

=item Inputs

=item * -unexport

(Optional) Unexport the volume if it is exported.  1 = True,  0 = False.  Default is 0.

=item * All inputs from San::SanFiler::Volume::move_start()

See POD documentation for San::SanFiler::Volume::move_start() for the complete list of options.

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Automatic Cutover with Wait

    $ndvm_obj->ndvm_start( -unexport  => 1,
                           -aggr      => <destination aggr>,
                           -wait_flag => 1);
    if ($ndvm_obj->error())
        ...
    }

    # Automatic Cutover with Wait and other options

    $ndvm_obj->ndvm_start( -unexport  => 1,
                           -aggr      => <destination aggr>,
                           -max_wait  => 900,
                           -interval  => 30,
                           -wait_flag => 1);
    if ($ndvm_obj->error())
        ...
    }

    # Manual Cutover

    $ndvm_obj->ndvm_start( -unexport       => 1,
                           -aggr           => <destination aggr>,
                           -manual_cutover => 1);
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<ndvm_status>

Get the status of a non-disruptive volume move

=over 4

=item Inputs

=item * -verbose

(Optional) Returns verbose or non-verbose status.  1 = True,  0 = False.  Default is 0.

=item Outputs

Returns reference to a hash containing the status output returned from a vol move status cli call or a
volume-move-status zapi command.  The hash reference will be in the form of:

Verbose Status:

$output_hash->{
    'migration-state => 'running',
    'vol-move-status-info' => {
        'cutover-attempts' => '6',
        'move-state' => ' setup (paused)',
        'cutover-window' => '88',
        'last-transfer-size' => '0',
        'Last-transfer-duration' => '0',
        'temp-volume' => 'ndm_dstvol_1267950718',
        'dest-aggregate' => 'destaggr',
        'source-volume-name' => 'NdvmSrcVol1',
        'src-aggregate' => 'testaggr',
    }
}

Non-Verbose Status:

$output_hash->{
    'migration-state => 'running',
    'vol-move-status-info' => {
        'cutover-attempts' => '6',
        'move-state' => ' setup (paused)',
        'cutover-window' => '88',
        'dest-aggregate' => 'destaggr',
        'source-volume-name' => 'NdvmSrcVol1',
        'src-aggregate' => 'testaggr',
    }
}


=item Supported Interfaces:

zapi and cli

=item Examples

    Verbose Status:

    my $output_hash = $ndvm_obj->ndvm_status(-verbose => 1);
    if ($ndvm_obj->error())
        ...
    }

    unless (defined($output_hash->{'vol-move-status-info'}->{'move-state'})
            and $output_hash->{'vol-move-status-info'}->{'move-state'} =~ /paused/) {
        $self->_set_err("Unable to resume the move of volume $volume.  Move state is currently " .
                        $output_hash->{'vol-move-status-info'}->{'move-state'});
        return $ClientUtil::FALSE;
    }

=back

=head2 B<ndvm_cutover(I<%args>)>

Perform a manual cutover of a non-disruptive volume move

=over 4

=item Inputs

=item * -unexport

(Optional) Unexport the volume if it is exported.  1 = True,  0 = False.  Default is 0.

=item * -num_cutover

(Optional) Specifies the number of cutover attempts. Default is 3.

=item * All inputs from San::SanFiler::Volume::move_cutover()

See POD documentation for San::SanFiler::Volume::move_cutover() for the complete list of options.

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Cutover with modified  cutover window and unexport

    $ndvm_obj->ndvm_cutover( -unexport       => 1,
                             -num_cutover    => 5,
                             -cutover_window => 90);
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<ndvm_abort();

Abort a non-disruptive volume move

=over 4

=item Inputs

None

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Abort a move

    $ndvm_obj->ndvm_abort();
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<ndvm_pause();

Pause a non-disruptive volume move

=over 4

=item Inputs

None

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Pause a move

    $ndvm_obj->ndvm_pause();
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<ndvm_resume(I<%args>)>

Resume a paused non-disruptive volume move

=over 4

=item Inputs

=item * -unexport

(Optional) Unexport the volume if it is exported.  1 = True,  0 = False.  Default is 0.

=item * All inputs from San::SanFiler::Volume::move_resume()

See POD documentation for San::SanFiler::Volume::move_resume() for the complete list of options.

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Resume with an automatic cutover and wait

    $ndvm_obj->ndvm_resume( -unexport  => 1,
                            -aggr      => <destination aggr>,
                            -wait_flag => 1);
    if ($ndvm_obj->error())
        ...
    }

    # Resume with an automatic cutover, wait and other options

    $ndvm_obj->ndvm_resume( -unexport        => 1,
                            -aggr            => <destination aggr>,
                            -max_wait        => 900,
                            -interval        => 30,
                            -num_cutover     => 5,
                            -cutover_window  => 90,
                            -wait_flag       => 1);
    if ($ndvm_obj->error())
        ...
    }

    # Resume with manual Cutover

    $ndvm_obj->ndvm_resume( -unexport       => 1,
                            -aggr           => <destination aggr>,
                            -manual_cutover => 1);
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<get_dest_aggr()>

Retrieve and return a string containing the destination aggregate for the ndvm object

=over 4

=item Inputs

None

=item Outputs

Returns a string containing the destination aggregate.  If there is an error it returns undef
and sets error.

=item Supported Interfaces:

N/A

=item Examples

    $dest_aggr = $ndvm_obj->get_dest_aggr()
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<get_filer_name()>

Returns the filer name off of the ndvm object.

=over 4

=item Inputs

None

=item Outputs

Returns a string containing the filer name.  If there is an error it returns undef and sets error.

=item Supported Interfaces:

N/A

=item Examples

    $filer_name = $ndvm_obj->get_filer_name()
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<get_filer_object()>

Returns the filer object off of the ndvm object.

=over 4

=item Inputs

None

=item Outputs

Returns a filer object.  If there is an error it returns {} and sets error.

=item Supported Interfaces:

N/A

=item Examples

    $filer_obj = $ndvm_obj->get_filer_object()
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<get_vol_name()>

Returns the volume name off of the ndvm object.

=over 4

=item Inputs

None

=item Outputs

Returns a string containing the volume name.  If there is an error it returns undef and sets error.

=item Supported Interfaces:

N/A

=item Examples

    $vol_name = $ndvm_obj->get_vol_name()
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<get_vol_object()>

Returns the volume object off of the ndvm object.

=over 4

=item Inputs

None

=item Outputs

Returns a volume object.  If there is an error it returns {} and sets error.

=item Supported Interfaces:

N/A

=item Examples

    $vol_obj = $ndvm_obj->get_vol_object()
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<get_interval()>

Returns the sleep interval defined on the ndvm object.

=over 4

=item Inputs

None

=item Outputs

Returns a string containing the sleep interval.  If there is an error it returns undef and sets error.

=item Supported Interfaces:

N/A

=item Examples

    $interval = $ndvm_obj->get_interval()
    if ($ndvm_obj->error())
        ...
    }

=back

=head2 B<get_max_wait()>

Returns the maximum wait time defined on the ndvm object.

=over 4

=item Inputs

None

=item Outputs

Returns a string containing the maximum wait time.  If there is an error it returns undef and sets error.

=item Supported Interfaces:

N/A

=item Examples

    $max_wait = $ndvm_obj->get_max_wait()
    if ($ndvm_obj->error())
        ...
    }

=back

=cut
