# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/San/Task/PortsetHelper.pm#1 $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Utility for testing portset functionality
## To provide library functions for the testing portset.
## All the cmode portset testing script will use this library
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::Task::PortsetHelper;

# compiler directives
use strict;
use Params::Validate qw( :all );
use San::Application::Portsetconstant;
use San::SanUtil;
use Data::Dumper;

our @ISA = qw( SanUtil );

# constant variable
use constant NOERRORFOUND => 'NoErrorFound';

###############################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {};

    # bless instance into class
    bless $self, $class;

    return $self;
}

###############################################################################
# Method:    create_portset_in_vserver
# Objective: to create portset object in given vserver and build the expected
#            portset attributes (as user requested) for later verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_portset_in_vserver {
    my ($self, %args) = @_;
    my ($vs_obj, $vs_name, $ps_attr_expected, $fcp_expt_attr, $port_names);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(PS_PARAM_VSERVER_OBJ) => {type     => OBJECT,
                                               required => 1,
                                               isa      => 'San::ControllerCluster::Filer10M::Vserver'},
                    (PS_ATTR_PORTSET_NAME) => {type => SCALAR, optional => 1},
                    (PS_ATTR_PROTOCOL)     => {type => SCALAR, optional => 1},
                    (PS_ATTR_PORT_NAME)    => {type => SCALAR, optional => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return {};
    }

    # remove this arg from args list, since we are using same args list to
    # send it to vs_obj->create_portset_obj(), which is invalid argument.
    $vs_obj = $args{(PS_PARAM_VSERVER_OBJ)};
    delete $args{(PS_PARAM_VSERVER_OBJ)};

    # get the vserver object and the name of the vserver
    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get vserver name.\nReason:\n" . $vs_obj->errmsg());
        return {};
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg       => _dump_message(\%args),
                  -msg_level => 3);

    # for zapi portocol is must parameter. if user does not pass
    # and cmd_interface is zapi, set it to default(mixed);
    my $cmd_interface = $vs_obj->get_cmd_interface();

    if ($cmd_interface =~ /zapi/i) {
        unless (exists($args{(PS_ATTR_PROTOCOL)})) {
            $args{(PS_ATTR_PROTOCOL)} = 'mixed';
        }

        # zapi does not supoort to create portset with port-name
        # so remove it here and add those port later...
        if (exists($args{(PS_ATTR_PORT_NAME)})) {
            $port_names = delete $args{(PS_ATTR_PORT_NAME)};
        }
    }

    # create portset object
    my $ps_obj = $vs_obj->create_portset_obj(%args);
    if ($vs_obj->error()) {
        $self->_set_err(  "Failed to create portset("
                        . $args{'portset'} . ") "
                        . "on vserver($vs_name)\nReason:\n"
                        . $vs_obj->errmsg());
        return {};
    }

    # build the hash - expected portset attributes for later verification
    # portset show command will have following attributes for given portset.
    # vserver|portset|protocol-type|port-name|port-count|igroups|
    # if any attribute is not passed fill with "-" as default,
    # since portset show output will have "-"
    # Sample output
    # ps_basic_vs|ps_basic_create_fcp_2|fcp|-|0|-|

    $ps_attr_expected->{(PS_ATTR_VSERVER_NAME)} = $vs_name;
    $ps_attr_expected->{(PS_ATTR_PORTSET_NAME)} = $args{(PS_ATTR_PORTSET_NAME)};

    # if protocol is not specified 'portset create' default to 'mixed'
    if (exists $args{(PS_ATTR_PROTOCOL)}) {
        $ps_attr_expected->{(PS_ATTR_PROTOCOL)} = $args{(PS_ATTR_PROTOCOL)};
    } else {
        $ps_attr_expected->{(PS_ATTR_PROTOCOL)} = 'mixed';
    }

    # if port-name is not specified 'portset create' default to '-' and
    # port-count to 0.
    if (exists $args{(PS_ATTR_PORT_NAME)}) {
        $ps_attr_expected->{(PS_ATTR_PORT_NAME)} = $args{(PS_ATTR_PORT_NAME)};
        my @ports = split(",", $args{(PS_ATTR_PORT_NAME)});
        my $no_of_ports_added = @ports;
        $ps_attr_expected->{(PS_ATTR_PORT_COUNT)} = $no_of_ports_added;
    } else {
        $ps_attr_expected->{(PS_ATTR_PORT_NAME)}  = '-';
        $ps_attr_expected->{(PS_ATTR_PORT_COUNT)} = 0;
    }

    # igroups always default to "-"
    $ps_attr_expected->{(PS_ATTR_IGROUP)} = '-';

    # fill up the portset object also....
    $ps_attr_expected->{(PS_ATTR_PORTSET_OBJ)} = $ps_obj;

    # if cmd_interface is zapi and user requested to add ports, then do it after creating the portset.
    if ($cmd_interface =~ /zapi/i) {
        if (defined($port_names)) {
            $self->add_ports_to_portset((PS_PARAM_PS_ATTR_INFO) => $ps_attr_expected,
                                        (PS_ATTR_PORT_NAME)     => $port_names);
            if ($self->error()) {
                my $err_str = $self->errmsg();
                $self->_clear_err();
                $self->_set_err(  "Failed to add ports while creating portset through zapi("
                                . $args{'portset'} . ") "
                                . "on vserver($vs_name)\nReason:\n"
                                . $err_str);

                # since portset is created, remove the portset
                # and return the empty string...
                $ps_obj->delete('-force' => "true");
                return {};
            }
        }
    }

    # dump the expected portset attributes for debug purpose
    $self->sanlog(-msg       => "Expected Portset Attributes after create:\n" . _dump_message($ps_attr_expected, 1),
                  -msg_level => 3);
    return $ps_attr_expected;
}    # end sub create_portset_in_vserver

###############################################################################
# Method:    add_ports_to_portset
# Objective: to add fcp or iscsi ports to existing portset and update the
#            expected portset attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_ports_to_portset {
    my ($self, %args) = @_;
    my ($ps_obj, $ps_name, $ps_attr_expected, $fcp_expt_attr);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(PS_PARAM_PS_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Attributes not qualified" => sub {
                         $_[0]->{(PS_ATTR_PORTSET_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Portset/
                           and exists $_[0]->{(PS_ATTR_PORTSET_NAME)}
                           and exists $_[0]->{(PS_ATTR_VSERVER_NAME)}
                           and exists $_[0]->{(PS_ATTR_PROTOCOL)}
                           and exists $_[0]->{(PS_ATTR_PORT_NAME)}
                           and exists $_[0]->{(PS_ATTR_IGROUP)}
                           and exists $_[0]->{(PS_ATTR_PORT_COUNT)};
                       }
                 }},
             (PS_ATTR_PORT_NAME) => {type => SCALAR, required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # remove this arg($ps_attr_expected) from args list,
    # since we are using same args list to
    # send it to $ps_obj->add(), which is invalid argument.
    $ps_attr_expected = $args{(PS_PARAM_PS_ATTR_INFO)};
    delete $args{(PS_PARAM_PS_ATTR_INFO)};
    $ps_obj  = $ps_attr_expected->{(PS_ATTR_PORTSET_OBJ)};
    $ps_name = $ps_attr_expected->{(PS_ATTR_PORTSET_NAME)};

    my @ports = split(",", $args{(PS_ATTR_PORT_NAME)});

    # add requested ports to portset
    my $cmd_interface = $ps_obj->get_cmd_interface();

    # through zapi we can add only one port at a time
    if ($cmd_interface =~ /zapi/i) {
        foreach my $port_name (@ports) {
            $ps_obj->add((PS_ATTR_PORT_NAME) => $port_name);
            if ($ps_obj->error()) {
                $self->_set_err(
                             "Failed to add port to " . "portset($ps_name) using zapi \nReason:\n" . $ps_obj->errmsg());
                return $SanUtil::FALSE;
            }
            if ($ps_attr_expected->{(PS_ATTR_PORT_NAME)} eq "-") {
                $ps_attr_expected->{(PS_ATTR_PORT_NAME)} = $port_name;
            } else {

                # if some information already exist just append the current with comma
                # separated to expected portset attribute
                $ps_attr_expected->{(PS_ATTR_PORT_NAME)} .= "," . $port_name;
            }
            $ps_attr_expected->{(PS_ATTR_PORT_COUNT)}++;
        }    # end of foreach my $port_name (@ports)
    } else {
        $ps_obj->add(%args);
        if ($ps_obj->error()) {
            $self->_set_err("Failed to add port(s) to " . "portset($ps_name)\nReason:\n" . $ps_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # update the expected portset attributes - PORT_NAME
        # if no information exists just add the current port info to
        # expected portset attribute
        my $no_of_ports_added = @ports;
        if ($ps_attr_expected->{(PS_ATTR_PORT_NAME)} eq "-") {
            $ps_attr_expected->{(PS_ATTR_PORT_NAME)}  = $args{(PS_ATTR_PORT_NAME)};
            $ps_attr_expected->{(PS_ATTR_PORT_COUNT)} = $no_of_ports_added;
        } else {

            # if some information already exist just append the current with comma
            # separated to expected portset attribute
            $ps_attr_expected->{(PS_ATTR_PORT_NAME)} .= "," . $args{(PS_ATTR_PORT_NAME)};
            $ps_attr_expected->{(PS_ATTR_PORT_COUNT)} += $no_of_ports_added;
        }
    }

    # dump the expected portset attributes for debug purpose
    # just dump the first level of hash
    $self->sanlog(-msg       => "Expected Portset Attributes after add:\n" . _dump_message($ps_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub add_ports_to_portset

###############################################################################
# Method:    remove_ports_from_portset
# Objective: to remove fcp or iscsi ports from existing portset and update the
#            expected portset attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_ports_from_portset {
    my ($self, %args) = @_;
    my ($ps_obj, $ps_name, $ps_attr_expected, $fcp_expt_attr);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(PS_PARAM_PS_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Attributes not qualified" => sub {
                         $_[0]->{(PS_ATTR_PORTSET_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Portset/
                           and exists $_[0]->{(PS_ATTR_PORTSET_NAME)}
                           and exists $_[0]->{(PS_ATTR_VSERVER_NAME)}
                           and exists $_[0]->{(PS_ATTR_PROTOCOL)}
                           and exists $_[0]->{(PS_ATTR_PORT_NAME)}
                           and exists $_[0]->{(PS_ATTR_IGROUP)}
                           and exists $_[0]->{(PS_ATTR_PORT_COUNT)};
                       }
                 }},
             (PS_ATTR_PORT_NAME) => {type => SCALAR, required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # remove this arg($ps_attr_expected) from args list,
    # since we are using same args list to
    # send it to $ps_obj->remove(), which is invalid argument.
    $ps_attr_expected = $args{(PS_PARAM_PS_ATTR_INFO)};
    delete $args{(PS_PARAM_PS_ATTR_INFO)};

    $ps_obj  = $ps_attr_expected->{(PS_ATTR_PORTSET_OBJ)};
    $ps_name = $ps_attr_expected->{(PS_ATTR_PORTSET_NAME)};

    my @ports_to_remove = split(',', $args{(PS_ATTR_PORT_NAME)});

    # add requested ports to portset
    my $cmd_interface = $ps_obj->get_cmd_interface();

    # through zapi we can only remove one port at a time
    # remove requested ports from portset
    if ($cmd_interface =~ /zapi/i) {
        foreach my $port_name (@ports_to_remove) {
            $ps_obj->remove((PS_ATTR_PORT_NAME) => $port_name);
            if ($ps_obj->error()) {
                $self->_set_err(
                        "Failed to remove port from " . "portset($ps_name) using zapi \nReason:\n" . $ps_obj->errmsg());
                return $SanUtil::FALSE;
            }
            $ps_attr_expected->{(PS_ATTR_PORT_NAME)} =~ s/\b$port_name\b,*//;
            $ps_attr_expected->{(PS_ATTR_PORT_COUNT)}--;

            # remove if there is any comma at end...
            $ps_attr_expected->{(PS_ATTR_PORT_NAME)} =~ s/,$//;
            unless ($ps_attr_expected->{(PS_ATTR_PORT_NAME)}) {
                $ps_attr_expected->{(PS_ATTR_PORT_NAME)} = "-";
            }
        }
    } else {
        $ps_obj->remove(%args);
        if ($ps_obj->error()) {
            $self->_set_err("Failed to remove port(s) from " . "portset($ps_name)\nReason:\n" . $ps_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # update the expected portset attributes

        my $no_of_ports_removed = @ports_to_remove;

        # update the port count
        $ps_attr_expected->{(PS_ATTR_PORT_COUNT)} -= $no_of_ports_removed;
        foreach my $port (@ports_to_remove) {
            $ps_attr_expected->{(PS_ATTR_PORT_NAME)} =~ s/\b$port\b,*//;
        }

        # remove if there is any comma at end...
        $ps_attr_expected->{(PS_ATTR_PORT_NAME)} =~ s/,$//;
        unless ($ps_attr_expected->{(PS_ATTR_PORT_NAME)}) {
            $ps_attr_expected->{(PS_ATTR_PORT_NAME)} = "-";
        }
    }

    # dump the expected portset attributes for debug purpose
    # just dump the first level of hash
    $self->sanlog(-msg       => "Expected Portset Attributes after remove:\n" . _dump_message($ps_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub remove_ports_from_portset

###############################################################################
# Method:    bind_portset_to_igroup
# Objective: to bind the portset to igroup and update the
#            expected portset attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub bind_portset_to_igroup {
    my ($self, %args) = @_;
    my ($ps_obj, $ps_name, $ps_attr_expected, $igroup_obj, $igroup_name);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(PS_PARAM_PS_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Attributes not qualified" => sub {
                         $_[0]->{(PS_ATTR_PORTSET_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Portset/
                           and exists $_[0]->{(PS_ATTR_PORTSET_NAME)}
                           and exists $_[0]->{(PS_ATTR_VSERVER_NAME)}
                           and exists $_[0]->{(PS_ATTR_PROTOCOL)}
                           and exists $_[0]->{(PS_ATTR_PORT_NAME)}
                           and exists $_[0]->{(PS_ATTR_IGROUP)}
                           and exists $_[0]->{(PS_ATTR_PORT_COUNT)};
                       }
                 }},
             (PS_PARAM_IGROUP_OBJ) => {type     => OBJECT,
                                       required => 1,
                                       isa      => 'San::ControllerCluster::Filer10M::Vserver::Igroup'}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    $ps_attr_expected = $args{(PS_PARAM_PS_ATTR_INFO)};
    $ps_obj           = $ps_attr_expected->{(PS_ATTR_PORTSET_OBJ)};
    $ps_name          = $ps_attr_expected->{(PS_ATTR_PORTSET_NAME)};

    $igroup_obj  = $args{(PS_PARAM_IGROUP_OBJ)};
    $igroup_name = $igroup_obj->get_name();
    if ($igroup_obj->error()) {
        $self->_set_err("Unable to get igroup name. Reason:\n" . $igroup_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # bind portset to the given igroup

    $igroup_obj->bind((PS_ATTR_PORTSET_NAME) => $ps_name);
    if ($igroup_obj->error()) {
        $self->_set_err(
            "Unable to bind portset($ps_name) " . "to igroup($igroup_name)
                  . Reason:\n"
              . $igroup_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected portset attributes
    if ($ps_attr_expected->{(PS_ATTR_IGROUP)} eq "-") {
        $ps_attr_expected->{(PS_ATTR_IGROUP)} = $igroup_name;
    } else {
        $ps_attr_expected->{(PS_ATTR_IGROUP)} .= "," . $igroup_name;
    }

    # dump the expected portset attributes for debug purpose
    # just dump the first level of hash
    $self->sanlog(-msg       => "Expected Portset Attributes after bind:\n" . _dump_message($ps_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub bind_portset_to_igroup

###############################################################################
# Method:    unbind_portset_from_igroup
# Objective: to unbind the portset from the igroup and update the
#            expected portset attributes for verification
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unbind_portset_from_igroup {
    my ($self, %args) = @_;
    my ($ps_obj, $ps_name, $ps_attr_expected, $igroup_obj, $igroup_name);
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(PS_PARAM_PS_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Attributes not qualified" => sub {
                         $_[0]->{(PS_ATTR_PORTSET_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Portset/
                           and exists $_[0]->{(PS_ATTR_PORTSET_NAME)}
                           and exists $_[0]->{(PS_ATTR_VSERVER_NAME)}
                           and exists $_[0]->{(PS_ATTR_PROTOCOL)}
                           and exists $_[0]->{(PS_ATTR_PORT_NAME)}
                           and exists $_[0]->{(PS_ATTR_IGROUP)}
                           and exists $_[0]->{(PS_ATTR_PORT_COUNT)};
                       }
                 }},
             (PS_PARAM_IGROUP_OBJ) => {type     => OBJECT,
                                       required => 1,
                                       isa      => 'San::ControllerCluster::Filer10M::Vserver::Igroup'}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }
    $ps_attr_expected = $args{(PS_PARAM_PS_ATTR_INFO)};
    $ps_obj           = $ps_attr_expected->{(PS_ATTR_PORTSET_OBJ)};
    $ps_name          = $ps_attr_expected->{(PS_ATTR_PORTSET_NAME)};

    $igroup_obj  = $args{(PS_PARAM_IGROUP_OBJ)};
    $igroup_name = $igroup_obj->get_name();
    if ($igroup_obj->error()) {
        $self->_set_err("Unable to get igroup name. Reason:\n" . $igroup_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # unbind portset from the given igroup

    $igroup_obj->unbind();
    if ($igroup_obj->error()) {
        $self->_set_err(
                "Unable to unbind portset($ps_name) " . "from igroup($igroup_name). Reason:\n" . $igroup_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # update the expected portset attributes
    # remove the igroup from portset attributes
    $ps_attr_expected->{(PS_ATTR_IGROUP)} =~ s/\b$igroup_name\b,*//;

    # remove if there is any comma at end...
    $ps_attr_expected->{(PS_ATTR_IGROUP)} =~ s/,$//;
    unless ($ps_attr_expected->{(PS_ATTR_IGROUP)}) {
        $ps_attr_expected->{(PS_ATTR_IGROUP)} = "-";
    }

    # dump the expected portset attributes for debug purpose
    # just dump the first level of hash
    $self->sanlog(-msg       => "Expected Portset Attributes after unbind:\n" . _dump_message($ps_attr_expected, 1),
                  -msg_level => 3);
    return $SanUtil::TRUE;
}    # end sub unbind_portset_from_igroup

###############################################################################
# Method:    verify_portset_attributes
# Objective: to verify the portset attributes are intact.
#            compare the requested attributes with the controller
#            return portset attributes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_portset_attributes {
    my ($self, %args) = @_;
    my ($ps_obj, $ps_name, %ps_attr_user, $ps_attr_controller, @err_str);
    my $err_cnt = 0;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(PS_PARAM_PS_ATTR_INFO) => {
                 type      => HASHREF,
                 required  => 1,
                 callbacks => {
                     "Attributes not qualified" => sub {
                         $_[0]->{(PS_ATTR_PORTSET_OBJ)} =~ /San::ControllerCluster::Filer10M::Vserver::Portset/
                           and exists $_[0]->{(PS_ATTR_PORTSET_NAME)}
                           and exists $_[0]->{(PS_ATTR_VSERVER_NAME)}
                           and exists $_[0]->{(PS_ATTR_PROTOCOL)}
                           and exists $_[0]->{(PS_ATTR_PORT_NAME)}
                           and exists $_[0]->{(PS_ATTR_IGROUP)}
                           and exists $_[0]->{(PS_ATTR_PORT_COUNT)};
                       }
                 }}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # copy the attributes, since we need to delete the portset_obj,
    # do not operate on reference
    %ps_attr_user = %{$args{(PS_PARAM_PS_ATTR_INFO)}};
    $ps_obj       = $ps_attr_user{(PS_ATTR_PORTSET_OBJ)};

    # since portset_obj is not part of portset attribute,
    # discard from $ps_attr_user
    delete $ps_attr_user{(PS_ATTR_PORTSET_OBJ)};

    # get the name of portset from "$ps_attr_user"
    $ps_name = $ps_attr_user{(PS_ATTR_PORTSET_NAME)};

    # get the portset info from controller, return reference to hash
    $ps_attr_controller = $ps_obj->get_list_info();
    if ($ps_obj->error()) {
        $self->_set_err("Unable to get portset information." . $ps_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # in case, if there is no requested portset,
    # 'get_list_info' won't return error, so make sure we have valid info.
    # ex:
    # > lun portset show -vserver ps_basic_mv_vs0 -portset _DELETE_
    # > There are no entries matching your query.
    # this will not return error
    unless (ref($ps_attr_controller) =~ /HASH/
            and exists $ps_attr_controller->{(PS_ATTR_PORTSET_NAME)}) {
        $self->_set_err("Not a valid portset information.\n" . "Controller return: $ps_attr_controller");
        return $SanUtil::FALSE;
    }

    # display the user request portset attributes if debug_level is high
    $self->sanlog(-msg       => "verify_portset_attributes: Expected:\n" . _dump_message(\%ps_attr_user),
                  -msg_level => 3);

    # display the controller portset attributes if debug_level is high
    $self->sanlog(-msg       => "verify_portset_attributes: Controller:\n" . _dump_message($ps_attr_controller),
                  -msg_level => 3);

    # verify the portset attributes from user expected list
    # with controller list and return one error message with
    # list of attributes which is/are not matching. User can set
    # _SKIP_CHECK_ value for particular key to skip checking.
    # for keys fc_lif|wwpn|tpg, the value will be comma separated list.
    # since order of that list will be different, need to check individual
    # values present in the controller list. "_verify_portset_key_values"
    # does that.
    @err_str = ("Portset($ps_name) attributes does not match");
    foreach my $key (keys %ps_attr_user) {

        # if caller did not care about this particular attribute
        if ($ps_attr_user{$key} =~ /_SKIP_CHECK_/) {
            next;
        }
        unless (defined $ps_attr_controller->{$key}) {
            $err_cnt++;
            push(@err_str, "\t$key is not defined in the controller list info");
            next;
        }

        # if the checking key (port-name or igroups) may have
        # comma separated values and
        # it may return in  different order,
        # so check for existence in the list.
        if ($key =~ /[${\PS_ATTR_PORT_NAME}|${\PS_ATTR_IGROUP}]/) {
            my @verify_array = _verify_portset_key_values($key, $ps_attr_user{$key}, $ps_attr_controller->{$key});
            if ($verify_array[0] ne NOERRORFOUND) {
                $err_cnt++;
                push(@err_str, @verify_array);
            }
        } else {
            if ($ps_attr_user{$key} ne $ps_attr_controller->{$key}) {
                $err_cnt++;
                push(@err_str,
                         "\t$key => Expected: "
                       . "'$ps_attr_user{$key}' <-> "
                       . "Controller: '$ps_attr_controller->{$key}'");
            }
        }
    }
    if ($err_cnt) {
        my $text = "No. of errors encountered: $err_cnt\n" . join("\n", @err_str);
        $self->_set_err($text);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}    # end sub verify_portset_attributes

###############################################################################
# Method:    verify_portset_igroup_bind
# Objective: to verify the portset and igroup bind
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_portset_igroup_bind {
    my ($self, %args) = @_;

    my ($ps_obj, $igroup_obj, $igrp_attr_controller, $ps_attr_controller);
    my ($ps_name, $igroup_name, $verify_presence);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(PS_PARAM_IGROUP_OBJ) => {type     => OBJECT,
                                              required => 1,
                                              isa      => 'San::ControllerCluster::Filer10M::Vserver::Igroup'},
                    (PS_ATTR_PORTSET_OBJ) => {type     => OBJECT,
                                              required => 1,
                                              isa      => 'San::ControllerCluster::Filer10M::Vserver::Portset'},
                    'verify_presence' => {type     => SCALAR,
                                          optional => 1,
                                          default  => "yes"}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }
    $verify_presence = $args{'verify_presence'};
    $igroup_obj      = $args{(PS_PARAM_IGROUP_OBJ)};
    $igroup_name     = $igroup_obj->get_name();
    if ($igroup_obj->error()) {
        $self->_set_err("Unable to get igroup name. Reason:\n" . $igroup_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $ps_obj  = $args{(PS_ATTR_PORTSET_OBJ)};
    $ps_name = $ps_obj->get_name();
    if ($ps_obj->error()) {
        $self->_set_err("Unable to get portset name. Reason:\n" . $ps_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # get the portset info from controller, return reference to hash
    $ps_attr_controller = $ps_obj->get_list_info();
    if ($ps_obj->error()) {
        $self->_set_err("Unable to get portset information." . $ps_obj->errmsg());
        return $SanUtil::FALSE;
    }
    if ($verify_presence =~ /yes/i) {

        # just need to make sure this igroup exist in the list.
        # verify igroup info has the portset name on it
        if ($ps_attr_controller->{(PS_ATTR_IGROUP)} !~ /$igroup_name/) {
            $self->_set_err(  "igroup name: $igroup_name is not presence "
                            . "in portset information:\n"
                            . "Controller return: "
                            . $ps_attr_controller->{(PS_ATTR_IGROUP)});
            return $SanUtil::FALSE;
        }
    } else {

        # just need to make sure this igroup exist in the list.
        # verify igroup info has the portset name on it
        if ($ps_attr_controller->{(PS_ATTR_IGROUP)} =~ /\b$igroup_name\b/) {
            $self->_set_err(  "Portset is still bound to "
                            . "igroup name: $igroup_name\n"
                            . "Controller return: "
                            . $ps_attr_controller->{(PS_ATTR_IGROUP)});
            return $SanUtil::FALSE;
        }
    }
    return
      $self->verify_igroup_bind('-igroup_obj'          => $igroup_obj,
                                (PS_ATTR_PORTSET_NAME) => $ps_name,
                                'verify_presence'      => $verify_presence);
}    # end sub verify_portset_igroup_bind

###############################################################################
# Method:    verify_igroup_bind
# Objective: to verify igroup does/does not have portset binding info
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_igroup_bind {
    my ($self, %args) = @_;

    my ($igroup_obj, $igrp_attr_controller);
    my ($ps_name, $igroup_name, $verify_presence);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(PS_PARAM_IGROUP_OBJ) => {type     => OBJECT,
                                              required => 1,
                                              isa      => 'San::ControllerCluster::Filer10M::Vserver::Igroup'},
                    (PS_ATTR_PORTSET_NAME) => {type     => SCALAR,
                                               required => 1},
                    'verify_presence' => {type     => SCALAR,
                                          optional => 1,
                                          default  => "yes"}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }
    $verify_presence = $args{'verify_presence'};
    $igroup_obj      = $args{(PS_PARAM_IGROUP_OBJ)};
    $igroup_name     = $igroup_obj->get_name();
    if ($igroup_obj->error()) {
        $self->_set_err("Unable to get igroup name. Reason:\n" . $igroup_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $ps_name = $args{(PS_ATTR_PORTSET_NAME)};

    # get the igroup info from controller, return reference to hash
    $igrp_attr_controller = $igroup_obj->get_list_info();
    if ($igroup_obj->error()) {
        $self->_set_err("Unable to get igroup information." . $igroup_obj->errmsg());
        return $SanUtil::FALSE;
    }

    if ($verify_presence =~ /yes/i) {

        # verify igroup info has the portset name on it
        if ($igrp_attr_controller->{(PS_ATTR_PORTSET_NAME)} ne $ps_name) {
            $self->_set_err(  "portset name: $ps_name is not presence "
                            . "in igroup information:\n"
                            . "Controller return: "
                            . $igrp_attr_controller->{(PS_ATTR_PORTSET_NAME)});
            return $SanUtil::FALSE;
        }
    } else {
        if ($igrp_attr_controller->{(PS_ATTR_PORTSET_NAME)} ne "-") {
            $self->_set_err(  "Igroup has portset bind to it\n"
                            . "Controller return: "
                            . $igrp_attr_controller->{(PS_ATTR_PORTSET_NAME)});
            return $SanUtil::FALSE;
        }
    }
    return $SanUtil::TRUE;
}    # end sub verify_igroup_bind

###############################################################################
# Method:    get_portset_attributes
# Objective: To get the Portset attributes hash for a given a portset object for
#            using PortsetHelper function. Return portset expected attribute
#            hash same like create_portset_in_vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_portset_attributes {
    my ($self, %args) = @_;

    my ($ps_obj, $ps_attrs_controller, $ps_attr_expected);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(PS_ATTR_PORTSET_OBJ) => {type     => OBJECT,
                                              required => 1,
                                              isa      => 'San::ControllerCluster::Filer10M::Vserver::Portset'}});
    };
    if ($@) {
        $self->_set_err($@);
        return {};
    }
    $ps_obj = $args{(PS_ATTR_PORTSET_OBJ)};

    # get the portset info from controller, return reference to hash
    $ps_attrs_controller = $ps_obj->get_list_info();
    if ($ps_obj->error()) {
        $self->_set_err("Unable to get portset information." . $ps_obj->errmsg());
        return {};
    }

    foreach my $ps_attr (@ExpectedPortsetAttributes) {
        $ps_attr_expected->{$ps_attr} = $ps_attrs_controller->{$ps_attr};
    }
    $ps_attr_expected->{(PS_ATTR_PORTSET_OBJ)} = $ps_obj;

    # dump the expected portset attributes for debug purpose
    $self->sanlog(
         -msg       => "Expected Portset Attributes in get_portset_attributes:\n" . _dump_message($ps_attr_expected, 1),
         -msg_level => 3);
    return $ps_attr_expected;
}    # end sub get_portset_attributes

###############################################################################
#                        Private Functions
###############################################################################

###############################################################################
# Method:    _verify_portset_key_values
# Objective: to verify the given portset attributes are intact.
#            compare the requested attributes with the controller
# Parameter:  all param are required
#   $key                   -> key [lifs|wwpn|tpg] name of the verifying key
#   $list_to_check         -> list of lif|wwpn|tgp name with comma separated
#   $list_from_controller  -> list of lif|wwpn|tgp name with comma separated
# Return:  array of error string. If nothing found return array with
#          single element NOERRORFOUND.
# Details:   POD is not required for local(internal) function.
###############################################################################
sub _verify_portset_key_values {
    my ($key, $list_to_check, $list_from_controller) = @_;

    my (@err_str, @arry_to_check, @arry_from_controller);
    my $error_found = 0;

    # since it is local function no need to do param validation.
    @arry_to_check        = split(',', $list_to_check);
    @arry_from_controller = split(',', $list_from_controller);

    # "portset show" might have attributes (fc_lifs and/or tpg) with multiple
    # name with comma separated. The list is not same as created. So we need
    # verify all the name is presented as expected(instead simply checking
    # the whole string, since ORDER is different)., so sort the arrays
    @arry_to_check        = sort(@arry_to_check);
    @arry_from_controller = sort(@arry_from_controller);

    # form the error message with what key we verified, what is expected and
    # what controller sees
    push(@err_str,
         "\t$key =>\n\tExpected: " . join(",", @arry_to_check) . "\n\tController: " . join(",", @arry_from_controller));

    if (scalar(@arry_to_check) != scalar(@arry_from_controller)) {
        push(@err_str, "\t\tNumber of entries in '$key' does not match");
        return @err_str;
    }

    # now check all entries are presented as expected
    my $noofentries = @arry_to_check;
    for (my $i = 0; $i < $noofentries; $i++) {
        if ($arry_to_check[$i] ne $arry_from_controller[$i]) {
            $error_found++;
            push(@err_str, "\t\t$arry_to_check[$i] not found");
        }
    }
    if ($error_found) {
        return @err_str;
    }
    return (NOERRORFOUND);
}    # end sub _verify_portset_key_values

###############################################################################
# Method:    _dump_message
# Objective: Get output from object
# Input:     Any hash or array object
# Return:    string from the dump
###############################################################################
sub _dump_message {
    my ($obj, $hash_level) = @_;

    # create new Data::Dumper object
    my $d_obj = Data::Dumper->new([$obj], [ref($obj)]);

    # Make Data::Dumper pretty, if it is hash
    if (ref($obj) =~ /HASH/) {
        $d_obj->Sortkeys(1);
        $d_obj->Indent(1);
    }
    $d_obj->Maxdepth($hash_level) if (length($hash_level));
    return $d_obj->Dump();
}    # end of _dump_message

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

PortsetHelper.pm

=head1 Author

Muthukumaran Govindarajan (muthug@netapp.com)

=head1 Description

PortsetHelper.pm provides an utility function interface to test portset functionality
with the main intent to avoid repeated code from thpl level. It is design as Object Oriented model.
Other object(s) can derive this as base class or user can create object from it.

=head1 See Also

PortsetHelper.pm package is used as base class from San::Application::PortsetMgr. See San::Application::PortsetMgr
POD for more details.

B<San::Application::PortsetMgr>

=head1 Attributes

None

=head1 Revision History

Nov-04-2009 muthug@netapp.com: Original template created
Nov-30-2009 muthug@netapp.com: modified after initial code review
Dec-15-2009 muthug@netapp.com: used Params::Validate for input params
Jan-04-2010 muthug@netapp.com: Modified for new UI portset command
May-19-2010 muthug@netapp.com: Modified for another UI portset command
Oct-26-2010 muthug@netapp.com: Modified for zapi support

=head1 Public Methods

=head2 B<new()>

Class Constructor method. To set the debug level to print debug message at particular level, call
'set_debug_level' method. See Example for more details

=over 4

=item Inputs

None


=item Outputs

Returns a PortsetHelper blessed object reference.

=item Examples

   #
   # Create a PortsetHelper object
   #
   my $ps_helper_obj = new PortsetHelper();
   if ($ps_helper_obj->error()) {
      my $err_msg = $ps_helper_obj->errmsg();
      $Test->nlog("Failed to instantiate PortsetHelper object:\nReason: $err_msg");
      return $TCD::FAIL;
   }
   #
   # set the debug level to print debug message at particular level
   #
   $ps_helper_obj->set_debug_level($debug_level);

=back

=head2 B<create_portset_in_vserver(%args)>

Create a portset in a given vserver and return the hash reference to expected portset attributes
for later verification.

=over 4

=item Inputs

=item * I<-vserver_obj> <required> 

Reference to vserver object, where portset will be created.

=item * I<-portset> <optional>                    

Name of the portset to create, portset is must parameter. Since this function can be used
      for negative test case, code would not check for mandatory parameter

=item * I<-protocol_type> <optional> <fcp | iscsi | mixed>

Name of the portset protocol type. Can be one of these(fcp,iscsi or mixed).

NOTE: to create portset, protocol_type is must parameter. Since this function can be used
      for negative test case, code would not check for mandatory parameter.

=item * I<-port_name> <optional but required if creating "non empty" portset> 

Name of the fcp lifs and/or iSCSI Target Portal Group (tpgroup) name to add to portset while creating. 
One or more lifs/tpg can be specified. If more than one lifs/tpg, should specify with comma separated as one string.

=item Outputs

Returns a reference to hash contains portset expected attributes info if successful.
If an error occurred the error flag is set, errmsg is set to the failure reason, and {} (empty hash reference)
is passed back as the return value.

Expected attributes to simulate what controller return for portset attribute.
portset show command will have following attributes for given portset.
   vserver|portset|protocol-type|port-name|port-count|igroups|

if any attribute is not passed fill with "-" as default. Since portset show output will have "-" 

Sample output from controller.

    ps_basic_vs|ps_basic_create_fcp_2|fcp|-|0|-|
     
Hash reference looks like following:

  $ps_attr_expected->{'vserver'}
  $ps_attr_expected->{'portset'}
  $ps_attr_expected->{'protocol_type'}
  $ps_attr_expected->{'port_name'}
  $ps_attr_expected->{'port_count'}
  $ps_attr_expected->{'igroups'}
  
  additional info, which is not part of portset attributes, but we need this for later verification
  $ps_attr_expected->{'portset_obj'} - reference to portset object

=item Examples

    #
    # Instantiate PortsetHelper Object (see above _new() pod)
    # or use PortsetMgr Object as below
    #
    
    $PSMgr_obj = new PortsetMgr();
    if ($PSMgr_obj->error) {
     my $err_msg = $PSMgr_obj->errmsg();
     $Test->nlog("Failed to instantiate PortsetMgr object:\nReason: $err_msg");
     return $TCD::FAIL;
    }
    #
    # Call setup to create environment to testing
    #
    $PSMgr_obj->setup(...);
    if ($PSMgr_obj->error) {
     $Test->nlog($PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    # get vserver object
    $vs_obj = ...;
    
    
    #
    # create default(mixed) portset
    #
    my $ps_attr_exp = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_default_empty"); 
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to create default portset: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    # verify portset created as expected
    $PSMgr_obj->verify_portset_attributes('-ps_attr_info' => $ps_attr_exp);
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    #
    # create empty fcp portset
    #
    my $ps_attr_exp_fcp = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_fcp_empty", 
                                            '-protocol_type' => "fcp");
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to create empty fcp portset: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    # verify portset created as expected
    $PSMgr_obj->verify_portset_attributes('-ps_attr_info' => $ps_attr_exp_fcp);
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    #
    # create fcp portset with fcp lifs name
    #
    # get all available fcp lifs
    my $fcp_lif_names = $PSMgr_obj->get_protocol_lifs_name_in_vserver('-vserver_obj' => $vs_obj,
                                                                   'protocol_type' => 'fcp');
    
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to get fcp lifs: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    
    my $ps_attr_exp_1 = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_fcp_lif", 
                                            '-protocol_type' => "fcp",
                                            '-port_name'     => $fcp_lif_names);
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to create fcp portset: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    # verify portset created as expected
    $PSMgr_obj->verify_portset_attributes('-ps_attr_info' => $ps_attr_exp_1);
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ....
    ....
    #
    # create iscsi portset with iscsi using iscsi lif's tpg name
    #
    # get all available iscsi lifs(iscsi lif name is default tpg name)
    my $iscsi_lif_tpg_name = $PSMgr_obj->get_protocol_lifs_name_in_vserver('-vserver_obj' => $vs_obj,
                                                                   'protocol_type' => 'iscsi');
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to get iscsi lifs tpg name: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    
    my $ps_attr_exp = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_iscsi", 
                                            '-protocol_type' => "iscsi",
                                            '-port_name'     => $iscsi_lif_tpg_name);
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to create portset: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }

=back

=head2 B<add_ports_to_portset(%args)>

To add fcp or iscsi ports to existing portset and update the expected portset attributes for verification if ports added successfully

=over 4

=item Inputs

=item * I<-ps_attr_info> <required>

Hash reference to portset info, basically expects same info which return from
the function 'create_portset_in_vserver'. See create_portset_in_vserver POD for more details.
This hash reference will be updated upon successful addition of port(s) to portset.

Hash reference looks like following:
  $ps_attr_expected->{'portset_obj'} - reference to portset object, this attribute is MUST.
  $ps_attr_expected->{'vserver'}
  $ps_attr_expected->{'portset'}
  $ps_attr_expected->{'protocol_type'}
  $ps_attr_expected->{'fc_lif'}
  $ps_attr_expected->{'tpg'}
  $ps_attr_expected->{'igroups'}
  
=item * I<-fc_lif> <optional but required if adding port(s) to FCP or MIXED portset> 

Name of the fcp lifs to add to existing FCP or MIXED portset. One or more fcp lifs can be specified.
If more than one lifs, should specify with comma separated as one string.

=item * I<-tpg> <optional> 

Name of the iSCSI Target Portal Group (tpgroup) name to add to portset. 
One or more tpgroup can be specified. If more than one tpgroup, should specify with comma separated as one string.

=item Outputs

Returns TRUE  if port(s) added to portset successful. If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    # Instantiate PortsetHelper Object (see above _new() pod)
    # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
    ....
    ....
    # get vserver object
    $vs_obj = ...;
    
    # get all available fcp lifs
    my $fcp_lif_names = $PSMgr_obj->get_protocol_lifs_name_in_vserver('-vserver_obj' => $vs_obj,
                                                               'protocol_type' => 'fcp');
    
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to get fcp lifs: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    my $tpg = = $PSMgr_obj->get_protocol_lifs_name_in_vserver('-vserver_obj' => $vs_obj,
                                                               'protocol_type' => 'iscsi');
    
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to get iscsi default tpg name: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    
    # create empty fcp portset 
    my $ps_attr_exp_fcp = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_fcp_empty", 
                                            '-protocol_type' => "fcp");
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to create portset: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    # create empty iscsi portset 
    my $ps_attr_exp_iscsi = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_iscsi_empty", 
                                            '-protocol_type' => "iscsi");
    ...
    ...
    
    # now add ports to fcp portset using fc_lifs
    $PSMgr_obj->add_ports_to_portset('-ps_attr_info'      => $ps_attr_exp_fcp,
                                     '-fc_lif'          => $fcp_lif_names);
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to add port to portset: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    ...
    ...
    # now add ports to iscsi portset
    $PSMgr_obj->add_ports_to_portset('-ps_attr_info'   => $ps_attr_exp_iscsi,
                                     '-tpg'          => $tpg);
    ...
    ...

=back


=head2 B<remove_ports_from_portset(%args)>

To remove fcp or iscsi ports from existing portset and update the expected portset attributes for verification if ports removed successfully

=over 4

=item Inputs

=item * I<-ps_attr_info> <required>

Hash reference to portset info, basically expects same info which return from
the function 'create_portset_in_vserver'. See create_portset_in_vserver POD for more details.
This hash reference will be updated upon successful removal of port(s) from portset.

Hash reference looks like following:
  $ps_attr_expected->{'portset_obj'} - reference to portset object, this attribute is MUST.
  $ps_attr_expected->{'vserver'}
  $ps_attr_expected->{'portset'}
  $ps_attr_expected->{'protocol_type'}
  $ps_attr_expected->{'fc_lif'}
  $ps_attr_expected->{'tpg'}
  $ps_attr_expected->{'igroups'}
  
=item * I<-fc_lif> <optional but required if removing port(s) from FCP or MIXED portset> 

Name of the fcp lifs to remove from existing FCP or MIXED portset. One or more fcp lifs can be specified.
If more than one lifs, should specify with comma separated as one string.

=item * I<-tpg> <optional> 

Name of the iSCSI Target Portal Group (tpgroup) name to remove from portset. 
One or more tpgroup can be specified. If more than one tpgroup, should specify with comma separated as one string.

=item Outputs

Returns TRUE  if port(s) removed from portset successful. If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    # Instantiate PortsetHelper Object (see above _new() pod)
    # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
    ....
    ....
    # get vserver object
    $vs_obj = ...;
    
    # get all available fcp lifs
    my $fcp_lif_names = $PSMgr_obj->get_protocol_lifs_name_in_vserver('-vserver_obj' => $vs_obj,
                                                               'protocol_type' => 'fcp');
    
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to get fcp lifs: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    my $tpg = = $PSMgr_obj->get_protocol_lifs_name_in_vserver('-vserver_obj' => $vs_obj,
                                                               'protocol_type' => 'iscsi');
    
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to get iscsi default tpg name: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
      
    # create fcp portset using fc_lifs
    my $ps_attr_exp_fcp = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_fcp_lifs", 
                                            '-protocol_type' => "fcp",
                                            '-fc_lif'          => $fcp_lif_names);
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to create portset: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    # create iscsi portset 
    my $ps_attr_exp_iscsi = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_iscsi_tpg", 
                                            '-protocol_type' => "iscsi",
                                            '-tpg'           => $tpg);
    ...
    ...
    # now remove port from fcp portset
    $PSMgr_obj->remove_ports_from_portset('-ps_attr_info'   => $ps_attr_exp_fcp,
                                        '-fc_lif'            => "fcp_lif_1");
    ...
    ...                                        
    # verify it is removed from controller perspective
    $PSMgr_obj->verify_portset_attributes('-ps_attr_info' => $ps_attr_exp_fcp);
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    ...
    ...
    
    # now remove port from iscsi portset
    $PSMgr_obj->remove_ports_from_portset('-ps_attr_info'   => $ps_attr_exp_iscsi,
                                        '-tpg'          => "ps_iscsi_lif_e4a1_vs_ps_default");
    ...
    ...                                        
    # verify it is removed from controller perspective
    $PSMgr_obj->verify_portset_attributes('-ps_attr_info' => $ps_attr_exp_iscsi);
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to verify: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }

=back

=head2 B<bind_portset_to_igroup(%args)>

To bind the portset to given igroup and update the expected portset attributes

=over 4

=item Inputs

=item * I<-ps_attr_info> <required>

Hash reference to portset info, basically expects same info which return from
the function 'create_portset_in_vserver'. See create_portset_in_vserver POD for more details

Hash reference looks like following:
  $ps_attr_expected->{'portset_obj'} - reference to portset object, this attribute is MUST.
  $ps_attr_expected->{'vserver'}
  $ps_attr_expected->{'portset'}
  $ps_attr_expected->{'protocol_type'}
  $ps_attr_expected->{'fc_lif'}
  $ps_attr_expected->{'tpg'}
  $ps_attr_expected->{'igroups'}
  
=item * I<-igroup_obj> <required>

Reference to cmode igroup object.

=item Outputs

Returns TRUE  if portset bind to igroup successful. 
If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    #
    # Instantiate PortsetHelper Object (see above _new() pod)
    # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
    #
    ....
    ....
    ....
    # get vserver object
    $vs_obj = ...;
    
    # get all available fcp lifs
    my $fcp_lif_names = $PSMgr_obj->get_protocol_lifs_name_in_vserver('-vserver_obj' => $vs_obj,
                                                                      'protocol_type' => 'fcp');
    
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to get fcp lifs: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
      
    #
    # create fcp portset with fcp lifs name
    #
    my $ps_attr_exp = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_fcp", 
                                            '-protocol_type' => "fcp",
                                            '-fc_lif'        => $fcp_lif_names);
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to create portset: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
   
    # get the *fcp* igroup object...
    my @igroup_objs = $PSMgr_obj->get_igroup_objs('-protocol_type' => "fcp",
                                              '-vserver_name'  => $vs_name);
    if ($PSMgr_obj->error()) {
        $err =
            "CONFIG ERROR: Test Suite Requirements:\n"
          . "Unable to get '$protocol_type' igroup(s) "
          . "on vserver: $vs_name\nReason:\n"
          . $PSMgr_obj->errmsg();
        $Test->nlog($err);
        return $TCD::FAIL;
    }
    #
    # bind the portset to igroup
    #
      
    $PSMgr_obj->bind_portset_to_igroup('-ps_attr_info' => $ps_attr_exp,
                              '-igroup_obj'   => $igroup_objs[0]);
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to bind: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    #
    # verify controller see as expected about igroup bind info's
    #
    $PSMgr_obj->verify_portset_igroup_bind('-igroup_obj'  => $igroup_objs[0],
                                       '-portset_obj' => $ps_attr_exp->{'portset_obj'});
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to verify the bind: " . $PSMgr_obj->errmsg());
        # destroy the portset using force option ...
        return $TCD::FAIL;
    }


=back

=head2 B<unbind_portset_from_igroup(%args)>

To unbind the portset from the given igroup and update the expected portset attributes

=over 4

=item Inputs

=item * I<-ps_attr_info> <required>

Hash reference to portset info, basically expects same info which return from
the function 'create_portset_in_vserver'. See create_portset_in_vserver POD for more details

Hash reference looks like following:
  $ps_attr_expected->{'portset_obj'} - reference to portset object, this attribute is MUST.
  $ps_attr_expected->{'vserver'}
  $ps_attr_expected->{'portset'}
  $ps_attr_expected->{'protocol_type'}
  $ps_attr_expected->{'fc_lif'}
  $ps_attr_expected->{'tpg'}
  $ps_attr_expected->{'igroups'}
  
=item * I<-igroup_obj> <required>

Reference to cmode igroup object.

=item Outputs

Returns TRUE  if portset unbind from igroup successful. 
If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples


    #
    # Instantiate PortsetHelper Object (see above _new() pod)
    # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
    #
    ....
    ....
    ....
    # get vserver object
    $vs_obj = ...;
    
    # get all available fcp lifs
    my $fcp_lif_names = $PSMgr_obj->get_protocol_lifs_name_in_vserver('-vserver_obj' => $vs_obj,
                                                                      'protocol_type' => 'fcp');
    
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to get fcp lifs: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
      
    #
    # create fcp portset with fcp lifs name
    #
    my $ps_attr_exp = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_fcp", 
                                            '-protocol_type' => "fcp",
                                            '-fc_lif'        => $fcp_lif_names);
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to create portset: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
   
    # get the *fcp* igroup object...
    my @igroup_objs = $PSMgr_obj->get_igroup_objs('-protocol_type' => "fcp",
                                              '-vserver_name'  => $vs_name);
    if ($PSMgr_obj->error()) {
        $err =
            "CONFIG ERROR: Test Suite Requirements:\n"
          . "Unable to get '$protocol_type' igroup(s) "
          . "on vserver: $vs_name\nReason:\n"
          . $PSMgr_obj->errmsg();
        $Test->nlog($err);
        return $TCD::FAIL;
    }
    
    #
    # bind the portset to igroup
    #
      
    $PSMgr_obj->bind_portset_to_igroup('-ps_attr_info' => $ps_attr_exp,
                              '-igroup_obj'   => $igroup_objs[0]);
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to bind: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    #
    # verify controller see as expected about igroup bind info's
    #
    $PSMgr_obj->verify_portset_igroup_bind('-igroup_obj'  => $igroup_objs[0],
                                       '-portset_obj' => $ps_attr_exp->{'portset_obj'});
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to verify the bind: " . $PSMgr_obj->errmsg());
        # destroy the portset using force option ...
        return $TCD::FAIL;
    }
    ...
    ...
    # unbind the igroup
    $PSMgr_obj->unbind_portset_from_igroup('-ps_attr_info' => $ps_attr_exp,
                          '-igroup_obj'   => $igroup_obj);
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to unbind: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
    #
    # verify controller see as expected about igroup unbind
    #
    $PSMgr_obj->verify_portset_attributes('-ps_attr_info' => $ps_attr_exp);
    if ($PSMgr_obj->error()) {
         $Test->nlog("Failed to verify: " . $PSMgr_obj->errmsg());
         return $TCD::FAIL;
    }

=back


=head2 B<verify_portset_attributes(%args)>

To verify the requested portset attributes are intact with controller return attributes

NOTE:

    # verify the portset attributes from user expected list 
    # with controller list and return one error message with  
    # list of attributes which is/are not matching. User can set 
    # _SKIP_CHECK_ value for particular key to skip checking.
    
    
    # for keys fc_lif|wwpn|tpg, the value will be comma separated list.
    # since order of that list will be different, need to check individual 
    # values present in the controller list. "_verify_portset_key_values"
    # does that.

=over 4

=item Inputs

=item * I<-ps_attr_info> <required>

Hash reference to portset info, basically expects same info which return from
the function 'create_portset_in_vserver'. See create_portset_in_vserver POD for more details

Hash reference looks like following:
  $ps_attr_expected->{'portset_obj'} - reference to portset object, this attribute is MUST.
  rest are optional or can skip the checking by setting the value of attribute to  _SKIP_CHECK_
  $ps_attr_expected->{'vserver'}
  $ps_attr_expected->{'portset'}
  $ps_attr_expected->{'protocol_type'}
  $ps_attr_expected->{'fc_lif'}
  $ps_attr_expected->{'tpg'}
  $ps_attr_expected->{'igroups'}

Caller can set value _SKIP_CHECK_ if does not want to check particular attribute.
Ex: 
   $ps_attr_expected->{'igroups'} = '_SKIP_CHECK_';

=item Outputs

Returns TRUE  if all expected portset attributes matches, if successful. 
If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    #
    # Instantiate PortsetHelper Object (see above _new() pod)
    # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
    #
    ....
    ....
    ....
    # get vserver object
    $vs_obj = ...;
    
    # get all available fcp lifs
    my $fcp_lif_names = $PSMgr_obj->get_protocol_lifs_name_in_vserver('-vserver_obj' => $vs_obj,
                                                                      'protocol_type' => 'fcp');
    
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to get fcp lifs: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }
      
    #
    # create fcp portset with fcp lifs name
    #
    my $ps_attr_exp = $PSMgr_obj->create_portset_in_vserver(
                                            '-vserver_obj'   => $vs_obj,
                                            '-portset'       => "ps_fcp", 
                                            '-protocol_type' => "fcp",
                                            '-fc_lif'        => $fcp_lif_names);
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to create portset: " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
    #
    # verify the portset attributes against controller
    #
    $PSMgr_obj->verify_portset_attributes('-ps_attr_info' => $ps_attr_exp);
    if ($PSMgr_obj->error()) {
        $Test->nlog("Failed to verify: " . $PSMgr_obj->errmsg());
        return $TCD::FAIL;
    }

=back

=head2 B<verify_portset_igroup_bind(%args)>

To verify the portset and igroup binding information are intact with controller return attributes

=over 4

=item Inputs

=item * I<-igroup_obj> <required>

Reference to cmode igroup object.


=item * I<-portset_obj> <required>

Reference to cmode portset object.

=item * I<-verify_presence> <optional> <defualt=yes>

To check the portset presence in the given igroup or not and igroup presence in the portset or not.
If verify_presence = yes it will check for the presence of portset in the given igroup
                     and igroup is bound to the given portset
If verify_presence = no it will check there is no portset bind to igroup and portset does not have
                     the igroup

=item Outputs

Returns TRUE  if all expected portset attributes matches, if successful. 
If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    see <bind_portset_to_igroup> pod Example above
   
=back

=head2 B<verify_igroup_bind(%args)>

To verify the igroup binding information has or not portset information

=over 4

=item Inputs

=item * I<-igroup_obj> <required>

Reference to cmode igroup object.

=item * I<-portset> <required>

Name of the portset to check in igroup.

=item * I<-verify_presence> <optional> <defualt=yes>

To check the portset presence in the given igroup or not.
If verify_presence = yes it will check for the presence of portset in the given igroup.
If verify_presence = no it will check there is no portset bind to igroup

=item Outputs

Returns TRUE, if successful. 
If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

    see <bind_portset_to_igroup> pod Example above
   
=back

=head2 B<get_portset_attributes(%args)>

To get the Portset attributes hash for a given a portset object for
using PortsetHelper function. Return portset expected attribute 
hash same like create_portset_in_vserver

=over 4

=item Inputs

=item * I<-portset_obj> <required>

Reference to cmode portset object.

=item Outputs

Returns a reference to hash contains portset expected attributes info if successful.
If an error occurred the error flag is set, errmsg is set to the failure reason, and {} (empty hash reference)
is passed back as the return value.

See create_portset_in_vserver POD for more deails

=item Examples

    #
    # Instantiate PortsetHelper Object (see above _new() pod)
    # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
    #
    ....
    ....
    ....
    # get vserver object
    $vs_obj = ...;
    
    $ps_obj = ...;
    
    #
    # get portset expected attributes using portset oject
    #
    my $ps_attr_exp = $PSMgr_obj->get_portset_attributes(
                                            '-portset_obj'   => $ps_obj);
    if ($PSMgr_obj->error()) {
     $Test->nlog("Failed to get portset attributes : " . $PSMgr_obj->errmsg());
     return $TCD::FAIL;
    }
   
=back

=cut

