# -*- perl -*-
#  $Id: $
#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
#
## @summary Just a bunch of helper methods for testing C-mode VVOLs functionality
## @author (sanap) dl-ssan-test@netapp.com
## @status Private
## @pod Not here yet.

package San::Task::VVolHelper;

use strict;
use Exporter;
use List::MoreUtils qw(any);
use Error;
use San::ControllerCluster;
use San::SanClient;
use San::SanLog;
use San::SanUtil;
use Data::Dumper;
use Tharn;
use NACL::C::Lun;
use NACL::CS::Lun;
use NACL::C::LunBind;
use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::Volume;
use NACL::C::LunIgroup;
use NACL::C::LunAttribute;
use NACL::STask::Aggregate;
use NACL::STask::Volume;
use NACL::STask::Lun;
use NACL::STask::LunIgroup;
use NACL::STask::NetworkInterface;
use NACL::STask::Cluster;
use NACL::CS::LunMapped;
use NACL::APISet;
use Params::Validate qw(:all);

#
# Exceptions
#
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

use vars qw( @EXPORT @ISA );

@ISA = qw( Error SanLog SanUtil );

###############################################################################
# P U B L I C   M E T H O D S
###############################################################################

###############################################################################
# Method:     new
# Objective:  Creates new VVolHelper object.
# Parameters: None.
# Returns:    A reference to the object.
###############################################################################

sub new {
    my ($class, %args) = @_;

    my $self = {};

    bless $self, $class;

    my $debug_level = 2;

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /debug_level/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                 "ERROR: Missing value for key '-debug_level'");
            }
        }
    }

    $self->set_debug_level($debug_level);

    $self->sanlog(
           -msg => "Create VVolHelper Object with debug level " . $debug_level);

    return $self;
} ## end sub new

###############################################################################
# Method:     fetch_vvol_pe_info
#
# Objective:  Create a hash to store following info
#             Vservers, VVOLs, PEs, PE-iGroup Maps, VVOL-PE binds
#
# Parameters:
#             vserver - vs name (optional)
#             ci - NACL Stask command interface obj (required)
#
# Returns:    Hash with VVOL PE info
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
###############################################################################

sub fetch_vvol_pe_info {

    my ($self, %args) = @_;
    my ($vs_name, $ci);

    $self->_clear_err();

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                      "ERROR: Missing value for key 'vserver'");
            }
            $vs_name = $value;
        } elsif ($key =~ /ci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'ci'");
            }
            $ci = $value;
        } else {
            NATE::Exceptions::Argument->throw("ERROR: Invalid key " . $key);
        }
    } ## end while (my ($key, $value) ...

    unless (defined($ci)) {
        NATE::Exceptions::Argument->throw("ERROR: Missing key 'ci'");
    }

    my @vservers;

    # if vs name given just work on it or else all vs
    $self->sanlog(msg => "Fetching VVOL-PE hash for vservers");

    if (defined($vs_name)) {
        @vservers =
          NACL::CS::Vserver->fetch(command_interface => $ci,
                                   filter            => {
                                              type    => 'data',
                                              vserver => $vs_name});
    } else {
        @vservers =
          NACL::CS::Vserver->fetch(command_interface => $ci,
                                   filter            => {type => 'data'});

    }

    my $vs_hash = {vs => []};

    # Vserver hash for a single vserver would look like this.
    #    'vs' => [
    #        {'vvol' => ['/vol/vol_1/vvol_0', '/vol/vol_1/vvol_1',
    #                    '/vol/vol_2/vvol_0', '/vol/vol_2/vvol_1'],
    #         'name' => 'vvol_vs_0',
    #         'pe'   => [
    #                  {'maps' => [
    #                       {'id'     => '2000',
    #                        'ig'     => 'auto_anyiscsi',
    #                        'vvolid' => ['0x07d0d20000010000', '0x07d0d20000020000']
    #                       }],
    #                   'path'  => '/vol/vol_1/auto_pe',
    #                   'binds' => [
    #                              {'pair' => '/vol/vol_1/auto_pe-/vol/vol_1/vvol_0',
    #                               'slun' => 'd20000010000',
    #                               'optimal' => 'true'},
    #                              {'pair' => '/vol/vol_1/auto_pe-/vol/vol_1/vvol_1',
    #                               'slun' => 'd20000020000',
    #                               'optimal' => 'true'}
    #                               ]
    #                  },
    #                  ]
    #        }
    #        ]

    foreach (@vservers) {
        my $vs = $_->vserver();
        my $vsh = {name => $vs};
        my @vols;
        my @vvols;
        my @vvols_offline;
        my @pes;
        my $vvol_hash = {};

        # Get all flexvols
        @vols =
          NACL::CS::Volume->fetch(command_interface => $ci,
                                  filter            => {vserver => $vs},
                                  allow_empty       => 'true');
        foreach (@vols) {
            my $vol_name = $_->volume();
            push(@{$vsh->{"vol"}}, $vol_name);
        }

        # Get all VVOLS
        @vvols =
          NACL::CS::Lun->fetch(command_interface => $ci,
                               filter            => {
                                          class   => 'vvol',
                                          vserver => $vs},
                               allow_empty => 'true');
        foreach (@vvols) {
            my $vvol_path = $_->path();
            push(@{$vsh->{"vvol"}}, $vvol_path);
        }

        # Get all offlined VVOLS
        @vvols_offline =
          NACL::CS::Lun->fetch(command_interface => $ci,
                               filter            => {
                                          class   => 'vvol',
                                          vserver => $vs,
                                          state   => 'offline'},
                               allow_empty => 'true');
        foreach (@vvols_offline) {
            my $vvol_path = $_->path();
            push(@{$vsh->{"vvol_offline"}}, $vvol_path);
        }

        @pes = NACL::CS::Lun->fetch(command_interface => $ci,
                                    filter            => {
                                               class   => 'protocol_endpoint',
                                               vserver => $vs},
                                    allow_empty => 'true');

        # For every PE see which igroups it is mapped to
        # for every igroup get LUN ID, create VVOL ID
        # also for every PE see what binds it has

        foreach (@pes) {
            my @binds   = ();
            my @lunmaps = ();
            my $pe_path = $_->path();
            my $pe_hash = {"path" => $pe_path};

            @binds =
              NACL::CS::LunBind->fetch(
                                      command_interface => $ci,
                                      filter            => {
                                          vserver                  => $vs,
                                          'protocol-endpoint-path' => $_->path()
                                      },
                                      allow_empty => 'true');
            foreach (@binds) {
                my $bind_pair =
                  $_->protocol_endpoint_path() . "-" . $_->vvol_path();
                my $bind_hash = {};
                $bind_hash->{"pair"}    = $bind_pair;
                $bind_hash->{"slun"}    = $_->secondary_lun();
                $bind_hash->{"optimal"} = $_->is_optimal();
                push(@{$pe_hash->{"binds"}}, $bind_hash);
            }

            # Get all lun maps
            @lunmaps =
              NACL::CS::LunMapped->fetch(command_interface => $ci,
                                         filter            => {
                                                    path    => $pe_path,
                                                    vserver => $vs},
                                         allow_empty => 'true');
            foreach (@lunmaps) {
                my $ig          = $_->igroup();
                my $id          = $_->lun_id();
                my $lunmap_hash = {};
                $lunmap_hash->{"ig"} = $ig;
                $lunmap_hash->{"id"} = $id;

                # Create VVOL ID = hex(LunID) . bind secondary lun
                foreach my $bind (@binds) {
                    my $pe_map_id = sprintf("%#06x", $id);
                    my $vvol_id = $pe_map_id . $bind->secondary_lun();
                    push(@{$lunmap_hash->{"vvolid"}}, $vvol_id);
                }
                push(@{$pe_hash->{"maps"}}, $lunmap_hash);
            } ## end foreach (@lunmaps)
            push(@{$vsh->{"pe"}}, $pe_hash);
        } ## end foreach (@pes)
        push(@{$vs_hash->{"vs"}}, $vsh);
    } ## end foreach (@vservers)

    $self->sanlog(-msg => "Dumper of PE_VVOL Info hash" . Dumper($vs_hash),
                  -msg_level => 2);
    return ($vs_hash);
} ## end sub fetch_vvol_pe_info

###############################################################################
# Method:     verify_scsi_rw
#
# Objective:  Fetch VVOL-PE info hash, run SCSI W-R on every VVOL using
#             anyiscsi from Linux host, compare Read data with written data
#             simple write of 512 byte pattern
#
# Parameters:
#             vserver - vs name (optional)
#             ci - NACL Stask command interface obj (required)
#             IO_Pattern_File - Path to pattern file (required)
#             anyiscsi - path to the tool,
#                        default /u/dsharma/p4/any_iscsi/any_iscsi/any_iscsi_linux
#             client - anyscsi installed on this client,
#                      default ssan-rhel5u6-09
#
# Returns:    SanUtil::TRUE on success to run RW on all VVOLs
#             SanUtil::FALSE if IO errors found
# Exceptions: throws NATE::Exceptions::Argument on arguments errors
#             throws NATE::BaseException on IO pattern mismatch
###############################################################################

sub verify_scsi_rw {

    my ($self, %args) = @_;
    my ($vs_name, $client, $pe_path, $vvol_path, $ci, $anyiscsi_path);
    my $vs_hash;
    my @iscsi_lifs;
    my $IO_pattern_file;
    my $read_file = "read.out";

    $self->_clear_err();

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                             "Missing value for key 'vserver'");
            }
            $vs_name = $value;
        } elsif ($key =~ /client/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                            "Incorrect value for key 'client'");
            }
            $client = $value;
        } elsif ($key =~ /pe/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                           "Incorrect value for key 'pe path'");
            }
            $pe_path = $value;
        } elsif ($key =~ /vvol/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "Incorrect value for key 'vvol path'");
            }
            $vvol_path = $value;
        } elsif ($key =~ /ci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                                "Incorrect value for key 'ci'");
            }
            $ci = $value;
        } elsif ($key =~ /IO_Pattern_File/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                   "Incorrect value for key 'IO_Pattern_File'");
            }
            $IO_pattern_file = $value;
        } elsif ($key =~ /anyisci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                          "Incorrect value for key 'anyiscsi'");
            }
            $anyiscsi_path = $value;
        } else {
            NATE::Exceptions::Argument->throw("Invalid key " . $key);
        }
    } ## end while (my ($key, $value) ...

    unless (defined($ci)) {
        NATE::Exceptions::Argument->throw("Missing key 'ci'");
    }

    unless (defined($IO_pattern_file)) {
        NATE::Exceptions::Argument->throw("Missing key 'IO_Pattern_File'");
    }

    unless (defined($client)) {
        $client = "ssan-rhel5u6-09";
        $self->sanlog(msg => "Using anyiscsi client " . $client);
    }

    unless (defined($anyiscsi_path)) {
        $anyiscsi_path = "/u/dsharma/p4/any_iscsi/any_iscsi/any_iscsi_linux";
        $self->sanlog(msg => "Using anyiscsi path " . $anyiscsi_path);
    }

    # Fetch VVOL PE info
    if (defined($vs_name)) {
        $vs_hash =
          $self->fetch_vvol_pe_info(vserver => $vs_name,
                                    ci      => $ci);

    } else {
        $vs_hash = $self->fetch_vvol_pe_info(ci => $ci);

    }

    my @vservers = @{$vs_hash->{"vs"}};
    my $ssh      = "/x/eng/kc-users/samt/bin/essh " . $client . " root netapp";

    # Travel VVOL PE info hash to get VVOL ID and then run SCSI W followed by R
    foreach my $vs (@vservers) {

        $vs_name = $vs->{"name"};

        # Make sure at least 1 PE is present in the vserver
        if (not defined(@{$vs->{"pe"}})) {
            $self->sanlog(  msg => "We did not find any PE LUN for vserver " 
                          . $vs_name
                          . " skipping IO");
            next;
        }

        my @pes = @{$vs->{"pe"}};

        # Need to capture iSCSI Data LIF, select one of them to run IO, TODO select all
        my @data_protocol = qw(iscsi);
        @iscsi_lifs =
          NACL::CS::NetworkInterface->fetch(
                                         command_interface => $ci,
                                         filter            => {
                                             role            => 'data',
                                             'data-protocol' => \@data_protocol,
                                             vserver         => $vs_name});

        foreach my $pe (@pes) {
            if (defined($pe->{"maps"})) {
                my @maps = @{$pe->{"maps"}};

                foreach my $map (@maps) {
                    if (defined($map->{"vvolid"})) {
                        my @vvol_ids = @{$map->{"vvolid"}};

                        # Select first iSCSI LIF to run anyiscsi from
                        my $iscsi_ip = $iscsi_lifs[0]->address();

                        foreach my $vvol_id (@vvol_ids) {
                            my $anyiscsi_write =
                                $anyiscsi_path . " "
                              . $iscsi_ip . " "
                              . $vvol_id
                              . " 512 -d "
                              . $IO_pattern_file
                              . " -v 0x2a 0 0 0 0 0 63 0 1 0";
                            my $anyiscsi_read =
                                $anyiscsi_path . " "
                              . $iscsi_ip . " "
                              . $vvol_id
                              . " 512 -d "
                              . $read_file
                              . " -v 0x28 0 0 0 0 0 63 0 1 0";

                            $self->sanlog(msg => "executing anyiscsi write cmd "
                                          . $anyiscsi_write);
                            my $cmd_out = `$ssh $anyiscsi_write`;
                            if ($cmd_out =~ /Errorful status returned:/) {
                                $self->sanlog(msg =>
                                    "IO error during Write: Writing data to lun failed!!!!!!"
                                );
                                return $SanUtil::FALSE;
                            }
                            $self->sanlog(msg => $cmd_out, -msg_level => 2);

                            $self->sanlog(msg => "executing anyiscsi read cmd "
                                          . $anyiscsi_read);
                            $cmd_out = `$ssh $anyiscsi_read`;
                            if ($cmd_out =~ /Errorful status returned:/) {
                                $self->sanlog(msg =>
                                    "IO error during Read: Reading data from lun failed!!!!!!"
                                );
                                return $SanUtil::FALSE;
                            }
                            $self->sanlog(msg => $cmd_out, -msg_level => 2);


                            $self->sanlog(msg =>
                                "comparing pattern file with read.out file created by anyiscsi read"
                            );
                            $cmd_out = `$ssh diff $read_file $IO_pattern_file`;
                            if ($cmd_out == "") {
                                $self->sanlog(msg => "Pattern Matched !!! for "
                                              . $vs->{name} . " - "
                                              . $vvol_id);
                                $cmd_out = `$ssh rm -rf $read_file`;
                            } else {
                                NATE::BaseException->throw(
                                                 "Pattern is not matching !!!");
                            }
                        } ## end foreach my $vvol_id (@vvol_ids)
                    } ## end if (defined($map->{"vvolid"...
                } ## end foreach my $map (@maps)
            } ## end if (defined($pe->{"maps"...
        } ## end foreach my $pe (@pes)
    } ## end foreach my $vs (@vservers)

    return $SanUtil::TRUE;
} ## end sub verify_scsi_rw

###############################################################################
# Method:     create_pe_bind_vvol
#
# Objective:  create one PE per aggr per vserver and one anysicsi igroup per vs
#             then depenind on defined map-bind or bind-map order
#             bind the vvols to pe and map pe to anysicsi igroup
#
# Parameters:
#             vserver - vs name (optional)
#             Ci - NACL Stask command interface obj (required)
#             bind_order - "MAP-BIND" or "BIND-MAP" (first one is default)
#
# Returns:    SanUtil::TRUE on success
# Exceptions: throws NATE::Exceptions::Argument on arguments errors
###############################################################################

sub create_pe_bind_vvol {

    my ($self, %args) = @_;
    my ($vs_name, $ci, $bind_order);

    $self->_clear_err();

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                             "Missing value for key 'vserver'");
            }
            $vs_name = $value;
        } elsif ($key =~ /bind_order/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                        "Incorrect value for key 'bind_order'");
            }
            $bind_order = $value;
        } elsif ($key =~ /ci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                                "Incorrect value for key 'ci'");
            }
            $ci = $value;
        } else {
            NATE::Exceptions::Argument->throw("Invalid key " . $key);
        }
    } ## end while (my ($key, $value) ...

    unless (defined($ci)) {
        NATE::Exceptions::Argument->throw("Missing key 'ci'");
    }

    unless (defined($bind_order)) {
        $bind_order = "MAP-BIND";
    }

    my @vservers;

    if (defined($vs_name)) {
        $self->sanlog(msg => "Fetching VVOL-PE hash for vserver " . $vs_name);
        @vservers =
          NACL::CS::Vserver->fetch(command_interface => $ci,
                                   filter            => {
                                              type    => 'data',
                                              vserver => $vs_name});
    } else {
        $self->sanlog(msg => "Fetching VVOL-PE hash for all vservers");
        @vservers =
          NACL::CS::Vserver->fetch(command_interface => $ci,
                                   filter            => {type => 'data'});

    }

    # Lun map id for PE has to be greater than 256, assuming 2000 is a good start
    # As regular LUNs start to get mapped from LUN ID 0, having more than 2000 LUNs
    # mapped to one igroup is rare case

    my $lun_id = "2000";

    foreach (@vservers) {
        my $vs = $_->vserver();
        my @aggrs;
        my @vols;
        my $anyiscsi_ig;

        $self->sanlog(msg => "Find anyiscsi igroup if any for vserver " . $vs);

        $anyiscsi_ig =
          NACL::C::LunIgroup->find(
                           command_interface => $ci,
                           filter            => {
                               vserver   => $vs,
                               initiator => ['iqn.2007-12.com.netapp.any-iscsi']
                           });

        # Get all aggrs for this vs
        @aggrs =
          NACL::CS::Aggregate->fetch(command_interface => $ci,
                                     allow_empty       => 'true');

        # create one pe per aggr, choose any volume from that aggr
        foreach (@aggrs) {
            my @vvols;
            my $aggr = $_->aggregate();

            # Find all volumes for this aggr and this vserver
            @vols =
              NACL::CS::Volume->fetch(command_interface => $ci,
                                      filter            => {
                                                 vserver   => $vs,
                                                 aggregate => $aggr},
                                      allow_empty => 'true');

            # Find all vvols for all the volumes from this aggr and vs
            foreach (@vols) {
                my @vvols_per_vol =
                  NACL::CS::Lun->fetch(command_interface => $ci,
                                       filter            => {
                                                  class   => 'vvol',
                                                  vserver => $vs,
                                                  volume  => $_->volume()},
                                       allow_empty => 'true');
                push(@vvols, @vvols_per_vol);
            } ## end foreach (@vols)

            if (scalar(@vvols) == 0) {

                # There are no VVOLs in this aggr, move to next aggr
                $self->sanlog(msg => "No vvols found in aggr " . $aggr);
                next;

            } else {

                # There are VVOLs in this aggr, create PE and bind all VVOLS
                my $pe_path = "/vol/" . $vols[0]->volume() . "/auto_pe";

                $self->sanlog(
                    msg => "Creating PE for aggr " . $aggr . " at " . $pe_path);

                # Using Stask function, so that if PE already exists it will reuse it
                my $pe =
                  NACL::STask::Lun->create(command_interface => $ci,
                                           path              => $pe_path,
                                           vserver           => $vs,
                                           size              => '35m',
                                           class     => 'protocol-endpoint',
                                           ostype    => 'linux',
                                           if_exists => 'reuse');

                if ($bind_order eq "MAP-BIND") {

                    # Map the PE first and then Bind VVOLS to PE
                    $self->sanlog(
                                 msg => "Map PE to anyiscsi igroup with LUN ID "
                                   . $lun_id);

                    # Using Stask function, so that if PE already mapped then it will reuse it
                    my $lunmap =
                      NACL::STask::Lun->map(command_interface => $ci,
                                            path              => $pe_path,
                                            vserver           => $vs,
                                            igroup   => $anyiscsi_ig->igroup(),
                                            'lun-id' => $lun_id,
                                            nacltask_if_mapped_igroup => 'reuse'
                                           );

                    # Increament Lun id for next map
                    $lun_id++;

                    foreach (@vvols) {

                        my $vvol_path = $_->path();
                        $self->sanlog(  msg => "Bind VVOL "
                                      . $vvol_path
                                      . " to PE "
                                      . $pe_path);

                        $self->sanlog(msg =>
                             "Creating the bind because it doesnt already exist"
                        );
                        my $bind =
                          NACL::C::LunBind->create(
                                           command_interface        => $ci,
                                           'protocol-endpoint-path' => $pe_path,
                                           'vvol-path' => $vvol_path,
                                           vserver     => $vs,);

                    } ## end foreach (@vvols)
                } else {

                    # Bind VVOLS to PE first and them MAP PE
                    foreach (@vvols) {
                        my $vvol_path = $_->path();
                        $self->sanlog(  msg => "Bind VVOL "
                                      . $vvol_path
                                      . " to PE "
                                      . $pe_path);

                        $self->sanlog(msg =>
                             "Creating the bind because it doesnt already exist"
                        );
                        my $bind =
                          NACL::C::LunBind->create(
                                           command_interface        => $ci,
                                           'protocol-endpoint-path' => $pe_path,
                                           'vvol-path' => $vvol_path,
                                           vserver     => $vs,);
                    } ## end foreach (@vvols)

                    $self->sanlog(
                                 msg => "Map PE to anyiscsi igroup with LUN ID "
                                   . $lun_id);

                    # Using Stask function, so that if PE already mapped then it will reuse it
                    my $lunmap =
                      NACL::STask::Lun->map(command_interface => $ci,
                                            path              => $pe_path,
                                            vserver           => $vs,
                                            igroup   => $anyiscsi_ig->igroup(),
                                            'lun-id' => $lun_id,
                                            nacltask_if_mapped_igroup => 'reuse'
                                           );

                    # Increament Lun id for next map
                    $lun_id++;
                } ## end else [ if ($bind_order eq "MAP-BIND")
            } ## end else [ if (scalar(@vvols) == ...
        } ## end foreach (@aggrs)
    } ## end foreach (@vservers)
    return $SanUtil::TRUE;
} ## end sub create_pe_bind_vvol

###############################################################################
# Method:     cleanup_pe_bind_vvol
#
# Objective:  cleanup PE map and VVOL bind based on order given
#
# Parameters:
#             vserver - vs name (optional)
#             ci - NACL Stask command interface obj (required)
#             bind_order - "UNMAP-UNBIND" or "UNBIND-UNMAP" (first one is default)
#             unbind_with_force - use force while unbinding, default 1
#             delete_pe - to delete PE lun or not, default delete it (1)
#
# Returns:   SanUtil::TRUE on success
# Exceptions: throws NATE::Exceptions::Argument on arguments errors
###############################################################################

sub cleanup_pe_bind_vvol {

    my ($self, %args) = @_;
    my ($vs_name, $ci, $unbind_order, $vs_hash, $unbind_with_force, $delete_pe);

    $self->_clear_err();

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                             "Missing value for key 'vserver'");
            }
            $vs_name = $value;
        } elsif ($key =~ /unbind_order/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                        "Incorrect value for key 'bind_order'");
            }
            $unbind_order = $value;
        } elsif ($key =~ /ci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                                "Incorrect value for key 'ci'");
            }
            $ci = $value;
        } elsif ($key =~ /unbind_force/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                      "Incorrect value for key 'unbind_force'");
            }
            $unbind_with_force = $value;
        } elsif ($key =~ /delete_pe/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "Incorrect value for key 'delete_pe'");
            }
            $delete_pe = $value;
        } else {
            NATE::Exceptions::Argument->throw("Invalid key " . $key);
        }
    } ## end while (my ($key, $value) ...

    unless (defined($ci)) {
        NATE::Exceptions::Argument->throw("Missing key 'ci'");
    }

    unless (defined($unbind_order)) {
        $unbind_order = "UNMAP-UNBIND";
    }

    unless (defined($unbind_with_force)) {
        $unbind_with_force = 1;
    }

    unless (defined($delete_pe)) {
        $delete_pe = 1;
    }

    if (defined($vs_name)) {
        $vs_hash =
          $self->fetch_vvol_pe_info(vserver => $vs_name,
                                    ci      => $ci);

    } else {
        $vs_hash = $self->fetch_vvol_pe_info(ci => $ci);
    }

    my @vservers = @{$vs_hash->{"vs"}};

    foreach my $vs (@vservers) {

        $vs_name = $vs->{"name"};

        # if no PE then no VVOL binds and no PE mapping to be cleaned up
        if (not defined(@{$vs->{"pe"}})) {
            $self->sanlog(  msg => "No PE LUN for vserver " 
                          . $vs_name
                          . " skipping PE Bind cleanup");
            next;
        }

        my @pes = @{$vs->{"pe"}};

        foreach my $pe (@pes) {
            my $path = $pe->{"path"};

            if ($unbind_order eq "UNMAP-UNBIND") {

                if (defined($pe->{"maps"})) {
                    my @maps = @{$pe->{"maps"}};

                    foreach my $map (@maps) {
                        my $ig = $map->{"ig"};

                        $self->sanlog(msg => "Unmap PE from igroup " . $ig);

                        # Using Stask function, so that if PE already unmapped then it will just gracefully exit
                        my $lunmap =
                          NACL::STask::Lun->unmap(
                                           command_interface => $ci,
                                           path              => $path,
                                           vserver           => $vs_name,
                                           igroup            => $ig,
                                           nacltask_if_unmapped_igroup => 'pass'
                          );

                    } ## end foreach my $map (@maps)
                } ## end if (defined($pe->{"maps"...

                if (defined($pe->{"binds"})) {
                    my @binds = @{$pe->{"binds"}};

                    foreach my $bind (@binds) {

                        my @pair      = split(/-/, $bind->{"pair"});
                        my $pe_path   = $pair[0];
                        my $vvol_path = $pair[1];
                        if ($unbind_with_force) {
                            $self->sanlog(  msg => "Unbind (with force) VVOL "
                                          . $vvol_path
                                          . " to PE "
                                          . $pe_path);

                            NACL::C::LunBind->destroy(
                                           command_interface        => $ci,
                                           'protocol-endpoint-path' => $pe_path,
                                           'vvol-path' => $vvol_path,
                                           vserver     => $vs_name,
                                           force       => 'true');
                        } else {
                            $self->sanlog(  msg => "Unbind VVOL "
                                          . $vvol_path
                                          . " to PE "
                                          . $pe_path);

                            NACL::C::LunBind->destroy(
                                           command_interface        => $ci,
                                           'protocol-endpoint-path' => $pe_path,
                                           'vvol-path' => $vvol_path,
                                           vserver     => $vs_name);

                        } ## end else [ if ($unbind_with_force)

                    } ## end foreach my $bind (@binds)
                } ## end if (defined($pe->{"binds"...

            } elsif ($unbind_order eq "UNBIND-UNMAP") {

                if (defined($pe->{"binds"})) {
                    my @binds = @{$pe->{"binds"}};

                    foreach my $bind (@binds) {

                        my @pair      = split(/-/, $bind->{"pair"});
                        my $pe_path   = $pair[0];
                        my $vvol_path = $pair[1];

                        $self->sanlog(  msg => "Unbind VVOL "
                                      . $vvol_path
                                      . " to PE "
                                      . $pe_path);

                        NACL::C::LunBind->destroy(
                                           command_interface        => $ci,
                                           'protocol-endpoint-path' => $pe_path,
                                           'vvol-path' => $vvol_path,
                                           vserver     => $vs_name,);

                    } ## end foreach my $bind (@binds)
                } ## end if (defined($pe->{"binds"...
                if (defined($pe->{"maps"})) {
                    my @maps = @{$pe->{"maps"}};

                    foreach my $map (@maps) {
                        my $ig = $map->{"ig"};
                        $self->sanlog(msg => "Unmap PE from igroup " . $ig);

                        # Using Stask function, so that if PE already unmapped then it will just gracefully exit
                        my $lunmap =
                          NACL::STask::Lun->unmap(
                                           command_interface => $ci,
                                           path              => $path,
                                           vserver           => $vs_name,
                                           igroup            => $ig,
                                           nacltask_if_unmapped_igroup => 'pass'
                          );

                    } ## end foreach my $map (@maps)
                } ## end if (defined($pe->{"maps"...
            } ## end elsif ($unbind_order eq "UNBIND-UNMAP")

            if ($delete_pe) {
                $self->sanlog(msg => "Delete PE");
                NACL::C::Lun->delete(command_interface => $ci,
                                     path              => $path,
                                     vserver           => $vs_name,
                                     force             => 'true');
            }

        } ## end foreach my $pe (@pes)

        # Delete igroup
    } ## end foreach my $vs (@vservers)
    return $SanUtil::TRUE;
} ## end sub cleanup_pe_bind_vvol

sub generate_random_string {
    my $length_of_randomstring = shift;
    my @chars = ('a' .. 'z', 'A' .. 'Z', '0' .. '9', '_');
    my $random_string;

    foreach (1 .. $length_of_randomstring) {

        # rand @chars will generate a random
        # number between 0 and scalar @chars
        $random_string .= $chars[rand @chars];
    }
    return $random_string;
} ## end sub generate_random_string

###############################################################################
# Method:     set_attr
#
# Objective:  set user defined attribute (key, value pair) to VVOLs
#             verify the set attr value if verify flag is set
#
# Parameters:
#             vserver - vs name (optional)
#             verify - read the set attr and verify
#             ci - NACL Stask command interface obj (required)
#
# Returns:    None
# Exceptions: throws NATE::Exceptions::Argument on Argument errors
#             throws NATE::Base::Exception on attr verification failure
###############################################################################

sub set_attr {

    my ($self, %args) = @_;
    my ($vs_name, $ci, $verify, $max_length);

    $self->_clear_err();

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                      "ERROR: Missing value for key 'vserver'");
            }
            $vs_name = $value;
        } elsif ($key =~ /ci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'ci'");
            }
            $ci = $value;
        } elsif ($key =~ /verify/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                     "ERROR: Incorrect value for key 'verify'");
            }
            $verify = $value;
        } elsif ($key =~ /max_length/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                "ERROR: Incorrect value for key 'attr_length'");
            }
            $max_length = $value;
        } else {
            NATE::Exceptions::Argument->throw("ERROR: Invalid key " . $key);
        }
    } ## end while (my ($key, $value) ...

    unless (defined($ci)) {
        NATE::Exceptions::Argument->throw("ERROR: Missing key 'ci'");
    }

    unless (defined($verify)) {
        $verify = 0;
    }

    unless (defined($max_length)) {

        # Setting default for SN.0, user can pass this if it changes in future
        $max_length = 4092;
    }

    my $attr_name = 'VMWareAttr';

    # Total limit on size of name and value pair is 4092 chars in SN.0
    # Here name of attribute is 10 char long, so adjusting max range 4092 - 10
    my @attr_length_range = ('512', '2048', $max_length - 10, $max_length - 9);

    my @vservers;

    # if vs name given just work on it or else all vs
    $self->sanlog(msg => "Fetching VVOL-PE hash for vservers");

    if (defined($vs_name)) {
        @vservers =
          NACL::CS::Vserver->fetch(command_interface => $ci,
                                   filter            => {
                                              type    => 'data',
                                              vserver => $vs_name});
    } else {
        @vservers =
          NACL::CS::Vserver->fetch(command_interface => $ci,
                                   filter            => {type => 'data'});

    }

    foreach (@vservers) {
        my $vs = $_->vserver();
        my @vvols;

        # Get all VVOLS
        @vvols =
          NACL::CS::Lun->fetch(command_interface => $ci,
                               filter            => {
                                          class   => 'vvol',
                                          vserver => $vs},
                               allow_empty => 'true');

        # index to iterate over range array
        my $i              = 0;
        my $command_passed = 1;

        foreach (@vvols) {

            # Lets create attribute value using random pattern
            my $attr_length = $attr_length_range[$i++];
            if ($i eq @attr_length_range) {

                # i has incremented past length of range array, reset i to 0
                $i = 0;
            }

            my $attr_val_set = generate_random_string($attr_length);
            $self->sanlog(msg => "Atrribute length = " . Dumper($attr_length));
            $self->sanlog(msg => "Atrribute value = " . Dumper($attr_val_set));

            my $vvol_path = $_->path();
            try {
                NACL::C::LunAttribute->set(command_interface => $ci,
                                           vserver           => $vs,
                                           path              => $vvol_path,
                                           name              => $attr_name,
                                           value             => $attr_val_set);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                if ($attr_length <= 4082) {

                    # attr length is within set limits, the command should have not failed
                    my $e = shift;
                    NATE::BaseException->throw("Attr set failed !!!" . $e);
                } else {

                    # command should fail as att name + value is > max length
                    # so skip the verification
                    $command_passed = 0;
                    $self->sanlog(
                            msg => "Attr set failed due to exceeding max length"
                              . " skipping verification");
                }
            };

            if ($verify && $command_passed) {

                # Get the attr value and compare
                my $attr_show =
                  NACL::CS::LunAttribute->fetch(command_interface => $ci,
                                                filter            => {
                                                           vserver => $vs,
                                                           path => $vvol_path,
                                                           name => $attr_name});
                my $attr_val_get = $attr_show->value();

                if ($attr_val_get ne $attr_val_set) {
                    NATE::BaseException->throw(
                                         "Attribute value is not matching !!!");
                } else {
                    $self->sanlog(  msg => "Attribute value set for vvol "
                                  . $vvol_path
                                  . " matched");
                }
            } ## end if ($verify && $command_passed)
            $command_passed = 1;
        } ## end foreach (@vvols)
    } ## end foreach (@vservers)
} ## end sub set_attr
################################################################################
#  get_data_aggrs
# Objective: - get_data_aggrs
# Parameters:
#     ci - NACL Stask command interface obj (required)
# Returns:  - Array of data aggrs on a cluster.
#           - Empty array if none found
# Exceptions: throws NATE::Exceptions::Argument on arguments errors
################################################################################

sub get_data_aggrs {
    my ($self, %args) = @_;
    my @data_aggr;    # array to return
    my $ci;

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /ci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'ci'");
            }
            $ci = $value;
        }
    }
    
    my @Aggrs = NACL::C::StorageAggregate->find(command_interface => $ci);

    # Get all aggrs

    foreach my $aggr (@Aggrs) {
        if ($aggr->aggregate() =~ /data_aggr/) {
            push(@data_aggr, $aggr);
        }
    }

    # Make sure there are some data aggrs returned
    unless (scalar(@data_aggr)) {
        $self->sanlog(msg => "No data aggrs found. ");
        return (wantarray()) ? () : [];
    }

    my $i = 0;

    # Print out the data aggr array
    for ($i = 0 ; $i < (scalar(@data_aggr)) ; $i++) {
        $self->sanlog(  "Name of aggregate at " 
                    . $i 
                    . " is : "
                    . $data_aggr[$i]->aggregate());
    }

    $self->sanlog(msg => "We found data aggrs " . scalar(@data_aggr));
    return (wantarray()) ? @data_aggr : \@data_aggr;

} ## end sub get_data_aggr

################################################################################
#  get_volume_with_pe
# Objective: - get_volume_with_pe
# Parameters:
#     ci - NACL command interface obj (required)
#     aggr - name of aggregate from which to find volume (required)
#     vserver - name of vserver from which to find volume (optional)
# Returns:  - NACL::C::Volume object that has PE in it
#           - NACL::C::Lun object of class protocol-endpoint
# Exceptions: throws NATE::Exceptions::Argument on arguments errors
################################################################################

sub get_volume_with_pe {
    my ($self, %args) = @_;
    my $aggr;
    my $ci;
    my $vserver;
    my ($source_vol, $pes);    # return variables
    my @Vol;

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /ci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'ci'");
            }
            $ci = $value;
        } elsif ($key =~ /aggr/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                       "ERROR: Incorrect value for key 'aggr'");
            }
            $aggr = $value;

        } elsif ($key =~ /vserver/i) {
            $vserver = $value;

        }
    } ## end while (my ($key, $value) ...

    #Step :- Source vol-> Get vol that has PE in it. for source aggr. Dest vol is just a name.
    $self->sanlog(msg => "Find Vols associated with the source aggr " . $aggr);

    if (length($vserver)) {
        @Vol =
          NACL::C::Volume->find(
                            command_interface => $ci,
                            filter => {aggregate => $aggr, vserver => $vserver},
                            allow_empty => 'true');
    } else {
        @Vol =
          NACL::C::Volume->find(command_interface => $ci,
                                filter            => {aggregate => $aggr},
                                allow_empty       => 'true');

    }
    unless (scalar(@Vol)) {
        $self->sanlog(msg => "No Vols associated with the source aggr found ");
        return ($source_vol, $pes);
    }

    $self->sanlog(msg => "Find the Vol that has the PE in it");
    foreach my $find_pe_vol (@Vol) {

          # Search for PEs in each vol
            $pes = NACL::C::Lun->find(command_interface => $ci,
                                      filter            => {
                                                class  => "protocol-endpoint",
                                                volume => $find_pe_vol->volume(), 
                                                vserver =>$find_pe_vol->vserver(),
                                      },
                                      allow_empty => 'true');

        if (length($pes)) {
            $self->sanlog(msg => "We got our source vol and found our PE");
            $source_vol = $find_pe_vol;
            last;
        }

    } ## end foreach my $find_pe_vol (@Vol)

    return ($source_vol, $pes);
} ## end sub get_volume_with_pe

################################################################################
#  get_binds_for_pe
# Objective: - Get NACL::CS::LunBind objects for a PE and verify the is_optimal value on all
#            - Create a true array for vvols in same volume
#            - Create a false array for vvols in different volume
# Parameters:
#     ci - NACL Stask command interface obj (required)
#     vserver - name of vserver (required)
#     pe_path - protocol-endpoint-path to which vvols are bound (required)
# Returns:  - Array of NACL::CS::LunBind objects with is_optimal true
#           - Array of NACL::CS::LunBind objects with is_optimal false
# Exceptions: throws NATE::Exceptions::Argument on arguments errors
################################################################################

sub get_binds_for_pe {
    my ($self, %args) = @_;
    my @lun_bind_objs;      # array of all binds for PE
    my @true_bind_objs;     # array of all vvols in same vol as PE
    my @false_bind_objs;    # array of all vvols in same vol as PE
    my $vserver;            # vserver in which the PE belongs
    my $pes_path;
    my $ci;                 #

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /ci/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                         "ERROR: Incorrect value for key 'ci'");
            }
            $ci = $value;
        }
        if ($key =~ /vserver/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                    "ERROR: Incorrect value for key 'vserver'");
            }
            $vserver = $value;
        }
        if ($key =~ /pe_path/i) {
            unless (length($value)) {
                NATE::Exceptions::Argument->throw(
                                    "ERROR: Incorrect value for key 'pe_path'");
            }
            $pes_path = $value;
        }

    } ## end while (my ($key, $value) ...

    # get all lun bind objs associated with PE in source vol and verify is_optimal
    @lun_bind_objs =
      NACL::C::LunBind->find(command_interface => $ci,
                             filter            => {
                                        'vserver'                => $vserver,
                                        'protocol-endpoint-path' => $pes_path,},
                             allow_empty => 'true');

    unless (scalar(@lun_bind_objs)) {
        $self->sanlog(msg => "No binds found for PE " . $pes_path);
        return (((wantarray()) ? () : []), ((wantarray()) ? () : []));
    }

    # Go over all the bind objects
    # If vvol is in the same vol as PE it goes in truearray
    # If vvol is in different vol as PE it goes in falsearray
    foreach my $bind (@lun_bind_objs) {
        my $vvol_path_bindobj = $bind->state()->vvol_path();
        $vvol_path_bindobj =~ m/(\/.*?\/.*?\/).*/;
        my $vvol_vol_path = $1;
        $pes_path =~ m/(\/.*?\/.*?\/).*/;
        my $pe_vol_path = $1;
        if ($pe_vol_path ne $vvol_vol_path) {
            $self->sanlog(  msg => "PE and VVol are in different volumes"
                          . $pe_vol_path . " and "
                          . $vvol_vol_path);
            push(@false_bind_objs, $bind);
        } else {
            $self->sanlog(  msg => "PE and VVol are in same volumes"
                          . $pe_vol_path . " and "
                          . $vvol_vol_path);
            push(@true_bind_objs, $bind);

        }
    } ## end foreach my $bind (@lun_bind_objs)

    $self->sanlog(  msg => "Number of vvols in same vol as PE : "
                  . scalar(@true_bind_objs)
                  . "Number of vvols in diff vol as PE : "
                  . scalar(@false_bind_objs));
    return (\@true_bind_objs, \@false_bind_objs);

} ## end sub get_binds_for_pe

1;

