#
# $Id: //depot/prod/test/Rsierranevada/lib/San/VdiskSanity/Checks.pm#1 $
#
# Copyright (c) 2008 NetApp, Inc.
# All rights reserved.
#
# san/VdiskSanity/Checks.pm
##
## @summary Library that contains routines to perform UUID and VTOC inconsistencies
## @description
##   This library provides necessary routines that would perform the UUID and VTOC
##      consistency checks based on the operation performed.
##
## @description Objective:
## To provide library functions that would perform UUID and VTOC inconsistencies check.
## @author Yeshanth Jayakumar(yeshanth@netapp.com)
## @status public
## @change 2009-06-09 yeshanth: Original template created

# Package Name
##############
package San::VdiskSanity::Checks;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::VdiskSanity::lun_attr_parse;
use San::SanLog;
use Error;

use vars qw( @ISA );

@ISA=qw(Error SanLog);

# Package Export contents
####################
@ISA = Exporter;
require Exporter;
@EXPORT = qw(check_for_UUID_VTOC_consistencies);

# Set this is to "1" if you want debug mode turned on
my $debug = 0;

=head2 B<check_for_UUID_VTOC_consistencies>

Check for UUID and VTOC consistencies in a regular/mirrored LUN, 
provided the LUN show -i output is supplied.

=over 4
    
=item Inputs

=item * lun_show_i_output

LUN show -i output

=item * lun_state

regularLUN/mirroredLUN  

=item * operation

"SFSR-in-place", "SFSR-out-of-place", "VBSR", NDVSR"...

=item Outputs
    
Returns $ClientUtil::TRUE and (vol_obj) on success
Returns $ClientUtil::FALSE on error message on failure

=item Examples

     ($status, $return_value) = check_for_UUID_VTOC_consistencies(  $lun_show_i_output,
                                                                    "regularLUN",
                                                                    "SFSR-in-place"
                                                                );

    if($status == $ClientUtil::FALSE) {
        return ($status, $return_value);
    }
    
=back

=cut

##################################################
## Function     : check_for_UUID_VTOC_consistencies
##
## Description  :   This function performs the following consistency checks
##                      1) Serial number consistency in the Attribute stream and
##                          UUID db.
##                      2) UUID value consistency in in-core(vdisk_t), UUID db and
##                          the VTOC.
##      Rules:
##              a) If LUN is regular LUN, then (1) and (2) should be satisfied
##              a) If LUN is mirrored LUN, then (1) and (2) should be satisfied
##
## Arguments        :
##      $lun_show_i_output  :   The LUN show -i output
##      $lun_state          :   The state of LUN(regularLUN/mirroredLUN)
##      $operation          :   What operation went through(VSM,QSM,VBSR,NDVSR,SFSR...)
##
## Notes            :
##################################################
sub check_for_UUID_VTOC_consistencies
{
     my (
         $lun_show_i_output,
         $lun_state,
         $operation
        )
        = @_;

    my $value1                      = "";
    my $value2                      = "";
    my $value3                      = "";
    my $error                       = 0;
    my @array                       = ();

    dbg_print("Checks.pm: LUN show output: $lun_show_i_output :: LUN state:$lun_state :: Operation:$operation");

    @array = split("\n", $lun_show_i_output);

    ############
    # CHECK 1 : Serial # consistency in Attribute stream and UUID db
    ############

    $value1 = lun_showInfo_parse_2_key(\@array, "Attribute stream", "serial");
    $value2 = lun_showInfo_parse_2_key(\@array, "UUID", "serial");

    if($value1 != $value2) {
            return ($ClientUtil::FALSE, "VTOC - UUID inconsistencies");
    }

    ############
    # CHECK 2 : UUID consistency in Vdisk details(vdisk_t), UUID(vdisk_uuid_entry_t) and VTOC(vdisk_vtoc_entry_info_t)
    ############

    $value1 = lun_showInfo_parse_2_key(\@array, "Vdisk details", "UUID");
    $value2 = lun_showInfo_parse_2_key(\@array, "UUID", "UUID");
    $value3 = lun_showInfo_parse_2_key(\@array, "VTOC Info", "UUID");

    #
    # For the LUN in the mirrored destination location the UUID value is updated only in
    #               a) Vdisk incore(vdisk_t)
    #               b) UUID (vdisk_uuid_entry_t) - (which lives in the root volume of the destination controller"
    #
    # And the UUID value remains DIFFERENT in VTOC info, because the destination LUN is read-only.
    if($lun_state == "mirroredLUN") {
        if($value1 != $value2 && $value2 == $value3) {
            return ($ClientUtil::FALSE, "VTOC - UUID inconsistencies");
        }
    } else {
        if($value1 != $value2 || $value1 != $value3) {
            return ($ClientUtil::FALSE, "VTOC - UUID inconsistencies");
        }
    }

    return ($ClientUtil::TRUE, "match");
}

#######################################################################
#
# Function: dbg_print
#
# Description: This internal function will print the passed line
#              of output if the DEBUG flag is set.
#
# Input:  String to print.
#
# Return: None
#
# History:
# 03/05/2009  Glenn M. Gobeli
#
#######################################################################
sub dbg_print {
    if ($debug) {
        print "\n Checks.pm: DBG: ", $_[0], "\n";
    }
}

# Need to include 1; because Perl complains if the
# package doesn't have return value
1;
