#
# $Id: //depot/prod/test/Rsierranevada/lib/San/VdiskSanity/lun_attr_parse.pm#1 $
#
# Copyright (c) 2008 NetApp, Inc.
# All rights reserved.
#
# san/VdiskSanity/lun_attr_parse.pm
##
## @summary Utility for testing SNMP functionality
## @description
##   This utility provides a variety of functions to test the SNMP functionality
##
## @description Objective:
# This file contains LUN attribute parser routines. 
# The intent is to provide a means to parse LUN attribute information
# which may have been collected through various LUN interfaces. This 
# may be based upon a single <key>/<value> paring, such as the output
# from a "lun attribute list" command, or it may be based upon a dual 
# key and value pairing (<key1>, <key2/<value>), such as the output 
# from the "lun show -i" command.
#
# Notes: Without an export list, the user would access by:
# attribute_parse::<sub routine name>(); 
# However, we can export the "global" functions we want the user to 
# see so that they can call the function without scoping. This is by
# <sub routine name>();
## @author Glenn Gobeli(glenng@netapp.com) and Yeshanth Jayakumar(yeshanth@netapp.com)
## @status public
## @change 2009-02-09 Glenng: Original template created
## @change 2009-06-09 yeshanth: Added attr_diff and attr_check
#
#######################################################################

package San::VdiskSanity::lun_attr_parse;

use San::SanLog;
use Error;

# Define our mappings structure. The user should declare a pointer to
# this structure in their module, then dereference the hash pointer.
our %mapStruct = {"LUN_Name"  => "undef",
                  "Num_Maps"  => 0,
                  "Mapped_To" => [],
                  "LUN_ID"    => [],
                  "Protocol"  => []
                 }; 

@ISA = Exporter;
require Exporter;
our @ISA = qw(Exporter);

# Let importer see these specified functions and the hash structure.
our @EXPORT = qw(   lun_showInfo_parse_1_key 
                    lun_showInfo_parse_2_key 
                    lun_attribute_parse 
                    attr_diff
                    lun_showMapped_parse 
                    %mapStruct);


# Forward declarations of internal sub-routines.
sub strip_whitespace;
sub dbg_print;

# A couple of definitions for our delimiters.
my $LUN_SHOW_DELIMITER = ":";
my $LUN_ATTR_DELIMITER = "=";
my $LUN_MAP_DELIMITER  = ";";

# Debug print flag: flip to 1 to see debug output.
my $DEBUG_FLAG = 0;

=head2 B<lun_showInfo_parse_1_key>

Given the "lun show -i  <lun_name>" output, this module
parses and returns the attribute value corresponding to the
specified keys.

=over 4
    
=item Inputs
 
=item *attrArrayPtr

Pointer to array of attributes

=item *key

Scaler for <key> to locate

=item Outputs
    
Returns PASS and (vol_obj) on success
Returns FAIL on error message on failure

=item Examples

    $key = "Volume Stats";
    
    $value = lun_showInfo_parse_1_key(\@attr_stream, $key);
    if ($value) {
        print "Parser returned value $value for key ;.chprt<$key>.\n";
    }
    else {
        print "Parser routine returned error indication.\n";
    }

=back

=cut

#######################################################################
#
# Function: lun_showInfo_parse_1_key
#
# Description: This function will parse the passed list of LUN 
#              attributes for the specified key, returning the 
#              corresponding value. 
#
# Input:  Pointer to array of attributes.
#         Scaler for <key> to locate.
#
# Return: Value related to passed in key, if found.
#         0 if key is not found.
#
# History:
# 03/04/2009  Glenn M. Gobeli
#
#######################################################################
sub lun_showInfo_parse_1_key {
    my ($attrArrayPtr, $key) = @_; # Pull off attributes pointer and key.
    my $value = 0;
    my ($line, @linePieces);

    dbg_print("lun_showInfo_parse_1_key() called...");
        
    foreach $line (@$attrArrayPtr) {
	if ($line =~ /$key/) {
	    # We have found the key.
	    dbg_print("Key located in: \"$line\"");
	    
	    # Split the line into separate pieces.
	    @linePieces = split($LUN_SHOW_DELIMITER, $line);
	    
	    # Assign the latter piece to our "value".
	    chomp($value = $linePieces[1]);
	    
            # Now remove any leading or trailing whitespaces.
	    strip_whitespace($value);
	    last;
        }
    }
        
    return $value;    
}

=head2 B<lun_showInfo_parse_2_key>

Given the "lun show -i  <lun_name>" output, this module
parses and returns the attribute value corresponding to the
specified keys.

=over 4
    
=item Inputs
 
=item *attrArrayPtr

Pointer to array of attributes

=item *key1

Scaler for <key1> to locate

=item *key1

Scaler for <key2> to locate

=item Outputs
    
Returns PASS and (vol_obj) on success
Returns FAIL on error message on failure

=item Examples

    $key  = "VTOC Info";
    $key2 = "UUID";
    
    $value = lun_showInfo_parse_2_key(\@attr_stream, $key, $key2);
    if ($value) {
        print "Parser returned value $value for keys <$key><$key2>.\n";
    }
    else {
        print "Parser routine returned error indication.\n";
    }

=back

=cut

#######################################################################
#
# Function: lun_showInfo_parse_2_key
#
# Description: This function will parse the passed list of LUN 
#              attributes for the specified keys, returning the 
#              corresponding value.
#
# Input:  Pointer to array of attributes.
#         Scaler for first <key> to locate.
#         Scaler for second <key> to locate.
#
# Return: Value related to passed in keys, if found.
#         0 if keys are not found.
#
# History
# 03/04/2009  Glenn M. Gobeli
#
#######################################################################
sub lun_showInfo_parse_2_key {
    my ($attrArrayPtr, $key1, $key2) = @_; # Pull off attributes pointer and keys.
    my $value = 0;
    my ($line, @linePieces);
    my $firstKeyFound = 0;
  
    dbg_print("lun_showInfo_parse_2_key() called...");

    foreach $line (@$attrArrayPtr) {
	if ($line =~ /$key1/) {
	    # We have found the first key.
	    $firstKeyFound = 1;
	    dbg_print("Key 1 located in: \"$line\"");
        }
	    	    
	if ($line =~ /$key2/ && $firstKeyFound) {
	    dbg_print("Key 2 has been found in \"$line\"");
	    
	    # Split the line into separate pieces.
	    @linePieces = split($LUN_SHOW_DELIMITER, $line);
	    
	    # Assign the latter piece to our "value".
	    chomp($value = $linePieces[1]);
	    
            # Now remove any leading or trailing whitespaces.
	    strip_whitespace($value);
	    last;
        }
    }
        
    return $value;    
}

=head2 B<lun_attribute_parse>

Given the "lun attribute list <lun_name>" output, this module
parses and returns the attribute value corresponding to the
specified key.

=over 4
    
=item Inputs
 
=item *attrArrayPtr

Pointer to array of attributes

=item *key

Scaler for <key> to locate

=item Outputs
    
Returns $value for the corresponding key in the Attribute-Key pair

=item Examples

    $key = "path_last";

    $value = lun_attribute_parse(\@list_stream, $key);
    if ($value) {
        print "Parser returned value $value for key <$key>.\n";
    }
    else {
        print "Parser routine returned error indication.\n";
    }

=back

=cut

#######################################################################
#
# Function: lun_attribute_parse
#
# Description: This function will parse the passed list of LUN 
#              attributes for the specified key, returning the 
#              corresponding value. 
#
# Input:  Pointer to array of attributes.
#         Scaler for <key> to locate.
#
# Return: Value related to passed in key, if found.
#         0 if key is not found.
#
# History:
# 03/05/2009  Glenn M. Gobeli
#
#######################################################################
sub lun_attribute_parse {
    my ($attrArrayPtr, $key) = @_; # Pull off attributes pointer and key.
    my ($line, @linePieces, $value);

    $value = 0;
  
    dbg_print("lun_attribute_parse() called...");

    foreach $line (@$attrArrayPtr) {
	    if ($line =~ /$key/) {
	        # We have found the key.
	        dbg_print("Key located in: \"$line\"");
	    
	        # Split the line into separate pieces.
	        @linePieces = split($LUN_ATTR_DELIMITER, $line);
	    
	        # Assign the latter piece to our "value".
	        chomp($value = $linePieces[1]);
	    
   	        dbg_print("chomped value: $value");
   	        
            # Now remove any leading or trailing whitespaces.
	        strip_whitespace($value);

   	        dbg_print("stripped value: $value");
	        
	        # Now that we have located our key, get out.
	        last;
        }
    }
        
    return $value;    
}

=head2 B<attr_diff>

Given the pointers to two attribute lists, this module returns
the pointers to the common and the difference in the specified
lists.

=over 4
    
=item Inputs
 
=item *ref_attributes_1

Pointer to first attributes output

=item *ref_attributes_2

Pointer to second attributes output

=item Outputs
    
Returns PASS and (vol_obj) on success
Returns FAIL on error message on failure

=item Examples

    ($ref_common_attributes, $ref_difference_attributes) = attr_diff(
                                                            \@attributes_before,
                                                            \@attributes_after
                                                            );

    # Dereferencing
    @common_attributes      = @{$ref_common_attributes};
    @difference_attributes  = @{$ref_difference_attributes};

=back

=cut

##################################################
# Function     : attr_check
#
# Arguments        :
#      @attributes_1   -   attribute list 1
#      @attributes_2   -   attribute list 2
#
# Notes            :
#      This routine returns the common and difference
# attributes between attributes_1 and attributes_2
#
# History:
# 03/05/2009  Yeshanth Jayakumar
#
#
##################################################
sub attr_diff
{
    my (
         $ref_attributes_1,
         $ref_attributes_2
      )
      = @_;

    my $errmsg          = "";
    my @common          = ();
    my @difference      = ();
    my %count           = ();
    my $element         = "";
    my @attributes_1    = ();
    my @attributes_2    = ();

    @attributes_1      = @{$ref_attributes_1};
    @attributes_2      = @{$ref_attributes_2};

    foreach $element (@attributes_1, @attributes_2) {
        $count{$element}++;
    }

    foreach $element (keys %count) {
        push @{ $count{$element} > 1 ? \@common : \@difference }, $element;
    }

    dbg_print("attr_diff: Source Attributes:        @attributes_1");
    dbg_print("attr_diff: Destination Attributes:   @attributes_2");
    dbg_print("attr_diff: Common Attributes:        @common");
    dbg_print("attr_diff: Destination Attributes:   @difference");

    return (\@common, \@difference);
}

=head2 B<lun_showMapped_parse>

This function will parse the passed list of LUN 
mappings (lun show -m) for the specified LUN name, 
returning mapping information.

=over 4
    
=item Inputs
 
=item *mappedArrayPtr

Pointer to lun show -m output

=item *mapPtr

Pointer to hash mapStruct

=item Outputs
    
Returns PASS and (vol_obj) on success
Returns FAIL on error message on failure

=item Examples

    $mapPtr->{"LUN_Name"} = "/vol/vol1/lun2";
    $mapPtr->{"Num_Maps"} = 0;
    $value = lun_showMapped_parse(\@mapped_stream, $mapPtr);
    if ($value) {
        print qq(Returned with number of maps equal $mapPtr->{"Num_Maps"}.\n);
        for (my $x = 0; $x < $mapPtr->{"Num_Maps"}; $x++) {
            print qq(LUN $mapPtr->{"LUN_Name"} is mapped to: $mapPtr->{"Mapped_To"}->[$x], with LUN ID $mapPtr->{"LUN_ID"}->[$x], and protocol $mapPtr->{"Protocol"}->[$x].\n);
        }
    }
    else {
        print "Parser routine returned error indication.\n";
    }

=back

=cut

#######################################################################
#
# Function: lun_showMapped_parse
#
# Description: This function will parse the passed list of LUN 
#              mappings (lun show -m) for the specified LUN name, 
#              returning mapping information.
#
# Input:  Pointer to array of LUN mapping output.
#         Pointer to hash mapStruct.
#
# Return: 1 if key (LUN name) is found.
#         0 if key is not found.
#         mapStruct loaded if successful.
#
# History:
# 03/05/2009  Glenn M. Gobeli
#
#######################################################################
sub lun_showMapped_parse {
    my ($mappedArrayPtr) = shift(@_); # Pull off mapped info array.
    my ($mapPtr) = shift(@_);         # Pull off mappings pointer.
    my ($x, $retVal, $line, @linePieces, $numPieces);
  
    dbg_print("lun_showMapped_parse() called...");

    $x = $retVal = 0;

    foreach $line (@$mappedArrayPtr) {
        # First, pull any possible whitespaces off of the front.
        $line =~ s/^\s+//g;
        
	# Now we need to get rid of the extra whitespaces between items.
	# We will do this by changes whitespaces to ';'.
	$line =~ s/\s+/$LUN_MAP_DELIMITER/g;
	    
	# Now split the line into separate pieces.
	@linePieces = split($LUN_MAP_DELIMITER, $line);
        $numPieces = @linePieces;
        dbg_print("Line broken into $numPieces");
        dbg_print("@linePieces");
        	    
	if ($numPieces == 4 && @linePieces[0] =~ /$mapPtr->{"LUN_Name"}/) {
	    # We have found the LUN.
	    dbg_print("LUN Name located in: \"$line\"");
	    
	    # Assign the piece to our mappings pointer.
	    $mapPtr->{"Num_Maps"} += 1;
	    chomp($mapPtr->{"Mapped_To"}->[$x] = @linePieces[1]);
	    chomp($mapPtr->{"LUN_ID"}->[$x]    = @linePieces[2]);
	    chomp($mapPtr->{"Protocol"}->[$x]  = @linePieces[3]);
 	    
            # Now remove any leading or trailing whitespaces.
	    strip_whitespace($mapPtr->{"Mapped_To"}->[$x]);
	    strip_whitespace($mapPtr->{"LUN_ID"}->[$x]);
	    strip_whitespace($mapPtr->{"Protocol"}->[$x]);
	    
	    $x++;
	    
	    $retVal = 1;
        }
        elsif ($retVal == 1 && $numPieces == 3) {
	    # This means we have previously found a matching line
	    # And this is another line for the same LUN name.
	    # Assign the piece to our mappings pointer.
            dbg_print("Adding another mapping to $mapPtr->{\"LUN_Name\"}");
	    $mapPtr->{"Num_Maps"} += 1;
	    chomp($mapPtr->{"Mapped_To"}->[$x] = @linePieces[0]);
	    chomp($mapPtr->{"LUN_ID"}->[$x]    = @linePieces[1]);
	    chomp($mapPtr->{"Protocol"}->[$x]  = @linePieces[2]);
 	    
            # Now remove any leading or trailing whitespaces.
	    strip_whitespace($mapPtr->{"Mapped_To"}->[$x]);
	    strip_whitespace($mapPtr->{"LUN_ID"}->[$x]);
	    strip_whitespace($mapPtr->{"Protocol"}->[$x]);
	    
	    $x++;	    
       }
    }
        
    return $retVal;    
}


#######################################################################
#
# Function: strip_whitespace
#
# Description: This internal function will strip any leading or 
#              trailing whitespaces from the passed in string.
#
# Input:  String array (used by reference).
#
# Return: Same string array.
#
# History:
# 03/04/2009  Glenn M. Gobeli 
#
#######################################################################
sub strip_whitespace {
    @_[0] =~ s/^\s+//; # Substitute leading whitespace with nothing.
    @_[0] =~ s/\s+$//; # Substitute trailing whitespace with nothing.
}


#######################################################################
#
# Function: dbg_print
#
# Description: This internal function will print the passed line
#              of output if the DEBUG flag is set.
#
# Input:  String to print.
#
# Return: None
#
# History:
# 03/05/2009  Glenn M. Gobeli
#
#######################################################################
sub dbg_print {
    if ($DEBUG_FLAG) {
	    print "DBG: ", @_[0], "\n";
    }
}


# End of module.
1;
