package SayUtil;
use strict;
use Exporter;
use vars ('@ISA','@EXPORT');
@ISA = qw(Exporter);
@EXPORT = qw(say_retry);

use Tharn;
use Time::HiRes qw(gettimeofday tv_interval);
use vars ('@ISA','@EXPORT');
@ISA = qw(Exporter);
@EXPORT = qw(say_retry
             say_time);

## @name ::say_retry
## @usage say_retry(@say_params,,"-retry",3,"-iter_sleep",2,"-cmds_time","local")
## @synopsis
##    Adds the following new functionality to THARN say call
##
##       1. Retries the say command until either the call returns or the retry limit is reached.
##       2. Ability to calculate the time taken to execute each say command and 
##          time taken for any command with the command list separated by '; '
##    Adds these extra arguments to the existing say call: iter_sleep, retry_limit, cmds_time
##    Retries in case of timeout of the say command and in case connection is refused because 
##    all rsh sockets/ports are currently in use.
##    It does not retry if the prompt string appears and then the connection times out
##    It only retries in cases where the connection does not see a prompt string but times out
##    because the filer is too busy to respond.
##    Used if running against a filer under heavy load which takes along time to respond for the say call
##    The value of retry_limit specifies the number of times the say call must be retried.
##    The value of iter_sleep specifies the amount of time to sleep between consecutive say retries.

##    It adds a cmds_time flag to the say call. This call can be used to calculate the
##    amount of time taken to run a particular command. This flag takes three values: filer, local and both
##    Default is filer
##    If filer is specified, it use the time command on the filer to find the time taken to execute a command
##    If the command is a ; separated list, it will find the time taken to execute each individual command
##    If local is specified, it will use the time on the local machine to find the time taken to execute the command
##    If both is specified, it will use both the above specified methods
##    If timing information is requested, it will append a array reference at the end of the say call output
##    The array referenced by this array ref will have the following time information:
##    If cmds_time is filer, the array will contain the times to execute each of the ';' separated 
##    list of commands in the cmds param
##    If cmds_time is local, the array will contain one element which will be 
##    the times to execute the say call
##    If cmds_time is both, the last element of the array array will be  the times to execute the say call
##    All the preceding elements will contain the times to execute each of the ';' separated 
##    list of commands in the cmds param
 
##    Returns the output of say in an array  (@out)

## @usage  
##    In these examples $cmd can be a single command like vol status or a ; separated list like
##    vol status;snap list test vol;aggr status
##    my(undef,undef,$result)=say_retry($fc,$cmd,"-wait","done","-retry",3,"-iter_sleep",2 );
##    The above call will retry the calling the say function 3 times 
##    with sleep of 2 seconds between each retry
##    my(undef,undef,$result,undef,$ref)=say_retry($fc,$cmd,"-wait","done","-retry",3,"-iter_sleep",2,"-cmds_time","both");
##    This will do the retries as specifed above  and also find time to execute each command
##    as measured from the filer and host(since the value is set to both)
##    $ref is a reference to an array of the time information collected.
##    $ref->[scalar(@$ref)-1] wil be the time taken to execute the complete say command as measured from the test host
##    $ref->[0] to $ref->[scalar(@$ref)-2] will contain the times taken to execute each of the 
##    individual command seperated by the ;
##
## @Example
##     my $cmd = "snap list max_test_vol_348;snap list max_test_vol_348;snap list max_test_vol_348;snap list max_test_vol_348;snap list max_test_vol_348;snap list max_test_vol_348;snap list max_test_vol_348;snap list max_test_vol_348;echo done";
##     my $cmd = "snap list max_test_vol_80;snap list max_test_vol_81;snap list max_test_vol_82;snap list max_test_vol_83";
##     my(undef,undef,$result,undef,$ref)=say_retry($fc,$cmd,"-retry",3,"-iter_sleep",2,"-cmds_time","both");
##     logcomment("Result is $result");
##     my @times = @$ref;
##     # Print inividual execution times  
##     for (my $i=0;$i<(scalar(@times) - 1) ;$i++) {
##     	print($times[$i] . "==");
##     }
##     # Print time to execute the say call  
##     print("\n $times[(scalar(@times) - 1)] \n");
##
#################################
sub say_time(@) {
   my @args = @_;
   my @push_args = ("-retry",0,"-iter_sleep",0);
   push(@args,@push_args);
   my @out = say_retry(@args);
   return(@out);
}
##################################
sub say_retry(@) {

   my @args = @_ ;
   my $retry =5;
   my $sleep_time=30;
   my $cmds_time=0;
   my $cmds_time_args="filer";
   my @splice_ind;
   my @splice_flags;
   my $is_wait=0;
   my @times;
   my $out_index=2;
   my $stime;
   my $etime;
   for (my $i=0;$i <= scalar(@args); $i++){
           if ($args[$i] =~ /retry/) {
                   $retry = @args[$i+1];
                   splice(@args,$i,2);
           }
           if ($args[$i] =~ /iter_sleep/) {
                   $sleep_time = @args[$i+1];
                   splice(@args,$i,2);
           }
           if ($args[$i] =~ /cmds_time/) {
                   $cmds_time = 1;
                   splice(@args,$i,1);
                   if ($args[$i] =~ /filer|local|both/) {
                      $cmds_time_args=$args[$i];
                      splice(@args,$i,1);
                   }
           }
           if ($args[$i] =~ /wait/) {
                   $is_wait = 1;
           }


   }
   my @out;
   my $i=0;
   if($cmds_time ==1 && $cmds_time_args =~ /filer|both/){
      my $cmd_str = $args[1];
      $cmd_str =~ s/;/;time /g;
      $cmd_str = "priv set diag; time " . $cmd_str;
      $args[1]=$cmd_str;
   }
   while(1){
      eval { 
            $stime = [gettimeofday];
            @out = say(@args);
            $etime = [gettimeofday];
         };
      $i++;
      if($@){
            
           if(($@ =~ /Err=TIMEOUT waiting for waitstrings\/prompts/i) ||($@ =~ /Connection refused/)){
                 if($i < ($retry)){
                    sleep($sleep_time);
                    logcomment("Say returned retriable error:$@ . \nRetry No: $i");
                 }
                 else{
                   # If the retry fails on last retry cycle too then we return what ever say retuens to us
                   logcomment("Retry No: $i - Last Retry");
                   sleep($sleep_time);
                   @out = say(@args);
                   last;
                 }
            }
            else {
               logresult("FAIL", abort=>1,"Say returned a non-retry error:\n $@");
               last;   
            }
      }
      else{
            if ($is_wait==1) {
               $out_index=1;
            }
            else {
               $out_index=2;
            }
            if($cmds_time ==1){
               if ($cmds_time_args =~ /filer|both/) {
                  @times = ($out[$out_index] =~ /\d+ms/g) ;
               }
               if($cmds_time_args =~ /local|both/)  {
                  push(@times,tv_interval($stime, $etime));
               }
               push(@out,\@times);
            }
            last;
      }
   }
   return(@out);
}

1;
