# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/SayWrap.pm#1 $

## @summary Library of say-wrapper subroutines
##
## @description
##    This package defines routines which wrap the Tharn say command.
##
## @author remi@netapp.com dl-test-infra@netapp.com
## @status public

package SayWrap;
use Params::Validate qw(validate validate_pos validate_with BOOLEAN OBJECT SCALAR ARRAYREF UNDEF);

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(sayEOF sayWAIT sayPerl);
our @EXPORT_OK = qw(chop_debug chop_priv_warning chop_syslog chop_wouldbe_syslog
		    $SYSLOG_DATE prompt_loop
		    );

use strict;
use Tharn;
use Carp qw(croak);

use constant PERL => 'X:\perl\bin\perl.exe';

## @name ::sayEOF
## @usage sayEOF($connobj, $command, %options)
## @synopsis
##    Sends the provided command to the established connection, collects all
##    output from that command, and then returns that output.  Prior to
##    returning the output, attempts to prune the output such that it will
##    always be the same for a given command, regardless of the type of
##    connection used.  Hence, it prunes the command echo (if present), and
##    then any other recognizable messages.  By default, all options which prune
##    the command output are used, but may be selectively disabled (by setting
##    the respective options to zero).
## @param TIMEOUT Optional: timeout period (in seconds) before failing
## @param CHOP_PRIV Optional: whether to chop any priviledge warnings from the output
## @param CHOP_DEBUG Optional: whether to chop any debug messages from the output
## @param CHOP_SYSLOG Optional: whether to chop any syslog messages from the output
## @param CHOP_WOULDBE Optional: whether to chop any wouldbe-syslog messages from the output
sub sayEOF($$;%) {
  # XXX - As the first step to converting parameter names, I'm going to
  # allow both upper and lower cases
  my $conn  = shift;
  my $cmd   = shift;
  my %param = validate_with(
    params => \@_,
    spec   => {
      timeout      => { optional => 1 },
      chop_priv    => { type => BOOLEAN, optional => 1 },
      chop_debug   => { type => BOOLEAN, optional => 1 },
      chop_syslog  => { type => BOOLEAN, optional => 1 },
      chop_wouldbe => { type => BOOLEAN, optional => 1 }
    },
    ignore_case => 1
  );

  my @args = ();

  # We force the command to echo a string at the end and wait for
  # that string, since it catches the case where the connection dies
  # in the middle of the output.  (Not sure why NATE misses this.)
  # Solaris clients seem to have rsh connections drop often.
  #
  # The only time we don't bother doing this extra work is if
  # we're using a good, reliable connection type for which NATE
  # will throw errors if the connection is abnormally dropped.
  unless ($conn->conntype() eq "rsh-socket") {

    my $echo_str = "END_COMMAND_OUTPUT";
    if ( $conn->hostp->hosttype() =~ /^cache/ ) { 
        # XXX - NetCaches behave differently here.
        #    1 - 'echo' is not available by default, so you need to
        #        priv set advanced
        #    2 - The command 'echo FOO_BLAH_BAZ' actually prints
        #        'FOO _BLAH _BAZ' (notice the spaces)
        #

        $cmd .= "; priv set -q advanced; END'_'COMMAND'_'OUTPUT";
        $echo_str = "END _COMMAND _OUTPUT";
    } elsif ( $conn->hostp->hosttype() =~ /windows/i ) {
        # windows/DOS uses && to seperate commands on the same line, so we need
        # to feed say the proper command seperator when the target is Windows
        $cmd .= "&& echo END_COMMAND_OUTPUT";
    } else {
        $cmd .= "; echo END'_'COMMAND'_'OUTPUT";
    }

    $echo_str .= "\\r*\\n"; # make sure to catch any whitespace introduced by echo
    push @args, ('-wait', $echo_str);

  }

  push @args, ('-timeout', $param{'timeout'})
    if ((defined $param{'timeout'}) && ($param{'timeout'} =~ /\A-?\d+\z/));

  # Send the command and get all the output.
  my ($match, $before, $after, $prompt) = say($conn, $cmd, @args);
  my $text = $before;

  # Burt 109859 -- rsh connections send back output on two channels,
  # stdout and stderr, but echo only operates on stdout.  Hence, ignoring
  # $after for rsh connections can inadvertently truncate the output.
  # (However, the echo/wait still catches the Solaris dropped connection bug.)
  # This also gets the output if we didn't do the echo/wait at all.
  $text .= $after if ($conn->conntype() =~ /\A(rsh-|rlogin\z)/);

  # Burt 116286 -- we are never sure what kind of line separator the resulting
  # output will be using.  It seems to be dependent upon the a variety of
  # factors, but for the moment we shall accept "\r*\n", but only return text
  # that uses a simple "\n" as the line separator.
  $text =~ s/\r+(\n|\z)/$1/g;

  # Escape all metacharacters in the command before matching with it.
  my $cmd_exp = quotemeta($cmd);

  # Chop command echo.  If the command echo is present, we also chop everything
  # before it, under the presumption that we only care about the output between
  # the echo and END_COMMAND_OUTPUT.  This deals with the case where our output
  # splits a syslog message in two.
  $text =~ s/^(.*)$cmd_exp\s*(\n|\z)//m;

  # Avoid doing all chops if possible.
  return "" if ($text =~ /^\s*$/);

  chop_priv_warning($text)
    unless ((exists $param{'chop_priv'}) && !$param{'chop_priv'});
  chop_debug($text)
    unless ((exists $param{'chop_debug'}) && !$param{'chop_debug'});
  chop_syslog($text)
    unless ((exists $param{'chop_syslog'}) && !$param{'chop_syslog'});
  chop_wouldbe_syslog($text)
    unless ((exists $param{'chop_wouldbe'}) && !$param{'chop_wouldbe'});

  $text =~ s/^\s*//; # chop leading whitespace
  $text =~ s/\s*$//; # chop trailing whitespace
  return $text;
}

## @name ::sayPerl
## @description
##  wrapper for say command to allow sending multiline Perl scripts
##  over a connection and not have to worry about the restrictions
##  imposed by the shell running on the connection (i.e. user's default
##  shell). Specifically the csh is very restrictive.
##
## @synopsis 
## 	$($match, $before, $after, $prompt) =
##          sayPerl($cc, $cmd, '/usr/local/bin/sudo', -timeout, 30); 
##
## @arg
##  cc		Required: connection object
##
##  cmd		Required: Perl script to run on connection.
##
##  cmdprefix	Required: Commands to put before the call to perl. Typically
##		this would be sudo or could be a shell like /bin/csh. Pass
##              in an empty string if there is no command to prefix with.
##
##  opts	Optional: Say parameters
##
## @return	
##  returns the script output on success or an error message in $@ on failure
##
## @notes	
##  The perl script is run with the -MEnglish option so that scripts passed
##  in should use the English mneumonics like $OS_ERROR instead of $! because
##  in the csh we cannot escape the $! (Joy's revenge %^)).
##
##  Not recommended for telnet connections because there may be line
##  length restrictions depending on the system being used.

sub sayPerl ( $$$;% ) {
    use constant TARGET_WORD_LEN => 768; # target chars per -e quoted word
    my ($cc, $cmd, $cmdprefix, @opts) = validate_pos(@_,
        { type => OBJECT },
        { type => SCALAR },
        { type => SCALAR | UNDEF },
        (0) x (@_ - 3)
    );
    my (@out, $bline, @lines, $perl, $quote, $cmdline, $i);
    my $hosttype = $cc->hostp()->hosttype();

    if ( $hosttype =~ /^unix/ ) {
        # Use single quotes for unix clients
        $perl = "perl";
        $quote = q(');
    } elsif ( $hosttype =~ /^windows/ ) {
        # Windows client requires the perl script
        # to be enclosed by double quotes
        $perl  = PERL;
        $quote = q(");
    } else {
        croak "Hosttype '$hosttype' is not supported";
    }

    $cmdline = "$cmdprefix $perl -MEnglish ";

    # Remove leading spaces:
    # On windows the command fails if there is any leading space
    $cmdline =~ s/^\s+//;

    # csh can only handle a word length of 1024 bytes so break up the
    # command so as not to exceed the word length. Each "word" or section
    # of perl code is prefixed with the -e option. Perl assembles all
    # the -e options into one program and runs it. The major gotcha is
    # that if you break up in the middle of string or name it will insert
    # a newline at that point. The best place to break is after a ;\n
    #
    # The algorithm used breaks the cmd up into lines ended by ; and 
    # builds it up until it exceeds TARGET_WORD_LEN and then it adds it
    # to the command line as a word.
    if (length($cmd) > TARGET_WORD_LEN) {
        ($bline, @lines) = split(/(;)\n/, $cmd);
        for ($i = 0; $i <= $#lines; $i++) {
            if ((length($bline) + length($lines[$i])) > TARGET_WORD_LEN) {
                $bline .= $lines[$i++] if $lines[$i] eq ';';
                $cmdline .= "-e $quote$bline$quote ";
                $bline = '';
            }
            $bline .= $lines[$i];
        }
        if (length($bline) > 0) {
            $cmdline .= "-e $quote$bline$quote ";
        }
    } else {
        $cmdline .= "-e $quote$cmd$quote";
    }

    # Strip off any remaining newlines from the command
    $cmdline =~ s/\n//g;

    @out = say($cc, $cmdline, @opts);

    # Remove command echo from output if any
    $cmdline = quotemeta($cmdline);
    s/^.*$cmdline\s*// foreach @out;

    return @out;
}

## @name ::sayWAIT
## @usage sayWAIT($connobj, $command, $waitstring, %options)
## @synopsis
##    Says the given command to the given connection and waits for the output
##    to match one of the given regexp strings -- returns as soon as one of
##    the given waitstrings or the next prompt is seen.  By default, all
##    options which prune the command output are used, but may be selectively
##    disabled (by setting the respective options to zero).  Returns a tuple
##    of the matched waitstring and the text before it.  If a match is never
##    seen, or if an error occurs, the same tuple is returned, but the matched
##    waitstring is empty and there is an error  message in $@.
##
##    The 'waitstring' argument is passed to say() using '-prompt'. It can be
##    'one or more strings, treated as Perl regular expressions, expected to
##    appear in the connection's output. The list may include string arrays
##    and/or array references.' (from the say() docs)
##
## @param TIMEOUT Optional: timeout period (in seconds) before failing
## @param CHOP_PRIV Optional: whether to chop any priviledge warnings from the output
## @param CHOP_DEBUG Optional: whether to chop any debug messages from the output
## @param CHOP_SYSLOG Optional: whether to chop any syslog messages from the output
## @param CHOP_WOULDBE Optional: whether to chop any wouldbe-syslog messages from the output
sub sayWAIT($$$;%) {
  my ($conn, $cmd, $waitstring, %param) = @_;

  my ($matched, $text) = ("", "");
  eval {
    # Setup arguments to say().
    # We need to set the provided waitstring as a prompt in order to get the
    # same behavior on persistent and persay connections.  If we passed it as
    # a real waitstring to say(), then we'd need to use -noprompt to make say()
    # return as soon as the waitstring was seen.  However, that would also
    # cause us to wait forever on persistent connections if the waitstring
    # never appeared (whereas persay connections get closed and say() returns).
    my @args = ('-prompt', $waitstring);
    push @args, ('-timeout', $param{'TIMEOUT'})
      if (defined $param{'TIMEOUT'});
    push @args, '-eval_error';

    # Say the command and wait for the expected string/regexp
    my ($match, $before, $after, $prompt) = say($conn, $cmd, @args);
    if ($@) {
      # Some error occurred -- get the error message and fail.
      $text .= $after if (defined $after);
      $text .= $prompt if (defined $prompt);
      if ($@ =~ /^Err=(.+)\s*$/) {
	die "$1\n";
      } else {
	die "$@\n";
      }
    }
    # We successfully matched either the string or a prompt.
    # Matched text is in $prompt, and everything before it is in $after.
    if (
	(
	    (ref($waitstring) eq 'ARRAY') &&
	    (grep $prompt =~ /$_/, @{$waitstring})
	) ||
	    ($prompt =~ /$waitstring/)
    ) {
      $matched = $prompt;
    }
    $text = $after;

    return 1;
  };
  # If there was an error, then $@ was set.
  # Make sure not to clear it.

  # Escape all metacharacters in the command before matching with it.
  my $cmd_exp = quotemeta($cmd);

  # Chop command echo.  If the command echo is present, we also chop everything
  # before it, under the presumption that we only care about the output between
  # the echo and waitstring.  This deals with the case where our output
  # splits a syslog message in two.
  $text =~ s/^(.*)$cmd_exp\s*?(\n|\z)//m;

  # Avoid doing all chops if possible.
  unless ($text =~ /^\s*$/) {
    # Prune output
    chop_priv_warning($text)
      if ((defined $param{'CHOP_PRIV'}) ? $param{'CHOP_PRIV'} : 1);
    chop_debug($text)
      if ((defined $param{'CHOP_DEBUG'}) ? $param{'CHOP_DEBUG'} : 1);
    chop_syslog($text)
      if ((defined $param{'CHOP_SYSLOG'}) ? $param{'CHOP_SYSLOG'} : 1);
    chop_wouldbe_syslog($text)
      if ((defined $param{'CHOP_WOULDBE'}) ? $param{'CHOP_WOULDBE'} : 1);
  }

  $text =~ s/^\s*//; # chop leading whitespace
  $text =~ s/\s*$//; # chop trailing whitespace

  return ($matched, $text);
}

## @name ::chop_priv_warning
## @usage chop_priv_warning($string)
## @synopsis
##    Chops out the warning that can appear in command output when
##    using "priv set" to change the privilege state.
##    It receives its argument string by reference and modifies it directly.
##    Returns whether it chopped out the warning (if it was present).
##    Will only chop one warning from the string.
sub chop_priv_warning($;) {
  if (($_[0] =~ s/^Warning\:\s+These\s+advanced\s+commands\s+are\s+potentially\s+dangerous\;\s+use\s+them\s+only\s+when\s+directed\s+to\s+do\s+so\s+by\s+Network\s+Appliance\s+personnel\.\s*//)
      || ($_[0] =~ s/^Warning\:\s+These\s+diagnostic\s+commands\s+are\s+for\s+use\s+by\s+Network\s+Appliance\s+personnel\s+only\.\s*//)
      || ($_[0] =~ s/^Warning\:\s+These\s+test\s+commands\s+are\s+for\s+internal\s+testing\s+only,\s+and\s+are\s+not\s+present\s+on\s+customers\'\s+systems\.\s*//)) {
    return 1;
  } else {
    return 0;
  }
}

## @name ::chop_debug
## @usage chop_debug($string)
## @synopsis
##    Chops out any/all console debug messages from the given string.
##    It recieves its argument string by reference and modifies it directly.
##    Returns whether it chopped out any messages (if any were present).
sub chop_debug($;) {
  return ($_[0] =~ s/^DBG\:.*?(\n|\z)//mg);
}

# e.g. Wed Jul 18 15:23:00 EST
our $SYSLOG_DATE = '(?:Mon|Tue|Wed|Thu|Fri|Sat|Sun) (?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) [ \d]\d \d\d\:\d\d\:\d\d [A-Z]{3}';

{
  # Basically, [(%s@)?(%s/)?(%s: )?%s:%s]:
  my $syslog_type = '\[(?:\S+?\@)?(?:\S+?\/)?(?:\S+?\: )?\w+?\:\w+?]:';

  ## @name ::chop_syslog
  ## @usage chop_syslog($string)
  ## @synopsis
  ##    Chops out any/all console syslog messages from the given string.
  ##    It recieves its argument string by reference and modifies it directly.
  ##    Returns whether it chopped out any messages (if any were present).
  sub chop_syslog($;) {
    return ($_[0] =~ s/^$SYSLOG_DATE $syslog_type\s+\S.*?(\n|\z)//mg);
  }

  ## @name ::chop_wouldbe_syslog
  ## @usage chop_wouldbe_syslog($string)
  ## @synopsis
  ##    Chops out any/all wouldbe syslog messages from the given string.
  ##    These are messages which SHOULD be syslog messages, but aren't.
  ##    Instead, they are pieces of bare text that get printed to the console.
  ##    It receives its argument string by reference and modifies it directly.
  ##    Returns whether it chopped out any messages (if any were present).
  sub chop_wouldbe_syslog($;) {
    my $cnt = 0;
    $cnt++ if ($_[0] =~ s/^Power Supply \#\d+ is removed\s*?(\n|\z)//mg);
    $cnt++ if ($_[0] =~ s/^blacksmith \#\d+\:.*?(\n|\z)//mg);
    $cnt++ if ($_[0] =~ s/^CIFS (local|partner) server is running\.\s*?(\n|\z)//mg);
    $cnt++ if ($_[0] =~ s/^$SYSLOG_DATE last message repeated \d+ times\s*?(\n|\z)//mg);
    $cnt++ if ($_[0] =~ s/^$SYSLOG_DATE $syslog_type\s+Autosupport\s+is\s+not\s+configured\s+to\s+send\s+to\s+Network\s+Appliance\.\s+Enabling\s+Autosupport\s+can\s+significantly\s+speed\s+problem\s+determination\s+and\s+resolution\s+should\s+a\s+problem\s+occur\s+on\s+your\s+system\.\s+\s+For\s+further\s+information\s+on\s+Autosupport,\s+please\s+see:\s+http\:\/\/now\.netapp\.com\/autosupport\/\s*?(\n|\z)//mg);
    return $cnt;
  }
}

## @name ::prompt_loop
## @synopsis
##   prompt_loop(connection => $connectrec,
##                table => [
##                        "prompt1>"    => "answer to prompt1",
##                        "prompt2>"    => sub { "answer to $_[0]" },
##                ],
##                done => ["prompt3>"],     # end when prompt3 seen
##                max_tries => 50,
##                say_opts => ['-timeout' => 600,
##                             '-max-idle' => 60,
##                             '-ignore' => 'bad_output']);
## 
## @description
## 
## A routine for answering a long series of various prompts that may
## occur optionally or in any order, particularly "setup" types of
## questions.
## 
## Listen for each of the prompts in the keys in $table and in
## $done.  If a key in %$table matches, send the string given by
## its corresponding value.  If a string in $done matches, then
## return, returning the string that matched.  If an error occurs
## (timeout, maximum tries exceeded, unexpected EOF, etc.) then an
## exception is thrown.
## 
## @param connection Required: The connection record object to use
## (such as is returned by Tharn::connect)
## 
## @param table Required: An array reference containing a list of
## prompts (regular expressions) and their responses, where each
## prompt is followed by its corresponding response.  This is an array
## instead of a hash in order to preserve order, in that earlier
## prompts are checked for in the output before later prompts.
## A response value that is a string will send that string.  A response
## value which is a code reference will cause that code reference to be
## called, and its return value used as the string to send.  The response
## or computed response should not include a newline or carriage return
## (prompt_loop will itself supply a carraige return).  If the
## response or computed response is undef, then a response will not be
## sent.
##
## @param done Required: An array reference containing a list of
## prompts (regular expressions) for which if prompt_loop sees them
## it should return.  The return value of prompt_loop is whichever
## prompt in done matched.
## 
## @param max_tries Required: Maximum number of times that prompt_loop
## may read and respond to a prompt.  (This exists separate from
## timeout as a mechanism to prevent furiously busy prompt and
## response loops when the target of the connection refuses a given
## answer and asks the same prompt again)
## 
## @param say_opts Optional: An array reference containing an optional
## list of options to pass to the Tharn::say command when waiting for
## new output.

sub prompt_loop {
    my %opts = validate(@_, {
                             connection => { isa => "Connectrec" },
                             table => { type => ARRAYREF },
                             done => { type => ARRAYREF },
                             max_tries => { type => SCALAR },
                             say_opts => { type => ARRAYREF, default => [] },
                             });
    my %table;
    my @keys;
    while (@{$opts{table}}) {
        my $key = shift @{$opts{table}};
        my $value = shift @{$opts{table}};
        $table{$key} = $value;
        push @keys, $key;
    }
    my $prompt;
    my $tries;
    LOOP:
    for ($tries=0;$tries < $opts{max_tries}; $tries++) {
        (undef, undef, undef, $prompt) =
            say ($opts{connection}, "",
                 '-noprompt',
                 '-prompt', [@keys, @{$opts{done}}],
                 @{$opts{say_opts}},
                );
        foreach my $key (@keys) {
            if ($prompt =~ /$key/) {
                my $value = $table{$key};
                my $send = (ref($value) eq "CODE") ? &$value($prompt) : $value;
                if (defined($send)) {
                    say ($opts{connection}, "$send\r",
                         '-nocr',
                         '-blind',
                         '-swallow');
                }
                next LOOP;
            }
        }
        # $prompt did not match any of the keys in @keys;
        # so it must be from $opts{done}.
        last LOOP;
    }
    
    if ($tries==$opts{max_tries}) {
        die "Too many prompts ($opts{max_tries}), probably a loop\n";
    }
    return ($prompt);
}

1;
