#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Scalability.pm#1 $
#
## @summary	Scalability - Class for running tests on many clients
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Scalability - Object oriented interface for running tests on multiple clients.

=head1 SYNOPSIS

  use Scalability;

  my @clients = (
      { client => 'client1' },
      { client => 'client2' },
  );
  my %plan = (
      # Global parameters (all tests on all sequence)
      # Parameter GLOBAL_PARAM is always 'global' while
      # VARY_GLOBAL is either 'vglobalA' or 'vglobalB'
      parameters => { GLOBAL_PARAM => 'global',
                      VARY_GLOBAL  => [ qw( vglobalA vglobalB ) ] },
      sequence => [
          # Sequence point 0
          {
              # Block parameters (all tests on this sequence)
              # Parameter BLOCK_PARAM is always 'block' while
              # VBLOCK is 'valA' on 40% and 'valB' on the other 60%
              parameters => { BLOCK_PARAM => 'block',
                              VBLOCK      => { valA => '40%',
                                               valB => '60%' } },
              # Tests to run in parallel
              # Sequence point is done when all three tests finish
              list => [
                  { script => $testA }, # Run test on 1/3 of clients
                  { script => $testB }, # Run test on 1/3 of clients
                  { script => $testC }, # Run test on 1/3 of clients
              ]
          },
          # Sequence point 1
          {
              # Tests to run in parallel
              # Sequence point is done when first test finishes
              list => [
                  {
                      # Run each thpl on the exports.plan file
                      # sequentially. This test is done when all
                      # thpl's are finished
                      script => 'exports.plan',

                      # Run this test on one client only
                      nclients => 1,

                      # Parameters for this test
                      parameters => { LOCAL_PARAM => 'local' }
                  },
                  # The following tests must be restarted if they finish
                  {
                      # Run testD on 30% of the total
                      # number of clients - 1
                      script => $testD, continuous => 1,
                                        nclients   => '30%' },
                   {
                      # Run testE on 70% of the total
                      # number of clients - 1
                      script => $testE, continuous => 1 },
              ]
          },
          # Sequence point 2
          {
              # Finish this sequence if it
              # is still running in 300 seconds
              duration => 300,

              # Tests to run in parallel
              # Sequence point is done when first test finishes
              list => [
                  {
                      # Run testF then testG
                      scripts => [ $testF, $testG ],

                      # Run this test on 10 clients
                      nclients => 10
                  },
                  {
                      # Run testH on all clients but 10
                      script => $testH, continuous => 1 },
              ]
          }
      ]
  );

  # Create the scalability object
  $obj = Scalability->new( plan    => \%plan,
                           clients => \@clients,
                           vipaddr => '10.0.0.1',
                           vnumber => 100 );

  # Start the test
  $obj->start();

  # Get the status from the main test manager when
  # the test is run with the background option set
  $obj->status();

  # Parse results and create summary
  $obj->scalability_report(message  => "Scalability message",
                           rslt_obj => $@);

=head1 DESCRIPTION

Provides the common routines and functionality for scalability testing.
To run tests on many clients the constructor is called with the plan
of execution followed by a call to the start method which will plan the
execution on all clients, create the virtual clients as necessary, create
intermediate test managers, distribute the load according to relative
performance of each physical client, monitor tests and test managers,
restart tests and test managers as needed, propagate failures and finally
remove any virtual interfaces created on the clients.

All tests have the following parameters available by default:

  INTERNAL    => always set to 1,
  TIMESTART   => time to start the test on sequence point,
                 used to synchronize all tests in sequence,
  VIRTUAL_IP  => the virtual ip of current instance
                 (only if virtual clients are used),
  NFS_VERSION => NFS version to use (only if nfs_version is specified)

=head1 STATUS

  Review

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/Scalability.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2006, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Scalability;

use strict;
use warnings;

use IO::Select;
use Carp qw(croak);
use Storable qw(dclone);
use Socket qw(inet_aton);
use nfs::Utils qw(range_ipaddrs);
use SPARSE::Util::Logger qw(logTrc logDbg logWarn);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF UNDEF OBJECT);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#-------------------------------------------------------------------------------
# Function prototypes
#-------------------------------------------------------------------------------
sub new($;%);
sub start($;);
sub status($;%);
sub clients_check($;);
sub client_info_get($;);
sub available_ipaddrs($;);
sub _plan($$;);
sub _monitor($;);
sub _next_sequence($$$;);
sub _job_compare($$$$$;);
sub _get_params($$$;$);
sub _propagate($;);
sub _start_tm($$;$);
sub _start_test($$;);
sub _get_test($;);
sub _plan_sequence($$;$);
sub _plan_level($$;);
sub _info_process($$$;);
sub _virtual_clients_run($$;);
sub _virtual_clients_create($;);
sub _virtual_interface_down($;);
sub _setral($;);
sub _process_params($$;$);
sub _get_pct_parameter($;);
sub _total_clients($;);
sub _check_errors($;);
sub _propagate_results($$;);
sub scalability_report($%;);
sub DESTROY;
sub testme();

#-------------------------------------------------------------------------------
# Package local variables
#-------------------------------------------------------------------------------
my $sudo     = '/usr/local/bin/sudo';
my $ifconfig = '/sbin/ifconfig';
my $MAX_NUM  = 50;
my $tm_total = 0;
my $MAX_RETRIES = 5;
my $report_file = "$main::TOP_LOGDIR_UNIX/report.log";
my $reports_dir = "$main::TOP_LOGDIR_UNIX/reports";

# WAIT_TIMESTART built in function
my $WAIT_TIMESTART = qq{\{
    use vars qw(\$TIMESTART);
    param('TIMESTART', '-default', undef);
    my \$stime = \$TIMESTART - time;
    sleep \$stime if \$stime > 0;
    logresult('PASS', 'WAIT_TIMESTART');
\}};

# List of valid built in functions
my %builtin_functions = (
    WAIT_TIMESTART => $WAIT_TIMESTART
);

#-------------------------------------------------------------------------------

=head2 new

  $obj = Scalability->new( [, plan        => \%plan ]
                           [, clients     => \@clients ]
                           [, nfs_version => $nfs_version_mix ]
                           [, vipaddr     => $vipaddr ]
                           [, vnumber     => $vnumber ]
                           [, background  => 0|1 ] );

Scalability class constructor which initializes scalable test environment.

The input plan is a hash with the following items:

  {
      parameters => global parameters (optional),
      setral     => global resource agitation levels (optional),
      sequence   => list of sequences
  }

The parameters option is a hash with the parameter name as the key
and the value can be either a scalar, array or hash. If it is a scalar,
the parameter has this value. For an array, the parameter can be any
value in the list (random distribution). For a hash, the values are
specified with its corresponding percentage. e.g.,

  parameters => {
      PARAM_1 => value,
      PARAM_2 => [ value_1, value_2, value_3 ],
      PARAM_3 => { value_1 => 'N%', value_2 => 'M%' }
  }

where N and M are numbers and N + M = 100.

The setral option is a hash with the resource name as the key and the
value is a hash of options for such resource. e.g.,

  setral => {
      cpu    => { consume_pct => 30 },
      malloc => { consume_pct => 5, random_N => 100 }
  }

The sequence option is an array of sequences to be executed
sequentially. Each item in the array is a hash with the following:

  {
      parameters => block parameters (optional),
      setral     => block resource agitation levels (optional),
      duration   => limit the runtime for this sequence (optional),
      list       => list of tests to run in parallel in this sequence
  }

Once the sequence has been running for the number of seconds
specified in the duration option, the next sequence starts terminating
the tests as necessary. Each test in the list is a hash with the following:

  {
      parameters    => local parameters (optional),
      script        => scriptspec of test to run,
      scripts       => list of scripts to run sequentially,
      nclients      => number of instances for this test to run (optional),
      continuous    => the test should run continuously or not (optional),
      ignore_errors => Ignore any errors from this test (optional),
      username      => run this test as the specified user (optional)
  }

The script option could be either a codeblock, thpl or .plan file.
A .plan file is a list of thpl's to run sequentially. Instead of using
the script option with a .plan file, the scripts option can be used instead
with a list of tests to run. These two options script and scripts are
mutually exclusive unless the script is pointing to a .plan file,
in this case all the tests in the .plan file will be appended to the
scripts option.
The script option can also be a built-in function, as of this moment
there is only one built-in function '<WAIT_TIMESTART>', which it logs
a pass result when the TIMESTART value is reached.

The nclients option gives the number of instances this test must be run.
This option can be either a fixed number of instances (N) or a percentage (N%)
of the number of available clients. It allocates the tests with a fixed number
of instances first, then all remaining clients are allocated to the tests
with percentages. If this option is not specified, it is evenly distributed
among all remaining clients.

If a test is continuous, the test will be restarted when it finishes.
If all tests are continuous, the sequence will never finish so
a duration must be provided in order to limit the runtime of the test.
The sequence is done when either the duration of the sequence has
expired or when all non-continuous tests have finished. Default is 0.

The following named arguments are accepted:

=over

=item Arguments

=over

=item plan

(Optional) : Reference to a hash having the plan of which tests to run,
their parameters and sequence information. See above for detail information
about this hash.

=item clients

(Optional) : Reference to a list hashes of real clients. Each hash must have
the key 'client'.

=item nfs_version

(Optional) : NFS version mix to use. If values (2, 3 or 4) are specified
the NFS version always be that version. Value 'all' gives an even
distribution of all three versions. Percentage distribution:
'NFSver:Percentage%;NFSver:Percentage%;...',
e.g., '2:30%;3:50%;4:20%' for a distribution of 30% for version 2,
50% for version 3 and 20% for version 4.

=item vipaddr

(Optional) : Initial ip address to use for virtual clients.

=item vnumber

(Optional) : Number of virtual clients to use.

=item background

(Optional) : Run the test in the background. The state of the test can
be checked by calling status(). Defaults to 0.

=back

=item Returns

Scalability object.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------------------
sub new($;%) {
    my $class = shift;
    my (%opts) = validate (
        @_,
        {
            plan        => { type => HASHREF,      optional => 1 },
            clients     => { type => ARRAYREF,     optional => 1 },
            nfs_version => { type => SCALAR|UNDEF, optional => 1 },
            vipaddr     => { type => SCALAR|UNDEF, optional => 1 },
            vnumber     => { type => SCALAR|UNDEF, optional => 1 },
            background  => { type => SCALAR,        default => 0 },
            # The following parameters are for internal use only
            sid         => { type => SCALAR, default => 0 },
            tskew       => { type => SCALAR, default => 0 },
            leaf        => { type => SCALAR, default => 0 },
            slave       => { type => SCALAR, default => 0 },
            bg          => { type => SCALAR, default => 0 },
        }
    );
    my ($count, $client);

    # Bless this object
    my $self = bless { %opts }, $class;

    if ( $self->{slave} ) {
        # Get total number of clients
        $count = 0;
        foreach $client ( @{$self->{clients}} ) {
            $count += $client->{nclients};
        }
        $self->{nclients} = $count;
    }

    # Return this object
    return $self;
}

#-------------------------------------------------------------------------------

=head2 start

  $obj->start();

This function gets a list of available ip addresses to used for virtual
clients, then checks all physical clients for availability, CPU speed, and
memory size to determine the potential load that the real client can handle.
It plans the distribution of clients and tests and
propagates specific parts of the plan to each intermediate test manager
where the leaf test manager starts the tests on either the physical or
virtual clients.

=over

=item Returns

Nothing.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------------------
sub start($;) {
    my $self = shift;
    my (%opts, $cmd);
    my $virtual_ips = [];

    # First instance creates the directory for failure reports
    if ( ! $self->{slave} ) {
        eval { qx(mkdir -p $reports_dir) };
        croak "Could not create reports directory: $@" if $@;
    }

    if ( $self->{background} ) {
        # Load library for select
        require IO::Select;

        # Pass all options to the main test manager
        %opts = (
            plan        => $self->{plan},
            clients     => $self->{clients},
            nfs_version => $self->{nfs_version},
            vipaddr     => $self->{vipaddr},
            vnumber     => $self->{vnumber},
            background  => 0,
            bg          => 1,
        );

        # Start the test in the background
        $cmd = qq{\{
            use Scalability;
            my \%opts = \@_;
            my \$obj = Scalability->new(\%opts);
            \$obj->start();
        \}};

        # Initialize transaction id
        $self->{xid} = 0;

        # Start the subtest on physical client
        $self->{tm} = eval { Tharn::subtest($cmd, '-runid', 'tm', '-bg',
                                            '-ipc', '-onexit', sub{},
                                            '--', %opts) };

        # Return immediately
        return;
    }

    if ( $self->{vipaddr} && $self->{vnumber} ) {
        # Get all available ip addresses to use as virtual addresses
        $virtual_ips = range_ipaddrs(ipaddr => $self->{vipaddr},
                                     number => $self->{vnumber});
        croak "No physical clients are available" if ! @{$self->{clients}};
        croak "No virtual clients are available" if ! @$virtual_ips;
    }

    # Only do the planing if this is the first instance (not a slave)
    if ( ! $self->{slave} ) {
        # Get information from all clients
        $self->clients_check();
        croak "No physical clients are available" if ! @{$self->{clients}};

        # Plan test
        $self->_plan($virtual_ips);
    }

    if ( ! $self->{leaf} ) {
        # This is not a leaf, so propagate to next branch
        $self->_propagate();

        # Monitor tests and test managers
        $self->_monitor();
    }

    return;
}

#-------------------------------------------------------------------------------

=head2 status

  $obj->status( [ stop => 0|1 ]
                [, timeout => $timeout ] );

Get the status of the currently running test. This function is used only
when the test is run using the background option.

The following named arguments are accepted:

=over

=item Arguments

=over

=item stop

(Optional) : Send a STOP signal to the main test manager.
Defaults to 0.

=item timeout

(Optional) : Timeout value waiting for response from the main test manager.
Defaults to 30.

=back

=item Returns

A reference to a hash containing the current state of the test.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------------------
sub status($;%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            stop    => { type => SCALAR, default => 0 },
            timeout => { type => SCALAR, default => 30 },
        }
    );
    my ($xid, $read_set, $read, $out, $sid, $phase, $cont, $tstamp, $ret);
    my $timeout = $opts{timeout};
    my $status  = $opts{stop} ? 'STOP' : 'STATUS';

    if ( exists($self->{tm}) && $self->{tm}->is_alive() ) {
        # Send message to main test manager
        $xid = $self->{xid}++;
        eval { $self->{tm}->ipc->write_message("$status--$xid") };

        # Add main test manager to list for select
        $read_set = IO::Select->new();
        $read_set->add($self->{tm}->ipc);

        do {
            # Wait for message from main test manager
            ($read, undef, undef) =
              IO::Select->select($read_set, undef, undef, $timeout);

            if ( defined($read) ) {
                # Make sure main test manager is still alive
                return $ret if ! $self->{tm}->is_alive();

                # Read the message from the main test manager
                $out = eval { $self->{tm}->ipc->read_message() };
                if ( $out ) {
                    # Process status from main test manager
                    ($sid, $phase, $cont, $xid, $tstamp) = split(/--/, $out);

                    # Return status
                    $ret = { sid        => $sid,
                             xid        => $xid,
                             seq_done   => $phase,
                             continuous => $cont,
                             done       => 0,
                             tstamp     => $tstamp };
                } else {
                    # Wait for one second when it
                    # is a socket error then retry
                    sleep 1;
                }
            }

            # Change timeout to 200ms for retry
            $timeout = 0.2;
        } while ( defined($read) );
    } else {
        # Test manager is not running
        $sid = @{$self->{plan}{sequence}};
        $ret = { sid        => $sid,
                 xid        => 0,
                 seq_done   => 1,
                 continuous => 0,
                 done       => 1,
                 tstamp     => time };
    }

    return $ret;
}

#-------------------------------------------------------------------------------

=head2 clients_check

  $obj->clients_check();

Check the performance for every client. It gets the cpu speed and available
memory to determine the potential load that the real client can handle.
It also gets the clock skew for each client.

=over

=item Returns

Nothing.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------------------
sub clients_check($;) {
    my $self = shift;
    my (%sobjs, $sobj, $client, $cmd, $runid, $tdelta, $client_name);
    my $index = 0;
    my $count = 0;
    my $rindex = 1;

    # Start all information subtest in parallel up to
    # a maximum of MAX_NUM at any given time
    while ( $index < @{$self->{clients}} ) {
        $tdelta = time();
        if ( $count < $MAX_NUM ) {
            # Start subtest for this client
            $client = $self->{clients}[$index]{client};

            # Script to run on each physical client to get the
            # client information about the cpu speed and available memory
            $cmd = qq{\{
                use Scalability;
                my \$obj = Scalability->new(slave=>1, leaf=>1);
                my \$out = \$obj->client_info_get();
                my \$cpuinfo = \$out->\{cpuinfo\};
                my \@cpu = map \{ \$cpuinfo->\{\$_\}\{"core id"\},
                                  \$cpuinfo->\{\$_\}\{bogomips\} \}
                                  keys %\{\$cpuinfo\};
                my \$res = join('--', \$out->\{meminfo\}\{MemTotal\}, \@cpu);
                logresult('PASS', \$res);
            \}};

            # Start the subtest on physical client
            $client_name = Tharn::host($client)->id();
            $runid = sprintf('%03d_info_%s', $rindex++, $client_name);
            $sobj = eval { Tharn::subtest($cmd, '-runid', $runid, '-bg',
                                          '-client', $client,
                                          '-onexit', sub{}) };
            if ( $@ ) {
                # Error using this client
                # log a warning and remove it from the list
                Tharn::logresult('WARN',
                                 "Client <$client_name> will not be used:", $@);
                splice(@{$self->{clients}}, $index, 1);
            } else {
                $sobjs{$index} = $sobj;
                # To keep track of all subtests
                $index++;
                # To keep track on the number of
                # subtest running concurrently
                $count++;
            }
        } else {
            # Reach a maximum of MAX_NUM
            # subtests running concurrently
            sleep(1);
        }

        # Process subtests as soon as possible to get a better
        # measurement of the skew on the clocks
        $tdelta = time() - $tdelta;
        $count -= $self->_info_process(\%sobjs, $tdelta);
    }

    # Done starting subtests on all the clients
    # just process the last batch
    while ( %sobjs ) {
        $tdelta = 1;
        sleep($tdelta);
        $self->_info_process(\%sobjs, $tdelta);
    }

    # Remove dead clients
    $index = 0;
    while ( $index < @{$self->{clients}} ) {
        if ( ! defined($self->{clients}[$index]) ) {
            splice(@{$self->{clients}}, $index, 1);
        } else {
            ++$index;
        }
    }

    return;
}

#-------------------------------------------------------------------------------

=head2 client_info_get

  $info = $obj->client_info_get();

Get cpu and memory information for local client.

=over

=item Returns

A reference to a hash of the client information.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------------------
sub client_info_get($;) {
    my $self = shift;
    my (%ret, $cpu, $out);

    # Get memory information
    open(FH, "</proc/meminfo") or croak "Could not open meminfo: $!";
    $out = join('', <FH>);
    close(FH);
    $ret{meminfo} = { $out =~ /(\w+):\s+(\S+)/g };

    # Get processors information
    open(FH, "</proc/cpuinfo") or croak "Could not open cpuinfo: $!";
    $out = join('', <FH>);
    close(FH);
    $out =~ s/\s+:/:/g;

    # Split by processor
    $out =~ s/\n\n/\n---/g;
    foreach $out ( split /---/, $out ) {
        # Skip if processor key is not found
        next if $out !~ /^processor/;
        my %cpuinfo = ( $out =~ /(.*):\s+(.*)/g );

        # Create key as the processor number
        # while deleting the processor key
        $cpu = $cpuinfo{processor};
        delete $cpuinfo{processor};
        $ret{cpuinfo}{$cpu} = \%cpuinfo;
    }

    return \%ret;
}

#-------------------------------------------------------------------------------

=head2 available_ipaddrs

  $virtual_ipaddrs = $obj->available_ipaddrs();

This is a wrapper for nfs::Utils::available_ipaddrs since it is required
to be run as root (icmp protocol). It starts the check at the ip address
given in the constructor (vipaddr) and returns until (vnumber) of free
ip addresses is found. If the number of virtual clients is large, then
the physical clients will be used to parallelize the search for free
ip addresses.

=over

=item Returns

A reference to an array of available virtual ip addresses.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------------------
sub available_ipaddrs($;) {
    my $self = shift;
    my (@sobjs, @vips, %sopts, $sobj, $client, $virtual_ips, $cmd);
    my ($count, $cnt, $index, $number, $runid, $ip, $n, $client_name);
    my $vipaddr = $self->{vipaddr};
    my $vnumber = $self->{vnumber};
    my $clients = $self->{clients};

    # Code to run at each client to get the available
    # ip addresses for the given range
    $cmd = qq{\{
        use nfs::Utils qw(available_ipaddrs);
        my (\$vipaddr, \$count) = \@_;
        my \$virtual_ips = available_ipaddrs(ipaddr  => \$vipaddr,
                                             max_num => \$count);
        logcomment("Found " . scalar(\@\$virtual_ips) . " available ip's " .
                   "for range \$vipaddr:\$count");
        Subtest::ipc->write_obj(\$virtual_ips);
    \}};

    $index = 0;
    while ( @vips < $vnumber && @$clients ) {
        # Get the number of ip's left to check
        # Add a 10% for those ip's that could be already allocated
        $number = int(1.1*($vnumber-@vips));

        # Get the number of ip's to check for each client (minimum 200)
        $cnt = int($number / @$clients) + 1;
        $cnt = 200 if $cnt < 200;

        # Convert the ip address to a number
        $ip = 0;
        ( $ip = ($ip<<8)|$_ ) foreach split(/\./, $vipaddr);

        # Check for available ip addresses in parallel so
        # start a subtest on each client up to the maximum allowed
        while ( $number > 0 && @sobjs < $MAX_NUM && @$clients ) {
            # Check for available ip's on this client
            $client = $clients->[$index]{client};

            # Get number of ip's to check for this client
            # with a minimum of 10 ip's
            $count = $cnt > $number ? $number : $cnt;
            $count = 10 if $count < 10;

            # Get client and runid (log file name) for this subtest
            %sopts = ( '-client' => $client );
            $client_name = Tharn::host($client)->id();
            $runid = sprintf('%03d_getvips_%s', $index+1, $client_name);

            if ( $vnumber < 200 ) {
                # For a small number of virtual
                # clients use only the localhost
                $count = $vnumber;
                $number = $count;
                $cmd =~ s/max_num/number/;
                %sopts = ();
            }

            # Start subtest as root since the library function
            # uses icmp ping to determined ip availability
            # Use ipc sockets to transfer the list of available ip's
            $sobj = eval { Tharn::subtest($cmd, '-runid', $runid, '-bg',
                                          '-ipc', '-username', 'root',
                                          %sopts, '-onexit', sub{},
                                          '--', $vipaddr, $count) };
            if ( $@ ) {
                # Error using this client
                # log a warning and remove it from the list
                Tharn::logresult('WARN',
                                 "Client <$client_name> will not be used:", $@);
                splice(@$clients, $index, 1);
                $index = 0 if ($index >= @$clients);
                next;
            } else {
                push @sobjs, $sobj;
            }

            # Get last octet
            $n = $ip & 0xFF;

            # Add 2 more for every transition to
            # account for last octet 0 and 255
            $ip += $count + 2 * int(($count + $n) / 255);

            # Make sure last octet is 1 (on next second to last octet)
            # if last octet is either 0 or 255
            $n = ($n + 1) & 0xFF;

            # Set the initial ip address for next client
            $ip += $n < 2 ? 2 - $n : 0;

            # Convert ip number back to an ip address for next iteration
            $vipaddr = join('.', map {($ip>>$_)&0xFF} (24, 16, 8, 0));

            # Decrement the number of ip addresses needed for next iteration
            $number -= $count;

            # Use next client for next iteration and if this is
            # the last, cycle back to the beginning of the list
            $index++;
            $index = 0 if ($index >= @$clients);
        }

        # Get the list of virtual ip's from each client
        foreach $sobj ( @sobjs ) {
            next if ! $sobj->is_alive();
            $virtual_ips = $sobj->ipc->read_obj();
            push @vips, @$virtual_ips;
        }

        # Wait for all the subtests to finish
        if ( @sobjs ) {
            Subtest::wait_finish(subtests=>\@sobjs);
            @sobjs = ();
        }
    }

    # Get exactly the number of virtual ip's requested
    splice @vips, $vnumber if @vips > $vnumber;

    return \@vips;
}

#-------------------------------------------------------------------------------
# Internal function
# This function plans how the workload is distributed among all
# the physical clients. First, it calculates the performance factors
# for each of the clients. Then the number of levels are calculated
# based on the number of physical clients and the maximum number
# of subtest processes a client can handle. Finally, the clients
# are arranged in a hierarchy and the virtual clients are distributed
# among the leaf clients according to the performance factors.
#-------------------------------------------------------------------------------
sub _plan($$;) {
    my ($self, $virtual_ips) = @_;
    my (@clients, @list, $levels, $item, $ref, $index);
    my ($client, $sum, $count, $seq, $nfs_ver, %nfs, $max_count);
    my ($n, $m, $k, $line, $total_clients, $need_clients, $sn_clients);
    my $nvips = @$virtual_ips;

    # Check correct format of nfs version mix
    if ( defined($self->{nfs_version}) ) {
        $nfs_ver = $self->{nfs_version};
        if ( $nfs_ver =~ /%/ ) {
            # Format is given in terms of percentages
            # Convert to a hash
            %nfs = map { split(/:/, $_) } split(/;/, $nfs_ver);

            # Check the sum of the percentages equals 100%
            $sum = 0;
            foreach $item ( values %nfs ) {
                if ( ! ($item =~ s/%$//) ) {
                    # This version is not given in percentage
                    croak "Error parsing NFS version: <$nfs_ver>";
                }
                $sum += $item;
            }
            croak "NFS version distribution <$sum%> must be 100%"
              if $sum != 100;
        } elsif ( $nfs_ver !~ /^([234]|all|not2|not3|not4)$/ ) {
            # If not in percentages the value specified is invalid
            croak "NFS version not supported: <$nfs_ver>";
        } elsif ( $nfs_ver eq 'all' ) {
            # Special case 'all' 33.33% for all versions
            %nfs = map { $_ => 100/3 } (2, 3, 4);
        } elsif ( $nfs_ver eq 'not2' ) {
            # Special case 'not2' 50% for version 3 and 4
            %nfs = map { $_ => 50 } (3, 4);
        } elsif ( $nfs_ver eq 'not3' ) {
            # Special case 'not3' 50% for version 2 and 4
            %nfs = map { $_ => 50 } (2, 4);
        } elsif ( $nfs_ver eq 'not4' ) {
            # Special case 'not4' 50% for version 2 and 3
            %nfs = map { $_ => 50 } (2, 3);
        } else {
            # Only one version was specified -- 100%
            %nfs = ( $nfs_ver => 100 );
        }

        # Save the NFS version mix in the plan
        $self->{plan}{nfs_params} = \%nfs;
    }

    # Calculate performance factors for each client
    $self->{factor_sum} = 0;
    foreach $client ( @{$self->{clients}} ) {
        # Information for client is not available
        next if ( exists($client->{factor}) && $client->{factor} == 0 );

        # Calculate performance factor for client
        $client->{factor} = $client->{mips} * $client->{mem};
        $self->{factor_sum} += $client->{factor};
    }

    # Calculate the number of levels necessary
    # for given number of clients
    $count = @{$self->{clients}};
    $m = $MAX_NUM;

    # Geometric series solving for N (x + x^2 + ... + x^N)
    # where x is the maximum number of clients directly
    # connected under a client (MAX_NUM)
    $n = (log($count*($m-1) + $m) / log($m)) - 1;

    # Round up the result
    $levels = int($n > int($n) ? $n+1 : $n);

    # Start list of levels with a minimum
    # of 2 clients per client on each level
    @list = map { 2**$_ } (1..$levels-1);

    # Calculate the number of clients on each level
    if ( @list ) {
        # Redo block
        {
            $sum = 0;
            # Get the total number of clients
            # needed for intermediate levels
            map { $sum += $_ } @list;

            # Total number of leaf clients
            $n = $count - $sum;

            # Number of parent branches
            $m = $list[-1];
            if ( $n > $MAX_NUM*$m ) {
                # The number of parent branches can not support the current
                # number of clients so add a branch and propagate the change
                # by adding additional branches to the tree as needed
                $index = @list;
                do {
                    ++$list[--$index];
                    $n = $list[$index];
                    $m = $list[$index-1];
                } while ( $n > $MAX_NUM*$m );

                # Re-check the new changes
                redo;
            }
        }
    }

    # Arrange clients for each level starting with the leaf level
    foreach $item ( @list ) {
        # Get all clients from this level
        my @clnts = splice(@{$self->{clients}}, 0, $item);

        # Put them in front of the list
        unshift @clients, \@clnts;
    }
    # Put the leaf clients in front of the list
    unshift @clients, $self->{clients};

    # Calculate total performance for leaf elements only to get the
    # relative performance for the distribution of the virtual clients
    $sum = 0;
    foreach $client ( @{$clients[0]} ) {
        $sum += $client->{factor};
    }

    # Make sure at least one client has a non-zero performance factor
    croak "Could not get any performance info for the clients" if ! $sum;

    # Add the proper number of virtual clients to each leaf real client
    $n = 0;
    $max_count = 0;
    $total_clients = 0;
    foreach $client ( @{$clients[0]} ) {
        # Make performance factor relative
        $client->{factor} /= $sum;
        # Calculate total number of virtual clients for this client
        $m = $client->{factor} * $nvips;
        $count = int($m);

        # Accumulate fraction
        $n += $m - $count;
        if ( $n > 0.9 ) {
            # Fraction has accumulated one more client
            # so allocate it in this physical client
            $n -= 1;
            $count++;
        }
        # Make sure to get only exactly what is remaining
        $count = @$virtual_ips if $count > @$virtual_ips;
        $max_count = $count if $count > $max_count;
        $client->{vclients} = [ splice(@$virtual_ips, 0, $count) ];
        $total_clients += $count ? $count : 1;
    }

    # Move clients to their corresponding levels.
    # Up to this point, the clients are stored in an array where
    # each of the elements in the array are the individual levels,
    # now create a hierarchical structure on the hash of all levels
    foreach $index ( 0..@clients-2 ) {
        $item = $clients[$index];   # Current level
        $ref  = $clients[$index+1]; # Parent level
        $n = @$ref;                 # Number of elements in parent level
        $m = int(@$item / $n);      # Number of elements for each parent
                                    #   element to have
        $k = @$item % $n;           # Number of unallocated elements
        foreach my $itm ( @$ref ) {
            # Each parent elements has the same number of elements and the
            # unallocated elements are allocated to the first parent elements
            $count = $m + ($k-->0?1:0);
            $itm->{clients} = [ splice(@$item, 0, $count) ];
        }
    }

    if ( defined($ref) ) {
        # Replace input clients by the hierarchical clients
        $self->{clients} = $ref;
    }

    # Calculate total number of clients at each branch
    $count = _total_clients($self->{clients});
    # including the main branch
    $self->{nclients} = $count;

    # Add the number of clients to the total number
    # of test managers calculated at _total_clients
    $tm_total += @{$self->{clients}};

    # Initial startup time for test managers
    $self->{plan}{tm_dtime} = 2 * $tm_total;

    # Startup time for tests
    $self->{plan}{dtime} = 0.2 * $max_count;

    # Sanity check for the test plan: process '.plan' entries,
    # verify correct percentages of plan
    $need_clients = 0;
    foreach $seq ( @{$self->{plan}{sequence}} ) {
        $m = 0;
        $sum = 0;
        $index = 0;
        $sn_clients = 0;

        # Check every test
        foreach $item ( @{$seq->{list}} ) {
            if ( defined($item->{script}) && $item->{script} =~ /.plan$/ ) {
                # This is a .plan file
                open(FH, "<$item->{script}")
                  or croak "Could not open plan file: $!";
                while ( <FH> ) {
                    $line = $_;
                    if ( $line =~ s/(\S+\.thpl).*/$1/ ) {
                        # Add only the thpl to the list
                        push @{$item->{scripts}}, $line;
                    }
                }
                close(FH);
            }

            if ( defined($item->{scripts}) ) {
                # Initialize the starting test index for sequential tests
                $item->{sindex} = 0;
            }

            if ( !defined($item->{nclients}) || $item->{nclients} =~ /%/ ) {
                # This is a percentage test:
                # implicit (no nclients) or explicit (%)
                if ( defined($item->{nclients}) ) {
                    # Explicit percentage
                    $n = $item->{nclients};
                    $n =~ s/%//;
                    $sum += $n;
                    $m++;  # Number of explicit entries
                }
                push @{$seq->{pct_list}}, $index;
                $sn_clients++;
            } else {
                $sn_clients += $item->{nclients};
            }
            $index++;
        }
        croak "Percentages on plan exceed 100 ($sum)" if $sum > 100;

        # Minimum number of clients needed
        $need_clients = $sn_clients if $sn_clients > $need_clients;

        if ( defined($seq->{pct_list}) ) {
            # Distribute evenly the implicit percentage entries
            $seq->{pct_sum} = $sum;

            # Unallocated percentage
            $sum = 100 - $sum;

            # Calculate the number of implicit entries
            $count = @{$seq->{pct_list}} - $m;
            if ( $count > 0 ) {
                # Even distribution for implicit entries
                $n = $sum / $count;
                foreach $index ( @{$seq->{pct_list}} ) {
                    $item = $seq->{list}[$index];
                    if ( ! defined($item->{nclients}) ) {
                        # Make it explicit
                        $item->{nclients} = "$n%";
                    }
                }
            }
        }
    }

    # Check if there is enough clients to run this plan
    croak "\nThere is not enough clients to run this plan.\n" .
          "A minimum of $need_clients clients is needed\n"
      if $need_clients > $total_clients;

    return;
}

#-------------------------------------------------------------------------------
# Internal function
# This function monitors all its children: being tests or test
# managers. It relays all messages from/to parent test manager to/from
# slave test managers.
#-------------------------------------------------------------------------------
sub _monitor($;) {
    my $self = shift;
    my (@tests, @monitors, $client, $item, $status, $dtime, $tdef, $seq);
    my ($phase, $cont_test, $cont_phase, $read_set, $read, $sock, $done);
    my (%filenos, $sid_r, $phase_r, $cont_r, $tstamp_r, $done_time, $fc);
    my ($save_status, $out, $msg, $msg_expire, $xid, $xid_n);
    my $sid = $self->{sid};
    my $next_phase = 0;
    my $sleep_time = 10;
    my $log_on_main = time();
    my $killmsg = ['PASS', 'Test killed: received STOP message'];
    my $last_results;

    # Renice process so test manager has
    # higher priority than other processes
    qx($sudo renice -10 $$);

    if ( ! $self->{slave} ) {
        # Start resource agitation only if
        # this is the main test manager
        $fc = $self->_setral();

        # Log step on main log
        Tharn::logcomment("Start monitoring tests");
    }

    # Create a list of tests and test managers under this test manager
    foreach $client ( @{$self->{clients}} ) {
        push @tests, $client->{test} if $client->{test};
        push @tests, @{$client->{vtests}} if $client->{vtests};
        push @monitors, $client->{tm} if $client->{tm};
    }

    # Connect to parent test manager
    if ( $self->{slave} || $self->{bg} ) {
        $self->{ipc} = Subtest::ipc();
    }

    # Get the file numbers of sockets for the slave test managers
    foreach $item ( @monitors ) {
        # This hash is to make it easier to translate from the
        # socket ready to be read to the correct test in the list
        $filenos{$item->{sobj}->ipc->fileno} = $item;
    }

    # The following label is needed to execute the following code
    # up to the main monitor loop after every change in sequence
    LABEL:

    # Get current sequence id
    $seq = $self->{plan}{sequence}[$sid];

    # Get end time for this sequence
    undef $done_time;  # Clear duration flag for this sequence
    if ( defined($seq->{duration}) && @tests ) {
        $dtime = time() + $self->{plan}{dtime} if ! defined($dtime);
        $done_time = $seq->{duration} + $dtime + $self->{tskew} - $sleep_time;
    }

    # Find if any test in sequence is not continuous
    # This flag is set to 1 if all tests in sequence are continuous
    # and set to 0 if at least one test in sequence must finish
    $cont_phase = 1;
    foreach $item ( @{$seq->{list}} ) {
        if ( ! $item->{continuous} ) {
            # This test is not continuous
            $cont_phase = 0;
            last;
        }
    }

    # Find if any test in client is not continuous
    # This flag is set to 1 if all tests currently running are continuous
    $cont_test = @tests ? 1 : 0;
    foreach $item ( @tests ) {
        $tdef = $seq->{list}[$item->{testid}];
        if ( $tdef->{continuous} == 0 ) {
            # This test is not continuous
            $cont_test = 0;
            last;
        }
    }

    # This flag is set to 1 when the monitor should exit the main loop
    # after waiting to receive an acknowledge message from its slave test
    # manager and to send and acknowledge message to its parent test manager
    $done = 0;

    # This flag is the sequence status. It is set to 1 when the
    # sequence is done and 0 if it is still running
    $phase = 0;

    # This is the message that is sent to slave test managers
    # values are: STATUS, NEXT and STOP
    $status = 'STATUS';

    # Monitor loop
    while ( 1 ) {
        # Check the status every 'sleep_time' seconds
        sleep $sleep_time if ( ! $next_phase && ! $phase );

        # Create object for list of slave
        # test managers to test for messages
        $read_set = IO::Select->new();

        # Check if sequence is done in order to propagate the
        # next message either a NEXT or a STOP if all sequences are done
        # Only the main test manager can make this decision
        if ( ! $self->{slave} && $phase && $status !~ /STOP/ ) {
            # Current sequence is complete so
            # increment current sequence id
            $sid++;
            $self->{sid} = $sid;
            if ( $sid == @{$self->{plan}{sequence}} ) {
                # Done with all sequences: send STOP message
                Tharn::logcomment('Done with all sequences');
                $status = 'STOP';

                # Decrement sequence id in order to
                # log it correctly on the main test manager
                $sid--;
                $self->{sid} = $sid;
            } else {
                # Go to next sequence: send NEXT message
                Tharn::logcomment('Request start of next sequence');

                # Calculate start time for next sequence
                $dtime = time() + $self->{plan}{dtime};
                $status = "NEXT--$sid--$dtime";

                # Set resource agitation for next sequence
                $fc = ( $self->_setral() || $fc );
            }
            if ( ! @monitors ) {
                # This test manager does not have any slave test
                # managers so just kill all tests still alive
                if ( $status eq 'STOP' ) {
                    foreach $item ( @tests ) {
                        if ( $item->{state} && $item->{sobj}->is_alive() ) {
                            $item->{sobj}->kill(message=>$killmsg);
                        }
                    }
                    last;
                }

                # and change to next sequence
                $self->_next_sequence(\@tests, $dtime);
                goto LABEL;
            }
        }

        # Expire time for test managers which do not respond
        $msg_expire = time() + 3*$sleep_time;

        # Send message to all test managers
        foreach $item ( @monitors ) {
            # Reset state of slave test manager back to running
            # in order to propagate the NEXT or STOP messages
            if ( $status =~ /(STOP|NEXT)/
              || ( $item->{state} == 2 && $item->{tstamp} > $msg_expire ) ) {
                $item->{state} = 1;
            }
            # Skip dead test managers, this is just a safety precaution
            # since a test manager should be restarted when dead
            next if ( $item->{state} == 0 );

            # Check process status of test manager
            if ( $item->{sobj}->is_alive() ) {
                # This test manager is alive
                if ( $item->{state} == 1 ) {
                    if ( $status =~ /STATUS/ && $item->{nextseq} > 3 ) {
                        # Need to change to current sequence;
                        # save the current message with the previous
                        # NEXT message to this client
                        $save_status = $status;
                        $status = $item->{nextmsg};
                    }
                    # This test manager is running and ready to
                    # to receive a message
                    $xid = $item->{xid}++;
                    logDbg("Sending <$status--$xid> to slave test manager " .
                           "[$item->{runid}]");

                    # Save next message on item for retransmission
                    if ( $status =~ /NEXT/ ) {
                        $item->{nextmsg} = $status;
                        $item->{nextseq} = 0;
                    }

                    # Send message to test manager
                    eval { $item->{sobj}->ipc->write_message("$status--$xid") };

                    # Restore current message for next test manager
                    if ( defined($save_status) ) {
                        $status = $save_status;
                        undef $save_status;
                    }

                    # Add socket to list to be used with select
                    $read_set->add($item->{sobj}->ipc);

                    # Change the state for this test manager to waiting for
                    # response so a second message is not retransmitted
                    $item->{state}  = 2;
                    $item->{tstamp} = time();
                    $item->{nmsgs}++;
                } elsif ( $item->{state} == 2 ) {
                    # This test manager has not responded,
                    # keep adding it to the list for select
                    $read_set->add($item->{sobj}->ipc);
                }
            } else {
                # This test manager is somehow dead. Change the state
                # to dead in case it is unable to be restarted
                $item->{state} = 0;
                if ( ! $done )  {
                    # This test manager should be alive so:
                    # Check for errors
                    _check_errors($item->{sobj});

                    if ( $item->{restart}++ < $MAX_RETRIES ) {
                        # Restart test manager
                        logDbg("Restarting test manager");
                        delete $filenos{$item->{sobj}->ipc->fileno};
                        $self->_start_tm($self->{clients}[$item->{clientid}]);
                        $filenos{$item->{sobj}->ipc->fileno} = $item;
                    }
                }
            }
        }

        # The main test manager must re-initialize for next sequence
        if ( ! $self->{slave} && $status =~ /NEXT/ ) {
            goto LABEL;
        }

        # Wait for all slave test managers
        # to reply at the STOP message
        $done = 1 if $status eq 'STOP';

        # Change to default request
        $status = $done ? 'STOP' : 'STATUS';

        # Check if any test has finished
        foreach $item ( @tests ) {
            # Skip already marked dead tests
            next if ( $item->{state} == 0 || $item->{sobj}->is_alive() );

            # The test has finished, so change the state to dead
            $item->{state} = 0;

            # Propagate test results to parent
            $self->_propagate_results($item);

            # Check if this test is continuous or if this is a .plan
            $tdef = $seq->{list}[$item->{testid}];
            if ( ( $tdef->{continuous} || defined($tdef->{sindex}) )
              && ! $done ) {
                # This test should be restarted
                logDbg("Restarting testid: <$item->{testid}>");
                $self->_start_test($item);
            }
        }

        # Check if sequence has finished for the tests
        if ( $done ) {
            # All tests should have been killed by now
            $phase = 1;
        } elsif ( defined($done_time) && time > $done_time ) {
            # Duration has expired for this sequence
            logTrc("Duration has expired for sequence $sid");
            $phase = 1;
        } elsif ( $cont_test ) {
            # All test are continuous, so sequence is never done
            $phase = 0;
        } else {
            # Check if all tests which are not continuous are done
            $phase = 1;
            foreach $item ( @tests ) {
                $tdef = $seq->{list}[$item->{testid}];
                if ( !$tdef->{continuous} && $item->{state} ) {
                    # Sequence point is not done yet at least
                    # one non-continuous test is still running
                    $phase = 0;
                    last;
                }
            }
        }

        # Check if sequence has finished for the test managers
        if ( $read_set->count ) {
            READ_FROM_SLAVES:
            # Wait for one second for any connection
            # from the slave test managers
            ($read, undef, undef) =
              IO::Select->select($read_set, undef, undef, 1.0);

            if ( defined($read) ) {
                # At least on slave test manager is ready
                foreach $sock ( @$read ) {
                    # Translate ready socket to test structure
                    $item = $filenos{$sock->fileno};

                    # Make sure test manager is still alive
                    if ( ! $item->{sobj}->is_alive() ) {
                        $read_set->remove($sock);
                        next;
                    }

                    # Ignore error, let _check_errors handle it
                    $out = eval { $sock->read_message() };
                    if ( $@ ) {
                        sleep 1;
                        next;
                    }
                    logDbg("Message from slave test manager " .
                           "[$item->{runid}]: <$out>");

                    # The slave test manager has answered, change state
                    $item->{state}   = 1;
                    $item->{tstamp}  = time;
                    $item->{nmsgs}   = 0;
                    $item->{message} = $out;

                    # Process status from slave test manager
                    ($sid_r, $phase_r, $cont_r, $tstamp_r) = split(/--/, $out);
                    if ( $sid_r != $sid ) {
                        # Sequence point of slave test manager is not current
                        # Count the number of times this has happened in order
                        # to retransmit the change of sequence
                        $item->{nextseq}++;
                    } elsif ( $phase_r || ($cont_r && !$cont_phase) ) {
                        # The sequence on the slave manager has
                        # finished or still not done but all tests under
                        # it are continuous -- mark it as this sequence
                        # being done
                        $item->{state} = 3;
                    }
                }

                # Keep on reading until there are no more messages
                goto READ_FROM_SLAVES;
            }
        }

        # Check if sequence is done
        foreach $item ( @monitors ) {
            if ( $item->{state} && $item->{state} != 3 ) {
                $phase = 0;
            }
        }

        # Log current status of sequence
        $msg = "Sequence point $sid status: " . ($phase ? 'done' : 'running');
        if ( ! $self->{slave} && ($phase || time() > $log_on_main + 300) ) {
            # Log status on main log every 5 mins
            Tharn::logcomment($msg);
            $log_on_main = time();
        } else {
            logDbg($msg);
        }

        if ( $done && $phase ) {
            # Slave test managers have reply
            # to STOP message so exit
            last;
        }

        # Get messages from parent
        $read_set = IO::Select->new();
        $read_set->add($self->{ipc});

        READ_FROM_PARENT:
        # Wait for one second for any message
        # from parent test manager
        ($read, undef, undef) =
          IO::Select->select($read_set, undef, undef, 1.0);

        if ( defined($read) || $next_phase ) {
            # There is a message or a change in
            # sequence has been requested previously
            if ( $next_phase ) {
                # A change in sequence has been requested previously
                # so send acknowledgment since due to the 'goto LABEL'
                # an acknowledgment was not sent previously
                $out = "STATUS--$xid_n";
            } else {
                # Read the message from the parent test manager
                # Ignore error, let _check_errors handle it
                $out = eval { $self->{ipc}->read_message() };
                next if $@;
            }
            logDbg("Message from parent test manager: <$out>");
            if ( $out =~ /STOP/ ) {
                # Exit test manager: notify all slave test managers
                $status = $out;
                $status =~ s/--\d+$//;

                # Kill all tests if still alive
                foreach $item ( @tests ) {
                    if ( $item->{state} && $item->{sobj}->is_alive() ) {
                        $item->{sobj}->kill(message=>$killmsg);
                    }
                }
            } elsif ( $out =~ /NEXT/ ) {
                # A change of sequence is requested
                (undef, $sid, $dtime) = split(/--/, $out);

                # Change to next sequence only if
                # current sequence is different
                if ( $sid != $self->{sid} ) {
                    logTrc("Change to next sequence <$sid>")
                      if ! $self->{leaf};

                    # Set new sequence id
                    $self->{sid} = $sid;

                    # Notify all slave test managers
                    $status = $out;
                    $status =~ s/--\d+$//;
                    if ( $self->{leaf} ) {
                        # Make sure next time around the
                        # acknowledgment is sent
                        $next_phase = 1;
                        ($xid_n) = ($out =~ /--(\d+)$/);

                        # Call function to change sequence
                        $self->_next_sequence(\@tests, $dtime);
                        goto LABEL;
                    } else {
                        # Initialize status for this sequence until
                        # the real status gets propagated from the leaf
                        $phase = 0;
                        $cont_test = 0;
                    }
                }
            }
            # Prepare status for this test manager: send the
            # sequence currently running, the sequence
            # status and all continuous tests flag
            ($xid) = ($out =~ /--(\d+)$/);
            $out = "$sid--$phase--$cont_test--$xid--" . time;
            logDbg("Sending status to parent test manager: <$out>");
            eval { $self->{ipc}->write_message($out) };
            $next_phase = 0;
        }

        # Keep on reading until there are no more messages
        goto READ_FROM_PARENT if defined($read);
    }

    if ( defined($fc) ) {
        # Disable resource agitation
        filer::filer_resource_agitation_set($fc, 'reset');
    }

    if ( @monitors ) {
        $out = @monitors;
        $msg = "Waiting for $out slave test managers to finish\n" .
               "This could take a few minutes while trying to " .
               "collect all the results";
        $self->{slave} ? logDbg($msg) : Tharn::logcomment($msg);
    }

    # Initialize the last result counter
    $last_results = time if !$last_results;

    while ( $item = shift @monitors ) {
        if ( $item->{sobj}->is_alive() ) {
            # If test manager is still alive, check how long we have
            # been waiting for one to finish
            if ( (time - $last_results) > 900 ) {
                # More than 15 minutes? Discard the test manager and try
                # the next
                next;
            } else {
                # Less than 15 minutes, put the test manager at the end
                # of the list and try the next
                push @monitors, $item;
                sleep 1;
            }
        } else {
            # Otherwise, test manager has completed, propagate the results
            $self->_propagate_results($item);
            $last_results = time;
        }
    }

    return;
}

#-------------------------------------------------------------------------------
# Internal function
# This function starts the tests in the next sequence.
#-------------------------------------------------------------------------------
sub _next_sequence($$$;) {
    my ($self, $tests, $dtime) = @_;
    my (@new, @found, $item, $new_item, $script, $count, $out, $client_name);
    my ($client, $index, $runid, $stime, $kill_test, $test);
    my ($ntests, $pct_tests, $multiplier, $killmsg, $m, $n);
    my $njob = $self->{plan}{sequence}[$self->{sid}];
    my $cjob = $self->{plan}{sequence}[$self->{sid}-1];

    logTrc("Starting next sequence <$self->{sid}>");

    # Get a list of new tests to run and calculate how
    # many clients are available for percentage runs
    $pct_tests = $self->{nclients};
    foreach $item ( @{$njob->{list}} ) {
        # Get list of new tests to run
        push @new, $item->{script};
        if ( $item->{nclients} !~ /%/ ) {
            # Get number of tests to run with percentage ratios
            $pct_tests -= $item->{nclients};
        }
    }
    $pct_tests = 0 if $pct_tests < 0;

    # Find which tests must be killed (the tests currently running
    # which are not in the next sequence). These include same tests
    # but with different set of parameters (jobs are different).
    $killmsg = ['PASS', 'Test killed: not in new sequence'];
    foreach $item ( @$tests ) {
        # Skip finished tests
        next unless $item->{state};

        # Get test name of current job
        $script = $cjob->{list}[$item->{testid}]{script};

        # Find all tests currently running which are in next sequence
        @found = grep { $script eq $_ } @new;

        $kill_test = 0;
        if ( ! @found ) {
            # No overlapping tests
            $kill_test = 1;
        } else {
            # Check tests for same parameters
            $index = 0;
            $kill_test = 1;
            foreach $new_item ( @{$njob->{list}} ) {
                if ( $script eq $new_item->{script} ) {
                    # Compare test definitions
                    $out = $self->_job_compare($self->{sid}-1, $item->{testid},
                                               $self->{sid}, $index);
                    if ( $out ) {
                        # Tests are the same, so increment the number
                        # of instances this test is currently running
                        $new_item->{count}++;

                        # Change the testid of the currently running
                        # job with the testid for next sequence
                        $item->{testid} = $index;

                        # Save current job object for future reference
                        push @{$new_item->{current_tests}}, $item;
                        $kill_test = 0;
                    }
                }
                $index++;
            }
        }
        if ( $kill_test ) {
            # Kill test not in new sequence
            $item->{sobj}->kill(message=>$killmsg);
            $item->{state} = 0;
        }
    }

    # Step down number of instances currently running
    $killmsg = ['PASS', 'Test killed: step down number of instances'];
    foreach $item ( @{$njob->{list}} ) {
        # Skip tests in new sequence which do not
        # have any instances currently running
        next unless $item->{count};

        # Get number of tests needed for next sequence
        $n = $item->{nclients};
        if ( $n =~ s/%// ) {
            $m = $n * $pct_tests / 100;
            # Round the number to the nearest integer
            # with probabilistic bias
            $n = int($m);
            $n += ($m - $n) > rand() ? 1 : 0;
        }

        if ( $item->{count} > $n ) {
            # Step down: calculate number of tests to kill
            $m = $item->{count} - $n;
            foreach ( 1..$m ) {
                # Select test to kill at random
                $index = int(rand(@{$item->{current_tests}}));
                $item->{current_tests}[$index]{sobj}->kill(message=>$killmsg);
                $item->{current_tests}[$index]{state} = 0;
                splice @{$item->{current_tests}}, $index, 1;
                $item->{count}--;
            }
        }

        # Remove reference to array of current tests
        delete $item->{current_tests};
    }

    # Remove dead tests from list
    $index = 0;
    while ( $index < @$tests ) {
        if ( $tests->[$index]{state} == 0 ) {
            splice @$tests, $index, 1;
            next;
        }
        $index++;
    }

    # Adjust the plan for next sequence

    # Number of tests currently running
    $count = @$tests;

    # Number of new tests that can be started
    $ntests = ($self->{nclients} - $count);

    if ( $count > 0 && $ntests > 0 ) {
        # Adjust the plan for tests with exact number of instances to run
        foreach $item ( @{$njob->{list}} ) {
            if ( $item->{nclients} !~ /%/ ) {
                # Number of instances this test is currently running
                $n = defined($item->{count}) ? $item->{count} : 0;

                # Calculate how many more instances are needed
                $item->{nclients} -= $n;

                # Adjust number of available spots
                $ntests -= $item->{nclients};
            }
        }

        # Adjust the plan for tests with percentages
        if ( $ntests > 0 && $pct_tests ) {
            # Calculate multiplier: the ratio between the
            # total number of instances over the number of
            # instances to be started (total - current)
            $multiplier = $pct_tests / $ntests;
            foreach $item ( @{$njob->{list}} ) {
                $n = $item->{nclients};
                if ( $n =~ s/%// ) {
                    # This is a percentage entry
                    # Calculate the percentage difference
                    # of already running instances of this test
                    $m = defined($item->{count})
                      ? (100*$item->{count} / $pct_tests) : 0;
                    # Calculate new percentage for this test
                    $n = ($n - $m) * $multiplier;
                    $n = 0 if $n < 0;
                    $item->{nclients} = "$n%";
                }
            }
        }
    }

    # Create list of clients with already running instances
    # Save it on the object so _virtual_clients_run has access to it
    $self->{current_tests} = { map { $_->{client} => 1 } @$tests };

    # Start all new tests for next sequence
    $index = 1;
    foreach $client ( @{$self->{clients}} ) {
        # Skip clients already running
        next if $self->{current_tests}{$client->{client}};

        # Start time for this client
        $stime = int($dtime + $client->{tskew});

        if ( ! exists($client->{clients})
          && ( ! exists($client->{vclients}) || ! @{$client->{vclients}} ) ) {
            # This is a leaf physical client
            # Plan this sequence for this client
            $self->_plan_sequence($client);

            # Run id for subtest
            $client_name = Tharn::host($client->{client})->id();
            $runid = sprintf('%03d_%03d_%s',
                             $self->{sid}+1, $index++, $client_name);

            # Save test information
            $test = { client => $client->{client},
                      runid  => $runid,
                      stime  => $stime };

            # Run the test script on the client
            $out = $self->_start_test($test);
            if ( $out ) {
                # Save information on client structure
                # as well as in the list of tests to monitor
                $client->{test} = $test;
                push @$tests, $test;
            }
        } elsif ( exists($client->{vclients}) && @{$client->{vclients}} ) {
            # Start new tests on virtual clients
            undef $client->{vtests};
            $self->_virtual_clients_run($stime);
            push @$tests, @{$client->{vtests}} if $client->{vtests};
        }
    }

    # Remove reference to array of current tests
    delete $self->{current_tests};

    return;
}

#-------------------------------------------------------------------------------
# Internal function
# This function compares two jobs, returns true if both jobs
# are the same. It compares both the fixed and variable parameters.
#-------------------------------------------------------------------------------
sub _job_compare($$$$$;) {
    my ($self, $sid1, $idx1, $sid2, $idx2) = @_;
    my ($params1, $params2);

    # Get list of both fixed and variable parameters for job 1
    $params1 = join("\n", sort $self->_get_params($sid1, $idx1, 1));

    # Get list of both fixed and variable parameters for job 2
    $params2 = join("\n", sort $self->_get_params($sid2, $idx2, 1));

    return ($params1 eq $params2);
}

#-------------------------------------------------------------------------------
# Internal function
# This function returns all parameters for the given test taking
# into account precedence of global, block and local parameters.
#-------------------------------------------------------------------------------
sub _get_params($$$;$) {
    my ($self, $sid, $index, $mode) = @_;
    my (%params, @params, @vparams, $param, $val);
    my $job     = $self->{plan}{sequence}[$sid];
    my $jscript = $job->{list}[$index];

    # Process the special parameter for NFS versions
    if ( !$mode && defined($self->{plan}{nfs_params}) ) {
        $val = _get_pct_parameter($self->{plan}{nfs_params});
        $params{NFS_VERSION} = $val if defined($val);
    }

    # Global parameters
    if ( defined($self->{plan}{parameters}) ) {
        push @vparams,
          _process_params($self->{plan}{parameters}, \%params, $mode);
    }

    # Overwrite global parameters by block parameters
    if ( defined($job->{parameters}) ) {
        push @vparams, _process_params($job->{parameters}, \%params, $mode);
    }

    # Overwrite parameters by local parameters
    if ( defined($jscript->{parameters}) ) {
        push @vparams, _process_params($jscript->{parameters}, \%params, $mode);
    }

    # Make list of parameters
    @params = map { "$_=$params{$_}" } keys %params;

    return (@params, @vparams);
}

#-------------------------------------------------------------------------------
# Internal function
# This function propagates the test manager to all physical clients
# and run the test script on all leaf clients and/or virtual clients
#-------------------------------------------------------------------------------
sub _propagate($;) {
    my $self = shift;
    my ($plan, $client, $test, $runid, $stime, $out, $client_name);
    my $index = 1;
    my $tm_index = 1;
    my $client_id = 0;

    # Start time for all tests
    my $dtime = time + $self->{plan}{tm_dtime} + $self->{plan}{dtime};

    # Clone plan in order to modify the plan for the next
    # level without modifying the plan for this level
    $self->{cplan} = dclone($self->{plan});

    # Loop through each client in current branch
    foreach $client ( @{$self->{clients}} ) {
        $client_name = Tharn::host($client->{client})->id();
        # Start time for this client
        $stime = int($dtime + $client->{tskew});
        if ( ! exists($client->{clients})
          && ( ! exists($client->{vclients}) || ! @{$client->{vclients}} ) ) {
            # This client has no physical nor virtual clients underneath
            # so this must be a physical leaf client and the test script
            # must be started on this client
            $self->_plan_sequence($client);

            # Run id for subtest
            $runid = sprintf('%03d_%03d_%s',
                             $self->{sid}+1, $index++, $client_name);

            # Save test information
            $test = { client => $client->{client},
                      runid  => $runid,
                      stime  => $stime };

            # Run the test script on the client
            $out = $self->_start_test($test);
            $client->{test} = $test if $out;
        } else {
            # Plan this level
            $plan = $self->_plan_level($client);

            # Run id for subtest
            $runid = sprintf('%03d_tm_%s', $tm_index++, $client_name);

            # Save test manager information
            $client->{tm} = { plan     => $plan,
                              clientid => $client_id,
                              runid    => $runid };
            # Propagate to next level
            $self->_start_tm($client, $stime);
        }
        $client_id++;
    }

    # Remove cloned plan, no longer needed
    delete $self->{cplan};

    return;
}

#-------------------------------------------------------------------------------
# Internal function
# This function starts a test manager on the given client
# It takes a hash as an argument (client structure) with
# a key 'tm' with the following info:
#  {
#    plan     => Reference to a hash of the plan for this test manager,
#    clientid => Client id,
#    runid    => Runid for subtest
#  }
# and an optional starting time for the test under this test manager.
#-------------------------------------------------------------------------------
sub _start_tm($$;$) {
    my ($self, $client, $stime) = @_;
    my ($cmd, $sobj);
    my $sid = $self->{sid};
    my $tskew = $client->{tskew};

    # Save the test manager info and delete it from the client
    # so the new instance will not have this copy of the test manager
    my $tm = $client->{tm};
    delete $client->{tm};

    # Set default value for stime
    $stime = 0 if ! defined($stime);

    # The following is the code to run on the remote client to serve as
    # a slave test manager. It calls the constructor for this object and
    # propagates to the next level if this is a middle branch. If this is
    # a leaf with virtual clients, the constructor returns to call the
    # function to create all virtual interfaces on the test manager client
    # and then run the tests on the virtual interfaces.
    $cmd = qq{\{
        use Scalability;
        use Carp qw(croak);
        use SPARSE::Util::Logger qw(logTrc);
        my (\$plan, \$client) = \@_;
        my (\$clients, \%sopts);
        my \$vclients = 0;
        if ( exists(\$client->\{clients\}) ) \{
            \$clients = \$client->\{clients\};
        \} else \{
            \$clients = [ \$client ];
            \$vclients = 1;
            \%sopts = ( leaf => 1 );
        \}
        logTrc("Starting test manager on \$client->\{client\}");
        my \$obj = Scalability->new(clients => \$clients,
                                    plan    => \$plan,
                                    sid     => $sid,
                                    tskew   => $tskew,
                                    slave   => 1, \%sopts);
        if ( \$vclients ) \{
            eval \{ \$obj->_virtual_clients_create() \};
            croak "TM_SETUP: \$@" if \$@;
            \$obj->_virtual_clients_run($stime);
            \$obj->_monitor();
        \} else \{
            \$obj->start();
        \}
        logresult('PASS', 'Slave Test Manager');
    \}};

    # Start the test manager on the client
    $sobj = Tharn::subtest($cmd, '-runid', $tm->{runid}, '-bg', '-ipc',
                           '-client', $client->{client}, '-onexit', sub{},
                           '--', $tm->{plan}, $client);

    # Save test manager information
    $tm->{state}   = 1;
    $tm->{nextseq} = 0;
    $tm->{xid}     = 0;
    $tm->{sobj}    = $sobj;
    $client->{tm}  = $tm;

    return;
}

#-------------------------------------------------------------------------------
# Internal function
# This function starts the test on the physical or virtual client
# It only takes a hash as an argument with the following info:
#  {
#    testid  => Test id of the test to run for the current sequence id,
#    client  => Physical client spec or virtual ip address,
#    runid   => Runid for subtest,
#    stime   => Time for test to start,
#    params  => Reference to an array of parameters,
#    sopts   => Reference to a hash of subtest parameters,
#    virtual => True if this test must be run on a virtual client
#  }
# If testid is not in the hash, _get_test is called to get the
# testid, sopts and params
#-------------------------------------------------------------------------------
sub _start_test($$;) {
    my ($self, $test) = @_;
    my ($job, $script, $testid, $user, $bfunc, @params, %sopts, $rslt);
    my $seq = $self->{plan}{sequence}[$self->{sid}];
    my %client = $test->{virtual} ? () : ( '-client' => $test->{client} );
    my $runid = $test->{runid};

    if ( ! exists($test->{testid}) ) {
        # Get the parameters for this test
        ($testid, $user, @params) = $self->_get_test();
        return if ! defined($testid);

        # Add username option to subtest
        %sopts = ( '-username' => $user ) if defined($user);

        # Save testid and sopts
        $test->{testid} = $testid;
        $test->{sopts}  = \%sopts;

        # Append the parameters to the list
        push @{$test->{params}}, @params;
    }

    # Get job to run
    $job = $seq->{list}[$test->{testid}];

    # Script to run
    $script = $job->{script};

    # Check if this is a '.plan'
    if ( defined($job->{sindex}) ) {
        # Check if all tests have been run
        return if $job->{sindex} >= @{$job->{scripts}};

        # Get the next script to run on the list
        $script = $job->{scripts}[$job->{sindex}];
        $job->{sindex}++;

        # Append script index to runid so each test
        # in the '.plan' file has its own log file
        $runid .= sprintf('_%03d', $job->{sindex});
    }

    # Check if script is a built in function
    if ( $script =~ /^<(.*)>$/ ) {
        $bfunc = $1;
        if ( exists($builtin_functions{$bfunc}) ) {
            $script = $builtin_functions{$bfunc};
        } else {
            $test->{state} = 0;
            $test->{sobj} = undef;
            Tharn::logresult('SCRIPT', abort=>0,
                             "Built in function $script does not exist");
            return;
        }
    }

    # Run the test script on the client
    my $sobj = eval {
      Tharn::subtest($script, 'INTERNAL=1', "TIMESTART=$test->{stime}",
                     @{$test->{params}}, '-runid', $runid,
                     '-bg', %{$test->{sopts}}, %client, '-onexit', sub{});
    };
    $rslt = $@;

    if ( $rslt ) {
        # Log the error so it gets propagated but do not
        # abort in order not to kill the test manager
        if ( ref($rslt) ) {
            # Already a results object
            Tharn::logresult($rslt->type, abort=>0, $rslt->messages);
        } else {
            # Not a results object
            Tharn::logresult('FATAL', abort=>0, $rslt);
        }
        return;
    }

    # Set the state back to running
    # and save the new subtest object
    $test->{state} = 1;
    $test->{sobj} = $sobj;

    return 1;
}

#-------------------------------------------------------------------------------
# Internal function
# This function selects which test in the sequence to run.
# It gives precedence to tests with an explicit number of instances,
# if none is found then it selects a test at random based on its
# percentage.
#-------------------------------------------------------------------------------
sub _get_test($;) {
    my $self = shift;
    my ($jscript, @params, %params, $param, $rval, $sum, $index);
    my $job = $self->{plan}{sequence}[$self->{sid}];
    my $idx = 0;
    my $count = 0;

    # Go through all tests in sequence and give priority to the
    # ones which have an explicit number of instances to be run
    foreach $jscript ( @{$job->{list}} ) {
        if ( $jscript->{nclients} !~ /%/ && $jscript->{instances} > 0 ) {
            $index = $idx;
            last;
        }
        $idx++;
    }

    if ( defined($index) ) {
        # A test with an explicit number of instances was found
        $jscript = $job->{list}[$index];
        $jscript->{instances}--;
    } else {
        # Select test based on percentages
        $sum = 0;
        $rval = rand(100);
        undef $jscript;

        # Select test from the percentage list
        foreach $idx ( @{$job->{pct_list}} ) {
            $count = $job->{list}[$idx]{nclients};
            $count =~ s/%//;
            $sum += $count;
            if ( $sum > $rval ) {
                # This test is selected, save the
                # job structure and its index
                $jscript = $job->{list}[$idx];
                $index = $idx;
                last;
            }
        }
    }

    # Return if test was not found
    return if ! defined($jscript);

    # Get the parameters for this test
    @params = $self->_get_params($self->{sid}, $index);

    return ($index, $jscript->{username}, @params);
}

#-------------------------------------------------------------------------------
# Internal function
# This function is used to plan each sequence. It only plans
# or allocates the tests with explicit number of clients since tests
# with percentages do not need any more planing.
#-------------------------------------------------------------------------------
sub _plan_sequence($$;$) {
    my ($self, $client, $seq) = @_;
    my ($item, $count, $freeclients);

    # Get correct sequence
    $seq = defined($seq) ? $seq : $self->{plan}{sequence}[$self->{sid}];

    # Get the total number of available clients for this branch
    $freeclients = defined($client->{nclients}) ? $client->{nclients}
      : (defined($client->{vclients}) && @{$client->{vclients}})
      ? @{$client->{vclients}} : 1;

    # Allocate the tests to be run under this client
    foreach $item ( @{$seq->{list}} ) {
        $item->{instances} = 0;
        $item->{continuous} = 0 unless exists($item->{continuous});
        $item->{ignore_errors} = 0 unless exists($item->{ignore_errors});

        # Only process tests which have an explicit number of clients to run
        next if (! $item->{nclients} || $item->{nclients} =~ /%/);

        # Get the number of instances to run under this client
        if ( $freeclients > $item->{nclients} ) {
            $count = $item->{nclients};
        } else {
            $count = $freeclients;
        }

        # Allocate this test
        $item->{nclients} -= $count;
        $item->{instances} = $count;
        $freeclients -= $count;
    }

    return;
}

#-------------------------------------------------------------------------------
# Internal function
# This function plans or allocates all tests in all sequences
# for a given branch or test monitor.
#-------------------------------------------------------------------------------
sub _plan_level($$;) {
    my ($self, $client) = @_;
    my ($plan, $seq, $item);

    # Use the cloned plan
    $plan = $self->{cplan};

    # Allocate tests for this client in all sequences
    foreach $seq ( @{$plan->{sequence}} ) {
        $self->_plan_sequence($client, $seq);
    }

    # Clone the plan just processed in order to change the number
    # of instances to be run on this branch
    $plan = dclone($self->{cplan});

    # Get cloned plan ready for test manager at next level
    foreach $seq ( @{$plan->{sequence}} ) {
        foreach $item ( @{$seq->{list}} ) {
            if ( $item->{instances} ) {
                # Set the total number clients running
                # this test on given client
                $item->{nclients} = $item->{instances};
            }
        }
    }

    return $plan;
}

#-------------------------------------------------------------------------------
# Internal function
# This function process the output of each subtest on clients_check
# and calculate the clock skew between the main test manager client
# and the each of the physical clients.
# It returns the total number of subtests processed.
#-------------------------------------------------------------------------------
sub _info_process($$$;) {
    my ($self, $sobjs, $tdelta) = @_;
    my ($sobj, @results, $reftime, $idx, @info, %cpuinfo, $client, $out);
    my $count = 0;

    # Check all available subtests
    foreach $idx ( keys %$sobjs ) {
        $sobj = $sobjs->{$idx};
        $reftime = time() - $tdelta;
        if ( !$sobj->is_alive() ) {
            # This subtest has already finished so process its output
            my $result;
            @results = $sobj->own_results();
            foreach my $res ( @results ) {
                if ( $res->type() eq 'PASS' ) {
                    # Get the PASS result where the
                    # information is located
                    $result = $res;
                    last;
                }
            }
            # Remove this subtest object from list so it is not processed
            # again and keep track of how many objects have been processed
            $count++;
            delete $sobjs->{$idx};

            if ( !defined($result) ) {
                # Error using this client, log a warning
                $result = NATE::Result::worst(@results);
                $client = $self->{clients}[$idx]{client};
                my $client_name = Tharn::host($client)->id();
                Tharn::logresult('WARN',
                                 "Client <$client_name> will not be used:",
                                 $result->messages());

                # Remove this client from the list
                $self->{clients}[$idx] = undef;
                next;
            }

            # Calculate the clock skew with the timestamp of the PASS result
            $out = $result->{timestamp} - $reftime;
            $self->{clients}[$idx]{tskew} = $out;

            # Get the cpuinfo and the available memory from the result
            ($out) = ($result->messages())[0];
            @info = split /--/, $out;
            $self->{clients}[$idx]{mem}  = shift @info;

            # Put the information of all the CPU's into a hash in order
            # to discriminate from 'ghost' CPU's. On linux a single
            # CPU can be seen as two different entities but with the
            # same core id -- the information from the client has the
            # performance of each CPU and its core id. By putting this
            # information on a hash it removes any 'ghost' CPU's
            %cpuinfo = @info;
            $out = 0;
            ($out += $_) foreach values %cpuinfo;
            $self->{clients}[$idx]{mips} = $out;
        }
    }

    return $count;
}

#-------------------------------------------------------------------------------
# Internal function
# This function runs the test script on all virtual interfaces
#-------------------------------------------------------------------------------
sub _virtual_clients_run($$;) {
    my ($self, $stime) = @_;
    my ($test, $vclient, $runid);
    my $client = $self->{clients}[0];
    my $index  = 1;

    # Subtest locally for every virtual client
    foreach $vclient ( @{$client->{vclients}} ) {
        if ( defined($self->{current_tests})
          && $self->{current_tests}{$vclient} ) {
            next;
        }

        # Run id for subtest
        $runid = sprintf('%03d_%03d_%s', $self->{sid}+1, $index++, $vclient);

        # Save test information
        $test = { client  => $vclient,
                  runid   => $runid,
                  stime   => $stime,
                  params  => [ "VIRTUAL_IP=$vclient" ],
                  virtual => 1 };

        # Run the test script on the client
        $self->_start_test($test) or next;

        # Save test on list
        push @{$client->{vtests}}, $test;
    }

    return;
}

#-------------------------------------------------------------------------------
# Internal function
# This function creates the virtual ip address interfaces on the
# local host
#-------------------------------------------------------------------------------
sub _virtual_clients_create($;) {
    my $self = shift;
    my (%devs, $interface, $addr, $bcast, $mask, $dev, $cmd, $out);
    my ($device, $netmask, $broadcast, $net_ip, $net_mask);

    my $first_vip = $self->{clients}[0]{vclients}[0];
    my $last_vip  = $self->{clients}[0]{vclients}[-1];

    # Save the id for the process that configures
    # the virtual ip addresses
    $self->{vips_setup_id} = $$;

    # Get interface name, ip address, broadcast address and netmask
    $out = qx($sudo $ifconfig -a 2>&1);
    croak "Could not get interface information: $out" if $?;

    foreach ( split /\n/, $out ) {
        # Get interface name
        $interface = $1 if ( /^(\S+)/ );

        # Get the ip address, broadcast and mask for interface
        ($addr, $bcast, $mask) = (/inet addr:(\S+)\s+Bcast:(\S+)\s+Mask:(\S+)/);
        if ( $addr && $bcast && $mask ) {
            # Get the network part of the interface
            $net_mask = inet_aton($mask);
            $net_ip   = inet_aton($addr) & $net_mask;

            # Save list of virtual interfaces for each real interface
            if ( $interface =~ /(.*):/ ) {
                push @{$devs{$1}}, $interface;
            }
            # Check if this is a valid interface for the requested range
            if ( ! defined($device) && $interface !~ /:/
              && (inet_aton($first_vip) & $net_mask) eq $net_ip
              && (inet_aton($last_vip) & $net_mask) eq $net_ip ){
                # This is the main interface where all
                # virtual ip addresses will be created
                $device    = $interface;
                $netmask   = $mask;
                $broadcast = $bcast;
            }
        }
    }

    if ( ! defined($device) ) {
        croak "Could not find proper interface for virtual ip range" .
              " [$first_vip, $last_vip]";
    }

    # Delete all virtual ip's on the chosen interface
    _virtual_interface_down($devs{$device});

    # Create virtual ip addresses on client
    $dev = 0;
    foreach $addr ( @{$self->{clients}[0]{vclients}} ) {
        # Save virtual device for cleanup
        push @{$self->{vdevs}}, "$device:$dev";

        # Command to create virtual device
        $cmd = "$sudo $ifconfig $device:$dev $addr " .
               "netmask $netmask broadcast $broadcast 2>&1";
        logDbg($cmd);
        $out = qx($cmd);
        croak "Could not create virtual interface: $out" if $?;
        ++$dev;
    }

    return;
}

#-------------------------------------------------------------------------------
# Internal function
# This function deletes the virtual ip address interfaces on the local host
#-------------------------------------------------------------------------------
sub _virtual_interface_down($;) {
    my $vdevs = shift;
    my ($dev, $cmd, $out);

    foreach $dev ( @{$vdevs} ) {
        # Command to remove virtual device
        $cmd = "$sudo $ifconfig $dev down 2>&1";
        logDbg($cmd);
        $out = qx($cmd);
        print "Could not remove virtual interface: $out" if $?;
    }
}

#-------------------------------------------------------------------------------
# Internal function
# Sets resource agitation levels on the filer for current sequence
# point. The plan has two types of setral options: global and sequence.
# Global setral options apply to all sequences and sequence setral
# options apply only to the given sequence. Sequence setral
# options have precedence over global options
#-------------------------------------------------------------------------------
sub _setral($;) {
    my $self = shift;
    my ($fc, $item, $curr, $prev, %agitation, %setral_opts, @reset);
    my $sid = $self->{sid};

    # Add global setral settings
    if ( defined($self->{plan}{setral}) ) {
        %agitation = (%{$self->{plan}{setral}});
    }

    # Get setral settings for current sequence
    $curr = $self->{plan}{sequence}[$sid]{setral};
    if ( defined($curr) ) {
        # Add setral settings for current sequence
        # these have higher priority than global settings
        %agitation = (%agitation, %{$curr});

        # Mark all global settings that are masked out by
        # current sequence settings
        foreach $item ( keys %{$curr} ) {
            if ( exists($self->{plan}{setral}{$item}) ) {
                # Current sequence setting will be used instead
                $self->{plan}{setral}{$item}{state} = 0;
            }
        }
    }

    if ( $sid > 0 ) {
        # Get setral settings for previous sequence
        $prev = $self->{plan}{sequence}[$sid-1]{setral};

        # Get the setral settings that need to be reset
        @reset = grep { !exists($agitation{$_}) } keys %{$prev};
    }

    if ( %agitation || @reset ) {
        # Use require in order to dynamically load the module
        # to save memory usage when setral is not being used
        require filer;

        # Get filer connection object
        Tharn::host($main::FILER);
        $fc = Tharn::connect('#console');

        # Reset setral settings which do not
        # exist on current sequence
        foreach $item ( @reset ) {
            # Make a copy of the options and delete internal option
            %setral_opts = %{$prev->{$item}};
            delete $setral_opts{state};

            # Clear all options for setting
            $setral_opts{$_} = 0 foreach keys %setral_opts;

            # Reset setral setting
            filer::filer_resource_agitation_set($fc, $item, %setral_opts);
        }

        # Set setral settings for current sequence
        foreach $item ( keys %agitation ) {
            # Skip settings which have already been set
            next if $agitation{$item}{state};

            # Make a copy of the options and delete internal option
            %setral_opts = %{$agitation{$item}};
            delete $setral_opts{state};

            # Enable setral setting and flag it
            filer::filer_resource_agitation_set($fc, $item, %setral_opts);
            $agitation{$item}{state} = 1;
        }
    }

    return $fc;
}

#-------------------------------------------------------------------------------
# Internal function
# This function process the 'parameters' option of the plan.
# The last argument is the mode for this function.
# If mode is not specified or it is false, only one value is
# selected for every parameter. If the mode if true all values are
# selected for every parameter, this is used for comparing parameter
# definitions for different tests.
#-------------------------------------------------------------------------------
sub _process_params($$;$) {
    my ($parameters, $params, $mode) = @_;
    my (@vparams, $param, $item, $val, $count);

    # Process all variable parameters
    foreach $param ( keys %{$parameters} ) {
        if ( $mode ) {
            # In this mode, just return a list of parameters
            # with all their values
            if ( ref($parameters->{$param}) eq 'HASH' ) {
                # Add all parameters in hash but append the
                # percentage to the value
                foreach $item ( keys %{$parameters->{$param}} ) {
                    # Append the percentage to the value
                    $val = $parameters->{$param}{$item};
                    push @vparams, "$param=$item;$val";
                }
            } elsif ( ref($parameters->{$param}) eq 'ARRAY' ) {
                # Add all parameters in list
                foreach $val ( @{$parameters->{$param}} ) {
                    push @vparams, "$param=$val";
                }
            } else {
                # Just add the parameter to the list
                push @vparams, "$param=$parameters->{$param}";
            }
        } else {
            # In this mode, select only one value from the list
            if ( ref($parameters->{$param}) eq 'HASH' ) {
                # Select one according to percentages
                $val = _get_pct_parameter($parameters->{$param});
                $params->{$param} = $val if defined($val);
            } elsif ( ref($parameters->{$param}) eq 'ARRAY' ) {
                # Select one from list in a random but uniform manner
                $count = @{$parameters->{$param}};
                if ( $count > 0 ) {
                    # Select index at random
                    $count = int($count * rand());
                    $params->{$param} = $parameters->{$param}[$count];
                }
            } else {
                # Use the value
                $params->{$param} = $parameters->{$param};
            }
        }
    }

    return @vparams;
}

#-------------------------------------------------------------------------------
# Internal function
# This function selects a value based on percentages
#-------------------------------------------------------------------------------
sub _get_pct_parameter($;) {
    my $parameters = shift;
    my ($item, $value);
    my $sum  = 0;
    my $rval = rand(100);

    # Select which value of parameter to use based on its percentage
    foreach $item ( keys %$parameters ) {
        # Get value and remove the percentage sign
        $value = $parameters->{$item};
        $value =~ s/%//;

        # Accumulate value for comparison to a random value
        $sum += $value;
        if ( $sum > $rval ) {
            return $item;
        }
    }

    return;
}

#-------------------------------------------------------------------------------
# Internal function
# Recursive function to calculate the number of leaf clients (physical
# and/or virtual)
#-------------------------------------------------------------------------------
sub _total_clients($;) {
    my $clients = shift;
    my ($client, $count, $ncount);
    my $sum = 0;
    my $max_count = 0;

    # Get the total number of clients on this branch
    foreach $client ( @{$clients} ) {
        if ( defined($client->{clients}) ) {
            # Recurse into next level
            $count = _total_clients($client->{clients});
            $ncount = @{$client->{clients}};
            $max_count = $ncount if $ncount > $max_count;
        } else {
            # This is a leaf so add the number of virtual clients and if this
            # client does not have virtual clients just add the client itself
            $count = @{$client->{vclients}} ? @{$client->{vclients}} : 1;
        }
        # Save the number of clients for this branch
        $sum += $count;
        $client->{nclients} = $count;
    }

    # Add to total number of test managers
    $tm_total += $max_count;

    return $sum;
}

#-------------------------------------------------------------------------------
# Internal function
# Check for errors on subtest
#-------------------------------------------------------------------------------
sub _check_errors($;) {
    my $sobj = shift;
    my ($result, $msgs);

    # Get worst result for subtest
    $result = NATE::Result::worst($sobj->own_results);
    if ( $result && $result->severity > NATE::Result::Warn->severity ) {
        $msgs = join("\n", $result->msgs);
        if ( $msgs =~ /TM_SETUP/ ) {
            # This is a test manager setup error
            croak $msgs;
        }
    }
}

#-------------------------------------------------------------------------------
# Internal function
# Propagate subtest results up
#-------------------------------------------------------------------------------
sub _propagate_results($$;) {
    my $self = shift;
    my $item = shift;

    my ($r_host, $r_sid, $r_script, $r_log, $r_type, $r_lg, @r_messages);
    my (@item_res, $r_obj, @r_msgs, $r_res, $virtual_ip);

    # Given the current object (which should be a task manager) and a named
    # item (a sub-task-manager or a test object) collect the host information,
    # sequence id, script name, test log address, result type, result
    # messages, and continuous status of the item.

    # Get the host, sid, script, and log
    if ( defined($item->{sobj}{client}) ) {
        $r_host = $item->{sobj}{client};
    } else {
        $r_host = $self->{clients}[0]{client};
        $virtual_ip = $item->{sobj}{params}{VIRTUAL_IP};
    }
    $r_host = (split(/;/, $r_host))[0];
    $r_host = "virtual ip $virtual_ip on $r_host" if defined($virtual_ip);
    $r_sid = $self->{sid};
    $r_script = $item->{sobj}{scriptspec};
    $r_log = $item->{sobj}{logfile};

    # Get the results
    @item_res = $item->{sobj}->own_results();

    # Special handling for slave test managers
    if ($r_script =~ /Slave Test Manager/) {
        # Reduce the script info, since all slave test managers have the same
        # script text. Reduces output in final report.
        $r_script = 'Slave Test Manager';
        $r_sid = 0; # No sequence id, use 0 (everything stored as initial seq)
        $r_lg  = 0; # Not a load generator

        # Propagate all information messages upwards
        foreach $r_obj (@item_res) {
            @r_msgs = $r_obj->messages();
            if (($r_obj->type() eq 'INFO') &&
                ($r_msgs[0] =~ /^Result message propagation/)) {
                Tharn::addresult($r_obj->type(), $r_obj->messages());
            }
        }
    } else {
        # Set the load generator flag
        my $seq = $self->{plan}{sequence}[$r_sid];
        $r_lg = ($seq->{list}[$item->{testid}]{ignore_errors}) ? 1 : 0;
    }

    # Get the result type and messages
    $r_res = NATE::Result::worst(@item_res);
    if (defined $r_res) {
        if ( $r_res->severity() < 2 ) {
            # The worst result is either an INFO or a WARN
            # so look if there exists a PASS result
            foreach $r_obj ( @item_res ) {
                if ( $r_obj->type() eq 'PASS' ) {
                    # Propagate the PASS result instead
                    $r_type = 'PASS';
                    @r_messages = $r_obj->messages();
                    last;
                }
            }
        }
        if ( ! defined($r_type) ) {
            $r_type = $r_res->type();
            @r_messages = $r_res->messages();
        }
    } else {
        $r_type = 'FATAL';
        @r_messages = ("No results found for test");
    }

    # Add an information result to the current object that contains
    # the information of the sub object.
    Tharn::addresult("INFO", "Result message propagation ---- " .
                             "$r_host ---- " .
                             "$r_sid ---- " .
                             "$r_script ---- " .
                             "$r_log ---- " .
                             "$r_type ---- " .
                             "$r_lg ---- " .
                             $r_messages[0]);
}

#-------------------------------------------------------------------------------

=head2 scalability_report

  $obj->scalability_report(message  => "Scalability message",
                           rslt_obj => $@);

Process the information results returned from the scalability tests
and generate a report page summarizing the test results from each
sequence performed.

Determine the PASS/FAIL status of the entire scalability test and
log an appropriate result.

The following named arguments are accepted:

=over

=item Arguments

=over

=item message

(Required) : Test description message to be printed with result.

=item rslt_obj

(Optional) : A result object or string that is alternately the
result of the test. If the scalability tests fail, any rslt_obj will
be logged as a warning. If the scalability tests pass, any rslt_obj
will be logged as the test result.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------------------
sub scalability_report($%;) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            message  => { type => SCALAR },
            rslt_obj => { type => SCALAR|OBJECT, optional => 1 },
        }
    );
    my $msg = $opts{message};
    my $rslt_obj = $opts{rslt_obj};
    my $maxsid = 0;
    my $error = 0;
    my (@results, @rslt_types, @logfiles, @logtimes, $file, $first, $start,
        $end, @localtime, $numlogs, @msgs, $result, $report, $filename,
        %lg_count, %tm_count, %task_count, @res_info, %lg_reports, %tm_reports,
        %task_reports, @lg_summary, @tm_summary, @task_summary, @lg_report_str,
        @tm_report_str, @task_report_str, $host, $sid, $script, $log, $type,
        $lg, $message, $test, @log_info, $test_start, $test_end, $lo, $hi,
        $curr, $max, $min, $i);

    # Load library for find
    require File::Find;

    if ( $self->{background} ) {
        Tharn::logcomment('Waiting for main test manager to finish');
        eval { $self->{tm}->wait_finish(timeout => 300) };
        @results = $self->{tm}->own_results();
    } else {
        @results = Tharn::own_results();
    }

    @rslt_types = ('PASS', 'FAIL', 'SCRIPT', 'CONFIG', 'FATAL',
                   'INSPECT', 'NA', 'WARN', 'INFO', 'PARAM');

    Tharn::logcomment("BEGIN: scalability_report");

    Tharn::logcomment("Generate list of logs, start and end times");

    # Get the list of relevant log files
    File::Find::find(sub { push @logfiles, $File::Find::name
                             if /^\d+_\d+_.*\.log$/ },
                     $main::TOP_LOGDIR_UNIX);

    foreach $file (@logfiles) {
        # Read the first line of the log file to get the start time
        if (open(TLOG, "< $file")) {
            $first = <TLOG>;
            close(TLOG);
            $first =~ /^(\d+ \d+)/;
            $start = $1;
            # Remove the space between date and time
            $start =~ s/ //;
            # Use stat to get the end time, and convert to YYYYMMDDhhmmss format
            $end = (stat($file))[9];
            @localtime = localtime($end);
            $end = sprintf("%04d%02d%02d%02d%02d%02d", $localtime[5]+1900,
                           $localtime[4]+1, $localtime[3], $localtime[2],
                           $localtime[1], $localtime[0]);
            push @logtimes, "${start}--${end}--$file";
        } else {
            logWarn("Could not open log file $file: $!");
        }
    }
    @logtimes = sort @logtimes;
    $numlogs = $#logtimes;

    Tharn::logcomment("Generate test results");

    # Scan the results and generate the report accordingly
    foreach $result (@results) {
        @msgs = $result->messages();
        @res_info = split(' ---- ', $msgs[0]);
        if ($res_info[0] ne 'Result message propagation' ||
            $result->type() ne 'INFO') {
            # Result does not indicate it is a propagation or is not an INFO
            logWarn("Invalid result propagation encountered");
            next;
        }

        (undef, $host, $sid, $script, $log, $type, $lg, $message) = @res_info;

        # Ignore INFO results completely
        next if $type =~ /INFO/;

        # Is there a failure on any non-load generation test
        # (including task managers)
        $error = 1 if ($type !~ /PASS|PARAM|INFO/ && !$lg);

        $maxsid = ($sid > $maxsid) ? $sid : $maxsid;
        $report = "Script: $script\n  Host: $host\n  " .
                  "Log file: $log\n  Message: $message\n\n";

        if ($lg) {
            # Result is from a load generator:
            $lg_count{$type}[$sid]++;
            $lg_reports{$type}[$sid] .= $report;
        } elsif ($script eq 'Slave Test Manager') {
            # Result is from a task manager:
            $tm_count{$type}[$sid]++;
            $tm_reports{$type}[$sid] .= $report;
        } else {
            # Result is from a scalability task:
            if ($type !~ /PASS|PARAM/) {
                # On a subtest error, get all the tests that were
                # running during that time frame. To do this, we get the
                # time frame from the test log, and find all logs that have
                # timestamps within that time frame.

                Tharn::logcomment("Found failure in $log");

                # Get the time frame of the failed test
                $test = (grep { m/$log/ } @logtimes)[0];
                @log_info = split('--', $test);
                $test_start = $log_info[0];
                $test_end = $log_info[1];
                if ( $log_info[2] ne $log ) {
                    logWarn("scalability_report: log data corrupt");
                }
                $filename = "$reports_dir/$test_start-$test_end.log";

                # If the given filename exists, don't duplicate the work.
                # Results will be the same.
                if ( ! -e $filename ) {
                    Tharn::logcomment("Calculating test ranges");

                    # Here we are going to do a binary search on the sorted
                    # @logtimes list.  First find the max entry that has
                    # a start time that is less than the end time of the
                    # test in question.
                    $lo = 0;
                    $hi = $numlogs;
                    $curr = int(($hi-$lo)/2);

                    while ( $curr != $lo && $curr != $hi ) {
                        if ( (split('--', $logtimes[$curr]))[0] < $test_end ) {
                            $lo = $curr;
                            $curr += int(($hi-$lo)/2);
                        } else {
                            $hi = $curr;
                            $curr -= int(($hi-$lo)/2);
                        }
                    }
                    # If the value at $hi fits the range, include it, else
                    # use the $lo index for $max
                    $max = ( (split('--', $logtimes[$hi]))[0] <= $test_end ) ?
                      $hi : $lo;

                    # Next, find the min entry, between 0 and max, that has
                    # an end time that is later than the start time of the
                    # test in question.
                    $lo = 0;
                    $curr = int(($hi-$lo)/2);

                    while ( $curr != $lo && $curr != $hi ) {
                        if ( (split('--', $logtimes[$curr]))[1] > $test_start ) {
                            $hi = $curr;
                            $curr -= int(($hi-$lo)/2);
                        } else {
                            $lo = $curr;
                            $curr += int(($hi-$lo)/2);
                        }
                    }
                    # If the value at $lo fits the range, include it, else
                    # use the $hi index for $min
                    $min = ( (split('--', $logtimes[$lo]))[1] >= $test_start ) ?
                      $lo : $hi;

                    # All logs with indices from min to max have a start time
                    # before the test in question ends, and an end time after
                    # the test in question starts, so some part of the log
                    # occurred during the execution of the test in question.
                    # Put the information into a text file with the start-end
                    # timestamp as the name.
                    if (open(TFILE, "> $filename")) {
                        foreach $i ($min .. $max) {
                            print TFILE (split('--', $logtimes[$i]))[2] . "\n";
                        }
                        close(TFILE);
                        # Modify the report string to point
                        # to the file
                        $report =~ s/  Message/  Active tests: $filename\n  Message/;
                    } else {
                        logWarn("Could not open log file $filename: $!");
                    }
                }

            }
            # Save the scalability task report information
            $task_count{$type}[$sid]++;
            $task_reports{$type}[$sid] .= $report;
        }
    }

    Tharn::logcomment("Generate test summaries and reports");

    foreach $i (0..$maxsid) {
        # Generate the test summaries
        $lg_summary[$i] = '';
        $tm_summary[$i] = '';
        $task_summary[$i] = '';

        foreach $type (@rslt_types) {
            $lg_summary[$i] .= "  $type: $lg_count{$type}[$i]\n"
              if $lg_count{$type}[$i];
            $tm_summary[$i] .= "  $type: $tm_count{$type}[$i]\n"
              if $tm_count{$type}[$i];
            $task_summary[$i] .= "  $type: $task_count{$type}[$i]\n"
              if $task_count{$type}[$i];
        }

        # Generate the reports
        $lg_report_str[$i] = '';
        $tm_report_str[$i] = '';
        $task_report_str[$i] = '';

        foreach $type (keys %lg_count) {
            next if $type eq 'PASS';
            $lg_report_str[$i] .= "-- $type --\n$lg_reports{$type}[$i]\n\n"
              if $lg_reports{$type}[$i];
        }
        foreach $type (keys %tm_count) {
            next if $type eq 'PASS';
            $tm_report_str[$i] .= "-- $type --\n$tm_reports{$type}[$i]\n\n"
              if $tm_reports{$type}[$i];
        }
        foreach $type (keys %task_count) {
            next if $type eq 'PASS';
            $task_report_str[$i] .= "-- $type --\n$task_reports{$type}[$i]\n\n"
              if ($task_reports{$type}[$i]);
        }
    }

    Tharn::logcomment("Print report to file");

    # Print the reports to the file
    if ( open(REPORT, "> $report_file") ) {
        print REPORT "=================\n";
        print REPORT "    SUMMARIES\n";
        print REPORT "=================\n\n";
        foreach $i (0..$maxsid) {
            $lg_summary[$i] = "  No tests executed\n" if !$lg_summary[$i];
            print REPORT "-- Sequence $i --\n\n";
            print REPORT "Tests Run With Errors Checked:\n$task_summary[$i]\n";
            print REPORT "Tests Run With Errors Ignored:\n$lg_summary[$i]\n";
            print REPORT "\n";
        }
        print REPORT "-- Task managers: --\n$tm_summary[0]\n";
        if ( $maxsid || $tm_report_str[0] ) {
            print REPORT "\n\n";
            print REPORT "===============\n";
            print REPORT "    REPORTS\n";
            print REPORT "===============\n\n";
        }
        foreach $i (0..$maxsid) {
            if ($task_report_str[$i] || $task_report_str[$i]) {
                print REPORT "-- Sequence $i --\n\n";
                print REPORT "Tests run with errors checked - test reports:" .
                  "\n\n$task_report_str[$i]\n" if $task_report_str[$i];
                print REPORT "Tests run with errors ignored - test reports:" .
                  "\n\n$lg_report_str[$i]\n" if $lg_report_str[$i];
            }
        }
        print REPORT "-- Task manager reports: --\n\n$tm_report_str[0]\n"
          if $tm_report_str[0];

        close(REPORT);
    } else {
        logWarn("Could not open report file $report_file: $!");
    }

    # Log a result
    if ($error) {
        # An error from one of the tests was found
        if ($rslt_obj) {
            # Get any $rslt_obj messages
            @msgs = $rslt_obj->messages() if ref($rslt_obj);
            unshift @msgs, ("Error reported from Scalability execution");
            # Log a warning
            logWarn(\@msgs);
        }
        # Log a fail result for the test
        Tharn::logresult('FAIL', $msg);
    } else {
        # Test passed, log any $rslt_obj
        if(ref($rslt_obj)) {
            Tharn::logresult($rslt_obj);
        } elsif ($rslt_obj) {
            Tharn::logresult('FATAL', $rslt_obj);
        } else {
            Tharn::logresult('PASS', $msg);
        }
    }

    Tharn::logcomment("END: scalability_report");
}

#-------------------------------------------------------------------------------
# Destructor for object
# It removes all virtual ip address interfaces that this object created
#-------------------------------------------------------------------------------
sub DESTROY {
    my $self = shift;
    local($@);

    # Only remove the virtual interfaces created by this object
    if ( defined($self->{vips_setup_id}) && $self->{vips_setup_id} == $$ ) {
        _virtual_interface_down($self->{vdevs});
    }
}

#---------------------------------------------------------------------

=head1 testme

Self test routine for this library

Use test_lib.thpl to invoke this routine.

  Example
    ntest -hw PKGNAME=Scalability SUBNAME=testme util/test_pkg.thpl

=over

=item Arguments

None

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub testme() {
    my ($client, @clients, %plan, $sc_obj);

    # Get information about localhost for the client list
    $client = 'localhost;hosttype=' . Tharn::host('localhost')->hosttype() .
              ';default_ip=' . Tharn::host('localhost')->default_ip();
    @clients = ( { client => $client } );

    # Create a plan that requires too many clients
    %plan = ( sequence => [
                           { list => [
                                      { script => "nfsop_leaf.thpl",
                                        nclients => 1000
                                      },
                                     ]
                           }
                          ] );

    # Run Scalability
    eval { $sc_obj = Scalability->new( clients => \@clients, plan => \%plan ) };

    # Validate that the correct error was returned
    if ($@ !~ /not enough clients to run this plan/) {
        Tharn::logresult('FAIL', "Scalability did not report too few clients");
    } else {
        Tharn::logresult('PASS', "SELF TEST of Scalability.pm");
    }
}

1;

