# $Id: //depot/prod/test/Rsierranevada/lib/Sentinel/Utils.pm#1 $

## @summary Sentinel::Utils - Sentinel utility library
## @author jzhuge@netapp.com sandeep@netapp.com sriram@netapp.com
## @pod here

=head1 NAME

Utils - Sentiel Utility Functions

=head1 SYNOPSIS

    use Sentinel::Utils;

    report_error($error[, $cleanup_error]);
    $num_vvols = aggr_num_vvols($conn, $aggr);
    $num_vvols = filer_num_vvols($conn);
    $num_vols = filer_num_vols($conn);

    create_volumes($conn, $aggr, $num_vols, $vol_size, %opts);
    delete_volumes($conn, $num_vols);
    $num_vols can also be a volume name iterator generated by one of the iterator functions below.

    $iter = default_name_iterator($count[, $start_index[, $name_prefix]]);
    $iter = list_name_iterator(@name_list);

=head1 DESCRIPTION

This module defines utility routines for Sentinel test cases

This library is still in beta status, testers welcome, but please be
warned that the behavior and interface may change.

=head1 FUNCTIONS

Many functions below take C<$conn>, which is a Connection object
(Connectrec) that is created by calling C<Tharn::connect()>.

=cut

package Sentinel::Utils;
use strict;

use Tharn;
use SPARSE::Util::Logger qw(logStep logWarn logDbg);
use Volume qw(vol_create_flex vol_purge vol_list);
use Aggregate qw(aggr_status);
use SayWrap qw(sayEOF);

BEGIN {
    use Exporter ();
    our ( $VERSION, @ISA, @EXPORT_OK );

    $VERSION = 2.00;
    @ISA     = qw(Exporter);

    @EXPORT_OK = qw(
      report_error
      aggr_num_vvols
      filer_num_vols
      filer_num_vvols
      create_volumes
      delete_volumes
      default_name_iterator
      list_name_iterator
      snap_pininfo_count
      my_aggr_copy
    );
}

=head2 report_error($error, $cleanup_error)

Log an error string or a result object.

=cut

sub report_error {
    my ( $error, $cleanup_error ) = @_;
    $error
      and ref($error) ? logresult($error) : logresult( 'FATAL', $error );
    $cleanup_error and logresult( 'WARN', "Cleanup returned error: $error" );
    1;
}

=head2 aggr_num_vvols($conn, $aggr)

Return the number of VVols in an aggregate
Use aggr_status routine

=cut

sub aggr_num_vvols {
    my ( $conn, $aggr ) = @_;
    my $status_obj = aggr_status( $conn, $aggr );
    my $vols_list = $status_obj->volumes();
    1 + $#{$vols_list};
}

=head2 filer_num_vols($conn)

Return the number of 'Volumes (trad + flex)' on a filer

=cut

sub filer_num_vols {

    my ($conn) = @_;

    my @vols     = vol_list($conn);
    my $num_vols = $#vols + 1;
    logStep("filer has $num_vols volumes");

    return $num_vols;
}

=head2 filer_num_vvols($conn)

Return the number of VVols on a filer

=cut

sub filer_num_vvols {

    my ($conn) = @_;

# TODO: Use sayEOF for now
# Switch to library function once support is added
    my $text = sayEOF( $conn, "vol status" );
    my $num_vvols = 0;
    foreach ( split( /\r?\n/, $text ) ) {
        /^\s*(\S+).*\sflex\s/ and ++$num_vvols;
    }

    return $num_vvols;
}

sub my_aggr_copy {

    my ( $conn, $src, $tgt, $timeout ) = @_;
    use constant ABORT_STR => "Aborting";

    my ( $match, $before, undef ) = say( $conn, "aggr copy start $src $tgt",
        ABORT_STR, "100% done", -timeout, $timeout, -fail,
        "aggr copy from $src to $tgt failed" );
    $match eq ABORT_STR and do { chomp $before; die $before };
}

=head2 create_volumes($conn, $aggr, $num_vols, $vol_size, %opts)

Create many volumes in an aggregate

=cut

sub create_volumes {
    my ( $conn, $aggr, $num_vols, $vol_size, %opts ) = @_;
    exists $opts{guarantee} or $opts{guarantee} = 'none';
    my $vol_name_iter =
      ref($num_vols) ? $num_vols : default_name_iterator($num_vols);

    while ( my $vol_name = $vol_name_iter->() ) {
        logDbg("create vvol $vol_name");
        vol_create_flex( $conn, $vol_name, $aggr, $vol_size, %opts )
          or die "Failed to create volume $vol_name";
    }
    1;
}

=head2 delete_volumes($conn, $num_vols)

Delete many volumes

=cut

sub delete_volumes {
    my ( $conn, $num_vols ) = @_;
    my $vol_name_iter =
      ref($num_vols) ? $num_vols : default_name_iterator($num_vols);

    while ( my $vol_name = $vol_name_iter->() ) {
        logDbg("delete volume $vol_name");
        vol_purge( $conn, $vol_name );
    }
    1;
}

=head2 default_name_iterator($count[, $start_index[,$name_prefix]])

Return the default name iterator.
Each name has the format: $PREFIX$INDEX, e.g. vv_1, vv_2, ...

For example, default_name_iterator(10, 5, "vol_") returns
an iterator for vol_5, vol_6, vol_7, ..., vol_14.

=cut

sub default_name_iterator {
    my $count  = shift;
    my $index  = shift || 0;        # start index
    my $prefix = shift || "vv_";    # name prefix
    $count += $index;               # end index

    return sub {
        ++$index;
        return $index > $count ? undef: "$prefix$index";
      }
}

=head2 list_name_iterator(@name_list)

Return an iterator for names listed in parameters

For example, list_name_iterator(qw(testvol1 temp2 nv)) returns
an iterator for testvol1, temp2, and nv.

=cut

sub list_name_iterator {
    my $list = \@_;

    return sub {
        return shift @$list;
      }
}

=head2 snap_pininfo_count ($filer, $volume)


This function runs the snap pininfo command on the filer and returns
the count of the number of snapshot root inodes that are pinned
in-core.

parmeters:
filer - required connection object of filer
$volume - volume name

returns the number of pinned roots


=cut

sub snap_pininfo_count {
    my ( $filer, $volume ) = @_;
    my $text;
    $text = sayEOF( $filer, "priv set -q test; snap pininfo $volume" );
    chomp $text;
    $text =~ /Number of pinned roots: (.*)/;

    # Return the count we get back
    return $1;
}

1;

=head1 SEE ALSO

=head1 BUGS

=head1 STATUS

Shared among Sentinel test cases

=head1 ATTRIBUTES

=head1 AUTHOR

John Zhuge (jzhuge@netapp.com)
Sandeeep Yadav (sandeep@netapp.com)

=head1 COPYRIGHT

Copyright 2005, Network Appliance Inc., All rights reserved.
