#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Server.pm#2 $

#***********************************************************************
#
# NAME
#   Server.pm - remote programmatic administration of a Server.
#
# SYNOPSIS
#   Enables remote programmatic administration of a Server
#
# DESCRIPTION
#   This Perl module is meant to be programmatic way of administering a
#   Server.
#
# AUTHOR
#	dl-central-libs
#
#***********************************************************************

## @summary     Programmatic administration of a Server.
## @author dl-central-libs
## @pod here

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDESERVER}) {
    my $newModule = $ENV{OVERRIDESERVER};
    delete $ENV{OVERRIDESERVER};
    print "The main Server module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);
package Server;

use Node::Classic::6_X;
use Node::Classic::7_X;
use Node::Classic::7_X::FlexCache;
use Node::NG;
use TracePkg;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF BOOLEAN);
use Options;
use strict;
use IO::Socket;
use Carp qw(confess croak);
use SayWrap;
use Data::Dumper;
use Params::Common; # pick up NODES/FILER
use Net::Ping;

# These don't have to match the values in Filer.pm
use constant CLASSIC => 1;
use constant NG => 2;

#***********************************************************************
# NOTE
# The following pieces of code have been hacked to make FlexCache work
# with GX. This needs to be cleaned up by completely rewriting the 
# topology code. 
# This is an ugly hack because the full Topology hash needs to be moved out of here.
# If there is nothing in the clustered array, then the topology has been filled in on the
# nodes side. However, most of the code in the other libraries still treats a single 7G
# filer as a one-node cluster and looks for information in {cluster} so we need to move
# the data over for now.
# 
#  Server: discover_topology
#  Server: create_topology
#  Server: delete_topology
#  Node: pick_aggregates
#***********************************************************************

=pod

=head1 NAME

Server.pm - Programmatic interface to GX node or cluster

=head1 SYNOPSIS

  use Server;
  my $server_obj = Server->new;

  # If you want to execute on specific node.
  $server_obj->set_execution_server(-Server => "node1");

  # Get data from cluster.
  # Get all volumes in cluster
  my @vols = $server_obj->volume_show;

  # Get only volumes in vs0 VServer.
  my @vs0_vols = $server_obj->volume_show(vserver => 'vs0');

  # Iterate over return data structures.
  foreach my $vol (@vols) {
       print "Volume:     $vol->{volume}\n";
       print "VServer:    $vol->{vserver}\n";
       print "Used:       $vol->{used}\n";
       print "Permissions:   $vol->{'unix-permissions'}";
       ...
  }

  # Create objects in the cluster - wrap in eval() if you want
  # to trap errors.
  eval {
      $server_obj->volume_create(volume  => 'my_new_vol',
				 vserver => 'vs0',
				 size    => '200mb',
				 'unix-permissions' => '777',
				 'junction-path'    => '/my_new_vol');
  };

  if ($@) {
      # Error path - do something other than die, otherwise, there
      # is not much point in trapping the error via 'eval'.
  }

  # Issue commands to FreeBSD shell.
  my $shell_output = $server_obj->shell(cmd => 'ls -l /mroot/log');
  $server_obj->shell(cmd      => 'sysctl debug.panic=1',   # Force a panic.
                     noprompt => 1);     # Don't wait for prompt to return.

  # Issue commands to RLM.
  my $rlm_output = $server_obj->rlm(cmd => 'version');

  # Issue commands to dbladecli
  my $dblade_output = $server_obj->dblade(cmd => 'disk show');

  # Issue commands to CFE/LOADER (must already be at the CFE/LOADER prompt).
  $server_obj->cfe(cmd => 'boot_backup');

=head1 DESCRIPTION

Server class is an interface to GX clusters as well as 7G filers.   It
contains interfaces to communicate with the CLI, FreeBSD shell (GX only),
RLM, console, dbladecli (GX only) and the CFE/LOADER.

Refer to the B<METHODS> section below for an explanation on using the
Autoloaded methods (provides access to all CLI commands), as well as
the predefined methods.

The Server class also provides backwards compatibility for GX command syntax
for CLI commands that are issued through the Autoloaded methods.  Backwards
compatibility is not provided for direct calls made using the cli, shell,
rlm, console, dblade, or cfe methods.

=head1 METHODS

In the calling sequences below, square brackets B<[]> represent
optional parameters.

=head1 Autoloaded methods

The Server class contains a small set of predefined methods (see B<Predefined
Methods> below), but the majority of functionality comes from the ability
of Server to make any GX CLI command look like a method call through the
use of the Perl AUTOLOAD feature.  The basic syntax of Autoloaded methods
is as follows:

    $server_obj-><method>([-Server   => <node name>],
                          [-Help     => <1 | 0>],
			  [output    => <variable reference>],
			  [-Timeout  => <seconds>],
			  [nocr      => <1 | 0>],
			  [poke      => <1 | 0>],
			  [noprompt  => <1 | 0>],
			  [-Verification => <undef>],
			  [exit_code => <variable reference>],
			  [<Command Specific Arguments>]);

where B<method> is a base CLI command that has been converted to the proper
format for Server (see B<Method Construction> below).

Arguments passed to the autoloaded method are described below in the
B<Command Specific Arguments> and B<Standard Server Arguments> sections.

=head2 Method Construction

If you know the syntax of the GX CLI, you can issue calls using a Server
object that appear to be methods as follows:

The method name is the base portion of the CLI command joined together
with underscores.  For example, a simple 'volume show' command in the
CLI, becomes 'volume_show' as the method name, as in the following example

    $server_obj->volume_show();

Any command in the CLI can be made into a method using the above syntax.  If
the command has a dash ('-') in it, as in 'vserver name-mapping', you need
to convert the dash to an underscore as well.  So, in this case, the
method becomes:

    $server_obj->vserver_name_mapping();

=head2 Arguments

There are 2 flavors of arguments that can be passed to the autoloaded methods,
the standard set of Server arguments that are applicable to all autoloaded
methods (these will be gobbled up by Server and acted on appropriately),
and the command specific arguments that are visible in the GX CLI (can be
seen when you use the '?' operator after the base command to see what options
that command takes).

=head3 Command Specific Arguments

Command specific arguments are passed to the methods in the following manner.
First, remove the leading dash, use the name as it appears in the CLI as
the left hand operand, use the '=>' operator, then, specify a value
as the right hand operand.  For example, if I want to use the volume show
command as above, but would like to provide a specific vserver and
volume name, the method call will look as follows:

    $server_obj->volume_show(vserver => 'vs0',
			     volume  => 'myvol');

If the argument contains a dash in it, you need to quote the left side:

    $server_obj->volume_show('security-style' => 'unix');

Otherwise, Perl will complain about the use of a bareword since it will
interpret the dash a a minus operator.

=head3 Standard Server Arguments

The standard Server arguments shown in the basic syntax above can be passed
to any autoloaded method, and many of the predefined methods (the specific
methods that accept the standard Server arguments are shown below).

=over 4

=item -Server

Execute the command on the specified node.  Must be a node that is already
known to Server.

=item -Help

Ask the filer for help on the given command. This currently only works with
GX/10mode and will pass the '?' to the command. The command will return a
structure that represents the possible fields/options available to the command.

=item output

Variable reference used for returning raw output.  Use this if you want the
raw output as well as the parsed.  Value must be a reference to a variable,
as follows:

   my $output;
   @vols = $server_obj->volume_show(output => \$output);
   print "Raw output from 'volume show' command is: $output\n";

=item -Timeout

How long to wait for output from specified command.  Default value is 20
seconds.  Supplied value must be in seconds.

=item nocr (boolean - default is false)

If set to true (1), do not append a carriage return to the end of the
specified command.

=item poke (boolean - default is false)

Send the poke command after issuing the specified command, before
looking for the prompt.

=item noprompt (boolean - default is false)

Issue the specified command without waiting for a prompt.

=item exit_code

Variable reference used to store the exit code for shell commands.

   my $rc;
   $output = $server_obj->shell(cmd       => '/path/to/someprog',
                                exit_code => \$rc)
   print "Exit code from 'someprog' is: $rc\n";

=item -Verification

Causes create or delete object verification to be skipped.

=back

=head3 Output

The output back from Server calls will come in two forms.  If the method is
predefined (see B<Predefined Methods> below), the form of the output will be
described along with the method description.  For all Autoloaded methods, the
output is returned as an array of hash references.  You can determine what
the keys to the hashes are by the following procedure.

1. log into ngsh

2. issue 'set -privilege diagnotstic'

3. issue 'set -showallfields true'

4. issue the command of interest (e.g. 'storage aggregate show')

5.  The output is difficult to read, but you will see two levels of  headings.

The first set of header information that you see are the names of the hash
keys that will be returned in the output.  If the name (key) contains a
dash, you will need to quote the key, as in:

    print $aggr_out->{'raid-block-checksum'}

Since the Server class sets the 'showallfields' attribute to true, and is
in diagnostic mode when issuing the commands, you have access to all the
output fields that are listed when you issue the command of interest.

The second set of headers is what you will see as the labels in the
CLI, and are for display purposes only.

=head1 Predefined methods

=cut

my @Server_array;
my %Hostspecs;
my %Servers;
my $Current_server;
my $Object;

my $Trace = new TracePkg;

# Find the name of the script running this code
my $Testname = $Tharn::Scriptspec;
$Testname =~ s/.+\///; # just get the name of the test
$Testname =~ s/\.thpl//; # strip off the .thpl extension


my $Options = new Options;
$Options->parameters(definitions=>
		     [
		      {
                       name => "FORCE_DELETE_RESOURCES",
                       description => "delete_topology() should delete the resources",
                       spec => "boolean",
                       default => undef,
		      },
		      {
                       name => "FIND_ALL_NODES",
                       description => "try to find all of the nodes in the cluster",
                       spec => "boolean",
                       default => undef,
		      },
		      {
                       name => "NODES_ON_DEMAND",
                       description => "should the node connections be made on demand",
                       spec => "boolean",
                       default => undef,
		      },
		      {
                       name => "RANDOM_NODE",
                       description => "use a random node for each call",
                       spec => "boolean",
                       default => undef,
		      },
		      {
                       name => "SKIP_NODES",
                       description => "skip this list of nodes (useful with FIND_ALL_SERVERS)",
# This should be turned on after a bug in Options is fixed
#                       dependencies => [qw(FIND_ALL_NODES)],
                       spec => "s@",
		      },
		      ]
		     );
$Options->compatability(definitions=>
			{
                         SERVERS_ON_DEMAND => "NODES_ON_DEMAND",
                         FILERS_ON_DEMAND => "NODES_ON_DEMAND",
                         FIND_ALL_SERVERS => "FIND_ALL_NODES",
                         FIND_ALL_FILERS => "FIND_ALL_NODES",
			});

#***********************************************************************

=head2 new

Create a new Server object

    $server_obj = Server->new([nodes => <node1[,node2,...]>],
			      [no_cli => <1 | 0>],
			      [nodes_on_demand => <1 | 0>]);

This is the constructor for Server objects.  A new object is returned on
success.  On failure, the C<new> method will die, causing your program
to die as well, unless you have trapped it inside an eval block.

As part of the C<new> call, the Server library will create connections
to the nodes specified in the B<NODES> parameter passed to C<ntest>.  You
can specify that any creation of connections beyond the first node
specified should be deferred until they are used by setting the
B<filers_on_demand> parameter to true (1).

When logging into a GX filer, the C<new> method will automatically make
connections to the shell and cli of each node. For each node, the
username/password can be separate for the two connections. Server will
default to the default shipping values. This is overridden by information
found in the nate_host entries for the node. The connection entry for the
cli should be called 'cli'. The connection entry for the shell should be
'shell' or 'ssh' (the old way). Here are examples:

conn shell { conntype=ssh-persist ; connhost=10.98.206.216 ; username=root ; password=admin }
conn cli { conntype=ssh-persist ; connhost=10.98.206.216 ; username=snider ; password="admin" }

The values can be overridden by CLI_USER/CLI_PASSWORD and
SHELL_USER/SHELL_PASSWORD. See the C<PARAMS> section below for info
on these.

By default, cli connections are attempted by logging into the cli connection.
If this fails, the connection is made via the shell connection and via
a single ngsh instance. The connection is then named 'ngsh'. See the
C<CLI_FROM_SHELL> param below to run multiple 'ngsh -c' commands.

=over 4

=item nodes

Specifies a list of nodes which will be managed by this object. The values specified
here will override any value passed in NODES.

=item no_cli

This parameter is passed on to Node::NG and is used to determine whether or
not to establish a connection to the CLI.  If set to 1 (true), cli
connections will not be established up front, but non-generated cdefs
(used primarily for 7G) will be loaded.  It set to 0 (false - this is the
default), CLI connections will be established and all generated cdefs will
be loaded.  This parameter can be used to defer creating connections, making
startup time for Server object creation quicker.  Also useful if you are
communicating only with the FreeBSD shell, dbladecli, or RLM and have no
need for the CLI.  However, if this parameter is used and set to true,
you will not have access to the CLI unless you explicitly call the
C<establish_connections> method prior to issuing any calls to the CLI.

=item nodes_on_demand (boolean - default is false)

Discover only the first node specified in the NATE NODES parameter.
Other nodes will go through the discovery phase as they are used (via
set_execution_server, or using the '-Server' option on other autoloaded
calls, as in $server_obj->volume_show(-Server => 'node1').  Default is
false (0), meaning that all nodes will be discovered up front.  By setting
this to true (1), you can significantly speed up the startup time when you
have specified a good number of nodes in the NODES param. This param does
not have an effect in a non-clustered environment.

=back

=cut

sub new {
    $Trace->entry();
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {
		nodes => undef,
                no_cli => 0,
                nodes_on_demand => undef,
		@_, # overlay defaults with what was explicitly passed
		};
    bless($self, $class);

    if ($Object) {
        $Trace->exit(msg=>"return $Object");
        return $Object;
    }

    $Options->process();
    my $options = $Options->get_hash();

    # create empty arrays so that we do not hit script errors when trying
    # to dereference them later
    @Server_array = ();
    $self->{types} = {
                      clustered => {},
                      node => {},
                     };

    # A user might be passing in hostspecs rather than simply nodenames so just assume
    #  this to be the case.
    # This logic now also means that the user cannot override this node list with the NODES
    #  parameter. If the code passes in a value, it will override anything in NODES.
    my @hostspecs;
    if (defined($self->{nodes})) {
        if (!ref($self->{nodes})) {
            @hostspecs = split(/,/, $self->{nodes});
        } elsif (ref($self->{nodes}) ne "ARRAY") {            
            $Trace->fatal(msg=>"The 'nodes' parameter to new() must be a scalar or an array reference. Got nodes='$self->{nodes}'",
                          result=>"SCRIPT");
        } else {
            @hostspecs = @{$self->{nodes}};
        }
    } else {
        if ($Options->defined(option=>"NODES")) {
            @hostspecs = @{$options->{NODES}};
        }
        if (!@hostspecs) {
            $Trace->fatal(msg=>"Could not find any nodes in either the NODES parameter\n".
                          "or the nodes method param", result=>"CONFIG");
        }
    }
    # save the information regarding the host
    foreach my $hostspec (@hostspecs) {
        $self->_add_host_to_lists($hostspec);
    }
    # at this point, @Server_array is populated

    my $skip_array = $options->{SKIP_NODES};

    # this loop is not needed if the dependencies check above can be turned on
    foreach my $server (@Server_array) {
        if (grep($server eq $_, @$skip_array)) {
            $Trace->fatal(msg=>"$server is in NODES and SKIP_NODES arrays", result=>"CONFIG");
        }
    }

    if ($options->{RANDOM_NODE}) {
        # define a random node to be the default
        $Current_server = $Server_array[int rand(scalar @Server_array)];
    } else {
        # define the first server as the default server
        $Current_server = $Server_array[0];
    }

    # setup the object for the current server.
    my $object = $self->discover(server=>$Current_server, _no_cli => $self->{no_cli});

    my $on_demand = $options->{NODES_ON_DEMAND} || $self->{nodes_on_demand};

    if (! keys %{$self->{types}{clustered}}) {
        # if there is more than one 7mode filer, certain options will not work
        if (keys %{$self->{types}{node}}) {
            if ($options->{FIND_ALL_NODES}) {
                $Trace->warn(msg=>"The FIND_ALL_NODES param was set but will not work when running against 7-mode filers.\n Disabling it.");
                $Options->set(FIND_ALL_NODES=>undef);
            }
            if ($on_demand) {
                $on_demand=undef;
                $Options->set(NODES_ON_DEMAND=>undef);
                $Trace->warn(msg=>"The NODES_ON_DEMAND param was set but will not work when running against 7-mode filers.\n Disabling it");
            }
        }
    }

    # If FIND_ALL_NODES is still set, then this is a cluster. Otherwise, the code above would have turned the switch off.
    # If no_cli is set, Node::NG nodes will load basic cdefs which may not work for performing discovery.
    if ($options->{FIND_ALL_NODES} && $self->{no_cli}) {
        $Trace->warn(msg=>"The FIND_ALL_NODES param was set but will not work when Server->new() is called with no_cli=>1.\n Disabling it");
        $Options->set(FIND_ALL_NODES=>undef);
    }        

    # call find_all_nodes if we still have FIND_ALL_NODES set.  only pass in the clustered nodes
    # because they are the only ones which we currently support doing a find_all_nodes on
    if ($options->{FIND_ALL_NODES}) {
        $self->find_all_nodes(); # this will default to only run against the clustered nodes
    }

    if (!$on_demand) {
        # discover the rest of the servers at start time
        foreach my $server (@Server_array) {
            $self->discover(server=>$server, _no_cli => $self->{no_cli});
        }

        # pick a new random server from the list to be the default
        if ($options->{RANDOM_NODE}) {
            $Current_server = $Server_array[int rand(@Server_array)];
        }
    } else {
        # we really should do some kind of health check on the nodes
        # in this case since we lost that check which came for free
        # before since all nodes would at least get a cli connection opened.
    }
    
    $Object = $self;
    # if the running script is not gx_boot, there exists more than one filer running 7G/7-mode and this
    #  this is not a FlexCache test, then error out. Until Topology is removed from Server, the code
    #  cannot support more than one 7G filer very well. But gx_boot has a requirement for more than one.
    if ($Testname !~ /^gx_boot|^gsi|/) {
        if (!$options->{CACHE_TRAFFIC} && keys %{$self->{types}{node}} > 1) {
            $Trace->fatal(msg => "Server has only been tested to work on more than one 7-mode filer when a ".
                          "FlexCache is involved or with the gx_boot.thpl script");
        }
        if (!$options->{CACHE_TRAFFIC} && keys %{$self->{types}{node}} && keys %{$self->{types}{clustered}}) {
            $Trace->fatal(msg => "Server has only been tested to work with a combination of 7-mode and 10-mode ".
                          "nodes when the 7G filer is a Flexcache or with the gx_boot.thpl script");
        }
    }

  done:
    $Trace->exit(msg=>"return $Object");
    return $Object;
} # end new

=head2 servers

Get current list of nodes that have been discovered by the Server class.  This
is a list of nodes that were either passed in via the command line, or
discovered in the cluster because the 'FIND_ALL_FILERS' parameter was
specified.

Does not take any arguments.

  @nodes = $svr_obj->servers;
  print "List of currently known nodes:\n";
  foreach my $node (@nodes) {
      print "\t$node\n";
  }

=cut

sub servers {
    my $self = shift;
    $Trace->entry();
    $Trace->exit(msg=>"Server_array = " . join(" ", @Server_array));
    return @Server_array;
} # end servers

#***********************************************************************

=head2 set_execution_server

Reset the default execution server.

    $orig_server = $server_obj->set_execution_server(-Server => "server1");

This routine will specify which node should be set as the default for all
future calls through Server. The previous execution server is returned to
the caller.

=over 4

=item -Server

The name of the server to set as the new default

=back

=cut

sub set_execution_server {
    my $self = shift;
    my %params = (
                  -Server => undef,
                  @_,
                  );
    my $output="";

    my $server = $params{-Server} || $params{Server};
    $Trace->entry(msg=>"$server");

    if (!$server) {
	$Trace->fatal(msg=>"A \"-Server\" param is required", result=>"SCRIPT");
    } 
    if (!grep($server eq $_, @Server_array)) {
	my $msg = "looking for ".$server." in ".join(":", @Server_array)."\n";
	$Trace->fatal(msg=>"The server '$server' must be in the known server list\n$msg", 
		      result=>"CONFIG");
    } 

    my $orig_server = $Current_server;
    $Current_server = $server;
    $Trace->trace(msg=>"Current_server = $Current_server");
	
    $Trace->exit(msg=>"return $orig_server");
    return $orig_server;
} # end set_execution_server

#***********************************************************************
=head2 get_execution_server

Get the current execution server.

    my $original_node = $server_obj->get_execution_server();

This routine will return current execution server node.

=cut

sub get_execution_server {
    my $self = shift;

    $Trace->entry();
    $Trace->exit(msg=>"return $Current_server");
    return $Current_server;
} # end get_execution_server

#***********************************************************************
=head2 get_clustered_nodes

Get a list of the clustered nodes.

    my @clusters = $server_obj->get_clustered_nodes()

This routine will return the list of clustered nodes

=cut

sub get_clustered_nodes {
    my $self = shift;
    $Trace->entry(msg=>"");

    # maintain the order of server names as specified by the user and
    #  stored in @Server_array.
    my @nodes;
    foreach my $node (@Server_array) {
        push(@nodes, $node) if (exists($self->{types}{clustered}{$node}));
    }
    $Trace->exit(msg=>join(", ", @nodes));
    return @nodes;
}

#***********************************************************************

=head2 get_legacy_nodes

Get a list of the non-clustered nodes.

    my @nodes = $server_obj->get_legacy_nodes()

This routine will return the list of non-clustered nodes

=cut

sub get_legacy_nodes {
    my $self = shift;
    $Trace->entry(msg=>"");

    # To work properly when running against FlexCaches and 7G origin filers,
    # create_topology() needs a list of all nodes, including the caches.
    # However, it needs the list to have the non-caches first so that it
    # can create the origin volumes first. Then it wants the caches so it
    # can create the cached-volumes which are depependent on the origin
    # volumes. This will be thrown away when the topology is removed.
    my @nodes = ();
    my $options = $Options->get_hash();
    if ($options->{CACHE_TRAFFIC}) {
        my @cache_nodes = ();
        foreach my $node (keys %{$self->{types}{node}}) {
            if ($self->is_cache(-Server=>$node)) {
                push(@cache_nodes, $node);
            } else {
                push(@nodes, $node);
            }
        }
        push(@nodes, @cache_nodes);
    } else {
        @nodes = keys %{$self->{types}{node}};
    }
    $Trace->exit(msg=>join(", ", @nodes));
    return @nodes
}

#***********************************************************************

=head2 get_topology

Get the current topology structure

    my $topology = $server_obj->get_topology()

This routine will return the reference to the current topology structure 
being maintained by this object. 

=cut

sub get_topology {
    my $self = shift;
    $Trace->entry(msg=>"");
    $Trace->exit();
    return $self->{Topology};
}

#***********************************************************************

=head2 discover

Run discovery on the specified server.

    my $object = $server_obj->discover(server => "server1")

This routine will determine what type of server has been specified and
will create an internal object of that type. These objects are stored
internally and are referenced by calling methods via the '-Server'
param. This really should only be used internally.

The routine returns the object handle.

=cut

sub discover {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'Server' } ],
					 {
                                          server => {type => SCALAR},
                                          _no_cli => {type => SCALAR | UNDEF,
                                                      optional => 1,
                                                      default => undef,
                                                     },
					 }
                                        );
    my $server = $params{server};
    my $object;

    $Trace->entry(msg=>"server = $params{server}");
    if (!$server) {
	$Trace->fatal(msg=>"\"server\" parameter is required", result=>"SCRIPT");
    }
    
    if ($Servers{$server}) { # already been processed
	$Trace->exit();
	return;
    }

    $Trace->trace(msg=>"Running discover($server)");

    ##########################################################################
    # This is a horrible hack until NATE no longer caches the host information.
    # See burt285311
    ##########################################################################

    # retrieve the hostspec for this servername
    my $hostspec = $Hostspecs{$server};
    $Trace->trace(msg=>"host ($hostspec)");

    # delete the entry out of the Tharn data structure (clear the cache)
    delete $Tharn::Hosts{$server};

    # call host() again to make NATE aware of the new hostspec
    #  and obtain the values we need for this hostspec
    my $host_object = main::host($hostspec);
    my $ip = $host_object->default_ip();
    my $hosttype = $host_object->hosttype();
    my $hostname = $host_object->hostname();

    # END HACK

    $params{server} = $hostname;
    $params{server_object} = $self;
    my ($server_array, $find_all) = $Options->get(option=>["FILER", "FIND_ALL_FILERS"]);


    # An object needs to be created for this server even if the CLI is not going to be 
    #  opened. This is so that future calls through AUTOLOAD will correctly make the calls
    #  to the object which should then properly make the necessary connections.
    eval {
        $Trace->trace(msg=>"$hostname has hosttype = $hosttype");

        # check the hosttype to determine if this is a 7-mode or 10-mode host
	if ($host_object->hosttype() =~ /filer-ng|vsim-ng/) {
	    $Trace->trace(msg=>"Create a Node::NG object");
            $self->{types}{clustered}{$server} = 1;
	    $object = new Node::NG(no_cli=>$self->{no_cli}, %params);
	} elsif ($host_object->hosttype() =~ /filer|vsim/) {
            $self->{types}{node}{$server} = 1;
	    $Trace->trace(msg=>"Create a Node::Classic type object");
            if (!$params{_no_cli}) {
                # since no_cli is not set, connect to the node and try to determine the version
                #  of Ontap and check for the Flexcache license so that the proper object type
                #  will be created.
                $Trace->trace(msg=>"no_cli is not set so try to contact the node");
                eval "use Filer::Version qw(get_wafl_version WV_SCRIMSHAW_MAX);";
                die $@ if ($@);
                eval "use filer qw(filer_service_is_licensed);";
                die $@ if ($@);
                eval "use filer qw(filer_service_is_licensable);";
                die $@ if ($@);
                
                my $version;
                no strict 'subs';
                
                # wrap the call in eval since useless failure messages come out
                eval {
                    $version = get_wafl_version($host_object);
                }; 
                if ($@) {
                    $Trace->fatal(msg=>"Could not create the appropriate Node::Classic object: $@");
                }
                if ($version > WV_SCRIMSHAW_MAX) {
                    # determine if this filer is running as a FlexCache or just Classic
                    my $fc = main::connect('#rsh-socket');
                    if ((filer_service_is_licensable($fc, 'flex_cache') && filer_service_is_licensed($fc, 'flex_cache')) ||
                        (filer_service_is_licensable($fc, 'flexcache_nfs') && filer_service_is_licensed($fc, 'flexcache_nfs'))) {
                        $Trace->trace(msg=>"$hostname is running a version of Ontap > scrimshaw (version = $version) and is running FlexCache");
                        $object = new Node::Classic::7_X::FlexCache(%params);
                    } else {
                        $Trace->trace(msg=>"$hostname is running a version of Ontap > scrimshaw (version = $version)");
                        $object = new Node::Classic::7_X(%params);
                    }
                } else {
                    $Trace->trace(msg=>"$hostname is running a version of Ontap <= scrimshaw (version = $version)");
                    $object = new Node::Classic::6_X(%params);
                }
            } else {
                $Trace->comment(msg=>"Since no_cli has been set for $hostname, the software version cannot be determined\n");
                $Trace->comment(msg=>"The hosttype specifies a 7-mode system so assuming an Ontap 7.X Server object");
                # this is an interesting state. The user has said to not connect to the CLI. But that connection 
                # is needed so that the proper object can be created (6_x vs 7_x). Since gx_boot is the primary 
                # user of no_cli and it is only used for Boilermaker and later builds, just assume that it will
                # be a 7_x node so create that object. Also pass it the no_cli setting.
                $object = new Node::Classic::7_X(no_cli => 1, %params);
            }
	} else {
	    $Trace->fatal(msg=>"could not determine the platform type for $hostname\n",
			  result=>"CONFIG");
	}
    };
    if ( $@ ) {
	# there was an error contacting the server
	if ( $find_all && ! grep($hostname eq $_, @$server_array )) {
	    # skip this host since it was not in the input list and we are discovering all servers
	    $Trace->warn(msg=>"Skipping '$hostname' because it could not be reached: $@");
	    goto done;
	} else {
	    # this server was passed in so fatal out
	    $Trace->fatal(msg=>"Could not reach '$hostname': $@");
	}
    }

    $Servers{$hostname} = $object;
    
  done:
    $Trace->exit();
    return $object;
} 

#***********************************************************************

=head2 is_cache

Determine whether or not the specified object is a cache device.

    if ($server_obj->is_cache(-Server => "server1")) {
       print "server1 is a cache\n";
    }

This routine will return true if the server is a cache device. 

=over 4

=item -Server

Optional. This method will work against the current default server unless
this parameter specifies which server to query.

=back

=cut

sub is_cache {
    my $self = shift;
    my %params = (
		  -Server => undef,
		  @_, );
    $Trace->entry(msg=>"Entered with -Server set to:  $params{-Server}");

    # find the name of the server we are talking to
    my $server = $params{-Server} || $params{Server} || $Current_server;
    my $object = $self->_get_server_object(server=>$server);
    my $cache = $object->is_cache();

    $Trace->exit(msg=>"return '$cache'");
    return $cache;
}

# This code is going to undergo a major change so I do not want to
# advertise it yet.
sub discover_topology {
    my $self = shift;
    my %params = (
		  cmd_nodes => undef,
		  @_,
		  );
    my %topology; # to be returned
    
    no strict 'refs';
    
    $Trace->entry();
    $Trace->trace(msg=>"Discovery in progress");

    # predefine the empty hashes for cluster and node
    $topology{cluster} = {};
    $topology{node} = {};

    my @discovered_nodes = $self->_get_nodes();

    foreach my $node (@discovered_nodes) {
        my $orig_server = $self->set_execution_server(-Server => $node);

        # this hash contains the method names to call for each cmdnode resource that needs to be created
        my %cmd_node_definitions = $self->get_topology_definitions(definition => "discover");
    
        my @cmd_nodes = ($params{cmd_nodes}) ? @{$params{cmd_nodes}} : $self->_process_cmdnode_params();

        foreach my $cmd_node (@cmd_nodes) {
            $Trace->trace(msg=>"node = $cmd_node");
            if (!$cmd_node_definitions{$cmd_node}) {
                $Trace->comment(msg=>Dumper(\%cmd_node_definitions));
                $Trace->fatal(msg=>"got a bad definition for $cmd_node", result=>"SCRIPT");
            }
            
            my ($method, %method_params) = $self->_find_method_definition($cmd_node_definitions{$cmd_node});
            eval {
                # Please see the NOTE in Server.pm regarding the Topology hack
                if ($self->get_type() == $self->ONTAP_NG()) {
                    $Trace->trace(msg=>"Run '$method' against clustered node '$node'");
                    # just run against the first node in the list
                    $topology{cluster}{$cmd_node} = $self->$method(-Server=>$node, %method_params);
                } else {
                    $Trace->trace(msg=>"Run '$method' against single node '$node'");
                    $topology{node}{$node}{$cmd_node} = [] if (!$topology{node}{$node}{$cmd_node});
                    push(@{$topology{node}{$node}{$cmd_node}}, @{$self->$method(-Server=>$node, %method_params)});
                }
            };
            if ($@ && $@ !~ /This table is currently empty/) {
                $Trace->fatal(msg=>"$@\n", result=>"CONFIG");
            }
	}
        $self->set_execution_server(-Server => $orig_server);
    }

    if (! $self->get_clustered_nodes() ) {
        foreach my $node ($self->get_legacy_nodes()) {
            foreach my $cmd_node (keys %{$topology{node}{$node}}) {
                push(@{$topology{cluster}{$cmd_node}}, @{$topology{node}{$node}{$cmd_node}});
            }
            delete $topology{node}{$node};
        }
    }

    # save to the file if it was specified
    $self->_write_topology_files(\%topology);

  done:
    $Trace->exit();
    $self->{Topology} = \%topology;
    return \%topology;

} # end discover_topology

# This code is going to undergo a major change so I do not want to
# advertise it yet.
sub create_topology {
    my $self = shift;
    my %params = (
		  print_node => 1, #undef,
		  cmd_nodes => undef,
		  topology => undef,
		  @_,
		  );
    
    no strict 'refs';
    my %topology; # to be returned
    
    $Trace->entry();
    $Trace->comment(msg=>"Creation in progress") if ($params{print_node});

    # make a local copy which we will modify
    my $new_topology;
    foreach my $cmd_node (keys %{$params{topology}{cluster}}) {
        $new_topology->{cluster}{$cmd_node} = $params{topology}->{cluster}{$cmd_node};
    }
    foreach my $node (keys %{$params{topology}{node}}) {
        foreach my $cmd_node (keys %{$params{topology}{node}{$node}}) {
            $new_topology->{node}{$node}{$cmd_node} = $params{topology}->{node}{$node}{$cmd_node};
        }
    }

    # We need to run the create on any of the clustered nodes and against each of the 
    # single nodes.
    my @create_nodes = $self->_get_nodes();

    if (!$self->get_clustered_nodes()) {
        # look for any nodes listed in @create_nodes but not in the topology nodes
        my @missing_nodes = ();
        foreach my $node (@create_nodes) {
            if (!$new_topology->{node}{$node}) {
                push(@missing_nodes, $node);
            }
        }
        if (@missing_nodes > 1) {
            $Trace->comment(msg=>"legacy=".Dumper(\@create_nodes));
            $Trace->comment(msg=>"topology=".Dumper($new_topology));
            $Trace->fatal(msg=>"This is a strange error where we found more than one node in the master list of 7G nodes than in the topology list. See snider to figure it out.");
        } 
        foreach my $node (@missing_nodes) {
            foreach my $cmd_node (keys %{$new_topology->{cluster}}) {
                next if (ref($new_topology->{cluster}{$cmd_node}) ne "ARRAY");
                if (!exists($new_topology->{node}{$node}{$cmd_node})) {
                    $new_topology->{node}{$node}{$cmd_node} = [];
                }
                push(@{$new_topology->{node}{$node}{$cmd_node}}, @{$new_topology->{cluster}{$cmd_node}});
                delete $new_topology->{cluster}{$cmd_node};
            }
        }
    }

    foreach my $node (@create_nodes) {
        my $orig_server = $self->set_execution_server(-Server => $node);
        
        # this hash contains the method names to call for each cmdnode resource that needs to be created
        my %create_definitions = $self->get_topology_definitions(definition => "create");
        
        # this hash contains the method names to call for each cmdnode resource that needs to be checked for existence
        my %exists_definitions = $self->get_topology_definitions(definition => "exists");
        
        # obtain the list of cmdnodes to process
        my @cmd_nodes = ($params{cmd_nodes}) ? @{$params{cmd_nodes}} : $self->_process_cmdnode_params();

        # foreach cmdnode, try to create each resource. The creation routine should be idempotent.
        foreach my $cmd_node (@cmd_nodes) {
            $Trace->comment(msg=>"node = $cmd_node") if ($params{print_node});
            next if (!$create_definitions{$cmd_node}); # we don't know how to create this resource
            
            # we have a cmdnode to examine, so try to create each resource
            my @resource_hashes;
            # Please see the NOTE in Server.pm regarding the Topology hack
            if ($self->get_type() == $self->ONTAP_NG()) {
                next if (!$new_topology->{cluster}{$cmd_node}); # this cmdnode does not have an entry in the topology
                @resource_hashes = @{$new_topology->{cluster}{$cmd_node}};
            } else {
                next if (!$new_topology->{node}{$node}{$cmd_node}); # this cmdnode does not have an entry in the topology
                @resource_hashes = @{$new_topology->{node}{$node}{$cmd_node}};
            }
            foreach my $resource_params (@resource_hashes) {
                my ($exists_method, %exists_params) = $self->_find_method_definition($exists_definitions{$cmd_node});
                my $exists_cdef = $self->get_command_definition(command=>$exists_method); # this is the command definition for the exists() routine
                
                # go through each of the input params for the new resource and see if it should be passed to exists()
                # if the routine is a real method then pass all of the params. If it is not then check out the definition
                # and only pass valid params
                while (my ($param, $value) = each %$resource_params) {
                    $exists_params{$param} = $value;
                }
                
                # check to see if the resource exists. If not, try to create it.
                # The methods are called via the Server object so that multiple Classic
                #  filers can be managed.
                $Trace->debug(msg=>"call $exists_method for $cmd_node\n"); 
                if (!$self->$exists_method(%exists_params)) {
                    my ($method, %create_params) = $self->_find_method_definition($create_definitions{$cmd_node});
                    $Trace->trace(msg=>"call $method for $cmd_node on $node\n");
                    
                    # just run against the first node in the list
                    my $resource_array = $self->$method(%{$resource_params}, Topology=>$new_topology,
                                                        -Idempotent=>1, -IgnoreTransient => 1);

                    # We successfully created the resource so mark it for deletion
                    $resource_params->{Delete_Resource} = 1;

                    if ($resource_array && ref($resource_array) eq "ARRAY") {
                        push(@{$topology{$cmd_node}}, @$resource_array);
                    }
                }
            }
        }
        $self->set_execution_server(-Server => $orig_server);
    }
    
  done:
    $Trace->exit();
    # this next line is just wrong. However, we should add the recently created resources to the existing
    # topology that is stored
#    $self->{Topology} = \%topology;
    return $self->{Topology};

} # end create_topology

# This code is going to undergo a major change so I do not want to
# advertise it yet.
sub delete_topology {
    my $self = shift;
    my %params = (
		  print_node => 1, #undef,
		  cmd_nodes => undef,
		  topology => undef,
		  @_,
		  );
    
    no strict 'refs';
    
    $Trace->entry();
    $Trace->comment(msg=>"Creation in progress") if ($params{print_node});

    my $options = $Options->get_hash();
    my $topology = $params{topology};

    # We need to run the deletes on any of the clustered nodes and against each of the 
    # single nodes.
    my @delete_nodes = $self->_get_nodes(reverse_legacy => 1);

    foreach my $node (@delete_nodes) {
        $Trace->trace(msg=>"Setting server to $node\n");
        my $orig_server = $self->set_execution_server(-Server => $node);

        # call node-specific code to perform any pre-cleanup checks
        $self->pre_delete_topology();
        
        # this hash contains the method names to call for each cmdnode resource that needs to be created
        my %delete_definitions = $self->get_topology_definitions(definition => "delete");
        
        # this hash contains the method names to call for each cmdnode resource that needs to be checked for existence
        my %exists_definitions = $self->get_topology_definitions(definition => "exists");
        
        # obtain the list of cmdnodes to process
        my @cmd_nodes = ($params{cmd_nodes}) ? @{$params{cmd_nodes}} : $self->_process_cmdnode_params();

        # foreach cmdnode, try to create each resource. The creation routine should be idempotent.
        foreach my $cmd_node (@cmd_nodes) {
            $Trace->trace(msg=>"node = $cmd_node");
            next if (!$delete_definitions{$cmd_node}); # we don't know how to delete this resource
            # we have a cmdnode to examine, so try to delete each resource
            my @resource_hashes;
            # Please see the NOTE in Server.pm regarding the Topology hack
            if ($self->get_type() == $self->ONTAP_NG() || !$self->get_clustered_nodes()) {
                $Trace->trace(msg=>"type=NG, cmd_node=$topology->{cluster}{$cmd_node}\n");
                next if (!$topology->{cluster}{$cmd_node}); # this cmdnode does not have an entry in the topology
               @resource_hashes = @{$topology->{cluster}{$cmd_node}};
            } else {
                $Trace->trace(msg=>"type=Classic, cmd_node=$topology->{node}{$node}{$cmd_node}\n");
                next if (!$topology->{node}{$node}{$cmd_node}); # this cmdnode does not have an entry in the topology
                @resource_hashes = @{$topology->{node}{$node}{$cmd_node}};
            }
            foreach my $resource_params (@resource_hashes) {
                # send a warning if we decide not to delete a resource and we are not being told to
                if (!$resource_params->{Delete_Resource} && !$options->{FORCE_DELETE_RESOURCES}) {
                    $Trace->comment(msg=>Dumper $resource_params);
                    $Trace->warn(msg=>"Did not delete a resource because this test did not create it and the FORCE_DELETE_RESOURCES param was not set");
                    next;
                }
                my ($exists_method, %exists_params) = $self->_find_method_definition($exists_definitions{$cmd_node});
                my $exists_cdef = $self->get_command_definition(command=>$exists_method); # this is the command definition for the exists() routine

                # go through each of the input params for the new resource and see if it should be passed to exists()
                # if the routine is a real method then pass all of the params. If it is not then check out the definition
                # and only pass valid params
                while (my ($param, $value) = each %$resource_params) {
                    $exists_params{$param} = $value;
                }
                
                # check to see if the resource exists. If so, try to delete it.
                # The methods are called via the Server object so that multiple Classic
                #  filers can be managed.
                $Trace->debug(msg=>"call $exists_method for $cmd_node\n"); 
                if ($self->$exists_method(%exists_params)) {
                    my ($method, %delete_params) = $self->_find_method_definition($delete_definitions{$cmd_node});
                    $Trace->debug(msg=>"call $method for $cmd_node\n");
                    eval {
                        $topology->{$cmd_node} = $self->$method(%{$resource_params}, -Idempotent => 1,
                                                                Topology=>$topology);
                    };
                    if ($@) {
                        if ($@ =~ /Ruleset is in use by a volume/) {
                            # I really hate commenting this out but it is causing warns to always appear
                            # in the smokes
#                            $Trace->warn(msg=>$@);
                        } else {
                            die $@;
                        }
                    }
                }
            }
        }
        $Trace->debug(msg=>"Setting server back to $orig_server\n");
        $self->set_execution_server(-Server => $orig_server);
    }
    
  done:
    $Trace->exit();

} # end delete_topology

# This method will either search a topology for a given key and value
#  or a list of key:value pairs. The default is the Topology located
#  inside the current object, but it will also search one that is passed
sub search_topology {
    my $self = shift;
    my (%params) = validate_mixed(
				  @_,
				  [ ],
				  {
				      cmd_node => {type => SCALAR},
				      key_pairs => {type => ARRAYREF | UNDEF, optional => 1},
				      key => {type => SCALAR | UNDEF, optional =>1 },
				      value => {type => SCALAR | UNDEF , optional =>1 },
				      topology => {type => HASHREF | UNDEF, optional => 1},
				  }
				  );
    my @matches;
    $Trace->entry();

    # figure out which topology to use
    my $topology = $params{topology};
    if (!$topology) {
	# one was not specified, so use the current one
        $topology = $self->get_topology();
	# if the current one is empty, discover it but only for the given cmd_node
	if (!exists($topology->{cluster})) { # this is not complete
	    $topology = $self->discover_topology(cmd_nodes=>[$params{cmd_node}]);
	}
    } 

    # create a temporary combined topology that we will search
    my %tmp_topology;
    foreach my $cmd_node (keys %{$topology->{cluster}}) {
        $tmp_topology{$cmd_node} = [@{$topology->{cluster}{$cmd_node}}];
    }
    foreach my $node (keys %{$topology->{node}}) {
        foreach my $cmd_node (keys %{$topology->{node}{$node}}) {
            $tmp_topology{$cmd_node} = [] if (!$tmp_topology{$cmd_node});
            if ($topology->{node}{$node}{$cmd_node}) {
                push(@{$tmp_topology{$cmd_node}}, @{$topology->{node}{$node}{$cmd_node}});
            }
        }
    }

    # if the cmd_node exists then search it, otherwise just return nothing
    if (exists($tmp_topology{$params{cmd_node}})) {
	# look at each resource for this command node
	foreach my $entry (@{$tmp_topology{$params{cmd_node}}}) {
	    # if an input key was specified, see if this resource matches it
	    if ($params{key}) {
		if ($entry->{$params{key}} eq $params{value}) {
		    push(@matches, $entry);
		}
	    } else {
		# an input key was not passed, so use the key_pairs
		# Look at each key_pair to see if it exists in this resource.
		# If it does, then increment the match param.
		my $match = 0;
		foreach my $pairing (@{$params{key_pairs}}) {
		    my ($key, $value) = split(/:/, $pairing);
                    my $negation = undef;
                    if ($value =~ /^!(.+)/) {
                        $value = $1;
                        $negation = 1;
                    }
		    if (!$negation && ($entry->{$key} eq $value)) {
			$match++;
		    } elsif ($negation && ($entry->{$key} ne $value)) {
			$match++;
                    }
		}
		# only push this resource onto the list if all of the specified
		#  key/value pairs were found
		if ($match == scalar @{$params{key_pairs}}) {
		    push(@matches, $entry);
		}
	    }
	}
    }
    $Trace->exit();
    return @matches;
}

sub gsi {
    # Gather support info on each of the servers
    my $self = shift;
    # this can't be a my because it is used by subroutines in here
    our %params = (
		  -Servers => undef,
                
		  @_, );
    my @servers;
    my @args = @_;
    $Trace->entry(msg=>"Entered with -Servers set to:  $self->{-Servers}");

    # make sure the server(s) passed in make sense
    if ( $params{-Servers} ) {
	if (ref($params{-Servers}) ne "ARRAY") {
	    $Trace->fatal(msg=>"Passed a Server parameter that was not an array reference",
			  result=>"SCRIPT");
	} else {
	    @servers = @{$params{-Servers}};
	}
    } else {
	$Trace->debug( msg=>"Passed an empty Server list, using internal servers() "
		       . "method instead");
	@servers = $self->servers();
    }

    # Call OntapNG->gsi() on all the servers in the @servers array
    my $gather_attempts = 0;
    my $gather_limit = 3;
    my @server_success = ();
    my $no_unknown_errors = 1;

    # Rather than cycling through each node one at a time, let's fire off a subtest for
    # each node that will handle the processing for that node.
    my @transit_subtests;
    my %subtest_to_host_map;
	SERVERLOOP: foreach our $servername (@servers){
		$Trace->comment(msg=>"Servername=$servername");

		our $clustered = 0;
		my $mode;
		my $platform;
		my $hosttype = $Tharn::Hosts{$servername}->hosttype();
		my @prompt_norm = $Tharn::Hosts{$servername}->prompt_norm();
		$Trace->comment(msg=>"BEFORE\n***$hosttype***");
		$Trace->debug(msg=>"***\n".Dumper(@prompt_norm)."***");
		if ($hosttype =~ /sim/) {
			$platform = "vsim";
		} elsif ($hosttype =~ /filer/){
			$platform = "cfe";
		}		

		my $max_retry=2;
		my $cur_retry=0;
		while ($cur_retry < $max_retry) {
			$cur_retry++;
			$self->set_execution_server(-Server => $servername);
			$mode = Tharn::param("MODE");
		        if (defined($mode)){
				$Trace->comment(msg=>"Mode is defined as $mode, so no mode detection, considering the node as $mode node");
				$self->_setup_login_credentials(-Server =>$servername);
                        } else {
				$Trace->comment(msg=>"Trying to detect the mode of the node as Mode is not defined");
				$mode = $self->mode_detector(node => $servername,);
			}
			if (!$mode) {
				$Trace->warn(msg=>"Failed to detect the mode of '$servername' ".
								  "Let's try and cold boot the node and then detect mode.\n".
								  "$cur_retry out of $max_retry!");
				my $state=_get_state($servername);
				my $gdb_on_panic = Tharn::param("DUMP");
                                if ($state =~ /HANG|UNKNOWN/ && defined($gdb_on_panic) && $gdb_on_panic==0) {
                                	$Trace->warn(msg=>"Seems like the system is in GDB_ON_PANIC as STATE: $state GDB_ON_PANIC:$gdb_on_panic ".
							   "This will cause gsi to not gather your files because we don.t want to mess with the ".
							   "system that you want to gdb (debug)");
                               		exit();
				}
				if ($params{hold_jig} ) {
                                	$Trace->warn(msg=>"HOLD_JIG requested, so not going to reboot, " .
                                        "skipping machine: $servername");
					 next SERVERLOOP;
                                }

				eval {
                    	        	_coldboot_this_node($servername);
	
				};
				next;
			} else {
				delete $Tharn::Hosts{$servername};
				if ($mode == 10) {
					main::host("$servername;hosttype=filer-ng-$platform");
				} elsif ($mode == 7) {
					main::host("$servername;hosttype=filer-$platform");
				}
				delete $Tharn::Connections{$servername."#console"};

				$hosttype = $Tharn::Hosts{$servername}->hosttype();
				@prompt_norm = $Tharn::Hosts{$servername}->prompt_norm();
				$Trace->comment(msg=>"AFTER\n***$hosttype***");
				$Trace->debug(msg=>"***\n".Dumper(@prompt_norm)."***");

				if ($hosttype =~ /ng/) {
					$clustered = 1;
				}
				last;
			}
		}

		my $state;

		sub _get_state {
			my $lservername=shift;
			my $panicstr;
			my $lstate;

			# The state method will fatal if it sees a panic or a dumpcore. We don't want
			# to stop just because there was a fatal. If we do, we will never try to get 
			# a core.
			eval {
				$lstate = $Servers{$lservername}->agnostic_state(panic_string => \$panicstr,
																 node => $lservername,
																 clustered => $clustered,);
			};

			if (($@)&&($panicstr)) {
				$Trace->warn(msg=>"Filer appears to have hit a panic, $panicstr");
				$lstate = "HANG";
			}
			if (($@)&&!($panicstr)) {
				$Trace->warn(msg=>"The state of the filer returned an error. The error is $@");
				$lstate = "HANG";
			}
			# If HOLD_JIG was requested, and the machine is hung, skip it.  Otherwise
			# proceed normally.
			if ( ($lstate eq 'HANG') && $params{hold_jig} ) {
				$Trace->comment(msg=>"HOLD_JIG requested, and machine is found to be HUNG, " .
					"skipping machine: $servername");
				next SERVERLOOP;
			}
			# If we have a down'ed node, only GATHER makes sense
			if ( ($lstate !~ /SHELL|LOGIN/) && ($params{operation} ne "GATHER") ) {
				$Trace->warn(msg=>"Can't perform operation $params{operation} on non-operational node");
				next SERVERLOOP;
			}
			return $lstate;
		}

		$state=_get_state($servername);
		
		print ")"x80 . "\n" . ")"x80 . "\n" ."Node=" . $servername ." State=" . $state . " Mode=".$mode."\n" . ")"x80 . "\n" . ")"x80 . "\n";
	       
		 sub _maint_boot_this_node {
                        my ($servername) = @_;
                        my $dest_dir=Testinfo::dir();
                        my $dest_server =$params{dest_server};
			_coldboot_this_node($servername);
		        my $state=_get_state($servername);
			$Trace->comment(msg=>"After shutdown cold My state is $state");
			my $output;
			my $rc ;
                        $Trace->comment(msg=>"Attempting to go to maintenence mode...");
                        if ($state =~ /CFE/){  
                            $Servers{$servername}->agnostic_transit_to(state => $Servers{$servername}->IBM(),
                                                                                                   no_connections => 1,
                                                                                                   node => $servername,
                                                                                                   clustered => $clustered,);
                        }
                        my $cmd = 5;
                        eval {
                             $Servers{$servername}->console(cmd=>$cmd,
                                           additional_prompts_timeout => "999",
                                           additional_prompts      => {
                                                  "Continue with boot?"     => "y",
                                                  "DUMPCORE"                => "FATAL_PROMPT",
                                                  "(unconnected)"           =>"FATAL_PROMPT",
                                                   "Last login:"            => "FINAL_PROMPT",
                                                   "^#"                     => "FINAL_PROMPT",
                                                   "^$servername#"          => "FINAL_PROMPT",
                                                   "^$servername>"          => "FINAL_PROMPT",
                                                   "^$servername\\*>"       => "FINAL_PROMPT",
                                                    "^\\>"                  => "FINAL_PROMPT",
                                                    "^\\*>"                 => "systemshell",
                                                    "login:"                => "diag",
                                                    "Password:"             => "netapp1!",
                                                    "^$servername%"         => "FINAL_PROMPT",
                                        });
                         };
			 $state=_get_state($servername);
                         if ($state =~ /SHELL/) {
                            $Trace->comment(msg=>"Trying to copy cores.....");
		            my $cmd;
			    my $path;
			    if ($hosttype =~ /sim/i) {
			        $cmd = "ls /sim/kcores";
				$path ="/sim/kcores";
			    } else {
			        $cmd = "ls /mroot/etc/crash";
				$path = "/mroot/etc/crash";
			    }				    
                            eval {
			    $output = $Servers{$servername}->console(cmd=>"$cmd", additional_prompts_timeout => "120");
			    };
		            if ($output =~ /No such file or directory/){

				$Trace->comment(msg=>" Tried to get cores, but failed beacuse of $@ ..");
				return $state;
			    } else { 		
                               my @core_files = split(/\n/, $output);
			       $Trace->comment(msg=>"Found cores @core_files");
			       if (@core_files) {	
                               foreach my $core (@core_files) {
                                 if($core =~ /saving/) {
				    $Trace->comment(msg=>"Found cores, but still saving so will sleep for 4 mins.");
                                    sleep(240);
                                 } elsif ($core =~ /^(core)/) {
				    $Trace->comment(msg=>"Trying to copy the cores to dest_dir");
				    $Servers{$servername}->console(cmd=>"cd $path", additional_prompts_timeout => "10");
				    my $cmd = "sudo scp -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no  $core root\@$dest_server\:$dest_dir\/$servername" ;
                                    eval {
					$Servers{$servername}->console(cmd=>$cmd, additional_prompts_timeout => "120",
									           additional_prompts      => {
										   "password:"     => "sm0ke", 
									           "$servername%" => "FINAL_PROMPT",});						
				    };
                                    if ($@) {
					$Trace->comment(msg=>"Failed to get copy cores: $@");
                                        return($state);
                                    } else  {
					$Trace->comment(msg=>"Copied the cores sucessfully, so exiting GSI");
				        Tharn::logresult("PARAM", "code=1");
					exit();
                                    }
                                }
                             }
			     } else { $Trace->comment(msg=>"Booted in Mant and tried to get cores, but there are no cores found in $path, seems like panic did not leave the cores, so exciting GSIwith logresult as pass"); 
					Tharn::logresult("PARAM", "code=1");
					exit(); }
			  }
                        }
			$Trace->comment(msg=>"Tried booting in maint mode, but did not reach to shell state");
                        return($state);
                }


 
		sub _giveback_this_node {

                      my ($servername,$state,$mode) = @_;

		      my $partner_node ;
                      my $orig_node ;
	              my $state_partner_node ;
               	      $Trace->comment(msg=>"@servers");
                      if ($servername eq $servers[0]){
                          $partner_node = $servers[1];
                       } else {$partner_node = $servers[0];}
                       $Trace->comment(msg=>"I am $servername, My partner Node is $partner_node, and My state is $state, and My mode is $mode");

		       $Trace->comment(msg=>"Checking to see if the systems are in takeover/giveback state");
		       eval {
                               $Servers{$servername}->console(cmd=>"\r",
                                               additional_prompts_timeout => "30",
                                               additional_prompts => {
                                                    "Waiting for giveback...(Press Ctrl-C to abort wait)" => "FATAL_PROMPT",});
			 };
                         if ($@){
			      $Trace->comment(msg=>"System is in give back state, will sleep for 2 mins");
                              sleep(120);
			      $Trace->comment(msg=>"After 4 mins of sleep, now try to get the state of partner node");
			      $orig_node = $servername;
			      $servername = $partner_node;
			      $self->set_execution_server(-Server => $servername);
			      $state_partner_node = _get_state($servername);
                              
			      $Trace->comment(msg=>"State of the partner Node is $state_partner_node, if it is not CLI, try to transit to CLI");
			      if ($state_partner_node !~ /CLI/) {
			        $Servers{$servername}->{RR_FLAG} = 1 ;
				$Servers{$servername}->agnostic_transit_to(state => $Servers{$servername}->CLI(),
                                                                                                       no_connections => 1,
                                                                                                       node => $servername,
                                                                                                       clustered => $clustered,);

			      
			      }
			      $state_partner_node=_get_state($servername);	
			      $Trace->comment(msg=>"The state of the partner node is $state_partner_node, and if state is CLI trying giveback");
                              if ($state_partner_node =~ /CLI/) { 
		                  $Trace->comment(msg=>"Trying giveback now...");
                                  eval {
                                         my $output = $self->console(cmd=>"storage failover show", additional_prompts_timeout => "10");
                                          if ($output =~ /giveback/i){
                                              my $cmd  = "storage failover giveback -ofnode $orig_node";
                                              my %prompts =
                                                  ("waiting for giveback state? {y|n}:"         => "y",);

                                               $self->console(cmd=>$cmd, 
                                                                      additional_prompts_timeout => "240",
                                                                      additional_prompts =>{%prompts}, );
					       sleep(240);
                                           }
                                   };
                                  if ($@ =~ /Error: command failed/) {
                                      $Trace->warn(msg=>"Giveback failed!! $@");
		                      $self->console(cmd=>"storage failover show", additional_prompts_timeout => "10");
                                  } else {
				       $Trace->comment(msg=>"Giveback sucessfull");
				  }
			      }
			  $self->set_execution_server(-Server => $orig_node);	
                          $Trace->comment(msg=>"back to the orignal server");	 
		          $state=_get_state($orig_node);
			  $servername = $orig_node;
			  return ($servername,$state);
			  
			
		   }    
                }	

		sub _coldboot_this_node {
			my $lservername=shift;
			$Trace->trace(msg=>"Resetting '$servername' via gx_boot's shutdown_cold task. (no servername)");

			main::subtest( 
				"os/boot/gx_boot.thpl",
				-runid=>"$lservername"."_gxboot_shutdown_cold", 
				-client=>"localhost", 
				-onexit, sub {},
				"NODES=$lservername", 
				"TASKS=shutdown_cold",
				"NO_ADD_TASKS=1",
                "GX_BOOT_RETRY=0",
				"PING=0",
				"CLUSTERED_ONTAP=$clustered",
			); 

		}


               sub _netboot_this_node {
                        my $lservername=shift;
                        $Trace->trace(msg=>"Resetting '$servername' via gx_boot's netboot task. (no servername)");
			 my $netboot_image;
                         my $netboot_param;

                        if (defined (Tharn::param("BUILDROOT"))) {
                            if (Tharn::param("BUILDROOT") =~ /prefetch/) {
                                $netboot_image = Tharn::param("BUILDROOT");
                                $netboot_param = "NETBOOT_UNTAR_DIR";
                            }
                        }

                        if (defined(Tharn::param("NETBOOT_IMAGE"))) {
                           $netboot_image = Tharn::param("NETBOOT_IMAGE");
                           $netboot_param = "NETBOOT_IMAGE";
                        }

                        if (defined($netboot_param) && defined($netboot_param)){
                                main::subtest(
                                        "os/boot/gx_boot.thpl",
                                        -runid=>"$lservername"."_gxboot_netboot",
                                        -client=>"localhost",
                                        -onexit, sub {},
                                        "NODES=$lservername",
                                        "TASKS=netboot",
                                        "NO_ADD_TASKS=1",
                                        "$netboot_param=$netboot_image",
                                        "GX_BOOT_RETRY=0",
                                        "PING=0",
                                        "CLUSTERED_ONTAP=$clustered",
                                );
                        } else {
                                main::subtest(
                                        "os/boot/gx_boot.thpl",
                                        -runid=>"$lservername"."_gxboot_netboot",
                                        -client=>"localhost",
                                        -onexit, sub {},
                                        "NODES=$lservername",
                                        "TASKS=netboot",
                                        "NO_ADD_TASKS=1",
                                        "GX_BOOT_RETRY=0",
                                        "PING=0",
                                        "CLUSTERED_ONTAP=$clustered",
                                );

                        }

                }


		my $scriptspec = $Tharn::Scriptspec;
		# Light up the subtest
		my $st = main::subtest(
			sub {
				# Keep track of the number of times we try to bring the box up.
				my $transit_attempts = 0;
				my $expect_core = 0;
				my $keep_trying=1;

				while (($keep_trying) and ($transit_attempts < $gather_limit) and ($Servers{$servername}->agnostic_state(node => $servername, clustered => $clustered) !~ /SHELL|SINGLEUSER|DBLADECLI|CLI/)) {


					# Try to bring the box up to the shell a couple of times
					$Trace->comment(msg => "Transitting to SHELL attempt $transit_attempts of $gather_limit");
					$transit_attempts++;
					$Servers{$servername}->{RR_FLAG} = $self->{RR_FLAG} if $self->{RR_FLAG};

					if ( $state =~ /^(HANG|UNKNOWN)$/ ) {
						# In the HANG case, reset via the RLM and try to get back to CFE.
						$Trace->warn(msg=>"Upon entry into gsi, we found the node in a HUNG state.");
						($servername, $state) = _giveback_this_node($servername,$state,$mode);	
						$Trace->comment(msg=>"My State is $state");			
						if ($state =~ /^(HANG|UNKNOWN)$/) {
							_coldboot_this_node($servername); 
						}
					}
					elsif ( $state eq 'SINGLEUSER' ) {
						#We might end up here in the end, but I want to at least try 'normal' mode.  Since transit_to
						#doesn't know how to go from SINGLEUSER to NORMAL, I will stop at IBM along the way.
						$Trace->comment(msg=>"Trying to transit_to IBM since I landed at SINGLEUSER right off the bat...");
						$Servers{$servername}->agnostic_transit_to(state => $Servers{$servername}->IBM(),
														           no_connections => 1,
														           node => $servername,
														           clustered => $clustered,);
					}
					elsif ( $state =~ /CFE|IBM/ ) {
						#If I got here, I assume something bad happened.  I will try some specific steps
						#to get to either shell or maint to collect the information.
						$Trace->trace(msg=>"Trying to transit_to the boot menu...");
						$Servers{$servername}->agnostic_transit_to(state => $Servers{$servername}->IBM(),
																   no_connections => 1,
																   node => $servername,
																   clustered => $clustered,);

						my $error=0;

						$Trace->trace(msg=>"Trying to get to a normal login prompt...");
						#choose the normal option and wait for LOGIN

						my $cmd = 1;
						if ($clustered) {
							$cmd = "normal";
						}
						eval {
							$Servers{$servername}->console(cmd=>$cmd,
								additional_prompts_timeout => "360",
								additional_prompts => {
									"Waiting for send"              => "FATAL_PROMPT",
									"Giving up waiting for mroot"   => "FATAL_PROMPT",
									"This option cannot be chosen"  => "FATAL_PROMPT",
									"DUMPCORE"			=> "FATAL_PROMPT",
									"(unconnected)"		        => "FATAL_PROMPT", 
								        "Please make a selection:"      => "normal",
								        'Selection \(1-\d\)\?'          => "1",
									"Data ONTAP GX"                 => "FINAL_PROMPT",
								        "Data ONTAP"                    => "FINAL_PROMPT",
									"login:"                        => "FINAL_PROMPT",
                                                                        "Password:"                     => "FINAL_PROMPT",
									"^$servername>"                 => "FINAL_PROMPT",
									"^$servername\\*>"              => "FINAL_PROMPT",
									"^\\>"                          => "FINAL_PROMPT",
									"^\\*>"                         => "FINAL_PROMPT",
									"Waiting for giveback...(Press Ctrl-C to abort wait)" => "FATAL_PROMPT",});
						};
						
						if ($@ =~ "failed becasue: Waiting for send") {
							$Trace->warn(msg => "TRIED TO RUN BOOT_MENU:NORMAL BUT NODE IS IN A WAITING FOR SEND-HOME STATE!");
							my $lstate=undef;
							my $count=0;
							my $max=3;
							while (($count<$max) and ($lstate ne "LOGIN")) {
								$count++;
								$Trace->trace(msg=>"Waiting 60 seconds for a login prompt after sendhome ($count of $max) ...");
								sleep 60;
								$lstate=_get_state($servername);
							}
							$error=1 unless ($lstate eq "LOGIN");
						} elsif ($@ =~/^( Waiting for giveback)/) {
							$Trace->warn(msg => "TRIED TO RUN BOOT_MENU:NORMAL BUT NODE IS IN A GIVEBACK STATE!");
							($servername, $state) = _giveback_this_node($servername,$state,$mode);
							if ($state =~ /^(HANG|UNKNOWN)$/) {
								#_coldboot_this_node($servername);
                                                        	$error=-1;
                                                	}
						} elsif ($@ =~ "say: Err=TIMEOUT") {
							$Trace->warn(msg => "\nTRIED TO RUN BOOT_MENU:NORMAL BUT IT TOOK TOO LONG.\n");

							eval {
                                                                 $Servers{$servername}->console(cmd=>"\r\r",
                                                                        additional_prompts_timeout => "15", 
                                                                        additional_prompts => {
                                                                        "DUMPCORE"                      => "FATAL_PROMPT",
                                                                        "(unconnected)"                 => "FATAL_PROMPT",
                                                                        "Please make a selection:"      => "normal",
                                                                        'Selection \(1-\d\)\?'          => "1",
                                                                        "login:"                        => "FINAL_PROMPT",
                                                                        "Password:"                     => "FINAL_PROMPT",
                                                                        "^$servername>"                 => "FINAL_PROMPT",
                                                                        "^$servername\\*>"              => "FINAL_PROMPT",
                                                                        "^\\>"                          => "FINAL_PROMPT",
                                                                        "^\\*>"                         => "FINAL_PROMPT",});

											     
                                                         };
						         if ($@) {
								#_coldboot_this_node($servername);
								$error=1;
							 }else { 
								$error=0;
							 }

						} elsif ($@ =~ "Giving up waiting for mroot") {
							$Trace->warn(msg => "\nTRIED TO RUN BOOT_MENU:NORMAL BUT THE MROOT DID A NO SHOW!: Giving up waiting for mroot.\n");
							#_coldboot_this_node($servername); 
							$error=1;
						} elsif ($@ =~ "This option cannot be chosen") {
							$Trace->warn(msg => "\nTRIED TO RUN BOOT_MENU:NORMAL But, got \"This option cannot be chosen\"");
							$error=1;
						} elsif ($@ =~ "(unconnected)") {
							$Trace->warn(msg => "\nTRIED TO RUN BOOT_MENU:NORMAL But, got \"(unconnected)\"");
							#_coldboot_this_node($servername); 
							$error=1;
						} elsif ($@ =~ "DUMPCORE") {
							$Trace->warn(msg => "\nTRIED TO RUN BOOT_MENU:NORMAL But, got \"DUMPCORE\"");
							#_coldboot_this_node($servername); 
							$error=1;
						}
						if ($error) {
							$Trace->comment(msg => "Trying to boot in Maintance Mode");
							($state) =_maint_boot_this_node($servername);
                                                         if ($state != /SHELL/) {
                                                            $error=1;
                                                         }
						         if ($state =~ /SHELL/){
							    $Trace->comment(msg=>"Systems are booted in maint mode, but failed to see or copy cores, even though we are in SHELL, but ssh as admin won't work");
							    _coldboot_this_node($servername);
							    $error=1;
							 }
						}
						if ($error) {
							if ($hosttype !~ /sim/i) {
								$Trace->warn(msg => "\nTRYING NETBOOT \"");
								_coldboot_this_node($servername);
								_netboot_this_node($servername);
								$cmd = "syncflash";
								my %prompts =
									 ("Are you sure you want to continue? {y|n}:"         => "y",);
	
								eval {
                		                                        $Servers{$servername}->console(cmd=>$cmd,
		                	                                                additional_prompts_timeout => "15", 
											additional_prompts =>{%prompts}, )};
							} else  {
			                                        $Trace->comment(msg => "HOSTTYPE $hosttype");
							}
						}


					} else {
						#Try to just transit.  Hopefully a good state
						#
						$Trace->trace(msg=>"Using transit_to to get to a shell prompt...");
						$Servers{$servername}->agnostic_transit_to(state => $Servers{$servername}->SHELL(),
																   no_connections => 1,
																   node => $servername,
																   clustered => $clustered,);
					}

					$state=_get_state($servername);
					print ")"x80 . "\nEnd of while loop\n" . ")"x80 . "\n" ."State=$state\n" . ")"x80 . "\n" . ")"x80 . "\n";

				} #end of while


				print "("x10 . "FINISHED THE WHILE LOOP WITH STATE '$state' AND ATTEMPTS '$gather_limit'" . ")"x10 . "\n";


				if ($state =~ /^(CFE|HANG|UNKNOWN)$/){
					$expect_core = 1;
				}
				if (($Servers{$servername}->agnostic_state(node => $servername, clustered => $clustered) =~ /SHELL|SINGLEUSER|DBLADECLI|CLI/) && ($mode == 10)) {
                    if (grep(@args,/SNAP/)) {
                        $Servers{$servername}->gsi_snap(@args, server_ref=> $Servers{$servername}, expect_core=>$expect_core, scriptspec => $scriptspec);
                    } elsif (grep(@args,/INSTALL/)) {
                        $Servers{$servername}->gsi_install(@args, server_ref=> $Servers{$servername}, expect_core=>$expect_core, scriptspec => $scriptspec);
                    } else {
                        # fire off the real gsi
                        $Servers{$servername}->gsi_gather(@args, server_ref=> $Servers{$servername}, expect_core=>$expect_core, scriptspec => $scriptspec);
                    }
					# Set a param that indicates that we passed
					Tharn::logresult("PARAM", "code=1");
                } elsif (($Servers{$servername}->agnostic_state(node => $servername, clustered => $clustered) =~ /DBLADECLI|CLI/) && ($mode == 7)) {
                    if (grep(@args,/SNAP|INSTALL/)) {
                        $Trace->fatal(msg=>"Gsi can't be told to do OP=INSTALL or OP=SNAP if the node is in 7mode!",result=>"CONFIG");
                    }
                    $Trace->comment(msg=>"Transit.pm says we are in DBLADECLI, but is this a 7mode node or a 10mode node in dbladecli?");
                    my $mode7_ip = $Tharn::Hosts{$servername}->default_ip_7g();
                    my $nfs_ping = Net::Ping->new();
                    $nfs_ping->{port_num} = 2049;
                    $nfs_ping->service_check(1);
                    if ($nfs_ping->ping($mode7_ip, 5)) {
                        $Trace->comment(msg=>"7mode IP '$mode7_ip' of node '$servername' is pingable on NFS port 2049.");
                        push(@args,'mount_ip',"$mode7_ip");
                        #switch the GATHER op with NFS_GATHER op
                        map(s/GATHER/NFS_GATHER/,@args);
                        $Trace->trace(msg=>"\@ARGS:\n@args");
                        $Servers{$servername}->gsi_nfs_gather(@args, server_ref=> $Servers{$servername}, expect_core=>$expect_core, scriptspec => $scriptspec);
                        Tharn::logresult("PARAM", "code=1");
                    } else {
                        $Trace->warn(msg=>"7mode ip '$mode7_ip' is not pingable on NFS port 2049!");
                        # Set a param that indicates that we failed
                        Tharn::logresult("PARAM", "code=0");
                    }
				} else {
					# Set a param that indicates that we failed
					Tharn::logresult("PARAM", "code=0");
				}
			},
			-bg,
			-runid, $servername."_gsi",
			-onexit, sub {},
		); #end of subtest
		# Save the subtest handle
		push(@transit_subtests, $st);
		# And the name of the server that corresponds to the subteest
		$subtest_to_host_map{$st} = $servername;
	}

    if (@transit_subtests){
	# Wait for all of the background reboots to finish
	Subtest::wait_finish(subtests => \@transit_subtests);
	foreach my $st (@transit_subtests){
	    my $exit_status = undef;
	    my @results = $st->results();
	    foreach my $result (@results){
		if (($result->type() eq "PARAM") && ((my $message = ($result->messages())[0]) =~ /^code/)){
		    $Trace->comment(msg => $result->type() . " => $message");
		    my ($key, $value) = split(/=/, $message);
		    $exit_status = $value;
		    if ($value == 1){
			push @server_success, $subtest_to_host_map{$st};
		    }
		}
	    }
	    NATE::Result::store(@results);

	    # The following is an internal Tharn sub that does what we want with the results.  Should
	    # talk to them about making it non-internal or allowing the default exit handler not kill
	    # all subtests when one fatals (even though fatals are abortive....at least in some cases)
	    $st->_relog_severe();
	    $Trace->comment(msg => "Exit status for $subtest_to_host_map{$st} is $exit_status");
	}
    }

    $Trace->exit();
    return @server_success;
}


#***********************************************************************
#
# This method is used to automatically load new methods which are not defined
#   in the namespace. It basically just takes the name of the method being
#   called and passes it into RunCommand as a special paramater. It has to
#   be called via an anonymous function to work correctly.
#
#***********************************************************************

sub AUTOLOAD {
    my $method = our $AUTOLOAD;
    $Trace->entry(msg=>"method = $method");
    if ($method =~ /::DESTROY/) {
	$Trace->exit();
	return;
    }

    no strict 'refs';
    my $self = shift @_;
    my %params = (
		  -Server => undef, # intentionally capitalized
		  @_,
		  );
    $method =~ s/.*:://; # trim the package name

    # should a random server be picked?
    if ($Options->get(option=>"RANDOM_NODE")) {
        $Current_server = (@Server_array)[int rand(@Server_array)];
    }

    # find the name of the server we are talking to
    my $server = $params{-Server} || $params{Server} || $Current_server;

    my $host_object = main::host($server);
    $server = $host_object->hostname();

    # complain if this is not a known server
    if (!grep($server eq $_, @Server_array)) {
	$Trace->fatal(msg=>"did not find server '$server' in the server array for method '$method'", 
		      result=>"CONFIG");
    }

    my $object = $self->_get_server_object(server=>$server, method=>$method);

    delete $params{-Server};

    $Trace->exit(msg=>"Server::AUTOLOAD calling $method in $object for $server");

	if ($method =~ m/^debug_smdb_table_(\w+|_)+_(show|run|create|modify|delete)$/) {
	   return $object->_exec_debug_smdb_table(
		   table => $1,
		   operation => $2,
		   %params
	   );
   }	
    return $object->$method(%params);
} # end AUTOLOAD

sub Servers {
    my $self = shift;
    print "*** The Servers() method is being deprecated in the next release of the Server object.\n";
    print "*** Please switch to calling servers() instead.\n";
    return @Server_array;
}

sub getCLIConnection {
    my $self = shift;
    my %args = @_;
    print "*** The getCLIConnection(host=>) method is being deprecated in the next release of the Server object.\n";
    print "*** Please switch to calling get_cli_connection(server=>) instead.\n";
    $self->get_cli_connection(server=>$args{host});
}

sub SetExecutionServer {
    my $self = shift;
    my %args = @_;
    print "*** The SetExecutionServer(Server=>) method is being deprecated in the next release of the Server object.\n";
    print "*** Please switch to calling set_execution_server(server=>) instead.\n";
    $self->set_execution_server(server=>$args{Server});
}

#***********************************************************************

=head2 add_nodes

Add additional nodes to a  Server object

    $server_obj->add_nodes([nodes => <node1[,node2,...]>],
			   [nodes_on_demand => <1 | 0>]);

This is a function for adding new nodes to an existing server objects.  Given a
node via the C<nodes> parameter it will add that node to the Server object and
make it available for us by Server.  This should only be needed if someone else
has create the Server object there are nodes you need to talk to which were 
either unavailable during the inital Server->new() call, or were not detected
because of incompatible use of FIND_ALL_NODES=false and what your code is 
expecting.

=over 4

=item nodes

Specifies a list of nodes which will be added to the Server object.  Accepts 
a comma seperated list of nodes or an array of nodes.  In this context we
are expecting a nodename, not a fully qualified hostname or a NATE hostspec.

=item nodes_on_demand (boolean - default is false)

Do not do discovery on newly added nodes. The discovery phase will be run 
as they are used (via set_execution_server, or using the '-Server' option 
on other autoloaded calls, as in $server_obj->volume_show(-Server => 'node1').
Default is false (0), meaning that newly added nodes will be discovered as added.
By setting this to true (1), you can significantly speed up the startup time when you
have specified a larger number of nodes. This param is ignored if the node being
added is a 7G/7Mode node.  It can also be toggled via the param NODES_ON_DEMAND 
as well.  If NODES_ON_DEMAND is set globally, you will not override this setting
via the nodes_on_demand parameter for this method.

=back

=cut

# TODO - This code should probably handle being given hostspecs and not just
# short nodenames.  But until we lay out what we are doing for RR where we'll
# need to stop thinking about MODES, this will suffice.

sub add_nodes {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Server' } ],
        {
            nodes => { type => SCALAR | ARRAYREF},
            nodes_on_demand => { 
                type        => BOOLEAN,
                optional    => 1,
                default     => undef,
            },
        },
    );

    $Trace->entry();

    # the global option for NODES_ON_DEMAND will override the locally called version.  this
    # matches the function of Server->new()
    my $on_demand = $Options->get(option => "NODES_ON_DEMAND") || $params{nodes_on_demand};
    # if someone passes in something we use that, otherwise whatever is set in the server_object
    # already is what we will use
    my @current_nodes = $self->servers();


    # handle the nodes we got in.  be nice and accept a comma seperated list, or an arrayref
    my @nodes;
    if ( defined $params{nodes} ) {
        if ( ref($params{nodes}) eq "ARRAY" ) {
            @nodes = @{$params{nodes}};
        } else {
            @nodes = split(/,/, $params{nodes});
        }
    }
    # if we don't have a node to add - throw an exception
    if ( scalar @nodes == 0 ) {
        $Trace->fatal(msg => "Unable to determine nodes that are to be added.", result => "SCRIPT");
    }

    foreach my $node ( @nodes ) {
        if ( grep /$node/, @current_nodes ) {
            # don't bother adding a node that we already know about
            $Trace->trace(msg => "Asked to add a node, '$node', which is already known to Server.  Nothing to do for this node.");
            next;
        } elsif ( grep /$node/, $Options->get(option => "SKIP_NODES") ) {
            # make sure this node isn't on the SKIP_NODES list that we have
            $Trace->fatal(msg => "Asked to add a node, '$node', which is on the SKIP_NODES list.  This is illegal.", result => "CONFIG");
        }
        # add the node to the internal list
        $Trace->trace(msg => "Adding node '$node' to server list.");
        my $node_type = $self->_add_host_to_lists($node);
        # if the node added is a 7G node, we have to run discovery on it
        if ( $node_type eq "node" ) {
            $Trace->debug(msg => "Node '$node' is a 7G/7Mode node - calling discovery for it.");
            $self->discover(server => $node,_no_cli => $self->{no_cli});  # no_cli is ignored for a 7G node
        }

    }

    # run discovery on the nodes unless we are in on_demand mode
    if ( !$on_demand ) {
        # this will be a no-op if its already been done, so do it on all the nodes
        # rather than trying to keep track of who's new or old
        foreach my $node ( $self->servers() ) {
            $self->discover(server => $node, _no_cli => $self->{no_cli});
        }
    }

    $Trace->exit();
}

#***********************************************************************

=head2 find_all_nodes

Discover all the nodes in the same cluster as the nodes we already know
about, or just the ones passed in to it.

        $server_obj->find_all_nodes(nodes => "NODE1");

This function will find all the nodes in a cluster with the nodes passed
in to it.  By default it will look for all the clustered nodes that the
object knows about and will look for all the clusters it encounters and 
find all the nodes from all the clusters.  Any node that is defined in the
SKIP_NODES global Option will be skipped and not added.  Otherwise all
node are added.  The function is optimized so that we don't accidently
cause a discovery on a node by connecting to it unecessarily.  If we have
seen the node name in the cluster already, we don't connect to it and
ask it about its cluster partners.

=over 4

=item nodes

Specifies a list of nodes to discover the cluster members of.  Defaults
to all the clustered nodes that the Server object currently knows about.

=back

=cut

sub find_all_nodes {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Server' } ],
        {
		    nodes => { 
                        type     => SCALAR|ARRAYREF,
                        default  => undef,
                        optional => 1,
                    },
        },
    );

    $Trace->entry();

    # we'll need to set this back later
    my $orig_server = $self->get_execution_server();
    my @found_nodes;
    my $skip_nodes = $Options->get(option => "SKIP_NODES");

    # handle the nodes we got in.  be nice and accept a comma seperated list, or an arrayref
    my @nodes;
    if ( defined $params{nodes} ) {
        if ( ref $params{nodes} eq "ARRAY" ) {
            @nodes = $params{nodes};
        } else {
            @nodes = split /\,/, $params{nodes};
        }
    } else {
        @nodes = $self->get_clustered_nodes();
    }

    # if we don't have any nodes to work with - throw an exception
    if ( scalar @nodes == 0 ) {
        $Trace->fatal(msg => "Unable to determine nodes that are to be used to run against.", result => "SCRIPT");
    }

    # walk the list of nodes we were given
    foreach my $seed_node ( @nodes ) {
        # if we've already seen this node and added it, we can skip to the next node
        next if ( grep /$seed_node/, @found_nodes );

        $Trace->debug(msg => "Checking for other nodes in a cluster with node '$seed_node'");
        $self->set_execution_server(-Server => $seed_node);
        my @cluster = $self->_find_all_servers();

        # walk through the nodes we have and process them
        foreach my $node ( @cluster ) {
            if (grep($node eq $_, @{$skip_nodes})) {
                $Trace->trace(msg=>"Skipping $node because it exists in SKIP_NODES");
                next;
            }
            # push it onto our list so we won't look for it, and add it to the internal list
            push @found_nodes, $node;
            $self->add_nodes(nodes => $node);
        }
    }

    # set the execution server back to its original setting
    $self->set_execution_server(-Server => $orig_server);

    $Trace->exit();
} # end find_all_nodes

# this routine will take a server/hostspec and add it to the appropriate lists
# returns the type of host that was added to the object
sub _add_host_to_lists {
    my $self = shift;
    my ($hostspec) = @_;
    my $hosttype = "";
    
    $Trace->entry(msg=>"server = $hostspec");

    # call host on the entry to properly parse any hostspec
    my $host_object = main::host($hostspec);

    # find the name of the actual node
    my $hostname = $host_object->hostname();

    # discover the hosttype and store it
    my $type = $host_object->hosttype();
    if ($type =~ /filer-ng|vsim-ng/) {
        $hosttype = "clustered";
        $self->{types}{clustered}{$hostname} = 1;
    } elsif ($type =~ /filer|vsim/) {
        $hosttype = "node";
        $self->{types}{node}{$hostname} = 1;
    }

    # store the hostspec in case it is needed later
    $Hostspecs{$hostname} = $hostspec;

    # add this hostname to @Server_array if it is not already there
    if (!grep($_ eq $hostname, @Server_array)) {
        push(@Server_array, $hostname);
    }

    $Trace->exit();
    return $hosttype;

} # end _add_host_to_lists

# this routine is used to get a list of all nodes, including
#  one clustered and all legacy nodes. This code will pay attention to
#  the RANDOM_NODE option as well when picking a clustered node to use
sub _get_nodes {
    my $self = shift;
    my %params = @_;
    my $options = $Options->get_hash();

    my @nodes;
    my @clustered = $self->get_clustered_nodes();
    if (@clustered) {
        if ($options->{RANDOM_NODE}) {
            @nodes = ($clustered[int rand(@clustered)]);
        } else {
            @nodes = ($clustered[0]);
        }
    }
    if ($params{reverse_legacy}) {
        push(@nodes, reverse $self->get_legacy_nodes());
    } else {
        push(@nodes, $self->get_legacy_nodes());
    }
    return @nodes;
}

# This routine will return the object handle for the specified server name.
# This routine is intentionally marked as internal because I really do not
#  want to provide the direct handle back to the end user yet because it
#  could cause unexpected behaviour. We should force interaction via this
#  object.
# server: the name of the server
# method: the name of the method that is asking for the object
# returns the object handle on success
sub _get_server_object {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'Server' } ],
					 {
                                          server => {type => SCALAR},
                                          method => {type => SCALAR, optional => 1},
					 }
                                        );
    my $server = $params{server};
    
    # Locate the object for this server.
    # If it is not already known (and stored) then call discover on it. If it returns NULL,
    #  then die. More than likely, the underlying discover code will die with a more useful
    #  error so this fatal will not hit.
    my $object = $Servers{$server};
    if (!$object && !($object = $self->discover(server=>$server))) {
        my @caller = caller(1);
        my $calling_method = $caller[3];
        my $error_msg = "The method '$calling_method' was trying to find the object handle for server '$server'";
        if ($params{method}) {
            $error_msg .= " (on behalf of $params{method})";
        }
	$Trace->fatal(msg=>$error_msg, result=>"CONFIG");
    }
    return $object;
}

sub _discover {
    my $self = shift;
    print "*** The _discover() method is being deprecated in the next release of the Server object.\n";
    print "*** Please switch to calling discover() instead.\n";
    $self->discover(@_);
}

1;

=head2 cli

Issue raw cli commands.  This method returns a string which contains
raw output from the ngsh command, just as you would see it if you
issued the command by hand within ngsh.

  $cli_output = $server_obj->cli(cmd => <raw cli command>,
				 [output    => <variable reference>],
				 [-Timeout  => <seconds>],
				 [nocr      => <1 | 0>],
				 [poke      => <1 | 0>],
				 [noprompt  => <1 | 0>]);

=over 4

=item cmd

The raw cli command to send to the node.  Example of this is:

  $raw_output = $server_obj->cli(cmd => 'storage aggregate show');
  print "Raw output from 'storage aggregate show':\n\n $raw_output\n";

=item other arguments

All the other arguments available to the B<cli> method are described in detail
under the B<Standard Server Arguments> section above.

=back

=head2 shell

Issue FreeBSD shell commands.  This method returns a string which contains
raw output from the shell command, just as you would see it if you
issued the command by hand at the FreeBSD prompt.  Some commands do not
return any output (e.g. 'sysctl debug.panic=1')

  $shell_output = $server_obj->shell(cmd => <FreeBSD command>,
			  	    [output    => <variable reference>],
				    [-Timeout  => <seconds>],
				    [nocr      => <1 | 0>],
				    [poke      => <1 | 0>],
				    [noprompt  => <1 | 0>],
				    [exit_code => <variable reference>]);

=over 4

=item cmd

The raw shell command to send to the node.  Example of this is:

  $shell_output = $server_obj->shell(cmd => 'ls /mroot/log');
  print "Raw shell output from 'ls /mroot/log':\n\n $shell_output\n";

=item other arguments

All the other arguments available to the B<shell> method are described in
detail under the B<Standard Server Arguments> section above.

=back

=head2 rlm

Issue raw RLM commands.  This method returns a string which contains
raw output from the rlm command, just as you would see it if you
issued the command by hand at the RLM prompt.

  $rlm_output = $server_obj->rlm(cmd => <RLM command>,
				 [output    => <variable reference>],
				 [-Timeout  => <seconds>],
				 [nocr      => <1 | 0>],
				 [poke      => <1 | 0>],
				 [noprompt  => <1 | 0>]);

=over 4

=item cmd

The raw rlm command to send to the node.  Example of this is:

  $rlm_output = $server_obj->rlm(cmd => 'events newest 10');
  print "Raw rlm output from 'events newest 10':\n\n $rlm_output\n";

=item other arguments

All the other arguments available to the B<rlm> method are described in
detail under the B<Standard Server Arguments> section above.

=back

=head2 dblade

Issue raw dbladecli commands.  This method returns a string which contains
raw output from the dbladecli, just as you would see it if you
issued the command directly within the dbladecli.

  $dblade_output = $server_obj->dblade(cmd => <dblade command>,
				       [output    => <variable reference>],
				       [-Timeout  => <seconds>],
				       [nocr      => <1 | 0>],
				       [poke      => <1 | 0>],
				       [noprompt  => <1 | 0>]);

=over 4

=item cmd

The raw dbladecli command to send to the node.  Example of this is:

  $dblade_output = $server_obj->dblade(cmd => 'disk show');
  print "Raw dbladecli output from 'disk show':\n\n $dblade_output\n";

=item other arguments

All the other arguments available to the B<rlm> method are described in
detail under the B<Standard Server Arguments> section above.

=back

=head2 dblade_set_privilege

Globally change the privilege level for all dbladecli commands. This 
method sets a global <priv set level> string that will get prepended
to any raw command sent to the dblade method.

    dblade_set_privilege(level=> <admin | advanced | diag | test>, 
                         quiet=> <off | on>);

=over 4

=item level

Sets the privilege level for all dblade commands sent to the dblade method. 
At this time the only options available for level are admin, advanced, diag, 
and test.

=item quiet

Sets quiet mode on|off. 
"on"   suppresses the warnings associated with entering advanced, diag, or test levels.
"off"  allows warnings to be throttled.  

=back

=head2 console

Talk directly to the console.    This method returns a string which contains
raw output from the console, just as you would see it if you
issued the command directly on the console.

  $console_output = $server_obj->console(cmd => <console command>,
					[output    => <variable reference>],
					[-Timeout  => <seconds>],
					[nocr      => <1 | 0>],
					[poke      => <1 | 0>],
					[noprompt  => <1 | 0>]);

Be aware that since you have a connection to the console, you may get
spurious messages in your output that are typically seen on the console.

=over 4

=item cmd

The raw console command to send to the node.  Example of this is:

  $console_output = $server_obj->console(cmd => 'ngsh -c "storage aggregate show"');
  print "Raw console output from 'ngsh command':\n\n $console_output\n";

Using the console, you can access the FreeBSD shell, dbladecli, or ngsh.  Note
that you need to be at the right state for the command you issue to work
properly.  Depending on your current state, you may need to transit to
the proper state before calling the B<console> method.

=item other arguments

All the other arguments available to the B<console> method are described in
detail under the B<Standard Server Arguments> section above.

=back

=head2 cfe

This is syntatic sugar for the console method.  Refer to the B<console>
method description above for details.

Note that the B<cfe> method does not transit to CFE/LOADER.  In order
to use this method and issue console commands to the CFE/LOADER, you must
already have transited to those states.

=head1 Convenience Methods from Other Packages

There are a handful of other packages that provide convenience methods
available directly through a Server object.  These are described in 
detail below.

=head2 Volume

=head3 volume_offline

Offline specified volume.

   $server->volume_offline(volume => <vol name>);

=over 4

=item volume

Volume name you wish to put in the offline state.  Vserver is not necessary,
this method will locate the vserver for the specified volume.

=back

=head3 volume_purge

Ummount, Offline and delete specified volume.

   $server->volume_purge(%params);

The params passed in a scrubbed to only pass in the params that are valid for
the volume_unmount, volume_offline and volume_delete commands.  They aren't a
full pass-through because volume_delete can query on some of the striped fields
on a non-striped volume and fail.

=head3 get_volumes

Find all volume in the cluster that match specified parameters.

   @vols = $server->get_volumes(%params);

%params are pass-through, so whatever parameters are available to the
'volume show' command are valid.

If called in list context, return an array of volume hashes that match
the specified criteria passed in.  If called in scalar context, it will
return the first volume that matches the specified criteria.

=head3 volume_mount_point

Return the junction path for specified volume.

   $mount_point = $server->volume_mount_point(volume => 'myvol');

=over 4

=item volume

Name of volume for which you want to get the mount point.

=back

=head3 is_root_volume

Search all available VServers to determine if specified volume is a root
volume for any of them.

   $is_root_vol = $server->is_root_volume(volume => 'myvol');

Returns 1 if specified volume is a root volume, 0 otherwise.

=over 4

=item volume

Volume name.

=back

=head3 create_volume_from_ip

If anyone uses this method, or has a clue what it really does, please
add documentation for it here.

=head3 wait_on_jobs

   $job_hashref = $server->wait_on_jobs([timeout => <seconds>]);

Wait on volume operation jobs.  Any of the volume operations that are
performed in the background will automatically be placed in an internal
Server volume job queue, which can then be "waited" on by calling
this method.  Issues the 'job show' command in the CLI and waits for
jobs in the job queue to be in a completed state (succeeded, failed, quit,
died, etc.), or until a timeout has occurred.

Returns a reference to a hash that contains job info.  Job hash has the 
following structure:

 $jobhash = {
     <jobid> => {
	 jobid => <job>,
	 completion => <event>,
	 state      => <completion state>,
     }
     ...
 }

=over 4

=item timeout

Optional - specify the amount of time (in seconds) to wait for any
individual job to complete.  Default is 600.

=back

=head3 volumes_vserver

Locate the vserver for first volume matching the specified criteria.
If multiple volumes match the criteria you will get a warning the the
code will return the first volume that it matched.  Not ideal so your
query should be really specific.  Ideally you should always keep track 
of the VServer you are using.

  $server->volumes_vserver(%params);

Returns a vserver name.

=over 4

=item %params

This is a pass-through parameter, so any parameter that is available for
the 'volume show' command is legal.  However, you *must* specify any 
parameter that is required, such as B<volume>.  Since it returns only
the vserver of the first volume, wildcarding the volume name and specifying
other volume parameters will not return what you might think if your search
criteria span multiple vservers.

=back

=head3 find_min_max_volume_size

Determine various sizes that can be used for volume creation.

  $size = $server->find_min_max_volume_size(aggr_name, sizetype);

Returns a number that represents the minperm, maxperm, aggrsize or maxavail,
depending on what you specify for the 'sizetype' parameter.
If the aggregate type is flex, returns the defaults for minperm/maxperm.

=over 4

=item aggr_name

Name of aggregate which you are attempting to get the volume size info for.

=item sizetype

Specifies what will be returned for the output.  Can be one of:

=over 4

=item B<minperm>

=item B<maxperm>

=item B<aggrsize>  (total size of specified aggregate)

=item B<maxavail>  (available size of aggregate)

=back

=back

=head3 volume_size

Get volume size and used size in blocks.

   my($blocksize, $size, $used) = $server->volume_size($volume);

Returns blocksize (currently hardcoded to 4096), size of the volume
(converted to blocks) and the amount (in blocks) used in the volume.

=over 4

=item $volume

Volume name you wish to get the size information for.

=back

=head3 volume_new_and_modify

Create volume and modify the 'files' parameter, since this parameter
is not supported on the vol new call.

  $server->volume_new_and_modify(volume  => <volname>,
				 vserver => <vserver name>,
				 files   => <max files>)

=over 4

=item volume

Name of volume to operate on.

=item vserver

Name of vserver to operate on.

=item files

Max number of files.  The volume will be modified to to increase the max
files to this number after the create is complete.

=back

=head3 pick_aggregates

Search through a list of aggregates and only return the list that match the
specified criteria. Some of the criteria can be specified via method params 
detailed below. Others are specified with NATE params which will override
the method params. 

  @aggr_list = $server->pick_aggregates(state              => <state>,
		   	                voltype            => <volume type>,
                                        voltype_discovery_mode  =>  <discovery mode>,
			                avail_size         => <available aggregate size>,
                                        topology           => <topology hash to use>,
                                        aggregate_list     => <list of aggregates to pick from>,
                                        flexcache_involved => <is a flexcache in the mix>,
					block_type         => <aggregate block type>,
                                        blocktype_discovery_mode => <discovery mode for aggregate block type>)

=over 4

=item state

Optional: This parameter tries to find aggregates with their state flag set to
this value. This type is not checked but should be values such as online,
offline, restricted, etc. Defaults to 'online'.

=item voltype

Optional: This parameter tries to find aggregates with their voltype set to the
specified value. This can be overriden with the C<VOLUME_STYLE> parameter. 
Defaults to 'flex'.

=item voltype_discovery_mode

Optional: This parameter determines the order in which the volume type is 
discovered. This will decide if 'voltype' or VOLUME_STYLE should be used, and
in what order. Defaults to 0.

0 = Look for VOLUME_STYLE first, then 'voltype'

1 = Look for 'voltype' first, then VOLUME_STYLE

2 = Only look at VOLUME_STYLE

3 = Only look at 'voltype'

=item avail_size

Optional: This parameter tries to find aggregates with at least this much space
available. This value may have a modifier, such as mb on it.

=item topology

Optional: If this parameter is specified, the aggregates listed in this hash will
also be searched. Used by L<TestManager>.

=item aggregate_list

Optional: If this parameter is specified, then this list is searched rather than 
the full list of aggregates found during discovery.

=item flexcache_involved

Optional: This parameter specifies that a flexcache is in the configuration. If an
aggregate is found to be on the cache, then it will mark it as so and will skip
any traditional volumes and not treat them as aggregates. Used by L<TestManager>.

=item block_type

Optional: This parameter tries to find aggregates with their block type set to the
specified value this can be overridden with the C<AGGR_BLOCK_TYPE> parameter.
Defaults to undef.

=item blocktype_discovery_mode

Optional:This parameter determines the order in which the block type is
discovered. This will decide if 'block_type' or AGGR_BLOCK_TYPE should be used, and
in what order. Defaults to 0.

0 = Look for AGGR_BLOCK_TYPE first, then 'block_type'

1 = Look for 'block_type' first, then AGGR_BLOCK_TYPE

2 = Only look at AGGR_BLOCK_TYPE

3 = Only look at 'block_type'

=item aggregate 

This NATE parameter specifies a list of aggregate names or regular expressions.
When filtering the aggregates, the list will be reduced to the ones that match
these names.

If the value is a simple string then it is specifying the actual names of 
aggregates. Placing a '!' or '~' in front of the name will cause it to skip
aggregates of that name.

If the value is of the form /string/ then string is interpreted as a perl
regular expression and any aggregate name that matches the regexp will
match.

If the value is of the form !/string/ then string is interpreted as a perl
regular expression and any aggregate name that does not match the regexp will
match.

=item SKIP_MROOT_AGGREGATE

This NATE param indicates that any aggregate hosting an mroot will not match.

=item VOLUME_STYLE

This NATE param specifies that the aggregate volume type must match this value.
This will override the method param 'voltype'.

=item AGGR_BLOCK_TYPE

This NATE param specifies that the aggregate block type must match this value.
This will override the method param 'block_type'.

=back

=head2 Network

From the Network module description:

Network.pm - remote programmatic administration of Network functions

Network is documented in within the Node::NG::Network module.  That
documentation can be found at:

L<http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-Node-NG-Network-pm.html>

=head2 Storage

From the Storage module description:

Storage.pm - remote programmatic administration of Storage functions

Storage is documented within the Node::NG::Storage module.  That documentation
can be found at:

L<http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-Node-NG-Storage-pm.html>

=head2 Transit

From the Transit module description:

This module contains implementations of filer state and state transition methods.

Transi is documented within the Node::NG::Transit module.  That documentation
can be found at:

L<http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-Node-NG-Transit-pm.html>

=head1 PARAMS 

=over

=item NODES

This param defines a list of nodes. The list can be separated by a C<,> 
or a C<:>. The hostspecs in this parameter can be overriden with the value
of nodes passed into Server->new()

=item FIND_ALL_NODES

This param will allow the user to only specify one node name and the
code will attempt to discover the nodes in the cluster and manage them
as well. This has no effect in a non-clustered environment such as 7G.

=item NODES_ON_DEMAND

This param will cause the code to only connect to the first node in the
C<NODES> list upon calling C<new>. It will then connect to any other specified
node as needed. By default, C<new> will connect to all C<NODES> it knows about.
This param does not have an effect in a non-clustered environment.

=item RANDOM_NODE

This param will cause C<new> to define the default node as a random one in the
C<NODES> list rather than the default action of defining the first one in the list
as the default.

=item SKIP_NODES

This param defines a list of nodes to be skipped. The list should be a comma-
separated list of node names. The nodes must exist in the C<NODES> list or
discovered via the C<NODES_ON_DEMAND> param.

=item CLI_USER

This param specifies the username to use when connecting to the cli. This will
override any value found in the nate_host file. 

=item CLI_PASSWORD

This param specifies the password to use when connecting to the cli. This will
override any value found in the nate_host file. If you need to specify an empty
string (the Tricky default), then set this param to the string EMPTY_STRING.

=item CLI_PARAMETER_VALIDATION

This param specifies that the parameters passed to the cdef-generated methods
will be validated prior to sending them to the UI. This is off by default.

=item SHELL_USER

This param specifies the username to use when connecting to the shell. This will
override any value found in the nate_host file. 

=item SHELL_PASSWORD

This param specifies the password to use when connecting to the shell. This will
override any value found in the nate_host file.

=item CLI_FROM_SHELL

This param specifies that all commands sent to the CLI be sent via "ngsh -c" 
commands from the shell prompt.

=item aggregate 

See C<pick_aggregates>.

=item SKIP_MROOT_AGGREGATE

See C<pick_aggregates>.

=item VOLUME_STYLE

See C<pick_aggregates>.

=item CONNECTION_MAXBUFSIZE

Output from say() will be limited to just the last \$maxbufsize bytes. This value is
defaulted to 1MB by NATE. Setting the CONNECTION_MAXBUFSIZE param will specify a new
maxbufsize for the NATE connections to the values specified.

=item LIBSSH_OVERRIDE

When LIBSSH_OVERRIDE is set to "1", it will force the Node::NG library to use the 
libssh-persit connection type for the filer connections.  By default, this PARAM is set
to "0" which uses the existing ssh-persist connection type.  ssh-persist connections can
cause SSH sessions to hang around after the script exits and take up valuable resources
on the filer.

=back

=head3 verify_aggregate_state


              verify_aggregate_state( aggregate   =>$aggregate_name,
                                      nodes       =>$nodes,
                                      wait_for    =>"50", #40 seconds is default
                                      state       =>"online",
                                     );
C<verify_aggregate_state> is used to verify the state of an aggregate is
as expected before moving on to other operations. When an aggregate is created,
there is a duration in which it is in a 'creating' state'. During this state, the
addition of volumes will fail. This method will wait for the duration given, and
poll every 10 seconds for aggregate state. It will trace fatal either if the
converged state is not expected, or if the aggregate remains in 'creating' and
time exhausts.

This method has been generalized for anyone who wants to make sure their aggregate
is in the proper state.

=over 4

=item state

(Required) This is the expected aggregate "target" state. Choices include:
aggr show -state ?
  creating                    creating
  failed                      failed
  offline                     offline
  online                      online
  partial                     partial
  restricted                  restricted
  unknown                     unknown

=item aggregate

(Required) This is the aggregate name.

=item nodes

(Optional) This is the node name.

=item wait_for

(Optional) This is the duration in seconds in which you want to wait for
the aggregate if it is still in a 'creating' state. This number will be
divided by 10 to calculate the number of retries, and perform an aggregate_show
every 10 seconds until either the retries are exhausted, or the aggregate has
changed its state.

=item CMDPROF

     Given a file will enable cli command profiling to dump hash block to the file, 
     default is undef.
     Time values are stored as epoch seconds
     Logic for CMDPROF features is in Node.pm
     Hash block example
     {
          'server'      => 'rtpgx-3050-15',      # name of filer
          'begin'       => 1213733229,           # begin time in epoch seconds
          'end'         => 1213733232,           # end time in epoch seconds
          'cmdtime'     => 3,                    # calculated time (end-begin)
          'begin_micro' => '1213733229.92422',   # begin time in epoch seconds and microseconds
          'end_micro'   => '1213733232.09315',   # end  time in epoch seconds and microseconds
          'cmd'         => 'storage disk show',  # filer command for this profile
          'args'        => '',                   # command arguments
     };
     xxEND_RECORDxx


=item CMDPROF_CSV

     Given a file will enable cli command profiling to dump values to the csv file
     default is undef.
     CSV fields - server,begin,end,cmdtime,begin_micro,end_micro,cmd,args
     Note that the last field may (command arguments) may contain ','.
     Account for this in your parsing of the csv file.
     See CMDPROF for description of values

=item CMDPROF_RT

     If non-zero, will keep N amount of profiles in the server object, default = 0
     All profiles are stored in a list in the server object under the 'cmdprof' key.
     Each profile is an array element, most recent profile is the 0th element of the array.
     The array only keeps N elements as defined by the CMDPROF_RT param. Each element is a hash
     block, as exampled below.
     See CMDPROF for description of values
     $server->{'cmdprof'}[0]{key} = value
     {
          'server'         => 'rtpgx-3050-15', 
          'begin'          => 1213733229,
          'end'            => 1213733232,
          'cmdtime'        => 3,
          'begin_micro'    => '1213733229.92422',
          'end_micro'      => '1213733232.09315',
          'command_retried =>
          'cmd'            => 'storage disk show',
          'args'           => '',
     };

=item RETRY_SAY_FAILURE

If non-zero, specifies the amount of time that a command should be retried. The default is 0.
When this param is specified, Server will continue to retry opening the connection and sending
the command until the specified amount of time has been elapsed. A fatal result is thrown if the
timeout occurs. When using this parameter, only the connection is reestablished; the CDEFS and
version are not reset on the new connection. So, this method is not useful for handling an upgrade.

=item RETRY_SAY_FAILURE_INTERVAL

When C<RETRY_SAY_FAILURE> is set, this param specifies how long to wait in between reconnection
attempts. This value is set to 30 seconds by default.

=back

=head1 SEE ALSO

L<TestManager> defines the C<CACHE_TRAFFIC> parameter.

=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/Server.pm#2 $
  Last revised $Date: 2013/05/24 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut
