#Hey emacs, this is basically -*- perl -*-
# $Id$

#***********************************************************************
#
# NAME
#   Shoc.pm 
#
# SYNOPSIS
#   Shoc (show-output-compare) routines.
#
# DESCRIPTION
#   This library provides routines to compare NACL component states for
#   use in the shoc_save and shoc_diff thpls.
#
# AUTHOR
#   dl-coati-core
#***********************************************************************

=pod

=head1 Library Name

Shoc.pm

=head1 Author

dl-coati-core

=head1 Description

This library contains routines that will compare NACL::CS objects.  How the comparisons
are done are based on Comparison Specifications stored in Shoc::CompareSpecTable.pm.
The entries in a comparison spec (see lib/Shoc/CompareSpecTable.pm) control how various
ComponentStates are compared, which fields are ignored, etc...  

In this library is the compare() routine, that takes a compare spec, along with pointers
to the 2 data structures (generally arrays of component specs, but the routine supports
arrays of hashs (aoh) and hash elements which are arrays and does comparisons on those
lower level items as well.

For array items which are not ComponentState objects, it is assumed that the items are in 
sorted order (or at least the SAME order).  For ComponentState objects, we get the primary
keys from the associated Component object and use those to match up items in arrays.

For objects which are NOT ComponentState objects (for example, if a ComponentState object
has an array of hashes as the value of one of it's hash keys, ex: Aggregate{plexes} in 7Mode)
but we can call "keys" on these objects, we will compare these objects as hashes.  This also
works if the objects are Perl Objects which are essentially blessed hashes.

Fields of a hash (or hash-like object) can be ignored, or routine references can be specified
if the comparison should be something other than the default, which is to compare 2 scalars as
strings.  (see lib/Shoc/CompareSpecTable.pm for examples)

The compare() routine accumulates compare and miscompare information into arrays (references
to which the caller passes in) and returns a 1 for no miscompares, and a 0 for miscompares.
The accumulated compare data returned to the user (in errors and validation arrays) contain
Shoc::CompareResult objects which consist of a status (1|0) and a string.  The string, as
built by compare() attempts to provide an index to the element of the comparison.

For example, let's say that we are comparing an array of NACL::C::Cluster entries (which
have as their primary key, the node).  We would build comparison entry strings like so:

  [node:csiqa-3070-31]{eligibility} "true" == "true"
  [node:csiqa-3070-31]{epsilon} "false" == "false"
  [node:csiqa-3070-31]{health} "true" == "true"
  [node:csiqa-3070-31]{node} "csiqa-3070-31" == "csiqa-3070-31"
  [node:csiqa-3070-31]{uuid} "ce8e44eb-3b74-11e0-a0eb-97a553547fd1" == "ce8e44eb-3b74-11e0-a0eb-97a553547fd1"
  [node:csiqa-3070-32]{eligibility} "true" == "true"
  [node:csiqa-3070-32]{epsilon} "false" == "false"
  [node:csiqa-3070-32]{health} "true" == "true"
  [node:csiqa-3070-32]{node} "csiqa-3070-32" == "csiqa-3070-32"
  [node:csiqa-3070-32]{uuid} "cea13e56-3b74-11e0-a7e9-8710e713b7bd" == "cea13e56-3b74-11e0-a7e9-8710e713b7bd"

The data inside of the square brackets [] represents the primary key index of the ComponentState.  The
string is [key1:value1,key2:value2,...keyN:valueN].  After that is the field name in {}, followed by the
value of that field in the first data structure, followed by a relational operator (==|!=|=~|!~), followed
by the value of that field in the second data structure.  

For an example where a ComponentState field contains an array of strings, the message string will be formatted
like so:

  [aggregate:aggr0]{disklist}[0] "0a.16" == "0a.16"
  [aggregate:aggr0]{disklist}[1] "0a.32" == "0a.32"

For a really involved structure, the 7Mode aggregate "plexes" value is an array of hashes, of arrays, of hashes, etc...

  [aggregate:aggr0]{plexes}[0]{plex-info}[0]{is-online} "true" == "true"
  [aggregate:aggr0]{plexes}[0]{plex-info}[0]{name} "/aggr0/plex0" == "/aggr0/plex0"
  [aggregate:aggr0]{plexes}[0]{plex-info}[0]{raid-groups}[0]{raid-group-info}[0]{disks}[0]{disk-info}[0]{name} "0a.16" == "0a.16"
  [aggregate:aggr0]{plexes}[0]{plex-info}[0]{raid-groups}[0]{raid-group-info}[0]{disks}[0]{disk-info}[1]{name} "0a.32" == "0a.32"
  [aggregate:aggr0]{plexes}[0]{plex-info}[0]{raid-groups}[0]{raid-group-info}[0]{name} "/aggr0/plex0/rg0" == "/aggr0/plex0/rg0"
  [aggregate:aggr0]{plexes}[0]{plex-info}[0]{raid-groups}[0]{raid-group-info}[0]{status} "normal, block checksums" == "normal, block checksums"
  [aggregate:aggr0]{plexes}[0]{plex-info}[0]{status} "online, normal, active, pool0" == "online, normal, active, pool0"

Although it might be possible to infer whether a string is a match or mismatch based on the relational operator, 
you should really only look at the CompareResult status which will be either 0 (mismatch) or 1 (match).  The 
message part of the CompareResult is really just there for displaying to the user.  For an example of how 
these results are handled, please see tools/shoc_diff.thpl.

If you are only concerned with the mismatches, you could just run through the error array, or if you want
to see all comparisons, look through the validation array.

Finally in this file are routines to build the data structures from a list of Components.  We do a find on
each component, then call $component->state() to retrieve the ComponentState object.  Each of those objects
are stored in a hash, keyed by the fully qualified NACL Component name (i.e. NACL::C::Volume).

Also in this hash is stored information regarding the release of the cluster or 7mode node for which we are
saving state.  This information is stored so that at a later date we can run the tools/shoc_diff.thpl against
these files and there will be enough information stored in the SHOC files so that we will know how to do
the comparisons.  Even when we are dealing with 2 SHOC files from the same release, certain fields will not
match (uptime, usedsize, etc...) and when we are dealing with 2 SHOC files from different releases, even more
fields will normally mis-compare (fields that exist in one release, but not another...).  Please see
lib/Shoc/CompareSpecTable.pm for more information.


=head1 See Also

See these modules and scripts for further detail.

tools/shoc_diff.thpl
tools/shoc_save.thpl

lib/Shoc.pm
lib/Shoc/CompareResult.pm
lib/Shoc/CompareSpec.pm
lib/Shoc/CompareSpecTable.pm
lib/DataStructureStore.pm

=cut

package Shoc;

use strict;
use Tharn;
use XML::LibXML;
use Data::Dumper;
use Shoc::CompareSpec;
use Shoc::CompareResult;

## @_shouldDoCompare()
## @description Routine which will determine a given field should be compared.
##              Note that if "compare" is defined, it is used instead of ignore.
##              If neither compare or ignore are specified, return 1 (true).
##
## @arg key     Required: hash key
## @arg compare Optional: ARRAY ref, contains fields to compare
## @arg ignore  Optional: ARRAY ref, contains fields to ignore
##
## @synopsis
##
##    my $rc = Shoc::_shouldDoCompare( key=>$key, ignore=>\@ignore );
##
## @return       :  1|0
##
##------------------------------------------------------------------------------
sub _shouldDoCompare {
    my %opt = ( key     => undef,
                compare => undef,   # Compare only these fields
                ignore  => undef,   # Ignore these fields
                @_ 
              );

    my $key     = $opt{key};
    my $compare = $opt{compare};
    my $ignore  = $opt{ignore};

    my $rc = 1;

    # if you have specified compare, and $key isn't in there, return 0
    if (     $compare && !grep(/^$key$/, @$compare) ) {
        $rc = 0;
    # if you have specified ignore, and $key is in there, return 0
    } elsif ( $ignore &&  grep(/^$key$/, @$ignore) ) {
        $rc = 0;
    }

    return $rc;
}

## @hasKeys()
## @description Routine which will determine if you can retrieve keys from 
##              the reference (i.e. if it is a hash, or an object based on a hash)
##
## @arg ref Required: A reference to the data
##
## @synopsis
##
##    my $rc = Shoc::hasKeys( $ref );
##
## @return       :  1|0
##
##------------------------------------------------------------------------------
sub hasKeys {
    my $ref = shift;
    my @keys;
    my $rc = 0;

    # Can't just call ref($ref) and expect it to be HASH, it might be
    # an arbitrary class which has a HASH as it's core structure.
    eval { @keys = keys %$ref; };
    if ( !$@ ) {
        if ( @keys ) {
            $rc = 1;
        }
    }
    return $rc;
}

## @buildKeyMap()
## @description Routine which will build a key:value pair string
##              (separated by ';') given the "primary" keys of a hash
##              based object, then assign the object to a hash, keyed by that
##              string.
##
##              The assumption here is that there is a hash object which
##              has certain key values which are "primary keys" which 
##              serve to uniquely identify an object and the other keys
##              are "data keys".
##
## @arg objRef   Required: A reference to the hash-based object
## @arg results  Required: A reference to the results hash, can be null
## @arg primkeys Required: A reference to primary keys array
##
## @synopsis
##
##    my $keyString = Shoc::buildKeyMap( $objRef, \%hash, \@primkeys );
##
## @return       :  1|0
##
##------------------------------------------------------------------------------
sub buildKeyMap {
    my $item = shift;
    my $hptr = shift;
    my $kptr = shift;

    my $kstr="";
    foreach my $key (@$kptr) {
        if ( $kstr ) {
            $kstr = $kstr.";".$key.":".$item->{$key};
        } else {
            $kstr = $key.":".$item->{$key};
        }
    }

    if ( $hptr ) {
        $hptr->{$kstr} = $item;
    }

    return $kstr;
}

## @removeDuplicates()
## @description Routine which removes duplicate entries from an array.
##
## @arg @items - array of items (possibly with duplicates)
##
## @synopsis
##
##    my @uniqueItems = Shoc::removeDuplicates( @items );
##
## @return       :  array 
##
##------------------------------------------------------------------------------
sub removeDuplicates {
    my @items = @_;

    my %seen = (); 
    foreach my $item (@items) { 
        $seen{$item}++; 
    }
    return (keys %seen);
}

## @name compare()
## @description Routine which will verify that 2 data structures are equivalent.
##              The structures can be any combination of array, hash or scalar.
##
## @arg ref1         Required: A reference to the first structure
## @arg ref2         Required: A reference to the second structure
##                             Terminal scalars (leafs) can be strings or numeric
##                             representing a single value.
## @arg compare      Optional: A reference to an array of hash fields to be compared.
## @arg ignore       Optional: A reference to an array of hash fields to be ignored.
## @arg errors       Optional: A reference to an array in which errors are stored.
## @arg validation   Optional: A reference to an array in which errors and successful validations are stored.
## @arg compare_spec Optional: A reference to a hash which specifies how to do the comparison (see Shoc::CompareSpec)
##
## @synopsis
##
##    my $rc = Shoc::compare( ref1 => \%act, ref2 => \%exp );
##
## @return       :  1|0
##
##------------------------------------------------------------------------------
sub compare {
    # index is "private" argument used for recursive calls
    my %opt = (   ref1            => undef, # DataRef:  HashRef or ArrayRef of first data
                  ref2            => undef, # DataRef:  HashRef or ArrayRef of second data
                  compare         => undef, # ArrayRef: Compare only these fields
                  ignore          => undef, # ArrayRef: Ignore these fields
                  errors          => undef, # ArrayRef: Store errors in here
                  validation      => undef, # ArrayRef: Store verbose results of compare in here
                  compare_spec    => undef, # HashRef:  Instructions on how to do the comparison.
                  index           => "",    # String:   Private tag, used in recursion
                  @_ 
              );

    my $whoami         = (caller(0))[3];
    my $calledby       = (caller(1))[3];
    my $ref1           = $opt{ref1};
    my $ref2           = $opt{ref2};
    my $index          = $opt{index};
    my $errors         = $opt{errors};
    my $validation     = $opt{validation};
    my $compare_spec   = $opt{compare_spec};
    my $compare        = $opt{compare};
    my $ignore         = $opt{ignore};

    my $rc = 1;
    my $ref1_type = ref($ref1);
    my $ref2_type = ref($ref2);

    if ( !ref($errors) =~ /ARRAY/i ) {
        $errors = [];
    }

    if ( !ref($validation) =~ /ARRAY/i ) {
        $validation = [];
    }

    # If a compare spec was specified, set compare/ignore
    # based on it, then add any missing compare/ignore
    # that are passed in via option array 
    if ( $compare_spec ) {

        $ignore = [$compare_spec->{IGNORE_KEYS}];

        # If we also have ignores passed in, (possibly overrides
        # to what was specified in compare_spec) add them
        if ( $opt{ignore} ) {
            foreach my $ign (@{$opt{ignore}}) {
                if ( !grep(/^$ign$/, @$ignore) ) {
                    push( @$ignore, $ign );
                }
            }
        }

        $compare = $compare_spec->{COMPARE_KEYS};

        # If we also have compares passed in, (possibly overrides
        # to what was specified in compare_spec) add them
        if ( $opt{compare} ) {
            foreach my $cmp (@{$opt{compare}}) {
                if ( !grep(/^$cmp$/, @$compare) ) {
                    push( @$compare, $cmp );
                }
            }
        }
    }
 
    if ( $ref1_type ne $ref2_type) {
       my $nomatch_msg = "$index STRUCTURE TYPE ($ref1_type) != ($ref2_type)";
       my $nomatch = Shoc::CompareResult->new( status=>0, message=>$nomatch_msg );
       $rc = 0;
       push( @$validation, $nomatch );
       push( @$errors,     $nomatch );
    } else {
        if ( $ref1_type =~ /ARRAY/i ) {
            
            my $last1 =  scalar(@$ref1) - 1;
            my $last2 =  scalar(@$ref2) - 1;
            my $last  = ($last1 > $last2) ? $last1 : $last2;
            my $greater_last = ($last1 > $last2) ? $last1 : $last2;

            # Compare ComponentStates as if they are not ordered, by using their
            # primary keys, even if they are sorted by key, if one list is missing
            # an entry, we would end up mis-comparing items.
            if ( ref($$ref1[0]) =~ /NACL::CS::/ ) {
                my $cType = cs2CType( ref($$ref1[0]) );
                eval "use $cType";
                if ($@) {
                    die "Failed to load module $cType '$@'";
                }

                # If any of the primary keys are ones we are to ignore
                # (i.e. if a primary key changed names between releases, 
                #       we may have a new and old key in one CS but not the other)
                my @keys = $cType->get_primary_keys();
                my @tmp;
                foreach my $key (@keys) {
                    # hash keys use _ not -, but primary keys returned may have -
                    $key =~ s/-/_/g;

                    if ( !grep( /^$key$/, @$ignore ) ) {
                        push( @tmp, $key );
                    }
                }
                @keys = @tmp;

                # build a hash key string out of the primary key names and values
                my (%h1,%h2);
                map( buildKeyMap( $_, \%h1, \@keys ), @$ref1 );
                map( buildKeyMap( $_, \%h2, \@keys ), @$ref2 );

                my @instance_keys1 = sort keys %h1;
                my @instance_keys2 = sort keys %h2;
                my @all_keys = removeDuplicates(@instance_keys1, @instance_keys2);
                @all_keys = sort @all_keys;

                my $numChecks = 0;
                my $numPasses = 0;
                foreach my $key (@all_keys) {
                    if ( defined $h1{$key} ) {
                        if ( defined $h2{$key} ) {
                            $numChecks++;
                            $numPasses += compare( ref1           => $h1{$key}, 
                                                   ref2           => $h2{$key},
                                                   index          => $index."[".$key."]", 
                                                   errors         =>$errors,
                                                   validation     =>$validation,
                                                   compare        =>$compare,
                                                   ignore         =>$ignore,
                                                   compare_spec   =>$compare_spec );
                        } else {
                            $rc = 0;
                            my $nomatch_msg = $index."[".$key."] CS (".$key.") != (non-existant)";
                            my $nomatch = Shoc::CompareResult->new( status=>0, message=>$nomatch_msg );
                            push( @$validation, $nomatch );
                            push( @$errors,     $nomatch );
                        }
                    } else {
                        $rc = 0;
                        my $nomatch_msg = $index."[".$key."] CS (non-existant) != (".$key.")";
                        my $nomatch = Shoc::CompareResult->new( status=>0, message=>$nomatch_msg );
                        push( @$validation, $nomatch );
                        push( @$errors,     $nomatch );
                    }
                }
                if ( $numChecks != $numPasses ) {
                    $rc = 0;
                }
            } else {

                my $numChecks = 0;
                my $numPasses = 0;
                my $missing = ($last1==$greater_last) ? 2 : 1;
                foreach my $i (0 .. $last) {
                    $numChecks++;
                    $numPasses += compare( ref1           => $$ref1[$i], 
                                           ref2           => $$ref2[$i], 
                                           index          => $index."[".$i."]", 
                                           errors         =>$errors,
                                           validation     =>$validation,
                                           compare        =>$compare,
                                           ignore         =>$ignore,
                                           compare_spec   =>$compare_spec );
                }

                if ( $numChecks != $numPasses ) {
                    $rc = 0;
                }

                if ( $greater_last > $last ) {
                    $rc = 0;
                    foreach my $i ($last+1 .. $greater_last) {
                        my $nomatch_msg;
                        if ( $missing == 1 ) {
                            $nomatch_msg = "$index CS (non-existant) != (".$index."[".$i."])";
                        } else {
                            $nomatch_msg = "$index CS (".$index."[".$i."]) != (non-existant)";
                        }

                        my $nomatch = Shoc::CompareResult->new( status=>0, message=>$nomatch_msg );
                        push( @$validation, $nomatch );
                        push( @$errors,     $nomatch );
                    }
                }
            }
          # The hasKeys is in there to allow Perl Objects that have
          # hashes at their core
        } elsif ( ($ref1_type =~ /HASH/i) || hasKeys($ref1) ) {
            my $numChecks = 0;
            my $numPasses = 0;
            my @actKeys = keys %$ref1;
            my @expKeys = keys %$ref2;
            foreach my $key (sort keys %$ref2) {
                $numChecks++;
                # check for elements in second that are missing from first
                if ( !exists $ref1->{$key} ) {
                    my $do_compare = 1;
        
                    $do_compare = _shouldDoCompare( key => $key, compare => $compare, ignore => $ignore );
        
                    if ( $do_compare ) {
                        $rc = 0;
                        my $nomatch_msg = "$index MISSING key \"$key\"";
                        my $nomatch = Shoc::CompareResult->new( status=>0, message=>$nomatch_msg );
                        push( @$validation, $nomatch );
                        push( @$errors,     $nomatch );
                    } else {
                        $numPasses++;
                    }
                } else {
                    $numPasses += compare( ref1           => $ref1->{$key}, 
                                           ref2           => $ref2->{$key}, 
                                           index          => $index."{".$key."}", 
                                           errors         =>$errors,
                                           validation     =>$validation,
                                           compare        =>$compare,
                                           ignore         =>$ignore,
                                           compare_spec   =>$compare_spec );
                }
            }
            if ( $numChecks != $numPasses ) {
                $rc = 0;
            }
            # check for elements in first that are extra as compared to second
            foreach my $key (sort keys %$ref1) {
                if ( !exists $ref2->{$key} ) {
                    my $do_compare = 1;
        
                    if ( $ignore && grep(/^$key$/, @$ignore) ) {
                        $do_compare = 0;
                    }

                    if ( $do_compare ) {
                        $rc = 0;
                        my $nomatch_msg = "$index EXTRA key \"$key\"";
                        my $nomatch = Shoc::CompareResult->new( status=>0, message=>$nomatch_msg );
                        push( @$validation, $nomatch );
                        push( @$errors,     $nomatch );
                    }
                }
            }
        } else {
            my $ref1val = $ref1;
            my $ref2val = $ref2;
            if ( ref( $ref1 ) =~ /SCALAR/ ) {
                $ref1val = $$ref1;
            }
            if ( ref( $ref2 ) =~ /SCALAR/ ) {
                $ref2val = $$ref2;
            }
            $ref1val = "undef" if ( !defined $ref1val );
            $ref2val = "undef" if ( !defined $ref2val );

            my $do_compare = 1;

            my $fieldname = $index;
            if ( $fieldname =~ /{(\S+)}/ ) {
                $fieldname = $1;
            }

            $do_compare = _shouldDoCompare( key => $fieldname, compare => $compare, ignore => $ignore );

            if ( $do_compare ) {
                # Do the compare
                my $match_msg   = "$index \'$ref1val\' == \'$ref2val\'";
                my $nomatch_msg = "$index \'$ref1val\' != \'$ref2val\'";
                my $match   = Shoc::CompareResult->new( status=>1, message=>$match_msg );
                my $nomatch = Shoc::CompareResult->new( status=>0, message=>$nomatch_msg );
                if ( $compare_spec && 
                     (defined $compare_spec->{COMPARE_RTNS}) &&
                     (ref($compare_spec->{COMPARE_RTNS}) =~ /HASH/) && 
                     (defined $compare_spec->{COMPARE_RTNS}->{$fieldname}) ) {
                    my $cmpRtns = $compare_spec->{COMPARE_RTNS}->{$fieldname};
                    my $cmpRtn = $cmpRtns;
                    my @cmpArgs = ( $ref1val, $ref2val );
                    if ( ref($cmpRtns) =~ /ARRAY/i ) {
                        $cmpRtn = $cmpRtns->[0];
                        my $last = $#{$cmpRtns};

                        if ( $last >= 1 ) {
                            push( @cmpArgs, @{$cmpRtns}[1..$last] );
                        }
                    }

                    if ( ref($cmpRtn) =~ /CODE/i ) {
                        if ( $cmpRtn->( @cmpArgs ) ) { 
                            # If someone has supplied a compare code block, most likely these
                            # 2 fields are not literally identical.  Using ~= to signify that they
                            # are equivalent
                            $match_msg = "$index \'$ref1val\' ~= \'$ref2val\'";
                            my $match = Shoc::CompareResult->new( status=>1, message=>$match_msg );
                            push( @$validation, $match );
                        } else {
                            $rc = 0;
                            push( @$validation, $nomatch );
                            push( @$errors,     $nomatch );
                        }
                    } else {
                        die $compare_spec->{$fieldname}." does not refer to a closure or routine reference";
                    }
                    
                } elsif ( $ref1val eq $ref2val ) {
                    push( @$validation, $match );
                } else {
                    $rc = 0;
                    push( @$validation, $nomatch );
                    push( @$errors,     $nomatch );
                }
            }
        }
    }

    return $rc;
}

## @buildComponentHash()
## @description Routine which aggregates the component states into a single hash
##              keyed by NACL component type, and containing an array of the
##              associated component state objects.
##
## @arg node   - REQUIRED - NACL::C::Node component to use as a command-interface to query the other
##                        components.
## @arg components - OPTIONAL - Array reference to list of NACL::C::* component names
## @arg ignore     - OPTIONAL - Array reference to list of NACL::C::* component names
##
## @synopsis
##
##    my $hptr = Shoc::buildComponentHash( node => $node );
##
## @return       :  pointer to a hash, keyed by the NACL::C::* component names, and
##                  value is an array of the ComponentState objects associated with 
##                  each Component.
##
##------------------------------------------------------------------------------
sub buildComponentHash {
    my $whoami       = (caller(0))[3];
    my %opt = ( node        => undef,  # pass in one node of the cluster
                components  => undef,  # arrayref, list of components
                ignore      => undef,  # arrayref, list of components to ignore
                @_ 
              );
    my $node = $opt{node};
    if ( !$node ) {
        die "$whoami - Must provide \"node\" option";
    }

    # No real reason to specify both components and ignore
    my @comps = ();
    if ( ref($opt{components}) =~ /ARRAY/ ) {
        @comps = @{$opt{components}};
    }

    my @ignore = ();
    if ( ref($opt{ignore}) =~ /ARRAY/ ) {
        @ignore = @{$opt{ignore}};
    }

    my $data;

    my $components = Shoc::CompareSpec::findSaveComponentList( node => $node, data => \$data );

    # If component list passed in use it instead.
    # (note, findSaveComponentList also adds the version to $data)
    if ( @comps ) {
        $components = \@comps;
    }

    foreach my $component (@$components) {

        if ( grep( /^$component$/, @ignore ) ) {
            $Log->comment("Told to IGNORE component $component");
        } else {
            eval "use $component";
            if ($@) {
                die "Failed to load module $component '$@'";
            }
        
            my @objects;
            my $cs = c2CsType( $component );
            @objects = $cs->fetch( command_interface => $node, allow_empty => 1 );
        
            foreach my $object (@objects) {
                addObject( \$data, \$object );
            }
        
            if ( @objects ) {
                @{$data->{$component}{OBJECT_DATA}} = sort cmpCSByKeys @{$data->{$component}{OBJECT_DATA}};
            }
        }
    }

    return $data;
}

## @addObject()
## @description Routine which will add a component into a hash which is keyed by that
##              object type.
##
## @arg data - REQUIRED - hash pointer
## @arg obj  - REQUIRED - NACL::C::Component based object pointer
##
## @synopsis
##
##    my %data;
##    my $node = NACL::CS::Node->fetch();
##    Shoc::addObject( \%data, $node );
##
## @return       :  Nothing.
##
##------------------------------------------------------------------------------
sub addObject {
    my $data = shift;
    my $obj = shift;

    my $ref = cs2CType( ref($$obj));
    if ( !defined $$data->{COMPONENTS} ) {
        $$data->{COMPONENTS} = [];
    }

    if ( !defined $$data->{$ref}{OBJECT_DATA} ) {
        $$data->{$ref}{OBJECT_DATA} = [];
    }

    if ( !grep(/^$ref$/, @{$$data->{COMPONENTS}}) ) {
        push( @{$$data->{COMPONENTS}}, $ref );
    }

    my $state = $$obj;

    # There is an issue with externalizing NACL::CS objects
    # with the Storable class.  Recently a reference to 
    # a NACL::C::CommandInterface was added to CS objects and
    # now Storable won't externalize them because of GLOB
    # data stored in the command_interface.  So here we are
    # simply deleting that reference if it exists.
    if ( exists $state->{command_interface} ) {
        delete $state->{command_interface};
    }
    push( @{$$data->{$ref}{OBJECT_DATA}}, $state );
}

## @cs2CType()
## @description Routine which will add a component into a hash which is keyed by that
##              object type.  Useful if you need the primary_keys and you only have a
##              ComponentState object.
##
## @arg cstype - REQUIRED - ComponentState class name
##
## @synopsis
##
##    my $node = NACL::CS::Node->fetch();
##    my $cType = Shoc::cs2CType( ref($node) );
##    my @prim_keys = $cType->get_primary_keys();
##
## @return       :  Will return associated NACL::C::Component for a NACL::CS::Component
##                  (ex: NACL::C:Node given NACL::CS:Node)
##
##------------------------------------------------------------------------------
sub cs2CType {
    my $csType = shift;
    my $cType = $csType; 
    $cType =~ s/:CS:/:C:/g;
    return $cType;
}

## @c2CsType()
## @description Routine which will add a component into a hash which is keyed by that
##              object type.  Useful if you need the primary_keys and you only have a
##              ComponentState object.
##
## @arg ctype - REQUIRED - Component class name
##
## @synopsis
##
##    my $node = NACL::C::Node->find();
##    my $csType = Shoc::cs2CType( ref($node) );
##
## @return       :  Will return associated NACL::CS::Component for a NACL::C::Component
##                  (ex: NACL::CS:Node given NACL::C:Node)
##
##------------------------------------------------------------------------------
sub c2CsType {
    my $cType = shift;
    my $csType = $cType; 
    $csType =~ s/:C:/:CS:/g;
    return $csType;
}

## @cmpCSByKeys()
## @description Routine which will compare 2 ComponentState objects by
##              primary keys.
##
## @arg data1 - REQUIRED - NACL::CS::* pointer
## @arg data2 - REQUIRED - NACL::CS::* pointer
##
## @synopsis
##
##    my @volStates = # some list of NACL::CS::Volume objects...
##    @volStates = sort cmpCSByKeys @volStates;
##
## @return       :  0 if data1 is lt data2, 1 otherwise
##------------------------------------------------------------------------------
sub cmpCSByKeys {
    my $csType = ref($a);
    my $cType = cs2CType( $csType );

    my @keys  = sort $cType->get_primary_keys();

    foreach my $key (@keys) {

        # hash keys use _ not -, but primary keys returned may have -
        $key =~ s/-/_/g;

        if ( $a->{$key} lt $b->{$key} ) {
            return 0;
        }
    }

    return 1;
}
    
1;
