#Hey emacs, this is basically -*- perl -*-
# $Id$
#***********************************************************************
#
# NAME
#   Shoc::CompareSpec.pm 
#
# SYNOPSIS
#   Shoc (show-output-compare) Compare Specification hashes.
#
# DESCRIPTION
#   This library provides routines to use in comparing NACL ComponentStates.
#
# AUTHOR
#   dl-coati-core
#***********************************************************************

=pod

=head1 Library Name

Shoc::CompareSpec.pm

=head1 Author

dl-coati-core

=head1 Description

This library contains routines that retrieve the list of components that should
have their states saved as part of SHOC save (tools/shoc_save.thpl), retrieves the
Compare Specification for a given LESSER/GREATER release pair, and routines that
actually do some of the comparisons defined in Shoc::CompareSpecTable.pm.

=head1 See Also

See these modules and scripts for further detail.

tools/shoc_diff.thpl
tools/shoc_save.thpl

lib/Shoc.pm
lib/Shoc/CompareResult.pm
lib/Shoc/CompareSpec.pm
lib/Shoc/CompareSpecTable.pm
lib/DataStructureStore.pm

=cut

package Shoc::CompareSpec;

use Tharn;
use Data::Dumper;
use Clone;

use Shoc::CompareSpecTable;

## @findSaveComponentList()
## @description Routine which will return the component list to be saved for a given release.
##
## @arg node Required: NACL::C::Node object
## @arg data Optional: reference to hash where release string will be stored
##
## @synopsis
##
##    my $data;
##    my $rc = Shoc::CompareSpec::@findSaveComponentList( node => $node, data => \$data );
##
## @return       :  reference to an array of NACL::C::* object types.
##
##------------------------------------------------------------------------------
sub findSaveComponentList {
    my $whoami = (caller(0))[3];
    my %opt = (
        node  => undef,
        data  => undef,
        @_,
    );

    my $node = $opt{node};
    if ( !$node ) {
        die "$whoami - Must provide \"node\" option";
    }

    my $verMgr = $node->get_version_manager();

    # Should be a simple way to get this
    my $mode    = (ref($verMgr) =~ /::CMode/) ? "CMode" : "7Mode";

    my $release = $verMgr->get_version_attribute( attribute => "release" );

    # Store release information in there so shoc_diff.thpl can retrieve it
    # later
    if ( $opt{data} ) {
        ${$opt{data}}->{RELEASE} = $release;
    }

    foreach my $item (@{$Shoc::CompareSpecTable::saveComponents}) {
        if ( $release  =~ $item->{RELEASE} ) {
            return $item->{$mode}->{COMPONENT_LIST};
        }
    }
    die "ComponentList for release=$release mode=$mode was not found";
}

## @findComponentCompareSpec()
## @description Routine which will return the compare specification for a given
##              set of lesser/greater releases.
##
## @arg ref Required: 
##
## @synopsis
##
##    my $rc = Shoc::CompareSpec::findComponentCompareSpec( lesser => $lesser, greater => $greater );
##
## @return       :  reference to a CompareSpec hash
##
##------------------------------------------------------------------------------
sub findComponentCompareSpec {
    my %opt = (
        lesser  => undef,
        greater => undef,
        @_,
    );

    if ( !defined $opt{lesser} || !defined $opt{greater} ) {
        die "Must supply options lesser and greater";
    }

    foreach my $item (@{$Shoc::CompareSpecTable::compareSpecs}) {

        if ( ( $opt{lesser}  =~ $item->{LESSER} ) &&
             ( $opt{greater} =~ $item->{GREATER} ) ) {
            my $rc = $item;

            # If the version strings are different, there is no way that system image show 
            # will compare
            if ( $opt{lesser} ne $opt{greater} ) {
                $rc = Clone::clone($item);
                if ( !defined $rc->{COMPONENTS} ) { 
                    $rc->{COMPONENTS} = { IGNORE  => [ qw(NACL::C::SystemNodeImage) ], };
                } elsif ( !defined $rc->{COMPONENTS}->{IGNORE} ) {
                    $rc->{COMPONENTS}->{IGNORE} = [ qw(NACL::C::SystemNodeImage) ];
                } else {
                    if ( !grep( /^NACL::C::SystemNodeImage$/, @{$rc->{COMPONENTS}->{IGNORE}} ) ) {
                        push( @{$rc->{COMPONENTS}->{IGNORE}}, 'NACL::C::SystemNodeImage' );
                    }
                }
            }
            return $rc;
        }
    }
    die "ComponentCompareSpec for releases lesser=$opt{lesser} and greater=$opt{greater} was not found";
}

## @findReleaseString()
## @description Routine which return the release string in a Shoc config hash.
##
## @arg data Required: A reference to the data structure created by shoc_save.thpl
##
## @synopsis
##
##    my $rc = Shoc::CompareSpec::findReleaseString( $data );
##
## @return       :  release string
##
##------------------------------------------------------------------------------
sub findReleaseString {
    my $data = shift;
    my $rc;
    my ($num_nodes, $num_match)=(0,0);
    
    # RELEASE should be defined, but keeping SystemNodeImage in there in 
    # for backward compatability
    if ( defined $data->{RELEASE} ) {
        $rc = $data->{RELEASE};
        $num_nodes = 1;
        $num_match = 1;
    } elsif ( defined $data->{'NACL::C::SystemNodeImage'} ) {
        foreach my $item (@{$data->{'NACL::C::SystemNodeImage'}->{OBJECT_DATA}}) {
            if ( $item->{iscurrent} =~ /true/i ) {
                $num_nodes++;
                if ( !defined $rc ) {
                    $num_match=1;
                } elsif ( $rc eq $item->{version} ) {
                    $num_match++;
                }
                $rc = $item->{version};
            }
        }
    }

    if ( $num_match && ($num_nodes == $num_match) ) {
        return $rc;
    }
    die "Can't find release string in Compare Data Structure";
}

## @cmp_lifs()
## @description Routine which will compare the fields of the network interface
##              table, minus the fields which might differ if interfaces are home
##              in one set of saved ComponentStates, but not-home in another.
##
## @arg ref1         Required: A reference to the first structure
## @arg ref2         Required: A reference to the second structure
##                             Terminal scalars (leafs) can be strings or numeric
##                             representing a single value.
## @arg compare      Optional: A reference to an array of ComponentState fields to be compared.
## @arg ignore       Optional: A reference to an array of ComponentState fields to be ignored.
## @arg errors       Optional: A reference to an array in which errors are stored.
## @arg validation   Optional: A reference to an array in which errors and successful validations are stored.
## @arg compare_spec Optional: A reference to a hash which specifies how to do the comparison (see Shoc::CompareSpec)
##
## @synopsis
##
##    my $rc = Shoc::cmp_lifs( ref1 => \%act, ref2 => \%exp );
##
## @return       :  1|0
##
##------------------------------------------------------------------------------
sub cmp_lifs {
    my %opt = (
                  ref1            => undef,
                  ref2            => undef,
                  compare         => undef,
                  ignore          => undef,
                  errors          => undef,
                  validation      => undef,
                  compare_spec    => undef,
                  @_
              );

    my $ref1           = $opt{ref1};
    my $ref2           = $opt{ref2};
    my $ignore         = $opt{ignore};
    my $errors         = $opt{errors};
    my $validation     = $opt{validation};
    my $compare_spec   = $opt{compare_spec};

    if ( !$ref1 || !$ref2 ) {
        die "Shoc::CompareSpec::cmp_lifs() must provide both ref1 and ref2 arguments";
    }

    if ( !ref($errors) =~ /ARRAY/i ) {
        $errors = [];
    }

    if ( !ref($validation) =~ /ARRAY/i ) {
        $validation = [];
    }

    my $rc = Shoc::compare( ref1         => $ref1,
                            ref2         => $ref2,
                            ignore       => $ignore,
                            errors       => $errors,
                            validation   => $validation,
                            compare_spec => $compare_spec );

    if ( $compare_spec ) {

        $ignore = [$compare_spec->{IGNORE_KEYS}];

        # If we also have ignores passed in, (possibly overrides
        # to what was specified in compare_spec) add them
        if ( $opt{ignore} ) {
            foreach my $ign (@{$opt{ignore}}) {
                if ( !grep(/^$ign$/, @$ignore) ) {
                    push( @$ignore, $ign );
                }
            }
        }
    }

    my $cType = Shoc::cs2CType( ref($$ref1[0]) );
    eval "use $cType";
    if ($@) {
        die "Failed to load module $cType '$@'";
    }

    # If any of the primary keys are ones we are to ignore
    # (i.e. if a primary key changed names, we may have a new
    #       and old key in here)
    my @keys = $cType->get_primary_keys();
    my @tmp;
    foreach my $key (@keys) {
        # hash keys use _ not -, but primary keys returned may have -
        $key =~ s/-/_/g;

        if ( !grep( /^$key$/, @$ignore ) ) {
            push( @tmp, $key );
        }
    }
    @keys = @tmp;

    my $last = $num - 1;
    my $num_passes = 0;

    # build a hash key string out of the primary key names and values
    my (%h1,%h2);
    map( Shoc::buildKeyMap( $_, \%h1, \@keys ), @$ref1 );
    map( Shoc::buildKeyMap( $_, \%h2, \@keys ), @$ref2 );

    my @instance_keys1 = sort keys %h1;
    my @instance_keys2 = sort keys %h2;
    my @all_keys = Shoc::removeDuplicates(@instance_keys1, @instance_keys2);
    @all_keys = sort @all_keys;

    foreach my $key (@all_keys) {
        if ( defined $h1{$key} ) {
            if ( defined $h2{$key} ) {
                my $looprc=0;
                if ( ( $h1{$key}->{is_home} ne 
                       $h2{$key}->{is_home} ) ) {
                    if ( ( $h1{$key}->{curr_node} ne 
                           $h2{$key}->{curr_node} ) ||
                         ( $h1{$key}->{curr_port} ne 
                           $h2{$key}->{curr_port} ) ) {
                        $looprc++;
                    }
                } else {
                    if ( ( $h1{$key}->{curr_node} eq 
                           $h2{$key}->{curr_node} ) &&
                         ( $h1{$key}->{curr_port} eq 
                           $h2{$key}->{curr_port} ) ) {
                        $looprc++;
                    }
                }
                my $oper   = ( $looprc ) ? ' =~ ' : ' !~ ';

                my $msg = "\[$key\]\{is_home\;curr_node\;curr_port\} ".
                          $h1{$key}->{is_home}.";".
                          $h1{$key}->{curr_node}.";".
                          $h1{$key}->{curr_port}.$oper.
                          $h2{$key}->{is_home}.";".
                          $h2{$key}->{curr_node}.";".
                          $h2{$key}->{curr_port};

                my $result;
                if ( !$looprc ) {
                    $rc = 0;
                    $result = Shoc::CompareResult->new( status=>0, message=>$msg );
                    push( @$errors, $result );
                } else {
                    $result = Shoc::CompareResult->new( status=>1, message=>$msg );
                }

                push( @$validation, $result );
            } else {
                $rc = 0;
                # call to Generic Shoc::compare() will catch the missing elements, just set $rc to fail
            }
        } else {
            $rc = 0;
            # call to Generic Shoc::compare() will catch the missing elements, just set $rc to fail
        }
    }

    return $rc;
}

## @treat_as_equiv()
## @description Routine which will determine if 2 values are equivalent
##
## @arg first        Required: First  scalar (treated as string)
## @arg second       Required: Second scalar (treated as string)
## @arg equiv_values Required: Array of value arrays which are to be treated
##                             as equivalent
##
## @synopsis
##
##    my $equiv = [ [qw( yes on true )], [qw( no off false )] ];
##    my $rc = Shoc::CompareSpec::treat_as_equiv( $first, $second, equiv_values=>$equiv );
##
## @return       :  1|0
##
##------------------------------------------------------------------------------
sub treat_as_equiv {
    my $first  = shift;
    my $second = shift;
    my $whoami = (caller(0))[3];
    my ($firstMatch, $secondMatch) = (0,0);

    my %opt = (
        equiv_values  => [ [] ],
        @_,
    );

    my $equiv = $opt{equiv_values};

    if ( !ref($equiv) =~ /ARRAY/ ) {
        die "$whoami: equiv_values must be a reference to an array of arrays!";
    }

    foreach $item (@$equiv) {
        if ( !ref($item) =~ /ARRAY/ ) {
            die "$whoami: equiv_values must be a reference to an array of arrays!";
        }
        ($firstMatch, $secondMatch) = (0,0);

        # Maybe we got lucky and they are identical?
        if ( $first eq $second ) {
            return 1;
        }

        if ( grep(/^$first$/, @$item ) ) {
            $firstMatch=1;
        }

        if ( grep(/^$second$/, @$item ) ) {
            $secondMatch=1;
        }

        if ( $firstMatch && $secondMatch ) {
            return 1;
        }
    }

    return 0;
}

## @cmp_csv()
## @description Routine which treat 2 given strings as comma-separated lists of elements
##
## @arg first        Required: First  scalar (treated as string)
## @arg second       Required: Second scalar (treated as string)
## @arg allow_extra  Required: Array of strings which can be ignored
##
## @synopsis
##
##    my $first  = "raid_dp, parity uninit'd!"
##    my $second = "raid_dp, parity uninit'd!, redirect"
##    my $rc = Shoc::CompareSpec::cmp_csv( $first, $second, allow_extra =>[qw(redirect)] );
##
## @return       :  1|0
##
##------------------------------------------------------------------------------
sub cmp_csv {
    my $first  = shift;
    my $second = shift;
    my $whoami = (caller(0))[3];
    my ($firstMatch, $secondMatch) = (0,0);

    my %opt = (
        allow_extra  => [ ],
        @_,
    );

    my $extra = $opt{allow_extra};

    my @first  = split( ',', $first );
    my @second = split( ',', $second );

    foreach $item (@$extra) {
        @first  = !grep(/^$item$/, @first );
        @second = !grep(/^$item$/, @second );
    }

    @first  = sort @first;
    @second = sort @second;

    $first  = join( ',', @first );
    $second = join( ',', @second );

    if ( $first eq $second ) {
        return 1;
    }

    return 0;
}
1;
