#Hey emacs, this is basically -*- perl -*-
# $Id$
#***********************************************************************
#
# NAME
#   Shoc::CompareSpecTable.pm 
#
# SYNOPSIS
#   Shoc (show-output-compare) Compare Specification hashes.
#
# DESCRIPTION
#   This library provides tabling to use in comparing NACL components.
#
# AUTHOR
#   dl-coati-core
#***********************************************************************

=pod

=head1 Library Name

Shoc::CompareSpecTable.pm

=head1 Author

dl-coati-core

=head1 Description

This library defines what components are to be saved for a given release and mode, and 
how comparisons between components are to be accomplished.

These structures support cross release comparisons, by:

=over 4

=item * Identifying components to be compared.
=item * Identifying components to be ignored.
=item * Identifying components to be compared using a non-standard method.
=item * Identifying component fields which are to be ignored.
=item * Identifying component fields which are to be compared using a non-standard method.

=back

By default, SHOC comparisons are done on all ComponentState objects that have been saved, 
and the method is a simple string compare.  There are certain fields and ComponentStates 
for which this will not suffice.  For example, if we are comparing a SystemNodeImage
ComponentState when a filer was running BR against the same table when the filer was
running RR, many of the fields will not compare between the 2 releases.  Or if we are
trying to compare the Interface ComponentState after a lif was migrated from it's home port,
a straight string compare will show a mis-match, but it is possible that this is perfectly
acceptable.  Or even the case of Aggregate availsize, it is very likely that the availsize
will fluctuate over the course of multiple queries so that a strict comparison on this field
will fail.

The compareSpecs table is an array of hashes.  At the top level, the keys are:

=over 4

=item * LESSER  - a regex which matches an Ontap Release version for the lesser release.
=item * GREATER - a regex which matches an Ontap Release version for the greater release.
=item * COMPONENT_SPEC - a hash which details how the comparisons are to be accomplished.

=back

When doing a comparison usine Shoc::compare(), we find the appropriate compare spec by
searching the compareSpecs table from the beginning until we find an entry that matches 
the LESSER and GREATER releases.  Intra-release (same release) compares are handled by
adding table entries with the same regex for LESSER and GREATER while inter-release (different
release) compares will have different regexs for LESSER and GREATER.

Within the COMPONENT_SPEC, the following keys are possible (all top level keys are optional):

=over 4

=item * COMPONENTS  - a hash consisting of keys COMPARE or IGNORE, which are lists of components.
=item * NACL::C::component_name - a hash with keys:

=over 4

=item * COMPARE_RTNS - either a subroutine reference or a hash of subroutines keyed by ComponentState field name.
=item * IGNORE_KEYS - a list of ComponentState fields for this component which are to be ignored entirely.
=item * COMPARE_KEYS - a list of ComponentState fields for this component which are to be compared.

=back
=back

Starting with the COMPONENTS key, this is if you want to restrict the components to compare to 
something less than the components that are provided in the data structures passed into Shoc::compare().
You can either define a set of components to compare (COMPARE), or a set of components to ignore
(IGNORE).  These 2 keys should not be specified together, but if they were, the COMPARE would win out
and the IGNORE would be ignored.

At the next level, are hash entries keyed by fully qualified NACL component name (for example
NACL::C::Volume or NACL::C::Aggregate).  In those hashes, you can provide a IGNORE_KEYS or a COMPARE_KEYS
entry which detail the ComponentState fields which are to be ignored, or compared respectively.  As 
before, these should not be specified together, but if they are, the COMPARE_KEYS trumps the IGNORE_KEYS.

Note, it is generally better to define IGNORE parameters instead of COMPARE parameters.  This is because
if a new ComponentState field is added, if you specified a discrete set of COMPARE_KEYS, you would 
automatically ignore the new ComponentState fields, but if you specified IGNORE_KEYS, that newly 
added ComponentState field would automatically get compared.  Perhaps a little more maintenence 
work, but you are much less likely to silently and unintentionally ignore changes.

Also at this level, you can define COMPARE_RTNS.  The data element for this key can either be a subroutine
reference (if you want to do the comparison for this table using this routine) or it can be a hash
of subroutine references, keyed by ComponentState field name.  In the case of NACL::C::Interface, 
is is possible that 2 lif states should be considered the same even if one lif has migrated.  In this 
case, we specify the curr_node, curr_port, is_home fields in the IGNORE_KEYS so that when Shoc::compare()
is called, it will not compare these fields, then the routine tabled for the NACL::C::Interface
component can look at multiple fields of the 2 ComponentState objects at the same time (namely is_home,
curr_node and curr_port) to decide if these differences are acceptable.

=head1 Limitations

Currently this file only supports one way to compare 2 components given a LESSER/GREATER release pair.
Nothing prevents multiple different specifications from being created, it is just that the 
Shoc::findComponentCompareSpec() routine does not support it.  If this were desired, multiple tables
could be defined.

=head1 Table Maintenence

Changes to these tables should be undertaken with great care.  Preferably, fields should only be
ignored if one release does not have the field at all.  If a field exists for both LESSER and GREATER
releases, but there is a value difference which can be calculated soley from those 2 fields and
knowledge of the 2 releases (i.e. change in units or format), then a perl closure (i.e. an unnamed
subroutine) can generally be used to convert the LESSER into the GREATER format, allowing comparison.

If there are ComponentState differences which require examination of multiple fields of the same
ComponentState, it is better to ignore all of those keys, and table a subroutine reference for COMPARE_RTNS
(see the NACL::C::Interface example).  In this case, the tools/shoc_diff.thpl will do the generic
compare on all of the fields except is_home, curr_node and curr_port and then subsequently call
Shoc::CompareSpec::cmp_lifs() method to verify that if is_home is true, then curr_node and curr_port
are the same in the LESSER/GREATER components, and if is_home is false, that curr_node and/or curr_port
are different.

It is also important that this file is not modified due to every change in every branch.  If differences
are found during development, it is better to temporarily ignore fields which miscompare in development
branches in tools/shoc_diff.thpl (please see POD in that file for SHOC_IGNORE) until changes have
stabilized, been merged into main, and we are fairly confident that those fields are reasonably
stable.

It is also important, as mentioned before, that mismatching fields are not automatically just 
added to the IGNORE_KEYS list in this file.  Also any changes made to this file need to be approved by
the library owner.

=head1 See Also

See these modules and scripts for further detail.

tools/shoc_diff.thpl
tools/shoc_save.thpl

lib/Shoc.pm
lib/Shoc/CompareResult.pm
lib/Shoc/CompareSpec.pm
lib/DataStructureStore.pm

=cut


package Shoc::CompareSpecTable;

use Tharn;
use Data::Dumper;

## @summary saveComponents
##
## @synopsis 
##     Table which specifies what NACL components to save for a given release
##     and mode.
##
## @description 
##     The table is an array of hashes.  Each entry of the array has a RELEASE regex
##     and a 7Mode and CMode entry.
##
##     When looking up the entry to get the components to save, we start at the
##     beginning of the array and stop at the first match.  So you can have a
##     "generic" entry for a given release, like this:
##
##        RELEASE  => qr/^Boilermaker__8.0/,
##  
##     and later, can add more specific release entries (like for 8.0.1 and/or 8.0.2)
##     before this entry, so that they are found before the generic entry if you find
##     issues with a specific release.
##
##    {
##        RELEASE  => qr/^Boilermaker__8.0/,
##        '7Mode'  => {
##            COMPONENT_LIST => [ qw( 
##                                    NACL::C::Aggregate
##                                    NACL::C::Volume
##                                  )
##            ],
##        },
##        'CMode'  => {
##            COMPONENT_LIST => [ qw( 
##                                    NACL::C::Cluster
##                                    NACL::C::Node
##                                    NACL::C::SystemNodeImage
##                                    NACL::C::NetworkPort
##                                    NACL::C::Interface
##                                    NACL::C::Aggregate
##                                    NACL::C::Volume
##                                    NACL::C::Vserver
##                                    NACL::C::SystemServicesFirewall
##                                    NACL::C::VserverCifs
##                                    NACL::C::VserverCifsShare
##                                    NACL::C::VserverCifsDomainPreferredDc 
##                                  )
##            ],
##        },
##    },
##
##------------------------------------------------------------------------------
our $saveComponents = [
    {
        RELEASE  => qr/^Boilermaker__8.0/,
        '7Mode'  => {
            COMPONENT_LIST => [ qw( 
                                    NACL::C::Aggregate
                                    NACL::C::Volume
                                  )
            ],
        },
        'CMode'  => {
            COMPONENT_LIST => [ qw( 
                                    NACL::C::Cluster
                                    NACL::C::Node
                                    NACL::C::SystemNodeImage
                                    NACL::C::NetworkPort
                                    NACL::C::Interface
                                    NACL::C::Aggregate
                                    NACL::C::Volume
                                    NACL::C::Vserver
                                    NACL::C::SystemServicesFirewall
                                    NACL::C::VserverCifs
                                    NACL::C::VserverCifsShare
                                    NACL::C::VserverCifsDomainPreferredDc 
                                  )
            ],
        },
    },
    {
        RELEASE  => qr/^RollingRock__8.1/,
        '7Mode'  => {
            COMPONENT_LIST => [ qw( 
                                    NACL::C::Aggregate
                                    NACL::C::Volume
                                  )
            ],
        },
        'CMode'  => {
            COMPONENT_LIST => [ qw( 
                                    NACL::C::Cluster
                                    NACL::C::Node
                                    NACL::C::SystemNodeImage
                                    NACL::C::NetworkPort
                                    NACL::C::Interface
                                    NACL::C::Aggregate
                                    NACL::C::Volume
                                    NACL::C::Vserver
                                    NACL::C::SystemServicesFirewall
                                    NACL::C::VserverCifs
                                    NACL::C::VserverCifsShare
                                    NACL::C::VserverCifsDomainPreferredDc 
                                  )
            ],
        },
    },
    {
        RELEASE  => qr/^SierraNevada__8.2/,
        '7Mode'  => {
            COMPONENT_LIST => [ qw( 
                                    NACL::C::Aggregate
                                    NACL::C::Volume
                                  )
            ],
        },
        'CMode'  => {
            COMPONENT_LIST => [ qw( 
                                    NACL::C::Cluster
                                    NACL::C::Node
                                    NACL::C::SystemNodeImage
                                    NACL::C::NetworkPort
                                    NACL::C::Interface
                                    NACL::C::Aggregate
                                    NACL::C::Volume
                                    NACL::C::Vserver
                                    NACL::C::SystemServicesFirewall
                                    NACL::C::VserverCifs
                                    NACL::C::VserverCifsShare
                                    NACL::C::VserverCifsDomainPreferredDc 
                                  )
            ],
        },
    },
    {
        RELEASE  => qr/^Fullsteam__8.3/,
        'CMode'  => {
            COMPONENT_LIST => [ qw( 
                                    NACL::C::Cluster
                                    NACL::C::Node
                                    NACL::C::SystemNodeImage
                                    NACL::C::NetworkPort
                                    NACL::C::Interface
                                    NACL::C::Aggregate
                                    NACL::C::Volume
                                    NACL::C::Vserver
                                    NACL::C::SystemServicesFirewall
                                    NACL::C::VserverCifs
                                    NACL::C::VserverCifsShare
                                    NACL::C::VserverCifsDomainPreferredDc 
                                  )
            ],
        },
    },
];

# Note, Fullsteam does not support 7Mode

## @summary compareSpecs
##
## @synopsis 
##     Table which specifies, on a from/to release basis, the components to be
##     compared, and how to do the comparison.
##
## @description 
##     The table is an array of hashes.  Each entry of the array has a LESSER and GREATER
##     release regex.  When looking up the entry to do the comparisons, we start at the
##     beginning of the array and stop at the first match.
##
##        LESSER  => qr/^Boilermaker__8.0/,
##        GREATER => qr/^RollingRock__8.1.0/,
##
##     We can also specify either the list of components to compare or ignore with the COMPONENTS
##     hash element.  If not present, the components present in the LESSER list will be compared.
##
##        # Can specify that given components should be ignored
##        COMPONENTS => {
##            IGNORE  => [ qw(NACL::C::SystemServicesFirewall) ],
##        },
##
##     Following this is the COMPONENT_SPEC hash element.  This allows us to override, on a per
##     component basis, the keys that we want to ignore or compare (IGNORE_KEYS or COMPARE_KEYS),
##     comparison routines to be used for the entire component (see cmp_lifs example), or if we
##     only need to change the comparison used for specific fields (see COMPARE_RTNS for 
##     NACL::C::Aggregate below).  The comparison routines can be Perl closures, or routine 
##     references.  If you specify a closure or a simple routine name, the 2 fields you are 
##     attempting to compare are passed in (in lesser, greater order).  If you want to specify 
##     options to be passed in to the routine, specify these using an unnammed array
##     (see treat_as_equiv example below).
##
##        COMPONENT_SPEC => { 
##            COMPONENTS => {
##                IGNORE  => [ qw(NACL::C::SystemNodeImage) ],
##            },
##            "NACL::C::Node" => {
##                IGNORE_KEYS => [ qw(uptime) ],
##            },
##            "NACL::C::Interface" => {
##                COMPARE_RTNS =>  \&Shoc::CompareSpec::cmp_lifs,
##                IGNORE_KEYS => [ qw(curr_node curr_port is_home auto data_protocol 
##                                    status_extended vserver wwpn sticky),
##                ],
##            },
##            "NACL::C::Aggregate" => {
##                COMPARE_RTNS =>  { raidgroups => 
##                                       sub { 
##                                               if ("\"".$_[0] ." (block)\"" eq $_[1]) {
##                                                   return 1;
##                                               }
##                                               return 0;
##                                           },
##                                   thorough_scrub => [ \&Shoc::CompareSpec::treat_as_equiv, 
##                                                       equiv_values => [ [qw(- on), qw(- off)] ] 
##                                                     ],
##                },
##                IGNORE_KEYS => [ # May differ in normal operation
##                                 qw(availsize usedsize percent_used 
##                                    iron_last_start_errno iron_last_start_error_info),
##                                 # In BR but not RR (or vice versa)
##                                 qw(iron_scan_percentage
##                                    iron_status mirror_count_dp_destinations 
##                                    mirror_count_dp_qtree_destinations
##                                    mirror_count_ls_destinations mirror_count_move_destinations
##                                    volcount_not_online volcount_quiesced)
##                ],
##            },
##        },
##
##------------------------------------------------------------------------------
our $compareSpecs = [
    {
        LESSER  => qr/^Boilermaker__8.0/,
        GREATER => qr/^Boilermaker__8.0/,
        COMPONENT_SPEC => { 
            "NACL::C::Node" => {
                IGNORE_KEYS => [ qw(uptime) ],
            },
            "NACL::C::Interface" => {
                COMPARE_RTNS =>  \&Shoc::CompareSpec::cmp_lifs,
                IGNORE_KEYS => [ qw(curr_node curr_port is_home vserver sticky),
                               ],
            },
            "NACL::C::Aggregate" => {
                IGNORE_KEYS => [ qw(availsize usedsize percent_used 
                                    iron_last_start_errno iron_last_start_error_info),
                                 qw(iron_scan_percentage iron_status mirror_count_dp_destinations 
                                    mirror_count_dp_qtree_destinations
                                    mirror_count_ls_destinations mirror_count_move_destinations
                                    volcount_not_online volcount_quiesced)
                ],
             },
            "NACL::C::Volume" => {
                IGNORE_KEYS => [ qw(available used files_used snapshot_space_used
                                    percent_used overwrite_reserve unix_permission snapshot_count)
                ],
             },
        },
    },
    {
        LESSER  => qr/^RollingRock__8.1/,
        GREATER => qr/^RollingRock__8.1/,
        COMPONENT_SPEC => { 
            "NACL::C::Interface" => {
                COMPARE_RTNS =>  \&Shoc::CompareSpec::cmp_lifs,
                IGNORE_KEYS => [ qw(curr_node curr_port is_home server wwpn sticky),
                ],
            },
            "NACL::C::Aggregate" => {
                COMPARE_RTNS =>  { thorough_scrub      => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raid_block_checksum => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                },
                IGNORE_KEYS => [ # May differ in normal operation
                                 qw(availsize usedsize percent_used 
                                    sis_metadata_space_used
                                    volume_footprints volume_footprints_percent
                                    aggregate_metadata aggregate_metadata_percent
                                    used_including_snapshot_reserve used_including_snapshot_reserve_percent
                                    iron_last_start_errno iron_last_start_error_info),
                ],
            },
            "NACL::C::Volume" => {
                COMPARE_RTNS =>  { antivirus_on_access_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - none)]] ],
                },
                IGNORE_KEYS => [ qw(available used files_used snapshot_space_used
                                    percent_used overwrite_reserve unix_permission snapshot_count),
                ],
             },
            "NACL::C::Vserver" => {
                COMPARE_RTNS =>  { quota_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - default)]] ],
                },
                IGNORE_KEYS => [ qw(admin_state aggr_list allowed_protocols disallowed_protocols max_volumes)],
             },
            "NACL::C::StorageFailover" => {
                IGNORE_KEYS => [ qw(bnormal pnormal firmware_received 
                                    local_firmware_progress partner_firmware_progress transit_time
                                    logs_unsynced)],
             },
            "NACL::C::StorageDisk" => {
                IGNORE_KEYS => [ qw(initiator_iops initiator_io_kbps initiator_lun_in_use_count 
                                    is_reconstructing is_replacing is_scrubbing is_zeroing
                                    path_error_count path_link_errors
                                    path_iops path_io_kbps path_lun_in_use_count
                                    diskpathnames
                                    percent_rated_life_used percent_spares_consumed percent_spares_consumed_limit
                                    poweron_hours poweron_seconds
                                    effective_rpm
                                    average_latency position
                                    media_scrub_count media_scrub_last_done
                                    disk_io_kbps_total disk_iops_total
                                    shm_time_interval shm_timeinterval
                                    scrubcount
                                    scrub_count scrub_last_done scrub_last_done_seconds
                                    sectors_read sectors_written
                                    initiator nodelist target_wwpn tpgn
                                    target_io_kbps target_iops
                                    target_lun_in_use_count target_port_access_state
                                    blocksread blockswritten power_on_hours
                                    grown_defect_list_count
                                    usedblocks usedsize zeroingpercent
                                    lun_iops lun_io_kbps scrub_last_done)],
             },
            "NACL::C::Node" => {
                IGNORE_KEYS => [ qw(uptime) ],
            },
            "NACL::C::SystemNode" => {
                IGNORE_KEYS => [ qw(uptime)],
             },
            "NACL::C::SystemNodeAutosupport" => {
                IGNORE_KEYS => [ qw(last_subject last_time)],
             },
            "NACL::C::SystemServicesWebPackages" => {
                IGNORE_KEYS => [ qw(modified)],
             },
            "NACL::C::VserverCifsShare" => {
                IGNORE_KEYS => [ qw(acl) ],
             },
        },
    },
    {
        LESSER  => qr/^Boilermaker__8.0/,
        GREATER => qr/^RollingRock__8.1/,
        COMPONENT_SPEC => { 
            "NACL::C::Node" => {
                IGNORE_KEYS => [ qw(uptime) ],
            },
            "NACL::C::Interface" => {
                COMPARE_RTNS =>  \&Shoc::CompareSpec::cmp_lifs,
                IGNORE_KEYS => [ qw(curr_node curr_port is_home auto data_protocol status_extended vserver wwpn sticky),
                ],
            },
            "NACL::C::Aggregate" => {
                COMPARE_RTNS =>  { raidgroups => sub { if ("\"".$_[0] ." (block)\"" eq $_[1]) {return 1;} return 0;},
                                   thorough_scrub      => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raid_block_checksum => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                },
                IGNORE_KEYS => [ # May differ in normal operation
                                 qw(availsize usedsize percent_used 
                                    sis_metadata_space_used
                                    volume_footprints volume_footprints_percent
                                    aggregate_metadata aggregate_metadata_percent
                                    used_including_snapshot_reserve used_including_snapshot_reserve_percent
                                    iron_last_start_errno iron_last_start_error_info),
                                 # In BR but not RR
                                 qw(iron_scan_percentage
                                    iron_status mirror_count_dp_destinations mirror_count_dp_qtree_destinations
                                    mirror_count_ls_destinations mirror_count_move_destinations
                                    volcount_not_online volcount_quiesced)
                ],
            },
            "NACL::C::Volume" => {
                COMPARE_RTNS =>  { antivirus_on_access_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - none)]] ],
                },
                IGNORE_KEYS => [ qw(available used files_used snapshot_space_used
                                    percent_used overwrite_reserve unix_permission snapshot_count),
                                 qw(files_used percent_used snapshot_space_used volume_ownership
                                    auto_snapshots_enabled block_type compression_space_saved
                                    compression_space_saved_percent convert_ucode
                                    dedupe_space_saved dedupe_space_saved_percent
                                    extent_enabled filesys_size_fixed flexcache_cache_policy
                                    fractional_reserve in_nvfailed_state instance_uuid
                                    is_copied_for_transition is_quiesced_in_memory
                                    is_quiesced_on_disk is_sis_logging_enabled
                                    is_sis_volume is_transitioned create_ucode nvfail
                                    overwrite_reserve physical_replica read_realloc
                                    sched_snap_name provenance_uuid sis_space_saved
                                    sis_space_saved_percent snapshot_clone_dependency
                                    space_mgmt_try_first transition_state
                                    )
                ],
             },
            "NACL::C::Vserver" => {
                COMPARE_RTNS =>  { quota_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - default)]] ],
                },
                IGNORE_KEYS => [ qw(admin_state aggr_list allowed_protocols disallowed_protocols max_volumes)],
             },
            "NACL::C::VserverCifsShare" => {
                IGNORE_KEYS => [ qw(acl) ],
             },
        },
    },
    {
        LESSER  => qr/^SierraNevada__8.2/,
        GREATER => qr/^SierraNevada__8.2/,
        COMPONENT_SPEC => { 
            "NACL::C::Node" => {
                IGNORE_KEYS => [ qw(uptime) ],
            },
            "NACL::C::SystemNode" => {
                IGNORE_KEYS => [ qw(uptime)],
             },
            "NACL::C::Interface" => {
                COMPARE_RTNS =>  \&Shoc::CompareSpec::cmp_lifs,
                IGNORE_KEYS => [ qw(curr_node curr_port is_home server wwpn sticky),
                ],
            },
            "NACL::C::Aggregate" => {
                COMPARE_RTNS =>  { thorough_scrub      => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raid_block_checksum => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raidstatus          => [ \&Shoc::CompareSpec::cmp_csv,        allow_extra  => [ qw( redirect) ] ],
                },
                IGNORE_KEYS => [ # May differ in normal operation
                                 qw(availsize usedsize percent_used 
                                    sis_metadata_space_used
                                    volume_footprints volume_footprints_percent
                                    aggregate_metadata aggregate_metadata_percent
                                    used_including_snapshot_reserve used_including_snapshot_reserve_percent
                                    iron_last_start_errno iron_last_start_error_info),
                ],
            },
            "NACL::C::Volume" => {
                COMPARE_RTNS =>  { antivirus_on_access_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - none)]] ],
                },
                IGNORE_KEYS => [ qw(available used files_used snapshot_space_used
                                    percent_used overwrite_reserve unix_permission snapshot_count),
                ],
             },
            "NACL::C::Vserver" => {
                COMPARE_RTNS =>  { quota_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - default)]] ],
                },
                IGNORE_KEYS => [ qw(admin_state aggr_list allowed_protocols disallowed_protocols max_volumes)],
             },
            "NACL::C::StorageFailover" => {
                IGNORE_KEYS => [ qw(bnormal pnormal firmware_received
                                    local_firmware_progress partner_firmware_progress transit_time
                                    logs_unsynced)],
             },
            "NACL::C::StorageDisk" => {
                IGNORE_KEYS => [ qw(initiator_iops initiator_io_kbps initiator_lun_in_use_count
                                    is_reconstructing is_replacing is_scrubbing is_zeroing
                                    path_error_count path_link_errors
                                    path_iops path_io_kbps path_lun_in_use_count
                                    diskpathnames
                                    percent_rated_life_used percent_spares_consumed percent_spares_consumed_limit
                                    poweron_hours poweron_seconds
                                    effective_rpm
                                    average_latency position
                                    media_scrub_count media_scrub_last_done
                                    disk_io_kbps_total disk_iops_total
                                    shm_time_interval shm_timeinterval
                                    scrubcount rserver rserver_id
                                    scrub_count scrub_last_done scrub_last_done_seconds
                                    sectors_read sectors_written
                                    initiator nodelist target_wwpn tpgn
                                    target_io_kbps target_iops
                                    target_lun_in_use_count target_port_access_state
                                    blocksread blockswritten power_on_hours
                                    grown_defect_list_count
                                    usedblocks usedsize zeroingpercent
                                    lun_iops lun_io_kbps scrub_last_done)],
             },
            "NACL::C::SystemNodeAutosupport" => {
                IGNORE_KEYS => [ qw(last_subject last_time)],
             },
            "NACL::C::SystemServicesWebPackages" => {
                IGNORE_KEYS => [ qw(modified)],
             },
            "NACL::C::VserverCifsShare" => {
                IGNORE_KEYS => [ qw(acl) ],
             },
        },
    },
    {
        LESSER  => qr/^RollingRock__8.1/,
        GREATER => qr/^SierraNevada__8.2/,
        COMPONENT_SPEC => { 
            "NACL::C::Interface" => {
                COMPARE_RTNS =>  \&Shoc::CompareSpec::cmp_lifs,
                IGNORE_KEYS => [ qw(curr_node curr_port is_home server wwpn sticky),
                ],
            },
            "NACL::C::Aggregate" => {
                COMPARE_RTNS =>  { thorough_scrub      => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raid_block_checksum => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raidstatus          => [ \&Shoc::CompareSpec::cmp_csv,        allow_extra  => [ qw( redirect) ] ],
                },
                IGNORE_KEYS => [ # May differ in normal operation
                                 qw(availsize usedsize percent_used 
                                    sis_metadata_space_used
                                    volume_footprints volume_footprints_percent
                                    aggregate_metadata aggregate_metadata_percent
                                    used_including_snapshot_reserve used_including_snapshot_reserve_percent
                                    iron_last_start_errno iron_last_start_error_info),
                ],
            },
            "NACL::C::Volume" => {
                COMPARE_RTNS =>  { antivirus_on_access_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - none)]] ],
                },
                IGNORE_KEYS => [ qw(available used files_used snapshot_space_used
                                    percent_used overwrite_reserve unix_permission snapshot_count),
                ],
             },
            "NACL::C::Vserver" => {
                COMPARE_RTNS =>  { quota_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - default)]] ],
                },
                IGNORE_KEYS => [ qw(admin_state aggr_list allowed_protocols disallowed_protocols max_volumes)],
             },
            "NACL::C::StorageFailover" => {
                IGNORE_KEYS => [ qw(bnormal pnormal firmware_received 
                                    local_firmware_progress partner_firmware_progress transit_time
                                    logs_unsynced) ],
             },
            "NACL::C::StorageDisk" => {
                IGNORE_KEYS => [ qw(initiator_iops initiator_io_kbps initiator_lun_in_use_count 
                                    is_reconstructing is_replacing is_scrubbing is_zeroing
                                    path_error_count path_link_errors
                                    path_iops path_io_kbps path_lun_in_use_count
                                    diskpathnames
                                    percent_rated_life_used percent_spares_consumed percent_spares_consumed_limit
                                    poweron_hours poweron_seconds
                                    effective_rpm
                                    average_latency position
                                    media_scrub_count media_scrub_last_done
                                    disk_io_kbps_total disk_iops_total
                                    shm_time_interval shm_timeinterval
                                    scrubcount
                                    scrub_count scrub_last_done scrub_last_done_seconds
                                    sectors_read sectors_written
                                    initiator nodelist target_wwpn tpgn
                                    target_io_kbps target_iops
                                    target_lun_in_use_count target_port_access_state
                                    blocksread blockswritten power_on_hours
                                    grown_defect_list_count
                                    usedblocks usedsize zeroingpercent
                                    lun_iops lun_io_kbps scrub_last_done)],
             },
            "NACL::C::Node" => {
                IGNORE_KEYS => [ qw(uptime)],
             },
            "NACL::C::SystemNode" => {
                IGNORE_KEYS => [ qw(uptime)],
             },
            "NACL::C::SystemNodeAutosupport" => {
                IGNORE_KEYS => [ qw(last_subject last_time)],
             },
            "NACL::C::SystemServicesWebPackages" => {
                IGNORE_KEYS => [ qw(modified)],
             },
            "NACL::C::VserverCifsShare" => {
                IGNORE_KEYS => [ qw(acl) ],
             },
        },
    },
    {
        LESSER  => qr/^Fullsteam__8.3/,
        GREATER => qr/^Fullsteam__8.3/,
        COMPONENT_SPEC => { 
            "NACL::C::Node" => {
                IGNORE_KEYS => [ qw(uptime) ],
            },
            "NACL::C::SystemNode" => {
                IGNORE_KEYS => [ qw(uptime)],
             },
            "NACL::C::Interface" => {
                COMPARE_RTNS =>  \&Shoc::CompareSpec::cmp_lifs,
                IGNORE_KEYS => [ qw(curr_node curr_port is_home server wwpn sticky),
                ],
            },
            "NACL::C::Aggregate" => {
                COMPARE_RTNS =>  { thorough_scrub      => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raid_block_checksum => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raidstatus          => [ \&Shoc::CompareSpec::cmp_csv,        allow_extra  => [ qw( redirect) ] ],
                },
                IGNORE_KEYS => [ # May differ in normal operation
                                 qw(availsize usedsize percent_used 
                                    sis_metadata_space_used
                                    volume_footprints volume_footprints_percent
                                    aggregate_metadata aggregate_metadata_percent
                                    used_including_snapshot_reserve used_including_snapshot_reserve_percent
                                    iron_last_start_errno iron_last_start_error_info),
                ],
            },
            "NACL::C::Volume" => {
                COMPARE_RTNS =>  { antivirus_on_access_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - none)]] ],
                },
                IGNORE_KEYS => [ qw(available used files_used snapshot_space_used
                                    percent_used overwrite_reserve unix_permission snapshot_count),
                ],
             },
            "NACL::C::Vserver" => {
                COMPARE_RTNS =>  { quota_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - default)]] ],
                },
                IGNORE_KEYS => [ qw(admin_state aggr_list allowed_protocols disallowed_protocols max_volumes)],
             },
            "NACL::C::StorageFailover" => {
                IGNORE_KEYS => [ qw(bnormal pnormal firmware_received
                                    local_firmware_progress partner_firmware_progress transit_time
                                    logs_unsynced)],
             },
            "NACL::C::StorageDisk" => {
                IGNORE_KEYS => [ qw(initiator_iops initiator_io_kbps initiator_lun_in_use_count
                                    is_reconstructing is_replacing is_scrubbing is_zeroing
                                    path_error_count path_link_errors
                                    path_iops path_io_kbps path_lun_in_use_count
                                    diskpathnames
                                    percent_rated_life_used percent_spares_consumed percent_spares_consumed_limit
                                    poweron_hours poweron_seconds
                                    effective_rpm
                                    average_latency position
                                    media_scrub_count media_scrub_last_done
                                    disk_io_kbps_total disk_iops_total
                                    shm_time_interval shm_timeinterval
                                    scrubcount rserver rserver_id
                                    scrub_count scrub_last_done scrub_last_done_seconds
                                    sectors_read sectors_written
                                    initiator nodelist target_wwpn tpgn
                                    target_io_kbps target_iops
                                    target_lun_in_use_count target_port_access_state
                                    blocksread blockswritten power_on_hours
                                    grown_defect_list_count
                                    usedblocks usedsize zeroingpercent
                                    lun_iops lun_io_kbps scrub_last_done)],
             },
            "NACL::C::SystemNodeAutosupport" => {
                IGNORE_KEYS => [ qw(last_subject last_time)],
             },
            "NACL::C::SystemServicesWebPackages" => {
                IGNORE_KEYS => [ qw(modified)],
             },
            "NACL::C::VserverCifsShare" => {
                IGNORE_KEYS => [ qw(acl) ],
             },
        },
    },
    {
        LESSER  => qr/^SierraNevada__8.2/,
        GREATER => qr/^Fullsteam__8.3/,
        COMPONENT_SPEC => { 
            "NACL::C::Node" => {
                IGNORE_KEYS => [ qw(uptime) ],
            },
            "NACL::C::SystemNode" => {
                IGNORE_KEYS => [ qw(uptime)],
             },
            "NACL::C::NetworkPort" => {
                IGNORE_KEYS => [ qw(dcb_state)],
             },
            "NACL::C::Interface" => {
                COMPARE_RTNS =>  \&Shoc::CompareSpec::cmp_lifs,
                IGNORE_KEYS => [ qw(curr_node curr_port is_home server wwpn sticky),
                ],
            },
            "NACL::C::Aggregate" => {
                COMPARE_RTNS =>  { thorough_scrub      => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raid_block_checksum => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [ [qw(- on), qw(- off)] ] ],
                                   raidstatus          => [ \&Shoc::CompareSpec::cmp_csv,        allow_extra  => [ qw( redirect) ] ],
                },
                IGNORE_KEYS => [ # May differ in normal operation
                                 qw(availsize usedsize percent_used 
                                    sis_metadata_space_used
                                    volume_footprints volume_footprints_percent
                                    aggregate_metadata aggregate_metadata_percent
                                    used_including_snapshot_reserve used_including_snapshot_reserve_percent
                                    iron_last_start_errno iron_last_start_error_info),
                                  # in FS but not in SN
                                  qw(cluster cluster_id),
                ],
            },
            "NACL::C::Volume" => {
                COMPARE_RTNS =>  { antivirus_on_access_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - none)]] ],
                },
                IGNORE_KEYS => [ qw(available used files_used snapshot_space_used
                                    percent_used overwrite_reserve unix_permission snapshot_count),
                                 # in FS but not in SN
                                 qw(caching_policy junctionFileHandle snapshot_policy_uuid),
                                 # in SN but not in FS
                                 qw(external_cache node),
                ],
             },
            "NACL::C::Vserver" => {
                COMPARE_RTNS =>  { quota_policy => [ \&Shoc::CompareSpec::treat_as_equiv, equiv_values => [[qw( - default)]] ],
                },
                IGNORE_KEYS => [ qw(admin_state aggr_list allowed_protocols disallowed_protocols max_volumes),
                                 # in FS but not in SN
                                 qw(caching_policy config_lock snapshot_policy_uuid subtype),
                                 # in SN but not in FS
                                 qw(external_cache),
                ],
             },
            "NACL::C::StorageFailover" => {
                IGNORE_KEYS => [ qw(bnormal pnormal firmware_received
                                    local_firmware_progress partner_firmware_progress transit_time
                                    logs_unsynced)],
             },
            "NACL::C::StorageDisk" => {
                IGNORE_KEYS => [ qw(initiator_iops initiator_io_kbps initiator_lun_in_use_count
                                    is_reconstructing is_replacing is_scrubbing is_zeroing
                                    path_error_count path_link_errors
                                    path_iops path_io_kbps path_lun_in_use_count
                                    diskpathnames
                                    percent_rated_life_used percent_spares_consumed percent_spares_consumed_limit
                                    poweron_hours poweron_seconds
                                    effective_rpm
                                    average_latency position
                                    media_scrub_count media_scrub_last_done
                                    disk_io_kbps_total disk_iops_total
                                    shm_time_interval shm_timeinterval
                                    scrubcount rserver rserver_id
                                    scrub_count scrub_last_done scrub_last_done_seconds
                                    sectors_read sectors_written
                                    initiator nodelist target_wwpn tpgn
                                    target_io_kbps target_iops
                                    target_lun_in_use_count target_port_access_state
                                    blocksread blockswritten power_on_hours
                                    grown_defect_list_count
                                    usedblocks usedsize zeroingpercent
                                    lun_iops lun_io_kbps scrub_last_done)],
             },
            "NACL::C::SystemNodeAutosupport" => {
                IGNORE_KEYS => [ qw(last_subject last_time)],
             },
            "NACL::C::SystemServicesWebPackages" => {
                IGNORE_KEYS => [ qw(modified)],
             },
            "NACL::C::VserverCifsShare" => {
                IGNORE_KEYS => [ qw(acl) ],
             },
        },
    },
];

1;
