package Smoke;

use strict;
use warnings;

use Tharn;
use TestUtils;
use Options;
use ndisp::API;
use NATE::TestInfo;
use Mail::Sendmail qw(sendmail);
use Data::Dumper;

BEGIN {
	use Exporter;
	@Smoke::ISA = qw(Exporter);
	@Smoke::EXPORT_OK = qw(whichBubble runningInBubble cleanup parse_result move_out_of_pool send_cleanup_email);
}

sub whichBubble();
sub runningInBubble();
sub add_cleanup;
sub cleanup($$);
sub send_cleanup_email($$);
sub parse_result($);
sub move_out_of_pool();

sub _log   { $Tharn::Log->log(@_); }
sub _info  { $Tharn::Log->log(info_result(@_)); }
sub _warn  { $Tharn::Log->log(warn_result(@_)); }

sub whichBubble() {
	if (runningInBubble()) {
		if (open (WB, "/u/smoke/.site")) {
			my $bubble = <WB>;
			close(WB);
			return $bubble;
		} else {
			_warn("Failed to open /u/smoke/.site file.");
			return '';
		}
	} else {
		_warn("Not able to find which smoke bubble");
		return '';
	}
}

sub runningInBubble() {
	if (-e "/u/smoke/.site") {
		return 1;
	} else {
		return 0;
	}
}

sub add_cleanup {
	my $thpl_to_execute = shift;
	my $client = shift;
	my $script_owner = shift;
	my $params = shift;
	my @nodes = @_;
	if (!@nodes || !scalar(@nodes)) {
		_warn("No nodes specified to clean.");
		return undef;
	}
	my $ndispd = Tharn::param("TESTBED_NDISPD");

	if (!$ndispd) {
		_warn("No ndispd information, cannot proceed!!\n");
		return undef;
	}

	my $options = new Options;
	$options->parameters(
		definitions => [
		{ name => "NATE_LIB_UNIX", description => "NATE lib UNIX", default => undef },
		{ name => "MAIL_FROM", description => "Mail from", default => undef },
		{ name => "MAIL_TO", description => "Mail to", default => undef },
		{ name => "NATE_LIB_WINDOWS", description => "NATE lib windows", default => undef },
		{ name => "SITE_FILE", description => "Site specific config file", default => undef }
	]);
	$options->set_default("SITE_FILE" => "/u/smoke/p4/smoke/testbed/config/smoke_site_info.dat");
	$options->process();
	my $nlu = $options->get(option => "NATE_LIB_UNIX");
	my $nlw = $options->get(option => "NATE_LIB_WINDOWS");
	my $mail_to = $options->get(option => 'MAIL_TO');
	$mail_to =~ s'\\'\\\\'g;
	my $mail_from = $options->get(option => 'MAIL_FROM');
	$mail_from =~ s'\\'\\\\'g;
	my %Site_Info = TestUtils::get_site_info();
	my $smtp_server = $Site_Info{'smoke_bubble'}{'smtp_server'};
	if (!$client || $client =~ /^\s*$/) {
		_log("Getting the test controller from site info file.");
		$client = $Site_Info{'smoke_bubble'}{'general_use_linux_client'};
	}
	if (!$client) {
		_warn("TestController not specified.");
		return undef;
	}
	_log("Using TestController: $client");
	# Need test name without /
	my $test_name;
	if ($thpl_to_execute =~ /.*\//) {
		$test_name = $';
	} else {
		$test_name = $thpl_to_execute;
	}
	# Remove extension if any in test name
	$test_name =~ s/\.thpl$//;
	my $cmd = "`/u/smoke/p4/smoke/scripts/tcpath -branch NATE5 -p4 -script`/bin/ntest -s v2all.info -hw -noconsole CLIENT=\%e -client \\\$CLIENT ";
	$cmd .= "POSTJOB=\'tools/smoke/cleanup_result.thpl CLEAN_TEST_NAME=$test_name\' ";
	$cmd .= "NATE_LIB_UNIX=\'$nlu\' ";
	$cmd .= "NATE_LIB_WINDOWS=\'$nlw\' ";
	$cmd .= "SCRIPT_OWNER=$script_owner ";
	foreach my $param (keys %$params) {
		if ($params->{$param}) {
			$cmd .= "$param=$params->{$param} ";
		}
	}
	for (my $i = 0; $i < @nodes; $i++) {
		$cmd .= "CLEAN_NODE$i=%{join(',',res([(q(\$_->hostname =~ /($nodes[$i])/))]))} ";
	}
	my $nodesStr = join('-', @nodes);

	my $jobname = $ENV{JOBID};
	_log("Jobname is $jobname");

	my $api = new ndisp::API($ndispd);
	my $job_ref = $api->call('list jobs', [$jobname], 0);
	my ($reqname, $username, $action, $path);
	if($job_ref) {
		$reqname = $job_ref->{$jobname}->{'request'};
		if(!$reqname || $reqname =~ /^\s*$/) {
			_warn("Failed to find NDISP request name for the job $jobname.");
			return undef;
		}
		$action = $job_ref->{$jobname}->{'SMOKE_ACTION'};
		$path = $job_ref->{$jobname}->{'PATH'};
		my $req_ref = $api->call('list requests', [$reqname], 0);
		if($req_ref && defined $req_ref->{$reqname}) {
			$username = $req_ref->{$reqname}->{'username'};
			if (!$username || ($username =~ /^\s*$/)) {
				_warn("Failed to find username. Use smoke as username.");
				$username = 'smoke';
			}
		} else {
			_warn("Failed to get request reference from NDISP for the request $reqname");
			return undef;
		}
		_log("Request details are: Request name: $reqname, Username: $username, Smoke action: $action, Path: $path");
	} else {
		_warn("Failed to get job reference from NDISP for the job $jobname.");
		return undef;
	}
	$cmd .= "ORGI_USER=$username MAIL_FROM=$mail_from MAIL_TO=$mail_to SMTP_SERVER=$smtp_server ORIGINAL_JOB=$jobname ";
	$cmd .= $thpl_to_execute;
	my $rpri;
	my $rnum = int(rand(2000));
	if ($params->{'CleanUpJobOrder'} && ($params->{'CleanUpJobOrder'} =~ /^\s*$/)) {
		_log("Default priority used.");
		$rpri = 2000;
	} elsif ($params->{'CleanUpJobOrder'} && ($params->{'CleanUpJobOrder'} =~ /random/i)) {
		_log("Random priority used.");
		# Generate a random number between 2000 and 3000.
		$rpri = int( rand( 3000 - 2000 + 1 ) ) + 2000;
	} elsif ($params->{'CleanUpJobOrder'} && ($params->{'CleanUpJobOrder'} =~ /^\d+$/)) {
		_log("User defined priority used.");
		$rpri = 2100 - $params->{'CleanUpJobOrder'};
	} else {
		_log("Default priority used.");
		$rpri = 2000;
	}
	_log("Priority assigned to the new job is: $rpri");
	my $req = $api->call("add request",
		{	id                     => "TO_CLEAN.$nodesStr-$username-$rnum",
			jobname                => "TO_CLEAN.$nodesStr-$username-$rnum" . '.%{time()}',
			charges                => 1,
			start                  => "!req_exists('$jobname')",
			disable                => 0,
			SMOKE_ACTION           => $action,
			TAG                    => "Clean",
			priority               => $rpri,
			PATH                   => $path,
			username               => $username,
			reserve_all_or_nothing => 0,
			command                => $cmd,
			endpoint               => $client,
		},
		interactive => 0,
	);
	if (defined($req)) {
		_log("Launched NDISP request \"TO_CLEAN.$nodesStr-$username-$rnum\" with command: $cmd");
		return 1;
	} else {
		_warn("Could not submit cleanup request!!. Error: $@");
		return undef;
	}
}

sub cleanup($$) {
	my $thpl_to_execute = shift;
	if (!$thpl_to_execute || ($thpl_to_execute =~ /^\s*$/)) {
		_warn("No clean up thpl provided to execute.");
		return undef;
	} else {
		_log("Cleanup script to be executed: $thpl_to_execute");
	}
	my $params = shift;
	if (!$params || (ref($params) ne 'HASH')) {
		_warn("No parameters provided for the cleanup thpl. Parameters must be a reference to HASH");
		return undef;
	} else {
		_log("Parameters provided to the cleanup script: " . Dumper($params));
	}
	my @nodes = ();
	my $test_ctrl;
	my $script_owner;
	foreach my $param (keys %$params) {
		if ($param =~ /^TOCLEAN\d+$/i) {
			push @nodes, $params->{$param};
		} elsif ($param =~ /TestController/i) {
			$test_ctrl = $params->{$param};
		} elsif ($param =~ /CleanupScriptMaintainerUsername/i) {
			$script_owner = $params->{$param};
		}
	}
	if (!scalar(@nodes)) {
		_warn("No nodes specified to clean.");
		return undef;
	}
	if (!$script_owner || ($script_owner =~ /^\s*$/)) {
		_warn("Cleanup Script maintainer email must be provided.");
		return undef;
	}
	if (runningInBubble()) {
		_log("Running in " . whichBubble());
	} else {
		return undef;
	}
	add_cleanup($thpl_to_execute, $test_ctrl, $script_owner, $params, @nodes);
}

sub  parse_result($) {
	my $test_name = shift;
	if (!$test_name || ($test_name =~ /^\s*$/)) {
		_warn("Test name not specified.");
		return undef;
	}
	my $root = NATE::TestInfo->root();
	my @children = $root->children();
	_log("Getting the result for the test $test_name");
	foreach my $child (@children) {
		my $run_id = $child->runid();
		if ($run_id =~ /\d+_$test_name$/) {
			my @results = $child->all_results();
			if (scalar(@results)) {
				my $worst_result = NATE::Result::worst(@results);
				my $test_result = $worst_result->type();
				_log("Test \"$test_name\" worst result is: $worst_result and its type is: " . $test_result);
				if ("FATAL|FAIL|SCRIPT|CONFIG|INSPECT|UNKNOWN" =~ /$test_result/) {
					return 'FAIL';
				} else {
					return 'PASS';
				}
			} else {
				_log("No results reported by the test \"$test_name\". Treat it as PASS.");
				return 'PASS';
			}
		}
	}
	return undef;
}

sub move_out_of_pool() {
	my $jobname = $ENV{'JOBID'};
	my $ndispd = Tharn::param("TESTBED_NDISPD");
	my $api = new ndisp::API($ndispd);
	my $job_ref = $api->call('list jobs', [$jobname], 0);
	my @nodes;
	if($job_ref) {
		print Dumper($job_ref);
		my $ntest_cmd = $job_ref->{$jobname}->{'command'};
		@nodes = $ntest_cmd =~ /TOCLEAN\d+=(.*?) /g;
	}

	if (!-e "/u/smoke/presub/tharnhosts/temp_remove/cleanup_failed") {
		my @out = `/bin/mkdir -p /u/smoke/presub/tharnhosts/temp_remove/cleanup_failed 2>&1`;
		if ($? > 0) {
			_warn("Failed to create the directory /u/smoke/presub/tharnhosts/temp_remove/cleanup_failed. Exit status: $?, Output: @out");
			return undef;
		}
		_log("Created the directory: /u/smoke/presub/tharnhosts/temp_remove/cleanup_failed");
	}
	my $all_moved = 1;
	foreach my $node (@nodes) {
		if (-e "/u/smoke/presub/tharnhosts/$node") {
			my @out = `/bin/mv -f /u/smoke/presub/tharnhosts/$node /u/smoke/presub/tharnhosts/temp_remove/cleanup_failed 2>&1`;
			if ($? >0) {
				_warn("Failed to move tharnhost file u/smoke/presub/tharnhosts/$node to /u/smoke/presub/tharnhosts/temp_remove/cleanup_failed. Exit status: $?, Output: @out");
				$all_moved = 0;
			} else {
				_log("Moved tharnhost file /u/smoke/presub/tharnhosts/$node to /u/smoke/presub/tharnhosts/temp_remove/cleanup_failed");
			}
		} else {
			_warn("Tharnhost file /u/smoke/presub/tharnhosts/$node does not exists.");
			$all_moved = 0;
		}
	}
	return $all_moved;
}

sub send_cleanup_email($$) {
	my $subject = shift;
	my $message = shift;
	if (!$subject || ($subject =~ /^\s*$/)) {
		_warn("No subject specified to send email.");
		return undef;
	}
	if (!$message || ($message =~ /^\s*$/)) {
		_warn("No message specified to send email.");
		return undef;
	}

	my $options = new Options;
	$options->parameters(
		definitions => [
		{ name => "SMTP_SERVER", description => "Email server", default => undef },
		{ name => "MAIL_FROM", description => "Mail from", default => undef },
		{ name => "MAIL_TO", description => "Mail to", default => undef },
		{ name => "ORIGINAL_JOB", description => "Original job name", default => undef },
		{ name => "ORGI_USER", description => "Username that submitted the job", default => undef },
		{ name => "CGI_PATH", description => "Web link for the logs", default => undef },
		{ name => "LOGDIR", description => "Root directory of the logs volume", default => undef },
		{ name => "SCRIPT_OWNER", description => 'Script owner email', default => undef }
	]);
	$options->process();
	my %mail;
	$mail{'To'} = "dl-bundle-test-triage\@netapp.com";
	$mail{'Cc'} = '';
	if ($options->get(option => 'ORGI_USER') && $options->get(option => 'ORGI_USER') !~ /smoke/) {
		$mail{'Cc'} = $options->get(option => 'ORGI_USER') . "\@netapp.com";
	}
	if ($options->get(option => 'SCRIPT_OWNER') && ($options->get(option => 'SCRIPT_OWNER') !~ /^\s*$/)) {
		$mail{'Cc'} .= ", " . $options->get(option => 'SCRIPT_OWNER') . "\@netapp.com";
	}
	if ($options->get(option => 'MAIL_TO') && ($options->get(option => 'MAIL_TO') !~ /dl-bundle-test-triage/)) {
		if ($mail{'Cc'}) {
			$mail{'Cc'} .= ", " . $options->get(option => 'MAIL_TO');
		} else {
			$mail{'Cc'} = $options->get(option => 'MAIL_TO');
		}
	}
	$mail{'From'} = $options->get(option => 'MAIL_FROM');
	$mail{'Subject'} = $subject;
	my $cgi = $options->get(option => "CGI_PATH");
	my $logs_vol = $options->get(option => "LOGDIR");
	my $ojob = $options->get(option => "ORIGINAL_JOB");
	if ($logs_vol =~ '/logs') {
		$logs_vol = $`;
	}
	my $extra_message = "To view the jobs click the below links:\n\n";
	# http://web.smoke.eng.btc.netapp.in/natejobs_gx.cgi?testbed=/u/smoke/presub&logdir=TO_CLEAN.asdf-lkjh.1378470027
	$extra_message .= "Link to the original job that invoked the cleanup. $cgi?testbed=$logs_vol&logdir=$ojob\n";
	$extra_message .= "Link to the job that did the cleanup. $cgi?testbed=$logs_vol&logdir=$ENV{'JOBID'}\n";

	$message .= "\n\n$extra_message";

	$mail{'Message'} = $message;
	$Mail::Sendmail::mailcfg{'debug'} = 1;

        if (sendmail %mail) {
		_log("Mail successfully sent to $mail{'To'} and $mail{'Cc'}");
		return 1;
        } else {
		_warn("Failed to send email. $Mail::Sendmail::error");
		return undef;
        }
}

1
