# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/SnapCmds.pm#2 $

## @summary SnapCmds library -- modification subroutines
##
## @description
##    This package defines routines which attempt to perform snap*
##    operations that actually modify filer state, instead of
##    just gathering information from the filer and returning it.
##
##    Note: this module was designed to work only with volumes, but
##    it can work with aggregates if "-A " is prepended to
##    the name of the aggregate and passed whereever a
##    volume is passed:
##
##        snapshot_create($filer, "-A ".$aggrname, $snapname);
##
##
## @dependencies
##    lib/SayWrap.pm
##    FSR/FSRutils.pm
##
## @bugs
##
## @author dl-dpgqa-automation@netapp.com
## @status Shared
## MODIFIED FOR BURT (294627) by smallik@netapp.com
package SnapCmds;

use strict;

use vars qw(@EXPORT);
use Exporter qw(import);
BEGIN {
    @EXPORT = qw(snapshot_create snapshot_delete snaprestore_volume
		 snaprestore_file snaprestore_lun
		 snapreserve_volume snapsched_volume
		 snapshot_status snapshot_list snapshot_list_n snapshot_flags
		 snapshot_exists snaplist_identical snapshot_rename
                 snapshot_list_b snapshot_list_info
		 snapshot_autodelete_cmd snapshot_autodelete_options
                 snapshot_multicreate snapshot_multicreate_validate
		 snapshot_latest_id snapshot_list_loaded
		 restore_saved_aggr_snap_reserve
		);
}

use SayWrap;
use FSR::FSRutils;
use Tharn;
use Carp qw(croak);
use Filer::Version qw(get_wafl_version WV_TSINGTAO_MIN WV_IRONCITY_MIN);
use SPARSE::Util::Logger qw( logWarn logTrc logStep logDbg logInfo );
use Zapi;
use ZapiErrno;
use NaElement;
use Data::Dumper;

my $ERR_NO_SNAP = "No snapshots exist";


############################################################
#
# BURT 427608 -- "Accomodate change of default aggr snap reserve to 0"
#
#  Add an internal fcn: '_increase_aggr_snap_resv_if_needed', an exported
#  fcn: 'restore_saved_aggr_snap_reserve', and a (internal) hashtable,
#  %_saved_aggr_snap_reserve.
#
#  '_increase_aggr_snap_resv_if_needed' is called (only by) 'snapshot_create',
#  and its purpose is to verify for 'aggregate snapshots' that the 'snap reserve'
#  value for the aggregate is >= "$min_resv_pct" (3%), and if not, to
#  first store the current 'snap reserve' value for the aggregate in
#  %_saved_aggr_snap_reserve, and then change the 'snap reserve' value for the
#  aggregate to "$dflt_resv_pct" (5%).
#
#  If the option: "skip_aggr_snap_resv_check => 1" is passed to 'snapshot_create',
#  then no check of the snap reserve is done, and 'snapshot_create' functions
#  exactly as it did before this change. ("skp_aggr_snap_resv_check" is 0 by default)
#
#  'restore_saved_aggr_snap_reserve' is to be called from the 'cleanup' routine
#  of a script, either directly or by calling FS::Util::Setup::common_cleanup, which
#  then calls this fcn. 'restore_saved_aggr_snap_reserve' restores the original
#  'snap reserve' values for aggregates used in the script by accessing 
#  %_saved_aggr_snap_reserve.
#
#   change 10/20/10:  If increasing aggr snap reserve to "$dflt_resv_pct"
#                     Then will try to increase snap reserve to highest
#                     possible value less than "$dflt_resv_pct", but
#                     greater than current value.
#
###############################################################

use vars qw(%_saved_aggr_snap_reserve);

sub _increase_aggr_snap_resv_if_needed($$) {
    my ($fc, $aggr) = @_;
    my $min_resv_pct = 3;
    my $dflt_resv_pct = 5;
    my $timeout = 300;

    # 'normalize' the aggr name.
    $aggr =~ s/^\s+//;
    $aggr =~ s/\s+$//;
    $aggr =~ s/-A\s+/-A /;
    my $bare_aggr = $aggr;
    $bare_aggr =~ s/-A //;

    # Return unless the '$aggr' arg  starts with '-A'
    unless ($aggr =~ /^-A /) {
        return;
    }

    # Return if the aggr name is empty string.
    unless ($bare_aggr) {
        logresult('WARN', "SnapCmds::_increase_aggr_snap_resv_if_needed: Invalid aggr name: \"$bare_aggr\"");
        return;
    }

    # Return if the global param 'SKIP_AGGR_SNAP_RESV_CHECK' is set
    if ($main::SKIP_AGGR_SNAP_RESV_CHECK) {
        return;
    }

    # get existing snap resv for aggr
    my ($orig_resv_pct, undef) = snapreserve_volume($fc, $aggr);
    unless (defined($orig_resv_pct)) {
        return;
    }

    # Return if the current snap reserve value is greater than the minimum required value.
    if ($orig_resv_pct >= $min_resv_pct) {
        return;
    }

    # Initialize the '%_saved_aggr_snap_reserve' hash if necessary.
    if (!%_saved_aggr_snap_reserve) {
        %_saved_aggr_snap_reserve = ();
    }

    # Create a hashtable for this host, if it doesn't exist
    my $hostname = $fc->hostp()->hostname();
    my $aggr_reserve_for_host = $_saved_aggr_snap_reserve{$hostname};
    unless ($aggr_reserve_for_host) {
        $aggr_reserve_for_host = {};
        $_saved_aggr_snap_reserve{$hostname} = $aggr_reserve_for_host;
    }

    #  Return if there is already a 'snap reserve' value stored for this aggr
    if (exists($aggr_reserve_for_host->{$aggr})) {
        return;
    }

    # Store the current 'snap reserve' value for the aggregate so that it may be restored later.
    $aggr_reserve_for_host->{$bare_aggr} = $orig_resv_pct;

    # Now increase the 'snap reserve' value for the aggregate to that of "$min_resv_pct" (initially 5%)
    #
    #   change 10/20/10:  If can't increase aggr snap reserve to "$dflt_resv_pct"
    #                     Then will try to increase snap reserve to highest
    #                     possible value less than "$dflt_resv_pct", but
    #                     greater than current value.
    #
    my $resv_is_set = 0;
    my $new_resv_pct = 0;
    for ($new_resv_pct=$dflt_resv_pct; ($new_resv_pct > $orig_resv_pct); $new_resv_pct--) {
        $resv_is_set = snapreserve_volume($fc, $aggr, $new_resv_pct);
        last if ($resv_is_set);
    }
    if ($resv_is_set) {
        my ($verify_resv_pct, undef) = snapreserve_volume($fc, $aggr);
        if ($verify_resv_pct == $new_resv_pct) {
            logresult('INFO', "SnapCmds::_increase_aggr_snap_resv_if_needed:\n".
                      "Increased 'snap reserve percent' for aggr: \"$bare_aggr\" ".
                      "from $orig_resv_pct to $new_resv_pct: $@");
            return;
        }
    }
    logresult('WARN', "SnapCmds::_increase_aggr_snap_resv_if_needed:\nCould not increase\n".
                          "'snap reserve percent' for aggr: \"$bare_aggr\" from $orig_resv_pct: $@");
}


## @name ::restore_saved_aggr_snap_reserve
## @usage restore_saved_aggr_snap_reserve($fc)
## @arg fc   (Required):   Connection to filer (CONNECTREC). 
## @synopsis
##    Introduced for BURT 427608 -- "Accomodate change of default aggr snap reserve to 0"
##
##    Intended to be called from script "cleanup" routine.
##
##    Will restore as necessary any "snapshot reserve" settings for aggregates which were
##    modified as a result of a call to "snapshot_create". Has no effect if no aggregate
##    snapshots were created. May be called multiple times.
sub restore_saved_aggr_snap_reserve($) {
    my ($fc) = @_;
    my $hostname = $fc->hostp()->hostname();
    my $aggr_reserve_for_host = $_saved_aggr_snap_reserve{$hostname};
    my $timeout = 300;

    # Return if the '%_saved_aggr_snap_reserve' table is undefined.
    if (!%_saved_aggr_snap_reserve) {
        return;
    }

    # Return if there are no stored values for this host.
    unless ($aggr_reserve_for_host) {
        return;
    }

    # Loop over each entry in the 'aggr snap reserve' table for this host to restore the original values.
    my @aggrs = keys(%$aggr_reserve_for_host);
    foreach my $bare_aggr (@aggrs) {
        my $aggr = "-A " . $bare_aggr;
        my $saved_resv_pct = $aggr_reserve_for_host->{$bare_aggr};

        # Skip to next aggr if this aggr does not exist.
	unless (Aggregate::aggr_exists($fc, $bare_aggr)) {
            next;
        }

        # Restore the original snap reserve value.
        unless (snapreserve_volume($fc, $aggr, $saved_resv_pct)) {
            logresult('WARN', "SnapCmds::restore_saved_aggr_snap_reserve: Could not restore\n".
                              "'snap reserve' for aggr: \"$bare_aggr\" to $saved_resv_pct: $@");
            next;
        }

        # Verify we restored the value.
        my ($resv_pct, undef) = snapreserve_volume($fc, $aggr);
        if ($resv_pct != $saved_resv_pct) {
            logresult('WARN', "SnapCmds::restore_saved_aggr_snap_reserve: Could not set\n".
                              "'snap reserve' for aggr: \"$bare_aggr\" to $saved_resv_pct");
            next;
        }
    }

    # Delete the 'aggr snap reserve' table for this host, so we don't try to restore again later.
    delete($_saved_aggr_snap_reserve{$hostname});
    return(1);
}


## @name ::snapshot_create
## @usage snapshot_create($filer, $volume, $snapname, %options)
## @arg filer     (Required):   Connection to filer (CONNECTREC). 
## @arg volume    (Required):   Name of volume or aggregate for which to create the snapshot.
##                              For aggregates, name must be preceeded by "-A".
## @arg snapname  (Required):   Name of snapshot to create.
## @arg use_zapi (optional): use the ZAPI interface instead of the CLI interface
## @arg async (optional): return immediately instead of waiting (ZAPI only)
## @arg lock_backing (optional): lock the LUN backing snapshots (ZAPI only)
## @arg skip_aggr_snap_resv_check (optional): If set, restores behavior prior
##      to change made for BURT 427608 -- "Accomodate change of default aggr snap reserve to 0"
##      Default is 0. 
## @synopsis
##    Attempts to create a snapshot on the given volume.
##    Returns success or failure, with an error message in $@ on failure.
sub snapshot_create($$$;%) {
  my ($filer, $volume, $snapname, %options) = @_;

  # Check for 'aggregate' snapshot:

  ########################################################
  #
  # BURT 427608 -- Accomodate change of default aggr snap reserve to 0
   unless ($options{skip_aggr_resv_check}) {
       _increase_aggr_snap_resv_if_needed($filer, $volume);
  }
  #
  ########################################################

  if ($options{use_zapi} or $main::USE_ZAPI_IF_AVAILABLE) {
      my $zcmd = new NaElement("snapshot-create");
      $zcmd->child_add(new NaElement("volume", $volume));
      $zcmd->child_add(new NaElement("snapshot", $snapname));
      my $async = $options{async} ? $options{async} : "false";
      $zcmd->child_add(new NaElement("async", $async));
      $zcmd->child_add(new NaElement("is-valid-lun-clone-snapshot", "true"))
        if $options{lock_backing};

      logDbg("command sent: ". $zcmd->toString());
      my $zapi = Zapi->new($filer->hostp()->default_ip(),
                           "root",
                           $filer->hostp()->default_password());
      $zapi->{fail} = \&_fail_default;

      my $xo = $zapi->invoke_elem($zcmd, "any");
      logDbg("snapshot-create ZAPI returns: ". $xo->toString());
      if ( ! $xo->results_ok() ) {
          $@ = "ZAPI snapshot-create FAILed: ". $xo->toString();
          return 0;
      }
      return 1;
  }
  return eval {
    # Let timeout default to at least 5 minutes
    my $timeout = $filer->timeout > 300 ? $filer->timeout : 300;
    my $text = sayEOF($filer, "snap create $volume $snapname", timeout => $timeout);
    $text =~ s/^creating snapshot\.\.\.(\n|$)//;
    die "$text\n" if ($text);
    return 1;
  }
}

## @name ::snapshot_delete
## @usage snapshot_delete($filer, $volume, $snapname, [$options], [$use_zapi])
## @synopsis
##    Attempts to delete a snapshot or all snapshots on the given volume.
##    Returns success or failure, with an error message in $@ on failure.
sub snapshot_delete($$$;$$) {
  my ($filer, $volume, $snapname, $options, $use_zapi) = @_;

  if ($use_zapi or $main::USE_ZAPI_IF_AVAILABLE) {
      return eval {
          logDbg("snapshot-delete snap $snapname on vol $volume");
          my $zcmd = new NaElement("snapshot-delete");
          $zcmd->child_add(new NaElement("volume", $volume));
          $zcmd->child_add(new NaElement("snapshot", $snapname));

          my $zapi = Zapi->new($filer->hostp()->default_ip(),
                               "root",
                               $filer->hostp()->default_password());
          $zapi->{fail} = \&_fail_default;

          my $xo = $zapi->invoke_elem($zcmd, "any");
          logDbg("snapshot-delete ZAPI returns: ". $xo->toString());
          if ( ! $xo->results_ok() ) {
              die "ZAPI snapshot-delete returns: ". $xo->toString();
          }
          return 1;
      }
  }
  return eval {
    my $text;
    if (defined($options)) {
        if ($options =~ /\-a/) {
            $text = sayEOF($filer, "snap delete $options -f -q $volume", timeout=>$::CMDTIMEOUT);
        }
        else {        # "-A" or "-V"
            $text = sayEOF($filer, "snap delete $options $volume $snapname", timeout=>$::CMDTIMEOUT);
        }
    }
    else {
        $text = sayEOF($filer, "snap delete $volume $snapname", timeout=>$::CMDTIMEOUT);
    }
    $text =~ s/^deleting snapshot\.\.\.(\n|$)//;
    die "$text\n" if ($text);

    return 1;
  }
}

## @name ::snapshot_rename
## @usage snapshot_rename($filer, $vol, $snap_from, $snap_to, %options)
## @synopsis
##    Attempts to rename a snapshot on the give volume.
##    Returns success or failure, with error message in $@ on failure.
sub snapshot_rename($$$$;%) {
  my ($filer, $volume, $snap_from, $snap_to, %options) = @_;

  if ($options{use_zapi} or $main::USE_ZAPI_IF_AVAILABLE) {
      my $zcmd = new NaElement("snapshot-rename");
      $zcmd->child_add(new NaElement("volume", $volume));
      $zcmd->child_add(new NaElement("current-name", $snap_from));
      $zcmd->child_add(new NaElement("new-name", $snap_to));

      my $zapi = Zapi->new($filer->hostp()->default_ip(),
                           "root",
                           $filer->hostp()->default_password());
      $zapi->{fail} = \&_fail_default;

      my $xo = $zapi->invoke_elem($zcmd, "any");
      logDbg("snapshot-rename ZAPI returns: ". $xo->toString());
      if ( ! $xo->results_ok() ) {
          $@ = "ZAPI snapshot-rename returns: ". $xo->toString();
          return 0;
      }
      return 1;
  }
  return eval {
    my $text = sayEOF($filer, "snap rename $volume $snap_from $snap_to");
    die "$text\n" if ($text);
    return 1;
  }
}

## @name ::snapshot_exists
## @usage snapshot_exists($filer, $volume, $snap_name)
## @synopsis
##    Return whether a snapshot of specific name exists.
##    On error, an error message is placed in $@
sub snapshot_exists($$$) {
  my ($filer, $volume, $snap_name) = @_;

  return eval {
    my @dst_snap_list = snapshot_list($filer, $volume);
    if ($@ =~ /volume $volume is offline/) {
      # Due to a bug in slideruler
      sleep 10;
      @dst_snap_list = snapshot_list($filer, $volume);
    }
    die $@ if ($@);

    my $snap;
    while (defined ($snap = pop(@dst_snap_list))) {
      last if ($snap eq $snap_name)
    }

    if (defined $snap) {
      return 1;
    } else {
      return 0;
    }
  };
}

## @name ::snaplist_identical
## @usage snaplist_identical($filer1, $vol1, $filer2, $vol2)
## @synopsis
##	Return whether the list of snapshots on $vol1 and $vol2
##	are identical.
sub snaplist_identical($$$$) {
  my ($filer1, $vol1, $filer2, $vol2) = @_;

  my (@snap_list1, @snap_list2);
  my ($snap1, $snap2);

  @snap_list1 = snapshot_list($filer1, $vol1);
  @snap_list2 = snapshot_list($filer2, $vol2);

  while (1) {
    $snap1 = pop(@snap_list1);
    $snap2 = pop(@snap_list2);

    return 1 if (!$snap1 && !$snap2);

    return 0 if ($snap1 ne $snap2);
  }
}

## @name ::snaprestore_volume
## @usage snaprestore_volume($filer, $volume, $snapshot)
## @synopsis
##    Does a snaprestore of the given volume to the given snapshot.
##    Returns success or failure, with an error message in $@ on failure.
sub snaprestore_volume($$$;%) {
    my ($filer, $volume, $snapname, %opts) = @_;

    $opts{timeout} = $::CMDTIMEOUT unless defined $opts{timeout};
    return eval {
	my $text = sayEOF($filer, "snap restore -f -s $snapname $volume", %opts);
	die "No output\n" unless ($text);
	die "$text\n" unless ($text =~ /revert successful\./);
	return 1;
    }
}

## @name ::snaprestore_file
## @summary Restore a file from a snapshot
## @usage snaprestore_file($filer, $file_path, $snapshot [, restore_as_path=><path> ])
## @args filer Required: hostspec of filer
## @args lun_path Required: path to the lun to be restored
## @args snapshot Required: name of snapshot from where to be restored
## @args restore_as_path Optional: File will be restored to this path if specfied
## @synopsis
##    Does a snaprestore of the given file to the given snapshot.
##    The file is restored to its original path if restore_as_path is not specified.
##    If restore_as_path is  specified, the path must be a full path to a filename,
##    and must be in the same volume as the volume used for the restore.
##    Returns success or failure, with an error message in $@ on failure.
##
## @notes Dependencies for sfsr
##    snaprestore must be licensed
##    volume must be online
##    volume must not be full
##    restore path if specified must be in same volume
##    cannot restore directories, files with NT Streams
##    exisiting files will be overwritten

# Possible error messages (for reference)
# 1. SNAPRESTORE is not licensed.
#    (Use the "license" command to license it.)
# 2. snap restore: File does not exist in snapshot.
# 3. snap restore: Must specify file name to be restored as.
# 4. snap restore: Volume <volname> does not have a snapshot named <snapname>.
# 5. snap restore: volume name or path have to be specified.
# 6. snap restore: Snapshot name is required for -s option.
# 7. snap restore error: No space left on device.
# 8. snap restore: volname or pathname <path> does not exist.
# 9. snap restore: restore_as_path and restore_from_path have to be specified in the same volume.

# Usage: snap restore [-A | -V] [-f] [-t vol | file ] [-s name]
#                     [-r <restore_as_path>] <vol-name> | <restore_from_path>

sub snaprestore_file($$$%) {

  my ($filer, $path, $snapname, %args) = @_;

  my $restore_as_path = defined($args{restore_as_path}) ? $args{restore_as_path} : '';

  return eval {

    my $cmd;
    if ($restore_as_path) {
        $cmd = "snap restore -f -s $snapname -r $restore_as_path $path";
    } else {
        $cmd = "snap restore -f -s $snapname $path";
    }
    my $text = sayEOF($filer, $cmd, timeout => $::CMDTIMEOUT);
    # this command does not put out any output via rsh - never has
    # so success if there is no output
    # burt 97022
    die "$text\n" if ($text);
    return 1;
  }
}

## @name ::snaprestore_lun
## @summary Restore a lun from a snapshot
## @usage snaprestore_lun($filer, $lun_path, $snapshot [, restore_as_path=><path> ])
## @args filer Required: hostspec of filer
## @args lun_path Required: path to the lun to be restored
## @args snapshot Required: name of snapshot from where to be restored
## @args restore_as_path Optional: Lun will be restored to this path if specfied
## @synopsis
##    Does a snaprestore of the given lun to the given snapshot.
##    The lun is restored to its original path if restore_as_path is not specified.
##    If restore_as_path is  specified, the path must be a full lun path,
##    and must be in the same volume as the volume used for the restore.
##    Returns success or failure, with an error message in $@ on failure.

sub snaprestore_lun($$$%) {
    &snaprestore_file(@_);
}

## @name ::snapreserve_volume
## @usage snapreserve_volume($filer, $volume, $reserve_percent)
## @synopsis
##   Gets or sets the snap reserve for the specified volume.
##   The third argument is optional.  If given, it specifies the
##   percentage to set the snap reserve to, and the function returns
##   success or failure with an error message in $@ on failure. If the
##   third argument is omitted, and the function returns a tuple of
##   ($reserve_percent, $kilobytes), with an error message in $@ on failure.
sub snapreserve_volume($$;$) {
  my ($filer, $volume, $reserve) = @_;

  return eval {

    if (defined $reserve) {
      # Set the snap reserve
      my $text = sayEOF($filer, "snap reserve $volume $reserve", timeout => $::CMDTIMEOUT);
      die "$text\n" if ($text);
      return 1;

    } else {
      # Determine the snap reserve
      my $text = sayEOF($filer, "snap reserve $volume", timeout => $::CMDTIMEOUT);
      die "$text\n"
	unless ($text =~ /current snapshot reserve is (\d+)% or (\d+) k-bytes\./);
      my ($percent, $kb) = ($1, $2);
      return ($percent, $kb);
    }

  }
}

## @name ::snapsched_volume
## @usage snapsched_volume($filer, $volume, @settings)
## @synopsis
##   Gets or sets the snap schedule for the specified volume.
##   If given settings specify the snap schedul for the volume.
##   Return success or failure with an error message in $@ on failure.
##   Success is either 1 or a list of the current settings if no @settings
##   are given.
##   Failure sets $@.
sub snapsched_volume($$;@) {
  my ($filer, $volume, @settings) = @_;

  return eval {

    if (defined $settings[0]) {
      # Set the snap schedule
      my $text = sayEOF($filer, "priv set -q advanced; snap sched $volume @settings",
                        timeout => $::CMDTIMEOUT);
      die "$text\n" if ($text);
      return 1;

    } else {
      # Determine the snap schedule
      my $text = sayEOF($filer, "snap sched $volume", timeout => $::CMDTIMEOUT);
      die "$text\n"
	unless ($text =~ s/(Aggregate|Volume) .*: //);
      return (split(' ', $text));
    }

  }
}

## @name ::snapshot_status
## @usage snapshot_list($filer, $volume)
## @synopsis
##    Returns a list of all the snapshots in the specified volume
##    returned by advance command 'snap status' as array of hash
##    references, keys are taken from header
##
##    Volume vol0
##    snapid  status     date           ownblks release fsRev name
##    ------  ------     ------------   ------- ------- ----- --------
##    245     complete   Nov 03 20:00      6941     7.2    71 hourly.0
##    243     complete   Nov 03 16:00      2141     7.2    71 hourly.1
##
##    Volume vl00327_874b_c95a8134 (cleaning summary map)
##    snapid  status     date           ownblks release fsRev name
##    ------  ------     ------------   ------- ------- ----- --------
##      1     creating   Nov 03 21:53         0     7.2    71 fsr-u60(0033604436)_vl00327_874b_c95a8134.1 (no map)
##

sub snapshot_status($$) {
	my ($filer, $volume) = @_;
	my @snapshots = ();
	my $text = sayEOF($filer, "priv set -q advanced;snap status $volume");
	croak($text) if( $text =~ /does not exist or is not online/ );	

	my @hkeys = ();
	my $sncnt=0;
	foreach( split( '[\n\r]',$text) ) {
		chomp; s/^\s+//;
		s/(\w\w\w)\s+(\d\d)\s+(\d\d:\d\d)/$1_$2_$3/;	# work around dates
		@hkeys = split('\s+',$_) if( /snapid/ );
		if( /^[0-9]/ ) {
			my $i=-1;
			foreach( split('\s+',$_) ) {
				s/_/ /g if($hkeys[++$i] =~ /date/ );	# restore space in dates
				$snapshots[$sncnt]->{ $hkeys[$i] } = $_ if($hkeys[$i]);
			}
			$sncnt++;
		}
	}
	return(@snapshots);
}

## @name ::snapshot_list
## @usage snapshot_list($filer, $volume, %options)
## @synopsis
##    Returns a list of all the snapshots in the specified volume
sub snapshot_list($$;%) {
  my ($filer, $volume, %options) = @_;

  if ($options{use_zapi} or $main::USE_ZAPI_IF_AVAILABLE) {
      return eval {
          my @snapshots = ();
          my @snaps = snapshot_list_info($filer, $volume, %options);
          foreach my $snap (@snaps) {
              my %list = %{$snap};
              push(@snapshots, $list{name});
          }
          return @snapshots;
      }
  }
  return eval {
    my @snapshots = ();
    my $text = sayEOF($filer, "snap list $volume", timeout => $::CMDTIMEOUT);

    die "$text" if( $text =~ /Volume.*does not exist or is not online/ );

    # Return if list is empty
    return @snapshots
      if ($text =~ /$ERR_NO_SNAP/i);

    # Strip everything up to the beginning of the list
    # (Keep everything after the "----------  ----------  ------------  --------")
    die "Error stripping output header: $text\n"
      unless ($text =~ s/.*?\s*\-+\s+\-+\s+\-+\s+\-+\s*(.*)$/$1/s);

    chomp $text;

    my @lines = split(/\r?\n/,$text);
    @snapshots = map{
      die "Error parsing snapshot list: $_\n"
      unless (/^(\s*\S+\s*)*?(\S+)\s*(\(\S+\))?$/);
      $2
    } @lines;
    return @snapshots;
  }
}

## @name ::snapshot_list_info
## @usage snapshot_list_info($filer, $volume, %args)
## @arg type (optional): "volume" (default) or "aggregate"
## @arg terse (optional): don't include computed statistics
## @arg snapowners (optional): return an array of snapshot owners
## @arg lun_clone (optional): check if snapshot contains LUN-clones
## @synopsis Uses ZAPI to return an array of hashes of all the snapshot
##    info available or requested.
sub snapshot_list_info($$;%)
{
    my ($filer, $volume, %options) = @_;

    my $zcmd = new NaElement("snapshot-list-info");
    $zcmd->child_add(new NaElement("target-name", $volume));
    $zcmd->child_add(new NaElement("target-type", ($options{type} ? $options{type} : "volume")));
    $zcmd->child_add(new NaElement("terse", $options{terse}))
      if $options{terse};
    $zcmd->child_add(new NaElement("snapowners", $options{snapowners}))
      if $options{snapowners};
    $zcmd->child_add(new NaElement("lun-clone-snapshot", $options{lun_clone}))
      if $options{lun_clone};

    my $zapi = Zapi->new($filer->hostp()->default_ip(),
                         "root",
                         $filer->hostp()->default_password());
    $zapi->{fail} = \&_fail_default;

    my $xo = $zapi->invoke_elem($zcmd, "any");
    logDbg("snapshot-list-info ZAPI returns: ". $xo->toString());

    my @snapresult;
    if ( ! $xo->results_ok() ) {
        if ($xo->name() eq "results") {
            die "snapshot-list-info FAILed: status=>$xo->attr_get('status'), errno=>$xo->attr_get('errno'), reason=>$xo->attr_get('reason')";
        } else {
            die "snapshot-list-info FAILed for unknown reason";
        }
    }
    my $vo = $xo->child_get("snapshots");
    if ($vo) {
        my $num_elems = $vo->child_max_number();
        logDbg("child-get(snapshots) returned $num_elems children");
        for (my $i=0; $i < $num_elems; $i++) {
            my %result = ();
            my $vl = $vo->child_get_nth($i);
            $result{name} = $vl->child_get("name")->{content};
            $result{total} = $vl->child_get("total")->{content}
              if $vl->child_get("total");
            $result{"cumulative-total"} = $vl->child_get("cumulative-total")->{content}
              if $vl->child_get("cumulative-total");
            $result{"access-time"} = $vl->child_get("access-time")->{content};
            $result{busy} = $vl->child_get("busy")->{content};
            $result{dependency} = $vl->child_get("dependency")->{content};
            $result{"percentage-of-used-blocks"} = 
              $vl->child_get("percentage-of-used-blocks")->{content};
            $result{"cumulative-percentage-of-used-blocks"} = 
              $vl->child_get("cumulative-percentage-of-used-blocks")->{content};
            $result{"percentage-of-total-blocks"} =
              $vl->child_get("percentage-of-total-blocks")->{content};
            $result{"cumulative-percentage-of-total-blocks"} =
              $vl->child_get("cumulative-percentage-of-total-blocks")->{content};
            $result{"contains-lun-clones"} =
              $vl->child_get("contains-lun-clones")->{content}
                if $vl->child_get("contains-lun-clones");
            my $vol = $vl->child_get("snapshot-owners-list");
            if ($vol) {
                my $numelem = $vol->child_max_number();
                my @owners;
                for (my $j=0; $j < $numelem; $j++) {
                    my $voll = $vol->child_get_nth($j);
                    my $volo = $voll->child_get("snapshot-owner")->{content};
                    push(@owners, $volo);
                }
                $result{"snapshot-owners-list"} = (@owners);
            }
            push(@snapresult, {%result});
        }
    }
    return(@snapresult);
}

## @name ::snapshot_list_n
## @usage snapshot_list_n($filer, $volume)
## @synopsis
##    Returns a list of all the snapshots in the specified volume
##    without regard to the sizes of the snapshots.  Returns much faster if
##    you just want to know if a snapshot exists or not.
##    TODO: In the future the function snapshot_exists should probably be
##    changed to use this function.
sub snapshot_list_n($$;%) {
  my ($filer, $volume, %options) = @_;

  if ($options{use_zapi} or $main::USE_ZAPI_IF_AVAILABLE) {
      return eval {
          my @snapshots = ();
          my @snaps = snapshot_list_info($filer, $volume, terse => "true", %options);
          foreach my $snap (@snaps) {
              my %list = %{$snap};
              push(@snapshots, $list{name});
          }
          return @snapshots;
      }
  }
  return eval {
    my @snapshots = ();
    my $text = sayEOF($filer, "snap list -n $volume");

    # Return if list is empty
    return @snapshots
      if ($text =~ /$ERR_NO_SNAP/i);

    # Strip everything up to the beginning of the list
    # (Keep everything after the "------------  --------")
    die "Error stripping output header: $text\n"
      unless ($text =~ s/.*?\s*\-+\s+\-+\s*(.*)$/$1/s);

    chomp $text;

    my @lines = split(/\r?\n/,$text);
    @snapshots = map{
      die "Error parsing snapshot list: $_\n"
      unless (/^(\S+\s\d+\s\d+\:\d+\s+)(\S+)/);
      $2
    } @lines;
    return @snapshots;
  }
}

## @name ::snapshot_list_b
## @usage snapshot_list_b($filer, $volume)
## @synopsis
##    Returns an array of snapshot lock descriptors for all the snapshots in
##    the specified volume. Each snapshot lock descriptor is a hash of the
##    form (name owner)
sub snapshot_list_b($$) {
  my ($filer, $volume) = @_;

  return eval {
    my @snapshots = ();
    my $text = sayEOF($filer, "snap list -b $volume", timeout => $::CMDTIMEOUT);

    # Return if list is empty
    return @snapshots
      if ($text =~ /$ERR_NO_SNAP/i);

    # Strip everything up to the beginning of the list
    # (Keep everything after the "----------  -----------")
    die "Error stripping output header: $text\n"
      unless ($text =~ s/.*?\s*\-+\s+\-+\s*(.*)$/$1/s);

    chomp $text;

    foreach my $line (split(/\n/, $text)) {
      if ($line =~ /^\s*(\S+)\s*(.*)$/) {
	  my %snap;

	  $snap{'name'} = $1;
	  $snap{'owner'} = $2;
	  push (@snapshots, {%snap});
      }
    }
    return @snapshots;
  }
}

## @name ::snapshot_autodelete_cmd
## @usage snapshot_autodelete_cmd($filer, $volume[, $cmd])
## @synopsis
##   snap autodelete cmds: on, off, reset, show
##   If cmd is "show", it returns a hash of autodelete options value
##   For other cmds, return 1 when succeed; otherwise croak.
##
sub snapshot_autodelete_cmd($$;$) {
  my ($filer, $volume, $in_cmd) = @_;

  # snap autodelete is available after TT
  my $MAX = 999999;
  my $TT_MIN = WV_TSINGTAO_MIN;
  my %mapping = ("on"=> {$TT_MIN=>"on", $MAX=>"on"},
                 "off"=> {$TT_MIN=>"off", $MAX=>"off"},
                 "reset"=> {$TT_MIN=>"reset", $MAX=>"reset"},
                 "show"=> {$TT_MIN=>"show", $MAX=>"show"}
                 );

  my %output = ("on", "autodelete enabled",
                "off", "autodelete disabled",
                "reset", "reset to default",
                "show", "autodelete settings");

  if (!defined($in_cmd)) {
      $in_cmd = "show";
  }

  if (!defined($output{$in_cmd})) {
      croak("invalid input for snap autodelete: $in_cmd");
  }

  my $version = get_wafl_version($filer->{hostp});
  my $cmd_list_p = $mapping{$in_cmd};
  my %cmd_list = %$cmd_list_p;
  my $cmd;
  foreach my $cmd_version (sort (keys %cmd_list)) {
      if ($version < $cmd_version) {
          last;
      }
      $cmd = $cmd_list{$cmd_version};
  }
  if (!defined($cmd)) {
      croak("autodelete is not supported at version $version");
  }

  my $say_cmd = "snap autodelete $volume $cmd";
  my (undef, undef, $after) = say($filer, $say_cmd);
  if ($after !~ $output{$in_cmd}) {
      croak("snap autodelete return unexpected result: $after");
  }

  if ($in_cmd eq "show") {
      if ($after =~ /state\t+:\s+(\w+).*
                     commitment\t+:\s+(\w+).*
                     trigger\t+:\s+(\w+).*
                     target_free_space\t+:\s+(\d+)%.*
                     delete_order\t+:\s+(\w+).*
                     defer_delete\t+:\s+(\w+).*
                     prefix\t+:\s+(.*?)/xs){ 
          my %show_result;
          $show_result{state} = $1;
          $show_result{commitment} = $2;
          $show_result{trigger} = $3;
          $show_result{target_free_space} = $4;
          $show_result{delete_order} = $5;
          $show_result{defer_delete} = $6;
          my $pref = $7;
          $pref =~ s/(.*)\s?\(.*$/$1/;
          $show_result{prefix} = $pref;
          if ($version > WV_IRONCITY_MIN){
              if ($after =~ /destroy_list\t+:\s+(\w+)\s+\n/xs){
                    $show_result{destroy_list} = $1;
              }
              else{
                    croak("snap autodelete show return unexpected result  destroy_list: $after");
              }
          }
          return (%show_result);
      }
      else {
          croak("snap autodelete show return unexpected result: $after");
      }
  }

  return 1;

}

## @name ::snapshot_autodelete_options
## @usage snapshot_autodelete_options($filer, $volume, %opts)
## @synopsis
##   set snap autodelete options
##   example: snapshot_autodelete_options($fc, $vol,
##                                        trigger=>"volume",
##                                        defer_delete=>"user_create");
##   return 1 when succeed; otherwise croak.
##
sub snapshot_autodelete_options($$%) {
  my ($filer, $volume, %opts) = @_;
  my %options = ("state", ["on", "off"],
                 "commitment", ["try", "disrupt","destroy"],
                 "trigger", ["volume", "snap_reserve", "space_reserve"],
                 "target_free_space", [1..100],
                 "delete_order", ["oldest_first","newest_first"],
                 "defer_delete", ["scheduled","user_created", "prefix", "none"],
                 "prefix",[],
                 "destroy_list", ["lun_clone", "vol_clone", "sfsr", "cifs_share", "none"]);
  my $say_cmd = "snap autodelete $volume ";

  if (!%opts) {
      croak("no option specified for snap autodelete");
  }

  while (my ($opt_key, $opt_value) = each(%opts)) {
      if (!defined($opt_value)) {
          croak("no value passed to key $opt_key");
      }
      my $found_key = 0;
      foreach my $preset_key (keys %options) {
          if ($opt_key eq $preset_key) {
              $found_key = 1;
              if ($opt_key eq "prefix") {
                  if ($opt_value) {
                   $say_cmd .= $opt_key." " .$opt_value." ";
                  }
              }
              elsif ($opt_key eq "target_free_space") {
                  if ($opt_value <=100 && $opt_value >=1) {
                    $say_cmd .= $opt_key." ".$opt_value." ";
                  }
                  else {
                    croak("invalid value of option target_free_space: ".
                          $opt_value);
                  }
              }
              else {
                my $value_list_p = $options{$opt_key};
                my $found_value = 0;
                foreach my $value (@$value_list_p) {
                  if ($opt_value eq $value) {
			$found_value = 1;
                        $say_cmd .= $opt_key." ".$opt_value." ";
                        last;
                  }
                }
                if (!$found_value) {
                    croak("invalid value of option $opt_key: ".
                          $opt_value);
                }
              }

              last;
          }
      }
      if (!$found_key) {
          croak("invalid option: $opt_key");
      }
  }
  my (undef, undef, $after) = say($filer, $say_cmd);
  if ($after !~ /snap autodelete configuration options set/) {
      croak("snap autodelete fail: $after");
  }

  return 1;
}

## @name ::snapshot_flags
## @usage snapshot_flags($filer, $volume, $snapshot)
## @synopsis
##    For a given snapshot name, returns the list of all the active flags on it
##    e.g., for remi2(0007462702)_rum0_dreg0-dst.48 returns ("busy","snapmirror")
##    Does 'die' if the snapshot does not exist.
sub snapshot_flags($$$) {
  my ($filer, $volume, $snapshot) = @_;
  return eval {
    my @flags = ();
    my $text = sayEOF($filer, "snap list $volume", timeout => $::CMDTIMEOUT);

    # Return if list is empty
    die "No snapshots on volume $volume\n"
      if ($text =~ /$ERR_NO_SNAP/i);

    # Strip everything up to the beginning of the list
    # (Keep everything after the "----------  ----------  ------------  --------")
    die "Error stripping output header: $text\n"
      unless ($text =~ s/.*?\s*\-+\s+\-+\s+\-+\s+\-+\s*(.*)$/$1/s);

    my $snapexp = regx_escape($snapshot);

    if ($text =~ /$snapexp\s*(\(\S+\))?/s) {
      my $flagstr = ($1 ? $1 : undef);
      if ($flagstr && ($flagstr =~ /^\((\S+)\)$/)) {
        # Return the list of flags
        @flags = split(/,/,$1);
        return @flags;
      }
      else {
        # No flags, return empty list
        return @flags;
      }
    }
    else {
      # Return if snapshot isn't there
      die "No such snapshot \"$snapshot\"\n";
    }
  }
}

# zapi can fail under so many circumstances, so shut it up
# since we dont want FAIL results showing up from libs
sub _fail_default {
      my $msg = shift;
      logDbg("zapi failed: $msg ($@)");
}

## @name ::snapshot_multicreate
## @usage @snaps = snapshot_multicreate(args)
## @arg host - required. hostspec of filer
## @arg snap_name - required. name of snapshot to take
## @arg cleanup - optional. true|false. cleanup on failure
## @arg volumes - required. ref to array of volume names
## @returns list of (volume name, snap was created) pairs
## @synopsis
##    Create a "common" snapshot across multiple volumes (VSS)
sub snapshot_multicreate {
    my (%params) = @_;

    # prepare the ZAPI
    my $zcmd = new NaElement("snapshot-multicreate");
    $zcmd->child_add(new NaElement("snapshot","$params{snapname}"));
    $zcmd->child_add(new NaElement("cleanup",$params{cleanup}?$params{cleanup}:"true"));
    my $volumes = new NaElement("volume-names");
    my @vols = @{$params{volumes}};
    foreach my $vol (@vols) {
        $volumes->child_add(new NaElement("volume-name","$vol")) if $vol;
    }
    $zcmd->child_add($volumes);

    my $zapi = Zapi->new($params{host}->default_ip(),
			 "root",
			 $params{host}->default_password());
    $zapi->{fail} = \&_fail_default;

    my $xo = $zapi->invoke_elem($zcmd, "any");
    logDbg("snapshot-multicreate ZAPI returns: ". $xo->toString());
    my @snapsmade;
    if ($xo->name() eq "results") {
        push(@snapsmade, { 'status' => $xo->attr_get("status"),
                           'errno' => $xo->attr_get("errno"),
                           'reason' => $xo->attr_get("reason") });
    }
    if ( ! $xo->results_ok() ) {
        my $vo = $xo->child_get("volume-is-snapcreated-list");
        if ($vo) {
            my $num_elems = $vo->child_max_number();
            logDbg("child-get(volume-is-snapcreated-list) returned $num_elems children");
            for (my $i=0; $i < $num_elems; $i++) {
                my $vl = $vo->child_get_nth($i);
                my $vn = $vl->child_get("name")->{content};
                my $vs = $vl->child_get("is-snapcreated")->{content};
                push(@snapsmade, { 'name' => $vn,
                                   'is-snapcreated' => $vs });
            }
        }
    }
    return(@snapsmade);
}

## @name ::snapshot_multicreate_validate
## @usage @snaps = snapshot_multicreate_validate(args)
## @arg filer - required. hostspec of filer
## @arg snap_name - required. name of snapshot to take
## @arg volumes - required. ref to array of volume names
## @synopsis
##    Validate that can create a "common" snapshot across multiple
##    volumes (VSS) without actually creating snapshots
sub snapshot_multicreate_validate {
    my (%params) = @_;

    logTrc("volumes=>$params{volumes},snapname=>$params{snap_name}");
    # prepare the ZAPI
    my $zcmd = new NaElement("snapshot-multicreate-validate");
    $zcmd->child_add(new NaElement("snapshot","$params{snapname}"));
    my $volumes = new NaElement("volume-names");
    my @vols = @{$params{volumes}};
    foreach my $vol (@vols) {
        $volumes->child_add(new NaElement("volume-name","$vol")) if $vol;
    }
    $zcmd->child_add($volumes);

    my $zapi = Zapi->new($params{host}->default_ip(),
			 "root",
			 $params{host}->default_password());
    $zapi->{fail} = \&_fail_default;

    my $xo = $zapi->invoke_elem($zcmd, "any");
    logDbg("snapshot-multicreate-validate ZAPI returns: ". $xo->toString());
    my @snapresult;
    if ( ! $xo->results_ok() ) {
        if ($xo->name() eq "results") {
            push(@snapresult, { 'status' => $xo->attr_get("status"),
                                'errno' => $xo->attr_get("errno"),
                                'reason' => $xo->attr_get("reason") });
        }
    }
    my $vo = $xo->child_get("volume-errors");
    if ($vo) {
        my $num_elems = $vo->child_max_number();
        logDbg("child-get(volume-errors) returned $num_elems children");
        for (my $i=0; $i < $num_elems; $i++) {
            my $vl = $vo->child_get_nth($i);
            my $vn = $vl->child_get("name")->{content};
            my $ve = $vl->child_get("errno") ? $vl->child_get("errno")->{content} : undef;
            my $vr = $vl->child_get("reason") ? $vl->child_get("reason")->{content} : undef;
            push(@snapresult, { 'name' => $vn, 'errno' => $ve,
                                'reason' => $vr });
        }
    }
    return(@snapresult);
}

## @name ::snapshot_list_loaded
## @usage @snaps = snapshot_list_loaded(args)
## @arg filer - required. hostspec of filer
## @arg volume - required. name of volume
## @synopsis
##    uses the "snap test" to get and return the list of loaded snapshots
##    since this uses priv set -q test command, please make sure you are
##    a debug build
sub snapshot_list_loaded {
  my ($filer, $volume, %options) = @_;
  my $cmd = "priv set -q test;snap test show $volume snap_count";
  return eval {
    my $output  = sayEOF($filer, $cmd , timeout => $::CMDTIMEOUT);
    my @snap_ids = ($output =~ /snapid.*?(\d+)/g);
    return @snap_ids;
  }
}

## @name ::snapshot_latest_id
## @usage $snap_latest_id = snapshot_latest_id(args)
## @arg filer - required. hostspec of filer
## @arg volume - required. name of volume
## @synopsis
##    uses the "snap test" to get and return the latest loaded snapshot
##    since this uses priv set -q test command, please make sure you are
##    a debug build
sub snapshot_latest_id {
  my ($filer, $volume, %options) = @_;
  my $cmd = "priv set -q test;snap test show $volume latest_snapid";
  return eval {
    my $output  = sayEOF($filer, $cmd , timeout => $::CMDTIMEOUT);
    $output =~ /Latest snapid for.*?=.*?(\d+)/;
    my $latest_snapid  = $1;
    return $latest_snapid;
  }
}
1;
