#
#  $Id: //depot/prod/test/main/lib/SnapLock/Datamanip.pm#1 $
#
## @summary Functions to populate and modify worm data.
##
## @usage
##   use SnapLock::Datamanip;
##
## @status
##    Private
##
## @author kbala@netapp.com
##

package SnapLock::Datamanip;

use strict;
use warnings;
use Data::Dumper;
use NOTE;

use vars ('@ISA', '@EXPORT', '@EXPORT_OK');
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
             populate
             modify
            );

use constant APPEND_SIZE => '262144';
use constant MIN_RETENTION => '1200s';
use constant DEF_RETENTION => '2400s';
use constant MAX_RETENTION => '2400s';
my @retention = (MIN_RETENTION, DEF_RETENTION, MAX_RETENTION);
my (%data, $data, @files_modified);
my $file_count = 0;

use Carp qw(confess croak);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF);
use Tharn qw(say script_result host logcomment logresult param);
use Module::Common qw(validate_mixed);

BEGIN {
    Params::Validate::validation_options(
                                         on_fail => sub { my @msgs = @_; chomp(@msgs);
                                         Carp::confess( Tharn::script_result(@msgs) ) }
                                        );
}

#---------------------------------------------------------------------

=head2 populate 

populate(path => <vol|qtree>, worm => 10, app_worm => 10, non_worm => 10)

Populate volume or qtree with worm files, non worm files and appendable worm files

The following arguments are accepted:

=over

=item path  

(Required): Specify volume object or a qtree object  

=item worm

(Optional): Number of worm files to create. 

=item app_worm

(Optional): Number of appendable worm files to create.

=item non_worm 

(Optional): Number of regular files to create. 

=back

=cut

#---------------------------------------------------------------------
# populate 
#---------------------------------------------------------------------

sub populate {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                worm => { type=>SCALAR, optional=>1 },
                                app_worm => { type=>SCALAR, optional=>1 },
                                non_worm => { type=>SCALAR, optional=>1 },
                               }
                         );
    
    my $path = $opts{path};
    my $worm = $opts{worm};
    my $app_worm = $opts{app_worm};
    my $non_worm = $opts{non_worm};

    ## @step create a worm files
    my $i=0;
    while ($worm) {
        my $file = $path->nfs->create->fill(max => "1m");
        $i = 0 if ($i > 2);
        $file->commit(retention => $retention[$i], force => 1);
        $worm--;
        $i++;
    }

   ## @step Create a non worm files
    my $size=0;
    while ($non_worm) {
        $size = $non_worm % 2;
        if ($size eq 1) {
            my $file = $path->nfs->create->fill(max => "1m");
        } else {
            my $file = $path->nfs->create->fill(max => "1k");
        }
        $non_worm--;
    }

    ## @step Create a appendable worm files
    while ($app_worm) {
        my $file = $path->nfs->create;
        $file->commit(retention => MIN_RETENTION, force => 1);       
        $file->chmod(0666);

        NOTE::logcomment("Debug: File $file size before append : %d", $file->size);

        _append (    
                 file  => $file->fullpath, 
                 from_byte => 0,
                 to_byte => APPEND_SIZE,
                 byte => "N",
                ); 

        $app_worm--;
        NOTE::logcomment("Debug: File $file size after append : %d", $file->size);
    }
}

#---------------------------------------------------------------------
# _append
#---------------------------------------------------------------------

sub _append {

    my (%opts) = validate( @_, {
                                file => { type => SCALAR, },
                                from_byte  => { type => SCALAR },
                                to_byte => { type => SCALAR },
                                byte =>  { type => SCALAR, optional => 1, default => "N" },
                               }
                         );

    my $file = $opts{file};
    my $from_byte = $opts{from_byte};
    my $to_byte = $opts{to_byte};
    my $byte = $opts{byte};
    my ($data, $i) = (0, 0);

    open(FH, "+<$file") or die "Could not open file [$file]: $!"; 
    seek(FH, $from_byte, 0);

    while ($i < $to_byte) {
        print FH "$byte";
        $i++;
    }

    seek (FH, $from_byte, 0);
    read(FH, $data, $to_byte); 
    close(FH)

}

#---------------------------------------------------------------------

=head2 modify

modify(path => <vol|qtree>, worm => 1, app_worm => 1, non_worm => 1,
       commit_non_worm => 1, commit_app_worm => 1, expire_worm => 1, expire_app_worm => 1, extend_worm => 1, 
       extend_app_worm => 1, delete_non_worm => 1, delete_worm => 1, delete_app_worm => 1 )

Modify volume or qtree with worm files, non worm files and appendable worm files

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item worm

(Optional): Number of worm files to create. 

=item app_worm

(Optional): Number of appendable worm files to create.

=item non_worm

(Optional): Number of regular files to create.

=back

=cut

#---------------------------------------------------------------------
# modify
#---------------------------------------------------------------------

sub modify {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                worm => { type=>SCALAR, optional=>1, default=>0 },
                                app_worm => { type=>SCALAR, optional=>1, default=>0 },
                                non_worm => { type=>SCALAR, optional=>1, default=>0 },
                                commit_non_worm => { type=>SCALAR, optional=>1 },
                                commit_app_worm => { type=>SCALAR, optional=>1 },
                                expire_worm => { type=>SCALAR, optional=>1 },
                                expire_app_worm => { type=>SCALAR, optional=>1 },
                                extend_worm => { type=>SCALAR, optional=>1 },
                                extend_app_worm => { type=>SCALAR, optional=>1 },
                                delete_non_worm => { type=>SCALAR, optional=>1 },
                                delete_worm => { type=>SCALAR, optional=>1 },
                                delete_app_worm => { type=>SCALAR, optional=>1 },
                               }
                         );

    my $path = $opts{path};
    my $worm = $opts{worm};
    my $app_worm = $opts{app_worm};
    my $non_worm = $opts{non_worm};
    my $commit_non_worm = $opts{commit_non_worm};
    my $commit_app_worm = $opts{commit_app_worm};
    my $expire_worm = $opts{expire_worm};
    my $expire_app_worm = $opts{expire_app_worm};
    my $extend_worm = $opts{extend_worm};
    my $extend_app_worm = $opts{extend_app_worm};
    my $delete_non_worm = $opts{delete_non_worm};
    my $delete_worm = $opts{delete_worm};
    my $delete_app_worm = $opts{delete_app_worm};
    my ($data, @children);

    ## @step files_modified array contains mapping of files->status [status: 0 - not modified, 1 - modified ] 
    ## @step Initialize files_modified array to zero.
    @files_modified = ();
    @children = $path->nfs->children;
    foreach my $child (@children) {
        my $file = $child->name;
        next if ($file =~ /.snapshot/);
        $files_modified[$file_count][0] = $file;
         $files_modified[$file_count][1] = 0;
        $file_count++;
    }

    NOTE::logcomment("Debug: Data before modify");
    $data = _get_file_info (
                            path => $path,
                           );
    print Dumper($data);

    ## @step commit non_worm file to worm
    commit_non_worm ( 
                     path => $path,
                     commit_non_worm => $commit_non_worm,
                    ) if defined ($commit_non_worm);

    
    ## @step commit app_worm file to worm
    commit_app_worm (
                     path => $path,
                     commit_app_worm => $commit_app_worm,
                    ) if defined ($commit_app_worm);

    ## @step Extend retrention period for a worm file 
    extend_worm (
                 path => $path,
                 extend_worm => $extend_worm,
                ) if defined ($extend_worm);

    ## @step Extend retrention period for a appendable worm file
    extend_app_worm (
                     path => $path,
                     extend_app_worm => $extend_app_worm,
                    ) if defined ($extend_app_worm);

    ## @step Expire worm file
    expire_worm (
                 path => $path,
                 expire_worm => $expire_worm,
                ) if defined ($expire_worm);

    ## @step Expire appendable worm file
    expire_app_worm (
                 path => $path,
                 expire_app_worm => $expire_app_worm,
                ) if defined ($expire_app_worm);

    ## @step Delete non worm file
    delete_non_worm (
                     path => $path,
                     delete_non_worm => $delete_non_worm,
                    ) if defined ($delete_non_worm);

    ## @step Delete expired worm file
    delete_worm (
                 path => $path,
                 delete_worm => $delete_worm,
                ) if defined ($delete_worm);

    ## @step Delete expired appendable worm file
    delete_app_worm (
                     path => $path,
                     delete_app_worm => $delete_app_worm,
                    ) if defined ($delete_app_worm);

    populate (
              path => $path,
              worm => $worm,
              non_worm => $non_worm,
              app_worm => $app_worm,
             );

    NOTE::logcomment("Debug: Data after modify");
    $data = _get_file_info (
                            path => $path,
                           );
    print Dumper($data);
}


#---------------------------------------------------------------------
# _get_data
#---------------------------------------------------------------------

sub _get_data {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                type => { type=>SCALAR },
                                count => { type=>SCALAR },
                                task => => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $type = $opts{type};
    my $count = $opts{count};
    my $task = $opts{task};
    my (@file_obj, $file_obj);

    @file_obj = _get_file_info (
                                path => $path,
                                type => $type,
                               );

    $file_obj = $#file_obj;
    my $untouched = 0;
    foreach my $file_obj (@file_obj) {
        my $file = $file_obj->name;
        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] =~ $file);
        }
        $untouched++ if ($files_modified[$row][1] eq 0);
    }

    if ($count > $untouched) {
        NOTE::logcomment("Debug: Untouched $type files $untouched");
        NOTE::logcomment("Debug: Need to $task [$count] $type files");
        my $num = ($count - $untouched);
        NOTE::logcomment("Debug: Creating [$num] $type files");

        populate (
                  path => $path,
                  $type => $num,
                 );

        _add_to_array(path => $path);

        @file_obj = _get_file_info (
                                     path => $path,
                                     type => $type,
                                    );
    }
  
    return @file_obj;
}

#---------------------------------------------------------------------
# _add_to_array
#---------------------------------------------------------------------

sub _add_to_array {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                               }
                         );

    my $path = $opts{path};
    ## @step files_modified array contains mapping of files->status [status: 0 - not modified, 1 - modified ]
    ## @step Initialize files_modified array to zero for newly added non worm files.
    my $count = $file_count;
    my @children = $path->nfs->children;
    my $flag = 0;

    foreach my $child (@children) {
        my $file = $child->name;
        next if ($file =~ /.snapshot/);

        foreach my $i (0..$count) {
            if ( $file eq $files_modified[$i][0] ) {
                $flag = 1;
                last;
            } else {
                $flag = 0;
            }
        }

        if ($flag eq 0) {
            $files_modified[$file_count][0] = $file;
            $files_modified[$file_count][1] = 0;
            $file_count++;
        }  
    }
}

#---------------------------------------------------------------------

=head2 commit_non_worm

commit_non_worm ( 
                 path => <vol|qtree>, 
                 commit_non_worm => 1, 
                )

Commit a non worm file to worm

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item commit_non_worm

(Optional): Number of regular files to commit to worm.

=back

=cut

#---------------------------------------------------------------------
# commit_non_worm
#---------------------------------------------------------------------

sub commit_non_worm {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                commit_non_worm => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $commit_non_worm = $opts{commit_non_worm};
    my (@file_obj, $file_obj);

    @file_obj = _get_data (
                            path => $path,
                            type => 'non_worm',
                            task => 'commit',
                            count => $commit_non_worm,
                          );

    $file_obj = $#file_obj;
    my $i = 0;
    while ($commit_non_worm) {
        $i = 0 if ($i > 2);
        my $file = $file_obj[$file_obj]->name;

        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] =~ /$file/);
        }
 
        if ($files_modified[$row][1] == 0) {
            NOTE::logcomment("Debug: committing worm file [$file]");
            $file_obj[$file_obj]->commit(retention => $retention[$i], force => 1);   
            $files_modified[$row][1] = 1;
            $commit_non_worm--;
        }
        $file_obj--;
        $i++;
    }
    NOTE::logcomment("Debug: commit non_worm is successful");
}

#---------------------------------------------------------------------

=head2 commit_app_worm

commit_app_worm (
                 path => <vol|qtree>,
                 commit_app_worm => 1,
                )

Commit an appendable worm file to worm

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item commit_app_worm

(Optional): Number of appendable worm files to worm. 

=back

=cut

#---------------------------------------------------------------------
# commit_app_worm
#---------------------------------------------------------------------

sub commit_app_worm {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                commit_app_worm => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $commit_app_worm = $opts{commit_app_worm};
    my (@file_obj, $file_obj);

    @file_obj = _get_data (
                            path => $path,
                            type => 'app_worm',
                            task => 'commit',
                            count => $commit_app_worm,
                          );

    $file_obj = $#file_obj;
    while ($commit_app_worm) {
        my $file = $file_obj[$file_obj]->name;

        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] =~ $file);
        }

        if ($files_modified[$row][1] == 0) { 
            NOTE::logcomment("Debug: Adding data to appendable worm file [$file] before commit");
            my $size = $file_obj[$file_obj]->size;
          
            _append (
                     file  => $file_obj[$file_obj]->fullpath,
                     from_byte => $size,
                     to_byte => APPEND_SIZE,
                     byte => "N",
                    );
  
            my $file = $file_obj[$file_obj]->name;
            NOTE::logcomment("Debug: committing appendable worm file [$file]");
            $file_obj[$file_obj]->chmod(040); 
            $files_modified[$row][1] = 1;
            $commit_app_worm--;
        }
        $file_obj--;
    }
    NOTE::logcomment("Debug: commit app_worm is successful");
}

#---------------------------------------------------------------------

=head2 extend_worm

extend_worm (
             path => <vol|qtree>,
             extend_worm => 1,
            )

Extend the retention period of worm file

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item extend_worm

(Optional): Number of worm files whose retention needs to be extended.

=back

=cut

#---------------------------------------------------------------------
# extend_worm
#---------------------------------------------------------------------

sub extend_worm {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                extend_worm => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $extend_worm = $opts{extend_worm};
    my (@file_obj, $file_obj);

    @file_obj = _get_data (
                            path => $path,
                            type => 'worm',
                            task => 'extend',
                            count => $extend_worm,
                          );

    $file_obj = $#file_obj;
    while ($extend_worm) {
        my $file = $file_obj[$file_obj]->name;

        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] =~ $file);
        }

        if ($files_modified[$row][1] == 0) {
             NOTE::logcomment("Extending retention period for worm file [$file]");
            $file_obj[$file_obj]->retention(MAX_RETENTION);
            $files_modified[$row][1] = 1;
            $extend_worm--;
        }
        $file_obj--;
   }
    NOTE::logcomment("Debug: Extend retention period for worm is successful");
}

#---------------------------------------------------------------------

=head2 extend_app_worm

extend_app_worm (
                 path => <vol|qtree>,
                 extend_app_worm => 1,
                )

Extend the retention period of appendable worm file

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item extend_app_worm

(Optional): Number of worm appendable files whose retention needs to be extended.

=back

=cut


#---------------------------------------------------------------------
# extend_app_worm
#---------------------------------------------------------------------

sub extend_app_worm {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                extend_app_worm => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $extend_app_worm = $opts{extend_app_worm};
    my (@file_obj, $file_obj);

    @file_obj = _get_data (
                            path => $path,
                            type => 'app_worm',
                            task => 'extend',
                            count => $extend_app_worm,
                          );

    $file_obj = $#file_obj;
    while ($extend_app_worm) {
        my $file = $file_obj[$file_obj]->name;

        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] =~ $file);
        }

        if ($files_modified[$row][1] == 0) {
            NOTE::logcomment("Debug: Adding data to appendable worm file [$file] before commit");
            my $size = $file_obj[$file_obj]->size;

            _append (
                     file  => $file_obj[$file_obj]->fullpath,
                     from_byte => $size,
                     to_byte => APPEND_SIZE,
                     byte => "N",
                    );

            NOTE::logcomment("Extending retention period for app_worm file [$file]");
            $file_obj[$file_obj]->retention(MAX_RETENTION);
            $files_modified[$row][1] = 1;
            $extend_app_worm--;
        }
        $file_obj--;
   }
    NOTE::logcomment("Debug: Extend retention period for app_worm file is successful");
}

#---------------------------------------------------------------------

=head2 expire_worm

expire_worm (
             path => <vol|qtree>,
             expire_worm => 1,
            )

Make sure worm file is expired

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item expire_worm

(Optional): Number of worm files which gets expired. 

=back

=cut

#---------------------------------------------------------------------
# expire_worm
#---------------------------------------------------------------------

sub expire_worm {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                expire_worm => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $expire_worm = $opts{expire_worm};
    my (@file_obj, $file_obj, @expire_file_obj);

    @file_obj = _get_data (
                            path => $path,
                            type => 'worm',
                            task => 'expire',
                            count => $expire_worm,
                          );

    $file_obj = $#file_obj;
    while ($expire_worm) {
        my $file = $file_obj[$file_obj]->name;

        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] =~ $file);
        }

        if ($files_modified[$row][1] == 0) {
            NOTE::logcomment("Debug: Expire worm file [$file]");
            push (@expire_file_obj, $file_obj[$file_obj]);
            $file_obj[$file_obj]->expire;
            $files_modified[$row][1] = 1;
            $expire_worm--;
        }
        $file_obj--;
   }
    NOTE::logcomment("Debug: Expire worm file is successful");
    return (\@expire_file_obj);
}

#---------------------------------------------------------------------

=head2 expire_app_worm

expire_app_worm (
             path => <vol|qtree>,
             expire_app_worm => 1,
            )

Make sure appendable worm file is expired

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item expire_app_worm

(Optional): Number of appendable worm files which gets expired.

=back

=cut

#---------------------------------------------------------------------
# expire_app_worm
#---------------------------------------------------------------------

sub expire_app_worm {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                expire_app_worm => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $expire_app_worm = $opts{expire_app_worm};
    my (@file_obj, $file_obj, @expire_file_obj);

    @file_obj = _get_data (
                            path => $path,
                            type => 'app_worm',
                            task => 'expire',
                            count => $expire_app_worm,
                          );

    $file_obj = $#file_obj;
    while ($expire_app_worm) {
        my $file = $file_obj[$file_obj]->name;

        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] =~ $file);
        }

        if ($files_modified[$row][1] == 0) {
            NOTE::logcomment("Debug: Adding data to appendable worm file [$file] before commit");
            my $size = $file_obj[$file_obj]->size;

            _append (
                     file  => $file_obj[$file_obj]->fullpath,
                     from_byte => $size,
                     to_byte => APPEND_SIZE,
                     byte => "N",
                    );

            NOTE::logcomment("Debug: Expire appenable worm file [$file]");
            push (@expire_file_obj, $file_obj[$file_obj]);
            $file_obj[$file_obj]->expire;
            $files_modified[$row][1] = 1;
            $expire_app_worm--;
        }
        $file_obj--;
    }
    NOTE::logcomment("Debug: Expire appendable worm file is successful");
    return (\@expire_file_obj);
}

#---------------------------------------------------------------------

=head2 delete_non_worm

delete_worm (
             path => <vol|qtree>,
             delete_non_worm => 1,
            )

Delete expired non-worm file

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item delete_worm

(Optional): Number of expired non-worm files which gets deleted. Default 1.

=back

=cut

#---------------------------------------------------------------------
# delete_non_worm
#---------------------------------------------------------------------

sub delete_non_worm {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                delete_non_worm => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $delete_non_worm = $opts{delete_non_worm};
    my (@file_obj, $file_obj);

    @file_obj = _get_data (
                            path => $path,
                            type => 'non_worm',
                            task => 'delete',
                            count => $delete_non_worm,
                          );

   
    $file_obj = $#file_obj; 
    while ($delete_non_worm) {
        my $file = $file_obj[$file_obj]->name;
        NOTE::logcomment("Debug: Deleting non-worm [$file]");

        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] =~ $file);
        }

        if ($files_modified[$row][1] == 0) {
            $file_obj[$file_obj]->rm;
            $delete_non_worm--;
            delete $files_modified[$row][0];
            delete $files_modified[$row][1];
        }
        $file_obj--;
    }
    NOTE::logcomment("Debug: Delete non-worm is successful");
}

#---------------------------------------------------------------------

=head2 delete_worm

delete_worm (
             path => <vol|qtree>,
             delete_worm => 1,
            )

Delete expired worm file 

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item delete_worm

(Optional): Number of expired worm files which gets deleted. 

=back

=cut

#---------------------------------------------------------------------
# delete_worm
#---------------------------------------------------------------------

sub delete_worm {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                delete_worm => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $delete_worm = $opts{delete_worm};
  
    my $file_obj = expire_worm (
                                 path => $path,
                                 expire_worm => $delete_worm,
                               );

    my $count = scalar(@$file_obj);
    foreach my $i (0..$count-1) {
        my $file = $file_obj->[$i]->name;
        NOTE::logcomment("Debug: Deleting worm file [$file]");
        $file_obj->[$i]->rm;
 
        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] eq $file);
        }
        
        delete $files_modified[$row][0];
        delete $files_modified[$row][1];
    }   
    NOTE::logcomment("Debug: delete worm file is successful"); 
} 

#---------------------------------------------------------------------

=head2 delete_app_worm

delete_app_worm (
                 path => <vol|qtree>,
                 delete_app_worm => 1,
                )

Delete expired appenable worm file

The following arguments are accepted:

=over

=item path

(Required): Specify volume object or a qtree object

=item delete_worm

(Optional): Number of expired appendable worm files which gets deleted.

=back

=cut

#---------------------------------------------------------------------
# delete_app_worm
#---------------------------------------------------------------------

sub delete_app_worm {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                delete_app_worm => { type=>SCALAR },
                               }
                         );

    my $path = $opts{path};
    my $delete_app_worm = $opts{delete_app_worm};

    my $file_obj = expire_app_worm (
                                    path => $path,
                                    expire_app_worm => $delete_app_worm,
                                  );

    my $count = scalar(@$file_obj);
    foreach my $i (0..$count-1) {
        my $file = $file_obj->[$i]->name;
        NOTE::logcomment("Debug: Deleting appendable worm file [$file]");
        $file_obj->[$i]->rm;

        my $row = 0;
        foreach my $j (0..$#files_modified) {
            $row = $j;
            last if ($files_modified[$j][0] eq $file);
        }

        delete $files_modified[$row][0];
        delete $files_modified[$row][1];
    }
    NOTE::logcomment("Debug: delete appendable worm file is successful");
}

#---------------------------------------------------------------------
# _get_file_info 
#---------------------------------------------------------------------

sub _get_file_info {
    my (%opts) = validate( @_, {
                                path => { type=>OBJECT },
                                type => { type=>SCALAR, optional=>1 },
                               }
                         );

    my $path = $opts{path};
    my $type = $opts{type};
    my (%data, $data, $file, @children);
    my (@worm_obj, @app_worm_obj, @non_worm_obj) = ();

    @children = $path->nfs->children;
    foreach my $child (@children) {
        $file = $child->name;
        next if ($file =~ /.snapshot/);

        my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = $child->stat;
        my $perms = sprintf("%04o", $mode & 07777);

        my $worm = $child->is_worm;

        if ( ($child->is_worm) && ($perms =~ /0040/) ) {
            $file = $child->name;
            $data{worm}{$file}{ino} = $ino;
            $data{worm}{$file}{mode} = $perms;
            $data{worm}{$file}{size} = $size;
            $data{worm}{$file}{atime} = $atime;
            $data{worm}{$file}{mtime} = $mtime;
            $data{worm}{$file}{ctime} = $ctime; 
            push (@worm_obj, $child);
        } elsif ( ($child->is_worm) && ($perms =~ /0666/) ) {
            $file = $child->name;
            $data{app_worm}{$file}{ino} = $ino;
            $data{app_worm}{$file}{mode} = $perms;
            $data{app_worm}{$file}{size} = $size;
            $data{app_worm}{$file}{atime} = $atime;
            $data{app_worm}{$file}{mtime} = $mtime;
            $data{app_worm}{$file}{ctime} = $ctime;
            push (@app_worm_obj, $child);
        } else {
            $file = $child->name;
            $data{non_worm}{"$file"}{ino} = $ino;
            $data{non_worm}{"$file"}{mode} = $perms;
            $data{non_worm}{"$file"}{size} = $size;
            $data{non_worm}{"$file"}{atime} = $atime;
            $data{non_worm}{"$file"}{mtime} = $mtime;
            $data{non_worm}{"$file"}{ctime} = $ctime;
            push (@non_worm_obj, $child);
        }
    }
     
    if ($type eq 'non_worm') {
        return @non_worm_obj;
    } elsif ($type eq 'app_worm') {
        return @app_worm_obj;
    } elsif ($type eq 'worm') {
        return @worm_obj;
    }
    return (\%data);
}
1;
