#
#  $Id: //depot/prod/test/main/lib/SnapLock/Snaplock_datamanip.pm#1 $
#
## @summary Functions to populate and modify worm data.
##
## @usage
##   use Snaplock_datamanip;
##
## @status 
##    Private 
##
## @author kasasrik@netapp.com, kbala@netapp.com
##

=head1 NAME

SnapLock::Snaplock_datamanip

=head1 DESCRIPTION

These are all the functions which help to populate and modify worm data.

=head1 METHODS

=cut

package SnapLock::Snaplock_datamanip;

use strict;
use base qw(Exporter);
use Cwd qw(getcwd);
use Carp qw(croak);
use Tharn;
use filer;
use boot;
use Filer::Version qw(codename_to_version);
use SPARSE::Util::Logger qw(logStep logError logWarn logDbg);
use NOTE;
use SnapLock::Helper;
use Dense::WackIron;
use wack;
use constant FILE_COUNT => 1;
use constant MAX_FILE_COUNT => 3;
use constant APPEND_SIZE => '262144';
use constant OVERWRITE_SIZE=> '200';
use snaplock qw(compliance_clock_status);
use Date::Manip;
use Time::ParseDate;
use Time::Local;

BEGIN {
   our @EXPORT = qw(
		    get_worm_data
                    populate_worm_data
		    populate_regular_data
                    modify_worm_data
		    wait_for_dump_image
		    restore_dump_image
		    verify_dump_data
		    flip_setup
		    wack_and_iron
 		    create_file_with_name
                    populate_worm_files
		    extend_retention_period
                    dump_vol_image
                    restore_vol_dump_image
                    set_filer_cc
                    get_filer_cc
                    get_qt_cc_sec
		   );
}
                     
=head2 ::get_worm_data

Get all types of worm data on the given qtree
Worm data types includes 
	- worm files
	- expired worm files
	- worm append files
	- expired worm append files

The format is as follows
         get_worm_data(qtree, populated data );

=item Arguments

B<qtree> => I<Qtree object> | 'NOTE:Qtree'

I<Required.>
A Qtree object.
 
B<data> => I<populated_data>

I<Required.>
Populated data by populate_data function.

=over 

=cut

sub get_worm_data($$) {
    my ($qt_tmp, $data) = @_;

    my @files = $qt_tmp->nfs->children;
    my %hash;

    foreach my $file (@files) {
        if($file->name !~ /snapshot/) {
            foreach my $type (keys %$data){
                my $types = $data->{$type};
                foreach my $file_name (keys %$types){
                    my $fobj = $types->{$file_name}->{obj};
                    if($file->name eq $file_name){
                        $hash{$type}->{$file->name}->{obj} = $file;
                        $hash{$type}->{$file->name}->{name} = $file->name;
                        $hash{$type}->{$file->name}->{atime} = $file->atime;
                        $hash{$type}->{$file->name}->{mtime} = $file->mtime;
                        $hash{$type}->{$file->name}->{ctime} = $file->ctime;
                    }
                }
            }
        }
    }
    return \%hash;
}

=head2 ::populate_worm_data

Populate all types of worm data on the given qtree
Worm data types includes
        - worm files
        - expired worm files
        - worm append files
        - expired worm append files

The format is as follows
         populate_worm_data(qtree);

=item Arguments

B<qtree> => I<Qtree object> | 'NOTE:Qtree'

I<Required.>
A Qtree object.

=item Examples

=back

     $populate_worm_data($qtree);

=head2 ::status

Returns a hash reference which contains populated worm data

=over

=cut

sub populate_worm_data($$) {
    my ($src,$retention) = @_;
    my (%hash, %args );

    ## Create Regular Files and save the data into hash->{regular}
    $hash{regular} = create_setup_with_regular_files($src, \%args);

    ## Create WORM Files and save the data into hash->{worm}
    $hash{worm} = create_setup_with_worm_files($src, $retention);

    ## Create WORM_EXTEND Files and save the data into hash->{worm_extend}
    $hash{worm_extend} = create_setup_with_worm_files($src,$retention);

    ## Create EXPIRED_WORM Files and save the data into hash->{worm_expired}
    $hash{worm_expired} = create_setup_with_expired_worm_files($src);

    ## Create REGULAR_FILES with SIZE as 0 bytes and save the data into hash->{regular_zero}
    %args = ( size => 0 );
    $hash{regular_zero} = create_setup_with_regular_files($src, \%args);

    my $IS_WORM_APPEND_REQ = check_filer_version($src);
    if ( !$IS_WORM_APPEND_REQ ) {
        ## Create WORM APPEND Files and save the data into hash->{worm_append}
        $hash{worm_append} = create_setup_with_worm_append_files($src, $retention);

        ## Create WORM APPEND Files and save the data into hash->{worm_append_extend}
        $hash{worm_append_extend} = create_setup_with_worm_append_files($src, $retention);
    }

    return \%hash;
}

=head2 :: populate_regular_data

Populate Regular file on the given qtree/volume.

The format is as follows 
	populate_regular_data(qtree)

=item Arguments

B<qtree> => I<Qtree object> | 'NOTE:Qtree'

I<Required.> 
A Qtree object.

=item Examples

=back

	$populate_worm_data($qtree);

=head2 ::status

Returns a hash reference which contains populated worm data

=over 

=cut 

sub populate_regular_data($) {
     my ($src) = @_;
     my (%hash, %args);

     ## Create Regular Files and save the data into hash->{regular}
     $hash{regular} = create_setup_with_regular_files($src, \%args);

      return \%hash;
 }
	
# This is an internal utility function which converts hash to an array
sub _hash_to_array($) {
    my($hash) = @_;
 
    my @files;
    foreach my $file (keys %$hash) {
        NOTE::logcomment("File Name Inside the Hash_to_array : $file : %s ",$hash->{$file}->{obj}->name );
        push(@files, $hash->{$file}->{obj});
    }
    return \@files;
}

=head2 ::modify_worm_data

modify all types of worm data on the given qtree
Worm data types includes
        - worm files
        - expired worm files
        - worm append files
        - expired worm append files

The format is as follows
         modify_worm_data(qtree);

=item Arguments

B<qtree> => I<Qtree object> | 'NOTE:Qtree'

I<Required.>
A Qtree object.

=item Examples

=back

     $modify_worm_data($qtree);

=head2 ::status

Returns a hash reference which contains modified worm data

=over

=cut

sub modify_worm_data($$$) {
    my ($src, $data, $retention) = @_;
    my (@files, %hash, $files, $codename);


    ## Modify Non-WORM file to WORM
    if( exists $data->{regular} ) { 
        $files = _hash_to_array( $data->{regular} );
        commit_to_worm($files, $retention->{MIN_RETENTION});
        $hash{worm} = _save_file_data_to_hash($files);
    
	    foreach my $key (sort keys %hash) {
 	        foreach my $file (@$files) {
		        $data->{$key}->{$file->name}->{obj} = $file;
		        $data->{$key}->{$file->name}->{name} = $file->name;			
		        $data->{$key}->{$file->name}->{atime} = $file->atime;
		        $data->{$key}->{$file->name}->{mtime} = $file->mtime;
		        $data->{$key}->{$file->name}->{ctime} = $file->ctime;
		        NOTE::logcomment("In Regular to WORM loop for file %s and type %s.", $file->name, $key);
		        delete ( $data->{regular}->{$file->name} );
		        last;
	        }
	    }
    } 

    ## Modify WORM file to WORM_EXTEND
    if (exists $data->{worm} ) {
        my $files = _hash_to_array( $data->{worm} );
        $hash{worm_extend} = extend_retention_period($data->{worm});
	
        foreach my $key (sort keys %hash) {
            foreach my $file (@$files) {
                $data->{$key}->{$file->name}->{obj} = $file;
                $data->{$key}->{$file->name}->{name} = $file->name;
                $data->{$key}->{$file->name}->{atime} = $file->atime;
                $data->{$key}->{$file->name}->{mtime} = $file->mtime;
                $data->{$key}->{$file->name}->{ctime} = $file->ctime;
		NOTE::logcomment("In WORM to WORM Extend loop for file %s and type %s.", $file->name, $key);
                delete ( $data->{worm}->{$file->name} );
		        last;
            }
        }
    }

    ## Modify WORM_EXTEND file to WORM_EXPIRED  
    if (exists $data->{worm_extend} ) {
        $files = create_expired_worm_files($src, FILE_COUNT);
        $hash{worm_expired} = _save_file_data_to_hash($files);
	
        foreach my $key (sort keys %hash) {
            foreach my $file (@$files) {
                $data->{$key}->{$file->name}->{obj} = $file;
                $data->{$key}->{$file->name}->{name} = $file->name;
                $data->{$key}->{$file->name}->{atime} = $file->atime;
                $data->{$key}->{$file->name}->{mtime} = $file->mtime;
                $data->{$key}->{$file->name}->{ctime} = $file->ctime;
		NOTE::logcomment("In WORM Extend to Expired File loop for file %s and type %s.", $file->name, $key);
                delete ( $data->{worm_extend}->{$file->name} );
                last;
            }
        }
    }

    
    ## Modify Non-WORM files to WORM_APPEND
    if (exists $data->{regular_zero} ) {
        $files = _hash_to_array($data->{regular_zero});
        commit_to_worm_append($files, $retention->{MIN_RETENTION});
        $hash{worm_append} = _save_file_data_to_hash($files);
        
        foreach my $key (sort keys %hash) {
            foreach my $file (@$files) {
                $data->{$key}->{$file->name}->{obj} = $file;
                $data->{$key}->{$file->name}->{name} = $file->name;
                $data->{$key}->{$file->name}->{atime} = $file->atime;
                $data->{$key}->{$file->name}->{mtime} = $file->mtime;
                $data->{$key}->{$file->name}->{ctime} = $file->ctime;
		NOTE::logcomment("In Non-WORM to WORM_APPEND loop for file %s and type %s.", $file->name, $key);
                delete ( $data->{regular_zero}->{$file->name} );
                last;
            }
        }
    }

    my $IS_WORM_APPEND_REQ = check_filer_version($src);
    if ( !$IS_WORM_APPEND_REQ ) {

   	    ## Modify WORM APPEND 
   	    if (exists $data->{worm_append} ) {
    	        $files = _hash_to_array($data->{worm_append});
                $hash{worm_append} = overwrite_worm_append_files($hash{worm_append});

                foreach my $key (sort keys %hash) {
	            foreach my $file (@$files) {
        	        $data->{$key}->{$file->name}->{obj} = $file;
                        $data->{$key}->{$file->name}->{name} = $file->name;
               	        $data->{$key}->{$file->name}->{atime} = $file->atime;
                        $data->{$key}->{$file->name}->{mtime} = $file->mtime;
                        $data->{$key}->{$file->name}->{ctime} = $file->ctime;
		        NOTE::logcomment("In modify WORM_Append loop for file %s and type %s.", $file->name, $key);
                        last;
                    }
                }
           }

         ## Extend the Retention Period for WORM Append Files
         if (exists $data->{worm_append} ) {
             $files = _hash_to_array($data->{worm_append});
             $hash{worm_append_extend} = extend_retention_period($data->{worm_append});

              foreach my $key (sort keys %hash) {
                  foreach my $file (@$files) {
               	     $data->{$key}->{$file->name}->{obj} = $file;
               	     $data->{$key}->{$file->name}->{name} = $file->name;
               	     $data->{$key}->{$file->name}->{atime} = $file->atime;
               	     $data->{$key}->{$file->name}->{mtime} = $file->mtime;
               	     $data->{$key}->{$file->name}->{ctime} = $file->ctime;
		     NOTE::logcomment("In extend WORM_APPEND loop for file %s and type %s.", $file->name, $key);
               	     delete ( $data->{worm_append}->{$file->name} );
               	     last;
                  }
       	      }
         }

   	    ## Make WORM APPEND Files to WORM
   	    if (exists $data->{worm_append} ) {
       	    $files = _hash_to_array($data->{worm_append});
       	    commit_worm_append_to_worm($files);
       	    $hash{worm} = _save_file_data_to_hash($files);

       	    foreach my $key (sort keys %hash) {
                foreach my $file (@$files) {
         	    $data->{$key}->{$file->name}->{obj} = $file;
               	    $data->{$key}->{$file->name}->{name} = $file->name;
               	    $data->{$key}->{$file->name}->{atime} = $file->atime;
               	    $data->{$key}->{$file->name}->{mtime} = $file->mtime;
               	    $data->{$key}->{$file->name}->{ctime} = $file->ctime;
		    NOTE::logcomment("In WORM_APPEND to WORM loop for file %s and type %s.", $file->name, $key);
               	    delete ( $data->{worm_append}->{$file->name} );
               	    last;
           	}
       	    }
        }
    }
    return \%hash;
}

sub populate_worm_files($$) {

        my ($src, $args) = @_;
        my %return_hash;
        my %Inargs;

        if ( $args->{worm_expired}  ) {

                my $expired_worm_files = create_expired_worm_files($src, FILE_COUNT);

                ## Save all the file level data into Hash
                my $hash = _save_file_data_to_hash($expired_worm_files);
                $return_hash{worm_expired} = $hash;
        }

        if ( $args->{regular} ) {
                %Inargs = ( count => $args->{count} );
                $return_hash{regular} = create_setup_with_regular_files($src , \%Inargs );
        }

        if ( $args->{worm} ) {
            my $worm_files ;
            if ( defined $args->{size} ) {
                $worm_files = create_worm_files($src, $args->{count}, $args->{retention}, $args->{size} );
            }else {    
                $worm_files = create_worm_files($src, $args->{count}, $args->{retention} );
            }
            my @files;
            push(@files, @$worm_files);

                ## Save all the file level data inti Hash
             my  $hash = _save_file_data_to_hash(\@files);

             $return_hash{worm} = $hash;
        }

        if ( $args->{worm_append} ){

                my $worm_append_files = create_worm_append_files($src, $args->{count}, $args->{retention} );
                my @files;

                push(@files, @$worm_append_files);
                my  $hash = _save_file_data_to_hash(\@files);
                $return_hash{worm_append} = $hash;
        }

        return \%return_hash;
}

sub create_file_with_name($$) {

        my ($root, $path ) = @_;

        my $file = $root->nfs->create("abcd")->commit(retention =>"300", force=>1);;

        return $file;
}


## @name ::commit_worm_append_to_worm
##
## @description
##  Commits worm append files to worm file by changing the permission on the file
##
## @synopsis
##  commit_worm_append_to_worm($$) 

sub commit_worm_append_to_worm($$) {
    my ($files) = @_;
    foreach  my $file (@$files) { 
        $file->chmod(044);
	last;
    }
}

## @name ::extend_retention_period
##
## @description
##  Extends the retention period for the given files by 2 hour.
##
## @synopsis
##   extend_retention_period($) 

sub extend_retention_period($) {
    my ($data) = @_;
    my ($file, @files);
    foreach  my $file_name (keys %$data) { 
        $file = $data->{$file_name}->{obj};
	my $retention = $file->atime - $file->ctime;
        logcomment("Old retention = $retention\n");
	$retention = $retention * 2; 
	$retention = $retention."s";
        logcomment("new retention = $retention\n");
        my $current = $file->atime;
        logcomment("Current atime = $current\n");
	$file->retention($retention);
        push(@files, $file);
	last;
    }
    my $hash = _save_file_data_to_hash(\@files);
    return $hash;
}

## @name ::create_setup_with_worm_files
##
## @description
##   Create worm files with min/def/max retention period. 
##   Number of worm files created  depends on FILE_COUNT
##   
## @synopsis
##   create_setup_with_worm_files($)

sub create_setup_with_worm_files($$) {
    my ($src,$retention) = @_;
    my @files;

    ## Create WORM files with different retention dates
    my $worm_files = create_worm_files($src, FILE_COUNT, $retention->{MIN_RETENTION});
    push(@files,@$worm_files);

    $worm_files = create_worm_files($src, FILE_COUNT, $retention->{DEF_RETENTION});
    push(@files,@$worm_files);

    $worm_files = create_worm_files($src, FILE_COUNT, $retention->{MAX_RETENTION});
    push(@files,@$worm_files);

    ## Save all the file level data into Hash
    my $hash = _save_file_data_to_hash(\@files);

    return $hash;
}

## @name ::create_setup_with_worm_append_files
##
## @description
##   Create worm append files with min/def/max retention period.
##   Number of worm files created  depends on FILE_COUNT
##
## @synopsis
##   create_setup_with_worm_append_files($)

sub create_setup_with_worm_append_files($$) {
    my ($src,$retention) = @_;
    my @files;

    ## Create WORM files with different retention dates
    my $worm_files = create_worm_append_files($src, FILE_COUNT, $retention->{MIN_RETENTION});
    push(@files, @$worm_files);

    $worm_files = create_worm_append_files($src, FILE_COUNT, $retention->{DEF_RETENTION});
    push(@files, @$worm_files);

    $worm_files = create_worm_append_files($src, FILE_COUNT, $retention->{MAX_RETENTION});
    push(@files, @$worm_files);

    ## Save all the file level data into Hash
    my $hash = _save_file_data_to_hash(\@files);

    return $hash;
}

## @name ::create_setup_with_regular_files
##
## @description
##   Create regular files of size 1m.
##   Number of worm files created  depends on FILE_COUNT
##   Regular file of size 0 is created if $args{size}=0. 
##    
## @synopsis
##   create_setup_with_regular_files($$) 

sub create_setup_with_regular_files($$) {
    my ($src, $args) = @_;
    my (@files, $file, $count, $size);

    ## Create WORM files with different retention dates
    if( defined $args->{count} ) { 
        $count = $args->{count} }
    else { 
        $count = FILE_COUNT; 
    }

    while($count) {
        if( defined $args->{size} ) { 
            if($args->{size} == 0) {
                $file = create_file($src);   
            } else {
                $file = create_and_fill_file($src,$args->{size});
            }
        } else {
            $size = '1m';
            $file = create_and_fill_file($src,$size);
        }
        push(@files,$file);
        $count--;
    }

    ## Save all the file level data into Hash
    my $hash = _save_file_data_to_hash(\@files);
    return $hash;
}

## @name ::create_setup_with_expired_worm_files
##
## @description
##   Create expired worm files.
##   Number of worm files created  depends on FILE_COUNT
##
## @synopsis
##   create_setup_with_expired_worm_files($)

sub create_setup_with_expired_worm_files($) {
    my ($src) = @_;
    my @files;

    ## Create WORM files with different retention dates
    my $expired_worm_files = create_expired_worm_files($src, FILE_COUNT);

    ## Save all the file level data into Hash
    my $hash = _save_file_data_to_hash($expired_worm_files);
    return $hash;
}

## @name ::_append
##
## @description
##   This is an internal function which append data to the worm file
##
## @synopsis
##  append($$) 

sub _append($$) {
    my ($file, $file_to_append) = @_;

    my $i = 0;
    open(OUT, ">>$file") or $@ = $!, return 0;
    open(IN, "$file_to_append") or $@ = $!, return 0;
    while (<IN>) {
       print OUT $_;
    }
    close(OUT);
    close(IN);
}

## @name ::_overwrite
##
## @description
##  This is an internal function which overwrite n bytes of a worm file with 
##  a given string
##
## @synopsis
##  _overwrite($$$$)

sub _overwrite($$$$) {
    my ($filename, $frombyte, $tobyte, $byte) = @_;
    my ($i, $data, $count);

    $byte = "a" if(!defined $byte);
    die "Usage: <filename> <frombyte> <tobyte> [string]" if(!defined $tobyte or !defined $frombyte);
    open(fh, "+<$filename") or die "Could not open file [$filename]: $!";
    seek(fh, $frombyte, 0);
    $i = 0;
    while ($i < $tobyte) {
        print fh "$byte";
        $i++;
    }
        
    seek(fh, $frombyte, 0);
    $count = $frombyte - $tobyte;
    read(fh, $data, $tobyte); 
    close(fh);
}

## @name ::overwrite_worm_append_files
##
## @description
##  This is an function which overwrite n bytes of a worm file with
##  a given string
##
## @synopsis
##  overwrite_worm_append_files($) 

sub overwrite_worm_append_files($) {
    my ($files) = @_;
    
    my (@new_files, $size, $to_byte, $mtime, $byte, $cmd, $out);
 
    foreach my $file_name (keys %$files) {
        my $file = $files->{$file_name}->{obj};
        push(@new_files, $file);
        $size = $file->size;
        $to_byte = $size + APPEND_SIZE;
        $mtime = $file->mtime;
        $byte = 'A';

        NOTE::logcomment("Appending 200b data to WORM APPEND file : %s and MTIME = %d ", $file->fullpath, $file->mtime);
        _overwrite($file->fullpath, $size, $to_byte, $byte);

        NOTE::logcomment("File Size Before Appending : $size \n File Size After Appending : %s", $file->size);
        NOTE::logcomment("After Appending data to WORM APPEND file : %s and NEW MTIME = %d ", $file->fullpath, $file->mtime);
        NOTE::logcomment("MTIME is getting modified after appending data to WORM Append Files") if($mtime != $file->mtime);

        $cmd = $file->fullpath;
        $out = `stat $cmd`;
        NOTE::logcomment("File Stat After Overwriting : $out "); 
 
        $mtime = $file->mtime;
        $byte = 'O';
        sleep(30);

        NOTE::logcomment("OVERWRITING 200b data in WORM APPEND file : %s and \n MTIME = %d ",$file->fullpath,$file->mtime);
        _overwrite($file->fullpath,APPEND_SIZE, '100', $byte);

        $cmd = $file->fullpath;
        $out = `stat $cmd`;
        NOTE::logcomment("File Stat After Overwriting : $out ");  
        
        NOTE::logcomment("After OVERWRITING data in WORM APPEND file : %s and MTIME = %d ",$file->fullpath,$file->mtime);
        NOTE::logcomment("MTIME is getting modified after overwriting data to WORM Append Files") if($mtime != $file->mtime);
	last;
    }   

    my $hash = _save_file_data_to_hash(\@new_files);
    return $hash;
}

## @name ::create_worm_append_files
##
## @description
##  Create a worm append file
##
## @synopsis
##  create_worm_append_files($$$) 

sub create_worm_append_files($$$) {
    my ($root, $count, $retention) = @_;
    my (@files, $dir, $file_to_append, $vol);

    ## Create a Directory
    $dir = $root->nfs;
    
    ## Create a File for appending the data
    $file_to_append = $dir->create->fill(max=>'1m');

    ## Set the snaplock_default_period for the snaplock volume to retention (default is 600s) 
    eval { $vol = $root->Volume; };
    $vol = $root if($@);
    $vol->options("snaplock_default_period"=>$retention);    
  
    ## Iterate to create the Worm Append files
    while($count) {
        my $file = $dir->create->chmod(0777)->chmod(0444)->chmod(0666);
               
        NOTE::logcomment("File size after creation : %d and FILE COUNT = $count ",$file->size);  
        my $mtime = $file->mtime;
        my $ret = $file->retention;
                    
        NOTE::logcomment("WORM Append Files before Appending Data : ....$mtime :: Retention = $ret");
        _append($file->fullpath,$file_to_append->fullpath);

        $mtime = $file->mtime;

        $ret = $file->retention;
        NOTE::logcomment("WORM Append Files After Appending Data : $mtime :: Retention = $ret");
        eval{$file->rm;};
        if($@) {
            NOTE::logcomment("As expected Not able to delete Append file $@");
        }
        else { 
            die "ABle to ddlete worm append file after creation";
        }
        $count--;
        push(@files,$file);
    }
        
    ## Set the snaplock_default_period back to default values
    if ($vol->snaplock eq 'compliance') {
        $vol->options("snaplock_default_period"=>'max');
    } elsif($vol->snaplock eq 'enterprise') {
         $vol->options("snaplock_default_period"=>'min');
    }
    $file_to_append->rm;
    return \@files;
}

=head2 ::wait_for_dump_image

Wait for the dump image to get created when resync happens between slc to slc volumes
The input arguments are Destination volume name, Dump Path, Destination qtree name.
Return dump file if it gets created.

The format is as follows
         wait_for_dump_image($vol, $path, $qtree);

=item Arguments

B<vol> => I<Vol object> | 'NOTE:Volume'

I<Required.>
A Volume object.

B<path> => I<Path to dump image>

I<Required.>
Path to dump image.

B<qtree> => I<Qtree object> | 'NOTE:Qtree'

I<Required.>
A Qtree object.

=item Examples

=back

     $wait_for_dump_image($vol, $path, $qtree);

=head2 ::status

Returns the dump file which got created during resync

=over

=cut

sub wait_for_dump_image($$$) {
    my ($vol, $path, $qtree) = @_;
   
    my ($dump_image_file, $MAX_TIME_OUT) = (1, 20);
    my $dump_image;
    sleep(10);
    
    while($MAX_TIME_OUT) {
        my $dir = NOTE::FileSystem::NFS::File->init($vol,$path);
        my @dirs = $dir->children;
     
        foreach my $dir  (@dirs) {
           if($dir->name =~ /$qtree/) {
               my @files = $dir->children;
               foreach my $file (@files) {
                   if($file->name =~ /dump_image/) { 
                       return $file if( $file->is_worm );
                   }
               }  
           }
       }
        
       NOTE::logcomment("Sleeping for 10s as the script is waiting for the dump_image to get created");
       sleep(10);
       $MAX_TIME_OUT= $MAX_TIME_OUT - 10;
       NOTE::logcomment("Check for the Dump Image again under %s");
   }    
   die "Not able to get the dump_image Details";
}

=head2 ::restore_vol_dump_image

Restore the dump file created to another volume.
It takes the vol object of dump vol on which image file resides.
The dump file name on which dump image of vol resides.
And restore vol object where dump image has to be restored.

The format is as follows
         restore_vol_dump_image($dump_vol_obj, $dump_file_name, $restore_vol_obj);

=item Arguments

B<dump_vol_obj> => I<Dump volume  Object>

I<Required.>
Dump volume  Object.

B<dump_file_name> => I<dump file name>

I<Required.>
A dump file.

B<restore_vol_obj> => I<Restore vol  object>

I<Required.>
A Volume object to which dump image will be restored.

=item Examples

=back

     $restore_vol_dump_image($dump_vol_obj, $dump_file_name, $restore_vol_obj);

=head2 ::status

Returns 1 on success, 0 on fail

=over

=cut



sub restore_vol_dump_image {
    my ($dump_vol_obj, $dump_file_name, $restore_vol_obj) = @_;

    my $cmd = 'restore rfD ';
    my $dump_vol_name = $dump_vol_obj->name;
    my $dump_file_path = '/vol/' . $dump_vol_name . "/$dump_file_name" ;
    my $restore_vol_path = '/vol/' . $restore_vol_obj->name;
    my $TIMEOUT = 300;

    my $full_cmd = $cmd . " $dump_file_path" . " $restore_vol_path" ;

    my $response = $restore_vol_obj->Filer->say ($full_cmd);

    if ( $response =~ /RESTORE IS DONE/ && $response !~ /read-only/i){
        logcomment("RESTORE COMPLETED \n $response");
        return 1;
    }else { return 0; }

}

=head2 ::dump_vol_image

Dump the volume image on file.
It takes the source vol object of which image has to be taken.
The dump file name on which dump image of vol is to be taken.
And dump vol object where dump image has to be restored.

The format is as follows
         dump_vol_image ($src_vol_obj, $dump_file_name, $dump_vol_obj);

=item Arguments

B<src_vol_obj> => I<source volume  Object>

I<Required.>
source volume Object.

B<dump_file_name> => I<dump file name>

I<Required.>
A dump file.

B<dump_vol_obj> => I<dump vol object>

I<Required.>
A dump Volume object on which dump image will be taken.

=item Examples

=back

     $dump_vol_image($src_vol_obj, $dump_file_name, $dump_vol_obj);

=head2 ::status

Returns dump file object on success, 0 on fail

=over

=cut

sub dump_vol_image {

   my ($src_vol_obj, $dump_file_name, $dump_vol_obj)  = @_;

   my $cmd = 'dump f ';
   my $dump_vol_name = $dump_vol_obj->name;
   my $dump_file_path = '/vol/' . $dump_vol_name . "/$dump_file_name" ;
   my $src_vol_path = '/vol/' . $src_vol_obj->name;
   my $TIMEOUT = 300;
   my $filer_name = $dump_vol_obj->Filer->host;
   my $full_cmd = $cmd . " $dump_file_path" . " $src_vol_path" ;
   my $response = $dump_vol_obj->Filer->say ($full_cmd);

   if ( $response =~ /DUMP IS DONE/ && $response !~ /read-only/i){
        logcomment("DUMP COMPLETED \n $response");
        $dump_vol_obj->offline; $dump_vol_obj->online;
        $dump_vol_obj = NOTE::Volume->attach(
                                Filer => $filer_name,
                                name  => $dump_vol_name,
                       );
        $dump_vol_obj->unexportfs;
        $dump_vol_obj->exportfs( 'rw,nosuid,anon=0');
        my @file_objs = $dump_vol_obj->nfs->children;
        foreach my $file_obj (@file_objs){
          my $f_name = $file_obj->name;
          print "Children files in the $dump_vol_name $f_name \n";
          if ($f_name =~  $dump_file_name){
            return $file_obj;
          }
        }
    }else { return 0; }
}


=head2 ::restore_dump_image

Restore the dump file created to another qtree.
It takes the snapmirror object, dump image and qtree name to be restored.

The format is as follows
         restore_dump_image($sm, $dump_image, $qtree);

=item Arguments

B<sm> => I<Snapmirror Object>

I<Required.>
Snapmirror object.

B<dump_image> => I<dump file name> 

I<Required.>
A dump file.

B<qtree> => I<Qtree object> | 'NOTE:Qtree'

I<Required.>
A Qtree object to which dump image will be restored.

=item Examples

=back

     $restore_dump_image($sm, $dump_image, $qtree);

=head2 ::status

Returns the qtree at which dump file got restored

=over

=cut

sub restore_dump_image($$$) {
    my ($sm, $dump_file, $qtree) = @_;
    my ($cmd, $src, $dst, $full_cmd, $result);
  
    $cmd = 'restore rfD /vol/';
    $src = $sm->dst->Volume->name.$dump_file->path;
    $dst = ' /vol/'.$qtree->Volume->name.'/'.$qtree->name;
    $full_cmd = $cmd.$src.$dst;

    $result = $sm->dst->Filer->say($full_cmd);
    if ( $result =~ /RESTORE IS DONE/ && $result !~ /read-only/i){
        NOTE::logcomment("RESTORE COMPLETED \n $result");     
        return $qtree;
    }
    die "RESTORE FAILED - $result";
}

=head2 ::verify_dump_data

Verify dump data is intact with the original file system

The format is as follows
         verfify_dump_data($org_data, $qtree);

=item Arguments

B<org_data> => I<\$data>

I<Required.>
Data before resync happened

B<qtree> => I<Qtree object> | 'NOTE:Qtree'

I<Required.>
Restored qtree object 

=item Examples

=back

     $verify_dump_data($data, $qtree);

=head2 ::status

Returns 1 if dump data is intact with the file system.
Returns 0 if dump data is not intact with the file system.

=over

=cut

sub verify_dump_data ($$) {
    my ($orig_data, $qtree) = @_;
    my ($ref_count, $count) = (0, 0);
    my ($data, @restored_files); 

    $data = $orig_data;
    @restored_files = $qtree->nfs->children;
    foreach my $restored_file (@restored_files) {
        if($restored_file->name !~ /snapshot/ &&
       	   $restored_file->name !~ /restore_symboltable/
        ) {
            NOTE::logcomment("For RefCount - Restore File : %s : $ref_count", $restored_file->name );
            foreach my $type (keys %$data) {
                my $types = $data->{$type};
                foreach my $file_name (keys %$types) {
                    if($file_name eq $restored_file->name) {
            		$ref_count++;
                        NOTE::logcomment("VERIFYING *** $type *** DATA ");
                        NOTE::logcomment("Restored File: %s : ATIME = %d",$restored_file->name,$restored_file->atime);
                        NOTE::logcomment("Original File: $file_name : ATIME = %d ",$types->{$file_name}->{atime});
                        if($types->{$file_name}->{atime} == $restored_file->atime) {
                            $count++;
                        }
                    }   
                }
            }
        } 
     } 
    return 1 if($count == $ref_count);
    return 0;
}

# This is an internal utility function which converts array to an hash

sub _save_file_data_to_hash($) {
    my ($files) = @_;
    my %data;
    
    foreach my $file (@$files) {
        $data{$file->name}->{obj} = $file;   
        $data{$file->name}->{name} = $file->name;   
        $data{$file->name}->{atime} = $file->atime;
        $data{$file->name}->{mtime} = $file->mtime;
        $data{$file->name}->{ctime} = $file->ctime;
    }
    return \%data;
}

## @ name :: create_files
## @ Arguments - NOTE container Object ( volume or qtree object )

sub create_file($) {
    my ($root) = @_;

    my $file = $root->nfs->create;
    return $file;
}

## @ name :: create and fill the files
## @ Arguments - NOTE container Object ( volume or qtree object )
##               size of the file

sub create_and_fill_file($$) {
    my ($root, $size) = @_;

    my $file = create_file($root);
    $file->fill(max=>$size);
    return $file;
}

## @ name :: Commit the files to WORM
## @ Arguments - Array reference of NOTE File Object
##               retention period

sub commit_to_worm($$) {
    my ($files, $retention) = @_;

    foreach my $file (@$files){
        $file->commit(retention=>$retention);
	last;
    }
}

## @ name :: Commit the files to WORM Append
## @ Arguments - Array reference of NOTE File Object ( with Zero length)
##               retention period

sub commit_to_worm_append($$) {
    my ($files,$retention) = @_;

    my $set = 0;
    my ($vol, $default_retention);

    foreach my $file (@$files) {
        if (!$set) {
            $vol = $file->Volume;
            $default_retention = $vol->options("snaplock_default_period");
            $vol->options("snaplock_default_period" => $retention);
            $set=1;
        }
        $file->chmod(0777)->chmod(0444)->chmod(0666);
	last;
    }
}

## @name ::create_expired_worm_files
##
## @description
##   Create expired worm files by making retention of the file to be 5s.
##
## @synopsis
##   create_expired_worm_files($)

sub create_expired_worm_files($$) {
    my ($root, $count) = @_;

    my @files;
    while ($count) {
        my $file = $root->nfs
                        ->create
                        ->fill(max=>"1m")
                        ->commit(retention =>"5s");
        push(@files, $file);
        $count--;
	last;
    }
    wait_till_expiry(\@files);
    return \@files
}

sub flip_setup($$)
{
    my ($data,$dst_qtree) = @_;

    my @files = $dst_qtree->nfs->children;
    my %hash;

    foreach my $file (@files) {
        if($file->name !~ /snapshot/) {
            foreach my $type (keys %$data){
                my $types = $data->{$type};
                foreach my $file_name (keys %$types){
                    my $fobj = $types->{$file_name}->{obj};
                    if($file->name eq $file_name){
                        $hash{$type}->{$file->name}->{obj} = $file;
                        $hash{$type}->{$file->name}->{name} = $file->name;
                        $hash{$type}->{$file->name}->{atime} = $file->atime;
                        $hash{$type}->{$file->name}->{mtime} = $file->mtime;
                        $hash{$type}->{$file->name}->{ctime} = $file->ctime;
                        $hash{$type}->{$file->name}->{parent}= $file->parent;
                    }
                }
            }
        }

    }
    return \%hash;

}

sub wack_and_iron($$$)
{
    my ($filer,$vol, $error_exp) = @_;

    require "Dense/setup_simple.pl";

    my ($vol_obj,$vol_type,$target);

    #$vol_obj = $vol->status;
    #$vol_type = $vol_obj->type();
    if($vol->type eq 'trad')
    {
        $target = $vol->name;
    } else {
        $target = $vol->Aggregate->name;
    }
    if ( $error_exp ) {
    	_wack_and_report(filerhost=>$filer,aggr=>$target,error_expected=>$error_exp );
    } else {
	_wack_and_report(filerhost=>$filer,aggr=>$target);
    }
    return 1;
}

## Local Wack and report subroutine
## Dense::Wack_and_iron epxects ASIS license on all code lines

sub _wack_and_report($@) {
	    my (%opts) = @_;

    my $host = host($opts{filerhost});
    my $aggrName = $opts{aggr};
    my $conn = connect("#rsh");

    my $wack_result;
    my $output = '';

    $opts{error_expected} = 0 unless $opts{error_expected};
    $opts{commit} = 1 unless $opts{commit};

    $wack_result = wack($host, args=>$aggrName, commit=>$opts{commit}, output=>\$output, timeout=>$opts{timeout});

    if (!$wack_result) {
        die "Wack problem:$@";
    } elsif ($wack_result eq WACK_CLEAN) {
        if ($opts{error_expected} == 1){
            die "File system is unexpectedly clean by Wack\n";
        } else {
            logcomment("File system is clean as expected by Wack");
        }
    } elsif ($wack_result eq WACK_DIRTY) {
       if ($opts{error_expected} == 0){
           die "File system is unexpectedly dirty by Wack\n";
       } else {
           logcomment("File system is corrupt and fixed as expected by Wack")
       }
    } else {
        die "Wack is in unknown state: $wack_result\n";
    }

    logcomment("Filer boot after wack");
    boot($host) or die $@;
}

sub check_filer_version($) {
    my ($src) = @_;
    my ($version, $codename);
    my $is_as_version = 0;
    $version = filer_version_get($src->Filer->conn);

    NOTE::logcomment("Version = $version");
    $codename =  codename_to_version($version);
   
    my @codename_split = split(/\./,$codename); 
 
    if ( $codename_split[1] == 0 ) {
	$is_as_version = 1 ;
    } 
   return $is_as_version;
}

=head2 ::get_filer_cc

Get the filer complice clock date in the format of yyyymmddhhmm.ss.
This date can be used in setting file cc time in v-sims for testing.

The format is as follows
         get_filer_cc ($Filer_obj);

=item Arguments

B<Filer_obj> => I<Filer NOTE object>

I<Required.>
A Filer NOTE object.


=item Examples

=back

    get_filer_cc ($Filer_obj);

=head2 ::status

Returns cc date in format of yyyymmddhhmm.ss

=over

=cut


sub get_filer_cc {

 my ($Filer_obj) = @_;
 my $cc_time = compliance_clock_status($Filer_obj->conn);
 my $gmt = gmtime ("$cc_time");
 my $FilerCCtime = &ParseDate($gmt);
 my @temp_ar = split /:/,$FilerCCtime;
  $FilerCCtime = $temp_ar[0].$temp_ar[1].'.'. $temp_ar[2];
  return $FilerCCtime;
}

=head2 ::set_filer_cc

Set the filer complice clock date to a specified date in the format of yyyymmddhhmm.ss.
This works in the Simulators only and in the test mode.

The format is as follows
         set_filer_cc ($Filer_obj,$date);

=item Arguments

B<Filer_obj> => I<Filer NOTE object>

I<Required.>
A Filer NOTE object.

B<date> => I<cc date of filer to be changed>

I<Required.>
A date of filer

=item Examples

=back

    set_filer_cc ($Filer_obj,$date);

=head2 ::status

Returns 1 on success and 0 on Fail.

=over

=cut


sub set_filer_cc {

  my ($filer,$date) =  @_;
  my %op_hash ;
  my $is_sim  =  $filer->is_a_simulator;
      if ( $is_sim != 1){
        return 0;
      }

  my $cmnd = 'priv set test; '. 'date -c ' . $date ;
  my (undef,undef,$response) = say($filer->conn,$cmnd);
  my @date_ar = split (//,$date);
  my $year = join (//,@date_ar[0..3]);
  my $month = join (//,@date_ar[4..5]);
  my $day = join (//,@date_ar[6..7]);
  my $hr = join (//,@date_ar[8..9]);
  my $min = join (//,@date_ar[10..11]);
  my $sec = join (//,@date_ar[13..14]);
  my $epoc_time = timegm($sec,$min,$hr,$day,$month - 1,$year);

  if ( $response =~ /WARNING: resetting the compliance clock is only permitted on simulators/){
       my $cc_time = compliance_clock_status($filer->conn);
       if ( ($cc_time - $epoc_time) < 20) { return 1;}
       else { return 0; }
  }else { return 0; }
}


=head2 ::get_qt_cc_sec

Get the qtree cc metafile sec and nsec values of both
source and destination qtree's in QSM relation.

The format is as follows
         get_qt_cc_sec ($qt);

=item Arguments

B<qt> => I<Qtree object>

I<Required.>
A Qtree NOTE object.

=item Examples

=back

    get_qt_cc_sec ($Filer_obj,$date);

=head2 ::status

Returns hash ref having dst/src sec & nsec time .

=over

=cut

sub get_qt_cc_sec {
  my ($qt) = @_;
  my $sec_i = 5;
  my $nsec_i = 8;
  my %qt_cc_hash ;
  my $qt_name  = $qt->name;
  my $vol_name = $qt->Volume->name;
  my $cmnd = "priv set test;qtree_metafile_read";
  my $full_cmnd = "$cmnd " . "$vol_name " . "$qt_name ";
  my $response = $qt->Filer->say ($full_cmnd);

  my @message_lines = split /\n/, $response;
  foreach my $line (@message_lines){
   if ( $line =~ /Source secure time/){
     my @seconds = split / /, $line ;
     $seconds[$sec_i] =~ s/,//;
     $qt_cc_hash{'src_secure_time'}{'sec'} = $seconds[$sec_i];
     $qt_cc_hash{'src_secure_time'}{'nsec'} = int ($seconds[$nsec_i]) ;
   }
   if ( $line =~ /Destination secure time/){
     my @seconds = split / /, $line ;
     $seconds[$sec_i] =~ s/,//;
     $qt_cc_hash{'dst_secure_time'}{'sec'} = $seconds[$sec_i];
     $qt_cc_hash{'dst_secure_time'}{'nsec'} = int ($seconds[$nsec_i]);
   }
  }

 return \%qt_cc_hash;
}


1;

