# Hey emacs, this is basically -*- perl -*-
#
#Copyright (c) 2004 Network Appliance, Inc.
#All rights reserved.
#
## @summary Snaplock Qsm setup functions 
##
## @description
##
## $Id: //depot/prod/test/main/lib/SnapLock/Snaplock_qsm.pm
##
## @usage
##   use Snaplock_qsm;
##
## @author kbala@netapp.com
##
## @status private

package SnapLock::Snaplock_qsm;

use strict;
use base qw(Exporter);
use Cwd qw(getcwd);
use Carp qw(croak);
use Tharn;
use NOTE;
use SnapLock::Snaplock_constants;
use SnapLock::Snaplock_datamanip;
use snaplock qw(compliance_clock_status);
use Date::Manip;
use Time::ParseDate;
use Time::Local;
use Wreck;
use constant SLC_AGGR => "wormaggr_compliance";
use constant SLE_AGGR => "wormaggr_enterprise";
use constant REG_AGGR => "testaggr";


BEGIN
{
  our @EXPORT = qw(
		   sl_qsm_setup
		   corrupt_atime_and_insert_disk_flags_on_nonsl_volume
		   corrupt_atime_and_insert_disk_flags_on_sl_volume
		   inject_wreck
                   generate_qt_name
                   filer_cc
                   create_src_dst_qt  
  );
}

use constant FLEX_SIZE => '350m';

=head2 ::sl_qsm_setup

This function initializes the compliance clock on both source and destination filer
It also creates source volume, destination volume and source qtree 

The format is as follows
         sl_qsm_setup(\%args);

=item Arguments

B<src_filer> => I<$FILER_SRC> 

I<Required.>
Source filer

B<dst_filer> => I<$FILER_DST>

I<Required.>
Destination filer

B<src_vol_sl_type> => I<regular> | I<compliance> | I<enterprise>

I<Required.>
It can be either regular or compliance or enterprise

=item Examples

=back

     $sl_qsm_setup(\%args);

=head2 ::status

Returns an array of filer, volume and qtree objects for both source and destination 

=over

=cut

#---------------------------------------------------------------------
# sl_qsm_setup
#---------------------------------------------------------------------
sub sl_qsm_setup($) {
    my ($args) = @_;

    my ($setup, $params, $vol_src, $vol_dst, $filer_src, $filer_dst, $qt_src, $qt_dst, @data);
    my $QTREE_NAME = 'src_qtree';
    ## @step initialize compliance clock on source and destination filer
    $filer_src = NOTE::Filer->connect(
                                       host => $args->{src_filer},
                                       mode => '#rsh',
                                       stats => 0
                                     );

    $filer_dst = NOTE::Filer->connect(
                                       host => $args->{dst_filer},
                                       mode => '#rsh',
                                       stats => 0
                                     );

    push(@data, $filer_src);
    push(@data, $filer_dst);

    snaplock::snaplock_enterprise_license($filer_src->conn);
    snaplock::snaplock_enterprise_license($filer_dst->conn);

    snaplock::compliance_clock_initialize($filer_src->conn, strict=>0, reinit_ok=>1);
    snaplock::compliance_clock_initialize($filer_dst->conn, strict=>0, reinit_ok=>1);

    #Instantiate the Setup object to get the global parameters
    $setup = dp::snapmirror::Setup->new();
    $params = $setup->get_parameters();

    my ($SRC_AGGR_NAME , $DST_AGGR_NAME);
    if ( $args->{src_vol_sl_type} =~ /compliance/ ) {
	$SRC_AGGR_NAME = SLC_AGGR;
    }elsif( $args->{src_vol_sl_type} =~ /enterprise/ ) {
	$SRC_AGGR_NAME = SLE_AGGR;
    }else {
	$SRC_AGGR_NAME = REG_AGGR;
    }

    if ( $args->{dst_vol_sl_type} =~ /compliance/ ) {
	$DST_AGGR_NAME = SLC_AGGR;
    }elsif($args->{dst_vol_sl_type} =~ /enterprise/ ) {
	$DST_AGGR_NAME =  SLE_AGGR;
    }else {
        $DST_AGGR_NAME = REG_AGGR;
    }


    ## @step Create source volume
    $vol_src = NOTE::Volume->create(
                                     Filer     => { host => $params->{FILER_SRC} },
                                     tag       => 'vol_src',
                                     type      => $params->{VOL_TYPE},
                                     name      => $params->{FILER_SRC_VOL},
                                     snaplock  => $args->{src_vol_sl_type},
                                     flex      => { size    => FLEX_SIZE },
                                     Aggregate => { name    => $SRC_AGGR_NAME, raidtype => 'raid4',
                                                    recycle => 1 },
                                     trad      => { disks   => $params->{TRAD_SIZE},
                                                    recycle => 1 },
                                    );


    ## @step Create destination volume
    $vol_dst = NOTE::Volume->create(
                                     Filer     => { host => $params->{FILER_DST} },
                                     tag       => 'vol_dst',
                                     type      => $params->{VOL_TYPE},
                                     snaplock  => $args->{dst_vol_sl_type}, 
                                     name      => $params->{FILER_DST_VOL},
                                     flex      => { size    => FLEX_SIZE },
                                     Aggregate => { name    => $DST_AGGR_NAME, raidtype => 'raid4',
                                                    recycle => 1 },
                                     trad      => { disks   => $params->{TRAD_SIZE},
                                                    recycle => 1 },
                                   );

    push(@data, $vol_src);
    push(@data, $vol_dst);

    ## @step Create qtree on soruce volume
    $qt_src = NOTE::Qtree->create(
                                   Volume => $vol_src,
                                   name   => $QTREE_NAME,
                                   # Force anonymous, otherwise, parent
                                   # volume will not be recyclable
                                   anonymous => 1,
                                 );

    # This qtree will be available because it is part
    # of the dataset of its containing volume
    $qt_src = NOTE::Qtree->attach(
                                   Volume => $vol_src,
                                   name   => $QTREE_NAME,
                                   tag    => 'qt_src',
                                 );

    # Destination qtree object definition,
    # the qtree is not actually created on the filer
    # since it will be created with snapmirror
    $qt_dst = NOTE::Qtree->new(
                                Volume => $vol_dst,
                                tag    => 'qt_dst',
                               );

    push(@data, $qt_src);
    push(@data, $qt_dst);
    return @data;
}

sub corrupt_atime_and_insert_disk_flags_on_nonsl_volume ($$$) {

    my ( $qt_src , $vol_src , $conn ) = @_;

    ## Populate data on source volume and corrupt the data set with
    ## Inserting EXCL flag and change the sec and nsec values of atime.

    my $data1 = populate_regular_data($qt_src);

    ## Create first set of data.
    my $atime_nsec = NSEC_LESS_THAN_ONE_BILLION;
    my $atime_sec = 0;
    my $disk_flag = NFS_VERIFIER_FLAG ;
    inject_wreck($conn, $data1 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec ,$disk_flag);

    ## Second data set
    my $data2 = populate_regular_data($qt_src);
    $atime_nsec = NSEC_ONE_BILLION;
    inject_wreck($conn, $data2 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec ,$disk_flag);

    ### Third set
    my $data3 =  populate_regular_data($qt_src);
    $atime_nsec = NSEC_GREATER_THAN_ONE_BILLION;
    inject_wreck($conn, $data3 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec ,$disk_flag);

    ### Fourth Set
    my $data4 = populate_regular_data($qt_src);
    $atime_nsec = NSEC_NEGETIVE_VALUE;
    inject_wreck($conn, $data4 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec ,$disk_flag);

    ### Fifth set , Without EXCL disk flag set on the inode.
    $atime_nsec = NSEC_LESS_THAN_ONE_BILLION;
    my $data5 = populate_regular_data($qt_src);
    inject_wreck($conn, $data5 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec);

    ### Sixth Data Set.
    $atime_nsec = NSEC_ONE_BILLION;
    my $data6 = populate_regular_data($qt_src);
    inject_wreck($conn, $data6 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec);

    ### Seventh Data Set.
    $atime_nsec = NSEC_GREATER_THAN_ONE_BILLION;
    my $data7 = populate_regular_data($qt_src);
    inject_wreck($conn, $data7 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec);

    ### Eight Data Set.
    $atime_nsec = NSEC_NEGETIVE_VALUE;
    my $data8 = populate_regular_data($qt_src);
    inject_wreck($conn, $data8 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec);

    ### Ninenth Data Set.
    ### Corrupt only disk flag.
    my $atime = 0;
    my $data9 = populate_regular_data($qt_src);
    inject_wreck($conn, $data9 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec, $disk_flag , $atime );

    ### Tenth Data Set.
    ### Files without any corruption.
    my $data10 = populate_regular_data($qt_src);

}

sub corrupt_atime_and_insert_disk_flags_on_sl_volume ($$$$) {
    my ($qt_src , $vol_src, $conn, $ret_ref) = @_;
    
    ## Create first set of data.
    my $atime_nsec = NSEC_LESS_THAN_ONE_BILLION;
    my $atime_sec = 0;
    my $disk_flag = NFS_VERIFIER_FLAG;

    my $data1 = populate_worm_data($qt_src, $ret_ref);
    inject_wreck($conn, $data1 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec ,$disk_flag);

    ## Second data set
    my $data2 = populate_worm_data($qt_src, $ret_ref);
    $atime_nsec = NSEC_ONE_BILLION;
    inject_wreck($conn, $data2 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec ,$disk_flag);

    ### Third set
    my $data3 =  populate_worm_data($qt_src,$ret_ref);
    $atime_nsec = NSEC_GREATER_THAN_ONE_BILLION;
    inject_wreck($conn, $data3 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec ,$disk_flag);

    ### Fourth Set
    my $data4 = populate_worm_data($qt_src,$ret_ref);
    $atime_nsec = NSEC_NEGETIVE_VALUE;
    inject_wreck($conn, $data4 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec ,$disk_flag);

    ### Fifth set , Without EXCL disk flag set on the inode.
    $atime_nsec = NSEC_LESS_THAN_ONE_BILLION;
    my $data5 = populate_worm_data($qt_src,$ret_ref);
    inject_wreck($conn, $data5 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec);

    ### Sixth Data Set.
    $atime_nsec = NSEC_ONE_BILLION;
    my $data6 = populate_worm_data($qt_src,$ret_ref);
    inject_wreck($conn, $data6 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec);

    ### Seventh Data Set.
    $atime_nsec = NSEC_GREATER_THAN_ONE_BILLION;
    my $data7 = populate_worm_data($qt_src,$ret_ref);
    inject_wreck($conn, $data7 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec);

    ### Eight Data Set.
    $atime_nsec = NSEC_NEGETIVE_VALUE;
    my $data8 = populate_worm_data($qt_src,$ret_ref);
    inject_wreck($conn, $data8 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec);

    ### Ninenth Data Set.
    ### Corrupt only disk flag.
    my $atime = 0;
    my $data9 = populate_worm_data($qt_src,$ret_ref);
    inject_wreck($conn, $data9 , $vol_src , $qt_src ,  $atime_sec , $atime_nsec, $disk_flag , $atime );

    ### Tenth Data Set.
    ### Files without any corruption.
    my $data10 = populate_worm_data($qt_src,$ret_ref);

}

sub inject_wreck ($$$$$$;$;$) {

       my $lines_atime;
       my $lines_flag;
       my ( $conn, $data, $vol_obj ,$qtree_obj , $sec , $nsec, $flag , $noatime ) = @_;
       my $qtree_name = $qtree_obj->name;
       foreach my $type ( keys %$data ) {
                my $types = $data->{$type};
                foreach my $file_name ( keys %$types ) {
                        my $complete_path = "$qtree_name/$file_name";
                        if ( $sec == 0 ) {
                                my $file_obj = $types->{$file_name}->{obj};
                                $sec = $file_obj->atime;
                        }
                        $lines_atime.= "inode_atime inum=path:/$complete_path sec=$sec nsec=$nsec\n";
                        if ( defined $flag ) {
                            $lines_flag.="inode_flags_insert inum=path:/$complete_path flags=$flag\n";
                        }
                }
        }
        if ( defined $flag ) {
                wreck ( $conn, $vol_obj->name , "type" => 'flex' , "inject_script"=> $lines_flag ) or logresult("fatal" , "Wreck failed: $@" );
        }
        $vol_obj->online;
        if ( !defined $noatime ) {
                wreck ( $conn, $vol_obj->name , "type" => 'flex' , "inject_script" => $lines_atime ) or logresult("fatal" , "Wreck failed: $@" );
        }

        $vol_obj->online;
}


=head2 ::create_src_dst_qt

Create src and destination qtree object, which can be used for setting qsm relationship
using NOTE::SnapMirror.

The format is as follows
         create_src_dst_qt ($src_vol_obj, $dst_vol_obj,$qt_name)

=item Arguments

B<src_vol_obj> => I<Volume NOTE object>

I<Required.>
A Volume  NOTE object.

B<dst_vol_obj> => I<Volume NOTE object>

I<Required.>
A Volume  NOTE object.


B<qt_name> => I<string>

I<Required.>
A stirng.

=item Examples

=back

   create_src_dst_qt ($src_vol_obj, $dst_vol_obj,$qt_name);

=head2 ::status

Returns ref to hash having values of source and destination NOTE::Qtree objects

=over

=cut


sub create_src_dst_qt {

 my ( $src_vol_obj, $dst_vol_obj,$qt_name) = @_;
 my %qt ;

 $qt{'src_qt_obj'} = NOTE::Qtree->create (
                                    Volume => $src_vol_obj,
                                    name => $qt_name,
                               );
   $qt{'dst_qt_obj'} = NOTE::Qtree->new(
         Volume => $dst_vol_obj,
         name => $qt_name,
         tag    => 'destination'
     );


      return \%qt;
}



=head2 ::filer_cc

Get the vcc (volume complice clock) date in the format of yyyymmddhh:mm:ss.
This date can be used for debugging purpose.

The format is as follows
         filer_cc ($vol_obj);

=item Arguments

B<Volume_obj> => I<Volume NOTE object>

I<Required.>
A Volume NOTE object.


=item Examples

=back

    filer_cc ($Filer_obj);

=head2 ::status

Returns cc date in format of yyyymmddhh:mm:ss

=over

=cut


sub filer_cc {

 	my ($vol_obj) = @_;
	my $version = $vol_obj->Filer->version;	
	my $cc_time;
	if ($version > 8 ){
		$cc_time = $vol_obj->compliance_clock;	
	}else {
	 	$cc_time = compliance_clock_status($vol_obj->Filer->conn,$vol_obj->name);
	}
 	my $gmt = gmtime ("$cc_time");
 	my $FilerCCtime = &ParseDate($gmt);
  	return $FilerCCtime;
}



=head2 ::generate_qt_name

Creates the qtree name by conncatnating type of src and dest and interarion id.

The format is as follows

   generate_qt_name ($src_vol_type,$dst_vol_type,$run_id)

=item Arguments

B<src_vol_type> => I<string>

I<Required.>
A string.


B<dst_vol_type> => I<string>

I<Required.>
A string.


B<run_id> => I<string>

I<Required.>
A string.

=item Examples

=back

   generate_qt_name ($src_vol_type,$dst_vol_type,$run_id)

=head2 ::status

Returns string.

=over

=cut


sub generate_qt_name {
       my ($src_vol_type,$dst_vol_type,$run_id) = @_;
       my $qt_name = $src_vol_type. '_' . $dst_vol_type . '_' . $run_id;
       return $qt_name;
 }


1;

