#$Id:$

## @summary  SnapLock PrivDel and WORMLogging library.
## @pod      here
## @status   Public
## @author   Anusuya K ( anusuya.k@netapp.com, dl-snaplock-qa(DRG QA))

=head1 NAME

SnapLock::WORMLogging

=head1 DESCRIPTION

These Library functions help in configuring and verifying 
privileged delete and WORM logging functionality.

=head1 METHODS

=cut

use strict;

package SnapLock::WORMLogging;

require Exporter;
use vars ('@ISA','@EXPORT');

BEGIN {
 @ISA = qw(Exporter);
 @EXPORT = qw(
	create_volume
	common_privdel_setup
	get_conn
	create_user
	delete_user
	delete_group
	add_role
	modify_role
        delete_role
	get_worm_log_vol
	get_worm_log_file_path 
	snaplock_log_status
	set_worm_log_vol 
	privdel_file privdel_option
	get_previous_log_file
	is_exists_log_file
	set_distinct_telnet
	config_ssl
	config_ssh
	login_to_console
	check_files_got_deleted
	verify_privdel_execution
	fill_the_volume
        delete_the_file
	verify_state_transition_log
	verify_system_log
	snaplock_log_archive
	worm_log_file_size
	snaplock_log_close
        get_filer_connection
	set_volume_security_style
	get_ntfs_file_details
        populate_worm_files_over_cifs
	calculate_record_expiry
	verify_snaplock_scan_log_file
	insert_nfs_verifier_flag
	read_snaplock_scan_file
	get_snaplock_scan_file
	
    );

}

use Tharn;
use NOTE;
use SayWrap;
use Zapi;
use ZapiErrno;
use NaElement;
use Volume qw(vol_online);
use XML::Simple;
use warnings;
use Date::EzDate;
use Data::Dumper;
use SnapLock::fingerprint;
use SnapLock::Snaplock_constants;
use OntapSSH::SecureAdmin;
use snaplock;
use Utils qw(make_host make_conn);
use Users qw(addCIFSUser findCIFSUser removeCIFSUser);
use Params::Validate qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF
UNDEF);
use wafl_utils qw(df);
use FS::Util::Utilities qw(write_file);
use FS::Util::Utilities qw(util_mkfile
                           nfs_cifs_mkfile
                           nfs_cifs_rdonly
                           nfs_cifs_rd_write
                           nfs_cifs_append);
use Filer::Version qw(get_wafl_version WV_ROLLINGROCK_MIN) ;
use boot qw( transit_to UP);
use Os::OsUtils;
use Wreck;
use Carp qw(confess croak);
use constant LOG_VOL  => "WORMLog_vol";
use constant SLE_VOL  => "SleVol";
use constant TRAD_SLC_VOL => "trad_compliance";
use constant TRAD_SLE_VOL => "trad_enterprise";
use constant TRAD_REG_VOL => "trad_regular";
use constant PRIVDEL_PASS => "No error";
use constant PRIVDEL_STATE_FAIL => "Illegal capability transition";
use constant HW_WORM_LOG_VOL_RET => "1800s";
use constant SIM_WORM_LOG_VOL_RET => "1800s";
use constant DEFAULT_USER => "root";
use constant SLC_AGGR => "wormaggr_compliance";
use constant SLE_AGGR => "wormaggr_enterprise";
use constant REG_AGGR => "testaggr";

## Global hash to maintain the VCC of the log volume at the time of privdel.
my %log_vol_vcc_comp;
=head2 B<privdel/wormlogging setup>

=cut

=head2 ::create_volume

Creates a volume and returns an object representing the created volume. 

=over

=item Arguments

B<Filer> => I<host spec>

I<Required.>
The filer on which to create the volume.

B<name> => I<string>

I<Optional.>
The name of the volume.
If absent, an anonymous volume will be generated.
If present the volume and the containing aggregate (if there is one)
will be spared during cleanup.

B<snaplock> => C<compliance | enterprise>

I<Optionsl.>
If set, specifies which type of snaplock volume to create.

B<size> => I<size spec>

I<optional.>
The size to make the volume.
Only valid for flexible volumes.
If absent the default size is the minimum flexible volume size, 20MB.

B<type> => C<flex> | C<trad>

I<Optional.>
The type of volume to create.
The default is C<flex>.

B<trad> => C<{> I<::create options> c<}>

I<Optional.>
Additional C<::create> options that only take effect
if the type of the volume being created is C<trad>.

=item Examples

=back

	$vol = create_volume( Filer => $FILER,
      		              type => 'flex',
                	      size => '20m',
        );

=cut 

sub create_volume(%) {
    my (%vol_data) = @_;
    my $vol_obj;
    my $aggr_name;

    if ( $vol_data{snaplock} eq 'compliance' ) {
	$aggr_name = SLC_AGGR;
    } elsif ( $vol_data{snaplock} eq 'enterprise' ) {
	$aggr_name = SLE_AGGR;
    } else {
	$aggr_name = REG_AGGR;
    }
 
    if ( $vol_data{type} eq 'trad' ) {
	if ( $vol_data{snaplock} eq 'compliance' ) {
		$vol_data{name} = TRAD_SLC_VOL;
        }elsif ( $vol_data{snaplock} eq 'enterprise' ) {
		$vol_data{name} = TRAD_SLE_VOL;
        } else {
		$vol_data{name} = TRAD_REG_VOL;
	}
    }

    $vol_obj = NOTE::Volume->create (
		Filer 		=> $vol_data{Filer},
		name		=> $vol_data{name},
		snaplock 	=> $vol_data{snaplock},
		type 		=> $vol_data{type},
		flex 		=> { size => $vol_data{size}} ,
		Aggregate 	=> { name => $aggr_name ,raidtype => 'raid4', 
				     recycle => 1 },
		trad 		=> { disks => 2 , raidtype => 'raid4', },
	);
    die "Failed to create Volume\n" if ( !$vol_obj ) ;
    return $vol_obj;
}

=head2 ::common_privdel_setup

Create common privileged delete setup.
Setup includes:
 Create required volumes,
 Configure WORMLog volume,
 Enable privileged delete,
 Configure and enable SSH/SSL setup,
 Create user belongs to Compliance administrator group,
 Connection to perform privileged delete.
Returns WormLog Volume and Privdelete enabled volume
objects, Default filer connection object and returns privdel
filer connection if I<PrivDelConn> is specified.

=over

=item Arguments

B<Filer> => I<host spec>.

I<Required.>
Filer on which test will run

B<Client> => I<host spec>

I<Required.>
Unix Client on which the test will get triggered.

B<Zapi> => C< 0 | 1 >

I<Optional.>
If set uses ZAPI calls to perform privileged operation.
Default is 0.

B<User> => I<compliance user name>.

I<Optional.>
Required only for deleting the file.

B<Passwd> =>I<compliance user password>.

I<Optional.>
Required only for deleting the file.

B<Group> => I<Group to which user should belong to>.

I<Optional.>
Required only for creating new user.

B<PrivDelConn> => C<ssh | zapi | telnet | console | rsh >

I<Optional.>
Network protocol to perform privileged delete of the file.
Required only for deleting the file.

=item Examples

=back

	($log_vol, $sle_vol, $filer_conn) = common_privdel_setup(
						Filer      => $FILER,
	);

=cut 

sub common_privdel_setup($) {
    my ($args) = @_;
	
    my ( $log_vol , $sle_vol ,$priv_del_conn, @arr );
    ### Validate check needs to be added to validate the parameters
    NOTE::logcomment("Creating Setup required for PrivDel execution\n");
    my $filer_conn = NOTE::Filer->connect(
			host => $args->{Filer},
			mode => '#rsh',
			status => 0
		);
    snaplock::snaplock_enterprise_license($filer_conn->conn);
    snaplock::compliance_clock_initialize($filer_conn->conn, strict=>0, reinit_ok=>1);
    ## @Step  Get/Create Required Volums and Aggregate.
    my %vol_info = (
			Filer => $args->{Filer},
			type  => 'flex',
			size  => '30m',
			snaplock => 'compliance',
    );

    #Using volume size passed from the script
    if (defined ($args->{'slc_vol_size'})){
        $vol_info{size} = $args->{slc_vol_size} ; 
    }

    if ( defined $args->{volume_type} && $args->{volume_type} ) {
        ### If Volume type is defined Create trad volumes
        ### Create Traditional Compliance volume.
        $vol_info{type} = 'trad';
	$log_vol = create_volume(%vol_info);
	### Create Traditional Enterprise volume.
	$vol_info{snaplock} = 'enterprise';
	$sle_vol = create_volume(%vol_info);
    } else {
	## Default create flex volumes.
	##( $log_vol , $sle_vol ) = flex_volume_setup($args->{Filer});
	## Create flex compliance volume.
	if (defined ($args->{'slc_vol_name'})) {
	    $vol_info{name} = $args->{'slc_vol_name'};
	}else {    
	    $vol_info{name} = LOG_VOL;
	}   
	$log_vol = create_volume(%vol_info);
	## Create flex Enterprise volume.
	$vol_info{name} = SLE_VOL;
	$vol_info{snaplock} = 'enterprise';
	$sle_vol = create_volume(%vol_info);
    }
    ## @Step Set WormLog Volume.
    my $out = set_worm_log_vol($filer_conn->conn, $log_vol->name , use_zapi => $args->{zapi});
    if ( $out ) {
        die "Failed to set WORM LOG Volume $out\n";
    }else {
        ### Sucessfully set worm log volume, select $log_vol for execution.
	NOTE::logcomment("logvol object selected\n");
        push(@arr, $log_vol);
    }
    ## @Step Enable privdel on Sle volume if not enabled before.
    my $priv_del = privdel_option($sle_vol, use_zapi =>$args->{zapi});	
    if ( !$priv_del ) {
	die "snaplock priv_delete failed to get option: $priv_del\n";
    } else {
	   ## Enter here if the privdel option is off/uninitialized.
	if ( $priv_del ne "on" && $priv_del !~ /disallowed/){
            $out = privdel_option($sle_vol , option => "on", use_zapi => $args->{zapi});
            if ( $out ) {
	        die "Failed to enable privdel $out";
            } else {
 		## Privdel is enabled, select $sle_vol for execution
		NOTE::logcomment("slevol object selected\n");
                push(@arr, $sle_vol);
            }
        } else {
	    ### Enter here if the privdel option is "disallowed".
            if ( $priv_del =~ /disallowed/ ) {
                die "Privdel state is disallowed, cannot use this volume for privdel the file\n";
            }else {
 		## If the privdel is already "on" , select $sle_vol for execution.
		NOTE::logcomment("slevol object selected\n");
		push(@arr, $sle_vol);
           }
        }
    }
    push(@arr, $filer_conn);
    ## @Step  Configure Required connection..
    if ( $args->{priv_del_conn} ) {
        if ( $args->{user} && $args->{passwd} && $args->{group} ) {
	## @Step  Create user belongs to Compliance Administrator Group.
	    create_user($filer_conn->conn, $args->{user}, $args->{passwd}, $args->{group});
	}
	else {
	    die "Require user details to get privileged delete connection and to perform privleged delete\n"
	}

	## @Step Configure SSH or SSL based on privdel connection.
        $priv_del_conn = get_filer_connection($args->{Filer}, $args->{client}, $filer_conn , $args->{user} , $args->{passwd} , $args->{priv_del_conn} );

	push(@arr, $priv_del_conn);
    }
    NOTE::logcomment("Steup Completed\n");
    return @arr;
}

=head2 ::get_filer_connection

Perofrms the resuired setup and returns the required connection

=over

=item Arguments

B<Filer> => Hostspec of the filer.

I<Required.>
Host spec of the filer.

B<client> => Hostspec of the linux client.

I<Required.>
Host spec of the linux client.

B<conn_type> => C<rsh | telnet | ssh>

I<Required.>
Network Protocol type to connect to filer.

B<user_name> => I<Login as user>

I<Required.>
Login to filer as a specified user.

B<Password> => I<User password>

I<Required.>
User password.

=item Examples

=back

        $conn = get_filer_connection($Filer , $client ,$filer_conn , $user_name, $password , $conn_type);

=cut

sub get_filer_connection($$$$$$) {

    my ( $Filer , $client , $filer_conn, $user, $passwd , $conn_type ) = @_;	
    my $priv_del_conn;

    NOTE::logcomment("Entered get_filer_connection\n"); 
    if ( $conn_type =~ /zapi/i ) {
	## Verify and Configure SSL.
	config_ssl($filer_conn->host, $filer_conn->conn);
        $priv_del_conn = get_conn($conn_type , $Filer , $user, $passwd );
    } elsif ( $conn_type =~ /ssh/i ) {
	config_ssh($Filer, $client, $filer_conn->conn);
	set_distinct_telnet($filer_conn->conn , "on");
	my $default_user = DEFAULT_USER;
	my $default_passwd = $filer_conn->conn->hostp()->default_password();
        my $hosttype = $filer_conn->{hostp}->{hosttype};
	my $con_conn = login_to_console($Filer, $default_user , $default_passwd );
	disconnect($con_conn);

	$priv_del_conn = get_conn($conn_type , $Filer , $user, $passwd );
    } elsif ( $conn_type =~ /console/i ) {
	set_distinct_telnet($filer_conn->conn , "on");
        my $hosttype = $filer_conn->{hostp}->{hosttype};
	$priv_del_conn = login_to_console($Filer , $user, $passwd );
    } else {
	$priv_del_conn = get_conn($conn_type , $Filer , $user, $passwd );
    }

    NOTE::logcomment("Exitting from get_filer_connection\n");
    return $priv_del_conn;

}

	
=head2 ::get_conn

Connects to filer with the given connection as a given user.

=over

=item Arguments

B<conn_type> => C<rsh | telnet | ssh>

I<Required.>
Network Protocol type to connect to filer.

B<filer> => I<host spec>

I<Required.>
Filer on which test should run.

B<user_name> => I<Login as user>

I<Required.>
Login to filer as a specified user.

B<Password> => I<User password>

I<Required.>
User password.

=item Examples

=back

	$conn = get_conn($conn_type , $filer , $user_name, $password);

=cut

sub get_conn($$$$) {

    my ($con_type ,$NODE, $user , $pass) = @_;

    my $hconn = make_host($NODE);
    my $c_type;
    if ($con_type =~ /ssh/i)
    {
        $c_type = "ssh-persist";
    }
    elsif ($con_type =~ /rlm/i)
    {
        $c_type = "ssh-persist";
    }
    elsif ($con_type =~ /telnet/i)
    {
        $c_type = "telnet";
    }
    elsif ($con_type =~ /rsh/i)
    {
        $c_type = "rsh-persay";
    }
    my $conn = undef;
    my $c_obj = undef;
    # make a random connection ID so that a new connection is made everytime
    my $con_id = rand();

    if ( ($con_type =~ "rlm") ) {
        my $RLM = "-rlm";
        my $rlm_user;
        if ( $user =~ "root" ) {
	     $rlm_user = "naroot" ;
        } else {
	     $rlm_user = $user;
	}
        $hconn->{"connection_fields"} = {$con_id => {conntype => $c_type, username => $rlm_user, password => $pass}};
        $conn = "#$con_id;conntype=$c_type;connhost=$NODE$RLM;username=$rlm_user;password=$pass";
    } else {
        $hconn->{"connection_fields"} = {$con_id => {conntype => $c_type, username => $user, password => $pass}};
        $conn = "#$con_id;conntype=$c_type;username=$user;password=$pass";
    }
    eval { $c_obj = Tharn::connect($conn, "-login",  "-timeout", 30); };
    if ($@) {
       logcomment ("INFO : Tharn::connect failed. Return conObj = undef. $@");
       return undef;
    }
    logcomment ("INFO : Tharn::connect successful. cObj after Login = $c_obj");

    # For some reason once in a while the login strings show up after the return from
    # this routine. Not sure why but it screws up the other commands. So wait a few
    # seconds after the connect returns and attempt to clear the buffers before continuing.

    sleep (5);
    eval { say( $c_obj, -blind, -swallow); };

    if ( ($con_type =~ "rlm") ) {
        logcomment("INFO : Execute RLM startConsole");
        if (startConsole ($c_obj, $user, $pass) == 0) {
	    logcomment ("INFO : startConsole failed. Return conObj = undef");
        } else {
            logcomment ("INFO : startConsole successful.");
	}
    }
    logcomment("INFO :  get_conn successfully completed. Return conObj = $c_obj");
    return $c_obj;
}


# Internal function set the worm log volume retention
sub _set_log_volume_retention($;$) {
    my ($filer_conn , $retention ) = @_;

    my $hosttype = $filer_conn->{hostp}->{hosttype};

    NOTE::logcomment("Printing Host type : $hosttype\n");

    my $ret;
    if ( defined $retention ) {
        $ret = $retention
    } elsif ( $hosttype =~ /(simulator|vsim)/i ) { 
        $ret = SIM_WORM_LOG_VOL_RET; 
    } else {
        $ret  = HW_WORM_LOG_VOL_RET;
    }
	my $output = sayEOF($filer_conn, "options snaplock.log.default_retention $ret");

}
	

=head2 ::config_ssh

Configure and enable SSH on the filer.

=over

=item Arguments

B<filer> => I<Host spec>

I<Required.>
Hostspec of the filer.

B<client> => I<client host spec>
Hostspec of the unix client

B<filer_conn> => I<rsh connection>
Filer connection to execute commands on the filer.

=item Examples

=back
	
	config_ssh($filer_name , $client_name , $filer_conn);

=cut	

sub config_ssh($$$) {

    my ($filer, $client, $filer_conn) = @_;

    NOTE::logcomment(" Check SSH Setup");
    my $is_ssh_setup = _check_ssh_setup($filer, $client, $filer_conn);	
    if ( !$is_ssh_setup ) {
        die "Failed to setup SSH.  Can not run Secure test.";
    } else {
        NOTE::logcomment("SSH is set up and ready.");
	my $output = sayEOF( $filer_conn, "options ssh.enable on" );
    }
}

=head2 ::config_ssl

Configure and enable SSL on the filer.

=over

=item Arguments

B<filer_host> => I<Host of the filer>

I<Required.>
Hout of the filer.

B<filer_conn> => I<rsh connection>

I<Required>
RSH connection of the filer.

=item Examples

=back

        config_ssl($filer_host , $filer_conn);

=cut
sub config_ssl($$) {

    my ($filer_host, $filer_conn ) = @_;

    NOTE::logcomment(" Check SSL Setup");
    my $is_ssl_setup = _check_ssl_setup($filer_host,$filer_conn);
    if ( !$is_ssl_setup ) {
 	die "Failed to setup SSL.  Can not run https test.";
    } else {
        NOTE::logcomment("SSL is set up and ready.");
    	my $out = sayEOF( $filer_conn, "options httpd.admin.ssl.enable on" );
    }
}

=head2 ::create_user

Creates the given user belongs to given group only if
the user does not exists before.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required.>
Connection to the filer to execute the command.

B<username> => I<user name>

I<Required.>
Name of the user to be created.

B<password> => I<password>

I<Required>
Password for the new user.

B<group> => I<user group>

I<Required>
User Group to which new user should belong.

=item Examples

=back

        create_user($filer_conn, $user_name, $passwd , $user_group );

=cut

sub create_user($$$$) {

    NOTE::logcomment("Entering create_user function\n");
    my ( $filer_conn, $user_name, $passwd , $user_group ) = @_;

    ## Enable distinct telnet session
    set_distinct_telnet($filer_conn , "on");

    my $find_user= findCIFSUser($filer_conn, $user_name); 
    if ( !defined $find_user ) {
	addCIFSUser($filer_conn, $user_name , $passwd , $user_group);
    }
    NOTE::logcomment("Exiting from create_user function\n");
}

=head2 ::delete_user

Deletes the given user from the filer.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required.>
Connection to the filer to execute the command.

B<username> => I<user name>

I<Required.>
Name of the user to be deleted.

=item Examples

=back

        delete_user($filer_conn, $user_name );

=cut
sub delete_user($$) {
    my ( $filer_conn, $username ) = @_;

    my $find_user = findCIFSUser($filer_conn, $username);
    print Dumper($find_user);
    if ( defined $find_user ) {
        NOTE::logcomment("Deleting user $username\n");
        removeCIFSUser($filer_conn, $username );
    } else {
        NOTE::logcomment("User $username does not exists\n");
        return 1;
    }

    ## Verify User has deleted or not
    $find_user = findCIFSUser($filer_conn, $username);
    if ( !defined $find_user ) {
        NOTE::logcomment("Sucessfully deleted $username\n");
    } else {
        NOTE::logcomment("Failed to delete $username\n");
        return 1;
    }
}


=head2 ::delete_group

Deletes the given group from the filer.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required.>
Connection to the filer to execute the command.

B<groupname> => I<group name>

I<Required.>
Name of the group to be deleted.

=item Examples

=back
	
	$result = delete_group($filer_conn, $group_name);

=cut
sub delete_group($$) {

    my ( $filer_conn, $group_name )=  @_;

    my $out = sayEOF($filer_conn , "useradmin group delete $group_name");
    if ( $out =~ /Error: Cannot delete default group/ ) {
        NOTE::logcomment("$group_name deletion Failed\n");
        return $out;
    } else {
        NOTE::logcomment("The group '$group_name' has been deleted.\n");
        return;
    }
}

=head2 ::add_role

Add a role to the filer.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required.>
Connection to the filer to execute the command.

B<rolename> => I<role name>

I<Required.>
Name of the role to be added.

B<rolestring> => I<roles to create>

I<Required.>
role(s) to create.

=item Examples

=back

	$result = add_role ($filer_conn , $rolename , $rolestring );

=cut
sub add_role( $$$ ) {

    my ( $filer_conn , $rolename , $rolestring ) = 
                            validate_pos(@_,
                                    {type => SCALAR|HASHREF},
                                    {type => SCALAR},
                                    {type => SCALAR}
                            );
                                                

    my $out = sayEOF($filer_conn , "useradmin role add $rolename -a $rolestring");
    if ( $out =~ /Could not add role/ ) {
        NOTE::logcomment("Failed to add role\n");
        return $out;
    } else {
        NOTE::logcomment("The role $rolename has been added.\n");
        return ;
    }
}

=head2 ::modify_role

Add a role to the filer.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required.>
Connection to the filer to execute the command.

B<rolename> => I<role name>

I<Required.>
Name of the role to be added.

B<rolestring> => I<roles to create>

I<Required.>
role(s) to create.

=item Examples

=back

	$out = modify_role($filer_conn , $rolename , $rolestring );

=cut

sub modify_role($$$) {

    my ($filer_conn , $rolename , $rolestring ) =
                        validate_pos(@_,
                            { type => SCALAR|HASHREF },
                            { type => SCALAR },
                            { type => SCALAR }
                        );
    my $out = sayEOF($filer_conn, "useradmin role modify $rolename -a $rolestring" );
    if ( $out =~ /Could not modify role/ ) {
        NOTE::logcomment("Failed to modify role\n");
        return $out;
    } else {
        NOTE::logcomment("The role $rolename has been modified.\n");
        return ;
    }
}

=head2 ::delete_role

Delete a role from the filer.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required.>
Connection to the filer to execute the command.

B<rolename> => I<role name>

I<Required.>
Name of the role to be deleted.

=item Examples

=back

        $out = delete_role($filer_conn , $rolename );

=cut

sub delete_role($$) {

    my ($filer_conn , $rolename ) =
                        validate_pos(@_,
                            { type => SCALAR|HASHREF },
                            { type => SCALAR },
                        );
    my $out = sayEOF($filer_conn, "useradmin role delete $rolename " );
    ( $out =~ /role.*deleted/i ) ? return 1 : return $out;

}


	
=head2 ::get_worm_log_vol

Gets the name of the snaplock WORM Log volume. Returns
WORM log volume name on success or failure, with an error
message in $@ on failure.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required>
Filer connection to execute the command on the filer.

B<options> => C<{> I<::log options> C<}>

I<Optional.>
$options{use_zapi} if specified ZAPI is used to get the
log volume name. By default CLI is used.

=item Examples

=back

	$log_vol = ( $filer_conn,
			use_zapi => 1 );


=cut
sub get_worm_log_vol($;%) {

    my ($filer_conn, %options) = @_;

    NOTE::logcomment("Entering get_worm_log_vol function\n");
    # ZAPI EXECUTION
    if ( $options{use_zapi} ) {
	my $zcmd = new NaElement("snaplock-get-log-volume");
	#logcomment("command sent: ". $zcmd->toString());
	my $zapi = Zapi->new($filer_conn->hostp()->default_ip(),
				DEFAULT_USER,
				$filer_conn->hostp()->default_password());
	$zapi->{fail} = \&_fail_default;
	my $xo = $zapi->invoke_elem($zcmd,"any");
	#logcomment("snaplock-get-log-volume ZAPI returns: ". $xo->toString()); 
	if ( ! $xo->results_ok() ) {
	    $@  = "ZAPI snaplock-get-log-volume". $xo->toString();
	    return 0;
	}
	my $log_vol = $xo->child_get("log-volume")->{content};
	logcomment("log_vol is : $log_vol");
	if ( $log_vol !~ /Not configured/ ) {
	    ##Validate that the volume exists on the filer_conn
	    return $log_vol if ( vol_online($filer_conn, $log_vol) ); 
	}
	return 0;
    }

    ### CLI Execution
    return eval {
	my $output = sayEOF($filer_conn, "snaplock log volume");
	die "$output\n" if (!$output);
	#return ($output) if ( $output !~ /Not configured/ && vol_online($filer_conn, $output ));
	return ($output) if ( $output !~ /Not configured/ );
	return 0;
    }
}

=head2 ::set_worm_log_vol

Make Compliance volume as a WORM Log volume.

=over

=item Arguments

B<conn> => I<NOTE Connection object>

I<Required.>
Filer connection is required to execute the command on the filer.

B<vol_name>  => I<Name of volume>

I<Required.>
Given volume will be set as worm log volume.

B<options> => C<{> I<::log options> C<}>

I<Optional.>
$options{use_zapi} if specified ZAPI is used to set the

$options{force_option} : If specified , skip the get_worm_log_vol 
                        verification.
$options{retention} : If specified, the given retention is 
log volume. By default CLI is used.

=item Examples

=back

	$out = set_worm_log_vol ( $filer_conn , $vol_name,
				   use_zapi => 0 , force_option => 0,
				   retention => "300s" );


=cut

sub set_worm_log_vol($$;%) {
    my ($filer_conn,$vol_name, %options) = @_;
    NOTE::logcomment("Entering set_worm_log_vol function\n");
    if ( !defined $options{force_option} ) {
        my $log_vol = get_worm_log_vol($filer_conn);
	if ( defined($log_vol) && $log_vol =~ /$vol_name/ ) {
            my $out = snaplock_log_status($filer_conn, $vol_name);
	    if ( $out =~ /snaplock log status failed/i) {
		NOTE::logcomment("Volume $vol_name has not set\n");
            } else {
		NOTE::logcomment("Volume $vol_name has already been set as a worm log volume");
		return 0;
            }
        }
    }
    ## If the log volume is new, set the default retention period before setting
    ## volume as a log volume
    if ( defined $options{retention} ) {
	    _set_log_volume_retention($filer_conn, $options{retention});
    } else {
        _set_log_volume_retention($filer_conn);
    }

    ## Make sure that snaplock license is there in the filer
    snaplock::snaplock_license($filer_conn);
	
    if ( $options{use_zapi}) {

	my $zcmd = new NaElement("snaplock-set-log-volume");
	$zcmd->child_add(new NaElement("log-volume", $vol_name));
	$zcmd->child_add(new NaElement("is-force", 1 ));
	my $zapi = Zapi->new($filer_conn->hostp()->default_ip(),
                                        DEFAULT_USER,
                                       $filer_conn->hostp()->default_password());
        $zapi->{fail} = \&_fail_default;
	my $xo = $zapi->invoke_elem($zcmd,"any");
        #logcomment("snaplock-set-log-volume ZAPI returns: ". $xo->toString());
		
	if ( ! $xo->results_ok() ) {
            my $error  = "ZAPI snaplock-set-log-volume". $xo->toString();
            return $error;
        }
	return 0;
    }
    if ( $options{force_option} eq "skip" ) {
        return eval {
	    my $output = sayEOF($filer_conn, "snaplock log volume $vol_name");
	    return $output if ( $output );
	    return 0;
        }
    } else {
        return eval {
	    my $output = sayEOF($filer_conn, "snaplock log volume -f $vol_name");
	    return $output if ( $output );
	    return 0;
        }
    }  
}

=head2 ::get_worm_log_file_path

Gets the worm log files path( system log file , privdel log file )
and returns the required path.

=over

=item Argument

B<filer_conn>  => I<Filer connection>

I<Required.>
Filer connection to execute command on the filer.

B<wormlog> => I<Log volume name>

I<Required.>
Name of Active wormlog volume.

B<basename>  => C<priv_delete | system_log>

I<Required.>
Basename of the worm log file. It can be either priv_delete or
system_log.

B<options>

I<Optional.>
$options{use_zapi} if specified ZAPI is used to get the
worm log files. By default CLI is used.

=item Examples

=back

	## Get the complete path of "priv_delete" file.
	$log_file_path = ( $filer_conn , $worm_log_vol , "priv_delete" , 
			    use_zapi => 0 );

=cut
sub get_worm_log_file_path($$$;%) {
    my ( $filer_conn, $worm_log_vol, $basename , %options ) = @_;

    NOTE::logcomment("Entering get_worm_log_files_path  function\n");
    $options{basename} = $basename;
    my $output = snaplock_log_status($filer_conn, $worm_log_vol, %options );
    if ( $output =~ /snaplock log status failed/i ) {
        NOTE::logcomment("Failed to get log status output\n");
        return;
    }else { 
        my $file_path = $output->{$basename}->{"log-complete-path"};
        NOTE::logcomment("exit from get_worm_log_files_path  function\n");
	return $file_path;
   }     
}

=head2 :: snaplock_log_status

Get complete output of "snaplock log status" command.

=over

=item Arguments

B<filer_conn>  => I<Filer connection>

I<Required.>
Filer connection to execute command on the filer.

B<wormlog> => I<Log volume name>

I<Required.>
Name of Active wormlog volume.

B<options> => C<{> I<log options> C<}>

I<Optional.>
log options includes
use_zapi : To get log status output using ZAPI
basename : Basename of the wormlog file.
Default is to get all active files.

=item Examples

=back

	$log_status = snaplock_log_status($filer_conn, $log_vol_name , )
			     use_zapi => 0 , basename => "priv_delete" );

=cut 
sub snaplock_log_status($$;%) {

    my ( $filer_conn, $worm_log_vol, %options ) = @_;

    NOTE::logcomment("Entering snaplock log status  function\n");

    ## Snaplock log status can be executed on enterprise volume
    ## if the snaplock scanner fix has created snaplock scanner log file.
    ## Make sure that snaplock/snaplock_enterprise license is installed
   
    if ( defined $options{snaplock_type} && $options{snaplock_type} =~ /enterprise/ ) {
	snaplock::snaplock_enterprise_license($filer_conn);
    } else {
    	snaplock::snaplock_license($filer_conn);
    }

    ## ZAPI Execution.
    if ( $options{use_zapi}  && $options{basename} ) {
        my $basename = $options{basename};
        my $zcmd = new NaElement("snaplock-log-status-list-info");
	$zcmd->child_add(new NaElement("log-volume", $worm_log_vol));
	$zcmd->child_add(new NaElement("log-basename", $basename));
	my $zapi = Zapi->new($filer_conn->hostp()->default_ip(),
                                        DEFAULT_USER,
                                        $filer_conn->hostp()->default_password());
        $zapi->{fail} = \&_fail_default;
	my $xo = $zapi->invoke_elem($zcmd,"any");
	if ( ! $xo->results_ok() ) {
            return $xo->toString();
        }
	my $v1 = $xo->child_get("log-file-details");
	my $v2 = $v1->child_get("log-file-info");

	my %hash = ( 
		"file-size" => $v2->child_get("file-size")->{content},
		"formatted-retention-date" => $v2->child_get("formatted-retention-time")->{content},
		"log-complete-path" => $v2->child_get("log-complete-path")->{content},
		"log-basename" => $v2->child_get("log-basename")->{content},
	);
	my %log_status = (
		$basename => \%hash,
	);
	return \%log_status;
    }
    ## CLI execution.
    ## Execute "snaplock log status" for Given basename.
    if ( defined $options{basename} ) {
	my $basename = $options{basename};
        NOTE::logcomment("Printing Basename, $basename \n");
	return eval {
  	    my $output = sayEOF($filer_conn, "snaplock log status $worm_log_vol $basename" );
	    if ( $output =~ /snaplock log status failed/i) {
	        NOTE::logcomment("Failed to Get log status output: $output");
                return $output;
	    }
	    # Strip everything up to the beginning of the list
	    # (Keep everything after the"-------------")
	    die "Error stripping OutPut header: $output\n"
		unless ( $output =~ s/.*?\s*\-+\s*(.*)$/$1/s );
	    chomp $output;
            print $output , "\n";
            my ( %hash, %log_status );
	    foreach my $line (split(/\n/, $output)) {
	        if ( $line =~ /^\s*(\S+)\s*(.*)$/) {
		    my $name = $1;
		    my $value = $2;
		    my $newname;
		    if ( $name eq "Basename:" ) { $newname = "log-basename";
		    } elsif ( $name eq "Complete_path:" ) {
                        $newname = "log-complete-path"; 
		    } elsif ( $name eq "Expiry:" ) {     
                        $newname = "formatted-retention-date"; 
		    } elsif ( $name eq "Size:" ) { $newname = "file-size"; }
  		    $hash{$newname} = $value;
	        }	
	    }
	    $log_status{$basename} = \%hash; 
	    return \%log_status;
	}
    } else {
        ## Execute "snaplock log status" without basename.
	return eval {
	    my $output = sayEOF($filer_conn, "snaplock log status $worm_log_vol");
	    if ( $output =~ /snaplock log status failed/i) {
		NOTE::logcomment("Failed to Get log status output :$output");
		return $output;
	    }
            my $output1 = $output;
            my @split_output = split(/\n/, $output1 );
            my @split_output_line1 = split(/:/ , $split_output[0] );  
	    my $Active_log_file_count = $split_output_line1[2];
	    die "Error stripping OutPut header: $output\n"
		unless ( $output =~ s/.*?\s*\-+\s*(.*)$/$1/s );
	    chomp $output;
	    my $basename;
	    my (%hash0 , %hash1, %hash2,%log_status); 
	    foreach my $line (split(/\n/, $output)) {
		if ( $line =~ /^\s*(\S+)\s*(.*)$/) {
		    my $name = $1;
		    my $value = $2;
                    my $newname;
		    if ( $name eq "Basename:" ) { $newname = "log-basename";
						  $basename = $value ;}
		    elsif ( $name eq "Complete_path:" ) { $newname = "log-complete-path" ; }
		    elsif ( $name eq "Expiry:" ) { $newname = "formatted-retention-date" ; }
		    elsif ( $name eq "Size:" ) { $newname = "file-size" ; }

		    if ( $basename =~ /snaplock_scan/ ) {
			$hash0{$newname} = $value;
		    }elsif( $basename =~ /system_log/ ) {
			$hash1{$newname} = $value;
		    }elsif ( $basename =~ /priv_delete/ ) {
			$hash2{$newname} = $value;
		    }
		 
		}
	    }
	    $log_status{snaplock_scan} = \%hash0 if (%hash0);
	    $log_status{system_log} = \%hash1 if ( %hash1 );
            $log_status{priv_delete} = \%hash2 if ( %hash2 ) ;
	    $log_status{active_log_file_count} = $Active_log_file_count if ( defined ($Active_log_file_count));
	    return \%log_status;
	};
    }        
}

=head2 :: privdel_option

Gets or Set privdel option for the given volume.

The format is as follows
    privdel_option($vol_obj , %args)

=over

=item Arguments

B<vol object> => I<NOTE Volume object>

I<Required.>
NOTE Volume object.

B<args> => I<arguments to privdel option>

I<Optional>
Arguments include
use_zapi : Optional, If specied ZAPI is used for execution.
do_it    : Optional, If specified uses force option "-f" to change the
           state of privdel.
option   : Optional, If specified chage the privdel state of the volume.
           privdel option includes on|off|disallowed.
Default is to get the privdel state of the volume.           

=item Examples

=back

        ## Enable privdel on the given volume.
	privdel_option($vol_obj ,
			use_zapi => 0, option => "on" );
 
        ## Get the privdel option set on the volume.
        privdel_option($vol_obj ,
                        use_zapi => 0 );

=cut
sub privdel_option($;%) {

    my ( $vol_obj, %args ) = @_;
    my $vol_name = $vol_obj->name;
    my $filer_conn = $vol_obj->Filer->conn;

    NOTE::logcomment("Entering privdel_option function\n");
    ### Make sure that snaplock license is installed.
    snaplock::snaplock_license($filer_conn);
    ### Execute ZAPI
    if ( $args{use_zapi} ) {
	my $zapi = Zapi->new($filer_conn->hostp()->default_ip(),
                         DEFAULT_USER,
                         $filer_conn->hostp()->default_password());
	$zapi->{fail} = \&_fail_default;

        ## Get privdel state of the given volume.
	if ( !defined $args{option} ) {
	    my $zcmd = new NaElement("snaplock-get-options");
	    $zcmd->child_add(new NaElement("volume", $vol_name));
	    $zcmd->child_add(new NaElement("option", "privdel"));
	    my $xo = $zapi->invoke_elem($zcmd, "any");
	    #logDbg("snaplock-get-options ZAPI returns: ". $xo->toString());
	    if ( $xo->results_ok() ) {
		return $xo->child_get("option-value")->content;
	    }
	    return 0;
	} else { ## Change privdel state.
	    my $zcmd = new NaElement("snaplock-set-options");
	    $zcmd->child_add(new NaElement("volume", $vol_name));
	    $zcmd->child_add(new NaElement("option", "privdel"));
	    $zcmd->child_add(new NaElement("option-value", $args{option}));
	    if ( defined $args{do_it} && $args{do_it} eq "true" ) {
		$zcmd->child_add(new NaElement("do-it", $args{do_it} ));
	    }
	    my $xo = $zapi->invoke_elem($zcmd, "any");
          
	    NOTE::logcomment("Exit from privdel_option function");
            if ( $xo->results_ok() ) {
		return;
            }else {
                return $xo->toString();
            }
        }
    }
    ## CLI execution
    ## Get privdel state of the given volume.
    my $force_option;
    if ( ! defined $args{option} ) {
        return eval {
	    my $output = sayEOF($filer_conn, "snaplock options $vol_name privdel");
            return $output;
	}; 
    }else { ## Change the privdel state.
        if (( defined $args{do_it}) &&( $args{do_it} eq "true") ) {
	    $force_option = "-f";
	    return eval {
	        my $output = sayEOF($filer_conn, "snaplock options $force_option $vol_name privdel $args{option}");
	        NOTE::logcomment("Exit from privdel_option function");
	        return $output;
            };
        } else {
	    return eval {
	        my $output = sayEOF($filer_conn, "snaplock options $vol_name privdel $args{option}");
	        NOTE::logcomment("Exit from privdel_option function");
	        return $output;
            };
        }
    }
         
}

=head2 :: priv_delete_file

Delete the given file using ZAPI interface or CLI interface.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required>
Filer connection to execute commands on the filer.

B<file>  => I<File object>

I<Required.>
File object has complete file path, and file attributes like
atime, ctime, mtime , size etc.

B<options> => I<privdel options>

I<Required.>
Privdel option is hash name value pair, which includes
use_zapi : Optional < 0 | 1 >. If the value is 1, delete the file 
           through ZAPI API.
ZAPI_USES_HTTPS : Optional , If specified HTTPS is user, otherwise HTTP 
            will be used as a protocol.
do_it    : Required < true | false >. If true delete the file.
force_option: Optional < 1 | 0 >. If set delete the file with "-f" option.

There is default value, Either "force_option" or "do_it" should be present.

=item Example

=back

	privdel_file($filer_conn , $file_obj ,
		      use_zapi => 0 , force_option => 1 );

=cut
sub privdel_file($$$) {
    my ( $filer_conn, $file, $args ) = validate_pos( @_,
						{ type => SCALAR | HASHREF },
						{ type => HASHREF },
						{ type => HASHREF },
					);

    NOTE::logcomment("Entering privdel_file function\n");
    my $delete_file = $file->{file_path};
    ## Get the volume name from the file path.
    my @vol_split = split(/\// , $delete_file);
    my $data_volume = $vol_split[2];
    NOTE::logcomment("Deleting file $delete_file\n");	
    my  ( $rsh, $fh) ;
    if ( $args->{use_zapi} ) {
	$rsh = $filer_conn;
        $fh = $filer_conn->conn->{hostp};
    } else {
    	my $filer = $filer_conn->hostp()->{hostname};
        $rsh = NOTE::Filer->connect(
                        host => $filer,
                        mode => '#rsh',
                        status => 0
                );
        $fh = $rsh->conn->{hostp};
    }
    my $ver = get_wafl_version($fh);
    my $log_volume = get_worm_log_vol($rsh->conn);
    my $action;
    ## ZAPI Execution
    if ( $args->{use_zapi} ) {
        ## Privdel requires secure connection.
	if ( $args->{zapi_uses_https}) {
		$ENV{ZAPI_USES_HTTPS} = "yes";
	}
	my $zapi = Zapi->new($filer_conn->conn->hostp()->default_ip(),
				$args->{username}, $args->{password});
	$zapi->{fail} = \&_fail_default;
	$action = $args->{do_it};
	my $zcmd = new NaElement("snaplock-privileged-delete-file");
	$zcmd->child_add(new NaElement("path" , $delete_file ));
	$zcmd->child_add(new NaElement("do-it" , $action));
	my $xo = $zapi->invoke_elem($zcmd, "any");
	if ($xo->results_ok() ) {
	    if ( $ver >= WV_ROLLINGROCK_MIN ) {
                $log_vol_vcc_comp{$delete_file}{log_vol} = snaplock::compliance_clock_status($rsh->conn, $log_volume);
                $log_vol_vcc_comp{$delete_file}{data_vol} = snaplock::compliance_clock_status($rsh->conn,$data_volume);
            }
            return ;
        }
	my $error = $xo->toString();
	return $error;
    }

    return eval {
	if ( $args->{force_option}) {
	    ## Force delete the file.
	    my $output  = sayEOF($filer_conn, "snaplock privdel -f $delete_file", {timeout=>3600});
            if ( $ver >= WV_ROLLINGROCK_MIN ) {
                $log_vol_vcc_comp{$delete_file}{log_vol} = snaplock::compliance_clock_status($rsh->conn, $log_volume);
                $log_vol_vcc_comp{$delete_file}{data_vol} = snaplock::compliance_clock_status($rsh->conn,$data_volume);
	    }
	    return $output;
	} else {
	    ## Decide wether file should be deleted or not.
	    my $proceed="Are you sure you want to delete it (y/n)?";
	    if ( $args->{do_it} eq "true" ) { 	
		$action	= "y"; 
	    }else {
		## By default delete the file.
		$action = "n";
	    }
	    my ($match, $before, $after, $prompt ) =
				say($filer_conn, "snaplock privdel $delete_file", $proceed, -swallow, -noprompt, -eval_error);
	    die $@ if $@;
	    my $output;
	    if ( $match  =~ /$proceed/ ) {
	        ( undef, undef,$output)=
				 say($filer_conn, $action);
	    }
            if ( $ver >= WV_ROLLINGROCK_MIN ) {
                $log_vol_vcc_comp{$delete_file}{log_vol} = snaplock::compliance_clock_status($rsh->conn, $log_volume);
                $log_vol_vcc_comp{$delete_file}{data_vol} = snaplock::compliance_clock_status($rsh->conn,$data_volume);
            }
	    return $output;
	}
    };
}

=head2 ::snaplock_log_archive

Archive worm log files

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required>
Filer connection to execute commands on the filer.

B<worm_log_vol> => I<worm log volume>

I<Required>
Name of the worm log volume.

B<arg> => I<options for command>

I<Optional.>
arg{basename} : Optional: If specified, achive only the 
		given basename log file.

=item Examples

=back
        
        ## Archive only "priv_delete" file.
	$output = snaplock_log_archive ( $filer_conn, $worm_log_vol
					  basename => "priv_delete);

        ## Archive only "system_log" file.
	$output = snaplock_log_archive ( $filer_conn, $worm_log_vol
                                          basename => "system_log");
 
        ## Archive only "snaplock_scan" file.
	$output = snaplock_log_archive ( $filer_conn, $worm_log_vol
					  basename => "system_log");

        ## Archive Both Active Worm log files.
	$output = snaplock_log_archive ( $filer_conn, $worm_log_vol);

=cut

sub snaplock_log_archive($$;%){

    my ( $filer_conn, $worm_log_vol , %arg ) = @_;
    my $log_vol = $worm_log_vol->name;
    return eval{
        if ( defined $arg{basename} ) {
            my $basename = $arg{basename};
            my $output = sayEOF($filer_conn, "snaplock log archive $log_vol $basename");
            return $output;
        } else {
            my $output = sayEOF($filer_conn, "snaplock log archive $log_vol");
            return $output;
        }
    }

}
=head2 ::snaplock_log_close

Close worm log files

=over

=item Arguments

B<worm_log_vol> => I<worm log volume>

I<Required>
Note log volume object.

=item Examples

=back
	
	$output = snaplock_log_close($vol_obj);

=cut
sub snaplock_log_close($) {

    my ( $worm_log_vol ) = @_;
    my $logvol_name = $worm_log_vol->name;
    my $filer_conn = $worm_log_vol->Filer->conn;
    
    return eval {
        my $output = sayEOF($filer_conn, "priv set diag;snaplock log close $logvol_name");
        return $output;
    }
}

=head2 ::worm_log_file_size

Get or change the maximum size of WORM log files.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required>
Filer connection to execute commands on the filer.

B<file_size> => I<size of the file>

I<Required>
Specify the size of the file to be set.

=item Examples

=back 

	$output = worm_log_file_size($filer_conn , $file_size );

=cut

sub worm_log_file_size($;$){
    my ( $filer_conn, $file_size ) = @_;

    use constant DEFAULT_SIZE => "10m";
    my $out_re = qr(snaplock\.log\.maximum\_size\s+(\d+\S));
    my $cmd = "options snaplock.log.maximum_size";
 
    # Get the current setting
    my $output = sayEOF($filer_conn, $cmd);

    my ($current_size) = $output =~ /$out_re/;
    defined ( $current_size) or croak "$cmd failed: $output";

    return ($current_size ) if ( !defined ($file_size ) );
   
    # Change the setting 
    $file_size = DEFAULT_SIZE if ( $file_size eq 'default' );

    $cmd .= " $file_size ";
    $output = sayEOF($filer_conn, $cmd );
    croak "$cmd failed: $output" if (sayEOF($filer_conn, $cmd) ne "");
    
    return $current_size;
}
## Internal Function which converts the XML string to HASH

sub _XML_string_to_hash($) {
    my ( $xml_string) = @_;	
	
    die "String is not provided" if (!$xml_string);
    my $hash;
    my $xml_obj = XML::Simple->new( KeepRoot => 1, KeyAttr => 1) ;
    eval {
	$hash = $xml_obj->XMLin($xml_string);
    };
    NOTE::logresult('FATAL', "Failed to Convert XML to HASH $@" ) if ($@);
    return $hash;
}

## Internal function which breaks the Privdel log file.	
sub _split_logentries($){
    my ($logout) = @_;

    NOTE::logcomment("In _split_logentries function\n");
    my @logarray= ();
    my $end_string = "<\/LogEntry>";
    my @arr = split(/$end_string/, $logout);

    my $first_index = 0;
    while ( $first_index < $#arr ) {
	my $log_entry = $arr[$first_index].$end_string;
	my $log_hash	= _XML_string_to_hash($log_entry);
	$first_index = $first_index + 1;
	push(@logarray , $log_hash );
    }
    my @priv_del_logs = _get_pre_post_entry_of_log(@logarray); 
    NOTE::logcomment("Exiting from _split_logentries function\n");
    return @priv_del_logs;
}
	
## Internal function to seperate the pre and post entry of the privdel log.
sub _get_pre_post_entry_of_log(@) {
    my (@logs) = @_;
    my @priv_del_logs_arr = ();
    my $first_index = 0;
    my $second_index = 1;
    NOTE::logcomment("In _get_pre_post_entry_of_log function call\n");

    while ( $first_index < $#logs ) {
	my $pre_out = $logs[$first_index];
       	if (defined $pre_out->{LogEntry}->{snaplock_pre_privileged_delete_1} ) {
	    while ( $second_index <= $#logs ) {
		my $post_out = $logs[$second_index];
		if (defined $post_out->{LogEntry}->{snaplock_post_privileged_delete_1}  ) {
		    if ( $pre_out->{LogEntry}->{snaplock_pre_privileged_delete_1}->{sequence_number} eq $post_out->{LogEntry}->{snaplock_post_privileged_delete_1}->{sequence_number}) {
			my %priv_del_log = (
				pre => $pre_out,
				post => $post_out
			);
			push(@priv_del_logs_arr, \%priv_del_log);
			$second_index++;
			last; ## Got Pre Post Match, Break the loop.
		    }	
		}
		$second_index++;
	    }
	## New loop added for snaplock scan entries.
	} elsif ( defined $pre_out->{LogEntry}->{snaplock_pre_option_privileged_delete_1}) {
	    while ( $second_index <= $#logs ) {
		my $post_out = $logs[$second_index];
		if (defined $post_out->{LogEntry}->{snaplock_post_option_privileged_delete_1} ) {
		    if ( $pre_out->{LogEntry}->{snaplock_pre_option_privileged_delete_1}->{sequence_number} eq $post_out->{LogEntry}->{snaplock_post_option_privileged_delete_1}->{sequence_number} ) {
		        my %priv_del_log = (
				pre => $pre_out,
				post => $post_out
		        );
			push(@priv_del_logs_arr, \%priv_del_log);
			$second_index++;
			last;## Got Pre Post Match, Break the loop.
		    }
		}
		$second_index++;
	    }
	}elsif ( defined $pre_out->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}) {
            while ( $second_index <= $#logs ) {
                my $post_out = $logs[$second_index];
                if (defined $post_out->{LogEntry}->{snaplock_post_scanner_fix_retention_1} ) {
                    if ( $pre_out->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{sequence_number} eq $post_out->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{sequence_number} ) {
                        my %priv_del_log = (
                                pre => $pre_out,
                                post => $post_out
                        );
                        push(@priv_del_logs_arr, \%priv_del_log);
                        $second_index++;
                        last;## Got Pre Post Match, Break the loop.
                    }
                }
                $second_index++;
            }
        }
	$first_index++;
    }
    NOTE::logcomment("In exit from _get_pre_post_entry_of_log function call\n");
    return @priv_del_logs_arr;
}
		
=head2 ::check_files_got_deleted

Check file deleted through privdel still exists in the volume.

The format is as follows
    check_files_got($volobj, $list_of_deleted_files)

=over
   
=item Arguments

B<vol object> => I<NOTE volume object>

I<Required.>
NOTE Volume object

B<files> => I<Array of deleted files>

I<Required.>
Array of deleted files.

=item Examples

=back
	
	$out = check_files_got_deleted($vol_obj , $deleted_files );

=cut
sub check_files_got_deleted( $$ ) {
    my ( $vol_obj , $deleted_files ) =@_;
    my $vol_name = $vol_obj->name;
    my $status= 0;
    my @files_in_volume= get_data($vol_obj);
    foreach my $del_file ( @$deleted_files ) {
        foreach my $existing_file ( @files_in_volume ) {
	    if ( $del_file->{file_path} eq $existing_file->{file_path} ) {
		NOTE::logcomment("Privdelete has Failed to delete the file [%s]", $del_file->{file_path});
		$status =1 ;
	    }
	}
    }
    return $status;
}
	
=head2 ::get_previous_log_file

Get last archived worm log files. 

=over

=item Arguments

B<file path> => I<Worm log file path>

I<Required.>
Active worm log file path.

B<Vol object> => I<vol object>

I<Required.>
NOTE log volume object.

B<Basename> => I<Basename>

I<Required.>
Log file Basename. ( priv_delete | system_log )

=item Examples

=back
 
        ## Get the previous/last archived  "priv_delete" file.
	$prev_log_file = get_previous_log_file($active_file_path , $log_vol , "priv_delete");

=cut
sub get_previous_log_file($$$) {

	my ($new_file_path, $vol_obj,$basename ) = @_;

	my @split_path = split(/\//, $new_file_path );
	my $present_file = pop(@split_path);

	my @split_present_file = split(/_/, $present_file );
	pop(@split_present_file);
	my $time_stamp = pop(@split_present_file);
	
	my @files = $vol_obj->nfs->children;

	foreach my $file (@files) {
		if ( $file->name =~ /$time_stamp/ && $file->name !~ /present/ &&
        $file->name =~ /$basename/) {
			my $file_name=$file->path;
			my $dir_path = join("/", @split_path);
			my $previous_file = $dir_path.$file_name;
			return $previous_file;
		}
	}
    ## Failed to get previous file
    return;
}

=head2 ::is_exists_log_file

Check given logfile exists or not.If exists return 1,
otherwise, return the path of new logfile.

=over

=item Arguments

B<File path > => I<log file path>.

I<Required.>
Log file path returned by get_worm_log_file_path.

B<vol object> => I<volume object>

I<Required.>
NOTE log volume object.

=item Examples

=back

	$out = is_exists_log_file($file_path, $log_vol);
	
=cut
sub is_exists_log_file($$) {

	my ($FilePath , $vol_obj ) =  
			validate_pos(@_,
			{ type => SCALAR },
		   	{ type => HASHREF }
		);

	my $filer_conn = $vol_obj->Filer->conn;
	my $vol_name = $vol_obj->name;
	my $basename;
	if ( $FilePath =~ /priv_delete/ ) {
		$basename = "priv_delete";
	} elsif ( $FilePath =~ /system_log/) {
		$basename = "system_log";
	}
	my $CheckFile = get_worm_log_file_path( $filer_conn, $vol_name, $basename );

	if ( $FilePath ne $CheckFile ) {
		return $CheckFile;
	}
	return 1;
}

=head2 ::set_distinct_telnet

Set the option telnet.distinct.enable. to ( on | off ).

=over

=item Arguments

B<filer_conn> => I<Connect onject>

I<Required.>
Filer Connect object to execute commands on the filer.

B<Value>  => I <on | off >

I<Required.>
Enable or disable telnet distinct session.

=item Examples

=back

	$out = set_distinct_telnet($filer_conn , "on");

=cut

sub set_distinct_telnet($$) {
	
    my ($filer_conn , $value) = @_;
    my $status;

    ## Verify the options telnet.distinct.enable
    my $newout = sayEOF($filer_conn, "options telnet.distinct.enable");
    my @split = split(/ /, $newout);
    $status = pop(@split);
    while ( $status ne "on" && $status ne "off" ) {
 	$status = pop(@split);
    }
    if ( $status eq $value ) {
	NOTE::logcomment(" telnet.distinct.enable is already set to $value\n");
    } else {
	my $output = sayEOF($filer_conn , "options telnet.distinct.enable $value");
	## Verify the options telnet.distinct.enable
	$newout = sayEOF($filer_conn, "options telnet.distinct.enable");
	@split = split(/ /, $newout);
        $status = pop(@split);
	while( ($status ne "on" ) && ( $status ne "off" )) {
            $status = pop(@split);
	}
	if ( $status eq $value ) { 
	    NOTE::logcomment(" Sucessfully set telnet.distinct.enable to $value\n");
	}else {
	    NOTE::logresult('FATAL',"Failed to Set telnet.distinct.enable to $value\n");
	}
    }
}

=head2 ::login_to_console

Login to console with given username and password.

=over

=item Arguments

B<filer> => I<Hostspec of the filer>

I<Required.>
Hostspec of the filer.

B<username> => I<Username>

I<Required.>
Username to login to filer.

B<password> => I<password>

I<Required.>
Password for the given user to login to the filer.

=item Examples

=back
	
	$console_conn = ( $filer_name , $username , $password );

=cut	

sub login_to_console($$$){

    NOTE::logcomment("In function login_to_console\n");
    my ($filer, $username , $password ) = @_;
    my $login_user;
    my $console_conn;

    my $host = host($filer);
    eval { $console_conn = connect("#console; username=$username; password=$password", "-login", "-timeout", 30);};
    if ( $@ ) {
	logcomment("Failed to get console connection");
	return undef;
    }
    if ( $host->hosttype() =~ /simulator/){ 
        ## From IC.2 onwards, relogin to console is not required.
        ## Since relogin to console on simulator gave lot of problems
        ## Avoiding the relogin part to simulator.
        ## Privdel script cannot be run on IC.1 release.
	return $console_conn;
    } else {	
        ## Relogin to console as a different user
        say($console_conn , "\cD" , '-prompt',"Data ONTAP",'-timeout',10);
        say($console_conn ,  $username , '-prompt', "Password" , '-timeout',10);
        say($console_conn , $password , '-prompt', ">",'-timeout',10);
    }
    $login_user = sayEOF($console_conn, "useradmin whoami");
    $login_user = sayEOF($console_conn, "useradmin whoami");
    if ( $username =~ /$login_user/ ) { logcomment("Sucessfully logged in to console as $login_user"); }
    else { die "Failed to login to console as $login_user "; }
    return $console_conn;
}

## Internal function which checks the SSL configuration
sub _check_ssl_setup($$) {
    my ($host, $filer_conn) = @_;

    my $ret_ssl_configured = 0;

    # Check whether or not SSL is already set up.
    my ( undef, undef, $output ) = say( $filer_conn, "options httpd.admin.ssl.enable on" );
    if (  $output =~ /SSL is not configured./ )
    {
        eval { _ssl_setup($host); };
        if ($@) {
            my $msg = "Error configuring SSL";
            die "$msg $@";
        }
        else {
            $ret_ssl_configured = 1;
        }
    }
    else {
        $ret_ssl_configured = 1;
    }
    return $ret_ssl_configured;
}

## Internal function to setup SSL on the filer
sub _ssl_setup($) {

    my ($host) = @_;
   #my (%args) = validate( 
   #     @_,
   #     {
   #         host         => { type => SCALAR | HASHREF },
         #   root_pass    => { type => SCALAR, optional => 1 },
          #  vfiler       => { type => SCALAR, optional => 1 },
          #  country      => { type => SCALAR, optional => 1 },
           # state        => { type => SCALAR, optional => 1 },
           # locality     => { type => SCALAR, optional => 1 },
           # company      => { type => SCALAR, optional => 1 },
           # division     => { type => SCALAR, optional => 1 },
           # common_name  => { type => SCALAR, optional => 1 },
           # admin_email  => { type => SCALAR, optional => 1 },
           # expire_days  => { type => SCALAR, optional => 1 },
           # key_length   => { type => SCALAR, optional => 1 },
           # re_run_setup => { type => SCALAR, optional => 1 },
   #     }
   # );

    # Create the telnet connection for the filer and fill up the
    # root_pass in our args if we don't know it yet.
    my $filer_host = make_host( $host );
    my $telnet_conn = make_conn( $host, "#telnet" );
    die($@) if ($@);    # remove when the return eval is removed
    if ( defined($main::CMDTIMEOUT) ) {
        $telnet_conn->timeout($main::CMDTIMEOUT);
    }

    my $ctrl_C = sprintf( "%c", 003 );    # [Ctrl]+C char;
                                          # Create a ReplyTable.
    my %reply_table = (
        "SSL Setup has already been done before. Do you want to proceed?" =>
          "yes",
        "Country Name \\(2 letter code\\)"              => "",
        "State or Province Name \\(full name\\)"        => "",
        "Locality Name \\(city, town, etc\\.\\)"        => "",
        "Organization Name \\(company\\)"               => "",
        "Organization Unit Name \\(division\\)"         => "",
        "Common Name \\(fully qualified domain name\\)" => "",
        "Administrator email"                           => "",
        "Days until expires"                            => "",
        "Key length \\(bits\\)"                         => "",
    );

    # wait for "secureadmin setup ssl" prompts in %ReplyTable, and act on
    # them, until the prompt coming back is one of the built-in
    # say prompts (i.e. the filer's normal prompt, showing that
    # the setup is done)

    say( $telnet_conn, " secureadmin setup ssl", '-blind' );
    my ( $tries, $prompt, $reply, $after );
    for ( $tries = 50 ; $tries ; $tries-- ) {
        ( undef, undef, $after, $prompt ) =
          say( $telnet_conn, "", '-prompt', [ keys %reply_table ] );

        last if ( matchstrings( $prompt, [ $telnet_conn->prompts() ] ) );
        if ( !matchstrings( $prompt, [ keys %reply_table ] ) ) {
            say( $telnet_conn, "\r", '-nocr', '-blind' )
              ;    # if we don't have a match, just
                   # take the default
        }
        foreach my $reg_exp ( keys %reply_table ) {
            next unless ( matchstrings( $prompt, $reg_exp ) );
            if ( $tries <= 1 ) {

                # if got stuck, try ^C to terminate the setup.
                $reply = $ctrl_C;
            }
            else {
                $reply = $reply_table{$reg_exp};
            }
            $reply .= "\r";
            say( $telnet_conn, $reply, '-nocr', '-blind' );
        }
    }

    # If we are out of tries, something is wrong.
    if ( !$tries ) {
        die "ssl_setup(): too many prompts during \"secureadmin setup ssl\"\n";
    }

    # Check whether the setup worked.
    my ( undef, undef, $output ) =
      say( $telnet_conn, "options httpd.admin.ssl.enable on" );
    disconnect($telnet_conn);
    if ( $output =~ /SSL is not configured./ ) {
        die "ssl_setup() unsuccessful : \"$output\"\n";
    }
    return 1;
                                                                
}

## Internal function to verify SSH configuration.
sub _check_ssh_setup($$$) {

    my ($filer, $client, $filer_conn ) = @_;
	
    my $ret_ssh_configured = 0;

    # Check whether or not SSH is already set up.

    my ( undef , undef, $output ) = 
	say($filer_conn, "options ssh.enable on");
    if ( $output =~ /SSH Server is not configured./ ) {
	eval{ _ssh_setup($filer, $client, $filer_conn ) ;};
	if ($@) {
            my $msg = "Error configuring SSH";
            logresult('FAIL', "$msg : $@");
        }
        else {
            $ret_ssh_configured = 1;
        }
    }
    else {
        $ret_ssh_configured = 1;
    }
    return $ret_ssh_configured;
}

## Internal function to setup SSH on the filer
sub _ssh_setup($$$) {

    my ($filer, $client, $conn ) = @_;

    NOTE::logcomment("In _ssh_setup\n");
    my $ssh_config;

    $ssh_config = OntapSSH::SecureAdmin->get_config($filer,$client);
    $ssh_config->compare_config() ? logcomment("PASS: Match!") : logresult("FATAL", "Mismatch! SSH setup");

    $ssh_config->set_ssh1_hostkeysize();
    $ssh_config->set_ssh1_serverkeysize();
    $ssh_config->set_ssh2_hostkeysize();
    $ssh_config->update_filer_settings();

    # Check whether the setup worked.
    my ( undef , undef, $output ) =
        say($conn, "options ssh.enable.on");
    if ( $output =~ /SSH Server is not configured./ ) {
	die "ssl_setup() unsuccessful : \"$output\"\n";
    }
    return 1;
}

=head2 ::fill_the_volume

100% fill the given volume.

=over

=item Arguments

B<vol_obj> => I<Volume object>

I<Required.>
NOTE volume object.

=item Examples

=back

	@files = fill_the_volume($vol_obj);

=cut
sub fill_the_volume($) {

    my ( $vol_obj ) =  validate_pos( @_,
                                    { type => HASHREF},
	);
    my @big_files;
    ## Changed the library used to fill the volume.	
    my $vol = $vol_obj;
    $vol->fill(percent=>100, proto=>"local");
    logcomment("filled volume $vol->name");

    ## Fill always creates single file named __FILL__
    ## Get this file and return the path.
    my $file = $vol->fillfile( proto => 'local' );
    push(@big_files ,$file);

    ## Run sync to ensure the all new writes have been written to disk
    $vol_obj->Filer->say("priv set -q diag; sync; sync");

    my %disk_free = df ( $vol_obj->Filer->conn, $vol_obj->name , options => "-k" );
    my $avail = $disk_free{volume}->{avail};
    my $tries = 30;
    my @file_size = split(/KB/, $avail);
    my $vol_size = $file_size[0];
    $file_size[0] = $file_size[0]."k";
    my $size = $file_size[0];

       my $vol_name = $vol_obj->name;
       #Below call not used, as '-b' option requires diag mode
       #"df -b" is executed using 'say' with priv set as diag 
       #my $df_b = df ( $vol_obj->Filer->conn, $vol_obj->name , options => "-b" );
       my $df_b = $vol_obj->Filer->say("priv set -q diag; df -b $vol_name");        

       NOTE::logcomment("OutPut of #> df -b <slc_Worm_vol>  $df_b");
       
       #Removing the white spaces
	$df_b =~ s/\s+/ /g;
	$df_b =~ s/^\s//;
	$df_b =~ s/\s$//;
        my @words_in_op = split / / , $df_b ;
	my $blocks_free = $words_in_op[10];
        NOTE::logcomment("Blocks Free found from  df -b <slc_Worm_vol>  $blocks_free");
        ## Fill the remaining space left in the volume with KB.

    if ( $blocks_free > 64 ){
      while ( ( $vol_size > 0 ) && $tries  != 0 ) {
	my $file_name = "/vol/".$vol_obj->name."/BigFile_".$avail.$tries;
	write_file(filer=>$vol_obj->Filer->conn, path=>$file_name , size=>$size);
	## Run sync to ensure the all new writes have been written to disk
	$vol_obj->Filer->say("priv set -q diag; sync; sync");        
        my %disk_free1 = df ( $vol_obj->Filer->conn, $vol_obj->name , options => "-k" );
        $avail = $disk_free1{volume}->{avail};
        my @file_size = split(/KB/, $avail);
        $vol_size = $file_size[0];
        NOTE::logcomment("remaining vol size = $vol_size\n");
        $size = $file_size[0]*1024;
        push(@big_files , $file_name);
        $tries--;
     }
    }else {
      for (my $x = 0 ; $x < $blocks_free ; $x++ ){
        my $file_name = "/vol/".$vol_obj->name."/$x";
	my $size = "1k" ;
	write_file(filer=>$vol_obj->Filer->conn, path=>$file_name , size=>$size);
	$vol_obj->Filer->say("priv set -q diag; sync; sync");
        push(@big_files , $file_name);
      }
    }
    
    $vol_obj->Filer->say("priv set -q diag; sync; sync");        
    return @big_files;
}


=head2 ::delete_the_file

Delete the given file from the volume.

=over

=item Auguments.

B<filer_conn> => I<filer_conn>

I<Required.>
Filer connection to execute commands on the filer.

B<file_path> => I<File_Path>

I<Required.>
File path to be deleted.

=item Examples

=back

	delete_the_file($filer_conn , @non_worm_files );

=cut
sub delete_the_file($@) {
    my ( $filer_conn, @files) = @_;
	
    foreach my $file_path ( @files ) {
        my $cmd = "priv set diag; rm $file_path";
	my $text = sayEOF($filer_conn , $cmd);
    }
}

=head2 :: verify_privdel_execution

Verify the privileged execution.

=over

=item Arguments

B<log_output> => I<WORM log output>

I<Required.>.
Log string returned by log reader.

B<Files> => I<files>

I<Required.>.
Privileged deleted files details.

B<input_hash> => I<input hash>

I<Required>
Inputs to verify the privileged delete operation.
user     : Name of the user who deleted the file.
client   : linux client used to perform operation.
conn_type: Network protocol used ( SSH | ZAPI )
volume   :   Name of the volume
result   : Expected result (PASS | FAIL )

B<Vol obj > => I<Volume object>

I<Required.>
NOTE Enterprise volume object.

=back

=cut

sub verify_privdel_execution($$$$) {

    my ( $log_output , $files, $input_hash, $vol_obj ) = @_;

    NOTE::logcomment(" In verify_privdel_execution function call\n");
    ## @Step Verify File has deleted on the file system.
    my $is_file_deleted = check_files_got_deleted($vol_obj, $files);
    ($is_file_deleted == 0 ) ? logresult("PASS", " All files got deleted"):logresult('FAIL', "Files Exists Even After Privilege Delete" ); 
     NOTE::logcomment("Verifying PrivDel file log entry\n");
    ## @Step Get the pre and post log OutPut in a hash.
    my @log_entries  = _split_logentries($log_output);
    ## @Step Update input Hash with File specific details
    my @input_entries = _update_input_hash($files, $input_hash, $vol_obj->Filer->conn);

    my $index = 0;
    while ( $index <= $#log_entries && $index <= $#input_entries) {
 	my $log_entry = $log_entries[$index];
	my $input_values = $input_entries[$index];
	_verify_privdel_log($log_entry , $input_values);
	$index++;	
    }
    NOTE::logcomment("Exit from verify_privdel_execution execution\n");
}

=head2 ::verify_system_log

Verify the entries of system log files.

=over

=item Arguments

B<log output> => I<system log output>

I<Required.>
System log output returned by LogReader.

B<input hash> => I<input hash>

I<Required.>
Inputs to verify the system log operation.
old_log_volume : Required to verify snaplock_log_volume_change operation
new_log_volume : Required to verify snaplock_log_volume_change operation
whodidit       : Required to verify snaplock_user_added_deleted operation
username       : Required to verify snaplock_user_added_deleted operation
action         : Required to verify snaplock_user_added_deleted operation

=back 

=cut
sub verify_system_log($$){

    my ( $log_output, $input_hash) = @_;

    my $log_entry = _XML_string_to_hash($log_output);
    

    if ( $input_hash->{verify} eq "log_volume_change" ) {
        ## Verify log_volume_change result
        if ( defined $log_entry->{LogEntry}->{snaplock_log_volume_changed_1}) {

            NOTE::logcomment("PASS : Log entry snaplock_log_volume_changed_1 is available\n");
            ( $log_entry->{LogEntry}->{snaplock_log_volume_changed_1}->{old_log_volume} eq $input_hash->{old_log_volume} ) ? NOTE::logcomment("PASS : Old Logvolume name matched\n") : NOTE::logresult("FAIL", "Old log volume name did not match\n");

            ( $log_entry->{LogEntry}->{snaplock_log_volume_changed_1}->{new_log_volume} eq $input_hash->{new_log_volume} ) ? NOTE::logcomment("PASS : New Log volume name matched\n" ) : NOTE::logresult("FAIL", "New log volume name did not match\n");
        }else {
            NOTE::logresult("FAIL", "Log entry snaplock_log_volume_changed_1 is not available\n");
        }
    }elsif ( $input_hash->{verify} eq "user_added_deleted" ) {
        ## Verify snaplock_user_added_deleted result.
        if ( defined $log_entry->{LogEntry}->{snaplock_user_added_deleted_1} ) {
            NOTE::logcomment("PASS : Log entry snaplock_user_added_deleted_1 is available\n");
            ( $log_entry->{LogEntry}->{snaplock_user_added_deleted_1}->{whodidit} eq $input_hash->{whodidit} ) ? NOTE::logcomment("PASS : whodidit matched\n") : NOTE::logresult("PASS", "whodidit did not match\n");

            ( $log_entry->{LogEntry}->{snaplock_user_added_deleted_1}->{username} eq $input_hash->{username} ) ?  NOTE::logcomment("PASS : username matched\n") : NOTE::logresult("PASS", "username did not match\n");

            ( $log_entry->{LogEntry}->{snaplock_user_added_deleted_1}->{action} eq $input_hash->{action} ) ? NOTE::logcomment("PASS : Action matched\n") : NOTE::logresult("FAIL", "action did not matched\n");
        } else {
            NOTE::logresult("FAIL", "Log entry snaplock_user_added_deleted_1 is  not available\n");
        }
    }
        
}

=head2 :: verify_state_transition_log

Verify Log entries of privdel state transition operation.

=over

=item Arguments.

B<Logoutput> => I<logoutput>

I<Required.>
Log string returned by LogReader.

B<input_hash> => I<input_hash>

I<Required.>
Inputs to verify the state transition log operation.
volume       : NOTE volume object.
changed_from : Privdel state changed from ( Previous state)
changed to   : Privdel state changed to ( Current/New state )
result       : Expected result PASS|FAIL 

=back 

=cut    

sub verify_state_transition_log($$) {

    my ( $log_output, $input_hash ) = @_;
    ##  Get the pre and post log OutPut in a hash.
    my @log_entries  = _split_logentries($log_output);

    foreach my $logentry ( @log_entries ) {
	my $prelog = $logentry->{pre};
	my $postlog = $logentry->{post};

	## Verify the Pre PrivDel Log Entry.
	NOTE::logcomment("Verifying Pre PrivDel State Transition Log Entry");
	( $prelog->{LogEntry}->{snaplock_pre_option_privileged_delete_1}->{volume} eq $input_hash->{volume}->name ) ? NOTE::logcomment("PASS : Volume Name Matched\n") : NOTE::logresult('FAIL', "Volume Name did not match\n");
	my $changed_from = $input_hash->{from_state};
	( $prelog->{LogEntry}->{snaplock_pre_option_privileged_delete_1}->{changed_from}  =~ /$changed_from/ ) ? NOTE::logcomment("PASS : Previous State  Matched\n" ) : NOTE::logresult('FAIL', "Previous state did not match\n" );
        my $changed_to = $input_hash->{to_state};
	( $prelog->{LogEntry}->{snaplock_pre_option_privileged_delete_1}->{changed_to} =~ /$changed_to/ ) ? NOTE::logcomment("PASS : Changed State Matched\n"): logresult('FAIL', "Changed state did not match\n");

	## Verify the Post Privdel Log Entry.
	logcomment("Verifying Post Privdel  State Transition Log Entry\n");
	
	( $postlog->{LogEntry}->{snaplock_post_option_privileged_delete_1}->{volume} eq $input_hash->{volume}->name ) ? logcomment("PASS : Volume Name Matched\n") : logresult('FAIL', "Volume Name did not match\n");
	( $postlog->{LogEntry}->{snaplock_post_option_privileged_delete_1}->{changed_from} =~ /$changed_from/  ) ? logcomment("PASS : Previous State Matched\n" ) : logresult('FAIL', "Previous state did not match\n" );
	( $postlog->{LogEntry}->{snaplock_post_option_privileged_delete_1}->{changed_to}  =~ /$changed_to/ ) ? logcomment("PASS : Changed State Matched\n"): logresult('FAIL', "Changed state did not match\n");
	if ( $input_hash->{result} eq "PASS" ) {
 	    ( $postlog->{LogEntry}->{snaplock_post_option_privileged_delete_1}->{result} eq PRIVDEL_PASS ) ? logcomment("PASS : Result Entry Matched\n") : logresult('FAIL', "Result Did not match\n");
	} else {
	    ( $postlog->{LogEntry}->{snaplock_post_option_privileged_delete_1}->{result} eq  PRIVDEL_STATE_FAIL ) ? logcomment("PASS : Result Entry Matched\n") : logresult('FAIL', "Result Did not match\n");
	}
    }
}
	
## Internal function to verify the privdel delete execution.
sub _verify_privdel_log($$) {
    my ( $logout, $inputval ) = @_;

    ##@Step Verify Pre PrivDel Log Entry.
    _verify_pre_privdel_log($logout->{pre} , $inputval);
    ##@Step Verify Post PrivDel Log Entry.
    _verify_post_privdel_log($logout->{post}, $inputval);
    ##@Step Compare Pre and Post Log Entries.
    _compare_pre_post_log($logout->{pre} ,  $logout->{post} );
}

### Internal function to update the input hash which is privided later to verify
### priv delete execution.
sub _update_input_hash ($$$) {

    my ( $files , $input_hash, $filer_conn ) = @_;

    logcomment("In _update_input_hash \n");
    ## Get the ip address of the client
    my $client = $input_hash->{client};
    print("client = $client\n");
    my $host = make_host($client);
    my $client_ip_address = $host->default_ip();
    if ( $input_hash->{conn_type} eq "CONSOLE" ) {
	$input_hash->{client} = "0.0.0.0";
    } else {
        $input_hash->{client} = $client_ip_address;		
    }
    $input_hash->{filer_conn} = $filer_conn;
    my @input_array=();
    foreach my $file ( @$files ) {
	my $file_path =$file->{file_path};
	my $file_expires = $file->{file_atime};
        if ( $file_expires =~ /infinite/ ) {
		$file_expires = 'never';
	}
	logcomment(" Update_hash FilePath = $file_path FileExpires = $file_expires\n");
	$input_hash->{file_pathname} = $file_path;
	$input_hash->{file_expires} = $file_expires;
        $input_hash->{file_obj} = $file;

	my %temp_hash = %$input_hash;
	push(@input_array, \%temp_hash); 
    }
    logcomment("Exitting  _update_input_hash\n");
    return @input_array;
}

## Internal function which actually verifies the every entry of pre privdel log.
sub _verify_pre_privdel_log($$) {

    my ( $log_entry, $input_values ) = @_;
    logcomment("Verifying PrivDel file pre log entry\n");
    print Dumper($log_entry);
    my $date_string = $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_expires};
    my $time_stamp ;
    if ( $date_string =~ /^(\w+)\s*((\w+)\s*(\d+))\s*((..):(..):(..))\s*(\w+)\s*(\d+)/){
	$time_stamp = Date::Manip::UnixDate($date_string , "%s");
    } else {
	$time_stamp = $date_string ;
    }

    logresult('FAIL', "snaplock_pre_privileged_delete_1->client_transport type did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_transport}, $input_values->{conn_type}) if ($log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_transport} ne $input_values->{conn_type} );
    logresult('FAIL', "snaplock_pre_privileged_delete_1->client_ip_address did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_ip_address}, $input_values->{client}) if ( $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_ip_address} ne $input_values->{client} );
    logresult('FAIL',"snaplock_pre_privileged_delete_1->client_user did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_user} , $input_values->{user} ) if ( $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_user} ne $input_values->{user} );
    logresult('FAIL', "snaplock_pre_privileged_delete_1->file_pathname did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_pathname}, $input_values->{file_pathname}) if ( $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_pathname} ne $input_values->{file_pathname} );
    logresult('FAIL', "snaplock_pre_privileged_delete_1->file_expires did not match: LogEntry = [%s] , InputEntry = [%s]", $time_stamp , $input_values->{file_expires})	if ( $time_stamp ne $input_values->{file_expires} );

    ####  Verify file fingerprint info###
    my $ver = get_wafl_version($input_values->{filer_conn}->{hostp});
    if ( $ver >= WV_ROLLINGROCK_MIN ) {
	my $scc_temp = $log_entry->{LogEntry}->{SCCdate};
        my @split_scc = split(/ / , $scc_temp );
        my $scc_string = $split_scc[0].$split_scc[1].$split_scc[2];
	my $scc = Date::Manip::UnixDate($scc_string , "%s");
	$log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_fingerprint}->{fingerprint}->{system_cc} = $scc ;
        my $key = $input_values->{file_pathname};
        $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_fingerprint}->{fingerprint}->{volume_cc} = $log_vol_vcc_comp{$key}{data_vol};
        ## Compare the VCC of log volume in audit log record with VCC taken after privdel.
        my $vcc_log_entry = $log_entry->{LogEntry}->{VCCdate};
        my @split_vcc = split(/ / , $vcc_log_entry );
        my $vcc_string = $split_vcc[0].$split_vcc[1].$split_vcc[2];
        my $log_vcc = Date::Manip::UnixDate($vcc_string , "%s");
        my $cmd_vcc = $log_vol_vcc_comp{$key}{log_vol};
        
        my $vcc_diff = $log_vcc - $cmd_vcc;
        if ( abs ($vcc_diff)  == 0 or abs ($vcc_diff) < 30 ) {
  	    logcomment("PASS: VCC on Audit log recored matched with VCC taken from 'snaplock clock status' command");
        } else {
	    logresult('FAIL', "VCC on Audit log recored doesnot match with VCC taken from 'snaplock clock status' command");
	}
   }
    verify_fingerprint($input_values->{filer_conn} , $log_entry->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_fingerprint}->{fingerprint} , $input_values->{file_obj} );
    logcomment("Verified Privdel file pre log entry\n");

}

## Calculate the audit log file priv_delete expiry date.
sub calculate_record_expiry($$) {

    my ($logentry, $pre_post_type) = @_;
	
    ## Formula to calculate the record expiry date.
    ## record_expiry = file_expiry + ( vcc_of_log_vol - vcc_of_data_vol )
    ## Get the Vcc of log volume in seconds.
    my $vcc_temp = $logentry->{LogEntry}->{VCCdate};
    my @split_vcc = split(/ / , $vcc_temp );
    my $vcc_string = $split_vcc[0].$split_vcc[1].$split_vcc[2];
    my $vcc_of_log_vol_in_secs = Date::Manip::UnixDate($vcc_string , "%s" );
    ## Get the expiry of deleted file in seconds.
    my $file_expiry_in_secs = Date::Manip::UnixDate($logentry->{LogEntry}->{$pre_post_type}->{file_expires} , "%s");
    ## Get the VCC of SLE data volume in seconds.
    my $vcc_of_data_vol_in_secs = $logentry->{LogEntry}->{$pre_post_type}->{file_fingerprint}->{fingerprint}->{"fingerprint-info"}->{"snaplock-volume-compliance-clock"};
    my $record_expiry = ( $file_expiry_in_secs + ( $vcc_of_log_vol_in_secs - $vcc_of_data_vol_in_secs ));
    return $record_expiry; 

}

## Internal function which actually verifies the every entry of post privdel log.
sub _verify_post_privdel_log($$) {

    my ( $log_entry, $input_values ) = @_;
    logcomment("Verifying PrivDel file post log entry\n");
    print Dumper($log_entry);
    my $date_string = $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_expires};
    my $time_stamp; 
    if ( $date_string =~ /^(\w+)\s*((\w+)\s*(\d+))\s*((..):(..):(..))\s*(\w+)\s*(\d+)/ ){
	$time_stamp = Date::Manip::UnixDate($date_string , "%s");
    } else {
	$time_stamp = $date_string ;
    }

    logresult('FAIL', "snaplock_post_privileged_delete_1->client_transport type did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_transport}, $input_values->{conn_type})
    if ( $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_transport} ne $input_values->{conn_type} );

    logresult('FAIL', "snaplock_post_privileged_delete_1->client_ip_address did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_ip_address}, $input_values->{client})
    if ( $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_ip_address} ne $input_values->{client} );

    logresult('FAIL',"snaplock_post_privileged_delete_1->client_user did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_user} , $input_values->{user} )
    if ( $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_user} ne $input_values->{user} );

    logresult('FAIL', "snaplock_post_privileged_delete_1->file_pathname did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_pathname}, $input_values->{file_pathname})
    if ( $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_pathname} ne $input_values->{file_pathname} );

    logresult('FAIL', "snaplock_post_privileged_delete_1->file_expires did not match: LogEntry = [%s] , InputEntry = [%s]", $time_stamp , $input_values->{file_expires})
    if ( $time_stamp ne $input_values->{file_expires} );
	
    ####  Verify file fingerprint info###
    my $ver = get_wafl_version($input_values->{filer_conn}->{hostp});
    if ( $ver >= WV_ROLLINGROCK_MIN ) {
        my $scc_temp = $log_entry->{LogEntry}->{SCCdate};
        my @split_scc = split(/ / , $scc_temp );
        my $scc_string = $split_scc[0].$split_scc[1].$split_scc[2];
	my $scc = Date::Manip::UnixDate($scc_string , "%s");
        $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_fingerprint}->{fingerprint}{system_cc} = $scc ;
        my $key = $input_values->{file_pathname};
        $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_fingerprint}->{fingerprint}{volume_cc} = $log_vol_vcc_comp{$key}{data_vol};
        ## Compare the VCC of log volume in audit log record with VCC taken after privdel.
        my $vcc_log_entry = $log_entry->{LogEntry}->{VCCdate};
        my @split_vcc = split(/ / , $vcc_log_entry );
        my $vcc_string = $split_vcc[0].$split_vcc[1].$split_vcc[2];
        my $log_vcc = Date::Manip::UnixDate($vcc_string , "%s");
        my $cmd_vcc = $log_vol_vcc_comp{$key}{log_vol};
        my $vcc_diff = $log_vcc - $cmd_vcc;
        if ( abs($vcc_diff) == 0 or abs ($vcc_diff) < 30 ) {
            logcomment("PASS: VCC on Audit log recored matched with VCC taken from 'snaplock clock status' command");
        } else {
            logresult('FAIL', "VCC on Audit log recored doesnot match with VCC taken from 'snaplock clock status' command");
        }

   }

    verify_fingerprint($input_values->{filer_conn} , $log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_fingerprint}->{fingerprint} , $input_values->{file_obj} );
    logresult('FAIL', "snaplock_post_privileged_delete_1->result did not match:")
    if ($log_entry->{LogEntry}->{snaplock_post_privileged_delete_1}->{result} ne PRIVDEL_PASS );
    logcomment("Verified Privdel file post log entry\n");
}	

## Internal function which compares pre and post privdel log entries.
sub _compare_pre_post_log($$){

    my ( $prelog , $postlog )= @_; 

    logresult('FAIL',"snaplock_pre_privileged_delete_1->client_transport did not match with snaplock_post_privileged_delete_1->client_transport type Pre[%s] , Post[%s]" , $prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_transport}, $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_transport})
    if ($prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_transport} ne $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_transport});
    logresult('FAIL', "snaplock_pre_privileged_delete_1->client_ip_address did not match with snaplock_post_privileged_delete_1->client_ip_address Pre[%s], Post[%s]", $prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_ip_address} , $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_ip_address} )
    if ( $prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_ip_address} ne $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_ip_address} );

    logresult('FAIL',"snaplock_pre_privileged_delete_1->client_user did not match with snaplock_post_privileged_delete_1->client_user Pre[%s] Post[%s]", $prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_user}, $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_user})
    if ( $prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{client_user} ne $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{client_user});

    logresult('FAIL', "snaplock_pre_privileged_delete_1->file_pathname did not match with snaplock_post_privileged_delete_1->file_pathname, Pre[%s] Post[%s]", $prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_pathname} , $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_pathname})
    if ( $prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_pathname} ne $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_pathname});

    logresult('FAIL', "snaplock_pre_privileged_delete_1->file_expires did not match with snaplock_post_privileged_delete_1->file_expires , Pre[%s] , POST[%s]", $prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_expires},  $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_expires} )
    if ( $prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_expires} ne $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_expires});
     ######### Compare Pre and Post Fingerprint OutPut #######################
     my @fields_did_not_match = compare_fingerprint($prelog->{LogEntry}->{snaplock_pre_privileged_delete_1}->{file_fingerprint} , $postlog->{LogEntry}->{snaplock_post_privileged_delete_1}->{file_fingerprint} );
    
     if ( $#fields_did_not_match > 0 ) {
	logresult("FAIL" , "File fingerprint did not match\n");
		print Dumper(@fields_did_not_match);
     } else {
	logcomment("PASS : File fingerprint output matched\n");
     }

}

=head2 ::set_volume_security_style

Set the security style for given volume.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required>
Filer connection to execute commands on the filer.

B<worm_log_vol> => I<worm log volume>

I<Required>
NOTE Vol object of the worm log volume.

B<style> => I<Security Style>

I<Required>
Security style to be set on the volume unix/ntfs/mixed

=item Examples

=back

	set_volume_security_style($conn , $volume , "ntfs");

=cut 
sub set_volume_security_style( $$$ ) {

    my ( $conn , $volume , $style ) =  validate_pos( @_,
                                    {type => SCALAR|HASHREF},
                                    {type => HASHREF},
                                    {type => SCALAR, default=>'unix',regex=>qr/^(ntfs|unix|mixed)$/},
                            );
    
    ## Change the security style of the worm log volume.
    my $cmd = "qtree security /vol/".$volume->name." ".$style;
    my $out = sayEOF($conn , $cmd );

    ### Verify the security style set on the volume
    $cmd = "qtree security /vol/".$volume->name;
    $out = sayEOF($conn , $cmd );
    if ( $out !~ /$style/ ) {
	logcomment("FAIL:Failed to set security style to $style\n");    
        return 0;
    } else {
	logcomment("PASS: Set the security style of the volume to $style\n");
        return $style;
    }

}

=head2 ::get_ntfs_file_details

Get the files from the volume , whose security style was set to NTFS

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required>
Filer connection to execute commands on the filer.

B<files> => I<Populated files>

I<Required>
Ref to array of files returned from populate_worm_files_over_cifs().

=item Examples

=back

	@file_details = get_ntfs_file_details($filer_conn $files );

=cut

sub get_ntfs_file_details ($$) {

    my ( $filer_conn ,  $files ) = @_;

    my $fp_options = create_fp_options();
    my @file_details;

    ## Execute fingerprint and get required file inode details.
    foreach my $file ( @$files ) {
        my $output = execute_fingerprint($filer_conn, $file , $fp_options->{'option_xml_fp'});
        my %fp = (
                file_path   => $output->{"fingerprint-info"}->{"fingerprint-input-path"},
                file_mtime  => $output->{"fingerprint-info"}->{"metadata-files"}->{"file-fingerprint-info"}->{"modified-time"},
                file_ctime  => $output->{"fingerprint-info"}->{"metadata-files"}->{"file-fingerprint-info"}->{"changed-time"},
                file_atime  => $output->{"fingerprint-info"}->{"metadata-files"}->{"file-fingerprint-info"}->{"retention-time"},
                volume_fsid => $output->{"fingerprint-info"}->{"metadata-files"}->{"file-fingerprint-info"}->{"fsid"},
                volume_uuid => $output->{"fingerprint-info"}->{"volume-uuid"},
        );
        push(@file_details , \%fp );
    }
    return @file_details ;
}

=head2 ::populate_worm_files_over_cifs

Population of worm files over CIFS

=over

=item Arguments

B<prefix> => I<Cifs mount path>

I<Required>
Mount path returned from cifs_map_share.

B<vol> => I<Note Volume Object>

I<Required.>
Volume object on which files has to be created.

B<client> => I<CIFS Client>

I<Required.>
Windows client name.

B<file_count> => I<file count>

I<Required.>
Number of files to be created.

=item Examples

=back

	my @files = populate_worm_files_over_cifs($prefix,$vol, $client , $file_count);

=cut

sub populate_worm_files_over_cifs($$$$) {
	
    my ( $prefix,$vol, $client , $file_count ) = @_;
    my $content = "Populate worm data on SLE volume.";
    my $file_name = "ntfs_file";
    my @files;
    my $i;
    for ( $i = 0 ; $i <= $file_count ; $i++ ) {
          my $file = $file_name.$i;
          my $file_path = "/vol/".$vol->name."/".$file;
          nfs_cifs_mkfile("$prefix$file", protocol => 'cifs',
                        client => $client, content=>$content)
                or die($@);
          NOTE::logcomment("Commit the file to worm");
          nfs_cifs_rdonly("$prefix$file", protocol => 'cifs',
                        client => $client)
                or die fail_result("Was unable to commit the file to worm: $@");
          push(@files, $file_path);
    }

    return @files;

}

=head2 ::get_snaplock_scan_file

Get the snaplock scan log file.

=over

=item Arguments

B<vol_obj> => I<Volume object>

I<Required>
NOTE volume object.

=item Examples

=back

        my @files = get_snaplock_scan_file($vol_obj);

=cut

sub get_snaplock_scan_file($) {

    my ($vol_obj) = @_;
    my $vol_name = $vol_obj->name;
    my $conn = $vol_obj->Filer->conn;
    my $cmd = "priv set -q diag;ls /vol/".$vol_name;
    my $ls_of_volume;
    eval {
    	$ls_of_volume =  sayEOF($conn , $cmd );
    };
    if ( $@ or $ls_of_volume =~ /No such file or directory/ ) {
	croak "Failed to get snaplock scan file";
    }
    my @files = split(/\n/ , $ls_of_volume);
    my @scan_files;
    foreach my $file (@files) {
	if ( $file =~ /snaplock_scan/ ) {
		push (@scan_files , $file);
        }
    }
    return @scan_files;
}

=head2 ::read_snaplock_scan_file

Read the given file and return the file content as a string.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required>
Filer connection to execute commands on the filer.

B<file> => I<Path>

I<Required>
Absolute path of the snaplock scan file to be read.

=item Examples

=back
       sample file path "/vol/vl11998_6c61_46cc9354/snaplock_scan.20110125_062747_GMT-present"
       $string  = read_snaplock_scan_file($filer_conn $file_path );

=cut

sub read_snaplock_scan_file($$) {
    my ($conn, $filepath ) = @_;
    
    my $out = sayEOF($conn , "priv set -q diag; rdfile $filepath");
    my @data = split(/(\n)/, $out);
    my $string;
    for (@data) {
	if ( $_ =~ /SnapLockLogFile/i || $_ =~ /LogFileFooter LogVersion/i || $_ =~ /LogFileHeader LogVersion/ ) {
                        ## Do not add this line to string
                } else {
                        $string .= "$_";
                }
    }
    return $string;
}

=head2 ::insert_nfs_verifier_flag

Insert NFS verifier flags to all files in the volume/qtree.

=over

=item Arguments

B<filer_conn> => I<filer connection>

I<Required>
Filer connection to execute commands on the filer.

B<obj> => I<Volume/Qtree object>

I<Required>
Note Volume or Qtree object.

B<is_qtree>  => I<0|1>

I<Optional>
If qtree object is specified, this argument is required, otherwise not.

=item Examples

=back

       insert_nfs_verifier_flag($conn, $obj , $is_qtree);

=cut

sub insert_nfs_verifier_flag($$;$) {
    my ( $conn, $obj , $is_qtree) = @_;
    my $disk_flag = NFS_VERIFIER_FLAG ;

    my ( $vol_name , @files, @file_names, @inode_nums ,@file_details, $qtree_name , $lines_flag );
    @files = $obj->nfs->children;
    foreach my $file (@files) {
        if ( $file->name !~ /snapshot/ ) {
            if ( ! $file->is_dir ) {
                push(@file_names , $file->name);
		push(@inode_nums, $file->inode);
            }
        }
    }

    if ( $is_qtree ) {
        $qtree_name = $obj->name;
        $vol_name = $obj->Volume->name;
    } else {
	$vol_name = $obj->name;
    }

    for( my $i=0 ; $i <= $#file_names ;$i++ ) { 
        my $file_name = $file_names[$i];
        my $complete_path;
        if ( $is_qtree ) {
	    $complete_path = $qtree_name.'/'.$file_name;
        } else {
	    $complete_path = $file_name;
        }

        $lines_flag.="inode_flags_insert inum=path:/$complete_path flags=$disk_flag\n";
    }      

    wreck ( $conn, $vol_name , "type" => 'flex' , "inject_script"=> $lines_flag ) or logresult("fatal" , "Wreck failed: $@" );
        
    ## Online the volume after injecting corruption.
    eval {
    	vol_online($conn, $vol_name);
    };
    if ( $@ ) {
	die (" Failed to online the volume after injecting the corruption to files using wreck");
    }
    ## Verify that NFS verifier falg is properly injected on all the files.
    my $count = SnapLock::Helper::verify_nfs_verifier_flag_on_inode($conn, $vol_name , \@inode_nums );    
    if ( $count != scalar @inode_nums ) { die "NFS flag is not injected on all the inodes";}

}


=head2 ::verify_snaplock_scan_log_file

Verify the pre and post fields in the log file.

=over

=item Arguments

B<log_output> => I<File content>

I<Required>.
Content of the log file.

B<files> => I<file list>

I<Required>
List of the files populated for the particular test.

=item Examples

=back

       $result  = verify_snaplock_scan_log_file($filer_conn $file_path );

=cut

sub verify_snaplock_scan_log_file($$$) {

    my ( $log_output , $files , $conn) = @_;;
    my @input_entries;
    my %final_hash;
    ## @Step Get the pre and post log OutPut in a hash.
    my @log_entries  = _split_logentries($log_output);
    my @input_files = @$files;
    my $count = 0;
    foreach my $file ( @input_files ) {
	foreach my $logentry ( @log_entries ) {
	 if ( $logentry->{pre}->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{file_pathname} eq $file->{file_path} ) {
             my (%input_hash, %comp_hash);
             $input_hash{file_pathname} = $file->{file_path}; 
             $input_hash{incorrect_retention_time} = $file->{file_atime};
             $input_hash{file_expires} = "never";
             $input_hash{record_expires} = "never";
             $input_hash{file_obj} = $file;
             $input_hash{filer_conn} = $conn;
             $comp_hash{log_entry} = $logentry;
             $comp_hash{input_value} = \%input_hash;
             push ( @input_entries , \%comp_hash );
             $count++;
          }
      }
    }
    logcomment (" Count = $count , number of input files = $#input_files \n");
    if ( ( $count - 1 )  ne $#input_files ) {
	logcomment("Number of entries in the log file did not match with number of infected files");
	return 0;
    }
    my $log_result;
    foreach my $entry ( @input_entries ) {
        my $log_entry = $entry->{log_entry};
        my $input_value = $entry->{input_value};
        my $pre_status = _verify_pre_snaplock_scan_log($log_entry->{pre}, $input_value);
        my $post_status = _verify_post_snaplock_scan_log($log_entry->{post}, $input_value);
        if ( $pre_status > 0 or $post_status > 0 ) {
	    $log_result++;
        }
    }
   return $log_result; 
}


## Internal function to verify the snaplock scan pre log entries.
## Here all the entries of snaplock scan pre log will be verified.
## Require: expected output hash
## Require: Actual pre log entry hash.

sub _verify_pre_snaplock_scan_log($$) {

    my ( $log_entry , $input_values) = @_;

    logcomment("Verifying snaplock scan file pre log entry\n");
    print Dumper($log_entry);
    my $date_string = $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{incorrect_retention_time};
    my $time_stamp = Date::Manip::UnixDate($date_string , "%s");
    my $count=0;

    logcomment("Checking pre_scanner_fix file_pathname field");
    if ( $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{file_pathname} ne $input_values->{file_pathname} ) {
	logresult('FAIL',"snaplock_pre_scanner_fix_retention_1->>file_pathname did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{file_pathname}, $input_values->{file_pathname});
	$count++;
    }
    logcomment("Checking pre_scanner_fix incorrect_retention_time field");
    if ( $time_stamp ne $input_values->{incorrect_retention_time} ) {
	logresult('FAIL',"snaplock_pre_scanner_fix_retention_1->incorrect_retention_time did not match: LogEntry = [%s] , $time_stamp , $input_values->{incorrect_retention_time}" );
	$count++;
    }
    logcomment("Checking pre_scanner_fix record_expires field");
    if ( $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{record_expires} ne  $input_values->{record_expires} ) {
	logresult('FAIL', "snaplock_pre_scanner_fix_retention_1->record_expiry did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{record_expires}, $input_values->{record_expires});
        $count++;
    }
    my $scc_temp = $log_entry->{LogEntry}->{SCCdate};
    my @split_scc = split(/ / , $scc_temp );
    my $scc_string = $split_scc[0].$split_scc[1].$split_scc[2];
    my $scc = Date::Manip::UnixDate($scc_string , "%s");
    $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{file_fingerprint}->{fingerprint}{system_cc} = $scc ;
    my $vcc_log_entry = $log_entry->{LogEntry}->{VCCdate};
    my @split_vcc = split(/ / , $vcc_log_entry );
    my $vcc_string = $split_vcc[0].$split_vcc[1].$split_vcc[2];
    my $log_vcc = Date::Manip::UnixDate($vcc_string , "%s");
    $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{file_fingerprint}->{fingerprint}{volume_cc} = $log_vcc;

    if ( $time_stamp eq $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{file_fingerprint}->{fingerprint}->{"fingerprint-info"}->{"metadata-files"}->{"file-fingerprint-info"}->{"retention-time"} ) {
	logresult('FAIL' , "incorrect retention time field and fingerprint retention time fileld should not be same in snaplock_pre_scanner_fix_retention");
	$count++;
    } else {
	logcomment("As expected incorrect retention time field and fingerprint retention time filelds are not same");
        ## Change the atime field in file object.
        $input_values->{file_obj}{file_atime} = $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{file_fingerprint}->{fingerprint}->{"fingerprint-info"}->{"metadata-files"}->{"file-fingerprint-info"}->{"creation-time"};
    }
        
   verify_fingerprint($input_values->{filer_conn} , $log_entry->{LogEntry}->{snaplock_pre_scanner_fix_retention_1}->{file_fingerprint}->{fingerprint} , $input_values->{file_obj} );
 
   logcomment("Verified scanner_fix_retention pre log entry\n");
   return $count;
}

## Internal function to verify the snaplock scan post log entries.
## All the entries of snaplock scan post log will be verified.
## Require: Expected output hash
## Require: Snaplock scan post log entry hash.
sub _verify_post_snaplock_scan_log($$) {

    my ( $log_entry, $input_values ) = @_;
    my $count = 0;
    logcomment("Checking post_scanner_fix file_pathname field");
    print Dumper($log_entry);
    if ( $log_entry->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{file_pathname} ne $input_values->{file_pathname} ) {
	logresult('FAIL',"snaplock_post_scanner_fix_retention_1->>file_pathname did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{file_pathname}, $input_values->{file_pathname});
	$count++;
    }
    logcomment("Checking post_scanner file_expires field");
    if ( $log_entry->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{file_expires} ne $input_values->{file_expires} ) {
	logresult('FAIL',"snaplock_post_scanner_fix_retention_1->file_pathname did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{file_expires}, $input_values->{file_expires}) ;
        $count++;
    }

    logcomment("Checking post_scanner_fix record_expiry field");
    if ( $log_entry->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{record_expires} ne $input_values->{record_expires} ) {
	logresult('FAIL', "snaplock_post_scanner_fix_retention_1->record_expiry did not match: LogEntry = [%s] , InputEntry = [%s]", $log_entry->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{record_expires}, $input_values->{record_expires});
	$count++;
    }
    my $scc_temp = $log_entry->{LogEntry}->{SCCdate};
    my @split_scc = split(/ / , $scc_temp );
    my $scc_string = $split_scc[0].$split_scc[1].$split_scc[2];
    my $scc = Date::Manip::UnixDate($scc_string , "%s");
    $log_entry->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{file_fingerprint}->{fingerprint}{system_cc} = $scc ;
    my $vcc_log_entry = $log_entry->{LogEntry}->{VCCdate};
    my @split_vcc = split(/ / , $vcc_log_entry );
    my $vcc_string = $split_vcc[0].$split_vcc[1].$split_vcc[2];
    my $log_vcc = Date::Manip::UnixDate($vcc_string , "%s");
    $log_entry->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{file_fingerprint}->{fingerprint}{volume_cc} = $log_vcc;

    ## Post log entry for retention time has to be infinite.
    $input_values->{file_obj}{file_atime} = "infinite";

   verify_fingerprint($input_values->{filer_conn} , $log_entry->{LogEntry}->{snaplock_post_scanner_fix_retention_1}->{file_fingerprint}->{fingerprint} , $input_values->{file_obj} );

   logcomment("Verified scanner_fix_retention post log entry\n");
   return $count;

}


=head1 SEE ALSO

L<C<NOTE>|engineering::ontap_tests::main::docs::lib-NOTE-pm/>

L<C<NOTE::Aggregate>|engineering::ontap_tests::main::docs::lib-NOTE-Aggregate-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>

L<C<NOTE::Qtree>|engineering::ontap_tests::main::docs::lib-NOTE-Qtree-pm/>

L<C<NOTE::Volume>|engineering::ontap_tests::main::docs::lib-NOTE-Volume-pm/>

=head1 STATUS

Review


$Date: 2009/02/06 $

=head1 AUTHOR

  dl-nb-sl-qa@netapp.com

=cut

1;


