# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/SnapMirror.pm#1 $

## @summary SnapMirror utilities library -- action subroutines
##
## @author dl-dpgqa-automation@netapp.com
## @status Shared
## @pod here

=head1 NAME

SnapMirror - functions for using SnapMirror

=head1 SYNOPSIS

 use SnapMirror;

 snapmirror_abort(host=>$dst_hostrec, dst=>$dst);

 snapmirror_break(host=>$dst_hostrec, dstpath=>$dstpath);

 my $obj = snapmirror_check(host=>dst_hostrec, dst=>$dst);

 snapmirror_fix(host=>dst_hostrec, dst=>$dst);

 snapmirror_check_abort(host=>$dst_hostrec, dstpath=>$dstpath);

 snapmirror_convert(host=>$src_hostrec, dstpath=>$dstpath);

 my $array_ref = snapmirror_destinations(host=>$src_hostrec, 
                                         srcpath=>$srcpath);

 my $array_ref = snapmirror_destinations_list(host=>$src_hostrec);

 snapmirror_disable(host=>$src_hostrec);

 snapmirror_enable(host=>$src_hostrec);

 my $obj = snapmirror_initialize(host=>$dst_hostrec, dst=>$dst, src=>$src);

 my $obj = snapmirror_migrate(host=>$src_hostrec, src=>$src, dst=>$dst);

 my $obj = snapmirror_quiesce(host=>$dst_hostrec, dstpath=>$dstpath);

 snapmirror_release(host=>$src_hostrec, srcpath=>$srcpath, dst=>$dst);

 snapmirror_resume(host=>$dst_hostrec, dstpath=>$dstpath);

 snapmirror_retrieve(host=>$hostrec, vol=>$vol, restorefile=>$file_or_drive);

 my $obj = snapmirror_resync(host=>$dst_hostrec, src=>$src, dst=>$dst);

 my $hash_ref = snapmirror_status(host=>$dst_hostrec, dst=>$dst);

 my $hash_ref = snapmirror_status_vol(host=>$dst_hostrec, vol=>$vol);

 my @array_ref = snapmirror_status_list(host=>$dst_hostrec);

 my $tape_ref = snapmirror_store(host=>$hostrec, vol=>$vol, dumpfile=>$file_or_drive,
			     geometry=>$dst_disk_geometry, size=>$limit_in_megs);

 my $status = snapmirror_tape_status(host=>$hostrec, vol=>$vol,
			type=>$action, msg=>$expected_msg,
			wait=>$flag, timeout=>$timeout);

 my $obj = snapmirror_update(host=>$dst_hostrec, src=>$src, dst=>$dst);

Please note that parameters...

1) 'src' and 'dst' are of the form <filer>:<volume> or 
<filer>:</vol/volname/qtreename>
2) 'srcpath' and 'dstpath' are of the form <volume> or </vol/volname/qtreename>

The main difference between (1) and (2) is that the string does not include 
the '<filer>:' prefix.

To trap exceptions thrown by the functions in this library, enclose the 
function call inside an eval block and check $@ for the error message.

 eval {
   snapmirror_enable(host=>$src_hostrec);
 };
 die $@ if ($@);

=head1 DESCRIPTION

This package defines basic routines for using SnapMirror.

Please consider looking at the convenient functions in SnapMirror::Helper.pm
first to see if those functions meet your needs. It may save you time on
writing the code to use SnapMirror in your script.

=head1 ROUTINES

=cut

package SnapMirror;

use strict;
use warnings;
use Carp qw(croak);
use Params::Validate qw(validate validate_pos BOOLEAN SCALAR OBJECT HASHREF);
use base qw(Exporter);

# for debugging purposes. pls do NOT use it in testscripts.
# it may be removed in the future.
our $Use_Rsh = 0;

BEGIN {
    Params::Validate::validation_options(
         on_fail => sub { my @msgs = @_; chomp(@msgs);
			  Carp::confess(Tharn::script_result(@msgs)) }
    );

    our @EXPORT_OK = qw(
        snapmirror_abort
        snapmirror_break
        snapmirror_check
        snapmirror_fix
        snapmirror_check_abort
        snapmirror_convert
        snapmirror_destinations
        snapmirror_destinations_list
        snapmirror_disable
        snapmirror_enable
        snapmirror_initialize
        snapmirror_migrate
        snapmirror_quiesce
        snapmirror_release
        snapmirror_resume
        snapmirror_resync
        snapmirror_retrieve
        snapmirror_status
        snapmirror_status_list
        snapmirror_status_vol
        snapmirror_store
	snapmirror_tape_status
        snapmirror_update
    );
}


use ATI::ati_ontapi qw(ati_ontapi_create_context 
                       ati_ontapi_call_api
                       ATI_SUCCESS 
                       ATI_FAILURE
                       ATI_ONTAPI_TRANSPORT_DEFAULT
                       ATI_ONTAPI_DEBUG_NONE
                       ATI_ONTAPI_DEBUG_TERSE);

use Data::Dumper qw( Dumper );
use Date::Manip qw(ParseDateString UnixDate);
use FSR::rpl_cmds qw(:all :error_codes);
use Filer::Version qw(get_wafl_version 
                      WV_FULLSAIL_MIN
                      WV_TSINGTAO_MIN
                      WV_GORDONBIERSCH_MIN
                      WV_GORDONBIERSCH_MAX);
use Module::Common qw(validate_mixed);
use SayWrap qw(sayEOF);
use SnapMirror::Helper qw(sm_get_config _sm_extract_subtest_err _sm_extract_volname);
use Tharn;
use Utils qw(make_conn);
use Volume qw(vol_status vol_restrict vol_online);
use LogReader::Messages;
use Tapeinfo qw( getTapeStatus );

autoflush STDOUT 1;
# cmd prefix for snapmirror
my $sm = "snapmirror";
my ($log_rdr, $reuse_log) = (undef, 0); # (log reader)

=head2 snapmirror_status

 my $hash_ref = snapmirror_status( host=>$host, dst=>$dst, [ src=>$src ] );

Retrieves snapmirror status for the specified source/destination
relationship.

It returns a hash reference containing the status on success and dies
with the error string on failure.

Refer to this page (snapmirror-status-info) for the hash keys:

http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/mainN/web/snapmirror/index.html#snapmirror-status-info

Note: The hash reference returned by snapmirror_status() will not contain
the status of 'snapmirror check' if the verification is running.

=over

=item host

(Required) : Hostrec Object of filer.

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item src

(Optional) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=back

=cut

sub snapmirror_status {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
			         dst =>  { type => SCALAR },
			         src =>  { type => SCALAR,
					   optional => 1  }, } );

    _hostname_present($args{dst});
    _hostname_present($args{src}) if (exists $args{src});

    if (my $context = _ontapi_support($args{host})) {
	# zapi
	my $sm_status;
	my $out = ();

	# not using snapmirror-get-status with 'location' input.
	# because if there are multiple SM relationships from
	# one filer, specifying one end of the relationship
	# can potentially return a list of all relationships.
	$sm_status = snapmirror_status_list(host=>$args{host});

	foreach my $status (@{$sm_status}) {
	    # find the relationship
	    if ($status->{'destination-location'} eq $args{dst}) {
		# if source if provided, do a more stringent check
		if (exists $args{src} &&
		    $status->{'source-location'} eq $args{src}) {
		    $out = $status;
		    last;
		} else {
		    $out = $status;
		    last;
		}
	    }
	}
	_burt248930_work_around($out)if defined $out;
	return $out if defined $out;

    } else {
	# rsh
	my $conn = make_conn($args{host});
	my $rc;

	if (exists $args{src}) {
	    $rc = rpl_details_internal($sm, $conn, $args{dst}, $args{src});
	} else {
	    $rc = rpl_details_internal($sm, $conn, $args{dst});
	}
	croak $@ if ($@);
#	return (_transform_data_struct($rc)) if defined $rc;
	if ( defined $rc ) {
		my $out = ();
		$out = _transform_data_struct($rc);
		_burt248930_work_around($out);
		return $out;
	}
		
    }
    croak "SnapMirror status for non-existent relationship: "
	.(exists $args{src} ? $args{src}." -> " : "").$args{dst};
}

my %bad_cnt; # Hash to prevent multiple error due to burt 248930 work around below

sub _burt248930_work_around {
    my ($stat) = @_;

    return; # BURT 248930 fixed
    $bad_cnt{$stat->{'base-snapshot'}} = 0 unless exists $bad_cnt{$stat->{'base-snapshot'}};
    return if $bad_cnt{$stat->{'base-snapshot'}};
    return unless exists $stat->{'current-transfer-error'};
    if ($stat->{'status'} =~ /^idle$/i && $stat->{'current-transfer-type'} !~ /-/ ) {
	logcomment("SM: Ignore false error for ".$stat->{'source-location'}
		." (work-around for BURT248930)");
	$stat->{status} = "transferring";
	delete $stat->{'current-transfer-error'};
	$bad_cnt{$stat->{'base-snapshot'}}++;
    } else {
	logcomment("SM: Real Failure ".$stat->{'source-location'}." (not BURT248930)");
    }
}

=head2 snapmirror_status_vol

 my $hash_ref = snapmirror_status_vol( host=>$host, vol=>$vol );

Retrieves snapmirror status values for a given volume. Prior to the 
execution of the command, this function will check that a jump-ahead is 
not taking place. The check will always take place regardless of the 
type of the relationship, ie. qtree, asynchronous, or synchronous.

It returns a hash reference containing the volume status on success and dies
with the error string on failure.

Refer to this page (snapmirror-get-volume-status) for the hash keys:

http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/mainN/web/snapmirror/index.html#snapmirror-get-volume-status

=over

=item host

(Required) : Hostrec Object of filer.

=item vol

(Required) : Name of the volume.

=back

=cut

sub snapmirror_status_vol {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
			         vol =>  { type => SCALAR }, } );

    # specific to RSM only.
    # make sure volume is not frozen before we move on
    _poll_for_thawed_vol($args{host}, $args{vol});

    if (my $context = _ontapi_support($args{host})) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-get-volume-status',
				 args => [volume => $args{vol}]);

	# $out[1] contains the results (if any)
	# see _invoke_ontapi() for descriptions on the other elements.
	return $out[1];

    } else {
	# rsh
	my $conn = make_conn($args{host});
	my $dst = $args{host}->hostname() . ":" . $args{vol};
	my $rc = rpl_details_internal($sm, $conn, $dst);
	croak $@ if ($@);

	my $s = ();
	my $struct = _transform_data_struct($rc);

	# NOTE: if there's a QSM transfer inside the volume, $args{vol}, 
	# rpl_details_internal() will return an emptry $rc data structure.

	if (ref($rc) eq "HASH") {
	    # create data structure to return to caller. data structure has
	    # to be consistent to Zapi's.
	    $s->{'is-broken-destination'} = 
	      ($struct->{'state'} =~ /broken-off/) ? 'true' : 'false';

	    $s->{'is-transfer-in-progress'} = 
	      ($struct->{'transfer-progress'} > 0) ? 'true' : 'false';

	    # get the 2nd element from the 'split' array
	    my $src_vol = ( split(/:/, $struct->{'source-location'}) )[1];
	    $s->{'is-source'} = ($src_vol eq $args{vol}) ? 'true' : 'false';

	    my $dst_vol = ( split(/:/, $struct->{'destination-location'}) )[1];
	    $s->{'is-destination'} = 
	      ($dst_vol eq $args{vol}) ? 'true' : 'false';

	} else {
	    # ZAPI returns all keys with value as 'false' if there's no
	    # VSM activity on the volume
	    $s->{'is-broken-destination'}   = 'false';
	    $s->{'is-transfer-in-progress'} = 'false';
	    $s->{'is-source'}               = 'false';
	    $s->{'is-destination'}          = 'false';
	}
	
	return $s;
    }
}

=head2 snapmirror_status_list

 my $hash_ref = snapmirror_status_list( host=>$host );

 foreach my $mirror ( @{$array_ref} ) {
     print "$mirror->{source-location} " .
           "$mirror->{destination-location}";
 }

Retrieves snapmirror status for *all*  source/destination relationships.

It returns an array reference containing the status on success and dies
with the error string on failure.

Refer to this page (snapmirror-status-info) for the hash keys:

http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/mainN/web/snapmirror/index.html#snapmirror-status-info

Note: The hash reference returned by snapmirror_status() will not contain
the status of 'snapmirror check' if the verification is running.

=over

=item host

(Required) : Hostrec Object of filer.

=back

=cut

sub snapmirror_status_list {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
						    \&_validate_hostrec } 
					 },
			       } );
    if (my $context = _ontapi_support($args{host})) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-get-status');

	foreach my $s (@{$out[1]->{'snapmirror-status'}}) {
		_burt248930_work_around($s);
	}
	return $out[1]->{'snapmirror-status'};
    } else {
	my @stats;
	my $conn = make_conn($args{host});
	my @rc = rpl_details_internal($sm, $conn);
	# Convert array of statuses in zapi equivalent
	if ( @rc ) {
		foreach my $s (@rc) {
#			push @stats, _transform_data_struct($s);
			my $out = _transform_data_struct($s);
			_burt248930_work_around($out);
			push @stats, $out;
		}
	}
	return \@stats;
    }
}

=head2 snapmirror_destinations

 my $array_ref = snapmirror_destinations( host=>$srchost, srcpath=>$srcpath );

Retrieves snapmirror destinations (for a given source).

It returns an array reference containing the destinations on success and dies
with the error string on failure.

Refer to this page (destination-info) for the hash keys:

http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/mainN/web/snapmirror/index.html#destination-info

=over

=item host

(Required) : Hostrec Object of source filer.

=item srcpath

(Required) : Source volume or path for the SnapMirror relationship, 
i.e. <volname> or </vol/volname/qtreename>

=back

=cut

sub snapmirror_destinations {
    my (%args) = validate( @_, { host => { type => OBJECT, 
 					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         srcpath => { type => SCALAR } } );

   if (my $context = _ontapi_support($args{host})) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-list-destinations',
				 args => ['source-location' => $args{srcpath}]
				);
	
	my $dest = $out[1]->{'destinations'};
        $dest = [] if ref $dest ne 'ARRAY';

	return $dest;

    } else {
	my $conn = make_conn($args{host});
	my @rc = rpl_destinations_internal($sm, $conn, $args{srcpath});
	croak $@ if ($@);

	# @rc data structure
	#
	# @rc is an array containing the list of destinations for a given
	# source, ie. @rc = ('destination1', 'destination2', ...);
	#
	# need to convert the above to the Zapi equivalent which is
	# an array reference of...
	#
	# [ { 'source-location'      => $source,
	#     'destination-location' => $destination,
	#     'source-snapshot'      => $source-snapshot },
	#   { .... },
	# ]
	my @out = ();
	foreach my $dst (@rc) {
	    my $s = ();
	    $s->{'destination-location'} = $dst;

	    # make sure it's of this form to be consistent with
	    # Zapi's
	    my $srcpath = $args{host}->hostname() . ":" . $args{srcpath};
	    $s->{'source-location'} = $srcpath;

	    # now get the name of source snapshot
	    my $src_snapshot = rpl_dst_snap_internal($sm, $conn,
						     $args{srcpath},
						     $dst);
	    croak $@ if ($@);
	    $s->{'source-snapshot'} = $src_snapshot;
	    push(@out, $s);
	}
	
	return \@out;
    }
}


=head2 snapmirror_destinations_list

 my $array_ref = snapmirror_destinations_list( host=>$srchost ); 

Retrieves all snapmirror destinations on the source host.

It returns an array reference containing the destinations on success and dies
with the error string on failure.

Refer to this page (destination-info) for the hash keys:

http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/mainN/web/snapmirror/index.html#destination-info

=over

=item host

(Required) : Hostrec Object of filer.

=back

=cut

sub snapmirror_destinations_list {
    my (%args) = validate( @_, { host => { type => OBJECT, 
 					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } 
					 },
				 } );

   if (my $context = _ontapi_support($args{host})) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-list-destinations');

	return $out[1]->{'destinations'};

    } else {
	my $conn = make_conn($args{host});
	my $rc = rpl_destinations_internal($sm, $conn);
	croak $@ if ($@);

	# $rc data structure
	#
	# source is key to hash. the key points to an array
	# containing the destinations.
	#   'source' => [ 'destination1', 'destination2', ... ]
	#
	# need to convert the above to the Zapi equivalent which is
	# an array reference of...
	#
	# [ { 'source-location'      => $source,
	#     'destination-location' => $destination,
	#     'source-snapshot'      => $source-snapshot },
	#   { .... },
	# ]
	my @out = ();
	foreach my $srcpath (keys %{$rc}) {
	    foreach my $dst ( @{$rc->{$srcpath}} ) {
		my $s = ();
		if ( $dst =~ /\[dump\]/ ) {
	    		logresult('WARN', 'Work-around for BURT 269590: Skipping invalid dump dst $dst.');
			next;
		}
		$s->{'destination-location'} = $dst;

		# make sure it's of this form to be consistent with
		# Zapi's
		if ($srcpath !~ /\S+:\S+/) {
		    $srcpath = $args{host}->hostname() . ":" . $srcpath;
		}
		$s->{'source-location'} = $srcpath;

		# now get the name of source snapshot
		my $src_snapshot = rpl_dst_snap_internal($sm, $conn,
							 $srcpath,
							 $dst);
		croak $@ if ($@);
		$s->{'source-snapshot'} = $src_snapshot;
		push(@out, $s);
	    }
	}
	
	return \@out;
    }
}


=head2 snapmirror_quiesce

 my $subtest_obj = snapmirror_quiesce( host=>$host, dstpath=>$dstpath, 
		     [ background=>0|1 ], 
                     [ timeout=>$timeout ] );

 # if background option is set to 1
 $subtest_obj->wait_finish();

Quiesces the given SnapMirror relationship. Prior to the execution of the
quiesce command, this function will check that a jump-ahead is not taking
place. The check will always take place regardless of the type of the
relationship, ie. qtree, asynchronous, or synchronous.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item dstpath

(Required) : Destination volume or path for the SnapMirror relationship, 
i.e. <volname> or </vol/volname/qtreename>

=item timeout

(Optional) : Timeout for commands to complete. Defaults is 300 seconds.

=item background

(Optional) : If set, returns while transfer is still running in the
background.  The result of the transfer needs to be extracted manually
later. Defaults to 0.

=back

=cut

sub snapmirror_quiesce {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
			         dstpath =>  { type => SCALAR },
   			         background => { type => BOOLEAN,  
					       optional => 1,
					       default => 0 },
				 timeout => { type => SCALAR,
					      optional => 1,
					      default => 300,
					      regex => qr/^\d+$/ } } );


    my $obj;
    my $dstpath = $args{dstpath};

    # specific to RSM only.
    # make sure volume is not frozen before we move on
    my $vol = _sm_extract_volname($args{dstpath});
    _poll_for_thawed_vol($args{host}, $vol);

    my ($cnt, $pause) = (10, 2);
    my $jump_ahead_error = "volume is temporarily in a transitional state";
    if (my $context = _ontapi_support($args{host})) {
	# zapi
	# may need to implement retry as for rsh - to be determine later
	$obj = subtest( sub {	my $retry = $cnt;
				my @out;
				do {
					if ( $retry < $cnt ) {
						sleep $pause;
						logcomment("retry after hit jump-ahead");
					}
					eval {
						@out = _invoke_ontapi(context => $context,
							api => 'snapmirror-quiesce',
							args => [ 
								'destination-location' => $dstpath,
								] )
					};
				} while $@ =~ qr/$jump_ahead_error/si && $retry-- > 0;
				die "$@" if $@;
				return @out;
				},
			'-bg', '-onexit', sub{}, # don't pass results to parent
			);
    } else {
	my $cmd = "$sm quiesce $dstpath";

	my $conn = make_conn($args{host});
	$obj = subtest(
	    sub {
                my ($text,$retry) = (undef,$cnt);
		do {
		    sleep $pause if( $retry < $cnt );	# don't sleep unless we loop at least ones
		    $text = sayEOF($conn,$cmd);
		}
		while( $text =~ qr/$jump_ahead_error/si && $retry-- > 0 );

	        die "no output from 'snapmirror quiesce'" unless $text;
	        die "$text" unless ($text =~ /Successfully quiesced/i);
	    },
	    '-bg', '-onexit', 
	    sub{}, # don't pass results to parent
	);
    }

    if (!$args{background}) {
	$obj->wait_finish();
	_sm_extract_subtest_err($obj);
    }
    
    return $obj;
}


=head2 snapmirror_resume

 snapmirror_resume( host=>$host, dst=>$dst);

Resumes transfer to destination.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item dstpath

(Required) : Destination volume or path for the SnapMirror relationship, 
i.e. <volname> or </vol/volname/qtreename>

=back

=cut

sub snapmirror_resume {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
			         dstpath => { type => SCALAR } } );

   if (my $context = _ontapi_support($args{host})) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-resume',
				 args => [ 
                                   'destination-location' => $args{dstpath}
                                 ] );
    } else {
	# rsh
	my $cmd = "$sm resume " . $args{dstpath};
	my $conn = make_conn($args{host});
	my $text   = sayEOF($conn, $cmd,
			    TIMEOUT => $args{timeout});
	croak $text unless ($text =~ /Successfully resumed/);
    }
}



=head2 snapmirror_break

 snapmirror_break( host=>$host, dstpath=>$dstpath,
                   [ hard=>0|1 ] );

Breaks the SnapMirror relationship and make the given SnapMirror 
destination writable.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item dstpath

(Required) : Destination volume or path for the SnapMirror relationship, 
i.e. <volname> or </vol/volname/qtreename>

=item hard

(Optional) : If set to 1, does a "hard" break, which will break the
relationship even if the destination is a qtree that has not been quiesced.
Default is 0.

=back

=cut

sub snapmirror_break {
    my (%args) = validate( @_, { host => { type => OBJECT, 
 					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         dstpath => { type => SCALAR },
			         hard => { type => BOOLEAN,
					   optional => 1,
					   default  => 0  } } );

   my $context;

   # There is no support for hard break in zapi because it is a 
   # privileged command. In that case, use CLI interface.
   if (!$args{hard} && ($context = _ontapi_support($args{host}))) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-break',
				 args => [ 
                                   'destination-location' => $args{dstpath},
                                 ] );
    } else {
	# rsh
	my $conn = make_conn($args{host});
	rpl_break_internal($sm, $conn, $args{dstpath}, _uppercase_keys(%args));
	croak $@ if ($@);
    }
}


=head2 snapmirror_abort

 snapmirror_abort( host=>$host, dst=>$dst, 
                   [ hard=>0|1 ] );

Aborts an active SnapMirror transfer.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item hard

(Optional) : If set to 1, does a "hard" abort, such that the transfer is 
not restartable. Default is 0.

=back

=cut

sub snapmirror_abort {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
			         dst =>  { type => SCALAR },
			         hard => { type => BOOLEAN,
					   optional => 1,
					   default  => 0  } } );
    _hostname_present($args{dst});

   if (my $context = _ontapi_support($args{host})) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-abort',
				 args => [ 
                                   'destination-location' => $args{dst},
				   'clear-checkpoint' => $args{hard},
                                 ] );
    } else {
	# rsh
	my $conn = make_conn($args{host});
	rpl_abort_internal($sm, $conn, $args{dst}, _uppercase_keys(%args));
	croak $@ if ($@);
    }
}


=head2 snapmirror_release

 snapmirror_release( host=>$srchost, srcpath=>$srcpath, dst=>$dst );

Releases the source-side resources being held for a particular
SnapMirror destination.  

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item srcpath

(Required) : Source volume or path for the SnapMirror relationship, 
i.e. <volname> or </vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=back

=cut


sub snapmirror_release {
    my (%args) = validate( @_, { host => { type => OBJECT,
 					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         srcpath => { type => SCALAR },
			         dst     => { type => SCALAR } } );
    _hostname_present($args{dst});
   if (my $context = _ontapi_support($args{host})) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-release',
				 args => [
                                     'source-location' => $args{srcpath},
				     'destination-location' => $args{dst} ]);
    } else {
	# rsh
	my $conn = make_conn($args{host});
	rpl_release_internal($sm, $conn, $args{srcpath}, $args{dst});
	croak $@ if ($@);
    }
}

=head2 snapmirror_convert

 snapmirror_convert( host=>$srchost, dstpath=>$dstpath );

Convert the named replica destination into a SnapMirror destination.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item dstpath

(Required) : Source qtree path for the SnapMirror relationship, 
i.e. </vol/volname/qtreename>

=back

=cut


sub snapmirror_convert {
    my (%args) = validate( @_, { host => { type => OBJECT,
 					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         dstpath => { type => SCALAR } } );

    my $conn = make_conn($args{host});
    rpl_convert_internal($sm, $conn, $args{dstpath});
    croak $@ if ($@);
}


=head2 snapmirror_migrate

 my $subtest_obj = snapmirror_migrate( host=>$srchost, src=>$src, dst=>$dst, 
		     [ background=>0|1 ], 
                     [ throttle=>$throttle ], [ timeout=>$timeout ] );

 # if background option is set to 1
 $subtest_obj->wait_finish();

Performs a snapmirror migrate of the specified snapmirror relationship
on the source filer.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of source filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item throttle

(Optional) : Kbs rate to throttle transfer with.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=item background

(Optional) : If set, returns while transfer is still running in the
background.  The result of the transfer needs to be extracted manually
later. Defaults to 0.

=back

=cut

# XXX/tapl make this function async so that we can test nfs client access
# while migration is taking place
sub snapmirror_migrate {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
			         src =>  { type => SCALAR },
			         dst =>  { type => SCALAR },
   			         throttle => { type => SCALAR,  
					       optional => 1 },
   			         background => { type => BOOLEAN,  
					       optional => 1,
					       default => 0 },
				 timeout => { type => SCALAR,
					      optional => 1,
					      default => 300,
					      regex => qr/^\d+$/ } } );

    _hostname_present($args{src}, $args{dst});
    # construct the command line
    my $cmd_throttle = (exists $args{throttle}) 
      ? "-k " . $args{throttle} : "";
    my $cmd = "$sm migrate -f $cmd_throttle " . $args{src} .
      " " . $args{dst};

    # try doing the migrate
    my $conn   = make_conn($args{host});
    my $obj = subtest(
	sub {
	    my $text = sayEOF($conn, $cmd, TIMEOUT => $args{timeout});
	    die $text unless ($text =~ /migration complete/);
	},
	'-bg', '-onexit', sub{},
    );
    if ( !$args{background} ) {
	$obj->wait_finish();
	_sm_extract_subtest_err($obj);
    }
    
    return $obj;
}


=head2 snapmirror_resync

 my $subtest_obj = snapmirror_resync( host=>$srchost, dst=>$dst, 
                                      [ src=>$src ], 
                      [ snapshot=>$snapshot_name ],
                      [create_snapshot=>$snapshot_name],
                                      [ throttle=>$throttle ], 
                                      [ background=>0|1 ], 
                                      [ parallel=>0|1 ], 
                                      [ timeout=>$timeout ] );

 # if background option is set to 1
 $subtest_obj->wait_finish();

Performs a snapmirror migrate of the specified snapmirror relationship
on the source filer.

If the background option is set and the resync has successfully
started, returns a subtest object for an active subtest which is waiting 
for the transfer to complete. Important: The caller is responsible for 
checking for any exception thrown (if any) by the subtest.

On failure, dies with an error string.

=over

=item host

(Required) : Hostrec Object of source filer.

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item src

(Optional) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item snapshot

(Optional) : Name of specific snapshot on src to resync from, this parameter
is for use with QSM only, it does not work for VSM or RSM.

=item create_snapshot

(Optional) : Name of specific snapshot to create on destination, this parameter
is for use with QSM only, it does not work for VSM or RSM.

=item throttle

(Optional) : Kbs rate to throttle transfer with.

=item background

(Optional) : If set, returns while transfer is still running in the
background.  The result of the transfer needs to be extracted manually
later. Defaults to 0.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=back

=cut

sub snapmirror_resync {
    my (%args) = validate( @_, { host => { type => OBJECT,
                       callbacks => {
                           'Hostrec Object of filer' =>
                            \&_validate_hostrec } },
                     dst =>  { type => SCALAR },
                     src =>  { type => SCALAR,
                       optional => 1 },
                 snapshot => { type => SCALAR,
                           optional => 1 },
                                 create_snapshot => { type => SCALAR,
                                                      optional => 1 },
                     throttle => { type => SCALAR,  
                           optional => 1 },
                     background => { type => BOOLEAN,  
                           optional => 1,
                           default => 0 },
                 parallel => { type => BOOLEAN,
                           optional => 1,
                           default => 0 },
                 timeout => { type => SCALAR,
                          optional => 1,
                          default => 300,
                          regex => qr/^\d+$/ } } );


    if ( $args{snapshot} ) {
    my $wafl_version = get_wafl_version($args{host});

    if ($wafl_version < WV_GORDONBIERSCH_MIN) {
        croak "snapshot option only from Gordonbiersch release " .
          "onwards";
    }
    }
    _hostname_present( $args{dst} );
    _hostname_present( $args{src} ) if ( defined( $args{src} ));  

    if ( $args{parallel} and my $context = _ontapi_support($args{host})) {
    _invoke_ontapi(context => $context,
             api => 'snapmirror-resync',
             args => _zapi_args(%args) );
    return 1;
    }
    # construct the command line
    my $cmd_src = (exists $args{src}) ? "-S $args{src}" : "";
    my $cmd_snapshot = (exists $args{snapshot} ? "-s $args{snapshot}" : "");
    my $cmd_dest_snapshot = (exists $args{create_snapshot} ? "-c $args{create_snapshot}": "");
    my $cmd_throttle = (exists $args{throttle}) 
      ? "-k " . $args{throttle} : "";
    my $cmd = "$sm resync -f $cmd_throttle $cmd_snapshot $cmd_dest_snapshot $cmd_src -w " . $args{dst};

    my %params = ();
    $params{timeout} = $args{timeout};
    my $ignore_err = 'source contains no new data; ' .
                     'suspending transfer to destination';
    
    my $conn = make_conn($args{host});
    my $text = rpl_start_xfer($conn, $sm, $cmd, _uppercase_keys(%params));
    if ( $@ =~ /$ignore_err/ ) {
        return;
    }
    croak $@ if $@;
    croak $text if($text);

    my $obj = subtest( sub { rpl_finish_xfer($conn, $sm, 
                         _uppercase_keys(%params));},
               '-bg', '-onexit', 
               sub{}, # don't pass results to parent
             );

    if (!$args{background}) {
    $obj->wait_finish();
    _sm_extract_subtest_err($obj);
    }
    
    return $obj;
}


=head2 snapmirror_initialize

 my $subtest_obj = snapmirror_initialize( host=>$host,
                                          dst=>$dst,
                                          [ src=>$src ],
                                          [ throttle=>$throttle ],
                                          [ snapshot=>$snapshot ],
                                          [ create_snapshot=>$dst_snapshot ],
                                          [ background=>0|1 ],
					  [ parallel=>0|1 ],
                                          [ allow_defer=>0|1 ],
                                          [ timeout=>$timeout ] );

 # if background option is set to 1
 $subtest_obj->wait_finish();

Starts a baseline transfer for a SnapMirror relationship.

If the background option is set and the initialization has successfully
started, returns a subtest object for an active subtest which is waiting 
for the transfer to complete. Important: The caller is responsible for 
checking for any exception thrown (if any) by the subtest.

On failure, dies with an error string.

=over

=item host

(Required) : Hostrec Object of filer.

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item src

(Optional) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>
If 'src' parameter is not specified, then snapmirror (not this function)
will use the source in /etc/snapmirror.conf for the destination path.

=item throttle

(Optional) : Kbs rate to throttle transfer with

=item snapshot

(Optional) : Name of specific snapshot on source to initialize from

=item create_snapshot

(Optional) : Create a snapshot with this name on the destination after 
transfer completes. Only works for an initialize to a qtree.

=item background

(Optional) : If set, returns while transfer is still running in the
background.  The result of the transfer needs to be extracted manually
later. Defaults to 0.

=item parallel

(Optional) : Sends command to filer using zapi if ontap version supports it,
otherwise send via rsh without -w option. Defaults to 0.

=item allow_defer

By default, transfers that cannot start due to resource limitations
are queued, and that is interpreted here as a failure.  If this
option is set, then a queued transfer will be intepreted as a
successfully started transfer.  The subtest object will still
wait for it to complete, as usual.

Note, tests that try to run other operations while transfers are
active cannot safely use this option.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=back

=cut

sub snapmirror_initialize {
    # XXX need a -L option for compliance (snaplock)
    my (%args) = validate( @_, { host => { type => OBJECT,
 					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         dst  => { type => SCALAR },
			         src  => { type => SCALAR,
					   optional => 1 },
   			         throttle => { type => SCALAR,  
					       optional => 1 },
 			         snapshot => { type => SCALAR,  
					       optional => 1 },
 			         create_snapshot => { type => SCALAR,
						      optional => 1 },
 			         connection_mode => { type => SCALAR,
                                                      optional => 1 },
 			         compression=> { type => SCALAR,
                                                      optional => 1 },
                                 background => { type => BOOLEAN,
						 optional => 1,
					         default => 0 },
				 parallel => { type => BOOLEAN,
					       optional => 1,
					       default => 0 },
 			         allow_defer => { type => BOOLEAN,
						  optional => 1 },
				 timeout => { type => SCALAR,
					      optional => 1,
					      default => 300,
					      regex => qr/^\d+$/ } } );

    _hostname_present($args{dst});

    if (exists $args{src}) {
	# source is specified, so use the full form for the cmd
	# cmd: snapmirror initialize -S src_filer:src_vol dst_filer:dst_vol
	_hostname_present($args{src});
    } else {
	# no source specified. assume /etc/snapmirror.conf is there and
	# use short form for the cmd
	# cmd: snapmirror initialize dst_vol
	$args{dst} =~ /^\S+:(\S+)$/;
	$args{dst} = $1;
    }

    if ( $args{parallel} and my $context = _ontapi_support($args{host})) {
	_invoke_ontapi(context => $context,
			 api => 'snapmirror-initialize',
			 args => _zapi_args(%args) );
	return 1;
    }

    $args{wait} = ! $args{parallel};
    delete $args{parallel};

    my %params = ();
    $params{timeout} = $args{timeout};

    # used in rpl_cmds::rpl_start_xfer()
    $params{allow_defer} = $args{allow_defer}; 

    # XXX/tapl this is a workaround. fix rpl_cmds.pm.
    $args{source}  = $args{src};
    
    my $cmd = rpl_cmd_from_opts($sm, "initialize", $args{dst}, 
				_uppercase_keys(%args));
    
    my $conn = make_conn($args{host});
    my $text = sayEOF($conn,"$sm on", timeout => 600);	# force registry into snapmirror.conf
    $text = rpl_start_xfer($conn, $sm, $cmd, _uppercase_keys(%params));
    
    croak $@ if $@;
    croak $text if($text);

    my $obj = subtest( sub { rpl_finish_xfer($conn, $sm, 
					     _uppercase_keys(%params));},
		       '-bg', '-onexit', 
		       sub{}, # don't pass results to parent
		     );

    if (!$args{background}) {
	$obj->wait_finish();
	_sm_extract_subtest_err($obj);
    }
    
    return $obj;
}


=head2 snapmirror_update

 my $subtest_obj = snapmirror_update( host=>$host,
                                      src=>$src,
                                      dst=>$dst,
                                      [ throttle=>$throttle ],
                                      [ snapshot=>$snapshot ],
                                      [ create_snapshot=>$dst_snapshot ],
                                      [ background=>0|1 ],
				      [ parallel=>0|1 ],
                                      [ timeout=>$timeout ] );

 # if background option is set to 1
 $subtest_obj->wait_finish();

Starts an incremental transfer for a SnapMirror relationship.

If the background option is set and the update has successfully
started, returns a subtest object for an active subtest which is waiting 
for the transfer to complete. Also, if SnapMirror relationship is synchronous,
user is responsible for ensuring that jump-ahead has taken-place after
snapmirror update has completed. Important: The caller is responsible for 
checking for any exception thrown (if any) by the subtest.

On failure, dies with an error string.

=over

=item host

(Required) : Hostrec Object of filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item throttle

(Optional) : Kbs rate to throttle transfer with

=item snapshot

(Optional) : Name of specific snapshot on source to initialize from

=item create_snapshot

(Optional) : Create a snapshot with this name on the destination after 
transfer completes. Only works for an update to a qtree.

=item background

(Optional) : If set, returns while transfer is still running in the
background.  The result of the transfer needs to be extracted manually
later. Defaults to 0.

=item parallel

(Optional) : Sends command to filer using zapi if ontap version supports it,
otherwise send via rsh without -w option. Defaults to 0.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=back

=cut

sub snapmirror_update {
    my (%args) = validate( @_, { host => { type => OBJECT,
 					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         dst  => { type => SCALAR },
			         src  => { type => SCALAR },
   			         throttle => { type => SCALAR,  
					       optional => 1 },
 			         snapshot => { type => SCALAR,  
					       optional => 1 },
 			         create_snapshot => { type => SCALAR,
						      optional => 1 },
 			         background => { type => BOOLEAN,
						 optional => 1,
					         default => 0 },
				 parallel => { type => BOOLEAN,
					       optional => 1,
					       default => 0 },
				 timeout => { type => SCALAR,
					      optional => 1,
					      default => 300,
					      regex => qr/^\d+$/ } } );
    _hostname_present($args{dst}, $args{src});

    if ( $args{parallel} and my $context = _ontapi_support($args{host})) {
	_invoke_ontapi(context => $context,
			 api => 'snapmirror-update',
			 args => _zapi_args(%args) );
	return 1;
    }
    $args{wait} = ! $args{parallel};
    delete $args{parallel};

    my %params = ();
    $params{timeout} = $args{timeout};

    # XXX/tapl this is a workaround. fix rpl_cmds.pm.
    $args{source}  = $args{src};

    my $cmd = rpl_cmd_from_opts($sm, "update", $args{dst}, 
				_uppercase_keys(%args));
    # Try starting the transfer.
    my $conn = make_conn($args{host});

    # The transfer successfully started,
    # so now we just need to wait for the completion string.
    my $obj = subtest( sub {
    			my $text = rpl_start_xfer($conn, $sm, $cmd, _uppercase_keys(%params));
    			die $@ if ($@);
    			die "$text\n" if ($text);

			rpl_finish_xfer($conn, $sm, 
					     _uppercase_keys(%params));
			}, '-bg', '-onexit', 
		       sub{},	# don't pass results to parent
		     );

    # If specified, wait for the transfer to complete
    if (!$args{background}) {
	$obj->wait_finish();
	_sm_extract_subtest_err($obj);

	# wait for jump-ahead to complete before returning control
	# to user
	my $vol = _sm_extract_volname($args{dst});
	_poll_for_thawed_vol($args{host}, $vol);
    }

    return $obj;
}


sub _zapi_args {
    my (%args) = @_;

    my %param;
    $param{'destination-location'} = $args{dst};
    $param{'destination-snapshot'} = $args{create_snapshot}
						if exists $args{create_snapshot};
    $param{'max-transfer-rate'} = $args{throttle} if exists $args{throttle};
    $param{'source-location'} = $args{src} if exists $args{src};
    $param{'source-snapshot'} = $args{snapshot} if exists $args{snapshot};
    return [ %param ];
}

=head2 snapmirror_store

 $tape_ref = snapmirror_store(host=>$hostrec,
			  vol=>$vol,
			  dumpfile=>$file_or_drive,
			  [ geometry=>$dst_disk_geometry ],
			  [ size=>$limit_in_megs ]
			 );

Start a tape dump of src volume to either a localfile or tape drive on the given
filer $hostrec. The operation run in asynch mode

On successful start of the dump, the routine will return a tape reference (not to be
confused with Perl references) of the SM vol-tape relationship.

On failure, dies with error string

=over

=item host

(Required) : Hostrec Object of filer.

=item vol

(Required) : Source volume for the 'SnapMirror store', 
i.e. <volname>

=item dumpfile

(Required) : Localfile or a comma separated list of tape drives, to hold the dump

=item geometry

(Optional) : Indicates dst disk geometry

=item size

(Optional) : Size in MBs to write per tape / localfile

=back

=cut

sub snapmirror_store {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callback => {
						'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
				 vol => { type => SCALAR },
				 dumpfile => { type => SCALAR },
				 geometry => { type => SCALAR, optional=>1 },
				 size => { type => SCALAR, optional=>1 }
				}
		 );
    
    my $conn    = make_conn($args{host});
    my $src = $args{vol};
    my $dmpfile = $args{dumpfile};
    my $cmd = _device_chk_or_prefix ($conn, $dmpfile)
		. "snapmirror store "
		. (exists $args{geometry} ? "-g $args{geometry} " : "")
		. (exists $args{size} ? "-s $args{size} " : "")
		. "$src $dmpfile";
    _start_msg_log_rdr($args{host});
    my $tape_ref = sayEOF($conn, $cmd);
    croak "$tape_ref\n"
	unless $tape_ref =~ /^Reference: /;

    $tape_ref =~ s/^Reference: //;
    chomp $tape_ref;
    return $tape_ref;
}


=head2 snapmirror_retrieve

 my $info = snapmirror_retrieve(host=>$hostrec,
				vol=>$vol,
				[ restorefile=>$file_or_drive ],
				[ geometry=>$flag ], 
				);

If <restorefile> is supplied, retrieves the dump from the localfile or tape_drive,
and restore the data to dst volume on filer $hostrec. The volume must be in
restricted mode prior to calling this function.  The operation is async.

If <geometry> is supplied, returns the geometry of the vol.

For a regular retrieve, the command runs in async mode only, and extra logic is added to
handle a special error concerned with 'disk check sum error', where the filer does an
automatic reversion of the check sum. In this situation, the routines waits for the revert to
complete and then resubmits the 'retrieve' command.

On failure, dies with error string

=over

=item host

(Required) : Hostrec Object of filer.

=item vol

(Required) : Volume name where 'SnapMirror retrieve' will restore the data to
i.e. <volname>

=item restorefile

(Optional) : Localfile or a comma separated list of tape drives, where the data is

=item geometry

(Optional) : Indicates to return disk geometry instead of doing a restore from tape

=back

=cut

sub snapmirror_retrieve{
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callback => {
						'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
				 vol => { type => SCALAR },
				 restorefile => { type => SCALAR, optional=>1 },
				 geometry => { type => BOOLEAN, optional=>1 },
				}
		);
    
    
    croak "Must supply either <restorefile> or <geometry>"
	unless (exists $args{restorefile} || exists $args{geometry});
    croak "<restorefile> or <geometry> are mutually exclusive"
	if exists $args{restorefile} && exists $args{geometry};

    my $conn    = make_conn($args{host});
    my $dst = $args{vol};
    my $cmd = "snapmirror retrieve ";
    if ( exists $args{restorefile} ) {
	$cmd = _device_chk_or_prefix ($conn, $args{restorefile})
		. $cmd
		. "$dst $args{restorefile}";
	_start_msg_log_rdr($args{host});
	sayEOF($conn, $cmd);
	if ( _redo_due_to_revert($args{host}, $args{vol}, "retrieve") ) {
		logcomment("Redo retrieve command due to automatice revert of disk");
		sayEOF($conn, $cmd);
	}
	return 1;
    } else {
	$cmd .= "-g $dst";
	my $geo = sayEOF($conn, $cmd);
	croak "Unable to retrieve geometry for $dst: $geo"
		unless $geo =~ /-g /;
	$geo =~ s/^.+-g //;
	chomp $geo;
	return $geo;
    }
}

sub _device_chk_or_prefix {
    my ($conn, $tape_device) = @_;

    if ( $tape_device =~ /^\// ) {
	# Assume local file
	return "priv set -q diag; "
    } else {
	# Assume tape device
	my %info = getTapeStatus($conn, $tape_device);
	croak "Tape device $tape_device does not exists or not ready"
		unless $info{status} =~ /ready/i;
	return "";
    }
}

=head2 snapmirror_tape_status

 my $status = snapmirror_tape_status(host=>$hostrec,
				vol=>$vol,
				[ type=>$action ],
				[ wait=>$flag ],
				[ timeout=>$timeout ]
				);

Check backup of snapmirror store or retrieve cmds. It will check /etc/messages for
successful completion log entries for these cmds, success or failure.
The routine will wait for the successful completion of the cmd or can run in poll mode,
with wait = 0, meaning in will return result after one check, default is to wait until
<timeout> for the completion msg.

An special failure message is also checked for, this is the 'disk check error, with auto revert'
message, this is returned as a separate error so that the calling routine can handle this
error.

Returns 1 if successful completion msg is found
On failure, dies with an error message

=over

=item host

(Required) : Hostrec Object of filer.

=item vol

(Required) : Volume name of the 'SnapMirror store/retrieve' cmd

i.e. <volname>

=item type

(Optional) : The action to check the status on, store or retrieve, Default store.

=item wait

(Optional) : Indicates to wait for comletion msg or return after one check. Default 1

=item timeout

(Optional) : Timeout in sec to wait for completion msg if wait = 1, Default 900

=back

=cut

sub snapmirror_tape_status {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callback => {
						'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
				 vol => { type => SCALAR },
				 type => { type => SCALAR, regex => qr/store|retrieve/,
					   default => "store" },
				 wait => { type => BOOLEAN, default => 1 },
				 timeout => { type => SCALAR, default => 900 }
				}
		);

    croak "snapmirror_store or snapmirror_retrieve must be called before here"
	unless $log_rdr;

    my $conn = make_conn($args{host});

    my $base_msg = ( $args{type} eq "store"
	     ? "SnapMirror store "
		."(from $args{vol} to tape|to tape from $args{vol})"
		." "
	     : "SnapMirror retrieve "
		."(to $args{vol} from tape|from tape to $args{vol})"
		." "
	     );
    my ($good_msg, $fail_msg) = ($base_msg."was successful", $base_msg."failed");
    my $rvt_msg = "A SnapMirror transfer to destination volume ".$args{vol}
		. " has found that the source and destination volumes"
		. " have differing checksum states."
		. " Data ONTAP is aborting the transfer and reverting"
		. " the checksum state for the destination volume.";

    my $lines = _log_search($args{timeout}, $good_msg, $fail_msg, $rvt_msg);

    return 1 if $lines =~ /$good_msg/;

    # XXX extra code to handle the 'automatic revert' error
    croak "Found: $rvt_msg" if $lines =~ /$rvt_msg/;

    croak "Found: $fail_msg" if $lines =~ /$fail_msg/;

    croak "Unable to detect expected msgs";
}

sub _redo_due_to_revert {
    my ($host, $vol, $type, $timeout) = @_;

    # It is possible to get disks with different check sums, which will cause
    # an automatic reversion of the disks, so check for it, and wait for completion
    # of the reversion, then return flag to indicate resubmission of the command
    # that caused the error.
    my $complete_msg = "Checksum revert of volume ".$vol." completed";
    $reuse_log = 0;
    eval {
	snapmirror_tape_status(host => $host,
		vol => $vol,
		type => $type,
		timeout => 15);
    };
    # Determine that the 'reverting' msg exists but not due to a time out error
    if ( $@ =~ /^Found.*$vol.*reverting/ ) {
    	$timeout = (3600*3) unless $timeout;
    	croak "Time out waiting for: $complete_msg"
		unless _log_search($timeout, $complete_msg);
    } else {
	$reuse_log = 1;
    }
    undef $@ if $@;
    return ! $reuse_log;
};

sub _log_search {
    my ($wait_time, @strs) = @_;

    my $regex = join "|", @strs;
    my $timeout_msg = "Timed out after $wait_time secs, failed to detect: $regex";

    my $loop_cnt = 0;
    do {
	if ( $loop_cnt++ == 12 ) {
		$loop_cnt = 0;
		logcomment("Waiting for msg: $regex");
	}
	my $interval = ($wait_time > 5 ? 5 : $wait_time);
	$wait_time -= $interval;
	do {
		$interval -= sleep($interval);
	} while $interval > 0;

	if ( $reuse_log ) {
		# reuse last log snippet
		$reuse_log = 0;
	} else {
		$log_rdr->restart;
	}
	my $lines = $log_rdr->matches($regex);
	if ( @$lines ) {
		my $found = join "", @$lines;
		logcomment("Found in log: $found");
		return $found;
	}
    } while $wait_time;
    croak $timeout_msg;
}

sub _start_msg_log_rdr {
    my ($host) = @_;

    unless ( $log_rdr && $log_rdr->hostname eq $host ) {
    	my $conn = make_conn($host);
	$log_rdr = LogReader::Messages->new($conn);
	$LogReader::Base::Verbose = 0; # turns off extra logging
    }
    $log_rdr->start();
    return $log_rdr;
}

=head2 snapmirror_check

 my $subtest_obj = snapmirror_check( host=>$dsthost,
                                     dst=>$dst,
                                     [ src=>$src ],
                                     [ throttle=>$throttle ],
                                     [ snapshot=>$snapshot ],
                                     [ quick_mode=>0|1 ],
                                     [ check_level=>'data'|'checksum' ],
                                     [ background=>0|1 ],
                                     [ block_based=>0|1 ],
                                     [ fix=>0|1 ],
                                     [ timeout=>$timeout ] );

 # if background option is set to 1
 $subtest_obj->wait_finish();

Starts either 'snapmirror check' or 'snapmirror check -F' (fix) process for
the specified mirror destination based on the options passed. Again,
if 'block_based' option set then block based checker/fixer is called otherwise
logical checker/fixer is invoked. The destination must exist and be in a
"snapmirrored" or "quiesced" state. Function must be invoked on the destination filer.

Prior to the execution of the command, this function will check that a 
jump-ahead is not taking place. The check will always take place regardless
of the type of the relationship, ie. qtree, asynchronous, or synchronous.

If the background option is set and the snapmirror check has successfully 
started, returns a subtest object for an active subtest which is waiting 
for the snapmirror check process to complete. Important: The caller is 
responsible for checking for any exception thrown (if any) by the subtest.

On failure or any data mismatch, dies with an error string or the data 
mismatch in $@ respectively.

Example 1. Quick mode comparison using one specified snapshot

 snapmirror_check(host=>$dsthost, dst=>$dst,
                  snapshot=>$snapshot_name,
                  quick_mode=>1);

Example 2. Quick mode comparison using all common snapshots

 snapmirror_check(host=>$dsthost, dst=>$dst,
                  quick_mode=>1);

Example 3. Checksum comparison using one specified snapshot

 snapmirror_check(host=>$dsthost, dst=>$dst,
                  snapshot=>$snapshot_name,
                  check_level=>'checksum');

Example 4. Checksum comparison using all common snapshots

 snapmirror_check(host=>$dsthost, dst=>$dst,
                  check_level=>'checksum');

Example 5. Data comparison using one specified snapshot

 snapmirror_check(host=>$dsthost, dst=>$dst,
                  snapshot=>$snapshot_name,
                  check_level=>'data');

Example 6. Data comparison using all common snapshots

 snapmirror_check(host=>$dsthost, dst=>$dst,
                  check_level=>'data');

Example 7. Data comparison using block based checker

 snapmirror_check(host=>$dsthost, dst=>$dst, block_based=>1);

Example 8. Mismatch rectification using block based fixer

 snapmirror_check(host=>$dsthost, dst=>$dst, fix=>1, block_based=>1);

Example 9. Mismatch rectification using logical fixer

 snapmirror_check(host=>$dsthost, dst=>$dst, fix=>1);

=over

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item src

(Optional) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item throttle

(Optional) : Maximum speed in kbs at which data can be transferred  
over network.

=item snapshot

(Optional) : Specify a specific snapshot snapname on the mirror destination 
to be used. By default base snapshot for a qtree-based mirror and all common 
snapshots for a volume-based mirror will be used.

=item quick_mode

(Optional) : Starts snapmirror check in quick mode. Defaults to 0.

=item check_level

(Optional) : Specify data comparison level as either data or checksum.
Defaults to data.

=item background

(Optional) : If set, returns while verification is still running in the
background.  The result of the verification needs to be extracted manually
later. Defaults to 0.

=item block_based

(Optional) : If set, then we call the block based checker/fixer. We also
skip prepending the dst path with "/vol/" in such case. Defaults to 0.

=item fix

(Optional) : If set, 'snapmirror check start -F' (snapmirror fixer) is called.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=back

=cut

sub snapmirror_check {
    my (%args) = validate( @_, { host => { type => OBJECT,
 					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         dst  => { type => SCALAR },
			         src  => { type => SCALAR,
					   optional => 1 },
   			         throttle => { type => SCALAR,  
					       optional => 1 },
 			         snapshot => { type => SCALAR,  
					       optional => 1 },
 			         quick_mode => { type => BOOLEAN,
						 optional => 1,
					         default => 0 },
 			         check_level => { type => SCALAR,
						 optional => 1 },
 			         background => { type => BOOLEAN,
						 optional => 1,
					         default => 0 },
 			         block_based => { type => BOOLEAN,
						 optional => 1,
					         default => 0 },
 			         fix => { type => BOOLEAN,
					  optional => 1,
				          default => 0 },
				 timeout => { type => SCALAR,
					      optional => 1,
					      default => 300,
					      regex => qr/^\d+$/ } } );


    _hostname_present($args{dst});
    _hostname_present($args{src}) if (exists $args{src});

    my $wafl_version = get_wafl_version($args{host});

    if ($wafl_version < WV_TSINGTAO_MIN) {
	croak "snapmirror check command only exist in Tsingtao release " .
	  "onwards";
    }
    if ($args{block_based} && ($args{dst} =~ /\/vol\//)) {
        croak "block based checker/fixer invoked when destination path contains '/vol/'";
    }

    my %args_234529 = %args; # Make copy for args for work-around before it is changed

    # snapmirror check is a diag cmd
    my $priv   = "priv set -q diag";
    my $sm_cmd = "snapmirror check start";
    my $conn   = make_conn($args{host});
    my $vol    = _sm_extract_volname($args{dst});
    my $status = vol_status($conn, $vol) or croak $@;

    my $vol_regex = qr{[a-zA-Z_]\w*};
    my $qtree_regex = qr{[^/]*[^/\s]};

    # following check saves time. ONTAP sends back different error messages
    # based on different code paths traversed.
    # we need to make sure that dst is not a qtree because the containing
    # volume could be a flex volume too and 'snapmirror fix' IS supported
    # on qtrees
    if ($args{fix} && ($status->type() eq "flex") && ($wafl_version <= WV_GORDONBIERSCH_MAX)
	&& ($args{dst} !~ qr{/vol/($vol_regex)(?:/($qtree_regex))?$}) ) {
	croak "SnapMirror Fix is NOT supported for flexible volumes in GB or prior versions";
    }

    # for 'snapmirror check' on volumes, if 'block_based' option is zero and
    # paths does not start with "/vol/" then we prepend them with "/vol/"
    # 
    # For flex vols, only logical check is supported. The code below will still
    # work because "block_based" param in that case would be 0 and we would add
    # "/vol/" to the path names.
    if ($args{block_based} == 0) { 
        if ($args{dst} =~ /^(\S+:)?([\w_]+)$/) {
 	    my $dst_vol = _sm_extract_volname($args{dst});
            $args{dst} =~ s|$dst_vol$|/vol/$dst_vol|;
        }

        if (exists $args{src} &&
            $args{src} =~ /^(\S+:)?([\w_]+)$/) {
            my $src_vol = _sm_extract_volname($args{src});
            $args{src} =~ s|$src_vol$|/vol/$src_vol|;
        }
    }

    # construct the command line. wish there was a zapi interface to this...
    my $cmd_src = (exists $args{src}) ? "-S $args{src}" : "";
    my $cmd_throttle = (exists $args{throttle}) 
      ? "-k " . $args{throttle} : "";

    my ($cmd_sshot, $cmd_quick_mode, $cmd_check_level, $fix, $cmd_force);
    if ($args{dst} =~ /^\S+:\/vol\/[\w_]+\S+$/) {
        if ($args{fix}) {
            # logical fixer	
	    $fix = "-F ";
            # uncomment the following line when burt 155411 is fixed
    	    $cmd_force = "-f";
        } else {
            # logical checker	
            $cmd_sshot = (exists $args{snapshot}) ? "-d $args{snapshot}" : "";
            $cmd_quick_mode = ($args{quick_mode}) ? "-Q" : "";
            $cmd_check_level = (exists $args{check_level}) ?
				 "-L $args{check_level}" : "-L data";
        } 	
    } else {
        if ($args{fix}) {
            # block based fixer	
	    $fix = "-F ";
    	    $cmd_force = "-f";
        } else {
            # block based checker	
            $cmd_check_level = (exists $args{check_level}) ?
				 "-L $args{check_level}" : "-L data";
        } 	
    }

    $fix = ($fix) ? $fix: "";
    $cmd_sshot = ($cmd_sshot) ? $cmd_sshot: "";
    $cmd_force = ($cmd_force) ? $cmd_force: "";
    $cmd_quick_mode = ($cmd_quick_mode) ? $cmd_quick_mode : ""; 
    $cmd_check_level = ($cmd_check_level) ? $cmd_check_level : "";

    # Logical Check:
    # snapmirror check start [-F] [-f] [ -Q ] [ -L {data | checksum} ] [ -d snapname  ]
    # [ -k kilobytes ] [ -w ] [ -S source ] destination
    #
    # Block Based Check:
    # snapmirror check start [-F] [-f] [ -k kilobytes ] [ -w ] [ -S source ] destination
    my $cmd = "$priv; $sm_cmd $fix $cmd_force $cmd_quick_mode $cmd_check_level " .
      "$cmd_sshot $cmd_throttle -w $cmd_src " . $args{dst};

    my %params = ();
    $params{timeout} = $args{timeout};

    my $need_resume = _burt234529_quiesce(%args_234529);

    my $cnt = 0;
    my $obj;
    do {
	# specific to RSM only.
	# make sure volume is not frozen before we move on
	_poll_for_thawed_vol($args{host}, $vol);
	eval {
    		_start_msg_log_rdr($args{host}); # start the log reader
		my $text = rpl_start_xfer($conn, $sm, $cmd, _uppercase_keys(%params));
		croak "$@" if $@;
		croak "$text" if($text);

		$obj = subtest( sub { rpl_finish_xfer($conn, $sm, 
						     _uppercase_keys(%params));},
			       '-bg', '-onexit', 
			       sub{}, # don't pass results to parent
			     );
	
		my $dst_obj;
		if ( $args{dst} =~ /([\w_]+)$/ ) {
			$dst_obj = $1;
		}
		if (!$args{background}) {
			$obj->wait_finish();
			_sm_extract_subtest_err($obj);
			my @err_str =
				( "$dst_obj found mismatches"
				, "$dst_obj completed and no mismatches were found"
				, "$dst_obj completed successfully");
			my $str = _log_search($args{timeout}, @err_str); 
			croak "snapmirror check found: $str"
				if $str =~ /$dst_obj found mismatches/;
		}
	};
	if ( $@ ) {
		# Work-around for BURT 245013. Retry if found transient error
		# This work-around does not support background invokation
		croak "snapmirror check/fix error: $@"
			unless $@ =~ /destination volume has been modified. Please try .+ again./;
		sleep (1); # pause a little before retrying...
		logcomment("Retry snapmirror check command due to transient error.");
        }
    } while ++$cnt < 6 && $@;
    croak "snapmirror check/fix error after $cnt attempts: $@" if $@;

    _burt234529_resume(%args_234529) if $need_resume;

    return $obj;
}


sub _burt234529_quiesce {
    my (%args) = @_;

    my $stat = snapmirror_status(host=>$args{host}, dst=>$args{dst});
    if ( $stat->{status} =~ /in-sync/i ) {
	logcomment("Quiesce in-sync mirror (work-around for BURT234529)");
        eval {
		# Ignore any errors during abort
		snapmirror_abort(host=>$args{host}, dstpath=>$args{dst});
	};
        snapmirror_quiesce(host=>$args{host}, dstpath=>$args{dst});
	return 1;
    } else {
	# No work-around needed
	return 0;
    }
}

sub _burt234529_resume {
    my (%args) = @_;

    if ( $args{background} ) {
	logcomment("Need to resume in-sync mirror after background sm check/fix completes");
    } else {
	logcomment("Resume in-sync mirror (work-around for BURT234529)");
	snapmirror_resume(host=>$args{host}, dstpath=>$args{dst});
	my $sm = snapmirror_status(host=>$args{host}, dst=>$args{dst});
	my $cfg = sm_get_config(host=>$args{host},
					src=>$sm->{'source-location'}, dst=>$args{dst});
	my $cp_sync = $cfg->{'ops-throttle'};
	$cp_sync =~ s/s//;
	my $timeout = ( exists $args{timeout} ? $args{timeout} : 30*60 );
	SnapMirror::Helper::sm_wait_sync(host=>$args{host},
					src=>$sm->{'source-location'}, dst=>$args{dst},
					cp_sync=>$cp_sync, timeout=>$timeout);
    }
}

=head2 snapmirror_fix

 my $subtest_obj = snapmirror_fix( host=>$dsthost,
                                     dst=>$dst,
                                     [ src=>$src ],
                                     [ throttle=>$throttle ],
                                     [ background=>0|1 ],
                                     [ block_based=>0|1 ],
                                     [ skip_restrict=>0|1 ],
                                     [ timeout=>$timeout ] );

 # if background option is set to 1
 $subtest_obj->wait_finish();

=over

=item host

(Required) : Hostrec Object of filer.

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item src

(Optional) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item throttle

(Optional) : Maximum speed in kbs at which data can be transferred  
over network.

=item background

(Optional) : If set, returns while verification is still running in the
background.  The result of the verification needs to be extracted manually
later. Defaults to 0.

=item block_based

(Optional) : If set, then we call the block based checker/fixer. We also
skip prepending the dst path with "/vol/" in such case. Defaults to 0.

=item skip_restrict
(Optional) : If set, then the destination traditional volume will not be
restricted before running snapmirror fixer. Defaults to 0.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=back

=cut

sub snapmirror_fix {
    my (%args) = validate( @_, { host => { type => OBJECT,
 					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         dst  => { type => SCALAR },
			         src  => { type => SCALAR,
					   optional => 1 },
   			         throttle => { type => SCALAR,  
					       optional => 1 },
 			         background => { type => BOOLEAN,
						 optional => 1,
					         default => 0 },
 			         block_based => { type => BOOLEAN,
						 optional => 1,
#					         default => 0 },
					         default => 1 },
 			         skip_restrict => { type => BOOLEAN,
						 optional => 1,
					         default => 0 },
				 timeout => { type => SCALAR,
					      optional => 1,
					      default => 300,
					      regex => qr/^\d+$/ } } );

    _hostname_present($args{dst});
    _hostname_present($args{src}) if (exists $args{src});

    my $conn = make_conn($args{host});
    my $dstvol = _sm_extract_volname($args{dst});
    my $status = vol_status($conn, $dstvol) or croak $@;

    # if dstpath is not a qtree, check if it is  a trad volume,
    # if yes, then restrict
    my $vol_restricted = 0;
    if ($args{dst} !~ /^\S+:\/vol\/([\w_]+)\/\S+$/) {
	my $wafl_version = get_wafl_version($args{host});
        if (($status->type() eq "trad") || $wafl_version > WV_GORDONBIERSCH_MAX ) {
            $args{block_based} = 1 if $wafl_version <= WV_GORDONBIERSCH_MAX; # ie force block-based for trad vol
            unless ($args{skip_restrict}) {
                # restrict destination volume unless otherwise specified
                if ($status->state() ne "restricted") {
                    vol_restrict($conn, $dstvol) or croak $@;
		    $vol_restricted = 1;
                }
            }
        } else {
	    # Error if flex vol and GB or prior
            croak "SnapMirror Fix is NOT supported for flexible volumes in GB or prior versions";
        }
    }

    # get rid of the 'skip_restrict' entry in the params hash
    # before passing it to 'snapmirror_check' function
    delete $args{skip_restrict};

    # now call snapmirror check with 'fix' option set
    my $obj;
    eval {
	$obj = snapmirror_check(
		fix => 1,
		%args);
	if ( $vol_restricted ) {
		if ( $args{background} ) {
			logcomment("Running snapmirror fix in background,"
				." need to online dst vol when it completes");
		} else {
			vol_online($conn, $dstvol);
		}
	}
    };
    croak $@ if $@;
    return $obj;
}

=head2 snapmirror_check_abort

 snapmirror_check_abort( host=>$host, dstpath=>$dstpath );

Aborts a currently running snapmirror check process for destination.   
It does not abort currently  running SnapMirror transfers. 

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item dstpath

(Required) : Destination volume or path for the SnapMirror relationship, 
i.e. <volname> or </vol/volname/qtreename>

=back

=cut

sub snapmirror_check_abort {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
						    \&_validate_hostrec } },
			         dstpath =>  { type => SCALAR } } );

    my $wafl_version = get_wafl_version($args{host});

    if ($wafl_version < WV_TSINGTAO_MIN) {
	croak "snapmirror check command only exist in Tsingtao release " .
	  "onwards";
    }

    if ($args{dstpath} !~ /^\/vol\/([\w_]+)\/\S+$/) {
	# not a QSM transfer. do the replacement
	my $dst_vol = 
	  _sm_extract_volname($args{dstpath});
	$args{dstpath} =~ s|$dst_vol$|/vol/$dst_vol|;
    }

    my $cmd = "priv set -q diag; $sm check abort " . $args{dstpath};
    my $conn = make_conn($args{host});
    my $text = sayEOF($conn, $cmd);
    croak "$text\n" if ($text);
}

=head2 snapmirror_enable

 snapmirror_enable( host=>$host );

On the given filer, turns SnapMirror on.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=back

=cut

sub snapmirror_enable {
    my %args = validate( @_, { host => { type => OBJECT,
					 callbacks => {
					     'Hostrec Object of filer' =>
					         \&_validate_hostrec }
					},
			      } );

   if (my $context = _ontapi_support($args{host})) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-on');
    } else {
	# rsh
	my $conn = make_conn($args{host});
	my $cmd = $sm . " " . "on";
	my $text = sayEOF($conn, $cmd);
	croak "$text\n" if ($text);
    }
}


=head2 snapmirror_disable

 snapmirror_disable( host=>$host );

On the given filer, turns SnapMirror off.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=back

=cut

sub snapmirror_disable {
    my %args = validate( @_, { host => { type => OBJECT,
					 callbacks => {
					     'Hostrec Object of filer' =>
					         \&_validate_hostrec }
					},
			      } );

   if (my $context = _ontapi_support($args{host})) {
	# zapi
	my @out = _invoke_ontapi(context => $context,
				 api => 'snapmirror-off');
    } else {
	# rsh
	my $conn = make_conn($args{host});
	my $cmd = $sm . " " . "off";
	my $text = sayEOF($conn, $cmd);
	croak "$text\n" if ($text);
    }
}

=head1 STATUS

 Shared

=head1 ATTRIBUTES

 ONTAP_COMMANDS = snapmirror

 ONTAP_ZAPI_APIS = snapmirror-abort
                   snapmirror-break
                   snapmirror-get-status,
                   snapmirror-get-volume-status,
                   snapmirror-list-destinations
                   snapmirror-off
                   snapmirror-on
                   snapmirror-quiesce
                   snapmirror-release
                   snapmirror-resume

=head1 AUTHOR

 DP QA (dl-dpgqa-automation)
 $Id: //depot/prod/test/Rsierranevada/lib/SnapMirror.pm#1 $
 Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

 Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut


# ---------------------------------------------------------------------
#
#                          PRIVATE FUNCTIONS
#
# ---------------------------------------------------------------------

# Function to upper-case keys in a hash
# The functions in this perl module has hashkeys in lowercase. However,
# some of these functions need to pass the hash into rpl_cmds.pm which
# has hashkeys in uppercase. _uppercase_keys() will copy the hash passed
# in and return a new hash with uppercased keys.
# XXX/tapl will fix rpl_cmds.pm to all lower-cased keys.
# Args   : Hash structure
# Returns: Copy of hash structure with upper-cased keys
sub _uppercase_keys {
    my (%old_hash) = @_;

    my %new_hash = ();
    
    foreach my $key (keys %old_hash) {
	$new_hash{uc($key)} = $old_hash{$key};
    }
    return %new_hash;
}

# Private function to validate that the hostrec object
# is indeed a hostrec object
sub _validate_hostrec {
    my (@args) = @_;

    return (UNIVERSAL::isa($args[0], 'Hostrec')) ? 1 : 0;
}

# Function to determine if filer has support for snapmirror ZAPIs
# The support will bt there if ONTAP is ver 6.4 and above
# Args   : Hostrec object
# Returns: Context if filer has support for snapmirror ZAPIs.
#          Otherwise, 0.
sub _ontapi_support {
    my ($hostrec) = @_;

    # for debugging purposes.
    return 0 if ($Use_Rsh);

    if (get_wafl_version($hostrec) >= WV_FULLSAIL_MIN) {
	# create a new context whenever we call any of these functions
	# in this package.
	# XXX when this becomes a performance bottleneck, implement
	# a caching version
	my $context = 0;
	my $passwd  = $hostrec->default_password();

	eval {
	    $context = ati_ontapi_create_context
	      (filer     => $hostrec->default_ip(),
	       transport => ATI_ONTAPI_TRANSPORT_DEFAULT,
	       user      => 'root',
	       password  => $passwd,
	       timeout   => 300,
	       debug     => ATI_ONTAPI_DEBUG_NONE);
	};

	if (!$@) {
	    return $context;
	} else {
	    # can't get context. don't croak but fallback to use 
	    # rsh instead.
	    my $call_func = (caller(1))[3];
	    logresult('WARN', 'Unable to create ATI ONTAPI context for ' .
		      "$call_func. Falling back to RSH:\n$@");
	    return 0;
	}

    } else {
	return 0;
    }
}

# wrapper to invoke Zapi call and to check Zapi results. it also prints out
# the result from the call.
#
# The usage is identical to ati_ontapi_call_api().
#   _invoke_ontapi(context=>$context, api=>'zapi-api-name', args=>[...]);
#          
# Returns: arrayref to a hash if successful. The error string is returned if 
#          not successful.
sub _invoke_ontapi {
    my (%params) = @_;

    # 'args' parameter is optional to ati_ontapi_call_api() but
    # we passing the 'args' parameter explicitly to it below. 
    # workaround this but setting this to a null reference.
    if (! exists $params{args}) {
	$params{args} = [];
    }

    # $out[0] contains Zapi outcome of the issued Zapi
    # $out[1] contains the results (if any)
    # $out[2] contains Zapi errno
    # $out[3] contains the error string which is identical to CLI's

    my @out;
    eval {
	@out = ati_ontapi_call_api(context=>$params{context},
				   api=>$params{api},
				   args=>$params{args});
    };
    my $err = $@;

    # print out what API was sent to the filer before we croak
    # on any failure, if any.
    $params{result} = $out[1]; # only concern with the results
    _ontapi_comment(%params);
    croak $@ if ($err);
	
    if ($out[0] == ATI_FAILURE) {
	croak "$out[2]: $out[3]";
    }

    return @out;
}

# Convenient function to print out what Zapi was called and its results.
# This function is called from _invoke_ontapi().
#
#   _ontapi_comment(context=>$context, api=>$api, args=>$args, 
#                   result=>$result);
#
# Returns: None
sub _ontapi_comment {
    my (%params) = @_;

    # do minimal indentation. the default value indents too much.
    local $Data::Dumper::Indent  = 1;

    my $string = "Executed ONTAPI API...\n" .
      "Host   = " . $params{context}->{filer} . "\n" .
      "Api    = " . $params{api} . "\n";

    # rework the data structure to do a pretty print
    my $h = ();
    my @arr = @{$params{args}};

    # is there a better way to do this?
    for (my $i = 0; $i < @arr; $i += 2) { 
	$h->{$arr[$i]} = $arr[$i+1];
    }
    local $Data::Dumper::Varname = 'Args';
    $string .= Dumper($h);

    # print the results from the call
    local $Data::Dumper::Varname = 'Rslt';
    $string .= Dumper($params{result});

    # The output will look something like this...
    #
    # ........................................
    # . Comment: Executed ONTAPI API...
    # . Host   = tapl
    # . Api    = snapmirror-list-destinations
    # . $Args1 = {
    # .   'source-location' => '/vol/s/qtree'
    # . };
    # . $Rslt1 = {
    # .   'destinations' => [
    # .     {
    # .       'source-snapshot' => 'tapl3(0123456193)_d_qtree-src.1',
    # .       'destination-location' => 'tapl3:/vol/d/qtree',
    # .       'source-location' => 'tapl:/vol/s/qtree'
    # .     }
    # .   ]
    # . };
    # . 
    # ........................................
    logcomment($string);
}

# during a jump-ahead in RSM, the volume is frozen and it will
# appear temporarily 'offlined'. this function polls to determine 
# if a volume involved in a RSM relationship is still frozen.
# see burt 133571 and burt 133463 for more details.
#
# **NOTE: this function is specific to snapmirror only. 
# do not copy it wholesale.**
#
# Args   : hostrec object of filer
#          name of volume to poll for
#          timeout value for the poll (default 300 secs)
# Returns: return undef if volume has unthawed, croak with a message that
#          says the timeout value has reached.
sub _poll_for_thawed_vol {
    my ($host, $vol, $timeout) = @_;

    # set default timeout to 300 secs (5 mins)
    $timeout = (defined($timeout)) ? $timeout : 300;

    my $call_func = (caller(1))[3];  # calling func name
    my $curr_func = (caller(0))[3];  # current func name

    my $check_interval   = 5;
    my $fut_time         = time() + $timeout;
    my $wait_str_printed = 0;

    logcomment("$call_func: Checking for frozen volume: $vol");

    my @zapi_error;
    while (time() < $fut_time) {

	if (my $context = _ontapi_support($host)) {
	    # directly call ati_ontapi_call_api() instead of _invoke_ontapi()
	    # to avoid printing out the status.
	    my @out;

	    # make sure that the volume is in a snapmirror relationship first
	    # otherwise just return as the volume is definitely not frozen
	    { 
		eval {
		    @out = ati_ontapi_call_api(context=>$context,
					       api=>'snapmirror-get-status');
		};
		croak $@ if ($@);

		if ($out[0] == ATI_FAILURE) {
		    croak "$out[2]: $out[3]";
		}

		my @sm_list = ();
		if (exists ($out[1]->{'snapmirror-status'})) {
			@sm_list = @{$out[1]->{'snapmirror-status'}};
		}
		else {}

		my $vol_snapmirrored = 1;
		
		foreach my $sm (@sm_list) {
		    if ($sm->{'source-location'}      =~ /\S+\:$vol$/ ||
			$sm->{'destination-location'} =~ /\S+\:$vol$/) {
			# last check. make sure it's initialized
			if ($sm->{'state'} =~ 
			    /uninitialized|broken-off|unknown/) {
			    # nope, vol is not part of snapmirror relationship
			    # stop polling and return.
			    $vol_snapmirrored = 0;
			    last;
			}
		    }
		}
		
		if (!$vol_snapmirrored) {
		    return undef;
		}
	    }

	    eval {
		# if the volume is in jump-ahead, getting SM status from that
		# volume will end up with a msg saying volume is offline.
		@out = ati_ontapi_call_api(context=>$context,
					   api=>'snapmirror-get-volume-status',
					   args=>[volume => $vol]);
	    };
	    croak $@ if ($@);

	    # ATI_SUCCESS will indicate that we didn't get an error saying
	    # that the volume is 'offline'.
	    if ($out[0] == ATI_SUCCESS ||
		# Allow Volume to be in 'restricted' mode. See BURT 236161 for details
		($out[0] != ATI_SUCCESS && $out[2] == 13061) ||
		# Allow volume to be nonexistent
		($out[0] != ATI_SUCCESS && $out[2] == 13040)) { 
		return;
	    } else {
		unless ( $zapi_error[2] == $out[2] ) {
			@zapi_error = @out;
			logcomment("snapmirror-get-volume-status error: ("
					.$out[1].", ".$out[2].")");
		}
	    }

	} else {
	    # rsh code portion
	    my $dst = $host->hostname() . ":" . $vol;
	    my $sm;

	    eval { $sm = snapmirror_status(host=>$host, dst=>$dst) };
	    if ( $@ ) {
		# no snapmirror relationship established on the host
		return undef $@ if ( $@ =~ /non-existent relationship/ ); # Clear $@ and return undef
		croak $@; # Some other error
	    }

	    if ($sm->{'state'} =~ /uninitialized|broken-off|unknown/) {
		# nope, vol is not part of snapmirror relationship
		# stop polling and return.
		return;
	    }

	    my $conn = make_conn($host);
	    my $status_obj = vol_status($conn, $vol);

	    if ($status_obj->state() ne 'frozen') {
		return;
	    }
	}

	# print this only once
	if (!$wait_str_printed) {
	    # print the calling function and what's it doing
	    logcomment("$call_func: Waiting for frozen volume '$vol' " .
		       "to be thawed");
	    $wait_str_printed = 1;
	}
	sleep $check_interval;
    }

    croak "$call_func: Volume is still frozen after polling for " .
      "$timeout secs";
}

# Function to convert CLI's datas structure to Zapi's equivalent
# Args   : hashref of the CLI's 'snapmirror status' output
# Returns: hashref of Zapi's equivalent
sub _transform_data_struct {
    my ($org) = @_;

    # new data structure conforming to zapi data struct
    my $zapi = ();

    foreach my $key (keys %{$org}) {

      SWITCH: for ($key) {
	    /^Base Snapshot$/ && do {
		# this is optional in ZAPI. don't populate it if it's empty
		if (length($org->{$key}) > 0) {
		    $zapi->{'base-snapshot'} = $org->{$key};
		}
		last;
	    };

	    /^Check Progress$/ && do {
		# only seen with 'snapmirror check'
		# ignore it for now since there's no zapi equivalent for it.
		last;
	    };

	    /^Contents$/ && do {
		# note: CLI shows '-'. rpl_cmds.pm return '' while 
		# zapi shows '-' for no progress. to be consistent,
		# we map this back to '-'.
		if ($org->{$key} eq '') {
		    # is this correct?
		    $zapi->{'contents'} = '-';
		} else {
		    $zapi->{'contents'} = $org->{$key};
		}
		last;
	    };

	    /^Current Transfer Error$/ && do {
		# this is optional in ZAPI. don't populate it if it's empty
		if (length($org->{$key}) > 0) {
		    $zapi->{'current-transfer-error'} = $org->{$key};
		}
		last;
	    };

	    /^Current Transfer Type$/ && do {
		# this is optional in ZAPI. don't populate it if it's empty
		if (length($org->{$key}) > 0) {
		    $zapi->{'current-transfer-type'} = lc($org->{$key});
		}
		last;
	    };

	    /^Destination$/ && do {
		$zapi->{'destination-location'} = $org->{$key};
		last;
	    };

	    /^Lag$/ && do {
		$zapi->{'lag-time'} = _convert_time2secs($org->{$key});
		last;
	    };

	    /^Last Transfer From$/ && do {
		# this is optional in ZAPI. don't populate it if it's empty
		if (length($org->{$key}) > 0) {
		    $zapi->{'last-transfer-from'} = $org->{$key};
		}
		last;
	    };

	    /^Last Transfer Duration$/ && do {
		# zapi represents last-transfer-duration in seconds
		$zapi->{'last-transfer-duration'} = 
		  _convert_time2secs($org->{$key});
		last;
	    };

	    /^Last Transfer Size$/ && do {
		$org->{$key} =~ /^(\d+)\s*\w+$/;
		$zapi->{'last-transfer-size'} = $1;
		last;
	    };

	    /^Last Transfer Type$/ && do {
		my $type = lc($org->{$key});
		$zapi->{'last-transfer-type'} = $type;
		last;
	    };

	    /^Mirror Timestamp$/ && do {
		# note: CLI shows '-'. rpl_cmds.pm return '' while 
		# zapi shows '0' for no progress. to be consistent,
		# we map this back to 0.
		if ($org->{$key} eq '') {
		    # is this correct?
		    $zapi->{'mirror-timestamp'} = 0;
		} else {
		    # convert CLI's date string to epoch
		    my $date  = ParseDateString($org->{$key});
		    $zapi->{'mirror-timestamp'} = UnixDate($date, "%s");
		}
		last;
	    };

	    /^Progress$/ && do {
		# note: CLI shows '-'. rpl_cmds.pm return '' while 
		# zapi shows '0' for no progress. to be consistent,
		# we map this back to 0.
		if ($org->{$key} eq '') {
		    # is this correct?
		    $zapi->{'transfer-progress'} = 0;
		} else {
		    # may show up as 'XX KB'
		    if ($org->{$key} =~ /(\d+) \w+/) {
			$zapi->{'transfer-progress'} = $1;
		    } else {
			$zapi->{'transfer-progress'} = $org->{$key};
		    }
		}
		last;
	    };

	    /^Snapshots Checked$/ && do {
		# only seen with 'snapmirror check'
		# ignore it for now since there's no zapi equivalent for it.
		last;
	    };
	    
	    /^Source$/ && do {
		$zapi->{'source-location'} = $org->{$key};
		last;
	    };

	    /^State$/ && do {
		my $state = lc($org->{$key});
		$zapi->{'state'} = $state;
		last;
	    };

	    /^Status$/ && do {
		my $status = lc($org->{$key});
		$zapi->{'status'} = $status;
		last;
	    };

	    /^Inodes Replicated$/ && do {
		# ignore it for now since there's no zapi equivalent for it.
		last;
	    };

	    # default switch
	    /.*/ && do {
		my $curr_func = (caller(0))[3];
		my $call_func = (caller(1))[3];
		logresult('WARN', "$curr_func called from $call_func: " .
			  "Unknown key: $key");
	    };
	} # end of SWITCH block
    }

    # nested function
    sub _convert_time2secs {
	my ($timestr) = @_;

	# Example of output: '00:01:27'
	if ( $timestr =~ /(\d\d):(\d\d):(\d\d)/ ) {
	    my $hrs  = $1 * 60 * 60;
	    my $mins = $2 * 60;
	    my $secs = $3;

	    return ($hrs + $mins + $secs);
	}
	return 0;
    }

    return $zapi;
}

# internal function that does a crude check for the hostname
# in the snapmirror paths.
# Args: array of paths
# Returns 1 if all of paths have ':', 0 croaks otherwise
sub _hostname_present {
	my @entries = @_;
	my $call_func = (caller(1))[3];
	foreach my $entry (@entries) {
		if ($entry !~ /^\S+:\S+$/) {
			croak ("$call_func: Has to be of the form <filer>:<volname> or " .
			       "<filer>:</vol/volname/qtree>");

		}
	}
	return 1;
}

1;
