# -*- perl -*-
#  $Id: $

## @summary SnapMirror Helper library
##
## @author dl-dpgqa-automation@netapp.com
## @status Shared
## @pod here

=head1 NAME

SnapMirror::Helper - Helper functions for using SnapMirror

=head1 SYNOPSIS

 use SnapMirror::Helper;

 my $obj = sm_init(dsthost=>$dst_hostrec, srchost=>$src_hostrec, dst=>$dst, src=>$src);

 my $obj = rsm_init(host=>$dst_hostrec, dst=>$dst, src=>$src);

 sm_set_config(host=>$dst_hostrec, dst=>$dst, src=>$src);

 my $hash_ref = sm_get_config(host=>$dst_hostrec, dst=>$dst, src=>$src);

 sm_del_config(host=>$dst_hostrec, dst=>$dst);

 sm_set_connection(host=>$dst_hostrec, name=>$name);

 sm_del_connection(host=>$dst_hostrec, name=>$name);

 sm_set_permissions(host=>$src_hostrec);

 my $array_ref = sm_get_permissions(host=>$src_hostrec);

 sm_destination_snapshot(host=>$src_hostrec, src=>$src, dst=>$dst);

 sm_get_snapid(host=>$src_hostrec, src=>$src, dst=>$dst);

 my %cfg_param = cfg_to_param($cfg_ref);

 sm_wait_sync(host=>$dst_hostrec, dst=>$dst, src=>$src);

 sm_wreck_dstqt( host=>$dsthost, src=>$src, dst=>$dst, snapid=>0);

 sm_compare_src_dst(srchost=>$srchost, dsthost=>$dsthost, src=>$src, dst=>$dst);

 sm_cleanup(srchost=>$src_hostrec, dsthost=>$dst_hostrec, src=>$src, dst=>$dst);

Please note that parameters...

1) 'src' and 'dst' are of the form <filer>:<volume> or 
<filer>:</vol/volname/qtreename>
2) 'srcpath' and 'dstpath' are of the form <volume> or </vol/volname/qtreename>

The main difference between (1) and (2) is that the string does not include 
the '<filer>:' prefix.

To trap exceptions thrown by the functions in this library, enclose the 
function call inside an eval block and check $@ for the error message.

 eval {
   sm_set_permissions(host=>$src_hostrec);
 };
 die $@ if ($@);

=head1 DESCRIPTION

A set of helper or convenient functions to aid in testing of SnapMirror.

=head1 ROUTINES

=cut

package SnapMirror::Helper;

use strict;
use warnings;

BEGIN {
    Params::Validate::validation_options(
         on_fail => sub { my @msgs = @_; chomp(@msgs);
			  Carp::confess(Tharn::script_result(@msgs)) }
    );

    our @EXPORT_OK = qw(
        rsm_init
	cfg_to_param
        sm_compare_src_dst
        sm_del_config
        sm_del_connection
        sm_destination_snapshot
        sm_get_config
        sm_get_permissions
	sm_init_pre
	sm_init_proc
        sm_init
        sm_set_config
        sm_set_connection
        sm_set_permissions
        sm_wait_sync
        sm_cleanup
        sm_wreck_dstqt
        sm_get_snapid
        _sm_extract_subtest_err
        _sm_extract_volname
    );
    # NOTE: do *NOT* import any function with the _sm_ prefix!
}


use Carp qw(croak);
use Params::Validate qw(validate validate_pos BOOLEAN SCALAR OBJECT HASHREF);
use base qw(Exporter);
use Wreck qw(wreck);
use Admin::License qw(is_licensed);
use Data::Dumper qw(Dumper);
use Filer::Version qw(get_wafl_version 
                      WV_SCRIMSHAW_MIN
                      WV_ANCHORSTEAM_MAX
                      WV_TSINGTAO_MIN
                      WV_IRONCITY_MIN);
use QtreeCmds qw(qtreeComponents qtreePresent);
use RegistryCmds qw(getOption setOption registryGet registrySet 
                    registryDelete registryWalk);
use SayWrap qw(sayEOF);
use SnapCmds qw(snapshot_delete snapshot_status snapshot_exists snapshot_list);
use SnapMirror qw(snapmirror_status snapmirror_enable snapmirror_initialize
                  snapmirror_destinations snapmirror_check snapmirror_release
		  snapmirror_quiesce snapmirror_break snapmirror_abort snapmirror_resync);
use Tharn;
use Utils qw(make_conn);
use Volume qw(vol_online vol_restrict vol_status vol_size);
use cat qw(contains);
use fileSystem qw(treecompare);


=head2 sm_init

 my $subtest_obj = sm_init( srchost=>$srchost,
                            dsthost=>$dsthost,
			    src=>$src,
			    dst=>$dst,
			    [ no_qt_src=>0|1 ],
			    [ throttle=>$throttle ],
			    [ snapshot=>$snapshot ],
			    [ create_snapshot=>$dst_snapshot ],
			    [ background=>0|1 ],
			    [ parallel=>0|1 ],
			    [ allow_defer=>0|1 ]
			    [ timeout=>$timeout ] 
			    [ legacy_access=>0|1 ] );

 # if background option is set to 1
 $subtest_obj->wait_finish();

Performs a level 0 snapmirror transfer between a source and destination.

This helper function will perform the following actions:

 1) Checks if both $src and $dst are either volumes or qtrees respectively.
 2) Sets up snapmirror access permission if necessary.
 3) Creates entries in the snapmirror.conf if necessary.
 4) Enables snapmirror.
 5) Onlines the source volume.
 6) If it's a volume snapmirror...
 6.2) Restricts the destination volume.
 7) If it's a logical snapmirror...
 7.1) Verifies that the source qtree exist.
 7.2) Verifies that the destination qtree does not exist.
 7.3) Onlines the destination volume.
 8) Executes SnapMirror::snapmirror_initialize()

If the background option is set and the initialization has successfully
started, returns a subtest object for an active subtest which is waiting 
for the transfer to complete.

This function have been broken up into 3 functions:

B<sm_init_pre> - Perform preprocessing of admin commands suchs setting registry entries
for the snapmirror command

B<sm_init_post> - Perform the snapmirror command

B<sm_init> - Calls I<sm_init_pre> and I<sm_init_post> functions to carry out snapmirror initialize.

On failure, dies with an error string.

=over

=item srchost

(Required) : Hostrec Object of source filer.

=item dsthost

(Required) : Hostrec Object of destination filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item no_qt_src

(Optional) : Indicates not to check for qtree src.

=item throttle

(Optional) : Kbs rate to throttle transfer with

=item snapshot

(Optional) : Name of specific snapshot on source to initialize from

=item create_snapshot

(Optional) : Create a snapshot with this name on the destination after 
transfer completes. Only works for an initialize to a qtree.

=item background

(Optional) : If set, returns while transfer is still running in the
background.  The result of the transfer needs to be extracted manually
later. Defaults to 0.

=item parallel

(Optional) : If set, runs in parallel mode.

=item allow_defer

By default, transfers that cannot start due to resource limitations
are queued, and that is interpreted here as a failure.  If this
option is set, then a queued transfer will be interpreted as a
successfully started transfer.  The subtest object will still
wait for it to complete, as usual.

Note, tests that try to run other operations while transfers are
active cannot safely use this option.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=item legacy_access

(Optional) : Use legacy snapmirror settings for snapmirror.access and
snapmirror.allow. Defaults to 0.

=back

=cut

sub _sm_init_args {
    # XXX need a -L option for compliance (snaplock)
    my (%args) = validate( @_, { srchost => { type => OBJECT,
					      callbacks => {
						   'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
				 dsthost => { type => OBJECT,
					      callbacks => {
					           'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         src  => { type => SCALAR },
			         dst  => { type => SCALAR },
				 no_qt_src => { type => SCALAR,
						optional => 1 },
   			         throttle => { type => SCALAR,  
					       optional => 1 },
 			         snapshot => { type => SCALAR,  
					       optional => 1 },
                                 connection_mode => { type => SCALAR,
                                             optional => 1 },
                                 compression=> { type => SCALAR,
                                             optional => 1 },
 			         create_snapshot => { type => SCALAR,
						      optional => 1 },
 			         background => { type => BOOLEAN,
						 optional => 1,
					         default => 0 },
				 parallel => { type => BOOLEAN,
					       optional => 1,
					       default => 0 },
 			         allow_defer => { type => BOOLEAN,
						  optional => 1 },
				 timeout => { type => SCALAR,
					      optional => 1,
					      default => 600,
					      regex => qr/^\d+$/ },
			         legacy_access => { type => BOOLEAN,
						    optional => 1,
						    default => 0 } } );
    return %args;
}

sub sm_init_pre {
    my %args = _sm_init_args(@_);

    _hostname_present($args{dst}, $args{src});

    my ($srcvol, $dstvol, $srcqt, $dstqt);

    my $curr_func = (caller(0))[3];
    my $logical_mirror = 0;

    # checks if $args{src} and $args{dst} are qtree paths
    if (_is_logical_mirror($args{src}, $args{dst})) {
	# QSM. it's a logical snapmirror
	$logical_mirror = 1;

	# grab the qtree path from the split().
	$srcqt = ( split(/:/, $args{src}) )[1];
	$dstqt = ( split(/:/, $args{dst}) )[1];

	# now extract the volume name
	$srcvol = _sm_extract_volname($args{src});
	$dstvol = _sm_extract_volname($args{dst});
    } elsif (_is_volume_mirror($args{src}, $args{dst})) {
	# VSM
	$logical_mirror = 0;

	# now extract the volume name
	$srcvol = _sm_extract_volname($args{src});
	$dstvol = _sm_extract_volname($args{dst});
    } else {
	croak "$curr_func: Either both source and destination should be " .
	  "volumes or qtree paths";
    }

    # create connection to filer(s)
    my $dstconn = make_conn($args{dsthost});
    my $srcconn;
    my $local_xfer = 0;
    if (_is_local_xfer($args{srchost}, $args{dsthost})) {
	$srcconn = $dstconn;
	$local_xfer = 1;
    } else {
	$srcconn = make_conn($args{srchost});
    }

    # check and make sure both filer have 'snapmirror' licensed
    if (!is_licensed($dstconn, 'snapmirror')) {
	croak "SnapMirror is not licensed on destination " .
	  "filer: " . $dstconn->hostp()->hostname();
    }

    if (!is_licensed($srcconn, 'snapmirror')) {
	croak "SnapMirror is not licensed on source " .
	  "filer: " . $srcconn->hostp()->hostname();
    }

    # setup snapmirror permissions if transfer is b/w 2 filers
    if (!$local_xfer) {
	sm_set_permissions(host=>$args{srchost}, 
			   allowed_host=>$args{dsthost}->hostname(),
			   legacy_access=>$args{legacy_access});
    }

    # creates entries in /etc/snapmirror.conf
    # make sure entry for transfer exists
    my $conf = sm_get_config(host=>$args{dsthost},
			     src=>$args{src},
			     dst=>$args{dst});

    # entry does not exist or if config sched is set to 'sync',
    # reset it back to '- - - -'.
    if (scalar(keys %{$conf}) == 0 ||
	_is_config_sched_sync($conf)) {

	sm_set_config(host=>$args{dsthost},
		      src=>$args{src},
		      dst=>$args{dst});
    } else {
	# found an entry with the same destination location. abort.
	if ($conf->{'destination-location'} eq $args{dst}) {
	    if ($conf->{'source-location'} ne $args{src}) {
		# source mismatch, abort
		croak "Duplicate entry for destination ". $args{dst} .
		      " in snapmirror.conf, source mismatch!! ".
		      "Existing source ". $conf->{'source-location'}.
		      " != ". "new source entry ". $args{src} .
		      " Aborting... ";
	    } else {
		# source identical, log msg
		logcomment("Duplicate entry for the same src, dst combo ".
			   $conf->{'source-location'} . " == " . $args{src}.
			   " Ignoring...");
	    }
	}
    }
     #Check if connection_mode paramter exists, either as a part of environment or the script itself. If it exists, call sm_set_config to set the
     #connection_mode parameter. This is valid only for IPv6 and other releases that support IPv6.
     if(exists $args{connection_mode}) {
     sm_set_config(host=>$args{dsthost},
                       src=>$args{src},
                       dst=>$args{dst},
                       connection_mode=>$args{connection_mode});
     }

    if(exists $args{compression}) {
    my $reg_key = "options.files.snapmirror.conf.".$args{dst};
    my $enable_string=$args{compression} ? 'enable':'disable' ;
    my $reg_args_str="compression=".$enable_string ;
    my $conn = make_conn($args{dsthost});
    my $get_string=registryGet($conn, "$reg_key.args") ;
    croak $@ if(($@) && ($@!~/does not exist/));
    unless($@){
        my @get_string= split(/args\=/, $get_string);
        $get_string[0]=~s/compression\=(en|dis)able//;
        $get_string[0]=~s/\-//;
        $reg_args_str=$get_string[0].",$reg_args_str" ;
        $reg_args_str=~s/\,\,/\,/;
        $reg_args_str=~s/^\,//;
        }
    registrySet($conn, "$reg_key.args",$reg_args_str) or croak $@;
    }
    


    # enable snapmirror
    snapmirror_enable(host=>$args{srchost});
    snapmirror_enable(host=>$args{dsthost});

    # online source volume if it's not in that state
    my $status = vol_status($srcconn, $srcvol) or croak $@;
    if ($status->state() ne "online" && 
	$status->status() !~ /snapmirrored/) {
	vol_online($srcconn, $srcvol) or croak $@;
    }

    if (!$logical_mirror) {
	# VSM. restrict dst volume
	my $status = vol_status($dstconn, $dstvol) or croak $@;
	if ($status->state() ne "restricted") {
	    vol_restrict($dstconn, $dstvol) or croak $@;
	}

    } else {
	# QSM
	# make sure src qtree exists
	unless ( $srcqt =~ /\/-$/ ||  $args{no_qt_src} ) {
		# Check for qtree presence if not the special root qt '-'
		my $rslt = qtreePresent($srcconn, $srcqt, TIMEOUT=>$args{timeout});
		croak "Source qtree " . $srcqt . " does not exist." if (!$rslt);
	}

	# online destination volume
	vol_online($dstconn, $dstvol) or croak $@;
    }
}

sub sm_init_proc {
    my %args = _sm_init_args(@_);

    # ok, we're done with the setup.
    # need to manipulate some keys from %args hash before we pass the %args
    # hash to snapmirror_initialize(), otherwise snapmirror_initialize()
    # params validation will complain abt unknown new keys. check
    # snapmirror_initialize() for list of valid parameters.
    $args{host} = $args{dsthost}; # initialization is done on dst filer
    delete $args{src}; 
    delete $args{dsthost};
    delete $args{srchost};
    delete $args{legacy_access};

    return snapmirror_initialize(%args);
}

sub sm_init {
    sm_init_pre(@_);
    return sm_init_proc(@_);
}

=head2 rsm_init

 my $subtest_obj = rsm_init( srchost=>$srchost,
                             dsthost=>$dsthost,
			     src=>$src,
			     dst=>$dst,
			     [ cp_sync=>$cp_sync ],
			     [ throttle=>$throttle ],
			     [ snapshot=>$snapshot ],
			     [ create_snapshot=>$dst_snapshot ],
  			     [ background=>0|1 ],
			     [ allow_defer=>0|1 ],
			     [ parallel=>0|1 ],
			     [ timeout=>$timeout ] );

 # if background option is set to 1
 $subtest_obj->wait_finish();

Performs a level 0 snapmirror transfer between a source and destination and
sets the relationship into sync mode.

This helper function will perform identical actions as those in sm_init() 
(see documentation for sm_init()). On top of that, it will verify that the
source and destination volumes are at least 10Gb and it also edit 
/etc/snapmirror.conf to transition the relationship into sync mode while the
baseline transfer has completed.

If the background option is set and the *baseline transfer* has successfully
completed, returns a subtest object for an active subtest which is waiting 
for the mirror to go into full sync or CP sync mode. This behavior is 
*different* from sm_init().

On failure, dies with an error string.

=over

=item srchost

(Required) : Hostrec Object of source filer.

=item dsthost

(Required) : Hostrec Object of destination filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item cp_sync

(Optional) : Boolean value to determine if the function should be checking
if the mirror goes into full sync (default) or CP sync. Defaults to 0.

=item semi_sync

(Optional) : Boolean value to determine if a semi-sync (a.k.a CP sync)
relationship should be set up rather than a fully synced. Defaults to 0.

=item throttle

(Optional) : Kbs rate to throttle transfer with

=item snapshot

(Optional) : Name of specific snapshot on source to initialize from

=item create_snapshot

(Optional) : Create a snapshot with this name on the destination after 
transfer completes. Only works for an initialize to a qtree.

=item background

(Optional) : If set, returns while transition to full or CP sync mode is 
still happening in the background. Defaults to 0.

=item parallel

(Optional) : If set, runs in parallel mode.

=item allow_defer

(Optional) : By default, transfers that cannot start due to resource limitations
are queued, and that is interpreted here as a failure.  If this 
option is set, then a queued transfer will be interpreted as a
successfully started transfer.  The subtest object will still
wait for it to complete, as usual.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=back

=cut

sub rsm_init {
    my (%args) = validate( @_, { srchost => { type => OBJECT,
					      callbacks => {
						   'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
				 dsthost => { type => OBJECT,
					      callbacks => {
					           'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         src  => { type => SCALAR },
			         dst  => { type => SCALAR },
   			         cp_sync => { type => BOOLEAN,  
					      optional => 1,
					      default => 0 },
   			         semi_sync => { type => BOOLEAN,  
					      optional => 1,
					      default => 0 },					      
   			         connection_mode =>{ type => SCALAR,
                                               optional => 1 },
   			         compression=>{ type => SCALAR,
                                               optional => 1 },
                                 throttle => { type => SCALAR,  
					       optional => 1 },
 			         snapshot => { type => SCALAR,  
					       optional => 1 },
 			         create_snapshot => { type => SCALAR,
						      optional => 1 },
 			         background => { type => BOOLEAN,
						 optional => 1,
					         default => 0 },
				 parallel => { type => BOOLEAN,
					       optional => 1,
					       default => 0 },
 			         allow_defer => { type => BOOLEAN,
						  optional => 1 },
				 timeout => { type => SCALAR,
					      optional => 1,
					      default => 300,
					      regex => qr/^\d+$/ } } );

    if ($args{semi_sync}){
        # Make sure version of ontap supports the implemented semi-sync setup
        # For Ontap versions earlier than Ironcity, semi-sync relationships (a.k.a CP-sync) 
        # are set up using the registry key 'outstanding'        
        # TODO: Are there requirements on the source too?
        my $dst_version = get_wafl_version($args{dsthost});
        if ($dst_version < WV_IRONCITY_MIN){
            croak "Semi-Synchronous SnapMirror not supported for versions earlier than Ironcity";
        }
        
        # Override cp_sync for semi-sync relationships
        # since it doesn't make sense to wait for full sync mode        
        $args{cp_sync} = 1;
    }

    _hostname_present($args{dst}, $args{src});

    my $src_conn = make_conn($args{srchost});
    my $dst_conn = make_conn($args{dsthost});

    # check and make sure both filer have 'snapmirror_sync' licensed
    # if this is not done here, snapmirror will never go into sync mode
    # and sm_wait_sync() (below) will never complete and only timeout
    # much much later.
    if (!is_licensed($dst_conn, 'snapmirror_sync')) {
	croak "Synchronous SnapMirror is not licensed on destination " .
	  "filer: " . $dst_conn->hostp()->hostname();
    }

    if (!is_licensed($src_conn, 'snapmirror_sync')) {
	croak "Synchronous SnapMirror is not licensed on source " .
	  "filer: " . $src_conn->hostp()->hostname();
    }

    # check and ensure that both src and dst volumes are at least 10g
    # i got burned too many times because of this requirement...
    my $src_vol = _sm_extract_volname($args{src});
    my $dst_vol = _sm_extract_volname($args{dst});

    my $src_size = vol_size($src_conn, $src_vol) or croak $@;
    my $dst_size = vol_size($dst_conn, $dst_vol) or croak $@;

    my $min_size = 10 * 1024 * 1024; # 10g in Kb

    if ($src_size < $min_size) {
	if( $src_conn->{'hostp'}->{'hosttype'} eq 'filer-simulator' ) {
	    sayEOF($src_conn, "priv set -q diag;setflag rsm_allow_small_flex_vols 1");
	} else {
	    croak "Source volume must be at least 10g";
	}
    }

    if ($dst_size < $min_size) {
	if( $dst_conn->{'hostp'}->{'hosttype'} eq 'filer-simulator' ) {
	    sayEOF($dst_conn, "priv set -q diag;setflag rsm_allow_small_flex_vols 1");
	} else {
	    croak "Destination volume must be at least 10g";
	}
    }

    #if ($src_size < $min_size || $dst_size < $min_size) {
    #	croak "Both source and destination volume must be at least 10g";
    #}

    # delete the background option before we pass to sm_init().
    # the background option is only used to wait for a mirror to
    # go into full sync mode below. also delete cp_sync option.
    my %sm_init_args = %args; # make a copy of the args
    delete $sm_init_args{background};
    delete $sm_init_args{cp_sync};
    delete $sm_init_args{semi_sync};
    sm_init(%sm_init_args);

    if ($args{semi_sync}){
        sm_set_config(host=>$args{dsthost},
		  src=>$args{src},
		  dst=>$args{dst},
		  sched=> { semi_sync=>1 });        
    }
    else{
        sm_set_config(host=>$args{dsthost},
		  src=>$args{src},
		  dst=>$args{dst},
		  sched=> { sync=>1 });
    }

    # do the background thingy here. gives the caller an option to wait
    # for mirror to go to full sync or CP sync mode, or to do other
    # things while waiting for it to go full sync or CP sync mode.
    my $obj = subtest( sub { sm_wait_sync(host=>$args{dsthost},
					  src=>$args{src},
					  dst=>$args{dst},
					  cp_sync=>$args{cp_sync});},
		       '-bg', '-onexit', sub {});

    if (!$args{background}) {
	$obj->wait_finish();
	_sm_extract_subtest_err($obj);
    }

    return $obj;
}


=head2 sm_wait_sync

Waits for the snapmirror relationship to enter sync or CP sync mode. The
timer does not start if the mirror is in the last async transfer right
before entering sync mode.

Dies with an error string on failure.

 sm_wait_sync( host=>$host,
               src=>$src,
               dst=>$dst,
               [ cp_sync=>0|1 ],
               [ timeout=>$timeout ]);

=over

=item host

(Required) : Hostrec Object of filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item cp_sync

(Optional) : Boolean value to determine if the function should be checking
if the mirror goes into full sync (default) or CP sync. Defaults to 0.

=item timeout

(Optional) : Timeout value to wait for the mirror to go into either full
sync or CP sync. Defaults to 600 (10 mins).

=back

=cut

sub sm_wait_sync {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         src  => { type => SCALAR },
			         dst  => { type => SCALAR },
   			         cp_sync => { type => BOOLEAN,  
					      optional => 1,
					      default => 0 },
			         timeout => { type => SCALAR,
					      optional => 1,
					      default => 600 } } );

    _hostname_present($args{dst}, $args{src});

    my $check_interval = 15;

    # smaller check interval for simulators
    if ($args{host}->hosttype eq 'filer-simulator') {
	$check_interval = 5;
    }

    while (1) {
	my $s = snapmirror_status(host=>$args{host}, 
				  src=>$args{src}, 
				  dst=>$args{dst});

	croak $s->{'current-transfer-error'} if( $s->{'current-transfer-error'} && $s->{'status'} !~ /pending/i );

	# check for full sync mode
	if ( !$args{cp_sync} && 
	     ($s->{'state'}  eq 'snapmirrored' || $s->{'state'} eq 'source') &&
	     ($s->{'status'} eq 'in-sync') &&
	     ($s->{'lag-time'} == 0) ) {
	    last;
	}

	# check for CP sync mode
	if ( $args{cp_sync} && 
	     ($s->{'state'}  eq 'snapmirrored' || $s->{'state'} eq 'source') &&
	     ($s->{'status'} eq 'in-sync') &&
	     ($s->{'lag-time'} >= 10) ) {
	    last;
	}

	if ($args{timeout} <= 0) {
	    croak "Timeout waiting for mirror to retain sync mode";
	}

	if ( ($s->{'state'}  eq 'snapmirrored'   ||
	      $s->{'state'}  eq 'source')        &&
	     ($s->{'status'} eq 'syncing'        || 
	      $s->{'status'} eq 'transferring')  &&
	     ($s->{'transfer-progress'} != 0) ) {
	    $args{timeout} -= $check_interval
	}

	if ( $s->{'state'} eq 'snapmirrored' &&
	     (	$s->{'status'} =~ /(idle|pending)/i
	     && $s->{'transfer-progress'} =~ /^[0-]$/
	     )
	) {
	    # Also start timer if in idle or pending mode with no transfer in progress
	    $args{timeout} -= $check_interval;
        }
	sleep $check_interval;
    }
}


=head2 sm_set_config

 sm_set_config( host=>$host,
                src=>$src,
		dst=>$dst,
		[ cksum=>$cksum ],
		[ kbs=>$kbs ],
                [ tries=>$tries ],
		[ restart=>$restart ],
		[ wsize=>$wsize ],
   	        [ outstanding=>$outstanding ],
                [ connection_mode=>$connection_mode],
		[ visibility_interval=>$visibility_interval ],
		[ sched=> {
                      sync=>1|0
                      ... or ...
                      mins=>0-59,
                      hours=>0-23,
                      days_of_month=>1-31,
                      days_of_week=>0-6    # 0-Sun to 6-Sat
                } ] );

This helper function defines the mirror relationship for destination volumes
on the filer. It specifies the volume or qtree to be replicated, arguments for 
the replication, and the schedule for updating the mirror.

Dies with an error string on failure.

Example 1. /etc/snapmirror.conf has the following relationship defintion:

 netapp0-gig:home netapp1:home_mirror kbs=2000 30 9,13,19 * 1,2,3,4,5

 sm_set_config(host=>'netapp1', 
               src=>'netapp0-gig:home',
               dst=>'netapp1:home_mirror',
               kbs=>2000,
               sched=> {
                   mins=>30,
                   hours=>'9,13,19',
                   days_of_month=>'*',
                   days_of_week=>'1,2,3,4,5'
               });

Example 2. /etc/snapmirror.conf has the following relationship defintion to
           setup synchronous mirroring:

 netapp0-gig:home netapp1:home_mirror - sync

 sm_set_config(host=>'netapp1', 
               src=>'netapp0-gig:home',
               dst=>'netapp1:home_mirror',
               sched=> {
                   sync=>1 
               });

=over

=item host

(Required) : Hostrec Object of filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item cksum

(Optional) : Controls which checksum algorithm is used to protect the data 
transmitted by SnapMirror.

=item kbs

(Optional) : The value for this argument specifies the maximum speed 
(in kilobytes per second) at which SnapMirror data is transferred over 
the network.

=item tries

(Optional) : The value for this argument specifies the maximum number of 
attempts that the destination will make to complete a scheduled snapmirror 
update.

=item restart

(Optional) : This controls the behavior of the SnapMirror scheduler with 
respect to restartability.

=item wsize

(Optional) : This sets the TCP window size to use for the connection. 

=item outstanding

(Optional) : This argument controls the performance versus synchronicity 
trade-off for synchronous mirrors. This argument is ignored for asynchronous 
mirrors.

=item visibility_interval

(Optional) : This argument controls the amount of time before an automatic 
snapshot is created on the source volume that is synchronously mirrored. 
This argument is ignored for asynchronous mirrors.

=item sched

(Optional) : This is the schedule used by the destination filer for updating 
the mirror.

=over

=item sync

(Optional) : If this value is set, it will indicate that synchronous mirroring
is desired. Likewise, other parameters such as 'mins', 'hours', 'days_of_month',
and 'days_of_week' will be ignored. Defaults to 0.

=item mins

(Optional) : Which minutes in each hour to update on. Values are from 0 to 59.
Defaults to '*'.

=item hours

(Optional) : Which hours in the day to update on. Values are from 0 to 23.
Defaults to '*'.

=item days_of_month

(Optional) : Which days in the month to update on. Values are from 1 to 31.
Defaults to '*'.

=item days_of_week

(Optional) : Which days in the week to update on. Values are from 0 (Sunday) 
to 6 (Saturday). Defaults to '*'.

=back

=back

=cut

sub sm_set_config {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         src  => { type => SCALAR },
			         dst  => { type => SCALAR },
   			         cksum => { type => SCALAR,  
					    optional => 1 },
 			         kbs => { type => SCALAR,  
					  optional => 1 },
 			         tries => { type => SCALAR,  
					    optional => 1 },
 			         connection_mode =>{ type => SCALAR,
                                           optional => 1 },
 			         compression=>{ type => SCALAR,
                                           optional => 1 },
				 restart => { type => SCALAR,
					      optional => 1 },
 			         wsize => { type => SCALAR,
					    optional => 1 },
 			         outstanding => { type => SCALAR,
						  optional => 1 },
				 visibility_interval => { type => SCALAR,
							  optional => 1 },
				 group => { type => SCALAR,
					    optional => 1 },
 			         sched => { type => HASHREF,
					    optional => 1,
					    callbacks => {
					        'sched validation' =>
						    \&_validate_config_sched }
                                           },
			       } );

    _hostname_present($args{dst}, $args{src});

    my $conn = make_conn($args{host});

    if (exists $args{sched} && $args{sched}{semi_sync}){
        my $version = get_wafl_version($args{host});
        if ($version < WV_IRONCITY_MIN){
            croak "Semi-Synchronous SnapMirror not supported for versions earlier than Ironcity";
        }
    }

    # registry key
    my $reg_key = "options.files.snapmirror.conf." . $args{dst};

    # set options.files.snapmirror.conf.<dstfiler>:<dstvol>.dest
    registrySet($conn, "$reg_key.dest", $args{dst}) or croak $@;


    # set options.files.snapmirror.conf.<dstfiler>:<dstvol>.source   
    registrySet($conn, "$reg_key.source", $args{src}) or croak $@;


    # set options.files.snapmirror.conf.<dstfiler>:<dstvol>.args   
    my @reg_args = ();
    #connection_mode is added as one of the parameter that needs to be set in the config file. Supported only for IC.3 and other IPv6 supporting releases.
   
    foreach my $key (qw(connection_mode cksum kbs restart wsize outstanding
                       group visibility_interval)) {
	next if (!exists $args{$key});
	push(@reg_args, "$key=$args{$key}");
    }

    #When mode is 'semi-sync' or 'sync' all the arguments set in the conf file is cleaned up. To have a work-around for that, we do a registry get for the conf file
    # on the destination volume, if it has connection_mode entry, then it will not be replaced by '-'
    my $return_string=registryGet($conn, "$reg_key");
    my $reg_args_str='';
    $reg_args_str = '-' unless($return_string=~/inet/); # default value if there's no SM args
    $reg_args_str = join(',', @reg_args) if (@reg_args);
    if($reg_args_str)
    {
         registrySet($conn, "$reg_key.args",$reg_args_str) or croak $@;
    }

    # set options.files.snapmirror.conf.<dstfiler>:<dstvol>.sched
    my $reg_sched_str = '- - - -'; # default value if there's no SM sched

    # override the default value above if schedule is set
    if (exists $args{sched}) {
	my @reg_sched = ();

	# check if sched's sync is set. if so, it's RSM and we can ignore
	# the rest of the schedule
	if (exists $args{sched}{sync} && $args{sched}{sync}) {
	    $reg_sched_str = 'sync';
	}
	elsif (exists $args{sched}{semi_sync} && $args{sched}{semi_sync}){
	    $reg_sched_str = 'semi-sync';
	}
	else {
	    # do not change the order of the loop!
	    foreach my $key (qw(mins hours days_of_month days_of_week)) {

		if (exists $args{sched}{$key}) {
		    push(@reg_sched, $args{sched}{$key});
		} else {
		    push(@reg_sched, '*');
		}
	    }
	    $reg_sched_str = join(' ', @reg_sched);
	}
    }
    registrySet($conn, "$reg_key.sched", $reg_sched_str) or croak $@;
}


=head2 sm_get_config

 my $hash_ref = sm_get_config( host=>$host,
                               dst=>$dst,
                               src=>$src );

Retrieves snapmirror configuration for a given relationship.

It returns a hash reference containing the snapmirror configuration on success 
and dies with the error string on failure. The hash keys are:

 destination-location
 source-location
 visibility-frequency (optional)
 max-transfer-rate (optional)
 ops-throttle (optional)
 minutes (optional)
 hours (optional)
 days-of-month (optional)
 days-of-week (optional)

The following are made up as there're no equivalent keys defined in Zapi yet.
 restart (optional)
 checksum (optional)
 tries (optional)
 window-size (optional)
 connection_mode (optional)
Keys with the '(optional)' suffix indicate that these keys may not or
may not appear in the hash reference returned by sm_get_config().

If the relationship configuration is for synchronous snapmirror, the keys 
'minutes', 'hours', 'days-of-month', and 'days-of-week' will not exist.

=over

=item host

(Required) : Hostrec Object of filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=back

=cut

sub sm_get_config {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         src  => { type => SCALAR },
			         dst  => { type => SCALAR } } );

    _hostname_present($args{dst}, $args{src});

    my $conn = make_conn($args{host});
    my $s = ();
    my $reg_key = "options.files.snapmirror.conf." . $args{dst};

    # check to see if the key is there. if not, return an empty hash
    # reference
    eval {
	registryGet($conn, "$reg_key") or croak $@;
    };
    return $s if ($@);

    # get the source and destination info
    $s->{'source-location'} = registryGet($conn, "$reg_key.source")
      or croak $@;
    $s->{'destination-location'} = registryGet($conn, "$reg_key.dest")
      or croak $@;

    # get the arguments info
    my $args_str = registryGet($conn, "$reg_key.args") or croak $@;

    my %mapping = ('cksum'   => 'checksum',
		   'kbs'     => 'max-transfer-rate',
		   'tries'   => 'tries',
		   'group'   => 'group',
		   'restart' => 'restart',
		   'wsize'   => 'window-size',
		   'outstanding' => 'ops-throttle',
                   'connection_mode' =>'connection_mode',
                   'compression' =>'compression',
		   'visibility_interval' => 'visibility-frequency');

    # $args_str will look like 'key=value,key2=value2,key3=value3,...'
    foreach my $key_value_pair (split(/,/, $args_str)) {
	$key_value_pair =~ /(\S+)=(\S+)/;
	if ($1 && exists $mapping{$1}) {
	    $s->{$mapping{$1}} = $2;
	} elsif ($key_value_pair eq '-') {
	    # do nothing
	} else {
	    logcomment("sm_get_config(): Unknown args key-value pair: " .
		       "$key_value_pair");
	}
    }

    # get the schedule information
    my $sched_str = registryGet($conn, "$reg_key.sched") or croak $@;

    # shows cron-style schedule
    if ($sched_str =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/) {
	$s->{'minutes'}       = $1;
	$s->{'hours'}         = $2;
	$s->{'days-of-month'} = $3;
	$s->{'days-of-week'}  = $4;

    } elsif ($sched_str =~ /semi-sync/) {
	$s->{'ops-throttle'}  = 1;
    } elsif ($sched_str =~ /sync/) {
	# do nothing
    } else {
	logcomment("sm_get_config(): Unknown schedule: $sched_str");
    }
    return $s;
}

=head2 cfg_to_param

Convert a config hash returned by 'sm_get_config' to the format of parameters suitable for
sm_set_config.

It removes 'source-location' and destination-location' entries as these are passed
separately to 'sm_set_config'.

Returns a hash of entries suitable to be pass to 'sm_set_config' routine.

=item $old_cfg

A reference to a has of config entries as returned by sm_get_config


=item Example

    my $old_cfg = sm_get_config(host=>$dst_hostrec, src=>$src, dst=>$dst);
    my %new_cfg = cfg_to_param($old_cfg);

=cut

sub cfg_to_param {
    my ($old_cfg) = @_;

    my %mapping = (
		# Regular entries
		'checksum' => 'cksum',
		'max-transfer-rate' => 'kbs',
		'window-size' => 'wsize',
		'ops-throttle' => 'outstanding',
		'visibility-frequency' => 'visibility_interval',
		# schedule entries
		'minutes' => 'sched.mins',
		'hours' => 'sched.hours',
                'connection_mode'=> 'connection_mode',
                'compression'=> 'compression',
		'days-of-month' => 'sched.days_of_month',
		'days-of-week' => 'sched.days_of_week'
   );
   my %param;
   foreach my $del_key ( 'source-location', 'destination-location' ) {
	delete $old_cfg->{$del_key} if exists $old_cfg->{$del_key};
   }
   foreach my $key (keys %$old_cfg) {
	my $real_key = exists $mapping{$key} ? $mapping{$key} : $key;
	if ( $real_key =~ /sched\./ ) {
		my ($k1, $k2) = split /\./, $real_key;
		$param{$k1}{$k2} = $old_cfg->{$key};
		next;
	}
	$param{$real_key} = $old_cfg->{$key};
   }
   return %param;
}

# XXX/tapl I'll implement this later.
# =head2 _sm_get_config_list

#  my $array_ref = sm_get_config_list( host=>$host,
#                                      dst=>$dst );

# Retrieves all snapmirror configurations for a given destination filer.

# It returns an array reference containing the snapmirror configuration on 
# success and dies with the error string on failure. Please see sm_get_config()
# for the full list of hash keys.

# =over

# =item host

# (Required) : Hostrec Object of filer.

# =item dst

# (Required) : Destination volume or path for the SnapMirror relationship. Must
# include the filer name, ie. 
# <filer>:<volume> or <filer>:</vol/volname/qtreename>

# =item src

# (Optional) : Source volume or path for the SnapMirror relationship. Must
# include the filer name, ie. 
# <filer>:<volume> or <filer>:</vol/volname/qtreename>


# =back

# =cut

#sub _sm_get_config_list {
#}


=head2 sm_del_config

 sm_del_config( host=>$host,
                dst=>$dst );

Deletes the snapmirror configuration for a given destination filer.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=back

=cut

sub sm_del_config {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         dst  => { type => SCALAR } } );

    _hostname_present($args{dst});

    my $conn = make_conn($args{host});

    # deletes key 'options.files.snapmirror.conf.<dst>'
    my $reg_key = "options.files.snapmirror.conf." . $args{dst};
    registryDelete($conn, $reg_key) or croak $@;
}


=head2 sm_set_connection

 sm_set_connection( host=>$host,
                    name=>$name,
                    mode=>$mode,
                    [ ips1=>$source_ip_addr1,
                      ipd1=>$dest_ip_addr1   ],
                    [ ips2=>$source_ip_addr2,
                      ipd2=>$dest_ip_addr2   ] 
                  } );

This helper function is used to specify specific network connections to 
the source volume and to allow the specification of dual paths to the 
source volume.

Dies with an error string on failure.

Example. /etc/snapmirror.conf has the following connection defintion:

 fridge-con = failover(fridge-gig,icebox-gig)(fridge-slow,icebox-slow)

 sm_set_connection(host=>'icebox-gig',
                   name=>'fridge-con',
                   mode=>'failover',
                   ips1=>'fridge-gig',
                   ipd1=>'icebox-gig',
                   ips2=>'fridge-slow',
                   ipd2=>'icebox-slow');

=over

=item host

(Required) : Hostrec Object of filer.

=item name

(Required) : This is the name of the connection you would like to define. 
This name is to be used as the source filer in relationship definitions.

=item mode

(Optional) : Specifies the mode in which two IP address pairs will be used.

=item ips1

=item ipd1

(Optional) : These are resolvable network names or IP addresses that define 
a path through the network between the source and the destination. The 
pairing denotes a path from source to destination. 'ips1' and 'ipd1' must 
be specified together.

=item ips2

=item ipd2

(Optional) : These are resolvable network names or IP addresses that define
a path through the network between the source and the destination. The 
pairing denotes a path from source to destination. 'ips2' and 'ipd2' must 
be specified together.

=back

=cut

sub sm_set_connection {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
				 name => { type => SCALAR },
				 mode => { type => SCALAR,
					   optional => 1 },
				 ips1 => { type => SCALAR,
					   optional => 1,
					   depends => ['ipd1'], },
				 ipd1 => { type => SCALAR,
					   optional => 1,
					   depends => ['ips1'], },
				 ips2 => { type => SCALAR,
					   optional => 1,
					   depends => ['ipd2'], },
				 ipd2 => { type => SCALAR,
					   optional => 1,
					   depends => ['ips2'], } } );

    my $conn = make_conn($args{host});

    # registry key
    my $reg_key = "options.files.snapmirror.connection." . $args{name};

    # set options.files.snapmirror.connection.<name>.*
    registrySet($conn, "$reg_key.mode", $args{mode}) or croak $@;

    if (exists $args{ips1} &&
	exists $args{ipd1}) {
	registrySet($conn, "$reg_key.ips1", $args{ips1}) or croak $@;
	registrySet($conn, "$reg_key.ipd1", $args{ipd1}) or croak $@;
    }

    if (exists $args{ips2} &&
	exists $args{ipd2}) {
	registrySet($conn, "$reg_key.ips2", $args{ips2}) or croak $@;
	registrySet($conn, "$reg_key.ipd2", $args{ipd2}) or croak $@;
    }
}


=head2 sm_del_connection

 sm_del_connection( host=>$host,
                    name=>$name );

Deletes the snapmirror connection definition for a given destination filer.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.


=item name

(Required) : This is the name of the connection you would like to delete. 
This name is to be used as the source filer in relationship definitions.

=back

=cut

sub sm_del_connection {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         name  => { type => SCALAR } } );

    my $conn = make_conn($args{host});

    # deletes key 'options.files.snapmirror.connection.<conn_name>.mode'
    my $reg_key = "options.files.snapmirror.connection." . 
      $args{name} . ".mode";
    registryDelete($conn, $reg_key) or croak $@;
}


=head2 sm_get_permissions

 my $array_ref = sm_get_permissions( host=>$host,
                                     [ legacy_access=>0|1 ] );

Retrieves snapmirror access permissions from snapmirror.access option.
If legacy_access option is set, the list of allowed hosts will be
pulled from the registry key 'options.files.snapmirror.allow'.

It returns an array reference containing the hostname of the allowed
filers on success.

Dies with the error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item legacy_access

(Optional) : Pull list of allowed hosts from /etc/snapmirror.allow instead
of the option 'option.snapmirror.access'. Defaults to 0.

=back

=cut

sub sm_get_permissions {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         legacy_access => { type => BOOLEAN,
						    optional => 1,
						    default => 0 } } );

    my $conn  = make_conn($args{host});
    my @hosts = ();

    my $opt = getOption($conn, "snapmirror.access");
    croak "Error getting snapmirror.access option: $@" if ($@);

    if (!$args{legacy_access}) {

	if ($opt =~ s/^host=(\S*)$/$1/) {
	    @hosts = split(/,/, $opt);
	} else {
	    push(@hosts, $opt);
	}

    } else {
	if ($opt !~ /legacy\b/i) {
	    croak "options.snapmirror.access is not set to 'legacy': $opt";
	}

	my $entry = registryWalk($conn, 
				 "options.files.snapmirror.allow"
				);
	# if a key does not exist in the registry, registryWalk() just croak
	# with an emptry string and a newline. for this function, if the key
	# doesn't exist, just return an empty arrary ref to the caller.
	if ($@ && $@ !~ /^\n$/) {
	    croak "Error getting snapmirror.allow entry: $@" ;
	}

	foreach my $key (keys %{$entry}) {
	    # options.files.snapmirror.allow.fridge:=fridge#
	    $entry->{$key} =~ s/#$//; # remove the # at the end
	    push(@hosts, $entry->{$key});
	}
    }
    return \@hosts;
}


=head2 sm_set_permissions

 sm_set_permissions( host=>$host,
                     [ allowed_host=>$hostname ],
	    	     [ legacy_access=>0|1 ] );

Gives the specified destination host permission to receive snapmirror transfer 
from the source filer.

Call this function multiple times if list of hosts need to be allowed access.
For example,

 # adds host1, host2, host3 to access list
 sm_set_permissions(host=>$srchost, allowed_host=>"host1");
 sm_set_permissions(host=>$srchost, allowed_host=>"host2");
 sm_set_permissions(host=>$srchost, allowed_host=>"host3");

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of filer.

=item allowed_host

(Optional) : Hostname (*not* Hostrec object) of allowed destination filer. Call
sm_set_permissions() multiple times if list of hosts need to be allowed 
access. Defaults to 'all'.

=item legacy_access

(Optional) : Use legacy snapmirror settings for snapmirror.access and
snapmirror.allow. Defaults to 0.

=back

=cut

sub sm_set_permissions {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
				 allowed_host => { type => SCALAR,
						   optional => 1,
						   default => 'all' },
			         legacy_access => { type => BOOLEAN,
						    optional => 1,
						    default => 0 } } );

    my $conn = make_conn($args{host});
    my $allowed_host = $args{allowed_host};

    if (!$args{legacy_access}) {
	my $hosts = sm_get_permissions(host=>$args{host});
	my $value = '';
	my @keywords = qw(all legacy none * -);

	# conditions which we do not permit appending to existing list
	# of allowed hosts:
	#
	# 1) $allowed_host is not 'all'
	# 2) $allowed_host is not 'legacy'
	# 3) existing allowed hosts has keyword 'all'
	# 4) existing allowed hosts has keyword 'legacy'
	# 5) existing allowed hosts has keyword '*'
	#
        if ( scalar @{$hosts} > 0                         &&
            !(grep { lc($allowed_host) eq $_ } @keywords) &&
            !contains($allowed_host, @{$hosts})           &&
            !(grep { contains($_, @{$hosts}) } @keywords) ) {

            push(@{$hosts}, $allowed_host);
            $value = join(',', @{$hosts});

        } elsif ( contains($allowed_host, @{$hosts}) || # Already on list
		  contains( '*',   @{$hosts} ) ||  # Already allowed
		  contains( 'all', @{$hosts} ) ) { # Already allowed
            return;
        } else {
            $value = $allowed_host;
        }

        if ( !grep { lc($value) eq $_ } @keywords ) {
	    setOption($conn, "snapmirror.access", "host=$value");
	} else {
	    setOption($conn, "snapmirror.access", "$value");
	    croak "Error setting snapmirror.access option: $@" if ($@);
	}

    } else {
	# set snapmirror legacy access
	setOption($conn, "snapmirror.access", "legacy");
	croak "Error setting snapmirror.access option: $@" if ($@);

	# update /etc/snapmirror.allow
	my $hosts = sm_get_permissions(host=>$args{host},
				       legacy_access=>$args{legacy_access});

	if (!contains($allowed_host, @{$hosts})) {
	    # bad or non-existent snapmirror.allow entry -- make a new one
	    registrySet($conn, "options.files.snapmirror.allow.$allowed_host",
			$allowed_host);
	    croak "Error setting snapmirror.allow entry: $@" if ($@);
	}
    }
}


=head2 sm_destination_snapshot

 sm_destination_snapshot( host=>$srchost,
                          src=>$src,
                          dst=>$dst );

Returns the name of the snapshot being retained for the given 
source/destination pair. Returns undef if there is no entry for
the source/destination pair. This function must be executed on the
source filer.

Note: that this function returning a snapshot name does not guarantee that 
the snapshot exists. The destinations entry hangs around until a release 
occurs, and will return the name of a deleted snapshot.

Dies with an error string on failure.

=over

=item host

(Required) : Hostrec Object of source filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=back

=cut

sub sm_destination_snapshot {
    my (%args) = validate( @_, { host => { type => OBJECT,
					   callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         src  => { type => SCALAR },
			         dst  => { type => SCALAR } } );

    _hostname_present($args{dst}, $args{src});

    my $srcpath = $args{src};
    $srcpath =~ s/^\S+:(\S+)$/$1/; # Chop off filer name

    my $src_destinations = snapmirror_destinations(host=>$args{host},
						   srcpath=>$srcpath);

    if (defined($src_destinations)) {
	foreach my $dest ( @{$src_destinations} ) {
	    if ($dest->{'destination-location'} eq $args{dst}) {
		return $dest->{'source-snapshot'};
	    }
	}
    }

    return undef;
}

=head2 sm_get_snapid

 sm_get_snapid( host=>$dsthost,
                src=>$src,
                dst=>$dst );

Returns the id of the base snapshot for the given source/destination pair.
Returns undef if there is no entry for the source/destination pair.

=over

=item host

(Required) : Hostrec Object of destination filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=back

=cut

sub sm_get_snapid {
    my (%args) = validate( @_, { host => { type => OBJECT,
                                           callbacks => {
                                               'Hostrec Object of filer' =>
                                                   \&_validate_hostrec } },
                                 src  => { type => SCALAR },
                                 dst  => { type => SCALAR } } );

    _hostname_present($args{dst}, $args{src});

    # get src base snapshot
    my $dst_status_ref = snapmirror_status(host=>$args{host},
                                       src=>$args{src},
                                       dst=>$args{dst});
    my $dst_snapshot = $dst_status_ref->{'base-snapshot'};
    my $dstconn = make_conn($args{host});
    my $dstvol = _sm_extract_volname($args{dst});
    my @status = snapshot_status($dstconn, $dstvol);
    croak $@ if ($@);
    return undef unless ( scalar( @status ) );

    my %snap_status;

    for ( @status ) {
        if ($_->{name} eq $dst_snapshot) {
            $snap_status{$dst_snapshot} = $_;
        }
    }

    my $len = scalar(keys %snap_status);
    croak "No base snapshot found for the snapmirror relation"
        if ($len < 0);
    croak "Snapshot name collision found"
        if ($len > 1);
    my $snap_id = $snap_status{$dst_snapshot}{snapid};
    return $snap_id;
}

=head2 sm_wreck_dstqt

 sm_wreck_dstqt( host=>$dsthost,
                  qtsrc=>$src,
                  qtdst=>$dst,
		  snapid=>$snapid,
		  [timeout=>$timeout]   );

Please do not call this function unless you know what 'wreck' command does.
Uses 'wreck' command to make the destination qtree read/write or read-only
based on the value of snapid. If snapid is 0, then the function first finds
out snapshot id of the base snapshot and then calls wreck on the volume so as
to make the destination qtree writeable. Upon successful completion it
returns the snapshot id.
If snapid is any other number other than 0, then wreck command is called
with that snapshot id, basically to make the destination qtree read-only once
again. Thus the destination qtree becomes read-only once again. In this case,
the function returns the same snapshot id that was passed to it just to be
consistent.
For more information, pl see: http://web.netapp.com/~arur/wreck_usage.htm

=over

=item host

(Required) : Hostrec Object of destination filer.

=item qtsrc

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:</vol/volname/qtreename>

=item qtdst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:</vol/volname/qtreename>

=item snapid

(Required) : Id of the target snapshot or 0.   If 0, will use sm_get_snapid() 
to find the base snapshot of the source/destination pair.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=back

=cut

sub sm_wreck_dstqt {
    my (%args) = validate( @_, { host => { type => OBJECT,
                                           callbacks => {
                                               'Hostrec Object of filer' =>
                                                    \&_validate_hostrec } },
                                 qtsrc =>  { type => SCALAR },
                                 qtdst =>  { type => SCALAR },
                                 snapid => { type => SCALAR,
                                             regex => qr/\d+$/ },
                                 timeout => { type => SCALAR,
                                              optional => 1,
                                              default => 300,
                                              regex => qr/^\d+$/ } } );

    my $dstconn = make_conn($args{host}); 
    my $dstpath = $args{qtdst};
    $dstpath =~ s/^\S+:\/vol\/([\w_]+)\/(\S+$)/$2/;  # Chop off filer name
    my $dstvol = $1;

    my $snapid_of_target_snapshot;
    if ($args{snapid} == 0) {
        # get the actual snapid of the target snapshot first
        $snapid_of_target_snapshot = sm_get_snapid(
				     host=>$args{host},
				     src=>$args{qtsrc},
  			             dst=>$args{qtdst});
        croak $@ if $@;
    } else {
        $snapid_of_target_snapshot = $args{snapid};
    }

    # create the wreck command
    my $cmd = "qtree_snapid snapid=$args{snapid} name=$dstpath";
    logcomment("wreck cmd: $cmd");

    # find out vol type (flex or trad)
    my $status_obj = vol_status($dstconn, $dstvol) or croak $@;
    my $vol_type   = $status_obj->type();

    # issue the wreck command
    eval { wreck($dstconn, $dstvol, type => $vol_type, inject_script => $cmd, timeout => $args{timeout}) };
    croak "$@" if $@;

    # online the destination volume
    vol_online($dstconn, $dstvol) or croak $@;

    return $snapid_of_target_snapshot;
}

=head2 sm_compare_src_dst

 sm_compare_src_dst( srchost=>$srchost,
                     dsthost=>$dsthost,
                     src=>$src,
                     dst=>$dst,
                     [ throttle=>$throttle ],
                     [ treecompare=>0|1 ],
                     [ timeout=>$timeout ] );

Helper function that will call 'snapmirror check' command if both destination
and source filers are running Tsingtao and above. Otherwise, it will call 
'treecompare' command if both destination and source filers are each running
release between Scrimshaw and Anchorsteam.

This function will perform the following actions:

1) Determines the version of both source and destination filer and type of
relationship to decide whether to use either 'snapmirror check' or 
'treecompare'.

2) If 'snapmirror check' is used, it will do data comparison. Using 
'snapmirror check' default, base snapshot for a qtree-based mirror and all 
common snapshots for a volume-based mirror will be used for comparison.

3) If 'treecompare' is used and ...

3.1) If relationship is QSM, the base snapshots on both source and destination
filers are used. Comparison will ignore ctime and mtime mismatches.

3.2) If relattionship is VSM/RSM, the base snapshot on the destination filer 
is used. Comparison will ignore vdisk metadata/times mismatches.

On failure or any data mismatches, dies with an error string or the data
mismatches stored in $@ respectively.

=over

=item srchost

(Required) : Hostrec Object of source filer.

=item dsthost

(Required) : Hostrec Object of destination filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item throttle

(Optional) : Maximum speed in kbs at which data can be transferred  
over network.

=item treecompare

(Optional) : Forces the routine to use 'treecompare' command *regardless* of
the version of either source or destination filer. Defaults to 0.

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 600 seconds.

=back

=cut

sub sm_compare_src_dst {
    my (%args) = validate( @_, { srchost => { type => OBJECT,
					      callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
				 dsthost => { type => OBJECT,
					      callbacks => {
					       'Hostrec Object of filer' =>
					           \&_validate_hostrec } },
			         src  => { type => SCALAR },
			         dst  => { type => SCALAR },
				 timeout => { type => SCALAR,
					      optional => 1,
					      default => 600,
					      regex => qr/^\d+$/ },
			         treecompare => { type => BOOLEAN,
						        optional => 1,
						        default => 0 } } );

    _hostname_present($args{dst}, $args{src});

    my $logical_mirror = 0;
    my $curr_func = (caller(0))[3];

    # check the destination host's version. that's where snapmirror check 
    # will be executing. check the source host also in case we need to
    # do treecompare on a QSM relationship. (see comment below abt treecompare
    # and QSM).
    my $dst_version = get_wafl_version($args{dsthost});
    my $src_version = get_wafl_version($args{srchost});

    # checks if $args{src} and $args{dst} are qtree paths
    if (_is_logical_mirror($args{src}, $args{dst})) {
	# QSM. it's a logical snapmirror
	$logical_mirror = 1;

    } elsif (_is_volume_mirror($args{src}, $args{dst})) {
	# VSM
	$logical_mirror = 0;

    } else {
	croak "$curr_func: Either both source and destination should be " .
	  "volumes or qtree paths";
    }

    # get dst base snapshot
    my $dst_status = snapmirror_status(host=>$args{dsthost},
				       src=>$args{src},
				       dst=>$args{dst});
    my $dst_snapshot = $dst_status->{'base-snapshot'};

    if ( !$args{treecompare}             && 
	 ($dst_version >= WV_TSINGTAO_MIN    &&
	  $src_version >= WV_TSINGTAO_MIN) ) {
	# execute snapmirror check cmd on dsthost
	my %opts;
	$opts{timeout} = $args{timeout};
	snapmirror_check(host=>$args{dsthost},
			 dst=>$args{dst},
			 src=>$args{src},
			 check_level=>'data',
			 background=>0,
			 %opts);

    } elsif ( $args{treecompare}                   || 
	      ($dst_version >= WV_SCRIMSHAW_MIN    &&
	       $src_version >= WV_SCRIMSHAW_MIN) ) {

	# check if base snapshot is in place...
	my $itr = 0;	# try to get based snapshot 4 times before forcing filers to panic
	my $src_vol=$args{src};
	$src_vol =~ s/.*://;
	my $dst_vol=$args{dst};
	$dst_vol =~ s/.*://;

	my $max_rtrs = 4;   # how many times are we trying to get base snapshot
	my $slp_time = 3;   # delay between retries. Don't forget, snap list adds time too

	while ( ! $dst_snapshot && $itr < $max_rtrs ) {
	    logcomment( "Warning, found no base snapshot, trying ". ++$itr ." of ". $max_rtrs );
	    logcomment("Source Snapshots:\n" .  
		Dumper(
		    SnapCmds::snapshot_list(
			make_conn($args{srchost}),
			$src_vol
		    )
		)
	    ); 

	    logcomment("Destination Snapshots:\n" .
		Dumper(
		    SnapCmds::snapshot_list(
			make_conn($args{srchost}),
			$src_vol
		    )
		)
	    );

	    my $dst_status = snapmirror_status(
		host=>$args{dsthost},
		src=>$args{src},
		dst=>$args{dst}
	    );

	    sleep $slp_time;

	    $dst_snapshot = $dst_status->{'base-snapshot'};
	}

	if( ! $dst_snapshot ) {
	    filer::filer_panic( make_conn($args{dsthost}) );
	    filer::filer_panic( make_conn($args{srchost}) );
	    croak "Missing base snapshot, unable to perform treecompare without it, BURT 201868.\nPanic was forced...";
	}

	# execute treecompare cmd
	#-----------------------------------------------------------------------------------------------
	my $needresync=0;
	my $waitinsync=0;
	if( $args{dst} !~ /\/vol\/.*\// ) {

		logcomment("Checking if we need to break relationship for treecompare\n");
		my ( $state, $status ) = _get_sm_stat( $args{dsthost}, $args{src}, $args{dst} );
		if( $state eq 'snapmirrored' ) {

			croak "Tree compare is not applicable when status is 'transferring'" if ( defined ($status) && $status eq 'transferring' );

			$needresync++;

			if( defined ($status) && $status =~ /sync/ ) { # in-sync, syncing ... ??? more
				$waitinsync++;
				eval { snapmirror_quiesce( host => $args{dsthost}, dstpath => $args{dst} ); };
				croak $@ if ($@);
			}

			( $state, $status ) = _get_sm_stat( $args{dsthost}, $args{src}, $args{dst} ); # we need this here!
			eval { snapmirror_break( host => $args{dsthost}, dstpath => $args{dst} ); };
			croak $@ if ($@);
		}
	}

	#-----------------------------------------------------------------------------------------------

	my $srcvol = _sm_extract_volname($args{src});
	my $dstvol = _sm_extract_volname($args{dst});

	if ($logical_mirror) {
	    # QSM
	    # grab the qtree path from the split().
	    my $srcqtpath = ( split(/:/, $args{src}) )[1];
	    my $dstqtpath = ( split(/:/, $args{dst}) )[1];

	    # now extract the volume name & qtree name
	    my ($srcvol, $srcqt) = qtreeComponents($srcqtpath) or croak $@;
	    my ($dstvol, $dstqt) = qtreeComponents($dstqtpath) or croak $@;

	    # get src base snapshot
	    my $src_snapshot = sm_destination_snapshot(host=>$args{srchost},
						       src=>$args{src},
						       dst=>$args{dst});

	    # create snappath to be of the form 
	    # /vol/<volname>/.snapshot/<snapshot>/<qtree>
	    my $src_snappath = "/vol/$srcvol/.snapshot/$src_snapshot/$srcqt";
	    my $dst_snappath = "/vol/$dstvol/.snapshot/$dst_snapshot/$dstqt";

	    my $srcconn = make_conn($args{srchost});

	    # we need to run the treecompare on source filer since the 
	    # destination would have an entry with the given source and 
	    # destination paths in the snapmirror conf file. Due to 
	    # ONTAP burt 113199 and 139519, treecompare fails by saying 
	    # destination is busy if there is an entry in the conf file.
	    treecompare($srcconn,
			$src_snappath,
			$dst_snappath,
			args=>'-c',
			skip=>["mtime"],  # ignore mtime mismatch for .
			timeout=>$args{timeout},
			filer_dst=>$args{dsthost}) or croak $@;

	} else {
	    # VSM
	    # create snappath to be of the form 
	    # /vol/<volname>/.snapshot/<snapshot>

	    my $src_snappath = "/vol/$srcvol/.snapshot/$dst_snapshot";
	    my $dst_snappath = "/vol/$dstvol/.snapshot/$dst_snapshot";

	    my $dstconn = make_conn($args{dsthost});

	    treecompare($dstconn,
			$dst_snappath,
			$src_snappath,
			args=>'-V',
			timeout=>$args{timeout},
			filer_dst=>$args{srchost}) or croak $@;
	}

	#-----------------------------------------------------------------------------------------------
	if( $args{dst} !~ /\/vol\/.*\// ) {
		if( $needresync ) {
			eval { snapmirror_resync( host => $args{'dsthost'}, src => $args{src}, dst => $args{'dst'} ); };
			croak $@ if ($@);
			my ( $state, $status ) = _get_sm_stat( $args{dsthost}, $args{src}, $args{dst} );
			if( $waitinsync ) {
				local $SIG{ALRM} = sub { croak "Status Timeout waiting for in-sync after break" };
				alarm $args{timeout};
				while( $status !~ /in-sync/ ) {
					( $state, $status ) = _get_sm_stat( $args{dsthost}, $args{src}, $args{dst} );
					sleep 5;
				}
				alarm 0;
			}
		}
	}
	#-----------------------------------------------------------------------------------------------


    } else {
	# neither 'snapmirror check' or 'treecompare' exists in earlier
	# releases
	croak "$curr_func: Both src and dst filers must be running ONTAP " .
	  "release Scrimshaw and onwards to use either 'snapmirror check' " .
	    "or 'treecompare' command.";
    }
}


=head2 sm_cleanup

 sm_cleanup( srchost=>$srchost,
             dsthost=>$dsthost,
	     src=>$src,
	     dst=>$dst,
	     [ timeout=>$timeout ] );

Removes a snapmirror relationship, regardless of the state.

=over

=item srchost

(Required) : Hostrec Object of source filer.

=item dsthost

(Required) : Hostrec Object of destination filer.

=item src

(Required) : Source volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item dst

(Required) : Destination volume or path for the SnapMirror relationship. Must
include the filer name, ie. 
<filer>:<volume> or <filer>:</vol/volname/qtreename>

=item timeout

(Optional) : Timeout for commands to complete. Defaults to 300 seconds.

=back

=cut

# SnapMirror states and actions
my $states_to_do = {
    'unknown' => sub {
	# my $self = shift;
	# logcomment( "Destination volume may be offline; trying to online ".
        #             ":$self->{'dst'}" );
	# unless ( vol_online( $self->{'dsthost'}, _sm_extract_volname(
        #          $self->{'dst'} ) ) ) {
	#     croak "Can't shutdown, unknown state";
	# }
	my $c = (caller(0))[3];
	croak "$c: Can't shutdown, unknown state";
    },
    'uninitialized' => sub { 
	# unreached
    },
    'snapmirrored' => sub {
	my $self    = shift;
	my $dstpath = (split(/:/, $self->{dst}))[1] ||
                      _sm_extract_volname( $self->{dst} );
	logcomment( "Issuing snapmirror quiesce $self->{'src'} " .
		    "-> $self->{'dst'}");
	snapmirror_quiesce(
	    host    => $self->{'dsthost'},
	    dstpath => $dstpath
	);
    },
    'quiesced' => sub {
	my $self = shift;
	my $dstpath = (split(/:/, $self->{dst}))[1] ||
                      _sm_extract_volname( $self->{dst} );
	logcomment( "Issuing snapmirror break $self->{'src'} " .
		    "-> $self->{'dst'}" );
	snapmirror_break(
	    host => $self->{'dsthost'}, dstpath => $dstpath
	);
    },
    'source' => sub {
	my $self = shift;
	my $srcpath = (split(/:/, $self->{src}))[1] ||
                      _sm_extract_volname( $self->{src} );
	logcomment( "Issuing snapmirror release $self->{'src'} " .
		    "-> $self->{'dst'}" );
	snapmirror_release(
	    host    => $self->{'srchost'},
	    srcpath => $srcpath,
	    dst     => $self->{'dst'}
	);
    },
    'broken-off' => sub {
	my $self = shift;
	my $srcpath = (split(/:/, $self->{src}))[1] ||
                      _sm_extract_volname( $self->{src} );

	my $base_sshot;

	# must be called before the release takes place
	if (_is_volume_mirror($self->{src}, $self->{dst})) {
	    # for VSM, RSM
	    $base_sshot = sm_destination_snapshot(
	        host    => $self->{'srchost'},
	        src     => $self->{'src'},
	        dst     => $self->{'dst'}
	    );

	} else {
	    # for QSM, the base snapshot on the dst is different
	    # from VSM. use 'snapmirror status' on the dst to get it
	    my $sref = snapmirror_status( host => $self->{'dsthost'}, 
					  src  => $self->{'src'}, 
					  dst  => $self->{'dst'} );
	    # XXX QSM may not have a base-snapshot
	    $base_sshot = $sref->{'base-snapshot'};
	}	    

	if (defined($base_sshot)) {
		logcomment( "Issuing snapmirror release $self->{'src'} " .
			    "-> $self->{'dst'}" );
		snapmirror_release(
		    host    => $self->{'srchost'},
		    srcpath => $srcpath,
		    dst     => $self->{'dst'}
		);
	}

	my $dstconn = make_conn( $self->{'dsthost'} );
	my $dstvol  = _sm_extract_volname( $self->{dst} );
	my $key = "status.snapmirror.active." . $self->{dst} . ".snapshot";
	my $dst_base_sshot = registryGet($dstconn, "$key") or croak $@;
	if (snapshot_exists($dstconn, $dstvol, $dst_base_sshot)) {
		logcomment( "Removing base snapshot from $self->{'dst'}" );
		snapshot_delete( $dstconn, $dstvol, $dst_base_sshot ) or croak $@;
	}
    },
};

sub _get_sm_stat {
    my ( $host, $src, $dst) = @_;
    my $sref;
    eval { $sref = snapmirror_status( host=>$host, src=>$src, dst=>$dst ) };
    return undef $@ if ( $@ =~ /non-existent relationship/ ); # Clear $@ and return undef
    croak $@ if ($@);

    my ($state,$status) = @{$sref}{'state','status'};
    croak "Weird snapmirror state: '$state'" unless (
        grep { $state } keys %{$states_to_do}
    );

    croak "Weird status: '$status'" unless (
	    $status =~ /^(idle|transferring|pending|aborting|in-sync)$/
	    ||
	    $status =~ /^(migrating|quiescing|resyncing|waiting|syncing)$/
	    ||
	    $status =~ /^(checking|idle with restart checkpoint)$/
    );
    logcomment( "current snapmirror state:status is $state:$status" );
    return ( $state, $status );
}

sub sm_cleanup {
    my (%args) = validate( @_,
	{
	    srchost => {
		type => OBJECT,
		callbacks => { 'Hostrec Object of filer' =>
                               \&_validate_hostrec }
	    },
	    dsthost => {
		type => OBJECT,
		callbacks => { 'Hostrec Object of filer' =>
                               \&_validate_hostrec }
	    },
	    src => { type => SCALAR },
	    dst => { type => SCALAR },
	    timeout => { type => SCALAR, optional => 1,
                         default => 300, regex => qr/^\d+$/ },
	},
    );
    if ( sm_get_config( host => $args{'dsthost'},
			src  => $args{'src'},
			dst  => $args{'dst'} ) ) {
	# Do this first to stop the scheduled transfers
	eval {
	    sm_del_config(
		host    => $args{'dsthost'},
		dst     => $args{'dst'}
	    );
	};
	croak $@ if ($@);  # shouldn't happen
    }
    # Abort, if currently transferring
    my ($state,$status) = _get_sm_stat( @args{'dsthost', 'src', 'dst'} );
    if ( defined ($status) && $status eq 'transferring' ) {
	logcomment( "Aborting snapmirror transfer for $args{'dsthost'}" );
	eval {
	    snapmirror_abort( host => $args{'dsthost'},
			      dst  => $args{'dst'} )
	};
	if ( $@ ) {
	    croak $@ unless (
	        # This might happen because sm_del_config shuts transfer down
		$@ =~ /destination is already idle/
	    );
	}
    }
    my $curtime = $args{'timeout'};
    my $sleep_interval = 5; # seconds
    while ( ((($state,$status) =
            _get_sm_stat(@args{'dsthost','src','dst'})),$state) ) {
	# - If status is 'idle' and state is 'uninitialized', the transfer
	# hasn't started yet.
	# - If the status is 'idle with restart checkpoint' and the state
	# is 'uninitialized',  the level 0 transfer was aborted, so exit
	# this loop.
	last if ( $status =~ /^idle with restart checkpoint/ );
        # sleep-loop until steady-state
	if ( ($status !~ /^idle/) || ($state eq 'uninitialized') ) {
	    logcomment( "snapmirror state:status for $args{'dst'} is ",
                        "$state:$status; waiting" );
            if ( ($curtime -= $sleep_interval) <= 0 ) {
    	        croak "Timeout waiting for idle status; caller is ".(caller(0))[3];
	    }
	    sleep( $sleep_interval );
	    next;
	}
	eval { &{$states_to_do->{$state}}( \%args ) };
	last if $@;
    	$curtime = $args{'timeout'};
    }
    if ( $@ ) {
        my $failstring = $@;
        logcomment( "cleanup failed in state $state for $args{'dst'}" );
        croak $failstring;
    }
}


=head1 STATUS

 Shared

=head1 ATTRIBUTES

 ONTAP_COMMANDS = snapmirror

 ONTAP_ZAPI_APIS = none

=head1 AUTHOR

 DP QA (dl-dpgqa-automation)
 $Id: Helper.pm,v 1.2 2005/09/30 22:03:10 tapl Exp $
 Last revised $Date: 2005/09/30 22:03:10 $

=head1 COPYRIGHT

 Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

# ---------------------------------------------------------------------
#
#                     EXPORTED PRIVATE FUNCTIONS
#
#            DO *NOT* IMPORT THESE FUNCTIONS INTO A TESTSCRIPT!
#
# ---------------------------------------------------------------------
# routine to extract subtest result and then repropagate it back to
# the caller if any.
# Args   : subtest object
# Returns: returns undef if there's not error string. croak with the
#          error string embedded in the subtest object.
sub _sm_extract_subtest_err {
    my ($subtest_obj) = @_;

    # grep for only the abortive errors, ignore the rest
    my @rslt_objs = grep { $_->type() =~ /^SCRIPT|CONFIG|FATAL$/i }
      $subtest_obj->own_results();

    if (scalar(@rslt_objs) == 0) {
	return;
    }

    my $worst_rslt_obj = NATE::Result::worst(@rslt_objs);

    if ($worst_rslt_obj) {
	my @err_array = $worst_rslt_obj->messages();
	my $err_str   = join("\n", @err_array);
	croak $err_str if (scalar(@err_array));
    }
}

# routine to extract volume name from either form:
# <filer>:<volume> or <filer>:/<vol>/<volname>/<qtree> 
# ... OR ...
# <volume> or /<vol>/<volname>/<qtree>
# ... OR ...
# /<vol>/<volname>
sub _sm_extract_volname {
    my ($str) = @_;

    if ($str =~ /^(\S+:)?\/vol\/([\w_]+)\/\S+$/) {
	return $2;

    } elsif ($str =~ /^(\S+:)?\/vol\/([\w_]+)$/) {
	return $2;

    } elsif ($str =~ /^(\S+:)?([\w_]+)$/) {
	return $2;

    } else {
	my $curr_func = (caller(0))[3];
	my $call_func = (caller(1))[3];
	croak "$curr_func called from $call_func: Unrecognized string: $str";
    }
}

# ---------------------------------------------------------------------
#
#                      INTERNAL PRIVATE FUNCTIONS
#
# ---------------------------------------------------------------------

sub _is_local_xfer {
    my ($srchost, $dsthost) = @_;

    return ($srchost->hostname() eq $dsthost->hostname()) ? 1 : 0;
}

# taken from QtreeCmds.pm with some modifications
my $VOL_REGEX = qr{[a-zA-Z_]\w*};
my $QTREE_REGEX = qr{[^/]*[^/\s]};  # no space at end - even though ontap 
                                    # allow but it messes up getting the name
                                    # of qtree from qtree status
my $QTREE_PATH_REGEX = qr{/vol/($VOL_REGEX)(?:/($QTREE_REGEX))?$};

sub _is_logical_mirror {
    my ($src, $dst) = @_;

    return ($src =~ $QTREE_PATH_REGEX && $dst =~ $QTREE_PATH_REGEX) ? 1 : 0;
}

sub _is_volume_mirror {
    my ($src, $dst) = @_;

    return ($src !~ $QTREE_PATH_REGEX && $dst !~ $QTREE_PATH_REGEX) ? 1 : 0;
}

# routine to check /etc/snapmirror.conf sched to see if it's
# set to 'sync'. for sm_get_config(), the keys 'minutes', 'hours',
# 'days-of-month', and 'days-of-week' will not exist if it's
# set to 'sync'.
sub _is_config_sched_sync {
    my ($config) = @_;

    if (exists $config->{'minutes'}       || 
	exists $config->{'hours'}         ||
	exists $config->{'days-of-month'} ||
	exists $config->{'days-of-week'}) {
	return 0;
    } else {
	return 1;
    }
}

# Private function to validate that the hostrec object
# is indeed a hostrec object
sub _validate_hostrec {
    my (@args) = @_;

    return (UNIVERSAL::isa($args[0], 'Hostrec')) ? 1 : 0;
}

# Second level sched params validation for sm_set_config().
# if anyone can think of a better way to do this, pls let Tap knows.
sub _validate_config_sched {
    # fudge the error msg to say validation failed in sm_set_config()
    # rather than _validate_config_sched().
    Params::Validate::validation_options(
         on_fail => 
	     sub { my @msgs = @_; chomp(@msgs);
		   @msgs = 
		     grep { s/_validate_config_sched/sm_set_config/ } @msgs;
		   Carp::confess(Tharn::script_result(@msgs)) }
    );

    # the first element in @_ contains the 2nd level params that
    # we want to validate. other elements are ignored.
    my @params = %{ $_[0] };

    validate( @params, {semi_sync => { type => BOOLEAN,
				 optional => 1,
				 default => 0 }, 
                        sync => { type => BOOLEAN,
				 optional => 1,
				 default => 0 },
			 mins  => { type => SCALAR,
				   optional => 1,
				   default => '*' },
			 hours => { type => SCALAR,
				   optional => 1,
				   default => '*' },
			 days_of_month => { type => SCALAR,
					   optional => 1,
					   default => '*' },
			 days_of_week => { type => SCALAR,
					  optional => 1,
					  default => '*' } } );

    # yahoo, we're here. return 1 to indicate success!
    return 1;
}


# internal function that does a crude check for the hostname
# in the snapmirror paths.
# Args: array of paths
# Returns 1 if all of paths have ':', 0 croaks otherwise
sub _hostname_present {
    my @entries = @_;
    my $call_func = (caller(1))[3];
    foreach my $entry (@entries) {
	if ($entry !~ /^\S+:\S+$/) {
	    croak ("$call_func: Has to be of the form " .
		   "<filer>:<volname> or " .
		   "<filer>:</vol/volname/qtree>");

	}
    }
    return 1;
}

1;
