# this is basically -*- perl -*-
#
# $Id: //depot/prod/test/main/lib/SnapMirror/SnapMirrorMonitor.pm $
#
# Copyright (c) 2009 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc
#
## @summary      Library used to monitor the current status of SnapMirror 
##               relationships
## 
## @dependencies That the relationships are pre-configured. This is not
##               to be used to do any sort of "setup."
##
## @author       dl-qa-cst
##
## @pod          here
##

=head1 NAME

SnapMirrorMonitor.pm

=head1 AUTHOR

dl-paloma-fit

=head1 DESCRIPTION

C<SnapMirrorMonitor> will handle checking the health of both LS and DP
SnapMirror Relationships. It will look for 2 main things.

1. For LS mirrors, it will make sure the destination mirrors are all in
sync by verifying that their exported-snapshot-timestamp, exported-snapshot, 
schedule, etc. all match

2. For both LS and DP mirrors it will examine the timestamp of the exported-
snapshot and compare that with the current time. If there is a difference
larger than what the configured schedule will allow, then it will notify the
calling method. 

=head1 SYNOPSIS

Library used to monitor the current status of SnapMirror relationships

    use SnapMirror::SnapMirrorMonitor;
    $SM_Monitor = new SnapMirror::SnapMirrorMonitor(
        source_vserver => $params{vserver},
        source_volume  => $params{volume},
        type           => $params{type},
    );
    if (!$SM_Monitor->fix_lagging_mirrors(
                sleep     => $Opts_Hash->{JOB_SLEEP},
                max_tries => $Opts_Hash->{JOB_RETRIES},                        
            )) {
        $Trace->warn(msg => "Unable to fix lagging mirrors");
        $Trace->exit();
        return 0;
    } else {
        $Trace->comment(msg => "Successfully fixed lagging mirrors");
    }

=head1 ARGUMENTS

=over 4

=item source_vserver
(Optional)

Useage: [source_vserver => vs0]
Description: Source Vserver of the SnapMirror relationships

=back 

=over 4

=item source_volume
(Optional)

Useage: [source_volume => $test_volume]
Description: Source Volume of the SnapMirror relationships

=back

=over 4

=item type
(Optional)

Useage: [type => "LS"]
Description: Type of SnapMirror relationship we are dealing with

=back

=over 4

=item source_path 
(Optional)

Useage: [source_path => "rtpgx-3050-5-cluster://vs0/root_vs0"]
Description: Source Path of the SnapMirror relationships

=back

=cut

package SnapMirror::SnapMirrorMonitor;

# Standard Libs

use strict;
use warnings;
use Data::Dumper;
use Switch;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Time::Local;
use Date::Parse;

# Netapp Specific Libs

use TracePkg;
use Server;

# Class Data, only accessible from this class

{
    my $refCount = 0;

    sub incRef {
        $refCount++;
    }

    sub getRef {
        return $refCount;
    }
}

# Globals

my $Trace  = new TracePkg();
my $Server = new Server();
my $self;

=head1 PUBILC METHODS 

=head2 new

To instantiate and configure the SnapMirrorMonitor object

=cut

#############################################################################
# Subroutine: new
# Parameters: None
# Returns:    Reference to this object
#############################################################################

sub new {

    $Trace->entry();
    my $class = shift(@_);

    # Check to see if this class has already been declared. If so, then we can
    # just return its reference.

    if (getRef()) {
        $Trace->exit(msg => "Package has already been declared. Returning...");
        return $self;
    }

    # If the calling method passes these options into new, they will be
    # the defaults for most of the subroutines found within this lib.
    # The options can be overridden at a later point (i.e. verifying 
    # different vservers, etc.) but this also gives them the option to 
    # set it once and not worry about it.

    my %params = validate(
        @_,
        {
            source_vserver => {
                type     => SCALAR,
                optional => 1,
                default  => "*",
            },
            source_volume => {
                type     => SCALAR,
                optional => 1,
                default  => "*",
            },
            type => {
                type     => SCALAR,
                optional => 1,
                default  => "*",
            },
            source_path => {
                type     => SCALAR,
                optional => 1,
                default  => undef,
            },
        }
    );

    # Copy specified parameters globally into object

    $self->{_source_vserver} = $params{source_vserver};
    $self->{_source_volume}  = $params{source_volume};
    $self->{_type}           = $params{type};

    # Placeholder for lagging mirrors

    $self->{_lagging_mirrors};
    
    # Defined release specific values
    my $release = $Server->get_release( format => "cooked" );
    if ($release =~ /boilermaker.0/i) {
        $self->{'_mirror-timestamp'} = 'mirror-timestamp';
    } else {
        $self->{'_mirror-timestamp'} = 'exported-snapshot-timestamp';
    }

    # Bless and return

    bless($self, $class);
    incRef();

    $Trace->exit(msg => "Initialized SnapMirror::SnapMirrorMonitor object "
                 . "with the following configuration:\n".Dumper($self));
    return $self;

} ## end sub new

=head2 health_check

This subroutine will handle examining the state of the mirrors 
and verifying that any LS sets are in sync with each other and
also checking that all sets (LS and DP) are current

=cut

#############################################################################
# Subroutine: health_check
# Parameters: Source Vserver and Source Volume
# Returns:    1 if health check was successful, 0 on error
#############################################################################

sub health_check {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            vs => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_vserver},
            },
            vol => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_volume},
            },
            type => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_type},
            },
        }
    );

    # A hash (keyed off of vserver) to hold all lagging mirrors. This will 
    # allow us to check all destinations before erroring out (in the event 
    # that there are 'unhealthy' mirrors).

    my %fix_these;

    # It is possible that the calling method could have pased in wildcard
    # characters for any of these values. We will do checks on a SnapMirror
    # relationship basis. First extract a list of vservers

    my $total_vservers = $Server->vserver_show(vserver => $params{vs});

    foreach my $vs (@{$total_vservers}) {

        my $current_vs = $vs->{vserver};
        $Trace->debug(msg => "Current Vserver: $current_vs");

        # Now extract a list of relevant volumes

        my $total_volumes = _get_source_volumes(
            vs   => $current_vs,
            vol  => $params{vol},
            type => $params{type}, 
        );

        my @lagging_vols;

        # Iterate through each relevant volume on this vserver

        foreach my $current_vol (@{$total_volumes}) {       

            $Trace->debug(msg => "Current Volume: $current_vol");

            # Save the current state (so it does not change during analysis)

            if (!save_state($self, vs => $current_vs, vol => $current_vol, 
                    type => $params{type})) {
                $Trace->warn(msg => "Unable to get current state for vserver ".
                             "$current_vs, volume $current_vol.");
                return 0;
            }

            # If we are running with LS mirrors, we will make sure they
            # are all in sync with each other. If not, push on the
            # list of mirrors we need to fix.
            
            if ($params{type} eq "LS") {
                if (!check_ls_synced($self, 
                        vs => $current_vs, vol => $current_vol)) {
                    push(@lagging_vols, $current_vol);
                    $Trace->debug(msg => "Health check failed. SnapMirrors ".
                                 "sourced from $current_vs, $current_vol, ".
                                 "of type $params{type} are not in sync with ".
                                 "each other.");
                }
            }

            # If we just found that an LS set is not synced ($current_vol
            # will be in @lagging_vols), we will skip this check. Otherwise,
            # we can now look at each destination and make sure they are 
            # up to date. If they are not up to date, we will push them onto 
            # the list of mirrors that we need to fix

            if (!grep(/^$current_vol$/, @lagging_vols)) {
                if (!check_if_current($self, skip_all_dest => 1,
                        vs => $current_vs, vol => $current_vol)) {
                    push(@lagging_vols, $current_vol);
                    $Trace->debug(msg => "Health check failed. SnapMirrors ".
                                 "sourced from $current_vs, $current_vol, of ".
                                 "type $params{type} are not current");
                }
            }
        }

        # Save the mirrors that we need to fix on a per-vserver basis

        if (scalar(@lagging_vols) > 0) {
            $fix_these{$current_vs} = \@lagging_vols;
        }
    }

    # Check results and return

    if (scalar(keys(%fix_these)) > 0) {
        $self->{_lagging_mirrors} = \%fix_these;
        $Trace->exit(msg => "Health check failed for the following SnapMirrors: ".
                     Dumper($self->{_lagging_mirrors}));
        return 0;
    }

    $Trace->exit(msg => "Health check successful");
    return 1;

} ## end sub health_check

=head2 save_state

To grab a point in time snapshot of all mirrors that have been specified to 
the library

=cut

#############################################################################
# Subroutine: save_state
# Parameters: None
# Returns:    1 if save was successful, 0 on error
#############################################################################

sub save_state {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            vs => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_vserver},
            },
            vol => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_volume},
            },
            type => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_type},
            },
        }
    );

    # Get the source-path for this particular vserver and vserver

    my $source_path = build_mirror_path($self,
        vs  => $params{vs}, 
        vol => $params{vol},
    );
    $Trace->comment(msg => "About to save state for relationship sourced from ".
                    "$source_path");

    # Issue the show for the set of defined snapmirrors

    my @sm;
    eval {
        @sm = $Server->snapmirror_show(
            'source-path' => $source_path,
            type          => $params{type},
        );
    };
    if ($@ =~ /There are no entries matching your query./) {
        $Trace->warn(msg => "Snapmirror with source path does not exist.");
        return 0;
    } elsif ($@) {
        $Trace->warn(msg => "Encountered unknown error while searching for ".
                     "snapmirror sourced from $source_path: $@");
        return 0;
    }

    my %state_hash;

    # Loop through everything that was returned from the show command and save
    # the information that we care about.

    foreach my $mirror (@sm) {
        my %mirror_hash = %{$mirror};

        # Save relevant information
        
        $state_hash{$mirror_hash{'destination-path'}}->{'source-path'} =
            $mirror_hash{'source-path'};
        $state_hash{$mirror_hash{'destination-path'}}->{'type'} =
            $mirror_hash{'type'};
        $state_hash{$mirror_hash{'destination-path'}}->{schedule} =
            $mirror_hash{schedule};
        $state_hash{$mirror_hash{'destination-path'}}->{state} =
            $mirror_hash{state};
        $state_hash{$mirror_hash{'destination-path'}}->{'exported-snapshot-timestamp'} =
            $mirror_hash{$self->{'_mirror-timestamp'}};
        $state_hash{$mirror_hash{'destination-path'}}->{status} =
            $mirror_hash{status};
        $state_hash{$mirror_hash{'destination-path'}}->{'exported-snapshot'} =
            $mirror_hash{'exported-snapshot'};
    }

    # Save it globally
    
    $self->{_current_state} = \%state_hash;

    $Trace->exit(msg => "Successfully saved state for mirrors sourced from ".
                 "source-path $source_path: ".Dumper(\%state_hash));
    return 1;

} ## end sub save_state

=head2 check_ls_synced

To check consistencies of a point in time 'snapshot' taken of the
mirror relationship from save_state()

=cut

#############################################################################
# Subroutine: check_ls_synced
# Parameters: None
# Returns:    1 if save was successful, 0 on error
#############################################################################

sub check_ls_synced {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            vs => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_vserver},
            },
            vol => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_volume},
            },
        }
    );

    # Get the source-path for this particular vserver and vserver

    my $source_path = build_mirror_path($self,
        vs  => $params{vs}, 
        vol => $params{vol},
    );

    # Extract the current_state

    my %state;
    if (defined($self->{_current_state})) {
        %state = %{$self->{_current_state}};
    } else {
        $Trace->warn(msg => "Current state for LS set sourced from $source_path ".
                     "has not been initialized yet. save_state() must be called ".
                     "on this set first. Returning...");
        return 0;
    }

    # First we should verify that there is more than one destination for this LS set.
    # When there is only one destination, we do not need to check if they are in sync.

    if (scalar(keys(%state)) <= 1) {
        $Trace->exit(msg => "There is only one destination for the SnapMirrorr ".
                     "relationship sourced from $source_path. Returning success.");
        return 1;
    }

    # Loop through the destinations (of this particular source) and check for
    # consistencies. 

    my %baseline;
    my %errors;

    # Call out what we are going to check. Checking the "status" will be skipped
    # as some updates can complete before others in the set.

    my @check_these = qw/exported-snapshot
                         state
                         schedule/;

    foreach my $destination (keys(%state)) {
        
        # Pull out all of the relevant SnapMirror information for this particular
        # destination.

        my %sm_state = %{$state{$destination}};

        # Set the baseline if we have not yet already (done on first iteration)

        if (scalar(keys(%baseline)) < 1 ) {

            $baseline{'exported-snapshot-timestamp'}  = $sm_state{$self->{'_mirror-timestamp'}};
            $baseline{'exported-snapshot'}            = $sm_state{'exported-snapshot'};
            $baseline{'status'}                       = $sm_state{'status'};
            $baseline{'state'}                        = $sm_state{'state'};
            $baseline{'schedule'}                     = $sm_state{'schedule'};
            $Trace->comment(msg => "Destination $destination used for baseline: ".
                            Dumper(\%baseline));

            # We can continue. No need to compare the baseline to itself

            next;
        }

        # Compare the baseline to the current destination

        my $error_str = "";
        foreach my $key (@check_these) {

            if ($baseline{$key} ne $sm_state{$key}) {
                $error_str .= "Found mismatch while comparing baseline $key ".
                              "($baseline{$key}) to current destination ".
                              "$destination $key ($sm_state{$key}).\n";
            }
        }

        # Append any error conditions if we found any

        if ($error_str ne "") {
            $errors{$destination} = $error_str;
        }
    }

    # Check for errors

    if (scalar(keys(%errors)) > 0 ) {
        foreach my $key (keys(%errors)) {
            $Trace->warn(msg => "Destination $key is out of sync. ".
                         "Details: $errors{$key}");
        }
        $Trace->exit(msg => "Relationship sourced from $source_path is not ".
                     "in sync.");
        return 0;
    }

    $Trace->exit(msg => "Relationship sourced from $source_path is in sync");
    return 1;

} ## end sub check_ls_synced

=head2 check_if_current

To make sure that the specified LS or DP sets are up to date

=cut

#############################################################################
# Subroutine: check_if_current
# Parameters: None
# Returns:    1 if save was successful, 0 on error
#############################################################################

sub check_if_current {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            vs => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_vserver},
            },
            vol => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_volume},
            },
            skip_all_dest => {
                type     => SCALAR,
                optional => 1,
                default  => 0,
            },
        }
    );

    # Get the source-path for this particular vserver and vserver

    my $source_path = build_mirror_path($self,
        vs  => $params{vs}, 
        vol => $params{vol},
    );

    # Extract the current_state

    my %state;
    if (defined($self->{_current_state})) {
        %state = %{$self->{_current_state}};
    } else {
        $Trace->warn(msg => "Current state for LS set sourced from $source_path ".
                     "has not been initialized yet. save_state() must be called ".
                     "on this set first. Returning...");
        return 0;
    }

    # Loop through each destination, grab it's schedule and make sure that is 
    # is up to date.

    my @lagging_destinations;
    my $difference = 0;
    foreach my $destination (keys(%state)) {

        # Pull out all of the relevant SnapMirror information for this 
        # particular destination.

        my %sm_state = %{$state{$destination}};

        # First make sure we are not in an unknown state

        if (($sm_state{'exported-snapshot-timestamp'} eq "-") || ($sm_state{state} eq "Unknown") ||
                ($sm_state{'exported-snapshot'} eq "-")) {
            $Trace->comment(msg => "Destination: $destination is in an Unknown state. ".
                            "Reporting as if it is lagging.");
            push(@lagging_destinations, $destination);
            next;
        }

        # Grab the current time (epoch)

        my $current_time = time();

        # Calculate the timestamp of the exported snapshot

        my $es_epoch = _calculate_exported_snapshot_epoch( 
            snapshot => $sm_state{'exported-snapshot'},
        );
        if ($es_epoch == 0) {
            $Trace->exit(msg => "Unable to obtain the epoch for the exported-".
                         "snapshot. Returning.");
            return 0;
        }

        # Find the number of seconds between scheduled updates

        my $last_update = _calculate_last_update(
            schedule => $sm_state{'schedule'},
        );
        if ($last_update == 0) {
            $Trace->exit(msg => "Unable to calculate the number of seconds since ".
                         "the last update should have occurred");
            return 0;
        }

        # Take the difference. The number of seconds in between the current
        # time and the timestamp of the exported-snapshot should be less than
        # 2x the schedule (in seconds). Allowing for 2x's in case we are 
        # currently in the middle of a transfer (the exported snapshot may not
        # have been updated yet, but that does not mean we are behind. 

        $difference = $current_time - $es_epoch;
        my $mins = $difference/60;

        if ($difference > (2*$last_update)) {
            $Trace->comment(msg => "Destination $destination is $difference seconds ".
                            "($mins minutes) behind the current time: ".
                            localtime($current_time).". The last update should have ".
                            "occurred within $last_update seconds (and we gave it ".
                            2*$last_update." seconds). Marking this destination as ".
                            "lagging.");
            push(@lagging_destinations, $destination);
        }

        # If we are dealing with an LS set (remember, this is called one per source-path),
        # then we do not need to check every destination. If the calling method already 
        # made sure that they are all in sync (via check_ls_synced - as health_check()
        # does), then we can exit out here to save time because if one is up to date
        # or behind, then they all are.

        if (($sm_state{'type'} eq "LS") && ($params{skip_all_dest})) {
            $Trace->comment(msg => "Skip_all_dest was flagged. Since this is ".
                            "an LS set, we are only checking one destination");
            last;
        }
    }

    # Check for errors

    if (scalar(@lagging_destinations > 0)) {
        my $dest_str = join(", ", @lagging_destinations);
        $Trace->warn(msg => "Mirrors sourced from $source_path, destined to: ".
                     "$dest_str are behind schedule. See above for details.");
        return 0;        
    }

    # Return on success

    $Trace->exit(msg => "Relationship sourced from $source_path is current");
    return 1;

} ## end sub check_if_current

=head2 build_mirror_path

Quick utility to generate the path (which is used as a key for accessing/saving 
snapmirror data)

=cut

#############################################################################
# Subroutine: build_mirror_path
# Parameters: Source vserver and volume
# Returns:    Full name of the source/destination paths
# Note:       Assumes all relationships are intra-cluster - current 
#             functionality for BR.0
#############################################################################

sub build_mirror_path {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            vs => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_vserver},
            },
            vol => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_volume},
            },
        }
    );

    # Get the cluster name (again, assumes intra-cluster snapmirror relationships)
    # and save it for future calls.

    if (!defined($self->{_cluster_name})) {
        my @cluster = $Server->cluster_identity_show();
        $self->{_cluster_name} = $cluster[0]{name};
    }

    # This is not something that is saved globally because we can have multiple
    # source paths that we are keeping tabs on.
    
    my $path = "$self->{_cluster_name}://$params{vs}/$params{vol}";
    
    # Return

    $Trace->exit(msg => "Returning with path: $path");
    return $path;
   
} ## end sub build_mirror_path

=head2 get_job_id

Quick procedure to parse out a job ID from the SnapMirror command
output

=cut

#############################################################################
# Subroutine: get_job_id
# Parameters: Output from a snapmirror command
# Returns:    Job ID
#############################################################################

sub get_job_id {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            output => {
                type     => SCALAR,
                optional => 0,
            },
        }
    );

    my $id;

    if ($params{output} =~ m/^\[Job (\d+)\] .+/) {
        $id = $1;
    } else {
        $Trace->fatal(msg => "Invalid job output: $params{output}");
    }

    $Trace->exit(msg => "Returning with Job ID $id");
    return $id;

} ## end sub get_job_id

=head2 wait_on_snapmirror_job

Subroutine to handle checking the job status until it has completed

=cut

#############################################################################
# Subroutine: wait_on_snapmirror_job
# Parameters: Output from a snapmirror initialize/update command, OR a list
#             of job IDs to monitor. The calling method must specify one or
#             the other.
# Returns:    1 if the job(s) was successful, 0 on failure
#############################################################################

sub wait_on_snapmirror_job {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            output => {
                type     => SCALAR,
                optional => 1,
            },
            job_ids => {
                type     => ARRAYREF,
                optional => 1,
            },
            sleep => {
                type     => SCALAR,
                optional => 1,
                default  => 60,
            },
            max_tries => {
                type     => SCALAR,
                optional => 1,
                default  => 500,
            },
        }
    );

    # Array to hold all of the jobs we are waiting on

    my @job_ids;

    # If the calling method specified output, we will parse out the
    # job id

    if (defined($params{output})) {

        my $id = get_job_id($self, output => $params{output});
        push(@job_ids, $id);

    } elsif (defined($params{job_ids})) {
        
        # Otherwise, we will append the job ids that have already been
        # specified

        push(@job_ids, @{$params{job_ids}});

    } else {
        $Trace->fatal(msg => "The calling method must either specific ".
                      "\"output\" from which to parse the job ids, or ".
                      "an array containing the job ids (\"job_ids\") ".
                      "themselves. Neither were definied.");
    }

    # Save the job ids for future reference

    my @check_these = @job_ids;

    # Loop until the job is completed, or until we have hit the max tries

    my $output;
    my $current_id;
    my $max_tries = 0;

    # List of any jobs that may have failed

    my @failed_jobs;

    while (scalar(@check_these) > 0) {

        my $current_id = shift(@check_these);

        # Did we hit max_tries yet?
        
        if ($max_tries++ > $params{max_tries}) {
            $Trace->warn(msg => "We have reached the max tries ($params{max_tries}) ".
                         "before the following SnapMirror jobs have completed: ".
                         Dumper(\@check_these)." Returning with failure");
            return 0;
        }

        # Check the job

        eval {
            $output = $Server->job_show(id => $current_id);
        };

        # Case where job has completed and moved into the job history

        if ($@ =~ /There are no entries matching your query/) {

            $Trace->debug(msg => "Job completed, checking job history");
            $output = $Server->job_history_show(id => $current_id);

            # Now that the job has finished, we can check if it succeed

            if (!_did_job_succeed(completion => $output->[0]{completion})) {
                push(@failed_jobs, $current_id);
                next;
            } 

        } elsif ($@) {
            $Trace->fatal(msg => "Encountered unexpected error while checking ".
                          "job status: $@");
        } else {

            # Case where job is either still running or has just finished (and
            # as not yet moved into the job history).
    
            if ($output->[0]{state} =~ m/Running|Queued|Unknown/g) {
    
                $Trace->comment(msg => "SnapMirror Job [$current_id] is still ".
                                "running. Sleeping for $params{sleep} seconds");
                Tharn::snooze($params{sleep});
                push(@check_these, $current_id);
    
            } else {
                $Trace->comment(msg => "SnapMirror Job [$current_id] completed with ".
                                "state $output->[0]{state}");
    
                # Since the job has finished, we can compare the job completion string
    
                if (!_did_job_succeed(completion => $output->[0]{completion})) {
                    push(@failed_jobs, $current_id);                
                }
            }
        }

    } ## end while

    if (scalar(@failed_jobs > 0)) {

        $Trace->exit(msg => "SnapMirror Job(s): ".Dumper(\@failed_jobs).
                     " did not complete as expected.");
        return 0;
        
    } else {

        $Trace->exit(msg => "SnapMirror Job(s): ".Dumper(\@job_ids).
                     " completed as expected.");
        return 1;
    }

} ## end sub wait_on_snapmirror_job

=head2 fix_lagging_mirrors

This procedure will handle the manual steps needed to bring an LS relationship 
back in sync and up to date

=cut

#############################################################################
# Subroutine: fix_lagging_mirrors
# Parameters: Source Vserver and Source Volume
# Returns:    1 if the job was successful, 0 on failure
# NOTE:       This really should only be called for LS sets, but it will also
#             bring DP sets more up to date as well.
#############################################################################

sub fix_lagging_mirrors {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            vs => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_vserver},
            },
            vol => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_volume},
            },
            type => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_type},
            },
            sleep => {
                type     => SCALAR,
                optional => 1,
                default  => 60,
            },
            max_tries => {
                type     => SCALAR,
                optional => 1,
                default  => 500,
            },
            use_saved_mirrors => {
                type     => SCALAR,
                optional => 1,
                default  => 0,
            },
        }
    );

    # If the calling method specified to use_saved_mirrors, we will pull
    # the information out of _lagging_mirrors. Otherwise, we will query
    # for the vserver and volumes as normal.

    my @total_vservers;
    my %fix_these = %{$self->{_lagging_mirrors}};
    
    if ($params{use_saved_mirrors}) {
        @total_vservers = keys(%fix_these);
    } else {

        # It is possible that the calling method could have pased in wildcard
        # characters for any of these values. We will do checks on a SnapMirror
        # relationship basis. First extract a list of vservers

        my $tmp_vservers = $Server->vserver_show(vserver => $params{vs});
        foreach my $vs (@{$tmp_vservers}) {
            push(@total_vservers, $vs->{vserver});
        }
    }

    foreach my $current_vs (@total_vservers) {

        $Trace->debug(msg => "Current Vserver: $current_vs");

        # Now grab a list of the volumes

        my @total_volumes;
        if ($params{use_saved_mirrors}) {
            @total_volumes = @{$fix_these{$current_vs}};
        } else {
            @total_volumes = @{_get_source_volumes(
                vs   => $current_vs,
                vol  => $params{vol},
                type => $params{type}, 
            )};  
        }

        # Begin iterations through the various combinations

        foreach my $current_vol (@total_volumes) {       

            $Trace->debug(msg => "Current Volume: $current_vol");

            # Get the source-path for this particular vserver and vserver

            my $source_path = build_mirror_path($self,
                vs  => $current_vs, 
                vol => $current_vol,
            );

            # Save the current state. When we re-enable the schedules, we will need
            # to know which schedule was saved for each destination (in case we have
            # a mix of DP and LS destinations for this source volume).

            if (!save_state($self, vs => $current_vs, vol => $current_vol)) {
                $Trace->warn(msg => "Unable to get current state for vserver ".
                             "$current_vs, volume $current_vol.");
                return 0;
            }

            # Disable the schedule (to disable the schedule, set it to the empty
            # string).

            $Server->snapmirror_modify(
                'source-path'      => $source_path,
                'destination-path' => "*",
                schedule           => "\"\"",
            );

            # Grab the current state (since the save was done before we disabled
            # the schedules, the current state will have the old schedules (which
            # is what we want).

            my %sm = %{$self->{_current_state}};

            # First we will abort any scheduled tranfers so that we can issue
            # ours manually. Ignore any errors - we will get errors in cases
            # where we are aborting tranfers that have just finished.
            # Destinations need to be aborted on an individual basis.

            my $did_abort = 0;
            foreach my $destination (keys(%sm)) {

                # If there is an active transfer at the time we check, will
                # will try to abort it 

                if ($sm{$destination}->{status} eq "Transfer") {

                    # Aborts are run in the foreground
                    
                    eval {
                        $Server->snapmirror_abort(
                            'source-path'      => $source_path,
                            'destination-path' => $destination,
                        );
                        $did_abort = 1;
                    };
                    if ($@ =~ /no transfer to abort/) {
                        $Trace->comment(msg => "Transfer completed before it ".
                                        "could be aborted.");
                    } elsif ($@) {
                        $Trace->warn(msg => "SnapMirror abort failed unexpectedly ".
                                   "with: $@");
                        return 0;
                    } else {
                        $Trace->comment(msg => "SnapMirror abort for destination ".
                                        "$destination succeeded");
                    }
                }
            }

            # If we ran any aborts, we will sleep to allow transfers to cease

            if ($did_abort) {
                Tharn::snooze(30);
                $Server->snapmirror_show('source-path' => $source_path);
            }

            # If we are dealing with a load-sharing set, we will issue an 
            # update-ls-set here to bring the entire set up to date

            my $output;
            if ($params{type} eq "LS") {

                $output = $Server->snapmirror_update_ls_set(
                    'source-path' => $source_path,
                );

                # Wait for the Job to finish. Since we know that we have lagging
                # mirrors here, update-ls-set will fail. We will simply wait for
                # it to continue and move on. 

                wait_on_snapmirror_job($self,
                    output    => $output,
                    sleep     => $params{sleep},
                    max_tries => $params{max_tries},
                );
                $Trace->comment(msg => "Update-ls-set sourced from $source_path ".
                                "completed. Will now issue individual updates to ".
                                "recover from errors during update-ls-set");
            }

            # Remember, we are pulling the schedule from the saved state 
            # (_current_state). Not from "snapmirror show". That would already 
            # have the schedule turned off.

            my $schedule = undef;
            my @job_ids;

            # Fire off an individual update on each destination in parallel.
            # Will return before the job is complete

            foreach my $destination (keys(%sm)) {

                # Save the schedule if we have not already (first iteration).

                if (!defined($schedule)) {
                    $schedule = $sm{$destination}->{schedule};
                }

                # Individual update. This will fail for any mirrors that
                # are current, so we will ignore errors and continue

                $output = $Server->snapmirror_update(
                    'source-path'      => $source_path,
                    'destination-path' => $destination,
                );
                push(@job_ids, get_job_id($self, output => $output));
            }

            wait_on_snapmirror_job($self,
                job_ids => \@job_ids,
                sleep     => $params{sleep},
                max_tries => $params{max_tries},
            );
            $Trace->comment(msg => "SnapMirror update completed as shown ".
                            "above for all destinations.");

            # Cannot re-enable the schedules one by one because some updates will
            # start before all lagging mirrors are fixed.

            $Server->snapmirror_modify(
                'source-path'      => $source_path,
                'destination-path' => "*",
                schedule           => $schedule,
            );

            # Issuing one last show for any future debugging

            $Trace->comment(msg => "SnapMirror Set sourced from $source_path ".
                            "has been updated.");
            $Server->snapmirror_show('source-path' => $source_path);

        } # end foreach my $current_vol
    } # end foreach my $vs

    $Trace->exit(msg => "Lagging Mirrors have been fixed for Vserver: $params{vs}, ".
                 "Volume: $params{vol}");
    return 1;

} ## end sub fix_lagging_mirrors

=head2 verify_volume_attrs

Subroutine to verify that volume attributes are successfully replicated from the source
volume to all of the destination volumes

=cut

#############################################################################
# Subroutine: verify_volume_attrs
# Parameters: Source Vserver and Source Volume
# Returns:    1 if the comparison was successful, 0 on failure
#############################################################################

sub verify_volume_attrs {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            vs => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_vserver},
            },
            vol => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_volume},
            },
            type => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_type},
            },
        }
    );

    # A list of volume attributes that we will compare. We are basically checking
    # the ones that take part in volume configuration replication, as well as a 
    # few others.
    #  This is the volume replication table for volume attributes in different 
    #  mirroring scenarios, as of 06/18/2010 (see burt426639).
    #  *                                     VolMove   VolCopy   DPMirror   LSMirror
    #  * noupdate_atime                         X        X                     X
    #  * min_read_ahead                         X        X                     X
    #  * no_snapdir                             X        X                     X
    #  * snap_reserve_pct                       X        X                     X
    #  * maxdirsize                             X        X          X          X
    #  * files-total (maxfiles)                 X        X          X          X
    #  * security-style (striped vol only)      X        X          X          X

    my @check_these_attrs;
    if ($params{type} =~ /LS/i) {
        @check_these_attrs = qw /
            state
            volume-style
            policy
            language
            security-style
            min-readahead
            atime-update
            snapdir-access
            files
            maxdir-size
            percent-snapshot-space
        /;
    } elsif ($params{type} =~ /DP/i) {
        @check_these_attrs = qw /
            state
            volume-style
            policy
            language
            security-style
            files
            maxdir-size
        /;
    } else {
        $Trace->warn(msg => "Invalid mirror type: ".$params{type});
        return 0;
    }

    # Get a list of destination volumes for this particular snapmirror 
    # relationship

    my @destination_vols = @{_get_destination_volumes(
        vs   => $params{vs},
        vol  => $params{vol},
        type => $params{type},
    )};

    # Issue the volume show on the source volume. All destinations will be 
    # compared to this

    my $source_vol = $Server->volume_show(
        vserver => $params{vs},
        volume  => $params{vol},
    );

    # Iterate through each destination and compare
    
    my @errors;
    foreach my $destination (@destination_vols) {

        # Get the destination volume attributes
        
        my $dest_vol = $Server->volume_show(
            vserver => $params{vs},
            volume  => $destination,
        );       

        # Compare

        foreach my $attr (@check_these_attrs) {
            if ($source_vol->[0]{$attr} ne $dest_vol->[0]{$attr}) {
                push(@errors, "Found mismatch on destination $destination, ".
                     "Attribute: $attr. Source value: $source_vol->[0]{$attr},".
                     " Destination Value: $dest_vol->[0]{$attr}.");
            }
        }
    }

    if (scalar(@errors)) {
        $Trace->warn(msg => Dumper(\@errors));
        $Trace->exit();
        return 0;
    }

    # Exit

    $Trace->exit(msg => "All destinations successfully match the source volume");
    return 1;

} ## end sub verify_volume_attrs

=head2 run_wafliron

Subroutine to run wafliron on a single aggregate specificed by the calling
method. This will also take care of monitoring the health (both before and after
running wafliron) of any LS or DP sets that contain destination volumes on the 
aggregate that is being ironed. It will also resync any Broken-off DP sets and 
verify everything was cleaned up in vreport.

=cut

#############################################################################
# Subroutine: run_wafliron
# Parameters: Aggregate(s) to be ironed
# Returns:    1 if the iron and iron verification was successful, 0 on 
#             failure
#############################################################################

sub run_wafliron {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize parameters

    my %params = validate(
        @_,
        {
            aggregate => {
                type     => SCALAR,
                optional => 0,
            },
            wafliron_sleep => {
                type     => SCALAR,
                optional => 1,
                default  => 30,
            },
            max_tries => {
                type     => SCALAR,
                optional => 1,
                default  => 500,
            },
            skip_health_check => {
                type     => SCALAR,
                optional => 1,
                default  => 0,
            },
        }
    );

    # Get a list of all load sharing mirrors hosted on this aggregate so that
    # post-wafliron, we can make sure that these sets were not affected.

    my $ls_volumes;
    eval {
        $ls_volumes = $Server->volume_show(
            aggregate => $params{aggregate},
            type      => "LS",
        );
    };
    if ($@ =~ /There are no entries matching your query./) {
        $Trace->comment(msg => "Aggregate $params{aggregate} does not contain ".
                        "any Load Sharing volumes");
    } elsif ($@) {
        $Trace->exit(msg => "Received unexpected error message while querying ".
                     "for LS volumes on aggregate $params{aggregate}: $@");
        return 0;
    }
    
    # Get a separate list of all data protection mirrors hosted on this aggr
    # so that we can re-sync the relationships post-wafliron.

    my $dp_volumes;
    eval {
        $dp_volumes = $Server->volume_show(
            aggregate => $params{aggregate},
            type      => "DP",
        );
    };
    if ($@ =~ /There are no entries matching your query./) {
        $Trace->comment(msg => "Aggregate $params{aggregate} does not contain ".
                        "any Data Protection volumes");
    } elsif ($@) {
        $Trace->exit(msg => "Received unexpected error message while querying ".
                     "for DP volumes on aggregate $params{aggregate}: $@");
        return 0;
    }

    # For the initial health check, we can have these two lists combined since
    # we are performing the same actions for both. We do need to keep them
    # seperate so that only DP destinations will be resynced after running 
    # wafliron.

    my @all_destinations;
    if (defined($ls_volumes)) {
        push(@all_destinations, @{$ls_volumes});
    }
    if (defined($dp_volumes)) {
        push(@all_destinations, @{$dp_volumes});
    }

    # We will keep a list of source volumes that we have already checked so
    # that we can avoid doing that same check multiple times. This will help us
    # in the situation whereby we have multiple destination volumes that belong 
    # to the same source volume on the aggregate that we are ironing. Since we
    # have to worry about vserver AND volume, this will be hash keyed off of a
    # vserver holding the volumes already checked for that vserver. 

    my %sources_checked;

    # Verify that all LS and DP relationships are currently healthy before 
    # running wafliron. Also check that vreport does not have anything 
    # logged for this destination vserver/volume combination.

    foreach my $dest_vol (@all_destinations) {

        # Vreport

        my $object = $dest_vol->{vserver}.":".$dest_vol->{volume};
        eval {
            $Server->debug_vreport_show(
                type   => "volume",
                object => $object,
            );
        };
        if ($@ =~ /There are no entries matching your query./) {
            $Trace->debug(msg => "Vreport did not find descrepencies for object: ".
                          "$object prior to running wafliron as expected");
        } elsif ($@) {
            $Trace->exit(msg => "Received unexpected error message while running ".
                         "vreport for object: $object: $@");
            return 0;
        } else {
            $Trace->exit(msg => "Vreport found descrepencies for object: $object ".
                          "prior to running wafliron. Exiting.");
            return 0;
        }

        # Only run the health check if flagged to

        if (!$params{skip_health_check}) {
        
            # Get the source volume so that we can run the health check from there

            my $sm_source;
            eval {
                $sm_source = $Server->snapmirror_show(
                    'destination-vserver' => $dest_vol->{vserver},
                    'destination-volume'  => $dest_vol->{volume},
                );
            };
            if ($@ =~ /There are no entries matching your query./) {
                $Trace->comment(msg => "Volume $dest_vol->{volume} in vserver ".
                                "$dest_vol->{vserver} is a $dest_vol->{type} volume, ".
                                "but is not configured to be part of a SnapMirror ".
                                "relationship. Skipping it...");
                next;
            } elsif ($@) {
                $Trace->warn(msg => "Received unexpected result when querying for ".
                             "$dest_vol->{volume} on $dest_vol->{vserver}: $@");
                return 0;
            }
            
            if (scalar(@{$sm_source}) > 1) {
                $Trace->warn(msg => "Only one entry should be returned when querying ".
                             "for a SnapMirror set via its destination volume & ".
                             "vserver. Found: ".scalar(@{$sm_source}));
                return 0;
            }

            # Check if we've already called heath check for this source volume.
            # If not, run health check and add it to the list so we don't waste
            # time checking it again. Also handle the case where this is the first
            # source volume for this particular vserver. 

            my @vols_checked;
            if (defined($sources_checked{$sm_source->[0]{'source-vserver'}})) {
                @vols_checked = @{$sources_checked{$sm_source->[0]{'source-vserver'}}};
            }
            
            if (!grep(/$sm_source->[0]{'source-volume'}/, @vols_checked)) {

                # SnapMirror Health Check
                
                if (!health_check($self,
                            vs   => $sm_source->[0]{'source-vserver'},
                            vol  => $sm_source->[0]{'source-volume'},
                            type => $sm_source->[0]{'type'},
                        )) {
                    $Trace->warn(msg=> "$sm_source->[0]{'type'} Set destined to ".
                                 "$dest_vol->{volume} on vserver ".
                                 "$dest_vol->{vserver} is not healthy. ".
                                 "Returning before running wafliron on aggregate ".
                                 "$params{aggregate}.");
                    $Trace->exit();
                    return 0;
                }
            
                # Add it to the list of source volumes that we have already checked

                push(@vols_checked, $sm_source->[0]{'source-volume'});
                $sources_checked{$sm_source->[0]{'source-vserver'}} = \@vols_checked;
            }

            $Trace->comment(msg => "SnapMirror set destined to $dest_vol->{vserver}, ".
                            "$dest_vol->{volume} is heathly before running wafliron.")

        } ## end if (!$params{skip_health_check})
    }

    # Run wafliron and wait for completion (or until we've hit the max_tries).

    $Server->storage_aggregate_wafliron_start(
        aggregate => $params{aggregate},
        'include-mirrors' => "true",
    );
 
    my $running   = 1;    
    my $iteration = 0;
    my $wafliron_status;

    while ($running) {

        # Check our breakout case

        if ($iteration++ > $params{max_tries}) {
            $Trace->exit(msg => "Wafliron did not complete after checking $iteration ".
                         "times with $params{wafliron_sleep} seconds in between.");
            return 0;
        }
        
        # Issue the show

        $wafliron_status = $Server->storage_aggregate_wafliron_show(
            aggregate => $params{aggregate},
        );

        # Copmare the results

        if ($wafliron_status->[0]{'iron-status'} eq "not_running") {
            $Trace->comment(msg => "Wafliron is no longer running for aggregate ".
                            "$params{aggregate}");
            $running = 0;
        } else {
            $Trace->comment(msg => "Wafliron is still running for aggregate ".
                            "$params{aggregate}. Sleeping...");
            Tharn::snooze($params{wafliron_sleep});
        }

    } ## end while

    # Verify success

    if ($wafliron_status->[0]{'iron-last-start-errno'} !~ /^0$/ || 
            $wafliron_status->[0]{'iron-last-start-error-info'} !~ /^Successful.$/) {
        $Trace->exit(msg => "Wafliron did not complete successfully. Last Start Error ".
                     "Number was: $wafliron_status->[0]{'iron-last-start-errno'} ".
                     "(expected: 0), Last Start Error Info was: ". 
                     $wafliron_status->[0]{'iron-last-start-error-info'}.
                     "(expected: Successful.)");
        return 0;
    }

    # Ultimately, we want to resync all of the DP relationships that wafliron had
    # just broken-off, even if we find errors. We will use the following to hold
    # any errors encountered along the way. This will allow us to resync all 
    # relationships that we broke before erroring out.

    my @errors_resyncing;

    # Resync the DP relationships (wafliron had to break them) and check
    # their health

    foreach my $dest_vol (@{$dp_volumes}) {

        # Skip it if it was just a DP volume out there that was not configured
        # to be part of a SnapMirror relationship.

        my $sm_source;
        eval {
            $sm_source = $Server->snapmirror_show(
                'destination-vserver' => $dest_vol->{vserver},
                'destination-volume'  => $dest_vol->{volume},
            );
        };
        if ($@ =~ /There are no entries matching your query./) {
            $Trace->comment(msg => "Volume $dest_vol->{volume} in vserver ".
                            "$dest_vol->{vserver} is a $dest_vol->{type} volume, ".
                            "but is not configured to be part of a SnapMirror ".
                            "relationship. Skipping it...");
            next;
        } elsif ($@) {
            $Trace->warn(msg => "Received unexpected result when querying for ".
                         "$dest_vol->{volume} on $dest_vol->{vserver}: $@");
            $Trace->exit();
            return 0;
        }

        # First make sure that the state is "Broken-off"

        if ($sm_source->[0]{state} ne "Broken-off") {
            push(@errors_resyncing, "Expected DP relationship destined to: ".
                 "$dest_vol->{vserver},$dest_vol->{volume} to be Broken-off. ".
                 "Found that it was $sm_source->[0]{state}");
            next;
        }

        # Resync

        my $dest_path = build_mirror_path($self,
            vs  => $dest_vol->{vserver}, 
            vol => $dest_vol->{volume},
        );        
        my $output = $Server->snapmirror_resync(
            'source-path'      => $sm_source->[0]{'source-path'},
            'destination-path' => $dest_path,
        );
        if (!wait_on_snapmirror_job($self, output => $output)) {
            push(@errors_resyncing, "SnapMirror resync faild for DP set sourced from ".
                 "$sm_source->[0]{'source-path'}.");
            next;
        } else {
            $Trace->comment(msg => "SnapMirror resync succeeded for DP set sourced from ".
                            "$sm_source->[0]{'source-path'}.");            
        }

        # SnapMirror Health Check. This time we do want to run health_check for
        # each destination volume (because we are resyncing).

        if (!$params{skip_health_check}) {

            if (!health_check($self,
                        vs   => $sm_source->[0]{'source-vserver'},
                        vol  => $sm_source->[0]{'source-volume'},
                        type => $sm_source->[0]{'type'},
                    )) {
                push(@errors_resyncing, "DP Set destined to $dest_vol->{volume} on ".
                     "vserver $dest_vol->{vserver} is not healthy ".
                     "after running wafliron.");
                next;
            }
        }

        # Vreport

        my $object = $dest_vol->{vserver}.":".$dest_vol->{volume};
        eval {
            $Server->debug_vreport_show(
                type   => "volume",
                object => $object,
            );
        };
        if ($@ =~ /There are no entries matching your query./) {
            $Trace->debug(msg => "Vreport did not find descrepencies for object: ".
                          "$object after resyncing DP set as expected");
        } elsif ($@) {
            push(@errors_resyncing, "Received unexpected error message while ".
                 "running vreport for object: $object: $@");
        } else {
            push(@errors_resyncing, "Vreport found descrepencies for object: $object ".
                 "prior to running wafliron.");
        }

    } ## end foreach my $dest_vol

    if (scalar(@errors_resyncing)) {
        $Trace->warn(msg => "Encountered the following errors while trying to ".
                     "resync the Broken-off DP sets on aggregate $params{aggregate} ".
                     "after running wafliron: ".Dumper(\@errors_resyncing));
        $Trace->exit();
        return 0;
    }
    $Trace->comment(msg => "Broken-off DP SnapMirrors hosted on aggregate ".
                    "$params{aggregate} have been successfully resynced.");

    # Verify the load sharing mirrors are still healthy

    undef %sources_checked;
    foreach my $dest_vol (@{$ls_volumes}) {

        # Vreport

        my $object = $dest_vol->{vserver}.":".$dest_vol->{volume};
        eval {
            $Server->debug_vreport_show(
                type   => "volume",
                object => $object,
            );
        };
        if ($@ =~ /There are no entries matching your query./) {
            $Trace->debug(msg => "Vreport did not find descrepencies for object: ".
                          "$object after running wafliron as expected");
        } elsif ($@) {
            $Trace->exit(msg => "Received unexpected error message while running ".
                         "vreport for object: $object: $@");
            return 0;
        } else {
            $Trace->exit(msg => "Vreport found descrepencies for object: $object ".
                          "prior to running wafliron. Exiting.");
            return 0;
        }

        # Only run the health check if flagged to

        if (!$params{skip_health_check}) {

            # Get the source volume so that we can run the health check from there
            
            my $sm_source;
            eval {
                $sm_source = $Server->snapmirror_show(
                    'destination-vserver' => $dest_vol->{vserver},
                    'destination-volume'  => $dest_vol->{volume},
                );
            };
            if ($@ =~ /There are no entries matching your query./) {
                $Trace->comment(msg => "Volume $dest_vol->{volume} in vserver ".
                                "$dest_vol->{vserver} is a $dest_vol->{type} volume, ".
                                "but is not configured to be part of a SnapMirror ".
                                "relationship. Skipping it...");
                next;
            } elsif ($@) {
                $Trace->warn(msg => "Received unexpected result when querying for ".
                             "$dest_vol->{volume} on $dest_vol->{vserver}: $@. ".
                             "Failed to iron aggregate $params{aggregate}.");
                $Trace->exit();
                return 0;
            }

            if (scalar(@{$sm_source}) > 1) {
                $Trace->warn(msg => "Only one entry should be returned when querying ".
                             "for a SnapMirror set via its destination volume & ".
                             "vserver. Found: ".scalar(@{$sm_source}));
                $Trace->exit();
                return 0;
            }

            # Check if we've already called heath check for this source volume.
            # If not, run health check and add it to the list. so we don't waste
            # time checking it again. Also handle the case where this is the first
            # source volume for this particular vserver. 

            my @vols_checked;
            if (defined($sources_checked{$sm_source->[0]{'source-vserver'}})) {
                @vols_checked = @{$sources_checked{$sm_source->[0]{'source-vserver'}}};
            }
            
            if (!grep(/$sm_source->[0]{'source-volume'}/, @vols_checked)) {

                # SnapMirror Health Check

                if (!health_check($self,
                            vs   => $sm_source->[0]{'source-vserver'},
                            vol  => $sm_source->[0]{'source-volume'},
                            type => $sm_source->[0]{'type'},
                        )) {
                    $Trace->warn(msg=> "$sm_source->[0]{'type'} Set destined to ".
                                 "$dest_vol->{volume} on vserver ".
                                 "$dest_vol->{vserver} is not healthy after ".
                                 "running wafliron on aggregate $params{aggregate}.");
                    return 0;
                }

                # Add it to the list of source volumes that we have already checked
                
                push(@vols_checked, $sm_source->[0]{'source-volume'});
                $sources_checked{$sm_source->[0]{'source-vserver'}} = \@vols_checked;
            }

            $Trace->comment(msg => "SnapMirror set destined to $dest_vol->{vserver}, ".
                            "$dest_vol->{volume} is heathly after running wafliron.")

        } ## end if (!$params{skip_health_check})
    } ## end foreach my $dest_vol

    # Log result and return

    $Trace->exit(msg => "Successful ironed aggregate $params{aggregate}. All ".
                 "SnapMirror sets on this aggregate continue to be healthy.");
    return 1;

} ## end sub sub run_wafliron

=head1 PRIVATE METHODS

=head2 _did_job_succeed

Single procedure to handle the job completion string comparison

=cut

#############################################################################
# Subroutine: _did_job_succeed
# Parameters: Completion string from a job show command
# Returns:    1 if the job was successful, 0 on failure
#############################################################################

sub _did_job_succeed {

    $Trace->entry();

    # Initialize parameters

    my %params = validate(
        @_,
        {
            completion => {
                type     => SCALAR,
                optional => 0,
            },
        }
    );

    # Verify it has indeed finished

    if ($params{completion} =~ m/SnapMirror: done/g) {

        # Check for update failures (should be proceeded by a "NOTE:"

        if ($params{completion} =~ m/NOTE:/g && $params{completion} !~ m/NOTE: load-sharing destination volume .+ is current/g) {
            $Trace->warn(msg => "SnapMirror Job completed but errored with: ".
                         $params{completion});
            return 0;
        } else {
            $Trace->exit(msg => "SnapMirror Job successfully completed.");
            return 1;
        }

    } else {

        # Job did not report it was finished
        
        $Trace->warn(msg => "SnapMirror Job did not complete. Failed with: ".
                         $params{completion});
        return 0;
    }

} ## end sub _did_job_succeed

=head2 _get_source_volumes

Subroutine to pare down a list of source volumes after issuing 
a snapmirror show (since source volume is returned for each entry
in the same relationship). Allows the calling method to specify
wildcard characters to do verification in chunks.

=cut

#############################################################################
# Subroutine: _get_source_volumes
# Parameters: Source vserver, volume, and type
# Returns: Returns a list of source volumes used in these relationships
#############################################################################

sub _get_source_volumes {

    $Trace->entry();

    # Initialize parameters

    my %params = validate(
        @_,
        {
            vs => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_vserver},
            },
            vol => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_volume},
            },
            type => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_type},
            },
        }
    );

    my $raw_output = $Server->snapmirror_show(
        'source-vserver' => $params{vs},
        'source-volume'  => $params{vol},
        type             => $params{type},
    );

    # Loop through the raw output returned and push that value onto the
    # list if it is not a part of the list already.

    my @source_vols;
    foreach my $sm (@{$raw_output}) {
        push(@source_vols, $sm->{'source-volume'}) unless 
            (grep(/^$sm->{'source-volume'}$/, @source_vols));
    }

    $Trace->exit(msg => "Returning with source volumes: ".Dumper(\@source_vols));
    return \@source_vols;

} ## end sub _get_source_volumes

=head2 _get_destination_volumes

Subroutine to obtain a list of destination volumes after issuing 
a snapmirror show on a particular source volume

=cut

#############################################################################
# Subroutine: _get_destination_volumes
# Parameters: Source vserver, volume, and type
# Returns: Returns a list of destination volumes for this relationship
#############################################################################

sub _get_destination_volumes {

    $Trace->entry();

    # Initialize parameters

    my %params = validate(
        @_,
        {
            vs => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_vserver},
            },
            vol => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_source_volume},
            },
            type => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_type},
            },
        }
    );

    my $raw_output = $Server->snapmirror_show(
        'source-vserver' => $params{vs},
        'source-volume'  => $params{vol},
        type             => $params{type},
    );

    # Loop through the raw output returned and push that value onto the
    # list if it is not a part of the list already.

    my @dest_vols;
    foreach my $sm (@{$raw_output}) {
        push(@dest_vols, $sm->{'destination-volume'}) unless 
            (grep(/^$sm->{'destination-volume'}$/, @dest_vols));
    }

    $Trace->exit(msg => "Returning with destination volumes: ".
                 Dumper(\@dest_vols));
    return \@dest_vols;

} ## end sub _get_destination_volumes

=head2 _calculate_exported_snapshot_epoch

Covert the timestamp to seconds so that we can do simple math to 
see if we are up to date.

=cut

#############################################################################
# Subroutine: _calculate_exported_snapshot_epoch
# Parameters: A SnapMirror's exported snapshot
# Returns:    Epoch (number of seconds) representing the timestamp of the exported
#             snapshot
#############################################################################

sub _calculate_exported_snapshot_epoch {

    $Trace->entry();

    # Initialize parameters

    my %params = validate(
        @_,
        {
            snapshot => {
                type     => SCALAR,
                optional => 0,
            },
   });

    # An example exported-snapshot can look like the following:
    # "snapmirror.10_2147484712.2009-05-11_143707". We want to parse out the 
    # date portion and convert that to seconds. Also, verifying that we are
    # in the correct format.

    my %snapshot_timestamp;
    if ($params{snapshot} =~ m/^snapmirror.\d+_\d+\.(\d{4})-(\d{2})-(\d{2})_(\d{6})$/) {
        
        # Save the values (still need to break out the time even further)
        
        $snapshot_timestamp{year}  = $1;
        $snapshot_timestamp{month} = $2;
        $snapshot_timestamp{day}   = $3;
        my $temp_time              = $4;

        if ($temp_time =~ m/(\d{2})(\d{2})(\d{2})/) {
            $snapshot_timestamp{hour} = $1;            
            $snapshot_timestamp{min}  = $2;
            $snapshot_timestamp{sec}  = $3;
        } else {
            $Trace->warn(msg => "Exported-snapshot timestamp in unexpected ".
                         "format. Expected 6 digits (HHMMSS), found $temp_time. ".
                         "Exiting.");
            return 0;
        }

    } else {
        $Trace->warn(msg => "Exported-snapshot not in correct format. ".
                     "Expected: snapmirror.XX_XXXXX.YYYY-DD-MM_HHMMSS, found ".
                     "$params{snapshot}");
        return 0;
    }

    # Now that everything is parsed out, we can fetch the epoch

    my $epoch = timelocal(
        $snapshot_timestamp{sec},
        $snapshot_timestamp{min},
        $snapshot_timestamp{hour},
        $snapshot_timestamp{day},
        $snapshot_timestamp{month} - 1,
        $snapshot_timestamp{year},
    );

    $Trace->exit(msg => "Returning with exported-snapshot's epoch of $epoch");
    return $epoch;

} ## end sub _calculate_exported_snapshot_epoch

=head2 _calculate_last_update

Convert interval & cron jobs to seconds

=cut

#############################################################################
# Subroutine: _calculate_last_update
# Parameters: A SnapMirror's schedule
# Returns:    Number of seconds since the last update should have occurred
#############################################################################

sub _calculate_last_update {

    $Trace->entry();

    # Initialize parameters

    my %params = validate(
        @_,
        {
            schedule => {
                type     => SCALAR,
                optional => 0,
            },
    });

    # Make sure we have a schedule to calculate

    if ($params{schedule} eq '""' || $params{schedule} eq '-') {
        $Trace->warn(msg => "This SnapMirror relationship does not have a ".
                     "schedule associated with it. Returning...");
        $Trace->exit();
        return 0;
    }

    # Get the Job information

    if ($params{schedule} eq '-') {
        $Trace->config(msg => "");
    }
    my $job_info = $Server->job_schedule_show(
        name => $params{schedule},
    );
    if (scalar(@{$job_info}) > 1) {
        $Trace->warn(msg => "Found more than one job named: ".
                     "$params{schedule}. Defaulting to the first one.");
    }

    # We will have to handle interval schedules differently than cron 
    # schedules.

    my $total_seconds = 0;
    my $this_job;
    my $date;

    if ($job_info->[0]{type} eq "interval") {

        $this_job = $Server->job_schedule_interval_show(
            name => $params{schedule},
        );

        $total_seconds = ($this_job->[0]{seconds}) + 
                         ($this_job->[0]{minutes}*60) +
                         ($this_job->[0]{hours}*3600) + 
                         ($this_job->[0]{days}*86400);

    } elsif ($job_info->[0]{type} eq "cron") {

        # Get the job information

        $this_job = $Server->job_schedule_cron_show(
            name => $params{schedule},
        );

        # Grab the localtime and break it out. We are doing it this way instead
        # of asking for the array because we want the names of the month (i.e.
        # "May") opposed to the number (i.e. "5").

        my $current_time = localtime();
        my %cron_timestamp;
        if ($current_time =~ m/^(\w+)\s(\w+)\s+(\d+)\s(\d{2}):(\d{2}):(\d{2})\s(\d{4})$/) {
            $cron_timestamp{dayofweek} = $1;
            $cron_timestamp{month}     = $2;
            $cron_timestamp{day}       = $3;
            $cron_timestamp{hour}      = $4;
            $cron_timestamp{minute}    = $5;
            $cron_timestamp{second}    = $6;
            $cron_timestamp{year}      = $7;
        }

        # total_interval will be used to hold the number of seconds for the
        # cron interval we are currently working with. Since we *DO* assume
        # basic cron schedules, we aren't handling special cases. If the
        # curr_epoch is less than the cron_epoch, that means the schedule
        # has not yet fired for this interval. We will take the difference between
        # the two times and subtract them from the total_interval. This will give
        # us the number of seconds since the LAST update and not the number of
        # seconds until the NEXT update. 

        my $total_interval = 0;

        # Whenever we find a "-" in the cron job, we will substitute today's
        # value.

        if ($this_job->[0]{month} ne "-") {

            $cron_timestamp{month} = $this_job->[0]{month}; 

            # total_interval will be set to 12 months to account for another
            # year (because if its not set yet, we are on a yearly schedule).

            if ($total_interval == 0) {
                $total_interval = 31536000;
            }
        }

        if ($this_job->[0]{day} ne "-") {

            $cron_timestamp{day} = $this_job->[0]{day};

            # total_interval will be set to 31 days to account for another
            # month (because if its not set yet, we are on a monthly schedule).
            # NOTE: 31 days is an estimate for the worst case.

            if ($total_interval == 0) {
                $total_interval = 2678400;
            }
        }

        if ($this_job->[0]{dayofweek} ne "-") {

            $cron_timestamp{dayofweek} = $this_job->[0]{dayofweek};

            # total_interval will be set to 7 days to account for another
            # week (because if its not set yet, we are on a weekly schedule).

            if ($total_interval == 0) {
                $total_interval = 604800;
            }
        }

        if ($this_job->[0]{hour} ne "-") {

            $cron_timestamp{hour} = $this_job->[0]{hour};

            # total_interval will be set to 24 hours to account for another
            # day (because if its not set yet, we are on a daily schedule).

            if ($total_interval == 0) {
                $total_interval = 86400;
            }
        }

        if ($this_job->[0]{minute} ne "-") {

            $cron_timestamp{minute} = $this_job->[0]{minute};

            # total_interval will be set to 60 mins to account for another
            # hour (because if its not set yet, we are on an hourly schedule).

            if ($total_interval == 0) {
                $total_interval = 3600;
            }
        }

        # Verify that all of timestamps are in the expected format. We currently
        # do not support special cases when checking cron jobs

        my $error = "";
        
        foreach my $key (keys(%cron_timestamp)) {
            if ($cron_timestamp{$key} !~ /^\w+$/) {
                $error .= "Cron Job $params{schedule} contains unsupported format ".
                          "for $key. Found: $cron_timestamp{$key}.\n";
            }
        }
        if ($error ne "") {
            $Trace->warn(msg => $error);
            return 0;
        }

        # Convert to an epoch

        $date = sprintf("%s %s %2u %s:%s:%s %u", $cron_timestamp{dayofweek},
                       $cron_timestamp{month}, $cron_timestamp{day},
                       $cron_timestamp{hour}, $cron_timestamp{minute}, 
                       $cron_timestamp{second}, $cron_timestamp{year}
        );

        my $cron_epoch = str2time($date);
        my $curr_epoch = str2time($current_time);

        # If the curr_epoch is less than the cron epoch then we haven't hit the
        # scheduled update time yet for this interval. We will account for that
        # here.

        if ($curr_epoch < $cron_epoch) {

            $Trace->comment(msg => "The current epoch: $curr_epoch is less than ".
                            "the cron job's epoch: $cron_epoch. This means that the ".
                            "job has not fired yet for this interval. Accounting ".
                            "for the total interval: $total_interval as well");
            $total_seconds = $total_interval - ($cron_epoch - $curr_epoch);

            $Trace->exit(msg => "Last scheduled update should have occurred within ".
                         "$total_seconds seconds. (Cron date: $date, Current time: ".
                         "$current_time)");

        } else {

            # Grab the difference. 
            
            $total_seconds = $curr_epoch - $cron_epoch;            

            $Trace->exit(msg => "Last scheduled update should have occurred within ".
                 "$total_seconds seconds.");

        }

    } else {
        $Trace->warn(msg => "Found unexpected job type: $job_info->[0]{type}. ".
                     "Expected interval|cron. Returning.");
        return 0;
    }

    return $total_seconds;

} ## end sub _calculate_last_update

1;
