# $Id: //depot/prod/test/Rsierranevada/lib/SnapMirrorZapi.pm#1 $
# Copyright (c) 2010 Network Appliance, Inc.
#  All rights reserved.
#

#-----------------------------------------------------------
## @summary Routine to test Snapmirror c-mode ZAPIs
##
## @description
## This module implements routines to test SnapMirror c-mode ZAPIs
## Testing a ZAPI involves the following steps
## 1- Instantiate a SnapMirrorZapi Object 
## 2- Set expected results returned by the ZAPI
## 3- Set whether to check output parameters and their expected 
##    values where applicable
## 4- If applicable Set whether completion of Asynchronous ZAPI's
##    should be monitored and expected job status and state 
## 5- Set the ZAPI input parameters
## 6- Invoke the entry point that tests the ZAPI
## 7- Check whether the test was successful or not
## 8- Get additional  information about the test result including the ZAPI output
##
## The following sections explains each of these steps in more details
##
## 1- Instantiate a SnapMirrorZapi object
##    -------------------------------------
##    my $sm = new SnapMirrorZapi( $filer_ip, $user, $password );
##
## 2- Set expected results
##    ------------------------
##    This includes setting the expected results returned by the ZAPI including:
##    status, errno
##    - To set expected status: $sm->set_exp_result($exp_status). Default: 'passed'
##    - To set expected errno: $sm->set_exp_errno($exp_errno). Default: any errno 
##     value is OK i.e value of errno returned is not checked
## 
## 3- Set whether to check output parameter and possible expected values
##    ------------------------------------------------------------------
##    By default the content of output parameters returned by the ZAPI is 
##    checked according to the following rules:
##       - snapmirror-get/get-iter: 
##           + Check that only desired attributes returned if specified
##           + Check that enpoints parameters match corresponding input 
##             parameters. 
##             Note: Check of wild cards query is not implemented yet
##       - snapmirror-create:
##           + if return-record set true, check that endpoints and conf parameters
##             match corresponding input parameters
##       - Other query ZAPI (basically xxx-iter): the check of the content is not yet 
##         implemented yet
##       - Non query Asynchronous ZAPIs: See 'Set verification of Asynchronous ZAPIs jobs'
##       - snapmirror-get-volume-status: Possible expected values are 'true', 'false', 
##           '_nocare_'. '_nocare_' means the parameter should not be checked
##           + To set expected is-destination: $sm->set_vol_stat_exp_is_dst($exp_is_dest);
##           + To set expected is-source: $sm->set_vol_stat_exp_is_dst($exp_is_src);
##           + To set expected is-transfer-in-progress: 
##                                  $sm->set_vol_stat_exp_is_xfer($exp_is_xfer);
##           + To set expected is-transfer-is-broken: 
##                                  $sm->set_vol_stat_exp_is_broken($exp_is_broken);
##          
## 4- Set verification of the completion of Async ZAPIs' jobs
##    -------------------------------------------------------
##    By default completion of the job associated to a non query (not xxx-iter) async
##    ZAPI is monitored. The job status is expected to be '0' and the expected job state 
##    is 'success'
##    This behavior can be changed as follow:
##      - To disable verification of job completion i.e test returns when the ZAPI returns:
##           $sm->set_job_monitor("false");
##      - To re-enable verification of job completion: $sm->set_job_monitor("true");
##      - To set the expected job status: $sm->set_job_exp_status($exp_job_status);
##        Note: to just indicate that the job is expected to fail regardless of a
##             specific value, you must specify $SnapMirrorZapi::FAIL as expected job status
##      - To set the expected job state: $sm->set_job_exp_state($exp_job_state);  
##        Note: If you don't care about a specific job state,  you must specify
##             $SnapMirrorZapi::ANY as the expected job state
## 
## 5- Set the ZAPI input parameters
##    -----------------------------
##    To set the ZAPI input parameter: $sm->set_req_param($ref_parameter_hash);
##    $ref_paramter_ref is the reference to a hash containing values of the ZAPI input
##    parameters. The input parameter names should be used as keys to the hash.
##    The following should be noted:
##    - desired-attributes values should be specified as a comma separated list
##      myhash{"desired-attributes"} = "is-healthy,relation-ship,mirror-state";
##    - specifying a value for a desired-attribute is not supported
##    - For ZAPI with query, the attributes to the query should be specified as for
##      ZAPI without query. They are put in query automatically. For example to
##      to query relationships with destination location starting with 'cl://vs/vol1':
##      myhash{"destination-location} = "cl://vs/vol1*';
##    Note: Also see examples below
##
## 6- Invoke the entry point that test the ZAPI
##    ----------------------------------------
##    The name of entry that test a ZAPI is the name of the ZAPI with '-' replaced with '_'
##    for example for snapmirror-get the entry point is snapmirror_get, for 
##    snapmirror-initialize-ls-set, the entry point is snapmirror_initialize_ls_set
##    To invoke ZAPI test entry point: my $result = $sm-><zapi_entry_point>();
##    example for snapmirror-get: my $result = $sm->snapmirror_get();
##
## 7- Check whether the test succeeded or fail
##    ----------------------------------------
##    If test succeeded, $TCD::PASS is returned. If the test failed, $TCD::FAIL is 
##    returned. In the latter case, the cause of the failure can be retrieved by invoking:
##    $sm->get_last_errmsg();
## 8- Get additional information about the result of the test
##    ------------------------------------------------------
##    - To get the XML output of the last ZAPI: $sm->get_last_xo();
##    - To get the XML output for last Async Zapi job: $sm->get_last_job_xo();
##    - To get a string with nicely formatted information about the last async ZAPI job: 
##        $sm->sprint_job_info(); 
## 
## 9- Example: To test snapmirror-get with success expected and the content is checked
##    ---------------------------------------------------------------------------------
##    my $sm = new SnapMirrorZapi ($ip, "admin", "netapp1!");
##    my %get_parm; 
##    $get_parm{"destination-location"} = "cl://vs0/vol00";
##    $sm->set_req_param(\%get_param);
##    my $result = $sm->snapmirror_get();
##    my $xo = $sm->get_last_xo();
##    if ($verbose > 0) {
##       my $xo = $sm->get_last_xo();
##       if (defined($xo)) {
##          printf "snapmirror-get output: " . $xo->sprint_pretty();
##       }
##    }
##    if ($result eq $TCD::FAIL) {
##       my $errmsg = $sm->get_last_errmsg();
##       printf "Test failed: $errmsg\n";
##    } else {
##      print "Test succeeded\n";
##    }
##
## 10- Example: To test snapmirror-get-iter with: success expected and content checked
##     ------------------------------------------------------------------------------
##    # Relations with source:  'cl://vs0/vol01'
##    my $sm = new SnapMirrorZapi ($ip, "admin", "netapp1!");
##    my %get_iter_parm;
##    $get_iter_parm{"source-location"} = "cl://vs0/vol01";
##    $sm->set_req_param(\%get_iter_param);
##    my $result = $sm->snapmirror_get_iter();
##    if ($verbose > 0) {
##       my $xo = $sm->get_last_xo();
##       if (defined($xo)) {
##          printf "snapmirror-get-iter output: " . $xo->sprint_pretty();
##       }
##    }
##    if ($result eq $TCD::FAIL) {
##       my $errmsg = $sm->get_last_errmsg();
##       printf "Test failed: $errmsg\n";
##    } else {
##      print "Test succeeded\n";
##    }
##
## 11- Example: To test snapmirror-update with: failure of the job expected (e.g. xfer disabled)
##     ------------------------------------------------------------------------------
##    my $sm = new SnapMirrorZapi ($ip, "admin", "netapp1!");
##    $sm->set_job_exp_status($SnapMirrorZapi::FAIL);
##    $sm->set_job_exp_state("failure");
##    my %update_parm;
##    $get_iter_parm{"destination-location"} = "cl://vs0/vol02";
##    $sm->set_req_param(\%update_param);
##    my $result = $sm->snapmirror_update();
##    if ($verbose > 0) {
##       my $xo = $sm->get_last_xo();
##       if (defined($xo)) {
##          printf "snapmirror-get-iter output: " . $xo->sprint_pretty();
##       }
##       my $jobinfo = $sm->sprint_job_info();
##       if (defined($jobinfo)) {
##          printf "snapmirror-get-iter job information:\n$jobinfo ";
##       }
##    }
## 
##    if ($result eq $TCD::FAIL) {
##       my $errmsg = $sm->get_last_errmsg();
##       printf "Test failed: $errmsg\n";
##    } else {
##      print "Test succeeded\n";
##    }
##
## @author alain.nonga@netapp.com

package SnapMirrorZapi;

use strict;

use misc;
use ZapiTest;
use XML::Parser;
use NaElement;
use Zapi;
use TCD;
use Tharn qw($Log);

use vars ('@ISA', '@EXPORT');
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(new);
@EXPORT = qw(set_exp_result);
@EXPORT = qw(set_exp_errno);
@EXPORT = qw(get_last_errmsg);
@EXPORT = qw(set_check_content);
@EXPORT = qw(set_req_param);
@EXPORT = qw(sprint_job_info);
@EXPORT = qw(set_job_monitor);
@EXPORT = qw(set_job_exp_state);
@EXPORT = qw(set_job_exp_status);
@EXPORT = qw(get_last_job_xo);
@EXPORT = qw(set_vol_stat_exp_is_src);
@EXPORT = qw(set_vol_stat_exp_is_dst);
@EXPORT = qw(set_vol_stat_exp_is_xfer);
@EXPORT = qw(set_vol_stat_exp_is_broken);
@EXPORT = qw(get_last_xo);
@EXPORT = qw(snapmirror_get);
@EXPORT = qw(snapmirror_get_iter);
@EXPORT = qw(snapmirror_abort);
@EXPORT = qw(snapmirror_abort_iter);
@EXPORT = qw(snapmirror_break);
@EXPORT = qw(snapmirror_break_iter);
@EXPORT = qw(snapmirror_check);
@EXPORT = qw(snapmirror_check_iter);
@EXPORT = qw(snapmirror_initialize);
@EXPORT = qw(snapmirror_initialize_ls_set);
@EXPORT = qw(snapmirror_update);
@EXPORT = qw(snapmirror_update_ls_set);
@EXPORT = qw(snapmirror_promote);
@EXPORT = qw(snapmirror_destroy);
@EXPORT = qw(snapmirror_get_volume_status);
@EXPORT = qw(snapmirror_create);
@EXPORT = qw(snapmirror_modify);
@EXPORT = qw(snapmirror_quiesce);
@EXPORT = qw(snapmirror_resume);
@EXPORT = qw(snapmirror_resync);

# Global variables
$SnapMirrorZapi::SUCCESS = "0";
$SnapMirrorZapi::ANY = "_nocare_";
$SnapMirrorZapi::FAIL = "fail";

sub new {
    my $class = shift;
    my $ip = shift;
    my $user = shift;
    my $passwd = shift;
    my $self = {};

    $self->{ip} = $ip;
    $self->{user} = $user;
    $self->{passwd} = $passwd;
    $self->{exp_result} = "passed";
    $self->{exp_errno} = $Zapi::ANY;
    $self->{check_content} = "false";
    $self->{endpoint}->{"destination-cluster"} = "";
    $self->{endpoint}->{"destination-location"} = "";
    $self->{endpoint}->{"destination-vserver"} = "";
    $self->{endpoint}->{"destination-volume"} = "";
    $self->{endpoint}->{"source-cluster"} = "";
    $self->{endpoint}->{"source-location"} = "";
    $self->{endpoint}->{"source-vserver"} = "";
    $self->{endpoint}->{"source-volume"} = "";
    $self->{job_flags}->{monitor} = "true";
    $self->{job_flags}->{exp_job_state} = "success";
    $self->{job_flags}->{exp_job_status} = $SnapMirrorZapi::SUCCESS;
    $self->{job_result}->{job_id} = "_unknown_";
    $self->{job_result}->{job_status} = "_unknown_";
    $self->{job_result}->{job_state} = "_unknown_";
    $self->{job_result}->{job_completion} = "_unknown_";
    $self->{job_result}->{job_name} = "_unknown_";
    $self->{exp_vol_status}->{is_source} = $SnapMirrorZapi::ANY;
    $self->{exp_vol_status}->{is_dest} = $SnapMirrorZapi::ANY;
    $self->{exp_vol_status}->{is_xfer} = $SnapMirrorZapi::ANY;
    $self->{exp_vol_status}->{is_broken} = $SnapMirrorZapi::ANY;
    bless $self, $class;
    return $self; 
}

sub set_exp_result {
    my $self = shift;
    my $exp_result = shift;
    if (defined($exp_result)) {
        $self->{exp_result} = $exp_result;
    }
}

sub get_exp_result {
    my $self = shift;
    return $self->{exp_result};
}

sub set_exp_errno {
    my $self = shift;
    my $exp_errno = shift;
    if (defined($exp_errno)) {
        $self->{exp_errno} = $exp_errno;
    }
}

sub get_exp_errno {
    my $self = shift;
    return $self->{exp_errno};
}

sub set_check_content {
    my $self = shift;
    my $check_content = shift;
    if (defined($check_content)) {
        $self->{check_content} = $check_content;
    }
}

sub get_check_content {
    my $self = shift;
    return $self->{check_content};
}


sub set_req_param {
    my $self = shift;
    $self->{req_params} = shift;
}

sub get_last_errmsg {
   my $self = shift;
   return $self->{last_errmsg};
}

sub get_last_xo {
   my $self = shift;
   return $self->{last_xo};
}

sub check_result{
    my $self = shift;
    my $server = shift;
    my $xo = shift;

    my $result;
    my $msg;

    ($result, $msg) = $server->check_result($xo, @_);

    if ( $result eq  $Zapi::FAIL) {
        $self->{last_errmsg} = $msg;
        return $TCD::FAIL;
    }
    return $TCD::PASS;
}

sub sprint_job_info {
    my $self = shift;
    my $job_name = $self->{job_result}->{job_name};
    my $job_id = $self->{job_result}->{job_id};
    my $job_state = $self->{job_result}->{job_state};
    my $job_status = $self->{job_result}->{job_status};
    my $job_completion = $self->{job_result}->{job_completion};
    my $msg = "\njob-name = $job_name\n" 
             . "job-id = $job_id\n"
             . "job-state = $job_state\n"
             . "job-status-code = $job_status\n"
             . "job-completion = $job_completion\n";
    return $msg;
}

sub sprint_job_flags {
   my $self = shift;
   my $monitor = $self->{job_flags}->{monitor};
   my $job_state = $self->{job_flags}->{exp_job_state};
   my $job_status = $self->{job_flags}->{exp_job_status};
   my $msg = "\nmonitor=$monitor\nexp_state=$job_state\nexp_status=$job_status\n";
   return $msg;
}


sub set_job_monitor {
   my $self = shift;
   my $monitor = shift;
   if (defined($monitor)) {
       $self->{job_flags}->{monitor} = $monitor;
   }
}

sub set_job_exp_state {
   my $self = shift;
   my $job_state = shift;
   if (defined($job_state)) {
       $self->{job_flags}->{exp_job_state} = $job_state;
   }
}

sub set_job_exp_status {
   my $self = shift;
   my $job_status = shift;
   if (defined($job_status)) {
       $self->{job_flags}->{exp_job_status} = $job_status;
   }
}

sub set_vol_stat_exp_is_src {
   my $self = shift;
   my $is_source = shift;
   if (defined($is_source)) {
       $self->{exp_vol_status}->{is_source} = $is_source;
   }
}

sub set_vol_stat_exp_is_dst {
   my $self = shift;
   my $is_dest = shift;
   if (defined($is_dest)) {
       $self->{exp_vol_status}->{is_dest} = $is_dest;
   }
}

sub set_vol_stat_exp_is_xfer {
   my $self = shift;
   my $is_xfer = shift;
   if (defined($is_xfer)) {
       $self->{exp_vol_status}->{is_xfer} = $is_xfer;
   }
}

sub set_vol_stat_exp_is_broken {
   my $self = shift;
   my $is_broken = shift;
   if (defined($is_broken)) {
       $self->{exp_vol_status}->{is_broken} = $is_broken;
   }
}

sub get_last_job_xo {
    my $self = shift;
    return $self->{last_job_xo};
}

sub get_last_out_params {
   my $self = shift;
   return $self->{out_params};
}

sub check_elmt {
    my $self = shift;
    my $xo = shift;
    my $elmt_name = shift;
    my $exp_val = shift;
    my $attr_cont = $xo->child_get_string($elmt_name);
    if (!defined($attr_cont)) {
         $self->{last_errmsg} = "$elmt_name missing from output";
         return $TCD::FAIL;
     }
     if ($attr_cont ne $exp_val) {
         $self->{last_errmsg} = "Unexpected value returned for $elmt_name:" 
                                   . " $attr_cont, expected: $exp_val";
         return $TCD::FAIL; 
     }
     return $TCD::PASS;
}

sub check_std_output{
    my $self = shift;
    my $sm_info = shift;
    my $res;

    if ( defined ($self->{std_req}->{"destination-location"})) {
        $res = $self->check_elmt($sm_info, "destination-location", 
                                 $self->{std_req}->{"destination-location"});
        if ($res eq $TCD::FAIL) {
            return $res
        }
    }        
    if ( defined ($self->{std_req}->{"destination-cluster"})) {
        $res = $self->check_elmt($sm_info, "destination-cluster", 
                                 $self->{std_req}->{"destination-cluster"});
        if ($res eq $TCD::FAIL) {
            return $res
        }
    }        
    if ( defined ($self->{std_req}->{"destination-vserver"})) {
        $res = $self->check_elmt($sm_info, "destination-vserver", 
                                 $self->{std_req}->{"destination-vserver"});
        if ($res eq $TCD::FAIL) {
            return $res
        }
    }        
    if ( defined ($self->{std_req}->{"destination-volume"})) {
        $res = $self->check_elmt($sm_info, "destination-volume", 
                                 $self->{std_req}->{"destination-volume"});
        if ($res eq $TCD::FAIL) {
            return $res
        }
    }        


    if ( defined ($self->{std_req}->{"source-location"})) {
        $res = $self->check_elmt($sm_info, "source-location", 
                                 $self->{std_req}->{"source-location"});
        if ($res eq $TCD::FAIL) {
            return $res
        }
    }        
    if ( defined ($self->{std_req}->{"source-cluster"})) {
        $res = $self->check_elmt($sm_info, "source-cluster", 
                                 $self->{std_req}->{"source-cluster"});
        if ($res eq $TCD::FAIL) {
            return $res
        }
    }        
    if ( defined ($self->{std_req}->{"source-vserver"})) {
        $res = $self->check_elmt($sm_info, "source-vserver", 
                                 $self->{std_req}->{"source-vserver"});
        if ($res eq $TCD::FAIL) {
            return $res
        }
    }        
    if ( defined ($self->{std_req}->{"source-volume"})) {
        $res = $self->check_elmt($sm_info, "source-volume", 
                                 $self->{std_req}->{"source-volume"});
        if ($res eq $TCD::FAIL) {
            return $res
        }
    }        
    return $TCD::PASS;
}

sub check_sm_des_attrs {
    my $self = shift;
    my $sm_info = shift;
    my $req_type = shift;

    # if desired-attributes specified, check the following:
    # Any attribute returned is an end-point attribute or a desired attribute
    my $ref_des_attrs = $self->{$req_type}->{"desired-attributes"}{"snapmirror-info"};
    if (defined($ref_des_attrs)) { 
        my $nb_ret_attrs = $sm_info->child_max_number;
        my %des_attrs = %$ref_des_attrs;
        for (my $i=0; $i < $nb_ret_attrs; $i++) {
            my $child = $sm_info->child_get_int($i);
            if (defined($child)) {
                my $aname = $child->name();
                if ((!defined($des_attrs{$aname})) && 
                    (!defined($self->{endpoint}->{$aname}))) {
                    $self->{last_errmsg} = "attribute \'$aname\' returned," 
                                      . " but not in the list of desired attributes";
                    return $TCD::FAIL;
                }
            } 
        }
    } 
    return $TCD::PASS;

}

sub check_sm_get_cont {
    my $self = shift;
    my $xo = shift;
    my $attributes = $xo->child_get("attributes");
    if (!defined($attributes)) {
         $self->{last_errmsg} = "attributes missing from snapmirror-get output\n";
         return $TCD::FAIL;
    }    
    my $sm_info = $attributes->child_get("snapmirror-info");
    if (!defined($sm_info)) {
         $self->{last_errmsg} = "snapmirror-info missing from snapmirror-get output\n";
         return $TCD::FAIL;
    }    
    
    my $result;
    # Check desired-attributes if specified
    if (defined ($self->{std_req}->{"desired-attributes"})) {
        $result = $self->check_sm_des_attrs($sm_info, "std_req");
        if ($result eq $TCD::FAIL) {
            return $result;
        }
    }

    ## Check endpoint output parameters against corresponding input parameters
    return $self->check_std_output($sm_info);
}

sub clear_req_buffers {
    my $self = shift;
    
    $self->{std_req} = {};
    $self->{query_req} = {};
}

sub save_sm_info_out_params {
    my $self = shift;
    my $sm_info = shift;
    my $elmt_val;
    $elmt_val  = $sm_info->child_get_string("destination-location");
    if (defined($elmt_val)) {
       $self->{out_params}->{"destination-location"} = $elmt_val;
    }
    $elmt_val  = $sm_info->child_get_string("destination-cluster");
    if (defined($elmt_val)) {
       $self->{out_params}->{"destination-cluster"} = $elmt_val;
    }
    $elmt_val  = $sm_info->child_get_string("destination-vserver");
    if (defined($elmt_val)) {
       $self->{out_params}->{"destination-vserver"} = $elmt_val;
    }
    $elmt_val  = $sm_info->child_get_string("destination-volume");
    if (defined($elmt_val)) {
       $self->{out_params}->{"destination-volume"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("source-location");
    if (defined($elmt_val)) {
       $self->{out_params}->{"source-location"} = $elmt_val;
    }
    $elmt_val  = $sm_info->child_get_string("source-cluster");
    if (defined($elmt_val)) {
       $self->{out_params}->{"source-cluster"} = $elmt_val;
    }
    $elmt_val  = $sm_info->child_get_string("source-vserver");
    if (defined($elmt_val)) {
       $self->{out_params}->{"source-vserver"} = $elmt_val;
    }
    $elmt_val  = $sm_info->child_get_string("source-volume");
    if (defined($elmt_val)) {
       $self->{out_params}->{"source-volume"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("exported-snapshot");
    if (defined($elmt_val)) {
       $self->{out_params}->{"exported-snapshot"} = $elmt_val;
    }
    $elmt_val  = $sm_info->child_get_string("exported-snapshot-timestamp");
    if (defined($elmt_val)) {
       $self->{out_params}->{"exported-snapshot-timestamp"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("is-healthy");
    if (defined($elmt_val)) {
       $self->{out_params}->{"is-healthy"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("max-transfer-rate");
    if (defined($elmt_val)) {
       $self->{out_params}->{"max-transfer-rate"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("mirror-state");
    if (defined($elmt_val)) {
       $self->{out_params}->{"mirror-state"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("newest-snapshot");
    if (defined($elmt_val)) {
       $self->{out_params}->{"newest-snapshot"} = $elmt_val;
    }
    $elmt_val  = $sm_info->child_get_string("newest-snapshot-timestamp");
    if (defined($elmt_val)) {
       $self->{out_params}->{"newest-snapshot-timestamp"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("relationship-status");
    if (defined($elmt_val)) {
       $self->{out_params}->{"relationship-status"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("schedule");
    if (defined($elmt_val)) {
       $self->{out_params}->{"shedule"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("snapshot-checkpoint");
    if (defined($elmt_val)) {
       $self->{out_params}->{"snapshot-checkpoint"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("snapshot-progress");
    if (defined($elmt_val)) {
       $self->{out_params}->{"snapshot-progress"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("transfer-snapshot");
    if (defined($elmt_val)) {
       $self->{out_params}->{"transfer-snapshot"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("tries");
    if (defined($elmt_val)) {
       $self->{out_params}->{"tries"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("type");
    if (defined($elmt_val)) {
       $self->{out_params}->{"type"} = $elmt_val;
    }

    $elmt_val  = $sm_info->child_get_string("vserver");
    if (defined($elmt_val)) {
       $self->{out_params}->{"vserver"} = $elmt_val;
    }

}

## @name ::snapmirror_get
## @sypnopsis: Invokes snapmirror-get API and check results 
## @description:
## 
sub snapmirror_get {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %get_req;
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();
    # Prepare enpoint related input parameters 
    $self->prep_std_req();
    # Prepare desired-attributes parameters if specified
    if ( defined ($input_parm{"desired-attributes"})) {
        my @attrs = split /,/, $input_parm{"desired-attributes"};
        my $attr;
        foreach $attr (@attrs) {
            $self->{std_req}->{"desired-attributes"}{"snapmirror-info"}{$attr}="";
        }
    }
    
    # Invoke snapmirror-get API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-get");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();
    if (! defined ($xo)) {
        $self->{last_errmsg} ="No output returned from invoke_zapi\n";
        return $TCD::FAIL;
    }

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    # Save output parameters returned 
    my $attributes = $xo->child_get("attributes");
    if (defined($attributes)) {
        my $sm_info = $attributes->child_get("snapmirror-info");
        if (defined($sm_info)) {
            $self->save_sm_info_out_params($sm_info);
        }
    }  

    # Check content of the result if requested
    if ($self->{check_content} eq "true" ) {
        $result =  $self->check_sm_get_cont($xo);
    }
    return $result;
}

sub check_sm_get_iter_cont {
    my $self = shift;
    my $xo = shift;
    my $attr_list = $xo->child_get("attributes-list");
    if (!defined($attr_list)) {
         $self->{last_errmsg} = "attributes-list missing from snapmirror-get-iter output\n";
         return $TCD::FAIL;
    }    
    my $num_records = $xo->child_get_string("num-records");
    if (!defined($num_records)) {
         $self->{last_errmsg} = "num-records  missing from snapmirror-get-iter output\n";
         return $TCD::FAIL;
    }
    my $max_records;    
    if (defined($self->{query_req}->{"max-records"})) {
        $max_records = $self->{query_req}->{"max-records"};
        if ($num_records > $max_records) {
            $self->{last_errmsg} = "More records returned than expected. " 
                                  . "max-records:$max_records, num-records:$num_records";
            return $TCD::FAIL;
        }
    }

    for (my $i=0; $i < $num_records; $i++) {
        my $sm_info = $attr_list->child_get_nth($i);
        if (!defined($sm_info)) {
            $self->{last_errmsg} = "Cannot extract record $i from snapmirror-get-iter output\n";
            return $TCD::FAIL;
        }    
        my $result;
        # Check desired-attributes if specified
        if (defined ($self->{inquiry_req}->{"desired-attributes"})) {
            $result = $self->check_sm_des_attrs($sm_info, "query_req");
            if ($result eq $TCD::FAIL) {
                return $result;
            }
        }

        #TODO: Check output against query input
    }
    
    return $TCD::PASS;
}

## @name ::snapmirror_get_iter
## @sypnopsis: Invokes snapmirror-get-iter API and check results 
## @description:
## 
sub snapmirror_get_iter {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare query related  input parameters 
    $self->prep_query_req();
    
    # Prepare desired-attributes parameters if specified
    if ( defined ($input_parm{"desired-attributes"})) {
        my @attrs = split /,/, $input_parm{"desired-attributes"};
        my $attr;
        foreach $attr (@attrs) {
            $self->{query_req}->{"desired-attributes"}{"snapmirror-info"}{$attr}="";
        }
    }
    # Prepare other input  parameters if specified
    if ( defined ($input_parm{"max-records"})) {
        $self->{query_req}->{"max-record"} = $input_parm{"max-records"};
    }

    if ( defined ($input_parm{"tag"})) {
        $self->{query_req}->{"tag"} = $input_parm{"tag"};
    }

    # Invoke snapmirror-get-iter API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-get-iter");
    $server->set_zapi_param($self->{query_req});
    my $xo = $server->invoke_zapi();
    if (! defined ($xo)) {
        $self->{last_errmsg} ="No output returned from invoke_zapi\n";
        return $TCD::FAIL;
    }

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    # Check content of the result if requested
    if ($self->{check_content} eq "true" ) {
        $result =  $self->check_sm_get_iter_cont($xo);
    }
    return $result;
}

## @name ::snapmirror_abort
## @sypnopsis: Invokes snapmirror-abort API and check results 
## @description:
## 
sub snapmirror_abort {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    #Prepare input parameters
    $self->prep_std_req();
    if ( defined($input_parm{"check-only"})) {
        $self->{std_req}->{"check-only"} = $input_parm{"check-only"};
    }
    if ( defined($input_parm{"clear-checkpoint"})) {
        $self->{std_req}->{"clear-checkpoint"} = $input_parm{"clear-checkpoint"};
    }

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-abort");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    return $result;
}


## @name ::snapmirror_abort_iter
## @sypnopsis: Invokes snapmirror-abort-iter API and check results 
## @description:
## 
sub snapmirror_abort_iter {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_set_query_req();
    if ( defined($input_parm{"check-only"})) {
        $self->{query_req}->{"check-only"} = $input_parm{"check-only"};
    }
    if ( defined($input_parm{"clear-checkpoint"})) {
        $self->{query_req}->{"clear-checkpoint"} = $input_parm{"clear-checkpoint"};
    }

    #invoke API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-abort-iter");
    $server->set_zapi_param($self->{query_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    # TODO: Check content of the result if requested
    return $result;
}

## @name ::snapmirror_break
## @sypnopsis: Invokes snapmirror-break API and check results 
## @description:
## 
sub snapmirror_break {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();

    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-break");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    return $result;
}

## @name ::snapmirror_break_iter
## @sypnopsis: Invokes snapmirror-break-iter API and check results 
## @description:
## 
sub snapmirror_break_iter {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();
    $self->prep_set_query_req();

    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-break-iter");
    $server->set_zapi_param($self->{query_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }

    if ($self->{exp_result} eq "failed") {
        return $result;
    }
  
    # TODO: Check content of the result if requested
    return $result;
}

## @name ::snapmirror_check
## @sypnopsis: Invokes snapmirror-check API and check results 
## @description:
## 
sub snapmirror_check {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();
    if ( defined ($input_parm{"target-snapshot"})) {
        $self->{std_req}->{"target-snapshot"} = 
			$input_parm{"target-snapshot"};
    }

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-check");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    # Check  job completion
    $result = $self->check_job_completion($xo, "snapmirror-check");
    return $result;
}

## @name ::snapmirror_check_iter
## @sypnopsis: Invokes snapmirror-check-iter API and check results 
## @description:
## 
sub snapmirror_check_iter {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameter
    $self->prep_set_query_req();
    if ( defined ($input_parm{"target-snapshot"})) {
        $self->{query_req}->{"target-snapshot"} = 
			$input_parm{"target-snapshot"};
    }
    
    # invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-create");
    $server->set_zapi_param($self->{query_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    # TODO: Check content of the result if requested
    return $result;
}

## @name ::snapmirror_initialize
## @sypnopsis: Invokes snapmirror-initialize API and check results 
## @description:
## 
sub snapmirror_initialize {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();
    if ( defined ($input_parm{"max-transfer-rate"})) {
        $self->{std_req}->{"max-transfer-rate"} = 
			$input_parm{"max-transfer-rate"};
    }

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-initialize");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    # Check initialize job completion
    $result = $self->check_job_completion($xo, "snapmirror-initialize");
    return $result;
}

## @name ::snapmirror_initialize_ls_set
## @sypnopsis: Invokes snapmirror-initialize-ls-set API and check results 
## @description:
## 
sub snapmirror_initialize_ls_set {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    if ( defined ($input_parm{"source-location"})) {
        $self->{std_req}->{"source-location"} = 
			$input_parm{"source-location"};
    }
    if ( defined ($input_parm{"source-cluster"})) {
        $self->{std_req}->{"source-cluster"} = 
			$input_parm{"source-cluster"};
    }
    if ( defined ($input_parm{"source-vserver"})) {
        $self->{std_req}->{"source-vserver"} = 
			$input_parm{"source-vserver"};
    }
    if ( defined ($input_parm{"source-volume"})) {
        $self->{std_req}->{"source-volume"} = 
			$input_parm{"source-volume"};
    }

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-initialize-ls-set");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    # Check initialize-ls-set job completion
    $result = $self->check_job_completion($xo, "snapmirror-initialize-ls-set");
    return $result;
}

## @name ::snapmirror_update
## @sypnopsis: Invokes snapmirror-update API and check results 
## @description:
## 
sub snapmirror_update {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();
    if ( defined ($input_parm{"max-transfer-rate"})) {
        $self->{std_req}->{"max-transfer-rate"} = 
			$input_parm{"max-transfer-rate"};
    }

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-update");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    # Check update job completion
    $result = $self->check_job_completion($xo, "snapmirror-update");
    return $result;
}

## @name ::snapmirror_update_ls_set
## @sypnopsis: Invokes snapmirror-update-ls-set API and check results 
## @description:
## 
sub snapmirror_update_ls_set {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    if ( defined ($input_parm{"source-location"})) {
        $self->{std_req}->{"source-location"} = 
			$input_parm{"source-location"};
    }
    if ( defined ($input_parm{"source-cluster"})) {
        $self->{std_req}->{"source-cluster"} = 
			$input_parm{"source-cluster"};
    }
    if ( defined ($input_parm{"source-vserver"})) {
        $self->{std_req}->{"source-vserver"} = 
			$input_parm{"source-vserver"};
    }
    if ( defined ($input_parm{"source-volume"})) {
        $self->{std_req}->{"source-volume"} = 
			$input_parm{"source-volume"};
    }

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-update-ls-set");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    # Check initialize-ls-set job completion
    $result = $self->check_job_completion($xo, "snapmirror-update-ls-set");
    return $result;
}

## @name ::snapmirror_promote
## @sypnopsis: Invokes snapmirror-promote API and check results 
## @description:
## 
sub snapmirror_promote {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-promote");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    return $result;
}

## @name ::snapmirror_destroy
## @sypnopsis: Invokes snapmirror-destroy API and check results 
## @description:
## 
sub snapmirror_destroy {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-destroy");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    return $result;
}

sub check_sm_get_vol_status_cont {
    my $self = shift;
    my $xo = shift;

    my $is_source = $xo->child_get_string("is-source");
    if (!defined($is_source)) {
        $self->{last_errmsg} = "is-source missing from snapmirror-get-volume-status output";
        return $TCD::FAIL;
    }

    my $exp_is_source = $self->{exp_vol_status}->{is_source};
    if (($exp_is_source ne $SnapMirrorZapi::ANY) && 
        ($exp_is_source ne $is_source)) {
        $self->{last_errmsg} = "Unexpected is-source value: $is_source," 
        . "expected: $exp_is_source";
        return $TCD::FAIL;
    }

    my $is_dest = $xo->child_get_string("is-destination");
    if (!defined($is_dest)) {
        $self->{last_errmsg} = "is-destination missing from snapmirror-get-volume-status output";
        return $TCD::FAIL;
    }

    my $exp_is_dest = $self->{exp_vol_status}->{is_dest};
    if (($exp_is_dest ne $SnapMirrorZapi::ANY) && 
        ($exp_is_dest ne $is_dest)) {
        $self->{last_errmsg} = "Unexpected is-destination value: $is_dest," 
        . "expected: $exp_is_dest";
        return $TCD::FAIL;
    }

    my $is_xfer = $xo->child_get_string("is-transfer-in-progress");
    if (!defined($is_xfer)) {
        $self->{last_errmsg} = "is-transfer-in-progress missing from " 
               . "snapmirror-get-volume-status output";
        return $TCD::FAIL;
    }

    my $exp_is_xfer = $self->{exp_vol_status}->{is_xfer};
    if (($exp_is_xfer ne $SnapMirrorZapi::ANY) && 
        ($exp_is_xfer ne $is_xfer)) {
        $self->{last_errmsg} = "Unexpected is-transfer-in-progress value: "
              . " $is_xfer, expected: $exp_is_xfer";
        return $TCD::FAIL;
    }

    my $is_broken = $xo->child_get_string("is-transfer-broken");
    if (!defined($is_broken)) {
        $self->{last_errmsg} = "is-transfer-broken missing from " 
               . "snapmirror-get-volume-status output";
        return $TCD::FAIL;
    }

    my $exp_is_broken = $self->{exp_vol_status}->{is_broken};
    if (($exp_is_broken ne $SnapMirrorZapi::ANY) && 
        ($exp_is_broken ne $is_xfer)) {
        $self->{last_errmsg} = "Unexpected is-transfer-broken value: "
              . " $is_broken, expected: $exp_is_broken";
        return $TCD::FAIL;
    }
    return $TCD::PASS; 
}


## @name ::snapmirror_get_volume_status
## @sypnopsis: Invokes snapmirror-get-volume-status API and check results 
## @description:
## 
sub snapmirror_get_volume_status {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    if ( defined($input_parm{"volume"})) {
        $self->{std_req}->{"volume"} = $input_parm{"volume"};
    }

    #invoke API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-get-volume-status");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }

    if ($self->{exp_result} eq "failed") {
        return $result;
    }
  
    # Check content of the result if requested
    if ($self->{check_content} eq "true") {
        $result = $self->check_sm_get_vol_status_cont($xo);
    }
    return $result;
}

sub check_sm_create_cont {
    my $self = shift;
    my $xo = shift;
   
    if ((!defined ($self->{std_req}->{"return-record"})) ||
        ($self->{std_req}->{"return-record"} eq "false")) {
        my $nb_child = $xo->child_max_number();
        if ($nb_child != 1) {
            $self->{last_errmsg} = "snapmirror-create output include unexpected number "
                                 . "of elements:$nb_child, expected:1";
            return $TCD::FAIL;
        }
    }
   
    my $result_elt = $xo->child_get("result");
    if (!defined($result_elt)) {
         $self->{last_errmsg} = "result missing from snapmirror-create output\n";
         return $TCD::FAIL;
    }    
    my $sm_info = $result_elt->child_get("snapmirror-info");
    if (!defined($sm_info)) {
         $self->{last_errmsg} = "snapmirror-info missing from snapmirror-create output\n";
         return $TCD::FAIL;
    }    
    
    ## Check endpoint output parameters against corresponding input parameters
    ## TODO: Check other input parameters as well
    return $self->check_std_output($sm_info);
}

## @name ::snapmirror_create
## @sypnopsis: Invokes snapmirror-create API and check results 
## @description:
## 
sub snapmirror_create {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();
    if ( defined ($input_parm{"max-transfer-rate"})) {
        $self->{std_req}->{"max-transfer-rate"} = 
			$input_parm{"max-transfer-rate"};
    }
    if ( defined ($input_parm{"schedule"})) {
        $self->{std_req}->{"schedule"} = 
			$input_parm{"schedule"};
    }
    if ( defined ($input_parm{"tries"})) {
        $self->{std_req}->{"tries"} = 
			$input_parm{"tries"};
    }
    if ( defined ($input_parm{"type"})) {
        $self->{std_req}->{"type"} = 
			$input_parm{"type"};
    }
    if ( defined ($input_parm{"vserver"})) {
        $self->{std_req}->{"vserver"} = 
			$input_parm{"vserver"};
    }
    if ( defined ($input_parm{"return-record"})) {
        $self->{std_req}->{"return-record"} = 
			$input_parm{"return-record"};
    }

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-create");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }

    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    if ($self->{check_content} eq "true") {
        $result = $self->check_sm_create_cont($xo);
    }
    return $result;
}

## @name ::snapmirror_modify
## @sypnopsis: Invokes snapmirror-modify API and check results 
## @description:
## 
sub snapmirror_modify {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();
    if ( defined ($input_parm{"max-transfer-rate"})) {
        $self->{std_req}->{"max-transfer-rate"} = 
			$input_parm{"max-transfer-rate"};
    }
    if ( defined ($input_parm{"schedule"})) {
        $self->{std_req}->{"schedule"} = 
			$input_parm{"schedule"};
    }
    if ( defined ($input_parm{"tries"})) {
        $self->{std_req}->{"tries"} = 
			$input_parm{"tries"};
    }
    if ( defined ($input_parm{"vserver"})) {
        $self->{std_req}->{"vserver"} = 
			$input_parm{"vserver"};
    }

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-modify");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    return $result;
}

## @name ::snapmirror_quiesce
## @sypnopsis: Invokes snapmirror-quiesce API and check results 
## @description:
## 
sub snapmirror_quiesce {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();

    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-quiesce");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    return $result;
}

## @name ::snapmirror_resume
## @sypnopsis: Invokes snapmirror-resume API and check results 
## @description:
## 
sub snapmirror_resume {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();

    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-resume");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    return $result;
}

## @name ::snapmirror_resync
## @sypnopsis: Invokes snapmirror-resync API and check results 
## @description:
## 
sub snapmirror_resync {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    # Clear buffers
    $self->clear_req_buffers();

    # Prepare input parameters
    $self->prep_std_req();
    if ( defined ($input_parm{"max-transfer-rate"})) {
        $self->{std_req}->{"max-transfer-rate"} = 
			$input_parm{"max-transfer-rate"};
    }

    # Invoke the API
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("snapmirror-resync");
    $server->set_zapi_param($self->{std_req});
    my $xo = $server->invoke_zapi();

    # Process results
    $self->{last_xo} = $xo;
    my $result = $self->check_result($server, $xo, $self->{exp_result}, 
                                     $self->{exp_errno}, $Zapi::ANY);
    if ($result eq $TCD::FAIL) {
        return $result;
    }
  
    if ($self->{exp_result} eq "failed") {
        return $result;
    }

    # Check  resync job completion
    $result = $self->check_job_completion($xo, "snapmirror-resync");
    return $result;
}

sub check_job_completion {
    my $self = shift;
    my $xo = shift;
    my $api_name = shift;

    my $job_id = $xo->child_get_string("result-jobid");
    if (!defined($job_id)) {
       $self->{last_errmsg} = "jobid missing from the output of $api_name";
       return $TCD::FAIL;
    }
    my $res_status = $xo->child_get_string("result-status");
    if (!defined($res_status)) {
       $self->{last_errmsg} = "result-status missing from the output of $api_name";
       return $TCD::FAIL;
    }

    $self->{job_result}->{job_id} = $job_id;
    $self->{job_result}->{result_status} = $res_status;

    # TODO: Process error in result against expected behavior
    my $error_code = $xo->child_get_string("result-error-code");
    if (defined($error_code)) {
        my $err_msg = $xo->child_get_string("result-error-message");
        if (!defined($err_msg)) {
           $err_msg = " ";
        }
        $self->{last_errmsg} = "$api_name caused an error. jobid=$job_id, errno=$error_code,"
                              . " reason=$err_msg";
        return $TCD::FAIL;
    } 
    
   
    if ($self->{job_flags}->{monitor} ne "true") {
        return $TCD::PASS;
    }

    # query job information until completion
    my $result;
    $self->{job_result}->{int_state} = "notdone";
    while($self->{job_result}->{int_state} ne "done") {
        $result = $self->query_job_status($api_name);
        sleep 2;
    }

    if ($result eq $TCD::FAIL) {
        return $result;
    }

    my $job_state = $self->{job_result}->{job_state};
    my $job_status = $self->{job_result}->{job_status};
    my $exp_job_state = $self->{job_flags}->{exp_job_state};
    my $exp_job_status = $self->{job_flags}->{exp_job_status};
    my $job_name = $self->{job_result}->{job_name};
    my $job_compl = $self->{job_result}->{job_completion};

    if ($exp_job_status eq $SnapMirrorZapi::SUCCESS) {
        if($job_status ne "0") {
            $self->{last_errmsg} = "Unexpected job-status-code:$job_status. "
             . "API=$api_name, jobid=$job_id, exp-job-status-code=$exp_job_status,"
             . "job-name=$job_name, job-completion=$job_compl.";
            return $TCD::FAIL;
        }
        if ($exp_job_state eq $SnapMirrorZapi::ANY) {
            return $TCD::PASS;
        } 
    } elsif ($exp_job_status eq $SnapMirrorZapi::FAIL) {
        if($job_status eq "0") {
            $self->{last_errmsg} = "Unexpected job success."
             . "API=$api_name, jobid=$job_id"
             . "job-name=$job_name, job-completion=$job_compl.";
            return $TCD::FAIL;
        }
        if ($exp_job_state eq $SnapMirrorZapi::ANY) {
            return $TCD::PASS;
        } 
    } else {
        if($job_status ne $exp_job_status) {
            $self->{last_errmsg} = "Unexpected job-status-code:$job_status."
             . "API=$api_name, jobid=$job_id, exp-job-status-code=$exp_job_status,"
             . "job-name=$job_name, job-completion=$job_compl.";
            return $TCD::FAIL;
        }
        if ($exp_job_state eq $SnapMirrorZapi::ANY) {
            return $TCD::PASS;
        } 
    }

    if ($job_state ne $exp_job_state) {
        $self->{last_errmsg} = "Unexpected job state: $job_state. API=$api_name,"
              . "jobid=$job_id, exp_state:$exp_job_state, "
              . "job-name=$job_name, job-completion=$job_compl.";
        return $TCD::FAIL; 
    }

    return $TCD::PASS;
}

sub query_job_status {
    my $self = shift;
    my $api_name = shift;
    my $job_id = $self->{job_result}->{job_id};
    my %job_param = ("job-id" => $job_id);
    my $server = new ZapiTest($self->{ip}, $self->{user}, $self->{passwd});
    $server->set_zapi_name("job-get");
    $server->set_zapi_param(\%job_param);
    my $xo = $server->invoke_zapi();
    
    if (!defined($xo)) {
        $self->{last_errmsg} = "Cannot get job information for $api_name";
        $self->{job_result}->{int_state} = "done";
        return $TCD::FAIL;
    }
    (my $result, my $msg)  = $server->check_result($xo, "passed", 
					$self->{exp_errno}, $Zapi::ANY);
    if ($result ne $Zapi::OK) {
        $self->{last_errmsg} = $msg . " API=$api_name";
        $self->{job_result}->{int_state} = "done";
        return $TCD::FAIL;
    }

    $self->{last_job_xo} = $xo;
    my $job_attrs = $xo->child_get("attributes");
    if (!defined($job_attrs)) {
        $self->{last_errmsg} = "attributes missing from output job=$job_id, API=$api_name";
        $self->{job_result}->{int_state} = "done";
        return $TCD::FAIL;
    }
    my $job_info = $job_attrs->child_get("job-info");
    if (!defined($job_info)) {
        $self->{last_errmsg} = "job-info missing from output job=$job_id, API=$api_name";
        $self->{job_result}->{int_state} = "done";
        return $TCD::FAIL;
    }

    my $job_state = $job_info->child_get_string("job-state");
    if (!defined($job_state)) {
       $self->{last_errmsg} = "job-state missing from output job=$job_id, API=$api_name";
        $self->{job_result}->{int_state} = "done";
       return $TCD::FAIL;
    }
    $self->{job_result}->{job_state} = $job_state;
    my $job_status = $job_info->child_get_string("job-status-code");
    if (defined($job_status)) {
       $self->{job_result}->{job_status} = $job_status;
    } 
    my $job_name = $job_info->child_get_string("job-name");
    if (defined($job_name)) {
        $self->{job_result}->{job_name} = $job_name;
    } 
    my $job_completion = $job_info->child_get_string("job-completion");
    if (defined($job_completion)) {
        $self->{job_result}->{job_completion} = $job_completion;
    }

    # check that job state is a final state
    if (($job_state eq "initial") || ($job_state eq "success") ||
        ($job_state eq "failure") || ($job_state eq "dead") ||
        ($job_state eq "error")) {
        $self->{job_result}->{int_state} = "done";
    }
    return $TCD::PASS;
}


sub prep_std_req {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    if ( defined ($input_parm{"destination-location"})) {
        $self->{std_req}->{"destination-location"} = 
			$input_parm{"destination-location"};
    }

    if ( defined ($input_parm{"destination-cluster"})) {
        $self->{std_req}->{"destination-cluster"} = 
			$input_parm{"destination-cluster"};
    }

    if ( defined ($input_parm{"destination-vserver"})) {
        $self->{std_req}->{"destination-vserver"} = 
			$input_parm{"destination-vserver"};
    }

    if ( defined ($input_parm{"destination-volume"})) {
        $self->{std_req}->{"destination-volume"} = 
			$input_parm{"destination-volume"};
    }


    if ( defined ($input_parm{"source-location"})) {
        $self->{std_req}->{"source-location"} = 
			$input_parm{"source-location"};
    }

    if ( defined ($input_parm{"source-cluster"})) {
        $self->{std_req}->{"source-cluster"} = 
			$input_parm{"source-cluster"};
    }

    if ( defined ($input_parm{"source-vserver"})) {
        $self->{std_req}->{"source-vserver"} = 
			$input_parm{"source-vserver"};
    }
    if ( defined ($input_parm{"source-volume"})) {
        $self->{std_req}->{"source-volume"} = 
			$input_parm{"source-volume"};
    }
}


sub prep_query_req {
    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    if ( defined ($input_parm{"destination-location"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"destination-location"} = 
			$input_parm{"destination-location"};
    }

    if ( defined ($input_parm{"destination-cluster"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"destination-cluster"} = 
			$input_parm{"destination-cluster"};
    }

    if ( defined ($input_parm{"destination-vserver"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"destination-vserver"} = 
			$input_parm{"destination-vserver"};
    }

    if ( defined ($input_parm{"destination-volume"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"destination-volume"} = 
			$input_parm{"destination-volume"};
    }


    if ( defined ($input_parm{"source-location"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"source-location"} = 
			$input_parm{"source-location"};
    }

    if ( defined ($input_parm{"source-cluster"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"source-cluster"} = 
			$input_parm{"source-cluster"};
    }

    if ( defined ($input_parm{"source-vserver"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"source-vserver"} = 
			$input_parm{"source-vserver"};
    }

    if ( defined ($input_parm{"source-volume"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"source-volume"} = 
			$input_parm{"source-volume"};
    }
    if ( defined ($input_parm{"exported-snapshot"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"exported-snapshot"} = 
			$input_parm{"exported-snapshot"};
    }
    if ( defined ($input_parm{"exported-snapshot-timestamp"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"exported-snapshot-timestamp"} = 
			$input_parm{"exported-snapshot-timestamp"};
    }
    if ( defined ($input_parm{"is-healthy"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"is-healthy"} = 
			$input_parm{"is-healthy"};
    }
    if ( defined ($input_parm{"max-transfer-rate"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"max-transfer-rate"} = $input_parm{"max-transfer-rate"};
    }
    if ( defined ($input_parm{"mirror-state"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"mirror-state"} = 
			$input_parm{"mirror-state"};
    }
    if ( defined ($input_parm{"newest-snapshot"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"newest-snapshot"} = 
			$input_parm{"newest-snapshot"};
    }
    if ( defined ($input_parm{"newest-snapshot-timestamp"})) {
        $self->{query_req}->{query}{"snapmirror-info"}{"exported-snapshot-timestamp"} = 
		$input_parm{"exported-snapshot-timestamp"};
    }
}

sub prep_set_query_req {

    my $self = shift;
    my $paramref = $self->{req_params};
    my %input_parm = %$paramref;

    $self->prep_query_req();
    if ( defined ($input_parm{"max-records"})) {
        $self->{query_req}->{"max-record"} = $input_parm{"max-records"};
    }

    if ( defined ($input_parm{"tag"})) {
        $self->{query_req}->{"tag"} = $input_parm{"tag"};
    }
    if ( defined ($input_parm{"continue-on-failure"})) {
        $self->{query_req}->{"continue-on-failure"} = $input_parm{"continue-on-failure"};
    }
    if ( defined ($input_parm{"max-failure-count"})) {
        $self->{query_req}->{"max-failure-count"} = $input_parm{"max-failure-count"};
    }
    if ( defined ($input_parm{"return-failure-list"})) {
        $self->{query_req}->{"return-failure-list"} = $input_parm{"return-failure-list"};
    }
    if ( defined ($input_parm{"return-success-list"})) {
        $self->{query_req}->{"return-success-list"} = $input_parm{"return-success-list"};
    }
}

1;
