# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/SnapVault.pm#1 $

## @summary SnapVault utilities library -- action subroutines
##
## @description
##    This package defines routines for using SnapVault.
##    Most of the functions take an established connection object
##    to a filer as their first argument.
##
##    Synchronous commands typically interact with the filer and then
##    return immediately, indicating success or failure and passing
##    back an error message in $@ on error.
##
##    Longer-running and asynchronous commands will interact with the
##    filer on the provided connection, and kick off a backgrounded
##    subtest to continue interacting with the filer and monitor
##    the operation for completion.  Once the operation has
##    successfully started the routine returns the subtest object for
##    the backgrounded subtest.  This way, scripts can invoke
##    long-running or asynchronous SnapVault operations and still
##    do other work while the operations are in progress.
##
##    However, since the backgrounded subtest waiting for operation
##    completion still uses the connection passed in to the original
##    routine, scripts must take care to not to clash with the subtest
##    while it's active.  If the connection is persistent, scripts
##    must not attempt to use it until the backgrounded subtest
##    completes, or else the results are undefined.  If the connection
##    is persay, then it is safe to use the connection while the subtest
##    is still active.
##
##    ZAPI requires to differentiate between primary and secondary filers
##    on its API. This knowledge is not known from the connectrec object,
##    so in order to call the proper API a hash reference could be passed
##    instead of the connectrec object, e.g.,
##
##        snapvault_status_detailed($conn, $dst);
##
##        snapvault_status_detailed({conn => $conn, source => 0}, $dst);
##
##    The hash reference has the connectrec object and the source flag
##    (1 for a primary call, 0 for a secondary call).
##
##    For backwards compatibility if called with just the connectrec
##    object, for any API which runs on both the primary and secondary
##    the snapvault licenses are checked. If filer is licensed as a primary,
##    the primary API is called and if it is licensed as a secondary then
##    the secondary API is called instead. In the case where the filer is
##    licensed for both primary and secondary the call will fall back to
##    use rsh instead and in case where calls are restricted to use ZAPI
##    only then the function call will croak with an error.
##
##    Only functions which are supported on both primary and secondary
##    have this alternate first argument and it is clearly noted under
##    the documentation for each function.
##
## @dependencies
##    lib/SayWrap.pm
##    lib/RegistryCmds.pm
##    FSR/FSRutils.pm
##    FSR/rpl_cmds.pm
##
## @bugs
##    The internal functions for start, update, and resync aren't very strict
##    about the type of output they'll expect, and don't prune for specific
##    error messages.
##
##    Most functions don't do strict params checking, so users can shoot
##    themselves in the foot if they try using internal parameters.
##
## @author dl-dpgqa-automation@netapp.com
## @status Shared

package SnapVault;

use warnings;
use strict;
use Data::Dumper;

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
        conf => [ qw(sv_toggle snapvault_modify
                     snapvault_snap_sched snapvault_snap_unsched) ],
        status => [ qw(snapvault_status snapvault_status_config
                       snapvault_status_detailed sv_detailed_status
                       snapvault_status_snapshot sv_status_snapshot
                       snapvault_destinations snapvault_dst_snap
                       snapvault_status_list) ],
        action => [ qw(snapvault_abort snapvault_abort_snapshot
                       snapvault_release snapvault_convert) ],
        xfer => [ qw(sv_start sv_update sv_resync sv_stop sv_restore
                     snapvault_snap_create sv_snap_create
                     sv_wait_idle) ],
        manual => [ qw(snapvault_start snapvault_update snapvault_resync
                       snapvault_stop snapvault_restore snapvault_wait) ],
        check => [ qw(snapvault_check snapvault_check_abort 
		      snapvault_compare) ],
        utils => [ qw(schedule_info schedule_string
                      use_zapi_rsh use_zapi_only use_rsh_only
		      get_filer_interface set_filer_interface) ]
  );
}

use Tharn;
use Carp;
use SayWrap;
use SnapMirror qw();
use RegistryCmds qw(getOption setOption registryGet);
use FSR::FSRutils qw(:DEFAULT split_filer_path split_vol_path);
use FSR::rpl_cmds qw(:all);
use Date::Manip qw(ParseDateString UnixDate);
use Filer::Version qw(get_wafl_version
		      WV_GORDONBIERSCH_MIN WV_IRONCITY_MIN);
use ATI::ati_ontapi qw(ati_ontapi_create_context
                       ATI_ONTAPI_TRANSPORT_DEFAULT
                       ATI_ONTAPI_DEBUG_NONE);
use fileSystem qw(treecompare);
use QtreeCmds qw(qtreeComponents);


use constant UNLIMITED_STR => "unlimited";
#use constant UNLIMITED_VAL => 4294967295;#(2^32 - 1)
use constant UNLIMITED_VAL => 2147483646; #(2^31 - 2) Per BURT 261796

# Function prototype to force the reference on the argument
sub _get_conn(\$;);
sub _ontapi_support(\$;$);

# Package global variables
# SnapVault command on the filer
my $sv = "snapvault";
# This flag is used to decide how to execute the command, via ZAPI or rsh.
# With a value of 0 only rsh will be used, with a value of 1 only ZAPI will
# be used while a value of 2 (default) will try ZAPI first and if it is not
# available rsh will be used instead.
my $filer_interface = 2;

# This flag is used to decide if the snapvault destination is vfiler.
# By default this flag is set to zero indicating that the snapvault destination
# is pfiler. This flag has only been added to be used by the 
# NOTE::Vfiler::Snapvault library.
# Beware of its perils, and use at your risk..!
my $vfiler_dst = 0;

# Vfiler arguments are declared as global arguments here. This is because though
# these arguments are only needed by the internal function _ontapi_support, this
# function itself is called many times through various methods to be able to 
# pass theses arguments through all of them to reach this method.
# So making them global instead.
my %vfiler_args;
  
# Translation table CLI -> ZAPI
# This table is used to translate the arguments and outputs from the
# CLI interface to ZAPI outputs, so all functions take and return
# the same arguments and outputs independent of interface.
my %zapi_h = (
     # Status
    'Contents'               => { value => sub { $_[0] } },
    'Base Snapshot'          => { value => sub { $_[0] } },
    'Last Transfer Duration' => { value => \&_convert_time2secs },
    'Last Transfer Size'     => { value => sub { $_[0] =~ /^(\d+)/; $1 } },
    'Source'                 => { value => \&_host_path },
    'Destination'            => { value => \&_host_path },
    'Last Transfer From'     => { value => \&_host_path },
    'Lag' => { key => 'lag-time', value => \&_convert_time2secs },
    'Progress' => {
        key   => 'transfer-progress',
        value => sub { $_[0] =~ /^(\d+)/; $1 }
    },
    'Mirror Timestamp' => {
        value => sub { UnixDate(ParseDateString($_[0]), '%s') }
    },

    # Configuration
    'tries'       => { key => 'tries-count' },
    'tries_count' => { key => 'tries-count' },
    'dstpath'     => { key => 'secondary-path' },
    'kbs'         => {
	key => 'max-transfer-rate',
	value => sub { return lc($_[0]) eq UNLIMITED_STR ? UNLIMITED_VAL : $_[0]; }
    },
    'source'      => { key => 'primary', value => \&_host_path },
    'back_up_open_files' => {
        key   => 'is-open-file-backup-allowed',
        value => sub { $_[0] =~ /(on|1)/ ? 'true' : 'false' }
    },
    'compression'	=> { key => 'use-compression' },
    'connection_mode'   => { key => 'connection-mode' },
    'ignore_atime'      => { key => 'is-access-time-change-ignored' ,
                             value => sub { $_[0] =~ /(on|1)/  ? 'true' : 'false' }},
    'utf8_primary_path' => { key => 'is-primary-path-utf8-encoded',
                             value => sub { $_[0] =~ /(on|1)/  ? 'true' : 'false' }},  

    # Setting configuration
    'throttle' => {
	key => 'max-transfer-rate',
	value => sub { return lc($_[0]) eq UNLIMITED_STR ? UNLIMITED_VAL : $_[0]; }
    }
);

## @name ::sv_toggle
## @usage sv_toggle($filerConn, $on)
## @synopsis
##    The second argument ($on) is optional.  If specified, then this
##    function turns SnapVault on or off, depending upon the value of
##    $on, and returns success or failure, with an error message in $@
##    on failure.  If $on is not provided, then this function instead
##    determines and returns whether or not SnapVault is currently
##    turned on.  On error, an error message is returned in $@.
sub sv_toggle($;$) {
    my ($conn, $on) = @_;
    my $sv_key = "snapvault.enable";
    my ($context, $api, $args, $value, @out);

    return eval {
        $value = ($on ? 'on' : 'off') if defined $on;

        if ( $context = _ontapi_support($conn, 1) ) {
            # ZAPI - Enable or get status of SnapVault
            if ( defined $on ) {
                $api = 'options-set';
                $args = [ name => $sv_key, value => $value ];
            } else {
                $api = 'options-get';
                $args = [ name => $sv_key ];
            }
            @out = SnapMirror::_invoke_ontapi(context => $context,
                                              api     => $api,
                                              args    => $args);

            $value = $out[1]{value} if ! defined $on;
        } else {
            # CLI - Enable or get status of SnapVault
            if ( defined $on ) {
                setOption($conn, $sv_key, $value);
                croak $@ if ($@);
            } else {
                $value = getOption($conn, $sv_key);
                croak $@ if ($@);
            }
        }

        # Return success when setting option
        return 1 if defined $on;

        # Return a boolean
        if ($value eq 'on') {
            return 1;
        } elsif ($value eq 'off') {
            return 0;
        } else {
            croak "$sv_key=$value";
        }
    };
}

## @name ::snapvault_status
## @usage snapvault_status($filerConn, $dst, $src)
## @synopsis
##    Returns a list that comprises the SnapVault status entry
##    for the specified source/destination combination.
##    If the source is unspecified, it returns the first entry with
##    the given destination (and presumably there's only one anyway).
##    Returns an empty list if there is no status entry for the given destination
##    (and possibly source).  On failure, returns an error message in $@.
## @example
##    my ($src, $dst, $state, $lag, $status) = snapvault_status($filerConn, $dst, $src);
sub snapvault_status($$;$) {
    my ($conn, $dst, $src) = @_;
    return rpl_status_internal($sv, $conn, $dst, $src);
}

{
    my $pair_pat = '([A-Za-z_]+)\=(\S+)';
    my $conf_pat = '\A(\S+)((?: ' . $pair_pat . ')*)\s*(?:\n|\z)';

    # Strips a conf entry line from the start of $_[0]
    # and returns a hash for the parsed conf entry.
    # Croaks if the entry cannot be recognized or parsed.
    sub parse_conf_entry {
        unless ($_[0] =~ s/$conf_pat//o) {
            croak "Malformed configuration entry: " . $_[0] . "\n";
        }
        my ($dstpath, $pairs) = ($1, $2, $3);
        my %conf = (dstpath => $dstpath);

        $pairs =~ s/\A //; # strip leading space
        foreach my $pair (split(/ /, $pairs)) {
            unless ($pair =~ /\A$pair_pat\z/o) {
                croak "Malformed configuration option: $pair\n";
            }
	        if($pair =~ /($pair_pat,\s*)*/){ 
			    my @key_value_pairs = split(/,/,$pair);
			    foreach my $key_value_pair (@key_value_pairs){
			        my ($key, $value) = split(/=/,$key_value_pair);
                    $value = 0 if ($value eq "off");
                    $value = 1 if ($value eq "on");
                    $conf{$key} = $value;
			    }
			}else{
			    my ($key, $value) = ($1, $2);
                $value = 0 if ($value eq "off");
                $value = 1 if ($value eq "on");
                $conf{$key} = $value;  
            }
		}

        # Some code depends on having a source.
        unless (exists $conf{source}) {
            croak "No source in configuration entry: $pairs\n";
        }
        return %conf;
    }
}

## @name ::snapvault_status_config
## @usage snapvault_status_config($conn, %options)
## @synopsis
##    Given a connection to a filer, obtains and returns a list of SnapVault
##    configuration destinations.
##
##    If the details option is specified, returns a list of SnapVault
##    configuration entries.
##
##    If the destination path is specified, returns a hash of all the
##    configuration values for the specified destination, or an empty
##    hash if the destination is not configured.
##
##    If the destination path is not specified, returns a list of all
##    configured destination paths.
##
##    On any error, returns undef with an error message in $@.
##
## @arg details Optional: Returns a list of configuration entries if
##    only the connection to the filer is given.
##
## @arg dstpath Optional: the destination path
##
## @arg source Optional: The configured source, specified as <filer>:<path>
##      Causes the function to only return entries with this source.
##
## @example
##    my %conf = snapvault_status_config($conn,
##                                       dstpath => $dstpath,
##                                       source => $src);
##    if (scalar keys %conf == 0) {
##      # $dstpath is not configured
##    } else {
##      # examine $conf{dstpath} $conf{source}, $conf{kbs}, etc.
##    }
sub snapvault_status_config( $;% ) {
    my ($conn, %opts) = @_;
    my ($context, $api, %args, @out, $list, $conf);
    my ($cmd, $text, $ret, $source);

    return eval {
        if ( $context = _ontapi_support($conn, 1) ) {
            # ZAPI - Get list of all configurations
            if ( $opts{dstpath} ) {
                # Only the configuration of the specified destination
                $api = 'snapvault-secondary-get-configuration';
                $args{args} = [ 'secondary-path' => $opts{dstpath} ];
                $conf = 'configuration';
            } else {
                # Get all configurations
                $api = 'snapvault-secondary-configuration-list-info';
                $conf = 'configurations';
            }
            @out = eval { SnapMirror::_invoke_ontapi(context => $context,
                                                     api     => $api, %args) };
            return () if ($@ =~ /No primaries are configured/);
            croak $@ if $@;
            $list = $out[1]{$conf};
            $list = [] if ref $list ne 'ARRAY';
        } else {
            # CLI - Get list of all configurations
            $cmd = "snapvault status -c";
            if ( $opts{dstpath} ) {
                # Only the configuration of the specified destination
                $cmd .= " " . $opts{dstpath};
            }
            $text = sayEOF($conn, $cmd);
            $text = rpl_strip_header($sv, $text);

            # No status entry for the given dst or dst/src
            if ( $text eq "No primaries are configured." ) {
                return ();
            }

            # Make result consistent with ZAPI results
            while ( $text ) {
                $conf = { parse_conf_entry($text) };
                push @$list, _transform_data_struct($conf);
                if ( exists $list->[-1]{'max-transfer-rate'}
                  && $list->[-1]{'max-transfer-rate'} eq UNLIMITED_STR ) {
                    $list->[-1]{'max-transfer-rate'} = UNLIMITED_VAL;
                }
            }
        }

        if ( $opts{dstpath} ) {
            # Just return the configuration data for the specified destination
            $ret = {};
            foreach $conf ( @$list ) {
                # Host path of the source
                $source = $conf->{'primary-system'}.':'.$conf->{'primary-path'};
                if ( $opts{dstpath} eq $conf->{'secondary-path'}
                  && ( !$opts{source} || $opts{source} eq $source ) ) {
                    if (%$ret) {
                        croak "Duplicate destination entries found"
                    }
                    $ret = $conf;
                }
            }
            return %$ret;
        } elsif ( $opts{details} ) {
            # No destination path specified, so return a list of configs.
            return @$list;
        } else {
            # No destination path and no details specified,
            # so return a list of destinations.
            return map { $_->{'secondary-path'} } @$list;
        }
    };
}

## @name ::snapvault_status_detailed
## @usage snapvault_status_detailed($filerConn, $dst, $src)
## @synopsis
##    Obtains and returns the detailed status entry for a SnapVault
##    relationship as a hash reference.
##    If the source is unspecified, it returns the first entry with
##    the given destination (and presumably there's only one anyway).
##    Replaces fields that are "-" with "" (that way the keys is still there,
##    but you can also do a truth test to see if the field was set).
##    Returns undef if there was no status entry for the given destination
##    (and possibly source).  On failure, returns an error message in $@.
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_status_detailed($$;$) {
    my ($conn_h, $dst, $src) = @_;
    my ($context, $api, $args, $type, @out);
    my ($status, $path, $host, $conn);

    return eval {
        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Get status of relationship
            ($conn, $type) = _get_conn($conn_h);
            $api = "snapvault-$type-get-relationship-status";
            $args = [ 'system-path' => $dst ];
            @out = SnapMirror::_invoke_ontapi(context => $context,
                                              api     => $api,
                                              args    => $args);
            $out[1]{status} = [] if ref $out[1]{status} ne 'ARRAY';
            if ( defined $src ) {
                ($path, $host) = FSR::FSRutils::split_filer_path($src);
                foreach $status ( @{$out[1]{status}} ) {
                    if ( $status->{'source-system'} eq $host
                      && $status->{'source-path'} eq $path ) {
                        return $status;
                    }
                }
                # Source was not found on list
                return undef;
            } else {
                $status = $out[1]{status}[0];
            }
        } else {
            # CLI - Get status of relationship and convert it
            $conn = _get_conn($conn_h);
            $status = rpl_details_internal($sv, $conn, $dst, $src);
            croak $@ if $@;
            $status = _transform_data_struct($status);
        }

        return $status;
    };
}

## @name ::snapvault_status_list
## @usage snapvault_status_list($filerConn)
## @synopsis
##    Returns a reference to an array of the detailed status entries for all
##    SnapVault relationships.
##    Returns a reference to an empty array if no status entries were found.
##    On failure, returns an error message in $@.
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_status_list($;) {
    my ($conn_h) = @_;
    my ($context, $api, $args, $type, @out, $tag);
    my (@stats, @list, $status, $conn);

    return eval {
        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Get status of relationship
            ($conn, $type) = _get_conn($conn_h);

            # Request start of iteration through the list of status entries
            $api = "snapvault-$type-relationship-status-list-iter-start";
            @out = SnapMirror::_invoke_ontapi(context => $context,
                                              api     => $api);
            $tag = $out[1]{tag};
            if ( $tag && $out[1]{records} ) {
                # Request all status entries 100 at a time
                $api = "snapvault-$type-relationship-status-list-iter-next";
                $args = [ 'tag' => $tag, 'maximum' => 100 ];
                do {
                    @out = SnapMirror::_invoke_ontapi(context => $context,
                                                      api     => $api,
                                                      args    => $args);
                    push @stats, @{$out[1]{'status-list'}}
                      if ref $out[1]{'status-list'} eq 'ARRAY';
                } while ( 0 && $out[1]{records} );
            }

            $api = "snapvault-$type-relationship-status-list-iter-end";
            $args = [ 'tag' => $tag ];
            # Terminate the status list iteration
	    # Burt 237949: API does not work right
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);
        } else {
            # CLI - Get status of relationship and convert it
            $conn = _get_conn($conn_h);
            @list = rpl_details_internal($sv, $conn);
            croak $@ if $@;
            foreach $status ( @list ) {
                push @stats, _transform_data_struct($status);
            }
        }

        return \@stats;
    };
}

## @name ::sv_detailed_status
## @usage sv_detailed_status($filerConn, $dst, $src)
## @synopsis
##    A simple wrapper for snapvault_status_detailed(), for scripts that
##    always want to fail if they can't get the status of a relationship.
##    Logs a fatal result if it cannot obtain the status entry.
##    Otherwise, returns the status hash reference.
sub sv_detailed_status($$;$) {
    my ($conn, $dst, $src) = @_;
    my $statref = snapvault_status_detailed($conn, $dst, $src);
    logresult('FATAL', "EVAL_ERROR=$@") if ($@);
    $src = '' unless defined($src);
    logresult('FATAL', "EVAL_ERROR=No status entry for $src -> $dst")
      unless ($statref);
    return $statref;
}

## @name ::snapvault_status_snapshot
## @usage snapvault_status_snapshot($filerConn, %options)
## @synopsis
##    Returns a list of configured SnapVault snapshot targets for the filer.
##    Returns undef if the snapshot target is not configured.
##    On failure, returns an error message in $@.
## @param VOLUME Optional: get list of schedules for this volume only.
## @param SNAPSHOT Optional: get the schedule for this volume/snapshot only.
## @param STATUS_ONLY Optional: get the schedule status only.
## @example
##    my @list = snapvault_status_snapshot($filerConn);
##    my @list = snapvault_status_snapshot($filerConn, VOLUME=>$volname);
##    my $sched_info = snapvault_status_snapshot($filerConn,
##                                               VOLUME=>$volname,
##                                               SNAPSHOT=>$snapname);
##    my $sched_info = snapvault_status_snapshot($filerConn,
##                                               VOLUME=>$volname,
##                                               SNAPSHOT=>$snapname,
##                                               STATUS_ONLY=>1);
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_status_snapshot($;%) {
    my ($conn_h, %param) = @_;
    my ($context, $api, $args, @out, $type);
    my ($cmd, $options, $text, $sched_list, $list, $item, $itm);
    my ($vol, $snap, $regex, $status, $index, %opts, $conn);

    return eval {
        croak "Option 'SNAPSHOT' requires 'VOLUME' to be defined"
          if ( exists $param{SNAPSHOT} && ! exists $param{VOLUME} );

        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Get snapshot schedule status
            ($conn, $type) = _get_conn($conn_h);
            $args = $param{VOLUME} ? [ 'volume-name' => $param{VOLUME} ] : [];
            $api = "snapvault-$type-snapshot-schedule-status-list-info";
            @out = eval { SnapMirror::_invoke_ontapi(context => $context,
                                                     api     => $api,
                                                     args    => $args) };
            return if ($@ =~ 'No snapshot targets are configured');
            croak $@ if $@;
            # Get schedule list (only need the status from this list)
            $sched_list = $out[1]{'snapshot-schedule-status'};
            $sched_list = [] if ref $sched_list ne 'ARRAY';
            if ( ! $param{STATUS_ONLY} ) {
                $api = "snapvault-$type-snapshot-schedule-list-info";
                @out = SnapMirror::_invoke_ontapi(context => $context,
                                                  api     => $api,
                                                  args    => $args);
                # Merge both lists
                $list = $sched_list;
                $sched_list = $out[1]{'snapshot-schedules'};
                $sched_list = [] if ref $sched_list ne 'ARRAY';
                foreach $item ( @$sched_list ) {
                    $index = 0;
                    foreach $itm ( @$list ) {
                        if ( $itm->{'schedule-name'} eq $item->{'schedule-name'}
                          && $itm->{'volume-name'} eq $item->{'volume-name'} ) {
                            $item = { %$item, %$itm };
                            splice(@$list, $index, 1);
                            $item->{'schedule'} = $item->{'schedule'}[0];
                            last;
                        }
                        $index++;
                    }
                }
            }
        } else {
            # CLI - Get snapshot schedule status
            $conn = _get_conn($conn_h);
            $options = '';
            $options .= ' ' . $param{VOLUME} if $param{VOLUME};
            # The snapshot name could be the empty '' string
            if ( defined $param{SNAPSHOT} ) {
                $options .= ' ' .
                  (($param{SNAPSHOT} eq '') ? '""' : $param{SNAPSHOT});
            }

            # Run command
            $cmd = "$sv status -s$options";
            $text = sayEOF($conn, $cmd);

            $text = rpl_strip_header($sv, $text);

            return if ($text eq "No snapshot targets are configured.");

            croak "Malformed output: $text\n"
              if ! ($text =~ s/^Volume\s+Snapshot\s+Status\s+Schedule\s*(\n|\z)//);
            croak "Malformed output: $text\n"
              if ! ($text =~ s/^-+\s*\-+\s*\-+\s*\-+\s*(\n|\z)//);

            $regex = qr{^(\S+)\s+(\S*)\s+(\S+)\s+(\S+)\s*(\S*)\s*(\n|\z)};
            while ($text) {
                # It's legal for the second field (snapshot name) to be the
                # empty string, but we still want to parse it that way, and
                # keep the third field where it is.  Hence, the \S* surrounded
                # by mandatory whitespace should force this behavior.
                croak "Malformed output: $text\n"
                  if ! ($text =~ s/$regex//);

                %opts = $param{STATUS_ONLY} ? () : %{schedule_info($4)};

                push @$sched_list, {
                    'volume-name'   => $1,
                    'schedule-name' => $2,
                    'status'        => $3,
                    %opts
                };
            }

            if ( ! $param{STATUS_ONLY} ) {
                # Get the type of schedule
                $text = sayEOF($conn, "$sv snap sched$options");
                $regex = qr{^(\S+)\s+(\S+)\s+(\S*)\s+(\S+)\s*(\S*)\s*(\n|\z)}; 
                while ($text) {
                    croak "Malformed output: $text\n"
                      if ! ($text =~ s/$regex//);
                    ($type, $vol, $snap) = ($1, $2, $3);
                    foreach $item ( @$sched_list ) {
                        if ( $snap eq $item->{'schedule-name'}
                          && $vol eq $item->{'volume-name'} ) {
                            $item->{'is-auto-update'} = $type eq 'xfer' ?
                                                        'true' : 'false';
                            last;
                        }
                    }
                }
            }
        }

        # Select correct entries from list according to input parameters
        # only if the SNAPSHOT parameter is specified since the ZAPI
        # API does not have the schedule name as a parameter
        $list = [];
        if ( defined $param{SNAPSHOT} ) {
            foreach $item ( @$sched_list ) {
                if ( $item->{'volume-name'} eq $param{VOLUME}
                  && ( ! defined $param{SNAPSHOT}
                       || $item->{'schedule-name'} eq $param{SNAPSHOT} ) ) {
                    push @$list, $item;
                }
            }
        } else {
            $list = $sched_list;
        }

        # Return a hash reference if SNAPSHOT is specified
        # otherwise, return a list
        return defined $param{SNAPSHOT} ? $list->[0] : @$list;
    };
}

## @name ::sv_status_snapshot
## @usage sv_status_snapshot($filerConn, $volname, $snapname)
## @synopsis
##    A simple wrapper for snapvault_status_snapshot(), for scripts that
##    always want to fail if they can't get the status of a snapshot target.
##    Logs a fatal result if it cannot obtain the status for the given
##    snapshot on the given volume. Otherwise, returns the status.
sub sv_status_snapshot($$$) {
    my ($conn, $volname, $snapname) = @_;
    my $sched = snapvault_status_snapshot($conn,
                                          VOLUME => $volname,
                                          SNAPSHOT => $snapname,
                                          STATUS_ONLY => 1);
    logresult('FATAL', "EVAL_ERROR=$@") if ($@);
    logresult('FATAL', "EVAL_ERROR=snapshot target is not configured")
      unless $sched;
    return $sched->{status};
}

## @name ::snapvault_destinations
## @usage snapvault_destinations($filerConn, %options)
## @synopsis
##    Return a list of all snapvault destinations for the given source
##    If source is unspecified, returns a reference to a hash.
##    Each key in the hash is the source and the value is the list
##    of destinations for that source.
##    On failure, returns an error message in $@.
## @arg source Optional: The configured source, specified as <filer>:<path>
##      causes the function to only return destinations for this source.
## @example
##    my $dest = snapvault_destinations($conn);
##    my @dest = snapvault_destinations($conn, source => $src);
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_destinations($;%) {
    my ($conn_h, %opts) = @_;
    my ($context, $api, %args, @out, $type);
    my (%dests, $src, $list, $item, $conn);

    # Get the source path from the argument
    ($src) = FSR::FSRutils::split_filer_path($opts{source}) if exists $opts{source};

    return eval {
        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Get destinations from filer
            ($conn, $type) = _get_conn($conn_h);
            $api = "snapvault-$type-destinations-list-info";
            $args{args} = [ 'source-path' => $src ] if $src;
            @out = SnapMirror::_invoke_ontapi(context => $context,
                                              api     => $api,
                                              %args);
            $list = $out[1]{destinations};
            $list = [] if ref $list ne 'ARRAY';
            foreach $item ( @$list ) {
                $dests{$item->{'source-path'}} = [
                    map { $_->{'destination-system'} . ':' .
                          $_->{'destination-path'}
                        } @{$item->{'chained-destinations'}}
                ];
            }
            @out = $src ? $dests{$src} ? @{$dests{$src}} : () : ( \%dests );
        } else {
            # CLI - Get destinations from filer
            $conn = _get_conn($conn_h);
            @out =  rpl_destinations_internal($sv, $conn, $src);
            croak $@ if $@;
        }

        if ( $src ) {
            return @out;
        } else {
            return $out[0];
        }
    };
}

## @name ::snapvault_dst_snap
## @usage snapvault_dst_snap($filerConn, $src, $dst)
## @synopsis
##    Obtains and returns the name of the snapshot being retained for
##    the given source/destination pair.  Returns undef if there is no
##    entry for the given source and destination.
##    On failure, returns an error message in $@.
##    The destination must be specified as <filer>:<path>.
##    Note that this function returning a snapshot name in no way guarantees
##    that the snapshot still exists -- the destinations entry hangs around
##    until a release occurs.
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_dst_snap($$$) {
    my ($conn_h, $src, $dst) = @_;
    my ($context, $api, $args, @out, $type, $snapname, $conn);

    # Get the source path from the argument
    ($src) = FSR::FSRutils::split_filer_path($src);

    return eval {
        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Get snapshot of relationship from filer
            ($conn, $type) = _get_conn($conn_h);
            $api = "snapvault-$type-destinations-list-info";
            $args = [ 'source-path' => $src ];
            @out = SnapMirror::_invoke_ontapi(context => $context,
                                              api     => $api,
                                              args    => $args);
            $out[1]{destinations} = [] if ref $out[1]{destinations} ne 'ARRAY';
            $snapname = $out[1]{destinations}[0]{'source-snapshot'};
        } else {
            # CLI - Get snapshot of relationship from filer
            $conn = _get_conn($conn_h);
            $snapname = rpl_dst_snap_internal($sv, $conn, $src, $dst);
            croak $@ if $@;
        }

        return $snapname;
    };
}

## @name ::snapvault_snap_sched
## @usage snapvault_snap_sched($filerConn, $volname, $snapname, $sched, $xfer)
## @synopsis
##    Configures or lists the SnapVault backup schedule for a particular
##    volume.  The filer connection and volume name are the only required
##    arguments, and the behavior of the function depends upon how many
##    if any of the optional arguments are provided.  Note, it is legal
##    for a SnapVault snapshot target name to be the empty string.
##    On failure (in all cases) returns undef with error message in $@.
##    Case 1: With just the filer connection and volume name, obtains and
##            returns a list of all the scheduled snapshots for the volume.
##    Case 2: With the filer connection, volume name, and snapshot
##            name, returns the SnapVault schedule entry for the given
##            snapshot on the given volume as a reference to a hash,
##            or undef if the snapshot is not configured.  (See the examples.)
##    Case 3: With the filer connection, volume name, snapshot name, and
##            schedule (as a string or a hash reference), sets the schedule
##            for the given snapshot. The optional $xfer argument in this
##            case is a boolean that specifies whether to configure the
##            snapshot target as a "xfer" snapshot target or a "create"
##            snapshot target.
## @example
##    my @scheduled_snaps = snapvault_snap_sched($filerConn, $volname);
##    my $sched_h = snapvault_snap_sched($filerConn, $volname, $snapname);
##    my ($type, $volname, $snapname, $sched) =
##          snapvault_snap_sched($filerConn, $volname, $snapname);
##    snapvault_snap_sched($filerConn, $volname, $snapname, $sched, 1)
##      or die "Error configuring snapshot target: $@";
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_snap_sched($$;$$$) {
    my ($conn_h, $volname, $snapname, $sched, $xfer,$preserve,$warn) = @_;
    my ($context, $api, $args, @out, $type, $sched_h, $regex);
    my ($sched_list, $list, $item, $volexp, $cmd, $text, $conn,$sub_command);
    my $o = 0;
	
    return eval {
        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Get/set snapshot schedule
            ($conn, $type) = _get_conn($conn_h);
            if ( defined $sched ) {
                $api = "snapvault-$type-set-snapshot-schedule";
                if ( ref($sched) eq 'HASH' ) {
                    $sched_h = $sched;
                } else {
                    # Convert schedule string into a hash
                    $sched_h = schedule_info($sched);
                }
                # Convert schedule hash into array
                $sched_h->{'schedule'} = [ %{$sched_h->{'schedule'}} ];
                $sched_h->{'volume-name'} = $volname;
                $sched_h->{'schedule-name'} = $snapname;
                $sched_h->{'is-auto-update'} = ($xfer ? 'true' : 'false')
                  if $type eq 'secondary';
		$sched_h->{'preserve-snapshots'} = $preserve 
		  if $type eq 'secondary' && defined $preserve;
                $sched_h->{'warn-at-count'} = $warn
                  if $type eq 'secondary' && defined $warn;				
                $args = [ 'snapshot-schedule' => [ %$sched_h ] ];
            } else {
                $api = "snapvault-$type-snapshot-schedule-list-info";
                $args = [ 'volume-name' => $volname ];
            }			
            @out = SnapMirror::_invoke_ontapi(context => $context,
                                              api     => $api,
                                              args    => $args);
            if ( ! defined $sched ) {
                # Get schedule list
                $sched_list = $out[1]{'snapshot-schedules'};
                $sched_list = [] if ref $sched_list ne 'ARRAY';
                $_->{'schedule'} = $_->{'schedule'}[0] foreach @$sched_list;
            }
        } else {
            # CLI - Get/set snapshot schedule
            $conn = _get_conn($conn_h);
            $volexp = quotemeta($volname);

            if ( ref($sched) eq 'HASH' ) {
                # Change the input schedule structure to a string
                $sched = schedule_string($sched);
            }
			
	    if( defined $preserve ) {
	        $sub_command = " -o preserve=$preserve";
		$o = 1;
	    }
	
	    if(defined $warn){
		if($o){
		    $sub_command .= ",warn=$warn";
		}else{
		    $sub_command = " -o warn=$warn";
		    $o = 1;
		}
            }

            $cmd = "snapvault snap sched -f" .
                      ($xfer ? " -x" : "") . ($o ? $sub_command : "").
                      " $volname" .
                      (defined $snapname ? (($snapname eq "") ? ' ""' :
                                           " $snapname") : "") .
                      ($sched ? " $sched" : "");

            $text = sayEOF($conn, $cmd);

            return if ($text =~ /No snapshots scheduled for volume $volname\b/);

            # Check error when setting the schedule
            croak $text if (defined $sched && $text);

             $regex = qr{^(\S+)\s+(\S+)\s+(\S*)\s+(\S+)\s*(\S*)\s*(\n|\z)};
             while ($text) {
                # It's legal for the third field (snapshot name) to be the
                # empty string, but we still want to parse it that way,
                # and keep the fourth field where it is.  Hence, the \S*
                # surrounded by mandatory whitespace should force this
                # behavior.
                croak "Malformed output: $text\n"
                  if ! ($text =~ s/$regex//);
                push @$sched_list, {
                    'volume-name' => $volname,
                    'schedule-name' => $3,
                    'is-auto-update' => ($1 eq 'xfer' ? 'true' : 'false'),
                    %{schedule_info($4)}
                };
            }
        }

        # Select correct entries from list according to input parameters
        # only if the snapname parameter is specified since the ZAPI
        # API does not have the schedule name as a parameter
        if ( defined $snapname ) {
            $list = [grep { $_->{'schedule-name'} eq $snapname } @$sched_list];
        } else {
            $list = $sched_list;
        }

        return defined $sched ? 1 : defined $snapname ?
               (wantarray ? _snap_sched_out($list) : $list->[0]) : @$list;
    };
}

## @name ::snapvault_snap_unsched
## @usage snapvault_snap_unsched($filerConn, $volname, $snapname)
## @synopsis
##    Given the volume and snapshot target name, unsets the SnapVault
##    backup schedule.  Returns success or failure, with error message
##    in $@ on failure.
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_snap_unsched($$$) {
    my ($conn_h, $volname, $snapname) = @_;
    my ($context, $api, $args, $type, $cmd, $text, $conn);

    return eval {
        croak "Snapshot name not defined" unless (defined $snapname);

        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Get/set snapshot schedule
            ($conn, $type) = _get_conn($conn_h);
            $api = "snapvault-$type-delete-snapshot-schedule";
            $args = [ 'volume-name'   => $volname,
                      'schedule-name' => $snapname ];
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);
        } else {
            # CLI - Get/set snapshot schedule
            $conn = _get_conn($conn_h);
            $snapname = '""' if ($snapname eq "");

            $cmd = "snapvault snap unsched -f $volname $snapname";

            # Try unsetting the schedule
            $text = sayEOF($conn, $cmd);
            croak $text if $text;
        }

        return 1;
    };
}

## @name ::snapvault_abort
## @usage snapvault_abort($filerConn, $dst, %options)
## @synopsis
##    Aborts an active SnapVault transfer.
##    Returns success or failure, with an error message in $@ on failure.
## @param HARD Optional: if set, makes sure the transfer is not restartable
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_abort($$;%) {
    my ($conn_h, $dst, %param) = @_;
    my ($context, $api, $args, $type, $ret, %opts, $conn, $dstname, $dstpath);

    return eval {
        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Abort transfer
            ($conn, $type) = _get_conn($conn_h);
            $api = "snapvault-$type-abort-transfer";
            # Add the hard abort option if specified
            $opts{'is-hard-abort'} = $param{HARD} ? 'true' : 'false';
	    ($dstname, $dstpath) = split(":", $dst);
	    $args = [ "target-system" => $dstname,
	 	      "system-path" => $dstpath,
                      %opts ];
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);
            return 1;
        } else {
            # CLI - Abort transfer
            $conn = _get_conn($conn_h);
            $ret = rpl_abort_internal($sv, $conn, $dst, %param);
            if ( $@ =~ /This may take a long time to complete/ ) {
                $@ = '';
                return 1;
            }
            croak $@ if $@;
            return $ret;
        }
    };
}

## @name ::snapvault_abort_snapshot
## @usage snapvault_abort_snapshot($filerConn, $volname, $snapname)
## @synopsis
##    Aborts the creation of a snapshot (i.e. an active snapshot target).
##    Returns success or failure, with an error message in $@ on failure.
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_abort_snapshot($$$) {
    my ($conn_h, $volname, $snapname) = @_;
    my ($context, $api, $args, $type, $text, $conn);

    return eval {
        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Abort snapshot creation
            ($conn, $type) = _get_conn($conn_h);
            $api = "snapvault-$type-abort-snapshot-create";
            $args = [ 'volume-name' => $volname,
                      'schedule-name' => $snapname ];
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);
        } else {
            # CLI - Abort snapshot creation
            $conn = _get_conn($conn_h);
            croak "Snapshot name not defined"
              unless (defined $snapname);
            $snapname = '""' if ($snapname eq "");

            $text = sayEOF($conn, "$sv abort -s $volname $snapname");
            croak $text if $text;
        }

        return 1;
    };
}

## @name ::snapvault_modify
## @usage snapvault_modify($conn, $dstpath, %options)
## @synopsis
##    Given a connection to a filer and a configured destination path,
##    modifies the existing SnapVault configuration for the path.
##
##    Returns success or failure, with error message in $@ on failure.
##
##    All options are just the configuration values.
## @example
##    if (snapvault_modify($conn, $dstpath, source => "$host:$path")) {
##        # Successfully updated configuration entry.
##    else {
##        # Could not modify configuration entry, error is in $@.
##    }
##    snapvault_modify($conn, $dstpath, back_up_open_files => 0);
sub snapvault_modify( $$;% ) {
    my ($conn, $dstpath, %opts) = @_;
    my ($context, $api, $args, $in_conf);
    my ($cmd, $text, $conf, $val, $options);
	my $opt = '';
	my $zapi = 1;
	my $o    = 0;

    return eval {
        if ( exists($opts{wsize}) ) {
            $zapi = 0 ; # wsize cannot be modified using zapi
        }
	
        if ( exists($opts{back_up_open_files}) ) {
            $opt = " -o back_up_open_files=" .($opts{back_up_open_files} ? 'on':'off');
	    $o   = 1;
        }
		
 	if ( exists($opts{ignore_atime}) ) {
	    if ( $o ){
  	        $opt = $opt . "," . "ignore_atime=" .($opts{ignore_atime} ? 'on':'off');
	    }else{
                $opt = " -o ignore_atime=" . $opts{ignore_atime} ? 'on':'off';
	        $o  = 1;
	    }
        }
		
 	if ( exists($opts{utf8_primary_path}) ) {
	    if ( $o ){
  	        $opt = $opt . "," . "utf8_primary_path=" .($opts{utf8_primary_path} ? 'on':'off');
	    }else{
                $opt = " -o utf8_primary_path=" . $opts{utf8_primary_path} ? 'on':'off';
	        $o  = 1;
	    }
        }

        # Translate input arguments
        $in_conf = _transform_data_struct(\%opts);

        if ( ($context = _ontapi_support($conn, 1) ) && $zapi ) {
            # ZAPI - Modify configuration of relationship
            $api = 'snapvault-secondary-modify-configuration';
            $conf = { 'secondary-path' => $dstpath, %$in_conf };
            $args = [ 'configuration' => [ %$conf ] ];
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);
        } else {
            # CLI - Modify configuration of relationship and convert it
            $cmd = rpl_cmd_from_opts($sv, "modify$opt", $dstpath, %opts);
			
            # Try modifying the configuration
            $text = sayEOF($conn, $cmd);

            $text =~ s/\ANo changes in the configuration\.(\n|\z)//;

            unless ($text =~ s/\AConfiguration for qtree \Q$dstpath\E is:\s*(\n|\z)//) {
                croak $text;
            }

            # Translate configuration structure
            $conf = _transform_data_struct({ parse_conf_entry($text) });

            # Check changes did occur
            while ( ($opt, $val) = each %$in_conf ) {		
                if ( !exists($conf->{$opt}) || $conf->{$opt} ne $val ) {
                    croak "returned configuration did not match " .
                          "specified values";
                }
            }

            unless ($text eq "") {
                croak "Garbage after configuration entry: $text";
            }
        }

        return 1;
    };
}

## @name ::snapvault_release
## @usage snapvault_release($filerConn, $srcpath, $dst)
## @synopsis
##    Releases the source-side resources being held for a particular
##    SnapVault destination.  The destination must be specified as
##    <filer>:<path>, while the source is a simple path.
##    Returns success or failure, with an error message in $@ on failure.
## @note
##    The first argument can be either a connectrec object or a hash reference.
sub snapvault_release($$$) {
    my ($conn_h, $srcpath, $dst) = @_;
    my ($context, $api, $args, $ret, $type, $otype);
    my $conn = _get_conn($conn_h);

    return eval {
        if ( $context = _ontapi_support($conn_h) ) {
            # ZAPI - Release primary-side resources
            ($conn, $type) = _get_conn($conn_h);
            $otype = $type eq 'primary' ? 'secondary' : 'primary';
            $api = "snapvault-$type-release-relationship";
            $args = [ "$type-path"   => $srcpath,
                      %{_host_path($dst, $otype)} ];
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);
            $ret = 1;
        } else {
            # CLI - Release primary-side resources
            $ret = rpl_release_internal($sv, $conn, $srcpath, $dst);
            croak $@ if $@;
        }

        return $ret;
    };
}

## @name ::snapvault_convert
## @usage snapvault_convert($filerConn, $dst)
## @synopsis
##    Convert the named replica destination into a SnapVault destination.
##    Returns success or failure, with an error message in $@ on failure.
## @note
##    This function does not use ZAPI since there is no API for convert.
sub snapvault_convert($$) {
    my ($conn, $dst) = @_;
    rpl_convert_internal($sv, $conn, $dst);
}

## @name ::snapvault_start
## @usage snapvault_start($filerConn, $dst, %options)
## @synopsis
##    This is an internal function, meant only to be called from
##    specific libraries.  This probably isn't what you're looking for.
##    Configures a SnapVault backup relationship, and starts a baseline
##    transfer if necessary.  Returns success or failure, with error
##    message in $@ on failure.
## @param SOURCE Optional: the source <filer>:<path> for the SnapVault relationship
## @param THROTTLE Optional: kbs rate to configure and throttle transfers with
## @param TRIES_COUNT Optional: SnapVault tries count for this relationship
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param WAIT Optional: specifies to wait for the transfer to complete if set
sub snapvault_start($$;%) {
    my ($conn, $dst, %param) = @_;

    return eval {
        my $cmd = rpl_cmd_from_opts($sv, "start", $dst, %param);

        # Configure and try starting the baseline transfer
        my $text = sayEOF($conn, $cmd, TIMEOUT => $param{'TIMEOUT'});
        croak $text unless ($text =~ /$RPL_SUCCESS/);
        return 1;
    };
}

## @name ::sv_start
## @usage sv_start($filerConn, $dst, %options)
## @synopsis
##    Configures a SnapVault backup relationship, and starts a baseline
##    transfer.  If successfully started, returns a subtest object for
##    an active subtest which is waiting for the transfer to complete.
##    On failure, returns undef with error message in $@.
##    If the COMPLETE option is set, then this function returns whether
##    the transfer both started and completed successfully.
##    There is no need to extract results from the returned subtest object,
##    since $@ will contain the error if the transfer failed.
##    The destination should be specified as a <filer>:<path>, but the
##    filer name isn't technically required.
## @param SOURCE Optional: the source <filer>:<path> for the SnapVault relationship
## @param THROTTLE Optional: kbs rate to configure and throttle transfers with
## @param TRIES_COUNT Optional: SnapVault tries count for this relationship
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @param PARALLEL Optional: start the transfer and return, the COMPLETE option
##    is ignored and no subtest object is ever returned. Use only when starting
##    a lot of transfers concurrently.
## @param COMPRESSION Optional: Enable compression for this relationship
## @param CONNECTION_MODE Optional: 'inet' or 'inet6'
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub sv_start($$;%) {
    my ($conn, $dst, %param) = @_;
    my ($context, $api, $args, $in_conf, $conf);
    my ($dstpath, $cmd, $text, %opts, $obj);
    my $parallel = $param{PARALLEL};
    delete $param{PARALLEL};

    return eval {
        # Create relationship configuration options
        map { $opts{lc($_)} = $param{$_} if exists $param{$_} }
            qw(THROTTLE TRIES_COUNT SOURCE COMPRESSION CONNECTION_MODE);
        $opts{'connection-mode'}=$param{connection_mode} if exists($param{connection_mode});
        # Translate input arguments
        $in_conf = _transform_data_struct(\%opts);
        if ( $context = _ontapi_support($conn, 1) ) {
            # ZAPI - Configure and initiate transfer
            ($dstpath) = FSR::FSRutils::split_filer_path($dst);
            if ( $param{RESTART} ) {
                $api = 'snapvault-secondary-resync-relationship';
            } else {
                $api = 'snapvault-secondary-create-relationship';
            }
            $conf = { 'secondary-path' => $dstpath, %$in_conf };
	    $Log->comment("sv_start conf is " . Data::Dumper->Dump([ $in_conf ]));
            $args = [ 'configuration' => [ %$conf ] ];
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);
            return 1 if $parallel;
            $obj = sv_wait_idle({conn=>$conn, source=>0}, $dst, %param);
            croak $@ if $@;
        } else {
            # CLI - Configure and initiate transfer
            $param{'WAIT'} = $parallel ? 0 : 1;

            %opts = ();
            $opts{TIMEOUT} = $param{TIMEOUT} if defined $param{TIMEOUT};

            $cmd = rpl_cmd_from_opts($sv, "start", $dst, %param);

            # Try starting the transfer.
            $text = rpl_start_xfer($conn, $sv, $cmd, %opts);
            croak $@ if ($@);

            # It's possible that 'snapvault start' just needed to configure
            # the destination, and the replica was already initialized.
            # Or, perhaps the destination was both configured and initialized
            # already, so this command has nothing to do.  Since this function
            # is largely concerned with starting the baseline transfer,
            # we choose to treat both cases as an error and fail above.
            # In lieu of trying to detect the error messages above and then
            # fake up a null transfer object below, scripts which test these
            # cases should just expect the relevant error messages.

            if ($param{'RESTART'}) {
                # Whether this string appears is dependent upon whether
                # a resync-rollback was necessary.  If the negotiated
                # common snapshot is the currently base snapshot, and
                # the tree is "in sync", then it is not printed.  However,
                # it doesn't affect whether the following transfer was started,
                # so we just strip it out.
                $text =~ s/^The resync base snapshot will be: (?:.+?)\s*(\n|\z)//m;
            }

            # Whether this string appears is dependent upon whether or not
            # the destination was already configured, and it does not affect
            # whether or not the transfer was successfully started -- so just
            # strip it out.
            $text =~ s/^Snapvault configuration for the qtree has been set\.\s*(\n|\z)//m;

            # Any other text that we got must be an error message.
            croak $text if $text;
            return 1 if $parallel;

            # The transfer successfully started,
            # so now we just need to wait for the completion string.
            $obj = subtest( sub { rpl_finish_xfer($conn, $sv, %opts); },
                               '-bg', '-onexit', sub{} );

            # If specified, wait for the transfer to complete
            # and check the results once it does.
            if ($param{'COMPLETE'}) {
                Subtest::wait_finish(subtests=>[$obj]);
                croak $@ unless subtestResults($obj);
            }

        }
        return $obj;
    };
}

## @name ::snapvault_update
## @usage snapvault_update($filerConn, $dst, %options)
## @synopsis
##    This is an internal function, meant only to be called from
##    specific libraries.  This probably isn't what you're looking for.
##    Starts an incremental backup transfer to the given SnapVault destination.
##    Returns success or failure, with error message in $@ on failure.
## @param SNAPSHOT Optional: specific snapshot to update from
## @param THROTTLE Optional: kbs rate to configure and throttle transfers with
## @param ALL_BLOCKS Optional: whether or not to send all blocks for changed files
## @param CKSUM Optional: whether or not to send checksums for changed files
## @param TIMEOUT Optional: timeout period waiting for the command to complete
sub snapvault_update($$;%) {
    my ($conn, $dst, %param) = @_;

    return eval {
        my $cmd = rpl_cmd_from_opts($sv, "update", $dst, %param);

        # Try starting the incremental transfer
        my $text = sayEOF($conn, $cmd, TIMEOUT => $param{'TIMEOUT'});
        croak $text unless ($text =~ /$RPL_SUCCESS/);
        return 1;
    };
}

## @name ::sv_update
## @usage sv_update($filerConn, $dst, %options)
## @synopsis
##    Tries starting an incremental backup transfer to the given SnapVault
##    destination.  If successfully started, returns a subtest object for
##    an active subtest which is waiting for the transfer to complete.
##    On failure, returns undef with error message in $@.
##    If the COMPLETE option is set, then this function returns whether
##    the transfer both started and completed successfully.
##    There is no need to extract results from the returned subtest object,
##    since $@ will contain the error if the transfer failed.
##    The destination should be specified as a <filer>:<path>, but the
##    filer name isn't technically required.
## @param SNAPSHOT Optional: specific snapshot to update from
## @param THROTTLE Optional: kbs rate to configure and throttle transfers with
## @param ALL_BLOCKS Optional: whether or not to send all blocks for changed files
## @param CKSUM Optional: whether or not to send checksums for changed files
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @param PARALLEL Optional: start the transfer and return, the COMPLETE option
##    is ignored and no subtest object is ever returned. Use only when starting
##    a lot of transfers concurrently.
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub sv_update($$;%) {
    my ($conn, $dst, %param) = @_;
    my ($context, $api, $args, $dstpath, %opts, $cmd, $text, $obj);
    my $parallel = $param{PARALLEL};
    delete $param{PARALLEL};

    return eval {
        # Create update options
        $opts{'max-transfer-rate'} = $param{THROTTLE}
          if exists $param{THROTTLE};
        $opts{'primary-snapshot'} = $param{SNAPSHOT}
          if exists $param{SNAPSHOT};

        if ( $context = _ontapi_support($conn, 1) ) {
            # ZAPI - Initiate an incremental transfer
            ($dstpath) = FSR::FSRutils::split_filer_path($dst);
            $api = 'snapvault-secondary-initiate-incremental-transfer';
            $args = [ 'secondary-path' => $dstpath, %opts ];
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);
            return 1 if $parallel;
            $obj = sv_wait_idle({conn=>$conn, source=>0}, $dst, %param);
            croak $@ if $@;
        } else {
            # CLI - Initiate an incremental transfer
            $param{'WAIT'} = $parallel ? 0 : 1;

            %opts = ();
            $opts{TIMEOUT} = $param{TIMEOUT} if defined $param{TIMEOUT};

            $cmd = rpl_cmd_from_opts($sv, "update", $dst, %param);

            # Try starting the transfer.
            $text = rpl_start_xfer($conn, $sv, $cmd, %opts);
            croak $@ if ($@);
            croak $text if $text;
            return 1 if $parallel;

            # The transfer successfully started,
            # so now we just need to wait for the completion string.
            $obj = subtest( sub { rpl_finish_xfer($conn, $sv, %opts); },
                               '-bg', '-onexit', sub{} );

            # If specified, wait for the transfer to complete
            # and check the results once it does.
            if ($param{'COMPLETE'}) {
                Subtest::wait_finish(subtests=>[$obj]);
                croak $@ unless subtestResults($obj);
            }
        }

        return $obj;
    };
}

## @name ::snapvault_resync
## @usage snapvault_resync($filerConn, $dst, %options)
## @synopsis
##    This is an internal function, meant only to be called from
##    specific libraries.  This probably isn't what you're looking for.
##    Restarts protection to a SnapVault destination by starting a
##    resync-rollback transfer.  Returns success or failure, with error
##    message in $@ on failure.
## @param SOURCE Optional: the source <filer>:<path> for the SnapVault relationship
## @param THROTTLE Optional: kbs rate to configure and throttle transfers with
## @param TRIES_COUNT Optional: SnapVault tries count for this relationship
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param WAIT Optional: specifies to wait for the transfer to complete if set
sub snapvault_resync($$;%) {
    my ($conn, $dst, %param) = @_;

    return eval {
        $param{'RESTART'} = 1; # make this a resync instead of a start
        $param{'FORCE'} = 1; # never answer questions, just do it

        my $cmd = rpl_cmd_from_opts($sv, "start", $dst, %param);

        # Configure and try starting the baseline transfer
        my $text = sayEOF($conn, $cmd, TIMEOUT => $param{'TIMEOUT'});
        croak $text unless ($text =~ /$RPL_SUCCESS/);
        return 1;
    };
}

## @name ::sv_resync
## @usage sv_resync($filerConn, $dst, %options)
## @synopsis
##    Restarts backups the given SnapVault destination, which involves starting
##    a resync-rollback transfer.  If successfully started, returns a subtest
##    object for an active subtest which is waiting for the transfer to
##    complete.  On failure, returns undef with error message in $@.
##    If the COMPLETE option is set, then this function returns whether
##    the transfer both started and completed successfully.
##    There is no need to extract results from the returned subtest object,
##    since $@ will contain the error if the transfer failed.
##    The destination should be specified as a <filer>:<path>, but the
##    filer name isn't technically required.
## @param SOURCE Optional: the source <filer>:<path> for the SnapVault relationship
## @param THROTTLE Optional: kbs rate to configure and throttle transfers with
## @param TRIES_COUNT Optional: SnapVault tries count for this relationship
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub sv_resync($$;%) {
    my ($conn, $dst, %param) = @_;

    $param{'RESTART'} = 1;
    $param{'FORCE'} = 1;
    return sv_start($conn, $dst, %param);
}

## @name ::snapvault_stop
## @usage snapvault_stop($filerConn, $dst, %options)
## @synopsis
##    This is an internal function, meant only to be called from
##    specific libraries.  This probably isn't what you're looking for.
##    Unconfigures the given SnapVault destination, and and then deletes
##    the destination path from the active filesystem.
##    Returns success or failure, with error message in $@ on failure.
## @param TIMEOUT Optional: timeout period waiting for the command to complete
sub snapvault_stop($$;%) {
    my ($conn, $dst, %param) = @_;

    return eval {
        $param{'FORCE'} = 1; # never answer questions, just do it

        my $cmd = rpl_cmd_from_opts($sv, "stop", $dst, %param);

        # Try stop the qtree, and wait until it's done.
        my $text = sayEOF($conn, $cmd, TIMEOUT => $param{'TIMEOUT'});

        my ($path) = FSR::FSRutils::split_filer_path($dst);
        my $pathexp = quotemeta($path);

        if ($text =~ /^qtree is not a snapvault qtree replica: $pathexp\s+Snapvault configuration for the qtree has been deleted\.$/) {
            # First success case --
            # destination doesn't exist, but was successfully unconfigured.
            return 1;
        } elsif ($text =~ /^Protection to $pathexp has been stopped and the qtree deleted\.$/) {
            # Second success case --
            # destination existed and was both unconfigured and deleted.
            return 1;
        } else {
            # Some other error occurred.
            # It could be the case that the destination is neither configured,
            # nor existent, but the container volume exists.  In this case,
            # we would just see the message "qtree is not a snapvault qtree
            # replica" above, with no other output, and no action would be
            # taken.  For script purposes, we're going to treat this as a
            # failure -- it's unlikely that scripts will be running
            # 'snapvault stop' without wanting to actually perform some action.
            $text =~ s/^Could not delete qtree: (.+?)\s*(\n|\z)/$1/;
            croak $text if $text;
            croak "unexpected end of output";
        }
    };
}

## @name ::sv_stop
## @usage sv_stop($filerConn, $dst, %options)
## @synopsis
##    Tries to start an operation which stops a particular backup relationship
##    to the given SnapVault destination.  Stopping a backup relationship
##    entails unconfiguring the destination (if it is configured) and deleting
##    the destination replica (if it exists).  If the operation is successfully
##    started, returns a subtest object for an active subtest which is waiting
##    for the deletion to complete.  On failure, returns undef with error
##    message in $@.  If the COMPLETE option is set, then this function returns
##    whether the destination was successfully unconfigured and deleted.
##    There is no need to extract results from the returned subtest object,
##    since $@ will contain the error if the deletion failed.
##    The destination should be specified as a <filer>:<path>, but the
##    filer name isn't technically required.
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub sv_stop($$;%) {
    my ($conn, $dst, %param) = @_;
    my ($context, $api, $args, %opts);
    my ($cmd, $path, $pathexp, $waitstring, $obj);

    return eval {
        if ( $context = _ontapi_support($conn, 1) ) {
            # ZAPI - Delete relationship
            $api = 'snapvault-secondary-delete-relationship';
            $args = [ 'secondary-path' => $dst ];
            $obj = subtest( sub {
                                eval { SnapMirror::_invoke_ontapi(
                                           context => $context,
                                           api     => $api,
                                           args    => $args
                                       ) };
                                logresult('FATAL', "EVAL_ERROR=$@") if $@;
                            }, '-bg', '-onexit', sub{} );
        } else {
            # CLI - Delete relationship
            $param{'FORCE'} = 1; # never answer questions, just do it

            %opts = ();
            $opts{TIMEOUT} = $param{TIMEOUT} if defined $param{TIMEOUT};

            # Construct the command from the options
            $cmd = rpl_cmd_from_opts($sv, "stop", $dst, %param);

            # Try starting the unconfigure/deletion.
            say($conn, $cmd, '-blind', '-eval_error');
            croak $@ if ($@);

            # Theoretically, the deletion successfully started -- actually,
            # there is no way to tell for sure, since nothing gets printed
            # when it starts.
            # It could have failed immediately, but wouldn't know it yet.
            # Anyway, unless we change ONTAP, the only thing to do now is
            # to construct the completion string and wait for it.

            ($path) = FSR::FSRutils::split_filer_path($dst);
            $pathexp = quotemeta($path);

            # This is a hardcoded regexp which corresponds to the ONTAP code.
            # The parentheses also determine how we determine errors below.
            # Don't change these unless you know exactly what you're doing.
            $waitstring = "(?m)" .
              "(?:^Snapvault configuration for the qtree has been deleted\\.\\r?\\n)?" .
              "(?:" .
                "^Protection to $pathexp has been stopped and the qtree deleted\\.|" .
                "^This command does not delete the data in whole-volume SnapVault replicas\\.\\nTo remove the data, destroy the entire volume\\.|" .
                "^(Could not unconfigure $pathexp\\.)\\nSnapVault may be active in the volume, or the root volume may be full\\.|" .
                "^Could not delete qtree: (.+?)\\s*" .
              ")(?:\\r?\\n|\\z)";

            # We're interpreting the whole-volume "failure" message as success.
            # Stopping a whole-volume destination doesn't delete any data,
            # but for test scripts it's fine if unconfigures successfully
            # and doesn't choke.

            $obj = subtest( sub { rpl_finish_op($conn, $waitstring, %opts); },
                            '-bg', '-onexit', sub{} );
        }

        # If specified, wait for the stop command to complete
        # and check the results once it does.
        if ( $param{'COMPLETE'} ) {
            Subtest::wait_finish(subtests=>[$obj]);
            croak $@ unless subtestResults($obj);
        }

        return $obj;
    };
}

## @name ::snapvault_restore
## @usage snapvault_restore($filerConn, $dst, %options)
## @synopsis
##    This is an internal function, meant only to be called from
##    specific libraries.  This probably isn't what you're looking for.
##    Restores a backup dataset by performing a baseline transfer from the
##    SnapVault secondary backup to a primary.  In this case, the filer
##    connection should be a connection to a SnapVault primary, and the
##    destination should be the primary-side <filer>:<path> to restore to.
##    Returns success or failure, with error message in $@ on failure.
## @param SOURCE Optional: source (secondary) <filer>:<path> to restore from
## @param SNAPSHOT Optional: specific snapshot to update from
## @param THROTTLE Optional: kbs rate to configure and throttle transfers with
## @param TIMEOUT Optional: timeout period waiting for the command to complete
sub snapvault_restore($$;%) {
    my ($conn, $dst, %param) = @_;

    return eval {
        croak "source not specified" unless $param{'SOURCE'};

        # Setup escaped versions of paths for matching.
        my ($path) = FSR::FSRutils::split_filer_path($dst);
        my $pathexp = quotemeta($path);
        my $srcexp;
        if (defined $param{'SOURCE'}) {
            my ($srcpath, $srcfiler) = FSR::FSRutils::split_filer_path($param{'SOURCE'});
            if (defined $srcfiler) {
                my $srcpathexp = quotemeta($srcpath);
                my $srcfilerexp = quotemeta($srcfiler);
                $srcexp = "$srcfilerexp:$srcpathexp";
            }
        }
        unless ($srcexp) {
            $srcexp = ".+?:.+?";
        }

        my $cmd = rpl_cmd_from_opts($sv, "restore", $dst, %param);

        # Try starting the baseline restore
        my $text = sayEOF($conn, $cmd, TIMEOUT => $param{'TIMEOUT'});

        # Catch the easy errors first.
        if ($text =~ /^Restore to $pathexp could not be started: (.+?)\s*(\n|\z)/m) {
            croak $1;
        } elsif ($text =~ /^Restore to $pathexp failed: (.+?)\s*(\n|\z)/m) {
            croak $1;
        } elsif ($text =~ /^Could not make qtree $pathexp writable: (.+?)\s*(\n|\z)/) {
            croak $1;
        }

        # Run through the lines we expect to see on success.
        croak $text
          unless ($text =~ s/^Restore from $srcexp to $pathexp started\.\s*//);
        croak $text
          unless ($text =~ s/^Monitor progress with the 'snapvault status' command\.\s*//);
        croak $text
          unless ($text =~ s/^Abort the restore with \^C\.\s*//);
        # This "might" appear, on older releases.
        $text =~ s/^Data restored to $pathexp\.\s+//;
        croak $text
          unless ($text =~ s/^Made qtree $pathexp writable\.\s+//);
        croak $text
          unless ($text =~ s/^Restore to $pathexp completed successfully\.\s*(\n|\z)//);

        # The string should now be empty unless it contains some error message.
        croak $text if $text;
        return 1;
    };
}

## @name ::sv_restore
## @usage sv_restore($filerConn, $dst, %options)
## @synopsis
##    Starts a baseline restore transfer to the specified destination, which is
##    on a SnapVault primary.  If successfully started, returns a subtest object
##    for an active subtest which is waiting for the transfer to complete.
##    On failure, returns undef with error message in $@.
##    If the COMPLETE option is set, then this function returns whether
##    the transfer both started and completed successfully.
##    There is no need to extract results from the returned subtest object,
##    since $@ will contain the error if the transfer failed.
##    The destination should be specified as a <filer>:<path>, but the
##    filer name isn't technically required.
## @param SOURCE Required: the source <filer>:<path> for the SnapVault relationship
## @param SNAPSHOT Optional: specific snapshot to update from
## @param INCREMENTAL Optional: incremental restore. By default, baseline restore
## @param THROTTLE Optional: kbs rate to configure and throttle transfers with
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub sv_restore($$;%) {
    my ($conn, $dst, %param) = @_;
    my ($context, $api, $args, %opts,$in_conf);
    my ($cmd, $path, $pathexp, $srcexp, $text, $obj);
    my ($srcpath, $srcfiler, $srcpathexp, $srcfilerexp);
    my ($restart_break, $start_wait, $finish_wait);

    return eval {
        # Create restore options
        $opts{'max-transfer-rate'} = $param{THROTTLE}
          if exists $param{THROTTLE};
        $opts{'secondary-snapshot'} = $param{SNAPSHOT}
          if exists $param{SNAPSHOT};
        $opts{'connection-mode'} = $param{connection_mode}
          if exists $param{connection_mode};
        # Translate input arguments
        $in_conf = _transform_data_struct(\%opts);
        if ( $context = _ontapi_support($conn, 1) ) {
            ($path) = FSR::FSRutils::split_filer_path($dst);
	    if ($param{INCREMENTAL}) {
                # ZAPI - Initiate a incremental restore transfer
                $api = 'snapvault-primary-initiate-incremental-restore-transfer';
	    } else {
                # ZAPI - Initiate a restore transfer
            	$api = 'snapvault-primary-initiate-restore-transfer';
	    }
            $args = [ 'primary-path' => $path, %$in_conf,
                      %{_host_path($param{SOURCE}, 'secondary')} ];
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);

            # Poll the primary system
            $obj = sv_wait_idle({conn=>$conn, source=>1}, $dst, %param);
            croak $@ if $@;
        } else {
            # CLI - Initiate a restore transfer
            %opts = ();
            $opts{TIMEOUT} = $param{TIMEOUT} if defined $param{TIMEOUT};

	    # New CLI option for incremental restore starting in Ironcity
            $param{'RESTART'} = 1 if $param{INCREMENTAL};
	    # Allow no user interaction
            $param{'FORCE'} = 1 if $param{INCREMENTAL};
            # Force ONTAP command to wait for completion
            $param{'WAIT'} = 1;

            # Setup escaped versions of paths for matching.
            ($path) = FSR::FSRutils::split_filer_path($dst);
            $pathexp = quotemeta($path);
            if (defined $param{'SOURCE'}) {
              ($srcpath, $srcfiler) = FSR::FSRutils::split_filer_path($param{'SOURCE'});
              if (defined $srcfiler) {
                $srcpathexp = quotemeta($srcpath);
                $srcfilerexp = quotemeta($srcfiler);
                $srcexp = "$srcfilerexp:$srcpathexp";
              }
            }
            unless ($srcexp) {
              $srcexp = ".+?:.+?";
            }

            $cmd = rpl_cmd_from_opts($sv, "restore", $dst, %param);
            if ($vfiler_dst){
                # Starting the restore
                $text = rpl_start_xfer($conn, $sv, $cmd, %opts);
                croak $@ if ($@);
                croak $text if $text;

                # The check successfully started,
                # so now we just need to wait for the completion string.
                $obj = subtest( sub { rpl_finish_xfer($conn, $sv, %opts); },
                                   '-bg', '-onexit', sub{} );
            } else {
            if ( get_wafl_version($conn->hostp) < WV_IRONCITY_MIN ) {
                # These are a hardcoded regexps which corresponds to the ONTAP code.
                # The parentheses also determine how we determine errors below.
                # Don't change these unless you know exactly what you're doing.
                $restart_break = "(?:" .
                  "The restore may be restartable, try again\\.\\s*(?:\\n|\$)|" .
                  "Made qtree $pathexp writable\\.\\s*(?:\\n|\$)|" .
                  "Could not make qtree $pathexp writable: (.+?)\\s*(?:\\n|\$)" .
                  ")";

                $start_wait = "(?m)" .
                  "^Restore from $srcexp to $pathexp started\\.\\s+" .
                  "Monitor progress with the 'snapvault status' command\\.\\s+" .
                  "Abort the restore with \\^C\\.\\s*(?:\\n|\$)|" .
                  "^Restore to $pathexp could not be started: (.+?)\\s+$restart_break";

                $finish_wait = "(?m)" .
                  "^(?:Data restored to $pathexp\\.\\s+)?" .
                  "Made qtree $pathexp writable\\.\\s+" .
                  "Restore to $pathexp completed successfully\\.\\s*(?:\\n|\$)|" .
                  "^(?:Data restored to $pathexp\\.\\s+)?" .
                  "Could not make qtree $pathexp writable: (.+?)\\s*(?:\\n|\$)|" .
                  "^Restore to $pathexp failed: (.+?)\\s+$restart_break";

                # Try starting the transfer.
                $text = rpl_start_op($conn, $cmd, $start_wait, %opts);
                croak $@ if ($@);
                croak $text if $text;

                # The transfer successfully started,
                # so now we just need to wait for the completion string.
                $obj = subtest( sub { rpl_finish_op($conn, $finish_wait, %opts); },
                                   '-bg', '-onexit', sub{} );
            } else {
                # Starting the restore
                $text = rpl_start_xfer($conn, $sv, $cmd, %opts);
                croak $@ if ($@);
                croak $text if $text;

                # The check successfully started,
                # so now we just need to wait for the completion string.
                $obj = subtest( sub { rpl_finish_xfer($conn, $sv, %opts); },
                                   '-bg', '-onexit', sub{} );
            }
            }
            # If specified, wait for the transfer to complete
            # and check the results once it does.
            if ($param{'COMPLETE'}) {
                Subtest::wait_finish(subtests=>[$obj]);
                croak $@ unless subtestResults($obj);
            }
        }
        return $obj;
    };
}

## @name ::snapvault_snap_create
## @usage snapvault_snap_create($filerPri, $volname, $snapname)
## @synopsis
##    Creates a configured snapshot target on a SnapVault primary volume.
##    Again, note that the connection to the filer must be to the
##    SnapVault primary, not the secondary.  This function cannot be
##    used to create SnapVault snapshot targets on a secondary.
##    Returns success or failure, with error message in $@ on failure.
sub snapvault_snap_create($$$) {
    my ($conn, $volname, $snapname) = @_;
    my ($context, $api, $args, $text);

    return eval {
        if ( $context = _ontapi_support($conn, 1) ) {
            # ZAPI - Initiate snapshot creation on primary
            $api = "snapvault-primary-initiate-snapshot-create";
            $args = [ 'volume-name' => $volname,
                      'schedule-name' => $snapname ];
            SnapMirror::_invoke_ontapi(context => $context,
                                       api     => $api,
                                       args    => $args);
        } else {
            $snapname = '""' if ($snapname eq "");
            $text = sayEOF($conn, "$sv snap create $volname $snapname");
            croak $text if ! ($text =~ /Snapshot creation started/);
            croak $text if ! ($text =~ /Monitor progress with 'snapvault status -s'/);
        }

        return 1;
    };
}

## @name ::sv_snap_create
## @usage sv_snap_create($filerSec, $volname, $snapname, %options)
## @synopsis
##    Tries starting the creation of a configured SnapVault snapshot target
##    on the given SnapVault secondary volume.  Again, note that the
##    connection to the filer must be to the SnapVault secondary, not the
##    primary.  This function cannot be used to create SnapVault snapshot
##    targets on a primary.  If the snapshot creation is successfully started,
##    returns a subtest object for an active subtest which is waiting for the
##    snapshot creation to complete.  On failure, returns undef with error
##    message in $@.  If the COMPLETE option is set, then this function
##    returns whether the snapshot target was successfully created.
##    There is no need to extract results from the returned subtest object,
##    since $@ will contain the error if the snapshot creation failed.
## @param DELAY Optional: number of seconds between polls to status (default=15)
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub sv_snap_create($$$;%) {
    my ($conn, $volname, $snapname, %param) = @_;
    my ($context, $api, $args, %opts, $cmd, $text, $obj, $sched);
    my $delay = defined $param{DELAY} ? $param{DELAY} : 15;
    my $timeout = $param{TIMEOUT};

    return eval {
        if ( $context = _ontapi_support($conn, 1) ) {
            # ZAPI - Initiate snapshot creation
            $api = "snapvault-secondary-initiate-snapshot-create";
            $args = [ 'volume-name' => $volname,
                      'schedule-name' => $snapname ];
            my $hostname = $conn->hostp->hostname;
            # Function to start snap create, poll snapshot status for completion.
            # Also reading /etc/messages log to check for failure msgs as there is
            # no way to determine failures with snapshot status cmd
            $cmd = sub {
                _log_start($hostname);
                SnapMirror::_invoke_ontapi(context => $context,
                                           api     => $api,
                                           args    => $args);
                # Set expiration time
                my $etime = time + $timeout
                  if (defined $timeout && $timeout > 0);
                # Give time for creating the snapshot when no transfer
                # is required
                sleep 2;
                eval {
                    while (1) {
                        # Get status of snapshot
                        $sched = snapvault_status_snapshot(
                                          { conn => $conn, source => 0},
                                          VOLUME => $volname,
                                          SNAPSHOT => $snapname,
                                          STATUS_ONLY => 1
                        );
                        last if (defined $sched && $sched->{status} eq 'Idle');
                        if ( defined $etime && time > $etime ) {
                            croak 'TIMEOUT';
                        }
                        sleep $delay;
                    }
                };
                logresult('FATAL', "EVAL_ERROR=$@") if $@;
                _log_error($snapname);
                logresult('PASS');
            };
            # Start subtest to poll the snapshot status
            $obj = subtest( $cmd, '-bg', '-onexit', sub{} );
            sleep(2); # Pause a little to allow snap create to start before returning
        } else {
            # CLI - Initiate snapshot creation
            $snapname = '""' if ($snapname eq "");

            %opts = ();
            $opts{TIMEOUT} = $param{TIMEOUT} if defined $param{TIMEOUT};

            $cmd = "$sv snap create -w $volname $snapname";

            # Try starting the transfer.
            $text = rpl_start_snap($conn, $sv, $cmd, %opts);
            croak $@ if ($@);
            croak $text if $text;

            # The transfer successfully started,
            # so now we just need to wait for the completion string.
            $obj = subtest( sub { rpl_finish_snap($conn, $sv, %opts); },
                               '-bg', '-onexit', sub{} );
        }

        # If specified, wait for the transfer to complete
        # and check the results once it does.
        if ( defined $obj && $param{'COMPLETE'} ) {
            Subtest::wait_finish(subtests=>[$obj]);
            croak $@ unless subtestResults($obj);
        }

        return $obj;
    };
}

## @name ::snapvault_check
## @usage snapvault_check($filerConn, $dst, %options)
## @synopsis
##    Starts a snapvault check session for the specified secondary qtree.
##    This function can only be run on the secondary filer.
## @param SOURCE Required: the source <filer>:<path> for the SnapVault relationship
## @param THROTTLE Optional: kbs rate to configure and throttle transfers with
## @param VERIFIER_FIX Optional: run in fix mode.
## @param VERIFIER_QUICK Optional: run in quick mode.
## @param VERIFIER_CKSUM Optional: data comparison level.
## @param VERIFIER_SNAPSHOT Optional: specific snapshot on the secondary
##        filer to be used.
## @param VERIFIER_IGNORE Optional: ignore some type of comparisons (ctime).
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub snapvault_check($$;%) {
    my ($conn, $dst, %param) = @_;
    my ($cmd, $text, $obj, %opts);

    return eval {
        $param{'WAIT'} = 1;  # force ONTAP command to wait for completion
        $param{'FORCE'} = 1; # never answer questions, just do it
        $cmd = rpl_cmd_from_opts("$sv check", 'start', $dst, %param);

        # Set options for rpl commands
        $opts{TIMEOUT} = $param{TIMEOUT} if exists $param{TIMEOUT};

        # Try starting the check
        $text = rpl_start_xfer($conn, $sv, $cmd, %opts);
        croak $@ if ($@);
        croak $text if $text;

        # The check successfully started,
        # so now we just need to wait for the completion string.
        $obj = subtest( sub { rpl_finish_xfer($conn, $sv, %opts); },
                           '-bg', '-onexit', sub{} );

        # If specified, wait for the transfer to complete
        # and check the results once it does.
        if ($param{'COMPLETE'}) {
            $obj->wait_finish();
            subtestResults($obj) or croak $@;
        }

        return $obj;
    };
}

## @name ::snapvault_check_abort
## @usage snapvault_check_abort($filerConn, $dst)
## @synopsis
##    Aborts a currently running snapvault check session for secondary qtree.
##    It does not abort currently running SnapVault transfers. This function
##    can be called from both primary or secondary filers.
##    Returns success or failure, with error message in $@ on failure.
sub snapvault_check_abort($$;) {
    my ($conn, $dst) = @_;
    my $text;

    return eval {
        $text = sayEOF($conn, "priv set -q diag; $sv check abort $dst");
        croak $text if $text;

        return 1;
    };
}

## @name ::snapvault_compare
## @usage snapvault_compare($conn_dst, $conn_src, $dst, $src, %options)
## @synopsis
##    Do treecompare on the pair of snapshots on primary and secondary
##    after restore (either baseline or incremental) operation.
##    This function is meant to be called from SnapVault secondary only.
##    $conn_dst connection to destination filer
##    $conn_src connection to source filer
##    $dst destination path
##    $src source path
## @param TIMEOUT Optional: timeout period waiting for the destination to be idle
##
sub snapvault_compare($$$$;%) {
    my ($conn_dst, $conn_src, $dst, $src, %param) = @_;
    my ($result, $src_basesnap, $dst_basesnap);
    my ($srchost, $srcpath, $dsthost, $dstpath, $srcvol, $srcqt, 
        $dstvol, $dstqt, $src_snappath, $dst_snappath, $conn);
        
    return eval {
	# Determine the base snapshot created on source filer 
        # (restore target)
        $result = snapvault_status_detailed($conn_src, $src, $dst);
        $src_basesnap = $result->{'base-snapshot'};

	# Determine the base snapshot used on destination filer 
	# (restore origin)
        $result = snapvault_dst_snap($conn_dst, $dst, $src);
        $dst_basesnap = $result;

	# Split the host name and path
        ($srcpath, $srchost) = FSR::FSRutils::split_filer_path($src);
        ($dstpath, $dsthost) = FSR::FSRutils::split_filer_path($dst);

	# Split the volume name and qtree name
        ($srcvol, $srcqt) = qtreeComponents($srcpath) or croak $@;
        ($dstvol, $dstqt) = qtreeComponents($dstpath) or croak $@;

        # create snappath to be of the form
        # /vol/<volname>/.snapshot/<snapshot>/<qtree>
        $src_snappath = "/vol/$srcvol/\.snapshot/$src_basesnap/$srcqt";
        $dst_snappath = "/vol/$dstvol/\.snapshot/$dst_basesnap/$dstqt";

	# Get the hostspec connection 
        $conn = _get_conn($conn_src);

	# Issue treecompare on source filer
        treecompare($conn,
                    $src_snappath,
                    $dst_snappath,
                    args=>'-c',
                    timeout=>$param{timeout},
                    filer_dst=>$dsthost) or croak $@;
	return 1;
    };
}


## @name ::snapvault_wait
## @usage snapvault_wait($filerConn, $dst, $delay, $timeout)
## @synopsis
##    Internal function which polls SnapVault, waiting for the given
##    destination and all snapshot targets in its volume to become idle.
##    Logs NATE results to report success/failure.
## @return
##    On failure, logs a FATAL result key EVAL_ERROR.
sub snapvault_wait($$$$) {
    my ($conn_h, $dst, $delay, $timeout) = @_;
    my ($conn) = _get_conn($conn_h);

    # Record when we started so that we can timeout properly.
    my $start_time = time;

    # Determine the SnapVault volume we're in.
    my ($path) = FSR::FSRutils::split_filer_path($dst);
    my $volname = FSR::FSRutils::split_vol_path($path);
    unless ($volname) {
        chomp $@;
        logresult('FATAL', "EVAL_ERROR=$@");
    }

    while (1) {
        # Check whether the destination is idle.
        my $statref = sv_detailed_status($conn_h, $dst);
        if ( exists $statref->{'current-transfer-error'} ) {
            logresult('FATAL', 'EVAL_ERROR=' .
                               $statref->{'current-transfer-error'});
        }
        if ( $statref->{'status'} eq 'idle' ) {
            # Check whether all snapshot targets are idle
            my $notidle = grep { $_->{status} ne 'Idle' }
              snapvault_status_snapshot($conn_h, VOLUME=>$volname,
                                                 STATUS_ONLY=>1);
            last if ($notidle == 0); # everything is idle
        }

        if ( defined $timeout && $timeout >= 0 ) {
            my $now = time;
            logresult('FATAL', "EVAL_ERROR=TIMEOUT")
              if (($now - $start_time) > $timeout);
        }
        sleep $delay;
    }
    logresult('PASS');
}

## @name ::sv_wait_idle
## @usage sv_wait_idle($filerConn, $dst, %options)
## @synopsis
##    SnapVault manages sets of read-only backups (replicas).  In doing so,
##    it often starts various operations in response to other operations
##    completing or failing.  This can cause problems for scripts which
##    expect the particular destinations they're using to be idle.
##    This function exists to alleviate that problem.  It returns a subtest
##    object for an active subtest which is waiting for the destination to
##    to be idle (by polling status).  On failure, returns undef with error
##    message in $@.  If the COMPLETE option is set, then this function
##    returns once the destination is idle.  There is no need to extract
##    results from the returned subtest object, since $@ will contain the
##    error if the something failed (e.g. if the timeout was exceeded).
##    The destination should be specified as a <filer>:<path>.
## @param DELAY Optional: number of seconds between polls to status (default=15)
## @param TIMEOUT Optional: timeout period waiting for the destination to be idle
## @param COMPLETE Optional: waits for destination to be idle before returning if set
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the subtest succeeded.
sub sv_wait_idle($$;%) {
    my ($conn, $dst, %param) = @_;

    return eval {
        $param{'DELAY'} = 15
          unless ((defined $param{'DELAY'}) && ($param{'DELAY'} =~ /^\d+$/));
        $param{'TIMEOUT'} = 60
          unless (exists $param{'TIMEOUT'});

        # Kick off a subtest to poll SnapVault until it's idle.
        my $obj = subtest( sub { snapvault_wait($conn, $dst, $param{'DELAY'},
                                                $param{'TIMEOUT'}); },
                           '-bg', '-onexit', sub{} );

        # If specified, wait for the transfer to complete
        # and check the results once it does.
        if ($param{'COMPLETE'}) {
            Subtest::wait_finish(subtests=>[$obj]);
            croak $@ unless subtestResults($obj);
        }

        return $obj;
    };
}

## @name ::schedule_info
## @usage schedule_info($schedule_string)
## @synopsis
##    Convert schedule string into a schedule-info.
sub schedule_info {
    my $sched = shift;
    my ($day, $hour, %ret);
    my @list = split /@/, $sched;

    # Split values
    $ret{'retention-count'} = shift @list;
    ($day) = grep { !/\d/ } @list;
    ($hour) = grep { /\d/ } @list;

    # Set defaults
    $ret{'schedule'}{'days-of-week'} = defined $day ? $day : 'mon-sun';
    $ret{'schedule'}{'hours-of-day'} = defined $hour ? $hour : 0;

    return \%ret;
}

## @name ::schedule_string
## @usage schedule_string($schedule_info)
## @synopsis
##    Convert schedule-info into a schedule string
sub schedule_string {
    my $info = shift;
    my $ret = $info->{'retention-count'};

    # Check for defaults
    if ( $info->{'schedule'}{'days-of-week'} ne 'mon-sun' ) {
        $ret .= '@' . $info->{'schedule'}{'days-of-week'};
    }
    if ( $info->{'schedule'}{'hours-of-day'} ne '0' ) {
        $ret .= '@' . $info->{'schedule'}{'hours-of-day'};
    }

    return $ret;
}

## @name ::use_zapi_rsh
## @usage use_zapi_rsh()
## @synopsis
##    Run SnapVault commands on ZAPI and if it is not available fall back
##    back to use rsh.
sub use_zapi_rsh() {
    $filer_interface = 2;
    return 1;
}

## @name ::use_zapi_only
## @usage use_zapi_only()
## @synopsis
##    Run SnapVault commands on ZAPI only.
sub use_zapi_only() {
    $filer_interface = 1;
    return 1;
}

## @name ::use_rsh_only
## @usage use_rsh_only()
## @synopsis
##    Run SnapVault commands on rsh only.
sub use_rsh_only() {
    $filer_interface = 0;
    return 1;
}

## @name ::get_filer_interface
## @usage get_filer_interface()
## @synopsis
##    Return the current chosen interface for SnapVault commands
##    (either ZAPI or rsh).
sub get_filer_interface() {
    return $filer_interface;
}

## @name ::set_filer_interface
## @usage set_filer_interface($interface_type)
## @synopsis
##    Set the interface for SnapVault commands
##    (either ZAPI or rsh).
sub set_filer_interface {
    $filer_interface = shift;
    return 1;
}

## @name ::set_vfiler_args
## @usage set_vfiler_args($vfiler_dst,%vfiler_args)
## @synopsis
##    Set the $vfiler_dst flag and also set the vfiler arguments.
sub set_vfiler_args {
    my ($is_vfiler_dst,$vf_args) = (shift,shift);

    if (($is_vfiler_dst) && ((!defined $vf_args->{_user_name}) || (!defined $vf_args->{_password})) ){
        croak 'Set the vfiler flag or arguments properly';
    }

    $vfiler_dst = $is_vfiler_dst;
    $vfiler_args{user_name} = $vf_args->{_user_name};
    $vfiler_args{password} = $vf_args->{_password};

    return 1;
}

# Internal function
# All functions in this package that are supported on both the primary
# and secondary systems need to differentiate between the two types since
# ZAPI has a different API according to type:
#   snapvault-primary-abort-transfer
#   snapvault-secondary-abort-transfer
# Adding a named parameter to specify the type is not possible for functions
# which have positional optional parameters since the named parameter must
# be specified only when all optional parameters are specified.
#
# This function checks the SnapVault license on the filer connection to
# derive the type (primary or secondary). The argument could be either a
# filer connection object or a hash reference having the connection object
# and a flag (true if connection is for the primary system).
# If the argument is a hash reference, the license check is skipped.
#
# If this function is called in an array context the license check is
# done if necessary and returns both the connection object and the flag.
# Otherwise, if it is called in a scalar context the license check is
# skipped and the connection object is returned.
#
# This function is used in a scalar context when the command runs on CLI
# to avoid the license check. On ZAPI, this it is called in an array context
# to force the check to called the appropriate API.
#
# If both the primary and secondary licenses are enabled, it croaks with
# an error.
sub _get_conn(\$;) {
    my $conn_h = shift;
    my ($conn, $src, $out, $pri);

    if ( ref($$conn_h) eq 'HASH' ) {
        $conn = $$conn_h->{conn};
        $src  = $$conn_h->{source};
    } else {
        $conn = $$conn_h;
        if ( wantarray ) {
            $out = registryGet($conn, 'options.license.sv_ontap_sec');
            if ( $out eq 'on' ) {
                $pri = registryGet($conn, 'options.license.sv_ontap_pri');
                # Error if both licenses are enabled
                croak "Both primary and secondary snapvault licenses are " .
                      "enabled: use alternate first argument as a hash " .
                      "having the connectrec and the source." if $pri eq 'on';
            }
            $src = $out eq 'off' ? 1 : 0;

            # Replace current connection object by a hash reference
            $$conn_h = { conn => $conn, source => $src };
        }
    }

    # The flag is the type of system
    $src = $src ? 'primary' : 'secondary';

    return wantarray ? ($conn, $src) : $conn;
}

# Internal function
# Convert output of snap schedule to old format to be backwards compatible
sub _snap_sched_out($;) {
    my $list = shift;

    my $type     = $list->[0]{'is-auto-update'} eq 'true' ? 'xfer' : 'create';
    my $volname  = $list->[0]{'volume-name'};
    my $snapname = $list->[0]{'schedule-name'};
    my $sched    = schedule_string($list->[0]);

    return ($type, $volname, $snapname, $sched);
}

# Internal function
# Convert a time into seconds
sub _convert_time2secs {
    my $timestr = shift;
    if ( $timestr =~ /(-)?(\d\d):(\d\d):(\d\d)/ ) {
        return ($1 ? $1 : '') . (3600*$2 + 60*$3 + $4);
    }
    return;
}

# Internal function
# Convert a hostpath into a hash having a system and path keys.
sub _host_path {
    my ($hostpath, $key) = @_;
    my ($path, $host) = FSR::FSRutils::split_filer_path($hostpath);
    return { "$key-system" => $host, "$key-path" => $path };
}

# Internal function
sub _transform_data_struct {
    my $data = shift;
    my (%ret, $conv, $key, $value, $nkey, $out);

    return unless defined $data;

    foreach $key ( keys %$data ) {
        $value = $data->{$key};
        next if $value eq '';
        # Convert the current key
        $nkey = lc($key);
        $nkey =~ s/\s+/-/g;
        if ( exists($zapi_h{$key}) ) {
            $conv = $zapi_h{$key};
            if ( exists($conv->{key}) ) {
                # Request for a change of key
                $nkey = $conv->{key};
                if ( !exists($conv->{value}) ) {
                    # Convert value to lower case
                    $ret{$nkey} = lc($value);
                }
            }
            if ( exists($conv->{value}) ) {
                $out = &{$conv->{value}}($value, $nkey);
                if ( ref($out) eq 'HASH' ) {
                    %ret = (%ret, %$out);
                } elsif ( defined($out) ) {
                    $ret{$nkey} = $out;
                }
            }
            if ( !exists($conv->{key}) && !exists($conv->{value}) ) {
                croak "Error parsing translation entry for item '$key'";
            }
        } else {
            # Convert value to lower case
            $ret{$nkey} = lc($value);
        }
    }

    return \%ret;
}

# Internal function
# Function to determine if filer has support for snapvault ZAPIs
# The support will be there if ONTAP is ver R7.2.1x and above
# Args   : Connectrec object
#        : (optional) Do not check snapvault licenses
# Returns: Context if filer has support for snapvault ZAPIs.
#          Otherwise, 0.
sub _ontapi_support(\$;$) {
    my ($conn_h, $no_check) = @_;
    my ($context, $passwd, $err);
    my $conn = &_get_conn($conn_h);
    my $timeout = $conn->timeout >= 0 ? $conn->timeout : 0x7fffffff;

    # No ZAPI support if disabled for package
    return 0 if $filer_interface < 1;

    if ( $filer_interface == 2 && ! $no_check ) {
        ($conn, undef) = eval { &_get_conn($conn_h) };
        # Default to rsh if both snapvault licenses are enabled
        return 0 if $@ =~ /Both primary and secondary snapvault licenses/;
    }

    if ($vfiler_dst) {
        $context = 0;

        eval {
            $context =
              ati_ontapi_create_context(
                                      filer     => $conn->hostp->default_ip(),
                                      transport => ATI_ONTAPI_TRANSPORT_DEFAULT,
                                      user      => $vfiler_args{user_name},
                                      password  => $vfiler_args{password},
                                      timeout   => $timeout,
                                      debug     => ATI_ONTAPI_DEBUG_NONE);
        };
        $err = $@;
    } else {
        if (get_wafl_version($conn->hostp) >= WV_GORDONBIERSCH_MIN) {

            # create a new context whenever we call any of these functions
            # in this package.
            $context = 0;
            $passwd  = $conn->hostp->default_password();

            eval {
                $context =
                  ati_ontapi_create_context(
                                      filer     => $conn->hostp->default_ip(),
                                      transport => ATI_ONTAPI_TRANSPORT_DEFAULT,
                                      user      => 'root',
                                      password  => $passwd,
                                      timeout   => $timeout,
                                      debug     => ATI_ONTAPI_DEBUG_NONE);
            };
            $err = $@;
        } else {
            $err = "Version of filer does not support ZAPI";
        }
    } ## end else [ if ($vfiler_dst)

    if ( ! $err ) {
        return $context;
    } elsif ( $filer_interface == 1 ) {
        # Only using ZAPI and if not available this is an error
        croak "Unable to create ATI ONTAPI context: $err";
    } else {
        logresult('WARN', "Unable to create ATI ONTAPI context, ".
                          "falling back to rsh: $err");
    }

    return 0
}

sub _log_start {
    my ( $hostname) = @_;

    # Start to message log reader on hostname
    SnapMirror::_start_msg_log_rdr($hostname);
}

sub _log_error {
    my ($snapname) = @_;

    # Wait for upto 15 secs for error messages to be logged into /etc/messages
    eval {
        # considered failure with any messages involving snapshot name
        my $str = SnapMirror::_log_search(15, $snapname);
       croak $str if($str !~ /A backup Snapshot copy was created/);
    };
    logresult('FATAL', "EVAL_ERROR=$@")
        if $@ && $@ !~ /^Timed out/;
}

1;

