#!/usr/bin/perl

#***********************************************************************
# This is the comment section for tharndoc
#***********************************************************************
## @summary Perl module used by most STEF test scripts     
## @author System Test
## @pod here

=head1 NAME

SpinTestTool.pm

=head1 AUTHOR

System Test   

=head1 DESCRIPTION

These test tools can be used to do some miscellaneous things from
within a Perl test script. Generally, this Perl module should be
included into every ST (Spinnaker Test) test that we write.

=head1 SYNOPSIS

This module is instantiated by all STEF test scripts, and is not directly called.
See the descriptions below if you need the syntax for using a parameter

=head1 EXAMPLES

No examples needed

=cut

#***********************************************************************
# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDESPINTESTTOOL}) {
    my $newModule = $ENV{OVERRIDESPINTESTTOOL};
    delete $ENV{OVERRIDESPINTESTTOOL};
    print "The main SpinTestTool module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package SpinTestTool;
use Server;
use Print;
use strict;
use English;
use Getopt::Long;
use SystestCommon;
use Data::Dumper;
use Socket;
use Cwd qw(getcwd realpath);
use POSIX "_exit";
no strict 'refs';

my $print = new Print;

my @protocols = ("udp", "tcp");
my @NFSversions = (2, 3);

my $prompt = '\[.+@.+\][$#] ';
my $CleanupSTT = undef;
my $CleanupCode = undef;
my $stt;

use Options;
my $Options;


#***********************************************************************

=head1 C<new()>

=over 4

=item Description

This is the constructor

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {Verbosity => 1, # optional
	        logVerbosity => 3, # optional
		testObjective => undef, # optional
                OutputHandler => undef, # optional
		@_, # get any parameters that were explicitly passed
	        };
    bless($self, $class); 

    # when running inside of nate, $0 is set to 'full script options' so
    #  we need to first split this by spaces and split the first element
    #  (which is the actual script) by the '/' character and take the
    #  last element which will leave us with the real script name
    $self->{FullScript} = (split(/\s+/, $PROGRAM_NAME))[0];
    my @program = split("/", $self->{FullScript});
    $self->{Program} = pop @program;

    # Initialize some of the "instance variables" for this object.
    $self->{MountStackSize} = 0;
    $self->{DirStackSize} = 0;
    # the following key is for the RunST method. The idea is that some
    # new methods can be added to interface and modify these definitions
    $self->{PhaseDefinitions} = {SetupServers => \&_SetupServers,
				 SetupClients => \&_SetupClients,
				 Test => \&_RunTest,
				 CleanupClients => \&_CleanupClients,
				 CleanupServers  => \&_CleanupServers,
			     };

    # check to see if we are running inside of NATE
    if (exists($main::{'Tharn::'})) {
	$self->{NATE} = 1;
	# If this file is running inside of NATE and logging to a file
	#  then set the default verbosity level to 3
	$self->{Verbosity} = 3 if ($ENV{LOGDIR} ne '');
    }

    $stt = $self;
    return $self;
} # end new

#***********************************************************************

=head1 C<DESTROY()>

=over 4

=item Description

This is the Destructor

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************


sub DESTROY {
    my $self = shift;
    # no parameters
    $self->Print(3, "SpinTestTool::DESTROY entered\n");
    $self->Print(3, "SpinTestTool::DESTROY exited\n");
} # end DESTROY

#***********************************************************************

=head1 C<GetOptions()>

=over 4

=item Description

This method sets the default values for all parameters and reads in any
values passed in

=item Standard Test Arguments

=over 4

=item -client 

Usage: -client <host[,host,..]> [-client <host[,host,..]>] | -clientfile <filename> 

Description: -client defines which clients are used for the test.
-clientfile specifies a file with the list of clients listed on each line

Default: NULL

=item -server 

Usage: -server<host[,host,..]> [-server <host[,host,..]>] 

Description: defines which servers are used for the test

Default: NULL

=item -quick

Usage: [-quick] 

Description: specifies that the test should run its shortest form

Default: 0 (not specified)

=item -verbosity

Usage: [-verbosity <0..n>] 

Description: specifies how much output should be printed 

Default: 1

=item -stverbosity

Usage: [-stverbosity <0..n>] 

Description: specifies the verbosity level for the ST file itself. 

Default: Same as -verbosity

=item -logverbosity

Usage: [-logverbosity <0..n>] 

Description: specifies the verbosity level for the log messages. 

Default: 3

=item -color|-nocolor

Usage: [-color|-nocolor]

Description: specifies colored output or not.  Defaults to on.

=item -clientuser 

Usage: [-clientuser <user>] 

Description: specifies the login name for the clients 

Default: root

=item -clientpassword

Usage: [-clientpassword <password>]

Description: specifies the login password for the clientuser 

Default: spinwin

=item -serveruser

Usage: [-serveruser <user>] 

Description: specifies the login name for the servers 

Default: root

=item -serverpassword

Usage: [-serverpassword <password>]

Description: specifies the login password for the serveruser

Default: admin

=item -cliuser 

Usage: [-cliuser <user>]

Description: specifies the login name for the admin user 

Default: admin

=item -clipassword 

Usage: [-clipassword <password>]

Description: specifies the login password for the cliuser 

Default: admin

=item -source

Usage: [-source <sourcedir>]

Description: specifies where the source code can be found

Default: 

=item -share

Usage: UNKNOWN

Description: UNKNOWN

Default: /

=item -target

Usage: [-target <dir | host:/share/dir | \\\\host\\share\\dir>]

Description: specifies where the source code should be copied before executing

Default: /tmp

=item -seed

Usage:[-seed <randomseed>]

Description: specifies a seed to use by the random number generator

Default: undef

=item -notify

Usage: [-notify <email>]

Description: specifies an email account to notify with the results

Default: undef

=item -help

Usage: [-help]

Description: prints this message

Default: 0, meaning don't execute

=item -podhelp

Usage: [-podhelp]

Description: prints the onscreen help using POD tags

Default: 0, meaning don't execute

=item -podusage

Usage: [-podusage]

Description: displays the usage string using POD tags

Default: 0, meaning don't execute

=item -podmethod

Usage: [-podmethod <method>]

Description: displays documentation on specified method.

Default: 0, meaning don't execute

=item -getrequirements

Usage: [-getrequirements]

Description: will print out the requirements to run this test

Default: 0, meaning don't execute

=item -initFile

Usage: [-initFile <filename>]

Description: specifies a file to read the server configuration information from

Default: undef

=item -duration

Usage: [-duration <time>]

Description: specifies how long the test should try to run

Default: undef

=item -iterations

Usage: [-iterations <count>]

Description: specifies the number of times the test should run

Default: undef

=item -testOptions
(Optional)

Usage: [-testOptions <comma-separated string of options> [-testOptions ...]]

Description: is a way to pass special options to an underlying program

=item -testcase
(Optional)

Usage: [-testcase <name of test case>]

Description: specifies a specific testcase to run

=item -listtestcases
(Optional)

Usage: [-listtestcases]

Description: will print out the possible testcases for the test

=item -testResults

Usage: [-testResults <0|1>]

Description: this option is obsolete

=item -tester

Usage: [-tester <username>]

Description: specifies the name to store in the database as the tester 

Default: [${ $self->{options}}{tester}]

=item -nBladeIP
(Optional)

Usage: [-nbladeIP <IP address[:range]> | -nbladeIP <IP address[,IP address ...]]

Description: will specify a specific server IP address to use for the N blade(can optionally specify a range with :digit)

=item -dbladeIP
(Optional)

Usage: [-dbladeIP <IP address[:range]> | -dbladeIP <IP address[,IP address ...]]

Description: will specify a specific server IP address to use for the D blade(can optionally specify a range with :digit)

=item -serverIP
(Optional)

Usage: [-serverIP <IP address[:range]> | -serverIP <IP address[,IP address ...]] 

Description: is an old flag that will go away

=item -skiperrors
(Optional)

Usage: [-skiperrors]

Description: will skip any errors in Iterate() and continue

=item -stoponerror
(Optional)

Usage: [-stoponerror]

Description: will cause the test to stop when an iteration encounters an error

=item -verify

Usage: [-verify]

Description: will try and verify the data for VFS ops

Default: undef

=item -dataType

Usage: [-dataType <Short|Long>]

Description: specifies how much data to use when calling ModifyPath

Default: Short

=item -release 

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -saveoutput

Usage: [ -saveoutput [/path/logfile]]

Description: Instruct Print() to save output to a log file as well as print the
	     output to the screen.  Optionally takes a log file name to use,
	     if not given one will create one in the format "/tmp/testname.$$"

Default: undef

=item -VfsServer

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -VirtualServer

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -StoragePool

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -BuildVersion

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -StopOnErrors

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -FindAllServers 

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -timeout

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -TestTimeout

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -benchmark

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -Reconnect

Usage: UNKNOWN

Description: can be '-1' for forever, 0 for never or # for wait # seconds

Default: 0 

=item -CLIAsRoot

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=item -maxClientLoad

Usage: UNKNOWN

Description: UNKNOWN

Default: undef

=back

=item CIFS Test Arguments

=over 4

=item -cifsdomain

Usage: [-cifsdomain <CIFS domain>]

Description: specifies the CIFS domain to use

=item -cifsshare

Usage: [-cifsshare <CIFS share>]

Description: specifies the share on the CIFS server to use

=item -cifsserver

Usage: [-cifsserver <CIFS server>]

Description: specifies the Netbios name of the CIFS server to use

=item -cifsclient 

Usage: UNKNOWN

Description: UNKNOWN

Default: NULL

=item -cifsuser 

Usage: [-cifsuser <username>]

Description: specify a specific CIFS user to use

Default: undef

=item -cifsgroup 

Usage: [-cifsgroup <group name>]

Description: specify a specific CIFS group to use

Default: undef

=item -cifspassword

Usage: [-cifspassword <password>]

Description: specify a specific password for the CIFS user

Default: undef

=back

=item NFS Test Arguments

=over 4

=item -nfsversion

Usage: [-nfsversion <2|3>]

Description: specifies which version of NFS to mount

Default: undef

=item -protocol 

Usage: [-protocol <tcp|udp>]

Description: specifies which transport protocol to use when mounting

Default: undef

=item -nfsclient 

Usage: UNKNOWN

Description: UNKNOWN

Default: NULL

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************

sub GetOptions {
    my $self = shift;
    my $badparams;
    if (ref($_[0]) ne "HASH") {
	$self->Print(0, "SpinTestTool::GetOptions ERROR, options hash not passed in as first argument\n");
	exit 1;
    }
    my $optionsRef = shift;
    my %params = (
		  Help => undef,
		  GetRequirements => undef,
		  AdditionalOpts => [],
		  Required => [],
		  STRequired => [],
		  TestCases => [],
		  Cleanup => undef,
		  PostProcess => undef,
		  ExitAfterList => 1,
		  DuplicationOK => undef,
		  @_, # get any parameters that were explicitly passed
		  );
    my $optionsEntry;
    my $rc;
    my $myclient;
    my $myserver;
    my %Requirements;
    my $key;
    my $member;

    # Define the options (parameters) defaults.
    my $defoptions = {
	clientuser => "root",
	clientpassword => "spinwin",
	target => "/tmp",
	share => "/",
	TestTimeout => 300, # default to 5 minutes for no real reason,
	dataType => "Short",
	%$optionsRef,
    };

    my %fss_definitions = (
			   NBLADE_IP => "nbladeIP",
			   NBLADE_DATA_IP => "nbladeIP",
			   DATA_IP1 => "nbladeIP",
			   LINUX_CLIENT => "client",
			   CLIENT => "client",
			   );

    my @option_definitions = (
			      'quick',
			      'testdir=s',
			      'share=s',
			      'stverbosity=i',
			      'client=s@',
			      'nfsclient=s@',
			      'cifsclient=s@',
			      'clientfile=s',
			      'clientuser=s',
			      'clientpassword=s',
			      'cifsuser=s',
			      'cifsgroup=s',
			      'cifspassword=s',
			      'source=s',
			      'target=s',
			      'seed=i',
			      'notify=s',
			      'podhelp',
			      'podusage',
			      'podmethod=s',
			      'getrequirements',
			      'initFile=s',
			      'duration=s',
			      'iterations=i',
			      'testOptions=s@',
			      'serverIP=s@',
			      'testResults=i',
			      'tester=s',
			      'nfsversion=i',
			      'release=s',
			      'protocol=s',
			      'saveoutput:s',
			      'phase=s@',
			      'VfsServer=s',
			      'VirtualServer=s',
			      'StoragePool=s@',
			      'cifsserver=s',
			      'cifsdomain=s',
			      'cifsshare=s',
			      'BuildVersion=s',
			      'StopOnErrors',
			      'verify=i',
			      'timeout=i',
			      'TestTimeout=i',
			      'nbladeIP=s@',
			      'dbladeIP=s@',
			      'dataType=s',
 			      'benchmark',
			      'vfs=s@',
			      'vfsfile=s',
			      'Reconnect=i',
			      'CLIAsRoot=i',
			      'maxClientLoad=i',
			      'testFile=s',
			      'STmode=i',
			      @{$params{AdditionalOpts}},
			      );
    
    # Convert the old-style ARRAY of test cases into the new style HASH
    #  If the reference is for an array, loop through each entry (which is
    #  a hash reference) and look for a TCName key in it. If the name exists
    #  then the name is the key in a new hash and the value is the hash reference
    #  itself. Also remove the TCName from this reference. When finished, set the
    #  original reference to this new hash
    if (ref($params{TestCases}) eq "ARRAY") { 
        my %tcases;
        foreach my $tchash (@{$params{TestCases}}){
            if ($tchash->{TCName}) {
                $tcases{$tchash->{TCName}} = $tchash;
                delete $tchash->{TCName};
            } elsif ($tchash->{Objective}) {
                if (!$self->{testObjective}) {
                    $self->{testObjective} = $tchash->{Objective};
                }
                delete $tchash->{Objective};
            } else {
                my $text = Dumper($tchash);
                $self->Print(1,"SpinTestTool::GetOptions encountered an old-style testcase without a testcase name:\n$text\n", "bad");
                exit;
            }
        }
        $params{TestCases} = \%tcases;
    } elsif (ref($params{TestCases}) ne "HASH") { 
        my $text = Dumper($params{TestCases});
        $self->Print(1,"SpinTestTool::GetOptions encountered a TestCases entry it did not understand:\n$text\n", "bad");
        exit;
    }

    my @definitions;
    my @require_one_of;
    foreach my $option (@option_definitions) {
	my %definition;
	my ($name,$type) = split(/=|:|!|\+/, $option);
	$definition{name} = $name;
	$definition{spec} = $type if ($type);
	$definition{description} = "unknown description for $name";
	$definition{default} = $$defoptions{$name} if ($$defoptions{$name});
	foreach my $type qw(required st_required) {
	    foreach my $option (grep(/^$name/, @{$params{$type}})) {
		$definition{$type} = "The '$option' option is required";
		if ( $option =~ /->|\||>=|</ ) { # catch the special options
		    if ( $option =~ /->/ ) { # if x then y is required
			my ($opt, $dep) = split(/->/, $option);
			$definition{dependencies} = $dep;
		    } elsif ( $option =~ /\|/ ) { # x or y (or z) is required
			push(@require_one_of, [split(/\|/, $option)]);
			delete $definition{$type};
		    } elsif ( $option =~ />=/ ) {
			my ($opt, $val) = split(/>=/, $option);
			$definition{greater_equal} = $val;
		    } elsif ( $key =~ /</ ) {
			my ($opt, $val) = split(/</, $option);
			$definition{less} = $val;
		    }
		} elsif ($option =~ /=/) {
		    my ($option, $operators) = split(/=/, $option);
		    $definition{file_tests} = $operators;
		}
	    }
	}	
	push(@definitions, \%definition);
    }
    my $Options = new Options;
    $Options->define(options=>$optionsRef);
    $Options->parameters(definitions=>\@definitions, require_one_of=>\@require_one_of, 
			 duplication_ok=>$params{DuplicationOK});
    $Options->compatability(definitions=>\%fss_definitions);
    $Options->test_cases(definitions=>$params{TestCases}, exit_after_list=>$params{ExitAfterList});
    $Options->process();
    $optionsRef = $Options->get_hash();
    $self->{options} = $optionsRef;

    # A .thpl files is a combination of an ST and a PL. 
    # If this is a THPL and clients were passed in then we are running in ST mode
    $self->{STmode} = 1 if ($self->{Program} =~ /\.thpl$/ && scalar @{$optionsRef->{client}});
    $self->{STmode} = $$optionsRef{STmode} if (defined($$optionsRef{STmode}));

    if ($self->{STmode} && $$optionsRef{testcase}) {
	my $test_case = $$optionsRef{testcase};
	my $tc = $params{TestCases}{$test_case};
	eval "use SharedTestBed::TestResults::Burt qw(testcase_start);";
	die $@ if ($@);
	eval {
	    Burt::testcase_start($test_case, $$tc{Description});
	  };
	die $@ if ($@);
    }
    # Set the verbosity of this SpinTestTool object based on the
    # verbosity that was just read from the parsed options. This 
    # SpinTestTool verbosity is used for the Print() method. We also
    # allow a different verbosity level for the ST (Spinnaker Test) file
    # than for a script that the ST file invokes.

    if ($self->{Program} =~ /\.st$/ || $self->{STmode}) {
	if (defined($$optionsRef{stverbosity})) {
	    $self->_SetVerbosity(Verbosity => $$optionsRef{stverbosity});
	} elsif (defined($$optionsRef{verbosity})) {
	    $self->_SetVerbosity(Verbosity => $$optionsRef{verbosity});
	}
    } elsif (defined($$optionsRef{verbosity})) {
	$self->_SetVerbosity(Verbosity => $$optionsRef{verbosity});
    }

    # if we are doing benchmarking then we should load the module
    if ( $$optionsRef{benchmark} ) {
	$self->Print(1, "SpinTestTool::GetOptions: Enabling Benchmark module.\n");
	eval("use Benchmark;");
	if ($@) { # eval sets this if an error is encountered
            print $@, "\n\n\n";
	    exit(1);
        }
    }
    
    $$optionsRef{testdir} =~ s!/+$!! if ($$optionsRef{testdir}); # get rid of any trailing /'s

    if (defined($$optionsRef{protocol}) && !grep(/$$optionsRef{protocol}/i, @protocols)) {
	print "The protocol $$optionsRef{protocol} is not one of: ", join(" ", @protocols), "\n";
	exit(1);
    }

    if (defined($$optionsRef{nfsversion}) && !grep(/$$optionsRef{nfsversion}/i, @NFSversions)) {
	print "The NFS version $$optionsRef{nfsversion} is not one of: ", join(" ", @NFSversions), "\n";
	exit(1);
    }

    # In this routine, we set $badparams to non-zero if we find any problems
    # rather than exiting immediately. This is so that we can continue
    # to verify all the options and report all problems in a single pass.
    # At the end of the routine, if any problems have occurred, we'll
    # do a hard exit, otherwise, we'll just do a soft return to our 
    # invoker.

    if ( $optionsRef->{vfsfile} ) {
	$stt->Print(2, "Using vfsfile = $optionsRef->{vfsfile}\n");
	open(IF, $optionsRef->{vfsfile}) || die "Cannot open $optionsRef->{vfsfile}: $!\n";
	while ( <IF> ) {
	    chomp; # remove the \n from each input line
	    push(@{$optionsRef->{vfs}}, "$_");
	}
	close(IF);
    }

    # we need to pop all the nfsclients and cifsclients onto the clients list.  the assumtion would be that
    # you shouldn't care if you have a cifs or nfs client if you use the client list, but if you do care you will
    # use the nfsclient and cifsclient array instead
    my @clients;
    my $useClient = 1;
    if (@{$optionsRef->{nfsclient}} != ()) {
	$useClient = undef;
	if (@{$optionsRef->{client}} != ()) {
	    $self->Print(1, "SpinTestTool::GetOptions is overriding '-client' with '-nfsclient'\n", "warn");
	}
	push(@clients, @{$$optionsRef{nfsclient}});
    }
    if ($useClient) {
	push(@clients, @{$$optionsRef{client}});
    }
    if ($$optionsRef{clientfile}){
	if (! -r "$$optionsRef{clientfile}") {
	    $self->Print(0, "SpinTestTool::GetOptions ERROR, cannot open $$optionsRef{clientfile}, skipping it\n");
	} else {	    
	    if (open(IF, $$optionsRef{clientfile})){
		while (<IF>){
		    next if /^\s*#/;
		    chomp;
		    push(@clients, $_);
		}
		close(IF);
	    }
	}
    }    
    $$optionsRef{client} = \@clients;

    # If -help, call the caller's Help subroutine.
    if ( $$optionsRef{help} == 1 ) {
  	if ( ref($params{Help}) eq "CODE" ) {
  	    $params{Help}->(); # invoke the passed routine
  	    exit 0;
	} elsif ( ref($params{Help}) eq "HASH" ) {
	    $self->Help($params{Help}->{Options}, $params{Help}->{Descriptions}, $params{Help}->{SpecificHelp});
	    exit(0);
	} else {
	    $self->Print(1, "Help is not supported.\n");
	    exit(0);
	}
    }

    # If -podhelp, call the internal HelpFromPod subroutine.
    if ( $$optionsRef{podhelp} == 1 ) {
	$self->HelpFromPod(Filename=>$self->{FullScript});
        exit(0);
    }

    # If -podusage, call the internal UsageFromPod subroutine.
    if ( $$optionsRef{podusage} == 1 ) {
        $self->UsageFromPod(Filename=>$self->{FullScript});
        exit(0);
    }

    # If -podmethod, call the internal MethodFromPod subroutine.
    if ( $$optionsRef{podmethod}) {
        $self->MethodFromPod(Filename=>$self->{FullScript}, Method=>$$optionsRef{podmethod});
        exit(0);
    }

    # If -getrequirements, call the caller's GetRequirements subroutine.
    if ( $$optionsRef{getrequirements} == 1 ) {
	if ( ref($params{GetRequirements}) eq "CODE" ) {
	    $params{GetRequirements}->($optionsRef);
	    exit 0;
	} else {
	    $self->Print(0, "SpinTestTool::GetOptions ERROR, GetRequirements subroutine not defined\n");
	    $badparams = 1;
	}
    }

    $CleanupSTT = $self;
    $CleanupCode = $params{Cleanup};

    # setup a subroutine to handle interrupt signals
    $SIG{INT} = \&InterruptHandler; # use the passed routine

    $self->HandleInitFile();

    if ($$optionsRef{serverIP}) {
	$$optionsRef{nbladeIP} = [] if (!$$optionsRef{nbladeIP});
	push(@{$$optionsRef{nbladeIP}},@{$$optionsRef{serverIP}});
    }
    my @nbladeIPs;
    if ($$optionsRef{nbladeIP}) {
	foreach my $ip (@{$$optionsRef{nbladeIP}}) {
	    if ($ip =~ /local/) {
		$self->{localNbladeIP} = 1;	
	    } elsif ($ip =~ /remote/) {
		$self->{remoteNbladeIP} = 1;
	    } elsif ($ip =~ /random/) { # this is the default so do nothing
		# this now means to use all of the vifs for the virtual server
		;
	    } elsif ($ip =~ /(.+):(.+)/) {
		my ($ipaddress, $range) = ($1,$2);
		$ipaddress = Socket::inet_ntoa(scalar(gethostbyname($ipaddress))); 
		my ($a,$b,$c,$d) = split(/\./, $ipaddress);
		if (($d + $range - 1) > 255) {
		    $self->Print(1, "The IP address range ($range) extends beyond $a.$b.$c.255!\n", "bad");
		    exit(1);
		}
		foreach my $address ($d..($d+$range-1)) {
		    push(@nbladeIPs, "$a.$b.$c.$address");
		}
	    } else {
		push(@nbladeIPs, $ip);
	    }
	    }
	$$optionsRef{nbladeIP} = \@nbladeIPs;
    }
    my @dbladeIPs;
    if ($$optionsRef{dbladeIP}) {
	foreach my $ip (@{$$optionsRef{dbladeIP}}) {
	    if ($ip =~ /(.+):(.+)/) {
		my ($ipaddress, $range) = ($1,$2);
		$ipaddress = Socket::inet_ntoa(scalar(gethostbyname($ipaddress))); 
		my ($a,$b,$c,$d) = split(/\./, $ipaddress);
		if (($d + $range - 1) > 255) {
		    $self->Print(1, "The IP address range ($range) extends beyond $a.$b.$c.255!\n", "bad");
		    exit(1);
		}
		foreach my $address ($d..($d+$range-1)) {
		    push(@dbladeIPs, "$a.$b.$c.$address");
		}
	    } else {
		push(@dbladeIPs, $ip);
	    }
	}
	$$optionsRef{dbladeIP} = \@dbladeIPs;
    }
    if ($$optionsRef{release} && ($$optionsRef{release} !~ /\d+\.\d+\.\d+/
				  && $$optionsRef{release} !~ /^R/)) {
	$self->Print(1, "Please specify the release as x.x.x: $$optionsRef{release}\n", "bad");
	exit(1);
    }
    
    if (scalar(@{$optionsRef->{StoragePool}})) {
	# SPNames is used to indicate that the options were passed in
	$self->{SPNames} = join(",", @{$optionsRef->{StoragePool}});
    }

    if ( $badparams != 0 ) {
	exit 1;
    }

    $self->{notify} = $$optionsRef{notify};

    if (defined($params{PostProcess})) {
	if (ref($params{PostProcess}) ne "CODE") {
	    $self->Print(0, "SpinTestTool::GetOptions ERROR, PostProcess is not a CODE reference\n", "bad");
	    exit 1;
	}
	
	exit 1 if (!&{$params{PostProcess}}());
    }

    # Check to see if the file is a ST (Spinnaker Test) file. If so,
    # then this file is required to have at least one client and at
    # least one server specified in its options. If this is not a
    # .st file, then there are no specific requirements imposed by
    # this method.
    if ( $self->{Program} =~ /\.st$/ ) {
	if ( @clients == () ) {
	    $self->Print(0, "SpinTestTool::GetOptions ERROR, ST files require at least one client\n");
	    exit 1;
	}
	if ( @{$optionsRef->{server}} == () ) {
	    $self->Print(0, "SpinTestTool::GetOptions ERROR, ST files require at least one server\n");
	    exit 1;
	}
    } elsif ($self->{STmode}) {
        # If this file is called with client parameters then it is being
        # run as an ST. Don't do this if the file is an actual ST file.
#	$$optionsRef{testFile} = $ENV{TESTNAME};
#	$$optionsRef{testFile} =~ s!.+/!!;
#	$$optionsRef{source} = $ENV{TESTNAME};
#	$$optionsRef{source} =~ s!/$$optionsRef{testFile}!!;
#	if (! -r "$$optionsRef{source}/$$optionsRef{testFile}") {
#	    die "SpinTestTool::GetOptions Error: testFile '$$optionsRef{source}/$$optionsRef{testFile}' does not exist\n";
#	}
	$self->Print(2, "SpinTestTool::GetOptions will now call RunST()\n", "info");
	$self->RunST(%params, TestCommand=>$ENV{TESTNAME});
	# this does not return
    }
} # end GetOptions

sub systype {
    my $self = shift;
    my $systype;
    my $ostype = `uname -s`;
    chomp($ostype);
    if ($ostype =~ /sunos/i) {
	$systype = "solaris";
	my $arch = `uname -p`;
	chomp $arch;
	$systype .= "$arch";
    } elsif ($ostype =~ /hp-ux/i) {
	$systype = "hpux";
    } elsif ($ostype =~ /irix64/i) {
	$systype = "irix";
    } elsif ($ostype =~ /aix/i) {
	$systype = "aix";
    } elsif ($ostype =~ /linux/i) {
	$systype = "linux";
    } else {
	$self->Print(message=>"SpinTestTool::systype Unknown result from `uname -s`: $ostype\n", status=>"bad", die=>1);
    }
    return $systype;
}

sub log_error {
    my $self = shift;
    $self->{ResultError} = shift;
}

sub cwd {
    my $self = shift;
    my $path = $ENV{TESTNAME};
    if (!$path) {
	return undef;
    }
    my @path = split(m!/!, $path);
    pop @path;
    $path = join("/", @path);
}

#***********************************************************************

=head1 C<IterateAndExit()>

=over 4

=item Description

Run the given routine a # of iterations or for a period
of time specified by duration. Iterations or Durations
has to be specified. The method will report results,
log everything and exit with the appropriate code

=item Arguments

=over 4

=item Options
(Required)

A reference to the options hash. The Iterations,
Duration, and Sleep will be pulled directly from
here.

=item Subroutine
(Required)

a reference to the subroutine to run

=item Exit
(Optional)

should the program exit? (default:true)

=item PostIterate
(Optional)

a reference to a subroutine to run prior to
exiting the program

=back

=item Returns

0 when both iteration and any PostIterate subroutine pass

1 otherwise

=back

=cut

#***********************************************************************
sub IterateAndExit {
    my $self = shift;
    my %params = (
		  Options => undef, # required
		  Subroutine => undef, # required
		  Exit => 1, # optional
		  PostIterate => undef, #optional
		  Message => undef,
		  @_, # overlay any parameters that were explicitly passed
		  );

    my $rc = undef;
    $self->Print(3, "SpinTestTool::IterateAndExit entered\n");

    if ( ref($params{Subroutine}) ne "CODE" ) {
	$self->Print(0, "SpinTestTool::IterateAndExit ERROR, Subroutine is not a code reference\n");
	goto done;
    }
    if ( $params{PostIterate} && ref($params{PostIterate}) ne "CODE" ) {
	$self->Print(0, "SpinTestTool::IterateAndExit ERROR, PostIterate is not a code reference\n");
	goto done;
    }
    if ( ref($params{Options}) ne "HASH" ) {
	$self->Print(0, "SpinTestTool::IterateAndExit ERROR, Options is not a hash reference\n");
	goto done;
    }

    $rc = $self->Iterate(
			 Duration => $params{Options}{duration},
			 Iterations => $params{Options}{iterations},
			 Subroutine => $params{Subroutine},
			 Sleep => $params{Options}{sleep},
			 );
    
    if (! $rc ) {
	$stt->Print(0, "SpinTestTool::IterateAndExit ERROR, Iterate() failed\n", "bad");
    }
    $rc &= $params{PostIterate}->($rc) if ($params{PostIterate});
    if ($self->{ResultError}) {
	$params{Message} = $self->{ResultError};
	$self->{ResultError} = undef;
    }
    if (! $rc ) {
	$stt->HandleResults(ReturnCode => 1, Exit => $params{Exit}, Message => $params{Message});
    } else {
	$stt->HandleResults(ReturnCode => 0, Exit => $params{Exit}, Message => $params{Message});
    }
    
  done:
    $self->Print(3, "SpinTestTool::IterateAndExit exited\n");
    return($rc);
} # IterateAndExit

#***********************************************************************

=head1 C<Iterate()>

=over 4

=item Description

Run the given routine a # of iterations or for a period
of time specified by duration. Iterations or Durations
has to be specified.

=item Arguments

=over 4

=item Subroutine
(Required)

a reference to the subroutine to run

=item Iterations
(Depracated)

the number of iterations to run.

=item Duration
(Depracated)

the period of time to run. (default is in minutes)

Syntax:  #s or #S for seconds

         #m or #M for minutes

         #h or #H for hours


=item Sleep
(Depracated)

how long to sleep in between iterations.
Units are seconds of time

=back

=item Returns

true on success

undef on failure

=back

=cut

#***********************************************************************
sub Iterate {
    my $self = shift;
    my %params = (
		  Iterations => undef, # optional
		  Duration => undef, # optional
		  Subroutine => undef, # required
		  Sleep => 0, # optional
		  @_, # overlay any parameters that were explicitly passed
		  );
    my $testStartTime;
    my $testDuration;
    my $iterStartTime;
    my $iterDuration;
    my $durationUnit;
    my $durationTarget;
    my $iterations;
    my $i;
    my $rc = 1;

    $self->Print(3, "SpinTestTool::Iterate entered\n");

    if ( ref($params{Subroutine}) ne "CODE" ) {
	$self->Print(0, "SpinTestTool::Iterate ERROR, $params{Subroutine} subroutine not defined\n");
	return undef;
    }

    if ( ($params{Iterations} eq undef) && ($params{Duration} eq undef) ) {
	$iterations = 1; # default to doing one iteration
    } elsif ( $params{Iterations} eq undef ) {
	$iterations = -1;
    } else {
	$iterations = $params{Iterations};
    }

    if ( $params{Duration} eq undef ) {
	$durationTarget = -1;
    } else {
	# Normalize the duration string into numerical seconds
	if ( $params{Duration} =~ /[Hh]/ ) { # if value is hours
	    $durationUnit = chop($params{Duration});
	    $durationTarget = $params{Duration} * 3600;
	} elsif ( $params{Duration} =~ /[Mm]/ ) { # if value is minutes
	    $durationUnit = chop($params{Duration});
	    $durationTarget = $params{Duration} * 60;
	} elsif ( $params{Duration} =~ /[Ss]/ ) { # if value is seconds
	    $durationUnit = chop($params{Duration});
	    $durationTarget = $params{Duration};
	} else {
	    $durationUnit = "m"; # default unit of duration is minutes
	    $durationTarget = $params{Duration} * 60; # convert to seconds
	}
    }

    if ( $iterations >= 0 ) {
	$self->Print(1, "Starting a $iterations-iteration test.\n");
    }
    if ( $durationTarget >= 0 ) {
	if ( $durationUnit =~ /[Hh]/ ) {
	    $self->Print(1, "Starting a " . $durationTarget/3600 . "-hour test.\n");
	} elsif ( $durationUnit =~ /[Mm]/ ) {
	    $self->Print(1, "Starting a " . $durationTarget/60 . "-minute test.\n");
	} else {
	    $self->Print(1, "Starting a " . $durationTarget . "-second test.\n");
	} 
    }

    $testStartTime = time;
    $testDuration = 0;

    # This loop is controlled as follows: If the -duration parameter is used
    # and -iterations is not, then the -duration value is used, and vice
    # versa. If both are passed in, then the test will end when the "shorter"
    # of the two is hit. If neither is passed in, the test will run until it
    # dies or is killed.
    $i = 1;
    while ( ($durationTarget >= 0 ? $testDuration < $durationTarget : 1) &&
	    ($iterations >= 0 ? $i <= $iterations : 1) &&
	    ($rc == 1 || !$ {$self->{options}}{StopOnErrors}) ) {
	sleep $params{Sleep} unless ($i == 1);
	if ( $iterations >= 0 ) {
	    $self->Print(1, "Begin iteration $i (of $iterations).\n");
	}
	if ( $durationTarget >= 0 ) {
	    $self->Print(1, "Begin iteration $i ");
	    if ( $self->{Verbosity} >= 1 ) {
		if ( $durationUnit =~ /[Hh]/ ) {
		    printf "(%.2f of " . $durationTarget/3600 . " hours)\n", $testDuration/3600;
		} elsif ( $durationUnit =~ /[Mm]/ ) {
		    printf "(%.2f of " . $durationTarget/60 . " minutes)\n", $testDuration/60;
		} else {
		    printf "(%.2f of " . $durationTarget . " seconds)\n", $testDuration;
		}
	    }
	}
	$iterStartTime = time;
	
	#***************************************************************
	# Invoke the subroutine that was passed in.
	#***************************************************************
	$rc &= $params{Subroutine}->(); # routine failure must return undef
	$self->Print(1, "End iteration $i (took ");
	if ( $self->{Verbosity} >= 1 ) {
	    $iterDuration = time - $iterStartTime;
	    if ( $durationUnit =~ /[Hh]/ ) {
	        printf "%.2f hours)\n", $iterDuration/3600;
	    } elsif ( $durationUnit =~ /[Mm]/ ) {
	        printf "%.2f minutes)\n", $iterDuration/60;
	    } else {
	        printf "%.2f seconds)\n", $iterDuration;
	    }
        }
	$i++;
	$testDuration = time - $testStartTime;
    } # end while

    $self->Print(1, "Test took ");
    if ( $self->{Verbosity} >= 1 ) {
        $testDuration = time - $testStartTime;
        if ( $durationUnit =~ /[Hh]/ ) {
	    printf "%.2f hours\n", $testDuration/3600;
        } elsif ( $durationUnit =~ /[Mm]/ ) {
	    printf "%.2f minutes\n", $testDuration/60;
        } else {
	    printf "%.2f seconds\n", $testDuration;
        } 
    }

    $self->Print(3, "SpinTestTool::Iterate exited\n");
    return($rc);
} # end Iterate

#***********************************************************************

=head1 C<Notify()>

=over 4

=item Description    

Notify will send a message to a user using a number
of methods. The most primary method is via Email.  Notify will
also notify a user via AOL Instant messenger.  Support for Yahoo
Messenger is also planned as well. Depending on what options are 
passed in, Notify will handle things a bit differently.
    
If only an email address is given the message will be delivered
directly to the email address.

If only an IM address is given the message will be delivered
directly to the IM account.

If both an IM and one email address is given, the IM account is
checked to see if the IM user is online, and hasn't been idle
for more than a set amount of time (defaults to 30 minutes). If
the user is online and hasn't been idle longer than that time the
message is delievered to the IM account.  If not, then the message
is delivered to the email address.

If more than one email address and one or more IM screennames are
passed in the method will notify all users becuase we aren't tieing
screen names and email addresses at all and can't pop an email address
off the stack if we get a message to the IM user.

=item Arguments 

=over 4

=item Message 

message to be sent

=item File

file to send as message

=item To

email address to send email to

=item From

email from address (defaults to logged in user name)

=item Subject

email subject to set

=item HTML

should the email be sent as HTML mail

=item SmtpServer

SMTP mail server to use to send the mail

=item ToAIM

AOL IM Screen Name to send message to

=item MyAIM

AIM Screen Name to use to log in to AIM

=item MyAIMPassword

AIM Password for the screen name

=item ToYMsg

Yahoo Messenger Screen Name

=item MyYMsg

Yahoo Messenger Screen Name to use to log in to Yahoo Mesenger

=item MyYMsgPassworD

Yahoo Messenger Password for the screen name

=item IMIdle

idle time to cause cancel to IM messenger

=back

=item Returns

None

=back

=cut

#********************************************************

sub Notify {
    my $self = shift;
    my %params = (To => undef,
		  From => $ENV{LOGNAME},
		  Subject => undef,
		  Message => undef,
		  File => undef,
		  HTML => undef,
		  SmtpServer => "sepia.hq.netapp.com",
		  ToAIM   => undef,
		  MyAIM   => "spinsystest",
		  MyAIMPassword => "spin_systest",
		  ToYMsg  => undef,
		  MyYMsg  => "spinsystest",
		  MyYMsgPassword    => "fooaccount",
		  IMIdle    => 30,
		  @_, # get any parameters that were explicitly passed
		  );
    $self->Print(3, "SpinTestTool::Notify entered\n");
    # we need perl 5.8 to use AIM or Yahoo.
    if ($] ne "5.008" && ($params{ToAIM} || $params{ToYMsg})){
	$self->Print(1, "SpinTestTool::Notify Must use Perl 5.8 to use AIM or Yahoo Notification.\n","warn");
	delete $params{ToAIM};
	delete $params{ToYMsg};
    };

    # load the modules that we'll need based on the options passed in
    if ($params{To}) {
	$self->Print(3, "SpinTestTool::Notify Loading SMTP Module.\n");
	eval "use Net::SMTP";
	if ($@) { # eval sets this if an error is encountered
	    print $@, "\n\n\n";
	}
    }
    if ($params{ToAIM}) {
	$self->Print(3, "SpinTestTool::Notify Loading AOL Instant Messenger Module.\n");
	eval "use Net::OSCAR";
	if ($@) { # eval sets this if an error is encountered
	    print $@, "\n\n\n";
	}
    }
    if ($params{ToYMsg}) {
	$self->Print(3, "SpinTestTool::Notify Loading Yahoo Messenger Module.\n");
	warn "Unable to deal with Yahoo messenger yet.\n";
	if ($@) { # eval sets this if an error is encountered
	    print $@, "\n\n\n";
	}
    }

    # if there was a AIM Screen Name passed in, log in to AIM, check to see if the user is online,
    # check to be sure they haven't been idle for too long and send the message.
    if ($params{ToAIM}) {
	# create a new AIM session if there currently isn't one
	if (!$self->{aim}) {
	    $self->{aim} = Net::OSCAR->new;
	    $self->{aim}->set_callback_error(\&_aim_error);
	    $self->{aim}->set_callback_admin_error(\&_aim_admin_error);
	    $self->{aim}->set_callback_admin_ok(\&_aim_admin_ok);
	    $self->{aim}->loglevel(1);
	    $self->{aim}->signon(screenname=>"$params{MyAIM}", password=>"$params{MyAIMPassword}");
	}
	my $aim = $self->{aim};
	$self->Print(3, "SpinTestTool::Notify connecting to AIM Server.\n");
	my $sign_on_timeout = 5000; # 5 seconds
	until ( $aim->is_on() || ($sign_on_timeout == 0)){ # keep looping through the socket until we get logged on
	    $aim->do_one_loop();
	    $sign_on_timeout--;
	}
	print "Took ", 5000-$sign_on_timeout, " milliseconds to get logged in.\n";
	# check to see if the user is logged on
	my $lookup_timeout = 5000; # 5 seconds
	my $buddy = $aim->buddy($self->{ToAIM});
	until ( $lookup_timeout == 0 ) {
	    $aim->do_one_loop();
	    $lookup_timeout--;
	    print "Checking to see if \"$params{ToAIM}\" is online\n";
	    foreach my $key ( keys %{%$buddy}){
		print "$key -> $buddy->{$key}\n";
	    }
	    if ($buddy->{online}) {
		print "$params{ToAIM} is online.\n";
		print "IDLE Time: $buddy->{idle}\n";
		$lookup_timeout = 0;
	    }
#	    my $info = $aim->get_info("$params{ToAIM}");
#	    if ($info->{online}) {
#		print "$params{ToAIM} is online.\n";
#		print "IDLE Time: $info->{idle}\n";
#	    }
	}

	sleep 90; # debug sleep
	
#	foreach my $key (keys %{%$self->{aim}}) {
#	    print "$key -> $self->{aim}{$key}\n";
#	}
    }
    # if there was a Yahoo Messenger Screen Name passed in, log in to Yahoo Messenger, check to see
    # if the user is online, check to be sure they haven't been idle for too long, and send the message

    
    # if we are here, and the $params{To} still has any people on it, send mail to them
    if ($params{To}) {
	my $to;
	my $mailserver = $params{SmtpServer};
	my $smtp;
	my @toList;
	@toList = split(/,/, $params{To});
	$smtp = Net::SMTP->new($mailserver, Port => 25) || die "Couldn't open SMTP connection to $mailserver";
	if ( $params{From} eq "" ) {
	    # do nothing. The user intentionally wants the From field to be blank.
	} else {
	    if ( !($params{From} =~ /@/) ) {
		$params{From} .= "\@netapp.com";
	    }
	}
	$smtp->mail($params{From}) || die "Couldn't set From: to $params{From}";
	foreach $to (@toList){
	    $smtp->to($to) || die "Couldn't add $to as recipient";
	}
	$smtp->data() || die "Couldn't start message";
	$smtp->datasend("To: $params{To}\n") || die "Couldn't add To: line to message";
	$smtp->datasend("Subject: $params{Subject}\n") || die "Couldn't add Subject: line to message";
	# this will tell the mail client to interpret the mail as html formatted mail instead of text
	# as soon as we send a blank line the header is complete and we start the body
	if ($params{HTML}) { 
	    $smtp->datasend("Content-type: text/html\nMIME-Version: 1.0\n");
	    $smtp->datasend("\n") || die "Couldn't insert blank line in message";
	    $smtp->datasend("<html><body>\n");
	} else {
	    $smtp->datasend("\n") || die "Couldn't insert blank line in message";
	}
	if ( $params{File} eq "" ) {
	    $smtp->datasend("$params{Message}\n") || die "Couldn't append body to message";
	} else {
	    open(IF, $params{File});
	    while (my $line = <IF>){
		if ( $params{HTML} ) { # strip any ANSI codes and change \n to <BR>\n
		    $line = $self->ANSItoHTML($line);
		}
		$smtp->datasend($line) || die "Couldn't append body to message";
	    }
	    close(IF);
	}
	if ($params{HTML}) { 
	    $smtp->datasend("</body></html>\n");
	}
	$smtp->dataend() || die "Couldn't end the message";
	$smtp->quit;
    }
    $self->Print(3, "SpinTestTool::Notify exited\n");
    return 1;
} # end Notify

#***********************************************************************

=head1 C<Print()>

=over 4

=item Description

This subroutine shifts off the parameters differently from our
other subroutines. Rather than force the caller of Print to name
the parameters on every invocation (and there will be many), we
allow them to just pass the two parameters to save typing and to
make their source code cleaner.

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

sub Print {
    my ($self, $verbosityMin, $message, $status) = @_;

    $print->Print(min=>$verbosityMin, message=>$message, status=>$status);
}

#***********************************************************************

=head1 C<RunST()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut   

#***********************************************************************
sub RunST { # Version 2
    my $self = shift;

    my $options = $self->{options};
    my $rc = 1;
    my %params = (
		  SkipPhases => [],
		  Phases => [qw(SetupServers SetupClients Populate Test CleanupClients CleanupServers)],
		  TestCommand => "$options->{source}/$options->{testFile}",
		  TestOptions => "",
		  TestTimeout => $options->{TestTimeout}, 
		  TestDir => $options->{TestDir},
		  MirrorTest => undef,
		  ModifyTestOptions => undef,
		  Configuration => undef,
		  SetupConfiguration => 1,
		  @_, # get any parameters that were explicitly passed
		  );
    # this hash will be used to define the configuration to create during SetupServers phase (if defined)
    my $Configuration;
    if (ref($params{Configuration}) eq "CODE") {
	$Configuration = &{$params{Configuration}};
	if (ref($Configuration) ne "HASH") {
	    $self->Print(1, "SpinTestTool::RunST error with Configuration: the code did not return a hash reference\n", "bad");
	    goto fail;
	}
    } elsif (ref($params{Configuration}) eq "HASH") {
	$Configuration = $params{Configuration};
    } elsif (defined($params{Configuration})) {
	$self->Print(1, "SpinTestTool::RunST error with Configuration: '$params{Configuration}' is not a hash or code reference\n", "bad");
	goto fail;
    }

    my %PhaseDefinitions = %{$self->{PhaseDefinitions}};
    my @DiscoveryNodes = qw(cluster vif storagepools virtualserver);
    push(@DiscoveryNodes, "ports") if ($$Configuration{VIF});
    push(@DiscoveryNodes, ("CIFSServer", "CIFSShare", "CIFSDomain", "kerberosrealm")) if ($self->{testObjective} eq "CIFS");
    
    $self->Print(3, "SpinTestTool::RunST entered\n");
    $self->{RunST} = 1;
    
    srand($options->{seed});

    # we need to make sure that ModifyTestOptions is a CODE reference since we will call it if defined
    if ($params{ModifyTestOptions} && ref($params{ModifyTestOptions}) ne "CODE") {
	$self->Print(1, "SpinTestTool::RunST error with ModifyTestOptions: '$params{ModifyTestOptions}' is not a code reference\n", "bad");
	goto fail;
    }
    
    # It doesn't make much sense to have an ST defined with no Test phase. The Test phase is also special because the
    #  ModifyTestOptions code will be called right before the Test phase. 
    if (!$PhaseDefinitions{Test}) {
	$self->Print(1, "SpinTestTool::RunST requires that a Test phase be defined\n", "bad");
	goto fail;
    }

    # this array will define the list of valid phase names
    my @phaseNames = ();
    
    # this array will define the list of phases to run
    my @phasesToRun = @{$params{Phases}};
    # Check out the Phase Definitions 
    foreach my $name (keys %PhaseDefinitions) {
	my $code = $PhaseDefinitions{$name};
	if (ref($code) ne "CODE") {
	    $self->Print(1, "SpinTestTool::RunST error with Phase '$name': '$code' is not a code reference\n", "bad");
	    goto fail;
	}
	push(@phaseNames, $name);
    }
    
    # for compatability with version 1
    $params{SetupConfiguration} = undef if ($params{Unconfigured});
    my $compatability = undef;
    my @compatPhaseNames = ();
    my %DefaultCode = ();
    foreach my $phase (keys %PhaseDefinitions) {
	if (exists($params{$phase})) {
	    push(@compatPhaseNames, $phase);
	    $compatability = 1;
	    if (!$params{$phase}) { # this phase is not supposed to run
		$PhaseDefinitions{$phase} = undef;
	    } else { # the phase was probably overridden
		$DefaultCode{$phase} = $PhaseDefinitions{$phase};
		$PhaseDefinitions{$phase} = $params{$phase};
	    }
	}
    }
    @phaseNames = @compatPhaseNames if ($compatability);

    # check for any phases to skip before checking the phase option so that it can complain about unsupported phases
    if ($params{SkipPhases}) {
	foreach my $phase (@{$params{SkipPhases}}) {
	    @phasesToRun = grep(!/$phase/, @phasesToRun);
	}
    }

    # If this option is defined, then the user only wants to run certain phases. Check them out for validity.
    if (scalar @{$options->{phase}}) {
	my @phases = @phasesToRun;
	@phasesToRun = ();
	foreach my $phase (@{$options->{phase}}) {
	    my @tphases = grep(/$phase/i, @phases);
	    if ($#tphases > 0) {
		my $phases = join(",", @tphases);
		$self->Print(1, "SpinTestTool::RunST error: conflict with $phase: $phases\n", "bad");
		goto fail;
	    }
	    if ($#tphases == -1) {
		$self->Print(1, "SpinTestTool::RunST error: $phase is an unsupported phase for this ST\n", "bad");
		my $supported = join(", ", @phases);
		$self->Print(-1, "  supported phases: $supported\n", "bad");
		goto fail;
	    }
	    push(@phasesToRun, $tphases[0]);
	}
    }

    # store these in the object for later use
    $self->{TestCommand} = $params{TestCommand};
    $self->{TestTimeout} = $params{TestTimeout};

    # Print out some information that is only needed in extreme debug scenariors
    $self->Print(4, "SpinTestTool:RunST TestCommand initialized as \"$self->{TestCommand}\"\n");
    $self->Print(4, "SpinTestTool:RunST TestTimeout initialized as \"$self->{TestTimeout}\"\n");
    $self->Print(4, "SpinTestTool:RunST testObjective initialized as \"$self->{testObjective}\"\n");

    # We need a server object so that we can talk to the servers
    if (!($self->{ServerObject} = Server->new(Options => $options, SpinTestTool => $self))) {
	$self->Print(1, "Could not create a SpinServer object\n", "bad");
	goto fail;
    }

    my $NServer;
    my $DServer;
    my @StoragePools;
    my @NBladeIPs;

    # we need to know the ST name
    my $name = $self->{Program};
    $name =~ s/\.thpl//;
    $name =~ s/\.st//;

    my @serverList = $self->{ServerObject}->Servers();
    $options->{release} = $self->{ServerObject}->GetRelease() if (!$options->{release});
    $self->{BuildVersion} = $options->{BuildVersion};
    $self->{BuildVersion} = $self->{ServerObject}->GetBuildVersion() if (!$options->{BuildVersion});
    
    if ($params{SetupConfiguration}) {
	# This call will find the cluster configuration and store it in $self->{Config}
	$self->{Config} = $self->{ServerObject}->DiscoverConfig(Nodes=>\@DiscoveryNodes);
	
	# pick a random storage pool in the cluster first. We will then later pick a vif to satisfy the 
	#  local/remote requirement
	# let's look for an online storage pool with available space, or the one that was passed in
	foreach my $poolHash (@{$self->{Config}{StoragePool}}) {
	    # skip any pool that has a server we don't know about for its primary
	    next if (!grep($_ eq $$poolHash{PrimaryServer}, @serverList));
	    if ( scalar(@{$options->{StoragePool}} )) {
		push(@StoragePools, $poolHash) if (grep($$poolHash{Name} eq $_, @{$options->{StoragePool}}));
	    } elsif ($$poolHash{Status} eq "online" && $$poolHash{Avail} > 200) {
		push(@StoragePools, $poolHash);
	    }
	}

	# catch the case where we couldn't find any pools
	if ( scalar(@{$options->{StoragePool}} ) && (@StoragePools == ()) ) {
	    my $name = join(",", @{$options->{StoragePool}});
	    $self->Print(1,"SpinTestTool::RunST did not detect the specified pools as online: $name\n", "bad");
	    goto fail;
	} elsif (@StoragePools == ()) {
	    my $text = "Available servers: ".join(",", @serverList)."\n";
	    foreach my $poolHash (@{$self->{Config}{StoragePool}}) {
		$text .= "$$poolHash{Name} $$poolHash{PrimaryServer} $$poolHash{Avail} $$poolHash{Status}\n";
	    }
	    $self->Print(1,"SpinTestTool::RunST could not find an online storage pool with at least 200MB free\n$text", "bad");
	    goto fail;
	}

	# pick a random one from the list for version 1 compatability
	my $poolHash = $StoragePools[int rand(scalar @StoragePools)];
	if (!scalar(@{$options->{StoragePool}})) {
	    push(@{$options->{StoragePool}}, $$poolHash{Name});
	}
	$options->{VfsServer} = $$poolHash{PrimaryServer};
	if (!$options->{VfsServer} || $options->{VfsServer} eq "") {
	    $self->Print(1, "Could not determine the primary for $$poolHash{Name}\n", "bad");
	}
	
	# set our D server as the server which will be hosting the VFS
	$DServer = $options->{VfsServer};
	
	# now that we have a Dserver, remove it from our list and pick a random one for our Nblade server.
	# use the same server if their is only a single node
	if (!$#{$options->{server}}) {
	    $NServer = $DServer;
	} else {
	    if ($self->{remoteNbladeIP}) { # we need to find a remote IP address in the cluster
		my @serverList = grep($_ ne $DServer, @ {$options->{server}});
		$NServer = $serverList[int rand(scalar @serverList)];
	    } elsif ($self->{localNbladeIP}) { # we need to find a local IP address
		$NServer = $DServer;
	    } else {
		$NServer = undef; # let us use all of the vifs
	    }		
	}
	# logic for determining the correct Virtual Server taking into account the CIFS needs for a test.
	if ( $self->{testObjective} eq "CIFS" ) { # we know we are going to be dealing with CIFS for sure
	    if ( !defined $options->{VirtualServer} ) {
	        if ( $options->{cifsserver} ) { # we have a netbios name and need the VS for it
		    $options->{VirtualServer} = $self->{ServerObject}->GetVirtualServers(CIFSServer=>$options->{cifsserver});
		    if ( !$options->{VirtualServer} ) { # if we don't have a valid Virtual Server break out
		        $self->Print(1,"SpinTestTool::RunST could not find a Virtual Server for CIFS server $options->{cifsserver}.\n", "bad");
		        goto fail;
		    }
		} elsif ( $options->{cifsdomain} ) { # we have a domain, find the VS and CIFSServer that are on it
		    foreach my $serverHash (@{$self->{Config}{CIFS}}) {
			$options->{cifsserver} = $$serverHash{NetBiosName},last if ($$serverHash{Domain} eq $options->{cifsdomain});
		    }
		    unless ( $options->{cifsserver} ) { # try to find one that has the domain we are looking for as a Visible Domain if we didn't find one in the domain we are looking for
			foreach my $serverHash (@{$self->{Config}{CIFSServers}}) {
			    $options->{cifsserver} = $$serverHash{NetBiosName},last if ($$serverHash{VisibleDomains} =~ /$options->{cifsdomain}/);
			}
			if (!$options->{cifsserver}) {
			    $self->Print(1,"SpinTestTool::RunST could not find a CIFS Server for CIFS domain $options->{cifsdomain}.\n", "bad");
			    goto fail;
			}
		    }
		    $options->{VirtualServer} = $self->{ServerObject}->GetVirtualServers(CIFSServer=>$options->{cifsserver});
		    if ( !$options->{VirtualServer} ) { # if we don't have a valid Virtual Server break out
			$self->Print(1,"SpinTestTool::RunST could not find a Virtual Server for CIFS server $options->{cifsserver} in $options->{cifsdomain}.\n", "bad");
			goto fail;
		    }
		} else { # get any VS that has CIFS bound to it
		    $options->{VirtualServer} = $self->{ServerObject}->GetVirtualServers(NServer=>$NServer, DServer=>$DServer, CIFS=>1);
		    foreach my $serverHash (@{$self->{Config}{CIFSServers}}) {
			$options->{cifsserver} = $$serverHash{NetBiosName},last if ($options->{VirtualServer} eq $$serverHash{VirtualServer});
		    }
		    if (!$options->{cifsserver}) {
			$self->Print(1,"SpinTestTool::RunST could not find a Virtual Server with desired CIFS options enabled.\n", "bad");
			goto fail;
		    }
		}
	    } else { # see if CIFS is bound to the Virtual Server we were passed in
		$options->{cifsserver} = $self->{ServerObject}->GetCIFSServers(VirtualServer => $options->{VirtualServer});
		if ( !$options->{cifsserver} ) {
		    $self->Print(1,"SpinTestTool::RunST did not find a CIFS Server bound to Virtual Server $options->{VirtualServer}.\n", "bad");
		    goto fail;
	        }
	    }
	    # make sure that if a cifsshare was passed in, it exists on the Virtual Server that we found
	    my $found = 0;
	    foreach my $shareHash ( @{$self->{Config}{CIFSShares}} ) {
		next if ( $$shareHash{VirtualServer} ne $options->{VirtualServer} ); # don't even bother if this is the wrong Virtual Server
		next if ( $$shareHash{Status} ne "OK" ); # if this share is broken skip it
		if ( defined $options->{cifsshare} && ( $$shareHash{ShareName} =~ /$options->{cifsshare}/i ) ) {
		    $found = 1;
		    last;
		} elsif ( !defined $options->{cifsshare} && ( $$shareHash{Path} =~ /^\/\.readwrite$/ ) ) { # if we have a share at /.readwrite - use it
		    # if this share was created by an ST we need to avoid it unless it was created for this ST
		    if ( $$shareHash{Remark} =~ /Created by RunST for (.+)\.st/) {
			if ( $1 eq $name ) {
			    # print this as a warn because if two STs are running on the same cluster (something that isn't supposed to happen) this could cause trouble
			    $self->Print(1, "SpinTestTool::RunST found existing CIFS Share '$$shareHash{ShareName}' which belongs to this ST '$$shareHash{Remark}'\n", "warn");
			    $options->{cifsshare} = $$shareHash{ShareName};
			    $found = 1;
			    last;
			} else {
			    $self->Print(4, "SpinTestTool::RunST ignoring CIFS Share '$$shareHash{ShareName}' which belongs to ST '$1' : '$$shareHash{Remark}'\n");
			}
		    } else {
			$options->{cifsshare} = $$shareHash{ShareName};
			$found = 1;
			last;
		    }
		}
	    }
	    # if we didn't see a share that was passed in we need to error out, otherwise we will create a CIFS Share are /.readwrite later
	    if ( !$found && defined $options->{cifsshare} ) {
		$self->Print(1,"SpinTestTool::RunST could not find CIFS Share \"$options->{cifsshare}\" on CIFS Server \"$options->{cifsserver}\".\n", "bad");
		goto fail;
	    }
	# if the objective isn't CIFS we can use any Virtual Server
	} elsif (!$options->{VirtualServer}) {
	    my $ip = undef; # look for a vserver with this IP if one is provided
	    $ip = ${$options->{nbladeIP}}[0] if (scalar(@{$options->{nbladeIP}}));
	    $options->{VirtualServer} = $self->{ServerObject}->GetVirtualServers(IP=>$ip,NServer=>$NServer,RunST=>1);
	} elsif (scalar(@{$options->{nbladeIP}})) {
	    # we were given an IP and a vserver; make sure they are OK
	    my %vserver;
	    foreach my $vserver (@{$self->{Config}{VirtualServer}}) {
		%vserver = %$vserver if ($$vserver{Name} eq $options->{VirtualServer});
	    }
	    foreach my $vif (@{$self->{Config}{VIF}}) {
		my $vifname = $$vif{SpinServerID}.":".$$vif{Name};
		if ($vifname && $vserver{NetworkInterfaces} !~ /$vifname/) {
		    $self->Print(1, "The specified IP ($$vif{IPAddress}) is not bound to the given vserver ($options->{VirtualServer})\n", "bad");
		    goto fail;
		}
	    }
	}
	if (!$options->{VirtualServer}) { 
	    $self->Print(1,"SpinTestTool::RunST could not find a virtual server on $NServer\n", "bad");
	    goto fail;
	}
	
	# Now it is time to find some VIFs to use. If addresses were passed in on the command line then use them.
	#  If not, then we need to find some based on our local/remote requirement
	if (!$options->{nbladeIP} || @{$options->{nbladeIP}} == ()) {
	    foreach my $vif (@{$self->{Config}{VIF}}) {
		push(@NBladeIPs, $$vif{IPAddress}) if ($$vif{VirtualServer} eq $options->{VirtualServer} && (!$NServer || $$vif{SpinServerID} eq $NServer));
	    }
	    if (@NBladeIPs == ()) {
		$self->Print(1, "RunST ERROR getting a remote IP address for Virtual Server $options->{VirtualServer}\n", "bad");
		goto fail;
	    }
	    $options->{nbladeIP} = \@NBladeIPs;
	} else {
	    @NBladeIPs = @{$options->{nbladeIP}};
	}
	my $ips = join(",", @{$options->{nbladeIP}});
	my $spnames = join(",", @{$options->{StoragePool}});
	
	$self->Print(3,"SpinTestTool::RunST localNbladeIP set to $self->{localNbladeIP}\n", "info");
	$self->Print(3,"SpinTestTool::RunST remoteNbladeIP set to $self->{RemoteNbladeIP}\n", "info");
	$self->Print(3,"SpinTestTool::RunST set VfsServer to $options->{VfsServer}\n", "info");
	$self->Print(3,"SpinTestTool::RunST set VirtualServer to $options->{VirtualServer}\n", "info");
	$self->Print(3,"SpinTestTool::RunST set StoragePool to $spnames\n", "info");
	$self->Print(3,"SpinTestTool::RunST set DServer to $options->{VfsServer}\n", "info");
	$self->Print(3,"SpinTestTool::RunST set NServer to $NServer\n", "info");
	$self->Print(3,"SpinTestTool::RunST set nbladeIP to $ips\n", "info");
	$self->Print(3,"SpinTestTool::RunST cifsserver set to $options->{cifsserver}\n", "info");
	$self->Print(3,"SpinTestTool::RunST cifsdomain set to $options->{cifsdomain}\n", "info");
	$self->Print(3,"SpinTestTool::RunST cifsshare set to $options->{cifsshare}\n", "info") if ( defined $options->{cifsshare} );
    }
    $self->{Clients} = $options->{client};
    
    my $prevclient;
    my $instance = 0;

    foreach my $client (sort @ {$self->{Clients}}) {
	if ($client ne $prevclient) {
	    $instance = 0;
	} else {
	    $instance++;
	}
	$prevclient = $client;
	my $rwdir = "";
	if (!$params{MirrorTest} && $options->{release} >= 2) { # we might have a mirror so let's deal with it
	    $rwdir = ".readwrite/";
	}
	my $testdir = $params{TestDir};
	if ( !$testdir ) {
	    $testdir = "$name.$NBladeIPs[0].$instance";
	    if ($self->{testObjective} eq "CIFS"){
		$testdir = "/cifs/$testdir";
	    } else {
		$testdir = "/nfs/$testdir";
	    }
	}
	$self->Print(3,"SpinTestTool::RunST testdir set to $testdir on $client\n", "info");

	my %testOptions = $self->_CreateTestHash($params{TestOptions});
	$testOptions{testdir} = $testdir if ($testdir);
	$testOptions{VfsServer} = $DServer if ($DServer);
	$testOptions{cifsshare} = $name if ( !defined $options->{cifsshare} );


	if ($params{SetupConfiguration}) {
	    $self->{serverIP}{$client}{$instance} = $NBladeIPs[int rand(scalar @NBladeIPs)];
	    if (!exists $testOptions{vfs}) {
		$testOptions{vfs} = "${name}_${client}_$instance";
		$testOptions{share} = $self->{ServerObject}->RootShare($DServer, $rwdir, "${name}_${client}_$instance");
		$self->{share}{$client}{$instance} = $testOptions{share};
		push(@{$self->{VFSName}}, "${name}_${client}_$instance");
	    } else {
		$testOptions{share} = $options->{share};
		push(@{$self->{VFSName}}, $testOptions{vfs});
		$self->{share}{$client}{$instance} = "/$rwdir$name.$client.$instance"; #should be determined from vfs
	    }
	    $self->{MountPoint}{$client}{$instance} = $testdir if (!$self->{MountPoint}{$client}{$instance});
	}
	$self->{testOptions}{$client}{$instance} = \%testOptions;
    }
    
    my $CreateVFS = 1;
    # save the IP so that the SetACLS call can be made. Loop through all available nbladeIPs and find one that is configured
    #  since the test script might have many nbladeIPs set but not configured.
    my $ServerIP;
    foreach my $vif (@{$self->{Config}{VIF}}, @{$$Configuration{VIF}}) {
	if (grep($$vif{IPAddress} eq $_, @{$options->{nbladeIP}})) {
	    $ServerIP = $$vif{IPAddress};
	    $self->Print(4,"SpinTestTool::RunST set \$ServerIP to '$ServerIP'\n", "info");
	    last;
	}
    }
    if (%{$Configuration} != ()) {    
	foreach my $vshash (@{$$Configuration{VirtualServer}}) {
	    if ($$vshash{Name} !~ /\.st$/) {
		$self->Print(1, "The virtual server '$$vshash{Name}' should be of the form test.st\n", "bad");
		goto fail;
	    }
	    my $poolhash = $StoragePools[int rand(scalar @StoragePools)];
	    $$vshash{SPName} = $$poolhash{Name} if (!$$vshash{SPName});
	    $$vshash{User} = 99 if (!$$vshash{User});
	}
	if (exists($$Configuration{VIF}) && scalar(@{$$Configuration{VIF}})) {
	    # set port in vif creation
	    # set ssid in vif creation
	    my %ports;
	    foreach my $server (@serverList) {
		$ports{$server} = [];
		foreach my $porthash (@{$self->{Config}{$server}{Ports}}) {
		    push(@{$ports{$server}}, $$porthash{Name}) if ($$porthash{Name} =~ /serv/ &&
								   $$porthash{Status} eq "up");
		}
	    }
	    my $servers = ($NServer) ? [$NServer] : \@serverList;
	    foreach my $vifhash (@{$$Configuration{VIF}}) {
		$$vifhash{SSID} = $$servers[int rand(scalar @$servers)] if (!$$vifhash{SSID});
		$$vifhash{VirtualServer} = $options->{VirtualServer} if (!$$vifhash{VirtualServer});
		my @ports = @{$ports{$$vifhash{SSID}}};
		$$vifhash{NetPortName} = $ports[int rand(scalar @ports)] if (!$$vifhash{NetPortName});
	    }
	}
	if (exists($$Configuration{VFS}) && scalar(@{$$Configuration{VFS}})) {
	    foreach my $vfshash (@{$$Configuration{VFS}}) {
		my $poolhash = $StoragePools[int rand(scalar @StoragePools)];
		$$vfshash{StoragePool} = $$poolhash{Name} if (!$$vfshash{StoragePool});
		$$vfshash{VirtualServer} = $options->{VirtualServer} if (!$$vfshash{VirtualServer});
		$$vfshash{User} = 99 if (!$$vfshash{User});
		$$vfshash{ServerIP} = $ServerIP if (!$$vfshash{ServerIP});
	    }
	    $CreateVFS = undef;
	}
    }
    if ($CreateVFS) {
	foreach my $vfsname (@{$self->{VFSName}}) {
	    my $poolhash = $StoragePools[int rand(scalar @StoragePools)];
	    my $pool = $$poolhash{Name};
	    my $dserver = $$poolhash{PrimaryServer};
	    my $chmod = 777;
	    my $user = 99;
	    my $group = 99;
	    my $domain = 0;
	    my $usertype = "SPIN";
	    
	    # handle the CIFS case where we want to use the cifs properties passed in, or NETAPP\SpinDexD
	    if ( $self->{testObjective} eq "CIFS" ) {
		$user = (defined $options->{cifsuser}) ? $options->{cifsuser} : "spindexd";
		$group = (defined $options->{cifsgroup}) ? $options->{cifsgroup} : "\"Domain Users\"";
		$domain = (defined $options->{cifsdomain}) ? $options->{cifsdomain} : "Netapp";
		$usertype = "CIFS";
		$chmod = undef;
	    }
	    push(@{$$Configuration{VFS}}, 
		 {
		     Name => $vfsname,
		     StoragePool => $pool,
		     VirtualServer => $options->{VirtualServer},
		     Mount => "/$vfsname",
		     MaxQuota => 100000,
		     Chmod => $chmod,
		     UserType => $usertype,
		     User => $user,
		     Group => $group,
		     Domain => $domain,
#		     Server => $dserver,
		     ServerIP => $ServerIP,
		 });
	}
    }
    
    # if we don't have a share, we need to create one at "/.readwrite"
    if ( $self->{testObjective} eq "CIFS" ) {  
        if ( !defined $options->{cifsshare} ) {
	    $options->{cifsshare} = "$name";
	    push(@{$$Configuration{CIFSShares}}, {
		VirtualServer	=> $options->{VirtualServer},
		Name	=> "$name",
		Path	=> "/.readwrite",
		Remark  => "\"Created by RunST for $name.st\"",
	    });
	}
    }

    $self->{ServerObject}->SetConfiguration($Configuration);

    # we don't use spindex when running under nate
    if (!$self->{NATE}) {
	eval "use SpinDexClient";
	if  ($@) {
	    print $@, "\n\n\n";
	    exit(1);
	}
	$self->{SpinDex} = SpinDexClient->new(DefaultUser => $options->{user},
					      DefaultPassword => $options->{password});
    }
    foreach my $phase (@phasesToRun) {
	my $rc;
	my $code = $PhaseDefinitions{$phase};
	next if (!$code);
	$self->Print(2, "SpinTestTool::RunST in $phase phase\n");
	if ($phase eq "Test" && $params{ModifyTestOptions}) {
	    if (!($rc=&{$params{ModifyTestOptions}}($self))) {
		$self->Print(1, "SpinTestTool::RunST The ModifyTestOptions code returned with rc=$rc\n", "bad");
		goto fail;
	    }
	}
	if (!($rc=&$code($self, $DefaultCode{$phase}))) {
	    $self->Print(1, "SpinTestTool::RunST The $phase code returned with rc=$rc\n", "bad");
	    goto fail;
	}
    }

  pass:
    $self->HandleResults(ReturnCode => 0, ReportResults=>!$self->{NATE});
    # we have now exited
  fail:
    _Cleanup($self);
    $self->HandleResults(ReturnCode => 1, ReportResults=>!$self->{NATE});

    # we have now exited
} # end RunST

#***********************************************************************

=head1 C<ClientMounts>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************

sub ClientMounts {
    my ($self, $options, $spindex) = @_; # spindex defined here for old ST files xxREMOVExx
    my $retval = 1;
    my $msg = undef;
    my $destroySpindex;
    if (!$spindex && !$self->{NATE}) {
	eval "use SpinDexClient";
	if ($@) {
	    print $@, "\n\n\n";
	    exit(1);
	}
	$spindex = SpinDexClient->new();
	$destroySpindex = 1;
    }

    $self->Print(3, "SpinTestTool::ClientMounts entered\n");
    my %Clients = ();
    foreach my $client (@{$options->{client}}) {
	next if ($Clients{$client});
	$Clients{$client} = 1;
	if ($options->{Mounts}{$client}) {
	    foreach my $mount (@{$options->{Mounts}{$client}}) {
		our $retry;
		$retry = 1 if (!defined($retry));

		# the print is conditional on the type of mount we're doing (we only expect to see {nfsmountoptions}
		# or {cygmountmountoptions} passed in so we print the options next to each other
		unless ($mount->{mounttype} eq "win32") {
		    $self->Print(3, "SpinTestTool::ClientMounts => mount $mount->{nfsmountoptions}$mount->{cygwinmountoptions} $mount->{server}:$mount->{share} $mount->{mountpoint} on $client\n");
		} else { # this will really call "net use"
		    $self->Print(3, "SpinTestTool::ClientMounts => net use $mount->{mountpoint} \\\\$mount->{server}\\$mount->{share} $mount->{win32mountoptions} on $client\n");
		}
		if ($self->{NATE}) {
		    eval "use mount (mount)";
		    if ($@) {
			print $@, "\n\n\n";
			exit(1);
		    }
		    main::host($client);
		    my $cc = main::connect('#rsh');
		    my $mount_point = mount($mount->{server}, $mount->{share},
					    client => $client,
					    mountpoint => $mount->{mountpoint},
					    mountopts => $mount->{nfsmountoptions});
		} else {
		    ($retval,$msg) = $spindex->Rmount(Host => $client,
						      Server => $mount->{server},
						      Share => $mount->{share},
						      MountPoint => $mount->{mountpoint},
						      NFSMountOptions => $mount->{nfsmountoptions},
						      Win32MountOptions => $mount->{win32mountoptions},
						      CygwinMountOptions => $mount->{cygwinmountoptions},
						      MountType => $mount->{mounttype},
						      );
		    if ($retry && $retval == 1) {
			$retry = 0;
			redo;
		    }
		    if ( $retval < 0 || $retval > 0 ) {
			my $error = "ERROR doing Rmount on client $client, rc = $retval";
			$error .= ": $msg" if ($msg);
			$self->Print(0, "$error\n", "bad");
			# We should likely undo here
#		    $self->ClientUmounts($options->{Mounts});
			next if (!$options->{StopOnErrors});
			$spindex->DESTROY() if ($destroySpindex);
			return undef;
		    }
		}
		push(@{ $self->{createdMountPoints}{$client}}, $mount->{mountpoint}) if ($mount->{mountpoint});
	    }
	} else {

            $self->Notify(To=>"snider\@netapp.com",
                          From=>"$ENV{USER}\@netapp.com",
                          Subject=>"ClientMounts usage",
                          Message=>"$0 is not using option->{Mounts}\n");
 
	    $self->Print(3, "SpinTestTool::ClientMounts => mount $options->{nfsmountoptions} $options->{server}[0]:$options->{share} $options->{mountpoint}\n");
	    # XXX
	    # {mountoptions} was deprecated on 9/24/2002, remove warning after a month or so
	    # XXX   
	    if ($options->{mountoptions}) {
		$self->Print(1, "SpinTestTool::ClientMounts \"\$options->{mountoptions}\" deprecated. Please use \"\$options->{nfsmountoptions}\" instead.  Handling this for you right now.\n", "warn");
		$options->{nfsmountoptions} = $options->{mountoptions};
	    }
	    ($retval, $msg) = $spindex->Rmount(Host => $client,
				      Server => $options->{server}[0],
				      Share => $options->{share},
				      MountPoint => $options->{mountpoint},
				      NFSMountOptions => $options->{nfsmountoptions},
				      Win32MountOptions => $options->{win32mountoptions},
				      CygwinMountOptions => $options->{cygwinmountoptions},
				      MountType => $options->{mounttype},
					);
	    if ( $retval < 0 || $retval > 0 ) {
		my $error = "ERROR doing Rmount on client $client, rc = $retval";
		$error .= ": $msg" if ($msg);
		$self->Print(0, "$error\n", "bad");
		# We should likely undo here
#		$self->ClientUmounts();
		$spindex->DESTROY() if ($destroySpindex);
		return undef;
	    }
	}
    }
    $self->Print(3, "SpinTestTool::ClientMounts exited\n");
    $spindex->DESTROY() if ($destroySpindex);
    if (!$retval) {
	return ($msg =~ /\w+/) ? $msg : 1;
    } elsif ($retval =~ /\w+/) {
	return $retval;
    } else {
	return undef;
    }
}

#***********************************************************************

=head1 C<ClientUmounts>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub ClientUmounts {
    my ($self, $Mounts) = @_;
    my $retval = 1;
    my $rc = 1;
    my $msg = undef;
    my $destroySpindex;
    my $spindex = $self->{SpinDex};
    my %Mounts;

    $self->Print(3, "SpinTestTool::ClientUMounts entered\n");
    if (!$spindex && !$self->{NATE}) {
	$spindex = SpinDexClient->new();
	$destroySpindex = 1;
    }

    if ($Mounts) {
	%Mounts = %$Mounts;
    } elsif ($self->{MountPoint}) {
	foreach my $client (keys %{$self->{MountPoint}}) {
	    $Mounts{$client} = [];
	    foreach my $instance (keys %{$self->{MountPoint}{$client}}) {
		push(@{$Mounts{$client}}, $self->{MountPoint}{$client}{$instance});
	    }
	}
    }
    foreach my $client (keys %Mounts) {
	foreach my $mntpnt (@{ $Mounts{$client}}) {
	    $self->Print(3, "SpinTestTool::ClientUMounts: umount $mntpnt on $client\n");
	    
	    if ($self->{NATE}) {
		eval "use mount (umount)";
		if ($@) {
		    print $@, "\n\n\n";
		    exit(1);
		}
		main::host($client);
		my $cc = main::connect('#rsh');
		$retval = umount($mntpnt, 
				 client => $client);
		$msg = $@ if (!$retval);
	    } else {
		($retval, $msg) = $spindex->Rumount(Host => $client,
						    MountPoint => $mntpnt,
						    );
		# convert $retval to perl return code
		$retval = undef if ( $retval < 0 || $retval > 0);
	    }
	    $self->Print(3, "SpinTestTool::ClientUMounts returned ($retval, $msg)\n");
	    next if ($msg =~ /not mounted|cannot find/i); # not mounted
	    if (!$retval) {
		$self->Print(0, "ERROR doing Rumount on client $client, rc = $retval\n");
		$rc = undef;
		# Drive on to make sure that we clean up all clients
		if ($msg =~ /\w+/) {
		    $self->Print(1, "SpinTestTool::ClientUmounts error on $client => $msg\n", "bad");
		}
	    }
	}
    }
    $self->Print(3, "SpinTestTool::ClientUMounts exited with $rc\n");
    $spindex->DESTROY() if ($destroySpindex);
    return $rc;
}


#***********************************************************************

=head1 C<HandleResults()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub HandleResults {
    my $self = shift;
    my $status = 1;
    my $msg = "Failed";
    my $result = "bad";
    my %params = (SpinDexHandle => $self->{SpinDex},
		  SpinServerHandle => $self->{ServerObject},
		  ReturnCode => undef,
		  ReportResults => 1,
		  Message => undef,
		  Exit => 1,
		  @_, # get any parameters that were explicitly passed
		  );
    $self->Print(3, "SpinTestTool::HandleResults entered\n");
    # We need to call this now before we exit because Expect->DESTOY calls 
    # waitpid which can cause the $? to be modified upon exit. The SpinServer 
    # DESTROY will clean up properly.
    if ($params{SpinServerHandle}) { 
	$params{SpinServerHandle}->DESTROY();
    }

    if ($params{ReturnCode} == 0) {
	$status = 0;
	$msg = "Passed";
	$result = "good";
	if ($self->{NATE} && $params{ReportResults}) {
	    main::logresult('PASS', $params{Message});
	  }
    } else {
	if ($self->{NATE}) {
	    main::logresult('FAIL', $params{Message});
	  }
    }
    $self->Print(1, "$msg\n", $result);
    if ( $self->{notify} ne undef ) {
	$self->Print(1, "Notifying $self->{notify}\n");
	$self->Notify(
		     To => $self->{notify},
		     From => "",
		     Subject => "$self->{FullScript} $msg at " . localtime(),
		     );
    }
    if ($params{Exit} && $self->{STmode}) {
	$self->Print(2, "SpinTestTool::HandleResults exiting with $status\n");
	_exit $status; # _exit
    } else {
	$self->Print(3, "SpinTestTool::HandleResults returning\n");
    }
}

#***********************************************************************
=head1 C<StartTestResults()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub StartTestResults {
    my $spinserver;
    my $self = shift;
    my %params = (Objective => $self->{testObjective},
		  Test_Name => $ {$self->{options}}{testcase},
		  Release => undef,
		  Tester => $ {$self->{options}}{tester},
		  Platform => "Linux",
		  SpinServer => $self->{ServerObject},
		  @_, # get any parameters that were explicitly passed
		  );
    my $program = shift;

    if (!defined($params{Release})){
	# Try to retrieve the release from the server
	$params{Release} = $params{SpinServer}->GetRelease();
    }

    # if we don't know the tester, make a best guess
    if ( !defined $params{Tester} ) {
	if ( $^O =~ /mswin32/i ) { # windows
	    # LoginName is part of the core perl on Win32 and does require Win32 to be loaded
	    $params{Tester} = Win32::LoginName();
	} else { # *nix
	    $params{Tester} = (getpwuid($<))[0];
	}
    }
	

    if ($self->{testResults}) {
	eval "use TestResults";
	if ($@){
	    print $@, "\n\n\n";
	    print "An error occurred while loading TestResults.  Try setting LD_LIBRARY_PATH in\n";
	    print "your environment to /usr/local/lib and trying again\n";
	    exit;
	}
	if ($spinserver = $params{SpinServer}) {
	    $self->{Ops} = $spinserver->GetOps();
	    $self->{spinserver} = $spinserver;
	}
	my $connection = DBConnection->new(Release=>$params{Release});
	$self->{results} = Result->new(Release=>$params{Release}, Connection=>$connection,
				       Test_Objective=>$params{Objective});
	$self->{results}->start(Test_Name=>$params{Test_Name},Tester=>$params{Tester}, Platform=>$params{Platform});
    }
}

#***********************************************************************

=head1 C<StopTestResults()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub StopTestResults {
    my $self = shift;
    my %params = (MountPoint => undef,
		  Ops => undef,
		  @_, # get any parameters that were explicitly passed
		  );
    my $spinserver;
    if ($self->{results}) {
	$self->{results}->finish();
    }
    if (defined($self->{Ops}) && ($spinserver = $self->{spinserver})) {
	my $ops = $self->{Ops};
	$self->{Ops} = $spinserver->GetOps() - $ops;
    }
    $self->{Ops} = $params{Ops} if (defined($params{Ops}));
}

#***********************************************************************

=head1 C<InterruptHandler()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub InterruptHandler {
    exit 1 if (!defined($CleanupSTT));
#    if ( $CleanupNotifier ne undef ) {
#	$CleanupSTT->Print(1, "Notifying $CleanupNotifier\n");
#	$CleanupSTT->Notify(
#		     To => $CleanupNotifier,
#		     From => "",
#		     Subject => "$self->{FullScript} at " . localtime(),
#		     );
#    }
    $SIG{INT} = undef;
    $CleanupSTT->Print(0, "\n");
    _Cleanup($CleanupSTT) if $CleanupSTT->{RunST};
    $CleanupCode->(shift) if $CleanupCode;
    exit 1;
}

#***********************************************************************
=head1 C<Help()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************

sub Help {
    my $self = shift;
    my $Options = shift;
    my $Descriptions = shift;
    my $SpecificHelp = (shift || "ALL");
    
    print << "EOF";

Usage:

$self->{Program}
    -client <host[,host,..]> [-client <host[,host,..]>] | -clientfile <filename>
    -server <host[,host,..]> [-server <host[,host,..]>] 
    -quick
    [-verbosity <0..n>] [-stverbosity <0..n>]
    [-clientuser <user>] [-clientpassword <password>]
    [-serveruser <user>] [-serverpassword <password>]
    [-cliuser <user>] [-clipassword <password>]
    [-source <sourcedir>]
    [-target <dir | host:/share/dir | \\\\host\\share\\dir>]
    [-seed <randomseed>]
    [-notify <email>]
    [-help]
    [-getrequirements]
    [-initFile <filename>]
    [-duration <time>]
    [-iterations <count>]
    [-testOptions <comma-separated string of options> [-testOptions ...]]
    [-testcase <name of test case>] [-listtestcases]
    [-testResults <0|1>]
    [-tester <username>]
    [-nbladeIP <IP address[:range]> | -nbladeIP <IP address[,IP address ...]]
    [-dbladeIP <IP address[:range]> | -dbladeIP <IP address[,IP address ...]]
    [-serverIP <IP address[:range]> | -serverIP <IP address[,IP address ...]] This is old
    [-vfsserver <ssid>]
    [-virtualserver <vsname>]
    [-storagepool <spname>]
    [-StopOnErrors]
    [-verify]
    [-dataType <Short|Long>]
EOF
    if($SpecificHelp =~ /CIFS|ALL/){
	print << "EOF";
    [-cifsdomain <CIFS domain>]
    [-cifsshare <CIFS share>]
    [-cifsserver <CIFS server>]
EOF
    }
    if($SpecificHelp =~ /NFS|ALL/){
	print << "EOF";
    [-nfsversion <2|3>]
    [-protocol <tcp|udp>]
EOF
    }
    print $Options;
    print << "EOF";

    -client defines which clients are used for the test
    -clientfile specifies a file with the list of clients listed on each line
    -server defines which servers are used for the test
    -quick speciifes that the test should run its shortest form
    -verbosity specifies how much output should be printed [1]
    -stverbosity specifies the verbosity level for the ST file itself. Same as verbosity by default.
    -clientuser specifies the login name for the clients [root]
    -clientpassword specifies the login password for the clientuser [spinwin]
    -serveruser specifies the login name for the servers [root]
    -serverpassword specifies the login password for the serveruser [admin]
    -cliuser specifies the login name for the admin user [admin]
    -clipassword specifies the login password for the cliuser [admin]
    -source specifies where the source code can be found
    -target specifies where the source code should be copied before executing
    -seed specifies a seed to use by the random number generator
    -notify specifies an email account to notify with the results
    -help prints this message
    -getrequirements will print out the requirements to run this test
    -initFile specifies a file to read the server configuration information from
    -duration specifies how long the test should try to run
    -iterations specifies the number of times the test should run
    -testOptions is a way to pass special options to an underlying program
    -testcase specifies a specific testcase to run
    -listtestcase will print out the possible testcases for the test
    -testResults specifies whether or not the results should be place in the database.
      the default for a .st file is true and false otherwise
    -tester specifies the name to store in the database as the tester [${ $self->{options}}{tester}]
    -nbladeIP will specify a specific server IP address to use for the N blade(can optionally specify a range with :digit)
    -dbladeIP will specify a specific server IP address to use for the D blade(can optionally specify a range with :digit)
    -serverIP is an old flag that will go away
    -skiperrors will skip any errors in Iterate() and continue
    -verify will try and verify the data for VFS ops
    -dataType specifies how much data to use when calling ModifyPath
EOF
    if($SpecificHelp =~ /CIFS|ALL/){
	print << "EOF";
    -cifsdomain specifies the CIFS domain to use
    -cifsserver specifies the Netbios name of the CIFS server to use
    -cifsshare specifies the share on the CIFS server to use
EOF
    }
     if($SpecificHelp =~ /NFS|ALL/){
	print << "EOF";
    -nfsversion specifies which version of NFS to mount
    -protocol specifies which transport protocol to use when mounting
EOF
    }
	
    print $Descriptions;
    print "\n";
    
    return;
} # end Help

#***********************************************************************

=head1 C<CleanupTarget()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub CleanupTarget {
    my $self = shift;
    my %params = (Host => undef,
		  File => [],
		  Target => "/tmp",
		  @_, # get any parameters that were explicitly passed
		  );

    # Parse the Target to see if it is using an NFS syntax, a CIFS
    # syntax, or a local syntax. If it is a simple path (like /tmp/my/path)
    # then simply FTP the file(s) to that directory on the client (Host).
    # If the Target is NFS or CIFS, then that path will be "mounted"
    # appropriately on the Host, and then the file(s) will be accessible
    # (access/read/write/execute ability must be verified by this method)
    # by the SpinDex server on that remote Host.

    # NFS path - host:/share/dir
    # CIFS (UNC) path - \\host\share\dir
    # Simple path - /simple/dir

    my @fields;
    my $remainder;
    my $host;
    my $share;
    my $targetPath;
    my $rc = 0;
    my $spindex;
    my $file;
    my $fileList;
    my $command;
    my $pid;
    my @path;
    my $pathComp;
    my $createDir;
    my $dir;
    my $rootMount = "/tmp/"; # use /tmp so we can do it as non-root user

    eval "use SpinDexClient";
    if ($@) {
	print $@, "\n\n\n";
	exit(1);
    }
    $spindex = main::SpinDexClient->new(
    # okeefe: bryan says I cannot use the DefaultUser and DefaultPassword
    # when this is an "embedded" spindex object. By "embedded", I mean using
    # the SpinDexClient object within this object, while another SpinDexObject
    # object is (or will be) in use by the test that has created this
    # SpinTestTool object. Thus, there are actually two SpinDexObjects
    # instantiated by the same calling script: one directly instantiated, 
    # and one indirectly instantiated.
#				  DefaultUser => "root",
#				  DefaultPassword => "spinwin",
				  );

    # --------------------------------------------------------------
    # Is this an NFS path?
    # --------------------------------------------------------------
    if ( index($params{Target}, ":") >= 0 ) {
	$self->Print(3, "SpinTestTool::CleanupTarget $params{Target} is an NFS path\n");
	@fields = split(/:/, $params{Target});
	$host = @fields[0];
	$share = @fields[1];
	$targetPath = "$rootMount$host$share";

	foreach $file (@{$params{File}}) {
	    $command = "rm -rf $targetPath/$file";
	    $rc = $spindex->Rexec(Host => $params{Host},
				  Command => $command,
				  Wait => 1,
				  );
	    if ( $rc < 0 ) {
		$self->Print(0, "SpinTestTool::CleanupTarget ERROR, Rexec $command, rc = $rc\n");
		goto fail_and_umount;
	    } else {
		$pid = $rc;
	    }
	    $rc = $spindex->ExitValue(Pid => $pid);
	    if ( $rc != 0 ) {
		$self->Print(0, "SpinTestTool::CleanupTarget ERROR, $command, rc = " . eval($rc >> 8) . "\n");
		goto fail_and_umount;
	    } else {
		$self->Print(3, "SpinTestTool::CleanupTarget $command, rc = " . eval($rc >> 8) . "\n");
	    }
	}

	# Unmount the target on the client.
	$command = "umount $rootMount$host$share";
	$rc = $spindex->Rexec(Host => $params{Host},
			      Command => $command,
			      Wait => 1,
			      );
	if ( $rc < 0 ) {
	    $self->Print(0, "SpinTestTool::CleanupTarget ERROR, Rexec $command, rc = $rc\n");
	    goto fail;
	} else {
	    $pid = $rc;
	}
	$rc = $spindex->ExitValue(Pid => $pid);
	if ( $rc != 0 ) {
	    $self->Print(0, "SpinTestTool::CleanupTarget ERROR, $command, rc = " . eval($rc >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::CleanupTarget $command, rc = " . eval($rc >> 8) . "\n");
	}

    # --------------------------------------------------------------
    # Is this a CIFS path?
    # --------------------------------------------------------------
    } elsif ( index($params{Target}, "\\") >= 0 ) {
	$self->Print(3, "$params{Target} is a CIFS path\n");
	@fields = split(/\\\\/, $params{Target});
	$remainder = $fields[1];
	$self->Print(3, "remainder = $remainder\n");
	@fields = split(/\\/, $remainder);
	$host = $fields[0];
	$self->Print(3, "host = $host\n");

    # --------------------------------------------------------------
    # This is a simple path.
    # --------------------------------------------------------------
    } else {
	$self->Print(3, "$params{Target} is a simple path\n");
	$targetPath = $params{Target};
	foreach $file (@{$params{File}}) {
	    $command = "rm -rf $targetPath/$file";
	    $rc = $spindex->Rexec(Host => $params{Host},
				  Command => $command,
				  Wait => 1,
				  );
	    if ( $rc < 0 ) {
		$self->Print(0, "SpinTestTool::CleanupTarget ERROR, Rexec $command, rc = $rc\n");
		goto fail;
	    } else {
		$pid = $rc;
	    }
	    $rc = $spindex->ExitValue(Pid => $pid);
	    if ( $rc != 0 ) {
		$self->Print(0, "SpinTestTool::CleanupTarget ERROR, $command, rc = " . eval($rc >> 8) . "\n");
		goto fail;
	    } else {
		$self->Print(3, "SpinTestTool::CleanupTarget $command, rc = " . eval($rc >> 8) . "\n");
	    }
	} # end foreach
    } # end if

  pass:
    $spindex->DESTROY(); # for some reason, we need to call this explicitly
    return 0;

  fail_and_umount:
    $command = "cd /";
    $rc = $spindex->Rexec(Host => $params{Host},
			  Command => $command,
			  Wait => 1,
			  );
    $command = "umount $rootMount$host$share";
    $rc = $spindex->Rexec(Host => $params{Host},
			  Command => $command,
			  Wait => 1,
			  );
    if ( $rc < 0 ) {
	$self->Print(0, "SpinTestTool::CleanupTarget ERROR, Rexec $command, rc = $rc\n");
	goto fail;
    } else {
	$pid = $rc;
    }
    $rc = $spindex->ExitValue(Pid => $pid);
    if ( $rc != 0 ) {
	$self->Print(0, "SpinTestTool::CleanupTarget ERROR, $command, rc = " . eval($rc >> 8) . "\n");
	goto fail;
    } else {
	$self->Print(3, "SpinTestTool::CleanupTarget $command, rc = " . eval($rc >> 8) . "\n");
    }

  fail:
    $spindex->DESTROY(); # for some reason, we need to call this explicitly
    return 1;
} # end CleanupTarget


#***********************************************************************

=head1 C<Mount()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************

sub Mount {
    my $self = shift;
    my %params = (Server => undef,
		  Share => undef,
		  MountPoint => undef, # if specified, must be fully qualified
		  MountOptions => undef,
		  User => undef,
		  Password => undef,
		  @_, # get any parameters that were explicitly passed
		  );

    my @path;
    my $pathComp = undef;
    my $createDir;
    my $spinRoot = "/spin";
    my $command;
    my $rc = 0;

    $self->Print(3, "SpinTestTool::Mount entered\n");
    if ( $params{Server} eq undef ) {
	$self->Print(0, "SpinTestTool::Mount ERROR, Server is required\n");
	return undef;
    }

    # Set up the default Share to use. Give the user some leeway by adding
    # a leading "/" even if it's not there, since that's what they meant to
    # do.
    if ( $params{Share} =~ /\// ) {
	# do nothing, since the leading "/" was already specified
    } elsif ( $params{Share} ne undef ) {
	# no leading "/", but there was a share given
	$params{Share} = "/" . $params{Share};
    } elsif ( $params{Share} eq undef ) {
	# nothing specified, choose a share name for a user
	$params{Share} = "/" . $self->{InitialShare}++;
    }

    # Set up the default MountOptions to use. Give the user some leeway by
    # adding a "-o" even if it's not there, since that's what they meant to
    # do.
    if ( $params{MountOptions} =~ /-o/ ) {
	# do nothing, since the -o option was already specified
    } elsif ( $params{MountOptions} ne undef ) {
	$params{MountOptions} = "-o " . $params{MountOptions};
    }

    # Set up the default MountPoint name to use.
    if ( $params{MountPoint} eq undef ) {
	$params{MountPoint} = "$spinRoot/$params{Server}$params{Share}";
#	$params{MountPoint} = "$spinRoot/$params{Server}/" . $self->{InitialMountPoint}++;
    }

    @path = split(/\//, $params{MountPoint});
    
    # Create any directories that are needed for this mount.
    foreach $pathComp (@path) {
	next if ( $pathComp eq undef );
	$createDir .= "/$pathComp";
	if ( !(-e $createDir) ) {
	    $self->{DirStack}[$self->{DirStackSize}] = $createDir;
	    $self->{DirStackSize}++;
	    $self->Print(3, "SpinTestTool::Mount mkdir $createDir\n");
	    if ( !mkdir($createDir, 0777) ) {
		$self->Print(0, "SpinTestTool::Mount ERROR, mkdir $createDir failed: $!.\n");
		return undef;
	    }
	}
    }

    # Mount it.
    if ( $params{Server} ne undef ) {
	if ( $OSNAME =~ /win/ ) { # CIFS mount (on Cygwin)
	    $command = "mount $params{MountOptions} //$params{Server}$params{Share} $params{MountPoint}";
	} else { # NFS mount
	    $command = "mount $params{MountOptions} $params{Server}:$params{Share} $params{MountPoint}";
	}
	$rc = system $command;
	if ( $rc != 0 ) {
	    $self->Print(0, "SpinTestTool::Mount ERROR, $command, rc = " . eval($rc >> 8) . "\n");
	    return undef;
	} else {
	    $self->Print(3, "SpinTestTool::Mount $command, rc = " . eval($rc >> 8) . "\n");
	}
	$self->{MountStack}[$self->{MountStackSize}] = $params{MountPoint};
	$self->{MountStackSize}++;
    }
    $self->Print(3, "SpinTestTool::Mount exited\n");
    return $params{MountPoint};
} # end Mount

#***********************************************************************

=head1 C<SetupTarget()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub SetupTarget {
    my $self = shift;
    my %params = (Host => undef,
		  File => [],
		  Source => ".",
		  Target => "/tmp",
		  @_, # get any parameters that were explicitly passed
		  );

    # Parse the Target to see if it is using an NFS syntax, a CIFS
    # syntax, or a local syntax. If it is a simple path (like /tmp/my/path)
    # then simply FTP the file(s) to that directory on the client (Host).
    # If the Target is NFS or CIFS, then that path will be "mounted"
    # appropriately on the Host, and then the file(s) will be accessible
    # (access/read/write/execute ability must be verified by this method)
    # by the SpinDex server on that remote Host.

    # NFS path - host:/share/dir
    # CIFS (UNC) path - \\host\share\dir
    # Simple path - /simple/dir

    my @fields;
    my $remainder;
    my $host;
    my $share;
    my $targetPath;
    my $rc = 0;
    my $error = 0;
    my $spindex;
    my $file;
    my $fileList;
    my $command;
    my $pid;
    my @path;
    my $pathComp;
    my $createDir;
#    my @dirList;
    my $dir;
    my $rootMount = "/tmp/"; # use /tmp so we can do it as non-root user
#    my $telnet;

    $spindex = SpinDexClient->new(
    # okeefe: bryan says I cannot use the DefaultUser and DefaultPassword
    # when this is an "embedded" spindex object. By "embedded", I mean using
    # the SpinDexClient object within this object, while another SpinDexObject
    # object is (or will be) in use by the test that has created this
    # SpinTestTool object. Thus, there are actually two SpinDexObjects
    # instantiated by the same calling script: one directly instantiated, 
    # and one indirectly instantiated.
#				  DefaultUser => "root",
#				  DefaultPassword => "spinwin",
				  );

    # Create a list of file(s) to transfer. The resulting list is relative
    # to the source host, not the destination.
    foreach $file (@{$params{File}}) {
	$fileList .= "$file ";
	if (! -r "$params{Source}/$file") {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, $params{Source}/$file does not exist\n");
	    $error |= 1;
	}
    }
    goto fail if ($error);
    $self->Print(3, "fileList = $fileList\n");

    # --------------------------------------------------------------
    # Is this an NFS path?
    # --------------------------------------------------------------
    if ( index($params{Target}, ":") >= 0 ) {
	$self->Print(3, "SpinTestTool::SetupTarget $params{Target} is an NFS path\n");
	@fields = split(/:/, $params{Target});
	$host = @fields[0];
	$share = @fields[1];
	$targetPath = "$rootMount$host$share";

	# Create any directories that are needed for this mount.
	@path = split(/\//, $targetPath);
	foreach $pathComp (@path) {
	    if ( $pathComp eq undef ) {
		$createDir = "/";
		next; # jump to the next iteration of the foreach loop
	    }
	    $createDir .= $pathComp . "/";
	    
	    # Determine if a directory already exists on the remote host.
	    # okeefe: need a better way to determine if a dir exists.
#	    $command = "stat $createDir";
	    $command = "perl -e 'exit (-d \"$createDir\")'";
	    $rc = $spindex->Rexec(Host => $params{Host},
				  Command => $command,
				  Wait => 1, # flag, wait for cmd to finish
				  );
	    if ( $rc < 0 ) {
		$self->Print(0, "SpinTestTool::SetupTarget ERROR, Rexec \"$command\", rc = $rc\n");
		goto fail;
	    } else {
		$pid = $rc;
	    }
	    $rc = $spindex->ExitValue(Pid => $pid);
	    $self->Print(3, "SpinTestTool::SetupTarget \"$command\", rc = " . eval($rc >> 8) . "\n");
	    if ( $rc == 0 ) {
		$self->Print(3, "SpinTestTool::SetupTarget directory $createDir does not yet exist\n");
	    } else {
		$self->Print(3, "SpinTestTool::SetupTarget directory $createDir already exits\n");
		next; # goto the next iteration of this loop
	    }

	    # Create the directory.
	    $command = "mkdir -m 0777 $createDir";
	    $rc = $spindex->Rexec(Host => $params{Host},
				  Command => $command,
				  Wait => 1, # flag, wait for cmd to finish
				  );
	    if ( $rc < 0 ) {
		$self->Print(0, "SpinTestTool::SetupTarget ERROR, Rexec \"$command\", rc = $rc\n");
		goto fail;
	    } else {
		$pid = $rc;
	    }
	    $rc = $spindex->ExitValue(Pid => $pid);
	    if ( $rc != 0 ) {
		$self->Print(0, "SpinTestTool::SetupTarget ERROR, \"$command\", rc = " . eval($rc >> 8) . "\n");
		goto fail;
	    } else {
		$self->Print(3, "SpinTestTool::SetupTarget \"$command\", rc = " . eval($rc >> 8) . "\n");
	    }
	} # end foreach

	# Mount the target on the client.
	$command = "mount -o soft,intr $host:$share $targetPath";
	$rc = $spindex->Rexec(Host => $params{Host},
			      Command => $command,
			      Wait => 1,
			      );
	if ( $rc < 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, Rexec \"$command\", rc = $rc\n");
	    goto fail;
	} else {
	    $pid = $rc;
	}
	$rc = $spindex->ExitValue(Pid => $pid);
	if ( $rc != 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, \"$command\", rc = " . eval($rc >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::SetupTarget \"$command\", rc = " . eval($rc >> 8) . "\n");
	}

	# FTP the file(s) to their destination.
	$command = "tar cf - -C $params{Source} $fileList | ncftpput -u systest -p test4spin -c $params{Host} $targetPath/SpinTestTool.tar";
	$rc = system $command;
	if ( $rc != 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, \"$command\" failed, rc = " . eval($rc >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::SetupTarget \"$command\", rc = " . eval($rc >> 8) . "\n");
	}

	# Untar the file(s).
	# I use the "m" flag here to NOT extract the file modified times. Most
	# of the time this does not matter, but in the case where the original
	# file is modified, tared up, and then FTPed to a machine that has a
	# system clock set slightly behind the host where the tar was created,
	# this can cause a tar error when extracting a file that has a mod
	# time that is "in the future".
	$command = "cd $targetPath; tar -x -m --no-same-owner -f SpinTestTool.tar";
	$rc = $spindex->Rexec(Host => $params{Host},
			      Command => $command,
			      Wait => 1,
			      );
	if ( $rc < 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, Rexec \"$command\", rc = $rc\n");
	    goto fail;
	} else {
	    $pid = $rc;
	}
	$rc = $spindex->ExitValue(Pid => $pid);
	if ( $rc != 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, \"$command\", rc = " . eval($rc >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::SetupTarget \"$command\", rc = " . eval($rc >> 8) . "\n");
	}

	# Remove the tar file.
	$command = "rm $targetPath/SpinTestTool.tar";
	$rc = $spindex->Rexec(Host => $params{Host},
			      Command => $command,
			      Wait => 1,
			      );
	if ( $rc < 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, Rexec \"$command\", rc = $rc\n");
	    goto fail;
	} else {
	    $pid = $rc;
	}
	$rc = $spindex->ExitValue(Pid => $pid);
	if ( $rc != 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, \"$command\", rc = " . eval($rc >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::SetupTarget \"$command\", rc = " . eval($rc >> 8) . "\n");
	}

    # --------------------------------------------------------------
    # Is this a CIFS path?
    # --------------------------------------------------------------
    } elsif ( index($params{Target}, "\\") >= 0 ) {
	$self->Print(3, "SpinTestTool::SetupTarget $params{Target} is a CIFS path\n");
	@fields = split(/\\\\/, $params{Target});
	$remainder = $fields[1];
	$self->Print(3, "SpinTestTool::SetupTarget remainder = $remainder\n");
	@fields = split(/\\/, $remainder);
	$host = $fields[0];
	$self->Print(3, "SpinTestTool::SetupTarget host = $host\n");

    # --------------------------------------------------------------
    # This is a simple path.
    # --------------------------------------------------------------
    } else {
	$self->Print(3, "SpinTestTool::SetupTarget $params{Target} is a simple path\n");
	$targetPath = $params{Target};

	# okeefe: need to create any directories in the $targetPath that
	# do not already exist. And need to clean them up too.

	# FTP the file(s) to their destination.
	$command = "tar cf - -C $params{Source} $fileList | ncftpput -u systest -p test4spin -c $params{Host} $targetPath/SpinTestTool.tar";
	$rc = system $command;
	if ( $rc != 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, \"$command\" failed, rc = " . eval($rc >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::SetupTarget \"$command\", rc = " . eval($rc >> 8) . "\n");
	}

	# Untar the file(s).
	# I use the "m" flag here to NOT extract the file modified times. Most
	# of the time this does not matter, but in the case where the original
	# file is modified, tared up, and then FTPed to a machine that has a
	# system clock set slightly behind the host where the tar was created,
	# this can cause a tar error when extracting a file that has a mod
	# time that is "in the future".
	$command = "cd $targetPath; tar -x -m --no-same-owner -f SpinTestTool.tar";
	$rc = $spindex->Rexec(Host => $params{Host},
			      Command => $command,
			      Wait => 1,
			      );
	if ( $rc < 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, Rexec \"$command\", rc = $rc\n");
	    goto fail;
	} else {
	    $pid = $rc;
	}
	$rc = $spindex->ExitValue(Pid => $pid);
	if ( $rc != 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, \"$command\", rc = " . eval($rc >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::SetupTarget \"$command\", rc = " . eval($rc >> 8) . "\n");
	}

	# Remove the tar file.
	$command = "rm $targetPath/SpinTestTool.tar";
	$rc = $spindex->Rexec(Host => $params{Host},
			      Command => $command,
			      Wait => 1,
			      );
	if ( $rc < 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, Rexec \"$command\", rc = $rc\n");
	    goto fail;
	} else {
	    $pid = $rc;
	}
	$rc = $spindex->ExitValue(Pid => $pid);
	if ( $rc != 0 ) {
	    $self->Print(0, "SpinTestTool::SetupTarget ERROR, \"$command\", rc = " . eval($rc >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::SetupTarget \"$command\", rc = " . eval($rc >> 8) . "\n");
	}
    }
    $spindex->DESTROY(); # for some reason, we need to call this explicitly
    return $targetPath;
#    return "$targetPath/$params{Source}";
fail:
    $spindex->DESTROY(); # for some reason, we need to call this explicitly
    return undef;
} # end SetupTarget

#***********************************************************************

=head1 C<Umount()> 

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub Umount {
    my $self = shift;
    my %params = (MountPoint => undef,
		  Rmdir => undef,
		  @_, # get any parameters that were explicitly passed
		  );
    my $i;
    my $command;
    my $rc = undef;
    my $code;

    $self->Print(3, "SpinTestTool::Umount entered\n");
    if ( $params{MountPoint} ne undef ) {
	# Unmountz this mountpoint.
	$command = "umount $params{MountPoint}";
	$code = system $command;
	if ( $code != 0 ) {
	    $self->Print(0, "SpinTestTool::Umount ERROR, $command, rc = " . eval($code >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::Umount $command, rc = " . eval($code >> 8) . "\n");
	}
    } else {
	# Unmount everything that was mounted by this object.
	for ( $i = $self->{MountStackSize}-1; $i >= 0; $i-- ) {
	    $command = "umount $self->{MountStack}[$i]";
	    $code = system $command;
	    if ( $code != 0 ) {
		$self->Print(0, "SpinTestTool::Umount ERROR, $command, rc = " . eval($code >> 8) . "\n");
		goto fail;
	    } else {
		$self->Print(3, "SpinTestTool::Umount $command, rc = " . eval($code >> 8) . "\n");
	    }
	    $self->{MountStackSize}--;
	} # end for
    }
    if ( $params{Rmdir} ne undef ) {
	$command = "rmdir $params{MountPoint}";
	$code = system $command;
	if ( $code != 0 ) {
	    $self->Print(0, "SpinTestTool::Umount ERROR, $command, rc = " . eval($code >> 8) . "\n");
	    goto fail;
	} else {
	    $self->Print(3, "SpinTestTool::Umount $command, rc = " . eval($code >> 8) . "\n");
	}
    }
    $rc = 1;
fail:
    $self->Print(3, "SpinTestTool::Umount exited with $rc\n");
    return $rc;
} # end Umount

#***********************************************************************

=head1 C<ModifyPath()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub ModifyPath {
    my ($self, $path, $type) = @_;
    my $maxFiles = 20;
    my $maxFileSize = 50 * 1024;
    $type = $self->{options}->{dataType} unless ($type);
    our @files;
    our $index; 
    my $loopCount = 1;
    $index = 0 if (!defined($index));

    if ($type =~ /Long/i) {
	$maxFiles = 100;
	$loopCount = 20;
	$maxFileSize = 1024 * 1024;
    }

    for (my $i=0; $i < $loopCount; $i++) {
	$index++; # index will be 1,2.3...
	my $createfile = "$path/file.$index";
	my $size = int rand($maxFileSize);
	if ($#files >= $maxFiles) { # keep a limit on the number of files
	    my $oldfile = shift @files;
	    unlink $oldfile;
	}
	
	my $length = 0;
	my $offset = 0;
	if (!open(OF, ">$createfile")) {
	    $self->Print(1, "Could not create $createfile: $!\n", "bad");
	    if (! -d $path) {
		$self->Print(1, "The directory '$path' does not exist\n", "warn");
	    }
	    if (! -w $path) {
		$self->Print(1, "Cannot write the directory '$path'\n", "warn");
	    }
	    return undef;
	}
	my $sig = "$createfile $size ";
	$length = length($sig);
	while (($offset + $length) <= $size){
	    if (!(print OF $sig)) {
		$self->Print(1, "Could not write to $createfile: $!\n", "bad");
		return undef;
	    }
	    $offset += $length;
	}
	if (!(print OF substr($sig, 0, ($size - $offset)))) {
	    $self->Print(1, "Could not write to $createfile: $!\n", "bad");
	    return undef;
	}
	close(OF);
	push(@files, $createfile);
    }
    return 1;
}

#***********************************************************************

=head1 C<CompareReleases()>

=over 4

=item Description

Given two releases and an optional comparison routine,
determine if one release is >= to another. IE:
ReleaseA >= ReleaseB

=item Arguments

=over 4

=item ReleaseA

=item ReleaseB

=item Comparison

Accepted values are ">=" and "<". Default is ">="

=back

=item Returns

true when comparison is true
undef otherwise

=back

=cut

#***********************************************************************
sub CompareRelease {
    my $self = shift;
    my %params = (
	    ReleaseA    => undef,
	    ReleaseB	=> undef,
	    Comparison	=> ">=",
	    @_,
	);
    my $name = "SpinTestTool::CompareRelease:";

    # make sure we have 2 releases to check
    if ( !$params{ReleaseA} || !$params{ReleaseB} ) {
	$self->Print(0, "$name requires ReleaseA and ReleaseB.\n", "bad");
	return undef;
    } else {
	$self->Print(3, "$name comparing release \"$params{ReleaseA}\" $params{Comparison} \"$params{ReleaseB}\"\n");
    }

    # break up the version numbers into an array
    my @ReleaseA = split(/\./, $params{ReleaseA});
    my @ReleaseB = split(/\./, $params{ReleaseB});
    # any version number that was empty needs to be zero filled in the array
    while ($#ReleaseA gt $#ReleaseB) {
	push(@ReleaseB, 0);
    }
    while ($#ReleaseB gt $#ReleaseA) {
	push(@ReleaseA, 0);
    }
    # now walk through the arrays and do the comparison
    for (my $index=0; $index <= $#ReleaseA; $index++) {
	if ($params{Comparison} eq ">=") { 
	    if ($ReleaseA[$index] gt $ReleaseB[$index]) {
		return 1;
	    } elsif ($ReleaseA[$index] lt $ReleaseB[$index]) {
		return undef;
	    }
	} elsif ($params{Comparison} eq "<") {
	    if ($ReleaseA[$index] lt $ReleaseB[$index]) {
		return 1;
	    } elsif ($ReleaseA[$index] gt $ReleaseB[$index]) {
		return undef;
	    }
	} else {
	    $self->Print(0, "$name cannot handle comparison \"$params{Comparison}\".\n", "bad");
	    return undef;
	}
    }
    # if we are here, the releases must be equal
    if ( $params{Comparison} eq ">=" ) {
	return 1;
    } else {
	return undef;
    }
} # CompareRelease
#***********************************************************************

=head1 C<HandleInitFile()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub HandleInitFile {
    my $self = shift;
    my $optionsRef = $self->{options};
    # Time to check for SpinServer parameters
    our %Servers = ();

    if ($$optionsRef{initFile}) {
	if (! -r $$optionsRef{initFile}) {
	    $self->Print(1, "Could not open $$optionsRef{initFile}\n");
	    exit(1);
	} else {
	    eval { require ($$optionsRef{initFile}); };
	    if ($@) {
		warn "Error loading $$optionsRef{initFile}: $@";
	    }
	}
    }
    $$optionsRef{Configs} = \%Servers;
}
#***********************************************************************

=head1 C<PrintStructure()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub PrintStructure {
    my $self = shift;
    my $struct = shift;
    my $name = shift;
    
    $Data::Dumper::Varname = $name if ($name);
    print Dumper($struct);
    return;
}
#***********************************************************************

=head1 C<ANSItoHTML()>

=over 4

=item Description

Convert any ANSI color strings to HTML <font color=XXX> codes as well as newlines etc.

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub ANSItoHTML {
    my $self = shift;
    my $message = shift;

    # BASE colors
    $message =~ s!\[30m(.+)\[0m!<font color=black>$1</font>!g;
    $message =~ s!\[31m(.+)\[0m!<font color=red>$1</font>!g;
    $message =~ s!\[32m(.+)\[0m!<font color=green>$1</font>!g;
    $message =~ s!\[33m(.+)\[0m!<font color=yellow>$1</font>!g;
    $message =~ s!\[34m(.+)\[0m!<font color=blue>$1</font>!g;
    $message =~ s!\[35m(.+)\[0m!<font color=purple>$1</font>!g; # magenta
    $message =~ s!\[36m(.+)\[0m!<font color=cyan>$1</font>!g;
    $message =~ s!\[37m(.+)\[0m!<font color=silver>$1</font>!g; # white
    # BOLD Colors
    $message =~ s!\[1;30m(.+)\[0m!<bold><font color=black>$1</font></bold>!g;
    $message =~ s!\[1;31m(.+)\[0m!<bold><font color=red>$1</font></bold>!g;
    $message =~ s!\[1;32m(.+)\[0m!<bold><font color=green>$1</font></bold>!g;
    $message =~ s!\[1;33m(.+)\[0m!<bold><font color=yellow>$1</font></bold>!g;
    $message =~ s!\[1;34m(.+)\[0m!<bold><font color=blue>$1</font></bold>!g;
    $message =~ s!\[1;35m(.+)\[0m!<bold><font color=purple>$1</font></bold>!g; # magenta
    $message =~ s!\[1;36m(.+)\[0m!<bold><font color=cyan>$1</font></bold>!g;
    $message =~ s!\[1;37m(.+)\[0m!<bold><font color=white>$1</font></bold>!g; # white
    # Special Colors (IE colors with a background set)
    $message =~ s!\[34;46m(.+)!<font color=blue face=\"comic sans ms\">$1</font>!g;
    $message =~ s!\[1;34;46m(.+)!<bold><font color=blue face=\"comic sans ms\">$1</font></bold>!g;
    $message =~ s!\[34;47m(.+)!<font color=blue face=\"comic sans ms\">$1</font>!g;
    $message =~ s!\[1;34;47m(.+)!<bold><font color=blue face=\"comic sans ms\">$1</font></bold>!g;
    # End ANSI
    $message =~ s!\[0m!!g;

    # Other Terminal to HTML conversions
    # if the message contains a pointer to a file, turn it into an HTML link
    $message =~ s!(/x/eng/systest/[\S+/]*)!<a href="http://zoo.hq.netapp.com/cgi-bin/readSTOutput.cgi?File=$1">$1</a>!g;
    # Newline
    $message =~ s!\n!<br>\n!g;

    return $message;
}
#***********************************************************************

=head1 C<RunCommand()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub RunCommand {
    my $self = shift;
    my $command = shift;
    my $outFile = shift;
    my $oldfh;
    $self->Print(2, "$command\n");

    if ($outFile && $self->{Verbosity} <= 1) {
	$command .= " >> $outFile 2>&1";
    }
    $self->Print(2, "Running $command\n");
    my $rc = system $command;
    my $exit_value = $? >> 8;
    my $signal_num = $? & 127;
    my $dumped_core = $? & 128;
    if ( $rc != 0 ) {
	$self->Print(0, "RunCommand: $command exit_value=$exit_value signal=$signal_num dumped_core=$dumped_core\n");
	$self->Print(0, "ERROR, $command, rc = " . eval($rc>>8) . "\n", "bad");
	return undef;
    }
    return 1;
} # end RunCommand
#***********************************************************************

=head1 C<CheckClient()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub CheckClient {
    my ($output, $client) = @_;
    if ($output =~ /load average: (\d+\.*\d+), (\d+\.*\d+),/) {
	my $uptime = $2;
	if ($uptime > $stt->{options}{maxClientLoad}) {
	    $stt->Print(1, "Skipping client '$client' because the load ($uptime) is higher than the defined max ($stt->{options}{maxClientLoad})\n");
	    delete $stt->{testOptions}{$client};
	}
    }
}
#***********************************************************************

=head1 C<HelpFromPod()>

=over 4

=item Description

Obtain the on-screen help display by scanning POD tags. This subroutine 
displays the usage string and parameter description

=item Arguments

=over 4

=item Filename

The name of the file (usually .pl) to print help on

=back

=item Returns

None

=back

=cut

#***********************************************************************
sub HelpFromPod {
    my $self = shift;
    my $synopsis = ""; 
    my $parameters = ""; 
    my $iterator = 1;
    my %params = (Filename => undef,
                  @_, # get any parameters that were explicitly passed
                  );

    $stt->Print(3, "SpinTestTool::HelpFromPod entered\n");
    if (!defined($params{Filename})){
	$stt->Print(1, "SpinTestTool::HelpFromPod ERROR: Must supply a testname\n","bad");
	goto done;
    }
    my @output = `pod2text $params{Filename}`;
    for (my $i=0; $i<=$#output; $i++) {
        next if ($output[$i] =~ /^\s*$/);  ## skip blank lines
        chomp($output[$i]);
        if ($output[$i] =~ /SYNOPSIS/) {
 	    my $line = $output[$i+1];
            while ( $line !~ /PARAMETERS/) {
     	        chomp ($line);
	        $line =~ s/^\s+/ /;
     	        $synopsis .= $line;
     	        $iterator++;
     	        $line = $output[$i+$iterator];
 	    }
            print "usage:\n $synopsis\n";
            $i = $i + $iterator - 1;
        }
        $iterator = 1;
        if ($output[$i] =~ /PARAMETERS/) {
	    my $line = $output[$i+1];
	    while ( $line !~ /EXAMPLES/) {
	        $iterator++,$line = $output[$i+$iterator],next if ($line =~ /This program takes the following arguments/);
	        $iterator++,$line = $output[$i+$iterator],next if ($line =~ /\.html\#parameters/); ## This line is handled later
	        ##############################################################
	        # If the .pl script uses STT.pm parameters, run pod2text
	        # on STT.pm and parse all of the Parameters
	        ##############################################################
	        if ($line =~ /All standard parameters from SpinTestTool\.pm/) {
	            my @STT_output = `pod2text /x/eng/systest/sb/export/perl/SpinTestTool.pm`;
		    my $STT_iterator=0;
	            foreach (my $j=0; $j<=$#STT_output; $j++) {
		        next if ($STT_output[$j] =~ /^\s*$/);  ## skip blank lines
    		        chomp($STT_output[$j]);
    		        if ($STT_output[$j] =~ /GetOptions()/) {
        		    my $STT_line = $STT_output[$j+1];
			    my $iter = $j;
			    while ( $STT_line !~ /Arguments/) {
				$iter++;
				$STT_line = $STT_output[$iter];
			    };
			    print "\n";
        		    while ( $STT_line !~ /CIFS Test Arguments/) {	
			        if ($STT_line !~ /http|html/) {
			            $parameters .= $STT_line;
			        }
	                        $STT_iterator++;
           		        $STT_line = $STT_output[$iter+$STT_iterator];
			    }
			    last;
        	        }
		    }
	        } elsif ($line =~ /\.html\#item_(.*)/) {
		    my $param = $1;
		    my @STT_output = `pod2text /x/eng/systest/sb/export/perl/SpinTestTool.pm`;
                    foreach (my $j=0; $j<=$#STT_output; $j++) {
                        if ($STT_output[$j] =~ /\-($param)/) {
        		    my $STT_line = $STT_output[$j];
			    my $STT_iterator=0;
			    while ($STT_line !~ /Description/) {
                                $parameters .= $STT_line;
                                $STT_iterator++;
                                $STT_line = $STT_output[$j+$STT_iterator];
                            }
			    while ($STT_line !~ /^\s+\-/) {	
			        $parameters .= $STT_line;
			        $STT_iterator++;
                                $STT_line = $STT_output[$j+$STT_iterator];
			    }
			    last;
                        }
                    }
		} elsif ($line =~ /http/) {
		    $stt->Print(3, "SpinTestTool::HelpFromPod Skipping HTTP line since it is assumed that .html part will be on separate line\n","warn");
	        } else {
	            $parameters .= $line;
	        }
	        $iterator++;
	        $line = $output[$i+$iterator];
	    }
	    print "Parameters:\n $parameters\n";
        }
    }

  done:
    $stt->Print(3, "SpinTestTool::HelpFromPod exited\n");
}

#***********************************************************************

=head1 C<UsageFromPod()>

=over 4

=item Description

Obtain the on-screen help display by scanning POD tags. This subroutine 
displays only the usage string

=item Arguments

=over 4

=item Filename

The name of the file (usually .pl) to print help on

=back

=item Returns

None

=back

=cut

#***********************************************************************
sub UsageFromPod {
    my $self = shift;
    my $synopsis = ""; 
    my $parameters = ""; 
    my $iterator = 1;
    my %params = (Filename => undef,
                  @_, # get any parameters that were explicitly passed
                  );

    $stt->Print(3, "SpinTestTool::UsageFromPod entered\n");
    if (!defined($params{Filename})){
	$stt->Print(1, "SpinTestTool::UsageFromPod ERROR: Must supply a testname\n","bad");
	goto done;
    }
    my @output = `pod2text $params{Filename}`;
    for (my $i=0; $i<=$#output; $i++) {
        next if ($output[$i] =~ /^\s*$/);  ## skip blank lines
        chomp($output[$i]);
        if ($output[$i] =~ /SYNOPSIS/) {
 	    my $line = $output[$i+1];
            while ( $line !~ /PARAMETERS/) {
     	        chomp ($line);
	        $line =~ s/^\s+/ /;
     	        $synopsis .= $line;
     	        $iterator++;
     	        $line = $output[$i+$iterator];
 	    }
            print "usage:\n $synopsis\n";
            $i = $i + $iterator - 1;
        }
    }
  done:
    $stt->Print(3, "SpinTestTool::UsageFromPod exited\n");
}
#***********************************************************************

=head1 C<MethodFromPod()>

=over 4

=item Description

Obtain the on-screen help display for a particular method by scanning 
POD tags for the specified file, and displaying help for specified method. 
The contents of Filename are scanned first and then the contents of any
.pm modules are scanned. Currently, only Filename and STT.pm are scanned.

=item Arguments

=over 4

=item Filename

The name of the file (usually .pl) to print help on

=item Method 

The name of the method to print help on. Case insensitive.

=back

=item Returns

None

=back

=cut

#***********************************************************************
sub MethodFromPod {
    my $self = shift;
    my $info = "";
    my $iterator = 1;
    my $found = 0;
    my @systest_files = ("/x/eng/systest/sb/export/perl/SpinTestTool.pm", "/x/eng/systest/sb/export/perl/Filer.pm", "/x/eng/systest/sb/export/perl/SpinServer.pm", "/x/eng/spin-cores/documentation/SpinServer.podtext","/x/eng/spin-cores/documentation/Filer.podtext");
    my %params = (Filename => undef,
		  Method => undef,
                  @_, # get any parameters that were explicitly passed
                  );

    $stt->Print(3, "SpinTestTool::MethodFromPod entered\n");
    if (!defined($params{Filename})){
        $stt->Print(1, "SpinTestTool::MethodFromPod ERROR: Must supply a testname\n","bad");
        goto done;
    }
    if (!defined($params{Method})){
        $stt->Print(1, "SpinTestTool::MethodFromPod ERROR: Must supply a method\n","bad");
        goto done;
    }
    #***********************************************************************
    # Look at local file and each file in the @systest_files array for the method
    # Order of preference:
    #    1) Calling .pl file
    #    2) SpinTestTool.pm
    #    3) Filer.pm
    #    4) SpinServer/*.pm
    #    5) Filer/*.pm 
    #***********************************************************************
    foreach my $file ($params{Filename}, @systest_files) {
        my @output = `pod2text $file`;
        for (my $i=0; $i<=$#output; $i++) {
            next if ($output[$i] =~ /^\s*$/);  ## skip blank lines
            if ($output[$i] =~ /$params{Method}\(\)/i) {
	        $found = 1;
	        $info .= $output[$i];	
                my $line = $output[$i+1];
                while ( $line !~ /^\".+\(\)/) {
                    $info .= $line;
                    $iterator++;
                    $line = $output[$i+$iterator];
                }
                print "$info\n";
            }
	    last if ($found);
        }
	last if ($found);
    }
    if (!$found) {
        $stt->Print(1, "SpinTestTool::MethodFromPod ERROR: Method $params{Method} was not found\n","bad");
        goto done;
    }
  done:
    $stt->Print(3, "SpinTestTool::MethodFromPod exited\n");
}
#***********************************************************************

=head1 C<FileTest()>

=over 4

=item Description

Given a file and a list of tests to perform on the file, return true if the file passes all the tests,
and undef if any of the tests fail.
If not give a Tests section, defaults to make sure the file exists

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub FileTest {
    my $self = shift;
    my %params = (
	File	=> undef,
	Tests	=> "e",
	@_,
    );
    my $name = "SpinTestTool::FileTest";
    my $failed = undef;

    # we need to have the file defined
    if ( !defined $params{File} ) {
	$self->Print(1, "$name called without a file name.\n", "bad");
	return undef;
    }
    
    # go through the list of operators that we were passed and test them all.  if someone passed in
    # "erw" and the file doesn't exists then they should expect to get back a failure and see that 
    foreach my $operator ( split(//, $params{Tests}) ) {
	# I wanted to make this generic but couldn't quite get it
	# if (!eval("if(-$operator($$optionsRef{$option})) {return 1; } else { return 0;}")) {
	if (!grep($operator eq $_, qw(e d f r w x z s l S p b c u g k T B))) {
	    $self->Print(2, "$name cannot handle file test operator '$operator'\n", "bad");
	    $failed .= $operator;
	}
	if ($operator eq 'e' && ! -e $params{File}) {
	    $self->Print(2, "$name $params{File} does not exist\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'd' && ! -d $params{File}) {
	    $self->Print(2, "$name $params{File} is not a directory\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'f' && ! -f $params{File}) {
	    $self->Print(2, "$name $params{File} is not a file\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'r' && ! -r $params{File}) {
	    $self->Print(2, "$name $params{File} is not readable\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'w' && ! -w $params{File}) {
	    $self->Print(2, "$name $params{File} is not writable\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'x' && ! -x $params{File}) {
	    $self->Print(2, "$name $params{File} is not executable\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'z' && ! -z $params{File}) {
	    $self->Print(2, "$name $params{File} does not exist or is not zero length\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 's' && ! -s $params{File}) {
	    $self->Print(2, "$name $params{File} does not exist or is zero length\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'l' && ! -l $params{File}) {
	    $self->Print(2, "$name $params{File} is not a symbolic link \n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'S' && ! -S $params{File}) {
	    $self->Print(2, "$name $params{File} is not a socket\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'p' && ! -p $params{File}) {
	    $self->Print(2, "$name $params{File} is not a name pipe\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'b' && ! -b $params{File}) {
	    $self->Print(2, "$name $params{File} is not a block special file\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'c' && ! -c $params{File}) {
	    $self->Print(2, "$name $params{File} is not a character special file\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'u' && ! -u $params{File}) {
	    $self->Print(2, "$name $params{File} is not setuid \n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'g' && ! -g $params{File}) {
	    $self->Print(2, "$name $params{File} is not setgid\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'k' && ! -k $params{File}) {
	    $self->Print(2, "$name $params{File} is not sticky\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'T' && ! -T $params{File}) {
	    $self->Print(2, "$name $params{File} is not a text file\n", "bad");
	    $failed .= $operator;
	}	    
	if ($operator eq 'B' && ! -B $params{File}) {
	    $self->Print(2, "$name $params{File} is not a binary file\n", "bad");
	    $failed .= $operator;
	}	    
    }

    # return true if we didn't detect failure
    $failed ? return undef : return 1;
}
 
#***********************************************************************

=head1 C<_RunST()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _RunTest {
    my ($self) = @_;
    my @pidList;
    my $rc = 0;
    my $command = $self->{TestCommand};
    my $spindex = $self->{SpinDex};
    my $timeout = $self->{TestTimeout};
    my $fail = undef;

    if (defined($self->{options}{duration})) {
	my $duration = $self->{options}{duration};
	if ( $duration =~ /[Hh]/ ) { # if value is hours
	    chop($duration);
	    $timeout = $duration * 3600;
	} elsif ( $duration =~ /[Mm]/ ) { # if value is minutes
	    chop($duration);
	    $timeout = $duration * 60;
	} elsif ( $duration =~ /[Ss]/ ) { # if value is seconds
	    chop($duration);
	    $timeout = $duration;
	} else {
	    $timeout = $duration * 60; # convert to seconds
	}
    } else {
	my $iterations = (defined($self->{options}{iterations})) ? $self->{options}{iterations} : 1;
	$timeout *= $iterations;
    }

    $self->Print(3, "SpinTestTool::_RunTest entered\n");
    
    $self->StartTestResults();

    # still need to fix this for NATE
    if (defined($self->{options}{maxClientLoad})) {
	$spindex->{PrintFormat} = undef;
	foreach my $client (keys %{$self->{testOptions}}) {
	    if (($spindex->Rexec(Host => $client, Command => "uptime", EonOut => \&CheckClient)) < 0){
		$self->Print(1, "Skipping client '$client' because it cannot be reached\n");
		delete $self->{testOptions}{$client};
	    }
	}
	$spindex->WaitGroup();
	delete $spindex->{PrintFormat};
    }
    my @subtests;
    my $eonout = 0; # the normal case for SpinDex
    my $runseq = "01"; # for nate
    foreach my $client (keys %{$self->{testOptions}}) {
	foreach my $instance (keys %{$self->{testOptions}{$client}}) {	
	    my $hashref = $self->{testOptions}{$client}{$instance};
	    if ($self->{NATE}) {
		my @options = $self->_CreateTestOptions(%$hashref);
		my $options = $self->_CreateTestOptions(%$hashref);
		my $runid = "$self->{Program}_${client}_${instance}";
		$self->Print(3, "SpinTestTool::_RunTest Running \'$command -runid $runid -bg -client $client $options\' on client $client\n");
		push(@subtests, main::subtest($command, '-runid', $runid, '-bg', '-client', $client, '--', @options)); # -iffatal?
	    } else {
		my $options = $self->_CreateTestOptions(%$hashref);
		my $cmd = $command . $options;
		if ($self->{NATE}) {
		    my $nateflags = "-noconsole";
		    $nateflags .= " NATE_LIB='$ENV{NATE_LIB}'" if ($ENV{NATE_LIB});
		    if ($ENV{LOGDIR}) {
			$nateflags .= " LOGDIR=$ENV{LOGDIR}";
			$nateflags .= " JOBID=${client}_${instance}_test_phase";
			$eonout = undef;
		    } else {
			$nateflags .= " LOGDIR=-";
		    }
		    $cmd = "/usr/local/test/share/nate/latest/bin/ntest $nateflags \"$cmd\"";
		}
		$rc = $spindex->Rexec(Host => $client,
				      Command => $cmd,
				      EonOut => $eonout,
				      );
		if ( $rc < 0 ) {
		    $self->Print(1, "SpinTestTool::_RunTest ERROR kicking off \"$cmd\" on client $client, rc = $rc\n");
		} else {
		    $self->Print(1, "SpinTestTool::_RunTest Kicking off command on client $client (pid = $rc)\n");
		    $self->Print(2, "SpinTestTool::_RunTest Running \'$cmd\' on client $client (internal pid = $rc)\n");
		    push(@pidList, $rc); # push the pid onto our list
		}
	    }
	}
    }
    $self->StopTestResults();

    if ($self->{NATE}) {# running inside of nate so get each subtest exit status
	main::Subtest::wait_finish(subtests=>\@subtests,
			   timeout=>$timeout);
	  # now get the return codes
	foreach my $t (@subtests) {
	    $rc = $t->exit_status();
	    if ( !defined($rc) || $rc ) {
		my $pid = $t->pid();
		$self->Print(1, "SpinTestTool::_RunTest ERROR, process (internal pid = $pid) got rc = " . eval($rc >> 8) . "\n");
		$fail++;
	    }
	}
	  goto fail if ( $fail != 0 );
      } else {
	$self->Print(1, "Waiting for all clients to finish\n");
	$rc = $spindex->WaitGroup(#WaitGroupName => "mywaitgroup",
				  Timeout => $timeout,
				  );
	if ( $rc < 0 ) {
	    $self->Print(1, "ERROR waiting for all clients to finish after $timeout seconds, rc = $rc\n");
	    $spindex->Rkill(Pid => "all");
	    goto fail;
	} else {
	    $self->Print(1, "All clients have finished\n");
	}
        # Check the return codes (exit values) from the actual remote tests.
	foreach my $pid (@pidList) {
	    $rc = $spindex->ExitValue(Pid => $pid);
	    if ( !defined($rc) || $rc ) {
		$self->Print(1, "ERROR, process (internal pid = $pid) got rc = " . eval($rc >> 8) . "\n", "bad");
		$fail++;
	    }
	}
	
	unlink("$ENV{LOGDIR}/latest") if ($self->{NATE} && $ENV{LOGDIR});

        # If any of the remote processes finished with non-zero return codes, fail.
	if ( $fail || $spindex->SeenError() != 0 ) {
	    goto fail;
	}
    }
    
    $self->Print(3, "SpinTestTool::_RunTest exited with 1\n");
    return 1;
fail:
    $self->Print(3, "SpinTestTool::_RunTest exited with 0\n");
    return undef;
}

#***********************************************************************

=head1 C<_CreateTestHash()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _CreateTestHash {
    my $self = shift;
    my $additionalOptions = shift;
    my $size;
    my %testOptions;
    my %options = %{$self->{options}};
    if ($options{testOptions}) {
	$testOptions{testOptions} = "";
	foreach my $testoption (@{$options{testOptions}}) {
	    $testOptions{testOptions} .= " $testoption";
	}
    }
    $testOptions{testOptions} .= "$additionalOptions" if ($additionalOptions);
    foreach my $option qw(iterations StopOnErrors quick seed duration nfsversion protocol release source
			  benchmark verify FindAllServers cifsserver cifsdomain cifsshare cifsuser cifspassword
			  dataType verbosity VirtualServer share StopOnErrors) {
	$testOptions{$option} = $options{$option} if ($options{$option});
    }
    
    if ($options{testdir} && $options{mountpoint}) {
	$testOptions{testdir} =  "$options{mountpoint}/$options{testdir}";
    }

    if ($options{nbladeIP} && ($size = scalar(@{$options{nbladeIP}}))) {
	$self->{nbladeIP} = $options{nbladeIP}[rand($size)];
	$testOptions{nbladeIP} =  $self->{nbladeIP};
    }
    
    if ($options{dbladeIP} && ($size = scalar(@{$options{dbladeIP}}))) {
	$self->{dbladeIP} = $options{dbladeIP}[rand($size)];
	$testOptions{dbladeIP} = $self->{dbladeIP};
    }
    
    if ($options{serverIP} && ($size = scalar(@{$options{serverIP}}))) {
	$self->{serverIP} = $options{serverIP}[rand($size)];
	$testOptions{serverIP} = $self->{serverIP};
    }

    if ($self->{SPNames}) {
	$testOptions{StoragePool} = $self->{SPNames};
    }
    
    $testOptions{server} = join(",", @ {$options{server}});
    $testOptions{vfs} = join(",", @ {$options{vfs}}) if (scalar(@{$options{vfs}}));
    if ($self->{BuildVersion}) {
	$testOptions{testOptions} .= " -BuildVersion $self->{BuildVersion}";
    }
    return %testOptions;
}

#***********************************************************************

=head1 C<_CreateTestOptions()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _CreateTestOptions {
    my $self = shift;
    my %testOptions = @_;
    my $testOptions;
    my @testOptions;

    foreach my $option (keys %testOptions) {
	if ($testOptions{$option} =~ /\s+/) {
	    $testOptions .= " -$option '$testOptions{$option}'";
	    push(@testOptions, "-$option", "'$testOptions{$option}'");
	}
	if (grep($option eq $_, qw(benchmark StopOnErrors))) {
	    $testOptions .= " -$option";
	    push(@testOptions, "-$option");
	} elsif ($option eq "testOptions") {
	    $testOptions .= " $testOptions{$option}";
	    push(@testOptions, "$testOptions{$option}");
	} else {
	    $testOptions .= " -$option $testOptions{$option}";
	    push(@testOptions, "-$option", $testOptions{$option});
	}
    }
    if ($self->{STmode}) {
	$testOptions .= " -STmode 0";
	push(@testOptions, "-STmode", 0);
	if (!$ENV{LOGDIR} || $ENV{LOGDIR} eq "") {
	    $testOptions .= " -noheader";
	    push(@testOptions, "-noheader");
	}
    }
	
    return wantarray ? @testOptions : $testOptions;
}

#***********************************************************************

=head1 C<_SetupServers()>

=over 4

=item Description

This is the default method for the SetupServers phase of
RunST. It will basically pass the %Config hash to Configure

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _SetupServers {
    my $self = shift;
    my $rc = undef;

    $self->Print(3, "SpinTestTool::_SetupServers entered\n");
    my $spinserver = $self->{ServerObject};
    my $server => $spinserver->{Server};

    $self->{options}{skipNode} = ["Cluster","NTP"];	
#    foreach my $server ($spinserver->Servers()) {
#	$spinserver->{Configuration}{$server}{SpinServer}{ID} = $server;
#    }
					    
    # if the objective is CIFS, we need to make sure that the CIFS Domain, CIFS Server, and CIFS
    # Share are defined for the test to use
    if ($self->{testObjective} eq "CIFS") {
	if (!$self->{options}{cifsserver}) { # we should have this already
	    $self->Print(1, "SpinTestTool::_SetupServers should have never got here.  Found a Virtual Server that claimed to have CIFS enabled, but could not resolve the Netbios name.\n", "bad");
	    goto done;
	}
    }

    goto done if (!$spinserver->Configure(PrintNode=>undef));

    $rc = 1;

  done:
    $self->Print(3, "SpinTestTool::_SetupServers exited with $rc\n");
    return $rc;
} #_SetupServers

#***********************************************************************

=head1 C<_SetupClients()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _SetupClients {
    my $self = shift;
    $self->Print(3, "SpinTestTool::_SetupClients entered\n");
    my $rc = undef;
    my $options = $self->{options};

    if (!$options->{Mounts}) {
	foreach my $client (keys %{$self->{MountPoint}}) {
	    foreach my $instance (keys %{$self->{MountPoint}{$client}}) {
		my $mountpoint = $self->{MountPoint}{$client}{$instance};
		my %mount = ();
		# here we need to decide if we're CIFS or NFS and build the mount hash up apropriately
		if ($self->{testObjective} eq "CIFS") {
		    # build the mount options up for a CIFS Cygwin style mount
		    # we don't add the full path to the individual VFS onto the share because that isn't supported in
		    # Windows versions older than W2K
		    $mount{server} = $options->{cifsserver};
		    $mount{share} = $options->{cifsshare};
		    #$mount{mounttype} = "win32";
		    $mount{cygmountoptions} = "-x -b -s";
		    $mount{win32mountoptions} = "/PERSISTENT:NO";
		} else {
		    $mount{server} = $self->{serverIP}{$client}{$instance};
		    $mount{share} = $self->{share}{$client}{$instance};
		    $mount{nfsmountoptions} = "hard,intr";
		    $mount{nfsmountoptions} .= ",nfsvers=$options->{nfsversion}" if ($options->{nfsversion});
		    $mount{nfsmountoptions} .= ",$options->{protocol}" if ($options->{protocol});
		}
		if (!$mount{share} || !$mount{server}) {
		    $self->Print(1, "SpinServer::_SetupClients Did not find a share($mount{share}) or server($mount{server})\n", "bad");
		    goto done;
		}
		
		# XXX mountpoint is fine like it is for Cygwin, but for win32 it needs to be a drive letter
		$mount{mountpoint} = $mountpoint;
		if (!$mount{mountpoint}) {
		    $self->Print(1, "SpinServer::_SetupClients Did not find a mountpoint for $client\n", "bad");
		    goto done;
		}
		push(@{$options->{Mounts}{$client}}, \%mount);
	    }
	}
    } else {
	my @NBladeIPs = @{$options->{nbladeIP}};
	foreach my $server ( keys %{$options->{Mounts}}) {
	    foreach my $mount (@{$options->{Mounts}{$server}}) {
		if (!$mount->{server}) {
		    $mount->{server} = $NBladeIPs[int rand(scalar @NBladeIPs)];
		}
	    }
	}	
    }

    if (!$options->{Mounts}) {
	# nothing to do
	$rc = 1;
    } else {
	$rc = $self->ClientMounts($options);
    }

done:    
    $self->Print(3, "SpinTestTool::_SetupClients exited\n");
    # perform any specific client setup here
    return $rc;
} # _SetupClients

#***********************************************************************

=head1 C<_CleanupClients()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _CleanupClients {
    my $self = shift;

    # perform any specific client cleanup here
    $self->Print(3, "SpinTestTool::_CleanupClients entered\n");

    my $rc = $self->ClientUmounts();

    $self->Print(3, "SpinTestTool::_CleanupClients exited\n");

    return $rc;
} # _CleanupClients

#***********************************************************************

=head1 C<_CleanupServers()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _CleanupServers {
    my $self = shift;

    my $rc = 1;
    # perform any specific server cleanup here
    $self->Print(3, "SpinTestTool::_CleanupServers entered\n");

    my $rc = $self->{ServerObject}->Unconfigure();

    $self->Print(3, "SpinTestTool::_CleanupServers exited with $rc\n");
    return $rc;
} # _CleanupServers

#***********************************************************************

=head1 C<_Cleanup()>

=over 4

=item Description

This is a method

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _Cleanup {
    my $self = shift;
    $self->Print(3, "SpinTestTool::_Cleanup entered.\n");
    my $spindex = $self->{SpinDex};
    $spindex->DESTROY() if ($spindex); # for some reason, we need to call this explicitly
    $self->Print(3, "SpinTestTool::_Cleanup exited.\n");
} # end Cleanup

#***********************************************************************

=head1 C<_SetVerbosity()>

=over 4

=item Description

internal function to set the Verbosity in the object

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _SetVerbosity {
    my $self = shift;
    my %params = (Verbosity => 1,
		  @_, # get any parameters that were explicitly passed
		  );
    $print->SetVerbosity($params{Verbosity});
    return 1;
} # end _SetVerbosity

#***********************************************************************

=head1 C<_SetLogVerbosity()>

=over 4

=item Description

internal function to set the log verbosity

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _SetLogVerbosity {
    my $self = shift;
    my %params = (LogVerbosity => 3,
		  @_, # get any parameters that were explicitly passed
		  );
    $print->SetLogVerbosity(log_verbosity=>$params{LogVerbosity});
    return 1;
} # end _SetLogVerbosity


#***********************************************************************

=head1 C<_aim_error()>

=over 4

=item Description

Callback function for AOL Instant Messanger as part of the Notify()

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _aim_error {
    my($oscar, $connection, $errno, $error, $fatal) = @_;
    if($fatal) {
	    die "Fatal error $errno in ".$connection->{description}.": $error\n";
    } else {
	    print STDERR "Error $errno: $error\n";
    }

    return 0;
}
#***********************************************************************

=head1 C<_aim_admin_error()>

=over 4

=item Description

Callback function for AOL Instant Messanger as part of the Notify()

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub _aim_admin_error($$$$) {
	my($oscar, $reqtype, $error, $errurl) = @_;

	print STDERR "Your $reqtype request was unsuccessful (", 0+$error, "): $error.";
	print STDERR "  See $errurl for more info." if $errurl;
	print STDERR "\n";
}
#***********************************************************************

=head1 C<_aim_admin_ok()>

=over 4

=item Description

Callback function for AOL Instant Messanger as part of the Notify()

=item Arguments

=over 4

=item List first argument

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************

sub _aim_admin_ok($$) {
	my($oscar, $reqtype) = @_;

	print "Your $reqtype request was successful.\n";
}

#***********************************************************************


# all good modules return true
1;
