#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/Statistics.pm#1

#***********************************************************************
#
# NAME
#   Statistics.pm - 
#
# SYNOPSIS
#   General helper package for measuring statistics on the filer.
#
# DESCRIPTION
#   This library provides helper functions to measure statisctics.
#
#
# AUTHOR
#   dl-cluster-qa
#
# HISTORY
# 7/13/2009 Initial template
# 4/05/2011 Initial function added
#***********************************************************************

## @summary     Statistics.pm

package Statistics;

use strict;

use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::C::Node;
use Tharn qw(host snooze subtest $Parent $Log fatal_result param);
use Time::HiRes qw(gettimeofday tv_interval);

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use File::Basename;
use Spreadsheet::WriteExcel;
use Spreadsheet::WriteExcel::Utility;
use Spreadsheet::WriteExcel::Chart;

my %spreadsheet;
my @vm_row = (0);
my @top_row = (0);
my @swap_row = (0);
my @command_row = (0);
my @command_row_rows;
my @connections;

# new method - Create a new helper object
#  Parameter hash (* means required):
#   *nodes              - NACL::C::Node array of where to run statistics
#   write_xls           - Expected string of "true" or undef to write and xls file
#   filename            - Overridden filenam, if undef one will be created
#
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my %options = (
        write_xls=>undef,
        filename=>undef,
        nodes=>{ type => ARRAYREF, default =>undef},
        @_
    );

    my @subArgs;
    if (!$options{nodes}) {
        $Log->error("Nodes required for Statistics\n");
        return 0;
    }

    if($options{write_xls} && $options{write_xls} eq "true") {
        create_workbook(filename=>$options{filename}, nodes=>$options{nodes});
    }

    my $self = {
        "nodes"     => $options{nodes},
        "write_xls" => $options{write_xls},
        "filename"  => $options{filename},
        "spreadsheet" => %spreadsheet,
        @_,
    };

    bless($self, $class);
    return $self;
}

# close_workbook method - finalizes the workbook and will close it
#
sub close_workbook {
    my $self = shift;
    if($self->{write_xls} && $self->{write_xls} eq "true") {
        write_charts(self=>$self);
        $spreadsheet{workbook}->close();
    }
}

# write_charts method - outputs charts in the xls file
#  Parameter hash (* means required):
#   *self              - Main Statsistics object
#
sub write_charts {
    my %options = (
        @_ );

    my $self = $options{self};
    if($self->{'nodes'} != undef) {
        my $worksheet;
        my $node_count = 1;
        my @nodes = @{$self->{'nodes'}};
        foreach(@nodes) {
            $worksheet = "worksheet".$node_count;
            my $worksheet_name = $worksheet . "_name";

            if($vm_row[$node_count] != 1) {
                my $paging_chart = $spreadsheet{workbook}->add_chart(name=>'Paging'.$node_count,          type=>'line', embedded => 1);
                $paging_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$vm_row[$node_count],1,1),name=>"Paging In");
                $paging_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$vm_row[$node_count],2,2),name=>"Paging out");
                $spreadsheet{$worksheet}->insert_chart(0,0,$paging_chart);
            }

            if($top_row[$node_count] != 1) {
                my $top_chart    = $spreadsheet{workbook}->add_chart(name=>'Physical Memory'.$node_count, type=>'line', embedded => 1);
                $top_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$top_row[$node_count],3,3),name=>"Active Memory");
                $top_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$top_row[$node_count],4,4),name=>"Inactive Memory");
                $top_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$top_row[$node_count],5,5),name=>"Wired Memory");
                $top_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$top_row[$node_count],6,6),name=>"Cache Memory");
                $top_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$top_row[$node_count],7,7),name=>"Free Memory");
                $spreadsheet{$worksheet}->insert_chart(20,0,$top_chart);
            }

            if($swap_row[$node_count] != 1) {
                my $swap_chart   = $spreadsheet{workbook}->add_chart(name=>'Swap Memory'.$node_count,     type=>'line', embedded => 1);
                $swap_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$swap_row[$node_count],8,8), name=>"Total swap");
                $swap_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$swap_row[$node_count],9,9), name=>"Used swap");
                $swap_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$swap_row[$node_count],10,10), name=>"Available swap");
                $spreadsheet{$worksheet}->insert_chart(0,9,$swap_chart);
            }

            if($command_row[$node_count] != 1) {
                my $cmd_chart    = $spreadsheet{workbook}->add_chart(name=>'Command Latency'.$node_count,     type=>'line', embedded => 1);
                $cmd_chart->add_series(values=>xl_range_formula($spreadsheet{$worksheet_name},1,$command_row[$node_count],0,0),name=>"Command Latency");
                $spreadsheet{$worksheet}->insert_chart(20,9,$cmd_chart);
            }

            $node_count++;
        }
    }
}

# command_latency method - Executes test to see how quick the command prompt responds
#  Parameter hash (* means required):
#   *cmd              - String command
#    count            - Used if you want to specify more than one command.
#                       Used with a loop to move the column.
#    row              - Number of rows the prompt is set to
#
sub command_latency {
    my $self = shift;
    my %params = (
            cmd=>"",
            count=>1,
            row=>0,
            show_output=>1,
            @_,
    );

    if (!$params{cmd}) {
        $Log->error("Command required for checking command latency");
        return 0;
    }
    my $worksheet = "worksheet1";

    my $hostobj;
    my $res = "";
    my $exec_time;

    if($self->{'nodes'} != undef) {
        my $node_count = 1;
        my @nodes = @{$self->{'nodes'}};
        if (@connections == undef) {
            my $tmp_count = 0;
            foreach(@nodes) {
                $hostobj = $_->hostrec();
                my $apiset = NACL::APISet->new(
                hostobj => $hostobj,
                category => "Node",
                interface => "CLI",
                set => "Systemshell",
                shell_type => "CMode"
                    );
                $connections[$tmp_count++] = $apiset;
            }
        }

        foreach(@nodes) {
            $worksheet = "worksheet".$node_count;
            my $apiset = $connections[$node_count - 1];
            eval {
                my $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows ".$params{row} .";"
                    . $params{cmd} . "\"";
                if($params{show_output} != 1) {
                    $cmd = $cmd . " \>/dev/null";
                }
                my $t0 = [gettimeofday];
                my $res = $apiset->execute_raw_command(command => $cmd); # Anything over 60 seconds comes in blank... and gets killed...
                my $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $elapsed_time;
               };

            if($self->{write_xls} && $self->{write_xls} eq "true") {
                #Write out a dynamic heading of the command and number of rows
                if($command_row_rows[$node_count][$params{count}] == undef) {
                    $command_row_rows[$node_count][$params{count}] = 1;
                    my $row = 10 + $params{count};
                    my $heading = $params{cmd} . "/" 
                                    . $params{row};
                    $spreadsheet{$worksheet}->write(0, $row, $heading);
                }
               $spreadsheet{$worksheet}->write($command_row_rows[$node_count][$params{count}], (10 + $params{count}), $exec_time);
            }

            $Log->comment("Command:\"" . $params{cmd} . "\" took " . $exec_time . " to run.");
            $command_row_rows[$node_count++][$params{count}]++;
        }
    }
    return $exec_time;
}

# average_commands method - Average of show commands (Can take a long time to run)
#
sub average_commands {
    my $self = shift;
    my %params = (
            @_,
    );

    my $worksheet = "worksheet1";

    my $hostobj;
    my $res = "";
    my $exec_time;

    if($self->{'nodes'} != undef) {
        my $node_count = 1;
        my @nodes = @{$self->{'nodes'}};
        foreach(@nodes) {
            $worksheet = "worksheet".$node_count;
            $hostobj = $_->hostrec();

            my $apiset = NACL::APISet->new(
                hostobj => $hostobj,
                category => "Node",
                interface => "CLI",
                set => "Systemshell",
                shell_type => "CMode"
            );

            # Total of commands run
            my $cmd_size = 0;
            eval {
                my $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80; vserver cifs show\" \>/dev/null";
                my $t0 = [gettimeofday];
                my $res = $apiset->execute_raw_command(command => $cmd);
                my $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;vserver cifs share show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;cluster peer show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;export-policy show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;export-policy rule show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;fcp adapter show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;fcp show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;iscsi show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;job show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;job history show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;


                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;job schedule cron show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;job schedule show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;job schedule interval show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;nfs show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;nfs status\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;nis show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;performance show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;quota report\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;sis show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;sis policy show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;snapmirror show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;snapshot show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;snapshot policy show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;storage disk show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;system show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;

                $cmd = "echo 'q' | /usr/bin/time ngsh -c \"set -rows 80;volume show\" \>/dev/null";
                $t0 = [gettimeofday];
                $res = $apiset->execute_raw_command(command => $cmd);
                $elapsed_time =  tv_interval($t0, [gettimeofday]);
                $exec_time = $exec_time + sprintf("%.2f", $elapsed_time);
                $cmd_size++;
           };
            if($cmd_size != 0) {
               $exec_time = $exec_time / $cmd_size;
            }
            if($self->{write_xls} && $self->{write_xls} eq "true") {
               $spreadsheet{$worksheet}->write($command_row[$node_count], 0, $exec_time);
            }

           $Log->comment("Command:\"" . $params{cmd} . "\" took " . $exec_time . " to run.");
           $command_row[$node_count++]++;
        }
    }
    return $exec_time;
}

# collect_vmstat method - Executes vmstat a number of iterations
#  Parameter hash (* means required):
#  iter         - Number of iterations the vmstat should run
#
sub collect_vmstat {
    my $self = shift;
    my %opts = (
            iter=>1,
            @_ );
    my $worksheet = "worksheet1";

    my $hostobj;
    my $res = "";
    if($self->{'nodes'} != undef) {
        my $node_count = 1;
        my @nodes = @{$self->{'nodes'}};
        foreach(@nodes) {
            $worksheet = "worksheet".$node_count;
            $hostobj = $_->hostrec();

            my $apiset = NACL::APISet->new(
                hostobj => $hostobj,
                category => "Node",
                interface => "CLI",
                set => "Systemshell",
                shell_type => "CMode"
            );
    
            eval {
                $res = $apiset->execute_raw_command(command => "vmstat 5 " . $opts{iter});
            };

            if($self->{write_xls} && $self->{write_xls} eq "true") {
                #Parse the paging and write it
                foreach my $line (split(/\n/, $res)) {
                    if ($line =~ m/(\d+)\s+(\d+)\s+(\d+)\s+(\d+)M\s+(\d+)M\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s/s) {
                        $spreadsheet{$worksheet}->write($vm_row[$node_count], 1, $8);
                        $spreadsheet{$worksheet}->write($vm_row[$node_count]++, 2, $9);
                    }
                }
            }

            $node_count++;
        }
    }

    return $res;
}

sub collect_ps {
    my $self = shift;
    my $hostobj;
    my $res = "";
    if($self->{'nodes'} != undef) {
        my @nodes = @{$self->{'nodes'}};
        foreach(@nodes) {
            $hostobj = $_->hostrec();

            my $apiset = NACL::APISet->new(
                hostobj => $hostobj,
                category => "Node",
                interface => "CLI",
                set => "Systemshell",
                shell_type => "CMode"
            );

            eval {
                $res = $apiset->execute_raw_command(command => "ps -aux");
            };
        }
    }
    return $res;
}

# collect_top_memory method - Executes top and can store the memory
#
sub collect_top_memory {
    my $self = shift;

    my $top_cmd = "top -d1 -b -S";
    my %opts = (
            @_ );
    my $worksheet = "worksheet1";

    my $hostobj;
    my $res = "";
    if($self->{'nodes'} != undef) {
        my $node_count = 1;
        my @nodes = @{$self->{'nodes'}};
        foreach(@nodes) {
            $worksheet = "worksheet".$node_count;
            $hostobj = $_->hostrec();

            my $apiset = NACL::APISet->new(
                hostobj => $hostobj,
                category => "Node",
                interface => "CLI",
                set => "Systemshell",
                shell_type => "CMode"
            );

            eval {
                $res = $apiset->execute_raw_command(command => $top_cmd);
            };

            my @parsed_output = {};
            my $active_mem;
            my $inactive_mem;
            my $wired_mem;
            my $cache_mem;
            my $free_mem;

            if($self->{write_xls} && $self->{write_xls} eq "true") {
                #Parse up the memory line
                foreach my $line (split(/\n/, $res)) {
                    if ($line =~ m/Mem:\s+(\d+)(\D+)\s+Active,\s+(\d+)(\D+)\s+Inact,\s+(\d+)(\D+)\s+Wired,\s+(\d+)(\D+)\s+Cache,\s+(\d+)(\D+)\s+Buf,\s+(\d+)(\D+) Free/s) {
                        $active_mem = $1;
                        $inactive_mem = $3;
                        $wired_mem = $5;
                        if( $8 eq "k" ||
                            $8 eq "kb" ||
                            $8 eq "K" ||
                            $8 eq "KB" ||
                            $8 eq "Kb" ) {
                            $cache_mem = $7 / 1024;
                        } else {
                            $cache_mem = $7;
                        }
                        if( $12 eq "k" ||
                            $12 eq "kb" ||
                            $12 eq "K" ||
                            $12 eq "KB" ||
                            $12 eq "Kb" ) {
                            $free_mem = $11 / 1024;
                        } else {
                            $free_mem = $11;
                        }
                    } elsif ($line =~ m/Mem:\s+(\d+)(\D+)\s+Active,\s+(\d+)(\D+)\s+Inact,\s+(\d+)(\D+)\s+Wired,\s+(\d+)(\D+)\s+Cache,\s+(\d+)(\D+) Free/s) {
                        $active_mem = $1;
                        $inactive_mem = $3;
                        $wired_mem = $5;
                        if( $8 eq "k" ||
                            $8 eq "kb" ||
                            $8 eq "K" ||
                            $8 eq "KB" ||
                            $8 eq "Kb" ) {
                            $cache_mem = $7 / 1024;
                        } else {
                            $cache_mem = $7;
                        }
                        if( $10 eq "k" ||
                            $10 eq "kb" ||
                            $10 eq "K" ||
                            $10 eq "KB" ||
                            $10 eq "Kb" ) {
                            $free_mem = $9 / 1024;
                        } else {
                            $free_mem = $9;
                        }
                    }
                }

                $spreadsheet{$worksheet}->write($top_row[$node_count], 3, $active_mem);
                $spreadsheet{$worksheet}->write($top_row[$node_count], 4, $inactive_mem);
                $spreadsheet{$worksheet}->write($top_row[$node_count], 5, $wired_mem);
                $spreadsheet{$worksheet}->write($top_row[$node_count], 6, $cache_mem);
                $spreadsheet{$worksheet}->write($top_row[$node_count], 7, $free_mem);
                $top_row[$node_count]++;
            }
            $node_count++;
        }
    }

    return $res;
}

# collect_swapinfo method - Executes swapinfo and can store it in worksheet
#
sub collect_swapinfo {
    my $self = shift;
    my %opts = (
            @_ );
    #my $worksheet = "worksheet".$opts{id};
    my $worksheet = "worksheet1";

    my $top_cmd = "swapinfo -m";
    my $hostobj;
    my $res = "";
    if($self->{'nodes'} != undef) {
        my $node_count = 1;
        my @nodes = @{$self->{'nodes'}};
        foreach(@nodes) {
            $worksheet = "worksheet".$node_count;
            $hostobj = $_->hostrec();

            my $apiset = NACL::APISet->new(
                hostobj => $hostobj,
                category => "Node",
                interface => "CLI",
                set => "Systemshell",
                shell_type => "CMode"
            );


            eval {
                $res = $apiset->execute_raw_command(command => $top_cmd);
            };

            if($self->{write_xls} && $self->{write_xls} eq "true") {
                foreach my $line (split(/\n/, $res)) {
                    if($line =~ m/.*dev.*\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)%/s) {
                            $spreadsheet{$worksheet}->write($swap_row[$node_count], 8, "$1");
                            $spreadsheet{$worksheet}->write($swap_row[$node_count], 9, "$2");
                            $spreadsheet{$worksheet}->write($swap_row[$node_count]++, 10, "$3");
                    }
                }
            }
            $node_count++;
        }
    }

    return $res;

}

# create_workbook method - Creates the workbook, should be wrapped by check of write_xls
#  Parameter hash (* means required):
#       write_xls   - expected "true" or undef - determines to write to a file
#       filename    - String of a filename that is optional
#      *nodes       - NACL::Node list that will be used for gathering stats on
#
sub create_workbook {
    my %opts = (
        write_xls=>undef,
        filename=>undef,
        nodes=>{ type => ARRAYREF, default =>undef},
        @_
    );

    # Build directory path
    my ($script_name, $path, $ext) = fileparse($0, qr/\.[^.]*/);
    my $logpath = param('LOGDIR');
    my $reportdir = $logpath.'/'.$script_name.'_report_dir/';

    # Create the directory
    $Log->comment("Creating subdir for report files: ".$reportdir);
    if (! mkdir($reportdir)) {
        $Log->error("Failed to create sub-directory for report files: ".$!);
        exit(1);
    }

    # Number of stats worksheets
    $spreadsheet{num_worksheets} = 1;

    # Get time
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    my $now = sprintf("%4d-%02d-%02d %02d:%02d:%02d", $year+1900, $mon+1, $mday, $hour, $min, $sec);

    my $filename =$now."-stats.xls";
    if($opts{'filename'}) {
        $filename = $opts{'filename'};
    }

    # Create a new Excel workbook
    $spreadsheet{workbook} = Spreadsheet::WriteExcel->new($reportdir.$filename);
    if (! defined($spreadsheet{workbook})) {
        $Log->error("Failed to create new Excel workbook : ".$!);
        exit(1)
    }

    # Set properties
    $spreadsheet{workbook}->set_properties(title    => 'Performance Statistics',
                                           author   => 'Statistics.pm',
                                           comments => 'Created with Perl and Spreadsheet::WriteExcel',
                                           company  => 'NetApp, Inc.');

    if($opts{'nodes'} != undef) {
        my $node_count = 1;
        my @nodes = @{$opts{'nodes'}};
        foreach(@nodes) {
            my $worksheet = "worksheet".$node_count;
            $vm_row[$node_count] = 0;
            $top_row[$node_count] = 0;
            $swap_row[$node_count] = 0;
            $command_row[$node_count] = 0;

            $spreadsheet{$worksheet."_name"} = $_->name();
            $spreadsheet{$worksheet} = $spreadsheet{workbook}->add_worksheet($spreadsheet{$worksheet."_name"});

            # Create the headings
            $spreadsheet{$worksheet}->write($command_row[$node_count]++, 0, "Command Latency");
            $spreadsheet{$worksheet}->write($vm_row[$node_count],      1, "Page In");
            $spreadsheet{$worksheet}->write($vm_row[$node_count]++,    2, "Page Out");
            $spreadsheet{$worksheet}->write($top_row[$node_count],     3, "Active Mem");
            $spreadsheet{$worksheet}->write($top_row[$node_count],     4, "Inactive Mem");
            $spreadsheet{$worksheet}->write($top_row[$node_count],     5, "Wired Mem");
            $spreadsheet{$worksheet}->write($top_row[$node_count],     6, "Cache Mem");
            $spreadsheet{$worksheet}->write($top_row[$node_count]++,   7, "Free Mem");
            $spreadsheet{$worksheet}->write($swap_row[$node_count],    8, "Total swap");
            $spreadsheet{$worksheet}->write($swap_row[$node_count],    9, "Used swap");
            $spreadsheet{$worksheet}->write($swap_row[$node_count]++, 10, "Available swap");

            $node_count++;
        }
    }
}

=pod

=head1 NAME

Statistics.pm - Statistic helper package.


=head1 SYNOPSIS

    use Statistics;
    my @nodes = NACL::C::Node->find();
    my $stat = Statistics->new(nodes => \@nodes, write_xls =>"true");
    my $command_seconds = $stat->command_latency(cmd => "volume show", row=>24);

=head1 DESCRIPTION

This library provides a basic statistic helper class in perl.
  
    
=head1 METHODS

new(node=>Node);
        Description:
            Creates a new instance of the Statistics.pm.
        Inputs:
            node => The NACL component Node to run tests on.
        Outputs:
            Statistics object instance
        Examples:
            <see synopsis above>

close_workbook();
        Description:
            Closes the document also writes out the graphs at the end.
            Does this only if told to write_xls.
        Inputs:
            None.
        Outputs:
            File will be closed pending it was told to write_xls
        Examples:
            $stat->close_workbook();

command_latency(cmd=>"command line");
        Description:
            Tests the duraction of a command to return specified rows, default is all.
        Inputs:
            cmd   => The actual filer command to run.
            row   => Number of rows to expect to return
            count => Allows the additional columns to be used for various commands
        Outputs:
            Returns the time in seconds it took to run.
        Examples:
            my @cmds = ("vol show", "net int show");
            my $cmd_count = 1;
            foreach(@cmds) {
                $stat->command_latency(cmd => $_, count=>$cmd_count++, row=>24);
            }
           
average_commands(); 
        Description:
            Averages a set of commands that are hard coded. Can take a long time to run.
        Inputs:
            None.
        Outputs:
            returns the average latency.
        Examples:
            $stat->average_commands();

collect_vmstat();
        Description:
            Will collect the vmstat info for paging. (Higher iter means longer pause.
                if you pass in 24 it will take 2 minutes PER node).
        Inputs:
            iter=> The number of iterations to run the vmstat command.
        Outputs:
            Will return the raw output
        Examples:
            my $vm = $stat->collect_vmstat(iter=>1);

collect_top_memory();
        Description:
            Uses the command top to get the total memory, active memory, wired and etc..
        Inputs:
            None.
        Outputs:
            Raw top output
        Examples:
            my $top_output = $stat->collect_top_memory();

collect_swapinfo();
        Description:
            Uses the command swapinfo on the filer to return the amount of swap space used
                and free.
        Inputs:
            None.
        Outputs:
            Raw swapinfo output
        Examples:
            my $swap_out = $stat->swapinfo();
=cut

1;
