## @summary ACP.pm - library module for ACP(Alternate Control Path) automation
##
## @status in-progress
##
## @author dl-storsub-qa,dl-storage-verification,anishv,anosh,dennis1,darshanm
##
## @pod here
## changes made on acpadmin_configure 08/19/2011 amal@netapp.com
## @change 20-01-2012 gururajc: modified acpadmin_voltage_status sub to handle IOM6E module output too
## @change 31-08-2012 gururajc: modified acp_reset sub with respect to Burt 634552
## @change 23-01-2013 nandeesh: replaced say() with filer_cmd() wherever possible.
## @change 05-04-2013 nandeesh: added support for the handling prompts while enabling acp in sub acp_enable().
## @change 25-10-2013 nandeesh: Corrected the sentence to check the acp connectivity of partner in sub acp_state() Burt #765507.
## @change 02-12-2013 nandeesh: Removed the required params (say INTERFACE\NETMASK\DOMAIN) from sub acp_enable() see burt 745855
##

=head1 NAME

Storage::ACP

=head1 SYNOPSIS

    use ACP qw(acpadmin_list_all);

    my $host = host($FILER);
    my $conn = connect($FILER);
    my $hash = acpadmin_list_all(host=>$host,....);


=head1 DESCRIPTION

    This is a library module for ACP(Alternate Control Path) automation

=head1 ROUTINES

=cut

package Storage::ACP;

use strict;
use warnings;

BEGIN {
    use base qw(Exporter);
    our @EXPORT_OK = qw(
      acpadmin_list_all
      get_acp_ids
      get_active_acp_ids
      acp_state
      acp_enable
      acp_disable
      acp_reset
      acp_login
      acpadmin_post_data
      acpadmin_voltage_status
      acpp_cmd
      acpadmin_configure
      );

    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess(Tharn::script_result(@msgs));
        });
} ## end BEGIN

use Carp qw(confess croak);
use Params::Validate qw(validate validate_pos OBJECT SCALAR HASHREF);
use Storage::SASUtils qw(set_filer_options randomize_array get_disks);
use Storage::Misc qw(filer_cmd);
use Storage::Storage qw(storage_show_acp);
use Tharn qw(say script_result host logcomment logresult param);
use Utils qw(make_conn);
use Data::Dumper qw(Dumper);
use ATI::ati_ontapi qw(ATI_SUCCESS
  ATI_ONTAPI_PRIVILEGE_DIAG
  ati_ontapi_create_context
  ati_ontapi_call_api);
use Params::Common;
use filer;
use NACL::APISet;
use NACL::C::Node;
use vars qw(@ISA @EXPORT
  $FILER_TEST_UI $INTERFACE $NETMASK $DOMAIN $IP_ADDRESS
  );

=head2 acpadmin_list_all

    my $hash = acpadmin_list_all(host=>$host, expander_name=>$expander_name);

This routine returns a pointer to hash table of acpadmin list_all command
It croaks on failure.
Following is an example of the hash table returned by the function:

$VAR1 = {

          '198.15.1.244' => {

                              'Version' => '1.1.5.0',

                              'MACAddress' => '00:50:cc:14:04:f4',

                              'InbandID' => '4d.12.A',

                              'ResetCnt' => '000',

                              'State' => '0x5',

                              'ShelfSN' => 'SHU9415700271E1',

                              'AssignerID' => '118060110',

                              'LastContact' => '960'

                            },

          '198.15.1.16' => {

                             'Version' => '1.1.5.0',

                             'MACAddress' => '00:50:cc:14:05:10',

                             'InbandID' => '4d.12.B',

                             'ResetCnt' => '000',

                             'State' => '0x5',

                             'ShelfSN' => 'SHU9415700271E1',

                             'AssignerID' => '118060110',

                             'LastContact' => '375'

                           },

          '198.15.1.146' => {

                              'Version' => '1.1.5.0',

                              'MACAddress' => '00:50:cc:62:61:92',

                              'InbandID' => '4d.22.B',

                              'ResetCnt' => '000',

                              'State' => '0x5',

                              'ShelfSN' => 'SHX0931422G000D',

                              'AssignerID' => '118060110',

                              'LastContact' => '351'

                            },

          '198.15.1.223' => {

                              'Version' => '1.1.5.0',

                              'MACAddress' => '00:50:cc:62:61:df',

                              'InbandID' => '4d.22.A',

                              'ResetCnt' => '000',

                              'State' => '0x5',

                              'ShelfSN' => 'SHX0931422G000D',

                              'AssignerID' => '118060110',

                              'LastContact' => '345'

                            },

          '198.15.1.180' => {

                              'Version' => '1.1.5.0',

                              'MACAddress' => '00:50:cc:14:05:b4',

                              'InbandID' => '4d.2.A',

                              'ResetCnt' => '000',

                              'State' => '0x5',

                              'ShelfSN' => 'SHU9415700271D3',

                              'AssignerID' => '118060110',

                              'LastContact' => '351'

                            },

          '198.15.1.224' => {

                              'Version' => '1.1.5.0',

                              'MACAddress' => '00:50:cc:14:04:e0',

                              'InbandID' => '4d.2.B',

                              'ResetCnt' => '000',

                              'State' => '0x5',

                              'ShelfSN' => 'SHU9415700271D3',

                              'AssignerID' => '118060110',

                              'LastContact' => '345'

                            }

        };


=over

=item host

(Required) : Filer connnection to use.


=item filermode

(Optional) : The mode of filer: "normal" or "maint". Default is "normal".
If filer in maintenance mode, the connection type will be "#console".

=item conntype

(Optional) : The connection type to be created by this function to talk to
the filer. Default is "#rsh".

=item timeout

(optional) : timeout value. Default is 300 secs.

=back

=cut

sub acpadmin_list_all {
    my (%opts) =
      validate(@_,
               {host      => {type => OBJECT, isa     => 'Hostrec'},
                filermode => {type => SCALAR, default => "normal", optional => 1},
                conntype  => {type => SCALAR, default => "#rsh", optional => 1},
                timeout   => {type => SCALAR, default => 300, optional => 1},});

    my $hash;
    my $output = filer_cmd(host => $opts{host}, filermode => $opts{filermode}, conntype => $opts{conntype}, timeout => $opts{timeout}, cmd => "acpadmin list_all");
    if ($output =~ /Alternate Control Path Management is disabled/) {
        return $output;
    }
    if ($output !~ /IP/) {
        croak $output;
    }
    my @acplist = split(/\r*\n/, $output);
    foreach my $line (@acplist) {
        if ($line =~ /\d+\.\d+\.\d+\.\d+/) {
            my @info = split(/\s+/, $line);
            $hash->{$info[0]}->{MACAddress}  = $info[1];
            $hash->{$info[0]}->{ResetCnt}    = $info[2];
            $hash->{$info[0]}->{LastContact} = $info[3];
            $hash->{$info[0]}->{Version}     = $info[4];
            $hash->{$info[0]}->{AssignerID}  = $info[5];
            $hash->{$info[0]}->{ShelfSN}     = $info[6];
            $hash->{$info[0]}->{State}       = $info[7];

            # $hash->{$info[0]}->{InbandID} = $info[8];
            my @InbandID = split(/\./, $info[8]);
            $InbandID[1] = int($InbandID[1]);
            $hash->{$info[0]}->{InbandID} = join(".", @InbandID);
            if (defined($info[9])) {
                $hash->{$info[0]}->{ModuleType} = $info[9];
            }

        } ## end if ($line =~ /\d+\.\d+\.\d+\.\d+/)
    } ## end foreach my $line (@acplist)
    return $hash;

} ## end sub acpadmin_list_all

=head2 get_acp_ids

    my @ids = get_acp_ids(hash=>$hash);

This routine returns an array of acp ids in the system.

=over

=item hash

(Required) : reference to the hash returned by storage_show_acp() function

=back

=cut

sub get_acp_ids {
    my (%opts) = validate(@_, {hash => {type => HASHREF},});
    my @arr = keys %{$opts{hash}};
    my @arr1 = grep(/\w+\.\w+\.\w+/, @arr);
    return @arr1;
}

=head2 get_active_acp_ids

    my @ids = get_active_acp_ids(hash=>$hash);

This routine returns an array of active acp ids in the system.

=over

=item hash

(Required) : reference to the hash returned by storage_show_acp() function

=back

=cut

sub get_active_acp_ids {
    my (%opts) = validate(@_, {hash => {type => HASHREF},});
    my @active;
    my @arr = keys %{$opts{hash}};

    #    print "@arr \n";
    my @arr1 = grep(/\w+\.\w+\.\w+/, @arr);
    foreach my $arr1 (@arr1) {
        if ($opts{hash}->{$arr1}->{Status} eq "active") {
            push @active, $arr1;
        }
    }
    return @active;

} ## end sub get_active_acp_ids

=head2 acp_disable

    acp_disable($host);

This routine disables ACP.

=over

=item host

(Required) : reference to the hostspec to filer returned by host() function

=back

=cut

sub acp_disable {

    my ($h) = @_;

    set_filer_options(host => $h, options => "acp.enabled", value => "off");

}

=head2 acp_enable

    my $err = acp_enable($host);

This routine enables ACP then check the state. It returns 0 for success and >0 for failure.

=over

=item host

(Required) : reference to the hostspec to filer returned by host() function

=item INTERFACE

(optional) : Ethernet Interface. Default is e0P

=item NETMASK

(optional) : ACP Netmask. Default is "255.255.252.0"

=item DOMAIN

(optional) : ACP Subnet. Default is undef

=item IP_ADDRESS

(optional) : ACP IP Address. Default is undef

=back

=cut

sub acp_enable {

    my ($h) = @_;
    param ('INTERFACE', '-default', "e0P");
    param ('IP_ADDRESS', '-default', "");
    param ('DOMAIN', '-default', "");
    param ('NETMASK', '-default', "");
    # param ('IP_ADDRESS', '-default', undef);  #commented for a while
    # param ('DOMAIN', '-default', undef);
    # param ('NETMASK', '-default', "255.255.252.0");
    eval {
       set_filer_options(host => $h, options => "acp.enabled", value => "on");
    };
    my $error = $@;
    # if($error && (!defined(Tharn::param("INTERFACE")) || !defined(Tharn::param("DOMAIN")) || !defined(Tharn::param("NETMASK")))){
        # logresult("config", "Please pass INTERFACE\\DOMAIN\\NETMASK ip\'s as a parameters through ntest");
    # } elsif($error){
    if($error){
        $INTERFACE = Tharn::param("INTERFACE");
        $NETMASK = Tharn::param("NETMASK");
        $DOMAIN = Tharn::param("DOMAIN");
        print "Interface is $INTERFACE\nNetmask is $NETMASK\nDomain is $DOMAIN\n";
        acpadmin_configure(host => $h,
                      interface => $INTERFACE,
                      domain    => $DOMAIN,
                      netmask   => $NETMASK );
    }
    # sleep to allow ACP initialization
    sleep 30;

    return acp_state($h);
} ## end sub acp_enable

=head2 acp_state

    my $err = acp_state($host);

This routine checks ACP state and returns 0 for normal and >0 for abnormal.

=over

=item host

(Required) : reference to the hostspec to filer returned by host() function

=back

=cut

sub acp_state {

    my ($h) = @_;
    my $a = storage_show_acp(host => $h);
    sleep 90;
    my $err = 0;

    if (!defined $a) {
        logresult('warn', "storage_show_acp returned empty hash.");
        return $err++;
    }

    if ($a->{Status} ne "Active") {
        logresult('warn', "ACP Status is:  $a->{Status}");
        $err++;
    }
    if ($a->{ACPStatus} ne "Enabled") {
        logresult('warn', "ACP State is:  $a->{ACPStatus}");
        $err++;
    }
    if ($a->{Interface} !~ m/e\d\D/) {
        logresult('warn', "ACP Ethernet Interface is:  $a->{Interface}");
        $err++;
    }
    if ($a->{IPAddress} !~ m/\d+\.\d+\.\d+\.\d/) {
        logresult('warn', "ACP IP Address is:  $a->{IPAddress}");
        $err++;
    }
    if ($a->{Domain} !~ m/\d+\.\d+\.\d+\.\d/) {
        logresult('warn', "ACP Domain is:  $a->{Domain}");
        $err++;
    }
    if ($a->{Netmask} !~ m/\d+\.\d+\.\d+\.\d/) {
        logresult('warn', "ACP Netmask is:  $a->{Netmask}");
        $err++;
    }
    if ($a->{Connectivity} ne "Full Connectivity") {
        logresult('warn', "ACP Connectivity Status is:  $a->{Connectivity}");
        $err++;
    }

    if ($a->{PartnerConnectivity} !~ m/Full Connectivity/) {
        logresult('warn', "ACP Partner Connectivity Status is:  $a->{PartnerConnectivity}");
        $err++;
    }


    my @acp_ids = get_acp_ids(hash => $a);
    my @active_acp_ids = get_active_acp_ids(hash => $a);

    if ($#acp_ids eq -1) {
        logresult('warn', "get_acp_ids returned empty list.");
        return $err++;
    }
    if ($#active_acp_ids eq -1) {
        logresult('warn', "get_active_acp_ids returned empty list.");
        return $err++;
    }
    if ($#active_acp_ids ne $#acp_ids) {
        logresult('warn', "Mismatched ACP IDs. All: @acp_ids; Active: @active_acp_ids");
        return $err++;
    }

    my $mcnt = 0;
    foreach my $a (@acp_ids) {
        foreach my $b (@active_acp_ids) {
            if ($a eq $b) {
                $mcnt++;
            }
        }
    }

    $mcnt -= 1;
    if ($#acp_ids ne $mcnt) {
        logresult('warn', "Mismatched ACP IDs. All: @acp_ids; Active: @active_acp_ids");
        $err++;
    }

    return $err;
} ## end sub acp_state

=head2 acp_reset

    my $hash = acp_reset(host=>$host,expander_name=>$expander_name);

This routine resets expanders using acpadmin expander_reset or expander_power_cycle commands
It returns 0 if no errors are found and 1 or more if errors are found

=over

=item host

(Required) : Filer connnection to use.


=item filermode

(Optional) : The mode of filer: "normal" or "maint". Default is "normal".
If filer in maintenance mode, the connection type will be "#console".

=item conntype

(Optional) : The connection type to be created by this function to talk to
the filer. Default is "#rsh".

=item timeout

(optional) : timeout value. Default is 300 secs.

=item reset_type

(optional) : "reset" for expander_reset and "power_cycle for expander_power_cycle. Default is "reset"

=item domain_reset

(optional) : Resetting all A/B modules in parallel. Default 0.

=item module

(optional) : "," or " " seperated list of modules to reset (Example: 4a.01.A. Module id from storage show acp). Default is "all".

=item cnt_check

(optional) : To check for reset count after resetting an expander? Default 1.

=item random_reset

(optional) : Reset list of modules in random order. Default 0.

=item interval

(optional) : Sleep time in secs between resets. Default is 180 secs

=item reset_both

(optional) : Reset all the modules in the list without sleep interval. Default 0.

=back

=cut

sub acp_reset {
    my (%opts) =
      validate(@_,
               {host         => {type => OBJECT, isa     => 'Hostrec'},
                filermode    => {type => SCALAR, default => "normal", optional => 1},
                conntype     => {type => SCALAR, default => "#rsh", optional => 1},
                timeout      => {type => SCALAR, default => 300, optional => 1},
                reset_type   => {type => SCALAR, default => "reset", optional => 1},
                domain_reset => {type => SCALAR, default => "0", optional => 1},
                module       => {type => SCALAR, default => "all", optional => 1},
                cnt_check    => {type => SCALAR, default => 1, optional => 1},
                random_reset => {type => SCALAR, default => 0, optional => 1},
                interval     => {type => SCALAR, default => 180, optional => 1},
                reset_both   => {type => SCALAR, default => 0, optional => 1},});
    my $output;
    my @modules;
    my @a_modules;
    my @b_modules;
    my @a_list;
    my @b_list;
    my $a_modules;
    my $b_modules;
    my $reset_type;
    my $err = 0;


    if ($opts{reset_type} eq "reset") {
        $reset_type = "expander_reset";
    } elsif ($opts{reset_type} eq "power_cycle") {
        $reset_type = "expander_power_cycle";
    } else {
        croak "Invalid ACP reset type: Valid options are reset or power_cycle";
    }

    my $hash1 = storage_show_acp(host => $opts{host}, filermode => $opts{filermode}, conntype => $opts{conntype}, timeout => $opts{timeout});
    my @active_modules = get_active_acp_ids(hash => $hash1);    #Get a list of active ACP ids from storage show acp
    if (($opts{module} eq "all") || ($opts{domain_reset})) {    #Reset all the modules.
        @modules = @active_modules;
    } else {
        @modules = split(/\s+|,/, $opts{module});
    }
    if ($opts{random_reset}) {
        randomize_array(\@modules);
    }

    if ($opts{domain_reset}) {
        foreach my $module (@modules) {
            if ((substr($module, -1, 1)) eq "A") {
                push(@a_modules, "acpadmin $reset_type " . $module);
                push(@a_list,    $module);
            } else {
                push(@b_modules, "acpadmin $reset_type " . $module);
                push(@b_list,    $module);
            }
            $a_modules = join(';', @a_modules);
            $b_modules = join(';', @b_modules);
        } ## end foreach my $module (@modules)
        if ($#a_list >= 0) {
            logcomment("Resetting Domain A: @a_list with $reset_type");
            my $output = filer_cmd(host => $opts{host}, filermode => $opts{filermode}, conntype => $opts{conntype}, timeout => $opts{timeout}, cmd => "$a_modules");
            if ($output =~ /Usage|Unable to find ACPP|Alternate Control Path is not active|Cannot perform reset/) {
                logcomment("WARNING: $output");
                $err++;
            }
            unless ($opts{reset_both}) {    #This is for cases where we want to reset all the modules in parallel
                logcomment("Sleeping for $opts{interval} secs");
                sleep $opts{interval};
            }
        } ## end if ($#a_list >= 0)
        if ($#b_list >= 0) {
            logcomment("Resetting Domain B: @b_list");
            my $output = filer_cmd(host => $opts{host}, filermode => $opts{filermode}, conntype => $opts{conntype}, timeout => $opts{timeout}, cmd => "$b_modules");
            if ($output =~ /Usage|Unable to find ACPP|Alternate Control Path is not active|Cannot perform reset/) {
                logcomment(" WARNING: $output");
                $err++;
            }
            logcomment("Sleeping for $opts{interval} secs");
            sleep $opts{interval};
        } ## end if ($#b_list >= 0)
        if (($#a_list < 0) && ($#b_list < 0)) {    #We shouldn't be running into this condition.
            logresult("config", "No active acps found on either of A or B modules");
        }
    } else {
        foreach my $module (@modules) {
            logcomment("Resetting module $module with $reset_type");
            my $output = filer_cmd(host => $opts{host}, filermode => $opts{filermode}, conntype => $opts{conntype}, timeout => $opts{timeout}, cmd => "acpadmin $reset_type $module");
            if ($output =~ /Usage|Unable to find ACPP|Alternate Control Path is not active|Cannot perform reset/) {
                logcomment(" WARNING: $output");
                $err++;
            }
            unless ($opts{reset_both}) {           #This is useful if we want to reset both modules of a shelf in parallel
                logcomment("Sleeping for $opts{interval} secs");
                sleep $opts{interval};
            }
        } ## end foreach my $module (@modules)
    } ## end else [ if ($opts{domain_reset...

    if ($opts{reset_both}) {                       # We didn't give enough time above. Let's sleep for some time before checking the reset count again.
        logcomment("Sleeping for $opts{interval} secs");
        sleep $opts{interval};
    }

    my $hash2 = storage_show_acp(host => $opts{host}, filermode => $opts{filermode}, conntype => $opts{conntype}, timeout => $opts{timeout});
    my $befval;
    my $aftval;
    if ($opts{cnt_check}) {
        foreach my $module (@modules) {
            if(($module =~ /\.\d+0\./)|($module =~ /\.0\d+\./)){
                $befval = $hash1->{$module}->{ResetCnt};
                $aftval  = $hash2->{$module}->{ResetCnt};
            } else{
                $module =~ s/(\w+)\.(\w)\.(\w+)/$1.0$2.$3/;
                $befval = $hash1->{$module}->{ResetCnt};
                $aftval  = $hash2->{$module}->{ResetCnt};
            }
            if ($aftval != $befval + 1) {
        logresult('warn', "ResetCount didn't increment OR multiple unexpected resets happened for $module. Before it was $befval and after reset it is $aftval.");
        ##
        ## burt 526358: Commenting out the following line to avoid fatal fail during count, see burt516358 for more details.
        ## Once fix is available for this, remove comment lines.
        ##
                ## $err++;
            }
        }
    } ## end if ($opts{cnt_check})
    foreach my $active (@active_modules) {    # Modules should remain active after reset.
        if ($hash2->{$active}->{Status} ne "active") {
            logcomment("WARNING: Module $active is not active");
            $err++;
        }
    }
    return $err;
} ## end sub acp_reset

=head2 acp_login

    my $hash = acp_reset(host=>$host,expander_name=>$expander_name);

This routine logs in into the ACPP console.

=over

=item host

(Required) : ACPP console connnection to use.

=item conntype

(Optional) : The connection type to be created by this function to talk to
the ACPP . Default is "#console".

=item timeout

(optional) : timeout value. Default is 300 secs.

=item retry

(optional) : Number of retires if login fails. Default 1.

=back

=cut

sub acp_login {
    my (%opts) =
      validate(@_,
               {host     => {type => OBJECT, isa     => 'Hostrec'},
                conntype => {type => SCALAR, default => "#console", optional => 1},
                timeout  => {type => SCALAR, default => 60, optional => 1},
                retry    => {type => SCALAR, default => 1, optional => 1},});

    my $fc = make_conn($opts{host}, $opts{conntype});
    my ($a, $b, $o);
    my $retry = 1;
  ACP_LOGIN_LABEL:;
    ($a, $b, $o) = say($fc, "\n", -swallow, "-noprompt", -timeout, "$opts{timeout}", '-wait', "local login:|Password:|local#|Login incorrect");
    print " A: $a \n";
    print " B: $b \n";
    print " o: $o \n";

    if (($b =~ /Login incorrect/) || ($a =~ /Login incorrect/)) {
        sleep 10;
        ($a, $b, $o) = say($fc, "\r", -swallow, "-noprompt", -timeout, "$opts{timeout}", '-wait', "local login:");
    }

    if ($o =~ /local login:/) {
        $o = say($fc, "root", -swallow, -noprompt, '-wait', "Password:", -timeout, "$opts{timeout}");
        say($fc, "$opts{host}->{conn}->{console}->{password}", -swallow, -timeout, "$opts{timeout}");
    } elsif ($o =~ /Password:/) {
        say($fc, "$opts{host}->{conn}->{console}->{password}", -swallow, -timeout, "$opts{timeout}");
    } elsif ($o =~ /local#\s*$/) {
        logcomment("Logged in");
    } else {
        if ($retry <= $opts{retry}) {
            sleep 30;
            $retry++;
            goto ACP_LOGIN_LABEL;
        } else {
            die "Unable to login into ACPP console";
        }
    }
} ## end sub acp_login

=head2 acpadmin_post_data

    my $hash = acpadmin_post_data(host=>$host, expander_name=>$expander_name);

This routine returns a pointer to hash table of acpadmin post_data command or croaks on failure.
Following is an example of the hash table returned by the function:

$VAR1 = {

          '4a.0.A' => {
                        'ACPPostStatusRegister1' => {
                                                        'PostTestCompleted' => 'PASSED'
                                                        'ACPECCStatus' => 'PASSED'
                                                        'PowerRegisterReportingErrorStatus' => 'PASSED'
                                                        'BiggerMemoryTest' => 'PASSED'
                                                    }

                        'ACPPostStatusRegister2' => {
                                                        'CPU'sEthernetTest' => 'PASSED'
                                                        'CPU'sUartLoopTest' => 'PASSED'
                                                        'CPU'sCacheTest' => 'PASSED'
                                                        'I2CBusTest' => 'PASSED'
                                                        'SimpleMemoryTest' => 'PASSED'
                                                        'ACPCPLDTest' => 'PASSED'
                                                    }

                        'EBODPostStatusRegister1' => {
                                                        'PostTestCompleted' => 'PASSED'
                                                        'EBODActiveBootImage' => 'B'
                                                        'EBODSystemMonitorStatus' => 'PASSED'
                                                        'EBODInternalMemoryTest' => 'PASSED'
                                                        'EBODMemoryTest' => 'PASSED'
                                                     }

                        'EBODPostStatusRegister2' => {
                                                        'HALInit' => 'PASSED'
                                                        'I2CTest' => 'PASSED'
                                                        'MidPlaneVPDState' => 'NON-DEFAULT'
                                                        'MidPlaneCPLDTest' => 'PASSED'
                                                        'EBODCPLDTest' => 'PASSED'
                                                     }

                        'ACPHW/SWINFO' => {
                                                        'CPLDVersion' => '0x02'
                                                        'U-BootVersion' => 'sable-44'
                                                     }
                        }
        }

=over

=item host

(Required) : Host Spec of Filer.

=item expander_name

(Required) : The ACP module from which the post data needs to be retrieved.

=item conntype

(Optional) : The connection type to be created by this function to talk to
the filer. Default is "#console".

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item wait_timeout

(Optional) : The duration for which the logs would be captured after the command as we dont get the prompt automatically.
             Default is 20 secs.

=item return_raw

(Optional) : Instructs the command not to process the output.  The retrieved values would be retured to the calling routine.
             Default is 0.

=back

=cut

sub acpadmin_post_data {
    my (%opts) =
      validate(@_,
               {host      => {type => OBJECT, isa     => 'Hostrec'},
                filermode => {type => SCALAR, default => "normal", optional => 1},
                conntype  => {type => SCALAR, default => "#console", optional => 1},
                timeout   => {type => SCALAR, default => 300, optional => 1},
                expander_name => {type => SCALAR},
                wait_timeout  => {type => SCALAR, default => 20, optional => 1},
                return_raw    => {type => SCALAR, default => 0, optional => 1},});

    my $hash;

    my $host = $opts{host};
    my $conn = make_conn($host, $opts{conntype});
    my $cmd  = "priv set -q diag;acpadmin post_data $opts{expander_name}";
    #check the version of ONTAP. If the version is <8.0 then the command should be executed as given in the if condition
    #otherwis as provided in the "else" part of the code. This is due the change in the command output of ONTAP

    my $label_version = _get_label_version(host => $host, conntype => "#rsh");

    my $output2;

    if ($label_version < 11) {

        say($conn, "$cmd", '-blind');
        sleep 5;
        my (undef, undef, $output1, undef) = say($conn, "\r", -timeout, $opts{timeout});
        print Dumper $output1;
        if ($output1 =~ /\[acpadmin post data\]/) {
            ($output1) = grep(/post data/, split(/\r+\n/, $output1));
            return $output1;
        } elsif ($output1 =~ /Data Received from ACP module/) {
            if ($opts{return_raw} == 1) {
                return $output1;
            }
            $output2 = $output1;
        } else {
            sleep $opts{wait_timeout};
            (undef, undef, $output2, undef) = say($conn, "\r", -timeout, $opts{timeout});

            if (($opts{return_raw} == 1) || ($output2 !~ /Data Received from ACP module/)) {
                my $output = $output1 . $output2;
                return $output;
            }
        }
    } else {
        my $output1;
        my $modeType = $host->hosttype();
        if ($modeType =~ /filer-ng/) {
            $output1 = filer_cmd(host => $opts{host}, conntype => $opts{conntype}, timeout => $opts{timeout}, cmd => $cmd);
        } else {
            $output1 = filer_cmd(host => $opts{host}, timeout => $opts{timeout}, cmd => $cmd);

        }
        sleep 5;
        print Dumper $output1;
        if ($output1 =~ /\[acpadmin post data\]/) {
            ($output1) = grep(/post data/, split(/\r+\n/, $output1));
            return $output1;
        } elsif ($output1 =~ /Data Received from ACP module/) {
            if ($opts{return_raw} == 1) {
                return $output1;
            }
            $output2 = $output1;
        } else {
            sleep $opts{wait_timeout};
            (undef, undef, $output2, undef) = say($conn, "\r", -timeout, $opts{timeout});

            if (($opts{return_raw} == 1) || ($output2 !~ /Data Received from ACP module/)) {
                my $output = $output1 . $output2;
                return $output;
            }
        }
    }
    my @postdatalist = grep(/\w+/, split(/\r*\n/, $output2));
    my $PostStatusRegisterFlag;
    my $count   = 0;
    my $hashref = {};
    my $inband_id;
    my $tempo;
    while ($count <= $#postdatalist) {
        if ($postdatalist[$count] =~ /Data Received from ACP module/) {
            $postdatalist[$count] =~ s/ //g;
            (undef, $inband_id) = split(/:/, $postdatalist[$count]);
            if($inband_id =~ /\.(\d)\./){
                $tempo=$1;
                $inband_id =~ s /\.\d+\./.0$tempo./;
            }
        } elsif ($postdatalist[$count] =~ /(ACP Post Status Register|EBOD Post Status Register|ACP HW\/SW INFO)/) {
            $postdatalist[$count] =~ s/ //g;
            $PostStatusRegisterFlag = $postdatalist[$count];
            $PostStatusRegisterFlag =~ s/:$//;
        } elsif ($postdatalist[$count] =~ /\w+/) {
            $postdatalist[$count] =~ s/ //g;
            my @line_split = split(/:/, $postdatalist[$count]);
            $hashref->{$inband_id}->{$PostStatusRegisterFlag}->{$line_split[0]} = $line_split[1];
        } else {

        }
        $count++;
    } ## end while ($count <= $#postdatalist)
    return $hashref;

} ## end sub acpadmin_post_data

=head2 acpadmin_voltage_status

    my $hash = acpadmin_voltage_status(host=>$host, expander_name=>$expander_name);

This routine returns a pointer to hash table of acpadmin voltage_status command or croaks on failure.
Following is an example of the hash table returned by the function:

$VAR1 = {
          '3b.1.B' => {
                        'ShelfVoltageACPStatus' => {
                                                     'ACPHotswap' => 'Good',
                                                     'ACP1.8v' => 'Good',
                                                     'ACP3.3v' => 'Good',
                                                     'ACP2.5v' => 'Good',
                                                     'ACP1.2v' => 'Good',
                                                     'ACP1.5v' => 'Good'
                                                   },
                        'ShelfVoltageEBODStatus' => {
                                                      'EBOD3.3v' => 'Good',
                                                      'EBOD1.2v' => 'Good',
                                                      'EBOD5.0v' => 'Good',
                                                      'EBODHotswap' => 'Good'
                                                   }
                      }
        };
=over

=item host

(Required) : Host Spec of Filer.

=item expander_name

(Required) : The ACP module from which the voltage status needs to be retrieved.

=item conntype

(Optional) : The connection type to be created by this function to talk to
the filer. Default is "#console".

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item wait_timeout

(Optional) : The duration for which the logs would be captured after the command as we dont get the prompt automatically.
             Default is 20 secs.

=back

=cut

sub acpadmin_voltage_status {
    my (%opts) =
      validate(@_,
               {host      => {type => OBJECT, isa     => 'Hostrec'},
                filermode => {type => SCALAR, default => "normal", optional => 1},
                conntype  => {type => SCALAR, default => "#console", optional => 1},
                timeout   => {type => SCALAR, default => 300, optional => 1},
                expander_name => {type => SCALAR},
                wait_timeout  => {type => SCALAR, default => 20, optional => 1},});

    my $hash;
    my $host = $opts{host};
    my $conn = make_conn($host, $opts{conntype});
    my $cmd = "priv set -q diag;acpadmin voltage_status $opts{expander_name}";

    #check the version of ONTAP. If the version is <8.0 then the command should be executed as given in the if condition
    #otherwis as provided in the "else" part of the code. This is due the change in the command output of ONTAP

    my $label_version = _get_label_version(host => $host, conntype => "#rsh");

    my ($output1, $output2);

    if ($label_version < 11) {

        #This means the ONTAP version is of prior to BR
        say($conn, "$cmd", '-blind');
        sleep 5;
        (undef, undef, $output1, undef) = say($conn, "\r", -timeout, $opts{timeout});
        if ($output1 =~ /\[acpadmin voltage status\]/) {
            ($output1) = grep(/voltage status/, split(/\r+\n/, $output1));
            return $output1;
        }
        sleep $opts{wait_timeout};
        (undef, undef, $output2, undef) = say($conn, "\r", -timeout, $opts{timeout});

    } else {

        #This means the ONTAP version is BR or its later verison
        $output2 = filer_cmd(host=>$opts{host}, cmd=>$cmd, timeout=>$opts{timeout});
    }
    if ($output2 !~ /Received from ACP module/) {
        my $output = $output1 . $output2;
        return $output;
    }
    my @voltagestatuslist = grep(/\w+/, split(/\r*\n/, $output2));
    my $VoltageStatusRegisterFlag;
    my $count   = 0;
    my $hashref = {};
    my $inband_id;
    my $tempo;
    while ($count <= $#voltagestatuslist) {
        if ($voltagestatuslist[$count] =~ /Received from ACP module/) {
            $voltagestatuslist[$count] =~ s/ //g;
            (undef, $inband_id) = split(/:/, $voltagestatuslist[$count]);
            if($inband_id =~ /\.(\d)\./){
                $tempo=$1;
                $inband_id =~ s /\.\d+\./.0$tempo./;
            }
        } elsif ($voltagestatuslist[$count] =~ /(Shelf Voltage ACP Status|Shelf Voltage EBOD Status|Module Voltage ACP Status|Module Voltage EBOD Status|Module Voltage ACP Latched Status|Module Voltage EBOD Latched Status|Module PC Voltage Status)/) {
            $voltagestatuslist[$count] =~ s/ //g;
            $VoltageStatusRegisterFlag = $voltagestatuslist[$count];
            $VoltageStatusRegisterFlag =~ s/:$//;
        } elsif ($voltagestatuslist[$count] =~ /\w+/) {
            $voltagestatuslist[$count] =~ s/ //g;
            my @line_split = split(/:/, $voltagestatuslist[$count]);
            $hashref->{$inband_id}->{$VoltageStatusRegisterFlag}->{$line_split[0]} = $line_split[1];
        } else {

        }
        $count++;
    } ## end while ($count <= $#voltagestatuslist)
    return $hashref;

} ## end sub acpadmin_voltage_status

=head2 acpp_cmd

    my $scalar =  acpp_cmd(host=>$host, cmd => $cmd);

This routine tries to login to the acp console object specified, sends a command and returns the output of the command as a string.
It uses the acp_login function to login to the console.

=over

=item host

(Required) : Host Spec of acp console.

=item cmd

(Required) : The command to be executed on the acp console.

=item conntype

(Optional) : The connection type to be created by this function to talk to the console. Default is "#console".

=item timeout

(Optional) : timeout value. Default is 60 secs.

=item retry

(Optional) : The number of tries that could be done to login to the host object.
             Default is 1.

=back

=cut

sub acpp_cmd {
    my (%opts) =
      validate(@_,
               {host     => {type => OBJECT, isa     => 'Hostrec'},
                conntype => {type => SCALAR, default => "#console", optional => 1},
                timeout  => {type => SCALAR, default => 60, optional => 1},
                retry    => {type => SCALAR, default => 1, optional => 1},
                cmd => {type => SCALAR},});

    acp_login(host => $opts{host}, conntype => $opts{conntype}, timeout => $opts{timeout}, retry => $opts{retry});

    my $acpp_con = make_conn($opts{host}, $opts{conntype});

    my ($match, $before, $after) = say($acpp_con, "$opts{cmd}", -timeout, "$opts{timeout}");

    return $after;
} ## end sub acpp_cmd

sub acpadmin_configure {

    my (%opts) =
      validate(@_,
               {host      => {type => OBJECT, isa     => 'Hostrec'},
                filermode => {type => SCALAR, default => "normal", optional => 1},
                conntype  => {type => SCALAR, default => "#console", optional => 1},
                timeout   => {type => SCALAR, default => 300, optional => 1},
                interface => {type => SCALAR, default => 'e0P', optional => 1},
                netmask   => {type => SCALAR, default => "", optional => 1},
                domain    => {type => SCALAR, default => "", optional => 1},
                subnet    => {type => SCALAR, default => "", optional => 1},});
    my $filer_ci = NACL::C::Node->find();
    my $api_set_obj = $filer_ci->get_7m_or_nodescope_apiset();
    my $cmd = "acpadmin configure";
    $api_set_obj->set_timeout('connectrec-timeout' => $opts{timeout});

    my @match_table = (
        'Do you want to configure the Shelf Alternate Control Path Management interface for SAS shelves' => 'y',
        'Do you want to reconfigure it\?' => 'y',
        'Do you want to configure the Shelf Alternate Control Path Management using cluster partner subnet and netmask' => 'y',
        'Enter the network interface you want to use for the Alternate Control Path Management' => "$opts{interface}",
        'Enter the netmask for Network Interface' => "$opts{netmask}",
        'Enter the domain for Network Interface' => "$opts{domain}",
        'Enter the subnet for Network Interface' => "$opts{subnet}",
    );

    $api_set_obj->execute_command(command => $cmd,
                              'connectrec-match_table' => \@match_table,);
} ## end sub acpadmin_configure

#This function returns the label-version of the disk from the command ouput of label examine $disk -a.
sub _get_label_version {
    my (%opts) =
      validate(@_,
               {host     => {type => OBJECT, isa     => 'Hostrec'},
                conntype => {type => SCALAR, default => "#rsh", optional => 1},
                timeout  => {type => SCALAR, default => 300, optional => 1},});
    my $host          = $opts{host};
    my $conntype      = $opts{conntype};
    my $label_version = undef;
    my $output;
    my $conn = make_conn($host, $conntype);
    my @disks = get_disks(host      => $host,
                          raid_type => "AGGREGATE",
                          conntype  => $conntype);

    foreach my $disk (@disks) {
        my $cmd = "priv set -q diag; label examine $disk -a";
        $output = filer_cmd(host=>$opts{host}, cmd=>$cmd, timeout=>$opts{timeout});
        if ($output =~ /label_version=(\d+)/i) {
            $label_version = $1;
            logcomment("label-version: $label_version");
            last;
        } else {
            logcomment("Label examine is not found for $disk");

        }
    } ## end foreach my $disk (@disks)
    return $label_version;
} ## end sub _get_label_version

=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = "storage show acp", "acpadmin"

    ONTAP_ZAPI_APIS =

=head1 AUTHOR
    dl-storage-verification,dl-storsub-qa,anishv,anosh,dennis1
    $Id: //depot/prod/test/Rsierranevada/lib/Storage/ACP.pm#5 $
    Last revised $Date: 2013/12/02 $


=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;

