## @summary Alternate Control Path Log File parsing Library module
##
## @description
##  Library of routines for parsing the ACP Log files acplog_master and acplog to search for specific records
##
## @synopsis
##  use Storage::ACPLog
##
## @author darshanm@netapp.com
##
## @status TESTME
##
## @pod here
## @change 07/01/2013 nandeesh: gave C-mode support for sub acp_log_mount() and implemented "scp" copy operation for the same in c-mode

=head1 NAME

Storage::ACPLog

=head1 SYNOPSIS

    use ACPLog qw(acp_log_extract, acp_log_mount);

    my $host = host($FILER);
    my $conn = connect($FILER);
    my $hash = acp_log_mount(host=>$host,....);


=head1 DESCRIPTION

    This is a library module for Alternate Control Path Log Parsing Automation

=head1 EXAMPLES

    The following example shows the usage of different subroutines to check for the logs

        #Define the global structure for storing information on acp_logs
        my $acp_logs {};

        #Mark the ACP Log file
        my $mark_test_start = acp_log_mark(filer => $filer, acp_log_file_name => "acplog_master");

        #Run the test here
        my $test_result = test_run();

        #Mark the ACP Log file again
        my $mark_test_end = acp_log_mark(filer => $filer, acp_log_file_name => "acplog_master");

        #Get the matching logs from the ACP logs
    my $log_extract = acp_log_extract(filer => $filer, acp_log_file_name => "acplog_master", ips => ['198.15.1.4', '198.15.1.15'],
                        default_string => 0, include_strings => ["polllink"], acp_log_start_marker => $mark_test_start,
                        acp_log_end_marker => $mark_test_end);

    #Log extracted would have information related to the searched logs

    $log_extract = {
                         'count' => 2,
                         'records' => [
                                        {
                                          'acpp_serial_number' => 'NA',
                                          'inband_id' => '1a.1.B',
                                          'mac' => '00:50:cc:62:62:11'
                                          'timestamp' => 'Sun Dec 20 00:05:42 GMT 2009 ',
                                          'ip' => '198.15.1.4',
                                          'end_block' => 48,
                                          'start_block' => 42,
                                          'data' => [
                                                      'IMS9421400001AF; 0032225427029; 00000000000000; 0000000f; disc; 08; 62; polllink_thread_func:773 // ;2518',
                                                      'IMS9421400001AF; 0032225429029; 00000000000000; 0000000e; disc; 08; 65; flush_switch_routing_tab // ;2A79',
                                                      'IMS9421400001AF; 0032225429029; 00000000000000; 0000000f; disc; 08; 62; polllink_thread_func:773 // ;251C',
                                                      'IMS9421400001AF; 0032225651037; 00000000000000; 0000000e; disc; 08; 65; flush_switch_routing_tab // ;2A77'
                                                    ],
                                        },
                                        {
                                          'acpp_serial_number' => 'NA',
                                          'inband_id' => '1a.2.B',
                                          'mac' => '00:50:cc:72:62:11'
                                          'timestamp' => 'Sun Dec 20 00:07:25 GMT 2009 ',
                                          'ip' => '198.15.1.15',
                                          'end_block' => 61,
                                          'start_block' => 55,
                                          'data' => [
                                                      'IMS9421400001AF; 0032230051081; 00000000000000; 00000026; test; 10; 20; same ip - do nothing(428 // ;176B',
                                                      'IMS9421400001AF; 0032230051082; 00000000000000; 00000008; disc; 08; 37; bootp_client_recoverable // ;2A79',
                                                      'IMS9421400001AF; 0032230051082; 00000000000000; 0000001e; disc; 10; 72; write :acpa:c60f0156 acp // ;251C',
                                                      'IMS9421400001AF; 0032230051082; 00000000000000; 0000000c; disc; 20; 38; polllink_thread_func:773 // ;2A77'
                                                    ],
                                        },
                                      ]
                       };


=head1 ROUTINES

=cut

#Define Package name
package Storage::ACPLog;

#Define required packages and modules
use strict;
use warnings;
use Carp qw(confess croak);
use Params::Validate qw(validate validate_pos OBJECT SCALAR HASHREF ARRAYREF);
use Storage::Misc qw(filer_cmd);
use Tie::File;
use Volume qw(vol_root);
use mount;
use Sys::Hostname;
require Exporter;
use Data::Dumper;
use Tharn qw(say script_result host logcomment logresult $Log param);
use Utils qw(make_conn);
use NACL::MTask::SCP;
use NACL::C::Client;
use NACL::APISet;
use NACL::C::Node;

#Define exportable subroutines
BEGIN{
    use base qw(Exporter);
    our @EXPORT_OK = qw(
        acp_log_mount
        acp_log_umount
        acp_log_exist
        acp_log_mark
        acp_log_file_get_order
        acp_log_read_line
        acp_log_mark_only
        acp_log_read_first_line
        acp_log_read_next_line
        acp_log_generate_records
        acp_log_process_header
        acp_log_search_miis
        acp_log_search_element
        acp_log_search_strings
        acp_log_extract_records
        acp_log_search
        acp_log_search_all
        acp_log_print_records
        );
};

#Define begin block
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}

=head2 acp_log_mount

     my $acp_log_file = acp_log_mount({ filer => $filer, filer_conn => $filer_connection });

This routine connects to the filer, shares the root volume and mounts the acp log directory onto the local client and returns the absolute path of the acp log file.  If the subroutine encounters any error it shows a warning and returns 1.  The filer_conn, sudo_path and acp_log_file_name have default values rsh, /usr/local/bin and acplog_master respectively.

The following named arguments are accepted,

=over

=item filer

(Required) : The filer from which the logs have to be parsed.

=item filer_conn

(Optional) : The type of connection to be established with the Filer.  Default is RSH.

=item acp_log_file_name

(Optional) : The name of the acp log file to be parsed, Default is acplog_master, but could also be acplog

=item sudo_path

(Optional) : The path for the sudo command on the client. Default is /usr/local/bin

=back

=cut

#########################################################################################################
#
# Subroutine to mount the filer root volume.
# The root volume is mounted to /t/<filer_name>/root and the path for the acp log file would be returned.
#
#########################################################################################################
sub acp_log_mount {
        my (%opts) = validate(@_, {
                                filer => { type => SCALAR },
                                filer_conn => { type => SCALAR, optional => 1, default => "#rsh"},
                                sudo_path => { type => SCALAR, optional => 1, default => "/usr/local/bin"},
                acp_log_file_name => { type => SCALAR, optional => 1, default => "acplog_master"}
                                });
        logcomment("acp_log_mount: Mounting the acp log directory from filer");

    #Retrieve filer and connection details
        my $filer = $opts{filer};
        my $filer_conn = $opts{filer_conn};
        my $acp_log_file_name = $opts{acp_log_file_name};

    #Warn if correct acp log file name is not specified
        if($acp_log_file_name ne "acplog" && $acp_log_file_name ne "acplog_master")
        {
            $Log->warn("acp_log_mount: \"$acp_log_file_name\" doesn't appear like ACP log file");
        }
        my $filer_ci;
        my $client=hostname();
        my @clienta = split(/\./,$client);
        my $CLIENT=$clienta[0];
        logcomment("Hostname is $CLIENT");
        my $unix_ci = NACL::C::Client->new(client=> $CLIENT);
        $filer_ci = NACL::C::Node->find();
        my $scp=NACL::MTask::SCP->new(source_command_interface => $unix_ci,
                                        target_command_interface => $filer_ci,
                                        set_directory_on_source  => '/x/eng/cores',
                                    );
        my $files=["/mroot/etc/log/acp/$acp_log_file_name"];
        if ( $filer_ci->is_cmode() ) {
            $scp->copy_from_target_to_source('files_to_copy'=>$files);
            $Log->comment("The file /x/eng/cores/$acp_log_file_name is copied from filer $filer");
            my $acp_log_files = "/x/eng/cores/$acp_log_file_name";
            $acp_log_file_name = [$acp_log_files];
            logcomment ("acp_log_mount: ACP log file path is $acp_log_file_name");
        } else {
        #Retrieve the current client hostname and log into the client machine for mounting the filer
            my $client_hostname = hostname();
            my ($client) = split(/\./,$client_hostname);

            my $sudo = "$opts{sudo_path}"."/sudo";

            my $client_connection = make_conn($client, "#rlogin");
            my $say_timeout = 180;

            my $filer_connection = make_conn($filer, $filer_conn);
            my $filer_root_vol = vol_root($filer_connection);

        #Export root volume for mouting the root volume on the clients
            say($filer_connection, "priv set -q diag; exportfs -p sec=sys,rw,anon=0 /vol/$filer_root_vol", '-timeout',  $say_timeout );
            my $mount_flag = 1;
            my $mountpoint = mount ($filer,"/",client=>$client)
                                                        or $mount_flag =0;

            #Return error if the mount is not successful
            if($mount_flag == 0){
                $Log->warn ("acp_log_mount: Could not mount $filer on $client");
                return 1;
            }
        #Add acp log file with path to generate the absolute path
            logcomment ("acp_log_mount: ACP log file mount point is $mountpoint");
            $acp_log_file_name = "$mountpoint/etc/log/acp/$acp_log_file_name";
            logcomment ("acp_log_mount: ACP log file path is $acp_log_file_name");
        }  ## end of if ( $filer_ci->is_cmode() )
        return $acp_log_file_name;
}

=head2 acp_log_umount

     my $result = acp_log_umount({ acp_log_file_name => $acp_log_file_name });

This routine unmounts the Filer root volume from the mount point of the local client. This uses the absolute log path given by the acp_log_mount routine.  It returns a 0 on successful unmount and returns 1 if unmount is unsuccessful.

The following named arguments are accepted,

=over

=item acp_log_file_name

(Required) : The absolute path of the ACP log file.

=cut


#########################################################################################################
#
# Subroutine to unmount the filer root volume from the client.
#
#########################################################################################################

sub acp_log_umount
{
        my (%opts) = validate(@_, {
                                acp_log_file_name => { type => SCALAR },
                                });

        logcomment("acp_log_umount: Unmounting the ACP log file directory");

    #Retrieve the current client hostname and log into the client machine for unmouting the filer
    my $acp_log_file_name = $opts{acp_log_file_name};
        my $client_hostname = hostname();
        my ($client) = split(/\./,$client_hostname);
        my $client_connection = make_conn($client,"#rlogin");
        my $say_timeout = 180;

    #Get the mount point and unmount the filer from client
        my ($mount_point) = split( /\/etc\/log\/acp/, $acp_log_file_name);
        my $return_value = umount ( $mount_point, client => $client );

        if((defined $return_value) && ($return_value eq "1"))
        {
                logcomment ("acp_log_umount: Unmount successful");
                return 0;
        }
        else
        {
                logcomment ("acp_log_umount: Unmount failed");
                return 1;
        }
}

=head2 acp_log_mark

     my $acp_log_start_mark = acp_log_mark({ filer => $filer, filer_conn => $filer_connection });

This routine connects to the filer, shares the root volume, mounts the acp log directory onto the local client, checks the status of the ACP log file and returns a hash reference providing the details of the ACP log file and current line number.

The returned hash reference would have following parameters
1. no_file flag      : set if the ACP log file doesnt exist
2. line_number       : the last line number of the log file being marked
3. acp_log_file      : the current log file name
4. acp_log_reference : the first time stamp found in the ACP log file that would be used while parsing to check if the log file is rotated

Example Return Value:
    {
               no_file => '0',
               line_number => '0',
               acp_log_file => '/t/barnum/root/etc/log/acp/acplog_master',
               acp_log_reference => '*Fri Dec 25 23:58:14 GMT 2009; ACP trasaction cmd timeout(ms) value is: 5000',
       };

The following named arguments are accepted,

=over

=item filer

(Required) : The filer from which the logs have to be parsed

=item filer_conn

(Optional) : The type of connection to be established with the Filer.  Default is RSH

=item acp_log_file_name

(Optional) : The name of the acp log file to be parsed, Default is acplog_master, but could also be acplog

=item sudo_path

(Optional) : The path for the sudo command on the client. Default is /usr/local/bin

=item timeout

(Optional) : The timeout value for checking the number of lines.  Default is 180 seconds

=back

=cut

#########################################################################################################
#
# Subroutine to mount the filer and mark the current position of acp log file.
#
# The return value would be a hash reference with flags to indicate if no file existed or the file details
# like the line nmber and refernce time stamp from the log file for checking if the log file got rotated.
#
# Return value:
#    {
#               no_file => '0',
#               line_number => '0',
#               acp_log_file => '/t/barnum/root/etc/log/acp/acplog_master',
#               acp_log_reference => '*Fri Dec 25 23:58:14 GMT 2009; ACP trasaction cmd timeout(ms) value is: 5000',
#       };
#
#########################################################################################################

sub acp_log_mark {
        my (%opts) = validate(@_, {
                                filer => { type=>SCALAR },
                                filer_conn => { type => SCALAR, optional => 1},
                                sudo_path => { type => SCALAR, optional => 1},
                acp_log_file_name => { type => SCALAR, optional => 1},
                timeout => { type => SCALAR, optional => 1, default => 180},
                                });

        logcomment("acp_log_mark: Marking the current line from the acp log file");

    #Mount the filer root volume and get acp log file path from the filer
    my $parameters = {};
    my $timeout = $opts{timeout} ;
    $parameters->{filer} = $opts{filer};

    if(defined $opts{filer_conn})
    {
        $parameters->{filer_conn} = $opts{filer_conn};
    }

    if(defined $opts{sudo_path})
    {
        $parameters->{sudo_path} = $opts{sudo_path};
    }

    if(defined $opts{acp_log_file_name})
    {
        $parameters->{acp_log_file_name} = $opts{acp_log_file_name};
    }

    my $acp_log_file = acp_log_mount($parameters);
    logcomment("acp_log_mark: Current log file is $acp_log_file");

    #Try to retrieve the line count from the acp log file to set the marker
    my $client_hostname = hostname();
    my ($client) = split(/\./,$client_hostname);
    my $connection = make_conn($client, "#rsh");
    my (undef,undef,$result) = say($connection, "wc -l $acp_log_file", -timeout, $timeout);
    my $rvalue;

    #Check the response, if the acp log file doesnt exist set $no_file flag to 1 and return
    if ($result =~ /No such file or directory/)
    {
        $rvalue->{no_file} = 1;
        $rvalue->{line_number} = 0;
        $rvalue->{acp_log_reference} = 0;
        $Log->warn("acp_log_mark: ACP log file was not found");
        return $rvalue;
    }
    my ($count) = grep (/$acp_log_file/, split(/\r?\n/, $result));
    $count =~ s/(\s*)(\d+)(.*)$/$2/g;

    #Retrieve the first time stamp available in the ACP log file and store it for reference
    #This could be used at a later stage to find if the log file has been rotated.

    $rvalue->{acp_log_file} = $acp_log_file;
    $rvalue->{no_file} = 0;
    if($count !~ /\d+/)
    {
        $rvalue->{line_number} = 0;
    }
    else
    {
        $rvalue->{line_number} = $count;
    }

    my @acp_log_array;
    use Fcntl 'O_RDONLY';
    tie @acp_log_array, 'Tie::File', $acp_log_file, mode => O_RDONLY, autochomp => 0;

    for(my $i = 0; $i <= $#acp_log_array; $i++)
    {
        my $line = $acp_log_array[$i];
        chomp($line);
        if($acp_log_file =~ /acplog_master$/)
        {
            if($line =~ /^\*\w{3} \w{3} \d{1,2} \d{2}:\d{2}:\d{2} \w{3} \d{4}/)
            {
                $rvalue->{acp_log_reference} = $line;
                last;
            }
        }
        else
        {
            if($line =~ /\w{3} \w{3} \d{1,2} \d{2}:\d{2}:\d{2} \w{3} \d{4}/)
            {
                $rvalue->{acp_log_reference} = $line;
                last;
            }
        }
    }
    untie @acp_log_array;

    #Report warning if no reference time stamp is found
    if(!(defined $rvalue->{acp_log_reference}))
    {
        $Log->warn("acp_log_mark: Reference timestamp for marking the file not found, if this is used as start marker then only one file would be considered for processing");
        $rvalue->{acp_log_reference} = "0";
    }

    #Unmount the filer root volume from the client
    acp_log_umount({acp_log_file_name => $acp_log_file});
    return $rvalue;
}


=head2 acp_log_file_get_order

     my $ordered_list = acp_log_file_get_order({ acp_logs => $acp_logs });

This routine accepts the acp_logs data structure which hosts all information regarding the ACP log parsing parameters.  It checks if the acp_log_start_marker and acp_log_end_marker are defined and tries to get the order of the acp log files if the files are rotated based on the acp_log_reference and returns an array reference to the list of relevant ACP log files.

The following named arguments are accepted,

=over

=item acp_logs

(Required) : The acp_logs data structure with acp_log_start_marker and acp_log_end_marker

=back

=cut

#########################################################################################################
#
# Subroutine to get the list of acp log files in order if the files have been rotated
# between start and end markers
#
#########################################################################################################

sub acp_log_file_get_order {

    my (%opts) = validate(@_, {
            acp_logs => { type => HASHREF}
            });

    logcomment("acp_log_file_get_order: Checking the log file order");

    #Get the values of start and end markers
    my $acp_logs = $opts{acp_logs};
    my $acp_log_start_marker;
    my $acp_log_end_marker;

    #Handle unexpected conditions for start marker, end marker and start marker having no files defined.
    if(defined $acp_logs->{acp_log_start_marker})
    {
        $acp_log_start_marker = $acp_logs->{acp_log_start_marker};
    }
    else
    {
        logresult("FATAL","acp_log_file_get_order: Start marker not defined in acp_logs");
    }

    if(defined $acp_logs->{acp_log_end_marker})
    {
        $acp_log_end_marker = $acp_logs->{acp_log_end_marker};
    }

    if (!defined($acp_log_start_marker->{acp_log_reference})) {
        $acp_log_start_marker->{acp_log_reference} = 0;
    }

    my $ordered_list = ();
    if($acp_log_start_marker->{no_file} == 1)
    {
        $Log->warn("acp_log_file_get_order: No file defined in start marker, current ACP log file would be parsed");
        my $rvalue = [$acp_log_start_marker->{acp_log_file}];
        return $rvalue;
    }

    if($acp_log_start_marker->{acp_log_reference} eq "0")
    {
        $Log->warn("acp_log_file_get_order: Start marker doesnt have a reference, current ACP log file would be taken for parsed");
        my $rvalue = [$acp_log_start_marker->{acp_log_file}];
        return $rvalue;
    }

    #Generate an array of all the log file names that could be seen on the filer
    my $acp_log_files = [$acp_log_start_marker->{acp_log_file}];
    for(my $i = 0; $i <= 10; $i++)
    {
        my $next_file = $acp_log_start_marker->{acp_log_file}.'.'.$i ;
        push @$acp_log_files, $next_file;
    }

    #Run through each file to find the time reference recorded in start marker.
    #Return the file names that are between the current log file and
    #the log file on which the timestamp is seen.  If the timestamp is not
    #found in any file, return 1 with log result as fail.

    my $match_flag = 0;
    for(my $i = 0; $i <= $#$acp_log_files; $i++)
    {
        my @acp_log_array;
        use Fcntl 'O_RDONLY';
        tie @acp_log_array, 'Tie::File', $$acp_log_files[$i], mode => O_RDONLY, autochomp => 0;
        #print "$$acp_log_files[$i]\n";
        for(my $j = 0; $j <= $#acp_log_array; $j++)
        {
            my $line = $acp_log_array[$j];
            chomp($line);
            #print "$line\n";
            if($$acp_log_files[$i] =~ /acplog_master$|acplog_master\.\d$/)
            {
                if($line =~ /^\*\w{3}\s\w{3}\s+\d{1,2}\s\d{2}:\d{2}:\d{2}\s\w{3}\s\d{4}/)
                {
                    if($line eq $acp_log_start_marker->{acp_log_reference})
                    {
                        #print "$line\n";
                        $match_flag = 1;
                        push @$ordered_list, $$acp_log_files[$i];
                    }
                    last;
                }
            }
            else
            {
                if($line =~ /\w{3}\s\w{3}\s+\d{1,2}\s\d{2}:\d{2}:\d{2}\s\w{3}\s\d{4}/)
                {
                    if($line eq $acp_log_start_marker->{acp_log_reference})
                    {
                        #print "$line\n";
                        $match_flag = 1;
                        push @$ordered_list, $$acp_log_files[$i];
                    }
                    last;
                }
            }
        }
        untie @acp_log_array;
        #print "\n";
        if($match_flag == 1)
        {
            last;
        }
        push @$ordered_list, $$acp_log_files[$i];
    }
    @$ordered_list = reverse(@$ordered_list);

    if($match_flag != 1)
    {
        logresult("FAIL","Unable to find the file marked by acp_log_start_marker");
        return 1;
    }

    if(($#$ordered_list == 0) && ($acp_log_start_marker->{line_number} == $acp_log_end_marker->{line_number}))
    {
        logcomment("acp_log_file_get_order: ACP log file is not updated with any data");
    }
    $acp_logs->{acp_log_start_marker}->{acp_log_file} = $$ordered_list[0];
    return $ordered_list;
}


=head2 acp_log_exist

     my $result = acp_log_exist({ filer => $filer, filer_conn => $filer_connection });

This routine checks for the acp log files' existance and updates acp_logs data structure with the log file names, number of lines in each file and virtual file name line numbers for the log files (Please refer to acp_log_read_line for help on virtual file).

It returns 0 if files are accessible and returns 1 if there is any problem reading the files.

Reference acp_logs datastructure:
         {
          'acp_log_end_marker' => {
                                    'no_file' => 0,
                                    'line_number' => '0',
                                    'acp_log_file' => '/t/barnum/root/etc/log/acp/acplog_master',
                                    'acp_log_reference' => '*Fri Dec 25 23:58:14 GMT 2009; ACP trasaction cmd timeout(ms) value is: 5000'
                                  },
          'total_files' => '5',
          'total_lines' => 893017,
          'acp_log_files' => [
                               {
                                 'end_line' => 273,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master.3',
                                 'start_line' => 0,
                                 'line_count' => 274
                               },
                               {
                                 'end_line' => 205088,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master.2',
                                 'start_line' => 274,
                                 'line_count' => 204815
                               },
                               {
                                 'end_line' => 619783,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master.1',
                                 'start_line' => 205089,
                                 'line_count' => 414695
                               },
                               {
                                 'end_line' => 892997,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master.0',
                                 'start_line' => 619784,
                                 'line_count' => 273214
                               },
                               {
                                 'end_line' => 893016,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master',
                                 'start_line' => 892998,
                                 'line_count' => 19
                               }
                             ],
          'acp_log_start_marker' => {
                                      'no_file' => 0,
                                      'line_number' => '0',
                                      'acp_log_file' => '/t/barnum/root/etc/log/acp/acplog_master',
                                      'acp_log_reference' => '*Fri Dec 25 23:58:14 GMT 2009; ACP trasaction cmd timeout(ms) value is: 5000'
                                    },
        };

The following named arguments are accepted,

=over

=item acp_logs

(Required) : The reference of acp_logs data structure

=item filer

(Optional) : The filer from which the logs have to be parsed

=item acp_log_files

(Optional) : List of acp log files that need to be processed if the filer is not specified

=item timeout

(Optional) : The timeout value for checking the number of lines.  Default is 180 seconds

=back

=cut

#########################################################################################################
#
# Subroutine to check and return details of the log files specified in acp_logs structure
#
#########################################################################################################

sub acp_log_exist
{
    my (%opts) = validate(@_, {
            acp_log_files => { type => ARRAYREF, optional => 1},
            filer => {type => SCALAR, optional => 1},
            timeout => { type => SCALAR, optional => 1, default => 180},
            acp_logs => { type => HASHREF }
            });

    logcomment("acp_log_exist: Checking ACP log files for existence and accessibility");
    my $flag = 0;
    my $acp_logs = $opts{acp_logs};

    #Retrieve the list of log files for parsing either from the filer or from a list of log files from the command line
    my @acp_log_files;
    if(defined $opts{filer})
    {
        if(defined $acp_logs->{acp_log_start_marker})
        {
            my $filer_acp_log_files = acp_log_file_get_order({acp_logs => $acp_logs});
            @acp_log_files = @$filer_acp_log_files;
        }
        else
        {
            $Log->warn("acp_log_exist: acp log start marker is not defined, current ACP log file would be parsed");
            @acp_log_files = @{$opts{acp_log_files}};
        }
    }
    else
    {
        @acp_log_files = @{$opts{acp_log_files}};
    }

    #Check if the log files are defined
    if($#acp_log_files < 0 )
    {
        logresult ("FAIL","acp_log_exist: No log files are defined");
        return 1;
    }

    #Check if the files exist and have read permissions
    foreach my $file (@acp_log_files)
    {
        if ((-e $file) && (-r $file))
        {
            logcomment "acp_log_exist: File  : $file\nStatus: OK";
        }
        else
        {
            logcomment "acp_log_exist: File  : $file\nStatus: NOK";
            $flag = 1;
        }
    }

    if($flag == 1)
    {
        logresult("FAIL","acp_log_exist: Please check the log files for existence and accessibility");
        return 2;
    }

    my $client_hostname = hostname();
    my ($client) = split(/\./,$client_hostname);
    my $connection = make_conn($client, "#rsh");
    my @log_references = ();

    #Retrieve the number of lines the acp log files have and build the acp_logs data structure
    #The data structure would host the important parameter like acp log files, their start and
    #end line numbers, start and end markers etc.
    for(my $i = 0; $i <= $#acp_log_files; $i++)
    {
        my $log_file_details = {};
        my $timeout = $opts{timeout};

        my (undef,undef,$result) = say($connection, "wc -l $acp_log_files[$i]", -timeout, $timeout);
        if ($result =~ /No such file or directory/)
        {
            return 0;
        }
        my ($count) = grep (/$acp_log_files[$i]/, split(/\r?\n/, $result));
        $count =~ s/(\s*)(\d+)(.*)$/$2/g;
        if($count !~ /\d+/)
        {
            $flag = 1;
        }

        $log_file_details->{name} = $acp_log_files[$i];
        $log_file_details->{line_count} = $count;
        $log_references[$i] = $log_file_details;
    }
    $acp_logs->{acp_log_files} = \@log_references;
    $acp_logs->{total_files} = $#{$acp_logs->{acp_log_files}} + 1;

    for(my $i = 0; $i <= $#acp_log_files; $i++)
    {
        if($i == 0)
        {
            $acp_logs->{acp_log_files}[$i]->{start_line} = 0;
        }
        else
        {
            $acp_logs->{acp_log_files}[$i]->{start_line} = $acp_logs->{acp_log_files}[$i - 1]->{end_line} + 1;
        }

        $acp_logs->{acp_log_files}[$i]->{end_line} = $acp_logs->{acp_log_files}[$i]->{start_line} + $acp_logs->{acp_log_files}[$i]->{line_count} - 1;
    }

    $acp_logs->{total_lines} = 0;
    for(my $i = 0; $i <= $#acp_log_files; $i++)
    {
        $acp_logs->{total_lines} = $acp_logs->{total_lines} + $acp_logs->{acp_log_files}[$i]->{line_count};
    }

    if($flag == 1)
    {
        logresult ("FAIL","acp_log_exist: Please check the log files exist and have read permissions");
        return 1;
    }
    return 0;
}


=head2 acp_log_read_line

     my $line = acp_log_read_line({ acp_logs => $acp_logs, line => $line });

This routine reads and returns a line from the ACP log files.  This routine conceals the handling of multiple log files.  For a user the file is shown as a single large concatenated log file from the start marker till the end marker.  When any line is read the line number is converted from the virtual line number to actual log file and line number.

Virtual acp log file


                |       virtual single acp log file from start marker to end marker       |
                |+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|
       |        !      |                         |                          |             !               |                    |
       |        !<= acp_log_end_marker           |                          |             !<= acp_log_start_marker             |
       |        !      |                         |                          |             !               |                    |
       |---------------|-------------------------|--------------------------|-----------------------------|--------------------|
       | acplog_master |   acplog_master.0       |     acplog_master.1      |     acplog_master.2         |  acplog_master.3   |


On calling this routine with a line number, the line number would be translated to the corresponding file and actual line number and line is fetched from the appropriate log file.  If the line number falls below or above the start marker or the end marker respectively an undef is returned.  The routine also updates the current file position and current line number in the acp_logs data structure.

Example:

In the following example, if the routine is called with line=300, then line 26 from second file '/t/barnum/root/etc/log/acp/acplog_master.2' would be retrieved.

Reference acp_logs data structure:
         {
          'acp_log_end_marker' => {
                                    'no_file' => 0,
                                    'line_number' => '0',
                                    'acp_log_file' => '/t/barnum/root/etc/log/acp/acplog_master',
                                    'acp_log_reference' => '*Fri Dec 25 23:58:14 GMT 2009; ACP trasaction cmd timeout(ms) value is: 5000'
                                  },
          'total_files' => '5',
          'total_lines' => 893017,
          'acp_log_files' => [
                               {
                                 'end_line' => 273,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master.3',
                                 'start_line' => 0,
                                 'line_count' => 274
                               },
                               {
                                 'end_line' => 205088,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master.2',
                                 'start_line' => 274,
                                 'line_count' => 204815
                               },
                               {
                                 'end_line' => 619783,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master.1',
                                 'start_line' => 205089,
                                 'line_count' => 414695
                               },
                               {
                                 'end_line' => 892997,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master.0',
                                 'start_line' => 619784,
                                 'line_count' => 273214
                               },
                               {
                                 'end_line' => 893016,
                                 'name' => '/t/barnum/root/etc/log/acp/acplog_master',
                                 'start_line' => 892998,
                                 'line_count' => 19
                               }
                             ],
          'current_file_pos' => 0,
          'acp_log_start_marker' => {
                                      'no_file' => 0,
                                      'line_number' => '0',
                                      'acp_log_file' => '/t/barnum/root/etc/log/acp/acplog_master',
                                      'acp_log_reference' => '*Fri Dec 25 23:58:14 GMT 2009; ACP trasaction cmd timeout(ms) value is: 5000'
                                    },
          'current_line' => 1
        };

The following named arguments are accepted,

=over

=item acp_logs

(Required) : The reference of acp_logs data structure

=item line

(Required) : The line number from the virtual single log file

=back

=cut

#########################################################################################################
#
# Subrotine to read a line from the acp log files.  The management of information retrieval from all the acp log files
# is handled by this subroutine and the other subroutines are shown a single virtual log file.
#
# Virtual acp log file information
#
#                |       virtual single acp log file from start till the end marker        |
#                |+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|
#       |        !      |                         |                          |             !               |                    |
#       |        acp_log_end_marker               |                          |             acp_log_start_marker                 |
#       |        !      |                         |                          |             !               |                    |
#       |---------------|-------------------------|--------------------------|-----------------------------|--------------------|
#       | acplog_master |   acplog_master.0       |     acplog_master.1      |     acplog_master.2         |  acplog_master.3   |
#
#
#
#########################################################################################################

sub acp_log_read_line {
    my(%opts) = validate(@_, {
            acp_logs => { type=>HASHREF },
            line => { type=>SCALAR }
            });
    my $acp_logs = $opts{acp_logs};
    my $current_line = $opts{line} - 1;
    my $temp = $acp_logs->{acp_log_files};
    my @acp_log_files = @$temp;
    my $current_file_pos;


    #Donot return any line if the line number requested is outside the scope of the log files specified
        if(($current_line < 0) || ($current_line >= $acp_logs->{total_lines}))
        {
                return undef;
        }

    #Calcule and set $current_file_pos to the correct file name to read the line from the line number requested
        for(my $i = 0; $i <= $#acp_log_files; $i++)
        {
                if(($current_line >= $acp_log_files[$i]->{start_line}) && ($current_line <= $acp_log_files[$i]->{end_line}))
                {
                        $current_file_pos = $i;
                }
        }
        #print "current_file_pos = $current_file_pos\n";

    #Handle different errors encountered
        if($current_file_pos !~ /\d+/)
        {
                $Log->warn("acp_log_read_line: Script is trying to access line outside the scope of files: Dumping current data from variables");
                print Dumper($acp_logs);
                print "\ncurrent line: $current_line\n";
                print "current_file_pos: $current_file_pos\n";
                logresult("Script","acp_log_read_line: Encountered an unexpected error(1).  Please contact the author with the logs");
                return undef;
        }

        #Move the $current_file_pos and the file handle fh to appropriate files based on the line number.
    #All the variables are stored as references in data stracure acp_logs.
    if(defined $acp_logs->{current_file_pos})
        {
                if($acp_logs->{current_file_pos} != $current_file_pos)
                {
                        if(defined $acp_logs->{fh})
                        {
                                my @log_array = @{$acp_logs->{fh}};
                                untie @log_array;
                                delete $acp_logs->{fh};
                        }
                        my @log_array;
                        $acp_logs->{current_file_pos} = $current_file_pos;
                        use Fcntl 'O_RDONLY';
                        tie @log_array, 'Tie::File', $acp_logs->{acp_log_files}[$acp_logs->{current_file_pos}]->{name}, mode => O_RDONLY, autochomp => 0;
                        $acp_logs->{fh} = \@log_array;
                }
                else
                {
                        if(!(defined $acp_logs->{fh}))
                        {
                                my @log_array;
                                $acp_logs->{current_file_pos} = $current_file_pos;
                                use Fcntl 'O_RDONLY';
                                tie @log_array, 'Tie::File', $acp_logs->{acp_log_files}[$acp_logs->{current_file_pos}]->{name}, mode => O_RDONLY, autochomp => 0;
                                $acp_logs->{fh} = \@log_array;
                        }
                }
        }
        else
        {
                if(defined $acp_logs->{fh})
                {
                        my @log_array = @{$acp_logs->{fh}};
                        untie @log_array;
                        delete $acp_logs->{fh};
                }
                my @log_array;
                $acp_logs->{current_file_pos} = $current_file_pos;
                use Fcntl 'O_RDONLY';
                tie @log_array, 'Tie::File', $acp_logs->{acp_log_files}[$acp_logs->{current_file_pos}]->{name}, mode => O_RDONLY, autochomp => 0;
                $acp_logs->{fh} = \@log_array;
        }

    #If start marker is defined, return no lines until the requested line number is greater than acp log start marker
        if(defined $acp_logs->{acp_log_start_marker})
        {
                if($acp_logs->{acp_log_files}[$acp_logs->{current_file_pos}]->{name} eq $acp_logs->{acp_log_start_marker}->{acp_log_file})
                {
                        if(($current_line + 1) <= $acp_logs->{acp_log_start_marker}->{line_number})
                        {
                                return undef;
                        }
                }
        }

        my $log_array = $acp_logs->{fh};
        $acp_logs->{current_line} = $current_line + 1;

    #Translate/Calculate the correct line number in the particular log file to be read from the line number
    #of the overall virtual log file

        for(my $i = 0; $i < $acp_logs->{current_file_pos}; $i++)
        {
                $current_line = $current_line - $acp_logs->{acp_log_files}[$i]->{line_count};
        }

    #If end marker is defined, return no lines until the read line is equal to or less than acp log end marker
        if(defined $acp_logs->{acp_log_end_marker})
        {
                if($acp_logs->{acp_log_files}[$acp_logs->{current_file_pos}]->{name} eq $acp_logs->{acp_log_end_marker}->{acp_log_file})
                {
                        if(($current_line + 1)> $acp_logs->{acp_log_end_marker}->{line_number})
                        {
                                return undef;
                        }
                }
        }

    #Handle errors like if the current line that should be read is out of scope of all the log files
        if($current_line > $#$log_array)
        {
                if($acp_logs->{current_file_pos} < ($acp_logs->{total_files} - 1))
                {
                        $Log->warn("acp_log_read_line: Script is trying to access line outside the scope of files: Dumping current data from variables");
                        print Dumper($acp_logs);
                        print "\ncurrent_line: $current_line\n";
                        print "current_file_pos: $current_file_pos\n";
                        logresult("Script","acp_log_read_line: Encountered an unexpected error(2).  Please contact the author with the logs");
                        return undef;
                }
                else
                {
                        return undef;
                }
        }
        return $$log_array[$current_line];
}

=head2 acp_log_read_first_line

     my $line = acp_log_read_first_line({ acp_logs => $acp_logs });

This routine reads the first line from the ACP log files.  It initializes the values for the current file position and first line to be read and returns the first line from the ACP log file.  This routine would set the current line to first line or to the line pointed to by the start marker and returns the line.

The following named arguments are accepted,

=over

=item acp_logs

(Required) : The reference of acp_logs data structure

=back

=cut

#########################################################################################################
#
# Subrotine to read first line from the acp log files.  This would initialize the different parameters
# of the acp_logs data structure.
#
#########################################################################################################

sub acp_log_read_first_line {
        my (%opts) = validate(@_, {
                                acp_logs => { type => HASHREF}
                                });

    #Set the current line number to 1 or as specified by the acp log start marker
        my $acp_logs = $opts{acp_logs};
        my $current_line = 1;

        if(defined $acp_logs->{acp_log_start_marker})
        {
        if(defined $acp_logs->{acp_log_start_marker}->{line_number})
            {
                    $current_line = $acp_logs->{acp_log_start_marker}->{line_number} + 1;
            }
    }

        #Read the line and return to the caller
    my $line;
        if(!($line = acp_log_read_line({acp_logs => $acp_logs, line => $current_line})))
        {
                return undef;
        }
        else
        {
                return $line;
        }
}

=head2 acp_log_mark_only

     my $acp_log_end_mark = acp_log_mark_only({ acp_log_file => $acp_log_file });

This routine is same as acp_log_mark, but doesnt attempt to mount the filer.  The input parameter would be the acp log file path for marking.

The returned hash reference would have following parameters
1. no_file flag      : set if the ACP log file doesnt exist
2. line_number       : the last line number of the log file being marked
3. acp_log_file      : the current log file name
4. acp_log_reference : the first time stamp found in the ACP log file that would be used while parsing to check if the log file is rotated

Example Return Value:
    {
               no_file => '0',
               line_number => '0',
               acp_log_file => '/t/barnum/root/etc/log/acp/acplog_master',
               acp_log_reference => '*Fri Dec 25 23:58:14 GMT 2009; ACP trasaction cmd timeout(ms) value is: 5000',
       };

The following named arguments are accepted,

=over

=item acp_log_file

(Required) : The acp log file absolute path

=item timeout

(Optional) : The timeout value for checking the number of lines.  Default is 180 seconds

=back

=cut


#########################################################################################################
#
# Subroutine to mark the current position of acp log file.
#
# The return value would be a hash reference with flags to indicate if no file existed or the file details
# like the line nmber and refernce time stamp from the log file for checking if the log file got
# rotated.
#
# Return value:
#    {
#               no_file => '0',
#               line_number => '0',
#               acp_log_file => '/t/barnum/root/etc/log/acp/acplog_master',
#               acp_log_reference => '*Fri Dec 25 23:58:14 GMT 2009; ACP trasaction cmd timeout(ms) value is: 5000',
#       };
#
#########################################################################################################

sub acp_log_mark_only {
        my (%opts) = validate(@_, {
                                acp_log_file => { type=>SCALAR },
                                timeout => { type => SCALAR, optional => 1, default => 180},
                                });
        logcomment("acp_log_mark_only: Marking the current line from the log file");
    my $timeout = $opts{timeout};
        my $acp_log_file = $opts{acp_log_file};
        my $client_hostname = hostname();
        my ($client) = split(/\./,$client_hostname);
        my $connection = make_conn($client, "#rsh");

        my $change_flag = 1;
        my $change_count = 0;
        my $old_count = -100;
        my $rvalue;

    #Try to retrieve the line count from the acp log file to set the marker
    #If the read file number is not consistent for two reads successively for 5 times,
    #report a warning and get the line number
        while(($change_flag == 1) && ($change_count < 5))
        {
                my (undef,undef,$result) = say($connection, "wc -l $acp_log_file", -timeout, $timeout);

                if ($result =~ /No such file or directory/)
                {
                        $rvalue->{no_file} = 1;
                        $rvalue->{line_number} = 0;
                        $rvalue->{log_reference} = 0;
            $Log->warn("acp_log_mark_only: ACP log file was not found");
                        return $rvalue;
                }
                my ($count) = grep (/$acp_log_file/, split(/\r?\n/, $result));
                $count =~ s/(\s*)(\d+)(.*)$/$2/g;

                $rvalue->{acp_log_file} = $acp_log_file;
                $rvalue->{no_file} = 0;
                if($count !~ /\d+/)
                {
                        $rvalue->{line_number} = 0;
                }
                else
                {
                        $rvalue->{line_number} = $count;
                        if($count != $old_count)
                        {
                                $change_flag = 1;
                                $change_count++;
                                $old_count = $count;
                        }
                        else
                        {
                                $change_flag = 0;
                        }
                }
                sleep 1;
        }

        $rvalue->{change_count} = $change_count;

    #Get the reference for the log file by retrieving the first line having a timestamp
        my @acp_log_array;
        use Fcntl 'O_RDONLY';
        tie @acp_log_array, 'Tie::File', $acp_log_file, mode => O_RDONLY, autochomp => 0;

        for(my $i = 0; $i <= $#acp_log_array; $i++)
        {
                my $line = $acp_log_array[$i];
                chomp($line);
                if($acp_log_file =~ /acplog_master$/)
        {
            if($line =~ /^\*\w{3} \w{3} \d{1,2} \d{2}:\d{2}:\d{2} \w{3} \d{4}/)
                    {
                            $rvalue->{acp_log_reference} = $line;
                            last;
                    }
        }
        else
        {
            if($line =~ /\w{3} \w{3} \d{1,2} \d{2}:\d{2}:\d{2} \w{3} \d{4}/)
            {
                $rvalue->{acp_log_reference} = $line;
                last;
            }
        }
        }
        untie @acp_log_array;

        if(!(defined $rvalue->{acp_log_reference}))
        {
                logcomment("acp_log_mark_only: Reference timestamp for marking the file not found, if this is used as start marker then only one file would be considered for processing");
                $rvalue->{acp_log_reference} = "0";
        }

        return $rvalue;
}


=head2 acp_log_read_next_line

     my $line = acp_log_read_next_line({ acp_logs => $acp_logs });

This routine increments the current_line parameters in acp_logs data structure and returns the next line from the acp log file.  It returns an undef is the line is outside the scope of start marker or end marker.

The following named arguments are accepted,

=over

=item acp_logs

(Required) : The reference of acp_logs data structure

=back

=cut


#########################################################################################################
#
# Subroutine to read the next line from the acp log file
#
#########################################################################################################

sub acp_log_read_next_line {
        my (%opts) = validate(@_, {
                                acp_logs => {type => HASHREF}
                                });
        my $acp_logs = $opts{acp_logs};
        #my $current_line = $acp_logs->{current_line} + 1;

    #Increment the log file current line value in the acp_logs data strcture and retrieve the line
        $acp_logs->{current_line}++;
        my $current_line = $acp_logs->{current_line};
        my $line;

        if(!($line = acp_log_read_line({acp_logs => $acp_logs, line => $current_line})))
        {
                return undef;
        }
        else
        {
                return $line;
        }
}

=head2 acp_log_read_current_line

     my $line = acp_log_read_current_line({ acp_logs => $acp_logs });

This routine returns the line represented by current_line in acp_logs data structure.  If the line is outside the scope of the start or end marker, an undef is returned.

The following named arguments are accepted,

=over

=item acp_logs

(Required) : The reference of acp_logs data structure

=back

=cut



#########################################################################################################
#
# Subroutine to read the current line pointed to by current_line in $acp_logs data structure
#
#########################################################################################################

sub acp_log_read_current_line {
        my (%opts) = validate(@_, {
                                acp_logs => { type=>HASHREF }
                                });
        my $acp_logs = $opts{acp_logs};
        my $line;
        if(!($line = acp_log_read_line({acp_logs => $acp_logs, line => $acp_logs->{current_line}})))
        {
                return undef;
        }
        else
        {
                return $line;
        }
}


=head2 acp_log_generate_records

     my $array_ref = acp_log_generate_records({ acp_logs => $acp_logs });

This routine reads the ACP log file and generates an array of hash references to the logs of each acpp module.  The hash references would be arranged in the same order as in the ACP log file and would contain acpp serial number, acpp ip, acpp inband id, acpp mac, acpp log timestamp, start line number of the acp log in the log file and the end line number where the log ends.  If acp log start mraker and acp log end marker are specified, the set of log files would be processed, else a warning would be displayed and the current acplog_master or acplog file would be processed.

Reference array reference for an ACP log file:

         [
          {
            'acpp_serial_number' => 'NA',
            'timestamp' => 'Sun Dec 20 00:05:45 GMT 2009 ',
            'ip' => '198.15.1.2',
            'end_block' => 9,
            'start_block' => 6,
            'inband_id' => '1a.2.A',
            'mac' => '00:50:cc:14:06:02'
          },
          {
            'acpp_serial_number' => 'NA',
            'timestamp' => 'Sun Dec 20 00:05:50 GMT 2009 ',
            'ip' => '198.15.1.3',
            'end_block' => 18,
            'start_block' => 14,
            'inband_id' => '1a.3.A',
            'mac' => '00:50:cc:14:06:03'
          },
          {
            'acpp_serial_number' => 'NA',
            'timestamp' => 'Sun Dec 20 00:06:40 GMT 2009 ',
            'ip' => '198.15.1.4',
            'end_block' => 27,
            'start_block' => 23,
            'inband_id' => '1a.4.A',
            'mac' => '00:50:cc:14:06:04'
          },
          {
            'acpp_serial_number' => 'NA',
            'timestamp' => 'Sun Dec 20 00:06:30 GMT 2009 ',
            'ip' => '198.15.1.5',
            'end_block' => 37,
            'start_block' => 32,
            'inband_id' => '1a.5.A',
            'mac' => '00:50:cc:14:06:05'
          }
        ];

The following named arguments are accepted,

=over

=item acp_logs

(Required) : The reference of acp_logs data structure

=back

=cut



#########################################################################################################
#
# Subroutine to generate the records between the start and end markers from the specified acp log file
#
#########################################################################################################

sub acp_log_generate_records {

        my (%opts) = validate(@_, {
                                acp_logs => { type => HASHREF}
                                });

        logcomment("acp_log_generate_records: Generating records from the ACP log file");
        my $acp_logs = $opts{acp_logs};

        my $rvalue = ();
        my $line;
        my $block = 0;

    #Read the first line from the log file and initialize the current_line to start marker if defined
        acp_log_read_first_line({acp_logs => $acp_logs});
        $acp_logs->{current_line}--;

        my $inside_acpp_flag = 0;

    #Read the lines from the log file and parse for header and body
        while($line = acp_log_read_next_line({acp_logs => $acp_logs}))
        {
        #Start parsing the header on encountering a line having ACPP Log Start
                if($line =~ /=+ ACPP Log Start =+/)
                {
                        if($inside_acpp_flag == 1)
                        {
                                $Log->warn("acp_log_generate_records: Log file is corrupt, encountered ACPP Log Start mark twice");
                                return undef;
                        }

            my $hash_ref;

            #If header information is not processed, on returning undef, return undef to the calling program
                if(!($hash_ref = acp_log_process_header({acp_logs => $acp_logs})))
                {
                        return undef;
            }

            #Define the hash reference for the current record
                        $$rvalue[$block] = $hash_ref;
                        $$rvalue[$block]->{start_block} = $acp_logs->{current_line}++;
                        $inside_acpp_flag = 1;
                }

        #End parsing the record on encountering ACPP Log End
                if($line =~ /=+ ACPP Log End =+/)
                {
            #If ACPP Log End is found twice without ACPP Log Start, declare the file is corrupt
                        if($inside_acpp_flag == 0)
                        {
                                if($block != 0)
                {
                    $Log->warn("acp_log_generate_records: Log file is corrupt, encountered ACPP Log End mark twice");
                                    return undef;
                }
                        }
            #If inside a block then mark the end of block and increament the record value
            if($inside_acpp_flag == 1)
                        {
                                $$rvalue[$block]->{end_block} = $acp_logs->{current_line} - 1;
                                $block++;
                                $inside_acpp_flag = 0;
                        }
                }
        }

    #On reaching EOF while inside a record, declare the file may be corrupt
        if($inside_acpp_flag == 1)
        {
                $Log->warn("acp_log_generate_records: Line 'ACPP Log End' not found, the acplog file may be corrupt");
        $$rvalue[$block]->{end_block} = $acp_logs->{current_line} - 1;
        }
        return $rvalue;

}


=head2 acp_log_process_header

     my $array_ref = acp_log_process_header({ acp_logs => $acp_log });

This routine reads the ACP log header information and generates a hash reference with the acpp serial number, acpp ip, acpp inband id, acpp mac and acpp log timestamp information.  On encountering any error it returns an undef.

Reference hash reference for a header:

          {
            'acpp_serial_number' => 'NA',
            'timestamp' => 'Sun Dec 20 00:05:45 GMT 2009 ',
            'ip' => '198.15.1.2',
            'end_block' => 9,
            'start_block' => 6,
            'inband_id' => '1a.2.A',
            'mac' => '00:50:cc:14:06:02'
          },

The following named arguments are accepted,

=over

=item acp_logs

(Required) : The reference of acp_logs data structure

=back

=cut


#########################################################################################################
#
# Subroutine to process the header information of the acp log file record
#
#########################################################################################################

sub acp_log_process_header {
        my(%opts) = validate(@_, {
                                acp_logs => { type=>HASHREF },
                                });
        my $acp_logs = $opts{acp_logs};
        my $rvalue = {};
        my $pos;
        my $line;

    #Process and retrieve the header information into a structure, on encountering a read failure return undef
        if(!($line = acp_log_read_next_line({acp_logs => $acp_logs})))
        {
                return undef;
        }
        chomp($line);
        {
                $rvalue->{mac} = $line;
                $pos = log_search_error_check($line, "MAC: ");
                if($pos == "-1")
        {
            return undef;
        }
        $rvalue->{mac} =~ s/(.{$pos})(\S+)(.*)$/$2/g;
                #print "rvalue->{mac} = $rvalue->{mac}\n";
        }

        {
                $rvalue->{ip} = $line;
                $pos = log_search_error_check($line, "IP: ");
                if($pos == "-1")
        {
            return undef;
        }
                $rvalue->{ip} =~ s/(.{$pos})(\S+)(.*)$/$2/g;
                #print "rvalue->{ip} = $rvalue->{ip}\n";
        }

        {
                $rvalue->{inband_id} = $line;
                $pos = log_search_error_check($line, "Inband ID: ");
                if($pos == "-1")
        {
            return undef;
        }
                $rvalue->{inband_id} =~ s/(.{$pos})(\S+)(.*)$/$2/g;
                #print "rvalue->{inband_id} = $rvalue->{inband_id}\n";
        }

        if(!($line = acp_log_read_next_line({acp_logs => $acp_logs})))
        {
                return undef;
        }
        chomp($line);
        {
                $rvalue->{acpp_serial_number} = $line;
                $pos = log_search_error_check($line, "ACPP Serial Number: ");
                if($pos == "-1")
        {
            return undef;
        }
                $rvalue->{acpp_serial_number} =~ s/(.{$pos})(\S+)(.*)$/$2/g;
                #print "rvalue->{acpp_serial_number} = $rvalue->{acpp_serial_number}\n";
        }

        {
                $rvalue->{timestamp} = $line;
                $pos = log_search_error_check($line, "Timestamp: ");
                if($pos == "-1")
        {
            return undef;
        }
                $rvalue->{timestamp} =~ s/(.{$pos})(.*)$/$2/g;
                #print "rvalue->{timestamp} = $rvalue->{timestamp}\n";
        }

        if(!($line = acp_log_read_next_line({acp_logs => $acp_logs})))
        {
                return undef;
        }

    chomp($line);

    if($line !~ /=+/)
    {
        $Log->warn("acp_log_process_header: ACP log file seems to be corrupt, please check the log file");
    }

    return $rvalue;
}



=head2 log_search_error_check

     my $position = log_search_error_check( $line, $pattern);

This routine searches for the pattern in line and returns the positon of the pattern in the line.

The following named arguments are accepted,

=over

=back

=cut


#########################################################################################################
#
# Subroutine to check for a pattern and return the position of the pattern in a line
#
#########################################################################################################

sub log_search_error_check($$)
{

        my ($line, $search) = @_;

        my $position = index($line, $search);
        if($position == -1)
        {
                $Log->warn("log_search_error_check: Error retrieving $search, the acp log file is corrupt");
                return -1;
        }
        $position = $position + length($search);

        return $position;
}



=head2 acp_log_search_miis

     my $array_ref = acp_log_search_miis({ acp_log_records => $acp_log_records, inband_ids => $array_ref_inband_ids });

This routine accepts the array reference of generated records for the acp logs and returns a list of all the records that match the given criteria.  The routine tries to search for specified inband ids, macs, ips and serial numbers.  By default if this routine is called without specifying search criteria all the logs would be returned.  The return value would be a hash reference containing the count of matched records and an array reference to the list of matched record ids from acp_log_records.

Reference return value:
        {
          'count' => '2',
          'records' => [
                         4,
                         5
                       ]
        };

The following named arguments are accepted,

=over

=item acp_log_records

(Required) : The reference of acp log records.  This can be generated from acp_log_generate_records

=item macs

(Optional) : An array reference to the list of ACPP MAC addresses that need to be searched for in the acp logs

=item ips

(Optional) : An array reference to the list of ACPP IP addresses that need to be searched for in the acp logs

=item inband_ids

(Optional) : An array reference to the list of ACPP Inband ids that need to be searched for in the acp logs

=item serial_numbers

(Optional) : An array reference to the list of ACPP serial numbers that need to be searched for in the acp logs

=back

=cut


#########################################################################################################
#
# Subroutine to search for specific records based on MACS, IPS, INBAND IDs or Serial Numbers
#
#########################################################################################################

sub acp_log_search_miis {
        my (%opts) = validate(@_, {
                        acp_log_records => { type => ARRAYREF },
                        macs => { type => ARRAYREF, optional => 1},
                        ips => { type => ARRAYREF, optional => 1},
                        inband_ids => {type => ARRAYREF, optional => 1},
                        serial_numbers => {type => ARRAYREF, optional => 1},
                        });
        my $acp_log_records = $opts{acp_log_records};
        my $matched_records;
        my $match_count = 0;

    #Return found records as zero if no logs could be parsed
    if($#$acp_log_records < 0)
    {
        my $rvalue = {};
        $rvalue->{count} = 0;
        return $rvalue;
    }

    #Search through the records for specific IPs if IPs are specified, else return the complete set of records for further search
        if((defined $opts{ips}) && ($#{$opts{ips}} >= 0))
        {
                my $ips = $opts{ips};

                logcomment("acp_log_search_miis: Searching for IPs: @$ips");
                for(my $i = 0; $i <= $#$acp_log_records; $i++)
                {
                        foreach my $ip (@$ips)
                        {
                                if($$acp_log_records[$i]->{ip} eq $ip)
                                {
                                        push @$matched_records, $i;
                                        last;
                                }
                        }
                }
                $match_count = $#$matched_records + 1;
        }
        else
        {
                for(my $i = 0; $i <= $#$acp_log_records ; $i++)
                {
                        push @$matched_records, $i;
                }
                $match_count = $#$matched_records + 1;
        }

    #Search through the records for specific MACs if MACs are specified, else return the complete set of records for further search
        if((defined $matched_records) && (defined $opts{macs}) && ($#{$opts{macs}} >= 0))
        {
                my $macs = $opts{macs};
                logcomment("acp_log_search_miis: Searching for MACs: @$macs");
                $matched_records = acp_log_search_element({ acp_log_records => $acp_log_records, records => $matched_records, parameter => "mac", elements => $macs });
        }

    #Search through the records for specific Inband Ids if they are specified, else return the complete set of records for further search
        if((defined $matched_records) && (defined $opts{inband_ids}) && ($#{$opts{inband_ids}} >= 0))
        {
                my $inband_ids = $opts{inband_ids};
                logcomment("acp_log_search_miis: Searching for Inband IDs: @$inband_ids");
                $matched_records = acp_log_search_element({ acp_log_records => $acp_log_records, records => $matched_records, parameter => "inband_id", elements => $inband_ids });
        }

    #Search through the records for specific Serial Numbers if they are specified, else return the complete set of records for further search
        if((defined $matched_records) && (defined $opts{serial_numbers}) && ($#{$opts{serial_numbers}} >=0))
        {
                my $serial_numbers = $opts{serial_numbers};
                logcomment("acp_log_search_miis: Searching for Module Serial Numbers: @$serial_numbers");
                $matched_records = acp_log_search_element({ acp_log_records => $acp_log_records, records => $matched_records, parameter => "serial_number", elements => $serial_numbers });
        }

        #Caluculate the number of records found and return a hash reference with count and recordxs ids
    my $rvalue;
        if(defined $matched_records)
        {
                $match_count = $#$matched_records + 1;
                $rvalue = {
                        count => $match_count,
                        records => $matched_records
                        };
        }
        else
        {
                $rvalue->{count} = 0;
        }

#    print "From: acp_log_search_miis\n";
#    print Dumper($rvalue);

        return $rvalue;

}


=head2 acp_log_search_element

     my $array_ref = acp_log_search_element({ acp_log_records => $acp_log_records, records => $records, parameter => $parameter, elements => $elements });

This routine accepts the array reference of the generated records from the acp logs, a set of selected records, the header parameter for which the matching has to be performed and the values that need to be matched for.  The header contains a set of parameters like acpp serial number, acpp ip, acpp macs, acpp inband ids and acpp serial numbers.  This routine can be used to search for any of these parameters for a set of matching values.

The return value would be an array reference to the set of matched record ids from the acp_log_records.

Reference return value:
        {
          'count' => '2',
          'records' => [
                         4,
                         5
                       ]
        };

The following named arguments are accepted,

=over

=item acp_log_records

(Required) : The reference of acp log records.  This can be generated from acp_log_generate_records

=item records

(Required) : An array reference to a list of matched records

=item parameter

(Required) : A scalar representing either acpp serial number or acpp mac or acpp inband id or acpp ip

=item elements

(Required) : An array reference to the list of parameter values that need to be searched for

=back

=cut

#########################################################################################################
#
# Subroutine to search for a particular header element and return the matched set of records
#
#########################################################################################################

sub acp_log_search_element {
        my (%opts) = validate(@_, {
                        acp_log_records => { type => ARRAYREF},
                        records => { type => ARRAYREF },
                        parameter => { type => SCALAR },
                        elements => {type => ARRAYREF }
                        });

        my $acp_log_records = $opts{acp_log_records};
        my $records = $opts{records};
        my $parameter = $opts{parameter};
        my $elements = $opts{elements};

    if($#$acp_log_records < 0)
    {
        my $rvalue = {};
        $rvalue->{count} = 0;
        return $rvalue;
    }
        my $matched_records = ();
        foreach my $record (@$records)
        {
                my $match_flag = 0;
                foreach my $element (@$elements)
                {
                        if($$acp_log_records[$record]->{$parameter} eq $element)
                        {
                                $match_flag = 1;
                        }
                }
                if($match_flag == 1)
                {
                        push @$matched_records, $record;
                }
        }

        if($#$matched_records != -1)
        {
                return $matched_records;
        }
        else
        {
                return undef;
        }
}



=head2 acp_log_search_strings

     my $array_ref = acp_log_search_strings({ acp_logs => $acp_logs, acp_log_records => $acp_log_records, strings => $strings_ref });

This routine searches for a set of specified strings in the acp log records and returns the list of record ids in which strings are found.

The routine accepts the following optional parameters
1. selected records which can be an array reference to the set of selected record ids which can be refine the search.
2. events which can be an array reference to set of events
3. acpp_modules which can be an array reference to set of acpp module ids

The return value is a hash reference which contains the count of matched records and an array reference to the set of matched record ids.

Reference return value:
        {
          'count' => '2',
          'records' => [
                         4,
                         5
                       ]
        };

The following named arguments are accepted,

=over

=item acp_log

(Required) : The acp logs data structure hash reference

=item acp_log_records

(Required) : The reference of acp log records.  This can be generated from acp_log_generate_records

=item selected_records

(Optional) : Reference to a set of selected record ids

=item strings

(Optional) : An array reference to the list of strings that need to be searched for in the acp logs

=item events

(Optional) : An array reference to the list of event ids that need to be searched for in the acp logs

=item acpp_modules

(Optional) : An array reference to the list of ACPP module names that need to be searched for in the acp logs

=back

=cut


#########################################################################################################
#
# Subroutine to search for specific records containing specified events, acpp modules or strings
#
#########################################################################################################

sub acp_log_search_strings {
        my (%opts) = validate(@_, {
                        acp_logs => { type => HASHREF},
                        acp_log_records => { type => ARRAYREF },
                        selected_records => { type => HASHREF, optional => 1},
                        strings => { type => ARRAYREF, optional => 1},
            events => {type => ARRAYREF, optional => 1},
            acpp_modules => {type => ARRAYREF, optional => 1},
                        });

        my $acp_logs = $opts{acp_logs};
        my $acp_log_records = $opts{acp_log_records};
        my $strings;
    my $events;
    my $acpp_modules;
    my $rvalue = {};

    #Return number of records found as zero if no logs could be parsed
    if($#$acp_log_records < 0)
    {
        my $rvalue = {};
        $rvalue->{count} = 0;
        return $rvalue;
    }

    #Return number of records found as zero if no search criterial is specified
    if((!(defined $opts{strings})) &&  (!(defined $opts{events})) && (!(defined $opts{acpp_modules})))
    {
        $Log->warn("acp_log_search_strings: No search criteria specified");
    }

        my $match_count = 0;
        my $line;
        my $records;

    #Handle if already selected records are zero and return number of records as zero
        if(defined $opts{selected_records})
        {
                my $selected_records = $opts{selected_records};
        if(defined $selected_records->{count})
        {
            if($selected_records->{count} == 0)
            {
                return $selected_records;
            }
            else
            {
                $records = $selected_records->{records};
            }
        }
        else
        {
            my $rvalue = {};
            $rvalue->{count} = 0;
            return $rvalue;
        }
        }
        else
        {
                $records = [0..$#$acp_log_records];
        }

    #Search for specified strings in the set of records
    if((defined $opts{strings}) && ($#{$opts{strings}} >= 0))
    {
            my @matched_records;
            $" = ',';
        $strings = $opts{strings};
        logcomment("acp_log_search_strings: Searching for Strings: @$strings");
            $" = ' ';
        foreach my $record (@$records)
            {
            my $match_flag = 0;
            for (my $lineno = $$acp_log_records[$record]->{start_block} + 1; $lineno <= $$acp_log_records[$record]->{end_block}; $lineno++)
            {
                if($line = acp_log_read_line({acp_logs => $acp_logs, line => $lineno}))
                {
                    chomp($line);
                    foreach my $string (@$strings)
                    {
                        if($line =~ /$string/)
                        {
                            if($match_flag == 0)
                            {
                                                               $match_flag = 1;
                                                               push @matched_records, $record;
                                last;
                            }
                        }
                    }
                }
            }
        }
        if($#matched_records < 0)
        {
            $rvalue->{count} = 0;
            return $rvalue;
        }
        else
        {
            $records = \@matched_records;
        }
    }

    #Search for specified events in the set of records
    if((defined $opts{events}) && ($#{$opts{events}} >= 0))
    {
            my @matched_records;
            $" = ',';
        $events = $opts{events};
        logcomment("acp_log_search_strings: Searching for Events: @$events");
            $" = ' ';
        foreach my $record (@$records)
            {
            my $match_flag = 0;
            for (my $lineno = $$acp_log_records[$record]->{start_block} + 1; $lineno <= $$acp_log_records[$record]->{end_block}; $lineno++)
            {
                if($line = acp_log_read_line({acp_logs => $acp_logs, line => $lineno}))
                {
                    chomp($line);
                    foreach my $event (@$events)
                    {
                        if($line =~ /^\w+; \d+; \d+; \w+; \w+; $event;/)
                        {
                            if($match_flag == 0)
                            {
                                                               $match_flag = 1;
                                                               push @matched_records, $record;
                                last;
                            }
                        }
                    }
                }
            }
        }
        if($#matched_records < 0)
        {
            $rvalue->{count} = 0;
            return $rvalue;
        }
        else
        {
            $records = \@matched_records;
        }
    }

    #Search for specified acpp modules in the set of records
    if((defined $opts{acpp_modules}) && ($#{$opts{acpp_modules}} >= 0))
    {
            my @matched_records;
            $" = ',';
        $acpp_modules = $opts{acpp_modules};
        logcomment("acp_log_search_strings: Searching for ACPP Modules: @$acpp_modules");
            $" = ' ';
        foreach my $record (@$records)
            {
            my $match_flag = 0;
            for (my $lineno = $$acp_log_records[$record]->{start_block} + 1; $lineno <= $$acp_log_records[$record]->{end_block}; $lineno++)
            {
                if($line = acp_log_read_line({acp_logs => $acp_logs, line => $lineno}))
                {
                    chomp($line);
                    foreach my $acpp_module (@$acpp_modules)
                    {
                        if($line =~ /^\w+; \d+; \d+; \w+; $acpp_module;/)
                        {
                            if($match_flag == 0)
                            {
                                                               $match_flag = 1;
                                                               push @matched_records, $record;
                                last;
                            }
                        }
                    }
                }
            }
        }
        if($#matched_records < 0)
        {
            $rvalue->{count} = 0;
            return $rvalue;
        }
        else
        {
            $records = \@matched_records;
        }
    }

    $rvalue = {};
    if($#$records < 0)
    {
        $rvalue->{count} = 0;
    }
    else
    {
        $match_count = $#$records + 1;
        $rvalue = {
                    count => $match_count,
                    records => $records,
                    };
    }


#    print "From: acp_log_search_strings\n";
#    print Dumper($rvalue);
        return $rvalue;
}

=head2 acp_log_extract_records

     my $result = acp_log_extract_records({ filer => $filer, acp_log_start_marker => $log_start_marker, acp_log_end_marker => $log_end_marker, inband_ids => $inband_ids, include_strings => $include_strings, exclude_strings => $exclude_strings});

This routine is a wrapper for searching and extracting specific logs from the acp log file.  If the acp log file is specified, the contents of the log file $log_file are parsed.  If the filer with filer connection and acp log file are specified, the subroutine mounts the filer root volume and parses the logs.  The parameter sudo path is specified if the sudo command is not in default location /usr/local/bin.  The parameters acp_log_start_marker and acp_log_end_marker specify the starting and ending line numbers in the log file for parsing.  The logs are categorised into log records and stored as an array of hash references linked with the log file based on the line numbers.  Parameters like inband_ids can be specified to retrieve particular inband id records from the log file.  A default set of strings are always searched in the log file for all the controllers.  If the strings set needs to be modified, the parameters include_strings and exclude strings can be used to include or exclude a set of strings respectively.  These parameters need to be reference to an array of strings.  The parameter header_only when set to 1 prints only the header part of the logs that match the criteria.

The subroutine returns the count of matched controllers and the records that match the specified criteria.

Reference return value:
$VAR1 = {
          'count' => 2,
          'records' => [
                         {
                           'acpp_serial_number' => 'NA',
                           'timestamp' => 'Sat Mar 20 14:37:20 GMT 2010 ',
                           'ip' => '198.15.1.223',
                           'end_block' => 131200,
                           'start_block' => 131191,
                           'data' => [
                                       'IMS94214000016B; 0000013243057; 00000000000000; 00000010; disc; 08; 63; send~;27A9',
                                       'IMS94214000016B; 0000013243057; 00000000000000; 00000023; disc; 08; 65; cm_e~;284F',
                                       'IMS94214000016B; 0000013243057; 00000000000000; 0000000a; disc; 10; 68; read~;25FF',
                                       'IMS94214000016B; 0000013243057; 00000000000000; 0000000a; disc; 10; 43; read~;1EAC',
                                       'IMS94214000016B; 0000013306070; 00000000000000; 00000026; test; 10; 26; curr~;179F',
                                       'IMS94214000016B; 0000013306073; 00000000000000; 00000026; test; 10; 22; new ~;15E6',
                                     ],
                           'inband_id' => '4d.63.A',
                           'mac' => '00:50:cc:62:61:df'
                         },
            {
                           'acpp_serial_number' => 'NA',
                           'timestamp' => 'Sat Mar 20 14:37:21 GMT 2010 ',
                           'ip' => '198.15.1.244',
                           'end_block' => 131219,
                           'start_block' => 131210,
                           'data' => [
                                       'IMS94214C1404F4; 0000071131051; 00000000000000; 00000010; disc; 08; 63; send~;27C9',
                                       'IMS94214C1404F4; 0000071131051; 00000000000000; 00000023; disc; 08; 65; cm_e~;286F',
                                       'IMS94214C1404F4; 0000071131051; 00000000000000; 0000000a; disc; 10; 68; read~;25EF',
                                       'IMS94214C1404F4; 0000071131051; 00000000000000; 0000000a; disc; 10; 43; read~;1E9C',
                                       'IMS94214C1404F4; 0000071146064; 00000000000000; 00000026; test; 10; 26; curr~;17D1',
                                       'IMS94214C1404F4; 0000071146067; 00000000000000; 00000026; test; 10; 22; new ~;1618',
                                       'IMS94214C1404F4; 0000071146070; 00000000000000; 00000026; test; 10; 20; same~;174B',
                                       'IMS94214C1404F4; 0000071146070; 00000000000000; 00000008; disc; 08; 37; log_~;1CF0',
                                     ],
                           'inband_id' => '4d.0.A',
                           'mac' => '00:50:cc:14:04:f4'
                         }
              ]
    }

The following named arguments are accepted,

=over

=item filer

(Optional) : Host spec of the filer.

=item acp_log_files

(Optional) : An array reference for the list of acp log files.

=item filer_conn

(Optional) : Connection type for the filer. Default is #rsh.

=item sudo_path

(Optional) : path for the sudo command on the client.

=item acp_log_file_name

(Optional) : name of the acp log file for parsing eg. acplog_master or acplog

=item macs

(Optional) : an array reference to the list of mac addresses that need to be searched for in the acp logs

=item ips

(Optional) : an array reference to the list of ip addresses that need to be searched for in the acp logs

=item inband_ids

(Optional) : an array reference to the list of inband ids that need to be searched for in the acp logs

=item serial_numbers

(Optional) : an array reference to the list of serial numbers that need to be searched for in the acp logs

=item acp_log_start_marker

(Optional) : start marker of the acp log file got from acp_log_mark or acp_log_mark_only

=item acp_log_end_marker

(Optional) : end marker of the acp log file got from acp_log_mark or acp_log_mark_only

=item events

(Optional) : an array reference to the set of events that need to be searched for in the acp logs

=item acpp_modules

(Optional) : an array reference to the set of acpp modules that need to be searched for in the acp logs

=item include_strings

(Optional) : An array reference to the list of strings that need to be searched for in the acp logs

=item default_strings

(Optional) : flag to search for the default set of strings in the acp logs.  Default is 1 (TRUE).

=item exclude_strings

(Optional) : An array reference to the list of strings that need to be excluded during search of acp logs

=item monitor_priority_events

(Optional) : flag to monitor criticial events in the log file.  Default is 1 (TRUE).

=item header_only

(Optional) : flag to retreive only headers of the matching records.  Default is 0 (FALSE).

=back

=cut


#########################################################################################################
#
# Subroutine to extract specified records from logs
#
#########################################################################################################

sub acp_log_extract_records {
        my (%opts) = validate(@_, {
                        filer => { type => SCALAR, optional => 1},
                        acp_log_files => { type => ARRAYREF, optional => 1},
            filer_conn => { type => SCALAR, optional => 1},
            sudo_path => { type => SCALAR, optional => 1},
            acp_log_file_name => { type => SCALAR, optional => 1},
            timeout  => { type => SCALAR, optional => 1},
            macs => { type => ARRAYREF, optional => 1},
            ips => { type => ARRAYREF, optional => 1},
            inband_ids => {type => ARRAYREF, optional => 1},
            serial_numbers => {type => ARRAYREF, optional => 1},
            acp_log_start_marker => { type => HASHREF, optional => 1},
            acp_log_end_marker => { type => HASHREF, optional => 1},
            events => { type => ARRAYREF, optional => 1},
            acpp_modules => { type => ARRAYREF, optional => 1},
            include_strings => { type => ARRAYREF, optional => 1 },
            default_strings => { type => SCALAR, optional => 1, default => 1 },
            exclude_strings => { type => ARRAYREF, optional => 1 },
            monitor_priority_events => { type => SCALAR, optional => 1, default => 1},
            header_only => { type => SCALAR, optional => 1, default => 0},
                        });

        my $acp_logs = {};
    my $filer = $opts{filer};
        if(defined $opts{acp_log_start_marker})
        {
                $acp_logs->{acp_log_start_marker} = $opts{acp_log_start_marker};
                logcomment("acp_log_extract_records: Log start marker is set to $acp_logs->{acp_log_start_marker}->{line_number}");
        #print Dumper($opts{acp_log_start_marker});
        }
        if(defined $opts{acp_log_end_marker})
        {
                $acp_logs->{acp_log_end_marker} = $opts{acp_log_end_marker};
                logcomment("acp_log_extract_records: Log end marker is set to $acp_logs->{acp_log_end_marker}->{line_number}");
        #print Dumper($opts{acp_log_end_marker});
        }

    #Return count as 0 if no search criteria is specified
    if((!defined $opts{macs}) && (!defined $opts{ips}) && (!defined $opts{inband_ids}) && (!defined $opts{serial_numbers}) && (!defined $opts{include_strings}) &&
        ($opts{default_strings} == 0) && (!defined $opts{events}) && (!defined $opts{acpp_modules}) && ($opts{monitor_priority_events} == 0))
    {
        my $rvalue = {};
        $rvalue->{count} = 0;
        return $rvalue;
    }

    #If filer is specified mount the filer and get the log file path
    #If log files are specified get the list of log files
    #Fail if nothing is specified
        my $acp_log_files;
        if(defined $opts{acp_log_files})
        {
                 $acp_log_files = $opts{acp_log_files};
        }
        elsif(defined $opts{filer})
        {
                my $parameters;
                $parameters->{filer} = $opts{filer};
                if(defined $opts{filer_conn})
                {
                        $parameters->{filer_conn} = $opts{filer_conn};
                }
                if(defined $opts{acp_log_file_name})
                {
                        my $temp = $opts{acp_log_file_name};
            $temp =~ s/(.*)\/(\w+)/$2/g;
                        $parameters->{acp_log_file_name} = $temp;
                }
                if(defined $opts{sudo_path})
                {
                        $parameters->{sudo_path} = $opts{sudo_path};
                }
                $acp_log_files = acp_log_mount($parameters);

        }
        else
        {
                logresult("FAIL", "acp_log_extract_records: Please provide atleast a filer name or a log file");
                return undef;
        }

    #Check if the log files exist and are accessible
        my $check;
        if(defined $opts{filer})
        {
                $check = acp_log_exist({ filer => $opts{filer}, acp_log_files => $acp_log_files, acp_logs => $acp_logs});
        }
        else
        {
                $check = acp_log_exist({ acp_log_files => $acp_log_files, acp_logs => $acp_logs});

        }

    #Return search count as zero if log file is not updated with any data
    if((defined $acp_logs->{acp_log_start_marker}) && (defined $acp_logs->{acp_log_start_marker}))
    {
        if($#{$acp_logs->{acp_log_files}} == 0 && ($acp_logs->{acp_log_start_marker}->{line_number} == $acp_logs->{acp_log_end_marker}->{line_number}))
            {
                    logcomment("acp_log_extract_records: No new information has been logged between the start and end markers");
                    my $result = {};
            $result->{count} = 0;
            return $result;
        }
    }

    #Generate records from the ACP log file
    my $myrecords;
    if(!($myrecords = acp_log_generate_records({acp_logs => $acp_logs})))
    {
                my $result;
        $result->{count} = 0;
        logcomment("acp_log_extract_records: No new records found while parsing the ACP logs");
        return $result;
    }

    #Search for specified MAC, IP, Inband id or ACPP serial number in the records from the logs, else continue to search for strings, events or acpp modules
    my $selected_records;
        if((defined $opts{macs}) || (defined $opts{ips}) || (defined $opts{inband_ids}) || (defined $opts{serial_numbers}))
        {
                my $parameters;
                $parameters->{acp_log_records} = $myrecords;
                if(defined $opts{macs})
                {
                        $parameters->{macs} = $opts{macs};
                }
                if(defined $opts{ips})
                {
                        $parameters->{ips} = $opts{ips};
                }
                if(defined $opts{inband_ids})
                {
                        $parameters->{inband_ids} = $opts{inband_ids};
                }
                if(defined $opts{serial_numbers})
                {
                        $parameters->{serial_numbers} = $opts{serial_numbers};
                }
                $selected_records = acp_log_search_miis($parameters);

        if(defined $selected_records)
        {
            if(defined $selected_records->{count})
            {
                if($selected_records->{count} == 0)
                {
                    logcomment("acp_log_extract_records: No records found in search (1)");
                    return $selected_records;
                }
            }
            else
            {
                my $rvalue = {};
                $rvalue->{count} = 0;
                return $rvalue;
            }
        }
        else
        {
            my $rvalue = {};
            $rvalue->{count} = 0;
            return $rvalue;
        }

    }
    else
    {
        $selected_records->{count} = $#$myrecords +1;
        $selected_records->{records} = [0..$#$myrecords];
    }

    #Search for default set of strings if default_strings is set to 1
        my $strings;

        if($opts{default_strings} eq "1")
        {
                $strings = ["test"]; # To be taken from SVL
        }

    #Add specified strings to search if include_strings are specified
        if(defined $opts{include_strings})
        {
                my $include_strings = $opts{include_strings};
                foreach my $string (@$include_strings)
                {
                        push @$strings, $string;
                }
        }

    #Remove specified strings from search if exclude_strings are specified
        if(defined $opts{exclude_strings})
        {
                my $unmatched_strings;
                my $exclude_strings = $opts{exclude_strings};
                foreach my $string (@$strings)
                {
                        my $match_flag = 0;
                        foreach my $exclude_string (@$exclude_strings)
                        {
                                if($string eq $exclude_string)
                                {
                                        $match_flag = 1;
                                }
                        }
                        if($match_flag == 0)
                        {
                                push @$unmatched_strings, $string;
                        }
                }
                $strings = $unmatched_strings;
        }

    #Search for default set of event types if monitor_priority_events is set to 1, else use the specified list of events for search
        my $events;
        if($opts{monitor_priority_events} eq "1")
        {
                $events = ["01","02"];
                if(defined $opts{events})
                {
                        $events = [@$events, @{$opts{events}}];
                }
        }
        else
        {
                $events = $opts{events};
        }

    #Search for specified strings, events or acpp modules int the log records
    if(($#$strings >= 0) || ($#$events >= 0) || (defined $opts{acpp_modules}))
    {
        my $parameters;
        $parameters->{acp_logs} = $acp_logs;
        $parameters->{acp_log_records} = $myrecords;
        if(defined $selected_records)
        {
            if(defined $selected_records->{count})
            {
                if($selected_records->{count} != 0)
                {
                    $parameters->{selected_records} = $selected_records;
                }
            }
        }
        if($#$strings >= 0)
        {
            $parameters->{strings} = $strings;
        }
        if($#$events >= 0)
        {
            $parameters->{events} = $events;
        }
        if(defined $opts{acpp_modules})
        {
            $parameters->{acpp_modules} = $opts{acpp_modules};
        }

        $selected_records = acp_log_search_strings($parameters);

        if(defined $selected_records)
        {
            if(defined $selected_records->{count})
            {
                if($selected_records->{count} == 0)
                {
                    logcomment("acp_log_extract_records: No records found in search (2)");
                    return $selected_records;
                }
            }
            else
            {
                my $rvalue = {};
                $rvalue->{count} = 0;
                return $rvalue;
            }
        }
        else
        {
            my $rvalue = {};
            $rvalue->{count} = 0;
            return $rvalue;
        }
    }

    #Extract the records that match the criterial to an array of hash references and return to the caller
    my $result = {};
    $result->{count} = $selected_records->{count};
    my @result_array = ();
    my $records_to_be_extracted = $selected_records->{records};
    my @records_to_be_extracted = ();
       @records_to_be_extracted = @$records_to_be_extracted;

    foreach my $record (@records_to_be_extracted)
    {
        my $extracted_record = $$myrecords[$record];

        if($opts{header_only} == 0)
        {
            my @data = ();

            for(my $lineno = $$myrecords[$record]->{start_block} + 1; $lineno <= $$myrecords[$record]->{end_block}; $lineno++)
            {
                my $line;
                if(!($line = acp_log_read_line({ acp_logs => $acp_logs, line => $lineno })))
                {
                            $Log->warn("acp_log_extract_records: Unexpected undef encountered: Dumping current data from variables");
                            print Dumper($acp_logs);
                            logresult("Script","acp_log_read_line: Encountered an unexpected error(3).  Please contact the author with the logs");
                            return undef;
                }
                chomp($line);
                push @data, $line;
            }

            $extracted_record->{data} = \@data;
        }
        push @result_array, $extracted_record;
    }

    $result->{records} = \@result_array;
    return $result;
}


=head2 acp_log_print_records

     my $result = acp_log_print_records({ acp_log_records => $acp_log_records, acp_logs =>$acp_logs, records => $records});

This routine prints the specified log records.

The following named arguments are accepted,

=over

=item acp_logs

(Required) : The acp_logs data structure with acp_log_start_marker and acp_log_end_marker

=item acp_log_records

(Required) : The reference of acp log records.  This can be generated from acp_log_generate_records

=item records

(Required) : An array reference to a list of matched records

=back

=cut




#########################################################################################################
#
# Subroutine to print the specified records
#
#########################################################################################################

sub acp_log_print_records {
    my (%opts) = validate(@_, {
            acp_logs => { type => HASHREF },
            acp_log_records => { type => ARRAYREF },
            records    => { type => ARRAYREF },
            });

    my $acp_logs = $opts{acp_logs};
    my $acp_log_records = $opts{acp_log_records};
    my $records = $opts{records};

    logcomment("acp_log_print_records: Printing Records");
    for(my $num = 0; $num <=$#$records; $num++)
    {
        if(!(defined $$acp_log_records[$$records[$num]]->{start_block}))
        {
            $Log->warn("Start block line number for $$records[$num] is missing: Dumping current data from variables");
            print Dumper($acp_logs);
            logresult("Script","acp_log_print_records: Encountered an unexpected error(4). Please contact the author with the logs");
            return undef;
        }
        my $start_line = $$acp_log_records[$$records[$num]]->{start_block} - 3;

        if(!(defined $$acp_log_records[$$records[$num]]->{end_block}))
        {
            $Log->warn("End block line number for $$records[$num] is missing: Dumping current data from variables");
            print Dumper($acp_logs);
            logresult("Script","acp_log_print_records: Encountered an unexpected error(5). Please contact the author with the logs");
            return undef;
        }
        my $end_line = $$acp_log_records[$$records[$num]]->{end_block} + 1;
        my $line;

        my $lineno;
        for($lineno = $start_line; $lineno <= $end_line; $lineno++)
        {
            if(!($line = acp_log_read_line({acp_logs => $acp_logs, line => $lineno})))
            {
                $Log->warn("Encountered EOF while reading a record: Dumping current data from variables");
                print Dumper($acp_logs);
                logresult("Script","acp_log_print_records: Encountered an unexpected error(6). Please contact the author with the logs");
                return undef;
            }
            print $line;
        }
    }
    return 0;
}


=head2 acp_log_search

     my $result = acp_log_search({ filer => $filer, acp_log_start_marker => $log_start_marker, acp_log_end_marker => $log_end_marker, inband_ids => $inband_ids, include_strings => $include_strings, exclude_strings => $exclude_strings});

This routine is a wrapper for searching and printing specific log records from the acp log file.  If the acp log file is specified, the contents of the log file $log_file are parsed.  If the filer with filer connection and acp log file are specified, the subroutine mounts the filer root volume and parses the logs.  The parameter sudo path is specified if the sudo command is not in default location /usr/local/bin.  The parameters acp_log_start_marker and acp_log_end_marker specify the starting and ending line numbers in the log file for parsing.  The logs are categorised into log records and stored as an array of hash references linked with the log file based on the line numbers.  Parameters like inband_ids can be specified to retrieve particular inband id records from the log file.  A default set of strings are always searched in the log file for all the controllers.  If the strings set needs to be modified, the parameters include_strings and exclude strings can be used to include or exclude a set of strings respectively.  These parameters need to be reference to an array of strings.  The parameter header_only when set to 1 prints only the header part of the logs that match the criteria.

The subroutine returns the count of matched controllers that match the specified criteria.

Reference return value:
$VAR1 = {
          'count' => 2,
    }

The following named arguments are accepted,

=over

=item filer

(Optional) : Host spec of the filer.

=item acp_log_files

(Optional) : An array reference for the list of acp log files.

=item filer_conn

(Optional) : Connection type for the filer. Default is #rsh.

=item sudo_path

(Optional) : path for the sudo command on the client.

=item acp_log_file_name

(Optional) : name of the acp log file for parsing eg. acplog_master or acplog

=item macs

(Optional) : an array reference to the list of mac addresses that need to be searched for in the acp logs

=item ips

(Optional) : an array reference to the list of ip addresses that need to be searched for in the acp logs

=item inband_ids

(Optional) : an array reference to the list of inband ids that need to be searched for in the acp logs

=item serial_numbers

(Optional) : an array reference to the list of serial numbers that need to be searched for in the acp logs

=item acp_log_start_marker

(Optional) : start marker of the acp log file got from acp_log_mark or acp_log_mark_only

=item acp_log_end_marker

(Optional) : end marker of the acp log file got from acp_log_mark or acp_log_mark_only

=item events

(Optional) : an array reference to the set of events that need to be searched for in the acp logs

=item acpp_modules

(Optional) : an array reference to the set of acpp modules that need to be searched for in the acp logs

=item include_strings

(Optional) : An array reference to the list of strings that need to be searched for in the acp logs

=item default_strings

(Optional) : flag to search for the default set of strings in the acp logs.  Default is 1 (TRUE).

=item exclude_strings

(Optional) : An array reference to the list of strings that need to be excluded during search of acp logs

=item monitor_priority_events

(Optional) : flag to monitor criticial events in the log file.  Default is 1 (TRUE).

=item header_only

(Optional) : flag to retreive only headers of the matching records.  Default is 0 (FALSE).

=back

=cut



#########################################################################################################
#
# Subroutine to search for specified records from logs
#
#########################################################################################################

sub acp_log_search {
        my (%opts) = validate(@_, {
                        filer => { type => SCALAR, optional => 1},
                        acp_log_files => { type => SCALAR, optional => 1},
            filer_conn => { type => SCALAR, optional => 1},
            sudo_path => { type => SCALAR, optional => 1},
            acp_log_file_name => { type => SCALAR, optional => 1},
            timeout  => { type => SCALAR, optional => 1},
            macs => { type => ARRAYREF, optional => 1},
            ips => { type => ARRAYREF, optional => 1},
            inband_ids => {type => ARRAYREF, optional => 1},
            serial_numbers => {type => ARRAYREF, optional => 1},
            acp_log_start_marker => { type => HASHREF, optional => 1},
            acp_log_end_marker => { type => HASHREF, optional => 1},
            events => { type => ARRAYREF, optional => 1},
            acpp_modules => { type => ARRAYREF, optional => 1},
            include_strings => { type => ARRAYREF, optional => 1 },
            default_strings => { type => SCALAR, optional => 1, default => 1 },
            exclude_strings => { type => ARRAYREF, optional => 1 },
            monitor_priority_events => { type => SCALAR, optional => 1, default => 1},

                        });

        my $acp_logs = {};
    my $filer = $opts{filer};
        if(defined $opts{acp_log_start_marker})
        {
                $acp_logs->{acp_log_start_marker} = $opts{acp_log_start_marker};
                logcomment("acp_log_search: Log start marker is set to $acp_logs->{acp_log_start_marker}->{line_number}");
        }
        if(defined $opts{acp_log_end_marker})
        {
                $acp_logs->{acp_log_end_marker} = $opts{acp_log_end_marker};
                logcomment("acp_log_search: Log end marker is set to $acp_logs->{acp_log_end_marker}->{line_number}");
        }

        #If filer is specified mount the filer and get the log file path
        #If log files are specified get the list of log files
        #Fail if nothing is specified
        my $acp_log_files;
        if(defined $opts{acp_log_files})
        {
                 $acp_log_files = $opts{acp_log_files};
        }
        elsif(defined $opts{filer})
        {
                my $parameters;
                $parameters->{filer} = $opts{filer};
                if(defined $opts{filer_conn})
                {
                        $parameters->{filer_conn} = $opts{filer_conn};
                }
                if(defined $opts{acp_log_file_name})
                {
                        $parameters->{acp_log_file_name} = $opts{acp_log_file_name};
                }
                if(defined $opts{sudo_path})
                {
                        $parameters->{sudo_path} = $opts{sudo_path};
                }

                $acp_log_files = acp_log_mount($parameters);
                $acp_log_files = [$acp_log_files];
        }
        else
        {
                logresult("FAIL", "acp_log_search: Please provide atleast a filer name or a log file");
                return undef;
        }

    #Check if the log files exist and are accessible
        my $check;
        if(defined $opts{filer})
        {
                $check = acp_log_exist({ filer => $opts{filer}, acp_log_files => $acp_log_files, acp_logs => $acp_logs});
        }
        else
        {
                $check = acp_log_exist({ acp_log_files => $acp_log_files, acp_logs => $acp_logs});

        }

    #Return search count as zero if log file is not updated with any data
    if((defined $acp_logs->{acp_log_start_marker}) && (defined $acp_logs->{acp_log_start_marker}))
    {
        if($#{$acp_logs->{acp_log_files}} == 0 && ($acp_logs->{acp_log_start_marker}->{line_number} == $acp_logs->{acp_log_end_marker}->{line_number}))
            {
                    logcomment("acp_log_search: No new information has been logged between the start and end markers");
                    my $result;
            $result->{count} = 0;
            return $result;
        }
        }

    #Generate records from the ACP log file
    my $myrecords;
    if(!($myrecords = acp_log_generate_records({acp_logs => $acp_logs})))
    {
                my $result;
        $result->{count} = 0;
        logcomment("acp_log_search: No new records found while parsing the ACP logs");
        return $result;
    }

    #Search for specified MAC, IP, Inband id or ACPP serial number in the records from the logs, else continue to search for strings, events or acpp modules
    my $selected_records = {};
    if((defined $opts{macs}) || (defined $opts{ips}) || (defined $opts{inband_ids}) || (defined $opts{serial_numbers}))
    {
        my $parameters;
        $parameters->{acp_log_records} = $myrecords;
        if(defined $opts{macs})
        {
            $parameters->{macs} = $opts{macs};
        }
        if(defined $opts{ips})
        {
            $parameters->{ips} = $opts{ips};
        }
        if(defined $opts{inband_ids})
        {
            $parameters->{inband_ids} = $opts{inband_ids};
        }
        if(defined $opts{serial_numbers})
        {
            $parameters->{serial_numbers} = $opts{serial_numbers};
        }
        $selected_records = acp_log_search_miis($parameters);

                if(defined $selected_records)
                {
                        if(defined $selected_records->{count})
                        {
                                if($selected_records->{count} == 0)
                                {
                                        logcomment("acp_log_search: No records found in search (1)");
                                        return $selected_records;
                                }
                        }
                        else
                        {
                                my $rvalue = {};
                                $rvalue->{count} = 0;
                                return $rvalue;
                        }
                }
                else
                {
                        my $rvalue = {};
                        $rvalue->{count} = 0;
                        return $rvalue;
                }

        }
    else
    {
        $selected_records->{count} = $#$myrecords +1;
        $selected_records->{records} = [0..$#$myrecords];
    }

    #Search for default set of strings if default_strings is set to 1
    my $strings;
    if($opts{default_strings} eq "1")
        {
                $strings = ["test"]; # To be taken from SVL
        }

    #Add specified strings to search if include_strings are specified
        if(defined $opts{include_strings})
        {
                my $include_strings = $opts{include_strings};
                foreach my $string (@$include_strings)
                {
                        push @$strings, $string;
                }
        }

    #Remove specified strings from search if exclude_strings are specified
        if(defined $opts{exclude_strings})
        {
                my $unmatched_strings;
                my $exclude_strings = $opts{exclude_strings};
                foreach my $string (@$strings)
                {
                        my $match_flag = 0;
                        foreach my $exclude_string (@$exclude_strings)
                        {
                                if($string eq $exclude_string)
                                {
                                        $match_flag = 1;
                                }
                        }
                        if($match_flag == 0)
                        {
                                push @$unmatched_strings, $string;
                        }
                }
                $strings = $unmatched_strings;
        }

    #Search for default set of event types if monitor_priority_events is set to 1, else use the specified list of events for search
    my $events;
    if($opts{monitor_priority_events} eq "1")
    {
        $events = ["01","02"];
        if(defined $opts{events})
        {
            $events = [@$events, @{$opts{events}}];
        }
    }
    else
    {
        $events = $opts{events};
    }

    #Search for specified strings, events or acpp modules in the log records
    if(($#$strings >= 0) || ($#$events >= 0) || (defined $opts{acpp_modules}))
    {
        my $parameters;
        $parameters->{acp_logs} = $acp_logs;
        $parameters->{acp_log_records} = $myrecords;
        if(defined $selected_records)
        {
            if(defined $selected_records->{count})
            {
                $parameters->{selected_records} = $selected_records;
            }
        }
        if($#$strings >= 0)
        {
            $parameters->{strings} = $strings;
        }
        if($#$events >= 0)
        {
            $parameters->{events} = $events;
        }
        if(defined $opts{acpp_modules})
        {
            $parameters->{acpp_modules} = $opts{acpp_modules};
        }

        $selected_records = acp_log_search_strings($parameters);
                if(defined $selected_records)
                {
                        if(defined $selected_records->{count})
                        {
                                if($selected_records->{count} == 0)
                                {
                                        logcomment("acp_log_search: No records found in search (2)");
                                        return $selected_records;
                                }
                        }
                        else
                        {
                                my $rvalue = {};
                                $rvalue->{count} = 0;
                                return $rvalue;
                        }
                }
                else
                {
                        my $rvalue = {};
                        $rvalue->{count} = 0;
                        return $rvalue;
                }

    }

    #Printing the set of records which match the specified criteria in the logs
    logcomment("acp_log_search: Start of Records");
    acp_log_print_records({ acp_logs => $acp_logs, acp_log_records => $myrecords, records => $selected_records->{records}});
    logcomment("acp_log_search: End of Records");
    return $selected_records;
}



=head2 acp_log_search_all

     my $result = acp_log_search_all({ filer => $filer, acp_log_start_marker => $log_start_marker, acp_log_end_marker => $log_end_marker, inband_ids => $inband_ids, include_strings => $include_strings, exclude_strings => $exclude_strings});

This routine searches for specified strings in the log file from the start marker to the end marker.

The subroutine returns the 1 on success and 0 on failure.

The following named arguments are accepted,

=over

=item filer

(Optional) : Host spec of the filer.

=item acp_log_files

(Optional) : An array reference for the list of acp log files.

=item filer_conn

(Optional) : Connection type for the filer. Default is #rsh.

=item sudo_path

(Optional) : path for the sudo command on the client.

=item acp_log_file_name

(Optional) : name of the acp log file for parsing eg. acplog_master or acplog

=item acp_log_start_marker

(Optional) : start marker of the acp log file got from acp_log_mark or acp_log_mark_only

=item acp_log_end_marker

(Optional) : end marker of the acp log file got from acp_log_mark or acp_log_mark_only

=item include_strings

(Optional) : An array reference to the list of strings that need to be searched for in the acp logs

=item default_strings

(Optional) : flag to search for the default set of strings in the acp logs.  Default is 1 (TRUE).

=item exclude_strings

(Optional) : An array reference to the list of strings that need to be excluded during search of acp logs

=back

=cut


#########################################################################################################
#
# Subroutine to search for a speicified set of strings the log file
#
#########################################################################################################

sub acp_log_search_all {
        my (%opts) = validate(@_, {
                        filer => { type => SCALAR, optional => 1},
                        acp_log_files => { type => SCALAR, optional => 1},
            filer_conn => { type => SCALAR, optional => 1},
            sudo_path => { type => SCALAR, optional => 1},
            acp_log_file_name => { type => SCALAR, optional => 1},
            timeout  => { type => SCALAR, optional => 1},
            strings => { type => ARRAYREF, optional => 1},
            acp_log_start_marker => { type => HASHREF, optional => 1},
            acp_log_end_marker => { type => HASHREF, optional => 1},
            include_strings => { type => ARRAYREF, optional => 1 },
            default_strings => { type => SCALAR, optional => 1, default => 1 },
            exclude_strings => { type => ARRAYREF, optional => 1 },
                        });

        my $acp_logs = {};
    my $filer = $opts{filer};
        if(defined $opts{acp_log_start_marker})
        {
                $acp_logs->{acp_log_start_marker} = $opts{acp_log_start_marker};
                logcomment("acp_log_search: Log start marker is set to $acp_logs->{acp_log_start_marker}->{line_number}");
        }
        if(defined $opts{acp_log_end_marker})
        {
                $acp_logs->{acp_log_end_marker} = $opts{acp_log_end_marker};
                logcomment("acp_log_search: Log end marker is set to $acp_logs->{acp_log_end_marker}->{line_number}");
        }

        #If filer is specified mount the filer and get the log file path
        #If log files are specified get the list of log files
        #Fail if nothing is specified
        my $acp_log_files;
        if(defined $opts{acp_log_files})
        {
                 $acp_log_files = $opts{acp_log_files};
        }
        elsif(defined $opts{filer})
        {
                my $parameters;
                $parameters->{filer} = $opts{filer};
                if(defined $opts{filer_conn})
                {
                        $parameters->{filer_conn} = $opts{filer_conn};
                }
                if(defined $opts{acp_log_file_name})
                {
                        $parameters->{acp_log_file_name} = $opts{acp_log_file_name};
                }
                if(defined $opts{sudo_path})
                {
                        $parameters->{sudo_path} = $opts{sudo_path};
                }

                $acp_log_files = acp_log_mount($parameters);
                $acp_log_files = [$acp_log_files];
        }
        else
        {
                logresult("FAIL", "acp_log_search: Please provide atleast a filer name or a log file");
                return undef;
        }

    #Check if the log files exist and are accessible
        my $check;
        if(defined $opts{filer})
        {
                $check = acp_log_exist({ filer => $opts{filer}, acp_log_files => $acp_log_files, acp_logs => $acp_logs});
        }
        else
        {
                $check = acp_log_exist({ acp_log_files => $acp_log_files, acp_logs => $acp_logs});

        }

    #Return search count as zero if log file is not updated with any data
    if((defined $acp_logs->{acp_log_start_marker}) && (defined $acp_logs->{acp_log_start_marker}))
    {
        if($#{$acp_logs->{acp_log_files}} == 0 && ($acp_logs->{acp_log_start_marker}->{line_number} == $acp_logs->{acp_log_end_marker}->{line_number}))
            {
                    logcomment("acp_log_search: No new information has been logged between the start and end markers");
                    my $result;
            $result->{count} = 0;
            return $result;
        }
        }

    #Search for default set of strings if default_strings is set to 1
    my $strings;
    if($opts{default_strings} eq "1")
        {
                $strings = ["test"]; # To be taken from SVL
        }

    #Add specified strings to search if include_strings are specified
        if(defined $opts{include_strings})
        {
                my $include_strings = $opts{include_strings};
                foreach my $string (@$include_strings)
                {
                        push @$strings, $string;
                }
        }

    #Remove specified strings from search if exclude_strings are specified
        if(defined $opts{exclude_strings})
        {
                my $unmatched_strings;
                my $exclude_strings = $opts{exclude_strings};
                foreach my $string (@$strings)
                {
                        my $match_flag = 0;
                        foreach my $exclude_string (@$exclude_strings)
                        {
                                if($string eq $exclude_string)
                                {
                                        $match_flag = 1;
                                }
                        }
                        if($match_flag == 0)
                        {
                                push @$unmatched_strings, $string;
                        }
                }
                $strings = $unmatched_strings;
        }

    #Search through the logs and return 0 if a string is found else return 1
    acp_log_read_first_line({acp_logs => $acp_logs});
    $acp_logs->{current_line}--;

    my $line;
    if($#$strings >= 0)
    {
                $" = ',';
                logcomment("acp_log_search_all: Searching for Strings: @$strings");
                $" = ' ';
        while($line = acp_log_read_next_line({acp_logs => $acp_logs}))
        {
                    chomp($line);
            foreach my $string (@$strings)
            {
                if($line =~ /$string/)
                {
                    logcomment("acp_log_search_all: Found a matching string");
                    return 0;
                }
            }
        }
    }
    logcomment("acp_log_search_all: Did not find any matching strings");
    return 1;
}

1;
