# Library to extract useful information from "storage show adapter -p"
## @summary Library of routines to extract  Storage adapter information and Inject/Check  Storage adapter errors
##
## @author dl-sar-qa@netapp.com, dl-step@netapp.com
##
## @synopsis
## use Storage::AdaptUtils;
##
## @change 10-27-2004 vivienne: add subroutines.
##  get_online_adapters_maint
##  get_active_adapters_maint
##  get_online_lports_maint
##  reduncheck
##  enablecheck
##  disablecheck
## @change 03-05-2008 vchintal: add subroutines
##      get_single_attach
##      turn_onoff_adapters
##      get_adapters_makesa
## @change 09-11-2012 amal@netapp.com
##      added Cmode support for get_online_adapters
##
## @change 01-23-2013 ssrinath
##      added Cmode support for get_adapters
##
## @change 06-14-2013 nandeesh
##      added Cmode support for get_online_adapters_maint(), get_active_adapters_maint(), get_online_lports_maint()
##
####################################################################


package Storage::AdaptUtils;
use strict;
require Exporter;
use vars ('@ISA','@EXPORT_OK');
@ISA = ('Exporter');
@EXPORT_OK = qw(
        get_adapter_sysconfig
        get_link_rate
        get_adapters
        get_online_adapters
        get_online_adapters_maint
        get_active_adapters_maint
        get_online_lports
        get_online_lports_maint
        get_adapt_hash
        get_adapt_version
        get_adapt_description
        get_adapt_width
        get_adapt_speed
        get_adapt_nname
        get_adapt_csize
        get_adapt_psize
        get_adapt_drate
        get_adapt_parity
        get_adapt_state
        get_adapt_use
        get_adapt_redundant
                turn_onoff_adapters
        reduncheck
        enablecheck
        disablecheck
                get_single_attach
                get_adapters_makesa
        );
use Tharn;
use Carp qw(confess croak);
use Storage::Misc qw(filer_cmd);
use NACL::C::Node;
use Params::Validate qw(validate OBJECT SCALAR);

BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}

#### Function to get the output of sysconfig for a particular adapter port
## @name Storage::AdaptUtils::get_adapter_sysconfig
## @summary Returns either an array or a string with the sysconfig output
## @synopsis
##  get_adapter_sysconfig( conn => $conn, adapter => $adapter, timeout => $timeout);
##
## @arg conn: connectrec to use for the connection
## @arg adapter: adapter port name
## @arg timeout: say timeout to use

sub get_adapter_sysconfig(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                adapter => { type=>SCALAR },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $adapter = $opts{adapter};
    my $timeout = $opts{timeout};

    my $slot = $adapter;
    $slot =~ s/\D//g;

    my (undef,undef,$sysconf_out) =say($conn,"sysconfig -v $slot", "-timeout", $timeout);
    chomp( my @sysconf_lines = split(/\r?\n/,"$sysconf_out"));
    if ($slot ne $adapter){
        my $sysconf_start = -1;
        my $sysconf_end = -1;
        foreach my $i (0 .. $#sysconf_lines){
            if ($sysconf_lines[$i] =~ /slot\s+$slot:\s+FC\s+Host\s+Adapter\s+$adapter\s+/){
                $sysconf_start = $i;
            } elsif (($sysconf_start != -1) && ($sysconf_lines[$i] =~ /slot $slot: FC Host Adapter/)){
                # this is the case where the adapter is not the last one listed for the slot
                $sysconf_end = ($i-1);
                if (wantarray) {
                    return  @sysconf_lines[$sysconf_start .. $sysconf_end];
                } else {
                    return join ("\n", @sysconf_lines[$sysconf_start .. $sysconf_end]);
                }
            }
        }
        if ($sysconf_start != -1){ # handle the case where the adapter is the last listed for the slot
            if (wantarray) {
                return  @sysconf_lines[$sysconf_start .. $#sysconf_lines];
            } else {
                return join ("\n", @sysconf_lines[$sysconf_start .. $#sysconf_lines]);
            }
        }
        die "CONFIG: adapter sysconfig not found for $adapter: @sysconf_lines";
    }
    # now handle the case where slot = $adapter (no a, b,... suffix)
    if ($sysconf_out =~  /slot $adapter: FC Host Adapter $adapter/){
        if (wantarray) {
            return @sysconf_lines;
        } else {
            return join ("\n", @sysconf_lines);
        }
    } else {
        die "CONFIG: adapter sysconfig not found for $adapter: @sysconf_lines";
    }
}

#### Function to get the link rate for an adapter
## @name Storage::AdaptUtils::get_link_rate
## @summary Returns the link speed for the adapter
## @synopsis
##  get_link_rate( conn => $conn, adapter => $adapter, adapter_sysconfig => $adapter_sysconfig,
##              timeout => $timeout);
##
## @arg conn: connectrec to use for the connection
## @arg adapter: adapter port name
## @arg adapter_sysconfig: sysconfig output for adapter
## @arg timeout: say timeout to use

sub get_link_rate(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                adapter => { type=>SCALAR},
                                adapter_sysconfig => { type=>SCALAR, default=>"", optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $adapter = $opts{adapter};
    my $timeout = $opts{timeout};
    my $adapter_sysconfig = $opts{adapter_sysconfig} || get_adapter_sysconfig(conn => $conn, adapter => $adapter, timeout => $timeout);

    chomp( my @adapter_syslines = split(/\r?\n/,"$adapter_sysconfig"));
    my ($link_line) = grep(/Link Data Rate/, @adapter_syslines);
    return $1 if ($link_line =~ /Link Data Rate:\s+(\d+)\s+Gbit/);
    die "CONFIG: link rate not found for adapter $adapter: $adapter_sysconfig";
}


#### Function to get the list adapters in a filer.
## @name ::get_adapters
## @summary Returns an array of storage adapters in the filer
## @synopsis
##  get_adapters($filer)
##
## @arg $filer Required: Filer connnection to use

sub get_adapters {
    my ( $a , $b , $c , $adap , @adap , @adap1 , @adap2 , @adap3);
    my $host_rec = $_[0]->hostp();
    $c = filer_cmd(host=>$host_rec, cmd=>"storage show adapter");
    @adap = split (/\r*\n/,$c);
    @adap1 =  grep /Slot/, @adap;
    @adap3 = ();

    foreach $adap (@adap1) {
        my @adap2 = split(/Slot:\s+/ ,$adap);
        push @adap3 , @adap2;
    }
    return (@adap3);

}

#### Function to get a list of online adapters.
## @name ::get_online_adapters
## @summary Returns an array of online storage adapters in the filer
## @synopsis
##  get_online_adapters($filer)
##
## @arg $filer Required: Filer connnection to use

sub get_online_adapters {

        my $host_rec = $_[0]->hostp( );

        my $c = filer_cmd(host=>$host_rec, cmd=>"fcadmin link_state");

    my (@adap , @adap1 , @adap2 , @adap3 , $findad , $item , %seen ,$i );
        tstlog -comment, "$c \n";
    @adap1 = split( /\r*\n/, $c);
    @adap1 = grep/Link State for adapter/ , @adap1;
    $i =0;
    foreach $findad (@adap1) {
        if ( $findad =~ /UP/) {
            @adap2 = split(/\s+/,$findad);
            $adap3[$i] = $adap2[5];
            $adap3[$i] =~ s/://;
            $i++;
        }
    }
    tstlog -comment, "Online Adapters: @adap3 \n";
    return (@adap3);
}

#### Function to get a list of online adapters.
## @name ::get_online_adapters_maint
## @summary Returns an array of online storage adapters in the filer
##          running in maintenance mode (console connection)
## @synopsis
##  get_online_adapters_maint($filer)
##
## @arg $filer Required: Filer connnection to use (console connection)

sub get_online_adapters_maint {
        my ( $c, @adap , @adap1 , @adap2 , @adap3 , $findad , $item , %seen ,$i, $fc );
        my $filer_ci = NACL::C::Node->find();
       # if ( $filer_ci->is_cmode() ) {
     #       (undef,undef,$c) = say( $_[0], "run local \"priv set -q advanced;fcadmin link_state\"",'-noprompt','-prompt' , "^\\*>","-timeout", "400" , -ok);
    #    } else {
            (undef,undef,$c) = say( $_[0], "priv set -q advanced;fcadmin link_state",'-noprompt','-prompt' , "^\\*>","-timeout", "400" , -ok);
     #   }
    @adap1 = split( /\r*\n/, $c);
    @adap1 = grep/Link State for adapter/ , @adap1;
    $i =0;
    foreach $findad (@adap1) {
        if ( $findad =~ /UP/) {
            @adap2 = split(/\s+/,$findad);
            $adap3[$i] = $adap2[5];
            $adap3[$i] =~ s/://;
            $i++;
        }
    }
    tstlog -comment, "Active adapters in the system: @adap3 \n";
    return (@adap3);
}

#### Function to get a list of adapters from "environ shelf".
## @name ::get_active_adapters_maint
## @summary Returns an array of active storage adapters in the filer
##          running in maintenance mode (console connection)
## @synopsis
##  get_active_adapters_maint($filer)
##
## @arg $filer Required: Filer connnection to use (console connection)

sub get_active_adapters_maint {
        my ( @adap , @adap1 ,$adap1, @adap2 , @adap3 , $findad , $item , %seen, $i);
        my $filer_ci = NACL::C::Node->find();
       # if ( $filer_ci->is_cmode() ) {
      #      (undef,undef,$adap1) = say( $_[0], "run local \"priv set -q advanced;environ shelf\"",'-noprompt','-prompt' , "^\\*>","-timeout", "400" , -ok);
     #   } else {
            (undef,undef,$adap1) = say($_[0], "environ shelf",'-noprompt','-prompt' , "^\\*>","-timeout", "400" , -ok);
     #   }
        @adap1 = split(/\n+/,$adap1);
        @adap1 = grep/Environment for channel/ , @adap1;
        $i =0;
        foreach $findad (@adap1) {
            @adap2 = split(/\r*\s+/,$findad);
            $adap3[$i] = $adap2[4];
            $adap3[$i] =~ s/://;
            $i++;
        }
        %seen =();
        foreach $item (@adap3) {
            push(@adap,$item) unless $seen{$item}++;
    }
    tstlog -comment, "Active adapters in the system : @adap \n";
    return (@adap);
}

## @name ::get_online_lports
## @summary Returns an array of storage adapters in L-port mode on the filer
## @synopsis
##  get_online_lports($filer)
##
## @arg $filer Required: Filer connnection to use
sub get_online_lports {
    my ($c, @adap1 , $findad , @adap2 , @adap3 , $abcd , $i, @adap );
    (undef,undef,$c)  = say($_[0] , "priv set -q advanced;fcadmin device_map");
    @adap1 = split(/\r*\n/,$c);
    $i =0;
     foreach $findad (@adap1){
        if ( !($findad =~ /failed|Cannot complete/) &&   ($findad =~ /channel|adapter/)){
            $adap2[$i] = $findad;
            $i++;
        }
    }
    $i =0;
    foreach $abcd (@adap2){
        @adap3 = split(/\s+/,$abcd);
        $adap[$i] = $adap3[4];
        $adap[$i] =~ s/:$//;
        $i++;
    }
    tstlog -comment, "Online Adapters: @adap \n";
    return (@adap);
}

## @name ::get_online_lports_maint
## @summary Returns an array of storage adapters in L-port mode on the filer
##          that running in maintenance mode from "fcadmin device_map".
##          Gives switch ports (console connection)
## @synopsis
##  get_online_lports_maint($filer)
##
## @arg $filer Required: Filer connnection to use (console connection)
sub get_online_lports_maint {
        my( @adap1 ,$adap1, $findad, $a , $b , $o , @adap2 , @adap3 , $abcd , $i, @adap , @adap4);
        my $filer_ci = NACL::C::Node->find();
       # if ( $filer_ci->is_cmode() ) {
      #      (undef,undef,$adap1) = say( $_[0], "run local \"priv set -q advanced;fcadmin device_map\"",'-noprompt','-prompt' , "^\\*>","-timeout", "400" , -ok);
     #   } else {
            (undef,undef,$adap1) = say($_[0], "fcadmin device_map",'-noprompt','-prompt' , "^\\*>","-timeout", "400" , -ok);
     #   }
        @adap1 = split(/\r*\n+/,$adap1);
        $i = 0;
        foreach $findad (@adap1){
            $findad =~ s/\r\r\n//;
            if ( !($findad =~ /failed/) && ($findad =~ /channel|adapter/) ) {
                $adap2[$i] = $findad;
                $i++;
            }
        }
        $i =0;
        foreach $abcd (@adap2){
            @adap3 = split(/\s+/,$abcd);
            $adap4[$i] = $adap3[4];
            $adap4[$i] =~ s/:$//;
          #  if ( $filer_ci->is_cmode() ) {
         #       (undef,undef,$o) = say( $_[0], "run local \"priv set -q advanced;fcadmin device_map $adap4[$i]\"",'-noprompt','-prompt' , "^\\*>","-timeout", "400" , -ok);
         #   } else {
                (undef,undef,$o) = say($_[0], "fcadmin device_map $adap4[$i]",'-noprompt','-prompt' , "^\\*>","-timeout", "400" , -ok);
         #   }
            if ($o =~ /Shelf \d+:/) {
                push @adap , $adap4[$i];
            }
            $i++;
        }
        return (@adap);
}

#### Function to create a hash table
## @name ::get_adapt_hash
## @summary Returns a hash table of storage adapters  information in the filer
## @synopsis
##  get_adapt_hash($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required
sub get_adapt_hash {

    my ( $a , $b , $c , $adap , @adap,@adap1, %adap);
    my $filer_ci = NACL::C::Node->find();
    if ( $filer_ci->is_cmode() ) {
        ($a , $b , $c) = say ( $_[0] ,"run local \"priv set -q advanced;storage show adapter $_[1]\"");
    } else {
        ($a , $b , $c) = say ( $_[0] ,"storage show adapter $_[1]");
    }
    @adap = split (/\r*\n/,$c);
    foreach my $line (@adap){
        logcomment("The line : $line");
        @adap1 = split(/:\s+/,$line);
        logcomment("key and value : $adap1[0] and $adap1[1]");
        $adap{"$adap1[0]"}=$adap1[1];
    }
    return(%adap);
}


#### Function to return the Description of an adapter
## @name ::get_adapt_description
## @summary Returns description of a storage adapter
## @synopsis
##  get_adapt_description($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_description {

    my %info = get_adapt_hash($_[0],"$_[1]");
    if ( exists($info{"Description"})) {
        my $description = $info{"Description"};
        return ($description);
    }
    else {
        tstlog (-comment , "This adapter doen't have any description");
        return undef;
    }

}

#### Function to return the Adapter FW version
## @name ::get_adapt_version
## @summary Returns fw version of a storage adapter.
## @synopsis
##  get_adapt_version($filer)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_version {

    my %info = get_adapt_hash($_[0],"$_[1]");
    if ( exists($info{"Firmware Rev"})) {
        my $fw_ver = $info{"Firmware Rev"};
        return ( $fw_ver );
    }
    else {
        tstlog (-comment , "FW version doesn't exist for this adapter.");
        return undef;
    }

}

#### Function to get the PCI bus width of an adapter
## @name ::get_adapt_width
## @summary Returns PCI bus width of an adapter
## @synopsis
##  get_adaptwidth($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_width {

    my %info = get_adapt_hash($_[0],"$_[1]");

    if ( exists($info{"PCI Bus Width"})) {

        my $width = $info{"PCI Bus Width"};
        return ($width);
    }
    else {
        tstlog (-comment , "Unable to obtain the PCI bus width of this adapter");
        return undef;
    }

}

####  Function to get the speed of an adapter
## @name ::get_adapt_speed
## @summary Returns PCI clock speed of an adapter
## @synopsis
##  get_adapt_speed($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_speed {

    my %info = get_adapt_hash($_[0],"$_[1]");

    if ( exists($info{"PCI Clock Speed"})) {

        my $speed = $info{"PCI Clock Speed"};
        return ($speed);
    }
    else {
        tstlog (-comment , "Unable to obtain the clock speed of this adapter");
        return undef;
    }

}

#### Function to get the FC Node name of an adapter
## @name ::get_adapt_nname
## @summary Returns FC node name of an adapter
## @synopsis
##  get_adapt_nname($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_nname {
    my %info = get_adapt_hash($_[0],"$_[1]");
    if ( exists($info{"FC Node Name"})) {

        my $nname = $info{"FC Node Name"};
        return ($nname);
    }
    else {
        tstlog (-comment , "FC Node name doesnot exist for this adapter");
        return undef;
    }
}


#### Function to get the Cacheline size of an adapter
## @name ::get_adapt_csize
## @summary Returns cacheline size of an adapter
## @synopsis
##  get_adapt_csize($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_csize {

    my %info = get_adapt_hash($_[0],"$_[1]");

    if ( exists($info{"Cacheline Size"})) {

        my $csize =  $info{"Cacheline Size"};
        return ($csize);
    }
    else {
        tstlog (-comment , "Unable to obtain the Cacheline size of this adapter");
        return undef;
    }

}


#### Function to get the packet size of an adapter
## @name ::get_adapt_psize
## @summary Returns FC packet size of an adapter
## @description Returns FC packet size of an adapter
##      Returns "NO
## @synopsis
##  get_adapt_psize($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_psize {

    my %info = get_adapt_hash($_[0],"$_[1]");
    if ( exists($info{"FC Packet Size"})) {

        my $psize = $info{"FC Packet Size"};
        return ($psize);
    }
    else {
        tstlog (-comment , "Unable to obtain the packet size of the adapter");
        return undef;
    }

}

#### Function to get the Data Link rate of an adapter
## @name ::get_adapt_drate
## @summary Returns link data rate of an adapter
## @description Returns link data rate of an adapter
##      Returns "2 Gbit" or "1 Gbit".Returns "undef" if unable to obtain the infomation.
## @synopsis
##  get_adapt_drate($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_drate {

    my %info = get_adapt_hash($_[0],"$_[1]");

    if ( exists($info{"Link Data Rate"})) {

        my $drate = $info{"Link Data Rate"};
        return ($drate);
    }
    else {
        tstlog (-comment , "Unable to obtain the Data link rate of this adapter");
        return undef;

    }
}


####  Function to get the SRAM parity info of an adapter
## @name ::get_adapt_parity
## @summary Returns SRAM parity of an adapter
## @description Returns SRAM parity of an adapter
##      Returns "Yes" or "No" if information found otherwise returns "undef"
## @synopsis
##  get_adapt_parity($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_parity {

    my %info = get_adapt_hash($_[0],"$_[1]");

    if ( exists($info{"SRAM Parity"})) {

        my $parity = $info{"SRAM Parity"};
        return($parity);
    }
    else {
        tstlog (-comment , "Unable to obtain the SRAM parity of this adapter");
        return undef ;
    }

}


#### Function to get the state of an adapter
## @name ::get_adapt_state
## @summary Returns Enabled/Disabled state of an adapter
## @description Returns Enabled/Disabled state of an adapter
##      Returns "Enabled" if adapter is online or "Disabled" if Offline. Returns "undef" if unable to obtain the information.
## @synopsis
##  get_adapt_state($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_state {

    my %info = get_adapt_hash($_[0],"$_[1]");

    if ( exists($info{"State"})) {

        my $state = $info{"State"};
        return($state);
    }
    else {
        tstlog (-comment , "Unable to obtain the state of this adapter");
        return undef;
    }

}

##### Function to check whether an adapter is in use or not
## @name ::get_adapt_use
## @summary Returns  Inuse information of an adapter
## @description Returns  Inuse information of an adapter
##      Returns "Yes" if in use otherwise "No". Returns "undef" if it's unable to obtain the information.
##
## @synopsis
##  get_adapt_use($filer,$adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_use {

    my %info = get_adapt_hash($_[0],"$_[1]");

    if ( exists($info{"In Use"})) {

        my $use = $info{"In Use"};
        return($use);
    }
    else {

        tstlog (-comment , "Unable to obtain the use info of this adapter");
        return undef;
    }

}

####  Function to check whether an adapter is redundant or not
## @name ::get_adapt_redundant
## @summary Returns redundancy state of an adapter
## @description Returns redundancy state of an adapter
##      Returns "Yes" if adapter is redundant or "No" if adapter is not redundant. Returns "undef" if it is unable to obtain the information.
##      Can be used to check if the system is dual attached or not.
## @synopsis
##  get_adapt_redundant($filer, $adapter)
##
## @arg $filer Required: Filer connnection to use
## @arg $adapter Required: Adapter for which the information is required

sub get_adapt_redundant {

    my %info = get_adapt_hash($_[0],"$_[1]");

    if ( exists($info{"Redundant"})) {

        my $redundant = $info{"Redundant"};
        return($redundant);
    }
    else {
        tstlog (-comment , "Unable to obtain the redudancy state of this adapter");
        return undef;
    }
}

####  Function to check the redundancy of loop.
## @name ::reduncheck
## @summary Returns script error if adapter is not redundant
## @description Returns script error if adapter is not redundant
##      Can be used to make sure the system is dual attached.
## @synopsis
##  reduncheck($filer, $adapter)
##
## @arg $adapter Required: Adapter for which the information is required

sub reduncheck  {
#   my ($a , $o , $c);
#   ($a ,$o,$c) = say($fc , "priv set advanced;storage show adapter $_[0]");

    my $c = get_adapt_redundant($_[0],"$_[1]");
    if ( $c !~ /Yes/){
        tstlog(-comment, "Adapter $_[1] is not redundant");
    } else {
        tstlog(-comment, "Adapter $_[1] is redundant");
    }
}


#### Function to check that the loop is enabled.
## @name ::enablecheck
## @summary Returns fatal error if adapter is not get enabled
## @description Returns fatal error if adapter is not get enabled(online).
## @synopsis
##  enablecheck($filer, $adapter)
##
## @arg $adapter Required: Adapter for which the information is required

sub enablecheck {
    my $c = get_adapt_state($_[0],"$_[1]");
    tstlog -comment, "State: $c \n";
    if ( $c !~ /Enabled/){
        tstlog(-fatal , "Adapter $_[1] didnot get enabled.");
    }
}

###### Subroutine to check that the loop is disabled.
#### Function to check that the loop is disabled.
## @name ::disablecheck
## @summary Returns fatal error if adapter is not get disabled
## @description Returns fatal error if adapter is not get disabled(offline).
## @synopsis
##  disablecheck($filer, $adapter)
##
## @arg $adapter Required: Adapter for which the information is required

sub disablecheck {
    my $c = get_adapt_state($_[0],"$_[1]");
    tstlog -comment, "State: $c \n";
    if ( $c !~ /Disabled/){
        tstlog(-fatal , "Adapter $_[1] didnot get disabled.");
    }
}


#### Function to turn on/off the specified adapters
## @name ::turn_onoff_adapters
## @summary turn on/off the adapters
## @description Given a list of adapters, and specified whether to turn on/off (0 to off and 1 to on), the adapters will be turned on/off accordingly
## @synopsis
##          turn_onoff_adapters($filer, $flag, @adapters)
##
## @arg $filer Required: Filer connection to use
## @arg $flag Required: Specify whether the adapters have to be turned on (pass 1) or turned off (pass 0)
## @arg @adapters Required: List of adapters

sub turn_onoff_adapters{
      my($r, $flag, @adapt) = @_;
      my ($c, $j);
      my($itersToBlock) = 6;
      my($tflag) = 0;
      if($flag == 0)
      {
        tstlog -comment, "Need to turn off the adapters..The adapters are: @adapt \n";
        for (my $i = 0; $i <= $#adapt; $i++)
        {
                $tflag = 0;
                say($r, "priv set -q test; fcadmin offline $adapt[$i]");
                for(my $j = 1; $j <= $itersToBlock; $j++)
                {
                    sleep 5;
                    (undef, undef, $c) = say($r, "priv set -q test; fcadmin link_state  $adapt[$i]");
                    if ($c =~ /OFFLINE/) {
                       tstlog -comment, "Adaper $adapt[$i] is successfully offlined. \n";
                       $tflag = 1;
                       last;
                     }
                }
                if($tflag == 0)
                {
                    tstlog -fatal, "Adaper $adapt[$i] is not offlined. \n";
                }
        }

     }
          elsif($flag == 1)
     {
         tstlog -comment, "Need to turn on the adapters..The adapters are: @adapt \n";
         for (my $i = 0; $i <= $#adapt; $i++)
         {
            $tflag = 0;
            say($r, "priv set -q test; fcadmin online $adapt[$i]");
            for($j = 1; $j <= $itersToBlock; $j++)
            {
               sleep 5;
               (undef, undef, $c) = say($r, "priv set -q test; fcadmin link_state  $adapt[$i]");
               if ($c =~ /UP/) {
                     tstlog -comment, "Adaper $adapt[$i] is successfully onlined. \n";
                     $tflag = 1;
                     last;
                   }
               elsif ($c =~ /physical/){
                        tstlog -comment, "Adapter $adapt[$i] is physically offlined. \n";
                        $tflag = 1;
                        last;
                        }
             }
            if($tflag == 0)
            {
                tstlog -fatal, "Fail to online Adaper $adapt[$i] \n";
            }
        }
     }
}


#### Function to see if the disks are dual attached or not
## @name ::get_single_attach
## @summary Returns whether the disks are single_attached or dual attached
## @description Returns 1 if single attached
##              Retuns 0 if dual attached
## @synopsis
##        get_single_attach($filer)
## @arg $filer Required: Filer connection to use

sub get_single_attach {
       my ($o, $flag, @lines, @str);
       (undef, undef, $o) = say($_[0], "storage show disk -p");
       (@lines) = split(/\n/, $o);
        for(my $i = 2; $i <= $#lines; $i++){
              (@str) = split(" ", $lines[$i]);                 # The output prints primary port secondary port shelf bay in that order. If secondary isn't there there would be exactly 4 elements.
               if( $#str != 3){
                   $flag = 1;
                   return 0;
               }
       }
      return 1;
}
#end of single_attach


#### Function that returns adapters to be disabled to make it single_attached
## @name ::get_adapters_makesa
## @summary Returns adapters to be disabled to make the dual attached single attached
## @description Returns the adapters to be disabled as an array
## @synopsis
##         get_adapters_makesa($filer)
## @arg $filer Required: Filer connection to use

sub get_adapters_makesa{
        if (get_single_attach($_[0])){
              tstlog -fatal, "This is single attached..Cannot find adapters";
        }
        my (%pri_hash, %sec_hash);
        my ($adapt, $o, $flag, $pr, $sc, $port1, $port2, $shelf, $bay, $diskid, $href, $href_sec, @final, @lines, @str);
        (undef, undef, $o) = say($_[0], "storage show disk -p");
        (@lines) = split(/\n/, $o);
        for(my $i = 2; $i <= $#lines; $i++)
        {
            ($pr, $port1, $sc, $port2, $shelf, $bay) = split(" ", $lines[$i]);
            ($adapt, $diskid) = split(/\./, $pr);
            $pri_hash{$port1}{$adapt} = undef;
            ($adapt, $diskid) = split(/\./, $sc);
            $pri_hash{$port2}{$adapt} = undef;
        }
       $href = \%pri_hash;
       $href_sec = $href->{$port2};
       @final = keys (%$href_sec);
       logcomment("hash is @final");
       return @final;
}

#end of det_adapters_makesa

1;
