# _*_ emacs _*_
# $Id: //depot/prod/test/Rsierranevada/lib/Storage/Disk.pm#8 $

## @summary Disk - the library provides routines to manipulate
## the disk command line interface.
## @author tomdoan@netapp.com
## @status testing
## @author yingz@netapp.com: function "disk_fw_update"
## @change sreedhar@netapp.com: function "_disk_fwcommand()" added event dfu.fwDownloaded
## @change shilpab@netapp.com: Enhanced function "disk_show" when no-options are passed
## @change amrita@netapp.com:disk_show not returning empty hash in case there are no disks to show.
## @change amitg@netapp.com: Fixed _diskfwcommand function. This function was waiting longer than sent wait_time.
## @change faslur@netapp.com: function "disk_shm_stats_asup" is added
## @burt 354489 Enhanced function "disk_show" when no-options are passed
## @burt 358976 disk_show not returning empty hash in case there are no disks to show.
## @change faslur: Burt#409611 added the support for secondary path if the system is multipathed in few functions
## @change faslur: Burt#473010 commented "last" at line in two places under _diskfwcommand() function
## @change gurus@netapp.com: Burt#521645 Replaced say command with filer_cmd in disk_show() function so that it supports C-mode.
## @change gurus@netapp.com: Burt#509853 Added support to disk_assign() and disk_remove_ownership() functions to handle prompts.
## @change gurus@netapp.com: Burt#594293 Modified disk_show() function to parse the output completely
## @change amal@netapp.com: Modified disk_fw so that it will fail, if disk_fw already in progress
## @change hansen: modified disk_show() to remove connection type regex,
## @change nbalacha: Modified disk_show() function, added last elseif part to parse the output
## @change nandeesh: Added Cmode support to disk_sanitize() function
## @change ssrinath: modified sub disk_show to support cmode even after reboot
##         by using apiset obj instead of filer_cmd
## @change villas: modified sub disk_scrub, disk_fail and disk_unfail to support cmode
## simply pass through and remove maintenance mode console connection creation.
## filer_cmd should handle both of these correctly.
#
#updates: 8/6/2013: jolie - changed subroutine disk_shm_stats_asup; regex wasn't matching for I/O Completion Time Table output which requires 14 columns.  See BURTs 742828 and 741294
## @change 2013-10-04: gururajc - Modified disk_remove_ownership sub to support cmode

=head1 NAME

Storage::Disk.pm - The disk library routine.

=head1 SYNOPSIS

use Storage::Disk qw(disk_* ...);

my $host = host($FILER);

disk_fail(host=>$host, disk=>$disk);

disk_remove(host=>$host, disk=>$disk);

disk_scrub(host=>$host);

disk_unfail(host=>$host, disk=>$disk);

...

=head1 DESCRIPTION

The library supports routines to manipulate the disk CLI.

The defined routines should work on both filers and "simulators".

=head1 ROUTINES

All routines, unless otherwise stated, return 1
on success and 0 on failure. On failure "$@" is set
to the corresponding error messages via "croak()".

All routines shall take the first argument "$host" as
the hostrec.

=cut

package Storage::Disk;

use strict;
use warnings;
our @EXPORT_OK;

BEGIN {
    use base qw(Exporter);
    @EXPORT_OK = qw(
      disk_assign
      disk_fail
      disk_fw_update
      disk_maint
      disk_remove
      disk_remove_ownership
      disk_replace
      disk_resume
      disk_sanitize
      disk_scrub
      disk_shm_stats
      disk_show
      disk_simpull
      disk_simpush
      disk_suspend
      disk_unfail
      disk_zero_spares
      disk_dump
      disk_unfail_during_reconstruction
      disk_wait_for_reconstruction
      disk_shm_stats_asup
      );

    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess(Tharn::script_result(@msgs));
        });
} ## end BEGIN

use Carp qw(croak confess);
use Params::Validate qw(validate validate_pos ARRAYREF OBJECT SCALAR );
use Module::Common qw(validate_mixed);
use Volume qw(vol_create vol_offline vol_destroy vol_exists);
use Data::Dumper;
use boot;
use filer;
use Storage::Ems qw(ems_log_dump_by_type ems_log_get_mark);
use Sysconfig;
use Aggregate qw(aggr_offline aggr_online);
use NACL::C::Node;
use NACL::APISet;
use Utils qw(make_host make_conn);
use Tharn;
use RaidInfo;
use Raid::RaidUtils qw(RaidUtils_is_sim);
use Storage::SASUtils qw(filer_cmd);
use RaidSqa;
use vars qw($LOGDIR $VERBOSE $FUNCTION_NAMES $FILERMODE $ITERATIONS);

print "\nUsing this module from " . __FILE__ . "\n";

# ---------------
# define constant
# ---------------
use constant SAY_TIMEOUT     => 300;
use constant PASS            => 1;
use constant FAIL            => 0;
use constant NOOP            => 2;
use constant MAINT_PROMPT    => "\\*\>";
use constant WAIT_UNTIL_DONE => -1;         # wait until done
use constant NORMAL_MODE     => "normal";
use constant MAINT_MODE      => "maint";
use constant YES             => 1;
use constant NO              => 0;
use constant VOLNAME         => "foo";

# ======================= DEFINE ROUTINES ===========================

=head2 disk_assign

disk_assign(host=>$host, disk=>$disk, ...);

The routine provides the interface of the disk assign command. It
initiates the assign ownership of the specified disk(s) to the
specified system. The routine supports both maint and normal mode.

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : hostspec of filer

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item timeout

(Optional) : connection timeout. Default is 300.

=item disk

(Required) : Specify a disk name, multiple disks separated by space, "all", or "-n <count>" of unowned_disks.

=item pool

(Optional) : Specify a pool name id 0 or 1. Default is none.

=item ownername

(Optional) : Specify the owner filer. Default is none

=item sysid

(Optional) : Specify a sysid. Default is none

=item blockzone

(Optional) : Specify either block or zoned. Default is none.

=item foption

(Optional) : Specify the [-f] option. Default is none.

=back

=item Returns

Returns 1 on success. It returns an error message via $@ if there is any error.

=back

=cut

sub disk_assign {
    _debug("Entering disk_assign: ...");
    my (%pars) =
      validate(@_,
               {host => {isa  => 'Hostrec'},
                disk => {type => SCALAR, optional => NO},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => "#rsh"},
                timeout   => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                filermode => {type     => SCALAR,
                              optional => YES,
                              regex    => qr/maint|normal/,
                              default  => NORMAL_MODE},
                pool       => {type => SCALAR, optional => YES},
                ownername  => {type => SCALAR, optional => YES},
                sysid      => {type => SCALAR, optional => YES},
                blockzoned => {type => SCALAR, optional => YES, regex => qr/block|zoned/},
                foption    => {type => SCALAR, optional => YES, regex => qr/-f/},});

    my $conn;
    my $pool       = defined($pars{pool})       ? " -p " . $pars{pool}       : "";
    my $ownername  = defined($pars{ownername})  ? " -o " . $pars{ownername}  : "";
    my $sysid      = defined($pars{sysid})      ? " -s " . $pars{sysid}      : "";
    my $blockzoned = defined($pars{blockzoned}) ? " -c " . $pars{blockzoned} : "";
    my $cmd        = "disk assign $pars{disk} $pool $ownername $sysid $blockzoned $pars{foption}";
    my $re_continue = "Do you wish to continue\\? ";
    my $output;
    if ($pars{filermode} eq NORMAL_MODE) {
        # $conn = make_conn($pars{host}, $pars{conntype});
        $cmd = "priv set -q diag; " . $cmd;
        # if($pars{host}->{hosttype} =~ /filer-ng/){
            # $cmd = "run local  \"$cmd\"";
        # }
        # else{
        # $cmd = "$cmd";
        # }
        # (undef, undef, $output) = say($conn, $cmd, -prompt, MAINT_PROMPT, '-timeout', $pars{timeout});
        $output = filer_cmd(host=>$pars{host}, conntype =>$pars{conntype}, cmd=>$cmd ,timeout => $pars{timeout});
    } else {
        $conn = make_conn($pars{host}, "#console");
        $output= $conn->execute(command=> $cmd, match_table=>[$re_continue=>'y'],prompts => [("\\*\\>")]);
    }



    if ($output =~ /disk assign:/                         # this should catch most cases
        || $output =~ /disk: Did not recognize option/
        || $output =~ /Ignoring -c option as it is valid only for Gateway Filer/ || $output =~ /Ownership for $pars{disk} cannot be changed/
      ) {
        croak("Disk assign command failed due to: $output");
    }

    return PASS;
}    #disk_assign

=head2 disk_fail

    disk_fail(host=>$host, disk=>$disk, ...);

The routine supports the interface of the disk fail CLI. It fails the specified disk.
Option -i can be specified to remove disk immediately. Option -f to skip any warning.
Refer to the na_disk man page for details information.

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : Hostspec to filer.

=item disk

(Required) : diskname.

=item options

(Optional) : -i|-f fails the disk immediately or turns off warnings.

=item conntype

(Optional) : use of the 3 connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item timeout

(Optional) : connection timeout. Default is 300.

=item waittime

(Optional) : specifies time in seconds. Default is set to wait until done.

=item Notes:

The command is not supported in the maintenance mode.

=back

=item Returns

Return 1 on success or 0 on failure due to one of the following errors:
   1. disk not found
   2. disk is not a file system disk
   3. disk already been prefailed

=item Example:

disk_fail(host=>$host, disk=>$disk, waittime=>'300');

disk_fail(host=>$host, disk=>$disk, options=>'-i', conntype=>"#rsh");

=back

=cut

sub disk_fail {
    _debug("Entering disk_fail: ...");
    my (%pars) =
      validate(@_,
               {host => {isa  => 'Hostrec'},
                disk => {type => SCALAR, optional => NO},
                options => {type     => SCALAR,
                            optional => YES,
                            regex    => qr/-i/,
                            default  => '-f'},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => "#rsh"},
                timeout => {type     => SCALAR,
                            optional => YES,
                            regex    => qr/\d+/,
                            default  => SAY_TIMEOUT},
                waittime => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\d+/,
                             default  => WAIT_UNTIL_DONE},
                api_set_obj => {type     => OBJECT,
                                optional => YES,},
          }
    );

    my $disk     = $pars{disk};
    my $conn     = make_conn($pars{host}, $pars{conntype});
    my $waittime = $pars{waittime};
    my $apiset_obj = $pars{api_set_obj};
    if(!$apiset_obj){
        my $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
    }
    $apiset_obj->set_timeout('connectrec-timeout' => $pars{timeout});
    # ------------------------------------
    # Using -f option as the default value
    # ------------------------------------
    my $options = $pars{options};
    if ($options ne '-f') {
        $options = $options . " -f";
    }

#    my (undef, undef, $output) = say($conn, "disk fail $options $disk", '-timeout', $pars{timeout});
     my $output = $apiset_obj->execute_command(command => "disk fail $options $disk");

    # -----------------------
    # checking for bad result
    # -----------------------
    if (   $output =~ /not a file system/
        || $output =~ /usage: disk <options>/
        || $output =~ /not found/
        || $output =~ /Unrecognized option/
        || $output =~ /cannot fail disk/i
        || $output =~ /disk fail: Has extra arguments/) {
        croak("Disk fail command failed due to: $output");
    }

    # -------------------------
    # checking for valid result
    # -------------------------
    if ($output =~ /disk fail:/) {

        # ----------------------------------------------------
        # if the waittime is set we should wait until the disk
        # appears on the fail list
        # ----------------------------------------------------
        if (defined($pars{waittime})) {
            my $waittime = $pars{waittime};
            my $status = _disk_wait_until($conn, $disk, $waittime, "failed");
            if ($status) {
                tstlog(-comment, "Disk <$disk> was listed on the failed list");
                return PASS;
            }
            tstlog(-comment, "Disk <$disk> was not listed on the failed list after <$waittime> seconds");
            return FAIL;
        }    # waittime
    }    # validate result

    # ---------------------------------------
    # Unable to determine the result. At this
    # point the code needs to be updated to
    # handle this specific scenario
    # ---------------------------------------
    croak("Unable to determine the result of <$output>");
}    #disk_fail

=head2 disk_fw_update

Function is implemented to issue disk_fw_update on the FILER according to different option.Required console connection. And the function will decide whether we will wait for finishing firmware updated according to the parameter $wait. All the paramters will pass into the function through the hash %disk_options. When using the function, the script need to define the parameters according to the different command option of disk_fw_update.

=over

=item host

(Required): Host spec to the FILER;

=item option_name

(Optional) : the possible values can be:
f  represents -f <comma separated disk list>
FT represents -F <firmware-file> -t <product id>
F  represents -F <firmware-file> <disk-list>
Q  represents qualification test
R  represents revert firmware
S  stop qualification test

=item firmware_file

(Optional) : firmware file.

=item product_id:

(Optional) : disk product id.

=item disks

(Optional) : disk list for updating firmware

=item timeout

(optional) : timeout value. Default is 300 secs.

=item wait

(Optional) : 1 - wait for disk_fw_udpate finish. 0 - will not wait for disk_fw_update finish. Default is 0.

=item wait_time

(Optional) : time for waiting disk_fw_update. Default is 300 secs.

=item Example

$ret = disk_fw_update(%disk_option;)

=back

=cut

sub disk_fw_update {
    my (%disk_options) =
      validate(@_,
               {host  => {type => OBJECT,   isa      => 'Hostrec'},
                disks => {type => ARRAYREF, optional => 1, default => []},
                option_name   => {optional => 1},
                timeout       => {type     => SCALAR, default => 300, optional => 1},
                firmware_file => {optional => 1},
                product_id    => {optional => 1},
                wait          => {type     => SCALAR, default => 1, optional => 1},
                wait_time     => {type     => SCALAR, default => 300, optional => 1},});
    my ($host, $cmd, $v, $output, $output1, $output2, $output4, $output3, $conntype);
    $host = host($disk_options{host});
    $conntype = ($host->hosttype =~ /filer-ng/ )? "#ssh" : "#rsh";
    my $conn = make_conn($disk_options{host}, $conntype);
    my @mydate;
    my ($opt)        = $disk_options{option_name};
    my ($Firmware)   = $disk_options{firmware_file};
    my ($product_id) = $disk_options{product_id};
    my ($wait)       = $disk_options{wait};
    my ($wait_time)  = $disk_options{wait_time};
    my ($timeout)    = $disk_options{timeout};
    my $disk_list    = $disk_options{disks};

    if (!defined($opt)) {

        # Issue disk_fw_update without option
        printf "Disk firmware update without any option.\n";
        $cmd = "disk_fw_update -f";
        _diskfwcommand($disk_options{host}, $conntype, $cmd, $wait, $wait_time, $timeout);
    }

    elsif ($opt eq "f") {

        #Issue disk_fw_update -f <disk-list>
        logcomment("Disk firmware update by using option -f");
        $cmd = "disk_fw_update -f @$disk_list";
        # This call fails if the build is not debug.
        _diskfwcommand($disk_options{host}, $conntype, $cmd, $wait, $wait_time, $timeout);
    }

    elsif ($opt eq "FT") {

        #Issue disk_fw_update -F <firmware_file> -t <product_id>
        logcomment("Disk firmware update by using option -F -t");
        $cmd = "disk_fw_update -f -F $Firmware -t $product_id";
        _diskfwcommand($disk_options{host}, $conntype, $cmd, $wait, $wait_time, $timeout);
    }

    elsif ($opt eq "F") {

        #Issue disk_fw_update -F <firmware-file><disklist>
        logcomment("Disk firmware update by using option -F");
        $cmd = "disk_fw_update -f -F $Firmware @$disk_list";
        _diskfwcommand($disk_options{host}, $conntype, $cmd, $wait, $wait_time, $timeout);
    }

    elsif ($opt eq "Q") {

        # Issue disk_fw_update -Q <disk-list> without wait
        logcomment("Disk firmware update by using option -Q for qualification test");
        $cmd  = "disk_fw_update -f -Q @$disk_list";
        $wait = 0;
        _diskfwcommand($disk_options{host}, $conntype, $cmd, $wait, $wait_time, $timeout);
    }

    elsif ($opt eq "R") {

        # Issue disk_fw_update -R <disk-list>
        logcomment("Disk firmware update by using option -R to revert firmware");
        $cmd = "disk_fw_update -f -R @$disk_list";
        _diskfwcommand($disk_options{host}, $conntype, $cmd, $wait, $wait_time, $timeout);
    } elsif ($opt eq "S") {

        # Issue disk_fw_update -f -S <disk-list>
        logcomment("Disk firmware update by using option -S to stop qulification test");
        $wait = 0;
        $cmd  = "disk_fw_update -f -S";
        #say($conn, $cmd, -timeout, $timeout);
        eval{
            _diskfwcommand($disk_options{host}, $conntype, $cmd, $wait, $wait_time, $timeout);
         }
       } else {
        logcomment("there is no such option $opt on the command disk_fw_update");
        croak $opt;
    }
} ## end sub disk_fw_update

=head2 disk_maint

    disk_maint(host=>$host, disklist=>$disk, ...);

The routine performs disk maint on the specified list of disks.
If either maint list or maint status is specified, the routine
shall return the reference of a hash table with the following
keys and values:

=begin html

Output of the maint list command:

 $VAR1 = {
          '0' => {
                   'Test_Id' => 'ws',
                   'Test_name' => 'Write Same Test'
                 },
          1 => {
                 'Test_Id' => 'ndst',
                 'Test_name' => 'NDST Test'
               },
          2 => {
                 'Test_Id' => 'endst',
                 'Test_name' => 'Extended NDST Test'
               },
          3 => {
                 'Test_Id' => 'vt',
                 'Test_name' => 'Verify Test'
               },
          4 => {
                 'Test_Id' => 'ss',
                 'Test_name' => 'StartStop Test'
               },
          5 => {
                 'Test_Id' => 'dt',
                 'Test_name' => 'Data Integrity Test'
               },
          6 => {
                 'Test_Id' => 'rdt',
                 'Test_name' => 'Read Test'
               }
          7 => {
                 'Test_Id' => 'pc',
                 'Test_name' => 'Power Cycle Test'
               }
        };

Output of the maint status command:

    $VAR1 = {
    '10a.38' => {
                        'cycle' => '1/1',
                        'Test_name:' => {
                                        'Data_Integrity_Test' => {
                                                                  'Cycle:' => 1,
                                                                  'Iteration:' => '0/1',
                                                                  'Status:' => 'Not_started'
                                 },
                    'NDST_Test' => {
                                                           'Time:' => '00:45:42',
                                                           'Cycle:' => 1,
                                                           'Iteration:' => '1/1',
                                                           'Status:' => 'Running'
                               },
                    'Verify_Test' => {
                                                             'Time:' => '00:00:42',
                                                             'Cycle:' => 1,
                                                             'Iteration:' => '1/1',
                                                             'Status:' => 'Success'
                             },
                    'StartStop_Test' => {
                                                                'Time:' => '00:00:34',
                                                                'Cycle:' => 1,
                                                                'Iteration:' => '1/1',
                                                                'Status:' => 'Success'
                                },
                    'Write_Same_Test' => {
                                                                 'Time:' => '00:00:42',
                                                                 'Cycle:' => 1,
                                                                 'Iteration:' => '0/1',
                                                                 'Status:' => 'Not_started'
                                 }
            },
                        'test' => '4/6'
            },
...

=end html

=over

=item Arguments

=over

=item host

(Required) : Hostspec of filer

=item maintcmd

(Required) : Accepts one of the four available commands: start, abort, status, and list.

=item disklist

(Optional) : Lists one or more disks. However, the disklist argument is required for both 'disk maint start' and 'disk maint abort'.

=item conntype

(Optional) : Supports connection type of: '#rsh', '#console', and '#telnet'. Default is '#rsh'.

=item timeout

(Optional) : Connection time out in seconds. (Default is 300)

=item testlist

(Optional) : Specify a list of tests to run (test index or test id). A list of test could be found using "disk maint list".

=item cyclecount

(Optional) : Specify a number of cycles to run. Default is "1".

=item confirmation

(Optional) : Skip warning and confirmation using -f. Default is "-f".

=item immediate

(Optional) : Avoids recovery and removes the disk immediately. Default is "".

=item fullstatus

(Optional) : This option can only be specified with the disk maint status.
Specify -v to display full status of disk maint. Default is "".

=item waittime

(Optional) : specifies time in seconds. Default is set to wait until done.

=back

=item Notes

The command is available for both filer and simulator.

=item Returns

It returns the reference of a hash table if the maint status or maint list
is specified.

=item Example:

    # starts disk maint test on the specified disks list and waits until done
    disk_maint(host=>$host, maintcmd=>'start', disklist=>$disk);

    # starts disk maint without waiting
    disk_maint(host=>$host, maintcmd=>'start', disklist=>$disk, waittime=>'0');

    # returns a disk maint status via a hash reference
    my $hr = disk_maint(host=>$host, maintcmd=>'status');

    # returns a maint test_list via a hash reference
    my $hr = disk_maint(host=>$host, maintcmd=>'list');

=back

=cut

sub disk_maint {
    _debug("Entering disk_maint: ...");
    my (%pars) =
      validate(@_,
               {host     => {isa => 'Hostrec'},
                maintcmd => {type     => SCALAR,
                             optional => NO,
                             regex    => qr/start|abort|status|list/,
                             default  => ""},
                disklist     => {type => SCALAR, optional => YES, default => ""},
                testlist     => {type => SCALAR, optional => YES},
                confirmation => {type => SCALAR, optional => YES, default => "-f"},
                immediate    => {type => SCALAR, optional => YES, default => ""},
                testlist     => {type => SCALAR, optional => YES},
                cyclecount   => {
                               type     => SCALAR,
                               optional => YES,
                               regex    => qr/\d+/,
                               default  => "1"},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => "#rsh"},
                fullstatus => {type     => SCALAR,
                               optional => YES,
                               regex    => qr/-v/,
                               default  => ""},
                timeout => {type     => SCALAR,
                            optional => YES,
                            regex    => qr/\d+/,
                            default  => SAY_TIMEOUT},
                waittime => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\d+/,
                             default  => WAIT_UNTIL_DONE},});

    my $conn = make_conn($pars{host}, $pars{conntype});

    # -------------------
    # process maint start
    # -------------------
    if ($pars{maintcmd} eq "start") {
        my $testlist   = $pars{testlist}   ? "-t " . $pars{testlist}   : "";
        my $cyclecount = $pars{cyclecount} ? "-c " . $pars{cyclecount} : "";
        my $output     = "";

        # --------------------------------------
        # if waittime == 0 don't specify timeout
        # --------------------------------------
        if ($pars{waittime} == 0) {
            (undef, undef, $output) = say($conn, "disk maint start $testlist $cyclecount $pars{immediate} $pars{confirmation} -d $pars{disklist}");
        } else {
            (undef, undef, $output) = say($conn, "disk maint start $testlist $cyclecount $pars{immediate} $pars{confirmation} -d $pars{disklist}", '-timeout', $pars{timeout});
        }

        if ($output =~ /disk maint:/) {
            croak($output);
        }

        # --------------------------------------
        # if waittime == 0 bail out
        # --------------------------------------
        if ($pars{waittime} == 0) {
            return PASS;
        }

        my $cnt       = WAIT_UNTIL_DONE;
        my $sleeptime = 30;

        while ($cnt < $pars{waittime} || $pars{waittime} == WAIT_UNTIL_DONE) {

            # --------------------------------
            # check if the disk state is ready
            # --------------------------------
            my $hr;
            eval {

                # -----------------------------------------------------------
                # Must delay checking maint status for ATA drive & simmulator
                # -----------------------------------------------------------
                sleep(5);
                $hr = disk_maint(host     => $pars{host},
                                 disklist => $pars{disklist},
                                 maintcmd => 'status');
            };
            if ($@) {
                croak($@);
            }

            # --------------------------------------------------------
            # if the status list is empty then disk maint test is done
            # --------------------------------------------------------
            if (keys(%$hr) < 1) {
                return PASS;
            }

            # -----------------------------------------------
            # Checking the status of the requested disks only
            # -----------------------------------------------
            my $no_testing = 1;
            foreach (keys(%$hr)) {
                if ($hr->{$_}->{Status} eq "Testing") {
                    $no_testing = 0;
                    last;
                }
            }

            if ($no_testing) {
                tstlog(-comment, "Disk maint testing completed in <$cnt> secs");
                return PASS;
            }

            sleep($sleeptime);
            $cnt += $sleeptime;
        } ## end while ($cnt < $pars{waittime...

        # waiting until disk maint is completed

        tstlog(-comment, "Maint testing of <$pars{disklist} exceeds waittime of <$pars{waittime}>");
        return PASS;
    }    #disk maint start

    # -------------------
    # process maint abort
    # -------------------
    if ($pars{maintcmd} eq "abort") {
        my (undef, undef, $output) = say($conn, "disk maint abort $pars{disklist}", '-timeout', $pars{timeout});

        # -----------------------------------------
        # Handling errors.
        # If one of the disk failed, then the whole
        # request is considered as having problem.
        # -----------------------------------------
        if (   $output =~ /at least one/
            || $output =~ /disk maint: Not testing/
            || $output =~ /Could not find/
            || $output =~ /usage: disk <options>/) {
            croak($output);
        }

        if ($output =~ /disk maint: Aborting testing/) {
            return PASS;
        }
    }    # disk maint abort

    # --------------------
    # process maint status
    #
    # --------------------
    if ($pars{maintcmd} eq "status") {
        my (undef, undef, $output) = say($conn, "disk maint status $pars{fullstatus} $pars{disklist}", '-timeout', $pars{timeout});
        if (   $output =~ /disk maint: Could not find/
            || $output =~ /duplicate disk name/
            || $output =~ /disk maint: Unregonized option/
            || $output =~ /disk maint: Status has failed on disk/
            || $output =~ /usage: disk <options>/) {
            croak($output);
        }

        # -----------------------
        # when there is no status
        # -----------------------
        if ($output eq "") {
            return "";
        }

        my %ht;

        my @lines = split(/\r*\n/, $output);

        # ---------------------
        # parse each maint disk
        # ---------------------
        my @header;
        foreach (@lines) {

            # --------------------------------------------
            # retrieve the disk name, test, and cycle info
            # --------------------------------------------
            if (/^Testing/) {
                (@header) = split(/ /, $_);
                $ht{$header[1]}->{Status}     = $header[0];
                $ht{$header[1]}->{$header[2]} = $header[3];
                $ht{$header[1]}->{$header[4]} = $header[5];
                next;
            }

            if (/^Copying/) {
                (@header) = split(/ /, $_);
                $ht{$header[1]}->{Status} = $header[0];
                next;
            }

            # handling disk is not being tested
            if (/disk maint: Not testing/) {
                (@header) = split(/ /, $_);
                $ht{$header[4]}->{Status} = "Not_testing";
                next;
            }

            # Replacing space with _ for the test name
            s/([a-zT]) ([a-zA-Z])/$1_$2/g;
            my @cols = split(/\s+|\t/, $_);
            $ht{$header[1]}->{$cols[0]}->{$cols[1]}->{$cols[2]} = $cols[3];
            $ht{$header[1]}->{$cols[0]}->{$cols[1]}->{$cols[4]} = $cols[5];
            $ht{$header[1]}->{$cols[0]}->{$cols[1]}->{$cols[6]} = $cols[7];
            if (defined($cols[8])) {
                $ht{$header[1]}->{$cols[0]}->{$cols[1]}->{$cols[8]} = $cols[9];
            }
        }    # for each status line
        return \%ht;
    }    # process maint status

    # ------------------
    # process maint list
    # ------------------
    if ($pars{maintcmd} eq "list") {
        my (undef, undef, $output) = say($conn, "disk maint list", '-timeout', $pars{timeout});

        if ($output =~ /disk: Did not recognize option/) {
            croak($output);
        }

        my %ht;
        $output =~ s/Test index:/Test_index/g;
        $output =~ s/Test Id:/Test_Id/g;
        $output =~ s/Test name:/Test_name/g;

        my @lines = split(/\r*\n/, $output);
        shift(@lines);

        foreach (@lines) {
            my @cols = split(/\s+|\t/, $_);

            # ------------------------
            # create the hash of index
            # ------------------------
            $ht{$cols[1]}->{$cols[2]} = $cols[3];
            if ($#cols > 6) {
                $ht{$cols[1]}->{$cols[4]} = $cols[5] . " " . $cols[6] . " " . $cols[7];
            } else {
                $ht{$cols[1]}->{$cols[4]} = $cols[5] . " " . $cols[6];
            }
        } ## end foreach (@lines)
        return \%ht;
    } ## end if ($pars{maintcmd} eq...

}    #disk_maint

=head2 disk_remove

    disk_remove(host=>$host, disk=>$disk, ...);

The routine removes the specified disk with either option -w or -f.
Option -w is only be used for the Gateway filer.

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : Hostspec of filer

=item disk

(Required) : Disk name

=item options

(Optional) : Options -f|-w options. Default is none.

=item conntype

(Optional) : Supports connection type of: '#rsh', '#console', and '#telnet'. Default is '#rsh'.

=item timeout

(Optional) : Connection time out in seconds. (Default is 300)

=item waittime

(Optional) : Wait until the disk is being removed or out of time. Default is wait until the disk is completely been removed.

=back

=item Notes

The command is available for all priv modes.

=item Returns

Returns 1 on success or 0 on failure due to multiple errors:
   1. disk not found
   2. disk is not a file system disk
   3. disk already been prefailed

=item Example:

    $ret = disk_remove(host=>$host, disk=>$disk, options=>'-f', timeout=>'300', waittime=>'300');

=item See Also

disk_fail, disk_unfail

=back

=cut

sub disk_remove {
    _debug("Entering disk_remove: ...");
    my (%pars) =
      validate(@_,
               {host    => {isa  => 'Hostrec'},
                disk    => {type => SCALAR, optional => NO},
                options => {type => SCALAR, optional => YES, regex => qr/-w|-f/, default => '-f'},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => "#rsh"},
                timeout => {type     => SCALAR,
                            optional => YES,
                            regex    => qr/\d+/,
                            default  => SAY_TIMEOUT},
                waittime => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\d+/,
                             default  => WAIT_UNTIL_DONE},});

    my $disk = $pars{disk};
    my $conn = make_conn($pars{host});

    # ------------------------------------
    # Using -f option as the default value
    # ------------------------------------
    my $options = $pars{options};
    if ($options ne '-f') {
        $options = $options . " -f";
    }

    my (undef, undef, $output) = say($conn, "disk remove $options $disk", '-timeout', $pars{timeout});

    # ------------------------
    # checking for bad results
    # ------------------------
    if (   $output =~ /is not a spare disk/
        || $output =~ /usage: disk <options>/
        || $output =~ /not found/
        || $output =~ /Unrecognized option/
        || $output =~ /disk remove: Cannot remove disk (raid group degraded)/
        || $output =~ /disk remove: Has extra arguments/) {
        croak("Disk remove command failed due to: $output");
    }

    # --------------------------
    # checking for the option -w
    # --------------------------
    if ($output =~ /-w options as it is valid only for Gateway Filer/) {
        tstlog(-comment, "Using the option -w is only for the Gateway Filer");
    }

    # --------------------------
    # checking for valid results
    # --------------------------
    if ($output =~ /disk remove:/) {

        # ----------------------------------------------------
        # if the waittime is set we should wait until the disk
        # appears on the fail list
        # ----------------------------------------------------
        if (defined($pars{waittime})) {
            my $waittime = $pars{waittime};
            my $status = _disk_wait_until($conn, $disk, $waittime, "failed");
            if ($status) {
                tstlog(-comment, "The removed disk <$disk> was listed on the failed list");
                return PASS;
            }
            tstlog(-comment, "The removed disk <$disk> was not listed on the failed list after <$waittime> seconds");
            return FAIL;
        } ## end if (defined($pars{waittime...
    }    # validate result

    # ---------------------------------------
    # Unable to determine the result. At this
    # point the code needs to be updated to
    # handle this specific scenario
    # ---------------------------------------
    croak("Unable to determine the result of disk remove command: <$output>");
}    #disk_remove

=head2 disk_remove_ownership

    disk_remove_ownership(host=>$hostname, $disk=>disk, ...);

The routine provides the interface of "disk remove_ownership" CLI.
The command is available in both maint and normal mode.

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : Hostspec of filer

=item disk

(Required) : Disk name or "all". "all" can only be used in Maintenance mode.

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item timeout

(Optional) : connection timeout. Default is 300.

=item sysid

(Optional) : Specify a sysid. This option is only allowed in Maintenance mode.

=item foption

(Optional) : specify [-f] option. Default is "".

=back

=item Returns

Returns an error message via $@ upon any error.

=back

=cut

sub disk_remove_ownership {
    _debug("Entering disk_remove_ownership: ...");
    my (%pars) =
      validate(@_,
               {host => {isa  => 'Hostrec'},
                disk => {type => SCALAR, optional => NO},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => "#rsh"},
                timeout => {type     => SCALAR,
                            optional => YES,
                            regex    => qr/\d+/,
                            default  => SAY_TIMEOUT},
                filermode => {type     => SCALAR,
                              optional => YES,
                              regex    => qr/maint|normal/,
                              default  => NORMAL_MODE},
                sysid   => {type => SCALAR, optional => YES},
                foption => {type     => SCALAR,
                            optional => YES,
                            regex    => qr/-f/,
                            default  => ""},});

    my $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
    my $api_set_obj_cli = $filer_node->get_7m_or_nodescope_apiset();
    my $sysid = defined($pars{sysid}) ? " -s " . $pars{sysid} : "";
    my ($cmd, $conn,$re_continue);
    $re_continue = "Volumes must be taken offline. Are all impacted volumes offline\\(y/n\\)\\?\\? ";
    if ($pars{filermode} eq MAINT_MODE) {
        $conn = make_conn($pars{host}, "#console");
        $cmd = "disk remove_ownership $pars{disk} $sysid $pars{foption}";
        my $re_continue1 = "Do you wish to continue\\? ";
        my $output = $conn->execute(command=> $cmd, match_table=>[$re_continue1=>'y',$re_continue=>'y'],prompts => [("\\*\\>")]);
    } else {
        $conn = make_conn($pars{host}, "#console");
        $cmd = "priv set -q diag; disk remove_ownership $pars{disk} $sysid $pars{foption}";
        if($pars{host}->{hosttype} =~ /filer-ng/){
            $cmd = "run local  \"$cmd\"";
        } else{
            $cmd = "$cmd";
        }
        my $output = $conn->execute(command=> $cmd, match_table=>[$re_continue=>'y'],timeout => 300,);
        my $response = $api_set_obj_cli->disk_show("not-owned" =>1, disk => "$pars{disk}");
        if($response->{output} !~ /$pars{disk}/){
            croak("Unable to remove the ownership of disk <$pars{disk}>");
        }
    }
    return PASS;
}    #disk_remove_ownership

=head2 disk_replace

   disk_replace(host=>$host, cmd=>$cmd, disk=>$disk_name, spare_drive=>$spare);

The routine supports the interface of the disk replace CLI. It replaces
the specified disk. Option -f to skip any warning.
Refer to the na_disk man page for detailed information.

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : Hostspec to filer.

=item disk

(Required) : diskname.

=item spare_drive

(Required) : spare drive name

=item options

(Optional) : -f turns off warnings.

=item conntype

(Optional) : use of the 3 connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item timeout

(Optional) : connection timeout. Default is 300.

=item waittime

(Optional) : specifies time in seconds. Default is set to wait until done.

=item Notes:

The command is not supported in the maintenance mode.

=back

=item Returns

Return 1 on success or 0 on failure:

=cut

sub disk_replace {
    _debug("Entering disk_replace: ...");
    my (%pars) =
      validate(@_,
               {host => {isa  => 'Hostrec'},
                cmds => {type => SCALAR, optional => NO, regex => qr/start|stop/},
                disk        => {type => SCALAR, optional => NO},
                spare_drive => {type => SCALAR, optional => NO},
                options     => {type => SCALAR, optional => YES, default => "-f"},
                timeout     => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                waittime => {type => SCALAR, optional => YES, default => WAIT_UNTIL_DONE},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => '#rsh'},});

    my $disk  = $pars{disk};
    my $spare = $pars{spare_drive};
    my $conn  = make_conn($pars{host}, $pars{conntype});

    # ------------------------------------
    # Using -f option as the default value
    # ------------------------------------
    my $options = $pars{options};
    if ($options ne '-f') {
        $options = $options . " -f";
    }

    my $output;
    if ($pars{cmds} eq 'start') {
        (undef, undef, $output) = say($conn, "disk replace start $options $disk $spare", '-timeout', $pars{timeout});
    }

    # -------------------------
    # checking for valid result
    # -------------------------
    if ($output =~ /disk replace:/) {

        # ----------------------------------------------------
        # if the waittime is set will wait until the disk
        # appears on the spare list
        # ----------------------------------------------------
        if (defined($pars{waittime})) {
            my $waittime = $pars{waittime};
            my $status = _disk_wait_until($conn, $disk, $waittime, "spare");
            if ($status == PASS) {
                tstlog(-comment, "Disk <$disk> was listed on the spare list");
                return PASS;
            }
            tstlog(-comment, "Disk <$disk> was not listed on the spare list after <$waittime> seconds");
            return FAIL;
        }    # waittime
    } else {
        croak("Disk replace command failed due to: $output");
    }
}    #disk_replace

=head2 disk_resume

    disk_resume(host=>$host, disk=>$disk, ...);

The routine initiates the resume of I/O to the specified disk via the 'disk resume' CLI.

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : Hostspec to filer

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item disk

(Required) : Specify the disk to be resumed.

=item timeout

(Optional) : connection timeout. Default is 300.

=item waittime

(Optional) : Specify time in secs to wait until IO is resumed on the specified disk.
Default is set to wait_until_done.

=back

=item Returns

Returns 1 upon success else returns 0.
On any error the routine returns the error message via $@.

=back

=cut

sub disk_resume {
    _debug("Entering disk_resume: ...");
    my (%pars) =
      validate(@_,
               {host     => {isa  => 'Hostrec'},
                disk     => {type => SCALAR, optional => NO},
                timeout  => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                waittime => {type => SCALAR, optional => YES, regex => qr/\d+/, default => WAIT_UNTIL_DONE},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => '#rsh'},});

    my $conn = make_conn($pars{host}, $pars{conntype});
    my (undef, undef, $output) = say($conn, "priv set -q diag; disk resume $pars{disk}", '-timeout', $pars{timeout});

    # check for errors
    if (   $output =~ /disk:/
        || $output =~ /disk resume\/resume:/) {
        croak("Unable to resume disk <$pars{disk}> due to: <$output>");
    }

    my $cnt       = WAIT_UNTIL_DONE;
    my $sleeptime = 10;
    while ($cnt < $pars{waittime} || $pars{waittime} == WAIT_UNTIL_DONE) {

        # --------------------------------
        # check if the disk state is ready
        # --------------------------------
        my $hr;
        eval { $hr = disk_shm_stats(host => $pars{host}, disk => $pars{disk}); };
        if ($@) {
            tstlog(-fail, "Unable to resume disk <$pars{disk}> due to: <$@>");
            return FAIL;
        } else {
            tstlog('-pass', "Able to resume disk <$pars{disk}>");
            return PASS;
        }

        sleep($sleeptime);
        $cnt += $sleeptime;
    }    # waiting until IO is resumed or run out of time
    return PASS;
}    #disk_resume

=head2 disk_sanitize

    my $ht = disk_sanitize(host=>$host, cmd=>$cmd, disk=>$disk, ...);

To start, abort, release, or obtain status of the disk sanitization.
One or more disks need to be specified per command. There are certain
patterns to be used for sanitizing the disk. Refer to the na_disk man
pages for the specific patterns specification.

The routine takes the following parameters.

Note - The routine assumes that the sanitize license has been activated.

=over

=item Arguments

=over

=item host

(Required) : Hostrec of filer

=item cmds

(Required) : Specify one of the sanitize commands: start, abort, status, release.

=item disk

(Required) : Specify one or multiple disk names.

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item timeout

(Optional) : connection timeout. Default is 300.

=item waittime

(Optional) : Specify time in secs to wait until the disks are sanitized completely or run out of time. Default is set to wait_until_done. Set to 0 if no wait.

=item options

(Optional) : Available options are -f, -j, -s , -p <pattern>, -r, and -c <count>. The options can be specified all together or separated.

=back

=item Returns

Returns 1 on success. If there is an error the error shall be return as an exception via $@.

=back

=cut

sub disk_sanitize {
    _debug("Entering disk_sanitize: ...");
    my (%pars) =
      validate(@_,
               {host    => {isa  => 'Hostrec'},
                cmds    => {type => SCALAR, optional => NO, regex => qr/start|abort|status|release/},
                disk    => {type => SCALAR, optional => YES, default => ""},
                timeout => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                waittime => {type => SCALAR, optional => YES, default => WAIT_UNTIL_DONE},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => '#rsh'},
                options => {type => SCALAR, optional => YES, default => ""},});

    my $options = $pars{options};

    # The connection type must be persistent for start if the cluster is
    # disabled to answer the question. If the tests are rebooting the
    # filer then there is a problem with telnet connections beign disconnected
    # and NATE doesn't know to do a login again. Easiest just to run
    # with the console.
    if ($pars{cmds} eq "start") {
        $pars{conntype} = "#console";
    }
    my $conn;
    my $filer_ci;
    my $api_set_obj;
    $filer_ci = NACL::C::Node->find();
    $api_set_obj = $filer_ci->get_7m_or_nodescope_apiset();
    my $host = host($pars{host});
    #If there are multiple disk, get the diks list in to an array and then check the active path of each disks
    my @disks = split(/\s+/, $pars{disk});
    my $disk_list = ' ';
    foreach my $disk_name (@disks) {
        my $disk_name1 = RaidSqa_find_active_path_disk_name(host => $pars{host}, conntype => $pars{conntype}, disk_name => $disk_name);
        if ($disk_name1) {
            logcomment("RaidSqa.pm: Using path name $disk_name1 for any operations on $disk_name.\n");
            $disk_list .= $disk_name1 . " ";
        } else {
            logcomment("RaidSqa.pm: The disk_name $disk_name specified does not exist. Please check again.\n");
            return (0);
        }
    } ## end foreach my $disk_name (@disks)

    # ----------------------
    # process sanitize start
    # ----------------------
    if ($pars{cmds} eq "start") {
        if ($options !~ /-f/) {
            $options = $options . " -f ";
        }
        my $output;
        my $cmd = "disk sanitize $pars{cmds} $options $disk_list";
        my $prompts_answers = [ '.*Would you like to continue \(y/n\)\?*' => 'y'];
        $output = $api_set_obj->execute_command('command' => $cmd, 'connectrec-match_table' => $prompts_answers);

        # Sometimes we get DBG messages interspersed with the output.
        # This code will remove the debug messages.
        my @lines = split(/\n/, $output);

        # Replace the DBG: message with a XXX marker
        foreach my $line (@lines) {
            $line =~ s/DBG:.*$/XXX/;
        }

        # Remove empty lines as this may stick whitespace in the middle
        # of a word after doing the join
        my $i = 0;
        while (defined($lines[$i])) {
            if ($lines[$i] eq 'XXX') {
                splice(@lines, $i, 1);
            } else {
                $i++;
            }
        }

        # Look for lines where a debug message split a line so rejoin it here
        $i = 0;
        while (defined($lines[$i])) {
            if ($lines[$i] =~ /XXX$/) {
                $lines[$i] =~ s/XXX$//;
                $lines[$i] .= $lines[$i + 1];
                $i++;
                splice(@lines, $i, 1);
            } else {
                $i++;
            }
        } ## end while (defined($lines[$i]...

        # Join the lines back together
        $output = join(' ', @lines);

        if ($output !~ /sanitization\s+process\s+has\s+been\s+initiated/) {
            croak("Disk sanitize failed due to: <$output>");
        }

        # No wait
        if ($pars{waittime} == 0) {
            return PASS;
        }

        # ---------------------------------------------
        # wait until done or exceeds the specified time
        # ---------------------------------------------
        my $cnt       = WAIT_UNTIL_DONE;
        my $sleeptime = 60;

        # -----------------------------------------------
        # create a table for tracking status of each disk
        # -----------------------------------------------
        my %hd;
        foreach (split(/ /, $disk_list)) {
            $hd{$_} = 0;
        }
        while ($cnt <= $pars{waittime} || $pars{waittime} == WAIT_UNTIL_DONE) {
            foreach (keys %hd) {
                $output = filer_cmd(host=>$pars{host}, conntype =>$pars{conntype}, cmd=>"disk sanitize status $_" ,timeout => $pars{timeout});
                print Dumper $output;

                # check the output just in case thing goes really bad
                if ($output !~ /sanitization for/) {
                    croak("Disk sanitize status failed due to: <$output>");
                }
                if ($output =~ /sanitization for \S+ is 100/) {
                    $hd{$_} = 1;
                }
            }    # checking each disk

            # --------------------------------------------------
            # if sanitize is completed within the specified time
            # --------------------------------------------------
            my $is_done = 1;
            foreach (keys %hd) {
                if ($hd{$_} == 0) {
                    $is_done = 0;
                    last;
                }
            }

            if ($is_done) {
                tstlog(-comment, "disk sanitize completed from start to finish of <$disk_list>");
                return PASS;
            }
            _debug("waiting for disk sanitize status after spending <$cnt> secs");
            sleep($sleeptime);
            $cnt += $sleeptime;
        }    #wait_until_done

        # ----------------------------------------------------------------
        # By now the sanitize has not completed within the specified time
        # ----------------------------------------------------------------
        tstlog(-comment, "Sanitizing disk <$disk_list> did not complete within the specified time of <$pars{waittime}> secs");

    }    # sanitize start

    _debug("after start sanitize: ...");

    # ---------------------------------------
    # initiate disk sanitize release or abort
    # ---------------------------------------
    if ($pars{cmds} =~ /release|abort/) {
        my $output = filer_cmd(host=>$pars{host}, conntype =>$pars{conntype}, cmd=>"disk sanitize $pars{cmds} $disk_list" ,timeout => $pars{timeout});
        if (   $output =~ /Respared sanitized disks/
            || $output =~ /Sanitization abort initiated/) {
            tstlog(-comment, "Initiated disk sanitize release/abort successfully");
            return PASS;
        }
        croak("Disk sanitize release/abort <$disk_list> failed due to: <$output>");
    }    # disk sanitize release

    _debug("after release sanitize: ...");

    # ---------------------------------
    # retrieve the disk sanitize status
    # ---------------------------------
    if ($pars{cmds} eq "status") {
        _debug("sanitize status ...");
        my $output = filer_cmd(host=>$pars{host}, conntype =>$pars{conntype}, cmd=>"disk sanitize status" ,timeout => $pars{timeout});
        print Dumper $output;
        my %ht = ();

        # -------------------------------
        # when there is no sanitized disk
        # -------------------------------
        if ($output eq "") {
            tstlog(-comment, "There is no sanitized disk");
            return \%ht;
        }

        if ($output !~ /sanitization for/) {
            croak("Retrieving disk sanitize status failed due to: <$output>");
        }

        my @lines = split(/\r*\n/, $output);
        foreach (@lines) {
            my @cols = split(/ /, $_);
            $ht{$cols[2]}->{percent_complete} = $cols[4];
        }    #each sanitized disk
        return \%ht;
    }    # init disk sanitize status
}    #disk_sanitize

=head2 disk_shm_stats

    my $ht = disk_shm_stats(host=>$host, disk=>$disk);

The routine retrieves the health status of the specified disk or all disks.
The retrieved statuses will be returned via a hash table by reference.
The routine should handles different output formats of the specified disk.

The routine takes the following parameters.

=begin html

 The general format of multiple disks:

 $VAR1 = {
          '2a.20' => {
                       'Sns_ErrKey_4' => '0',
                       'IO_TOUT_COUNT' => '000',
                       'ThruPut_RateKBPS' => 56797,
                       'IO_RETRY_COUNT' => '000',
                       'IO_COMP_TIME' => '008',
                       'IO_ABT_COUNT' => '000',
                       'Sns_ErrKey_1' => '0',
                       'Sns_ErrKey_2' => '0',
                       'Sns_ErrKey_3' => '0'
                     },

         '2a.35' => {
                       'Sns_ErrKey_4' => '0',
                       'IO_TOUT_COUNT' => '000',
                       'ThruPut_RateKBPS' => '0842',
                       'IO_RETRY_COUNT' => '000',
                       'IO_COMP_TIME' => '031',
                       'IO_ABT_COUNT' => '000',
                       'Sns_ErrKey_1' => '0',
                       'Sns_ErrKey_2' => '0',
                       'Sns_ErrKey_3' => '0'
                     },
            ...

However, the individual disk information could be formatted differently.
There are 2 general formats:

 Format 1:
 $VAR1 = {
          '2a.20' => {
                       'DiskSerialNumber' => 'A829MRLE',
                       'ATTR' => {
                                   29 => {
                                           'VALUE' => '004h',
                                           'STATUS' => '64h',
                                           'RAW_VAL' => 253,
                                           'WRST_VAL' => 253,
                                           'ATTR_ID' => ''
                                         },
                                   1 => {
                                          'VALUE' => '027h',
                                          'STATUS' => '03h',
                                          'RAW_VAL' => 204,
                                          'WRST_VAL' => 204,
                                          'ATTR_ID' => ''
                                        },
                                    ...

                                  28 => {
                                           'VALUE' => '004h',
                                           'STATUS' => '63h',
                                           'RAW_VAL' => 253,
                                           'WRST_VAL' => 253,
                                           'ATTR_ID' => ''
                                         }
                                 },
                       'HA' => '2a',
                       'BAY' => 4,
                       'SHELF' => 1,
                       'DataStructRevNumber' => '0010h',
                       'HA_ID' => '2a.20',
                       'CHAN' => 'FC:B'
                     }
        };

Format 2:

 $VAR1 = {
          '11a.22' => {
                        'DiskSerialNumber' => '3HW1KLZ6000074525WH9',
                        'BAY' => 6,
                        'Read' => {
                                    'Tot_UNCErr' => '000000000',
                                    'Tot_Retry' => '000000000',
                                    'Tot_Err_Cor' => '000000000',
                                    'Tot_BytProc' => '000000000',
                                    'Err_CWPD' => '000000000'
                                  },
                        'Thruput_KBPS' => 52610,
                        'IO_Abt_Num' => '000',
                        'HA_ID' => '11a.22',
                        'Err_SKey1' => '0',
                        'Err_SKey3' => '0',
                        'Err_Skey2' => '0',
                        'WrtVfy' => {
                                      'Tot_UNCErr' => '000000000',
                                      'Tot_Retry' => '000000000',
                                      'Tot_Err_Cor' => '000000000',
                                      'Tot_BytProc' => '000000000',
                                      'Err_CWPD' => '000000000'
                                    },
                        'Err_SKey4' => '0',
                        'IO_Long_msec' => 647,
                        'IO_Avg_msec' => '010',
                        'IO_Tout_Cnt' => '000',
                        'Write' => {
                                     'Tot_UNCErr' => '000000000',
                                     'Tot_Retry' => '000000000',
                                     'Tot_Err_Cor' => '000000000',
                                     'Tot_BytProc' => '000000000',
                                     'Err_CWPD' => '000000000'
                                   },
                        'HA' => '11a',
                        'SHELF' => 1,
                        'Disk_Log_Header' => 'Disk LOG Sense Stats',
                        'CHAN' => 'FC:A',
                        'IO_Retry_Cnt' => '000'
                      }
        };

 Format 3:

 $VAR1 = {
          '0a.33' => {
                       'DiskSerialNumber' => '3KR0TVL6000076045P5D',
                       'Disk_Log' => 'Disk LOG Sense Statistics',
                       'BAY' => 1,
                       'Read' => {
                                   'Corr_Algorithm' => 92384,
                                   'Errs_corrected' => 92384,
                                   'Errs_w/delay' => '0',
                                   'Uncorrected' => '0',
                                   'Bytes_processed' => '691916096040'
                                 },
                       'Thruput_KBPS' => 745,
                       'IO_Abt_Num' => '000',
                       'HA_ID' => '0a.33',
                       'Err_SKey1' => '0',
                       'Err_SKey3' => '0',
                       'Err_Skey2' => '0',
                       'Err_SKey4' => '0',
                       'IO_Long_msec' => 63988,
                       'IO_Avg_msec' => '029',
                       'IO_Tout_Cnt' => '000',
                       'Write' => {
                                    'Corr_Algorithm' => '0',
                                    'Errs_corrected' => '0',
                                    'Errs_w/delay' => '0',
                                    'Uncorrected' => '0',
                                    'Bytes_processed' => '1204615676160'
                                  },
                       'HA' => '0a',
                       'SHELF' => 2,
                       'Verify' => {
                                     'Corr_Algorithm' => 1089522,
                                     'Errs_corrected' => 1089522,
                                     'Errs_w/delay' => '0',
                                     'Uncorrected' => '0',
                                     'Bytes_processed' => '5344189074640'
                                   },
                       'CHAN' => 'FC:A',
                       'IO_Retry_Cnt' => '000'
                     }
        };

=end html

=over

=item Arguments

=over

=item host

(Required) : Hostspec to filer

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item disk

(Optional) : Specify the disk to be retrieved.

=item timeout

(Optional) : connection timeout. Default is 300.

=back

=item Returns

Returns 1 upon success else returns 0.
On any error the routine returns the error message via $@.

=back

=cut

sub disk_shm_stats {
    _debug("Entering disk_shm_stats: ...");
    my (%pars) =
      validate(@_,
               {host    => {isa  => 'Hostrec'},
                disk    => {type => SCALAR, optional => YES, default => ""},
                timeout => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => '#rsh'},});

    my $conn = make_conn($pars{host}, $pars{conntype});
    my (undef, undef, $output) = say($conn, "priv set -q diag; disk shm_stats $pars{disk}", '-timeout', $pars{timeout});

    # check for errors
    if (   $output =~ /Disk State not READY/
        || $output =~ /disk shm_stat:/) {
        croak("Retrieving shm_stats of disk <$pars{disk}> failed due to: <$output>");
    }

    # ---------------------------------------------------------------
    # parsing a list of disk. Let simply the parsing process for now
    # by assuming that the header is not going to be changed for a while
    # The Header block looks like: DISK, IO_COMP_TIME, IO_TOUT_COUNT
    # IO_RETRY_COUNT, IIO_ABT_COUNT, ThruPut_RateKBPS, and Sns_ErrKey
    # ---------------------------------------------------------------
    my %ht;
    my @lines = split(/\r*\n/, $output);
    if ($output =~ /DISK/ && $output =~ /IO_COMP/) {
        for (my $i = 0 ; $i < 5 ; $i++) { shift(@lines) }
        foreach (@lines) {
            my (@cols) = split(/\s+|\t/);
            $ht{$cols[0]}->{IO_COMP_TIME}     = $cols[1];
            $ht{$cols[0]}->{IO_TOUT_COUNT}    = $cols[2];
            $ht{$cols[0]}->{IO_RETRY_COUNT}   = $cols[3];
            $ht{$cols[0]}->{IO_ABT_COUNT}     = $cols[4];
            $ht{$cols[0]}->{ThruPut_RateKBPS} = $cols[5];
            $ht{$cols[0]}->{Sns_ErrKey_1}     = $cols[6];
            $ht{$cols[0]}->{Sns_ErrKey_2}     = $cols[7];
            $ht{$cols[0]}->{Sns_ErrKey_3}     = $cols[8];
            $ht{$cols[0]}->{Sns_ErrKey_4}     = $cols[9];
        }    #foreach
        return \%ht;
    }    #if

    # -------------------------------------
    # parsing individual disk health status
    # which could be in one of the following
    # formats:
    # -------------------------------------
    if ($output =~ /ATTR_ID/ && $output =~ /STATUS/) {

        # skip the header HA.ID, HA SHELF BAY ...
        shift(@lines);
        shift(@lines);

        # parse HA.ID, HA, SHELF, BAY, CHAN, & Disk Serial Number
        my @cols = split(/\s+|\t/, $lines[0]);
        my $disk = $cols[0];
        $ht{$disk}->{HA_ID}            = $disk;
        $ht{$disk}->{HA}               = $cols[1];
        $ht{$disk}->{SHELF}            = $cols[2];
        $ht{$disk}->{BAY}              = $cols[3];
        $ht{$disk}->{CHAN}             = $cols[4];
        $ht{$disk}->{DiskSerialNumber} = $cols[5];
        shift(@lines);

        #parse Data Struct. Rev. Number:
        @cols = split(/\s+|\t/, $lines[0]);
        $ht{$disk}->{DataStructRevNumber} = $cols[4];

        # remove the header
        shift(@lines);
        shift(@lines);
        shift(@lines);

        #parse ATTR_ID STATUS VALUE WRST_VAL and RAW_VAL
        my $cnt = 1;
        foreach (@lines) {
            @cols = split(/\s+|\t/);
            _debug("Parsing $_");
            $ht{$disk}->{ATTR}->{$cnt}->{ATTR_ID}  = $cols[0];
            $ht{$disk}->{ATTR}->{$cnt}->{STATUS}   = $cols[1];
            $ht{$disk}->{ATTR}->{$cnt}->{VALUE}    = $cols[2];
            $ht{$disk}->{ATTR}->{$cnt}->{WRST_VAL} = $cols[3];
            $ht{$disk}->{ATTR}->{$cnt}->{RAW_VAL}  = $cols[4];
            $cnt++;
        }    #foreach

        return \%ht;
    } ## end if ($output =~ /ATTR_ID/...

    # ------------------------------
    # parse status of an active disk
    # ------------------------------
    if ($output =~ /IO_Avg_msec.*IO_Long_msec/) {

        # skip the header line HA.ID HA SHELF BAY ...
        shift(@lines);
        shift(@lines);

        # parse HA.ID, HA, SHELF, BAY, CHAN, & Disk Serial Number
        my @cols = split(/\s+|\t/, $lines[0]);
        my $disk = $cols[0];
        $ht{$disk}->{HA_ID}            = $cols[0];
        $ht{$disk}->{HA}               = $cols[1];
        $ht{$disk}->{SHELF}            = $cols[2];
        $ht{$disk}->{BAY}              = $cols[3];
        $ht{$disk}->{CHAN}             = $cols[4];
        $ht{$disk}->{DiskSerialNumber} = $cols[5];
        shift(@lines);

        # ------------------------------------------------
        # Save the column headers of IO_Avg_msec, IO_Long_msec, ...
        # ------------------------------------------------
        my @headers = split(/\s+|\t/, $lines[0]);
        shift(@lines);
        shift(@lines);

        @cols = split(/\s+|\t/, $lines[0]);
        $ht{$disk}->{$headers[0]} = $cols[1];
        $ht{$disk}->{$headers[1]} = $cols[2];
        $ht{$disk}->{$headers[2]} = $cols[3];
        $ht{$disk}->{$headers[3]} = $cols[4];
        $ht{$disk}->{$headers[4]} = $cols[5];
        shift(@lines);
        shift(@lines);

        # -------------------------------------------------------
        # Save the column headers of  Thruput_KBPS, Err_SKey1, ...
        # -------------------------------------------------------
        @headers = split(/\s+|\t/, $lines[0]);
        shift(@lines);
        shift(@lines);

        @cols = split(/\s+|\t/, $lines[0]);
        $ht{$disk}->{$headers[0]} = $cols[1];
        $ht{$disk}->{$headers[1]} = $cols[2];
        $ht{$disk}->{$headers[2]} = $cols[3];
        $ht{$disk}->{$headers[3]} = $cols[4];
        $ht{$disk}->{$headers[4]} = $cols[5];
        shift(@lines);
        shift(@lines);

        # ------------------------------------------------
        # save the Disk LOG header since this information
        # could be varied.
        # ------------------------------------------------
        my $disk_log = $lines[0];
        $ht{$disk}->{Disk_Log_Header} = $disk_log;
        shift(@lines);

        # --------------------------------------------------
        # Save the column headers since they could be varied
        # --------------------------------------------------
        $lines[0] =~ s/([A-z]) ([A-z])/$1_$2/g;
        @headers = split(/\s+|\t/, $lines[0]);
        shift(@lines);

        # skip the blank line
        shift(@lines);

        # parse the disk log stats
        foreach (@lines) {
            @cols = split(/\s+|\t/, $_);
            $ht{$disk}->{$cols[0]}->{$headers[1]} = $cols[1];
            $ht{$disk}->{$cols[0]}->{$headers[2]} = $cols[2];
            $ht{$disk}->{$cols[0]}->{$headers[3]} = $cols[3];
            $ht{$disk}->{$cols[0]}->{$headers[4]} = $cols[4];
            $ht{$disk}->{$cols[0]}->{$headers[5]} = $cols[5];
        }
        return \%ht;
    } ## end if ($output =~ /IO_Avg_msec.*IO_Long_msec/)
    croak("Unable to determine the disk shm_format: $output");
}    #disk_shm_stats

=over

=item Arguments

=over

=item host

(Required) : Hostspec to filer

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item timeout

(Optional) : connection timeout. Default is 300.

=back

=item Returns

Returns reference to the hash containing the parsed output upon success else returns 0.
On any error the routine returns the error message via $@.

=back

=cut

sub disk_shm_stats_asup {
    my (%pars) =
      validate(@_,
               {host     => {isa  => 'Hostrec'},
                timeout  => {type => SCALAR, default => 300, optional => 1},
                conntype => {type => SCALAR, default => "#rsh", optional => 1},
                ,});

    my $output = filer_cmd(host     => $pars{host},
                           conntype => $pars{conntype},
                           timeout  => $pars{timeout},
                           cmd      => "priv set -q diag; disk shm_stats asup");

    # check for errors
    if (   $output =~ /Disk State not READY/
        || $output =~ /disk shm_stat:/) {
        croak("Retrieving shm_stats of disk <$pars{disk}> failed due to: <$output>");
    }
    my @lines = split(/\r*\n/, $output);

    my (@resultant_array, %resultant_hash);

    foreach my $line (@lines) {
        if ($line =~ /(\w+(\.\w+)+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/) {
            my %hash;
            push(@{$resultant_hash{general_details}}, \%hash);
            $hash{disk}             = $1;
            $hash{disk_state}       = $3;
            $hash{average_io}       = $4;
            $hash{max_io}           = $5;
            $hash{retry_count}      = $6;
            $hash{timeout_count}    = $7;
            $hash{sense_data_1}     = $8;
            $hash{sense_data_2}     = $9;
            $hash{sense_data_3}     = $10;
            $hash{sense_data_4}     = $11;
            $hash{sense_data_5}     = $12;
            $hash{sense_data_9}     = $13;
            $hash{sense_data_B}     = $14;
            $hash{sense_data_other} = $15;
        } elsif ($line =~ /(\d+)\.\s+(\<\s+\d+\s+ms)\:\s+CPIO\s+reads\:\s+(\d+)\s+Reads\:\s+(\d+)\s+Writes\:\s+(\d+)\s*$/) {
            my %hash;
            push(@{$resultant_hash{system_io_completion_times}}, \%hash);
            $hash{index}      = $1;
            $hash{time_index} = $2;
            $hash{cpio_reads} = $3;
            $hash{reads}      = $4;
            $hash{writes}     = $5;
            $hash{time_index} =~ s/\s+/ /g;
        } elsif($line =~
            /(\d+)\.\s+(\<\s+\d+\s+ms)\:\s+CPIO\s+reads\:\s+(\d+)\s+Reads\:\s+(\d+)\s+Writes\:\s+(\d+)\s+Verifies\:\s+(\d+)\s*$/
          ) {
            my %hash;
            push(@{$resultant_hash{system_io_completion_times}}, \%hash);
            $hash{index}      = $1;
            $hash{time_index} = $2;
            $hash{cpio_reads} = $3;
            $hash{reads}      = $4;
            $hash{writes}     = $5;
            $hash{verifies}   = $6;
            $hash{time_index} =~ s/\s+/ /g;
        } elsif ($line =~ /(\d+)\.\s+(\w+(\.\w+)+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$/) {
            my %hash;
            push(@{$resultant_hash{io_completion_per_disk}}, \%hash);
            $hash{index}      = $1;
            $hash{disk}       = $2;
            $hash{cpio_reads} = $4;
            $hash{reads}      = $5;
            $hash{writes}     = $6;
            $hash{maxq}       = $7;
            $hash{ios_queued} = $8;
        } elsif ($line =~ /(\d+)\.\s+(\w+(\.\w+)+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$/) {
            my %hash;
            push(@{$resultant_hash{io_completion_per_disk}}, \%hash);
            $hash{index}      = $1;
            $hash{disk}       = $2;
            $hash{cpio_reads} = $4;
            $hash{reads}      = $5;
            $hash{writes}     = $6;
            $hash{verifies}   = $7;
            $hash{maxq}       = $8;
            $hash{ios_queued} = $9;
    } elsif ($line =~ /(\d+)\.\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$/) {
            my %hash;
            push(@{$resultant_hash{io_completion_time_table}}, \%hash);
            $hash{disk_index}  = $1;
            $hash{"4ms"}       = $2;
            $hash{"8ms"}       = $3;
            $hash{"16ms"}      = $4;
            $hash{"30ms"}      = $5;
            $hash{"50ms"}      = $6;
            $hash{"100ms"}     = $7;
            $hash{"200ms"}     = $8;
            $hash{"400ms"}     = $9;
            $hash{"800ms"}     = $10;
            $hash{"2000ms"}    = $11;
            $hash{"4000ms"}    = $12;
            $hash{"16000ms"}   = $13;
        $hash{"30000ms"}   = $14;
        $hash{"45000ms"}   = $15;
        $hash{"60000ms"}   = $16;
            $hash{"1000000ms"} = $17;
        } elsif ($line =~ /(\d+)\.\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$/) {
            my %hash;
            push(@{$resultant_hash{io_completion_time_table_2}}, \%hash);
            $hash{index} = $1;
            push(@{$hash{interval_1}}, {average_io_latency => $2,  utilization_pct => $3});
            push(@{$hash{interval_2}}, {average_io_latency => $4,  utilization_pct => $5});
            push(@{$hash{interval_3}}, {average_io_latency => $6,  utilization_pct => $7});
            push(@{$hash{interval_4}}, {average_io_latency => $8,  utilization_pct => $9});
            push(@{$hash{interval_5}}, {average_io_latency => $10, utilization_pct => $11});
            push(@{$hash{interval_6}}, {average_io_latency => $12, utilization_pct => $13});
        } elsif ($line =~ /Disk\s+(\w+(\.\w+)+)\s+grown\s+defect\s+list\s+has\s+(\d+)\s+entries$/) {
            my %hash;
            push(@{$resultant_hash{disk_defect_list}}, \%hash);
            $hash{disk}          = $1;
            $hash{no_of_entries} = $3;
        } elsif ($line =~ /^\s*(\d+)\:\s+\(\s+(\d+)\,\s+(\d+)\,\s+(\d+)\)\s*$/) {
            push(@{$resultant_hash{disk_defect_list}->[$#{$resultant_hash{disk_defect_list}}]->{defect_details}}, {defect_index => $1, cyl => $2, head => $3, sector => $4});
        } elsif ($line =~ /Error\s+\d+\s+type\s+\d+\:\s+command/) {
            my %hash;
            push(@{$resultant_hash{error_data_for_all_disks}}, \%hash);
            $resultant_hash{error_data_for_all_disks}->[$#{$resultant_hash{error_data_for_all_disks}}]->{data} = $line;
        } elsif ($line =~ /Sense\s+data\:/) {
            $resultant_hash{error_data_for_all_disks}->[$#{$resultant_hash{error_data_for_all_disks}}]->{sense_data} = $line;
        } elsif ($line =~ /^\s*(\d+x\w+\s*)+$/) {
            $resultant_hash{error_data_for_all_disks}->[$#{$resultant_hash{error_data_for_all_disks}}]->{sense_data} .= "$line";
        }

    } ## end foreach my $line (@lines)
    push(@resultant_array, \%resultant_hash);

    print Dumper(\@resultant_array);
    return \@resultant_array;
} ## end sub disk_shm_stats_asup

=head2 disk_show

    my $ht = disk_show(host=>$host, options=>'-n', ...);

The routine supports the 'disk show' CLI which displays information about the ownership all disks. This routine is only available on system with software-based disk ownership or SANOWN. The routine returns a reference of a hash table which has the common format of:

=begin html
$VAR1 = {
          '0c.01.6' => {
                         'HOME' => 'stor-u62',
                         'SYSID' => '175287864',
                         'UpCnt' => '4',
                         'DUID' => '5000C500:01823B67',
                         'POOL' => 'Pool0',
                         'PRKEY' => '175287864',
                         'SERIAL_NUMBER' => '3LN08T7Q00009720EMX6',
                         'HOME_SYSID' => '175287864',
                         'OWNER' => 'stor-u62'
                       },
          '0c.01.14' => {
                          'HOME' => 'stor-u63',
                          'SYSID' => '162218031',
                          'UpCnt' => '5',
                          'DUID' => '5000C500:0182414B',
                          'POOL' => 'Pool0',
                          'PRKEY' => '162218031',
                          'SERIAL_NUMBER' => '3LN093FN00009721JG0B',
                          'HOME_SYSID' => '162218031',
                          'OWNER' => 'stor-u63'
..........



=end html

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : Hostrec of filer

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item owername

(Optional) : Specifies the filername (owner of the disks). If ownername is specified then sysid and options cannot be specified.

=item sysid

(Optional) : Specifies a filer system id. If sysid is specified then both ownername and options cannot be specified.

=item options

(Optional) : There are 3 options: -n (unassigned disks), -a (assigned disks), and -v for all disks. Default is -v. If this is set then both ownername and sysid cannot be specified.

Notes - There is a hidden -x option and it is not be instrumented.

=item nooptions
(Optional) : This option need to be set when disk show need to be executed without any options.

=item timeout

(Optional) : connection timeout. Default is 300.

=back

=item Returns

If there is any error, returns an error message via $@.
If there is no record, the routine returns an empty hash.
Returns a reference of hash table of disks ownership. The
hash table has the following format:

=back

=cut

sub disk_show {
    _debug("Entering disk_show: ...");
    my (%pars) =
      validate(@_,
               {host    => {isa  => 'Hostrec'},
                timeout => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                conntype => {type     => SCALAR,
                             optional => YES,
                             default  => '#rsh'},
                filermode => {type     => SCALAR,
                              optional => YES,
                              regex    => qr/maint|normal/,
                              default  => NORMAL_MODE},
                ownername => {type => SCALAR, optional => YES},
                sysid     => {type => SCALAR, optional => YES},
                options   => {type => SCALAR, optional => YES, regex => qr/-[avnz]/, default => "-v -z"},
                nooptions => {type => SCALAR, optional => YES},});

    my $sysid     = defined($pars{sysid})     ? " -s " . $pars{sysid}     : "";
    my $ownername = defined($pars{ownername}) ? " -o " . $pars{ownername} : "";
    if ($sysid ne "" || $ownername ne "") {
        $pars{options} = "";
    }
    my $cmd;

    # Check if disk show needs to be executed without any options

    if (defined($pars{nooptions})) {
        $pars{options} = "";
    }
    $cmd = "disk show $ownername $sysid $pars{options}";
    my $host = $pars{host};
    my $mode = '7Mode';
    $mode = 'Nodescope' if ($host->{hosttype} =~ /filer\-ng/);
    my $api_set_obj = NACL::APISet->new(
            hostobj   => $host,
            category  => 'Node',
            interface => 'CLI',
            set       => $mode,
        );
    my $output = $api_set_obj->execute_command(
        command => $cmd,
        "connectrec-timeout" => $pars{timeout}
    );
    # --------------------------------------------
    # The empty list is considered as not an error.
    # ---------------------------------------------
    my %ht = ();
    if ($output =~ /disk show: No disks match option -n/) {
        tstlog(-comment, "$output");
        return \%ht;
    }
    if (   $output =~ /disk show:/
        || $output !~ /DISK/
        || $output eq ""
        || $output =~ /disk:/) {
        croak("Disk show command failed due to: $output");
    }

    my @lines = split(/\r*\n/, $output);

    # --------------------------------------------------------
    # removes the header lines
    # --------------------------------------------------------
    if ($pars{filermode} eq MAINT_MODE) {
        shift(@lines);
        my @arr=split(/:\s+/,$lines[0]);
        #$ht{Local_System_ID}=$arr[1];
        shift(@lines);
        shift(@lines);
        shift(@lines);
        shift(@lines);
    } else {
        shift(@lines);
        shift(@lines);
    }
    # -------------------------------------------------------
    # parsing each line and adding the result to a hash table
    # -------------------------------------------------------
    foreach (@lines) {
        my $line=$_;
        # Replaces open/close parentheses with spaces
        s/\(|\)/ /g;
        my (@cols) = split(/\s+/);
        if ($_=~/^\S+\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+(\S+)$/) {#0b.00.0      stor-u133 (1574774957)    Pool0  PNG4MVKA              stor-u133 (1574774957)    Block     PRkey=1574774957   UpCnt=  241 duid=5000CCA0:12087018
            my @prkey=split(/=/,$8);
            my @duid=split(/=/,$10);
            $ht{$cols[0]}->{OWNER}=$1;
            $ht{$cols[0]}->{SYSID}=$2;
            $ht{$cols[0]}->{POOL}=$3;
            $ht{$cols[0]}->{SERIAL_NUMBER}=$4;
            $ht{$cols[0]}->{HOME}=$5;
            $ht{$cols[0]}->{HOME_SYSID}=$6;
            $ht{$cols[0]}->{CHKSUM}=$7;
            $ht{$cols[0]}->{UpCnt}=$9;
            $ht{$cols[0]}->{PRKEY}=$prkey[1];
            $ht{$cols[0]}->{DUID}=$duid[1];
        } elsif ($_=~/^\S+\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+\S+\s+(\S+)\s+(\S+)$/) {#1a.32        stor-u133 (1574774957)    Pool0  3LN7CY0M00009925EMAZ  stor-u133 (1574774957)    Block     PRkey=        0   UpCnt=  244 duid=2000001D:381076
            my @duid=split(/=/,$10);
            $ht{$cols[0]}->{OWNER}=$1;
            $ht{$cols[0]}->{SYSID}=$2;
            $ht{$cols[0]}->{POOL}=$3;
            $ht{$cols[0]}->{SERIAL_NUMBER}=$4;
            $ht{$cols[0]}->{HOME}=$5;
            $ht{$cols[0]}->{HOME_SYSID}=$6;
            $ht{$cols[0]}->{CHKSUM}=$7;
            $ht{$cols[0]}->{PRKEY}=$8;
            $ht{$cols[0]}->{UpCnt}=$9;
            $ht{$cols[0]}->{DUID}=$duid[1];
        } elsif ($_=~/^\S+\s+(\S+)\s+(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+\S+\s+(\S+)\s+(\S+)$/) {#1a.00.0                                 FAILED PPG5L0EA             Block       PRkey=        0   UpCnt=    0 duid=5000CCA0:120A25E8
            my @duid=split(/=/,$6);
            $ht{$cols[0]}->{OWNER}="";
            $ht{$cols[0]}->{SYSID}="";
            $ht{$cols[0]}->{POOL}=$1;
            $ht{$cols[0]}->{SERIAL_NUMBER}=$2;
            $ht{$cols[0]}->{HOME}="";
            $ht{$cols[0]}->{HOME_SYSID}="";
            $ht{$cols[0]}->{CHKSUM}=$3;
            $ht{$cols[0]}->{PRKEY}=$4;
            $ht{$cols[0]}->{UpCnt}=$5;
            $ht{$cols[0]}->{DUID}=$duid[1];
        } elsif (($_=~/^\S+\s+(\S+\s+\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+(\S+)$/)) {#0b.45        Not Owned                  NONE   3QD0P7HK             Zoned       PRkey=        0   UpCnt=    0 duid=D7509942:44757B7C:E885AED2:0D6A2AE9
            if ($1 eq 'Not Owned') {
                my @duid=split(/=/,$8);
                $ht{$cols[0]}->{OWNER}=$1;
                $ht{$cols[0]}->{SYSID}="";
                $ht{$cols[0]}->{POOL}=$2;
                $ht{$cols[0]}->{SERIAL_NUMBER}=$3;
                $ht{$cols[0]}->{HOME}="";
                $ht{$cols[0]}->{HOME_SYSID}="";
                $ht{$cols[0]}->{CHKSUM}=$4;
                $ht{$cols[0]}->{PRKEY}=$6;
                $ht{$cols[0]}->{UpCnt}=$7;
                $ht{$cols[0]}->{DUID}=$duid[1];
            } else {#0b.36        svtl-2040-2(135095648)    Pool0  3QD0NZBS              svtl-2040-2(135095648)      PRkey=135095648   UpCnt=    2 duid=B72A47A7:6EB7A9AC:2FD1E5BB:2C2B75A9
                my @owner=split(/\s+/,$1);
                my @duid=split(/=/,$8);
                my @prkey=split(/=/,$6);
                $ht{$cols[0]}->{OWNER}=$owner[0];
                $ht{$cols[0]}->{SYSID}=$owner[1];
                $ht{$cols[0]}->{POOL}=$2;
                $ht{$cols[0]}->{SERIAL_NUMBER}=$3;
                $ht{$cols[0]}->{HOME}=$4;
                $ht{$cols[0]}->{HOME_SYSID}=$5;
                $ht{$cols[0]}->{PRKEY}=$prkey[1];
                $ht{$cols[0]}->{UpCnt}=$7;
                $ht{$cols[0]}->{DUID}=$duid[1];
            }
        } elsif (($line=~/\)$/)&&($_=~/^\S+\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/)) {#0b.37        svtl-2040-2(135095648)    Pool0  3QD0NZCP              svtl-2040-2(135095648)
            $ht{$cols[0]}->{OWNER}=$1;
            $ht{$cols[0]}->{SYSID}=$2;
            $ht{$cols[0]}->{POOL}=$3;
            $ht{$cols[0]}->{SERIAL_NUMBER}=$4;
            $ht{$cols[0]}->{HOME}=$5;
            $ht{$cols[0]}->{HOME_SYSID}=$6;
        } elsif ($_=~/^\S+\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)$/) {#0b.37        svtl-2040-2(135095648)    Pool0  3QD0NZCP              svtl-2040-2(135095648)
            $ht{$cols[0]}->{OWNER}=$1;
            $ht{$cols[0]}->{SYSID}=$2;
            $ht{$cols[0]}->{POOL}=$3;
            $ht{$cols[0]}->{SERIAL_NUMBER}=$4;
            $ht{$cols[0]}->{HOME}=$5;
            $ht{$cols[0]}->{HOME_SYSID}=$6;
            $ht{$cols[0]}->{CHKSUM}=$7;
        } elsif ($_=~/^\S+\s+(\S+\s+\S+)\s+(\S+)\s+(\S+)$/) {#0b.100        Not Owned                  NONE   3QD0P7HK
            $ht{$cols[0]}->{OWNER}=$1;
            $ht{$cols[0]}->{SYSID}="";
            $ht{$cols[0]}->{POOL}=$2;
            $ht{$cols[0]}->{SERIAL_NUMBER}=$3;
            $ht{$cols[0]}->{HOME}="";
            $ht{$cols[0]}->{HOME_SYSID}="";
        }elsif($_=~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/){  #0a.01.5      stor-2240-5 1786428317     Pool0  PPG3L87A              stor-2240-5 1786428317
            $ht{$cols[0]}->{OWNER}=$2;
            $ht{$cols[0]}->{SYSID}=$3;
            $ht{$cols[0]}->{POOL}=$4;
            $ht{$cols[0]}->{SERIAL_NUMBER}=$5;
            $ht{$cols[0]}->{HOME}="";
            $ht{$cols[0]}->{HOME_SYSID}="";
        }elsif($_=~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/){  #0a.01.5      stor-2240-5 1786428317     Pool0  PPG3L87A              stor-2240-5 1786428317
            $ht{$cols[0]}->{OWNER}=$2;
            $ht{$cols[0]}->{SYSID}=$3;
            $ht{$cols[0]}->{POOL}=$4;
            $ht{$cols[0]}->{SERIAL_NUMBER}=$5;
            $ht{$cols[0]}->{HOME}="";
            $ht{$cols[0]}->{HOME_SYSID}="";
        }else {
            logcomment("WARNING: Unmatching field found in the disk show output:  @cols ");
        }
    }    # Parse line by line

    #print "\n ht:".Dumper %ht;
    return \%ht;

}    #disk_show

=head2 disk_simpull

    disk_simpull(host=>$host, disk=>$disks);

The routine performs pulling simulator disk. One or more disks could be specified.

=over

=item Arguments

=over

=item host

(Required) : Hostrec of filer

=item disk

(Required) : Specifies one of multiple disk names to be pulled.

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item timeout

(Optional) : connection timeout. Default is 300.

=back

=item Returns

Returns a message if fails via $@.

=item Notes

The routine is for simulator use only.

=back

=cut

sub disk_simpull {
    _debug("Entering disk_simpull: ...");
    my (%pars) =
      validate(@_,
               {host    => {isa  => 'Hostrec'},
                timeout => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => '#rsh'},
                disk => {type => SCALAR, optional => YES},});

    my ($conn);
    $conn = make_conn($pars{host}, $pars{conntype});

    my (undef, undef, $output) = say($conn, "disk simpull $pars{disk}", '-timeout', $pars{timeout});

    # -----------------------
    # Should catch all errors
    # -----------------------
    if (   $output =~ /usage: /
        || $output =~ /disk simpull: Failed/
        || $output =~ /disk simpull: Disk \S+ does not exist/) {
        croak($output);
    }
    return PASS;
}    #disk_simpull

=head2 disk_simpush

    disk_simpush(host=>$host, diskserialnum=>$serialnum);

The routine performs disk pushing. One or more disk serial numbers could be specified.

=over

=item Arguments

=over

=item host

(Required) : Hostrec of filer

=item diskserialnum

(Required) : Specifies one of multiple disk serial numbers.

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".

=item timeout

(Optional) : connection timeout. Default is 300.

=back

=item Returns

Returns a message if fails via $@.

=item Notes

The routine is for simulator use only.

=back

=cut

sub disk_simpush {
    _debug("Entering disk_simpush: ...");
    my (%pars) =
      validate(@_,
               {host    => {isa  => 'Hostrec'},
                timeout => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => '#rsh'},
                diskserialnum => {type => SCALAR, optional => YES},
                simdir        => {type => SCALAR, optional => YES, default => ""},});

    my ($conn);
    $conn = make_conn($pars{host}, $pars{conntype});

    my $disks = _sim_get_disk_fullname($pars{diskserialnum}, $pars{simdir});
    if ($disks eq "") {
        croak("Unable to retrieve the disk fullname of disk serial num: <$pars{diskserialnum}>");
    }

    my (undef, undef, $output) = say($conn, "disk simpush $disks", '-timeout', $pars{timeout});

    # -----------------------
    # Should catch all errors
    # -----------------------
    if (   $output =~ /usage: disk/
        || $output =~ /disk simpush: Failed to add disk/) {
        croak($output);
    }
    return PASS;
}    #disk_simpush

=head2 disk_scrub

    disk_scrub(host=>$host);

The routine starts or stops scrub on all raidgroups. The routine shall wait until scrub is completed unless the scrub_nowait option is specified.

The routine checks for any PI, checksum, and media errors only without the option 'scrub_nowait'. The routine returns FATAL if there is any PI, checksum, or media error.

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : Hostspec of the filer.

=item disk

(Required) : Disk name.

=item start

(Optional) : to start scrub

=item stop

(Optional) : to stop scrub

=item conntype

(Optional) : connection type: '#rsh', '#console', or '#telnet'. Default is '#rsh".

=item timeout

(Optional) : connection timeout. Default is 300.

=item waittime

(Optional) : -1 wait until scrub is done. Default is (-1).
Set waittime=>'0' not to wait.

=back

=item Returns

Returns 1 on success of either start or stop scrub

=item Example:

# start scrub and wait until scrub is completed.
disk_scrub(host=>$host, start=>1);

# start scrub but not wait
disk_scrub(host=>$host, start=>1, waittime=>0);

=back

=cut

sub disk_scrub {
    _debug("Entering disk_scrub: ...");
    my (%pars) =
      validate(@_,
               {host     => {isa  => 'Hostrec', optional => NO},
                start    => {type => SCALAR,    optional => YES},
                stop     => {type => SCALAR,    optional => YES},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => "#rsh"},
                timeout  => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                waittime => {type => SCALAR, optional => YES, regex => qr/\d+/, default => WAIT_UNTIL_DONE},
                api_set_obj => {type => OBJECT, optional => YES},
         },
    );

    my $disk     = $pars{disk};
    my $conn     = make_conn($pars{host}, $pars{conntype});
    my $waittime = $pars{waittime};
    my $apiset_obj = $pars{api_set_obj};
    if(!$apiset_obj){
       my $node_obj = NACL::C::Node->new(hostrec => $pars{host});
       $apiset_obj = $node_obj->get_7m_or_nodescope_apiset();
    }
   $apiset_obj->set_timeout('connectrec-timeout' => $pars{timeout});
    if (defined($pars{start})) {
#        my (undef, undef, $output) = say($conn, "disk scrub start", '-timeout', $pars{timeout});
        my $output = $apiset_obj->execute_command(command => 'disk scrub start');
        # --------------------------------------------
        # starts scrub, verify it is started then exit
        # --------------------------------------------
        if ($waittime == 0) {
            if (_disk_scrub_is_started($conn)) {
                _debug("scrub is started without any waiting");
                return PASS;
            }

            logresult('fail', 'Unable to start scrub without any wait');
            return FAIL;
        }    # without wait

        # ----------------------------------------------
        # verify that disk scrub is going on all vols/aggr
        # and wait until scrub is done
        # ----------------------------------------------
        if (!_disk_scrub_wait($conn, $waittime)) {
            logresult('fatal', $@);
        }
        return PASS;
    }    # scrub start

    # -----------
    # stop scrub
    # -----------
    _debug("disk_scrub: stopping scrub ...");
    if (defined($pars{stop})) {
        my (undef, undef, $output) = say($conn, "disk scrub stop", '-timeout', $pars{timeout});
        if (!_disk_scrub_wait($conn)) {
            logresult('fatal', $@);
        }
        return PASS;
    }    # scrub stop

}    #disk_scrub

=head2 disk_suspend

    disk_suspend(host=>$host, disk=>$disk);

The routine initiates the request of suspending I/O to the specified disk using
the 'disk suspend' CLI.

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : Hostspec to filer

=item disk

(Required) : Specify the disk to be suspended.

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".


=item timeout

(Optional) : connection timeout. Default is 300.

=back

=item Returns

Returns 1 upon success else returns 0.
On any error the routine returns the error message via $@.

=back

=cut

sub disk_suspend {
    _debug("Entering disk_suspend: ...");
    my (%pars) =
      validate(@_,
               {host    => {isa  => 'Hostrec'},
                disk    => {type => SCALAR, optional => NO},
                timeout => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => '#rsh'},});

    my $conn = make_conn($pars{host}, $pars{conntype});
    my (undef, undef, $output) = say($conn, "priv set -q diag; disk suspend $pars{disk}", '-timeout', $pars{timeout});

    # Check for errors
    if (   $output =~ /disk:/
        || $output =~ /disk suspend\/resume:/) {
        croak("Disk suspend <$pars{disk}> failed due to: <$output>");
    }

    # ----------------------------------
    # verify the disk state is not ready
    # ----------------------------------
    eval { disk_shm_stats(host => $pars{host}, disk => $pars{disk}); };
    if ($@ =~ /Disk State not READY/) {
        return PASS;
    } else {
        croak("The state of <disk=>$pars{disk}> supposed to be NOT READY but it was not");
    }
}    #disk_suspend

=head2 disk_unfail

    disk_unfail(host=>$host, disk=>disk, ...);

The routine unfails the specified disk. The routine tries
to unfail the disk with option "-s". If the option -s is
not supported in the current kernel, will try again without -s.
It also handles the maint mode.

The routine takes the following parameters.

=over

=item Arguments

=over

=item host

(Required) : Hostspec of the filer.

=item disk

(Required) : diskname.

=item conntype

(Optional) : 3 available connection types: #rsh, #console, or '#telnet'.

=item timeout

(Optional) : connection time out. Default is 300.

=item waittime

(Optional) : how long to wait for the disk to listed as spared. Default is (-1) wait until completed. Specify the n seconds to wait for.

=item options

(Optional) : -s to be spared. The -s option is only available on certain
release kernels.

=item filermode

(Optional) : filer mode could be either maint or up.

=back

=item Returns

Returns 1 on success or 0 on failure due to multiple errors:

=item Example:

# wait until the disk is listed on the spare pool.
 disk_unfail(host=>$host, disk=>$disk);

# no wait
disk_unfail(host=>$host, disk=>$disk, waittime=>0);

=item See Also

disk_fail

=back

=cut

sub disk_unfail {
    _debug("Entering disk_unfail: ...");
    my (%pars) =
      validate(@_,
               {host     => {isa  => 'Hostrec', optional => NO},
                disk     => {type => SCALAR,    optional => NO},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => "#rsh"},
                options   => {type => SCALAR, optional => YES, regex => qr/-s/,           default => "-s"},
                timeout   => {type => SCALAR, optional => YES, regex => qr/\d+/,          default => SAY_TIMEOUT},
                waittime  => {type => SCALAR, optional => YES, regex => qr/\d+/,          default => WAIT_UNTIL_DONE},
                filermode => {type => SCALAR, optional => YES, regex => qr/maint|normal/, default => "normal"},
                api_set_obj => {type => OBJECT, optional => YES},
               });

    my $conn = make_conn($pars{host}, $pars{conntype});
    my $options = defined($pars{options}) ? $pars{options} : "-s";    #hidden option
    my $output;
    my $apiset_obj = $pars{api_set_obj};
    if(!$apiset_obj){
        my $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
    }
    $apiset_obj->set_timeout('connectrec-timeout' => $pars{timeout});
    # -----------------------------------------------
    # Try with option -s first and if it failed then
    # try without the option -s. This seems to be
    # a reasonable approach to determine whether or
    # not the current kernel supports the -s option.
    # -----------------------------------------------
    if ($pars{filermode} eq "normal") {
#        (undef, undef, $output) = say($conn, "priv set -q diag; disk unfail $options $pars{disk}", '-timeout', $pars{timeout});
        $output = $apiset_obj->execute_command(command => "priv set -q diag; disk unfail $options $pars{disk}",);
        if ($output =~ /disk unfail: Has extra arguments/) {
            _debug("disk unfail <-s> is not supported in this kernel");
#            (undef, undef, $output) = say($conn, "priv set -q diag; disk unfail $pars{disk}", '-timeout', $pars{timeout});
            $output = $apiset_obj->execute_command(command => "priv set -q diag; disk unfail $pars{disk}",);
        }
    } else {    #proceed with maint mode
        $conn = make_conn($pars{host}, "#console");
        (undef, undef, $output) = say($conn, "disk unfail $pars{disk}", -prompt, MAINT_PROMPT, '-timeout', $pars{timeout});
    }

    # ---------------------------
    # check for the output result
    # ---------------------------
    if (   $output =~ /disk: No argument supplied for unfail option/
        || $output =~ /disk unfail: Disk not found/
        || $output =~ /disk unfail: Disk is not currently failed/
        || $output =~ /disk unfail: Has extra arguments/
        || $output =~ /disk unfail: Can\'t unfail a/
        || $output =~ /Drive could not be unfailed/
        || $output =~ /No failed drives were found/
        || $output =~ /not found/) {
        croak("Disk unfail command failed due to: $output");
    } ## end if ($output =~ /disk: No argument supplied for unfail option/...

    # -----------------------------
    # wait until the disk is spared unless the filermode is maint
    # -----------------------------
    if ($output =~ /disk unfail:/ && $pars{filermode} eq "normal") {

        # ----------------------------------------------------
        # if the waittime is set will wait until the disk is
        # listed on the spare list.
        # ----------------------------------------------------
        if (defined($pars{waittime})) {
            my $waittime = $pars{waittime};
            my $status = _disk_wait_until($conn, $pars{disk}, $waittime, "spare");
            if ($status) {
                tstlog(-comment, "Disk <$pars{disk}> was listed on the spare list");
                return PASS;
            }
            tstlog(-comment, "Disk <$pars{disk}> was not listed on the spare list after <$waittime> seconds");
            return FAIL;
        } ## end if (defined($pars{waittime...
        return PASS;
    }    # validate result
}    #disk_unfail

=head2 disk_zero_spares

    disk_zero_spares(host=>$host);

The routine initiates the zeroing disks and wait until zeroing completed.

The routine takes the following parameters.

=over

=item Argument

=over

=item host

(Required) : Hostspec of filer

=item conntype

(Optional) : Specify one of the connection types: '#rsh', '#console', or '#telnet'. Default is "#rsh".


=item timeout

(Optional) : connection timeout. Default is 300.

=item waittime

(Optional) : Specify time in secs to wait for spares being zeroed. Default is set to
wait_until_done.

=back

=item Returns

Returns 1 on zeroing done or 0 if not completed.

=back

=cut

sub disk_zero_spares {
    _debug("Entering disk_zero_spares: ...");
    my (%pars) =
      validate(@_,
               {host     => {isa  => 'Hostrec'},
                timeout  => {type => SCALAR, optional => YES, regex => qr/\d+/, default => SAY_TIMEOUT},
                waittime => {type => SCALAR, optional => YES, regex => qr/\d+/, default => WAIT_UNTIL_DONE},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => '#rsh'},});

    my $conn = make_conn($pars{host}, $pars{conntype});
    say($conn, "disk zero spares", '-timeout', $pars{timeout});

    my $cnt       = -1;
    my $sleeptime = 60;
    if (RaidUtils_is_sim($conn)) {
        $sleeptime = 10;
    }
    my $ro = new RaidInfo($conn);

    # -------------------------------------------------------
    # check until zeroing spares completed or ran out of time
    # -------------------------------------------------------
    while ($cnt <= $pars{waittime} || $pars{waittime} == WAIT_UNTIL_DONE) {
        my %spares     = $ro->spare_disks();
        my $is_zeroing = 0;
        foreach my $disk (keys %spares) {
            _debug("checking disk <$disk> and its status <$spares{$disk}->{fsm_state}>");
            if ($spares{$disk}->{fsm_state} ne "ZEROED") {
                $is_zeroing = 1;
                last;
            }
        }

        # -----------------------------------------
        # when zeroing is completed then it's done.
        # -----------------------------------------
        if (!$is_zeroing) {
            _debug("All disks are now zeroed");
            return PASS;
        }
        $cnt += $sleeptime;
        sleep($sleeptime);
        $ro->refresh();
    }    #wait until zeroing all spares

    tstlog(-comment, "Disk zeroing exceeded the specified waittime");
    return PASS;
}    #disk_zero_spares

=head2 scrub_status

    scrub_status($host);

The routine checks for the scrub status and returns 1 is running and returns 0 otherwise. The routine also checks for any PI, media errors, or checksum error and fatal out if there is any error.

The routine takes the following parameters.

=over

Example:

# check if scrub is going and wait until it is completed.
scrub_status($host);

# check if scrub is going and DO not wait
scrub_status($host, waittime=>0);

=item Arguments

=over

=item host

(Required) : Hostspec of the filer.

=item conntype

(Optional) : connection type: '#rsh', '#telnet', or '#console'.
Default is '#rsh'.

=item timeout

(Optional) : Connection timeout. Default is 300.

=item waittime

(Optional) : wait until the specified time if scrub has not been
completed yet. Default (-1) to wait until scrub is done.

=back

=item Returns

The routine returns 1 if the scrub is still going and 0 if not.

=back

=cut

sub scrub_status {
    _debug("Entering disk_unfail: ...");
    my (%pars) =
      validate(@_,
               {host     => {isa => 'Hostrec', optional => NO},
                conntype => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\#rsh|\#telnet|\#console/,
                             default  => "#rsh"},
                timeout => {type     => SCALAR,
                            optional => YES,
                            regex    => qr/\d+/,
                            default  => SAY_TIMEOUT},
                waittime => {type     => SCALAR,
                             optional => YES,
                             regex    => qr/\d+/,
                             default  => WAIT_UNTIL_DONE},});

    my $timeout;
    my $conn     = make_conn($pars{host}, $pars{conntype});
    my $ro       = new RaidInfo($conn);
    my $cnt      = -1;
    my $waittime = $pars{waittime};

    my $sleep_time = 60;
    if (RaidUtils_is_sim($conn)) {
        $sleep_time = 10;
    }

    my %aggrs;
    while ($cnt <= $waittime || $waittime == WAIT_UNTIL_DONE) {
        $ro->refresh();
        %aggrs = $ro->all_raid_groups();
        my $scrub_state = 0;

        while (my ($key, $rg) = each(%aggrs)) {
            _debug("raidgroup key=$key ; $rg->{scrub_progress}");
            if ($rg->{scrub_progress} ne "-1") {
                _debug("scrub is in progress <$rg->{scrub_progress}>");
                $scrub_state = 1;
            }

            # ----------------------------
            # check for any type of errors
            # ----------------------------
            if ($rg->{scrub_pi} ne "0") {
                croak("Found scrub PI error in raidgroup <$key>");
            }
            if ($rg->{scrub_media_err} ne "0") {
                croak("Found scrub media error in raidgroup <$key>");
            }
            if ($rg->{scrub_cksum_err} ne "0") {
                croak("Found scrub check sum error in raidgroup <$key>");
            }
        }    #each raidgroup

        # --------------------
        # finish when no scrub
        # --------------------
        if ($scrub_state == 0) {
            last;
        }

        # sleep a little while
        _debug("sleeping <$sleep_time> seconds waiting for scrub to be completed.");
        sleep($sleep_time);
        $cnt += $sleep_time;
    }    # while scrubbing

    while (my ($key, $rg) = each(%aggrs)) {
        _debug("raidgroup key=$key ; $rg->{scrub_progress}");
        if ($rg->{scrub_progress} ne "-1") {
            _debug("scrub_status: scrub is in progress <$rg->{scrub_progress}>");
            return PASS;
        }
    }    #each raidgroup

}    #scrub_status

# internal function and should not be exported
# helper function to execute commands
# and return result
#
sub execute_cmd {
    my %opts = validate(@_,
                        {host            => {type => OBJECT, isa      => 'Hostrec'},
                         conntype        => {type => SCALAR, default  => "#console"},
                         cmd             => {type => SCALAR},
                         options         => {type => SCALAR, optional => 1},
                         expected_string => {type => SCALAR, optional => 1},
                         prompt    => {type => SCALAR, optional => 1, default => "\\*>"},
                         timeout   => {type => SCALAR, optional => 1, default => 7200},
                         priv_mode => {type => SCALAR, optional => 1, default => ""},});

    my ($cmd, $subcmd);

    ($cmd, $subcmd) = (split / /, $opts{cmd});

##  die if( ! is_cmd_supported(host => $opts{host}, conntype => $opts{conntype},
##          cmd => $cmd, subcmd => $subcmd,
##          expected => "$cmd\\s+(\\w+\\s+)+",
##          prompt => $opts{prompt})
##  );

    my ($fc, @str);

    $fc = make_conn($opts{host}, $opts{conntype});

    push @str, "$opts{expected_string}" if (defined $opts{expected_string});
    push @str, "-timeout"               if (defined $opts{timeout});
    push @str, "$opts{timeout}"         if (defined $opts{timeout});
    push @str, "-prompt"                if (defined $opts{prompt} and $opts{prompt} ne "");
    push @str, "$opts{prompt}"          if (defined $opts{prompt} and $opts{prompt} ne "");

    my $output = "";
    my $prompt = "";

    $cmd = "";

    #------------------------------------------------------------------------------------
    # Replacing switch statement with if and else from as it is not working on win client
    # please find burt330766 for more details
    #-------------------------------------------------------------------------------------
    if    ("$opts{priv_mode}" eq "diag")     { $cmd = "priv set diag; "; }
    elsif ("$opts{priv_mode}" eq "advanced") { $cmd = "priv set advanced; "; }
    elsif ("$opts{priv_mode}" eq "test")     { $cmd = "priv set test; "; }
    else                                     { $cmd = ""; }
    $cmd .= " $opts{cmd}";
    logcomment("Executing Command=$cmd, other=[@str]");

    my $total_output = "";
    eval {
        if ($#str > -1) {
            (undef, undef, $output, $prompt) = say($fc, $cmd, @str);
        } else {
            (undef, undef, $output) = say($fc, $cmd);
        }
    };

    $total_output .= $output;
    if ($@) {
        (undef, undef, $output, $prompt) = say($fc, "\r\n");
    }

    $total_output .= $output;

    $total_output =~ s/$cmd\s+//g;
    $total_output =~ s/\s+$//g;

    logcomment("output=$total_output");

    return $total_output;
} ## end sub execute_cmd

=head2 disk_dump

disk_dump (     host    => $host,
        conntype=> $conntype,
        src => $disk,
        dst => $disk
    )

disk_dump (     host    => $host,
        conntype=> $conntype,
        src => $disk,
        dst => $disk,
        count   => $count
    )

disk_dump (     host    => $host,
        conntype=> $conntype,
        blk_start=> $disk,
        blk_end => $disk,
    )

Dump a contents of the disk (using dd) from source disk to destination disk.

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use.
    Default is "#rsh".
    Default for filermode "MAINT" is "#console"

=item src

(Optional) : Name of source disk
    'src' and 'dst' go together, should be specified when complete
    disk need to be dumped to the destination disk

=item dst

(Optional) : Name of destination disk
    'src' and 'dst' go together, should be specified when complete
    disk need to be dumped to the destination disk

=item count

(Optional) : Number of blocks to be copied to destination
    'count' goes hand in hand with 'src', 'dst' when only few blocks
    need to be copied to destination
    Default is set to 1 block (for testing purposes)

=item blk_start

(Optional) : Block number from where the disk dump should begin
    'blk_start' and 'blk_end' go together, should be specified when only
    a portion of the disk need to be copied to the destination disk

=item blk_end

(Optional) : Block number from where the disk dump should end
    'blk_start' and 'blk_end' go together, should be specified when only
    a portion of the disk need to be copied to the destination disk

=item timeout

(Optional) : time in secs, maximum time the command should wait for the response
        Default is set to 7200 secs

=back

=cut

#------------------------------------------------------------------------------

sub disk_dump {
    my %opts = validate(@_,
                        {host      => {type => OBJECT, isa      => 'Hostrec'},
                         conntype  => {type => SCALAR, default  => "#console"},
                         src       => {type => SCALAR, optional => 1},
                         dst       => {type => SCALAR, optional => 1},
                         count     => {type => SCALAR, optional => 1, default => 1},
                         blk_start => {type => SCALAR, optional => 1},
                         blk_end   => {type => SCALAR, optional => 1},
                         timeout   => {type => SCALAR, optional => 1, default => 7200},});

    # Local variables here
    my ($cmd, $output);

    # src/dst and count go together
    # blk_start/blk_end go together
    # other combinations are invalid
    #croak "Invalid set of parameters [@_], specify either ('src' and 'dst' or count) or ('blk_start' and 'blk_end')"
    #   if( ( defined $opts{src} or $opts{dst} or $count )  and
    #       ( defined $opts{blk_start} or $opts{blk_end} ) );

    # count doesnot go with blk_start/blk_end
    #croak "Invalid set of parameters [@_], specify 'count' along with 'src' and 'dst'"
    #   if( defined $opts{count} and
    #       ( defined $opts{blk_start} or defined $opts{blk_end} ) );

    # forumulate the command
    # when 'src' and 'dst' are specified
    $cmd .= "dd din=$opts{src} dout=$opts{dst}"
      if (defined $opts{src} and $opts{dst});

    # when count is specified in conjunction with src/dst pair
    # add count parameter to the command
    if (defined $opts{count}
        and (defined $opts{src} and defined $opts{dst})) {
        $cmd .= " count=$opts{count} ";

        # change timeout to lower values as number of blks to copy < max blks
        $opts{timeout} = ($opts{count} < 100) ? 30 : $opts{timeout};
    }

    # formulate the command
    # when block start/end are specified
    $cmd .= "dd bin=$opts{blk_start} bout=$opts{blk_end}"
      if (defined $opts{blk_start} and defined $opts{blk_end});
    print("------the command $cmd------\n");

    # handle all command arguments related cases here
    croak "Invalid/No arguments specified"
      if ($cmd eq "");

    # DD command does not return any message when the command is successful
    # making sure __OK__ is printed and prompt is obtained for 'say' not to timeout
    $cmd .= " ; echo _OK_ ";

    #$cmd .= "timeout => $opts{timeout}";
    $output = execute_cmd(host      => $opts{host},
                          conntype  => $opts{conntype},
                          cmd       => $cmd,
                          timeout   => $opts{timeout},
                          priv_mode => "diag");

    # Return the error message to the calling function/script
    # to validate with the know error/output message, especially required for snaplock to
    # verify the error/output message to match with the operation performed
    return $output;

}    # end of disk_dump

=head2 disk_unfail_during_reconstruction

disk_unfail_during_reconstruction (
        host        => $host,
        conntype    => "#console",
        failed_disk     => $failed_disk,
        aggr        => $aggr,
        aggr_disks  => $aggr_disks,
        make_spare  => FALSE,
        recons_range    => "0-10",
        timeout     => "7200",
        forced      => TRUE,
        offline     => FALSE,
        reboot      => FALSE,
        panic       => FALSE
    )

disk_unfail_during_reconstruction (
        host        => $host,
        failed_disk     => $failed_disk,
        aggr        => $aggr,
        offline     => TRUE,
    )

disk_unfail_during_reconstruction (
        host        => $host,
        failed_disk     => $failed_disk,
        aggr        => $aggr,
        panic       => TRUE,
    )

disk_unfail_during_reconstruction (
        host        => $host,
        failed_disk     => $failed_disk,
        aggr        => $aggr,
        reboot      => TRUE,
    )

fail a disk belonging to the aggregate, wait for raid reconstruction to kick-in and
within the range specified and do one of the operations like reboot/panic/offline volume/nothing
and subsequently unfail the failed disk. disk should not get assimilated to aggregate and move to
spare pool

returns the output message on success, croaks/die on failure

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use.
    Default is "#rsh".
    for filermode in "MAINT" is "#console"

=item failed_disk

(Required) : Name of disk to be failed

=item aggr

(Required) : Name of aggregate under test

=item aggr_disks

(Optional) : List of all disks belonging to the aggregate "aggr"

=item make_spare

(Optional) : flag, to include -s option in the disk unfail command to move disk to spare
    Default is FALSE (0)

=item recons_range

(Optional) : wait time for disk reconstruction to reach the % range (0-100) before going ahead with
    other operations

=item timeout

(Optional) : time in secs, maximum time the command should wait for the response
        Default is set to 7200 secs

=item forced

(Optional) : flag, to force the disk unfail operation
    Default is TRUE (1)

=item offline

(Optional) : flag, to indicate the volume need to be made offline before performing disk unfail
    Default is FALSE (0)

=item reboot

(Optional) : flag, to reboot the filer before performing disk unfail operation
    Default is FALSE (0)

=item panic

(Optional) : flag, to panic the filer before performing disk unfail operation
    Default is FALSE (0)

=back

=cut

sub disk_unfail_during_reconstruction {
    my %opts = validate(@_,
                        {host        => {type => OBJECT,   isa      => 'Hostrec'},
                         conntype    => {type => SCALAR,   default  => "#console"},
                         failed_disk => {type => SCALAR},
                         aggr_disks  => {type => ARRAYREF, optional => 1},
                         aggr        => {type => SCALAR},
                         makespare   => {type => SCALAR,   optional => 1},
                         recons_range => {type => SCALAR, default => "1-10", optional => 1},
                         timeout      => {type => SCALAR, default => "7200", optional => 1},
                         forced       => {type => SCALAR, default => "1",    optional => 1},
                         offline      => {type => SCALAR, default => "0",    optional => 1},
                         panic        => {type => SCALAR, default => "0",    optional => 1},
                         reboot       => {type => SCALAR, default => "0",    optional => 1},});

    # Local variables
    my ($output, $offline, $status);

    #Bring the filer to normal mode
    logcomment("Boot the appliance to Normal mode");
    transit_to($opts{host}, "UP");

    # Establish the connection to the device
    my $conn = make_conn($opts{host}, $opts{conntype});
    logcomment("conn=$conn");

    # Fail the disk 'failed_disk' belonging to the aggregate 'aggr'
    logcomment("Fail the disk $opts{failed_disk} belonging to the aggregate $opts{aggr}");

    $output = disk_fail(host     => $opts{host},
                        conntype => $opts{conntype},
                        disk     => $opts{failed_disk},
                        options  => "-i",
                        timeout  => $opts{timeout});

    # Wait for raid reconstruction to kick-in and
    # disk reconstruction should be in the range specified

    $status = disk_wait_for_reconstruction(host         => $opts{host},
                                           conntype     => $opts{conntype},
                                           failed_disk  => $opts{failed_disk},
                                           aggr_disks   => $opts{aggr_disks},
                                           aggr         => $opts{aggr},
                                           recons_range => "$opts{range}",
                                           timeout      => $opts{timeout});

    # if disk reconstruction did not get started due to low spares or
    # aggregate is in partial state or whatever
    if (!$status) {
        logcomment("disk_wait_for_reconstruction returned [$status], which was not expected");
        logresult("fail", "disk_wait_for_reconstruction returned [$status], which was not expected");

        #die("disk_wait_for_reconstruction returned [$status], which was not expected");
    }

    # if OFFLINE flag was specified
    # then offline the volume contained in the aggregate
    if ($opts{offline}) {

        # offline aggregate
        logcomment("Trying to offline using aggr command: $opts{aggr}");
        $offline = aggr_offline($conn, $opts{aggr});

        # if aggregate offline did not succeed
        if (!$offline) {

            logcomment("aggr command seems to have failed, Trying to offline " . "using vol command: $opts{aggr}, status=$offline");

            # try to offline volume
            $offline = vol_offline(${conn}, $opts{aggr});

            # even if this fails, then terminate the test
            logresult("fail", "Could not offline aggr/vol $opts{aggr}, status=$offline")
              if (!$offline);

            # offline aggregate
            logcomment("Trying to offline using aggr command: $opts{aggr}");
            aggr_offline($conn, $opts{aggr})
              or die "Could not offline aggregate: $!";

        }    # if ( ! offline )

        logcomment("aggr/vol was successfully offlined, status=$offline");

    }    # if ( offline )

    # if REBOOT flag was specified
    # then perform reboot and wait for appliance to be UP
    if ($opts{reboot}) {

        # perform reboot
        $status = reboot($opts{host});

        # re-establish the connection to the device
        $conn = make_conn($opts{host}, $opts{conntype});

    }    # if ( reboot )

    # if PANIC flag was specified
    # then perform 'panic' using the diag mode
    # make sure appliance recovers from the user initiated panic
    if ($opts{panic}) {

        # Ensure connection still persists
        $conn = make_conn($opts{host}, $opts{conntype});
        logcomment("conn=$conn");

        # perform panic
        eval {

            # perform panic, timeout may need some fine-tuning
            (undef, undef, $status) = say($conn, "priv set diag; panic", '-timeout', "600", '-prompt', "ok", '-prompt', "Loader", '-prompt', "CFE");
        };

        # if say could not recover or obtain the prompt for what-so-ever reason
        if ($@) {

            # it is expected that above say will fail after panic
            # since it would not get the prompt
            # nothing to handle here
        }

        # if after panic, either appliance was left in loader/ok/CFE prompt
        # then making sure appliance is up
        transit_to($opts{host}, "UP");

        # re-establishing the connection
        $conn = make_conn($opts{host});

    }    # if ( panic )

    # Bring the appliance to normal mode
    logcomment("Going to Normal mode");
    transit_to($opts{host}, "UP");

    # if the MAKESPARE flag was enabled
    # then include -s in the command
    # some release(s) do-not support -s, in that case the option would be removed from
    # the command
    logcomment("Reconstruction has started for disk $opts{failed_disk}");
    $output = ($opts{makespare})
      ?

      disk_unfail(host     => $opts{host},
                  conntype => $opts{conntype},
                  disk     => $opts{failed_disk},
                  options  => "-s",
                  timeout  => $opts{timeout})

      :

      disk_unfail(host     => $opts{host},
                  conntype => $opts{conntype},
                  disk     => $opts{failed_disk},
                  timeout  => $opts{timeout});

    # if the output from disk unfail was a failure
    croak "$output" if ($output =~ /Drive could not be failed/);

    # if OFFLINE flag was enabled
    # then aggr/volume was made offline initially before disk unfail
    # now bring the aggr/vol online
    if ($opts{offline}) {

        # offline aggr/vol
        logcomment("Trying to online using aggr command: $opts{aggr}");
        my $online = aggr_online($conn, $opts{aggr});

        if (!$online) {

            logcomment("aggr command seems to have failed, Trying to online " . "using vol command: $opts{aggr}, status=$online");
            $offline = vol_online(${conn}, $opts{aggr});

            logresult("warn", "Could not online aggr/vol $opts{aggr}, status=$online")
              if (!$online);
        }

        logcomment("aggr/vol was successfully online, status=$online");

    }    # if ( offline )

    # for the disk already in spare pool
    # zero spare disks
    eval { disk_zero_spares(host => $opts{host}, conntype => $opts{conntype}); };

    if ($@) {

        # dont worry about this failure
    }

    # return the output of disk unfail command
    return $output;

}    # end of disk_unfail_during_reconstruction

=head2 disk_wait_for_reconstruction

disk_wait_for_reconstruction (
        host        => $host,
        conntype    => "#console",
        aggr        => $aggr,
        failed_disk     => $failed_disk,
        aggr_disks  => $aggr_disks,
        recons_range    => "0-10",
        timeout     => "7200",
    )

disk_wait_for_reconstruction (
        host        => $host,
        failed_disk     => $failed_disk,
        aggr        => $aggr,
        recons_range    => "0-30",
    )

disk_wait_for_reconstruction (
        host        => $host,
        failed_disk     => $failed_disk,
        aggr        => $aggr,
    )

wait for raid reconstruction to kick-in and within the range specified.
Ensure only one disk reconstruction occurs at a time for simplicity.

if there is a situation where two reconstruction are taking place, then
this fuction needs to be enhanced to handle the case by specifying the
disk under reconstruction which this function can monitor

returns TRUE on success and FALSE on failure

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use.
    Default is "#rsh".
    for filermode in "MAINT" is "#console"

=item failed_disk

(Required) : Name of disk to be failed

=item aggr

(Required) : Name of aggregate under test

=item aggr_disks

(Optional) : List of all disks belonging to the aggregate "aggr"

=item recons_range

(Optional) : wait time for disk reconstruction to reach the % range (0-100) before going ahead with
    other operations

=item timeout

(Optional) : time in secs, maximum time the command should wait for the response
        Default is set to 7200 secs

=item count

(Optional) : currently un-used, the idea is to have maximum iterations to wait for
        reconstruction to get kicked-in, which can override timeout
        Default is set to 0 (zero)

=back

=cut

# TODO: This function must be enhanced to in-corporate
# situations when there are no usable spare disks, this case reconstruction will never get started
sub disk_wait_for_reconstruction {
    my %opts = validate(@_,
                        {host     => {type => OBJECT, isa      => 'Hostrec'},
                         conntype => {type => SCALAR, optional => 1, default => "#console"},
                         aggr         => {type => SCALAR},
                         failed_disk  => {type => SCALAR},
                         recons_range => {type => SCALAR, optional => 1, default => "0-10"},
                         timeout      => {type => SCALAR, optional => 1, default => "900"},
                         aggr_disks => {type => ARRAYREF, optional => 1},
                         count      => {type => SCALAR,   optional => 1, default => "0"},});

    # Local variables
    my ($output, $recons_disk, $per_recons, $timeout);
    my $cmd = " aggr status -r $opts{aggr} ";
    my ($range_start, $range_end) = split /-/, $opts{range};
    print("2------the command $cmd------\n");

    # Check if reconstruction has started for the failed disk
    # if yes, then look for the % to be within the expected range
    eval {

        # Dump the aggregate disk status
        $output = execute_cmd(host     => $opts{host},
                              conntype => $opts{conntype},
                              cmd      => $cmd);

        # Check if the raid reconstruction has started
        # if yes, retrieve the disk name and reconstruction %
        # currently expects only one disk reconstruction to occur at a time
        # b'cos if two disks are being reconstructed (for raid_dp) then it might occur that
        # the failed disk under test might not have started reconstruction but other reconstruction
        # could have kicked in and within the range specified, which would defy the logic here
        if (($output =~ /\w+\s+(\w+(\.\w+)+).*reconstruction (\d+)% completed/mi)) {
            $recons_disk = $1;    # disk name
            $per_recons  = $2;    # reconstruction % completion
        }

        $timeout = $opts{timeout};

        my $iteration = 1;

        # wait for reconstruct to kick-off
        # check if reconstruction is within the expected range
        while (($per_recons eq "" or $per_recons < $range_start or $per_recons < $range_end)
               and $timeout) {

            logcomment("Waiting for reconstruction, iteration=$iteration, timeout=$timeout");
            $output = execute_cmd(host     => $opts{host},
                                  conntype => $opts{conntype},
                                  cmd      => $cmd);

            if (($output =~ /\w+\s+(\w+(\.\w+)+).*reconstruction (\d+)% completed/mi)) {

                $recons_disk = $1;    # disk name
                $per_recons  = $3;    # % completion

                logcomment("disk=$recons_disk, percent=$per_recons");

                last if (    $per_recons >= $range_start
                         and $per_recons <= $range_end);
            } ## end if (($output =~ /\w+\s+(\w+(\.\w+)+).*reconstruction (\d+)% completed/mi...

            # Either disk reconstruction did not start (or)
            # did not meet the reconstruction % range
            sleep(60);

            $timeout -= 60;

            croak "Either disk reconstruction did not start (or) did not reach the % expected range ($opts{recons_range})"
              if (!$timeout);

            $iteration++;    # keep a measure of elapsed-time

        }    # while

    };    # eval

    # Something bailed out waiting for disk reconstruction
    if ($@) {

        # just propogate the error for now
        croak("$@");
    }

    # if TIMEOUT has elapsed
    if (!$timeout) {
        logcomment(  "Either Disk reconstruction did not get started due to low spares (OR) "
                   . "it did not reach $range_start % before timeout. "
                   . "current values are: disk=$recons_disk, percent=$per_recons, timeout=$timeout");

        # return FALSE
        return 0;
    }

    # if everything went well
    # disk reconstruction is in-progress and within specified range
    logcomment("Disk reconstruction must be in-progress, disk=$recons_disk, percent=$per_recons");

    # return TRUE
    return 1;

}    # end of disk_wait_for_reconstruction

# ====================== DEFINE LOCAL ROUTINES ===================
#-----------------------------------------------------------------
# _diskfwcommand
#
# Usage:
#
# The function is called by disk_fw_update. It is used to implement
# disk_fw_update command according to the input. And according to the
# value $wait, it will decide whether the function will wait for the
# command finished.
# Issue disk_fw_update and waiting for the message:
# firmware update?      command issuing successful
# already in progress   there is disk_fw_update in progress
# Usage:                command issue with syntax error
# not exist:            invalid disk id
# ignore disk error:    command issuing successful for disk_fw_udpate -f
#
# -----------------------------------------------------------------
sub _diskfwcommand($$$$$$) {
    my ($host, $conntype, $cmd, $wait, $wait_time, $timeout) = @_;
    my ($output1, $output3, $output2, $output4, $event_by_type, $count, $state_flag, $end_event, $start_event, $text, $dw_failed_event);
    $count      = 0;
    $state_flag = 0;
    my $count_start = 0;
    my $start_time  = 240;
    my $sleep_time  = 60;
    my $start_flag  = 0;
    my $mark        = ems_log_get_mark(host => $host);
    #($output1, $text) = filer_cli_advanced_say($conn, $cmd);
    $output1 = filer_cmd(host=>$host, conntype =>$conntype, cmd=>"priv set advanced; $cmd");

    if($output1 =~ /Disk firmware update already in progress/){
        croak $output1;
    } elsif ($output1 =~ /Usage/) {
        croak "This must be run on a debug build.";
    }
    while ($count_start < $start_time) {
        $count_start = $count_start + $sleep_time;
        logcomment("Waiting for 60 secs");
        sleep $sleep_time;
        $event_by_type = ems_log_dump_by_type(host => $host, timestamp => $mark);

        #--------------------------------------
        # Check for disk fw failure events
        #-----------------------------------
        $dw_failed_event = $event_by_type->{"dfu.fwDownloadFailedWbuf"};
        if ($dw_failed_event) {
            logresult('FATAL', "Disk firmware update failed and reported event dfu.fwDownloadFailedWbuf");
        } else {
            logcomment("No download failure events reported so far...");
        }
        $start_event = $event_by_type->{"dfu.firmwareDownloading"};
        if ($start_event) {
            logcomment("Reported event : dfu.firmwareDownloading");
            logcomment("=======Starting firmware update in $count=====");
            $start_flag = 1;
            last;
        }

        # anishv: Following code was causing false pass even though there is no download done. Which is highly undesirable. With this change we will see a lot of failures which were previously reporting wrong pass. :-)

        #else { #now handle the r200 case, where dfu.firmwareUpToDate occurs immediately

        #$start_event = $event_by_type->{"dfu.firmwareUpToDate"};

        #

        #if ("$start_event" ne "") {

        #logcomment("Reported event : dfu.firmwareUpToDate");

        #logcomment ("=======Starting firmware update in $count=====");

        #$start_flag = 1;

        #goto DFU_LABEL;

        #}

        #}
    } ## end while ($count_start < $start_time)

    #DFU_LABEL: ;

    if ($start_flag == 0) {
        croak $output1;
    }

    # Decide whether it is needed to wait for the disk_fw_update finished
    printf "Now the value of wait is $wait.\n";

    #=============================================================================================
    # end_event always not a dfu.firmwareUpToDate,if firmware file is higer revision
    # then firmware update never reports the above  event, instead it reports dfu.fwDownloaded
    # script will wait long enough and report download failed.
    # Now adding checking either dfu.fwDownloaded or dfu.firmwareUpToDate event for the specified wait time
    # always check recent event assuming one or more drives selected,if there are no failed download events
    # at least one succesful download completion found then make it pass,otherwise fail
    # caller has to make sure the value for  wait_time based on number of disks selected for fw download
    #======================================================================================================
    if ($wait != 0) {
        logcomment("waiting for $wait_time secs till  no download failure events reported,checking for download completion events");
        while ($count < $wait_time) {
            $event_by_type = ems_log_dump_by_type(host => $host, timestamp => $mark);

            #--------------------------------------
            # Check for disk fw failure events
            #-----------------------------------
            $dw_failed_event = $event_by_type->{"dfu.fwDownloadFailedWbuf"};
            if ($dw_failed_event) {
                logresult('FATAL', "Disk firmware update failed and reported event dfu.fwDownloadFailedWbuf");
            } else {
                logcomment("No  download failure  events reported so far...");
            }
            $end_event = $event_by_type->{"dfu.firmwareUpToDate"};
            if ($end_event ne "") {
                logcomment("found event : dfu.firmwareUpToDate,which indicates firmware is up to date");
                logcomment("=======Finishing firmware update in $count=====");
                $state_flag = 1;

                #      last;
            } else {
                $end_event = $event_by_type->{"dfu.fwDownloaded"};
                if ($end_event ne "") {
                    logcomment("found event : dfu.fwDownloaded which indicates firmware updated successfully");
                    logcomment("=======Finishing firmware update in $count=====");
                    $state_flag = 1;

                    #         last;
                }
            } ## end else [ if ($end_event ne "")
            $count = $count + 120;
            printf "Now the count is $count.\n";
            sleep 120;
            logcomment("wait time still not yet finished,checking for any new failure events");
        }    #while loop

        if ($state_flag == 0) {
            croak "Can not finish disk firmware update in $wait_time.";
        }
    }    # $wait
    logcomment("Disk firmware update successfully finished on all disks,no failure events reported");
}    #_diskfwcommand


# --------------------------------------------------
# _disk_scrub_is_started - verifies scrub should be
# within a few seconds. Return FAIL if it could be
# started.
#
# parameters -
#  $conn - connection obj
# --------------------------------------------------
sub _disk_scrub_is_started($) {
    my ($conn) = $_[0];
    my $waittime = 5;

    _debug("_disk_scrub_is_started: ...");

    my $ro = new RaidInfo($conn);

    # -----------------------------------------
    # checking if scrub is started
    # -----------------------------------------
    my %aggrs = $ro->all_raid_groups();
    while (my ($key, $rg) = each(%aggrs)) {
        _debug("raidgroup key=$key ; $rg->{scrub_progress}");
        if ($rg->{scrub_progress} ne "-1") {
            _debug("_disk_scrub_is_started: scrub is in progress <$rg->{scrub_progress}>");
            return PASS;
        }
    }    #each raidgroup

    # -----------------------------------------------
    # sleep at most 5 seconds for scrub to be started
    # if it still not then exit
    # -----------------------------------------------
    sleep($waittime);

    $ro->refresh();
    while (my ($key, $rg) = each(%aggrs)) {
        _debug("raidgroup key=$key ; $rg->{scrub_progress}");
        if ($rg->{scrub_progress} ne "-1") {
            _debug("_disk_scrub_is_started: scrub is in progress <$rg->{scrub_progress}>");
            return PASS;
        }
    }    #each raidgroup
    logresult('fail', "Scrub could not be started within <$waittime> seconds");
    return FAIL;
}    #_disk_scrub_is_started

# --------------------------------------------------
# _disk_scrub_wait - waits until all scrubs are done
# and reports any media, pi, and checksum errors.
# parameter
# $conn - connection obj
# --------------------------------------------------
sub _disk_scrub_wait($$) {
    _debug("_disk_scrub_wait: ...");
    my ($conn, $waittime) = @_;

    my $sleep_time = 120;
    if (RaidUtils_is_sim($conn)) {
        $sleep_time = 30;
    }

    my $ro  = new RaidInfo($conn);
    my $cnt = -1;

    # -----------------------------------------
    # checking if scrub is still going.
    # Verify if there is any PI, media errors, or
    # check sum error.
    # -----------------------------------------
    while ($cnt <= $waittime || $waittime == WAIT_UNTIL_DONE) {
        $ro->refresh();
        my %aggrs       = $ro->all_raid_groups();
        my $scrub_state = 0;

        while (my ($key, $rg) = each(%aggrs)) {
            _debug("raidgroup key=$key ; $rg->{scrub_progress}");
            if ($rg->{scrub_progress} ne "-1") {
                _debug("scrub is in progress <$rg->{scrub_progress}>");
                $scrub_state = 1;
            }

            # ----------------------------
            # check for any type of errors
            # ----------------------------
            if ($rg->{scrub_pi} ne "0") {
                croak("Found scrub PI error in raidgroup <$key>");
            }
            if ($rg->{scrub_media_err} ne "0") {
                croak("Found scrub media err in raidgroup <$key>");
            }
            if ($rg->{scrub_cksum_err} ne "0") {
                croak("Found scrub check sum error in raidgroup <$key>");
            }
        }    #each raidgroup

        # --------------------
        # finish when no scrub
        # --------------------
        if ($scrub_state == 0) {
            last;
        }

        # sleep a little while
        _debug("sleeping <$sleep_time> seconds waiting for scrub to be completed.");
        sleep($sleep_time);
        $cnt += $sleep_time;
    }    # while scrubbing

    tstlog("-comment", "Scrub is now completed.");
    return PASS;
}    #_disk_scrub_wait

# ------------------------------------------------------------
# _disk_wait_until - waits until the specified disk to be
# listed on the spare or broken pool.
# ------------------------------------------------------------
sub _disk_wait_until($$$$) {
    _debug("_disk_wait_until: <$_[1]>");
    my ($conn, $disk, $waittime, $target) = @_;
    my $ro  = new RaidInfo($conn);
    my $cnt = -1;

    my $sleep_time = 60;
    if (RaidUtils_is_sim($conn)) {
        $sleep_time = 20;
    }

    # --------------------------------------------
    # checking for the unfailed disk to be listed
    # as part of the spare pool.
    # --------------------------------------------
    my %fdisks;
    while ($cnt <= $waittime || $waittime == WAIT_UNTIL_DONE) {
        if ($target eq 'failed') {
            %fdisks = $ro->failed_disks;
        } elsif ($target eq 'spare') {
            %fdisks = $ro->spare_disks;
        } else {
            tstlog(-fatal, "Unable to find the target <$target>");
        }

        if (defined($fdisks{$disk})) {
            logresult('pass', "Disk <$disk> has been moved to the $target list");
            return PASS;
        }
        _debug("_disk_wait_until: wait for <$disk> to be listed on the $target list in <$cnt> seconds");
        sleep($sleep_time);
        $cnt += $sleep_time;
        $ro->refresh();    # this is too expensive

    }    #checking the unfailed disk or out of time
    logresult('fail', "Disk <$disk> did not make to the $target list in <$waittime> seconds");
    return FAIL;
}    #_disk_wait_until

# ------------------------------------------------------------
# _sim_get_disk_fullname - retrieves a full path name of the
# specified disks based on their unique disk serial number.
# This routine is needed for simulator use only.
#
# Arguments:
#   $serialnum - (Required) the disk serial number
#   $simdir    - (Optional) the full path of the sim directory
#
# Returns - a list of specified disks full path name.
#
# ------------------------------------------------------------
sub _sim_get_disk_fullname($$) {
    my ($serialnum, $simdir) = @_;

    my @disks_list;
    if ($simdir eq "") {
        @disks_list = glob(",disks/,pulled/*:*");
    } else {
        @disks_list = glob("$simdir/,disks/,pulled/*:*");
    }

    my @serial_list = split(/ /, $serialnum);
    my $return_list;

    foreach my $diskpath (@disks_list) {
        foreach my $serialnum (@serial_list) {
            if ($diskpath =~ /$serialnum/) {
                _debug("Found a disk match <$diskpath>");
                my @sel_disk = split("/", $diskpath);
                $return_list = $return_list . $sel_disk[$#sel_disk] . " ";
            }
        }
    }    # search for each disk path
    return $return_list;
}    #_sim_get_disk_fullname

# -----------------------------------------------------------
# _debug - providing additional information for debugging.
# Debug messages will be controlled via a parameter VERBOSE.
# By default VERBOSE is set to
# -----------------------------------------------------------
sub _debug($) {
    if ($main::VERBOSE) {
        print("\nDEBUG: $_[0]\n");
    }
}    #_debug

# ------------------------------------------------------------
# _verify_result - verifies the result based on the passing
# error msg and prints the appropriate messages.
# Args -
#   err - passing the error message $@
#   msg - passing the info to be written out
#   exception - set if expecting an exception.
# Return
#   PASS - no error
#   FAIL - upon error
# ------------------------------------------------------------
sub _verify_result {
    my (%pars) =
      validate(@_,
               {err       => {type => SCALAR, optional => YES},
                msg       => {type => SCALAR, optional => YES},
                exception => {type => SCALAR, optional => YES},});

    # -------------------------------------------
    # Handling the expected exceptional scenarios
    # -------------------------------------------
    if (defined($pars{exception})) {
        if ($pars{err}) {
            tstlog(-pass, "Caught the expected exception: $pars{err}");
            return PASS;
        }

        tstlog(-fail, "Expecting an exception of the $pars{msg}. However, it did not occur.");
        return FAIL;
    }

    if ($pars{err}) {
        tstlog(-fail, "The request of $pars{msg} failed due to: $pars{err}");
        return FAIL;
    }

    tstlog(-pass, "The request of $pars{msg} succeeded");
    return PASS;
}    #_verify_result

# ------------------------------------------------------------
# ===================== DEFINE TESTME MODULES ================

#
# testing the function disk_dump
#
sub testme_disk_dump ($$$$) {

    package main;

    our ($FILER, $FILER_CONN, $SRC_DISK, $DST_DISK);

    param('FILER',      '-mesg',    "Filer to test");
    param('FILER_CONN', '-default', '#console');
    param('SRC_DISK',   '-mesg',    "Source Disk");
    param('DST_DISK',   '-mesg',    "Destination Disk");

    package Storage::Disk;

    my ($host, $conntype, $src_disk, $dst_disk) = (make_host($FILER), $FILER_CONN, $SRC_DISK, $DST_DISK);

    my $status;
    eval {
        tstlog(-comment, "start disk dump from $src_disk disk to $dst_disk disk");

        $status = disk_dump(host     => $host,
                            conntype => $conntype,
                            src      => $src_disk,
                            dst      => $dst_disk);

        tstlog(-comment, "start disk dump from $src_disk disk to $dst_disk disk");
        $status = disk_dump(host     => $host,
                            conntype => $conntype,
                            src      => $src_disk,
                            dst      => $dst_disk,
                            count    => 1,);

        tstlog(-comment, "start disk dump from $src_disk disk to $dst_disk disk from blk 10 to 20");
        $status = disk_dump(host      => $host,
                            conntype  => $conntype,
                            src       => $src_disk,
                            dst       => $dst_disk,
                            blk_start => "10",
                            blk_end   => "20",);
    };

    if ($@) {
        logresult('fail', "$@");
    } else {
        logresult('pass', "Able to disk dump from $src_disk to $dst_disk [$status]");
    }
} ## end sub testme_disk_dump ($$$$)

#
# testing the function disk_unfail_during_reconstruction
#
sub testme_disk_unfail_during_reconstruction ($$$$$) {

    package main;

    our ($FILER, $FILER_CONN, $DISK, $AGGR, $OPERATION);

    param('FILER',      '-mesg',    "Filer to test");
    param('FILER_CONN', '-default', '#console');
    param('DISK',       '-mesg',    "DISK to FAIL");
    param('AGGR',       '-mesg',    "Aggregate");
    param('OPERATION',  '-default', "");

    package Storage::Disk;

    my ($host, $conntype, $failed_disk, $aggr, $operation) = (make_host($FILER), $FILER_CONN, $DISK, $AGGR, $OPERATION);

    my $status;
    eval {
        if ($operation eq "offline") {

            $status =
              disk_unfail_during_reconstruction(host        => $host,
                                                failed_disk => $failed_disk,
                                                aggr        => $aggr,
                                                offline     => 1,);
        } elsif ($operation eq "panic") {

            $status =
              disk_unfail_during_reconstruction(host        => $host,
                                                failed_disk => $failed_disk,
                                                aggr        => $aggr,
                                                panic       => 1,);
        } elsif ($operation eq "reboot") {

            $status =
              disk_unfail_during_reconstruction(host        => $host,
                                                failed_disk => $failed_disk,
                                                aggr        => $aggr,
                                                reboot      => 1,);
        } else {

            $status =
              disk_unfail_during_reconstruction(host        => $host,
                                                failed_disk => $failed_disk,
                                                aggr        => $aggr,);
        }
    };

    if ($@) {
        logresult('fail', "$@");
    } else {
        logresult('pass', "Disk unfail succeeded [$status]");
    }
} ## end sub testme_disk_unfail_during_reconstruction ($$$$$)

#
# testing the function disk_unfail_during_reconstruction
#
sub testme_disk_wait_for_reconstruction ($$$$$) {

    package main;

    our ($FILER, $FILER_CONN, $DISK, $AGGR, $RANGE, $TIMEOUT);

    param('FILER',      '-mesg',    "Filer to test");
    param('FILER_CONN', '-default', '#console');
    param('DISK',       '-mesg',    "DISK to FAIL");
    param('AGGR',       '-mesg',    "Aggregate");
    param('RANGE',      '-default', "0-10");
    param('TIMEOUT',    '-default', "600");

    package Storage::Disk;

    my ($host, $conntype, $failed_disk, $aggr, $range, $timeout) = @_;

    #( make_host($FILER), $FILER_CONN, $DISK, $AGGR, $RANGE, $TIMEOUT );

    my $status;

    $timeout = $timeout || 7200;
    $range   = $range   || "0-10";

    eval {
        $status = disk_wait_for_reconstruction(host         => $host,
                                               conntype     => $conntype,
                                               failed_disk  => $failed_disk,
                                               aggr         => $aggr,
                                               recons_range => "$range",
                                               timeout      => $timeout);

    };

    if ($@) {
        logresult('FAIL', "$@");
    } else {
        logresult('PASS', "Reconstruction must have started now and in range: $status");
    }
} ## end sub testme_disk_wait_for_reconstruction ($$$$$)

# ------------------------------
# testing the command disk_scrub
# ------------------------------
sub testme_disk_scrub($$) {
    _debug("Entering testme_disk_scrub: ...");
    my ($host, $filermode) = @_;

    # ------------------------
    # start scrub on all disks
    # ------------------------
    tstlog(-comment, "start scrub without waiting");
    if (!disk_scrub(host => $host, start => 1, waittime => 0)) {
        logresult('fail', "unable to start disk scrub: $@");
    }

    tstlog(-comment, "start scrub and wait until completion");
    if (!disk_scrub(host => $host, start => 1)) {
        logresult('fail', "unable to start disk scrub: $@");
    }

    tstlog(-comment, "start scrub without wait");
    if (!disk_scrub(host => $host, start => 1, waittime => 0)) {
        logresult('fail', "unable to start disk scrub: $@");
    }

    tstlog(-comment, "stop scrub when it is still going");
    if (!disk_scrub(host => $host, stop => 1)) {
        logresult('fail', "unable to stop disk scrub: $@");
    }
}    #testme_disk_scrub

# -----------------------------
# testing the command disk_fail
# -----------------------------
sub testme_disk_fail($$) {
    _debug("Entering testme_disk_fail: ...");
    my ($host, $filermode) = @_;

    tstlog(-comment, "testing disk_fail command");
    my $conn = make_conn($host);
    my $ro   = new RaidInfo($conn);

    if (vol_exists($conn, VOLNAME)) {
        vol_offline($conn, VOLNAME);
        vol_destroy($conn, VOLNAME);
    }

    # ----------------------------------------------
    # need at least 3 spares to work on both filers
    # and simulators
    # ----------------------------------------------
    my %spares = $ro->spare_disks();
    if (keys(%spares) < 3) {
        tstlog(-fail, "There are not enough spares for testing");
        return;
    }
    my @disks_list;

    # create a test volume with 3 disks
    foreach my $disk (keys %spares) {
        if ($#disks_list >= 2) {
            last;
        }
        push(@disks_list, $disk);
    }    #get spares

    vol_create($conn, VOLNAME, \@disks_list);
    sleep(10);

    # fail one of the disks
    tstlog(-comment, "Try to fail the first disk");

    # ---------------------------------
    # testing fail and wait until done
    # ---------------------------------
    eval { disk_fail(host => $host, disk => $disks_list[0]); };
    _verify_result(err => $@, msg => "disk_fail command");

    # ------------------------------------------------
    # fails and waits until the disk is listed on the
    # failed list or exceeding the specified time.
    # ------------------------------------------------
    eval { disk_fail(host => $host, disk => $disks_list[1], options => '-i', waittime => '1200'); };
    _verify_result(err => $@, msg => "disk_fail command");

    vol_offline($conn, VOLNAME);
    vol_destroy($conn, VOLNAME);

    # ---------------------------------------------------------
    # By now we should have a failed disks. Let bring them back
    # on spare list.
    # ---------------------------------------------------------
    eval { disk_unfail(host => $host, disk => $disks_list[0]); };
    _verify_result(err => $@, msg => "disk_fail command");

    eval { disk_unfail(host => $host, disk => $disks_list[1], waittime => '1200'); };
    _verify_result(err => $@, msg => "disk_fail command");

    # perform abnormally test
    tstlog(-comment, "** Creating an exception case of disk_fail **");
    my $dd = "foo";
    eval { disk_fail(host => $host, disk => $dd); };
    _verify_result(err => $@, msg => "disk_fail command <$dd> ", exception => 1);
}    #testme_disk_fail

# ------------------------
# Need to be added by Ying
# ------------------------
sub testme_disk_fw_update($$) {
    _debug("Entering testme_disk_fw_update: ...");

}    #testme_disk_fw_update

sub testme_disk_sanitize($$) {
    _debug("Entering testme_disk_sanitize: ...");

    my ($host, $filermode) = @_;
    my $conn = make_conn($host);
    my $ro   = new RaidInfo($conn);

    my %spares = $ro->spare_disks();
    if (keys(%spares) < 2) {
        tstlog(-fail, "There are not enough spares for testing");
        return;
    }

    my @disks_list;

    # create a test volume with 4 disks
    foreach (keys %spares) {
        if ($#disks_list >= 4) {
            last;
        }
        _debug("Testing disk sanitize on disk <$_>");
        push(@disks_list, $_);
    }    #get spares

    my $hr;
    eval { disk_sanitize(host => $host, cmds => 'start', disk => $disks_list[0], waittime => '60'); };
    _verify_result(err => $@, msg => "disk sanitize start command");

    eval { $hr = disk_sanitize(host => $host, cmds => 'status'); };
    if (_verify_result(err => $@, msg => "disk sanitize status command")) {
        print Dumper($hr);
    }

    # --------------------------------------------
    # start sanitizing a disk and quickly abort it
    # --------------------------------------------
    eval { disk_sanitize(host => $host, cmds => 'start', disk => $disks_list[1], waittime => 10); };
    _verify_result(err => $@, msg => "disk sanitize start command <$disks_list[1]>");

    eval { disk_sanitize(host => $host, cmds => 'abort', disk => $disks_list[1]); };
    _verify_result(err => $@, msg => "disk sanitize abort command <$disks_list[1]>");

    # -----------------------------------------
    # let check out the list of sanitized disks
    # -----------------------------------------
    $hr = disk_sanitize(host => $host, cmds => 'status');
    if (_verify_result(err => $@, msg => "disk sanitize status command")) {
        print Dumper($hr);
    }

    # --------------------------------------------------------------
    # if the hash table is empty then there is no reason to continue
    # on with the testing. No licensing for disk sanitize could well
    # be one of the scenarios.
    # --------------------------------------------------------------
    if (keys(%$hr) < 1) {
        tstlog(-fail, "No disk sanitize is going on. Skip the rest of the test.");
        return FAIL;
    }

    # --------------------------------------
    # let abort all on going disk sanitizing
    # except for those are 100% complete
    # --------------------------------------
    my $disks_list;
    foreach (keys(%$hr)) {
        if ($hr->{$_}->{percent_complete} ne "100") {
            $disks_list .= $_ . " ";
        } else {
            print("Skiping disk <$_> since it was sanitized fully\n");
        }
    }
    eval { $hr = disk_sanitize(host => $host, cmds => 'abort', disk => $disks_list); };
    _verify_result(err => $@, msg => "disk sanitize abort command <$disks_list>");

    # -------------------------------------------------
    # Sanitize a few disks and wait until they are done
    # -------------------------------------------------
    my $disks = $disks_list[2] . " " . $disks_list[3];
    eval { disk_sanitize(host => $host, cmds => 'start', disk => $disks); };
    _verify_result(err => $@, msg => "disk sanitize start <$disks>");

    eval { $hr = disk_sanitize(host => $host, cmds => 'status'); };

    # ------------------------------------------------
    # Release all disks sanitized before exit the test
    # ------------------------------------------------
    if (_verify_result(err => $@, msg => "disk sanitize status command")) {
        print Dumper($hr);
        my $ongoing_list = "";
        my $full_list    = "";
        foreach (keys(%$hr)) {
            if ($hr->{$_}->{percent_complete} eq "100") {
                $full_list .= $_ . " ";
            } else {
                $ongoing_list .= $_ . " ";
            }
        }

        if ($ongoing_list ne "") {
            eval { $hr = disk_sanitize({host => $host, cmds => 'abort', disk => $ongoing_list}); };
            _verify_result(err => $@, msg => "disk sanitize abort command <$ongoing_list>");
        }

        if ($full_list ne "") {
            eval { $hr = disk_sanitize({host => $host, cmds => 'release', disk => $full_list}); };
            _verify_result(err => $@, msg => "disk sanitize release command <$full_list>");
        }
    } ## end if (_verify_result(err...
}    #testme_disk_sanitize

# -------------------------------------------------------------------
# testme_disk_shm_stats - performs test of the routine disk_shm_stats
# -------------------------------------------------------------------
sub testme_disk_shm_stats($$) {
    _debug("Entering testme_disk_shm_stats: ...");

    my ($host, $filermode) = @_;

    my $hr;
    eval { $hr = disk_shm_stats(host => $host); };
    if (_verify_result(err => $@, msg => "disk shm_state command")) {
        print Dumper($hr);
    }

    my ($key, $val) = each(%$hr);
    my $hr2;
    eval { $hr2 = disk_shm_stats(host => $host, disk => $key); };
    if (_verify_result(err => $@, msg => "disk shm_state command")) {
        print Dumper($hr2);
    }
}    #testme_disk_shm_stats

# --------------------------
# testing disk_maint command
# --------------------------
sub testme_disk_maint($$) {
    _debug("Entering testme_disk_maint:...");

    my ($host, $filermode) = @_;
    my $hr;

    # ----------------------
    # verify disk maint list
    # ----------------------
    eval { $hr = disk_maint(host => $host, maintcmd => "list"); };
    if ($@) {

        # ------------------------------------------------
        # In case disk maint is not support for the kernel
        # then simply exit the test
        # ------------------------------------------------
        if ($@ =~ /Did not recognize option "maint"/) {
            tstlog(-fail, "Skip the disk_maint test since there is no disk maint command available for the current kernel");
            return;
        }
        tstlog(-fail, "disk maint list failed due to <$@>");
    } else {
        print Dumper ($hr);
    }

    # ----------------------
    # verify disk maint list
    # ----------------------
    eval { $hr = disk_maint(host => $host, maintcmd => "status"); };
    if (_verify_result(err => $@, msg => "disk maint status command")) {
        print Dumper($hr);
    }

    # -----------------------
    # verify disk maint start
    # -----------------------
    my $conn = make_conn($host);
    my $ro   = new RaidInfo($conn);
    my $disklist;
    my %spares = $ro->spare_disks();

    if (keys(%spares) < 2) {
        tstlog(-fail, "Need 2 spares to run disk maint start");
        return;
    }
    my $cnt = 1;
    foreach my $key (keys(%spares)) {
        $disklist .= " " . $key;
        if ($cnt++ >= 2) {
            last;
        }
    }

    eval { disk_maint(host => $host, disklist => $disklist, maintcmd => 'start', waittime => '30'); };
    _verify_result(err => $@, msg => "disk maint start command <$disklist>");

    # let abort all disk maint testing
    eval { disk_maint(host => $host, disklist => $disklist, maintcmd => 'abort'); };
    _verify_result(err => $@, msg => "disk maint abort command <$disklist>");

    # ---------------------------------------------------------------
    # Creating a test vol, starting disk maint one of the file system
    # disk, and waiting until disk maint is done.
    # ---------------------------------------------------------------

    $ro->refresh();
    %spares = $ro->spare_disks();
    my @disks;

    # create a test volume with 3 disks
    foreach my $disk (keys %spares) {
        if ($#disks >= 2) {
            last;
        }
        push(@disks, $disk);
    }    #get spares

    if (vol_exists($conn, VOLNAME)) {
        vol_offline($conn, VOLNAME);
        vol_destroy($conn, VOLNAME);
    }
    vol_create($conn, VOLNAME, \@disks);

    eval { disk_maint(host => $host, maintcmd => 'start', disklist => $disks[0], immediate => "-i"); };
    _verify_result(err => $@, msg => "disk maint start command <$disks[0]>");

    if (vol_exists($conn, VOLNAME)) {
        vol_offline($conn, VOLNAME);
        vol_destroy($conn, VOLNAME);
    }
    tstlog(-comment, "Finishing testme_disk_maint");
}    #testme_disk_maint

# -------------------------------
# testing the disk_unfail command
# -------------------------------
sub testme_disk_unfail($$) {
    _debug("Entering testme_disk_unfail: ...");

    my ($host, $filermode) = @_;
    tstlog(-comment, "testing disk_unfail command");

    my $conn = make_conn($host);
    my $ro   = new RaidInfo($conn);

    # requires a minimum of 3 spares
    my %spares = $ro->spare_disks_pool0();
    if (keys(%spares) < 3) {
        %spares = $ro->spare_disks_pool1();
        if (keys(%spares) < 3) {
            tstlog(-fail, "There is not enough spares for testing");
        }
    }

    my @disks_list;

    # create a test volume using 3 disks
    foreach my $disk (keys %spares) {
        if ($#disks_list >= 2) {
            last;
        }
        push(@disks_list, $disk);
    }    #get spares

    vol_create($conn, "foo", \@disks_list);
    sleep(10);

    my $sel_disk = $disks_list[0];

    # -----------
    # fail a disk
    # -----------
    eval { disk_fail(host => $host, disk => $sel_disk, options => '-i', waittime => '1200'); };
    if ($@ ne "") {
        tstlog(-fail, "disk_fail failed due to: <$@>");
        vol_offline($conn, VOLNAME);
        vol_destroy($conn, VOLNAME);
        return FAIL;
    }

    eval { disk_unfail(host => $host, disk => $sel_disk, filermode => 'normal', waittime => 200); };
    if ($@ ne "") {
        tstlog(-fail, "disk_unfail failed due to: <$@>");
        vol_offline($conn, VOLNAME);
        vol_destroy($conn, VOLNAME);
        return FAIL;
    }

    vol_offline($conn, VOLNAME);
    vol_destroy($conn, VOLNAME);

    # ---------------------------
    # create an abnormal scenario
    # ---------------------------
    tstlog(-comment, "** Creating an exception case of disk_unfail **");
    my $dd = "foo";

    eval { disk_unfail(host => $host, disk => $dd); };
    _verify_result(err => $@, msg => 'disk unfail command <$dd>', exception => '1');
}    #testme_disk_unfail

# --------------------------------
# testing the function disk_remove
# --------------------------------
sub testme_disk_remove($$) {
    _debug("Entering testme_disk_remove: ...");

    my ($host, $filermode) = @_;

    tstlog(-comment, "testing disk_remove function");
    my $conn = make_conn($host);
    my $ro   = new RaidInfo($conn);

    # get 2 spares
    my %spares = $ro->spare_disks_pool0();
    if (keys(%spares) < 1) {
        %spares = $ro->spare_disks_pool1();
        if (keys(%spares) < 1) {
            tstlog(-fail, "There is no sufficient spare for testing");
            return;
        }
    }

    my @sel_disk = %spares;
    eval { disk_remove(host => $host, disk => $sel_disk[0], waittime => '1200'); };
    _verify_result(err => $@, msg => 'disk remove command');

    tstlog(-comment, "** Creating an exception case of disk remove **");
    my $dd = " ";
    eval { disk_remove(host => $host, disk => $dd); };
    _verify_result(err => $@, msg => 'disk remove command', exception => '1');
}    #testme_disk_remove

# --------------------------------------------------
# testing both functions disk_assign and disk_remove
# --------------------------------------------------
sub testme_assign_remove($$) {
    my ($host, $filermode) = @_;

    my $conn      = make_conn($host);
    my $filername = $host->hostname();

    my $ro     = new RaidInfo($conn);
    my %spares = $ro->spare_disks();
    if (keys(%spares) < 2) {
        tstlog(-fail, "There are not enough spares for testing");
        return FAIL;
    }

    # --------------------------------------------
    # select a couple disks to be used for testing
    # --------------------------------------------
    my @disks_list;
    foreach my $disk (keys %spares) {
        if ($#disks_list >= 1) {
            last;
        }
        push(@disks_list, $disk);
    }    #get spares

    eval { disk_remove_ownership(host => $host, disk => $disks_list[0]); };
    _verify_result(err => $@, msg => "disk remove ownership command");

    eval { disk_remove_ownership(host => $host, disk => $disks_list[1], foption => '-f'); };
    _verify_result(err => $@, msg => "disk remove ownership command");

    tstlog(-comment, "** Creating an exception case **");
    eval { disk_remove_ownership(host => $host, disk => 'foo'); };
    _verify_result(err => $@, msg => "disk remove ownership <foo>", exception => '1');

    eval { disk_assign(host => $host, disk => $disks_list[0]); };
    _verify_result(err => $@, msg => "disk assign command <disk=>$disks_list[0]>");

    # ----------------------------------------------------
    # testing with "all" which only works under maint mode
    # ----------------------------------------------------
    eval { disk_assign(host => $host, disk => "all"); };
    _verify_result(err => $@, msg => "disk assign command <all>");
}    #testme_assign_remove

# ---------------------------------------
# testing the disk_assign and disk remove
# ---------------------------------------
sub testme_disk_assign($$) {
    my ($host, $filermode) = @_;

    testme_assign_remove($host, $filermode);
}

# ----------------------------------------------------
# testing both the disk_resume and disk_suspend module
# ----------------------------------------------------
sub _test_disk_suspend_resume($$) {
    my ($host, $filermode) = @_;
    my $conn = make_conn($host);
    my $ro   = new RaidInfo($conn);

    my %spares = $ro->spare_disks();
    if (keys(%spares) < 2) {
        tstlog(-fail, "There are not enough spares for testing");
        return;
    }

    my @disks_list;

    # create a test volume with 3 disks
    foreach my $disk (keys %spares) {
        if ($#disks_list >= 2) {
            last;
        }
        push(@disks_list, $disk);
    }    #get spares

    disk_suspend(host => $host, disk => $disks_list[0]);
    disk_resume(host => $host, disk => $disks_list[0]);

}    #_test_disk_suspend_resume

# ---------------------------------
# testing the disk_suspend function
# ---------------------------------
sub testme_disk_suspend($$) {
    my ($host, $filermode) = @_;

    _test_disk_suspend_resume($host, $filermode);
}    #testme_disk_suspend

# ---------------------------------
# testing the disk_resume function
# ---------------------------------
sub testme_disk_resume($$) {
    my ($host, $filermode) = @_;

    _test_disk_suspend_resume($host, $filermode);

}    #testme_disk_resume

# ---------------------------------------
# testing the disk_assign and disk remove
# ---------------------------------------
sub testme_disk_remove_ownership($$) {
    my ($host, $filermode) = @_;

    testme_assign_remove($host, $filermode);
}

# ----------------------------
# testing the disk_zero_spares
# ----------------------------
sub testme_disk_zero_spares($$) {
    my ($host, $filermode) = @_;
    _debug("testme_disk_zero_spares: ...");

    my $waittime = 40;
    tstlog(-comment, "zero disks and wait up to $waittime secs");
    eval { disk_zero_spares(host => $host, waittime => $waittime); };
    _verify_result(err => $@, msg => "disk zero spares command");

    eval { disk_zero_spares(host => $host); };
    _verify_result(err => $@, msg => "disk zero spares command and wait until done");
} ## end sub testme_disk_zero_spares($$)

# ------------------------------
# testing the disk_show function
# ------------------------------
sub testme_disk_show($$) {
    _debug("testme_disk_show: ...");
    my ($host, $filermode) = @_;
    my $dh;

    eval { $dh = disk_show(host => $host, options => '-a', filermode => $filermode); };
    if (_verify_result(err => $@, msg => "disk show -a command")) {
        print Dumper($dh);
    }

    eval { $dh = disk_show(host => $host, options => '-n', filermode => $filermode); };
    if (_verify_result(err => $@, msg => "disk show -n command")) {
        print Dumper($dh);
    }

    eval { $dh = disk_show(host => $host, options => '-v', filermode => $filermode); };
    if (_verify_result(err => $@, msg => "disk show -v command")) {
        print Dumper($dh);
    }

    my $so = Sysconfig->new($host->hostname);
    eval { $dh = disk_show(host => $host, sysid => $so->{system_id}, filermode => $filermode); };
    if (_verify_result(err => $@, msg => "disk show sysid <$so->{system_id}>")) {
        print Dumper($dh);
    }

}    #testme_disk_show

# ------------------------------------------
# testing disk simpull is only for simulator
# ------------------------------------------
sub testme_pull_push($$) {
    _debug("Entering testme_disk_simpull");
    my ($host, $filermode) = @_;

    my $conn = make_conn($host, "#rsh");
    if (!RaidUtils_is_sim($conn)) {
        tstlog(-comment, "disk_simpull only runs on simulator");
        return;
    }

    my $ro = new RaidInfo($conn);

    # get a few spares to be used for pull/push
    my %spares = $ro->spare_disks();
    if (keys(%spares) < 3) {
        tstlog(-fail, "There is not enough spares to perform disk pull/push");
        return;
    }

    # Try to pull a disk and then push back
    my $cnt = 0;
    foreach my $disk (keys %spares) {
        eval { disk_simpull(host => $host, disk => $disk); };
        _verify_result(err => $@, msg => "disk simpull <$disk>");

        # get the disk serial numer
        my $diskserialnum = $spares{$disk}->{serialno};
        eval { disk_simpush(host => $host, diskserialnum => $diskserialnum); };
        _verify_result(err => $@, msg => "disk simpush <$diskserialnum>");

        if ($cnt++ > 2) {
            last;
        }
    }    # push/pull each spare

    tstlog(-comment, "*** Testing abnormal disk sim pull/push ***");
    eval { disk_simpull(host => $host, disk => "foo"); };
    _verify_result(err => $@, msg => "disk simpull <foo>", exception => '1');

    eval { disk_simpush(host => $host, diskserialnum => "foo"); };
    _verify_result(err => $@, msg => "disk simpush <foo>", exception => '1');

    return PASS;
}    #testme_pull_push

# ------------------------------------------
# testing disk simpush is only for simulator
# ------------------------------------------
sub testme_disk_simpull($$) {
    my ($host, $filermode) = @_;
    testme_pull_push($host, $filermode);
}    #testme_disk_simpull

# ------------------------------------------
# testing disk simpush is only for simulator
# ------------------------------------------
sub testme_disk_simpush($$) {
    my ($host, $filermode) = @_;
    testme_pull_push($host, $filermode);
}    #testme_disk_simpush

=head1 TESTME

This self test can be used to run all or a few selected define routines.
Make sure to have several spare disks since they will be utilized for
testing disk fail, unfail, remove, sanitize, and so on.

=over

=item Example:

 # to test all the defined routines
 ntest PKGNAME=Storage::Disk SUBNAME=testme FILER=step-u14 util/test_pkg.thpl

 # test a list of routines
 ntest PKGNAME=Storage::Disk FUNCTION_NAMES='disk_fail disk_scrub' SUBNAME=testme FILER=step-u14 util/test_pkg.thpl

=back

=cut

sub testme() {
    testme_driver("Storage::Disk", @EXPORT_OK);
}    #testme

=head2 testme_driver

This routine is written to load and execute selectable testme_<function>
defined within a library. Upon invocation, the routine automatically
loads and executes the testme_<function>. Each test module testme_<function>
should be written to test/verify a specific external <function> in the lib.
The test driver should be named accordingly testme_<function>.

Invocation - refer to the implementation of the testme module in Disk.pm for
invoking the testme_driver using ntest.

=over

=item Arguments

=over

=item lib_name

(Required) : specify the lib full name qualifier. Ex: "Storage::Disk"

=item export_list

(Required) : Specify a list of selected functions to be tested.
This should be passed with the global var EXPORT_OK of the lib.

=back

=item Returns

None

=back

=cut

sub testme_driver($$) {
    my ($lib_name, @export_list) = @_;

    param('FILER', '-mesg', "hostpec of filer");

    #    param('FILERA', '-mesg', "hostpec of filer");
    #    param('FILERB', '-mesg', "hostpec of filer");
    param('FUNCTION_NAMES', '-default', "@EXPORT_OK");
    param('FILERMODE',      '-default', NORMAL_MODE);
    param('LOGDIR',         '-default', './natelogs');

    #    param('ITERATIONS', '-default', '1');
    param('VERBOSE', '-default', '0');

    # define a callback hash table
    my %hcb = ();
    my $fc;
    my $host = host($main::FILER);

    if ($main::FILERMODE eq MAINT_MODE) {
        $host = host($main::FILER);
        $fc   = connect("#console");
        reboot_to($fc, MAINT);
    }

    # --------------------------
    # load the supported routines
    # --------------------------
    foreach my $fn (@export_list) {
        my $tfcn = $lib_name . "::testme_" . $fn;
        $hcb{$fn} = \&$tfcn;
    }

    foreach my $fn (split(/\s\t*/, $main::FUNCTION_NAMES)) {
        _debug("Testing <$fn>");

        if (!defined($hcb{$fn})) {
            tstlog(-fail, "Undefined function name <$fn>");
            next;
        }
        my $fcb = $hcb{$fn};
        &$fcb($host, $main::FILERMODE);
    } ## end foreach my $fn (split(/\s\t*/...

    if ($main::FILERMODE eq MAINT_MODE) {
        my $state = boot_state($fc);
        if ($state eq "UP") {
            tstlog -comment, "Filer has booted successfully";
        }
    }

}    #testme_driver

=head1 STATUS

    TESTME

=head1 ATTRIBUTES

    ONTAP_COMMANDS = disk

=head1 AUTHOR

    Tom Doan (tomdoan@netapp.com)
    $Id: //depot/prod/test/Rsierranevada/lib/Storage/Disk.pm#8 $
    Last revised $Date: 2013/10/04 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;
