# $Id: //depot/prod/test/Rsierranevada/lib/Storage/DiskUtils.pm#4 $
## @summary Library of routines to extract disk information  and inject/check disk errors.
##
## @description
## Library of routines to extract disk information and inject/check disk errors.
##
## @synopsis
##
## use Storage::DiskUtils;
##
## @change 01-26-2004 anishv:add get_disk_struct and get_failed_disks subroutines
## @change 09-29-2004 tomdoan: modified the sanitize_* routines.
##
## @change 10-27-2004 vivienne: add get_disk_list_maint subroutine.
## @change 2-23-2006 erickson: add disk_power_ctl function
## @change 04-07-2009 sreedhar: added led_device method to issue led_on and led_off command on a device
## @change 15-12-2010 faslur: Burt#409611, added the support for secondary path if the system is multipathed in few functions
## @change 09-11-2012 amal: added support for Cmode get_disk_struct and get_failed_disks
## @change 31-01-2013 nandeesh: added support for Cmode in sanitize_startup, sanitize_status and sanitize_monitor
## @change 19-03-2013 villas: added support for Cmode in sub get_disk_hash
## @change 02-0402013 villas: modified sub get_directory_files and rm_directory_files to support cmode
## @author dl-sar-qa@netapp.com, dl-step@netapp.com, dl-storsub-qa@netapp.com
## @pod here

=head1 NAME

Storage::DiskUtils

=head1 SYNOPSIS

        use DiskUtils qw(
        get_drive_sysconfig
        check_scrub_error
                compare_disks
                check_fwfile_exist
                create_vol
                disk_power_ctl
                get_directory_files
                get_disk_list
                get_disk_list_maint
                get_disk_hash
                get_disk_bay
                get_disk_wwn
                get_disk_downrev
                get_disk_version
                get_disk_serial
                get_disk_redundant
                get_disk_poweron
                get_disk_fwfile
                get_disk_raidstatus
                get_disk_model
                get_disk_shelf
                get_disk_glist
                get_disk_struct
                get_failed_disks
                get_filerdisk_model
                get_filer_fwfile
                license_lookup
                led
                rm_directory_files
                sanitize_start
                sanitize_status
                sanitize_abort
                sanitize_release
                sanitize_monitor
                sanitize_startup
                sanitize_response
                sanitize_failure
                vol_stop_scrub
                wait_for_scrub
        led_device
                );

=head1 DESCRIPTION

This is a library module for disk upper layer utilities

=head1 ROUTINES

=cut

package Storage::DiskUtils;

#use strict;
use warnings;

#require Exporter;
#use vars ('@ISA','@EXPORT_OK');
use Tharn;

use filer;
use Aggregate qw(aggr_create aggr_scrub aggr_offline aggr_destroy);
use Storage::AdaptUtils qw(get_adapter_sysconfig);
use Volume;
use Carp qw(confess croak);
use Params::Validate qw(validate OBJECT SCALAR);
use Tharn qw(say script_result host logcomment logresult param);
use Storage::Disk qw(
    disk_assign
    disk_unfail
    disk_remove_ownership
);
use Storage::Misc qw(filer_cmd);
use RaidSqa;
use NACL::APISet;
use NACL::STask::Node;
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}

use Utils qw(make_conn);
use Data::Dumper;

use base qw(Exporter);

#@ISA = ('Exporter');
our @EXPORT_OK = qw(
        get_drive_sysconfig
        check_scrub_error
        compare_disks
        check_fwfile_exist
        create_vol
                get_directory_files
        get_disk_list
        get_disk_list_maint
        get_disk_hash
        get_disk_bay
        get_disk_wwn
        get_disk_downrev
        get_disk_version
        get_disk_serial
        get_disk_redundant
        get_disk_poweron
        get_disk_fwfile
        get_disk_raidstatus
        get_disk_model
        get_disk_shelf
        get_disk_glist
        get_disk_struct
        get_failed_disks
        get_filerdisk_model
        get_filer_fwfile
        license_lookup
        led
                rm_directory_files
             sanitize_start
             sanitize_status
             sanitize_abort
             sanitize_release
             sanitize_monitor
             sanitize_startup
             sanitize_response
             sanitize_failure
        vol_stop_scrub
        wait_for_scrub
        led_device
        );

# library utilities

sub get_loops(%){
    my (%opts) = validate(@_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 },
                                });

        # no zapi so make connection
        my $conn = make_conn($opts{host},$opts{conntype});

        my (undef,undef,$shelf_out) =say($conn,"environ shelf", "-timeout", $opts{timeout});
      chomp( my @shelf_lines = split(/\r?\n/,"$shelf_out"));

    my %loops;
    my $channel;
    my $shelf;
    my $shelf_id;

      foreach my $shelf_line (@shelf_lines) {

            if( $shelf_line =~ /Channel: (\d\[a-h])$/ ){
            $channel = $1;
            $loops{$channel} = {};
            $shelf=0;
        }elsif ($shelf_line =~ /^\s+Shelf: (\d+)$/ ){
            $shelf_id = $1;
            $loops{$channel}{$shelf_id} = {};
            $loops{$channel}{$shelf_id}{shelf_no} = $shelf++;
        }elsif ($shelf_line =~ /SES device path: local access: $channel:(\d+)$/ ){
            $loops{$channel}{$shelf_id}{SES_device} = $1;
        }elsif ($shelf_line =~ /Module type: (\w+);/) {
            $loops{$channel}{$shelf_id}{module_type} = $1;
        }elsif ($shelf_line =~ /product identification=([\w-]+)$/ ){
            $loops{$channel}{$shelf_id}{shelf_type} = $1;
        }elsif ($shelf_line =~ /^\s+Shelf $loops{$channel}[$shelf]{shelf_id}: (.*)$/ ){
            $loops{$channel}{$shelf_id}{disk_ids} = $1;
        }
    }
    return %loops;
}


=head2 disk_power_ctl

        my $ret = disk_power_ctl(host=>$host, disk_id=>$disk_id, action=>$action,
                                                conntype=>$conntype, timeout=>$timeout);
This routine returns 1 on success or croaks on failure.  It calls
    fcadmin_bridge_cmd_cemi for at-fcx and esh4 controllers
and it calls
    sasadmin_dev_power_cycle for shasta

It is not exported, there are three exported library functions that are wrappers,
to do power up, power down and power cycle.

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer.

=item disk_id

(Required) : Device id of disk to be powered

=item action

(Required) : command string to be issued

=item conntype

(Optional) : Connection type for filer. Default : #rsh

=item timeout

(Optional) : Timeout value for the say. Default : 30

=back

=cut

sub disk_power_ctl ( % ) {
        my (%opts) = validate(@_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                disk_id => { type=>SCALAR, regex => qr/^(\w+\.\w+)$/},
                controller =>  { type=>SCALAR, regex => qr/^ESH4|SHASTA/},
                                action => { type=>SCALAR, regex => qr/^(pup|pdown|pcycle)$/},
                                conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 },
                                }
                        );

        # no zapi so make connection
        my $conn = make_conn($opts{host},$opts{conntype});

    my $host = $opts{host};
    my $disk_id = $opts{disk_id};
    my $controller = $opts{controller};
    my $action = $opts{action};
    my $conntype = $opts{conntype};
    my $timeout = $opts{timeout};

    my $ret = 0;
    if ($controller eq 'ESH4') {
        $ret = Storage::Fcadmin::fcadmin_bridge_cmd_cemi(host => $host, disk_id => $disk_id,
                            action => $action, conntype => $conntype,
                            timeout => $timeout);
    }elsif ($controller eq 'SHASTA') {
        if ($action eq 'pcycle') {
                $ret = Storage::Sasadmin::sasadmin_dev_power_cycle(host => $host, dev_id => $disk_id,timeout => $timeout);
        }
    }

        return $ret;
}


## Function to return the output line from sysconfig for a particular drive
##  @name Storage::DiskUtils::get_drive_sysconfig
##  @summary Returns the sysconfig output line for a particuler drive
##  @synopsis
##    get_drive_sysconfig(conn => $conn,
##                          adapter => $adapter,
##                          drive_id => $drive_id,
##                          adapter_sysconfig => $adapter_sysconfig,
##                          timeout => $timeout);
##
##  @arg $conn Required: Filer connection to use
##  @arg $adapter Required: adapter that the drive is accessed by
##  @arg $drive_id Required: drive_id (e.g. 8b.24)
##  @arg $adapter_sysconfig Optional: sysconfig output for the adapter
##  @arg $timeout Optional: timeout for say
##  @status TESTME

sub get_drive_sysconfig(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                adapter => { type=>SCALAR },
                                drive_id => { type=>SCALAR },
                                adapter_sysconfig => { type=>SCALAR, default=>"", optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $drive_id = $opts{drive_id};
    my $adapter = $opts{adapter};
    my $timeout = $opts{timeout};
    my $adapter_sysconfig = $opts{adapter_sysconfig} || get_adapter_sysconfig(conn => $conn, adapter => $adapter, timeout => $timeout);
    chomp( my @adapter_syslines = split(/\r?\n/,"$adapter_sysconfig"));

    my $drive_num = $1 if ($drive_id =~ /\.(\d+)$/);

    foreach my $drive_line (@adapter_syslines){
        return $drive_line if ($drive_line =~ /$drive_id|$drive_num:/);
    }

}


# Function to return an array of disks in the system
##  @name ::get_disk_list
##  @summary Returns an array of all the disks in the filer
##  @synopsis
##    get_disk_list($filer)
##
##  @arg $filer Required: Filer connection to use
##  @status TESTME

sub get_disk_list {
    my ($a , $b , $c, @disk ,  $disk1 , @disk3 ,  $i);
    ($a , $b , $c) = say ($_[0] , "storage show disk -p");
    @disk =  split (/\r*\n/,$c);
    $i = 0;
    foreach $disk1 (@disk) {
        if ( $disk1 =~ /\./) {
            my @disk2 = split(/\s+/,$disk1);
            $disk3[$i] = $disk2[0];
            $i++;
        }
    }
    return(@disk3);
}

# Function to return an array of disks in the system from "disk_list"
##  @name ::get_disk_list_maint
##  @summary Returns an array of all the disks in the filer
##           running in maintenance mode. (console connection)
##  @synopsis
##    get_disk_list_maint($filer)
##
##  @arg $filer Required: Filer connection to use (console connection)
##  @status TESTME

sub get_disk_list_maint {
        my $timeout = 300;
        my @disk_list2 =();
        my ($a, $b, $o, $disk);
    tstlog -comment, "timeout = $timeout \n";
    ($a,$b,$o) = say($_[0], "disk_list", -noprompt, -prompt, "^\\*>", -timeout, $timeout );
    my @disk_list = split(/\r*\n/, $o);
    splice ( @disk_list, 0, 3);
    foreach $disk (@disk_list){
        my  @disk_list1 = split (/\s+/,$disk);
        push @disk_list2 ,$disk_list1[0];
    }
    tstlog -comment, "Disks are : @disk_list2 \n";
    return (@disk_list2);
}

# Function to return a hash of disk information
## @name ::get_disk_hash
## @summary Returns a hash of disk properties.
## @synopsis
##     get_disk_hash($filer,$disk_id)
##
## @arg $filer Required: Filer connection to use
## @arg $disk_id Required: Disk id whose information is required

sub get_disk_hash {
    my ( $i ,  $a , $b , $c , @disk, @disk1 , %disk);
    #Using api set object to support cmode and 7Mode
    my $host = $_[0]->hostp();
    my $mode = '7Mode';
    $mode = 'Nodescope' if ($host->{hosttype} =~ /filer\-ng/);
    my $api_set_obj = NACL::APISet->new(
            hostobj   => $host,
            category  => 'Node',
            interface => 'CLI',
            set       => $mode,
    );
    $c = $api_set_obj->execute_command(command => "storage show disk $_[1]");
    if ($c =~ /does not exist/) {
        tstlog('-fatal' ,"$_[1] doesnot exist.Please check that correct argument is supplied as disk name.");
    }
    @disk = split (/:\s+|\r*\n/,$c);
    @disk1 =  grep /\S+/, @disk;
    %disk = @disk1;
    return(%disk);

}

# Function to return bay id. of a disk
## @name ::get_disk_bay
## @summary Returns bay information of a disk.
## @synopsis
##    get_disk_bay($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required:  Disk id whose information is required

sub get_disk_bay {
    my %info = get_disk_hash($_[0],"$_[1]");
    if ( exists($info{"Bay"})) {
        my $bay = $info{"Bay"};
        return ($bay);
    }
    else {
    tstlog (-comment , "Unable to extract the bay information fo disk $_[1]");
    return ("NOTFOUND");
    }
}


#Function to return downrev status of a disk

## @name ::get_disk_downrev
## @description Returns bay information of a disk.
## @synopsis
##    get_disk_downrev($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required
sub get_disk_downrev {
    my %info = get_disk_hash($_[0],"$_[1]");
     if ( exists($info{"Downrev"})) {
        my $downrev = $info{"Downrev"};
        return($downrev);
    }
    else {
    tstlog (-comment , "Unable to extract the Downrev information for disk $_[1]");
    return ("NOTFOUND");
    }

}

# Function to return disk model


## @name ::get_disk_model
## @summary Returns model information of a disk.
## @synopsis
##    get_disk_model($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required
sub get_disk_model {
    my %info = get_disk_hash($_[0],"$_[1]");
    if ( exists($info{"Model"})) {
        my $model = $info{"Model"};
        return($model);
    }
    else {
        tstlog (-comment , "Unable to extract the model info of disk $_[1]");
         return ("NOTFOUND");
    }

}


# Function to return firmware version of a disk

## @name ::get_disk_version
## @summary Returns fw version of a disk.
## @synopsis
##    get_disk_version($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required

sub get_disk_version {
    my %info = get_disk_hash($_[0],"$_[1]");
    if ( exists($info{"Rev"})) {
        my $version = $info{"Rev"};
        return($version);
    }
    else {
        tstlog (-comment , "Unable to extract the fw rev.  of disk $_[1]");
         return ("NOTFOUND");
    }

}

# Function to return serial no of a disk

## @name ::get_disk_serial
## @summary Returns serial number of a disk.
## @synopsis
##    get_disk_serial($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required
sub get_disk_serial {
    my %info = get_disk_hash($_[0],"$_[1]");
    if ( exists($info{"Serial"})) {
        my $serial = $info{"Serial"};
        if ( $serial =~ /^Vendor$/) { # Unable to obtain serial no.
            $serial = "$_[1]"; # Will use disk_id as serial no.
        }

        return($serial);
    }
    else {
        tstlog (-comment , "Unable to extract the serial no.  of  disk $_[1]");
         return ("NOTFOUND");
    }

}


# Function to return shelf id of a disk

## @name ::get_disk_shelf
## @summary Returns shelf id of a disk.
## @synopsis
##    get_disk_shelf($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required
sub get_disk_shelf {
    my %info = get_disk_hash($_[0],"$_[1]");
    if ( exists($info{"Shelf"})) {
        my $shelf = $info{"Shelf"};
        return($shelf);
    }
    else {
        tstlog (-comment , "Unable to extract the shelf no.  of  disk $_[1]");
         return ("NOTFOUND");
    }

}



# Function to return WWN no. of a disk


## @name ::get_disk_wwn
## @summary Returns WWN number  of a disk.
## @synopsis
##    get_disk_wwn($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required
sub get_disk_wwn {
    my %info = get_disk_hash($_[0],"$_[1]");
    if ( exists($info{"WWN"})) {
        my $wwn = $info{"WWN"};
        return($wwn);
    }
    else {
        tstlog (-comment , "Unable to extract the WWN  of  disk $_[1]");
         return ("NOTFOUND");
    }

}

# Function to return  redundancy state of a disk

## @name ::get_disk_redundant
## @summary Returns redundancy state of a disk.
## @synopsis
##    get_disk_redundant($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required
sub get_disk_redundant {
    my %info = get_disk_hash($_[0],"$_[1]");
    if ( exists($info{"Sec Name"})) {
        my $shelf = $info{"Sec Name"};
        return("REDUNDANT");
    }
    else {
        tstlog (-comment , "disk $_[1] is not redundant");
         return ("NOT REDUNDANT");
    }

}


# Function to return  poweron hours of a disk

## @name ::get_disk_poweron
## @summary Returns poweron hours of a disk.
## @synopsis
##    get_disk_poweron($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required

sub get_disk_poweron {
    my %info = get_disk_hash($_[0],"$_[1]");
    if ( exists($info{"Power-on Hours"})) {
        my $poweron = $info{"Power-on Hours"};
        return($poweron);
    }
    else {
        tstlog (-comment , "Unable to extract the power-on hoursof disk $_[1]");
         return ("NOTFOUND");
    }

}

# Function to return  raidstatus of a disk

## @name ::get_disk_raidstatus
## @summary Returns raid information of a disk.
## @synopsis
##    get_disk_raidstatus($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required
sub get_disk_raidstatus {

    my ( $a , $b , $c);

    ($a , $b , $c) = say ($_[0] , "vol status -r");
    my @vol = split(/\r*\n/,$c);
    my @vol1 = grep /\s+$_[1]\s+/ , @vol;
    my @vol2 = split(/\s+$_[1]\s+/,$vol1[0]);
    return($vol2[0]);

}

# Function to return  disk fw file on a disk

## @name ::get_disk_fwfile
## @summary Returns firmware file name of a disk.
## @synopsis
##    get_disk_fwfile($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required
sub get_disk_fwfile {

    my $model = get_disk_model($_[0],"$_[1]");
    my $version = get_disk_version($_[0],"$_[1]");
    my $filename = "$model"."."."$version"."." . "LOD";
    return($filename);

}

# Function to return  disk fw file on a disk

## @name ::get_disk_glist
## @summary Returns firmware file name of a disk.
## @synopsis
##    get_disk_glist($filer,$disk_id)
## @arg $filer Required: Filer conenction to use
## @arg $disk_id Required: Disk id whose information is required
sub get_disk_glist {

    my %info = get_disk_hash($_[0],"$_[1]");
    if ( exists($info{"Glist entries"})) {
        my $glist = $info{"Glist entries"};
        return($glist);
    }
    else {
        tstlog (-comment , "Unable to extract the power-on hoursof disk $_[1]");
         return ("NOTFOUND");
    }


}

# Function to return disk information data structure.

## @name ::get_disk_struct
## @summary Returns a reference to hashes-of-hashes containing disk related information on a filer
## @description
## Returns a reference to hashes-of-hashes containing attributes of each disk in the system
## Following is a datadumper of a disk from the hash table returned by the subroutine.
## 'raids12:6.28' => {
##                              'IO_TOUT' => '000',
##                              'IN' => '0',
##                              'PriPort' => 'A',
##                              'PowerOnHours' => '15544',
##                              'WWN' => '2:000:0004cf:17851a',
##                              'Blocks read' => '492051',
##                              'Time interval' => '08:21:51',
##                              'ThruPut' => '15218',
##                              'Model' => 'X235_SCHT5073F10',
##                              'Serial' => '3EK0G34T000072031H7A',
##                              'Blocks written' => '856747568',
##                              'CRC' => '0',
##                              'BlkWritten' => '856747568',
##                              'LinkFail' => '0',
##                              'Shelf' => '1',
##                              'Scrub last done' => '00:00:00',
##                              'BlkRead' => '492051',
##                              'Rev' => 'NA06',
##                              'Glist entries' => '0',
##                              'Scrub count' => '143',
##                              'Vendor' => 'NETAPP',
##                              'IO_ABT' => '000',
##                              'IO_RETRY' => '000',
##                              'Sense1' => '0',
##                              'IO_COMP' => '034',
##                              'Sense2' => '0',
##                              'Bay' => '12',
##                              'Sec Port' => 'B',
##                              'Sense3' => '0',
##                              'LossofSync' => '0',
##                              'Sense4' => '0',
##                              'Underrun' => '0',
##                              'SecPort' => 'B',
##                              'Downrev' => 'no',
##                              'Pri Port' => 'A',
##                              'OUT' => '0',
##                              'Power-on Hours' => '15544',
##                              'Glist' => '0',
##                              'SecPath' => 'raids11:6.28',
##                              'Sec Name' => 'raids11:6.28'
##                            },

## @synopsis
##    get_disk_struct($filer)
## @arg $filer Required: Filer connection to use


sub get_disk_struct {

    my $timeout = 300;
    my $disk_struct = {};
    my $disklist;
    my $diskid;
    my @disk;
    my $diskserial;

        my $host_rec = $_[0]->hostp( );

        my $o = filer_cmd(host=>$host_rec, cmd=>"storage show disk -a");

    my @disklist = split(/\r*\n/,$o);
    @disklist = grep /\S+/,@disklist;
    foreach $disklist (@disklist) {
        @disk = split (/:\s+|\r\n/,$disklist);
        if (defined($disk[1])) {$disk[1] =~ s/\s+//g;}
        else {
            #print "Undefined info : $disk[0] \n";
             $disk[1] = "Undefined";}
        if ( $disklist =~ /^Disk:/){
            $diskid = $disk[1];
            $diskid =~ s/\s+//g;
            #print "\n\n\n $diskid \n\n\n";
            goto LABEL;
        }
        if ($disk[0] =~ /Pri Port/){
            $disk_struct->{$diskid}->{PriPort} = $disk[1];
        }
        elsif ($disk[0] =~ /Glist/){
            $disk_struct->{$diskid}->{Glist} = $disk[1];
        }
        elsif ($disk[0] =~ /Power-on/){
            $disk_struct->{$diskid}->{PowerOnHours} = $disk[1];
        }
        elsif ($disk[0] =~ /Blocks read/){
            $disk_struct->{$diskid}->{BlkRead} = $disk[1];
        }
        elsif ($disk[0] =~ /Blocks written/){
            $disk_struct->{$diskid}->{BlkWritten} = $disk[1];
        }
        elsif ($disk[0] =~ /Sec Name/){
            $disk_struct->{$diskid}->{SecPath} = $disk[1];
        }
        elsif ($disk[0] =~ /Sec Port/){
            $disk_struct->{$diskid}->{SecPort} = $disk[1];
        }
        elsif ($disk[0] =~ /Serial/){
            $diskserial = "$disk[1]";
            if($diskserial =~ /Undefined/) {
                $diskserial = $diskid;
            #print "SERIAL: $diskserial \n ";
            }
            $disk_struct->{$diskid}->{Serial} = $diskserial;
        }
        else {
            $disk_struct->{$diskid}->{$disk[0]} = $disk[1];
        }

        LABEL: ;
    }
    my @disks = keys %$disk_struct; # List of disks in the system.
    $disk_struct->{DISK_LIST} = [@disks];
    my $disks;
    my $secpath;
    my @seriallist =();
    my $array;
    my $serial;
    my @disk_paths = ();
    foreach $array (@disks) {
        if ( exists $disk_struct->{$array}->{Serial}){
            $serial = $disk_struct->{$array}->{Serial};
            #print "SERIAL : $serial \n";
            @disk_paths = ();
            push @disk_paths, $array;
            if ( exists $disk_struct->{$array}->{SecPath}) {
                push @disk_paths, $disk_struct->{$array}->{SecPath};
            }

            $disk_struct->{SERIAL_LIST}->{$serial} = [@disk_paths];


        }
    }
        $o = filer_cmd(host=>$host_rec, cmd=>"fcadmin link_stats", timeout => $timeout);

    my @linkstat1 = split(/\r*\n/,$o);
    my $linkstat1;
    my @linkstat2;
    my $i;
    foreach $linkstat1 (@linkstat1) {
        if ($linkstat1 =~ /\./) {
            if ($linkstat1 !~ /\.ha/) {
        #print "$linkstat1 \n";
                @linkstat2 = split(/\s+/,$linkstat1);
                for($i =0; $i <= $#linkstat2 ; $i++) {
                    $linkstat2[$i] =~ s/\s+//g;
                }
                $disk_struct->{$linkstat2[0]}->{LinkFail} = $linkstat2[1];
                $disk_struct->{$linkstat2[0]}->{Underrun} = $linkstat2[2];
                $disk_struct->{$linkstat2[0]}->{LossofSync} = $linkstat2[3];
                $disk_struct->{$linkstat2[0]}->{CRC} = $linkstat2[4];
                $disk_struct->{$linkstat2[0]}->{IN} = $linkstat2[5];
                $disk_struct->{$linkstat2[0]}->{OUT} = $linkstat2[6];
            }
            elsif ($linkstat1 =~ /\.ha/) {
                #print "$linkstat1 \n";
                @linkstat2 = split(/\s+/,$linkstat1);
                for($i =0; $i <= $#linkstat2 ; $i++) {
                $linkstat2[$i] =~ s/\s+//g;
            }
            $linkstat2[0] =~ s/\.ha//g;
            $disk_struct->{$linkstat2[0]}->{LinkFail} = $linkstat2[1];
            $disk_struct->{$linkstat2[0]}->{Underrun} = $linkstat2[2];
            $disk_struct->{$linkstat2[0]}->{LossofSync} = $linkstat2[3];
            $disk_struct->{$linkstat2[0]}->{CRC} = $linkstat2[4];
            $disk_struct->{$linkstat2[0]}->{IN} = $linkstat2[5];
            $disk_struct->{$linkstat2[0]}->{OUT} = $linkstat2[6];
            }
        }
    }
        $o = filer_cmd(host=>$host_rec, cmd=>"priv set -q advanced;disk shm_stats", timeout => $timeout);

    my @shmstat1 = split(/\r*\n/,$o);
    my $shmstat1;
    my @shmstat2;

    foreach $shmstat1 (@shmstat1) {
        if($shmstat1 =~ /\./) {
            @shmstat2 = split(/\s+/,$shmstat1);
            for($i =0; $i <= $#shmstat2 ; $i++) {
                $shmstat2[$i] =~ s/\s+//g;
            }
            $disk_struct->{$shmstat2[0]}->{IO_COMP} = $shmstat2[1];
            $disk_struct->{$shmstat2[0]}->{IO_TOUT} = $shmstat2[2];
            $disk_struct->{$shmstat2[0]}->{IO_RETRY} = $shmstat2[3];
            $disk_struct->{$shmstat2[0]}->{IO_ABT} = $shmstat2[4];
            $disk_struct->{$shmstat2[0]}->{ThruPut} = $shmstat2[5];
            $disk_struct->{$shmstat2[0]}->{Sense1} = $shmstat2[6];
            $disk_struct->{$shmstat2[0]}->{Sense2} = $shmstat2[7];
            $disk_struct->{$shmstat2[0]}->{Sense3} = $shmstat2[8];
            $disk_struct->{$shmstat2[0]}->{Sense4} = $shmstat2[9];
        }

    }

    return ($disk_struct);
}


## @name ::get_failed_disks
## @summary Returns a hash of failed disks with the serial numbers
## @synopsis
##        get_failed_disks($filer);
## @arg $filer Required: Filer connection to use

sub get_failed_disks {
    my $timeout = 300;

        my $host_rec = $_[0]->hostp( );
        my $o = filer_cmd(host=>$host_rec, cmd=>"vol status -f");

    my @failed = split(/\r*\n/,$o);
    my $failed;
    my @failed1;
    my %faileddisks = ();
    my $serial;
    foreach $failed (@failed) {
        if ($failed =~ /\./) {
            @failed1 = split(/\s+/,$failed);
            if ($failed1[0] =~ /failed/) {
                    $serial = get_disk_serial($_[0],$failed1[1]);
                $faileddisks{$failed1[1]} = $serial;
                }
            if (($failed !~ /^admin/) && ($failed1[2] =~ /\./)) {
                $serial = get_disk_serial($_[0],$failed1[2]);
                $faileddisks{$failed1[2]} = $serial;
            }

        }
    }
    return (%faileddisks);
}

## @name ::sanitize_status() -- check the status of the
## sanitized disk and return the status code.
## @usage sanitize_status($filer, $disk)
## @arg $filer Required - Specified the required filer
## @arg $disk Required - Specified the disk name
## Return 1 - if 100% complete.
## Return 0 - if no sanitizing for the specified disk
## Return 2 - if sanitize is still going.
sub sanitize_status($$$) {
    my ($filer, $conn, $disk) = @_;
    # $Fc = connect( $FILER_CONN, -timeout, $TIMEOUT );
    # my $disk1= RaidSqa_find_active_path_disk_name(host =>$conn->{hostp} , disk_name => $disk);
    my $disk1= RaidSqa_find_active_path_disk_name(host =>$filer , disk_name => $disk);
    if($disk1) {
        logcomment("RaidSqa.pm: Using path name $disk1 for any operations on $disk.\n");
        $disk=$disk1;
    } else {
        logcomment("RaidSqa.pm: The disk_name $disk specified does not exist. Please check again.\n");
                return (0);
       }
    my $output = filer_cmd(host=>$filer, conntype =>$conn, cmd=>"disk sanitize status $disk");
    tstlog(-comment, "sanitize_status: <$output> of <$disk>");
    if ($output =~ /100 % complete/) {
    tstlog(-comment, "found 100% complete");
    return 1; #100% complete
    }
    if ($output =~ /% complete/) {
    tstlog(-comment, "sanitize is still going");
    return 2; # not complete
    }

    # No sanitizing disk for specified disk
    tstlog(-comment, $output);
    return 0; #fail or no disk sanitizing
}

## @name ::sanitize_abort()
## @summary Abort the sanitize on the specified disk(s)
## @usage sanitize_abort($filer, $disk)
## @arg $filer Required - Specified the filer name
## @arg $disk Required - Specified the abort disk
## Return 1 -- success
## Return 0 -- failed
sub sanitize_abort($$$) {
    my ($filer, $conn, $disk) = @_;
        # my $disk1= RaidSqa_find_active_path_disk_name(host =>$conn->{hostp} , disk_name => $disk);
        my $disk1= RaidSqa_find_active_path_disk_name(host =>$filer , disk_name => $disk);
        if($disk1) {
                logcomment("RaidSqa.pm: Using path name $disk1 for any operations on $disk.\n");
                $disk=$disk1;
        }else {
                logcomment("RaidSqa.pm: The disk_name $disk specified does not exist. Please check again.\n");
                return (0);
       }

    tstlog (-comment, "executing disk sanitize abort <$disk>");
    my (undef, undef, $output) = say($filer, "priv set -q diag; disk sanitize abort $disk");
    if ($output =~ /abort initiated/) {
    tstlog (-comment, "<$disk> is/are being aborted");
    return 1; #success
    }
    tstlog (-fail, "Unable to activate disk sanitize abort <$disk>");
    return 0; #fail
}

## @name ::sanitize_release()
## @summary Release the specified sanitized disks
## @usage sanitize_release($filer, $disk)
## @arg $filer Required
## @arg $disk Required - Specified the required disk
## Return 1 - success
## Return 0 - fail
sub sanitize_release($$) {
    my ($filer, $disk) = @_;

    foreach my $sdisk ($disk) {
        tstlog (-comment, "release disk <$disk>");
        my (undef, undef, $output) = say($filer, "priv set -q diag; disk sanitize release $disk");
        if ($output =~ /Respared/) {
            tstlog (-comment, "<$disk> has been successfully respared.");
            next;
        }
        elsif ( $output =~ /disk sanitize release: failed for UNKNOWN reason: CR_DISK_ERROR.|disk sanitize release: Disk failed. If disk is unowned, run \'disk assign\' command/ ) {
            # sometimes (GB particularly with sanown) release fails
            # disk becomes unowned, must be assigned and unfailed.
            # we should issue a WARN result here but allow the test to succeed
            logresult( 'WARN', "disk sanitize release appears to have failed, trying to assign disk" );
            disk_assign(
                host => $filer->hostp(),
                disk => $sdisk,
            )
                or logresult( 'FATAL', "Failed to assign disk" );

            my $status = get_disk_raidstatus( $filer, $sdisk );

            if ($status eq '') {
                disk_remove_ownership (
                    host => $filer->hostp(),
                    disk => $sdisk,
                )
                    or logresult( 'FATAL', "Failed to remove ownership" );

                disk_assign(
                    host => $filer->hostp(),
                    disk => $sdisk,
                )
                    or logresult( 'FATAL', "Failed to assign disk" );

            }

            $status = get_disk_raidstatus( $filer, $sdisk );

            # check disk status if it appears owned but doesn't appear in "vol
            # status -r" or "vol status -f" then we need to "disk
            # remove_ownership -f" and "disk assign" again.

            # once it appears in "vol status -f" then we can unfail the disk
            if ($status =~ /bad label/) {
                disk_unfail(
                    host => $filer->hostp(),
                    disk => $sdisk,
                )
                    or logresult( 'FATAL', "Failed to unfail disk" );
                next;
            }
            else {
                # second attempt at restoring disk status seems to have
                # failed.
                logresult( 'FATAL', "Failed to fix disk ownership" );
            }
        }
=head
     elsif ( $output =~ /disk sanitize release: Disk failed. If disk is unowned, run \'disk assign\' command/ ){
      # looks like the disk will be going to broken pool, assign the ownership if it is unowned and then unfail the disk
            logresult( 'WARN', "disk sanitize release appears to have sent the disk to the failed pool, verify whether the disk is in broken pool or not" );
        tstlog (-comment, "Wait for 60 second and then check the status");
        sleep 60; #let the disk to go to failed pool

        my $status = get_disk_raidstatus( $filer, $sdisk );
            # once it appears in "vol status -f" then we can unfail the disk
            if ($status =~ /bad label/) {
                # looks like the disk went to broken pool, unfail the disk
            logresult( 'WARN', "disk sanitize release sent the disk to the failed pool, unfail the disk to sent it the spare pool" );
                disk_unfail(
                    host => $filer->hostp(),
                    disk => $sdisk,
                )
                    or logresult( 'FATAL', "Failed to unfail disk" );
             tstlog (-comment, "<$disk> is unfailed");
            }

     else {
                # The disk is not found in the broken pool, something went wrong.
        # Display the vol status -r output so that the user can easily debug
        tstlog (-comment, "The disk is not found in the broken pool, display the vol status -r output so that the user can easily debug");
        get_disk_raidstatus( $filer);

                logresult( 'FATAL', "The disk is not found in the broken pool, something went wrong, see the above vol status -r output" );
            }
}
=cut
        else {
            tstlog (-comment, "output: $output");
            tstlog (-fatal, "Unable to release disk <$disk>");
            return 0; #fail
        }
    }
    tstlog (-comment, "Successfully released all disks <$disk>");
    return 1; #success
}

## @name ::led() - do an led_on_all or led_off_all
## @summary do an led_on_all or led_off_all on $adapter
## @usage led($filer, $adapter, $action)
## @arg $filer Required
## @arg $adapter Required: channel on which to execute led_off_all or led_on_all
## @arg $action Required: on | off
## Return:
##      1 (success) if command succeeds
##      0 (failed)
##

sub led($$$) {
    my($filer, $adapter, $action) = @_;
    my($output, $ret);

    $ret = 1;
    if ($action eq 'on') {
        $output = filer_cmd(host=>$filer->hostp(), cmd=>"led_on_all $adapter");
    } elsif ($action eq 'off') {
        $output = filer_cmd(host=>$filer->hostp(), cmd=>"led_off_all $adapter");
    } else {
        # Invalid action parameter given
        $ret = 0;
    }

    if ($output && ( $output =~ /invalid channel:/ ||
        $output =~ /Usage: /)) {
        $ret = 0;
    }

    return $ret;
}

## @name ::led_device() - do a led_on or led_off on device or led_on_off
## @summary do an led_on or led_off or led_on_off  on $device
## @usage led($filer, $device, $action)
## @arg $filer Required
## @arg $device Required: device on which to execute led_on or led_off or led_on_off
## @arg $action Required: on | off | blink
## Return:
##      1 (success) if command succeeds
##      0 (failed)
##

sub led_device($$$) {
    my($filer, $device, $action) = @_;
    my($output, $ret);

    $ret = 1;
    if ($action eq 'on') {
        $output = filer_cmd(host=>$filer->hostp(), cmd=>"led_on $device");
      } elsif ($action eq 'off') {
        $output = filer_cmd(host=>$filer->hostp(), cmd=>"led_on $device");

    } elsif ($action eq 'blink'){
        $output = filer_cmd(host=>$filer->hostp(), cmd=>"led_on $device");
    } else {
        $ret = 0;
    }

    if ($output && ( $output =~ /invalid disk name:/ ||
        $output =~ /Usage: /)) {
        $ret = 0;
    }

    return $ret;
}

## @name ::license_lookup() - lookup for the specified license
## @summary lookup for the specified license for authorization
## @usage license_lookup($filer, $lincese_key)
## @arg $filer Required
## @arg $license_key Required: "disk_sanitization" or "cluster" or ...
## Return:
##      1 (success) if the specified license is authorized
##      0 (failed)
##

sub license_lookup($$) {
    my($filer, $license_key) = @_;
    my(@info);

    my (undef, undef, $output) = say($filer, "license");

    #
    # identify the EOL characters before splitting the output
    #
    if ($output =~ /\r\n/) {
    @info = split(/\r\n/, $output);
    } else  {
    @info = split(/\n/, $output);
    }
    tstlog(-comment, ">>licenses: @info");
    foreach my $line (@info) {
        tstlog(-comment, "LINE:" . $line);
        if ($line =~ /$license_key/) {
            if ($line =~ /not licensed/) {
                return 0;
            } else {
                return 1;
            }
        }
    }
    return 0;
} #license_lookup

## @name ::sanitize_start()
## @summary Sanitize the specified disks with the specified options.
## Wait until the sanitize process to start within the specified
## amount of time.
## @usage sanitize_start($filer, $options, $disk, @wait_time)
## @arg $filer Required:
## @arg $options Required: "disk sanitize start -f -s -..."
## @arg $disk Required: one or more disks to be santized.
## @arg $wait_time Optional: how long to wait for the sanitize to start.
##   default: is 60 seconds
## Return:
##   1 - success
##   0 - fail

sub sanitize_start($$$$$) {
    my ($filer, $conn, $options, $disk, $wait_time) = @_;
        # my $disk1= RaidSqa_find_active_path_disk_name(host =>$conn->{hostp} , disk_name => $disk);
        my $disk1= RaidSqa_find_active_path_disk_name(host =>$filer , disk_name => $disk);
        if($disk1) {
                logcomment("RaidSqa.pm: Using path name $disk1 for any operations on $disk.\n");
                $disk=$disk1;
        }else {
                logcomment("RaidSqa.pm: The disk_name $disk specified does not exist. Please check again.\n");
                return (0);
       }

    my $priv = "priv set -q diag; ";
    my ($sanitize_start, undef, $output) = say($filer, $priv . $options . " " . $disk);

    # Check for a valid disk sanitize license
    #
    if ($output =~ /not licensed/) {
    tstlog (-fatal, "No license to run disk sanitize");
    }

    if ($output !~ /been initiated/) {
    tstlog (-comment, "status=$output");
    tstlog (-comment, "Failed to start disk sanitize for <$disk>");
    return 0; #fail
    }

    logcomment("$sanitize_start");
    my $stime = 0;
    my $cmd = "priv set -q diag; disk sanitize status " . $disk;
    while ( $stime < $wait_time) {
    my (undef, undef, $output) = say ($filer, $cmd);
    if ($output =~ /\% complete/) {
        last;
    }
    $stime++;
    sleep(2);
    }

    if (length($wait_time) <=0) {
    $wait_time = 60;
    }

    if ($stime > $wait_time) {
    tstlog -fatal, "Unable to start disk sanitize for <$disk> after <$wait_time> seconds";
    return 0; #fail
    }
    return 1; #success
} #sanitize_start


## @name ::sanitize_monitor
## @summary Monitor the progress of your sanitizations
## @usage sanitize_monitor($filer)
## @arg $filer Required
sub sanitize_monitor{
    logcomment ("Reading the sanitized_disks file from the filer");
    my $output = filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"priv set -q diag;rdfile /etc/sanitized_disks");
    if($output =~ /Can't open directory|No such file or directory/) {
        logcomment("Check in the directory /etc/log/ ");
        filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"priv set -q diag;rdfile /etc/log/sanitized_disks");
    }
    logcomment ("Reading the sanitization log");
    $output = filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"priv set -q diag;rdfile /etc/sanitization.log");
    if($output =~ /Can't open directory|No such file or directory/) {
        logcomment("Check in the directory /etc/log/ ");
        filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"priv set -q diag;rdfile /etc/log/sanitization.log");
    }
}

## @name ::sanitize_startup
## @summary Set the system up for sanitization
## @usage sanitize_startup($filer)
sub sanitize_startup{
    logcomment ("Cleaning up the log files before starting the test.");
    logcomment ("Files to be cleaned up are sanitized_disks and sanitization.log");
    sanitize_monitor($_[0],$_[1]);
    logcomment ("emptying the log files");
    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"rm /etc/sanitization.log");
    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"rdfile /etc/sanitization.log");

    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"rm /etc/log/sanitization.log");
    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"rdfile /etc/log/sanitization.log");

    logcomment ("Sanitization log file empty");

    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"rm /etc/sanitized_disks");
    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"rdfile /etc/sanitized_disks");

    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"rm /etc/log/sanitized_disks");
    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"rdfile /etc/log/sanitized_disks");

    logcomment ("Sanitized_disks file empty");
    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"mkfile 0 /etc/sanitization.log");
    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"mkfile 0 /etc/sanitized_disks");
    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"mkfile 0 /etc/log/sanitization.log");
    filer_cmd(host=>$_[0], conntype =>$_[1], cmd=>"mkfile 0 /etc/log/sanitized_disks");
    logcomment ("Ready to test");
}

## @name ::sanitize_response
## @summary wait for sanitization to complete
## @usage sanitize_response($filer,$spares)
sub sanitize_response{
    logcomment ("Waiting for Sanitization of $_[1] to complete");
# Check that the format phase completes
    #my ($format_complete,undef,undef) = say($_[0],"",-noprompt,-timeout,14400,"-wait","format complete, begin pattern writes", -fatal, "Format never completed for $_[1] within four hours!");
    #logcomment ("$format_complete");
# Check the pattern write phase completes and the drive is sanitized
    my ($response,undef,undef) = say ($_[0],"", -noprompt,-timeout,14400,"-wait","Disk sanitization on $_[1] .+ has completed successfully", -fatal, "Pattern writes never completed within 4 hours on $_[1], please test on smaller faster drives if possible.Or file a burt against the owner to extend the duration for pattern writes" );
    logcomment ("$response");
    logcomment ("$_[1] will now be released to the spares pool");
    &sanitize_release($_[0],$_[1]);
}
## @name ::sanitize_failure
## @summary checks that the disk_sanitize code behaves correctly after induced sanitization failure.
## @usage sanitize_failure($filer,$spares)

sub sanitize_failure{
    logcomment ("Checking that the system is behaving correctly after the sanitization failure");
    my ($sanitize_restart,undef,undef) = say($_[0],"",-noprompt,-timeout,1800,"-wait","shm: Disk $_[1] is being reformatted due to an error during disk sanitize", -fatal, "Format never restarted for $_[1] within 30 minutes!");
    logcomment ("$sanitize_restart");
    my ($format_complete,undef,undef) = say($_[0],"",-noprompt,-timeout, 3600, "-wait","shm: Disk$_[1] has been successfully reformatted.  To sanitize disk, restart the disk sanitize command", -fatal, "Format never completed for $_[1] within one hour!");
    logcomment ("$format_complete");
    logcomment ("Reformatted drive after returned to spares pool");
}

## @name :: vol_stop_scrub
## @summary  stop all the aggregates scrub
sub vol_stop_scrub {
    my ($conn) = @_;
    printf "Now the connection is $conn.\n";
    my $cmd_scrubstop = "aggr scrub stop";
    filer_cli_advanced_say($conn, $cmd_scrubstop);
    sleep 30;
    my $cmd_scrubstatus = "aggr scrub status";
    my ($output, $text) = filer_cli_advanced_say($conn, $cmd_scrubstatus);
    if ($output =~ /no scrubs in progress/) {
    logcomment ("There is no scrub running on volumes.");
    } else {
    logresult ("FATAL", "vol scrub stop is not working.");
    }

}

## @name :: wait_for_scrub
## @summary  wait for scrub to be ended
sub wait_for_scrub {
    my($conn) = @_;
    my $cmd_scrubstatus = "aggr scrub status";
    my ($output, $text) = filer_cli_advanced_say($conn, $cmd_scrubstatus);
    my $state_flag = 0;
    while(!$state_flag) {
    tstlog -comment, "vol scrub  is still going on.\n";
    sleep 120;
    my ($output, $text) = filer_cli_advanced_say($conn, $cmd_scrubstatus);
    if ($output =~ /no scrubs in progress/) {
        $state_flag = 1;
        logcomment ("There is no scrub running on volumes.");
    } else {
        logcomment ("aggr scrub is still going on.");
    }
    }
}

## @name :: get_filerdisk_model
## @summary  Get all the disk model on the filer
sub get_filerdisk_model(%) {
    my(%disks) = @_;
    my %disk_model;
    foreach my $disk (keys %disks) {
    my $diskname = $disks{$disk}->name();
    my $model = $disks{$disk}->parameter("model");
    $model =~ s/\s+//;    # remove the leading/trailing space
    push(@{$disk_model{$model}}, $diskname);
    }
    return (%disk_model);
}

## @name :: get_filer_fwfile
## @summary  Get all firmware file on the filer
sub get_filer_fwfile($) {
    my ($rsh_conn) = @_;
    my $cmd_ls = "ls /etc/disk_fw";
    my ($after, $result) = filer_cli_advanced_say($rsh_conn, $cmd_ls);
    my @ls_lines = split(/\s\n+/, $after);
    my %fw_files;
    my ($fwmodel, $na, $rev, $file);
    foreach my $line (@ls_lines) {
    if ($line=~ /([^\.]+)\.([^\.]{2})([^\.]{2})\.LOD/){
        my ($fwmodel, $na, $rev) = ($1, $2, $3);
        $file = "$fwmodel\.$na" . sprintf($rev) . "\.LOD";
        $fw_files{$fwmodel} = [$file, $na, $rev];
    }
    }
    return (%fw_files);
}

## @name :: check_scrub_error
## @summary  check error after scrub
sub check_scrub_error {
    my ($rsh_conn) = @_;
    my $rscrub = new RaidInfo($rsh_conn);
    my %groups = $rscrub->all_raid_groups();
    my $group_name;
    foreach $group_name (keys(%groups)) {
    if ($groups{$group_name}{scrub_cksum_err} != 0 ) {
        logresult ("FATAL", "scrub find checksum errors in $group_name, test fail.");
    }
    if ($groups{$group_name}{scrub_pi} != 0 ) {
        logresult ("FATAL", "scrub find pi errors in $group_name, test fail.");
    }
    if ($groups{$group_name}{scrub_media_err} != 0 ) {
        logresult ("FATAL", "scrub find media error in $group_name, test fail.");
    }
    }
}

## @name :: compare_disks
## @summary  compare the disks parameters before and after disk_fw_update
sub compare_disks {
    my ($r_before, $r_after) = @_;
    my %disk_before = $r_before->all_disks();
    my %disk_after = $r_after->all_disks();
    foreach my $key (keys (%disk_before)){
    my $diskname = $disk_before{$key}->name();
    #Get the disk parameter before disk firmware update
    my $model_before = $disk_before{$key}->parameter("model");
    my $shelf_before = $disk_before{$key}->parameter("shelf");
    my $bay_before = $disk_before{$key}->parameter("bay");
    my $rpm_before = $disk_before{$key}->parameter("rpm");
    my $serial_before = $disk_before{$key}->parameter("serialno");
    my $vendor_before = $disk_before{$key}->parameter("vendor");
    my $wwn_before = $disk_before{$key}->parameter("wwn") || "";
    my $size_before = $disk_before{$key}->parameter("size");
    #Get the disk parameter after disk firmware update
    if (! $disk_after{$key}){
        logresult("FATAL", "$diskname (key = $key, shelf = $shelf_before, bay = $bay_before) does not exist after firmware update.");
    }
    my $model_after = $disk_after{$key}->parameter("model");
    my $shelf_after = $disk_after{$key}->parameter("shelf");
    my $bay_after = $disk_after{$key}->parameter("bay");
    my $rpm_after = $disk_after{$key}->parameter("rpm");
    my $serial_after = $disk_after{$key}->parameter("serialno");
    my $vendor_after = $disk_after{$key}->parameter("vendor");
    my $wwn_after = $disk_after{$key}->parameter("wwn") || "";
    my $size_after = $disk_after{$key}->parameter("size");
    if ($model_before ne $model_after) {
        logresult("FATAL", "$diskname has model $model_before before firmware update and has $model_after after firmware update.");
    }
    if ($shelf_before ne $shelf_after) {
        logresult("FATAL", "$diskname has shelf $shelf_before before firmware update and has $shelf_after after firmware update.")
    }
    if ($bay_before ne $bay_after) {
        logresult("FATAL", "$diskname has bay $bay_before before firmware update and has $bay_after after firmware update.");
    }
    if ($rpm_before ne $rpm_after) {
        logresult("FATAL", "$diskname has rpm $rpm_before before firmware update and has $rpm_after after firmware update.");
    }
    if ($serial_before ne $serial_after) {
        logresult("FATAL", "$diskname has serial $serial_before before firmware update and has serial $serial_after after firmware update.");
    }
    if ($vendor_before ne $vendor_after) {
        logresult("FATAL", "$diskname has vendor $vendor_before before firmware update and has vendor $vendor_after after firmware update.");
    }
    if ("$wwn_before" ne "$wwn_after") {
        logresult("FATAL", "$diskname has wwn $wwn_before before firmware update and has wwn $wwn_after after firmware update.");
    }
    if ($size_before ne $size_after) {
        logresult("FATAL", "$diskname has rpm $size_before before firmware update and has $size_after after firmware update.");
    }

    }
}

## @name :: check_fwfile_exist
## @summary  check whether firmware file exist
sub check_fwfile_exist {
    my ($Firmware, %fwfile) = @_;
    my $flag = 0;
    foreach my $key (keys (%fwfile)){
    my $fw_ref = $fwfile{$key};
    my ($filename, $na, $rev) = @$fw_ref;
    if ($filename eq $Firmware) {
        printf "Now we are matching.\n";
        $flag = 1;
    }
    }
    return ($flag);
}

## @name :: create_vol
## @summary  create volume according to requirement
sub create_vol{
    my ($rsh_conn, $VOL_CREATE) = @_;
    my ($FILER_VOL_raid4non, $FILER_VOL_dp, $FILER_VOL_raid4, $FILER_VOL_raiddp);
    my @volname_array = ();
    # Create volume according to the requirements
    if ($VOL_CREATE eq "all") {
    #create non-mirrored raid4 volume
    my $num_raid4_non=3;
    my (%option1, %option2, %option_raid4, %option_raid4non);
    $option_raid4non{raidtype} = 'raid4';
    $option_raid4non{mirror} = 0;
    $FILER_VOL_raid4non = "testvol_raid4non";
    my $rv_raid4non = aggr_create($rsh_conn,$FILER_VOL_raid4non,$num_raid4_non, %option_raid4non);
    if(!$rv_raid4non){
        logresult("FATAL", "aggr create non-mirrored raid4 volume fail for VOL_CREATE equal to all.");
    }
    push (@volname_array, $FILER_VOL_raid4non);
    #create non-mirrored raiddp volume
    my $num_dp = 3;
    $FILER_VOL_dp = "testvol_dp";
    $option1{raidtype} = 'raid_dp';
    $option1{mirror} = 0;
    my $rv1 = aggr_create($rsh_conn,$FILER_VOL_dp,$num_dp, %option1);
    if(!$rv1){
        logresult("FATAL", "aggr create non-mirrored raiddp volume fail for VOL_CREATE equal to all.");
    }
    #create mirrored raid4 volume
    my $num_raid4 = 6;
    $FILER_VOL_raid4 = "testvol_raid4";
    $option_raid4{raidtype} = 'raid4';
    $option_raid4{mirror} = 1;
    my $rv_raid4 = aggr_create($rsh_conn,$FILER_VOL_raid4,$num_raid4, %option_raid4);
    if (!$rv_raid4){
        logresult("FATAL", "aggr create mirrored raid4 volume fail for VOL_CREATE equal to all. Check whether you have mirrored license or enough spares.");
    }
    push (@volname_array, $FILER_VOL_raid4);
    #create mirrored raiddp volume
    my $num_raiddp = 6;
    $FILER_VOL_raiddp = "testvol_raiddp";
    $option2{raidtype} = 'raid_dp';
    $option2{mirror} = 1;
    my $rv2 = aggr_create($rsh_conn,$FILER_VOL_raiddp,$num_raiddp, %option2);
    if (!$rv2){
        logresult("FATAL", "aggr create mirrored raiddp volume fail for VOL_CREATE equal to all. Check whether you have mirrored license or enough spares.");
    }
    push (@volname_array, $FILER_VOL_raiddp);
    } elsif ($VOL_CREATE eq "NONE") {
    printf ("=============Now I am in the $VOL_CREATE eq to none.\n=========");
    logcomment ("No need to creat new volume. Just use current volume type.");
    }
    return (@volname_array);
}

## Function to return file names under the specified directory
##  @name    Storage::DiskUtils::get_directory_files
##  @summary Returns an array reference with the file names under the
##           specified directory. Croaks on error.
##  @synopsis
##      get_directory_files(fc   => $fc,
##                          path => $path);
##
##  @arg $fc   Required: Filer connection to use
##  @arg $path Required: Path to the directory

sub get_directory_files(%;) {
    my (%opts) = validate(@_,
                         {fc   => {type => OBJECT},
                          path => {type => SCALAR},
                          api_set_obj => {
                                   type     => OBJECT,
                                   optional => YES,
                                 },
                     });

    my ($file, $i, $output, @names);
    my $apiset_obj = $opts{api_set_obj};
    if(!$apiset_obj){
        my $filer_node = NACL::STask::Node->new( hostrec => $opts{fc}->hostp() );
        $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
    }
    $output = $apiset_obj->execute_command(command => "priv set -q advanced;ls $opts{path}");
    if ($output =~ /Can't open directory/i) {
        croak "Cannot open directory $opts{path} on filer";
    }
    @names = split(/\n/, $output);
    # Remove trailing white space on each filename
    for ($i = 0; $i < scalar(@names); $i++) {
        $names[$i] =~ s/\s+$//;
    }

    return \@names;
}

## Function to remove all files under the specified directory
##  @name    Storage::DiskUtils::rm_directory_files
##  @summary Removes all files under the specified directory
##           Croaks on error.
##  @synopsis
##      rm_directory_files(fc   => $fc,
##                         path => $path);
##
##  @arg $fc   Required: Filer connection to use
##  @arg $path Required: Path to the directory

sub rm_directory_files(%;) {
    my (%opts) = validate(@_,
                         {fc   => {type => OBJECT},
                          path => {type => SCALAR},
                          api_set_obj => {
                          type     => OBJECT,
                          optional => YES,
                          },
                       });

    my ($output, $file, $files);
    my $apiset_obj = $opts{api_set_obj};
    if(!$apiset_obj){
        my $filer_node = NACL::STask::Node->new( hostrec => $opts{fc}->hostp() );
        $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
    }
    $files = get_directory_files(fc => $opts{fc}, path => $opts{path}, api_set_obj => $apiset_obj);
    foreach $file (@{$files}) {
        next if (($file eq '.') || ($file eq '..'));
#        $output = filer_cli_advanced_say($opts{fc}, "rm $opts{path}/$file");
        $output = $apiset_obj->execute_command(command => "priv set -q advanced; rm $opts{path}/$file");
        if ($output =~ /(failed|Can not)/i) {
            croak "Error removing file $opts{path}/$file on filer: $output";
        }
    }

    return;
}

1;
