#
# $Id: //depot/prod/test/Rsierranevada/lib/Storage/FDE/DiskEncrypt.pm#1 $
#

## @summary DiskEncrypt routine
##
## @description
## This module provides all the capabilities
## required for FDE DiskEncrypt work flow.
##
## @synopsis
## use strict;
## use warnings;
## use Storage::FDE::DiskEncrypt qw(fde_lock fde_unlock fde_rekey fde_show verify_locked print_key power_cycle_disk);
## param('FILER', -mesg, "hostspec of FILER");
## my $host_obj = host($FILER);
## fde_lock(host=> $host_obj); 
## fde_unlock(host=> $host_obj);
## fde_rekey(host=> $host_obj, key => "DEADBEEFDEADBEEFDEADBEEFDEADBEEF");
## fde_show(host=> $host_obj);
## verify_locked(host=> $host_obj);
## print_key(host=> $host_obj);
## power_cycle_disk(host=> $host_obj);

## @author niveetas@netapp.com
## @status private (in-progress)

package Storage::FDE::DiskEncrypt;
use strict;
use warnings;
use Storage::Disk;
use filer;
use Tharn qw(say logcomment logresult);
use Utils qw(make_conn);
use Data::Dumper;
use base qw(Exporter);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use vars;
use EMS;
use Storage::Disk qw(disk_show);
use Storage::Ems qw(ems_log_get_mark ems_log_dump_by_type ems_log_check ems_error_monitor ems_log_dump);


BEGIN{
    use base qw(Exporter);
    our @EXPORT_OK = qw(
        fde_lock
        fde_unlock
        fde_rekey
        fde_show
        verify_locked
        print_key
        power_cycle_disk
    );
};

# ---------------
# define constants 
# ---------------
use constant SAY_TIMEOUT => 300; 
use constant PASS => 1; 
use constant FAIL => 0; 
use constant NOOP => 2; 
use constant MAINT_PROMPT => "\\*\>"; 
use constant WAIT_UNTIL_DONE => -1; # wait until done
use constant NORMAL_MODE => "normal"; 
use constant MAINT_MODE => "maint"; 
use constant YES => 1; 
use constant NO => 0; 
use constant VOLNAME => "foo"; 

# ======================= DEFINE ROUTINES ===========================


## @name ::fde_lock
## @summary runs the CLI disk encrypt lock command
## 
## @description
## Reports the expected output of the command
##
## @return output 
## 
## @arg host Required: Hostrec object
##
## @arg disk Optional: Takes the disk name provided.
## If not provided, considers all the disk in the filer
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## fde_lock(host=> $host_obj, disk => "0c.00.8");
## fde_lock(host=> $host_obj, disk => "*");
## fde_lock(host=> $host_obj);
##
## @author niveetas@netapp.com
## @status private (in-progress)
##

sub fde_lock {
my (%pars) = validate(@_, {
    host => {  type=>OBJECT, isa=>'Hostrec'}, 
    disk => { type => SCALAR, optional => YES},
    conntype => { type => SCALAR, optional => YES,
        regex => qr/\#rsh|\#telnet|\#console/, default => "#rsh"}, 
    timeout => {type => SCALAR, optional => YES,
        regex => qr/\d+/, default => SAY_TIMEOUT },
    waittime => { type => SCALAR, optional => YES,
        regex => qr/\d+/, default => WAIT_UNTIL_DONE},
    }
        );

    my $conn = make_conn($pars{host}, $pars{conntype});
    my $cmd = "disk encrypt lock";
    my $disklist_given = $pars{disk};
    if (!defined ($disklist_given)){
        $cmd = $cmd . " ". "*";
    } else {
        $cmd = $cmd . " ". $disklist_given;
    }
    my (undef, undef, $output) = say($conn, $cmd , '-timeout', $pars{timeout});
    if (($output =~ m/successful unlock/)||($output =~ m/0 failures/)) {
        logcomment ("DiskEncrypt:fde_lock: Disk Locked");
    }else {
        logresult ("Fail", "Disk: $disklist_given does not exist");
    }
}

## @name ::fde_unlock
## @summary runs the CLI disk encrypt unlock command
## 
## @description
## Reports the expected output of the command
##
## @return output 
## 
## @arg host Required: Hostrec object
##
## @arg disk Optional: Takes the disk name provided.
## If not provided, considers all the disk in the filer
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## fde_unlock(host=> $host_obj, disk => "0c.00.8");
## fde_unlock(host=> $host_obj, disk => "*");
## fde_unlock(host=> $host_obj);
##
## @author niveetas@netapp.com
## @status private (in-progress)
##

sub fde_unlock {
my (%pars) = validate(@_, {
    host => {  type=>OBJECT, isa=>'Hostrec'}, 
    disk => { type => SCALAR, optional => YES},
    conntype => { type => SCALAR, optional => YES,
        regex => qr/\#rsh|\#telnet|\#console/, default => "#rsh"}, 
    timeout => {type => SCALAR, optional => YES,
        regex => qr/\d+/, default => SAY_TIMEOUT },
    waittime => { type => SCALAR, optional => YES,
        regex => qr/\d+/, default => WAIT_UNTIL_DONE},
    }
        );
    my $conn = make_conn($pars{host}, $pars{conntype});
    my $cmd = "disk encrypt unlock";
    my $disklist_given = $pars{disk};
    if (!defined ($disklist_given)){
        $cmd = $cmd . " ". "*";
    } else {
        $cmd = $cmd . " ". $disklist_given;
    }
    my (undef, undef, $output) = say($conn, $cmd , '-timeout', $pars{timeout});
    if ($output =~ m/successful unlock/) {
        logcomment ("DiskEncrypt:fde_unlock: Disk unlocked");
    }else {
        logresult ("Fail", "Disk: $disklist_given does not exist");
    }
}

## @name ::fde_rekey
## @summary runs the CLI disk encrypt rekey command
## 
## @description
## Reports the expected output of the command
##
## @return output 
## 
## @arg host Required: Hostrec object
##
## @arg disk Optional: Takes the disk name provided.
## If not provided, considers all the disk in the filer
##
## @arg key Required: Must be 32 hex characters or special value '0x0'
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## fde_rekey(host=> $host_obj, key => "DEADBEEFDEADBEEFDEADBEEFDEADBEEF");
## fde_rekey(host=> $host_obj, key => "DEADBEEFDEADBEEFDEADBEEFDEADBEEF", disk => '*');
## fde_rekey(host=> $host_obj, key => "DEADBEEFDEADBEEFDEADBEEFDEADBEEF", disk => "0c.00.9");
##
## @author niveetas@netapp.com
## @status private (in-progress)
##

sub fde_rekey {
my (%pars) = validate(@_, {
    host => {  type=>OBJECT, isa=>'Hostrec'}, 
    disk => { type => SCALAR, optional => YES},
    key => {type => SCALAR },
    conntype => { type => SCALAR, optional => YES,
        regex => qr/\#rsh|\#telnet|\#console/, default => "#rsh"}, 
    timeout => {type => SCALAR, optional => YES,
        regex => qr/\d+/, default => SAY_TIMEOUT },
    waittime => { type => SCALAR, optional => YES,
        regex => qr/\d+/, default => WAIT_UNTIL_DONE},
    }
        );

    my $conn = make_conn($pars{host}, $pars{conntype});
    my $key = $pars{key};
    my $cmd = "disk encrypt rekey $key";
    my $disklist_given = $pars{disk};
    if (!defined ($disklist_given)){
        $cmd = $cmd . " ". "*";
    } else {
        $cmd = $cmd . " ". $disklist_given;
    }
    my (undef, undef, $output) = say($conn, $cmd , '-timeout', $pars{timeout});
    if (($output =~ m/successful unlock/)||($output =~ m/0 failures/)) {
        logcomment ("DiskEncrypt:fde_rekey: Disk rekeyed");
    } elsif($output =~ /new_key_id must be/) { #TODO check for an invalid key_id, verify the output
        logresult ("Fail", "Disk: Key $key is invalid");
    } else {
        logresult ("Fail", "Disk: $disklist_given does not exist");
    }
}


## @name ::fde_show
## @summary runs the CLI disk encrypt show command
## 
## @description
## Reports the expected output of the command
##
## @return output 
## 
## @arg host Required: Hostrec object
##
## @arg disk Optional: Takes the disk name provided.
## If not provided, considers all the disk in the filer
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## fde_show(host=>$host_obj, disk => "0c.00.*"); 
## fde_show(host=>$host_obj); for all the disk in the filer
## fde_show(host=> $host_obj, disk => "*"); for all the disk in the filer
##
##
## @author niveetas@netapp.com
## @status private (in-progress)
##

sub fde_show  {
    my (%pars) = validate(@_, {
    host => { type=>OBJECT, isa=>'Hostrec'}, 
    disk => { type => SCALAR|UNDEF, optional => YES},
    conntype => { type => SCALAR, optional => YES,
        regex => qr/\#rsh|\#telnet|\#console/, default => "#rsh"}, 
    timeout => {type => SCALAR, optional => YES,
        regex => qr/\d+/, default => SAY_TIMEOUT },
    waittime => { type => SCALAR, optional => YES,
        regex => qr/\d+/, default => WAIT_UNTIL_DONE},
    }
        );
    my $conn = make_conn($pars{host}, $pars{conntype});
    my $cmd = "disk encrypt show";
    my $disklist_given = $pars{disk};
    if (!defined ($disklist_given)){
        $cmd = $cmd . " ". "*";
    } else {
        $cmd = $cmd . " ". $disklist_given;
    }
        my (undef, undef, $output) = say($conn, $cmd , '-timeout', $pars{timeout});
        logcomment ("DiskEncrypt:show \n $output");
        my @field1=split(/\n/, $output); 
        my $disk_name = $field1[1];
        if (!defined ($disk_name)) {
        logresult ("Fail", "Disk: $disklist_given does not exist");
        } else { 
        return $output;
        }
}

## @name ::verify_locked
## @summary verifies if a disk is locked or unlocked
## 
## @description
## Reports the Locked or Unlocked state of the drive
##
## @return output 
## 
## @arg host Required: Hostrec object
##
## @arg disk Optional: Takes the disk name provided.
## If not provided, considers all the disk in the filer
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## verify_locked(host=>$host_obj, disk => "0c.00.7 0c.00.9"); please note the space in between the disk numbers.
## verify_locked(host=>$host_obj); for all the disk in the filer
## verify_locked(host=> $host_obj, disk => "*"); for all the disk in the filer
##
## @author niveetas@netapp.com
## @status private (in-progress)
##
sub verify_locked
{
    my (%pars) = validate(@_, {
    host => { type=>OBJECT, isa=>'Hostrec'}, 
    disk => { type => SCALAR, optional => YES},
    conntype => { type => SCALAR, optional => YES,
        regex => qr/\#rsh|\#telnet|\#console/, default => "#rsh"}, 
    timeout => {type => SCALAR, optional => YES,
        regex => qr/\d+/, default => SAY_TIMEOUT },
    waittime => { type => SCALAR, optional => YES,
        regex => qr/\d+/, default => WAIT_UNTIL_DONE},
    }
        );

    my (@num, $n, $i, $disk_name, $locked_state, $output);
        $output = fde_show(host=> $pars{host}, disk => $pars{disk}, conntype => $pars{conntype}, timeout => $pars{timeout}, waittime => $pars{waittime} );
    @num = split(/\n/, $output);
    $n = scalar (@num);
    chomp($output); 
    my @field1=split(/\n/, $output); 
    my $header=$field1[0];
    for ($i=1;$i<$n;$i++) {
        my $value=$field1[$i];
        chomp($value);
        my @field2=split(/\s+/, $value);
        my $disk_name=$field2[0];
        my $locked_state=$field2[2];
        if ($locked_state eq "Yes"){
            logcomment ("DiskEncrypt:verify_locked, Disk: $disk_name is LOCKED");
        } else {
            logcomment ("DiskEncrypt:verify_locked, Disk: $disk_name is  NOT LOCKED");
        }
    }

}

## @name ::print_key
## @summary prints the key_id associated with the disk
## 
## @description
## Reports the key id corresponding to the disk
##
## @return output 
## 
## @arg host Required: Hostrec object
##
## @arg disk Optional: Takes the disk name provided.
## If not provided, considers all the disk in the filer
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## print_key(host=>$host_obj, disk => "0c.00.7");
## print_key(host=>$host_obj); for all the disk in the filer
## print_key(host=> $host_obj, disk => "*"); for all the disk in the filer
## print_key(host=> $host_obj, disk => $disk_name);
##
## @author niveetas@netapp.com
## @status private (in-progress)
##

sub print_key
{
    my (%pars) = validate(@_, {
    host => { type=>OBJECT, isa=>'Hostrec'}, 
    disk => { type => SCALAR, optional => YES},
    conntype => { type => SCALAR, optional => YES,
        regex => qr/\#rsh|\#telnet|\#console/, default => "#rsh"}, 
    timeout => {type => SCALAR, optional => YES,
        regex => qr/\d+/, default => SAY_TIMEOUT },
    waittime => { type => SCALAR, optional => YES,
        regex => qr/\d+/, default => WAIT_UNTIL_DONE},
    }
        );

    my (@num,$n, $i, $disk_name, $locked_state, $output);
    $output = fde_show(host=> $pars{host}, disk => $pars{disk}, conntype => $pars{conntype}, timeout => $pars{timeout}, waittime => $pars{waittime} );
    @num = split(/\n/, $output);
    $n = scalar(@num);
    chomp($output); 
    my @field1=split(/\n/, $output); 
    my $header=$field1[0];
    for ($i=1;$i<$n;$i++) {
        my $value=$field1[$i];
        chomp($value);
        my @field2=split(/\s+/, $value);
        my $disk_name=$field2[0];
        my $key_id=$field2[1];
        logcomment ("DiskEncrypt:print_key, Disk: $disk_name is associated with key: $key_id");
        }
}


## @name ::power_cycle_disk
## @summary power cycles disk
## 
## @description
## power cycles disk and waits till the disk are up
##
## @return output 
## 
## @arg host Required: Hostrec object
##
## @arg disk Optional: Takes the disk name provided.
## If not provided, considers all the disk in the filer
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## power_cycle_disk(host=>$host_obj, disk => "0c.00.7"); 
## power_cycle_disk(host=>$host_obj); for all the disk in the filer
## 
## @author niveetas@netapp.com
## @status private (in-progress)
##

sub power_cycle_disk{
    my (%pars) = validate(@_, {
    host => { type=>OBJECT, isa=>'Hostrec'}, 
    disk => { type => SCALAR, optional => YES},
    aggregate => { type => SCALAR, optional => YES},
    conntype => { type => SCALAR, optional => YES,
        regex => qr/\#rsh|\#telnet|\#console/, default => "#rsh"}, 
    timeout => {type => SCALAR, optional => YES,
        regex => qr/\d+/, default => SAY_TIMEOUT },
    waittime => { type => SCALAR, optional => YES,
        regex => qr/\d+/, default => WAIT_UNTIL_DONE},
    }
        );

    my $conn = make_conn($pars{host}, $pars{conntype});
    my $cmd = "priv set -q test; disk power_cycle -f -d ";
    my $disklist_given = $pars{disk};
    my $disklist = disk_show(host=> $pars{host});
    my @disklist_all = ();
        while ( my ($key, $value) = each(%$disklist) )
        {
            push(@disklist_all, "$key");
        }
    my $mark=ems_log_get_mark(host=>$pars{host});
    if (!defined ($disklist_given)) {
        $cmd = $cmd ." ". "@disklist_all";
    } else {
        $cmd = $cmd . " ". $disklist_given;
    }
    say($conn, $cmd , '-timeout', $pars{timeout}, $pars{waitime});
    sleep 20;
    if (!defined ($disklist_given)) {
        foreach my $disk (@disklist_all)
        {
            my @ems_events = "sas.adapter.debug";
            my $test = ems_log_check(host=>$pars{host}, eventtype=>@ems_events, key=>"debug_string", value=>"Powercycle on device $disk complete: status 0", timestamp=>$mark);
            if ($test >= 1)
            {
                logcomment ("DiskEncrypt:power_cycle_disk: disk $disk is up after the powercycle");
            } else {
                logcomment ("DiskEncrypt:power_cycle_disk: disk $disk is not up after the powercycle");
            }
        }
    } else {
          my @ems_events = "sas.adapter.debug";
          my $test = ems_log_check(host=>$pars{host}, eventtype=>@ems_events, key=>"debug_string", value=>"Powercycle on device $disklist_given complete: status 0", timestamp=>$mark);
          if ($test >= 1)
          {
              logcomment ("DiskEncrypt:power_cycle_disk: disk $disklist_given is up after the powercycle");
          }else {
              logcomment ("DiskEncrypt:power_cycle_disk: disk $disklist_given is not up after the powercycle");
          }
       }
}

# ---------------------------------------
# routine to test the module
# ---------------------------------------


sub testme
{
    package main; 
    our ($FILER); 
    package Storage::FDE::DiskEncrypt;
    use Tharn qw(say logcomment logresult host param); 
    param('FILER', '-mesg', "Filer to test"); 
    my $host_obj = host($FILER); 
    fde_lock(host=> $host_obj, disk=>'0c.00.1'); 
    fde_show(host=> $host_obj); 
    fde_unlock(host=> $host_obj); 
    fde_show(host=> $host_obj); 
    fde_rekey(host=> $host_obj, key => "0x0"); #TODO: test with an invalid key_id, it is tested only with key_id 0x0
    verify_locked(host=> $host_obj);
    print_key(host=> $host_obj);
    power_cycle_disk(host=> $host_obj, disk=>'0c.00.1'); 
    }
1;
