#
# $Id: //depot/prod/test/Rsierranevada/lib/Storage/FDE/KeyManager.pm#11 $
#

## @summary KeyManager routine
##
## @description
## This module provides all the capabilities
## required for FDE KeyManger work flow.
##
## @synopsis
## use strict;
## use warnings;
## use Storage::FDE::KeyManager qw( keymanager_add keymanager_remove keymanager_setup keymanager_show keymanager_query keymanager_rekey keymanager_restore keymanager_status keymanager_sync verify_uniquekey);
## param('FILER', -mesg, "hostspec of FILER");
## my $host_obj = host($FILER);
## keymanager_add(host=>$host_obj, kmipserver_ip1 => "0.0.0.0");
## keymanager_remove(host=>$host_obj, kmipserver_ip1 => "0.0.0.0");
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", kmipserver_ip2 =>"1.1.1.1", kmipserver_ip3 =>"2.2.2.2", kmipserver_ip4 =>"3.3.3.3", enterpassphrase => TRUE, passphrase => "testtesttesttesttesttest", keytagname => "FDE", lockdrives => TRUE);
## keymanager_show(host=>$host_obj);
## keymanager_query(host=>$host_obj);
## keymanager_rekey(host=>$host_obj);
## keymanager_status(host=>$host_obj);
## keymanager_sync(host=>$host_obj);

## @author niveetas@netapp.com
## @status private (in-progress)
## @update toc@netapp.com 10/21/10: removed SSL certificate check, no longer part of key_manager setup.
## @update toc@netapp.com 10/22/10: added TCP port support since key_manager setup now requires TCP port setup.
## @update toc@netapp.com 10/25/10: changed "-wait" to "-prompt" with "$promptstring while setting up kmip server ip addreses for key_manager setup.
## @update toc@netapp.com 11/01/10: removed SSL certificate check, no longer part of key_manager add.
## @update toc@netapp.com 11/03/10: changed key_manager restore to be able to restore more than one kmip server and add another error check.
## @update toc@netapp.com 06/20/11: added client_private.pem passphrase protect into key_manager setup.
## @update toc@netapp.com 06/20/11: key_manager add does not have fix for passphrase protect check since assuming key_manager setup has been run.
## @update toc@netapp.com 06/23/11: added keymgr install cert cmd to install client.pem and client_private.pem since key_manager setup library does not do that.
## @update gurus@netapp.com 03/09/12: Updated keymanager_setup() to handle prompts
## @update gurus@netapp.com 10/18/12: Updated keymanager_rekey(), keymanager_show(), keymanager_remove(), keymanager_restore() and keymanager_add() subroutines to be mode agnostic
## @update nandeesh@netapp.com 30/10/12: Modified sub keymanager_add() such that it won't exit at first success in case of multiple servers.
## @update nandeesh@netapp.com 2013-02-27: Modified sub keymanager_query() subroutines to be mode agnostic.
## @update nandeesh@netapp.com 2013-02-27: Modified sub keymanager_add() to handle addtional prompts.
## @update nandeesh@netapp.com 2013-05-22: Modified sub keymanager_setup() subroutines to be mode agnostic and converted say() to NACL execute().
## @update gururajc@netapp.com 2013-06-18: Added apiset param to all subroutines.
## @update gururajc@netapp.com 2013-07-26: 721146 as per this burt given support to run local.
## @update nandeesh@netapp.com 2013-10-23: Modified sub keymanager_setup() to handle a new prompt encountered in FullSteam; currently IPv6 is being ignored

package Storage::FDE::KeyManager;
use strict;
use warnings;
use NACL::C::Node;
use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use Storage::Disk;
use filer;
use Tharn qw(say logcomment logresult);
use Tharn;
use Utils qw(make_conn);
use Data::Dumper;
use base qw(Exporter);
use Params::Validate
    qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use vars;
use Storage::Disk qw(disk_show);
use Connectrec qw();

#use FS::Util::CIFS_Session qw(disconnect);
our ($FILER);
our @EXPORT_OK;

BEGIN {
    use base qw(Exporter);
    our @EXPORT_OK = qw(
        keymanager_add
        keymanager_remove
        keymanager_setup
        keymanager_show
        keymanager_query
        keymanager_rekey
        keymanager_restore
        keymanager_status
        keymanager_sync
        verify_uniquekey
    );
} ## end BEGIN

# ---------------
# define constant
# ---------------
use constant SAY_TIMEOUT     => 300;
use constant PASS            => 1;
use constant FAIL            => 0;
use constant NOOP            => 2;
use constant MAINT_PROMPT    => "\\*\>";
use constant WAIT_UNTIL_DONE => -1;         # wait until done
use constant NORMAL_MODE     => "normal";
use constant MAINT_MODE      => "maint";
use constant YES             => 1;
use constant NO              => 0;
use constant VOLNAME         => "foo";
use constant TRUE            => 1;
use constant FALSE           => 0;

# ======================= DEFINE ROUTINES ===========================

## @name ::keymanager_add
## @summary runs the CLI key_manager add command
##
## @assumption
## certificates <keyserverip>_CA.pem exist in vol0/home/certs/
## certificate for client.pem, client_private.pem is installed in the filer
##
## @description
## Reports the expected output of the command
##
## @return 1 incase of key_manager add command sucess
##         0 incase of key_manager add command fail.
##
## @arg host Required: Hostrec object
##
## @arg kmipserver_ip1 Required: The ip of the first KMIP server that has to be added to store the keys
##
## @arg kmipserver_ip2 Optional: The ip of the second KMIP server that has to be added to store the keys
## If not provided, just adds one KMIP server
##
## @arg kmipserver_ip3 Optional: The ip of the third KMIP server that has to be added to store the keys
## If not provided, just adds two KMIP server
##
## @arg kmipserver_ip4 Optional: The ip of the fourth KMIP server that has to be added to store the keys
## If not provided, just adds three KMIP server
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## keymanager_add(host => host($FILER), kmipserver_ip => "172.18.42.31");
##
## @author niveetas@netapp.com
## @status private (in-progress)
##
my $filer_node;
sub keymanager_add {
    my (%pars) = validate(
        @_,
        {   host => { type => OBJECT, isa => 'Hostrec' },
            kmipserver_ip1 =>
                { type => SCALAR, regex => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/ },
            kmipserver_ip2 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip3 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip4 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            conntype => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\#rsh|\#telnet|\#console/,
                default  => "#rsh"
            },
            timeout => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            waittime => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => WAIT_UNTIL_DONE
            },
            passphrase_answer => {
                type     => SCALAR,
                optional => YES,
                default  => 'No'
            },
            passphrase => {
                type     => SCALAR,
                optional => YES,
                default  => "\n"
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
        }
    );
    my ( $match, $before, $after, $prompt, $cmd, $output );
    my $apiset_obj = $pars{api_set_obj};
    my $cmode_bit = 0;
    my $apiset_created = 0;
    $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
    if((!$apiset_obj)||($apiset_obj->{actual_set} =~ /Nodescope/i)){
        $apiset_created = 1;
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        if ($filer_node->is_cmode){
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    set       => "CMode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
            my $response = $apiset_obj->execute_command(command => "run local");
            if ($response =~ /Too many users logged in\!\s+Please try again later\./i){
                $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                die();
            }
            $apiset_obj->execute_command(command => "priv set test");
            $cmode_bit = 1;
        } else {
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    connid    => "console",
                                                    set       => "7Mode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
        }
    }elsif($apiset_obj->{actual_set} =~ /CMode/i){
        my $response = $apiset_obj->execute_command(command => "run local");
        if ($response =~ /Too many users logged in\!\s+Please try again later\./i){
            $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
            die();
        }
        $apiset_obj->execute_command(command => "priv set test");
        $cmode_bit = 1;
    }elsif($apiset_obj->{actual_set} =~ /7Mode/i){
        $apiset_created = 1;
        $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                category  => "Node",
                                                interface => "CLI",
                                                connid    => "console",
                                                set       => "7Mode");
        $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
    }
    # if(!$apiset_obj){
        # my $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
    # }
    my @servers = ("$pars{kmipserver_ip1}");
    if ( $pars{kmipserver_ip2} ) {
        push( @servers, "$pars{kmipserver_ip2}" );
    }
    if ( $pars{kmipserver_ip3} ) {
        push( @servers, "$pars{kmipserver_ip3}" );
    }
    if ( $pars{kmipserver_ip4} ) {
        push( @servers, "$pars{kmipserver_ip4}" );
    }
    foreach my $server_ip (@servers) {
        my $num_servers = keymanager_show( host => $pars{host},api_set_obj => $apiset_obj );
        if ( $num_servers == 4 ) {
            logcomment("No more than 4 key server allowed");
            return 0;
        } else {
                if ($filer_node->is_cmode){
                    my $output = $apiset_obj->execute_command(command => "run local");
                    if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
                        $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                        die();
                    }
                    $apiset_obj->execute_command(command => "priv set test");
                }
            my $cmd = "key_manager add -key_server $server_ip";
            my $prompts_answers = ['.*Is this file protected by a passphrase\?*'=> $pars{passphrase_answer},
                                   'Please enter passphrase for client_private.pem:' => $pars{passphrase},
                                   'Please reenter the passphrase:' => $pars{passphrase}];

            my $output = $apiset_obj->execute_command(
                                           'command'                => $cmd,
                                           'connectrec-timeout'     => $pars{timeout},
                                           'connectrec-match_table' => $prompts_answers);

            if ( $output =~ /is not a registered key server/ ) {
                logcomment(
                    "KeyManager: keymanager_add: $server_ip is not a registered key server"
                );
                logcomment("Command original output is $output");
                return 0;
            } elsif ( $output =~ /An important file has not been installed./ )
            {
                logcomment(
                    "KeyManager: keymanager_add: Certificate for $server_ip is not present at /vol0/home/certs/"
                );
                logcomment("Command original output is $output");
                return 0;
            } elsif ( $output =~ /Registration complete./ ) {
                logcomment(
                    "KeyManager: keymanager_add: Registration for $server_ip complete"
                );
            } ## end elsif ( $output =~ /Registration complete./)
        }    # end main else block
    }    # end foreach
    if ($cmode_bit){
        $apiset_obj->execute_command(command => "exit");
    }
    if($apiset_created){
        $apiset_obj->cleanup_APISet();
    }
    return 1;
}    # end sub keymanager_add

## @name ::keymanager_remove
## @summary runs the CLI key_manager remove command
##
## @description
## Reports the expected output of the command
##
## @return output
##
## @arg host Required: Hostrec object
##
## @arg kmipserver_ip1 Required: The ip of the first KMIP server that has to be removed
##
## @arg kmipserver_ip2 Optional: The ip of the second KMIP server that has to be removed
## If not provided, removes one KMIP server
##
## @arg kmipserver_ip3 Optional: The ip of the third KMIP server that has to be removed
## If not provided, removes two KMIP server
##
## @arg kmipserver_ip4 Optional: The ip of the fourth KMIP server that has to be removed
## If not provided, removes three KMIP server
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## keymanager_remove(host=>$FILER, kmipserver_ip => "172.18.42.31");
##
## @author niveetas@netapp.com
## @status private (in-progress)
##

sub keymanager_remove {
    my (%pars) = validate(
        @_,
        {   host => { type => OBJECT, isa => 'Hostrec' },
            kmipserver_ip1 =>{
                type => SCALAR,
                regex => qr/(^\b(?:\d{1,3}\.){3}\d{1,3}\b)||(q)/,
                optional => YES
                },
            kmipserver_ip2 => {
                type     => SCALAR,
                regex    => qr/(^\b(?:\d{1,3}\.){3}\d{1,3}\b)||(q)/,
                optional => YES
            },
            kmipserver_ip3 => {
                type     => SCALAR,
                regex    => qr/(^\b(?:\d{1,3}\.){3}\d{1,3}\b)||(q)/,
                optional => YES
            },
            kmipserver_ip4 => {
                type     => SCALAR,
                regex    => qr/(^\b(?:\d{1,3}\.){3}\d{1,3}\b)||(q)/,
                optional => YES
            },
            conntype => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\#rsh|\#telnet|\#console|\#ssh/,
                default  => "#rsh"
            },
            timeout => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            waittime => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => WAIT_UNTIL_DONE
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
        }
    );
    my $apiset_obj = $pars{api_set_obj};
    my $cmode_bit = 0;
    my $apiset_created = 0;
    $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
    if((!$apiset_obj)||($apiset_obj->{actual_set} =~ /Nodescope/i)){
        $apiset_created = 1;
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        if ($filer_node->is_cmode){
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    set       => "CMode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
            my $output = $apiset_obj->execute_command(command => "run local");
            if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
                $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                die();
            }
            $apiset_obj->execute_command(command => "priv set test");
            $cmode_bit = 1;
        } else {
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    connid    => "console",
                                                    set       => "7Mode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
        }
    }elsif($apiset_obj->{actual_set} =~ /CMode/i){
        my $output = $apiset_obj->execute_command(command => "run local");
        if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
            $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
            die();
        }
        $apiset_obj->execute_command(command => "priv set test");
        $cmode_bit = 1;
    }elsif($apiset_obj->{actual_set} =~ /7Mode/i){
        $apiset_created = 1;
        $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                category  => "Node",
                                                interface => "CLI",
                                                connid    => "console",
                                                set       => "7Mode");
        $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
    }
    my @servers = ("$pars{kmipserver_ip1}");
    if ( $pars{kmipserver_ip2} ) {
        push( @servers, "$pars{kmipserver_ip2}" );
    }
    if ( $pars{kmipserver_ip3} ) {
        push( @servers, "$pars{kmipserver_ip3}" );
    }
    if ( $pars{kmipserver_ip4} ) {
        push( @servers, "$pars{kmipserver_ip4}" );
    }
    foreach my $server_ip (@servers) {
            if($server_ip eq "q") {
                # $quitbit = "q";
                last;
            }
        my $num_servers = keymanager_show( host => $pars{host},api_set_obj => $apiset_obj );
        if ( $num_servers == 0 ) {
            logresult( "FAIL",
                "KeyManager: keymanager_remove: Key server not found" );
        } else {
# my $delete_cert_cmd = "keymgr delete cert"." ".$server_ip ."_CA.pem";
# (undef, undef, $output) = say($conn,$delete_cert_cmd , '-timeout', $pars{timeout});
# if ($output =~ m/An error occurred while attempting/) {
#     logresult ("FAIL", "KeyManager: keymanager_remove: Certificate for $server_ip does not exist.");
# } else {
            if ($filer_node->is_cmode){
                my $output = $apiset_obj->execute_command(command => "run local");
                if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
                    $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                    die();
                }
                $apiset_obj->execute_command(command => "priv set test");
            }
            my $cmd = "key_manager remove -key_server $server_ip";
            my $output = $apiset_obj->execute_command('command'   => $cmd,
                'connectrec-timeout'     => $pars{timeout},
            );
            logresult( "PASS", "KeyManager: keymanager_remove:\n $output" );

            # }
        } ## end else [ if ( $num_servers == 0)
    } ## end foreach my $server_ip (@servers)
    if ($cmode_bit){
        $apiset_obj->execute_command(command => "exit");
    }
    if($apiset_created){
        $apiset_obj->cleanup_APISet();
    }
} ## end sub keymanager_remove

## @name ::keymanager_setup
## @summary runs the CLI key_manager setup command
##
## @description
## Reports setup completed
##
## @assumption
## certificates <keyserverip>_CA.pem exist in vol0/home/certs/
## certificate for client.pem, client_private.pem is installed in the filer
##
## @return output : 1, on success or 0 on failure
##
## @arg host Required: Hostrec object
##
## @arg kmipserver_ip1 Optional: Takes the ip provided.
## If not provided, considers no kmip server is available
##
## @arg kmipserver_ip2 Optional: Takes the ip provided.
## If not provided, considers that only one kmip server is available
##
## @arg kmipserver_ip3 Optional: Takes the ip provided.
## If not provided, considers that only two kmip server is available
##
## @arg kmipserver_ip4 Optional: Takes the ip provided.
## If not provided, considers that only three kmip server is available
##
## @arg quit1 : This is required if there are no kmipserver_ip specified
##
## @arg quit2 : This is required if kmipserver_ip1 is specified
##
## @arg quit3 : This is required if kmipserver_ip1 and kmipserver_ip2 are specified
##
## @arg quit4 : This is required if kmipserver_ip1, kmipserver_ip2 and kmipserver_ip3 are specified
##
## @arg keytagname Optional: Takes the value provided.
## If not provided, considers host name as the keytagname
##
## @arg enterpassphrase Optional: Takes the value provided.
## If not provided, the script follows the branch where no passphrase is required during key_manager setup
##
## @arg passphrase Optional: Takes the value provided.
## If not provided, key_manager setup automatically uses the machine generated passphrase
##
## @arg synckeyservers Optional: Takes the value provided and syncs key servers
## If not provided, key_manager setup doesnot sync key server
##
## @arg lockdrives Optional: Takes the value provided and locks the drives
## If not provided, key_manager setup doesnot lock the drives at that point of time
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## keymanager_setup(host=>$host_obj, quit1 => TRUE);
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", quit2 => TRUE);
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", kmipserver_ip2 =>"1.1.1.1", quit3 => TRUE);
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", kmipserver_ip2 =>"1.1.1.1", kmipserver_ip3 =>"2.2.2.2", quit4 => TRUE);
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", kmipserver_ip2 =>"1.1.1.1", kmipserver_ip3 =>"2.2.2.2", kmipserver_ip4 =>"3.3.3.3");
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", kmipserver_ip2 =>"1.1.1.1", kmipserver_ip3 =>"2.2.2.2", kmipserver_ip4 =>"3.3.3.3", enterpassphrase => TRUE);
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", kmipserver_ip2 =>"1.1.1.1", kmipserver_ip3 =>"2.2.2.2", kmipserver_ip4 =>"3.3.3.3", enterpassphrase => TRUE, passphrase => "testtesting");
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", quit2 => TRUE, enterpassphrase => TRUE, passphrase => "ttesttesttesttesttesttest", keytagname => "FDE");
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", quit2 => TRUE, enterpassphrase => TRUE, passphrase => "testtesttesttesttesttest", keytagname => "FDE", synckeyservers => TRUE);
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", quit2 => TRUE, enterpassphrase => TRUE, passphrase => "testtesttesttesttesttest", keytagname => "FDE", synckeyservers => TRUE, lockdrives => TRUE);
## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", kmipserver_ip2 =>"1.1.1.1", kmipserver_ip3 =>"2.2.2.2", kmipserver_ip4 =>"3.3.3.3", enterpassphrase => TRUE, passphrase => "testtesttesttesttesttest", keytagname => "FDE", synckeyservers => TRUE, lockdrives => TRUE);
##
## @author niveetas@netapp.com
## @status private (in-progress)

sub keymanager_setup {
    my (%pars) = validate(
        @_,
        {   host => { type => OBJECT, isa => 'Hostrec' },
            file_protect =>
                { type => BOOLEAN, optional => YES, default => FALSE },
            file_passphrase => { type => SCALAR, optional => YES },
            kmipserver_ip1  => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES,
            },
            kmipserver_ip2 => {
                type     => SCALAR,
                regex    => qr/(^\b(?:\d{1,3}\.){3}\d{1,3}\b)||(q)/,
                optional => YES,
            },
            kmipserver_ip3 => {
                type     => SCALAR,
                regex    => qr/(^\b(?:\d{1,3}\.){3}\d{1,3}\b)||(q)/,
                optional => YES,
            },
            kmipserver_ip4 => {
                type     => SCALAR,
                regex    => qr/(^\b(?:\d{1,3}\.){3}\d{1,3}\b)||(q)/,
                optional => YES,
            },
            quit1 => { type => BOOLEAN, optional => YES, default => FALSE },
            quit2 => { type => SCALAR,  optional => YES, default => FALSE },
            quit3 => { type => SCALAR,  optional => YES, default => FALSE },
            quit4 => { type => SCALAR,  optional => YES, default => FALSE },
            tcp_port     => { type => SCALAR, optional => YES },
            keytagname   => { type => SCALAR, optional => YES },
            nw_interface => { type => SCALAR, optional => YES },
            ip_address   => { type => SCALAR, optional => YES },
            netmask      => { type => SCALAR, optional => YES },
            gateway      => { type => SCALAR, optional => YES },
            ipv6         => { type => SCALAR, optional => YES, default => "no" },   #Currently IPv6 support is ignored
            enterpassphrase =>
                { type => BOOLEAN, optional => YES, default => FALSE },
            passphrase => { type => SCALAR, optional => YES, default  => undef },
            synckeyservers =>
                { type => BOOLEAN, optional => YES, default => FALSE },
            lockdrives =>
                { type => BOOLEAN, optional => YES, default => FALSE },
            conntype => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\#rsh|\#telnet|\#console/,
                default  => "#rsh"
            },
            timeout => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            waittime => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => WAIT_UNTIL_DONE
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
			skip_cert_install => {
                type => BOOLEAN,
                optional => YES,
				default => FALSE,
            },
        }
    );

    my ($match,        $before,          $after,    $prompt,
        $file_protect, $file_passphrase, $tcp_port, $keytagname,
        $key_tagname, $nw_interface, $passphrase, $invalid_passphrase,  $output,$kmipserver_ip1, $kmipserver_ip2, $kmipserver_ip3, $kmipserver_ip4,
        $keytaganswer, $ip_address, $netmask, $gateway, $enterpassphrase, $passprompt, $lockdrives, $passphrase_answer, $api_set_obj,
        $kmipserver_ip1_quit, $kmipserver_ip2_quit, $kmipserver_ip3_quit, $kmipserver_ip4_quit, $invalidflag, $response, $Test, $ipv6YrN,
    );
    my $lockbit = ();
    my $conn_tel = make_conn( $pars{host}, "#console" );
    my $apiset_obj = $pars{api_set_obj};
    $ipv6YrN = $pars{ipv6};
    # if($ipv6YrN ne "yes") {
        # $ipv6YrN = "no";
    # }
    my $cmode_bit = 0;
    my $apiset_created = 0;
    $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
    if((!$apiset_obj)||($apiset_obj->{actual_set} =~ /Nodescope/i)){
        $apiset_created = 1;
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        if ($filer_node->is_cmode){
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    set       => "CMode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
            my $output = $apiset_obj->execute_command(command => "run local");
            if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
                $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                die();
            }
            $apiset_obj->execute_command(command => "priv set test");
            $cmode_bit = 1;
        } else {
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    connid    => "console",
                                                    set       => "7Mode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
        }
    }elsif($apiset_obj->{actual_set} =~ /CMode/i){
        my $output = $apiset_obj->execute_command(command => "run local");
        if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
            $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
            die();
        }
        $apiset_obj->execute_command(command => "priv set test");
        $cmode_bit = 1;
    }elsif($apiset_obj->{actual_set} =~ /7Mode/i){
        $apiset_created = 1;
        $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                category  => "Node",
                                                interface => "CLI",
                                                connid    => "console",
                                                set       => "7Mode");
        $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
    }
    # if(!$apiset_obj){
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        # $apiset_obj->set_timeout( 'connectrec-timeout' => 600 );
        # $apiset_created = 1;
    # }
    #adding the certificates for all the kmip_servers
    my @servers;
	if ( $pars{kmipserver_ip1} ) {
        push( @servers, "$pars{kmipserver_ip1}" );
    }
    if ( $pars{kmipserver_ip2} ) {
        push( @servers, "$pars{kmipserver_ip2}" );
    }
    if ( $pars{kmipserver_ip3} ) {
        push( @servers, "$pars{kmipserver_ip3}" );
    }
    if ( $pars{kmipserver_ip4} ) {
        push( @servers, "$pars{kmipserver_ip4}" );
    }
    logcomment("Available key servers are @servers");
    my $check_bit = ();
    my $quitbit = ();
    my $len = @servers;
    # print $len;
    if ( $len eq "4" ) {    #recently added ; Fix the issue seen when four key servers are installed in normal testcase
        push( @servers, "q" );
    }
	if($pars{skip_cert_install} eq FALSE ) {
    foreach my $server_ip (@servers) {
        my $num_servers = keymanager_show( host => $pars{host},api_set_obj => $apiset_obj );
        my $promptstring = $pars{host}->{id};
        if ( $num_servers > 4 ) {
            logcomment( "No more than 4 key server allowed" );
            return 0;
        } else {
                if ($filer_node->is_cmode){
                    my $output = $apiset_obj->execute_command(command => "run local");
                    if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
                        $Log->comment("FAIL: The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                        die();
                    }
                    $apiset_obj->execute_command(command => "priv set test");
                }
            if($server_ip eq "q") {
                $quitbit = "q";
                last;
            } else {
                my $install_cert_cmd
                    = "keymgr install cert /home/certs/" . $server_ip . "_CA.pem";
                $apiset_obj->execute_command(  'command'                => $install_cert_cmd,
                                               'connectrec-timeout'     => $pars{timeout} );
            if(!$check_bit){
                $install_cert_cmd = "keymgr install cert /home/certs/client.pem";
                $apiset_obj->execute_command(  'command'                => $install_cert_cmd,
                                               'connectrec-timeout'     => $pars{timeout} );
                $install_cert_cmd
                    = "keymgr install cert /home/certs/client_private.pem";
                $output = $apiset_obj->execute_command(  'command'     => $install_cert_cmd,
                                               'connectrec-timeout'     => $pars{timeout} );
                $check_bit = 1;
            }
                if ( $output =~ m/does not exist/ ) {
                    logresult( "FAIL",
                        "KeyManager: keymanager_setup: Certificate for $server_ip is not present at /home/certs/"
                    );
                }
            }
        } ## end else [ if ( $num_servers == 4)
    } ## end foreach my $server_ip (@servers)
	}

    if ($pars{file_protect} eq "TRUE") {
        $passphrase_answer = "yes";
        $file_passphrase = $pars{file_passphrase};
        logcomment ("file_protect passphrase is :$file_passphrase");
        $file_protect = "yes";
    } else {
        $passphrase_answer = "no";
        $file_protect = $pars{host}->{id};
        logcomment ("file_protect has no passphrase :'$pars{host}->{id}'");
        $file_protect = "no";
    }

    my $first_bit = ();
    # This is for the first set of ip/quit
    if ( $pars{kmipserver_ip1} ) {
        if ( $quitbit ne $pars{kmipserver_ip1}) {
            $kmipserver_ip1 = $pars{kmipserver_ip1};
            logcomment ("Selected  kmipserver_ip1 is :$kmipserver_ip1");
            $first_bit = 1;
        }
    } else {
        $kmipserver_ip1 = '';
        $first_bit = 1;
    }

    if ( $pars{quit1}  eq "TRUE" ) {
        $kmipserver_ip1_quit = "1";
        $first_bit = 1;
    }
    if(!$first_bit) {
        logcomment ("Invalid option: You must enter at least one key server to continue");
        return 0;
    }

    # This is for the second set of ip/quit
    if(!$kmipserver_ip1_quit){
        if ( $pars{kmipserver_ip2} ) {
            if ( $quitbit ne $pars{kmipserver_ip2}) {
                $kmipserver_ip2 = $pars{kmipserver_ip2};
                logcomment ("Selected  kmipserver_ip2 is :$kmipserver_ip2");
            } else {
                $kmipserver_ip2 = "q";
            }
        } else {
            $kmipserver_ip2 = "q";
        }
        if ( $pars{quit2}  eq "TRUE" ) {
            $kmipserver_ip2_quit = "1";
            $kmipserver_ip2 = "q";
        }
    } else {
        $kmipserver_ip2 = "q";
    }
    # This is for the third set of ip/quit
    if(!$kmipserver_ip2_quit){
        if ( $pars{kmipserver_ip3} ) {
            if ( $quitbit ne $pars{kmipserver_ip3}) {
                $kmipserver_ip3 = $pars{kmipserver_ip3};
                logcomment ("Selected  kmipserver_ip3 is :$kmipserver_ip3");
            } else {
                $kmipserver_ip3 = "q";
            }
        } else {
            $kmipserver_ip3 = "q";
        }
        if ( $pars{quit3}  eq "TRUE" ) {
            $kmipserver_ip3_quit = "1";
            $kmipserver_ip3 = "q";
        }
    } else {
        $kmipserver_ip3 = "q";
    }
    # This is for the fourth set of ip/quit
    if(!$kmipserver_ip3_quit){
        if ( $pars{kmipserver_ip4} ) {
            if ( $quitbit ne $pars{kmipserver_ip4}) {
                $kmipserver_ip4 = $pars{kmipserver_ip4};
                logcomment ("Selected  kmipserver_ip4 is :$kmipserver_ip4");
            } else {
                $kmipserver_ip4 = "q";
            }
        } else {
            $kmipserver_ip4 = "q";
        }
        if ( $pars{quit4}  eq "TRUE" ) {
            $kmipserver_ip4 = "q";
        }
    } else {
        $kmipserver_ip4 = "q";
    }
    # For TCP port setup
    if ( $pars{tcp_port} ) {
        $tcp_port = $pars{tcp_port};
        logcomment("tcp_port :$tcp_port");
    } else {
        $tcp_port = '';
    }

    # For keytagname
    if ( $pars{keytagname} ) {
        $keytagname = $pars{keytagname};
        logcomment("keytagname :$keytagname");
        $keytaganswer = $keytagname;
    } else {
        $key_tagname = $pars{host}->{id};
        logcomment("Default keytagname name is :'$key_tagname'");
        $keytaganswer = "yes";
    } ## end else [ if ( $pars{keytagname})

    #For Enter network interface [e0a]:
    if ( $pars{nw_interface} ) {
        $nw_interface = $pars{nw_interface};
        logcomment ("Selected  nw_interface is :$nw_interface");
    } else {
        $nw_interface = '';
    }

    #For Enter IP address [172.18.96.222]:
    if ( $pars{ip_address} ) {
        $ip_address = $pars{ip_address};
        logcomment ("Selected  ip_address is :$ip_address");
    } else {
        $ip_address = '';
    }

    #For Enter netmask [255.255.254.0]:
    if ( $pars{netmask} ) {
        $netmask = $pars{netmask};
        logcomment ("Selected  netmask is :$netmask");
    } else {
        $netmask = '';
    }

    #For Enter gateway [172.18.96.1]:
    if ( $pars{gateway} ) {
        $gateway = $pars{gateway};
        logcomment ("Selected  gateway is :$gateway");
    } else {
        $gateway = '';
    }
    my $enterphrase = ();
    # For the enter passphrase and generate passphrase section
    $passphrase = $pars{passphrase};
    if ( $pars{enterpassphrase} eq "TRUE" ) {
            $enterpassphrase = "yes";
        if ($passphrase) {
            my @arraychar = split( //, $passphrase );
            my $length_passphrase = @arraychar;
            if ( $length_passphrase >= '20' ) {
                logcomment("Passphrase :$passphrase");
                $passprompt = "no";
            } else {
                logcomment ("FAIL : Passphrase must be 20 characters or longer");
                return 0;
            }
        } else {
            $passprompt = "yes";
        }
        #This is to lock all encrypting drives
        if ( $pars{lockdrives} eq "TRUE") {
            $lockdrives = "yes";
            $lockbit = 1;

        } else {
            $lockdrives = "no";

        } ## end else [ if ( $pars{lockdrives}...
        $enterphrase = 1;
    } else {
        $enterpassphrase = "no";
        $enterphrase = 1;

    }
my ($flag, $ip1, $ip2, $ip3) = 0;
my $flag2 = 1;
my $cmd = "key_manager setup";
my $prompts_answers = [ '.*Is this file protected by a passphrase\?*'=> $passphrase_answer,
                        'Please enter passphrase for client_private.pem\:' => $file_passphrase,
                        '.*Enter the IP address for a key server*' => sub {
                                                                            if ( $flag2 ) {
                                                                                $ip1=1;
                                                                                $flag2 = ();
                                                                                return $kmipserver_ip1;
                                                                            } elsif ( $ip1 ) {
                                                                                $ip2=1;
                                                                                $ip1=();
                                                                                return $kmipserver_ip2;
                                                                            } elsif ($ip2){
                                                                                $ip3=1;
                                                                                $ip2=();
                                                                                return $kmipserver_ip3;
                                                                            } elsif ($ip3){
                                                                                $ip3=();
                                                                                return $kmipserver_ip4;
                                                                            }  else {
                                                                                return $kmipserver_ip1;
                                                                            }
                                                                        },
                        'Enter the TCP port number for kmip server' => $tcp_port,
                        '^Would you like to use' => $keytaganswer,
                        '.*Enter network interface*' => $nw_interface,
                        '.*Enter IP address*' => $ip_address,
                        '.*Enter netmask*' => $netmask,
                        '.*Enter gateway*' => $gateway,
                        '.*Would you like to configure an IPv6 address\?' => $ipv6YrN,
                        '.*encrypting drives at this time\?' => $enterpassphrase,
                        '.*Would you like the system to autogenerate a passphrase\?' => $passprompt,
                        '.*Please enter a passphrase \(20 characters or longer\)' => $passphrase,
                        '.*Please reenter the passphrase\:' => sub {
                                                                    if($flag) {
                                                                        return $passphrase;
                                                                    }else{
                                                                        $flag++;
                                                                        return $file_passphrase;
                                                                    }
                                                                },
                        '.*Should the system lock all encrypting drives at this time\?' => $lockdrives,
                        ];
    try {
        $response = $apiset_obj->execute_command('command' => $cmd,
                                       'connectrec-timeout' => $pars{timeout},
                                   'connectrec-match_table' => $prompts_answers,
                                                   'ignore' => ["\<$key_tagname\> as"],
                                                  'timeout' => 200,
                                    );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $Test->nlog("Error is :" . $exception_object->text());
        return 0;
    };
    my $failbit = ();
    print "Output dumped so far is" . Dumper $response;
    if ($response =~ /Oops\: Detected by kmip_send_msg_rekey at fde_keymgr\.c\:(\S+)/i) {
        logcomment ("Oops: Detected by kmip_send_msg_rekey at fde_keymgr.c:$1");
        $failbit = 1;
    } elsif ($response =~ /Oops\: kmip_client reply unexpected status\: (.*)/i) {
        logcomment ("Oops: kmip_client reply unexpected status: $1");
        $failbit = 1;
    }
    if($failbit) {
        return 0;
    }
    if($lockbit) {
        logcomment ("KeyManager:keymanager_setup:Completed key_manager setup and the drives are locked");
        return 1;
    } else {
        logcomment ("KeyManager:keymanager_setup:Completed key_manager setup but the drives are not locked");
        return 1;
    }
    if($enterphrase) {
        logcomment ("KeyManager:keymanager_setup:Key_manager setup completed");
        return 1;
    }
    if ($cmode_bit){
        $apiset_obj->execute_command(command => "exit");
    }
    if($apiset_created){
        $apiset_obj->cleanup_APISet();
    }
    # if($apiset_created){
        # $apiset_obj->cleanup_APISet();
    # }
} ## end sub keymanager_setup
## @name ::keymanager_show
## @summary runs the CLI key_manager show command
##
## @description
## Reports the expected output of the command
##
## @return output
##
## @arg host Required: Hostrec object
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## keymanager_show(host=>$FILER);
##
## @author niveetas@netapp.com
## @status private (in-progress)
##

sub keymanager_show {
    my (%pars) = validate(
        @_,
        {   host     => { type => OBJECT, isa => 'Hostrec' },
            conntype => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\#rsh|\#telnet|\#console/,
                default  => "#rsh"
            },
            timeout => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            waittime => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => WAIT_UNTIL_DONE
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
        }
    );
    my @n;
    my $apiset_obj = $pars{api_set_obj};
    my $cmode_bit = 0;
    my $apiset_created = 0;
    $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
    if((!$apiset_obj)||($apiset_obj->{actual_set} =~ /Nodescope/i)){
        $apiset_created = 1;
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        if ($filer_node->is_cmode){
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    set       => "CMode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
            my $output = $apiset_obj->execute_command(command => "run local");
            if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
                $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                die();
            }
            $apiset_obj->execute_command(command => "priv set test");
            $cmode_bit = 1;
        } else {
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    connid    => "console",
                                                    set       => "7Mode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
        }
    }elsif($apiset_obj->{actual_set} =~ /CMode/i){
        my $output = $apiset_obj->execute_command(command => "run local");
        if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
            $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
            die();
        }
        $apiset_obj->execute_command(command => "priv set test");
        $cmode_bit = 1;
    }elsif($apiset_obj->{actual_set} =~ /7Mode/i){
        $apiset_created = 1;
        $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                category  => "Node",
                                                interface => "CLI",
                                                connid    => "console",
                                                set       => "7Mode");
        $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
    }
    # if(!$apiset_obj){
        # my $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
    # }
    my $cmd = "key_manager show";
    my $output = $apiset_obj->execute_command(
           'command'                => $cmd,
           'connectrec-timeout'     => $pars{timeout}
    );
    if ( $output =~ m/No key servers registered/ ) {
        return 0;
    } else {
        @n = split( /\n/, $output);
    }
    if ($cmode_bit){
        $apiset_obj->execute_command(command => "exit");
    }
    if($apiset_created){
        $apiset_obj->cleanup_APISet();
    }
    return @n;
} ## end sub keymanager_show

## @name ::keymanager_query
## @summary runs the CLI key_manager query command
##
## @description
## Reports the expected output of the command
##
## @return output
##
## @arg host Required: Hostrec object
##
## @arg kmipserver_ip1 Optional: The ip of the first KMIP server that has to be queried
## If not provided, it considers all the keyservers
##
## @arg kmipserver_ip2 Optional: The ip of the second KMIP server that has to be queried
## If not provided, it considers all the keyservers
##
## @arg kmipserver_ip3 Optional: The ip of the third KMIP server that has to be queried
## If not provided, it considers all the keyservers
##
## @arg kmipserver_ip4 Optional: The ip of the fourth KMIP server that has to be queried
## If not provided, it considers all the keyservers
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## keymanager_query(host=>$FILER);
##
## @author niveetas@netapp.com
## @status private (in-progress)
##
sub keymanager_query {
    my (%pars) = validate(
        @_,
        {   host           => { type => OBJECT, isa => 'Hostrec' },
            kmipserver_ip1 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip2 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip3 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip4 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            conntype => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\#rsh|\#telnet|\#console/,
                default  => "#rsh"
            },
            timeout => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            waittime => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => WAIT_UNTIL_DONE
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
        }
    );
    my $conn = make_conn( $pars{host}, $pars{conntype} );
    my $cmd = "key_manager query";
    if ( $pars{kmipserver_ip1} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip1}";
    }
    if ( $pars{kmipserver_ip2} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip2}";
    }
    if ( $pars{kmipserver_ip3} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip3}";
    }
    if ( $pars{kmipserver_ip4} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip4}";
    }
    my $apiset_obj = $pars{api_set_obj};
    my $cmode_bit = 0;
    my $apiset_created = 0;
    $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
    if((!$apiset_obj)||($apiset_obj->{actual_set} =~ /Nodescope/i)){
        $apiset_created = 1;
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        if ($filer_node->is_cmode){
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    set       => "CMode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
            my $output = $apiset_obj->execute_command(command => "run local");
            if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
                $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                die();
            }
            $apiset_obj->execute_command(command => "priv set test");
            $cmode_bit = 1;
        } else {
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    connid    => "console",
                                                    set       => "7Mode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
        }
    }elsif($apiset_obj->{actual_set} =~ /CMode/i){
        my $output = $apiset_obj->execute_command(command => "run local");
        if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
            $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
            die();
        }
        $apiset_obj->execute_command(command => "priv set test");
        $cmode_bit = 1;
    }elsif($apiset_obj->{actual_set} =~ /7Mode/i){
        $apiset_created = 1;
        $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                category  => "Node",
                                                interface => "CLI",
                                                connid    => "console",
                                                set       => "7Mode");
        $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
    }
    # if(!$apiset_obj){
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
    # }
    my $after = $apiset_obj->execute_command(
           'command'                => $cmd,
           'connectrec-timeout'     => $pars{timeout}
    );
    if ( $after =~ /not a registered key server/ ) {
        logresult( "FAIL",
            "KeyManager:keymanager_query:\nThe key server is not registered"
        );
    } elsif ( $after =~ /No key servers registered./ ) {
        logresult( "PASS",
            "KeyManager:keymanager_query:\nThere are no key server defined" );
    } else {
        logresult( "PASS", "KeyManager:keymanager_query:\n$after" );
    }
    if ($cmode_bit){
        $apiset_obj->execute_command(command => "exit");
    }
    if($apiset_created){
        $apiset_obj->cleanup_APISet();
    }
    return ($after);
} ## end sub keymanager_query

## @name ::keymanager_rekey
## @summary runs the CLI key_manager rekey command
##
## @description
## Reports the expected output of the command and stores the new key_id and key_tag in a file
##
## @return output
##
## @arg host Required: Hostrec object
##
## @arg manual Optional: This is a flag used for manual key_manager rekey
## If not provided, the key_manager rekey is auto
##
## @arg passphrase Optional: Passphrase required for manual key_manager rekey
## If not provided, the key_manager rekey is auto
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @examples
## keymanager_rekey('host'=>$FILER); # for auto
## keymanager_rekey('host'=>$host_obj,'manual'=> 'TRUE', 'passphrase' => "testtesttesttesttesttest"); #for manual
##
## @author niveetas@netapp.com
## @author dl-storsub-qa
## @status private (in-progress)
##

sub keymanager_rekey {
    my (%pars) = validate(
        @_,
        {   host   => { type => OBJECT,  isa      => 'Hostrec' },
            manual => { type => BOOLEAN, optional => YES, default => FALSE },
            passphrase => { type => SCALAR, optional => YES, },
            timeout    => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
        }
    );

    my $apiset_obj = $pars{api_set_obj};
    my $cmode_bit = 0;
    my $apiset_created = 0;
    $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
    if((!$apiset_obj)||($apiset_obj->{actual_set} =~ /Nodescope/i)){
        $apiset_created = 1;
        $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        if ($filer_node->is_cmode){
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    set       => "CMode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
            my $output = $apiset_obj->execute_command(command => "run local");
            if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
                $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                die();
            }
            $apiset_obj->execute_command(command => "priv set test");
            $cmode_bit = 1;
        } else {
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    connid    => "console",
                                                    set       => "7Mode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
        }
    }elsif($apiset_obj->{actual_set} =~ /CMode/i){
        my $output = $apiset_obj->execute_command(command => "run local");
        if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
            $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
            die();
        }
        $apiset_obj->execute_command(command => "priv set test");
        $cmode_bit = 1;
    }elsif($apiset_obj->{actual_set} =~ /7Mode/i){
        $apiset_created = 1;
        $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                category  => "Node",
                                                interface => "CLI",
                                                connid    => "console",
                                                set       => "7Mode");
        $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
    }
    # if(!$apiset_obj){
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        # $apiset_created = 1;
    # }

    my $disklist = disk_show( host => $pars{host} );
    my @disklist_all = ();

    while ( my ( $key, $value ) = each(%$disklist) ) {
        push( @disklist_all, "$key" );
    }
    my $disk_num = scalar(@disklist_all);
    my $timeout  = ( 100 * $disk_num );
    my $cmd      = "key_manager rekey";
    my ($output);
    if ( $pars{manual} == TRUE ) {
        $cmd = $cmd . " " . "-manual";
        my $passphrase        = $pars{passphrase};
        my @arraychar         = split( //, $passphrase );
        my $length_passphrase = @arraychar;
        if ( $length_passphrase >= 20 ) {
            logcomment("Passphrase:$passphrase");

            my $prompts_answers = [
                '.*Please enter a new passphrase.*'     => $passphrase,
                '.*Please reenter the new passphrase.*' => $passphrase,
            ];

            $output = $apiset_obj->execute_command(
                'command'                => $cmd,
                'connectrec-timeout'     => $pars{timeout},
                'connectrec-match_table' => $prompts_answers,
            );

        } else {
            logresult( "FATAL",
                "Passphrase must be 20 characters or longer" );
        }
    } else {
        $output = $apiset_obj->execute_command(
            'command'            => $cmd,
            'connectrec-timeout' => $pars{timeout}
        );
    } ## end else [ if ( $pars{manual} == ...
    my @output_formated = split( /\n/, $output );
    my ($key_id,$key_tag,$key,$junk);
    foreach (@output_formated) {
        if($_=~/Key\s+ID:/) {
            ( $junk, $key ) = split( /:\s+/, $_);
            $key_id=$_;
        } elsif ($_=~/Key\s+tag:/) {
            ($junk,$key_tag) = split( /:\s+/, $_);
        }
    }
    open( FILEOUT, ">>keyinfo.txt" ) or die "Error:  $!\n";
    my $current    = localtime();
    my $outputline = "$current: $pars{host}->{id}: $key_id, $key_tag\n";
    print FILEOUT $outputline, "";
    close(FILEOUT);
    logcomment("The values are: \n $key_id \n $key_tag ");

    if ( $output =~ m/Warning/ ) {
        logresult( "FAIL",
            "KeyManager:keymanager_rekey:Key Servers are down" );
    } else {
        logcomment( "PASS", "KeyManager:keymanager_rekey:$output" );
    }
    logcomment("key is $key");
    # if($apiset_created){
        # $apiset_obj->cleanup_APISet();
    # }
    if ($cmode_bit){
        $apiset_obj->execute_command(command => "exit");
    }
    if($apiset_created){
        $apiset_obj->cleanup_APISet();
    }
    return $key;
} ## end sub keymanager_rekey

## @name ::keymanager_restore
## @summary runs the CLI key_manager restore command
##
## @description
## Reports the expected output of the command
##
## @return output
##
## @arg host Required: Hostrec object
##
## @arg kmipserver_ip Required: The ip of the KMIP server from where the key_id is restored
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## keymanager_restore(host=>$FILER, kmipserver_ip => '1.1.1.1');
##
## @author niveetas@netapp.com
## @status private (in-progress)

sub keymanager_restore {
    my (%pars) = validate(
        @_,
        {   host => { type => OBJECT, isa => 'Hostrec' },
            kmipserver_ip1 =>
                { type => SCALAR, regex => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/ },
            kmipserver_ip2 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip3 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip4 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            conntype => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\#rsh|\#telnet|\#console|\#ssh/,
                default  => "#rsh"
            },
            timeout => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            waittime => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => WAIT_UNTIL_DONE
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
        }
    );

    my ( $cmd, $output );
    my $apiset_obj = $pars{api_set_obj};
    my $cmode_bit = 0;
    my $apiset_created = 0;
    $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
    if((!$apiset_obj)||($apiset_obj->{actual_set} =~ /Nodescope/i)){
        $apiset_created = 1;
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        if ($filer_node->is_cmode){
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    set       => "CMode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
            my $response = $apiset_obj->execute_command(command => "run local");
            if ($response =~ /Too many users logged in\!\s+Please try again later\./i){
                $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                die();
            }
            $apiset_obj->execute_command(command => "priv set test");
            $cmode_bit = 1;
        } else {
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    connid    => "console",
                                                    set       => "7Mode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
        }
    }elsif($apiset_obj->{actual_set} =~ /CMode/i){
        my $response = $apiset_obj->execute_command(command => "run local");
        if ($response =~ /Too many users logged in\!\s+Please try again later\./i){
            $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
            die();
        }
        $apiset_obj->execute_command(command => "priv set test");
        $cmode_bit = 1;
    }elsif($apiset_obj->{actual_set} =~ /7Mode/i){
        $apiset_created = 1;
        $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                category  => "Node",
                                                interface => "CLI",
                                                connid    => "console",
                                                set       => "7Mode");
        $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
    }
    # if(!$apiset_obj){
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
    # }
    my @servers = ("$pars{kmipserver_ip1}");
    if ( $pars{kmipserver_ip2} ) {
        push( @servers, "$pars{kmipserver_ip2}" );
    }
    if ( $pars{kmipserver_ip3} ) {
        push( @servers, "$pars{kmipserver_ip3}" );
    }
    if ( $pars{kmipserver_ip4} ) {
        push( @servers, "$pars{kmipserver_ip4}" );
    }
    foreach my $server_ip (@servers) {
        my $num_servers = keymanager_show( host => $pars{host},api_set_obj => $apiset_obj );
        if ( $num_servers == 0 ) {
            logresult( "FAIL",
                "KeyManager: keymanager_restore: Key server not found" );
        } else {
            if ($filer_node->is_cmode){
                my $response = $apiset_obj->execute_command(command => "run local");
                if ($response =~ /Too many users logged in\!\s+Please try again later\./i){
                    $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                    die();
                }
                $apiset_obj->execute_command(command => "priv set test");
            }
            my $cmd = "key_manager restore -key_server " . $server_ip;
            $output = $apiset_obj->execute_command(command=>$cmd, 'connectrec-timeout' => $pars{timeout});
            if ( $output =~ /Error: Can't connect to KMIP server/ ) {
                logresult( "FAIL",
                    "KeyManager:keymanager_restore:\n Not able to connect to "
                        . $server_ip
                        . " key server" );
            } elsif ( $output =~ /No key servers registered./ ) {
                logresult( "PASS",
                    "KeyManager:keymanager_restore:\nThere are no key server defined"
                );
            } else {
                logresult( "PASS",
                    "KeyManager:keymanager_restore:\n$output" );
            }
        } ## end else [ if ( $num_servers == 0)
    } ## end foreach my $server_ip (@servers)
} ## end sub keymanager_restore

## @name ::keymanager_status
## @summary runs the CLI key_manager status command
##
## @description
## Reports the expected output of the command
##
## @return output
##
## @arg host Required: Hostrec object
##
## @arg kmipserver_ip1 Optional: The ip of the first KMIP server that has to be queried for status
## If not provided, it considers all the keyservers
##
## @arg kmipserver_ip2 Optional: The ip of the second KMIP server that has to be queried for status
## If not provided, it considers all the keyservers
##
## @arg kmipserver_ip3 Optional: The ip of the third KMIP server that has to be queried for status
## If not provided, it considers all the keyservers
##
## @arg kmipserver_ip4 Optional: The ip of the fourth KMIP server that has to be queried for status
## If not provided, it considers all the keyservers
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @examples
## keymanager_status(host=>$FILER, kmipserver_ip1 => '1.1.1.1');
## keymanager_status(host=>$FILER);
##
## @author niveetas@netapp.com
## @author dl-storsub-qa
## @status private (in-progress)

sub keymanager_status {
    my (%pars) = validate(
        @_,
        {   host           => { type => OBJECT, isa => 'Hostrec' },
            kmipserver_ip1 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip2 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip3 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip4 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            timeout => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
        }
    );

    my $apiset_obj = $pars{api_set_obj};
    my $cmode_bit = 0;
    my $apiset_created = 0;
    $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
    if((!$apiset_obj)||($apiset_obj->{actual_set} =~ /Nodescope/i)){
        $apiset_created = 1;
        # $filer_node = NACL::C::Node->new( hostrec => $pars{host} );
        # $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
        if ($filer_node->is_cmode){
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    set       => "CMode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
            my $output = $apiset_obj->execute_command(command => "run local");
            if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
                $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
                die();
            }
            $apiset_obj->execute_command(command => "priv set test");
            $cmode_bit = 1;
        } else {
            $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                    category  => "Node",
                                                    interface => "CLI",
                                                    connid    => "console",
                                                    set       => "7Mode");
            $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
        }
    }elsif($apiset_obj->{actual_set} =~ /CMode/i){
        my $output = $apiset_obj->execute_command(command => "run local");
        if ($output =~ /Too many users logged in\!\s+Please try again later\./i){
            $Log->comment("The filer console already in run local prompt so please manualy take back to CMode prompt and rerun the script");
            die();
        }
        $apiset_obj->execute_command(command => "priv set test");
        $cmode_bit = 1;
    }elsif($apiset_obj->{actual_set} =~ /7Mode/i){
        $apiset_created = 1;
        $apiset_obj = NACL::APISet->new(hostobj   => $pars{host},
                                                category  => "Node",
                                                interface => "CLI",
                                                connid    => "console",
                                                set       => "7Mode");
        $apiset_obj->set_timeout( 'connectrec-timeout' => 1800 );
    }

    my $cmd = "key_manager status";
    if ( $pars{kmipserver_ip1} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip1}";
    }
    if ( $pars{kmipserver_ip2} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip2}";
    }
    if ( $pars{kmipserver_ip3} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip3}";
    }
    if ( $pars{kmipserver_ip4} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip4}";
    }
    my $output = $apiset_obj->execute_command(
        'command'            => $cmd,
        'connectrec-timeout' => $pars{timeout}
    );
    if ( $output =~ /not a registered key server/ ) {
        logresult( "FAIL",
            "KeyManager:keymanager_status:\nThe key server is not registered"
        );
    } elsif ( $output =~ /No key servers registered./ ) {
        logresult( "PASS",
            "KeyManager:keymanager_status:\nThere are no key server defined"
        );
    } else {
        logresult( "PASS",
            "KeyManager:keymanager_status:\nThe set of key Servers and their status are: \n$output"
        );
    }
    # if($apiset_created){
        # $apiset_obj->cleanup_APISet();
    # }
    if ($cmode_bit){
        $apiset_obj->execute_command(command => "exit");
    }
    if($apiset_created){
        $apiset_obj->cleanup_APISet();
    }
    return ($output);
} ## end sub keymanager_status

## @name ::keymanager_sync
## @summary runs the CLI key_manager sync command
##
## @description
## Reports the expected output of the command
##
## @return output
##
## @arg host Required: Hostrec object
##
## @arg kmipserver_ip1 Optional: The ip of the first KMIP server that has to be synced
## If not provided, it considers all the keyservers
##
## @arg kmipserver_ip2 Optional: The ip of the second KMIP server that has to be synced
## If not provided, it considers all the keyservers
##
## @arg kmipserver_ip3 Optional: The ip of the third KMIP server that has to be synced
## If not provided, it considers all the keyservers
##
## @arg kmipserver_ip4 Optional: The ip of the fourth KMIP server that has to be synced
## If not provided, it considers all the keyservers
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## keymanager_sync(host=>$FILER, kmipserver_ip1 => '1.1.1.1');
## keymanager_sync(host=>$FILER);
##
## @author niveetas@netapp.com
## @status private (in-progress)

sub keymanager_sync {
    my (%pars) = validate(
        @_,
        {   host           => { type => OBJECT, isa => 'Hostrec' },
            kmipserver_ip1 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip2 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip3 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            kmipserver_ip4 => {
                type     => SCALAR,
                regex    => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/,
                optional => YES
            },
            conntype => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\#rsh|\#telnet|\#console/,
                default  => "#rsh"
            },
            timeout => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            waittime => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => WAIT_UNTIL_DONE
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
        }
    );
    my ( $match, $before, $after, $prompt, $output );
    my $promptstring = $pars{host}->{id};
    my $conn         = make_conn( $pars{host}, $pars{conntype} );
    my $cmd          = "key_manager sync";
    if ( $pars{kmipserver_ip1} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip1}";
    }
    if ( $pars{kmipserver_ip2} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip2}";
    }
    if ( $pars{kmipserver_ip3} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip3}";
    }
    if ( $pars{kmipserver_ip4} ) {
        $cmd = $cmd . " " . "-key_server $pars{kmipserver_ip4}";
    }

#    my (undef, undef, $output) = say($conn, $cmd , '-timeout', $pars{timeout}, '-prompt', $promptstring, '-waittime', $pars{waittime});
    ( undef, undef, $output )
        = say( $conn, $cmd, '-timeout', $pars{timeout}, '-prompt',
        $promptstring, '-waittime', $pars{waittime} );
    if ( $output =~ /not a registered key server/ ) {
        logresult( "FAIL",
            "KeyManager:keymanager_sync:\nThe key server is not registered" );
    } elsif ( $output =~ /No key servers registered./ ) {
        logresult( "PASS",
            "KeyManager:keymanager_sync:\nThere are no key server defined" );
    } else {
        logresult( "PASS", "KeyManager:keymanager_sync:\n$output" );
    }
} ## end sub keymanager_sync

## @name ::verify_uniquekey
## @summary it is used to verify that there are no duplicate key_id in the kmip_server
##
## @description
## Reports if there exist a duplicate key_id in the kmip_server or not
##
## @return duplicate key_id if there exist one
##
## @arg host Required: Hostrec object
##
## @arg kmipserver_ip Required: The ip of the KMIP server that has to be checked for duplicate key_id
##
## @arg conntype Optional: Takes the connection type provided.
## If not provided, rsh is the default connection
##
## @arg timeout Optional: Takes the timeout provided.
## If not provided, the default is 300
##
## @arg waittime Optional: Takes the waitime provided.
## If not provided, it waits till the job is done
##
## @examples
## verify_uniquekey(host=>$FILER, kmipserver_ip => '1.1.1.1');
##
## @author niveetas@netapp.com
## @status private (in-progress)

sub verify_uniquekey {
    my (%pars) = validate(
        @_,
        {   host => { type => OBJECT, isa => 'Hostrec' },
            kmipserver_ip =>
                { type => SCALAR, regex => qr/^\b(?:\d{1,3}\.){3}\d{1,3}\b/ },
            conntype => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\#rsh|\#telnet|\#console/,
                default  => "#rsh"
            },
            timeout => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => SAY_TIMEOUT
            },
            waittime => {
                type     => SCALAR,
                optional => YES,
                regex    => qr/\d+/,
                default  => WAIT_UNTIL_DONE
            },
            api_set_obj => {
                type => OBJECT,
                optional => YES,
            },
        }
    );

    my $conn = make_conn( $pars{host}, $pars{conntype} );
    my $cmd = "key_manager restore -key_server $pars{kmipserver_ip}"
        ; # TODO: assuming that key restore gives a list of key_id stored in the particular KMIP server
    my ( undef, undef, $output )
        = say( $conn, $cmd, '-timeout', $pars{timeout} );
    if ( $output =~ /not a registered key server/ ) {
        logresult( "FAIL",
            "KeyManager:keymanager_restore:\n $pars{kmipserver_ip} is not a registered key server"
        );
    } else {
        chomp($output);
        my @list = ();
        my %seen = ();
        my @uniq = ();
        @list = split( /\n/, $output );
        foreach my $item (@list) {
            unless ( $seen{$item} ) {
                $seen{$item} = 1;
            } else {
                logresult( "FATAL",
                    "There is a duplicate item in the list of key_id for $item"
                );
            }
        } ## end foreach my $item (@list)
    } ## end else [ if ( $output =~ /not a registered key server/)
} ## end sub verify_uniquekey

sub server_status {

    #TODO: check the status of the key_servers
}

sub get_passpharse {

    #TODO: get the cleartext password and the key_id from the KMIP server
}

sub certificate_setup {

    #TODO: setup a certificate between the filer and the KMIP server
}

#################################################################
# Internal subroutines
# 06/20/11 toc@netapp.com:  comment this testme out from keymanager.pm file
################################################################
#
#sub testme
#{
#    package main;
#    our ($FILER);
#    package Storage::FDE::KeyManager;
#    use Tharn qw(say logcomment logresult host param);
#    param('FILER', '-mesg', "Filer to test");
#    my $host_obj = host($FILER);
#    keymanager_add(host=>$host_obj, kmipserver_ip1 =>"172.18.99.175");
#    keymanager_show(host=>$host_obj);
#    keymanager_query(host=>$host_obj);
#    keymanager_restore(host=>$host_obj, kmipserver_ip =>"172.18.99.175");
#    keymanager_sync(host=>$host_obj, kmipserver_ip1 =>"172.18.99.175" );
#    keymanager_remove(host=>$host_obj, kmipserver_ip1 =>"172.18.99.175");
#    keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"172.18.99.175", quit2 => TRUE, enterpassphrase => TRUE, passphrase => "12345678901234567890", keytagname => "FDE");
#    keymanager_rekey(host=>$host_obj,manual=> TRUE, passphrase => "12345678901234567890");
#    keymanager_status(host=>$host_obj, kmipserver_ip1 =>"172.18.99.175" );
#    verify_uniquekey(host=>$host_obj, kmipserver_ip =>"172.18.99.175");
#    ## DONOT use this example as it locks all the drives when lockdrives => TRUE
#    ## keymanager_setup(host=>$host_obj, kmipserver_ip1 =>"0.0.0.0", kmipserver_ip2 =>"1.1.1.1", kmipserver_ip3 =>"2.2.2.2", kmipserver_ip4 =>"3.3.3.3", enterpassphrase => TRUE, passphrase => "testtesttesttesttesttest", keytagname => "FDE", lockdrives => TRUE);
#}
1;
