# _*_ emacs _*_
# $Id: //depot/prod/test/Rsierranevada/lib/Storage/Fiji.pm#2 $

## @summary Libray to set fiji error injection points
## the information of environment 
## @author yingz@netapp.com 
## 
## @change 10-01-2009 anishv: Added functions set_fiji_filter,check_fiji_filter and get_fiji_slot
## @change 10-01-2009 anishv: Modified fiji_clear and fiji_show to use filer_cmd. Now it should work for 10-mode as well
## @change 11-11-2009 anishv: Added check_fiji_hit function
## @pod here

=head1 NAME

Fiji - Fiji Commands Routines

=head1 SYNOPSIS

use Storage::Fiji qw(fiji_clear fiji_filter_sas_ses fiji_show set_fiji_filter check_fiji_filter get_fiji_slot); 

$s = fiji_show(%input);
    
=head1 DESCRIPTION

The library supports routines to inject fiji on different elements.  

The defined routines should work on filers.
  
Here is the list of provided routines are listed 
in alphabetical order. Please refer to each routine
description for details information.  

=head1 FUNCTION PROTOTYPES

$s = fiji_clear(%input)
 
=cut 

=head1 ROUTINES
 
=cut

package Storage::Fiji;

use strict;
use warnings;

use Carp qw(confess croak);
use Params::Validate qw(validate validate_pos OBJECT SCALAR ARRAYREF);
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}
use Utils qw(make_conn make_conn_and_complain); 
use Storage::SASUtils qw(filer_cmd); 
use Tharn; 
use Data::Dumper;
use RaidInfo;

use vars qw(@ISA @EXPORT_OK $VERBOSE); 

require Exporter; 
@ISA = qw(Exporter); 
@EXPORT_OK = qw(
		fiji_clear
		fiji_filter_sas_ses
		fiji_show
		$VERBOSE
        set_fiji_filter
        check_fiji_filter
        get_fiji_slot
        check_fiji_hit
		); 

our $MAINT_PROMPT = "\\*\>";

# ======================= DEFINE ROUTINES ===========================

=head1 fiji_clear(%input);

This sub-routine clears the fiji point.
command

=item Return

Return whether the action is successful.

=item host
(Required) : Host spec to the FILER

=item conntype
(Optional) : Connection type to the Filer. Default is rsh.

=item timeout
(Optional) : say timeout. Default is 300 secs.


=item fiji_point
(Optional) : fiji point to be cleared. Default is "all"

=item fiji_type
(Optional) : nvfilter or normal (filter)



=item Example: 
    
    $ret = fiji_clear(%input);

=item SEE ALSO
   None

=back

=cut

sub fiji_clear {
    my (%opts) = validate(@_, {
            host => { type=>OBJECT, isa=>'Hostrec' },
            filermode => { type=>SCALAR, default=>"normal", optional=>1 },
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            timeout => {type=>SCALAR,default=>"300",optional=>1},
            fiji_point => { type=>SCALAR,default=>"all",optional=>1 },
            filter_type => { type=>SCALAR,default=>"normal",optional=>1 },
            }
            );
    my $cmd;
    my $fiji_point = $opts{fiji_point};
    my $output;
    if($fiji_point eq "all"){
        if($opts{filter_type} eq "nvfilter"){
            $cmd = "fiji nvclear all";
        }
        else{
            $cmd = "fiji clear all";
        }

        $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>"$cmd");
    }
    else{

        my @fiji_slot = get_fiji_slot(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},fiji_point=>"$fiji_point");
        foreach my $slot (@fiji_slot){
            if($opts{filter_type} eq "nvfilter"){
                $cmd = "fiji nvclear $slot";
            }
            else{
                $cmd = "fiji clear $slot";
            }
            $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>"$cmd");
        }

    }

    if(check_fiji_filter(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},fiji_point=>$fiji_point)){
        logcomment("Fiji filter $fiji_point Not cleared");
        croak $output;
    } else {
        return 1;
    } 
}

=head2 fiji_show(%input);

The routine to show the current fiji in the system;

=item host
(Required) : Host spec to the FILER

=item conntype
(Optional) : Connection type to the Filer. Default is rsh.

=item Return

Return the hash of the environment information

=cut

=back

=item Example: 
    
    $ret = fiji_show(%input);

=item SEE ALSO
   None

=back

=cut
sub fiji_show {
    my (%opts) = validate(@_, {
            host => { type=>OBJECT, isa=>'Hostrec' },
            filermode => { type=>SCALAR, default=>"normal", optional=>1 },
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            timeout => {type=>SCALAR,default=>"300",optional=>1},
            }
            );
    my $cmd;
    my $output;
    $cmd = "fiji show";
    $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>"$cmd");
    return $output;
}

 
=head2 fiji_filter_sas_ses(%input);

The routine to inject the fiji in the system;

=item host
(Required) : Host spec to the FILER

=item conntype
(Optional) : Connection type to the Filer. Default is rsh.

=item Return

Return the hash of the environment information

=cut

=back

=item Example: 
    
    $ret = fiji_filter_sas_ses(%input);

=item SEE ALSO
   None

=back

=cut
sub fiji_filter_sas_ses(%) {
    my(%input) = validate ( @_, {
	      host => {type=>OBJECT, isa=>'Hostrec'},
	      conntype => {type=>SCALAR, default=>"#rsh", optional=>1},
	      timeout => {type=>SCALAR, default=>300, optional=>1},
	      shelfid => {type=>SCALAR},
	      channel => {type=>SCALAR},
	      element => {type=>SCALAR},
	      status => {type=>SCALAR},
	      repetition => {type=>SCALAR, default=>1, optional=>1},
	      number => {type=>SCALAR},
	      byte => {type=>SCALAR, default=>0, optional=>1},
	  }
			   );
    my $outputinfo;
    my $cmd;
    my $conn = make_conn($input{host},$input{conntype});
    my $byte_num = $input{byte};
    if ($byte_num != 0) {
	if ($input{repetition} !=1) {
	    $cmd = "priv set -q diag; fiji filter \"(repeat $input{repetition} (ses.status.page.byte$byte_num.$input{element}.sensor(&&(eq \@channel $input{channel})(== \@shelf $input{shelfid})(== \@element_number $input{number}))(return $input{status})))\" "; 
	} else {
	 $cmd = "priv set -q diag; fiji filter \"(ses.status.page.byte$byte_num.$input{element}.sensor(&&(eq \@channel $input{channel})(== \@shelf $input{shelfid})(== \@element_number $input{number}))(return $input{status}))\" ";
	}
    } else {
	if ($input{repetition} != 1) {
	    $cmd = "priv set -q diag; fiji filter \"(repeat $input{repetition} (ses.status.page.common.$input{element}.sensor(&&(eq \@channel $input{channel})(== \@shelf $input{shelfid})(== \@element_number $input{number}))(return $input{status})))\" ";
	} else {
	    $cmd = "priv set -q diag; fiji filter \"(ses.status.page.common.$input{element}.sensor(&&(eq \@channel $input{channel})(== \@shelf $input{shelfid})(== \@element_number $input{number}))(return $input{status}))\" ";
	    
	}
    }
    (undef, undef, $outputinfo) = say($conn, "$cmd", "-timeout", $input{timeout});
    if ($outputinfo =~ /filter set/) {
	logcomment ("Fiji is successful set on channel $input{channel}, shelf $input{shelfid} on $input{element}");
	return 1;
    } else {
	croak $outputinfo;
    }
}


=head2 set_fiji_filter

set_fiji_filter(host=>$host, filermode=>$filermode, conntype=>$conntype, fiji_point=>$fiji_point,timeout=>$timeout);

This subroutine add a defined fiji filter to the filer.

Return 1 on success and 0 on failure.

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item filermode

(Optional) : Specifies to run tests in normal or maintenance mode. Default is normal.
For maintenance mode the variable will be "MAINT"

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item fiji_point

(Required): Name of the fiji_point which we want to add to the filer.

=item timeout

(Optional): Timeout to say command. Default 300 secs


=back

=cut

#------------------------------------------------------------------------------

sub set_fiji_filter {
    my (%opts) = validate(@_, {
            host => { type=>OBJECT, isa=>'Hostrec' },
            filermode => { type=>SCALAR, default=>"normal", optional=>1 },
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            timeout => {type=>SCALAR,default=>"300",optional=>1},
            filter_type => {type=>SCALAR,default=>"normal",optional=>1},
            fiji_point => { type=>SCALAR },
            }
            );
    my $fiji_point  = $opts{fiji_point};
    my $cmd;

    if($opts{filter_type} eq "nvfilter"){
        $cmd = "fiji nvfilter \'$fiji_point\'";
    }
    else{
        $cmd = "fiji filter \'$fiji_point\'";
    }

    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>"$cmd");

    if(check_fiji_filter(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},fiji_point=>$fiji_point)){
        return 1;
    } else {
        logcomment("Fiji filter \"  $fiji_point  \" not set in the fiji show");
        croak $output;
    } 

}

=head2 check_fiji_filter

check_fiji_filter(host=>$host, filermode=>$filermode, conntype=>$conntype, fiji_point=>$fiji_point,timeout=>$timeout);

This subroutine check that filer is having fiji set.

It returns 1 on success and 0 on failure

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item filermode

(Optional) : Specifies to run tests in normal or maintenance mode. Default is normal.
For maintenance mode the variable will be "MAINT"

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item fiji_point

(Required): Name of the fiji_point for which we want to check fiji.

=item timeout

(Optional): Timeout for say command. Default is 300 secs.


=back

=cut

#------------------------------------------------------------------------------
sub check_fiji_filter {
    my (%opts) = validate(@_, {
            host => { type=>OBJECT, isa=>'Hostrec' },
            filermode => { type=>SCALAR, default=>"normal", optional=>1 },
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            timeout => {type=>SCALAR,default=>"300",optional=>1},
            fiji_point => { type=>SCALAR },
            }
            );
    my $fiji_point  = $opts{fiji_point};
    my $cmd = "fiji show";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>"$cmd");
    #print "Output is $output \n";
    #print "FIJI point is  is $fiji_point \n";
    if(index($output, $fiji_point) ne "-1") {
       # logcomment("fiji filter/nvfilter $fiji_point is set\n");
        return 1;
    } else {
       # logcomment("fiji filter/nvfilter $fiji_point is not set\n");
        return 0;
    }
}

#------------------------------------------------------------------------------

=head2 get_fiji_slot 

get_fiji_slot(host=>$host, filermode=>$filermode, conntype=>$conntype, fiji_point=>$fiji_point,timeout=>$timeout);

This subroutine get the fiji slot numbers for a particular fiji point.

It returns an array on success 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item filermode

(Optional) : Specifies to run tests in normal or maintenance mode. Default is normal.
For maintenance mode the variable will be "MAINT"

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item fiji_point

(Required) : Name of fiji_point .

=back

=cut

#------------------------------------------------------------------------------

sub get_fiji_slot {
    my (%opts) = validate(@_, {
            host => { type=>OBJECT, isa=>'Hostrec' },
            filermode => { type=>SCALAR, default=>"normal", optional=>1 },
            conntype => {type=>SCALAR, default=>"#rsh", optional=>1},
            timeout => {type=>SCALAR, default=>"300", optional=>1},
            fiji_point => { type=>SCALAR },
            }
            );
    my $fiji_point = $opts{fiji_point};

    my $n=0;
    my @fiji_slot;
    my $cmd = "fiji show";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>"$cmd");
    print "Output is $output \n";
    my @info = split(/\r*\n/, $output);
    foreach my  $line (@info) {
        my (@data) = split(": ", $line);
        if (index($line, "$fiji_point") ne "-1") {
            if (index($data[0], " (nvram)") ne "-1") {
                substr($data[0], index($data[0], " (nvram)"))="";
            }
            $fiji_slot[$n]=$data[0];
            $n++;
        }
    }
    return @fiji_slot;
}


=head1 check_fiji_hit(%input);

This sub-routine uses fiji dump command to find out if we have hit the fiji point for a particular slot
command
Returns 1 if it has hit the fiji point and 0 if not.

=item host
(Required) : Host spec to the FILER

=item conntype
(Optional) : Connection type to the Filer. Default is rsh.

=item timeout
(Optional) : say timeout. Default is 300 secs.


=item slot
(required) : Slot id for which we want to check the hit

=back

=cut

sub check_fiji_hit {
    my (%opts) = validate(@_, {
            host => { type=>OBJECT, isa=>'Hostrec' },
            filermode => { type=>SCALAR, default=>"normal", optional=>1 },
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            timeout => {type=>SCALAR,default=>"300",optional=>1},
            slot => { type=>SCALAR},
            }
            );
    my $output;
    my $slot1;
    my $IN = 0;
    my $cmd = "fiji dump";
        $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>"$cmd");
        my @op = split(/\r*\n/,$output);
        foreach my $op (@op){
            if($op =~ /^\d+:/){
                $slot1 = $op;
                $slot1 =~ s/://g;
            }
            if($slot1 eq "$opts{slot}"){
                $IN = 1;
                if($op =~ /--+>/){
                    return 0;
                }
            }
        }
        if($IN){ 
            return 1;
        }
        else{
            logcomment("Slot $opts{slot} not in the fiji dump");
            return 0;
        }
}


=head1 SEE ALSO

    None

=head1 BUGS

    None

=head1 STATUS

    Testing

=head1 ATTRIBUTES

    ONTAP_COMMANDS = fiji

=head1 AUTHOR

    Ying Zhang (yingz@netapp.com)
    $Id: //depot/prod/test/Rsierranevada/lib/Storage/Fiji.pm#2 $
    Last revised $Date: 2013/02/28 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;

