# $Id: //depot/prod/test/Rsierranevada/lib/Storage/Misc.pm#9 $

## @summary Library for misc. test, internal use only, filer commands and queries.
## @author schulz@netapp.com, dl-step-internal@netapp.com, dl-ssg-testdev@netapp.com
## @status testme
### @status public
## @pod here
# Have tharndoc use POD:
# - @summary, @author, @status still used when indexing
# - Lines below beginning with "##<any non-blank text>" will give tharndoc error; ### OK
# - To display documentation:  tharndoc Misc.pm <==> perldoc -t Misc.pm

# Design assumptions, standards:
# - http://wikid.netapp.com/w/ShastaQAAutomation
# - original:  pass connection object, assume script already connected to filer
#   later:  host rec for future compatibility with ZAPI
# - only export symbols on request
# - return hashes, not objects
# - default timeout 300 if needed
#@change 29-03-2012 hinduja: Adding a support for eth_ctrl_sp.thpl script in filer_cmd subroutine
# @change 2013-04-10 villas: modified sub misc_disk_zero_spares_background to support cmode
# @change 2013-11-5 jolie: modified script to export subs properly.  Nothing was being exported.  See BURTs 698492,741127
=head1 NAME

Misc.pm

=head1 SYNOPSIS

Library for misc. test, internal use only, commands and system queries.

=for text
use Storage::Misc;

=head1 DESCRIPTION

Library for misc. test, internal use only, commands:

    panic, printflag, setflag, version

and system queries:

    is simulator

This library has a self test routine called "testme".

=head1 STATUS

Shared

=head1 ATTRIBUTES

=for text
ONTAP_COMMANDS = panic, printflag, setflag, version
ONTAP_ZAPI_APIS =

=head1 AUTHOR

=for text
Charles Schulz, schulz@netapp.com, dl-step-internal@netapp.com
$Id: //depot/prod/test/Rsierranevada/lib/Storage/Misc.pm#9 $
Last revised $Date: 2013/11/05 $, $Author: jolie $

=head1 COPYRIGHT

Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

package Storage::Misc;

# Standard libs
use strict;
use warnings;
use Carp qw(confess croak);
# http://search.cpan.org/~drolsky/Params-Validate-0.78/lib/Params/Validate.pm
use Params::Validate qw(validate validate_pos SCALAR OBJECT HASHREF);
BEGIN {
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_;
                chomp(@msgs);
                Carp::confess(Tharn::script_result(@msgs))
                }
    );
        use base qw(Exporter);
        our @EXPORT_OK = qw(
                        testme
                        misc_printflag
                        misc_setflag
                        misc_setflags
                        misc_panic
                        misc_is_cluster
                        misc_is_sim
                        misc_filer_version
                        misc_get_filer_non_root_vols
                        misc_disk_maint_start_again
                        misc_aggr_status_detected_admin_disk
                        misc_release_all_maintenance_disks
                        misc_aggr_status_detected_spare_disk
                        misc_raid_config_info_showfdr
                        misc_export_root_fs
                        misc_raid_media_scrub_enable
                        misc_disk_zero_spares_background
                        filer_cmd
                        client_cmd
                        rename_fw_files
                        ping_host
                        get_acp_modules_from_disk_hba_shelf_type
                        get_sas_shelves_from_disk_hba_acp_shelf_type
                        get_disks_from_disk_hba_acp_shelf_type
                        );      # export on request

}
use Data::Dumper;

use Net::Ping;


# NetApp libs
use Tharn;            # http://web.netapp.com/engineering/nate/releases/latest/html/Tharn.html
# //depot/prod/ontap/main/test/lib/*.pm
use Utils qw(make_conn);    # make a host connection with name of host, Hostspec, or Connspec
use boot;

use Storage::Disk
qw(
    disk_maint
    disk_scrub
);

use Sys::Hostname;

use NACL::APISet;
use NACL::C::Node;

our $MAINT_PROMPT = "\\*\>";
our $LIB_TIMEOUT = 300;

=head1 ROUTINES

=head2 testme

Self test routine for this library.  This routine determines filer configuration and whether
a console is in use to choose certain tests.
This routine works on these configurations:

=over

=item Running sim

Will not run panic tests as sim would dumpcore from trying to run two sims on same sim disks
(terminal running it is console).

=item Not running sim

Sim will be booted by this routine if a filer console connection spec. is defined.
Will not run rsh tests as the sim cannot network with itself.

=item HW

If a console is not connected, will run panic tests.
if a console is connected, will not run console tests as ntest would give a config error anyway.

=back

ARGUMENTS

None.

EXAMPLE

Use standard tester script:

    ntest PKGNAME=Storage::Misc SUBNAME=testme util/test_pkg.thpl
    ntest PKGNAME=Storage::Misc SUBNAME=testme util/test_pkg.thpl FILER=step-ata01

=cut

sub testme() {
    logcomment("Self test of Storage::Misc. \n");

    # Test POD
    _test_POD();

    # Filer connections
    param("FILER", "-default", "");
    my $host = host($::FILER);    # hostrec object
    my $b = connect("#console");    # connectrec object for boot connection
    #my $c = connect("#rsh");    # connectrec object for regular connection

    # Test if filer is bootable
    my ($boot, $sim) = _test_bootable();
    # Boot if needed
    if ($boot) { boot($b) }

    # Test subs with rsh and console connections
    my ($rc, @rc);    # return codes
    my $co;        # connection object
    foreach my $co_type ("#rsh", "#console") {
        logcomment("Testing subs with connection type $co_type...\n");
        if ((!$boot) && $co_type eq "#console") {
            logcomment("Sim already has terminal console, can't use this type as console test -- " .
                "skipping subs tests.");
            next;
        } elsif ($boot && $sim && $co_type eq "#rsh") {
            logcomment("Can't rsh to itself on sim which was booted via console connection -- " .
                "skipping subs tests.");
            # and can't switch connections or it loses the sim
            next;
        }

        # Test misc_panic() if HW or bootable sim
        logcomment("Testing misc_panic... \n");
        # Need console connection to test or can't reboot
        if ($boot && $co_type eq "#console") {
            eval { $rc = misc_panic(host => $host, conntype => $co_type, option => "-D") };
            if ($@) { logresult("FAIL", $@) } else { logresult("PASS", "Returns:", Dumper($rc)) }
            boot($b);

            eval { $rc = misc_panic(host => $host, conntype => $co_type) };
            if ($@) { logresult("FAIL", $@) } else { logresult("PASS", "Returns:", Dumper($rc)) }
            boot($b);
        } else {
            logcomment("Need console connection to test panic so can reboot -- " .
                "skipping panic/boot tests.");
        }

        logcomment("Testing misc_is_sim... \n");    # no croak, so no eval
        my $simrc = misc_is_sim(host => $host, conntype => $co_type);
        if ($simrc)     { logcomment("Is simulator.") }
        else         { logcomment("Is not simulator.") }
        if ($simrc ne "") { logresult("PASS", "Returns:", Dumper($sim)) }

        logcomment("Testing misc_filer_version... \n");    # no croak, so no eval
        my $ver = misc_filer_version(host => $host, conntype => $co_type);    # returns hash ref
        if (defined($ver->{debug})) {
            logcomment("This is a debug build.");
        } else {
            logcomment("This is a non debug build.");
        }
        if ($ver ne "") { logresult("PASS", "Returns:", Dumper($ver))}

        logcomment("Testing misc_printflag... \n");
        my ($flag, $flags_hash);
        eval { $flag = misc_printflag(host => $host, conntype => $co_type, flag => "disk_verify_rate") };    # 1 flag
        _eval_log($flag);
        eval { $flag = misc_printflag(host => $host, conntype => $co_type, flag => "disk_verify_rat") };    # bad flag
        _eval_log($flag);
        eval { $flags_hash = misc_printflag(host => $host, conntype => $co_type) };    # all flags
        _eval_log($flags_hash);
        #logcomment("flag wafl_read_ahead:  " . $flags_hash->{wafl_read_ahead});    # lookup flag
        #logcomment("flag wafl_read_ahead:  " . $$flags_hash{wafl_read_ahead});
        # do 2 flags after all, so setflags below will only set 2 not all
        eval { $flags_hash = misc_printflag(host => $host, conntype => $co_type, flag => "disk_verify_rate disk_sort_disable") };    # 2 flags
        _eval_log($flags_hash);
        #logcomment("Flag values:  " . join(", ", values(%$flags_hash)));

        logcomment("Testing misc_setflag... \n");
        eval { $rc = misc_setflag(host => $host, conntype => $co_type, flag => "disk_sort_disable", value => 1) }; # default 0
        _eval_log($rc);
        eval { $rc = misc_setflag(host => $host, conntype => $co_type, flag => "disk_sort_disable", value => 0) }; # default 0
        _eval_log($rc);
        eval { $rc = misc_setflag(host => $host, conntype => $co_type, flag => "disk_verify_rate",  value => 300) };
        # default 600, can't always change?
        _eval_log($rc);
        eval { $rc = misc_setflag(host => $host, conntype => $co_type, flag => "disk_verify_rat",   value => 300) }; # bad flag
        _eval_log($rc);

        logcomment("Testing misc_setflags... \n");
        eval { @rc = misc_setflags(host => $host, conntype => $co_type, flags => $flags_hash) }; # 2 flags
        _eval_log(@rc);

    } # foreach connection

    # shutdown w/o core if booted sim
    if ($boot && $sim) { halt($b) }
}

# Local subroutines -----------------------------------------------------------

# Test POD for this lib.
sub _test_POD () {
    logcomment("Testing POD...");
    param("ONTAPDIR", "-default", "");
    param("LOGDIR",   "-default", "");
    my $infile = $::PKGNAME;
    $infile =~ s/::/\//g;
    my $outfile = `basename $infile`;
    $infile = "$::ONTAPDIR/test/lib/$infile.pm";
    chomp($outfile);
    $outfile = "$::LOGDIR/$outfile.html";
    # scalar context to avoid extra blank lines from list + \n's
    logcomment("Writing POD text into log...");
    logcomment("" . `perldoc -t $infile`);
    logcomment("Writing POD html file to log dir ($outfile)...");
    `pod2html -infile=$infile -outfile=$outfile -title=$::PKGNAME -htmldir=$::LOGDIR -htmlroot=$::LOGDIR`;
}

# Test if filer is bootable.
# Returns:  1 or 0 for bootable, 1 or 0 for sim.
# Use to avoid:  boot tests on sim already running (terminal is console), rsh after booting sims
sub _test_bootable () {
    logcomment("Checking if filer is bootable...");
    my ($boot, $sim);
    param("FILER", "-default", "");
    my $hostobj  = host($::FILER);
    my $filername = $hostobj->hostname();
    my $filertype = $hostobj->hosttype();
    $sim = 0 + $filertype =~ "simulator";
    logcomment("Filer $filername type is $filertype.");
    my $ntesthost = `hostname`;
    #my $h = connect("$ntesthost; hosttype=unix-solaris#rsh");    # use CLIENT if present
    param('CLIENT', '-default', $ntesthost);
    my $h = connect("$::CLIENT#rsh");
    my (undef, undef, $output) = say($h, "ping $filername 2");
    if ($output !~ "alive") {
        logcomment("Filer not up, trying to boot.");
        $boot = 1;    # need to boot sim or HW to run tests
    } elsif ($filertype =~ "simulator") {
        logcomment("Sim already running, assuming terminal console, not bootable.");
        $boot = 0;    # already has terminal console, so can't boot
    } else {
        logcomment("HW, bootable.");
        $boot = 1;    # HW can boot
    }
    return ($boot, $sim);
}

# Log and croak on privilege error.
# Promoted version shows admin, advanced privileges; diag not shown but can be used; no test.
# Un-promoted debug version adds test privilege, and shows all.
# croak shows die from perspective of lib (package) caller, not sub here.
# Arg. is say command output string.
sub _priv_croak ($) {
    my ($output) = @_;
    if ($output =~ /unrecognized privilege/) {
        my $emsg = "Privilege level and command not available for this version:";
        logcomment($emsg, $output);
        croak($emsg);
    }
}

# Log eval results.
# Arg. is data object to dump.
sub _eval_log (@) {
    if ($@) { logresult("FAIL", $@) }
    else     { logresult("PASS", "Returns: ", Dumper(@_)) }
}

#------------------------------------------------------------------------------

=head2 misc_printflag

Prints debugging flags.

RETURNS

    Value for one flag
    Reference to data hash for multiple flags
    Null on failure (to distinguish from 0 flag value)

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item conn

(Optional):  Named connection type hash.

=item flags

(Optional):  Named string of flags to return values.  Multiple flags are separated by spaces.
Default null returns all flag values.

=item privmode

(optional) : "priv set" mode for the flag. Default is "diag".
In case flag is visible only in test mode then we need to use "test".

=back

EXAMPLE

    $host = host($FILER);
    my $flag = printflag(host => $host, flags => "disk_verify_rate");
    my $flags_hash = printflag(host => $host, flags => "disk_verify_rate disk_sort_disable");
    my $flags_hash = printflag(host => $host);

=cut

# man na_printflag/setflag doesn't have details of debugging flags' use.
# printflag command can print for multiple flags separated by spaces (not documented).
# printflag works in maint. mode.
sub misc_printflag {
    my (%opts) = validate(@_, {
        host      => { type => OBJECT, isa    => 'Hostrec' },
        conntype => { type => SCALAR, default    => "#rsh",     optional => 1 },
        flag     => { type => SCALAR, default    => "",        optional => 1 },
        privmode => { type => SCALAR, default    => "diag",    optional => 1 },
        });
    my $conn = make_conn($opts{host}, $opts{conntype});

    my $flag = $opts{flag};
    my (undef, undef, $output) = say($conn, "priv set -q $opts{privmode}; printflag $flag");
    # parse output and fill hash
    my %flags = ();
    my (@output, $line);
    @output = split("\r*\n", $output);
    foreach $line (@output) {
        _priv_croak($output);
        if ($line =~ /no such flag: (.*)/) {
            logcomment("Unknown flag $1.");
            $flags{$1} = "";
        }
        # flag = value (may be trailing blank)
        elsif ($line =~ /(.*) = ([^\s][^\s]*)/) {
            $flags{$1} = $2;
        }
    }
    if (1 == scalar(@output)) {
        return $flags{$flag};    # single flag, just return value
    } else {
        return \%flags;        # multiple flags, return data structure ref.
    }
}
#------------------------------------------------------------------------------

=head2 misc_setflag

Sets a debugging flag.

RETURNS

    1 on success
    0 on failure (some flags cannot be changed, or get automatically reset?)

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item conn

(Optional):  Named connection type hash.

=item flag

(Required):  Named single flag to set.

=item value

(Required):  Named value to set for flag.

=item privmode

(optional) : "priv set" mode for the flag. Default is "diag".
In case flag is visible only in test mode then we need to use "test".

=back

EXAMPLE

    $host = host($FILER);
    my $rc = setflag(host => $host, flag => "disk_verify_rate", $value => 300);

=cut

# man na_printflag/setflag doesn't have details of debugging flags' use.
# setflag works in maint. mode.

# SPARSE/Resource/Filer.pm has setflag():
# - comments it should be in filer.pm,
# - uses non-standard connection object, filer say, logging
# - no printflag, no return code
# - args don't match prototype, uses has for multi-set, but not seeing that in use
sub misc_setflag {
    my (%opts) = validate(@_, {
        host      => { type => OBJECT, isa    => 'Hostrec' },
        conntype => { type => SCALAR, default    => "#rsh",     optional => 1 },
        flag     => { type => SCALAR },
        value    => { type => SCALAR },
        privmode => { type => SCALAR, default    => "diag",    optional => 1 },
        });
    my $conn = make_conn($opts{host}, $opts{conntype});

    my ($flag, $value) = ($opts{flag}, $opts{value});
    my (undef, undef, $output) = say($conn, "priv set -q $opts{privmode}; setflag $flag $value");
    _priv_croak($output);
    if ($output =~ /no such flag/) {
        logcomment("Unknown flag $flag.");
        return 0;
    }
    # verify -- not all flags are setable?
    my $setval = misc_printflag(host => $opts{host}, conntype => $opts{conntype}, flag => $flag, privmode => $opts{privmode});
    if ($setval eq $value) {
        return 1;
    } else {
        return 0;
    }
}
#------------------------------------------------------------------------------

=head2 misc_setflags

Sets multiple debugging flags.

RETURNS

List of return codes for each flag:

    1 on success
    0 on failure (some flags cannot be changed, or get automatically reset?)

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item conn

(Optional):  Named connection type hash.

=item flags_hash

(Required):  Reference to hash of flags and values, as produced by printflag().

=item privmode

(optional) : "priv set" mode for the flags. Default is "diag".
In case flags are visible only in test mode then we need to use "test".

=back

EXAMPLE

    $host = host($FILER);
    my $flags_hash = misc_printflag(host => $host, flags => "disk_verify_rate disk_sort_disable");
    my @rc = misc_setflags(host => $host, flags => $flags_hash);

=cut

# man na_printflag/setflag doesn't have details of debugging flags' use.
# setflag works in maint. mode.
sub misc_setflags {
    my (%opts) = validate(@_, {
        host      => { type => OBJECT, isa    => 'Hostrec' },
        conntype => { type => SCALAR, default    => "#rsh",     optional => 1 },
        flags    => { type => HASHREF },
        privmode => { type => SCALAR, default    => "diag",    optional => 1 },
        });
    my $conn = make_conn($opts{host}, $opts{conntype});

    my @rc = ();
    my ($rc, $flag, $value);
    foreach $flag (sort keys %{$opts{flags}}) {
        $value = ${$opts{flags}}{$flag};
        my (undef, undef, $output) = say($conn, "priv set -q $opts{privmode}; setflag $flag $value");
        _priv_croak($output);
        if ($output =~ /no such flag/) {
            logcomment("Unknown flag $flag.");
            $rc = 0;
        } else {
            # verify -- not all flags are setable?
            my $setval = misc_printflag(host => $opts{host}, conntype => $opts{conntype}, flag => $flag, privmode => $opts{privmode});
            if ($setval eq $value) {
                $rc = 1;
            } else {
                $rc = 0;
            }
        }
        @rc = (@rc, $rc);
    }
    return @rc;
}
#------------------------------------------------------------------------------

=head2 misc_panic

Bring the system down, bypassing standard shutdown procedures, dumping core.
Panic -D option skips core dump (not documented in man na_panic).
Only works in normal mode, not maintenance mode.
Can panic from non-console, but then won't see shutdown, and can't boot.
-D only works with test privilege versions (debug, not promoted).

RETURNS

    1 on success
    0 on failure

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item conn

(Optional):  Named connection type hash.

=item option

(Optional):  Named option -D to skip core dump.

=back

EXAMPLE

    $host = host($FILER);
    panic(host => $host);
    panic(host => $host, option => "-D");

=cut

# panic -D option not documented in man na_panic.
# panic only works in normal mode.
# No CLI argument help for panic:  "panic --" just performs the panic.
# Behavior, messages:
# - panic
#   same as "reboot -d panic" but man na_panic says use "halt -d <string>"
#    PANIC: reboot -d panic without any arguments
#   HW
#    DUMPCORE: START
#    DUMPCORE: END -- coredump written.
#   sim
#    Abort (core dumped)
# - panic -d
#   any option besides -D just aborts
#    Wait...Filer will panic at the end of the next CP...
#    PANIC: CP Panicking as requested for test purposes
#   HW
#    DUMPCORE: START
#    DUMPCORE: END -- coredump *NOT* written.
#   sim
#    Abort (core dumped)
# - panic -D
#    PANIC: panic without core
#   HW
#    *** Warm reboot...
#   sim
#    (...nothing...)

# filer.pm filer_panic() sleeps 320, no -D
# -D only works with priv test, not diag.
sub misc_panic {
    my (%opts) = validate(@_, {
        host      => { type => OBJECT, isa    => 'Hostrec' },
        conntype => { type => SCALAR, default    => "",     optional => 1 },
        option   => { type => SCALAR, default    => "",     optional => 1 },
        });
    my $conn = make_conn($opts{host}, $opts{conntype});
    my $option = $opts{option};

    my $priv = ("diag", "test")[0 + ($option eq "-D")]; # -D only works with test
    my $timeout;
    # default say timeout is 120 sec.;
    # need more for real filer panic since it automatically reboots also
    if (misc_is_sim(host => $opts{host}, conntype => $opts{conntype})) {
        $timeout = 120;        # default
    } else {
        $timeout = 360;        # filer
    }
    #filer.pm filer_panic() sleeps 320, no -D, uses -blind on say so no confirmation?
    my (undef, undef, $output) = say($conn, "priv set -q $priv; panic $option", -timeout, $timeout);
    #logcomment $output;     # notes on behavior, messages above
    _priv_croak($output);
    if ($output =~ /PANIC: /) {
        return 1;
    } elsif ($output =~ /Wait\.\.\.Filer will panic/) {
        sleep(30);    # wait for panic to start after next CP
        return 1;
    } else {
        return 0;
    }
}
#------------------------------------------------------------------------------

=head2 misc_is_cluster

Checks if running on cluster.

RETURNS

    1 for cluster
    0 for not (HW)

ARGUMENTS

=over

=item hosta

(Required):  Named host object hash.

=item hostb

(Required):  Named host object hash.

=back

EXAMPLE

    $hosta = host($FILERA);
    $hostb = host($FILERB);
    $sim = misc_is_cluster(hosta => $hosta, hostb => $hostb);

=cut

# Could check via ntest hostrec hosttype method, but this checks via filer.
# version works in maint. mode.
sub misc_is_cluster {
    my (%opts) = validate(@_, {
        hosta      => { type => OBJECT, isa    => 'Hostrec',  optional => 1 },
        hostb      => { type => OBJECT, isa    => 'Hostrec',  optional => 1 },
        });

    my $TIMEOUT = 240;
    my $hosta = $opts{'hosta'};
    my $hostb = $opts{'hostb'};

    if ((! $hosta) || (! $hostb)){
        return 0;
    }

    foreach my $host ($hosta, $hostb){
    my $fc = make_conn($host, "#rsh");
        my (undef, undef, $output) = say($fc, "cf status", '-timeout', $TIMEOUT);

        if ($output =~ /Cluster enabled/ ||
            $output =~ /Cluster disabled/ ||
            $output =~ /has disabled takeover by/ ||
            $output =~ /attempting takeover/ ||
            $output =~ /giveback in progress/ ) {
        next;
        } else {
            return 0;
        }
    }

    return 1;

}
#------------------------------------------------------------------------------

=head2 misc_is_sim

Checks if running on simulator.  Use to check behavior differences in some commands
such as timeouts or certain features not available.

RETURNS

    1 for simulator
    0 for not (HW)

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item conn

(Optional):  Named connection type hash.

=back

EXAMPLE

    $host = host($FILER);
    $sim = misc_is_sim(host => $host);

=cut

# Could check via ntest hostrec hosttype method, but this checks via filer.
# version works in maint. mode.
sub misc_is_sim {
    my (%opts) = validate(@_, {
        host      => { type => OBJECT, isa    => 'Hostrec' },
        conntype => { type => SCALAR, default    => "#rsh",     optional => 1 },
        });
    my $conn = make_conn($opts{host}, $opts{conntype});

    # -v to always show compile flags, incl. promoted builds
    # Flags:  see /rlse/mainN/prod/build/build -h
    # Sims:  t = Sun, L = Linux, R = FreeBSD, o = UNIX/Alpha (obsolete), P = Linux w. NGPT (obsolete)
    my (undef, undef, $output) = say($conn, "priv set -q admin; version -v");
    if ($output =~ /<.*[tLRoP].*>/) {
        return 1;
    } else {
        return 0;
    }
}
#------------------------------------------------------------------------------
=head2 misc_filer_version

Gets release, vendor, build time, debug from filer version.

RETURNS

    For a debug build:
    $VAR1 = {
    'release' => 'mainN_050910_2300',
    'buildTimestamp' => 'Sat Sep 10 23:42:26 PDT2005',
    'vendor' => 'NetApp',
    'debug' => '3d'
    };

    For a non-debug build:
    $VAR1 = {
    'release' => '7.0.3X5',
    'buildTimestamp' => 'Mon Nov 28 14:48:41 PST2005',
    'vendor' => 'NetApp',
    'debug' => undef
    };

EXAMPLE

To check if the build is debug or not:

    my $host = host($FILER);
    my $ver = filer_version(host => $host);
    if (defined($ver{debug})) {
       #This is a debug build
    } else {
       #This is a non debug build
    }

=cut

# Adjusted from code, examples by Sajan Agrawal.
# version works in maint. mode.
# Filer::Version does not work in maintenance mode, confirmed by Lincoln via Haripriya.
sub misc_filer_version {
    my (%opts) = validate(@_, {
        host       => { type => OBJECT, isa    => 'Hostrec' },
        filermode => { type => SCALAR, default    => "normal",     optional => 1 },
        conntype  => { type => SCALAR, default    => "#rsh",     optional => 1 },
        timeout   => { type => SCALAR, default    => $LIB_TIMEOUT, optional => 1 },
        });

    if ($opts{filermode} ne "normal") {
        $opts{conntype} = "#console";
    }
    my $conn = make_conn($opts{host}, $opts{conntype});

    my $output;
    my $cmd = "version -v"; # -v to always show compile flags, incl. promoted builds
    if ($opts{filermode} ne "normal") {
        (undef, undef, $output) = say($conn, "$cmd", -prompt, "$MAINT_PROMPT", "-timeout", $opts{timeout});
    } else {
        $output = Storage::SASUtils::filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);
#        (undef, undef, $output) = say($conn, "$cmd", "-timeout", $opts{timeout});
    }

    my %ver = ();
    my(@lines) = split(/\r*\n/, $output);     # cmd is echoed if it is sent over console connection
    foreach (@lines) {
        if ($_ =~ /Release/) {
            my (@v) = split(/\s+/);
            if ($_ =~ /IBM/) { #check for IBM release else Netapp.
                chop ($v[11]);
                substr($v[11], 0, 1) = "";
                $ver{vendor} = $v[11];
                chop($v[3]) if ($v[3] =~ /.*:$/);
                $ver{release} = $v[3];
                $ver{buildTimestamp} = "$v[4]" . " $v[5]" . " $v[6]" . " $v[7]" . " $v[8]" . "$v[9]";
                $ver{debug} = undef;
                if ($v[12] =~ /<(\w+)d/) {
                    $ver{debug} = $1 . "d";
                }
            } else {
                $ver{vendor} = $v[0];
                chop($v[2]) if ($v[2] =~ /.*:$/);
                $ver{release} = $v[2];
                $ver{buildTimestamp} = "$v[3]" . " $v[4]" . " $v[5]" . " $v[6]" . " $v[7]" . "$v[8]";
                $ver{debug} = undef;
                if ($v[9] =~ /<(\w+)d/||$v[10]=~ /<(\w+)d/ ) {
                    $ver{debug} = $1 . "d";
                }
            }
        }
    }

    return (\%ver);
}

#------------------------------------------------------------------------------
=head2 misc_get_filer_non_root_vols

Gets release, vendor, build time, debug from filer version.

RETURNS

EXAMPLE

To check if the build is debug or not:

    my $host = host($FILER);
    my @vols = misc_get_filer_non_root_vols();

=cut

#
#
#
sub misc_get_filer_non_root_vols{
    my (%opts) = validate(@_, {
        fc => { type => OBJECT, isa    => 'Connectrec' },
        timeout   => { type => SCALAR, default    => $LIB_TIMEOUT, optional => 1 },
        });

    my $fc = $opts{'fc'};
    my $TIMEOUT = $opts{'timeout'};
    my (undef,undef,$vol_status_out) =say($fc, "vol status", "-timeout", $TIMEOUT);
    chomp( my @vol_status = split(/\r?\n/,"$vol_status_out"));
    my $found_root = 0;
    my @vols = ();

    foreach my $line (@vol_status) {
        if ($line =~ /^\s*([\w_]+)\s+(online|offline|restricted|failed)\s+raid([\w_]+),\s+(flex|trad)\s+(root)*/){
            print "line: $line\nname:$1\nstate:$2\nraid_type:$3\ntype:$4\n";
            if ($5){
                $found_root = 1
            } else {
                push (@vols, $1);
            }
        }
    }


    # Let's make sure we found the root volume.
    if (! $found_root){
        print "found $FILER's vols: @vols\n";
        die "CONFIG: Couldn't identify root volume\n";
    }
    print "found vols: @vols\n";

    return (@vols);
}


#------------------------------------------------------------------------------

=head2 misc_disk_zero_spares_background

Start disk zero spares in the background

RETURNS

     nothing

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=back

EXAMPLE

    $host   = host($FILER);
    $fc     = connect($FILER_CONN);
    $output = misc_disk_zero_spares_background($fc);

=cut

sub misc_disk_zero_spares_background
{
    my ($fc, %pars)   = @_;
    my $apiset_obj = $pars{api_set_obj};
    if(!$apiset_obj){
        my $filer_node = NACL::C::Node->new( hostrec => $fc->hostp() );
        $apiset_obj = $filer_node->get_7m_or_nodescope_apiset();
    }
    my $output = "";
    # There is a diskZero_spares in Disk.pm but it waits for the task to complete.
    # I just want to fire this off as part of a cleanup routine and don't want
    # to wait for it to complete.
    logcomment("Initiating disk zero spares");
#    (undef, undef, $output) = say($fc, "disk zero spares");
   $apiset_obj->execute_command(command => "disk zero spares");
}

#------------------------------------------------------------------------------

=head2 misc_raid_media_scrub_enable

Set the raid.media_scrub.enable to off or on

RETURNS

    nada

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item option

(Required):  on or off

=back

EXAMPLE

    $host   = host($FILER);
    $fc     = connect($FILER_CONN);
    $output = misc_raid_media_scrub_enable($fc, "on");

=cut
sub misc_raid_media_scrub_enable
{
    # set raid.media_scrub.enable to off or on
    my ($fc, $option) = @_;
    say($fc,"options raid.media_scrub.enable $option");
}
#------------------------------------------------------------------------------

=head2 misc_export_root_fs

Export the root fs

RETURNS

    nada

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item root_vol

(Required):  The name of the root volume on the filer

=item timeout

(Required): A timeout value for say

=back

EXAMPLE

    param ('FILER_ROOT_VOL',     '-default', "vol0");
    param ('TIMEOUT',            '-default', "300");
    $host   = host($FILER);
    $fc     = connect($FILER_CONN);
    $output = misc_export_root_fs($fc, $FILER_ROOT_VOL, $TIMEOUT);

=cut

sub misc_export_root_fs
{
    # Export the root file system from the filer
    my ($fc, $root_vol, $timeout) = @_;
    say($fc, "priv set -q diag; exportfs -p sec=sys,rw,anon=0 /vol/$root_vol",
             '-timeout',  $timeout );
}

#------------------------------------------------------------------------------

=head2 misc_raid_config_info_showfdr

See if the disk is showing up in raid_config info showfdr

RETURNS

    true (TRUE)
    false (null value or 0)

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item disk_name

(Required):  Disk name

=item timeout

(Required):  Timeout value for say

=back

EXAMPLE

    $host   = host($FILER);
    $fc     = connect($FILER_CONN);
    $output = misc_raid_config_info_showfdr($fc, $disk_name, $timeout);

=cut
sub misc_raid_config_info_showfdr
{

    my ($fc, $disk_name, $timeout) = @_;

    logcomment("misc_raid_config_info_showfdr timeout = $timeout");
    $timeout = 300;

    my $device = "";
    my $found  = "";
    my $line   = "";
    my $output = "";

    (undef, undef, $output) = say($fc, "priv set -q diag; raid_config info showfdr", "-timeout" , $timeout);

    if ($output =~ /$disk_name/)
    {
        $found = "TRUE";

        logcomment("$disk_name found in raid_config info showfdr");
    }
    else
    {
        logcomment("$disk_name NOT found in raid_config info showfdr");
    }
    return $found;
}
#------------------------------------------------------------------------------

=head2 misc_aggr_status_detected_spare_disk

See if the disk is a pare

RETURNS

    true (TRUE)
    false (null value or 0)

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item disk_name

(Required):  Disk name

=back

EXAMPLE

    $host   = host($FILER);
    $fc     = connect($FILER_CONN);
    $output =  misc_aggr_status_detected_spare_disk($fc, $disk_name);

=cut
sub misc_aggr_status_detected_spare_disk
{
    my ($fc, $disk_name) = @_;

    die "aggr_status_detected_spare_disk : disk_name not defined!\n"
        if (! $disk_name);

    my $aggr_device;
    my $found  = "";
    my $line   = "";
    my $output = "";

    (undef, undef, $output) = say($fc, "aggr status -r");

    my @output = split(/\n/, $output);

    foreach $line (@output)
    {
        next if ($line !~ /^spare\s+(\S+)/);
        $aggr_device =$1;
        logcomment($line);
        if ($aggr_device =~ /$disk_name/)
        {
            $found = "TRUE";
            logcomment("Found device : $aggr_device");
        }
        last if ($found);
    }
    if ($found =~ /TRUE/)
    {
        logcomment("$disk_name found in aggr status -r Spare disks output");
    }
    else
    {
        logcomment("$disk_name NOT found in aggr status -r Spare Disks output");
    }
    return $found;
}
#------------------------------------------------------------------------------

=head2 misc_release_all_maintenance_disks

Release all the disks in the maint center

RETURNS

    No return value

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=back

EXAMPLE

    $host   = host($FILER);
    $fc     = connect($FILER_CONN);
    $output =  misc_release_all_maintenance_disks($fc);

=cut
sub misc_release_all_maintenance_disks
{
        my ($host, $fc) = @_;

        my $device;
        my $line;
        my $output;

        logcomment("Releasing all disks from the maintenance pool");

        (undef, undef, $output) = say($fc, "aggr status -r");

        my @output = split(/\n/, $output);

        foreach $line (@output)
        {
            next if ($line !~ /^admin testing\s+(\S+)/);

            $device =$1;

            logcomment("Releasing $device");

            #say($fc, "disk maint abort $aggr_device");

            $output = disk_maint( host     => $host,
                                  maintcmd => 'abort',
                                  disklist => $device
                                );

        }
}
#------------------------------------------------------------------------------

=head2 misc_aggr_status_detected_admin_disk

Check that aggr status detected an admin disk

RETURNS

    true (non zero value)
    false (0 value)


ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item disk_name

(Required):  Disk name

=back

EXAMPLE

    $host   = host($FILER);
    $fc     = connect($FILER_CONN);
    $output =  misc_aggr_status_detected_admin_disk($fc, $disk_name);

=cut
sub misc_aggr_status_detected_admin_disk
{
    my ($fc, $disk_name) = @_;

    my $aggr_device;
    my $found  = "";
    my $line;
    my $output;

    (undef, undef, $output, undef) = say($fc, "aggr status -r", "Maintenance disks");

    my @output = split(/\n/, $output);


    foreach $line (@output)
    {
        next if ($line !~ /^admin testing\s+(\S+)/);

        $aggr_device =$1;

        logcomment($line);

        if ($aggr_device =~ /$disk_name/)
        {
            $found = "TRUE";
        }
        last if ($found);
    }
    if ($found =~ /TRUE/)
    {
        logcomment("PASS $disk_name found in aggr status -r Maintenance disks output");

    } else {

        logcomment("FAIL $disk_name NOT found in aggr status -r Maintenance disks output");

    }
    return $found;
}
#------------------------------------------------------------------------------

=head2 misc_disk_maint_start_again

This routine resubmits disk to the disk maint routine.

RETURNS

    Output from disk maint

ARGUMENTS

=over

=item host

(Required):  Named host object hash.

=item disk_name

(Required):  Disk name to submit

=back

EXAMPLE

    $host = host($FILER);
    $fc   = connect($FILER_CONN);
    $output = misc_disk_maint_start_again($fc, $disk_name);

=cut

sub misc_disk_maint_start_again
{
    my($fc, $disk_name) = @_;
    # This is a negative test case. If I call the library directly
    # the library checks will circumvent the tests
    my $output;
    (undef, undef, $output) = say($fc, "disk maint start -f -d $disk_name");
    return $output;
}

=head2 filer_cmd

    my $ret = filer_cmd();

This routine is to issue command to filer and return the output

=over

=item host

(Required) : Filer connnection to use. Hostrec

=item filermode

(Optional) : The mode of filer: "normal" or "maint" or "ngsh". Default is "normal".

If filer in maintenance mode, the connection type will be "#console".

In case of "normal". If filer is running 10-mode (hostype filer-ng in the host file). It will use "nodescope". Otherwise use regular say via rsh.

filermode "ngsh" is not yet implemented

=item conntype

(Optional) : The connection type to be created by this function to talk to
the filer. Default is "#rsh".

=item timeout

(optional) : timeout value. Default is 300 secs.

=item privmode

(optional) : "priv set" mode for the command. Default is "diag". In case command is vaiable only in test mode then we need to use "test".

=item noprompt

(optional) : In case we just want to issue to command and don't want to wait for the prompt. Default is 0, i.e, it waits for the prompt.


=item cmd


(Required) : command to be issued.

=back

=cut


sub filer_cmd {
    my (%opts) = validate(@_,{
            host => {type=>OBJECT, isa=>'Hostrec'},
            node => {type=>HASHREF, default=>undef, optional=>1},
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            timeout => {type=>SCALAR, default=>300,optional=>1},
            ignore_error => {type=>SCALAR, default=>1,optional=>1},
            filermode => {type=>SCALAR,default=>"normal",optional=>1 },
            privmode => {type=>SCALAR,default=>"diag",optional=>1 },
            noprompt => {type=>SCALAR,default=>0,optional=>1 },
            cmd => {type=>SCALAR},
            sp_check => {type=>SCALAR,default=>0,optional=>1 },
            prompt_check => {type=>SCALAR,default=>0,optional=>1 }
            }
            );
    my $output;
    my $conn;
    if ($opts{filermode} eq "maint") {
        $opts{conntype} = "#console";
        $conn = make_conn($opts{host},$opts{conntype});
        my $cmd = "$opts{cmd}";
        if($opts{noprompt}){
            (undef, undef, $output) = say($conn, "$cmd", -noprompt , "-timeout", $opts{timeout});
        }
        else{
            (undef, undef, $output) = say($conn, "$cmd", -prompt, "$MAINT_PROMPT", "-timeout", $opts{timeout});
        }

    }
    elsif ($opts{filermode} eq "normal"){
        if($opts{host}->{hosttype} =~ /filer-ng/){
            my $server= new Server;
            if (defined $opts{node}) {
                $server->set_execution_server(-Server => $opts{node});
            } else {
                $server->set_execution_server(-Server => $opts{host}->{hostname});
            }

            my $cmd = "run local  \"priv set -q $opts{privmode};$opts{cmd}\"";
            if($opts{noprompt} && ($opts{conntype} eq "#rsh")){
                $output = $server->cli(cmd=>"$cmd",-Timeout=>$opts{timeout},-Ignore_errors=>$opts{ignore_error},noprompt=>1);
            }
            elsif($opts{noprompt} && ($opts{conntype} eq "#console")){
                $output = $server->console(cmd=>"$cmd",-Timeout=>$opts{timeout},-Ignore_errors=>$opts{ignore_error},noprompt=>1);
            }
            elsif($opts{conntype} eq "#console"){
                $output = $server->console(cmd=>"$cmd",-Timeout=>$opts{timeout},-Ignore_errors=>$opts{ignore_error});
            }
            else{
                $output = $server->cli(cmd=>"$cmd",-Timeout=>$opts{timeout},-Ignore_errors=>$opts{ignore_error});
            }
	    
        }
        else{
            $conn = make_conn($opts{host},$opts{conntype});
            my $cmd = "priv set -q $opts{privmode};$opts{cmd}";
            if($opts{noprompt}){
                (undef, undef, $output) = say($conn, "$cmd",-noprompt,"-timeout", $opts{timeout});
            }elsif($opts{sp_check}){
                my $cmd = "$opts{cmd}";
                if($opts{prompt_check}){
                    (undef, undef, undef, $output) = say($conn, "$cmd", -prompt, "$opts{sp_check}", "-timeout", $opts{timeout});
                } else{
                    (undef, undef, $output) = say($conn, "$cmd", -prompt, "$opts{sp_check}", "-timeout", $opts{timeout});
                }
            }
            else{
                (undef, undef, $output) = say($conn, "$cmd","-timeout", $opts{timeout});
            }

        }
    }
    elsif($opts{filermode} eq "ngsh"){
        logcomment("NOT YET IMPLEMENTED");
    }
    else{
        die "Unsupported filermode";
    }
    return $output;
}


=head2 client_cmd

    my $ret = client_cmd();

This routine is to issue command to client and return the output

=over

=item client

(Optional) : Client connnection to use. Hostrec. Default is return value of hostname() without domain.

=item conntype

(Optional) : The connection type to be created by this function to talk to
the client. Default is "#rlogin".

=item timeout

(optional) : timeout value. Default is 300 secs.

=item cmd

(Required) : command to be issued on client.

=back

=cut


sub client_cmd {
    my (%opts) = validate(@_,{
            #client => {type=>OBJECT, isa=>'Hostrec'},
            client => {type=>SCALAR,default=>undef,optional=>1},
            conntype => {type=>SCALAR,default=>undef,optional=>1},
        sudo => {type=>SCALAR,default=>undef,optional=>1},
            timeout => {type=>SCALAR, default=>300,optional=>1},
            cmd => {type=>SCALAR},
            }
            );

    my $client ;
    if(defined($opts{client})) {
        $client="$opts{client}";
    } else {
        my $c=hostname();
        my @ca = split(/\./,$c);
        $client = $ca[0];
       print "\n Hostname = $client\n";
    }
    my $sudo ;
    if(defined($opts{sudo})){
        $sudo="$opts{sudo}";
    } else {
        $sudo="/usr/local/bin/sudo";
    }

    #my $conn = "$opts{conntype}";
    my $conn;
    if(defined($opts{conntype})){
        $conn = make_conn($client, $opts{conntype});
    } else {
        $opts{conntype} = "#rlogin";
        $conn = make_conn($client, $opts{conntype});
    }

    my $cmd = "$opts{cmd}";

    my $output;

    (undef, undef, $output) = say($conn, "$sudo $cmd","-timeout", $opts{timeout});

    if($output) {
        return $output;
    } else {
        return 1;
    }
}


=head2 rename_shelf_fw_files

    my $ret = rename_shelf_fw_files();

This routine is to rename FW files on filer.

=over

=item host

(Required) : Filer connnection to use. Hostrec

=item filermode

(Optional) : The mode of filer: "normal" or "maint" or "ngsh". Default is "normal".

If filer in maintenance mode, the connection type will be "#console".

In case of "normal". If filer is running 10-mode (hostype filer-ng in the host file). It will use "nodescope". Otherwise use regular say via rsh.

filermode "ngsh" is not yet implemented

=item conntype

(Optional) : The connection type to be created by this function to talk to
the filer. Default is "#rsh".

=item timeout

(optional) : timeout value. Default is 300 secs.

=item privmode

(optional) : "priv set" mode for the command. Default is "diag". In case command is vaiable only in test mode then we need to use "test".

=item noprompt

(optional) : In case we just want to issue to command and don't want to wait for the prompt. Default is 0, i.e, it waits for the prompt.


=item cmd


(Required) : command to be issued.

=back

=cut


sub rename_fw_files {
    my (%opts) = validate(@_,{
            host => {type=>OBJECT, isa=>'Hostrec'},
            dir => {type=>SCALAR},
            #type => {type=>SCALAR,default=>undef,optional=>1,regex=>qr/^(disk|shelf)$/},
            type => {type=>SCALAR,regex=>qr/^(disk|shelf|acp)$/},
            #p_type => {type=>SCALAR,default=>undef,optional=>1,regex=>qr/^(ESH|ESH2|ESH4|AT-FCX|AT-FC|IOM3|LRC|R100Shelf|R1XXShelf|SAS|IOM6)$/ },
            p_type => {type=>SCALAR,default=>undef,optional=>1},
            major => {type=>SCALAR,default=>0,optional=>1},
            timeout => {type=>SCALAR, default=>300,optional=>1},
            #privmode => {type=>SCALAR,default=>"diag",optional=>1 },
            #noprompt => {type=>SCALAR,default=>0,optional=>1 },
            #cmd => {type=>SCALAR},
            }
            );

    my $dir = "$opts{dir}";
    #my $p_type = "$opts{p_type}";
    my $cmd = "ls /etc/$dir";

    my $list = filer_cmd (host=>$opts{host}, cmd=>$cmd, timeout=>$opts{timeout});

    if($list =~ /Can't open directory/) {
        my $emsg = "Directory /etc/$dir doesn't exist \n";
        croak($emsg);
    }

    my @fw_files = ();
    my @all_fw_files_1 = split(/\s\n+/,$list);
    my @all_fw_files = grep {!/FVF/} @all_fw_files_1; # Remove FVF file(s)

    if(defined $opts{p_type}) {
        @fw_files = grep (/$opts{p_type}/, @all_fw_files); # Need to handle ESHx, IOMx, AT-FCx later.
    } else {
        @fw_files = @all_fw_files;
    }

    #if(scalar @fw_files == 0) { ###
    if($#fw_files < 0) {
        my $emsg = "Directory /etc/$dir doesn't contain any file OR any of specified type $opts{p_type}.\n";
        croak($emsg);
    }

    my %fwfiles= ();
    my $type;
    my $na;
    my $na_sub;
    my $rev;
    my $fw_file;
    my $fw_new;
    my $rename_cmd;
    my $minor = 0;

    # sort firmware files in reverse order to ensure download picks up the highest version of firmware
    my @sorted_fw_files = sort {$b cmp $a} (@fw_files);

    foreach my $file(@sorted_fw_files){
        if($opts{type} =~ /shelf/) {
            if($file =~ /([^\.]+)\.([^\.]{2})([^\.]{2})\.SFW/){
                ($type, $na, $rev) = ($1, $2, $3);
                $fwfiles{$type} = [$file, $na, $rev];
                $fw_file = "/etc/$dir/$type\.$na" . sprintf($rev) . "\.SFW"; # Absolute path.

                if($file =~ /IOM/) {
                    $minor = 1; # Defaults to minor version bump.
                    if($opts{major}) {
                        $minor = 0;
                    }
                }

                if($minor) {
                     $fw_new = "/etc/$dir/$type\.$na" . sprintf(++$rev) . "\.SFW";
                } else {
                     $na_sub = substr($na,1,1);
                     $na = "Z";
                     $na .= $na_sub;
                     $fw_new = "/etc/$dir/$type\.$na" . sprintf($rev) . "\.SFW";
                }

                $rename_cmd = "mv $fw_file $fw_new";
                filer_cmd (host=>$opts{host}, cmd=>$rename_cmd,timeout=>$opts{timeout});
             }
        } elsif($opts{type} =~ /acp/) {
            if($file =~ /([^\.]+)\.([^\.]{2})([^\.]{2})\.AFW/){
                ($type, $na, $rev) = ($1, $2, $3);
                $fwfiles{$type} = [$file, $na, $rev];
                $fw_file = "/etc/$dir/$type\.$na" . sprintf($rev) . "\.AFW";

                if($file =~ /IOM/) {
                    $minor = 1;
                    if($opts{major}) {
                        $minor = 0;
                    }
                }

                if($minor) {
                     $fw_new = "/etc/$dir/$type\.$na" . sprintf(++$rev) . "\.AFW";
                } else {
                     $na_sub = substr($na,1,1);
                     $na = "Z";
                     $na .= $na_sub;
                     $fw_new = "/etc/$dir/$type\.$na" . sprintf($rev) . "\.AFW";
                }

                $rename_cmd = "mv $fw_file $fw_new";
                filer_cmd (host=>$opts{host}, cmd=>$rename_cmd,timeout=>$opts{timeout});
            }
        } else {
            if ($file =~ /([^\.]+)\.([^\.]{2})([^\.]{2})\.LOD/){
                ($type, $na, $rev) = ($1, $2, $3);
                $fwfiles{$type} = [$file, $na, $rev];
                $fw_file = "/etc/$dir/$type\.$na" . sprintf($rev) . "\.LOD";
                $rev = 99;
                $fw_new = "/etc/$dir/$type\.$na" . sprintf($rev) . "\.LOD";


                    $rename_cmd = "mv $fw_file $fw_new";
                    filer_cmd (host=>$opts{host}, cmd=>$rename_cmd,timeout=>$opts{timeout});
                }

        }
     }
     logcomment ("Firmware files are renamed.");
     return 1;
}

sub ping_host {


    my (%opts) = validate(@_,{
            host => {type=>SCALAR},
            retry => {type=>SCALAR,default=>3,optional=>1},
            }
            );
    my $retry = 0;
    my $p = Net::Ping->new()
        or die "Can't create new ping object: $!\n";
    while($retry < $opts{retry}){
        $retry++;
        if ($p->ping($opts{host})) {
            $p->close;
            print "$opts{host} is alive \n" ;
            return 1;
        }
        else {
            print "\n";
            print $p->ping($opts{host});

            print "Unable to ping $opts{host}  retry $retry\n";
        }
        sleep 5;
    }
    $p->close;
    return 0;
}

=head2 get_acp_modules_from_disk_hba_shelf_type

 my $ret = get_acp_modules_from_disk_hba_shelf_type(FILER => $FILER, disk_type => $DISK_TYPE, shelf_type => $SHELF_TYPE, hba_type => $HBA_TYPE, acp_type => $ACP_TYPE);
This subrutine returns an hash reference of acp modules in the system with specific parameters.

The following arguments are accepted:

=over

=item FILER
(Required): filer name

=item shelf_type

(Optional): Shelf types for which the acp modules has to be returned

=item hba_type

(Optional): HBA types for which the acp modules has to be returned

=item disk_type

(Optional): Disk types for which the acp modules has to be returned

=back

=cut

sub get_acp_modules_from_disk_hba_shelf_type {

        my (%opts) = validate(@_, {
            FILER => {type=>SCALAR},
            shelf_type => {type=>SCALAR,optional=>1},
            hba_type => {type=>SCALAR,optional=>1},
            disk_type => {type=>SCALAR,optional=>1},
        }
        );
                my @disk_types;
                my @search_results;
                my @hba_types;
                my @shelf_types;
                my $result;
                my $acp_modules;
                my $response;

    my $FILER = $opts{FILER};
    my $filer_ci = NACL::C::Node->find();
    my $api_set_obj = $filer_ci->get_7m_or_nodescope_apiset();

    if($opts{disk_type}) {
        my $test_disks;
        @disk_types = split(",",$opts{disk_type});
        $response = $api_set_obj->aggr_status("raid-disk" => 1);
        @search_results = $response->search('type' => join('|',@disk_types));
        foreach $result(@search_results){
            if($result->{raid_disk} ne "partner") {
                if($result->{"device"} =~ /(\w+\.\w+)\.\w+/) {
                    $test_disks->{$1} = 1;
                }

            }
        }
        my @test_disk = keys(%$test_disks);
        @search_results = ();
        foreach my $test_disk (@test_disk){
            $response = $api_set_obj->acpadmin_list_all();
            my @search_results = $response->search( inband_id => ".*" );
            foreach my $result (@search_results) {
                if ( $result->{iom_type} ne "IOM6E" ) {
                    if ( $result->{inband_id} =~ /(\S+)\.\S+/ ) {
                        if($1 =~ /$test_disk/){
                            $acp_modules->{$result->{iom_type}} = 1;
                        }
                    }
                }
            }
        }
        if(!keys(%$acp_modules)){
            die config_result("No acp module present for the disk type passed");
            exit(0);
        }
    }

    if ($opts{hba_type}) {
        foreach (split(',',$opts{hba_type})) {
            if($_ eq 'FC') {
                push(@hba_types,'.*Fibre Channel.*');
            } elsif ($_ eq 'SAS') {
                push(@hba_types,'.*SAS.*');
            } else {
                die config_result("Only allowed types of HBA are FC and SAS");
                exit(0);
            }
        }
        my $response = $api_set_obj->storage_show_adapter("all" => 1);
        my $channels;
        @search_results = ();
        @search_results = $response->search('description' => join('|',@hba_types));
        foreach (@search_results) {
            $channels->{$_->{slot}} = 1;
        }
        my @test_channels = keys(%$channels);
        print "The test channels : @test_channels";
        foreach my $test_channel (@test_channels){
            $response = $api_set_obj->acpadmin_list_all();
            @search_results = $response->search( inband_id => ".*" );
            foreach $result (@search_results) {
                if ( $result->{iom_type} ne "IOM6E" ) {
                    if ( $result->{inband_id} =~ /(\w+)\.\w+\.\w+/) {
                        if($1 =~ /$test_channel/){
                            $acp_modules->{$result->{iom_type}} = 1;
                        }
                    }
                }
            }
        }
        if(!keys(%$acp_modules)){
            die config_result("No acp module present for the hba type passed");
            exit(0);
        }       
    }

    if ($opts{shelf_type}) {
        @shelf_types = split(',',$opts{shelf_type});
        $response = $api_set_obj->environment_status_shelf();
        @search_results = ();
        @search_results = $response->search('product_identification' => join('|',@shelf_types));
        foreach (@search_results) {
            if ( $_->{module_type} =~/IOM6|IOM3/) {
                $acp_modules->{$_->{module_type}} = 1;
            }
        }
        if(!keys(%$acp_modules)){
            die config_result("No acp module present for the shelf type passed");
            exit(0);
        }       
    }
    
    return $acp_modules;
}


sub get_sas_shelves_from_disk_hba_acp_shelf_type {
    my (%opts) = validate(@_, {
            FILER => {type=>SCALAR},
            shelf_type => {type=>SCALAR,optional=>1},
            hba_type => {type=>SCALAR,optional=>1},
            acp_type => {type=>SCALAR,optional=>1},
            disk_type => {type=>SCALAR,optional=>1},
            }
            );
    my $output;
    my @disk_types;
    my @search_results;
    my $result;
    my @channels = ();
    my @shelf_types;
    my @hba_types;
    my @acp_types;
    my $test_shelves;
    my @shelfs;
    
    my $FILER = $opts{FILER};
    
    my $filer_ci = NACL::C::Node->find();
    my $api_set_obj = $filer_ci->get_7m_or_nodescope_apiset();
    
    
    
    if($opts{disk_type}) {
        my $output_aggr_status = $api_set_obj->aggr_status("raid-disk" => 1);
        @disk_types = split(",",$opts{disk_type});
        @search_results = $output_aggr_status->search('type' => join('|',@disk_types));
        foreach $result(@search_results) {
            if($result->{raid_disk} ne "partner" && $result->{device} =~ /(\w+\.\w+)\.\w+/) {
                  push(@shelfs, $1)
            }
        }
    }
    
     if($opts{hba_type}) {
        foreach (split(',',$opts{hba_type})) {
            if($_ eq 'FC') {
                push(@hba_types,'.*Fibre Channel.*');
            } elsif ($_ eq 'SAS') {
                push(@hba_types,'.*SAS.*');
            } else {
                die config_result("Only allowed types of HBA are FC and SAS");
                exit(0);
            }
        }
        $output = $api_set_obj->storage_show_adapter("all" => 1);
        @search_results = $output->search('description' => join('|',@hba_types));
        foreach (@search_results) {
            my $output_aggr_status = $api_set_obj->aggr_status("raid-disk" => 1);
            my @srch_rs = $output_aggr_status->search('ha' => $_->{slot});
            foreach $result(@srch_rs) {
                if($result->{raid_disk} ne "partner" && $result->{device} =~ /(\w+\.\w+)\.\w+/) {
                        push(@shelfs, $1)
                }
            }
        }
        
     }
     
     if ($opts{shelf_type}) {
        $output = $api_set_obj->environment_status_shelf();
        @search_results = ();
        @shelf_types = split(',',$opts{shelf_type});
        @search_results = $output->search('product_identification' => join('|',@shelf_types));
        
        foreach (@search_results) {
            my $output_aggr_status = $api_set_obj->aggr_status("raid-disk" => 1);
            my @srch_rs = $output_aggr_status->search('ha' => $_->{channel});
            foreach $result(@srch_rs) {
                if($result->{raid_disk} ne "partner" && $result->{device} =~ /(\w+\.\w+)\.\w+/) {
                    push(@shelfs, $1)
                }
            }
       }
    } 
    
    if ($opts{acp_type}) {
        @acp_types = split(',',$opts{acp_type});
        $output = $api_set_obj->storage_show_acp("all" => 1);
        @search_results = $output->search('acp_status' => ".*");
    
        if($search_results[0]->{acp_status} eq "Active") {
            @search_results = $output->search('module_type' => join('|',@acp_types));
            foreach (@search_results) {
                if($_->{shelf_module} =~ /(\w+\.\w+)\.\w+/) {
                    push(@shelfs,$1); 
                }
            }
        } else {
            logcomment("ACP status is inactive");
        }
    } 
 
     foreach my $shelf (@shelfs){
        my $response = $api_set_obj->storage_show_adapter('adapter-name' => $shelf);
        my @search_results = $response->search('slot' => ".*");
        foreach my $result(@search_results){
            if(($result->{"description"} =~ /^SAS Host Adapter/i) &&($result->{'redundant'} eq 'Yes')){
                $test_shelves->{$shelf} = 1;
            }
        }
    }
    return $test_shelves;   
}

=head2 get_disks_from_disk_hba_acp_shelf_type

 my @ret = get_disks_from_disk_hba_acp_shelf_type(FILER => $FILER, disk_type => $DISK_TYPE, shelf_type => $SHELF_TYPE, hba_type => $HBA_TYPE, acp_type => $ACP_TYPE);
This subrutine returns an hash reference of disks in the system with specific parameters.

The following arguments are accepted:

=over

=item FILER
(Required): filer name

=item shelf_type

(Optional): Shelf types for which the disks has to be returned

=item hba_type

(Optional): HBA types for which the disks has to be returned

=item disk_type

(Optional): Disk types for which the disks has to be returned

=item acp_type

(Optional): ACP types for which the disks has to be returned

=back

=cut

sub get_disks_from_disk_hba_acp_shelf_type {
    
    my (%opts) = validate(@_, {
            FILER => {type=>SCALAR},
            shelf_type => {type=>SCALAR,optional=>1},
            hba_type => {type=>SCALAR,optional=>1},
            acp_type => {type=>SCALAR,optional=>1},
            disk_type => {type=>SCALAR,optional=>1},
            }
            );
    my $output;
    my @disk_types;
    my @search_results;
    my $result;
    my @channels = ();
    my @shelf_types;
    my @hba_types;
    my @acp_types;
    my $test_disks;
    
    my $FILER = $opts{FILER};
    
    my $filer_ci = NACL::C::Node->find();
    my $api_set_obj = $filer_ci->get_7m_or_nodescope_apiset();
    
    my $output_aggr_status = $api_set_obj->aggr_status("raid-disk" => 1);
    
    if($opts{disk_type}) {
        @disk_types = split(",",$opts{disk_type});
        @search_results = $output_aggr_status->search('type' => join('|',@disk_types));
        foreach $result(@search_results){
            if($result->{raid_disk} ne "partner") {
                $test_disks->{$result->{"device"}} = 1;
            }
        }
    } 
    
    if ($opts{shelf_type}) {
        $output = $api_set_obj->environment_status_shelf();
        @search_results = ();
        @shelf_types = split(',',$opts{shelf_type});
        @search_results = $output->search('module_type' => join('|',@shelf_types));
        
        foreach (@search_results) {
            push(@channels, $_->{channel}); 
        }
    
        @search_results = $output_aggr_status->search('ha' => join("|",@channels));
        foreach (@search_results) {
            if($_->{raid_disk} ne "partner") {
                $test_disks->{$_->{"device"}} = 1;
            }
        }
    } 
    
    if ($opts{hba_type}) {
        foreach (split(',',$opts{hba_type})) {
            if($_ eq 'FC') {
                push(@hba_types,'.*Fibre Channel.*');
            } elsif ($_ eq 'SAS') {
                push(@hba_types,'.*SAS.*');
            } else {
                die fail_result("Only allowed types of HBA are FC and SAS");
                exit(0);
            }
        }
            
        $output = $api_set_obj->storage_show_adapter("all" => 1);
        @channels = ();
        @search_results = $output->search('description' => join('|',@hba_types));
        foreach (@search_results) {
            push(@channels,$_->{slot}); 
        }
    
        $output = $api_set_obj->aggr_status("raid-disk" => 1);
        @search_results = $output_aggr_status->search('ha' => join("|",@channels));
        foreach (@search_results) {
            if($_->{raid_disk} ne "partner") {
                $test_disks->{$_->{"device"}} = 1;
            }
        }
    } 
    
    if ($opts{acp_type}) {
        @acp_types = split(',',$opts{acp_type});
        $output = $api_set_obj->storage_show_acp("all" => 1);
        @search_results = $output->search('acp_status' => ".*");
    
        if($search_results[0]->{acp_status} eq "Active") {
            @channels = ();
            @search_results = $output->search('module_type' => join('|',@acp_types));
            foreach (@search_results) {
                if($_->{shelf_module} =~ /(\w+)\.\w+\.\w+/) {
                    push(@channels,$1); 
                }
            }
            $output = $api_set_obj->aggr_status("raid-disk" => 1);
            @search_results = $output_aggr_status->search('ha' => join("|",@channels));
            foreach (@search_results) {
                if($_->{raid_disk} ne "partner") {
                    $test_disks->{$_->{"device"}} = 1;
                }
            }
        
        } else {
            logcomment("ACP status is inactive");
        }
    } 
    
    return $test_disks;
}


1;
