#$Id: //depot/prod/test/Rsierranevada/lib/Storage/Scsi.pm#1 $
##
## @summary Scsi - Library module for scsi command
## @author sajan@netapp.com
## @status testme
## @pod here
## @change 2008-05-06 kbala: Added scsi_format subroutine
## @change 2010-02-17 anishv: Change functions to use filer_cmd.
## @change 2012-03-18 rpradeep: modified the scsi_modeselcet subroutine to support boab dongle power cycle.

=head1 NAME

Storage::Scsi

=head1 DESCRIPTION

This library provide functions to implement ONTAP scsi command. This 
command is not availbale in admin mode at all, in diag mode some of the scsi 
options are available and quite a few of them are available only in debug 
build (test privilege).
=head1 SYNOPSIS

use Storage::Scsi qw(
                      scsi_format
                      scsi_capacity
                      scsi_inquiry
                      scsi_test
                      scsi_dumpfailures
                      scsi_dumpfailures_s
                      scsi_readdefect
                      scsi_readdefect_b
                      scsi_modesense
                      scsi_modeselect
                      scsi_set_error
                      scsi_force_error_e
                      scsi_force_error_d
                      scsi_force_error_s
                      scsi_writefailure
                      scsi_drive_log
                      scsi_reset
                      scsi_stop
                      scsi_start
                      scsi_reserve
                      scsi_release
                      scsi_preempt
                      scsi_primary_select
                      scsi_verify
                    );
  
my $host = host($FILER);
my $device = "5.0";
my $capacity = scsi_capacity(host=>$host, device=>$device);

=head1 ROUTINES

=cut

package Storage::Scsi;

use strict;
use warnings;
BEGIN {
    use base qw(Exporter);
    our @EXPORT_OK = qw(
            scsi_format
            scsi_capacity 
            scsi_inquiry
            scsi_test
            scsi_dumpfailures
            scsi_dumpfailures_s
            scsi_readdefect
            scsi_readdefect_b
            scsi_modesense
            scsi_modeselect
            scsi_set_error
            scsi_force_error_e
            scsi_force_error_d
            scsi_force_error_s
            scsi_writefailure
            scsi_drive_log
            scsi_reset
            scsi_stop
            scsi_start
            scsi_reserve
            scsi_release
            scsi_preempt
            scsi_primary_select
            scsi_verify
            );

Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
};


use Storage::SASUtils qw(filer_cmd);
use Carp qw(confess croak);
use Params::Validate qw(validate OBJECT SCALAR);
use Tharn qw(say script_result host logcomment logresult param);
use Storage::Misc qw(filer_cmd);


use Utils qw(make_conn);
use Data::Dumper;
#use Switch;

our $MAINT_PROMPT = "\\*\>";

#----------------------------------------------------------------------
# The exported rotines are coded in the following section
#----------------------------------------------------------------------

=head2 scsi_format

This routine formats single or mulitple disks

scsi_format( 
             host=>$host, 
             device=>"device_name1 device_name2. ...",
	     timeout=>"2000",
           );

The return value will be 

$VAR1 = {
            '4b.444' => 'Format Failed. Disk not found',
            '4b.28' => 'Un-Formatted. Snaplock Compliance Disk'
            '4b.29' => 'Formatted"
        };

The command in MAIN mode is scsi format <device_name1> [<device_name2> ... [<device_nameN]]

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device_name

(Required) : Get device name. 

=item conntype

(Optional) : The connection type to be created by this function to talk to
the filer. Defualt is "#console".

=back

=cut

sub scsi_format {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device_name => { type=>SCALAR },
                     conntype => { type=>SCALAR, default=>"#console", optional=>1 },
                     timeout => { type=>SCALAR, default=>6000, optional=>1 },	 
                     blocksize => { type=>SCALAR, default=>undef, optional=>1 },	 
                 });

    my $host = $opts{host};
    my $device_name = $opts{device_name};
    my $conntype = $opts{conntype};
    my $timeout = $opts{timeout};
    my $conn = make_conn($host, $conntype);
    my $count = 0;
    my $cmd ;
    if(defined($opts{blocksize})){
        if(($opts{blocksize} eq "512") || ($opts{blocksize} eq "520")){
            $cmd = "scsi format -b $opts{blocksize} -t 10000000 $device_name";
        }
        else{
            croak "Only valid blocksize options are 512 or 520";
        }
    }
    else{
            $cmd = "scsi format -t 10000000 $device_name";
    }



    my ($result, $output, @output, @device_name, %output);
    (undef, undef, $result) = say($conn, "$cmd", -prompt, "$MAINT_PROMPT", "-timeout", $opts{timeout});

    @output = split(/\n/, $result);
    @device_name = split(/ /, $device_name);

    foreach my $disk (@device_name) {
        foreach my $out (@output) {
            if ($out =~ $disk) {
                if ( $out =~ /scsi format: succeeded/ ) { 
                    $output{$disk} = "Formatted";
                } elsif ( $out =~ /SnapLock Compliance volume. It will not be overwritten or formatted/ ) {
                    $output{$disk} = "Un-Formatted. Snaplock Compliance Disk";
                } elsif ( $out =~ /scsi format failed/ ) {
                    $output{$disk} = "Format Failed. Disk not found";
                }
            }
        }
    }    
                    
    return (\%output);
}

=head2 scsi_capacity

This routine returns ponter to hash having device capacity information

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_capacity(host=>$host, device=>$device_name);

The output looks like:

$VAR1 = {
          '9a.16' => {
                       'totalBlocks' => 562884301,
                       'totalBytes' => '292699836520',
                       'bytesPerBLock' => 520
                     }
        };

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_capacity {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi capacity $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);
    if ($output =~ /device $opts{device} not found/) {
        croak $output;
    }

    # The scsi capacity generates only one line of output if this command is 
    # issued through console then it gets echoed back you get an additional 
    # line
    my(@lines) = split(/\r*\n/, $output);
    my %capacity;
    foreach (@lines) {
        if ($_ =~ /^\s*(\w+) blocks x (\w+) bytes\/blk = (\w+) bytes/) {
            $capacity{$opts{device}}{totalBlocks} = $1;
            $capacity{$opts{device}}{bytesPerBLock} = $2;
            $capacity{$opts{device}}{totalBytes} = $3;
        }
    }

    return (\%capacity);
}

=head2 scsi_inquiry

This routine returns ponter to hash having device inquiry information incliding
the serial number, the command is issued again with -s option to get the serial
number.

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_inquiry(host=>$host, device=>$device_name);

The output looks like:
$VAR1 = {
          '9a.16' => {
                       'product' => 'X276_HPYTA288F10',
                       'support' => 3,
                       'revision' => 'NA02',
                       'vendor' => 'NETAPP',
                       'serialNumber' => 'VAV0VM8A',
                       'type' => '0',
                       'vendorUnique' => 'VAV0VM8A'
                     }
        };

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_inquiry {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                     get_serial => { type=>SCALAR, default=>"yes", optional=>1 },
                 });

    my $cmd = "scsi inquiry $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);

    if ($output =~ /device $opts{device} not found/) {
        croak $output;
    }

    my(@lines) = split(/\r*\n/, $output);
    my %inquiry;
    foreach (@lines) {
        if ($_ =~ /^\s*type (\w+) support (\w+)/) {
            $inquiry{$opts{device}}{type} = $1;
            $inquiry{$opts{device}}{support} = $2;
        }

        if ($_ =~ /^\s*vendor (\w+)/)  {
            if ($_ =~ /^\s*vendor unique (\w+)/)  {
                $inquiry{$opts{device}}{vendorUnique} = $1;
            } else {
                $inquiry{$opts{device}}{vendor} = $1;
            }
        }

        if ($_ =~ /^\s*product /)  {
		my @split = split(/\s+/,$_);
            $inquiry{$opts{device}}{product} = $split[1];
        }

        if ($_ =~ /^\s*revision (\w+)/)  {
            $inquiry{$opts{device}}{revision} = $1;
        }
    }

    #
    #Now get the serial number
    #
    if($opts{get_serial} eq "yes") {
     $cmd = "scsi inquiry -s  $opts{device}";
     $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);
    if ($output =~ /device $opts{device} not found/) {
        croak $output;
    }
    
    @lines = ();
    (@lines) = split(/\r*\n/, $output);
    foreach (@lines) {
        if ($_ =~ /^\s*serial #: (\w+)/) {
            $inquiry{$opts{device}}{serialNumber} = $1;
        }

    }
    }

    return (\%inquiry);
}

=head2 scsi_test

This routine returns 1 if drive is ready otherwise 1

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_test(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_test {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi test $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);

    if ($output =~ /device $opts{device} not found/) {
        croak $output;
    }

    my(@lines) = split(/\r*\n/, $output);
    foreach (@lines) {
        if ($_ =~ /device|drive ready/) {
            return (1);
        }
    }

    return (0);
}

=head2 scsi_dumpfailures

This routine returns pointer to hash table

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_dumpfailures(host=>$host, device=>$device_name);

The output looks like:

$VAR1 = {
          '9a.16' => {
                       1 => 'Disk failed due to SCSI:hardware error(0x04), ASC(0x80), ASCQ(0x85), FRU(0x32)',
                       2 => 'Disk failed with error 0x1d:0x03  (0xf0, 0x1d, 0x03, 0x0b) ',
                       3 => 'Disk unfailed by operator '
                     }
        };

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 1200

=back

=cut

sub scsi_dumpfailures {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>1200, optional=>1 },
                 });

    my %dumpfailure;
    my $cmd = "scsi dumpfailures $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);

    if ($output =~ /device $opts{device} not found/) {
        croak $output;
    }

    my(@lines) = split(/\r*\n/, $output);
    foreach (@lines) {
        if ($_ =~ /^\s*(\d+).\s*$opts{device}:\s*(.*)/)  {
            $dumpfailure{$opts{device}}{$1} = $2;
        }
    }
    return(\%dumpfailure);

}

=head2 scsi_dumpfailures_s

The output of command scsi dumpfailures -s <device-name> is returned in a 
pointer to hash table by this routine. 

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_dumpfailures_s(host=>$host, device=>$device_name);

The output looks like:
$VAR1 = {
          '9a.16' => {
                       1 => {
                              'FRU' => ' 0x32',
                              'ASC' => ' 0x80',
                              'ASCQ' => ' 0x85',
                              'errorType' => '0xf4'
                            },
                       2 => {
                              'FRU' => ' 0x0b',
                              'ASC' => ' 0x1d',
                              'ASCQ' => ' 0x03',
                              'errorType' => '0xf0'
                            },
                       3 => {
                              'FRU' => ' 0x00',
                              'ASC' => ' 0x00',
                              'ASCQ' => ' 0x00',
                              'errorType' => '0xe0'
                            }
                     }
        };

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item timeout

(Optional) : Timeout value for the say. Default : 1200

=back

=cut

sub scsi_dumpfailures_s {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>1200, optional=>1 },
                 });

    my %dumpfailure;
    my $cmd = "scsi dumpfailures -s  $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);

    if ($output =~ /device $opts{device} not found/) {
        croak $output;
    }

    my(@lines) = split(/\r*\n/, $output);
    foreach (@lines) {
        if ($_ =~ /^\s*(\d+).\s*$opts{device}:\s*\((.*)\)/)  {
            my @temp = split(',\s*', $2);
            $dumpfailure{$opts{device}}{$1}{errorType} = $temp[0];
            $dumpfailure{$opts{device}}{$1}{ASC} = $temp[1];
            $dumpfailure{$opts{device}}{$1}{ASCQ} = $temp[2];
            $dumpfailure{$opts{device}}{$1}{FRU} = $temp[3];
        }
    }
    return(\%dumpfailure);

}

=head2 scsi_readdefect

This routine returns pointer to hash table 
my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_readdefect(host=>$host, device=>$device_name);

The output looks like:
$VAR1 = {
          '9a.16' => {
                      'defectLength' => 11516,
                       'bytes' => 11516,
                       'format' => 5,
                       '0' => {
                                'head' => '0',
                                'cyl' => 286,
                                'sector' => 683
                              },
                       1 => {
                              'head' => '0',
                              'cyl' => 286,
                              'sector' => 684
                            },
                       2 => {
                              'head' => '0',
                              'cyl' => 286,
                              'sector' => 685
                            },
                       1248 => {
                                 'head' => 4,
                                 'cyl' => 49648,
                                 'sector' => 605
                               },
                       715 => {
                                'head' => '0',
                                'cyl' => 22528,
                                'sector' => 424
                              },
                       1249 => {
                                 'head' => 4,
                                 'cyl' => 49649,
                                 'sector' => 604
                               },
                        .
                        .
        }

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item defecttype

(Optional) : primary|grown. Default : primary

=item timeout

(Optional) : Timeout value for the say. Default : 1800

=back

=cut

sub scsi_readdefect {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     defecttype => { type=>SCALAR,default=>"primary",optional=>1 },
                     timeout => { type=>SCALAR, default=>1800, optional=>1 },
                 });


    my %readdefect = ();
    my $cmd = "scsi readdefect -t $opts{defecttype} $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);

    if ($output =~ /device $opts{device} not found/) {
        croak $output;
    }

    my(@lines) = split(/\r*\n/, $output);
    foreach (@lines) {
        if ($_ =~ /^\s*defect length\s*(\w+)/)  {
            $readdefect{$opts{device}}{defectLength} = $1;
        }

        if ($_ =~ /^\s*Disk $opts{device} \((\w+) bytes format (\w+)/)  {
            $readdefect{$opts{device}}{bytes} = $1;
            $readdefect{$opts{device}}{format} = $2;
        }

        if ($_ =~ /^\s*(\d+):\s*\(\s*(.*)\)/)  {
            my @temp = split(',\s*', $2);
            $readdefect{$opts{device}}{$1}{cyl} = $temp[0];
            $readdefect{$opts{device}}{$1}{head} = $temp[1];
            $readdefect{$opts{device}}{$1}{sector} = $temp[2];
        }
    }
    return(\%readdefect);

}

=head2 scsi_resddefect_b

This routine returns pointer to hash table having information by 
blocks (scsi readdefect -b -t primary|grown <device-name> 

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_readdefect_b(host=>$host, device=>$device_name);

The output looks like:

$VAR1 = {
          '9a.16' => {
                       1248 => 480876498,
                       715 => 241695265,
                       1249 => 484802483,
                       716 => 241695266,
                       570 => 234426274,
                       717 => 241695267,
                       'defectLength' => 5688,
                       'format' => '0',
                       'bytes' => 5688,
                        .
                        .
                        .
    }

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item defecttype

(Optional) : primary|grown. Default : primary

=item timeout

(Optional) : Timeout value for the say. Default : 1800

=back

=cut

sub scsi_readdefect_b {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     defecttype => { type=>SCALAR,default=>"primary",optional=>1 },
                     timeout => { type=>SCALAR, default=>1800, optional=>1 },
                 });


    my %readdefect = ();
    my $cmd = "scsi readdefect -b -t $opts{defecttype} $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);

    if ($output =~ /device $opts{device} not found/) {
        croak $output;
    }

    my(@lines) = split(/\r*\n/, $output);
    foreach (@lines) {
        if ($_ =~ /^\s*defect length\s*(\w+)/)  {
            $readdefect{$opts{device}}{defectLength} = $1;
        }

        if ($_ =~ /^\s*Disk $opts{device} \((\w+) bytes format (\w+)/)  {
            $readdefect{$opts{device}}{bytes} = $1;
            $readdefect{$opts{device}}{format} = $2;
        }

        if ($_ =~ /^\s*(\d+):\s*(\w+)/)  {
            $readdefect{$opts{device}}{$1} = $2;
        }
    }
    return(\%readdefect);

}

=head2 scsi_modesense

This routine returns pointer to hash table  

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_modesense(host=>$host, device=>$device_name);

The output looks like:
$VAR1 = {
          '9a.16' => {
                       'pageCode' => 3,
                       'header' => '23 00 10 08',
                       'value' => '9f a6 6b 30 00 68 06 80 03 7e 02 08 00 01 00 69 00 c4 40 00 00 00',
                       'len' => 16,
                       'blockDescriptor' => '21 8c ee ce 00 00 02 08'
                     }
        };

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item page

(Required) : page number

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item tvalue

(Optional) : Value for -t option. Default : undef (not needed)

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_modesense {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     page => { type=>SCALAR }, 
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     tvalue => { type=>SCALAR,default=>undef,optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my %modesense = ();
    my $tvalue = undef;

    if (defined($opts{tvalue})) {
        $tvalue = "-t $opts{tvalue}";
    }
    my $cmd = "scsi modesense -p $opts{page} $tvalue $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);

    if ($output =~ /device $opts{device} not found/) {
        croak $output;
    }

    if ($output =~ /scsi modesense error/) {
        croak $output;
    }

    if ($output =~ /Invalid -t argument/) {
        croak $output;
    }

    my(@lines) = split(/\r*\n/, $output);
    my $next_is_value = 0;
    foreach (@lines) {
        if ($_ =~ /^\s*header\s*(.*)/)  {
            $modesense{$opts{device}}{header} = $1;
            chop($modesense{$opts{device}}{header});
        }

        if ($_ =~ /^\s*block descriptor\s*(.*)\s*/)  {
            $modesense{$opts{device}}{blockDescriptor} = $1;
            chop($modesense{$opts{device}}{blockDescriptor});
        }

        if ($next_is_value == 1)  {
            $modesense{$opts{device}}{value} = $modesense{$opts{device}}{value} . $_;
        }

        if ($_ =~ /^\s*page code\s*(\w+)\s*len\s*(\w+)/)  {
            $modesense{$opts{device}}{pageCode} = $1;
            $modesense{$opts{device}}{len} = $2;
            $modesense{$opts{device}}{value} = undef;
            $next_is_value = 1;
        }

    }

    if (exists($modesense{$opts{device}})) {
        chop($modesense{$opts{device}}{value});
    }
    return(\%modesense);

}

=head2 scsi_modeselect

This routine returns 1 for successful completion otherwise croak with the
error messages reported back by the command.

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_modeselect(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item page

(Required) : page number

=item offset

(Optional) : byte offset to change. Default undef (will not be used).

=item value

(Optional) : value to change byte to. Default undef (will not be used).

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_modeselect {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     page => { type=>SCALAR },
                     offset => { type=>SCALAR,default=>undef,optional=>1 },
                     value => { type=>SCALAR,default=>undef,optional=>1 },
                     s_page => {type=>SCALAR,default=>undef,optional=>1 },
                     offset1 => {type=>SCALAR,default=>undef,optional=>1 },
                     value1 => { type=>SCALAR,default=>undef,optional=>1 },
                     cmd_opt => { type=>SCALAR,default=>undef,optional=>1 },
                     device => { type=>SCALAR },
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                     'length' => { type => SCALAR, default=>undef, optional=>1},
                     'boab' => { type => SCALAR, default=>undef, optional=>1}
                 });

    my %modesense = ();
    my $output;
    my $value = undef;
    my $offset = undef;
    my $cmd;

# Added these veriable for Boab

    my $s_page = undef;
    my $value1 = undef;
    my $offset1 = undef;
# Added for BOAB shelf

    if (defined($opts{offset}) && defined($opts{value}) && defined($opts{s_page}) && defined($opts{offset1}) && defined($opts{value1})){

    $offset = "-o $opts{offset}";
    $value = "-v $opts{value}";
    $s_page = "-s $opts{s_page}";
    $offset1 = "-o $opts{offset1}";
    $value1 = "-v $opts{value1}";
    $cmd = "scsi modeselect -p $opts{page} $s_page $offset $value $offset1 $value1 $opts{device}";
     }
    elsif (defined($opts{offset}) && defined($opts{value})) {
        $offset = "-o $opts{offset}";
        $value = "-v $opts{value}";
        $cmd = "scsi modeselect -p $opts{page} $offset $value $opts{device}"; 
    }
    elsif(defined($opts{'length'})) {
        $cmd = "scsi modeselect -f $opts{page} -l $opts{length} $opts{cmd_opt} $opts{device}";
    }
    elsif(defined($opts{cmd_opt})) {
        $cmd = "scsi modeselect -p $opts{page} $opts{cmd_opt} $opts{device}"; 
    }
    $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>"$cmd");

    if ($output !~ /modifying offset/){
        croak $output;
    }
# Added for boab shelf

    if (($output =~ /modeselect error:/) && (defined($opts{boab}))){
    return $output;
    }
#####

    if ($output =~ /modeselect error:/){
        croak $output;
    }
    return(1);
}

=head2 scsi_set_error

This routine returns 1 for successful completion otherwise croak with the
error messages reported back by the command. 

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_set_error(host=>$host, errordesc=>"WriteScsiAbort2904C" device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item errordesc

(Required) : Name of the error sequence to set

=item loopcount

(Optional) : Number of times to loop error sequence; -1 forever. Default: undef.

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_set_error {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     errordesc => { type=>SCALAR }, 
                     loopcount => {type=>SCALAR,default=>undef,optional=>1 },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });


    my $loopcount = "";

    if (defined($opts{loopcount})) {
        $loopcount = "-l $opts{loopcount}";
    }
    my $cmd = "scsi set_error -e $opts{errordesc} $loopcount $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /Error sequence $opts{errordesc} set for device $opts{device}/) {
        return (1);
    } else {
        croak $output;
    }
}

=head2 scsi_force_error_e

Implements the command "scsi force_error -e <device-name>". This routine returns 1 for 
successful completion otherwise croak with the error messages reported back by the command. 

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_force_error_e(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_force_error_e {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });
    my $cmd = "scsi force_error -e $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /Error injection enabled for device $opts{device}/) {
        return (1);
    } else {
        croak $output;
    }
}

=head2 scsi_force_error_d

Implements the command "scsi force_error -d <device-name>". This routine returns 1 for 
successful completion otherwise croak with the error messages reported back by the command. 

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_force_error_d(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_force_error_d {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });
    my $cmd = "scsi force_error -d $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /Error injection disabled for device $opts{device}/) {
        return (1);
    } else {
        croak $output;
    }
}

=head2 scsi_force_error_s

Implements the command "scsi force_error -s <device-name>". This routine  returns pointer 
to a hash table for command completed successfully otherwise croak with the error messages 
reported back by the command. 

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_force_error_s(host=>$host, device=>$device_name);

The output looks like:
$VAR1 = {
          '9a.16' => {
                       '0' => {
                                'execution_count' => '0x1',
                                'resudue' => '0x0',
                                'target_status' => '0x2',
                                'out_status_flags' => '0x24',
                                'opcode' => '0x2a',
                                'sense_data' => {
                                                  'segment' => '0x0',
                                                  'sense_code_qualifier' => '0x4',
                                                  'length' => '0xa',
                                                  'class' => '0x70',
                                                  'info' => '0x0000',
                                                  'sense_key' => '0xb',
                                                  'sense_code' => '0x29'
                                                },
                                'sense_len' => '0x12',
                                'adapter_action' => '0x0',
                                'error_type' => '0x1',
                                'ha_status' => '0x0',
                                'Error state information for device 9a.16' => ''
                              },
                       '1' => {
                                'execution_count' => '0x0',
                                'resudue' => '0x0',
                                'target_status' => '0x0',
                                'ha_status' => '0x0',
                                'out_status_flags' => '0x0',
                                'opcode' => '0x0',
                                'sense_data' => {
                                                  'segment' => '0x0',
                                                  'sense_code_qualifier' => '0x0',
                                                  'length' => '0x0',
                                                  'class' => '0x0',
                                                  'info' => '0x0000',
                                                  'sense_key' => '0x0',
                                                  'sense_code' => '0x0'
                                                },
                                'sense_len' => '0x0',
                                'adapter_action' => '0x0',
                                'error_type' => '0x0'
                              },
                        .
                        .
                        .
       } 

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_force_error_s {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my %HoH = ();
    my $cmd = "scsi force_error -s $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /Error state information for device $opts{device}/) {
        my $dev = $opts{device};
        my (@lines) = split(/\r*\n/, $output);
        my $st;
        $st = 0;
        foreach (@lines) {
            if ($_  =~ /Error Injection:\s*DISABLED/) {
                $HoH{$dev}{errInjState} = "DISABLED";
            } elsif ($_ =~ /Error Injection Loop Count:\s*(\d+)/) {
                $HoH{$dev}{loopCount} = $1;
            } elsif ($_ =~ /Error Injection State\s*\#(\d+)/) {
                $st = $1;
            } elsif ($_ =~ s/sense_data://) {
                my @temp = split(/;\s+/);
                foreach (@temp) {
                   if ($_ =~ /^\s*(.*):\s*(.*)/) {
                       $HoH{$dev}{$st}{sense_data}{$1} = $2; 
                   }
                }
            } elsif ($_ =~ /sense_code/) {
                my @temp = split(/;\s+/);
                foreach (@temp) {
                   if ($_ =~ /^\s*(.*):\s*(.*)/) {
                       $HoH{$dev}{$st}{sense_data}{$1} = $2; 
                   }
                }
            } else {
                my @temp = split(/;\s+/);
                foreach (@temp) {
                   if ($_ =~ /^\s*(.*):\s*(.*)/) {
                       $HoH{$dev}{$st}{$1} = $2; 
                   }
                }
            }
        }
        return(\%HoH);
    } else {
        croak $output;
    }
}

=head2 scsi_writefailure

Implements the command "scsi writefailure <device-name>". This routine  returns 1
for command completed successfully otherwise croak with the error messages 
reported back by the command. 

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_writefailure(host=>$host, param=>s device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item param

(Required) : parameter to the command. It could be either s|t|u

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_writefailure {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     param => { type=>SCALAR }, 
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi writefailure -$opts{param} $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /Failure bytes written successfully/) {
        return (1);
    } else {
        croak $output;
    }
}

=head2 scsi_drive_log

Implements the command "scsi drive_log <device-name>". At the moment this routine  returns 
the command output in a scalar variable if command is successful otherwise it croaks if the 
command is failed

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_drive_log(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_drive_log {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });
    my $cmd = "scsi drive_log $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /Device $opts{device} not found/) {
        croak $output;
    } else {
        return ($output);
    }
}

=head2 scsi_reset

Implements the command "scsi reset <device-name>". It returns 1 if reset is done otherwise  
croaks.

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_reset(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_reset {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi reset $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /scsi reset $opts{device}: done/) {
        return (1);
    } else {
        croak $output;
    }
}

=head2 scsi_stop

Implements the command "scsi stop <device-name>". It returns 1 for success otherwise croaks.

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_stop(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_stop {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi stop $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /scsi stop: completed/) {
        return (1);
    } else {
        croak $output;
    }
}

=head2 scsi_start

Implements the command "scsi start <device-name>". It returns 1 for success otherwise croaks.

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_start(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_start {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi start $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /scsi start: completed/) {
        return (1);
    } else {
        croak $output;
    }
} 

=head2 scsi_reserve

Implements the command "scsi reserve <device-name>". It returns 1 for success otherwise croaks.

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_reserve(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_reserve {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi reserve $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /action succeeded/) {
        return (1);
    } else {
        croak $output;
    }
} 

=head2 scsi_release

Implements the command "scsi release <device-name>". It returns 1 for success otherwise croaks.

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_release(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_release {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi release $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /action succeeded/) {
        return (1);
    } else {
        croak $output;
    }
} 

=head2 scsi_preempt

Implements the command "scsi preempt <device-name>". It returns 1 for success otherwise croaks.

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_preempt(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_preempt {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi preempt $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /action succeeded/) {
        return (1);
    } else {
        croak $output;
    }
} 

=head2 scsi_select_primary

Implements the command "scsi select_primary -h <adapter-name> <device-name>". It returns 1 for success otherwise croaks.

my $host = host($FILER);
my $device = "5.0";
my $ret = scsi_select_primary(host=>$host, device=>$device_name);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item adapter

(Required) : The adapter this device is connected to  

=item device

(Required) : Get information for this device 

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in 
maintenance mode)

=item conntype

(Optional) : The connection type to be created by this function to talk to 
the filer. Defualt is "#rsh".

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub scsi_select_primary {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     adapter => { type=>SCALAR }, 
                     device => { type=>SCALAR }, 
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my $cmd = "scsi select_primary -h $opts{adapter} $opts{device}";
    my $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");

    if ($output =~ /action succeeded/) {
        return (1);
    } else {
        croak $output;
    }
} 

=head2 scsi_verify

This routine runs scsi verify on a single disk

scsi_format( 
             host=>$host, 
             device=>"device_name",
	        timeout=>"2000",
           );

The return value will be 


The command in MAIN mode is scsi format <device_name1> [<device_name2> ... [<device_nameN]]

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item device_name

(Required) : Get device name. 

=item conntype

(Optional) : The connection type to be created by this function to talk to
the filer. Default is "#rsh".

=back

=cut

sub scsi_verify {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     device => { type=>SCALAR },
                     filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                     conntype => { type=>SCALAR,default=>"#rsh",optional=>1 },
                     timeout => { type=>SCALAR, default=>"10000", optional=>1 },	 
                     lba => { type=>SCALAR, default=>undef, optional=>1 },	 
                     blocks => { type=>SCALAR, default=>undef, optional=>1 },	 
                 });

    my ($result, $output, @output, @device_name, %output);
    my $cmd;
    if((defined($opts{lba})) && (defined($opts{blocks}))){
        $cmd = "scsi verify -l $opts{lba} -n $opts{blocks} $opts{device}";
    }
    else{
        $cmd = "scsi verify $opts{device}";
    }

    $output = filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd,privmode=>"test");
    if(($output !~ /\./) || ($output =~ /scsi verify error:/)){
        croak $output;
    }
    else{
        return 1;
    }

}



#----------------------------------------------------------------------
# The testme routine to test the exported functions of this library
#
# This routine can be invoked in the following manner
#
# ntest -hw FILER=raid-u24 PKGNAME=Storage::Scsi util/test_pkg (test all functions in normal filer mode)
#
# ntest -hw FILER=raid-u24 FILERMODE=maint PKGNAME=Storage::Scsi util/test_pkg (test all functions in maintenance mode)
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=scsi_capacity DEVICE=9a.16 PKGNAME=Storage::Scsi util/test_pkg (test function scsi_capacity in normal mode)
#
# ntest -hw FILER=raid-u24 FILERMODE=maint FUNCTION_NAME=scsi_capacity DEVICE=9a.16 PKGNAME=Storage::Scsi util/test_pkg (test function scsi_capacity in maint mode)
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=scsi_inquiry DEVICE=9a.16 PKGNAME=Storage::Scsi util/test_pkg (test function scsi_inquiry in normal mode)
#
# ntest -hw FILER=raid-u24 FILERMODE=maint FUNCTION_NAME=scsi_inquiry DEVICE=9a.16 PKGNAME=Storage::Scsi util/test_pkg (test function scsi_inquiry in maint mode)
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=scsi_test DEVICE=9a.16 PKGNAME=Storage::Scsi util/test_pkg (test function scsi_test in normal mode)
#
# ntest -hw FILER=raid-u24 FILERMODE=maint FUNCTION_NAME=scsi_test DEVICE=9a.16 PKGNAME=Storage::Scsi util/test_pkg (test function scsi_test in maint mode)
#----------------------------------------------------------------------

sub testme {

    package main;
    our ($FILER);
    our ($FUNCTION_NAME);
    our ($DEVICE);
    our ($FILERMODE);
    our ($ADAPTER);
    package Storage::Scsi;
    param('FILER',       '-mesg',    "Filer to test");
    param('FUNCTION_NAME',       '-default',    "all");
    param('DEVICE',       '-mesg',    "Device name is required");
    param('FILERMODE',       '-default',    "normal");
    param('ADAPTER',       '-default',    undef);

    logcomment("Testing on FILER $FILER");
    logcomment("Testing Function $FUNCTION_NAME");
    logcomment("Testing Device $DEVICE");
    logcomment("Testing Filermode $FILERMODE");
    logcomment("Testing Adapter $ADAPTER");

    my $host = host($FILER);
    my $ret;

    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_capacity") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_capacity(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_capacity failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_capacity");
            }

            #test rsh connection + normal mode
            eval {
                $ret = scsi_capacity(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_capacity failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_capacity");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_capacity(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_capacity failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_capacity");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_capacity(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_capacity failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_capacity");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_capacity(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_capacity: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_capacity: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_capacity(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_capacity: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_capacity: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_capacity(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_capacity failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_capacity");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_capacity(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_capacity failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_capacity");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_capacity(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_capacity: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_capacity: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_capacity(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_inquiry
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_inquiry") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_inquiry(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_inquiry failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_inquiry");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_inquiry(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_inquiry failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_inquiry");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_inquiry(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_inquiry failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_inquiry");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_inquiry(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_inquiry: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inquiry");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_inquiry(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_inquiry: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_inquiry(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_inquiry failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_inquiry");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_inquiry(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_inquiry failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_inquiry");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_inquiry(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_inquiry: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inquiry: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_inquiry(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_inquiry: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_inquiry: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_test
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_test") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_test(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_test failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_test");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_test(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_test failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_test");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_test(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_test failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_test");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_test(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_test: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_test: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_test(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_test: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_test: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_test(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_test failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_test");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_test(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_test failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_test");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_test(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_test: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_test: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_test(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_test: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_test: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_dumpfailures
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_dumpfailures") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_dumpfailures(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_dumpfailures(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_dumpfailures(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_dumpfailures(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_dumpfailures: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_dumpfailures");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_dumpfailures(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_dumpfailures: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_dumpfailures(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_dumpfailures(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_dumpfailures(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_dumpfailures: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_dumpfailures: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_dumpfailures(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_dumpfailures: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_dumpfailures: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_dumpfailures_s
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_dumpfailures_s") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_dumpfailures_s(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures_s");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_dumpfailures_s(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures_s");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_dumpfailures_s(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures_s");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_dumpfailures_s(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_dumpfailures_s: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_dumpfailures_s");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_dumpfailures_s(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_dumpfailures_s: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_dumpfailures_s(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures_s");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_dumpfailures_s(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_dumpfailures_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_dumpfailures_s");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_dumpfailures_s(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_dumpfailures_s: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_dumpfailures_s: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_dumpfailures_s(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_dumpfailures_s: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_dumpfailures_s: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_readdefect
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_readdefect") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_readdefect(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect");
            }

            #test rsh connection + normal mode + grown defect
            eval {
                $ret = scsi_readdefect(host=>$host, device=>$DEVICE, filermode=>$FILERMODE, defecttype=>"grown");
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_readdefect(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_readdefect(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_readdefect(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_readdefect: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_readdefect");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_readdefect(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_readdefect: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_readdefect(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_readdefect(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_readdefect(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_readdefect: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_readdefect: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_readdefect(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_readdefect: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_readdefect: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_readdefect_b
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_readdefect_b") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_readdefect_b(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect_b failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect_b");
            }

            #test rsh connection + normal mode + grown defect
            eval {
                $ret = scsi_readdefect_b(host=>$host, device=>$DEVICE, filermode=>$FILERMODE, defecttype=>"grown");
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect_b failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect_b");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_readdefect_b(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect_b failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect_b");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_readdefect_b(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect_b failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect_b");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_readdefect_b(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_readdefect_b: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_readdefect_b");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_readdefect_b(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_readdefect_b: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_readdefect_b(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect_b failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect_b");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_readdefect_b(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_readdefect_b failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_readdefect_b");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_readdefect_b(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_readdefect_b: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_readdefect_b: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_readdefect_b(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_readdefect_b: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_readdefect_b: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_modesense
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_modesense") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_modesense(host=>$host, device=>$DEVICE, filermode=>$FILERMODE, page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modesense failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modesense");
            }

            #test rsh connection + normal mode + tvalue 
            eval {
                $ret = scsi_modesense(host=>$host, device=>$DEVICE, filermode=>$FILERMODE, tvalue=>"saved", page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modesense failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modesense");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_modesense(host=>$host, device=>$DEVICE, page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modesense failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modesense");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_modesense(host=>$host, device=>$DEVICE, conntype=>"#console", page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modesense failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modesense");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_modesense(host=>$host, device=>"abc", page=>0x83);
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_modesense: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_modesense");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_modesense(host=>$host, device=>"abc", conntype=>"#console", page=>0x83);
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_modesense: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }

            #test wrong tvalue
            eval {
                $ret = scsi_modesense(host=>$host, device=>$DEVICE, page=>0x83, tvalue=>"zzz");
            };

            if ($@) {
                logresult('pass', "Got expected failure from function scsi_modesense: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_modesense: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_modesense(host=>$host, device=>$DEVICE, filermode=>"maint", page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modesense failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modesense");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_modesense(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console", page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modesense failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modesense");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_modesense(host=>$host, device=>"abc", filermode=>"maint", page=>0x83);
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_modesense: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_modesense: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_modesense(host=>$host, page=>0x83);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_modesense: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_modesense: $@");
        }

        #test missing mandotary page number
        eval {
            $ret = scsi_modesense(host=>$host, device=>$DEVICE);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_modesense: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_modesense: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_modeselect
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_modeselect") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_modeselect(host=>$host, device=>$DEVICE, filermode=>$FILERMODE, page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modeselect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modeselect");
            }

            #test rsh connection + normal mode + value + offset 
            eval {
                $ret = scsi_modeselect(host=>$host, device=>$DEVICE, filermode=>$FILERMODE, value=>"0x12", offset=>"0x100", page=>0x83);
            };

            if ($@) {
                logresult('pass', "Got expected result from function scsi_modeselect");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_modeselect");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_modeselect(host=>$host, device=>$DEVICE, page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modeselect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modeselect");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_modeselect(host=>$host, device=>$DEVICE, conntype=>"#console", page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modeselect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modeselect");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_modeselect(host=>$host, device=>"abc", page=>0x83);
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_modeselect: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_modeselect");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_modeselect(host=>$host, device=>"abc", conntype=>"#console", page=>0x83);
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_modeselect: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }

            #test wrong value
            eval {
                $ret = scsi_modeselect(host=>$host, device=>$DEVICE, page=>0x83, offset=>"Ox12");
            };

            if ($@) {
                logresult('pass', "Got expected failure from function scsi_modeselect: $@");
            } else { 
                 logresult('fail', "Got unexpected result from function scsi_modeselect: $@");
            }

            #test for missing offset
            eval {
                $ret = scsi_modeselect(host=>$host, device=>$DEVICE, page=>0x83, value=>"0x12");
            };

            if ($@) {
                logresult('pass', "Got expected failure from function scsi_modeselect: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_modeselect: $@");
            }

            #test for missing value
            eval {
                $ret = scsi_modeselect(host=>$host, device=>$DEVICE, page=>0x83, offset=>"0x12");
            };

            if ($@) {
                logresult('pass', "Got expected failure from function scsi_modeselect: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_modeselect: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_modeselect(host=>$host, device=>$DEVICE, filermode=>"maint", page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modeselect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modeselect");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_modeselect(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console", page=>0x83);
            };

            if ($@) {
                logresult('fail', "function scsi_modeselect failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_modeselect");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_modeselect(host=>$host, device=>"abc", filermode=>"maint", page=>0x83);
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_modeselect: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_modeselect: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_modeselect(host=>$host, page=>0x83);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_modeselect: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_modeselect: $@");
        }

        #test missing mandotary page number
        eval {
            $ret = scsi_modeselect(host=>$host, device=>$DEVICE);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_modeselect: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_modeselect: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_set_error
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_set_error") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_set_error(host=>$host, device=>$DEVICE, filermode=>$FILERMODE, errordesc=>"WriteScsiAbort2904C");
            };

            if ($@) {
                logresult('fail', "function scsi_set_error failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_set_error");
            }

            #test rsh connection + normal mode + loopcount 
            eval {
                $ret = scsi_set_error(host=>$host, device=>$DEVICE, filermode=>$FILERMODE, loopcount=>2, errordesc=>"WriteScsiAbort2904C");
            };

            if ($@) {
                logresult('fail', "function scsi_set_error failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_set_error");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_set_error(host=>$host, device=>$DEVICE, errordesc=>"WriteScsiAbort2904C");
            };

            if ($@) {
                logresult('fail', "function scsi_set_error failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_set_error");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_set_error(host=>$host, device=>$DEVICE, conntype=>"#console", errordesc=>"WriteScsiAbort2904C");
            };

            if ($@) {
                logresult('fail', "function scsi_set_error failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_set_error");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_set_error(host=>$host, device=>"abc", errordesc=>"WriteScsiAbort2904C");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_set_error: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_set_error");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_set_error(host=>$host, device=>"abc", conntype=>"#console", errordesc=>"WriteScsiAbort2904C");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_set_error: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }

            #test wrong error descriptor
            eval {
                $ret = scsi_set_error(host=>$host, device=>$DEVICE, errordesc=>"abc");
            };

            if ($@ =~ /invalid error descriptor abc/ ) {
                logresult('pass', "Got expected failure from function scsi_set_error: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_set_error: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_set_error(host=>$host, device=>$DEVICE, filermode=>"maint", errordesc=>"WriteScsiAbort2904C");
            };

            if ($@) {
                logresult('fail', "function scsi_set_error failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_set_error");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_set_error(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console", errordesc=>"WriteScsiAbort2904C");
            };

            if ($@) {
                logresult('fail', "function scsi_set_error failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_set_error");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_set_error(host=>$host, device=>"abc", filermode=>"maint", errordesc=>"WriteScsiAbort2904C");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_set_error: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_set_error: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_set_error(host=>$host, errordesc=>"WriteScsiAbort2904C");
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_set_error: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_set_error: $@");
        }

        #test missing mandotary error descriptor
        eval {
            $ret = scsi_set_error(host=>$host, device=>$DEVICE);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_set_error: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_set_error: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_force_error_e
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_force_error_e") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_force_error_e(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_e failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_e");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_force_error_e(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_e failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_e");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_force_error_e(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_e failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_e");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_force_error_e(host=>$host, device=>"abc");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_force_error_e: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_force_error_e");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_force_error_e(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_force_error_e: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_force_error_e(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_e failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_e");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_force_error_e(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_e failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_e");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_force_error_e(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_force_error_e: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_force_error_e: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_force_error_e(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_force_error_e: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_force_error_e: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_force_error_d
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_force_error_d") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_force_error_d(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_d failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_d");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_force_error_d(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_d failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_d");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_force_error_d(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_d failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_d");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_force_error_d(host=>$host, device=>"abc");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_force_error_d: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_force_error_d");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_force_error_d(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_force_error_d: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_force_error_d(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_d failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_d");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_force_error_d(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_d failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_d");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_force_error_d(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_force_error_d: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_force_error_d: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_force_error_d(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_force_error_d: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_force_error_d: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_force_error_s
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_force_error_s") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_force_error_s(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_s");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_force_error_s(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_s");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_force_error_s(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_s");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_force_error_s(host=>$host, device=>"abc");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_force_error_s: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_force_error_s");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_force_error_s(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_force_error_s: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_force_error_s(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_s");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_force_error_s(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_force_error_s failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_force_error_s");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_force_error_s(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /no such device/ ) {
                logresult('pass', "Got expected failure from function scsi_force_error_s: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_force_error_s: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_force_error_s(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_force_error_s: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_force_error_s: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_writefailure
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_writefailure") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode + param = s
            eval {
                $ret = scsi_writefailure(host=>$host, device=>$DEVICE, param=>"s", filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_writefailure failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_writefailure");
            }

            #test rsh connection + normal mode + param = s
            eval {
                $ret = scsi_writefailure(host=>$host, device=>$DEVICE, param=>"t", filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_writefailure failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_writefailure");
            }

            #test rsh connection + normal mode + param = s
            eval {
                $ret = scsi_writefailure(host=>$host, device=>$DEVICE, param=>"u", filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_writefailure failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_writefailure");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_writefailure(host=>$host, device=>$DEVICE, param=>"s");
            };

            if ($@) {
                logresult('fail', "function scsi_writefailure failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_writefailure");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_writefailure(host=>$host, device=>$DEVICE, conntype=>"#console", param=>"t");
            };

            if ($@) {
                logresult('fail', "function scsi_writefailure failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_writefailure");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_writefailure(host=>$host, device=>"abc", param=>"u");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_writefailure: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_writefailure");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_writefailure(host=>$host, device=>"abc", conntype=>"#console", param=>"u");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_writefailure: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_inqury: $@");
            }

            #test wrong param value
            eval {
                $ret = scsi_writefailure(host=>$host, device=>$DEVICE, param=>"z");
            };

            if ($@ =~ /unknown option/ ) {
                logresult('pass', "Got expected failure from function scsi_writefailure: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_writefailure: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_writefailure(host=>$host, device=>$DEVICE, filermode=>"maint", param=>"u");
            };

            if ($@) {
                logresult('fail', "function scsi_writefailure failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_writefailure");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_writefailure(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console", param=>"t");
            };

            if ($@) {
                logresult('fail', "function scsi_writefailure failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_writefailure");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_writefailure(host=>$host, device=>"abc", filermode=>"maint", param=>"u");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_writefailure: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_writefailure: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_writefailure(host=>$host, param=>"s");
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_writefailure: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_writefailure: $@");
        }

        #test missing mandotary param
        eval {
            $ret = scsi_writefailure(host=>$host, device=>$DEVICE);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_writefailure: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_writefailure: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_drive_log
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_drive_log") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_drive_log(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_drive_log failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_drive_log");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_drive_log(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_drive_log failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_drive_log");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_drive_log(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_drive_log failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_drive_log");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_drive_log(host=>$host, device=>"abc");
            };

            if ($@ =~ /Device abc not found/) {
                logresult('pass', "Got expected failure from function scsi_drive_log: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_drive_log: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_drive_log(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /Device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_drive_log: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_drive_log: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_drive_log(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_drive_log failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_drive_log");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_drive_log(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_drive_log failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_drive_log");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_drive_log(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /Device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_drive_log: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_drive_log: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_drive_log(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_reset
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_reset") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_reset(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_reset failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reset");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_reset(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_reset failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reset");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_reset(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_reset failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reset");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_reset(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/) {
                logresult('pass', "Got expected failure from function scsi_reset: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_reset: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_reset(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_reset: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_reset: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_reset(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_reset failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reset");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_reset(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_reset failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reset");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_reset(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_reset: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_reset: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_reset(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_stop
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_stop") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_stop(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_stop failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_stop");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_stop(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_stop failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_stop");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_stop(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_stop failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_stop");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_stop(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/) {
                logresult('pass', "Got expected failure from function scsi_stop: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_stop: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_stop(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_stop: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_stop: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_stop(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_stop failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_stop");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_stop(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_stop failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_stop");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_stop(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_stop: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_stop: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_stop(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_start
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_start") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_start(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_start failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_start");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_start(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_start failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_start");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_start(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_start failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_start");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_start(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/) {
                logresult('pass', "Got expected failure from function scsi_start: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_start: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_start(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_start: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_start: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_start(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_start failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_start");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_start(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_start failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_start");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_start(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_start: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_start: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_start(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_reserve
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_reserve") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_reserve(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_reserve failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reserve");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_reserve(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_reserve failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reserve");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_reserve(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_reserve failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reserve");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_reserve(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/) {
                logresult('pass', "Got expected failure from function scsi_reserve: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_reserve: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_reserve(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_reserve: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_reserve: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_reserve(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_reserve failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reserve");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_reserve(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_reserve failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_reserve");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_reserve(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_reserve: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_reserve: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_reserve(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_release
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_release") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_release(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_release failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_release");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_release(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_release failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_release");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_release(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_release failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_release");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_release(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/) {
                logresult('pass', "Got expected failure from function scsi_release: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_release: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_release(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_release: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_release: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_release(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_release failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_release");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_release(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_release failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_release");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_release(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_release: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_release: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_release(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_preempt
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_preempt") {
        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_preempt(host=>$host, device=>$DEVICE, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_preempt failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_preempt");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_preempt(host=>$host, device=>$DEVICE);
            };

            if ($@) {
                logresult('fail', "function scsi_preempt failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_preempt");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_preempt(host=>$host, device=>$DEVICE, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_preempt failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_preempt");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_preempt(host=>$host, device=>"abc");
            };

            if ($@ =~ /device abc not found/) {
                logresult('pass', "Got expected failure from function scsi_preempt: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_preempt: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_preempt(host=>$host, device=>"abc", conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_preempt: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_preempt: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_preempt(host=>$host, device=>$DEVICE, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_preempt failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_preempt");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_preempt(host=>$host, device=>$DEVICE, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_preempt failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_preempt");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_preempt(host=>$host, device=>"abc", filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_preempt: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_preempt: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_preempt(host=>$host);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }
    }

    #----------------------------------------------------------------
    # test scsi_select_primary
    #----------------------------------------------------------------
    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "scsi_select_primary") {
        if (!defined($ADAPTER)) {
            if ($DEVICE =~ /^\s*(\w+).*/) {
                $ADAPTER = $1;
            }
        }

        if ($FILERMODE eq "normal") {
            #test rsh connection + normal mode
            eval {
                $ret = scsi_select_primary(host=>$host, device=>$DEVICE, adapter=>$ADAPTER, filermode=>$FILERMODE);
            };

            if ($@) {
                logresult('fail', "function scsi_select_primary failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_select_primary");
            }

            #test rsh connection + normal mode  (default connection 
            #type and default connection mode
            eval {
                $ret = scsi_select_primary(host=>$host, device=>$DEVICE, adapter=>$ADAPTER);
            };

            if ($@) {
                logresult('fail', "function scsi_select_primary failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_select_primary");
            }

            #test console connection + normal mode
            eval {
                $ret = scsi_select_primary(host=>$host, device=>$DEVICE, adapter=>$ADAPTER, conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_select_primary failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_select_primary");
            }

            #test invalid device name over rsh connection
            eval {
                $ret = scsi_select_primary(host=>$host, device=>"abc", adapter=>$ADAPTER);
            };

            if ($@ =~ /device abc not found/) {
                logresult('pass', "Got expected failure from function scsi_select_primary: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_select_primary: $@");
            }

            #test invalid device name over console connection
            eval {
                $ret = scsi_select_primary(host=>$host, device=>"abc", adapter=>$ADAPTER, conntype=>"#console");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_select_primary: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_select_primary: $@");
            }
        } else {     #maintenance mode testing
            #test implicit console connection + maint mode
            eval {
                $ret = scsi_select_primary(host=>$host, device=>$DEVICE, adapter=>$ADAPTER, filermode=>"maint");
            };

            if ($@) {
                logresult('fail', "function scsi_select_primary failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_select_primary");
            }

            #test explicit console connection + maint mode
            eval {
                $ret = scsi_select_primary(host=>$host, device=>$DEVICE, adapter=>$ADAPTER, filermode=>"maint", conntype=>"#console");
            };

            if ($@) {
                logresult('fail', "function scsi_select_primary failed: $@");
            } else { 
                print Dumper($ret);
                logresult('pass', "Got expected result from function scsi_select_primary");
            }

            #test invalid device name in maint mode
            eval {
                $ret = scsi_select_primary(host=>$host, device=>"abc", adapter=>$ADAPTER, filermode=>"maint");
            };

            if ($@ =~ /device abc not found/ ) {
                logresult('pass', "Got expected failure from function scsi_select_primary: $@");
            } else { 
                logresult('fail', "Got unexpected result from function scsi_select_primary: $@");
            }

        }

        #test missing mandotary device name
        eval {
            $ret = scsi_select_primary(host=>$host, adapter=>$ADAPTER);
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }

        #test missing mandotary adapter name
        eval {
            $ret = scsi_select_primary(host=>$host, device=>"abc");
        };

        if ($@) {
            logresult('pass', "Got expected failure from function scsi_capacilty: $@");
        } else { 
            logresult('fail', "Got unexpected result from function scsi_capacilty: $@");
        }
    }
}

=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Testme

=head1 ATTRIBUTES

    ONTAP_COMMANDS = scsi

    ONTAP_ZAPI_APIS = none 

=head1 AUTHOR

    Sajan Agrawal (sajan@netapp.com)

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;

