# $Id: //depot/prod/test/Rsierranevada/lib/Storage/ShelfLog.pm#17 $

# Library of storage Shelf Log related functions.
## @summary Shelf Log - shelf logs library module
## @description
##  Library of routines for Shelf Log related tests.
## @synopsis
##
## use Storage::ShelfLog;
## @status TESTME
## @author divyesh@netapp.com
## @updated by darshan.murthy@netapp.com

#
#  7/12/13  A note motivation for inclusion of param "client_hostname" throughout this library, and a not about the dependencies which determine which
#  subroutines have this param..
#  previous to this edit, this library was making calls to the hostname() method, which returns the hostname of the current client.
#  Now, if a user was doing an ntest on a .thpl script, from the client that they were specifying on the CLI or in their test bed ,
#  then this wouldn't be a problem.  But suppose you were running an ntest on a .thpl and specifying on the CLI CLIENT=<some other client>... (a client other than the one you're on)
#  then if it made calls to this library, you would end up having a problem. So, instead of assuming that an outside script shuold be
#  running on a certain client, we need to have the option to specify which client to use.
#
#  Now... there are two distinct "islands" of subroutines in this library: subroutines that calls hostname() directly or through call stacks,
#  and those which make no calls to subroutines in a call stack that ultimately call hostname().    Those subroutines in the ladder category
#  are not equipped with the client_hostname option HOWEVER, it's important to note that for future changes of this library, if these dependencies change,
#  so should the parameters.
#
#  Currently, the following subroutines call hostname() DIRECTLY:
#    sub log_mount
#    sub log_umount
#    sub log_exist
#    sub log_mark_only
#    sub log_mark
#
#  the following subroutines call these subroutines:
#    sub log_search (calls log_mount, log_exist, log_umount)
#    sub log_exstract_records (log_mount, log_exist, log_umount)
#    sub log_search_stats (log_mount, log_exist)
#    sub log_mark (log_mount, log_umount)
#    sub log_mark_only (log_mount, log_umount)
#
#    so just keep this in mind if the script goes under any major changes.


=head1 NAME

     Storage::ShelfLog

=head1 SYNOPSIS

     use ShelfLog qw(log_mark log_read_line log_read_current_line log_read_previous_line log_read_next_line log_process_main_block log_read_first_line log_generate_records log_search_serials_csms log_search_strings log_search_serial_strings log_print_records log_search log_generate_stats log_process_stat_block log_search_stats log_search_element log_search_stat_records log_print_stats log_mount log_file_get_order );

     my $hash_ref = log_search(host => $FILER,...);

=head1 DESCRIPTION

     This is a library module for parsing the Shelf Logs.

=head1 ROUTINES

=cut

package Storage::ShelfLog;
use Tharn;
use strict;
use Tie::File;
use Volume qw(vol_root);
use mount;
use Sys::Hostname;
require Exporter;
use Data::Dumper;
use vars ( '@ISA', '@EXPORT_OK' );
@ISA       = ('Exporter');
@EXPORT_OK = qw(log_exist
    log_mark
    log_read_line
    log_read_current_line
    log_read_previous_line
    log_read_next_line
    log_process_main_block
    log_read_first_line
    log_generate_records
    log_search_serials_csms
    log_search_strings
    log_search_serial_strings
    log_print_records
    log_search
    log_generate_stats
    log_process_stat_block
    log_search_stats
    log_search_element
    log_search_stat_records
    log_print_stats
    log_mount
    log_file_get_order
    log_mark_only
    log_extract_records
);
my $timeout = 300;
my $talk = 0;#for debugging, toggle this on if you want some really detailed logs

use Carp qw(confess croak);
use Params::Validate qw(validate OBJECT SCALAR HASHREF ARRAYREF GLOB GLOBREF);
use Tharn qw(say script_result host logcomment logresult $Log param);

BEGIN {
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess( Tharn::script_result(@msgs) );
        }
    );
} ## end BEGIN

=head2 log_mount

     my $log_file = log_mount({ filer => $filer, filer_conn => $filer_connection, sudo_path => $sudo_path, shelf_type => $shelf_type});

This routine connects to the filer, shares the root volume and mounts the shelf log directory onto the local client based on the shelf type specified and returns the absolute path of the log file.  If the subroutine encounters any error it logs a fatal result and exits.  The filer_conn, sudo_path and shelf_type have default values rsh, /usr/local/bin and SAS respectively.

The following named arguments are accepted,

=over

=item filer

(Required) : The filer from which the logs has to be parsed.

=item filer_conn

(Optional) : The type of connection to be established with the Filer.  Default is RSH.

=item shelf_type

(Optional) : The shelf type to be parsed. Default is SAS

=item sudo_path

(Optional) : The path for the sudo command on the client. Default is /usr/local/bin

=item client_hostname
 
(Optional) : The name of the client. Default will be to the current host

=back

=cut

sub log_mount {
    my (%opts) = validate(
        @_,
        {   filer => { type => SCALAR },
            filer_conn =>
                { type => SCALAR, optional => 1, default => "#rsh" },
            sudo_path => {
                type     => SCALAR,
                optional => 1,
                default  => "/usr/local/bin"
            },
	    client_hostname => {type => SCALAR, optional => 1},
            shelf_type => { type => SCALAR, optional => 1, default => "ESAS" }
        }
    );
    logcomment("Mounting the log directory from filer");
    my $filer      = $opts{filer};
    my $filer_conn = $opts{filer_conn};
    my $shelf_type = $opts{shelf_type};

  
    my $client_hostname;
    if( ! defined( $opts{client_hostname} )){
	    $client_hostname = hostname();
    }
    else{
	    $client_hostname = $opts{client_hostname};
    }
    #my $client_hostname = hostname() if(! defined( $opts{client_hostname} ) );
    my ($client) = split( /\./, $client_hostname );

    my $sudo = "$opts{sudo_path}" . "/sudo";
    my $log_file;

    my $client_connection = connect("$client#rlogin");
    my $say_timeout       = 180;

    my $filer_connection = connect("$filer $filer_conn");
    my $filer_root_vol   = vol_root($filer_connection);

    say( $filer_connection,
        "priv set -q diag; exportfs -p sec=sys,rw,anon=0 /vol/$filer_root_vol",
        '-timeout',
        $say_timeout
    );
    my $mountpoint = mount( $filer, "/", client => $client )
        or logresult( 'fatal', "Could not mount $filer on $client" );
    logcomment("Log file mount point is $mountpoint") if($talk);

    if ( $shelf_type =~ /^(ata|ATA|sata|SATA)$/ ) {
        $log_file = "$mountpoint/etc/log/shelflog/shelflog_ata";
    } elsif ( $shelf_type =~ /^(esh|esh2|esh4|ESH|ESH2|ESH4|FC)$/ ) {
        $log_file = "$mountpoint/etc/log/shelflog/shelflog_esh";
    } elsif ( $shelf_type =~ /^(sas|SAS)$/ ) {
        $log_file = "$mountpoint/etc/log/shelflog/shelflog_iom_master";
    } elsif ( $shelf_type =~ /^(esas|ESAS)$/ ) {
        $log_file = "$mountpoint/etc/log/shelflog/shelflog_iom_master";
    } else {
        logresult( 'fatal',
            "Failed to recognize the shelf type $shelf_type" );
    }
    return $log_file;
} ## end sub log_mount

sub log_file_get_order {

    my (%opts) = validate(
        @_,
        {   log_start_marker => { type => HASHREF, optional => 1 },
            log_end_marker   => { type => HASHREF, optional => 1 }
        }
    );
    logcomment("Checking the log file order") if($talk);

    my $log_start_marker = $opts{log_start_marker};
    my $log_end_marker   = $opts{log_end_marker};

    if ( !defined( $log_start_marker->{log_reference} ) ) {
        $log_start_marker->{log_reference} = 0;
    }

    if ( $log_start_marker->{no_file} == 1 ) {
        $Log->warn(
            "log_file_get_order: No file defined in start marker, current Shelf log file would be parsed"
        );
        my $rvalue = [ $log_start_marker->{log_file} ];
        return $rvalue;
    } ## end if ( $log_start_marker...

    if ( $log_start_marker->{log_reference} eq "0" ) {
        $Log->warn(
            "The start marker doesnt have a reference, the current shelflog file would be taken for parsing"
        );
        my $rvalue = [ $log_start_marker->{log_file} ];
        return $rvalue;
    } ## end if ( $log_start_marker...

    my $log_files = [ $log_start_marker->{log_file} ];
    for ( my $i = 0; $i <= 10; $i++ ) {
        my $next_file = $log_start_marker->{log_file} . '.' . $i;
        push @$log_files, $next_file;
    }
    my $ordered_list = ();

    my $match_flag = 0;
    for ( my $i = 0; $i <= $#$log_files; $i++ ) {
        my @log_array;
        use Fcntl 'O_RDONLY';
        tie @log_array, 'Tie::File', $$log_files[$i],
            mode      => O_RDONLY,
            autochomp => 0;

        #print "$$log_files[$i]\n";
        for ( my $j = 0; $j <= $#log_array; $j++ ) {
            my $line = $log_array[$j];
            chomp($line);

            #print "$line\n";
            if ( $line =~ /Timestamp:/ ) {
                if ( $line eq $log_start_marker->{log_reference} ) {

                    #print "$line\n";
                    $match_flag = 1;
                    push @$ordered_list, $$log_files[$i];
                } ## end if ( $line eq $log_start_marker...
                last;
            } ## end if ( $line =~ /Timestamp:/)
        } ## end for ( my $j = 0; $j <= ...
        untie @log_array;

        #print "\n";
        if ( $match_flag == 1 ) {
            last;
        }
        push @$ordered_list, $$log_files[$i];
    } ## end for ( my $i = 0; $i <= ...
    @$ordered_list = reverse(@$ordered_list);
    if ( $match_flag != 1 ) {
        logresult( "FAIL",
            "Unable to find the file marked by the log_start_marker in the default and shelflog files 0 through 10"
        );
        return 1;
    } ## end if ( $match_flag != 1 )

    if (( $#$ordered_list == 0 )
        && ( $log_start_marker->{line_number}
            == $log_end_marker->{line_number} )
        )
    {
        logcomment("The log file is not updated with any data");
    } ## end if ( ( $#$ordered_list...
    return $ordered_list;
} ## end sub log_file_get_order

sub log_umount {
    my (%opts) = validate( @_, { log_file => { type => SCALAR },
	    			 client_hostname => {type => SCALAR, optional => 1}
			} );
    my $log_file = $opts{log_file};

    my $client_hostname;
    if(! defined( $opts{client_hostname} ) ){
	    $client_hostname = hostname();
    }
    else{
	    $client_hostname = $opts{client_hostname};
    }
    my ($client) = split( /\./, $client_hostname );
    my $client_connection = connect("$client#rlogin");
    my $say_timeout       = 180;
    my ($mount_point) = split( /\/etc\/log\/shelflog/, $log_file );
    logcomment("the mount point is: $mount_point");
    #Note: put in an eval block so if we haven't actually mount the script won't stop.  Example of when this could happen (as of 7/16/13) in the log_search subroutine, 
    #if filer were defined but so was file_log, it would attempt to unmount by calling this sub without ever having called log_mount
    #my $return_value;
    eval{
	    #$return_value = umount( $mount_point, client => $client, force => 1 );
       umount( $mount_point, client => $client, force => 1 );
    };
    if($@){
	    logcomment("attempting to unmount but it's not there");
	    print $@;
    }
    #warn $@ if $@;
    #if ( $return_value ne "1" ) {
    #    logresult( "FAIL", "Unmount failed" );
    #    return 1;
    #} else {
    print "returning 0\n";
        return 0;
    #}
} ## end sub log_umount

=head2 log_exist

     my $check = log_exist({ file => $log_file, shelf_logs => $shelf_logs});

This routine checks the existance and accessability of log files.  It returns 0 on success and 1 on failure.

$VAR1 = {
          'total_files' => '2',
          'log_files' => [
                           {
                             'end_line' => 2670,
                             'name' => '/u/darshanm/Automation/shelflog_sas.0',
                             'start_line' => 0,
                             'line_count' => 2670
                           },
                           {
                             'end_line' => 1171,
                             'name' => '/u/darshanm/Automation/shelflog_sas.1',
                             'start_line' => 1,
                             'line_count' => 1170
                           }
                         ]
        };


The following named arguments are accepted,

=over

=item file

(Required) : The log file name with absolute path.

=item shelf_logs

(Required) : The reference to the data structure the would be created after verification of log files for generating records.

=item client_hostname

(optional) : The name of the client to connect with.  If not specified, default to the current client
=back

=cut

sub log_exist {
    my (%opts) = validate(
        @_,
        {   log_files  => { type => ARRAYREF },
            shelf_logs => { type => HASHREF },
            filer      => { type => SCALAR, optional => 1 },
            timeout    => { type => SCALAR, optional => 1, default => 60 },
	    client_hostname => {type => SCALAR, optional => 1 }
        }
    );

    logcomment(">>> Checking the log files for existence and accessibility (ShelfLog::log_exist)");

    my $flag       = 0;
    my $shelf_logs = $opts{shelf_logs};

    my @log_files;
    if ( defined $opts{filer} ) {
        if ( defined $shelf_logs->{log_start_marker} ) {
            my $parameters;
            $parameters->{log_start_marker} = $shelf_logs->{log_start_marker};
            if ( defined $shelf_logs->{log_end_marker} ) {
                $parameters->{log_end_marker} = $shelf_logs->{log_end_marker};
            }
	    logcomment(">>> calling ShelfLog::log_file_get_order and sending the following: (ShelfLog::log_exist)") if($talk); 
	    print Dumper %{$parameters} if($talk);
            my $filer_log_files = log_file_get_order($parameters);
	    logcomment(">>> returning from log_file_get_order with the following filer_log_files hash (ShelfLog::log_exist)") if($talk);
	    print Dumper $filer_log_files if($talk);
            @log_files = @$filer_log_files;
        } else {
            logcomment(
                "log_start_marker is not defined, current shelflog file would be parsed"
            );
            @log_files = @{ $opts{log_files} };
        } ## end else [ if ( defined $shelf_logs...
    } else {
        @log_files = @{ $opts{log_files} };
    }

    if ( $#log_files < 0 ) {
        logresult( "FAIL", "No log files defined" );
        return 1;
    }

    foreach my $file (@log_files) {
	logcomment(">>> current file in log file: $file (ShelfLog::log_exist)") if($talk);
        if ( ( -e $file ) && ( -r $file ) ) {
            logcomment "File  : $file\nStatus: OK";
        } else {
            logcomment "File  : $file\nStatus: NOK";
            $flag = 1;
        }
    } ## end foreach my $file (@log_files)

    if ( $flag == 1 ) {
        logresult( "FAIL",
            "Please check the log files for existence and accessibility" );
        return 2;
    }

    my $client_hostname;
    if( ! defined( $opts{client_hostname} )){
            $client_hostname = hostname();
    }
    else{
            $client_hostname = $opts{client_hostname};
    }
    #my $client_hostname = hostname() if(! defined( $opts{client_hostname} ) );
    my ($client) = split( /\./, $client_hostname );
    my $connection     = connect("$client#rsh");
    my @log_references = ();

    for ( my $i = 0; $i <= $#log_files; $i++ ) {
	logcomment(">>> in for loop to generated log file details for each log filei, iteration $i. (ShelfLog::log_exist)") if($talk);
        my $log_file_details = {};

        my ( undef, undef, $result )
            = say( $connection, "wc -l $log_files[$i]", -timeout, $timeout );
        if ( $result =~ /No such file or directory/ ) {
            return 0;
        }
        my ($count) = grep ( /$log_files[$i]/, split( /\r?\n/, $result ) );
        $count =~ s/(\s*)(\d+)(.*)$/$2/g;
        if ( $count !~ /\d+/ ) {
            $flag = 1;
        }

        $log_file_details->{name}       = $log_files[$i];
        $log_file_details->{line_count} = $count;
        $log_references[$i]             = $log_file_details;
	logcomment(">>> and the generated log file details: (ShelfLog::log_exist)") if($talk);
	print Dumper $log_file_details;
    } ## end for ( my $i = 0; $i <= ...
    $shelf_logs->{log_files}   = \@log_references;
    $shelf_logs->{total_files} = $#{ $shelf_logs->{log_files} } + 1;

    for ( my $i = 0; $i <= $#log_files; $i++ ) {

        #print "$shelf_logs->{log_files}[$i]->{name}\n";
        if ( $i == 0 ) {
            $shelf_logs->{log_files}[$i]->{start_line} = 0;
        } else {
            $shelf_logs->{log_files}[$i]->{start_line}
                = $shelf_logs->{log_files}[ $i - 1 ]->{end_line} + 1;
        }

        $shelf_logs->{log_files}[$i]->{end_line}
            = $shelf_logs->{log_files}[$i]->{start_line}
            + $shelf_logs->{log_files}[$i]->{line_count} - 1;
    } ## end for ( my $i = 0; $i <= ...

    $shelf_logs->{total_lines} = 0;
    for ( my $i = 0; $i <= $#log_files; $i++ ) {
        $shelf_logs->{total_lines} = $shelf_logs->{total_lines}
            + $shelf_logs->{log_files}[$i]->{line_count};
    }

    if ( $flag == 1 ) {
        logresult( "FAIL",
            "Please check the log files exist and have read permissions" );
        return 1;
    }
    logcomment(">>> returning false (0) from log_exists.  The following was populated in the shelf_log hash (ShelfLog::log_exist)") if($talk);
    print Dumper $shelf_logs;
    return 0;
} ## end sub log_exist

sub log_mark_only {
    my (%opts) = validate(
        @_,
        {   log_file   => { type => SCALAR },
            timeout    => { type => SCALAR, optional => 1 },
            shelf_type => { type => SCALAR, optional => 1 },
	    client_hostname => {type => SCALAR, optional => 1}
        }
    );
    logcomment(">>> Marking the current line from the log file (ShelfLog::log_mark_only)") if($talk);
    my $log_file        = $opts{log_file};
    my $client_hostname;
    if( ! defined( $opts{client_hostname} )){
            $client_hostname = hostname();
    }
    else{
            $client_hostname = $opts{client_hostname};
    }
    #my $client_hostname = hostname() if(! defined( $opts{client_hostname} ) );
    my ($client) = split( /\./, $client_hostname );
    my $connection = connect("$client#rsh");

    my $change_flag  = 1;
    my $change_count = 0;
    my $old_count    = -100;
    my $rvalue;

    while ( ( $change_flag == 1 ) && ( $change_count < 5 ) ) {
        my ( undef, undef, $result )
            = say( $connection, "wc -l $log_file", -timeout, $timeout );

        if ( $result =~ /No such file or directory/ ) {
            $rvalue->{no_file}       = 1;
            $rvalue->{line_number}   = 0;
            $rvalue->{log_reference} = 0;
            $Log->warn("log_mark_only: Shelf log file was not found");
            return $rvalue;
        } ## end if ( $result =~ /No such file or directory/)
        my ($count) = grep ( /$log_file/, split( /\r?\n/, $result ) );
        $count =~ s/(\s*)(\d+)(.*)$/$2/g;

        $rvalue->{log_file} = $log_file;
        $rvalue->{no_file}  = 0;
        if ( $count !~ /\d+/ ) {
            $rvalue->{line_number} = 0;
        } else {
            $rvalue->{line_number} = $count;
            if ( $count != $old_count ) {
                $change_flag = 1;
                $change_count++;
                $old_count = $count;
            } else {
                $change_flag = 0;
            }
        } ## end else [ if ( $count !~ /\d+/ )
        sleep 1;
    } ## end while ( ( $change_flag ==...

    $rvalue->{change_count} = $change_count;

    my @log_array;
    use Fcntl 'O_RDONLY';
    tie @log_array, 'Tie::File', $log_file, mode => O_RDONLY, autochomp => 0;

    for ( my $i = 0; $i <= $#log_array; $i++ ) {
        my $line = $log_array[$i];
        chomp($line);
        if ( $line =~ /Timestamp:/ ) {
            $rvalue->{log_reference} = $line;
            last;
        }
    } ## end for ( my $i = 0; $i <= ...
    untie @log_array;

    if ( !( defined $rvalue->{log_reference} ) ) {
        logcomment(
            "A reference timestamp for marking the file couldn't be found, if this is used as start marker then only one file would be considered for processing"
        );
        $rvalue->{log_reference} = "0";
    } ## end if ( !( defined $rvalue...

    logcomment(">>> exiting subroutine log_mark_only and will send back the followin rvalue hash (ShelfLog::log_mark_only)") if($talk);
    print Dumper $rvalue;
    return $rvalue;
} ## end sub log_mark_only

=head2 log_mark

     my $line_number = log_mark({file => $log_file timeout => $timeout});

This routine returns the current count of lines in the log file.  This can be used to mark the starting and ending points in a log file for analysis.

=over

=item file

(Required) : The log file name with absolute path

=item timeout

(Optional) : The timeout value for running the command on the local client. Default is 60 seconds.

=back

=cut

sub log_mark {
    my (%opts) = validate(
        @_,
        {   filer      => { type => SCALAR },
            timeout    => { type => SCALAR, optional => 1 },
            filer_conn => { type => SCALAR, optional => 1 },
            sudo_path  => { type => SCALAR, optional => 1 },
            shelf_type => { type => SCALAR, optional => 1 },
	    client_hostname => {type => SCALAR, optional => 1 }
        }
    );
    logcomment(">>> Marking the current line from the log file (ShelfLog::log_mark)") if($talk);

    my $log_mount_parameters = {};
    $log_mount_parameters->{filer} = $opts{filer};
    if ( defined $opts{timeout} ) {
        $log_mount_parameters->{timeout} = $opts{timeout};
    }

    if ( defined $opts{filer_conn} ) {
        $log_mount_parameters->{filer_conn} = $opts{filer_conn};
    }

    if ( defined $opts{sudo_path} ) {
        $log_mount_parameters->{sudo_path} = $opts{sudo_path};
    }

    if ( defined $opts{shelf_type} ) {
        $log_mount_parameters->{shelf_type} = $opts{shelf_type};
    }

    if ( defined $opts{client_hostname} ) {
        $log_mount_parameters->{client_hostname} = $opts{client_hostname};
    }

    logcomment(">>> preparing to call log_mount and will send the following: (ShelfLog::log_mark)");
    print Dumper $log_mount_parameters;
    my $log_file        = log_mount($log_mount_parameters);
    
    my $client_hostname;
    if( !defined( $opts{client_hostname} )){
	    $client_hostname = hostname();
    }
    else{
	    $client_hostname = $opts{client_hostname};
    }
    my ($client) = split( /\./, $client_hostname );
    my $connection = connect("$client#rsh");
    my ( undef, undef, $result )
        = say( $connection, "wc -l $log_file", -timeout, $timeout );
    my $rvalue;

    if ( $result =~ /No such file or directory/ ) {
        $rvalue->{no_file}       = 1;
        $rvalue->{line_number}   = 0;
        $rvalue->{log_reference} = 0;
        $Log->warn("log_mark: Shelf Log file was not found");
        return $rvalue;
    } ## end if ( $result =~ /No such file or directory/)
    my ($count) = grep ( /$log_file/, split( /\r?\n/, $result ) );
    $count =~ s/(\s*)(\d+)(.*)$/$2/g;

    $rvalue->{log_file} = $log_file;
    $rvalue->{no_file}  = 0;
    if ( $count !~ /\d+/ ) {
        $rvalue->{line_number} = 0;
    } else {
        $rvalue->{line_number} = $count;
    }
    my @log_array;
    use Fcntl 'O_RDONLY';
    tie @log_array, 'Tie::File', $log_file, mode => O_RDONLY, autochomp => 0;

    for ( my $i = 0; $i <= $#log_array; $i++ ) {
        my $line = $log_array[$i];
        chomp($line);
        if ( $line =~ /Timestamp:/ ) {
            $rvalue->{log_reference} = $line;
            last;
        }
    } ## end for ( my $i = 0; $i <= ...
    untie @log_array;

    if ( !( defined $rvalue->{log_reference} ) ) {
        $Log->warn(
            "A reference timestamp for marking the file couldn't be found, if this is used as start marker then only one file would be considered for processing"
        );
        $rvalue->{log_reference} = "0";
    } ## end if ( !( defined $rvalue...

    my $log_umount_parameters;
    $log_umount_parameters->{log_file} = $log_file;
    $log_umount_parameters->{client_hostname} = $client_hostname if( defined $opts{client_hostname} );
    logcomment(">>> Preparing to call log_umount subroutine and will send the following params: (ShelfLog::log_mark)");
    print Dumper $log_umount_parameters;
    $result = log_umount( $log_umount_parameters );

#	if($result != 0)
#	{
#		logresult("FAIL","Unable to unmount $log_file"); #TBC can we put a warning here??
#	}
    logcomment(">>> Exiting log_mark subroutine and will return rvalue hash as follows: (ShelfLog::log_mark)") if($talk);
    print Dumper $rvalue if($talk);
    return $rvalue;
} ## end sub log_mark

=head2 log_read_current_line

     my $line = log_read_current_line({shelf_logs => $shelf_logs});

This routine returns the line represented by the line number $shelf_logs->{current_line}.  If this is outside the scope of the defined start and end markers of the log file, it returns an undef.

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=back

=cut

sub log_read_current_line {
    my (%opts) = validate( @_, { shelf_logs => { type => HASHREF }} );
    logcomment(">>> in log read current line (ShelfLog::log_read_current_line)") if($talk);
    #print Dumper %opts;
    my $shelf_logs = $opts{shelf_logs};
    my $line;
    logcomment(">>> preparing to call log_read_line, will set result to my variable line (ShelfLog::log_read_current_line)") if($talk);
    if (!(  $line = log_read_line(
                {   shelf_logs => $shelf_logs,
                    line       => $shelf_logs->{current_line}
                }
            )
        )
        )
    {
	logcomment(">>> an empty string was returned from the call to log_read_line, so we'll exit the subroutine now with an undef (ShelfLog::log_read_current_line)") if($talk);
        return undef;
    } else {
	logcomment(">>> a non-empty string was returned from the call to log_read_line, so we'll exit the subroutine now with the line-> $line (ShelfLog::log_read_current_line)") if($talk);
        return $line;
    }
} ## end sub log_read_current_line

=head2 log_read_next_line

     my $line = log_read_next_line({shelf_logs => $shelf_logs});

This routine increments the line number $shelf_logs->{current_line} and returns the line referenced by $shelf_logs->{current_line} from the log file.  The routine returns an undef, if the line is outside the scope of the defined start and end markers of the log file or on encountering an EOF.

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=back

=cut

sub log_read_next_line {
    my (%opts) = validate( @_, { shelf_logs => { type => HASHREF }} );
    my $shelf_logs = $opts{shelf_logs};

    logcomment(">>> Reading next line in log... (ShelfLog::log_read_next_line)") if($talk);
    #my $current_line = $shelf_logs->{current_line} + 1;
    $shelf_logs->{current_line}++;
    my $current_line = $shelf_logs->{current_line};
    my $line;
    if (!(  $line = log_read_line(
                { shelf_logs => $shelf_logs, line => $current_line }
            )
        )
        )
    {
	logcomment(">>> Received empty string from sub call to log_read_line, so exiting sub now with undef (ShelfLog::log_read_next_line)") if($talk);
        return undef;
    } else {
	logcomment(">>> Received non-empty string from sub call to log_read_line, so exit with the line -> $line (ShelfLog::log_read_next_line)") if($talk);
        return $line;
    }
} ## end sub log_read_next_line

=head2 log_read_previous_line

     my $line = log_read_previous_line({shelf_logs => $shelf_logs});

This routine decrements the line number $shelf_logs->{current_line} and returns the line referenced by $shelf_logs->{current_line} from the log file.  The routine returns an undef, if the line is outside the scope of the defined start and end markers of the log file or on encountering an EOF.

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=back

=cut

sub log_read_previous_line {
    my (%opts) = validate( @_, { shelf_logs => { type => HASHREF }} );
    my $shelf_logs = $opts{shelf_logs};

    logcomment(">>> Attempting to read previous line.  Current lnie set to $shelf_logs->{current_line} (ShelfLog:: log_read_previous_line)") if($talk);
    #my $current_line = $shelf_logs->{current_line} - 1;
    $shelf_logs->{current_line}--;
    my $current_line = $shelf_logs->{current_line};
    my $line;
    if (!(  $line = log_read_line(
                { shelf_logs => $shelf_logs, line => $current_line }
            )
        )
        )
    {
	logcomment(">>> Received empty string from call to log_read_line, so exit sub now with undef. (ShelfLog::log_read_previous_line)") if($talk);
        return undef;
    } else {
	logcomment(">>> Receive non-empty string from call to log_read_line, so exit with the following line -> $line (ShelfLog::log_read_previous_line)") if($talk);
        return $line;
    }
} ## end sub log_read_previous_line

=head2 log_read_first_line

     my $line = log_read_first_line({shelf_logs => $shelf_logs});

This routine is used to access the first line from the log file.  This routine sets the $shelf_logs->{current_line} to start marker in $shelf_logs data structure, opens the file if its not open and returns the line referenced by $shelf_logs->{current_line}.  The routine returns an undef on encountering an EOF.

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=back

=cut

sub log_read_first_line {
    my (%opts) = validate( @_, { shelf_logs => { type => HASHREF }} );
    my $shelf_logs   = $opts{shelf_logs};
    my $current_line = 1;

    logcomment(">>>  Attempting to read first line in log file. (ShelfLog::log_read_first_line)") if($talk);
    if ( defined $shelf_logs->{log_start_marker}->{line_number} ) {
        $current_line = $shelf_logs->{log_start_marker}->{line_number} + 1;
    }

    my $line;
    if (!(  $line = log_read_line(
                { shelf_logs => $shelf_logs, line => $current_line }
            )
        )
        )
    {
	logcomment(">>> empty string was returned.  Exit sub with undef (ShelfLog::log_read_first_line)") if($talk);
        return undef;
    } else {
	logcomment(">>> non-emnpty string was returned.  exit with line -> $line (ShelfLog::log_read_first_line)") if($talk);
        return $line;
    }
} ## end sub log_read_first_line

=head2 log_goto_line

     log_goto_line({shelf_logs => $shelf_logs, line => $line_number});

This routine sets the line pointer to $line_number.

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=item line

(Required) : The line number in the log file.

just a note:
notice this subroutine does not take in the client_hostname param which is being taken in in other subroutines.
The reason why is because there are no calls made to external subroutines from here.  The reason why client_hostname
was needing to be added was because scripts were calling this library externally, and some of those scripts needed
to use CLI specified clients.  They were calling subs in this library, which were calling log_mount internally,
which was then selecting the current client and trying to connect via that client.  Since it is the "culprit" but
not really being called from the scripts, we need to equip all scripts in this file which call log_mount with a host name
param.  We can't just make a global parameter because we have no gauarantee which subroutines are being called.
but anyway, FOR NOW, there's no way this subroutine can cause a problem as it's entirely self contained.

=back

=cut

sub log_goto_line {
    my (%opts) = validate(
        @_,
        {   shelf_logs => { type => HASHREF },
            line       => { type => SCALAR }
        }
    );
    my $shelf_logs = $opts{shelf_logs};
    $shelf_logs->{current_line} = $opts{line};

} ## end sub log_goto_line

=head2 log_read_line

     my $line = log_read_line({shelf_logs => $shelf_logs, line => $line_number});

This routine automatically opens the log file if it is not open and returns the line referenced by $line_number from the log file.  An undef is returned if the $line_number is outside the scope of start and end markers or when an EOF is encountered.

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=item line

(Required) : The line number in the log file.

=back

=cut

sub log_read_line {
    my (%opts) = validate(
        @_,
        {   shelf_logs => { type => HASHREF },
            line       => { type => SCALAR },
        }
    );
    logcomment(">>> Attempting to read a line in the log. shelf_logs too large to Dump, but line passed in is $opts{line} (ShelfLog::log_read_line)") if($talk);
    #print Dumper %opts;
    my $shelf_logs   = $opts{shelf_logs};
    my $current_line = $opts{line} - 1;
    my $temp         = $shelf_logs->{log_files};
    my @log_files    = @$temp;
    my $current_file_pos;

    if (   ( $current_line < 0 )
        || ( $current_line >= $shelf_logs->{total_lines} ) )
    {
	logcomment(">>> Exiting sub with undef because current_line specified < 0, or current_line specified ($current_line) is >= number of lines in log ($shelf_logs->(total_line}) (ShelfLog::log_read_line)") if($talk);
        return undef;
    }

    for ( my $i = 0; $i <= $#log_files; $i++ ) {
        if (   ( $current_line >= $log_files[$i]->{start_line} )
            && ( $current_line <= $log_files[$i]->{end_line} ) )
        {
            $current_file_pos = $i;
        }
    } ## end for ( my $i = 0; $i <= ...


    if ( $current_file_pos !~ /\d+/ ) {
        $Log->warn(
            "Script is trying to access line outside the scope of files: Dumping current data from variables"
        );
	#print Dumper($shelf_logs);
        print "\ncurrent line: $current_line\n";
        print "current_file_pos: $current_file_pos\n";
        logresult( "Script",
            "This is an unexpected error.  Please contact the author with the logs: line 558"
        );
        return undef;
    } ## end if ( $current_file_pos...

    if ( defined $shelf_logs->{current_file_pos} ) {
    	print(">>> shelf logs defines current file pos.. (ShelfLog::log_reaed_line)");
	  if ( $shelf_logs->{current_file_pos} != $current_file_pos ) {
		  print("but its not equal to $current_file_pos..");
            if ( defined $shelf_logs->{fh} ) {
		    print("but shelf_log dash fh is defined..");
                my @log_array = @{ $shelf_logs->{fh} };
                untie @log_array;
                delete $shelf_logs->{fh};
            }
            my @log_array;
            $shelf_logs->{current_file_pos} = $current_file_pos;
            use Fcntl 'O_RDONLY';
	    print("will tie..");
            tie @log_array, 'Tie::File',
                $shelf_logs->{log_files}[ $shelf_logs->{current_file_pos} ]
                ->{name},
                mode      => O_RDONLY,
                autochomp => 0;
		print("done tying..");
            $shelf_logs->{fh} = \@log_array;
        } else {
		print("else: its equal to $current_file_pos..");
            if ( !( defined $shelf_logs->{fh} ) ) {
		    print("shelf logs not defined with fh..");
                my @log_array;
                $shelf_logs->{current_file_pos} = $current_file_pos;
                use Fcntl 'O_RDONLY';
                tie @log_array, 'Tie::File',
                    $shelf_logs->{log_files}
                    [ $shelf_logs->{current_file_pos} ]->{name},
                    mode      => O_RDONLY,
                    autochomp => 0;
                $shelf_logs->{fh} = \@log_array;
            } ## end if ( !( defined $shelf_logs...
        } ## end else [ if ( $shelf_logs->{current_file_pos...
    } else {
        if ( defined $shelf_logs->{fh} ) {
            my @log_array = @{ $shelf_logs->{fh} };
            untie @log_array;
            delete $shelf_logs->{fh};
        }
        my @log_array;
        $shelf_logs->{current_file_pos} = $current_file_pos;
        use Fcntl 'O_RDONLY';
        tie @log_array, 'Tie::File',
            $shelf_logs->{log_files}[ $shelf_logs->{current_file_pos} ]
            ->{name},
            mode      => O_RDONLY,
            autochomp => 0;
        $shelf_logs->{fh} = \@log_array;
    } ## end else [ if ( defined $shelf_logs...

    if ( defined $shelf_logs->{log_start_marker} ) {
	    print("shelf logs defines log start marker..");
        if ( $shelf_logs->{log_files}[ $shelf_logs->{current_file_pos} ]
            ->{name} eq $shelf_logs->{log_start_marker}->{name} )
        {
	print("if statementI..");
            if ( ( $current_line + 1 )
                <= $shelf_logs->{log_start_marker}->{line_number} )
            {
		    print("returning undef..");
                return undef;
            }
        } ## end if ( $shelf_logs->{log_files...
    } ## end if ( defined $shelf_logs...

    my $log_array = $shelf_logs->{fh};
    $shelf_logs->{current_line} = $current_line + 1;

    print("Start for..");
    for ( my $i = 0; $i < $shelf_logs->{current_file_pos}; $i++ ) {
	    print("in for..$i..");
        $current_line
            = $current_line - $shelf_logs->{log_files}[$i]->{line_count};
	    print("for current line is $current_line\n");
    }

    if ( defined $shelf_logs->{log_end_marker} ) {
	    print("shelf loglog_end_marker defined..");
        if ( $shelf_logs->{log_files}[ $shelf_logs->{current_file_pos} ]
            ->{name} eq $shelf_logs->{log_end_marker}->{name} )
        {
            if ( ( $current_line + 1 )
                > $shelf_logs->{log_end_marker}->{line_number} )
            {
		    print("returning undef 2..");
                return undef;
            }
        } ## end if ( $shelf_logs->{log_files...
    } ## end if ( defined $shelf_logs...

    if ( $current_line > $#$log_array ) {
	    print("last if statement..");
        if ( $shelf_logs->{current_file_pos}
            < ( $shelf_logs->{total_files} - 1 ) )
        {
            $Log->warn(
                "Script is trying to access line outside the scope of files: Dumping current data from variables"
            );
	    #print Dumper($shelf_logs);
            print "\ncurrent_line: $current_line\n";
            print "current_file_pos: $current_file_pos\n";
            logresult( "Script",
                "This is an unexpected error.  Please contact the author with the logs: line 653"
            );
            return undef;
        } else {
		print("return undef3..");
            return undef;
        }
    } ## end if ( $current_line > $#$log_array)
    logcomment(">>> Returning from log_read_line with the following $$log_array[$current_line] (ShelfLog::log_read_line)") if($talk);
    return $$log_array[$current_line];
} ## end sub log_read_line

=head2 log_error_check

     my $pos = log_error_check($line, $search_pattern);

This routine finds the position of $search_pattern in $line and returns the position of the first character after the string $search_pattern in $line.  If the $search_pattern is not found, it exits from the script.

=over

=item line

(Required) : The line in which the search has to be done.

=item search

(Required) : The search pattern.

=back

=cut

sub log_search_error_check($$) {

    my ( $line, $search ) = @_;

    logcomment(">>> Attempting a log_search_error_check for line: -> ( $line ) and for search value -> ( $search )  (ShelfLog::log_search_error_check)") if($talk);
    my $position = index( $line, $search );
    if ( $position == -1 ) {
        print "Search value: $search not located in requested line to search $line\nexiting...\nPotential causes: the log file might be corrupt, or DATA Ontap might have changed to include different fields in the log file\n";
        exit(4);
    }
    $position = $position + length($search);
    my $length=length($search);
	logcomment(">>> Exiting log_search_error_check with value $position (ShelfLog::log_search_error_check)") if($talk);
    return $position;
} ## end sub log_search_error_check($$)

=head2 log_process_main_block

   @OVERVIEW:

   Subroutine which takes in a log file and parses out relevant information.

   @RETURNS:

   a hashref \%parseTable with key/value paris as param names and their parsed data, for a single shelf in the log file

   @UPDATING THIS SCRIPT

   For future versions of ONTAP that might change the way this block of logging is done, this code has been simplified so that the programmer only needs
   to specify three variables, which are marked in the sub below.  They are as follows, with an example given below:
    
     1) $delim:         : A string which will match the delimeter that lets us know when this block is over
     2) %expected_Vars  : A hash where the keys are the key values we send back in %parseTable, and corresponding values are regex designed so that
     			  data matched as a result of the regex will be sent back as the value for that key in %parseTable.  The data matched in association with the regex
			  will be the data immediately to the right of the regex, continuing on to either the next matching regex or the end of the line, whichever occurs first.
			  The regex values must conform to three rules:
    
                               1) these values should be stored in the format qr/regex/
                                  (see the following perl documentation to learn of this construct: http://perldoc.perl.org/perlop.html#Regexp-Quote-Like-Operators
                               2) Desired send back values must lay directly to the right of each specified regex (disregard whitespace)
                                  (example: if log output is Channel: 0a and you desire to keep the field '0a', then you would NOT want to use as
                                  your regex 'Channel', because you would end up getting ': 0a'.  Instead you'd want to use as your regex 'Channel:' or 'Channel: ')
                               3) regex patterns must be mutually exclusive and not overlap, nor must patterns overlap with desired values
                                  (meaning, don't use a regex that will match in places other than the desired place, and don't use a regex which contains one of the values you want to send back)
                                  (example: if one line of log output is 'Shelf Serial Number: xxxx' and another is 'Module Serial Number: xxx' you wouldn't
                                  want to use 'Shelf Serial Number' and the other as 'Serial Number'.  instead you'd want to make them mutually exclusive and use
                                  'Shelf Serial Number' and 'Module Serial Number'.)
                                  (example 2: if log output is 'Module X Serial Number: xxxxx' and you desire to get the value 'X' for module and 'xxxx' for module_serial_number,
                                  you wouldn't want to use the regex 'Module \D Serial Number' because this contains one of the values you actually want to send back)
    
      3) @allowed_ToSkip: A list of keys from %expected_Vars that we are allowed to skip
                          (for example, say some systems have the log_ID line but others don't and we don't want a fail if log_ID is never matched)
     /////////////
     EXAMPLE
     //////////////
    
     Suppose in the current version of ONTAP the shelf log output looks like this:
    
     --------------------------------------------------------
    
      Shelflog start time: Thu Jul 11 17:04:20 GMT 2013
      Controller Id: 6a7ab85b
      Channel: 0a Shelf: 1 Module type: IOM6 Firmware rev: 0162
      Shelf product id: DS224IOM6
      Shelf Serial Number: 9401089463
      Module A Serial Number: 8002276940
      Log ID: 9401089463_A_f7
      Timestamp: Tue Jul 16 21:13:31 PDT 2013
    
      -------------------------------------------------------
    
      Suppose your goal is to send back a hash populated with the following information as follows:
    
      my %parsedHash = (
                               channel => 0a,
                               shelf_id => 1,
                               module_type = IOM6,
                               firmare = 0162,
                               shelf_product_id => DS224IOM6,
                               shelf_serial => 941089463,
                               module => A,
                               module_serial => 8002276940,
                               timestamp => Tue Jul 11 17:04:20 GMT 2013
                       );
    
    
     Then you would want to setup your variables in the following ways (there are multiple ways you could do it, this is one):
    
     my $delim = '--------';
     my %expected_Vars = (
               channel          =>   qr/Channel:/,
               shelf_id         =>   qr/Shelf:/,
               module_type      =>   qr/Module type:/,
               firmware         =>   qr/Firmware rev:/,
               shelf_product_id =>   qr/Shelf product id:/,
               shelf_serial     =>   qr/Shelf Serial Number:/,
               module           =>   qr/^Module/,
               module_serial    =>   qr/(?>!Shelf )Serial Number:/,
               timestamp        =>   qr/Timestamp:/,
               log_ID           =>   qr/Log ID:/
       );
    
       Note that each of these regex follow the rules listed above! 1) all in the qr/regex/ format, 2) the patterns we desire lay directly to the right (whitespace irrelevant)
                                         3) none of the regex overlap (notice, module_serial does not overlap with module, as the (?!Shelf ) is not saved by the regex)
    
    
      /////////////////////////////
      IMPLEMENTATION
      ///////////////////////////////
    
      The way it will work: it will continue taking in the current line, and will terminate once the current line matches $delim.
      With each current line taken in, this line will be sent to a recursive method which attempts to match as many params as possible.
    
      Once this process has terminated, it will take the remaining keys in the hash, and for each one, search to see if that key is
      in the @allowed_To_Skip list.   If it is not, then we fail.  If it is, then we return (but will still print this information just in case)
    
      (note: the sub assumes the initial delimeter line is not being passed in, to keep in accordance with an older version of this library)
    
      //////////////
      questions about implementation can be directed to jolie@netapp.com
    

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=back

=cut

sub log_process_main_block {
    my (%opts) = validate( @_, { shelf_logs => { type => HASHREF }} );
    my $shelf_logs = $opts{shelf_logs};
    my %parseTable=();#will populate this with the parsed data and send back

    #///*** SET VARS HERE ***///#
    my $delim = '--------';
    my %expected_Vars = (
               channel          =>   qr/Channel:/,
               shelf_id         =>   qr/Shelf:/,
               module_type      =>   qr/Module type:/,
               firmware         =>   qr/Firmware rev:/,
               shelf_product_id =>   qr/Shelf product id:/,
               shelf_serial     =>   qr/Shelf Serial Number:/,
               module           =>   qr/^Module/,
               module_serial    =>   qr/(?<!Shelf )Serial Number:/,
               timestamp        =>   qr/Timestamp:/,
               log_ID           =>   qr/Log ID:/,
               request_type     =>   qr/Request Type:/
    );
    my @allowed_To_Skip = ('log_ID', 'request_type');
    #////**** END SET VARS  *****///#

    #//// go through the lines in the log and process them until you hit the delimeter ///////#
    my $line = &log_read_current_line( {shelf_logs => $shelf_logs } );
    while($line !~ /$delim/){
	chomp($line);#get rid of trailing new line chars
        &_recursive_reg( {line=>$line, vars=>\%expected_Vars, svalue=>\%parseTable} );#process the current line
        $line = &log_read_next_line( {shelf_logs => $shelf_logs } );
    }# end while, keep getting lines and processing them while they aren't the signal to end the process block
        
    #if the delimeter has been matched, then we are done processing the block.  But first make sure that
    #any keys that didn't end up getting matched we in the allowable skip list.  If they all are, return otherwise fail out
    foreach my $remaining (keys %{expected_Vars}){
        print "Warn: the following param was never matched: $remaining\n";
        if(!grep(/$remaining/, @allowed_To_Skip)){
            print "an error: we are not allowed to skip the parameter $remaining\n";
            exit 0;
        } #end if, one of the remaining keys didn't match
    } #end foreach, make sure any remaining keys we're allowed to skip

    print "will return the following parsed table:";
    &_print_hash_(\%parseTable);
    return \%parseTable;
} ## end sub log_process_main_block


=head2 _recursive_reg

   This is a recurvie method used to populate the hash table, passed in here as 'svalue' with particular values for variables specified in 'vars'.
   Refer to the documentation for log_process_block_main.

   EXAMPLE:

   Suppose we have the following inputs:

   $line = 'Channel: 0a Shelf: 1 Module type: IOM6 Firmware rev: 0162';
   my %vars = (
               channel          =>   qr/Channel:/,
               shelf_id         =>   qr/Shelf:/,
               log_ID           =>   qr/Log ID:/,
               firmware         =>   qr/Firmware rev:/,
               module_type      =>   qr/Module type:/,
               shelf_product_id =>   qr/Shelf product id:/,
               shelf_serial     =>   qr/Shelf Serial Number:/,
               module           =>   qr/^Module/,
               module_serial    =>   qr/(?<!Shelf )Serial Number:/,
               timestamp        =>   qr/Timestamp:/,
               request_type     =>   qr/Request Type:/
    );

    an initial call to _recursive_reg would result in the following recursion tree:

	1)   *CURRENT LINE* 'Channel: 0a Shelf: 1 Module type: IOM6 Firmware rev: 0162'
	     search through remaining k/v pairs... 'channel',
	     match will occur with 'Channel:', so channel k/v deleted from %vars
             recusive call -> 
	     1L)   *CURRENT LINE* '' 
             .      empty string so BASE CASE #1: 
	     .	    return ''
	     .	    //1L RECURSIVE CALL COMPLETE
	     recursive call ->
             1R)   *CURRENT LINE* '0a Shelf: 1 Module type: IOM6 Firmware rev: 0162'
	     .	    search through remaining k/v pairs... 'shelf_id',
             .      match will occur with 'Shelf:', so shelf_id k/v deleted from %vars
	     .	    recursive call ->
	     .	    1RL)  *CURRENT LINE* '0a'
	     .      .       search through remaining k/v pairs... 'log_ID','firmware','module_type','shelf_product_id','shelf_serial','module','module_serial','timestamp','request_type'
	     .      .       no matches found so BASE CASE #2:
	     .	    .       return '0a'
	     .      .	    //1RL RECURSIVE CALL COMPLETE//
             .      recursive call ->
	     .	    1RR)  *CURRENT LINE* '1  Module type: IOM6 Firmware rev: 0162'
	     .	    .      search through remaining k/v pairs... 'log_ID','firmware',
	     .	    .      match will occur with 'Firmware rev:', so firmware k/v deleted from %vars
	     .	    .      recursive call ->
	     .	    .      1RRL) *CURRENT LINE* '1 Module type: OM6'
	     .	    .      .       search through remaining k/v pairs... 'log_ID','module_type',
	     .      .      .       match will occur with 'Module type:', so log_ID k/v deleted from %vars
	     .      .	   .	   recursive call ->
	     .      .	   .	   1RRLL)  *CURRENT LINE* '1'
	     .      .	   .	   .	    search through remaining k/v pairs... 'log_ID','shelf_product_id','shelf_serial','module','module_serial','timestamp','request_type'
	     .	    .	   .	   .        no matches found so BASE CASE #2 (return current line):
	     .	    .	   .	   .        return '1'
	     .      .      .       .	    //1RRLL RECURSIVE CALL COMPLETE//
             .      .      .       recursive call -> 
	     . 	    .	   .       1RRLR)  *CURRENT LINE* 'OM6' 
	     .	    .	   .	   .        search through remaining k/v pairs... 'log_ID','shelf_product_id','shelf_serial','module','module_serial','timestamp','request_type'
	     .	    .	   .	   .        no matches found so BASE CASE #2 (return current line):
	     .	    .	   .	   .        return 'OM6'
             .      .      .       .        //1RRLR RECURSIVE CALL COMPLETE//  
	     .	    .	   .       set k/v pair {current match}{return of right recursive call}   : $svalue->{module_type} = 'OM6'   
	     .	    .	   .       non-base case exit: (return the return of left recursive call) : return '1'
             .      .      .       //1RRL RECURSIVE CALL COMPLETE//
             .      .      recursive call ->
	     .	    .      1RRR)  *CURRENT LINE* '0162'
	     . 	    .	   .      search through remaining k/v pairs... 'log_ID','shelf_product_id','shelf_serial','module','module_serial','timestamp','request_type'
	     . 	    .	   .      no matches found so BASE CASE #2 (return current line):
	     .	    .	   .      return '0162'
             .      .      .	  //1RRR RECURSIVE CALL COMPLETE//
	     .      .      //1RR RECURSIVE CALL COMPLETE//
	     .	    .      set k/v pair {current match}{return of right recursive call}    : $svalue->{firmware} = '0162'
	     .	    .      non-base case exit: (return the return of left recursive call)  : return '1'
	     .	    //1R RECURSIVE CALL COMPLETE
	     .	    set k/v pair {current match}{return of right recursive call}   : $svalue->{shelf_id} = '1'
	     .	    non-base case exit: (return the return of left recursive call) : return '0a'
	     //1 RECURSIVE CALL COMPLETE
	     set k/v pair {current match}{return of right recursive call}  : $svalue->{channel} = '0a'
	     non-base case exit: (return of return of left recursive call) : ''

            end result:
            %vars = (
               log_ID           =>   qr/Log ID:/,
               shelf_product_id =>   qr/Shelf product id:/,
               shelf_serial     =>   qr/Shelf Serial Number:/,
               module           =>   qr/^Module/,
               module_serial    =>   qr/(?<!Shelf )Serial Number:/,
               timestamp        =>   qr/Timestamp:/,
               request_type     =>   qr/Request Type:/
    	    );

	    %svalue = (
		module_type	=>   '0M6',
		firmware	=>   '0162',
		shelf_id	=>   '1',
		channel		=>   '0a'
	   );
  
		       


=over

=back

=cut

sub _recursive_reg{
    my (%opts) = validate(
        @_,
        {    line   => {type => SCALAR,  optional => 0 },
             vars   => {type => HASHREF, optional => 0 },
             svalue => {type => HASHREF, optional => 0 },
        }
    );#validate incoming params

    my $line     = $opts{line};  #this will be the string we are processing currently
    my $vars     = $opts{vars};  #this will be a dynamically changing hash with the remaining matches
    my $svalue   = $opts{svalue};#this will be dynamically changing as it is populated with more calls
    my $matchBoo = 0;            #will flag to 1 if any matches are found during execution

    #**BASE CASE 1: incoming string is empty.  Nothing left to process so return from sub.
    return 0 if(!$line);

    #**Search through remaining potential matches to try and see if any match in $line
    foreach my $var (keys %{$vars}){
        if($line =~ /(.*)$vars->{$var}(.*)/){#regex n00bs: match on $vars->{$var}, save everything to the left and right
            #we found a match : our line has an occurence of $vars->{$var}.  Delete this key now so we don't end up searching for it in other recursive calls
            delete $vars->{$var};

            #now, store what is to the left and the right (the parents in the regex let us do this)
            my $checkLeft  = $1;
            my $checkRight = $2;

            #trim whitespaces by calling our sub sub routine
            $checkLeft =  &_trim_($checkLeft);
            $checkRight = &_trim_($checkRight);

            #now make calls on these variables, the purpose being, our full line $line might have multiple matches on it, some to the right and some to the left
            my $resultL = &_recursive_reg( {line=>$checkLeft,  vars=>$vars, svalue=>$svalue} );
            my $resultR = &_recursive_reg( {line=>$checkRight, vars=>$vars, svalue=>$svalue} );

            #$resultR should be the desired value, so store it in $svalue
            $svalue->{$var}=$resultR;

            #return the result of the left, this should always be either an empty string (because there was nothing to process) or this will be the actual data the function that called this one needs
            return $resultL;
        }# end if for finding a match
    }#end foreach to search through all the keys

    #**BASE CASE 2: there were no matches.  Send back the string we came in with.  This will be getting sent up the stack and should end up being set as one of the values
    return $line if(!$matchBoo);

    #an internal subroutine: trim (perl doesn't have a built in trim function.  So all we do here is trim whitespace of either end of the string)
    sub _trim_{
        my $string = shift;
        $string =~ s/^\s+//;#regex n00bs: match any number of white space characters at the beginning of line and replace with empty string
        $string =~ s/\s$//;#regex n00bs: match any number of white space characters at the end of line and replace with empty string
        return $string;
    } #end internal sub _trim_

    #an internal subroutine to print out a hash (for debugging purposes)
    sub _print_hash_{
        my $hashRef = shift;
        foreach my $key (keys %{$hashRef}){
            print "[key]->[value]: [$key]->[$hashRef->{$key}]\n";
        }#end foreach
    } #end internal sub _print_hash_
} #end subroutine recursive_reg


=head2 log_generate_records

     my $array_ref = log_generate_records({shelf_logs => $shelf_logs});

This routine generates an array of hash references to the log instances in the log file between the start and end markers.  If the start and end markers are not defined the complete log file would be parsed.  This routine retrieves the information related to adapter, module, shelf id, serial number, module serial number, timestamp, the starting and the ending line numbers of each log instance and creates a hash reference.  All the log instance hash references are recorded into an array maintaining the order in which the logs are available in the log file and a reference to this array is returned to the calling program. On error this routine returns an undef.

Currenlty this subroutine supports only SAS logs.

     Example:
       [
          {
            "time" => "Mon Dec  1 23:35:24 GMT 2008",
            "end_block" => 2722,
            "start_block" => 113,
            "shelf_id" => "1",
            "shelf_serial" => "ABCD",
            "channel" => "3c",
            "firmware" => "a6",
            "shelf_product_id" => "SAS-Shelf24",
            "module_type" => "SAS",
            "module_serial" => "ims942140000034",
            "module" => "B"
          },
       ]

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=back

=cut

sub log_generate_records {
    my (%opts) = validate( @_, { shelf_logs => { type => HASHREF }} );
    my $shelf_logs = $opts{shelf_logs};

    my $rvalue = ();
    my $line;
    my $block = 0;
    print "Generating records...\n\n";

    log_read_first_line( { shelf_logs => $shelf_logs } );
    $shelf_logs->{current_line}--;
    my $inside_block_flag = 0;

    my $counter=0;
    while ( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) {
	    $counter++;
        if ( $line
            =~ /--------------------------------------------------------------/
            )
        {
            if ( $inside_block_flag == 0 ) {
                next;
            } elsif ( $inside_block_flag == 1 ) {
                $$rvalue[$block]->{end_block}
                    = $shelf_logs->{current_line} - 1;
                $block++;
                $inside_block_flag = 0;
            } ## end elsif ( $inside_block_flag...
        } ## end if ( $line =~ ...

        if ( ( $line =~ /Channel: / ) && ( $line =~ /Shelf: / ) ) {
            if ( $inside_block_flag == 1 ) {
                logresult( "FAIL", "Error in log file" );
                return undef;
            }

     #			if($block != 0)
     #			{
     #				my $old_block = $block - 1;
     #				$$rvalue[$old_block]->{end_block} = $shelf_logs->{current_line} - 1;
     #			}

     logcomment("calling log_process_main_block");
            my $hash_ref
                = log_process_main_block( { shelf_logs => $shelf_logs } );
            $$rvalue[$block]                = $hash_ref;
	logcomment("back from log_process_main_block with");
	print Dumper $hash_ref;
            $$rvalue[$block]->{start_block} = $shelf_logs->{current_line}++;
            $inside_block_flag              = 1;
        } ## end if ( ( $line =~ /Channel: /...
    } ## end while ( $line = log_read_next_line...

    if ( $inside_block_flag == 1 ) {
        $$rvalue[$block]->{end_block} = $shelf_logs->{current_line} - 1;
    }
    return $rvalue;

} ## end sub log_generate_records

=head2 log_search_serials_csms

     my $hash_ref = log_search_serials_csms({log_records => $log_records, serials => $serial_numbers});

This routine searches for the module serial numbers represented by the array reference $serial_numbers in the parsed logs log_records and returns a hash reference providing the match count and the matched record ids to the calling program.

     Example:
       {
          "count" => 8,
          "records" => [
                         1,
                         2,
                         3,
                         4,
                         5,
                         6,
                         7,
                         8
                       ]
        }

=over

=item log_records

(Required) : The reference to the array of log records generated by log_generate_records for the log file.

=item serials

(Required) : The reference to an array of the serial numbers to be searched for in the log records.

=back

=cut

sub log_search_serials_csms {
    my (%opts) = validate(
        @_,
        {   log_records => { type => ARRAYREF },
            serials     => { type => ARRAYREF, optional => 1 },
            channels    => { type => ARRAYREF, optional => 1 },
            modules     => { type => ARRAYREF, optional => 1 },
            shelf_ids   => { type => ARRAYREF, optional => 1 },
            shelf_uids  => { type => ARRAYREF, optional => 1 },
        }
    );
    my $log_records = $opts{log_records};
    my $matched_records;
    my $match_count = 0;

    if ( $#$log_records < 0 ) {
        my $rvalue = {};
        $rvalue->{count} = 0;
        return $rvalue;
    }

    if ( defined $opts{serials} ) {
        my $serials = $opts{serials};

        logcomment("Searching for Serial Numbers: @$serials");
        for ( my $i = 0; $i <= $#$log_records; $i++ ) {
            foreach my $serial (@$serials) {
                if ( $$log_records[$i]->{module_serial} eq $serial ) {
                    $match_count++;
                    push @$matched_records, $i;
                    last;
                }
            } ## end foreach my $serial (@$serials)
        } ## end for ( my $i = 0; $i <= ...
        $match_count = $#$matched_records + 1;
    } else {
        for ( my $i = 0; $i <= $#$log_records; $i++ ) {
            push @$matched_records, $i;
        }
        $match_count = $#$matched_records + 1;
    } ## end else [ if ( defined $opts{serials...

    if ( ( defined $matched_records ) && ( defined $opts{channels} ) ) {
        my $channels = $opts{channels};
        logcomment("Searching for Channels: @$channels");
        $matched_records = log_search_element(
            {   log_records => $log_records,
                records     => $matched_records,
                parameter   => "channel",
                elements    => $channels
            }
        );
    } ## end if ( ( defined $matched_records...

    if ( ( defined $matched_records ) && ( defined $opts{shelf_ids} ) ) {
        my $shelf_ids = $opts{shelf_ids};
        logcomment("Searching for Shelf IDs: @$shelf_ids");
        $matched_records = log_search_element(
            {   log_records => $log_records,
                records     => $matched_records,
                parameter   => "shelf_id",
                elements    => $shelf_ids
            }
        );
    } ## end if ( ( defined $matched_records...

    if ( ( defined $matched_records ) && ( defined $opts{modules} ) ) {
        my $modules = $opts{modules};
        logcomment("Searching for Modules: @$modules");
        $matched_records = log_search_element(
            {   log_records => $log_records,
                records     => $matched_records,
                parameter   => "module",
                elements    => $modules
            }
        );
    } ## end if ( ( defined $matched_records...

    if ( ( defined $matched_records ) && ( defined $opts{shelf_uids} ) ) {
        my $shelf_uids = $opts{shelf_uids};
        logcomment("Searching for Shelf UIDs: @$shelf_uids");
        $matched_records = log_search_element(
            {   log_records => $log_records,
                records     => $matched_records,
                parameter   => "shelf_uid",
                elements    => $shelf_uids
            }
        );
    } ## end if ( ( defined $matched_records...

    my $rvalue;
    if ( defined $matched_records ) {
        $match_count = $#$matched_records + 1;
        $rvalue      = {
            count   => $match_count,
            records => $matched_records
        };
    } else {
        $rvalue = { count => 0, };
    }
    return $rvalue;

} ## end sub log_search_serials_csms

=head2 log_search_strings

     my $hash_ref = log_search_strings({shelf_logs => $shelf_logs, log_records => $log_records, strings => $strings, selected_records => $selected_records});

This routine searches for the strings in all the log records and returns a hash reference with the match count and an array reference to all the instances of the log records that match the strings.  If the parameter selected_records is defined, the search is restricted to these records only instead of the complete log file.

     Example:
       {
          "count" => 8,
          "records" => [
                         1,
                         2,
                         3,
                         4,
                         5,
                         6,
                         7,
                         8
                       ]
        }

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=item log_records

(Required) : The reference to the array of log records generated by log_generate_records for the log file.

=item strings

(Required) : The reference to an array of the strings that have to be searched for in the log records.

=item selected_records

(Optional) : The reference to any array of the instances of log records in which the strings have to be searched.

=back

=cut

sub log_search_strings {
    my (%opts) = validate(
        @_,
        {   shelf_logs       => { type => HASHREF },
            log_records      => { type => ARRAYREF },
            selected_records => { type => ARRAYREF, optional => 1 },
            strings          => { type => ARRAYREF },
            string_stats => { type => SCALAR, optional => 1, default => 1 }
        }
    );

    my $shelf_logs  = $opts{shelf_logs};
    my $log_records = $opts{log_records};
    my $strings     = $opts{strings};

    my @matched_records;
    my $match_count = 0;
    my $line;
    my $records;

    if ( $#$log_records < 0 ) {
        my $rvalue = {};
        $rvalue->{count} = 0;
        return $rvalue;
    }

    if ( defined $opts{selected_records} ) {
        $records = $opts{selected_records};
    } else {
        $records = [ 0 .. $#$log_records ];
    }

    my $string_stats = {};

    if ( $opts{string_stats} == 1 ) {
        my @table = ();

        foreach my $record (@$records) {
            my $match_flag = 0;
            my $current_element
                = "$$log_records[$record]->{channel}" . '.'
                . "$$log_records[$record]->{shelf_id}" . '.'
                . "$$log_records[$record]->{module}";
            foreach my $element (@table) {
                if ( $current_element eq $element ) {
                    $match_flag = 1;
                }
            }
            if ( $match_flag == 0 ) {
                push @table, $current_element;
            }
        } ## end foreach my $record (@$records)

        @table = sort @table;

        $string_stats = {};
        foreach my $element (@table) {
            foreach my $string (@$strings) {
                $string_stats->{$element}->{$string} = 0;
            }
        }
    } ## end if ( $opts{string_stats...

    $" = ',';
    logcomment("Searching for Strings: @$strings");
    $" = ' ';

    foreach my $record ( sort @$records ) {
        my $match_flag = 0;
        for (
            my $lineno = $$log_records[$record]->{start_block};
            $lineno <= $$log_records[$record]->{end_block};
            $lineno++
            )
        {
            my $current_element
                = "$$log_records[$record]->{channel}" . '.'
                . "$$log_records[$record]->{shelf_id}" . '.'
                . "$$log_records[$record]->{module}";
            if ($line = log_read_line(
                    { shelf_logs => $shelf_logs, line => $lineno }
                )
                )
            {
                foreach my $string (@$strings) {
                    if ( $line =~ /$string/ ) {
                        $match_count++;
                        if ( $opts{string_stats} == 1 ) {
                            $string_stats->{$current_element}->{$string}++;
                        }
                        if ( $match_flag == 0 ) {
                            $match_flag = 1;
                            push @matched_records, $record;
                        }
                    } ## end if ( $line =~ /$string/)
                } ## end foreach my $string (@$strings)
            } ## end if ( $line = log_read_line...
        } ## end for ( my $lineno = $$log_records...
    } ## end foreach my $record ( sort @$records)

    my $rvalue;

    if ( $match_count > 0 ) {
        $rvalue = {
            count   => $match_count,
            records => \@matched_records,
        };
    } else {
        $rvalue = { count => 0, };
    }
    if ( $opts{string_stats} == 1 ) {
        $rvalue->{string_stats} = $string_stats;
    }
    return $rvalue;
} ## end sub log_search_strings

=head2 log_search_serial_strings

     my $hash_ref = log_search_serial_strings({shelf_logs => $shelf_logs, log_records => $log_records, strings => $strings, serials => $serials});

This routine retieves all log instances for the serial numbers given by the array reference $serials and searches for all the instances of strings given by the array reference $strings in them and returns the matched count and the record instances of the logs that match the strings as a hash reference.

     Example:
       {
          "count" => 5,
          "records" => [
                         1,
                         3,
                         4,
                         6,
                         8
                       ]
        }

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=item log_records

(Required) : The reference to the array of log records generated by log_generate_records for the log file.

=item serials

(Required) : The reference to an array of the module serial numbers that have to be searched for in the log records.

=item strings

(Required) : The reference to an array of the strings that have to be searched for in the log records.

=back

=cut

sub log_search_serial_strings {
    my (%opts) = validate(
        @_,
        {   shelf_logs   => { type => HASHREF },
            log_records  => { type => ARRAYREF },
            strings      => { type => ARRAYREF },
            serials      => { type => ARRAYREF, optional => 1 },
            channels     => { type => ARRAYREF, optional => 1 },
            modules      => { type => ARRAYREF, optional => 1 },
            shelf_ids    => { type => ARRAYREF, optional => 1 },
            shelf_uids   => { type => ARRAYREF, optional => 1 },
            string_stats => { type => SCALAR, optional => 1 }
        }
    );

    my $parameters = {};
    $parameters->{log_records} = $opts{log_records};
    if ( defined $opts{serials} ) {
        $parameters->{serials} = $opts{serials};
    }
    if ( defined $opts{channels} ) {
        $parameters->{channels} = $opts{channels};
    }
    if ( defined $opts{modules} ) {
        $parameters->{modules} = $opts{modules};
    }
    if ( defined $opts{shelf_ids} ) {
        $parameters->{shelf_ids} = $opts{shelf_ids};
    }
    if ( defined $opts{shelf_uids} ) {
        $parameters->{shelf_uids} = $opts{shelf_uids};
    }

    my $return_value = log_search_serials_csms($parameters);

    my $records = $return_value->{records};

    $parameters                     = {};
    $parameters->{shelf_logs}       = $opts{shelf_logs};
    $parameters->{log_records}      = $opts{log_records};
    $parameters->{strings}          = $opts{strings};
    $parameters->{selected_records} = $records;
    if ( defined $opts{string_stats} ) {
        $parameters->{string_stats} = $opts{string_stats};
    }
    $return_value = log_search_strings($parameters);

    return $return_value;
} ## end sub log_search_serial_strings

=head2 log_print_line

     log_print_line({dump_file => $dump_file, line => $line, log_also => $log_also});

This routine prints the line $line to the file represented by the file handle $dump_file.  If the file handle is not open, the line is printed on to the <stdout>. If the parameter log_also is defined and set to 1, then the line is attempted to print both in the dump_file as well as to the logs that is the <stdout>.

=over

=item line

(Required) : The line to be printed.

=item dump_file

(Optional) : The typeglob reference for the file handle for dumping the logs from the log file.

=item log_also

(Optional) : The parameter when set prints the line both to the dump file and <stdout>.  By default this is set to 0.

=back

=cut

sub log_print_line {
    my (%opts) = validate(
        @_,
        {   dump_file => { type => GLOBREF, optional => 1 },
            line      => { type => SCALAR },
            log_also => { type => SCALAR, optional => 1, default => 1 }
        }
    );
    my $fh;
    if ( defined $opts{dump_file} ) {
        $fh = $opts{dump_file};
    }
    if ( $opts{log_also} == 1 ) {
        print "$opts{line}";
    }
    if ( defined $opts{dump_file} ) {
        if ( tell($fh) != -1 ) {
            print $fh "$opts{line}";
        } elsif ( $opts{log_also} == 0 ) {
            print "$opts{line}";
        }
    } ## end if ( defined $opts{dump_file...
} ## end sub log_print_line

=head2 log_print_records

     my $result = log_print_records({shelf_logs => $shelf_logs, log_records => $log_records, records => $myrecords, file => $dump_file, log_also => $log_also, stats => $false});

This routine prints a block of data onto a dump file or to the <stdout>.  The records parameter accepts a reference to an array of record ids, these would be searched from the log_records and printed.  The file parameter is optional and if specified would print the log records onto the dump file.  If an error occurs when opening the file, the contents are dumped into the log file of the test being run.  The log_also parameter can be specified when the logs are to be printed both to the dump file and logs.  The parameter header_only can be used to print only the header information.  The stats parameter is used to print the statistical data from the log files.

     The subroutine returns 1 on success and 0 on failure.

=over

=item shelf_logs

(Required) : The reference to the data structure used for parsing the log file.

=item log_records

(Required) : The reference to the array of log records generated by log_generate_records for the log file.

=item records

(Required) : The reference to the array of record ids that have to be printed.

=item dump_file

(Optional) : The typeglob reference for the file handle for dumping the logs from the log file.

=item log_also

(Optional) : The parameter when set prints the line both to the dump file and <stdout>.  By default this is set to 0.

=item header_only

(Optional) : This parameter when specified prints only the log header information

=item stats

(Optional) : This parameter is used to print the statistical part of logs.  The default value of this variable is 0.

=back

=cut

sub log_print_records {
    my (%opts) = validate(
        @_,
        {   shelf_logs  => { type => HASHREF },
            log_records => { type => ARRAYREF },
            records     => { type => ARRAYREF },
            dump_file   => { type => SCALAR, optional => 1 },
            header_only => { type => SCALAR, optional => 1, default => 0 },
            log_also    => { type => SCALAR, optional => 1, default => 1 },
            stats       => { type => SCALAR, optional => 1, default => 0 },
            string_stats => { type => HASHREF, optional => 1 }
        }
    );

    print "log_print_records\n";
    my $shelf_logs  = $opts{shelf_logs};
    my $log_records = $opts{log_records};
    my $records     = $opts{records};
    my $log_also    = $opts{log_also};
    my $string_stats;

    if ( defined $opts{string_stats} ) {
        $string_stats = $opts{string_stats};
    }

    my $dump_file;
    my $FH;
    if ( defined $opts{dump_file} ) {
        $dump_file = $opts{dump_file};
        open $FH, ">> $dump_file"
            or logcomment(
            "Unable to open file due to: $!\nThe records would be printed to the test logs"
            );
    } ## end if ( defined $opts{dump_file...

    my $header_only_flag = 0;

    if ( defined $opts{header_only} ) {
        if ( $opts{header_only} == 1 ) {
            $header_only_flag = 1;
        }
    }

    my $return_value = 1;

    if ( ( $opts{log_also} == 0 ) && ( !( defined $FH ) ) ) {
        logcomment(
            "The data would not be printed to null as LOG_ALSO and DUMP_FILE are not defined"
        );
    }

    for ( my $num = 0; $num <= $#$records; $num++ ) {
        if ( defined $FH ) {
            if ( tell($FH) != -1 ) {
                logcomment(
                    "Printing Record #$$records[$num] to file $dump_file");
            } else {
                logcomment("Printing Record #$$records[$num]");
            }
        } else {
            logcomment("Printing Record #$$records[$num]");
        }

        my $start_line;
        if ( defined $$log_records[ $$records[$num] ]->{start_block} ) {
            $start_line = $$log_records[ $$records[$num] ]->{start_block};
        } else {
            logresult( "FAIL",
                "Starting block line for $$log_records[$$records[$num]] is not defined"
            );
	    print "log_print_records return 0";
            return 0;
        } ## end else [ if ( defined $$log_records...
        my $end_line;
        if ( defined $$log_records[ $$records[$num] ]->{end_block} ) {
            $end_line = $$log_records[ $$records[$num] ]->{end_block};
        } else {
            logresult( "FAIL",
                "Ending block line for $$log_records[$$records[$num]] is not defined"
            );
	    print "log_print_records return 0";
            return 0;
        } ## end else [ if ( defined $$log_records...

        if ( defined $FH ) {
            log_print_line(
                {   dump_file => $FH,
                    log_also  => $log_also,
                    line =>
                        "--------------------------------------------------------------\n"
                }
            );
        } else {
            log_print_line(
                {   log_also => $log_also,
                    line =>
                        "--------------------------------------------------------------\n"
                }
            );
        } ## end else [ if ( defined $FH )

        if ( $header_only_flag == 1 ) {
            foreach
                my $key ( sort keys %{ $$log_records[ $$records[$num] ] } )
            {
                if ( defined $FH ) {
                    log_print_line(
                        {   dump_file => $FH,
                            log_also  => $log_also,
                            line =>
                                "$key\t$$log_records[$$records[$num]]->{$key}\n"
                        }
                    );
                } else {
                    log_print_line(
                        {   log_also => $log_also,
                            line =>
                                "$key\t$$log_records[$$records[$num]]->{$key}\n"
                        }
                    );
                } ## end else [ if ( defined $FH )
            } ## end foreach my $key ( sort keys...
            if ( defined $FH ) {
                log_print_line(
                    {   dump_file => $FH,
                        log_also  => $log_also,
                        line =>
                            "--------------------------------------------------------------\n"
                    }
                );
            } else {
                log_print_line(
                    {   log_also => $log_also,
                        line =>
                            "--------------------------------------------------------------\n"
                    }
                );
            } ## end else [ if ( defined $FH )
        } else {
            my $header_line_start;
            if ( $opts{stats} == 0 ) {
                $header_line_start = $start_line - 5;
            } else {
                $header_line_start = $start_line - 10;
            }

            if ( defined $$log_records[ $$records[$num] ]->{request_type} ) {
                my $header_line_start = $start_line - 10;
            }

            my $line;
            for (
                my $line_num = $header_line_start;
                $line_num < $start_line;
                $line_num++
                )
            {
		    logcomment("CALLING log_read_line FROM IF STATEMENT 2");
                if ($line = log_read_line(
                        { shelf_logs => $shelf_logs, line => $line_num }
                    )
                    )
                {
                    if ( defined $FH ) {
                        log_print_line(
                            {   dump_file => $FH,
                                log_also  => $log_also,
                                line      => $line
                            }
                        );
                    } else {
                        log_print_line(
                            { log_also => $log_also, line => $line } );
                    }
                } else {
                    logresult( "FAIL",
                        "The script is trying to access a line that doesn't exist: $line_num"
                    );
		    print "log_print_records return 0";
                    return 0;
                } ## end else [ if ( $line = log_read_line...
            } ## end for ( my $line_num = $header_line_start...

            if ( defined $opts{string_stats} ) {
                if ( defined $FH ) {
                    log_print_line(
                        {   dump_file => $FH,
                            log_also  => $log_also,
                            line =>
                                "--------------------------------------------------------------\n"
                        }
                    );
                } else {
                    log_print_line(
                        {   log_also => $log_also,
                            line =>
                                "--------------------------------------------------------------\n"
                        }
                    );
                } ## end else [ if ( defined $FH )
                $line = "String search results:\n\n";
                if ( defined $FH ) {
                    log_print_line(
                        {   dump_file => $FH,
                            log_also  => $log_also,
                            line      => $line
                        }
                    );
                } else {
                    log_print_line(
                        { log_also => $log_also, line => $line } );
                }
                my $current_element
                    = "$$log_records[$$records[$num]]->{channel}" . '.'
                    . "$$log_records[$$records[$num]]->{shelf_id}" . '.'
                    . "$$log_records[$$records[$num]]->{module}";

                foreach
                    my $string ( keys %{ $string_stats->{$current_element} } )
                {
                    if ( $string_stats->{$current_element}->{$string} != 0 ) {
                        my $line
                            = "'$string': $string_stats->{$current_element}->{$string}\n";

                        if ( defined $FH ) {
                            log_print_line(
                                {   dump_file => $FH,
                                    log_also  => $log_also,
                                    line      => $line
                                }
                            );
                        } else {
                            log_print_line(
                                { log_also => $log_also, line => $line } );
                        }
                    } ## end if ( $string_stats->{$current_element...
                } ## end foreach my $string ( keys %...
            } ## end if ( defined $opts{string_stats...

            for (
                my $line_num = $start_line;
                $line_num <= $end_line;
                $line_num++
                )
            {
		    logcomment("CALLING log_read_line FROM IF STATEMENT 3");
                if ($line = log_read_line(
                        { shelf_logs => $shelf_logs, line => $line_num }
                    )
                    )
                {
                    if ( defined $FH ) {
                        log_print_line(
                            {   dump_file => $FH,
                                log_also  => $log_also,
                                line      => $line
                            }
                        );
                    } else {
                        log_print_line(
                            { log_also => $log_also, line => $line } );
                    }
                } else {
                    logresult( "FAIL",
                        "The script is trying to access a line that doesn't exist: $line_num"
                    );
		    print "log_print_records return 0\n";
                    return 0;
                } ## end else [ if ( $line = log_read_line...
            } ## end for ( my $line_num = $start_line...
            if ( defined $FH ) {
                log_print_line(
                    {   dump_file => $FH,
                        log_also  => $log_also,
                        line =>
                            "--------------------------------------------------------------\n"
                    }
                );
            } else {
                log_print_line(
                    {   log_also => $log_also,
                        line =>
                            "--------------------------------------------------------------\n"
                    }
                );
            } ## end else [ if ( defined $FH )
        } ## end else [ if ( $header_only_flag...
    } ## end for ( my $num = 0; $num...

    if ( defined $opts{string_stats} ) {
        logcomment("Cumulative String Statistics");

        my $cumulative_string_stats = {};
        my $first_flag              = 0;

        foreach my $current_element ( keys %$string_stats ) {
            foreach my $string ( keys %{ $string_stats->{$current_element} } )
            {
                if ( $first_flag == 0 ) {
                    $cumulative_string_stats->{$string}
                        = $string_stats->{$current_element}->{$string};
                } else {
                    $cumulative_string_stats->{$string}
                        = $cumulative_string_stats->{$string}
                        + $string_stats->{$current_element}->{$string};
                }
            } ## end foreach my $string ( keys %...
            $first_flag = 1;
        } ## end foreach my $current_element...

        foreach my $string ( keys %$cumulative_string_stats ) {
            my $line
                = "\t\t\t     '$string': $cumulative_string_stats->{$string}\n";
            if ( defined $FH ) {
                log_print_line(
                    {   dump_file => $FH,
                        log_also  => $log_also,
                        line      => $line
                    }
                );
            } else {
                log_print_line( { log_also => $log_also, line => $line } );
            }
        } ## end foreach my $string ( keys %$cumulative_string_stats)
    } ## end if ( defined $opts{string_stats...

    print "log_print_records return 1\n";
    return 1;
} ## end sub log_print_records

=head2 log_search

     my $result = log_search({ log_file => $log_file, log_start_marker => $log_start_marker, log_end_marker => $log_end_marker, dump_file => $dump_file, serials => serials, include_strings => $include_strings, exclude_strings => $exclude_strings, header_only => $header_only, log_also => $log_also, filer => $filer, filer_conn => $filer_connection_type, shelf_type => $shelf_type, sudo_path = $sudo_path});

This routine is a wrapper for searching and printing specific logs from the log file.  If the log file is specified, the contents of the log file $log_file are parsed.  If the filer with filer connection and shelf type are specified, the subroutine mounts the filer root volume and extracts the correct log file name and parses it.  The parameter sudo path is specified if the sudo command is not in default location /usr/local/bin.  The parameters log_start_marker and log_end_marker specify the starting and ending line numbers in the log file for parsing.  The logs are categorised into log records and stored as an array of hash references linked with the log file based on the line numbers.  The parameters serials can be specified to search for a particular module serial number records in the log file.  A default set of strings are always searched in the log file for all the modules and the subroutine returns 0 on finding these strings.  If the strings set needs to be modified, the parameters include_strings and exclude strings can be used to include or exclude a set of strings respectively.  These parameters need to be reference to an array of strings.  The parameter header_only when set to 1 prints only the header part of the logs that match the criteria.  The parameter log_also when set to 1 prints the logs both in the dump file as well as logs.

The subroutine returns

	* the count of matched module serial numbers or strings on success
	* "The log file is empty or not generated or there might be an error in the log file" when the start and end markers are same
	* "The log file might have been renewed or there might be an error in the log file" when the start marker is greater than end marker.
	* -1 on failure or error

Currenlty this subroutine supports only SAS logs.

=over

=item log_file

(Optional) : The log file with absolute path.

=item log_start_marker

(Optional) : This is the line number in the log file from which the script should parse data.  Default is start of log file.

=item log_end_marker

(Optional) : This is the line number in the log file until which the script should parse data.  Default is end of log file.

=item dump_file

(Optional) : The file to which the selected logs should be dumped.

=item serials

(Optional) : The module serial numbers that have to be selected for processing.  If not specified all the serial numbers are processed.

=item include_strings

(Optional) : The strings that have to be added to the default set of strings for searching.  Default strings that would be searched are "PMC Firmware assert", "watchdog" and "crash".

=item exclude_strings

(Optional) : The strings that have to be ignored even if they exist when searching the logs.

=item header_only

(Optional) : This parameter when set to 1 would print only the header information in the dump file or the log file after processing.  Default value of this parameter is 0.

=item log_also

(Optional) : This parameter when set to 1 would print the parsed records in the dump file as well as in the logs. Default is set to 0.

=item filer

(Optional) : The filer from which the log file has to be parsed.

=item filer_conn

(Optional) : This defines the way to connect to the filer.  For default value refer to section log_mount.

=item shelf_type

(Optional) : This defines the type of shelf.  For default value refer to section log_mount.

=item sudo_path

(Optional) : Defines the path of sudo on the client. For default value refer to section log_mount.


=back

=cut

sub log_search {
    my (%opts) = validate(
        @_,
        {   filer            => { type => SCALAR,   optional => 1 },
            filer_conn       => { type => SCALAR,   optional => 1 },
            shelf_type       => { type => SCALAR,   optional => 1 },
            sudo_path        => { type => SCALAR,   optional => 1 },
            channels         => { type => ARRAYREF, optional => 1 },
            modules          => { type => ARRAYREF, optional => 1 },
            shelf_ids        => { type => ARRAYREF, optional => 1 },
            shelf_uids       => { type => ARRAYREF, optional => 1 },
            log_files        => { type => ARRAYREF, optional => 1 },
            log_start_marker => { type => HASHREF,  optional => 1 },
            log_end_marker   => { type => HASHREF,  optional => 1 },
            dump_file        => { type => SCALAR,   optional => 1 },
            serials          => { type => ARRAYREF, optional => 1 },
            include_strings  => { type => ARRAYREF, optional => 1 },
            default_strings =>
                { type => SCALAR, optional => 1, default => 1 },
            exclude_strings => { type => ARRAYREF, optional => 1 },
            header_only     => { type => SCALAR,   optional => 1 },
            log_also        => { type => SCALAR,   optional => 1 },
            string_stats => { type => SCALAR, optional => 1, default => 1 },
	    client_hostname => {type => SCALAR, optional => 1 }
        }
    );

    my $shelf_logs = {};
    if ( defined $opts{log_start_marker} ) {
        $shelf_logs->{log_start_marker} = $opts{log_start_marker};
        logcomment(">>> Log start marker has been defined and passed in.  Setting: shelf_logs->{log_start_marker}->{line_number} as  $shelf_logs->{log_start_marker}->{line_number} (ShelfLog::log_search)") if($talk);
    } ## end if ( defined $opts{log_start_marker...
    if ( defined $opts{log_end_marker} ) {
        $shelf_logs->{log_end_marker} = $opts{log_end_marker};
        logcomment(">>> Log end marker has been defined and passed in.  Setting: shelf_logs->{log_end_marker}->{line_number} as $shelf_logs->{log_end_marker}->{line_numberi} (ShelfLog::log_search)") if($talk);
    } ## end if ( defined $opts{log_end_marker...

    my $log_files;
    logcomment(">>> attempt to get log_files (ShelfLog::log_search)") if($talk);
    if ( defined $opts{log_files} ) {
        $log_files = $opts{log_files};
	logcomment(">>> log_files is defined as $log_files and passed in (ShelfLog::log_search)") if($talk);
    } elsif ( defined $opts{filer} ) {
        my $log_mount_parameters;
        $log_mount_parameters->{filer} = $opts{filer};
        if ( defined $opts{filer_conn} ) {
            $log_mount_parameters->{filer_conn} = $opts{filer_conn};
        }
        if ( defined $opts{shelf_type} ) {
            $log_mount_parameters->{shelf_type} = $opts{shelf_type};
        }
        if ( defined $opts{sudo_path} ) {
            $log_mount_parameters->{sudo_path} = $opts{sudo_path};
        }
	if ( defined $opts{client_hostname} ) {
	    $log_mount_parameters->{client_hostname} = $opts{client_hostname};
	}

	logcomment(">>> opts filer was defined and log files was not, so we'll be mounting a client now.  We will call log_mount with the following params: (ShelfLog::log_search)") if($talk);
	print Dumper $log_mount_parameters;
        $log_files = log_mount($log_mount_parameters);
        $log_files = [$log_files];
	logcomment(">>> got log files as $log_files from the log mount call.  (ShelfLog::log_search)") if($talk);
    } else {
        logresult( "FAIL",
            "Please provide atleast a log file or a filer name" );
        return -1;
    }
    my $check;
    my $log_exist_parameters;
    $log_exist_parameters->{filer} = $opts{filer} if ( defined $opts{filer} );
    $log_exist_parameters->{log_files} = $log_files;
    $log_exist_parameters->{shelf_logs} = $shelf_logs;
    $log_exist_parameters->{client_hostname} = $opts{client_hostname} if ( defined $opts{client_hostname} );
    logcomment("sending to log_exist::") if($talk);
    #print Dumper $log_exist_parameters;
    $check = log_exist( $log_exist_parameters );
    logcomment("out of log exist") if($talk); 
    # Verify that if only 1 log file is prestent, if start mark and end of
    # file are the same then there has been no update.
    
    if ($#{ $shelf_logs->{log_files} } == 0) { 
        # There is only one log file
        if ( $shelf_logs->{log_start_marker}->{line_number}
            == $shelf_logs->{log_files}->[0]->{'end_line'} ) { 
            # The log file is the same size as when log_start_marker was
            # created
            logcomment(
                "There was no update in the log file, as the start and end markers are same"
            );
            return -2;
        }
    } ## end if ( $#{ $shelf_logs->...
    logcomment("calling log generate records") if($talk);
    my $myrecords = log_generate_records( { shelf_logs => $shelf_logs } );
    logcomment("out of log generated records") if($talk);
    my $selected_records;
    logcomment("got back from log generated records") if($talk);
    #print Dumper $selected_records;

    if ( $#$myrecords == -1 ) {
        logresult( "FAIL", "No records could be parsed" );
        return -1;
    }

    my $parameters = {};
    $parameters->{log_records} = $myrecords;
    if ( defined $opts{serials} ) {
        $parameters->{serials} = $opts{serials};
    }
    if ( defined $opts{channels} ) {
        $parameters->{channels} = $opts{channels};
    }
    if ( defined $opts{shelf_ids} ) {
        $parameters->{shelf_ids} = $opts{shelf_ids};
    }
    if ( defined $opts{modules} ) {
        $parameters->{modules} = $opts{modules};
    }
    if ( defined $opts{shelf_uids} ) {
        $parameters->{shelf_uids} = $opts{shelf_uids};
    }
    my $strings;

    if ( $opts{default_strings} eq "1" ) {
        $strings = [
            "PMC Firmware assert",
            "watchdog",
            "crash",
            "Failed to start Discovery service",
            "Failed to ping HAL watchdog",
            "Failed to configure drive phy",
            "DrvMgr service was unable to get CLI subsystem handle",
            "Unable to get number of PCM",
            "Customer region CRC error",
            "Failed to read from customer region",
            "Clock Fault",
            "Failed to get board ID, targeted download unavailable",
            "Download aborted. Bad data at offset",
            "Failed to get acpEnabled flag",
            "Failed to get acpmgr thread timeout",
            "Failed to get heatbeat info from HAL layer",
            "Fault processing failed",
            "Failed to get DrvMgr handle",
            "Failed to get HID handle",
            "Failed to get General handle",
            "Failed to get Expander Manager handle",
            "Failed to get HAL context",
            "Watchdog",
            "NMI_Type",
            "Temperature jump detected"
        ];
    } ## end if ( $opts{default_strings...

    if ( defined $opts{include_strings} ) {
        my $include_strings = $opts{include_strings};
        foreach my $string (@$include_strings) {
            push @$strings, $string;
        }
    } ## end if ( defined $opts{include_strings...

    if ( defined $opts{exclude_strings} ) {
        my $unmatched_strings;
        my $exclude_strings = $opts{exclude_strings};
        foreach my $string (@$strings) {
            my $match_flag = 0;
            foreach my $exclude_string (@$exclude_strings) {
                if ( $string eq $exclude_string ) {
                    $match_flag = 1;
                }
            }
            if ( $match_flag == 0 ) {
                push @$unmatched_strings, $string;
            }
        } ## end foreach my $string (@$strings)
        $strings = $unmatched_strings;
    } ## end if ( defined $opts{exclude_strings...

    if ((      defined $opts{serials}
            || defined $opts{channels}
            || defined $opts{shelf_ids}
            || defined $opts{modules}
            || $opts{shelf_uids}
        )
        && ( $#$strings != -1 )
        )
    {
        $parameters->{strings}    = $strings;
        $parameters->{shelf_logs} = $shelf_logs;
        if ( defined $opts{string_stats} ) {
            $parameters->{string_stats} = $opts{string_stats};
        }
        $selected_records = log_search_serial_strings($parameters);
    } elsif ( defined $opts{serials}
        || defined $opts{channels}
        || defined $opts{shelf_ids}
        || defined $opts{modules}
        || $opts{shelf_uids} )
    {
        $selected_records = log_search_serials_csms($parameters);
    } elsif ( $#$strings != -1 ) {
        $parameters->{strings}    = $strings;
        $parameters->{shelf_logs} = $shelf_logs;
        if ( defined $opts{string_stats} ) {
            $parameters->{string_stats} = $opts{string_stats};
        }
        $selected_records = log_search_strings($parameters);
	print "\nend: log_search_strings with \n";
	print Dumper $selected_records;
	print "finished dump\n";
    } else {
        $selected_records->{count}   = $#$myrecords + 1;
        $selected_records->{records} = [ 0 .. $#$myrecords ];
    }
    my $result;
    $parameters = {
        shelf_logs  => $shelf_logs,
        log_records => $myrecords,
        records     => $selected_records->{records}
    };

    logcomment("Filtered logs with error counts:");
    print $selected_records->{records};
    print Dumper $selected_records->{records};

    if ( defined $selected_records->{string_stats} ) {
        $parameters->{string_stats} = $selected_records->{string_stats};
	print "string_stats defined\n";
    }
    if ( defined $opts{log_also} ) {
        $parameters->{log_also} = $opts{log_also};
	print "log_also defined\n";
    }
    if ( defined $opts{dump_file} ) {
        $parameters->{dump_file} = $opts{dump_file};
	print "dump_file defined\n";
    }
    if ( defined $opts{header_only} ) {
        $parameters->{header_only} = $opts{header_only};
	print "header_only defined\n";
    }

    if (defined $parameters->{records}) { 
        # records will be undefined if no failures were identified.  records
        # must be an arrayref when calling log_print_records so check first
        $result = log_print_records($parameters);
    } else {
        logcomment( "No fault records found" );
        $result = 1;
    }

    if ( defined $opts{filer} ) {
	my $log_umount_parameters;
        if ( defined $shelf_logs->{fh} ) {
            my @log_array = @{ $shelf_logs->{fh} };
            untie @log_array;
            delete $shelf_logs->{fh};
        }
	$log_umount_parameters->{log_file} = $shelf_logs->{log_files}[0]->{name};
	#$log_umount_parameters->{filer} = $opts{filer};
	$log_umount_parameters->{client_hostname} = $opts{client_hostname} if( defined $opts{client_hostname} );
        log_umount( $log_umount_parameters );
    } ## end if ( defined $opts{filer...
    if ( $result != 1 ) {
        return -1;
    } else {
        return $selected_records->{count};
    }
} ## end sub log_search

=head2 log_extract_records

     my $result = log_extract_records({ log_files => $log_files, log_start_marker => $log_start_marker, log_end_marker => $log_end_marker, serials => serials, include_strings => $include_strings, exclude_strings => $exclude_strings, header_only => $header_only, filer => $filer, filer_conn => $filer_connection_type, shelf_type => $shelf_type, sudo_path = $sudo_path});

This routine is a wrapper for searching and retrieving specific logs from the log file.  If the log file is specified, the contents of the log file $log_file are parsed.  If the filer with filer connection and shelf type are specified, the subroutine mounts the filer root volume and extracts the correct log file name and parses it.  The parameter sudo path is specified if the sudo command is not in default location /usr/local/bin.  The parameters log_start_marker and log_end_marker specify the starting and ending line numbers in the log file for parsing.  The logs are categorised into log records and stored as an array of hash references linked with the log file based on the line numbers.  The parameters serials can be specified to search for a particular module serial number records in the log file.  A default set of strings are always searched in the log file for all the modules and the subroutine returns 0 on finding these strings.  If the strings set needs to be modified, the parameters include_strings and exclude strings can be used to include or exclude a set of strings respectively.  These parameters need to be reference to an array of strings.  The parameter header_only when set to 1 retrieves only the header part of the logs that match the criteria.  string_stats when set would provide the details of the strings and the count of strings found in the log records.

The subroutine returns

	* the count of matched module serial numbers or strings on success
	* a hash reference to the statistics of string for each shelf controller specified
	* the header and body of the log records that match the critieria

Currenlty this subroutine supports only SAS logs.

=over

=item log_file

(Optional) : The log file with absolute path.

=item log_start_marker

(Optional) : This is the line number in the log file from which the script should parse data.  Default is start of log file.

=item log_end_marker

(Optional) : This is the line number in the log file until which the script should parse data.  Default is end of log file.

=item serials

(Optional) : The module serial numbers that have to be selected for processing.  If not specified all the serial numbers are processed.

=item include_strings

(Optional) : The strings that have to be added to the default set of strings for searching.  Default strings that would be searched are "PMC Firmware assert", "watchdog" and "crash".

=item exclude_strings

(Optional) : The strings that have to be ignored even if they exist when searching the logs.

=item header_only

(Optional) : This parameter when set to 1 would print only the header information in the dump file or the log file after processing.  Default value of this parameter is 0.

=item filer

(Optional) : The filer from which the log file has to be parsed.

=item filer_conn

(Optional) : This defines the way to connect to the filer.  For default value refer to section log_mount.

=item shelf_type

(Optional) : This defines the type of shelf.  For default value refer to section log_mount.

=item sudo_path

(Optional) : Defines the path of sudo on the client. For default value refer to section log_mount.

= item string_stats

(Optional) : Flag for getting the statistics of each string found the shelf controllers.  Default is on

=back

=cut

sub log_extract_records {
    my (%opts) = validate(
        @_,
        {   filer            => { type => SCALAR,   optional => 1 },
            filer_conn       => { type => SCALAR,   optional => 1 },
            shelf_type       => { type => SCALAR,   optional => 1 },
            sudo_path        => { type => SCALAR,   optional => 1 },
            timeout          => { type => SCALAR,   optional => 1 },
            channels         => { type => ARRAYREF, optional => 1 },
            serials          => { type => ARRAYREF, optional => 1 },
            modules          => { type => ARRAYREF, optional => 1 },
            shelf_ids        => { type => ARRAYREF, optional => 1 },
            shelf_uids       => { type => ARRAYREF, optional => 1 },
            log_files        => { type => ARRAYREF, optional => 1 },
            log_start_marker => { type => HASHREF,  optional => 1 },
            log_end_marker   => { type => HASHREF,  optional => 1 },
            include_strings  => { type => ARRAYREF, optional => 1 },
            default_strings =>
                { type => SCALAR, optional => 1, default => 1 },
            exclude_strings => { type => ARRAYREF, optional => 1 },
            header_only => { type => SCALAR, optional => 1, default => 0 },
            string_stats => { type => SCALAR, optional => 1, default => 1 },
	    client_hostname => {type => SCALAR, optional => 1 }
        }
    );

    my $shelf_logs = {};
    if ( defined $opts{log_start_marker} ) {
        $shelf_logs->{log_start_marker} = $opts{log_start_marker};
        logcomment("log_extract_records: Log start marker is set to $shelf_logs->{log_start_marker}->{line_number}") if($talk);
    } ## end if ( defined $opts{log_start_marker...
    if ( defined $opts{log_end_marker} ) {
        $shelf_logs->{log_end_marker} = $opts{log_end_marker};
        logcomment("log_extract_records: Log end marker is set to $shelf_logs->{log_end_marker}->{line_number}") if($talk);
    } ## end if ( defined $opts{log_end_marker...

    #Return count as 0 if no search criteria is specified
    if (   ( !defined $opts{serials} )
        && ( !defined $opts{channels} )
        && ( !defined $opts{modules} )
        && ( !defined $opts{shelf_ids} )
        && ( !defined $opts{shelf_uids} )
        && ( !defined $opts{include_strings} )
        && ( $opts{default_strings} == 0 ) )
    {
        my $rvalue = {};
        $rvalue->{count} = 0;
        return $rvalue;
    } ## end if ( ( !defined $opts{...

    #If filer is specified mount the filer and get the log file path
    #If log files are specified get the list of log files
    #Fail if nothing is specified
    my $log_files;
    if ( defined $opts{log_files} ) {
        $log_files = $opts{log_files};
    } elsif ( defined $opts{filer} ) {
        my $log_mount_parameters;
        $log_mount_parameters->{filer} = $opts{filer};
        if ( defined $opts{filer_conn} ) {
            $log_mount_parameters->{filer_conn} = $opts{filer_conn};
        }
        if ( defined $opts{shelf_type} ) {
            $log_mount_parameters->{shelf_type} = $opts{shelf_type};
        }
        if ( defined $opts{sudo_path} ) {
            $log_mount_parameters->{sudo_path} = $opts{sudo_path};
        }
	if ( defined $opts{client_hostname} ) {
	    $log_mount_parameters->{client_hostname} = $opts{client_hostname};
	}

        $log_files = log_mount($log_mount_parameters);
        $log_files = [$log_files];
    } else {
        logresult( "FAIL",
            "log_extract_records: Please provide atleast a filer name or a log file"
        );
        my $rvalue = {};
        $rvalue->{count} = 0;
        return $rvalue;
    } ## end else [ if ( defined $opts{log_files...

    #Check if the log files exist and are accessible
    my $check;
    my $log_exist_parameters;
    $log_exist_parameters->{log_filers} = $log_files;
    $log_exist_parameters->{shelf_logs} = $shelf_logs;
    if ( defined $opts{filers} ){
        $log_exist_parameters->{filer} = $opts{filer};
    }
    if ( defined $opts{client_hostname} ){
        $log_exist_parameters->{client_hostname} = $opts{client_hostname};
    }
    $check = log_exist( $log_exist_parameters );  

    #Return search count as zero if log file is not updated with any data
    if (   ( defined $shelf_logs->{log_start_marker} )
        && ( defined $shelf_logs->{log_start_marker} ) )
    {
        if ($#{ $shelf_logs->{log_files} } == 0
            && ( $shelf_logs->{log_start_marker}->{line_number}
                == $shelf_logs->{log_end_marker}->{line_number} )
            )
        {
            logcomment("log_extract_records: No new information has been logged between the start and end markers");
            my $result = {};
            $result->{count} = 0;
            return $result;
        } ## end if ( $#{ $shelf_logs->...
    } ## end if ( ( defined $shelf_logs...

    #Generate records from the shelf log file
    my $myrecords;
    if (!(  $myrecords = log_generate_records( { shelf_logs => $shelf_logs } )
        )
        )
    {
        my $result;
        $result->{count} = 0;
        logcomment("log_extract_records: No new records found while parsing the Shelf logs");
        return $result;
    } ## end if ( !( $myrecords = log_generate_records...

#Search for specified channels, modeuls, shelf ids or shelf uids in the records from the logs, else continue to search for strings
    my $selected_records;
    if (   ( defined $opts{channels} )
        || ( defined $opts{modules} )
        || ( defined $opts{shelf_ids} )
        || ( defined $opts{shelf_uids} ) )
    {
        my $parameters;
        $parameters->{log_records} = $myrecords;
        if ( defined $opts{serials} ) {
            $parameters->{serials} = $opts{serials};
        }
        if ( defined $opts{channels} ) {
            $parameters->{channels} = $opts{channels};
        }
        if ( defined $opts{modules} ) {
            $parameters->{modules} = $opts{modules};
        }
        if ( defined $opts{shelf_ids} ) {
            $parameters->{shelf_ids} = $opts{shelf_ids};
        }
        if ( defined $opts{shelf_uids} ) {
            $parameters->{shelf_uids} = $opts{shelf_uids};
        }
        $selected_records = log_search_serials_csms($parameters);

        if ( defined $selected_records ) {
            if ( defined $selected_records->{count} ) {
                if ( $selected_records->{count} == 0 ) {
                    logcomment(
                        "log_extract_records: No records found in search (1)"
                    );
                    return $selected_records;
                } ## end if ( $selected_records...
            } else {
                my $rvalue = {};
                $rvalue->{count} = 0;
                return $rvalue;
            }
        } else {
            my $rvalue = {};
            $rvalue->{count} = 0;
            return $rvalue;
        }

    } else {
        $selected_records->{count}   = $#$myrecords + 1;
        $selected_records->{records} = [ 0 .. $#$myrecords ];
    }

    #Search for default set of strings if default_strings is set to 1
    my $strings;

    if ( $opts{default_strings} eq "1" ) {
        $strings = [
            "PMC Firmware assert",
            "watchdog",
            "crash",
            "Failed to start Discovery service",
            "Failed to ping HAL watchdog",
            "Failed to configure drive phy",
            "DrvMgr service was unable to get CLI subsystem handle",
            "Unable to get number of PCM",
            "Customer region CRC error",
            "Failed to read from customer region",
            "Clock Fault",
            "Failed to get board ID, targeted download unavailable",
            "Download aborted. Bad data at offset",
            "Failed to get acpEnabled flag",
            "Failed to get acpmgr thread timeout",
            "Failed to get heatbeat info from HAL layer",
            "Fault processing failed",
            "Failed to get DrvMgr handle",
            "Failed to get HID handle",
            "Failed to get General handle",
            "Failed to get Expander Manager handle",
            "Failed to get HAL context",
            "Watchdog",
            "NMI_Type",
            "Temperature jump detected"
        ];
    } ## end if ( $opts{default_strings...

    #Add specified strings to search if include_strings are specified
    if ( defined $opts{include_strings} ) {
        my $include_strings = $opts{include_strings};
        foreach my $string (@$include_strings) {
            push @$strings, $string;
        }
    } ## end if ( defined $opts{include_strings...

    #Remove specified strings from search if exclude_strings are specified
    if ( defined $opts{exclude_strings} ) {
        my $unmatched_strings;
        my $exclude_strings = $opts{exclude_strings};
        foreach my $string (@$strings) {
            my $match_flag = 0;
            foreach my $exclude_string (@$exclude_strings) {
                if ( $string eq $exclude_string ) {
                    $match_flag = 1;
                }
            }
            if ( $match_flag == 0 ) {
                push @$unmatched_strings, $string;
            }
        } ## end foreach my $string (@$strings)
        $strings = $unmatched_strings;
    } ## end if ( defined $opts{exclude_strings...

    #Search for specified strings in the log records
    if ( $#$strings >= 0 ) {
        my $parameters;
        $parameters->{shelf_logs}   = $shelf_logs;
        $parameters->{log_records}  = $myrecords;
        $parameters->{string_stats} = $opts{string_stats};
        if ( defined $selected_records ) {
            if ( defined $selected_records->{count} ) {
                if ( $selected_records->{count} != 0 ) {
                    $parameters->{selected_records}
                        = $selected_records->{records};
                }
            } ## end if ( defined $selected_records...
        } ## end if ( defined $selected_records)
        $parameters->{strings} = $strings;
        $selected_records = log_search_strings($parameters);
        if ( defined $selected_records ) {
            if ( defined $selected_records->{count} ) {
                if ( $selected_records->{count} == 0 ) {
                    logcomment(
                        "log_extract_records: No records found in search (2)"
                    );
                    return $selected_records;
                } ## end if ( $selected_records...
            } else {
                my $rvalue = {};
                $rvalue->{count} = 0;
                return $rvalue;
            }
        } else {
            my $rvalue = {};
            $rvalue->{count} = 0;
            return $rvalue;
        }
    } ## end if ( $#$strings >= 0 )

#Extract the records that match the criterial to an array of hash references and return to the caller
    my $result = {};
    $result->{count}        = $selected_records->{count};
    $result->{string_stats} = $selected_records->{string_stats};

    if ( $opts{shelf_type} eq 'SAS' ) {
        my @result_array            = ();
        my $records_to_be_extracted = $selected_records->{records};
        my @records_to_be_extracted = ();
        @records_to_be_extracted = @$records_to_be_extracted;

        foreach my $record (@records_to_be_extracted) {
            my $extracted_record = $$myrecords[$record];
            if ( $opts{header_only} == 0 ) {
                my @data = ();
                for (
                    my $lineno = $$myrecords[$record]->{start_block} + 1;
                    $lineno <= $$myrecords[$record]->{end_block};
                    $lineno++
                    )
                {
                    my $line;
		    logcomment("calling log read line 7");
                    if (!(  $line = log_read_line(
                                {   shelf_logs => $shelf_logs,
                                    line       => $lineno
                                }
                            )
                        )
                        )
                    {
                        $Log->warn(
                            "log_extract_records: Unexpected undef encountered: Dumping current data from variables"
                        );
			#print Dumper($shelf_logs);
                        logresult( "Script",
                            "log_read_line: Encountered an unexpected error(3).  Please contact the author with the logs"
                        );
                        return undef;
                    } ## end if ( !( $line = log_read_line...
                    chomp($line);
                    push @data, $line;
                } ## end for ( my $lineno = $$myrecords...
                $extracted_record->{data} = \@data;
            } ## end if ( $opts{header_only...
            push @result_array, $extracted_record;
        } ## end foreach my $record (@records_to_be_extracted)
        $result->{records} = \@result_array;
    } ## end if ( $opts{shelf_type}...

    if ( defined $opts{filer} ) {
        my $log_umount_parameters;
        $log_umount_parameters->{client_hostname} = $opts{client_hostname} if( !defined ( $opts{client_hostname} ));
        if ( defined $shelf_logs->{fh} ) {
            my @log_array = @{ $shelf_logs->{fh} };
            untie @log_array;
            delete $shelf_logs->{fh};
        }
        $log_umount_parameters->{log_file} = $shelf_logs->{log_files}[0]->{name};
        log_umount( $log_umount_parameters );
    } ## end if ( defined $opts{filer...
    return $result;
} ## end sub log_extract_records

#######################################################
#
#
#######################################################

sub log_generate_stats {
    my (%opts) = validate( @_, { shelf_logs => { type => HASHREF }, } );
    my $shelf_logs = $opts{shelf_logs};

    my $rvalue = ();
    my $line;
    my $block = 0;
    print "Generating records...\n\n";

    log_read_first_line( { shelf_logs => $shelf_logs } );
    $shelf_logs->{current_line}--;
    my $start_block_flag = 0;

    while ( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) {

        #print "#####\nline: $line\n#####\n";
        chomp($line);
        if ( $line
            =~ /--------------------------------------------------------------------------------------/
            )
        {
            if ( $start_block_flag == 0 ) {
                $start_block_flag = 1;
                next;
            }
        } ## end if ( $line =~ ...

        if ( ( $start_block_flag == 1 ) && ( $line =~ /Timestamp: / ) ) {
            my $line_number = $shelf_logs->{current_line} - 1;
            log_goto_line(
                { shelf_logs => $shelf_logs, line => $line_number } );

            my $hash_ref
                = log_process_stat_block( { shelf_logs => $shelf_logs } );
            $$rvalue[$block] = $hash_ref;
            $start_block_flag = 0;
            $block++;
        } ## end if ( ( $start_block_flag...
    } ## end while ( $line = log_read_next_line...

    return $rvalue;

} ## end sub log_generate_stats

#######################################################
##$VAR1 = [
#          {
#            'termswitch' => 'N/A',
#            'end_block' => 18,
#            'start_block' => 11,
#            'shelf_id' => '1',
#            'channel' => '3a',
#            'shelf_uid' => '50:05:0c:c1:02:00:02:6e',
#            'module_state' => 'OK',
#            'timestamp' => 'Mon Dec  1 23:35:12 GMT 2008',
#            'shelf_state' => 'ONLINE',
#            'shelf_name' => '3a.shelf1',
#            'module' => 'A'
#          },
#          {
#            'termswitch' => 'N/A',
#            'end_block' => 71,
#            'start_block' => 64,
#            'shelf_id' => '1',
#            'channel' => '3c',
#            'shelf_uid' => '50:05:0c:c1:02:00:02:6e',
#            'module_state' => 'OK',
#            'timestamp' => 'Mon Dec  1 23:35:12 GMT 2008',
#            'shelf_state' => 'ONLINE',
#            'shelf_name' => '3c.shelf1',
#            'module' => 'B'
#          }
#        ];
#
#
#######################################################

sub log_process_stat_block {
    my (%opts) = validate( @_, { shelf_logs => { type => HASHREF }, } );
    my $shelf_logs = $opts{shelf_logs};
    my $rvalue     = {};
    my $pos;
    my $line;

    $line = log_read_next_line( { shelf_logs => $shelf_logs } );
    chomp($line);
    {
        $rvalue->{timestamp} = $line;
        $pos = log_search_error_check( $line, "Timestamp:" );
        $rvalue->{timestamp} =~ s/(.{$pos})\s+(.*)$/$2/g;

        #print "rvalue->{timestamp} = $rvalue->{timestamp}\n";
    }

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    {
        $rvalue->{shelf_name} = $line;
        $pos = log_search_error_check( $line, "Shelf name:" );
        $rvalue->{shelf_name} =~ s/(.{$pos})\s+(\S+)(.*)$/$2/g;

        #print "rvalue->{shelf_name} = $rvalue->{shelf_name}\n";
    }

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    {
        $rvalue->{channel} = $line;
        $pos = log_search_error_check( $line, "Channel:" );
        $rvalue->{channel} =~ s/(.{$pos})\s+(\S+)(.*)$/$2/g;

        #print "rvalue->{channel} = $rvalue->{channel}\n";
    }

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    {
        $rvalue->{module} = $line;
        $pos = log_search_error_check( $line, "Module:" );
        $rvalue->{module} =~ s/(.{$pos})\s+(\S+)(.*)$/$2/g;

        #print "rvalue->{module} = $rvalue->{module}\n";
    }

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    {
        $rvalue->{shelf_id} = $line;
        $pos = log_search_error_check( $line, "Shelf id:" );
        $rvalue->{shelf_id} =~ s/(.{$pos})\s+(\S+)(.*)$/$2/g;

        #print "rvalue->{shelf_id} = $rvalue->{shelf_id}\n";
    }

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    {
        $rvalue->{shelf_uid} = $line;
        $pos = log_search_error_check( $line, "Shelf UID:" );
        $rvalue->{shelf_uid} =~ s/(.{$pos})\s+(\S+)(.*)$/$2/g;

        #print "rvalue->{shelf_uid} = $rvalue->{shelf_uid}\n";
    }

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    {
        $rvalue->{termswitch} = $line;
        $pos = log_search_error_check( $line, "Term switch:" );
        $rvalue->{termswitch} =~ s/(.{$pos})\s+(\S+)(.*)$/$2/g;

        #print "rvalue->{module} = $rvalue->{module}\n";
    }

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    {
        $rvalue->{shelf_state} = $line;
        $pos = log_search_error_check( $line, "Shelf state:" );
        $rvalue->{shelf_state} =~ s/(.{$pos})\s+(\S+)(.*)$/$2/g;

        #print "rvalue->{shelf_state} = $rvalue->{shelf_state}\n";
    }

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    {
        $rvalue->{module_state} = $line;
        $pos = log_search_error_check( $line, "Module state:" );
        $rvalue->{module_state} =~ s/(.{$pos})\s+(.*)$/$2/g;

        #print "rvalue->{module_state} = $rvalue->{module_state}\n";
    }

    $rvalue->{start_block} = $shelf_logs->{current_line} + 1;
    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    if ( $line !~ /^\n+|\r+/ ) {
        logresult( "FAIL",
            "The log is corrupted near line $shelf_logs->{current_line}" );
        return undef;
    }
    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    if ( $line !~ /^\n+|\r+/ ) {
        logresult( "FAIL",
            "The log is corrupted near line $shelf_logs->{current_line}" );
        return undef;
    }

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    if (!(     ( $line =~ /Partial Path/ )
            && ( $line =~ /Link/ )
            && ( $line =~ /Invalid/ )
            && ( $line =~ /Running/ )
            && ( $line =~ /Loss/ )
            && ( $line =~ /CRC/ )
            && ( $line =~ /Phy/ )
        )
        )
    {
        logresult( "FAIL",
            "The log is corrupted near line $shelf_logs->{current_line}" );
        return undef;
    } ## end if ( !( ( $line =~ /Partial Path/...

    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);
    if (!(     ( $line =~ /^Disk/ )
            && ( $line =~ /Port/ )
            && ( $line =~ /Timeout/ )
            && ( $line =~ /Rate/ )
            && ( $line =~ /Dword/ )
            && ( $line =~ /Disparity/ )
            && ( $line =~ /Reset/ )
            && ( $line =~ /Error/ )
            && ( $line =~ /Change/ )
        )
        )
    {
        logresult( "FAIL",
            "The log is corrupted near line $shelf_logs->{current_line}" );
        return undef;
    } ## end if ( !( ( $line =~ /^Disk/...
    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);

    if (!(     ( $line =~ /^Id/ )
            && ( $line =~ /State/ )
            && ( $line =~ /Value/ )
            && ( $line =~ /Count/ )
            && ( $line =~ /Problem/ )
        )
        )
    {
        logresult( "FAIL",
            "The log is corrupted near line $shelf_logs->{current_line}" );
        return undef;
    } ## end if ( !( ( $line =~ /^Id/...
    if ( !( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) ) {
        return undef;
    }
    chomp($line);

    if (!(  $line
            =~ /--------------------------------------------------------------------------------------------/
        )
        )
    {
        logresult( "FAIL",
            "The log is corrupted near line $shelf_logs->{current_line}" );
        return undef;
    } ## end if ( !( $line =~ ...
    my $count = 0;
    while ( $line = log_read_next_line( { shelf_logs => $shelf_logs } ) ) {
        if ( $line !~ /^\[/ ) {
            last;
        }
        $count++;

    } ## end while ( $line = log_read_next_line...
    if ( $count != 36 ) {
        logresult( "FAIL", "The log file is corrupt" );
        return undef;
    }
    my $line_number = $shelf_logs->{current_line} - 1;
    log_goto_line( { shelf_logs => $shelf_logs, line => $line_number } );

    $rvalue->{end_block} = $shelf_logs->{current_line};

    return $rvalue;
} ## end sub log_process_stat_block

#######################################################
#
#
#######################################################

sub log_search_stats {
    my (%opts) = validate(
        @_,
        {   channels   => { type => ARRAYREF, optional => 1 },
            modules    => { type => ARRAYREF, optional => 1 },
            shelf_ids  => { type => ARRAYREF, optional => 1 },
            shelf_uids => { type => ARRAYREF, optional => 1 },
            display => { type => SCALAR, optional => 1, default => "normal" },
            log_files        => { type => ARRAYREF, optional => 1 },
            log_start_marker => { type => HASHREF,  optional => 1 },
            log_end_marker   => { type => HASHREF,  optional => 1 },
            dump_file        => { type => SCALAR,   optional => 1 },
            log_also   => { type => SCALAR,   optional => 1, default => 0 },
            filer      => { type => SCALAR,   optional => 1 },
            filer_conn => { type => SCALAR,   optional => 1 },
            shelf_type => { type => SCALAR,   optional => 1 },
            sudo_path  => { type => SCALAR,   optional => 1 },
            phys       => { type => ARRAYREF, optional => 1 },
	    client_hostname => {type => SCALAR, optional => 1 }
        }
    );

    my $shelf_logs = {};
    if ( defined $opts{log_start_marker} ) {
        $shelf_logs->{log_start_marker} = $opts{log_start_marker};
        logcomment(
            "Log start marker is set to $shelf_logs->{log_start_marker}->{line_number}"
        ) if($talk);
    } ## end if ( defined $opts{log_start_marker...
    if ( defined $opts{log_end_marker} ) {
        $shelf_logs->{log_end_marker} = $opts{log_end_marker};
        logcomment(
            "Log end marker is set to $shelf_logs->{log_end_marker}->{line_number}"
        ) if($talk);
    } ## end if ( defined $opts{log_end_marker...

    my $log_files;
    if ( defined $opts{log_files} ) {
        $log_files = $opts{log_files};
    } elsif ( defined $opts{filer} ) {
        my $log_mount_parameters;
        $log_mount_parameters->{filer} = $opts{filer};
        if ( defined $opts{filer_conn} ) {
            $log_mount_parameters->{filer_conn} = $opts{filer_conn};
        }
        if ( defined $opts{shelf_type} ) {
            $log_mount_parameters->{shelf_type} = $opts{shelf_type};
        }
        if ( defined $opts{sudo_path} ) {
            $log_mount_parameters->{sudo_path} = $opts{sudo_path};
        }
	if ( defined $opts{client_hostname} ) {
	    $log_mount_parameters->{client_hostname} = $opts{client_hostname};
        }
        $log_files = log_mount($log_mount_parameters);
        $log_files = [$log_files];

    } else {
        logresult( "FAIL",
            "Please provide atleast the log file or the filer, filer connection and shelf type to the script"
        );
        return -1;
    } ## end else [ if ( defined $opts{log_files...
    my $check;
    my $log_exist_parameters;
    $log_exist_parameters->{log_files} = $opts{log_files};
    $log_exist_parameters->{shelf_logs} = $opts{shelf_logs};
    if( defined $opts{filer} ) {
        $log_exist_parameters->{filer} = $opts{filer};
    }
    if( defined $opts{client_hostname} ) {
        $log_exist_parameters->{client_hostname} = $opts{client_hostname};
    }
    $check = log_exist( $log_exist_parameters );

    if ($#{ $shelf_logs->{log_files} } == 0
        && ( $shelf_logs->{log_start_marker}->{line_number}
            == $shelf_logs->{log_end_marker}->{line_number} )
        )
    {
        logcomment(
            "There was no update in the log file, as the start and end markers are same"
        );
        return -2;
    } ## end if ( $#{ $shelf_logs->...

    my $myrecords = log_generate_stats( { shelf_logs => $shelf_logs } );

    if ( !( defined $myrecords ) ) {
        logresult( "FAIL", "No records were found" );
        return 0;
    }

    my $parameters = {};
    $parameters->{log_records} = $myrecords;
    if ( defined $opts{channels} ) {
        $parameters->{channels} = $opts{channels};
    }
    if ( defined $opts{shelf_ids} ) {
        $parameters->{shelf_ids} = $opts{shelf_ids};
    }
    if ( defined $opts{modules} ) {
        $parameters->{modules} = $opts{modules};
    }
    if ( defined $opts{shelf_uids} ) {
        $parameters->{shelf_uids} = $opts{shelf_uids};
    }
    my $selected_records = log_search_stat_records($parameters);

    if ( !( defined $selected_records ) ) {
        logresult( "FAIL", "No records were found" );
        return 0;
    }
    my $result;
    $parameters                = {};
    $parameters->{shelf_logs}  = $shelf_logs;
    $parameters->{log_records} = $myrecords;
    $parameters->{records}     = $selected_records->{records};
    $parameters->{log_also}    = $opts{log_also};

    if ( defined $opts{dump_file} ) {
        $parameters->{dump_file} = $opts{dump_file};
    }

    if ( $opts{display} eq "normal" ) {
        $parameters->{stats} = 1;
        $result = log_print_records($parameters);
    } else {
        if ( defined $opts{phys} ) {
            $parameters->{phys} = $opts{phys};
        }
        $result = log_print_stats($parameters);
    } ## end else [ if ( $opts{display} eq...

    logcomment("The match count is $selected_records->{count}") if($talk);
    if ( $result != 1 ) {
        return -1;
    } else {
        return $selected_records->{count};
    }
} ## end sub log_search_stats

#######################################################
#
#
#######################################################

sub log_search_element {
    my (%opts) = validate(
        @_,
        {   log_records => { type => ARRAYREF },
            records     => { type => ARRAYREF },
            parameter   => { type => SCALAR },
            elements    => { type => ARRAYREF }
        }
    );

    my $log_records = $opts{log_records};
    my $records     = $opts{records};
    my $parameter   = $opts{parameter};
    my $elements    = $opts{elements};

    if ( $#$log_records < 0 ) {
        my $rvalue = {};
        $rvalue->{count} = 0;
        return $rvalue;
    }

    my $matched_records = ();
    foreach my $record (@$records) {
        my $match_flag = 0;
        foreach my $element (@$elements) {
            if ( $$log_records[$record]->{$parameter} eq $element ) {
                $match_flag = 1;
            }
        }
        if ( $match_flag == 1 ) {
            push @$matched_records, $record;
        }
    } ## end foreach my $record (@$records)

    if ( $#$matched_records != -1 ) {
        return $matched_records;
    } else {
        return undef;
    }
} ## end sub log_search_element

#######################################################
#
#
#######################################################

sub log_search_stat_records {
    my (%opts) = validate(
        @_,
        {   log_records => { type => ARRAYREF },
            channels    => { type => ARRAYREF, optional => 1 },
            modules     => { type => ARRAYREF, optional => 1 },
            shelf_ids   => { type => ARRAYREF, optional => 1 },
            shelf_uids  => { type => ARRAYREF, optional => 1 },
        }
    );
    my $log_records = $opts{log_records};
    my $matched_records;
    my $match_count = 0;

    for ( my $i = 0; $i <= $#$log_records; $i++ ) {
        push @$matched_records, $i;
    }
    $match_count = $#$matched_records + 1;

    if ( defined $opts{channels} ) {
        my $channels = $opts{channels};
        logcomment("Searching for Channels: @$channels");
        $matched_records = log_search_element(
            {   log_records => $log_records,
                records     => $matched_records,
                parameter   => "channel",
                elements    => $channels
            }
        );
    } ## end if ( defined $opts{channels...

    if ( ( defined $matched_records ) && ( defined $opts{shelf_ids} ) ) {
        my $shelf_ids = $opts{shelf_ids};
        logcomment("Searching for Shelf IDs: @$shelf_ids");
        $matched_records = log_search_element(
            {   log_records => $log_records,
                records     => $matched_records,
                parameter   => "shelf_id",
                elements    => $shelf_ids
            }
        );
    } ## end if ( ( defined $matched_records...

    if ( ( defined $matched_records ) && ( defined $opts{modules} ) ) {
        my $modules = $opts{modules};
        logcomment("Searching for Modules: @$modules");
        $matched_records = log_search_element(
            {   log_records => $log_records,
                records     => $matched_records,
                parameter   => "module",
                elements    => $modules
            }
        );
    } ## end if ( ( defined $matched_records...

    if ( ( defined $matched_records ) && ( defined $opts{shelf_uids} ) ) {
        my $shelf_uids = $opts{shelf_uids};
        logcomment("Searching for Shelf UIDs: @$shelf_uids");
        $matched_records = log_search_element(
            {   log_records => $log_records,
                records     => $matched_records,
                parameter   => "shelf_uid",
                elements    => $shelf_uids
            }
        );
    } ## end if ( ( defined $matched_records...

    if ( defined $matched_records ) {
        $match_count = $#$matched_records + 1;
        my $rvalue = {
            count   => $match_count,
            records => $matched_records
        };
        return $rvalue;
    } else {
        return undef;
    }
} ## end sub log_search_stat_records

#######################################################
#
#
#######################################################

sub log_print_stats {
    my (%opts) = validate(
        @_,
        {   shelf_logs  => { type => HASHREF },
            log_records => { type => ARRAYREF },
            records     => { type => ARRAYREF },
            dump_file   => { type => SCALAR, optional => 1 },
            phys        => { type => ARRAYREF, optional => 1 },
            log_also    => { type => SCALAR, optional => 1, default => 1 }
        }
    );

    my $shelf_logs  = $opts{shelf_logs};
    my $log_records = $opts{log_records};
    my $records     = $opts{records};
    my $log_also    = $opts{log_also};

    my $dump_file;
    my $FH;
    if ( defined $opts{dump_file} ) {
        $dump_file = $opts{dump_file};
        open $FH, ">> $dump_file"
            or logcomment(
            "Unable to open file due to: $!\nThe records would be printed to the test logs"
            );
    } ## end if ( defined $opts{dump_file...

    my $return_value = 0;

    my @table = ();
    foreach my $record (@$records) {
        my $match_flag = 0;
        my $current_element
            = "$$log_records[$record]->{channel}" . '.'
            . "$$log_records[$record]->{shelf_id}" . '.'
            . "$$log_records[$record]->{module}";
        foreach my $element (@table) {
            if ( $current_element eq $element ) {
                $match_flag = 1;
            }
        }
        if ( $match_flag == 0 ) {
            push @table, $current_element;
        }
    } ## end foreach my $record (@$records)

    my @phys_to_lines = ();
    if ( defined $opts{phys} ) {
        my $phys_ref = $opts{phys};
        my @disk_ids = (
            '[IO 0]', '[IO 1]', '[IO 2]', '[IO 3]', '[IO 4]', '[IO 5]',
            '[IO 6]', '[IO 7]', '[  0 ]', '[  1 ]', '[  2 ]', '[  3 ]',
            '[  4 ]', '[  5 ]', '[  6 ]', '[  7 ]', '[  8 ]', '[  9 ]',
            '[ 10 ]', '[ 11 ]', '[ 12 ]', '[ 13 ]', '[ 14 ]', '[ 15 ]',
            '[ 16 ]', '[ 17 ]', '[ 18 ]', '[ 19 ]', '[ 20 ]', '[ 21 ]',
            '[ 22 ]', '[ 23 ]', '[DIS0]', '[DIS1]', '[DIS2]', '[DIS3]'
        );
        for ( my $i = 0; $i <= $#disk_ids; $i++ ) {
            foreach my $phy (@$phys_ref) {
                if ( $disk_ids[$i] eq $phy ) {
                    push @phys_to_lines, $i;
                }
            }
        } ## end for ( my $i = 0; $i <= ...
    } else {
        @phys_to_lines = ( 0 .. 35 );
    }

    for ( my $i = 0; $i <= $#phys_to_lines; $i++ ) {
        $phys_to_lines[$i] = $phys_to_lines[$i] + 6;
    }

    @table = sort @table;
    if ( defined $FH ) {
        if ( tell($FH) != -1 ) {
            logcomment("Printing Stats to file $dump_file");
        } else {
            logcomment("Printing Stats...");
        }
    } else {
        logcomment("Printing Stats...");
    }

    foreach my $element (@table) {
        my ( $channel, $shelf_id, $module ) = split( /_/, $element );
        my $line
            = "Statistics for Channel: $channel Shelf: $shelf_id Module: $module\n";
        if ( defined $FH ) {
            log_print_line(
                { dump_file => $FH, log_also => $log_also, line => $line } );
        } else {
            log_print_line( { log_also => $log_also, line => $line } );
        }

        $line
            = "--------------------------------------------------------------------------------------------------------------------------\n";
        if ( defined $FH ) {
            log_print_line(
                { dump_file => $FH, log_also => $log_also, line => $line } );
        } else {
            log_print_line( { log_also => $log_also, line => $line } );
        }

        for ( my $i = 0; $i < 3; $i++ ) {
            my $line_no = $$log_records[0]->{start_block} + 2 + $i;
            my $line;
	    logcomment("CALLING log_read_line FROM IF STATEMENT 5");
            if ($line = log_read_line(
                    { shelf_logs => $shelf_logs, line => $line_no }
                )
                )
            {
                chomp($line);
                if ( $i == 2 ) {
                    $line = $line . "           Timestamp";
                }
                $line = $line . "\n";
                if ( defined $FH ) {
                    log_print_line(
                        {   dump_file => $FH,
                            log_also  => $log_also,
                            line      => $line
                        }
                    );
                } else {
                    log_print_line(
                        { log_also => $log_also, line => $line } );
                }
            } ## end if ( $line = log_read_line...
        } ## end for ( my $i = 0; $i < 3...
        foreach my $phy_line (@phys_to_lines) {
            $line
                = "--------------------------------------------------------------------------------------------------------------------------\n";
            if ( defined $FH ) {
                log_print_line(
                    {   dump_file => $FH,
                        log_also  => $log_also,
                        line      => $line
                    }
                );
            } else {
                log_print_line( { log_also => $log_also, line => $line } );
            }
            for ( my $i = 0; $i <= $#$records; $i++ ) {
                if (   ( $$log_records[$i]->{channel} eq $channel )
                    && ( $$log_records[$i]->{shelf_id} eq $shelf_id )
                    && ( $$log_records[$i]->{module}   eq $module ) )
                {
                    my $line;
                    my $line_no
                        = $$log_records[$i]->{start_block} + $phy_line;
			logcomment("log_read_line 9");
                    if ($line = log_read_line(
                            { shelf_logs => $shelf_logs, line => $line_no }
                        )
                        )
                    {
                        chomp($line);
                        $line = $line . "\t"
                            . $$log_records[$i]->{timestamp} . "\n";
                        if ( defined $FH ) {
                            log_print_line(
                                {   dump_file => $FH,
                                    log_also  => $log_also,
                                    line      => $line
                                }
                            );
                        } else {
                            log_print_line(
                                { log_also => $log_also, line => $line } );
                        }
                    } ## end if ( $line = log_read_line...
                } ## end if ( ( $$log_records[$i...
            } ## end for ( my $i = 0; $i <= ...
            $line = "\n";
            if ( defined $FH ) {
                log_print_line(
                    {   dump_file => $FH,
                        log_also  => $log_also,
                        line      => $line
                    }
                );
            } else {
                log_print_line( { log_also => $log_also, line => $line } );
            }
        } ## end foreach my $phy_line (@phys_to_lines)
        $line
            = "--------------------------------------------------------------------------------------------------------------------------\n";
        if ( defined $FH ) {
            log_print_line(
                { dump_file => $FH, log_also => $log_also, line => $line } );
        } else {
            log_print_line( { log_also => $log_also, line => $line } );
        }
    } ## end foreach my $element (@table)
    return 0;
} ## end sub log_print_stats

return 1;
