# Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Storage/StorConfig.pm#1 $
##
## @summary Storage config library
## @author erickson@netapp.com
## @status testme
## @pod here

=head1 NAME

Storage::StorConfig

=head1 DESCRIPTION

This library defines a set of functions that provide hashes and lists of storage devices under test.

Devices under test are those devices on the filer under test that match the storage config types 
specified in the TESTPLAN_CONFIGS in the .plan.

Match has a somewhat "arbitrary" interpretation for now.  Once aggregate (loop) config types are 
supported this can be tightened up.  But for now it works like this:

TESTPLAN_CONFIGS can specify these storage components: 
	adapter model, 
	shelf aggregate, (e.g. ds-ds14mk2-fc-esh4-2g is a ds14mk2 w/esh4 running at 2G speed)
	controller, 
	path (single,dual,cluster), 
	drive model

The current interpretation of the TESTPLAN_CONFIGS (by TestX, nip and nuffo) is not loop-aware. 

For example, if the TESTPLAN_CONFIGS specify an adapter and a shelf, the config will match even 
if the shelf is not on the same loop as the adapter, as long as the filer has each item present 
in its config.  Unfortunately, that is probably not be what the tester wanted to test when he
specified the TESTPLAN_CONFIGS in the .plan.

It is my belief, after talking to other members of the storqa team, that the storage 
TESTPLAN_CONFIGS should be interpreted as specifying loop configurations.  Therefore, this 
library will look for loops with the specified combinations of storage components.  If none
are found, then it will relax the constraints and return all loops on the filer.  This is done
so that testX will still be able to use the "loose" interpretation of TESTPLAN_CONFIGS.

I don't think that in practice this will present too much of a problem.  Testbeds will just need to be
set up in such a way that they match the TESTPLAN_CONFIGS that are specified in the .plans.

Examples of what a storconfig hash will contain:
If a TESTPLAN_CONFIG includes:
    adapter		all loops from matching adapters, with all shelves
    adapter+shelves 	all loops from matching adapters that contain one or more of the shelves, with
				only the matching shelves, with all drives
    shelves+drives	all loops that contain one or more of the shelves with a matching drive, with 
				only the matching shelves, with only the matching drives
    drives		all loops that contain one or more of the drives, with only the shelves that contain
				matching drives and with only the matching drives
    no storage configs	all loops on the filer, all shelves, all drives

etc.

NOTE regarding storconfigs and how the testplan configs are interpreted:

Storconfigs are loop-based.  Testplan Config types apply to loops.  Config types are
restrictions on what must be in a loop.  All loops that match the config_types
for adapter, shelf, drive  and loop will be returned in the storconfig.

If multiple shelf, drive, adapter, or loop configs are included in the TESTPLAN_CONFIGS, they are
interpreted as "or".  This may not be what the test planner intended, but for now we have to go
one way or the other.  If this is not the best solution we can change it later, but it most closely
matches the existing behavior of testX, nip, etc.

If any storage config element (loop, shelf, drive or adapter config type) is not specified, there are 
no restrictions on storage configs for that kind of element.

If any of the storage config elements has a value other than "", then that element is
restricted.  The storconfig will not include any devices of that type other than listed.

NOTE regarding root volume: the default is not to include drives that are part of the root_volume.  
See function descriptions for how to override this behavior.

The storconfig hash will be accessed like this:

Examples of use:

my $storconfig = get_storconfig(conn => $conn);
(keys %{$storconfig{$channel_id}{$shelf_id}} will provide a list of the drives on a particular shelf
(keys %{$storconfig{$channel_id}} will provide a list of the shelves on a particular loop
(keys %$storconfig will provide a list of channel_ids that match the TESTPLAN_CONFIGS, either 
    explicitly or indirectly as a result of loop membership

=head1 SYNOPSIS

use Storage::StorConfig qw(get_storconfig, get_storconfig_drives, compare_storconfig_status, write_storconfig
				read_storconfig);
use boot qw(wait_for_rsh);
 
=head1 ROUTINES

=cut

package Storage::StorConfig;

use strict;
use warnings;

use Carp qw(confess croak);
use Params::Validate qw(validate BOOLEAN OBJECT SCALAR HASHREF);

BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}

use Data::Dumper;

use Tharn qw(say script_result host logcomment logresult param);
use Utils qw(make_conn);
use config_base qw(applies_to_testbed);
use SharedTestBed::StorConfigTypes qw(get_adapter_config get_loop_config get_shelf_config get_drive_config add_inferred_configs config_match);

use Storage::AdaptUtils qw(get_adapter_sysconfig get_link_rate);
use Storage::DiskUtils qw(get_drive_sysconfig);
use Burt qw(burt_begin burt_end ExitSoftError);

use nic::lib::NicCommands qw(compare_args);

use base qw(Exporter);

our @EXPORT_OK = qw(
			get_storconfig
			get_storconfig_uut
			get_storconfig_drives_uut
                   );
#			compare_storconfig_status
#			write_storconfig
#			read_storconfig

our $debug = 1;

#----------------------------------------------------------------------
# The internal routines are coded in the following section
#----------------------------------------------------------------------

######################################################################
#
# The command interface routines are coded below.  These will be broken
# out and added to various storage libraries.  One thing they have in common
# is that they return command output, either as a single string or as an 
# array.  They don't parse the command output and put it in a hash, etc.
# However, they may only return a portion of a command's output.  For example,
# get_adapter_sysconfig returns the portion of sysconfig -v that applies to
# a single adapter port (channel), since there is no sysconfig option to do
# that.
#
######################################################################


sub fabric_is_enabled(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};

    my (undef,undef,$storage_show_fabric) =say($conn, "priv set test; storage show fabric", "-timeout", $timeout);
    chomp( my @storage_show_fabric_lines = split(/\r?\n/,"$storage_show_fabric"));
    @storage_show_fabric_lines = grep (/Fabric:/, @storage_show_fabric_lines);
    return ($#storage_show_fabric_lines >= 0);
}


sub get_storage_show_disk_path(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};

    my (undef,undef,$storage_show_disk_path) = say($conn, "storage show disk -p", "-timeout", $timeout);
    chomp( my @storage_show_disk_path_lines = split(/\r?\n/,"$storage_show_disk_path"));

    if (wantarray){
        return @storage_show_disk_path_lines;
    } else {
        return join ("\n", @storage_show_disk_path_lines);
    }
}


sub get_vol_status_r(@){
    my (%opts) = validate(@_, { conn =>  { type=>OBJECT, isa=>'Connectrec' },
				timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};

    my (undef,undef,$vol_status_r) = say($conn, "vol status -r", "-timeout", $timeout);
    chomp( my @vol_status_r_lines = split(/\r?\n/,"$vol_status_r"));

    if (wantarray){
        return @vol_status_r_lines;
    } else {
        return join ("\n", @vol_status_r_lines);
    }
}


sub get_storage_show_adapter(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};

    my (undef,undef,$storage_show_adapter) = say($conn, "storage show adapter", "-timeout", $timeout);
    chomp( my @storage_show_adapter_lines = split(/\r?\n/,"$storage_show_adapter"));
    if (wantarray){
        return @storage_show_adapter_lines;
    } else {
        return join ("\n", @storage_show_adapter_lines);
    }
}


sub get_fcadmin_device_map(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};

    my (undef,undef,$fcadmin_device_map) = say($conn,"fcadmin device_map", "-timeout", $opts{timeout});
    chomp( my @fcadmin_device_map_lines = split(/\r?\n/,"$fcadmin_device_map"));
    if (wantarray){
        return @fcadmin_device_map_lines;
    } else {
        return join ("\n", @fcadmin_device_map_lines);
    }
}


sub environ_shelf_is_ok(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
    				environ => { type=>SCALAR },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $environ = $opts{environ};
    my $timeout = $opts{timeout};

    if (grep (/Shelf mapping/, $environ)) {
        if ( $environ =~ /Number of shelves monitored:\s+(\d+)\s+/){
            my $num_shelves = $1;
            chomp(my @environ_lines = split(/\r?\n/,"$environ"));
            my @shelf_lines = grep (/Shelf:/, @environ_lines);
            if ($num_shelves == @shelf_lines){
                return 1;
            }
        }
    }
    return 0;
}


sub get_environ_shelf_adapter(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                adapter => { type=>SCALAR },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $adapter = $opts{adapter};
    my $timeout = $opts{timeout};

    burt_begin(203702);
    my (undef,undef,$environ) = say($conn,"environ shelf $adapter", "-timeout", $timeout);

    return $environ if environ_shelf_is_ok(conn => $conn, environ => $environ, timeout => $timeout);

    # If we get this far, environ shelf output was missing information, burt 203702.
    # Now try rebooting filer and see if environ shelf output is ok.  If not, log FATAL CONFIG result.

    ExitSoftError('-burt',"203702", "rebooting filer due to bug 203702: environ shelf is missing shelves");
    burt_end(203702);

    boot::reboot($conn->hostp(), timeout => 180) or die $@;
    boot::wait_for_rsh($conn, timeout => 180) or die $@;
    sleep 30;

    (undef,undef,$environ) = say($conn,"environ shelf $adapter", "-timeout", $timeout);

    return $environ if environ_shelf_is_ok(conn => $conn, environ => $environ, timeout => $timeout);

    logresult ("CONFIG", "filer did not respond to two environ shelf commands with intervening reboot");
}


sub get_environ_shelf_adapter_shelf(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                adapter => { type=>SCALAR },
                                shelf_id => { type=>SCALAR },
				environ_shelf_adapter => { type=>SCALAR, optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $adapter = $opts{adapter};
    my $shelf_id = $opts{shelf_id};
    my $timeout = $opts{timeout};
    my $environ_shelf_adapter = $opts{environ_shelf_adapter} || 
		get_environ_shelf_adapter( conn => $conn, adapter => $adapter, timeout => $timeout);

    chomp( my @shelf_lines = split (/\r?\n/,"$environ_shelf_adapter"));

    my ($drive_map) = grep (/^\s+Shelf\s+$shelf_id:/, @shelf_lines);

    my $shelf_start= -1;
    my $shelf_end = -1;
    foreach my $i (0 .. $#shelf_lines){
	if (($shelf_lines[$i] =~ /^\s+Channel:\s+$adapter\s*$/) && ($shelf_lines[$i+1] =~ /^\s+Shelf:\s+$shelf_id\s*$/)){
	    $shelf_start = $i 
	} elsif (($shelf_start != -1) && ($shelf_lines[$i] =~ /^(\s*|\s+Shelf\smapping.*$adapter:\s*)$/)){
	    $shelf_end = ($i-1);

	    if (wantarray){
	        return  (@shelf_lines[$shelf_start .. $shelf_end], $drive_map);
	    } else {
		my $shelf_lines = join ("\n", @shelf_lines[$shelf_start .. $shelf_end]);
		return "$shelf_lines $drive_map";
	    }
	}
    }
    die "CONFIG: environ shelf not found for $adapter:$shelf_id\n$environ_shelf_adapter\n";

}

###############################################################
#
# The storconf routines are coded below
#
###############################################################


sub _get_storage_testplan_config_types {
    my $storplan_configs = {};

    my @testplan_configs = applies_to_testbed();

    print "DEBUG: testplan configs in _get_storage_testplan_config_types:\n" if $debug;
    print Data::Dumper->Dump([@testplan_configs]) if $debug;

    $storplan_configs->{LOOP} = join (" ", (grep /^dl-/, @testplan_configs));
    $storplan_configs->{ADAPTER} = join (" ", (grep /^fcha-/, @testplan_configs));
    $storplan_configs->{SHELF} = join (" ", (grep /^ds-/, @testplan_configs));
    $storplan_configs->{DRIVE} = join (" ", (grep /^(dd-|X)/, @testplan_configs));

    return $storplan_configs;
}


sub get_adapter_loops(%){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
				adapter => { type=>SCALAR, optional=>1 },
				environ_shelf_adapter => { type=>SCALAR, optional=>1 },
				loops => { type=>HASHREF, optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $adapter = $opts{adapter};
    my $timeout = $opts{timeout};
    my $loops = $opts{loops} || get_storage_loops(conn => $conn, timeout => $timeout);

    my $adapter_loops = {};

    if (! fabric_is_enabled(conn => $conn, timeout => $timeout)){
	$adapter_loops->{$adapter} = $loops->{$adapter};
    } else {
	#code below applies only for metro cluster

	my $environ_shelf_adapter = $opts{environ_shelf_adapter} || 
        		get_environ_shelf_adapter(conn => $conn, adapter => $adapter, 
					timeout => $timeout);

	chomp( my @shelf_lines = split(/\r?\n/,"$environ_shelf_adapter"));

	foreach my $shelf_line (@shelf_lines) {
            if( $shelf_line =~ /SES\s+device\s+path:\s+local\s+access:\s+([\w:]+)\.\d+$/ ){
        	$adapter_loops->{$1} = $loops->{$1};
        	print "DEBUG: get_adapter_loops found an adapter channel: $1\n" if $debug;
            }
	}
    }

    print "DEBUG: get_adapter_loops is returning:\n" if $debug;
    print Data::Dumper->Dump([%$adapter_loops]) if $debug;

    return $adapter_loops;
}


sub get_storage_loops(%){

    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
				fcadmin_device_map=> { type=>SCALAR,  optional=>1 },
				storage_show_disk_path => { type=>SCALAR,  optional=>1 },
				vol_status_r => { type=>SCALAR,  optional=>1 },
				drives => { type=>HASHREF,  optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
 
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};
    my $fcadmin_device_map = $opts{fcadmin_device_map} || get_fcadmin_device_map(conn => $conn, timeout => $timeout);
    my $storage_show_disk_path = $opts{storage_show_disk_path} || get_storage_show_disk_path(conn => $conn, timeout => $timeout);
    my $vol_status_r = $opts{vol_status_r} || get_vol_status_r(conn => $conn, timeout => $timeout);
    chomp( my @device_map = split(/\r?\n/,"$fcadmin_device_map"));
    my $drives = $opts{drives} || get_storage_drives(conn => $conn, 
                                                storage_show_disk_path => $storage_show_disk_path,
                                                fcadmin_device_map => $fcadmin_device_map,
                                                vol_status_r => $vol_status_r,
						timeout => $timeout);

    my $loops = {};

    my $channel;
    my $shelves;
    my $shelf_id;
    my $found_shelves = 0;

    foreach my $shelf_line (@device_map) {
 
        if( $shelf_line =~ /Loop\s+Map\s+for\s+channel\s+([\w:]+):\s*$/ ){
            $channel = $1;
        }elsif ($shelf_line =~ /Shelf\s+mapping:/ ){
	    $found_shelves = 1;
        }elsif ($shelf_line =~ /Target\s+SES\s+devices/ ){
	    $found_shelves = 0;
        }elsif (($shelf_line =~ /Shelf\s+(\d+):\s+(.*)$/ ) && $found_shelves){
	    $shelf_id = $1;
	    my $drive_string = $2;
    	    print "DEBUG: get_storage_loops found a shelf: $channel.$shelf_id\n" if $debug;
	    print "DEBUG: get_storage_loops found a list of drives for shelf $shelf_id: $drive_string\n" if $debug;
	    my @drive_ids = split(/\s+/, $drive_string);
	    @drive_ids = grep(/\d+/, @drive_ids); # drive_ids can also be XX & BYP

	    if (! defined $loops->{$channel}){
		$loops->{$channel} = {};
		$loops->{$channel}->{SHELVES} = {};
		$shelves = $loops->{$channel}->{SHELVES};
    	        print "DEBUG: get_storage_loops found a new channel: $channel\n" if $debug;
	    }
            $shelves->{$shelf_id}->{DRIVES} = {};
	    my $shelf_drives = $shelves->{$shelf_id}->{DRIVES};

	    foreach my $drive_id (@drive_ids){
    		$drive_id = "$channel." . $drive_id;
		if ($drives->{$drive_id}){ # the shelf map includes dead drives
		    $shelf_drives->{$drive_id}= "";
		}
	    }
        }
    }
   
    print "DEBUG: get_storage_loops is returning:\n" if $debug;
    print Data::Dumper->Dump([%$loops]) if $debug;
    return $loops;
}


sub get_storage_adapters(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
				storage_show_adapter => { type=>SCALAR,  optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};
    my $storage_show_adapter = $opts{storage_show_adapter} || get_storage_show_adapter(conn => $conn, timeout => $timeout);
    my @storage_show_adapter  = split (/\r?\n/, "$storage_show_adapter");

    my $adapters = {};

    my $adapter;
    my $link_rate;
    my $enabled;
    my $in_use;
    foreach my $adapter_line (@storage_show_adapter){
	if ($adapter_line =~ /^Slot:\s+(\w+)\s*$/){
	    $adapter = $1;
	    $enabled = 0;
            $in_use = 0;
	} elsif ($adapter_line =~ /Link Data Rate:\s+(\d+)\s+Gbit/){
	    $link_rate = $1;
	} elsif ($adapter_line =~ /State:\s+Enabled/){
	    $enabled = 1;
	} elsif ($adapter_line =~ /In Use:\s+Yes/){
	    $in_use = 1;
	} elsif ($adapter_line =~ /Redundant:\s+(\w+)$/){
	    my $redundant = ($1 eq "Yes");
	    if ($enabled && $in_use) {
		$adapters->{$adapter} = {};
		$adapters->{$adapter}->{LINK_RATE} = $link_rate;
		$adapters->{$adapter}->{PATH} = $redundant? "dual":"single";
	    }
	}
    }
    return $adapters;
}


sub get_storage_drives(@){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
				storage_show_disk_path => { type=>SCALAR,  optional=>1 },
				fcadmin_device_map => { type=>SCALAR,  optional=>1 },
				vol_status_r => { type=>SCALAR,  optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};
    my $storage_show_disk_path = $opts{storage_show_disk_path} || 
					get_storage_show_disk_path(conn => $conn, timeout => $timeout);
    my $fcadmin_device_map = $opts{fcadmin_device_map} ||
                                        get_fcadmin_device_map(conn => $conn, timeout => $timeout);
    my $vol_status_r = $opts{vol_status_r} ||
                                        get_vol_status_r(conn => $conn, timeout => $timeout);

    chomp(my @disk_paths = split (/\r?\n/, "$storage_show_disk_path"));
    chomp(my @vol_status_r_lines = split (/\r?\n/, "$vol_status_r"));
    chomp(my @fcadmin_device_map_lines = split (/\r?\n/, "$fcadmin_device_map"));

    my $drives = {};
    my $loop_config = "";

    foreach my $disk_path (@disk_paths){
	# there are several rules that determine how this output is parsed: 
	#     in a single pathed direct-attached storage cluster, A and B indicates ownership.
	#     in a dual pathed direct-attached single head storage config, A and B indicate primary/secondary path
	#     in metro cluster, A and B indicate primary & secondary, but ownership is not distinguished,
	#         other means need to be used to determine ownership.
	# this is all bogus, needs to be rewritten.  How do you determine primary/secondary?
 
        if ($disk_path =~ /^([\w:\.]+)\s+([\w]+)\s+([\w:\.]+)\s+([\w])+\s+([\d]+)\s+([\d]+)/){

            if ("$loop_config" eq "dl-single"){
                logresult ("WARN", "Filer has a mix of dual and single pathed disks:\n$storage_show_disk_path\n");
            } else {
                $loop_config = "dl-dual";
            }

	    $drives ->{$1} = {};
	    $drives ->{$1}->{PATH} = "dual";
	    $drives ->{$1}->{SHELF} = $5;
	    $drives ->{$1}->{BAY} = $6;
	    if ($2 eq "A"){
		$drives ->{$1}->{PRIMARY} = $1;
		$drives ->{$1}->{SECONDARY} = $3;
	    } elsif ($4 eq "A"){
		$drives ->{$1}->{PRIMARY} = $3;
		$drives ->{$1}->{SECONDARY} = $1;
	    }
	    $drives ->{$3} = $drives ->{$1};
        } elsif ($disk_path =~ /^([\w:\.]+)\s+([\w]+)\s+([\d]+)\s+([\d]+)/){

            if ("$loop_config" eq "dl-dual"){
                logresult ("WARN", "filer has a mix of dual and single pathed drives:\n$storage_show_disk_path\n");
            } else {
                $loop_config = "dl-single";
            }

	    $drives ->{$1} = {};
	    $drives ->{$1}->{PATH} = "single";
	    $drives ->{$1}->{SHELF} = $3;
	    $drives ->{$1}->{BAY} = $4;
	    $drives ->{$1}->{PRIMARY} = $1;
        } 
    }

    my $missing_drives = {};

    foreach my $vol_status_r_line (@vol_status_r_lines){
        if ($vol_status_r_line =~ /^\s*(partner|spare|data|parity|dparity|init failed|failed)\s+([\w\.\:]+)\s+.*(SATA|ATA|FCAL)\s+([\d]+)\s+/){
	    if ($drives ->{$2}){
                if ($1 eq 'partner'){
		    $drives ->{$2}->{OWNER} = $1;
		} else {
			$drives ->{$2}->{OWNER} = "filer";
		}
		$drives ->{$2}->{RAID_TYPE} = $1;
		$drives ->{$2}->{RPM} = $4;
		$drives ->{$2}->{DRIVE_TYPE} = $3;
	    } else {
		if (! $missing_drives -> {vol_status_r}){
		    $missing_drives -> {vol_status_r} = [];
		}
		push (@{$missing_drives -> {vol_status_r}}, $2);
	    }
	}
    }

    my ($channel,$found_shelves, $shelf_id);
    foreach my $fcadmin_device_map_line (@fcadmin_device_map_lines){
        if( $fcadmin_device_map_line =~ /Loop\s+Map\s+for\s+channel\s+([\w:]+):\s*$/ ){
            $channel = $1;
        }elsif ($fcadmin_device_map_line =~ /Shelf\s+mapping:/ ){
            $found_shelves = 1;
        }elsif ($fcadmin_device_map_line =~ /Target\s+SES\s+devices/ ){
            $found_shelves = 0;
        }elsif (($fcadmin_device_map_line =~ /Shelf\s+(\d+):\s+(.*)$/ ) && $found_shelves){
            $shelf_id = $1;
            my $drive_string = $2;
            print "DEBUG: get_storage_loops found a shelf: $channel.$shelf_id\n" if $debug;
            print "DEBUG: get_storage_loops found a list of drives for shelf $shelf_id: $drive_string\n" if $debug;
            my @drive_ids = split(/\s+/, $drive_string);
            @drive_ids = grep(/\d+/, @drive_ids); # drive_ids can also be XX & BYP

            foreach my $drive_id (@drive_ids){
                $drive_id = "$channel." . $drive_id;
                if (! $drives->{$drive_id}){ # the shelf map includes dead drives
                   if (! $missing_drives -> {fcadmin_device_map}){
                       $missing_drives -> {fcadmin_device_map} = [];
                }
                push (@{$missing_drives -> {fcadmin_device_map}}, $drive_id);
 
                }
            }
        }
    }

    foreach my $drive (keys %$drives){
	if (! $drives ->{$drive}->{RAID_TYPE}){
            if (! $missing_drives -> {show_disk_p}){
                $missing_drives -> {show_disk_p} = [];
            }
            push (@{$missing_drives -> {show_disk_p}}, $drive);
	    delete $drives ->{$drive};
	}
    }

    my $RESULT;
    if ($debug){
	$RESULT = 'WARN';
    } else {
	$RESULT = 'CONFIG';
    }

    if ($missing_drives -> {fcadmin_device_map}){
	logresult ($RESULT, "Drive listed by fcadmin device_map is not listed by show disk -p: @{$missing_drives -> {fcadmin_device_map}}\n");
    }
    if ($missing_drives -> {vol_status_r}){
	logresult ($RESULT, "Drive listed by vol status -r is not listed by show disk -p: @{$missing_drives -> {vol_status_r}}\n");
    }
    if ($missing_drives -> {show_disk_p}){
	logresult ($RESULT, "Drive listed by show disk -p is not listed by vol status -r: @{$missing_drives -> {show_disk_p}}\n");
    }

    return $drives;
}


sub loop_is_accessible (@) {
    my (%opts) = validate(@_, { loop => { type=>SCALAR },
    				adapters => { type=>HASHREF }});
    my $loop = $opts{loop};
    my $adapters = $opts{adapters};

    foreach my $adapter (keys %$adapters){
	if ($adapters->{$adapter}->{LOOPS}->{$loop}){
	    return 1;
	}
    }
    return 0;
}


#----------------------------------------------------------------------
# The exported routines are coded in the following section
#----------------------------------------------------------------------

=head2 get_storconfig

This routine returns a pointer to a hash containing the storage objects 

my $host = host($FILER);
my $conn = make_conn($host);
my %stor_hash = get_storconfig(conn => $conn);

=over

=item conn

(Required) : connspec of filer

#=item SKIP_ROOT_VOL
#
#(Optional) : Boolean.  1 means do not include root volume drives in the storhash.  Default: 1.
#
=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub get_storconfig(@) {
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                                timeout => { type=>SCALAR, default=>120, optional=>1 }});
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};

    my $storage_show_disk_path = get_storage_show_disk_path(conn => $conn, timeout => $timeout);
    my @storage_show_disk_path_lines = split (/\n/, $storage_show_disk_path);
    my $storage_show_adapter = get_storage_show_adapter(conn => $conn, timeout => $timeout);
    my $fcadmin_device_map = get_fcadmin_device_map(conn => $conn, timeout => $timeout);
    my $vol_status_r = get_vol_status_r(conn => $conn, timeout => $timeout);

    my $stor_act_config = {};
    $stor_act_config->{DRIVES} = get_storage_drives(conn => $conn, 
						storage_show_disk_path => $storage_show_disk_path, 
						fcadmin_device_map => $fcadmin_device_map, 
						vol_status_r => $vol_status_r, 
						timeout => $timeout);
    $stor_act_config->{LOOPS} = get_storage_loops(conn => $conn, 
						fcadmin_device_map => $fcadmin_device_map, 
                                                storage_show_disk_path => $storage_show_disk_path,
                                                vol_status_r => $vol_status_r,
						drives => $stor_act_config->{DRIVES}, 
						timeout => $timeout);
    $stor_act_config->{ADAPTERS} = get_storage_adapters(conn => $conn, storage_show_adapter => $storage_show_adapter);
    my $loops = $stor_act_config->{LOOPS};
    my $adapters = $stor_act_config->{ADAPTERS};
    my $drives = $stor_act_config->{DRIVES};

    print "DEBUG: stor_act_config in get_storconfig:\n" if $debug;
    print Data::Dumper->Dump([%$stor_act_config]) if $debug;

    foreach my $adapter (keys %$adapters){
	my $environ_shelf_adapter = get_environ_shelf_adapter(conn => $conn, adapter => $adapter, timeout => $timeout);
	my $adapter_sysconfig = get_adapter_sysconfig(conn => $conn, adapter => $adapter, timeout => $timeout);
 	$adapters->{$adapter}->{LOOPS} = get_adapter_loops(conn => $conn, adapter => $adapter,
						environ_shelf_adapter => $environ_shelf_adapter, 
						loops => $loops, timeout => $timeout);
	$adapters->{$adapter}->{ADAPTER_CONFIG} = get_adapter_config($adapter_sysconfig);
	$adapters->{$adapter}->{ADAPTER_CONFIG} = add_inferred_configs($adapters->{$adapter}->{ADAPTER_CONFIG});
	my $link_rate = $adapters->{$adapter}->{LINK_RATE};

	foreach my $channel (keys %{$adapters->{$adapter}->{LOOPS}}){
	    if (! (defined $loops ->{$channel}->{LOOP_CONFIG})){
		$loops ->{$channel}->{LOOP_CONFIG} = get_loop_config(@storage_show_disk_path_lines);
		$loops->{$channel}->{LOOP_CONFIG} = add_inferred_configs($loops->{$channel}->{LOOP_CONFIG});

		my $shelves = $loops->{$channel}->{SHELVES};
		foreach my $shelf_id (keys %$shelves){
		    my @environ_shelf = get_environ_shelf_adapter_shelf(conn => $conn, adapter => $adapter, shelf_id => $shelf_id, 
						environ_shelf_adapter => $environ_shelf_adapter, timeout => $timeout);
		    $shelves->{$shelf_id}->{SHELF_CONFIG} = get_shelf_config($shelf_id, $link_rate, @environ_shelf);
		    $shelves->{$shelf_id}->{SHELF_CONFIG} = add_inferred_configs($shelves->{$shelf_id}->{SHELF_CONFIG});

		    my $shelf_drives = $shelves->{$shelf_id}->{DRIVES};
		    foreach my $drive_id (keys %$shelf_drives){
			$shelf_drives->{$drive_id}= $drives->{$drive_id};
			my $drive_id_match;
			if ($shelf_drives->{$drive_id}->{SECONDARY}){
			    $drive_id_match = "$shelf_drives->{$drive_id}->{PRIMARY}|$shelf_drives->{$drive_id}->{SECONDARY}";
			} else { $drive_id_match = $drive_id; }
			$shelf_drives->{$drive_id}->{DISK_CONFIG} = 
				get_drive_config(get_drive_sysconfig(conn => $conn, adapter => $adapter,
							adapter_sysconfig => $adapter_sysconfig,
							drive_id => $drive_id_match, timeout => $timeout));
	        	$drives->{$drive_id}->{DISK_CONFIG} = add_inferred_configs($drives->{$drive_id}->{DISK_CONFIG});
		    }
		}
	    }
	}
    }

    return $stor_act_config;
}

=head2 get_storconfig_drives

This routine returns a list of drives present in a storconfig hash
If the storconfig hash is not passed in as argument stor_act_config, the
default is to return a list of all drives.

my $host = host($FILER);
my $ret = get_storconfig_drives(conn=>$conn);

The following named arguments are accepted,

=over

=item conn 

(Required) : Connspec to filer

=item stor_act_config

(Required) : StorConfig hash

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub get_storconfig_drives{
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
			stor_act_config => { type=>HASHREF, optional=>1 },
			timeout => { type=>SCALAR, default=>100, optional=>1 },
                 });
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};
    my $stor_act_config = $opts{stor_act_config} || get_storconfig (conn => $conn,
                                				timeout => $timeout);
    return $stor_act_config -> {DRIVES};
}

=head2 get_storconfig_drives_uut

This routine returns a list of drives under test

my $host = host($FILER);
my $ret = get_storconfig_drives(conn=>$conn);

The following named arguments are accepted,

=over

=item conn 

(Required) : Connspec to filer

=item stor_plan_config

(Optional) : hash with config types to test
These will get applied *in addition to* TESTPLAN_CONFIGS
Hash format is {DRIVE => "list of drive config types", SHELF "list of shelf config types", 
LOOP "list of loop config types", ADAPTER "list of adapter config types"}.
If any list of config types is empty, all category configs match.

#=item SKIP_ROOT_VOL
#
#(Optional) : Boolean.  1 means do not include root volume drives in the list.  Default: 1.
#
=item SPARES_ONLY =>1, 

(Optional) : boolean. Default is 0, which will return all disks.  If 1, only spares will be returned.

=item SAME_POOL 

(Optional) : boolean.  Default is 0, which will return all disks.  If 1, only disks in a single pool will be returned.

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub get_storconfig_drives_uut{
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
			stor_act_config => { type=>HASHREF, optional=>1 },
			stor_plan_config => { type=>HASHREF, default=>"", optional=>1 },
			timeout => { type=>SCALAR, default=>1200, optional=>1 },
			RAID_TYPE => { type=>SCALAR, default=>'spare', optional=>1 },
			OWNER => { type=>SCALAR, default=>'filer', optional=>1 },
                 });

    print "entering get_storconfig_drives_uut\n" if $debug;
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};
    my $stor_act_config = $opts{stor_act_config} || get_storconfig_uut (conn => $conn,  timeout => $timeout);
    my $stor_plan_config = $opts{stor_plan_config};

    delete $opts{conn};
    delete $opts{timeout};
    delete $opts{stor_act_config};
    delete $opts{stor_plan_config};

    # now %opts contains only drive attributes and will be used to select matching drives

    if ($stor_plan_config){
        print "calling get_storconfig_uut from get_storconfig_drives_uut\n" if $debug;
        $stor_act_config = get_storconfig_uut (conn => $conn, 
                                stor_plan_config => $stor_plan_config,
                                stor_act_config => $stor_act_config,
				timeout => $timeout);

	# This call to get_storconfig_uut occurs if an additional config hash is provided.  It is just provided
	# so that the "uut" functions have a uniform interface signature.  Normally the storconfig being passed
	# in will already have been pruned as needed, and the purpose of get_storconfig_drives_uut is just to select the
	# drives with matching drive attributes.
    }

    print "calling get_storconfig_drives from get_storconfig_drives_uut\n" if $debug;
    my $drives = get_storconfig_drives(conn => $conn, stor_act_config => $stor_act_config, timeout => $timeout);
    use Data::Dumper;
    print Data::Dumper->Dump([%$drives]) if $debug;

    foreach my $drive (keys %$drives){
	print "comparing args for $drive\n" if $debug;
        if (! compare_args (hash => $$drives{$drive}, patterns => \%opts)) {
	    print "deleting drive $drive\n" if $debug;
            delete $$drives{$drive};
        }else{
	    print "not deleting drive $drive\n" if $debug;
	}
    }

    print Data::Dumper->Dump([%$drives]) if $debug;
    return $drives;
}

=head2 compare_storconfig_status

This routine returns success (1), or it croaks if any device is new or missing or has different attributes

my $host = host($FILER);
my $prev_stor_hash = read_storconfig(conn => $conn);
my $stor_hash = get_storconfig(conn => $conn);
my $ret = compare_storconfig_status(prev_storconfig =>$prev_stor_hash, current_storconfig => $stor_hash);

The following named arguments are accepted,

=over

=item conn

(Required) : Connspec to filer

=item prev_storconfig

(Required) : stor_hash

=item current_storconfig

(Required) : stor_hash

=back

=cut

sub compare_storconfig_status{
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
				prev_storconfig => { type=>SCALAR },
				current_storconfig => { type=>SCALAR },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $prev_storconfig = $opts{prev_storconfig};
    my $current_storconfig = $opts{current_storconfig};
    my $timeout = $opts{timeout};

    my $status = 1;
    # do stuff
    return($status);
}

=head2 write_storconfig

This routine writes a stor_hash to a file and returns 1 if it was successful

my $host = host($FILER);
my $stor_hash = get_storconfig(conn => $conn);
my $ret = write_storconfig(host=>$host, stor_hash = $stor_hash);

=over

=item conn

(Required) : connspec to filer

=item stor_hash

(Required) : Storage config hash

=back

=cut

sub write_storconfig{
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
                     		stor_hash  => { type=>SCALAR },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $stor_hash = $opts{stor_hash};
    my $timeout = $opts{timeout};

    # do stuff
    return 1;
}

=head2 read_storconfig

This routine reads a stor_hash from a file and returns 1 if it was successful

my $host = host($FILER);
my $prev_storconfig = read_storconfig(conn=>$conn);

=over

=item conn

(Required) : Hostspec to filer

=back

=cut

sub read_storconfig{
    my (%opts) = validate(@_, {  conn => { type=>OBJECT, isa=>'Connectrec' } ,
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};

    my $stor_hash = {};
    # do stuff
    return $stor_hash;
}

=head2 get_storconfig_uut

This routine returns a pointer to a hash containing the storage objects 

my $host = host($FILER);
my $conn = make_conn($host);
my %stor_hash = get_storconfig_uut(conn => $conn);

=over

=item conn

(Required) : connspec of filer

#=item SKIP_ROOT_VOL
#
#(Optional) : Boolean.  1 means do not include root volume drives in the storhash.  Default: 1.
#
=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub get_storconfig_uut (%){
    my (%opts) = validate(@_, { conn => { type=>OBJECT, isa=>'Connectrec' },
				stor_plan_config => { type=>HASHREF, optional=>1 },
				stor_act_config => { type=>HASHREF, optional=>1 },
                                timeout => { type=>SCALAR, default=>30, optional=>1 }});

    print "entering get_storconfig_uut\n" if $debug;
    my $conn = $opts{conn};
    my $timeout = $opts{timeout};
    print "initializing stor_plan_config in get_storconfig_uut\n" if $debug;
    my $stor_plan_config = $opts{stor_plan_config} || _get_storage_testplan_config_types();
    print "initializing stor_act_config in get_storconfig_uut\n" if $debug;
    my $stor_act_config = $opts{stor_act_config} || get_storconfig(conn => $conn);

    print "DEBUG: stor_act_config in get_storconfig_uut:\n" if $debug;
    print Data::Dumper->Dump([%$stor_act_config]) if $debug;

    print "DEBUG: stor_plan_config in get_storconfig_uut:\n" if $debug;
    print Data::Dumper->Dump([%$stor_plan_config]) if $debug;

    my $adapters = $stor_act_config->{ADAPTERS};
    my $loops = $stor_act_config->{LOOPS};
    my $drives = $stor_act_config->{DRIVES};

    foreach my $adapter (keys %$adapters) {
	if (! config_match ($adapters->{$adapter}->{ADAPTER_CONFIG}, $stor_plan_config->{ADAPTER})){
	    print "DEBUG: deleting adapter $adapter\n" if $debug;
	    delete $adapters->{$adapter};
	}
    }

    foreach my $drive (keys %$drives){
	if (($drives->{$drive}) && (! config_match ($drives->{$drive}->{DISK_CONFIG}, $stor_plan_config->{DRIVE}))){
	    print "DEBUG: deleting drive $drive due to non-matching drive config\n" if $debug;
	    my $primary = $drives->{$drive}->{PRIMARY};
            delete $drives->{$primary};
	    if ($drives->{$drive}->{PATH} eq 'dual'){
	        my $secondary = $drives->{$drive}->{SECONDARY};
                delete $drives->{$secondary};
	    }
        }
    }

    foreach my $loop (keys %$loops){
	if ((! loop_is_accessible (loop => $loop, adapters => $adapters)) ||
		(! config_match ($loops->{$loop}->{LOOP_CONFIG}, $stor_plan_config->{LOOP}))){
	    my $shelves = $loops->{$loop}->{SHELVES};

            foreach my $shelf (keys %$shelves){
		my $shelf_drives = $shelves->{$shelf}->{DRIVES};

                foreach my $drive (keys %$shelf_drives) {
		    if ($drives->{$drive}){
	    		print "DEBUG: deleting $drive on non-accessible loop\n" if $debug;
		        my $primary = $drives->{$drive}->{PRIMARY};
			if ($drives->{$drive}->{PATH} eq 'dual'){
		            my $secondary = $drives->{$drive}->{SECONDARY};
        	            delete $drives->{$secondary};
			}
       		        delete $drives->{$primary};
		    }
	    	    print "DEBUG: deleting shelf_drive $drive on non-accessible loop\n" if $debug;
                    delete $shelf_drives->{$drive};
                }
	    	print "DEBUG: deleting shelf $shelf on non-accessible loop\n" if $debug;
                delete $shelves->{$shelf};
            }
	    print "DEBUG: deleting non-accessible loop $loop\n" if $debug;
	    delete $loops->{$loop};
            foreach my $adapter (keys %$adapters){
                print "DEBUG: deleting adapter loop $loop on non-accessible loop\n" if $debug;
                delete $adapters->{$adapter}->{LOOPS}->{$loop};
                if (! (keys %{$adapters->{$adapter}->{LOOPS}})){
                    print "DEBUG: deleting adapter $adapter on non-accessible loop\n" if $debug;
                    delete $adapters->{$adapter};
                }
            }

	} else  { # (config_match ($loops->{$loop}->{LOOP_CONFIG}, $stor_plan_config->{LOOP}))
	    my $shelves = $loops->{$loop}->{SHELVES};

	    foreach my $shelf (keys %$shelves){
		if (config_match ($shelves->{$shelf}->{SHELF_CONFIG}, $stor_plan_config->{SHELF})){
		    my $shelf_drives = $shelves->{$shelf}->{DRIVES};

		    foreach my $drive (keys %$shelf_drives) {
			if (! $drives->{$drive}){
	    		    print "DEBUG: deleting shelf_drive $drive due to missing drive key\n" if $debug;
			    delete $shelf_drives->{$drive};
			}
		    }
		    if (! keys %$shelf_drives){
	    		print "DEBUG: deleting shelf $shelf since no more drives remain\n" if $debug;
			delete $shelves->{$shelf};
		    }
		} else {
		    my $shelf_drives = $shelves->{$shelf}->{DRIVES};

		    foreach my $drive (keys %$shelf_drives) {
			if ($drives->{$drive}){
	    		    print "DEBUG: deleting drive $drive due to non-matching shelf config\n" if $debug;
		            my $primary = $drives->{$drive}->{PRIMARY};
			    if ($drives->{$drive}->{PATH} eq 'dual'){
		                my $secondary = $drives->{$drive}->{SECONDARY};
        	        	delete $drives->{$secondary};
			    }
       		            delete $drives->{$primary};
			}
	    		print "DEBUG: deleting shelf_drive $drive due to non-matching shelf config\n" if $debug;
			delete $shelf_drives->{$drive};
		    }
	    	    print "DEBUG: deleting shelf $shelf due to non-matching shelf config\n" if $debug;
		    delete $shelves->{$shelf};
		}
	    }
	    if (! keys %$shelves){
	    	print "DEBUG: deleting loop $loop since no more shelves remain\n" if $debug;
		delete $loops->{$loop};
		foreach my $adapter (keys %$adapters){
	    	    print "DEBUG: deleting adapter loop $loop since no more shelves remain\n" if $debug;
		    delete $adapters->{$adapter}->{LOOPS}->{$loop};
		    if (! (keys %{$adapters->{$adapter}->{LOOPS}})){
	    	        print "DEBUG: deleting adapter $adapter since no more loops remain\n" if $debug;
			delete $adapters->{$adapter};
		    }
		}
	    }
	} 
    }

    if (! keys %{$stor_act_config->{LOOPS}}){	# the point of this is that if we are calling this function we are
					# running a test, so rather than abort since we couldn't figure
					# out what to test, lets test everything
	if ((keys %{$stor_act_config->{ADAPTERS}}) || (keys %{$stor_act_config->{DRIVES}})){
	    print "DEBUG: stor_act_config has inconsistent cruft remaining:\n";
	    print Data::Dumper->Dump([%$stor_act_config]) if $debug;
	}

	print "DEBUG: returning full config\n" if $debug;
	$stor_act_config = get_storconfig(conn => $conn);
    }
    print "DEBUG: stor_act_config returned by get_storconfig_uut:\n" if $debug;
    print Data::Dumper->Dump([%$stor_act_config]) if $debug;
    return $stor_act_config;
}

1;
