# this is basically -*-perl-*-
# $Id: //depot/prod/test/Rsierranevada/lib/Storage/UiUtils.pm#1 $
#
## @summary	UiUtils.pm - Utility routines for storage UI tests
## @author	dl-storsub-qa@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod		here

=head1 NAME

UiUtils - Utility routines for storage UI tests

=head1 SYNOPSIS

    General purpose routines for storage UI tests.

    # Get the disk state
    $disk1 = get_disk_state(server => $Server, timeout => 600);

    ... Do some testing

    # Get the disk state again
    $disk2 = get_disk_state(server => $Server, timeout => 600);

    # Compare the disk state before and after testing to see if any
    # of the disks have changed state.
    $err = compare_disk_state(disks1 => $disks1, disks2 => $disks2);

    # Call 'event log show' on the filer to get the log messages
    $matches = search_filer_log(server      => $Server,
                                messagename => 'dfu.firmwareUpToDate',
                                severity    => 'INFORMATIONAL',
                                timeout     => $TIMEOUT);

    # Verify that all the disks specified by the disks parameter
    # were successfully upgraded.
    $matches = vrfy_disk_upgraded(server  => $Server,
                                  disks   => \@Disks,
                                  timeout => 0);

    # Verify that all the shelves specified by the shelves parameter
    # were successfully upgraded.
    $matches = vrfy_shelf_upgraded(nodes   => \@Nodes_with_shelves,
                                   server  => $Server,
                                   shelves => \%Shelves,
                                   timeout => $TIMEOUT);

=head1 DESCRIPTION

Library routines used in Storage UI tests.

=head1 STATUS

  Private

=head1 AUTHOR

  dl-storsub-qa@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/Storage/UiUtils.pm#1 $
  Last revised $Date: 2013/01/22 $

head1 COPYRIGHT

  Copyright 2008 Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Storage::UiUtils;

use strict;
use warnings;

use Date::Manip; # Perl Package
use base qw(Exporter);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF);
use Server;
use Carp qw(confess croak);

print "\nUsing this module from ". __FILE__ ."\n";

our @EXPORT_OK = qw(
    compare_disk_state
    get_disk_state
    search_filer_log
    vrfy_disk_upgraded
    vrfy_shelf_upgraded
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#---------------------------------------------------------------------
# Function prototypes
#----------------------------------------------------------------------------
sub compare_disk_state(%;);
sub get_disk_state(%;);
sub _filter_matches_by_time($$;);
sub _convert_log_to_str($;);
sub search_filer_log(%;);
sub vrfy_disk_upgraded(%;);
sub vrfy_shelf_upgraded(%;);

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------

#---------------------------------------------------------------------

=head2 compare_disk_state

    compare_disk_state(disks1 => $disks1, disks2 => $disks2);

Verify that the disk state provided in the first array matches
the disk state provided in the second array.

=over

=item Arguments

=over

=item disk1

(Required) : Reference to a hash of disk state returned from
get_disk_state() subroutine.

=item disk2

(Required) : Reference to a hash of disk state returned from
get_disk_state() subroutine.

=back

=item Returns

Returns any matches or an empty string if none were found.
Returns an empty string on success. On failure, returns a list of
error messages indicating which disks had a state change, or list
of disks that are new or missing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub compare_disk_state(%;) {
    my (%opts) = validate (
        @_,
        {
            disks1 => { type => HASHREF },
            disks2 => { type => HASHREF },
        }
    );
    my ($d1, $disks1, $disks2, $err);

    $err = '';
    $disks1 = $opts{disks1};
    $disks2 = $opts{disks2};
    foreach $d1 (keys %{$disks1}) {
        if (!defined($disks2->{$d1})) {
            $err .= "\nDisk state change disk: $disks1->{$d1}->{disk} sn: $d1".
                    "\n\tExpected state: $disks1->{$d1}->{state}\n" .
                    "\tActual state: not reported\n";
            next;
        }
        if ($disks1->{$d1}->{state} ne $disks2->{$d1}->{state}) {
            $err .= "\nDisk state change disk: $disks1->{$d1}->{disk} sn: $d1".
                    "\n\tExpected state: $disks1->{$d1}->{state}\n" .
                    "\tActual state: $disks2->{$d1}->{state}\n";
        }
    }
    return ($err);
}

#---------------------------------------------------------------------

=head2 get_disk_state

    $disks1 = get_disk_state(server => $Server, timeout => 600);

Get the disk state for the filer.

=over

=item Arguments

=over

=item server

(Required) : Server.pm object.

=item timeout

(Required) : Maximum number of seconds to wait for "storage disk show"
command to return data. The call could return sooner.

=back

=item Returns

Returns a reference to a hash of disks keyed by the serial number and value
is a reference to a hash of the disk information returned from
'storage disk show'

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub get_disk_state(%;) {
    my (%opts) = validate (
        @_,
        {
            server  => { type => OBJECT },
            timeout => { type => SCALAR, regex => qr/\d+/ },
        }
    );
    my ($disk, @disks, %ret);

    @disks = $opts{server}->storage_disk_show(state    => '!partner',
                                              -Timeout => $opts{timeout});
    foreach $disk (@disks) {
        $ret{$disk->{serialnumber}} = $disk;
    }
    return (\%ret);
}

#---------------------------------------------------------------------
# Internal Subroutine
# Remove any messages that occurred before start time from the $matches
# array. If the start time has is undefined or the matches array is
# empty then just return.
#---------------------------------------------------------------------
sub _filter_matches_by_time($$;) {
    my ($st, $matches) = @_;
    my ($date1, $date2, $entry, $flag, $idx);

    if (!defined($st) || scalar(@{$matches}) == 0) {
        # Return as there is nothing to do
        return;
    }

    # The log entries will be sorted with most recent first
    # Once we hit the point where the flag is greater than zero
    # which means the current log entry ($date2) is before the
    # starting date ($date1) of the test run, then all the
    # rest of the entries on the logs are no longer needed.
    $idx = 0;
    foreach $entry (@{$matches}) {
        $entry->{time} =~ s/"//g;
        $date2 = ParseDate($entry->{time});
        $date1 = $st->{$entry->{node}}{date};
        $flag = Date_Cmp($date1, $date2);
        if ($flag > 0) {
            splice @{$matches}, $idx;
            last;
        }
        $idx++;
    }
}

#---------------------------------------------------------------------
# Internal Subroutine
# Convert an array of log entry hashes into strings with one log
# entry per line.
#---------------------------------------------------------------------
sub _convert_log_to_str($;) {
    my ($matches) = @_;
    my ($entry, $str);

    $str = '';
    foreach $entry (@{$matches}) {
        $str .= "$entry->{node} $entry->{time} $entry->{severity} " .
                "$entry->{source} $entry->{event}\n";
    }
    return ($str);
}

#---------------------------------------------------------------------

=head2 search_filer_log

    $matches = search_filer_log(server      => $Server,
                                messagename => 'dfu.firmwareUpToDate',
                                severity    => 'INFORMATIONAL',
                                start_time  => $start_time,
                                timeout     => $TIMEOUT);

Use the "event log show' command to find the server log message.
This routine is needed because the Server.pm cdef for 'event log show'
will fail the command if it returns any errors including 'time is not
in sync' message.

=over

=item Arguments

=over

=item event

(Optional) : Event

=item messagename

(Optional) : Message name

=item node

(Optional) : Return only messages from this node.

=item server

(Required) : Server.pm object.

=item source

(Optional) : Source of the message.

=item start_time

(Optional) : All log entries from this time stamp going forward in
time will be returned. If it is not set, then all log entries will be
returned. This is typically a hash obtained from running the
'system node date' command. The expected format follows:

          {
            'utc-date' => '"2/4/2009 00:24:50"',
            'timezone' => 'GMT',
            'date' => '"2/4/2009 00:24:50"',
            'filer' => 'stor-u30',
            'node' => 'stor-u30'
          },
          {
            'utc-date' => '"2/4/2009 00:24:50"',
            'timezone' => 'GMT',
            'date' => '"2/4/2009 00:24:50"',
            'filer' => 'stor-u31',
            'node' => 'stor-u31'
          }

=item timeout

(Required) : Number of seconds to poll waiting for a message to appear.
Will poll every 15 seconds checking for the regex.
Set to 0 to check just once for a message.

=back


=item Returns

Returns an Hash reference with Event description and Event sequence number.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub search_filer_log(%;) {
    my (%opts) = validate (
        @_,
        {
            event       => { type => SCALAR, optional => 1 },
            messagename => { type => ARRAYREF, optional => 1 },
            node        => { type => SCALAR, optional => 1 },
            server      => { type => OBJECT },
            severity    => { type => SCALAR, optional => 1 },
            source      => { type => SCALAR, optional => 1 },
            start_time  => { type => ARRAYREF, optional => 1 },
            timeout     => { type => SCALAR, regex => qr/\d+/ },
        }
    );
    my ($cmd, $end_time, $entry, @matches, $out, $server, %st, $str,@msgs,$mesg,%evt_hash);
    print "The control is in the routine";

    # Pre-parse the date/time from the start_time if defined
    if (defined($opts{start_time})) {
        foreach $entry (@{$opts{start_time}}) {
            $entry->{date} =~ s/"//g;
            $st{$entry->{node}}{date} = ParseDate($entry->{date});
        }
    }
    if( defined($opts{messagename})){
        foreach $mesg (@{$opts{messagename}}){
            push(@msgs,$mesg);
       }
    }
    # Build the command
    $server = $opts{server};
    $end_time = time() + $opts{timeout};
    do {
        foreach $mesg (@msgs){
	$out="";
	@matches=();
        $cmd="";
	$cmd.="event log show ";
        $cmd .= "-event $opts{event} " if (defined($opts{event}));
        $cmd .= "-node $opts{node} " if (defined($opts{node}));
        $cmd .= "-severity $opts{severity} " if (defined($opts{severity}));
        $cmd .= "-source $opts{source} " if (defined($opts{source}));
        $cmd.="-messagename $mesg";
        #if(!defined($opts{event})){$opts{event}='*';}
        #if(!defined($opts{node})){$opts{node}='*';}
        #if(!defined($opts{severity})){$opts{severity}='*';}
        #if(!defined($opts{source})){$opts{source}='*';}
        #$opts{messagename}=$mesg;
        eval{
	    $server->cli(cmd => $cmd, '-Timeout' => 500, output => \$out);
        };
        if($@=~/There are no entries matching your query/){ $evt_hash{$mesg}=0;}
        else{
	        print "=======================\n";
	        print "$out\n";
	        print "=======================\n";
	        $server->_cli_output_handler(output_in  => $out,
                                     want_array => \@matches);

            if(scalar(@matches) > 0){
		    my @keys = keys %{$matches[0]};
		    print " keys is are @keys\n";
		    print " seq num is $matches[0]->{seqnum}\n";
                 $evt_hash{$mesg}=$matches[0]->{seqnum};
                 print "event seq number for $mesg is $evt_hash{$mesg}\n";
                }
        }
       }
            sleep(30);
    } while ((time() < $end_time));

    return (\%evt_hash);
}

#---------------------------------------------------------------------

=head2 vrfy_disk_upgraded

    $matches = vrfy_disk_upgraded(server  => $Server,
                                  disks   => \@Disks,
                                  timeout => 0);

Verify that the given array of disks has been upgraded.

=over

=item Arguments

=over

=item disks

(Required) : Reference to an array of disk names.

=item server

(Required) : Server.pm object.

=item start_time

(Optional) : All log entries from this time stamp going forward in
time will be returned. If it is not set, then all log entries will be
returned. This is typically a hash obtained from running the
'system node date' command. The expected format follows:

          {
            'utc-date' => '"2/4/2009 00:24:50"',
            'timezone' => 'GMT',
            'date' => '"2/4/2009 00:24:50"',
            'filer' => 'stor-u30',
            'node' => 'stor-u30'
          },
          {
            'utc-date' => '"2/4/2009 00:24:50"',
            'timezone' => 'GMT',
            'date' => '"2/4/2009 00:24:50"',
            'filer' => 'stor-u31',
            'node' => 'stor-u31'
          }

=item timeout

(Required) : Number of seconds to poll waiting for a message to appear.
Will poll every 15 seconds checking for the firmware downloaded message.

=back

=item Returns

Returns an empty string on success. On error returns a list of error
messages associated with the disks and/or a list of disks that did not
report a successful upgrade.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub vrfy_disk_upgraded(%;) {
    my (%opts) = validate (
        @_,
        {
            disks      => { type => ARRAYREF },
            server     => { type => OBJECT },
            drops => {type => SCALAR},
            occurs => { type=>SCALAR},
            start_time => { type => ARRAYREF, optional => 1 },
            timeout    => { type => SCALAR, regex => qr/\d+/ },
        }
    );
    my ($cnt, %disk_list, $dsk, $end_time, @m, $matches, $msg);

    # Build a hash of disk names to look for in the filer log
    $cnt = 0;
    foreach $dsk (@{$opts{disks}}) {
	my $serial=$dsk->{serialnumber};
	if (exists $disk_list{$serial}){ next;}
	else {
        	$disk_list{$serial}{found} = 0;
		    $disk_list{$serial}{name}=$dsk;
	}
        $cnt++;
    }

    # Mark all the disks that have had the firmware updated message
    # reported. Exit end of loop when all found or if timed out.
    $end_time = time() + $opts{timeout};
    do {
        $matches = search_filer_log(server      => $opts{server},
                                    messagename => 'dfu.fwDownloaded',
                                    severity    => 'DEBUG',
                                    start_time  => $opts{start_time},
                                    timeout     => $opts{timeout});
        foreach $dsk (keys %disk_list) {
            # Skip the disks that are already found
            next if ($disk_list{$dsk}{found} != 0);
            # Get all the matches for the disk name
            (@m) = ($matches =~ /$dsk/g);
            foreach $msg (@m) {
                if ($matches =~/^(Firmware\s+downloaded\s+to\s+Disk)(.+) (S\/N [$dsk] successfully)$/){
                    $disk_list{$dsk}{found} = 1;
                    $cnt--;
                    last;
                }
                # Ignore all other messages
            }
        }
        if ($cnt > 0) {
            sleep(15);
        }
    } while (($cnt > 0) && (time() < $end_time));

    # Generate errors for disks not found.
    $matches = '';
    if ($cnt > 0) {
    print(" Number of disks missing from event log show : $cnt\n");
    my $exp_disks = keys %disk_list;
    # Cheking number of drops and occurs after download
    my @event_stats_show = $opts{server}->event_status_show(messagename=>'dfu.fwDownloaded');
    my $a_occurs=($event_stats_show[0]->{'indications'}) + ($event_stats_show[1]->{'indications'});
    my $a_drops=($event_stats_show[0]->{'drops'}) + ($event_stats_show[1]->{'occurs'});
    my $delta_occurs=$a_occurs - $opts{'occurs'};
    my $delta_drops=$a_drops - $opts{'drops'};
    print("Total number of drops after download is : $delta_drops\n"); 
    print("Total number of occurs after download is : $delta_occurs\n"); 
        foreach $dsk (keys %disk_list) {
            # Skip the disks that are already found
            next if ($disk_list{$dsk}{found} != 0);
            $matches .= "Disk serial number ($dsk) and disk name $disk_list{$dsk}{name} firmware upgrade completion message " .
                        "not found in event log on filer.\n";
        }
    }
    
    return ($matches);
}

#---------------------------------------------------------------------

=head2 vrfy_shelf_upgraded

    $matches = vrfy_shelf_upgraded(nodes   => \@Nodes_with_shelves,
                                   server  => $Server,
                                   shelves => \%Shelves,
                                   timeout => $TIMEOUT);

Verify that the given list of shelves has been upgraded.

=over

=item Arguments

=over

=item nodes

(Required) : A reference to an array of nodes in the cluster connected
to a shelf.

=item server

(Required) : Server.pm object.

=item shelves

(Required) : A reference to a hash of the shelves undergoing testing.

=item start_time

(Optional) : All log entries from this time stamp going forward in
time will be returned. If it is not set, then all log entries will be
returned. This is typically a hash obtained from running the
'system node date' command. The expected format follows:

          {
            'utc-date' => '"2/4/2009 00:24:50"',
            'timezone' => 'GMT',
            'date' => '"2/4/2009 00:24:50"',
            'filer' => 'stor-u30',
            'node' => 'stor-u30'
          },
          {
            'utc-date' => '"2/4/2009 00:24:50"',
            'timezone' => 'GMT',
            'date' => '"2/4/2009 00:24:50"',
            'filer' => 'stor-u31',
            'node' => 'stor-u31'
          }

=item timeout

(Required) : Number of seconds to poll waiting for a message in the
filer messages file to appear.  Will poll every 30 seconds checking for
the message.

=back

=item Returns

Returns an empty string on success. On error returns a list of error
messages associated with the shelves and/or a list of shelves that did
not report a successful upgrade.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub vrfy_shelf_upgraded(%;) {
    my (%opts) = validate (
        @_,
        {
            nodes      => { type => ARRAYREF },
            server     => { type => OBJECT },
            shelves    => { type => HASHREF },
            start_time => { type => ARRAYREF, optional => 1 },
            timeout    => { type => SCALAR, regex => qr/\d+/ },
        }
    );
    my ($end_time, $line, @list, $matches, $name, $node, $shelves,
        %shelf_list, $uid);

    $shelves = $opts{shelves};
    $end_time = time() + $opts{timeout};
    do {
        # Get all the successful firmware download messages to shelves
        # Poll every 30 seconds
        sleep(30);
        $matches = search_filer_log(server      => $opts{server},
                                    messagename => 'sfu.downloadSuccess',
                                    severity    => 'INFORMATIONAL',
                                    start_time  => $opts{start_time},
                                    timeout     => 0);

        # matches will be returned with the node name on the first line
        # and all the matches for that node on each subsequent line
        # Setup the %shelf_list hash with a hash of shelves that were updated
        @list = split(/\n/, $matches);
        foreach $line (@list) {
            if ($line =~ /^(.*?)\s\d.*downloaded\son\s(.*)\./) {
                $shelf_list{$1}{$2} = 1;
            }
        }

        # The keys of the shelve hash are the UID. Then only use the ones
        # that match the node name. Search for each shelf for the node and
        # make sure it is found, saving a message if not found.
        $matches = '';
        foreach $uid (keys %{$shelves}) {
            foreach $node (keys %{$shelves->{$uid}}) {
                $name = $shelves->{$uid}{$node}{name};
                if (!defined($shelf_list{$node}{$name})) {
                    $matches .= "Node: $node Shelf ($name) firmware " .
                                "download completion message not found in " .
                                "event log on filer.\n";
                }
            }
        }
    } while ($matches && (time() < $end_time));

    return ($matches);
}

1;
