# this is basically -*-perl-*-
# $Id: //depot/prod/test/Rsierranevada/lib/Storage/Utils.pm#1 $
#
## @summary	Utils.pm - Utility with routines for pisces tests
## @author	dl-storsub-qa@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod		here

=head1 NAME

Utils - Routines for Storage QA tests

=head1 SYNOPSIS

  use Storage::Utils;

  # Common setup routine for pisces automated tests. Pass in a hash
  # 'setup_h' that holds the setup settings and other information
  # that may be important to the test. Pass the same hash into the
  # cleanup routine so we know what needs to be cleaned up.

  my %setup_h;
  stor_setup(
      setup_h        => \%setup_h,
      aggr_prefix    => $AGGR_PREFIX,
      aggr_size      => $AGGR_SIZE,
      aggr_timeout   => $AGGR_TIMEOUT,
      cifs_client    => $CIFS_CLIENT,
      cifs_drive     => $CIFS_DRIVE,
      cifs_passwd    => $CIFS_PASSWD,
      cifs_share     => $CIFS_SHARE,
      cifs_user      => $CIFS_USER,
      cifs_workgroup => $CIFS_WORKGROUP,
      do_wack        => $DO_WACK,
      server         => $Server,
      filer          => $FILER,
      flex_size      => $FLEX_SIZE,
      hammer_bin     => $HAMMER_BIN,
      hammer_exe     => $HAMMER_EXE,
      nfs_client     => $CLIENT,
      client_list    => $CLIENT_LIST,
      raid_type      => $RAID_TYPE,
      vol_prefix     => $VOL_PREFIX,
  );

  # Get the type of filer we are working with.
  $mode = filer_mode(server=>$server);

  # Used in conjunction with lib::LogReader to check for errors in the
  # EMS and /etc/messages log on the filer.

  $aref = check_log_errors(log => $entire_log, type => 'ems');
  if (defined(${$aref}[0])) {
      push @error_msgs, "\nFILER errors from /etc/log/ems:";
      push @error_msgs, @{$aref};
  }

  # Call stor_cleanup from within the test case cleanup routine
  # Cleans up everything stor_setup() created or changed.
  stor_cleanup(
      setup_h => \%setup_h,
      cleanup => $CLEANUP,
      server  => $Server,
  );

  # Use this routine to create an aggregate
  setup_aggr(server    => $Server,
             aggr_name => 'my_aggr',
             aggr_size => 3);

  # Use this routine to create a volume
  setup_vol(server    => $Server,
            aggr_name => 'my_aggr',
            vol_name  => 'my_vol,
            flex_size => '30m');

  # Power cycle a filer that is set up for RLM.
  rlm_power_cycle(server  => $Server,
                  timeout => $CMDTIMEOUT);

  # Check to see if the service is licensed.
  $result = server_is_licensed(server => $Server, service => 'nfs');

  # Install a new license.
  $result = server_license(server => $Server, service => 'nfs');

  # Remove filer license.
  $result = server_unlicense(server => $Server, service => 'nfs');

  # Gets the value for the option given.
  server_options_get(server => $Server, option => $option);

  # Sets the value for the option given.
  server_options_set(server => $Server,
                     option => $option,
                     value  => $value
                   [,verify => $verify ]);

  # To panic the filer and not wait for a prompt.
  server_panic(server => $Server);

  # Reboot a filer via a direct command call (reboot -t 0).
  server_reboot(server => $Server, timeout => $timeout);

  # Reboot a filer to a specified state.
  server_reboot_to(server => $Server, state => $state);

  # Calls transit_to to make sure the filer/node is up.
  server_transit_to(server => $Server);

  # Determines filer's ONTAP kernel is a "debug" build or not.
  $dbg_kernel = server_kernel_is_debug(server => $Server);

  # Determines filer's ONTAP kernel is a "domain crossing" build or not.
  $dom_xing = server_kernel_is_domain_xing(server => $Server);

  # Determines filer's ONTAP version supports $cmd or not.
  $result = server_supports_command(server  => $Server,
                                    command => $cmd,
                                    priv    => $priv);

  # Check to see that there are no configuration errors.
  $result = sysconfig_c(server => $Server);

=head1 DESCRIPTION

Provides various library routines for pisces automated tests.

=head1 STATUS

  Private

=head1 AUTHOR

  dl-platform-qa@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/Storage/Utils.pm#1 $
  Last revised $Date: 2013/01/22 $

head1 COPYRIGHT

  Copyright 2007 Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Storage::Utils;

use strict;
use warnings;

use CIFS::Setup qw(cifs_setup);
use filer qw(filer_service_is_licensed filer_service_license
             filer_service_unlicense filer_qtree_security_set);

use Aggregate qw(aggr_create aggr_exists aggr_online aggr_purge aggr_status);
use Carp qw(croak);
use Hammer qw(cifs_hammer_start cifs_hammer_stop
              nfs_hammer_start nfs_hammer_stop
              ontap_hammer_start ontap_hammer_abort);
use Params::Validate qw(validate OBJECT SCALAR HASHREF ARRAYREF UNDEF);
use TestEnv::MinWin qw(minwin_compatible);
use Tharn;
use Volume qw(vol_create_flex vol_purge vol_status);
use base qw(Exporter);
use boot qw(reboot_to transit_to);
use wack;

our @EXPORT_OK = qw(
    check_fw_dir
    check_log_errors
    set_execution
    setup_aggr
    setup_vol
    server_is_licensed
    server_license
    server_unlicense
    server_options_get
    server_options_set
    server_panic
    server_reboot
    server_reboot_to
    server_transit_to
    server_kernel_is_debug
    server_kernel_is_domain_xing
    server_supports_command
    sysconfig_c
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub check_fw_dir(%;);
sub check_log_errors(%;);
sub filer_mode(%;);
sub set_execution(%;);
sub setup_aggr(%;);
sub setup_vol(%;);
sub server_is_licensed(%;);
sub server_license(%;);
sub server_unlicense(%;);
sub server_options_get(%;);
sub server_options_set(%;);
sub server_panic(%;);
sub server_reboot(%;);
sub server_reboot_to(%;);
sub server_transit_to(%;);
sub server_kernel_is_debug(%;);
sub server_kernel_is_domain_xing(%;);
sub server_supports_command(%;);
sub sysconfig_c(%;);
sub _get_state($;);
sub _get_lic_code($$;);
sub _flex_scale($$$;);

#---------------------------------------------------------------------
# Local variables
#---------------------------------------------------------------------
# The following is a preamble used for all /etc/messages error only
# searches
my $msgs_err = '\[(?:\S+?\@)?(?:\S+?\/)?(?:\S+?\:\s+?)?\S+?\:error]:';

# The following are regex expressions used to search for iomem
# or extCache error messages in /etc/messages on the filer. The
# checks will be performed with the 'i' ignore case flag.
my @msgs_err = (
    'Failed to initialize accelerator card',
    'Accelerator card in slot',
    'WAFL external cache',
    'External cache has been taken offline on error',
);

# The following are event names that will be searched for in
# /etc/log/ems on the filer to look for pisces card errors
# that were logged.
my @ems_events = (
    'raid_disk_unload_done',
);

# Set the privilege level for these tests
my $Priv = 'diag';

# License map
my %Lic_map = (
    "Base"           => "Base",
    "CIFS"           => "CIFS",
    "cifs"           => "CIFS",
    "FlexScale"      => "FlexScale",
    "flex_scale"     => "FlexScale",
    "FlexVol_HPO"    => "FlexVol_HPO",
    "Mirror"         => "Mirror",
    "NFS"            => "NFS",
    "nfs"            => "NFS",
    "SnapRestore"    => "SnapRestore",
    "snaprestore"    => "SnapRestore",
    "Striped_Volume" => "Striped_Volume",
);

my %Server_cache;

# This hash defines the modes that are acceptable input/output
# for the functions below. If a new mode is implemented, it must
# be added to this hash as a key (the value does not matter).
my %modes = (
    '7'  => 1,
    '10' => 1,
);

# This converts the modes hash above to a regex for use in
# parameter validation
my $MODE_QR = '^(' . join("|", keys %modes) . ')$';

#---------------------------------------------------------------------

=head2 check_fw_dir 

    $fw_files = check_fw_dir(type => 'disk_fw'|'shelf_fw');

This function checks for known firmware directory on the filer.

The following named arguments are accepted:

=over

=item Arguments

=over

=item type

(Required) : The type of firmware directory to check. It is either 'disk_fw' or 'shelf_fw'.

=back

=item Returns

1 if directory exists or 0 if directory does not exist 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub check_fw_dir(%;) {
    my (%opts) = validate (
        @_,
        {
            server => { type => OBJECT },
            type   => { type => SCALAR, regex => qr/^(disk_fw|shelf_fw)$/ },
        }
    );
    my ($cmd, $entry, $fc, @matches, $out, $path, $server, $srv, $type);

    $server   = $opts{server};
    $type     = $opts{type};
    $path     = "/etc/$type";
    $srv      = _get_state($server);
    $cmd      = "priv set -q diag; ls $path";   

    if ($srv->{mode} eq "7") {
        $fc = $srv->{fc};
        (undef, undef, $out) = say($fc, $cmd); 
        @matches = split(/\n/,$out);
        foreach $entry (@matches) {
             if ($entry =~ /Can\'t\s+open\s+directory/) {
                 return 0;
             }
        }
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 check_log_errors

    @errors = check_log_errors(log => $log, type => 'ems'|'messages');

This function checks for known iomem or extCache error messages
in the specified log on the filer. It accepts a log reader object
that has already been setup and stopped so that the log messages
can be inspected.

The following named arguments are accepted:

=over

=item Arguments

=over

=item log

(Required) : A log reader object (see lib/LogReader.pm)

=item type

(Required) : The type of log to check. It is either 'ems' or 'messages'.

=back

=item Returns

A reference to an array of error messages relating to iomem or the extCache
or if no error messages were found then an array whose only element is
undefined.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub check_log_errors(%;) {
    my (%opts) = validate (
        @_,
        {
            log  => { type => OBJECT },
            type => { type => SCALAR, regex => qr/^(ems|messages)$/ },
        }
    );
    my ($aref, $log, @matches, $msg, @tmp);

    $log = $opts{log};
    @matches = ();
    if ($opts{type} eq 'ems') {
        #Check the EMS log
        $aref = $log->{'logs'}{'ems'}->find_events(@ems_events);
        if (defined(${$aref}[0])) {
            push @matches, @{$aref};
        }
    } else {
        #Check /etc/messages log - only look at error syslog entries
        foreach $msg (@msgs_err) {
            @tmp = ($log->{'logs'}{'messages'}->to_string =~
                    /($msgs_err\s*?$msg)/gi);
            if (defined($tmp[0])) {
                push @matches, @tmp;
            }
        }
    }

    return (\@matches);
}

#---------------------------------------------------------------------

=head2 filer_mode

  $mode = filer_mode(server => $server);

This routine gets the type of filer we are working with.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=back

=item Returns

Returns the mode of the filer or croaks on failure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub filer_mode(%;) {
    my (%opts) = validate (
        @_,
        {
            server => { type => OBJECT },
        }
    );

    my ($mode, $mode_list, $raw_output, $result);

    $mode_list = join(", ", keys %modes);

    eval {
        $result = $opts{server}->cli(cmd => '?', output => \$raw_output);
    };
    if ($@) {
        croak "Could not send '?' command to filer to test for filer mode: $@";
    }

    if ($raw_output =~ /up\s+Go up one directory/i) {
        $mode = 10;
    } elsif ($raw_output =~ /version/) {
        $mode = 7;
    } else {
        croak "Cannot figure out the filer mode (known types are " .
          "$mode_list):\n$raw_output";
    }

    if ($mode !~ /$MODE_QR/) {
        # If this message appears, it means that the filer_mode
        # function has been changed to allow another type of mode but
        # this mode is not defined in %modes above. The mode
        # value assigned here needs to be added above.
        croak "Filer type (e.g. 7 mode, 10 mode) specification '$mode' " .
          "not defined properly in library. Known modes are $mode_list.";
    }

    return $mode
}

#---------------------------------------------------------------------

=head2 set_execution

    $out = set_execution(server => $Server, pnode => $pnode);

This routine sets the execution node for 10-mode.
(If 7-mode it returns without error.)

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item pnode

(Required) : Name of the node to set execution to.

=back

=item Returns

Nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub set_execution(%;) {
    my (%opts) = validate (
        @_,
        {
            server => { type => OBJECT },
            pnode  => { type => SCALAR },
        }
    );
    my ($node, $srv);

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    if ($srv->{mode} ne "7") {
        $opts{server}->set_execution_server(-Server => $opts{pnode});
    }

    return;
}
#---------------------------------------------------------------------

=head2 setup_aggr

    setup_aggr(server       => $Server,
               aggr_name    => $aggr
               [,aggr_size  => $TRAD_SIZE ]
               [,force      => 0 ]
               [,raidtype   => 'raid4' ]
               [,timeout    => 7200 ]);

Usage: setup_aggr(server => $Server, aggr_name => 'my_aggr', aggr_size => 3);

This routine checks if the aggr_name exists and verifies that it is online.
If the aggregate does not exists, create it creates the aggregate.

=over

=item Arguments

=over

=item aggr_name

(Required) : Name of aggregate to create

=item aggr_size

(Optional) : Number of disks to use for the aggregate

=item force

(Optional) : Force aggregate creation

=item raidtype

(Optional) : Specify the raid type. The value must be either
             'raid4' or 'raid_dp'

=item server

(Required) : Name of the server object.

=item timeout

(Optional) : Specify the timeout for aggregate creation.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setup_aggr(%;) {
    my (%opts) = validate (
        @_,
        {
            aggr_name  => { type  => SCALAR },
            aggr_size  => { type  => SCALAR, default => '3' },
            force      => { type  => SCALAR, default => 1 },
            raid_type  => { type  => SCALAR, default => 'raid_dp' },
            server     => { type  => OBJECT },
            timeout    => { type  => SCALAR, default => 3600 },
        }
    );
    my ($aggr_exist, $fc, $server, $srv, $state, $status_obj);

    # Get the state of the server
    $server = $opts{server};
    $srv = _get_state($server);

    if ($srv->{mode} eq "7") {
        # 7-mode
        $fc = $srv->{fc};
        $aggr_exist = aggr_exists($fc, $opts{aggr_name});
        if ($aggr_exist) {
            $status_obj = aggr_status($fc, $opts{aggr_name});
            if (defined($status_obj)) {
                $state = $status_obj->state();
                if ($state ne 'online') {
                    aggr_online($fc, $opts{aggr_name});
                }
            }
        }

        if (!$aggr_exist) {
            $Log->comment("Creating aggregate $opts{aggr_name}, ".
                       "size $opts{aggr_size}");
            aggr_create(
                $fc,
                $opts{aggr_name},
                $opts{aggr_size},
                force    => $opts{force},
                raidtype => $opts{raid_type},
                timeout  => $opts{timeout},
            ) or croak $@;
        }
    } else {
        # 10-mode
        eval {
            $aggr_exist = $server->storage_aggregate_show(
                                  aggregate => $opts{aggr_name});
        };
        if (!$@) {
            # Make sure the aggregate is online
            $server->storage_aggregate_modify(
                                  aggregate => $opts{aggr_name},
                                  state     => 'online');
        } else {
            # Create a 10-mode stripped aggregate
            $server->storage_aggregate_create(
                       aggregate      => $opts{aggr_name},
                       'allow-mixed'  => 'true',
                       diskcount      => $opts{aggr_size},
                       raidtype       => $opts{raid_type},
                       'volume-style' => 'striped',
                       -Timeout       => $opts{timeout});
        }
    }

    return;
}

#---------------------------------------------------------------------

=head2 setup_vol

    setup_vol(server      => $Server,
              aggr_name   => $aggr_name,
              [,vol_name  => $vol_name ]
              [,flex_size => $flex_size ]);

Usage: setup_vol(server    => $Server,
                 aggr_name => 'my_aggr',
                 vol_name  => 'my_vol,
                 flex_size => '30m');

This routine checks to see if the given volume exists and purges it if it
does. The routine then creates a flex_size volume on the given aggregate.
If the volume created is used for cifs testing, the qtree security is
set to 'mixed'.

=over

=item Arguments

=over

=item aggr_name

(Required) : Name of aggregate to use when creating a volume

=item flex_size

(Optional) : Size of volume to create

=item server

(Required) : Name of the server object.

=item vol_name

(Required) : Name of volume to create

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub setup_vol(%;) {
    my (%opts) = validate (
        @_,
        {
            aggr_name  => { type  => SCALAR },
            flex_size  => { type  => SCALAR, default => '30m' },
            server     => { type  => OBJECT },
            vol_name   => { type  => SCALAR },
        }
    );
    my ($fc, %sec_opt, $server, $srv, $vol_exist, @vslist);

    # Get the state of the server
    $server = $opts{server};
    $srv = _get_state($server);

    if ($srv->{mode} eq "7") {
        # 7-mode
        $fc = $srv->{fc};
        $vol_exist = vol_status($fc, $opts{vol_name});
        if ($vol_exist) {
            vol_purge($fc, $opts{vol_name});
            if ($@) {
                # A copy of ontap hammer may still be running causing
                # the volume purge to fail. If the offline failed then
                # try aborting hammer and then purging again.
                if ($@ =~ /may not be offlined because/i) {
                    ontap_hammer_abort(fc => $fc);
                    vol_purge($fc, $opts{vol_name}) or croak $@;
                } else {
                    croak $@;
                }
            }
            $vol_exist = 0;
        }
        if (!$vol_exist) {
            $Log->comment("Creating volume $opts{vol_name} on containing ".
                       "aggregate $opts{aggr_name}, ".
                       "with size of: $opts{flex_size}");

            vol_create_flex(
                $fc,
                $opts{vol_name},
                $opts{aggr_name},
                $opts{flex_size},
            ) or croak $@;
        }
        if ($opts{vol_name} =~ /^cifs/) {
            if (filer_qtree_security_set($fc, "/vol/$opts{vol_name}",
                                         'mixed') == 0) {
                croak("QTREE set to mixed failed on /vol/$opts{vol_name}");
            }
        }
    } else {
        # 10-mode
        eval {
            $vol_exist = $server->volume_show(volume => $opts{vol_name});
        };
        if (!$@) {
            $server->volume_delete(volume => $opts{vol_name});
        }

        if ($opts{vol_name} =~ /^cifs/) {
            %sec_opt = ('security-style' => 'mixed');
        } else {
            %sec_opt = ();
        }
        @vslist = $server->vserver_show();
        $server->volume_create(volume          => $opts{vol_name},
                               vserver         => $vslist[0]->{vserver},
                               parent          => $vslist[0]->{rootvolume},
                               'junction-path' => $opts{vol_name},
                               aggregate       => $opts{aggr_name},
                               size            => $opts{flex_size},
                               %sec_opt);
    }

    return;
}

#---------------------------------------------------------------------

=head2 server_is_licensed

    $result = server_is_licensed(server => $Server, service => 'nfs');

This routine checks to see if the service is licensed.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item service

(Required) : Name of the service.

=back

=item Returns

On success, returns 0 if service is not licensed, 2 if it is a site licensed,
else returns 1. Otherwise returns undef and sets $@.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_is_licensed(%;) {
    my (%opts) = validate (
        @_,
        {
            server  => { type => OBJECT },
            service => { type => SCALAR },
        }
    );
    my ($fc, $srv);
    my $ret = 0;
    my $server = $opts{server};
    my $service = $opts{service};

    # Get the state of the server
    $srv = _get_state($server);

    if ($srv->{mode} eq "7") {
        # 7-mode
        $fc = $srv->{fc};
        $ret = 1 if (filer_service_is_licensed($fc, $service));
    } else {
        # XXX Special code to handle flex_scale. Remove when 10-mode
        # licenses are ported.
        if ($service eq "flex_scale") {
            $ret = _flex_scale($server, 'inq', $service);
        } else {
            eval { $server->system_license_show(feature=>$Lic_map{$service}) };
            $ret = 1 if (!$@);
        }
    }

    return $ret;
}

#---------------------------------------------------------------------

=head2 server_license

    $result = server_license(server => $Server, service => 'nfs');

Given a server object and a service name, uses the 'license' command
to install a new license.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item service

(Required) : Name of the service. It may be a single service name such as
'nfs'.

=back

=item Returns

On success, returns true. Otherwise returns 0 and sets $@.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_license(%;) {
    my (%opts) = validate (
        @_,
        {
            server  => { type => OBJECT },
            service => { type => SCALAR },
        }
    );
    my ($cmd, $code, $fc, $map_svc, $node, $raw_output, $srv);
    my $server = $opts{server};
    my $service = $opts{service};
    my $ret = 0;

    # Get the state of the server
    $srv = _get_state($server);

    # Get the node name
    $node = $srv->{node};

    eval {
        if (server_is_licensed(server => $server, service => $service)) {
            $ret = 1;
        } else {
            if ($srv->{mode} eq "7") {
                # 7-mode
                $fc = $srv->{fc};
                filer_service_license($fc, $service);
            } else {
                # XXX Special code to handle flex_scale. Remove when 10-mode
                # licenses are ported.
                if ($service eq "flex_scale") {
                    $ret = _flex_scale($server, 'inq', $service);
                    $ret = _flex_scale($server, 'lic', $service) if !$ret;
                } else {
                    # 10-mode
                    # XXX Remove below code when licenses are ported to 10-mode
                    # Map license
                    $map_svc = $Lic_map{$service};
                    croak("Unknown service mapping '$service'") if !$map_svc;

                    # Get the license code
                    $code = _get_lic_code($server, $map_svc);
                    croak("Unknown service '$service'") if !$code;

                    eval { $server->system_license_show(feature=>$map_svc) };
                    if ($@ =~ /There are no entries matching your query/){
                        eval {
                            $server->system_license_add("license-code"=>$code)
                        };
                        croak("Unable to add license: $@") if $@;
                    } elsif ($@) {
                        croak("Unable to run command: $@") if $@;
                    }
                    $ret = 1;
                }
            }
        }
    };

    return $ret;
}

#---------------------------------------------------------------------

=head2 server_unlicense

    $result = server_unlicense(server => $Server, service => 'nfs');

Given a server object and a service name, uses the 'license' command
to remove the corresponding license.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item service

(Required) : Name of the service. It may be a single service name such as
'nfs'.

=back

=item Returns

On success, returns 1. Otherwise returns 0 and sets $@.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_unlicense(%;) {
    my (%opts) = validate (
        @_,
        {
            server  => { type => OBJECT },
            service => { type => SCALAR },
        }
    );
    my ($cmd, $code, $fc, $node, $raw_output, $srv);
    my $server = $opts{server};
    my $service = $opts{service};
    my $ret = 0;

    # Get the state of the server
    $srv = _get_state($server);

    # Get the node name
    $node = $srv->{node};

    eval {
        if (!server_is_licensed(server => $server, service => $service)) {
            $ret = 1;
        } else {
            # Build the command
            if ($srv->{mode} eq "7") {
                # 7-mode
                $fc = $srv->{fc};
                filer_service_unlicense($fc, $service);
                $ret = 1 if !$@;
            } else {
                # XXX Special code to handle flex_scale. Remove when 10-mode
                # licenses are ported.
                if ($service eq "flex_scale") {
                    $ret = _flex_scale($server, 'ulic', $service);
                } else {
                    # 10-mode
                    eval {
                        $server->system_license_delete(
                            "feature"=>$Lic_map{$service});
                    };
                    croak("Unlicensing failed: $@") if $@;
                    $ret = 1;
                }
            }
        }
    };

    return $ret;
}

#---------------------------------------------------------------------

=head2 server_options_get

    server_options_get(server => $server, option => $option);

This routine gets the value for the option given.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item option

(Required) : Name of the option.

=back

=item Returns

Returns the value of the option or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_options_get(%;) {
    my (%opts) = validate (
        @_,
        {
            option => { type => SCALAR },
            server => { type => OBJECT },
        }
    );
    my ($cmd, @info, $line, $node, $raw_output, $srv, $value);

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    # Build the command
    # 7-mode
    $cmd = "options $opts{option}";
    # 10-mode
    $cmd = "run -node $node -command \"$cmd\"" if ($srv->{mode} ne "7");

    # Send the command to the filer
    eval {
        $opts{server}->cli(cmd    => $cmd,
                           output => \$raw_output);
    };
    croak("Unable to run command: $@") if $@;

    $value = "";
    @info = split /\n/, $raw_output;
    foreach my $line (@info) {
        if ($line =~ /$opts{option}/) {
            (undef, $value) = split(/\s+/, $line);
        }
    }

    return $value;
}

#---------------------------------------------------------------------

=head2 server_options_set

    server_options_set(server => $server,
                       option => $option,
                       value  => $value
                       [,verify => $verify ]);

This routine sets the value for the option given.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item option

(Required) : Name of the option.

=item value

(Required) : Value to set the option to.

=item verify

(Optional) : Set to 0 to skip verification that value was set.

=back

=item Returns

Returns 0 if verification determined the set operation failed or
croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_options_set(%;) {
    my (%opts) = validate (
        @_,
        {
            option => { type => SCALAR },
            server => { type => OBJECT },
            value  => { type => SCALAR },
            verify => { type => SCALAR, default => 1 },
        }
    );
    my ($cmd, $node, $raw_output, $status, $srv, $value);

    $status = 1;

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    # Build the command
    # 7-mode
    $cmd = "priv set -q $Priv; options $opts{option} $opts{value}";
    # 10-mode
    $cmd = "run -node $node -command \"$cmd\"" if ($srv->{mode} ne "7");

    # Send the command to the filer
    eval {
        $opts{server}->cli(cmd    => $cmd,
                           output => \$raw_output);
    };
    croak("Unable to run command: $@") if $@;

    # Verify that the option was set to the value provided
    if ($opts{verify}) {
        # Verify value has been changed
        $value = server_options_get(
            server => $opts{server},
            option => $opts{option},
        );

        # Some variables disappear when set to 'off'
        if ($opts{value} =~ /^off$/i) {
            $status = 0 if ($value !~ m/^off$|^$/i)
        } else {
            $status = 0 if ($value ne $opts{value});
        }
    }

    return $status;
}

#---------------------------------------------------------------------

=head2 server_panic

    server_panic(server    => $Server
                 [,timeout => $timeout ]);

This routine panics the filer.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item timeout

(Optional) : Specify the timeout for reboot to occur.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_panic(%;) {
    my (%opts) = validate (
        @_,
        {
            server  => { type => OBJECT },
            timeout => { type => SCALAR, default => 300, regex => qr/\d+/ },
        }
    );
    my ($cmd, %info, $node, $raw_output, $srv);

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    # Build the command
    # 7-mode
    $cmd = "priv set advanced; panic";
    # 10-mode
    $cmd = "run -node $node -command \"$cmd\"" if ($srv->{mode} ne "7");

    # Send the command to the filer
    # Use noprompt => 1, to replace the -blind option of the say command
    eval {
        $opts{server}->console(cmd        => $cmd,
                               output     => \$raw_output,
                               noprompt   => 1,
                               '-Timeout' => $opts{timeout});
    };
    croak("Unable to run command: $@") if $@;

    return;
}

#---------------------------------------------------------------------

=head2 server_reboot

    server_reboot(server    => $Server
                  [,timeout => $timeout ]);

This routine will reboot the filer via a direct command call (reboot -t 0).

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item timeout

(Optional) : Specify the timeout for reboot to occur.

=back

=item Returns

Returns nothing or sets $@ on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_reboot(%;) {
    my (%opts) = validate (
        @_,
        {
            server  => { type => OBJECT },
            timeout => { type => SCALAR, default => 300, regex => qr/\d+/ },
        }
    );
    my ($cmd, $node, $raw_output, $srv);
    my $server = $opts{server};

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    # Build the command
    # 7-mode
    $cmd = "reboot -t 0";
    # 10-mode
    $cmd = "run -node $node -command \"$cmd\"" if ($srv->{mode} ne "7");

    # Send command
    eval {
        $opts{server}->console(cmd        => $cmd,
                               output     => \$raw_output,
                               '-Timeout' => $opts{timeout});
    };
    croak("Unable to run command: $@") if $@;

    return;
}

#---------------------------------------------------------------------

=head2 server_reboot_to

    server_reboot_to(server => $Server, state => $state);

This routine will reboot filer to a specified state.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item state

(Required) : State to reboot the filer to.

=back

=item Returns

Returns nothing or sets $@ on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_reboot_to(%;) {
    my (%opts) = validate (
        @_,
        {
            server => { type => OBJECT },
            state  => { type => SCALAR, regex => qr/^(UP|LOADER_BANNER)$/ },
        }
    );
    my ($node, $srv);
    my $server = $opts{server};
    my $state = $opts{state};
    # 7 mode translated to 10 mode hashes
    my %filermode = (
        "UP"            => "REBOOT",
        "LOADER_BANNER" => "CFE",
    );
    my %wait_for = (
        "REBOOT" => "CLI",
        "CFE"    => "CFE",
    );

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    if ($srv->{mode} eq "7") {
        $Log->step("Reboot the filer '$state'");
        reboot_to(host($node), $state);
    } else {
        $Log->step("Reboot the filer '$filermode{$state}'");
        eval {
            $server->transit_to(state    => $filermode{$state},
                                wait_for => $wait_for{$filermode{$state}})
        };
    }

    return;
}

#---------------------------------------------------------------------

=head2 server_transit_to

    server_transit_to(server => $Server);

This routine calls transit_to make sure the filer/node is up.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=back

=item Returns

Returns nothing or sets $@ on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_transit_to(%;) {
    my (%opts) = validate (
        @_,
        {
            server => { type => OBJECT },
        }
    );
    my ($node, $org_ex_srv, $srv);

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    if ($srv->{mode} eq "7") {
        $Log->step("Calling transit_to for server $node");
        transit_to(host($node), 'UP');
    } else {
        $Log->step("Calling transit_to CLI for node $node");
        $org_ex_srv = $opts{server}->get_execution_server();
        $opts{server}->set_execution_server(-Server => $node);
        eval {
            $opts{server}->transit_to(node  => $node,
                                      state => 'CLI');
        };
        $opts{server}->set_execution_server(-Server => $org_ex_srv);
    }

    return;
}

#---------------------------------------------------------------------

=head2 server_kernel_is_debug

    $debug = server_kernel_is_debug(server => $Server);

This routine obtains the version of ONTAP on a filer to determine if kernel is
a "debug" build.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=back

=item Returns

On success returns 1 if ONTAP kernel is a "debug" build. Otherwise returns 0.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_kernel_is_debug(%;) {
   my (%opts) = validate (
        @_,
        {
            server => { type => OBJECT },
        }
    );
    my ($cmd, $node, $raw_output, $srv);
    my $server = $opts{server};

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    # Build the command
    # 7-mode
    $cmd = "version -v";
    # 10-mode
    $cmd = "run -node $node -command $cmd" if ($srv->{mode} ne "7");

    # Send the command to the filer
    eval {
        $opts{server}->cli(cmd    => $cmd,
                           output => \$raw_output);
    };
    croak("Unable to run command: $@") if $@;

    return 1 if ($raw_output =~ m/<.*d.*>/);

    return 0;
}

#---------------------------------------------------------------------

=head2 server_kernel_is_domain_xing

    $dom_xing = server_kernel_is_domain_xing(server => $Server);

This routine obtains the version of ONTAP on a filer to determine if kernel is
a "debug" build.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=back

=item Returns

On success returns 1 if ONTAP kernel is a "domain xing" build. Otherwise
returns 0.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_kernel_is_domain_xing(%;) {
   my (%opts) = validate (
        @_,
        {
            server => { type => OBJECT },
        }
    );
    my ($cmd, $node, $raw_output, $srv);
    my $server = $opts{server};

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    # Build the command
    # 7-mode
    $cmd = "version -v";
    # 10-mode
    $cmd = "run -node $node -command $cmd" if ($srv->{mode} ne "7");

    # Send the command to the filer
    eval {
        $opts{server}->cli(cmd    => $cmd,
                           output => \$raw_output);
    };
    croak("Unable to run command: $@") if $@;

    return 1 if ($raw_output =~ m/<.*A*>/);

    return 0;
}

#---------------------------------------------------------------------

=head2 server_supports_command

    $result = server_supports_command(server  => $Server,
                                      command => $cmd,
                                      priv    => $priv);

This routine determine if the command is supported on the filer.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item cmd

(Required) : Name of the command to check.

=item priv

(Required) : Priviledge level to check the command.

=back

=item Returns

Returns 'SUCCESS' if command is found in expected privilege level,
returns message with filer ONTAP version if command is not found.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub server_supports_command(%;) {
   my (%opts) = validate (
        @_,
        {
            server  => { type => OBJECT },
            command => { type => SCALAR },
            priv    => { type => SCALAR },
        }
    );
    my ($cmd, $node, $raw_output, $ret, $srv);
    my $server = $opts{server};
    my $command = $opts{command};
    my $priv = $opts{priv};

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    # Build the command
    # 7-mode
    $cmd = "priv set -q $priv; help $command";
    # 10-mode
    $cmd = "run -node $node -command $cmd" if ($srv->{mode} ne "7");

    # Send the command to the filer
    eval {
        $opts{server}->cli(cmd    => $cmd,
                           output => \$raw_output);
    };
    croak("Unable to run command: $@") if $@;

    if ($raw_output !~ /unrecognized command/) {
        $ret = "SUCCESS";
    } else {
        # Get Ontap version

        # Build the command
        # 7-mode
        $cmd = "version -v";
        # 10-mode
        $cmd = "run -node $node -command $cmd" if ($srv->{mode} ne "7");

        # Send the command to the filer
        eval {
            $opts{server}->cli(cmd    => $cmd,
                               output => \$raw_output);
        };
        croak("Unable to run command: $@") if $@;

        $ret = "Command ($command) is not supported in ($raw_output) ".
               $srv->{mode}."-Mode";
    }

    return $ret;
}

#---------------------------------------------------------------------

=head2 sysconfig_c

    sysconfig_c(server => $Server);

This routine checks to see that there are no configuration errors.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=back

=item Returns

Returns message or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub sysconfig_c(%;) {
    my (%opts) = validate (
        @_,
        {
            server => { type => OBJECT },
        }
    );
    my ($cmd, $node, $raw_output, $srv);

    # Get the state of the server
    $srv = _get_state($opts{server});

    # Get the node name
    $node = $srv->{node};

    # Build the command
    # 7-mode
    $cmd = "sysconfig -c";
    # 10-mode
    $cmd = "run -node $node -command \"$cmd\"" if ($srv->{mode} ne "7");

    # Send the command to the filer
    eval {
        $opts{server}->cli(cmd    => $cmd,
                           output => \$raw_output);
    };
    croak("Unable to run command: $@") if $@;

    return $raw_output;
}

#---------------------------------------------------------------------
# Internal function to get the filer state
#---------------------------------------------------------------------

sub _get_state($;) {
    my ($server) = shift;
    my (%entry, $fc, @nodes);

    if (!defined($Server_cache{$server})) {
        # Get the type of filer we are working with
        $entry{mode} = filer_mode(server => $server);

        # Save connection object
        $fc = $server->get_connection(type => 'cli_connection');
        if (!defined($fc)) {
            croak("_get_state: cannot get cli connection from Server object");
        }
        $entry{fc} = $fc;

        # Get the name of the first node
        $entry{node} = ($server->servers)[0];
        # Get an array of the names of all nodes
        @nodes = $server->servers;
        $entry{allnodes} = \@nodes;
        $Server_cache{$server} = \%entry;
    }
    return ($Server_cache{$server});
}

#---------------------------------------------------------------------
# Internal function to get the license code
#---------------------------------------------------------------------

sub _get_lic_code($$;) {
    my ($server) = shift;
    my ($service) = shift;
    my ($code);

#XXX Hard code the 10-mode licenses until "licenses" are ported to 10-mode
    my %lic_codes = (
        "Base"           => "NWUZZJPJYBFDAA",
        "CIFS"           => "JAMHCKPJYBFDAA",
        "FlexVol_HPO"    => "XFOSBKPJYBFDAA",
        "Mirror"         => "LLQDBKPJYBFDAA",
        "NFS"            => "TJFAEKPJYBFDAA",
        "SnapRestore"    => "VUJWCKPJYBFDAA",
        "Striped_Volume" => "DTYSFKPJYBFDAA",
    );

    $code = $lic_codes{$service};
    $code = 0 if !$code;

    return $code;
}

#---------------------------------------------------------------------
# Internal function to handle licensing flex_scale in 10-mode
#---------------------------------------------------------------------

sub _flex_scale($$$;) {
    my ($server) = shift;
    my ($action) = shift;
    my ($service) = shift;
    my ($cmd, @info, $line, $node, $raw_output, $srv);
    my $ret = 0;

#XXX Hard code the 10-mode licenses until "licenses" are ported to 10-mode
    # This is actually 7-mode license code since flex_scale does not exist in
    # 10-mode.
    my %lic_codes = (
        "flex_scale" => "MZGLGBN",
    );

    # Get the state of the server
    $srv = _get_state($server);

    # Get the node name
    $node = $srv->{node};

    # Use node scope for 10-mode
    # Build the command
    if ($action eq 'inq') {
        $cmd = "license";
    } elsif ($action eq 'lic') {
        $cmd = "license add $lic_codes{$service}";
    } elsif ($action eq 'ulic') {
        $cmd = "license delete $service";
    }
    $cmd = "run -node $node -command \"$cmd\"";

    # Send the command to the filer
    eval {
        $server->cli(cmd    => $cmd,
                     output => \$raw_output);
    };
    croak("Unable to run command: $@") if $@;

    if ($action eq 'inq') {
        @info = split(/\r*\n/,$raw_output);
        foreach $line (@info) {
            if ($line =~ /^$service\s|\s+$service\s/i) {
                if ($line =~ /not licensed|expired/) {
                    return 0;
                } elsif ($line =~ /\ssite\s/) {
                    return 2;
                } else {
                    return 1;
                }
            }
        }
    } elsif ($action eq 'lic') {
        $ret = 1 if ($raw_output =~ /$service licensed/i);
    } else {
        $ret = 1 if ($raw_output =~ /unlicensed $service\./i);
    }

    return $ret;
}

1;
