# $Id: //depot/prod/test/Rsierranevada/lib/String/CaseManip.pm#1 $

## @summary     Class to manipulate case of a given string.
## @author      astraub@netapp.com
## @pod here

=head1 NAME

String::CaseManip.pm

=head1 SYNOPSIS

=begin man

    use String::CaseManip;
    use strict;
     
    # define string to use
    my $string = 'abc';
     
    # create object
    my $nc = String::CaseManip->new('name' => $string);
    
    # loop through all possibilities
    foreach (0..$nc->max_mask()) {
       printf("get_next_name: %s\n", $nc->get_next_name());
    }
    
    Output:
    
    get_next_name: abc
    get_next_name: abC
    get_next_name: aBc
    get_next_name: aBC
    get_next_name: Abc
    get_next_name: AbC
    get_next_name: ABc
    get_next_name: ABC
    

=end man


=head1 DESCRIPTION

C<String::CaseManip> is a class for changing a string of characters
through all possible permutaions of upper/lower case.

Provided methods generate case variants of the provided string,
or an index can be manually set to apply specific case variants.

=head1 METHODS

=cut



package String::CaseManip;
use strict;
use Params::Validate qw(validate BOOLEAN SCALAR UNDEF);
use Tharn qw($Log);
use Data::Dumper;

=head2 new

C<< String::CaseManip->new(%params) >>

Invocant for the String::CaseManip class.

=over 4

=item Input

=over 4

=item * name => <scalar>

(Optional)
Default = undef
Defines string to use for manipulation

=item * random => <boolean>

(Optional)
Default = undef
Create a random string instead of providing one

=item * length => <scalar>

(Optional)
Default = 10
Length of random string to create if requested.

=item * index => <scalar>

(Optional)
Default = -1
Starting point to apply name mask.

=back

=item Output

Class Reference

Example

C<< my $nc_obj = String::CaseManip->new('name' => $string); >>

=back

=cut

################################################################################
#
# Subroutine:   new()
#
# Objective:    Invocant for the class
#
# Details:
#
#   Input:
#       name   - scalar containing string to be manipulated
#       random - boolean to request random string
#       length - length of random string to be generated
#       index  - set starting point for mask index
#
#   Return value:
#       object
#
################################################################################


sub new {
    $Log->enter();
    my $class = shift;
    my $self =  {validate( @_,
                           { name   => {type => SCALAR|UNDEF, optional =>1},
                             random => {type => BOOLEAN, optional => 1},
                             'length' => {type => SCALAR, optional => 1, default => 10},
                             'index'  => {type => SCALAR, optional => 1, default => "-1"},
                            }
                         )
                };


    $Log->debug("Input Params: ".Dumper($self));			
    if ($self->{random} && defined($self->{name})) {
    	die "'name' or 'random' are mutually exclusive. Only one can be defined"; 
    }
    if ($self->{random}) {
    	$self->{name} = _random_string($self->{length});
		$Log->trace("Generated random string: $self->{name}");
    } else {
		$Log->trace("Setting to lowercase,and verifying string: $self->{name}");
    	_verify($self->{name});
    	$self->{name} = lc($self->{name});
    }
    $self->{max_mask} = _max_mask($self->{name});
    if ($self->{index} > 0) {
		$Log->trace("Setting starting index: $self->{index}");
    	die "Index requested $self->{index} exceeds limit $self->{max_mask}" if ($self->{index} > $self->{max_mask}); 
    	$self->{maskedname} = _apply_name_mask($self->{name}, $self->{index});
    } else {
    	$self->{maskedname} = $self->{name};
    }
	$Log->debug("First iteration of string: $self->{name}");
    bless($self, $class);
	$Log->exit();
    return $self;

}

################################################################################
#
# Subroutine:   _random_string()
#
# Objective:    generate a random string
#
# Details:
#
#   Input:
#       <scalar> - length of string to generate
#   Return value:
#       scalar containing randomly generated string
#
################################################################################

sub _random_string {
	$Log->enter();
    my ($len) = shift;
    my $str;
    my @chars = ("a".."z");
    for (1..$len) {
	$str .= $chars[int(rand(@chars))];
    }
	$Log->exit("Returning random string: $str");
    return $str;

}

################################################################################
#
# Subroutine:   _verify()
#
# Objective:    private routine to verify that the string
#               contains only valid characters
#
# Details:
#
#   Input:
#       <scalar> - string to verify
#
#   Return value:
#       1 on success, dies on failure
#
################################################################################

sub _verify {
	$Log->enter();
    my $str = shift;
    die "Strings consisting of only ascii chars (a-z) are currently supported." if ($str =~ /[^a-zA-Z]+/);
    $Log->exit();
    return 1;
}

=head2 get_name


C<< $nc_obj->get_name() >`>

=over 4

=item Description

Get current permutation of string

=item Input

N/A

=item Output

Scalar containing the current case permutation of the string 

=back

=cut

################################################################################
#
# Subroutine:   get_name()
#
# Objective:    Get current permutation of string
#
# Details: Get the 
#       
#   Input:
#       none
#   Return value:
#       Scalar containing the current case permutation of the string 
#
################################################################################


sub get_name {
	$Log->enter();
    my $self = shift;
    $Log->exit("Returning name: $self->{maskedname}");
    return $self->{maskedname};
}

=head2 get_mask_index


C<< $nc_obj->get_mask_index() >>

=over 4

=item Description

Get current the current mask index that was applied to the string

=item Input

N/A

=item Output

Scalar containing the integer value of the current mask index

=back

=cut

################################################################################
#
# Subroutine:   get_mask_index()
#
# Objective:    retrieve the current index of the mask index
#
# Details:
#
#   Input:
#       none
#
#   Return value:
#       scalar containing the index value
#
################################################################################

sub get_mask_index {
	$Log->enter();
    my $self = shift;
	$Log->exit("Returning index: $self->{index}");
    return $self->{index};
}

=head2 set_mask_index


C<< $nc_obj->set_mask_index(%params) >>

=over 4

=item Description

Set current the current mask index and apply to the string

=item Input

=over 4

=item * index => <scalar>

(Optional)
Default = -1
Index number to apply. Must be within the valid range for the string.

=back

=item Output

Boolean 

=back

=cut

################################################################################
#
# Subroutine:   set_mask_index()
#
# Objective:    change the current mask index value 
#
# Details:
#
#   Input:
#       index - value to set the mask index will be set
#
#   Return value:
#       1 if successful; dies on failure.
#
#
################################################################################

sub set_mask_index {
	$Log->enter();
    my $self = shift;
    my %params = validate( @_,{ index => {type => SCALAR, optional =>1, default => "-1"}});
    $Log->debug("Input Params: ".Dumper(\%params));

    die "Mask exceeds max mask (Req: $params{index} Limit: $self->{max_mask})" if ($params{index} > $self->{max_mask});
    $self->{index} = $params{index};
    $self->{maskedname} = _apply_name_mask($self->{name}, $self->{index});
	$Log->exit();
    return 1;
}

=head2 max_mask

=over 4


=item Description

C<< $nc_obj->max_mask() >>

Get the largest mask index that may be applied to the string.

=item Input

N/A

=item Output

Scalar 

=back

=cut

################################################################################
#
# Subroutine:   max_mask() 
#
# Objective:    Return the maximum mask value
#               that can be applied to the string
#
# Details:
#
#   Input:
#       none 
#
#   Return value:
#       scalar containg the maximum mask value
#
################################################################################


sub max_mask {
	$Log->enter();
    my $self = shift;
	$Log->exit("Returning max_mask: $self->{max_mask}");
    return $self->{max_mask};
}

################################################################################
#
# Subroutine:   _max_mask() 
#
# Objective:    private routine to return the maximum mask value
#               that can be applied to the string
#
# Details:
#
#   Input:
#       <scalar> - string to be manipulated
#
#   Return value:
#       scalar containg the maximum mask value
#
################################################################################

sub _max_mask {
	$Log->enter();
    my $str = shift;
    my $length = length($str);
    my $combos = (1 << ($length)) -1;
	$Log->exit("Returning max_mask: $combos");
    return $combos;
    
}

=head2 get_random_name


C<< $nc_obj->get_random_name() >>

=over 4

=item Description

Randomly set the current the current mask index, apply it to the string, and return result 

Input

=over 4

=item * allow_dup => <boolean>

(Optional)
Default = 0
If set, Allow a string to be returned that is the same as the string previously generated.
Otherwise, retry until a string is returned that is not the same as the last. 

=back

Output

Scalar 

=back

=cut

################################################################################
#
# Subroutine:   get_random_name()
#
# Objective:    randomly select a case modified string
#
# Details:
#
#   Input:
#       allow_dup (boolean) - llow a string to be returned that is the
#       same as the string previously generated.
#
#   Return value:
#       Scalar containing the case modified string
#
################################################################################


sub get_random_name {
	$Log->enter();
    my $self = shift;
    my %params = validate( @_,{ allow_dup => {type => BOOLEAN, optional =>1, default => "0"}});

	$Log->debug("Input Params : ".Dumper(\%params));	 
	my $picked;
	my $newindex;
	until ($picked)	{
        $newindex = int(rand($self->{max_mask} + 1));
		$Log->trace("Random index chosen: $newindex previous: $self->{index}");
		$picked = 1 if ($params{allow_dup} or ($newindex ne $self->{index}));
	}
	$Log->debug("index is $newindex");
    $self->{index} = $newindex;
    $self->{maskedname} = _apply_name_mask($self->{name}, $self->{index});
	$Log->exit("Return: $self->{maskedname}");
    return $self->{maskedname};
}

=head2 get_next_name


C<< $nc_obj->get_next_name() >>

=over 4

=item Description

Increment the current the current mask index, apply it to the string, and return result 

=item Input

N/A

=item Output

Scalar 

=back

=cut

################################################################################
#
# Subroutine:  get_next_name()
#
# Objective:   return the next iteration of the case modified string
#
# Details:
#
#   Input:
#       None.
#
#   Return value:
#       Scalar containing the case modified string
#
################################################################################


sub get_next_name {
	$Log->enter();
    my $self = shift;
    $self->{index}++;
    if ($self->{index} > $self->{max_mask}) {
        $Log->debug("max_mask limit exceeded, resetting mask index to 0");
        $self->{index} = 0 ;
    }
    $self->{maskedname} = _apply_name_mask($self->{name}, $self->{index});
	$Log->exit("Returning maskedname : $self->{maskedname}");
    return $self->{maskedname};
}

################################################################################
#
# Subroutine:   _apply_name_mask()
#
# Objective:    Private subroutine to apply bitmask
#
# Details:
#
#   Input:
#       <scalar>  - string to be manipulated
#       <scalar>  - integer value of bitmask to apply
#
#   Return value:
#       scalar containing case modified string
#
################################################################################

sub _apply_name_mask {
    my ($instr, $mask) = @_;
	my $outstring;
    $Log->enter();
    my $length = length($instr);
    $Log->debug("Applying mask: $mask ".sprintf ("(%0${length}b)",$mask)." to name: $instr");
    my @maskarr = split("",sprintf ("%0${length}b",$mask)); # create an array of the bitmask
    my @namearr = split("",$instr); # create an array of the string
    #Iterate through the bitmask, transform chars in the name array to uppercase if bit is true.
    for my $bit (0..scalar(@maskarr)-1) {
        $namearr[$bit] = uc($namearr[$bit]) if ($maskarr[$bit]);
    }
    #Change array back into a string, and return it
	$outstring = join("",@namearr);
    $Log->exit("Return: $outstring");
    return $outstring;
}


1;
