#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work with IPC and subtests
##          This library provides a wrapper around NATE's subtest and IPC
##          calls. This package provides easy subtest creation and management
##          along with communication to and from a subtest. A common set of
##          functions are provided for a parent/child to notify each other
##          of changes in their states.
## @author latif@netapp.com
## @status public
## @pod here

package SubIPC;

# Compiler directives.
use strict;

# Module import.
use Tharn;
use Error;
use San::SanLog;
use San::SanUtil;
use ParamsMgr;
use NATE::STAF::Peer qw($Parent);
use Time::HiRes qw( gettimeofday );

use Data::Dumper;

# Define local/global variables
# The variable $_IntObject is used to store an internal object that is NOT accessible
# by callers/users. This object is mainly used internal for storing/holding
# subtests, messages to and from subtests and any errors that may have occurred.
# When a called wants to check error (or other things) on the package they simply
# need to call the package error function, this will check the internal object
# for any errors and return to the caller.
use vars qw(@ISA $_IntObject);
@ISA = qw(Error SanLog);

# Create internal object when this package is called/used "use SubIPC;".
# This should only need to be called once throughout the entire testcase because
# this package allows for management of subtests by handle.
$_IntObject = SubIPC->_new();

# Exported global variables
$SubIPC::TRUE   = $SanUtil::TRUE;
$SubIPC::FALSE  = $SanUtil::FALSE;

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    create
# Objective: Create new subtest.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    my @stest_args;

    &ParamsMgr::process_args(%args);

    my $run = &ParamsMgr::get_arg(-arg => "run", -required => 1, -default => "");
    my $id  = &ParamsMgr::get_arg(-arg => "id",  -required => 1, -default => "");
    my $bg  = &ParamsMgr::get_arg(-arg => "bg",  -required => 0, -default => "1");
    my $ipc = &ParamsMgr::get_arg(-arg => "ipc", -required => 0, -default => "0");
    my $esc = &ParamsMgr::get_arg(-arg => "esc", -required => 0, -default => $SubIPC::FALSE);
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # If there's a run id prefix defined prepend it to the
    # id passed in
    my $run_id_prefix = &SubIPC::get_run_id_prefix();
    if ((defined($run_id_prefix)) and (length($run_id_prefix))) {
        $id = $run_id_prefix . $id;
    }

    # See if the called has enabled subtest counts
    if ($esc == $SubIPC::TRUE) {
        $id = &SubIPC::_get_next_subtest_count() . "_" . $id;
    }

    # Start building out the argument set for subtest
    # Empty out the subtests array otherwise the subtests count will incorrectly increment
    push(@stest_args, sub { $_IntObject->{_subtests} = []; $run->()});
    push(@stest_args, "-runid");
    push(@stest_args, "$id");

    # If bg is requested add it to list
    if ($bg == 1) {
        push(@stest_args, "-bg");
    } else {
        push(@stest_args, "-nobg");
    }

    # This next line is used to make the subtest exit (on failure) without
    # killing the parent
    push(@stest_args, -onexit => sub{my ($self) = @_; $self->exithandler(1);});

    # Create a new SubIPC object for the new subtest
    my $subtest_Object = SubIPC->_new();

    # Save options for subtest on object
    $subtest_Object->{ipc} = $ipc;
    $subtest_Object->{bg} = $bg;

    # Create new subtest and add it to the internal container
    $subtest_Object->{_subtest} = &Tharn::subtest(@stest_args);

    # Add a unique ID on each subtest using Time::HiRes
    $subtest_Object->{_subtest}->{'uid'} = Time::HiRes::gettimeofday();

    $_IntObject->sanlog(-msg => "Created subtest " . $subtest_Object->{_subtest}->{'runid'} . " " .
                                "With unique id of " . $subtest_Object->{'uid'},
                        -msg_level => 5);

    if ($bg) {
        # Push SubIPC object onto internal stack
        push (@{$_IntObject->{_subtests}}, $subtest_Object);
    } else {
        # otherwise, wait for it to finish
        &SanUtil::nate_wait_finish($subtest_Object->{_subtest});
    }

    $_IntObject->sanlog(-exit => 1);

    return $subtest_Object;
}

###############################################################################
# Method:    wait_finish
# Objective: wait for subtests to finish
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wait_finish {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    my ($message, $all_complete, $counter) = ({}, 0, 0);

    &ParamsMgr::process_args(%args);

    my $events = &ParamsMgr::get_arg(-arg => 'events', -required => 0, -default => {});
    my $subtests = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0, -default => &SubIPC::get_subtests());
    my $parent = &ParamsMgr::get_arg(-arg => 'parent', -required => 0, -default => {});
    my $sync_enable = &ParamsMgr::get_arg(-arg => 'sync_enable', -required => 0, -default => 0);
    my $sleep_interval_time = &ParamsMgr::get_arg(-arg => "sleep_interval_time",
                                                  -required => 0, -default => "1");
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return $SubIPC::FALSE;
    }

    # Verify that some subtests have been passed in
    if (scalar(@{$subtests}) == 0) {
        return $SubIPC::FALSE;
    }

    # Loop through subtests until all have return status of 'passed' or 'failed'
    while (1) {
        my $subtest = ${$subtests}[$counter];

        $_IntObject->sanlog(-msg => "Processing subtest (" . $subtest->{_subtest}->{'runid'} . ") ($counter) Total subtests: " .
                                    scalar(@{$subtests}),
                            -msg_level => 3);

        if (not &SubIPC::is_alive(-subtest => [$subtest])) {
            $_IntObject->sanlog(-msg => "Subtest (" . $subtest->{_subtest}->{'runid'} . ") is no longer alive",
                                -msg_level => 3);
            # If the exit_status is not defined and the subtest is not alive, Huston [sic] we have a problem.
            if (not defined($subtest->{exit_status})) {
                $_IntObject->sanlog(-msg => "Subtest (" . $subtest->{_subtest}->{'runid'} . ") is no longer alive " .
                                            "AND no exit status was registered!!",
                                    -msg_type => "warn",
                                    -msg_level => 0);

                # Set the internal flag for the subtest status
                $subtest->{exit_status} = "failed";

                # Make sure the subtest is stopped
                $subtest->stop(-signal => "comment", -message => "Killed by parent");
            }
        } elsif ($subtest->can_read()) {
            # Determine if it has message to be read and processed
            if (&SubIPC::error()) {
                $_IntObject->sanlog(-msg => "Unable to properly determine if subtest (" .
                                            $subtest->{_subtest}->{'runid'} . ") could be read\n" .
                                            &SubIPC::get_error(),
                                    -msg_level => 0);
                return $SubIPC::FALSE;
            }

            # Read available message and process it
            $message = $subtest->read(-block => $SubIPC::TRUE);
            if (&SubIPC::error()) {
                $_IntObject->sanlog(-msg => "Unable to properly read subtest (" .
                                            $subtest->{_subtest}->{'runid'} . ")\n" . &SubIPC::get_error(),
                                    -msg_level => 0);
                return $SubIPC::FALSE;
            }

            # Process the messages by type
            if ($message->{'type'} eq "pass") {
                $_IntObject->sanlog(-msg => "Passed message received from child (" .
                                            $subtest->{_subtest}->{'runid'} . ")",
                                    -msg_level => 0);
                # If pass, issue subtest exit handshake
                $subtest->stop(-signal => "comment", -message => "Passed");

                # Save the subtest exit status for future processing
                $subtest->{exit_status} = "pass";

                $message->{event_occurred} = "on_pass";
            } elsif ($message->{'type'} eq "fail") {
                $_IntObject->sanlog(-msg => "Failed message received from child (" .
                                            $subtest->{_subtest}->{'runid'} . ")",
                                    -msg_level => 0);
                # If fail, issue subtest exit handshake, force enabled
                $subtest->stop(-signal => "fail", -message => "Failed");

                # Save the subtest exit status for future processing
                $subtest->{exit_status} = "fail";

                $message->{event_occurred} = "on_fail";
            } elsif ($message->{'type'} eq "sync") {
                $_IntObject->sanlog(-msg => "Sync requested from child (" .
                                            $subtest->{_subtest}->{'runid'} . ")",
                                    -msg_level => 0);

                # Save the subtest sync type for future processing
                $subtest->{sync_type} = $message->{sync_type};
            } else {
                # Message is not recognised, print it out and add subtest back to list
                # for further processing by parent
                $_IntObject->sanlog(-msg => "Message received from child subtest (" .
                                            $subtest->{_subtest}->{'runid'} . "):\n" .
                                        Dumper($message),
                                    -msg_level => 2);

                # If the event is unknown, set the event_occurred to the type of the message.
                # This provides flexibility to coders to be able to define their own events and event handlers.
                $message->{event_occurred} = $message->{'type'};
            }

            # Push all messages received from subtest on stack for future processing
            push(@{$subtest->{_messages}}, $message);

            # Save the message on the internal object for future processing
            $_IntObject->{'_child_messages'}->{$subtest->{_subtest}->{'uid'}}->{time()}->{$message->{type}} = $message;
        }

        # If an event occurred, check to see if a event handler was requested by caller
        if (defined($message->{event_occurred})) {
            # If the event type that occurred is other or misc, default to on_other
            if ($message->{event_occurred} =~ /misc|other/i) {
                $message->{event_occurred} = "on_other";
            }

            # If caller defines a event handler of 'on_all', override the event handler sub with the one provided
            # for all events that occurred EXCEPT for sync. The sync is a special case that should NOT be overridden
            # with the on_all event handler.
            if (defined($events->{'on_all'}) and $message->{event_occurred} !~ /sync/i) {
                $_IntObject->sanlog(-msg => "Event handler overriding event type (" .
                                            $message->{event_occurred} . ") with (on_all)\n" .
                                            "because event type of \"on_all\" provided by caller",
                                    -msg_level => 3);
                $message->{event_occurred} = "on_all";
            }

            # Run any events handler caller specified
            if (defined($events->{$message->{event_occurred}})) {
                # Get details about event handler
                my $exec = $events->{$message->{event_occurred}}->{'exec'};
                my $desc = $events->{$message->{event_occurred}}->{'desc'};
                $_IntObject->sanlog(-msg => "About to execute post subtest process for (" .
                                            $subtest->{_subtest}->{'runid'} . ")" .
                                            " because of (" . $message->{event_occurred} . "):\n\t$desc",
                                    -msg_level => 3);

                # Execute requested event, use the subtest as the "parent" or caller for the requested
                # event so that the event will get the subtest handle that the event is being called
                # for.
                $subtest->$exec(-message => $message, -parent => $parent);
                if ($subtest->error()) {
                    $_IntObject->sanlog(-msg => "An error occurred while post subtest process (" .
                                                $subtest->{_subtest}->{'runid'} . ")",
                                        -msg_level => 0);
                }
            }
        }

        # Check to see if caller enabled sync's
        if ($sync_enable == 1) {
            # Process subtests sync (if there is one)
            if (&SubIPC::process_sync(-subtests => $subtests) == $SubIPC::FALSE) {
                # Check to see if had an error during the sync attempt
                if (&SubIPC::error()) {
                    # Had an error, fail out
                    return $SubIPC::FALSE;
                }
            } else {
                # A sync was achieved, process on_sync event handler caller provided (if it was provided)
                # Run any events handler caller specified
                $message->{event_occurred} = "on_sync";
                if (defined($events->{$message->{event_occurred}})) {
                    # Get details about event handler
                    my $exec = $events->{$message->{event_occurred}}->{'exec'};
                    my $desc = $events->{$message->{event_occurred}}->{'desc'};
                    $_IntObject->sanlog(-msg => "About to execute sync event handler provided by caller.\n\t" .
                                                $desc,
                                        -msg_level => 3);

                    $_IntObject->$exec(-message => $message);
                    if (&SubIPC::error()) {
                        $_IntObject->sanlog(-msg => "An error occurred while executing sync event handler",
                                            -msg_level => 0);
                        return $SubIPC::FALSE;
                    }
                }
            }
        }

        # Clear out any event and message that previously occurred
        $message = {};

        # Initially set the all_completed flag to 1 and then verify it by
        # going through the subtest list
        $all_complete = $SubIPC::TRUE;
        # Check to see if all subtests have completed
        foreach my $subtest (@{$subtests}) {
            if ($subtest->{exit_status} eq "") {
                $all_complete = $SubIPC::FALSE;
                last;
            }
        }

        if ($all_complete) {
            $_IntObject->sanlog(-msg => "All subtests for caller (" . &ParamsMgr::_parent_caller() . ") have completed!",
                                -msg_level => 0);
            last;
        }

        # Move the counter to the next value, if exceed size of subtest array, reset counter to zero
        $counter++;
        if ($counter >= scalar(@{$subtests})) {
            $counter = 0;
            # Sleep n number of seconds before next subtest is processed
            &SanUtil::nate_sleep($sleep_interval_time);
        }
    }

    # Call the NATE wait finish to make sure all subtests have properly exited
    # Grab a list of subtests for the nate_wait_finish package.
    my @stests = ();
    foreach my $stest (@{$subtests}) {
        push(@stests, $stest->{_subtest});
    }
    &SanUtil::nate_wait_finish(@stests);

    $_IntObject->sanlog(-exit => 1);

    return $SubIPC::TRUE;
}

###############################################################################
# Method:    process_sync
# Objective: Process a sync (if once occurred) of children
# Details:   See POD documentation at the end of this file.
###############################################################################
sub process_sync {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    # $sync_requested_by_all is a hash used to gather the different sync types
    # presented by the subtest list or what the user codedync_requested_by_all
    my $sync_requested_by_all;
    my $total_subtests_running = 0;

    &ParamsMgr::process_args(%args);

    my $subtests = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0, -default => &SubIPC::get_subtests());
    my $sync_type = &ParamsMgr::get_arg(-arg => 'sync_type', -required => 0, -default => "sync");
    my $send_message = &ParamsMgr::get_arg(-arg => 'send_message', -required => 0, -default => "continue");
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return $SubIPC::FALSE;
    }

    # Verify that some subtests have been passed in
    if (scalar(@{$subtests}) == 0) {
        $_IntObject->_set_err("No valid subtests have been provided");
        return $SubIPC::FALSE;
    }

    my $no_sync_count = 0;

    # Check to see if a sync was requested
    foreach my $subtest (@{$subtests}) {
        # Check to make sure that the subtest is still alive
        if ($subtest->{exit_status} !~ /fail|pass/i) {
            $total_subtests_running++;
            if ($subtest->{sync_type} ne "") {
                $sync_requested_by_all->{$subtest->{sync_type}}++;
            } else {
                $no_sync_count++;
            }
        }
    }

    # If none of the subtests that are currently running have requested to
    # synchronize then don't do anything.
    if ($total_subtests_running == $no_sync_count) {
        return $SubIPC::FALSE;
    }

    # If we have more then one sync type found waiting by children subtests
    # this is a failure, the caller needs to call with only the subtests it
    # expects to get a single sync type.
    if (scalar(keys(%{$sync_requested_by_all})) > 1) {
        $_IntObject->_set_error("Unable to process multiple sync points found on supplied subtests");
        return $SubIPC::FALSE;
    }

    $_IntObject->sanlog(-exit => 1);

    # Check to see if the sync_type requested by caller matches the sync type found
    # on the subtests AND if all subtests are under the same sync type
    if ($sync_requested_by_all->{$sync_type} != $total_subtests_running) {
        # We don't have a quorum yet, return false.
        $_IntObject->sanlog(-msg => "No sync point from provided subtests",
                            -msg_level => 1);

        return $SubIPC::FALSE;
    } else {
        # We have a sync quorum, need to relay message back to children that sync
        # has been achieved by all and that children should continue running
        foreach my $subtest (@{$subtests}) {
            # Check to make sure that the subtest is still alive
            if ($subtest->{exit_status} !~ /fail|pass/i) {
                $subtest->write(-message => $send_message);
            }

            # reset this subtest's sync type since we've completed this sync type
            # for all subtests
            $subtest->{sync_type} = "";
        }

        return $SubIPC::TRUE;
    }
}

###############################################################################
# Method:    stop
# Objective: Stop subtest(s)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub stop {
    my ($self, %args) = @_;

    $_IntObject->sanlog(-enter => 1);

    # Check to see if the first item passed in is a SubIPC object, if not
    # reread the %args hash from the passed in arguments.
    if (ref($self) !~ /SubIPC/i) {
        (%args) = @_;
    }

    my $subtests;

    &ParamsMgr::process_args(%args);

    my $subtest_list = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0, -default => &SubIPC::get_subtests());
    my $signal = &ParamsMgr::get_arg(-arg => 'signal', -required => 0, -default => "comment");
    my $force = &ParamsMgr::get_arg(-arg => 'force', -required => 0, -default => $SubIPC::FALSE);
    my $message = &ParamsMgr::get_arg(-arg => 'message', -required => 0, -default => "Subtest Complete");
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return $SubIPC::FALSE;
    }

    if (ref($self) =~ /SubIPC/i) {
        @{$subtests} = ($self);
    } else {
        @{$subtests} = @{$subtest_list};
    }

    # If the subtest list is empty and the object calling this is NOT a child (subtest)
    # return failure. If the subtest list is empty and the object calling this is a child,
    # place the object in the subtest list for processing.
    if (scalar(@{$subtests}) == 0) {
        $_IntObject->sanlog(-msg => "No subtests provided to stop, list is empty",
                            -msg_level => 3);
        return $SubIPC::FALSE;
    }

    # If the force flag is specified, then use '-fail' as signal type
    if ($force == $SubIPC::TRUE) {
        $_IntObject->sanlog(-msg => "Overriding requested signal ($signal) with (fail) because -force flag was specified",
                            -msg_level => 0);
        $signal = "fail";
    }

    # Go through each subtest and issue the requested signal and message
    foreach my $subtest (@{$subtests}) {
        $_IntObject->sanlog(-msg => "About to issue subtest kill on (" . $subtest->{_subtest}->{'runid'} . ") " .
                                    "with signal ($signal) and message ($message)",
                            -msg_level => 3);
        $subtest->{_subtest}->kill(message=>[$signal, $message]);
    }

    # Call the NATE wait finish to make sure all subtests have properly exited
    # Grab a list of subtests for the nate_wait_finish package.
    my @stests = ();
    foreach my $stest (@{$subtests}) {
        push(@stests, $stest->{_subtest});
    }

    $_IntObject->sanlog(-exit => 1);

    return $SubIPC::TRUE;
}

###############################################################################
# Method:    fail
# Objective: Issue fail message to parent and hold for parent to handle child
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fail {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    &ParamsMgr::process_args(%args);

    my $message = &ParamsMgr::get_arg(-arg => 'msg', -required => 0, -default => "");
    my $object = &ParamsMgr::get_arg(-arg => 'object', -required => 0, -default => undef);
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return $SubIPC::FALSE;
    }

    # If the object is defined, see if there is an error get the error and append it to the
    # message to be sent to parent.
    if (defined($object)) {
        if ($object->can("error")) {
            if ($object->error()) {
                my @error = $object->errmsg();
                if (length($message)) {
                    $message .= "\n" . join("\n", @error);
                } else {
                    $message = join("\n", @error);
                }
            }
        }
    }

    # Set childs state to fail so that when the child exits the logs will turn red
    # in the online log viewer.
    &SanUtil::nate_log('-fail', "$message");

    &SubIPC::write(-msg => $message, -type => "fail");
    if (&SubIPC::error()) {
        # Notify the users that something has gone wrong with IPC
        &SanUtil::nate_log('-comment', "Unable to properly write message ($message) to parent: " . &SubIPC::get_error());
        # Clear the error on the internal package and continue, worst case the test blows up
        # but if IPC is having a problem, this is the least of the users problems.
        &SubIPC::clear_error();
    }


    $_IntObject->sanlog(-exit => 1);

    # block on eventloop drains, this will allow the parent to reap us
    while (1) {
        NATE::Events::eventloop_drain();
    }

    return $SubIPC::TRUE;
}

###############################################################################
# Method:    pass
# Objective: Issue pass message to parent and hold for parent to handle child
# Details:   See POD documentation at the end of this file.
###############################################################################
sub pass {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    &ParamsMgr::process_args(%args);

    my $message = &ParamsMgr::get_arg(-arg => 'msg', -required => 0, -default => "Passed");
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
    }

    &SubIPC::write(-msg => $message, -type => "pass");
    if (&SubIPC::error()) {
        # Notify the users that something has gone wrong with IPC
        &SanUtil::nate_log('-comment', "Unable to properly write message ($message) to parent: " . &SubIPC::get_error());
        # Clear the error on the internal package and continue, worst case the test blows up
        # but if IPC is having a problem, this is the least of the users problems.
        &SubIPC::clear_error();
    }

    # Set childs state to pass so that when the child exits the logs will turn green
    # in the online log viewer.
    &SanUtil::nate_log('-pass', "$message");

    $_IntObject->sanlog(-exit => 1);

    # block on eventloop drains, this will allow the parent to reap us
    while (1) {
        NATE::Events::eventloop_drain();
    }

    return $SubIPC::TRUE;
}

###############################################################################
# Method:    sync
# Objective: Issue sync message to parent and hold for parent to respond
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sync {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    &ParamsMgr::process_args(%args);

    my $message = &ParamsMgr::get_arg(-arg => 'msg', -required => 0, -default => "Syncing");
    my $sync_type = &ParamsMgr::get_arg(-arg => 'sync_type', -required => 0, -default => "sync");
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
    }

    # Write message to parent with sync type of $sync_type
    &SubIPC::write(-msg => $message, -type => "sync", -sync_type => $sync_type);
    if (&SubIPC::error()) {
        # Notify the users that something has gone wrong with IPC
        &SanUtil::nate_log('-comment', "Unable to properly write message ($message) to parent: " .
                                       &SubIPC::get_error());
        # Clear the error on the internal package and continue, worst case the test blows up
        # but if IPC is having a problem, this is the least of the users problems.
        &SubIPC::clear_error();
    }

    &SanUtil::nate_log('-comment', "Synchronizing with parent subtest...");

    # Start a blocking read call, this allows the parent time to handle the
    # request to sync and respond
    # Pass in an empty subtest list so that this only communicates to
    # the parent of this subtest, not any children this subtest may have
    &SubIPC::read(-block => $SubIPC::TRUE,
                  -subtest => []);

    &SanUtil::nate_log('-comment', "Continuing");

    $_IntObject->sanlog(-exit => 1);

    return $SubIPC::TRUE;
}

###############################################################################
# Method:    is_alive
# Objective: check to see if subtest is alive
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_alive {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    &ParamsMgr::process_args(%args);

    my $subtests = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0, -default => &SubIPC::get_subtests());
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return $SubIPC::FALSE;
    }

    # Verify that some subtests have been passed in
    if (scalar(@{$subtests}) == 0) {
        $_IntObject->_set_err("No valid subtests have been provided");
        return $SubIPC::FALSE;
    } elsif (scalar(@{$subtests}) > 1) {
        $_IntObject->_set_err("This function does not support more then one subtest passed in at a time");
        return $SubIPC::FALSE;
    }

    # Grab the first subtest passed in on the list
    my $subtest = ${$subtests}[0];

    $_IntObject->sanlog(-exit => 1);

    # Check to see if the subtest is alive or not
    if ($subtest->{_subtest}->isAlive()) {
        return $SubIPC::TRUE;
    } else {
        return $SubIPC::FALSE;
    }
}

###############################################################################
# Method:    get_subtests
# Objective: return array of stored subtests on internal object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_subtests {

    $_IntObject->sanlog(-msg => "Total number of subtests defined on internal object: " .
                                scalar(@{$_IntObject->{_subtests}}),
                        -msg_level => 3);

    return $_IntObject->{_subtests};
}

###############################################################################
# Method:    get_subtest_status
# Objective: returns count of passed, failed, and running subtests
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_subtest_status {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    my ($status);
    $status->{passed} = 0;
    $status->{failed} = 0;
    $status->{running} = 0;

    &ParamsMgr::process_args(%args);

    my $subtests = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0, -default => &SubIPC::get_subtests());
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # Verify that some subtests have been passed in
    if (scalar(@{$subtests}) == 0) {
        $_IntObject->_set_err("No subtests provided to process.");
        return {};
    }

    # Go through each subtest and count up passed, failed and running
    foreach my $subtest (@{$subtests}) {
        # If no status is defined then the subtest is still running
        if (not defined($subtest->{exit_status})) {
            $status->{running}++;
        } elsif ($subtest->{exit_status} =~ /pass/i) {
            $status->{passed}++;
        } elsif ($subtest->{exit_status} =~ /fail/i) {
            $status->{failed}++;
        } else {
            # Unknown status
            $status->{$subtest->{exit_status}}++;
        }
    }

    $_IntObject->sanlog(-exit => 1);

    # Return pass count, then fail count, and last is running count
    return $status;
}

###############################################################################
# Method:    get_messages
# Objective: returns array of messages matching type or child id as requested
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_messages {
    my (%args) = @_;

    my @messages;

    $_IntObject->sanlog(-enter => 1);

    &ParamsMgr::process_args(%args);

    my $subtests = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0, -default => &SubIPC::get_subtests());
    my $type = &ParamsMgr::get_arg(-arg => 'type', -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return @messages;
    }

    # Verify that some subtests have been passed in
    if (scalar(@{$subtests}) == 0) {
        $_IntObject->_set_err("No subtests provided to process.");
        return @messages;
    }

    # Get the list of subtest id's to grab messages for
    foreach my $subtest (@{$subtests}) {
        if (defined($_IntObject->{'_child_messages'}->{$subtest->{_subtest}->{'uid'}})) {
            # Go through the messages for the subtest and get a list of messages that match the requested type
            foreach my $time (keys(%{$_IntObject->{'_child_messages'}->{$subtest->{_subtest}->{'uid'}}})) {
                # Check to see if the type matches requested, if so push on return array
                if (defined($_IntObject->{'_child_messages'}->{$subtest->{_subtest}->{'uid'}}->{$time}->{$type})) {
                    push(@messages, $_IntObject->{'_child_messages'}->{$subtest->{_subtest}->{'uid'}}->{$time}->{$type});
                }
            }
        }
    }

    $_IntObject->sanlog(-exit => 1);

    return @messages;
}

###############################################################################
# Method:    remove_subtest
# Objective: Removes specified list of subtests from internal subtest list
#            Returns true when completed, false when failure to remove/locate
#            specified subtest.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_subtest {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    &ParamsMgr::process_args(%args);

    my $subtests = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0, -default => &SubIPC::get_subtests());
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return $SubIPC::FALSE;
    }

    # Verify that some subtests have been passed in
    if (scalar(@{$subtests}) == 0) {
        return $SubIPC::FALSE;
    }

    # Go through each subtest and count up passed, failed and running
    my $status = &SubIPC::get_subtest_status(-subtest => $subtests);
    if (&SubIPC::error()) {
        return $SubIPC::FALSE;
    }

    # If the subtests are running set error and let caller know that they have to be stopped first.
    if ($status->{running} != 0) {
        $_IntObject->_set_err("Unable to remove requested subtests, A total of (" . $status->{running} .
                              ") are still running.");
        return $SubIPC::FALSE;
    }

    # Grab total number of subtests requested to be removed
    my $counter = scalar(@{$subtests});

    # If all subtests requested equal all internal stored subtests simply clear out internal array
    if ($counter == scalar(@{$_IntObject->{_subtests}})) {
        @{$_IntObject->{_subtests}} = ();
        $counter = 0;
    }

    while ($counter != 0) {
        # Shift off one subtest to process
        my $int_subtest = shift(@{$_IntObject->{_subtests}});

        my $found = $SubIPC::FALSE;
        # Go through the caller supplied list and determine if any of those subtests match
        # $int_subtest, if it does set $found and exit the loop
        foreach my $subtest (@{$subtests}) {
            # Does the internal uid match the requested subtest uid?
            if ($int_subtest->{_subtest}->{'uid'} eq $subtest->{_subtest}->{'uid'}) {
                $found = $SubIPC::TRUE;
                last;
            }
        }

        # If the subtest was found in the list of to be removed, don't add it back to the internal
        # list, just leave it off.
        if ($found == $SubIPC::TRUE) {
            $_IntObject->sanlog(-msg => "Removing subtest (" . $int_subtest->{_subtest}->{'runid'} . ") " .
                                        "from internal list.",
                                -msg_level => 1);
            $counter--;
        } else {
            # It does not, push back on list
            push(@{$_IntObject->{_subtests}}, $int_subtest);
        }
    }

    $_IntObject->sanlog(-exit => 1);

    return $SubIPC::TRUE;
}

###############################################################################
# Method:    read
# Objective: Read message from parent or child subtest
# Details:   See POD documentation at the end of this file.
###############################################################################
sub read {
    my ($self, %args) = @_;

    $_IntObject->sanlog(-enter => 1);

    # Check to see if the first item passed in is a SubIPC object, if not
    # reread the %args hash from the passed in arguments.
    if (ref($self) !~ /SubIPC/i) {
        (%args) = @_;
    }

    my $data;
    my $subtests;

    &ParamsMgr::process_args(%args);

    my $block = &ParamsMgr::get_arg(-arg => 'block', -required => 0, -default => $SubIPC::TRUE);
    my $subtest_list = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0, -default => &SubIPC::get_subtests());
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        $data->{_success_read} = $SubIPC::FALSE;
        return $data;
    }

    # Populate the subtests array of the ref type of self is SubIPC, this means that the
    # parent wants to read from a child.
    if (ref($self) =~ /SubIPC/i) {
        # parent reading from child
        @{$subtests} = ($self);
    } else {
        @{$subtests} = @{$subtest_list};
    }

    # Determine if this is a parent or child making the call.
    # If NOT ref type of SubIPC and subtests are empty, this could be a child talking to parent
    if (scalar(@{$subtests}) == 0 and ref($self) !~ /SubIPC/i) {
        $_IntObject->sanlog(-msg => "Child->Parent - About to execute read_obj (Blocking: $block)",
                            -msg_level => 3);

        # This is a child trying to communicate with parent
        # check if we can read from the parent (let the can_read method handle any blocking)
        my $can_read = &SubIPC::can_read(block    => $block,
                                         subtest => []);
        if ($_IntObject->error()) {
            $_IntObject->_set_err("Failed to can_read()");
            $data->{_success_read} = $SubIPC::FALSE;
            return $data;
        }
        if ($can_read eq $SubIPC::TRUE) {
            # Found a message to read from the parent, read it
            $data = shift(@{$_IntObject->{_message_queue}});
            $data->{_success_read} = $SubIPC::TRUE;
        } else {
            # No message found from parent, set internal flag to false
            $data->{_success_read} = $SubIPC::FALSE;
        }
    } else {
        # Grab the first subtest off of the stack for processing
        my $subtest = ${$subtests}[0];
        my $runid = $subtest->{_subtest}->{'runid'};

        $_IntObject->sanlog(-msg => "Parent->Child(" . $runid . ") - About " .
                                    "to execute read_obj (Blocking: $block)",
                            -msg_level => 3);

        # This is a parent trying to communicate with a child
        # check if we can read from the child (let the can_read method handle any blocking)
        my $can_read = $subtest->can_read(block => $block);
        if ($_IntObject->error()) {
            $_IntObject->_set_err("Failed to can_read() from child subtest '$runid'");
            $data->{_success_read} = $SubIPC::FALSE;
            return $data;
        }
        if ($can_read eq $SubIPC::TRUE) {
            # Found a message to read from the child, read it
            $data = shift(@{$subtest->{_message_queue}});
            $data->{_success_read} = $SubIPC::TRUE;
        } else {
            # No message found from the child, set internal flag to false
            $data->{_success_read} = $SubIPC::FALSE;
        }
    }

    $_IntObject->sanlog(-exit => 1);

    return $data;
}

###############################################################################
# Method:    can_read
# Objective: Check to see if child or parent has message to be read
# Details:   See POD documentation at the end of this file.
###############################################################################
sub can_read {
    my ($self, %args) = @_;

    $_IntObject->sanlog(-enter => 1);

    # Check to see if the first item passed in is a SubIPC object, if not
    # reread the %args hash from the passed in arguments.
    if (ref($self) !~ /SubIPC/i) {
        (%args) = @_;
    }

    my $retval = $SubIPC::FALSE;
    my $subtests;

    &ParamsMgr::process_args(%args);

    my $subtest_list = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0, -default => &SubIPC::get_subtests());
    my $block = &ParamsMgr::get_arg(-arg => 'block', -default => $SubIPC::FALSE);
    my $poll_time = &ParamsMgr::get_arg(-arg => 'poll_time', -default => 0.1);

    if (&ParamsMgr::error()) {
        $_IntObject->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $SubIPC::FALSE;
    }

    if (ref($self) =~ /SubIPC/i) {
        @{$subtests} = ($self);
    } else {
        @{$subtests} = @{$subtest_list};
    }

    # Determine if this is a parent or child making the call.
    if (scalar(@{$subtests}) == 0 and ref($self) !~ /SubIPC/i) {
        $_IntObject->sanlog(-msg => "Child->Parent - About to execute can_read",
                            -msg_level => 3);

        # first check to see if there is a message sitting on the child subtest's message queue
        if (exists($_IntObject->{_message_queue}) and scalar(@{$_IntObject->{_message_queue}})) {
            $_IntObject->sanlog(-exit => 1);
            # there is something to read, return true
            return $SubIPC::TRUE;
        }

        # See if there is a message waiting to be read from the parent
        my $msg;
        if ($block == $SubIPC::TRUE) {
            # blocking read
            # wait for message_get to return a SubIPC message to use from
            # the parent NATE::STAF::Peer
            eval {
                if ($msg = $Parent->message_get(filter => sub { my $m = shift; return $m->{type} =~ /SubIPC/i; })) {
                    $retval = $SubIPC::TRUE;
                }
            };
            if ($@) {
                # failed to read from the parent
                $retval = $SubIPC::FALSE;
            }
        } else {
            # non blocking read
            # wait no more than 100ms for a SubIPC message to arrive from
            # the parent NATE::STAF::Peer
            eval {
                if ($msg = $Parent->message_get(filter => sub { my $m = shift; return $m->{type} =~ /SubIPC/i; },
                                                timeout => $poll_time,
                                                on_timeout => sub { return 0; })) {
                    $retval = $SubIPC::TRUE;
                }
            };
            if ($@) {
                # failed to read from the parent
                $retval = $SubIPC::FALSE;
            }
        }

        # store this message on the internal message queue of the IntObject
        # this queue represents messages sent to us from our parent
        if ($retval == $SubIPC::TRUE) {
            push(@{$_IntObject->{_message_queue}}, $msg);
        }
    } else {
        if (exists(${$subtests}[0])) {
            # Grab the first subtest off of the stack for processing
            my $subtest = ${$subtests}[0];
            $_IntObject->sanlog(-msg => "Parent->Child(". $subtest->{_subtest}->{'runid'} . ") " .
                                        " - About to execute can_read",
                                -msg_level => 3);

            # first check to see if there is a message sitting on the child subtest's message queue
            if (exists($subtest->{_message_queue}) and scalar(@{$subtest->{_message_queue}})) {
                $_IntObject->sanlog(-exit => 1);
                # there is something to read, return true
                return $SubIPC::TRUE;
            }

            # See if there is a message waiting to be read from the child
            my $msg;
            if ($block == $SubIPC::TRUE) {
                # blocking read
                # wait for message_get to return a SubIPC message to use from
                # the parent NATE::STAF::Peer
                eval {
                    if ($msg = $subtest->{_subtest}->message_get(filter => sub { my $m = shift; return $m->{type} =~ /SubIPC/i; })) {
                        $retval = $SubIPC::TRUE;
                    }
                };
                if ($@) {
                    # failed to read from the child
                    $retval = $SubIPC::FALSE;
                }
            } else {
                # non blocking read
                # wait no more than 100ms for a SubIPC message to arrive from
                # the parent NATE::STAF::Peer
                eval {
                    if ($msg = $subtest->{_subtest}->message_get(filter => sub { my $m = shift; return $m->{type} =~ /SubIPC/i; },
                                                                 timeout => $poll_time,
                                                                 on_timeout => sub { return 0; })) {
                        $retval = $SubIPC::TRUE;
                    }
                };
                if ($@) {
                    # failed to read from the child
                    $retval = $SubIPC::FALSE;
                }
            }

            # store this message on the internal message queue of this subtest object
            # this queue represents messages sent to us from this child subtest
            if ($retval == $SubIPC::TRUE) {
                push(@{$subtest->{_message_queue}}, $msg);
            }
        }
    }

    $_IntObject->sanlog(-exit => 1);

    return $retval;
}

###############################################################################
# Method:    write
# Objective: Write message to parent or child subtest
# Details:   See POD documentation at the end of this file.
###############################################################################
sub write {
    my ($self, %args) = @_;

    $_IntObject->sanlog(-enter => 1);

    # Check to see if the first item passed in is a SubIPC object, if not
    # reread the %args hash from the passed in arguments.
    if (ref($self) !~ /SubIPC/i) {
        (%args) = @_;
    }

    %args = &SanUtil::strip_dash(%args);

    my $subtests;

    &ParamsMgr::process_args(%args);

    my $msg = &ParamsMgr::get_arg(-arg => 'msg', -required => 1, -default => "");
    my $type = &ParamsMgr::get_arg(-arg => 'type', -required => 0, -default => "message");
    my $sync_type = &ParamsMgr::get_arg(-arg => 'sync_type', -required => 0, -default => "");
    my $subtest_list = &ParamsMgr::get_arg(-arg => 'subtest', -required => 0,
                                           -default => &SubIPC::get_subtests());
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return $SubIPC::FALSE;
    }

    if (ref($self) =~ /SubIPC/i) {
        @{$subtests} = ($self);
    } else {
        @{$subtests} = @{$subtest_list};
    }

    if (scalar(@{$subtests}) == 0) {
        $_IntObject->sanlog(-msg => "About to write message to parent",
                            -msg_level => 3);
    }

    # Define the data structure to be sent to parent/child during write_obj
    my $data;
    $data->{message} = $msg;
    $data->{type} = $type;
    $data->{sync_type} = $sync_type;
    $data->{caller} = (caller(2))[3];
    $data->{time} = time();
    $data->{extra} = \%args;

    $_IntObject->sanlog(-msg => "About to write message to parent\n" . Dumper($data),
                        -msg_level => 3);

    # Determine if this is a parent or child making the call.
    if (ref($self) !~ /SubIPC/i) {
        $_IntObject->sanlog(-msg => "Child->Parent - About to execute write_obj",
                            -msg_level => 3);
        # This is a child wanting to write to parent
        $Parent->message_put(type => "SubIPC/" . $type,
                             message => $data);
    } else {
        # This is a parent wanting to write to its children
        # Simply loop through the subtests and write message to each subtest
        foreach my $subtest (@{$subtests}) {
            $_IntObject->sanlog(-msg => "Parent->Child(" . $subtest->{_subtest}->{'runid'} . ") " .
                                        "About to execute write_obj",
                                -msg_level => 3);
            # If for some reason an element is found in the list to not be defined
            # skip it so that we don't get a syntax error.
            if (not defined($subtest)) {
                next;
            }

            # write the message data to the child
            $subtest->{_subtest}->message_put(type => "SubIPC/" . $type,
                                              message => $data);
        }
    }

    $_IntObject->sanlog(-exit => 1);

    return $SubIPC::TRUE;
}

###############################################################################
# Method:    sanlog
# Objective: Wrapper to SanLog::sanlog() function
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sanlog {
    my ($self, %args) = @_;

    my ($msg_type, $msg, $msg_level, $enter, $exit);

    &ParamsMgr::process_args(%args);
    $enter = &ParamsMgr::get_arg(-arg => "enter", -required => 0, -default => 0);
    $exit = &ParamsMgr::get_arg(-arg => "exit", -required => 0, -default => 0);
    if ($enter == 1 or $exit == 1) {
        # If enter or exit is either specified, msg is not required, if not defined
        # default it to an empty string
        $msg = &ParamsMgr::get_arg(-arg => "msg", -required => 0, -default => "");
    } else {
        $msg = &ParamsMgr::get_arg(-arg => "msg", -required => 1, -default => "");
    }

    $msg_type = &ParamsMgr::get_arg(-arg => "msg_type", -required => 0, -default => "");
    $msg_level = &ParamsMgr::get_arg(-arg => "msg_level", -required => 0, -default => 0);
    if (&ParamsMgr::error()) {
        return $SubIPC::FALSE;
    }

    # If the message is empty that could mean that the user specified an enter or exit flag
    # default it to _parent_caller
    if ($msg eq "") {
        $msg = $_IntObject->_parent_caller();
    } else {
        # Prepend the caller name/description to the msg
        $msg = $_IntObject->_parent_caller() . ": " . $msg;
    }

    # Determine if the caller called with an enter or exit request
    # Also set the msg_level to 3 if this is an enter/exit request
    if ($enter == 1) {
        $msg = "[Function Enter] " . $msg;
        $msg_level = 3;
    } elsif ($exit == 1) {
        $msg = "[Function Exit] " . $msg;
        $msg_level = 3;
    }

    # If the message type is warn then call SanUtil::nate_log with
    # warning type and message
    if ($msg_type =~ /warn/i) {
        &SanUtil::nate_log('-warn', $msg);
        return $SubIPC::TRUE;
    }

    # Call sanlog with the msg and msg_level
    $self->SanLog::sanlog(-msg => $msg, -msg_level => $msg_level);

    return $SubIPCr::TRUE;
}

###############################################################################
# Method:    debug_level
# Objective: Set the debug level and return it to caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub debug_level {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    &ParamsMgr::process_args(%args);
    my $debug_level = &ParamsMgr::get_arg(-arg => "debug_level", -required => 0, -default => 0);
    if (&ParamsMgr::error()) {
        return 0;
    }

    $_IntObject->{debug_level} = $debug_level;
    $_IntObject->sanlog(-msg => "Setting debug level to ($debug_level)",
                        -msg_level => 3);

    $_IntObject->sanlog(-exit => 1);

    return $debug_level;
}

###############################################################################
# Method:    _new
# Objective: Create new SubIPC object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($proto, %args) = @_;

    my $self = {};

    my $class = ref($proto) || $proto;
    bless ($self, $class);

    # Set attributes on self for Error.pm
    #   error_dont_overwrite_errors - Append errors until cleared.
    #   error_stack_style - 'Normal' means place first error at top of stack
    $self->{error_dont_overwrite_errors} = 1;
    $self->{error_stack_style} = "normal";

    # Init the run ID prefix
    $self->{run_id_prefix} = "";

    # Init the subtest prefix count
    $self->{subtest_count} = 1;

    # Clear out errors
    @{$self->{errors}} = ();

    # Clear our subtest reference
    @{$self->{_subtests}} = ();

    return $self;
}

###############################################################################
# Method:    _set_err
# Objective: Save error on internal error stack and Error package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _set_err {
    my ($self, @msg) = @_;

    return $_IntObject->Error::_set_err(@msg);
}

###############################################################################
# Method:    _clear_err
# Objective: If user calls, reroute to clear_error();
# Details:   None.
###############################################################################
sub _clear_err {

    return $_IntObject->clear_error();
}

###############################################################################
# Method:    error
# Objective: Return error flag on SubIPC object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub error {

    return $_IntObject->Error::error();
}

###############################################################################
# Method:    errmsg
# Objective: Return the errors string on the SubIPC object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub errmsg {

    return $_IntObject->Error::errmsg();
}

###############################################################################
# Method:    get_error
# Objective: Return the errors string on the SubIPC object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_error {

    return $_IntObject->errmsg();
}

###############################################################################
# Method:    clear_error
# Objective: Clear the error string(s) set on the SubIPC object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clear_error {

    return $_IntObject->_clear_err();
}

###############################################################################
# Method:    set_run_id_prefix
# Objective: Set the run_id_prefix to a given string
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_run_id_prefix {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);

    &ParamsMgr::process_args(%args);
    my $prefix = &ParamsMgr::get_arg(-arg => "prefix", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err(&ParamsMgr::get_error());
        return $SubIPC::FALSE;
    }

    $_IntObject->{run_id_prefix} = $prefix;
    $_IntObject->sanlog(-msg => "Setting run_id_prefix to ($prefix)",
                        -msg_level => 3);

    $_IntObject->sanlog(-exit => 1);

    return $prefix;
}

###############################################################################
# Method:    get_run_id_prefix
# Objective: Retreive the run_id_prefix to a given string
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_run_id_prefix {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);
    $_IntObject->sanlog(-exit  => 1);

    return $_IntObject->{run_id_prefix};
}

###############################################################################
# Method:    clear_run_id_prefix
# Objective: Set the run_id_prefix to a given string
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clear_run_id_prefix {
    my (%args) = @_;

    $_IntObject->sanlog(-enter => 1);
    $_IntObject->{run_id_prefix} = "";
    $_IntObject->sanlog(-exit  => 1);
    return $_IntObject->{run_id_prefix};
}

###############################################################################
# Method:    _next_subtest_count
# Objective: Helper method that returns a string such as 0001 or 0002 that is
#            a subtest count.  It will also increment $self->{subtest_count}
#            This is here so we can add an incrementing prefix to the top
#            level subtest run IDs.
# Details:   No inputs.  Ouput is a string, that is 3 digits that represents
#            the current subtest count.
###############################################################################
sub _get_next_subtest_count {

    my $subtest_count = $_IntObject->{subtest_count};
    my $return_string = $subtest_count;

    # Pad the subtest count with zero's for values less than 100
    $return_string = "00" . $subtest_count if (length($subtest_count) == 1);
    $return_string = "0"  . $subtest_count if (length($subtest_count) == 2);

    # Increment the subtest count
    $_IntObject->{subtest_count}++;
    return $return_string;
}

################################################################################
#                             POD Documentation
################################################################################

=pod

=head1 Package Name

SubIPC.pm

=head1 Author

Mike Latif

=head1 Description

SubIPC provides a common set of libraries to work with IPC and subtests.
This library provides a wrapper around NATE's subtest and IPC calls. The
intent is to provide API's that enable easy subtest creation, subtest
management and communication to and from a subtest. A common set of API's
are provided for a parent/child to notify each other of changes in their states.

Note that SubIPC creates an object that is stored internally and abstracted from
the user.  When the this library is included (i.e. used) by another library or
script, this internal object is created.  This is done to aid in the management
of subtests (by handle).  User's, however, will call simple API's
that appear to be procdural.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>
B<SanLog.pm>

=head1 Public Methods

=head2 B<create()>

Creates a new subtest.

=over 4

=item Inputs

=item * -id

Run ID.  String that is an identifier for the subtest.  Note that this ID is
used to name the directory that is created to hold the logs for this subtest.
If a prefix has been defined (see set_run_id_prefix(), get_run_id_prefix(),
clear_run_id_prefix()) then the ID will be the prefix followed by the
given string.

=item * -run

What to run in your subtest.  See the example below.

=item * I<-bg> (defaults to "1")

Background argument passed to Subtest.  If bg is set, subtest will create a
non-blocking subtest (i.e. it won't wait for the subtest to complete).  By
default, we will run in the background.

=item * I<-ipc> (defaults to "0")

ipc argument passed to Subtest.

=item * I<-esc> (defaults to "0")

Enable Subtest Count, if enabled, a 3 digit number is prepended to subtest run_id.

=item Outputs

Returns a subtest object.

=item Examples

    # Create a subtest
    my $subtest = &SubIPC::create(-id  => "Controller-setup",
                                  -run => sub {
                                                  $self->{sm_controller_object}->setup();
                                              },
                                  -ipc => $SanMgr::TRUE,
                                  -bg  => $SanMgr::TRUE);
    if (&SubIPC::error()) {
        $self->_set_err(&SubIPC::get_error());
        return $SanMgr::FALSE;
    }

    &SubIPC::wait_finish(-subtest => [$subtest]);
    if (&SubIPC::error()) {
        $self->_set_err("Unable to wait for Controller Setup subtest\n\t" .
                        &SubIPC::get_error());
        return $SanMgr::FALSE;
    }

=back

=head2 B<set_run_id_prefix(%args)>

Store a run ID prefix.  This will be used by the create() function.
The given string will be prepended to the subtest run ID.

=over 4

=item Inputs

=item * -prefix

String to prepend to subtest run ID.

=item Outputs

No value is returned.  There are no error conditions.

=item Examples

    use SubIPC;

    # Create two subtests: initial_setup and secondary_setup
    &SubIPC::set_run_id_prefix(-prefix => "initial_");
    # Create a subtest
    my $subtest = &SubIPC::create(-id  => "setup",
                                  -run => sub {
                                                  $self->setup(-phase => "inital");
                                              },
                                  -ipc => $SanMgr::TRUE,
                                  );

    &SubIPC::set_run_id_prefix(-prefix => "secondary_");
    # Create a subtest
    my $subtest = &SubIPC::create(-id  => "setup",
                                  -run => sub {
                                                  $self->setup(-phase => "secondary");
                                              },
                                  -ipc => $SanMgr::TRUE,
                                  );

=back

=head2 B<get_run_id_prefix()>

Retrieve the current run ID prefix.

=over 4

=item Outputs

Returns the current value of the run ID prefix.  There are no error conditions.
Note that if the prefix if undef, that's what will be returned.

=item Examples

    use SubIPC;

    # Create a bunch subtests with ever increasing run ids

    for (my $i = 1; $i < $max; $i++) {
        my $current_id = &SubIPC::get_run_id_prefix() . $i;
        # Create a subtest
        my $subtest = &SubIPC::create(-id  => "setup",
                                      -run => sub {
                                                      $self->setup(-phase => "inital");
                                                  },
                                      -ipc => $SanMgr::TRUE,
                                     );
    } # end for

=back

=head2 B<clear_run_id_prefix()>

Set the run ID prefix to an empty string.

=over 4

=item Outputs

Returns the current value of the run ID prefix.  There are no error conditions.
Note that if the prefix if undef, that's what will be returned.

=item Examples

   # Clear the SubIPC run id prefix so SubIPC::create()
   # won't prepend to the run id
   &SubIPC::clear_run_id_prefix();

=back

=cut

1;
