# Hey emacs, this is basically -*- perl -*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/Supdiag.pm#1 $
#
# Copyright (c) 2011 Network Appliance, Inc.
# All rights reserved.
#
## @summary Helper class for using the support diagnostics web service.
## @author samuele@netapp.com
## @pod here

=head1 NAME

Supdiag

=head1 SYNOPSIS

 my $supdiag = Supdiag->new(command_interface => $node);
 $supdiag->enable_access(vserver => $vserver_name, role => $rbac_role);
 my $hello_uri_path = $supdiag->add_web_file(
                                    file_content => '<?php echo "hello\n" ?>',
                                    rel_dst_path => "hello.php");
 # make HTTP request to $hello_uri_path
 $supdiag->delete_web_files("hello.php");
 $supdiag->undo_access_changes();
 $supdiag->disable($vserver_name);

=head1 DESCRIPTION

Helper class for using the support diagnostics (supdiag) web service.
This simplifies enabling and configuring access to supdiag as well as
installing files for use as web resources.

=cut

package Supdiag;

use strict;
use Tharn;
use Data::Dumper;
use NACL::APISet;
use NACL::C::Node;
use NACL::C::Client;
use NACL::C::VserverServicesWeb;
use NACL::C::VserverServicesWebAccess;
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::MTask::SCP;
use File::Basename;

# Global Constants
use constant SUPDIAG     => "supdiag";             # Web service name
use constant DEV_URI_DIR => "/support/dev/mroot/"; # URI path for web resources
use constant DEV_DOC_DIR => "/mroot/etc/www/dev/"; # Filer path for web resources

# Web resource for making localhost requests
# See add_localhost_proxy() method below
use constant LOCAL_PROXY_SCRIPT   => "proxy.php";
use constant LOCAL_PROXY_URI_PATH => DEV_URI_DIR.LOCAL_PROXY_SCRIPT;


=head1 CONSTRUCTOR

=head2 new

Create a Supdiag object.

=over

=item Synopsis

 my $supdiag = Supdiag->new(command_interface => $node);

=item Arguments

=over

=item command_interface (optional)

A NACL::C::Node object to use as a NACL command interface. By default, the
constructor will get this by calling NACL::C::Node->find().

=back

=item Returns

A reference to a new Supdiag object.

=back

=head1 METHODS

=cut

sub new($;%) {
    my ($this, %args) = @_;

    my $class = ref($this) || $this;
    my $self = {
        ci => $args{command_interface},
        access => {
            created => [],
            deleted => [],
        }
    };
    $self->{ci} = NACL::C::Node->find() unless $self->{ci};

    bless($self, $class);
    return $self;
}

=head2 modify

This is a wrapper for NACL::C::VserverServicesWeb->modify(). It supplies the
command_interface argument from the Supdiag instance, and sets the name argument
to "supdiag". The caller must supply the rest of the arguments including vserver
and whatever fields are to be modified (enabled and/or ssl-only).

=over

=item Synopsis

 $supdiag->modify(vserver    => $vserver_name,
                  enabled    => "true"|"false",
                  "ssl-only" => "true"|"false");

=item Arguments

=over

=item vserver

Modify supdiag on this vserver.

=item enabled (optional)

Modify the "enabled" field. The value may be "true" or "false". If the ssl-only
argument is not supplied, this argument is mandatory.

=item ssl-only (optional)

Modify the "ssl-only" field. The value may be "true" or "false". If the enabled
argument is not supplied, this argument is mandatory.

=back

=item Returns

Nothing.

=back

=cut

sub modify($%) {
    my ($self, %args) = @_;
    NACL::C::VserverServicesWeb->modify(
                    command_interface => $self->{ci},
                    name => SUPDIAG,
                    %args);
}

=head2 enable

Enable the supdiag web service on each vserver in the argument list.

=over

=item Synopsis

 $supdiag->enable(@vserver_names);

=item Arguments

=over

=item

A list of vserver names.

=back

=item Returns

Nothing.

=back

=cut

sub enable($@) {
    my $self = shift;
    $self->modify(vserver => $_, enabled => "true") foreach @_;
}

=head2 disable

Disable the supdiag web service on each vserver in the argument list.

=over

=item Synopsis

 $supdiag->disable(@vserver_names);

=item Arguments

=over

=item

A list of vserver names.

=back

=item Returns

Nothing.

=back

=cut

sub disable($@) {
    my $self = shift;
    $self->modify(vserver => $_, enabled => "false") foreach @_;
}

=head2 permit_access

Permit a given RBAC role to access the supdiag web service on a given vserver.
If the 'vserver services web access' table does not already have a row that
permits this access, the row is created. The Supdiag instance tracks the access
rows that it creates so that these changes can be reverted with the
undo_access_changes() method.

=over

=item Synopsis

 $supdiag->permit_access(vserver => $vserver_name,
                         role    => $rbac_role);

=item Arguments

=over

=item vserver

The name of the vserver that the role is defined for. To permit access on node
vservers, this should be the name of the admin vserver (aka cserver).

=item role

The RBAC role that will be permitted access to supdiag.

=back

=item Returns

Nothing.

=back

=cut

sub permit_access($%) {
    my ($self, %args) = @_;

    $args{name} = SUPDIAG;

    try {
        NACL::C::VserverServicesWebAccess->find(
                    command_interface => $self->{ci},
                    filter => \%args);
    }
    catch NACL::Exceptions::NoElementsFound with {
        $Log->trace("Creating access to ".SUPDIAG." web service for ".
                    "role $args{role}, vserver $args{vserver}");
        my $sa = NACL::C::VserverServicesWebAccess->create(
                                command_interface => $self->{ci},
                                %args);
        push(@{$self->{access}->{created}}, $sa);
    };
}

=head2 enable_access

Enable the supdiag web service on a given vserver and permit access to a given
RBAC role. To enable supdiag on a node vserver, pass the node vserver name to
the vserver argument and the admin vserver name to the cserver argument.

This method calls the permit_access() method. See that documentation for more
information about changes to the access table.

=over

=item Synopsis

 $supdiag->enable_access(vserver  => $vserver_name,
                         role     => $rbac_role,
                         [cserver => $admin_vserver_name]);

=item Arguments

=over

=item vserver

Enable supdiag on this vserver.

=item role

The RBAC role that will be permitted access to supdiag.

=item cserver (optional)

The name of the vserver that the role is defined for. By default, the vserver
argument is used for this. When enabling supdiag on a node vserver, the name of
the admin vserver (aka cserver) should be provided here.

=back

=item Returns

Nothing.

=back

=cut

sub enable_access($%) {
    my ($self, %args) = @_;

    my $vserver = $args{vserver};
    my $cserver = $args{cserver} || $vserver;
    my $role    = $args{role};

    $self->enable($vserver);
    $self->permit_access(vserver => $cserver, role => $role);
}

=head2 undo_access_changes

Reverts changes to the 'vserver services web access' table that were made by the
permit_access() and enable_access() methods.

=over

=item Synopsis

 $supdiag->undo_access_changes();

=item Arguments

None.

=item Returns

Nothing.

=back

=cut

sub undo_access_changes($) {
    my $self = shift;

    while (my $sa = shift(@{$self->{access}->{created}})) {
        $Log->trace("Delete access to ".SUPDIAG." web service for ".
                    "role $sa->{role}, vserver $sa->{vserver}");
        $sa->delete();
    }
}

################################################################################
# Private method:   _exec_shell
# Parameters:       Command line to execute.
# Returns:          Output from the command.
# Purpose:          Helper method for running commands on the filer's system
#                   shell.
################################################################################
sub _exec_shell($$) {
    my $self = shift;
    my $command = shift;

    if (!$self->{system}) {
        $self->{system} = NACL::APISet->new(
                            hostobj   => $self->{ci}->hostrec(),
                            category  => "Node",
                            interface => "CLI",
                            set       => "Systemshell");
    }

    return $self->{system}->execute_raw_command(command => $command);
}

=head2 add_web_file

Create a file as a web resource in the supdiag web service.

=over

=item Synopsis

 my $uri_path = $supdiag->add_web_file(
                             rel_dst_path => "relative/path/to/destination",
                             file_content => $contents_of_file_to_add);

=item Arguments

=over

=item rel_dst_path

The relative path to the destination file. The full path is created by appending
this to Supdiag::DEV_DOC_DIR. The URI path for this file Supdiag::DEV_URI_DIR
concatenated with this argument.

=item file_content

A string contain the contents of the file to be installed in supdiag. The
file will be created on the filer by echoing this string in the system shell and
redirecting the output to the destination file.

=back

=item Returns

The URI path of the new resource, or 0 on failure. The URI path will be
Supdiag::DEV_URI_DIR appended with the rel_dst_path argument.

=back

=cut

sub add_web_file($%) {
    my ($self, %args) = @_;

    my $dst_path = DEV_DOC_DIR.$args{rel_dst_path};
    my $uri_path = DEV_URI_DIR.$args{rel_dst_path};

    my $command = "echo '${args{file_content}}' > $dst_path";
    my $out = $self->_exec_shell($command);
    if ($out) {
        $Log->error("Failed to create test script: $out");
        return 0;
    }

    $Log->comment("Created file $dst_path");

    return $uri_path;
}

=head2 copy_web_file

Copy a file to the filer as a web resource in the supdiag web service.

=over

=item Synopsis

 my $uri_path = $supdiag->copy_web_file(
                             src_file     => "/local/path/to/source/file",
                             [rel_dst_dir => "relative/destination/directory"]);

=item Arguments

=over

=item rel_dst_dir

The destination directory relative to Supdiag::DEV_DOC_DIR. The basename of the
destination file will be the same as the basename of src_file. The URI path for
this file Supdiag::DEV_URI_DIR concatenated with this argument, concatenated
with the basename of src_file. By default, this will be ".", so the file will
be copied directly to Supdiag::DEV_URI_DIR.

=item src_file

The local path to the file to be installed in supdiag. When this argument is
used, this file will be copied to the filer using scp.

=back

=item Returns

The URI path of the new resource, or 0 on failure. The URI path will be
Supdiag::DEV_URI_DIR appended with the rel_dst_path argument.

=back

=cut

sub copy_web_file($%) {
    my ($self, %args) = @_;

    my $src_dir = dirname($args{src_file});
    my $file    = basename($args{src_file});

    my $dst_dir = DEV_DOC_DIR;
    $dst_dir .= $args{rel_dst_dir} if $args{rel_dst_dir};

    my $uri_path = DEV_URI_DIR.$args{rel_dst_dir}."/".$file;

    if (! -f $args{src_file}) {
        $Log->error("Source file ${args{src_file}} does not exist");
        return 0;
    }

    my $CLIENT;
    param("CLIENT", "-mesg", "client used for this test");
    my $client_ci = NACL::C::Client->find();

    my $scp = NACL::MTask::SCP->new(source_command_interface => $client_ci,
                                    set_directory_on_source  => $src_dir,
                                    target_command_interface => $self->{ci},
                                    set_directory_on_target  => $dst_dir);

    $scp->copy_from_source_to_target(files_to_copy => [$file]);

    $Log->comment("Copied file $args{src_file} to $dst_dir");

    return $uri_path;
}

=head2 delete_web_files

Delete web resource files from the supdiag web service.

=over

=item Synopsis

 $supdiag->delete_web_files(@relative_paths);

=item Arguments

=over

=item

An array of file paths relative to Supdiag::DEV_DOC_DIR. Use the same strings
that were passed to the rel_dst_path argument of add_web_file() to remove files
that were previously added.

=back

=item Returns

1 on success, 0 on failure.

=back

=cut

sub delete_web_files($@) {
    my $self = shift;
    my $status = 1;
    foreach (@_) {
        my $script = DEV_DOC_DIR.$_;
        my $out = $self->_exec_shell("rm -f $script");
        if ($out) {
            $Log->error("Failed to delete test script '$script':\n$out");
            $status = 0;
        }

        $Log->comment("Deleted file $script");
    }

    return $status;
}

=head2 add_local_proxy

Installs a supdiag web resource for proxying requests to the local filer, e.g.
to test the localhost interface. This copies the Supdiag::LOCAL_PROXY_SCRIPT
file to the filer. That file is expected to be in the same directory as the
Supdiag.pm file. See the comments at the top of the Supdiag::LOCAL_PROXY_SCRIPT
file for usage.

=over

=item Synopsis

 my $uri_path = $supdiag->add_local_proxy();

=item Arguments

None.

=item Returns

The URI path of the new resource, or 0 on failure. The URI path will be
Supdiag::LOCAL_PROXY_URI_PATH.

=back

=cut

sub add_local_proxy($) {
    my $self = shift;
    my $src_path = dirname(__FILE__)."/".LOCAL_PROXY_SCRIPT;
    return $self->copy_web_file(src_file => $src_path);
}

=head2 delete_local_proxy

Deletes the local proxy resource that is installed by add_local_proxy().

=over

=item Synopsis

 $supdiag->delete_local_proxy();

=item Arguments

None.

=item Returns

1 on success, 0 on failure.

=back

=cut

sub delete_local_proxy($) {
    my $self = shift;
    return $self->delete_web_files(LOCAL_PROXY_SCRIPT);
}

1;
