# $Id$
## @summary Common library for managing Switfttest tests
## @author bellizzi@netapp.com
## @pod here

=head1 NAME

SwiftTest.pm - Common library to manage Switfttest tests

=head1 DESCRIPTION

This package defines routines to manipulate Swifttest , XML test configurations
and test execution. Test execution can happen on Linux or Windows hosts. This 
moddule can only modify a pre existing SwiftTest project.  It camnot create one 
from scratch...yet. For an example on how to use this library see 
stress-scale/nas/swifttest.thpl.

=head1 SYNOPSIS

    use SwiftTest;
    my $SwiftTest = new SwiftTest(
                 'host'                 => $host,
                 'install_path'         => $install_path,
                 'mono_command'         => $mono_command,
                 'unix_log_path'        => $top_logdir,
                 'windows_log_path'     => $top_logdir_windows,
                 'unix_project_path'    => $top_logdir . '/' . $project,
                 'windows_project_path' => $top_logdir_windows . '\\' . $project
                 );

   Unix paths are required for project manipulation
   Windows paths are only required if using a windows host to run tests

=head1 REQUIREMENTS

Mono 2.x and specific files from the SwiftTest installation is required for i
runnng tests on Linux. 

A MinWin compliant Windows host with the SwiftTest TDE installed
is required for running tests usng Windows

=head1 SETUP

If you are gong to using a Windows host to run tests, install the SwiftTest
TDE on that host and note the install path.

If you are going to use a Linux host to run tests you wil the files in 
contained in SwiftCmdLinux.tgz that is included with your SwiftTest build

These files must match the version of the TDE that created yor SwiftTest project

The current plan is to place these files in /x/eng/localtest/arch/win32/SwiftTest_VERSION
where VERSION is the version string with periods replaced with an underscore

For example:

/x/eng/localtest/arch/win32/bin/SwiftTest_0_22_11624:

The exe and dll files must executable with lower case exceptions

If your Linux system does not have Mono 2.X, install Mono 2.x. See this link for 
how to get Mono. http://www.go-mono.com/mono-downloads/download.html

Run the SwiftTest TDE and create your test.

Click Generate Automation File in the Project menu to create the projects XML
files. 

Copy the project to somewhere accessable.

=head1 BUGS

As of 0.22.12086 running SwiftCmd on Linux using Monon causes it to fail with 
an xception while writing to the sqlite stats database

=head1 STATUS

  Public

=head1 AUTHOR

  David Bellizzi (bellizzi@netapp.com)
  $Id: $
  Created $Date: 2010/12/25 $

=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved

=head1 ROUTINES

=cut

package SwiftTest;

use strict;
use Tharn;
use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use File::Basename qw(basename);
use Data::Dumper;
use Options;
use POSIX qw(strftime);
use XML::Simple qw(:strict);
use Sort::Key::Natural qw(natsort);
use NATE::Child::Exec;
use YAML;

my $Options = new Options;

=head2 client_networks

    $SwiftTest->client_networks();

C<client_networks> Returns ClinetNetConfig as an array

=cut

sub client_networks {
    $Log->enter();
    my $self = shift;
    my @return;
    my $client_network_config =
      $self->{'AutomationConfig'}->{'ClientNetConfig'};
    if (ref($client_network_config) eq 'ARRAY') {
        @return = @{$client_network_config};
    } else {
        push(@return, $client_network_config);
    }
    return @return;
    $Log->exit();
}

=head2 client_network_ids

    $SwiftTest->client_network_ids()

C<client_network_ids> Returns an array of client network ids

=cut

sub client_network_ids {
    $Log->enter();
    my $self = shift;
    my @ids;
    foreach my $network ($self->client_networks()) {
        push(@ids, $network->{'NetID'});
    }

    return @ids;
    $Log->exit();
}

=head2 client_ports

    $SwiftTest->client_ports();

C<client_ports> Returns ClinetPortConfig as an array

=cut

sub client_ports {
    $Log->enter();
    my $self = shift;
    my @return;
    my $client_port_config = $self->{'AutomationConfig'}->{'ClientPortConfig'};
    if (ref($client_port_config) eq 'ARRAY') {
        @return = @{$client_port_config};
    } else {
        push(@return, $client_port_config);
    }
    return @return;
    $Log->exit();
}

=head2 client_port_ids

    $SwiftTest->client_ports_ids()

C<client_port_ids> Returns an array of client port ids

=cut

sub client_port_ids {
    $Log->enter();
    my $self = shift;
    my @ids;
    foreach my $port ($self->client_ports()) {
        push(@ids, $port->{'ClientPortID'});
    }

    return @ids;
    $Log->exit();
}

=head2 configure_stats

    $SwiftTest->configure_stats();

C<configure_stats> Confgure the specified ports stats

=over

=item template_file

(Required) An XML file containing the structure of the statistics

  <StatsOutput>
    <Port>
      <PortName></PortName>
      <StatElement>
        <StatType>Load Status</StatType>
        <StatFieldList>
          <StatField>SC_ATTEMPTS</StatField>
          <StatField>SC_SUCCEEDS</StatField>
          <StatField>SC_FAILS</StatField>
          <StatField>SC_ABORTS</StatField>
          <StatField>AC_ATTEMPTS</StatField>
          <StatField>AC_SUCCEEDS</StatField>
          <StatField>AC_FAILS</StatField>
          <StatField>AC_ABORTS</StatField>
          <StatField>CN_ATTEMPTS</StatField>
          <StatField>CN_SUCCEEDS</StatField>
          <StatField>CN_FAILS</StatField>
          <StatField>CN_ABORTS</StatField>
        </StatFieldList>
      </StatElement>
    </Port>
  </StatsOutput>

=cut

sub configure_stats {
    $Log->enter();

    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT}],
                     {template_file => {type => SCALAR, optional => 0},});

    my $path = "$self->{'unix_project_path'}/$opts{'template_file'}";
    my $stats_template = $self->read_xml_file('path' => $path);
    $self->{'stats_template'} = $stats_template;
    foreach my $port ($self->client_ports()) {
        my $stats = $self->{'stats_template'};
        $port->{'ClientPortID'} =~ /.+(\d)$/;
        $stats->{'Port'}->{'PortName'} = "Client Port " . $1;
        push(@{$self->{'AutomationConfig'}->{'StatsOutput'}}, $stats);
    }
    $Log->exit();
}

=head2 loads

    $SwiftTest->loads();

C<loads> Returns LoadConfig as an array

=cut

sub loads {
    $Log->enter();
    my $self = shift;
    my @return;
    my $loads = $self->{'AutomationConfig'}->{'LoadConfig'};
    if (ref($loads) eq 'ARRAY') {
        @return = @{$loads};
    } else {
        push(@return, $loads);
    }
    return @return;
}

=head2 load_ids

    $SwiftTest->load_ids()

C<load_ids> Returns an array of client port ids

=cut

sub load_ids {
    $Log->enter();
    my $self = shift;
    my @ids;
    foreach my $load ($self->loads()) {
        push(@ids, $load->{'LoadID'});
    }

    return @ids;
    $Log->exit();
}

=head2 modify_client_network_config

    $SwiftTest->modify_client_network_config(
                                  client_port_id => $port,
                                  enabled        => 1,
                                  net_id         => 5034,
                                  trace_param_id => undef, 
				  user_param_id => undef
                                  );

C<modify_client_network_config> Updates a client network

=over

=item client_port_id

(Optional) The id of the client port

=item enabled

(Optional) Is the port enabled

=item net_id

(Required) The id of the network defined in NetConfig

=item trace_param_id

(Optional) The id of the tracing parameters

=item user_param_id

(Optional) Arrayref of user parameter ids

=back

=cut

sub modify_client_network {
    $Log->enter();

    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {net_id         => {type => SCALAR, optional => 0},
                      client_port_id => {type => SCALAR, optional => 1},
                      enabled        => {type => SCALAR, optional => 1},
                      trace_param_id => {type => SCALAR, optional => 1},
                      user_param_id  => {type => SCALAR, optional => 1}
                     }
                     );

    my $client_net_config = $self->{'AutomationConfig'}->{'ClientNetConfig'};
    my $client_network;
    if (ref($client_net_config) eq 'ARRAY') {
        my @ids = $self->client_network_ids();
        for (my $id = 0 ; $id < @ids ; $id++) {
            $client_network = $client_net_config->[$id];
            if ($client_network->{'NetID'} eq $opts{'net_id'}) {
                last;
            }

        }
    } else {
        $client_network = $client_net_config;
    }

    if ($opts{'enabled'}) {
        $client_network->{'Enabled'} = $opts{'enabled'};
    }
    if ($opts{'client_port_id'}) {
        $client_network->{'ClientPortID'} = $opts{'client_port_id'};
    }
    if ($opts{'trace_param_id'}) {
        $client_network->{'TraceParametersID'} = $opts{'trace_param_id'};
    }
    if ($opts{'user_param_id'}) {
        $client_network->{'UserParametersID'} = $opts{'user_param_id'};
    }
    $Log->exit();
}

=head2 modify_client_port_config

    $SwiftTest->modify_client_port_config('path' => $path);

C<modify_client_port_config> Updates one or more fields of a user param file

=over

=item net_id

(Optional) The id of the network defined in NetConfig

=item client_port_id

(Required) The id of the client port

=item enabled

(Optional) Is the port enabled

=item appliance

(Optional) The IP address of the SwiftTest appliance

=item trace_param_id

(Optional) The id of the tracing parameters

=item user_param_id

(Optional) Arrayref of user parameter ids

=item port

(Optional) The physical SwiftTest port

=item enabled

(Optional) Enables/disables the port

=back

=cut

sub modify_client_port {
    $Log->enter();

    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {net_id         => {type => SCALAR, optional => 1},
                      client_port_id => {type => SCALAR, optional => 0},
                      appliance      => {type => SCALAR, optional => 1},
                      port           => {type => SCALAR, optional => 1},
                      enabled        => {type => SCALAR, optional => 1},
                      trace_param_id => {type => SCALAR, optional => 1},
                      user_param_id  => {type => SCALAR, optional => 1}
                     }
                     );
    my $client_port_config = $self->{'AutomationConfig'}->{'ClientPortConfig'};
    my $client_port;
    if (ref($client_port_config) eq 'ARRAY') {
        my @ids = $self->client_port_ids();
        for (my $id = 0 ; $id < @ids ; $id++) {
            $client_port = $client_port_config->[$id];
            if ($client_port->{'ClientPortID'} eq $opts{'client_port_id'}) {
                last;
            }

        }
    } else {
        $client_port = $client_port_config;
    }

    if ($opts{'appliance'}) {
        $client_port->{'Appliance'} = $opts{'appliance'};
    }
    if ($opts{'port'}) {
        $client_port->{'Port'} = $opts{'port'};
    }
    if ($opts{'enabled'}) {
        $client_port->{'Enabled'} = $opts{'enabled'};
    }
    $client_port->{'ClientPortID'} = $opts{'client_port_id'};
    if ($opts{'trace_param_id'}) {
        $client_port->{'TraceParametersID'} = $opts{'trace_param_id'};
    }
    if ($opts{'user_param_id'}) {
        $client_port->{'UserParametersID'} = $opts{'user_param_id'};
    }
    $Log->exit();
}

=head2 modify_load_config

    $SwiftTest->modify_load_config(load => \%hash);

C<modify_load_config> Updates one or more fields of a user param file

=over

=item load_id

(Required) The id of the client port

=item type

(Optional) The type of load

=item backwards

(Optional) The step direction

=item steps

(Optional) The number of steps for this load

=item values

(Optional) Pertains to load type

=back

=cut

sub modify_load {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {load_id   => {type => SCALAR, optional => 0},
                      backwards => {type => SCALAR, optional => 1},
                      steps     => {type => SCALAR, optional => 1},
                      type      => {type => SCALAR, optional => 0},
                      value     => {type => SCALAR, optional => 0}
                     }
                     );
    my $id;
    my $load;
    my $load_config = $self->{'AutomationConfig'}->{'LoadConfig'};
    if (ref($load_config) eq 'ARRAY') {
        my @ids = $self->load_ids();
        for ($id = 0 ; $id < @ids ; $id++) {
            $load = $load_config->[$id];
            if ($load->{'LoadID'} eq $opts{'load_id'}) {
                last;
            }
        }
    } else {
        $load = $load_config;
    }

    if ($opts{'backwards'}) {
        $load->{'Backwards'} = $opts{'backwards'};
    }
    if ($opts{'steps'}) {
        $load->{'Steps'} = $opts{'steps'};
    }
    if ($opts{'type'}) {
        $load->{'Type'} = $opts{'type'};
    }
    if ($opts{'value'}) {
        $load->{'Value'} = $opts{'value'};
    }
    $Log->exit();
}

=head2 modify_network_config

    $SwiftTest->modify_network_config(
                                  net_id => $id,
                                  ipv4_gw_addr => 1.2.3.1,
                                  ipv4_gw_enabled => 1,
                                  ipv4_netmask => 255.255.255.0,
                                  src_ipv4_addr_count => 254, 
                                  src_ipv4_addr_first => 1.2.3.1, 
                                  src_ipv4_addr_step => 1
                                  );

C<modify_network_config> Updates a client network

=over

=item net_id

(Required) The id of the network

=item ipv4_gw_addr

(Optional) The IPv4 gateway address

=item ipv4_gw_enabled

(Optional) Enables the IPv4 gateway

=item ipv4_netmask

(Required) The IPv4 netmask

=item src_ipv4_addr_count

(Required) How many IPv4 addresses to use

=item src_ipv4_addr_first

(Required) The first IPv4 addresses

=item src_ipv4_addr_step

(Required) How many IP addresses we increment before giving the next address

=item vlan_enabled

(Optional) Enables the use of VLAN tagging

=item vlan

(Optional) The vlan ID

=item vlan_count

(Optional) The nunber of vlans to use

=item vlan_step

(Optional) The step value used when incrementing the vlan ID

=back

=cut

sub modify_network {
    $Log->enter();

    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {net_id              => {type => SCALAR, optional => 0},
                      ipv4_gw_addr        => {type => SCALAR, optional => 1},
                      ipv4_gw_enabled     => {type => SCALAR, optional => 1},
                      ipv4_netmask        => {type => SCALAR, optional => 0},
                      src_ipv4_addr_count => {type => SCALAR, optional => 0},
                      src_ipv4_addr_first => {type => SCALAR, optional => 0},
                      src_ipv4_addr_step  => {type => SCALAR, optional => 0},
                      vlan_enabled =>
                        {type => SCALAR, optional => 1, default => 0},
                      vlan       => {type => SCALAR, optional => 1},
                      vlan_count => {type => SCALAR, optional => 1},
                      vlan_step  => {type => SCALAR, optional => 1},
                     }
                     );
    my $id;
    my $network;
    my $network_config = $self->{'AutomationConfig'}->{'NetConfig'};
    if (ref($network_config) eq 'ARRAY') {
        my @ids = $self->network_ids();
        for ($id = 0 ; $id < @ids ; $id++) {
            $network = $network_config->[$id];
            if ($network->{'NetID'} eq $opts{'net_id'}) {
                last;
            }
        }
    } else {
        $network = $network_config;
    }

    if ($opts{'ipv4_gw_enabled'}) {
        $network->{'IPv4GatewayAddr'}    = $opts{'ipv4_gw_addr'};
        $network->{'IPv4GatewayEnabled'} = $opts{'ipv4_gw_enabled'};
    }
    $network->{'IPv4Netmask'}         = $opts{'ipv4_netmask'};
    $network->{'SourceIPv4AddrCount'} = $opts{'src_ipv4_addr_count'};
    $network->{'SourceIPv4AddrFirst'} = $opts{'src_ipv4_addr_first'};
    $network->{'SourceIPv4AddrStep'}  = $opts{'src_ipv4_addr_step'};
    if ($opts{'vlan_enabled'}) {
        $network->{'VLANAddressEnabled'} = $opts{'vlan_enabled'};
        $network->{'VLANAddressFirst'}   = $opts{'vlan'};
        $network->{'VLANAddressCount'}   = $opts{'vlan_count'};
        $network->{'VLANAddressStep'}    = $opts{'vlan_step'};
    } else {
        $network->{'VLANAddressEnabled'} = 0;
        $network->{'VLANAddressFirst'}   = 1;
        $network->{'VLANAddressCount'}   = 1;
        $network->{'VLANAddressStep'}    = 1;
    }
    $Log->exit();
}

=head2 modify_user_parameter_file

    $SwiftTest->modify_user_parameter_file('path' => $path, );

C<modify_user_parameter_file> Updates one or more fields of a user param file

=over

=item file

(Required) The path to the user param file

=item data

(Required) Arrayref to array containing row data

=back

=cut

sub modify_user_parameter_file {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {file => {type => SCALAR,  optional => 0},
                      data => {type => HASHREF, optional => 0}
                     }
                     );

    # Read teh XML file
    my $path =
      "$self->{'unix_project_path'}/$self->{'config_dir'}/$opts{'file'}";
    $Log->comment("Modifying $opts{'file'}");
    my $params = $self->read_xml_file('path' => $path);

    # Convert to arrayref if needed
    my @existing_rows;
    if (ref($params->{'Row'}) ne 'ARRAY') {
        push(@existing_rows, $params->{'Row'});
    } else {
        @existing_rows = @{$params->{'Row'}};
    }

    # Convert the input rows to arrayref if needed
    my @input_rows;
    if (ref($opts{'data'}->{'Row'}) ne 'ARRAY') {
        push(@input_rows, $opts{'data'}->{'Row'});
    } else {
        @input_rows = @{$opts{'data'}->{'Row'}};
    }

    # Creat the hashref
    my $data = {};

    # Walk through each row and replace it if it exists in the input row
    # otherwise retain the existing data
    for (my $row = 0 ; $row < @input_rows ; $row++) {
        foreach my $column (keys(%{$input_rows[$row]})) {
            if (defined($input_rows[$row]->{$column})) {
                $data->{'Row'}->[$row]->{$column} =
                  $input_rows[$row]->{$column};
            } elsif (defined($existing_rows[$row]->{$column})) {
                $data->{'Row'}->[$row]->{$column} =
                  $existing_rows[$row]->{$column};
            } else {
                $data->{'Row'}->[$row]->{$column} = '';
            }
        }
    }

    $self->write_xml_file('path' => $path,
                          'data' => $data,
                          'root' => 'UserParameters'
                          );
    $Log->exit();

}

=head2 networks

    $SwiftTest->networks();

C<networks> Returns NetworkConfig as an array

=cut

sub networks {
    $Log->enter();
    my $self = shift;
    my @return;
    my $networks = $self->{'AutomationConfig'}->{'NetConfig'};
    if (ref($networks) eq 'ARRAY') {
        @return = @{$networks};
    } else {
        push(@return, $networks);
    }
    $Log->exit();
    return @return;
}

=head2 network_ids

    $SwiftTest->network_ids()

C<network_ids> Returns an array of network ids

=cut

sub network_ids {
    $Log->enter();
    my $self = shift;
    my @ids;
    foreach my $network ($self->networks()) {
        push(@ids, $network->{'NetID'});
    }

    return @ids;
    $Log->exit();
}

=head2 new

    my $SwiftTest = new SwiftTest(
                 'appliance'            => $appliance,
                 'host'                 => $host,
                 'install_path'         => $install_path,
                 'mono_command'         => $mono_command,
                 'unix_log_path'        => $top_logdir,
                 'windows_log_path'     => $top_logdir_windows,
                 'unix_project_path'    => $top_logdir . '/' . $project,
                 'windows_project_path' => $top_logdir_windows . '\\' . $project
                 );

C<new> Creates a SwiftTest object

=over

=item config_dir

(Required) The current projects configuration directory

=item install_path

(Required) The unix or windows path to SwiftCmd

=item mono_command

(Optional) The mono command to use

=item unix_log_path

(Required) The unix path to the swifttest log directory

=item unix_project_path

(Required) The unix path to the swifttest project directory

=item windows_log_path

(Optional) The windows path to the swifttest project directory

=item windows_project_path

(Optional) The windows path to the swifttest project directory

=back

=cut

sub new {
    $Log->enter();
    my ($invocant, %opts) =
      validate_mixed(@_,
                     [{type => SCALAR}],
                     {config_dir => {type     => SCALAR | UNDEF,
                                     optional => 1,
                                     default  => 'AutomationConfig'
                        },
                      'host' => {type     => SCALAR,
                                 optional => 0,
                                 default  => 'localhost'
                        },
                      install_path => {type => SCALAR, optional => 0},
                      mono_command => {type     => SCALAR | UNDEF,
                                       optional => 1,
                                       default  => 'mono'
                        },
                      unix_log_path    => {type => SCALAR, optional => 0},
                      windows_log_path => {type     => SCALAR | UNDEF,
                                           optional => 1,
                        },
                      unix_project_path    => {type => SCALAR, optional => 0},
                      windows_project_path => {type     => SCALAR | UNDEF,
                                               optional => 1,
                        }
                     }
                     );

    my $class = ref($invocant) || $invocant;
    my $self = {};

    $self->{'config_dir'}           = $opts{'config_dir'};
    $self->{'host'}                 = $opts{'host'};
    $self->{'install_path'}         = $opts{'install_path'};
    $self->{'mono_command'}         = $opts{'mono_command'};
    $self->{'unix_log_path'}        = $opts{'unix_log_path'};
    $self->{'unix_project_path'}    = $opts{'unix_project_path'};
    $self->{'windows_log_path'}     = $opts{'windows_log_path'};
    $self->{'windows_project_path'} = $opts{'windows_project_path'};

    # Get the project name from the unix project path
    $self->{'project_name'} = $self->{'unix_project_path'};
    $self->{'project_name'} =~ s/^.+\///;

    bless($self, $class);

    opendir(DIR, $self->{'unix_project_path'})
      || die "Cannont open $self->{'unix_project_path'}";
    my @files = grep(/\.swift_test/i, readdir(DIR));
    $self->{'project_file'} = $files[0];

    my $path =
      "$self->{'unix_project_path'}/$self->{'config_dir'}/AutomationConfig.xml";
    $Log->comment("Reading AutomationConfig.xml");
    $self->{'AutomationConfig'} = $self->read_xml_file('path' => $path);

    $Log->exit();
    return $self;
}

=head2 populate_user_parameter_files

    $self->populate_user_parameter_files(data => \%data, path => $path);

C<populate_user_parameter_files> Updates user parameter files based on 
user parameter specification and corressponding data

=over

=item data

(Required) Hasref of data

=item path

(Required) The path to teh user parameter specification file 

=back

=cut

sub populate_user_parameter_files {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT}],
                     {path => {type => SCALAR,  optional => 0},
                      data => {type => HASHREF, optional => 0}
                     }
                     );

    # Modify UserParameterFiles specified by SWIFTTEST_UP_SPEC
    $Log->comment("Modifying user parameter files");

    # Read the UP spec file
    my $specs = {};

    if (!$opts{'path'}) {
        die script_result("Cannot find $opts{'path'}");
    }

    open my $fh, '<', "$opts{'path'}"
      or die "Error opening $opts{'path'}: $!";
    my $yml = do { local $/; <$fh> };
    $specs = Load($yml);

    # Read the user parameter map and create a hash indexed by the
    # user parameter id
    my %map;
    foreach my $up ($self->user_parameter_map()) {
        $map{$up->{'UserParametersID'}} = $up->{'MapID'};
    }

    # Loop though each user parameter file
    foreach my $up ($self->user_parameters()) {

        # Get the UP files id
        my $up_id = $up->{'UserParametersID'};

        # If the current user parameter file is not in the map then skip it
        if (!grep(/$up_id/, keys(%map))) {
            next;
        }

        # Get the file name of the current user paramater file
        my $file = $up->{'FileName'};

        # Get the current user parameter file map id
        my $map_id = $map{$up_id};

        # Clear the the data hashref
        my $data = {};

        my $row = 0;
        if (defined($specs->{$map_id}->{'csv_file'})) {

            # If a CSV file is defined use it

            # Open the CSV file
            my $csv_file =
              "$self->{'unix_project_path'}/$specs->{$map_id}->{'csv_file'}";
            $Log->comment("Importing user parameters from $csv_file");
            open(FMT, $csv_file) || die "Unable to open $csv_file";

            # Get the column names from the spec file
            my @columns = natsort(keys(%{$specs->{$map_id}->{'columns'}}));

            # Red the contents fo the CSV file
            while (<FMT>) {
                my $line = $_;
                chomp($line);
                my @fields = split(/,/, $line);

                # Loop through the CSV files fields and modify the corresponding
                # field in the UP file
                for (my $i = 0 ; $i < @columns ; $i++) {
                    chomp($fields[$i]);
		    $fields[$i] =~ s/;/,/g;
                    $data->{'Row'}->[$row]->{$columns[$i]} = $fields[$i];
                }
                $row++;
            }

        } else {

            $Log->comment(  "Creating user parameter values "
                          . "for $specs->{$map_id}->{'name'}");

            foreach my $column (keys(%{$specs->{$map_id}->{'columns'}})) {
                my $key = $specs->{$map_id}->{'columns'}->{$column};
                $Log->debug("Adding $key data to column $column");
                if (defined($opts{'data'}->{$key})) {
                    $Log->debug("Iterating through "
                            . @{$opts{'data'}->{$key}}
                            . " values for $key");
                    for (
                        my $row = 0;
                        $row < @{$opts{'data'}->{$key}};
                        $row++
                        ) {
                        my $value = ${$opts{'data'}->{$key}}[$row];
                        $Log->debug(
                            "Saving $value for row:$row column:$column");
                        $data->{'Row'}->[$row]->{$column} = $value;
                    }
                }
            }
        }

        # Dump the data hashref so we can see whats going into the file
        $Log->debug(Dumper($data));

        # Update the user parameter file
        $self->modify_user_parameter_file('file' => $file,
                                          'data' => $data,);
    }

    $Log->exit();

}

=head2 read_xml_file

    my $data = $SwiftTest->read_xml_file(path => $path);

C<read_xml_file> returns a hashref of the specified XML file in 
perl hash form

=over

=item path

(Required) The path to the XML file to read

=back

=cut

sub read_xml_file {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {path => {type => SCALAR, default => undef},},
                     );

    if (!-e $opts{'path'}) {
        die script_result("File $opts{'path'} not found");
    }

    my $xml = new XML::Simple(NoAttr => 1, ForceArray => [], KeyAttr => []);
    my $data = $xml->XMLin($opts{'path'});
    $Log->exit();
    return $data;
}

=head2 write_config

    $SwiftTest->write_config();

C<write_config> Write the internal stucture to AutomationConfig.xml

=cut

sub write_config {
    $Log->enter();

    my $self = shift;

    my $path =
      "$self->{'unix_project_path'}/$self->{'config_dir'}/AutomationConfig.xml";
    my $data = $self->{'AutomationConfig'};
    $Log->comment("Writing AutomationConfig.xml");
    $self->write_xml_file('path' => $path,
                          'data' => $data,
                          'root' => 'SwiftCmd'
                          );
    $Log->exit();
}

=head2 start

    $SwiftTest->start();

C<start> Starts the current test

=cut

sub start {
    $Log->enter();

    my $self = shift;

    my $timestamp = strftime("%m%d%y%H%M%S", localtime());
    $self->{'swiftcmd_log'} = 'swiftcmd-' . $timestamp . '.log';
    $self->{'start_applog'} = 'start_applog-' . $timestamp . '.log';

    my ($cmd, $log, $stop_cmd);
    my $host_type = host($self->{'host'})->hosttype();
    if ($host_type =~ /lin/i) {
        $cmd =
          "$self->{'mono_command'} $self->{'install_path'}/SwiftCmd.exe --start "
          . "--project:$self->{'unix_project_path'}/$self->{'project_file'} "
          . "--config:$self->{'unix_project_path'}/$self->{'config_dir'}/AutomationConfig.xml "
          . "--applog:$self->{'unix_project_path'}/$self->{'start_applog'}";
        $stop_cmd =
          "$self->{'mono_command'} $self->{'install_path'}/SwiftCmd.exe --stop "
          . "--project:$self->{'unix_project_path'}/$self->{'project_file'} "
          . "--config:$self->{'unix_project_path'}/$self->{'config_dir'}/AutomationConfig.xml";
        $log = "$self->{'unix_project_path'}/$self->{'swiftcmd_log'}";
    } elsif ($host_type =~ /win/i) {
        $cmd =
            "$self->{'install_path'}\\SwiftCmd.exe /Start "
          . "/Project:$self->{'windows_project_path'}\\$self->{'project_file'} "
          . "/Config:$self->{'windows_project_path'}\\$self->{'config_dir'}\\AutomationConfig.xml "
          . "/Applog:$self->{'windows_project_path'}\\$self->{'start_applog'}";
        $stop_cmd =
            "$self->{'install_path'}\\SwiftCmd.exe /Stop "
          . "/Project:$self->{'windows_project_path'}\\$self->{'project_file'} "
          . "/Config:$self->{'windows_project_path'}\\$self->{'config_dir'}\\AutomationConfig.xml";
        $stop_cmd =~ s/\\/\\\\/g;
        $log = "$self->{'windows_project_path'}\\$self->{'swiftcmd_log'}";
    } else {
        die config_result("Hosttype $host_type not supported");
    }

    # Create a stop script so we can easially stop it.
    open(SCR, ">$self->{'unix_log_path'}/swiftcmd_stop.sh")
      || die "Cannot create swiftcmd_stop.sh; $!";
    print SCR '#!/bin/sh' . "\n";
    print SCR "rsh $self->{'host'} " . $stop_cmd . "\n";
    close SCR;

    my %args = (command              => $cmd,
                inherit_env          => 0,
                endpoint             => $self->{'host'},
                kill_if_parent_exits => 1,
                logfile              => $log
                );

    $Log->comment("Starting SwiftCmd: $cmd");
    $Log->debug("NATE::Child::Exec->new options\n" . Dumper(\%args));
    my $Staf = NATE::Child::Exec->new(%args);

    $Log->exit();
    return $Staf;
}

=head2 stop

    $SwiftTest->stop();

C<stop> Starts the current test

=cut

sub stop {
    $Log->enter();

    my $self = shift;
    my ($cmd, $log);
    my $timestamp = strftime("%m%d%y%H%M%S", localtime());
    $self->{'stop_applog'} = 'stop_applog-' . $timestamp . '.log';
    my $host_type = host($self->{'host'})->hosttype();
    if ($host_type =~ /lin/i) {
        $cmd =
          "sudo $self->{'mono_command'} $self->{'install_path'}/SwiftCmd.exe --stop "
          . "--project:$self->{'unix_project_path'}/$self->{'project_file'} "
          . "--config:$self->{'unix_project_path'}/$self->{'config_dir'}/AutomationConfig.xml "
          . "--applog:$self->{'unix_project_path'}/$self->{'stop_applog'}";
        $log = "$self->{'unix_project_path'}/$self->{'swiftcmd_log'}";
    } elsif ($host_type =~ /win/i) {
        $cmd =
            "$self->{'install_path'}\\SwiftCmd.exe /Stop "
          . "/Project:$self->{'windows_project_path'}\\$self->{'project_file'} "
          . "/Config:$self->{'windows_project_path'}\\$self->{'config_dir'}\\AutomationConfig.xml "
          . "/Applog:$self->{'windows_project_path'}\\$self->{'stop_applog'}";
        $log = "$self->{'windows_log_path'}\\$self->{'swiftcmd_log'}";
    } else {
        die config_result("Hosttype $host_type not supported");
    }

    my %args = (command              => $cmd,
                inherit_env          => 0,
                endpoint             => $self->{'host'},
                kill_if_parent_exits => 1,
                logfile              => $log
                );

    $Log->comment("Stopping SwiftCmd: $cmd");
    $Log->debug("NATE::Child::Exec->new options\n" . Dumper(\%args));
    my $Staf = NATE::Child::Exec->new(%args);

    $Log->exit();
    return $Staf;
}

=head2 user_parameters

    $SwiftTest->user_parameters();

C<user_parameters> Returns all the user parameter entries

=cut

sub user_parameters {
    $Log->enter();

    my $self   = shift;
    my $params = $self->{'AutomationConfig'}->{'UserParametersConfig'};
    if (ref($params) eq 'ARRAY') {
        return @{$params};
    } else {
        return ($params);
    }
    $Log->exit();
}

=head2 user_parameter_map

    $SwiftTest->user_parameter_map();

C<user_parameter_map> Returns the user parameter map

=cut

sub user_parameter_map {
    $Log->enter();

    my $self = shift;
    my $map  = $self->{'AutomationConfig'}->{'UserParametersMapConfig'};
    if (ref($map) eq 'ARRAY') {
        return @{$map};
    } else {
        return ($map);
    }
    $Log->exit();
}

=head2 write_xml_file

    my $data = $SwiftTest->write_xml_file(path => $path);

C<write_xml_file> returns a hashref of the specified XML file in 
perl hash form

=over

=item path

(Required) The path to the XML file to write

=back

=cut

sub write_xml_file {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {path => {type => SCALAR,  default => undef},
                      data => {type => HASHREF, default => undef},
                      root => {type => SCALAR,  default => undef}
                     }
                     );

    my $xml = new XML::Simple(NoAttr     => 1,
                              ForceArray => [],
                              KeyAttr    => [],
                              RootName   => $opts{'root'}
                              );
    my $data = $xml->XMLout($opts{'data'});

    open(XML, ">$opts{'path'}");
    print XML $data;
    close XML;
    $Log->exit();
}

1;
