use strict;
use Options;
use Tharn qw($Log);

package SysInit;

use Data::Dumper;
use Params::Validate
    qw(validate validate_with BOOLEAN SCALAR ARRAYREF HASHREF OBJECT SCALARREF);
use Params::TestBranch;
use NATE::Inc qw(find_file);
use Tie::IxHash;
use NACL::GeneralUtils;

my $NATE_LIB = Tharn::param("NATE_LIB");
my $bash;

my $Options = new Options;
$Options->process();

my $params = $Options->get_hash();

my %build2branch =    ( 'DOT/main'          => $params->{TEST_BRANCH_MAIN},
                        'DOT/Rsierranevada' => $params->{TEST_BRANCH_SN},
                        'DOT/Rrollingrock'  => $params->{TEST_BRANCH_RR},
                        'DOT/R8.1.2x'       => $params->{TEST_BRANCH_RR},
                        'DOT/R8.1x'         => $params->{TEST_BRANCH_RR0},
                        'DOT/Rboilermaker'  => $params->{TEST_BRANCH_BR},
                       );

my %testbranch2path = ( 'test/main'          => $params->{TEST_BRANCH_MAIN},
                        'test/Rsierranevada' => $params->{TEST_BRANCH_SN},
                        'test/Rrollingrock'  => $params->{TEST_BRANCH_RR},
                        'test/R8.1x'         => $params->{TEST_BRANCH_RR0},
                        'test/Rboilermaker'  => $params->{TEST_BRANCH_BR},
                       );

sub _is_cluster {
    my $type = shift;
    return ($type eq "C");
}

sub _is_seven {
    my $type = shift;
    return ($type eq "S");
}

# -----------------------------------------------
# Function: organize_system_params
#   Routine to group parameters that are defined as supporting
#   system prefixes into a hash that sys_init.thpl works off of.
#
# Usage:
#
#    my %system_params = SysInit::organize_system_params( param_names => \@sys_init_params )
#
# Parameters:
#
#    param_names => ARRAYREF, Options.pm parameters which support the sys_prefix attribute.
#
# Return:
#
#    Hash of hash $system_params{$systype}{$sysid}
#    Where systype is C or S and sysid is 1, 2, 3, etc...
#    corresponding to C1_parametername, or S2_parametername
#
# -----------------------------------------------

sub organize_system_params {
    my %opts = validate_with(
        params => \@_,
        spec   => {
            param_names  => { type    => ARRAYREF, 
                              default => [ qw( NODES 
                                               BUILDROOT
                                               IMAGE )
                                            ] },
        },
    );

    my %sys_init;
    my %sys_limit;
    my @param_names = @{$opts{param_names}};

    $sys_limit{C} = $Options->get_max_cmode_systems();
    $sys_limit{S} = $Options->get_max_7mode_systems();

    foreach my $systype ( "C", "S" ) {
        my $index = 1;

        while ( $index <= $sys_limit{$systype} ) {
            my $sysid = $systype . $index;
            $index++;
            my $prefix = $sysid . "_" ;

            foreach my $param ( @param_names ) {
                my $prefixed_param = $prefix . $param;

                if ( exists( $params->{$prefixed_param} ) ) {
                        $sys_init{$systype}{$sysid}{$param} = $params->{$prefixed_param};
                } elsif ( exists( $params->{$param} ) &&
                          ($param ne "NODES") ) {
                    $sys_init{$systype}{$sysid}{$param} = $params->{$param};
                }

                if ( ($param eq "IMAGE") && $sys_init{$systype}{$sysid}{$param} ) {
                    my $ibm_mode = 0;
                    if ( $sys_init{$systype}{$sysid}{$param} =~ /\.ibm/ ) {
                        $ibm_mode = 1;
                    }
                    $sys_init{$systype}{$sysid}{_ibm_mode} = $ibm_mode;
                }

                if ( ($param eq "NODES") && $sys_init{$systype}{$sysid}{$param} ) {
                    $sys_init{$systype}{$sysid}{_nodes} = $sys_init{$systype}{$sysid}{NODES};
                    $sys_init{$systype}{$sysid}{NODES} = join(",", @{$sys_init{$systype}{$sysid}{NODES}});
                }
            }

            # If no nodes were specified, skip
            if ( !$sys_init{$systype}{$sysid}{_nodes} ) {
                delete $sys_init{$systype}{$sysid};
                next;
            }

            $sys_init{$systype}{$sysid}{_is_vsim} = 0;
            my $nodes = $sys_init{$systype}{$sysid}{_nodes};

            # IxHash allows us to preserve hash key order
            my %nodespecs;
            tie %nodespecs, 'Tie::IxHash';

            %nodespecs = SysInit::_parse_hosts( $sys_init{$systype}{$sysid}{NODES} );
            foreach my $node (keys %nodespecs) {
                my $hostrec = Hostrec->new( id => $node );
                if ( $hostrec->{hosttype} =~ /vsim/ ) {
                    $sys_init{$systype}{$sysid}{_is_vsim} = 1;
                    last;
                }
            }

            # The assumption here is that the hostrec normally contains a Cmode
            # hosttype and that we need to change it to make it look like a 7mode
            # filer
            if ( _is_seven($systype) ) {
                my $hosttype = "filer-cfe";
                if ( $sys_init{$systype}{$sysid}{_is_vsim} ) {
                    $hosttype = "filer-vsim";
                }
                # Override hosttype for 7mode filers
                $sys_init{$systype}{$sysid}{NODES} = "";
                my $index = 0;
                foreach my $node ( @{$sys_init{$systype}{$sysid}{_nodes}} ) {
                    if ( $index ) {
                        $sys_init{$systype}{$sysid}{NODES} .= ',';
                    }
                    $index++;
                    $sys_init{$systype}{$sysid}{NODES} .= $node.';hosttype='.$hosttype;
                }
            }
        }
    }
    return %sys_init;
}

# -----------------------------------------------
# Function: ontap_build_to_automation_branch
#   Routine to determine the path to add to NATE_LIB so that
#   sys_init can subtest against the proper automation line
#   given an Ontap IMAGE name.
#
# Usage:
#
#    my $path = SysInit::ontap_build_to_automation_branch( image => $image )
#
# Parameters:
#
#    image    => path to Ontap image (ex: /x/eng/rlse/DOT/devN/final/bedrock/export/x86_64/tarball/image.tgz)
#    testfile => file we will look for in the test automation branch (default lib/SysInit.pm)
#
#    use_current_nate_lib => (yes|no|attempt) Default attempt.
#
#        If set to yes, this routine will simply return the root of the test automation
#        tree found first in NATE_LIB.
#
#        If set to no, this routine will look to see if the Ontap build has a test directory
#        if not it will use the parameters defined in lib/Params/TestBranch.pm
#
#        If set to attempt, this routine will try to use NATE_LIB when appropriate, but
#        fall back to looking at the parameters defined in lib/Params/TestBranch.pm
#        when necessary.
#
# Return:
#
#    Path to a test automation branch appropriate for this Ontap build.
#
# -----------------------------------------------
sub ontap_build_to_automation_branch {
    my $me = (caller(0))[3];
    my %opt = validate_with(
        params => \@_,
        spec   => {
            image    => { type => SCALAR },
            testfile => { type => SCALAR, default => 'lib/SysInit.pm' },
            use_current_nate_lib   => { type => SCALAR, 
                                        default => 'attempt',
                                        regex => qr/^(attempt|yes|no)$/i },
        },
    );

    my $automation_branch_path;

    my $image                  = $opt{image};
    my $testfile               = $opt{testfile};
    my $use_current_nate_lib   = $opt{use_current_nate_lib};

    my @nate_lib_path = split(';', $NATE_LIB );

    my $testfile_relative_to_nate_lib = find_file( $testfile, search => [@nate_lib_path] );
    my $testfile_root_from_nate_lib   = _find_p4_root( file => $testfile_relative_to_nate_lib );
    my $nate_lib_client_branch;

    my $P4BRANCH;

    if ( $use_current_nate_lib =~ /yes/i ) {
        if ( -e $testfile_root_from_nate_lib.'/ontap/main' && 
             -d $testfile_root_from_nate_lib.'/ontap/main' &&
             -r $testfile_root_from_nate_lib.'/ontap/main' ) {
            $testfile_root_from_nate_lib .= ';'.$testfile_root_from_nate_lib.'/ontap/main';
        }
        return $testfile_root_from_nate_lib;
    }

    if ( ! (  -e $image && 
             !-d $image &&
              -r $image ) ) {
        die "$me: $image does not exist";
    }

    my $image_root                  = _find_p4_root( file => $image );

    # If one of those new-fangled builds (like FS) where the automation directory is
    # stored in the Ontap build directory, skip the p4 examination
    if (  -e $image_root.'/test/'.$testfile &&
         !-d $image_root.'/test/'.$testfile &&
          -r $image_root.'/test/'.$testfile ) {
        $automation_branch_path = $image_root.'/test';
        goto RETURN;
    }

    my $p4info = $image_root.'/final/bedrock/export/common/p4info/p4info';
    $Tharn::Log->debug("$me: Looking for P4 Branch info in $p4info file");
    if (  -e $p4info &&
         !-d $p4info &&
          -r $p4info ) {
        open P4INFO_FILE, '<', $p4info or die $!;
        my @lines = <P4INFO_FILE>;
        my $P4CLIENT;
        foreach my $line ( @lines ) {
            if ( $line =~ /^P4BRANCH='(\S+)'/ ) {
                $P4BRANCH = $1;
                $Tharn::Log->debug("$me: P4BRANCH=$P4BRANCH");
                last;
            }
            if ( $line =~ /^P4CLIENT='(\S+)'/ ) {
                $P4CLIENT = $1;
                $Tharn::Log->debug("$me: P4CLIENT=$P4CLIENT");
            }
        }

        if ( $P4CLIENT && !$P4BRANCH ) {
            # If not one of the well-known clients
            if ( $P4CLIENT !~ /Rboilermaker|R8\.0/ ) {
                _get_bash_shell();

                my @results = `$bash -c 'P4NORC=1; P4CLIENT=noclient; export P4NORC; export P4CLIENT; /usr/software/rats/bin/p4 client -o $P4CLIENT'`;

                @results = reverse @results;
                my $found = 0;
                foreach my $result ( @results ) {
                    if ( $result =~ /(DOT\/.*?)\// ) {
                        $P4BRANCH = $1;
                        $Tharn::Log->debug(sub {"P4BRANCH=$P4BRANCH determined from examination of P4CLIENT: $P4CLIENT"});
                        last;
                    }
                }
            }
        }

        if ( $P4BRANCH ) {
            if ( exists $build2branch{$P4BRANCH} ) {
                $automation_branch_path = $build2branch{$P4BRANCH};
                goto RETURN;
            } else {
                _get_bash_shell();

                # Some users were running sys_init.thpl as root, and root typically does not
                # have a p4client setup, but if we specify P4CLIENT as noclient, and P4NORC
                # (which means do NOT look for a .b4p4rc in the user's home directory) then
                # codetree can be run by root.  This should also work for all other users 
                # too.
                $Tharn::Log->debug( "$me: $bash -c \'P4NORC=1; P4CLIENT=noclient; export P4NORC; export P4CLIENT; /usr/software/rats/bin/p4 codetree -tabs $P4BRANCH\'");
                my @results = `$bash -c 'P4NORC=1; P4CLIENT=noclient; export P4NORC; export P4CLIENT; /usr/software/rats/bin/p4 codetree -tabs $P4BRANCH'`;

                @results = reverse @results;
                my $found = 0;
                foreach my $result ( @results ) {
                    my @items = split(',', $result );
                    my $parent = $items[2];
                    if ( exists $build2branch{$parent} ) {
                        $automation_branch_path = $build2branch{$parent};
                        goto RETURN;
                    }
                }
            }
        } else {
            if ( $P4CLIENT ) {
                if ( $P4CLIENT =~ /Rboilermaker|R8\.0/ ) {
                    $automation_branch_path = $build2branch{'DOT/Rboilermaker'};
                    goto RETURN;
                } else {
                    print "Unknown Test Automation branch for $P4CLIENT\n";
                }
            } else {
                print "No P4BRANCH or P4CLIENT in $p4info\n";
            }
        }
    } else {
        print "$p4info does NOT exist\n";
    }

    return undef;

RETURN:

    # "new" style builds like devN where the test code is stored in the
    # Ontap build tree.  See if SysInit.pm library exists under the Ontap build.
    if (  -e $image_root.'/test/'.$testfile &&
         !-d $image_root.'/test/'.$testfile &&
          -r $image_root.'/test/'.$testfile ) {

        # If we found a "test" directory under the build directory, we know that should
        # work for this Ontap build, but if NATE_LIB is already setup with directories that
        # should work, see if they map to the same automation branch
        if ( $use_current_nate_lib =~ /attempt/i ) {

            my $image_client_branch   = _get_branch_from_client( root => $image_root );
            $nate_lib_client_branch   = _get_branch_from_client( root => $testfile_root_from_nate_lib );

            my $automation_branch_from_image    = _find_root_of_automation_branch( branch => $image_client_branch );
            my $automation_branch_from_nate_lib = _find_root_of_automation_branch( branch => $nate_lib_client_branch );

            # This probably will only work with DOT/mainN and test/main
            if ( $automation_branch_from_image eq $automation_branch_from_nate_lib ) {
                if ( -e $testfile_root_from_nate_lib.'/ontap/main' &&
                     -d $testfile_root_from_nate_lib.'/ontap/main' &&
                     -r $testfile_root_from_nate_lib.'/ontap/main' ) {
                    $testfile_root_from_nate_lib .= ';'.$testfile_root_from_nate_lib.'/ontap/main';
                }
                return $testfile_root_from_nate_lib;
            }

            if (  -e $testfile_root_from_nate_lib.'/test/'.$testfile &&
                 !-d $testfile_root_from_nate_lib.'/test/'.$testfile &&
                  -r $testfile_root_from_nate_lib.'/test/'.$testfile ) {
                # I don't think I can verify that this automation client is appropriate for the build 
                # as it was not directly derived from the build, just assuming that it is
                # because there is a test/$testfile under p4 root
                if ( -e $testfile_root_from_nate_lib.'/ontap/main' &&
                     -d $testfile_root_from_nate_lib.'/ontap/main' &&
                     -r $testfile_root_from_nate_lib.'/ontap/main' ) {
                    $testfile_root_from_nate_lib .= ';'.$testfile_root_from_nate_lib.'/ontap/main';
                }
                return $testfile_root_from_nate_lib;
            } else {
                $Tharn::Log->comment("$me: NATE_LIB does not appear to be appropriate for IMAGE $image");
                $Tharn::Log->comment("$me: Returning $automation_branch_path instead");
                $Tharn::Log->comment("$me: Control of test branches is documented at http://wikid/w/Coati/SysInit#NATE_LIB_per_Test_Automation_Branch_or_Why_isn.27t_sys_init.thpl_honoring_my_NATE_LIB.3F");
            }
        }
        if ( -e $automation_branch_path.'/ontap/main' &&
             -d $automation_branch_path.'/ontap/main' &&
             -r $automation_branch_path.'/ontap/main' ) {
            $automation_branch_path .= ';'.$automation_branch_path.'/ontap/main';
        }
        return $automation_branch_path;
    }

    if ( $use_current_nate_lib =~ /attempt/i ) {
       my $errors;
       eval {
           $nate_lib_client_branch   = _get_branch_from_client( root => $testfile_root_from_nate_lib );
           my $test_path             = _find_root_of_automation_branch( branch => $nate_lib_client_branch );
    
           if ( $test_path && ( $test_path eq $automation_branch_path ) ) {
               if ( -e $testfile_root_from_nate_lib.'/ontap/main' &&
                    -d $testfile_root_from_nate_lib.'/ontap/main' &&
                    -r $testfile_root_from_nate_lib.'/ontap/main' ) {
                   $testfile_root_from_nate_lib .= ';'.$testfile_root_from_nate_lib.'/ontap/main';
               }
               return $testfile_root_from_nate_lib;
           }
       };
       $errors = $@;
  
       if ( $errors ) {
           $errors = "\nErrors encountered = |$errors|";
       } else {
           $errors = "";
       }
       $Tharn::Log->comment("$me: NATE_LIB does not appear to be appropriate for IMAGE $image");
       $Tharn::Log->comment("$me: Returning $automation_branch_path instead");
       $Tharn::Log->comment("$me: Control of test branches is documented at http://wikid/w/Coati/SysInit#NATE_LIB_per_Test_Automation_Branch_or_Why_isn.27t_sys_init.thpl_honoring_my_NATE_LIB.3F");
    }

    if ( -e $automation_branch_path.'/ontap/main' &&
         -d $automation_branch_path.'/ontap/main' &&
         -r $automation_branch_path.'/ontap/main' ) {
        $automation_branch_path .= ';'.$automation_branch_path.'/ontap/main';
    }
    return $automation_branch_path;

}

# -----------------------------------------------
# Function: _find_root_of_automation_branch
#   Routine to determine the appropriate root automation path given an
#   branch name in the form DOT/branch or test/branch.
#
# Usage:
#
#    my $path = SysInit::_find_root_of_automation_branch( branch => 'DOT/dev' )
#
# Parameters:
#
#    branch    => branch name (like DOT/main, test/main, etc...)
#
# Return:
#
#    Path to the root of a test automation client appropriate for this Ontap build.
#
# -----------------------------------------------
sub _find_root_of_automation_branch {
    my $me = (caller(0))[3];
    my %opt = validate_with(
        params => \@_,
        spec   => {
            branch  => { type => SCALAR },
        },
    );

    my $branch = $opt{branch};

    if ( exists $testbranch2path{$branch} ) {
        return $testbranch2path{$branch};
    }
    if ( exists $build2branch{$branch} ) {
        return $build2branch{$branch};
    }

    _get_bash_shell();

    # Some users were running sys_init.thpl as root, and root typically does not
    # have a p4client setup, but if we specify P4CLIENT as noclient, and P4NORC
    # (which means do NOT look for a .b4p4rc in the user's home directory) then
    # codetree can be run by root.  This should also work for all other users
    # too.
    $Tharn::Log->debug( "$me: $bash -c \'P4NORC=1; P4CLIENT=noclient; export P4NORC; export P4CLIENT; /usr/software/rats/bin/p4 codetree -tabs $branch\'");
    my @results = `$bash -c 'P4NORC=1; P4CLIENT=noclient; export P4NORC; export P4CLIENT; /usr/software/rats/bin/p4 codetree -tabs $branch'`;

    @results = reverse @results;
    my $found = 0;
    foreach my $result ( @results ) {
        my @items = split(',', $result );
        my $parent = $items[2];

        if ( exists $testbranch2path{$parent} ) {
            return $testbranch2path{$parent};
        }
        if ( exists $build2branch{$parent} ) {
            return $build2branch{$parent};
        }
    }

    return undef;
}

# Just adding this routine so that when I merge these changes into older branches
# I can add my own _parse_hosts() in those versions of SysInit.pm
sub _parse_hosts {
    return NACL::GeneralUtils::_parse_hosts( @_ );
}

# -----------------------------------------------
# Function: _find_p4_root
#   Routine to determine the directory where P4ENV resides
#   given a file path.
#
# Usage:
#
#    my $path = SysInit::_find_p4_root( file => '/x/eng/rlse/DOT/devN/final/bedrock/export/x86_64/tarball/image.tgz' )
#
# Parameters:
#
#    file    => filepath at or deeper than where P4ENV resides.
#
# Return:
#
#    Path to the root of a P4 client.
#
# -----------------------------------------------
sub _find_p4_root {
    my $me = (caller(0))[3];
    my %opt = validate_with(
        params => \@_,
        spec   => {
            file    => { type => SCALAR },
        },
    );

    my $file = $opt{file};

    my @parts = split("/", $file );
    my $root = '';
    my $found_root = 0;
    foreach my $part (@parts) {
        if ( $part ) {
            $root .= '/'.$part;
            my $P4ENV = $root.'/P4ENV';
            if (  -e $P4ENV &&
                 !-d $P4ENV &&
                  -r $P4ENV ) {
                $found_root = 1;
                last;
            }
        }
    }

    if ( !$found_root ) {
        die "$me: could not find P4ENV from file $file";
    }

    return $root;
}

# -----------------------------------------------
# Function: _get_bash_shell
#   Routine to determine where the bash shell is.
#
# Usage:
#
#    my $bash = SysInit::_get_bash_shell();
#
# Parameters:
#
#    None
#
# Return:
#
#    Path to the bash shell
#
# -----------------------------------------------
sub _get_bash_shell {
    my $me = (caller(0))[3];
    if ( !$bash ) {
        # At least we can record which bash shell we are using
        $bash = `which bash`;
        chomp($bash);
        if (  -e $bash &&
             !-d $bash &&
              -r $bash &&
              -x $bash ) {
            $Tharn::Log->debug("$me: bash executable=|$bash|");
        } else {
            $Tharn::Log->warn("$me: bash executable does not appear to exist in path |$bash|");
            $bash = undef;
            foreach my $possible qw( /usr/local/bin/bash /bin/bash /usr/bin/bash ) {
                if (  -e $possible &&
                     !-d $possible &&
                      -r $possible &&
                      -x $possible ) {
                    $bash = $possible;
                    last;
                }
            }
            if ( !$bash ) {
                die "$me: Cannot find a bash executable";
            }
        }
    }
    return $bash;
}

# -----------------------------------------------
# Function: _get_branch_from_client
#   Routine to determine the p4 branch associated with 
#   a given p4 root directory.
#
# Usage:
#
#    my $path = SysInit::_get_branch_from_client( root => '/x/eng/rlse/DOT/devN' )
#
# Parameters:
#
#    root    => root directory of p4 client
#
# Return:
#
#    branch of p4 client
#
# -----------------------------------------------
sub _get_branch_from_client {
    my $me = (caller(0))[3];
    my %opt = validate_with(
        params => \@_,
        spec   => {
            root  => { type => SCALAR },
        },
    );

    my $root = $opt{root};
    my $P4BRANCH;

    _get_bash_shell();

    my $P4CLIENT;
    
    if (  -e "$root/P4ENV" &&
         !-d "$root/P4ENV" &&
          -r "$root/P4ENV" ) {
        $P4CLIENT = `$bash -c 'cat $root/P4ENV | grep P4CLIENT'`;
        chomp($P4CLIENT);
        if ( $P4CLIENT =~ /P4CLIENT\=(\S+)/ ) {
            $P4CLIENT = $1;
        } else {
            die "$me: Could not determine P4CLIENT from $root/P4ENV file";
        }
    } else {
        die "$me: P4 ENV file $root/P4ENV does not exist";
    }

    my @results = `$bash -c 'P4NORC=1; P4CLIENT=noclient; export P4NORC; export P4CLIENT; /usr/software/rats/bin/p4 client -o $P4CLIENT'`;

    @results = reverse @results;
    my $found = 0;
    foreach my $result ( @results ) {
        if ( $result =~ /(DOT\/.*?)\// ) {
            $P4BRANCH = $1;
            $Tharn::Log->debug(sub {"$me: P4BRANCH=$P4BRANCH determined from examination of P4 client from: $root"});
            last;
        }
        if ( $result =~ /\/\/depot\/prod\/(.*?\/.*?)\// ) {
            $P4BRANCH = $1;
            $Tharn::Log->debug(sub {"$me: P4BRANCH=$P4BRANCH determined from examination of P4 client from: $root"});
            last;
        }
    }

    if ( !$P4BRANCH ) {
        $Tharn::Log->comment("$me: Could not determine P4BRANCH from p4 client view of $P4CLIENT");
        my $p4info = $root.'/final/bedrock/export/common/p4info/p4info';
        $Tharn::Log->comment("$me: Looking for P4 Branch info in $p4info file");
        if (  -r $p4info &&
             !-d $p4info &&
              -r $p4info ) {
            open P4INFO_FILE, '<', $p4info or die $!;
            my @lines = <P4INFO_FILE>;
            my $P4CLIENT;
            foreach my $line ( @lines ) {
                if ( $line =~ /^P4BRANCH='(\S+)'/ ) {
                    $P4BRANCH = $1;
                    $Tharn::Log->comment("$me: P4BRANCH=$P4BRANCH");
                    last;
                }
            }
        }
    }

    if ( !$P4BRANCH ) {
        die "$me: Cannot determine P4 branch from path: $root";
    }

    return $P4BRANCH;
}

1;
