# -*- perl -*-
# $Id$

## @summary SysInit::Subtest
## @author  dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

SysInit::Subtest - Subtest class which supports chaining and retries

=head1 SYNOPSIS

  use SysInit::Subtest;

=head1 DESCRIPTION

This module defines routines related to Subtest object creation, chaining, execution
and results retrieval.

The Subtest object can be chained to other Subtest objects resulting in 
the sequential execution of scripts.  Also, a failure of a Subtest may
result in the re-running of that script (please see SysInit::SubtestMgr).

All Subtests are run as background processes, however the chaining allows
you to establish sequential series.

The Subtest class also exposes certain NATE Subtest methods like
kill, is_alive, isAlive, results, own_results, pid, and exit_status.

=cut

use strict;
use Options;
use NATE::OS;
use NATE::Inc qw(find_file);
use File::Basename;
use Tharn qw($Log);

package SysInit::Subtest;

our $_STAFProc_check              = 0;
our $CLIENT_CHECK                 = 1;
our $USER_CHECK                   = 2;
our $ALL_CHECKS                   = 3;
our $_is_STAFProc_running_as_root = 0;

use Data::Dumper;
use Params::Validate
    qw(validate validate_with BOOLEAN SCALAR ARRAYREF HASHREF OBJECT SCALARREF);

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => [
        {
            '-default' => undef,
            '-forward' => [qw/ kill is_alive isAlive results own_results pid exit_status scriptname /],
        },
        'subtest',
    ],
    scalar => [
        {
            '-default' => undef,
            '-type'    => "SysInit::Subtest",
        },
        'next',
    ],
    scalar => [
        {
            '-default' => 1,
        },
        'tries',
    ],
    scalar => [
        {
            '-default' => 0,
        },
        'retry_attempt',
    ],
    scalar => [
        {
            '-type'    => "NATE::Result",
        },
        'last_run_status',
    ],
    scalar => 'script',
    scalar => 'runid',
    array  => 'args',
             # redo_args is for something like deux_init and REDO=1
    array  => 'redo_args',
];

# -----------------------------------------------
# Function: init
#   Private method called by SysInit::Subtest->new();
#
# Usage:
#
#    my $st = SysInit::Subtest->new( script    => $script_relative_to_nate_lib,
#                                 runid     => $rid,
#                                 args      => \@ccl_params,
#                                 redo_args => \@ccl_params );
#
# Parameters:
#
#    script    - (required) path to script to be run
#    runid     - (optional) runid for script run
#    args      - (optional) Parameters to pass to script
#    redo_args - (optional) Parameters (if different from "args") to pass
#                           to script during a retry
#
# Return:
#
#    a SysInit::Subtest object
#
# Assumptions:
# -----------------------------------------------
sub init {
    my $self   = shift;

    # MethodMaker is a little weird in that when you pass in an
    # arrayref via the constructor, it isn't smart enough to 
    # properly initialize the array
    my @args = $self->args();
               $self->args( @{$args[0]} );

    # If someone passed in a different set of redo args (like 
    # adding REDO=1 for deux_init.thpl), use those, else
    # set the redo args to the primary args
    my @redo_args = $self->redo_args();
    if ( scalar(@redo_args) ) {
        $self->redo_args( @{$redo_args[0]} );
    } else {
        $self->redo_args( @{$args[0]} );
    }
}

# -----------------------------------------------
# Function: _incr_retry_attempt
#   Private method called by start();
#   The object keeps track of the number of retries that
#   it has performed.  This is used to limit the number of
#   retries, and to label the retries in the log directory.
#
# Usage:
#
#    $st->_incr_retry_attempt();
#
# Parameters:
#
#    None
#
# Return:
#
#    The incremented retry attempt count
#
# Assumptions:
# -----------------------------------------------
sub _incr_retry_attempt {
    my $self   = shift;
    my $retry_attempt = $self->retry_attempt();
    $retry_attempt++;
    $self->retry_attempt( $retry_attempt );
    return $retry_attempt;
}

# -----------------------------------------------
# Function: start
#   Public method to start a Subtest in the background.
#   Whether the subtest uses args or redo_args depends on
#   whether "retry" is 0 (use args) or 1 (use redo_args).
#
# Usage:
#
#    $st->start()
#
# Parameters:
#
#    retry    - (optional) whether or not this is a retry
#
# Return:
#
#    Nothing
#
# Assumptions:
# -----------------------------------------------
sub start {
    my $self = shift;
    my $me=(caller(0))[3];
    my %opts = validate_with(
        params => \@_,
        spec   => {
            retry      => { type => SCALAR, default => 0 },
            client     => { type => SCALAR, default => undef },
            user       => { type => SCALAR, default => undef },
        },
    );

    my @args;
    if ( $opts{retry} ) {
        @args = $self->redo_args();
    } else {
        @args = $self->args();
    }

    if ( !$opts{client} ) {

        if ( $_STAFProc_check != $ALL_CHECKS ) {
            &check_if_STAFProc_is_running_as_root( );
            if ( $_STAFProc_check == $USER_CHECK ) {
                $_STAFProc_check = $ALL_CHECKS;
            } else {
                $_STAFProc_check = $CLIENT_CHECK;
            }
        }

        if ( !$_is_STAFProc_running_as_root ) {
            my $different_versions = 0;
            my $args_string = join( " ", @args );

            if ( $args_string =~ /NATE_LIB=(\S+)/ ) {
                my $sub_NATE_LIB = $1;
                if ( $sub_NATE_LIB =~ /(\S+)\s+/ ) {
                    $sub_NATE_LIB = $1;
                }
            
                my @sub_nate_lib_path = split(';', $sub_NATE_LIB );
                my $lib_relative_to_sub_nate_lib = NATE::Inc::find_file( 'lib/SysInit/Subtest.pm', 
                                                                         search => [@sub_nate_lib_path] );
        
                my $our_NATE_LIB = Tharn::param("NATE_LIB");
                my @our_nate_lib_path = split(';', $our_NATE_LIB );
                my $lib_relative_to_our_nate_lib = NATE::Inc::find_file( 'lib/SysInit/Subtest.pm', 
                                                                         search => [@our_nate_lib_path] );

                $Tharn::Log->debug("$me: lib_relative_to_our_nate_lib = |$lib_relative_to_our_nate_lib|");
                $Tharn::Log->debug("$me: lib_relative_to_sub_nate_lib = |$lib_relative_to_sub_nate_lib|");

#print "FWR $me: lib_relative_to_our_nate_lib = |$lib_relative_to_our_nate_lib|\n";
#print "FWR $me: lib_relative_to_sub_nate_lib = |$lib_relative_to_sub_nate_lib|\n";


                if ( $lib_relative_to_sub_nate_lib ne $lib_relative_to_our_nate_lib ) {
                    $different_versions = 1;
                    eval {
                        my $our_p4_tree = SysInit::Subtest::DetermineP4TestAutomationBranch( $lib_relative_to_our_nate_lib );
                        my $sub_p4_tree = SysInit::Subtest::DetermineP4TestAutomationBranch( $lib_relative_to_sub_nate_lib );
    
                        if ( $our_p4_tree eq $sub_p4_tree ) {
                            $different_versions = 0;
                        }
                    };
    
                    if ( $@ ) {
                        $Tharn::Log->warn("$me Could not determine if subtest will access NATE_LIB via the same p4 client as the parent, assuming they are different, failure was |$@|");
                    }
                }
            } else {
                $Tharn::Log->warn("$me Could not determine if subtest will access NATE_LIB via the same p4 client as the parent, assuming they are different");
                $different_versions = 1;
            }

            if ( $different_versions ) {
                $Tharn::Log->comment("$me: subtest of script ".$self->script().
                                     " specified a NATE_LIB path different than ".
                                     "current NATE_LIB so must run with -client \'localhost\'");
                $opts{client} = 'localhost';
            }
        } else {
            $Tharn::Log->comment("$me: STAFProc is running as root, subtest of script ".$self->script()." will specify client=localhost");
            # If STAFProc is running as root, then no problem...
            $opts{client} = 'localhost';
        }
    }

    my @pre_args = ( -onexit => sub {}, -bg );
    if ( $opts{client} ) {
        @pre_args = ( @pre_args, -client, $opts{client} );
        if ( $_STAFProc_check != $ALL_CHECKS ) {
            &check_if_STAFProc_is_running_as_root( );
            if ( $_STAFProc_check == $USER_CHECK ) {
                $_STAFProc_check = $ALL_CHECKS;
            } else {
                $_STAFProc_check = $CLIENT_CHECK;
            }
        }
    }

    if ( $opts{user} ) {
        if ( &check_if_running_as_another_user( user => $opts{user} ) ) {
            @pre_args = ( @pre_args, -username, $opts{user} );
            if ( $_STAFProc_check != $ALL_CHECKS ) {
                &check_if_STAFProc_is_running_as_root( );
                if ( $_STAFProc_check == $CLIENT_CHECK ) {
                    $_STAFProc_check = $ALL_CHECKS;
                } else {
                    $_STAFProc_check = $USER_CHECK;
                }
            }
        }
    }

    @args = ( @pre_args, @args );

    if ( $self->runid() ) {
        my $runid = $self->runid();
        if ( $opts{retry} ) {
            $runid .= "_retry_".sprintf("%02d", $self->_incr_retry_attempt());
        }
        @args = ( '-runid', $runid, @args );
    }
#print "FWR $me args = ".Dumper(\@args)."\n";
    $self->subtest( Tharn::subtest( $self->script(), @args ) );
}

# -----------------------------------------------
# Function: check_if_running_as_another_user
#   If the user running sys_init (or the sys_init vsim scripts)
#   is different than the user who owns the vsim, STAFProc MUST 
#   be running as root.
#   We check here so that we can give a more meaningful error message
#   than the more generic STAF error message.
#
# Usage:
#
#    $st->check_if_running_as_another_user( user => $user )
#
# Parameters:
#
#    user    - (required) Username
#
# Return:
#
#    0|1
#
# -----------------------------------------------
sub check_if_running_as_another_user {
    my $me=(caller(0))[3];
    my $rc = 0;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            user       => { type => SCALAR },
        },
    );
    my $curr_user = NATE::OS::login_name();
    my $curr_uid  = getpwnam( $curr_user );
    my $uid       = getpwnam( $opts{user} );

    $Tharn::Log->debug("$me: curr_user=$curr_user, run_as_user=".$opts{user}.", curr_uid=$curr_uid, run_as_uid=$uid");
    if ( $uid != $curr_uid ) {
        $rc = 1;
    }

    return $rc;
}

# -----------------------------------------------
# Function: check_if_STAFProc_is_running_as_root
#   Check to see if STAFProc is running as root.
#   Try to only check when the user running the 
#   script is different from the vsim owner OR
#   when we would need to specify a different 
#   NATE_LIB to load an Ontap build other than
#   the one associated with the main script automation
#   branch.  (i.e. if sys_init is run from the main
#   automation branch (SN) and if you want to build 
#   RR.1 vsims, then sys_init needs to specify a
#   different NATE_LIB than what sys_init sees)
#
# Usage:
#
#    $st->check_if_STAFProc_is_running_as_root( force => <0|1> )
#
# Parameters:
#
#    force    - (optional) Do check even if we have done it before
#
# Return:
#
#    No return, just issue a warning if not running as root.
#
# -----------------------------------------------
sub check_if_STAFProc_is_running_as_root {
    my $me=(caller(0))[3];
    my $rc = 0;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            force       => { type => SCALAR, default => undef },
        },
    );

    if ( ($_STAFProc_check != $ALL_CHECKS) || ($opts{force}) ) {
        eval {
            my $STAFProc = `ps -aef | grep STAFProc | grep -v grep | grep '^root' | grep -v " su "`;
            if ( !$STAFProc ) {
                $Tharn::Log->warn("$me: STAF process authentication errors may result because ".
                           "STAFProc is not running as root.  ".
                           "Please see http://wikid/w/Coati/SysInit#A_Note_about_Vsims");
                $_is_STAFProc_running_as_root = 0;
            } else {
                $_is_STAFProc_running_as_root = 1;
            }
        };
        if ( $@ ) {
            $Tharn::Log->warn("$me: Could not determine how STAFProc is running $@");
            $Tharn::Log->warn("$me: STAF process authentication errors may result because ".
                       "STAFProc is not running as root.  ".
                       "Please see http://wikid/w/Coati/SysInit#A_Note_about_Vsims");
        }
    }
}

# -----------------------------------------------
# Function: DetermineP4TestAutomationBranch
#   Used to determine what p4 test automation 
#   branch the file is in.
#   The assumption is that if 2 files, even if they are
#   from different paths, if they are from clients of the
#   same test automation branch, we do not have to 
#   run the subtest with the client => localhost option
#   which will force a reload of all library files from
#   the second p4 path.  This is to allow people who do
#   NOT have STAFProc running as root to run without
#   encountering authentication errors.
#   If someone wants to assure that they are picking up
#   the libraries that they want, they can always run from
#   a host that has STAFProc running as root, and/or specify
#   TEST_BRANCH_* so it matches the NATE_LIB of the 
#   parent script (typically sys_init.thpl)
#
# Usage:
#
#    my $p4_root = DetermineP4TestAutomationBranch( $full_path_to_p4_file );
#
# Parameters:
#
#    full_path_to_p4_file    - (required) Fully specified path to a file in a p4 client
#
# Return:
#
#    Returns the root directory of the p4 depot that client is using
#
# -----------------------------------------------
sub DetermineP4TestAutomationBranch {
    my $me=(caller(0))[3];
    my $full_path = shift;
    my $root = $full_path;

    my ($file, $path) = File::Basename::fileparse( $full_path );

    my $cmd = "cd $path; p4 where $file";
    my $return = `$cmd`;

    if ( $return =~ /^(\/\/depot\/prod\/test\/\S+?)\// ) {
        $root = $1;
    } else {
        die "$me Could not determine p4 Path of $full_path";
    }
    return $root;
}

1;
