# -*- perl -*-
# $Id$

## @summary SysInit::SubtestMgr
## @author  dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

SysInit::SubtestMgr - SubtestMgr class which supports chaining and retries
                   of SysInit::Subtest objects

=head1 SYNOPSIS

  use SysInit::SubtestMgr;

=head1 DESCRIPTION

This module defines routines related to running SysInit::Subtest objects,
following the chain of related Subtests (in terms of sequential execution)
and retrying failed subtest.

=cut

use strict;
use Options;
use Tharn qw($Log);

package SysInit::SubtestMgr;

# $Log->debug() strips off leading spaces making Dumper
# output harder to read
sub _debug {
    my $message = shift;

    if ( $Tharn::Log->may_debug() ) {
        print $message."\n";
    }
}


use Data::Dumper;
use Params::Validate
    qw(validate validate_with BOOLEAN SCALAR ARRAYREF HASHREF OBJECT SCALARREF);
use SysInit::Subtest;

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    array  => 'subtests',
];

# -----------------------------------------------
# Function: init
#   Private method called by SysInit::SubtestMgr->new();
#
# Usage:
#
#    my $stm = SysInit::SubtestMgr->new( subtests => @subs );
#
# Parameters:
#
#    subtests  - (required) Array of subtests to run
#
# Return:
#
#    a SysInit::SubtestMgr object
#
# Assumptions:
# -----------------------------------------------
sub init {
    my $self   = shift;

    # MethodMaker is a little weird in that when you pass in an
    # arrayref via the constructor, it isn't smart enough to
    # properly initialize the array
    my @subtests = $self->subtests();
                   $self->subtests( @{$subtests[0]} );
}

# -----------------------------------------------
# Function: start
#   Public method to start the managed Subtests in the background.
#
# Usage:
#
#    $st->start()
#
# Parameters:
#
#    sleep    - (optional) Number of seconds to sleep between Subtest
#                          starts (default = 0)
#
# Return:
#
#    Nothing
#
# Assumptions:
# -----------------------------------------------
sub start {
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                               # Number of seconds to sleep between starts
            sleep           => { type => SCALAR, default => 0 },
        },
    );

    foreach my $s (@{$self->subtests()}) {
        $s->start( );

        if ( $opts{sleep} ) {
            Tharn::sleep( $opts{sleep} );
        }
    }
}

# -----------------------------------------------
# Function: wait_until_finished
#   Public method to wait until all of the background
#   subtests have completed.
#   Also allows retries if a script fails.
#
# Usage:
#
#    $st->wait_until_finished()
#
# Parameters:
#
#    worst_allowable - (optional) Class of type NATE::Result which specifies
#                                 the worst result a script can issue and still
#                                 be considered a success (defaut: NATE::Result::NA)
#    sleep           - (optional) Seconds to sleep between checks (default: 1)
#    retries         - (optional) If script fails, attempt to retry this many
#                                 times (default: 0)
#    kill_remaining  - (optional) If script fails for the last time, let the
#                                 remaining subtests that are running in 
#                                 parallel run to completion (0), or kill them (1)?
#                                 (default: 1)
#
# Return:
#
#    Nothing
#
# Assumptions:
# -----------------------------------------------
sub wait_until_finished {
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                                    # Any result with a serverity greater than this will be
                                    # considered a failure (i.e. Inspect and above)
            worst_allowable      => { isa => 'NATE::Result', default  => 'NATE::Result::NA' },

                                    # Number of seconds to sleep between checks
            sleep_between_checks => { type => SCALAR, default => 1 },

                                    # Number of seconds to sleep between retries
            sleep_between_retries => { type => SCALAR, default => 1 },

                                    # If a subtest fails, try to rerun it this many times
            retries              => { type => SCALAR, default => 0 },

                                    # If a subtest fails (and no more retries), should the remaining
                                    # subtests be killed, or left to complete on their own?
            kill_remaining       => { type => SCALAR, default => 1 },
        },
    );

    my @subs            = @{$self->subtests()};
    my $worst_allowable = $opts{worst_allowable};

    # Set initial number of retries for all subtests
    # including children
    foreach my $sub ( @subs ) {
        $sub->tries($opts{retries});
        my $next = $sub->next();
        while ( $next ) {
            $next->tries($opts{retries});
            $next = $next->next();
        }
    }

    my $done = 0;
    my @s = @subs;
    my @running;
    while ( !$done ) {
        $done = 1;
        @running = ();
        foreach my $s (@s) {
            if ($s->subtest()) {
                if (!$s->isAlive()) {
                    # Need to get own_results() because some scripts
                    # (like create_vsims.thpl) purposely mask failures
                    # and try again, so we only want the parent's errors,
                    # not all of the children's.
                    my @subResults = $s->own_results();
                    my $worst = NATE::Result::worst(@subResults);

                    $s->last_run_status( $worst );
                    _debug("Subtest: ".Dumper($s));
                    _debug("    Worst: ".Dumper($worst));
                    my $scriptname = $s->scriptname();
    
                    if ( $worst && 
                        ($worst->isa( 'NATE::Result' ) ) &&
                        ($worst->severity() > $worst_allowable->severity()) ) {

                        my $retries = $s->tries();
                        if ( $retries ) {
                            my $worst_type = $worst->type();
                            my $result = NATE::Result::Warn->new( "Child script, ".
                                                                  $scriptname.
                                                                  " had worst result of type ".
                                                                  $worst_type );

                            $result->log();
                            my $error_string = join(" ", $worst->messages() );
                            if ( $error_string =~ /ProcessAuthenticationDenied/ ) {
                                SysInit::Subtest::check_if_STAFProc_is_running_as_root( force => 1 );
                            }

                            $retries--;
                            $s->tries($retries);

                            Tharn::sleep( $opts{sleep_between_retries} );
    
                            $s->start( retry => 1 );
                            push( @running, $s );
                            $done = 0;
                        } else {
                            if ( $opts{kill_remaining} ) {
                                # die immediately and let the remaining children get
                                # cleaned up
                                my $error_string = join(" ", $worst->messages() );
                                if ( $error_string =~ /ProcessAuthenticationDenied/ ) {
                                    SysInit::Subtest::check_if_STAFProc_is_running_as_root( force => 1 );
                                }
                                die $worst;
                            } else {
                                $worst->log();
                            }
                        }
                    } else {
                        if ($s->next()) {
                            my $next = $s->next();
                            $next->start();
                            push( @running, $next );
                            $done = 0;
                        }
                    }

                } else {
                    push( @running, $s );
                    $done = 0;
                }
            } else {
                push( @running, $s );
                $done = 0;
            }
        }
        @s = @running;
        Tharn::sleep( $opts{sleep_between_checks} );
    }
}

1;
