# -*- perl -*-
# $Id$

## @summary SysInit::Vsim
## @author  dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

SysInit::Vsim - Class which interacts with vsims

=head1 SYNOPSIS

  use SysInit::Vsim;

=head1 DESCRIPTION

This module defines routines related to vsim creation, deletion and manipulation.

Note: 

No attempt is made here to document all vsim functionality.  For that please
see the following:

http://thebrewery-web.corp.netapp.com/Brewery/brewery/dev/Tools/VSim/index.html

=cut

package SysInit::Vsim;

use strict;
use NATE::OS;
use Tharn qw($Log);
use Data::Dumper;
use Connectrec qw();
use Hostrec qw();
use IPC::Cmd qw(can_run);
use NACL::C::Options;
use NACL::C::SystemNodeDate;
use NACL::C::SystemServicesNtpConfig;
use NACL::STask::Node;
use NACL::Transit;
use Tie::IxHash;
use NACL::GeneralUtils;
use NATE::BaseException qw(:try);
use Params::Validate
    qw(validate validate_with BOOLEAN SCALAR ARRAYREF HASHREF OBJECT SCALARREF);

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => [
        {   '-default_ctor' => sub { $_[0]->_get_node_obj() },
            '-type' => 'NACL::STask::Node',
        },
    'naclnode'
    ],
    scalar => [ 'cmode' ],
    scalar => [ 'node' ],
    scalar => [ 'saved_env_HOME' ],
    scalar => [ 'saved_env_USER' ],
    scalar => [ 'uhome' ],
    scalar => [ 'user' ],
    scalar => [ 'user_hash' ],
    scalar => [ 'id' ],
    scalar => [ 'blade' ],
    scalar => [ 'instance' ],
    scalar => [ 
        {   '-default_ctor' => sub { $_[0]->_create_transit() },
            '-type' => 'NACL::Transit',
        },
    'tran'
    ],
    scalar => [ 
        {   
            '-type' => 'Hostrec',
        },
    'hostrec'
    ],
    scalar => [ 
        {   
            '-default' => 0,
        },
    'ha'
    ],
    scalar => [ 
        {   
            '-default' => undef,
        },
    'num_nics'
    ],
];

our $exec_cmd       = 'tools/sys_init/vsim/utility/exec_cmd.thpl';
our $unstick_cmd    = 'tools/sys_init/vsim/utility/unstick_cmd.thpl';
our $vsim_cmd;
our %supports;

my $run_user  = NATE::OS::login_name();
my $run_uid   = getpwnam($run_user);
my $run_uhome = (getpwuid($run_uid))[7];

our $saved_env_HOME;
our $saved_env_USER;

if ( !$saved_env_HOME ) {

    $saved_env_HOME = $ENV{HOME};

    if ( !$saved_env_HOME ) {
        $Log->comment("SysInit::Vsim: ENV{HOME} not defined");
        $saved_env_HOME = Tharn::param("HOME");
        if ( $saved_env_HOME ) {
            $Log->comment("SysInit::Vsim: Setting ENV{HOME} with value of NATE HOME parameter($saved_env_HOME)");
            $ENV{HOME} = $saved_env_HOME;
        } elsif ( $run_uhome ) {
            $Log->comment("SysInit::Vsim: Setting ENV{HOME} with value of HOME associated with run user($run_uhome)");
            $ENV{HOME}      = $run_uhome;
            $saved_env_HOME = $run_uhome;
        } else {
            die "SysInit::Vsim: The vsim command requires that HOME and USER environment variables be set, but HOME is not set in ENV or as a NATE parameter";
        }
    }
}

if ( !$saved_env_USER ) {

    $saved_env_USER = $ENV{USER};

    if ( !$saved_env_USER ) {
        $Log->comment("SysInit::Vsim: ENV{USER} not defined");
        $saved_env_USER = Tharn::param("USER");
        if ( $saved_env_USER ) {
            $Log->comment("SysInit::Vsim: Setting ENV{USER} with value of NATE USER parameter($saved_env_USER)");
            $ENV{USER} = $saved_env_USER;
        } elsif ( $run_user ) {
            $Log->comment("SysInit::Vsim: Setting ENV{USER} with value of USER associated with run user($run_user)");
            $ENV{USER}      = $run_user;
            $saved_env_USER = $run_user;
        } else {
            die "SysInit::Vsim: The vsim command requires that HOME and USER environment variables be set, but USER is not set in ENV or as a NATE parameter";
        }
    }
}

# CTL and SMOKES use this variable to say where the vsim executable is.
# allows them to switch to an older version if necessary
my $vsim_script = Tharn::param('VSIM_SCRIPT');

if ( $vsim_script ) {
    $Log->comment("SysInit::Vsim: Was told to use the \"vsim\" command from VSIM_SCRIPT param, $vsim_script");
    $vsim_cmd = $vsim_script;
} else {
    $vsim_cmd = can_run('vsim');
}

# If not told via VSIM_SCRIPT, and can_run() can't find "vsim" in execution path, default it
if ( !$vsim_cmd ) {
    my $default_cmd = '/usr/software/test/bin/vsim';
    $Log->warn("SysInit::Vsim: Cannot find \"vsim\" command in execution PATH or VSIM_SCRIPT param, defaulting to $default_cmd");
    $vsim_cmd = $default_cmd;
}

my $VSIM_OVERRIDE_VERSION = Tharn::param('VSIM_OVERRIDE_VERSION');

# See if vsim command supports HA
if ( ! exists $supports{new_HA} ) {
    $Log->comment("$vsim_cmd version");
    my $result  = `$vsim_cmd version`;
    chomp $result;

    if ( $VSIM_OVERRIDE_VERSION ) {
        $Log->comment("Raw version of $vsim_cmd is |$result|" );
    }

    if ( $result =~ /Version: (\S+)/ ) {
        $result = $1;
    }

    $Log->comment("Version returned by $vsim_cmd is $result");

    if ( $VSIM_OVERRIDE_VERSION ) {
        $Log->comment("Version dictated by VSIM_OVERRIDE_VERSION parameter is $VSIM_OVERRIDE_VERSION");
        $result = $VSIM_OVERRIDE_VERSION;
    }

    my ($major, $minor, $rev);
    if ( $result =~ /(\d+)\.(\d+)\.(\d+)/ ) {
        $major = $1;
        $minor = $2;
        $rev   = $3;

        my $actual_rev     = $major*1000000 + $minor*1000 + $rev;
        my $ha_rev         =     18*1000000 +      8*1000 + 7;
        my $noicupdate_rev =     19*1000000 +      1*1000 + 0;
        my $mc_rev         =     20*1000000 +     11*1000 + 0;
        $Log->comment("$ha_rev : Version that supports new HA");
        $Log->comment("$noicupdate_rev : Version that supports hagroup -noicupdate flag");
        $Log->comment("$mc_rev : Version that supports mcgroups");

        $supports{new_HA}             = 0;
        $supports{hagroup_noicupdate} = 0;
        $supports{mcgroups}           = 0;

        if ( $actual_rev >= $ha_rev ) {
            $supports{new_HA} = 1;
        }

        if ( $actual_rev >= $noicupdate_rev ) {
            $supports{hagroup_noicupdate} = 1;
        }

        if ( $actual_rev >= $mc_rev ) {
            $supports{mcgroups} = 1;
        }

        # FWR Eventually when HA vsims support snapshots, add code for that check
        # in here too
    } elsif ($VSIM_OVERRIDE_VERSION) {
        $Log->warn("VSIM_OVERRIDE_VERSION must be given in the form major.minor.revision");
    } else {
        die "$vsim_cmd returned version string |$result| which is not in the form major.minor.revision where all components are decimal integers";
    }
}

our @nodes;
our $simdomain;

# The "vsim" scripts look for $ENV{VSIM_DEFAULT_SERVER_DOMAIN}
my $VSIM_DOMAIN = Tharn::param("VSIM_DEFAULT_SERVER_DOMAIN") || Tharn::param("VSIM_DOMAIN");
if ( $ENV{VSIM_DEFAULT_SERVER_DOMAIN} ) {
    $simdomain = $ENV{VSIM_DEFAULT_SERVER_DOMAIN};
} elsif ( $VSIM_DOMAIN ) {
    $simdomain = $VSIM_DOMAIN;
} else {
    # This is all just for the default domain
    my $domainname_cmd = can_run('domainname');
    if ( !$domainname_cmd ) {
        my $default_cmd = '/bin/domainname';
        $Log->comment("SysInit::Vsim: Cannot find \"domainname\" command in execution PATH, defaulting to $default_cmd");
        $domainname_cmd = $default_cmd;
    }
    
    my $hostname_cmd   = can_run('hostname');
    if ( !$hostname_cmd ) {
        my $default_cmd = '/bin/hostname';
        $Log->comment("SysInit::Vsim: Cannot find \"hostname\" command in execution PATH, defaulting to $default_cmd");
        $hostname_cmd = $default_cmd;
    }
    
    $Log->comment("$domainname_cmd");
    my $domain  = `$domainname_cmd`;
    chomp($domain);
    
    $Log->comment("$hostname_cmd");
    my $hostname = `$hostname_cmd`;
    chomp($hostname);

    $simdomain = "sim.netapp.com";

    my $name_to_check = $domain;
    if ( $hostname =~ /\.com$/ ) {
        $name_to_check = $hostname;
    }
    
    if ( $name_to_check =~ /rtp/ ) {
        $simdomain = "sim.rtp.netapp.com";
    } elsif ( $name_to_check =~ /spin/ ) {
        $simdomain = "data.spin.eng.netapp.com";
    } elsif ( $name_to_check =~ /btc/ ) {
        $simdomain = "sim.eng.btc.netapp.in";
    } elsif ( $name_to_check =~ /nane/ ) {
        $simdomain = "sim.nane.netapp.com";
    }
}

our $serial   = 1;

# -----------------------------------------------
# Function: get_seq
#   convenience routine to provide a serial (sequence) number
#   for subtest names
#
# Usage:
#
#    my $seq = $vsim->get_seq()
#
# Parameters:
#   
#    None
#
# Return:
#   
#    serial string in the range 0001 to 9999 incrementing each time
#    it is called.
#
# Assumptions:
# -----------------------------------------------
sub get_seq {
    my $self = shift;
    my $seq = sprintf( "%04d", $serial ); $serial++;
    return $seq;
}

# -----------------------------------------------
# Function: SFOPAIR
#   return value of SFOPAIR string
#
# Usage:
#
#    my $sfopair = $vsim->SFOPAIR();
#
# Parameters:
#   
#    None
#
# Return:
#   string
#
# Assumptions:
# -----------------------------------------------

sub SFOPAIR {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    $self->naclnode()->hostrec()->{SFOPAIR};
}

# -----------------------------------------------
# Function: _get_node_obj
#   return a NACL::STask::Node object
#
# Usage:
#
#    my $node = $vsim->_get_node_obj()
#
# Parameters:
#   
#    None
#
# Return:
#   NACL::STask::Node object
#
# Assumptions:
# -----------------------------------------------

sub _get_node_obj {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    foreach my $node (@nodes) {
        if ( $node->node() eq $self->node() ) {
            $self->naclnode( $node );
            last;
        }
    }
    $self->{naclnode};
}

# -----------------------------------------------
# Function: _create_transit
#   return a NACL::Transit object
#
# Usage:
#
#    my $trans = $vsim->_create_transit()
#
# Parameters:
#
#    None
#
# Return:
#   NACL::Transit object
#
# Assumptions:
# -----------------------------------------------

sub _create_transit {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    return NACL::Transit->new( name => $self->hostrec() );
}

# -----------------------------------------------
# Function: init
#   
#  routine used by the constructor
#
#  Note: Since there is not an explicit constructor
#        defined in this code (we're using init() and
#        MethodMaker's default constructor) I will 
#        provide the usage for the contructor here.
#
# Usage:
#   my $vsim = new( node => "ptcqa-vsim2" )
#
# Parameters:
#       node      - Name of the vsim
#
# Return: blessed reference
# -----------------------------------------------
sub init {
    my $self = shift;
    my $me=(caller(0))[3];

    my %opt = ( node => undef,
                @_ );

    my $node = $opt{node};

    # in 7mode, or even Cmode, people can pass in 
    # hostrec overrides, like "radish-vsim7;hosttype=filer-vsim"
    #   or                    "radish-vsim7;MGMT_PORT=e0f"
    #   or                    "radish-vsim7{MGMT_PORT=e0f}"

    # The assumption is that the hosttype stored in the Hostrec file
    # will be the CMode hosttype, and that 7Mode will be accomplished by
    # specifying hosttype=filer-vsim as a Hostrec override.
    my %nodespecs;
    tie %nodespecs, 'Tie::IxHash';

    %nodespecs = NACL::GeneralUtils::_parse_hosts( $node );
    my @nodespec_keys = (keys %nodespecs);
    $node = $nodespec_keys[0];

    if ( $nodespecs{$node} =~ /filer-vsim/ ) {
        $Log->comment("$me: The hosttype indicates $node is a 7Mode vsim");
        $self->cmode(0);
    } else {
        $Log->comment("$me: The hosttype indicates $node is a CMode vsim");
        $self->cmode(1);
    }

    $self->node( $node );

    my $hostrec;
    if ( $self->cmode() ) {
        $hostrec = Hostrec->new( id => $node );
    } else {
        $hostrec = Hostrec->new( id => $node, hosttype => 'filer-vsim' );
        if ( $hostrec->{hosttype} ne 'filer-vsim' ) {
            $Log->error("$me: For some reason 7mode filer has hosttype of ".$hostrec->{hosttype}." instead of filer-vsim");
            $hostrec->{hosttype} = 'filer-vsim';
        }
    }

    $self->hostrec( $hostrec );

    if ( !@nodes ) {
        @nodes       = NACL::STask::Node->find();
    }

    # Handle both public "radish-vsim3" and private "sti1-vsim-ucs049a" forms
    if ( $node =~ /^(.+?)[_-]vsim[_-]*(.+)$/ ) {

        $self->user( $1  );
        $self->id(   $2 );

        if ( $self->id() =~ /^(.+)([a-z])$/ ) {
            $self->blade( $1 );
            $self->instance( $2 );
        }
    } else {
        die "$me: Node name must be in the form <userid>-vsim<id> or <userid>-vsim-<id>";
    }

    my $uid   = getpwnam($self->user());
    my $uhome = (getpwuid($uid))[7];

    my $run_user  = NATE::OS::login_name();
    my $run_uid   = getpwnam($run_user);

    $self->uhome( $uhome );

    $self->saved_env_HOME( $ENV{HOME} );
    $self->saved_env_USER( $run_user );

    my %user_opts = ();

    if ( $uid != $run_uid ) {
        eval {
            my $STAFProc = `ps -aef | grep STAFProc | grep -v grep | grep '^root' | grep -v " su "`;
            if ( !$STAFProc ) {
                $Log->warn("$me: Vsim owned by |".$self->user()."|, user actually running this method is |".
                           $run_user."|, but STAFProc is not running as root.  ".
                           "Please see http://wikid/w/Coati/SysInit#A_Note_about_Vsims");
            }
        };
        if ( $@ ) {
            $Log->warn("$me: Could not determine how STAFProc is running $@");
            $Log->warn("$me: Vsim owned by |".$self->user()."|, user actually running this method is |".
                       $run_user."|, but STAFProc is not running as root.  ".
                       "Please see http://wikid/w/Coati/SysInit#A_Note_about_Vsims");
        }

        $user_opts{'-username'} = $self->user();
    }
    $self->user_hash( \%user_opts );

    $self;
}

# -----------------------------------------------
# Function: exists
#   
#  Has a vsim of this name been created?
#
# Usage:
#   $vsim->exists()
#
# Parameters: None
#
# Return: 0 | 1
# -----------------------------------------------
sub exists {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my $user  = $self->user();
    my $uhome = $self->uhome();
    my $id    = $self->id();

    $Log->comment("$me: USER=$user;HOME=$uhome;$vsim_cmd state $id");
    my $result       = `USER=$user;HOME=$uhome;$vsim_cmd state $id`;
    chomp $result;

    if ( $result !~ /does not exist/ ) {
        return 1;
    }

    return 0;
}

# -----------------------------------------------
# Function: spoof_user_env
#   
#  Set ENV{HOME} and ENV{USER} to look like
#  we are run as the owner of vsim.
#
#  The vsim command actually looks at ENV for HOME and 
#  USER, even if you are subtesting and specifying the 
#  user as which to run the subtest correctly.
#
# Usage:
#   $vsim->spoof_user_env()
#
# Parameters:
#       Node
#
# Return: nothing
# -----------------------------------------------
sub spoof_user_env {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    $ENV{HOME} = $self->uhome();
    $ENV{USER} = $self->user();
}

# -----------------------------------------------
# Function: restore_user_env
#   
#  Restore ENV{HOME} and ENV{USER} settings
#  as they were when we first created the vsim.
#
#  The vsim command actually looks at ENV for HOME and 
#  USER, even if you are subtesting and specifying the 
#  user as which to run the subtest correctly.
#
# Usage:
#   $vsim->restore_user_env()
#
# Parameters:
#       Node
#
# Return: nothing
# -----------------------------------------------
sub restore_user_env {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    $ENV{HOME} = $self->saved_env_HOME();
    $ENV{USER} = $self->saved_env_USER();
}

# -----------------------------------------------
# Function: make
#   
#  Make a vsim.  All "vsim make" parameters can
#  be passed via the create_args option, but only
#  a small subset are passable via the other named
#  options.  The create_args option is not to be used
#  with any of the other options.
#
# Usage:
#   $vsim->make( num_nics => 8 );
#
# Parameters:
#       
#   bubble_net        => Flag which indicates we're working on vsims which have their own ESX server
#                        and are not in the public vsim pool
#   build_root        => Root of the Ontap build tree
#   create_args       => If you want full control over the arguments passed to make
#                        you must include all arguments except the vsim index.
#   extra_create_args => If you just want to add some args to the defaults.
#   ddsize            => Data disk size  (default "1g" unless you specify -vstore options in "extra_create_args")
#   ddcount           => Data disk count (default "25" unless you specify -vstore options in "extra_create_args")
#   ha                => Make this an ha vsim
#   ibm               => Build with an IBM build? [0 | 1] (default 0)
#   num_nics          => Number of ports for vsim, if not specified num_nics=4, [5 | 6 | 7 | 8]
#   mode              => Ontap Mode [10m | 7m]
#   platform          => [x86_64 | x86_64.gcov | x86_64.debug | x86_64.debug.gcov]
#   smoke             => Flag which indicates we're working on vsims which have their own ESX server
#                        and are not in the public vsim pool
#   setenv            => Comma separated list of settings, variable1=setting1,variable2=setting2 (default "AUTOBOOT=false")
#   timeout           => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded make.
# -----------------------------------------------
sub make {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( build_root        => undef,
                ibm               => 0,
                mode              => "10m",
                platform          => "x86_64",
                num_nics          => undef,
                ddsize            => undef,
                ddcount           => undef,
                ha                => undef,
                smoke             => undef,
                bubble_net        => undef,
                create_args       => undef,
                extra_create_args => undef,
                setenv            => 'AUTOBOOT=false',
                timeout           => 900,
                @_ );

    my $alternate_args;

    if ( !$opt{create_args} ) {
        my ($build,$mode,$flavor);
        if ( !$opt{build_root} ) {
            die "$me If \"create_args\" is not specified, \"build_root\" must be";
        } else {
            $build = $opt{build_root};
        }

        if ( $opt{ibm} ) {
            if ( $opt{platform} !~ /ibm/ ) {
                $opt{platform} = 'x86_64.sim.ibm';
            }
        }

        $opt{create_args} = "-pcons -quickl -".$opt{mode}." -px ".$opt{platform}." -ir ".$build;
    }

    # If someone just wants to add something to the defaults
    if ( $opt{extra_create_args} ) {
        $opt{create_args} .= " ".$opt{extra_create_args};
    }

    if ( $opt{ha} ) {
        if ( !$supports{new_HA} ) {
            NATE::BaseException->throw("Attempting to create an HA vsim using $vsim_cmd which does not have the \"new\" HA support");
        } else {
            # There seems to be a file in buildroot that gives a version
            #     cat /x/eng/rlse/DOT/mainN/VSIMVER
            #     vsim_feature_version:3.1
            # probably just need to check to see if that version is >= 3.2
            # No point in adding this check before vsim.beta goes live.
            my $build;
            if ( $opt{build_root} ) {
                $build = $opt{build_root};
            } elsif ( $opt{create_args} =~ /-ir\s+(\S+)\s?/ ) {
                $build = $1;
            }
            if ( $build ) {
                my $build_vsim_ver_file = $build."/VSIMVER"; 
                if ( -e $build_vsim_ver_file ) {
                    $Log->comment("$me: grep vsim_feature_version: $build_vsim_ver_file");
                    my $result       = `grep vsim_feature_version: $build_vsim_ver_file`;
                    chomp $result;
                    if ( $result =~ /vsim_feature_version:(\S+)/ ) {
                        my $ontap_vsim_version = $1;
                        # Check if version is >= 3.2
                        if ( $ontap_vsim_version < 3.2 ) {
                            NATE::BaseException->throw("Cannot create an HA vsim because ".
                                       "ontap_vsim_version in $build_vsim_ver_file reports ".
                                       "$ontap_vsim_version which is less than 3.2");
                        }
                    } else {
                        $Log->warn("$me: $build_vsim_ver_file does not contain ".
                                   "vsim_feature_version: token, cannot verify if ".
                                   "Ontap build in $build supports new vsim HA");
                    }

                } else {
                    $Log->warn("$me: $build_vsim_ver_file does not exist, cannot ".
                                   "verify if Ontap build in $build supports new vsim HA");
                }
            } else {
                $Log->warn("$me: Cannot verify if Ontap build in $build supports new ".
                                   "vsim HA because cannot determine Ontap build directory");
            }

            $self->ha( 1 );
        }
    }

    if ( $opt{ibm} ) {
        $opt{create_args} .= " -ibm";
    }

    if ( $opt{num_nics} ) {
        my $num_nics = $opt{num_nics};
        if ( ($num_nics < 5) || ($num_nics > 8) ) {
            NATE::BaseException->throw("VSIM supports num_nics between 5 and 8, value supplied was $num_nics");
        }
        $self->num_nics($num_nics);
        if ( $self->ha() ) {
            $Log->comment("HA Vsims consume one nic, so nics available will be ".$num_nics-1);
        }
        $opt{create_args} .= " -nics ".$num_nics;
    } elsif ( $self->ha() ) {
        $self->num_nics(5);
        $opt{create_args} .= " -nics 5";
    }

    my $default_ddsize  = '1g';
    my $default_ddcount = 25;

    my $ddsize = $opt{ddsize};
    if ( !$ddsize ) {
        if ( $opt{create_args} !~ /-vstore/ ) {
            $ddsize = $default_ddsize;
            $opt{create_args} .= " -ddsize ".$ddsize;
        }
    } else {
        if ( $opt{create_args} !~ /-vstore/ ) {
            $opt{create_args} .= " -ddsize ".$ddsize;
        } else {
            die "$me: Cannot specify -vstore and -ddsize";
        }
    }

    my $ddcount = $opt{ddcount};
    if ( !$ddcount ) {
        if ( $opt{create_args} !~ /-vstore/ ) {
            $ddcount = $default_ddcount;
            $opt{create_args} .= " -ddcount ".$ddcount;
        }
    } else {
        if ( $opt{create_args} !~ /-vstore/ ) {
            $opt{create_args} .= " -ddcount ".$ddcount;
        } else {
            die "$me: Cannot specify -vstore and -ddcount";
        }
    }

    if ( $opt{platform} =~ /debug/ ) {
        $opt{create_args} .= " -d ";
    } else {
        $opt{create_args} .= " -nod ";
    }

    if ( $opt{bubble_net} ) {
        $opt{create_args} .= " -bubble_net ";
    }

    if ( $opt{smoke} ) {
        $opt{create_args} .= " -smoke ";
    }

    if ( $opt{setenv} ) {
        $opt{create_args} .= " -ev ".$opt{setenv};
    }

    # Not all vsims support ddsize and ddcount and rather than just ignoring them,
    # they fatal out!  So specify an alternate command to make without these options.
    $alternate_args = $opt{create_args};
    $alternate_args =~ s/-ddsize\s+\S+//g;
    $alternate_args =~ s/-ddcount\s+\S+//g;

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd make ".$opt{create_args}." ".$self->id(),
                    "ALTERNATE_COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd make ".$alternate_args." ".$self->id(),
                    #"EXPECTED=simulator created",
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_make_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: setenv
#
#  Invoke the "setenv" command on a vsim.
#
# Usage:
#
#   my $st = $vsim->setenv( setting => 'AUTOBOOT=false');
#   Subtest::wait_finish( subtests => [$st] );
#
# Parameters:
#
#   setting     => variable=value
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub setenv {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( setting => undef,
                smoke   => undef,
                timeout => 600,
                @_ );

    if ( !$opt{setting} || ($opt{setting} !~ /=/) ) {
        die "$me: setting must be provided in the form setting => \"VARIABLE=setting\"";
    }
    my $extra_args = "";
    if ( $opt{smoke} ) {
        $extra_args .= " -smoke";
    }

    my @setting = split('=', $opt{setting});
    my $variable = $setting[0];
    my $setting  = $setting[1];

    my $node     = $self->node();
    my $id       = $self->id();

    $self->spoof_user_env();

    my $user   = $self->user();
    my $uhome  = $self->uhome();

    $Log->comment("$me: USER=$user;HOME=$uhome;$vsim_cmd setenv $extra_args $id $variable $setting");
    my $result       = `USER=$user;HOME=$uhome;$vsim_cmd setenv $extra_args $id $variable $setting`;
    chomp $result;

    if ( $result ) {
        print "$me: raw setenv output = ".$result."\n";
    }

    $self->restore_user_env();
}

# -----------------------------------------------
# Function: state
#   
#  Invoke the "state" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   my $st = $vsim->state();
#   Subtest::wait_finish( subtests => [$st] );
#   @results = $st->results();
#   my $misc = $results[0]->misc();
#   my $out  = $misc->{VSIM_CMD_OUTPUT};
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub state {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 600,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    my $expected = "STATE";

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd state ".$self->id(),
                    #"EXPECTED=$expected",
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-bg',
                    '-onexit', sub { },
                    '-timeout', $opt{timeout},
                    '-runid', $self->get_seq()."_state_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: start
#   
#  Invoke the "start" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->start() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub start {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 900,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();
    my $blade    = $self->blade();

    my $start_args = "";
    my @sargs = ('NUM_RETRIES=3');

    # If this is a "private" vsim (smokes, ctl) we cannot rehost

    if ( !$blade ) {
        # In HA mode you can't initiate a rehost of the pair since the other
        # vsim is running and it would screw him up.
        if ( $self->ha() ) {
            push( @sargs, 'NO_RETRY_IF=please rehost the vsim' );
        } else {
            my $rehost_id = $self->id();
            push( @sargs, 'RETRY_IF=please rehost the vsim' );
            push( @sargs, "BEFORE_RETRY=HOME=".$self->uhome().";USER=".$self->user().
                          ";$vsim_cmd rehost -restype vsimesx -domain $simdomain ".$rehost_id );
        }
    } else {
        $start_args = " -smoke";
    }

    my $expected = "Power on";

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd start ".$self->id().$start_args,
                    #"EXPECTED=$expected",
                    'EXPECTED_RC=0',
                    @sargs,
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-bg',
                    '-onexit', sub { },
                    '-timeout', $opt{timeout},
                    '-runid', $self->get_seq()."_start_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: stop
#   
#  Invoke the "stop" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->stop() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub stop {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 900,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    # I don't have all of the specific error codes from vsim to 
    # reliably know what error code "is not powered on" corresponds to
    # so I am sticking with the strings for now
    my $expected = "stopped|is not powered on|is not bound";

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd stop ".$self->id(),
                    "EXPECTED=$expected",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-bg',
                    '-onexit', sub { },
                    '-timeout', $opt{timeout},
                    '-runid', $self->get_seq()."_stop_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: rm
#   
#  Invoke the "rm" command on a vsim to remove it.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->rm() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   force       => Force removal even if cores are present (default 1)
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub rm {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( force   => 1,
                timeout => 900,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    my $force = $opt{force} ? " -f" : "";

    # I don't have all of the specific error codes from vsim to 
    # reliably know what error code "does not exist" corresponds to
    # so I am sticking with the strings for now
    my $expected = "simulator removed|not allowed on a VSIM which is part of an hagroup";
    if ( $force ) {
        $expected .= "\|does not exist";
    }

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd rm ".$force." ".$self->id(),
                    "EXPECTED=$expected",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    'NO_RETRY_IF=not allowed on .* which is part .* hagroup',
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_rm_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}


# -----------------------------------------------
# Function: snapdelete
#   
#  Invoke the "snapdelete" command on a vsim to delete
#  a vsim snapshot.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->snapdelete( snapname => "post_deux_init" ) );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   snapname    => Name of snapshot (default VSIM_SNAPSHOT)
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub snapdelete {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( snapname        => "VSIM_SNAPSHOT",
                ignore_missing  => 1,
                timeout         => 900,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();
    my $snapshot = $opt{snapname};

    # I don't have all of the specific error codes from vsim to 
    # reliably know what error code "doesn't exist" corresponds to
    # so I am sticking with the strings for now
    my $expected = "successful";
    if ( $opt{ignore_missing} ) {
        $expected .= "|doesn't exist";
    }

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd snapdelete ".$self->id()." $snapshot",
                    "EXPECTED=$expected",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_snapdelete_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: snap
#   
#  Invoke the "snap" command on a vsim to create
#  a vsim snapshot.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->snap( snapname => "post_deux_init" ) );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   snapname    => Name of snapshot (default VSIM_SNAPSHOT)
#   timeout     => Number of seconds until command times out (default 3600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub snap {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( snapname => "VSIM_SNAPSHOT",
                timeout => 3600,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();
    my $snapshot = $opt{snapname};

    $self->spoof_user_env();

    # success message can be "successfully taken" or "successfully created"
    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd snap ".$self->id()." $snapshot",
                    #"EXPECTED=successful",
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_snap_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: snapback
#   
#  Invoke the "snapback" command on a vsim to restore
#  a previously saved vsim snapshot.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->snapback( snapname => "post_deux_init" ) );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   snapname    => Name of snapshot (default VSIM_SNAPSHOT)
#   timeout     => Number of seconds until command times out (default 3600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub snapback {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( snapname => "VSIM_SNAPSHOT",
                timeout => 3600,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();
    my $snapshot = $opt{snapname};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd snapback ".$self->id()." $snapshot",
                    #"EXPECTED=successful",
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    "RESUME_IF_SUSPENDED=1",
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_snapback_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: reserve
#   
#  Invoke the "reserve" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->reserve() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#        
#   domain      => An attempt is made to guess the correct simdomain
#                  based on the "domainname" of the client running
#                  this command.  Will be something like:
#                  "sim.rtp.netapp.com"
#                  This parameter can also be defaulted by specifying the
#                  VSIM_DEFAULT_SERVER_DOMAIN parameter (or ENV variable)
#                  
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub reserve {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( domain => $simdomain,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $simdomain = $opt{domain};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd reserve -restype vsimesx -domain $simdomain ".$self->id(),
                    #"EXPECTED=successful",
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_reserve_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: unreserve
#   
#  Invoke the "unreserve" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->unreserve() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#        
#   domain      => An attempt is made to guess the correct simdomain
#                  based on the "domainname" of the client running
#                  this command.  Will be something like:
#                  "sim.rtp.netapp.com"
#                  This parameter can also be defaulted by specifying the
#                  VSIM_DEFAULT_SERVER_DOMAIN parameter (or ENV variable)
#                  
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub unreserve {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( domain => $simdomain,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $simdomain = $opt{domain};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd unreserve -restype vsimesx -domain $simdomain ".$self->id(),
                    #"EXPECTED=successful",
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_reserve_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: rehost
#   
#  Invoke the "rehost" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->rehost() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#        
#   domain      => An attempt is made to guess the correct simdomain
#                  based on the "domainname" of the client running
#                  this command.  Will be something like:
#                  "sim.rtp.netapp.com"
#                  This parameter can also be defaulted by specifying the
#                  VSIM_DEFAULT_SERVER_DOMAIN parameter (or ENV variable)
#
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub rehost {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( domain => $simdomain,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $blade     = $self->blade();
    my $simdomain = $opt{domain};
    my $st;

    if ( !$blade ) {
        $self->spoof_user_env();
    
        $st = Tharn::subtest( $exec_cmd,
                        "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                                      ";$vsim_cmd rehost -restype vsimesx -domain $simdomain ".$self->id(),
                        'EXPECTED_RC=0',
                        "VSIM_INDEX=".$self->id(),
                        "VSIM_NODE=".$self->node(),
                        '-timeout', $opt{timeout},
                        '-bg',
                        '-onexit', sub { },
                        '-runid', $self->get_seq()."_rehost_".$self->node(),
                        %{$self->user_hash()} );
    
        $self->restore_user_env();
    }

    return $st;
}

# -----------------------------------------------
# Function: extend or renew
#   
#  Invoke the "renew" command on a vsim.
#  Note: the extend() and renew() methods are
#  synonyms for one another.
#
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->extend( days => 3 ) );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#        
#   days        => 1-7 (default 7)
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub extend {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }
    return $self->renew( @_ );
}

sub renew {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( days => 7,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $blade     = $self->blade();
    my $days      = $opt{days};
    my $st;

    # I don't have all of the specific error codes from vsim to 
    # reliably know what error code "End time must be less" corresponds to
    # so I am sticking with the strings for now
    if ( !$blade ) {
        $self->spoof_user_env();

        $st = Tharn::subtest( $exec_cmd,
                        "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd renew -res $days ".$self->id(),
                        "EXPECTED=Successful|End time must be less|aximum",
                        "VSIM_INDEX=".$self->id(),
                        "VSIM_NODE=".$self->node(),
                        '-timeout', $opt{timeout},
                        '-bg',
                        '-onexit', sub { },
                        '-runid', $self->get_seq()."_extend_".$self->node(),
                        %{$self->user_hash()} );
    
        $self->restore_user_env();
    }

    return $st;
}

# -----------------------------------------------
# Function: bind
#   
#  Invoke the "bind" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->bind() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#        
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub bind {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    # Vishal Kulkarni wants to be able to pass in extra bind options to allow
    # his own private ESX farm to be used.  This is controlled by
    # VSIM_EXTRA_BIND_ARGS in sys_init and vsim/bind_vsims.thpl
    my %opt = ( server_instance => undef,
                bind_args       => undef,
                timeout         => 900,
                @_ );

    my $server_instance = $opt{server_instance};
    my $node            = $self->node();
    my $id              = $self->id();
    my $blade           = $self->blade();
    my $inst            = $self->instance();
    my $extra_args      = $opt{bind_args};

    my $bind_args = "bind -esx";

    if ( $blade ) {
        $bind_args .= " -smoke ".$self->id()." ".$blade." ".$inst;
    } else {
        $bind_args .= " ".$self->id()." ".$server_instance;
    }

    if ( $extra_args ) {
        $bind_args .= " ".$extra_args;

        # Need to discuss with Vishal if there is a better way to accomplish this.
        # If not, should consider if I need to provide any other substitutions
        $bind_args =~ s/<node>/$node/g;
    }

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                    ";$vsim_cmd ".$bind_args,
                    #"EXPECTED=successful",
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_bind_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: unbind
#   
#  Invoke the "unbind" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->unbind() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#        
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub unbind {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( server_instance => undef,
                timeout         => 900,
                @_ );

    my $server_instance = $opt{server_instance};
    my $node            = $self->node();
    my $id              = $self->id();
    my $blade           = $self->blade();
    my $inst            = $self->instance();

    my $unbind_args = "unbind -esx";

    if ( $blade ) {
        #$unbind_args .= " ".$self->id()." ".$blade." ".$inst;
        $unbind_args .= " ".$self->id();
    } else {
        $unbind_args .= " ".$self->id()." ".$server_instance;
    }

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                    ";$vsim_cmd ".$unbind_args,
                    "EXPECTED=successful|is not bound",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_unbind_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: halt
#   
#  Use NACL::Transit to transition a vsim to the LOADER
#  (FIRMWARE) prompt.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->halt() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 900)
#   maxidle     => Maximum number of seconds that no response from the filer
#                  will be tolerated (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub halt {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 900,
                maxidle => 900,
                @_ );

    # force $self->{tran} to be created via the default_ctor method
    my $tran      = $self->tran();
    my $node      = $self->node();
    my $naclnode  = $self->naclnode();

    my $mode = $naclnode->mode();

    my $st = Tharn::subtest( sub {  my $initial = $tran->get_state(timeout => $opt{timeout}, maxidle => $opt{maxidle});
                                    $tran->change_state( to => 'FIRMWARE', 
                                                         timeout => $opt{timeout}, maxidle => $opt{maxidle}); },
                 "-runid", get_seq()."_halt_".$mode."_".$node,
                 '-timeout', $opt{timeout},
                 '-onexit', sub { },
                 "-bg");

    return $st;
}


# -----------------------------------------------
# Function: boot
#   
#  Use NACL::Transit to transition a vsim to the CLI
#  prompt.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->boot() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 900)
#   maxidle     => Maximum number of seconds that no response from the filer
#                  will be tolerated (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub boot {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 900,
                maxidle => 900,
                @_ );

    # force $self->{tran} to be created via the default_ctor method
    my $tran      = $self->tran();
    my $node      = $self->node();
    my $naclnode  = $self->naclnode();

    my $st = Tharn::subtest( sub { $tran->change_state( to => 'CLI', timeout => $opt{timeout}, maxidle => $opt{maxidle} ); },
                 "-runid", get_seq()."_boot_$node",
                 '-timeout', $opt{timeout},
                 '-bg',
                 '-onexit', sub { } );

    return $st;
}

# -----------------------------------------------
# Function: start_halt
#   
#  Start a vsim, leaving it at the LOADER
#  prompt.  Very similar to gx_boot's shutdown_cold
#  task.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->start_halt() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 900)
#   maxidle     => Maximum number of seconds that no response from the filer
#                  will be tolerated (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub start_halt {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my @subtests;
    my %opt = ( timeout      => 300,
                prehalt_wait => 0,
                halt_timeout => 2,
                @_ );

    my $node         = $self->node();
    my $id           = $self->id();
    my $blade        = $self->blade();
    my $hostrec      = $self->hostrec();
    my $prehalt_wait = $opt{prehalt_wait};
    my $timeout      = $opt{timeout};
    my $halt_timeout = $opt{halt_timeout};

    my $st = $self->start();

    push( @subtests, $st );

    $self->spoof_user_env();

    $st = Tharn::subtest( sub {
                                  if ( $prehalt_wait ) {
                                      Tharn::sleep( $prehalt_wait );
                                  }

                                  my $start    = time();
                                  my $prompts  = ['(SIM|V)LOADER'];
                                  my $finished = 0;
                                  my $conn     = Connectrec->new( host            => $hostrec,
                                                                  template        => 'console',
                                                                  connect         => 0 );

                                  # connect with verify takes a few extra seconds that
                                  # could possibly cause boot countdown to elapse
                                  $conn->connect( verify => 0 );

                                  while ( !$finished && (time() - $start < $timeout) ) {
                                      eval {
                                          $conn->execute( command => "W",
                                                                     match_table => [
                                                                         'to boot immediately' => 'X',
                                                                         'or any other key for command prompt' => 'Y',
                                                                         'Booting in' => 'Z',
                                                                     ],
                                                                 timeout => $halt_timeout,
                                                                 prompts => $prompts );
                                      };
                                      my $at_loader_rc = $@;

                                      # I think this should be in the Connectrec::disconnect()
                                      # method, but Mike Edmonds disagrees
                                      # The control-] is to cause pm_cons to exit
                                      eval {
                                          $conn->execute( command => "\c]",
                                                          match_table => [
                                                                         'telnet>' => 'quit',
                                                                     ],
                                                             timeout => 2,
                                                             prompts => [@{$prompts}, 'Connection closed'] );
                                      };

                                      # If this fails due to timeout, it isn't a big deal
                                      if ( $@ && ( $@ != /timeout/i ) ) {
                                          $Log->comment("$me: Ctrl-] failed with $@");
                                      }

                                      # Disconnect and delete Connectrec object
                                      $conn->disconnect();
                                      $conn->destroy();

                                      if ( !$at_loader_rc ) {
                                          $finished = 1;
                                      } else {
                                          $conn     = Connectrec->new( host            => $hostrec,
                                                                       template        => 'console',
                                                                       connect         => 0 );

                                          # connect with verify takes a few extra seconds that
                                          # could possibly cause boot countdown to elapse
                                          $conn->connect( verify => 0 );
                                      }
                                  }

                                  if ( !$finished ) {
                                      die "$me: Could not halt vsim ".$self->node();
                                  }
                    },
                    '-timeout', $timeout,
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_halt_".$self->node() );

    push( @subtests, $st );

    $self->restore_user_env();

    if ( wantarray() ) {
        return @subtests;
    } else {
        return \@subtests;
    }
}

# -----------------------------------------------
# Function: unstick
#   
#  Unstick, a "stuck" vsim.
#  Occasionally a vsim will become "stuck" when multiple
#  vsims attempt to write to the same pipe.  For a running
#  vsim this can be serious and should be looked into, however
#  a stuck vsim will even resist efforts to remove it.  This
#  routine will unstick a stuck vsim.
#
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   my $st = $vsim->unstick();
#   Subtest::wait_finish( subtests => [$st] );
#   @results = $st->results();
#   my $misc = $results[0]->misc();
#   my $out  = $misc->{VSIM_CMD_OUTPUT};
#
# Note: This routine is called automatically when another
#       command finds that the vsim is stuck, so it GENERALLY
#       does not need to be called explicitly.
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub unstick {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 900,
                @_ );

    $self->spoof_user_env();

    my $st = Tharn::subtest( $unstick_cmd,
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_unstick_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: set_ntp
#   
#  Enable or disable NTP (Network Time Protocol)
#  Used in the clock syncing process.
#
# Usage:
#
#   $vsim->set_ntp( enable => 0 );
#
# Parameters: 
#       
#   enable     => [0|1] (default 1)
#
# Return: Nothing
# -----------------------------------------------
sub set_ntp {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( enable => 1, 
                @_ );

    my $flag    = "true";
    my $enable  = "Enabling";
    my $flag7   = "on";

    if ( !$opt{enable} ) {
        $flag   = "false";
        $flag7  = "off";
        $enable = "Disabling";
    }

    my $retries = 3;

RETRY:
 
    $Log->comment($enable." NTP service on all nodes");

    if ( $self->naclnode()->mode() eq '7Mode' ) {
        eval {
            NACL::C::Options->option(command_interface => $self->naclnode(),
                                    'option-name'      => "timed.enable" ,
                                    'option-value'     => $flag7 );
        };
    } else {
        eval {
            NACL::C::SystemServicesNtpConfig->modify( command_interface => $self->naclnode(), enabled => $flag );
        };
    }

    if ( $@ && ($retries > 0) ) {
        $Log->warn("$me: Failed with $@\nRetrying...");
        $retries--;
        Tharn::snooze(5);
        goto RETRY;
    } elsif ( $@ ) {
        die "$me: $@";
    }
}

# -----------------------------------------------
# Function: sync_clocks
#   
#  Synchronize the clocks of all cluster nodes for
#  a vsim with those of the client running this routine.
#  Useful after snapback of a previous vsim
#  snapshot.
#
# Usage:
#
#   $vsim->sync_clocks();
#
# Parameters: None
#       
# Return: Nothing
# -----------------------------------------------
sub sync_clocks {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( @_ );

    my ($sec,$min,$hour,$day,$mon,$year,undef) = localtime();
    $year += 1900;
    $mon++;
    my $now = sprintf("\"%02d/%02d/%4d %02d:%02d:%02d\"", $mon, $day, $year, $hour, $min, $sec);

    my $naclnode = $self->naclnode();

    if ( $naclnode->is_cmode() &&
         $naclnode->has_uichange( uichange => 'date-system-node-to-cluster' ) ) {
        use NACL::C::ClusterDate;
        NACL::C::ClusterDate->modify( command_interface => $naclnode, date => $now );
        return;
    }

    sub _disable_ntp {
        $self->set_ntp( enable => 0 );
    }

    sub _enable_ntp {
        $self->set_ntp( enable => 1 );
    }

    my $ntp_must_be_disabled = 0;
    my $retries = 1;
    my @dates = NACL::C::SystemNodeDate->find( command_interface => $self->naclnode() );

AGAIN:

    if ( $ntp_must_be_disabled ) {
        _disable_ntp();
    }

    foreach my $date ( @dates ) {
        $Log->comment("Attempting to set date to $now on node ".$date->node());
        try {
            $date->modify( date => $now );
        } catch NATE::BaseException with {
            my $ex = shift;

            my $rethrow_exception = 1;
            if ( $ex->text() =~ /Cannot modify date on a node when ntp is enabled/i ) {
                if ( !$ntp_must_be_disabled ) {
                    $Log->comment("This build requires that NTP be disabled before the date can be modified, try again");
                    $ntp_must_be_disabled = 1;
                    $rethrow_exception = 0;
                }
            }

            if ( $rethrow_exception ) {
                $ex->throw();
            }
        };

        if ( $ntp_must_be_disabled && $retries ) {
            $retries = 0;
            goto AGAIN;
        }
    }

    if ( $ntp_must_be_disabled ) {
        _enable_ntp();
    }
}

# -----------------------------------------------
# Function: update
#   
#  Invoke the "update" command on a vsim to upgrade
#  to a new Ontap build.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->update( buildroot => "/x/eng/rlse/DOT/mainN" ) );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   buildroot   => Build root directory (ex: /x/eng/rlse/DOT/mainN)
#   platform    => platform string
#                    Ex:
#                       x86_64.debug.gcov.sim
#                       x86_64.gcov.sim
#                       x86_64.sim
#                       x86_64.domain.sim
#                       x86_64.sim.ovf
#                       x86_64.debug.sim.ibm
#                       x86_64.sim.ibm
#                       x86_64.debug.sim.ovf
#                       x86_64.debug.sim
#                       x86_64.debug.domain.sim
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub update {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    # maybe I can default buildroot in each test automation to the
    # appropriate path?
    my %opt = ( buildroot => "/x/eng/rlse/DOT/mainN",
                platform  => "x86_64.debug.sim",
                timeout   => 3600,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();

    $self->spoof_user_env();

    # -ir takes a buildroot-type path
    # Seems like I have to break down the path to what comes before final, i.e.
    # if IMAGE=/x/eng/rlse/DOT/mainN/final/bedrock/export/x86_64.debug.sim.ibm/tarball/image.tgz
    # then what goes to -ir is /x/eng/rlse/DOT/mainN
    #
    #  vsim update [Options] <id> - Update a VSim from a client.
    #
    #  Options:
    #    -bld                   - Build specified client during update.
    #    -d                     - Use debug build.
    #    -ibm                   - Use the ibm vendor build.
    #    -ir       <path>       - Specify the path to the client root.
    #    -iv       <view>       - Specify name of client view for make/update.
    #    -netapp                - Use the NetApp vendor build.
    #    -nightly               - Use the nightly build.
    #    -nod                   - Use the non-dbug build.
    #    -p        <platform_x> - Specify platform (x86/x86_64) for make/update.
    #    -promoted              - Select promoted build in the branch.
    #    -px       <platform_x> - Specify extended platform for make/update.
    #    -release  <branch>     - branch in conjunction with -nightly or -promoted.

    # success message is "... Done."
    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                    ";$vsim_cmd update -ir ".$opt{buildroot}.
                    " -px ".$opt{platform}." ".$self->id(),
                    "ALTERNATE_COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                    ";$vsim_cmd update -f -ir ".$opt{buildroot}.
                    " -px ".$opt{platform}." ".$self->id(),
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_update_".$self->node(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: verify_update
#   
#  Verify that after a vsim cluster is booted up, that
#  all of the Ontap upgrade processing has completed.
#
# Usage:
#
#   $vsim->verify_update();
#
# Parameters: 
#   retries   => Number of retries (default 120)
#   sleep     => Seconds to sleep (default 5)
#       
# Return: Nothing
# -----------------------------------------------
sub verify_update {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( retries => 120,
                sleep => 5,
                @_ 
              );

    my $node     = $self->naclnode();
    my $nodename = $node->node();
    my $retries  = $opt{retries};
    my $sleep    = $opt{sleep};

    while ( $retries > 0 ) {
           $retries--;
        my $again   = 0;
        my @up_rev  = NACL::CS::SystemNodeUpgradeRevert->fetch( command_interface => $node,
                                                                filter => { node => $nodename },
                                                                allow_empty => 1 );
        foreach my $entry ( @up_rev ) {
            if ( $entry->node_status() ne "complete" ) {
                $again = 1;
            }
        }

        if ( $again ) {
            if ( $retries > 0 ) {
                $Log->comment("$me: Node $nodename did not finish upgrading, retry in $sleep seconds, retries left $retries");
                Tharn::sleep( $sleep );
            } else {
                die "$me: Node $nodename did not finish upgrading";
            }
        } else {
            $retries = 0;
        }
    }
}

# -----------------------------------------------
# Function: wait_until_finished
#   
#  Convenience routine that will wait until all
#  subtested jobs are completed and 
#
# Usage:
#
#   SysInit::Vsim::wait_unti_finished( subtests = \@subs );
#
# Parameters: 
#
#    subtests             => ARRAYREF of NATE Subtest objects (required)
#    worst_allowable      => NATE::Result Worst allowable result (default NATE::Result::NA)
#    errors               => ARRAYREF into which error strings will be stored (optional)
#    empty_errors_before  => SCALAR Should errors arrayref be emptied before subtests are checked? (default 1)
#    empty_subtests_after => SCALAR Should subtests arrayref be emptied when wait_until_finished() returns? (default 1)
#       
# Return: 
# -----------------------------------------------
sub wait_until_finished {
    my $me=(caller(0))[3];
    my %opt = validate_with(
        params => \@_,
        spec   => {
                             # Any result with a severity greater than this will be
                             # considered a failure (i.e. Inspect and above)
            worst_allowable      => { isa => 'NATE::Result', default  => 'NATE::Result::NA' },
            subtests             => { type => ARRAYREF },
            errors               => { type => ARRAYREF, default => [] },
            die_if_errors        => { type => SCALAR,   default => 1 },
            empty_errors_before  => { type => SCALAR,   default => 1 },
            empty_subtests_after => { type => SCALAR,   default => 1 },
            kill_remaining       => { type => SCALAR,   default => 0 },
        },
    );

    my $worst_allowable = $opt{worst_allowable};

    if ( $opt{empty_errors_before} ) {
        @{$opt{errors}} = ();
    }

    my $done = 0;
    my $error = 0;
    my @living = @{$opt{subtests}};

    while ( !$done ) {
        my @tmp;
        foreach my $s (@living) {
            if ( !$s->is_alive() ) {
                my @subResults = $s->own_results();
                my $worst = NATE::Result::worst(@subResults);
        
                if ( $worst &&
                    ($worst->isa( 'NATE::Result' ) ) &&
                    ($worst->severity() > $worst_allowable->severity()) ) {

                    $error++;                    
        
                    # Add result strings into errors array
                    push( @{$opt{errors}}, "Runid: ".$worst->runid()." Messages: ".join(',', $worst->messages()) );
                }
            } else {
                push( @tmp, $s );
            }
        }

        if ( scalar(@tmp) ) {
            @living = @tmp;
            if ( $error && $opt{kill_remaining} ) {
                foreach my $s (@living) {
                    # Possible that subtest finished by now
                    if ( $s->is_alive() ) {
                        $s->kill();
                    }
                    # wait until subtest dies
                    while ( $s->is_alive() ) {
                        Tharn::sleep(1);
                    }
                    my @subResults = $s->own_results();
                    my $worst = NATE::Result::worst(@subResults);
        
                    if ( $worst &&
                        ($worst->isa( 'NATE::Result' ) ) &&
                        ($worst->severity() > $worst_allowable->severity()) ) {
    
                        # Add result strings into errors array, probably 'killed by parent'
                        # unless it ended before we got a chance to kill it
                        push( @{$opt{errors}}, "Runid: ".$worst->runid()." Messages: ".join(',', $worst->messages()) );
                    }
                }
                $done = 1;
            } else {
                Tharn::sleep(1);
            }
        } else {
            $done = 1;
        }
    }

    if ( $opt{empty_subtests_after} ) {
        @{$opt{subtests}} = ();
    }

    if ( scalar(@{$opt{errors}}) ) {
        my $errMsg = join('\n', @{$opt{errors}});
        if ( $opt{die_if_errors} ) {
            die "$me failed with $errMsg";
        } else {
            $Log->warn("$me failed with $errMsg");
        }
    }
}

# Returns the name of the hagroup that this vsim is in or undef if not in one
# Expect output to look like this:
# vsim hashow
# HaGroups                           VSIMs                              Mac used
# haconfig_new_ha                    ssurya_vsim_new1                   00:50:56:10:18:af
#                                    ssurya_vsim_new2                   00:50:56:10:18:ae
#
# Problem comes with using profiles like this (cha3 corresponds to radish-vsim3):
# How to definitively map cha3 to 3?
#
# vsim.beta hashow
# HaGroups                           VSIMs                              Mac used
# haconfig_radish-vsim1radish-vsim2  radish_vsim_1                      00:50:56:10:07:a8
#                                    radish_vsim_2                      00:50:56:10:20:ff
#
# haconfig_vsimchagrp34              radish_vsim_cha3                   00:50:56:10:07:a9
#                                    radish_vsim_cha4                   00:50:56:10:21:00

sub hagroup_from_hashow {
    my $self  = shift;
    my $me=(caller(0))[3];

    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my $user         = $self->user();
    my $uhome        = $self->uhome();
    my $hagroup_name = undef;

    $Log->comment("$me: USER=$user;HOME=$uhome;$vsim_cmd hashow");
    my $result       = `USER=$user;HOME=$uhome;$vsim_cmd hashow`;
    chomp $result;

    if ( $result ) {
        print "$me: raw hashow output = ".$result."\n";
        if ( $result !~ /Unknown command/ ) {
            my $expected_hagroup_name = "haconfig_".$self->SFOPAIR();
            my @lines = split("\n", $result);
            # shift off top line, just header info
            shift @lines;
            my $result = join(",", @lines);
            $result =~ s/\s+/,/g;
            $result =~ s/,,/,/g;
            $result =~ s/,haconfig/;haconfig/g;

            # BURT 643158 can cause long hagroup names to print out in
            # hashow commands with no space between the group name and the vsim name.
            # so we might get haconfig_sti2-vsim-ucs010gsti2-vsim-ucs010hsti2_vsim_ucs010g
            # instead of      haconfig_sti2-vsim-ucs010gsti2-vsim-ucs010h sti2_vsim_ucs010g
            # So if we find our expected hagroup name and it is not followed by a comma add one
            $result =~ s/$expected_hagroup_name([^,])/$expected_hagroup_name,$1/;
            $result =~ s/,,/,/g;

            @lines = split(";", $result);
            
            my $vsim_index = $self->id();
HA_OUTER:   foreach my $line (@lines) {
                if ( $line =~ /\*\* Warning \*\*/i ) {
                    $Log->warn("$me: hashow warning: $line");
                    next HA_OUTER;
                }
                my @grp = split(",", $line);
            
                # I am making the assumption that the vsim id
                # to vsim long name mapping is as defined below
                my $vsim_longname   = $user."_vsim_".$vsim_index;
                my $possible_suffix = $user."_vsim_";
                my $grpname = shift @grp;
                my $num_entries = scalar(@grp);
                foreach my $grp (@grp) {
                    if ( $grp =~ /$vsim_longname$/ ) {
                        $hagroup_name = $grpname;
                        last HA_OUTER;
                    }
                }
            }
        }
    }

    return $hagroup_name;
}

# Returns the name of the mcgroup that this vsim is in or undef if not in one
# Expect output to look like this:
# vsim mcshow
#   McGroups                  HaGroups                  vsims                       Mac Used 
# mcconfig_mcc1              haconfig_ha1         ponugoti_vsim_vsima1            00:50:56:10:05:41
#                                                 ponugoti_vsim_vsima2            00:50:56:10:05:42
#                            haconfig_ha2         ponugoti_vsim_vsimb1            00:50:56:10:05:43
#                                                 ponugoti_vsim_vsimb2            00:50:56:10:05:44
#
# If the vsim name created via profile is something other than the name we are looking for,
# this could fail

sub mcgroup_from_mcshow {
    my $self  = shift;
    my $me=(caller(0))[3];

    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    if ( !$supports{mcgroups} ) {
        return (undef, undef);
    }

    my $user   = $self->user();
    my $uhome  = $self->uhome();
    my ($mcgroup_name, $hagroup_name) = (undef, undef);

    $Log->comment("$me: USER=$user;HOME=$uhome;$vsim_cmd mcshow");
    my $result       = `USER=$user;HOME=$uhome;$vsim_cmd mcshow`;
    chomp $result;

    if ( $result ) {
        print "$me: raw mcshow output = ".$result."\n";
        if ( $result !~ /Unknown command/ ) {
            my @lines = split("\n", $result);
            # shift off top line, just header info
            shift @lines;
            my $result = join(",", @lines);
            $result =~ s/\s+/,/g;
            $result =~ s/,,/,/g;
            $result =~ s/,mcconfig/;mcconfig/g;

            $result =~ s/,,/,/g;

            @lines = split(";", $result);

            my $vsim_index = $self->id();
MC_OUTER:   foreach my $line (@lines) {

                if ( $line =~ /\*\* Warning \*\*/i ) {
                    next MC_OUTER;
                }
                my @grp = split(",", $line);

                # I am making the assumption that the vsim id
                # to vsim long name mapping is as defined below
                my $vsim_longname   = $user."_vsim_".$vsim_index;
                my $possible_suffix = $user."_vsim_";
                my $grpname = shift @grp;
                $hagroup_name = undef;
                my $num_entries = scalar(@grp);
                foreach my $grp (@grp) {
                    if ( $grp =~ /^haconfig_/ ) {
                        $hagroup_name = $grp;
                    }
                    if ( $grp =~ /$vsim_longname$/ ) {
                        $mcgroup_name = $grpname;
                        last MC_OUTER;
                    }
                }
            }
        }
    }

    return ($mcgroup_name, $hagroup_name);
}

# Returns the name of the mcgroup and hagroup that groups this vsim by examining the
# dparam output
sub mcgroup_from_dparam {
    my $self  = shift;
    my $me=(caller(0))[3];

    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    if ( !$supports{mcgroups} ) {
        return (undef, undef);
    }

    my $user       = $self->user();
    my $uhome      = $self->uhome();
    my $vsim_id    = $self->id();
    my ($mcgroup_name, $hagroup_name) = (undef, undef);

    $Log->comment("$me: USER=$user;HOME=$uhome;$vsim_cmd dparam $vsim_id mc_group");
    my $result       = `USER=$user;HOME=$uhome;$vsim_cmd dparam $vsim_id mc_group`;
    chomp $result;

    if ( $result ) {
        if ( $result !~ /directory does not exist/i ) {
            $Log->comment("$me: raw dparam mc_group output = |".$result."|");
            $mcgroup_name = $result; 
            if ( $mcgroup_name !~ /^mcgroup_/ ) {
                $mcgroup_name = 'mcconfig_'.$mcgroup_name;
            }
            $Log->comment("$me: USER=$user;HOME=$uhome;$vsim_cmd dparam $vsim_id ha_group");
            $result          = `USER=$user;HOME=$uhome;$vsim_cmd dparam $vsim_id ha_group`;
            chomp $result;
            if ( $result ) {
                if ( $result !~ /directory does not exist/i ) {
                    $Log->comment("$me: raw dparam ha_group output = |".$result."|");
                    $hagroup_name = $result; 
                    if ( $hagroup_name !~ /^hagroup_/ ) {
                        $hagroup_name = 'haconfig_'.$hagroup_name;
                    }
                }
            }
        }
    }

    return ($mcgroup_name, $hagroup_name);
}

# Returns the name of the hagroup that groups this vsim by examining the
# dparam output
sub hagroup_from_dparam {
    my $self  = shift;
    my $me=(caller(0))[3];

    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my $user         = $self->user();
    my $uhome        = $self->uhome();
    my $vsim_id      = $self->id();
    my $hagroup_name = undef;

    $Log->comment("$me: USER=$user;HOME=$uhome;$vsim_cmd dparam $vsim_id ha_group");
    my $result       = `USER=$user;HOME=$uhome;$vsim_cmd dparam $vsim_id ha_group`;
    chomp $result;

    if ( $result ) {
        if ( $result !~ /directory does not exist/i ) {
            $Log->comment("$me: raw dparam ha_group output = |".$result."|");
            $hagroup_name = $result; 
            if ( $hagroup_name !~ /^hagroup_/ ) {
                $hagroup_name = 'haconfig_'.$hagroup_name;
            }
        }
    }

    return $hagroup_name;
}

# -----------------------------------------------
# Function: sfopair_names
#   
#  Convenience routine that will return the unique list
#  of SFOPAIR names given a list of vsim objects
#
# Usage:
#
#   my @sfopair_names = SysInit::Vsim::sfopair_names( @vsims );
#
# Parameters: array of SysInit::Vsim objects
#
# Return: array of sfopair names for those vsims
# -----------------------------------------------
sub sfopair_names {
    my $me=(caller(0))[3];
    my @vsims = @_;

    my @sfopair_names;

    my %sfopairs;
    foreach my $vsim (@vsims) {
        my $SFOPAIR = $vsim->SFOPAIR();
        $sfopairs{$SFOPAIR} = 1;
    }

    @sfopair_names = (sort keys %sfopairs);

    return @sfopair_names;
}

# -----------------------------------------------
# Function: snapshots_supported
#   
#  Convenience routine that will tell if snapshots are
#  supported or not.  If you call on an instance, you do
#  not need to pass in anything, but if you make a static
#  (class call) you have to pass in the ha parameter.
#
# Usage:
#
#   my $snapshots_supported = $vsim->snapshots_supported();
#   my $snapshots_supported = SysInit::Vsim->snapshots_supported( ha => 1 );
#
# Parameters: 
#
#  ha    => is ha configured?
#
# Return: 1 if snapshots are supported, 0 if not.
# -----------------------------------------------
sub snapshots_supported {
    my $self  = shift;
    my $me=(caller(0))[3];

    my %opt = ( ha          => 0,
                @_ );

    # Currently HA vsims do not support snapshots
    # Eventually when this is supported, we will need to check the 
    # version of the vsim executable too
    my $snapshots_supported = 1;
    if ( ref($self) ) {
        $snapshots_supported = !$self->ha();
    } else {
        $snapshots_supported = !$opt{ha};
    }

    return $snapshots_supported;
}

# -----------------------------------------------
# Function: make_via_profile
#   
#  Convenience routine that will create using a vsim profile
#  This can only be called via a static (class) call because
#  profiles may create multiple vsim objects
#
# Usage:
#
#   SysInit::Vsim->make_via_profile( profile => "mc4n" );
#
# Parameters: 
#
#  Required:
#
#    profile           => vsim make profile name
#    vsims             => reference to an array of SysInit::Vsim objects
#    build_root        => root of the Ontap Build tree for passing to -ir
#
#  Optional:
#
#    bubble_net        => Option required for CTL and smokes vsims (boolean, 0)
#    create_args       => If you want to pass in your own create arguments (string, undef)
#    extra_create_args => If you want to append extra create arguments (string, undef)
#    dynamic_icmacs    => Should intercluster MAC addresses be created dynamically? (boolean, 0)
#    mode              => 10m or 7m (string 10m)
#    platform          => Vsim make platform for -px arg (string "x86_64")
#    noauto            => Do not auto-configure vsims (boolean 1)
#    hagrp             => Comma separated list of hagroup names (string undef)
#    ibm               => Do you want to use the IBM build? (boolean 0)
#    icmacs            => Comma separated list of intercluster MAC addresses (string undef)
#    icmacs_prefix     => Prefix for dynamic_icmacs (string "02:a0:98")
#    ids               => Comma separated list of vsim ids to override what the profile specifies (string undef)
#    mcgrp             => Comma separated list of mcgroup names (string undef)
#    nodelist          => Comma separated list of vsim node names to override what the profile specifies (string undef)
#    num_nics          => number of NICs to define on the vsim
#    setenv            => Comma separated list of settings, variable1=setting1,variable2=setting2 (default "AUTOBOOT=false")
#    smoke             => Option required for CTL and smokes vsims (boolean, 0)
#    timeout           => Timeout for the make command (integer 900 seconds)
#
# -----------------------------------------------
sub make_via_profile {
    my $self  = shift;
    my $me=(caller(0))[3];

    if ( ref($self) ) {
        die "$me Cannot be called as an instance method, instead call SysInit::Vsim->make_via_profile()";
    }

    my %opt = validate_with(
        params => \@_,
        spec   => {
            profile           => { type => SCALAR },
            build_root        => { type => SCALAR },
            vsims             => { type => ARRAYREF },
            mode              => { type => SCALAR, default => "10m", },
            platform          => { type => SCALAR, default => "x86_64", },
            noauto            => { type => SCALAR, default => 1, },
            ibm               => { type => SCALAR, default => 0, },
            mcgrp             => { type => SCALAR, default => undef, },
            hagrp             => { type => SCALAR, default => undef, },
            ids               => { type => SCALAR, default => undef, },
            nodelist          => { type => SCALAR, default => undef, },
            num_nics          => { type => SCALAR, default => undef, },
            icmacs            => { type => SCALAR, default => undef, },
            dynamic_icmacs    => { type => SCALAR, default => undef, },
            icmacs_prefix     => { type => SCALAR, default => "02:a0:98", },
            setenv            => { type => SCALAR, default => 'AUTOBOOT=false', },
            smoke             => { type => SCALAR, default => 0, },
            bubble_net        => { type => SCALAR, default => 0, },
            create_args       => { type => SCALAR, default => "", },
            extra_create_args => { type => SCALAR, default => "", },
            timeout           => { type => SCALAR, default => 1800 },
        },
    );

    my $vsim_owner;
    
    foreach my $vsim (@{$opt{vsims}} ) {
        if ( ref($vsim) ne 'SysInit::Vsim' ) {
            die "$me: objects passed in via the \"vsims\" option must be of type \"SysInit::Vsim\" but was |".ref($vsim)."|";
        }
        if ( !$vsim_owner ) {
            $vsim_owner = $vsim->user();
        }
    }

    if ( !$opt{create_args} ) {
        my ($build,$mode,$flavor);
        if ( !$opt{build_root} ) {
            die "$me If \"create_args\" is not specified, \"build_root\" must be";
        } else {
            $build = $opt{build_root};
        }

        if ( $opt{ibm} ) {
            if ( $opt{platform} !~ /ibm/ ) {
                $opt{platform} = $opt{platform}.'.ibm';
            }
        }

        # Brian McCarthy said that -10m and -pcons should not be necessary and -quickl is obsolete
        $opt{create_args} = "-px ".$opt{platform}." -ir ".$build;
    }

    # If someone just wants to add something to the defaults
    if ( $opt{extra_create_args} ) {
        $opt{create_args} .= " ".$opt{extra_create_args};
    }

    if ( $opt{ibm} ) {
        $opt{create_args} .= " -ibm";
    }

    if ( $opt{noauto} ) {
        $opt{create_args} .= " -noauto";
    }

    if ( $opt{num_nics} ) {
        my $num_nics = $opt{num_nics};
        if ( ($num_nics < 5) || ($num_nics > 8) ) {
            NATE::BaseException->throw("VSIM supports num_nics between 5 and 8, value supplied was $num_nics");
        }
        $opt{create_args} .= " -nics ".$num_nics;
    }

    if ( $opt{platform} =~ /debug/ ) {
        $opt{create_args} .= " -d ";
    } else {
        $opt{create_args} .= " -nod ";
    }

    if ( $opt{bubble_net} ) {
        $opt{create_args} .= " -bubble_net ";
    }

    if ( $opt{smoke} ) {
        $opt{create_args} .= " -smoke ";
    }

    if ( $opt{profile} ) {
        $opt{create_args} .= " -pro ".$opt{profile};
    }

    if ( $opt{ids} ) {
        $opt{create_args} .= " -ids ".$opt{ids};
    }

    if ( $opt{nodelist} ) {
        $opt{create_args} .= " -nodelist ".$opt{nodelist};
    }

    if ( $opt{mcgrp} ) {
        $opt{create_args} .= " -mcgrp ".$opt{mcgrp};
    }

    if ( $opt{hagrp} ) {
        $opt{create_args} .= " -hagrp ".$opt{hagrp};
    }

    if ( $opt{icmacs} ) {
        if ( ref($opt{icmacs}) =~ /array/i ) {
            $opt{create_args} .= ' -icmac '.join(',', @{$opt{icmacs}});
        } else {
            $opt{create_args} .= ' -icmac '.$opt{icmacs};
        }
    } elsif ( $opt{dynamic_icmacs} ) {
        my @icmacs;
        foreach my $vsim ( @{$opt{vsims}} ) {
            my $mgmt_ip=$vsim->naclnode()->hostrec()->{MGMT_IP};

            if ($mgmt_ip =~ /^([0-9]+?)\.([0-9]+?)\.([0-9]+?)\.([0-9]+?)$/) {
                my $mac = $opt{icmacs_prefix}.":".sprintf("%02x:%02x:%02x", $2, $3, $4);
                push(@icmacs, $mac);
            } else {
                die 'Dynamic IC MAC address calculation failed for VSIM HA, mgmt ip $mgmt_ip in nate host file is not proper';
            }
        }
        $opt{create_args} .= ' -icmac '.join(',', @icmacs);

        if ( $SysInit::Vsim::supports{hagroup_noicupdate} ) {
            $opt{create_args} .= ' -noicupdate';
        }
    }

    if ( $vsim_owner ) {
        $opt{create_args} .= " -u ".$vsim_owner;
    }

    if ( $opt{setenv} ) {
        $opt{create_args} .= " -ev ".$opt{setenv};
    }

    my $vsim_owner_uid   = getpwnam($vsim_owner);
    my $home_of_vsim_owner = (getpwuid($vsim_owner_uid))[7];

    my %user_opts = ();

    if ( $vsim_owner_uid != $run_uid ) {
        eval {
            my $STAFProc = `ps -aef | grep STAFProc | grep -v grep | grep '^root' | grep -v " su "`;
            if ( !$STAFProc ) {
                $Log->warn("$me: Vsim owned by |".$self->user()."|, user actually running this method is |".
                           $run_user."|, but STAFProc is not running as root.  ".
                           "Please see http://wikid/w/Coati/SysInit#A_Note_about_Vsims");
            }
        };
        if ( $@ ) {
            $Log->warn("$me: Could not determine how STAFProc is running $@");
            $Log->warn("$me: Vsim owned by |".$self->user()."|, user actually running this method is |".
                       $run_user."|, but STAFProc is not running as root.  ".
                       "Please see http://wikid/w/Coati/SysInit#A_Note_about_Vsims");
        }

        $user_opts{'-username'} = $vsim_owner;
    }

    # set home and user based on the owner of the vsims
    $ENV{USER} = $vsim_owner;
    $ENV{HOME} = $home_of_vsim_owner;

    # For debugging
    #$Log->warn("$me FWR do not submit with this strace command in here");
    #$vsim_cmd = '/usr/bin/strace -c '.$vsim_cmd;

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$ENV{HOME}.";USER=".$ENV{USER}.";$vsim_cmd make ".$opt{create_args},
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$opt{vsims}->[0]->id(),
                    "VSIM_NODE=".$opt{vsims}->[0]->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', SysInit::Vsim->get_seq()."_make_via_profile_".$opt{profile},
                    %user_opts );

    # restore home and user based on who is running this script
    $ENV{HOME} = $saved_env_HOME;
    $ENV{USER} = $saved_env_USER;

    return $st;
}

1;
