# -*- perl -*-
# $Id$

## @summary SysInit::VsimHaGroup
## @author  dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

SysInit::VsimHaGroup - Class which interacts with vsims

=head1 SYNOPSIS

  use SysInit::VsimHaGroup;

=head1 DESCRIPTION

This module defines routines related to vsim creation, deletion and manipulation.

Note: 

No attempt is made here to document all vsim functionality.  For that please
see the following:

http://thebrewery-web.corp.netapp.com/Brewery/brewery/dev/Tools/VSim/index.html

=cut

package SysInit::VsimHaGroup;

use strict;
use NATE::OS;
use Tharn qw($Log);
use Data::Dumper;
use Hostrec qw();
use IPC::Cmd qw(can_run);
use NACL::C::Options;
use NACL::C::SystemNodeDate;
use NACL::C::SystemServicesNtpConfig;
use NACL::STask::Node;
use NACL::Transit;
use Tie::IxHash;
use NACL::GeneralUtils;
use NATE::Exceptions::Argument qw(:try);
use NATE::BaseException qw(:try);
use SysInit::Vsim;

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    array  => [ 'pair' ],
    scalar => [ 
        {
            '-default' => undef,
            '-type'    => "SysInit::Vsim",
            '-read_cb' => sub { 
                                  my $vsim;
                                  my @pair = ($_[0]->pair());
                                  if ( @pair ) {
                                      $vsim = $pair[0];
                                      $_[0]->{first_vsim} = $vsim;
                                  }
                                  $vsim;
                               },
            '-forward' => [qw/ user uhome node naclnode id 
                               cmode get_seq
                               blade instance
                               spoof_user_env restore_user_env
                               saved_env_HOME saved_env_USER
                               SFOPAIR
                               user_hash /],
        },
        'first_vsim' ],
    scalar => [ 'hagroup' ],
];

# -----------------------------------------------
# Function: init
#   
#  routine used by the constructor
#
#  Note: Since there is not an explicit constructor
#        defined in this code (we're using init() and
#        MethodMaker's default constructor) I will 
#        provide the usage for the contructor here.
#
# Usage:
#   my $vsim = SysInit::VsimHaGroup->new( pair => [ $vsim1, $vsim2 ] )
#
# Parameters:
#       pair      - pair of vsims (required)
#       hagroup   - name of hagroup (optional, defaults to SFOPAIR)
#
# Return: blessed reference
# -----------------------------------------------
sub init {
    my $self = shift;
    my $me=(caller(0))[3];

    my %opt = ( pair    => undef,
                hagroup => undef,
                @_ );

    my $pair    = $opt{pair};

    if ( $pair ) {
        my $pair_specified_incorrectly = 0;
        # MethodMaker doesn't really handle array initialization properly.
        # it would store the arrayref as the zeroth element of the array
        if ( ref($pair) eq 'ARRAY' ) {
            if ( scalar(@$pair) == 2 ) {
                $self->{pair} = [];
                # Add to pair will check type and SFOPAIR
                $self->add_to_pair( @$pair );
            } else {
                $pair_specified_incorrectly = 1;
            }
        } else {
            $pair_specified_incorrectly = 1;
        }

        if ( $pair_specified_incorrectly ) {
            NATE::Exceptions::Argument->throw("Option \"pair\" must be an array reference ".
                                              "pointing to an array of exactly 2 SysInit::Vsim elements, pair = |".Dumper($pair)."|");
        }
    }

    $self;
}

# -----------------------------------------------
# Function: add_to_pair
#   
#  routine used to add a vsim to the hagroup object
#
#  Basically does checking to assure that the vsim objects
#  added are both in the same SFOPAIR.
#
# Usage:
#   $hagroup->add_to_pair( $vsim1, $vsim2 );
#
# Parameters:
#       One or two SysInit::Vsim objects
#
# Return: Nothing
# -----------------------------------------------
sub add_to_pair {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my @vsims = @_;
    if ( scalar(@vsims) > 2 ) {
        NATE::Exceptions::Argument->throw("$me: Can only add 2 elements to ".$self->hagroup()." \"pair\" but ".scalar(@vsims)." were specified ");
    }
    
    foreach my $vsim (@vsims) {
        if ( ref($vsim) ne 'SysInit::Vsim' ) {
            NATE::Exceptions::Argument->throw("$me: Can only add objects to ".$self->hagroup()." \"pair\" of type \"SysInit::Vsim\" but added ".Dumper(@vsims));
        }
    }

    my @pair = $self->pair();
    if ( (scalar(@pair) + scalar(@vsims)) > 2 ) {
        NATE::Exceptions::Argument->throw("$me: Can only add 2 objects to ".$self->hagroup()." but \"pair\" already contains ".
                                          scalar(@pair)." objects and you are attempting to add ".scalar(@vsims)." more.");
    }
    push(@pair, @vsims);
    my %sfopairs;
    my @sfopairs;
    foreach my $vsim (@pair) {
        if ( $vsim->SFOPAIR() ) {
            $sfopairs{$vsim->SFOPAIR()} = 1;
            push(@sfopairs, $vsim->node().",SFOPAIR=".$vsim->SFOPAIR());
        } else {
            $sfopairs{$vsim->node().",SFOPAIR=UNDEFINED"} = 1;
            push(@sfopairs, $vsim->node().",SFOPAIR=UNDEFINED");
        }
    }

    # Must have the same SFOPAIR definition
    if ( scalar(keys %sfopairs) != 1 ) {
            NATE::BaseException->throw("SysInit::Vsim objects stored in \"pair\" must have the same SFOPAIR hostrec ".
                                       "definition: ".Dumper(\@sfopairs));
    } else {
        # If an hagroup name was not passed into the constructor, use the SFOPAIR name
        if ( !$self->hagroup() ) {
            my $sfopair = (keys %sfopairs)[0];
            $self->hagroup($sfopair);
        }
    }

    $self->first_vsim();
    # Normally ha gets set in make, but want to make sure that
    # ha flag is set if we did not call make on the vsims
    foreach my $vsim (@pair) {
        $vsim->ha(1);
    }
    $self->pair( @pair );
}

# -----------------------------------------------
# Function: show
#   
#  Invoke the "hashow" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   my $st = $vsim->show();
#   SysInit::Vsim::wait_until_finished( subtests => [$st] );
#   @results = $st->results();
#   my $misc = $results[0]->misc();
#   my $out  = $misc->{VSIM_CMD_OUTPUT};
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub show {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 600,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    $self->spoof_user_env();

    my $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$SysInit::Vsim::vsim_cmd hashow ".$self->hagroup(),
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-bg',
                    '-onexit', sub { },
                    '-timeout', $opt{timeout},
                    '-runid', $self->get_seq()."_hashow_".$self->hagroup(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: rm
#   
#  Invoke the "rm" command on an hagroup to remove it.
#
# Usage:
#
#   push( @subtests, $hagroup->rm() );
#   SysInit::Vsim::wait_until_finished( subtests => \@subtest );
#
# Parameters: 
#       
#   force       => Force removal even if cores are present (default 1)
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub rm {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( force   => 1,
                timeout => 900,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    my $force = $opt{force} ? " -f" : "";

    # I don't have all of the specific error codes from vsim to 
    # reliably know what error code "does not exist" corresponds to
    # so I am sticking with the strings for now
    my $expected = "simulator removed";
    if ( $force ) {
        $expected .= "\|does not exist";
    }

    $self->spoof_user_env();

    my $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$SysInit::Vsim::vsim_cmd rm ".$force." ".$self->hagroup(),
                    "EXPECTED=$expected",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_rm_hagroup_".$self->hagroup(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: reserve
#   
#  Invoke the "reserve" command on an hagroup
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   push( @subtests, $hagroup->reserve() );
#   SysInit::Vsim::wait_until_finished( subtests => \@subtest );
#
# Parameters: 
#        
#   simdomain   => An attempt is made to guess the correct simdomain
#                  based on the "domainname" of the client running
#                  this command.  Will be something like:
#                  "sim.rtp.netapp.com"
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub reserve {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( domain => $SysInit::Vsim::simdomain,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $simdomain = $opt{domain};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                        ";$SysInit::Vsim::vsim_cmd reserve -restype vsimesx -domain $simdomain ".$self->hagroup(),
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_reserve_hagroup_".$self->hagroup(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: rehost
#   
#  Invoke the "rehost" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->rehost() );
#   }
#   SysInit::Vsim::wait_until_finished( subtests => \@subtest );
#
# Parameters: 
#        
#   simdomain   => An attempt is made to guess the correct simdomain
#                  based on the "domainname" of the client running
#                  this command.  Will be something like:
#                  "sim.rtp.netapp.com"
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub rehost {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( domain => $SysInit::Vsim::simdomain,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $blade     = $self->blade();
    my $simdomain = $opt{domain};
    my $st;

    if ( !$blade ) {
        $self->spoof_user_env();
    
        $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                        "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                                      ";$SysInit::Vsim::vsim_cmd rehost -restype vsimesx -domain $simdomain ".$self->hagroup(),
                        'EXPECTED_RC=0',
                        "VSIM_INDEX=".$self->id(),
                        "VSIM_NODE=".$self->node(),
                        '-timeout', $opt{timeout},
                        '-bg',
                        '-onexit', sub { },
                        '-runid', $self->get_seq()."_rehost_hagroup_".$self->hagroup(),
                        %{$self->user_hash()} );
    
        $self->restore_user_env();
    }

    return $st;
}

# -----------------------------------------------
# Function: extend or renew
#   
#  Invoke the "renew" command on an hagroup
#  Note: the extend() and renew() methods are
#  synonyms for one another.
#
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   push( @subtests, $hagroup->extend( days => 3 ) );
#   SysInit::Vsim::wait_until_finished( subtests => \@subtest );
#
# Parameters: 
#        
#   days        => 1-7 (default 7)
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub extend {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }
    return $self->renew( @_ );
}

sub renew {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( days    => 7,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $blade     = $self->blade();
    my $days      = $opt{days};
    my $st;

    # I don't have all of the specific error codes from vsim to 
    # reliably know what error code "End time must be less" corresponds to
    # so I am sticking with the strings for now
    if ( !$blade ) {
        $self->spoof_user_env();

        $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                        "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$SysInit::Vsim::vsim_cmd renew -res $days ".$self->hagroup(),
                        "EXPECTED=Successful|End time must be less|aximum",
                        "VSIM_INDEX=".$self->id(),
                        "VSIM_NODE=".$self->node(),
                        '-timeout', $opt{timeout},
                        '-bg',
                        '-onexit', sub { },
                        '-runid', $self->get_seq()."_extend_hagroup_".$self->hagroup(),
                        %{$self->user_hash()} );
    
        $self->restore_user_env();
    }

    return $st;
}

# -----------------------------------------------
# Function: create
#   
#  Invoke the "hacreate" command for an hagroup
#
# Usage:
#
#   push( @subtests, $ha->create());
#   SysInit::Vsim::wait_until_finished( subtests => \@subtest );
#
# Parameters: 
#      None 
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub create {
    my $self  = shift;
    my $me=(caller(0))[3];
    my %opt = ( server_instance => undef,
                icmacs          => undef,
                dynamic_icmacs  => undef,
                icmacs_prefix   => "02:a0:98",
                timeout         => 900,
                @_ );

    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my $hagroup   = $self->hagroup();
    my $node      = $self->node();
    my $blade     = $self->blade();
    my @pair      = ($self->pair());
    my $id1       = $pair[0]->id();
    my $id2       = $pair[1]->id();
    my $num_nics  = $pair[0]->num_nics();


    my %port_map = ( 5 => 'e0e',
                     6 => 'e0f',
                     7 => 'e0g',
                     8 => 'e0h' );

    my $port = $port_map{$num_nics};

    $self->spoof_user_env();

    my $create_args = "";

    # CTL HA vsims currently do not have the macs available via 'ypcat -k ethers.sim'

    # If a smoke/CTL vsim
    if ( $blade ) {
        $create_args .= " -smoke ";
        # can still turn this off by setting to 0 rather than undef
        if ( !defined($opt{dynamic_icmacs}) ) {
            $opt{dynamic_icmacs} = 1;
        }
    }

    if ( $opt{icmacs} ) {
        if ( ref($opt{icmacs}) =~ /array/i ) {
            $create_args .= ' -icmac '.join(',', @{$opt{icmacs}});
        } else {
            $create_args .= ' -icmac '.$opt{icmacs};
        }
    } elsif ( $opt{dynamic_icmacs} ) {
        my @icmacs;
        foreach my $vsim ( @pair ) {
            my $mgmt_ip=$vsim->naclnode()->hostrec()->{MGMT_IP};

            if ($mgmt_ip =~ /^([0-9]+?)\.([0-9]+?)\.([0-9]+?)\.([0-9]+?)$/) {
                my $mac = $opt{icmacs_prefix}.":".sprintf("%02x:%02x:%02x", $2, $3, $4);
                push(@icmacs, $mac);
            } else {
                die 'Dynamic IC MAC address calculation failed for VSIM HA, mgmt ip $mgmt_ip in nate host file is not proper';
            }
        }
        $create_args .= ' -icmac '.join(',', @icmacs);

        if ( $SysInit::Vsim::supports{hagroup_noicupdate} ) {
            $create_args .= ' -noicupdate';
        }
    }

    # success message is "... Done."
    my $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                    ";$SysInit::Vsim::vsim_cmd hacreate ".$create_args." -icport ".$port." ".$hagroup." $id1 $id2",
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$id1,
                    "VSIM_NODE=".$node,
                    "NO_RETRY_IF=Insufficient MAC Addresses allocated",
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_hacreate_hagroup_".$hagroup,
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: exists
#   
#  Has an hagroup of this name been created?
#
# Usage:
#   $hagroup->exists()
#
# Parameters: None
#
# Return: 0 | 1
# -----------------------------------------------
sub exists {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my $user  = $self->user();
    my $uhome = $self->uhome();
    my $id    = $self->id();

    $Log->comment("$me: USER=$user;HOME=$uhome;$SysInit::Vsim::vsim_cmd state $id");
    my $result       = `USER=$user;HOME=$uhome;$SysInit::Vsim::vsim_cmd state $id`;
    chomp $result;

    if ( $result !~ /does not exist/ ) {
        return 1;
    }

    return 0;
}

# -----------------------------------------------
# Function: groups_vsim
#   
#  Does this object group this particular vsim?
#
# Usage:
#   $hagroup->groups_vsim( $vsim )
#
# Parameters: SysInit::Vsim object reference
#
# Return: 0 | 1
# -----------------------------------------------
sub groups_vsim {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my $vsim  = shift;
    my @pair = $self->pair();

    foreach my $testvsim ( @pair ) {
        if ( $vsim->node() eq $testvsim->node() ) {
            return 1;
        }
    }

    return 0;
}


1;
