# -*- perl -*-
# $Id$

## @summary SysInit::VsimMcGroup
## @author  dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

SysInit::VsimMcGroup - Class which interacts with vsim mcgroups

=head1 SYNOPSIS

  use SysInit::VsimMcGroup;

=head1 DESCRIPTION

This module defines routines which allow interaction with vsim mcgroup objects

Note: 

No attempt is made here to document all vsim functionality.  For that please
see the following:

http://thebrewery-web.corp.netapp.com/Brewery/brewery/dev/Tools/VSim/index.html

=cut

package SysInit::VsimMcGroup;

use strict;
use NATE::OS;
use Tharn qw($Log);
use Data::Dumper;
use Hostrec qw();
use IPC::Cmd qw(can_run);
use NACL::C::Options;
use NACL::C::SystemNodeDate;
use NACL::C::SystemServicesNtpConfig;
use NACL::STask::Node;
use NACL::Transit;
use Tie::IxHash;
use NACL::GeneralUtils;
use NATE::Exceptions::Argument qw(:try);
use NATE::BaseException qw(:try);
use SysInit::Vsim;

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    array  => [ 'group' ],
    array  => [ 'vsims' ],
    scalar => [ 
        {
            '-default' => undef,
            '-type'    => "SysInit::Vsim",
            '-read_cb' => sub { 
                                  my $vsim;
                                  my @group = ($_[0]->group());
                                  my @vsims = ($_[0]->vsims());
                                  if ( @group ) {
                                      $vsim = $group[0]->first_vsim();
                                      $_[0]->{first_vsim} = $vsim;
                                  } elsif ( @vsims ) {
                                      $vsim = $vsims[0];
                                      $_[0]->{first_vsim} = $vsim;
                                  }
                                  $vsim;
                               },
            '-forward' => [qw/ user uhome node naclnode id 
                               cmode get_seq
                               blade instance
                               spoof_user_env restore_user_env
                               saved_env_HOME saved_env_USER
                               user_hash /],
        },
        'first_vsim' ],
    scalar => [ 'mcgroup' ],
];

# -----------------------------------------------
# Function: init
#   
#  routine used by the constructor
#
#  Note: Since there is not an explicit constructor
#        defined in this code (we're using init() and
#        MethodMaker's default constructor) I will 
#        provide the usage for the contructor here.
#
# Usage:
#   my $vsim = SysInit::VsimMcGroup->new( group => [ $hagrp1, $hagrp2 ] )
#
# Parameters:
#       mcgroup  - name of mcgroup (required)
#       group    - group of hagroups
#
# Return: blessed reference
# -----------------------------------------------
sub init {
    my $self = shift;
    my $me=(caller(0))[3];

    my %opt = ( group    => undef,
                mcgroup  => undef,
                @_ );

    my $group    = $opt{group};

    if ( $group ) {
        my $group_specified_incorrectly = 0;
        # MethodMaker doesn't really handle array initialization properly.
        # it would store the arrayref as the zeroth element of the array

        if ( ref($group) eq 'ARRAY' ) {
            if ( scalar(@$group) > 0 ) {
                $self->add_to_group( @$group );
            }
        } else {
            $group_specified_incorrectly = 1;
        }

        if ( $group_specified_incorrectly ) {
            NATE::Exceptions::Argument->throw("Option \"group\" must be an array reference ".
                                              "pointing to an array of SysInit::VsimHaGroup elements, group = |".Dumper($group)."|");
        }
    }

    $self;
}

# -----------------------------------------------
# Function: add_to_group
#   
#  routine used to add a VsimHaGroup to the mcgroup object
#
# Usage:
#   $mcgroup->add_to_group( $vsimhagroup1, $vsimhagroup2 );
#
# Parameters:
#       One or more SysInit::VsimHaGroup objects
#
# Return: Nothing
# -----------------------------------------------
sub add_to_group {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my @vsimhagroups = @_;
    my @groups = $self->group();
    my @hagroup_names = map {$_->hagroup()} @groups;
    foreach my $hagroup (@vsimhagroups) {
        if ( ref($hagroup) ne 'SysInit::VsimHaGroup' ) {
            NATE::Exceptions::Argument->throw("Can only add objects to \"group\" of type \"SysInit::VsimHaGroup\" but attempting to add ".Dumper(@vsimhagroups));
        } else {
            # Only add if not already there
            my $hagroup_name = $hagroup->hagroup();
            if ( !grep( /^$hagroup_name$/, @hagroup_names ) ) {
                push( @groups, $hagroup );
                my @vsims   = $self->vsims();
                my @havsims = $hagroup->pair();
                push( @vsims, @havsims );
                $self->vsims( @vsims );
            }
        }
    }

    $self->first_vsim();

    $self->group( @groups );
}

# -----------------------------------------------
# Function: show
#   
#  Invoke the "mcshow" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   my $st = $mcgroup->show();
#   SysInit::Vsim::wait_until_finished( subtests => [$st] );
#   @results = $st->results();
#   my $misc = $results[0]->misc();
#   my $out  = $misc->{VSIM_CMD_OUTPUT};
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub show {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 600,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    $self->spoof_user_env();

    my $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$SysInit::Vsim::vsim_cmd mcshow ".$self->mcgroup(),
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-bg',
                    '-onexit', sub { },
                    '-timeout', $opt{timeout},
                    '-runid', $self->get_seq()."_mcshow_".$self->mcgroup(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: rm
#   
#  Invoke the "rm" command on an mcgroup to remove it.
#
# Usage:
#
#   push( @subtests, $mcgroup->rm() );
#   SysInit::Vsim::wait_until_finished( subtests => \@subtest );
#
# Parameters: 
#       
#   force       => Force removal even if cores are present (default 1)
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub rm {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( force   => 1,
                timeout => 900,
                @_ );

    my $node     = $self->node();

    my $force = $opt{force} ? " -f" : "";

    # I don't have all of the specific error codes from vsim to 
    # reliably know what error code "does not exist" corresponds to
    # so I am sticking with the strings for now
    my $expected = "simulator removed";
    if ( $force ) {
        $expected .= "\|does not exist";
    }

    $self->spoof_user_env();

    my $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$SysInit::Vsim::vsim_cmd rm ".$force." ".$self->mcgroup(),
                    "EXPECTED=$expected",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_rm_mcgroup_".$self->mcgroup(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: exists
#   
#  Has an mcgroup of this name been created?
#
# Usage:
#   $mcgroup->exists()
#
# Parameters: None
#
# Return: 0 | 1
# -----------------------------------------------
sub exists {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my $user  = $self->user();
    my $uhome = $self->uhome();
    my $id    = $self->id();

    $Log->comment("$me: USER=$user;HOME=$uhome;$SysInit::Vsim::vsim_cmd state $id");
    my $result       = `USER=$user;HOME=$uhome;$SysInit::Vsim::vsim_cmd state $id`;
    chomp $result;

    if ( $result !~ /does not exist/ ) {
        return 1;
    }

    return 0;
}

# -----------------------------------------------
# Function: reserve
#
#  Invoke the "reserve" command on an hagroup
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   push( @subtests, $mcgroup->reserve() );
#   SysInit::Vsim::wait_until_finished( subtests => \@subtest );
#
# Parameters:
#
#   simdomain   => An attempt is made to guess the correct simdomain
#                  based on the "domainname" of the client running
#                  this command.  Will be something like:
#                  "sim.rtp.netapp.com"
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub reserve {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( domain => $SysInit::Vsim::simdomain,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $simdomain = $opt{domain};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                        ";$SysInit::Vsim::vsim_cmd reserve -restype vsimesx -domain $simdomain ".$self->mcgroup(),
                    'EXPECTED_RC=0',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-onexit', sub { },
                    '-runid', $self->get_seq()."_reserve_mcgroup_".$self->mcgroup(),
                    %{$self->user_hash()} );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: rehost
#
#  Invoke the "rehost" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->rehost() );
#   }
#   SysInit::Vsim::wait_until_finished( subtests => \@subtest );
#
# Parameters:
#
#   simdomain   => An attempt is made to guess the correct simdomain
#                  based on the "domainname" of the client running
#                  this command.  Will be something like:
#                  "sim.rtp.netapp.com"
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub rehost {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( domain => $SysInit::Vsim::simdomain,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $blade     = $self->blade();
    my $simdomain = $opt{domain};
    my $st;

    if ( !$blade ) {
        $self->spoof_user_env();

        $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                        "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                                      ";$SysInit::Vsim::vsim_cmd rehost -restype vsimesx -domain $simdomain ".$self->mcgroup(),
                        'EXPECTED_RC=0',
                        "VSIM_INDEX=".$self->id(),
                        "VSIM_NODE=".$self->node(),
                        '-timeout', $opt{timeout},
                        '-bg',
                        '-onexit', sub { },
                        '-runid', $self->get_seq()."_rehost_mcgroup_".$self->mcgroup(),
                        %{$self->user_hash()} );

        $self->restore_user_env();
    }

    return $st;
}

# -----------------------------------------------
# Function: extend or renew
#
#  Invoke the "renew" command on an mcgroup
#  Note: the extend() and renew() methods are
#  synonyms for one another.
#
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   push( @subtests, $mcgroup->extend( days => 3 ) );
#   SysInit::Vsim::wait_until_finished( subtests => \@subtest );
#
# Parameters:
#
#   days        => 1-7 (default 7)
#   timeout     => Number of seconds until command times out (default 900)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub extend {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }
    return $self->renew( @_ );
}

sub renew {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( days    => 7,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $blade     = $self->blade();
    my $days      = $opt{days};
    my $st;

    # I don't have all of the specific error codes from vsim to
    # reliably know what error code "End time must be less" corresponds to
    # so I am sticking with the strings for now
    if ( !$blade ) {
        $self->spoof_user_env();

        $st = Tharn::subtest( $SysInit::Vsim::exec_cmd,
                        "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$SysInit::Vsim::vsim_cmd renew -res $days ".$self->mcgroup(),
                        "EXPECTED=Successful|End time must be less|aximum",
                        "VSIM_INDEX=".$self->id(),
                        "VSIM_NODE=".$self->node(),
                        '-timeout', $opt{timeout},
                        '-bg',
                        '-onexit', sub { },
                        '-runid', $self->get_seq()."_extend_mcgroup_".$self->mcgroup(),
                        %{$self->user_hash()} );

        $self->restore_user_env();
    }

    return $st;
}

# -----------------------------------------------
# Function: groups_vsim
#
#  Does this object group this particular vsim?
#
# Usage:
#   $mcgroup->groups_vsim( $vsim )
#
# Parameters: SysInit::Vsim object reference
#
# Return: 0 | 1
# -----------------------------------------------
sub groups_vsim {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my $vsim  = shift;
    my @vsims = $self->vsims();

    foreach my $testvsim ( @vsims ) {
        if ( $vsim->node() eq $testvsim->node() ) {
            return 1;
        }
    }

    return 0;
}



1;
