# -*- perl -*-
# $Id$

## @summary SysInit::VsimUtils
## @author  dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

SysInit::VsimUtils - Utility routines used by SysInit vsim scripts

=head1 SYNOPSIS

  use SysInit::VsimUtils;

=head1 DESCRIPTION

This module defines routines that either work on more than 1 vsim class at a time
or are non-OO helper functions.

Note: 

No attempt is made here to document all vsim functionality.  For that please
see the following:

http://thebrewery-web.corp.netapp.com/Brewery/brewery/dev/Tools/VSim/index.html

=cut

package SysInit::VsimUtils;

use strict;
use NATE::OS;
use Tharn qw($Log);
use Data::Dumper;
use Connectrec qw();
use Hostrec qw();
use Storable;
use SysInit::Vsim;
use SysInit::VsimHaGroup;
use SysInit::VsimMcGroup;
use NATE::BaseException qw(:try);
use Params::Validate
    qw(validate validate_with BOOLEAN SCALAR ARRAYREF HASHREF OBJECT SCALARREF);

# -----------------------------------------------
# Function: 
#   convenience routine that creates in-memory SysInit::Vsim*Group objects
#   from the "vsim mcshow" and "vsim hashow" commands.
#   
#
# Usage:
#
#    my $objects = SysInit::VsimUtils->retrieve_vsim_objects( vsims => \@vsims )
#
# Parameters:
#   
#    None
#
# Return:
#   
#    a hash containing all related hagroups and mcgroups related to the
#    vsims passed in.
#
# Assumptions:
# -----------------------------------------------
sub retrieve_vsim_objects {
    my $self = shift;
    my $me=(caller(0))[3];
    if ( ref($self) ) {
        die "$me: Can only call as class method";
    }

    my %opt = validate_with(
        params => \@_,
        spec   => {
            vsims             => { type => ARRAYREF },
        },
    );

    my $vsims = $opt{vsims};

    my @nonha_vsims;
    my %nonmc_hagroups;
    my %mcgroups;
    my %mc_hagroups;
    my %hash;

    foreach my $vsim ( @$vsims ) {

        if ( ref($vsim) ne "SysInit::Vsim" ) {
            die "$me: Only objects of type SysInit::Vsim can be passed in via the -vsims option";
        }

        my ($mcgroup, $hagroup) = $vsim->mcgroup_from_dparam();

        if ( $mcgroup ) {
            if ( !exists $mcgroups{$mcgroup} ) {
                $mcgroups{$mcgroup} = SysInit::VsimMcGroup->new( mcgroup => $mcgroup );
            }
            if ( !exists $mc_hagroups{$mcgroup}{$hagroup} ) {
                $mc_hagroups{$mcgroup}{$hagroup} = SysInit::VsimHaGroup->new( hagroup => $hagroup );
            }
            $mc_hagroups{$mcgroup}{$hagroup}->add_to_pair( $vsim );
        } else {
            $hagroup  = $vsim->hagroup_from_dparam();
            if ( $hagroup ) {
                if ( !exists $nonmc_hagroups{$hagroup} ) {
                    $nonmc_hagroups{$hagroup} = SysInit::VsimHaGroup->new( hagroup => $hagroup );
                }
                $nonmc_hagroups{$hagroup}->add_to_pair( $vsim );
            } else {
                if ( $vsim->exists() ) {
                    push(@nonha_vsims, $vsim);
                }
            }
        }
    }

    if ( %mcgroups ) {
        foreach my $mcgroup (sort keys %mc_hagroups) {
            foreach my $hagroup (sort keys %{$mc_hagroups{$mcgroup}}) {
                $mcgroups{$mcgroup}->add_to_group( $mc_hagroups{$mcgroup}{$hagroup} );
            }
        }
    }

    if ( %mcgroups ) {
        $hash{mcgroups} = Storable::dclone(\%mcgroups);
    }

    if ( %nonmc_hagroups ) {
        $hash{nonmc_hagroups} = Storable::dclone(\%nonmc_hagroups);
    }

    if ( @nonha_vsims ) {
        $hash{nonha_vsims} = Storable::dclone(\@nonha_vsims);
    }
    

    if ( $Log->may_debug() ) {
        local $Data::Dumper::Sortkeys = 1;
        local $Data::Dumper::Maxdepth = 4;
        
        print "$me: Vsim objects = ".Dumper(\%hash)."\n";
    }

    return \%hash;
}

# -----------------------------------------------
# Function: decode_image
#
#   convenience routine that parses an image string and
#   determines the build root, whether it is an ibm build
#   or not, what platform it is.  It will also convert
#   a non-sim image string into a sim image string.
#   
#
# Usage:
#
#    my %image = SysInit::VsimUtils->decode_image( $image );
#
# Parameters:
#   
#    image => string representing an Ontap Vsim image
#             ex: /x/eng/rlse/DOT/devN_tested/final/bedrock/export/x86_64.sim/tarball/image.tgz
#
# Return:
#   
#    a hash containing the image, buildroot, platform and ibm_mode.
#
#    return (image => $image, buildroot => $buildroot, platform => $platform, ibm_mode => $ibm_mode);
#
# Assumptions:
# -----------------------------------------------
sub decode_image {
    my $self = shift;
    my $me=(caller(0))[3];
    if ( ref($self) ) {
        die "$me: Can only call as class method";
    }

    my $image = shift;
    my $platform;
    my $buildroot;

    my $ibm_mode = 0;

    my @platforms = qw( x86_64 x86_64.debug x86_64.debug.gcov x86_64.domain x86_64.debug.domain );

    if ( $image ) {
        # Allow callers to be a little sloppy.  Allow them to use the image name of a NON-sim build
        # and convert it to a sim build
        my $orig_image = $image;
        if ( $image !~ /x86(\S+?)\.sim\// ) {
            if ( $image =~ /x86(\S+?)\.ibm\// ) {
                if ( $image !~ /x86(\S+?)\.sim\.ibm\// ) {
                    $image =~ s/x86(\S+?)\.ibm\//x86$1\.sim\.ibm\//;
                }
            } else {
                $image =~ s/x86(\S+?)\//x86$1\.sim\//;
            }
            if ( !-e $image ) {
                die "$me: IMAGE $orig_image is not a vsim image, attempted to look for $image, but that does not exist";
            }
        } else {
            # image is already a sim image
            if ( !-e $image ) {
                die "$me: IMAGE $image does not exist";
            }
        }

        my @path = split( '/', $image );
        my @root;
        my $found_root = 0;
        foreach my $part (@path) {
            push( @root, $part );
            my $tmp = join( '/', @root );
            $tmp .= "/P4ENV";
            if ( -e $tmp ) {
                $found_root = 1;
                last;
            }
        }

        if ( !$found_root ) {
            die "$me: Could not find build root (where P4ENV resides) of IMAGE";
        }

        $buildroot = join( '/', @root );
    } else {
        die "$me: Must supply an Ontap image";
    }

    if ( !$buildroot ) {
        die "$me: Could not deduce the buildroot from the Ontap image";
    }
    
    foreach my $plat ( @platforms ) {
        if ( $image =~ /$plat\.sim/ ) {
            $platform = $plat;
            $Log->comment("$me: VSIM_PLATFORM determined from IMAGE is \"$platform\"");
            last;
        } elsif ( $image =~ /x86(\S+?)\.ibm\// ) {
            if ( $image =~ /debug/ ) {
                $platform = 'x86_64.debug.sim.ibm';
            } else {
                $platform = 'x86_64.sim.ibm';
            }
            last;
        }
    }

    if ( !$platform ) {
        $Log->comment("$me: Unable to determine platform from $image, defaulting to \"x86_64\"");
        $platform = 'x86_64';
    }

    if ( $image =~ /\.sim\.ibm/ ) {
        $Log->comment("$me: IMAGE is an IBM build");
        $ibm_mode = 1;
    }

    return (image => $image, buildroot => $buildroot, platform => $platform, ibm_mode => $ibm_mode);
}

1;
