# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Sysconfig.pm#1 $

## @summary OO interface to get the sysconfig output of the filer.
## @author dl-central-libs
## @status Shared
## @pod here


=head1 NAME

Sysconfig - OO interface to get the sysconfig output of the filer.

=head1 SYNOPSIS

    my $obj = Sysconfig->new('tada-u1');

    or

    my $h = host($FILER);
    my $obj = Sysconfig->new($h);


=head1 DESCRIPTION

METHOD new(FILER)

Given a filer creates a new object containing the "sysconfig -v"
output of the filer. The filer can be a DNS name or a NATE hostobject.
Be aware if you are passing in a connectrec. It internally converts every
connectrec to the conntype of "rsh-socket". So, don't expect it to work
over console/telnet.

    my $obj = Sysconfig->new('tada-u1');
    or
    my $h = host($FILER);
    my $obj = Sysconfig->new($h);

It can also be used to parse the raw output of sysconfig command got from
outside the Sysconfig class for Cmode filer. It can be invoked as:

	my $obj = Sysconfig->new($FILER,$sysconfig);  

Here $sysconfig is the sysconfig output of the Cmode filer.


METHOD slot('slot_number','argument_to_query')

Query a particular slot for various arguments of it.
For example a typical slot 0 on the filer looks like:

slot 0: System Board 2185 MHz (NetApp System Board VIII A0)
Model Name:         FAS960
Part Number:        110-00900
Revision:           A0
Serial Number:      BETA0008
Firmware release:   4.2_i2
BMC info:           IPMI v1.5, BMC rev 0.20, SDR Version 00.18
Processors:         2
Memory Size:        6144 MB

Different fields can be queried in the following fashion.
Please note how the field names are joined with '_'.

    my $model_name = $obj->slot('0','model_name');
    my $part_number = $obj->slot('0','part_number');
    my $revision = $obj->slot('0','revision');
    my $serial_number = $obj->slot('0','serial_number');

And the same for others. Every slot has a type and info field, which would
return the type and info of the card present in the slot.
For example:
    my $type = $obj->slot('0','type');
        would return $type as "System Board 2185 MHz".

    my $info = $obj->slot('0','info');
        would return $info as "NetApp System Board VIII A0".

For slots holding the disks, one can also query the disks:
For Example:
    my @disks = $obj->slot('5a','disks');
        in scalar context it return the number of disks present onto the slot.

In case you are querying for a slot that is not present on the filer
it would croak. And for any fields which is not present on the slot 
it would return an empty string.


OTHER Getter Methods

Apart from the above methods one also have access to these methods. 
Their names are self documenting for what they return.

    1.) release

    2.) system_id

    3.) partner_id

    4.) partner_name

    5.) system_serial_number

    6.) system_rev

    7.) backplane_part_number*

    8.) backplane_rev*

    9.) backplane_serial_number*

    10.) all_slots.

* Note that some methods like backplane_part_number, backplane_rev, 
backplane_serial_number do not exist on some platforms like Excelsior.


=cut


package Sysconfig;

use strict;
use warnings;

use English;
use Carp qw(croak);
use Tharn qw(connect host say logcomment logresult param tstlog);

use vars qw($AUTOLOAD);


sub new {
    my ($that,$filer,$sysconfig) = @_;
    my $class = ref($that) || $that;
    my $self = {};
    bless $self,$class;
    $self->{filer} = $filer || croak "The filer is undefined";
    $self->{sysconfig} = $sysconfig if($sysconfig);
    $self->_init();
    return $self;
}


sub _init {
    my ($self) = @_;
    my $out;
    if(!$self->{sysconfig}) {
        my $conn = $self->{filer};
        if (ref($conn) =~ /Connectrec/) {
            $conn = connect(host($conn->{hostp}->{id}) . "#rsh-socket");
        } elsif (ref($conn) =~ /Hostrec/) {
            $conn = connect(host($conn->{id}) . "#rsh-socket");
        } else {
            $conn = connect($conn . "#rsh-socket");
        }
        (undef,undef,$out) = say($conn,"sysconfig -v",'-eval_error');
        croak($EVAL_ERROR."\n") if($EVAL_ERROR);
    } else {
        $out = $self->{sysconfig};
    };
    my @lines = split(/\n/,$out);
    my $slot = undef;
    my $regex = qr{(.*?):\s*(.*?)\s*$};
    foreach my $line(@lines) {
        chomp($line);
        $line =~ s/^\s+//;
        if (($line =~ /slot/) || defined($slot)) {
            if($line =~ /slot (\d+):/) {
                my $slot_num = $1;
                my ($key,$value) = ($line =~ /$regex/);
                ($slot) = ($value =~ /Adapter (\w+)/) ? $1 : $slot_num;
                if($value =~ /(.*?)\((.*?)\)$/) {
                    ($self->{slots}->{$slot}->{type},
                        $self->{slots}->{$slot}->{info}) = ($1,$2); 
                } else {
                    $self->{slots}->{$slot}->{type} = $value;
                }
            } elsif ($line =~ /\s*(\d+):\s*(NETAPP.*)/) {
                $self->{slots}->{$slot}->{disks}->{$1} = $2;
            } else {
                if(my ($key,$value) = ($line =~ /$regex/)){
                    if(($key !~ /Shelf \d+/i) 
                            && ($value =~ /^(\w+)\s+(.*?):\s+(.*?)$/)) {
                        $self->{slots}->{$slot}->{modify_key($key)} = $1;
                        $self->{slots}->{$slot}->{modify_key($2)} = $3;
                    } elsif($key =~ /MAC Address/i) {
                        ($self->{slots}->{$slot}->{modify_key($key)},
                            $self->{slots}->{$slot}->{nic_status})
                                = ($value =~ /(.*?)\s*\((.*?)\)/) 
                    } else {
                        $self->{slots}->{$slot}->{modify_key($key)} = $value;
                    }
                } else {
                    $line =~ s/\)$//;
                    $line =~ s/^\(//;
                    $self->{slots}->{$slot}->{info} .= $line;
                }
            }
        } else {
            if($line =~ /NetApp Release/) { # does this have to change for IBM?
                ($self->{release}) = split(/:/,$line);
            } elsif ($line =~ /System ID/) {
                my ($selfid,$partner) = split(/;/,$line);
                ($self->{system_id}) = ($selfid =~ /\b(\d+)\b/);
                if (defined($partner)) {
                    ($self->{partner_id}, $self->{partner_name}) =
	                ($partner =~ /\s+(\d+)\s+\((.*)\)/);
                }
            } else {
                my ($key,$value) = ($line =~ /$regex/);
                $self->{modify_key($key)} = $value;
            }
        }
    }
}

sub slot {
    my ($self,$slot,$arg) = @_;
    croak "Either the slot or the arg is undefined"
        if(!defined($slot) || !defined($arg));
    $slot   = lc($slot);
    $arg    = lc($arg);
    if (!exists($self->{slots}->{$slot})) {
        croak "The slot \"$slot\" is not present on the filer";
    } elsif (!exists($self->{slots}->{$slot}->{$arg})) {
        return "";
    } elsif ($arg =~ /disk/) {
        my @disks;
        foreach my $key(keys %{$self->{slots}->{$slot}->{disks}}) {
            my $value = $self->{slots}->{$slot}->{disks}->{$key};
            push @disks,join(' ',$key,$value);
        }
        return @disks;
    } else {
        return $self->{slots}->{$slot}->{$arg};
    }
}

sub modify_key {
    my ($key) = @_;
    $key =~ s/\s+/_/g;
    return lc($key);
}

sub AUTOLOAD {
    my ($self,$slot,$arg) = @_;
    $AUTOLOAD =~ s/.*::(\w+)/$1/;
    return if($AUTOLOAD eq "DESTROY");
    $AUTOLOAD = lc($AUTOLOAD);
    return (keys %{$self->{slots}}) if($AUTOLOAD eq "all_slots");
    croak "Undefined method \"$AUTOLOAD\"" if(!exists($self->{$AUTOLOAD}));
    return $self->{$AUTOLOAD};
}


# one has to visually verify the output of this self test 'testme' routine
# use util/test_pkg.thpl to invoke the testme
# XXX - is there a better way to self test ?
sub testme() {

package main;
	
	my ($self,$sysconfig) = @_;
    our ($FILER);
    param('FILER',       '-mesg',    "Filer to test");

package Sysconfig;

    logcomment("Performing self test of package Sysconfig on FILER $FILER");
    my $obj = Sysconfig->new($FILER,$sysconfig);

    tstlog('-commline',"Release is ",$obj->release());
    tstlog('-commline',"System ID is ",$obj->system_id());
    tstlog('-commline',"Partner ID is ",$obj->partner_id());
    tstlog('-commline',"Partner name is ",$obj->partner_name());
    tstlog('-commline',"System Serial Number is ",$obj->system_serial_number());
    tstlog('-commline',"System Rev is ",$obj->system_rev());
    # backplanes dont exist on all filers so dont call the methods here

    tstlog('-commline', "Model name is ", $obj->slot('0','model_name'));

    my @all_slots = $obj->all_slots();
    tstlog('-commline',"All Slots is @all_slots");
    logresult('fail', 'No slot information') if (!scalar(@all_slots));

    # print the type and info for all slots
    my ($fc_host_found, $nvram_found); # watch out for some expected slots
    foreach my $slot (@all_slots) {
        my $type = $obj->slot($slot,'type');
        my $info = $obj->slot($slot,'info');
        tstlog('-commline', "Slot:$slot\tType:$type\tInfo:$info");

        if ($type =~ /FC Host/) {
            $fc_host_found = 1;
            my @disks = $obj->slot($slot,'disks');
            tstlog('-commline', "Disks are @disks");
        }
        if ($type =~ /NVRAM/) {
            $nvram_found = 1;
        }
    }

    # fail if some expected types are not found
    logresult('fail', "No Host adapter found") if (!$fc_host_found);
    logresult('fail', "No NVRAM found") if (!$nvram_found);
}

1;



=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = sysconfig

    ONTAP_ZAPI_APIS = none

=head1 BUGS

    Might not work for IBM models
    Does not work well when there is more than one type for a particular slot.
    Example,
    tada-c63% rsh -l root: tada-u36 "sysconfig -v 7"
            slot 7: NVRAM (NetApp NVRAM IV)
                    Revision:           C0
                    Serial Number:      72806
                    Memory Size:        256 MB
                    Battery Status:      OK
                    Running Microcode Revision: 0115
            slot 7: NetApp ATA/IDE Adapter 7 (0x0000efe0)
    tada-c63%

=head1 AUTHOR

    dl-central-libs
	$Id: //depot/prod/test/Rsierranevada/lib/Sysconfig.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut


