package Syslog::SyslogDaemon;

use strict;

require Exporter;
use Options;
use Tharn qw(host snooze fatal_result subtest $Parent $Log);

use NATE::Result;
use Data::Dumper;
use IO::Socket;
use IO::Socket::INET6;

sub new {
    my $this = shift;
    my $class = ref($this) || $this;

    my %options = (
        proto=>'udp',
        port=>'514',
        ipv6=>'',
        '-client'=>'',
        @_
    );

    my @subArgs;
    if (!$options{'-client'}) {
        $Log->error("SyslogDaemon requires a -client parameter \n");
        return 0;
    }

    my $addr;
    if ( $options{ipv6} ) {
        $addr = $options{ipv6};
    } else {
       # Not sure why, but with the LocalAddr option with the client name, name apparently does not
       # get resolved.  Maybe this is because client is windows?
        my $clientobj = host($options{-client});
        $addr = $clientobj->default_ip();
    }

    my @subArgs = ('{use Syslog::SyslogDaemon; Syslog::SyslogDaemon::run(proto=>"'.$options{proto}.'"'.
                                                                       ',port=>"'.$options{port}.'"'.
                                                                       ',client=>"'.$addr.'"'.
                                                                       ',ipv6=>"'.$options{ipv6}.'"'.
                                                                       '); }', '-bg', '-client', $options{'-client'} );

    my $subtest = subtest @subArgs;
    my $self = {
        "subtest" => $subtest,
        @_,
    };

    bless($self, $class);
    sleep(5);
    return $self;
}

sub ready {
    my $self = shift;
    my $message;
    my $i = 0;
    while ($i++ < 10) {
           eval {
                  $message = $self->{"subtest"}->message_get(type =>
                  "SYSLOGVERIFY", timeout => 3);
                  };
           if ($message =~ /SYSLOGREADY/) {
               $Log->comment("SYSLOGD started\n");
               return 1; 
           }
           sleep(20);

    }
    $Log->comment("Failed to start SYSLOGD\n");
    return 0;

}

sub stop {
    my $self = shift;
    $self->{"subtest"}->kill();
    return;
}

sub get_entry {
    my $self = shift;
    my $raw_msg;
    my $msg;
    my $head;
    my $event;
    my $pri;
    my $sev;
    my $fac;
    
    $raw_msg = $self->{"subtest"}->message_get(type => "Syslog", timeout => 3);
                                            #,on_timeout => sub { return undef; } );
    $Log->comment("Got message from daemon: ".Dumper($raw_msg)."\n");
    if ($raw_msg =~ /<(\d+)>(.*)\s\[(.*)\]:\s(.*)/) {
        $Log->trace("one: $1 two: $2 three: $3 four: $4\n");
        $event = $3;
        $msg = $4;
    } elsif ($raw_msg =~ /<(\d+)>(\w+\s*\d+\s*\d+:\d+:\d+)\s*(.*)/) {
        $Log->trace("Xone: $1 two: $2 three: $3\n");
        $event = $3;
        $msg = $3;
    } else {
        return undef;
    }

    $pri = $1;
    $head = $2;
    $sev = $pri % 8;
    $fac = ($pri - $sev) / 8;
    if (defined $raw_msg) {
        return {"pri"=>$pri,"head"=>$head,"event"=>$event,"msg"=>$msg,"sev"=>$sev,"fac"=>$fac};
    }
    return undef;
}

sub get_all {
    my $self = shift;
    my @msgConcat = ();
    my $msgTmp = undef;

    while (1) {
        $msgTmp=undef;
        #get an entry that is on the queue, a message reported from the server
        eval {
            $msgTmp = $self->get_entry(@_);
        };
        #if no message available, then exit
        if (!defined $msgTmp) {
            last;
        }
        #else add the message to the array
        else {
            $Log->trace("Adding message to array: ".Dumper($msgTmp));
            push @msgConcat, $msgTmp;
        }
    }
    #return the array of messages
    return \@msgConcat;

}

sub get_all_msg_concat {
    my $self = shift;
    my $concat = "";

    my $msgs = $self->get_all(@_);
    foreach my $i (@{$msgs}) {
        $concat = $concat."\n".$i->{event}." ".$i->{msg};
    }
    return $concat;
}

sub run {
    my %options = (
        @_
    );

    my $MAXLEN = 1524;
    my ($buf, $sock);

    my $addr;
    if ( $options{ipv6} ) {
        $addr = $options{ipv6};
    } else {
        $addr = $options{client};
    }
    $Log->comment("SyslogDaemon: Socket LocalAddr($addr), LocalPort($options{port}), Proto($options{proto})");

    $sock = IO::Socket::INET6->new(LocalAddr => "$addr",
                                   LocalPort => "$options{port}",
                                   Proto     => "$options{proto}",
                                   ReuseAddr => 1 );
    if (not defined $sock) {
        my $retries = 0;
        for ($retries = 1; not defined $sock && $retries < 10; $retries++) {
            $sock = IO::Socket::INET6->new(LocalAddr => "$addr",
                                           LocalPort => "$options{port}",
                                           Proto     => "$options{proto}",
                                           ReuseAddr => 1 );
            $Log->comment("Waiting $retries seconds for Socket LocalAddr($addr),".
                          " LocalPort($options{port}),".
                          " Proto($options{proto}) to free up");
            snooze($retries);
        }

        if (not defined $sock) {
            $Log->error("No socket");
            return fatal_result("No socket");
        }
    }
    $Parent->message_put(type => "SYSLOGVERIFY", message => "SYSLOGREADY");
    do {
        $sock->recv($buf,$MAXLEN);
        $Log->comment("Received message: $buf\n");
        $Parent->message_put(type => "Syslog", message => $buf);
    } while (1);

}

sub dump {
    my $self = shift;
    print Dumper($self);
};

1;

