# Hey emacs, this is basically -*- perl -*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/TableBackup.pm#1 $
#
# Copyright (c) 2011 Network Appliance, Inc.
# All rights reserved.
#
## @summary Cleanup modifications to simple SMF tables.
## @author samuele@netapp.com
## @pod here

=head1 NAME

TableBackup

=head1 SYNOPSIS

 my $Backup = {};
 sub setup {
     my $cli = NACL::APISet->new(hostobj => $CI->hostrec(),
                        category => "Node", interface => "CLI", set => "CMode");
     $Backup = TableBackup::get_table_states($cli, @table_commands);
 }
 sub cleanup {
     my $cli = NACL::APISet->new(hostobj => $CI->hostrec(),
                        category => "Node", interface => "CLI", set => "CMode");
     while (my ($command, $orig_state) = each(%$Backup)) {
         TableBackup::restore_table(apiset     => $cli,
                                    command    => $command,
                                    restore_to => $orig_state);
     }
 }

=head1 DESCRIPTION

Cleanup modifications to simple SMF tables.

For tables with the standard semantics for create, delete, and modify
operations,


=head1 ROUTINES

=cut

package TableBackup;

use strict;
use Tharn qw($Log);
use Data::Dumper;
use NACL::APISet;

our @EXPORT_OK = qw(
    get_table_states
    restore_table
);

=head2 get_table_states

Get the current states of a list of tables.

=over

=item Synopsis

 $Backup = TableBackup::get_table_states($cli, @table_commands);

=item Arguments

=over

=item

A NACL::APISet object for C-mode CLI.

=item

An array of tables as they are known in the CLI. E.g. the aggr table would be
"storage aggregate".

=back

=item Returns

A reference to a hash of APISet response objects keyed by the table commands
that were passed in.

=back

=cut

sub get_table_states($@) {
    my $cli = shift;
    my @tables = @_;
    my %states = map {$_ => _table_action($cli, $_, "show")} @tables;
    return \%states;
}

=head2 restore_table

Bring an SMF table to a given state by modifying, creating, and deleting rows
as necessary. This only works for tables with the standard semantics for
create, delete, and modify operations.

=over

=item Synopsis

 TableBackup::restore_table(apiset      => $cmode_cli,
                            command     => $table_command,
                            restore_to  => $orig_state,
                            [keys       => \@table_keys,]
                            [creatable  => \@creatable_fields,]
                            [modifiable => \@modifiable_fields,]);

=item Arguments

=over

=item apiset

A NACL::APISet object for C-mode CLI.

=item command

The part of a CLI command that identifies a table. E.g. the aggr table would be
"storage aggregate".

=item restore_to

An APISet response object containing the state that the table should be
restored to, e.g. a value from the hash returned by get_table_states().

=item keys (optional)

An array reference listing the table's keys. By default, keys are derived from
a show command with argument -fields "".

=item creatable (optional)

An array reference listing the fields to use when creating a row. By default,
all fields in the restore_to object are used in the create command.

=item modifiable (optional)

An array reference listing the fields to use when modifying a row. By default,
all fields in the restore_to object are used in the modify command.

=back

=item Returns

Nothing.

=back

=cut

sub restore_table(%) {
    my %args = @_;

    my $cli        = $args{apiset};     # APISet
    my $cmd        = $args{command};    # SMF command name
    my $orig_resp  = $args{restore_to}; # Response from $api->command_show()
    # optional
    my $keys       = $args{keys};
    my $creatable  = $args{creatable};
    my $modifiable = $args{modifiable};

    if (!$keys) {
        # Figure out what the keys are
        my $resp = _table_action($cli, $cmd, "show", fields => '""');
        my $row = $resp->get_parsed_output()->[0];
        my @keys_arr = keys(%$row);
        $keys = \@keys_arr;
    }

    my $curr_resp = _table_action($cli, $cmd, "show");

    # Delete rows from curr_resp that are not found in orig_resp
    foreach my $curr_row (@{$curr_resp->get_parsed_output()}) {
        my %curr_keys = _hash_slice($curr_row, @$keys);
        if (!$orig_resp->search(%curr_keys)) {
            _table_action($cli, $cmd, "delete", %curr_keys);
        }
    }

    # Create rows from orig_resp that are not found in curr_resp.
    # Modify rows that are in both, but have different values.
    foreach my $orig_row (@{$orig_resp->get_parsed_output()}) {
        _normalize_smdb_data($orig_row);
        my %orig_keys = _hash_slice($orig_row, @$keys);
        my %search_keys = _apiset_resp_search_exact(%orig_keys);
        my @curr_match_rows = $curr_resp->search(%search_keys);
        if (scalar(@curr_match_rows) > 1) {
            $Log->debug("Search $cmd for ".Dumper(\%search_keys));
            $Log->debug("Found ".Dumper(\@curr_match_rows));
            die("Incorrect keys for '$cmd': ".Dumper($keys));
        }

        my $curr_row = shift(@curr_match_rows);
        if ($curr_row) {
            # The original row still exists. Check if it was modified
            _normalize_smdb_data($curr_row);

            # If modifiable fields were provided, only use those. Otherwise,
            # use any field that has changed.
            my %orig_fields;
            if (ref($modifiable) eq "ARRAY") {
                %orig_fields = _hash_slice($orig_row, @$modifiable);
            } else {
                %orig_fields = %$orig_row;
            }

            # Find all fields that have been modified.
            my %mod_fields = ();
            while (my ($field, $value) = each(%orig_fields)) {
                if ($curr_row->{$field} ne $value) {
                    $mod_fields{$field} = $value;
                }
            }

            if (%mod_fields) {
                _table_action($cli, $cmd, "modify", %orig_keys, %mod_fields);
            }
        } else {
            # The original row no longer exists. Recreate it.
            # If creatable fields were provided, only use those. Otherwise, use
            # all fields in the create command.
            if (ref($creatable) eq "ARRAY") {
                my %orig_create = _hash_slice($orig_row, @$creatable);
                _table_action($cli, $cmd, "create", %orig_keys, %orig_create);
            } else {
                _table_action($cli, $cmd, "create", %$orig_row);
            }
        }
    }
}

################################################################################
# Subroutine:   _table_action
# Parameters:   cli     - APISet for c-mode CLI
#               command - The first part of the command, which identifies the
#                         table.
#               action  - The last part of the command, which identifies the
#                         action, e.g. "show", "create", etc.
#               args    - Hash of command arguments.
# Returns:      The APISet response object.
# Purpose:      Wrapper for running CLI commands with an APISet when the table
#               part of the command and the action part of the command are in
#               separate variables.
################################################################################
sub _table_action($$$%) {
    my $cli     = shift;
    my $command = shift;
    my $action  = shift;
    my %args    = @_;

    $command .= " $action";
    $Log->trace("$command ".Dumper(\%args)) unless ($action eq "show");
    $command =~ s/ |-/_/g;
    return $cli->$command(%args);
}

################################################################################
# Subroutine:   _normalize_smdb_data
# Parameters:   The array reference returned from APISet->get_parsed_output(),
#               or one of the hash references contained in that array.
# Returns:      Nothing. Modifies the input.
# Purpose:      Reformat the parsed output of an APISet response object so the
#               data may be used in other CLI commands.
#               - The APISet may create special fields that start and end with
#                 "__". These are deleted.
#               - Fields whose value is "-" are deleted.
#               - Fields whose value is an array are converted to a single csv
#                 string.
################################################################################
sub _normalize_smdb_data($) {
    my $obj = shift;
    if (ref($obj) eq "HASH") {
        while (my ($key, $val) = each(%$obj)) {
            if ($key =~ /__.+__/) {
                delete $obj->{$key};
            } elsif (!ref($val) && $val eq "-") {
                delete $obj->{$key};
            } elsif (ref($val) eq "ARRAY") {
                $obj->{$key} = join(",", @$val);
            }
        }
    } elsif (ref($obj) eq "ARRAY") {
        map(_normalize_smdb_data($_), @$obj);
    }
}

################################################################################
# Subroutine:   _hash_slice
# Parameters:   hash - Hash reference to get a slice of.
#               keys - The keys to be included in the slice.
# Returns:      A hash with the specified keys and values from the given hash.
# Purpose:      Get a subset (aka slice) of a hash. Used for readability.
################################################################################
sub _hash_slice($@) {
    my $hash = shift;
    my @keys = @_;
    return map {$_ => $hash->{$_}} @keys;
}

################################################################################
# Subroutine:   _apiset_resp_search_exact
# Parameters:   A hash of field/value pairs.
# Returns:      A copy of the input with regex anchors added to the values.
# Purpose:      Surround all values of the given hash with regex anchors, "^"
#               and "$", so the hash may be used to search an APISet response
#               object.
################################################################################
sub _apiset_resp_search_exact(%) {
    my %args = @_;
    return map {$_ => "^".$args{$_}."\$"} keys(%args);
}
