# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Tapeinfo.pm#1 $

## @summary Library for tape related information
##
## @description
## This library provides routines to get information about tape drives
## and medium changers.
##
## @dependencies
## The functions depend on exact output format to parse and extract results.
## Even a minor change in the UI can result in  functions not working properly.
##
## @author dl-dpgqa-automation@netapp.com
## 
## @changes 29/12/2009 nraju: added 10mode support for all subroutines.

package Tapeinfo;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(getTapes getTapeStatus getStorageTapes getStorageAliases);

use Tharn;
use strict;

## @name ::getTapes
## @synopsis getTapes($cfiler)
## description: Gets a list of all logical tape devices as provided by
##  the sysconfig -t filer command.
## param: $cfiler Filer connection object  (for 7mode filer)
##                Server.pm server object  (for 10mode filer)
## returns: Array containing a list of all tape devices.
sub getTapes($) {
  my ($filer) = @_;
  my $text;
  if( $filer->isa('Server') ) {
      $text = $filer->system_node_run(node     => "local",
                                      command  => "sysconfig -t",
                                      -Timeout => 60);
  } else {
      (undef, undef, $text) = say($filer, "sysconfig -t", -timeout => 60);
  }
  # Return if list is empty
  return ()
    if ($text =~ /No tape drives present on system./i);

  chomp $text;
  my @tapes = ();
  my @lines = split(/\n/,$text);
  foreach my $line (@lines) {
    $line =~ s/[\r\n]$//;
    # Parse this:
    #     rst0l  -  rewind device,        format is: 85937 bpi 35 GB
    if($line =~ /\s*([\w\.]+)\s+-\s+.+?format is:.+$/) {
      push(@tapes, $1);
    }
  }
  return @tapes;
}

## @name ::getTapeStatus
## @synopsis getTapeStatus($cfiler, $tape)
## description: Gets status of specified tape device as provided by
##  the mt filer command.
## param: $cfiler Filer connection object  (for 7mode filer)
##                Server.pm server object  (for 10mode filer)
## param: $tape Tape device whose status is returned
## returns: Hash with values: (status, fileno, blockno, resid, format)
## The status field contains 'unknown' in case of error.
## Otherwise that field, alongwith others, contains values provided by
## the "mt -f device status" filer command.
## If the tape device is reserved, only the status field is set.
sub getTapeStatus($$) {
  my ($filer, $tape) = @_;
  my $text;
  if( $filer->isa('Server') ) {
      $text = $filer->system_node_run(node     => "local",
                                      command  => "mt -f $tape status",
                                      -Timeout => 15);
  } else {
      (undef, undef, $text) = say($filer, "mt -f $tape status", -timeout => 15);
  }
  my %ret = (status => 'unknown');
  return (%ret) if($text =~ /is not a valid device name/i);
  if($text =~ /is reserved by another host/) {
    $ret{status} = "reserved";
    return (%ret);
  }

  $text =~ s/\r//g;

  # parse this:
  #Tape drive: Netapp Netapp Tape
  #  Status: ready, write enabled
  #  Format: VT-100MB (100 MB)
  #  fileno = -1  blockno = -1  resid = 0
  if($text =~/Status: (.+)$/m) {
    $ret{status} = $1;
  } else {
    return (%ret);
  }
  if($text =~/Format: (.+)$/m) {
      $ret{format} = $1;
  }
  if($text =~/fileno = ([\d\-]+)\s+blockno = ([\d\-]+)\s+resid = ([\d\-]+)/m) {
    $ret{fileno} = $1;
    $ret{blockno} = $2;
    $ret{resid} = $3;
  }
  return (%ret);
}

## @name ::getStorageTapes
## @synopsis getStorageTapes($cfiler)
## description: Gets list of physical tape names as provided by the
##  "storage show tape" filer command.
## param: $cfiler Filer connection object  (for 7mode filer)
##                Server.pm server object  (for 10mode filer)
## returns: A list of references to hashes. Each hash has following keys:
## (drive, desc, wwn, alias, state) which correspond to fields of
## the "storage show tape" command.
## The alias field contains only the first alias.
sub getStorageTapes($) {
  my ($cfiler) = @_;
  my $text;
  if( $cfiler->isa('Server') ) {
      $text = $cfiler->system_node_run(node     => "local",
                                       command  => "storage show tape",
                                       -Timeout => 60);
  } else {
      (undef, undef, $text) = say($cfiler, "storage show tape", -timeout => 60);
  }
  $text =~ s/\r//g;
  $text =~ s/^\s*\n//; # remove initial newline
  chomp $text;

# Parse:
#tagore> storage show tape

#Tape Drive:         0b.1
#Description:        Quantum DLT8000
#Serial Number:      CXA45P1500
#World Wide Name:    
#Alias Name(s):      st0  
#Device State:       available

#Tape Drive:         0b.2
#Description:        Quantum DLT8000
#Serial Number:      CXA45P1488
#World Wide Name:    
#Alias Name(s):      st1  st2  
#Device State:       available

  my @ret = ();
  my @tlist = split(/^\s*$/m, $text);
  foreach my $tapespec (@tlist) {
    my %tdesc;
    if($tapespec =~ /^Tape Drive:\s+(.+)$/m) {
      $tdesc{drive} = $1;
    } else {
       # Skip until valid entry
       next;
    }
    if($tapespec =~ /^Description:\s+(\S.+)$/m) {
      $tdesc{desc} = $1;
    }
    if(($tapespec =~ /(World Wide Name|WWNN):\s+(\S+)\s*$/m)) {
      $tdesc{wwn} = $2;
    }
    if($tapespec =~ /Alias Name\(s\):\s+(\w+)/m) {
      $tdesc{alias} = $1;
    }

    if($tapespec =~ /Device State:\s+(\w+)\s*$/m) {
      $tdesc{state} = $1;
    }
    push (@ret, \%tdesc);
  }
  return @ret;
}

## @name ::getStorageAliases
## @synopsis getStorageAliases($cfiler)
## description: Gets list of aliases as given by the
##  "storage alias" filer command.
## param: $cfiler Filer connection object  (for 7mode filer)
##                Server.pm server object  (for 10mode filer)
## returns: A hash with key value pairs corresponding to alias and real names.
sub getStorageAliases($) {
  my ($cfiler) = @_;
  my $text;
  if( $cfiler->isa('Server') ) {
      $text = $cfiler->system_node_run(node     => "local",
                                       command  => "storage alias",
                                       -Timeout => 60);
  } else {
      (undef, undef, $text) = say($cfiler, "storage alias", -timeout => 60);
  }
  $text =~ s/\r//g; # remove junk
  $text =~ s/^Alias\s+Mapping\s*$//m; # remove headers
  $text =~ s/^-+\s+-+\s*$//m; # remove headers
  $text =~ s/\s*$//m; # remove trailing blanks
  my %ret;
  my @lines = split(/\n/, $text);
  foreach my $line (@lines) {

    my($alias, $real) = split(/\s+/, $line);
    if(defined $alias and defined $real) {
      $ret{$alias} = $real;
    }
  }
  return (%ret);
}
