# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/TestEnv/MinWin.pm#4 $
##
## @summary Functions to interact with a minwin-compatible test client machine.
##
## @description
##	This library interacts with "MinWin" clients.  It will map the X: if necessary
##  and report back if the clients are in a good "MinWin" state.
##
## @synopsis
##	use minwin;
##
##	minwin_compatible($CLIENT);
##
## @status	public
## @author	tjt@netapp.com, schirato@netapp.com, kenny@netapp.com
##
## @notes
## This library is meant to be ran from a test controller and always invoked
## at a target CLIENT.  It is never meant to be ran from a local machine targetting
## itself as the CLIENT.
## 
## Future work:
## 1) Port to $Log for logging.  
##    This library should be updated to use $Log functionality in NATE rather than 
##    decrecated the older log*() format that is is currently using.
## 2) Add additional tracing and debug to support the library.

# 
# IF YOU ARE MAKING CHANGES TO THIS LIBRARY, PLEASE READ THIS:
#
# This library is a core library for all things NATE and Windows.  If you make
# a mistake and submit something bad to this library you are likley to cause a
# lot of different groups a lot of down time.  Be sure to follow proper procedure
# and have your change reviewed by multiple people and include the following in that
# list of reviewers:
#      Dan Schirato      schirato@netapp.com
#      Michael Latif     latif@netapp.com
#      Preston Lilley    preston@netapp.com
#      Thomas Masterson  thomasm@netapp.com
# 
# If you do not include these people in your review then you are not permitted to submit
# your changes.  If you want to add yourself to the above list, feel free, but you are
# required to review change requests as soon as possible.
#
# This package was originally located in //depot/prod/ontap/main/test/lib/TestEnv/MinWin.pm
# and was moved into the agnostic test tree to its current home later.  We had to rebase the
# source for the library in the agnostic tree at change #13 -  for the agnostic version of the
# library.  Unfortunately, "p4 filelog" doesn't show this easily but "p4 integ" does:
#     //depot/prod/test/main/lib/TestEnv/MinWin.pm#13 - integrate from //depot/prod/ontap/main/test/lib/TestEnv/MinWin.pm#30 using base //depot/prod/ontap/main/test/lib/TestEnv/MinWin.pm#29

package TestEnv::MinWin;

use strict;
use Tharn;
use Socket;
use Sys::Hostname;
use Module::Common;

BEGIN
{
    Module::Common::setup_export_tags(
        'basic' => [ qw(
        minwin_compatible
        map_drive
        unmap_drive
        execute_cmd
        execute_cmd_special
        minwin_xshare
        ) ]
    );
}

my $cifs_timeout = $main::TIMEOUT || 30;

#----------------------------
#   functions
#---------------------------

## @name ::minwin_xshare;
## @usage minwin_xshare()
## @synopsis
##    Returns the UNC name for the X: drive of minwin clients.
##    The value returned depends on the value of the SITE param,
##    or the value of the default_ip of the client referenced in
##    the connection spec that is passed in.
sub minwin_xshare {
    # TODO validate this is a conspec that we got in with Params::Validate
    my ($winconnspec) = @_;
    my $site = param("SITE", "-default", undef);
    my $x_unc = param("MINWIN_UNC", "-default", undef);
    my $x_user = param("MINWIN_USER", "-default", undef);
    my $x_password = param("MINWIN_PASSWORD", "-default", undef);

    # This hash which is keyed by the value we expect to see for SITE, contains
    # the information that is required to map the correct "local" version of X:
    my %minwin_info = (
        "nane"  => {    # Hosts in Boston will use these values
            x_unc       => "\\\\naninf02.nane.netapp.com\\localtest\\arch\\win32",
            x_user      => "naninf02\\minwin",
            x_password  => "qa4testing",
            networks    => ["10\\.97\\."],
        },
        "rtp"   => {    # Hosts in RTP (including KiloClient, rtpeng, stresScale, and ssqa) 
                        # will use these values.
                        # if the KiloClient gets its own mirror of this we could split
                        # it out of here, but today "kc" just maps to "rtp"
            x_unc       => "\\\\rtpinf02.rtp.netapp.com\\localtest\\arch\\win32",
            x_user      => "rtpinf02\\minwin",
            x_password  => "qa4testing",
            networks    => [
                "10\\.60\\.",  # Kilo - AGS - Lab1 - Lab2
                "10\\.62\\.",  # Lab 4
                "10\\.63\\.",  # VSIM
                "10\\.22[45678]\\.", # GDL, Lab1, Lab2
                "10\\.229\\.", # STI
                "10\\.230\\.", # STI expansion
                ],
        },
        "ptc"   => {    # Hosts in Pittsburgh will use these values
            x_unc       => "\\\\ptcinf02.spin.eng.netapp.com\\localtest\\arch\\win32",
            x_user      => "ptcinf02\\minwin",
            x_password  => "qa4testing",
            networks    => ["10\\.98\\."],
        },
        "smoke_bubble"   => {    # Hosts in any smoke bubble will use these values
            x_unc       => "\\\\smkmirror.mgmt.smoke\\localtest\\arch\\win32",
            x_user      => "win.smoke\\smoke",
            x_password  => "sm0ke",
            networks    => ["172\\.31\\."],
        },
        "btc"   => {    # Hosts in Bangalor will use these values
            x_unc       => "\\\\btcinf02.eng.btc.netapp.in\\localtest\\arch\\win32",
            x_user      => "btcinf02\\minwin",
            x_password  => "qa4testing",
            networks    => ["10\\.72\\."],
        },
        "svl"   => {    # Hosts in Sunnyvale will use these values
            x_unc       => "\\\\svlinf06.eng.netapp.com\\localtest\\arch\\win32",
            x_user      => "svlinf06\\minwin",
            x_password  => "qa4testing",
            networks    => [".+"], # any network that isn't matched is assumed to be SVL
        },
    );

    # map "kc", "rtpeng", and "stresScale" sites to "rtp" since we want to use 
    # the same values for those sites
    $minwin_info{kc}         = $minwin_info{rtp};
    $minwin_info{rtp_cqa}    = $minwin_info{rtp};
    $minwin_info{rtpeng}     = $minwin_info{rtp};
    $minwin_info{sti}        = $minwin_info{rtp};
    $minwin_info{ssqa}       = $minwin_info{rtp};
    $minwin_info{stresScale} = $minwin_info{rtp};

    # site gdl has the same data as site rtp
    $minwin_info{gdl} = $minwin_info{rtp};

    # if we don't have a site, or the site isn't set correctly, otherwise we can just
    # use the values defined in the hash for this site
    if ( !defined $site || !defined $minwin_info{$site} ) {
        # specifically warn if the site is bad rather than not there.
        if ( defined $site && !defined $minwin_info{$site} ) {
            logresult('WARN', "SITE is set to '$site' which isn't known to this library.  Please update " . caller() . " to include support for SITE '$site'");
            $site = undef;
        }
        # fall back to the old method of using the client's default_ip to decide the site
        my $client_ip = $winconnspec->{hostp}->{default_ip};
        # this is kind of tricky because we want to look at "svl" always last, so we need
        # to pull "svl" out of the list, check them, then toss "svl" on the end.
        SITE_KEY: foreach my $site_key ( grep(!/svl/, keys %minwin_info), "svl" ) {
            #print "Checking site '$site_key'\n";
            foreach my $site_network ( @{$minwin_info{$site_key}{networks}} ) {
                #print "Checking '$client_ip' against '$site_network'\n";
                if ( $client_ip =~ /$site_network/ ) {

                    $site = $site_key;
                    last SITE_KEY;
                }
            }
        }
        if ( !defined $site ) {
            logresult('FAIL', "Unable to match up a site for this client");
            return undef;
        } else {
            logcomment("Using the value derived from the clients network '$site' to assign path/user/password for X:");
        }
    } else {
        logcomment("Using the value specified in SITE '$site' to assign that path/user/password for X:");
    }

    my $site_ref = $minwin_info{$site};
    # override any of the UNC/USER/PASSWORD values we found for the site with the values found in
    # the params that we accept in from NATE.  Actually we're just telling the user that is what is
    # happening.
    if ( defined $x_unc ) {
        logcomment("Using value from MINWIN_UNC '$x_unc' as the UNC path for X: rather than the value defined for the site.");
    } else {
        $x_unc = $site_ref->{x_unc};
    }
    if ( defined $x_user ) {
        logcomment("Using value from MINWIN_USER '$x_user' as the user for X: rather than the value defined for the site.");
    } else {
        $x_user = $site_ref->{x_user};
    }
    if ( defined $x_password ) {
        logcomment("Using value from MINWIN_PASSWORD '$x_password' as the password for X: rather than the value defined for the site.");
    } else {
        $x_password = $site_ref->{x_password};
    }

    return($x_unc, $x_user, $x_password);
}

## @name ::minwin_compatible;
## @usage minwin_compatible($clientName)
## @synopsis
##    Returns 1 if the named windows client machine is minwin-compatible,
##    which means that it has an rshd server running, and now has
##    the X: share mounted containing various test utilities.
##    The named Windows client must be a name that can either matched to a
##    NATE_HOSTS entry, or can be resolved to an IP address via DNS.  The
##    latter of those two is very undesirable because it can cause you to 
##    map X: to the incorrect share.
sub minwin_compatible {

    my($winclient) = shift;
    my($persistent_map) = shift;

    if ( ! defined($winclient) ) {
        logcomment("minwin_compatible warning: Undefined value for winclient argument");
        return undef;
    } elsif ( $winclient =~ /localhost/ ) {
        # this is really a gross hack.  we don't have a chance in hell of determining the proper site
        # for a client called "localhost" unless SITE is defined.  but even on top of that there seems
        # to be a buffering problem with calling rsh on localhost under win32 (both inside of NATE and 
        # with rsh.exe directly) which causes output from the commands to be chopped up sometimes.  So 
        # after a grace period where we just issue a WARN on this, we will convert this to a SCRIPT error
        # because no one should be calling minwin_compatible with "localhost" as the client.
        logresult("WARN", "Calling minwin_compatible() with 'localhost' is not permitted.\nPlease update your script to pass in a real client name to this call.  On April 28, 2008 this warning message will be changed to a SCRIPT error.\nAttempting to determine the local hosts realname by using Sys::Hostname::hostname() as a temporary work around.");
        eval {
            $winclient = Sys::Hostname::hostname();
        };
        if ( $@ ) {
            logresult("FAIL", abort => 0, "Unable to determine the local client's host name. Cannot safely proceed.'$@'");
            return undef;
        }
    }

    # we specifically want to use RSH to connect to this client, and since it's possible we don't
    # have a NATE_HOSTS entry for it (I wonder if that is really true today) we have to specify the
    # clientspec for it in detail.
    my($spec) = $winclient;
    if ( $winclient !~ /;hosttype=windows/ ) {
        $spec .= ";hosttype=windows";
    }

    my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

    # Here's where all the magic information about the X drive is derived
    my $xdrive = "X";
    my($output);
    my($net_use_status);
    my($xshare, $xuser, $xpassword) = minwin_xshare($c);
    $xuser = "netapp\\ncqa" if (!defined($xuser));
    $xpassword = "netcache" if (!defined($xpassword));

    if ( $xshare eq "" ) {
        my $err = "minwin_compatible() doesn't know what "
        ."share to use for the $xdrive: drive, in your network";
        logcomment("minwin_compatible warning: $err");
        return undef;
    }

    my($num) = rand(999);	# We just need a unique string of some sort
    (undef, undef, $output) = say($c, "cmd /c echo $num", "-timeout", $cifs_timeout, "-eval_error");
    if ($@ && $@ =~ /Err=TIMEOUT/i) {
        logcomment("minwin_compatible warning: TIMEOUT during echo");
    }

    if ( $output !~ /$num/ ) {
        logcomment("minwin_compatible warning: The client $winclient is "
            ."not responding to rsh, its rshd needs resetting.");
        return undef;
    }

    # This is an attempt to address the situation for burt 116148.
    # The basic problem is that attempting to do a net use for a device that 
    # is already configured, will cause a prompt that we cannot answer.
    # The previous approach was to delete any existing X: drive
    # This got us in trouble when more than one subtest was running
    # since the second one would delete the configuration 
    # out from under the first one.
    # The new approach is to query the existing definition using 'net use'
    (undef, undef, $net_use_status) = say($c, "cmd.exe /c net use && echo NET_USE_RETURNED_OK", "-eval_error", "-timeout", $cifs_timeout);
    if ($@ && $@ =~ /Err=TIMEOUT/i) {
        logcomment("minwin_compatible warning: "
            ."TIMEOUT during net use $xdrive:");
    }
    # Then, only if the existing connection is disconnected or Unavailable
    # we delete it.
    my $need_delete = 0;
    my $need_connect = 1;
    my $did_stale_map_reconnect = 0;
    my $xshare_quotemeta = quotemeta $xshare;
    if ( $net_use_status =~ /Unavailable\s+$xdrive:\s+|Disconnected\s+$xdrive:\s+/i ) {
        # The drive is configured, but it is presently disconnected
        $need_delete = 1;
        logcomment("minwin_compatible warning: "
            . "'$xdrive' is in a bad state and will be unmapped.");
    } elsif ( $net_use_status !~ /\s+$xdrive:\s+$xshare_quotemeta/i ) {

        #Try to get the wrong mapping to show the user
        $net_use_status =~ /\s+$xdrive:\s+(\S+)/i;

        $need_delete = 1;
        logcomment("minwin_compatible warning: "
            . "'$xdrive' is mapped to '$1' which doesn't match '$xshare'.\n");
    } else {
        # if the drive map actually does look good, don't make the call to mount it
        $need_connect = 2;
    }

    while ($need_delete or $need_connect) {
        # set persistent mapping. if persistent
        # mapping is force diconnect and reconnect persistenly
        if ($persistent_map =~ /yes/i) {
            $persistent_map = "YES";
        } elsif ($persistent_map =~ /force/i) {
            $persistent_map = "YES";
            $need_delete = 1;
        } else {
            $persistent_map = "NO";
        }

        if ($need_delete) {
            $need_delete = 0;
            (undef, undef, $output) = say($c, "net use $xdrive: /DELETE /y", "-timeout", $cifs_timeout, "-eval_error");
            if ($@ && $@ =~ /Err=TIMEOUT/i) {
                logcomment("minwin_compatible warning: "
                    ."TIMEOUT during net use xdrive /DELETE");
            }

            $need_connect = 1;   # Force a reconnect next
        }
        # Next, we (re)establish the net use
        # if the drive is not presently configured, or it is and we just want to check it
        if ($need_connect == 1) {
            $need_connect = 0;
            (undef, undef, $output) = say($c, "net use $xdrive: $xshare $xpassword /persistent:$persistent_map /USER:$xuser", "-timeout", $cifs_timeout, "-eval_error");
            if ($@ && $@ =~ /Err=TIMEOUT/i) {
                logcomment("minwin_compatible warning: "
                    ."TIMEOUT during net use xdrive");
            }

            # The output may or may not contain the "local device name is
            # already in use" error - oddly, it's inconsistent.
            # But this problem always contains the "error.*85", so we key off that.

            # The 1909 error is intermittent, with increasing likelihood on busy systems
            # However, 3 bad attempts may lock out users - don't loop on retry here
            if ( $output =~ /System.*error.*1909/ ) {
                # Let's try it without the defined password in case access is already allowed
                # (This should be available to real user ids)
                (undef, undef, $output) = say($c, "net use $xdrive: $xshare /persistent:$persistent_map", "-timeout", $cifs_timeout, "-eval_error");
                if ($@ && $@ =~ /Err=TIMEOUT/i) {
                    logcomment("minwin_compatible warning: "
                        ."TIMEOUT during net use xdrive");
                }
                if ($output =~ /System.*error.*1909/) {
                    logcomment("minwin_compatible warning: "
                        ."Unable to map drive=$xdrive to "
                        ." share=$xshare : $output");
                    return undef;
                }
            }
            if ( $output =~ /successfully/ ) {
                # Verify the new mapping
                (undef, undef, $net_use_status) = say($c, "net use","-timeout", $cifs_timeout, "-eval_error");
                # fall through
            } elsif ( $output =~ /System.*error.*85/ ) {
                # Somehow, this drive became mapped since we looked a minute ago!
                # Let's verify it now has the right mapping
                (undef, undef, $net_use_status) = say($c, "net use","-timeout", $cifs_timeout, "-eval_error");
                if ($@ && $@ =~ /Err=TIMEOUT/i) {
                    logcomment("minwin_compatible warning: "
                        ."TIMEOUT during net use $xdrive:");
                }

                if ( $net_use_status !~ /\s+$xdrive:\s+\S+/ ) {
                    # It went away? Try to map the drive again
                    (undef, undef, $output) = say($c, "net use $xdrive: $xshare $xpassword /USER:$xuser", "-timeout", $cifs_timeout, "-eval_error");
                    if ($@ && $@ =~ /Err=TIMEOUT/i) {
                        logcomment("minwin_compatible warning: "
                            ."TIMEOUT during net use $xdrive $xshare");
                    }
                    if ( $output !~ /successfully/ ) {
                        logcomment("minwin_compatible warning: "
                            ."Unable to map $xdrive "
                            ."share: $output");
                        return undef;
                    }
                    # Verify the new mapping
                    (undef, undef, $net_use_status) = say($c, "net use","-timeout", $cifs_timeout, "-eval_error");
                }


            } elsif ( $output =~ /System.*error.*53/ ) {
                logcomment("minwin_compatible warning: "
                    ."Unable to map drive=$xdrive to "
                    ." share=$xshare : $output");
                return undef;
            } else {
                logcomment("minwin_compatible warning: "
                    ."Unexpected output when mapping "
                    ."$xdrive share: $output");
                return undef;
            }
        } 
        # if the status is OK or simply blank 
        if ( $net_use_status =~ /(:?OK\s+|\s+)$xdrive:\s+(\S+\\localtest\\arch\\win32)/m ) {
            my $present_map = $1;
            # We have verified that the directory has really been mapped
            # Let's make sure we can find some expected directories there
            # (There have been times where they do not appear!)
            (undef, undef, $output) = say($c, "cmd /c dir /on $xdrive:\\", "-timeout", $cifs_timeout, "-eval_error");
            if ($@ && $@ =~ /Err=TIMEOUT/i && (!defined($output) or ($output !~ /\s+bin\s+/ and $output !~ /\s+nate\s+/ and $output !~ /\s+perl\s+/)) ) {
                logcomment("minwin_compatible warning: TIMEOUT during dir");
            }
            logcomment("Windows client $winclient has been declared "
                ."minwin-compatible and the $xdrive: drive is mapped to $present_map");
            if ( $output =~ /The specified network name is no longer available/i ) {
                logcomment("minwin_compatible warning: "
                    . "Drive '$xdrive' was mapped and listed as OK, but really isn't.  Forcing a Delete and remap.\n");
                $need_delete = 1;
                $need_connect = 1;
                next; # make sure we don't return yet
            } elsif ( $output !~ /\s+bin\s+/ and $output !~ /\s+nate\s+/ and $output !~ /\s+perl\s+/) {
                logcomment("minwin_compatible warning: "
                    . "Drive $xdrive was mapped but it does not have the expected directories"
                );
                return undef;
            } else {
                (undef, undef, $output) = say($c, "cmd /c dir /on $xdrive:\\bin\\sio.exe", "-timeout", $cifs_timeout, "-eval_error");
                if ($@ && $@ =~ /Err=TIMEOUT/i && (!defined($output) or ($output !~ /\s+\d+\s+sio.exe/i )) ) {
                    logcomment("minwin_compatible warning: TIMEOUT during dir");
                } elsif($output =~ /\s+(\d+)\s+sio.exe/i ) {
                    if ($1 eq "0") {
                        # We have a stale copy. Drop the map and try again, but only once
                        if (!$did_stale_map_reconnect) {
                            $did_stale_map_reconnect = 1;
                            $need_connect = 1;
                        }
                    }
                }
            }
            return(1);
        } elsif ( $net_use_status !~ /\s+$xdrive:\s+\S+/ ) {
            # Second attempt to map it also failed, perhaps there is a script error here?
            logcomment("minwin_compatible warning: "
                ."Drive $xdrive could not be mapped: $net_use_status"
            );
            return undef;
        } else {
            # It is mapped somewhere that is not compatible
            logcomment("minwin_compatible warning: "
                ."Drive $xdrive is already mapped somewhere that is not compatible: $net_use_status"
            );
            return undef;
        }
    }
}

## @name ::map_drive;
## @usage map_drive($clientName,$driveLetter,$sharePath, %opts)
## @synopsis
##    Map the named share to the named drive on the named windows client.
## @param clientName required
## @param driveLetter required
## @param sharePath required
## @param opts optional
##	user and passwd credentials used to map drive maybe specified
##	user => "myuser", passwd => "mypasswd"
##	mapped drive can be PERSISTENT, defaults to NO
##	persistent => "YES" or "NO"
sub map_drive {

    my($winclient, $drive, $share, %opts) = @_;

    my($spec) = $winclient;
    $spec .= ";hosttype=windows"
    unless ($winclient =~ /;hosttype=windows/);
    my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

    my($output);
    (undef, undef, $output) = say($c, "cmd /c net use", "-eval_error");
    if ($@ && $@ =~ /Err=TIMEOUT/i) {
        logcomment("map_drive error: "
            ."TIMEOUT during net use in map_drive");
    }
    logcomment("currently mapped drives: $output");


    if ( $drive !~ /^[*A-Za-z]:?$/ ) {
        logresult('WARN',"Bad drive value ($drive) given to map_drive()");
        return undef;
    }

    # Add ":" if it's not there.
    if ( $drive =~ /^[A-Za-z]$/ ) {
        $drive = $drive.":";
    }

    my $cmd = "cmd /c net use $drive $share ";
    $cmd .= $opts{passwd} . " /USER:" . $opts{user}
    if ($opts{passwd} && $opts{user});

    $opts{persistent} = $opts{persistent} =~ /YES/i ? "YES" : "NO";
    $cmd .= " /PERSISTENT:".$opts{persistent};

    (undef, undef, $output) = say($c, $cmd, "-eval_error");
    if ($@ && $@ =~ /Err=TIMEOUT/i) {
        logcomment("map_drive error: "
            ."TIMEOUT during net use drive in map_drive");
    }
    logresult('WARN', $output)
    if ($output =~ /err|invalid/i);

    # This is silly, but works.
    $output =~ s/\r//g;
    $output =~ s/\n//g;

    # Different output from different versions of Windows.

    if ( $output !~ /successfully/ && $output !~ /now connected/ ) {
        logresult('WARN',"Unable to map $drive from $share : $output");
        return undef;
    } elsif ( $output =~ /now connected/ ) {
        if ( "$drive" eq "*" ) {
            # If we've let the windows client pick the drive
            # letter, we need to figure out which one was selected.

            # This is silly, but works.
            $output =~ s/.*Drive //;
            $output =~ s/:.*//;
            $drive = $output;
        }
    } else {
        if ( "$drive" eq "*" ) {
            # If we've let the windows client pick the drive
            # letter, we need to figure out which one was selected.
            $output =~ /Drive ([A-Z]):.*/;
            $drive = $output;
        }
    }
    $drive =~ s/://;
    logcomment("minwin_map has mapped $share as drive $drive");
    return($drive);
}

## @name ::unmap_drive;
## @usage unmap_drive($clientName,$driveLetter)
## @synopsis
##    Unmap the named drive from the named windows client.
sub unmap_drive {

    my($winclient) = shift;
    my($drive) = shift;

    if ( $drive !~ /^[A-Za-z]:?$/ ) {
        logresult('WARN',"Bad drive value ($drive) given to unmap_drive()");
        return undef;
    }

    my($spec) = $winclient;
    $spec .= ";hosttype=windows"
    unless ($winclient =~ /;hosttype=windows/);
    my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

    # Add ":" if it's not there.
    if ( $drive =~ /^[A-Za-z]$/ ) {
        $drive = $drive.":";
    }

    my($output);
    (undef, undef, $output) = say($c, "cmd /c net use $drive /delete", "-eval_error");
    if ($@ && $@ =~ /Err=TIMEOUT/i) {
        logcomment("TIMEOUT during net use /delete in unmap_drive");
    }

    if ( $output !~ /successfully/ ) {
        logresult('WARN',"Unexpected output when unmapping "
            ."$drive: $output");
        return("");
    } else {
        logcomment("minwin_mount has unmapped drive $drive");
        return($drive);
    }
}

## @name ::execute_cmd;
## @usage execute_cmd($clientName,$cmd)
## @synopsis
##    Run a command on the named minwin-compatible client.
sub execute_cmd {

    my($winclient) = shift;
    my($cmd) = shift;
    my ($timeout) = shift;

	if(!$timeout) {
        $timeout = $cifs_timeout;
	}
	my($spec) = $winclient;
    $spec .= ";hosttype=windows"
    unless ($winclient =~ /;hosttype=windows/);
    my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

    my($output);

    # We use the runcmd.pl utility in order to make sure
    # that the stderr output of the command is completely captured.

    (undef, undef, $output) = say($c,
        "x:\\perl\\bin\\perl.exe x:\\bin\\runcmd.pl $cmd", "-timeout", $timeout,
        "-eval_error");
    if ($@ && $@ =~ /Err=TIMEOUT/i) {
        logcomment("TIMEOUT during command execution: $cmd");
    }
    return($output);
}

## @name ::execute_cmd_special;
## @usage execute_cmd_special($clientName,$cmd)
## @synopsis
##    Run a command on the named minwin-compatible client. It is identical to 
##    execute_cmd except that the output of the command(stdout, stderr) is NOT 
##    redirected to special files for printing. 
##    Commands like echo "some_text" > filename, do not have the desired
##    output with the execute_cmd. This is useful for testing appending, 
##    overwriting to files.
sub execute_cmd_special {

    my($winclient) = shift;
    my($cmd) = shift;

    my($spec) = $winclient;
    $spec .= ";hosttype=windows"
    unless ($winclient =~ /;hosttype=windows/);
    my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

    my($output);

    (undef, undef, $output) = say($c,
        "cmd.exe /c $cmd", "-timeout", $cifs_timeout);
    if ($@ && $@ =~ /Err=TIMEOUT/i) {
        logcomment("TIMEOUT during command execution: $cmd");
    }
    return($output);
}
1;
