#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/TestManager.pm#1 $

## @summary     Class to control test execution
## @author      snider@netapp.com
## @pod here

=head1 NAME

Options - Class definition for a Test Object.

=head1 SYNOPSIS

    package Test;
    use base qw(TestManager);

    sub test_main {
        my $self = shift;
        my ($failure) = $self->get(parameter=>"INDUCE_FAILURE");
        # some test code here to implement the test case
        die "This test failed\n" if ($failure);
    }

    my $test = Test->new();

    $test->parameters(definitions=>
        [
         {
             name => "INDUCE_FAILURE",
             description => "force a failure in test_main",
             default => undef,
         }
        ]);
    $test->run_test(timeout=>600); # 10 minutes

=head1 DESCRIPTION

C<TestManager> is a base class for tests. It provides many of the base
functions needed by a test. More detailed documentation will be
available at http://wikid.netapp.com/w/Spin/QA/projects/TestManager.

C<TestManager> uses its own copy of the L<Options> and L<TracePkg> objects
and provides abstracted methods into L<Options>.

=head1 METHODS

=cut

#***********************************************************************
# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDETESTMANAGER}) {
    my $newModule = $ENV{OVERRIDETESTMANAGER};
    delete $ENV{OVERRIDETESTMANAGER};
    print "The main TestManager module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package TestManager;
use strict;
use English;
use Options qw(-Delayed_help); # bring this in first to force the Delayed_help action
use Params::Common;
use Params::NFS;
use Params::NS;
use Params::Volume;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Data::Dumper qw(Dumper);
use Tharn;
use NATE::ParamSet;
use mount qw(mount umount);
use MapShare qw(:all);
use SayWrap qw(sayEOF);
use Utils qw(make_conn);
use TestUtils qw(iterate expected_run_time);
use Sys::Hostname qw(hostname);
use Cwd qw(realpath);
use JobQueue;
use TracePkg;
use Network::Trace;
use Network::NodeTrace;
use cluster::enterprise::lib::TestUsers qw(get_users defined_lab);
my $Trace = new TracePkg;

use Server;
my $Server;

my $Options;
my $Options_hash;

use RemoteExecution;
my $Remote;
my $Remote_Ancillary; # Have to be on diff object, as it needs stopping externally.

# This hash will be used to cache the volume/share mapping
my %Shares;

use constant PER_INSTANCE => "per_instance";
use constant SINGLE => "single";
use constant TEST_DEFINED => "test_defined";

$Options = new Options;
$Options->parameters(definitions=>
          [
              {
                  name => "CACHE_TRAFFIC",
                  description => "Should traffic be routed via the FlexCache?",
                  spec => "boolean",
              },
              {
                  name=>"RO_PATH",
                  description=>"use the .admin path in the mounts",
                  spec=>"boolean",
                  default=>"false",
              },
              {
                  name=>"SKIP_192_ADDRESSES",
                  description=>"skip apparently-private 192.16 data interfaces",
                  spec=>"boolean",
                  default=>"false",
              },
              {
                  name=>"STAGE",
                  spec=>"s@",
                  description=>"change the default test stages to run",
              },
              {
                  name=>"SKIPSTAGE",
                  spec=>"s@",
                  description=>"skip these test stages",
              },
              {
                  name=>"TEST_MODE", # look at the comment before the new() regarding this
                  default=>"scalable",
                  description=>"dictates the execution of the test_main",
              },
              {
                  name=>"TEST_TIMEOUT",
                  spec=>"time",
                  description=>"how long before the test should timeout, in seconds",
              },
              {
                  name=>"VOLUME_LAYOUT",
                  description=>"how should new volumes be laid out",
                  valid_values => [PER_INSTANCE(), SINGLE(), TEST_DEFINED()],
              },
              {
                  name=>"TEST_BLOCK_LAYOUT",
                  description=>"how should test_setup/cleanup be executed",
                  valid_values => [PER_INSTANCE(), SINGLE()],
              },
              {
                  name => "testdir",
                  description => "The directory to run the test against",
              },
              {
                  name => "MOUNTPOINT_UNIQUIFIER",
                  description => "A string to embed in the auto-generated mountpoints. " .
                    "It can help distinguish multiple instances of the same test on the same client.",
              },
              {
                  name => "FILER_SETUP_POSTCHECK",
                  description => "List of thpls to be run at end of filer_setup",
                  spec => "s@",
              },
              {
                  name => "FILER_CLEANUP_PRECHECK",
                  description => "List of thpls to be run at beginning of filer_cleanup",
                  spec => "s@",
              },
              {
                  name => "CLIENT_INSTANCE",
                  description => "The instance of the client",
                  default => 0,
              },
              {
                  name => "VOLUME_OWNER",
                  description => "Sets the owner of a volume.  Can be \"self\" or a user name.",
              },
              {
                  name => "MOUNT_OPTIONS",
                  description => "The options to pass to mount",
              },
              {
                  name => "MNTPNT_MODE",
                  description => "The mode to set the client mountpoints to.",
                  default => "777",
              },
              {
                  name => "JUNCTION_PATH",
                  description => "The junction-path component of a volume",
              },
              {
                  name => "AUTOMOUNT",
                  spec => "boolean",
                  description => "Flag indicating that mounts should act like an automount",
                  default => "false",
              },
              # These are used when scalabled mode calls classic mode
              {
                  name => "VOLUME_SHARE",
                  spec => "s@",
                  description => "a comma-separated list of volume:share mappings;INTERNAL ONLY",
              },
              {
                  name => "MOUNT_LIST",
                  description => "list of mountpoints to create; INTERNAL ONLY",
              },
              {
                  name => "CHMOD_LIST",
                  description => "list of mountpoints to chmod; INTERNAL ONLY",
              },
              {
                  name => "NEED_NIS",
                  description => "Flag indicating this test requires NIS information to be filled in",
                  default => 0,
                  spec => "boolean",
              },
              {
                  name => "RUN_COMMAND_TIMEOUT",
                  spec => "time",
                  description => "how long before test run command should timeout, in seconds",
                  default => undef,
              },
              {
                  name => "RAND_MULTI_USER",
                  description => "Flag indicating this test should use test usernames to run commands",
                  default => 0,
                  spec => "boolean",
              },
              {
		  name => "CIFS_VSERVERS_ONLY",
                  description => "Only use vservers that have CIFS shares on them",
		  default => 0,
		  spec => "boolean",
  	      },
          ]
);


my $Full_testname = $Tharn::Scriptspec;
# this should only happen if we are outside of normal NATE processing
if (!$Full_testname || $Full_testname eq "") {
    # if we are running in natelint we don't want to blow up here
    if ( defined $INC{"NATE/Lint.pm"} ) {
        $Full_testname = "natelint";
        $Trace->warn( msg => "We seem to be running inside of natelint and will " .
              "continue execution with a fake testname. If you are seeing this " .
              "warning and are not executing natelint please alert the owner of " .
              "TestManager. Full_testname = '$Full_testname'"
        );
    } else {
        $Trace->fatal(msg => "Could not determine the TESTNAME", result=>"CONFIG");
    }
}
my $Testname = $Full_testname;
$Testname =~ s/.+\///; # just get the name of the test
$Testname =~ s/\.thpl//; # strip off the .thpl extension
my $Ruleset = "p_$Testname";

my $Discovered_topology;
my $Test_topology;
my @Aggregates;
my @Data_interfaces;
my $Virtual_server;
my $Cifs_Domain;
my $Cifs_Server;
my $Cifs_name_key = 'cifs-server'; #Setting default hash key name
my $Cifs_Share;
my $Kerberos_Realm;
my $Nis;
my $Self;

# Define the basic stages here. The %Stages hash can be changed, this one is not.
my %Stages_default = (
		      filer_setup => 1,
		      client_setup => 1,
		      test_setup => 1,
		      test_main => 1,
		      test_cleanup => 1,
		      client_cleanup => 1,
		      filer_cleanup  => 1,
		      );
# This hash will be used to manage the stages to run through. If a STAGE param is passed,
#  then the contents of this hash are replaced with just the specified stages.
my %Stages = %Stages_default;

=head2 new

    package Test;
    use base qw(TestManager);
    my $test = Test->new();

C<new> is the invocant for the Test object. The Test object should not
override this one; if it does, then it should call it as a SUPER first.
C<new> will first call process on L<Options> and then check for the value
of the C<TEST_MODE> param and store it for later useage. It will also define
references to the following internal structures so that the underlying Test
object can access them directly

    data_interfaces - an array reference of the interfaces that will be used
         aggregates - an array reference of aggregate hashes that will be used
     virtual_server - a hash reference containing information on the virtual
                      server that will be used
discovered_topology - a reference to the Topology hash which is discovered
                      on the filers. See L<Server> for more information.
      test_topology - a reference to the Topology hash as defined by the Test

=cut

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {
                cifs => undef,
                need_nis => undef,
                @_,
            };
    $Trace->entry();
    bless($self, $class);

    # Determine the test_mode at startup, prior to parameters() being called
    # The test_mode is based on an input parameter so call process() and then
    # check that option.
    # Since we are in the new(), test_cases have not yet been defined so
    # process() will die when it sees a TESTCASE; turn that check off
    $Options->process(test_cases=>undef);
    $self->{test_mode} = ($Options->get(option=>"TEST_MODE") =~ /classic/i) ? "classic" : "scalable";

    $Self = $self; # <- globalizing $self is horrible. We should stop doing this.
    $self->{data_interfaces} = \@Data_interfaces;
    $self->{aggregates} = \@Aggregates;
    $self->{discovered_topology} = undef;
    $self->{test_topology} = undef;
    $self->{testdir} = undef;
    $self->{testname} = $Testname;
    my @path = split(m!/!, $Tharn::Scriptspec);
    pop @path;
    $self->{test_location} = join("/", @path);
    $self->{_ancillary_tests} = {};
    $self->{ustr} = ""; # this will be filled in during run_test()

    if ($self->{test_mode} eq "scalable") {
        $self->{user} = getpwuid($<);
        $self->{euser} = getpwuid($>);
    }

    # As much state information as possible should be stored in our TestManager instance.
    # TM's reliance on globals should be reduced.

    # {mountpoint_uniquifier} is a string we embed in our auto-generated mount paths.
    $self->{mountpoint_uniquifier} = $Options->get(option=>"MOUNTPOINT_UNIQUIFIER");

    $Trace->exit(msg=>"return $self");
    return $self;
}

=head2 test_mode

    $mode = $self->test_mode();

C<test_mode> will return the current value of the C<TEST_MODE>.

=cut

sub test_mode {
    my $self = shift;
    return $self->{test_mode};
}

=head2 test_classic

  if ($self->test_mode() eq $self->test_classic()) {
    do something;
  }

C<test_classic> provides the current definition of the C<classic> C<TEST_MODE>

=cut

sub test_classic {
    return "classic";
}

=head2 test_scalable

  if ($self->test_mode() eq $self->test_scalable()) {
    do something;
  }

C<test_scalable> provides the current definition of the C<scalable> C<TEST_MODE>

=cut

sub test_scalable {
    return "scalable";
}

sub convert_share {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'TestManager' } ],
        {
            volume => {type => SCALAR},
            share=> {type => SCALAR},
            readonly => {type => SCALAR|UNDEF, default => $Options_hash->{RO_PATH}}
        }
    );

    if (defined($params{readonly}) && "$params{readonly}" =~ m/(1|yes|true)/i) {
        $params{readonly} = 1;
    } else {
        undef($params{readonly});
    }

    my $share;
    our %shares;
    $Trace->entry(msg=>"share=$params{share}, volume=>$params{volume}");

    # if this share is already cached, skip the check
    if (!exists($shares{$params{share}})) {
	# find the aggregate name for this volume. Classic needs this...
	my $vol_aggr;
	if ($self->{test_mode} eq "scalable") {
	    # we already have volume information in this mode so find the aggr for this volume from it
            foreach my $vol_hash (@{$Test_topology->{cluster}{volume}}) {
		my $cmp_vol;
		if ($vol_hash->{'volume-style'} eq "infinitevol") {
		    $cmp_vol=$vol_hash->{'volume'}."_ns";
		} else {
		    $cmp_vol = $params{volume};
		}
		if ($cmp_vol eq $vol_hash->{volume}) {
		    $vol_aggr = $vol_hash->{aggregate};
		    last;
		}
	    }
	} else {
	    # This won't work if there are multiple aggregates. But then again, this is only needed for qtrees
	    #  and the test can only work if the aggregate is specified.
            $vol_aggr = $self->get(parameter=>"aggregate");
	    $vol_aggr = $vol_aggr->[0] if (ref($vol_aggr) eq "ARRAY");
	}
	$Trace->fatal(result=>"SCRIPT", msg=>"Could not determine the aggregate name for $params{volume}") if (!$vol_aggr);

        # use the server object to actually convert a basic path into a platform-specific version
        # i.e. Tricky: /volume becomes /.admin/volume
        #     Classic: /volume becomes /vol/volume or maybe /vol/vol/qtree
        #     FlexCache: we need to add _fc to any volume name
        my $node;
        my $volume = $params{volume};
        # determine if we have flexcache nodes to deal with
        if ($self->{flexcaches_involved}) {
            my @legacy = $Server->get_legacy_nodes();
            $node = $legacy[0];
            $volume .= "_fc";
        } else {
            my @clustered = $Server->get_clustered_nodes();
            $node = $clustered[0];
        }
        ($shares{$params{share}}{rw},$shares{$params{share}}{ro}) = $Server->convert_export_path(-Server => $node,
                                                                                                 export => $params{share},
                                                                                                 aggregate => $vol_aggr,
                                                                                                 volume => $volume,
                                                                                                 readonly => $params{readonly});
    }
    $Trace->exit(msg=>"share=".($params{readonly} ? $shares{$params{share}}{ro} : $shares{$params{share}}{rw}));
    return ($params{readonly} ? $shares{$params{share}}{ro} : $shares{$params{share}}{rw});
}

sub find_volume_share {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'TestManager' } ],
        {
            volume => {type => ARRAYREF | SCALAR},
        }
    );

    $Trace->entry(msg=>"volume=$params{volume}");
    my $share;
    $params{volume} = $self->_normalize_volume_name(volume=>$params{volume});

    # if this share is already cached, then return that form
    if (!($share = $Shares{$params{volume}})) {
	# look for the volume/junction in the Topology since it may not be created yet
	my $volh = $self->search_topology(cmd_node=>"volume", key=>"volume", value=>$params{volume});
	if ($volh && $volh->{'junction-path'}) {
	    $share = $volh->{'junction-path'};
	} else {
	    # query the server object for the volume export path
	    $share = $Server->volume_mount_point(%params);
	}
	if ($self->{shares_available}) {
	    # this means that the %Shares hash has been setup so we can safely call convert_share now.
	    # If this is not set, then every client during a test could potentially start making CLI
	    #  calls to the filer to query for share information
	    $share = $self->convert_share(%params, share=>$share);
	}
    }
    $Trace->exit(msg=>"return $share");
    return $share;
}

=head2 parameters

    $test->parameters(definitions=>
        [
         {
             name => "INDUCE_FAILURE",
             description => "force a failure in test_main",
             default => undef,
         }
        ]);

C<parameters> provides a definition for a new parameter.

=over

=item definitions

An array of hashes which each define a unique parameter. Please see the
'definitions' method of the L<Options> library.

=back

=cut

sub parameters {
    my $self = shift;
    my %params = (@_);
    $Trace->entry();

    # pass the definitions to Options but only if the test_mode for the definition matches the test_mode
    #  that the TestManager is currently running. The default mode for the TM is "scalable" which causes
    #  the test to run through the test stages. The other mode for the TM is "classic" which runs just the
    #  test_main defined by the test.
    # The default value for the parameter test_mode is "scalable"
    # Just pass the params if not test_mode is specified.
    my @param_list;
    foreach my $param_hash (@{$params{definitions}}) {
	if (!defined($param_hash->{test_mode}) || ($self->{test_mode} =~ /$param_hash->{test_mode}/i)) {
	    push(@param_list, $param_hash);
	}
    }
    $Options->parameters(%params, definitions=>\@param_list);
    $Trace->exit(msg=>"");
}

# allow for a test to set a parameter
sub set {
    my $self = shift;
    my %params = (@_);

    $Trace->entry(msg=>"params: " . Dumper(\%params));
    $Options->set(%params);
    $Trace->exit();
}

# allow for a test to set a default
sub set_default {
    my $self = shift;
    my %params = (@_);

    $Trace->entry(msg=>"params: " . Dumper(\%params));
    $Options->set_default(%params);
    $Trace->exit();
}

sub get {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'TestManager' } ],
        {
            parameter => {type => ARRAYREF | SCALAR | UNDEF, optional => 1},
            testcase => {type => SCALAR | UNDEF, optional => 1},
        }
    );

    $Trace->entry(msg=>"parameter=>$params{parameter}, testcase=>$params{testcase}");
    if (wantarray) {
	my @options = $Options->get(option=>$params{parameter});
	$Trace->exit(msg=>"return @options");
	return @options;
    } else {
	my $option = $Options->get(option=>$params{parameter},
				    testcase=>$params{testcase});
	$Trace->exit(msg=>"return $option");
	return $option;
    }
}

sub get_hash {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'TestManager' } ],
        {
            parameter => {type => ARRAYREF | SCALAR, optional => 1},
        }
    );

    $Trace->entry();
    my $options = $Options->get_hash(option=>$params{parameter});
    $Trace->exit(msg=>"");
    return $options;
}

sub get_connection {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'TestManager' } ],
        {
            filer => {type => UNDEF | SCALAR, optional => 1},
        }
    );

    $Trace->entry();
    $Server = new Server if (!$Server);
    my $conn = $Server->get_connection(-Server=>$params{filer},
				       type=>"cli_connection");
    $Trace->exit(msg=>"return $conn");
    return $conn;
}

sub get_server {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'TestManager' } ],
        {
        }
    );

    $Trace->entry();
    $Server = new Server if (!$Server);
    $Trace->exit(msg=>"return $Server");
    return $Server;
}

sub test_cases {
    my $self = shift;
    $Trace->entry();
    $Options->test_cases(@_);
    $Trace->exit(msg=>"");
}


sub _start_ancillary_tests {
    my $self = shift;

    $Trace->entry(msg=>"");
    $Trace->debug(msg=>"Ancillary Tests: \n".Dumper($self->{_ancillary_tests}));
    my $options = $self->get_hash();

    foreach my $seq (sort keys %{$self->{_ancillary_tests}}) {
        my $testref = $self->{_ancillary_tests}{$seq};
        my %args;
        if($testref->{type} eq "coderef") {
            $args{nateeval} = 1;
        } else {
            $args{nateeval} = 0;
        }
        # Ancillary tests meant to run until real tests are done and then
        # parent can kill these. Hence set ITERATE to "infinity"
        if($testref->{nate_params}) {
            $args{nate_params} = {
                %$options,
                %{$testref->{nate_params}},
                DURATION => undef,
                ITERATE  => "infinity",
            };
        } else {
            $args{nate_params} = {
                %$options,
                DURATION => undef,
                ITERATE  => "infinity",
            };
        }

        $Trace->comment(msg=>"Starting Ancillary Test: $testref->{name}");
        $Remote_Ancillary->execute(host        => "localhost",
                                   cmd         => $testref->{cmd},
                                   runid       => $testref->{name},
                                   handler     => sub {},
                                   %args);
    }

    $Trace->exit(msg=>"");
    return;
}

=head2 add_ancillary_tests

    $testsref = [
                  {
                      name => "test1",
                      cmd  => $script_path,
                      type => "script",
                      sequence => 1,
                  },
                  {
                      name => "test2",
                      cmd  => "$code",
                      type => "coderef",
                      sequence => 2,
                  }
                 ]);
    $test->add_ancillary_tests(tests => $testsref);

C<add_ancillary_tests> is the the method of TestManager. This method
can be used by the test to add as many ancillary tests as needed
to run in parallel to test execution. These tests are started by
TestManager just before the real test is started and stoppped/killed
just after real test is completed.

=over

=item tests

The array ref of the hash entries that contains the test script path or
coderef and properties of the same.

=back

=cut

sub add_ancillary_tests {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'TestManager' } ],
        {
            tests => {type => ARRAYREF, optional => 0, default => []},
        }
    );
    $Trace->entry();

    foreach my $testref (@{$params{tests}}) {
        my $seq = $testref->{sequence};
        delete $testref->{sequence};
        $self->{_ancillary_tests}{$seq} = $testref;
    }

    $Trace->exit();
    return;
}

=head2 run_test

    $test->run_test(timeout=>600, # 10 minutes
                    volume_layout=>PER_INSTANCE(),
                    test_block_layout=>PER_INSTANCE(),
                    test_layout=>PER_INSTANCE(),
                    check_topology=>1);

C<run_test> is the main method of TestManager. This method controls all
execution of the test and will not return to the caller. More detailed
documentation will be available at
http://wikid.netapp.com/w/Spin/QA/projects/TestManager.

=over

=item timeout

C<timeout> specifies how long the test_main routine is expected to run. A
FATAL result will be returned if this time passes prior to the test_main
routine finishes. The time is in seconds.

=item volume_layout

C<volume_layout> specifies how the volumes should be layed out for this
particular test during C<filer_setup>. Please see C<VOLUME_LAYOUT> for more
information on the values and what they mean.

=item test_block_layout

C<test_block_layout> determines how and when the test_setup and test_cleanup
routines are called. The potential settings and behaviour are:

=over

=item PER_INSTANCE()

This will cause the test_setup to be called on each client, followed by
test_main and then test_cleanup. This entire block will be called during
each test iteration. This is the default.

=item SINGLE()

This will cause the test_setup to be called once on the local host prior to
remotely executing the test_main on each client. Once the clients have finished,
the test_cleanup is called once on the local host.

=back

=item test_layout

C<test_layout> determines how and when the test_main routine is called. The
potential settings and behaviour are:

=over

=item PER_INSTANCE()

This will cause the test_main to be remotely executed on each client instance.
The test_main routine will be run through the iterate method on each client.
This is the default action.

=item SINGLE()

This will cause the test_main to be run on the local host. The test_main
routine will be run through iterate locally. This should only be done when
taking advantage of C<run_queue> inside of the test_main routine.

=back

=item check_topology

C<check_topology> dictates whether or not C<run_test> should perform a discovery
on the current list of FILERs. It is on by default.

=back

=cut

sub run_test {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'TestManager' } ],
        {
		  objective => {type => ARRAYREF, optional => 1, default => []},
		  volume_layout => {type => SCALAR, default => PER_INSTANCE()},
		  test_block_layout => {type => SCALAR, default => PER_INSTANCE()},
		  check_topology => {type => SCALAR | UNDEF, default => 1},
		  test_layout => {type => SCALAR, default => PER_INSTANCE() },
		  timeout => {type => SCALAR, default => 1800}, # 1/2 hour for no good reason
        }
    );

    $Trace->entry();

    $self->{check_topology} = $params{check_topology};

    $Trace->set_options(separateNewline=>1);
    # first process the options
    # this is the main process call so all parameters should be defined. It is now safe to
    #  allow process to look for HELP.
    $Options->process(run_help=>1);

    # call process_params prior to setting up $Options_hash because that routine might change the options
    $self->process_params();
    $Options_hash = $Options->get_hash();
    $self->_define_stages();

    $Remote = new RemoteExecution;
    $Remote_Ancillary = new RemoteExecution;

    if ($self->{test_mode} eq "scalable") {
        # Set unique string to avoid directory name collision running
        # several tests in parallel from the same clients
        $self->{ustr} = $Options_hash->{TESTCASE} ? ${Testname}."_".$Options_hash->{TESTCASE} : ${Testname};

        $self->_start_ancillary_tests();
    }

    my $testcase = $Options->get(testcase=>$Options_hash->{TESTCASE});
    my $result_message = ($testcase && $testcase->{TC_description}) ? $testcase->{TC_description} : $Testname;
    $Trace->trace(msg=>"result_message set to $result_message");

    # save some of the params internal to the object so that the values can be leveraged in other code
    $self->{timeout} = $params{timeout};
    $self->{volume_layout} = $Options_hash->{VOLUME_LAYOUT} || $params{volume_layout};
    $self->{test_block_layout} = $Options_hash->{TEST_BLOCK_LAYOUT} || $params{test_block_layout};
    $self->{test_layout} = $params{test_layout};

    if ($self->{test_mode} eq "classic") {
	# scalable mode will pass in the VOLUME_SHARE information so that each client does not have to query the filer for it
	# Pull this information out of the param and store the mappings in %Shares
	foreach my $vol_share (@{$Options_hash->{VOLUME_SHARE}}) {
	    my ($vol, $share) = split("-", $vol_share);
	    $Trace->debug(msg=>"caching volume '$vol' with export '$share'");
	    $Shares{$vol} = $share;
	}
	$self->{shares_available} = 1; # flag find_volume_shares that it can now convert shares
	iterate( subroutine => \&_run_classic_test_mode);
	$Trace->exit();

	return 1;
    }

    # this check only occurs in scalable TEST_MODE due to this placement
    if ($Options_hash->{CLIENT_INSTANCE} && ( scalar @{$Options_hash->{CLIENT}} > 1 ) ) {
	$Trace->fatal(result=>"SCRIPT", msg=>"the 'CLIENT_INSTANCE' param can not be used when specifying more than one CLIENT");
    }

    if (ref($params{objective}) ne "ARRAY") {
	$Trace->fatal(result=>"SCRIPT", msg=>"the 'objective' parameter is not an ARRAY reference: $params{objective}");
    }

    # if the current volume layout is set to single but the test defined it as volume per instance, then the
    # test has been executed with an overridding param to use a single volume. Flag this condition.
    # Do nothing if the volume layout is set to test_defined.
    if ( ( $self->{volume_layout} eq $self->SINGLE() ) && $params{volume_layout} eq $self->PER_INSTANCE() ) {
        $self->{directory_per_instance} = 1;
    }

    # to allow for running test scripts with testcases in parallel, set the local testdir to a combination of
    #  the testname, the testcase, and the network interface on the filer
    my $testname = $Testname;
    if (my $tc = $Options_hash->{TESTCASE}) {
	$testname .= "_$tc";
	$self->{testname} = $testname;

	# now append the testcase name to the ruleset name to use as well
	$Ruleset .= "_$tc";
    }

    $self->_define_clients();

    # _define_clients will have set $self->{cifs} if it encountered CIFS clients.
    # Now, check the CIFS param to see if it should be set or unset
    if (defined($Options_hash->{CIFS})) {
        $self->{cifs} = $Options_hash->{CIFS};
    }

    # if the test needs NIS information filled in (not the default) make sure it happens
    if ( defined($Options_hash->{NEED_NIS})) {
        $self->{need_nis} = $Options_hash->{NEED_NIS};
    }

    # Try to uniquify the Ruleset by appending the first client and an instance.
    # If someone wants to run the same test from the same client, then they need
    #  to specify a different CLIENT_INSTANCE for each one
    my $client = ${$Options_hash->{CLIENT}}[0];
    if ($Options_hash->{CLIENT_INSTANCE}) {
	$Ruleset .= "_${client}_$Options_hash->{CLIENT_INSTANCE}";
    } else {
	$Ruleset .= "_${client}_0";
    }

    $Server = new Server if (!$Server);
    if ($params{check_topology}) {
        if ($self->{cifs} && !grep($_ =~ /cifs/i, @{$params{objective}})) {
            push(@{$params{objective}}, "cifs")
        }

	$Test_topology = $self->define_cluster_topology();
	if ($Test_topology->{cluster} && ref($Test_topology->{cluster}) ne "HASH") {
	    $Trace->fatal(result=>"SCRIPT", msg=>"the define_cluster_topology() method did not return a HASH reference: $Test_topology->{cluster}");
	}
	$self->{test_topology} = $Test_topology;

	$Discovered_topology = $self->discover_topology(objective=>$params{objective});
	$self->{discovered_topology} = $Discovered_topology;

	$self->_pick_aggregates();
	$self->_pick_vifs_and_vserver();
	$self->{virtual_server} = $Virtual_server;
        $self->_configure_cifs() if ($self->{cifs});
        $self->_find_nis() if ($self->{need_nis});
	$self->update_test_topology();
	my $aggregates;
	foreach my $aggrh (@Aggregates) {
	    $aggregates .= " ".$aggrh->{aggregate};
	}
	$Trace->debug(msg=>"Aggregates: $aggregates");
	$Trace->debug(msg=>"Virtual Server: $Virtual_server->{vserver}");
	$Trace->debug(msg=>"Data Interfaces: ".join(",", @Data_interfaces));

        # build up information for client_setup/cleanup

        # Build the unix mountpoint if it wasn't passed as a parameter (testdir)
        $self->{testdir} = $Options_hash->{testdir};
        unless (defined $self->{testdir}) {
            my $data_interface_path = $Data_interfaces[0];
            $data_interface_path =~ s/\:/_/g; # Replace IPv6 address colons with underscores (cthon can't handle colons)
            $self->{testdir} = "/t/$testname.$data_interface_path";

            # Append the user-specified uniquifier if MOUNTPOINT_UNIQUIFIER was provided
            if ((defined $self->{mountpoint_uniquifier}) && (length($self->{mountpoint_uniquifier}) > 0)) {
                $self->{testdir} .= '.' . $self->{mountpoint_uniquifier};
            }
        }

        $Options->set(testdir => $self->{testdir}) if $self->{volume_layout} eq SINGLE();
    }
    $self->setup_test_params(%params);

    if ($params{check_topology}) {
        $Test_topology->{clients} = $self->setup_mount_info();
        if ($Test_topology->{clients} && ref($Test_topology->{clients}) ne "HASH") {
            $Trace->fatal(msg=>"setup_mount_info returned something other than a HASH reference or undef: $Test_topology->{clients}",
                          result=>"CONFIG");
        }
        # now that the client topology has been specified, convert the information
        # into a common format. When AUTOMOUNT=1, this information is passed to each
        # of the clients. When the param is not set, the info is not passed to the clients
        # but is used locally by client_setup.
        $self->_convert_mount_info();
    }

    # the params for each client are generated prior to the mount information being
    #  generated. But some of the mount information generated during setup_mount_info()
    #  and _convert_mount_info() need to be passed along as well so perform that work
    #  now
    $self->_update_test_params(); # internal housekeeping

    $self->update_test_params(); # External, if test has to overwrite this.

    $self->{shares_available} = 1; # flag find_volume_shares that it can now convert shares

    my ($filer_setup_result, $client_setup_result, $test_result, $client_cleanup_result, $filer_cleanup_result);

    eval {
	if ($Stages{filer_setup}) {
	    $Trace->comment(msg=>"Running STAGE filer_setup");
	    $self->filer_setup();
	}
    };
    $filer_setup_result = $@;

    if (!$filer_setup_result) {
        eval {
            if ($Stages{client_setup}) {
                $Trace->comment(msg=>"Running STAGE client_setup");
                $self->client_setup();
            }
        };
        $client_setup_result = $@;
    }

    if (!$filer_setup_result && !$client_setup_result) {
        eval {
            if ($Stages{test_setup} || $Stages{test_main} || $Stages{test_cleanup}) {
                $self->_call_test_code();
            }
        };
        $test_result = $@;
    }

    eval {
	if ($Stages{client_cleanup}) {
	    $Trace->comment(msg=>"Running STAGE client_cleanup");
	    $self->client_cleanup();
	}
    };
    $client_cleanup_result = $@;

    eval {
	if ($Stages{filer_cleanup}) {
	    $Trace->comment(msg=>"Running STAGE filer_cleanup");
	    $self->filer_cleanup();
	}
    };
    $filer_cleanup_result = $@;

    # Before reporting the test result and exiting, kill all ancillary tests
    $Trace->comment(msg=>"Killing Ancillary Tests if Started Any");
    $Remote_Ancillary->kill();

    $self->_print_warnings();
    if ($test_result) {
        if ($client_setup_result) {
            $Trace->warn(msg=>"client_setup error: $client_setup_result");
        }
        if ($filer_setup_result) {
            $Trace->warn(msg=>"filer_setup error: $filer_setup_result");
        }
        if ($client_cleanup_result) {
            $Trace->warn(msg=>"client_cleanup error: $client_cleanup_result");
        }
        if ($filer_cleanup_result) {
            $Trace->warn(msg=>"filer_cleanup error: $filer_cleanup_result");
        }
    } else {
        my $result = "";
        $result .= "client_setup error: $client_setup_result" if ($client_setup_result);
        $result .= "filer_setup error: $filer_setup_result" if ($filer_setup_result);
        $result .= "client_cleanup error: $client_cleanup_result" if ($client_cleanup_result);
        $result .= "filer_cleanup error: $filer_cleanup_result" if ($filer_cleanup_result);
        if ($result) {
            $Trace->fatal(result=>"FATAL", msg=>"Encountered failures during setup or cleanup\n$result");
        }
    }

    if ($test_result) {
	$Trace->trace(msg=>"result = $test_result");
	if ($self->{stages_changed}) {
	    $result_message .= " (with STAGE changes) ";
	    $Trace->trace(msg=>"result_message set to $result_message");
	}
	$result_message .= " ";
	if (ref($test_result)) {
	    $result_message .= join("\n", ($test_result->messages));
	    $Trace->trace(msg=>"result_message set to $result_message");
	} else {
	    $result_message .= $test_result;
	    $Trace->trace(msg=>"result_message set to $result_message");
	}
	my $result_type = "FAIL";
	if (UNIVERSAL::isa($test_result, 'NATE::Result')) {
              $result_type = $test_result->type();
        }
	logresult($result_type, $result_message);
    } else {
	if ($self->{stages_changed}) {
	    $result_message .= " (with STAGE changes) ";
	    $Trace->trace(msg=>"result_message set to $result_message");
	}
	my $clients = $Options->get(option=>"CLIENT");
	my $numclients = scalar(@$clients);
	my $client_msg = " on $numclients client";
	$client_msg .= "s" if ($numclients > 1);
	# only log the result if it wasn't already logged (in a localhost mode)
	logresult("PASS", $result_message.$client_msg);
    }
    $Trace->exit();
}

sub _print_warnings {
    my $self = shift;
    if ($self->{warnings}) {
        foreach my $warn (@{$self->{warnings}}) {
            $Trace->warn(msg=>$warn)
        }
    }
}

sub run_queue {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'TestManager' } ],
					 {
					     jobs => {type => ARRAYREF},
					     runid => {type => SCALAR | UNDEF, optional => 1 },
					     timeout => {type => SCALAR},
					 }
					 );
    $Trace->entry();

    my $jqueue = new JobQueue(%$Options_hash, check_clients=>undef, runid=>($params{runid} || $Testname));
    $jqueue->Loop($params{jobs}, $self->_create_test_hash(), $params{timeout});

    $Trace->exit();
}

sub generate_volume_names {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'TestManager' } ],
					 {
					     base_name => {type => SCALAR},
					 }
					 );
    $Trace->entry(msg=>"base_name=$params{base_name}");

    my @volumes;
    foreach my $client (keys %{$self->{clients}}) {
        foreach my $instance (keys %{$self->{clients}{$client}}) {
            my $vclient = $self->_normalize_volume_name(volume=>$client);
            my $volume_name = $params{base_name};
            # this mode is supposed to use a unique volume for each client instance.
            $volume_name .= "_${vclient}_$instance";
            push(@volumes, $volume_name);
        }
    }

    $Trace->exit(msg=>Dumper(\@volumes));
    return @volumes;
}

sub define_cluster_topology {
    my $self = shift;
    $Trace->entry();

    my $topology = {};
    my $vol_size = ($Options_hash->{VOL_SIZE}) ? $Options_hash->{VOL_SIZE} : "1GB";
    my $vol_files = $Options_hash->{VOL_FILES} if ($Options_hash->{VOL_FILES});
    my $vol_maxdir = $self->size_in_mb(size=>$Options_hash->{VOL_MAXDIR_SIZE});
    $vol_maxdir .= "MB" if ($vol_maxdir);

    $Trace->trace(msg=>"volume_layout=$self->{volume_layout}");

    my $default_volume_name = "p_${Testname}";
    $default_volume_name .= "_$Options_hash->{TESTCASE}" if (defined($Options_hash->{TESTCASE}));
    if ($self->{volume_layout} eq PER_INSTANCE()) {
        my $base_volume_name;
        my $base_share;

        if (scalar(@{$Options_hash->{volume}})) {
            $base_volume_name = ${$Options_hash->{volume}}[0];
        } else {
            $base_volume_name = $default_volume_name;
        }
	foreach my $client (keys %{$self->{clients}}) {
	    foreach my $instance (keys %{$self->{clients}{$client}}) {
		my $vclient = $self->_normalize_volume_name(volume=>$client);
		my $volume_name = $base_volume_name;
                # this mode is supposed to use a unique volume for each client instance.
                $volume_name .= "_${vclient}_$instance";
		# in the PER_INSTANCE case, the volume_name and share need to be saved for later
		my $share = "/$volume_name";
                push(@{$self->{clients}{$client}{$instance}}, [$volume_name, $share]);
		push(@{$topology->{cluster}{volume}},
		     {
			 volume => $volume_name,
			 'junction-path' => $share,
			 size => $self->size_in_mb(size=>$vol_size)."MB",
			 files => $vol_files,
			 'maxdir-size' => $vol_maxdir,
			 user => 0,
			 group => 0,
		     });
	    }
	}
    } elsif ($self->{volume_layout} eq SINGLE()) {
        my @volumes;
        if (scalar(@{$Options_hash->{volume}})) {
            @volumes = @{$Options_hash->{volume}};
        } else {
            push(@volumes, $default_volume_name);
        }
        my $volume_array;
        my $share;
        my $vserver = $Options_hash->{vserver};
        foreach my $volume_name (@volumes) {
            $Trace->trace(msg=>"specified volume='$volume_name'\nspecified vserver='$vserver'");
            if ($vserver) {
                eval { $volume_array = $Server->volume_show(volume => $volume_name,
                                                            vserver => $vserver) };
                if ( $volume_array ) {
                    $share = $volume_array->[0]->{'junction-path'};
                    $Trace->trace(msg=>"share for '$volume_name' found to be '$share'");
                    if (!$share || $share eq "-") {
                        $Trace->fatal(msg=>"Did not get an junction-path for volume '$volume_name':".Dumper($volume_array));
                    }
                }
            } else {
                # try looking up the generic volume name
                eval { $volume_array = $Server->volume_show(volume => $volume_name); };
                if ( $volume_array ) {
                    if (scalar @$volume_array > 1) {
                        $Trace->fatal(result=>"CONFIG",
                                      msg=>"Found more than one volume of name '$volume_name':".Dumper($volume_array));
                    }
                    $share = $volume_array->[0]->{'junction-path'};
                    $Trace->trace(msg=>"share for '$volume_name' found to be '$share'");
                    if (!$share || $share eq "-") {
                        $Trace->fatal(msg=>"Did not get an junction-path for volume '$volume_name':".Dumper($volume_array));
                    }
                    # set the vserver choice to the vserver for this volume
                    $Options->set(vserver=>$volume_array->[0]->{vserver});
                    $Trace->trace(msg=>"vserver for '$volume_name' found to be '$volume_array->[0]->{vserver}'. This will be the default.");
                }
                if (!$volume_array && $volume_name !~ /^p_/) {
                    $Trace->warn(msg=>"The volume_name does not start with 'p_' indicating a private resource");
                }
            }

            # save this volume hash because _pick_aggregates() needs to see some of the information
            $self->{volume_hash} = $volume_array->[0];

            # catch the case where the volume does not exist yet. Fill in a sane share for it.
            if (!$share) {
                $share = $Options_hash->{JUNCTION_PATH} || "/$volume_name";
                $Trace->trace(msg=>"set the share for '$volume_name' to be '$share'");
            }
        }
	my $new_vol_size = 0;
	my $new_vol_files = 0;
	my $vol_size_unit;
        my $instance_count = 0;
        my $client_count = 0;
	if ($vol_size =~ m/\d+(MB|M|GB|G|TB|T)/i) {
	    $vol_size_unit = $1;
	}

        $Trace->debug(msg => "Volume size is $vol_size$vol_size_unit per instance per client.");

        # temporary hack to maintain the old way of just picking off the first volume
        my $volume_name = $volumes[0];
	foreach my $client (keys %{$self->{clients}}) {
            $client_count ++;
	    foreach my $instance (keys %{$self->{clients}{$client}}) {
                $instance_count ++;
                push(@{$self->{clients}{$client}{$instance}}, [$volume_name, $share]);

		# define the directories that should be created in this volume
		if ($self->{directory_per_instance}) {
                    $new_vol_size += $vol_size;
                    $new_vol_files += $vol_files;
		    push(@{$topology->{directories}{$volume_name}}, "${client}_${instance}_".$self->testcase_string());
		}
	    }
	}

        if ($new_vol_size) {
            $Trace->debug(msg => "Calculated new volume size $new_vol_size$vol_size_unit per $instance_count instance(s) on $client_count client(s) (".$self->size_in_mb(size=>$new_vol_size.$vol_size_unit).")");
            $vol_size = $new_vol_size.$vol_size_unit;
        }

        if ($new_vol_files) {
            $Trace->debug(msg => "Calculated new volume files $new_vol_files per $instance_count instance(s) on $client_count client(s)");
            $vol_files = $new_vol_files;
        }

	push(@{$topology->{cluster}{volume}},
	     {
		 volume => $volume_name,
		 'junction-path' => $share,
                 size => $self->size_in_mb(size=>$vol_size)."MB",
		 files => $vol_files,
		 'maxdir-size' => $vol_maxdir,
		 user => 0,
		 group => 0,
	     });
    } else {
	$Trace->fatal(msg=>"The volume layout you have specified ($self->{volume_layout}) is not supported",
		      result=>"SCRIPT");
    }
    $Trace->debug(msg=>"Topology=".Dumper($topology));
    $Trace->debug(msg=>"clients=".Dumper($self->{clients}));
    $Trace->exit();
    return $topology
}

sub discover_topology {
    my $self = shift;
    my %params = @_;

    my $topology;

    $Trace->entry(msg=>join(" ", @_));

    # Define the base list of command nodes to discover in the cluster
    my @discovery_nodes = qw(aggregate vserver network_interfaces);

    # Add network ports to the list if the new Topology includes any network interfaces. If it does, then
    #  a network port needs to be added to each network interface definition
    push(@discovery_nodes, "network_ports") if ($Test_topology->{cluster}{network_interfaces});

    # Add the cifs nodes to the list if the user is running a CIFS test
    push(@discovery_nodes, ("cifs_server", "cifs_share", "kerberos", "cifs_superuser")) if (grep("cifs" =~ /$_/i, @{$params{objective}}));

    # add nis to the nodes if the test needs NIS
    push(@discovery_nodes, "nis") if ($self->{need_nis});

    $topology = $Server->discover_topology(cmd_nodes=>\@discovery_nodes);

    # Determine if any of the filers are actually configured to be FlexCaches.
    # This check is only done if the CACHE_TRAFFIC is set since we don't want to
    #   make calls against each node object unless we absolutely have to since it
    #   will force connections to be made against each node.
    $self->{flexcaches_involved} = 0;
    if ($Options_hash->{CACHE_TRAFFIC}) {
        # Classic filers must be looked at via the server object since there is not a
        #  global view like there is in NG. Check each filer and flag whether or not
        #  a cache is involved since we do not want to create volumes on these filers.
        foreach my $filer ($Server->servers()) {
            $self->{flexcaches_involved}++ if ($Server->is_cache(-Server=>$filer));
        }
        if ($self->{flexcaches_involved} == $Server->servers()) {
            $Trace->fatal(msg=>"All of the specified filers are licensed for FlexCache.\n".
                          "This code requires that at least one is not licensed.",
                          result=>"CONFIG");
        }
        if (!$self->{flexcaches_involved}) {
            # supposed to use a cache but there are none
            $Trace->fatal(msg=>"The CACHE_TRAFFIC param has been set but no FlexCache filers were\n".
                          "specified with the FILER param. The current list of filers is:\n".
                          join(", ", $Server->servers()), result=>"CONFIG");
        }
    }

    $Trace->debug(msg=>Dumper($topology));
    $Trace->exit();
    return $topology;
}

sub _run_code_remotely {
    my $self = shift;
    my %params = @_;
    $Trace->entry();

    my $relative_test_name = $self->_find_relative_testname(testname=>$Full_testname);
    
    my $runseq = "01"; # for nate
    my $username = ($self->{cifs}) ? "root" : undef;
    while (my ($client, $chash) = each %{$self->{test_options}}) {
	while (my ($instance, $hash_ref) = each %$chash) {
            # This code used to pass a username into execute so that the remote test could be run as a different
            # user. This was causing some problems with CIFS tests so it has been removed for now
	    $Remote->execute(host=>$client, cmd=>$relative_test_name, jobid=>"${client}_${instance}_test_stage",
			     runid=>"${Testname}_test_${client}_${instance}", nate_params=> {%$hash_ref, %params},
			     handler=>sub {}, retry_cifs_errors=>0);
	}
    }

    my $timeout;
    if ($Options_hash->{TEST_TIMEOUT}) {
	$timeout = $Options_hash->{TEST_TIMEOUT};
    } else {
	$timeout = expected_run_time(timeout=>$self->{timeout});
    }
    $timeout *= $Options_hash->{TIMEOUT_MULTIPLIER}; # defined in Server (Node.pm)
    my $stats = $Remote->get_test_stats();
    $Trace->comment(msg=>"The following stats are to be considered beta.");
    $Trace->comment(msg=>sprintf("Test start averaged %01.3fs on each host.", $$stats{average_start_time}));
    $Trace->comment(msg=>sprintf("Total start time %01.3fs to start %d client(s).", $$stats{total_start_time}, $$stats{clients_started}));
    $Trace->comment(msg=>"STAGE test_main will run remotely");
    $Trace->comment(msg=>"wait for $timeout seconds on the whole test to run");
    my $result = $Remote->wait(timeout=>$timeout);
    if (ref($result)) {
        $Trace->fatal(msg=>"Encountered an error with the client tests:\n", result=>$result);
    } elsif ($result) {
        $Trace->fatal(msg=>"Encountered an error with the client tests: $result\n");
    }
    $Trace->exit();
}

# This routine will return a unique string comprised of the testname and testcase name
#  separated with an "_". If there is not testcase name, then it is simply the test name.
#  The value is initialized at the beginning of run_test() and only in scalable mode.
sub testcase_string {
    my $self = shift;
    return $self->{ustr};
}

# This routine gets called so many times, it is not wise to add tracing because it can fill a
# log file with all of the call statements
sub _normalize_volume_name {
    my $self = shift;
    my %params = @_;
    $params{volume} =~ s/-/_/g;
    return $params{volume};
}

sub _find_volume_for_client_instance {
    my ($self, $test_options, $client, $instance) = @_;
    $Trace->entry(msg=>"for $client $instance");

    my @volumes;
    my @shares;

    if ($test_options->{volume} && scalar(@{$test_options->{volume}})) {
	# the volume name is specified in the test_options hash for this client instance
	@volumes = @{$test_options->{volume}};
	if ($test_options->{share}) {
	    push(@shares, $test_options->{share});
	} else {
	    foreach my $volume (@volumes) {
		push(@shares, $self->find_volume_share(volume=>$volume));
	    }
	}
	$Trace->debug(msg=>"selected volumes from the test options for $client\:$instance\:".join(", ",@shares));
    } else {
	# find the list of volumes from the volumes in the Test_topology
	foreach my $volumeh (@{$Test_topology->{cluster}{volume}}) {
	    push(@volumes, $volumeh->{volume});
	    push(@shares, $volumeh->{'junction-path'});
	}
    }

    if (scalar @shares > 1) {
	# I don't know what to do with more that one share yet so I am just going to use the first one
	# if this code changes then the code in setup_test_params also needs to change
    }

    $Trace->exit(msg=>"volume=@volumes, share=@shares");
    return (\@volumes, \@shares);
}

sub automount_mount {
    my $self = shift;
    my %params = (
                  client => undef,
                  @_);
    my $client = $params{client};
    $Trace->entry();

    # if AUTOMOUNT is on and this is not a windows client, create the mountpoint now.
    # CIFS clients use UNC names so the mount is not needed
    # This step is performed outside of the eval blocks because a failure needs is
    #  abortive to the entire routine
    if ($Options_hash->{AUTOMOUNT}) {
        # only perform a mount if we are in localhost mode or test mode is classic and not windows
        if ( $self->{localhost} || (($self->{test_mode} eq $self->test_classic() && $OSNAME !~ /win32/i) )) {
            $self->_mount_client(client=>$client);
        }
    }

    $Trace->exit();
}

sub automount_umount {
    my $self = shift;
    my %params = (
                  client => undef,
                  @_);
    my $client = $params{client};
    $Trace->entry();
    $Trace->entry();

    # if AUTOMOUNT is on and this is not a windows client, cleanup the mountpoint now.
    # CIFS clients use UNC names so the umount is not needed
    # Perform the umount now, inside of the eval for test_cleanup. The downside is that a
    # failure in test_cleanup will leave it mounted...is this a good idea?

    if ($Options_hash->{AUTOMOUNT}) {
        # only perform a umount if we are in localhost mode or test mode is classic and not windows
        if ( $self->{localhost} || (($self->{test_mode} eq $self->test_classic() && $OSNAME !~ /win32/i) )) {
            $self->_umount_client(client=>$client);
        }
    }

    $Trace->exit();
}

sub _run_classic_test_mode {
    my ($self) = @_;
    $self = $Self if (!$self);
    $Trace->entry();
    # if this is running in local mode then the script is running on the client.
    # execute the test_main. return 1 if it succeeds. It should die on error
    my ($setup_result, $test_result, $cleanup_result);

    # find the name of the client that this code is currently executing on.
    #  when no CLIENT is specified, it is the equivalent of localhost.
    # Perform this now since it will be used to mount and umount.
    my $client = hostname();

    #
    # Start a network trace if desired.
    #
    my $traces;
    my $nodetraces = new Network::NodeTrace();
    if (ref($Options_hash->{TRACE_ADDRS}) eq 'ARRAY') {
        foreach my $addr (@{$Options_hash->{TRACE_ADDRS}}) {
            push(@{$traces}, new Network::Trace(address => $addr));
        }
    } else {
        push(@{$traces}, new Network::Trace(address => $Options_hash->{TRACE_ADDRS}));
    }

    $self->automount_mount(client=>$client);

    eval {
	if ($Stages{test_setup}) {
	    $Trace->comment(msg=>"Running STAGE test_setup");
	    $self->test_setup();
	}
    };
    $setup_result = $@;

    eval {
        if (!$setup_result && $Stages{test_main}) {
            $Trace->debug(msg=>"localhost=$self->{localhost}");
            $Trace->debug(msg=>"test_mode=$self->{test_mode}");
            $Trace->debug(msg=>"test_block_layout=$self->{test_block_layout}");
            $Trace->debug(msg=>"test_layout=$self->{test_layout}");

            # there was no error during test_setup and test_main is defined to run.
            # If the current mode is scalable and the test_layout is per_instance then the test_main needs to be run
            #   run remotely on each of the clients.
            if (!$self->{localhost} && $self->{test_mode} eq $self->test_scalable() &&
                $self->{test_block_layout} eq SINGLE() && $self->{test_layout} eq PER_INSTANCE()) {
                # run just the test_main remotely on each client if the setup/cleanup is to be done once (not on each client)
                #  and the test_main is to be executed on each client (rather than through the queue)
                $self->_run_code_remotely(STAGE => "test_main");
            } else {
                # otherwise, the test_main will run locally and not on each client (or this was localhost and will just
                # run locally
                $Trace->comment(msg=>"Running STAGE test_main on $client");
                if ($self->{test_layout} eq PER_INSTANCE()) {
                    $self->test_main();
                } else {
                    if ($Options_hash->{AUTOMOUNT} && $OSNAME !~ /win32/i) {
                        print Dumper $self;
                        print Dumper $Options_hash;
                        $Trace->fatal(result=>"SCRIPT",
                                      msg=>"ASSERT: Don't know how to handle AUTOMOUNT when test_layout is single");
                    }
                    $self->test_main();
                }
            }
        }
    };
    $test_result = $@;

    eval {
	if ($Stages{test_cleanup}) {
	    $Trace->comment(msg=>"Running STAGE test_cleanup");
	    $self->test_cleanup();
	}
        $self->automount_umount(client=>$client);
    };
    $cleanup_result = $@;

    if ($cleanup_result && ($test_result || $setup_result)) {
	$Trace->warn(msg=>"Cleanup failure: $cleanup_result");
    }

    if (ref($setup_result)) {
	$Trace->fatal(result=>$setup_result);
    } elsif ($setup_result) {
	$Trace->fatal(result=>'CONFIG', msg=>"test_setup failed: $setup_result");
    }

    if ($cleanup_result) {
	$Trace->fatal(msg=>"Cleanup failure: $cleanup_result");
    }

    if (ref($test_result)) {
        $Trace->fatal(result=>$test_result);
    } elsif ($test_result) {
	$Trace->fatal(result=>'FAIL', msg=>"test_main encountered an error: $test_result");
    }
    $Trace->exit();
}

sub _define_clients {
    my $self = shift;
    my %params = @_;

    $Trace->entry();

    my $clients = $self->get(parameter=>"CLIENT");

    # set the default client list to localhost so that a test run on a
    #  local client is a little faster and has cleaner output. This helps
    #  the wafl tests
    if (! scalar @$clients) {
	$clients = ["localhost"];
	$Options->set(CLIENT=>$clients);
    } elsif ($Stages{client_setup} || $Stages{client_cleanup} ||
	     $Stages{test_setup} || $Stages{test_main} || $Stages{test_cleanup}) {
	# for the stages that talk to clients, check them up front before running any of the stages
	my @clients = $Remote->check_clients(@$clients);
	$clients = \@clients;
	$self->set(CLIENT => $clients);
    }
    my $prev_client;
    my $instance = 0;
    foreach my $client (sort @$clients) {
	# if localhost is being used, replace it with the real hostname
	if ($client =~ "localhost") {
	    # set the localhost flag if there is a single host instance and it is localhost.
	    #  _call_test_code will not subtest in this case
	    if (scalar @$clients == 1) {
		$self->{localhost} = 1;
		$Trace->comment(msg=>"No clients have been specified so localhost will be used");
	    }
	    chomp($client = `hostname`);
	}
	# lose the domain name
	if ($client =~ /\./) {
	    $client = (split(/\./, $client))[0];
	}
	if ($client ne $prev_client) {
            my $c = Tharn::host($client);
            if ($c->hosttype() =~ /windows/) {
                $self->{cifs} = 1;
                $self->{Cifs_Clients}{$client} = 1;
            } else {
                $self->{Unix_Clients}{$client} = 1;
            }
	    $instance = 0;
	} else {
	    $instance++;
	}
	$prev_client = $client;
	$self->{clients}{$client}{$instance} = [];
    }
    $Trace->exit();
}

sub _wait_on_clients {
    my $self = shift;
    my $job = shift;
    my $timeout = 600; # wait for 10 minutes which has worked for 10 mounts on 140 hosts
    $Trace->trace(msg=>"wait for $timeout seconds on the $job to finish");
    my $result = $Remote->wait(timeout=>$timeout);
    if ($result) {
        $Trace->fatal(msg=>"Error with $job on clients: $result");
    }
}

#
# This routine is used for static and AUTOMOUNT mounts. It would be cleaner at execution
#  time to have the AUTOMOUNT mounts just make the calls to mount() directly but it ends
#  up making the code messier and duplicates the work in mount.thpl
# This code is assumed to be run against a single client, which might be the localhost.

# All of the code for chmod_list and CHMOD_LIST is probably no longer needed
sub _mount_client {
    my $self = shift;
    my %params = (
                  client => undef,
                  bg => undef, # should this be run in the background?
                  @_);
    my $client = $params{client};
    $Trace->entry(msg=>"client=$client");

    my $mount_list;
    my $chmod_list;
    if ($self->{test_mode} eq "scalable") {
        # in this mode, the mount information can be pulled directly from the internal
        #  hashes which are only available in scalable mode.
        if ($Test_topology->{clients}{$client}) {
            $mount_list = join("xMOUNTx", @{$Test_topology->{clients}{$client}{mount_list}});
            $chmod_list = join("xCHMODx", @{$Test_topology->{clients}{$client}{chmod_list}});
        }
    } else {
        # in this mode, the mount information is only available via the param that was
        #  created by TestManager in scalable mode and passed to each test
        if ($Options_hash->{AUTOMOUNT}) {
            # automount in classic mode
            $mount_list = $Options_hash->{MOUNT_LIST};
            $chmod_list = $Options_hash->{CHMOD_LIST};
        } else {
            # no automount in classic mode
            $Trace->fatal(msg=>"This code should not be called\n");
        }
    }


    # create a nate params hash to pass to the mount.thpl.
    my $params = {
                  # this is the delimiter for the list of mounts
                  MOUNT_LIST => $mount_list,
                  # specify a subdir in the LOGDIR to store these setup logs
                  LOGDIR => "$Options_hash->{LOGDIR}/client_setup",
                 };
    if (!$Options_hash->{RO_PATH}) {
        # this is the delimiter for the list of chmods
        $params->{CHMOD_LIST} = $chmod_list;
    }

    $Remote->execute(host=>$client, cmd=>"tools/mount.thpl", jobid=>"mount_${client}",
                     runid=>"mount_${client}", nate_params=>$params,
                     handler=>sub {});
    if (! $params{bg}) {
        $self->_wait_on_clients("mount");
    }

    $Trace->exit();
}

#
# This routine is used for static and AUTOMOUNT umounts. It would be cleaner at exection
#  time to have the AUTOMOUNT mounts just make the calls to umount() directly but it ends
#  up making the code messier and duplicates the work in umount.thpl
# This code is assumed to be run against a single client, which might be the localhost.
sub _umount_client {
    my $self = shift;
    my %params = (
                  client => undef,
                  bg => undef, # should this be run in the background?
                  @_);
    my $client = $params{client};
    $Trace->entry(msg=>"client=$client");

    my $mount_list;
    my @mount_list;
    if ($self->{test_mode} eq "scalable" && $Test_topology->{clients}{$client}) {
        # in this mode, the mount information can be pulled directly from the internal
        #  hashes which are only available in scalable mode.
        foreach my $mount_entry (reverse @{$Test_topology->{clients}{$client}{mounts}}) {
            push(@mount_list, $mount_entry->{mnt});
        }
    } else {
        # in this mode, the mount information is only available via the param that was
        #  created by TestManager in scalable mode and passed to each test
        if ($Options_hash->{AUTOMOUNT}) {
            # automount in classic mode
            foreach my $mount_entry (reverse split(/xMOUNTx/, $Options_hash->{MOUNT_LIST})) {
                # dig down to find the entry we are looking for
                my $mnt = (split(/:/, $mount_entry))[2];
                push(@mount_list, $mnt);
            }
        } else {
            # no automount in classic mode
            $Trace->fatal(msg=>"This code should not be called\n");
        }
    }
    $mount_list = join("xMOUNTx", @mount_list);

    my $params = {
                  # create a nate params hash to pass to the mount.thpl.
                  MOUNT_LIST => $mount_list,  # this is the delimiter for the list of mounts
                  NATE_LIB => $Options_hash->{NATE_LIB},
                  LOGDIR => "$Options_hash->{LOGDIR}/client_cleanup", # specify a subdir in the LOGDIR to store these cleanup logs
                 };
    $Remote->execute(host=>$client,
		    cmd=>"tools/umount.thpl",
		    jobid=>"umount_${client}",
                    runid=>"umount_${client}",
		    nate_params=>$params,
                    handler=>sub {}
		    );

    if (! $params{bg}) {
        $self->_wait_on_clients("umount");
    }

    $Trace->exit();
}

sub _configure_cifs {
    my $self = shift;
    $Trace->entry();

    # TODO -- This should be safe to remove once we enter RRT for BR
    if ( $Server->get_type() == $Server->ONTAP_NG() ) { # only do this when it is an NG system
        if ( $Server->has_change(change => "cifs-server") && !$Server->has_change(change => "cifs-server-rename") ) {
            $Cifs_name_key = 'name';
        }
    }

    # As of now, TestManager keys on the vserver first. This isn't always the best thing, but
    # for now, that's the way it is. So, we'll be consistent with that in this routine. At the
    # time we enter this routine, the vserver has already been "chosen". Given the vserver,
    # we then need to determine if there is a CIFS server, and then secondary to that,
    # determine the CIFS domain for that CIFS server.

    # coming into here, vserver is already known
    # Problems to still resolve:
    #  this vserver might not have a cifs server
    #  this vserver could be in the wrong domain
    #  this CIFS domain may not match the  Kerberos Realm we have -- unlikely unless user error
    #  cifs_server_show - needs default_unix_user
    #  vserver_show - name_service_switch should be NIS  or (files and services unix_user not null)

    # Determine the CIFS server
    my @servers = $Server->search_topology(cmd_node => "cifs_server",
                                           key_pairs => ["vserver:$Virtual_server->{vserver}"],
                                          );
    if (! scalar @servers) {
        $Trace->fatal(result=>"CONFIG", msg=>"Could not find a CIFS server for vserver=$Virtual_server->{vserver}");
    }
    if ($Options->defined(option=>"CIFS_SERVER")) {
        ($Cifs_Server) = grep($Options_hash->{CIFS_SERVER} eq $_->{$Cifs_name_key}, @servers);
        if (!$Cifs_Server) {
            $Trace->fatal(result=>"CONFIG", msg=>"Vserver $Virtual_server->{vserver} does not contain the specified CIFS server ($Options_hash->{CIFS_SERVER})");
        }
    } else {
        $Cifs_Server = $servers[0];
    }
    $self->{Cifs_Server} = $Cifs_Server;
    $Trace->trace(msg=>"Using CIFS server=".Dumper($Cifs_Server));

    # Determine the CIFS share
    # This only works with GX, not 7G
    if ( $Options->defined(option=>"CIFS_SHARE") ) {
        my @shares = $Server->search_topology(
            cmd_node => "cifs_share",
            key_pairs => ["vserver:$Virtual_server->{vserver}","share-name:$Options_hash->{CIFS_SHARE}"],
        );
        if ( ! scalar @shares ) {
            $Trace->fatal(result=>"CONFIG", msg=>"Vserver '$Virtual_server->{vserver}' does not contain the specified CIFS Share ($Options_hash->{CIFS_SHARE})");
        } elsif ( scalar @shares > 1 ) { # this shouldn't happen....
            $Trace->warn(msg => "There seems to be multiple SHARES on Virtual Server '$Virtual_server->{vserver}' that have the same name ($Options_hash->{CIFS_SHARE}) and that should not happen. Using the first one found.");
            $Cifs_Share = $shares[0]{'share-name'};
        } else {
            $Cifs_Share = $shares[0]{'share-name'};
        }
    } else {
        $Cifs_Share = uc "p_".$Testname;
    }
    # Put the CIFS Share onto the TestTopology so we create it if necessary
    my %share_hash = (
        vserver     => $Virtual_server->{vserver},
        'share-name' => $Cifs_Share,
        path        => "/.admin",
        comment     => '"created by TestManger"',
    );
    # Handle any of the per share parameters, default to passing in nothing
    if ( $Options->defined(option=>"CIFS_SHARE_SHARE_PROP") ) {
        $share_hash{'share-properties'} = join ",", @{$Options_hash->{CIFS_SHARE_SHARE_PROP}};
    }
    if ( $Options->defined(option=>"CIFS_SHARE_SYMLINK_PROP") ) {
        $share_hash{'symlink-properties'} = $Options_hash->{CIFS_SHARE_SYMLINK_PROP};
    }
    if ( $Options->defined(option=>"CIFS_SHARE_FILE_UMASK") ) {
        $share_hash{'file-umask]'} = $Options_hash->{CIFS_SHARE_FILE_UMASK};
    }
    if ( $Options->defined(option=>"CIFS_SHARE_DIR_UMASK") ) {
        $share_hash{'dir-umask]'} = $Options_hash->{CIFS_SHARE_DIR_UMASK};
    }
    push(@{$Test_topology->{cluster}{cifs_share}}, \%share_hash);
    $self->{Cifs_Share} = $Cifs_Share;
    $Trace->trace(msg=>"Using CIFS share=" . Dumper(\%share_hash));

    # Determine the CIFS domain
    $Cifs_Domain = $Cifs_Server->{'domain-workgroup'} || $Cifs_Server->{domain};
    if ( !defined($Cifs_Domain) || $Cifs_Domain eq "" ) {
        $Trace->fatal(result=>"CONFIG", msg=>"Could not find a CIFS domain for vserver=$Virtual_server->{vserver}");
    }
    if ( $Options->defined(option=>"CIFS_DOMAIN") && $Options_hash->{CIFS_DOMAIN} ne $Cifs_Domain ) {
	$Trace->fatal(result=>"CONFIG", msg=>"Discovered CIFS domain ($Cifs_Domain) differs from specified domain ($Options_hash->{CIFS_DOMAIN})");
    }
    $self->{Cifs_Domain} = $Cifs_Domain;
    $Trace->trace(msg=>"Using CIFS Domain='$Cifs_Domain'");

    $Trace->exit();
}

# this simple routine just checks to see if the NIS_DOMAIN and NIS_SERVER passed in matches what was found on
# the Virtual_server, or if there are none, we fill them in for the test.  This code only gets called if
# someone sets NEED_NIS (or $self->{need_nis})
sub _find_nis {
    my $self = shift;

    $Trace->entry();

    my @nis = $Server->search_topology( cmd_node => "nis",
        key_pairs => ["vserver:$Virtual_server->{vserver}", "active:true"],
    );

    if ( scalar(@nis) == 0 ) {
        $Trace->fatal(result => "CONFIG",
            msg => "The test has set the 'NEED_NIS' flag, but the Virtual Server '$Virtual_server->{vserver}' does not have a NIS Domain configured.");
    } elsif ( scalar(@nis) > 1 ) {
        $Trace->warn(msg => "The Virtual Server '$Virtual_server->{vserver}' seems to have more than 1 NIS domain configured, which isn't possible.  Using the first NIS Domain we found ($nis[0]->{domain}).");
    }
    $Nis = $nis[0];

    # now lets check to see if we were passed in a NIS_DOMAIN, or a NIS_SERVER and validate that they match
    # what we have discovered.
    if ( $Options->defined(option => "NIS_DOMAIN") && ($Options_hash->{NIS_DOMAIN} ne $Nis->{domain}) ) {
        $Trace->fatal(result => "CONFIG",
            msg => "Discovered NIS Domain for Virtual Server '$Virtual_server->{vserver}' ($Nis->{domain}) differs from specified NIS Domain ($Options_hash->{NIS_DOMAIN})");
    }
    if ( $Options->defined(option => "NIS_SERVER") && (scalar @{$Options_hash->{NIS_SERVER}} >= 1) ) {
        foreach my $nis_server ( @{$Options_hash->{NIS_SERVER}} ) {
            if ( !grep /$nis_server/, @{$Nis->{servers}} ) {
                $Trace->comment(msg => Dumper($Nis));
                $Trace->fatal(result => "CONFIG",
                    msg => "Discovered NIS Servers for NIS Domain '$Nis->{domain}' (@{$Nis->{servers}}) differ from specified NIS Servers (@{$Options_hash->{NIS_SERVER}})");
            }
        }
        # we need to overwrite which NIS servers are in list because it could contain more NIS servers than
        # were passed in, we only CONFIG out if the NIS Servers that we expect to be there aren't more is OK.
        $Nis->{servers} = $Options_hash->{NIS_SERVER};
    }
    $self->{Nis} = $Nis;
    $Trace->trace(msg => "Using Nis=" . Dumper($Nis));

    $Trace->exit();
}

sub _pick_aggregates {
    my $self = shift;
    my %params = @_;
    my $min_size = 200; # 200MB

    $Trace->entry(msg=>join(" ", @_));

    # skip for infinitevolume
    if (($self->{volume_hash}) && ($self->{volume_hash}{'volume-style'} eq "infinitevol")) {
            my $tmp_vol=$Self->{volume_hash}{'volume'}."_ns";
            my @ns_vol=$Server->volume_show('volume'=>$tmp_vol);
            push(@Aggregates,$Server->storage_aggregate_show('aggregate'=>$ns_vol[0]->{'aggregate'}));
	    # this code is now pushed down into Server.pm and friends
	    $Trace->exit(msg=>join(" ", map $_->{aggregate}, @Aggregates));

    } else {
        # find the cluster aggregates
        if (my @clustered = $Server->get_clustered_nodes()) {

            # check if the volume already exists. If so, there are some attributes that should be passed into
            # pick_aggregates()
            if ($self->{volume_hash}) {
                $params{voltype} = $self->{volume_hash}{'volume-style'};
            }
            push(@Aggregates, $Server->pick_aggregates(state=>"online", avail_size=>$min_size, 
                                                       flexcache_involved=>$self->{flexcaches_involved},
                                                       topology=>$Test_topology->{cluster}, 
                                                       %params,
                                                       -Server => $clustered[0]));
        }
        foreach my $node ($Server->get_legacy_nodes()) {
            push(@Aggregates, $Server->pick_aggregates(state=>"online", avail_size=>$min_size, 
                                                       flexcache_involved=>$self->{flexcaches_involved},
                                                       topology=>$Test_topology->{node}{$node}, %params,
                                                       -Server => $node));
        }
	$Trace->exit(msg=>join(" ", map $_->{aggregate}, @Aggregates));

    }
}

sub _pick_vifs_and_vserver {
    my $self = shift;
    my %params = @_;

    $Trace->entry(msg=>join(" ", @_));
    # Pick a random vserver from the discovered list to be used as the default
    # If a vserver is not specified by the user, look for ones that are not private.
    # If a vserver list is specified by the user then verify that they exist.

    my @available_vservers;
    if(($Options->get(option=>"CIFS_VSERVERS_ONLY")) == 1) {
        my %cifs_vservers;
        foreach (@{$Discovered_topology->{cluster}{cifs_server}}) {
            $cifs_vservers{$_->{vserver}} = "cifs vserver";
        }
	foreach (@{$Discovered_topology->{cluster}{vserver}}) {
            if (exists($cifs_vservers{$_->{vserver}})) {
	        push(@available_vservers,$_);
            }
        }
    } else {  
        @available_vservers = @{$Discovered_topology->{cluster}{vserver}};
    }
 
    push(@available_vservers, @{$Test_topology->{cluster}{vserver}}) if (defined @{$Test_topology->{cluster}{vserver}});

    my $vserver_interfaces; # the Classic style of specifying interfaces with the vfiler
    my %vservers;

    # Handle vserver role to type value change (storage to cluster to data)
    my $role = "cluster|data";
    if ( $Server->get_type() == $Server->ONTAP_NG() ) { # only do this when it is an NG system
        # if we don't have the "vserver-role-change" we need to switch back to "storage" for the role
        if ( !$Server->has_change(change => "vserver-role-change") ) {
            $role = "storage";
        }
    }

    foreach my $vserver (@available_vservers) {
        # if we don't have a vserver, then we need to pick a non-private one, which can serve data.  this is
        # made a lot more complicated because in BR a cserver vserver has "data" interfaces, but it can't
        # serve data.  so we need to look to see if the vserver has a role, if it does that role must be
        # "storage" and nothing else.  so this if has gotten ugly because of that.
	if ((!$Options_hash->{vserver} && (($vserver->{vserver} !~ /^p_/) && (!exists $vserver->{role} || ($vserver->{role} =~ /$role/)))) ||
	    ($Options_hash->{vserver} && $vserver->{vserver} eq $Options_hash->{vserver})) {
            $Trace->debug(msg => "Adding vserver '$vserver->{vserver}' to the list of valid vservers");
	    $vservers{$vserver->{vserver}} = $vserver;
	    # set the vserver_interfaces flag if the interfaces are specified in the vserver definition
	    if (!$vserver_interfaces && $vserver->{interfaces} && scalar @{$vserver->{interfaces}}) {
		$vserver_interfaces = 1;
		$Trace->debug(msg=>"detected interfaces in the vserver definition\n");
	    }
        } else {
            $Trace->debug(msg => "Rejected vserver:\n" . Dumper($vserver));
        }
    }

    if ($Options_hash->{vserver} && !scalar keys %vservers) {
	$Trace->fatal(result=>"CONFIG", msg=>"Could not find the vserver '$Options_hash->{vserver}' that you specified\n".Dumper(\@available_vservers));
    } elsif (!scalar keys %vservers) {
	$Trace->fatal(result=>"CONFIG", msg=>"Could not find any non-private vservers roled '$role' to use\n".Dumper(\@available_vservers));
    }

    # Now that a list of available vservers has been established, the interface requirements need to be examined.
    # The discovered list is looked at separately from the Topology because the status needs to be examined. For each
    #  entry in this list, skip private entries. Also, if interfaces have been specified, then skip an entry if it is
    #  not in the available list. Finally, skip any entries that are not data vifs and not up.
    my $topology_interfaces = (defined($Test_topology->{cluster}{network_interfaces}) && scalar @{$Test_topology->{cluster}{network_interfaces}});
    my $param_interfaces = (defined($Options_hash->{DATA_INTERFACES}) && scalar @{$Options_hash->{DATA_INTERFACES}});
    my @interfaces;
    my @servers = $Server->servers();

    my @discovered_vifs = @{$Discovered_topology->{cluster}{network_interfaces}};
    foreach my $node (keys %{$Discovered_topology->{node}}) {
        push(@discovered_vifs, @{$Discovered_topology->{node}{$node}{network_interfaces}});
    }

    VIF:
    foreach my $vif (@discovered_vifs) {
        $Trace->trace(msg=>"check vif '$vif->{address}' in vserver '$vif->{server}'");
        if ($vif->{vif} =~ /^p_/) { # this is a private vif
	    $Trace->trace(msg=>"skipping vif '$vif->{address}' because it is private");
	    next;
	}
	if ($topology_interfaces && (!grep($vif->{address} eq $_, @{$Test_topology->{cluster}{network_interfaces}}))) {
	    $Trace->trace(msg=>"skipping vif '$vif->{address}' because it is not in the Test_topology");
	    next;
	}
	if ($param_interfaces && (!grep($vif->{address} eq $_, @{$Options_hash->{DATA_INTERFACES}}))) {
	    $Trace->trace(msg=>"skipping vif '$vif->{address}' because it is not in the DATA_INTERFACES options");
	    next;
	}

	# skip any non-routable addresses such as 192.168.0.0 unless it is in the DATA_INTERFACES param
	if (($vif->{address} =~ /^192/) && $Options_hash->{SKIP_192_ADDRESSES} &&
	    (!$param_interfaces || ($param_interfaces && !grep($vif->{address} eq $_, @{$Options_hash->{DATA_INTERFACES}})))) {
	    $Trace->trace(msg=>"skipping vif '$vif->{address}' because it is not in the DATA_INTERFACES param and appears to be non-routable");
	    next;
	}

        if ($Options_hash->{CACHE_TRAFFIC}) {
            # If the vif has a current-filer set and that filer is a cache, then mark this vif as living on a cache.
            # This info will be used when filtering the vifs.
            # This loop will also skip vifs which are not currently on one of the specified filers
            if ($vif->{"curr-filer"}) {
                foreach my $filer (split(/,/, $vif->{"curr-filer"})){
                    if (!grep( {$_ eq $filer } @servers)) {
                        next VIF;
                    }
                    if ($Server->is_cache(-Server=>$filer)) {
                        $vif->{flexcache} = 1;
                        $Trace->trace( msg=>"marking $vif->{address} as a flexcache vif");
                    }
                }
            }
        }
        if ( $vif->{role} eq "data" && $vif->{"status-admin"} eq "up" ) {
            $Trace->debug(msg => "Adding vif '$vif->{address}' to interfaces list");
            push(@interfaces, $vif);
        }
    }

    # Prepend the Topology entries to any entries that matched the previous criteria since a vserver might be specified.
    # Look at each entry and find the first interface with a valid vserver. This vserver will be picked as the default.
    #  Find any other interfaces on this vserver and add them all to the master list.
    if ($topology_interfaces) {
	foreach my $interface (@{$Test_topology->{cluster}{network_interfaces}}) {
	    unshift(@interfaces, $interface) if (!grep($interface eq $_, @interfaces));
	}
    }

    # if interfaces were specified as nate params, verify that they were all found
    if ($param_interfaces) {
	my @failures;
	foreach my $interface (@{$Options_hash->{DATA_INTERFACES}}) {
	    if (!grep($interface eq $_, @interfaces)) {
		push(@failures, $interface);
	    }
	}
	if (!scalar @failures) {
	    $Trace->fatal(result=>"CONFIG", msg=>"Did not find the following data_interfaces that you specified:".
			  join(":", @failures)."\n");
	}
    }

    # if the interfaces are specified in the vserver definition then loop through the vserverss
    #  and find the first one with interfaces that matched the previous criteria
    # this is an NG-style definition. Look for the vserver in the vif definitions
    if ($vserver_interfaces) {
	$Trace->debug(msg=>"Look for interfaces in the vserver hashes");
	# save the IPs of the vifs since that is the information available in the
	#  vserver/vfiler definition
	my @ips;
	foreach my $vif (@interfaces) {
	    push(@ips, $vif->{address});
	}
	foreach my $vserver (keys %vservers) {
	    # skip this one if a vserver has been found already
	    next if ($Virtual_server && $vserver ne $Virtual_server->{vserver});
	    # check out each of the IP in this definition and see if it
	    #  exists in the IP list for this vserver/vfiler.
	    foreach my $interface (@{$vservers{$vserver}{interfaces}}) {
		if (grep($interface eq $_, @ips)) {
		    # this is the vserver/vfiler to use
		    push(@Data_interfaces, $interface);
		    $Virtual_server = $vservers{$vserver} if (!$Virtual_server);
		}
	    }
	}
    } else {
        # This is really a Classic vfiler definition
	$Trace->trace(msg=>"Look for vservers in the interfaces hashes");
	foreach my $vif (@interfaces) {
	    my $vserver = $vif->{server};

            # if the vif is a flexcache vif, we want to skip this because we don't care what vserver
            # we are using on the flexcache.  TestManager doesn't currently let you pick the vserver
            # on the flexcache.
            if (!$vif->{flexcache}) {
                next if (!$vservers{$vserver}); # this is not a valid vserver to use

                # found a valid vserver. Now, if a vserver has not been picked, pick this one
                $Virtual_server = $vservers{$vserver} if ($vservers{$vserver});
                next if ($Virtual_server && $vserver ne $Virtual_server->{vserver}); # this is not the right one
            }
            push(@Data_interfaces, $vif->{address});
	}
    }

    # catch the case where we couldn't find a virtual server or data_interfaces
    if ( ! $Virtual_server || ! scalar(@Data_interfaces)) {
	$Trace->fatal(result=>"CONFIG", msg=>"Could not find a virtual server ($Virtual_server) or data interfaces (".scalar(@Data_interfaces).")\n".Dumper(\%vservers)."\n".Dumper(\@interfaces).Dumper($Discovered_topology));
    }

    if ($param_interfaces) {
	@Data_interfaces = @{$Options_hash->{DATA_INTERFACES}};
    }
    $Trace->exit(msg=>"vserver: $Virtual_server->{vserver}, interfaces: ".join(" ", @Data_interfaces));
}

# This routine sets up the param hashes that will define the params to be passed to each client instance
sub setup_test_params {
    my $self = shift;
    my %params = @_;

    $Trace->entry();

    # unset the CLIENTFILE param because it causes a breakage on windows clients and should not be
    # needed any more
    $self->set(CLIENTFILE=>undef);

    # for now, TM does not know how to handle distributing more than one volume as an input parameter
    if (scalar(@{$Options_hash->{volume}}) > 1) {
	push(@{$self->{warnings}}, "Can't handle multiple volumes passed as a param yet".Dumper $Options_hash->{volume});
    }

    # for now, TM does not what to do if we have a multiple volume layout, a single volume input parameter and more than one client
    if (($self->{volume_layout} eq PER_INSTANCE()) && (scalar(@{$Options_hash->{volume}}) == 1) &&
	scalar(keys %{$self->{clients}}) > 1) {
	$Trace->fatal(msg=>"Can't handle a volume passed as a param when there are multiple clients specified", result=>"SCRIPT");
    }

    # the options for the test params.
    # if test_layout is single, then we don't need to setup info for each client.
    # The assumption is that the test will be doing everything necessary to generate
    if ($self->{test_layout} eq PER_INSTANCE()) {
	foreach my $client (keys %{$self->{clients}}) {
	    while (my ($instance, $volume_list) = each %{$self->{clients}{$client}}) {
		my ($volume_name, $share);

		# build a hash of all of the current test options we want to pass to the client
		my $classic_test_options = $self->_create_test_hash();
		$classic_test_options->{CLIENT} = [$client];
		$classic_test_options->{CLIENT_INSTANCE} = $instance;

		# pass volume/share information along to each test
		my $volumesref = [];
		my $sharesref = [];
		my $volume_share = [];
		if (scalar (@$volume_list)) {
                    foreach my $volume_pairing (@$volume_list) {
                        ($volume_name, $share) = @$volume_pairing;
                        push(@$volumesref, $volume_name);
                        $share = $self->convert_share(volume=>$volume_name, share=>$share);
                        push(@$sharesref, $share);
                        push(@$volume_share, "${volume_name}-$share");
                    }
		} else {
		    ($volumesref, $sharesref) = $self->_find_volume_for_client_instance($classic_test_options, $client, $instance);
		    $share = $sharesref->[0];
		    if ($#$volumesref != $#$sharesref) {
			$Trace->fatal(msg=>"the number of volumes and shares does not match for $client:$instance", result=>"SCRIPT");
		    }
		    my $i = 0;
                    foreach my $vol (@$volumesref) {
			my $share = $sharesref->[$i++];
			$share = $self->convert_share(volume=>$vol, share=>$share);
			push(@$volume_share, "${vol}-$share");
		    }
		}

		$classic_test_options->{volume} = $volumesref; # expects an array ref

                $classic_test_options->{DATA_INTERFACES} = [$self->filter_vifs(volume=>$volumesref->[0])];

                $share = $sharesref->[0];
		$classic_test_options->{share} = $share;
		if ($self->{directory_per_instance}) {
		    $classic_test_options->{share} .= "/${client}_${instance}_".$self->testcase_string();
		}
		$classic_test_options->{VOLUME_SHARE} = $volume_share;  # expects an array ref

		my $mnt_dir = "$self->{testdir}.$instance";
		$Trace->debug(msg=>"mnt set to $mnt_dir on $client");
                if ($self->{Cifs_Clients}{$client}) {
                    # maybe this should be done elsewhere
                    $share =~ s!/.admin!!;
                    # Specify the Cifs_Server for now; client_setup might change it to the IPs
                    $mnt_dir = "//$Cifs_Server->{$Cifs_name_key}/${Cifs_Share}$share";
                    if ($self->{directory_per_instance}) {
                        $mnt_dir .= "/${client}_${instance}_".$self->testcase_string();
                    }
                }
		$Trace->debug(msg=>"mnt_dir set to $mnt_dir on $client");
		$classic_test_options->{testdir} = $mnt_dir;

		$self->{test_options}{$client}{$instance} = $classic_test_options;
		# save the options in the localhost key if it was set. It will be needed in
		#  _call_test_code()
		if ($self->{localhost}) {
		    $self->{localhost} = $classic_test_options;
		}
	    }
	}
    }
    if ($self->{test_layout} eq SINGLE() || $self->{test_block_layout} eq SINGLE()) {
	# set the volume list so that it is available to the test setup/cleanup block
	my @volumes;
	foreach my $volumeh (@{$Test_topology->{cluster}{volume}}) {
	    push(@volumes, $volumeh->{volume});
	}
	$self->set(volume =>\@volumes) if (@volumes);
	$self->set(vserver => $Virtual_server->{vserver});
    }
    $Trace->exit(msg=>"");
}

sub filter_vifs
{
    my $allowed_type = ($Self->{check_topology}) ? {type=>SCALAR} : {TYPE=>SCALAR | UNDEF};
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'TestManager' } ],
					 {
					     volume => $allowed_type,
					 }
					 );
    $Trace->entry(msg => "Filtering vifs for $params{volume}");

    if (! $self->{check_topology} ) {
        $Trace->exit();
        return ();
    }
    #
    # Get the volume we're mounting by name
    #
    my ($volume) = grep({ $_->{volume} eq $params{volume} } @{$Test_topology->{cluster}{volume}});

    if (!defined($volume->{aggregate})) {
        my $volume_info = $Server->volume_show(volume => $volume->{volume});

        $volume->{aggregate} = $volume_info->[0]->{aggregate};
    }

    $Trace->debug(msg => "Got volume: ".Dumper($volume));

    #
    # Get the aggregate holding that volume
    #
    my ($aggr) = $Server->search_topology(cmd_node => "aggregate",
					  key_pairs => [ "aggregate:$volume->{aggregate}",
							 "filers:$volume->{'-Filer'}"],
					  );
    $Trace->debug(msg => "Got aggregate ".Dumper($aggr));
    $Trace->debug(msg => "Network interfaces to choose from: ".join(", ", @Data_interfaces));
    $Trace->debug(msg => "Vifs configured: ".join(", ", map({ $_->{vif}.": ".$_->{address} } @{$Discovered_topology->{network_interfaces}})));
    #
    # Get the vif-level details of the network interfaces we're using
    #

    # create a temporary combined vif topology that we will search
    my @topology_vifs;
    if ($Discovered_topology->{cluster}{network_interfaces}) {
        @topology_vifs = @{$Discovered_topology->{cluster}{network_interfaces}};
    }
    foreach my $node (keys %{$Discovered_topology->{node}}) {
        if ($Discovered_topology->{node}{$node}{network_interfaces}) {
            push(@topology_vifs, @{$Discovered_topology->{node}{$node}{network_interfaces}});
        }
    }
    my @vifs;
    foreach my $interface (@Data_interfaces) {
        foreach my $vif ( (@{$Test_topology->{cluster}{network_interfaces}}, @topology_vifs) ) {
            if ((($vif->{address} eq $interface) && !grep({$_->{address} eq $interface} @vifs)) && (grep( {$_ eq $vif->{'curr-filer'} } $Server->servers()))) {
                push (@vifs, $vif);
            }
        }
    }
    $Trace->debug(msg => "Vifs selected for filtering: ".join(", ", (map( { $_->{address} } @vifs))));

    if (!defined($volume) || !defined($aggr) || !scalar(@vifs)) {
        $Trace->fatal(msg => "One or more problems occured when filtering vifs:\n".
            "The volume $params{volume} was ".(defined($volume) ? "found" : "not found")."\n".
            "The volume's containing aggregate '$volume->{aggregate}' was".(defined($aggr) ? "" : " not" )." found.\n".
            "There are ".(scalar(@vifs) ? scalar(@vifs) : "no")." vifs available.", result=>'FATAL');
    }

    # find the list of nodes hosting the volume
    my @volume_nodes;
    if (ref($aggr->{filers}) eq "ARRAY"){
	@volume_nodes = @{$aggr->{filers}};
    } elsif($aggr->{filers} =~ /,/) {
        @volume_nodes = split(/,/, $aggr->{filers});
    } else {
	@volume_nodes = ($aggr->{filers});
    }

    # Run the following command on one of the nodes hosting the volume so that the correct filer_vifs() is called
    my @filtered_vifs = $Server->filter_vifs(-Server => $volume_nodes[0],
                                             volume => $volume,
					     interfaces => \@vifs,
					     volume_nodes => \@volume_nodes,
					     traffic => undef, # default to filter_vifs()
					     topology => $Test_topology,
                                             flexcache_involved=>$self->{flexcaches_involved},
					     );

    if (! scalar @filtered_vifs) {
        $Trace->fatal(msg=>"No vifs meet the criterea");
    }

    $Trace->exit(msg=>"return @filtered_vifs");
    return @filtered_vifs;
}

sub setup_mount_info {
    my $self = shift;

    $Trace->entry();
    my %client_topology;
    my $mount_number = 0;
    foreach my $client (keys %{$self->{clients}}) {
	my @mounts;
	while (my ($instance, $volume_list) = each %{$self->{clients}{$client}}) {
	    if (scalar(@$volume_list) == 0) {
		$Trace->fatal(result=>"SCRIPT", msg=>"$client $instance has undefined volume information");
	    }
            #            foreach my $volume_pairing (@$volume_list) {
            my $volume_pairing = $volume_list->[0];
            my ($volume_name, $share) = @$volume_pairing;
            my %mount;
            # build up the information for mount/umount

            my @interfaces = $self->filter_vifs(volume=>$volume_name);

            # pick vifs in a round-robin fashion rather than random
            $mount{ip} = $interfaces[ $mount_number++ % (scalar @interfaces) ];
            $mount{volume} = $volume_name;
	    $Trace->trace(msg=>"Finding testdir for client: $client instance: ".Dumper($self->{test_options}{$client}{$instance}));
            if (my $testdir = $self->{test_options}{$client}{$instance}{testdir}) {
                $mount{mnt} = $testdir;
            } else {
	        $Trace->trace(msg=>"No testdir defined. Using default");
                $mount{mnt} = $self->{testdir};
            }
	    $Trace->trace(msg=>"Testdir to use mount is : $mount{mnt} ");

            $mount{options} = $Options_hash->{MOUNT_OPTIONS} || "hard,intr";
            $mount{chmod} = $Options_hash->{MNTPNT_MODE};
            $mount{share} = $share;
            $mount{instance} = $instance;
            push(@mounts, \%mount);
            #            }
         }
        push(@{$client_topology{$client}{mounts}}, @mounts);
   }
    $Trace->exit(msg=>"");
    return \%client_topology;
}

sub update_test_topology {
    my $self = shift;
    my %striped_aggr_limits;
    my $mdv_tax;
    $Trace->entry();

    my $orig_server;
    if (my @clustered = $Server->get_clustered_nodes()) {
        $orig_server = $Server->set_execution_server(-Server=>$clustered[0]);
    }

    my $create_ruleset = undef;
    # set the initial ownership attributes of the volume to that of the caller. The mount call will
    #  then perform a chmod 777 to allow anyone to write into it
    foreach my $vs_hash (@{$Test_topology->{cluster}{vserver}}) {
	if ($vs_hash->{vserver} !~ /^p_/) {
	    # fail if the vserver is not private and not specified as an input param
	    if ($Options_hash->{vserver} && ($vs_hash->{vserver} ne $Options_hash->{vserver})) {
		$Trace->fatal(msg=>"The virtual server '$vs_hash->{vserver}' should be of the form p_*\n", result=>"CONFIG");
	    }
	}
	my $aggr_hash = $Aggregates[int rand(scalar @Aggregates)];
	$vs_hash->{aggregate} = $aggr_hash->{aggregate} if (!$vs_hash->{aggregate});
        $vs_hash->{comment} = '"created by TestManager"';
    }
    if (exists($Test_topology->{cluster}{network_interfaces}) && scalar(@{$Test_topology->{cluster}{network_interfaces}})) {
	# set port in vif creation
	# set ssid in vif creation
	my %ports;
	my @servers = $Server->servers();
	foreach my $server (@servers) {
	    $ports{$server} = [];
	    foreach my $port_hash (@{$Discovered_topology->{network_ports}}) {
		push(@{$ports{$server}}, $port_hash->{port}) if ($port_hash->{role} eq "data" &&
								 $port_hash->{link} eq "up");
	    }
	}
        my $vif_number = 0;
	foreach my $vif_hash (@{$Test_topology->{cluster}{network_interfaces}}) {
	    $vif_hash->{"home-filer"} = $servers[ $vif_number % (scalar @servers)] if (!$vif_hash->{"home-filer"});
            $vif_hash->{"curr-filer"} = $vif_hash->{"home-filer"};
	    $vif_hash->{server} = $Virtual_server->{vserver} if (!$vif_hash->{server});
	    my @ports = @{$ports{$vif_hash->{"home-filer"}}};
	    $vif_hash->{"home-port"} = $ports[$vif_number++ % (scalar @ports)] if (!$vif_hash->{"home-port"});
	}
    }
    $Trace->debug(msg=>"Aggregates Hash:".Dumper(\@Aggregates));

    # Some users will pass in a striped aggregate and forget to set VOLUME_STYLE to striped.
    # That should be OK, so set a volume_style variable based on the param and a possible
    #  user-specified aggregate. This "fix" is only a short-term fix; it only works because
    #  pick_aggregates skips the checks on voltype when a list of aggrs is passed in. Otherwise
    #  it would still be only picking the flex aggregates
    my $volume_style = "flex";
    if ($Options_hash->{VOLUME_STYLE} eq "striped") {
        $Trace->trace(msg=>"Setting volume_style to striped since VOLUME_STYLE was set to striped");
        $volume_style = "striped";
    }
    # this may be broken if someone passes a mixture of aggregate types in....
    if (grep($_->{voltype} eq "striped", @Aggregates)) {
        $volume_style = "striped";
        $Trace->trace(msg=>"Setting volume_style to striped since one of the aggregates is striped");
    }

    # Create a hash to store the minimum volume size allowed by the aggregates
    # Or calculate the minimum if option VOL_STRIPE_COUNT is provided.
    # Set the mdv_tax variable to use for calculation of volume sizes
    if ($volume_style eq 'striped') {
	if ($Server->striped_vol_create_size_includes_mdv()){
	    #Find out if the MDV is included in Volume size.
	    #This is here for older versions that took vol size / (DV_COUNT+MDV).
	    #To determine the size for each member volume.
	    $mdv_tax = 1;
	    $Trace->trace(msg=>"MDV is included in size calculations");
	}
	else {
	    $mdv_tax = 0;
	}
        for my $aggr (@Aggregates) {
            # skip non-striped aggregates
            next if ($aggr->{voltype} ne "striped");
            # Keep the minsize around to check against size requests
            # If a stripe count is specified, calculate the minsize from it.
            $striped_aggr_limits{$aggr->{aggregate}}{minperm}  = $$Options_hash{'VOL_STRIPE_COUNT'} ? ($$Options_hash{'VOL_STRIPE_COUNT'} + $mdv_tax) * (1.2 * 1024) : $Server->find_min_max_volume_size($aggr->{aggregate}, 'minperm');
            if ($striped_aggr_limits{$aggr->{aggregate}}{minperm} =~  /^(\d+\.?\d*)/) {
                $striped_aggr_limits{$aggr->{aggregate}}{minperm} = int($1);
		$Trace->debug(msg=>"Aggregate: $aggr->{aggregate} minsize: $striped_aggr_limits{$aggr->{aggregate}}{minperm}");

            } else {
                $Trace->fatal(result=>"CONFIG", msg=>"Couldn't determine minimum size" );
            }

            $Trace->debug(msg=>"Aggregate: $aggr->{aggregate} availsize: $aggr->{availsize}");
        }
    }

    if (exists($Test_topology->{cluster}{volume}) && scalar(@{$Test_topology->{cluster}{volume}})) {
        my $volume_number = 0;
	foreach my $vol_hash (@{$Test_topology->{cluster}{volume}}) {

            #
            # modify the volume name to comply with filer naming rules
            #
	    $vol_hash->{volume} = $self->_normalize_volume_name(volume=>$vol_hash->{volume});

            # see if the volume already exists in the cluster
            my $existing_volume;
            eval { $existing_volume = $Server->volume_show(volume => $vol_hash->{volume},
                                                           vserver => $Virtual_server->{vserver}) };
            # make sure that we have a volume array and that the volume hash is a real one.
            # For 7G, a volume_hash gets generated and is not complete....
            if ( $existing_volume && $existing_volume->[0]{volume}) {
                $existing_volume = $existing_volume->[0];
                my $aggr_hash;
                if ($existing_volume->{'volume-style'} eq "infinitevol") {
                    my $ns_vol=$existing_volume->{'volume'}."_ns";
                    my @tmp_vol=$Server->volume_show(volume=>$ns_vol);
                    $aggr_hash=$Server->storage_aggregate_show(aggregate => $tmp_vol[0]->{aggregate});
                    $aggr_hash=@$aggr_hash[0];
                    $Trace->comment(msg=>"Aggregate Hash ".Dumper $aggr_hash);
                } else {             
                    # find info on the hosting aggregate
                    ($aggr_hash) = grep($_->{aggregate} eq $existing_volume->{aggregate}, @Aggregates);
                }
                if (ref($aggr_hash) ne "HASH") {
                    $Trace->comment(msg=>Dumper $existing_volume);
                    $Trace->comment(msg=>Dumper \@Aggregates);
                    $Trace->comment(msg=>Dumper $aggr_hash);
                    $Trace->fatal(msg=>"Could not find the aggregate hosting volume '$existing_volume->{volume}'");
                }
                if ($existing_volume->{'volume-style'} eq "infinitevol") {
                    foreach my $key (keys %{$existing_volume}) {
                    #Don't cache volume params that are "-" or blank
                        $vol_hash->{$key} = $existing_volume->{$key} if ($existing_volume->{$key} !~ /^-$|^$/);
                        $vol_hash->{$key} = $aggr_hash->{$key} if ($aggr_hash->{$key} !~ /^-$|^$/);
                        $vol_hash->{-Filer} = $aggr_hash->{filers};
                    }
                } else {
                    foreach my $key (keys %{$existing_volume}) {
                    #Don't cache volume params that are "-" or blank 
                        $vol_hash->{$key} = $existing_volume->{$key} if ($existing_volume->{$key} !~ /^-$|^$/);
                        $vol_hash->{-Filer} = $aggr_hash->{filers};
                    }
                }
                next;
            }
       
            # Set Volume mode bits if not defined in volume hash,
	    # is supported by the filer and not ntfs security style
            if (!defined($vol_hash->{'unix-permissions'}) && $Options_hash->{MNTPNT_MODE} && $Server->supports_mode_in_volume_create() && $Options_hash->{VOL_SECURITY_STYLE} !~ /ntfs/i) {
                $vol_hash->{'unix-permissions'} = $Options_hash->{MNTPNT_MODE};
            }
	    # modify the maxdirsize of the volume if necessary
	    if (!defined($vol_hash->{'maxdir-size'}) && $Options_hash->{VOL_MAXDIR_SIZE}) {
		$vol_hash->{'maxdir-size'} = $self->size_in_mb(size=>$Options_hash->{VOL_MAXDIR_SIZE})."MB";
	    }
	    delete $vol_hash->{'maxdir-size'} if (!$vol_hash->{'maxdir-size'});

	    # find the current volume size
	    if (!defined($vol_hash->{size})) {
		$vol_hash->{size} = $Options_hash->{VOL_SIZE}."MB";
	    }
	    my $vol_size = $self->size_in_mb(size=>$vol_hash->{size});
            $Trace->trace(msg=>"Volume '$vol_hash->{volume}' has size $vol_size (from $vol_hash->{size})");

	    # The minimum volume size is 300MB so make sure it is at least that large
            # We will adjust this again later if striped aggregates are used
	    if ($vol_size < 300) {
		$Trace->comment(msg=>"The size of volume '$vol_hash->{volume}' was bumped from $vol_size to 300 since the volume size must be at least 300MB");
		$vol_size = 300;
	    }
	    if ($volume_style eq 'striped') {
		# fill in volume stripe-width option on volume new. This option only works if the volume is created on a striped aggregate.
		if ( defined ($$Options_hash{VOL_STRIPE_WIDTH}) ) {
		    $vol_hash->{'stripe-width'} = $$Options_hash{VOL_STRIPE_WIDTH};
		}
                if(defined ($$Options_hash{VOL_STRIPE_FORMAT})) {
                    $vol_hash->{'stripe-format'} = $$Options_hash{VOL_STRIPE_FORMAT};
                }
                if( defined ($$Options_hash{VOL_STRIPE_COUNT}) ) {
		    # fill in volume stripe-count option on volume new. This option only works if the volume is created on a striped aggregate.
			$vol_hash->{'stripe-count'} = $$Options_hash{VOL_STRIPE_COUNT};
		}
                if(defined ($$Options_hash{VOL_CONCURRENCY})) {
                    $vol_hash->{'concurrency'} = $$Options_hash{VOL_CONCURRENCY};
                }
                if(defined ($$Options_hash{VOL_OPTIMIZE})) {
                    $vol_hash->{'optimize'} = $$Options_hash{VOL_OPTIMIZE};
                }
                if( defined ($$Options_hash{VOL_LOOKUP_CACHE}) ) {
                    #enable/disable DNLC cache on volume creation
		    my $lookupCache;
		    if ($$Options_hash{VOL_LOOKUP_CACHE} =~ /yes|true|enabled|enable|1/i) {
			$lookupCache = "enabled";
		    }
		    elsif ($$Options_hash{VOL_LOOKUP_CACHE} =~ /no|false|disabled|disable|0/i) {
			$lookupCache = "disabled";
		    }
		    else {
			$Trace->fatal(msg=>"Invalid lookup cache setting");
		    }
                    $vol_hash->{'lookup-cache'} = $lookupCache;
		}


            }
            if (defined ($$Options_hash{VOL_INDEX_DIR_ENABLED})) {
                # Need test privilege for now, can remove this set if no longer needed
                $Trace->trace(msg=>"Need test privilege to use INDEX_DIR_ENABLED, ".
                                     "not usable on promoted builds for now");
                eval {
                    $Server->set("privilege" => "test");
                };
                my $error = $@;
                if ($error) {
                    if ($error =~ /failed to set field/) {
                        $Trace->fatal(msg=>"Unable to go to privilege test, ".
                                            " make sure this is not a promoted build.");
                    } else {
                        $Trace->fatal(msg=>"Got unknown error when going to priv test: $error");
                    }
                }
                $vol_hash->{'index-dir-enabled'} = "false";
                if ($$Options_hash{VOL_INDEX_DIR_ENABLED}) {
                    $vol_hash->{'index-dir-enabled'} = "true";
                }
            }

	    # modify the number of files if necessary
	    $vol_hash->{files} = $Options_hash->{VOL_FILES} if (!defined($vol_hash->{files}));
	    delete $vol_hash->{files} if (!$vol_hash->{files});

	    if (defined($vol_hash->{files})) {
		# each inode needs 4K so make sure we set the volume size large enough
		my $inode_space_needed = $self->size_in_mb(size=>($vol_hash->{files} * 4096));
		if ($inode_space_needed > $vol_size) {
		    $Trace->comment(msg=>"The size of volume '$vol_hash->{volume}' was bumped from $vol_size to $inode_space_needed due to the number of files specified ($vol_hash->{files})");
		    $vol_size = $inode_space_needed;
		}
	    }

            # we should be done with all of the volume size change so update it now
            $vol_hash->{size} = $vol_size."MB";

	    # fill in missing aggregate information
	    my @usable_aggrs;
	    if ($Stages{filer_setup}) {
                # this is a hack until we have a real feature test
                my $release = $Server->get_release(format=>"cooked");
                if ($release =~ /tricky.2/i) {
                    foreach my $aggr (@Aggregates) {
                        my $size = $self->size_in_mb(size=>$aggr->{availsize}."MB");
                        $aggr->{availsize} = $size;
                    }
                }

                if ($volume_style eq 'striped') {
                    #Make sure there is enough space to create the minimum allowed size
                    @usable_aggrs = grep( { $_->{availsize}  >= $striped_aggr_limits{$_->{aggregate}}{minsize}} @Aggregates );
                    #Check if there is enough space to meet the requested size
                    @usable_aggrs = grep( { $_->{availsize} >= $vol_size } @usable_aggrs );
                }
                else {
                    foreach my $aggr (@Aggregates) {
                        if ($self->size_in_mb(size =>$aggr->{availsize}) >= $vol_size) {
                            push(@usable_aggrs, $aggr);
                        }
                    }
                }

                my $aggr_info;
                map($aggr_info .= "$_->{aggregate} => $_->{availsize}\n", @usable_aggrs);
                $Trace->trace(msg=>$aggr_info);

		if (!scalar(@usable_aggrs)) {
		    $Trace->fatal(msg => "No aggregates with sufficient space are available, ".$vol_hash->{size}." needed.");
		}
	    } else {
		@usable_aggrs = @Aggregates;
	    }

            if ($Options_hash->{CACHE_TRAFFIC}) {
                # remove the cache aggregates from the usable list.
                my @non_cache_aggrs = ();
                foreach my $aggr (@usable_aggrs) {
                    if ($aggr->{flexcache}) {
                        $Trace->debug(msg=>"Removing $aggr->{aggregate} from the list since it is on a cache");
                        next;
                    }
                    push(@non_cache_aggrs, $aggr);
                }
                @usable_aggrs = @non_cache_aggrs;
            }

            if (!scalar(@usable_aggrs) && $Stages{filer_setup}) {
		$Trace->comment(msg=>"Aggregates=".Dumper(\@Aggregates));
		$Trace->comment(msg=>"usable_aggregates=".Dumper(\@usable_aggrs));
                $Trace->fatal(msg => Dumper(\@usable_aggrs)."Could not find any usable aggregates after filtering them");
            }
            # pick aggregates in a round-robin fashion rather than random
            my $aggr_hash = $usable_aggrs[ $volume_number++ % (scalar @usable_aggrs) ];
            $aggr_hash->{availsize} -= $vol_size;
	    $vol_hash->{aggregate} = $aggr_hash->{aggregate} if (!$vol_hash->{aggregate});

            $Trace->trace(msg=>"placed volume '$vol_hash->{volume}' on aggregate '$vol_hash->{aggregate}': its size is now $aggr_hash->{availsize}");

            # The goal is to restrict the aggregate list down to those that have enough space for the volume.
            # That had been done, but now that the list is available, the volume size might need to be adjusted
            # to meet the minimum requirements for the chosen aggregate
            if ( $vol_hash->{'size'} && $volume_style eq 'striped' ) {
                # make sure the size requested is above the minimum. If it's not, warn and adjust.
                # if a size is set for the volume, then we calculate the minimum size
                # based upon a supplied stripe count, or by calling find_min_max_volume_size
                # to check the minimum size allowed by the volume

		my $minsize = $striped_aggr_limits{$aggr_hash->{aggregate}}{minperm};
                if ( $minsize =~ /^(\d+\.?\d*)/) {
                    $minsize = $1;
                } else {
                    $Trace->fatal(result=>"CONFIG", msg=>"Couldn't determine minimum size;minsize=$minsize,stripe-count=$vol_hash->{'stripe-count'}" );
                }
                my $reqsize = $self->size_in_mb(size=>$vol_hash->{'size'});
                if ($reqsize < $minsize) {
                    $minsize = ($minsize + 100)."MB"; #increase the size to avoid conversion problem on filer.
                    $Trace->comment(msg=>"The requested Volume Size of ${reqsize}MB for volume $vol_hash->{volume} is below the minimum size allowed for striped aggregate $aggr_hash->{aggregate}($minsize). Adjusting the volume size to $minsize");
                    $vol_hash->{'size'} = $minsize;
                    $vol_size = $minsize;
                }
                # This is the final check. We already have an aggregate picked but make sure that it is still big enough
                # given our new volume size
                if ($vol_size > $self->size_in_mb(size=>$aggr_hash->{availsize})) {
                    $Trace->fatal(msg=>"The aggregate $aggr_hash->{aggregate} was picked but it does not have enough space available($aggr_hash->{availsize}) after bumping the volume size up to $vol_size MB");
                }
            }

	    # fill in missing vserver information
	    $vol_hash->{vserver} = $Virtual_server->{vserver} if (!$vol_hash->{vserver});

	    # fill in missing security_style information. It defaults to "unix" unless specified.
	    if ( !$vol_hash->{'security-style'} ) {
		$vol_hash->{'security-style'} = $$Options_hash{VOL_SECURITY_STYLE} || "unix";
	    }

	    # fill in user/group information
            if ($$Options_hash{VOLUME_OWNER} eq 'self') {
                $vol_hash->{user} = $ENV{LOGNAME};
                (undef(),undef(), undef(),$vol_hash->{group}, undef(), undef(), undef(), undef(), undef()) = getpwnam($ENV{LOGNAME});
            } else {
                $vol_hash->{user} = 0 if (!$vol_hash->{user});
                $vol_hash->{group} = 0 if (!$vol_hash->{group});
            }

	    # fill in junction-path information
	    if (!$vol_hash->{'junction-path'}) {
		$vol_hash->{'junction-path'} = "/$vol_hash->{volume}";
	    }

	    # fill in an export rule if one is not specified
	    if (!$vol_hash->{ruleset}) {
		$vol_hash->{ruleset} = $Ruleset;
		$create_ruleset++;
	    }

	    # Classic needs to know which filer the aggr/volume are on
	    $vol_hash->{-Filer} = $aggr_hash->{filers};

            # Indicate that it was created by TestManager
            if ($vol_hash->{comment}) {
                $vol_hash->{comment} =~ s/'//g;
                $vol_hash->{comment} =~ s/"//g;
                $vol_hash->{comment} = '"$vol_hash->{comment}, created by TestManager"';
            } else {
                $vol_hash->{comment} = '"created by TestManager"';
            }
	}
    }
    if (exists($Test_topology->{cluster}{volume_mirror}) && scalar(@{$Test_topology->{cluster}{volume_mirror}})) {
        my $volume_number = 0;
        foreach my $vmirror_hash (@{$Test_topology->{cluster}{volume_mirror}}) {

            #
            # modify the volume name to comply with filer naming rules
            #
            $vmirror_hash->{source} = $self->_normalize_volume_name(volume=>$vmirror_hash->{source});

            # set mirror type to LS if not defined
            $vmirror_hash->{type} = "LS" if (!$vmirror_hash->{type});

            # fill in missing vserver information
            $vmirror_hash->{vserver} = $Virtual_server->{vserver} if (!$vmirror_hash->{vserver});

            # Indicate that it was created by TestManager
            if ($vmirror_hash->{comment}) {
                $vmirror_hash->{comment} =~ s/'//g;
                $vmirror_hash->{comment} =~ s/"//g;
                $vmirror_hash->{comment} = '"$vmirror_hash->{comment}, created by TestManager"';
            } else {
                $vmirror_hash->{comment} = '"created by TestManager"';
            }
        }
    }

    if ($create_ruleset) {
	push(@{$Test_topology->{cluster}{exportruleset}}, {
	    vserver => $Virtual_server->{vserver},
	    rulesetname => $Ruleset,
	});
	push(@{$Test_topology->{cluster}{exportruleset_rules}}, {
	    ruleindex => 1,
	    vserver => $Virtual_server->{vserver},
	    rulesetname => $Ruleset,
	    anon => 0,
	    protocol => "any",
	    clientmatch => "0.0.0.0/0",
	    rorule => "any",
	    rwrule => "any",
	    superuser => "any",
	});
    }

    if ($orig_server) {
        $Server->set_execution_server(-Server=>$orig_server);
    }
    $Trace->exit(msg=>"");
}

sub size_in_mb {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'TestManager' } ],
        {
            size => {type => SCALAR | UNDEF},
        }
    );

    my $vol_size;
    if (!$params{size}) {
	$vol_size = 0;
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(MB|M|GB|G|TB|T)$/) {
	$vol_size = $1;
	if ($2 =~ m/GB?/i) {
	    $vol_size *= 1000;
	} elsif ($2 =~ m/TB?/i) {
	    $vol_size *= 1000;
	    $vol_size *= 1000;
	}
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(mb|m|gb|g|tb|t)$/) {
	$vol_size = $1;
	if ($2 =~ m/gb?/i) {
	    $vol_size *= 1024;
	} elsif ($2 =~ m/tb?/i) {
	    $vol_size *= 1024;
	    $vol_size *= 1024;
	}
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(KB|K)$/) {
	$vol_size = $1 / 1000;
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(kb|k)$/i) {
	$vol_size = $1 / 1024;
    } elsif ($params{size} =~ m/^(\d+\.?\d*)b?$/i) {
	$vol_size = $1 / 1024;
	$vol_size /= 1024;
    } else {
	$Trace->fatal(msg => "Unable to convert size to megabytes. Could not match input string '$params{size}'");
    }

    if ($vol_size > int($vol_size)) {
	return (int $vol_size) + 1;
    } else {
	return int $vol_size;
    }
}

sub process_params {
    my $self = shift;

    $Trace->entry();
    # placeholder
    $Trace->exit(msg=>"");
}

sub test_setup {
    my $self = shift;

    $Trace->entry();
    $Trace->exit(msg=>"");
    return 1;
}

sub test_main {
    my $self = shift;

    $Trace->entry();
    $Trace->fatal(msg=>"Please create a test-specific test_main routine", result=>"SCRIPT");
    $Trace->exit(msg=>"");
}

sub test_cleanup {
    my $self = shift;

    $Trace->entry();
    $Trace->exit(msg=>"");
    return 1;
}

sub filer_setup {
    my $self = shift;

    $Trace->entry();
    $Server->create_topology(print_node=>undef, topology=>$Test_topology);
    $Trace->debug(msg=>"Dumper ($Test_topology)");
    # if we are in directory_per_instance mode, then we need to create the directories
    #   after the volume has been created
    if (ref($Test_topology->{directories}) eq "HASH" ) {
        # Make sure nblade replay status is updated before accessing CIFS shares
        if (scalar(keys(%{$self->{Cifs_Clients}})) > 0) {
            my $replay_show;
            my $waiting = 1;
            while (($waiting > 0) && ($waiting < 5)) {
                # Run nblade replay command
                eval {
                    $replay_show = $Server->diag_nblade_replay_show();
                };
                my $error = $@;
                if ($error =~ /This table is currently empty/i) {
                    # Replay is currently running
                    $Trace->comment(msg => "Waiting on nblade replay to finish...");
                    $waiting++;
                    sleep(3);
                } elsif ($error) {
                    # Feature not supported
                    $waiting = 0;
                } else {
                    # Make sure it succeeded
                    if ($replay_show->[0]->{result} !~ /succeeded/i) {
                        $Trace->fatal(msg => "[diag nblade replay] failed: ".Dumper($replay_show));
                    }
                    $waiting = 0;
                }
            }
        }

	my @clients = keys %{$self->{clients}};

	foreach my $volume (keys %{$Test_topology->{directories}}) {
	    my $export = $self->find_volume_share(volume=>$volume);
            my $filerip = ($self->filter_vifs(volume=>$volume))[0];
	    my @directories = @{$Test_topology->{directories}{$volume}};
            my $client_idx = 0;
            my $client;
	    my $mntpnt;

            # Pick a number 500 directories per client as optimal value
            my $dirs_per_client = scalar(@directories) / scalar(@clients);
            $dirs_per_client = ($dirs_per_client > 500) ? $dirs_per_client+1 : 500;
            while (my @dirs = splice(@directories, 0, $dirs_per_client)) {
                if($client_idx >= scalar(@clients)) {
                    $client_idx = 0;
                }
                $client = $clients[$client_idx];

                if ($self->{Cifs_Clients}{$client}) {
                    $export =~ s/.admin//;
                    $mntpnt = "//$Cifs_Server->{$Cifs_name_key}/${Cifs_Share}$export";
		    # Changing the mntpoint to use the Ip address instead of the WINS name to see
		    # impact on CTL PB runs
		    $mntpnt = "//${filerip}/${Cifs_Share}$export";
                } else {
                    $mntpnt = "/t/${filerip}_${volume}_".$self->testcase_string()."_filer_setup";
                }

		# create directories on the client
                my $params = {
                    DIRS       => join(',', @dirs),
                    # change : to ; to avoid conflict with IPv6 addr format
                    MOUNT_LIST => $filerip.";".$export.";".$mntpnt.";rw;777",
                };
                $Trace->comment(msg=>"Run tools/TestManager/filer_directory_setup.thpl on $client to create ".join(",", @dirs));
                $Remote->execute(host=>$client,
                                 nate_params=>$params,
                                 cmd=>"tools/TestManager/filer_directory_setup.thpl",
                                 runid=>"filer_setup_directories",
                                 handler=>sub {});
                $client_idx++;
            }
	    my $fail_result = $Remote->wait(timeout=>300);

            if ($fail_result) {
                die $fail_result;
            }
	}

    }
	if ($Options_hash->{FILER_SETUP_POSTCHECK}) {
		foreach my $app (@{$Options_hash->{FILER_SETUP_POSTCHECK}}) {
			my $testcase;
			my $script;
			($script,$testcase) = split /\s+/,$app;
			if (!defined $testcase) {
				$testcase = "TESTCASE=testmanager_setup_check";
			}
			_checkers(testcase=>"$testcase", script=>$script);
		}

	}
    $Trace->exit(msg=>"");
}

sub _checkers {
    $Trace->entry();
	my %params = (
		testcase => undef,
		script  => undef,
        @_,
    );
	my @pc_params;
	push @pc_params, "$params{testcase}";
	my $num_vols = scalar @{$Test_topology->{cluster}->{volume}};
	my @volumes;
	for (my $i=0; $i < $num_vols; $i++) {
	    my $volume_name = $Test_topology->{cluster}->{volume}->[$i]->{volume};
	    push @volumes, $volume_name;
	}
    my $vols = join (",",@volumes);
	push @pc_params, "VOLUME=$vols";
	$params{testcase} =~ /TESTCASE=(\w+)/;
	my $tc = $1;
	$params{script} =~ /\/(\w+)\.\w+/;
	my $thpl = $1;
	my $runid = $thpl."_".$tc;
	subtest($params{script}, @pc_params,'-runid',$runid);

    $Trace->exit(msg=>"");
}

sub client_setup {
    my $self = shift;
    $Trace->entry();
    my $errors;

    $Trace->debug(msg=>Dumper($Test_topology->{clients}));
    # if the AUTOMOUNT flag is enabled then just return since the mounts will
    # occur with each iteration
    if ($Options_hash->{AUTOMOUNT}) {
        $Trace->comment(msg=>"STAGE client_setup skipped since AUTOMOUNT is enabled");
        $Trace->exit(msg=>"AUTOMOUNT enabled");
        return;
    }

    foreach my $client (keys %{$Test_topology->{clients}}) {

        # Authenticate CIFS user
        if ($self->{Cifs_Clients}{$client}) {
            if ($Options_hash->{RAND_MULTI_USER}) {
                # Determine which lab we are running in
                my $global_params = NATE::ParamSet->new(global=>1);
                my @all = $global_params->walk();
                my $labname = $global_params->get("LAB_NAME");
                $labname = $global_params->get("LAB") if (not defined($labname));

                my @users;
                if (defined($labname)) {
                    if (defined_lab(labname => $labname)) {
                        # Get a valid username from TestUsers.pm
                        @users = get_users(labname => $labname, count => 1, random => 1);
                        my $username = $users[0];
                        $Trace->comment(msg => "Running as user: $username") if (defined($username));

                        # Get parameters
                        my $cifs_domain = $global_params->get("CIFS_DOMAIN");

                        # Find UNC paths to authenticate
                        my %UNCpaths;
                        foreach my $opt_iter (keys(%{$self->{test_options}{$client}})) {
                            my $mntpnt = $self->{test_options}{$client}{$opt_iter}{testdir};
                            $mntpnt =~ tr{/}{\\};
                            my ($UNCbegin, $UNCend) = split(/$client/, $mntpnt);
                            my $separator = $/;
                            $/ = "\\";
                            chomp($UNCbegin);
                            $/ = $separator;
                            $UNCpaths{$UNCbegin} = 1;
                        }

                        # Un-authenticate UNC path if it is already authenticated
                        foreach my $UNCpath (keys(%UNCpaths)) {
                            eval {
                                unmap_cifs_share(mappath=>$UNCpath, client=>$client);
                            };
                        }

                        # Authenticate the client as the test username
                        foreach my $UNCpath (keys(%UNCpaths)) {
                            eval {
                                map_cifs_share(share=>$UNCpath,
                                               client=>$client,
                                               domain=>$cifs_domain,
                                               username=>$username,
                                               password=>$username);
                            };
                            my $error = $@;
                            if ($error && $error !~ /Multiple connections to a server or shared resource by the same user, using more than one user name, are not allowed/) {
                                $Trace->fatal(msg=>"map_cifs_share failed for $UNCpath on $client: ".$error);
                            }
                        }
                    }
                }
            }
        } else {
            # Mount NFS path
            $self->_mount_client(client=>$client, bg => 1);
        }
    }

    $self->_wait_on_clients("mount");

    $Trace->exit(msg=>"");
}

sub client_cleanup {
    my $self = shift;
    $Trace->entry();
    my $errors;

    # Cleanup any remaining mounts even if the AUTOMOUNT flag is enabled. The main loop
    # might have died so we still need to clean up the mess.

    foreach my $client (keys %{$Test_topology->{clients}}) {
        my $c = host($client);

        # Un-authenticate CIFS user
        if ($self->{Cifs_Clients}{$client}) {
            if ($Options_hash->{RAND_MULTI_USER}) {
                # Find UNC paths to un-authenticate
                my %UNCpaths;
                foreach my $opt_iter (keys(%{$self->{test_options}{$client}})) {
                    my $mntpnt = $self->{test_options}{$client}{$opt_iter}{testdir};
                    $mntpnt =~ tr{/}{\\};
                    my ($UNCbegin, $UNCend) = split(/$client/, $mntpnt);
                    my $separator = $/;
                    $/ = "\\";
                    chomp($UNCbegin);
                    $/ = $separator;
                    $UNCpaths{$UNCbegin} = 1;
                }

                # Un-authenticate UNC path
                foreach my $UNCpath (keys(%UNCpaths)) {
                    eval {
                        unmap_cifs_share(mappath=>$UNCpath, client=>$client);
                    };
                }
            }
        } else {
            # _umount_client will not throw errors if the export is no longer mounted.
            $self->_umount_client(client=>$client, bg => 1);
        }
    }

    $self->_wait_on_clients("umount");

    $Trace->exit(msg=>"");
}

sub filer_cleanup {
    my $self = shift;
    $Trace->entry();

    if (!$self->{check_topology}) {
        $Trace->exit();
        return;
    }

    # create a list of all node resources in the Test_topology
    my @resources;
    foreach my $resource_ref (values %{$Test_topology->{cluster}}) {
        push(@resources, @{$resource_ref}) if (ref($resource_ref) eq "ARRAY");
    }
    foreach my $node (keys %{$Test_topology->{node}}) {
        foreach my $resource_ref (values %{$Test_topology->{node}{$node}}) {
            push(@resources, @{$resource_ref}) if (ref($resource_ref) eq "ARRAY");
        }
    }

    # check all of the resources for a comment that indicates that the resource
    # was previously created by TestManager. Mark this resource for cleanup
    foreach my $resource (@resources) {
        next if (ref($resource) ne "HASH");
        if ($resource->{comment} =~ "created by TestManager") {
            $resource->{Delete_Resource} = 1;
        } elsif ($resource->{rulesetname}) {
            $resource->{Delete_Resource} = 1;
        }
    }
    if ($Options_hash->{FILER_CLEANUP_PRECHECK}) {
		foreach my $app (@{$Options_hash->{FILER_CLEANUP_PRECHECK}}) {
			my $testcase;
			my $script;
			($script,$testcase) = split /\s+/,$app;
			if (!defined $testcase) {
				$testcase = "TESTCASE=testmanager_cleanup_check";
			}
			_checkers(testcase=>"$testcase", script=>$script);
		}
	}
    $Server->delete_topology(print_node=>undef, topology=>$Test_topology);

    $Trace->exit(msg=>"");
}

sub update_test_params {
    my $self = shift;

    $Trace->entry();
    $Trace->exit(msg=>"");
}

# This is only for internal use
# The params for each test instance were created prior to the mount information being
#  generated. So, loop through each test instance and update the params with the mount
#  information if AUTOMOUNT is enabled.
sub _update_test_params {
    my $self = shift;

    $Trace->entry();

    foreach my $client (keys %{$self->{test_options}}) {
        foreach my $instance (keys %{$self->{test_options}{$client}}) {
            # This is being done in the loop so that other changes could be made here if necessary
            my $classic_test_options = $self->{test_options}{$client}{$instance};
            if ($Options_hash->{AUTOMOUNT}) {
                # create the appropriate params that define the mount information to pass to the clients
                $classic_test_options->{MOUNT_LIST} = join("xMOUNTx", @{$Test_topology->{clients}{$client}{mount_list}});
                $classic_test_options->{CHMOD_LIST} = join("xCHMODx", @{$Test_topology->{clients}{$client}{chmod_list}});
            }
            if (defined($self->{test_options}{$client}{$instance}{trace_addrs})) {
                $classic_test_options->{TRACE_ADDRS} = join(",", @{$self->{test_options}{$client}{$instance}{trace_addrs}});
            }
        }
    }
    $Trace->exit(msg=>"");

}

sub _call_test_code {
    my ($self) = @_;
    $Trace->entry();

    # If localhost is set then the local host is the only host being used so subtest/spindex
    #  can be skipped. Just call the routine direct and return.
    # * Warning *
    # Running with localhost currently bypasses any timeout mechanisms since they rely on
    #  RemoteExecution to handle the timeout and it is being bypassed.
    # * Warning *
    if ($self->{localhost}) {
        my $client = hostname();
	# if localhost is set, then it will contain the new param entries which need to be set
	#  in the Options object prior to calling the test code. Otherwise, the param entries
	#  would get passed via subtest/spindex.
	$Options->set(%{$self->{localhost}}) if (ref($self->{localhost}) eq "HASH");
        $Trace->comment(msg=>"Running test block on localhost ($client)");
	iterate( subroutine => \&_run_classic_test_mode);
	$Trace->exit();
	return 1;
    }

    # still need to fix this for NATE
    if (defined($Options_hash->{max_client_load})) {
	$Remote->uptime(hosts=>keys %{$Test_topology->{clients}});
    }

    if ($self->{test_block_layout} eq SINGLE() && $self->{test_layout} eq SINGLE()) {
        $Trace->comment(msg=>"Running STAGE test_main on remote hosts");
	iterate( subroutine => \&_run_classic_test_mode);
	$Trace->exit();
	return 1;
    }

    if ($self->{test_block_layout} eq PER_INSTANCE() && $self->{test_layout} eq SINGLE()) {
        $Trace->comment(msg=>"Running STAGE test_main on remote hosts");
	iterate( subroutine => \&_run_classic_test_mode);
	$Trace->exit();
	return 1;
    }

    if ($self->{test_block_layout} eq SINGLE() && $self->{test_layout} eq PER_INSTANCE()) {
        $Trace->comment(msg=>"Running STAGE test_main on remote hosts");
	_run_classic_test_mode();
	$Trace->exit();
	return 1;
    }

    # reaching here means that test_block_layout=PER_INSTANCE and test_layout=PER_INSTANCE
    $Trace->comment(msg=>"Running STAGE test_main on remote hosts");
    $self->_run_code_remotely();

    $Trace->exit(msg=>"");
}

sub run_command {
    my $self = shift;
    my %params = (command  => undef,
                  timeout  => 3600,
                  runid    => undef,
                  username => undef,
		  @_);
    my $command = $params{command};
    my $timeout = $params{timeout};
    my $runid = $params{runid};
    my $username = $params{username};
    my $rc;

    # If the tester wants to run with other another username, but has not provided
    # one, then use TestUsers.pm to find a valid username for the specific lab
    # we are running in (unix only, for now).
    if ((not defined($username) && $Options_hash->{RAND_MULTI_USER}) && $OSNAME !~ /win32/i) {
        # Determine which lab we are running in
        my $global_params = NATE::ParamSet->new(global=>1);
        my $labname = $global_params->get("LAB_NAME");
        $labname = $global_params->get("LAB") if (not defined($labname));

        # Get a valid username from TestUsers.pm
        my @users;
        if (defined($labname)) {
            if (defined_lab(labname => $labname)) {
                @users = get_users(labname => $labname, count => 1, random => 1);
                $username = $users[0];
                $Trace->comment(msg => "Running as user: $username") if (defined($username));
            }
        }
    }

    $Trace->entry(msg => "params= " . Dumper(\%params));
    $Trace->comment(msg=>"Running $command\n");
    my @command = split(/\s+/, $params{command});
    if ( defined($Options_hash->{RUN_COMMAND_TIMEOUT}) ) {
       $timeout = $Options_hash->{RUN_COMMAND_TIMEOUT};
       $Trace->trace(msg=>"timeout of $params{timeout}, overridden with RUN_COMMAND_TIMEOUT of $timeout\n");
    }
    my $subtest;
    local $@;
    #eval {
        # subtest on windows returns right away rather than waiting on the underlying
        #  command to finish. So, workaround the issue by running a code block inside
        #  of the subtest. This block will call say() to execute the command rather than
        #  system() so that nate can keep track of it.
        # Burt:658571 Changed the NATE::MAJOR_VERSION check to be >= 3 as suggested in the comment below.
        #  This should stop subtest to return right away which causes script error: subtest exited with status 5 issue
        if ($OSNAME =~ /win32/i && NATE::MAJOR_VERSION >= 3) {
            my $code_to_subtest = '{my ($client, $command, $timeout)=@_; host($client); my $conn = connect("$client#rsh"); say($conn, $command, -timeout, $timeout);}';
            $Trace->comment(msg=>"Using say workaround to subtest on win32; Use Nate v3 or greater for improved support");
            if ( defined $runid ) {
                $subtest = subtest($code_to_subtest, -runid, $runid, '--', hostname(), $command, $timeout);
            } else {
                $subtest = subtest($code_to_subtest, '--', hostname(), $command, $timeout);
            }
        }
        else {
            # Call subtest to run right away for linux clients or if we are using an old version of
            # NATE 
            my @params = (shift @command, -timeout, $timeout);
            if ( defined $runid ) {
                push(@params, "-runid", $runid);
            }
            if ( defined $username ) {
                push(@params, "-username", $username);
            }
            $subtest = subtest(@params, "--", @command);
        }
    #};
    my $error = $@;
    if ($error) {
        die "'$params{command}' resulted in an eval error: '$error'";
    }
    if ($subtest && ($rc = $subtest->exit_status())) {
        die "'$params{command}' exited with $rc";
    }
    $Trace->exit(msg=>"");
}

sub _define_stages {
    my $self = shift;
    $Trace->entry();
    my ($stages, $skip) = $self->get(parameter=>[qw(STAGE SKIPSTAGE)]);

    my %stages;

    my @default_stages = keys %Stages;
    my $changes = undef;

    # If this option is defined, then the user only wants to run certain stages. Check them out for validity.
    if (scalar @{$stages}) {
	$changes = 1;
	foreach my $stage (@{$stages}) {
	    my @tstages = grep(/$stage/i, @default_stages);
	    if ($#tstages > 0) {
		my $stages = join(",", @tstages);
		$Trace->fatal(msg=>"conflict with $stage: $stages\n",
			      result=>"CONFIG");
	    }
	    if ($#tstages == -1) {
		my $msg = "$stage is an unsupported stage for this test\n  supported stages: ";
		$msg .= join(", ", @default_stages)."\n";
		$Trace->fatal(msg=>$msg, result=>"CONFIG");
	    }
	    $stages{$tstages[0]} = 1;
	}
	%Stages = %stages;
    }
    # check for any stages to skip before checking the stage option so that it can complain about unsupported stages
    if (scalar @{$skip}) {
	$changes = 1;
	foreach my $stage (@{$skip}) {
	    my @tstages = grep(/$stage/i, @default_stages);
	    if ($#tstages > 0) {
		my $stages = join(",", @tstages);
		$Trace->fatal(msg=>"conflict with skip '$stage': $stages\n",
			      result=>"CONFIG");
	    }
	    if ($#tstages == -1) {
		my $msg = "$stage is an unsupported SKIPSTAGE for this test\n  supported stages: ";
		$msg .= join(", ", @default_stages)."\n";
		$Trace->fatal(msg=>$msg, result=>"CONFIG");
	    }
	    $Stages{$tstages[0]} = undef;
	}
    }
    $self->{stages_changed} = $changes;

    $Trace->exit(msg=>"");
}

sub modify_stage {
    # The following two lines setup the hash that is used by validate_mixed to validate the
    #  incoming parameter names. The resulting hash will have keys for each of the available
    #  stages that are defined in the master %Stages and specifies that they are optional to
    #  to this method and can either be scalars or undef. Doing this allows for new stages to
    #  be defined just in the master list and not here as well.
    my %stage_definitions;
    grep($stage_definitions{$_} = {type => SCALAR | UNDEF, optional => 1}, keys %Stages_default);
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'TestManager' } ],
					 \%stage_definitions,
					 );
    $Trace->entry();

    while (my ($stage, $value) = each %params) {
	$Stages{$stage} = $value;
    }

    $Trace->exit();
}

=head2 defined_stage

    if ($test->defined_stage(stage=>"filer_setup"))

C<defined_stage> is a utility function which will tell the caller whether
or not the specified stage is enabled for this test run.

=over

=item stage

Specifies the stage that should be checked. The question being asked is
whether or not this stage is enabled for this run.

=cut

sub defined_stage {
    $Trace->entry();
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'TestManager' } ],
                                         {
                                          stage => {type => SCALAR},
                                         }
                                        );

    $Trace->exit(msg => "return $Stages{$params{stage}}");
    return $Stages{$params{stage}};
}

sub _create_test_hash {
    my ($self, $pass_through) = shift;

    $Trace->entry(msg=>"");
    our $seed;
    my $size;
    my %test_options = ();
    my %options = %{$Options_hash};
    if ($options{test_options}) {
	$test_options{test_options} = "";
	foreach my $testoption (@{$options{test_options}}) {
	    $test_options{test_options} .= " $testoption";
	}
    }
    $test_options{test_options} .= "$pass_through" if ($pass_through);

    if ($options{testdir} && $options{mountpoint}) {
	$test_options{testdir} =  "$options{mountpoint}/$options{testdir}";
    }

    # since all options get passed to the children (for CLASSIC test_mode), make sure CLIENTFILE is
    # an absolute path or it will cause a CONFIG error on the child
    if ($options{CLIENTFILE} && $options{CLIENTFILE} !~ m!/!) {
	$test_options{CLIENTFILE} = realpath($options{CLIENTFILE});
    }
    # set the param which will cause the subtest to not execute all of the code again
    $test_options{TEST_MODE} = "classic";
    $test_options{SEED} = ++$seed;

    # fill in some missing values for chosen params
    $test_options{vserver} = $Virtual_server->{vserver};
    $test_options{aggregate} = [map $_->{aggregate}, @Aggregates];
    if ( $self->{cifs} ) {
        $test_options{CIFS_DOMAIN} = $Cifs_Domain;
        $test_options{KERBEROS_REALM} = $Kerberos_Realm;
        $test_options{CIFS_SERVER} = $Cifs_Server->{$Cifs_name_key};
        $test_options{CIFS_SHARE} = $Cifs_Share;
    }
    if ( $self->{need_nis} ) {
        $test_options{NIS_SERVER} = $Nis->{servers};
        $test_options{NIS_DOMAIN} = $Nis->{domain};
    }
    $Trace->exit(msg=>"");
    return \%test_options;
}

sub _create_test_params {
    my $self = shift;
    my %test_options = @_;
    my $test_options = [];

    $Trace->entry(msg=>"");
    foreach my $option (keys %test_options) {
	if (ref($test_options{$option}) eq "ARRAY") {
	    if (scalar @{$test_options{$option}}) {
		push(@$test_options, "$option=".join(",", @ {$test_options{$option}}));
	    }
	} else {
	    push(@$test_options, "$option=$test_options{$option}");
	}
    }

    $Trace->exit(msg=>"return ".join(" ", @$test_options));
    return $test_options;
}

# The mount information is stored internally as a hash. But when the information needs to be passed
#  to the clients, the information needs to be specified as a param.
# In addition, normalize the volume names
sub _convert_mount_info {
    my $self = shift;
    $Trace->entry();

    # Keep a cache of share names for volumes since we are likely to be mounting the same volumes on each client.
    #  If so, then a large number of volumes and clients can cause the tracing to get out of control when the convert_share
    #  routine is called.
    my %shares;

    if (keys %{$Test_topology->{clients}} && $self->{directory_per_instance} ) { #Add the share for caching
        $Trace->debug(msg=>"VOLUME_LAYOUT=single");
        my $volume=$Test_topology->{cluster}{volume}[0]{'volume'};
        my $share=$Test_topology->{cluster}{volume}[0]{'junction-path'};
        $shares{$volume} = $self->convert_share(volume=>$volume, share=>$share);
    }

    foreach my $client (keys %{$Test_topology->{clients}}) {
        my @mount_list = ();
        my @chmod_list = ();
        foreach my $mount_hash (@{$Test_topology->{clients}{$client}{mounts}}) {
            $Trace->debug(msg=>"$client original mount hash=".Dumper($mount_hash));
            # first, normalize the exports
	    # The volume and share should already be available in the mount entry
	    if ($mount_hash->{volume} && $mount_hash->{share}) {
		# make sure the volume name is legal
		$mount_hash->{volume} = $self->_normalize_volume_name(volume=>$mount_hash->{volume});
		if ($shares{$mount_hash->{volume}}) { # use the cached value
		    $mount_hash->{share} = $shares{$mount_hash->{volume}};
		} else {
		    # convert the share name to the real form and cache it as well
		    $mount_hash->{share} = $self->convert_share(volume=>$mount_hash->{volume}, share=>$mount_hash->{share});
		    $shares{$mount_hash->{volume}} = $mount_hash->{share}; # add the entry to our cache
		}

		# if we are in directory_per_instance mode rather than volume_per_instance then append

		#  the client_instance to the name of the share(export) to mount. This will insure that
		#  each client instance is in its own workspace
		if ($self->{directory_per_instance} && defined($mount_hash->{instance})) {
		    $mount_hash->{share} .= "/${client}_$mount_hash->{instance}_".$self->testcase_string();
		} elsif($self->{directory_per_instance}) { ##temp hack
                    $mount_hash->{share} .= "/${client}_".$self->testcase_string();
                }
	    } else {
		# This is the old code path. It had major problems and should no longer be used. Warn but still try to use it.
		$Trace->fatal(msg=>"This code path should no longer be used:".Dumper $mount_hash);
	    }

            # if this client is a windows box, then swap out the CIFS server name with specific IPs
            if ($self->{Cifs_Clients}{$client}) {
                my $instance = $mount_hash->{instance};
                $mount_hash->{mnt} =~ s/$Cifs_Server->{$Cifs_name_key}/$mount_hash->{ip}/;
                $self->{test_options}{$client}{$instance}{testdir} =~ s/$Cifs_Server->{$Cifs_name_key}/$mount_hash->{ip}/;
            }
            $Trace->debug(msg=>"$client modified mount hash=".Dumper($mount_hash));
            # store the data in the appropriate form

            $Trace->fatal(msg=> "No mount point defined") unless defined($mount_hash->{mnt}) && length($mount_hash->{mnt});
            $Trace->debug(msg=>"$client: mount $mount_hash->{options} $mount_hash->{ip}:$mount_hash->{share} $mount_hash->{mnt}");

            push(@{$self->{test_options}{$client}{$mount_hash->{instance}}{trace_addrs}}, $mount_hash->{ip});

            # If the server supports chmod on the volume creation, then skip the chmod step in the mount -or-
            #  the current value is undef, then specify the UNDEF since it will be passed in a string to
            #  mount.thpl
            if ($Server->supports_mode_in_volume_create() || !defined($mount_hash->{chmod})) {
                $mount_hash->{chmod} = "UNDEF";
            }

	    # Change delimiter : to ; to parse IPv6 addr
            push(@mount_list, join(";", $mount_hash->{ip}, $mount_hash->{share},
                                   $mount_hash->{mnt}, $mount_hash->{options}, $mount_hash->{chmod}));

        }
        $Test_topology->{clients}{$client}{mount_list} = \@mount_list;
        $Test_topology->{clients}{$client}{chmod_list} = \@chmod_list;
    }
    $Trace->exit();
}

sub say_kot {
    my ($cc, @sayopts) = @_;
    local $@;
    my (undef, undef, $out) = say($cc, @sayopts, '-eval_error');
    my $error = $@;
    if ($error && $error =~ /TIMEOUT/) {
	my $save = $error;
	$cc->send_intr();
	$@=$save;
    }
    return $out;
}

sub filers {
    my $self = shift;
    $Trace->entry();
    $Trace->exit();
    return $Server->servers();
}

sub systype {
    my $self = shift;
    my $systype;
    my $ostype = `uname -s`;
    chomp($ostype);
    if ($ostype =~ /sunos/i) {
	$systype = "sunos";
	my $arch = `uname -p`;
	chomp $arch;
	$systype .= "-$arch" if ($arch ne "sparc");
    } elsif ($ostype =~ /hp-ux/i) {
	$systype = "hpux";
    } elsif ($ostype =~ /irix64/i) {
	$systype = "irix";
    } elsif ($ostype =~ /aix/i) {
	$systype = "aix";
    } elsif ($ostype =~ /linux/i) {
	$systype = "linux";
	my $arch;
	chomp($arch = `uname -i`);
	if ($arch =~ /_64/) {
	    $systype .= "_$arch";
	}
    } else {
	$Trace->(msg=>"Unknown result from `uname -s`: $ostype\n", result=>"CONFIG");
    }
    return $systype;
}

sub search_topology {
    my $self = shift;
    my %params = (
		  topology => "cluster",
		  cmd_node => undef,
		  key => undef,
		  value => undef,
		  @_,
		  );

    my $hash;
    $Trace->entry(msg=>"looking for $params{cmd_node}->$params{key} = $params{value}");

    if (exists($Test_topology->{$params{topology}}{$params{cmd_node}})) {
	foreach my $entry (@{$Test_topology->{$params{topology}}{$params{cmd_node}}}) {
	    if ($entry->{$params{key}} eq $params{value}) {
		$hash = $entry;
		last;
	    }
	}
    }
    $Trace->exit(msg=>"return $hash");
    return $hash;
}

sub _find_relative_testname {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'TestManager' } ],
                                         {
                                           testname => { type => SCALAR },
                                         }
					 );
    my $testname;
    $Trace->entry(msg=>"test = $params{testname}");

    if ($params{testname} =~ m!^/!) {
        # get rid of any '..' entries in the path so that the foreach loop below will work
        if (my $real_testname = realpath($params{testname})) {
            $testname = $real_testname;
        } elsif ($self->{cifs}) {
            $testname = $params{testname};
            push(@{$self->{warnings}}, "The test may not work correctly on CIFS clients since realpath('$testname') failed");
        }

        # Try to find the original relative testname rather than the fqname.
        foreach my $path (split(/;/, $Options_hash->{NATE_LIB})) {
            $path = quotemeta($path);
            if ($params{testname} =~ m/^$path/) {
                $testname =~ s!$path(/)*!!;
                $Trace->trace(msg=>"Converted '$params{testname}' to '$testname' because of nate_lib '$path'");
            }
        }
        if (!$testname || $testname eq "") {
            push(@{$self->{warnings}}, "Could not locate '$params{testname}' in NATE_LIB");
            $testname = $params{testname};
        }
    } else {
        $testname = $params{testname};
    }

    $Trace->exit(msg=>"return $testname");
    return $testname;
}

=back

=head1 PARAMS

=over

=item STAGE

A list of specific stages to execute during this test script run. The list can
be separated by a ',' or a ':'. The list can be any combination of the following
which are all run by default:

filer_setup

client_setup

test_setup

test_main

test_cleanup

client_cleanup

filer_cleanup

=item SKIPSTAGE

A list of specific stages to not execute during this test script run. The list
can be separated by a ',' or a ':'. The list can be any combination of the
stages listed under C<STAGE>

=item TEST_MODE

This param controls the mode that the test is running in. By default, the test
is running in C<scalable> mode and will execute itself in C<classic> mode on
all of the provided clients.

=over

=item scalable

This test_mode indicates that the script will act more like a ST from the
Spinnaker days or a SDG thpl which performed setup/cleanup. The test is run
through multiple stages to perform setup on filers and clients, to subtest
the script to the provided clients and then to perform cleanup on the clients
and filers. The subtested test on the clients are actually run in classic
mode because the clients are presetup during the setup stages.

=item classic

This test_mode indicates that the script will act more like a classic thpl or
simple pl. It assumes it is already running on the client and will access the
mounted filesystem via the testdir parameter. This can be an NFS-mounted
filesystem or a local directory. No setup or cleanup on the filer or client
is performed in this mode. That is handled in scalable mode.

=back

=item TEST_TIMEOUT

Specifies the amount of time in seconds that the test case is expected to take.
If specified, this value overrides the value passed as C<timeout> in C<run_test>
as well as DURATION or ITERATE as defined in TestUtils.

=item VOLUME_OWNER

Specifies the default owner for a volume created by TestManager. If the value
is "self", then the current owner will be set. The default value is '0'.

=item MOUNT_OPTIONS

Specifies the options to pass to mount when running the C<client_setup> stage.
The default value is "hard,intr".

=item MNTPNT_MODE

Specifies the mode that should be applied to newly created mounts via the
chmod program. The default value is "777". Setting the param to nothing will
cause the chmod to not occur.

=item testdir

Specifies the path to the test directory a test should run against. Not all
tests will use this param but many do.

=item MOUNTPOINT_UNIQUIFIER

Specifies a string to be embedded in the mouthpoint paths. You can use this to
run multiple instances of the same test simultaneously on the same clients.
MOUNTPOINT_UNIQUIFIER is more robust than CLIENT_INSTANCE, which can't be used
with more than one client. Mountpoints with MOUNTPOINT_UNIQUIFIER specified have
the form: /t/<testname>.<data_if>.<uniquifier>.<client_instance>/

=item CLUSTER_TRAFFIC

Specifies whether or not client traffic should follow a specific data path.
The default is to use C<round-robin>. This parameter is actually defined in
L<Node::NG>.

=over

=item local

Force all client traffic to stay local on a filer and not cross the cluster.

=item remote

Force all client traffic to cross the cluster link.

=back

=item VOLUME_LAYOUT

Specifies how new volumes are created during C<scalable> C<TEST_MODE>. If the
value is "per_instance" then a new volume is created for each client instance
that has been passed in. If the value is "single" then only one volume is
created and all client instances will run against this one volume. If the value
is "test_defined" then the test is creating a special topology. The default
value is "per_instance". This value overrides the value specified in the
C<run_test> method.

=item CLIENT_INSTANCE

This param is mostly used to tell each client instance (when in classic
C<TEST_MODE>) which instance number it is. However, this param can also
be specified when you want to run the same test from the same client
at the same time. You need to set this to a unique value for each one
so that the rulesets are uniquely created for each one.

=item JUNCTION_PATH

C<JUNCTION_PATH> specifies the junction-path to use when you specify a
C<volume>. The default is /volume_name.

=item AUTOMOUNT

This param is used to specify that the mounts should appear as though
they are occurring from an automounter rather than from static mounts.
Setting the param will make the client_setup C<STAGE> and client_cleanup
C<STAGE> become a noop. The mounts will occur prior to test_main
and the umounts will occur after the test_main routines. Running
with ITERATE or DURATION will cause the mounts to occur prior to each
iteration of the test_main.

=back

=head1 SEE ALSO

L<TracePkg>
L<Options>
L<Server>

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/TestManager.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut


1;
