#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     Test Execution Controller for Stress Scalability Tests.
##
## @author      gvg@netapp.com,snider@netapp.com
##
## @pod here

=head1 NAME

TestManagerLite -  Test Execution Controller for Stress Scalability Tests. 
Controlls test execution and provides common plateform for stress and scalablility
related tests.

=head1 SYNOPSIS

    package Test;
    use base qw(TestManagerLite);

    sub test_main {
        my $self = shift;
        my ($failure) = $self->get(parameter=>"INDUCE_FAILURE");
        # some test code here to implement the test case
        die "This test failed\n" if ($failure);
    }

    my $test = Test->new();
   
    $test->parameters(definitions=>
        [
         {
             name => "INDUCE_FAILURE",
             description => "force a failure in test_main",
             default => undef,
         }
        ]);
    $test->run_test(timeout=>600); # 10 minutes


=head1 DESCRIPTION

C<TestManagerLite> is a base class for tests. It provides many of the base 
functions needed by a test. More detailed documentation will be 
available at http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite.

C<TestManagerLite> uses its own copy of the L<Options>
and provides abstracted methods into L<Options>.

=head1 STAGE

=head2 filer_setup

C<Topology> specific filer_setup function is called. If test wants to override it,
it is recommended to call SUPER.

=head2 client_setup

C<Topology> specific client_setup function is called. If test wants to override it,
it is recommended to call SUPER.

=head2 client_cleanup

C<Topology> specific client_cleanup function is called. If test wants to override it,
it is recommended to call SUPER.

=head2 filer_cleanup

C<Topology> specific filer_cleanup function is called. If test wants to override it,
it is recommended to call SUPER.

=head2 test_setup

Test is expected to override this function with test setup details.

=head2 test_main

Test is expected to override this function with test details.

=head2 test_cleanup

Test is expected to override this function with test cleanup details.

=head1 TODO

* have to port remaining functionality from TestManager.pm to this model.
* Result reporting should be revised and improved. 
* Detailed POD updates

=head1 SEE ALSO
    
Options Topology

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Created $Date: 2008/05/07 $

=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved

=head1 METHODS

=cut

package TestManagerLite;
use strict;
use English;

# bring this in first to force the Delayed_help action
use Options qw(-Delayed_help);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Data::Dumper qw(Dumper);
use Tharn qw ($Log config_result fail_result script_result subtest logresult);
use TestUtils qw(iterate expected_run_time);
use Cwd qw(realpath);
use String::Random;    # To create unique random string
use Network::Trace;
use Network::NodeTrace;
use TestManagerLite::ParentStaf;
use NATE::RPC::Freeze qw(freeze_and_escape);
use Topology;

my $Options;
my $Options_hash;

use RemoteExecution;

# To track Remote subtest objects (There could be local subtest objects too)
my $Remote;
my $Local_Ancillary;    # To track local ancillary subtest objects
my $tml_mgr_start_time;
$Options = new Options;
$Options->parameters(
    definitions => [
        {   name        => "MAX_CLIENT_LOSS",
            spec        => "s",
            default     => 0,
            description => "How many supervisors can fail before we die",
        },
        {   name        => "TOPO",
            description => "The topology to be used",
        },
        {   name        => "STAGE",
            spec        => "s@",
            description => "change the default test stages to run",
        },
        {   name        => "SKIPSTAGE",
            spec        => "s@",
            description => "skip these test stages",
        },
        {   name    => "SERIAL_SETUP",
            spec    => "s",
            default => 0,
            description =>
                "The test will not begin its own filer_setup method until the preceeding test has set SERIAL_SETUP "
                . "to match this value. This test will not move past filer_setup until something sets SETUP_COMPLETE to true.",
        },
        {   name    => "SERIAL_SETUP_COMPLETE",
            spec    => "s",
            default => 0,
            description =>
                "Setting this value to true causes this test to alert all tests waiting to move on from filer_setup once its own setup completes",
        },
        {   name => "TEST_TIMEOUT",
            spec => "time",
            description =>
                "how long before the test should timeout, in seconds",
        },
        {   name => "TOP_LOGDIR",
            spec => "s",
            description =>
                "The LOGDIR where test is logging. Takes it from NATE",
        },
        {   name        => "DURATION",
            default     => undef,
            description => "the length of time to iterate",
        },
        {   name        => "ITERATE",
            default     => 1,
            description => "an absolute number of iterations",
        },
        {   name         => "P_EXE_MODE",
            spec         => "s",
            default      => "mgr",
            valid_values => ["mgr", "supervisor", "worker"],
            description =>
                "dictates the execution flow. This is internal purpose only",
        },
        {   name        => 'TERMFILE',
            spec        => "s",
            default     => "TermFile.txt",
            description => "File to write TERMSTRINGS to",
        },
        {   name        => 'TERMSTRING',
            spec        => "s",
            default     => "TeRmStRiNg",
            description => "string to write in TERMFILE when complete",
        },
        {   name    => "P_USE_TOPO",
            spec    => "s",
            default => undef,
            description =>
                "Internal variable used to avoid non-mgr processes to call Server. Pre determined topology class path is set in this",
        },
        {   name    => "CLIENTNAME",
            spec    => "s",
            default => "localhost",
            description =>
                "The client name where this is running. Valid value of the hostname will be set on other P_EXE_MODEs",
        },
        {   name    => "USE_STAF",
            spec    => "s",
            default => "0.999",
            description =>
                "Should supervisors and mgr use STAF to share data?",
            subroutine => [\&_decide_staf_usage],
        },
        {   name        => "NATE_LIB_UNIX",
            spec        => "s",
            default     => undef,
            description => "Import from NATE params",
        },
        {   name        => "NATE_LIB_WINDOWS",
            spec        => "s",
            default     => undef,
            description => "Import from NATE params",
        },
    ]
);

my $Full_testname = $Tharn::Scriptname;
if (!$Full_testname || $Full_testname eq "") {
    if ( defined $INC{"NATE/Lint.pm"} ) {
        $Full_testname = "natelint";
        $Log->warn( "We seem to be running inside of natelint and will "
              . "continue execution with a fake testname. If you are seeing "
              . "this warning and are not executing natelint please alert the "
              . "owner of TestManagerLite. Full_testname = '$Full_testname'"
        );
    } else {
        die config_result("Could not determine the TESTNAME", stacktrace => 0);
    }
}
my $Testname = $Full_testname;
$Testname =~ s/.+\///;      # just get the name of the test
$Testname =~ s/\.thpl//;    # strip off the .thpl extension

# Define the basic stages here. The %Stages hash can be changed, this one is not.
my %Stages_default = (filer_setup    => 1,
                      client_setup   => 1,
                      test_setup     => 1,
                      test_main      => 1,
                      test_cleanup   => 1,
                      client_cleanup => 1,
                      filer_cleanup  => 1,);

# This hash will be used to manage the stages to run through. If a STAGE param is passed,
#  then the contents of this hash are replaced with just the specified stages.
my %Stages = %Stages_default;

=head2 new

    package Test;
    use base qw(TestManagerLite);
    my $test = Test->new();

C<new> is the invocant for the Test object. The Test object should not
override this one; if it does, then it should call it as a SUPER first.
C<new> will first call process on L<Options> and then check for the value
of the C<P_EXE_MODE> param and store it for later useage. It will also define
references to the following internal structures so that the underlying Test
object can access them directly

=cut

sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $self     = {@_,};
    $Log->enter();
    bless($self, $class);

    # Determine the test_mode at startup, prior to parameters() being called
    # The test_mode is based on an input parameter so call process() and then
    # check that option.
    # Since we are in the new(), test_cases have not yet been defined so
    # process() will die when it sees a TESTCASE; turn that check off
    $Options->process(test_cases => undef);
    $Options_hash = $Options->get_hash();
    $self->{testname} = $Testname;
    my @path = split(m!/!, $Tharn::Scriptspec);
    pop @path;
    $self->{test_location}    = join("/", @path);
    $self->{_ancillary_tests} = {};
    $self->{_local_tests}     = {};
    $self->{exe_mode}         = $self->get(parameter => "P_EXE_MODE");

    $self->{myself} =
      $self->_find_relative_testname(testname => $Full_testname);

    $Log->exit("return $self");
    return $self;
}

############## Routines that tests can override ###################

sub define_topology {
    $Log->enter();
    my ($self) = shift;

    $Log->comment("Running STAGE define_topology");

    if (my $serialSetup = $Options->get(option => "SERIAL_SETUP")) {
        $self->_define_topology_serial(setup_number => $serialSetup);
    } else {
        # define_topology is defined by Topology.
        $self->topology()->define_topology();
    }    

    $self->_handle_define_topology_complete();

    $Log->exit();
}


sub update_test_topology {
    $Log->enter();
    my ($self) = shift;

    # update_test_topology is defined by Topology.
    $self->topology()->update_test_topology();

    $Log->exit();
}

=head2 filer_setup

While this method should not be called directly by the user, there
are some important aspects of controlling its execution which 
are important to discuss here. 

C<filer_setup> controls the application of a topology on
to a filer. For example volumes are created. There are some
variables defined to offer the user some control over this function, 
when it is being used for multiple tests running at one time, 
especially under software such as stest.thpl

=over

=item SERIAL_SETUP 

Specifies that all tests should execute their setup portions in 
a serial fashion. Valid values for SERIAL_SETUP are anything greater
than zero, but should be sequetial. This will dictate the order
in which the filer_setup of the given script is performed. When 
a test has a TRUE value for SERIAL_SETUP, it will ALWAYS wait for 
SERIAL_SETUP_COMPLETE to be set on the parent

For example, in stest: 

parallel { 
    script1.thpl SERIAL_SETUP=1
    script2.thpl SERIAL_SETUP=2 SERIAL_SETUP_COMPLETE=1
}

This setup would cause script1.thpl to run until filer_setup has completed. 
script1.thpl will then set the value of SERIAL_SETUP (on the parent) to 
SERIAL_SETUP + 1. Scripts periodically check this value, and when 
it finds that SERIAL_SETUP on the parent is equal to the local value 
of SERIAL_SETUP, the script (script2.thpl in this example) begins
execution of filer_setup.

=item SERIAL_SETUP_COMPLETE

The last test to complete filer_setup (or the test with the highest value of 
SERIAL_SETUP) should set SERIAL_SETUP_COMPLETE. Setting this value to true, 
is what tells any waiting test that it should proceed from filer_setup to the 
next stages of the test.

*** If you override this method without calling SUPER class, then these values 
will be useless!!!***

=cut

sub filer_setup {
    $Log->enter();
    my ($self) = shift;

    $Log->comment("Running STAGE filer_setup");

    if (my $serialSetup = $Options->get(option => "SERIAL_SETUP")) {
        $self->_filer_setup_serial(setup_number => $serialSetup);
    } else {

        # filer_setup is defined by Topology.
        $self->topology()->filer_setup();
    }

    $self->_handle_setup_complete();

    $Log->exit();
}

sub filer_cleanup {
    my $self = shift;

    # filer_cleanup is defined by Topology.

    $Log->enter();
    $self->topology()->filer_cleanup();
    $Log->exit();
}

sub client_setup {
    my ($self) = shift;

    # client_setup is defined by Topology

    $Log->enter();
    $self->topology()->client_setup(@_);
    $Log->exit();
}

sub client_cleanup {
    my ($self) = shift;

    # client_cleanup is defined by Topology

    $Log->enter();
    $self->topology()->client_cleanup(@_);
    $Log->exit();
}

# The following three test functions are supposed to be
# overridden by test scripts. If they don't these
# functions do nothing but printing debug messages.

sub test_setup {
    $Log->enter();
    $Log->debug("Processing test_setup in TestManagerLite");
    $Log->exit();
}

sub test_main {
    $Log->enter();
    $Log->debug("Processing test_main in TestManagerLite");
    $Log->exit();
}

sub test_cleanup {
    $Log->enter();
    $Log->debug("Processing test_cleanup in TestManagerLite");
    $Log->exit();
}

############## A few APIs that tests can use ########################

=head1 LIMITED APIs 

These APIs can be called only from the following routines.
C<filer_setup>
C<filer_cleanup>
C<TestManagerLite> routines <P_EXE_MODE == mgr>

=head2 topology

    my $hashref = $test->topology();

C<topology> returns the hashref returned by C<Topology> object.

=cut

sub topology {
    my $self = shift;
    $Log->enter();

    $Log->exit();
    return $self->{topology};
}

=head2 clients

    my @clients = $test->clients();

C<clients> returns an array returned by C<Topology> object.

=cut

sub clients {
    my $self = shift;
    $Log->enter();

    if ($self->{exe_mode} ne "mgr") {
        die script_result(
                      "This routine is not available at $self->{exe_mode} mode",
                      stacktrace => 0);
    }

    $Log->exit();
    return $self->topology()->clients();
}

=head2 supervisor2start_next

    my $client = $test->supervisor2start_next();

C<supervisor2start_next> calls Topology method to get the next supervisor to start

=cut

sub supervisor2start_next {
    my $self = shift;
    $Log->enter();

    if ($self->{exe_mode} ne "mgr") {
        die script_result(
                      "This routine is not available at $self->{exe_mode} mode",
                      stacktrace => 0);
    }

    $Log->exit();
    return $self->topology()->supervisor2start_next();
}

=head2 filers

    my @filers = $test->filers();

C<filers> returns an array returned by C<Topology> object.

=cut

sub filers {
    my $self = shift;
    $Log->enter();

    if ($self->{exe_mode} ne "mgr") {
        die script_result(
                      "This routine is not available at $self->{exe_mode} mode",
                      stacktrace => 0);
    }

    $Log->exit();
    return $self->topology()->filers();
}

=head2 test_options

    my $hashref = $test->test_options();

C<test_options> returns a hashref returned by C<Topology> object.

=cut

sub test_options {
    my $self = shift;
    $Log->enter();

    if ($self->{exe_mode} ne "mgr" && $self->{exe_mode} ne "supervisor") {
        die script_result(
                      "This routine is not available at $self->{exe_mode} mode",
                      stacktrace => 0);
    }

    my $options = {};
    if (my $test = $self->topology()->test_options(@_)) {
        $options = $test;
    }

    $Log->exit("test options: " . Dumper($options));
    return $options;
}

=head2 clientsetup_options

    my $hashref = $test->clientsetup_options();

C<clientsetup_options> returns a hashref returned by C<Topology> object.

=cut

sub clientsetup_options {
    my $self = shift;
    $Log->enter();

    if ($self->{exe_mode} ne "mgr" && $self->{exe_mode} ne "supervisor") {
        die script_result(
                      "This routine is not available at $self->{exe_mode} mode",
                      stacktrace => 0);
    }

    my $options = {};
    if (my $test = $self->topology()->clientsetup_options(@_)) {
        $options = $test;
    }

    $Log->exit("client setup options: " . Dumper($options));
    return $options;
}

=head2 parameters

    $test->parameters(definitions=>
        [
         {
             name => "INDUCE_FAILURE",
             description => "force a failure in test_main",
             default => undef,
         }
        ]);

C<parameters> provides a definition for a new parameter.

=over

=item definitions

An array of hashes which each define a unique parameter. Please see the 
'definitions' method of the L<Options> library.

=back

=cut

sub parameters {
    my $self   = shift;
    my %params = (@_);
    $Log->enter();

    # pass the definitions to Options but only if the test_mode for the definition matches the test_mode
    #  that the TestManagerLite is currently running. The default mode for the TM is "mgr" which causes
    #  the test to run through the test stages. The other mode for the TM is "worker" which runs just the
    #  test_main defined by the test.
    # The default value for the parameter test_mode is "mgr"
    # Just pass the params if not test_mode is specified.
    my @param_list;
    foreach my $param_hash (@{$params{definitions}}) {
        if (!defined($param_hash->{test_mode})
            || ($self->{test_mode} =~ /$param_hash->{test_mode}/i)) {
            push(@param_list, $param_hash);
        }
    }
    $Options->parameters(%params, definitions => \@param_list);
    $Log->exit();
}

=head2 set

    $test->set(option => value);

C<set> sets a value to the option specified.

=over

=item option

Please see the C<set> method of the L<Options> library.

=back

=cut

# allow for a test to set a parameter
sub set {
    my $self   = shift;
    my %params = (@_);

    $Log->enter("params: " . Dumper(\%params));
    $Options->set(%params);
    $Log->exit();
}

=head2 set_default

    $test->set_default(option => value);

C<set_default> sets a value to the option specified.

=over

=item option

Please see the C<set_default> method of the L<Options> library.

=back

=cut

# allow for a test to set a default
sub set_default {
    my $self   = shift;
    my %params = (@_);

    $Log->enter("params: " . Dumper(\%params));
    $Options->set_default(%params);
    $Log->exit();
}

=head2 get

    $test->get(parameter => [options]);

C<get> gets a value for the option specified.

=over

=item parameter

Please see the C<get> method of the L<Options> library.

=back

=cut

sub get {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT, isa => 'TestManagerLite'}],
                     {parameter =>
                        {type => ARRAYREF | SCALAR | UNDEF, optional => 1},
                      testcase => {type => SCALAR | UNDEF, optional => 1},});

    $Log->enter("parameter=>$params{parameter}, testcase=>$params{testcase}");
    if (wantarray) {
        my @options = $Options->get(option => $params{parameter});
        $Log->exit("return @options");
        return @options;
    } else {
        my $option = $Options->get(option   => $params{parameter},
                                   testcase => $params{testcase});
        $Log->exit("return $option");
        return $option;
    }
}

=head2 get_hash

    $test->get_hash(parameter => [options]);

C<get_hash> gets values for the options specified.

=over

=item parameter

Please see the C<get_hash> method of the L<Options> library.

=back

=cut

sub get_hash {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT, isa => 'TestManagerLite'}],
                     {parameter => {type => ARRAYREF | SCALAR, optional => 1},}
                     );

    $Log->enter();
    my $options = $Options->get_hash(option => $params{parameter});
    $Log->exit();
    return $options;
}

=head2 test_cases

    $test->test_cases(option => value);

C<test_cases> of L<Options> library.

=cut

sub test_cases {
    my $self = shift;
    $Log->enter();
    $Options->test_cases(@_);
    $Log->exit();
}

=head2 add_ancillary_tests

    $testsref = [
                  {
                      name => "test1",
                      cmd  => $script_path,
                      sequence => 1,
                  },
                  {
                      name => "test2",
                      cmd  => "$code",
                      sequence => 2,
                  }
                 ]);                 
    $test->add_ancillary_tests(tests => $testsref);

C<add_ancillary_tests> is the the method of TestManagerLite. This method
can be used by the test to add as many ancillary tests as needed
to run in parallel to test execution. These tests are started by
TestManagerLite just before the real test is started and stoppped/killed
just after real test is completed.

=over

=item tests

The array ref of the hash entries that contains the test script path or
coderef and properties of the same.

=back

=cut

sub add_ancillary_tests {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT, isa => 'TestManagerLite'}],
                     {tests => {type => ARRAYREF, default => []},});
    $Log->enter();

    foreach my $testref (@{$params{tests}}) {
        my $seq = $testref->{sequence};
        delete $testref->{sequence};
        $self->{_ancillary_tests}{$seq} = $testref;
    }

    $Log->exit();
    return;
}

=head2 add_local_tests

    $testsref = [
                  {
                      name => "test1",
                      cmd  => $script_path,
                      sequence => 1,
                  },
                  {
                      name => "test2",
                      cmd  => "$code",
                      sequence => 2,
                  }
                 ]);                 
    $test->add_local_tests(tests => $testsref);

C<add_local_tests> is the the method of TestManagerLite. This method
can be used by the test to add as many tests as needed
to run in parallel to test execution. These tests are started by
TestManagerLite after the stage filer_setup and immediately after remote
tests are started.

=over

=item tests

The array ref of the hash entries that contains the test script path or
coderef or codestring and properties of the same.

=back

=cut

sub add_local_tests {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT, isa => 'TestManagerLite'}],
                     {tests => {type => ARRAYREF, default => []},});
    $Log->enter();

    foreach my $testref (@{$params{tests}}) {
        my $seq = $testref->{sequence};
        delete $testref->{sequence};
        $self->{_local_tests}{$seq} = $testref;
    }

    $Log->exit();
    return;
}

########################### run_test ############################

=head2 run_test

    $test->run_test(timeout=>600, # 10 minutes
                    cmd=>$cmd,);
(OR)    
    $test->run_test(timeout=>600, # 10 minutes
                    cmd=>$script_path,);

C<run_test> is the main method of C<TestManagerLite>. This method controls all 
execution of the test and will not return to the caller. More detailed 
documentation will be available at 
http://wikid.netapp.com/w/Stress-Scalability_QA/NAS/TestManagerLite

=over

=item timeout

C<timeout> specifies how long the test_main is expected to run. A
FATAL result will be returned if this time passes prior to the test_main
routine finishes. The time is in seconds.

=back

=over

=item cmd

Optional. C<cmd> specifies the coderef or script path to be executed instead of
executing test_setup, test_main and test_cleanup functions.

=back

=cut

sub run_test {
    my ($self, %params) = validate_mixed(
        @_,
        [{type => OBJECT, isa => 'TestManagerLite'}],
        {timeout => {type => SCALAR, default => 1800}
         ,    # 1/2 hour for no good reason
         reserve_clients => {type => SCALAR,           default => 1},
         cmd             => {type => SCALAR | CODEREF, default => undef},});

    $Log->enter();

    # first process the options
    # this is the main process call so all parameters should be defined. It is now safe to
    #  allow process to look for HELP.
    my %opts = (run_help => 1);

    # Process test_cases only in the mgr mode
    $opts{test_cases} = ($self->{exe_mode} eq "mgr") ? 1 : 0;

    # hack: The above doesn't seem to be sufficient
    $Options->{testcase_start} = ($self->{exe_mode} eq "mgr") ? 1 : 0;
    $Options->process(%opts);

    $Options_hash = $Options->get_hash();
    $self->_define_stages();

    my $timeout =
        $Options_hash->{TEST_TIMEOUT}
      ? $Options_hash->{TEST_TIMEOUT}
      : $params{timeout};
    if ($timeout ne "disabled") {
        $timeout = expected_run_time(timeout => $timeout);
    }

    # Create given topology
    $self->{topology} = new Topology(
                                 topology => $Options_hash->{TOPO},
                                 reserve_clients => ($self->{exe_mode} eq "mgr")
                                 ? $params{reserve_clients}
                                 : 0);

    if ($self->{exe_mode} eq "worker") {

        # If this code gets executed means external 'cmd' was not given,
        # hence we have to execute test_* functions.
        $Log->debug("Entering into Worker Mode");
        $self->run_worker_mode();
        return 1;
    } elsif ($self->{exe_mode} eq "supervisor") {
        $self->run_supervisor_mode(cmd     => $params{cmd},
                                   timeout => $timeout);
        return 1;
    } elsif ($self->{exe_mode} eq "mgr") {
        $self->run_mgr_mode(timeout => $timeout);
    } else {
        die fail_result(
            "Incorrect P_EXE_MODE found or was not set: $Options_hash->{P_EXE_MODE}. Allowed values are mgr, supervisor and worker",
            stacktrace => 0);
    }

    $Log->exit();
}

sub systype {
    my $self = shift;
    my $systype;
    my $ostype = `uname -s`;
    chomp($ostype);
    if ($ostype =~ /sunos/i) {
        $systype = "sunos";
        my $arch = `uname -p`;
        chomp $arch;
        $systype .= "-$arch" if ($arch ne "sparc");
    } elsif ($ostype =~ /hp-ux/i) {
        $systype = "hpux";
    } elsif ($ostype =~ /irix64/i) {
        $systype = "irix";
    } elsif ($ostype =~ /aix/i) {
        $systype = "aix";
    } elsif ($ostype =~ /linux/i) {
        $systype = "linux";
        my $arch;
        chomp($arch = `uname -i`);
        if ($arch =~ /_64/) {
            $systype .= "_$arch";
        }
    } else {
        die config_result("Unknown result from `uname -s`: $ostype\n",
                          stacktrace => 0);
    }
    return $systype;
}

# This function doesn't work on non-Unix and need STAF running
my $cmd_runid = 1;

sub run_command {
    my $self = shift;
    my %params = (command => undef,
                  runid   => "command",
                  timeout => 3600,
                  @_);
    $Log->enter();

    # Get it from NATE directly
    my $logdir = $self->get(parameter => [qw(TOP_LOGDIR)]);
    my $runid = Tharn::param("RUNID", -default, undef);
    my $child_logdir = $logdir . "/" . $runid;

    my $child_stdout = sprintf("%s/%03d_%s_stdout.log",
                               $child_logdir, $cmd_runid, $params{runid});
    my $child_stderr = sprintf("%s/%03d_%s_stderr.log",
                               $child_logdir, $cmd_runid++, $params{runid});

    my @params = split(/\s+/, $params{command});
    my $cmd = $params[0];
    delete $params[0];

    # Create directory like what subtest does to save command logs
    my $r = mkdir($child_logdir, oct("777"));
    if (!$r and not $!{EEXIST}) {
        die script_result("Unable to create log directory $child_logdir: $!",
                          stacktrace => 0);
    }

    $Log->comment("TestManagerLite executing: $cmd");

    my $RS = new String::Random;
    use Tharn qw($Staf);
    use NATE::STAF qw(kTimeout);

    my $done     = 0;
    my $workload = $RS->randregex('[a-z][a-z][a-z][a-z][a-z][a-z]');
    my $out;
    eval {
        $out = $Staf->process_start(
            command        => $cmd,
            params         => join(" ", @params),
            timeout        => $params{timeout},
            workload       => $workload,
            stdoutfile     => $child_stdout,
            stderrfile     => $child_stderr,
            stderrtostdout => 0,
            on_error       => sub {
                my (%result) = @_;
                if ($result{rc} == kTimeout) {
                    $Log->warn("$cmd: Timed out.");
                    $Log->comment(
                         "Calling process_stop to cleanup the running process");
                    $Staf->process_stop(workload => $workload);
                } else {
                    die Dumper(\%result);
                }
            });
    };
    my $err = $@;
    $Log->trace("Process start returned");

    open(OUT, "< $child_stderr")
      or die script_result("Could not open child logfile $child_stderr: $!",
                           stacktrace => 0);
    my @errs = <OUT>;
    close(OUT);

    if ($out->{rc} || $errs[0] || $err) {
        my $msg = "exited with $out->{rc}";
        $msg .= " with message: $errs[0]" if ($errs[0]);
        $msg .= " with exception: $err"   if ($err);
        die fail_result("$params{command} failed: $msg", stacktrace => 0);
    } else {
        $Log->comment("Command $params{command} completed successfully");
    }

    $Log->exit();
}

##################### Internal Routines ###############

sub _establish_channel {
    my ($self, %opts) = @_;

    if ($Options_hash->{USE_STAF}) {
        return;
    }

    # Save data correspond to all supervisors by name in TOP_LOGDIR
    foreach my $supervisor (@{$opts{supervisors}}) {
        my $datafile = $Options_hash->{TOP_LOGDIR} . "/" . $supervisor . ".dat";

        my $clientsetup_options =
          $self->clientsetup_options(client => $supervisor);
        my $test_options = $self->test_options(client => $supervisor);

        open(FHDATA, ">$datafile")
          or die "Could not open $datafile: $!\n";
        print FHDATA Data::Dumper->Dump([$clientsetup_options],
                                        ['CLIENTSETUP_OPTIONS'])
          . "\n";
        print FHDATA Data::Dumper->Dump([$test_options], ['TEST_OPTIONS'])
          . "\n";
        close(FHDATA);
    }
}

sub _start_ancillary_tests {
    my $self = shift;

    $Log->enter();
    $Log->debug("Ancillary Tests: \n" . Dumper($self->{_ancillary_tests}));
    my $options = $self->get_hash();

    foreach my $seq (sort keys %{$self->{_ancillary_tests}}) {
        my $testref = $self->{_ancillary_tests}{$seq};
        my %args;

        # Ancillary tests meant to run until real tests are done and then
        # parent can kill these. Hence set ITERATE to "infinity"
        if ($testref->{nate_params}) {
            $args{nate_params} = {%$options,
                                  %{$testref->{nate_params}},
                                  DURATION => undef,
                                  ITERATE  => "infinity",};
        } else {
            $args{nate_params} = {%$options,
                                  DURATION => undef,
                                  ITERATE  => "infinity",};
        }

        $Log->comment("Starting Ancillary Test: $testref->{name}");
        $Local_Ancillary->execute(
                    host     => "localhost",
                    cmd      => $testref->{cmd},
                    nateeval => $self->_detect_nateeval(cmd => $testref->{cmd}),
                    runid    => $testref->{name},
                    handler  => sub { },
                    %args);
    }

    $Log->exit();
    return;
}

sub _start_local_tests {
    my $self = shift;

    $Log->enter();

    #   $Log->debug("Local Tests: \n" . Dumper($self->{_local_tests}));
    my $options = $self->get_hash();

    foreach my $seq (sort keys %{$self->{_local_tests}}) {
        my $testref = $self->{_local_tests}{$seq};
        my %args;

        if ($testref->{nate_params}) {
            $args{nate_params} = {%$options, %{$testref->{nate_params}},};
        } else {
            $args{nate_params} = {%$options,};
        }

        $Log->comment(
              "Starting Local Test: $testref->{name} of type $testref->{type}");
        $Remote->execute(
                    host     => "localhost",
                    cmd      => $testref->{cmd},
                    nateeval => $self->_detect_nateeval(cmd => $testref->{cmd}),
                    runid    => $testref->{name},
                    handler  => sub { },
                    %args);
    }

    $Log->exit();
    return;
}

sub _init_supervisor {
    my $self = shift;
    $Log->enter();

    our ($CLIENTSETUP_OPTIONS, $TEST_OPTIONS);

    if ($Options_hash->{USE_STAF}) {
        my $msg;

        $msg =
          $Tharn::Parent->message_get(eval_message => 1,
                                      type         => "TML/Spr/ClntStOpts");
        $CLIENTSETUP_OPTIONS = ($msg->{message}) ? $msg->{message} : $msg;
        $Log->debug("_init_supervisor client setup options:"
                    . Dumper($CLIENTSETUP_OPTIONS));

        $msg =
          $Tharn::Parent->message_get(eval_message => 1,
                                      type         => "TML/Spr/TstOpts");
        $TEST_OPTIONS = ($msg->{message}) ? $msg->{message} : $msg;
        $Log->debug("_init_supervisor test options:" . Dumper($TEST_OPTIONS));
    } else {
        my $datafile =
            $Options_hash->{TOP_LOGDIR} . "/"
          . $Options_hash->{CLIENTNAME} . ".dat";
        require "$datafile";
    }

    $self->topology()->set_clientsetup_options(
                                           options => $CLIENTSETUP_OPTIONS,
                                           client => $Options_hash->{CLIENTNAME}
                                           );
    $self->topology()->set_test_options(options => $TEST_OPTIONS,
                                        client  => $Options_hash->{CLIENTNAME});

    #
    # Update TEST_OPTIONS into Staf, so workers can get them directly.
    #
    use Tharn qw($Staf);
    $Staf->var_set(handle => $Staf->handle(),
                   var    => $TEST_OPTIONS);
    $Log->exit();

}

sub run_mgr_mode {
    my ($self, %params) = @_;

    $Log->enter();
    $Log->debug("Entering into Mgr Mode");
    $tml_mgr_start_time = time();

    # Define Topology
    $self->define_topology();

    # Update test topology - In case test wants to change something
    $self->update_test_topology();

    my $testcase = $Options->get(testcase => $Options_hash->{TESTCASE});
    my $result_message =
      ($testcase && $testcase->{TC_description})
      ? $testcase->{TC_description} . ":"
      : $Testname . ":";
    $Log->trace("result_message set to $result_message");

    my @clients = $self->clients();
    my ($filer_setup_result, $filer_cleanup_result, $test_result);

    if ($Stages{filer_setup}) {
        $Log->comment("Running STAGE filer_setup");
        eval { $self->filer_setup(); };
        if ($@) { 
            $filer_setup_result = $@;
            $Log->warn("filer_setup_result: $filer_setup_result");
            goto FILER_CLEANUP;
        }
    }

    my $topoclass = $self->topology()->get_class();

    #
    # Establish communication channel to share between mgr and supervisor
    #
    $self->_establish_channel(supervisors => \@clients);

    $Remote          = new RemoteExecution;
    $Local_Ancillary = new RemoteExecution;

    if ($Stages{test_main}) {

        # Index them, so logviewer can show in the order it started
        my $index = 1;
        while (my $client = $self->supervisor2start_next()) {

            # Run supervisor always - XXX Need improvement later?
            my $options = $Options_hash;
            $options->{P_EXE_MODE} = "supervisor";
            $options->{P_USE_TOPO} = $topoclass;
            $options->{CLIENTNAME} = $client;
            $options->{CLIENTFILE} = undef;
            $options->{TML_MGR_START_TIME} = $tml_mgr_start_time;
	    delete($options->{NATE_LIB});

            $Log->debug("Launching supervisor on $client");

            my $SubTst =
              $Remote->execute(
                        host  => "$client",
                        cmd   => $self->{myself},
                        runid => sprintf("supervisor_%03d_%s", $index, $client),
                        nate_params => $options,
                        handler     => sub { },);
            $index++;

            die config_result("failed to start supervisor on $client: $@")
              if !$SubTst;

            if ($Options_hash->{USE_STAF} && $SubTst) {

                my $clientsetup_options =
                  $self->clientsetup_options(client => $client);
                $SubTst->message_put(
                             message => freeze_and_escape($clientsetup_options),
                             type    => "TML/Spr/ClntStOpts");

                my $test_options = $self->test_options(client => $client);
                $SubTst->message_put(
                                    message => freeze_and_escape($test_options),
                                    type    => "TML/Spr/TstOpts");

            }

        }

    }

    # Start local tests if any
    $self->_start_local_tests();

    # Start ancillary tests if any
    $self->_start_ancillary_tests();

    if ($Stages{test_main}) {

        # wait until timeout expires
        $test_result =
          $self->process_supervisors(timeout  => $params{timeout},
                                     testname => $Testname,
                                     object   => $Remote);
        $Log->debug("process_supervisors returned: " . Dumper($test_result));

        # Core testing is done. kill all ancillary tests
        $Log->comment("Killing Ancillary Tests if Started Any");
        $Local_Ancillary->kill();
    }

  FILER_CLEANUP:

    # Update TERMFILE and TERMSTRING, if specified
    if ($Options_hash->{'TERMFILE'} && $Options_hash->{'TERMSTRING'}) {
        my $termfile =
          $Options_hash->{TOP_LOGDIR} . "/" . $Options_hash->{TERMFILE};
        if (open(TERM, ">>$termfile")) {
            print TERM $Options_hash->{TERMSTRING} . "\n";
            close TERM;
        } else {
            $Log->warn("Updating $termfile failed: $!", stacktrace => 0);
        }
    }

    eval {
        if ($Stages{filer_cleanup}) {
            $Log->comment("Running STAGE filer_cleanup");
            $self->filer_cleanup();
        }
    };
    $filer_cleanup_result = $@;

    if ($test_result) {
        if ($filer_setup_result) {
            $Log->warn("filer_setup error: $filer_setup_result");
        }
        if ($filer_cleanup_result) {
            $Log->warn("filer_cleanup error: $filer_cleanup_result");
        }
    } else {
        my $result = "";
        $result .= "filer_setup error: $filer_setup_result"
          if ($filer_setup_result);
        $result .= "filer_cleanup error: $filer_cleanup_result"
          if ($filer_cleanup_result);
        if ($result) {
            die fail_result("filer setup or cleanup failed: $result",
                            stacktrace => 0);
        }
    }

    if ($test_result) {
        $Log->trace("result = $test_result");
        if ($self->{stages_changed}) {
            $result_message .= " (with STAGE changes) ";
            $Log->trace("result_message set to $result_message");
        }
        $result_message .= " ";
        if (ref($test_result)) {
            $result_message .= join("\n", ($test_result->messages));
            $Log->trace("result_message set to $result_message");
        } else {
            $result_message .= $test_result;
            $Log->trace("result_message set to $result_message");
        }
        my $result_type = "FAIL";
        if (UNIVERSAL::isa($test_result, 'NATE::Result')) {
            $result_type = $test_result->type();
        }
        logresult($result_type, $result_message);
    } else {
        if ($self->{stages_changed}) {
            $result_message .= " (with STAGE changes) ";
            $Log->trace("result_message set to $result_message");
        }
        $result_message .= " on " . scalar(@clients) . " client(s)";

        # only log the result if it wasn't already logged (in a localhost mode)
        logresult("PASS", $result_message);
    }

    $Log->exit();
}

# supervisor mode does not have topology object associated. This is to minimize
# the topology data copied across all supervisor processes run on all clients.
# To access any data, it has to make RPC call to mgr(parent).
sub run_supervisor_mode {
    my ($self, %params) = @_;
    my ($instances, $failed_instances);
    $Log->enter();

    $Log->trace("TML running supervisor code...");

    $self->_init_supervisor();

    my $host = $Options_hash->{CLIENTNAME};
    my %opts;
    $opts{nateeval} = 0;
    $opts{cmd}      = ($params{cmd}) ? $params{cmd} : $self->{myself};
    $opts{nateeval} = $self->_detect_nateeval(cmd => $opts{cmd});

    my ($client_setup_result, $client_cleanup_result);
    eval {
        if ($Stages{client_setup}) {
            $Log->comment("Running STAGE client_setup");
            my $cliopts = $self->clientsetup_options(client => $host);

            $self->client_setup(%{$cliopts});
        }
    };
    $client_setup_result = $@;
    $Log->warn("client_setup error: $client_setup_result")
      if ($client_setup_result);
    goto CLIENT_CLEANUP if ($client_setup_result);

    #
    # Start a network trace if desired
    #
    $Log->trace(
           "Starting network trace on $Options_hash->{TRACE_ADDRS} optionally");
    my $traces;
    my $nodetraces = new Network::NodeTrace();
    if (ref($Options_hash->{TRACE_ADDRS}) eq 'ARRAY') {
        foreach my $addr (@{$Options_hash->{TRACE_ADDRS}}) {
            push(@{$traces}, new Network::Trace(address => $addr));
        }
    } else {
        push(@{$traces},
             new Network::Trace(address => $Options_hash->{TRACE_ADDRS}));
    }

    my $tstopts = $self->test_options(client => $host);

    # I hope this would avoid generating result objects and saving them into
    # NATE database which is very much centralized and only the ancestor ntest
    # should handle all of them. Here, I disable the saving part and get only the summary
    # of the workers failed.

    sub _worker_exit_handler {
        my ($subtest) = @_;

        my $worst = $Tharn::Test_DB->get_own_worst_result(
                                             subtest_id => $subtest->test_id());
        $failed_instances++ if ($worst && $worst->severity() > 1);
    }

    $Remote = new RemoteExecution;
    foreach my $instance ($self->_instance_sort(keys %{$tstopts})) {
        my $options = $Options_hash;
        foreach my $key (keys %{$tstopts->{$instance}}) {

            # Set test specific Options
            $options->{$key} = $tstopts->{$instance}{$key};
        }
        $options->{P_EXE_MODE} = "worker";

        #
        # This is required for the worker to get test_options based on the index/instance
        #
        $options->{instance} = $instance;

        $opts{host}        = "localhost";
        $opts{runid}       = sprintf("%03d_worker", $instance);
        $opts{nate_params} = $options;
        $opts{handler}     = \&_worker_exit_handler;

        # If TESTUSER specified for any test instance, it means that the test
        # should be run as that specific user, hence $Remote will take care of that.
        $opts{username} = $tstopts->{$instance}{TESTUSER}
          if ($tstopts->{$instance}{TESTUSER});
        $opts{password} = $tstopts->{$instance}{TESTUSER_PASSWORD}
          if ($tstopts->{$instance}{TESTUSER_PASSWORD});

        my %dopts = %opts;    # just for the debug
                              # not to print code string
        delete $dopts{cmd} if ($dopts{cmd} =~ /^\s*\{.*\}\s*$/s);
        $Log->debug(
                  "Remote Execute options for $instance: \n" . Dumper(\%dopts));
        $Remote->execute(%opts);
        $instances++;
    }

    # wait until timeout expires
    my $worker_result =
      $self->process_workers(timeout => $params{timeout},
                             object  => $Remote);
    $Log->trace("process_workers returned: " . Dumper($worker_result));

  CLIENT_CLEANUP:

    eval {
        if ($Stages{client_cleanup}) {
            $Log->comment("Running STAGE client_cleanup");
            $self->client_cleanup();
        }
    };
    $client_cleanup_result = $@;
    $Log->warn("client_cleanup error: $client_cleanup_result")
      if ($client_cleanup_result);

    if ($client_setup_result) {
        logresult("CONFIG", "client_setup failed: " . $client_setup_result);
    } elsif ($failed_instances) {
        logresult("FAIL",
                  "$failed_instances (out of $instances) workers failed");
    } else {
        logresult("PASS", "All workers ($instances) passed");
    }

    $Log->exit();
}

sub process_supervisors {

    $Log->enter();

    my ($self, %params) = @_;

    $Options_hash = $Options->get_hash();

    if ($Options_hash->{'STOP_ON_ERROR'} == 0) {

        if ($params{timeout} ne "disabled") {
            return $params{'object'}->wait(timeout => $params{timeout});
        } else {
            return $params{'object'}->wait();
        }
    } else {

        my $clients = $self->clients('status' => 'inuse');

        # Instead of using $Remote->wait() keep parsing the subtests to find
        # failures so we can exit immeadiately
        $Log->comment("Waiting for supervisor subtests to finish");
        my $test_result;
        my $i = 0;
        my (@passed, @failed, @running, @complete);
        ($test_result, $passed[$i], $failed[$i], $running[$i], $complete[$i])
            = $self->_check_subtest_status(
            'obj'   => $params{'object'},
            'regex' => 'supervisor'
            );

        my $client_loss = sprintf("%.2f", ($failed[$i] / $clients) * 100);

        $Log->comment("Supervisor status - "
                . "Clients:$clients "
                . "Loss:$client_loss "
                . "Pass:$passed[$i] "
                . "Failed:$failed[$i] "
                . "Running:$running[$i] "
                . "Complete:$complete[$i]");

        Tharn::sleep 60;
        while ($complete[$i] < $clients && $running[$i] > 0) {
            $i++;
            (   $test_result, $passed[$i], $failed[$i],
                $running[$i], $complete[$i]
                )
                = $self->_check_subtest_status(
                'obj'   => $params{'object'},
                'regex' => 'supervisor'
                );
            $client_loss = sprintf("%.2f", ($failed[$i] / $clients) * 100);

            # If the state of a subtest has changed print status
            if (   $passed[$i] != $passed[$i - 1]
                || $failed[$i] != $failed[$i - 1]
                || $running[$i] != $running[$i - 1]
                || $complete[$i] != $complete[$i - 1]) {

                $Log->comment("Supervisor status - "
                        . "Clients:$clients "
                        . "Loss:$client_loss "
                        . "Pass:$passed[$i] "
                        . "Failed:$failed[$i] "
                        . "Running:$running[$i] "
                        . "Complete:$complete[$i]");

             # Figure out if we need to fail due client loss > MAX_CLIENT_LOSS
             # otherwise just pass the result object
                if ($failed[$i] > 0) {
                    my $msg;
                    if ($Options_hash->{'MAX_CLIENT_LOSS'}) {
                        if ($client_loss > $Options_hash->{'MAX_CLIENT_LOSS'})
                        {
                            $msg
                                = "Client loss ($client_loss\%) exceeds "
                                . "MAX_CLIENT_LOSS ("
                                . $Options_hash->{'MAX_CLIENT_LOSS'} . "\%)";
                            $Log->comment($msg);
                            if ($clients > $complete[$i]) {
                                $self->_terminate_subtests(
                                    'obj'   => $params{'object'},
                                    'regex' => 'supervisor',
                                    'text' =>
                                        "Subtest killed by parent: $msg",
                                    'type' => 'comment'
                                );

                                # Wait for subtests to report
                                eval {
                                    $params{'object'}->wait(timeout => 300);
                                };
                            }
                            return $test_result;
                        }
                    } else {
                        $msg = @failed . 'failed supervisors detected';
                        $Log->comment(
                            $msg . ': stopping all active supervisors');
                        if ($clients > $complete[$i]) {
                            $self->_terminate_subtests(
                                'obj'   => $params{'object'},
                                'regex' => 'supervisor',
                                'text'  => "Subtest killed by parent: $msg",
                                'type'  => 'comment'
                            );

                            # Wait for subtests to report
                            eval { $params{'object'}->wait(timeout => 300); };
                        }
                        return $test_result;
                    }
                }
            }

            Tharn::sleep 60;
        }
    }
} ## end sub process_supervisors

sub process_workers {
    my ($self, %params) = @_;
    my $RmObj = $params{object};

    my $result;
    if ($params{timeout} ne "disabled") {
        $result = $RmObj->wait(timeout => $params{timeout});
    } else {
        $result = $RmObj->wait();
    }

    return $result;
}

my $Self;    # Work around to pass coderef to iterate

sub run_worker_mode {
    my $self = shift;
    $Log->enter();
    $Self = $self;

    iterate(subroutine => \&run_test_code);

    $Log->exit();
}

sub run_test_code {
    my $self = $Self;

    $Log->enter();
    my ($setup_result, $test_result, $cleanup_result);

    eval {
        if ($Stages{test_setup}) {
            $Log->comment("Running STAGE test_setup");
            $self->test_setup();
        }
    };
    $setup_result = $@;

    eval {
        if (!$setup_result && $Stages{test_main}) {
            $Log->comment("Running STAGE test_main");
            $self->test_main();
        }
    };
    $test_result = $@;

    eval {
        if ($Stages{test_cleanup}) {
            $Log->comment("Running STAGE test_cleanup");
            $self->test_cleanup();
        }
    };
    $cleanup_result = $@;

    if ($cleanup_result && ($test_result || $setup_result)) {
        $Log->warn("Cleanup failure: $cleanup_result");
    }

    if (ref($setup_result)) {
        die $setup_result;
    } elsif ($setup_result) {
        die config_result("test_setup failed: $setup_result", stacktrace => 0);
    }

    if ($cleanup_result) {
        die fail_result("Cleanup failure: $cleanup_result", stacktrace => 0);
    }

    if ($test_result) {
        die fail_result("test_main failed: " . $test_result, stacktrace => 0);
    }
    $Log->exit();
}

sub _define_stages {
    my $self = shift;
    $Log->enter();
    my ($stages, $skip) = $self->get(parameter => [qw(STAGE SKIPSTAGE)]);

    my %stages;

    my @default_stages = keys %Stages;
    my $changes        = undef;

    # If this option is defined, then the user only wants to run certain stages.
    # Check them out for validity.
    if (scalar @{$stages}) {
        $changes = 1;
        foreach my $stage (@{$stages}) {
            my @tstages = grep(/$stage/i, @default_stages);
            if ($#tstages > 0) {
                my $stages = join(",", @tstages);
                die config_result("conflict with $stage: $stages\n",
                                  stacktrace => 0);
            }
            if ($#tstages == -1) {
                my $msg =
                  "$stage is an unsupported stage for this test\n  supported stages: ";
                $msg .= join(", ", @default_stages) . "\n";
                die config_result($msg, stacktrace => 0);
            }
            $stages{$tstages[0]} = 1;
        }
        %Stages = %stages;
    }

    # check for any stages to skip before checking the stage option so that it
    # can complain about unsupported stages
    if (scalar @{$skip}) {
        $changes = 1;
        foreach my $stage (@{$skip}) {
            my @tstages = grep(/$stage/i, @default_stages);
            if ($#tstages > 0) {
                my $stages = join(",", @tstages);
                die config_result("conflict with skip '$stage': $stages\n",
                                  stacktrace => 0);
            }
            if ($#tstages == -1) {
                my $msg =
                  "$stage is an unsupported SKIPSTAGE for this test\n  supported stages: ";
                $msg .= join(", ", @default_stages) . "\n";
                die config_result($msg, stacktrace => 0);
            }
            $Stages{$tstages[0]} = undef;
        }
    }
    $self->{stages_changed} = $changes;

    $Log->exit();
}

sub modify_stage {

    # The following two lines setup the hash that is used by validate_mixed to
    # validate the incoming parameter names. The resulting hash will have keys
    # for each of the available stages that are defined in the master %Stages
    # and specifies that they are optional to this method and can either be
    # scalars or undef. Doing this allows for new stages to be defined just in
    # the master list and not here as well.
    my %stage_definitions;
    grep($stage_definitions{$_} = {type => SCALAR | UNDEF, optional => 1},
         keys %Stages_default);
    my ($self, %params) =
      validate_mixed(@_, [{type => OBJECT, isa => 'TestManagerLite'}],
                     \%stage_definitions,);
    $Log->enter();

    while (my ($stage, $value) = each %params) {
        $Stages{$stage} = $value;
    }

    $Log->exit();
}

=head2 defined_stage

    if ($test->defined_stage(stage=>"filer_setup"))

C<defined_stage> is a utility function which will tell the caller whether
or not the specified stage is enabled for this test run.

=over

=item stage

Specifies the stage that should be checked. The question being asked is
whether or not this stage is enabled for this run.

=cut

sub defined_stage {
    $Log->enter();
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT, isa => 'TestManagerLite'}],
                     {stage => {type => SCALAR},});

    $Log->exit("return $Stages{$params{stage}}");
    return $Stages{$params{stage}};
}

sub _find_relative_testname {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT, isa => 'TestManagerLite'}],
                     {testname => {type => SCALAR},});
    my $testname;
    $Log->enter("test = $params{testname}");
    $Log->debug("Options_hash: " . Dumper($Options_hash));

    if ($params{testname} =~ m!^/!) {

        # get rid of any '..' entries in the path so that the foreach loop below will work
        if (my $real_testname = realpath($params{testname})) {
            $testname = $real_testname;
        } elsif ($self->{cifs}) {
            $testname = $params{testname};
            push(@{$self->{warnings}},
                 "The test may not work correctly on CIFS clients since realpath('$testname') failed"
                 );
        }

        $Log->debug('Matching the full path to paths defined in NATE_LIB');

        # Try to find the original relative testname rather than the fqname.
        foreach my $path (split(/;/, $Options_hash->{NATE_LIB})) {
            $path = quotemeta($path);
            $Log->debug("Matching $path against $params{testname}");
            if ($params{testname} =~ m/^$path/) {
                $testname =~ s!$path/!!;
                $Log->debug(
                    "Converted '$params{testname}' to '$testname' because of nate_lib '$path'"
                    );
            }
        }
        if (!$testname || $testname eq "") {
            push(@{$self->{warnings}},
                 "Could not locate '$params{testname}' in NATE_LIB");
            $testname = $params{testname};
        }
    } else {
        $testname = $params{testname};
    }

    $Log->exit("return $testname");
    return $testname;
}

sub _print_warnings {
    my $self = shift;
    if ($self->{warnings}) {
        foreach my $warn (@{$self->{warnings}}) {
            $Log->warn($warn);
        }
    }
}

sub _detect_nateeval {
    $Log->enter();
    my ($self, %opts) = @_;
    if (ref($opts{cmd}) eq "CODE") {    # Coderef
        return 1;
    } elsif ($opts{cmd} =~ /^\s*\{.*\}\s*$/s) {    # Codestring
        return 1;
    } elsif ($opts{cmd} =~ /\.(?:thpl|stpl)$/) {    # thpl
        $Log->exit();
        return 0;
    } else {                                        # Executable?
        $Log->exit();
        return 1;
    }
}

sub _filer_setup_serial {
    $Log->enter();
    my ($self, %opts) = @_;

    my $parentProcStaf = TestManagerLite::ParentStaf->new();

    my $runningSetupNum = undef;
    my $sleepTime       = 60;
    my $setupNum        = $opts{setup_number};
    $Log->comment(
        "Setup is to be run in a serial fashion. This test is number $setupNum in line to run."
        );

    # Look to see which script is currently running setup
    $runningSetupNum = $parentProcStaf->var_get(var => "SERIAL_SETUP");

    # if SERIAL_SETUP is undef, set it to one so that the first script will run 
    if (($runningSetupNum == undef) and ($setupNum == 1)) { 
        $parentProcStaf->var_set(var => "SERIAL_SETUP", val => 1);
    } elsif ($setupNum != 1) { 
        # wait our turn    
        $parentProcStaf->var_wait(var => "SERIAL_SETUP", val => $setupNum);
    }

    # run setup
    $self->topology()->filer_setup();

    # trigger the next one to run
    $parentProcStaf->var_set(var => "SERIAL_SETUP", val => ++$setupNum);

    $Log->exit();
    return;
}

sub _handle_setup_complete {
    $Log->enter();
    my ($self) = @_;

    my $parentProcStaf = TestManagerLite::ParentStaf->new();

    my $sleepTime = 60;
    if (my $set = $Options->get(option => "SERIAL_SETUP_COMPLETE")) {
        $Log->comment(
            "This script has completed filer_setup. Signaling SERIAL_SETUP_COMPLETE to parent. "
              . "Any scripts waiting to complete filer_setup should now proceed."
              );
        $parentProcStaf->var_set(var => "SERIAL_SETUP_COMPLETE", val => 1);

    } elsif (my $wait = $Options->get(option => "SERIAL_SETUP")) {
        $Log->comment(
            "waiting for signal SERIAL_SETUP_COMPLETE before completing filer_setup "
            );

        # Wait for the SERIAL_SETUP_COMPLETE value to appear in the parent
        $parentProcStaf->wait_setup_complete();
    }

    $Log->exit();
}

sub _define_topology_serial {
    $Log->enter();
    my ($self, %opts) = @_;

    my $parentProcStaf = TestManagerLite::ParentStaf->new();

    my $runningSetupNum = undef;
    my $sleepTime       = 60;
    my $setupNum        = $opts{setup_number};
    $Log->comment(
        "define_topology is to be run in a serial fashion. This test is number $setupNum in line to run."
        );   

    # Look to see which script is currently running setup
    $runningSetupNum = $parentProcStaf->var_get(var => "SERIAL_DEFINE");

    # if SERIAL_DEFINE is undef, set it to one so that the first script will run
    if (($runningSetupNum == undef) and ($setupNum == 1)) {
        $parentProcStaf->var_set(var => "SERIAL_DEFINE", val => 1);
    } elsif ($setupNum != 1) {
        # wait our turn
        $parentProcStaf->var_wait(var => "SERIAL_DEFINE", val => $setupNum);
    }

    # run define_topology
    $self->topology()->define_topology();

    # trigger the next one to run
    $parentProcStaf->var_set(var => "SERIAL_DEFINE", val => ++$setupNum);

    $Log->exit();
    return;
}

sub _handle_define_topology_complete {
    $Log->enter();
    my ($self) = @_;

    my $parentProcStaf = TestManagerLite::ParentStaf->new();

    my $sleepTime = 60;
    if (my $set = $Options->get(option => "SERIAL_SETUP_COMPLETE")) {
        $Log->comment(
            "This script has completed define_topology. Signaling SERIAL_DEFINE_COMPLETE to parent. "
              . "Any scripts waiting to complete define_topology should now proceed."
              ); 
        $parentProcStaf->var_set(var => "SERIAL_DEFINE_COMPLETE", val => 1);

    } elsif (my $wait = $Options->get(option => "SERIAL_SETUP")) {
        $Log->comment(
            "waiting for signal SERIAL_DEFINE_COMPLETE before completing define_topology "
            ); 

        # Wait for the SERIAL_DEFINE_COMPLETE value to appear in the parent
        $parentProcStaf->wait_define_complete();
    } 

    $Log->exit();
}


sub _decide_staf_usage {
    my $use_staf = shift;
    $Log->enter("use_staf = $use_staf");
    my $nate_version = NATE::version_string();
    $nate_version =~ /(\d+)\.+/;
    $nate_version = $1;
    if ($use_staf eq "0.999") {

        # This means user didn't set any value for this.
        if ($nate_version >= 3) {
            $use_staf = 1;
        } else {
            $use_staf = 0;
        }
    } elsif (($use_staf == 1) && ($nate_version < 3)) {
        $Log->warn(  "STAF is not available in NATE VERSION : "
                   . NATE::version_string()
                   . ". Setting USE_STAF to 0");
        $use_staf = 0;

    }
    $Options->set(USE_STAF => $use_staf);
    $Log->exit("USE_STAF = $use_staf");
}

###
# Instances sort
###

sub _instance_sort {
    my ($self, @instances) = @_;
    my %tobesorted;
    my $length = 3;
    foreach my $inst (@instances) {
        $length = length($inst) if ($length < length($inst));
    }

    foreach my $inst (@instances) {
        $tobesorted{$inst} = sprintf("%0${length}s", $inst);
    }

    return sort { $tobesorted{$a} <=> $tobesorted{$b} } (keys %tobesorted);
}

sub _check_subtest_status {

    $Log->enter();

    my ($self, %params) = validate_mixed(
        @_,
        [{type => OBJECT, isa => 'TestManagerLite'}],
        {   obj   => {type => OBJECT, optional => 0},
            regex => {type => SCALAR, optional => 0},
        }
    );

    my $passed   = 0;
    my $failed   = 0;
    my $running  = 0;
    my $complete = 0;

    # Parse all the subtest objects for the latest runids
    my @results;
    my @subtests = $params{'obj'}->Subtest::get_all();
    if (@subtests) {
        foreach my $s (@subtests) {
            if ($s->runid() =~ /$params{'regex'}/) {

                # If is_alive returns true then the subtest is considered
                # to still be running. Otherwise we figure out the worst
                # result and override the abortiveness if it's in the list
                # of errors
                $Log->debug("Found runid " . $s->runid());
                if ($s->is_alive()) {
                    $running++;
                } else {

                    # Save the worst results for all the children
                    $Log->debug("is_alive returned false "
                            . $s->is_alive()
                            . ". Getting the worst result");
                    my $worst = $Tharn::Test_DB->get_worst_result(
                        parent_id => $s->test_id());
                    if (defined($worst)) {
                        $Log->debug("Worst result for runid "
                                . $s->runid() . "\n"
                                . Dumper($worst));

                        # Change errors 1 and 30 to be non abortive
                        my $type = $worst->type();
                        if ($type eq 'SCRIPT') {
                            my @errors = (1, 30);
                            $Log->debug("Checking for the "
                                    . "following errors "
                                    . Dumper(\@errors));
                            my @messages = $worst->messages();
                            foreach my $error (@errors) {
                                my $msg = "Subtest exited with status $error";
                                if (grep { $_ eq $msg } @messages) {
                                    $Log->debug(
                                        "Overriding the abortive flag on subtest "
                                            . $s->runid());
                                    $worst->abortive(0);
                                    last;
                                }
                            }
                        }

                        push(@results, $worst);

                        if ($worst->type() !~ /(PASS|WARN|INFO)/) {
                            $failed++;
                            $complete++;
                        } else {
                            $complete++;
                            $passed++;
                        }
                    }
                }
            }
        }
    }

    # Return the worst result of all children with stats
    my $worst = NATE::Result::worst(@results);
    my @return = ($worst, $passed, $failed, $running, $complete);
    $Log->debug("Returning worst result with stats " . Dumper(\@return));
    $Log->exit();
    return @return;
} ## end sub _check_subtest_status

sub _terminate_subtests {

    $Log->enter();

    my ($self, %params) = validate_mixed(
        @_,
        [{type => OBJECT, isa => 'TestManagerLite'}],
        {   obj   => {type => OBJECT, optional => 0},
            regex => {type => SCALAR, optional => 0},
            text  => {type => SCALAR, optional => 0},
            type  => {type => SCALAR, optional => 0},
        }
    );

    my $text = $params{'text'};
    my $type = $params{'type'};

    # Parse all the subtest objects for the latest runids
    my @alive = $params{'obj'}->Subtest::get_live();
    foreach my $subtest (@alive) {
        my $runid = $subtest->runid();
        if ($runid =~ /$params{'regex'}/ && $subtest->is_alive()) {
            $Log->comment("Killing subtest $runid");
            eval { $subtest->kill('message' => [$type, $text]); };
        }
    }

    $Log->exit();
} ## end sub _terminate_subtests

1;
