#Hey emacs, this is basically -*- perl -*-
# $Id: $

package TestManagerLite::CIFS;

use strict;

use Tharn;
$Log->comment("Starting " . __FILE__);

use base qw(TestManagerLite);
use English;
use Params::Validate qw(
  validate
  SCALAR
  HASHREF
  ARRAYREF
  OBJECT
  UNDEF
  CODEREF);
use Data::Dumper qw(Dumper);
use TestUtils qw(iterate expected_run_time);

my $Options;
my $Options_hash;
my $Self;

$Options = new Options;
$Options->parameters(
           definitions => [
               {name         => "MGR_WAIT_MODE",
                spec         => "s",
                default      => 'active',
                valid_values => ['', 'active', 'passive'],
                description => "Determines how we wait for supervisor subtests",
               },
               {name    => 'TERMSTRINGS',
                spec    => "s@",
                default => "TeRmStRiNg",
                description =>
                  'A comma seperated list of strings to look for in TERMFILE',
               },
             ]
             );

# Define the basic stages here. The %Stages hash can be changed, this one is not.
my %Stages_default = (filer_setup    => 1,
                      client_setup   => 1,
                      test_setup     => 1,
                      test_main      => 1,
                      test_cleanup   => 1,
                      client_cleanup => 1,
                      filer_cleanup  => 1,
                      );

# This hash will be used to manage the stages to run through. If a STAGE param is passed,
#  then the contents of this hash are replaced with just the specified stages.
my %Stages = %Stages_default;

sub _check_subtest_status {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   obj   => {type => OBJECT, optional => 0},
            regex => {type => SCALAR, optional => 0},
        }
    );

    my $passed   = 0;
    my $failed   = 0;
    my $running  = 0;
    my $complete = 0;

    # Parse all the subtest objects for the latest runids
    my @results;
    my @subtests = $opts{'obj'}->Subtest::get_all();
    if (@subtests) {
        foreach my $s (@subtests) {
            if ($s->runid() =~ /$opts{'regex'}/) {

                # If is_alive returns true then the subtest is considered
                # to still be running. Otherwise we figure out the worst
                # result and override the abortiveness if it's in the list
                # of errors
                $Log->debug("Found runid " . $s->runid());
                if ($s->is_alive()) {
                    $running++;
                } else {

                    # Save the worst results for all the children
                    $Log->debug("is_alive returned false "
                            . $s->is_alive()
                            . ". Getting the worst result");
                    my $worst = $Tharn::Test_DB->get_worst_result(
                        parent_id => $s->test_id());
                    if (defined($worst)) {
                        $Log->debug("Worst result for runid "
                                . $s->runid() . "\n"
                                . Dumper($worst));

                        # Change errors 1 and 30 to be non abortive
                        my $type = $worst->type();
                        if ($type eq 'SCRIPT') {
                            my @errors = (1, 30);
                            $Log->debug("Checking for the "
                                    . "following errors "
                                    . Dumper(\@errors));
                            my @messages = $worst->messages();
                            foreach my $error (@errors) {
                                my $msg = "Subtest exited with status $error";
                                if (grep { $_ eq $msg } @messages) {
                                    $Log->debug(
                                        "Overriding the abortive flag on subtest "
                                            . $s->runid());
                                    $worst->abortive(0);
                                    last;
                                }
                            }
                        }
                    }

                    push(@results, $worst);

                    if ($worst->type() !~ /(PASS|WARN|INFO)/) {
                        $failed++;
                        $complete++;
                    } else {
                        $complete++;
                        $passed++;
                    }
                }
            }
        }
    }

    # Return the worst result of all children with stats
    my $worst = NATE::Result::worst(@results);
    my @return = ($worst, $passed, $failed, $running, $complete);
    $Log->debug("Returning worst result with stats " . Dumper(\@return));
    $Log->exit();
    return @return;
} ## end sub _check_subtest_status

sub _terminate_subtests {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   obj   => { type => OBJECT, optional => 0 },
            regex => { type => SCALAR, optional => 0 },
            text  => { type => SCALAR, optional => 0 },
            type  => { type => SCALAR, optional => 0 },
        }
    );

    my $text = $opts{'text'};
    my $type = $opts{'type'};

    #-----------------------------------------------------
    # Parse all the subtest objects for the latest runids
    #-----------------------------------------------------
    my @alive = $opts{'obj'}->Subtest::get_live();
    foreach my $subtest (@alive) {
        my $runid = $subtest->runid();
        if ( $runid =~ /$opts{'regex'}/ ) {
            $Log->comment("Killing subtest $runid");
            eval { $subtest->kill( 'message' => [ $type, $text ] ); };
        }
    }

    $Log->exit();
}

sub process_supervisors {

    $Log->enter();

    my ($self, %params) = @_;

    $Options_hash = $Options->get_hash();

    my $clients = $self->clients('status' => 'inuse');
    if ($self->topology()->{'max_clients'} > 0) {
        $clients = $self->topology()->{'max_clients'};
    }

    if ($Options_hash->{'MGR_WAIT_MODE'} eq 'passive') {

        if ($params{timeout} ne "disabled") {
            return $params{'object'}->wait(timeout => $params{timeout});
        } else {
            return $params{'object'}->wait();
        }
    } else {

        #-----------------------------------------------------------
        # Set the right path for the CSV files and create if needed
        #-----------------------------------------------------------
        my $csv_path = $Options_hash->{'TOP_LOGDIR'};
        if ($Options_hash->{'CSV_DIR'}) {
            $csv_path .= "/$Options_hash->{'CSV_DIR'}";
        }
        if (!-e $csv_path) {
            eval { mkdir($csv_path); };
            if ($@) {
                $Log->warn("Error creating $csv_path: $@");
            }
        }

        my $alt_csv_path = $Options_hash->{'ALT_LOGDIR'};
        if ($Options_hash->{'CSV_DIR'}) {
            $alt_csv_path .= "/$Options_hash->{'CSV_DIR'}";
        }
        if ($Options_hash->{'ALT_LOGDIR'} && !-e $alt_csv_path) {
            eval { mkdir($alt_csv_path); };
            eval { mkdir($csv_path); };
            if ($@) {
                $Log->warn("Error creating $alt_csv_path: $@");
            }
        }

        #----------------------------
        # Initialize the log file(s)
        #----------------------------
        my $csv_file =
          "$params{'testname'}-$Options_hash->{'SCRIPT_INDEX'}.csv";
        $Log->comment("Initializing CSV file $csv_path/$csv_file");
        open(LOG, ">>$csv_path/$csv_file")
          || return config_result("Unable to open $csv_path/$csv_file");
        print LOG "Date,Active Clients,Failed Clients,% Client Loss\n";
        close LOG;

        if ($Options_hash->{'ALT_LOGDIR'}) {
            $Log->comment(
                "Initializing alternate CSV file " . "$alt_csv_path/$csv_file");
            open(LOG, ">>$alt_csv_path/$csv_file")
              || return config_result("Unable to open $alt_csv_path/$csv_file");
            print LOG "Date,Active Clients,Failed Clients,% Client Loss\n";
            close LOG;
        }

        #----------------------------------------------------------------------
        # Instead of using $Remote->wait() keep parsing the subtests to find
        # failures so we can exit immeadiately
        #----------------------------------------------------------------------
        $Log->comment("Waiting for supervisor subtests to finish");
        my $test_result;
        my $i = 0;
        my (@passed, @failed, @running, @complete);
        ($test_result, $passed[$i], $failed[$i], $running[$i], $complete[$i]) =
          _check_subtest_status('obj'   => $params{'object'},
                                'regex' => 'supervisor');

        my $client_loss = sprintf("%.2f", ($failed[$i] / $clients) * 100);

        $Log->comment(  "Supervisor status - "
                      . "Clients:$clients "
                      . "Loss:$client_loss "
                      . "Pass:$passed[$i] "
                      . "Failed:$failed[$i] "
                      . "Running:$running[$i] "
                      . "Complete:$complete[$i]");

        snooze 60;
        while ($complete[$i] < $clients && $running[$i] > 0) {
            $i++;
            ($test_result, $passed[$i], $failed[$i],
             $running[$i], $complete[$i])
              = _check_subtest_status('obj'   => $params{'object'},
                                      'regex' => 'supervisor');
            $client_loss = sprintf("%.2f", ($failed[$i] / $clients) * 100);

            #-----------------------
            # Updated the CSV files
            #-----------------------
            use POSIX qw(strftime);
            my $log_line = strftime("%m%d%y%H%M%S", localtime());
            $log_line .= ",$running[$i],$failed[$i],$client_loss";
            open(LOG, ">>$csv_path/$csv_file")
              || return config_result("Unable to open $csv_path/$csv_file");
            print LOG "$log_line\n";
            close LOG;

            if ($Options_hash->{'ALT_LOGDIR'}) {
                open(LOG, ">>$alt_csv_path/$csv_file")
                  || return config_result(
                                      "Unable to open $alt_csv_path/$csv_file");
                print LOG "$log_line\n";
                close LOG;
            }

            #----------------------------------------------------
            # If the state of a subtest has changed print status
            #----------------------------------------------------
            if (   $passed[$i] != $passed[$i - 1]
                || $failed[$i] != $failed[$i - 1]
                || $running[$i] != $running[$i - 1]
                || $complete[$i] != $complete[$i - 1]) {

                $Log->comment(  "Supervisor status - "
                              . "Clients:$clients "
                              . "Loss:$client_loss "
                              . "Pass:$passed[$i] "
                              . "Failed:$failed[$i] "
                              . "Running:$running[$i] "
                              . "Complete:$complete[$i]");

                #-----------------------------------------------------------------
                # Figure out if we need to fail due client loss > MAX_CLIENT_LOSS
                # otherwise just pass the result object
                #-----------------------------------------------------------------
                if ( $failed[$i] > 0 ) {
                    my $msg;
                    if ( $Options_hash->{'MAX_CLIENT_LOSS'} ) {
                        if ( $client_loss
                            > $Options_hash->{'MAX_CLIENT_LOSS'} )
                        {
                            $msg
                                = "Client loss ($client_loss\%) exceeds "
                                . "MAX_CLIENT_LOSS ("
                                . $Options_hash->{'MAX_CLIENT_LOSS'} . "\%)";
                            $Log->comment($msg);
                            if ( $clients > $complete[$i] ) {
                                _terminate_subtests(
                                    'obj'   => $params{'object'},
                                    'regex' => 'supervisor',
                                    'text' =>
                                        "Subtest killed by parent: $msg",
                                    'type' => 'Fail'
                                );

                                # Wait for subtests to report
                                eval {
                                    $params{'object'}->wait( timeout => 60 );
                                };
                            }
                            return $test_result;
                        }
                    } else {
                        $msg = @failed . 'failed supervisors detected';
                        $Log->comment(
                            $msg . ': stopping all active supervisors' );
                        if ( $clients > $complete[$i] ) {
                            _terminate_subtests(
                                'obj'   => $params{'object'},
                                'regex' => 'supervisor',
                                'text'  => "Subtest killed by parent: $msg",
                                'type'  => 'Fail'
                            );

                            # Wait for subtests to report
                            eval { $params{'object'}->wait( timeout => 60 ); };
                        }
                        return $test_result;
                    }
                }
            }

            #-------------------------------------------------------------------
            # If infinity was specified, check termfile for termination strings
            #-------------------------------------------------------------------
            if (   $Options_hash->{'ITERATE'}
                && $Options_hash->{'TERMFILE'}
                && $Options_hash->{'TERMSTRINGS'}
                && $Options_hash->{'ITERATE'} eq 'infinity'
                && -e "$Options_hash->{'TOP_LOGDIR'}/$Options_hash->{'TERMFILE'}"
              ) {
                eval {
                    _check_term_file(
                        'term_file' =>
                          "$Options_hash->{'TOP_LOGDIR'}/$Options_hash->{'TERMFILE'}",
                        'term_strings' => $Options_hash->{'TERMSTRINGS'}
                        );
                };

                if ($@ =~ /^All TERMSTRINGS found/) {
                    return pass_result($@);
                } elsif ($@) {
                    return fatal_result($@);
                }

            }
            snooze 60;
        }
    }
}

sub _check_term_file {

    $Log->enter();

    my (%opts) =
      validate(@_,
               {term_file    => {type => SCALAR,   optional => 0},
                term_strings => {type => ARRAYREF, optional => 0},
               }
               );

    $Log->comment("Checking termination file $opts{'term_file'}");

    my @term_file_contents = ();
    open(TERM, "$opts{'term_file'}");
    while (<TERM>) {
        chomp;
        push(@term_file_contents, $_);
    }
    close TERM;

    my $all_strings_found = 1;
    foreach my $s (@{$opts{'term_strings'}}) {
        if (!grep(/$s/, @term_file_contents)) {
            $all_strings_found = 0;
            last;
        }
    }

    if (!$all_strings_found) {
        $Log->comment(  "Not all TERMSTRINGS exist in termination "
                      . "file, continuing\n@term_file_contents");
    }

    $Log->exit();

    if ($all_strings_found) {
        die "All TERMSTRINGS found in "
          . "termination file, exiting\n@term_file_contents";
    }

}

sub run_worker_mode {
    my $self = shift;
    $Log->enter();
    $Self = $self;

    iterate(subroutine => \&run_test_code);

    $Log->exit();
}

sub run_test_code {
    my $self = $Self;

    $Log->enter();
    my ($setup_result, $test_result, $cleanup_result);

    eval {
        if ($Stages{test_setup}) {
            $Log->comment("Running STAGE test_setup");
            $self->test_setup();
        }
    };
    $setup_result = $@;

    eval {
        if (!$setup_result && $Stages{test_main}) {
            $Log->comment("Running STAGE test_main");
            $self->test_main();
        }
    };
    $test_result = $@;

    eval {
        if ($Stages{test_cleanup}) {
            if (!$Options_hash->{'POST_FAILURE_CLEANUP'} && $test_result) {

                $Log->comment(
                         'POST_FAILURE_CLEANUP=0, Skipping STAGE test_cleanup');
            } else {

                $Log->comment("Running STAGE test_cleanup");
                $self->test_cleanup();
            }
        }
    };
    $cleanup_result = $@;

    if ($cleanup_result && ($test_result || $setup_result)) {
        $Log->warn("Cleanup failure: $cleanup_result");
    }

    if (ref($setup_result)) {
        die $setup_result;
    } elsif ($setup_result) {
        die config_result("test_setup failed: $setup_result", stacktrace => 0);
    }

    if ($cleanup_result) {
        die fail_result("Cleanup failure: $cleanup_result", stacktrace => 0);
    }

    if ($test_result) {
        die fail_result("test_main failed: " . $test_result, stacktrace => 0);
    }
    $Log->exit();
}

1;
