## @summary    Access values in the ntest parent
##
## @author      Jd dascenj@netapp.com
##
## @pod here

=head1 NAME

TestManagerLite::ParentStaf - Communicate with the Tharn::parent process

=head1 SYNOPSIS

    use TestManagerLite::ParentStaf;

    $comm = TestManagerLite::ParentStaf->new();

=head1 DESCRIPTION

    This library is meant to interact with variables in the process
    which is the parent to TestManagerLite (or others). Among other things, 
    this library is used to provide a means for (stest) scripts
    to communicate the status of their respective filer_setup methods. 

=head1 TODO

=head1 SEE ALSO

Options TestManagerLite Topology

=head1 BUGS

    Bug free code. 

=head1 STATUS

  Shared

=head1 AUTHOR

  Jd (dascenj@netapp.com)
  Created 18 Dec 2010 

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=head1 METHODS

=cut

package TestManagerLite::ParentStaf;

use base qw(TestManagerLite);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use Data::Dumper qw(Dumper);
use NATE::Events;
use Tharn qw ($Log script_result);

my $noExist = "VariableDoesNotExist";
my $options = new Options;

=head2 new

    my $comm = TestManagerLite::ParentStaf->new();

    C<new> is the invocant for the TestManagerLite::ParentStaf object.

=cut

sub new {
    my ($invocant, %opts) = @_;
    my $class = ref($invocant) || $invocant;
    my $self = {};
    bless($self, $class);

    return $self;
}

=head2 wait_setup_complete 

    my $comm = TestManagerLite::ParentStaf->new();
    $comm->wait_setup_complete();

    Use this method to cause a script to sleep until the value of 
    SERIAL_SETUP_COMPLETE is true at the parent.  This is useful
    specifically when a series of TestManagerLite tests must complete
    their filer_setup routines before this test may start. 

    Keep in mind that this method will have no affect if the
    SKIPSTAGE =~ /filer_setup/. 

=cut

sub wait_setup_complete {
    $Log->enter();
    my ($self) = @_;

    my @skipStage = @{$options->get(option => "SKIPSTAGE")};

    # Just return if we are skipping setup anyway.
    if (grep(/filer_setup/i, @skipStage )) {
        $Log->trace(
            "SKIPSTAGE filer_setup is true. There is no waiting on runway #2349er"
            );
        $Log->exit();
        return;
    }

    $Log->comment("Waiting for SERIAL_SETUP_COMPLETE to complete....");

    $self->var_wait(var => "SERIAL_SETUP_COMPLETE", val => 1);

    return;
}

=head2 wait_define_complete 

    my $comm = TestManagerLite::ParentStaf->new();
    $comm->wait_define_complete();

    Use this method to cause a script to sleep until the value of 
    SERIAL_DEFINE_COMPLETE is true at the parent. This is useful 
    specifically when a series of TestManagerLite tests must complete
    their define_topology routines before this test may start. 

    In contrast to wait_setup_complete, this method is still executed when
    SKIPSTAGE =~ /filer_setup/. 

=cut

sub wait_define_complete {
    $Log->enter();
    my ($self) = @_;

    $Log->comment("Waiting for SERIAL_DEFINE_COMPLETE to complete....");

    $self->var_wait(var => "SERIAL_DEFINE_COMPLETE", val => 1);

    return;
}

=head2 var_wait

    TestManagerLite::ParentStaf->wait_var(var = TEST, 
                                          val = 1);
 
    Wait for the given variable <var> to appear in the parent, and for 
    the value of said variable to be equal to <val>. The value of 
    TEST is expected to be set be some other script that needs to 
    communicate some information to this one.

=cut

sub var_wait {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {var => {type => SCALAR, optional => 0},
                      check_interval =>
                        {type => SCALAR, default => 60, optional => 1},
                      val => {type => SCALAR, optional => 0},},);

    my $wait = $opts{check_interval};
    my $var  = $opts{var};
    my $val  = $opts{val};

    $Log->comment(
           "Waiting for parent variable $var == $val before continuing...");

    # Wait for the value to appear in the parent
    while (1) {
        my $result = $self->var_get(var => $var);

        if (($result) and ($result =~ /^$val$/)) {
            $Log->exit();
            return;
        } else {  
            $Log->debug(
                "Waiting for parent (handle: " . $Tharn::Parent->handle() . " endpoint: " . $Tharn::Parent->endpoint() . 
                                     ") variable $var == $val. It is currently set as ($result). Sleeping $wait..."
                );
            # Just wait around for awhile
            NATE::Events::eventloop(seconds => $wait);
        }
    }
}

=head2 var_set

    TestManagerLite::ParentStaf->set_varl(var = TEST, 
                                           val = 1);
 
    Set the value of the given variable <var> in the parent 
    to the value of <val>.

=cut

sub var_set {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {var => {type => SCALAR, optional => 0},
                      val => {type => SCALAR, optional => 0},},);

    my $var = $opts{var};
    my $val = $opts{val};

    $Log->trace("Setting variable $var to value $val");
    my $result = eval{$Tharn::Staf->var_set(endpoint => $Tharn::Parent->endpoint(),
                                            handle   => $Tharn::Parent->handle(),
                                            var      => {$var => $val})};

    die script_result(
        "Could not set the value $var == $val at the parent. Error: $@"
        ) if ($@);

    $Log->exit();
    return;
}

=head2 var_get

    TestManagerLite::ParentStaf->var_get(var = TEST, 
                                           val = 1);
 
    Get the value of the given variable <var> in the parentx

=cut

sub var_get { 
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {var => {type => SCALAR, optional => 0},},);

    my $var = $opts{var};

    my $result = eval {
        $Tharn::Staf->var_get(endpoint => $Tharn::Parent->endpoint(),
                              handle   => $Tharn::Parent->handle(),
                              var      => "$var");
    };

    # check error           
    if ($@) { 
       if (grep(/$noExist/i, $@)) {
            $result = undef;
        } else { 
            die script_result("An error occurred communicating with the parent process $@");
        }
    }
    
    $Log->exit();
    return $result;
}

1;
