#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/TestManagerLite/Worker.pm#1 $

## @summary     Test Execution Controller for Stress Scalability Tests - Worker Class.
##
## @author      gvg@netapp.com
##
## @pod here

=head1 NAME

TestManagerLite::Worker - Test Execution Controller, Worker Class.
This is a helper class that interacts with supervisor process/Staf
on behalf of each worker to get the paramset for each execution.

=head1 SYNOPSIS

    package TestManagerLite::Worker;
    my $worker = new TestManagerLite::Worker(params =>
                                               { DEST_IP => { type => SCALAR }, },
                                               { VOLPATH => { type => SCALAR }, },
                                            );

    $worker->init();

    while (my $options = $worker->nextset()) {
        -- Code --
    }


=head1 DESCRIPTION

C<TestManageLite::Worker> is a helper class for worker code. This class helps
worker to get paramset from supervisor/Staf.

=head1 TODO

=head1 SEE ALSO
    
Options TestManagerLite Topology

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/TestManagerLite/Worker.pm#1 $
  Created $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=head1 METHODS

=cut

package TestManagerLite::Worker;
use strict;
use English;

# bring this in first to force the Delayed_help action
use Options qw(-Delayed_help);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Data::Dumper qw(Dumper);
use Tharn qw ($Log $Staf $Parent);

my $Options;
my $Options_hash;

$Options = new Options;

=head2 new

    my $worker = new TestManagerLite::Worker(params =>
                                               { DEST_IP => { type => SCALAR }, },
                                               { VOLPATH => { type => SCALAR }, },
                                            );

C<new> is the invocant for the TestManagerLite::Worker object.

=cut

sub new {
    my $invocant = shift;
    my $self;
    my (%params) = validate(@_, {params => {type => HASHREF},});
    my $class = ref($invocant) || $invocant;
    my $reqref = $params{params};
    $self->{req} = $reqref;    # Whatever test asked
    bless($self, $class);
    my (@params, @keys);

    $Log->enter(Dumper($reqref));

    $Options->parameters(
        definitions => [
            {   name         => "P_EXE_MODE",
                spec         => "s",
                default      => "worker",
                valid_values => ["worker"],
                description =>
                  "dictates the execution flow. This is for internal purpose only",
            },
            {name => "DATA_INTERFACES",
             spec => "s@",
             description =>
               "list of comma-separated data_interfaces, or range[specified as IP/range]",
             delimiter => ",",},
            {name        => "VOLPATH",
             spec        => "s",
             description => "Volpath for test to work on",},
            {name        => "DEST_PATHS",
             spec        => "s@",
             description => "list of comma-separated data_ip:volpath combos",
             delimiter   => ",",},
            {name        => "VOLPATHS",
             spec        => "s@",
             description => "Volpaths client should be mounted to",},
            {name        => "instance",
             spec        => "s",
             description => "The instance number of this work instance",},
            {name        => "RUNID",
             spec        => "s",
             description => "RUNID of this process set by NATE",},]);

    #
    # Put all options test asked through Options so we can pull the
    # values from environment as well in case it was passed as a param
    #
    foreach my $key (keys %$reqref) {
        my $spec = ($reqref->{$key}{type} == ARRAYREF) ? "s@" : "s";
        $Options->parameters(definitions => [
                                          {name    => $key,
                                           spec    => $spec,
                                           default => ($reqref->{$key}{default})
                                           ? $reqref->{$key}{default}
                                           : undef,
                                           description => "Name $key",},],);
    }

    #
    # We are ready to process all Options
    #
    $Options->process();
    $Options_hash = $Options->get_hash();

    #
    # Update $Options_hash by fetching values from $Staf (supervisor / parent)
    #
    $self->_fetch_options();
    $Log->debug("Options_hash: " . Dumper($Options_hash));

    #
    # The values available for test - Filter to only options requested by test
    #
    my $test_options = $self->_test_options();
    $Log->debug("Test Options: " . Dumper($test_options));

    #
    # Find possible permutation candidates. If requested is just SCALAR but
    # available value is an ARRAY, then assume that we have to permute on it
    #
    foreach my $key (keys %$reqref) {
        if (   ($reqref->{$key}{type} == SCALAR)
            && (ref($test_options->{$key}) eq "ARRAY")) {
            push @params, $test_options->{$key};
            push @keys,   $key;
        }
    }

    #
    # For 7M/CM, permute for DEST_PATHS. DEST_PATHS is kinda an internal variable that test may not be
    # expecting. Hence we have to check explictely for it and make sure they get converted into
    # test understable values (DEST_IP and VOLPATH)
    #
    # For 7M, permute for DATA_INTERFACES (Backward compatibility Compatibility) and VOLPATHS
    #
    if (   (!grep { $_ eq "DEST_PATHS" } @keys)
        && (ref($Options_hash->{DEST_PATHS}) eq "ARRAY")
        && (@{$Options_hash->{DEST_PATHS}})) {
        $Log->debug("DEST_PATHS added for permutations");
        push @keys,   "DEST_PATHS";
        push @params, $Options_hash->{DEST_PATHS};
    } else {
        if (   (!grep { $_ eq "DATA_INTERFACES" } @keys)
            && (ref($Options_hash->{DATA_INTERFACES}) eq "ARRAY")
            && (@{$Options_hash->{DATA_INTERFACES}})) {
            $Log->debug("DATA_INTERFACES added for permutations");
            push @keys,   "DATA_INTERFACES";
            push @params, $Options_hash->{DATA_INTERFACES};
        }
        if (   (!grep { $_ eq "VOLPATHS" } @keys)
            && (ref($Options_hash->{VOLPATHS}) eq "ARRAY")
            && (@{$Options_hash->{VOLPATHS}})) {
            $Log->debug("VOLPATHS added for permutations");
            push @keys,   "VOLPATHS";
            push @params, $Options_hash->{VOLPATHS};
        }
    }

    $Log->debug("Keys that are going to be permute: " . Dumper(\@keys));
    $Log->debug("The values : " . Dumper(\@params));

    #
    # Save required opts with all permutations
    #
    if (@params) {
        foreach (permute(@params)) {
            my %hash = %$test_options;
            @hash{@keys} = @$_;
            push @{$self->{params}}, {%hash};
        }
    } else {
        $Log->debug("There is no permutations qualified test_options found");
        push @{$self->{params}}, {%$test_options};
    }

    #
    # convert DEST_PATHS to VOLPATH and DEST_IP.
    # In case, DATA_INTERFACES has value, but DATA_IP doesn't, update DATA_IP
    #
    for (my $i = 0 ; $i < @{$self->{params}} ; $i++) {
        my $dest_path = $self->{params}->[$i]->{DEST_PATHS};
        if ($dest_path && !ref($dest_path)) {
            my ($dest_ip, $volpath) =
              split(":", $dest_path);
            $self->{params}->[$i]->{VOLPATH} = $volpath;
            $self->{params}->[$i]->{DEST_IP} = $dest_ip;
        }
        my $dataif = $self->{params}->[$i]->{DATA_INTERFACES};
        my $dstip  = $self->{params}->[$i]->{DEST_IP};
        if ($dataif
            && !$dstip) {
            $self->{params}->[$i]->{DEST_IP} = $dataif;
        }
    }

    $Log->debug("Possible Permutations: " . Dumper($self->{params}));

    $Log->exit("return $self = " . Dumper($self));
    return $self;
}

sub _fetch_options {
    my $self = shift;

    my $vlist = $Staf->var_list(handle => $Parent->handle());
    $Log->trace("VAR LIST: " . Dumper($vlist));

    if (grep { $_ eq "common" } keys %$vlist) {

        # Get all common test options from Staf
        my $topts = $Staf->var_get(var    => "common",
                                   handle => $Parent->handle());
        $Log->debug("VAR GET for 'common': " . Dumper($topts));

        # Gather into the same hashref
        foreach my $topt (keys %$topts) {
            $Options_hash->{$topt} = $topts->{$topt};
        }
    } else {
        $Log->debug("There is no common test_options set");
    }

    if (grep { $_ == $Options_hash->{instance} } keys %$vlist) {

        # Get all params specific to this instance
        my $topts = $Staf->var_get(var    => $Options_hash->{instance},
                                   handle => $Parent->handle());
        $Log->debug("VAR GET for $Options_hash->{instance}: " . Dumper($topts));

        # Gather into the same hashref
        foreach my $topt (keys %$topts) {
            $Options_hash->{$topt} = $topts->{$topt};
        }
    }
}

sub _test_options {
    my $self   = shift;
    my $reqref = $self->{req};

    #
    # Only keep the required params by the test, to keep namespace
    # atleast clean.
    # XXX: This part should be enhanced greatly. We no longer need to deal with environment
    # variables. We can do all through STAF in just a simple HASH form.
    #
    my $test_options;
    foreach my $key (keys %$Options_hash) {
        $test_options->{$key} = $Options_hash->{$key}
          if (defined($reqref->{$key}));
    }

    # Internal Variable to represent both DEST_IP and VOLPATH.
    $test_options->{DEST_PATHS} = $Options_hash->{DEST_PATHS};

    # Return The values available for test
    return $test_options;
}

=head2 init

    $test->init();

C<init> prepares a list of param set for C<nextset> to operate on.

=cut

sub init {
    my $self = shift;
    $Log->enter();

    $self->{permutations} = [];
    push @{$self->{permutations}}, @{$self->{params}};

    $Log->exit();
    return;
}

sub permute {
    my $last = pop @_;
    unless (@_) {
        return map [$_], @$last;
    }
    return map { my $left = $_; map [@$left, $_], @$last } permute(@_);
}

=head2 nextset

    my $href = $worker->nextset();

C<nextset> returns a hashref of a paramset required for a test run.

=cut

sub nextset {
    my $self = shift;
    $Log->enter();
    
    my $return = pop(@{$self->{permutations}});
    $Log->debug("nextset: " . Dumper($return));

    $Log->exit();
    return $return; 
}

=head2 id 

    my $id = $worker->id();

C<id> returns a scalar value of the id associated with this instance 

=cut

sub id {
    my $self = shift;
    $Log->enter();

    $Log->exit();
    return $Options_hash->{instance};
}

1;
