# $Id: //depot/prod/test/Rsierranevada/lib/TestUtils.pm#3 $
## @summary Perl module for test utilities
## @author Steven Ewing, Bill Snider, Brian O'Keefe
## @pod here

=head1 NAME

TestUtils.pm - Library of common test utilities.

=head1 DESCRIPTION

C<TestUtils> provides common routines used by the tests.

=head1 SYNOPSIS

    use TestUtils qw(expected_run_time iterate $Iteration $Duration);
    my $new_timeout = expected_run_time(timeout=>120);
    eval {
        iterate(subroutine=>\&test_main);
    };

    use TestUtils qw(filter_list);
    filter_list(original_list_ref=>[],
                filter_list_ref=>[]);

    use TestUtils qw(do_lockunlock);
    do_lockunlock($fh, $filename);

    sub test_main {
        print "Iteration=$Iteration\n";
    }

=head1 ROUTINES

=cut

my $LibraryLoaded;
if ($ENV{OVERRIDETESTUTILS}) {
    my $newModule = $ENV{OVERRIDETESTUTILS};
    delete $ENV{OVERRIDETESTUTILS};
    print "The main TestManager module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package TestUtils;

use strict;
use warnings;
use English qw(-no_match_vars);
use Options;
use TracePkg;
use Data::Dumper;
use List::Util qw(max min sum);
use Time::HiRes qw(gettimeofday tv_interval time);
use Net::SMTP;

my $Trace = new TracePkg();
my %Stats;

my $Options = new Options();
$Options->parameters(
    definitions => [
        {   name        => "DURATION",
            spec        => "s",
            description => "the length of time to iterate",
        },
        {   name => "SMART_DURATION",
            spec => "boolean",
            description =>
                "Stop iterations if average runtime from previous iterations will cause duration to be exceeded, even if there is time remaining",
            default => "false",
        },
        {   name        => "RUNTIME_HISTORY",
            spec        => "s",
            description => "Number of iterations to keep for runtime average",
            default     => "10",
        },
        {   name        => "ITERATE",
            spec        => "s",
            description => "an absolute number of iterations",
        },
        {   name        => "ITERATE_RESUME",
            spec        => "s",
            default     => 1,
            description => "if necessary, where to resume iterating",
        },
        {   name        => "ITERATION_TIMEOUT",
            spec        => "s",
            default     => undef(),
            description => "The time to wait for an individual iteration.",
        },
        {   name         => "ITERATION_TIMEOUT_INTERRUPT",
            spec         => "s",
            default      => undef(),
            dependencies => [qw(ITERATION_TIMEOUT)],
            description =>
                "Indicates whether or not to interrupt an over-time iteration.",
        },
        {   name    => "SLEEP",
            spec    => "s",
            default => 0,
            description =>
                "the number or seconds to sleep between iterations",
        },
        {   name         => "STOP_ON_ERROR",
            spec         => "s",
            valid_values => [qw(0 1 TRUE true FALSE false YES yes NO no)],
            default      => 1,
            description  => "stop iterating when an error is encountered",
        },
        {   name         => "FILE_READ_LOCK",
            valid_values => [qw(0 1)],
            default      => 1,
            description =>
                "obtain a read lock (then unlock) on a file before reading",
        },
        {   name        => "USE_O_DIRECT",
            spec        => "boolean",
            default     => undef,
            description => "use O_DIRECT on a file open before reading",
        },
        {   name        => "BUF_SIZE",
            spec        => "e",
            description => "The size of the read/write buffers",
            default     => 32768,
        },
        {   name => "SHOW_STATS",
            description =>
                "Causes manyread to display statitics every ten minutes.",
            spec    => 'boolean',
            default => "false",
        },
        {   name => "FILE_TIMEOUT",
            description =>
                "Individual read/write calls will cause fatal errors if they take longer than this.",
            spec    => 's',
            default => 0,
        },
        {   name => "SITE_FILE",
            description =>
                "Path to file that get_site_info() will load containing site information for servers (ftp,dns,cifs domain), image paths, etc",
            default => "TestUtils/site_info.dat",
        },
    ]
);

use Module::Common qw(validate_mixed);
use Params::Validate qw(CODEREF ARRAYREF SCALAR UNDEF OBJECT);
use Fcntl;
require Exporter;
our $Iteration;
our $Duration;
our @ISA    = qw(Exporter);
our @EXPORT = qw(
    iterate
    expected_run_time
    filter_list
    do_lockunlock
    read_file
    write_file
    size_in_unit
    get_site_info
    guess_site
    $Iteration
    $Duration
    $Iter_time
);
our @EXPORT_OK = qw(
    iterate
    expected_run_time
    filter_list
    do_lockunlock
    read_file
    write_file
    size_in_unit
    get_site_info
    guess_site
    $Iteration
    $Duration
    $Iter_time
);

#***********************************************************************

=head2 expected_run_time

C<expected_run_time> is usually called prior to calling C<iterate> to
calculate the expected run time of a routine. C<iterate> uses the params
C<ITERATE> and C<DURATION> to determine how many iterations to run a
given routine. This method will use the same logic to determine how
long the routine should run. If C<DURATION> has been specified, then that
is the time it should take. If C<ITERATE> has been specified then the
expected run time is the input C<timeout> * C<ITERATE>. Before returning 
the value, a fudge of 2 minutes is added. See C<ITERATE> for more info on
the params.

=over

=cut

sub expected_run_time {
    my %params = validate_mixed(@_, [], {timeout => {type => SCALAR}});
    $Trace->entry(msg => "timeout=>$params{timeout}");
    my $run_time = 120;                    # provide 2 minutes for overhead
    my $opt      = $Options->get_hash();
    if (defined($$opt{DURATION})) {
        $run_time += $params{timeout};    # allow for an iteration to complete
        $run_time += _normalize_duration($$opt{DURATION});
    } elsif ($$opt{ITERATE} =~ /infinity/i) {
        $run_time = 0;                    # cause nate to wait forever
    } elsif ($$opt{ITERATE} =~ /^\d+$/) {
        $run_time += $$opt{ITERATE} * $params{timeout};
    } elsif (defined($$opt{ITERATE})) {
        $Trace->fatal(msg => "ITERATE must be an integer or 'infinity'");
    } else {

        # default to ITERATE=1
        $run_time += $params{timeout};
    }
    $Trace->exit(msg => "return $run_time");
    return $run_time;
} ## end sub expected_run_time

sub _normalize_duration {
    my $duration = shift;
    my $duration_target;
    my $duration_unit;
    my $msg;

    $Trace->entry(msg => "duration=$duration");

    # Normalize the duration string into numerical seconds
    if ($duration =~ /^\d+[Hh]$/) {    # if value is hours
        $duration_unit   = chop($duration);
        $duration_target = $duration * 3600;
        $msg             = "Starting a ${duration}-hour test";
    } elsif ($duration =~ /^\d+[Dd]$/) {    # if value is days
        $duration_unit   = chop($duration);
        $duration_target = $duration * 86400;
        $msg             = "Starting a ${duration}-day test";
    } elsif ($duration =~ /^\d+[Mm]$/) {    # if value is minutes
        $duration_unit   = chop($duration);
        $duration_target = $duration * 60;
        $msg             = "Starting a ${duration}-minute test";
    } elsif ($duration =~ /^\d+[Ss]$/) {    # if value is seconds
        $duration_unit   = chop($duration);
        $duration_target = $duration;
        $msg             = "Starting a ${duration}-second test";
    } elsif ($duration =~ /^\d+$/) {
        $duration_unit = "m";    # default unit of duration is minutes
        $duration_target = $duration * 60;    # convert to seconds
        $msg = "Starting a ${duration}-minute test";
    } elsif (defined($duration) && $duration !~ /^\s*$/) {
        $Trace->fatal(msg =>
                "Could not parse the DURATION parameter. The value was '$duration'."
        );
    }

    $Trace->exit(msg => "return $duration_target, $duration_unit");
    return (wantarray)
        ? ($duration_target, $duration_unit, $msg)
        : $duration_target;
} ## end sub _normalize_duration

=head2 iterate

    eval {
        iterate(subroutine=>\&test_main, iteration_timeout => "30m");
    };

=back

C<iterate> is used to repeatedly call the given routine. The number of
iterations that a given routine is called is governed by the NATE parameters
C<ITERATE> and C<DURATION>. 

C<iteration_timeout> sets the iteration timeout.  An iteration which takes longer
than the specified time will be either logged as a failure or cause C<iterate> 
to terminate.  The NATE parameters C<ITERATION_TIMEOUT> and C<ITERATION_TIMEOUT_INTERRUPT> 
control this function as well.

A value for C<ITERATE> will cause the routine to be called that many times. 

A value for C<ITERATE_RESUME> will cause the routine to be called ITERATE
times, but beginning at ITERATE_RESUME. For example, if trying to create one
million files and an error occurs while creating file 600,002, you can rerun
the test with the ITERATE still set to 1000000, but with ITERATE_RESUME set
to 600002.

A value for C<DURATION> will cause the routine to be called repeatedly until
the amount of time has been specified. The total time will be at least the
amount of time specified, not shorter.

The value for C<DURATION> defaults to the number of minutes to run. The
duration can also be specified in seconds, minutes, hours or days. Examples:

=over

30s to indicate 30 seconds

90m to indicate 90 minutes

24h to indicate 24 hours

2d  to indicate 2 days

=back

If C<SMART_DURATION> is set, the average time of previous iterations
will be checked against the remaining time in the duration. Iterations will stop
if there is insufficient time remaining. C<RUNTIME_HISTORY> controls the number of
iteration runtimes are stored for calculating the runtime average.

A non-zero value assigned to C<ITERATION_TIMEOUT> will cause timeout checks to be
made for each iteration.  This parameter's value interacts with C<ITERATION_TIMEOUT_INTERRUPT>.

A true value for C<ITERATION_TIMEOUT_INTERRUPT> requires that C<ITERATION_TIMEOUT> be non-zero.
It will cause runaway iterations to be stopped once C<ITERATION_TIMEOUT> expires.  
This is implemented in terms of the ALRM signal.

During each iteration, the code block is run inside of an eval. It is 
assumed that the routine will die on a failure and returns nothing on
success. After each iteration, it checks for an error from the routine.
By default, iterate will die with the error if one is encountered. If
the NATE param C<STOP_ON_ERROR> is set to 0 or false, then C<iterate>
will store the error and continue running. At the end, it will return
a list of the failures. If C<STOP_ON_ERROR> is set to 1 or true, the
iterations will end immediately after the failure.

C<iterate> normally goes through each iteration with no pause in between
iterations. Specifying a value for C<SLEEP> will cause C<iterate> to
sleep for the given number of seconds in between iterations.

=over

=item subroutine

(Required) This is a reference to a subroutine to iterate around.

=back

=item iteration_timeout

(Optional) The amount of time a single iteration should require.

With specification of C<ITERATION_TIMEOUT> and C<ITERATION_TIMEOUT_INTERRUPT>, 
we can limit iterations and prevent tests from never returning.  This grid 
shows the differing behavior for C<ITERATION_TIMEOUT>, 
C<ITERATION_TIMEOUT_INTERRUPT>, and C<STOP_ON_ERROR>.

Parameters                         Effects on an iteration taking 
                                   more than X minutes.

ITERATION_TIMEOUT=0                 none
____________________________________________________________________

ITERATION_TIMEOUT=(X-1)m            iteration continues as long as
ITERATION_TIMEOUT_INTERRUPT=0       it wants, C<iterate> continues
                                    to next iteration
____________________________________________________________________

ITERATION_TIMEOUT=(X-1)m            iteration stopped after 10 min.
ITERATION_TIMEOUT_INTERRUPT=1       C<iterate> continues to next
                                    iteration
____________________________________________________________________

ITERATION_TIMEOUT=(X-1)m            iteration stopped after 10 min.
ITERATION_TIMEOUT_INTERRUPT=1       and C<iterate> terminates
STOP_ON_ERROR=1                     immediately
____________________________________________________________________

ITERATION_TIMEOUT=(X-1)m            iteration continues as long as
ITERATION_TIMEOUT_INTERRUPT=0       it wants, C<iterate> terminates
STOP_ON_ERROR=1                     once this iteration completes
____________________________________________________________________


Generally, the C<ITERATION_TIMEOUT> sets the error threshold for an iteration,
the C<ITERATION_TIMEOUT_INTERRUPT> sets whether or not to interrupt a runaway
iteration and C<STOP_ON_ERROR> indicates whether or not iterate terminates
when error (such as timeout expiriation) occurs.


=cut

sub iterate {
    my %params = validate_mixed(
        @_,
        [],
        {   subroutine        => {type => CODEREF},
            iteration_timeout => {type => SCALAR, default => undef()}
        }
    );

    my $testStartTime;
    my $testDuration;
    my $iterStartTime;
    my $iterDuration;
    my $durationUnit;
    my $durationTarget = 0;
    my $iterations     = "";
    my %error_hash;
    my $msg;
    my @iterate_times;
    my $iterate_timeout;
    my $avg_runtime;

    $Trace->entry();

    $Options->process();
    my $opt = $Options->get_hash();

# 1. If neither are specified, do one iteration.
# 2. If one or the other is specified, use it.
# 3. If both are specified and ITERATE=infinity, choose ITERATE and ignore.
# 4. If both are specified with other values, choose DURATION and ignore ITERATE.
    if (exists($$opt{ITERATE}) && $$opt{ITERATE} =~ m/infinity/i) {
        $iterations = $$opt{ITERATE};
        if ($iterations =~ m/infinity/i) {
            $Trace->comment(
                msg => "Starting an infinite test. Hit cntl-c to stop test");
        }
    } elsif (defined($$opt{DURATION})) {
        ($durationTarget, $durationUnit, $msg)
            = _normalize_duration($$opt{DURATION});
        $Trace->comment(msg => $msg);
        $iterations = undef;
    } elsif (exists($$opt{ITERATE})) {
        $iterations = $$opt{ITERATE};
        $Trace->comment(msg => "Starting a $iterations-iteration test");
    } else {
        $iterations = 1;
        $Trace->comment(msg => "Starting a $iterations-iteration test");
    }

    $testStartTime = time;
    $testDuration  = 0;

    if (defined($$opt{ITERATION_TIMEOUT})) {
        $iterate_timeout = _normalize_duration($$opt{ITERATION_TIMEOUT});
    } elsif (defined($params{iteration_timeout})) {
        $iterate_timeout = _normalize_duration($params{iteration_timeout});
    } else {
        $iterate_timeout = 0;
    }

    $Trace->comment(msg => "Iteration timeout is $iterate_timeout"
            . ($iterate_timeout ? "s" : "s (zero is unlimited)"));

#
# Iteration control:
# while
#    a.) We're iterating inifinitely; check the return code and see about stopping on errors.
#    b.) We're iterating for a finite period:
#        i.) Check for a remaining duration if we care about a duration.  Exit if there's none
#            left.
#       ii.) Check for remaining iterations if we care about iterations.  Exit if there's none
#            left.
#      iii.) Make sure that the most recent iteration was successful.  Bail out if it wasn't
#            and we're stopping on errors.
#

    $Iteration = $$opt{ITERATE_RESUME};
    $Duration  = $opt->{DURATION};
    while (
        (defined($iterations) && $iterations =~ m/infinity/i)
        || ((defined($$opt{DURATION}) ? ($testDuration < $durationTarget) : 1)
            && (defined($iterations) ? ($Iteration <= $iterations) : 1))
        ) {
        my $iter_time;
        sleep $$opt{SLEEP} unless ($Iteration == 1);
        if (defined($$opt{DURATION})) {
            if (defined($durationUnit) && $durationUnit =~ /h/i) {
                $iter_time = sprintf(
                    "(%.2f of " . $durationTarget / 3600 . " hours)",
                    $testDuration / 3600
                );
            } elsif (defined($durationUnit) && $durationUnit =~ /d/i) {
                $iter_time = sprintf(
                    "(%.2f of " . $durationTarget / 86400 . " days)",
                    $testDuration / 86400
                );
            } elsif (defined($durationUnit) && $durationUnit =~ /m/i) {
                $iter_time = sprintf(
                    "(%.2f of " . $durationTarget / 60 . " minutes)",
                    $testDuration / 60
                );
            } else {
                $iter_time = sprintf(
                    "(%.2f of " . $durationTarget . " seconds)",
                    $testDuration
                );
            }
            if ($$opt{SMART_DURATION} && int(@iterate_times) > 0) {

                if ($avg_runtime + $testDuration > $durationTarget) {
                    $Trace->comment(msg =>
                            "Stopping iterations. Another run would cause duration to be exceeded."
                    );
                    last;
                }
            }
            $Trace->comment(msg => "Begin iteration $Iteration $iter_time");
        } else {
            $Trace->comment(
                msg => "Begin iteration $Iteration (of $iterations)");
        }
        $iterStartTime = time;

        #***************************************************************
        # Invoke the subroutine that was passed in.
        #***************************************************************
        eval {

            #
            # Set an alarm to stop a runaway iteration.
            #
            local $SIG{ALRM};
            if (defined($$opt{ITERATION_TIMEOUT_INTERRUPT})
                && $$opt{ITERATION_TIMEOUT_INTERRUPT} =~ m/(true|yes|1)/i) {
                $SIG{ALRM} = sub {
                    $Trace->fatal(msg =>
                              "Iteration $Iteration is out of time (timeout: "
                            . $iterate_timeout
                            . "s) and was interrupted.");
                };
                alarm($iterate_timeout);
            }
            $params{subroutine}->();
            if (defined($$opt{ITERATION_TIMEOUT_INTERRUPT})
                && $$opt{ITERATION_TIMEOUT_INTERRUPT} =~ m/(true|yes|1)/i) {
                alarm(0);
            }
        };
        my $emsg = $@;
        $Trace->debug(msg => "Result returned: " . Dumper($emsg));

        $iterDuration = time - $iterStartTime;

# If the subroutine completes successfully, $@ is defined but set to "". As such, we
# need to check that specifically to see if there was a failure or not.
        if (defined($emsg)) {

            my $result_object
                ;    # this will hold a result object reference if we have one
            my $result_message
                ;    # this holds our message regardless of result type

            # Check the eval error that we saw
            if (!ref($emsg) && length($emsg)) {
                $result_message = $emsg;

     # Log the result reported with default FAIL, but control the abortiveness
                Tharn::logresult(
                    type  => "FAIL",
                    abort => 0,
                    msg   => $result_message
                );
            } elsif (UNIVERSAL::isa($emsg, 'NATE::Result')) {

              # Log the result reported as it is, but control the abortiveness
                $emsg->log();
                $emsg->store();
                if ($emsg->severity() >= NATE::Result::Fail->severity()) {
                    $result_object = $emsg;
                    $result_message
                        = $result_object->type()
                        . " Error.  Messages(s): "
                        . join(", ", $result_object->messages());
                } else {
                    undef($result_object);
                    undef($result_message);
                }
            }

            # If there was an abortive result, take the necessary action.
            if (defined($result_message) || defined($result_object)) {

        # Check for STOP_ON_ERROR and fatal with a result object if necessary.
                if ("$$opt{STOP_ON_ERROR}" =~ m/(false|no|0)/i) {
                    my $interrupted_message = "";
                    if ($emsg =~ m/Iteration \d+ is out of time/) {
                        $interrupted_message
                            = "  Iteration $Iteration was interrupted when the timeout expired.";
                    }
                    $error_hash{$Iteration} = $emsg;
                    $Trace->comment(
                        msg => "Failure, but continuing to next iteration."
                            . $interrupted_message);
                } else {
                    $Trace->comment(msg =>
                            "Failure, abandoning all subsequent iterations.\n$result_message"
                    );
                    $Trace->fatal(
                        msg    => "$result_message",
                        result => (
                            defined($result_object) ? $result_object : undef()
                        )
                    );
                }
            }
        }

        if (defined($durationUnit)) {
            if ($durationUnit =~ /h/i) {
                $iter_time = sprintf("%.2f hours)", $iterDuration / 3600);
            } elsif ($durationUnit =~ /d/i) {
                $iter_time = sprintf("%.2f days)", $iterDuration / 86400);
            } elsif ($durationUnit =~ /m/i) {
                $iter_time = sprintf("%.2f minutes)", $iterDuration / 60);
            } else {
                $iter_time = sprintf("%.2f seconds)", $iterDuration);
            }
        } else {
            $iter_time = sprintf("%.2f seconds)", $iterDuration);
        }
        $Trace->comment(msg => "End iteration $Iteration (took $iter_time");
        $testDuration = time - $testStartTime;

        #
        # Check the iteration timeout if one was provided.
        #
        if ($iterate_timeout && ($iterDuration > $iterate_timeout)) {
            if ("$$opt{STOP_ON_ERROR}" =~ m/(true|yes|1)/i) {
                $Trace->fatal(
                    msg => "Iteration $Iteration took too long; exiting.");
            } else {
                $Trace->warn(
                    msg => "Itertation $Iteration took too long "
                        . (
                        $$opt{ITERATION_TIMEOUT_INTERRUPT}
                        ? "and was interrupted"
                        : "but wasn't interrupted"
                        )
                        . ".  Continuing with next iteration."
                );
            }
        }

        $Iteration++;
        if (int(@iterate_times) >= $$opt{RUNTIME_HISTORY}) {
            shift(@iterate_times);
        }
        push(@iterate_times, int($iterDuration));
        $avg_runtime = _avg(@iterate_times);
        $Trace->comment(msg => "Average iteration runtime of last "
                . int(@iterate_times)
                . " run(s) : $avg_runtime seconds");

    }    # end while
    $testDuration = time - $testStartTime;
    my $test_time;
    if (defined($durationUnit)) {
        if ($durationUnit =~ /h/i) {
            $test_time = sprintf("%.2f hours", $testDuration / 3600);
        } elsif ($durationUnit =~ /d/i) {
            $test_time = sprintf("%.2f days", $testDuration / 86400);
        } elsif ($durationUnit =~ /m/i) {
            $test_time = sprintf("%.2f minutes", $testDuration / 60);
        } else {
            $test_time = sprintf("%.2f seconds", $testDuration);
        }
    } else {
        $test_time = sprintf("%.2f seconds", $testDuration);
    }
    $Trace->comment(msg => "Test took $test_time");
    my $msgs;
    foreach my $iteration (sort(keys %error_hash)) {
        $msgs .= "iteration $iteration failed: " . $error_hash{$iteration};
    }
    $Trace->comment(msg => "Error messages: \n" . $msgs) if ($msgs);
    $Trace->exit();
    return;
}    # end iterate

#***********************************************************************
# _avg() returns the rounded average of a list
#
#***********************************************************************

sub _avg {
    $Trace->entry();
    my @list    = @_;
    my $run_sum = 0;

    $Trace->debug(msg => "Items to average :" . int(@list) . " List: @list");
    map($run_sum += $_, @list);
    $Trace->exit();
    if (int(@list)) {
        return int($run_sum / int(@list));
    } else {
        return 0;
    }
} ## end sub _avg

#***********************************************************************
#
# filter_list takes two list references, uses the second list as a positive
# and negative filter against the first list, and returns a list (not a
# reference to a list) containing the filtered list.
#
# This can best be explained with an example. If original_list_ref is a
# reference to a list of (f0_aggr0, f0_aggr1, f1_aggr0, f1_aggr1),
# and filter_list_ref is a reference to a list of (f0, ~aggr1), this
# subroutine will keep from the original list all list items containing
# "f0" and throw away all list items containing "aggr1". Thus, the
# list that gets returned is (f0_aggr0).
#
# Note the use of ~ to be a "not". A ! can also be used, although it must
# be escaped with a \ when being passed in from a command line in a shell.
#
# An example of this would be, if a script has a parameter called AGGR,
# where the value(s) in AGGR are passed to this subroutine as the
# filter_list_ref, then the parameter would be specified as:
#    AGGR=f0,~aggr1 or AGGR=f0,\!aggr1
#
#***********************************************************************

sub filter_list {
    my %params = validate_mixed(
        @_,
        [],
        {   original_list_ref => {type => ARRAYREF},
            filter_list_ref   => {type => ARRAYREF},
        }
    );
    my $filter;
    my $tfilter;
    my @final_list;
    my $add;
    my $item;
    $Trace->entry();
    foreach $item (@{$params{original_list_ref}}) {
        $add = undef;
        $add = 1 if (!@{$params{filter_list_ref}});
        $Trace->debug(msg => "checking $item");
        foreach $filter (@{$params{filter_list_ref}}) {
            if ($filter =~ /^(!|~)/) {
                $tfilter = $filter;
                $tfilter =~ s/^(!|~)//;
                if ($item =~ /$tfilter/) {
                    $Trace->debug(msg => "matched negative, exclude");
                    $add = 0;
                    last;
                } else {
                    $Trace->debug(msg => "no match, include");
                    $add = 1;
                }
            } else {
                if ($item =~ /$filter/) {
                    $Trace->debug(msg => "matched positive, include");
                    $add = 1;
                }
            }
        }
        push(@final_list, $item) if defined($add) && $add;
    }
    $Trace->exit();
    return @final_list;
}    # end filter_list

=head2 do_lockunlock

    do_lockunlock($fh, $filename);

C<do_lockunlock> will obtain a read lock on a file and then release the lock.
This routine should be used prior to reading a file to force the client to
bypasss the local NFS client cache and perform the read on the filer.

=over

=item $fh

(Required) This is a filehandle to an already opened file.

=item $filename

(Required) This is the name of the file. It is used in the error case to
print out that the file could not be locked or unlocked.

=back

This routine is not currently checking the input parameters.

=cut

# put a read lock on $fh, then remove it immediately, for the purpose
# of convincing the client to forget any caching it may have done
sub do_lockunlock ( $$ ) {
    my ($fh, $filename) = @_;
    my $op;
    $Trace->entry(msg => $filename);

    my $opt = $Options->get_hash();

    if (!$opt->{FILE_READ_LOCK} || $OSNAME =~ /Win32/) {
        $Trace->exit(msg => "not obtaining locks");
        return;
    }

    eval {
        $Trace->debug(msg => "Acquiring read lock on $filename");
        $op = 'lock';
        my $struct_flock = pack("sx256", F_RDLCK);
        alarm(300);
        if (!fcntl($fh, F_SETLK, $struct_flock)) {
            $Trace->fatal(msg => "could not lock \"$filename\": $!\n");
        }
        alarm(0);

        $Trace->debug(msg => "Releasing read lock on $filename");
        $op = 'unlock';
        $struct_flock = pack("sx256", F_UNLCK);
        alarm(300);
        if (!fcntl($fh, F_SETLK, $struct_flock)) {
            $Trace->fatal(msg => "could not unlock \"$filename\": $!\n");
        }
        alarm(0);
    };

    my $emsg = $@;
    if ($emsg) {
        if ($emsg ne "alarm\n") {
            die($emsg);
        } else {
            $Trace->fatal(msg => ucfirst($op)
                    . " of $filename took more than 30 seconds to complete.");
        }
    }
    $Trace->exit();
} ## end sub do_lockunlock ( $$ )

=head2 read_file

    $bytes = read_file(file_name => "/path/to/file",
                       offset => 0,
                       length => 10000,
                       buf_size => 8192,
                       show_stats => 1,
                       timeout => 60
    );

C<read_file> provides a generic interface to a read routine. The routine has
the ability to lock the file to force the reads to bypass local client caches.
Timing characteristics can be printed for lengthy files.

=over

=item file_name

(Required) This is the name of the file to read.

=item offset

(Optional) The offset into the file to start reading. Defaults to 0.

=item length

(Optional) The amount of data to read. Defaults to the whole file.

=item buf_size

(Optional) How much data to read at a time. Defaults to the value of C<BUF_SIZE>.

=item show_stats

(Optional) Should the periodic stats be printed. Defaults to the value of C<SHOW_STATS>.

=item timeout

(Optional) How long to wait on an individual read before giving up. Default to the 
value of C<FILE_TIMEOUT>.

=back

The routine will return the total number of bytes read.

=cut

sub read_file {
    $Trace->entry();
    my $options = $Options->get_hash();
    my %params  = validate_mixed(
        @_,
        [],
        {   file_name => {type => SCALAR},
            offset => {type => SCALAR | UNDEF, optional => 1, default => 0},
            length   => {type => SCALAR | UNDEF, optional => 1},
            buf_size => {
                type     => SCALAR | UNDEF,
                optional => 1,
                default  => $options->{BUF_SIZE}
            },
            show_stats => {
                type     => SCALAR | UNDEF,
                optional => 1,
                default  => $options->{SHOW_STATS}
            },
            timeout => {
                type     => SCALAR | UNDEF,
                optional => 1,
                default  => $options->{FILE_TIMEOUT}
            },
        }
    );

    my $start_time            = time();
    my $stat_interval         = 300;
    my $next_stat_time        = $start_time + $stat_interval;
    my $first_5_minute_reads  = [];
    my $second_5_minute_reads = [];
    my $second_5_minute_bytes = 0;
    my $second_5_minute_count = 0;

    my $mode       = O_RDONLY;
    my $fname      = $params{file_name};
    my $offset     = $params{offset};
    my $length     = $params{length};
    my $bufSize    = $params{buf_size};
    my $show_stats = $params{show_stats};
    my $timeout    = $params{timeout};

    my @stat = stat $fname;
    my $fd;

    if ($options->{USE_O_DIRECT}) {
        $mode |= O_DIRECT;
    }

    if (!sysopen($fd, $fname, $mode)) {
        $Trace->fatal(
            msg    => "Could not open $fname: $! ($^E)\n",
            result => "FAIL"
        );
    }

    # this routine has a check on whether or not it should be run
    do_lockunlock($fd, $fname);

# Find the smallest time reported by Time::HiRes::tv_interval to avoid division by zero errors.
    my @tv_int_times;
    for (1 .. 1000) {
        push(@tv_int_times, tv_interval([gettimeofday()]));
    }

    my $min_time = min(@tv_int_times);

    if (!defined($length)) {
        $length = $stat[7];
    }
    my $bytesRemaining = $length;
    my $bytesRead      = 0;
    $Trace->trace(
        msg =>
            sprintf(
            "    reading %3.2f%s at offset $offset from $fname in $bufSize increments\n",
            size_in_unit($bytesRemaining))
    );

    if (!defined(sysseek($fd, $offset, 0))) {
        $Trace->fatal(
            msg    => "Error seeking to offset $offset in $fname: $! ($^E)\n",
            result => "FAIL"
        );
    }

    $Stats{reads} = [];

    while ($bytesRemaining > 0) {
        my $bytes = ($bytesRemaining < $bufSize) ? $bytesRemaining : $bufSize;
        my $buf;
        my $read;
        $Trace->debug(msg =>
                "      reading $bytes bytes, $bytesRemaining still to go\n");
        local $SIG{ALRM} = sub {
            $Trace->fatal(msg =>
                    "Read timeout.  The read didn't complete before ${timeout}s"
            );
        };
        alarm($timeout);
        my $begin_read_time = [gettimeofday()];
        if (!($read = sysread($fd, $buf, $bytes))) {
            $Trace->fatal(
                msg => "Error reading $bytes bytes from $fname: $! ($^E)\n",
                result => "FAIL"
            );
        }
        alarm(0);

        my $read_time = tv_interval($begin_read_time);
        $bytesRemaining -= $read;
        $bytesRead += $read;
        $Trace->debug(
            msg => "       $read bytes read, $bytesRead completed\n");

        if ($show_stats) {

            # Record the current read information.
            push(
                @{$Stats{reads}},
                {   duration => $read_time || $min_time,
                    size     => $read,
                    'time'   => $begin_read_time->[0],
                }
            );

            # Check to see if it's appropriate to print statistics.

            if (time() >= $next_stat_time) {

                # Place the information about reads in the correct lists.
                foreach my $read_info (@{$Stats{reads}}) {

                    # place the read throughput in a list
                    if ($read_info->{time}
                        >= ($next_stat_time - $stat_interval)) {
                        push(@{$first_5_minute_reads}, $read_info);
                    }
                }

                # calculate and display the data.
                my $stat = "Script-level Statistics summary\n";
                $stat .= _stats($first_5_minute_reads);
                if (scalar(@{$second_5_minute_reads})) {
                    $stat
                        .= _stats(
                        [@{$first_5_minute_reads}, @{$second_5_minute_reads}]
                        );
                }

                $Trace->comment(msg => $stat);

                $second_5_minute_reads = $first_5_minute_reads;
                $first_5_minute_reads  = [];
                $Stats{reads}          = [];
                $next_stat_time        = time() + $stat_interval;
            }
        }
    }

    $Trace->exit(msg => "return $bytesRead");
    return $bytesRead;
} ## end sub read_file

=head2 write_file

    $bytes = write_file(file_name => "/path/to/file",
                        offset => 0,
                        length => 10000,
                        buf_size => 8192,
                        show_stats => 1,
                        timeout => 60
    );

C<write_file> provides a generic interface to a write routine. 
Timing characteristics can be printed for lengthy files.

=over

=item file_name

(Required) This is the name of the file to write.

=item offset

(Optional) The offset into the file to write reading. Defaults to 0.

=item length

(Required) The amount of data to write. 

=item buf_size

(Optional) How much data to write at a time. Defaults to the value of C<BUF_SIZE>.

=item show_stats

(Optional) Should the periodic stats be printed. Defaults to the value of C<SHOW_STATS>.

=item timeout

(Optional) How long to wait on an individual write before giving up. Default to the 
value of C<FILE_TIMEOUT>.

=back

=cut

sub write_file {
    $Trace->entry();
    my $options = $Options->get_hash();
    my %params  = validate_mixed(
        @_,
        [],
        {   file_name => {type => SCALAR},
            offset    => {type => SCALAR | UNDEF, default => 0},
            length    => {type => SCALAR},
            buf_size =>
                {type => SCALAR | UNDEF, default => $options->{BUF_SIZE}},
            show_stats =>
                {type => SCALAR | UNDEF, default => $options->{SHOW_STATS}},
            timeout =>
                {type => SCALAR | UNDEF, default => $options->{FILE_TIMEOUT}},
        }
    );

    my $start_time             = time();
    my $stat_interval          = 300;
    my $next_stat_time         = $start_time + $stat_interval;
    my $first_5_minute_writes  = [];
    my $second_5_minute_writes = [];
    my $second_5_minute_bytes  = 0;
    my $second_5_minute_count  = 0;

    my $mode = O_WRONLY | O_CREAT;
    if ($OSNAME !~ /Win32/) {
        $mode |= O_DSYNC;
    }
    my $fname      = $params{file_name};
    my $offset     = $params{offset};
    my $length     = $params{length};
    my $bufSize    = $params{buf_size};
    my $show_stats = $params{show_stats};
    my $timeout    = $params{timeout};

    my $fd;
    if (!sysopen($fd, $fname, $mode)) {
        $Trace->fatal(
            msg    => "Could not open $fname: $! ($^E)\n",
            result => "FAIL"
        );
    }
    $Trace->trace(
        msg =>
            sprintf(
            "    writing %3.2f%s at offset $offset from $fname in $bufSize increments\n",
            size_in_unit($length))
    );
    if (!defined(sysseek($fd, $offset, 0))) {
        $Trace->fatal(
            msg    => "Error seeking to offset $offset in $fname: $! ($^E)\n",
            result => "FAIL"
        );
    }
    my $bytesRemaining = $length;

# Find the smallest time reported by Time::HiRes::tv_interval to avoid division by zero errors.
    my @tv_int_times;
    for (1 .. 1000) {
        push(@tv_int_times, tv_interval([gettimeofday()]));
    }

    my $min_time = min(@tv_int_times);

    $Stats{writes} = [];

    while ($bytesRemaining > 0) {
        my $bytes = ($bytesRemaining < $bufSize) ? $bytesRemaining : $bufSize;
        my $buf = '0' x $bytes;
        my $written;
        $Trace->debug(msg =>
                "      writing $bytes bytes, $bytesRemaining still to go\n");
        local $SIG{ALRM} = sub {
            $Trace->fatal(msg =>
                    "Write timeout.  The write didn't complete before ${timeout}s"
            );
        };
        alarm($timeout);
        my $begin_write_time = [gettimeofday()];
        if (!($written = syswrite($fd, $buf, $bytes))) {
            $Trace->fatal(
                msg    => "Error writing $bytes bytes to $fname: $! ($^E)\n",
                result => "FAIL"
            );
        }
        alarm(0);

        my $write_time = tv_interval($begin_write_time);
        $bytesRemaining -= $written;

        if ($show_stats) {

            # Record the current read information.
            push(
                @{$Stats{writes}},
                {   duration => $write_time || $min_time,
                    size     => $written,
                    'time'   => $begin_write_time->[0],
                }
            );

            # Check to see if it's appropriate to print statistics.

            if (time() >= $next_stat_time) {

                # Place the information about reads in the correct lists.
                foreach my $write_info (@{$Stats{writes}}) {

                    # place the write throughput in a list
                    if ($write_info->{time}
                        >= ($next_stat_time - $stat_interval)) {
                        push(@{$first_5_minute_writes}, $write_info);
                    }
                }

                # calculate and display the data.
                my $stat = "Script-level Statistics summary\n";
                $stat .= _stats($first_5_minute_writes);
                if (scalar(@{$second_5_minute_writes})) {
                    $stat .= _stats(
                        [   @{$first_5_minute_writes},
                            @{$second_5_minute_writes}
                        ]
                    );
                }

                $Trace->comment(msg => $stat);

                $second_5_minute_writes = $first_5_minute_writes;
                $first_5_minute_writes  = [];
                $Stats{writes}          = [];
                $next_stat_time         = time() + $stat_interval;
            }
        }
    }

    $Trace->exit();
} ## end sub write_file

=head2 size_in_unit

    ($size, $units) = size_in_unit($bytes);

C<size_in_unit> takes the number of bytes and returns a human readable form.
It returns the new $size and the unit of measure. For example, an input of 
1024 will return (1, "KB").

=over

=item bytes

(Required) The number of bytes.

=cut

sub size_in_unit {
    my ($size) = validate_mixed(@_, [{type => SCALAR}], {},);

    my $dividend = 1;
    my $suffix   = 'B';

    if ($size >= 1073741824) {
        $dividend = 1073741824;
        $suffix   = 'GB';
    } elsif ($size > 1048576) {
        $dividend = 1048576;
        $suffix   = 'MB';
    } elsif ($size > 1024) {
        $dividend = 1024;
        $suffix   = 'KB';
    }

    return (($size / $dividend), $suffix);
} ## end sub size_in_unit

=head2 get_site_info

returns a hash containing site specific information for configuration of filers and clients. Returns Image paths, server info (ftp/dns/nis/cifs domain), etc.  

=cut

sub get_site_info {
    $Trace->entry();
    my ($file, $return,);
    our %Siteinfo = ();

    # retrieve site file location
    $file = $Options->get(option => "SITE_FILE");
    my $code = NATE::Inc::get_file($file);
    if (!defined $code) {
        $Trace->fatal(
            result => "CONFIG",
            msg    => "Could not find SITE_FILE '$file'\n"
        );
    }
    eval $code;
    if ($@) {
        $Trace->fatal(
            result => "CONFIG",
            msg    => "Could not eval SITE_FILE '$file': $@\n"
        );
    }
    if (!keys %Siteinfo) {
        $Trace->fatal(
            result => "CONFIG",
            msg    => "SITE_FILE '$file' did not have a \%Siteinfo hash\n"
        );
    }

    $Trace->trace(msg => "Read site file: " . $file);

    # return site content as hash
    $Trace->debug(msg => "Site Info: " . Dumper(\%Siteinfo));
    return %Siteinfo;
} ## end sub get_site_info

=head2 guess_site
    
    $site = guess_site(host => $host);

returns the best guess it can for what $SITE would be based on the default ip of the
host passed in via the hostrec

=item host

(Required) A valid NATE hostrec for a client

=cut

sub guess_site {
    $Trace->entry();
    my ($host) = validate_mixed(@_, [{type => OBJECT},], {},);

    my $ip = $host->default_ip()
        || $Trace->fatal(
        msg => "Must pass in a valid hostrec to this method");
    my $site = "svl";

    # this was ripped right out of TestEnv/MinWin.pm
    if ($ip =~ /^10\.97\.*/) {    # NANE
        $site = "nane";
    } elsif ($ip =~ /^10\.60\.*/) {    # RTP
        $site = "rtp";
    } elsif ($ip =~ /^10\.98\.*/) {    # Pittsburgh
        $site = "ptc";
    } elsif ($ip =~ /^10\.72\.*/) {    # NB / BTC
        $site = "btc";
    } elsif ($ip =~ /^10\.229\.*/) {    # STI / CTL (GDL in RTP)
        $site = "sti";
    } elsif ($ip =~ /^10\.22[45678]\.*/)
    {    # RTP - GDL & Lab 1 (in rtpeng testbed)
        $site = "rtpeng";
    } elsif ($ip =~ /^172\.31\.*/) {    # smoke bubble
        $site = "smoke_bubble";
    }
    $Trace->trace(msg =>
            "Based on the default IP ($ip), the site appears to be '$site'");
    $Trace->exit();
    return $site;
} ## end sub guess_site

# this is an internal routine used to calculate the throughput
sub _stats {
    my $stat_string;

    my $io = shift();

    my $max_rate    = max(map({$_->{size} / $_->{'duration'}} @{$io}));
    my $min_rate    = min(map({$_->{size} / $_->{'duration'}} @{$io}));
    my $max_latency = max(map({$_->{'duration'}} @{$io}));
    my $min_latency = min(map({$_->{'duration'}} @{$io}));
    my $count       = scalar(@{$io});
    my $total       = sum(map({$_->{size}} @{$io}));
    my $total_time  = sum(map({$_->{'duration'}} @{$io}));
    if (!$total_time) {
        $Trace->warn(msg => "Encountered a 0 time");
        $Trace->comment(msg => Dumper($io));
        return;
    }
    my $avg  = $total / $total_time;
    my $time = int(($total_time + 59) / 60)
        ;    # make sure we get the next whole-number interval

    $stat_string = sprintf("%d-minute total bytes: %.2f%s", $time,
        size_in_unit($total));
    $stat_string
        = sprintf("  %d-min avg: %.2f %s\/s (%.2f%s over %ds in %d calls).\n",
        $time, size_in_unit($avg), size_in_unit($total), $total_time, $count);
    $stat_string .= sprintf(
        "  %d-min max rate: %-10s latency: %3.6fs\n",
        $time,
        (   defined($max_rate)
            ? sprintf("%.2f %s/s", size_in_unit($max_rate))
            : 'none'
        ),
        $max_latency,
    );
    $stat_string .= sprintf(
        "  %d-min min rate: %-10s latency: %3.6fs\n\n",
        $time,
        (   defined($min_rate)
            ? sprintf("%.2f %s/s", size_in_unit($min_rate))
            : 'none'
        ),
        $min_latency,
    );

    return (
        wantarray()
        ? ($avg, $min_rate, $max_rate, $total, $count, $total_time)
        : $stat_string
    );
} ## end sub _stats

=head2 email

fires of an email, includes engsup is asked to
=cut

sub email {
    $Trace->entry();
    my $default_from_addr;
    if ($0 =~ /\/(\w+\.thpl)/) {
        $default_from_addr = $1 . '@netapp.com';
    }
    my %params = validate_mixed(
        @_,
        [],
        {   to            => {type => ARRAYREF,},
            from          => {type => SCALAR, default => $default_from_addr},
            subject       => {type => SCALAR | UNDEF, optional => 1},
            body          => {type => SCALAR | UNDEF, optional => 1,},
            engsup_ticket => {type => SCALAR, optional => 1, default => 0},
            site       => {type => SCALAR | UNDEF, optional => 1,},
            host       => {type => OBJECT | UNDEF, optional => 1,},
            smtpserver => {type => SCALAR | UNDEF, optional => 1,},
        }
    );
    unless ($params{site}) {
        if ($params{host}) {
            $params{site} = guess_site($params{host});
        }
    }
    my %Siteinfo = get_site_info();

    #only look for the closest match if the $params{site}
    #does not exist in the site_info.dat file
    unless (exists($Siteinfo{$params{site}}->{smtp_server})) {
        if ($params{site} =~ /ptc/) {
            $params{site} = "ptc";
        } elsif ($params{site} =~ /nb/) {
            $params{site} = "nb";
        } elsif ($params{site} =~ /svl/) {
            $params{site} = "svl";
        } elsif ($params{site} =~ /rtp/) {
            $params{site} = "rtp";
        }
    }
    my $mailserver = "smtp.corp.netapp.com";
    if ($params{smtpserver}) {
        $mailserver = $params{smtpserver};
    } elsif ($Siteinfo{$params{site}}->{smtp_server}) {
        $mailserver = $Siteinfo{$params{site}}->{smtp_server};
    }

    $Trace->debug(msg => "DUMPER OF \%params:\n" . Dumper(\%params));

    my $smtp = Net::SMTP->new($mailserver, Port => 25)
        || die "Couldn't open SMTP connection to $mailserver";
    $smtp->mail("$params{from}");
    my $engsup_email = 'enghelp@netapp.com';
    if ($params{engsup_ticket}) {
        if ($params{site}) {
            $engsup_email = 'enghelp+' . $params{site} . '@netapp.com';
        } elsif ($params{host}) {
            my $site = guess_site($params{host});
            $engsup_email = 'enghelp+' . $site . '@netapp.com';
        }
        push(@{$params{to}}, $engsup_email);
    }
    foreach my $address (@{$params{to}}) {
        $smtp->to($address);
    }
    $smtp->data();
    $smtp->datasend("To: " . join(',', @{$params{to}}) . "\n");
    $smtp->datasend("Subject: $params{subject}\n");
    $smtp->datasend("\n\n");
    $smtp->datasend("$params{body}\n");
    $smtp->dataend();
    $smtp->quit;
} ## end sub email

=head2 temp_rm_nate_host

Moves a nate_host file to a temp_remove directory
=cut

sub temp_rm_nate_host {
    $Trace->entry();
    my %params = validate_mixed(
        @_,
        [],
        {   tharnhost_dir   => {type => SCALAR | UNDEF,   optional => 1,},
            temp_remove_dir => {type => SCALAR | UNDEF,   optional => 1},
            host_name       => {type => SCALAR,},
            host_type       => {type => SCALAR,},
            host            => {type => OBJECT | UNDEF,   optional => 1,},
            site            => {type => SCALAR | UNDEF,   optional => 1,},
            simhost         => {type => ARRAYREF | UNDEF, optional => 1,},
        }
    );
    $Trace->debug(msg => "DUMPER OF \%params:\n" . Dumper(\%params));
    my %Siteinfo = get_site_info();
    $Trace->debug(msg => "DUMPER OF \@simhost:\n" . Dumper($params{simhost}));

    #set the site
    unless ($params{site}) {
        if ($params{host}) {
            $params{site} = guess_site($params{host});
        }
    }

    #make sure we are operating in a testbed and set the tharnhosts dir
    unless ($params{tharnhost_dir}) {
        my @testbed_params     = qw(SITE_ROOT LAB TESTBED);
        my $working_in_testbed = 0;
        foreach my $testbed_param (@testbed_params) {
            if ((defined($ENV{$testbed_param})) && ($ENV{$testbed_param})) {
                ++$working_in_testbed;
            }
        }
        $Trace->comment(
            msg => "'$ENV{SITE_ROOT}':'$ENV{LAB}':'$ENV{TESTBED}'");
        if ($working_in_testbed == 3) {
            $params{tharnhost_dir}
                = $ENV{SITE_ROOT} . '/'
                . $ENV{LAB} . '/'
                . $ENV{TESTBED} . '/'
                . "tharnhosts";
        } else {
            $Trace->warn(msg =>
                    "It appears we are not running this test inside a testbed.\n"
                    . "Becasue of this we can not move the nate_host file to temp_remove!"
            );
            return 0;
        }
    }
    my $orig_path_to_nate_host;
    my $path_to_temp_remove;
    my $new_path_to_nate_host;
    my $vsims_moved = 0;
    $Trace->comment(msg => "SIMHOST='@{$params{simhost}}'");
    if (@{$params{simhost}}) {
        my $vsim_prefix = shift(@{$params{simhost}});
        foreach my $vsim_suffix (@{$params{simhost}}) {
            my $host_name = $vsim_prefix . $params{host_name} . $vsim_suffix;
            $Trace->comment(msg => "ATTEMPTING TO MOVE VSIM NATE_HOST FILE:"
                    . "'$host_name' TO TEMP_REMOVE!");

            #recursive call to this sub!
            my @empty_array;
            my $success = temp_rm_nate_host(
                host_name     => $host_name,
                host_type     => $params{host_type},
                site          => $params{site},
                host          => $params{host},
                tharnhost_dir => $params{tharnhost_dir},

                #must set to undef or will get stuck in loop
                simhost => \@empty_array,
            );
            if ($success) {
                $vsims_moved++;
                $Trace->comment(
                    msg => "SUCCESSFULLY MOVED VSIM NATE_HOST FILE:"
                        . "'$host_name' TO TEMP_REMOVE!");
            }
        }
        if (scalar(@{$params{simhost}}) == $vsims_moved) {
            $Trace->comment(
                msg => "SUCCESSFULLY MOVED *ALL* VSIM NATE_HOST FILES "
                    . "ON SIMHOST '$params{host_name}' TO TEMP_REMOVE!");
            return 1;
        } else {
            $Trace->warn(msg => "COULD NOT MOVE *ALL* VSIM NATE_HOST FILES "
                    . "ON SIMHOST '$params{host_name}' TO TEMP_REMOVE!");
            return 0;
        }
    } else {
        $orig_path_to_nate_host
            = $params{tharnhost_dir} . '/' . $params{host_name};
        $path_to_temp_remove = $params{tharnhost_dir} . '/temp_remove';
        if ($params{temp_remove}) {
            $path_to_temp_remove = $params{temp_remove};
        }
        $new_path_to_nate_host
            = $path_to_temp_remove . '/' . $params{host_name};
    }

    #connect to general use linux client
    my $gu_client = $Siteinfo{$params{site}}{general_use_linux_client};

    #print "GUI_CLIENT:'$gu_client'\n";
    my $host = Tharn::host($gu_client);

    #print "HOST:'$host'";
    my $gu_client_conn = Tharn::connect("#ssh");

    #check to make sure we can write to the nate_host file
    if (-w $orig_path_to_nate_host) {
        $Trace->comment(msg => "The file is writable!");
    } else {
        $Trace->comment(msg =>
                "The file is was not writable; trying to change mode bits.");
        Tharn::say($gu_client_conn, "ls -l $orig_path_to_nate_host",
            '-timeout', 60);
        Tharn::say($gu_client_conn, "sudo chmod 666 $orig_path_to_nate_host",
            '-timeout', 60);
        Tharn::say($gu_client_conn, "ls -l $orig_path_to_nate_host",
            '-timeout', 60);
        sleep 10;
        if (-w $orig_path_to_nate_host) {
            $Trace->comment(msg => "The file is now writable!");
        } else {
            $Trace->warn(msg =>
                    "Tried to correct file permission issues but could not.\n"
            );
        }
    }

    #check to make sure we have a temp_remove dir
    if (opendir(TEMP_REMOVE, $path_to_temp_remove)) {
        $Trace->comment(msg => "Found a temp_remove directory.");
        closedir(TEMP_REMOVE);
    } else {
        $Trace->comment(msg =>
                "Did *not* find a temp_remove directory.  Will attempt to create one."
        );
        if (mkdir($path_to_temp_remove, 0777)) {
            $Trace->comment(
                msg => "Created a temp_remove directory successfully.");
        } else {
            $Trace->warn(msg => "Unable to create a temp_remove directory.");
            return 0;
        }
    }

    #run some information commands on the nate_host file
    Tharn::say($gu_client_conn, "ls -l $orig_path_to_nate_host",
        '-timeout', 60);

    #let's move this sucker
    Tharn::say($gu_client_conn,
        "mv -fv $orig_path_to_nate_host $new_path_to_nate_host",
        '-timeout', 60);

    #did the move work
    my $match = Tharn::say(
        $gu_client_conn, "ls -l $new_path_to_nate_host",
        '-wait',         'No such file',
        '-wait',         "$new_path_to_nate_host \-\>",
        '-noprompt',     '-timeout',
        20
    );
    if ($match && ($match eq "No such file")) {
        $Trace->warn(msg =>
                "Unable to move hate_host file to the temp_remove directory."
        );
        return 0;
    }

    #bye bye
    Tharn::disconnect("#ssh", '-ok');
    return 1;
} ## end sub temp_rm_nate_host

=head2 find_nate_host_file

Finds the full path to a host's nate_host file
and returns it.
It also detects if the file is writable.
An attempt will be made to make the file writable.
=cut

sub find_nate_host_file {
    $Trace->entry();
    my %params = validate_mixed(
        @_,
        [],
        {   site     => {type => SCALAR | UNDEF, optional => 0,},
            hostname => {type => SCALAR | UNDEF, optional => 0,},
        }
    );
    my $full_path;
    my @stat;
    my $writable = 1;
    my %Siteinfo = get_site_info();
    $Trace->debug(msg => "NATE_HOSTS:'$ENV{NATE_HOSTS}");
    my @nate_host_paths = split(/;/, $ENV{NATE_HOSTS});

    foreach my $path (@nate_host_paths) {
        $Trace->trace(msg => "PATH:'$path'");
        if (-e "${path}\/$params{hostname}") {
            $full_path = "${path}\/$params{hostname}";
            $Trace->trace(msg => "FULL_PATH:'$full_path'");
            if (@stat = stat($full_path)) {
                $Trace->debug(msg => "MODE_BITS:'$stat[2]'");
                $Trace->debug(msg => "UID:GID:'$stat[4]':'$stat[5]'");
            }
            if (-w $full_path) {
                $Trace->comment(msg => "The file is writable!");
            } else {
                $Trace->comment(msg =>
                        "The file is was not writable; trying to change mode bits."
                );
                my $gu_client
                    = $Siteinfo{$params{site}}{general_use_linux_client};
                my $host           = Tharn::host($gu_client);
                my $gu_client_conn = Tharn::connect("#ssh");
                Tharn::say($gu_client_conn, "ls -l $full_path",
                    '-timeout', 60);
                Tharn::say($gu_client_conn, "sudo chmod 666 $full_path",
                    '-timeout', 60);
                Tharn::say($gu_client_conn, "ls -l $full_path",
                    '-timeout', 60);
                Tharn::disconnect("#ssh", '-ok');
                sleep 10;

                if (-w $full_path) {
                    $Trace->comment(msg => "The file is now writable!");
                } else {
                    $Trace->warn(msg =>
                            "Tried to correct file permission issues but could not.\n"
                            . "We will not be able to add REBOOTED_LAST time stamp to nate_host file."
                    );
                    $writable = 0;
                }
            }
            last;
        }
    }
    return ($full_path, $writable);
} ## end sub find_nate_host_file

=head1 PARAMS

=over

=item DURATION

This param is used by C<iterate> and specifies that the test routine will
run as many iterations as necessary to insure that the test runs at least
this long. The value can be specified as a number followed by either C<s>
for seconds, C<m> for minutes, C<h> for hours or C<d> for days. If no unit 
is specified, the value is in minutes. The C<STOP_ON_ERROR> param can cause 
the duration time to not be met. Please see the description of the C<iterate> 
method.

=item ITERATE

This param is used by C<iterate> and specifies that the test routine will
run this many iterations before returning. The C<STOP_ON_ERROR> param can
cause the number if iterations to not all be executed. Please see the
description of the C<iterate> method.

=item ITERATE_RESUME

This param is used by C<iterate> and specifies that the test routine will
resume running iterations beginning at this value, and continuing up to
the value of the C<ITERATE> param. This is typically only needed when the
test is creating a certain number of actual objects (volumes, files, etc.)
up to a certain limit, and when the test failed during the initial attempt.
For example, when you need 500 volumes on a filer, and the test created
400 and then failed. Assuming those 400 are still there, you simply want
to create 100 more, but probably want them to be named according to the
sequence that was already begun. Thus, the volumes will continue to be
named beginning at 401. Please see the description of the C<iterate>
method.

=item ITERATION_TIMEOUT

Specifies the amount of time an interation should take.  This parameter
is necessary, but isn't sufficient to stop a runaway iteration.  It must
be used with C<ITERATION_TIMEOUT_INTERRUPT> to achive that effect. Runaway
iterations are just another form of error and will stop C<iterate> based 
upon the C<STOP_ON_ERROR> setting.

=item ITERATION_TIMEOUT_INTERRUPT

Instructs C<iterate> to interrupt a runaway iteration after C<ITERATION_TIMEOUT>
has expired.

=item SLEEP

This param is used by C<iterate> and specifies that this much time should
elapse in between iterations of the test routine.

=item STOP_ON_ERROR

This param indicates that C<iterate> should immediately stop after an
iteration of a test routine if the routine encountered an error. The
default is set to 1. Turn this off by setting it to 0, false or no.

=item FILE_READ_LOCK

This param indicates that a read lock will be obtained and then
immediately released prior to reading a file. See C<do_lockunlock>.
The param defaults to 1 and should only be set to 0 if you are 
having problems on the filer which might be due to all of the locks
and need a way to turn them off for experimentation.

=item USE_O_DIRECT

This param indicates that read_file() will open the file with the
O_DIRECT flag prior to reading. On newer clients, this will cause
the client to bypass the local client NFS cache on the reads.

=item BUF_SIZE

This parameter specifies the size of each request for C<read_file> and
C<write_file>. Defaults to 32768.

=item SHOW_STATS

This parameter specifies that statistics should be printed out during
C<read_file> and C<write_file>.

=item FILE_TIMEOUT

This parameter specifies how long to wait on an individual read/write
requests in C<read_file> and C<write_file>.

=item SITE_FILE

Path to file that get_site_info() will load containing site information for
servers (ftp,dns,cifs domain), image paths, etc

=back

=head1 SEE ALSO


=head1 BUGS

C<iterate> could always handle NATE resutls better.

=head1 STATUS

  Shared


=head1 ATTRIBUTES


=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  Steve Ewing (ewing@netapp.com)
  Brian O'Keefe (okeefe@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/TestUtils.pm#3 $
  Last revised $Date: 2013/07/15 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
