#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     Common library to manage test Topologies
##
## @author      gvg@netapp.com
##
## @pod here

=head1 NAME

Topology - Test Topology Parent Class

=head1 SYNOPSIS

    package Topology::Topo1;
    use base qw(Topology); # Define super class

    sub filer_setup {
	# Define filer setup for the Topology Topo1
    }

    sub client_setup {
	# Define client setup for the Topology Topo1
    }

    sub client_cleanup {
	# Define client cleanup for the Topology Topo1
    }

    sub filer_cleanup {
	# Define filer cleanup for the topology Topo1
    }

    (AND)

    use qw(Topology);
    my $TOPO = new Topology(topology=>"topo1");

=head1 DESCRIPTION

This parent C<Topology> class creates and returns appropriate child
Topology object to the caller. This parent class also provides
a few functions for child classes and tests to use. 

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/Topology
http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite

=head1 TODO
* A lot. Still in design phase.

=head1 SEE ALSO

Options TestManagerLite

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Last revised $Date: 2008/05/16 $
    
=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved

=cut

package Topology;

use Tharn qw($Log config_result script_result host connect);
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Sort::Key::Natural qw(natsort);
use Params::Common;    # For DATA_INTERFACES and aggregate
use Params::Volume;    # For VOLUME
use Utils::HostManager;
use CIFS::Client qw(prep_for_windows_subtest check_connectivity);
use Server;
my $Server;

use Data::Dumper qw(Dumper);

use Options;
my $Options;
my $Options_hash;

my $Full_testname = $Tharn::Scriptname;
if (!$Full_testname || $Full_testname eq "") {
    if ( defined $INC{"NATE/Lint.pm"} ) {
        $Full_testname = "natelint";
        $Log->warn( "We seem to be running inside of natelint and will "
              . "continue execution with a fake testname. If you are seeing "
              . "this warning and are not executing natelint please alert the "
              . "owner of Topology.  Full_testname = '$Full_testname'"
        );
    } else {
        die script_result("Could not determine the TESTNAME", stacktrace => 0);
    }
}
my $Testname = $Full_testname;
$Testname =~ s/.+\///;      # just get the name of the test
$Testname =~ s/\.thpl//;    # strip off the .thpl extension

###################################################################

# Pick up the Topology names from the map file
our $Topos;

###################################################################

my $type;

$Options = new Options;
$Options->parameters(
    definitions => [
        {   name        => "TOPO",
            spec        => "s",
            description => "Topology name to be used",
        },
        {   name        => "SUPERVISOR2START",
            spec        => "s",
            description => "The personality of supervisor2start_next",
            default     => "basic",
        },
        {   name         => "REBOOT_CLIENTS",
            spec         => "s",
            default      => 0,
            valid_values => ["0", "1", "2"],
            description =>
                "Whether or not reboot all clients before starting the test. "
                . "This includes VERIFY_CLIENTS. "
                . "0 - Do not reboot no matter what. "
                . "1 - reboot if found unhealthy. "
                . "2 - reboot",
        },
        {   name    => "VERIFY_CLIENTS",
            spec    => "boolean",
            default => 1,
            description =>
                "Whether or not verify all clients before starting the test. "
                . "Verification includes checking for ping,rsh,staf. Unless "
                . "clients pass rate is 95%, test will not proceed. The "
                . "remaining 5% will be ignored hoping for the best",
        },
        {   name        => "NUM_VOLS",
            spec        => "s",
            description => "Number of the volume(s) to be created",
            default     => "1",
        },
    ]
);

sub new {
    my $invocant = shift;
    my %opts = (topology        => undef,
                clients         => [],
                test_options    => {},
                reserve_clients => 0,
                @_,);

    $Log->enter(Dumper(\%opts));
    $Options_hash = $Options->get_hash();
    # Create appropriate topology object
    my $topology = ($opts{topology}) ? $opts{topology} : $Options_hash->{TOPO};
    $Log->comment("Creating Topology for $topology");

    my $class = get_class(undef, $topology);
    $Log->comment("Using Topology library : $class");

    eval "require $class";
    die config_result("require $class failed: $@", stacktrace => 0) if ($@);
    my $self = $class->new();
    bless ($self, $class);
    foreach my $key (keys %opts) { $self->{$key} = $opts{$key} }
    $Options->process(run_help => 1);    # Bring in all Options

    if (!$self->{reserve_clients}) {
        goto RET;
    }

    # Save HostManager object here
    $self->{hostmanager} = new Utils::HostManager if (!$self->{hostmanager});
    my @oclients = $self->{hostmanager}->get_clients();

    my @clients;
    if ($Options_hash->{VERIFY_CLIENTS}) {
        use Utils::HostAdmin;

        $Log->comment("Number of clients given: " . scalar(@oclients));
        my $Ha = Utils::HostAdmin->new(clients => \@oclients);
        my $reboot = ($Options_hash->{REBOOT_CLIENTS} == 2) ? 1 : 0;
        my ($goods, $bads) =
          $Ha->clients_status(reboot  => $reboot,
                              timeout => ($reboot) ? 300 : 60,
                              retries => ($reboot) ? 3 : 1);
        push @clients, @$goods;
        $Log->comment("Number of healthy clients: " . scalar(@$goods));
        $Log->comment(  "Number of unhealthy clients: "
                      . scalar(@$bads)
                      . ". They are: "
                      . join(" ", @$bads));

        if (@$bads && ($Options_hash->{REBOOT_CLIENTS} == 1)) {
            $Log->comment(
                  "Number of unhealthy clients which are going to be rebooted: "
                    . scalar(@$bads));
            my ($bgoods, $bbads) =
              $Ha->clients_status(clients => $bads,
                                  reboot  => 1,
                                  timeout => 300,
                                  retries => 3);
            $Log->comment(  "Number of unhealthy clients even after reboot: "
                          . scalar(@$bbads)
                          . ". They are: "
                          . join(" ", @$bbads));
            push @clients, @$bgoods;
        } ## end if (@$bads && ($Options_hash...

        # Update Options hash with new set of clients.
        $Options->set(CLIENT => \@clients);

        my $ratio = scalar(@clients) / scalar(@oclients);
        if ($ratio == 1) {
            $Log->comment(
                      "GOOD NEWS: All given clients are healthy for the tests");
        } elsif ($ratio > 0.95) {
            $Log->warn(
                "Only the ratio $ratio of given clients found healthy. Proceeding to the test with the remaining, hoping for the best"
                );
        } else {
            die config_result(
                          "More than 5% of the given clients are not healthy.");
        }
    } else {
        $Log->comment(
                  "Using all the clients given without verifying their health");
        push @clients, @oclients;
    }
    @clients = $self->_hostname_sort(@clients);
    $self->{clients} = \@clients;
    my %clients = map { $_ => {"status" => "reserved"} } @clients;
    $self->{clients_status} = \%clients;

  RET:

    $Log->exit(Dumper($self));
    return $self;
} ## end sub new

sub get_class {
    my ($self, $topology) = @_;

    return $Options_hash->{P_USE_TOPO} if ($Options_hash->{P_USE_TOPO});

    $topology = $Options_hash->{TOPO} if (!$topology);
    
    my $Host = host(${$Options_hash->{'NODES'}}[0]);
    if ($Host->hosttype() !~ /ng/) {
        $type = "7M";
    } else {
        $type = "10M";
    }

    $Log->comment("Looking for class for $topology for type $type");

    # Pick up the Topology names from the map file
    my $topofile = NATE::Inc::find_file("Topology/topology.map");
    $Log->comment("Reading Topologies from $topofile");
    require "$topofile";

    my @valids = grep { $Topos->{$_}{$type} } keys %$Topos;

    if ((!$Topos->{$topology}) || (!$Topos->{$topology}{$type})) {
        die config_result(
            "Topology ($topology) not found for $type. Valid topology names are: @valids",
            stacktrace => 0);
    }

    return $Topos->{$topology}{$type};
} ## end sub get_class

########### A few common functions for sub classes and tests to use #####

=head1 METHODS_FOR_TESTS

=head2 define_topology

C<define_topology> is a core function. It defines the test topology that
is saved in hashref.

=cut

sub define_topology {
    my $self = shift;
    $Log->enter();

    $Log->exit();
}

sub update_test_topology {
    my $self = shift;
    $Log->enter();

    $Log->exit();
}

sub filer_setup {
    my $self = shift;
    $Log->enter();

    $Log->exit();
}

sub filer_cleanup {
    my $self = shift;
    $Log->enter();

    $Log->exit();
}

sub client_setup {
    my $self = shift;
    $Log->enter();

    $Log->exit();
}

sub client_cleanup {
    my $self = shift;
    $Log->enter();

    $Log->exit();
}

sub filers {
    my $self = shift;

    return @{$Options_hash->{FILER}};
}

sub is_10m {
    return 1 if ($type eq '10M');
    return 0;
}

sub is_7m {
    return 1 if ($type eq '7M');
    return 0;
}

=head2 get_connection

C<get_connection> returns NATE Connectrec of filer.

=over

=item filer

Optional. Filer Name.

=back

=cut

sub get_connection {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT}],
                     {filer => {type => UNDEF | SCALAR, optional => 1},});

    $Log->enter();
    $Server = new Server if (!$Server);
    my $conn = $Server->get_connection(-Server => $params{filer},
                                       type    => "cli_connection");
    $Log->exit(msg => "return $conn");
    return $conn;
} ## end sub get_connection

=head2 clients

C<clients> returns an array of clients used.

=cut

sub clients {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT}],
                     {status => {type => UNDEF | SCALAR, optional => 1},});
    my @clients = @{$self->{clients}};
    if ($params{status}) {
        return
          grep { $self->{clients_status}{$_}{status} eq $params{status} }
          @clients;
    } else {
        return @clients;
    }
} ## end sub clients

=head2 supervisor2start_next

C<supervisor2start_next> returns the client from the clients list that were reserved earlier. 
This function can be overloaded by any of the Topology sub classes to change the behavior of
how fast clients are consumed to subtest remotely. TestManagerLite.pm relies on this function
entirely to start each supervisor.

This solution was worked out to facilitate STAGING the client ramp up as per Topology needs.

=cut

sub supervisor2start_next {
    my $self = shift;
    $Log->enter();

    if ($Options_hash->{SUPERVISOR2START} eq 'basic') {

        # $self->{clients} is an ARRAYREF of list of clients in sorted form.
        foreach my $client (@{$self->{clients}}) {
            if ($self->{clients_status}{$client}{status} ne "inuse") {
                $self->{clients_status}{$client}{status} = "inuse";
                return $client;
            }
        }
    } else {

        my $client_index = 0;
        my $supervisor   = '';

        $Log->debug("Searching for clients");
        foreach my $client (@{$self->{'clients'}}) {

            if (!$self->{'clients_status'}{$client}{'status'}) {
                next;
            }

            $client_index++;

            if ($self->{'clients_status'}{$client}{'status'} eq 'reserved') {
                $self->{'clients_status'}{$client}{'status'} = 'inuse';
                $Log->comment("The next supervisor will be $client");
                $supervisor = $client;
                last;
            }
        }

        if (!$supervisor) { return; }

        if ($client_index > 1) {

            # Sleep for the specified amount of time between clients
            my $delay = $Options_hash->{'CLIENT_STARTUP_DELAY'};
            if ($delay) {
                $Log->comment(
                    "Delaying for CLIENT_STARTUP_DELAY ($delay) seconds");
                sleep $delay;
            }

            # TODO: This is where the LMR API call goes to get the appropriate
            #       stats so we can know if we need to start a new client
        }

        # Some CIFS clients use a login not known in the NETAPP domain so
        # so we need to map to the LOGDIR_WINDOWS share on the client so
        # the currently logged in user gets authenticated
        my $host = Hostrec->new(id => $supervisor);

        if ($host->hosttype() =~ /win/i
            && Tharn::param('LOGDIR_SHARE')) {

            $Log->comment("Preparing CIFS client for subtesting");

            # Connect to client
            my $conn = $host->connect(conntype => "staf-persay", id => $supervisor);

            # Verify connectivity to the client
            eval { check_connectivity('conn' => $conn); };

            # Perform any pre subtest tasks on the client
            eval { prep_for_windows_subtest('conn' => $conn); };
            if ($@) {
                die script_result("Error configuring the client "
                        . "for windows subtests: $@");
            }
        }
        $self->{clients_status}{$supervisor}{'status'} = "inuse";

    RET:

        my @total    = $self->clients();
        my @inuse    = $self->clients('status' => 'inuse');

        $Log->exit();
        return $supervisor;

    }

    $Log->exit();
} ## end sub supervisor2start_next

=head2 test_options

C<test_options> returns hashref of test options

=over

=item client

Optional. Client name the options corresponds to.

=back

=cut

sub test_options {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT,}],
                     {client => {type => SCALAR, optional => 1},});
   
    my $options = {}; 
    if ($params{client}) { 
        $options = $self->{test_options}{$params{client}};

    } elsif (my $test = $self->{test_options}) { 
        $options = $test;
    }

    return $options;
} ## end sub test_options

sub set_test_options {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT,}],
                     {options => {type => HASHREF | UNDEF},
                      client  => {type => SCALAR, optional => 1},});

    if ($params{client}) {
        $self->{test_options}{$params{client}} = $params{options};
    } else {
        $self->{test_options} = $params{options};
    }
} ## end sub set_test_options

sub clientsetup_options {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT,}],
                     {client => {type => SCALAR, optional => 1},});

    my $options = {};
    if ($params{client}) { 
        $options =  $self->{client_setup}{$params{client}};
    
    } elsif (my $test = $self->{client_setup}) { 
        $options = $test;
    }

    return $options;
} ## end sub clientsetup_options

sub set_clientsetup_options {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT,}],
                     {options => {type => HASHREF | UNDEF},
                      client  => {type => SCALAR, optional => 1},});

    if ($params{client}) {
        $self->{client_setup}{$params{client}} = $params{options};
    } else {
        $self->{client_setup} = $params{options};
    }
} ## end sub set_clientsetup_options

sub filersetup_options {
    my ($self) = @_;

    return $self->{filer_setup};
}

sub set_filersetup_options {
    my ($self, %params) = validate_mixed(
        @_,

        [{type => OBJECT,}],
        {options => {type => HASHREF | UNDEF},},);

    $self->{filer_setup} = $params{options};
}

#
# Returns total number of test instances defined by the topology
#

sub instances {
    my ($self) = shift;

    return $self->{_num_instances};
}

#
# Returns the volumes to be used for the topology
#
sub volumes {
    my ($self) = @_;
    $Log->enter();

    my @volumes  = @{$Options_hash->{VOLUME}};
    my $num_vols = $Options_hash->{NUM_VOLS};
    my $volstr   = $self->testcase_string();

    if (!@volumes) {
        $Log->debug("Volume was not given in cmdline");

        for (my $ind = 0 ; $ind < $num_vols ; $ind++) {
            push @volumes, sprintf("p_%03d_%s", $ind, $volstr);
        }
    }

    $Log->comment("Volumes: @volumes");

    $Log->exit();
    return @volumes;
} ## end sub volumes

sub testcase_string {
    my $self = shift;
    $Log->enter();

    $Log->exit();
    return ($Options_hash->{TESTCASE})
      ? ${Testname} . "_" . $Options_hash->{TESTCASE}
      : ${Testname};
}

sub size_in_mb {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT, isa => 'Topology'}],
                     {size => {type => SCALAR | UNDEF},});

    my $vol_size;
    if (!$params{size}) {
        $vol_size = 0;
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(MB|M|GB|G|TB|T)$/) {
        $vol_size = $1;
        if ($2 =~ m/GB?/i) {
            $vol_size *= 1000;
        } elsif ($2 =~ m/TB?/i) {
            $vol_size *= 1000;
            $vol_size *= 1000;
        }
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(mb|m|gb|g|tb|t)$/) {
        $vol_size = $1;
        if ($2 =~ m/gb?/i) {
            $vol_size *= 1024;
        } elsif ($2 =~ m/tb?/i) {
            $vol_size *= 1024;
            $vol_size *= 1024;
        }
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(KB|K)$/) {
        $vol_size = $1 / 1000;
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(kb|k)$/i) {
        $vol_size = $1 / 1024;
    } elsif ($params{size} =~ m/^(\d+\.?\d*)b?$/i) {
        $vol_size = $1 / 1024;
        $vol_size /= 1024;
    } else {
        die config_result(
            "Unable to convert size to megabytes. Could not match input string '$params{size}'",
            stacktrace => 0,);
    }

    if ($vol_size > int($vol_size)) {
        return (int $vol_size) + 1;
    } else {
        return int $vol_size;
    }
} ## end sub size_in_mb

sub size_in_kb {
    my ($self, %params) =
      validate_mixed(@_,
                     [{type => OBJECT, isa => 'Topology'}],
                     {size => {type => SCALAR | UNDEF},});

    my $vol_size;
    if (!$params{size}) {
        $vol_size = 0;
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(MB|M|GB|G|TB|T)$/) {
        $vol_size = $1 * 1000;
        if ($2 =~ m/GB?/i) {
            $vol_size *= 1000;
        } elsif ($2 =~ m/TB?/i) {
            $vol_size *= 1000;
            $vol_size *= 1000;
        }
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(mb|m|gb|g|tb|t)$/) {
        $vol_size = $1 * 1024;
        if ($2 =~ m/gb?/i) {
            $vol_size *= 1024;
        } elsif ($2 =~ m/tb?/i) {
            $vol_size *= 1024;
            $vol_size *= 1024;
        }
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(KB|K)$/) {
        $vol_size = $1;
    } elsif ($params{size} =~ m/^(\d+\.?\d*)(kb|k)$/i) {
        $vol_size = $1;
    } elsif ($params{size} =~ m/^(\d+\.?\d*)b?$/i) {
        $vol_size = $1 / 1024;
    } else {
        die config_result(
            "Unable to convert size to kilobytes. Could not match input string '$params{size}'",
            stacktrace => 0,);
    }

    if ($vol_size > int($vol_size)) {
        return (int $vol_size) + 1;
    } else {
        return int $vol_size;
    }
} ## end sub size_in_kb

###
# Host name sort
###
sub _hostname_sort {
    my ($self, @orighosts) = @_;

    return natsort @orighosts;
} ## end sub _hostname_sort

1;
