## @summary Discover the filer 
## @author dascenj@netapp.com
## @pod here

=head1 NAME

Topology::Discover - Discover the configuration of NetApp filers 

=head1 DESCRIPTION

Provide a library to utilize NACL in order to discover existing components on  
a NetApp filer.

=head1 SYNOPSIS

use Topology::Discover;
my $topo = Topology::Discover::run(node => $node);

=head1 EXAMPLES

=head2 discover all of the default components on the filer 

=over

=item *

C<< Topology::discover::run(node => $node); >>

=back

=head2 discover only the volumes on the filer

=over

=item *

C<< my $vols = Topology::Discover::volumes(node => $node); >>

=back

=head1 REQUIREMENTS

Based on NACL. $node must be a NACL::Node object;

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Jd (dascenj@netapp.com)
  Created

=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved

=head1 ROUTINES

This library uses an autoloader to translate the name of the function 
called into a NACL library and fetch the data from the contents. Therefore, 
you can actually call any function here that translates to an existing NACL::CS
library. There are some compatability options to help translate things like Volumes,
volumes, or volume, to the proper format before passing on to NACL. 
 
If the user calls Topology::Discover::foo(node => $node) then the library will attempt
to load NACL::CS::foo. We will issue a warning, if the requested library (NACL::CS::foo) is 
not found.

=cut


package Topology::Discover;

use strict;
use warnings;

use Tharn qw($Log config_result script_result fail_result);
use Params::Validate qw(validate UNDEF SCALAR OBJECT HASHREF ARRAYREF);
use Data::Dumper;

# Discover the layout of the filer
sub run { 
     my $topo = {};
     $Log->enter();
     my (%args) = validate(@_,
                     {node => {type => OBJECT, optional => 0},
                      targets => {type => ARRAYREF,
                              optional => 1,
                              # It is important to leave the word default as the last element
                              default => ['Volume', 'Aggregate', 'Vserver','NetworkPort', 'Interface', 'default'],},},);
    my $ci = $args{node};
    my $targets = $args{targets};
    my $targetsAreDefault = 0;

    # Check to see if the default values are being used for discovery targets. 
    # I use the "default" value as the last array element so that I can tell.
    if ($targets->[-1] eq "default") { 
        delete $targets->[-1];
        $targetsAreDefault = 1;
    }
   
    # If this is seven mode, we need to change the default discovery targets
    # these are the default discovery targets for 7Mode
    if (($ci->hostrec()->hosttype() !~ /-ng/) and ($targetsAreDefault)) { 
        my @defaults = ('Volume', 'Aggregate', 'Interface'); 
        $targets = \@defaults;
    }

    $Log->debug("Discovering targets " . Dumper($targets));

     # trigger the AUTOLOAD
     foreach my $item (@{$targets}) {  
        no strict 'refs';
        $topo->{$item} = &{$item}(node => $ci);
        $Log->debug("Found these (" . scalar(@{$topo->{$item}}) . ") $item items: "  .  Dumper($topo->{$item}));
        use strict;
    }
   
    $Log->exit();
    return $topo;
}

#
# What is my name? It's whatever you want it to be. 
#

sub AUTOLOAD { 
    our $AUTOLOAD;
    $Log->enter($AUTOLOAD);
     my (%args) = validate(@_,
                     {node => {type => OBJECT, optional => 0},},);

    my $ci = $args{node};
    my $topo;
    my @return;

    # strip out the package name
    my $p = $AUTOLOAD;
    my $currentPackage = __PACKAGE__;
    $p =~ s/${currentPackage}:://;

    # compatibility
    if ($p =~ /^volumes?$/i) { 
        $p = 'Volume';
    } elsif ($p =~ /^aggregates?$/i) { 
        $p = 'Aggregate';
    } elsif ($p =~ /^vservers?$/i) { 
        $p = 'Vserver';
    }elsif (($p =~ /^interfaces?$/i) or ($p =~ /^lifs?$/i)) { 
        $p = 'Interface';
    }elsif (($p =~ /^Ports?$/i) or ($p =~ /^networkports?$/i)) {
        $p = 'NetworkPort';
    } else { 
        $Log->warn("Attempting discovery of undocumented component type $p. " . 
                    "Please consider fixing the library to support this component " . 
                    "if it should truly be discoverable.");
    }
    $Log->trace("Discovering the available components of type $p...");

    # Only some of the components don't work for seven mode. Redirect them 
    # to the help function for 7Mode
    if (($ci->hostrec()->hosttype() !~ /-ng/) and ($p eq "Interface")) { 
        return &_discover_7Mode(node => $ci, target => $p);
    }

    # try to find a nacl library to support the requested component
    my $nacl_class = "NACL::CS::" . $p;
    $Log->debug("loading $nacl_class");

    # load 'er up
    eval "require $nacl_class";
    if ($@) { 
        $Log->warn("Failed to load library for $nacl_class. Discovery cannot be " . 
                     "performed for the requested component $p.");
    } else { 
        @return = $nacl_class->fetch(command_interface => $ci, allow_empty => 1) or
            fail_result ("failure to fetch $p from class $nacl_class");
    }

    # Save memory by knocking the command_interface off each of the components
    foreach my $obj (@return) { 
        $obj->command_interface_reset();
    }

    $Log->exit();
    return \@return;
}

# This private function is only used to handle the discovery targets that are not yet supported 
# by the NACL components.
sub _discover_7Mode { 
     my (%args) = validate(@_,
                     {node => {type => OBJECT,},
                      target => {type => SCALAR,},},);

     my $ci = $args{node};
     my $target = $args{target};
     my @return;

     # For a shortcut, I am embedding the api object into the node object, that is 
     # at least until the NACL team tells me how I shouldn't be doing that. :-) Oh, 
     # well, this way it will show up the second time through at least. 
     if (not $ci->{apiSet}->{cli}) { 
        require Options;
        my $options = Options->new();
        $ci->{apiSet}->{cli} = NACL::APISet->new(hostobj   => $ci->hostrec(),
                                                 category  => "Node",
                                                 interface => "CLI",
                                                 set       => "7Mode",
                                                 cdef_path => $options->get(option => 'CDEF_PATH'));
     }
     my $api = $ci->{apiSet}->{cli};

     if ($target eq "Interface") {  
          my $response = $api->ifconfig(all => 1) or 
              fail_result ("failure to execute ifconfig command");

          # The APISet embeds the return value funny. Do this to keep it uniform.
          return $response->get_parsed_output();
    }

}
  
1;
