## @summary navigate discovered data 
## @author dascenj@netapp.com
## @pod here

=head1 NAME

Topology::Lookup - lookup information returned by Topology::Discover

=head1 DESCRIPTION

Wade through the data returned by Topology::Discover, and return the 
requested targets. In the simplest form, this is just some easy "get" routines
for the data returned by the discovery library.

=head1 SYNOPSIS

use Topology::Lookup;
use Topology::Discover;

my $topo = Topology::Discover(node => $node);
my $vols = Topology::Lookup::volumes(topo => $topo);

=head1 EXAMPLES

=head2 lookup all of the volume targets on the filer

=over

=item *

C<< Topology::Lookup::volumes(topo => $topo); >>

=back

=head2 lookup all of the interface targets on the filer

=over

=item *

C<< Topology::Lookup::interface(topo => $topo); >>

=back

=head2 lookup all of the volume targets on the filer that reside on a given aggregate

=over

=item *

C<< Topology::Lookup::volume(topo => $topo, filter_key => 'aggregate', filter_value => 'testaggr_1'); >>

=back

=head1 REQUIREMENTS

The name of the routine that you call, should match 
the name of the hash key where the data is stored. Dumping the $topo
hashref is the best way to determine the proper name the key to 
request.

=head1 BUGS

If you find one, fix it. 

=head1 STATUS

  Shared

=head1 AUTHOR

  Jd (dascenj@netapp.com)
  Created

=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved

=head1 ROUTINES

This library uses an autoloader to translate the name of the function to the hash key 
from which the data should be retrieved.

For example:

If the user calls Topology::Lookup::volumes(topo => $topo) then the library 
will return all of the discovered volumes that exist in the given $topo (which was 
created by Topology::Discover).

=cut

package Topology::Lookup;

use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Tharn qw($Log config_result script_result fail_result);
use Data::Dumper;

use strict;
use warnings;

my $topo = {};

sub AUTOLOAD { 
    our $AUTOLOAD;
    my (%opts) =
        validate(@_,
                {topo => {type => HASHREF,},
                 filter_key => {type => SCALAR, optional => 1},
                 filter_value => {type => SCALAR, optional => 1},},);

    my $topo = $opts{topo};

    my $filterResults;
    my $fk = $opts{filter_key};
    my $fv = $opts{filter_value};
    if ((defined($fv)) and (defined($fk))) { 
        $Log->trace("Results will be filtered by key $fk value $fv");
        $filterResults = 1;
    }

    # strip out the package name
    my $p = $AUTOLOAD;
    my $currentPackage = __PACKAGE__;
    $p =~ s/${currentPackage}:://;

    # compatibility
    if ($p =~ /^volumes?$/i) {
        $p = 'Volume';
    } elsif ($p =~ /^aggregates?$/i) {
        $p = 'Aggregate';
    } elsif ($p =~ /^vservers?$/i) {
        $p = 'Vserver';
    }elsif (($p =~ /^ports?$/) or ($p =~ /^networkports?$/i)) {
        $p = 'NetworkPort';
    }elsif (($p =~ /^interfaces?$/) or ($p =~ /^lifs?$/i)) {
        $p = 'Interface';
    }

    $Log->debug("Lookup key $p");
    my $return = {};
    if (defined($topo->{$p})) { 
        $return  = $topo->{$p};

    } else { 
        $Log->debug("There were 0 instances of $p returned from the supplied topo");
        $filterResults = 0;
    } 

    if ($filterResults) { 
        $Log->debug("Filtering by key $fk for data $fv...");
        my @filtered;

        # Go through every object in the topo hash
        foreach my $data (@{$return}) { 
            # Check to see if the key/value we received as input
            # matches the key/value in the current hash
            if ((defined($data->{$fk})) and ($data->{$fk} =~ /^$fv$/i)) {
                push (@filtered, $data);
            }
        }
 
        $return = \@filtered;
    } 

    return $return;
}

1;
