#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     CIFS Stress Scalability Topology Library
##
## @author      bellizzi@netapp.com
##
## @pod here

=head1 NAME

Topology::NAS::CIFS - CIFS Stress Scalability test Topology Library

=head1 SYNOPSIS

    use qw(Topology);
    my $TOPO = new Topology(topology=>"topo1");

=head1 DESCRIPTION

This parent C<Topology::NAS::CIFS> class creates and returns appropriate child
Topology object to the caller. This parent class also provides
a few functions for child classes and tests to use. 

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/Topology
http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite

=head1 TODO

=head1 SEE ALSO

TestManager Options TestManagerLite

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  David Bellizzi (bellizzi@netapp.com)
    
=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved

=cut

package Topology::NAS::CIFS;

use strict;

use Tharn
    qw($Log connect host snooze fail_result config_result script_result say);
$Log->comment("Starting " . __FILE__);

use base qw(Topology);
use NATE::Inc qw(find_file);
use Options;
use Module::Common qw(validate_mixed);
use Params::Validate qw(
    SCALAR
    HASHREF
    ARRAYREF
    OBJECT
    UNDEF
    CODEREF);
use CIFS::Client qw(
    check_connectivity
    prep_for_windows_subtest
    ipv6_address_to_ipv6_literal);
use File::Basename;
use Data::Dumper;
use POSIX qw(strftime);
use Utils qw(make_conn);
use Utils::DNS qw(get_ip_by_name get_name_by_ip);
use Utils::IPManager;
use Utils::HostManager::Helper;
use Utils::Dump;
use YAML;
use Clone qw(clone);
use Sort::Key::Natural qw(natsort);
use Time::Local;
use Topology::NAS::CIFS::Parameters;

my $Options      = new Options;
my $Options_hash = $Options->get_hash();

sub new {
    my $invocant = shift;
    my $self     = {
        clients      => [],
        nodes        => [],
        topology     => undef,
        test_options => {},
        hostmanager  => {},
        @_,
    };

    $Log->enter();
    my $Options_hash = $Options->get_hash();

    $Log->debug("Executing new function in CIFS.pm");

    # Create appropriate topology object
    my $topology
        = ($self->{'topology'})
        ? $self->{'topology'}
        : $Options_hash->{'TOPO'};
    $Log->comment("Creating Topology for $topology");

    my $class = get_class(undef, $Options_hash->{'TOPO'});
    $Log->comment("Using Topology library : $class");

    eval "require $class";
    die config_result("require $class failed: $@", stacktrace => 0) if ($@);
    $self = $class->new();
    bless($self, $class);
    $self->{reserve_clients} = 0;
    $Options->process(run_help => 1);    # Bring in all Options

    $Log->exit();
    return $self;
} ## end sub new

sub get_delay_value {

    $Log->enter();

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   obj    => {type => OBJECT,   optional => 0},
            values => {type => ARRAYREF, optional => 0}
        }
    );

    my $delay;
    if (@{$opts{'values'}} > 1) {

        #-------------------------------------------------------
        # If multiple delay values are specified in the form of
        # config=seconds and one of those configs matches the
        # clients config then use that seconds value for delay
        #-------------------------------------------------------
        my %host_file_params = $opts{'obj'}->param_fields();
        my $matched          = 0;
        for (my $i = 1; $i < @{$opts{'values'}}; $i++) {
            my ($config, $seconds) = split(/=/, ${$opts{'values'}}[$i]);
            if ($host_file_params{'CONFIGS'} =~ /$config/) {
                $delay = $seconds;
            }
        }

        #-------------------------------------------------------------
        # If we did not match a config use the default, the 1st array
        # index
        #-------------------------------------------------------------
        if (!$matched) {
            $delay = ${$opts{'values'}}[0];
        }
    } else {
        $delay = ${$opts{'values'}}[0];
    }
    $Log->exit();

    return $delay;

} ## end sub get_delay_value

sub supervisor2start_next {
    my $self = shift;
    $Log->enter();

    $Log->debug("self: " . Dumper($self->{'clients'}));
    $Log->debug("self: " . Dumper($self->{'clients_status'}));
    my $Options_hash = $Options->get_hash();

    my $thresholds_met = 0;
    my $client_index   = 0;
    my $supervisor     = '';
    my $node_conn_obj;

    $Log->debug("Searching for clients");
    foreach my $client (@{$self->{'clients'}}) {

        if (!$self->{'clients_status'}{$client}{'status'}) {
            next;
        }

        $client_index++;

        $Log->debug("Client: $client "
                . "Index: $client_index "
                . "Max Clients: $self->{'max_clients'} "
                . "Status: $self->{'clients_status'}{$client}{'status'}");

        if ($self->{'clients_status'}{$client}{'status'} eq 'assigned') {
            $self->{'clients_status'}{$client}{'status'} = 'inuse';
            $Log->comment("The next supervisor will be $client");
            $supervisor = $client;
            last;
        }
    }

    if (!$supervisor) { return; }

    if ($client_index > 1) {

        #--------------------------------------------------------
        # Sleep for the specified amount of time between clients
        #--------------------------------------------------------
        if ($Options_hash->{'CLIENT_STARTUP_DELAY'}) {
            my $delay = $self->get_delay_value(
                'obj'    => host($supervisor),
                'values' => $Options_hash->{'CLIENT_STARTUP_DELAY'}
            );
            $Log->comment(
                "Delaying for CLIENT_STARTUP_DELAY ($delay) seconds");
            snooze $delay;
        }

    }

    #-----------------------------------------------------------------
    # Some CIFS clients use a login not known in the NETAPP domain so
    # so we need to map to the LOGDIR_WINDOWS share on the client so
    # the currently logged in user gets authenticated
    #-----------------------------------------------------------------
    if (host($supervisor)->hosttype() =~ /win/i
        && $Options_hash->{'LOGDIR_SHARE'}) {

        $Log->comment("Preparing CIFS client for subtesting");

        #-------------------
        # Connect to client
        #-------------------
        my $client_conn_obj
            = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
            '-nologin', -timeout, 300);

        #-----------------------------------
        # Verify connectivity to the client
        #-----------------------------------
        eval { check_connectivity('conn' => $client_conn_obj); };

        #---------------------------------------------
        # Perform any pre subtest tasks on the client
        #---------------------------------------------
        eval { prep_for_windows_subtest('conn' => $client_conn_obj); };
        if ($@) {
            die fail_result(
                "Error configuring the client " . "for windows subtests: $@");
        }
    }
    $self->{clients_status}{$supervisor}{'status'} = "inuse";

RET:

    my @total    = $self->clients();
    my @inuse    = $self->clients('status' => 'inuse');
    my @assigned = $self->clients('status' => 'assigned');

    $Log->debug("thresholds_met:$thresholds_met "
            . " \@assigned:"
            . @assigned
            . " \@inuse:"
            . @inuse
            . " \@total:"
            . @total);

  # Store and free the used and unused clients if the defined stats thresholds
  # are exceeded or there are no more assigned clients left.
    if ($thresholds_met || !@assigned) {
        my $clients = store_and_free_clients(
            'options' => $Options_hash,
            'script'  => "$self->{'testname'}",
            'total'   => \@total,
            'used'    => \@inuse
        );
    }

    return $supervisor;
    $Log->exit();
} ## end sub supervisor2start_next

sub wait_for_file {

    $Log->enter();

    $Options->process();
    my $Options_hash = $Options->get_hash();
    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   file  => {type => SCALAR, optional => 0},
            sleep => {
                type     => SCALAR,
                optional => 0,
                default  => 10
            }
        }
    );

    $Log->comment("Waiting for $opts{'file'}");
    while (!-e "$opts{'file'}") { snooze $opts{'sleep'}; }
    $Log->exit();

} ## end sub wait_for_file

sub write_term_string {

    $Log->enter();

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   term_file   => {type => SCALAR, optional => 0},
            term_string => {type => SCALAR, optional => 0},
        }
    );

    $Log->comment("Writing termination string $opts{'term_string'} "
            . "to termination file $opts{'term_file'}");
    my $status = 0;
    for (my $i = 1; $i <= 5; $i++) {
        if (open(TERM, ">>$opts{'term_file'}")) {
            print TERM "$opts{'term_string'}\n";
            close TERM;
            $status = 1;
            last;
        } else {
            sleep 10;
        }
    }

    if (!$status) {
        die
            "Error, Unable to write $opts{'term_string'} to $opts{'term_file'}";
    }

    $Log->exit();
} ## end sub write_term_string

sub convert_to_gb {

    $Log->enter();

    my ($self, %opts) = @_;
    my $converted_value = 0;

    if ($opts{'value'} =~ /(\d+)(^B$|^KB$|^MB$|^GB$|^TB$|^PB$)/) {
        my $value = $1;
        my $unit  = $2;
        if ($unit =~ /^B$|^KB$|^MB$/) {
            $converted_value = 0;
        } elsif ($unit eq 'GB') {
            $converted_value = $value;
        } elsif ($unit eq 'TB') {
            $converted_value = int($value * 1073741824);
        } elsif ($unit eq 'PB') {
            $converted_value = int($value * 1073741824000);
        }
    } elsif ($opts{'value'} =~ /^\d+$/) {
        $converted_value = int($opts{'value'} / 1073741824);
    }
    $Log->exit();

    return $converted_value;
} ## end sub convert_to_gb

sub get_array_index {
    $Log->enter();

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   array => {type => ARRAYREF, optional => 0},
            value => {type => SCALAR,   optional => 0}
        }
    );

    for (my $i = 0; $i < @{$opts{'array'}}; $i++) {
        if (${$opts{'array'}}[$i] eq $opts{'value'}) {
            $Log->exit();
            return $i;
        }
    }
    $Log->comment("Index not found for $opts{'value'}");
    $Log->exit();
} ## end sub get_array_index

sub parse_ifconfig {

    my $self   = shift;
    my $output = shift;
    my (%ifconfig, $interface);

    foreach my $line (@{$output}) {
        chomp($line);
        $line =~ s/\s+$//g;

# Linux
# eth0      Link encap:Ethernet  HWaddr 00:14:5E:6D:48:BA
# inet addr:172.16.17.46  Bcast:172.17.255.255  Mask:255.254.0.0
# inet6 addr: fe80::214:5eff:fe6d:48ba/64 Scope:Link
# UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
# RX packets:815794 errors:0 dropped:0 overruns:0 frame:0
# TX packets:344627 errors:0 dropped:0 overruns:0 carrier:0
# collisions:0 txqueuelen:1000
# RX bytes:540667306 (515.6 MiB)  TX bytes:26792813 (25.5 MiB)
# Interrupt:185 Memory:e2000000-e2011100
#
# bge1: flags=1004843<UP,BROADCAST,RUNNING,MULTICAST,DHCP,IPv4> mtu 1500 index 3
#       inet 10.60.88.172 netmask ffffff00 broadcast 10.60.88.255
#       ether 0:11:25:9e:5a:fd
# lo0:  flags=2002000849<UP,LOOPBACK,RUNNING,MULTICAST,IPv6,VIRTUAL> mtu 8252 index 1
#       inet6 ::1/128
# bge0: flags=2000841<UP,RUNNING,MULTICAST,IPv6> mtu 1500 index 2
#       inet6 fe80::211:25ff:fe9e:5afc/10
#       ether 0:11:25:9e:5a:fc

        if ($line =~ /^(\S+)\s+Link\s+encap:Ethernet\s+HWaddr\s+(.*)/) {
            $interface = $1;
            $ifconfig{$interface}{'mac'} = $2;
            if (!$ifconfig{$interface}{'ipv4'}) {
                @{$ifconfig{$interface}{'ipv4'}} = ();
            }
            if (!$ifconfig{$interface}{'ipv6'}) {
                @{$ifconfig{$interface}{'ipv6'}} = ();
            }
        } elsif ($line =~ /^(\S+)\s+flags/) {
            $interface = $1;
            if (!$ifconfig{$interface}{'ipv4'}) {
                @{$ifconfig{$interface}{'ipv4'}} = ();
            }
            if (!$ifconfig{$interface}{'ipv6'}) {
                @{$ifconfig{$interface}{'ipv6'}} = ();
            }
        }

        if ($interface) {
            if ($line =~ /.+inet\s+addr:(.+)\s+Bcast:(\S+)\s+Mask:(\S+)/) {
                push(@{$ifconfig{$interface}{'ipv4'}}, $1);
                $ifconfig{$interface}{'broadcast'} = $2;
                $ifconfig{$interface}{'mask'}      = $3;
            } elsif ($line =~ /.+inet6\s+addr:\s+(\S+)\/\d+\sScope:Link/) {
                $ifconfig{$interface}{'link_local_ipv6'} = $1;
            } elsif ($line =~ /.+inet6\s+addr:\s+(\S+)\/\d+\sScope:Global/) {
                push(@{$ifconfig{$interface}{'ipv6'}}, $1);
            } elsif ($line
                =~ /inet\s+(\S+)\s+(netmask|netmask-or-prefix)\s+(\S+)\s+broadcast\s+(\S+)/
                ) {
                push(@{$ifconfig{$interface}{'ipv4'}}, $1);
                $ifconfig{$interface}{'mask'}      = $3;
                $ifconfig{$interface}{'broadcast'} = $4;
            } elsif ($line =~ /.+inet6\s+(fe80.+)\//) {
                $ifconfig{$interface}{'link_local_ipv6'} = $1;
            } elsif ($line =~ /.+inet6\s+(\S+)\//) {
                push(@{$ifconfig{$interface}{'ipv6'}}, $1);
            } elsif ($line =~ /.+ether\s+(\S+)/) {
                $ifconfig{$interface}{'mac'} = $1;
            }
        }
    }
    return \%ifconfig;
} ## end sub parse_ifconfig

sub get_cifs_server_name {

    $Log->enter();

    my ($self, %options) = @_;

    if (!$options{'ip'}) {
        die config_result(
            "The parameter ip to get_cifs_server_name cannot be undef");
    }

    my $server = $options{'ip'};
    my $mode   = $options{'mode'};

    if ($mode =~ /name|fqdn/i) {

        # Perform the reverse lookup
        $Log->debug("Getting PTR record of $server");
        my $resolver = get_name_by_ip('ip' => $server);

        if ($resolver->{'ERROR'}) {

            # Die if ERROR
            die config_result(
                "IP address $server does not resolve to a name");
        } elsif ($mode =~ /name/
            && $resolver->{'PTR'} =~ /([^\.]+)\.(.+)/) {

            # Get the hostname from the PTR record
            $server = $1;
        } elsif ($mode =~ /fqdn/
            && $resolver->{'PTR'} =~ /([^\.]+\..+)/) {
            $server = $resolver->{'PTR'};
        }
    }

    $Log->exit();

    return $server;
} ## end sub get_cifs_server_name

sub get_io_configs {

    my ($self, %options) = @_;

    # If we specify a config file read that otherwise use global vars
    # for the load
    if ($Options_hash->{'IO_LOAD_CONFIG'}) {

        # Find  the YAML file
        $Log->comment("Finding $Options_hash->{'IO_LOAD_CONFIG'}");
        my $load_config_file = find_file($Options_hash->{'IO_LOAD_CONFIG'});

        # Open the YAML file
        $Log->comment("Reading $Options_hash->{'IO_LOAD_CONFIG'}");
        open my $fh, '<', $load_config_file
            or die "can't open config file $load_config_file: $!";

        # Convert the YAML file to perl hash format
        my $yml = do { local $/; <$fh> };

        # If TEST_MAIN_* is not used in the config then use TEST_MAIN*
        # from the NATE params
        my $hash = Load($yml);

        $self->{'defined_topology'}->{'io_configs'} = $hash;

        # Parameters we check for in the load config and NATE params
        my @parameters = (
            'MAX_CLIENTS',        'NODE_REGEX',
            'TEST_MAIN_DURATION', 'TEST_MAIN_SLEEP'
        );
        if (!Topology::is_7m()) { unshift(@parameters, 'AGGR_REGEX'); }

        # Get the number of loads
        my $loads = keys(%{$self->{'defined_topology'}->{'io_configs'}});

        # Count all paramters in all loads
        my $yaml_param_count = {};
        foreach my $p (@parameters) {
            $yaml_param_count->{$p} = 0;
            foreach my $load (keys(%{$hash})) {
                $yaml_param_count->{$p}
                    += grep { $_ eq $p } keys(%{$hash->{$load}});
            }
        }

        # Make sure parameters are defined correctly in the loads and or NATE
        foreach my $p (@parameters) {

            # Get all the values from the current param
            my @nate_param_values = split(/,/, $Options_hash->{$p});

            if ($loads != $yaml_param_count->{$p} && !@nate_param_values) {
                die script_result(
                          "\nThe param $p is not present in all loads and"
                        . "$p is not defined in the NATE params.\n"
                        . "Either define $p for all load configs\n"
                        . "or specify $loads values for $p in the\n"
                        . "form of a comma separated string of\n"
                        . "values. One for each load");
            } elsif ($loads != @nate_param_values) {

                die script_result(
                    "\nYou specified $loads loads but did not specify a "
                        . "$p\nvalue for each load using the $p parameter.\n"
                        . "Please make sure you specify $loads values for "
                        . "$p\nin the form of a comma separated string of "
                        . "values.\nOne for each load");
            } elsif (@nate_param_values) {

                # Override the values in the load with those specified in the
                # NATE params
                if ($p eq 'MAX_CLIENTS') { $p = 'max_clients'; }
                for (my $i = 0; $i < @nate_param_values; $i++) {

                    $self->{'defined_topology'}->{'io_configs'}->{$i}->{$p}
                        = $nate_param_values[$i];
                }
            }
        }

    } elsif ($Options_hash->{'IO_TOOLS_CONFIG'}) {
        $self->{'defined_topology'}->{'io_configs'}->{0} = {
            'IO_TOOLS_CONFIG'    => $Options_hash->{'IO_TOOLS_CONFIG'},
            'IO_TOOLS_REGEX'     => $Options_hash->{'IO_TOOLS_REGEX'},
            'IO_TOOLS_OVERRIDES' => $Options_hash->{'IO_TOOLS_OVERRIDES'},
            'NODE_REGEX'         => $Options_hash->{'NODE_REGEX'},
            'IO_MAP_SHARES'      => $Options_hash->{'IO_MAP_SHARES'},
            'IO_REMOVE_DATA'     => $Options_hash->{'IO_REMOVE_DATA'},
            'TEST_MAIN_DURATION' => $Options_hash->{'TEST_MAIN_DURATION'},
            'TEST_MAIN_SLEEP'    => $Options_hash->{'TEST_MAIN_SLEEP'},
            'CLIENT_HOSTTYPE_REGEX' =>
                $Options_hash->{'CLIENT_HOSTTYPE_REGEX'}
        };
        if ($Options_hash->{'MAX_CLIENTS'}) {
            my @max_clients = split(/,/, $Options_hash->{'MAX_CLIENTS'});
            $self->{'defined_topology'}->{'io_configs'}->{0}->{'max_clients'}
                = $max_clients[0];
        } elsif ($self->{'max_clients'}) {
            $self->{'defined_topology'}->{'io_configs'}->{0}->{'max_clients'}
                = $self->{'max_clients'};
        } else {
            $self->{'defined_topology'}->{'io_configs'}->{0}->{'max_clients'}
                = @{$self->{'clients'}};
        }
    }
} ## end sub get_io_configs

sub parse_io_configs {

    my ($self, %options) = @_;
    our $IO_TOOLS;

    my $io_configs = $self->{'defined_topology'}->{'io_configs'};
    for (
        my $io_config_index = 0;
        $io_config_index < keys(%{$io_configs});
        $io_config_index++
        ) {
        my $io_tools_filter;

        $Log->comment(
            "Finding $io_configs->{$io_config_index}->{'IO_TOOLS_CONFIG'}");
        my $io_tools_config_file
            = find_file($io_configs->{$io_config_index}->{'IO_TOOLS_CONFIG'});

        if (!$io_tools_config_file) {
            die script_result(
                "Unable to find $io_configs->{$io_config_index}->{'IO_TOOLS_CONFIG'}"
            );
        }

        $Log->comment(
            "Sourcing $io_configs->{$io_config_index}->{'IO_TOOLS_CONFIG'}");
        do "$io_tools_config_file";
        $io_configs->{$io_config_index}->{'IO_TOOLS'} = clone($IO_TOOLS);

        if (!keys(%{$IO_TOOLS})) {
            die
                'No tools defined! Check your IO_TOOLS_CONFIG and IO_TOOLS_REGEX parameters';
        } else {
            my $tool_list;
            $io_tools_filter
                = $io_configs->{$io_config_index}->{'IO_TOOLS_REGEX'};
            foreach my $tool (natsort(keys(%{$IO_TOOLS}))) {

                if ($io_tools_filter) {
                    $Log->debug("Checking $_ !~ /" . $io_tools_filter . "/");
                }
                if ($io_tools_filter && $tool !~ /$io_tools_filter/) {
                    next;
                }
                $Log->debug("Saving $tool");
                $tool_list .= "$tool\n";
            }
            $Log->comment("The following tools were imported:\n"
                    . "Source: $io_tools_config_file\n "
                    . "Tools : \n "
                    . $tool_list);

            my $io_tools_config
                = fileparse($io_tools_config_file, qr/\.[^.]*/);

            $io_tools_config
                .= "_$Options_hash->{'SCRIPT_INDEX'}" . "_$io_config_index";
            if ($io_tools_config_file =~ /(\.[^.]*)$/) {
                $io_tools_config .= $1;
            }

            $Log->comment("Saving tools hash to TOP_LOGDIR");
            write_dump(
                'file'    => "$Options_hash->{'TOP_LOGDIR'}/$io_tools_config",
                'creator' => $options{'script'},
                'var_name' => 'IO_TOOLS',
                'var_ref'  => $IO_TOOLS
            );

            $io_configs->{$io_config_index}->{'io_tools_config'}
                = $io_tools_config;

            my $max_shares_per_occurrence  = 0;
            my $max_occurrences_per_client = 0;
            foreach my $tool (keys(%{$IO_TOOLS})) {
                my $shares_per_client = 0;
                if (   $io_tools_filter
                    && $tool !~ /$io_tools_filter/) {
                    next;
                }

                my @occurrences = grep(/^\d+$/, keys(%{$IO_TOOLS->{$tool}}));
                if ($max_occurrences_per_client < @occurrences) {
                    $max_occurrences_per_client = @occurrences;
                }

                for (
                    my $i = 0;
                    $i < grep(/^\d+$/, keys(%{$IO_TOOLS->{$tool}}));
                    $i++
                    ) {

                    if (!$IO_TOOLS->{$tool}->{$i}->{'shares'}) {
                        die config_result(
                            "At least one share needs to "
                                . "be defined for $tool instance $i",
                            stacktrace => 0
                        );
                    }

                    $shares_per_client = $IO_TOOLS->{$tool}->{$i}->{'shares'};

                    if ($max_shares_per_occurrence < $shares_per_client) {
                        $max_shares_per_occurrence = $shares_per_client;
                    }
                }
            }
            $io_configs->{$io_config_index}->{'max_shares_per_occurrence'}
                = $max_shares_per_occurrence;
            $io_configs->{$io_config_index}->{'max_occurrences_per_client'}
                = $max_occurrences_per_client;
            $io_configs->{$io_config_index}->{'max_shares_per_client'}
                = $max_shares_per_occurrence * $max_occurrences_per_client;
        }
    }
} ## end sub parse_io_configs

sub update_test_options {

    my ($self, %options) = @_;
    my $client   = $options{'client'};
    my $instance = $options{'instance'};
    my $node     = $options{'node'};
    my $vfiler   = $options{'vfiler'};
    my $vserver  = $options{'vserver'};
    my $volume   = $options{'volume'};
    my $server   = $options{'server'};
    my $share    = $options{'share'};
    my $destip   = $options{'destip'};

    # Create the TO_NODES for the current client
    if ($node) {
        if (!$self->{'test_options'}->{$client}->{$instance}->{'TO_NODES'}) {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_NODES'}
                .= $node;
        } else {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_NODES'}
                .= ",$node";
        }
    }

    # Create the TO_VFILERS for the current client
    if ($vfiler) {
        if (!$self->{'test_options'}->{$client}->{$instance}->{'TO_VFILERS'})
        {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_VFILERS'}
                .= $vfiler;
        } else {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_VFILERS'}
                .= ",$vfiler";
        }
    }

    # Create the TO_VSERVERS for the current client
    if ($vserver) {
        if (!$self->{'test_options'}->{$client}->{$instance}->{'TO_VSERVERS'})
        {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_VSERVERS'}
                .= $vserver;
        } else {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_VSERVERS'}
                .= ",$vserver";
        }
    }

    # Create the TO_SERVERS for the current client
    if ($server) {
        if (!$self->{'test_options'}->{$client}->{$instance}->{'TO_SERVERS'})
        {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_SERVERS'}
                .= $server;
        } else {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_SERVERS'}
                .= ",$server";
        }
    }

    # Create the TO_VOLUMES entry for the current client
    if ($volume) {
        if (!$self->{'test_options'}->{$client}->{$instance}->{'TO_VOLUMES'})
        {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_VOLUMES'}
                .= $volume;
        } else {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_VOLUMES'}
                .= ",$volume";
        }
    }

    # Create the TO_SHARES entry for the current client
    if ($share) {
        if (!$self->{'test_options'}->{$client}->{$instance}->{'TO_SHARES'}) {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_SHARES'}
                .= $share;
        } else {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_SHARES'}
                .= ",$share";
        }
    }

    # Create the TO_DESTIPS entry for the current client
    if ($destip) {
        if (!$self->{'test_options'}->{$client}->{$instance}->{'TO_DESTIPS'})
        {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_DESTIPS'}
                .= $destip;
        } else {
            $self->{'test_options'}->{$client}->{$instance}->{'TO_DESTIPS'}
                .= ",$destip";
        }
    }
} ## end sub update_test_options

sub get_epoch_time {
    $Log->enter();
    my ($self, %options) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {'date' => {type => SCALAR, optional => 0},}
    );

    my $time;
    if ($options{'date'} =~ /(\d+)\/(\d+)\/(\d+)\s+(\d+):(\d+):(\d+)/) {
        my $month = $1;
        $month--;
        $time = timelocal($6, $5, $4, $2, $month, $3);
    }
    $Log->exit();
    return $time;
} ## end sub get_epoch_time

sub get_ra_config {

    my ($self, %options) = @_;

    # Find  the YAML file
    my $ra_config_file = find_file($Options_hash->{'RA_CONFIG'});

    # Open the YAML file
    open my $fh, '<', $ra_config_file
        or die "can't open config file $ra_config_file: $!";

    # Convert the YAML file to perl hash format
    my $yml = do { local $/; <$fh> };
    $self->{'defined_topology'}->{'ra_config'} = Load($yml);

} ## end sub get_ra_config

sub random_array_value {
    my ($self, %options) = @_;
    return ${$options{'array'}}[rand(@{$options{'array'}})];
}

# Let the topology decide when this is a good idea.
sub start_ip_manager {
    $Log->enter();
    my ($self) = @_;

    my ($ip_mgr, %args);

    # The potential to start the IPManager twice exists.
    # Don't do that.
    if ($self->{'ip_manager'}) {
        $ip_mgr = $self->{'ip_manager'};

    } else {
        my $mode = $Options->get(option => 'P_EXE_MODE');
        $args{'start_messenger'} = 0;

        $Log->comment(
            "Creating a new IPManager. The TestManagerLite mode is $mode.");

        # The IPManager should only be communicating with
        # the IPManager server when it is in manager mode.
        # supervisors get their information via Staph.
        if ($mode =~ /mgr/i) {
            $args{'start_messenger'} = 1;
        }

        # Start the client IP manager software
        $ip_mgr = Utils::IPManager::start_ip_manager(%args);
    }

    $Log->exit();
    return $ip_mgr;
} ## end sub start_ip_manager

sub random_string {

    $Log->enter();
    my ($self, %options) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {'chars' => {type => SCALAR, optional => 0},}
    );
    srand(time() ^ ($$ + ($$ << 15)));
    my @alphabet = (('A' .. 'Z'), 0 .. 9);
    my $string;
    for (my $i = 1; $i <= $options{'chars'}; $i++) {
        my $char = $alphabet[rand(@alphabet)];
        while ($i == 1 && $char =~ /\d/) {
            $char = $alphabet[rand(@alphabet)];
        }
        $string .= $char;
    }
    $Log->exit();
    return $string;
} ## end sub random_string

sub assign_test_clients {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Assigning test options for clients");

    $Options_hash = $Options->get_hash();
    my $instance     = 0;
    my $client_count = 0;

    for my $client (@{$self->{'clients'}}) {

        if (host($client)->hosttype()
            !~ /$Options_hash->{'CLIENT_HOSTTYPE_REGEX'}/) {
            next;
        }
        $client_count++;
        $Log->comment("Client: $client");
        $Log->comment("Instance: $instance");

        $self->{'clients_status'}->{$client}->{'status'} = 'assigned';
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_CONNECT_METHOD'}
            = $Options_hash->{'CLIENT_CONNECT_METHOD'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_HOSTTYPE_REGEX'}
            = $Options_hash->{'CLIENT_HOSTTYPE_REGEX'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'TEST_MAIN_DURATION'} = $Options_hash->{'TEST_MAIN_DURATION'};
        $self->{'test_options'}->{$client}->{$instance}->{'TEST_MAIN_SLEEP'}
            = $Options_hash->{'TEST_MAIN_SLEEP'};

        if ($client_count == $self->{'max_clients'}) {
            last;
        }
    }

    $Log->exit();
} ## end sub assign_test_clients

sub set_spn {
    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();
    my $cifs_server = $options{'cifs_server'};
    my $spn_name    = $options{'spn_name'};

    #Local variables
    my ($client_conn_obj, $output, $error);

    # Get connectivity to the DC.
    host($Options_hash->{'CIFS_DC'});
    $client_conn_obj = connect($Options_hash->{'DC_CONNECT_METHOD'},
        '-nologin', -timeout, 300);

    # Verify connectivity to the client
    eval { check_connectivity('conn' => $client_conn_obj); };
    $error = $@;
    if ($error) { die config_result($error); }

    # Here we set the spn for the cifs server.
    # Example set spn command:
    # setspn -a cifs/A13B.nas.ssqa.rtp.netapp.com A13

    my $setspn_l_cmd = "cmd.exe /c setspn -l $cifs_server";
    my $setspn_a_cmd = "cmd.exe /c setspn -a cifs/$spn_name $cifs_server";
    my $setspn_d_cmd = "cmd.exe /c setspn -d cifs/$spn_name $cifs_server";

    $Log->comment("Listing the spns for $cifs_server");
    eval {
        (undef, undef, $output)
            = say($client_conn_obj, $setspn_l_cmd, -timeout, 300);
    };

    $error = $@;
    if ($error) { die config_result($error); }

    $Log->comment(
        "Setting the spn cifs/$spn_name for the cifs server $cifs_server");
    eval {
        (undef, undef, $output)
            = say($client_conn_obj, $setspn_a_cmd, -timeout, 300);
    };
    $error = $@;
    if ($error) { die config_result($error); }

    $Log->comment("Listing the spns for $cifs_server to verify they are set");
    eval {
        (undef, undef, $output)
            = say($client_conn_obj, $setspn_l_cmd, -timeout, 300);
    };

    $error = $@;
    if ($error) { die config_result($error); }

    $Log->exit();
} ## end sub set_spn

sub query_spn {
    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    my $lif_name = $options{'lif_name'};

    #Local variables
    my ($client_conn_obj, $output, $error, $accountName);

    # Get connectivity to the DC.
    host($Options_hash->{'CIFS_DC'});
    $client_conn_obj = connect($Options_hash->{'DC_CONNECT_METHOD'},
        '-nologin', -timeout, 300);

    # Verify connectivity to the client
    eval { check_connectivity('conn' => $client_conn_obj); };
    $error = $@;
    if ($error) { die config_result($error); }

    #Example of removing the Querying the SPN to see if it exists.
    #setspn -T ssqa -Q */ns01-028-028.ssqa.gdl.englab.netapp.com

    my $setspn_q_cmd
        = "cmd.exe /c setspn -T $Options_hash->{'CIFS_DOMAIN'} -Q */$lif_name";

    eval {
        (undef, undef, $output)
            = say($client_conn_obj, $setspn_q_cmd, -timeout, 300);

    };

    $error = $@;
    if ($error) { die config_result($error); }
    if ($output) {
        if ($output =~ /No such SPN found. /) {
            $Log->comment("There is no SPN currently listed for $lif_name");
        } else {

# Expected output could be one of the following
#C:\Users\minwin>setspn -t ssqa.gdl.englab.netapp.com -Q */ns01-028-021.ssqa.gdl.englab.netapp.com
#Checking domain DC=ssqa,DC=gdl,DC=englab,DC=netapp,DC=com

            #No such SPN found.

     #Checking domain DC=ssqa,DC=gdl,DC=englab,DC=netapp,DC=com
     #CN=OP4DMSXN145242,CN=Computers,DC=ssqa,DC=gdl,DC=englab,DC=netapp,DC=com
     #cifs/nst-vsa-196-d4.ssqa.gdl.englab.netapp.com
     #cifs/nst-vsa-195-d2.ssqa.gdl.englab.netapp.com
     #HOST/op4dmsxn145242.ssqa.gdl.englab.netapp.com
     #HOST/OP4DMSXN145242

            foreach my $line (split(/\r/, $output)) {
                chomp($line);
                if ($line =~ /(\bHOST\b)(\/)(\w+)/) {
                    $accountName = $3;
                    $self->delete_old_computer_account(
                        'accountName' => $accountName);
                    last if $accountName ne '';
                }
            }
        }

    }
    $Log->exit();
} ## end sub lookup_spn

sub delete_old_computer_account {
    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    my $account = $options{'accountName'};

    #Local variables
    my ($client_conn_obj, $output, $error, $accountName);

    # Get connectivity to the DC.
    host($Options_hash->{'CIFS_DC'});
    $client_conn_obj = connect($Options_hash->{'DC_CONNECT_METHOD'},
        '-nologin', -timeout, 300);

    # Verify connectivity to the client
    eval { check_connectivity('conn' => $client_conn_obj); };
    $error = $@;
    if ($error) { die config_result($error); }

    #PowerShell Command we will run.
    my $powershell_cmd
        = "echo | powershell.exe c:\\powershell\\DeleteAccount.ps1 -ComputerName $account";

    eval {
        (undef, undef, $output)
            = say($client_conn_obj, $powershell_cmd, -timeout, 300);
    };

    $error = $@;

    if ($error) {
        die config_result("Error attempting to delete the computer account");
    }

    $Log->exit();
} ## end sub delete_old_computer_account
1;
