#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     CIFS IO Topology Library
##
## @author      bellizzi@netapp.com
##
## @pod here

=head1 NAME

CM.pm - CIFS Stress Scalability test Topology Library for ONTAP

=head1 SYNOPSIS

    See Topology.pm

=head1 DESCRIPTION

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/Topology

=head1 TODO

=head1 SEE ALSO

Options TestManagerLite Topology

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  David Bellizzi (bellizzi@netapp.com)
  $Id: $
  Last revised $Date: 2008/09/30 $

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=cut

package Topology::NAS::CIFS::CM;

use base qw(Topology::NAS::CIFS);

use strict;
use warnings;

use Module::Common qw(validate_mixed);
use Params::Validate
    qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Tharn
    qw($Log connect config_result host say snooze script_result fail_result);
use Data::Dumper qw(Dumper);
use Options;
use File::Basename;
use CIFS::Client;
use Net::IP;
use Clone qw(clone);
use YAML;
use CIFS::Setup qw(cifs_setup);
use Sort::Key::Natural qw(natsort);
use NetAddr::IP::Lite;
use Utils::IPManager;
use Utils::HostManager::Helper;
use Utils::DNS qw(get_name_by_ip);
use Limits;
use POSIX qw(strftime);
use SSQA::License;
use NACL::APISet;
use NACL::C::Volume;
use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::VserverCifs;
use NACL::C::VserverCifsShare;
use NACL::C::VserverServicesDns;
use NACL::C::VserverServicesNisDomain;
use NACL::C::Interface;
use NACL::C::Aggregate;
use NACL::C::Qtree;
use NACL::C::RoutingGroups;
use NACL::C::Route;
use NACL::C::ExportPolicyRule;
use NACL::C::SystemServicesNtpConfig;
use NACL::C::SystemServicesNtpServer;
use NACL::C::UnixGroup;
use NACL::C::UnixUser;
use NACL::C::VserverCifsDomainPreferredDc;
use NACL::C::NetworkPort;
use NACL::C::VserverCifsOptions;
use NACL::C::VserverCifsShareAccessControl;
use NACL::C::SystemNodeDate;
use NACL::C::SystemServicesDns;
use NACL::C::Antivirus;
use NACL::C::AntivirusEngine;
use NACL::C::AntivirusOnAccessPolicy;
use NACL::C::AntivirusUpdate;
use NACL::C::Quota;
use NACL::C::QuotaPolicy;
use NACL::C::QuotaPolicyRule;
use NACL::C::SystemNodeAutosupport;
use NACL::C::VserverCifsSecurity;
use NACL::CS::DiagNbladeReplay;
use NACL::CS::License;
use NACL::CS::VserverCifs;
use NACL::C::ClusterDate;
use NATE::BaseException qw(:try);
use if ($^O ne "MSWin32"), 'DBM::Deep';

$Data::Dumper::Sortkeys = 1;

my $Options = new Options;
$Options->process();

sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $self     = {@_,};
    $Log->enter();
    bless($self, $class);
    $self->{reserve_clients} = 0;
    $Log->exit();
    return $self;
} ## end sub new

my ($Options_hash, %apiset, $max_clients, $Response, $ci, $topology_database);
my @ontap_shares = ('admin$', 'ipc$');
my %releaseRegex;

our $IO_TOOLS;

sub define_topology {
    $Log->enter();

    my $self = shift;

    # Execute the one from the parent
    $self->SUPER::define_topology(@_);

    $Options_hash = $Options->get_hash();
    $Log->debug(Dumper($Options_hash));

    if ($Options_hash->{'CLIENT_ASSIGNMENT'} eq 'io'
        && (   !$Options_hash->{'TEST_MAIN_DURATION'}
            || !$Options_hash->{'TEST_MAIN_SLEEP'})
        ) {
        die script_result("TEST_MAIN_DURATION and/or "
                . "TEST_MAIN_SLEEP are required\n"
                . Dumper($Options_hash));
    }

    # Find calling script's name
    my $script = fileparse($0, qr/\.[^.]*/);
    if ($Options_hash->{'SCRIPT_INDEX'}) {
        $script .= "_$Options_hash->{'SCRIPT_INDEX'}";
    }
    $self->{'testname'} = $script;

    $topology_database
        = DBM::Deep->new("$Options_hash->{'TOP_LOGDIR'}/$script.db");
    $self->{'topology_database'} = $topology_database;
    $topology_database->put('shares', {});

    # Wait for other scripts to start if needed
    if ($Options_hash->{'WAIT_FOR_FILE'}) {
        $Log->comment("Waiting for $Options_hash->{'WAIT_FOR_FILE'}");
        $self->wait_for_file('file' => $Options_hash->{'TOP_LOGDIR'} . '/'
                . $Options_hash->{'WAIT_FOR_FILE'});
    }

    # Startup the IP Manager
    $self->{'ip_manager'} = $self->start_ip_manager();
    my $ip_manager = $self->{'ip_manager'};

    # Retrieve the clients from the client file or ndispd
    my @clients
        = retrieve_clients('script' => $script, 'options' => $Options_hash);
    $self->{'clients'} = 0;
    $self->{'clients'} = \@clients;

    # Calculate MAX_CLIENTS
    $self->{'max_clients'} = get_max_clients();
    $max_clients = $self->{'max_clients'};

    $ci = NACL::C::Node->find();

    my $api_set = $ci->apiset(
        category  => "Node",
        interface => "CLI",
        set       => "CMode"
    );

    $api_set->set_timeout(timeout => $Options_hash->{'APISET_TIMEOUT'});

    my $version_mgr = $api_set->get_version_manager();

    # Attribute can be any of branch; release; build_options; build_path;
    # change_number; wafl_version; node_version
    my $release
        = $version_mgr->get_version_attribute('attribute' => 'release');
    $self->{'release'}
        = $version_mgr->map_release_to_codename(release => $release);
    $Log->comment("Detected release: $self->{'release'}");

    # Discover the current config;
    my @targets = (
        'antivirus',   'aggregates', 'cifs',     'licenses',
        'lifs',        'qtrees',     'quotas',   'ports',
        'shares',      'volumes',    'vservers', 'routes',
        'ntp_servers', 'system'
    );

    $self->discover_topology('targets' => \@targets);

    # If we are here we are defining a topology so check for required params
    if (($Options_hash->{'VSERVER_COUNT'} * 2)
        > $Options_hash->{'VOLUME_COUNT'}) {
        die config_result('Not enough volumes specified for '
                . 'all vservers root and data volumes');
    }

    if (!$Options_hash->{'AGGREGATES'}) {
        die script_result('AGGREGATES is a required parameter');
    }

    if (!@{$Options_hash->{'DATA_INTERFACES'}}) {
        die script_result('DATA_INTERFACES is a required parameter');
    }

    # Make sure all specified aggregates exist
    foreach my $aggregate (split(/,/, $Options_hash->{'AGGREGATES'})) {
        if (!grep { $_ eq $aggregate }
            keys(%{$self->{'discovered_topology'}->{'aggregates'}})
            ) {
            die script_result(
                "Aggregate $aggregate specified in AGGREGATES does not exist"
            );
        }
    }

    if (   $Options_hash->{'PRE_TEST_CLEANUP'}
        && $Options_hash->{'PRE_TEST_CLEANUP'} eq 'discovered') {
        $Log->comment('Performing pretest clean up of discovered items');
        $self->filer_cleanup_discovered();
        snooze 1;
        $self->discover_topology('targets' => \@targets);
    }

    # Get the IO configs and store them in $self
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'io'
        && (   $Options_hash->{'IO_LOAD_CONFIG'}
            || $Options_hash->{'IO_TOOLS_CONFIG'})
        ) {
        $self->get_io_configs();
    }

    # Create the vserver names
    foreach my $vserver_index (1 ... $Options_hash->{'VSERVER_COUNT'}) {

        my $vserver = 'vs_' . $self->{'testname'} . '_' . $vserver_index;
        $self->{'defined_topology'}->{'vservers'}->{$vserver}->{'index'}
            = $vserver_index;
    }

    if (!$Options_hash->{'CIFS_SERVER_BASENAME'}) {
        my $index_digits = length($Options_hash->{'VSERVER_COUNT'});
        my $time         = strftime("%H%M%S", localtime());
        my $time_digits  = length($time);
        my $chars        = 15 - ($index_digits + $time_digits);
        $self->{'cifs_server_basename'}
            = $self->random_string('chars' => $chars) . $time;
    } else {
        $self->{'cifs_server_basename'}
            = $Options_hash->{'CIFS_SERVER_BASENAME'};

    }

    $self->{'vserver_count'}   = $Options_hash->{'VSERVER_COUNT'};
    $self->{'aggregates'}      = $Options_hash->{'AGGREGATES'};
    $self->{'data_interfaces'} = $Options_hash->{'DATA_INTERFACES'};
    $self->{'volume_count'}    = $Options_hash->{'VOLUME_COUNT'};
    $self->{'volume_size'}     = $Options_hash->{'VOLUME_SIZE'};

    $self->{'root_volume_count'} = $self->{'vserver_count'};
    $self->{'data_volume_count'}
        = $self->{'volume_count'} - $self->{'vserver_count'};
    if ($Options_hash->{'ORIGIN_VOLUME_PERCENT'}) {
        $self->{'origin_volume_percentage'}
            = $Options_hash->{'ORIGIN_VOLUME_PERCENT'};
        $self->{'origin_volume_count'} = ($self->{'data_volume_count'}
                * ($self->{'origin_volume_percentage'} / 100));
        $self->{'cache_volume_count'}
            = $self->{'origin_volume_count'} * @{$Options_hash->{'NODES'}};
    }

    # Calculate the max volumes per vserver
    $self->{'data_volumes_per_vserver'}
        = int($self->{'data_volume_count'} / $self->{'vserver_count'});

    $self->{'data_volumes_per_node'}
        = $self->{'data_volume_count'} / @{$Options_hash->{'NODES'}};

    # Use non stripped aggregates for root volumes
    my @root_aggregates;
    my $discovered_aggregates
        = $self->{'discovered_topology'}->{'aggregates'};
    foreach my $aggregate (split(/,/, $Options_hash->{'AGGREGATES'})) {
        if (@{$discovered_aggregates->{$aggregate}->{'nodes'}} > 1) {
            next;
        }
        push(@root_aggregates, $aggregate);
    }

    if (!@root_aggregates) {
        die config_result(
            'Unable to find a single node ' . 'aggregate for root volumes');
    }

    # Assign the vserver root vols
    my $volume_index = $self->assign_volumes_to_aggregates(
        'aggregates' => join(',', @root_aggregates),
        'count'      => $self->{'vserver_count'},
        'index'      => 1,
        'script'     => $script,
        'size'       => $Options_hash->{'ROOT_VOLUME_SIZE'},
        'type'       => 'RW'
    );

    # Assign the vserver data vols to
    $volume_index = $self->assign_volumes_to_aggregates(
        'aggregates' => $Options_hash->{'AGGREGATES'},
        'count'      => $self->{'data_volume_count'},
        'index'      => $volume_index,
        'script'     => $script,
        'size'       => $Options_hash->{'VOLUME_SIZE'},
        'type'       => 'RW'
    );

    # Spread the volumes across vservers
    $Log->comment("Adding volumes to vservers");

    my $volume_count     = 0;
    my $defined_shares   = $topology_database->{'shares'};
    my $defined_volumes  = $self->{'defined_topology'}->{'volumes'};
    my $defined_vservers = $self->{'defined_topology'}->{'vservers'};
    my @volumes          = natsort(keys(%{$defined_volumes}));
    my $share_index      = 1;

    # Assign the root volumes and shares to vservers
    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {

        if (my $volume = shift(@volumes)) {
            $Log->comment("Adding root volume $volume to $vserver");

            push(
                @{  $self->{'defined_topology'}->{'vservers'}->{$vserver}
                        ->{'volumes'}
                    },
                $volume
            );

            $self->{'defined_topology'}->{'vservers'}->{$vserver}
                ->{'rootvolume'} = $volume;

            $self->{'defined_topology'}->{'volumes'}->{$volume}
                ->{'rootvolume'} = 1;

            $self->{'defined_topology'}->{'volumes'}->{$volume}->{'vserver'}
                = $vserver;

            if ($Options_hash->{'SHARE_ROOT_VOLUMES'}) {
                my $share = 's_' . $vserver;
                $Log->comment("Adding vserver root share $share to $vserver");

                push(
                    @{  $self->{'defined_topology'}->{'vservers'}->{$vserver}
                            ->{'shares'}
                        },
                    $share
                );

                # Update the defined shares
                $topology_database->{'shares'}->{$share}->{'vserver'}
                    = $vserver;
                $topology_database->{'shares'}->{$share}->{'volume'}
                    = $volume;
                $topology_database->{'shares'}->{$share}->{'path'} = '/';
                $topology_database->{'shares'}->{$share}->{'filer_setup'} = 1;
            }
        }
    }

    my $origin_volume_count = $self->{'origin_volume_count'};

    # Assign the data volumes and their shares to vservers
    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {

        # Assign volumes to the defined vserver
        foreach (1 ... $self->{'data_volumes_per_vserver'}) {
            $volume_count++;
            if (my $volume = shift(@volumes)) {
                $Log->comment("Adding data volume $volume to $vserver");

                push(
                    @{  $self->{'defined_topology'}->{'vservers'}->{$vserver}
                            ->{'volumes'}
                        },
                    $volume
                );

                # Update the defined volumes with the owning vserver
                $self->{'defined_topology'}->{'volumes'}->{$volume}
                    ->{'vserver'} = $vserver;

                # Update the defined volumes rootvolume status
                $self->{'defined_topology'}->{'volumes'}->{$volume}
                    ->{'rootvolume'} = 0;

                my $style
                    = $self->{'defined_topology'}->{'volumes'}->{$volume}
                    ->{'style'};

                # Create cache volumes if needed
                if ($origin_volume_count && $style eq 'flex') {

                    # Decrement the origin count
                    $origin_volume_count--;

                    $volume_index = $self->define_cache_volumes(
                        'index'  => $volume_index,
                        'script' => $script,
                        'size'   => $Options_hash->{'VOLUME_SIZE'},
                        'volume' => $volume
                    );
                }

                if ($Options_hash->{'SHARE_DATA_VOLUMES'}) {
                    my $share = 's_' . $volume;
                    $Log->comment("Adding volume share $share to $vserver");

                    push(
                        @{  $self->{'defined_topology'}->{'vservers'}
                                ->{$vserver}->{'shares'}
                            },
                        $share
                    );

                    # Update the defined shares
                    $topology_database->{'shares'}->{$share}->{'vserver'}
                        = $vserver;
                    $topology_database->{'shares'}->{$share}->{'volume'}
                        = $volume;
                    $topology_database->{'shares'}->{$share}->{'path'}
                        = "/$volume";
                    $topology_database->{'shares'}->{$share}->{'filer_setup'}
                        = 1;
                }
            }
        }
    }

    # If we have left over volumes assign them to the 1st vserver
    my @vservers = natsort(keys(%{$defined_vservers}));
    my $vserver  = $vservers[0];
    my $defined_vserver
        = $self->{'defined_topology'}->{'vservers'}->{$vserver};
    while ($volume_count < $self->{'data_volume_count'}) {
        if (my $volume = shift(@volumes)) {
            $Log->comment("Adding data volume $volume to $vserver");
            push(@{$defined_vserver->{'volumes'}}, $volume);

            # Update the defined volumes with the owning vserver
            $self->{'defined_topology'}->{'volumes'}->{$volume}->{'vserver'}
                = $vserver;

            # Update the defined volumes rootvolume status
            $self->{'defined_topology'}->{'volumes'}->{$volume}
                ->{'rootvolume'} = 0;

            my $style = $self->{'defined_topology'}->{'volumes'}->{$volume}
                ->{'style'};

            # Create cache volumes if needed
            if ($origin_volume_count && $style eq 'flex') {

                # Decrement the origin count
                $origin_volume_count--;

                $volume_index = $self->define_cache_volumes(
                    'index'  => $volume_index,
                    'script' => $script,
                    'size'   => $Options_hash->{'VOLUME_SIZE'},
                    'volume' => $volume,
                );
            }
            $volume_count++;
        }
    }

    # Parse address ranges
    if (!$Options_hash->{'DESTIP'}) {
        die config_result('DESTIP is not defined');
    }

    my @dest_networks = $ip_manager->destination_networks();
    my $index         = 1;
    my $net_obj       = shift(@dest_networks);
    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {
        $Log->comment("Adding addresses to $vserver");
        my $defined_vserver = $defined_vservers->{$vserver};
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            my $discovered_ports
                = $self->{'discovered_topology'}->{'ports'}->{$node};
            my @ports = ();
            foreach (1 ... $Options_hash->{'DATA_LIFS_PER_NODE'}) {

                if (!@ports) {
                    foreach my $port (natsort(keys(%{$discovered_ports}))) {
                        if ((   grep { $_ eq $port }
                                @{$Options_hash->{'DATA_INTERFACES'}}
                            )
                            && $discovered_ports->{$port}->{'role'} eq 'data'
                            ) {
                            push(@ports, $port);
                        }
                    }
                }

                my $lif = $node . '_data' . $index;
                push(@{$defined_vserver->{'lifs'}}, $lif);

                my $ip_obj = $net_obj->assign_next_address();

                if (!ref($ip_obj)) {
                    if (@dest_networks) {
                        $net_obj = shift(@dest_networks);
                        $ip_obj  = $net_obj->assign_next_address();
                    } else {
                        die config_result('No more addresses available '
                                . 'please add ranges to DESTIP');
                    }
                }

                if (!$ip_obj->address()) {
                    die config_result('Unable to obtain an unassigned '
                            . 'address from for the network '
                            . $net_obj->networkID()
                            . ' . Check the value of DESTIP');
                }

                $self->{'defined_topology'}->{'lifs'}->{$vserver}->{$lif}
                    ->{'address'} = $ip_obj->address();
                $self->{'defined_topology'}->{'lifs'}->{$vserver}->{$lif}
                    ->{'netmask'} = $ip_obj->netmask();
                $self->{'defined_topology'}->{'lifs'}->{$vserver}->{$lif}
                    ->{'network'} = $net_obj->networkID();
                $self->{'defined_topology'}->{'lifs'}->{$vserver}->{$lif}
                    ->{'port'} = shift(@ports);
                $self->{'defined_topology'}->{'lifs'}->{$vserver}->{$lif}
                    ->{'vserver'} = $vserver;
                $self->{'defined_topology'}->{'lifs'}->{$vserver}->{$lif}
                    ->{'node'} = $node;

                $Log->debug(
                    "Assigning " . $ip_obj->address() . " to data LIF $lif");

                if (!$ip_obj->address() && !@dest_networks) {
                    last;
                }
                $index++;
            }
        }
    }

    # Create the CIFS server names
    $Log->comment("Creating the CIFS server name for all vservers");
    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {

        $self->{'defined_topology'}->{'vservers'}->{$vserver}->{'cifs_server'}
            = $vserver;
    }

    # If I/O loads are specified parse each load and get the max number of
    # shares per client per load and assign the clients to each load
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'io') {
        $self->parse_io_configs('script' => $script);
        $share_index = $self->assign_io_clients(
            'script'      => $script,
            'share_index' => $share_index
        );

        if (keys(%{$self->{'test_options'}}) == 0) {
            die config_result('No test options defined');
        }

    }

    # This block of code defines shares and assigns them to clients to be used
    # with the mapstress.exe win32 tool for shares limits testing
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'shares') {

        $self->define_shares(
            'recipe'      => 'clients',
            'script'      => $script,
            'share_index' => $share_index
        );

        $self->assign_share_verification_clients(
            'script'      => $script,
            'share_index' => $share_index
        );
        if (keys(%{$self->{'test_options'}}) == 0) {
            die config_result('No test options defined');
        }

    }

    # Assign clients for NTFS streams testing
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'streams') {

        $share_index = $self->assign_stream_clients(
            'script'      => $self->{'testname'},
            'clients'     => $self->{'clients'},
            'share_index' => $share_index
        );

        if (keys(%{$self->{'test_options'}}) == 0) {
            die config_result('No test options defined');
        }

    }

    # Define shares and assign them to swifttest user parameters
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'swifttest') {
        $self->define_shares(
            'recipe'      => 'swifttest',
            'script'      => $self->{'testname'},
            'share_index' => $share_index,
        );

        $Log->comment("Creating SwiftTest user parameters");
        $topology_database->put('swifttest', {});

        # Loop through each defined share
        $Log->comment("Creating server and share");
        foreach my $vserver (natsort(keys(%{$defined_vservers}))) {
            foreach my $share (
                natsort(@{$defined_vservers->{$vserver}->{'shares'}})) {

                # Only get the shares we want
                my $regex = '^s_' . $self->{'testname'};
                if ($share !~ /$regex/) {
                    next;
                }

                # Get the server name from the defined share
                my $server = $defined_shares->{$share}->{'server'};

                push(
                    @{  $topology_database->{'swifttest'}->{'user_parameters'}
                            ->{'share'}
                        },
                    $share
                );
                push(
                    @{  $topology_database->{'swifttest'}->{'user_parameters'}
                            ->{'server'}
                        },
                    $server
                );

            }
        }

        # Create the specified test domain and user passwd
        $Log->comment("Creating domain and user passwd");
        my $domain = $Options_hash->{'CIFS_DOMAIN'};
        $domain =~ s/^([^\.]+).+/$1/;
        push(
            @{  $topology_database->{'swifttest'}->{'user_parameters'}
                    ->{'domain'}
                },
            $domain
        );
        push(
            @{  $topology_database->{'swifttest'}->{'user_parameters'}
                    ->{'password'}
                },
            $Options_hash->{'CIFS_PASSWD'}
        );

        # Create the specified test users
        if ($Options_hash->{'TEST_USER_COUNT'}) {
            $Log->comment("Creating user");
            my $test_user_basename = $Options_hash->{'TEST_USER_BASENAME'};
            my $test_user_count    = $Options_hash->{'TEST_USER_COUNT'};
            for (my $row = 0; $row < $test_user_count; $row++) {

                # Create unique username and file
                my $user = $test_user_basename . ($row + 1);
                my $path = $self->{'testname'} . '_' . ($row + 1);

                push(
                    @{  $topology_database->{'swifttest'}->{'user_parameters'}
                            ->{'user'}
                        },
                    $user
                );

            }
        }

        # Create the specified test files
        if ($Options_hash->{'TEST_FILE_COUNT'}) {
            $Log->comment("Creating files");
            my $test_file_basename = $Options_hash->{'TEST_FILE_BASENAME'};
            my $test_file_count    = $Options_hash->{'TEST_FILE_COUNT'};
            for (my $row = 0; $row < $test_file_count; $row++) {

                # Create unique file
                my $file = $test_file_basename . ($row + 1);

                push(
                    @{  $topology_database->{'swifttest'}->{'user_parameters'}
                            ->{'file'}
                        },
                    $file
                );

            }
        }

    }

    # Assign clients for TML testing
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'test') {

        $self->assign_test_clients();

        if (keys(%{$self->{'test_options'}}) == 0) {
            die config_result('No test options defined ');
        }
    }

    #$Log->comment(Dumper($self));
    #   die(config_result('Dumping Topology'));

    $Log->exit();

} ## end sub define_topology

sub filer_setup {

    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::filer_setup(%options);

    $Options_hash = $Options->get_hash();

    if (   $Options_hash->{'PRE_TEST_CLEANUP'}
        && $Options_hash->{'PRE_TEST_CLEANUP'} eq 'defined') {
        $self->filer_cleanup_defined();
        my @targets = (
            'antivirus', 'aggregates', 'cifs',     'licenses',
            'lifs',      'qtrees',     'quotas',   'ports',
            'shares',    'volumes',    'vservers', 'routes',
            'ntp_servers'
        );
        $self->discover_topology('targets' => \@targets);
    }

    $self->time_setup();

    my $ip_manager       = $self->{'ip_manager'};
    my $defined_vservers = $self->{'defined_topology'}->{'vservers'};
    my $defined_volumes  = $self->{'defined_topology'}->{'volumes'};
    my $defined_qtrees   = $self->{'defined_topology'}->{'qtrees'};
    my $defined_shares   = $topology_database->{'shares'};
    my $defined_lifs     = $self->{'defined_topology'}->{'lifs'};
    my $discovered_system_settings
        = $self->{'discovered_topology'}->{'system'};
    my $discovered_vservers = $self->{'discovered_topology'}->{'vservers'};
    my $discovered_volumes  = $self->{'discovered_topology'}->{'volumes'};
    my $discovered_qtrees   = $self->{'discovered_topology'}->{'qtrees'};
    my $discovered_quotas   = $self->{'discovered_topology'}->{'quotas'};
    my $discovered_shares   = $self->{'discovered_topology'}->{'shares'};
    my $discovered_licenses = $self->{'discovered_topology'}->{'licenses'};
    my $discovered_ntp_servers
        = $self->{'discovered_topology'}->{'ntp_servers'};
    my $discovered_antivirus = $self->{'discovered_topology'}->{'antivirus'};
    my $discovered_aggregates
        = $self->{'discovered_topology'}->{'aggregates'};

    $Log->comment("Discovered licenses\n" . Dumper($discovered_licenses));

   #my @features = ('GX_Base', 'GX_CIFS', 'Striped_Volume', 'GX_FlexVol_HPO');
    my @features = ('CIFS');
    foreach my $feature (@features) {
        if (!grep { $_ eq $feature } keys(%{$discovered_licenses})) {
            SSQA::License::add(command_interface => $ci, service => $feature);
        }
    }

    if ($self->{'release'} =~ /boilermaker/) {
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            if (!grep { $_ eq 'dns' }
                keys(%{$discovered_system_settings->{$node}})
                ) {
                $Log->comment("Configuring DNS for node $node");
                NACL::C::SystemServicesDns->modify(
                    command_interface => $ci,
                    node              => $node,
                    domains => join(',', @{$Options_hash->{'DNS_SEARCH'}}),
                    'name-servers' => $Options_hash->{'DNS_SERVERS'}
                );
            }
        }
    } else {
        my @dns_servers = split(/,/, $Options_hash->{'DNS_SERVERS'});
        foreach my $vserver (natsort(keys(%{$discovered_vservers}))) {
            my $type = $discovered_vservers->{$vserver}->{'type'};
            if ($type ne 'admin') { next; }
            $Log->comment("Configuring DNS for vserver $vserver");
            NACL::C::VserverServicesDns->modify(
                command_interface => $ci,
                vserver           => $vserver,
                domains           => $Options_hash->{'DNS_SEARCH'},
                'name-servers'    => \@dns_servers
            );
        }
    }

    # Configure AV
    if ($Options_hash->{'ANTIVIRUS'} && $self->{'release'} !~ /boilermaker/) {

        if (grep { $_ eq 'state' } keys(%{$discovered_antivirus})) {
            if ($discovered_antivirus->{'state'} eq 'on') {
                NACL::C::AntivirusEngine->disable(command_interface => $ci);
            }

        }

        my %modify;
        $modify{'command_interface'} = $ci;
        $modify{'num-license'}       = @{$Options_hash->{'NODES'}};
        if ($Options_hash->{'AV_VENDOR'} eq 'sophos') {
            $modify{'vendor'}           = 'sophos';
            $modify{'sophos-http-url'}  = 'SOPHOS';
            $modify{'sophos-http-user'} = 'emnetapp01';
            $modify{'password'}         = '7ghNa0s5';
        } elsif ($Options_hash->{'AV_VENDOR'} eq 'mcafee') {
            $modify{'vendor'} = 'mcafee';
        }

        NACL::C::AntivirusEngine->modify(%modify);

        NACL::C::AntivirusEngine->enable(
            'command_interface' => $ci,
            'method-timeout'    => 3600
        );

        NACL::C::AntivirusUpdate->modify(
            'command_interface' => $ci,
            'auto-update'       => 'off'
        );

        my %policy;
        $policy{'command_interface'} = $ci;
        $policy{'name'}              = 'default';

        # $policy{'id'}                = 1;
        $policy{'scan'}            = 'on';
        $policy{'scan-ro-volume'}  = 'off';
        $policy{'protocols'}       = $Options_hash->{'AV_PROTOCOLS'};
        $policy{'cifs-file'}       = 'include';
        $policy{'cifs-file-list'}  = $Options_hash->{'AV_CIFS_FILE_REGEX'};
        $policy{'cifs-share'}      = 'include';
        $policy{'cifs-share-list'} = $Options_hash->{'AV_CIFS_SHARE_REGEX'};
        $policy{'scan-mandatory'}  = 'on';

        NACL::C::AntivirusOnAccessPolicy->modify(%policy);

    }

    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {
        my $discovered_lifs
            = $self->{'discovered_topology'}->{'lifs'}->{$vserver};
        my @lifs = @{$defined_vservers->{$vserver}->{'lifs'}};
        my $root_volume
            = $self->{'defined_topology'}->{'vservers'}->{$vserver}
            ->{'rootvolume'};
        if (!$root_volume) {
            die config_result("No root volume defined for vserver $vserver");
        }
        my $aggregate = $defined_volumes->{$root_volume}->{'aggregate'};
        my $size      = $defined_volumes->{$root_volume}->{'size'};
        my @nmSwitch  = ('file');
        my @nsSwitch  = ('file');

        if (!grep { $_ eq $vserver } keys(%{$discovered_vservers})) {
            NACL::C::Vserver->create(
                command_interface            => $ci,
                vserver                      => $vserver,
                rootvolume                   => $root_volume,
                aggregate                    => $aggregate,
                'ns-switch'                  => \@nsSwitch,
                'nm-switch'                  => \@nmSwitch,
                'rootvolume-security-style'  => 'mixed',
                language                     => 'C',
                'snapshot-policy'            => 'default',
                'antivirus-on-access-policy' => 'default',
            );

            my $vol = NACL::C::Volume->new(
                command_interface => $ci,
                volume            => $root_volume,
                vserver           => $vserver
            );
            if (!$vol) {
                die script_result("Unable to create the NACL volume object "
                        . "for volume $root_volume on vserver "
                        . $vserver);
            }

            # Change the perms to 777
            # Increase the size to ROOT_VOLUME_SIZE
            $vol->modify(
                'unix-permissions' => 777,
                'size' => $Options_hash->{'ROOT_VOLUME_SIZE'} . 'm'
            );

            my @any = ('any');
            NACL::C::ExportPolicyRule->create(
                command_interface => $ci,
                vserver           => $vserver,
                policyname        => 'default',
                protocol          => \@any,
                clientmatch       => '0.0.0.0/0',
                rorule            => \@any,
                rwrule            => \@any
            );
        }

        foreach my $lif (@lifs) {
            my $address = $defined_lifs->{$vserver}->{$lif}->{'address'};
            my $netmask = $defined_lifs->{$vserver}->{$lif}->{'netmask'};
            my $node    = $defined_lifs->{$vserver}->{$lif}->{'node'};
            my $port    = $defined_lifs->{$vserver}->{$lif}->{'port'};
            if (!grep { $_ eq $lif } keys(%{$discovered_lifs})) {
                NACL::C::Interface->create(
                    command_interface => $ci,
                    server            => $vserver,
                    address           => $address,
                    lif               => $lif,
                    role              => 'data',
                    'home-node'       => $node,
                    'home-port'       => $port,
                    netmask           => $netmask
                );

            }
        }

        $Log->comment("Creating routing groups for IP ranges ");
        my @dest_networks = $ip_manager->destination_networks();
        foreach my $net_obj (@dest_networks) {
            my $gateway = $net_obj->gateway();
            my $netmask = $net_obj->netmask();
            my $network = $net_obj->networkID();
            my $cidr    = $net_obj->cidr();

            $self->wait_for_replay();

            my @targets = ('routes');
            $self->discover_topology(
                'targets' => \@targets,
                'vserver' => $vserver
            );
            my $discovered_routing_groups
                = $self->{'discovered_topology'}->{'routing_groups'}
                ->{$vserver};

            if (!grep { $_ eq "d$network/$cidr" }
                keys(%{$discovered_routing_groups})
                ) {
                NACL::C::RoutingGroups->create(
                    command_interface => $ci,
                    server            => $vserver,
                    'routing-group'   => "d$network/$cidr",
                    subnet            => "$network/$cidr",
                    role              => 'data'
                );
            }
            if (!grep { $_ eq '0.0.0.0/0' }
                keys(
                    %{  $discovered_routing_groups->{"d$network/$cidr"}
                            ->{'routes'}
                        }
                )
                ) {

                NACL::C::Route->create(
                    command_interface => $ci,
                    server            => $vserver,
                    gateway           => $gateway,
                    'routing-group'   => "d$network/$cidr",
                    destination       => '0.0.0.0/0'
                );
            }
        }

        if ($self->{'release'} !~ /boilermaker/) {
            if (!grep { $_ eq 'dns' }
                keys(%{$discovered_vservers->{$vserver}})
                ) {
                $Log->comment("Configuring DNS for vserver $vserver");
                my @dns_servers = split(/,/, $Options_hash->{'DNS_SERVERS'});

                if (!keys(%{$discovered_vservers->{$vserver}->{'dns'}})) {
                    NACL::C::VserverServicesDns->create(
                        command_interface => $ci,
                        vserver           => $vserver,
                        domains           => $Options_hash->{'DNS_SEARCH'},
                        'state'           => 'enabled',
                        'timeout'         => '2',
                        'attempts'        => '1',
                        'name-servers'    => \@dns_servers,
                        );
                } else {
                    NACL::C::VserverServicesDns->modify(
                        command_interface => $ci,
                        vserver           => $vserver,
                        domains           => $Options_hash->{'DNS_SEARCH'},
                        'name-servers'    => \@dns_servers
                        );
                }
            }

            if (   $Options_hash->{'NIS_SERVERS'}
                && $Options_hash->{'NIS_DOMAINNAME'}) {
                if (!grep { $_ eq 'nis' }
                    keys(%{$discovered_vservers->{$vserver}})
                    ) {
                    $Log->comment("Configuring NIS for vserver $vserver");
                    my @nis_servers
                        = split(/,/, $Options_hash->{'NIS_SERVERS'});
                    NACL::C::VserverServicesNisDomain->create(
                        command_interface => $ci,
                        active            => 'true',
                        vserver           => $vserver,
                        domain    => $Options_hash->{'NIS_DOMAINNAME'},
                        'servers' => \@nis_servers
                    );
                }
            } else {
                $Log->comment(
                    "NIS_DOMAINNAME and NIS_SERVERS are not '
		    . 'defined skipping NIS configuration"
                );
            }
        }

        # Configure the local unix user accounts
        if ($self->{'release'} =~ /boilermaker|rollingrock/) {
            if ($Options_hash->{'LOCAL_UNIX_USERS'}) {
                $Log->comment(
                    "Adding local unix user accounts for vserver $vserver");
                my $discovered_users
                    = $discovered_vservers->{$vserver}->{'unix_users'};
                foreach my $record (split /,/, $Options_hash->{LOCAL_UNIX_USERS})
                {
                    my ($user, $uid, $gid, $name) = split /:/, $record;
                    if (!grep { $_ eq $user } keys(%{$discovered_users})) {
                        $Log->comment("Adding local unix user "
                                . "$user to vserver $vserver");
                        $self->unix_user_new(
                            'gid'     => $gid,
                            'name'    => $name,
                            'uid'     => $uid,
                            'user'    => $user,
                            'vserver' => $vserver
                        );

                    }
                }
            }

            # Configure the local unix group accounts
            if ($Options_hash->{'LOCAL_UNIX_GROUPS'}) {
                $Log->comment("Adding local unix groups to vserver $vserver");
                my $discovered_groups
                    = $discovered_vservers->{$vserver}->{'unix_groups'};
                foreach my $record (split /,/, $Options_hash->{LOCAL_UNIX_GROUPS})
                {
                    my ($group, $gid, $users) = split(/:/, $record);
                    if (!grep { $_ eq $group } keys(%{$discovered_groups})) {
                        $Log->comment("Adding local unix group "
                                . "$group to vserver $vserver");
                        $self->unix_group_new(
                            'gid'     => $gid,
                            'group'   => $group,
                            'vserver' => $vserver
                        );
                        foreach my $user (split /\|/, $users) {
                            $Log->comment("Adding local unix user $user to group "
                                    . "$group on vserver $vserver");
                            $self->unix_group_add_user(
                                'group'   => $group,
                                'user'    => $user,
                                'vserver' => $vserver
                            );
                        }
                    }
                }
            }

        }
        # Create the remaining data volumes
        foreach my $volume (@{$defined_vservers->{$vserver}->{'volumes'}}) {
            if ($defined_volumes->{$volume}->{'rootvolume'}) {
                next;
            } elsif ($defined_volumes->{$volume}->{'origin_volume'}) {
                next;
            }
            if (!grep { $_ eq $volume } keys(%{$discovered_volumes})) {
                $aggregate = $defined_volumes->{$volume}->{'aggregate'};
                $size      = $defined_volumes->{$volume}->{'size'};
                my %args = (
                    volume             => $volume,
                    command_interface  => $ci,
                    aggregate          => $aggregate,
                    'unix-permissions' => '777',
                    'junction-path'    => "/$volume",
                    vserver            => $vserver,
                    size               => $size . 'm',
                );
                if (@{$discovered_aggregates->{$aggregate}->{'nodes'}} > 1) {
                    $args{'stripe-format'} = $Options_hash->{'STRIPE_FORMAT'};
                }
                NACL::C::Volume->create(%args);
            }
        }

        $self->wait_for_replay();

        # Create cache volumes if needed
        if ($self->{'release'} !~ /boilermaker/) {
            foreach my $volume (@{$defined_vservers->{$vserver}->{'volumes'}})
            {
                if (!grep { $_ eq 'origin_volume' }
                    keys(%{$defined_volumes->{$volume}})
                    ) {
                    next;
                }

                if (!grep { $_ eq $volume } keys(%{$discovered_volumes})) {
                    my $origin
                        = $defined_volumes->{$volume}->{'origin_volume'};
                    $aggregate = $defined_volumes->{$volume}->{'aggregate'};
                    eval {
                        NACL::C::Volume->create(
                            command_interface         => $ci,
                            volume                    => $volume,
                            aggregate                 => $aggregate,
                            type                      => 'DC',
                            vserver                   => $vserver,
                            'flexcache-origin-volume' => $origin
                        );
                    };
                }
            }
        }

        if (!keys(%{$discovered_vservers->{$vserver}->{'cifs'}})) {

            my $user = $Options_hash->{'CIFS_USER'};
            $user =~ s/\w+\\(.+)/$1/g;

            if (@{$Options_hash->{'PREFERRED_DOMAIN_CONTROLLERS'}}) {

                my $domain_controllers = join(',',
                    @{$Options_hash->{'PREFERRED_DOMAIN_CONTROLLERS'}});
                my $disocvered_pref_dcs
                    = $self->{'discovered_topology'}->{'vservers'}->{$vserver}
                    ->{'cifs'}->{'preferred-dc'}
                    ->{$Options_hash->{'CIFS_DOMAIN'}}->{'preferred-dc'};

                if ($domain_controllers ne $disocvered_pref_dcs) {
                    $Log->comment(
                        "Configuring preferred-dc on vserver $vserver");

                    NACL::C::VserverCifsDomainPreferredDc->add(
                        command_interface => $ci,
                        vserver           => $vserver,
                        lif               => "\"\"",
                        domain            => $Options_hash->{'CIFS_DOMAIN'},
                        'preferred-dc' =>
                            $Options_hash->{'PREFERRED_DOMAIN_CONTROLLERS'}
                    );
                }
            }

            my $cifs_server_name = $self->{'cifs_server_basename'};
            $cifs_server_name .= $defined_vservers->{$vserver}->{'index'};
            my $attempts = 0;
            while (1) {
                $attempts++;
                my $success = 0;
                eval {
                    NACL::C::VserverCifs->create(
                        command_interface => $ci,
                        vserver           => $vserver,
                        domain            => $Options_hash->{'CIFS_DOMAIN'},
                        'cifs-server'     => $cifs_server_name,
                        username          => $user,
                        password          => $Options_hash->{'CIFS_PASSWD'},
                        ou                => 'CN=Computers'
                    );
                };

                snooze 1;
                $self->wait_for_replay();

                try {

                    NACL::CS::VserverCifs->fetch(
                        command_interface => $ci,
                        filter            => {vserver => $vserver}
                    );
                    $success = 1;
                }
                catch NATE::BaseException with {
                    my $Exception = shift;
                    $Log->warn(
                        'CIFS Server not created: ' . $Exception->text());

                    if ($attempts == 3) {
                        die config_result(
                                  "Failed to create a CIFS server for "
                                . "vserver $vserver after $attempts "
                                . "attempts");
                    }

                    # Wait for secd replay. See burt 483459
                    snooze 60;
                };
                if ($success) { last; }
            }

            $self->wait_for_replay();

            if ($Options_hash->{'DEFAULT_UNIX_USER'}) {
                NACL::C::VserverCifsOptions->modify(
                    command_interface => $ci,
                    vserver           => $vserver,
                    'default-unix-user' =>
                        $Options_hash->{'DEFAULT_UNIX_USER'}
                );
            }

            if (@{$Options_hash->{'WINS_SERVERS'}}) {
                NACL::C::VserverCifsOptions->modify(
                    command_interface => $ci,
                    vserver           => $vserver,
                    'wins-servers'    => $Options_hash->{'WINS_SERVERS'}
                );
            }
            if (   $self->{'release'} !~ /boilermaker/
                && $Options_hash->{'SMB_SIGNING'}) {
                $Log->comment("Enabing SMB signing on vserver $vserver");
                NACL::C::VserverCifsSecurity->modify(
                    command_interface     => $ci,
                    vserver               => $vserver,
                    'is-signing-required' => 'true',
                );
            }
        }

        if ($self->{'release'} !~ /boilermaker/ && $Options_hash->{'QUOTAS'})
        {

            foreach my $volume (keys(%{$defined_volumes})) {

                my $owning_vserver = $defined_volumes->{$volume}->{'vserver'};
                my $style          = $defined_volumes->{$volume}->{'style'};
                my $type           = $defined_volumes->{$volume}->{'type'};
                if (   $owning_vserver ne $vserver
                    || $style eq 'striped'
                    || $type  eq 'DC') {
                    next;
                }

                $Log->comment("Creating quota policy rules for $volume");

                #policy,target:type:disk:files:thold:sdisk:sfile
                #default,CIFS_USER:user@volume:-:-:700M:700M:5K

                foreach my $quota (split(/;/, $Options_hash->{'QUOTAS'})) {
                    if ($quota !~ /\@volume/) { next; }
                    $quota =~ s/CIFS_USER/$Options_hash->{'CIFS_USER'}/g;
                    my @fields = split(/:/, $quota);
                    $fields[2] =~ s/\@volume//g;

                    if ($fields[0] ne 'default'
                        && !grep { $_ eq $fields[0] }
                        keys(%{$discovered_quotas->{$vserver}})
                        ) {

                        $Log->comment(
                            "Creating quota policy $fields[0] for $volume");
                        NACL::C::QuotaPolicy->create(
                            command_interface => $ci,
                            vserver           => $vserver,
                            'policy-name'     => $fields[0],
                        );
                    }

                    if (!$fields[1]) { $fields[1] = '""'; }
                    my $quota_rules
                        = $discovered_quotas->{$vserver}->{$fields[0]}
                        ->{'rules'};

                    if (!grep { $_->{'volume'} eq $volume } @{$quota_rules}) {

                        NACL::C::QuotaPolicyRule->create(
                            command_interface => $ci,
                            vserver           => $vserver,
                            'policy-name'     => $fields[0],
                            target            => $fields[1],
                            type              => $fields[2],
                            'disk-limit'      => $fields[3],
                            'file-limit'      => $fields[4],
                            'threshold'       => $fields[5],
                            'soft-disk-limit' => $fields[6],
                            'soft-file-limit' => $fields[7],
                            qtree             => "",
                            volume            => $volume
                        );

                        push(
                            @{  $self->{'defined_topology'}->{'quotas'}
                                    ->{$vserver}->{$fields[0]}
                                },
                            $quota
                        );
                    }

                }

                foreach my $qtree (keys(%{$defined_qtrees})) {

                    my $owning_volume = $defined_qtrees->{$qtree}->{'volume'};
                    if ($volume ne $owning_volume) { next; }

                    #target:type:disk:files:thold:sdisk:sfile
                    #CIFS_USER:user@qtree:-:-:700M:700M:5K

                    foreach my $quota (split(/;/, $Options_hash->{'QUOTAS'}))
                    {
                        if ($quota !~ /\@qtree/) { next; }
                        $quota =~ s/CIFS_USER/$Options_hash->{'CIFS_USER'}/g;
                        my @fields = split(/:/, $quota);
                        $fields[2] =~ s/\@qtree//g;
                        $Log->comment(
                            "Adding quota to qtree $qtree on $volume");

                        if ($fields[0] ne 'default'
                            && !grep { $_ eq $fields[0] }
                            keys(%{$discovered_quotas->{$vserver}})
                            ) {
                            NACL::C::QuotaPolicy->create(
                                command_interface => $ci,
                                vserver           => $vserver,
                                'policy-name'     => $fields[0],
                            );
                        }

                        if (!$fields[1]) { $fields[1] = '""'; }

                        my $quota_rules
                            = $discovered_quotas->{$vserver}->{$fields[0]}
                            ->{'rules'};

                        if (!grep { $_->{'qtree'} eq $qtree } @{$quota_rules})
                        {

                            NACL::C::QuotaPolicyRule->create(
                                command_interface => $ci,
                                vserver           => $vserver,
                                'policy-name'     => $fields[0],
                                target            => $fields[1],
                                type              => $fields[2],
                                'disk-limit'      => $fields[3],
                                'file-limit'      => $fields[4],
                                'threshold'       => $fields[5],
                                'soft-disk-limit' => $fields[6],
                                'soft-file-limit' => $fields[7],
                                qtree             => $qtree,
                                volume            => $volume
                            );

                            push(
                                @{  $self->{'defined_topology'}->{'quotas'}
                                        ->{$vserver}->{$fields[0]}
                                    },
                                $quota
                            );

                        }
                    }
                }

                NACL::C::Quota->on(
                    command_interface => $ci,
                    vserver           => $vserver,
                    volume            => $volume
                );

            }

        }

        $self->wait_for_replay();
    }

    # Create the spn for the cifs server.
    if ($Options_hash->{'KERBEROS'}) {
        if (   $Options_hash->{'PRE_TEST_CLEANUP'}
            && $Options_hash->{'PRE_TEST_CLEANUP'} ne 'disabled'
            && $Options_hash->{'CIFS_SERVER_ACCESS_MODE'} ne 'ip') {
            $self->create_spn();

        }
    }

    my %mapped;

    # Create the vserver root volume shares
    my ($share_count, $last_share) = $self->create_shares(
        'regex'  => '^s_vs_' . $self->{'testname'},
        'mapped' => \%mapped
    );

    $self->wait_for_replay();

    $self->check_share_propagation('share' => $last_share);

    # Create the volume shares
    ($share_count, $last_share) = $self->create_shares(
        'regex'  => '^s_v_' . $self->{'testname'},
        'mapped' => \%mapped
    );

    $self->wait_for_replay();

    $self->check_share_propagation('share' => $last_share);

    # Create the data shares
    ($share_count, $last_share) = $self->create_shares(
        'regex'  => '^s_' . $self->{'testname'},
        'mapped' => \%mapped
    );

    $self->wait_for_replay();

    $self->check_share_propagation('share' => $last_share);

    $Log->debug(Dumper($self));

    $Log->exit();

} ## end sub filer_setup

sub filer_cleanup {

    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::filer_cleanup(%options);

    $Options_hash = $Options->get_hash();

    if (!$Options_hash->{'POST_FAILURE_CLEANUP'}) {
        $Log->comment(
            "Skipping filer_cleanup STAGE because POST_FAILURE_CLEANUP=0");
        return;
    }

    my @targets = ('antivirus', 'lifs', 'volumes', 'vservers', 'routes');
    $self->discover_topology('targets' => \@targets);

    $self->filer_cleanup_defined();

    $Log->exit();
} ## end sub filer_cleanup

sub client_setup {

    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::client_setup(%options);

    #Check for ets/hosts Clean
    $Options_hash = $Options->get_hash();

    if ($Options_hash->{'CLIENT_ETC_HOSTS_CLEAN_UP'}) {
        my ($error, $client_conn_obj, $cmd, $output, $client_name,
            $os_version, %fields);

        host($Options_hash->{'CLIENTNAME'});

        $client_name = $Options_hash->{'CLIENTNAME'};

        %fields = host($client_name)->param_fields();

        $Log->comment("Cleaning etc\\hosts file on $client_name");

        #-------------------------------
        # Connect to the client
        #-------------------------------
        $client_conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
            '-nologin', -timeout, 300);

        #-------------------------------
        # Verify connectivity to the client
        #-------------------------------
        eval { check_connectivity('conn' => $client_conn_obj); };
        $error = $@;
        if ($error) { die config_result($error); }

        # Clean etc\hosts file
        $cmd = "cmd.exe /c echo 127.0.0.1 localhost \> "
            . "\%SYSTEMROOT\%\\system32\\drivers\\etc\\hosts";
        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }

        $cmd = "cmd.exe /c echo ::1 localhost \>\> "
            . "\%SYSTEMROOT\%\\system32\\drivers\\etc\\hosts";
        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }

        # Verify etc\hosts file is clean
        $cmd
            = "cmd.exe /c type \%SYSTEMROOT\%\\system32\\drivers\\etc\\hosts";
        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }

        $Log->comment("Client etc\\hosts file on $client_name:\n $output");
    }

    $Log->exit();
} ## end sub client_setup

sub DESTROY {

    $Log->enter();

    my $self = shift;

    local ($@);

    $self->client_cleanup();

    $Log->exit();
} ## end sub DESTROY

sub client_cleanup {

    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::client_cleanup(%options);

    $Log->exit();
} ## end sub client_cleanup

sub discover_topology {

    $Log->enter();

    my ($self, %options) = @_;

    # Discover all existing vservers, volumes, aggregates, interfaces, cifs
    my @targets = @{$options{'targets'}};

    if (grep { $_ eq 'licenses' } @targets) {
        $Log->comment("Discovering licenses");
        delete($self->{'discovered_topology'}->{'licenses'});
        my @licenses = NACL::CS::License->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $license (@licenses) {
            $license->command_interface_reset();
            my %properties = %$license;
            $self->{'discovered_topology'}->{'licenses'}
                ->{$license->feature()} = \%properties;
        }
    }

    if (grep { $_ eq 'system' } @targets) {
        if ($self->{'release'} =~ /boilermaker/) {

            my @dns = NACL::CS::SystemServicesDns->fetch(
                command_interface => $ci,
                'allow_empty'     => 1
            );

            foreach my $row (@dns) {
                $row->command_interface_reset();
                my %properties = %$row;
                $self->{'discovered_topology'}->{'system'}->{$row->{'node'}}
                    ->{'dns'} = \%properties;
            }
        }
    }

    if (grep { $_ eq 'vservers' } @targets) {
        $Log->comment("Discovering vservers");
        delete($self->{'discovered_topology'}->{'vservers'});

        my @vservers = NACL::CS::Vserver->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $vserver (@vservers) {
            $vserver->command_interface_reset();
            my %properties = %$vserver;
            $self->{'discovered_topology'}->{'vservers'}
                ->{$vserver->vserver()} = \%properties;
        }

        my (@dns, @nis);
        if ($self->{'release'} !~ /boilermaker/) {
            @dns = NACL::CS::VserverServicesDns->fetch(
                command_interface => $ci,
                'allow_empty'     => 1
            );

            foreach my $row (@dns) {
                $row->command_interface_reset();
                my %properties = %$row;
                $self->{'discovered_topology'}->{'vservers'}
                    ->{$row->{'vserver'}}->{'dns'} = \%properties;
            }

            @nis = NACL::CS::VserverServicesNisDomain->fetch(
                command_interface => $ci,
                'allow_empty'     => 1
            );

            foreach my $row (@nis) {
                $row->command_interface_reset();
                my %properties = %$row;
                $self->{'discovered_topology'}->{'vservers'}
                    ->{$row->{'vserver'}}->{'nis'} = \%properties;
            }
        }

        my @users = NACL::CS::UnixUser->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $u (@users) {
            $u->command_interface_reset();
            my %properties = %$u;
            $self->{'discovered_topology'}->{'vservers'}->{$u->vserver()}
                ->{'unix_users'}->{$u->user()} = \%properties;
        }

        my @groups = NACL::CS::UnixGroup->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );
        foreach my $g (@groups) {
            $g->command_interface_reset();
            my %properties = %$g;
            $self->{'discovered_topology'}->{'vservers'}->{$g->vserver()}
                ->{'unix_groups'}->{$g->name()} = \%properties;
        }
    }

    if (grep { $_ eq 'lifs' } @targets) {
        $Log->comment("Discovering lifs");
        delete($self->{'discovered_topology'}->{'lifs'});
        my @lifs = NACL::CS::Interface->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $lif (@lifs) {
            $lif->command_interface_reset();
            my %properties = %$lif;
            $self->{'discovered_topology'}->{'lifs'}->{$lif->server()}
                ->{$lif->lif()} = \%properties;
        }
    }

    if (grep { $_ eq 'ports' } @targets) {
        $Log->comment("Discovering ports");
        delete($self->{'discovered_topology'}->{'ports'});
        my @ports = NACL::CS::NetworkPort->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $port (@ports) {
            $port->command_interface_reset();
            my %properties = %$port;
            $self->{'discovered_topology'}->{'ports'}->{$port->node()}
                ->{$port->port()} = \%properties;
        }

    }

    if (grep { $_ eq 'aggregates' } @targets) {
        $Log->comment("Discovering aggregates");
        delete($self->{'discovered_topology'}->{'aggregates'});
        my @aggregates = NACL::CS::Aggregate->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $aggregate (@aggregates) {
            $aggregate->command_interface_reset();
            my %properties = %$aggregate;
            $self->{'discovered_topology'}->{'aggregates'}
                ->{$aggregate->aggregate()} = \%properties;
        }
    }

    if (grep { $_ eq 'volumes' } @targets) {
        $Log->comment("Discovering volumes");
        delete($self->{'discovered_topology'}->{'volumes'});
        my @volumes = NACL::CS::Volume->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $volume (@volumes) {
            $volume->command_interface_reset();
            my %properties = %$volume;
            $self->{'discovered_topology'}->{'volumes'}->{$volume->volume()}
                = \%properties;
        }
    }

    if (grep { $_ eq 'qtrees' } @targets) {
        $Log->comment("Discovering qtrees");
        delete($self->{'discovered_topology'}->{'qtrees'});
        my @qtrees = NACL::CS::Qtree->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $qtree (@qtrees) {
            $qtree->command_interface_reset();
            my %properties = %$qtree;
            $self->{'discovered_topology'}->{'qtrees'}->{$qtree->qtree()}
                = \%properties;
        }
    }

    if ((grep { $_ eq 'quotas' } @targets)
        && $self->{'release'} !~ /boilermaker/) {
        $Log->comment("Discovering quota policies");
        delete($self->{'discovered_topology'}->{'quotas'});
        my @QuotaPolicies;
        eval {
            @QuotaPolicies = NACL::CS::QuotaPolicy->fetch(
                command_interface => $ci,
                allow_empty       => 1
            );
        };

        foreach my $qp (@QuotaPolicies) {
            $qp->command_interface_reset();
            my %properties = %$qp;
            $self->{'discovered_topology'}->{'quotas'}->{$qp->vserver()}
                ->{$qp->policy_name()} = \%properties;
        }

        $Log->comment("Discovering quota policy rules");
        my @QuotaPolicyRules;
        eval {
            @QuotaPolicyRules = NACL::CS::QuotaPolicyRule->fetch(
                command_interface => $ci,
                allow_empty       => 1
            );
        };

        foreach my $qpr (@QuotaPolicyRules) {
            $qpr->command_interface_reset();
            my %properties = %$qpr;
            push(
                @{  $self->{'discovered_topology'}->{'quotas'}
                        ->{$qpr->vserver()}->{$qpr->policy_name()}->{'rules'}
                    },
                \%properties
            );
        }
    }

    if (grep { $_ eq 'routes' } @targets) {
        if ($options{'vserver'}) {
            delete($self->{'discovered_topology'}->{'routing_groups'}
                    ->{$options{'vserver'}});
        } else {
            delete($self->{'discovered_topology'}->{'routing_groups'});
        }

        my %args = (command_interface => $ci, allow_empty => 1);
        if ($options{'vserver'}) {
            $args{'filter'} = {'vserver' => $options{'vserver'}};
        }

        my @routing_groups = NACL::CS::RoutingGroups->fetch(%args);
        foreach my $routing_group (@routing_groups) {
            $routing_group->command_interface_reset();
            my %properties = %$routing_group;
            $self->{'discovered_topology'}->{'routing_groups'}
                ->{$routing_group->server()}
                ->{$routing_group->routing_group()} = \%properties;
        }

        my @routes = NACL::CS::Route->fetch(%args);
        foreach my $route (@routes) {
            $route->command_interface_reset();
            my %properties = %$route;
            $self->{'discovered_topology'}->{'routing_groups'}
                ->{$route->server()}->{$route->routing_group()}->{'routes'}
                ->{$route->destination()} = \%properties;
        }
    }

    if (grep { $_ eq 'ntp_servers' } @targets) {
        $Log->comment("Discovering NTP servers");
        delete($self->{'discovered_topology'}->{'ntp_servers'});
        my @Servers = NACL::CS::SystemServicesNtpServer->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $s (@Servers) {
            $s->command_interface_reset();
            my %properties = %$s;
            $self->{'discovered_topology'}->{'ntp_servers'}->{$s->node()}
                ->{$s->server()} = \%properties;
        }

    }

    if ((grep { $_ eq 'antivirus' } @targets)
        && $self->{'release'} !~ /boilermaker/) {
        $Log->comment("Discovering Antivirus configuration");
        delete($self->{'discovered_topology'}->{'antivirus'});
        my @Antivirus = NACL::CS::Antivirus->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $av (@Antivirus) {
            $av->command_interface_reset();
            my %properties = %$av;
            $self->{'discovered_topology'}->{'antivirus'} = \%properties;
        }

    }

    if (grep { $_ eq 'cifs' } @targets) {
        $Log->comment("Discovering CIFS");
        my @vservers
            = natsort(keys(%{$self->{'discovered_topology'}->{'vservers'}}));
        foreach my $vserver (@vservers) {
            delete($self->{'discovered_topology'}->{'vservers'}->{$vserver}
                    ->{'cifs'});
        }
        my @cifs = NACL::CS::VserverCifs->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );

        foreach my $c (@cifs) {
            $c->command_interface_reset();
            my %properties = %$c;
            $self->{'discovered_topology'}->{'vservers'}->{$c->vserver()}
                ->{'cifs'} = \%properties;
        }

        $Log->comment("Discovering Preferred DCs");
        foreach my $vserver (@vservers) {
            delete($self->{'discovered_topology'}->{'vservers'}->{$vserver}
                    ->{'cifs'}->{'preferred-dc'});
        }
        my @domain_controllers
            = NACL::CS::VserverCifsDomainPreferredDc->fetch(
            command_interface => $ci,
            allow_empty       => 1
            );

        foreach my $d (@domain_controllers) {
            $d->command_interface_reset();
            my %properties = %$d;
            $self->{'discovered_topology'}->{'vservers'}->{$d->vserver()}
                ->{'cifs'}->{'preferred-dcs'}->{$d->domain()} = \%properties;
        }

    }

    if (grep { $_ eq 'shares' } @targets) {
        $Log->comment("Discovering CIFS shares");
        delete($self->{'discovered_topology'}->{'shares'});
        my @shares = NACL::CS::VserverCifsShare->fetch(
            command_interface => $ci,
            allow_empty       => 1
        );
        foreach my $s (@shares) {
            $s->command_interface_reset();
            my %properties = %$s;
            if ($s->share_name() =~ /^admin\$|^ipc\$/) { next; }
            $self->{'discovered_topology'}->{'shares'}->{$s->share_name()}
                = \%properties;
        }
    }

    $Log->exit();
} ## end sub discover_topology

sub assign_io_clients {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Assigning I/O client test options");

    my $share_index  = $options{'share_index'};
    my $volume_count = 1;
    my $client_count = 1;
    my @volumes;
    my $lif_index       = 0;
    my $io_config_index = 0;
    my $node_index      = 0;
    my $io_config;
    my $defined_vservers = $self->{'defined_topology'}->{'vservers'};
    my $defined_shares   = $topology_database->{'shares'};
    my $defined_volumes  = $self->{'defined_topology'}->{'volumes'};
    my $defined_lifs     = $self->{'defined_topology'}->{'lifs'};
    my $discovered_aggregates
        = $self->{'discovered_topology'}->{'aggregates'};
    my @node_regex = split(/,/, $Options_hash->{'NODE_REGEX'});
    my @aggr_regex = split(/,/, $Options_hash->{'AGGR_REGEX'});
    my @shares_per_volume;
    my $share_count = 1;
    my $volume;
    my $volume_index = 1;

    if ($Options_hash->{'SHARES_PER_VOLUME'}) {
        @shares_per_volume = split(/,/, $Options_hash->{'SHARES_PER_VOLUME'});
    }

    my @io_configs
        = natsort(keys(%{$self->{'defined_topology'}->{'io_configs'}}));

    foreach my $i (@io_configs) {

        # Create a pointer to the current loads config
        $io_config = $self->{'defined_topology'}->{'io_configs'}->{$i};

        # Get the clients used for this load
        my $clients = 0;
        if ($io_config->{'max_clients'}) {
            $clients = $io_config->{'max_clients'};
        } else {
            $clients = @{$self->{'clients'}};
            $io_config->{'max_clients'} = $clients;
        }

        # Get the max number of shares each client can possibly use
        my $shares = $io_config->{'max_shares_per_client'} * $clients;
        $self->{'defined_topology'}->{'io_configs'}->{$i}->{'total_shares'}
            = $shares;

        # Get the target nodes for this load
        my @nodes = grep(/$node_regex[$i]/, @{$Options_hash->{'NODES'}});

        # Die if we can't match any nodes
        if (!@nodes) {
            die config_result(
                "No nodes matched by NODE_REGEX $node_regex[$i]");
        }

        # Store the target nodes
        $io_config->{'target_nodes'} = \@nodes;

        # Get the target aggregates for this load
        my @aggregates = grep(/$aggr_regex[$i]/,
            split(/,/, $Options_hash->{'AGGREGATES'}));

        # Die if we can't match any aggrs
        if (!@aggregates) {
            die config_result(
                "No aggregates in AGGREGATES $Options_hash->{'AGGREGATES'} matched by AGGR_REGEX $aggr_regex[$i]"
            );
        }

        # Store the target aggregates
        $io_config->{'target_aggregates'} = \@aggregates;

        # Get all volumes who's aggregate matches the target aggregates
        my $volumes = 0;
        foreach my $volume (natsort(keys(%{$defined_volumes}))) {
            my $aggregate = $defined_volumes->{$volume}->{'aggregate'};
            if (grep { $_ eq $aggregate } @aggregates) {
                if (   $defined_volumes->{$volume}->{'rootvolume'} == 0
                    && $defined_volumes->{$volume}->{'type'} eq 'RW') {
                    $volumes++;
                    push(@{$io_config->{'target_volumes'}}, $volume);
                }

            }
        }
        $io_config->{'target_volume_count'} = $volumes;

        if (!@{$io_config->{'target_volumes'}}) {
            die script_result("No target volumes defined for this node\n"
                    . Dumper($io_config->{'target_volumes'}));
        }

    }

    if (!@shares_per_volume) {
        my $shares_per_volume = 0;
        my $required_shares   = 0;
        my $target_volumes    = 0;
        my @unique_volumes;

        foreach my $i (@io_configs) {

            $io_config = $self->{'defined_topology'}->{'io_configs'}->{$i};
            $required_shares += $io_config->{'total_shares'};
            foreach my $volume (@{$io_config->{'target_volumes'}}) {
                if (!grep { $_ eq $volume } @unique_volumes) {
                    push(@unique_volumes, $volume);
                }
            }
        }

        $target_volumes = @unique_volumes;

        if ($required_shares > $target_volumes) {
            $shares_per_volume = int($required_shares / $target_volumes);
            if ($required_shares % $target_volumes) {
                $shares_per_volume++;
                $Log->comment("Setting the number shares per volume to "
                        . "$shares_per_volume "
                        . "because the required number of shares ("
                        . $required_shares
                        . ")\ndoes not fit on all volumes evenly");
            } else {
                $Log->comment('Setting the number of shares per volume to '
                        . $shares_per_volume
                        . ' because the required '
                        . 'number of shares ('
                        . $required_shares
                        . ")\ndoes not fit on all target volumes ("
                        . $target_volumes
                        . ')');
            }
        } else {
            $shares_per_volume = 1;
        }

        foreach my $i (@io_configs) {

            # Create a pointer to the current loads config
            $io_config = $self->{'defined_topology'}->{'io_configs'}->{$i};

            # Setting shares per volume for the current config
            $io_config->{'shares_per_volume'} = $shares_per_volume;
        }
    } else {
        foreach my $i (@io_configs) {

            # Create a pointer to the current loads config
            $io_config = $self->{'defined_topology'}->{'io_configs'}->{$i};

            # Setting shares per volume for the current config
            $io_config->{'shares_per_volume'}
                = $Options_hash->{'SHARES_PER_VOLUME'};
        }

    }

    # Loop through each client and assign the correct info to it. Since the
    # tools themselves define the share needs we create share names and
    # allocate them to clients here.
    my $client_index = 1;
    for my $client (@{$self->{'clients'}}) {

        $Log->comment("Client: $client");
        $Log->comment("Client Index: $client_index");
        $client_index++;

        # Set the client state to reserved
        $self->{'clients_status'}->{$client}->{'status'} = 'reserved';

        # Get the I/O configuration
        $io_config
            = $self->{'defined_topology'}->{'io_configs'}->{$io_config_index};
        $Log->comment("Load Index: $io_config_index");

        # Populate a list of volumes
        if (!@volumes) {
            if (!@{$io_config->{'target_volumes'}}) {
                die script_result("Unable to get target volumes from "
                        . "\$self->{'defined_topology'}->{'io_configs'}"
                        . "->{$io_config_index}->{'target_volumes'}\n"
                        . Dumper($io_config->{'target_volumes'}));
            } else {
                @volumes = natsort(@{$io_config->{'target_volumes'}});
            }
        }

        if (host($client)->hosttype()
            !~ /$io_config->{'CLIENT_HOSTTYPE_REGEX'}/) {
            next;
        }

     # Added instance loop for future multiple client instances right now only
     # one instance is supported
        for (
            my $instance = 0;
            $instance < $Options_hash->{'INSTANCES_PER_CLIENT'};
            $instance++
            ) {

            $Log->comment("Instance: $instance");

            # Set current client to assigned so supervisor2startnext knows to
            # start it
            $self->{'clients_status'}->{$client}->{'status'} = 'assigned';

            # Configure how we connect to the client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'CLIENT_CONNECT_METHOD'}
                = $Options_hash->{'CLIENT_CONNECT_METHOD'};

            # Configure the desired hosttype for all clients in case this
            # script uses multiple client types. This is here for historical
            # purposes
            $self->{'test_options'}->{$client}->{$instance}
                ->{'CLIENT_HOSTTYPE_REGEX'}
                = $io_config->{'CLIENT_HOSTTYPE_REGEX'};

            # Set the length of time test_main executes on the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'TEST_MAIN_DURATION'} = $io_config->{'TEST_MAIN_DURATION'};

            # Set the length of time test_main sleeps on the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'TEST_MAIN_SLEEP'} = $io_config->{'TEST_MAIN_SLEEP'};

            # Provide the correct path the the tools config file
            if (host($client)->hosttype() =~ /win/) {
                $self->{'test_options'}->{$client}->{$instance}
                    ->{'IO_TOOLS_CONFIG'}
                    = "$Options_hash->{'TOP_LOGDIR_WINDOWS'}\\$io_config->{'io_tools_config'}";
            } else {
                $self->{'test_options'}->{$client}->{$instance}
                    ->{'IO_TOOLS_CONFIG'}
                    = "$Options_hash->{'TOP_LOGDIR'}/$io_config->{'io_tools_config'}";
            }

            # Set the tools regex for the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'IO_TOOLS_REGEX'} = $io_config->{'IO_TOOLS_REGEX'};

            # Set the tools overrides for the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'IO_TOOLS_OVERRIDES'} = $io_config->{'IO_TOOLS_OVERRIDES'};

            # Set the flag whether to map shares for the current client
            $self->{'test_options'}->{$client}->{$instance}->{'IO_MAP_SHARES'}
                = $io_config->{'IO_MAP_SHARES'};

            # Set the  wether to map shares for the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'IO_REMOVE_DATA'} = $io_config->{'IO_REMOVE_DATA'};

            # Determine the max shares per client for this load
            my $max_occurrences_per_client
                = $io_config->{'max_occurrences_per_client'};
            my $max_shares_per_occurrence
                = $io_config->{'max_shares_per_occurrence'};

            # Loop though the max number of shares per client for this load
            # This is where the share info is populated in defined_topology
            # and linked to volumes and vservers
            for (my $o = 0; $o < $max_occurrences_per_client; $o++) {

                $Log->comment("Occurence: $o");

                for (my $s = 1; $s <= $max_shares_per_occurrence; $s++) {

                    # Get the next volume if $volume is blank
                    if (!$volume) {
                        $volume = shift(@volumes);
                    }
                    if ($volume_index
                        > scalar(@{$io_config->{'target_volumes'}})) {
                        die script_result("
                        No Volumes left to add shares!! Increase"
                                . " SHARES_PER_VOLUME.");
                    }

                    $Log->comment("Volume: $volume");

                    # Get the vserver name from the volume entry
                    my $vserver = $defined_volumes->{$volume}->{'vserver'};
                    $Log->comment("Vserver: $vserver");

                    # Get the current node name
                    if ($node_index >= @{$io_config->{'target_nodes'}}) {
                        $Log->debug("Resetting node index to 0");
                        $node_index = 0;
                    }

                    my $node = ${$io_config->{'target_nodes'}}[$node_index];
                    if (!$node) {
                        die script_result("Unable to get the node from "
                                . "index $node_index\n"
                                . Dumper($io_config));
                    }

                    $Log->comment("Node: $node");

                    # Get the lif from the vserver entry
                    my $lif;
                    my $lif_owner;
                    my $node_found;
                    foreach my $element (
                        @{$defined_vservers->{$vserver}->{'lifs'}}) {
                        if ($defined_lifs->{$vserver}->{$element}->{'node'}
                            =~ $node) {
                            $lif        = $element;
                            $lif_owner  = $node;
                            $node_found = 1;
                        }
                    }
                    if (!$node_found) {
                        die script_result("Node:$node not available on "
                                . "Vserver LIF\n"
                                . Dumper($io_config));
                    }

                    $Log->comment("Logical interface: $lif");

                    # Get the ip address from the defined lifs
                    my $ip = $defined_lifs->{$vserver}->{$lif}->{'address'};
                    $Log->comment("IP address: $ip");

                    my $mode = $io_config->{'CIFS_SERVER_ACCESS_MODE'};
                    if (!$mode) {
                        $mode = $Options_hash->{'CIFS_SERVER_ACCESS_MODE'};
                    }

                    my $server = $self->get_cifs_server_name(
                        'ip'   => $ip,
                        'mode' => $mode
                    );
                    $Log->comment("CIFS Server: $server");

                    # Defined the share name
                    my $share
                        = 's_' . $options{'script'} . '_' . $share_index;
                    $Log->comment("Share: $share");

                    # Now that we have the name increment the share counter
                    $share_index++;

                    push(
                        @{  $self->{'defined_topology'}->{'vservers'}
                                ->{$vserver}->{'shares'}
                            },
                        $share
                    );

                    # Store the share path with the share
                    $topology_database->{'shares'}->{$share}->{'path'}
                        = "/$volume/$share";

                    # Store the volume with the share
                    $topology_database->{'shares'}->{$share}->{'volume'}
                        = $volume;

                    # Store the owning vserver with the share
                    $topology_database->{'shares'}->{$share}->{'vserver'}
                        = $vserver;

                    # Configure the filer_setup flag that determines if the
                    # share is created in filer_setup or some other way
                    $topology_database->{'shares'}->{$share}->{'filer_setup'}
                        = 1;

                    if ($defined_volumes->{$volume}->{'style'} eq 'flex') {
                        $topology_database->{'shares'}->{$share}->{'qtree'}
                            = $share;
                    }

                    $self->update_test_options(
                        'client'   => $client,
                        'instance' => $instance,
                        'node'     => $node,
                        'vserver'  => $vserver,
                        'server'   => $server,
                        'volume'   => $volume,
                        'share'    => $share,
                        'destip'   => $ip
                    );

                   # If we have reached the shares per volume threshold switch
                   # to the next node index
                    if ($share_count < $io_config->{'shares_per_volume'}) {
                        $share_count++;
                    } else {
                        $Log->debug(
                            "Clearing \$volume so another volume is retrieved"
                        );
                        $volume      = undef;
                        $share_count = 1;
                        $Log->debug("Resetting the share count");
                        $node_index++;
                        $volume_index++;
                    }
                }
            }
        }

        # Increment the client count and switch to the next load if needed
        $client_count++;
        if ($client_count > $io_config->{'max_clients'}) {
            $client_count = 1;
            if (   $Options_hash->{'IO_LOAD_CONFIG'}
                || $Options_hash->{'IO_TOOLS_CONFIG'}) {
                if ($io_config_index == $#io_configs) {

                    # No more loads so bail
                    last;
                } else {

                    # Switch to the next load
                    $volume = undef;
                    $io_config_index++;
                    $volume_index = 1;
                    $Log->debug(
                        "Incrementing the load index to $io_config_index");
                    if ($node_index >= @{$io_config->{'target_nodes'}}) {
                        $Log->debug("Resetting node index to 0");
                        $node_index = 0;
                    }
                }
            }
        }

    }

    $Log->enter();

    return $share_index;
} ## end sub assign_io_clients

sub assign_share_verification_clients {

    $Log->enter();

    $Log->comment("Assigning shares client test options");

    my ($self, %options) = @_;
    my $instance    = 0;
    my $share_index = $options{'share_index'};
    my %nodes;
    my $defined_shares   = $topology_database->{'shares'};
    my $defined_volumes  = $self->{'defined_topology'}->{'volumes'};
    my $defined_lifs     = $self->{'defined_topology'}->{'lifs'};
    my $defined_vservers = $self->{'defined_topology'}->{'vservers'};
    my @shares;

    my @filtered_nodes;

    if ($Options_hash->{'NODE_REGEX'}) {
        my $regex = $Options_hash->{'NODE_REGEX'};
        @filtered_nodes = grep(/$regex/, @{$Options_hash->{'NODES'}});
        if (!@filtered_nodes) {
            die config_result("No nodes found by NODE_REGEX $regex");
        }
    } else {
        @filtered_nodes = @{$Options_hash->{'NODES'}};
    }

    # Loop through each client and assign the correct info to it.
    my @addresses;
    my $client_count = 0;
    for my $client (@{$self->{'clients'}}) {

        if (host($client)->hosttype()
            !~ /$Options_hash->{'CLIENT_HOSTTYPE_REGEX'}/) {
            next;
        }

        $client_count++;
        $Log->comment("Client: $client");
        $Log->comment("Instance: $instance");

        $self->{'clients_status'}->{$client}->{'status'} = 'assigned';
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_CONNECT_METHOD'}
            = $Options_hash->{'CLIENT_CONNECT_METHOD'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_HOSTTYPE_REGEX'}
            = $Options_hash->{'CLIENT_HOSTTYPE_REGEX'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'TEST_MAIN_DURATION'} = $Options_hash->{'TEST_MAIN_DURATION'};
        $self->{'test_options'}->{$client}->{$instance}->{'TEST_MAIN_SLEEP'}
            = $Options_hash->{'TEST_MAIN_SLEEP'};

        my $shares_per_client = 0;
        if (   $client eq ${$self->{'clients'}}[0]
            && $self->{'shares_per_client_remainder'}) {
            $shares_per_client = $self->{'shares_per_client'}
                + $self->{'shares_per_client_remainder'};
        } else {
            $shares_per_client = $self->{'shares_per_client'};
        }
        $Log->comment("Shares: $shares_per_client");

        if (!@shares) {
            foreach my $vserver (natsort(keys(%{$defined_vservers}))) {
                foreach my $share (
                    natsort(@{$defined_vservers->{$vserver}->{'shares'}})) {
                    my $volume = $defined_shares->{$share}->{'volume'};
                    my $path   = $defined_shares->{$share}->{'path'};
                    my $root_volume
                        = $defined_vservers->{$vserver}->{'rootvolume'};
                    $Log->comment(
                        "\$share:$share \$volume:$volume \$root_volume:$root_volume \$path:$path "
                            . Dumper($defined_shares->{$share}));

                    # Dont use the vservers root share and any volume shares
                    if ($root_volume eq $volume) {
                        next;
                    } elsif ($path eq "/$volume") {
                        next;
                    }
                    push(@shares, $share);
                }
            }
        }

        my $starting_share_index;
        foreach my $share_count (1 ... $shares_per_client) {

            my ($next_node, $next_vserver, $next_volume, $next_share,
                $share_basename, $share);

            # Get the next share
            if (@shares) {
                $share = shift(@shares);
            } else {
                last;
            }

            $Log->comment("Share: $share");
            if ($share =~ /^(.+)_(\d+)$/) {
                $share_basename = $1;
                $share_index    = $2;
            }

            if (!$starting_share_index) {
                $starting_share_index = $share_index;
            }

            my $vserver = $defined_shares->{$share}->{'vserver'};
            if (!$vserver) {
                die script_result("unable to get the owning "
                        . "vserver from share $share\n"
                        . Dumper($defined_shares->{$share}));
            }

            # Get the addresses from the desired nodes
            if (!@addresses) {

                foreach my $lif (@{$defined_vservers->{$vserver}->{'lifs'}}) {
                    my $node = $defined_lifs->{$vserver}->{$lif}->{'node'};
                    if (grep { $_ eq $node } @filtered_nodes) {
                        my $address
                            = $defined_lifs->{$vserver}->{$lif}->{'address'};
                        $nodes{$address} = $node;
                        push(@addresses, $address);
                    }
                }
            }

            # Get the next address
            my $ip = shift(@addresses);
            if (!$ip) {
                die script_result("Unable to get an IP address from "
                        . "defined vserver $vserver\n"
                        . Dumper($defined_vservers->{$vserver})
                        . "\%nodes\n"
                        . Dumper(\%nodes)
                        . "\@filtered_nodes\n"
                        . Dumper(\@filtered_nodes));
            }

            # Get the node name from the address
            my $node = $nodes{$ip};

            # Convert to a hostname or FQDN if required
            my $mode = $Options_hash->{'CIFS_SERVER_ACCESS_MODE'};
            my $server
                = $self->get_cifs_server_name('ip' => $ip, 'mode' => $mode);

            # Get the volume name from the share
            my $volume = $defined_shares->{$share}->{'volume'};

            if (@shares) {
                $next_share   = $shares[0];
                $next_node    = $nodes{$next_share};
                $next_vserver = $defined_shares->{$next_share}->{'vserver'};
                $next_volume  = $defined_shares->{$next_share}->{'volume'};
            }

            # If any critical var has changed for the next share then we push
            # the currrent set of test options.
            if (   ($share_count == $shares_per_client)
                || $next_vserver ne $vserver
                || $next_volume  ne $volume) {

                my $share_info;
                if ($shares_per_client == 1) {
                    $share_info = $share;
                } else {
                    $share_info
                        = $share_basename . '_;'
                        . $starting_share_index . '-'
                        . $share_index;
                }

                $Log->comment("Updating test options");
                $self->update_test_options(
                    'client'   => $client,
                    'instance' => $instance,
                    'vserver'  => $vserver,
                    'server'   => $server,
                    'volume'   => $volume,
                    'node'     => $node,
                    'share'    => $share_info
                );

                # Reset the starting share index
                $starting_share_index = 0;

                # Clear the addresses so the we uses addesses for the next
                # vserver
                if ($next_vserver && $next_vserver ne $vserver) {
                    @addresses = ();
                }
            }
        }

        $self->update_test_options('process' => 'mapstress');

        if ($client_count == $self->{'max_clients'}) {
            last;
        }
    }

    $Log->exit();

    return $share_index++;
} ## end sub assign_share_verification_clients

sub get_object_properties {

    $Log->enter();

    my ($self, $obj) = @_;

    my $state = $obj->state();

    my %properties = %$state;

    $Log->exit();

    return \%properties;
} ## end sub get_object_properties

sub assign_volumes_to_aggregates {

    $Log->enter();

    my ($self, %options) = @_;
    my $count  = $options{'count'};
    my $index  = $options{'index'};
    my $script = $options{'script'};
    my $size   = $options{'size'};
    my $type   = $options{'type'};

    # Get an array of the discovered aggregates
    my $discovered_aggregates
        = $self->{'discovered_topology'}->{'aggregates'};

    # Get the target aggregates
    my @aggregates = split(/,/, $options{'aggregates'});
    my @aggregate_full;
    foreach my $i (0 ... $#aggregates) { $aggregate_full[$i] = 0; }

    my $aggregate;
    my $volume_index;
    my $aggregate_index = 0;

    # Loop though each volume assigning it to an  aggregate
    for (
        $volume_index = $index;
        $volume_index < ($index + $count);
        $volume_index++
        ) {

        my $aggregate_found = 0;
        while (!$aggregate_found) {
            $aggregate = $aggregates[$aggregate_index];

            # Get the current aggregates size
            my $aggregate_size
                = int($discovered_aggregates->{$aggregate}->{'availsize'}
                    * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));

            # Convert to MB
            $aggregate_size = int($aggregate_size / 1024);

            if (!$self->{'volume_usage'}->{$aggregate}) {
                $self->{'volume_usage'}->{$aggregate} = 0;
            }

            # Check to make sure we don't exceed the current aggregates space.
            if ($self->{'volume_usage'}->{$aggregate} + $size
                > $aggregate_size) {

                # Flag aggregate as full
                $aggregate_full[$aggregate_index] = 1;

                # Die if all aggregates are full
                if (!grep { $_ == 0 } @aggregate_full) {
                    die config_result(
                        "Not enough free aggregate space to create all voumes"
                    );
                }

            } else {

                # We found an avaiable aggregate
                $aggregate_found = 1;
            }

            # Decide if we need to switch aggregates
            my $switch_aggregate = 0;
            if (!$aggregate_found) {
                $switch_aggregate = 1;
            } elsif ($Options_hash->{'VOLUME_ASSIGNMENT'} eq 'round_robin') {
                $switch_aggregate = 1;
            }

            if ($switch_aggregate) {

                # Switch to the next aggregate in the list
                if ($aggregate_index == $#aggregates) {
                    $aggregate_index = 0;
                } else {
                    $aggregate_index++;
                }

            }
        }

        my $style = 'flex';
        if (@{$discovered_aggregates->{$aggregate}->{'nodes'}} > 1) {
            $style = 'striped';
        }

        # Create the volume name
        my $volume
            = 'v_' 
            . $script . '_'
            . $volume_index . '_'
            . $style . '_'
            . lc($type);
        $Log->comment("Adding volume $volume to aggregate $aggregate");

        # Populate the volume entries in the defined_topology hash
        $self->{'defined_topology'}->{'volumes'}->{$volume}->{'style'}
            = $style;
        $self->{'defined_topology'}->{'volumes'}->{$volume}->{'type'} = $type;
        $self->{'defined_topology'}->{'volumes'}->{$volume}->{'size'} = $size;
        $self->{'defined_topology'}->{'volumes'}->{$volume}->{'aggregate'}
            = $aggregate;

        $self->{'volume_usage'}->{$aggregate} += $size;

    }

    return $volume_index;

    $Log->exit();
} ## end sub assign_volumes_to_aggregates

sub define_shares {

    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    my $script      = $options{'script'};
    my $share_index = $options{'share_index'};
    my $recipe      = $options{'recipe'};
    my %shares;
    my $filer_setup;
    my $defined_vservers    = $self->{'defined_topology'}->{'vservers'};
    my $defined_volumes     = $self->{'defined_topology'}->{'volumes'};
    my $defined_shares      = $topology_database->{'shares'};
    my $defined_lifs        = $self->{'defined_topology'}->{'lifs'};
    my $discovered_vservers = $self->{'discovered_topology'}->{'vservers'};
    my $discovered_volumes  = $self->{'discovered_topology'}->{'volumes'};
    my $discovered_shares   = $self->{'discovered_topology'}->{'shares'};
    my $discovered_aggregates
        = $self->{'discovered_topology'}->{'aggregates'};

    # Count the existing shares
    $Log->comment('Counting existing shares');
    foreach my $vserver (natsort(keys(%{$discovered_vservers}))) {
        if ($discovered_vservers->{$vserver}->{'type'} ne 'cluster') {
            next;
        }

        # Account for hidden shares
        $shares{$vserver} = @ontap_shares;

        # Count the current vservers shares
        foreach my $s (natsort(keys(%{$discovered_shares}))) {

            # ignore shares created by this test previously
            if ($s =~ /^s_$script/) { next; }
            if ($discovered_shares->{$s}->{'vserver'} eq $vserver) {
                $shares{$vserver} += 1;
            }
        }

    }

    # Count the defined shares if they don't already exist;
    my @ds = keys(%{$discovered_shares});
    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {

        # Account for hidden shares
        $shares{$vserver} += @ontap_shares;

        foreach my $s (natsort(@{$defined_vservers->{$vserver}->{'shares'}}))
        {
            if (!grep { $_ eq $s } @ds) {
                $shares{$vserver} += 1;
            }
        }
    }

    # Total all the shares found
    my $total_existing_shares       = 0;
    my $max_existing_vserver_shares = 0;
    my $max_existing_vserver_shares_owner;
    foreach my $vs (keys(%shares)) {
        $total_existing_shares += $shares{$vs};
        if ($shares{$vs} > $max_existing_vserver_shares) {
            $max_existing_vserver_shares       = $shares{$vs};
            $max_existing_vserver_shares_owner = $vs;
        }
    }

    if ($total_existing_shares >= $Options_hash->{'SHARE_COUNT'}) {
        die config_result("Existing shares is greater or equal to "
                . "SHARES($Options_hash->{'SHARE_COUNT'})");
    }

    $Log->comment('Determining new shares to create');
    my $new_shares = $Options_hash->{'SHARE_COUNT'} - $total_existing_shares;

    $Log->comment('Spreading new shares across vservers');
    my $shares_per_vserver
        = int($new_shares / ($Options_hash->{'VSERVER_COUNT'}));
    if ($shares_per_vserver < 1) {
        $Log->warn(
                  "The $new_shares new shares to be created are not divisable"
                . "by VSERVER_COUNT $Options_hash->{'VSERVER_COUNT'}.\n All "
                . "shares will all be assigned to the 1st vserver. If this "
                . "is not what you want please increase SHARE_COUNT");
    }

    $Log->comment('Adding remainder of new shares to the first vserver');
    my $shares_per_vserver_remainder
        = $new_shares % ($Options_hash->{'VSERVER_COUNT'});
    my @dvs                         = natsort(keys(%{$defined_vservers}));
    my $first_vserver               = $dvs[0];
    my $shares_per_client           = 0;
    my $shares_per_client_remainder = 0;
    my @share_verification_clients;

    if ($recipe eq 'clients') {
        @share_verification_clients
            = grep { host($_)->hosttype() =~ /win/ } @{$self->{'clients'}};

        my $max_clients = $self->{'max_clients'};
        if (!$max_clients) {
            $max_clients = @share_verification_clients;
        } elsif ($max_clients < @share_verification_clients) {
            splice(@share_verification_clients, $max_clients);
        }

        if (@share_verification_clients > $new_shares) {

            $shares_per_client = 1;
            splice(@share_verification_clients, $new_shares);
            $self->{'clients'} = \@share_verification_clients;

        } else {
            $shares_per_client
                = int($new_shares / @share_verification_clients);
            $shares_per_client_remainder
                = $new_shares % @share_verification_clients;
        }
        $self->{'shares_per_client'}           = $shares_per_client;
        $self->{'shares_per_client_remainder'} = $shares_per_client_remainder;
    }

    $self->{'existing_shares'}             = $total_existing_shares;
    $self->{'new_shares'}                  = $new_shares;
    $self->{'shares_per_vserver'}          = $shares_per_vserver;
    $self->{'shares_per_vserver_reminder'} = $shares_per_vserver_remainder;
    #<<<

    my $summary = "Requested shares . . . . . . . . . . . . $Options_hash->{'SHARE_COUNT'}\n";
    $summary .= "Existing shares  . . . . . . . . . . . . $total_existing_shares\n";
    $summary .= "Shares to create   . . . . . . . . . . . $new_shares\n";
    $summary .= "Shares per vserver . . . . . . . . . . . $shares_per_vserver\n";
    $summary .= "Shares assigned to 1st vserver . . . . . " . ($shares_per_vserver + $shares_per_vserver_remainder) . "\n";
    if ($recipe eq 'clients') {
        $summary .= "Share verification clients . . . . . . . " . @share_verification_clients. "\n";
        $summary .= "Shares 1st client verifies . . . . . . . " . ($shares_per_client + $shares_per_client_remainder) . "\n";
        $summary .= "Shares verified by remaining client  . . $shares_per_client\n";
    }
    #>>>
    $Log->comment($summary);

    # Find the aggregates in AGGREGATES that match AGGR_REGEX
    my @target_aggregates;
    my @configured_aggregates = split(/,/, $Options_hash->{'AGGREGATES'});

    foreach my $aggregate (@configured_aggregates) {
        if ($aggregate =~ /$Options_hash->{'AGGR_REGEX'}/) {
            push(@target_aggregates, $aggregate);
        }
    }

    if (!@target_aggregates) {
        die config_result(
            "Unable to find any target target_aggregates using AGGREGATES "
                . $Options_hash->{'AGGREGATES'}
                . ' and AGGR_REGEX '
                . $Options_hash->{'AGGR_REGEX'} . ' '
                . Dumper(\@target_aggregates));
    }

    # Spread the shares across the nodes, vservers and volumes
    $Log->comment("Adding shares to vservers");
    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {
        $Log->comment("Vserver: $vserver");

        my $current_shares_per_vserver = $shares_per_vserver;
        if ($vserver eq $first_vserver) {
            $current_shares_per_vserver += $shares_per_vserver_remainder;
        }

        # Search for all volumes whos aggregate is listed in AGGREGATES
        my @target_volumes;
        foreach my $volume (natsort(keys(%{$defined_volumes}))) {
            my $aggregate = $defined_volumes->{$volume}->{'aggregate'};
            if (!$aggregate) {
                die config_result("Aggregate not defined for volume $volume "
                        . Dumper($defined_volumes));
            }
            my $vs = $defined_volumes->{$volume}->{'vserver'};
            if (!$vs) {
                die config_result("Vserver not defined for volume $volume\n"
                        . Dumper($defined_volumes));
            }

            if ((grep { $_ eq $aggregate } @target_aggregates)
                && $vs eq $vserver) {
                if (   $defined_volumes->{$volume}->{'rootvolume'} == 0
                    && $defined_volumes->{$volume}->{'type'} ne 'DC') {
                    $Log->debug(
                        "Saving $volume as a target volume for $vserver");
                    push(@target_volumes, $volume);
                }

            }
        }

        if (!@target_volumes) {
            die config_result("Unable to find any target volumes "
                    . "for vserver $vserver on aggreates "
                    . Dumper(\@target_aggregates));
        }

        $Log->comment('Volumes: ' . @target_volumes);
        my @volumes;
        my $volume;
        my $share_count = 0;
        $Log->comment("Shares: $current_shares_per_vserver");
        my $shares_per_volume           = 0;
        my $shares_per_volume_remainder = 0;

        if (@target_volumes > 1) {
            $shares_per_volume
                = int($current_shares_per_vserver / @target_volumes);
            $shares_per_volume_remainder
                = $current_shares_per_vserver % @target_volumes;
            $Log->comment("Shares assigned to 1st volume: "
                    . ($shares_per_volume + $shares_per_volume_remainder));
            $Log->comment(
                "Shares assigned to all other volmes: $shares_per_volume");
        } else {
            $shares_per_volume = $current_shares_per_vserver;
        }

        my $lif_index = 0;
        foreach (1 ... $current_shares_per_vserver) {

            # Create the share name
            my $share = 's_' . $script . '_' . $share_index;
            $share_index++;

            my $vs;
            my $path;
            my $vol;

            # If the share exists do not redefine but use it's vserver and
            # volume otherwise we keep using the current vserver and volume
            if (grep { $_ eq $share } keys(%{$discovered_shares})) {
                $path = $discovered_shares->{$share}->{'path'};
                $path =~ /^\/([^\/]+)(?:\/.+|$)/;
                $vol = $1;
                if (!$vol) {
                    die script_result("Unable to determine the "
                            . "volume for the share path $path");
                }
                $vs          = $discovered_shares->{$share}->{'vserver'};
                $filer_setup = 0;
                delete($discovered_shares->{$share});
                $Log->comment(
                    "Using existing share $share on vserver $vs with path $path"
                );
            } else {

                # If there are no more volumes in the array repopulate
                # the array
                if (!@volumes) { @volumes = @target_volumes; }
                if (!$volume)  { $volume  = shift(@volumes); }
                $filer_setup = 1;
                $vs          = $vserver;
                $vol         = $volume;
                if ($Options_hash->{'SHARE_PATH_CONFIG'} eq 'single') {
                    $path = "/$vol";
                } elsif ($Options_hash->{'SHARE_PATH_CONFIG'} eq 'multiple') {
                    $path = "/$vol/$share";
                }
                $Log->comment(
                    "Defining new share $share on vserver $vs with path $path"
                );

            }

            push(
                @{  $self->{'defined_topology'}->{'vservers'}->{$vs}
                        ->{'shares'}
                    },
                $share
            );

            # Store the the volume name with the share
            $topology_database->{'shares'}->{$share}->{'volume'} = $vol;

            # Store the share path
            $topology_database->{'shares'}->{$share}->{'path'} = $path;

            # Store the vserver name with the share
            $topology_database->{'shares'}->{$share}->{'vserver'} = $vs;

            # Store the flag to enable share creation in filer_setup
            $topology_database->{'shares'}->{$share}->{'filer_setup'}
                = $filer_setup;

            if ($recipe eq 'swifttest') {

                # Get the vserver address from the vserver entry
                my $lif = ${$defined_vservers->{$vs}->{'lifs'}}[$lif_index];

                # Increment or reset the interface index
                $lif_index++;
                if ($lif_index > $#{$defined_vservers->{$vs}->{'lifs'}}) {
                    $lif_index = 0;
                }

                # Get the ip address from the defined lifs
                my $ip = $defined_lifs->{$vs}->{$lif}->{'address'};
                if (!$ip) {
                    die config_result(
                              "Unable to get the IP addresss from vserver "
                            . "$vs LIF $lif\n"
                            . Dumper($defined_lifs->{$vs}));
                }

                my $mode   = $Options_hash->{'CIFS_SERVER_ACCESS_MODE'};
                my $server = $self->get_cifs_server_name(
                    'ip'   => $ip,
                    'mode' => $mode
                );

                $Log->comment(
                    "Assigning CIFS server $server for share $share");
                $topology_database->{'shares'}->{$share}->{'server'}
                    = $server;
            }

            $share_count++;
            if (   $shares_per_volume_remainder
                && @volumes
                && $volume eq $volumes[0]
                && $share_count == $shares_per_volume_remainder) {
                $volume      = undef;
                $share_count = 0;
            } elsif ($share_count == $shares_per_volume
                && @target_volumes > 1) {
                $volume      = undef;
                $share_count = 0;
            }

        }
    }

    $Log->exit();
} ## end sub define_shares

sub filer_cleanup_defined {

    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    my $defined_vservers = $self->{'defined_topology'}->{'vservers'};
    my $defined_lifs     = $self->{'defined_topology'}->{'lifs'};
    my $defined_routing_groups
        = $self->{'defined_topology'}->{'routing_groups'};
    my $defined_volumes     = $self->{'defined_topology'}->{'volumes'};
    my $discovered_lifs     = $self->{'discovered_topology'}->{'lifs'};
    my $discovered_volumes  = $self->{'discovered_topology'}->{'volumes'};
    my $discovered_vservers = $self->{'discovered_topology'}->{'vservers'};
    my $discovered_routing_groups
        = $self->{'discovered_topology'}->{'routing_groups'};
    my $discovered_antivirus = $self->{'discovered_topology'}->{'antivirus'};

    if ((grep { $_ eq 'state' } keys(%{$discovered_antivirus}))
        && $self->{'release'} !~ /boilermaker/) {
        if ($discovered_antivirus->{'state'} eq 'on') {
            NACL::C::AntivirusEngine->disable(command_interface => $ci);
        }

    }

    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {

   #Deterimine if the vserver exists first if it doesn't then we don't need to
   # anything.
        if (grep { $_ eq $vserver } keys(%{$discovered_vservers})) {
            if (keys(%{$discovered_vservers->{$vserver}->{'cifs'}})) {
                $Log->comment("Deleting CIFS server $vserver");
                my $cifs = NACL::C::VserverCifs->delete(
                    command_interface => $ci,
                    vserver           => $vserver,
                    username          => "administrator",
                    password          => $Options_hash->{'CIFS_PASSWD'}
                );
            }

            $self->wait_for_replay();

            $Log->comment("Deleting lifs for vserver $vserver");

            foreach my $lif (natsort(keys(%{$discovered_lifs->{$vserver}}))) {
                if (!grep { $_ eq $lif }
                    keys(%{$discovered_lifs->{$vserver}})
                    ) {
                    next;
                }
                my $l = NACL::C::Interface->new(
                    command_interface => $ci,
                    lif               => $lif,
                    server            => $vserver
                );

                # Delete the lif
                $l->delete();

            }

            $self->wait_for_replay();

            $Log->comment("Deleting routing groups for vserver $vserver");
            foreach my $rg (
                natsort(keys(%{$discovered_routing_groups->{$vserver}}))) {
                eval {
                    NACL::C::Route->delete(
                        command_interface => $ci,
                        'routing-group'   => $rg,
                        server            => $vserver,
                        destination       => '*'
                    );
                };
                NACL::C::RoutingGroups->delete(
                    command_interface => $ci,
                    'routing-group'   => $rg,
                    server            => $vserver
                );
            }

            $self->wait_for_replay();

            # Get the volumes for this vserver
            my @volumes = @{$defined_vservers->{$vserver}->{'volumes'}};

            # Get the vserver root volume
            my $root_volume = $defined_vservers->{$vserver}->{'rootvolume'};

            if ($self->{'release'} !~ /boilermaker/) {
                $Log->comment("Deleting cache volumes for $vserver");
                foreach my $volume (@volumes) {

                    if ($discovered_volumes->{$volume}->{'type'} ne 'DC') {
                        next;
                    }

                    if ($discovered_volumes->{$volume}->{'vserver'} ne
                        $vserver) {
                        next;
                    }

# Let's create an object so that we can simplify the performing of the other operations
# Generally, we'd suggest only 'find' or 'create' be used to get component objects
# but if we know the element exists and we have all the details necessary to identify it
# we can call new ourselves
                    my $vol = NACL::C::Volume->new(
                        command_interface => $ci,
                        volume            => $volume,
                        vserver           => $vserver
                    );

                    # Offline the volume
                    $vol->modify(state => 'offline');

                    # Wait for the offline state
                    while ($vol->state()->state() !~ /offline /) {
                        snooze 5;
                    }

                    # Delete the volume
                    $vol->delete();

                }
                $self->wait_for_replay();
            }

            $Log->comment("Deleting data volumes for $vserver");
            foreach my $volume (@volumes) {

                if ($defined_volumes->{$volume}->{'rootvolume'}) {
                    next;
                }

                if (!grep { $_ eq $volume } keys(%{$discovered_volumes})) {
                    next;
                }

                NACL::C::Volume->unmount(
                    command_interface => $ci,
                    volume            => $volume,
                    vserver           => $vserver
                );

                snooze 1;

# Let's create an object so that we can simplify the performing of the other operations
# Generally, we'd suggest only 'find' or 'create' be used to get component objects
# but if we know the element exists and we have all the details necessary to identify it
# we can call new ourselves
                my $vol = NACL::C::Volume->new(
                    command_interface => $ci,
                    volume            => $volume,
                    vserver           => $vserver
                );

                # Offline the volume
                $vol->modify(state => 'offline');

                # Wait for the offline state
                while ($vol->state()->state() !~ /offline/) {
                    snooze 5;
                }

                # Delete the volume
                $vol->delete();

            }

            $self->wait_for_replay();

            $Log->comment("Deleting root volume for vserver $vserver");
            if (grep { $_ eq $root_volume } keys(%{$discovered_volumes})) {

                my $vol = NACL::C::Volume->new(
                    command_interface => $ci,
                    volume            => $root_volume,
                    vserver           => $vserver
                );

                # Offline the root volume
                $vol->modify(state => 'offline');

                # Wait for the offline state
                while ($vol->state()->state() !~ /offline/) {
                    snooze 5;
                }

                # Delete the root volume
                $vol->delete();

                $self->wait_for_replay();
            }

            if (grep { $_ eq $vserver } keys(%{$discovered_vservers})) {

                $Log->comment("Deleting vserver $vserver");
                my $vs = NACL::C::Vserver->delete(
                    command_interface => $ci,
                    vserver           => $vserver
                );
                $self->wait_for_replay();
            }
        }
    }

    #Disable autosupport
    if ($Options_hash->{'AUTOSUPPORT'}) {
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            $Log->comment("Deconfiguring ASUP information for node $node");
            NACL::C::SystemNodeAutosupport->modify(
                command_interface => $ci,
                node              => $node,
                'state'           => 'disable',
                'support'         => 'disable',
            );
        }
    }

    $Log->exit();
} ## end sub filer_cleanup_defined

sub filer_cleanup_discovered {

    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    my $defined_vservers    = $self->{'defined_topology'}->{'vservers'};
    my $defined_volumes     = $self->{'defined_topology'}->{'volumes'};
    my $discovered_lifs     = $self->{'discovered_topology'}->{'lifs'};
    my $discovered_volumes  = $self->{'discovered_topology'}->{'volumes'};
    my $discovered_vservers = $self->{'discovered_topology'}->{'vservers'};
    my $discovered_routing_groups
        = $self->{'discovered_topology'}->{'routing_groups'};
    my $discovered_quotas    = $self->{'discovered_topology'}->{'quotas'};
    my @volumes              = natsort(keys(%{$discovered_volumes}));
    my $discovered_antivirus = $self->{'discovered_topology'}->{'antivirus'};

    if ((grep { $_ eq 'state' } keys(%{$discovered_antivirus}))
        && $self->{'release'} !~ /boilermaker/) {
        if ($discovered_antivirus->{'state'} eq 'on') {
            NACL::C::AntivirusEngine->disable(command_interface => $ci);
        }

    }

    foreach my $vserver (natsort(keys(%{$discovered_vservers}))) {

        my $root_volume = $discovered_vservers->{$vserver}->{'rootvolume'};

        if ($discovered_vservers->{$vserver}->{'type'} !~ /cluster|data/) {
            next;
        }

        if (keys(%{$discovered_vservers->{$vserver}->{'cifs'}})) {
            $Log->comment("Deleting CIFS chares for $vserver");
            eval {
                NACL::C::VserverCifsShare->delete(
                    command_interface => $ci,
                    vserver           => $vserver,
                    'share-name'      => 's_*'
                );
            };

            $self->wait_for_replay();

            $Log->comment("Deleting the CIFS server for $vserver");
            my $cifs = NACL::C::VserverCifs->delete(
                command_interface => $ci,
                vserver           => $vserver,
                username          => "administrator",
                password          => $Options_hash->{'CIFS_PASSWD'}
            );

            $self->wait_for_replay();
        } else {
            next;
        }

        if ($self->{'release'} !~ /boilermaker/) {
            $Log->comment("Deleting quota policy rules for $vserver");
            foreach
                my $policy (natsort(keys(%{$discovered_quotas->{$vserver}})))
            {
                foreach my $rule (
                    natsort(
                        @{  $discovered_quotas->{$vserver}->{$policy}
                                ->{'rules'}
                            }
                    )
                    ) {

                    NACL::C::QuotaPolicyRule->delete(
                        command_interface => $ci,
                        vserver           => $vserver,
                        'policy-name'     => $policy,
                        type              => $rule->{'type'},
                        target            => $rule->{'target'},
                        qtree             => $rule->{'qtree'},
                        volume            => $rule->{'volume'}
                    );
                }
                if ($policy ne 'default') {
                    $Log->comment(
                        "Deleting quota policy $policy for $vserver");
                    NACL::C::QuotaPolicy->delete(
                        command_interface => $ci,
                        vserver           => $vserver,
                        'policy-name'     => $policy
                    );

                }
            }
        }

        if ($self->{'release'} !~ /boilermaker/) {
            $Log->comment("Deleting cache volumes for $vserver");
            foreach my $volume (@volumes) {

                if ($discovered_volumes->{$volume}->{'type'} ne 'DC') {
                    next;
                }

                if ($discovered_volumes->{$volume}->{'vserver'} ne $vserver) {
                    next;
                }

# Let's create an object so that we can simplify the performing of the other operations
# Generally, we'd suggest only 'find' or 'create' be used to get component objects
# but if we know the element exists and we have all the details necessary to identify it
# we can call new ourselves
                my $vol = NACL::C::Volume->new(
                    command_interface => $ci,
                    volume            => $volume,
                    vserver           => $vserver
                );

                # Offline the volume
                $vol->modify(state => 'offline');

                # Wait for the offline state
                while ($vol->state()->state() !~ /offline/) {
                    snooze 5;
                }

                # Delete the volume
                $vol->delete();

            }
        }

        $self->wait_for_replay();

        #$Log->comment(Dumper($self));

        $Log->comment("Deleting data volumes for $vserver");
        foreach my $volume (@volumes) {

            if ($discovered_volumes->{$volume}->{'vsroot'} eq 'true') {
                next;
            }

            if ($discovered_volumes->{$volume}->{'type'} eq 'DC') {
                next;
            }

            if ($discovered_volumes->{$volume}->{'vserver'} ne $vserver) {
                next;
            }

            NACL::C::Volume->unmount(
                command_interface => $ci,
                volume            => $volume,
                vserver           => $vserver
            );

            snooze 1;

# Let's create an object so that we can simplify the performing of the other operations
# Generally, we'd suggest only 'find' or 'create' be used to get component objects
# but if we know the element exists and we have all the details necessary to identify it
# we can call new ourselves
            my $vol = NACL::C::Volume->new(
                command_interface => $ci,
                volume            => $volume,
                vserver           => $vserver
            );

            # Offline the volume
            $vol->modify(state => 'offline');

            # Wait for the offline state
            while ($vol->state()->state() !~ /offline/) {
                snooze 5;
            }

            # Delete the volume
            $vol->delete();

        }

        $self->wait_for_replay();

        $Log->comment("Deleting interfaces for vserver $vserver");
        foreach my $lif (natsort(keys(%{$discovered_lifs->{$vserver}}))) {
            my $l = NACL::C::Interface->new(
                command_interface => $ci,
                lif               => $lif,
                server            => $vserver
            );

            # Offline the volume
            $l->delete();

        }

        $self->wait_for_replay();

        $Log->comment("Deleting routing groups for vserver $vserver");
        foreach
            my $rg (natsort(keys(%{$discovered_routing_groups->{$vserver}})))
        {

            eval {
                NACL::C::Route->delete(
                    command_interface => $ci,
                    'routing-group'   => $rg,
                    server            => $vserver,
                    destination       => '*'
                );
            };
            NACL::C::RoutingGroups->delete(
                command_interface => $ci,
                'routing-group'   => $rg,
                server            => $vserver
            );
        }

        $self->wait_for_replay();

        # Offline the root volume
        $Log->comment("Deleting root volume for vserver $vserver");
        if (grep { $_ eq $root_volume } keys(%{$discovered_volumes})) {

            my $vol = NACL::C::Volume->new(
                command_interface => $ci,
                volume            => $root_volume,
                vserver           => $vserver
            );

            # Offline the volume
            $vol->modify(state => 'offline');

            # Wait for the offline state
            while ($vol->state()->state() !~ /offline/) {
                snooze 5;
            }

            # Delete the volume
            $vol->delete();

        }

        $self->wait_for_replay();

        $Log->comment("Deleting vserver $vserver");
        my $vs = NACL::C::Vserver->delete(
            command_interface => $ci,
            vserver           => $vserver
        );

        $self->wait_for_replay();
    }

    #Disable autosupport
    if ($Options_hash->{'AUTOSUPPORT'}) {
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            $Log->comment("Deconfiguring ASUP information for node $node");
            NACL::C::SystemNodeAutosupport->modify(
                command_interface => $ci,
                node              => $node,
                'state'           => 'disable',
                'support'         => 'disable',
            );
        }
    }

    #$Log->comment(Dumper($discovered_vservers));
       #die(config_result('Dumping Topology'));

    $Log->exit();
} ## end sub filer_cleanup_discovered

sub unix_user_new {

    $Log->enter();

    my ($self, %options) = @_;
    my $gid     = $options{'gid'};
    my $name    = $options{'name'};
    my $uid     = $options{'uid'};
    my $user    = $options{'user'};
    my $vserver = $options{'vserver'};

    $Options_hash = $Options->get_hash();

    $Log->comment("Creating local unix user $user wth UID "
            . "$uid GID $gid name $name on vserver $vserver");

    NACL::C::UnixUser->create(
        'command_interface' => $ci,
        vserver             => $vserver,
        user                => $user,
        id                  => $uid,
        'full-name'         => $name,
        'primary-gid'       => $gid,
    );
    $Log->exit();
} ## end sub unix_user_new

sub unix_group_new {

    $Log->enter();

    my ($self, %options) = @_;
    my $gid     = $options{'gid'};
    my $group   = $options{'group'};
    my $vserver = $options{'vserver'};

    $Log->comment("Adding local unix group $group with "
            . "GID $gid to vserver $vserver");

    NACL::C::UnixGroup->create(
        'command_interface' => $ci,
        vserver             => $vserver,
        name                => $group,
        id                  => $gid,
    );
    $Log->exit();
} ## end sub unix_group_new

sub unix_group_add_user {

    $Log->enter();

    my ($self, %options) = @_;
    my $group   = $options{'group'};
    my $user    = $options{'user'};
    my $vserver = $options{'vserver'};

    $Log->comment("Adding user $user to local unix "
            . "group $group on vserver $vserver");

    NACL::C::UnixGroup->adduser(
        command_interface => $ci,
        name              => $group,
        vserver           => $vserver,
        username          => $user
    );
    $Log->exit();
} ## end sub unix_group_add_user

sub time_setup {

    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    if (!$Options_hash->{'TIME_ZONE'} && !$Options_hash->{'TIME_SERVER'}) {
        $Log->comment('TIME_ZONE and TIME_SERVER are not '
                . ' defined skipping NTP configuration');
        return;
    }

    if (!$Options_hash->{'CIFS_DC'}) {
        die config_result("CIFS_DC is required");
    }

    host($Options_hash->{'CIFS_DC'});
    my $dc_conn_obj = connect('#rsh', '-nologin', -timeout, 300);

    $Log->comment("Getting date for DC");
    my ($date, $output);
    my $cmd = "cmd /c date /t";
    eval { (undef, undef, $output) = say($dc_conn_obj, $cmd, -timeout, 300); };

    foreach my $line (split(/\n/, $output)) {
        chomp($line);

        # Tue 07/17/2007
        if ($line =~ /\w+\s+(\d+\/\d+\/\d+)/) {
            $date = $1;
        }
    }

    $Log->comment("Getting time for DC");
    $cmd = "cmd /c time /t";
    eval { (undef, undef, $output) = say($dc_conn_obj, $cmd, -timeout, 300); };

    foreach my $line (split(/\n/, $output)) {
        chomp($line);

        # 07:00 PM
        if ($line =~ /^(\d+):(\d+)\s+(\w+)/) {
            my $hour     = $1;
            my $minutes  = $2;
            my $day_time = $3;
            if ($day_time =~ /AM/i && $hour == 12) { $hour = '00'; }
            if ($day_time =~ /PM/i && $hour != 12) {
                $hour = $hour + 12;
            }
            $date .= " $hour:$minutes:00";
        }
    }

    #Change for SN.0
    if ($self->{'release'} =~ /rollingrock|boilermaker/) {
        NACL::C::SystemServicesNtpConfig->modify(
            command_interface => $ci,
            enabled           => 'false'
        );
        snooze 1;

        my $discovered_ntp_servers
            = $self->{'discovered_topology'}->{'ntp_servers'};
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            my $server = $Options_hash->{'TIME_SERVER'};
            if (!grep { $_ eq $server }
                keys(%{$discovered_ntp_servers->{$node}})
                ) {
                NACL::C::SystemServicesNtpServer->create(
                    command_interface => $ci,
                    node              => $node,
                    server            => $server,
                    preferred         => 'true',
                    version           => 'max'
                );
            }
        }

        snooze 1;

        NACL::C::SystemNodeDate->modify(
            command_interface => $ci,
            node              => '*',
            date              => $date,
            timezone          => $Options_hash->{'TIME_ZONE'}
        );

        NACL::C::SystemServicesNtpConfig->modify(
            command_interface => $ci,
            enabled           => 'true'
        );
    } else {
        
        my $discovered_ntp_servers
            = $self->{'discovered_topology'}->{'ntp_servers'};
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            my $server = $Options_hash->{'TIME_SERVER'};
            if (!grep { $_ eq $server }
                keys(%{$discovered_ntp_servers->{$node}})
                ) {
                NACL::C::SystemServicesNtpServer->create(
                    command_interface => $ci,
                    node              => $node,
                    server            => $server,
                    preferred         => 'true',
                    version           => 'max'
                );
            }
        }

        snooze 1;
        
        NACL::C::ClusterDate->modify(
            command_interface => $ci,
            node              => '*',
            date              => $date,
            timezone          => $Options_hash->{'TIME_ZONE'}
        );
    }

    $Log->exit();

} ## end sub time_setup

sub list_shares {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment(
        "Discovering shares configured on vserver $options{'vserver'}");

    my @shares = NACL::CS::VserverCifsShare->fetch(
        command_interface => $options{'command_interface'},
        filter            => {vserver => $options{'vserver'}},
        allow_empty       => 1
    );

    my @shareInfo;
    foreach my $s (@shares) {
        my %properties = %$s;
        if ($options{'filter'} && $s->share_name() !~ /$options{'filter'}/) {
            next;
        }
        push(@shareInfo, \%properties);
    }

    $Log->exit();
    return \@shareInfo;
} ## end sub list_shares

sub list_share_acls {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Getting share ACLs for $options{'share'} "
            . "on vserver $options{'vserver'}");

    my @acls = NACL::CS::VserverCifsShareAccessControl->fetch(
        command_interface => $options{'command_interface'},
        'filter' =>
            {'share' => $options{'share'}, 'vserver' => $options{'vserver'}}
    );

    my @aclInfo;
    foreach my $a (@acls) {
        my %properties = %$a;
        push(@aclInfo, \%properties);
    }

    $Log->exit();
    return \@aclInfo;
} ## end sub list_share_acls

sub create_share_acl {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Creating share ACL $options{'permission'} for "
            . "group $options{'user_or_group'} on $options{'share'}");

    my $output;
    $options{'permission'}    =~ s/\s+/_/g;
    $options{'user_or_group'} =~ s/^(.*)$/"$1"/g;
    NACL::C::VserverCifsShareAccessControl->create(
        command_interface => $options{'command_interface'},
        'vserver'         => $options{'vserver'},
        'share'           => $options{'share'},
        'user-or-group'   => $options{'user_or_group'},
        'permission'      => $options{'permission'}
    );

    $Log->exit();
} ## end sub create_share_acl

sub delete_share_acl {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Deleting share ACL for user or group "
            . "$options{'user_or_group'} on $options{'share'}");

    my $output;
    $options{'permission'}    =~ s/\s+/_/g;
    $options{'user_or_group'} =~ s/^(.*)$/"$1"/g;
    NACL::C::VserverCifsShareAccessControl->create(
        command_interface => $options{'command_interface'},
        'vserver'         => $options{'vserver'},
        'share'           => $options{'share'},
        'user-or-group'   => $options{'user_or_group'},
    );

    $Log->exit();
} ## end sub delete_share_acl

sub modify_share_acl {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Modifying share ACL $options{'permission'} for "
            . "group $options{'user_or_group'} on $options{'share'}");

    $options{'permission'}    =~ s/\s+/_/g;
    $options{'user_or_group'} =~ s/^(.*)$/"$1"/g;
    NACL::C::VserverCifsShareAccessControl->modify(
        command_interface => $options{'command_interface'},
        'vserver'         => $options{'vserver'},
        'share'           => $options{'share'},
        'user-or-group'   => $options{'user_or_group'},
        'permission'      => $options{'permission'}
    );

    $Log->exit();
} ## end sub modify_share_acl

sub assign_share_acl_clients {

    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    $Log->comment("Assigning test options for share ACLs");

    my $instance     = 0;
    my $client_count = 0;

    for my $client (@{$self->{'clients'}}) {

        if (host($client)->hosttype()
            !~ /$Options_hash->{'CLIENT_HOSTTYPE_REGEX'}/) {
            next;
        }
        $client_count++;
        $Log->comment("Client: $client");
        $Log->comment("Instance: $instance");

        $self->{'clients_status'}->{$client}->{'status'} = 'assigned';
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_CONNECT_METHOD'}
            = $Options_hash->{'CLIENT_CONNECT_METHOD'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_HOSTTYPE_REGEX'}
            = $Options_hash->{'CLIENT_HOSTTYPE_REGEX'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'TEST_MAIN_DURATION'} = $Options_hash->{'TEST_MAIN_DURATION'};
        $self->{'test_options'}->{$client}->{$instance}->{'TEST_MAIN_SLEEP'}
            = $Options_hash->{'TEST_MAIN_SLEEP'};

        if ($client_count == $self->{'max_clients'}) {
            last;
        }
    }

    $Log->exit();
} ## end sub assign_share_acl_clients

sub create_shares {

    $Log->enter();

    my ($self, %options) = @_;

    my $regex  = $options{'regex'};
    my $mapped = $options{'mapped'};

    $Options_hash = $Options->get_hash();

    my $defined_lifs      = $self->{'defined_topology'}->{'lifs'};
    my $defined_vservers  = $self->{'defined_topology'}->{'vservers'};
    my $defined_shares    = $topology_database->{'shares'};
    my $discovered_qtrees = $self->{'discovered_topology'}->{'qtrees'};
    my $discovered_shares = $self->{'discovered_topology'}->{'shares'};

    $Log->comment("Creating shares matching $regex");

    my $client_conn_obj;
    my @share_creation_clients
        = grep { host($_)->hosttype() =~ /win/ } @{$self->{'clients'}};
    if (@share_creation_clients) {
        host($share_creation_clients[0]);
        $client_conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
            '-nologin', -timeout, 300);
        eval { check_connectivity('conn' => $client_conn_obj); };
        my $error = $@;
        if ($error) { die config_result($error); }
    }

    my $share_count = 0;
    my $last_share  = 0;
    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {
        foreach
            my $share (natsort(@{$defined_vservers->{$vserver}->{'shares'}}))
        {

            # Store the last share sowe can check that it has propagated
            # to all nodes
            $last_share = $share;

            # If the share exists skip it
            if (grep(/^$share$/, keys(%{$discovered_shares}))) { next; }

            # If filer_setup is not set then we skip share creation in
            # filer_setup
            if (!$defined_shares->{$share}->{'filer_setup'}) { next; }

            # These are not the droids we are looking for, move along...
            if ($share !~ /$options{'regex'}/) { next; }

            # If the qtree exists skip it
            my $qtree   = $defined_shares->{$share}->{'qtree'};
            my $vserver = $defined_shares->{$share}->{'vserver'};
            my $volume  = $defined_shares->{$share}->{'volume'};
            my $path    = $defined_shares->{$share}->{'path'};

            if (   $qtree
                && !grep(/^$share$/, keys(%{$discovered_qtrees}))
                && $share ne "s_$volume") {
                $Log->comment("Creating qtree $share on $volume");

                if (!grep { $_ eq $share } keys(%{$discovered_qtrees})) {
                    NACL::C::Qtree->create(
                        command_interface => $ci,
                        volume            => $volume,
                        vserver           => $vserver,
                        qtree             => $share,
                    );
                }

            } elsif (!$qtree
                && $share ne "s_$volume"
                && $share ne "s_$vserver") {

                @share_creation_clients
                    = grep { host($_)->hosttype() =~ /win/ }
                    @{$self->{'clients'}};
                if (!@share_creation_clients) {
                    die config_result(
                        "There are no windows clients to create share directories"
                    );
                }

                my @lifs = natsort(keys(%{$defined_lifs->{$vserver}}));
                my $server
                    = $defined_lifs->{$vserver}->{$lifs[0]}->{'address'};

                my $volume_share;
                if ($Options_hash->{'SHARE_ROOT_VOLUMES'}) {
                    $volume_share = "\\\\$server\\s_$vserver";
                } elsif ($Options_hash->{'SHARE_DATA_VOLUMES'}) {
                    $volume_share = "\\\\$server\\s_$volume";
                }

                if (!${$mapped}{$volume}) {
                    map_share(
                        'conn'     => $client_conn_obj,
                        'share'    => $volume_share,
                        'drive'    => '',
                        'username' => $Options_hash->{'CIFS_USER'},
                        'password' => $Options_hash->{'CIFS_PASSWD'},
                        'timeout'  => 300,
                        'reuse'    => 1,
                        'delay'    => 60,
                        'attempts' => 3,
                    );
                    ${$mapped}{$volume} = 1;
                }

                my $path;
                if ($Options_hash->{'SHARE_ROOT_VOLUMES'}) {
                    $path = "\\\\$server\\s_$vserver\\$volume\\$share";
                } elsif ($Options_hash->{'SHARE_DATA_VOLUMES'}) {
                    $volume_share = "\\\\$server\\s_$volume\\$share";
                }

                my $cmd
                    = 'cmd.exe /c if not exist ' . $path . ' mkdir ' . $path;

                my $output;
                eval {
                    (undef, undef, $output)
                        = say($client_conn_obj, $cmd, -timeout, 300);
                };
                my $error = $@;
                if ($output || $error) {
                    die config_result("Output:$output\nError:$error");
                }
            }

            $Log->comment(
                "Creating share $share on Vserver:$vserver Volume:$volume Path:$path"
            );
            if (!grep { $_ eq $share } keys(%{$discovered_shares})) {
                my @share_properties
                    = split(/,/, $Options_hash->{'SHARE_PROPERTIES'});
                NACL::C::VserverCifsShare->create(
                    command_interface  => $ci,
                    vserver            => $vserver,
                    'share-name'       => $share,
                    path               => $path,
                    'share-properties' => \@share_properties,

                    #'symlink-properties' => 'enable',
                );
                $share_count++;
            }

            snooze 1;
        }
    }

    return ($share_count, $last_share);

    $Log->exit();
} ## end sub create_shares

sub replay_status {
    $Log->enter();

    my ($self, %options) = @_;

    my @replays = NACL::CS::DiagNbladeReplay->fetch(
        command_interface => $ci,
        allow_empty       => 1
    );
    my $status = 1;
    if (@replays) {
        foreach my $r (@replays) {
            my %properties = %$r;
            if ($properties{'result'} ne 'succeeded') { $status = 0; }

        }
    } else {
        $status = 0;
    }

    $Log->exit();
    return $status;

} ## end sub replay_status

sub wait_for_replay {
    $Log->enter();

    my $self = shift;

    $Options_hash = $Options->get_hash();

    if ($Options_hash->{'WAIT_FOR_REPLAY'}) {
        my $interval;
        my $seconds;

        if (   ($Options_hash->{'WAIT_FOR_REPLAY'} =~ /(\d+),(\d+)/)
            && ($self->{'release'} !~ /boilermaker/)) {
            $interval = $1;
            $seconds  = $2;

            $Log->comment("Waiting $seconds seconds for nblade replay");
            my $elapsed_time = 0;
            while ($elapsed_time < $seconds) {
                if ($self->replay_status()) {
                    $Log->exit();
                    return;
                }
                $elapsed_time += $interval;
                snooze $interval;
            }
            $Log->exit();
            die fail_result(
                "Nblade replay did not succeed after $seconds seconds");
        } elsif (($Options_hash->{'WAIT_FOR_REPLAY'} =~ /(\d+),(\d+)/)
            && ($self->{'release'} =~ /boilermaker/)) {
            $Log->comment("Waiting $interval seconds for nblade replay");
            snooze $interval;
        } elsif ($Options_hash->{'WAIT_FOR_REPLAY'} =~ /^(\d+)$/) {
            $Log->comment("Waiting $seconds seconds for nblade replay");
            snooze $seconds;
        } else {

            #We will snooze for 30 seconds if all else fails.
            snooze 30;
        }
    }
} ## end sub wait_for_replay

sub define_cache_volumes {
    $Log->enter();

    my ($self, %options) = @_;
    my $starting_index  = $options{'index'};
    my $script          = $options{'script'};
    my $volume          = $options{'volume'};
    my $size            = $options{'size'};
    my $defined_volumes = $self->{'defined_topology'}->{'volumes'};
    my $discovered_aggregates
        = $self->{'discovered_topology'}->{'aggregates'};

    my $vserver = $defined_volumes->{$volume}->{'vserver'};

    # Get the node and aggrregate of the current volume
    my $local_aggregate = $defined_volumes->{$volume}->{'aggregate'};
    my $local_node
        = ${$discovered_aggregates->{$local_aggregate}->{'nodes'}}[0];

    # Search for all remote aggregates and nodes
    my @remote_aggregates;
    my @remote_nodes;
    foreach my $aggr (split(/,/, $self->{'aggregates'})) {
        if (@{$discovered_aggregates->{$aggr}->{'nodes'}} > 1) {
            next;
        }
        my $node = ${$discovered_aggregates->{$aggr}->{'nodes'}}[0];
        if ($node ne $local_node) {
            push(@remote_nodes,      $node);
            push(@remote_aggregates, $aggr);
        }
    }

    if (!@remote_aggregates) {
        die config_result('Unable to find a non local single '
                . 'node aggregate for cache volumes');
    }

    # Get the specified aggregates of the current remote node
    my @target_aggregates;
    foreach my $aggr (@remote_aggregates) {
        if (grep { $_ eq $aggr }
            split(/,/, $self->{'aggregates'})
            ) {
            push(@target_aggregates, $aggr);
        }

    }

    # Assign the cache volumes to aggregates of the remote nodes
    my $ending_index = $self->assign_volumes_to_aggregates(
        'aggregates' => join(',', @target_aggregates),
        'count'      => scalar(@remote_nodes),
        'index'      => $starting_index,
        'script'     => $script,
        'size'       => $size,
        'type'       => 'DC',
    );

    # Assign the cache volume for each remote node
    for (
        my $volume_index = $starting_index;
        $volume_index < $ending_index;
        $volume_index++
        ) {

        # Create the cache volume name
        my $cache_volume
            = 'v_' . $script . '_' . $volume_index . '_' . 'flex_dc';
        $Log->comment("Adding cache volume $cache_volume to $vserver");

        push(
            @{  $self->{'defined_topology'}->{'vservers'}->{$vserver}
                    ->{'volumes'}
                },
            $cache_volume
        );

        $self->{'defined_topology'}->{'volumes'}->{$cache_volume}
            ->{'rootvolume'} = 0;

        # Store the origin volume of the cache volume
        $self->{'defined_topology'}->{'volumes'}->{$cache_volume}
            ->{'origin_volume'} = $volume;

        # Store the origin volume of the cache volume
        $self->{'defined_topology'}->{'volumes'}->{$cache_volume}->{'vserver'}
            = $vserver;
    }

    $Log->exit();
    return $ending_index;
} ## end sub define_cache_volumes

sub assign_stream_clients {

    $Log->enter();

    my ($self, %options) = @_;
    my $instance = 0;
    my $share_index;
    my @volumes;

    # Get the target nodes
    my @target_nodes
        = grep(/$Options_hash->{'NODE_REGEX'}/, @{$Options_hash->{'NODES'}});

    if (!@target_nodes) {
        die config_result("Unable to find any nodes using the NODE_REGEX "
                . $Options_hash->{'NODE_REGEX'}
                . ' against '
                . Dumper($Options_hash->{'NODES'}));
    }

    # Get the target aggregates
    my @target_aggregates = grep(/$Options_hash->{'AGGR_REGEX'}/,
        split(/,/, $Options_hash->{'AGGREGATES'}));

    if (!@target_aggregates) {
        die config_result(
                  "Unable to find any aggregates using the AGGR_REGEX "
                . $Options_hash->{'AGGR_REGEX'}
                . ' against '
                . $Options_hash->{'AGGREGATES'});
    }

    # Get all volumes from the target aggregates
    my $defined_volumes = $self->{'defined_topology'}->{'volumes'};
    foreach my $volume (natsort(keys(%{$defined_volumes}))) {
        my $root_volume = $defined_volumes->{$volume}->{'rootvolume'};
        my $type        = $defined_volumes->{$volume}->{'type'};
        my $aggregate   = $defined_volumes->{$volume}->{'aggregate'};
        if ($root_volume || $type ne 'RW') { next; }
        if (grep { $_ eq $aggregate } @target_aggregates) {
            push(@volumes, $volume);
        }
    }

    # Loop through each client and assign the correct info to it.
    my $client_count    = 0;
    my $interface_index = 0;
    my %addresses;
    my $defined_vservers = $self->{'defined_topology'}->{'vservers'};
    my $defined_lifs     = $self->{'defined_topology'}->{'lifs'};
    my %nodes;
    foreach my $vserver (keys(%{$defined_vservers})) {
        @{$addresses{$vserver}} = ();
    }

    for my $client (@{$self->{'clients'}}) {

        if (host($client)->hosttype()
            !~ /$Options_hash->{'CLIENT_HOSTTYPE_REGEX'}/) {
            next;
        }

        $self->{'clients_status'}->{$client}->{'status'} = 'assigned';
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_CONNECT_METHOD'}
            = $Options_hash->{'CLIENT_CONNECT_METHOD'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_HOSTTYPE_REGEX'}
            = $Options_hash->{'CLIENT_HOSTTYPE_REGEX'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'TEST_MAIN_DURATION'} = $Options_hash->{'TEST_MAIN_DURATION'};
        $self->{'test_options'}->{$client}->{$instance}->{'TEST_MAIN_SLEEP'}
            = $Options_hash->{'TEST_MAIN_SLEEP'};

        # Get the next volume
        my $volume;
        if (@volumes) {
            $volume = shift(@volumes);
        } else {
            die script_result("Not enough volumes for all clients.\n"
                    . "Please specify a higher VOLUME_COUNT\n"
                    . "or change AGGR_REGEX to include more\n"
                    . "aggregates or decrease the number of\n"
                    . "clients by specifying less clients or\n"
                    . "limiting clients by using MAX_CLIENTS");
        }

        # Get the related information
        my $share           = 's_' . $volume;
        my $defined_volumes = $self->{'defined_topology'}->{'volumes'};
        my $aggregate       = $defined_volumes->{$volume}->{'aggregate'};
        my $vserver         = $defined_volumes->{$volume}->{'vserver'};

        # Get the addresses from the target nodes
        if (!@{$addresses{$vserver}}) {

            foreach my $lif (@{$defined_vservers->{$vserver}->{'lifs'}}) {
                my $node = $defined_lifs->{$vserver}->{$lif}->{'node'};
                if (grep { $_ eq $node } @target_nodes) {
                    my $address
                        = $defined_lifs->{$vserver}->{$lif}->{'address'};
                    $nodes{$address} = $node;
                    push(@{$addresses{$vserver}}, $address);
                }
            }
        }

        # Get the next address
        my $ip = shift(@{$addresses{$vserver}});

        my $node = $nodes{$ip};

        my $mode = $Options_hash->{'CIFS_SERVER_ACCESS_MODE'};
        my $server
            = $self->get_cifs_server_name('ip' => $ip, 'mode' => $mode);

        $Log->comment("Client: $client");
        $Log->comment("Aggregate: $aggregate");
        $Log->comment("Volume: $volume");
        $Log->comment("Vserver: $vserver");
        $Log->comment("Share: $share");
        $Log->comment("Node: $vserver");
        $Log->comment("IP: $ip");
        $Log->comment("Server: $server");
        $self->update_test_options(
            'client'   => $client,
            'instance' => $instance,
            'node'     => $node,
            'vserver'  => $vserver,
            'server'   => $server,
            'volume'   => $volume,
            'share'    => $share
        );

        $client_count++;
        if ($client_count == $self->{'max_clients'}) { last; }
    }

    $Log->exit();
    return $share_index++;
} ## end sub assign_stream_clients

sub check_share_propagation {

    $Log->enter();

    my ($self, %options) = @_;

    my $share = $options{'share'};

    $Options_hash = $Options->get_hash();

    my $defined_lifs      = $self->{'defined_topology'}->{'lifs'};
    my $defined_vservers  = $self->{'defined_topology'}->{'vservers'};
    my $defined_shares    = $topology_database->{'shares'};
    my $discovered_qtrees = $self->{'discovered_topology'}->{'qtrees'};
    my $discovered_shares = $self->{'discovered_topology'}->{'shares'};

    $Log->comment("Checking the propagation of $share");

    my $client_conn_obj;
    my @share_creation_clients
        = grep { host($_)->hosttype() =~ /win/ } @{$self->{'clients'}};
    if (@share_creation_clients) {
        host($share_creation_clients[0]);
        $client_conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
            '-nologin', -timeout, 300);
        eval { check_connectivity('conn' => $client_conn_obj); };
        my $error = $@;
        if ($error) { die config_result($error); }
    }

    my $vserver    = $defined_shares->{$share}->{'vserver'};
    my @lifs       = @{$defined_vservers->{$vserver}->{'lifs'}};
    my $max_checks = 60;
    for (my $i = 1; $i <= $max_checks; $i++) {
        my $failure = 0;
        foreach my $lif (@lifs) {
            my $ip = $self->{'defined_topology'}->{'lifs'}->{$vserver}->{$lif}
                ->{'address'};
            my $unc = '\\\\' . $ip . '\\' . $share;
            $Log->comment("Checking $share on LIF $lif:$ip");
            eval {
                map_share(
                    'conn'     => $client_conn_obj,
                    'share'    => $unc,
                    'drive'    => '',
                    'username' => $Options_hash->{'CIFS_USER'},
                    'password' => $Options_hash->{'CIFS_PASSWD'},
                    'timeout'  => 60,
                    'reuse'    => 1,
                    'delay'    => 60,
                    'attempts' => 1,
                    'tests'    => 'rw',
                );
            };
            if ($@) {
                $Log->comment("map_share_returned: " . Dumper($@));
                $failure = 1;
                last;
            }
        }
        if ($failure and $i <= $max_checks) {
            die script_result(
                "Shares not propagated to all nodes after 1 hour");
        } elsif ($failure) {
            sleep 60;
        } else {
            last;
        }
    }
} ## end sub check_share_propagation

sub create_spn {

    $Log->enter();

    #Adding a spn will use setspn -a

    my ($self, %options) = @_;

    #$Log->comment(Dumper($self));

    $Options_hash = $Options->get_hash();

    # Local Variables
    my ($cifs_server, $spn_name);

    my $defined_lifs     = $self->{'defined_topology'}->{'lifs'};
    my $defined_vservers = $self->{'defined_topology'}->{'vservers'};

    foreach my $vserver (natsort(keys(%{$defined_vservers}))) {

        #Hack from above to get the cifs server name.
        $cifs_server = $self->{'cifs_server_basename'};
        $cifs_server .= $defined_vservers->{$vserver}->{'index'};

        $Log->comment(
            "Getting lif data for the CIFS Server $cifs_server on Vserver $vserver"
        );

        my @vserver_lifs = @{$defined_vservers->{$vserver}->{'lifs'}};

     # make sure the cifs server name and ip match if they do then set the spn
     # on the DC.

        foreach my $lif (@vserver_lifs) {
            my $lif_ip
                = $self->{'defined_topology'}->{'lifs'}->{$vserver}->{$lif}
                ->{'address'};
            $Log->comment("lif ip is $lif_ip");
            my $resolved_address = get_name_by_ip('ip' => $lif_ip);

            if ($resolved_address->{'ERROR'}) {
                die config_result(
                    "IP address $resolved_address does not resolve to a name"
                );
            }

            $resolved_address->{'PTR'} =~ /([^\.]+)\.(.+)/;

            $spn_name = $1;
            $spn_name .= ".$Options_hash->{'CIFS_DOMAIN'}";

            $self->set_spn(
                'cifs_server' => $cifs_server,
                'spn_name'    => $spn_name,
            );
        }

    }

    $Log->exit();
} ## end sub create_spn

1;
