#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     NFS Stress Scalability Topology Library
##
## @author      gvg@netapp.com
##
## @pod here

=head1 NAME

Topology::NAS::NFS - NFS Stress Scalability test Topology Library

=head1 SYNOPSIS

    use qw(Topology);
    my $TOPO = new Topology(topology=>"topo1");

=head1 DESCRIPTION

This parent C<Topology::NAS::NFS> class creates and returns appropriate child
Topology object to the caller. This parent class also provides
a few functions for child classes and tests to use. 

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/Topology
http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite

=head1 TODO
* slow Name service fixes
* dup_clients support

=head1 SEE ALSO

TestManager Options TestManagerLite

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Last revised $Date: 2008/05/16 $
    
=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved

=cut

package Topology::NAS::NFS;

use base qw(Topology);

use strict;
use warnings;

use Tharn qw($Log config_result);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Params::NFS;
use RemoteExecution;

use Data::Dumper qw(Dumper);

use Options;
my $Options;
my $Options_hash;

###################################################################
# Notes:
# 1. This library should set the following test_options
#    VOLPATH or VOLPATHS_FILE
#    SRCIP (Optional)
###################################################################

$Options = new Options;
$Options->parameters(
    definitions => [
        {name        => "NFS_VERSION",
         description => "NFS version.",
         required    => "Specify NFS_VERSION",},
        {name        => "PROTOCOL",
         description => "NFS Protocol",
         required    => "Specify PROTOCOL for NFS_VERSION",},
        {name => "PERSISTENT_EXPORT",
         description =>
           "Whether or not exports should be persistent across reboots/to/gb",
         default => 1,
         spec    => "boolean",},
        {name    => "DIR_PER_INSTANCE",
         spec    => "boolean",
         default => 1,
         description =>
           "Whether or not a directory to be created for each test instance to mount exclusively for read/write/create operation ",
        },
        {name        => "NUM_EXPORTS",
         spec        => "s",
         default     => 1,
         description => "Number of exports to be created ",},
        {name    => "INSTANCES_PER_CLIENT",
         spec    => "s",
         default => 1,
         description =>
           "the maximum number of test instances per real client. Each nate script (test instance) takes about 14-19mb approx, if implemented efficiently. This number will not be exceeded on per client. If other config causes conflict, then error should be reported",
        },
        {name    => "MAX_NFS_CLIENTS",
         spec    => "s",
         default => undef,
         description =>
           "The maximum number of NFS clients to be used. Each instance is NFS client, hence INSTANCES_PER_CLIENT will be adjusted accordingly, if CLIENT*INSTANCES_PER_CLIENT > MAX_NFS_CLIENTS. In the event, CLIENT*INSTANCES_PER_CLIENT < MAX_NFS_CLIENTS, error will be reported",
        },
        {name => "DUP_CLIENTS",
         description =>
           "Number of duplicate client instances run per real client or synthetic client ip address",
         default => 1,
         spec    => "s",},
        {name    => "SLOW_NAME_SERVICE",
         spec    => "boolean",
         default => "0",
         dependencies =>
           ["DNS_SERVER", "NIS_SERVER", "PACKET_LOSS_RATE", "PACKET_DELAY"],
         description =>
           "Uses ipfw command on the filer to simulate slow name service by dropping name service requests",
        },
        {name    => "DNS_SERVER",
         spec    => "s",
         default => undef,
         description =>
           "IP Address of DNS Server. This is required when SLOW_NAME_SERVICE is set",
        },
        {name    => "NIS_SERVER",
         spec    => "s@",
         default => undef,
         description =>
           "IP Addresses of NIS Server. This is required when SLOW_NAME_SERVICE is set",
        },
        {name         => "PACKET_LOSS_RATE",
         description  => "Packets loss rate to simulate slow name service",
         default      => undef,
         dependencies => ["SLOW_NAME_SERVICE", "DNS_SERVER", "NIS_SERVER"],},
        {name => "PACKET_DELAY",
         description =>
           "Packet delay in milli seconds to simulate slow name service.",
         default      => undef,
         dependencies => ["SLOW_NAME_SERVICE", "DNS_SERVER", "NIS_SERVER"],},]);

sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $self     = {@_,};
    $Log->enter();
    bless($self, $class);

    $Log->exit();
    return $self;
} ## end sub new

########### A few common functions for sub classes and tests to use #####

sub create_dirs {
    my ($self, $filerip, $clients, $dirsref) = @_;

    my $Remote     = new RemoteExecution;
    my $client_idx = 0;
    my ($client, $mntpnt);
    my $num_subs = 0;
    foreach my $volume (keys %$dirsref) {
        my @directories = @{$dirsref->{$volume}};

        # Pick a number 500 directories per client as optimal value
        while (my @dirs = splice(@directories, 0, 500)) {
            if ($client_idx >= scalar(@$clients)) {
                $client_idx = 0;
            }
            $client = $clients->[$client_idx];

            $mntpnt =
              "/t/${filerip}_${volume}_${client}_${client_idx}_filer_setup";

            # create directories on the client
            my $params = {
                DIRS => join(',', @dirs),
                NFS_VERSION => 3,    # Force V3 so that all clients will work
                MOUNT_LIST => "${filerip};/${volume};${mntpnt};rw;777",};
            $Log->comment(
                "Run tools/TestManager/filer_directory_setup.thpl on $volume from $client to create dirs: "
                  . join(",", @dirs));
            my $SubTst =
              $Remote->execute(
                host        => $client,
                nate_params => $params,
                cmd         => "tools/TestManager/filer_directory_setup.thpl",
                runid =>
                  sprintf("%03d_%s_%s", $num_subs, "filer_setup_dirs", $client),
                handler => sub { });
            die config_result("Remote Execute on $client failed: $@")
              if (!$SubTst);
            $client_idx++;
            $num_subs++;
        } ## end while (my @dirs = splice(...
    } ## end foreach my $volume (keys %$dirsref)
    my $fail_result = $Remote->wait(timeout => 300);
    die config_result("dirs creation failed: $fail_result")
      if ($fail_result);

    return 1;
} ## end sub create_dirs

1;
