#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     NFS Stress Scalability 10 Mode Virtual Topology Library
##
## @author      gvg@netapp.com
##
## @pod here

=head1 NAME

Topology::NAS::NFS::10M_Virtual - NFS Stress Scalability test Virtual Topology Library for 10Mode ONTAP

=head1 SYNOPSIS

    See Topology.pm.

=head1 DESCRIPTION

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/Topology

=head1 TODO

=head1 SEE ALSO

Options TestManagerLite Topology

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Last revised $Date: 2008/09/30 $
    
=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved

=cut

package Topology::NAS::NFS::10M_Virtual;
use base qw(Topology::NAS::NFS::10M);

use strict;
use warnings;

use Tharn qw($Log config_result);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Socket;
use Net::IP;
use String::Random;    # To create unique export rules

use Data::Dumper qw(Dumper);

use Options;
my $Options_hash;

my $ifconfig = '/sbin/ifconfig';
my $sudo     = '/usr/software/bin/sudo';

###################################################################

my $Options = new Options;

$Options->parameters(
    definitions => [
        {name    => "SRCIP",
         spec    => "s",
         default => undef,
         description =>
           "The starting IP and number of IPs to be used separated by '/'",},
        {name         => "ADDR_TYPE",
         spec         => "s",
         default      => "inet4",
         valid_values => ["inet4", "inet6"],
         subroutine   => [\&_get_addr_type],
         description =>
           "Whether or not to use IPv4/IPv6 address in case hostname is given.",
        },
        {name         => "DEVICE",
         spec         => "s",
         default      => undef,
         dependencies => ["SRCIP"],
         description =>
           "Network Device of the clients where SRCIP to be configured. By default network interface is selected with some intelligence based on the first digit of SRCIP and the first digit of IPs assigned into the device. If specified, strictly only this DEVICE would be used blindly",
        },
        {name    => "ALLCLIENTS_2_ALLVOLS",
         spec    => "boolean",
         default => 0,
         description =>
           "This variable dictates whether or not all clients should go to all volumes and hence generate traffic evenly on all volumes by all clients. Warning: Unless worker processes mount each path given in VOLPATHS, this option will have no effect.",
        },
        {name => "VOLPATHS_FILE",
         spec => "s",
         description =>
           "file containing the list of volume paths for clients to mount to",
         file_tests => "fr",
         subroutine => [\&_read_volpaths_file],},
        {name        => "DUP_IPS",
         spec        => "s",
         default     => 1,
         description => "The number of duplicate clients per SRCIP",},
        {name        => "ACCESS_ALLOW",
         spec        => "s",
         default     => undef,
         description => "Set exports rule for rw and root",},
        {name => "MULTI_SERVER",
         description =>
           "Whether or not a client should run traffic against single server or multi server. Please see DATA_INTERFACES to specify multiple interface ips. Unless leaf script supports this feature this is no-op",
         default => 1,
         spec    => "s",},]);

sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $self     = {@_,};
    $Log->enter();
    bless($self, $class);

    $Log->exit();
    return $self;
} ## end sub new

sub define_topology {
    my $self = shift;
    $Log->enter();

    # Execute the one from the parent - This is not required since this forms the topology it's own
    # $self->SUPER::define_topology(@_);

    $Options_hash = $Options->get_hash();

    # Options used here
    my $num_exports          = $Options_hash->{NUM_EXPORTS};
    my $num_rules            = $Options_hash->{NUM_EXPORT_RULES};
    my $dup_ips              = $Options_hash->{DUP_IPS};
    my $instances_per_client = $Options_hash->{INSTANCES_PER_CLIENT};
    my $vol_size =
      $Options_hash->{VOL_SIZE};    # VOL_SIZE is considered per instance
    my @volumes         = $self->volumes();
    my $max_nfs_clients = $Options_hash->{MAX_NFS_CLIENTS};
    my $allclis2vols    = $Options_hash->{ALLCLIENTS_2_ALLVOLS};
    $Log->debug("Options: \n" . Dumper($Options_hash));

    # Adjust num_exports to @volumes
    $num_exports =
      ($num_exports < scalar(@volumes)) ? scalar(@volumes) : $num_exports;

    #
    # Spread Test Instances across available physical clients
    #

    my @clients = $self->clients();
    my $total_instances;
    if ($Options_hash->{SRCIP}) {
        my @ips = $self->_create_range($Options->get(option => "SRCIP"));
        if ($max_nfs_clients) {

            # This may mean it's confusing. let's not allow this condition to keep
            # it simple.
            die config_result(
                           "DUP_IPS can not be > 1 when MAX_NFS_CLIENTS is set")
              if ($dup_ips > 1);

            if (scalar(@ips) == $max_nfs_clients) {
                $Log->comment(
                          "MAX_NFS_CLIENTS($max_nfs_clients) is equal to SRCIP("
                            . scalar(@ips)
                            . "). Hence no change is requierd");
            } elsif (scalar(@ips) > $max_nfs_clients) {
                my $given_ips = scalar(@ips);
                @ips = splice(@ips, 0, $max_nfs_clients);  # Remove the remaining SRCIPs
                $Log->warn(
                    "SRCIPs($given_ips) > MAX_NFS_CLIENTS($max_nfs_clients). SRCIPs adjusted to ("
                      . scalar(@ips)
                      . ")");
            } else {
                die config_result("SRCIP("
                    . scalar(@ips)
                    . ") < $max_nfs_clients. Please adjust and run the test again"
                    );
            }
        } else {
            $Log->comment(
                      "MAX_NFS_CLIENTS is not set, hence using SRCIP as it is");
        }
        my $instances     = int((scalar(@ips) * $dup_ips) / scalar(@clients));
        my $remainder_ips = int((scalar(@ips) * $dup_ips) % scalar(@clients));
        if ($remainder_ips && (($instances + 1) > $instances_per_client)) {
            die config_result("SRCIP (" 
                . @ips
                . ") * DUP_IPS ($dup_ips) / CLIENTS ("
                . @clients
                . ") > INSTANCES_PER_CLIENT ($instances_per_client). Please adjust any of these params and rerun the test"
                );
        } elsif ($instances > $instances_per_client) {
            die config_result("SRCIP (" 
                . @ips
                . ") * DUP_IPS ($dup_ips) / CLIENTS ("
                . @clients
                . ") > INSTANCES_PER_CLIENT ($instances_per_client). Please adjust any of these params and rerun the test"
                );
        } else {
            $Log->debug(
                "INSTANCES_PER_CLIENT will be adjusted from $instances_per_client to $instances(+1) "
                );
        }
        $instances_per_client = $instances;

        ## It is not possible to spread DUP_IPS across different clients and hence
        ## it's tricky to divide the IPS per client if DUP_IPS is set.
        ## If this following formula is not met, then fail it asking user to setup
        ## the relevant params correctly.
        if (   ($dup_ips > 1)
            && ($remainder_ips || ($instances_per_client % $dup_ips))) {
            die config_result(
                "DUP_IPS could not be spread across evenly. Formular must be SRCIP "
                  . @ips
                  . " / CLIENT "
                  . @clients
                  . " % DUP_IPS $dup_ips == 0 ");
        }

        my $index       = 0;
        my $dupip_index = 0;
        foreach my $client (@clients) {
            my $instances = $instances_per_client;
            if ($remainder_ips) {
                $instances++;
                $remainder_ips--;
            }

            $self->{client_setup}{$client}{SRCIP} =
              $ips[$index] . "/" . $instances;
            $self->{client_setup}{$client}{DEVICE} = $Options_hash->{DEVICE};
            $self->{client_setup}{$client}{ADDR_TYPE} =
              $Options_hash->{ADDR_TYPE};

            for (my $inst = 1 ; $inst <= $instances ; $inst++) {
                $self->{test_options}{$client}{$inst}{SRCIP} = $ips[$index];

                # Get instances in here, so we can fill in the values in the next block
                $self->{test_options}{$client}{$inst}{VOLPATH} = "";
                $dupip_index++;
                if ($dupip_index >= $dup_ips) {
                    $index++;
                    $dupip_index = 0;
                }
            } ## end for (my $inst = 1 ; $inst...
        } ## end foreach my $client (@clients)
        $total_instances = int(scalar(@ips) * $dup_ips);
    } else {
        $Log->comment("SRCIP is not set");

        die config_result("DUP_IPS can not be > 1 when SRCIP is not set")
          if ($dup_ips > 1);

        my $instances;
        my $remainders;
        if ($max_nfs_clients) {
            if ($instances_per_client * scalar(@clients) < $max_nfs_clients) {
                die config_result(
                        "INSTANCES_PER_CLIENT($instances_per_client) * CLIENT("
                      . scalar(@clients)
                      . ") < MAX_NFS_CLIENTS($max_nfs_clients). Please adjust these variables and run the test."
                      );
            } else {
                $instances  = int($max_nfs_clients / scalar(@clients));
                $remainders = $max_nfs_clients % scalar(@clients);
                $Log->comment(
                    "INSTANCES_PER_CLIENT will be adjusted from $instances_per_client to $instances(+1) "
                    );
            }
        } else {
            $instances  = $instances_per_client;
            $remainders = 0;
        }
        foreach my $client (@clients) {
            my $inst = 0;
            for (; $inst < $instances ; $inst++) {

                # Get instances in here, so we can fill in the values in the next block
                $self->{test_options}{$client}{$inst}{VOLPATH} = "";
                $total_instances++;
            }
            if ($remainders) {
                $self->{test_options}{$client}{$inst}{VOLPATH} = "";
                $remainders--;
                $total_instances++;
            }
        } ## end foreach my $client (@clients)
    } ## end else [ if ($Options_hash->{SRCIP...

    #
    # Spread volume/dir layout across test instances
    #

    my $size = $self->size_in_mb(size => $vol_size);
    $size =
      ($size) ? $size : 30;    # Set 30MB per instance, in case if none is set.

    if ($allclis2vols) {

        # Directories should be created on all volumes for all clients.
        $Log->comment("All clients will go through all volumes");
        foreach my $volume (@volumes) {
            foreach my $client (@clients) {
                foreach my $inst (keys %{$self->{test_options}{$client}}) {
                    my $dir = "${client}_${inst}";
                    push @{$self->{filer_setup}{dirs}{$volume}}, $dir;
                    push @{$self->{test_options}{$client}{$inst}{VOLPATHS}},
                      "/${volume}/${dir}";
                }
            }
            $self->{filer_setup}{volume}{$volume}{name} = $volume;
            $self->{filer_setup}{volume}{$volume}{size} =
              $size * $total_instances;
        } ## end foreach my $volume (@volumes)
    } ## end if ($allclis2vols)

    my $instances_per_volume = int($total_instances / scalar(@volumes));
    my $remainder_vols       = $total_instances % scalar(@volumes);
    if ($instances_per_volume < 1 && (!$allclis2vols)) {
        die config_result(
            "The number of test instances $total_instances are given lesser than the volumes "
              . scalar(@volumes)
              . ". Please verify your options");
    }
    my $volume;
    my $instances = 0;
    my @vols      = @volumes;
    foreach my $client (@clients) {
        foreach my $inst (keys %{$self->{test_options}{$client}}) {
            if (!$instances) {
                die config_result(
                    "${client}_{$inst} couldn't find volume to run on. Some config/logic issue"
                    ) if (!($volume = pop @vols));
                $instances = $instances_per_volume;
                if ($remainder_vols) {
                    $instances++;
                    $remainder_vols--;
                }
                $self->{filer_setup}{volume}{$volume}{name} = $volume
                  if (!$allclis2vols);
                $self->{filer_setup}{volume}{$volume}{size} = $size * $instances
                  if (!$allclis2vols);
            } ## end if (!$instances)

            my $dir = "${client}_${inst}";
            push @{$self->{filer_setup}{dirs}{$volume}}, $dir
              if (!$allclis2vols);
            $self->{test_options}{$client}{$inst}{VOLPATH} =
              "/${volume}/${dir}";

            $instances--;
        } ## end foreach my $inst (keys %{$self...
    } ## end foreach my $client (@clients)

    #
    # Export Layout for all test instances
    #

    if ($num_exports > scalar(@volumes)) {
        die config_result("NUM_EXPORTS > NUM_VOLS. Failed topology");
    }

    my $RS = new String::Random;
    my @rulesets;
    for (my $i = 0 ; $i < $num_exports ; $i++) {
        my $rule_index = 1;
        if ($Options_hash->{ACCESS_ALLOW}) {

            push(@{$self->{filer_setup}{exportruleset}},
                 {vserver     => $Options_hash->{VSERVER},
                  rulesetname => $self->testcase_string()."_custom_" . $i,});

            for (; $rule_index < $num_rules ; $rule_index++) {
                push(@{$self->{filer_setup}{exportruleset_rules}},
                     {ruleindex   => $rule_index,
                      vserver     => $Options_hash->{VSERVER},
                      rulesetname => $self->testcase_string()."_custom_" . $i,
                      anon        => 0,
                      protocol    => "any",
                      clientmatch =>
                        $RS->randregex('[a-z][a-z][a-z][a-z][a-z][a-z]'),
                      rorule    => "any",
                      rwrule    => "any",
                      superuser => "any",});
            } ## end for (; $rule_index < $num_rules...

            push(@{$self->{filer_setup}{exportruleset_rules}},
                 {ruleindex   => $rule_index,
                  vserver     => $Options_hash->{VSERVER},
                  rulesetname => $self->testcase_string()."_custom_" . $i,
                  anon        => 0,
                  protocol    => "any",
                  clientmatch => "$Options_hash->{ACCESS_ALLOW}",
                  rorule      => "any",
                  rwrule      => "any",
                  superuser   => "any",});

            push @rulesets, $self->testcase_string()."_custom_" . $i;
        } else {
            if ($i == 0) {

                # This rule is added later
                push @rulesets, $self->testcase_string()."_allow_all";
                next;
            }

            push(@{$self->{filer_setup}{exportruleset}},
                 {vserver     => $Options_hash->{VSERVER},
                  rulesetname => $self->testcase_string()."_allow_all_" . $i,});

            for (; $rule_index < $num_rules ; $rule_index++) {
                push(@{$self->{filer_setup}{exportruleset_rules}},
                     {ruleindex   => $rule_index,
                      vserver     => $Options_hash->{VSERVER},
                      rulesetname => $self->testcase_string()."_allow_all_" . $i,
                      anon        => 0,
                      protocol    => "any",
                      clientmatch =>
                        $RS->randregex('[a-z][a-z][a-z][a-z][a-z][a-z]'),
                      rorule    => "any",
                      rwrule    => "any",
                      superuser => "any",});
            } ## end for (; $rule_index < $num_rules...

            push(@{$self->{filer_setup}{exportruleset_rules}},
                 {ruleindex   => 1,
                  vserver     => $Options_hash->{VSERVER},
                  rulesetname => $self->testcase_string()."_allow_all_" . $i,
                  anon        => 0,
                  protocol    => "any",
                  clientmatch => "0.0.0.0/0",
                  rorule      => "any",
                  rwrule      => "any",
                  superuser   => "any",});

            push @rulesets, $self->testcase_string()."_allow_all_" . $i;
        } ## end else [ if ($Options_hash->{ACCESS_ALLOW...
    } ## end for (my $i = 0 ; $i < $num_exports...

    my @rules;
    foreach my $volume (@volumes) {
        @rules = @rulesets if (!@rules);
        my $ruleset = pop @rules;
        my $opts = {vserver => $Options_hash->{VSERVER},
                    policy  => $ruleset,
                    volume  => $volume};
        $self->{filer_setup}{exports}{$volume} = $opts;
    }

    # Create export rule $self->testcase_string()."_allow_all" created.
    push(@{$self->{filer_setup}{exportruleset}},
         {vserver     => $Options_hash->{VSERVER},
          rulesetname => $self->testcase_string()."_allow_all",});

    push(@{$self->{filer_setup}{exportruleset_rules}},
         {ruleindex   => 1,
          vserver     => $Options_hash->{VSERVER},
          rulesetname => $self->testcase_string()."_allow_all",
          anon        => 0,
          protocol    => "any",
          clientmatch => "0.0.0.0/0",
          rorule      => "any",
          rwrule      => "any",
          superuser   => "any",});

    $self->{_num_instances} = $total_instances;

    $Log->debug("Topology: \n" . Dumper($self));
} ## end sub define_topology

sub filer_setup {
    my ($self, %options) = shift;

    # Execute the one from the parent
    $self->SUPER::filer_setup(%options);

    $Log->exit();
}

sub filer_cleanup {
    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::filer_cleanup(%options);

    $Log->exit();
}

sub client_setup {
    my ($self, %options) = @_;
    $Log->enter();

    # Execute the one from the parent
    $self->SUPER::client_setup(%options);

    if (!$options{SRCIP}) {
        $Log->comment("SRCIP is not set, no IPs need to be configured");
        return;
    }

    my @ips       = $self->_create_range($options{SRCIP});
    my $first_vip = $ips[0];
    my $last_vip  = $ips[-1];

    my ($device, $prefix) =
      $self->_pick_network_info(first_ip  => $first_vip,
                                last_ip   => $last_vip,
                                addr_type => $options{ADDR_TYPE},
                                device    => $options{DEVICE});

    $self->_setup_ips(ips       => \@ips,
                      device    => $device,
                      addr_type => $options{ADDR_TYPE},
                      prefix    => $prefix);

    $Log->exit();
} ## end sub client_setup

#-------------------------------------------------------------------------------
# Destructor for object
# It removes all virtual ip address interfaces that this object created
#-------------------------------------------------------------------------------
sub DESTROY {
    my $self = shift;
    local ($@);

    $self->client_cleanup();
}

sub client_cleanup {
    my ($self, %options) = @_;
    $Log->enter();

    $self->_cleanup_ips();

    # Execute the one from the parent
    $self->SUPER::client_cleanup(%options);

    $Log->exit();
} ## end sub client_cleanup

######################### Internal Routines ##################

sub _create_range {
    my ($self, $ip) = @_;
    $Log->enter();
    my ($begin, $range);
    if ($ip =~ /(.+)\/(\d+)/) {
        ($begin, $range) = ($1, $2);
    } elsif (!defined($ip)) {
        return ();
    } else {
        $begin = $ip;
        $range = 1;
    }
    $range--;

    my @range;
    ## This works for both IPv4 and IPv6
    my $ipobj = new Net::IP("$begin + $range")
      or die "Failed to create ip object: $@";

    do {
        push(@range, $ipobj->ip());
    } while (++$ipobj);

    $Log->exit();
    return (wantarray) ? @range : $range[$#range];
} ## end sub _create_range

sub _get_addr_type {
    my $addr_type = shift;
    $Log->enter("addr_type = $addr_type");
    if ($Options->get(option => "SRCIP") =~ /(.+)\/\d+/) {
        $addr_type = Net::IP::ip_is_ipv6($1) ? "inet6" : "inet4";
    } else {
        $addr_type =
          Net::IP::ip_is_ipv6(@{$Options->get(option => "DATA_INTERFACES")}[0])
          ? "inet6"
          : "inet4";
    }
    $Options->set(ADDR_TYPE => $addr_type);
    $Log->exit("addr_type = $addr_type");
} ## end sub _get_addr_type

sub _pick_network_info {
    my $self     = shift;
    my (%opts)   = @_;
    my $OS       = qx($sudo uname 2>&1);
    my $addrtype = ($opts{addr_type} eq "inet6") ? "inet6" : "inet";
    my ($EIP, $GFIP, $GLIP);
    $Log->enter(Dumper(\%opts));

    my ($device, $netmask);    # to be returned.
    my $interface;
    if ($OS =~ /SunOS/i) {

        my $out = qx($sudo $ifconfig -a $addrtype 2>&1);
        die config_result("Could not get interface information: $out") if $?;

        if ($opts{addr_type} eq "inet4") {

            # Primary IPv4 address is expected to be in primary interface,
            # hence look for only primary interfaces to pick appropriate interface
            foreach (split /\n/, $out) {

                # Get interface name
                next if (/^\S+:\d+:/);    # Ignore all logical interfaces
                $interface = $1 if (/^(\S+):/);

                next if ($opts{device} && $opts{device} ne $interface);

                # Get the ip address and mask for interface
                my ($addr, $mask) =
                  (/inet\s+(\S+)\s+netmask\s+(\S+)\s+broadcast\s+(\S+)/);
                next if ($addr =~ /127\.0\.0\.1/);    # Ignore if loopback
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    # SunOS prints mask in hex format, convert it to dotted decimal format
                    # to use it in inet_aton.
                    $mask =
                        hex(substr($mask, 0, 2)) . "."
                      . hex(substr($mask, 2, 2)) . "."
                      . hex(substr($mask, 4, 2)) . "."
                      . hex(substr($mask, 6, 2));

                    # Get the network part of the interface
                    my $net_mask = inet_aton($mask);
                    my $net_ip   = inet_aton($addr) & $net_mask;

                    # Check if this is a valid interface for the requested range
                    if (   (inet_aton($opts{first_ip}) & $net_mask) eq $net_ip
                        && (inet_aton($opts{last_ip}) & $net_mask) eq $net_ip) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                } ## end if ($addr && $mask)
            } ## end foreach (split /\n/, $out)
        } else {

            # Primary IPv6 address should be on one of the logical interface
            foreach (split /\n/, $out) {

                # Get interface name
                $interface = $1 if (/^(\S+):/);

                # Save only primary interface
                $interface = (split(":", $interface))[0];
                next if ($opts{device} && $opts{device} ne $interface);

                # Get the ip address and prefix for interface
                my ($addr, $mask) = (/inet6\s+(\S+)\/(\S+)/);

                # Ignore link and loopback
                next if ($addr =~ /^::1/ || $addr =~ /^fe80::/);
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    $EIP = $self->_get_inet6_iprange_obj($addr . "/" . $mask);
                    $GFIP =
                      $self->_get_inet6_iprange_obj(
                                                 $opts{first_ip} . "/" . $mask);
                    $GLIP =
                      $self->_get_inet6_iprange_obj(
                                                  $opts{last_ip} . "/" . $mask);

                    if ((   $EIP->overlaps($GFIP) == $IP_B_IN_A_OVERLAP
                         || $EIP->overlaps($GFIP) == $IP_IDENTICAL)
                        && (   $EIP->overlaps($GLIP) == $IP_B_IN_A_OVERLAP
                            || $EIP->overlaps($GLIP) == $IP_IDENTICAL)
                      ) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                } ## end if ($addr && $mask)
            } ## end foreach (split /\n/, $out)
        } ## end else [ if ($opts{addr_type} eq...
    } elsif ($OS =~ /Linux/i) {

        # Get interface name, ip address and netmask
        my $out = qx($sudo $ifconfig -a 2>&1);
        die config_result("Could not get interface information: $out") if $?;

        foreach (split /\n/, $out) {

            # Get interface name
            next if (/^\S+:\d+/);    # Ignore all logical interfaces
            $interface = $1 if (/^(\S+)/);
            next if ($opts{device} && $opts{device} ne $interface);

            if ($opts{addr_type} eq "inet4") {

                # Get the ip address and mask for interface
                my ($addr, $mask) =
                  (/inet addr:(\S+)\s+Bcast:\S+\s+Mask:(\S+)/);
                next
                  if ($addr && ($addr =~ /127\.0\.0\.1/));  # Ignore if loopback
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    # Get the network part of the interface
                    my $net_mask = inet_aton($mask);
                    my $net_ip   = inet_aton($addr) & $net_mask;

                    # Check if this is a valid interface for the requested range
                    if (   (inet_aton($opts{first_ip}) & $net_mask) eq $net_ip
                        && (inet_aton($opts{last_ip}) & $net_mask) eq $net_ip) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                } ## end if ($addr && $mask)
            } else {

                # Get the ip address and prefix for interface
                my ($addr, $mask) = (/inet6\s+addr:\s*(\S+)\/(\S+)/);
                next
                  if ($addr =~ /^::1/ || $addr =~ /^fe80::/)
                  ;    # Ignore link and loopback
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    $EIP = $self->_get_inet6_iprange_obj($addr . "/" . $mask);
                    $GFIP =
                      $self->_get_inet6_iprange_obj(
                                                 $opts{first_ip} . "/" . $mask);
                    $GLIP =
                      $self->_get_inet6_iprange_obj(
                                                  $opts{last_ip} . "/" . $mask);

                    if ((   $EIP->overlaps($GFIP) == $IP_B_IN_A_OVERLAP
                         || $EIP->overlaps($GFIP) == $IP_IDENTICAL)
                        && (   $EIP->overlaps($GLIP) == $IP_B_IN_A_OVERLAP
                            || $EIP->overlaps($GLIP) == $IP_IDENTICAL)
                      ) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                } ## end if ($addr && $mask)
            } ## end else [ if ($opts{addr_type} eq...
        } ## end foreach (split /\n/, $out)
    } else {
        die config_result("$OS is not supported by this topology yet");
    }

  FOUND:
    if (!defined($device)) {
        die config_result("Could not find proper interface for virtual ip range"
                          . " [$opts{first_ip}, $opts{last_ip}]");
    }

    $Log->exit("Selected: $device, $netmask");
    return ($device, $netmask);
} ## end sub _pick_network_info

sub _get_inet6_iprange_obj {
    my $self    = shift;
    my $iprange = shift;
    $Log->enter("iprange = $iprange");

    my ($ip, $prefix) = Net::IP::ip_splitprefix($iprange);
    my $octets = int($prefix / 16);
    my @octs = split(":", $ip);
    delete $octs[-1] while ($octets--);
    $ip = join(":", @octs) . "::/" . $prefix;
    my $ipobj = new Net::IP($ip)
      or die config_result("Net::IP failed for $ip");

    $Log->exit("returning Net::IP Object for $ip");
    return $ipobj;
} ## end sub _get_inet6_iprange_obj

sub _setup_ips {
    my $self = shift;
    my (%opts) = @_;
    $Log->enter();

    my ($cmd, $cleanupcmd);
    my $OS = qx($sudo uname 2>&1);

    # Let's assume no one wants to use logical interface on this system
    my $index = 0;
    foreach my $ip (@{$opts{ips}}) {
        if ($OS =~ /SunOS/i) {
            if ($opts{addr_type} eq "inet6") {
                $cmd =
                  "$sudo $ifconfig $opts{device} inet6 addif $ip/$opts{prefix} up";
                $cleanupcmd =
                  "$sudo $ifconfig $opts{device} inet6 removeif $ip";
            } else {
                $cmd =
                  "$sudo $ifconfig $opts{device} inet addif $ip netmask $opts{prefix} up";
                $cleanupcmd = "$sudo $ifconfig $opts{device} inet removeif $ip";
            }
        } elsif ($OS =~ /Linux/i) {
            if ($opts{addr_type} eq "inet6") {
                $cmd =
                  "$sudo $ifconfig $opts{device} inet6 add $ip/$opts{prefix} up";
                $cleanupcmd =
                  "$sudo $ifconfig $opts{device} inet6 del $ip/$opts{prefix}";
            } else {
                $cmd =
                  "$sudo $ifconfig $opts{device}:$index $ip netmask $opts{prefix} up";
                $cleanupcmd = "$sudo $ifconfig $opts{device}:$index down";
                $index++;
            }
        } else {
            die config_result("$OS is not supported by this topology yet");
        }

        my $output = qx($cmd 2>&1);
        die config_result("$cmd failed: $output") if $?;
        $Log->comment("$cmd : $output");

        push(@{$self->{ip_cleanupcmds}}, $cleanupcmd);
    } ## end foreach my $ip (@{$opts{ips...

    $Log->comment("IP Setup Complete : " . qx($sudo $ifconfig -a 2>&1));

    $self->{vdev_cleanup} = $opts{device};
    $self->{ip_setup_id}  = $$;             # Save this for cleanup to take care
    $Log->exit();
} ## end sub _setup_ips

sub _cleanup_ips {
    my $self = shift;
    $Log->enter();

    if ($self->{ip_setup_id} && ($self->{ip_setup_id} != $$)) {
        $Log->debug(
              "process $$ can not cleanup ips created by $self->{ip_setup_id}");
        return;
    }

    while (my $cmd = pop @{$self->{ip_cleanupcmds}}) {
        my $output = qx($cmd 2>&1);
        $Log->warn("$cmd failed: $output") if $?;
        $Log->comment("$cmd : $output");
    }

    $Log->exit();
} ## end sub _cleanup_ips

1;
