#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Topology/NAS/NFS/7M_Flexcache.pm#1 $

## @summary     NFS Stress Scalability 7Mode Topology Library
##
## @author      gvg@netapp.com
##
## @pod here

=head1 NAME

Topology::NAS::NFS::7M_Flexcache - NFS Stress Scalability test Topology Library for 7Mode ONTAP

=head1 SYNOPSIS

    use qw(Topology);
    my $topology = new Topology(topology=>"Topology::NAS::NFS::7M_Flexcache");

=head1 DESCRIPTION

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/Topology

=head1 TODO

=head1 SEE ALSO

Options TestManagerLite Topology

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Kha Nguyen (knguyenn@netapp.com)
  Samy Gv (gvg@netapp.com)   
    
=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved

=cut

package Topology::NAS::NFS::7M_Flexcache;
use base qw(Topology::NAS::NFS);

use strict;
use warnings;

use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Server;
use nfs::GenericExport qw(export_path);
use Utils qw(make_conn);
use SayWrap qw(sayEOF);
use Tharn qw($Log config_result);
use filer qw(filer_options_set);
use Dense qw(dense_start dense_toggle);
use Socket;
use Net::IP;
use String::Random;    # To create unique export rules

use Data::Dumper qw(Dumper);

use Options;
my $Options;
my $Options_hash;

my $ifconfig = '/sbin/ifconfig';
my $sudo     = '/usr/software/bin/sudo';

use RemoteExecution;
my $Remote = new RemoteExecution;

my $Server;

###################################################################

$Options = new Options;
$Options->parameters(
    definitions => [
        {name        => "CACHE_TRAFFIC",
         description => "Should traffic be routed via the FlexCache?",
         default     => "1",
         spec        => "boolean",},
        {name => "CLUSTER_TRAFFICS",
         description =>
           "Pick data interfaces that io traffic should be routed to",
         default      => "mixed",
         valid_values => ["local", "remote", "mixed"],
         spec         => "s",},
        {name        => "TESTUSER",
         spec        => "s@",
         description => "List of test users to use for the testing",
         default     => undef,},
        {name        => "TESTUSER_FILE",
         spec        => "s",
         description => "file containing the list of testusers who test run as",
         file_tests  => "fr",
         subroutine  => [\&_read_testuser_file],},
        {name        => "SIS_ENABLE",
         spec        => "s",
         description => "Param that is used to enable the sis volumes",
         default     => "0",},
        {name    => "SRCIP",
         spec    => "s",
         default => undef,
         description =>
           "The starting IP and number of IPs to be used separated by '/'",},
        {name         => "ADDR_TYPE",
         spec         => "s",
         default      => "inet4",
         valid_values => ["inet4", "inet6"],
         description =>
           "Whether or not to use IPv4/IPv6 address in case hostname is given.",
         subroutine => [\&_get_addr_type],},
        {name         => "DEVICE",
         spec         => "s",
         default      => undef,
         dependencies => ["SRCIP"],
         description =>
           "Network Device of the clients where SRCIP to be configured. By default network interface is selected with some intelligence based on the first digit of SRCIP           and the first digit of IPs assigned into the device. If specified, strictly only this DEVICE would be used blindly",
        },
        {name        => "DUP_IPS",
         spec        => "s",
         default     => 1,
         description => "The number of duplicate clients per SRCIP",},
        {name        => "ACCESS_ALLOW",
         spec        => "s",
         default     => undef,
         description => "Set exports rule for rw and root",},
        {name        => "ACCESS_UNIQUE",
         spec        => "boolean",
         default     => 1,
         description => "Whether or not to set export rules unique",},
        {name => "MULTI_SERVER",
         description =>
           "Whether or not a client should run traffic against single server or multi server. Please see DATA_INTERFACES to specify multiple interface ips. Unless leaf s            cript supports this feature this is no-op",
         default => 1,
         spec    => "s",},
        {name => "VOLPATHS_FILE",
         spec => "s",
         description =>
           "file containing the list of volume paths for clients to mount to",
         file_tests => "fr",
         subroutine => [\&_read_volpaths_file],},]);

sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $self     = {@_,};
    $Log->enter();
    bless($self, $class);

    $Log->exit();
    return $self;
} ## end sub new

sub define_topology {
    my $self = shift;
    $Log->enter();

    # Execute the one from the parent
    $self->SUPER::define_topology(@_);

    $Options_hash = $Options->get_hash();

    my @filers = $self->filers();
    if (scalar(@filers) != 2) {
        die config_result(
            "This Topology requires at least two FILERs, one as origin and other as Cache. Given: @filers"
            );
    }

    # Options used here
    my $num_exports          = $Options_hash->{NUM_EXPORTS};
    my $dup_ips              = $Options_hash->{DUP_IPS};
    my $instances_per_client = $Options_hash->{INSTANCES_PER_CLIENT};
    my $dir_per_instance     = $Options_hash->{DIR_PER_INSTANCE};
    my $vol_size =
      $Options_hash->{VOL_SIZE};    # VOL_SIZE is considered per instance
    my @volumes = $self->volumes();
    my $export_options =
      ($Options_hash->{EXPORT_OPTIONS})
      ? $Options_hash->{EXPORT_OPTIONS}
      : "rw,anon=0";
    my $max_nfs_clients = $Options_hash->{MAX_NFS_CLIENTS};
    $Log->comment("Options: \n" . Dumper($Options_hash));

    # Adjust num_exports to @volumes
    $num_exports =
      ($num_exports < scalar(@volumes)) ? scalar(@volumes) : $num_exports;

    #
    # Spread Test Instances across available physical clients
    #

    my @clients = $self->clients();
    my $total_instances;

    my $instances;
    my $remainders;
    if ($Options_hash->{SRCIP}) {
        my @ips = $self->_create_range($Options->get(option => "SRCIP"));
        if ($max_nfs_clients) {

            # This may mean it's confusing. let's not allow this condition to keep
            # it simple.
            die config_result(
                           "DUP_IPS can not be > 1 when MAX_NFS_CLIENTS is set")
              if ($dup_ips > 1);

            if (scalar(@ips) == $max_nfs_clients) {
                $Log->comment(
                          "MAX_NFS_CLIENTS($max_nfs_clients) is equal to SRCIP("
                            . scalar(@ips)
                            . "). Hence no change is requierd");
            } elsif (scalar(@ips) > $max_nfs_clients) {
                my $given_ips = scalar(@ips);
                my $offset    = $given_ips - $max_nfs_clients;
                splice(@ips, $offset);    # Remove the remaining SRCIPs
                $Log->warn(
                    "SRCIPs($given_ips) > MAX_NFS_CLIENTS($max_nfs_clients). SRCIPs adjusted to ("
                      . scalar(@ips)
                      . ")");
            } else {
                die config_result("SRCIP("
                    . scalar(@ips)
                    . ") < $max_nfs_clients. Please adjust and run the test again"
                    );
            }
        } else {
            $Log->comment(
                      "MAX_NFS_CLIENTS is not set, hence using SRCIP as it is");
        }
        my $instances     = int((scalar(@ips) * $dup_ips) / scalar(@clients));
        my $remainder_ips = int((scalar(@ips) * $dup_ips) % scalar(@clients));
        if ($remainder_ips && (($instances + 1) > $instances_per_client)) {
            die config_result("SRCIP (" 
                . @ips
                . ") * DUP_IPS ($dup_ips) / CLIENTS ("
                . @clients
                . ") > INSTANCES_PER_CLIENT ($instances_per_client). Please adjust any of these params and rerun the test"
                );
        } elsif ($instances > $instances_per_client) {
            die config_result("SRCIP (" 
                . @ips
                . ") * DUP_IPS ($dup_ips) / CLIENTS ("
                . @clients
                . ") > INSTANCES_PER_CLIENT ($instances_per_client). Please adjust any of these params and rerun the test"
                );
        } else {
            $Log->debug(
                "INSTANCES_PER_CLIENT will be adjusted from $instances_per_client to $instances(+1) "
                );
        }
        $instances_per_client = $instances;

        ## It is not possible to spread DUP_IPS across different clients and hence
        ## it's tricky to divide the IPS per client if DUP_IPS is set.
        ## If this following formula is not met, then fail it asking user to setup
        ## the relevant params correctly.
        if (   ($dup_ips > 1)
            && ($remainder_ips || ($instances_per_client % $dup_ips))) {
            die config_result(
                "DUP_IPS could not be spread across evenly. Formular must be SRCIP "
                  . @ips
                  . " / CLIENT "
                  . @clients
                  . " % DUP_IPS $dup_ips == 0 ");
        }

        my $index       = 0;
        my $dupip_index = 0;
        foreach my $client (@clients) {
            my $instances = $instances_per_client;
            if ($remainder_ips) {
                $instances++;
                $remainder_ips--;
            }

            $self->{client_setup}{$client}{SRCIP} =
              $ips[$index] . "/" . $instances;
            $self->{client_setup}{$client}{DEVICE} = $Options_hash->{DEVICE};
            $self->{client_setup}{$client}{ADDR_TYPE} =
              $Options_hash->{ADDR_TYPE};

            for (my $inst = 0 ; $inst < $instances ; $inst++) {
                $self->{test_options}{$client}{$inst}{SRCIP} = $ips[$index];

                # Get instances in here, so we can fill in the values in the next block
                $self->{test_options}{$client}{$inst}{VOLPATH} = "";
                $dupip_index++;
                if ($dupip_index >= $dup_ips) {
                    $index++;
                    $dupip_index = 0;
                }
            } ## end for (my $inst = 0 ; $inst...
        } ## end foreach my $client (@clients)
        $total_instances = int(scalar(@ips) * $dup_ips);
    } else {
        $Log->comment("SRCIP is not set");

        die config_result("DUP_IPS can not be > 1 when SRCIP is not set")
          if ($dup_ips > 1);

        my $instances;
        my $remainders;
        if ($max_nfs_clients) {
            if ($instances_per_client * scalar(@clients) < $max_nfs_clients) {
                die config_result(
                        "INSTANCES_PER_CLIENT($instances_per_client) * CLIENT("
                      . scalar(@clients)
                      . ") < MAX_NFS_CLIENTS($max_nfs_clients). Please adjust these variables and run the test."
                      );
            } else {
                $instances  = int($max_nfs_clients / scalar(@clients));
                $remainders = $max_nfs_clients % scalar(@clients);
                $Log->comment(
                    "INSTANCES_PER_CLIENT will be adjusted from $instances_per_client to $instances(+1) "
                    );
            }
        } else {
            $instances  = $instances_per_client;
            $remainders = 0;
        }
        foreach my $client (@clients) {
            my $inst = 0;
            for (; $inst < $instances ; $inst++) {

                # Get instances in here, so we can fill in the values in the next block
                $self->{test_options}{$client}{$inst}{VOLPATH} = "";
                $total_instances++;
            }
            if ($remainders) {
                $self->{test_options}{$client}{$inst}{VOLPATH} = "";
                $remainders--;
                $total_instances++;
            }
        } ## end foreach my $client (@clients)
    } ## end else [ if ($Options_hash->{SRCIP...

    #
    # Spread volume/dir layout across test instances
    #

    my $size = $self->size_in_mb(size => $vol_size);
    $size =
      ($size) ? $size : 30;    # Set 30MB per instance, in case if none is set.

    my $instances_per_volume = int($total_instances / scalar(@volumes));
    my $remainder_vols       = $total_instances % scalar(@volumes);
    if ($instances_per_volume < 1) {
        die config_result(
            "The number of test instances $total_instances are given lesser than the volumes "
              . scalar(@volumes)
              . ". Please verify your options");
    }

    my @vols = @volumes;
    $Log->debug("Setting up vol options for : @vols");
    my $volume;
    $instances = 0;

    foreach my $client (@clients) {
        foreach my $inst (keys %{$self->{test_options}{$client}}) {
            if (!$instances && !($volume = pop @vols)) {
                die config_result(
                    "${client}_{$inst} couldn't find volume to run on. Some config/logic issue"
                    );
            }

            if (!$instances) {
                $instances = $instances_per_volume;
                if ($remainder_vols) {
                    $instances++;
                    $remainder_vols--;
                }
                $self->{filer_setup}{volume}{$volume}{name} = $volume;
                $self->{filer_setup}{volume}{$volume}{size} =
                  $size * $instances;
            } ## end if (!$instances)

            if ($dir_per_instance) {
                my $dir = "${client}_${inst}";
                push @{$self->{filer_setup}{dirs}{$volume}}, $dir;
                $self->{test_options}{$client}{$inst}{VOLPATH} =
                  "/vol/${volume}/${dir}";
            } else {
                $self->{test_options}{$client}{$inst}{VOLPATH} =
                  "/vol/${volume}";
            }

            $instances--;
        } ## end foreach my $inst (keys %{$self...
    } ## end foreach my $client (@clients)

    #
    # Export Layout for all test instances
    #
    my $access = $Options_hash->{ACCESS_ALLOW};
    my $unique = $Options_hash->{ACCESS_UNIQUE};

    if ($num_exports == $total_instances) {

        # Number of test instances == number of exports. This means each instance will run
        # against exclusive directory/volume with exclusive export
        $Log->comment("An export per test instance detected");

        foreach my $client (@clients) {
            foreach my $inst (keys %{$self->{test_options}{$client}}) {
                my $volpath = $self->{test_options}{$client}{$inst}{VOLPATH};
                $self->{filer_setup}{exports}{$volpath}{options} =
                  $self->_export_options($access, $unique);
            }
        }
    } elsif ($num_exports < $total_instances) {
        if ($num_exports == scalar(@volumes)) {

            # Export each volume separately
            $Log->comment(
                "An export per volume is detected with export_options $export_options"
                );
            foreach my $volume (@volumes) {
                my $export = "/vol/" . $volume;
                $self->{filer_setup}{exports}{$export}{options} =
                  $self->_export_options($access, $unique);
            }
        } elsif ($num_exports > scalar(@volumes)) {

            # ??? - GV
            # Not sure if there are any other better ways to handle this. For now, export all volumes first
            # to make sure all are usable, then export sub directories until reach $num_exports

            my $exports = $num_exports;
            $Log->comment("exports $num_exports > volumes "
                . scalar(@volumes)
                . " detected. Exporting volumes first and then export directories"
                );
            foreach my $volume (@volumes) {
                my $export = "/vol/" . $volume;
                $self->{filer_setup}{exports}{$export}{options} =
                  $self->_export_options($access, $unique);
                $exports--;
            }
            foreach my $client (@clients) {
                foreach my $inst (keys %{$self->{test_options}{$client}}) {
                    my $volpath =
                      $self->{test_options}{$client}{$inst}{VOLPATH};
                    $self->{filer_setup}{exports}{$volpath}{options} =
                      $self->_export_options($access, $unique);
                    $exports--;
                    last if (!$exports);
                }
                last if (!$exports);
            } ## end foreach my $client (@clients)
        } else {
            die config_result(  "This doesn't make sense. NUM_VOLS("
                              . scalar(@volumes)
                              . ") && NUM_EXPORTS ($num_exports)");
        }
    } else {
        die config_result("This doesn't make sense. NUM_VOLS("
             . scalar(@volumes)
             . ") && NUM_EXPORTS ($num_exports) && INSTANCES ($total_instances)"
             );
    }

    # Spread test users across test instances if given
    my $testusers = $Options_hash->{TESTUSER};
    my $index     = 0;
    if (my $num_users = scalar(@$testusers)) {
        foreach my $client (@clients) {
            foreach my $inst (keys %{$self->{test_options}{$client}}) {
                $self->{test_options}{$client}{$inst}{TESTUSER} =
                  $testusers->[$index];
                $index = 0 if (++$index >= $num_users);
            }
        }
    }

    $self->{_num_instances} = $total_instances;

    $Log->comment(Dumper($self));
    $Log->exit();
} ## end sub define_topology

sub filer_setup {
    my ($self, %options) = @_;

    my $filer = ($self->filers())[0];

    $Options_hash = $Options->get_hash();

    # Execute the one from the parent
    $self->SUPER::filer_setup(%options);

    $Server = new Server;

    # Create conn object for the filer
    my $conn = make_conn($filer);

    # Define the base list of command nodes to discover in the cluster
    my @discovery_nodes = qw(aggregate vserver network_interfaces);

    $self->{discovered_topology} =
      $Server->discover_topology(cmd_nodes => \@discovery_nodes);
    $Log->debug("Discovered Topology: " . Dumper($self->{discovered_topology}));

    my $test_topology = $self->_update_test_topology();
    $Server->create_topology(print_node => undef, topology => $test_topology);

    my @clients = $self->clients();

    #Enable Dedupe for Volumes
    if ($Options_hash->{SIS_ENABLE} != 0) {
        my @volumes = $self->volumes();
        foreach my $dedup_vol (@volumes) {
            dense_toggle(host      => $self->{origin},
                         path_name => "/vol/$dedup_vol",
                         state     => "on");
            dense_start(host      => $self->{origin},
                        path_name => "/vol/$dedup_vol");
        }
    } ## end if ($Options_hash->{SIS_ENABLE...

    # Setup DATA_INTERFACES appropriately based on CLUSTER_TRAFFICS variable settings.
    if (grep { !defined($_) } @{$Options_hash->{DATA_INTERFACES}}) {
        my @discovered_vifs =
          @{$self->{discovered_topology}{cluster}{network_interfaces}};
        my (@local_ips, @remote_ips);
        foreach my $vif (@discovered_vifs) {
            if ($vif->{address} =~ /^172/ && $vif->{"status-admin"} eq "up") {
                if ($vif->{"curr-filer"} eq $self->{origin}) {
                    push(@local_ips, $vif->{address});
                    $Log->debug(
                        "Adding local vif '$vif->{address}' to local filer's interfaces list"
                        );
                } else {
                    push(@remote_ips, $vif->{address});
                    $Log->debug(
                        "Adding remote vif '$vif->{address}' to remote filer's interfaces list"
                        );
                }
            } ## end if ($vif->{address} =~...
        } ## end foreach my $vif (@discovered_vifs)
        if ($Options_hash->{CLUSTER_TRAFFICS} =~ "local") {
            $Options->set(DATA_INTERFACES => \@local_ips);
        } elsif ($Options_hash->{CLUSTER_TRAFFICS} =~ "remote") {
            $Options->set(DATA_INTERFACES => \@remote_ips);
        } else {
            push(@local_ips, @remote_ips);
            $Options->set(DATA_INTERFACES => \@local_ips);
        }
    } ## end if (grep { !defined($_...
    $Log->comment("Chosen DATA_INTERFACES are: "
                  . join(",", @{$Options_hash->{DATA_INTERFACES}}));

    # Create required directories
    if (keys %{$self->{filer_setup}{dirs}}) {
        $Log->comment("Directories need to be created on volumes");

        my $Remote     = new RemoteExecution;
        my $filer      = @{$Options_hash->{FILER}}[0];
        my $client_idx = 0;
        my ($client, $mntpnt);
        my $num_subs = 0;
        foreach my $volume (keys %{$self->{filer_setup}{dirs}}) {
            my @directories = @{$self->{filer_setup}{dirs}{$volume}};

            # If volume already exists and exported with restricted permission
            # create_topology() didn't seem to change anything, hence we need
            # override here so dir creation can succeed. This shouldn't be a
            # problem, since we will reexport them anyways.
            export_path($conn, "/vol/$volume", OPTIONS => "rw,anon=0")
              or die config_result("export_path of /vol/$volume failed: $@");

            # Pick a number 500 directories per client as optimal value
            while (my @dirs = splice(@directories, 0, 500)) {
                if ($client_idx >= scalar(@clients)) {
                    $client_idx = 0;
                }
                $client = $clients[$client_idx];

                $mntpnt =
                  "/t/${filer}_${volume}_${client}_${client_idx}_filer_setup";

                # create directories on the client
                my $params = {
                    DIRS => join(',', @dirs),
                    MOUNT_LIST => "${filer};/vol/${volume};${mntpnt};rw;777",};
                $Log->comment(
                    "Run tools/TestManager/filer_directory_setup.thpl on $volume from $client to create dirs: "
                      . join(",", @dirs));
                $Remote->execute(
                       host        => $client,
                       nate_params => $params,
                       cmd   => "tools/TestManager/filer_directory_setup.thpl",
                       runid => sprintf("%03d_%s_%s",
                                        $num_subs, "filer_setup_dirs", $client),
                       handler => sub { });
                $client_idx++;
                $num_subs++;
            } ## end while (my @dirs = splice(...
        } ## end foreach my $volume (keys %{...
        my $fail_result = $Remote->wait(timeout => 300);
        die config_result("dirs creation failed: $fail_result")
          if ($fail_result);
    } else {
        $Log->comment(
                "Directories weren't required to be created for this Topology");
    }

    # Setup exports appropriately
    # If one wants to export max exports allowed (10240), then we have to unexport
    # everything for max exports to succeed.
    if ((keys %{$self->{filer_setup}{exports}}) >= 10240) {
        unexport_all($conn);
    }

    if ($Options_hash->{PERSISTENT_EXPORT}) {
        my $output;
        eval {
            $output = sayEOF($conn, "exportfs -w /etc/exports", timeout => 180);
        };
        die config_result("export -w: $output: $@") if ($@);
        $Log->comment("export -w: $output");
    }

    # Set this option, so clients using non-previlaged ports also will have access
    filer::filer_options_set($conn, "nfs.mount_rootonly", "off");

    $Log->exit();
} ## end sub filer_setup

sub filer_cleanup {

    my ($self, %options) = @_;

    my $test_topology = $self->{test_topology};

    # create a list of all node resources in the test_topology
    my @resources;
    foreach my $resource_ref (values %{$test_topology->{cluster}}) {
        push(@resources, @{$resource_ref})
          if (ref($resource_ref) eq "ARRAY");
    }
    foreach my $node (keys %{$test_topology->{node}}) {
        foreach my $res_ref (values %{$test_topology->{node}{$node}}) {
            push(@resources, @{$res_ref})
              if (ref($res_ref) eq "ARRAY");
        }
    }

    # check all of the resources for a comment that indicates that the resource
    # was previously created by TestManagerLite. Mark this resource for cleanup
    foreach my $resource (@resources) {
        next
          if (ref($resource) ne "HASH");
        if ($resource->{comment} =~ "created by TestManagerLite") {
            $resource->{Delete_Resource} = 1;
        } elsif ($resource->{rulesetname}) {
            $resource->{Delete_Resource} = 1;
        }
    }

    $Server->delete_topology(print_node => undef,
                             topology   => $test_topology);

    # Execute the one from the parent
    $self->SUPER::filer_cleanup(%options);
} ## end sub filer_cleanup

sub client_setup {
    my ($self, %options) = @_;
    $Log->enter();

    # Execute the one from the parent
    $self->SUPER::client_setup(%options);

    if (!$options{SRCIP}) {
        $Log->comment("SRCIP is not set, no IPs need to be configured");
        return;
    }

    my @ips       = $self->_create_range($options{SRCIP});
    my $first_vip = $ips[0];
    my $last_vip  = $ips[-1];

    my ($device, $prefix) =
      $self->_pick_network_info(first_ip  => $first_vip,
                                last_ip   => $last_vip,
                                addr_type => $options{ADDR_TYPE},
                                device    => $options{DEVICE});

    $self->_setup_ips(ips       => \@ips,
                      device    => $device,
                      addr_type => $options{ADDR_TYPE},
                      prefix    => $prefix);

    $Log->exit();
} ## end sub client_setup

sub DESTROY {
    my $self = shift;
    local ($@);
    $self->_cleanup_ips();
    $self->client_cleanup();
}

sub client_cleanup {
    my ($self, %options) = @_;
    $Log->enter();

    # Execute the one from the parent
    $self->SUPER::client_cleanup(%options);

    $Log->exit();
}
########################################################
######################### Internal Routines ##################

sub _export_options {
    my ($self, $access, $unique) = @_;

    my $export_options;
    if ($access) {

        # If = is found, it means export options are given directly, hence use as it is
        if ($unique) {

            #
            # Make each export rule unique, hence use this technique: Attach some random
            # string to allow rule that is specified.
            #
            my $RS    = new String::Random;
            my $rstr1 = $RS->randregex('[a-z][a-z][a-z][a-z][a-z][a-z]');
            my $rstr2 = $RS->randregex('[a-z][a-z][a-z][a-z][a-z][a-z]');
            if ($access =~ /=|,/) {
                $export_options = $access . ":" . $rstr1;
            } else {
                $export_options =
                    'rw=' 
                  . $access . ':' 
                  . $rstr1 
                  . ',root=' 
                  . $access . ':'
                  . $rstr2;
            }
        } else {
            if ($access =~ /=|,/) {
                $export_options = $access;
            } else {
                $export_options = 'rw=' . $access . ',root=' . $access;
            }
        }
    } else {
        $export_options = "rw,anon=0";
    }

    return $export_options;
} ## end sub _export_options
########################################################
sub _create_range {
    my ($self, $ip) = @_;
    my ($begin, $range);
    if ($ip =~ /(.+)\/(\d+)/) {
        ($begin, $range) = ($1, $2);
    } elsif (!defined($ip)) {
        return ();
    } else {
        $begin = $ip;
        $range = 1;
    }
    $range--;

    my @range;
    ## This works for both IPv4 and IPv6
    my $ipobj = new Net::IP("$begin + $range")
      or die "Failed to create ip object: $@";

    do {
        push(@range, $ipobj->ip());
    } while (++$ipobj);

    return (wantarray) ? @range : $range[$#range];
} ## end sub _create_range

sub _get_addr_type {

    my $self = shift;
    if ($Options->get(option => "SRCIP") =~ /(.+)\/\d+/) {
        Net::IP::ip_is_ipv6($1)
          ? $Options->set(ADDR_TYPE => "inet6")
          : $Options->set(ADDR_TYPE => "inet4");
    } else {
        Net::IP::ip_is_ipv6(@{$Options->get(option => "DATA_INTERFACES")}[0])
          ? $Options->set(ADDR_TYPE => "inet6")
          : $Options->set(ADDR_TYPE => "inet4");
    }
} ## end sub _get_addr_type
#########################################################
sub _cleanup_ips {
    my $self = shift;
    $Log->enter();

    if ($self->{ip_setup_id} && $self->{ip_setup_id} != $$) {
        $Log->debug(
              "process $$ can not cleanup ips created by $self->{ip_setup_id}");
        return;
    }

    while (my $cmd = pop @{$self->{ip_cleanupcmds}}) {
        my $output = qx($cmd 2>&1);
        $Log->warn("$cmd failed: $output") if $?;
        $Log->comment("$cmd : $output");
    }

    $Log->exit();
} ## end sub _cleanup_ips

sub _update_test_topology {
    my $self = shift;

    $Log->enter();
    $Options_hash = $Options->get_hash();
    my $orig_server;
    if (my @clustered = $Server->get_clustered_nodes()) {
        $orig_server = $Server->set_execution_server(-Server => $clustered[0]);
    }
    my $test_topology;

    $self->{flexcaches_involved} = 0;
    my ($origin, $cache);

    # Classic filers must be looked at via the server object since there is not a
    #  global view like there is in NG. Check each filer and flag whether or not
    #  a cache is involved since we do not want to create volumes on these filers.
    foreach my $filer ($Server->servers()) {
        if ($Server->is_cache(-Server => $filer)) {
            $cache = $filer;
            $self->{flexcaches_involved}++;
        } else {
            $origin = $filer;
        }
    }
    if (!$origin) {
        die config_result(
            "All of the specified filers are licensed for FlexCache.\n.
                          This code requires that at least one is not licensed."
            );
    }
    if (!$cache) {

        # supposed to use a cache but there are none
        die config_result(
            "There were no FlexCache filers\n.
                          specified with the FILER param. The current list of filers is:\n.
                           $Server->servers()");
    }
    $self->{cache}  = $cache;
    $self->{origin} = $origin;

    # Find the minimum size required for an aggr using min vol size
    my $min_size = 300;
    foreach my $vol (keys %{$self->{filer_setup}{volume}}) {
        if ($min_size < $self->{filer_setup}{volume}{$vol}{size}) {
            $min_size = $self->{filer_setup}{volume}{$vol}{size};
        }
    }
    my @clustered = $Server->get_clustered_nodes();
    my @Aggrs =
      $Server->pick_aggregates(
                             state              => "online",
                             flexcache_involved => $self->{flexcaches_involved},
                             -Server            => $clustered[0],
                             avail_size         => "$min_size");
    $Log->debug("Auto Picked Aggrs: " . Dumper(\@Aggrs));

    my @Aggregates;
    if (@{$Options_hash->{aggregate}}) {

        # If aggrs are user-specified, then make sure they exist, otherwise fail
        foreach my $aggr (@{$Options_hash->{aggregate}}) {
            my $found = 0;
            foreach my $ag_hash (@Aggrs) {
                if ($ag_hash->{aggregate} eq $aggr) {
                    $found = 1;
                    push @Aggregates, $ag_hash;
                }
            }
            die config_result(
                    "Specified Aggr $aggr doesn't exist with size $min_size MB")
              if (!$found);
        } ## end foreach my $aggr (@{$Options_hash...
    } else {
        push @Aggregates, @Aggrs;
    }
    $Log->debug("Usable Aggr hash: " . Dumper(\@Aggregates));

    my $aggrs = $self->{discovered_topology}{cluster}{aggregate};

    my $volume_number = 0;
    foreach my $volume (keys %{$self->{filer_setup}{volume}}) {
        my $vol_hash;
        $vol_hash->{volume} = $volume;
        my $vol_size = $self->{filer_setup}{volume}{$volume}{size};
        if ($vol_size < 300) {
            $Log->comment(
                "Vol size is adjusted from $vol_size to 300 to keep min size of 300MB per volume"
                );
            $vol_size = 300;
        }
        $vol_hash->{size} = $vol_size . "MB";
        my $cache_vol_size = $vol_size . "m";

        # see if the volume already exists in the cluster
        my $existing_volume;
        eval { $existing_volume = $Server->volume_show(volume => $volume); };

        # make sure that we have a volume array and that the volume hash is a real one.
        # For 7G, a volume_hash gets generated and is not complete....
        if ($existing_volume && $existing_volume->[0]{volume}) {
            $existing_volume = $existing_volume->[0];

            # find info on the hosting aggregate
            my ($aggr_hash) =
              grep($_->{aggregate} eq $existing_volume->{aggregate},
                   @Aggregates);
            if (ref($aggr_hash) ne "HASH") {
                $Log->comment(Dumper $existing_volume);
                $Log->comment(Dumper \@Aggregates);
                die config_result(
                    "Could not find the aggregate hosting volume '$existing_volume->{volume}'"
                    );
            }
            foreach my $key (keys %{$existing_volume}) {

                #Don't cache volume params that are "-" or blank
                $vol_hash->{$key} = $existing_volume->{$key}
                  if ($existing_volume->{$key} !~ /^-$|^$/);
                $vol_hash->{-Filer} = $aggr_hash->{filers};
            }
            next;
        } ## end if ($existing_volume &&...

        # fill in missing aggregate information
        my @usable_aggrs;
        foreach my $agr (@Aggregates) {
            if ($agr->{availsize} >= $vol_size) {
                push(@usable_aggrs, $agr);
            }
        }

        my $aggr_info;
        map($aggr_info .= "$_->{aggregate} => $_->{availsize}\n",
            @usable_aggrs);
        $Log->trace($aggr_info);

        if (!scalar(@usable_aggrs)) {
            die config_result(
                           "No aggregates with sufficient space are available, "
                             . $vol_hash->{size}
                             . " needed.");
        }

        # pick aggregates in a round-robin fashion rather than random
        my $agr_hash = $usable_aggrs[$volume_number++ % (scalar @usable_aggrs)];
        $agr_hash->{availsize} -= $vol_size;
        $vol_hash->{aggregate} = $agr_hash->{aggregate}
          if (!$vol_hash->{aggregate});

        $Log->trace(
            "placed volume '$vol_hash->{volume}' on aggregate '$vol_hash->{aggregate}': its size is now $agr_hash->{availsize}"
            );

        # fill in vserver information - Should we make Server.pm happy?
        $vol_hash->{vserver} = 'N/A';

        # fill in missing security_style information. It defaults to "unix" unless specified.
        if (!$vol_hash->{'security-style'}) {
            $vol_hash->{'security-style'} = $$Options_hash{VOL_SECURITY_STYLE}
              || "unix";
        }

        # fill in user/group information
        if ($$Options_hash{VOLUME_OWNER} eq 'self') {
            $vol_hash->{user} = $ENV{LOGNAME};
            (undef(),            undef(), undef(),
             $vol_hash->{group}, undef(), undef(),
             undef(),            undef(), undef()) = getpwnam($ENV{LOGNAME});
        } else {
            $vol_hash->{user}  = 0 if (!$vol_hash->{user});
            $vol_hash->{group} = 0 if (!$vol_hash->{group});
        }

        # fill in junction-path information - Server.pm happy?
        $vol_hash->{'junction-path'} = "/$volume";
        $vol_hash->{ruleset} = "allow_all";

        # Classic needs to know which filer the aggr/volume are on
        $vol_hash->{-Filer} = $agr_hash->{filers};

        # Indicate that it was created by TestManagerLite
        $vol_hash->{comment} = '"created by TestManagerLite"';

        push @{$test_topology->{cluster}{volume}}, $vol_hash;

        # Build up the hash for cache filer and make Server.pm happy
        push(@{$test_topology->{node}{$cache}{exportruleset_rules}},
             {ruleindex   => 1,
              vserver     => 'N/A',
              rulesetname => "allow_all",
              anon        => 0,
              protocol    => "any",
              clientmatch => "0.0.0.0/0",
              rorule      => "any",
              rwrule      => "any",
              superuser   => "any",});
        push(@{$test_topology->{node}{$cache}{cache_volume}},
             {Delete_Resource => 1,
              source          => "$origin\:$volume",
              volume          => "$volume",
              vserver         => 'N/A',
              -Filer          => "$cache",
              size            => "$cache_vol_size",
              ruleset         => "allow_all",
              aggregate       => "$vol_hash->{aggregate}",});

    } ## end foreach my $volume (keys %{...

    push(@{$test_topology->{cluster}{exportruleset}},
         {vserver     => 'N/A',
          rulesetname => "allow_all",});
    push(@{$test_topology->{cluster}{exportruleset_rules}},
         {ruleindex   => 1,
          vserver     => 'N/A',
          rulesetname => "allow_all",
          anon        => 0,
          protocol    => "any",
          clientmatch => "0.0.0.0/0",
          rorule      => "any",
          rwrule      => "any",
          superuser   => "any",});

    $self->{test_topology} = $test_topology;

    $Log->debug("Test Topology: \n" . Dumper($test_topology));

    $Log->exit();
    return $test_topology;
} ## end sub _update_test_topology

######################### Internal Routines ##################

sub unexport_all {
    my ($conn) = @_;
    my $output = sayEOF($conn, "exportfs -ua", timeout => 180);

    $Log->comment("unexport all: $output");
    return $output;
}

sub export_all {
    my ($conn) = @_;
    my $output = sayEOF($conn, "exportfs -a", timeout => 180);

    $Log->comment("export all: $output");
    return $output;
}

my $File_read = undef;

sub _read_testuser_file {
    my $tfile = shift;
    my $file =
      NATE::Inc::find_file($tfile) ? NATE::Inc::find_file($tfile) : $tfile;
    $Log->enter("file=>$file");
    if ($File_read) {
        $Log->trace("file already processed");
        goto done;
    }
    my $testuser = $Options->get(option => "TESTUSER");

    if (open(IF, $file)) {
        while (<IF>) {
            next if (/^\s*\#/);
            chomp;
            push(@{$testuser}, $_);
        }
        close(IF);
    } else {
        die "Could not open TESTUSER_FILE: $tfile ($file) : $! / $@";
    }

    $File_read = 1;

  done: $Log->exit();
} ## end sub _read_testuser_file

sub _pick_network_info {
    my $self     = shift;
    my (%opts)   = @_;
    my $OS       = qx($sudo uname 2>&1);
    my $addrtype = ($opts{addr_type} eq "inet6") ? "inet6" : "inet";
    my ($EIP, $GFIP, $GLIP);
    $Log->enter(Dumper(\%opts));

    my ($device, $netmask);    # to be returned.
    my $interface;
    if ($OS =~ /SunOS/i) {

        my $out = qx($sudo $ifconfig -a $addrtype 2>&1);
        die config_result("Could not get interface information: $out")
          if $?;

        if ($opts{addr_type} eq "inet4") {

            # Primary IPv4 address is expected to be in primary interface,
            # hence look for only primary interfaces to pick appropriate interface
            foreach (split /\n/, $out) {

                # Get interface name
                next if (/^\S+:\d+:/);    # Ignore all logical interfaces
                $interface = $1 if (/^(\S+):/);

                next if ($opts{device} && $opts{device} ne $interface);

                # Get the ip address and mask for interface
                my ($addr, $mask) =
                  (/inet\s+(\S+)\s+netmask\s+(\S+)\s+broadcast\s+(\S+)/);
                next if ($addr =~ /127\.0\.0\.1/);    # Ignore if loopback
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    # SunOS prints mask in hex format, convert it to dotted decimal format
                    # to use it in inet_aton.
                    $mask =
                        hex(substr($mask, 0, 2)) . "."
                      . hex(substr($mask, 2, 2)) . "."
                      . hex(substr($mask, 4, 2)) . "."
                      . hex(substr($mask, 6, 2));

                    # Get the network part of the interface
                    my $net_mask = inet_aton($mask);
                    my $net_ip   = inet_aton($addr) & $net_mask;

                    # Check if this is a valid interface for the requested range
                    if (   (inet_aton($opts{first_ip}) & $net_mask) eq $net_ip
                        && (inet_aton($opts{last_ip}) & $net_mask) eq $net_ip) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                } ## end if ($addr && $mask)
            } ## end foreach (split /\n/, $out)
        } else {

            # Primary IPv6 address should be on one of the logical interface
            foreach (split /\n/, $out) {

                # Get interface name
                $interface = $1 if (/^(\S+):/);

                # Save only primary interface
                $interface = (split(":", $interface))[0];
                next if ($opts{device} && $opts{device} ne $interface);

                # Get the ip address and prefix for interface
                my ($addr, $mask) = (/inet6\s+(\S+)\/(\S+)/);

                # Ignore link and loopback
                next if ($addr =~ /^::1/ || $addr =~ /^fe80::/);
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");
                    $EIP = $self->_get_inet6_iprange_obj($addr . "/" . $mask);
                    $GFIP =
                      $self->_get_inet6_iprange_obj(
                                                 $opts{first_ip} . "/" . $mask);
                    $GLIP =
                      $self->_get_inet6_iprange_obj(
                                                  $opts{last_ip} . "/" . $mask);

                    if ((   $EIP->overlaps($GFIP) == $IP_B_IN_A_OVERLAP
                         || $EIP->overlaps($GFIP) == $IP_IDENTICAL)
                        && (   $EIP->overlaps($GLIP) == $IP_B_IN_A_OVERLAP
                            || $EIP->overlaps($GLIP) == $IP_IDENTICAL)
                      ) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                } ## end if ($addr && $mask)
            } ## end foreach (split /\n/, $out)
        } ## end else [ if ($opts{addr_type} eq...
    } elsif ($OS =~ /Linux/i) {

        # Get interface name, ip address and netmask
        my $out = qx($sudo $ifconfig -a 2>&1);
        die config_result("Could not get interface information: $out")
          if $?;

        foreach (split /\n/, $out) {

            # Get interface name
            next if (/^\S+:\d+/);    # Ignore all logical interfaces
            $interface = $1 if (/^(\S+)/);
            next if ($opts{device} && $opts{device} ne $interface);

            if ($opts{addr_type} eq "inet4") {

                # Get the ip address and mask for interface
                my ($addr, $mask) =
                  (/inet addr:(\S+)\s+Bcast:\S+\s+Mask:(\S+)/);
                next if ($addr =~ /127\.0\.0\.1/);    # Ignore if loopback
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    # Get the network part of the interface
                    my $net_mask = inet_aton($mask);
                    my $net_ip   = inet_aton($addr) & $net_mask;

                    # Check if this is a valid interface for the requested range
                    if (   (inet_aton($opts{first_ip}) & $net_mask) eq $net_ip
                        && (inet_aton($opts{last_ip}) & $net_mask) eq $net_ip) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                } ## end if ($addr && $mask)
            } else {

                # Get the ip address and prefix for interface
                my ($addr, $mask) = (/inet6\s+addr:\s*(\S+)\/(\S+)/);
                next
                  if ($addr =~ /^::1/ || $addr =~ /^fe80::/)
                  ;    # Ignore link and loopback
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    $EIP = $self->_get_inet6_iprange_obj($addr . "/" . $mask);
                    $GFIP =
                      $self->_get_inet6_iprange_obj(
                                                 $opts{first_ip} . "/" . $mask);
                    $GLIP =
                      $self->_get_inet6_iprange_obj(
                                                  $opts{last_ip} . "/" . $mask);

                    if ((   $EIP->overlaps($GFIP) == $IP_B_IN_A_OVERLAP
                         || $EIP->overlaps($GFIP) == $IP_IDENTICAL)
                        && (   $EIP->overlaps($GLIP) == $IP_B_IN_A_OVERLAP
                            || $EIP->overlaps($GLIP) == $IP_IDENTICAL)
                      ) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                } ## end if ($addr && $mask)
            } ## end else [ if ($opts{addr_type} eq...
        } ## end foreach (split /\n/, $out)
    } else {
        die config_result("$OS is not supported by this topology yet");
    }

  FOUND:
    if (!defined($device)) {
        die config_result("Could not find proper interface for virtual ip range"
                          . " [$opts{first_ip}, $opts{last_ip}]");
    }

    $Log->exit("Selected: $device, $netmask");
    return ($device, $netmask);
} ## end sub _pick_network_info

sub _setup_ips {
    my $self = shift;
    my (%opts) = @_;
    $Log->enter();

    my ($cmd, $cleanupcmd);
    my $OS = qx($sudo uname 2>&1);

    # Let's assume no one wants to use logical interface on this system
    my $index = 0;
    foreach my $ip (@{$opts{ips}}) {
        if ($OS =~ /SunOS/i) {
            if ($opts{addr_type} eq "inet6") {
                $cmd =
                  "$sudo $ifconfig $opts{device} inet6 addif $ip/$opts{prefix} up";
                $cleanupcmd =
                  "$sudo $ifconfig $opts{device} inet6 removeif $ip";
            } else {
                $cmd =
                  "$sudo $ifconfig $opts{device} inet addif $ip netmask $opts{prefix} up";
                $cleanupcmd = "$sudo $ifconfig $opts{device} inet removeif $ip";
            }
        } elsif ($OS =~ /Linux/i) {
            if ($opts{addr_type} eq "inet6") {
                $cmd =
                  "$sudo $ifconfig $opts{device} inet6 add $ip/$opts{prefix} up";
                $cleanupcmd =
                  "$sudo $ifconfig $opts{device} inet6 del $ip/$opts{prefix}";
            } else {
                $cmd =
                  "$sudo $ifconfig $opts{device}:$index $ip netmask $opts{prefix} up";
                $cleanupcmd = "$sudo $ifconfig $opts{device}:$index down";
                $index++;
            }
        } else {
            die config_result("$OS is not supported by this topology yet");
        }

        my $output = qx($cmd 2>&1);
        die config_result("$cmd failed: $output") if $?;
        $Log->comment("$cmd : $output");

        push(@{$self->{ip_cleanupcmds}}, $cleanupcmd);
    } ## end foreach my $ip (@{$opts{ips...
    $Log->comment("IP Setup Complete : " . qx($sudo $ifconfig -a 2>&1));

    $self->{vdev_cleanup} = $opts{device};
    $self->{ip_setup_id}  = $$;             # Save this for cleanup to take care
    $Log->exit();
} ## end sub _setup_ips

1;
