# $Id: $

## @summary     NFS Stress Scalability 10 Mode Virtual Topology Library
##
## @author      gvg@netapp.com
##
## @pod here

=head1 NAME

Topology::NAS::NFS::10M_Virtual - NFS Stress Scalability test Virtual Topology Library for 10Mode ONTAP

=head1 SYNOPSIS

    See Topology.pm.

=head1 DESCRIPTION

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/Topology

=head1 TODO

=head1 SEE ALSO

Options TestManagerLite Topology

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Last revised $Date: 2008/09/30 $
    
=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved

=cut

package Topology::NAS::NFS::CMode_Virtual;
use base qw(Topology::NAS::NFS::CMode);

use strict;
use warnings;

use Tharn qw($Log config_result);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Socket;
use Net::IP;
use Data::Dumper qw(Dumper);

use Options;
my $Options_hash;

my $ifconfig = '/sbin/ifconfig';
my $sudo     = '/usr/software/bin/sudo';

###################################################################

my $Options = new Options;

$Options->parameters(
    definitions => [
        {name => "DEST_PATHS_FILE",
         spec => "s",
         description =>
           "file containing the list of volume paths for clients to mount to",
         file_tests => "fr",
         subroutine => [\&_read_volpaths_file],},
        {name        => "DUP_IPS",
         spec        => "s",
         default     => 1,
         description => "The number of duplicate clients per SRCIP",},
        {name => "MULTI_SERVER",
         description =>
           "Whether or not a client should run traffic against single server or multi server. Please see DATA_INTERFACES to specify multiple interface ips. Unless leaf script supports this feature this is no-op",
         default => 1,
         spec    => "s",},]);

sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $self     = {@_,};
    $Log->enter();
    bless($self, $class);

    $Log->exit();
    return $self;
}

sub define_topology {
    my ($self) = shift;
    $Log->enter();

    # Startup the IP Manager
    $self->{ipManager} = $self->_start_ip_manager();

    $Options_hash = $Options->get_hash();
    $Log->debug("Topology Options: " . Dumper($Options_hash));
    
    my $ipManager = $self->{ipManager};
    my @netObjs = $ipManager->source_networks();
    if (! @netObjs) { 
        $Log->warn("The CMode_Virtual topology is called with no supplied SRCIP value.");
    }

    if (($Options_hash->{MAX_NFS_CLIENTS}) and ($Options_hash->{DUP_IPS}) > 1) { 
        die config_result ("The DUP_IPS value should not be greater than one when used with the MAX_NFS_CLIENTS value.");
    }
    my $dup_ips = $Options_hash->{DUP_IPS};
    my $max_nfs_clients = $Options_hash->{MAX_NFS_CLIENTS};
    my @clients = @{$self->{clients}};

    # How many IP addresses did we get?
    my $netObj = $netObjs[0];

    # It is possible to run synthetic clients without the SRCIP value. There may be no networks in the 
    # IP manager. In this case, any instances will use the localhosts ip address as opposed to one that is
    # given as SRCIP
    my $availableAddressesNum = 0;
    if ($netObj) { 
        $availableAddressesNum = $netObj->unassigned_addresses();
        if (! $availableAddressesNum > 0) { 
            die config_result("Could not determine source address from SRCIP. Check configuration.");
        }
       
        $Log->comment(
            "There were (" . $availableAddressesNum . ") test instances determined from the supplied IP addresses.");

        # complain if we have more clients than instances
        if (scalar(@clients) > $availableAddressesNum) { 
           $Log->warn("The number of supplied clients (" . scalar(@clients) .") was greater than the number" . 
              " of requested instances ($availableAddressesNum).");
        }

    }
    
    if ($max_nfs_clients) { 
        $Log->comment("Adjusting the available number of addresses to match the value of MAX_NFS_CLIENTS ($max_nfs_clients). " . 
                        "There were actually $availableAddressesNum addresses specified.");
        $availableAddressesNum = $max_nfs_clients;
    }

    # Check to see if will exceed the INSTANCES_PER_CLIENT value which should be a test (thpl) option
    if (defined($Options_hash->{INSTANCES_PER_CLIENT})) {
        # Cry if we exceed the INSTANCES_PER_CLIENT number
        if (int($availableAddressesNum/scalar(@clients)) > $Options_hash->{INSTANCES_PER_CLIENT}) {
            $Log->warn("There were ("
                . scalar(@clients)
                . ") clients and ("
                . $availableAddressesNum . ") "
                . "instances specified for this test.\nThis configuration will result in more instances "
                . "per client ("
                . int(($availableAddressesNum/scalar(@clients)))
                . "), than the number ("
                . $Options_hash->{INSTANCES_PER_CLIENT}
                . ") recommended by this test case."
                . "\nThis can be overridden with the INSTANCES_PER_CLIENT parameter on the command line."
                );
        }
    }

    # Execute the one from the parent
    $self->SUPER::define_topology(@_);

    # Setup the instances
    $self->_configure_instances();

    # Configure each instance with a DEST_PATHS
    $self->_assign_path_to_instance();

    # Set the size of each volume based on the number of instances.
    $self->_set_usable_volume_size();

    # Determine the users who will need quotas configured
    if ($Options->get(option => "QUOTA_VOLUME_ENABLE")) {
        $self->_define_volume_quotas();
    }

    # Get creative about logging all our information
    my $top_logdir = Tharn::param('TOP_LOGDIR');
    my $runid = Tharn::param('RUNID'); 
    my $lonely_log = $top_logdir . '/' . $runid . '_config_info.log';
    if ($Log->may_debug()) { 
        $Log->comment("define_topology Topology: \n" . Dumper($self), path => $lonely_log);

    } else {
        $Log->comment("filer_setup:" . Dumper($self->{filer_setup}), path => $lonely_log);
        $Log->comment("client_config:" . Dumper($self->{client_config}), path => $lonely_log);
        $Log->comment("test_options:" . Dumper($self->{test_options}), path => $lonely_log);
    }

    my $sleepTime = $Options_hash->{DELAY_SETUP_START};
    if ($sleepTime) {
        $Log->comment("Pausing for $sleepTime seconds before leaving define_topology");
        sleep($sleepTime);
    }
}

sub _define_volume_quotas {
    my ($self) = shift;

    # Execute the one from the parent
    $self->SUPER::_define_volume_quotas();

    $Log->exit();
}

sub filer_setup {
    my ($self, %options) = shift;

    # Execute the one from the parent
    $self->SUPER::filer_setup(%options);

    $Log->exit();
}

sub filer_cleanup {
    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::filer_cleanup(%options);

    $Log->exit();
}

sub client_setup {
    my ($self, %options) = @_;
    $Log->enter();

    # Execute the one from the parent
    $self->SUPER::client_setup(%options);

    my @clientSetupKeys = keys(%{$self->{client_setup}});

    # exit early if there are no SRCIP values
    return if (not $self->{client_setup}->{$clientSetupKeys[0]}->{SRCIP});

    my $ipManager = $self->_start_ip_manager();
    # Add the SRCIP information its in the client_setup options
    $Log->trace("Adding SRCIP $self->{client_setup}->{$clientSetupKeys[0]}->{SRCIP} to local IPManager." . 
                 " This address is assigned to this instance from the supervisor and passed through Staf.");
    $ipManager->address_add(SRCIP => $self->{client_setup}->{$clientSetupKeys[0]}->{SRCIP});

    my @netObjs = $ipManager->source_networks();

    if (not @netObjs) {
        $Log->warn("It would seem that there are no source networks. This is likely to be bad.");
        return;
    }

    my $netObj = $netObjs[0];
    my @ipObjs = $netObj->addresses();

    my ($device, $prefix) =
      $self->_pick_network_info(first_ip  => $ipObjs[0]->address(),
                                last_ip   => $ipObjs[-1]->address(),
                                addr_type => $ipObjs[0]->version());

    $self->_setup_ips(ips       => \@ipObjs,
                      device    => $device,
                      prefix    => $prefix);

    $Log->exit();
}

#-------------------------------------------------------------------------------
# Destructor for object
# It removes all virtual ip address interfaces that this object created
#-------------------------------------------------------------------------------
sub DESTROY {
    my $self = shift;
    local ($@);

    $self->client_cleanup();
}

sub client_cleanup {
    my ($self, %options) = @_;
    $Log->enter();

    $self->_cleanup_ips();

    # Execute the one from the parent
    $self->SUPER::client_cleanup(%options);

    $Log->exit();
}

######################### Internal Routines ##################


sub _pick_network_info {
    my $self     = shift;
    my (%opts)   = @_;
    my $OS       = qx($sudo uname 2>&1);
    my $addrVersion = $opts{addr_type};
    my $addrtype = ($addrVersion =~ /6/) ? "inet6" : "inet";
    my ($EIP, $GFIP, $GLIP);
    $Log->enter(Dumper(\%opts));

    my ($device, $netmask);    # to be returned.
    my $interface;
    if ($OS =~ /SunOS/i) {

        my $out = qx($sudo $ifconfig -a $addrtype 2>&1);
        die config_result("Could not get interface information: $out") if $?;

        if ($addrVersion =~ /^4$/) {

            # Primary IPv4 address is expected to be in primary interface,
            # hence look for only primary interfaces to pick appropriate interface
            foreach (split /\n/, $out) {

                # Get interface name
                next if (/^\S+:\d+:/);    # Ignore all logical interfaces
                $interface = $1 if (/^(\S+):/);

                next if ($opts{device} && $opts{device} ne $interface);

                # Get the ip address and mask for interface
                my ($addr, $mask) =
                  (/inet\s+(\S+)\s+netmask\s+(\S+)\s+broadcast\s+(\S+)/);
                next if ($addr =~ /127\.0\.0\.1/);    # Ignore if loopback
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    # SunOS prints mask in hex format, convert it to dotted decimal format
                    # to use it in inet_aton.
                    $mask =
                        hex(substr($mask, 0, 2)) . "."
                      . hex(substr($mask, 2, 2)) . "."
                      . hex(substr($mask, 4, 2)) . "."
                      . hex(substr($mask, 6, 2));

                    # Get the network part of the interface
                    my $net_mask = inet_aton($mask);
                    my $net_ip   = inet_aton($addr) & $net_mask;

                    # Check if this is a valid interface for the requested range
                    if (   (inet_aton($opts{first_ip}) & $net_mask) eq $net_ip
                        && (inet_aton($opts{last_ip}) & $net_mask) eq $net_ip) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                }
            }
        } else {

            # Primary IPv6 address should be on one of the logical interface
            foreach (split /\n/, $out) {

                # Get interface name
                $interface = $1 if (/^(\S+):/);

                # Save only primary interface
                $interface = (split(":", $interface))[0];
                next if ($opts{device} && $opts{device} ne $interface);

                # Get the ip address and prefix for interface
                my ($addr, $mask) = (/inet6\s+(\S+)\/(\S+)/);

                # Ignore link and loopback
                next if ($addr =~ /^::1/ || $addr =~ /^fe80::/);
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    $EIP = $self->_get_inet6_iprange_obj($addr . "/" . $mask);
                    $GFIP =
                      $self->_get_inet6_iprange_obj(
                                                 $opts{first_ip} . "/" . $mask);
                    $GLIP =
                      $self->_get_inet6_iprange_obj(
                                                  $opts{last_ip} . "/" . $mask);

                    if ((   $EIP->overlaps($GFIP) == $IP_B_IN_A_OVERLAP
                         || $EIP->overlaps($GFIP) == $IP_IDENTICAL)
                        && (   $EIP->overlaps($GLIP) == $IP_B_IN_A_OVERLAP
                            || $EIP->overlaps($GLIP) == $IP_IDENTICAL)
                      ) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                }
            }
        }
    } elsif ($OS =~ /Linux/i) {

        # Get interface name, ip address and netmask
        my $out = qx($sudo $ifconfig -a 2>&1);
        die config_result("Could not get interface information: $out") if $?;

        foreach (split /\n/, $out) {

            # Get interface name
            next if (/^\S+:\d+/);    # Ignore all logical interfaces
            $interface = $1 if (/^(\S+)/);
            next if ($opts{device} && $opts{device} ne $interface);

            if ($addrVersion =~ /^4$/) {

                # Get the ip address and mask for interface
                my ($addr, $mask) =
                  (/inet addr:(\S+)\s+Bcast:\S+\s+Mask:(\S+)/);
                next
                  if ($addr && ($addr =~ /127\.0\.0\.1/));  # Ignore if loopback
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    # Get the network part of the interface
                    my $net_mask = inet_aton($mask);
                    my $net_ip   = inet_aton($addr) & $net_mask;

                    # Check if this is a valid interface for the requested range
                    if (   (inet_aton($opts{first_ip}) & $net_mask) eq $net_ip
                        && (inet_aton($opts{last_ip}) & $net_mask) eq $net_ip) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                }
            } else {

                # Get the ip address and prefix for interface
                my ($addr, $mask) = (/inet6\s+addr:\s*(\S+)\/(\S+)/);
                next
                  if ($addr =~ /^::1/ || $addr =~ /^fe80::/)
                  ;    # Ignore link and loopback
                if ($addr && $mask) {
                    $Log->comment(
                             "Processing $interface addr $addr and mask $mask");

                    $EIP = $self->_get_inet6_iprange_obj($addr . "/" . $mask);
                    $GFIP =
                      $self->_get_inet6_iprange_obj(
                                                 $opts{first_ip} . "/" . $mask);
                    $GLIP =
                      $self->_get_inet6_iprange_obj(
                                                  $opts{last_ip} . "/" . $mask);

                    if ((   $EIP->overlaps($GFIP) == $IP_B_IN_A_OVERLAP
                         || $EIP->overlaps($GFIP) == $IP_IDENTICAL)
                        && (   $EIP->overlaps($GLIP) == $IP_B_IN_A_OVERLAP
                            || $EIP->overlaps($GLIP) == $IP_IDENTICAL)
                      ) {

                        # This is the main interface where all
                        # virtual ip addresses will be created
                        $device  = $interface;
                        $netmask = $mask;
                        goto FOUND;
                    } else {
                        $Log->comment(
                            "Addr $addr and prefix $mask didn't match with starting ip $opts{first_ip} and ending ip $opts{last_ip}"
                            );
                    }
                }
            }
        }
    } else {
        die config_result("$OS is not supported by this topology yet");
    }

  FOUND:
    if (!defined($device)) {
        die config_result("Could not find proper interface for virtual ip range"
                          . " [$opts{first_ip}, $opts{last_ip}]");
    }

    $Log->exit("Selected: $device, $netmask");
    return ($device, $netmask);
}

sub _get_inet6_iprange_obj {
    my $self    = shift;
    my $iprange = shift;
    $Log->enter("iprange = $iprange");

    my ($ip, $prefix) = Net::IP::ip_splitprefix($iprange);
    my $octets = int($prefix / 16);
    my @octs = split(":", $ip);
    delete $octs[-1] while ($octets--);
    $ip = join(":", @octs) . "::/" . $prefix;
    my $ipobj = new Net::IP($ip)
      or die config_result("Net::IP failed for $ip");

    $Log->exit("returning Net::IP Object for $ip");
    return $ipobj;
}

sub _setup_ips {
    my $self = shift;
    my (%opts) = @_;
    $Log->enter();

    my ($cmd, $cleanupcmd);
    my $OS = qx($sudo uname 2>&1);

    # Let's assume no one wants to use logical interface on this system
    my $index = 0;
    foreach my $ipObj (@{$opts{ips}}) {
        if ($OS =~ /SunOS/i) {
            if ($ipObj->version() =~ /6/) {
                $cmd =
                  "$sudo $ifconfig $opts{device} inet6 addif " . $ipObj->address() . "/$opts{prefix} up";
                $cleanupcmd =
                  "$sudo $ifconfig $opts{device} inet6 removeif " . $ipObj->address();
            } else {
                $cmd =
                  "$sudo $ifconfig $opts{device} inet addif " . $ipObj->address() . " netmask $opts{prefix} up";
                $cleanupcmd = "$sudo $ifconfig $opts{device} inet removeif " . $ipObj->address();
            }
        } elsif ($OS =~ /Linux/i) {
            if ($ipObj->version() =~ /6/) {
                $cmd =
                  "$sudo $ifconfig $opts{device} inet6 add " . $ipObj->address() . "/$opts{prefix} up";
                $cleanupcmd =
                  "$sudo $ifconfig $opts{device} inet6 del " . $ipObj->address() . "/$opts{prefix}";
            } else {
                $cmd =
                  "$sudo $ifconfig $opts{device}:$index " . $ipObj->address() . " netmask $opts{prefix} up";
                $cleanupcmd = "$sudo $ifconfig $opts{device}:$index down";
                $index++;
            }
        } else {
            die config_result("$OS is not supported by this topology yet");
        }

        my $output = qx($cmd 2>&1);
        die config_result("$cmd failed: $output") if $?;
        $Log->comment("$cmd : $output");

        push(@{$self->{ip_cleanupcmds}}, $cleanupcmd);
    }

    $Log->comment("IP Setup Complete : " . qx($sudo $ifconfig -a 2>&1));

    $self->{vdev_cleanup} = $opts{device};
    $self->{ip_setup_id}  = $$;             # Save this for cleanup to take care
    $Log->exit();
}

sub _cleanup_ips {
    my $self = shift;
    $Log->enter();

    if ($self->{ip_setup_id} && ($self->{ip_setup_id} != $$)) {
        $Log->debug(
              "process $$ can not cleanup ips created by $self->{ip_setup_id}");
        return;
    }

    while (my $cmd = pop @{$self->{ip_cleanupcmds}}) {
        my $output = qx($cmd 2>&1);
        $Log->warning("$cmd failed: $output") if $?;
        $Log->comment("$cmd : $output");
    }

    $Log->exit();
}

sub _configure_instances {
    $Log->enter();
    my ($self) = @_;

    $self->SUPER::_configure_instances();

    my @clients = @{$self->{clients}};

    # Now that each client has its instances assigned. Go through all the clients, and assign
    # each of their instances an IP address. Do it this way so that the ranges are contiguous
    # within a client.
    
    my $ipManager = $self->{ipManager};
    my @netObjs = $ipManager->source_networks();
    my $netObj = $netObjs[0];

    # Go through the physical clients records
    foreach my $clientKey (@clients) {
        my $client = $self->{client_config}->{$clientKey};

        # How many instances are on here? 
        my $numInstances = scalar(keys(%{$client->{instances}}));

        # Bail out if there are no instances on this client
        if (!$numInstances) { 
            $self->{client_setup}->{$clientKey}->{SRCIP} = 0;
            next;
        }
      

        # Now spread the ips we collected onto the instances on this client
        my $firstTime = 1;
        foreach my $instanceKey (keys(%{$client->{instances}})) {
            my $instance = $client->{instances}->{$instanceKey};

            # Make sure to note that all the instances are to be synthetic
            $instance->{isSynthetic} = 1;

            # We may not be configuring SRCIP values for the clients. If that is the case, 
            # leave now.  
            next if (not $netObj);
 
            my $ipObj = $netObj->assign_next_address();

            if ($firstTime) { 
                # The last value of the client_setup hash for synthetics. 
                # The values getting passed to client_setup are in the format
                # expected by IPManager. startAddr/cidr[numOfAddrs]
                $self->{client_setup}->{$clientKey}->{SRCIP} = $ipObj->address() . "/" . $ipObj->cidr() . "[" . $numInstances . "]:" . $ipObj->gateway();
                $firstTime = 0;
            }

            $instance->{srcip} = $ipObj->address();

            # Synthetic clients need the srcip value on the test_options hash
            $self->{test_options}->{$client->{name}}->{$instance->{name}}
              ->{SRCIP} = $instance->{srcip};
        }
    }
    $Log->exit();
}

sub _start_ip_manager { 
    my ($self) = @_;


    # Execute the one from the parent
    return $self->SUPER::_start_ip_manager();
}
1;
