#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     WAFL QA tests - Default topology
##
## @author      gvg@netapp.com
##
## @pod here

=head1 NAME

WaflDefault - WAFL tests' default topology

=head1 SYNOPSIS
    
use Topology;
my $self = new Topology(topology=>"wafl_default");

=head1 DESCRIPTION

C<WaflDefault> defines default topology for WAFL tests. Most of
the WAFL tests use this simple, default topology.

This topology consists of a aggr and/or a volume (either trad or flex).
The following variables are inputs to this topology.

    $FILER_AGGR     - Aggr name
    $AGGR_SIZE      - Size of aggr
    $VOL_TYPE       - Type of volume to be created (flex or trad)
    $FILER_VOL      - Name of the volume to be used for testing
    $TRAD_SIZE      - Size of traditional volume
    $FLEX_SIZE      - Size of flexible volume

Default values are set in new().

=head2 filer_setup:

Existence of $FILER_AGGR is independant of any $SETUP_POLICY. If
aggr doesn't exist, it creates one. The following $SETUP_POLICYs 
are supported for volumes in this topology.

By default, $VOL_TYPE is determined by the ONTAP release. If it is
pre-scrimshaw, $VOL_TYPE is set to 'trad', otherwise it is set to
'flex'.

By default $FILER_VOL is set to "testvol_".$VOL_TYPE.

  done
    Topology setup is already done. It checks the existence of 
    $FILER_VOL and uses it.

  create
    Topology needs to be setup from the scratch. It creates
    $FILER_VOL.

  recreate
    Topology needs to be re-created. This option first destroys 
    and recreates $FILER_VOL, if it exists. Otherwise,
    it justs creates the volume.

  reuse
    Topology may or may not exist. Try to reuse it, if it exists.
    This is default. It reuses the $FILER_VOL, if it exists.
    If it does not exist, it creates it and uses it.

  private
    Not supported now. Reserved for future use.

  This routine registers all the volumes created or existed
  using C<register_resource> function, so that resources
  can be used by tests without using environment variables.
  This registered list is also used by C<Cleanup>.

=head2 client_setup:
    None

=head2 client_cleanup:
    None

=head2 filer_cleanup:

This follows stateful cleanup model. It means that
C<Cleanup> actions are merely based on C<Setup> actions. If
a volume has been created during C<Setup> phase, then it
is destroyed in C<Cleanup> phase. This ensures original
state of test environment is always preserved. However,
there are different levels of cleanup are given by
$CLEANUP_POLICY.
  
  none
    No cleanup actions takes place.

  test
    If any test specific actions took place, they are reverted.
    This affects cleaning up neither $FILER_VOL nor
    $FILER_AGGR.

  topology
    If $FILER_VOL and/or $FILER_AGGR have been created
    during C<Setup>, then they are destroyed. If they were
    reused without creating them, then no actions take place.
    This options ensures C<Topology> is set to original state
    at the end of testing.

  all
    $FILER_AGGR and $FILER_VOL (Topology) are destroyed with
    other test specific resource. No resources are left behind 
    after testing.

=head1 SEE ALSO

Topology Options TracePkg TestManagerLite 

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Last revised $Date: 2006/04/12 $
    
=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

package Topology::WaflDefault;

use base qw(Topology); # Define Super class

use Tharn;
use strict;

use Data::Dumper qw(Dumper);

use Aggregate qw(aggr_exists aggr_create aggr_purge);
use Volume qw(vol_exists vol_create vol_create_flex vol_purge vol_status);
use Filer::Version qw(get_wafl_version supported_feature);
use ExportFS qw(export_path);
use Admin::License qw(license unlicense is_licensed);

use TracePkg;
my $Trace = new TracePkg;

use Options;
my $Options = new Options;
my $Options_hash;

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {};
    $Trace->entry();

    $Options->parameters(definitions =>
			 [
			  {
			      name => "FLEX_SIZE",
			      default => "30m",
			      description => "Flexible volume size",
			  },
			  {
			      name => "TRAD_SIZE",
			      default => "3",
			      description => "Traditional volume size in number of disks",
			  },			  
			  {
			      name => "VOL_TYPE",
			      default => undef(),
			      description => "Volume type used for testing",
			  },
			  {
			      name => "FILER_VOL",
			      default => undef(),
			      description => "Name of the volume to be used for testing",
			  },
			  {
			      name => "AGGR_SIZE",
			      default => "5",
			      description => "Aggregate size in number of disks",
			  },
			  {
			      name => "FILER_AGGR",
			      default => "testaggr",
			      description => "Name of aggregate",
			  },
			  ]);
    $Options->process();

    bless($self, $class);
    return $self;
}

##############################################################

sub filer_setup {
    my $self = shift;

    # Gather the values require for Setup
    my $SETUP_POLICY = $self->get(parameter=>"SETUP_POLICY");
    my $FILER_CONN = $self->get(parameter=>"FILER_CONN");
    my $TIMEOUT    = $self->get(parameter=>"TIMEOUT");
    my $CMDTIMEOUT = $self->get(parameter=>"CMDTIMEOUT");
    my $VOL_TYPE   = $self->get(parameter=>"VOL_TYPE");
    my $FILER_AGGR = $self->get(parameter=>"FILER_AGGR");
    my $AGGR_SIZE  = $self->get(parameter=>"AGGR_SIZE");
    my $TRAD_SIZE  = $self->get(parameter=>"TRAD_SIZE");
    my $FLEX_SIZE  = $self->get(parameter=>"FLEX_SIZE");
    my $LICENSES   = $self->get(parameter=>"LICENSES");
    my $FILER_VOL  = $self->get(parameter=>"FILER_VOL");

    # Multi filer support - For now rely on FILER_LIST (FILERS - when?)
    my $FILER_LIST      = $self->get(parameter=>"FILER_LIST");
    my @filers = split(",", $FILER_LIST);
    
    if(@filers == 0) {
	die(config_result("FILER_LIST is not set"));
    }

    $Trace->debug(msg=>"Processing filers @filers for Setup stage");
    foreach my $filer (@filers) {
	$Trace->debug(msg=>"Processing FILER($filer) for Setup...");
	my $FILERS     = $self->get(parameter=>"$filer");
	my $vol_name;

	# Currently, TestManager set FILER as a comma separated array
	# Hack to make it work for functional tests.
	my $FILER = @$FILERS[0];

	$Trace->debug(msg=>"Host connect - $FILER");
	my $fh = host($FILER); # thats how we access
	my $fc = connect($FILER_CONN, -timeout, $TIMEOUT);
	$fc->timeout($CMDTIMEOUT);
	my $filer_name = $fh->id();
	
	$self->register_resource(filer_name=>$filer_name,
				 type=>"CONNECTION",
				 resource=>$fc);

	# License the required services first
	my @licenses = split(",", $LICENSES);
	foreach my $license (@licenses) {
	    if (!is_licensed($fc, $license)) {
		license($fc, $license)
		    or die(config_result($@));
	    }
	}
	# When VOL_TYPE is not set, try finding appropriate
	# VOL_TYPE for this filer. If filer is down and VOL_TYPE
	# is not set, fail.
	if(!$VOL_TYPE) {
	    $Trace->comment(msg=>"VOL_TYPE=$VOL_TYPE is not specified. Contacting filer/sim to assign appropriate default VOL_TYPE value...");

	    use config_base;    
	    use Filer::Version qw(get_wafl_version WV_SCRIMSHAW_MAX);

	    my $wv;
	    my $fh = host($FILER);
	    eval { $wv = get_wafl_version($fh); };
	    if ($@) {
		die(config_result("filer/sim is down or unreachable -- define \$VOL_TYPE to continue"));
	    } elsif ($wv > WV_SCRIMSHAW_MAX) {
		$self->set(VOL_TYPE=>"flex");
	    } else {
		$self->set(VOL_TYPE=>"trad");
	    }
	}
	
	$VOL_TYPE   = $self->get(parameter=>"VOL_TYPE");

	# Define $FILER_VOL if not user specified
	if(!$FILER_VOL) {
	    $vol_name = "testvol_".$VOL_TYPE;
	} else {
	    $vol_name = $FILER_VOL;
	}

	# Create aggr, if it doesn't exist
	if($VOL_TYPE =~ /flex/i) {
	    my $aggr_created = 0;
	    if(!aggr_exists($fc, $FILER_AGGR)) {
		aggr_create($fc, $FILER_AGGR, $AGGR_SIZE)
		    or die(config_result($@));
		$aggr_created = 1;
	    }
	    
	    # Register aggr to cleanup later in filer_cleanup
	    $self->register_resource(filer_name=>$filer_name,
				     type=>"AGGR",
				     pool=>"TOPOLOGY",
				     resource=>$FILER_AGGR,
				     opts => {
					 created => $aggr_created,
				     },
				     );
	}

	# Setup required volume
	my $vol_created = 0;
	if($SETUP_POLICY =~ /^reuse$/i) {
	    $Trace->comment(msg=>"Mode - $VOL_TYPE: Generic resources are being reused");
	    if(!vol_exists($fc, $vol_name)) {
		$vol_created = 1;
	    }
	} elsif($SETUP_POLICY =~ /^recreate$/i) {
	    $Trace->comment(msg=>"Mode - $VOL_TYPE: Generic resources are being recreated");
	    if(vol_exists($fc, $vol_name)) {
		$Trace->debug(msg=>"Vol $vol_name exists. Removing...");
		vol_purge($fc, $vol_name)
		    or die(config_result($@));
	    }
	    $vol_created = 1;
	} elsif($SETUP_POLICY =~ /^done$/i) {
	    # Default is assert
	    $Trace->comment(msg=>"Mode - $VOL_TYPE: Generic resources existence is validated and asserted");
	    if(!vol_exists($fc, $vol_name)) {
		die(config_result("SETUP_POLICY=$SETUP_POLICY, but Vol $vol_name doesn't exist"));
	    }	
	} elsif($SETUP_POLICY =~ /^create$/i) {
	    $Trace->comment(msg=>"Mode - $VOL_TYPE: Generic resources are being created");
	    if(vol_exists($fc, $vol_name)) {
		die(config_result("SETUP_POLICY=$SETUP_POLICY, but Vol $vol_name exists"));
	    }
	    $vol_created = 1;
	} elsif($SETUP_POLICY =~ /^private$/i) {
	    die(script_result("SETUP_POLICY=$SETUP_POLICY is not IMPLEMENTED yet"));
	} else {
	    die(config_result("SETUP_POLICY=$SETUP_POLICY is not supported"));
	}
	
	if($vol_created) {
	    if($VOL_TYPE =~ /trad/i) {
		$Trace->debug(msg=>"Creating traditional volume");
		vol_create($fc, $vol_name, $TRAD_SIZE)
		    or die(config_result($@));   		    
	    } elsif ($VOL_TYPE =~ /flex/i) {
		supported_feature($fh, 'FLEXVOL')
		    or die(na_result($@));
		
		$Trace->debug(msg=>"Creating flexible volume");
		vol_create_flex($fc, $vol_name, $FILER_AGGR, $FLEX_SIZE)
		    or die(config_result($@));
		
	    }
	    
	    my $vol_path = "/vol/".$vol_name;
	    export_path($fc, "$vol_path")
		or die(config_result($@));
	}

	# Register the volumes created for this topology
	# this list is useful for client setup and also
	# tests can access them when required
	$self->register_resource(filer_name=>$filer_name,
				 type=>"VOLUME",
				 pool=>"TOPOLOGY",
				 resource=>$vol_name,
				 opts => {
				     vol_type => $VOL_TYPE,
				     created => $vol_created,
				 },
				 );

	# check if the existed/created volume matches the correct type
	my $obj  = vol_status($fc, $vol_name)
	    or die(config_result($@)); 
	my $type = $obj->type();
	if ($type ne $VOL_TYPE) {
	    die(config_result("Existed/Created volume '$vol_name' is of the ".
			      "wrong type: $type. Expected $VOL_TYPE"));
	}
    }
    $Trace->exit(msg=>"");
}

##############################################################

sub client_setup {
    my $self = shift;
    $Trace->debug(msg=>"This topology doesn't do any client setup");
}

sub client_cleanup {
    my $self = shift;
    $Trace->debug(msg=>"This topology doesn't do any client cleanup");
}
    
##############################################################

sub filer_cleanup {
    my $self = shift;
    
    $Trace->entry();
    $Trace->comment(msg=>"STAGE Cleanup");

    my $FILER_CONN     = $self->get(parameter=>"FILER_CONN");
    my $TIMEOUT        = $self->get(parameter=>"TIMEOUT");
    my $CMDTIMEOUT     = $self->get(parameter=>"CMDTIMEOUT");
    my $CLEANUP_POLICY = $self->get(parameter=>"CLEANUP_POLICY");
    $Trace->comment(msg=>"CLEANUP_POLICY = $CLEANUP_POLICY");
    if($CLEANUP_POLICY =~ /^none$/i) {
	$Trace->comment(msg=>"Skipping CLEANUP");
	return 1;
    }

    # Multi filer support - For now rely on FILER_LIST (FILERS - when?)
    my $FILER_LIST      = $self->get(parameter=>"FILER_LIST");
    my @filers = split(",", $FILER_LIST);
    foreach my $filer (@filers) {
	my $FILERS     = $self->get(parameter=>$filer);

	# Currently, TestManager set FILER as a comma separated array
	# Hack to make it work for functional tests.
	my $FILER = @$FILERS[0];

	my $fh = host($FILER); # thats how we access
	my $fc = connect($FILER_CONN, -timeout, $TIMEOUT);
	$fc->timeout($CMDTIMEOUT);
	my $filer_name = $fh->id();

	my $pool;
	my $created;
	if($CLEANUP_POLICY =~ /^test$/i) {
	    # Remove resources registerd in 'test' pool
	    $pool = "TEST";
	} elsif($CLEANUP_POLICY =~ /^topology$/i) {
            # Remove resources created(only) in 'topology' pool
	    $pool = "TOPOLOGY";
	    $created = 1;
	} elsif($CLEANUP_POLICY =~ /^all$/i) {
	    # Remove resources registerd in 'topology' pool
	    $pool = "TOPOLOGY";	    
	}

	# Remove volumes
	my $volumes = $self->get_resource_list(filer_name=>$filer_name,
					       pool=>$pool,
					       type=>"VOLUME",
					       opts => {
						   created=>$created,
					       },
					       );
	foreach my $vol_name (@$volumes) {
	    _vol_purge($fc, $vol_name)
		or logresult("WARN", $@);	    
	}
	
	# Remove aggrs
	my $aggrs = $self->get_resource_list(filer_name=>$filer_name,
					     pool=>$pool,
					     type=>"AGGR",
					     opts => {
						 created=>$created,
					     },
					     );
	foreach my $aggr_name (@$aggrs) {
	    aggr_purge($fc, $aggr_name)
		or logresult($@);	    
	}
    }
    $Trace->exit(msg=>"");
}

sub _vol_purge
{
    my ($fc, $vol_name) = @_;

    # 11/24/05 bobj -- It turns out that cleanup is failing in some tests
    # with vol_purge giving this message:
    #     vol offline: Volume is in use by other operations
    # We are seeing this in quota tests, where quotas were turned off on
    # a volume before common_cleanup() is run. "quota status" says that
    # quotas are off for the volume, but actually the "quota off" operation
    # hasn't finished, quota metadata files are still being truncated.

    my $MAX_RETRIES = 5;
    my $RETRY_PAUSE = 60;
    
    for (my $try=0; $try <= $MAX_RETRIES; $try++) { 
	$Trace->debug(msg=>"Attempt to delete volume $vol_name");
	if(vol_purge($fc, $vol_name)) {
	    $Trace->debug(msg=>"Deleted volume $vol_name");
	    last;
	} else {
	    logresult("WARN", "Can't delete volume: $@ (retrying...)");
	    $Trace->debug(msg=>"$@");
	}
	sleep($RETRY_PAUSE);
    }

    return 1;
}

##############################################################

1;
