#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     WAFL QA Snaplock topology
##
## @author      gvg@netapp.com
##
## @pod here

=head1 NAME

WaflDefault - WAFL snaplock topology

=head1 SYNOPSIS
    
use Topology;
my $self = new Topology(topology=>"wafl_snaplock");

=head1 DESCRIPTION

C<WaflSnaplock> defines snaplock topology for WAFL tests. Most of
the WAFL-snaplock tests can use this topology.

This topology consists of a aggr and/or a volume (either trad or flex).
The following variables are inputs to this topology.

    $WORM_AGGR      - Snaplock Aggr name
    $AGGR_SIZE      - Size of aggr
    $VOL_TYPE       - Type of volume to be created (flex or trad)
    $WORM_VOL       - Name of snaplock volume
    $TRAD_SIZE      - Size of traditional volume
    $FLEX_SIZE      - Size of flexible volume
    $SNAPLOCK_TYPE  - Type of snaplock volumes to be created (enterprice/compliace)
    $WORM_RETENTION - Retention time for worm files/vols
    $SKIP_SNAPLOCK_RETENTION_SETUP - Boolean to control snaplock retention setup
    $FILER_AGGR     - Aggr name
    $FILER_VOL      - Name of the volume

=head2 filer_setup:

Existence of $WORM_AGGR are independant of any $SETUP_POLICY. 
If aggr doesn't exist, it creates one with specified 
$SNAPLOCK_TYPE. The following $SETUP_POLICYs are supported for 
volumes in this topology.

By default, $VOL_TYPE is determined by the ONTAP release. If it is
pre-scrimshaw, $VOL_TYPE is set to 'trad' and $WORM_VOL is created
as traditional volume. Otherwise, $VOL_TYPE is set to 'flex' and 
$WORM_VOL is created as flexible volume on $WORM_AGGR.

Default name for $WORM_VOL is set to "wormvol_".$SNAPLOCK_TYPE."_".
$VOL_TYPE. Default name of $WORM_AGGR is set to "wormaggr_".$SNAPLOCK
_TYPE.

The type of $WORM_VOL is determined (and created) by $SNAPLOCK_TYPE.
Supported values are 'snaplock' and 'snaplock_enterprice'.

$FILER_VOL and/or $FILER_AGGR are inherited from 
WaflDefault topology. See WaflDefault.pm for details.

  done
    Topology setup is already done. It checks the existence of 
    $WORM_VOL and uses it.

  create
    Topology needs to be setup from the scratch. It creates
    $WORM_VOL.

  recreate
    Topology needs to be re-created. This option first destroys 
    and recreates $WORM_VOL, if it exists. Otherwise,
    it justs creates the volume.

  reuse
    Topology may or may not exist. Try to reuse it, if it exists.
    This is default. It reuses the $WORM_VOL, if it exists.
    If it does not exist, it creates it and uses it.

  private
    Not supported now. Reserved for future use.

  This routine registers all the volumes created or existed
  using C<register_resource> function, so that resources
  can be used by tests without using environment variables.
  This registered list is also used by C<Cleanup>.

=head2 client_setup:
    None

=head2 client_cleanup:
    None

=head2 filer_cleanup:

This follows stateful cleanup model. It means that
C<Cleanup> actions are merely based on C<setup> actions. If
a volume has been created during C<setup> phase, then it
is destroyed in C<cleanup> phase. This ensures original
state of test environment is always preserved. However,
there are different levels of cleanup are given by
$CLEANUP_POLICY.
  
  none
    No cleanup actions takes place.

  test
    If any test specific actions took place, they are reverted.
    This affects cleaning up neither $WORM_VOL nor
    $WORM_AGGR.

  topology
    If $WORM_VOL and/or $WORM_AGGR have been created
    during C<setup>, then they are destroyed. If they were
    reused without creating them, then no actions take place.
    This options ensures C<Topology> is set to original state
    at the end of testing.

  all
    $WORM_AGGR and $WORM_VOL (Topology) are destroyed with
    other test specific resource. No resources are left behind 
    after testing.

=head1 SEE ALSO

Topology Options TracePkg TestManagerLite WaflDefault

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Last revised $Date: 2006/04/12 $
    
=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

package Topology::WaflSnaplock;

use base qw(Topology::WaflDefault); # Define Super class

use Tharn;
use strict;

use Data::Dumper qw(Dumper);

use Aggregate qw(aggr_exists aggr_create aggr_purge);
use Volume qw(vol_exists vol_create vol_create_flex vol_purge vol_status);
use Filer::Version qw(
		      get_wafl_version 
		      supported_feature 
		      UNSUPPORTED_FEATURE_EXCEPTION);
use ExportFS qw(export_path);
use Admin::License qw(license unlicense is_licensed);
use snaplock qw(set_default_retention_period compliance_clock_initialize);

use TracePkg;
my $Trace = new TracePkg;

use Options;
my $Options = new Options;
my $Options_hash;

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = $invocant->SUPER::new;
    $Trace->entry();

    $Options->parameters(definitions =>
			 [
			  {
			      name => "WORM_AGGR",
			      default => undef(),
			      description => "Name for snaplock aggregate",
			  },
			  {
			      name => "WORM_RETENTION",
			      default => "30s",
			      description => "Retention time period for worm volumes",
			  },
			  {
			      name => "SKIP_SNAPLOCK_RETENTION_SETUP",
			      default => "0",
			      description => "Boolean to control snaplock retention setup",
			  },
			  {
			      name => "FLEX_SIZE",
			      default => "30m",
			      description => "Flexible volume size",
			  },
			  {
			      name => "TRAD_SIZE",
			      default => "3",
			      description => "Traditional volume size in number of disks",
			  },
			  {
			      name => "VOL_TYPE",
			      default => undef(),
			      description => "Type of volume (flex/trad)",
			      valid_values => ["trad","flex"],
			  },
			  {
			      name => "SNAPLOCK_TYPE",
			      default => "snaplock_enterprise",
			      description => "Type of snaplock volume ('snaplock_enterprice' for enterprise or 'snaplock' for compliance)",
			      valid_values => ["snaplock", "snaplock_enterprise"],
			  },
			  {
			      name => "WORM_VOL",
			      default => undef(),
			      description => "Name for snaplock volume",
			  },
			  ]);
    $Options->process();

    bless($self, $class);
    return $self;
}

##############################################################

sub filer_setup {
    my $self = shift;

    # Call super class to setup FILER_VOL and/or FILER_AGGR.
    $self->SUPER::filer_setup;

    # Gather the values require for Setup
    my $SETUP_POLICY = $self->get(parameter=>"SETUP_POLICY");
    my $FILER_CONN = $self->get(parameter=>"FILER_CONN");
    my $TIMEOUT    = $self->get(parameter=>"TIMEOUT");
    my $CMDTIMEOUT = $self->get(parameter=>"CMDTIMEOUT");
    my $VOL_TYPE   = $self->get(parameter=>"VOL_TYPE");
    my $WORM_AGGR  = $self->get(parameter=>"WORM_AGGR");
    my $AGGR_SIZE  = $self->get(parameter=>"AGGR_SIZE");
    my $TRAD_SIZE  = $self->get(parameter=>"TRAD_SIZE");
    my $FLEX_SIZE  = $self->get(parameter=>"FLEX_SIZE");
    my $LICENSES   = $self->get(parameter=>"LICENSES");
    my $WORM_VOL   = $self->get(parameter=>"WORM_VOL");

    my $SNAPLOCK_TYPE  = $self->get(parameter=>"SNAPLOCK_TYPE");
    my $SKIP_SNAPLOCK_RETENTION_SETUP = $self->get(parameter=>"SKIP_SNAPLOCK_RETENTION_SETUP");
    my $WORM_RETENTION = $self->get(parameter=>"SKIP_SNAPLOCK_RETENTION_SETUP");
    # Set the names for WORM_AGGR and WORM_VOL
    if(!$WORM_AGGR) {
	$WORM_AGGR = "wormaggr_".$SNAPLOCK_TYPE;	
    }
    if(!$WORM_VOL) {
	$WORM_VOL = "wormvol_".$SNAPLOCK_TYPE."_".$VOL_TYPE;
    }

    # Multi filer support - For now rely on FILER_LIST (FILERS - when?)
    my $FILER_LIST      = $self->get(parameter=>"FILER_LIST");
    my @filers = split(",", $FILER_LIST);
    
    if(@filers == 0) {
	die(config_result("FILER_LIST is not set"));
    }

    $Trace->debug(msg=>"Processing filers @filers for Setup stage");
    foreach my $filer (@filers) {
	$Trace->debug(msg=>"Processing FILER($filer) for Setup...");
	my $FILERS     = $self->get(parameter=>"$filer");
	my $vol_name = $WORM_VOL;

	# Currently, TestManager set FILER as a comma separated array
	# Hack to make it work for functional tests.
	my $FILER = @$FILERS[0];

	$Trace->debug(msg=>"Host connect - $FILER");
	my $fh = host($FILER); # thats how we access
	my $fc = connect($FILER_CONN, -timeout, $TIMEOUT);
	$fc->timeout($CMDTIMEOUT);
	my $filer_name = $fh->id();
	
	$self->register_resource(filer_name=>$filer_name,
				 type=>"CONNECTION",
				 resource=>$fc);

	# License the required services first
	my @licenses = split(",", $LICENSES);
	foreach my $license (@licenses) {
	    if (!is_licensed($fc, $license)) {
		license($fc, $license)
		    or die(config_result($@));
	    }
	}

	# License appropriate snaplock license
	if($SNAPLOCK_TYPE =~ /snaplock_enterprise/i) {
	    if (!is_licensed($fc, "snaplock_enterprise")) {
		license($fc, "snaplock_enterprise")
		    or die(config_result($@));
	    }
	} else {
	    if (!is_licensed($fc, "snaplock")) {
		license($fc, "snaplock")
		    or die(config_result($@));
	    }
	}

	# When VOL_TYPE is not set, try finding appropriate
	# VOL_TYPE for this filer. If filer is down and VOL_TYPE
	# is not set, fail.
	if(!$VOL_TYPE) {
	    $Trace->comment(msg=>"VOL_TYPE=$VOL_TYPE is not specified. Contacting filer/sim to assign appropriate default VOL_TYPE value...");

	    use config_base;    
	    use Filer::Version qw(get_wafl_version WV_SCRIMSHAW_MAX);

	    my $wv;
	    my $fh = host($FILER);
	    eval { $wv = get_wafl_version($fh); };
	    if ($@) {
		die(config_result("filer/sim is down or unreachable -- define \$VOL_TYPE to continue"));
	    } elsif ($wv > WV_SCRIMSHAW_MAX) {
		$self->set(VOL_TYPE=>"flex");
	    } else {
		$self->set(VOL_TYPE=>"trad");
	    }
	}
	
	$VOL_TYPE   = $self->get(parameter=>"VOL_TYPE");

	# Check for flexvol support to make sure correct VOL_TYPE
	# is set
	if($VOL_TYPE =~ /flex/i) {
	    supported_feature($fh, 'FLEXVOL')
		or die(na_result($@));
	}

	# Check if post-scrimshaw (ONTAP) is running. If it is,
	# initialize the compilance clock.
	my $ver = get_wafl_version($fc->{hostp});
	if ($ver > WV_SCRIMSHAW_MAX) {
	    $Trace->comment(msg=>"initialize the compliance clock (if it is not initialized");
	    compliance_clock_initialize($fc, reinit_ok => 1);
	}

	# Pass snaplock options appropriately to create aggr/vol
	my %args = ();
	$args{snaplock} = 1 if ($SNAPLOCK_TYPE =~ /^snaplock$/i);
	$args{snaplock_enterprise} = 1 if ($SNAPLOCK_TYPE =~ /^snaplock_enterprise$/i);

	# Create aggr, if it doesn't exist
	if($VOL_TYPE =~ /flex/i) {
	    my $aggr_created = 0;
	    if(!aggr_exists($fc, $WORM_AGGR)) {
		aggr_create($fc, $WORM_AGGR, $AGGR_SIZE, %args)
		    or die(config_result($@));
		$aggr_created = 1;
	    }
	    
	    # Register aggr to cleanup later in filer_cleanup
	    $self->register_resource(filer_name=>$filer_name,
				     type=>"AGGR",
				     pool=>"TOPOLOGY",
				     resource=>$WORM_AGGR,
				     opts => {
					 created => $aggr_created,
					 snaplock_type => $SNAPLOCK_TYPE,
				     },
				     );
	}

	# Setup required volume
	my $vol_created = 0;
	if($SETUP_POLICY =~ /^reuse$/i) {
	    $Trace->comment(msg=>"Mode - $VOL_TYPE: Generic resources are being reused");
	    if(!vol_exists($fc, $vol_name)) {
		$vol_created = 1;
	    }
	} elsif($SETUP_POLICY =~ /^recreate$/i) {
	    $Trace->comment(msg=>"Mode - $VOL_TYPE: Generic resources are being recreated");
	    if(vol_exists($fc, $vol_name)) {
		$Trace->debug(msg=>"Vol $vol_name exists. Removing...");
		vol_purge($fc, $vol_name)
		    or die(config_result($@));
	    }
	    $vol_created = 1;
	} elsif($SETUP_POLICY =~ /^done$/i) {
	    # Default is assert
	    $Trace->comment(msg=>"Mode - $VOL_TYPE: Generic resources existence is validated and asserted");
	    if(!vol_exists($fc, $vol_name)) {
		die(config_result("SETUP_POLICY=$SETUP_POLICY, but Vol $vol_name doesn't exist"));
	    }	
	} elsif($SETUP_POLICY =~ /^create$/i) {
	    $Trace->comment(msg=>"Mode - $VOL_TYPE: Generic resources are being created");
	    if(vol_exists($fc, $vol_name)) {
		die(config_result("SETUP_POLICY=$SETUP_POLICY, but Vol $vol_name exists"));
	    }
	    $vol_created = 1;
	} elsif($SETUP_POLICY =~ /^private$/i) {
	    die(script_result("SETUP_POLICY=$SETUP_POLICY is not IMPLEMENTED yet"));
	} else {
	    die(config_result("SETUP_POLICY=$SETUP_POLICY is not supported"));
	}
	
	if($vol_created) {
	    if($VOL_TYPE =~ /trad/i) {
		$Trace->debug(msg=>"Creating traditional volume");
		vol_create($fc, $vol_name, $TRAD_SIZE, %args)
		    or die(config_result($@));   		    
	    } elsif ($VOL_TYPE =~ /flex/i) {		
		$Trace->debug(msg=>"Creating flexible volume");
		vol_create_flex($fc, $vol_name, $WORM_AGGR, $FLEX_SIZE)
		    or die(config_result($@));
		
	    }
	    
	    my $vol_path = "/vol/".$vol_name;
	    export_path($fc, "$vol_path")
		or die(config_result($@));
	}

	# Register the volumes created for this topology
	# this list is useful for client setup and also
	# tests can access them when required
	$self->register_resource(filer_name=>$filer_name,
				 type=>"VOLUME",
				 pool=>"TOPOLOGY",
				 resource=>$vol_name,
				 opts => {
				     vol_type => $VOL_TYPE,
				     created => $vol_created,
				     snaplock_type => $SNAPLOCK_TYPE,
				 },
				 );

	# check if the existed/created volume matches the correct type
	my $obj  = vol_status($fc, $vol_name)
	    or die(config_result($@)); 
	my $type = $obj->type();
	if ($type ne $VOL_TYPE) {
	    die(config_result("Existed/Created volume '$vol_name' is of the ".
			      "wrong type: $type. Expected $VOL_TYPE"));
	}
	my %vopts = $obj->options();
	if ($SNAPLOCK_TYPE =~ /^snaplock$/i and !$vopts{snaplock_compliance}) {
	    die(config_result("volume $vol_name not a SnapLock volume"));
	}
	if ($SNAPLOCK_TYPE =~ /^snaplock_enterprise$/i and !$vopts{snaplock_enterprise}) {
	    die(config_result("volume $vol_name not a SnapLock Enterprise volume"));
	}
	
	# Set retention period on the volume created.
	my $retention = eval {
	    return set_default_retention_period(
						$fc,
						$WORM_RETENTION,
						$WORM_VOL
						);
	};
	if ($@) {
	    if (index($@, UNSUPPORTED_FEATURE_EXCEPTION) != -1) {
		logWarn(UNSUPPORTED_FEATURE_EXCEPTION);
	    } else {
		die(config_result($@));
	    }
	}
	$Trace->debug(msg=>"now, retention is '$retention'");
    }
    $Trace->exit(msg=>"");
}

##############################################################

sub client_setup {
    my $self = shift;
    $Trace->debug(msg=>"This topology doesn't do any client setup");
}

sub client_cleanup {
    my $self = shift;
    $Trace->debug(msg=>"This topology doesn't do any client cleanup");
}

##############################################################

sub filer_cleanup {
    my $self = shift;
    
    # Just call parent class to cleanup volumes and aggrs

    $self->SUPER::filer_cleanup;
}

##############################################################

1;
