# -*- perl -*-
# $Id: $
## @summary Package to control logging of messages.
## @author Kevin Olalde, Bill Snider
## @pod here

=head1 NAME

TracePkg.pm - Log message package

=head1 DESCRIPTION

C<TracePkg> provides an object and methods to allow fine control over the 
messages that are printed by tests and libraries.

A lot more information needs to be filled in here to better describe how
to use this package.

=head1 SYNOPSIS

    use TracePkg;
    my $trace = new TracePkg;

    $trace->entry(msg=>"input is $input");
    $trace->comment(msg=>"This message will be printed");
    $trace->trace(msg=>"This is a trace statement");
    $trace->debug(msg=>"This is a debugging statement");
    $trace->fatal(msg=>"This message is equivalent to a die");
    $trace->fatal(msg=>"This message is equivalent to a die script_result",
                  result=>"SCRIPT");
    $trace->warn(msg=>"This is a warning");

=head1 ROUTINES

=cut

#
#  Class that provides some tracing functions.

# Syntax	    
#<traceDetails> ::= <traceSpec> { ";" <traceSpec> }
#<traceSpec> ::= <packageNames>"="<levels>"="enabled|disabled","<traceStream>
#<packageNames> ::= <packageName> { "-" <packageName> }
#<packageName> ::= "*" | any character sequence
#<levels> ::= <level> { "-" <level> }
#<level> ::= all|normal|comment|error|warn|call|trace|debug|<custom>
#<custom> ::= any character sequence
#<traceStream> ::= STDOUT|STDERR|<fileSpec>
#<fileSpec> ::= valid path and file name
#

package TracePkg;
use strict;
use Carp;
use IO::File; 
use Env qw(NTAP_QA_TRACE);
use Data::Dumper;

# Class Data, only accessible from this class
{
    my $refCount = 0;
    
    sub incRef { 
	$refCount++;
    }
    sub getRef {
	return $refCount;
    }
}

# Fill the member vars for this instance
# Do not define any values for the vars because they will be
#   overridden when new() is called.
my $Object;

# the supported levels
my @Supported_levels = qw(comment warn call trace debug fatal);

# some level wildcards
my @All_levels = qw(comment warn call trace debug fatal);
my @Normal_levels = qw(comment warn call trace fatal);
my @Default_levels = qw(comment warn fatal);

my %Levels = (
              'C' => "comment", "comment" => "comment",
              'W' => "warn", "warn" => "warn",
              '>' => "call", "call" => "call",
              '<' => "call",
              'T' => "trace", "trace" => "trace",
              'D' => "debug", "debug" => "debug",
              'F' => "fatal", "fatal" => "fatal",
             );

=head2 new

C<new> is the costructor for the object.

=over

=item _traceD

(Optional)

=item _separateNewline

(Optional) Setting this to 1 will cause each line in a newline-separated
input message to be printed with the header information. The default is
undef which causes the first line to have the header and the remaining line
with none.

=item _nate

(Optional) Setting this to 1 will indicate that the script is running
inside of an environment in which Tharn has been loaded into the main
namespace. This check is normally done at startup time and set automatically.
If a script/library is coded in a manner that causes Tharn to be loaded later
then this flag can be set to indicate so.

=back

=cut

sub new {
    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    #  a hash.
    my ($self, %arg) = @_;
    if ( getRef()) {
	#$Object->event(msg=>"TracePkg already initialized\n"  );
	# BURT 389535 - Added a check for package. If called from 
	#               LogAnalyzer, it will return a new object. 
	my $package=caller;
	if ($package !~ /^LogAnalyzer::Parser::/){
	    return $Object;
	}
    }

    # Fill the member vars for this instance
    $Object = {
	_traceD         => $arg{traceD},
	_tPkgHash       => undef,
	_separateNewline => $arg{separateNewline},
    };

    # Decide if we are running inside of NATE.
    # We only need RUNID and logresult which are both available when running
    #  inside of ntest. If $::RUNID is available, then we are running inside
    #  of ntest, so both will be available. We will set an internal flag so
    #  that the rest of the code knows.
    $Object->{_nate} = 1 if (defined($::RUNID) && $::RUNID);

    # This tags the hash pointed to by $Object as a 
    #  type of TracePkg
    bless $Object, $self;

    if ( !(defined $Object->{_traceD}) ) {
	# Get the trace details from the env if none was passed
	if ( $NTAP_QA_TRACE ) {
	    $Object->{_traceD} = $NTAP_QA_TRACE;
	} else {
	    $Object->{_traceD} = '*=comment-fatal-warn=enabled';
	}
    } 
    # Set up the trace package hash of trace specs and streams
    my ($tDetail, $tSpec, $tPackage, $tLevel, $tStream);
    my ($tPkgs, $tLvls, $tEnabled);
    my $fh = undef;
    
    # The default levels
    foreach my $level (@Default_levels) {
	$Object->_set_level("*", $level, "STDOUT", "enabled");
    }
    
    # Peocess each spec, stream combo
    for $tDetail ( split( /;/, $Object->{_traceD}) ) {
	#print "Processig trace spec: $tDetail\n";
	
	# Get the stream
	($tSpec, $tStream) = split( /,/, $tDetail);
	if ( !defined $tSpec ) {
	    confess("\"traceSpec\" not found while processing\n\t$tDetail");
	} elsif (!defined $tStream) {
	    $tStream = "STDOUT";
	}

	# Get the package level info
	($tPkgs, $tLvls, $tEnabled) = split( /=/, $tSpec);
	if ( !(defined $tPkgs) ||
	     !(defined $tLvls) ||
	     !(defined $tEnabled) ) {
	    confess("\"pkgs=lvls=enabled\" format not found while processing\n\t$tSpec");
	}
	# Set trace for eack package, for this stream
	for $tPackage ( split( /-/, $tPkgs) ) {
	    
	    # Now open the trace streams
	    if ( $tStream ne "STDOUT" &&
		 $tStream ne "STDERR" ) {
		$fh = new IO::File;
		$fh->open("> $tStream") || die("Can't open $tStream: $!");
	    } else {
		$fh = $tStream;
	    }
	    for $tLevel ( split( /-/, $tLvls) ) {
		if ($tLevel eq "all") {
		    foreach my $level (@All_levels) {
			$Object->_set_level($tPackage, $level, $fh, $tEnabled);
		    }
		} elsif ( $tLevel eq "normal" ) {
		    foreach my $level ( @Normal_levels ) {
			$Object->_set_level($tPackage, $level, $fh, $tEnabled);
		    }
		} elsif ( grep ( $tLevel eq $_, @Supported_levels ) ) {
		    $Object->_set_level($tPackage, $tLevel, $fh, $tEnabled);
		} else {
		    croak "'$tLevel' is an unknown level for TracePkg\n";
		}
	    }
	}
    }
    
    # Mark that we've been here
    incRef();
    
    # And give the caller the referece to the bless object.
#    use Data::Dumper;
#    print Dumper($Object);
    return $Object;
}

sub _set_level {
    my ($self, $pkg, $level, $fh, $enabled) = @_;
#    print "Setting package: $pkg to $enabled for level ",
#          "$level on stream $fh\n";
    $self->{_tPkgHash}{$pkg}{$level}{_fh} = $fh;
    if ( $enabled =~ /enabled/ ) {
	$self->{_tPkgHash}{$pkg}{$level}{_isEnabled} = 1;
    } else {
	$self->{_tPkgHash}{$pkg}{$level}{_isEnabled} = 0;
    }	
}

=head2 set_options

C<set_options> allows internal options to be set

=cut

sub set_options {
    my ($self, %arg) = @_;
    while ( my ( $option, $value ) = each %arg ) {
	if ( exists $self->{"_$option"} ) {
	    $self->{"_$option"} = $value;
	} else {
	    confess("Unknown option '$option=>$value' encountered in TracePkg::set_options\n");
	}
    }
}

=head2 fatal

C<fatal> will cause a fatal message to be logged if the C<NTAP_QA_TRACE> spec
is set to include the fatal level.

The input message will be prepended with a 'F' character.

=over

=item msg

(Optional) This is the message that is printed out. It is optional but should
be included.

=item result

(Optional) This is a NATE result type. If it is specified then the call will
result in the equivalent of 

die log_result(result, msg);

List of possible NATE results types: SCRIPT, CONFIG, FATAL, FAIL, INSPECT, NA,
WARN, INFO, PARAM, PASS

For more info on these result types see:

http://web.netapp.com/engineering/nate/releases/latest/html/Tharn.html

=back

=cut

sub fatal {
    my ($self, %arg) = @_;
    $self->log_msg( %arg,
		    level => "fatal",
		    char  => "F",
		    fatal => 1);
}

=head2 comment

C<comment> will cause a  message to be logged if the C<NTAP_QA_TRACE> spec
is set to include the comment level.

The input message will be prepended with a 'C' character.

=over

=item msg

(Optional) This is the message that is printed out. It is optional but should
be included.

=back

=cut

sub comment {
    my ($self, %arg) = @_;
    $self->log_msg( %arg,
		    level => "comment",
		    char  => "C" );
}

=head2 warn

C<warn> will cause a  message to be logged if the C<NTAP_QA_TRACE> spec
is set to include the warn level.

The input message will be prepended with a 'W' character.

=over

=item msg

(Optional) This is the message that is printed out. It is optional but should
be included.

=back

=cut

sub warn {
    my ($self, %arg) = @_;
    $self->log_msg( %arg,
		    level => "warn",
		    char  => "W" );
}

=head2 entry

C<entry> will cause a  message to be logged if the C<NTAP_QA_TRACE> spec
is set to include the call level.

The input message will be prepended with a '>' character.

=over

=item msg

(Optional) This is the message that is printed out. It is optional but should
be included.

=back

=cut

sub entry {
    my ($self, %arg) = @_;
    $self->log_msg( %arg,
		    level => "call",
		    char  => ">" );
}

=head2 exit

C<exit> will cause a  message to be logged if the C<NTAP_QA_TRACE> spec
is set to include the call level.

The input message will be prepended with a '<' character.

=over

=item msg

(Optional) This is the message that is printed out. It is optional but should
be included.

=back

=cut

sub exit {
    my ($self, %arg) = @_;
    $self->log_msg( %arg,
		    level => "call",
		    char  => "<" );
}

=head2 trace

C<trace> will cause a  message to be logged if the C<NTAP_QA_TRACE> spec
is set to include the trace level.

The input message will be prepended with a 'T' character.

=over

=item msg

(Optional) This is the message that is printed out. It is optional but should
be included.

=back

=cut

sub trace {
    my ($self, %arg) = @_;
    $self->log_msg( %arg,
		    level => "trace",
		    char  => "T" );
}

=head2 debug

C<debug> will cause a  message to be logged if the C<NTAP_QA_TRACE> spec
is set to include the debug level.

The input message will be prepended with a 'D' character.

=over

=item msg

(Optional) This is the message that is printed out. It is optional but should
be included.

=back

=cut

sub debug {
    my ($self, %arg) = @_;
    $self->log_msg( %arg,
		    level => "debug",
		    char  => "D" );
}

# I believe this is normally only used internally so I am not
# providing POD for it
sub log_msg  {
    my $self = shift;
    my (%arg) = @_;
    my $level = $arg{level}; delete $arg{level};
    my $char = $arg{char}; delete $arg{char};
    my $msg = $arg{msg}; delete $arg{msg};
    my $result = $arg{result}; delete $arg{result};
    my $fatal = $arg{fatal}; delete $arg{fatal};
    if (%arg) {
	$self->warn(msg=>"Parameter not understood: " . Dumper(%arg));
	$msg = "<<<<<<< THIS LINE IN ERROR >>>>>>> Parameter not understood: " . 
        Dumper(%arg) . "\nOriginal message: $msg";
    }
    
    my ($stream, $now, $fh, $pkgHName);
    
    # Grab caller information
    my ($pkgName, $fileName, $line, $subName) = caller(2);
    (undef, undef, $line, undef) = caller(1);
    
    if ( !(defined $pkgName) || $subName eq "(eval)") {
	# Went back too far
	($pkgName, $fileName, $line) = caller(1);
	$subName = $pkgName. "::main";
    }
    $subName =~ /(\S+)::(\S+)$/;
    # pkgHName is used for searching in the hash of spec definitions
    $pkgHName = $1;
    my $script = ( $self->{_nate} ) ? "[$::RUNID] $char" : "$char $pkgName";
    
    #print "From caller $pkgName $pkgHName $subName $line\n";
    
    my $message_enabled = $self->message_enabled(level=>$level, package=>\$pkgHName);
    
    
    # Set up the stream to use
    if ( $level && $message_enabled ) {
	$fh = $self->{_tPkgHash}{$pkgHName}{$level}{_fh};
        if (!$fh) {
            croak "Could not find the filehandle for $pkgHName $level\n";
        }
    } else {
	$fh = "STDOUT";
    }

    my ($sec,$min,$hr,$day,$mon,$yr,$dow,$doy,$isdst) = localtime(time);
    $yr += 1900; $mon += 1;
    $now = sprintf("%04d%02d%02d %02d%02d%02d", $yr,$mon,$day,$hr,$min,$sec);
    no strict 'refs';

    my $header = "$now $script $subName:$line";
    my $message = "";

    # If msg is undefined then the user did not pass one so just
    # print a the header and a newline
    if (!$msg) {
	$message = "$header\n";
    } else {
	# a msg was specified
	# tag on a newline to the end of the string if it was not present
	if (substr($msg, -1) ne "\n") {
	    $msg .= "\n";
	}

	# Each newline-separated message needs to be prepended with the header
	#  if the option is enabled
	if ( $self->{_separateNewline} ) {
	    foreach my $msg ( split(/\n/, $msg) ) {
		$message .= "$header $msg\n";
	    }
	} else {
	    # the message is simply the header and the msg
	    $message = "$header $msg";
	}
    }
    
  # print the message with the header if we are not supposed to skip it
    print $fh $message if ($message_enabled);

    # handle returning results inside of NATE
    if ( $self->{_nate} ) {
	if ($level eq "warn" ) {
            # pass the message without the header to logresult
	    main::logresult("WARN", $msg);
	} elsif ($result && $fatal) {
            if (ref($result)) {
		die main::logresult($result);
	    } else {
                # pass the message without the header to logresult
		die main::logresult($result, abort=>1, $msg);
	    }
	} elsif ($result) {
	    if (ref($result)) {
		return main::logresult($result);
	    } else {
                # pass the message without the header to logresult
		return main::logresult($result, $msg);
	    }
	}
    } 

    if ( $fatal ) {
        # die with just the message and no header
	croak $msg;
    }

    return $message;

}	

=head2 message_enabled

C<message_enabled> is used to determine whether or not a trace statement
should be enabled.

=over

=item level

(Required) This specifies the given trace level for the trace statement.

item method

(Required) This specifies the given method for the trace statement.

=item package

(Required) This specifies the given package name for the trace statement.

=back

=cut

sub message_enabled {
    my $self = shift;
    my %params = (
                  level => undef,
                  method => undef,
                  package => undef,
                  @_,
                  );
    my $level = $Levels{$params{level}};
    my $method = $params{method};
    my $package;
    if (ref($params{package}) eq "SCALAR") {
        $package = ${$params{package}};
    } else {
        $package = $params{package};
    }

    # Does definition exist for pkg of executing sub 
  PKGSEARCH: {
	# First look for exact match
        if (!exists($self->{_tPkgHash}{$package})) {
            # Look for next best match, a package name in the spec
            #  that ends with a "*"
            for my $pkgKey ( keys %{ $self->{_tPkgHash} } ) {
                if ( $pkgKey =~ /(\S+)\*/ ) {
                    if ( $package =~ /^$1/ ) {
                        $package = $pkgKey;
                        last PKGSEARCH;
                        # If needed, we could optimize this so a successful
                        # search resulted in creating an exact match
                        # in the trace spec hash matching the wildcard spec.
                    }
                }
            }
            # Didn't find it, how about just *
            if ( exists $self->{_tPkgHash}{"*"}{$level}) {
                # Found it, set the pkg name for the hash
                $package = "*";
            } 
        } elsif (!exists($self->{_tPkgHash}{$package}{$level})) {
            # Didn't find it, how about just *
            if ( exists $self->{_tPkgHash}{"*"}{$level}) {
                # Found it, set the pkg name for the hash
                $package = "*";
            }
        }
    }
  
  #  print "$package $level $method\n";
  #  print Dumper $self->{_tPkgHash};
  
  
  my $message_enabled = undef;
  # decide whether or not the message should be printed
  if ( exists($self->{_tPkgHash}{$package}) &&
       exists($self->{_tPkgHash}{$package}{$level})) {
      $message_enabled = $self->{_tPkgHash}{$package}{$level}{_isEnabled};
  } elsif (grep($level eq $_, @Default_levels)) {
      $message_enabled = 1;
  }
  
  if (ref($params{package}) eq "SCALAR") {
      ${$params{package}} = $package;
  }
  return $message_enabled;
}


=head1 PARAMS 

=over

=item NTAP_QA_TRACE

This param defines the trace spec. More info is needed.

  Syntax	    
  <traceDetails> ::= <traceSpec> { ";" <traceSpec> }
  <traceSpec> ::= <packageNames>"="<levels>"="enabled|disabled","<traceStream>
  <packageNames> ::= <packageName> { "-" <packageName> }
  <packageName> ::= "*" | any character sequence
  <levels> ::= <level> { "-" <level> }
  <level> ::= all|normal|comment|error|warn|call|trace|debug|<custom>
	all=all levels
	normal=comment-warn-fatal
  <custom> ::= any character sequence
  <traceStream> ::= STDOUT|STDERR|<fileSpec>
  <fileSpec> ::= valid path and file name

=back

=head1 SEE ALSO

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Kevin Olalde (olalde@netapp.com)
  Bill Snider (snider@netapp.com)
  $Id: $
  Last revised $Date: $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;

