# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Traffic.pm#1 $

## @summary Various client traffic and stress routines
##
## @description
##    This package defines routines for starting/stopping various types
##    of traffic to a filer, usually from a client.  The idea is to be
##    able to start and stop the traffic as desired by the caller script
##    without relying on excessively large timeouts, to supply the traffic
##    to the desired mountpoints, and to properly clean up afterwards.
##
##    Another goal is to be able to run many types of traffic in the background
##    without interfering with other tests.  Hence, particular attention
##    is paid to using non-inherited connections and not overrunning the
##    console (and any parent scripts which might be reading from it).
##
##    The motivation for providing these routines as wrappers for the
##    backgrounded subtests which do the real work is that as subroutines
##    they can return more meaningful and complicated values (which improve
##    both readability and modularity) and serve as a generic interface
##    for traffic tests -- which makes them more easily extensible.
##
##    Each routine takes two arguments:  a boolean value of whether to start or
##    stop the traffic, and a reference to a hash of options for the traffic.
##    Certain options have default values, as specified by each routine.  Some
##    are required, but are filled in from existing parameters if possible.
##    Each routine returns a reference to a hash, suitable for passing
##    as the options field to subsequent invokations of the routine.
##    Known options are listed under parameters.
##
##    The primary expected use of this interface is that to start the traffic,
##    the routine is called with an anonymous hash to the options field, and
##    the return value is stored.  Later on, starting or stopping the traffic
##    is simply another call to the subroutine using the stored value (and then
##    re-storing it).  This paradigm effectively allows for shared state between
##    the individual traffic start/stop calls, and hence the subroutine can make
##    sure to properly clean-up stale mounts and data that it creates (if any).
##
##    If an error occurs within one of the traffic routines, the key 'ERROR' in
##    the returned hash is hashed to a string describing the error.  (However,
##    the key 'ERROR' is deleted upon entry into any of the routines, so you can't
##    "force an error" intentionally or accidentally simply by passing in a hash
##    with an existing ERROR entry.)
##
## @param CLIENT Required: Client to run traffic from
## @param MOUNTPOINT Required: Mountpoint on client where target resides.
##        This option can be omitted and will be filled in automatically if (and
##        only if) DO_MOUNT is set and both FILER and FILER_VOL_MNT are provided.
## @param DO_MOUNT Optional: Whether to mount the target when starting traffic
##        and unmount when stopping traffic (default=0)
## @param FILER Optional: Filer to receive traffic
## @param FILER_VOL_MNT Optional: Path on filer to be mounted/unmounted on client
## @param SEED Optional: Randomization seed to use (defaults to
##        current time meshed with pid)
##
## @example
##    # Mount source qtree and start NFS traffic
##    my $NFSobj = NFStraffic(1, { 'CLIENT' => $CLIENT,
##                                 'DO_MOUNT' => '1',
##                                 'FILER' => $FILER,
##                                 'FILER_VOL_MNT' => $FILER_VOL_MNT
##                               } );
##    tstlog('-fatal', "Error: " . $NFSobj->{'ERROR'})
##      if (defined($NFSobj->{'ERROR'});
##    # do lots of interesting tests
##    my $results = do_work();
##    # Stop NFS traffic and unmount
##    $NFSobj = NFStraffic(0, $NFSobj);
##    tstlog('-fatal', "Error: " . $NFSobj->{'ERROR'})
##      if (defined($NFSobj->{'ERROR'});
##
## @dependencies
##
## @bugs
##    CIFStraffic is unimplemented.
##
##    Stopping hammer traffic stops all hammers on the target filer.
##
##    Right now, Traffic tests can pollute trees with new data files, can
##    overwrite existing files, do not clean up their files or their mountpoints,
##    and can fail if targetted at trees with files from previous runs.
##    You've been warned.
##
## @author remi@netapp.com
## @cf     false

package Traffic;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(NFStraffic CIFStraffic HammerStress);

use SayWrap;
use Tharn;
use strict;

sub escape ($) {
  my ($arg) = @_;
  $arg =~ s/([\\\|\(\)\[\{\^\$\*\+\?\.])/\\$1/g;
  return $arg;
}

## @name ::NFStraffic
## @usage NFStraffic($startORstop, $options)
## @synopsis
##    This routine is not implemented.  Don't call it.
sub NFStraffic($$) {
  my ($on, $options) = @_;
  delete $options->{'ERROR'}; # delete ERROR if it exists

  tstlog('-fatal', "Unimplemented.");
  return $options;
}

## @name ::CIFStraffic
## @usage CIFStraffic($startORstop, $options)
## @synopsis
##    This routine is not implemented.  Don't call it.
sub CIFStraffic($$) {
  my ($on, $options) = @_;
  delete $options->{'ERROR'}; # delete ERROR if it exists

  tstlog('-fatal', "Unimplemented.");
  return $options;
}

## @name ::HammerStress
## @usage HammerStress($startORstop, $options)
## @synopsis
##    Hammer differs from NFS and CIFS traffic in that it doesn't have a client.
##    The filer just pounds itself.  Hence, FILER and FILER_VOL_MOUNT are
##    required options and all others are ignored.
##    Hammer-specific options:
##    HAMMER_FILE - name of file in FILER_VOL_MOUNT to hammer at (default=hammerfile)
##    BLOCK_COUNT - number of blocks (4k each) to write into HAMMER_FILE (default=256)
##    This routine is really hacked up and does a slipshod job.  Hammer traffic can
##    abort randomly for a lot of different reasons, and treats them all as errors.
##    At some point Filer/Hammer.thpl will get re-written and this will subtest it
##    instead of rolling its own.
sub HammerStress($$) {
  my ($on, $options) = @_;
  delete $options->{'ERROR'}; # delete ERROR if it exists
  my $priv = "priv set advanced ;";

  if ($on) {
    # Start hammer traffic
    my $filer = $options->{'filer'};
    if ($filer) {
      $options->{'ERROR'} = "Hammer traffic already in progress";
      return $options;
    }

    # Required options
    tstlog('-fatal', "FILER not specified")
      unless ($options->{'FILER'});
    tstlog('-fatal', "FILER_VOL_MNT not specified")
      unless ($options->{'FILER_VOL_MNT'});
    $options->{'FILER_VOL_MNT'} =~ s/\/+$//; # chop trailing slashes if present

    # Default options
    $options->{'HAMMER_FILE'} = 'hammerfile'
      unless ($options->{'HAMMER_FILE'});
    $options->{'BLOCK_COUNT'} = '256'
      unless ($options->{'BLOCK_COUNT'});

    # Connect to filer
    host($options->{'FILER'});
    $filer = connect("#rsh;conntype=rsh-persay");
    unless ($filer) {
      $options->{'ERROR'} = "Couldn't establish rsh connection to " . $options->{'FILER'};
      return $options;
    }

    # Start hammer
    my $target = $options->{'FILER_VOL_MNT'} . "/" . $options->{'HAMMER_FILE'};
    my $blocks = $options->{'BLOCK_COUNT'};
    my $output = sayEOF($filer, "$priv hammer -f -1 $target $blocks", chop_wouldbe => 0);
    if ($output) {
      $output =~ s/[\r\t\n]+/ /g;
      $options->{'ERROR'} = "Error starting hammer: $output";
      return $options;
    }
    $options->{'filer'} = $filer;

    return $options;
  }
  else {
    # Stop hammer traffic
    my $filer = $options->{'filer'};
    unless ($filer) {
      $options->{'ERROR'} = "No hammer traffic in progress";
      return $options;
    }

    # Make sure hammer is still running
    my $target = $options->{'FILER_VOL_MNT'} . "/" . $options->{'HAMMER_FILE'};
    my $targetexp = escape( $target );
    my $output = sayEOF($filer, "$priv hammer status", chop_wouldbe => 0);
    unless ($output =~ /^(?:blacksmith \#\d+: file $targetexp,.*)$/m) {
      $options->{'ERROR'} = "Hammer '$target' died prematurely. output is '${output}'";
      delete $options->{'filer'};
      return $options;
    }

    $output = sayEOF($filer, "$priv hammer abort", chop_wouldbe => 0);
    if ($output) {
      $output =~ s/[\r\t\n]+/ /g;
      $options->{'ERROR'} = "Error stopping hammer: $output";
      return $options;
    }
    delete $options->{'filer'};

    return $options;
  }
}
