# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Traffic_local.pm#1 $

## @summary NFS/CIFS traffic to locally mounted trees
##
## @description
##    This package defines routines for starting/stopping various types of
##    traffic to a locally mounted tree.  The routines are usually used to
##    generate traffic locally on a remote client with Traffic.pm -- see
##    the documentation in Traffic.pm for more information and examples.
##
##    Each routine takes two arguments:  a boolean value of whether to start or
##    stop the traffic, and a reference to a hash of options.
##    Known options are listed under parameters.  The options which act as
##    toggles for which type of stress to run work as follows.  If set to zero,
##    that traffic type will not be run.  If set to a non-zero number or a
##    non-empty string, then the traffic type will be run, and if indicated
##    by a non-empty string (instead of a number), will treat the string as
##    arguments to the traffic.
##
##    This library was intentionally written so as not to depend upon NATE/Tharn.
##    Please don't modify it to require/use Tharn.
##
## @param MOUNTPOINT Optional: Local mountpoint to hit with traffic. Typically
##        this is the only necessary option, but it can be omitted and will be
##        filled in automatically if (and only if) DO_MOUNT is set and both
##        FILER and FILER_VOL_MNT are provided
##
## @param DO_MOUNT Optional: Whether to mount the target when starting traffic
##        and unmount when stopping traffic (default=0)
##
## @param FILER Optional: Filer to receive traffic
##
## @param FILER_VOL_MNT Optional: Path on filer to be mounted
##
## @param SEED Optional: Randomization seed to use (defaults to
##        current time meshed with pid)
##
## @param READ_TRAFFIC Optional: Apply continuous read traffic to the
##        target. (default=1)  Will spew errors if things fail, but should
##        not die prematurely.
##
## @param WRITE_TRAFFIC Optional: Apply continuous write traffic to the
##        target. (default=1)  Will spew errors if things fail, but should
##        not die prematurely.
##
## @param RANDOM Optional: Run "random" NFS stress on the target. (default=0)
##        May die prematurely if things fail or if targetted at a contents of
##        a previous run. See /usr/local/bin/random for arguments.
##
## @dependencies
##    NFS/Mount.pm
##    /usr/local/bin/random
##
## @bugs
##    CIFStraffic is unimplemented.
##
##    NFS write_traffic should take parameters similar to DataManip::populate().
##
##    Right now, Traffic tests can pollute trees with new data files, can
##    overwrite existing files, do not clean up their files or their mountpoints,
##    and can fail if targetted at trees with files from previous runs.
##    You've been warned.
##
## @author remi@netapp.com
## @cf     false
## @status public

package Traffic_local;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(local_NFStraffic local_CIFStraffic);

use NFS::Mount;
use Fcntl;
use strict;

# Return current time meshed with pid
sub getSeed () {
  my $seed = time() ^ ($$ + ($$ << 15));
  return $seed;
}

sub setError($$) {
  my ($ref, $err) = @_;
  if ($ref->{'ERROR'}) {
    $ref->{'ERROR'} .= "\n$err";
  } else {
    $ref->{'ERROR'} = $err;
  }
}

# Returns the contents of a directory, or undef if an
# error occurred (with error message in $!)
sub ls ($) {
  my ($dir) = @_;
  return () unless (opendir DIR, $dir);
  my @contents = readdir DIR;
  return () unless (closedir DIR);
  return @contents;
}

# Set up a static hash for re-using filehandles
# left open between accesses
{
  my $cnt = 0;
  my %filehandles;
  my @array; # for random data

# Reads the contents of a file (regular or otherwise)
# Returns whether an error occurred (with error message in $!)
sub readfile ($) {
  my ($targ) = @_;

  if (-f $targ) {
    # Regular file
    sysopen(FD, $targ, O_RDONLY) or return undef;
  } else {
    # Something else
    sysopen(FD, $targ, (O_RDONLY|O_NONBLOCK)) or return undef;
  }

  # Read everything in it
  my $num;
  my $buf;
  while ($num = sysread(FD, $buf, 4096)) { };
  my $rc = (defined $num);

  # Close it (if no failure yet)
  $rc = close(FD) if ($rc);

  # Return success or failure
  return $rc;
}

# Writes $size random bytes to the target file (regular or otherwise)
# Returns whether an error occurred (with error message in $!)
sub writefile ($$) {
  my ($targ, $size) = @_;
  my $blocks = int ($size / 4096);
  my $extra = int ($size % 4096);

  if (! -e $targ) {
    # Doesn't exist, create a regular file
    sysopen(FD, $targ, (O_WRONLY|O_CREAT)) or return undef;
  } elsif (-f $targ) {
    # Regular file
    sysopen(FD, $targ, O_WRONLY) or return undef;
  } else {
    # Something else
    sysopen(FD, $targ, (O_WRONLY|O_NONBLOCK)) or return undef;
  }

  # Make some random data
  if (scalar(@array)) {
    my $i;
    my $next;
    my $tmp = $array[1];
    for ($i = 1; $i < 3968; $i = $next) {
      $next = $i + 127;
      $array[$i] = $array[$next];
    }
    $array[($next - 127)] = $tmp;
    $tmp = shift @array;
    push @array, $tmp;
  } else {
    for (my $i = 0; $i < 4096; $i++) {
      $array[$i] = int rand(256);
    }
  }
  my $buf = pack("C4096", @array);

  # Write out whole blocks while we can
  if ($blocks > 0) {
    for (my $i = 0; $i < $blocks; $i++) {
      return undef unless (syswrite FD, $buf);
    }
  }

  # Write out the rest if necessary
  if ($extra) {
    return undef unless (syswrite FD, $buf, $extra);
  }

  # Close it (if no failure yet)
  return undef unless close(FD);

  # Return success
  return 1;
}

}

# Tries starting the test of the given sub block with the given name
# Returns success or failure
sub start_test(&$$) {
  my ($sub, $options, $name) = @_;

  my $pid = $options->{$name};
  if ($pid) {
    setError($options, "$name already in progress");
    return undef;
  }

  $pid = fork();
  if (defined($pid) && ($pid == 0)) {
    # Child to exec the command
    die $@ if eval { &$sub };
    exit 0;
  }

  if (defined $pid) {
    sleep 3; # To allow forked process to get started
    if (kill(0, $pid)) {
      $options->{$name} = $pid;
    } else {
      waitpid($pid, 0);
      setError($options, "Error in test \"$name\": exited with code $?");
    }
  } else {
    setError($options, "Error starting test \"$name\": $!");
  }

  return(!$options->{'ERROR'});
}

sub stop_test($$) {
  my ($options, $name) = @_;

  my $pid = $options->{$name};
  unless ($pid) {
    setError($options, "No \"$name\" in progress");
    return undef;
  }
  delete $options->{$name};

  # Make sure the test is still alive
  unless (kill(0, $pid)) {
    # Test is already dead
    if (waitpid($pid, 0) < 0) {
      setError($options, "Could not locate test \"$name\": $!");
    } else {
      setError($options, "Test \"$name\" died prematurely, exited with code $?");
    }
    return undef;
  }

  # Kill it
  setError($options, "Could not signal test \"$name\": $!")
    unless (kill(9, $pid));
  setError($options, "Could not locate test \"$name\": $!")
    if (waitpid($pid, 0) < 0);

  return(!$options->{'ERROR'});
}

# Used for moving a target from the reading list to the explore list
sub schedule_exploration($$) {
  my ($explist, $targ) = @_;

  # Do nothing if a parent of $targ is already scheduled for exploration
  foreach my $et (@$explist) {
    return $explist if ($targ =~ /^$et(\/.+)?$/);
  }

  # Strip children of $targ from the list
  my @expl = grep !/^$targ\/.+$/o, @$explist;

  push @expl, $targ;
  return \@expl;
}

# Hits the specified target with read traffic.
# Does not return
sub read_traffic($) {
  my ($root) = @_;
  my $read = [ ];
  my $readinglist = [ ];
  my $expl = [ ];
  my $explorelist = [ $root ];

  while (1) {

    # Explore the tree
    #print "Exploring... (" . scalar(@$explorelist) . " items)\n";
    while (scalar(@$explorelist)) {
      my $targ = shift @$explorelist;

      if (-d $targ) {
	# Directories -- listing is reading, and add things to explore
	my @dir = ls($targ);
	if (scalar(@dir)) {
	  # schedule to be read again
	  push @$readinglist, $targ;
	} else {
	  # will explore again later
	  print STDERR "(read_traffic) Error listing $targ while exploring: $!\n";
	  push @$expl, $targ;
	  next;
	}
	my @stripped = grep !/^\.+$/, @dir; # strip out "." and ".."
	@dir = map { "$targ/$_" } @stripped; # prepend the target
	push @$expl, @dir; # schedule it all to be explored

      } else {
	# Files -- if it can be read, keep reading it
	if (readfile($targ)) {
	  push @$readinglist, $targ;
	} else {
	  # will explore again later
	  print STDERR "(read_traffic) Error reading $targ while exploring: $!\n";
	  unless ($targ =~ m|/.snapshot/|) {
	    push @$expl, $targ;
	  }
	}
      }

    }
    $explorelist = $expl;
    $expl = [ ];


    # Read what we've found
    #print "Reading...   (" . scalar(@$readinglist) . " items)\n";
    while (scalar(@$readinglist)) {
      my $targ = shift @$readinglist;
      if (-d $targ) {
	if (scalar(ls($targ))) {
	  push @$read, $targ;
	} else {
	  print STDERR "(read_traffic) Error listing $targ: $!\n";
	  # move back to "exploration"
	  $explorelist = schedule_exploration($explorelist, $targ);
	}
      } else {
	if (readfile($targ)) {
	  push @$read, $targ;
	} else {
	  print STDERR "(read_traffic) Error reading $targ: $!\n";
	  # move back to "exploration"
	  unless ($targ =~ m|/.snapshot/|) {
	    $explorelist = schedule_exploration($explorelist, $targ);
	  }
	}
      }
    }
    $readinglist = $read;
    $read = [ ];

  }

}

sub fs_swap($\@$) {
  my ($dir, $names, $tmp) = @_;
  my $a = shift @$names;
  my $c = shift @$names;
  my $b = shift @$names;
  unshift @$names, $c;
  print STDERR "Error renaming $dir/$a -> $dir/$tmp: $!\n"
    unless (rename("$dir/$a", "$dir/$tmp"));
  print STDERR "Error renaming $dir/$b -> $dir/$a: $!\n"
    unless (rename("$dir/$b", "$dir/$a"));
  print STDERR "Error renaming $dir/$tmp -> $dir/$b: $!\n"
    unless (rename("$dir/$tmp", "$dir/$b"));
  push @$names, $a;
  push @$names, $b;
}

# Creates file if not already present
# Returns success or failure (with error message in $!)
sub touch($) {
  return undef
    unless open(FH, ">".$_[0]);

  # Close the file and make sure it exists
  # should definitely succeed if we got this far
  return undef
    unless close(FH);
  unless (-f $_[0]) {
    $! = "Can't find created file.";
    return undef;
  }
  return 1;
}

# Hits the specified target with write traffic.
# Does not return
sub write_traffic($) {
  my ($root) = @_;
  my ($file_avg, $dir_avg, $breadth_avg) = (5, 3, 32768); # make into parameters later
  my ($dir_prefix, $file_prefix) = ("dir", "data");
  my $overwrite = 0;

  my @writelist = ();
  my $expl = [ ];
  my $explorelist = [ $root ];
  my @expandlist = ();

  if ((-e $root) && (! -d $root)) {
    while (1) {
      my $rc = writefile($root, $breadth_avg);
      print STDERR "Error writing root $root: $!\n" unless $rc;
    }
  }

  while (1) {

    if (scalar(@$explorelist)) {
      # Explore the tree
      while (scalar(@$explorelist)) {
	my $targ = shift @$explorelist;

	if (-d $targ) {
	  # If it's readable, schedule to be written
	  my @dir = ls($targ);
	  if (scalar(@dir)) {
	    # schedule to be written
	    push @writelist, $targ;
	  } else {
	    # will explore again later
	    print STDERR "Error listing $targ while exploring: $!\n";
	    push @$expl, $targ;
	    next;
	  }

	  # Schedule contents for exploration
	  @dir = grep !/^\.+$/, @dir; # strip out "." and ".."
	  unless ($overwrite) {
	    my @dir = grep /^$dir_prefix/, @dir; # strip out dirs to explore
	  }
	  @dir = map { "$targ/$_" } @dir; # prepend the target
	  @dir = grep { -d $_ } @dir; # only explore directories
	  if (scalar(@dir)) {
	    push @$expl, @dir; # schedule it all to be explored
	  } else {
	    push @expandlist, $targ; # schedule leaf to be expanded
	  }

	} else {
	  # will try again later
	  print STDERR "Error reading $targ while exploring: $!\n";
	  unless ($targ =~ m|/.snapshot/|) {
	    push @$expl, $targ;
	  }
	}

      }
      $explorelist = $expl;
      $expl = [ ];

    } elsif (scalar(@expandlist)) {

      # Expand the tree
      my $targ = shift @expandlist;

      if (-d $targ) {
	my @subdirs;
	my $error = 0;

	for (my $i = 0; $i < $file_avg; $i++) {
	  unless (touch("$targ/$file_prefix$i")) {
	    $error = 1;
	    last;
	  }
	}

	unless ($error) {
	  for (my $i = 0; $i < $dir_avg; $i++) {
	    my $d = "$targ/$dir_prefix$i";
	    unless (mkdir($d, 0777)) {
	      $error = 1;
	      last;
	    }
	    push @subdirs, $d;
	  }
	}

	if ($error) {
	  print STDERR "Error expanding $targ: $!\n";
	  unless ($targ =~ m|/.snapshot/|) {
	    push @expandlist, $targ; # try it again later
	  }
	} else {
	  push @$explorelist, @subdirs; # schedule it all to be explored
	}

      } else {
	print STDERR "Error reading $targ while expanding: $!\n";
	unless ($targ =~ m|/.snapshot/|) {
	  push @expandlist, $targ; # try it again later
	}
      }

    } else {
      die "Out of branches to explore/expand on\n";
    }

    # Write to everything we know of
    my @written = ();
    my $frac = sprintf("%.0f", (scalar(@writelist) * 0.75));
    for (my $i = 0; $i < $frac; $i++) {
      my $dirname = shift @writelist;

      # Get a temporary name
      my $num = 99;
      my $tmp_name = "tmp_file_" . $num;
      while (-e "$dirname/$num") {
	$num++;
	$tmp_name = "tmp_file_" . $num;
      }

      # Split the directory contents into subdirs and files
      my @dir = ls($dirname);
      if (scalar(@dir) > 0) {
	my @subdirs;
	my @files;
	while (scalar(@dir)) {
	  my $t = pop @dir;
	  next if (($t eq ".") || ($t eq ".."));
	  if (-d $t) {
	    push @subdirs, $t
	      if (($t =~ /^$dir_prefix\d+$/) || $overwrite);
	  } else {
	    push @files, $t
	      if (($t =~ /^$file_prefix\d+$/) || $overwrite);
	  }
	}

	# Swap 1 directory pair
	if (scalar(@subdirs) > 2) {
	  fs_swap($dirname, @subdirs, $tmp_name);
	}

	# Swap 2 file pairs
	if (scalar(@files) > 4) {
	  fs_swap($dirname, @files, $tmp_name);
	  fs_swap($dirname, @files, $tmp_name);
	}

	# Write to all the files
	foreach my $f (@files) {
	  print STDERR "Error writing $f: $!\n"
	    unless writefile("$dirname/$f", $breadth_avg);
	}

	# Mark as written
	unshift @written, $dirname;

      } else {
	# Move the directory back to the explore list
	print STDERR "Error reading $dirname during write phase: $!\n";
	unless ($dirname =~ m|/.snapshot/|) {
	  $explorelist = schedule_exploration($explorelist, $dirname);
	}
      }
    }
    # Schedule to be written again
    push @writelist, @written;

  }

}

## @name ::NFStraffic
## @usage NFStraffic($startORstop, $options)
## @synopsis
sub local_NFStraffic($$) {
  my ($on, $options) = @_;
  delete $options->{'ERROR'}; # delete ERROR if it exists

  if ($on) {
    # Start NFS traffic

    # Default options
    $options->{'SEED'} = getSeed() unless ($options->{'SEED'});
    $options->{'READ_TRAFFIC'} = 1 unless (defined $options->{'READ_TRAFFIC'});
    $options->{'WRITE_TRAFFIC'} = 1 unless (defined $options->{'WRITE_TRAFFIC'});
    $options->{'RANDOM'} = 0 unless (defined $options->{'RANDOM'});

    # Mount the NFS server if necessary
    if ($options->{'DO_MOUNT'}) {
      my $mntpt = mount($options->{'FILER'}, $options->{'FILER_VOL_MNT'});
      unless ($mntpt) {
	$options->{'ERROR'} = "Error mounting: $@";
	return $options;
      }
      $options->{'MOUNTPOINT'} = $mntpt;
    } else {
      unless ($options->{'MOUNTPOINT'}) {
	$options->{'ERROR'} = "No MOUNTPOINT to target traffic at.";
	return $options;
      }
    }

    # Start read traffic
    if ($options->{'READ_TRAFFIC'}) {
      return $options
	unless start_test( sub { read_traffic($options->{'MOUNTPOINT'});
				 die "read_traffic error\n";
			       }, $options, 'read_traffic');
    }

    # Start write traffic
    if ($options->{'WRITE_TRAFFIC'}) {
      return $options
	unless start_test( sub { write_traffic($options->{'MOUNTPOINT'});
				 die "write_traffic error\n";
			       }, $options, 'write_traffic');
    }

    # Start "random" test
    if ($options->{'RANDOM'}) {
      return $options
	unless start_test( sub {
			     my $cmd = "/usr/local/bin/random";
			     $cmd .= " -s " . $options->{'SEED'};
			     $cmd .= " " . $options->{'RANDOM'}
			       unless ($options->{'RANDOM'} =~ /^\d+$/);
			     $cmd .= " " . $options->{'MOUNTPOINT'};
			     exec $cmd or die "Error executing random: $!\n";
			   }, $options, 'random');
    }

    return $options;

  }
  else {
    # Stop NFS traffic

    foreach my $test (qw(read_traffic write_traffic random)) {
      my $up = $test;
      $up =~ tr/a-z/A-Z/;
      stop_test($options, $test)
	if ($options->{$up});
    }

    if ($options->{'DO_MOUNT'}) {
      $options->{'WARNING'} = "Error unmounting: $@"
	unless (umount($options->{'MOUNTPOINT'}));
      delete $options->{'MOUNTPOINT'};
    }

    return $options;
  }
}

## @name ::CIFStraffic
## @usage CIFStraffic($startORstop, $options)
## @synopsis
##    This routine is not implemented.  Don't call it.
sub local_CIFStraffic($$) {
  my ($on, $options) = @_;
  delete $options->{'ERROR'}; # delete ERROR if it exists

  $options->{'ERROR'} = "CIFS traffic is not implemented";
  return $options;
}


1;
