#
# ULT::Container --
#  
#

=pod

=head1 NAME

ULT::Container - Hold a collection of log/trace messages, and verify them against filer output.

=cut

use strict;

package ULT::Container;

=head2 new()

Construct a new Container object.

    my $container = new ULT::Container();

=cut
sub new {
    my $class = shift;
    my $self = bless {}, $class;

    return $self;
}

=head2 addMsg()

Add a new ULT::Message object to the Container.

    $container->addMsg(new ULT::Message( ... ));

=cut
sub addMsg {
    my $self = shift;
    my $msg = shift;
    my $key = $msg->key();

    $self->{'msgs'}{$key} = $msg;

    return undef;
}

=head2 populate()

Loop through the messages added to the Container, and add each
of them to the filer.

    $container->populate();

=cut
sub populate {
    my $self = shift;

    foreach my $row (values %{$self->{'msgs'}} ) {
	$row->populate();
    }
}

=head2 validateMsgs()

Validate the set of stored messages against the output of the filer
(as from the `logs show` command).
Messages are compared in the order that we expect them to come
out, as determined by thier keys.

    my $cmd_output = UNIFIED_LOG_TRACE::show_utl_log();

    if (!$container->validateMsgs($cmd_output, $Log)) {
         # fail test
         ...
    }

=cut
sub validateMsgs {
    my $self = shift;
    my $cmd_output = shift;
    my $Log = shift;

    my (@sorted_keys) = sort keys %{$self->{'msgs'}};

    if (@sorted_keys != @$cmd_output) {
	# different number of rows, so the results are not identical
	if ($Log) {
	    $Log->error("Expected ".scalar(@sorted_keys)." rows,".
			" but found ".scalar(@$cmd_output));
	}
	return 0;
    }

    my $i=0;
    foreach my $key (@sorted_keys) {
	my $row = $self->{'msgs'}{$key};
	my $output = $cmd_output->[$i++];

	$Log->comment("Expecting message ".$row->format());
	if (!$row->validate($output)) {
	    if ($Log) {
	      $Log->error("Expected message ".$row->format().
			  " but found ".ULT::Msg::_format($output));
	    }
	    # the row did not match the output
	    return 0;
	}
    }

    # everything checks out OK
    return 1;
}

# return value so perl doesn't have a cow
1;
