=pod

=head1 NAME

ULT::Msg - Represent a single log/trace message

=cut
use strict;

package ULT::Msg;

use Time::Local;

=head2 new()

Construct a new Msg object

    my $msg =
    new ULT::Msg(
		'time' => $time_or_date,
		'seq' => $sequence_number,
		'source' => $source_name,
		'node' => $node_name,
		'message' => $message,
		'tags' => $tags,
		);

=cut
sub new {
    my $class = shift;
    my (%args) = @_;
    my $self = bless {}, $class;

    my $time = $args{'time'};
    if ($time =~ m/^\d+$/) {
	$time = _epoch2date($time);
    }
    if ($time !~ m/^"/) {
	$time = "\"$time\"";
    }
    $self->{'time'}    = $time;
    $self->{'seq'}     = $args{'seq'};
    $self->{'source'}  = $args{'source'};
    $self->{'node'}    = $args{'node'};
    $self->{'message'} = $args{'message'};
    $self->{'tags'}    = $args{'tags'};

    $self->{'Key'}     = _hash2key($self);

    return $self;
}

=head2 validate()

Validate the message against one line of output.

=cut
sub validate {
    my $self = shift;
    my $output = shift;

    my $tags = $self->{'tags'};
    if ($tags eq '""') {
	$tags = '-';
    }

    my $mismatch =
      $self->{'time'} cmp $output->{'time'} ||
	$self->{'seq'} <=> $output->{'seq'} ||
	  $self->{'source'} cmp $output->{'source'} ||
	    $self->{'node'} cmp $output->{'node'} ||
	      $self->{'message'} cmp $output->{'message'} ||
		$tags cmp $output->{'tags'};

    return !$mismatch;
}

=head2 key()

Return the key of the message.

    my $key = $msg->key();

=cut
sub key {
    my $self = shift;

    return $self->{'Key'};
}

=head2 format()

Format the msg object to be printed.

    print "Msg:", $msg->format();

=cut
sub format {
    my $self = shift;

    return _format($self);
}

=head2 populate()

Populate the message on the filer

    $msg->populate()

=cut
{

# populate a message against the "test1" data source
sub __pop_test1 {
    my $self = shift;

    UNIFIED_LOG_TRACE::pop_utl_debug_smdb_table
	(
	 table_name => "cluster_logs_test_1",
	 node => $self->{'node'},
	 time => $self->{'time'},
	 seq => $self->{'seq'},
	 message => $self->{'message'},
	 tags=> $self->{'tags'}
	);
}

# populate a message against the "test2" data source
sub __pop_test2 {
    my $self = shift;

    UNIFIED_LOG_TRACE::pop_utl_debug_smdb_table
	(
	 table_name => "cluster_logs_test_2",
	 node => $self->{'node'},
	 time => $self->{'time'},
	 seq => $self->{'seq'},
	 message => $self->{'message'},
	 tags=> $self->{'tags'}
	);
}

my (%pop_methods) = (
		     'test1' => \&__pop_test1,
		     'test2' => \&__pop_test2,
		    );

# The populate() method is merely a dispatcher to specific population
#  methods.  We lookup the specific population method based on the name
#  of the data source.
sub populate {
    my $self = shift;

    my $method = $pop_methods{$self->{'source'}};

    $self->$method();
}
}
#### Utility functions below this mark ####

# Format the values of a hash for printing.  This can be used
#  on a ULT::Msg object, or on a row returned as filer output.
sub _format {
    my $hash = shift;

    return
      "[time:$hash->{'time'},".
	"seq:$hash->{'seq'},".
	  "src:$hash->{'source'},".
	    "node:$hash->{'node'},".
	      "msg:$hash->{'message'},".
		"tags:$hash->{'tags'}]";
}

# Calculate the key (that is, the sort order) of a given hash.
# This can be used on a ULT::Msg object, or on a row returned
#  as filer output.
sub _hash2key {
    my $hash = shift;

    my $time = $hash->{'time'};
    $time =~ s/^"//;
    $time =~ s/"$//;
    $time = _date2epoch($time);

    my $epoch = sprintf("%010d", $time);
    my $seq   = sprintf("%016s", $hash->{'seq'});
    my $source= $hash->{'source'};
    my $node  = $hash->{'node'};

    return "$epoch,$seq,$source,$node";
}

{
my (@DOW) = qw(Sun Mon Tue Wed Thu Fri Sat);
my (@MONTH) = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
my (%MONTH_INDEX) = (Jan => 0,
		     Feb => 1,
		     Mar => 2,
		     Apr => 3,
		     May => 4,
		     Jun => 5,
		     Jul => 6,
		     Aug => 7,
		     Sep => 8,
		     Oct => 9,
		     Nov => 10,
		     Dec => 11);

# Given a seconds-since-epoch value, convert it to a date as expected
#  by the Date SMF type.
#
# Expected output format is: Fri May 06 15:52:42 2011
sub _epoch2date {
    my $epoch_time = shift;

    my ($sec, $min, $hour, $dom, $month, $year, $dow, $yday, $isdet) =
      localtime($epoch_time);

    $year += 1900; # No Y2K problems in Perl

    # zero-pad numeric values
    $dom = sprintf("%02d", $dom);
    $hour = sprintf("%02d", $hour);
    $min = sprintf("%02d", $min);
    $sec = sprintf("%02d", $sec);

    my $date = "$DOW[$dow] $MONTH[$month] $dom $hour:$min:$sec $year";
}

# Given a date value, convert it to an epoch value
sub _date2epoch {
    my $date = shift;

    my ($month, $day, $hour, $min, $sec, $year) = 
      $date =~ m/\w+ (\w+) (\d+) (\d+):(\d+):(\d+) (\d+)/;

    $year -= 1900;

    my $epoch =
      timelocal($sec, $min, $hour, $day, $MONTH_INDEX{$month}, $year);

    return sprintf("%010d", $epoch);
}

} 

# return value so perl doesn't have a cow
1;
