#***********************************************************************
#
# NAME
#   UNIFIED_LOG_TRACE.pm - 
#
# SYNOPSIS
#   UNIFIED_LOG_TRACE Library for the invocation and verfication of 7M and CM SN ASUP.
#
# DESCRIPTION
#   This library is a wrapper for the typical UNIFIED_LOG_TRACE commands that an end user
#   may desire for the configuration, invocation, and verification of
#   autosupports.  The current library was designed to use NACL CM SN filer.
#
# AUTHOR
#   dl-cluster-qa
#
# HISTORY
# 05/02/2011 Initial revision
#***********************************************************************

## @summary     UNIFIED_LOG_TRACE library    

package UNIFIED_LOG_TRACE;

#use statements
use strict;
use Data::Dumper;
use Server;
use Tharn qw($Log snooze host connect script_result fatal_result);
use Options;
use Params::Common;
use Params::Validate qw(:all);

use Carp;
use Sys::Hostname;

use File::Path;
use File::Basename;
use NACL::APISet;
use NACL::C::Node;
use Clone;


#Global variables
our $ontap_version;
my $Server = new Server();

my @Nodes=$Server->servers();

#maybe we can load the new the cdefs on declaration
our $Api_Set_Obj; 
my $opts = new Options;

my $opt_href = $opts->get_hash();

my $enable_regex = qr/1|on|ON|yes|YES|true|TRUE|0|off|OFF|no|NO|false|FALSE|enable|disable/;
my $id_regex = qr/hostname|systemid|true|false/i;


###############################################################################
#Sub: show_utl_log
#Input Params(optional): node  source, time, seq, message
#Return values: string
#Summary: Display the UTL log information
###############################################################################
sub show_utl_log{
    my (%params) = @_;
    use NATE::BaseException qw(:try);
    my $ret;
    my $host = host("$Nodes[-1]");
    if (!$Api_Set_Obj){
        $Api_Set_Obj=NACL::APISet->new(hostobj   => $host,
                                     category  => "Node",
                                     interface => "CLI",
                                     set => "CMode",
                                     timeout=> undef,
                                     
                                  );
        $Api_Set_Obj->set_privilege('privilege-level' => "test");
        #$Log->comment("Setting the APISet variable");
    }

    if ($params{timeout}){
        #my $def = $Api_Set_Obj->get_timeout();
        #$Log->comment("Stopping the command after $params{timeout} seconds instead of the default timeout of $def");
        $Log->comment("Stopping the command after $params{timeout} seconds");
        $Api_Set_Obj->set_timeout(timeout=>$params{timeout});
        delete $params{timeout};        
        try {
            $ret = $Api_Set_Obj->logs_show(%params)->get_parsed_output();
        } catch NATE::BaseException with {
            my $temp = shift;
            my @trunc = split(/\n/,$temp);
            $ret = \@trunc;
           #$Log->comment("Ignoring exception because we forced the command to exit...");

        };        
    }
    else{
        eval {$ret = $Api_Set_Obj->logs_show(%params)->get_parsed_output();};
        
    }
    $Api_Set_Obj->set_timeout(timeout=>60);
    if ($@) { return []; }
    #$Log->comment("Return log: @{$ret}");
    return $ret;
}

###############################################################################
#Sub: mod_utl_log
#Input Params: source, node<optional>
#Return values: string
#Summary: Enables and disables the log file sources
###############################################################################
sub mod_utl_log{
    my %params = (@_);

    my $ret = 0;
    my $host = host("$Nodes[-1]");
    if (!$Api_Set_Obj){
        $Api_Set_Obj=NACL::APISet->new(hostobj   => $host,
                                     category  => "Node",
                                     interface => "CLI",
                                     set => "CMode",                                     
                                     
                                  );
        $Api_Set_Obj->set_privilege('privilege-level' => "test");
        #$Log->comment("Setting the APISet variable");
    }     
    eval {$ret = $Api_Set_Obj->logs_config_modify(%params)->get_raw_output();};
    if ($@) { return 0; }
    my ($count) = $ret =~ m/(\d+) entries were modified./;
    return $count;
}

###############################################################################
#Sub: show_utl_config
#Input Params: node
#Return values: 0 - Fail, 1 - Pass
#Summary: Display the log configuration settings
############################################################## #################
sub show_utl_config{
    my (%params) = @_;

    my $ret;
    my $host = host("$Nodes[-1]");
    if (!$Api_Set_Obj){
        $Api_Set_Obj=NACL::APISet->new(hostobj   => $host,
                                     category  => "Node",
                                     interface => "CLI",
                                     set => "CMode",                                     
                                     
                                  );
        $Api_Set_Obj->set_privilege('privilege-level' => "test");
        #$Log->comment("Setting the APISet variable");
    }    
    eval {$ret = $Api_Set_Obj->logs_config_show(%params)->get_parsed_output();};
    if ($@) { return []; }
    return $ret;
}

###############################################################################
#Sub: show_utl_debug_smdb_table
#Input Params: table name, node<optional>
#Return values: string
#Summary: Display the UTL SMDB Table information
###############################################################################
sub show_utl_debug_smdb_table{
    my (%params) = @_;
    my $table_name = $params{'table_name'};
    undef $params{'table_name'};

    my $ret;
    my $host = host("$Nodes[-1]");
    if (!$Api_Set_Obj){
        $Api_Set_Obj=NACL::APISet->new(hostobj   => $host,
                                     category  => "Node",
                                     interface => "CLI",
                                     set => "CMode",                                     
                                     
                                  );
        $Api_Set_Obj->set_privilege('privilege-level' => "test");
        #$Log->comment("Setting the APISet variable");
    }    
    if ($table_name eq 'cluster_logs_test_1') {
	eval {$ret = $Api_Set_Obj->debug_smdb_table_cluster_logs_test_1_show(%params)->get_parsed_output();};
    } elsif ($table_name eq 'cluster_logs_test_2') {
	eval {$ret = $Api_Set_Obj->debug_smdb_table_cluster_logs_test_2_show(%params)->get_parsed_output();};
    } else {
	die "Unknown table name: $table_name";
    }

    if ($@) { return []; }
    return $ret;
}

###############################################################################
#Sub: del_utl_debug_smdb_table
#Input Params(optional): table name, node, time, message, seq, tag
#Return values: string
#Summary: Delete the UTL SMDB Table
###############################################################################
sub del_utl_debug_smdb_table{
    my %params = (@_);

    my $table = $params{'table_name'};
    undef $params{'table_name'};
    
    if ($params{type}) {
	%params = ( 'node' => '*',
		    'time' => '*',
		    'seq' => '*',
		  );
    }

    my $ret = 0;
    my $host = host("$Nodes[-1]");
    if (!$Api_Set_Obj){
        $Api_Set_Obj=NACL::APISet->new(hostobj   => $host,
                                     category  => "Node",
                                     interface => "CLI",
                                     set => "CMode",                                     
                                     
                                  );
        $Api_Set_Obj->set_privilege('privilege-level' => "test");
        #$Log->comment("Setting the APISet variable");
    }
    if ($table eq 'cluster_logs_test_1') {
	eval {$ret = $Api_Set_Obj->debug_smdb_table_cluster_logs_test_1_delete(%params)->get_raw_output();};
    } elsif ($table eq 'cluster_logs_test_2') {
	eval {$ret = $Api_Set_Obj->debug_smdb_table_cluster_logs_test_2_delete(%params)->get_raw_output();};
    } else {
	die "Unknown table name: $table";
    }

    return undef;
}


###############################################################################
#Sub: pop_utl_debug_smdb_table
#Input Params(optional): node, table name, tags, time, seq, message
#Return values: string
#Summary: Populate the UTL SMDB Table
###############################################################################
sub pop_utl_debug_smdb_table{
    my %params = (@_);

    my $table = $params{'table_name'};
    delete $params{'table_name'};
    
    if ($params{type}) {
	%params = ( 'node' => '*',
		    'time' => '*',
		    'seq' => '*',
		  );
    }

    my $ret = 0;
    my $host = host("$Nodes[-1]");
    if (!$Api_Set_Obj){
        $Api_Set_Obj=NACL::APISet->new(hostobj   => $host,
                                     category  => "Node",
                                     interface => "CLI",
                                     set => "CMode",                                     
                                     
                                  );
        $Api_Set_Obj->set_privilege('privilege-level' => "test");
        #$Log->comment("Setting the APISet variable");
    }
    if ($table eq 'cluster_logs_test_1') {
	eval {$ret = $Api_Set_Obj->debug_smdb_table_cluster_logs_test_1_create(%params)->get_raw_output();};
    } elsif ($table eq 'cluster_logs_test_2') {
	eval {$ret = $Api_Set_Obj->debug_smdb_table_cluster_logs_test_2_create(%params)->get_raw_output();};
    } else {
	die "Unknown table name: $table";
    }

    return undef;
}


###############################################################################
#Sub: get_node_names
#Input Params: None
#Return values: Array of node names
#Summary: Gets the nodes names in the cluster, make the first node in the list
#the execution node
###############################################################################
sub get_node_names{
    my %params = (
            @_,
    );
    my @node_names;
    my $host = host("$Nodes[-1]");
    if (!$Api_Set_Obj){
        $Api_Set_Obj=NACL::APISet->new(hostobj   => $host,
                                     category  => "Node",
                                     interface => "CLI",
                                     set => "CMode",                                     
                                     
                                  );
        $Api_Set_Obj->set_privilege('privilege-level' => "test");
        #$Log->comment("Setting the APISet variable");
    }    
    my $response1 = $Api_Set_Obj->system_node_show(node=>"local");
    my $response2 = $Api_Set_Obj->cluster_show();
    my @parsed_output1 = @{$response1->get_parsed_output()};
    foreach my $hash (@parsed_output1){
        foreach my $hvalue (keys %{$hash}){
            ##$Log->comment("Hash Key for Cluster Show: $hvalue");
            if ( $hvalue eq "node"){
                my $node = $hash->{$hvalue};
                push(@node_names,$hash->{$hvalue});
               ##$Log->comment("Adding Execution Node Name $node"); 
            }
        }
    }    
    my @parsed_output2 = @{$response2->get_parsed_output()};
    foreach my $hash (@parsed_output2){
        foreach my $hvalue (keys %{$hash}){
            ##$Log->comment("Hash Key for Cluster Show: $hvalue");
            if ( $hvalue eq "node"){
                next if (grep /$hash->{$hvalue}/, @node_names);
                my $node = $hash->{$hvalue};
                push(@node_names,$hash->{$hvalue});
               ##$Log->comment("Adding Node Name $node"); 
            }
        }
    }
    return @node_names; 
}

###############################################################################
#Sub: verify_log_value
#Input Params: API Set Repsonse Object
#Return values: 1 for sucess, 0 for failure
#Summary: Reads the UTL log output and looks for a matching string
###############################################################################
sub verify_log_value{
    my %params = (
            output => undef,
            value => undef,
            @_,
    );
    my @values = split(/\n/,$params{output});
    my $found=0;
    ##$Log->comment("Checking for value $params{value} in log");
    foreach my $val (@values){
        ##$Log->comment("Value: $val");
        if ( $val =~ /$params{value}/){
            ##$Log->comment("Found Value: $val");
            $found=1;
            last;
        }        
    }
    return $found;
}

    
###############################################################################
#Sub: _exec_raw_apiset_cmd
#Input Params: table name, cmd, node, file handle
#Return values: option hash
#Summary: Reads the UTL log from a file  handle or the console
###############################################################################
sub _exec_raw_apiset_cmd{
    my %params = (
            cmd   => undef,
            @_,
    );
    if ( !$params{cmd} ){
        $Log->error("No Command passed to the APISet Object, returning an empty string");
        return "";
    }
   
    return $Api_Set_Obj->execute_raw_command(command=>$params{cmd}); 
}
=pod

=head1 NAME

UNIFIED_LOG_TRACE.pm - UNIFIED_LOG_TRACE configuration, generation, and verification library for 7mode and Cmode SN.
=head1 SYNOPSIS

    use UNIFIED_LOG_TRACE;
    ASUP::show_utl_log(
            node       => undef,
            source => undef,
            time => undef,
            seq => undef,
            message => undef,
            );


=head1 DESCRIPTION

The methods included in this library were developed in order to assist
engineers in generation and verification of Unified Tracing and Logging commands.  The current library
supports 7M and Cmode filers and invokes communication to the filer via the
Server.pm library.  The UNIFIED_LOG_TRACE.pm library allows the user to configure
and verify UTL settings and log data.
  
    
=head1 METHODS
            
    show_utl_log();
        Description:
            This method will return the current UTL logs for all the nodes in a cluster
                on the filer.
        Inputs:
            node       => undef,
            source => undef,
            time => undef,
            seq => undef,
            message => undef,
        Outputs:
            The raw output of the command.

    mod_utl_log();
        Description:
            This method will enable and disable the UTL logs
          Inputs:
            node       => undef,
            source  => undef,
            enable =>  undef,
        Outputs:
            The raw output of the command.
          
            
    show_utl_config();
        Description:
            This method will display the configuration values for all the logs on all the nodes in the cluster
                
        Inputs:
            node       => undef,
            maxbufsize => 1000000,
            table_name => "cluster_logs_test_1",
            cmd = > "show",
        Outputs:
            The raw output of the comand.
           
            
           

=head1 EXAMPLES

Typical example:

    use UNIFIED_LOG_TRACE.pm;
    use vars qw($LOGDIR $LOGDIR_UNIX $CLIENT);
    my $Server  = Server->new();
    my @Nodes  = $Server->servers();
    $Server->set_execution_server(-Server => $Nodes[0]);
    
    my $log_output = UNIFIED_LOG_TRACE::show_utl_log();
    if ( $log_output =~ /This table is currently empty\./){
        $Log->error("No outuput returned for log show of the local node(s) in the cluster");
        return $TCD::FAIL;
    }
=cut
