# -*- perl -*-
# /* $Id: //depot/prod/test/Rsierranevada/lib/USBUtils.pm#1 $ */
#
# Copyright (c) 2009 NetApp, Inc.
# All rights reserved.
##
## @summary			
## USBUtils.pm is a common module for all the USB tests. 
##
## @description
## This module provides common functions for use in testing USB related   
## functionalities.
##
## Sub-routines:
## 
## 1. ONTAP_USB_Devices_Info 
##  - To get the informations of the USB devices connected to the system.
##
## 2. LOADER_USB_Devices_Info
##	- At Firmware prompt, get information of the USB devices connected to 
##	  the system.
##
## 3. Bootfs_Dir
##	- To check if a specified file or directory exists in the specified device 
##	  directory structure.
##
## 4. Command_Response
##	- To get the output of a given command.
##
## 5. Run_Stress
##	- To start NFS, CIFS and CPU Hammer stress.
##
## 6. USB_Support
##	- To check, whether Filer supports USB boot media or not.
##
## 7. Drive_10Mode
##  - To get the drive name in 10mode
##
## 8. Bootfs_Cmd
##  - To Execute the passed Boot_fs commands
##
## @status private
##
## @author sandeepk@netapp.com,dl-platform-qa@netapp.com
##
###############################################################################

package USBUtils;

use strict;
use warnings;
use Shell;
use Tharn;
use BTS_Config;
use SPARSE::Util::Logger qw(logStep logWarn);

require Exporter;
our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

our %EXPORT_TAGS = ( 'all' => [ qw(

) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our (%dvc_info,%ldr_dvc_info, %drv_id);
my $gLOADER = "LOADER.*>";

our @EXPORT = qw(
	%dvc_info
	%ldr_dvc_info
	%drv_id
	ONTAP_USB_Devices_Info
	LOADER_USB_Devices_Info
	Bootfs_Dir
	Command_Response
	Run_Stress
	USB_Support
	USB_Drive_Letter
	Bootfs_Cmd
);

############### Function Prototypes ##########
sub ONTAP_USB_Devices_Info($);
sub LOADER_USB_Devices_Info($);
sub Bootfs_Dir($$$$);
sub Command_Response($$);
sub Run_Stress($$$$);
sub USB_Support($);
sub USB_Drive_Letter ();
sub Bootfs_Cmd($$$);


############### Package Version ############
our $VERSION = '1.0';

#--------------------------------------------------------------------------
# FUNCTION: To get the informations of the USB devices connected to the system     
# INPUT: 	$_[0] = Filer Connection Handle.
# RETURN: 	1 - Failed to get USB info. Else, updates %dvc_info
#--------------------------------------------------------------------------
sub ONTAP_USB_Devices_Info($) {
    my ($response,@usb_sys_info,@usb_info,$dvc_cnt,$pre_sort,
    		$post_sort,$post_match, $temp);
     
 	eval {
         (undef, undef, $response, undef) = say($_[0], "sysconfig -a", -swallow, '-timeout', 300);
 	};
 	if ($@) {
		my $err = $@;
		exit;
		return 1;
	}	
	$response = $' if ($response =~ /USB.*Adapter/);
     
	if ($response =~ /memory-mapped/) {
		$pre_sort = $`;
		$post_sort = $';
	}
	
	logcomment "pre_sort :  $pre_sort";
	logcomment "post_sort :  $post_sort";

	@usb_sys_info = split(/\n/,$pre_sort);
	$dvc_cnt = $#usb_sys_info - 1;
	
	@usb_info = split(/\s{1,}/,$usb_sys_info[1]);
	$dvc_info{ID}[1] = $usb_info[1];	
							# $dvc_info{ID}[1] - Embedded USB device Name
	$dvc_info{VENDOR}[1] = $usb_info[2];
							# $dvc_info{VENDOR}[1] - Embedded USB device Vendor
	$dvc_info{MODEL}[1] = $usb_info[3]; 
							# $dvc_info{MODEL}[1] - Embedded USB device Model
	$dvc_info{SIZE}[1] = $usb_info[5];	
							# $dvc_info{SIZE}[1] - Embedded USB device Size

	if ($post_sort =~ /USB.*Adapter/){
		$post_match = $';
		if ($post_match =~ /memory-mapped/) {
			$post_match = $`;
			@usb_sys_info = split(/\n/,$post_match);
			$temp = $#usb_sys_info - 1;
			$dvc_cnt += $#usb_sys_info - 1;
			@usb_info = split(/\s{1,}/,$usb_sys_info[1]);
			$dvc_info{ID}[2] = $usb_info[1];	
								# $dvc_info{ID}[2] - External USB device Name	
			$dvc_info{VENDOR}[2] = $usb_info[2];
								# $dvc_info{VENDOR}[2] - External USB device Vendor
			$dvc_info{MODEL}[2] = $usb_info[3];	
								# $dvc_info{MODEL}[2] - External USB device Model
			$dvc_info{SIZE}[2] = $usb_info[4];	
							# $dvc_info{SIZE}[2] - External USB device Size
		}
	}
	
	$dvc_info{TOTALDVCS}= $dvc_cnt; # Total number of USB devices available
}

#-----------------------------------------------------------------------------
# FUNCTION: At Firmware prompt, get information of the USB devices connected to 
#			the system     
# INPUT: 	$_[0] = Filer Connection Handle.
# RETURN: 	NONE. However, updates %ldr_dvc_info
#-----------------------------------------------------------------------------
sub LOADER_USB_Devices_Info($) {
	my ($response,@show_devices,$count);
	(undef,undef,$response,undef) = say($_[0],"show devices",-swallow,
											-prompt,$gLOADER,'-timeout',300);
	@show_devices = split(/\n/,$response);
	$count = 1;	
	
	foreach (@show_devices){		
		if($_ =~ /(^u0\w\.\d)/) {
			$ldr_dvc_info{ID}[$count] =  $1;
			$count = $count + 1;			
		}
	}
	$ldr_dvc_info{TOTALDVCS} = $count-1; # Total number of USB devices available
}

#--------------------------------------------------------------------------
# FUNCTION: To check if a specified file or directory exists in the specified
#           device directory structure
# INPUT:    $_[0] = Filer Connection Handle, $_[1] = File Name,
#           $_[2] = Dirve / Directory, Eg: 7G - 0e.0:1:/BACKUP, 7M - /BACKUP 
#           $_[3] = Filer Name.
# RETURN:   0 = Exists, 1 = Doesn't exist.
#--------------------------------------------------------------------------
sub Bootfs_Dir($$$$) {
    my ($response,$buildinfo,$mode,$fhh);

# Set priviledge in script before calling this function 
#     # Set ONTAP in test privilege mode
#     if (!(Set_Priv($_[0],"advanced"))){
#         return 1;
#     }

    # Get the filer mode
    $fhh = host ($_[3]);
    $mode = Check_Filer_Mode($fhh, $_[0], $_[3]);

    if($mode eq "7mode") {
    	$buildinfo = Get_Version_Info($_[0]);
		if ($$buildinfo{versionindex} eq "7") {
			# List contents of the boot drive
			(undef,undef,$response,undef) = say($_[0], "bootfs dir -r $_[2]",  
			-swallow, -timeout, 300);
			if ($response =~ /$_[1]/){
				return 0;
			} else {
				return 1;
			}
		} elsif ($$buildinfo{versionindex} eq "8") {
			# List contents of the boot drive
			(undef,undef,$response,undef) = say($_[0], "bootfs dir $_[2]", 
			-swallow, -timeout, 300);
			if ($response =~ /$_[1]/){
				return 0;
			} else {
				return 1;
			}
		} else {
			logStep("Failed to get the filer version");
			return 1;
		}
	} elsif ($mode eq "10mode") {
		(undef,undef,$response,undef) = say($_[0], "bootfs dir $_[2]", -swallow, -timeout, 300);
		if ($response =~ /$_[1]/){
			return 0;
		} else {
			return 1;
		}
 	}
}

#----------------------------------------------------------------------------
# FUNCTION: To get the output of a given command.
# INPUT: 	$_[0] = Filer Connection Handle, $_[1] = Command.
# RETURN:   $response = Command Output.
#----------------------------------------------------------------------------
sub Command_Response($$) {	
	my (undef,undef,$response)=say($_[0],"$_[1]",
									-swallow,-prompt,$gLOADER,'-timeout',900);
	return $response;
}

#----------------------------------------------------------------------------
# FUNCTION: To start NFS, CIFS and CPU Hammer stress.
# INPUT: 	$_[0] = Filer Connection Handle, $_[1] = Run CIFS Stress,
#			$_[2] = Run NFS stress, $_[3] = Run CPU hammer stress.
# RETURN:   $response = Command Output.
#----------------------------------------------------------------------------
sub Run_Stress($$$$) {
    # Validate hammer run parameters    
 	my ($fc, $run_cifs_hammer, $run_nfs_hammer, $run_ontap_hammer) = @_;
 	
    die ("RUN_CIFS_HAMMER value must either be 'yes' or 'no'. ".
         "Found: $run_cifs_hammer")
      if ($run_cifs_hammer !~ /^(yes|no)$/);

    die ("RUN_NFS_HAMMER value must either be 'yes' or 'no'. ".
         "Found: $run_nfs_hammer")
      if ($run_nfs_hammer !~ /^(yes|no)$/);

    die ("RUN_ONTAP_HAMMER value must either be 'yes' or 'no'. ".
         "Found: $run_ontap_hammer")
      if ($run_ontap_hammer !~ /^(yes|no)$/);

	if (($run_cifs_hammer eq 'yes') || ($run_nfs_hammer eq 'yes') || 
			($run_ontap_hammer eq 'yes')) {
		logWarn("Starting of Stress tests not implemented.");
    }
}


#----------------------------------------------------------------------------
# FUNCTION: To check, whether Filer supports USB boot media or not
# INPUT: 	$_[0] = Filer connection handle
# RETURN:   0 = Filer supports USB boot media, 
#           1 = Filer does NOT supports USB boot media
#----------------------------------------------------------------------------
sub USB_Support($) {
    my $prompt;    
    $prompt = Filer_Prompt($_[0]);
    
    if ($prompt !~/LOADER.*/) {
        if (ONTAP_USB_Devices_Info($_[0])) {
	        return 1;	# USB device not found
        } else {            
            return 0;	# USB device present            
        }        
    } else {
        LOADER_USB_Devices_Info($_[0]);
        if (!($ldr_dvc_info{TOTALDVCS})) {            
            return 1;	# USB device not found
        } else {            
            return 0;	# USB device present            
        }
    }
}

#----------------------------------------------------------------------------
# FUNCTION:  To get the drive name in 10mode
# INPUT:     NONE
# RETURN:    NONE, Update the hash %drv_id (hash is exported to scripts)
#----------------------------------------------------------------------------
sub USB_Drive_Letter () {
# vivekk - in 10mode no drive letter is used in bootfs dir command
#     $drv_id{EMB} = "boot_i:";
#     $drv_id{EXT} = "boot_e:";

	$drv_id{EMB} = "";
}
#----------------------------------------------------------------------------
# FUNCTION:  Execute the passed Boot_fs commands
# INPUT:     $_[0] = Filer connection handle,$_[1]=Bootfs command,$_[2]=Message
# RETURN:    NONE
#----------------------------------------------------------------------------
sub Bootfs_Cmd($$$) {
    my $result;
	$result = Command_Response($_[0],$_[1]);
	if ($result =~ $_[2]) {
		return 0;
	} else {
		return 1;
	}
}
1;


