# $Id: //depot/prod/test/Rsierranevada/lib/Users.pm#1 $

## @summary misc routines for nate script
##
## @description
##      Users.pm functions
##
## @status in-progress
## 
## @author  dl-maryv (FSS Protocols QA)
##
## @keywords User Group Account

package Users;
use strict;
use Carp qw(croak);
BEGIN
{
  use base qw(Exporter);
  our @EXPORT = qw(
	addRoles
	delRole
	createUsers
	createGroups
	setPrimaryGroup
	addPasswdEntry
	delPasswdEntry
	delPasswdList
	addGroupEntry
	delGroupEntry
	delGroupList
	addPasswdList
	addGroupEntry
	addGroupList
	addCIFSUser
	removeCIFSUser
	addCIFSGroup
	removeCIFSGroup
	listCIFSGroups
	listCIFSUsers
	findCIFSUser
	findCIFSGroup
	findCIFSUserOnClient
	removeCIFSUsersExcept
	getNextUID
	getNextGID
    	readGroupFile
    	readPasswdFile
	findUser
	findGroup
	writeDefaultPasswdFile
	writePasswdFile
	writeGroupFile
	appendPasswdFile
	appendGroupFile
  );
}
use Params::Validate qw(validate validate_pos 
			SCALAR OBJECT BOOLEAN ARRAYREF HASHREF);
use Tharn;
use SPARSE::Util::Logger;
use CLI::FileIO qw(rdfile wrfile);
use Utils qw(make_conn);
use Filer::Version qw(get_wafl_version WV_SCRIMSHAW_MAX);
use SayWrap qw(sayEOF);
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF UNDEF);




##
## @notes
##
##	Convenional hasth table representation of a 'user' (/etc/passwd entry) is
##	shown by the following snippet from createUsers. This convention is used
##	by all functions in this module
##
##	my $user_ref = {
##				login => $login,
##				passwd => "x",
##				uid => $uid,
##				gid => $gid,
##				name => "",
##				home => "",
##				shell => "",
##				group => $group,
##			};
##
##	Conventional representation of a 'group' (/etc/goup entry) is:
##
##	my $grp_hash = {
##				group => $group,
##				passwd => "x",
##				gid => $gid,
##				members => "",
##				members_list => [],
##			};
##


## @name ::addRoles()
## @synopsis addRoles($filerconn, $rolename, $rolestring)
## @arg $filerconn Required Filer Connection Object
## @arg $rolename: Name of the role to create.
## @arg $rolestring: role(s) to create.
## @description Add a role to the filer.
## @returns true if success
## @status public

sub addRoles ($$$)
{
	 my ($spec, $rolename, $rolestring) =
                validate_pos(@_,
                          {type => SCALAR|HASHREF },
                          {type => SCALAR },
                          {type => SCALAR } 
			  );

        my $added = "The role '$rolename' has been added|Role <$rolename> added";

        return eval {
                my $conn = make_conn($spec);
	        say($conn, "priv set -q test; useradmin role add ".
                         "$rolename -a $rolestring",'-wait', $added) or die $@;
		return 1;
        };
}


## @name ::delRole()
## @synopsis delRole($filerconn, $rolename)
## @arg $filerconn Required Filer Connection Object
## @arg $rolename: Name of the role to delete.
## @description Delete a role on the filer.
## @returns true if success
## @status public

sub delRole ($$)
{
	 my ($spec, $rolename) =
                validate_pos(@_,
                          {type => SCALAR|HASHREF },
                          {type => SCALAR }
			  );

        my $deleted = "The role '$rolename' has been deleted|Role <$rolename> deleted";

        return eval {
                my $conn = make_conn($spec);
	        say($conn, "useradmin role delete $rolename",'-wait', $deleted) or die $@;
		return 1;
        };
}


## @name ::createUsers()
## @synopsis @users = createUsers ($filerconn, $nusers)
## @arg $filerconn Required Filer Connection Object
## @arg $nusers: Number of unique users to create. One entry for each
##            user is created in the /etc/passwd file. Also, a unique
##            group is created for each of the user in the /etc/group
##            and that user is placed in this unique group in
##            the /etc/passwd file.
## @description creates specified number of user entries in
##            the /etc/passwd.
## @returns array to users created
## @status public
#
#
#   Creates $nusers new users, each in it's own new group
#   If 'gid' optional arg is specified, then no new groups are
#   created, and the specified gid is used for all new users
#
#  Args:
#
#	filerspec:		Required
#
#	nusers:			Required. Number of users to create.
#
#   Optional args:
#
#	login_prefix		Default: wafl_user
#
#	group_prefix		Default: wafl_group
#
#	group			Group name. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user.
#
#	gid			Group gid. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user. Ignored if 'group'
#				is specified.
#
#	samegroup		If 'group' and 'gid' are not specified, and 'samegroup' is 1,
#				then a single new group is created for all the new users.
#				Default is '0'.
#
#	current_groups		HASHREF which represents current groups. If defined,
#			    and the HASHREF is empty, then the HASHREF is updated with
#			    current groups from /etc/groups. Otherwise the  HASHREF is
#			    used to determine current groups, and /etc/groups is not read.
#			    The HASHREF is also updated with any new groups created
#
#	current_groups_list	ARRAYREF which represents current groups. If defined,
#			    and the ARRAYREF is empty, then the ARRAYREF is updated with
#			    current groups from /etc/groups. Otherwise the  ARRAYREF is
#			    used to determine current groups, and /etc/groups is not read.
#			    The ARRAYREF is also updated with any new groups created
#			    -- Only one of 'current_groups' or 'current_groups_list'
#			       should be defined.
#
#	current_users		HASHREF which represents current users. If defined,
#			    and the HASHREF is empty, then the HASHREF is updated with
#			    current users from /etc/passwd. Otherwise the  HASHREF is
#			    used to determine current users, and /etc/passwd is not read.
#			    The HASHREF is also updated with any new users created
#
#	current_users_list	HASHREF which represents current users. If defined,
#			    and the HASHREF is empty, then the HASHREF is updated with
#			    current users from /etc/passwd. Otherwise the  HASHREF is
#			    used to determine current users, and /etc/passwd is not read.
#			    The HASHREF is also updated with any new users created
#			    -- Only one of 'current_users' or 'current_users_list'
#			       should be defined.
#
#	created_users		HASHREF. If defined, then the created users are added
#			    to this hash.
#
#	created_users_list	ARRAYHREF. If defined, then the created users are added
#			    to this list.
#
#	created_groups		HASHREF. If defined, then the created groups are added
#			    to this hash.
#
#	created_groups_list	ARRAYREF. If defined, then the created groups are added
#			    to this list.
#
#	min_uid			Default value is set in _getNextUID
#
#	max_uid			Default value is set in _getNextUID
#
#	min_gid			Default value is set in _getNextGID
#
#	min_gid			Default value is set in _getNextGID
#
#	The following options apply only if 'nusers' is 1.
#	There are no defaults.
#
#	login			Login to use for new user
#
#	uid			Uid to use for new user
#
#	new_users		ARRAYREF, where the elements of the array are each a HASHREF
#				If supplied, these hashesare used as 'templates' to create the
#				new users.Defined fields are used, and undefined fields are
#				filled-in. This option is ignored unless 'nusers' is 0.
#
sub createUsers ($$%args)
{
  my ($filerspec, $nusers, %opts) = validate_mixed(@_,
			[
			        { type => SCALAR|HASHREF },
				{ type => SCALAR|UNDEF },
			],
			{ 
				login_prefix 		=> { type => SCALAR, optional => 1,
								default => "wafl_user" },
				group_prefix 		=> { type => SCALAR, optional => 1 },
				group 			=> { type => SCALAR|HASHREF, optional => 1 },
				gid 			=> { type => SCALAR, optional => 1 },
				samegroup 		=> { type => BOOLEAN,, optional => 1,
								default => 0 },
				current_groups		=> { type => HASHREF, optional => 1 },
				current_groups_list	=> { type => ARRAYREF, optional => 1 },
				current_users		=> { type => HASHREF, optional => 1 },
				current_users_list 	=> { type => ARRAYREF, optional => 1 },
				created_groups		=> { type => HASHREF, optional => 1 },
				created_groups_list	=> { type => ARRAYREF, optional => 1 },
				created_users		=> { type => HASHREF, optional => 1 },
				created_users_list	=> { type => ARRAYREF, optional => 1 },
				min_uid 		=> { type => SCALAR, optional => 1 },
				max_uid 		=> { type => SCALAR, optional => 1 },
				min_gid 		=> { type => SCALAR, optional => 1 },
				max_gid 		=> { type => SCALAR, optional => 1 },
				login 			=> { type => SCALAR, optional => 1 },
				uid 			=> { type => SCALAR, optional => 1 },
				new_users 		=> { type => ARRAYREF|UNDEF, optional => 1 },
			}
		);
		$opts{nusers} = $nusers;
    logDbg("Enter   sub createUsers  Line: ".__LINE__."\n\n".Dumper(\%opts));

  return eval {
    my $filerconn = make_conn($filerspec);
    my $spec_group_ref = undef;
    my $spec_gid;
    my $group_specified;
    my $next_login;
    my @new_users;
    my $login_prefix = $opts{login_prefix};
    #  'min_gid', 'max_gid', 'group_prefix' are defaulted by createGroup,
    #  which is passed these values

    my $created_users_hash = $opts{created_users} || {};
    my $created_users_list = $opts{created_users_list} || undef;
    my $created_groups_hash = $opts{created_groups} || {};
    my $created_groups_list = $opts{created_groups_list} || undef;

    my ($current_groups_hash, $current_groups_list, $current_users_hash, $current_users_list);
    if (defined($opts{current_groups})) {
    	$current_groups_hash = $opts{current_groups} || {};
        if (defined($opts{current_groups_list})) {
            $current_groups_list = [ values(%$current_groups_hash) ];
        } else {
            $current_groups_list = undef;
        }
    } elsif (defined($opts{current_groups_list})) {
    	$current_groups_list = $opts{current_groups_list} || {};
        $current_groups_hash = {};
        foreach ( @$current_groups_list ) {
            my $key = $_->{group};
            $current_groups_hash->{$key} = $_;
        }
    } else {
        $current_groups_hash = {};
        $current_groups_list = undef;
    }
    if (!keys(%$current_groups_hash)) {
         $current_groups_hash = readGroupFile($filerconn)
		or die($@);
         $current_groups_list = [ values(%$current_groups_hash) ];
    }
    if (defined($opts{current_users})) {
    	$current_users_hash = $opts{current_users} || {};
        if (defined($opts{current_users_list})) {
            $current_users_list = [ values(%$current_users_hash) ];
        } else {
            $current_users_list = undef;
        }
    } elsif (defined($opts{current_users_list})) {
      	    $current_users_list = $opts{current_users_list} || {};
            $current_users_hash = {};
            foreach ( @$current_users_list ) {
                my $key = $_->{login};
                $current_users_hash->{$key} = $_;
            }
        } else {
             $current_users_hash = {};
            $current_users_list = undef;
        }
        if (!keys(%$current_users_hash)) {
              $current_users_hash = readPasswdFile($filerconn)
  		    or die($@);
              $current_users_list = [ values(%$current_users_hash) ];
        }

        my $next_uid = undef;
        $nusers ||= 0; 

        #  4 cases:
        #	   1. $nusers is 0, and no 'new_users' option given
        #		==> return empty hashtable
        #	   2. $nusers is 0, and 'new_users' option given
        #		==> @new_users array is set from 'new_users option
        #		    Also, $nusers is set to size of @new_users list
        #	   3. $nusers is 1
        #		==> a singleton @new_users list is constructed using
        #	    the 'uid' and/or 'login' options, if they are given.
        #		    This is the only case where these options are used.
        #	   4. $nusers is > 0
        #		==> @new_users list is empty. If 'samegroup' option is 1,
        #		    then a single new group will be created for all new users.
        if ( (defined($opts{gid}) || defined($opts{group})) && $opts{samegroup} ) {
        }
        if ($nusers == 0) {
            if ($opts{new_users}) {
    	        @new_users = @{$opts{new_users}};
    	        $nusers = scalar(@new_users);
    	    }
        } elsif ($nusers == 1) {
    	     @new_users = ( { login=>$opts{login}, uid=>$opts{uid} } );
        } else {
    	    @new_users = ();
    	    foreach ( 1 .. $nusers ) {
    		    push(@new_users, {});
    	    }
        }
        return( () )
           if (!$nusers);

        my $num_uids_needed = 0;

        #  Calculate number of new UIDs needed for the new users.
        #  Also, for each user in @new_users, create a group in @new_groups,
        #  using the 'group' and 'gid' fields of the user.
        #
        #  @new_users is non-empty
        #	==> check each user in @new_users. Need a new UID
        #		    for each user whose 'uid' field is undefined.
         my @new_groups = ();
         foreach my $u (@new_users) {
            if (!defined($u->{uid})) {
       	        $num_uids_needed++;
            }
         }
         if ($num_uids_needed > 0) {
        	$next_uid = _getNextUID($num_uids_needed,
    				         $current_users_hash,
    				         min_uid   => $opts{min_uid},
    				         max_uid   => $opts{max_uid});
         }
         foreach my $u (@new_users) {
            if (!defined($u->{uid})) {
       	        $u->{uid} = $next_uid++;
            }
         }

#        if (!keys(%$current_groups_hash)) {
#    	    $current_groups_hash = readGroupFile($filerconn)
#	        or die($@);
#        }

        if ($opts{gid} || $opts{group} || $opts{samegroup}) {
	    $spec_group_ref =
	        _findOrCreateGroup($filerconn,
				     group_prefix           => $opts{group_prefix},
			             group                  => $opts{group},
				     gid                    => $opts{gid},
			             min_gid                => $opts{min_gid},
                                     max_gid                => $opts{max_gid},
			             current_groups         => $current_groups_hash,
			             current_groups_list    => $current_groups_list,
			             created_groups         => $created_groups_hash,
			             created_groups_list    => $created_groups_list);
   } else {
	my $ngroups = $nusers;
	foreach my $uref (@new_users) {
	    my $new_grp =
		_findOrCreateGroup($filerconn,
				     group_prefix         => $opts{group_prefix},
				     group                => $uref->{group},
				     gid                  => $uref->{gid},
			             min_gid              => $opts{min_gid},
				     max_gid              => $opts{max_gid},
			             current_groups       => $current_groups_hash,
			             current_groups_list  => $current_groups_list,
			             created_groups       => $created_groups_hash,
			             created_groups_list  => $created_groups_list);
	    push(@new_groups, $new_grp);
	}
    }

    for (my $i=0; $i < $nusers; $i++) {
	my $uid = undef;
	my $login = undef;
	my $group = undef;
	my $gid = undef;
	my $grp_ref = undef;

	$uid = $new_users[$i]->{uid};
	$login = $new_users[$i]->{login};
	if (!defined($uid)) {
		$uid = $next_uid + $i;
	}
	if (!defined($login)) {
		$login = sprintf "${login_prefix}%04d", $uid;
	}

	if (@new_groups) {
		$group = $new_groups[$i]->{group};
		$gid = $new_groups[$i]->{gid};
	} else {
		$group = $spec_group_ref->{group};
		$gid = $spec_group_ref->{gid};
		$grp_ref = $spec_group_ref;
        }
	if (!defined($group)) {
		$group = $grp_ref->{group};
	}
	if (!defined($gid)) {
		$gid = $grp_ref->{gid};
	}
	if (!defined($grp_ref->{members_list})) {
		$grp_ref->{members_list} = [];
	}
	my $user_ref = {
				login => $login,
				passwd => "x",
				uid => $uid,
				gid => $gid,
				name => "",
				home => "",
				shell => "",
				group => $group,
			};
	$$current_users_hash{$login} = $user_ref;
	$$created_users_hash{$login} = $user_ref;
        if (defined($current_users_list)) {
                push(@$current_users_list, $user_ref);
        }
        if (defined($created_users_list)) {
                push(@$created_users_list, $user_ref);
        }
	push(@{$grp_ref->{members_list}}, $login);
    }

    addPasswdList($filerconn, $created_users_hash)
	or die($@);

    # workaround for strange '$@' behavior ....
    logDbg("Users  Line: ".__LINE__."\ncreateUsers returning:\n\n".Dumper($created_users_hash));
    return %$created_users_hash;
  };
}



#  _findOrCreateGroup($filerspec, %opts)
#
#  Args:
#
#	filerspec:		Required
#
#   Optional args:
#
#	group			Group name. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user.
#
#	gid			Group gid. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user. Ignored if 'group'
#				is specified.
#
#	group_prefix		Default: wafl_group
#
#	min_gid			passed to _getNextGID
#
#	max_gid			passed to _getNextGID
#
#	current_groups		HASHREF which represents current groups. If defined,
#			    and the HASHREF is empty, then the HASHREF is updated with
#			    current users from /etc/passwd. Otherwise the  HASHREF is
#			    used to determine current users, and /etc/passwd is not read.
#
#	current_groups_list	ARRAYREF which represents current groups. If defined,
#			    and the ARRAYREF is empty, then the ARRAYREF is updated with
#			    current users from /etc/passwd. Otherwise the  ARRARYREF is
#			    used to determine current users, and /etc/passwd is not read.
#			    -- Only one of 'current_groups' or 'current_groups_list'
#			       should be defined.
#
#	created_groups		HASHREF. If defined, then the created groups are added
#			    to this hash.
#
#	created_groups_list	ARRAYREF. If defined, then the created groups are added
#			    to this list.
#
#	new_groups		ARRAYREF, where the elements of the array are each a HASHREF
#				If supplied, these hashesare used as templates to create the
#				new groups.Defined fields are used, and undefined fields are
#				filled-in. This option is ignored unless 'ngroups' is 0.
#
sub _findOrCreateGroup
{
  my ($filerspec, %opts) = validate_mixed(@_,
			[
			        { type => SCALAR|HASHREF },
			],
			{ 
				group_prefix 		=> { type => SCALAR|UNDEF, optional => 1 },
				group 			=> { type => SCALAR|UNDEF, optional => 1 },
				gid 			=> { type => SCALAR|UNDEF, optional => 1 },
				min_gid 		=> { type => SCALAR|UNDEF, optional => 1 },
				max_gid 		=> { type => SCALAR|UNDEF, optional => 1 },
				current_groups 		=> { type => HASHREF|UNDEF, optional => 1 },
				current_groups_list 	=> { type => ARRAYREF|UNDEF, optional => 1 },
				created_groups 		=> { type => HASHREF|UNDEF, optional => 1 },
				created_groups_list	=> { type => ARRAYREF|UNDEF, optional => 1 },
				new_groups 		=> { type => ARRAYREF|UNDEF, optional => 1 },
			}
		);
     logDbg("Enter   sub findOrCreateGroup  Line: ".__LINE__."\n\n".Dumper(\%opts));

  return eval {
    my $filerconn = make_conn($filerspec);
    my ($group_prefix) = $opts{group_prefix} || "wafl_group";
    my $created_groups_hash = $opts{created_groups} || {};
    my $created_groups_list = $opts{created_groups_list} || undef;

    my ($current_groups_hash, $current_groups_list);
    if (defined($opts{current_groups})) {
    	$current_groups_hash = $opts{current_groups} || {};
        if (defined($opts{current_groups_list})) {
            $current_groups_list = [ values(%$current_groups_hash) ];
        } else {
            $current_groups_list = undef;
        }
    } elsif (defined($opts{current_groups_list})) {
    	 $current_groups_list = $opts{current_groups_list} || {};
         $current_groups_hash = {};
         foreach ( @$current_groups_list ) {
             my $key = $_->{group};
             $current_groups_hash->{$key} = $_;
        }
    } else {
        $current_groups_hash = {};
        $current_groups_list = undef;
    }
    if (!keys(%$current_groups_hash)) {
         $current_groups_hash = readGroupFile($filerconn)
		or die($@);
         $current_groups_list = [ values(%$current_groups_hash) ];
    }

    my $grp_ref;
    if ( defined($opts{gid}) ) {
	my $gid = $opts{gid};
	($grp_ref) = _findGroup($current_groups_hash, gid => $gid);
    } elsif ( defined($opts{group}) ) {
	my $group = $opts{group};
	($grp_ref) = _findGroup($current_groups_hash, group => $group);
    }
    if ($grp_ref) {
        if (defined($current_groups_hash)) {
		$current_groups_hash->{$grp_ref->{group}} = $grp_ref;
	}
        if (defined($created_groups_hash)) {
		$created_groups_hash->{$grp_ref->{group}} = $grp_ref;
	}
        if (defined($current_groups_list)) {
                push(@$current_groups_list, $grp_ref);
        }
        if (defined($created_groups_list)) {
                push(@$created_groups_list, $grp_ref);
        }

	return ( $grp_ref );
    }
    my %new_group_hash = 
        createGroups($filerconn, 1, group_prefix=>$opts{group_prefix},
			group => $opts{group}, gid => $opts{gid},
			min_gid=>$opts{min_gid}, max_gid=>$opts{max_gid},
			current_groups=>$current_groups_hash,
			current_groups_list=>$current_groups_list,
			created_groups=>$created_groups_hash,
			created_groups_list=>$created_groups_list,
			new_groups=>$opts{new_groups});
    my @new_group_list = values(%new_group_hash);
    my $new_grp = $new_group_list[0];
    logDbg("Users  Line: ".__LINE__."\n\nfindOrCreateGroup  Returning:\n\n".Dumper($created_groups_hash));
    return ( $new_grp );
  }
}



## @name ::createGroups()
## @synopsis @groups = createGroups ($filerconn, $ngroups)
## @arg $filerconn Required Filer Connection Object
## @arg $ngroups: Number of unique groups to create. One entry for each
##            group is created in the /etc/group file
## @description creates specified number of groups entries in
##            /etc/group.
## @returns array to group created
## @status public
#
#  Args:
#
#	filerspec:		Required
#
#	ngroups:		Required. Number of groups to create.
#
#   Optional args:
#
#	group			Group name. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user.
#
#	gid			Group gid. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user. Ignored if 'group'
#				is specified.
#
#	group_prefix		Default: wafl_group
#
#	min_gid			passed to _getNextGID
#
#	max_gid			passed to _getNextGID
#
#	current_groups		HASHREF which represents current groups. If defined,
#			    and the HASHREF is empty, then the HASHREF is updated with
#			    current groups from /etc/group. Otherwise the  HASHREF is
#			    used to determine current groups, and /etc/group is not read.
#			    The HASHREF is also updated with any new groupss created
#
#	current_groups_list	ARRAYREF which represents current groups. If defined,
#			    and the ARRAYREF is empty, then the ARRAYREF is updated with
#			    current groups from /etc/group. Otherwise the  ARRARYREF is
#			    used to determine current groups, and /etc/group is not read.
#			    -- Only one of 'current_groups' or 'current_groups_list'
#			       should be defined.
#
#	created_groups		HASHREF. If defined, then the created groups are added
#			    to this hash.
#
#	created_groups_list	ARRAYREF. If defined, then the created groups are added
#			    to this list.
#
#	new_groups		ARRAYREF, where the elements of the array are each a HASHREF
#				If supplied, these hashesare used as templates to create the
#				new groups.Defined fields are used, and undefined fields are
#				filled-in. This option is ignored unless 'ngroups' is 0.
#
sub createGroups ($$%)
{
  my ($filerspec, $ngroups, %opts) = validate_mixed(@_,
			[
			        { type => SCALAR|HASHREF },
				{ type => SCALAR|UNDEF },
			],
			{ 
				group_prefix 		=> { type => SCALAR|UNDEF, optional => 1 },
				group 			=> { type => SCALAR|UNDEF, optional => 1 },
				gid 			=> { type => SCALAR|UNDEF, optional => 1 },
				current_groups 		=> { type => HASHREF|UNDEF, optional => 1 },
				current_groups_list	=> { type => ARRAYREF|UNDEF, optional => 1 },
				created_groups 		=> { type => HASHREF|UNDEF, optional => 1 },
				created_groups_list	=> { type => ARRAYREF|UNDEF, optional => 1 },
				min_gid 		=> { type => SCALAR|UNDEF, optional => 1 },
				max_gid 		=> { type => SCALAR|UNDEF, optional => 1 },
				new_groups 		=> { type => ARRAYREF|UNDEF, optional => 1 },
			}
		);
     logDbg("Enter   sub createGroups  Line: ".__LINE__."\n\n".Dumper(\%opts));

  return eval {
    my $filerconn = make_conn($filerspec);
    my ($group_prefix) = $opts{group_prefix} || "wafl_group";

    my $created_groups_hash = $opts{created_groups} || {};
    my $created_groups_list = $opts{created_groups_list} || undef;

    my ($current_groups_hash, $current_groups_list);
    if (defined($opts{current_groups})) {
    	$current_groups_hash = $opts{current_groups} || {};
        if (defined($opts{current_groups_list})) {
            $current_groups_list = [ values(%$current_groups_hash) ];
        } else {
            $current_groups_list = undef;
        }
    } elsif (defined($opts{current_groups_list})) {
    	 $current_groups_list = $opts{current_groups_list} || {};
         $current_groups_hash = {};
         foreach ( @$current_groups_list ) {
             my $key = $_->{group};
             $current_groups_hash->{$key} = $_;
        }
    } else {
        $current_groups_hash = {};
        $current_groups_list = undef;
    }
    if (!keys(%$current_groups_hash)) {
         $current_groups_hash = readGroupFile($filerconn)
		or die($@);
         $current_groups_list = [ values(%$current_groups_hash) ];
    }
    if (!keys(%$current_groups_hash)) {
         $current_groups_hash = readGroupFile($filerconn)
	     or die($@);
    }

    my @new_groups;
    $ngroups = 0
        unless (defined($ngroups));
    if ($ngroups == 0 && $opts{new_groups}) {
	@new_groups = @{$opts{new_groups}};
	$ngroups = scalar(@new_groups);
    } elsif ($ngroups == 1) {
	@new_groups = ( { group=>$opts{group}, gid=>$opts{gid} } );
    } else {
	@new_groups = ();
    }

    my $num_gids_needed = 0;
    if (@new_groups) {
        foreach my $g (@new_groups) {
	    if (!defined($g->{gid})) {
	        $num_gids_needed++;
	    }
	}
    } else {
	$num_gids_needed = $ngroups;
    }
    my $next_gid = _getNextGID($num_gids_needed, $current_groups_hash, %opts);

    for (my $i=0; $i < $ngroups; $i++) {
	my $group = undef;
	my $gid = undef;

	if (@new_groups) {
		$gid = $new_groups[$i]->{gid};
		$group = $new_groups[$i]->{group};
	}
	if (!defined($gid)) {
		$gid = $next_gid + $i;
	}
	if (! defined($group)) {
		$group = sprintf "${group_prefix}%04d", $gid;
	}

	logDbg("adding grp=$gid $group ");
	my $grp_hash = {
				group         => $group,
				passwd        => "x",
				gid           => $gid,
				members       => "",
				members_list  => [],
			};
	$$created_groups_hash{$group} = $grp_hash;
	$$current_groups_hash{$group} = $grp_hash;
        if (defined($current_groups_list)) {
                push(@$current_groups_list, $grp_hash);
        }
        if (defined($created_groups_list)) {
                push(@$created_groups_list, $grp_hash);
        }
    }

    addGroupList($filerconn, $created_groups_hash)
        or die($@);

    # workaround for strange '$@' behavior ....
    logDbg("Users  Line: ".__LINE__."\n\ncreateGroups Returning:\n\n".Dumper($created_groups_hash));
    return %$created_groups_hash;
  };
}



## @name ::listCIFSUsers
## @synopsis listCIFSGroups($filerconn, %opts);
## @arg $filerconn Required Filer Connection Object
## @arg $opts -- optional args
##
##		user	  --	Optional. Name of specific CIFS user. If given, then
##				only info for this user will be returned. Otherwise, info
##				for all CIFS users will be returned.
##
## @description  Returns hash version of 'useradmin user list' output
## @returns array of hashes of useradmin (CIFS) users
sub listCIFSUsers ($;%)
{
	 my ($filerspec, %opts) =
                validate_mixed(@_,
                          [  {type => SCALAR|HASHREF },
                          ],
			  {
				user   => { type => SCALAR|UNDEF, optional => 1},
			  });
        return eval {
	        my $filerconn = make_conn($filerspec);
		my $users_text;
		if ($opts{user}) {
			$users_text = sayEOF($filerconn, "useradmin user list $opts{user}");
			if ($users_text =~ /user .* does not exist/i) {
				$users_text = "";
			}
		} else {
			$users_text = sayEOF($filerconn, "useradmin user list");
		}
		my (@users) = ();
		my $user_info = {};
		my @user_list = split(/\n/, $users_text);
		for (split(/\n/, $users_text)) {
			$_ =~ s/^\s*//;
			$_ =~ s/\s*$//;
			if (!/\S+/) {
				next;
			}
			if (/Name:\s*(.*)/) {
				$user_info = {};
				$user_info->{Name} = $1;
				push(@users, $user_info);
			} elsif (/(\w+):\s*(.*)/) {
				$user_info->{$1} = $2;
			}
		}
		if (wantarray) {
			return(@users);
		} else {
			return(\@users);
		}
	};
}


## @name ::listCIFSGroups
## @synopsis listCIFSGroups($filerconn, %opts);
## @arg $filerconn Required Filer Connection Object
## @arg $opts -- optional args
##
##		group	  --	Optional. Name of specific CIFS group. If given, then
##				only info for this group will be returned. Otherwise, info
##				for all CIFS groups will be returned.
##
## @description  Returns hash version of 'useradmin group list' output
## @returns array of hashes of useradmin (CIFS) groups
sub listCIFSGroups ($;%)
{
	 my ($filerspec, %opts) =
                validate_mixed(@_,
                          [  {type => SCALAR|HASHREF },
                          ],
			  {
				group   => { type => SCALAR|UNDEF, optional => 1},
			  });
        return eval {
	        my $filerconn = make_conn($filerspec);
		my $groups_text;
		if ($opts{group}) {
			$groups_text = sayEOF($filerconn, "useradmin group list $opts{group}");
			if ($groups_text =~ /group .* does not exist/i) {
				$groups_text = "";
			}
		} else {
			$groups_text = sayEOF($filerconn, "useradmin group list");
		}
		my (@groups) = ();
		my $group_info = {};
		for (split(/\n/, $groups_text)) {
			$_ =~ s/^\s*//;
			$_ =~ s/\s*$//;
			if (!/\S+/) {
				next;
			}
			if (/Name:\s*(.*)/) {
				$group_info = {};
				$group_info->{Name} = $1;
				push(@groups, $group_info);
			} elsif (/(\w+):\s*(.*)/) {
				$group_info->{$1} = $2;
			}
		}
		if (wantarray) {
			return(@groups);
		} else {
			return(\@groups);
		}
	};
}


## @name ::findCIFSUser
## @synopsis findCIFSUser($filerconn, $username);
## @arg $filerconn Required Filer Connection Object
## @arg $filerconn Required Name of CIFS user
## @description  Returns info for user if found, undef otherwise
## @returns hash of useradmin (CIFS) user
sub findCIFSUser ($$)
{
	my ($filerspec, $username) = @_;
	my (@users) = listCIFSUsers($filerspec, user => $username);
	if (@users) {
		return($users[0]);
	}
	return(undef);
}


## @name ::findCIFSGroup
## @synopsis findCIFSGroup($filerconn, $Groupname);
## @arg $filerconn Required Filer Connection Object
## @arg $filerconn Required Name of CIFS Group
## @description  Returns info for Group if found, undef otherwise
## @returns hash of Groupadmin (CIFS) group
sub findCIFSGroup ($$)
{
	my ($filerspec, $groupname) = @_;
	my (@groups) = listCIFSGroups($filerspec, group => $groupname);
	if (@groups) {
		return($groups[0]);
	}
	return(undef);
}


## @name ::findCIFSUserOnClient
## @synopsis findCIFSUserOnClient($username, %opts);
## @arg $filerconn Required Name of CIFS user
## @arg $opts -- optional args
##		cifs_client	--	CIFS client being used. Default is CIFS_CLIENT
## @description  Returns info for user if found, undef otherwise
## @returns hash of useradmin (CIFS) user
sub findCIFSUserOnClient ($%)
{
	my ($username, %opts) = @_;
	my $cifs_client = $opts{cifs_client} || $main::CIFS_CLIENT;
	my $cc = make_conn($opts{cifs_client});
	die("make_conn failed: $@") if ($@ || !$cc);
	my $find_cmd = "net user $username ";
	my (undef, undef, $after) = say($cc, $find_cmd);
	if ($after =~ /user\s+name\s+not\s+found/i) {
		return(undef);
	}
	return( $after );
}


## @name ::removeCIFSUsersExcept
## @synopsis removeCIFSUsersExcept($filerconn);
## @synopsis removeCIFSUsersExcept($filerconn, $user1, $user2, ...);
## @description Removes (useradmin user delete) CIFS users except those listed
## in command line. If no users are listed, then all CIFS users are deleted.
## @returns list of deleted users if success
sub removeCIFSUsersExcept($@)
{
	my ($filerspec, @users_to_save) = @_;
        my $filerconn = make_conn($filerspec);
	my (@user_hash_list) = &listCIFSUsers($filerconn);
	my (@deleted_users) = ();
	for my $user_hash (@user_hash_list) {
		my $user = $$user_hash{Name};
		if (!grep(/$user/, @users_to_save)) {
			logTrc("Deleting CIFS User: \"$user\"");
			removeCIFSUser($filerconn, $user);
			push(@deleted_users, $user);
		}
	}
	return(@deleted_users);
}


## @name ::setPrimaryGroup()
## @synopsis setPrimaryGroup($filerconn, $user, $group)
## @arg $filerconn Required Filer Connection Object
## @arg $user: user name whose group need to be set
## @arg $group: group name of the above user. The user is added to the
##            group list of this group in the /etc/group file.
## @description sets the group of the specified user to the specified group
## @returns true if success
## @status public

sub setPrimaryGroup ($$$)
{
  my ($filerspec, $input_login, $new_gid) = @_;    

  return eval {
    my $filerconn = make_conn($filerspec);
    my $oldtext = "";
    my $newpasswdfile = "";
 
    my ($login, $passwd, $uid, $gid, $name, $home, $shell);
    $oldtext = rdfile($filerconn, "/etc/passwd");
    my @oldlines = grep(!/^rdfile/, split(/\n/, $oldtext));

    foreach my $line (@oldlines) {
	if ($line =~ /^\s*$/ || $line =~ /^\s*#/) {
		$newpasswdfile .= "$line\n";
		next;
	}
	($login, $passwd, $uid, $gid, $name, $home, $shell)
			= split(/:/, $line);

	if ($login eq $input_login) {
    		$newpasswdfile .= "${login}:${passwd}:${uid}:${new_gid}:".
			"${name}:${home}:${shell}\n";
		next;
	}
	$newpasswdfile .= "$line\n";
    }

    wrfile($filerconn, "/etc/passwd", $newpasswdfile)
		or die "can't write to /etc/passwd: $@";
    sleep(4);
  }
}


## @name ::addCIFSUser($filerspec, $user, $passwd, $group, %opts)
## @synopsis addCIFSUser($filerconn, $user, $pass, $group, %opts)
## @arg $filerconn Required Filer Connection Object
## @arg $user: user name to be added.
## @arg $pass: password for user
## @arg $group: group(s) for the user (comma separated list).
## @arg $opts -- optional args
##
##		cifs_client	  --	CIFS client being used. Default is CIFS_CLIENT
##
##		on_client	  --	If 1, then add user on CIFS client also
##					Default is 0.
##
##		check_existing    --	If 1, then check for existing user before adding
##					Action then depends on value of 'remove_if_exists'
##					Default is 1.
##
##		remove_if_exists  --	If 1, and if CIFS user with same name already exists,
##					then delete existing user before creating new one.
##					Default is 1. This option is ignored unless
##					'check_existing' option is also 1.
##
##
## @description Add a local CIFS user to the filer. If 'cifs_client' and 'on_client' options
##	are given, and 'on_client' is '1', then also add user on CIFS client
## @returns true if success
## @status public

sub addCIFSUser ($$$$;%opt)
{
	 my ($spec, $user, $pass, $group, %opts) =
                validate_mixed(@_,
                          [  {type => SCALAR|HASHREF },
                             {type => SCALAR },
                             {type => SCALAR },
                             {type => SCALAR } ],
			  {
				cifs_client	   => { type => SCALAR|UNDEF, optional => 1,
				                        default=>$main::CIFS_CLIENT },
				on_client	   => { type => SCALAR|UNDEF, optional => 1, default=>0 },
				remove_if_exists   => { type => SCALAR|UNDEF, optional => 1, default=>1 },
				check_existing     => { type => SCALAR|UNDEF, optional => 1, default=>1 },
			  });

        my $added = "User <$user> added.";

        return eval {
                my $conn = make_conn($spec);
		my $text;
		if (findCIFSUser($conn, $user)) {
			if ($opts{remove_if_exists}) {
				logresult('WARN', "addCIFSUser: User $user already exists -- removing");
				removeCIFSUser($conn, $user);
			}
		}
		my $ver = get_wafl_version($conn->{hostp});
		if ($ver > WV_SCRIMSHAW_MAX) {
			# for beyond SCRIMSHAW
#                	say($conn, "useradmin user add $user -p $pass -g $group", '-wait', $added)
                	$text = sayEOF($conn, "useradmin user add $user -p $pass -g $group")
			              or die("addCIFSUser  sayEOF: $@");
			if ($text !~ /$added/) {
				die("addCIFSUser:\nExpected reply to match pattern: \"$added\"\n".
				      "Actual Reply: \"$text\"");
			}
		} else { 
			# for SCRIMSHAW and before ..
#			say($conn, "useradmin useradd $user $pass", '-wait', $added)
			$text = sayEOF($conn, "useradmin useradd $user $pass")
			              or die("addCIFSUser   sayEOF: $@");
			if ($text !~ /$added/) {
				die("addCIFSUser:\nExpected reply to match pattern: \"$added\"\n".
				      "Actual Reply: \"$text\"");
			}
		}
                if ($opts{on_client} && $opts{cifs_client}
                      && $user !~ /administrator/i) {
			my $cc = make_conn($opts{cifs_client});
			die("make_conn failed: $@") if ($@ || !$cc);
			my $add_cmd = "net user \"$user\" \"$pass\" /add";
			my (undef, undef, $after) = say($cc, $add_cmd);
			if ($after !~ /completed\s+successfully/i) {
				die("addCIFSUser:\nWindows client \'add user\' command:\n".
				     "\"$add_cmd\"   FAILED: $@:\nReply: \"$after\"");
			}
		}
		return 1;
        };
}



## @name ::removeCIFSUser($filerspec, $user, %opts)
## @synopsis removeCIFSUser($filerconn, $user, %opts)
## @arg $filerconn Required Filer Connection Object
## @arg $user: user name to remove
## @arg $opts -- optional args
##
##		cifs_client	--	CIFS client being used. Default is undef
##
##		on_client	--	If 1, then remove user on CIFS client also
##					Default is 0.
##
## @description Remove a local CIFS user from filer. If 'cifs_client' and 'on_client' options
##	are given, and 'on_client' is '1', then also remove user on CIFS client
## @returns true if success
## @status public
sub removeCIFSUser ($$;%)
{
	 my ($spec, $user, %opts) =
                validate_mixed(@_,
			[
                             {type => SCALAR|HASHREF },
                             {type => SCALAR }
			],
			{
				cifs_client	=> { type => SCALAR|UNDEF, optional => 1,
							default=>$main::CIFS_CLIENT },
				on_client	=> { type => SCALAR|UNDEF, optional => 1,
							default=>0 },
			});

	 my $no_user;
	 my $ver = get_wafl_version($spec->{hostp})
		 or die $@;
	
	 my $deleted = "User <$user> deleted.";
	
	 return eval {
		 my $conn = make_conn($spec);
		 my $text;
		 if ($ver > WV_SCRIMSHAW_MAX) {
			 $no_user = "User does not exist";
#			 (undef, undef, $text) = say($conn, "useradmin user delete $user")
                	 $text = sayEOF($conn, "useradmin user delete $user")
			              or die("removeCIFSUser  sayEOF: $@");
			if ($text !~ /($deleted|$no_user)/) {
				die("removeCIFSUser:\nExpected reply to match pattern: \"$deleted\"\n".
				      "Actual Reply: \"$text\"");
			}
		 } else {
			 $no_user = "User \<$user\> not present";
#			 (undef, undef, $text) = say($conn, "useradmin userdel $user")
                	 $text = sayEOF($conn, "useradmin userdel $user")
			              or die("removeCIFSUser  sayEOF: $@");
			if ($text !~ /($deleted|$no_user)/) {
				die("removeCIFSUser:\nExpected reply to match pattern: \"$deleted\"\n".
				      "Actual Reply: \"$text\"");
			}
		 }

                if ($opts{on_client} && $opts{cifs_client}
                      && $user !~ /administrator/i
		      && findCIFSUserOnClient($user, cifs_client => $opts{cifs_client}) ) {
			my $cc = make_conn($opts{cifs_client});
			die("make_conn failed: $@") if ($@ || !$cc);
			my $del_cmd = "net user $user /delete";
			my (undef, undef, $after) = say($cc, $del_cmd);
			if ($after !~ /completed\s+successfully/i) {
				die("removeIFSUser:\nWindows client \'delete user\' command:\n".
				     "\"$del_cmd\"   FAILED: $@:\nReply: \"$after\"");
			}
		}
		return 1;
	 };
}



## @name ::addCIFSGroup($filerspec, $group, %opts)
## @synopsis addCIFSGroup($filerconn, $group, %opts)
## @arg $filerconn Required Filer Connection Object
## @arg $group: Name of group to create.
## @arg $role: roles(s) for the group. Default is "none".
## @arg $opts -- optional args
##
##		role	--	Role(s) for the new group (comma separated list).
##
## @description Add a local CIFS user to the filer. If 'cifs_client' and 'on_client' options
##
## @description Add a local group to the filer.
## @returns true if success
## @status public

sub addCIFSGroup ($$%opt)
{
	 my ($spec, $group, %opts) =
                validate_mixed(@_,
                          [  {type => SCALAR|HASHREF },
                             {type => SCALAR } ],
			  {
				role	           => { type => SCALAR, optional => 1, default => "" },
			  });

        my $added = "The group '$group' has been added|Group <$group> added";

        return eval {
                my $conn = make_conn($spec);
		my $role = $opts{role};
		my $text;
		if ($role) {
#	                say($conn, "useradmin group add $group -r $role", '-wait', $added)
                	$text = sayEOF($conn, "useradmin group add $group -r $role")
			              or die("addCIFSGroup  sayEOF: $@");
			if ($text !~ /$added/) {
				die("addCIFSGroup:\nExpected reply to match pattern: \"$added\"\n".
				      "Actual Reply: \"$text\"");
			}
		} else {
#	                say($conn, "useradmin group add $group", '-wait', $added)
			$text = sayEOF($conn, "useradmin group add $group")
			              or die("addCIFSGroup  sayEOF: $@");
			if ($text !~ /$added/) {
				die("addCIFSGroup:\nExpected reply to match pattern: \"$added\"\n".
				      "Actual Reply: \"$text\"");
			}
		}
		return 1;
        };
}



## @name ::removeCIFSGroup()
## @synopsis removeCIFSGroup($filerconn, $group)
## @arg $filerconn Required Filer Connection Object
## @arg $group: group name to remove
## @description Remove a local group from filer.
## @returns true if success
## @status public
sub removeCIFSGroup ($$)
{
	 my ($spec, $group) =
                validate_mixed(@_,
			[
                             {type => SCALAR|HASHREF },
                             {type => SCALAR }
			],
			{ },
			);

        my $deleted = "The group '$group' has been deleted.|Group <$group> deleted";

        return eval {
                my $conn = make_conn($spec);
		my $text;
#                say($conn, "useradmin group delete $group", '-wait', $deleted)
                $text = sayEOF($conn, "useradmin group delete $group")
		              or die("removeCIFSGroup  sayEOF: $@");
		if ($text !~ /$deleted/) {
			die("removeCIFSGroup:\nExpected reply to match pattern: \"$deleted\"\n".
			      "Actual Reply: \"$text\"");
		}
		return 1;
        };
}



## @name ::appendPasswdFile()
## @synopsis appendPasswdFile($filerspec, @text_list)
## @arg $filerconn Required Filer Spec or Connection Object
## @arg @text_list Required. Text to be appended to the /etc/passwd file.
## @description Append text to /etc/passwd file
## @returns true if success
## @status public
sub appendPasswdFile($@)
{
        my ($filerspec, @text_list) = @_;
        my $newtext = join("\n", @text_list);
        $newtext .= "\n" unless ($newtext =~ /\n$/);

	my $filerconn = make_conn($filerspec);
	my $oldtext = rdfile($filerconn, "/etc/passwd");
        my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
        $oldtext = join("\n", @lines);
	croak($@) if ($@);
	$oldtext .= "\n" unless ($oldtext =~ /\n$/);
	wrfile($filerconn, "/etc/passwd", $oldtext . $newtext)
		or croak("can't write to /etc/passwd : $@");
        sleep(4);
	return(1);
}


## @name ::writeDefaultPasswdFile()
## @synopsis writeDefaultPasswd($filerspec)
## @arg $filerconn Required Filer Spec or Connection Object
## @description Writes short default /etc/passwd file:
##    root::0:1::/:
##    pcuser::65534:65534::/:
##    nobody::65535:65535::/:
##    ftp::65533:65533:FTP Anonymous:/home/ftp:
## @returns true if success
## @status public
sub writeDefaultPasswdFile($@)
{
        my ($filerspec) = @_;
        my $text = <<__EOF__;
root::0:1::/:
pcuser::65534:65534::/:
nobody::65535:65535::/:
ftp::65533:65533:FTP Anonymous:/home/ftp:
__EOF__

	my $filerconn = make_conn($filerspec);
	wrfile($filerconn, "/etc/passwd", $text)
			or croak("can't write to /etc/passwd : $@");
	sleep(4);
	return(1);
}


## @name ::writePasswdFile()
## @synopsis writePasswd($filerspec, @text_list)
## @arg $filerconn Required Filer Spec or Connection Object
## @arg @text_list Required. Text to be appended to the /etc/passwd file.
## @description Write text to /etc/passwd file. Overwrites previous contents.
## @returns true if success
## @status public
sub writePasswdFile($@)
{
        my ($filerspec, @text_list) = @_;
        my $text = join("\n", @text_list);
        $text .= "\n" unless ($text =~ /\n$/);

	my $filerconn = make_conn($filerspec);
	wrfile($filerconn, "/etc/passwd", $text)
			or croak("can't write to /etc/passwd : $@");
        sleep(4);
	return(1);
}


## @name ::writeGroupFile()
## @synopsis writeGroup($filerspec, @text_list)
## @arg $filerconn Required Filer Spec or Connection Object
## @arg @text_list Required. Text to be appended to the /etc/group file.
## @description Write text to /etc/group file. Overwrites previous contents.
## @returns true if success
## @status public
sub writeGroupFile($@)
{
        my ($filerspec, @text_list) = @_;
        my $text = join("\n", @text_list);
        $text .= "\n" unless ($text =~ /\n$/);

	my $filerconn = make_conn($filerspec);
	wrfile($filerconn, "/etc/group", $text)
			or croak("can't write to /etc/group : $@");
	return(1);
}


## @name ::appendGroupFile()
## @synopsis appendGroupFile($filerspec, @text_list)
## @arg $filerconn Required Filer Spec or Connection Object
## @arg @text_list Required. Text to be appended to the /etc/group file.
## @description Append text to /etc/group file
## @returns true if success
## @status public
sub appendGroupFile($@)
{
        my ($filerspec, @text_list) = @_;
        my $newtext = join("\n", @text_list);
        $newtext .= "\n" unless ($newtext =~ /\n$/);

	my $filerconn = make_conn($filerspec);
	my $oldtext = rdfile($filerconn, "/etc/group");
        my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
        $oldtext = join("\n", @lines);
	croak($@) if ($@);
	$oldtext .= "\n" unless ($oldtext =~ /\n$/);
	wrfile($filerconn, "/etc/group", $oldtext . $newtext)
		or croak("can't write to /etc/group : $@");
	return(1);
}


## @name ::readPasswdFile
## @synopsis readPasswdFile($filerconn, %args)
#
#  Args:
#
## @arg	filerconn  Required.
## @arg Optional Args:
##	users:		If defined, then this is an HASHREF. The users
##			will be written to the HASHREF
## @description	Reads /etc/passwd and returns a hashtable represnting the users found there.
## @returns parsed passwd file as HASHREF
sub readPasswdFile($%)
{
  my ($filerspec, %args) = @_;
  my $filerconn = make_conn($filerspec);
  my (%users) = ();
  my ($line);
  my $usershash;
  if (defined($args{users})) {
	$usershash = $args{users};
  } else {
	$usershash = {};
  }

  return eval {
        my $oldtext = rdfile($filerconn, "/etc/passwd");
        my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
        foreach my $line (@lines) {
                if ($line =~ /^\s*$/ || $line =~ /^\s*#/) {
                        #print NEWPASSWDFILE $line;
                        next;
                }

                my ($login, $passwd, $uid, $gid, $name, $home, $shell)
                        = split(/:/, $line);
                my $user_hash = { login => $login, passwd => $passwd,
                                uid => $uid, gid => $gid, name => $name,
                                home => $home, shell => $shell };
		$$usershash{$login} = $user_hash;
        }
        return $usershash;
    };

}


#
# addPasswdEntry
#
#   Appends the entry to /etc/passwd for the user with the specified
#   login, uid, and gid
#
# Args
#
#	filerconn:		Required.
#
#	login:			Required.
#
#	uid:			Required.
#
#	gid:			Required
#
# Optional Args:
#
#	passwd:			Optional. Default is "x"
#

sub addPasswdEntry($$$$%)
{
  my ($filerspec, $login, $uid, $gid, %opts) = @_;
  my $filerconn = make_conn($filerspec);
  my (%user) = ();

  return eval {
    if (!defined($opts{CIFS_PASSWD})) {
        if ($::CIFS_PASSWD) {
            $opts{passwd} = $::CIFS_PASSWD;
        } else {
            logWarn("addPasswdEntry: no pawwd was given and \$::CIFS_PASSWD not defined. Passwd defaults to \"x\"");
            $opts{passwd} = "x";
        }
    }
    my $oldtext = rdfile($filerconn, "/etc/passwd");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    my $newline = "${login}:x:${uid}:${gid}:::";
    push(@lines, $newline);
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/etc/passwd", $oldtext)
		or croak("can't write to /etc/passwd : $@");

    sleep(4);
    return (login => $login, passwd => "x", uid => $uid,
		gid => $gid, name => "", home => "", shell => "");
  };
}


#
# delPasswdEntry
#
#   Delete the /etc/passwd entry for the user with the specified
#   login.
#
# Args
#
#	filerconn:		Required.
#
#	login:			Required.
#
sub delPasswdEntry($$)
{
  my ($filerconn, $login) = @_;

  return eval {

    my $oldtext = rdfile($filerconn, "/etc/passwd");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    @lines = grep(!/^${login}:/, @lines);
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/etc/passwd", $oldtext)
		or croak("can't write to /etc/passwd : $@");
    sleep(4);
  };
}


#
# delPasswdList
#
#   Delete specified users from /etc/passwd
#
# Args
#
#	filerconn:		Required.
#
#	userhash:		Required. HASHREF for list of users
#				to be deleted
#
sub delPasswdList($$)
{
  my ($filerconn, $userhash) = @_;

  return eval {

    my $oldtext = rdfile($filerconn, "/etc/passwd");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    my @newlines = ();
    foreach (@lines) {
        my ($login) = (/^([^:]+):/);
        push(@newlines, $_) unless (exists($userhash->{$login}));
    }
    $oldtext = join("\n", @newlines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/etc/passwd", $oldtext)
		or croak("can't write to /etc/passwd : $@");
    sleep(4);
  };
}


#
# addPasswdList
#
#	Adds entries to /etc/passwd for the users in userhash
#
# Args
#
#	filerconn:		Required.
#
#	userhash:		Required. HASHREF for list of users
#				to be added
#
# Optional Args:
#
#	cifs_user:		Optional. Default is 0. If '1', then
#				use the 'winlogin' and 'winpasswd' instead of
#				'login', and 'passwd' fields for the users.
#

sub addPasswdList($$%)
{
  my ($filerspec, $userhash, %opt) = @_;

  return eval {
    my $filerconn = make_conn($filerspec);
    my $oldtext = rdfile($filerconn, "/etc/passwd");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    my @keys = keys %$userhash;
    foreach my $u (@keys) {
	    my $info = $userhash->{$u};
	    my ($login, $passwd);
	    if ($opt{cifs_user}) {
	    	$login = $info->{winlogin};
	    	$passwd = $info->{winpasswd} || $main::CIFS_PASSWD || "cifspass123";
	    } else {
	    	$login = $info->{login};
	    	$passwd = $info->{passwd} || "x";
	    }
	    my $uid = $info->{uid};
	    my $gid = $info->{gid};
	    my $name = $info->{name} || "";
	    my $home = $info->{home} || "";
	    my $shell = $info->{shell} || "";

	    my $newline = join(":", $login, $passwd, $uid, $gid, $name, $home, $shell);
	    push(@lines, $newline);
    }

    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/etc/passwd", $oldtext)
	or croak("can't write to /etc/passwd : $@");
    sleep(4);

  }
}


#
#  readGroupFile
#
#	Reads /etc/group and returns a hashtable represnting groups found there.
#
#  Args:
#
#	filerconn:		Required
#
#  Optional Args:
#
#	groups:			If defined, then this is an HASHREF. The groups
#				will be written to the HASHREF
#
sub readGroupFile($%)
{
  my ($filerspec, %args) = @_;
  my $filerconn = make_conn($filerspec);
  my (%groups) = ();
  my ($line);
  my $groupshash;
  if (defined($args{groups})) {
	$groupshash = $args{groups};
  } else {
	$groupshash = {};
  }
  return eval {
	my $oldtext = rdfile($filerconn, "/etc/group");
	my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
	foreach my $line (@lines) {
		if ($line =~ /^\s*$/ || $line =~ /^\s*#/) {
	    		next;
		}
		my ($group, $passwd, $gid, $members)
			= split(/:/, $line);
		my $grp_hash = { group         => $group,
			        passwd         => $passwd, 
				gid            => $gid,
				members        => $members,
				members_list   => [split(/,/, $members)]
			     };
		$$groupshash{$group} = $grp_hash;
    	}
    return( $groupshash );
  }
}


#
# addGroupEntry
#
#   Appends the entry to /etc/group for the group with the specified
#   name and gid.
#
# Args
#
#	filerconn:		Required.
#
#	name:			Required.
#
#	gid:			Required
#
# Optional args:
#
#	members:		Optional. For "members" field, default is
#				empty string
#
sub addGroupEntry($$$;$)
{
  my ($filerspec, $name, $gid, $members) = @_;

  $members = "" if (!defined($members));

  return eval {

    my $filerconn = make_conn($filerspec);
    my $oldtext = rdfile($filerconn, "/etc/group");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    push(@lines, "${name}:x:${gid}:${members}");
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/etc/group", $oldtext)
		or croak("can't write to /etc/group : $@");

    return (group            => $name,
	      passwd         => "x",
	      gid            => $gid,
	      members        => $members,
	      members_list   => 
		((ref($members) =~ /ARRAY/) ?
		       $members :
		       [ split(/,/, $members) ] )
            );
  };
}


#
# delGroupEntry
#
#   Delete the existing /etc/group entry for the group with
#   the specified name
#
# Args
#
#	filerconn:		Required.
#
#	name:			Required.
#
sub delGroupEntry($$)
{
  my ($filerconn, $group) = @_;

  return eval {

    my $oldtext = rdfile($filerconn, "/etc/group");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    @lines = grep(!/^${group}:/, @lines);
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/etc/group", $oldtext)
		or croak("can't write to /etc/group : $@");
  };
}


#
# delGroupList
#
#   Delete specified groups from /etc/group
#
# Args
#
#	filerconn:		Required.
#
#	grouphash:		Required. HASHREF for list of groups
#				to be deleted
#
sub delGroupList($$)
{
  my ($filerconn, $grouphash) = @_;

  return eval {

    my $oldtext = rdfile($filerconn, "/etc/group");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    my @newlines = ();
    foreach (@lines) {
        my ($gname) = (/^([^:]+):/);
        push(@newlines, $_) unless (exists($grouphash->{$gname}));
    }
    $oldtext = join("\n", @newlines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/etc/group", $oldtext)
		or croak("can't write to /etc/group : $@");
  };
}


#
# addGroupList
#
#	Adds entries to /etc/group for the groups in grouphash
#
# Args
#
#	filerconn:		Required.
#
#	grouphash:		Required. HASHREF for list of groups
#				to be added
#
sub addGroupList($$)
{
  my ($filerspec, $grouphash) = @_;

  return eval {
    my $filerconn = make_conn($filerspec);
    my $oldtext = rdfile($filerconn, "/etc/group");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    my @keys = keys %$grouphash;
    foreach my $g (@keys) {
	    my $info = $grouphash->{$g};
	    my $group = $info->{group};
	    my $passwd = $info->{passwd} || "x";
	    my $gid = $info->{gid};
	    my $members = $info->{members} || "";
	    push(@lines, "${group}:${passwd}:${gid}:${members}");
    }
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/etc/group", $oldtext)
		or croak("can't write to /etc/group : $@");
  }
}


# ----------------------------
# Supporting local subroutines 
# ----------------------------


sub _array_from_hash($)
{
    my ($hashref) = @_;
    my $arrayref = [];
    while (my ($key, $val) = each(%$hashref)) {
	push(@$arrayref, $val);
    }
    return($arrayref);
}


#
#   _getNextUID
#
#  Args:
#
#	nusers:		Required. Number of new UIDs to create
#
#
#	current_users_hash		HASHREF which represents current users. If defined,
#			    and the HASHREF is empty, then the HASHREF is updated with
#			    current users from /etc/passwd. Otherwise the  HASHREF is
#			    used to determine current users, and /etc/passwd is not read.
#			    The HASHREF is also updated with any new users created
#
#	current_users_list	HASHREF which represents current users. If defined,
#   Optional args:
#
#	min_uid		Default is 10000
#	max_uid		Default is 0x7fffffff
#
#  If min_uid is not specified, and max_uid is specified, and is less than
#  10000, then the greatest of 1000, 100, 10 which is less than max_uid
#  is used as the value for min_uid
#
sub _getNextUID($$%)
{
  my ($nusers, $current_users_hash, %args) = @_;
  if (!$current_users_hash) {
          die("_getNextUID  current_users_hash is undefined");
  }
  if (!$nusers) {
	return(undef);
  }
  my ($max_uid) = (defined($args{max_uid}) ? $args{max_uid} : 0x7fffffff);
  my ($min_uid);
  if (defined($args{min_uid})) {
	$min_uid = $args{min_uid};
  } elsif ($max_uid > 10000) {
	$min_uid = 10000;
  } elsif ($max_uid > 1000) {
	$min_uid = 1000;
  } elsif ($max_uid > 100) {
	$min_uid = 100;
  } else {
	$min_uid = 10;
  }

  my (@uids) = ($min_uid-1, $max_uid);
  return eval {
    foreach my $login (keys %$current_users_hash) {
	my $u = $$current_users_hash{$login}->{uid};
	push @uids, $u if ($u >= $min_uid);
    }
    @uids = sort {$a <=> $b} @uids;

    #  Look for gaps in uid list
    my ($next_uid) = 0;
    my $found = 0;
    for (my $i = 0; $i < $#uids && !$found; $i++) {
	if ($uids[$i+1] - $uids[$i] > $nusers) {
	  $next_uid = $uids[$i] + 1;
          $found = 1;
    	}
    }
    if ($found) {
	return($next_uid);
    } else {
	return(undef);
    }
  }
}

#
#   Exported version of _getNextUID
#
sub getNextUID($%)
{
	my ($fc, %args) = @_;
	my $nusers = $args{nusers} || 1;
	delete($args{nusers});
	return( _getNextUID($nusers, readPasswdFile($fc), %args) );
}

#
#   _getNextGID
#
#  Args:
#
#	ngroups:	Required. Number of new GIDs to create
#
#   Optional args:
#
#	min_gid		Default is 10000
#	max_gid		Default is 0x7fffffff
#
#  If min_gid is not specified, and max_gid is specified, and is less than
#  10000, then the greatest of 1000, 100, 10 which is less than max_gid
#  is used as the value for min_gid
#
sub _getNextGID($$%)
{
  my ($ngroups, $current_groups_hash, %args) = @_;

  if (!$ngroups) {
	return(undef);
  }
  my ($max_gid) = (defined($args{max_gid}) ? $args{max_gid} : 0x7fffffff);
  my ($min_gid);
  if (defined($args{min_gid})) {
	$min_gid = $args{min_gid};
  } elsif ($max_gid > 10000) {
	$min_gid = 10000;
  } elsif ($max_gid > 1000) {
	$min_gid = 1000;
  } elsif ($max_gid > 100) {
	$min_gid = 100;
  } else {
	$min_gid = 10;
  }

  my (@gids) = ($min_gid-1, $max_gid);
  return eval {
    foreach my $group (keys %$current_groups_hash) {
	my $u = $$current_groups_hash{$group}->{gid};
	push @gids, $u if ($u >= $min_gid);
    }
    @gids = sort {$a <=> $b} @gids;

    #  Look for gaps in gid list
    my ($next_gid) = 0;
    my $found = 0;
    for (my $i = 0; $i < $#gids && !$found; $i++) {
	if ($gids[$i+1] - $gids[$i] > $ngroups) {
	  $next_gid = $gids[$i] + 1;
          $found = 1;
    	}
    }
    if ($found) {
	return($next_gid);
    } else {
	return(undef);
    }
  }
}

#
#   Exported version of _getNextGID
#
sub getNextGID($%)
{
	my ($fc, %args) = @_;
	my $ngroups = $args{ngroups} || 1;
	delete($args{ngroups});
	return( _getNextGID($ngroups, readGroupFile($fc), %args) ); 
}



#
# _findInList
#
#	Searches a given ARRAYREF for matching members, and returns
#	any matches.
#
# Args:
#
#	list:		Required. ARRAYREF from which matching members
#			are to be chosen
#
#	args:		Key, value pairs which are used to select
#			matching members

sub _findInList($%)
{
	my ($arraylist, %args) = @_;
	my (@keys) = keys(%args);
	my @matches = ();
	for my $key (@keys) {
		delete($args{$key}) unless (defined($args{$key}));
	}
	@keys = keys(%args);
	return() unless (scalar(@keys));
	for my $elt (@$arraylist) {
		my $match = 1;
		my $args_ref = { %args };
		while ( my ($key, $value) = each(%args)) {
			if ($elt->{$key} ne $value) {
				$match = 0;
				last;
			}
		}
		if ($match) {
			push(@matches, $elt);
		}
	}
	return(@matches);
}



#
# _findInHashList
#
#	Searches a given 'hash list' for matching members, and returns
#	any matches. The 'hash list' format is that returned by for example,
#	readPasswdFile, readGroupFile, createUsers, createGroups, etc.
#
# Args:
#
#	hashlist:	Required. HASHREF from which matching members
#			are to be chosen
#
#	args:		Key, value pairs which are used to select
#			matching members

sub _findInHashList($%)
{
	my ($hashlist, %args) = @_;
	my (@keys) = keys(%args);
	my @matches = ();
	my $arraylist = _array_from_hash($hashlist);
        return ( _findInList($arraylist, %args) );
}


#
# _findUser
#
#	Returns users from current users which match criterea determined
#	by %args
#
# Args:
#
#	current_users_hash:	Required. HASHREF from which matching users
#				are to be chosen
#
#	args:			Key, value pairs which are used to select
#				matching users

sub _findUser($%)
{
	my ($current_users_hash, %args) = @_;
	return ( _findInHashList($current_users_hash, %args) );
}


#
#  Exported version of _findUser
#
#  args:
#	filerspec	required
#
#	either:
#		(1)	name: 'login' (name) of user
#
#	or
#		(2)	%args: list of key-value pairs describing user
#
sub findUser($%)
{
	my $filerspec = shift;
        my $fc = make_conn($filerspec);
	if (@_ == 1) {
		my ($name) = @_;
		return ( _findUser(readPasswdFile($fc), login=>$name) );
	} else {
		my (%args) = @_;
		return ( _findUser(readPasswdFile($fc), %args) );
	}
}


#
# _findGroup
#
#	Returns groups from current_groups which match criterea determined
#	by %args
#
# Args:
#
#	current_groups_hash:	Required. HASHREF from which matching groups
#				are to be chosen
#
#	args:			Key, value pairs which are used to select
#				matching groups

sub _findGroup($%)
{
	my ($current_groups_hash, %args) = @_;
	return ( _findInHashList($current_groups_hash, %args) );
}


#
#  Exported version of _findGroup
#
#  args:
#	filerspec	required
#
#	either:
#		(1)	name: name of group
#
#	or
#		(2)	%args: list of key-value pairs describing group
#
sub findGroup($%)
{
	my $filerspec = shift;
        my $fc = make_conn($filerspec);
	if (@_ == 1) {
		my ($name) = @_;
		return ( _findGroup(readGroupFile($fc), group=>$name) );
	} else {
		my (%args) = @_;
		return ( _findGroup(readGroupFile($fc), %args) );
	}
}


1;
