
# $Id$
## @summary Common library for CSV functions
## @author bellizzi@netapp.com
## @pod here

=head1 NAME

Utils::CSV.pm - Library for CSV functions

=head1 DESCRIPTION

This package defines routines which can be used for CSV manipulation

=head1 SYNOPSIS

    use Utils::CSV qw(create_empty_data_set write_csv_file);

=head1 ROUTINES

=cut

package Utils::CSV;
use strict;
use Tharn;
$Log->comment("Starting " . __FILE__);
use POSIX qw(strftime);
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use Data::Dumper;

our $VERSION = 1.00;
our @ISA     = qw (Exporter);
our @EXPORT  = qw (create_empty_data_set write_csv_file);

=head2 create_empty_data_set

    $data = create_empty_data_set('headers' => \@headers, 'string' => ' ');

C<create_empty_data_set> 

=over

=item headers

(Required) Array of column headers

=item string

(Required) string to write in place or real data

=back

=cut

sub create_empty_data_set {

    $Log->enter();

    my ($headers, $string) = @_;
    my (%opts) =
      validate(@_,
               {headers => {type => ARRAYREF, optional => 0},
                string  => {type => SCALAR,   optional => 0}
               }
               );

    $Log->comment("Creating emtpy data set");

    # Create a new timestamp
    my %data;

    for (my $i = 1 ; $i < @{$opts{'headers'}} ; $i++) {
        $data{${$opts{'headers'}}[$i]} = $opts{'string'};
    }

    $Log->exit();

    return \%data;
}

=head2 write_csv_file

    This routine adds a timestamp to the header and data arrays, comma separates
    the items in both arrays, and writes them to a .csv file in TOP_LOGDIR and
    optionally ALT_LOGDIR.

    write_csv_file('conn' => $fc, 'node' => $node);

C<write_csv_file> Use NET_PREFER to determine the IP address to use.

=over

=item header

(Required) Hashref of column headers

=item data

(Required) Hashrefof data

=item file_name

(Required) The name of the CSV file

=item timestamp

(Required) The timestamp for the current record

=item write_headers

(Required) Toggles the wrting of column headers

=item subsys

(Optional) The stats subsystem 

=item timestamp_format

(Optional) The format of the timestamp

=item top_logdir

(Optional) NATE's TOP_LOGDIR

=item alt_logdir

(Optional) The alternate log path

=back

=cut

sub write_csv_file {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   headers          => {type => ARRAYREF,       optional => 0},
            data             => {type => ARRAYREF,       optional => 0},
            file_name        => {type => SCALAR,         optional => 0},
            timestamp        => {type => SCALAR,         optional => 0},
            write_headers    => {type => SCALAR,         optional => 0},
            timestamp_format => {type => SCALAR,         optional => 1},
            path             => {type => SCALAR | UNDEF, optional => 1},
            alt_path         => {type => SCALAR | UNDEF, optional => 1},
            timestamp_name   => {type => SCALAR,         optional => 1},
            overwrite        => {type => SCALAR | UNDEF, default  => 0},
        }
    );


    my $headers;

    if ($opts{'overwrite'}) {
	$opts{'write_headers'} = 1;
    }

    #------------------------------------------------------------
    # Prepend the timestamp and comma separate the headers values
    #------------------------------------------------------------
    if ($opts{'write_headers'}) {
        my $timestamp_name = 'Timestamp';
        if ($opts{'timestamp_name'}) {
            $timestamp_name = $opts{'timestamp_name'};
        }
        unshift(@{$opts{'headers'}}, $timestamp_name);
        $headers = join(",", @{$opts{'headers'}}) . "\n";
    }

    #----------------------------------------------------------
    # Prepend the timestamp and comma separate the data values
    #----------------------------------------------------------
    if ($opts{'timestamp_format'} eq "date_and_time") {

        #-------------------------------------------------
        # Convert epoch seconds to date/time if requested
        #-------------------------------------------------
        $opts{'timestamp'} =
          strftime("%m/%d/%y %H:%M:%S", localtime($opts{'timestamp'}));
    }

    unshift(@{$opts{'data'}}, $opts{'timestamp'});
    my $data = join(",", @{$opts{'data'}}) . "\n";

    #---------------------------------------------------------
    # Open the log file and alternate log file (if specified)
    #---------------------------------------------------------
    my $log_file     = $opts{'file_name'};
    my $log_path     = $opts{'path'};
    my $alt_log_path = $opts{'alt_path'};

    if (!-e $log_path) {
        eval { mkdir($log_path); };
        if ($@) {
            $Log->warn("Error creating $log_path: $@");
        }
    }

    if ($alt_log_path && !-e $alt_log_path) {
        eval { mkdir($alt_log_path); };
        if ($@) {
            $Log->warn("Error creating $alt_log_path: $@");
        }
    }

    $Log->debug("Opening $log_path/$log_file");
    if ($opts{'overwrite'}) {
        open(LOG, ">$log_path/$log_file");
    } else {
        open(LOG, ">>$log_path/$log_file");
    }

    if ($opts{'alt_path'}) {
        $Log->debug("Opening $alt_log_path/$log_file");
        open(ALTLOG, ">>$alt_log_path/$log_file");
    }

    #-----------------------------------------------------
    # Write headers data if it hasn't already been written
    #-----------------------------------------------------
    if ($opts{'write_headers'}) {
        $Log->debug("Writing headers to $log_path/$log_file");
        print LOG $headers;
        if ($opts{'alt_path'}) {
            $Log->debug("Writing headers to $alt_log_path/$log_file");
            print ALTLOG $headers;
        }
    }

    #--------------------------------------
    # Write data values to the log file(s)
    #--------------------------------------
    $Log->debug("Writing data to $log_path/$log_file");
    print LOG $data;
    if ($opts{'alt_path'}) {
        $Log->debug("Writing data to $alt_log_path/$log_file");
        print ALTLOG $data;
    }

    close LOG;
    if ($opts{'alt_path'}) {
        close ALTLOG;
    }

    $Log->exit();

}
1;
