# $Id$
## @summary Common library for DNS functions
## @author bellizzi@netapp.com
## @pod here

=head1 NAME

Utils::DNS.pm - Library for DNS functions

=head1 DESCRIPTION

This package defines routines which can be used for forward and reverse 
namerserver lookups;

=head1 SYNOPSIS

    use Utils::DNS qw(get_ip_by_name);

=head1 ROUTINES

=cut

package Utils::DNS;
use strict;
use Tharn;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);
use Module::Common qw(validate_mixed);
use TracePkg;
use Data::Dumper;
use Net::DNS;

my $Trace = new TracePkg;

our $VERSION = 1.00;
our @ISA     = qw (Exporter);
our @EXPORT  = qw (
    resolve
    get_ip_by_name
    get_name_by_ip
);

=head2 get_ip_by_name

    use Utils::DNS qw(get_ip_by_name);
    failed_results('name' => $hostname');

C<get_ip_by_name> Resolve IP addresses from a hostname

=over

=item name

(Required) The hostname to resolve

=back

=cut

sub get_ip_by_name {

    my (%opts) = validate(
        @_,
        {   name    => {type => SCALAR, optional => 0},
            records => {type => SCALAR, default  => 'first'}
        }
    );

    my @result;

    my $answer = resolve('host' => $opts{'name'});
    if ($answer->[0]->{'CNAME'}) {
        my $ca = resolve('host' => $answer->[0]->{'CNAME'});
        foreach my $a (@{$ca}) {
            $a->{CNAME} = $answer->[0]->{'CNAME'};
            push(@result, $a);
        }
    } else {
        foreach my $a (@{$answer}) {
            push(@result, $a);
        }
    }

    if ($opts{records} eq 'first') {
        return $result[0];
    } else {
        return \@result;
    }

} ## end sub get_ip_by_name

=head2 get_name_by_ip

    use Utils::DNS qw(get_name_by_ip);
    failed_results('ip' => $ip');

C<get_name_by_ip> Resolve a hostname from a IP address

=over

=item ip

(Required) The IP address to resolve

=back

=cut

sub get_name_by_ip {

    my (%opts) = validate(@_, {ip => {type => SCALAR, optional => 0},});

    my $answer = resolve('host' => $opts{'ip'});

    return $answer->[0];
} ## end sub get_name_by_ip

=head2 resolve

    use Utils::DNS qw(resolve);
    resolve('host' => $host');

C<resolve> Search subtest objects for failed results

=over

=item host

(Required) The hostname or IP address to resolve

=back

=cut

sub resolve {

    my (%opts) = validate(@_, {host => {type => SCALAR, optional => 0},});
    my @answer;

    my $res = Net::DNS::Resolver->new(dnsrch => 0);
    my $query = $res->search($opts{'host'}, 'ANY');

    if ($query) {
        foreach my $rr ($query->answer) {

            #print "RR:" . Dumper($rr) . "\n";
            if ($rr->type eq 'A') {
                push(
                    @answer,
                    {   A     => $rr->address,
                        INET4 => $rr->address,
                        AAAA  => '',
                        INET6 => '',
                        CNAME => '',
                        PTR   => '',
                        ERROR => ''
                    }
                );
            } elsif ($rr->type eq 'AAAA') {
                push(
                    @answer,
                    {   A     => '',
                        INET4 => '',
                        AAAA  => $rr->address,
                        INET6 => $rr->address,
                        CNAME => '',
                        PTR   => '',
                        ERROR => ''
                    }
                );
            } elsif ($rr->type eq 'CNAME') {
                push(
                    @answer,
                    {   A     => '',
                        INET4 => '',
                        AAAA  => '',
                        INET6 => '',
                        CNAME => $rr->cname,
                        PTR   => '',
                        ERROR => ''
                    }
                );
            } elsif ($rr->type eq 'PTR') {
                push(
                    @answer,
                    {   A     => '',
                        INET4 => '',
                        AAAA  => '',
                        INET6 => '',
                        CNAME => '',
                        PTR   => $rr->ptrdname,
                        ERROR => ''
                    }
                );
            }
        }
    } else {
        push(
            @answer,
            {   A     => '',
                INET4 => '',
                AAAA  => '',
                INET6 => '',
                CNAME => '',
                PTR   => '',
                ERROR => $res->errorstring
            }
        );
    }
    return \@answer;
} ## end sub resolve

#--------------------------------------------------------------------
1;
