#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Utils/Error.pm#1 $

## @summary Provides basic error handling methods   
## that should be inherited by other object oriented
## packages
##
## @author knopp@netapp.com
## @status Shared
## @pod here

package Utils::Error;

use lib();

# Compiler directives.
use strict;


###############################################################################
# Method:    errmsg
# Objective: Return the errmsg component from the object.
# Details:   See POD documentation at the end of this file.
###############################################################################

sub errmsg {
    my($self) = @_;

    if (ref($self->{errmsg}) eq "ARRAY") {
	 return @{$self->{errmsg}};
    }
    else {
	return ("NO ERROR MESSAGE");
    }
} # end sub errmsg



###############################################################################
# Method:    error
# Objective: Return error status of the err object.
# Details:   See POD documentation at the end of this file.
###############################################################################

sub error { 
    my($self) = @_;

    return $self->{error};

} # end sub error


###############################################################################
# Method:    _clear_err
# Objective: Clear error components in the object.
# Inputs:    None.
# OuTCaseuts:   None.
# Affected Data: Error components in the object have been cleared.
###############################################################################

sub _clear_err {
    my($self) = @_;

    $self->{error} = 0;
    $self->{errmsg} = '';

    return 1;
} # end sub _clear_err

###############################################################################
# Method:    _get_err_stack_depth
# Objective: return depth of call stack
# Inputs:    None.
# Outputs:   depth of output (scalar)
# Affected Data: Error components in the object have been cleared.
###############################################################################

sub _get_err_stack_depth {
    my ($self)=@_;
    my ($depth);
    
    $depth=(defined $self->{stack_depth})
	  ? $self->{stack_depth}
	  : 1;
    
    return $depth;
}

###############################################################################
# Method:    _get_err_stack_trace
# Objective: return depth of call stack
# Inputs:    None.
# Outputs:   depth of output (scalar)
# Affected Data: Error components in the object have been cleared.
###############################################################################

sub _get_err_stack_trace {
    my ($self)=@_;
    my ($trace);
    
    $trace=(defined $self->{stack_trace})
	  ? $self->{stack_trace}
	  : 1;
    
    return $trace;
}

###############################################################################
# Method:    _set_err
# Objective: Set error components in the object based on input errmsg.
# Inputs:    errmsg.
# OuTCaseuts:   None.
# Affected Data: Error components in the object have been set.
###############################################################################

sub _set_err {
    my ($self,%args)=@_;
    my ($stack_depth,
	$on_err,
	$stack_trace,
	$msg,
	@messages,
	@stack,
	);
    
    $self->{error} = 1;
    
    # parse input arguments    
    $on_err = (defined $args{on_err})
        ? $args{on_err} 
	: $self->{on_err};
	
    $on_err="die" if (not defined $on_err);
    
    $stack_depth = (defined $args{stack_depth})
        ? $args{stack_depth} 
	: $self->_get_err_stack_depth();
 	
    
    $stack_trace = (defined $args{stack_trace})
        ? $args{stack_trace} 
	: $self->_get_err_stack_trace();
	
    $stack_trace=0 if (not defined $stack_trace);
    
	
    $msg=(defined $args{msg})
	? $args{msg}
	: "NO ERROR MESSAGE";
	
    # retrieve error message 
    if (ref($msg) =~/NATE::Result/){
        push @messages,@{$@->{messages}};
    }
    elsif (ref($msg) eq "ARRAY"){
        push @messages,@$msg;
    }
    else {
        push @messages,$msg;
    }
    
    $messages[0]=~s/to \(eval\)//;
    $messages[0]=~s/callback$/check/;
    $messages[0]=~s/  / /g;
    
    # Check to see if we should include the call stack
    if ($stack_trace){
	if ($stack_depth==1){
	    my (undef,$file,$line,$sub)=caller($stack_depth);
	    push @stack,"at ".$file." in ".$sub."() line ".$line;
	}
	else {
	    my $cnt = 1;
	    push @stack,"-"x70, "Error Call Stack:", "-"x70;			
	    while ((my (undef,$file,$line,$sub)=caller($cnt-1))&&
		   ($cnt <= $stack_depth)){	
		push @stack,$cnt." ".$file."::".$sub.": ".$line;	 
		$cnt++;
	    }
	    push @stack,"-"x70,"\n";          
	}
    }
 
    # Remove ERROR header if it exists
    $messages[0]=~s/^ERROR: //i;
 
    if ($on_err=~/die/) {
      die "ERROR: ".$messages[0]."\n".join("\n",(@messages[1..scalar(@messages)],@stack));
    }
    else {
   
      #$self->_set_err("ERROR: ".@messages,@stack);
      $self->{errmsg} = ["ERROR: ".$messages[0],
			 @messages[1..scalar(@messages)],
			 @stack];
    }
    
    return 1;
}

################################################################
# Subroutine Name: _set_err_stack_depth()
# Objective:       set the depth of call stack
#                  
# Details:         See POD documentation at the end of this file
################################################################
sub _set_err_stack_depth {
    my ($self,$depth)=@_;
    
    $self->{stack_depth}=$depth;
    
    return 1;
}

################################################################
# Subroutine Name: _set_err_stack_trace()
# Objective:       turn stack trace on or off 
#                  
# Details:         See POD documentation at the end of this file
################################################################
sub _set_err_stack_trace {
    my ($self,$mode)=@_;
    
    $self->{stack_trace}=$mode;
    
    return 1;
}

1;

=pod

=head1 Package Name

Error.pm

=head1 Author

Gary Knopp

=head1 Description

Error provides an API interface for performing basic error handling.  Use @ISA to 
inherit these methods into another package.  Use the private methods 
_clear_err() and _set_err($error_msg) in your package to remove or set error messages.

=over 4

=item Examples

use vars qw( @EXPORT @ISA);

@ISA = qw(Error);

=back

=head1 Public Methods

=head2 B<errmsg()>

This method allows you to get the text of the error message if an error
occurred on the previous method call.

=over 4

=item Inputs

None

=item Outputs

Returns the text of the last error message set.  If no error message is
set, an empty string will be returned.

=item Examples

 # Check for errors.
 if ($myObj->error) {
    $errmsg = $myObj->errmsg;
    print "An error occurred: $errmsg\n";
 }

=back

=head2 B<error()>

This method allows you to determine if any errors occured in the
previous method call.  Note that the error flag is reset at the
beginning of each method, so it should be checked after B<every>
major method call.  If you do not check for errors, your calling
program may fail at some point and you will not know why.

=over 4 

=item Inputs

None.

=item Outputs

Returns 1 if an error occured in the previous call, 0 otherwise.

=item Examples

 # Determine if error occurred...
 if ($myObj->error) {
    print "Error encountered in previous method call\n";
 }

=back

=cut
