# $Id$
## @summary Common library to collect filer statistics
## @author bellizzi@netapp.com,gvg@netapp.com
## @pod here

=head1 NAME

Utils::FilerStats - Common library to handle filer statistics

=head1 DESCRIPTION

This package defines routines which attempt to get and parse
filer output using the "say" command in Tharn.

All the subroutines here expect a connection object as their
argument.  If the connection is invalid (causing the say
command to fail) or if any other errors occur while attempting
to get the requested information, these subroutines will return 0.

=head1 SYNOPSIS

    use Utils::FilerStats;

    $stats   = getNfsStats              ('conn' => $conn, timeout => 300);
    $stats   = getSysStats              ('conn' => $filerConObj, 'duration' => 10);
    $stats   = getCifsStats             ('conn' => $filerConObj, 'mode' => 'priv');
    $stats   = getvolStats              ('conn' => $filerConObj);
    $seconds = getUptime                ('conn' => $filerConObj);
    $ret     = checkStatisticThresholds ('conn' => $filerConObj, 'thresholds' => \%thresholds);

=head1 ROUTINES

=cut

package Utils::FilerStats;

use strict;
use Tharn qw(say $Log);
use vars qw (@ISA @EXPORT $VERSION $DEBUG);
use Module::Common qw(validate_mixed);
use Options;
use Params::Validate qw(validate UNDEF SCALAR BOOLEAN OBJECT HASHREF ARRAYREF);
use Volume qw( vol_disk_free DF_SPACE vol_list );
use Utils::Units qw(:constants);
use Dense::Utils qw(is_dense_enabled get_vol_savings);
use SayRetry qw(say_retry);
use Data::Dumper;
use filer qw(filer_vfiler_command);
use CIFS qw(get_cifs_limits);
use POSIX qw(strftime floor);
use Server;
my $Server;

$VERSION = 1.00;
@ISA     = qw (Exporter);
@EXPORT  = qw (
  getCifsStats
  getSysStats
  getNfsStats
  getLockStats
  getVolStats
  getUptime
  checkStatisticThresholds
  getNtapNbladeStats
  getNgshStats
  create_empty_data_set
  write_stats_data
  parse_sysstat_output
  getClusteredStats
  get_thresholds
  perct_or_average_counter_calculation
  rate_or_delta_counter_calculation
  raw_counter_calculation
  );

my $Options = new Options;
$Options->parameters(
    definitions => [
        {name        => "VOL_LIST",
         description => "Get stats on only these volumes",
         spec        => "s@",
         default     => undef,
        },
        {name        => "VOL_REGEX",
         description => "Get stats on only these volumes",
         spec        => "s",
         default     => '.*',
        },
        {name        => "PREFIX",
         description => "Prepend subsystem prefixes to headers",
         spec        => "s",
         default     => '1',
        },
        {name        => 'SYSSTAT_DURATION',
         spec        => "s",
         default     => '30',
         description => 'How long sysstat should gather statistics',
        },
        {name        => "RETRY",
         description => "Number of retry attempts to collect stats",
         spec        => "s",
         default     => '5',
        },
        {name        => "RETRY_INTERVAL",
         description => "The seconds between retry attempts to collect stats",
         spec        => "s",
         default     => '3',
        },
        {name        => 'MAX_CPU',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'MAX_CIFS_OPS',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'MAX_CIFS_SESSIONS',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'AVG_CLUSTERED_CPU',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'MAX_CLUSTERED_CPU',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'MAX_CLUSTERED_CIFS_OPS',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'SUM_CLUSTERED_CIFS_OPS',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'MAX_CLUSTERED_CIFS_SESSIONS',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'SUM_CLUSTERED_CIFS_SESSIONS',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'MAX_CLUSTERED_CIFS_USERS',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'SUM_CLUSTERED_CIFS_USERS',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'AVG_CLUSTERED_DISKUTIL',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name        => 'MAX_CLUSTERED_DISKUTIL',
         spec        => "s",
         default     => undef,
         description => 'Statistic threshold',
        },
        {name    => 'CLUSTERED_STATS_NODE_INDEX',
         spec    => "s",
         default => undef,
         description =>
           'Comma separated list of integers that correspond to the array index of NODES',
        },
      ]
      );

=head2 getCifsStats

C<getCifsStats> returns a hashref of the 'cifs stat' output.

$stats = getCifsStats('conn' => $filerConObj, 'mode' => 'priv');

=over

=item conn

(Required) Conn object of the filer required to be passed.

=item mode

(Required) Which CLI mode to use.

=item prefix	

(Optional). Default is 1. Toggle sub system prefixes in hash keys

=item timeout

(Optional). Default is 300 secs. The time 'nfsstat -d' takes to return.

=back

=cut

sub getCifsStats {

    $Log->trace("Entering getCifsStats at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) =
      validate(@_,
               {conn   => {type => OBJECT | UNDEF, optional => 1},
                mode   => {type => SCALAR,         optional => 1},
                node   => {type => SCALAR,         optional => 1},
                prefix => {type     => BOOLEAN,
                           default  => $options->{'PREFIX'},
                           optional => 1
                  },
                timeout => {type => SCALAR, default => 300, optional => 1},
               }
               );

    my %stats = ();
    my ($key, $value);
    if ($opts{'conn'}) {
        if ($opts{'mode'} ne 'priv') {

            #----------------------
            # Get 7mode CIFS stats
            #----------------------
            my $cmd = filer_vfiler_command($opts{'conn'}, 'cifs stat');
            my $output;
            eval {
                (undef, undef, $output) =
                  say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                      -noverbose);
            };

            $Log->debug("Command output: \n$output");

            if ($@) {
                die "Error executing $cmd\n";
            }

            # Search the commands output
            foreach my $line (split(/\n/, $output)) {
                $line =~ s/^\s+(.+)$/$1/;
                if ($line =~ /^([^=]+)\b\s*=\s*(\d+),?\s*(.*)$/) {
                    ($key, $value, $line) = ($1, $2, $3);
                    $key =~ s/\s+/_/g;
                    if ($opts{prefix}) {
                        $stats{"cifs:" . $key} = $value;
                    } else {
                        $stats{$key} = $value;
                    }
                    while ($line =~ /^([^=]+)\b\s*=\s*(\d+),?\s*(.*)$/) {
                        ($key, $value, $line) = ($1, $2, $3);
                        $key =~ s/\s+/_/g;
                        if ($opts{prefix}) {
                            $stats{"cifs:" . $key} = $value;
                        } else {
                            $stats{$key} = $value;
                        }
                    }
                } elsif ($line =~ /^(.+)\s+(\d+)\s+(\d+)\%/) {
                    $key   = $1;
                    $value = $2;
                    $key =~ s/^(.+\w+)\s+$/$1/;
                    $key =~ s/\s+/_/;
                    if ($opts{prefix}) {
                        $stats{"cifs:" . $key} = $value;
                    } else {
                        $stats{$key} = $value;
                    }
                } elsif ($line =~ /^(.+)\s+(\d+)/) {
                    $key   = $1;
                    $value = $2;
                    $key =~ s/^(.+\w+)\s+$/$1/;
                    $key =~ s/\s+/_/g;
                    if ($opts{prefix}) {
                        $stats{"cifs:" . $key} = $value;
                    } else {
                        $stats{$key} = $value;
                    }
                } ## end elsif ($line =~ /^(.+)\s+(\d+)/)
            } ## end foreach my $line (split(/\n/...

            $cmd = filer_vfiler_command($opts{'conn'}, 'cifs sessions -t');
            while (1) {
                eval {
                    (undef, undef, $output) =
                      say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                          -noverbose);
                };

                $Log->debug("Command output: \n$output");

                if ($@) {
                    sleep 10;
                } else {
                    last;
                }
            } ## end while (1)
            die "Error executing $cmd\n" if $@;

            # Search the commands output
            foreach my $line (split(/\n/, $output)) {
                if ($line =~ /^(.+):\s+(\d+)/) {
                    $key   = $1;
                    $value = $2;
                    $key =~ s/\s+/_/g;
                    if ($opts{prefix}) {
                        $stats{"cifs:" . $key} = $value;
                    } else {
                        $stats{$key} = $value;
                    }
                } ## end if ($line =~ /^(.+):\s+(\d+)/)
            } ## end foreach my $line (split(/\n/...

        } elsif ($opts{'mode'} eq 'priv') {

            #-----------------------------------------------------------------------
            # Get 7mode CIFS stats using privileged mode. This method is deprecated
            #-----------------------------------------------------------------------
            my $cmd = filer_vfiler_command($opts{'conn'},
                                        'priv set diag; stats show cifs_stats');
            my $output;
            eval {
                (undef, undef, $output) =
                  say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                      -noverbose);
            };

            $Log->debug("Command output: \n$output");

            if ($@) {
                die "Error executing $cmd\n";
            }

            # Search the commands output
            foreach my $line (split(/\n/, $output)) {
                if ($line =~ /^cifs_stats:vfiler0:(\S+)(:| :)(.+)$/) {
                    if ($opts{prefix}) {
                        $stats{"cifs:" . $1} = $3;
                    } else {
                        $stats{$1} = $3;
                    }
                }
            }

            $cmd = filer_vfiler_command($opts{'conn'}, 'cifs sessions -t');
            eval {
                (undef, undef, $output) =
                  say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                      -noverbose);
            };

            $Log->debug("Command output: \n$output");

            if ($@) {
                die "Error executing $cmd\n";
            }

            # Search the commands output
            foreach my $line (split(/\n/, $output)) {

                if ($line =~ /^CIFS\s+([^;]+):\s+(\d+)/) {
                    $key   = $1;
                    $value = $2;
                    $key =~ s/\s+/_/g;
                    if ($opts{prefix}) {
                        $stats{"cifs:" . $key} = $value;
                    } else {
                        $stats{$key} = $value;
                    }
                } ## end if ($line =~ /^CIFS\s+([^;]+):\s+(\d+)/)
            } ## end foreach my $line (split(/\n/...
        } ## end elsif ($opts{'mode'} eq 'priv')
    } else {
        $Server = new Server() if (!$Server);

        #---------------------------------
        # Get 10mode CIFS stats from ngsh
        #---------------------------------
        my $ngshStats;
        eval {
            $ngshStats = getNgshStats('srv_cmd'  => 'statistics_show',
                                      'srv_opts' => {'category' => 'cifs',
                                                     'node'     => $opts{'node'}
                                        }
                                        );
        };
        if ($@) { die $@; }

        foreach my $key (keys %{$ngshStats}) {
            $key =~ s/\s+/_/g;
            if ($opts{prefix}) {
                $stats{"cifs:" . $key} = $ngshStats->{$key};
            } else {
                $stats{$key} = $ngshStats->{$key};
            }
        }

        eval {
            $ngshStats = getNgshStats('srv_cmd'  => 'statistics_show',
                                      'srv_opts' => {
                                                'category' => 'cifs-currvalues',
                                                'node'     => $opts{'node'}
                                        }
                                        );
        };
        if ($@) { die $@; }

        foreach my $key (keys %{$ngshStats}) {
            $key =~ s/\s+/_/g;
            if ($opts{prefix}) {
                $stats{"cifs:" . $key} = $ngshStats->{$key};
            } else {
                $stats{$key} = $ngshStats->{$key};
            }
        }

    } ## end else [ if ($opts{'conn'})

    #----------------------------------------------------------------------
    # Create a common Sessions key so we don't have to figure this out in
    # every script
    #----------------------------------------------------------------------
    if ($opts{'prefix'}) {
        $stats{'cifs:Sessions'} = 0;
    } else {
        $stats{'Sessions'} = 0;
    }

    if ($stats{'Total_CIFS_sessions'}) {
        $stats{'Sessions'} = $stats{'Total_CIFS_sessions'};
    } elsif ($stats{'cifs:Total_CIFS_sessions'}) {
        $stats{'cifs:Sessions'} = $stats{'cifs:Total_CIFS_sessions'};
    } elsif ($stats{'CIFS_sessions'}) {
        $stats{'Sessions'} = $stats{'CIFS_sessions'};
    } elsif ($stats{'cifs:CIFS_sessions'}) {
        $stats{'cifs:Sessions'} = $stats{'cifs:CIFS_sessions'};
    } elsif ($stats{'CurrConnections'}) {
        $stats{'Sessions'} = $stats{'CurrConnections'};
    } elsif ($stats{'cifs:CurrConnections'}) {
        $stats{'cifs:Sessions'} = $stats{'cifs:CurrConnections'};
    }

    $Log->debug("stats: " . Dumper(\%stats));
    $Log->trace("Exiting getCifsStats at " . localtime);

    return \%stats;
} ## end sub getCifsStats

=head2 getSysStats

C<getSysStats> returns a hashref of the 'sysstat' output.

$stats = getSysStats('conn' => $filerConObj, 'duration' => 10);

=over

=item conn

(Required) Conn object of the filer required to be passed.

=item duration

(Required) The lenth of time to gather the stats.

=item prefix	

(Optional). Default is 1. Toggle sub system prefixes in hash keys

=back

=cut

sub getSysStats {

    $Log->trace("Entering getSysStats at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) =
      validate(@_,
               {conn     => {type => OBJECT | UNDEF, optional => 1},
                node     => {type => SCALAR,         optional => 1},
                duration => {type     => BOOLEAN,
                             default  => $options->{'SYSSTAT_DURATION'},
                             optional => 1
                  },
                prefix => {type     => BOOLEAN,
                           default  => $options->{'PREFIX'},
                           optional => 1
                  },
               }
               );

    my %stats = ();
    my $sysStats;
    if ($opts{'conn'}) {

        #-----------------------------------
        # Get the node's 7mode system stats
        #-----------------------------------
        my $cmd = filer_vfiler_command($opts{'conn'},
                                       'sysstat -x -c 1 ' . $opts{'duration'});
        my $output;

        while (1) {
            eval {
                (undef, undef, $output) =
                  say($opts{'conn'}, $cmd, -timeout, $opts{'duration'} + 60,
                      -noverbose);
            };

            $Log->debug("Command output: \n$output");

            if ($@) {
                sleep 10;
            } else {
                last;
            }
        } ## end while (1)
        die "Error executing $cmd\n" if $@;
        $sysStats = parse_sysstat_output($output);

        foreach my $key (keys(%{$sysStats})) {
            $key =~ s/\s+/_/g;
            if ($opts{prefix}) {
                $stats{"sys:" . $key} = $sysStats->{$key};

            } else {
                $stats{$key} = $sysStats->{$key};
            }
        }

    } else {
        $Server = new Server() if (!$Server);

        $Log->debug("Getting 10mode system statistics\n");

        #-----------------------------------------------
        # Get the node's 10mode system stats using ngsh
        #-----------------------------------------------
        my $ngshStats;
        eval {
            $ngshStats = getNgshStats('srv_cmd'  => 'statistics_show',
                                      'srv_opts' => {'category' => 'node',
                                                     'node'     => $opts{'node'}
                                        }
                                        );
        };

        if ($@) { die $@; }

        foreach my $key (keys(%{$ngshStats})) {
            $key =~ s/\s+/_/g;
            if ($opts{prefix}) {
                $stats{"sys:" . $key} = $ngshStats->{$key};

            } else {
                $stats{$key} = $ngshStats->{$key};
            }
        }

        $Log->comment("stats: " . Dumper(\%stats));
    } ## end else [ if ($opts{'conn'})

    if ($stats{'CIFS'}) {
        $stats{'CIFS_Operations'} = $stats{'CIFS'};
    } elsif ($stats{'sys:CIFS'}) {
        $stats{'CIFS_Operations'} = $stats{'sys:CIFS'};
    }

    if ($stats{'CPU_Busy'}) {
        $stats{'CPU'} = $stats{'CPU_Busy'};
    }

    $Log->debug("stats: " . Dumper(\%stats));
    $Log->trace("Exiting getSysStats at " . localtime);

    return \%stats;

} ## end sub getSysStats

=head2 getNfsStats

C<getNfsStats> returns a hashref of the 'nfsstat -d' output.

=over

=item conn

(Required) Conn object of the filer required to be passed.

=item access_cache

(Optional). Default is 1. access_cache stats are included.

=item mount

(Optional). Default is 1. mount stats are included.

=item timeout

(Optional). Default is 300 secs. The time 'nfsstat -d' takes to return.

=item prefix	

(Optional). Default is 1. Toggle sub system prefixes in hash keys

=back

=cut

sub getNfsStats {

    $Log->trace("Entering getNfsStats at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my ($conn, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT | SCALAR, optional => 1},],
                     {access_cache  => {type => BOOLEAN, default  => 1},
                      node          => {type => SCALAR,  optional => 1},
                      version       => {type => SCALAR,  optional => 1},
                      mount         => {type => BOOLEAN, default  => 1},
                      regex         => {type => SCALAR,  optional => 1},
                      search        => {type => SCALAR,  optional => 1},
                      timeout       => {type => SCALAR,  default  => 300},
                      prefix_string => {type => SCALAR,  optional => 1},
                      prefix        => {
                                 type     => BOOLEAN,
                                 default  => $options->{'PREFIX'},
                                 optional => 1
                        },
                     }
                     );

    my %stats = ();
    if ($conn) {
        my $cmd = filer_vfiler_command($conn, 'nfsstat -d');
        my $output;
        eval {
            (undef, undef, $output) =
              say($conn, $cmd, -timeout, $opts{timeout}, -noverbose);
        };

        $Log->debug("Command output: \n$output");

        if ($@) {
            die "Error executing $cmd\n";
        }

        foreach my $line (split(/\n/, $output)) {

            # get mount or access_cache stats which are similar to parse
            if (   $line =~ / mount | unmount |^mount|^unmount/
                || $line =~ /access cache/) {

                if (  !$opts{mount}
                    && $line =~ / mount | unmount |^mount|^unmount/) {
                    next;
                }

                if (!$opts{access_cache} && $line =~ /access cache/) {
                    next;
                }

                # Split keys and values by "="
                my ($allkeys, $values) = split("=", $line);

                # Seperate values
                my @vals = split(",", $values);

                # Seperate keys
                my ($prefix, $keys) = split('\(', $allkeys);
                my @ks = split(",", $keys);
                for (1 .. 10) {    # Have max of 10 for now.
                    my $key   = $prefix . $ks[$_];
                    my $value = $vals[$_];

                    if (!$key || !$value) {
                        last;
                    }

                    # Remove open/close paranthesis
                    $key   =~ s%\(%%g;
                    $key   =~ s%\)%%g;
                    $value =~ s%\(%%g;
                    $value =~ s%\)%%g;

                    $key   =~ s%\s+% %g;
                    $key   =~ s%\s$%%g;
                    $key   =~ s% %_%g;
                    $value =~ s% %%g;

                    if ($opts{'version'}) {

                        #-------------------------------
                        # Get the version specific keys
                        #-------------------------------
                        my $regex = 'v' . $opts{'version'} . '_';
                        if ($key =~ /$regex/) {
                            if ($opts{prefix}) {
                                $stats{"nfs$opts{'version'}:" . $key} = $value;
                            } else {
                                $stats{$key} = $value;
                            }
                        }
                    } else {

                        #--------------------------------------------------
                        # Fallback to the default behavior unless regex is
                        # used to limit the keys to a smaller subset
                        #--------------------------------------------------
                        if ($opts{'regex'} && $key !~ /$opts{'regex'}/) {
                            next;
                        }

                        if ($opts{prefix}) {
                            if ($opts{'prefix_string'}) {
                                $stats{$opts{'prefix_string'} . ':' . $key} =
                                  $value;
                            } else {
                                $stats{"nfs:" . $key} = $value;
                            }
                        } else {
                            $stats{$key} = $value;
                        }
                    } ## end else [ if ($opts{'version'})

                } ## end for (1 .. 10)
            } ## end if ($line =~ / mount | unmount |^mount|^unmount/...

        } ## end foreach my $line (split(/\n/...
    } else {
        $Server = new Server() if (!$Server);

        if ($opts{'version'}) {
            my $ngshStats;
            eval {
                $ngshStats =
                  getNgshStats('srv_cmd'  => 'statistics_show',
                               'srv_opts' => {
                                         'category' => 'nfs' . $opts{'version'},
                                         'node'     => $opts{'node'}
                                 }
                                 );
            };

            if ($@) { die $@; }

            foreach my $key (keys(%{$ngshStats})) {
                $key =~ s/\s+/_/g;
                if ($opts{prefix}) {
                    $stats{"nfs$opts{'version'}:" . $key} = $ngshStats->{$key};

                } else {
                    $stats{$key} = $ngshStats->{$key};
                }
            }

            my $ntapNbladeStats;
            eval {
                $ntapNbladeStats =
                  getNtapNbladeStats('node'   => $opts{'node'},
                                     'search' => $opts{'search'});
            };

            if ($@) { die $@; }

            $opts{'search'} =~ s/\./\\\./g;
            foreach my $key (keys(%{$ntapNbladeStats})) {
                $key =~ s/\s+/_/g;
                if ($key =~ /$opts{'search'}\.(.+)/) {
                    if ($opts{prefix}) {
                        $stats{"nfs$opts{'version'}:" . $1} =
                          $ntapNbladeStats->{$key};
                    } else {
                        $stats{$1} = $ntapNbladeStats->{$key};
                    }
                }
            } ## end foreach my $key (keys(%{$ntapNbladeStats...
        } else {
            my $ntapNbladeStats;
            eval {
                $ntapNbladeStats =
                  getNtapNbladeStats('node'   => $opts{'node'},
                                     'search' => $opts{'search'});
            };

            if ($@) { die $@; }

            $opts{'search'} =~ s/\./\\\./g;
            foreach my $key (keys(%{$ntapNbladeStats})) {
                $key =~ s/\s+/_/g;
                if ($key =~ /$opts{'search'}\.(.+)/) {
                    if ($opts{prefix}) {
                        $stats{"$opts{'prefix_string'}:" . $1} =
                          $ntapNbladeStats->{$key};
                    } else {
                        $stats{$1} = $ntapNbladeStats->{$key};
                    }
                }
            } ## end foreach my $key (keys(%{$ntapNbladeStats...
        } ## end else [ if ($opts{'version'})
    } ## end else [ if ($conn)

    $Log->debug("stats: " . Dumper(\%stats));
    $Log->trace("Exiting getNfsStats at " . localtime);

    return \%stats;
} ## end sub getNfsStats

=head2 getLockStats

C<getLockStats> returns a hashref of the 'lock status -c' on diag mode output.

=over

=item conn

(Required) Conn object of the filer required to be passed.

=item timeout

(Optional). Default is 300 secs. The time, the command takes to return.

=item prefix	

(Optional). Default is 1. Toggle sub system prefixes in hash keys

=back

=cut

sub getLockStats {

    $Log->trace("Entering getLockStats at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my ($conn, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {timeout => {type => SCALAR, default => 300},
                      retry =>
                        {type => SCALAR, default => $options->{'RETRY'},},
                      retry_interval => {
                                        type    => SCALAR,
                                        default => $options->{'RETRY_INTERVAL'},
                        },
                      prefix => {type     => BOOLEAN,
                                 default  => $options->{'PREFIX'},
                                 optional => 1
                        },
                     }
                     );

    my $cmd = filer_vfiler_command($conn, 'priv set -q diag; lock status -c');
    my (undef, undef, $output) =
      say_retry($conn,       $cmd,
                -retry,      $opts{retry},
                -iter_sleep, $opts{retry_interval},
                -timeout,    $opts{timeout},
                -noverbose,  -eval_error
                );
    my %lockstats;

    foreach my $line (split(/\n/, $output)) {
        foreach my $stat (split /;/, $line) {
            my @words = split(" ", $stat);
            my $value = $words[-1];
            delete $words[-1];
            my $key = join("_", @words);

            if ($opts{prefix}) {
                $lockstats{"lock:" . $key} = $value;
            } else {
                $lockstats{$key} = $value;
            }
        } ## end foreach my $stat (split /;/...
    } ## end foreach my $line (split(/\n/...

    $Log->trace("Exiting getLockStats at " . localtime);

    return \%lockstats;
} ## end sub getLockStats

=head2 getVolStats

C<getVolStats> returns a hashref of the 'df -g' output for all volumes.

$stats = getVolStats('conn' => $filerConObj);

=over

=item conn

(Required) Conn object of the filer required to be passed.

=item vol_list

(Optional). Default is VOL_LIST. An array of volumes to gets stats for.

=item vol_regex

(Optional). Default is VOL_REGEX. A regex to determine what vols have stats reported

=item prefix	

(Optional). Default is 1. Toggle sub system prefixes in hash keys

=item timeout

(Optional). Default is 300 secs. The time 'nfsstat -d' takes to return.

=back

=cut

sub getVolStats {

    $Log->trace("Entering getVolStats at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) =
      validate(@_,
               {conn     => {type => OBJECT | UNDEF, optional => 1},
                node     => {type => SCALAR,         optional => 1},
                vol_list => {type     => ARRAYREF,
                             default  => $options->{'VOL_LIST'},
                             optional => 1
                  },
                vol_regex => {type     => SCALAR,
                              default  => $options->{'VOL_REGEX'},
                              optional => 1
                  },
                prefix => {type     => BOOLEAN,
                           default  => $options->{'PREFIX'},
                           optional => 1
                  },
                timeout => {type => SCALAR, default => 300, optional => 1},
               }
               );
    my @volList;
    my %volStats;

    if (@{$opts{'vol_list'}}) {
        @volList = @{$opts{'vol_list'}};
    }

    if ($opts{'conn'}) {
        my %volSavings;
        eval { %volSavings = get_vol_savings('host' => $opts{'conn'}); };
        $Log->debug("volSavings: " . Dumper(\%volSavings));

        # Get the filer's volume stats
        my $output;
        my $cmd = filer_vfiler_command($opts{'conn'}, 'df -g');
        eval {
            (undef, undef, $output) = say($opts{'conn'}, $cmd, -timeout, 300);
        };
        $Log->debug("Command output: \n$output");

        if ($@) {
            die "Error executing $cmd\n";
        }

        # Search the commands output
        foreach my $line (split(/\n/, $output)) {

            #Filesystem               total       used      avail capacity  Mounted on
            #/vol/vol0/                43GB       18GB       24GB      42%  /vol/vol0/
            #/vol/san_io_fcp_raw_kc61b11/       24GB       12GB       11GB      50%  /vol/san_io_fcp_raw_kc61b11/
            if ($line =~
                /^\/vol\/([^\/]+)\/\s+(\d+)GB\s+(\d+)GB\s+(\d+)GB\s+(\S+)/) {
                my $vol      = $1;
                my $total    = $2;
                my $used     = $3;
                my $avail    = $4;
                my $capacity = $5;
                if (@volList && !grep(/^$vol$/, @volList)) { next; }
                if ($vol !~ /$opts{'vol_regex'}/) { next; }
                $volStats{$vol . '-total'}             = $total;
                $volStats{$vol . '-used'}              = $used;
                $volStats{$vol . '-avail'}             = $avail;
                $volStats{$vol . '-capacity'}          = $capacity;
                $volStats{$vol . '-sis_used'}          = 0;
                $volStats{$vol . '-sis_saved'}         = 0;
                $volStats{$vol . '-sis_percent_saved'} = '0%';

                if ($volSavings{'fs'}->{"/vol/$vol/"}->{'used'}) {
                    $volStats{$vol . '-sis_used'} =
                      $volSavings{'fs'}->{"/vol/$vol/"}->{'used'};
                }
                if ($volSavings{'fs'}->{"/vol/$vol/"}->{'saved'}) {
                    $volStats{$vol . '-sis_saved'} =
                      $volSavings{'fs'}->{"/vol/$vol/"}->{'saved'};
                }
                if ($volSavings{'fs'}->{"/vol/$vol/"}->{'%saved'}) {
                    $volStats{$vol . '-sis_percent_saved'} =
                      $volSavings{'fs'}->{"/vol/$vol/"}->{'%saved'};
                }
            } ## end if ($line =~ ...
        } ## end foreach my $line (split(/\n/...
    } else {
        $Server = new Server() if (!$Server);

        my ($volInfo, %args, $output);
        if ($opts{'node'}) {
            $args{'node'} = $opts{'node'};
        }
        $args{'output'}     = \$output;
        $args{'-Noverbose'} = 1;

        eval { $volInfo = $Server->volume_show(%args); };
        if ($@) {
            $Log->comment("$@ $output");
            die "Unable to get volume info";
        } else {
            $Log->debug($output);
        }

        foreach my $vol (@{$volInfo}) {
            if (@volList && !grep(/^$vol->{'volume'}$/, @volList)) { next; }
            if ($vol->{'volume'} !~ /$opts{'vol_regex'}/) { next; }
            my $id = $vol->{'vserver'} . '-' . $vol->{'volume'};
            $vol->{'total'}        =~ s/^-$//;
            $vol->{'used'}         =~ s/^-$//;
            $vol->{'percent-used'} =~ s/^-$//;
            $vol->{'available'}    =~ s/^-$//;
            $volStats{$id . '-total'}        = $vol->{'total'};
            $volStats{$id . '-used'}         = $vol->{'used'};
            $volStats{$id . '-percent-used'} = $vol->{'percent-used'};
            $volStats{$id . '-available'}    = $vol->{'available'};
        } ## end foreach my $vol (@{$volInfo...

    } ## end else [ if ($opts{'conn'})

    my %stats;
    foreach my $key (keys %volStats) {
        if ($opts{'prefix'}) {
            $stats{"vol:" . $key} = $volStats{$key};
        } else {
            $stats{$key} = $volStats{$key};
        }
    }

    $Log->debug("stats " . Dumper(\%volStats));
    $Log->trace("Exiting getVolStats at " . localtime);

    return \%stats;
} ## end sub getVolStats

=head2 getUptime

C<getUptime> returns the filers uptime in seconds

$stats = getUptime('conn' => $filerConObj);

=over

=item conn

(Required) Conn object of the filer required to be passed.

=back

=cut

sub getUptime {

    $Log->trace("Entering getUptime at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) =
      validate(@_,
               {conn => {type => OBJECT | UNDEF, optional => 1},
                node => {type => SCALAR,         optional => 1},
               }
               );

    my $uptime   = '';
    my $timeoday = '';
    my $output;
    if ($opts{'conn'}) {
        eval {
            (undef, undef, $uptime) =
              say($opts{'conn'}, 'uptime', -timeout, 300);
        };
    } else {
        $Server = new Server() if (!$Server);
        eval {
            $uptime =
              $Server->shell('-Server'    => $opts{'node'},
                             output       => \$output,
                             '-Noverbose' => 1,
                             cmd          => 'uptime'
                             );
        };

        if ($@) {
            $Log->comment($output);
            die "Error executing uptime\n";
        } else {
            $Log->debug($output);
        }
    } ## end else [ if ($opts{'conn'})

    my $days       = 0;
    my $hours      = 0;
    my $minutes    = 0;
    my $seconds    = 0;
    my $uptime_sec = '';

    # parse values from filer uptime string returned
    if ($uptime =~
        /(\d+:\d+[apAP][mM])\s+up\s+([0-9dayhrminsecs,:\s]+)\s+\d+\s+\w+/) {
        chomp($uptime);
        $Log->debug("Command $opts{'node'} output: [$uptime]");
        $timeoday = $1;
        $uptime   = $2;
        $uptime =~ s/,//g;       # strip ','
        $uptime =~ s/\s+/ /g;    # converge multi-space to single space
        $timeoday = lc($timeoday);
        $Log->debug(
               "Parsed output: timeoday [$timeoday] uptime_string [$uptime]\n");

        # parse components
        if ($uptime =~ /(\d+)\s+day[s]?/) { $days    = $1; }
        if ($uptime =~ /(\d+)\s+hr[s]?/)  { $hours   = $1; }
        if ($uptime =~ /(\d+)\s+min[s]?/) { $minutes = $1; }
        if ($uptime =~ /(\d+)\s+sec[s]?/) { $seconds = $1; }
        if ($uptime =~ /(\d+):(\d+)/)     { $hours   = $1; $minutes = $2; }

        # calculate seconds
        $uptime_sec =
          (($days * 86400) + ($hours * 3600) + ($minutes * 60) + $seconds);
        $Log->debug(
            "Uptime values output:  d: $days h: $hours m: $minutes s: $seconds ==> uptime in seconds: $uptime_sec\n"
            );
    } else {
        $Log->warn("### Uptime values not parsed from [$uptime]\n");
    }

    $Log->trace("Exiting getUptime at " . localtime);
    return (wantarray
            ? ($uptime_sec, $days, $hours, $minutes, $seconds, $timeoday)
            : $uptime_sec
            );

} ## end sub getUptime

sub perct_or_average_counter_calculation {

    my (%opts) =
      validate(@_,
               {object_id            => {type => SCALAR,  optional => 0},
                base_counter         => {type => SCALAR,  optional => 0},
                dependent_counter    => {type => SCALAR,  optional => 0},
                base_counter_id      => {type => SCALAR,  optional => 0},
                dependent_counter_id => {type => SCALAR,  optional => 0},
                parser               => {type => HASHREF, optional => 0},
                timestamp            => {type => SCALAR,  optional => 0},
                counter_type         => {type => SCALAR,  optional => 0},
               }
               );
    my ($object_id, $base_counter, $dependent_counter, $base_counter_id,
        $dependent_counter_id, $parser, $timestamp, $counter_type);

    $object_id            = $opts{'object_id'};
    $base_counter         = $opts{'base_counter'};
    $dependent_counter    = $opts{'dependent_counter'};
    $base_counter_id      = $opts{'base_counter_id'};
    $dependent_counter_id = $opts{'dependent_counter_id'};
    $parser               = $opts{'parser'};
    $timestamp            = $opts{'timestamp'};
    $counter_type         = $opts{'counter_type'};

    my %counter_values = ();
    my @counter_array  = ();
    @counter_array = ($base_counter, $dependent_counter);
    my @counter_array_id = ();
    @counter_array_id = ($base_counter_id, $dependent_counter_id);

    for (my $counter_id = 0 ; $counter_id <= $#counter_array_id ; $counter_id++)
    {
        my $slice = eval {
            $parser->filter($object_id, $counter_array_id[$counter_id],
                            $timestamp, $timestamp + 600);
        };

        for (my $iteration = 0 ; $iteration < 2 ; $iteration++) {

            if ($slice->[$iteration]->[2] == 0) {
                $slice->[$iteration]->[2] = 0;
                return (0, 0);
            }
            push(@{$counter_values{$counter_array[$counter_id]}},
                 $slice->[$iteration]->[2]
                 );
        } ## end for (my $iteration = 0 ...

    } ## end for (my $counter_id = 0...

    my $perct_value1 =
      ($counter_values{$base_counter}[1] - $counter_values{$base_counter}[0]);
    my $perct_value2 =
      ($counter_values{$dependent_counter}[1] -
        $counter_values{$dependent_counter}[0]);
    if ($perct_value2 == 0) {
        return (0, 0);
    } elsif ($counter_type =~ /perct/) {
        my $perct_value = ($perct_value1 / $perct_value2) * 100;
        return ($perct_value, $perct_value2);

    } elsif ($counter_type =~ /average/) {
        my $average_value = ($perct_value1 / $perct_value2);
        return ($average_value, $perct_value2);

    }

} ## end sub perct_or_average_counter_calculation

sub rate_or_delta_counter_calculation {
    $Options->process();
    my $options = $Options->get_hash();

    my (%opts) =
      validate(@_,
               {object_id       => {type => SCALAR,  optional => 0},
                base_counter    => {type => SCALAR,  optional => 0},
                base_counter_id => {type => SCALAR,  optional => 0},
                parser          => {type => HASHREF, optional => 0},
                timestamp       => {type => SCALAR,  optional => 0},
                counter_type    => {type => SCALAR,  optional => 0},
               }
               );
    my ($object_id, $base_counter, $base_counter_id, $parser, $timestamp,
        $counter_type);

    $object_id       = $opts{'object_id'};
    $base_counter    = $opts{'base_counter'};
    $base_counter_id = $opts{'base_counter_id'};
    $parser          = $opts{'parser'};
    $timestamp       = $opts{'timestamp'};
    $counter_type    = $opts{'counter_type'};

    my %counter_values = ();
    my $slice          = eval {
        $parser->filter($object_id, $base_counter_id,
                        $timestamp, $timestamp + 600);
    };

    for (my $iteration = 0 ; $iteration < 2 ; $iteration++) {

        if ($slice->[$iteration]->[2] == 0) {
            $slice->[$iteration]->[2] = 0;
            return 0;

        }

        push(@{$counter_values{'timestamp'}}, $slice->[$iteration]->[0]);

        push(@{$counter_values{$base_counter}}, $slice->[$iteration]->[2]);

    } ## end for (my $iteration = 0 ...

    my $rate_counter_value =
      ($counter_values{$base_counter}[1] - $counter_values{$base_counter}[0]);
    my $timestamp_value =
      ($counter_values{'timestamp'}[1] - $counter_values{'timestamp'}[0]);

    if ($counter_type =~ m/rate/) {
        my $data_value;
        if ($timestamp_value == 0) {
            $data_value = 0;
        } else {
            $data_value = ($rate_counter_value / $timestamp_value);
        }
        return $data_value;
    } elsif ($counter_type =~ m/delta/) {
        my $data_value = $rate_counter_value;
        return $data_value;
    }

} ## end sub rate_or_delta_counter_calculation

sub raw_counter_calculation {

    my (%opts) =
      validate(@_,
               {object_id       => {type => SCALAR,  optional => 0},
                base_counter    => {type => SCALAR,  optional => 0},
                base_counter_id => {type => SCALAR,  optional => 0},
                parser          => {type => HASHREF, optional => 0},
                timestamp       => {type => SCALAR,  optional => 0},
                counter_type    => {type => SCALAR,  optional => 0},
               }
               );
    my ($object_id, $base_counter, $base_counter_id, $parser, $timestamp,
        $counter_type);

    $object_id       = $opts{'object_id'};
    $base_counter    = $opts{'base_counter'};
    $base_counter_id = $opts{'base_counter_id'};
    $parser          = $opts{'parser'};
    $timestamp       = $opts{'timestamp'};
    $counter_type    = $opts{'counter_type'};

    my %counter_values = ();
    my $slice          = eval {
        $parser->filter($object_id, $base_counter_id,
                        $timestamp, $timestamp + 600);
    };

    for (my $iteration = 0 ; $iteration < 1 ; $iteration++) {

        if ($slice->[$iteration]->[2] == 0) {
            $slice->[$iteration]->[2] = 0;
            return 0;

        }
        push(@{$counter_values{'timestamp'}},   $slice->[$iteration]->[0]);
        push(@{$counter_values{$base_counter}}, $slice->[$iteration]->[2]);

    } ## end for (my $iteration = 0 ...

    my $raw_counter_value = $counter_values{$base_counter}[0];
    return $raw_counter_value;

} ## end sub raw_counter_calculation

=head2 get_thresholds

C<get_thresholds> Retrun 1 if the current filer stats exceed user defined thresholds

$thresholds = get_thresholds();

=over

Creates a thresholds hashref in the form of 

$thresholds{'MAX_CPU' => '60',
            'MAX_CIFS_OPS => '1200',
	    'MAX_CIFS_SESSIONS' => '50000|70%');

=back

=cut

sub get_thresholds {

    $Log->trace("Entering get_thresholds at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my %thresholds = (
           'AVG_CLUSTERED_CPU'      => $options->{'AVG_CLUSTERED_CPU'},
           'AVG_CLUSTERED_DISKUTIL' => $options->{'AVG_CLUSTERED_DISKUTIL'},
           'MAX_CIFS_OPS'           => $options->{'MAX_CIFS_OPS'},
           'MAX_CIFS_SESSIONS'      => $options->{'MAX_CIFS_SESSIONS'},
           'MAX_CPU'                => $options->{'MAX_CPU'},
           'MAX_DISKUTIL'           => $options->{'MAX_MAX_DISKUTIL'},
           'MAX_CLUSTERED_CIFS_OPS' => $options->{'MAX_CLUSTERED_CIFS_OPS'},
           'MAX_CLUSTERED_CIFS_SESSIONS' =>
             $options->{'MAX_CLUSTERED_CIFS_SESSIONS'},
           'MAX_CLUSTERED_CIFS_USERS' => $options->{'MAX_CLUSTERED_CIFS_USERS'},
           'MAX_CLUSTERED_CPU'        => $options->{'MAX_CLUSTERED_CPU'},
           'MAX_CLUSTERED_DISKUTIL'   => $options->{'MAX_CLUSTERED_DISKUTIL'},
           'SUM_CLUSTERED_CIFS_USERS' => $options->{'SUM_CLUSTERED_CIFS_USERS'},
           'SUM_CLUSTERED_CIFS_OPS'   => $options->{'SUM_CLUSTERED_CIFS_OPS'},
           'SUM_CLUSTERED_CIFS_SESSIONS' =>
             $options->{'SUM_CLUSTERED_CIFS_SESSIONS'},
             );

    $Log->trace("Exiting get_thresholds at " . localtime);
    return \%thresholds;

} ## end sub get_thresholds

=head2 checkStatisticThresholds

C<getUptime> Retrun 1 if the current filer stats exceed user defined thresholds

$ret = checkStatisticThresholds('conn' => $filerConObj, 'thresholds' => \%thresholds);

=over

=item conn

(Required) Conn object of the filer required to be passed.

=item thresholds

(Required) A hashref of threshholds in the form of 

$thresholds{'MAX_CPU' => '60',
            'MAX_CIFS_OPS => '1200',
	    'MAX_CIFS_SESSIONS' => '50000|70%');

=back

=cut

sub checkStatisticThresholds {

    $Log->trace("Entering checkStatisticThresholds at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) =
      validate(@_,
               {conn       => {type => OBJECT | UNDEF,  optional => 1},
                node       => {type => SCALAR,          optional => 1},
                client     => {type => SCALAR | OBJECT, optional => 1},
                p4_dir     => {type => SCALAR,          optional => 1},
                thresholds => {type => HASHREF,         optional => 1},
               }
               );

    #-----------------------------------------
    # Check if thresholds have been specified
    #-----------------------------------------
    if (grep($_ > 0, values(%{$opts{'thresholds'}}))) {

        #--------------------------------------------
        # Get the current system and CIFS statistics
        #--------------------------------------------
        my (%cifsLimits, $cifsStats, $sysStats, $clusteredStats);
        if ($opts{'conn'}) {
            if (grep(/cifs/, keys(%{$opts{'thresholds'}}))) {
                %cifsLimits =
                  get_cifs_limits('filer'      => $opts{'conn'},
                                  'client'     => $opts{'client'},
                                  'p4_dir'     => $opts{'p4_dir'},
                                  'avg_groups' => 0
                                  );

                $cifsStats =
                  getCifsStats('conn'   => $opts{'conn'},
                               'mode'   => '',
                               'prefix' => 0
                               );
            } ## end if (grep(/cifs/, keys(...

            $sysStats =
              getSysStats('conn'     => $opts{'conn'},
                          'duration' => 60,
                          'prefix'   => 0
                          );
        } else {

            $clusteredStats = getClusteredStats('prefix' => 0);

        }

        #---------------------------------------------------------------
        # Use common var names due to different 10/7 mode counter names
        #---------------------------------------------------------------
        my $cifsSessions = $cifsStats->{'Sessions'};
        my $cifsOps      = $sysStats->{'CIFS_Operations'};
        my $sysCpu       = $sysStats->{'CPU'};
        my $sysDiskUtil  = $sysStats->{'DiskUtil'};

        my ($avgClusteredCpu,          $maxClusteredCpu,
            $maxClusteredCifsOps,      $sumClusteredCifsOps,
            $avgClusteredDiskUtil,     $maxClusteredDiskUtil,
            $maxClusteredCifsSessions, $sumClusteredCifsSessions,
            $maxClusteredCifsUsers,    $sumClusteredCifsUsers,
            $nodeIndex
            );
        if (!$opts{'conn'}) {
            $avgClusteredCpu      = $clusteredStats->{'avg-cpu-busy'};
            $maxClusteredCpu      = $clusteredStats->{'max-cpu-busy'};
            $maxClusteredCifsOps  = $clusteredStats->{'max-cifs-ops'};
            $sumClusteredCifsOps  = $clusteredStats->{'sum-cifs-ops'};
            $avgClusteredDiskUtil = $clusteredStats->{'avg-DiskUtil'};
            $maxClusteredDiskUtil = $clusteredStats->{'max-DiskUtil'};
            $maxClusteredCifsSessions =
              $clusteredStats->{'max-curr-connections'};
            $sumClusteredCifsSessions =
              $clusteredStats->{'sum-curr-connections'};
            $maxClusteredCifsUsers = $clusteredStats->{'max-curr-users'};
            $sumClusteredCifsUsers = $clusteredStats->{'sum-curr-users'};

            if ($options->{'CLUSTERED_STATS_NODE_INDEX'}) {
                $nodeIndex = $options->{'CLUSTERED_STATS_NODE_INDEX'};
            } else {
                $nodeIndex = 0;
                for (my $i = 1 ; $i < @{$options->{'NODES'}} ; $i++) {
                    $nodeIndex .= ",$i";
                }
            }
        } ## end if (!$opts{'conn'})

        #--------------------------
        # Check if CPU% >= MAX_CPU
        #--------------------------
        if ($opts{'thresholds'}{'MAX_CPU'}) {

            $Log->comment(  "Checking if the current CPU% ($sysCpu) >= "
                          . "MAX_CPU ($opts{'thresholds'}{'MAX_CPU'})");

            if ($sysCpu >= $opts{'thresholds'}{'MAX_CPU'}) {

                $Log->comment(  "The current CPU% ($sysCpu) >= MAX_CPU "
                              . " ($opts{'thresholds'}{'MAX_CPU'})");
                return 1;
            }
        } ## end if ($opts{'thresholds'...

        #------------------------------------
        # Check if CPU% >= AVG_CLUSTERED_CPU
        #------------------------------------
        if ($opts{'thresholds'}{'AVG_CLUSTERED_CPU'}) {

            $Log->comment(  "Checking if the current average clustered "
                          . "CPU%  ($avgClusteredCpu) of nodes ($nodeIndex) "
                          . ">=  AVG_CLUSTERED_CPU "
                          . "($opts{'thresholds'}{'AVG_CLUSTERED_CPU'})");

            if ($avgClusteredCpu >= $opts{'thresholds'}{'AVG_CLUSTERED_CPU'}) {

                $Log->comment(  "The current average clustered CPU% "
                              . "($avgClusteredCpu) of nodes ($nodeIndex) "
                              . ">= AVG_CLUSTERED_CPU "
                              . "($opts{'thresholds'}{'AVG_CLUSTERED_CPU'})");
                return 1;
            }
        } ## end if ($opts{'thresholds'...

        #------------------------------------
        # Check if CPU% >= MAX_CLUSTERED_CPU
        #------------------------------------
        if ($opts{'thresholds'}{'MAX_CLUSTERED_CPU'}) {

            $Log->comment(  "Checking if the current max clustered CPU% "
                          . "($maxClusteredCpu) of nodes ($nodeIndex) "
                          . ">=  MAX_CLUSTERED_CPU "
                          . "($opts{'thresholds'}{'MAX_CLUSTERED_CPU'})");

            if ($maxClusteredCpu >= $opts{'thresholds'}{'MAX_CLUSTERED_CPU'}) {

                $Log->comment(  "The current max clustered CPU% "
                              . "($maxClusteredCpu) of nodes "
                              . "($nodeIndex) >= MAX_CLUSTERED_CPU "
                              . "($opts{'thresholds'}{'MAX_CLUSTERED_CPU'})");
                return 1;
            }
        } ## end if ($opts{'thresholds'...

        #------------------------------------
        # Check if DISKUTIL% >= MAX_DISKUTIL
        #------------------------------------
        if ($opts{'thresholds'}{'MAX_DISKUTIL'}) {

            $Log->comment(  "Checking if the current DISKUTIL% "
                          . "($sysDiskUtil)>= MAX_DISKUTIL "
                          . "($opts{'thresholds'}{'MAX_DISKUTIL'})");

            if ($sysDiskUtil >= $opts{'thresholds'}{'MAX_DISKUTIL'}) {

                $Log->comment(  "The current DISKUTIL% ($sysDiskUtil) "
                              . ">= MAX_DISKUTIL "
                              . " ($opts{'thresholds'}{'MAX_DISKUTIL'})");
                return 1;
            }
        } ## end if ($opts{'thresholds'...

        #----------------------------------------------
        # Check if DISKUTIL% >= AVG_CLUSTERED_DISKUTIL
        #----------------------------------------------
        if ($opts{'thresholds'}{'AVG_CLUSTERED_DISKUTIL'}) {

            $Log->comment(  "Checking if the current averaged clustered "
                          . "DSIKUTIL% ($avgClusteredDiskUtil) of "
                          . "nodes ($nodeIndex) >= AVG_CLUSTERED_DISKUTIL "
                          . "($opts{'thresholds'}{'AVG_CLUSTERED_DISKUTIL'})");

            if ($avgClusteredDiskUtil >=
                $opts{'thresholds'}{'AVG_CLUSTERED_DISKUTIL'}) {

                $Log->comment("The current averaged clustered DSIKUTIL% "
                           . "($avgClusteredDiskUtil) of nodes "
                           . "($nodeIndex) => AVG_CLUSTERED_CPU "
                           . "($opts{'thresholds'}{'AVG_CLUSTERED_DISKUTIL'})");
                return 1;
            }
        } ## end if ($opts{'thresholds'...

        #----------------------------------------------
        # Check if DISKUTIL% >= MAX_CLUSTERED_DISKUTIL
        #----------------------------------------------
        if ($opts{'thresholds'}{'MAX_CLUSTERED_DISKUTIL'}) {

            $Log->comment(  "Checking if the current max clustered "
                          . "DSIKUTIL% ($maxClusteredDiskUtil) of "
                          . "nodes ($nodeIndex) >= MAX_CLUSTERED_DISKUTIL "
                          . "($opts{'thresholds'}{'MAX_CLUSTERED_DISKUTIL'})");

            if ($maxClusteredDiskUtil >=
                $opts{'thresholds'}{'MAX_CLUSTERED_DISKUTIL'}) {

                $Log->comment("The current max clustered DSIKUTIL% "
                           . "($maxClusteredDiskUtil) of nodes "
                           . "($nodeIndex) => MAX_CLUSTERED_CPU "
                           . "($opts{'thresholds'}{'MAX_CLUSTERED_DISKUTIL'})");
                return 1;
            }
        } ## end if ($opts{'thresholds'...

        #------------------------------------------
        # Check if CIFS Operations >= MAX_CIFS_OPS
        #------------------------------------------
        if ($opts{'thresholds'}{'MAX_CIFS_OPS'}) {

            $Log->comment(  "Checking if the current CIFS operations "
                          . "($cifsOps) >= MAX_CIFS_OPS "
                          . "($opts{'thresholds'}{'MAX_CIFS_OPS'}");

            if ($cifsOps >= $opts{'thresholds'}{'MAX_CIFS_OPS'}) {

                $Log->comment(  "The current CIFS operations ($cifsOps) "
                              . ">= MAX_CIFS_OPS threshold "
                              . "($opts{'thresholds'}{'MAX_CIFS_OPS'})");
                return 1;
            }
        } ## end if ($opts{'thresholds'...

        #----------------------------------------------------
        # Check if CIFS Operations >= MAX_CLUSTERED_CIFS_OPS
        #----------------------------------------------------
        if ($opts{'thresholds'}{'MAX_CLUSTERED_CIFS_OPS'}) {

            $Log->comment(  "Checking if the current max clustered "
                          . "CIFS operations ($maxClusteredCifsOps) of nodes "
                          . "($nodeIndex) >= MAX_CLUSTERED_CIFS_OPS "
                          . "($opts{'thresholds'}{'MAX_CLUSTERED_CIFS_OPS'})");

            if ($maxClusteredCifsOps >=
                $opts{'thresholds'}{'MAX_CLUSTERED_CIFS_OPS'}) {

                $Log->comment("The current max clustered CIFS operations "
                           . "($maxClusteredCifsOps) of nodes ($nodeIndex) "
                           . ">= MAX_CLUSTERED_CIFS_OPS "
                           . "($opts{'thresholds'}{'MAX_CLUSTERED_CIFS_OPS'})");
                return 1;
            }
        } ## end if ($opts{'thresholds'...

        #----------------------------------------------------
        # Check if CIFS Operations >= SUM_CLUSTERED_CIFS_OPS
        #----------------------------------------------------
        if ($opts{'thresholds'}{'SUM_CLUSTERED_CIFS_OPS'}) {

            $Log->comment(  "Checking if the current sum of clustered "
                          . "CIFS operations ($sumClusteredCifsOps) of nodes "
                          . "($nodeIndex) >= SUM_CLUSTERED_CIFS_OPS "
                          . "($opts{'thresholds'}{'SUM_CLUSTERED_CIFS_OPS'})");

            if ($sumClusteredCifsOps >=
                $opts{'thresholds'}{'SUM_CLUSTERED_CIFS_OPS'}) {

                $Log->comment("The current sum of clustered CIFS operations "
                           . "($sumClusteredCifsOps) of nodes ($nodeIndex) "
                           . ">= SUM_CLUSTERED_CIFS_OPS "
                           . "($opts{'thresholds'}{'SUM_CLUSTERED_CIFS_OPS'})");
                return 1;
            }
        } ## end if ($opts{'thresholds'...

        #---------------------------------------------
        # Check if CIFS sessions >= MAX_CIFS_SESSIONS
        #---------------------------------------------
        if (   $opts{'thresholds'}{'MAX_CIFS_SESSIONS'}
            && $opts{'thresholds'}{'MAX_CIFS_SESSIONS'} !~ /%/) {

            $Log->comment(  "Checking if the current CIFS sessions "
                          . "($cifsSessions) >= "
                          . "MAX_CIFS_SESSIONS "
                          . "($opts{'thresholds'}{'MAX_CIFS_SESSIONS'})");

            if ($cifsSessions > $opts{'thresholds'}{'MAX_CIFS_SESSIONS'}) {

                $Log->comment(  "The current CIFS sessions ($cifsSessions) "
                              . ">= MAX_CIFS_SESSIONS "
                              . "($opts{'thresholds'}{'MAX_CIFS_SESSIONS'})");
                return 1;

            }
        } ## end if ($opts{'thresholds'...

        #-------------------------------------------------------
        # Check if CIFS sessions >= MAX_CLUSTERED_CIFS_SESSIONS
        #-------------------------------------------------------
        if (   $opts{'thresholds'}{'MAX_CLUSTERED_CIFS_SESSIONS'}
            && $opts{'thresholds'}{'MAX_CLUSTERED_CIFS_SESSIONS'} !~ /%/) {

            $Log->comment(" Checking if the current max clustered "
                      . "CIFS sessions ($maxClusteredCifsSessions) of "
                      . "nodes ($nodeIndex) >= MAX_CLUSTERED_CIFS_SESSIONS "
                      . "($opts{'thresholds'}{'MAX_CLUSTERED_CIFS_SESSIONS'})");

            if ($maxClusteredCifsSessions >=
                $opts{'thresholds'}{'MAX_CLUSTERED_CIFS_SESSIONS'}) {

                $Log->comment("The current max clustered CIFS sessions "
                      . "($maxClusteredCifsSessions) of nodes "
                      . "($nodeIndex) >= MAX_CLUSTERED_CIFS_SESSIONS "
                      . "($opts{'thresholds'}{'MAX_CLUSTERED_CIFS_SESSIONS'})");
                return 1;

            } ## end if ($maxClusteredCifsSessions...
        } ## end if ($opts{'thresholds'...

        #-------------------------------------------------------
        # Check if CIFS sessions >= SUM_CLUSTERED_CIFS_SESSIONS
        #-------------------------------------------------------
        if (   $opts{'thresholds'}{'SUM_CLUSTERED_CIFS_SESSIONS'}
            && $opts{'thresholds'}{'SUM_CLUSTERED_CIFS_SESSIONS'} !~ /%/) {

            $Log->comment(
                    " Checking if the current sum of clustered CIFS sessions "
                      . "($sumClusteredCifsSessions) of nodes ($nodeIndex) "
                      . ">= SUM_CLUSTERED_CIFS_SESSIONS "
                      . "($opts{'thresholds'}{'SUM_CLUSTERED_CIFS_SESSIONS'})");

            if ($sumClusteredCifsSessions >=
                $opts{'thresholds'}{'SUM_CLUSTERED_CIFS_SESSIONS'}) {

                $Log->comment("The current sum of clustered CIFS sessions "
                      . "($sumClusteredCifsSessions) of nodes ($nodeIndex) "
                      . ">= SUM_CLUSTERED_CIFS_SESSIONS "
                      . "($opts{'thresholds'}{'SUM_CLUSTERED_CIFS_SESSIONS'})");
                return 1;

            } ## end if ($sumClusteredCifsSessions...
        } ## end if ($opts{'thresholds'...

        #----------------------------------------------------
        # Check if CIFS sessions >= MAX_CLUSTERED_CIFS_USERS
        #----------------------------------------------------
        if (   $opts{'thresholds'}{'MAX_CLUSTERED_CIFS_USERS'}
            && $opts{'thresholds'}{'MAX_CLUSTERED_CIFS_USERS'} !~ /%/) {

            $Log->comment( " Checking if the current max clustered CIFS users "
                         . "($maxClusteredCifsUsers) of nodes ($nodeIndex) "
                         . ">= MAX_CLUSTERED_CIFS_USERS "
                         . "($opts{'thresholds'}{'MAX_CLUSTERED_CIFS_USERS'})");

            if ($maxClusteredCifsSessions >=
                $opts{'thresholds'}{'MAX_CLUSTERED_CIFS_USERS'}) {

                $Log->comment("The current max clustered CIFS users "
                         . "($maxClusteredCifsUsers) of nodes ($nodeIndex) >= "
                         . "MAX_CLUSTERED_CIFS_USERS "
                         . "($opts{'thresholds'}{'MAX_CLUSTERED_CIFS_USERS'})");
                return 1;

            } ## end if ($maxClusteredCifsSessions...
        } ## end if ($opts{'thresholds'...

        #----------------------------------------------------
        # Check if CIFS sessions >= SUM_CLUSTERED_CIFS_USERS
        #----------------------------------------------------
        if (   $opts{'thresholds'}{'SUM_CLUSTERED_CIFS_USERS'}
            && $opts{'thresholds'}{'SUM_CLUSTERED_CIFS_USERS'} !~ /%/) {

            $Log->comment( " Checking if the current sum of clustered "
                         . "CIFS users ($sumClusteredCifsUsers) of nodes "
                         . "($nodeIndex) >= SUM_CLUSTERED_CIFS_USERS "
                         . "($opts{'thresholds'}{'SUM_CLUSTERED_CIFS_USERS'})");

            if ($sumClusteredCifsUsers >=
                $opts{'thresholds'}{'SUM_CLUSTERED_CIFS_USERS'}) {

                $Log->comment("The current clustered CIFS users "
                         . "($sumClusteredCifsUsers) of nodes ($nodeIndex) >= "
                         . "SUM_CLUSTERED_CIFS_USERS "
                         . "($opts{'thresholds'}{'SUM_CLUSTERED_CIFS_USERS'})");
                return 1;

            } ## end if ($sumClusteredCifsUsers...
        } ## end if ($opts{'thresholds'...

        if (   %cifsLimits
            && $opts{'thresholds'}{'MAX_CIFS_SESSIONS'}) {

            #----------------------------------------------------------------------
            # Check if CIFS sessions >= MAX_CIFS_SESSIONS (percentage of allowable
            # sessions)
            #----------------------------------------------------------------------
            if ($opts{'thresholds'}{'MAX_CIFS_SESSIONS'} =~ /%/) {

                my $currentPercentOfMax;

                if (   $cifsLimits{'maxconnections'} > 0
                    && $cifsSessions > 0) {

                    $currentPercentOfMax = floor(
                         ($cifsSessions / $cifsLimits{'maxconnections'}) * 100);

                    my $maxSessionsThresholdPercent =
                      $opts{'thresholds'}{'MAX_CIFS_SESSIONS'};
                    $maxSessionsThresholdPercent =~ s/%//g;

                    my $maxSessionsThresholdValue =
                      int(($maxSessionsThresholdPercent / 100) *
                          $cifsLimits{'maxconnections'});

                    $Log->comment(
                         " Checking if the current CIFS sessions $cifsSessions "
                           . "($currentPercentOfMax%) >= "
                           . "MAX_CIFS_SESSIONS "
                           . "($maxSessionsThresholdPercent%)");

                    if ($currentPercentOfMax >= $maxSessionsThresholdPercent) {

                        $Log->comment("CIFS sessions $cifsSessions "
                                . "($currentPercentOfMax%) >= "
                                . "MAX_CIFS_SESSIONS "
                                . "($opts{'thresholds'}{'MAX_CIFS_SESSIONS'})");
                        return 1;

                    }
                } ## end if ($cifsLimits{'maxconnections'...
            } ## end if ($opts{'thresholds'...

            #----------------------------------------------
            # Check if CIFS sessions >= CIFS session limit
            #----------------------------------------------
            if (%cifsLimits && !$opts{'thresholds'}{'MAX_CIFS_SESSIONS'}) {

                $Log->comment(
                        "Checking if the current CIFS sessions ($cifsSessions) "
                          . ">= platform CIFS session limit "
                          . "($cifsLimits{'maxconnections'})");

                if ($cifsSessions > $cifsLimits{'maxconnections'}) {

                    $Log->comment(  "The current CIFS sessions ($cifsSessions) "
                                  . ">= platform CIFS session limit "
                                  . "($cifsLimits{'maxconnections'})");
                    return 1;

                }
            } ## end if (%cifsLimits && !$opts...
        } ## end if (%cifsLimits && $opts...
    } ## end if (grep($_ > 0, values...
    $Log->trace("Exiting checkStatisticThresholds at " . localtime);

    return 0;
} ## end sub checkStatisticThresholds

=head2 getNgshStats

C<getNgshStats> This routine obtains statistics for the specified protocol/node 
and returns them in a hash ref.

$ret = getNgshStats('srv_cmd' => $string, 'srv_opts' => \%hashref);

=over

=item srv_cmd

(Required) String containing the server command to call

=item srv_opts

(Required) Reference to hash containing statistics for the specified node if 
successfully obtained 0 if any error occurs

=back

=cut

sub getNgshStats {

    $Log->trace("Entering getNgshStats at " . localtime);

    my $serverInfo;
    my %stats;

    #--------------------
    # Process parameters
    #-------------------
    my %opts = validate(@_,
                        {srv_cmd  => {type => SCALAR,  optional => 0},
                         srv_opts => {type => HASHREF, optional => 0},
                        }
                        );

    $Log->trace(
          "getNgshStats called with the following arguments\n" . Dumper \%opts);

    $Server = new Server() if (!$Server);
    my $cmd = $opts{'srv_cmd'};
    my $output;

    #-----------------------------------------------
    # Execute the server command that was passed in
    #-----------------------------------------------
    eval {
        $serverInfo = $Server->$cmd(output       => \$output,
                                    '-Noverbose' => 1,
                                    %{$opts{'srv_opts'}}
                                    );
    };

    if ($@) {
        $Log->comment($output);
        die "Unable to execute $opts{'srv_cmd'} with $opts{'srv_opts'}: $@\n";
    } else {
        $Log->debug($output);

        #------------------------------------------------------
        # Extract the description and value for each statistic
        #------------------------------------------------------
        foreach my $stat (@$serverInfo) {

            # Remove quotation marks if they exist
            $stat->{'description'} =~ s/\"//g;
            $stat->{'description'} =~ s/\s+/_/g;

            # Remove quotation marks if they exist
            $stat->{'value'} =~ s/\%//g;

            $stats{$stat->{'description'}} = $stat->{'value'};
        } ## end foreach my $stat (@$serverInfo)

        $Log->trace(  "Successful completion of getNgshStats.  "
                    . "Returning "
                    . Dumper \%stats);

    } ## end else [ if ($@)

    $Log->trace("Exiting getNgshStats at " . localtime);
    return \%stats;

}    # end of getNgshStats

=head2 getNtapNbladeStats
            
C<getNtapNbladeStats> This routine calls the FreeBSD shell command to get ntap.nblade statistics for the specified node and writes the results to the specified file.
        
$ret = getNtapNbladeStats('node' => $opts{'node'});
        
=over                     
        
=item node
    
(Required) The node to gather stats from

=back

=cut

sub getNtapNbladeStats {

    $Log->trace("Entering getNtapNbladeStats at " . localtime);

    my $serverinfo;
    my %stats;

    #--------------------
    # Process parameters
    #--------------------
    my %opts =
      validate(@_,
               {node   => {type => SCALAR, optional => 0},
                search => {type => SCALAR, optional => 0}
               }
               );

    #---------------------------------------------------
    # Execute 'sysctl ntap.nblade' for the current node
    #---------------------------------------------------
    $Server = new Server() if (!$Server);
    my $output;
    eval {
        $Log->trace("Setting execution server to $opts{'node'}");
        $Server->set_execution_server(-Server      => $opts{'node'},
                                      '-Noverbose' => 1,
                                      output       => \$output,
                                      );
    };
    if ($@) {
        $Log->comment($output);
        die "Unable to set execution server to $opts{'node'}:" . $@;
    } else {
        $Log->debug($output);
        eval {
            $Log->trace("Executing sysctl ntap.nblade for $opts{'node'}");
            $serverinfo = $Server->shell('-Noverbose' => 1,
                                         output       => \$output,
                                         cmd => "sysctl $opts{'search'}"
                                         );
        };

        if ($@) {

            $Log->comment($output);
            die "Unable to execute sysctl $opts{'search'} for "
              . "$opts{'node'}"
              . $@;
            return;

        } else {
            $Log->debug($output);

            #-----------------------------------------------------
            # A newline-delimited string is returned.  Break into
            # individual statistics
            #-----------------------------------------------------
            my @fields = split /\n/, $serverinfo;

            #-------------------------------------------------
            # Create a hash based on the returned information
            #-------------------------------------------------
            foreach my $stat (@fields) {
                my @info        = split /: /, $stat;
                my $description = shift @info;
                my $value       = shift @info;

                # Remove "%" sign if it exists
                $value =~ s/\%//;
                $value =~ s/\s+/_/;
                $stats{$description} = $value;
            } ## end foreach my $stat (@fields)
        } ## end else [ if ($@)
    } ## end else [ if ($@)

    $Log->trace("Exiting GetNtapNbladeStats with " . Dumper \%stats);

    #--------------------------------
    # Return reference to stats hash
    #--------------------------------
    return \%stats;

}    # end of getNtapNbladeStats

=head2 create_empty_data_set

    create_empty_data_set('conn' => $fc, 'node' => $node);

C<create_empty_data_set> Use NET_PREFER to determine the IP address to use.

=over

=item header

(Required) Array of column headers

=item string

(Required) string to write in place or real data

=back

=cut

sub create_empty_data_set {

    $Log->trace("Entering create_empty_data_set at " . localtime);

    my ($headers, $string) = @_;
    my (%opts) =
      validate(@_,
               {headers => {type => ARRAYREF, optional => 0},
                string  => {type => SCALAR,   optional => 0}
               }
               );

    $Log->trace("Creating emtpy data set");

    # Create a new timestamp
    my %data;
    $data{'Uptime'} = 0;

    for (my $i = 1 ; $i < @{$opts{'headers'}} ; $i++) {
        $data{${$opts{'headers'}}[$i]} = $opts{'string'};
    }

    $Log->trace("Exiting create_empty_data_set at " . localtime);

    return \%data;
} ## end sub create_empty_data_set

=head2 write_stats_data

    This routine adds a timestamp to the header and data arrays, comma separates
    the items in both arrays, and writes them to a .csv file in TOP_LOGDIR and
    optionally ALT_LOGDIR.

    write_stats_data('conn' => $fc, 'node' => $node);

C<write_stats_data> Use NET_PREFER to determine the IP address to use.

=over

=item header

(Required) Hashref of column headers

=item data

(Required) Hashrefof data

=item node

(Required) Hostid of the node the stats belong to

=item timestamp

(Required) The timestamp for the current record

=item write_headers

(Required) Toggles the wrting of column headers

=item subsys

(Optional) The stats subsystem 

=item timestamp_format

(Optional) The format of the timestamp

=item top_logdir

(Optional) NATE's TOP_LOGDIR

=item alt_logdir

(Optional) The alternate log path

=item mode

(Optional) The stats mode 7m or 10m

=back

=cut

sub write_stats_data {

    $Log->trace("Entering write_stats_data at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();

    my (%opts) =
      validate(@_,
               {headers          => {type => ARRAYREF,       optional => 0},
                data             => {type => ARRAYREF,       optional => 0},
                node             => {type => SCALAR,         optional => 0},
                timestamp        => {type => SCALAR,         optional => 0},
                write_headers    => {type => SCALAR,         optional => 0},
                subsys           => {type => SCALAR,         optional => 1},
                timestamp_format => {type => SCALAR,         optional => 0},
                top_logdir       => {type => SCALAR | UNDEF, optional => 1},
                alt_logdir       => {type => SCALAR | UNDEF, optional => 1},
                mode             => {type => SCALAR,         optional => 0}
               }
               );

    my $headers;

    #------------------------------------------------------------
    # Prepend the timestamp and comma separate the headers values
    #------------------------------------------------------------
    if ($opts{'write_headers'}) {
        unshift(@{$opts{'headers'}}, "Timestamp");
        $headers = join(",", @{$opts{'headers'}}) . "\n";
    }

    #----------------------------------------------------------
    # Prepend the timestamp and comma separate the data values
    #----------------------------------------------------------
    if ($opts{'timestamp_format'} ne "epoch") {

        #-------------------------------------------------
        # Convert epoch seconds to date/time if requested
        #-------------------------------------------------
        $opts{'timestamp'} =
          strftime("%m/%d/%y %H:%M:%S", localtime($opts{'timestamp'}));
    }

    unshift(@{$opts{'data'}}, $opts{'timestamp'});
    my $data = join(",", @{$opts{'data'}}) . "\n";

    #---------------------------------------------------------
    # Open the log file and alternate log file (if specified)
    #---------------------------------------------------------
    my $logFile;
    if ($opts{'mode'}) {
        $logFile = "$opts{'mode'}-$opts{'subsys'}-stats-$opts{'node'}" . ".csv";
    } else {
        $logFile = "$opts{'subsys'}-stats-$opts{'node'}" . ".csv";
    }

    my $logPath = $options->{'TOP_LOGDIR'};
    if ($options->{'CSV_DIR'}) {
        $logPath .= "/$options->{'CSV_DIR'}";
    }

    $Log->debug("Opening $logPath/$logFile");
    open(LOG, ">>$logPath/$logFile");

    my $altLogPath = $options->{'ALT_LOGDIR'};
    if ($options->{'ALT_LOGDIR'}) {
        if ($options->{'CSV_DIR'}) {
            $altLogPath .= "/$options->{'CSV_DIR'}";
        }

        $Log->debug("Opening $logPath/$logFile");
        open(ALTLOG, ">>$altLogPath/$logFile");
    }

    #-----------------------------------------------------
    # Write headers data if it hasn't already been written
    #-----------------------------------------------------
    if ($opts{'write_headers'}) {
        $Log->debug("Writing headers to $logPath/$logFile");
        print LOG $headers;
        if ($options->{'ALT_LOGDIR'}) {
            $Log->debug("Writing headers to $altLogPath/$logFile");
            print ALTLOG $headers;
        }
    }

    #--------------------------------------
    # Write data values to the log file(s)
    #--------------------------------------
    $Log->debug("Writing data to $logPath/$logFile");
    print LOG $data;
    if ($options->{'ALT_LOGDIR'}) {
        $Log->debug("Writing data to $altLogPath/$logFile");
        print ALTLOG $data;
    }

    close LOG;
    if ($options->{'ALT_LOGDIR'}) { close ALTLOG; }

    $Log->trace("Exiting write_stats_data at " . localtime);

} ## end sub write_stats_data

=head2 parse_sysstat_output

C<parse_sysstat_output> Parseis the output of systat and returns a hashref

$stats = parse_sysstat_output($output);

=over

=item output

(Required) The scalar containing the text to parse

=back

=cut

sub parse_sysstat_output {

    $Log->trace("Entering parse_sysstat_output at " . localtime);

    my $output = shift;

    # Search the commands output
    my (%Stats, $node);
    foreach my $line (split(/\n/, $output)) {

        if ($line =~ /Node:\s+(\S+)/) {
            $node = $1;
        }

        # CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk    FCP iSCSI   FCP  kB/s
        #                                  in   out   read  write  read write   age   hit time  ty util                 in   out
        #  7%     0  1900     0    1900  1647  3328   2773     24     0     0     5   92%   0%  -   10%      0     0     0     0
        #
        # CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk    FCP iSCSI   FCP  kB/s\n
        #                                  in   out   read  write  read write   age   hit time  ty util                 in   out\n
        # 37%     0     0     0    2228  9857  4083    684  36949     0     0     7  100%  57%  F   40%   1983    54 20017 26600\n
        #
        if ($line =~ /^(\s+\d+|\d+)%/) {
            $line =~ s/^\s+(\d+%.*)/$1/;

            my @fields = split(/\s+/, $line, 21);
            for (my $f = 0 ; $f <= 21 ; $f++) {
                $fields[$f] =~ s/\s+//g;
            }

            my %stats;
            $stats{'CPU'}          = $fields[0];
            $stats{'NFS'}          = $fields[1];
            $stats{'CIFS'}         = $fields[2];
            $stats{'HTTP'}         = $fields[3];
            $stats{'Total'}        = $fields[4];
            $stats{'NetInKbs'}     = $fields[5];
            $stats{'NetOutKbs'}    = $fields[6];
            $stats{'DiskReadKbs'}  = $fields[7];
            $stats{'DiskWriteKbs'} = $fields[8];
            $stats{'TapeReadKbs'}  = $fields[9];
            $stats{'TapeWriteKbs'} = $fields[10];
            $stats{'CacheAge'}     = $fields[11];
            $stats{'CacheHit'}     = $fields[12];
            $stats{'CPTime'}       = $fields[13];
            $stats{'CPTY'}         = $fields[14];
            $stats{'DiskUtil'}     = $fields[15];
            $stats{'FCP'}          = $fields[16];
            $stats{'iSCSI'}        = $fields[17];
            $stats{'FCPInKbs'}     = $fields[18];
            $stats{'FCPOutKbs'}    = $fields[19];

            if ($node) {
                $Stats{$node} = \%stats;
            } else {
                %Stats = %stats;
                last;
            }
        } ## end if ($line =~ /^(\s+\d+|\d+)%/)
    } ## end foreach my $line (split(/\n/...

    $Log->trace("Exiting parse_sysstat_output at " . localtime);

    return \%Stats;

} ## end sub parse_sysstat_output

=head2 getClusteredStats

C<getSysStats> returns a hashref of specific stats averaged accross all or a i
subset of nodes in a cluster.

$stats = getClusteredStats('conn' => $filerConObj, 'duration' => 10);

=over

=item conn

(Required) Conn object of the filer required to be passed.

=item prefix	

(Optional). Default is 1. Toggle sub system prefixes in hash keys

=back

=cut

sub getClusteredStats {

    $Log->trace("Entering getClusteredStats at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my %opts = validate(@_,
                        {prefix =>
                           {type => BOOLEAN, default => $options->{'PREFIX'}}
                        }
                        );

    $Server = new Server() if (!$Server);

    $Log->debug("Getting 10mode clustered statistics\n");
    my ($counters, $categories, $counter_regex);
    if ($Server->get_release(cooked => 1) !~ /boilermaker/i) {
        $categories = 'node';
	$counter_regex = 'cpu-avg|-busy|-ops|connections|connected_shares|open-files';
    } else {
        $categories = 'node,cifs-currvalues';
	$counter_regex = 'cpu-avg|-busy|-ops|curr-connections|curr-users|curr-open-files';
    }

    my %args;
    if ($Server->get_release(cooked => 1) !~ /boilermaker/i) {
	$args{'object'} = $categories;
    } else {
	$args{'category'} = $categories;
    }

    eval {
        $counters =
          get_multi_node_stats(
             'srv_cmd' => 'statistics_show',
             'counter' => $counter_regex,
             'srv_opts' => {%args}
             );
    };

    if ($@) { die $@; }

    my %stats;
    foreach my $counter (keys(%{$counters})) {
        if ($opts{prefix}) {
            $stats{"cluster:" . $counter} = $counters->{$counter};
        } else {
            $stats{$counter} = $counters->{$counter};
        }
    }

    my $cmd = "sysstat -x -c 1 $options->{'SYSSTAT_DURATION'}";
    eval {
        $counters =
          get_multi_node_stats('srv_cmd'  => 'system_node_run',
                               'counter'  => 'CacheHit|CPTime|DiskUtil',
                               'srv_opts' => {'command' => $cmd, 'node' => "*"},
                               );
    };

    if ($@) { die $@; }

    foreach my $counter (keys(%{$counters})) {
        if ($opts{prefix}) {
            $stats{"cluster:" . $counter} = $counters->{$counter};
        } else {
            $stats{$counter} = $counters->{$counter};
        }
    }
    $Log->debug("stats: " . Dumper(\%stats));
    $Log->trace("Exiting _getClusteredStats " . localtime);

    return \%stats;

}

=head2 get_multi_node_stats

C<get_multi_node_stats> returns a hashref of specific stats across all or a 
subset of nodes in a cluster.

$stats = get_multi_node_stats(srv_cmd => "statistics show", 
                              srv_opts => {category => "node,cifs"});

=over

=item srv_cmd

(Required) The server command

=item srv_opts	

(Optional). Hash of server options for the command

=back

=cut

sub get_multi_node_stats {

    $Log->trace("Entering get_multi_node_stats at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();

    #--------------------
    # Process parameters
    #-------------------
    my %opts = validate(@_,
                        {srv_cmd  => {type => SCALAR},
                         srv_opts => {type => HASHREF},
                         counter  => {type => SCALAR},
                        }
                        );

    #---------------------------------------------------------------
    # Grab the ids that correspond to the indexs of the NODES array
    #---------------------------------------------------------------
    my ($node, @nodes);
    if ($options->{'CLUSTERED_STATS_NODE_INDEX'}) {
        foreach my $i (split(/,/, $options->{'CLUSTERED_STATS_NODE_INDEX'})) {
            if (!grep(/^${ $options->{'NODES'} }[$i]$/, @nodes)) {
                push(@nodes, ${$options->{'NODES'}}[$i]);
            }
        }
    } else {
        @nodes = @{$options->{'NODES'}};
    }

    if ($opts{'srv_opts'}->{'nodes'} eq '*') {
	$opts{'srv_opts'}->{'nodes'} = join(' ', @nodes);
    }

    my $cmd = $opts{'srv_cmd'};
    $Server = new Server() if (!$Server);

    #-----------------------------------------------
    # Execute the server command that was passed in
    #-----------------------------------------------
    my $Noverbose = 1;
    my ($serverInfo, $description, $value, $output);
    eval {
        $serverInfo = $Server->$cmd(output       => \$output,
                                    '-Noverbose' => $Noverbose,
                                    %{$opts{'srv_opts'}}
                                    );
    };

    if ($@) {
        $Log->comment($output);
        die "Unable to execute $opts{'srv_cmd'} with $opts{'srv_opts'}: $@\n";
    }
    $Log->debug($output);

    my %counters;
    if ($cmd !~ /system_node_run/) {
        foreach my $stat (@$serverInfo) {
            if (grep(/^$stat->{'node'}$/, @nodes)
                && $stat->{'counter'} =~ /$opts{'counter'}/) {
                $counters{$stat->{'counter'}}{'percent'} = 0;
                if ($stat->{'value'} =~ /\%/) {
                    $counters{$stat->{'counter'}}{'percent'} = 1;
                    $stat->{'value'} =~ s/\%//g;
                }
                push(@{$counters{$stat->{'counter'}}{'values'}},
                     $stat->{'value'}
                     );
            } ## end if (grep(/^$stat->{'node'}$/...
        } ## end foreach my $stat (@$serverInfo)
    } elsif (   $cmd =~ /system_node_run/
             && $opts{'srv_opts'}->{'command'} =~ /sysstat/) {
        my $sysStats = parse_sysstat_output($output);
        foreach my $node (keys(%{$sysStats})) {
            if (grep(/^$node$/, @nodes)) {
                foreach my $stat (keys(%{$sysStats->{$node}})) {
                    if ($stat =~ /$opts{'counter'}/) {
                        $counters{$stat}{'percent'} = 0;
                        if ($sysStats->{$node}->{$stat} =~ /\%/) {
                            $counters{$stat}{'percent'} = 1;
                            $sysStats->{$node}->{$stat} =~ s/\%//g;
                        }
                        push(@{$counters{$stat}{'values'}},
                             $sysStats->{$node}->{$stat}
                             );
                    } ## end if ($stat =~ /$opts{'counter'}/)
                } ## end foreach my $stat (keys(%{$sysStats...
            } ## end if (grep(/^$node$/, @nodes...
        } ## end foreach my $node (keys(%{$sysStats...
    } ## end elsif ($cmd =~ /system_node_run/...

    my %stats;
    foreach my $counter (keys(%counters)) {

        #--------------------------------------------------------
        # Find the maximum value of the counter across all nodes
        #--------------------------------------------------------
        my $max = 0;
        foreach my $value (@{$counters{$counter}{'values'}}) {
            if ($value > $max) { $max = $value; }
        }
        $stats{"max-$counter"} = $max;

        #--------------------------------------------------------
        # Find the minimum value of the counter across all nodes
        #--------------------------------------------------------
        my $min = $max;
        foreach my $value (@{$counters{$counter}{'values'}}) {
            if ($value < $min) { $min = $value; }
        }
        $stats{"min-$counter"} = $min;

        #----------------------------------------------
        # Sum the values of the counter from all nodes
        #----------------------------------------------
        my $sum = 0;
        foreach my $value (@{$counters{$counter}{'values'}}) {
            $sum += $value;
        }
        $stats{"sum-$counter"} = $sum;

        #--------------------------------------------------
        # Average the values of the counter from all nodes
        #--------------------------------------------------
        if ($counters{$counter}{'percent'}) {

            my $avg = 0;
            if (@{$counters{$counter}{'values'}} > 0) {
                $avg = sprintf("%.0f", $sum / @{$counters{$counter}{'values'}});
            }
            $stats{"avg-$counter"} = $avg;
        }
    } ## end foreach my $counter (keys(%counters...

    $Log->debug("counters: " . Dumper(\%counters));
    $Log->trace("Exiting _get_multi_node_stats " . localtime);
    return \%stats;

} ## end sub get_multi_node_stats

package Utils::FilerStats::10M;

use Server;
use Params::Validate qw(validate UNDEF SCALAR BOOLEAN OBJECT HASHREF ARRAYREF);
use Tharn qw($Log);
use Data::Dumper;

# Initialize the object includes collecting stats the first time
sub new {
    $Log->enter();
    my $class = shift;
    my (%opts) = validate(
        @_,
        {node => {type => SCALAR},
         category => {type => ARRAYREF,},},);

    my $self = bless {%opts}, $class;

    $Server = new Server if (!$Server);

    $self->collect_stats();

    $Log->exit();
    return $self;
} ## end sub new

# Updates saved stats by collecting it again
sub refresh {
    my $self = shift;
    $Log->enter();

    $self->collect_stats();

    $Log->exit();
}

#-------------------------------------
# Collect stats and save them by area
#-------------------------------------
sub collect_stats {
    my $self = shift;
    $Log->enter();

    $Log->comment("Getting stats for node $self->{node}");

    #------------------------------------------------------------
    # Run the command on the same node that we are interested in
    #------------------------------------------------------------
    $Server->set_execution_server(-Server => $self->{node});

    my %args;
    my $output;
    $args{'node'}       = $self->{node};
    $args{'output'}     = \$output;
    $args{'-Noverbose'} = 1;

    my $lmgr_ng = grep (/lmgr_ng/, @{$self->{category}});
    my @category = @{$self->{category}};

    if ($Server->get_release(cooked => 1) !~ /boilermaker/i) {
        my @objects = grep (!/replay|cifs-currvalues/, @category);
        $args{'object'} = join(",", @objects);

    } else {
        my @categories = grep (!/replay/, @category);
        $args{'category'} = join(",", @categories);
    }
    my @spl_category = grep (/replay/, @category);

    my ($stats, $output);
    eval { $stats = $Server->statistics_show(%args); };

    if ($@) {
        $Log->comment($output);
        die "Unable to get stats for node $self->{node}: $@";
    } else {
        $Log->debug($output);
    }
    #------------------------------------------------------
    # Extract the description and value for each statistic
    #------------------------------------------------------
    foreach my $stat (@$stats) {
        my ($category, $key);
        if ($Server->get_release(cooked => 1) !~ /boilermaker/i) {
            $category = $stat->{'object'};
            $key      = $stat->{'instance'} . "_" . $stat->{'counter'};
        } else {
            $category = $stat->{'category'};
            $key      = $stat->{'object'} . "_" . $stat->{'counter'};
        }
        my $value = $stat->{'value'};

        # Remove unwanted strings if they exist
        $key =~ s/(\"|\%|^latency_|^command_|^node_|^cifs-currvalues_)//g;
        $key =~ s/\s+/_/g;
        $self->{stats}{$category}{$key} = $value;
    }

    #-------------------------------------------------------------------------
    # XXX
    # lmgr_ng will not work keying based on the description since each object
    # like volumes, and node has the same description displayed and hence
    # only the last one would be saved from the above. We need to get just
    # node based stats to make sure we have some useful data in general.
    # When vol stats get added, may be the volume objects can go there.
    #-------------------------------------------------------------------------
    if ($lmgr_ng) {
        if ($Server->get_release(cooked => 1) !~ /boilermaker/i) {
            $args{'object'}   = 'lmgr_ng';
            $args{'instance'} = $self->{node};
        } else {
            $args{'category'} = 'lmgr_ng';
            $args{'object'}   = $self->{node};
        }

        eval { $stats = $Server->statistics_show(%args); };

        if ($@) {
            $Log->comment($output);
            die "Unable to get stats for node $self->{node}: $@";
        } else {
            $Log->debug($output);
        }

        foreach my $stat (@$stats) {
            my ($category, $key);
            if ($Server->get_release(cooked => 1) !~ /boilermaker/i) {
                $category = $stat->{'object'};
                $key      = $stat->{'instance'} . "_" . $stat->{'counter'};
            } else {
                $category = $stat->{'category'};
                $key      = $stat->{'object'} . "_" . $stat->{'counter'};
            }
            my $value = $stat->{'value'};

            # Remove unwanted strings if they exist
            $key =~ s/(\"|\%)//g;
            $key =~ s/\s+/_/g;

            $self->{stats}{$category}{$key} = $value;
        }
    }

    #------------------------------------------------------
    # Collet special category stats using diff commands
    #------------------------------------------------------
    foreach my $spl (@spl_category) {
        if ($spl eq "replay") {
            eval {
                $output =
                  $Server->system_node_run(node         => $self->{node},
                                           output       => \$output,
                                           '-Noverbose' => 1,
                                           command      => "spinnp_replay_stats"
                                           );
            };
        }
        if ($@) {
            $Log->comment($output);
            die "Unable to get stats for node $self->{node}: $@";
        } else {
            $Log->debug($output);
        }

        foreach (split /\n/, $output) {
            chomp;
            next if ($_ =~ /^$/);
            my ($key, $value) = split(":", $_);
            $key   =~ s/\s+/_/g if ($key);
            $value =~ s/\s+//g  if ($value);
            $self->{stats}{$spl}{$key} = $value if ($key && $value);
        }
    }
    $Log->exit();
}

#--------------------------------------
# returns the interested stats by area
#--------------------------------------
sub get_stats {
    my $self = shift;
    $Log->enter();

    my %opts =
      validate(@_,
               {category   => {type => SCALAR | ARRAYREF},
                add_prefix => {type => SCALAR, optional => 1}
               }
               );

    my @categories;
    if (ref($opts{category})) {
        @categories = @{$opts{category}};
    } else {
        push @categories, $opts{category};
    }

    my %rhash;
    foreach my $category (@categories) {
        foreach my $key (keys %{$self->{stats}{$category}}) {
            my $value = $self->{stats}{$category}{$key};
            $key = $opts{add_prefix} . ":" . $key if ($opts{add_prefix});
            $rhash{$key} = $value;
        }
    }

    return \%rhash;
    $Log->exit();
} ## end sub get_stats

1;
