# $Id$
## @summary Common library to collect filer statistics
## @author bellizzi@netapp.com,gvg@netapp.com
## @pod here

=head1 NAME

Utils::Graph.pm - Common library to generate images out of data collected

=head1 DESCRIPTION

This package defines routines which can be used to create images from
the data collected.

=head1 SYNOPSIS

    use Utils::Graph qw(csv2data plot_graph);

    my ($legends, $data) = csv2data(csvfile => $Options->{CSVFILE},
                                    xfield => $Options->{XFIELD},
                                    yfields => $Options->{YFIELDS})
        or die "csv2data failed: $@\n";

    plot_graph(imagefile => $Options->{IMAGEFILE},
               data => $data,
               legends => $legends)
        or die "draw_image failed: $@\n";

=head1 ROUTINES

=cut

package Utils::Graph;
use strict;
use vars qw (@ISA @EXPORT $VERSION $DEBUG);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR BOOLEAN OBJECT HASHREF ARRAYREF);
use Options;
use TracePkg;
use GD;
use GD::Text;
use GD::Graph::lines;
use GD::Graph::area;
use GD::Graph::bars;
use Data::Dumper;

my $Trace = new TracePkg;

my $Options = new Options;
$Options->parameters(
    definitions => [
        {   name        => "WIDTH",
            description => "Width of the image in pixels",
            spec        => "s",
            default     => "1024",
        },
        {   name        => "HEIGHT",
            description => "Height of the image in pixels",
            spec        => "s",
            default     => "768",
        },
        {   name        => "X_LABEL",
            description => "X Label",
            spec        => "s",
            default     => undef,
        },
        {   name        => "Y_LABEL",
            description => "Y Label",
            spec        => "s",
            default     => undef,
        },
        {   name        => "TITLE",
            description => "Title of the Image",
            spec        => "s",
            default     => undef,
        },
        {   name        => "IMAGEFILE",
            description => "Filepath to the image file",
            spec        => "s",
            default     => undef,
        },
        {   name        => "CSVFILE",
            spec        => "s",
            default     => undef,
            description => "The Data file with the statistics",
        },
        {   name        => "XFIELD",
            spec        => "s",
            default     => "Date",
            description => "X axis of the graph",
        },
        {   name        => "YFIELDS",
            spec        => "s",
            default     => undef,
            description => "Y axis of the graph",
        },
        {   name        => "YREGEX",
            spec        => "s",
            default     => undef,
            description => "Selected headers based on regex",
        },
        {   name        => "NOT_YFIELDS",
            spec        => "s",
            default     => undef,
            description => "! Y axis of the graph",
        },
        {   name        => "NO_XLABELS",
            spec        => "s",
            default     => 15,
            description => "Number of X labels to be appeared",
        },
        {   name        => "TTFPATH",
            spec        => "s",
            default     => '',
            description => 'The path to the true type font dir',
        },
        {   name        => "LASTRECORDS",
            spec        => "s",
            default     => '300',
            description => 'Only return the last n records',
        },
    ]
);

$VERSION = 1.00;
@ISA     = qw (Exporter);
@EXPORT  = qw (
    csv2data
    plot_graph
);

=head2 plot_graph

C<plot_graph> creates graph from the data provided.



=over

=item data

(Required) The ARRAYREF of the data

=item legends

(Required) The ARRAYREF of the legends

=item width

(Optional) The width of the graph in pixels.

=item height

(Optional) The height of the graph in pixels.

=item x_label

(Optional) Label of X axis of the graph.

=item y_label

(Optional) Label of Y axis of the graph.

=item title

(Optional) Title of the graph.

=item imagefile

(Optional) The filepath of the image to be created. By default, $image is returned.

=back

=cut

sub plot_graph {
    $Options->process();
    my $hash = $Options->get_hash();

    my (%opts) = validate_mixed(
        @_,
        [],
        {   width     => {type => SCALAR, default => $hash->{"WIDTH"}},
            height    => {type => SCALAR, default => $hash->{"HEIGHT"}},
            x_label   => {type => SCALAR, default => $hash->{"X_LABEL"}},
            y_label   => {type => SCALAR, default => $hash->{"Y_LABEL"}},
            title     => {type => SCALAR, default => $hash->{"TITLE"}},
            imagefile => {type => SCALAR, default => $hash->{"IMAGEFILE"}},
            ttf_path  => {
                type     => SCALAR,
                default  => $hash->{"TTFPATH"},
                optional => 1
            },
            type    => {type => SCALAR,   default  => 'line', optional => 1},
            data    => {type => ARRAYREF, optional => 0},
            legends => {type => ARRAYREF, optional => 0},
            graph_options => {type => HASHREF, optional => 1},
        },
    );

    if (@{$opts{data}} < 2) { return; }

    my $graphObj;

    if ($opts{type} eq 'bar') {
        $graphObj = GD::Graph::bars->new($opts{width}, $opts{height});
    } else {
        $graphObj = GD::Graph::lines->new($opts{width}, $opts{height});
    }

    # Adjust x_label_skip so it will be readable
    $Trace->trace(msg => "No of data records: " . scalar(@{$opts{data}[0]}));

    my $x_label_skip = int(@{$opts{data}[0]} / $hash->{NO_XLABELS});
    $Trace->trace(msg => "Adjusting x_label_skip to $x_label_skip");

    my %graphOptions = (
        x_labels_vertical => 'true',
        y_tick_number     => 10,
        y_label_skip      => 1,
        x_label_skip      => $x_label_skip,
        long_ticks        => 1,
        legend_placement  => 'BC',
        x_label_position  => 1 / 2,
        bgclr             => 'lgray',
        fgclr             => 'dgray',
        boxclr            => 'black',
        accentclr         => 'white',
        valuesclr         => 'lgreen',
        labelclr          => 'black',
        axislabelclr      => 'black',
        labelclr          => 'black',
        textclr           => 'black',
        line_width        => 3,
        logo              => 'logo.' . GD::Graph->export_format,
        logo_position     => 'BR',
        transparent       => 0,
        l_margin          => 10,
        b_margin          => 10,
        r_margin          => 10,
        t_margin          => 10,
        show_values       => 0,
        values_format     => "%4.1f"
    );

    if ($opts{'graph_options'}) {
        foreach my $opt (keys(%{$opts{'graph_options'}})) {
            $graphOptions{$opt} = ${$opts{'graph_options'}}{$opt};
        }
    }

    my @lineColors = qw(lgreen lred lblue white lyellow lpurple lorange pink
        marine cyan lbrown blue gold yellow green red purple
        orange dblue dyellow dgreen dred dpurple dpink dbrown);

    push(@{$graphOptions{'dclrs'}}, @lineColors);
    $graphOptions{'x_label'}
        = $opts{x_label} ? $opts{x_label} : $hash->{"XFIELD"};
    $graphOptions{'y_label'}
        = $opts{y_label} ? $opts{y_label} : join(",", @{$opts{legends}});
    $graphOptions{'title'}
        = $opts{title}
        ? $opts{title}
        : $graphOptions{'x_label'} . " Vs " . $graphOptions{'y_label'};

    $graphObj->set_legend(@{$opts{legends}});
    my $textObj = GD::Text->new();
    if ($opts{'ttf_path'}) {
        $textObj->font_path($opts{'ttf_path'});
    }
    $graphObj->set_title_font(['arial', 'verdana', 'comic', GD::Font->Giant],
        14);
    $graphObj->set_x_label_font(
        ['arial', 'verdana', 'comic', GD::Font->Large], 12);
    $graphObj->set_y_label_font(
        ['arial', 'verdana', 'comic', GD::Font->Large], 12);
    $graphObj->set_x_axis_font(['arial', 'verdana', 'comic', GD::Font->Tiny],
        9);
    $graphObj->set_y_axis_font(['arial', 'verdana', 'comic', GD::Font->Tiny],
        9);
    $graphObj->set_legend_font(
        ['arial', 'verdana', 'comic', GD::Font->MediumBold], 10);
    $graphObj->set(%graphOptions) or warn $graphObj->error;

    my $image = $graphObj->plot($opts{data})
        or die $graphObj->error;

    # If Image file is specified create a file
    if ($opts{imagefile}) {
        open(OUT, ">$opts{imagefile}") or die $!;
        binmode OUT;
        print OUT $image->png();
        close OUT;
    }

    return $image;
} ## end sub plot_graph

=head2 csv2data

C<csv2data> Parses csvfile into @legends and @data to pass to
C<plot_graph> function.

=over

=item csvfile

(Required) The filepath of the csvfile

=item xfield

(Optional) The X axis field of the graph to be created

=item yfields

(Optional) The comma separated Y axis fields of the graph to be created with.

=item not_yfields

(Optional) The comma separated Y axis fields of the graph not to be created with.

=back

=cut

sub csv2data {
    $Options->process();
    my $hash = $Options->get_hash();

    my (%opts) = validate_mixed(
        @_,
        [],
        {   csvfile => {
                type     => SCALAR,
                default  => $hash->{"CSVFILE"},
                optional => 0
            },
            xfield =>
                {type => SCALAR, default => $hash->{"XFIELD"}, optional => 1},
            yfields => {
                type     => SCALAR,
                default  => $hash->{"YFIELDS"},
                optional => 1
            },
            yregex => {
                type     => SCALAR,
                default  => $hash->{"YREGEXP"},
                optional => 1
            },
            not_yfields => {
                type     => SCALAR,
                default  => $hash->{"NOT_YFIELDS"},
                optional => 1
            },
            last_records => {type => SCALAR, optional => 1},
        },
    );

    # Read CSV data
    open(CSV, $opts{csvfile})
        or die "ERROR: Could not open CSV file: $opts{csvfile} \n";
    my @lines = <CSV>;
    close(CSV);

    # Store all Data into a hashref
    my @headers     = ();
    my $headerFound = 0;
    my $hashref;
    foreach my $line (@lines) {
        chomp($line);

        if ($line !~ /,/) {

            # Ignore non CSV lines
            next;
        } elsif ($headerFound == 0) {
            @headers = split(",", $line);
            if (!$opts{xfield}) {

             # No xfield is passed, so default to first column of csvdata file
                $opts{xfield} = $headers[0];
            }
            $Trace->debug(msg => "Headers: " . Dumper(\@headers));
            $headerFound = 1;
            next;
        } elsif ($headerFound == 1 && $line =~ /$opts{xfield}/) {

            # Pickup successive headers
            @headers = split(",", $line);
            $Trace->debug(msg => "Headers: " . Dumper(\@headers));
            next;
        }
        if (my @fields = split(",", $line)) {
            for (my $i = 0; $i < @headers; $i++) {
                $fields[$i] =~ s/(\n|^\s+|%|\s+$|>|<|-)//g;    # Remove stuff
                push(@{$hashref->{$headers[$i]}}, $fields[$i])
                    if (defined($fields[$i]));
            }
        }
    }
    if (   $hashref
        && $opts{'last_records'}
        && $opts{'last_records'} < @{$hashref->{$headers[0]}}) {
        my $shifts = @{$hashref->{$headers[0]}} - $opts{'last_records'};
        for (my $i = 0; $i < @headers; $i++) {
            for (my $s = $shifts; $s >= 0; $s--) {
                shift(@{$hashref->{$headers[$i]}});
            }
        }
    }
    $Trace->debug(msg => "CSV Data: " . Dumper($hashref));

    # Delete X field from @headers to get all possible yfields
    my @yfields = grep(!/$opts{xfield}/, @headers);
    $Trace->debug(
        msg => "X field: $opts{xfield}, Y Fields: " . Dumper(\@yfields));

    # Decide on legend keys based on yfields
    my @legends;
    if ($opts{not_yfields}) {
        for (my $i = 0; $i < @yfields; $i++) {
            my $match = 0;
            foreach my $not_field (split(',', $opts{not_yfields})) {
                if ($not_field =~ /$yfields[$i]/i) {
                    $match = 1;
                }
            }
            push(@legends, $yfields[$i]) if ($match == 0);
        }
    } elsif ($opts{yfields}) {
        chomp($opts{yfields});
        push(@legends, split(',', $opts{yfields}));
    } elsif ($opts{yregex}) {
        for (my $i = 0; $i < @yfields; $i++) {
            if ($yfields[$i] =~ /$opts{yregex}/) {
                push(@legends, $yfields[$i]);
            }
        }
    } else {
        push(@legends, @yfields);
    }

    # Filter out only required data
    my @data;
    if ($hashref->{$opts{xfield}}) {
        push(@data, $hashref->{$opts{xfield}});
    }
    foreach my $field (@legends) {
        if ($hashref->{$field}) {
            push(@data, $hashref->{$field});
        }
    }

    $Trace->debug(msg => "Legend Keys: " . Dumper(\@legends));
    $Trace->debug(msg => "Data: " . Dumper(\@data));
    return (\@legends, \@data);
} ## end sub csv2data

1;
