# $Id: //depot/prod/test/Rsierranevada/lib/Utils/Grid.pm#1 $

## @summary Help create 2d tables of test results
## @author lincoln@netapp.com, dl-test-infra@netapp.com
## @status Shared
## @pod here

package Utils::Grid;

=head1 NAME

Utils::Grid - Help create 2d tables of test results

=head1 SYNOPSIS

Help create 2d tables of test results

=head1 DESCRIPTION

This is a library to help in the generation of two-dimensional tables
of results, probably with multiple configurations/testbeds/etc. per
column and multiple builds/requests/events per row.

If you do not need multiple columns this way, you will probably be
better served by just pointing F<natelogs.cgi> to the top-level log
directory.  The typical test which uses this is F<util/grid.thpl>, but
you might need a different test using this library if you want to
customize the text which can appear in the table entries (see
F<bootlab/grid.thpl>).

=head2 Cells

A Utils::Grid object logically contains a set of table cells.  Each
cell contains an array of results, where each result has the following
properties:

=over

=item * 

row name (which row it will appear in)

=item *

column name (which column it it will appear in)

=item *

time (when it was created, a perl time() value)

=item *

text (what text should appear verbatim in the cell)

=item *

href (optional, what URL the text should be a link to)

=item *

color (optional, what background color should be used for text)

=back

Rows and columns are created and removed as necessary to show all of
the given table cells, but no more (no empty rows or columns).  When a
table is created, rows will be reverse asciibetically sorted (higher
values at the top), while columns will be asciibetically sorted
(higher values at the right).

=head2 Files

This library can read or write its data to/from a database file, and
can write its data to an HTML file.  When updates occur, both data and
html files are edited in place after being locked, so multiple tests
can safely attempt to update the table in parallel.

The database file is a dump of the table using Data::Dumper.
(Data::Dumper is used instead of Storable so that one can edit/remove
table entries by hand if one is careful).

The HTML file can contain any HTML, but should contain

	<!--Start Table-->

and

	<!--End Table-->

somewhere within it (if not, updates to the HTML file will not have
any real effect).  Whenever it is updated, all previous text between
those two directives is removed, and a new HTML table placed between
them instead.

=head1 METHODS

=over

=cut

use strict;
use warnings;

use Data::Dumper;
use Fcntl;
use IO::File;

=item new

    my $grid = Utils::Grid->new($datafile,$htmlfile);

create and return a new, empty Utils::Grid object.  C<$datafile> and
C<$htmlfile> are the filenames of the database and HTML files, to be
used by later methods.

=cut

sub new {
    my ($proto, $datafile, $htmlfile) = @_;
    my $class = ref($proto) || $proto;
    my $self = {
	datafile => $datafile,
	datafilefh => undef,
	htmlfile => $htmlfile,
	htmlfilefh => undef,
    };
    bless($self,$class);
    $self->clear();
    return $self;
}

=item open_lock

    $grid->open_lock();

Open and lock both the data file and HTML file.  It dies, with neither
file locked, on error.  The files are created (empty) if they did not
already exist, so it is safe to call C<open_lock> in a program which
does not have any data yet (though the HTML file will probably not be
very useful until manually populated with actual HTML).

=cut

sub open_lock {
    my ($self) = @_;
    my @handles;
    foreach my $filenamename ("datafile","htmlfile") {
	my $filename = $self->{$filenamename};
	my $fh = new IO::File($filename, O_CREAT|O_RDWR)
	    or die "Could not open $filename: $!";
	_lockfile($fh)
	    or die "Could not lock $filename: $!";
	push @handles, $fh;
    }
    # We don't store either handle until this point, so that if the
    # second fails to allocate, then the first will close
    # automatically.
    $self->{datafilefh}=$handles[0];
    $self->{htmlfilefh}=$handles[1];
}

=item save

    $grid->save();

write in-memory data to both the data file and the HTML file.  returns
on success, dies on error.

=cut

sub save {
    my ($self) = @_;
    $self->_save_data();
    $self->_save_html();
}

=item close

    $grid->close();

Close (and therefore unlock) the data file and html file, opened using
C<open_lock>.  returns on success, dies on error.

=cut

sub close {
    my ($self) = @_;
    foreach my $aref (["datafile","datafilefh"],
		      ["htmlfile","htmlfilefh"]) {
	my ($filenamename, $fhname) = @$aref;
	my $filename = $self->{$filenamename};
	close($self->{$fhname})
	    or die "Could not close $filename: $!\n";
	delete $self->{$fhname};
    }
}

=item load

    $grid->load();

Read data from the data file.  Returns on success, dies on error.  It
is safe to attempt to load from a newly created data file, so it is
safe to call C<load> after C<open_lock> in a program which does not
have any data yet. This function loads the data from the format of single
run per row/column pair and converts it to an aray to accomodate the results
of re-runs for the same row/column pair.

=cut

sub load {
    my ($self) = @_;
    my $filename = $self->{datafile};
    my $in = $self->{datafilefh};
    sysseek($in,0,SEEK_SET)
	or die "Could not seek in $filename: $!\n";
    my $size = -s $in;
    my $coded_data;
    if (!defined(sysread($in,$coded_data,$size))) {
	die "Could not read $filename: $!\n";
    }
    my $data = eval $coded_data;
    if ($@) {
	die "Error evaluating grid data from file \"$filename\": $@";
    }
    $self->{data}=$data;

# convert any older one-result-per-cell data to the new
# multiple-results-per-cell format    

    foreach my $r (keys %{$self->{data}}) {
        foreach my $c (keys %{$self->{data}{$r}}) {
            if (ref $self->{data}{$r}{$c} ne "ARRAY") {
                my $temp = [$self->{data}{$r}{$c}];
                $self->{data}{$r}{$c}=$temp;
            }
        }
    }

}

=item clear

    $grid->clear();

clear (in memory) the current cache of data.

=cut

sub clear {
    my ($self) = @_;
    $self->{data} = {};
}

=item add

    $grid->add($row,$column,time=>$time,text=>$string,color=>$color,href=>$href)

add a new table cell to the set.

C<time> defaults to the current C<time()> if not otherwise supplied.
C<text> has no default (the new table cell will be added but will have
no text).  If C<color> or C<href> are not provided, no background
color or URL will be emitted for this cell when writing the HTML
table.

=cut

sub add {
    my ($self, $row, $column, %hash) = @_;
    if (!exists($hash{time})) {
	$hash{time}=time();
    }
    push @{$self->{data}{$row}{$column}}, \%hash;
}

=item clear_older_than

    $grid->clear_older_than($time);

clear (in memory) cells if their first result has a C<time> which is
older than C<$time>.

=cut

sub clear_older_than {
    my ($self, $time) = @_;
    foreach my $r (keys %{$self->{data}}) {
	foreach my $c (keys %{$self->{data}{$r}}) {
	    my $t = $self->{data}{$r}{$c}[0]->{time};
	    if (defined($t) && ($t<$time)) {
		delete $self->{data}{$r}{$c};
	    }
	}
	if (!keys(%{$self->{data}{$r}})) {
	    delete $self->{data}{$r};
	}
    }
}

# update just the data file

sub _save_data {
    my ($self) = @_;
    my $filename = $self->{datafile};
    my $out = $self->{datafilefh};
    sysseek($out,0,SEEK_SET)
	or die "Could not seek in $filename: $!\n";
    my $dumper = new Data::Dumper([$self->{data}]);
    $dumper->Terse(1);
    my $coded_data = $dumper->Dump();
    if (!defined(syswrite($out,$coded_data))) {
	die "Could not write to $filename: $!\n";
    }
    truncate($out,length($coded_data))
	or die "Could not truncate $filename: $!\n";
}

# update just the HTML file

sub _save_html {
    my ($self) = @_;
    my $filename = $self->{htmlfile};
    my $inout = $self->{htmlfilefh};

    # Read Old HTML
    sysseek($inout,0,SEEK_SET)
	or die "Could not seek in $filename: $!\n";
    my $size = -s $inout;
    my $html;
    if (!defined(sysread($inout,$html,$size))) {
	die "Could not read $filename: $!\n";
    }

    # Substitute new Data
    my $table = $self->_data2table();
    my $start = "<!--Start Table-->";
    my $end = "<!--End Table-->";
    $html =~ s/$start.*$end/$start."\n".$table."\n".$end/seg;

    # Write New HTML
    sysseek($inout,0,SEEK_SET)
	or die "Could not seek in $filename: $!\n";
    if (!defined(syswrite($inout,$html))) {
	die "Could not write to $filename: $!\n";
    }
    truncate($inout,length($html))
	or die "Could not truncate $filename: $!\n";
}

# Return the text of an HTML table created from the contents of
# $self->{data}

sub _data2table {
    my ($self) = @_;
    my @rows = reverse sort keys %{$self->{data}};
    my %column_set;
    foreach my $row (@rows) {
	foreach my $column (keys %{$self->{data}{$row}}) {
	    $column_set{$column}=1;
	}
    }
    my @columns = sort keys %column_set;
    my $table = "<TABLE BORDER=\"1\" ALIGN=\"CENTER\">\n";
    $table .= "  <TR>\n";
    $table .= "    <TH>&nbsp;</TH>\n";
    foreach my $c (@columns) {
	$table .= "    <TH>" . &_entitize($c) . "</TH>\n";
    }
    $table .= "  </TR>\n";
    foreach my $r (@rows) {
	$table .= "  <TR>\n";
	$table .= "    <TD>" . &_entitize($r) . "</TD>\n";
	foreach my $c (@columns) {
	    $table .= "    " . $self->_htmlcell($r,$c) . "\n";
	}
	$table .= "  </TR>\n";
    }
    $table .= "</TABLE>";
    return $table;
}

# Return the html for one cell of $self->{data}
# Multiple reults of re-runs in the same cell

sub _htmlcell {
    my ($self, $row, $column) = @_;
    my $cells = $self->{data}{$row}{$column};
    if (!$cells) {
       return "<TD>&nbsp;</TD>";
    } else {
       my $count = 1;
       my $out = "";
       foreach my $cell (@{$cells}) {
       	  if ($count == 2) {
       	      $out = "<TR>" . $out . "</TR>";
       	  }
          my $text = $cell->{text};
          my $href = $cell->{href};
          my $color = $cell->{color};
          my $out1 = &_entitize($text);
          if ($href) {
              $out1 = "<A HREF=\"$href\">$out1</A>";
	  }
	  
          if ($color) {
              $out1 = "<TD BGCOLOR=\"$color\">$out1</TD>";
          } else {
              $out1 = "<TD>$out1</TD>";
          }
          if ($count > 1) {
	      $out1 = "<TR>" . $out1 . "</TR>";
       	  }
       	  
	  $out = $out . $out1;
	  $count++;
       }
       if ($count > 2) {
       	   $out = "<TD><TABLE ALIGN=\"CENTER\">" . $out . "</TABLE></TD>";
       }
       return $out;
    }
}

# Escape $str so that it will be displayed in html "as is"

sub _entitize ( $ ) {
    my ($str) = @_;
    $str =~ s/&/&amp;/g;
    $str =~ s/</&lt;/g;
    $str =~ s/>/&gt;/g;
    return $str;
}

# A lock mechanism which uses fcntl instead of flock (so it works on
# both Linux and Solaris over NFS)

sub _lockfile ( $ ) {
    my ($fh) = @_;
    my $struct_flock = pack("sx256",F_WRLCK);
    return fcntl($fh,F_SETLKW,$struct_flock);
}

=back

=head1 EXAMPLES

    my $grid = Utils::Grid->new("/u/myname/web/mylog.dat","/u/myname/web/mylog.html");
    $grid->open_lock();
    $grid->load();
    $grid->add("Comet","Regular","PASS","results/comet.log","#AAFFAA");
    $grid->add("Comet","Lemon Fresh","PASS","results/comet-lemon.log","#AAFFAA");
    $grid->add("Brand X","Regular","FAIL","results/brandx.log","#FF8888");
    $grid->save();
    $grid->close();

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    Lincoln Myers (lincoln@netapp.com), dl-test-infra@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/Utils/Grid.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;
