# $Id$
## @summary Common library for SSQA to do Admin tasks on Hosts (clients)
## @author gvg@netapp.com
## @pod here

=head1 NAME

Utils::HostAdmin.pm - Common library for SSQA to administer hosts

=head1 DESCRIPTION

This package defines routines to check for hosts health, reboot, rpower, etc

=head1 SYNOPSIS

    use Utils::HostAdmin;

    my $Ha = new Utils::HostAdmin (clients => \@clients);
    
    my ($goods, $bads) = $Ha->clients_status();

    my ($goods, $bads) = $Ha->clients_status(reboot => 1);

=head1 EXAMPLE

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Created $Date: 2009/11/20 $

=head1 COPYRIGHT

Copyright 2009, Network Appliance., All Rights Reserved

=head1 ROUTINES

=cut

package Utils::HostAdmin;
use strict;
use POSIX ":sys_wait_h";
use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use Tharn qw($Log config_result script_result);
use Data::Dumper;

=head2 new

    my $Ha = new Utils::HostAdmin(clients => [client1 client2 ...],);

C<new> creates HostAdmin object with the list of clients and some defaults.

=over

=item clients

(Optional) List of clients. Arrayref.

=item tests

(Optional) The tests to be performed on the clients to declare the clients status. Default: ping, rsh, staf.

=item timeout

(Optional) How long wait is allowed to get clients status. Default: 360 secs

=item retries

(Optional) How many retries can be attempted to get status of the clients. Default: 3

=back

=cut

sub new {
    my ($invocant, %opts) =
      validate_mixed(@_,
                     [{type => SCALAR},],
                     {clients => {type => ARRAYREF, default => undef},
                      tests =>
                        {type => ARRAYREF, default => ["ping", "rsh", "staf"],},
                      timeout => {type => SCALAR, default => 120},
                      retries => {type => SCALAR, default => 3},},);
    my $class = ref($invocant) || $invocant;
    my $self = {%opts};
    bless($self, $class);

    $self->{logdir} = Tharn::param("LOGDIR", -default, "/tmp") . "/host_admin/";

    my $r = mkdir($self->{logdir}, oct("777"));
    if (!$r and not $!{EEXIST}) {
        die script_result("Unable to create log directory $self->{logdir}: $!",
                          stacktrace => 0);
    }

    return $self;
} ## end sub new

=head2 clients_status
    
    my ($goods, $bads) = $Ha->clients_status(clients => @clients, refresh => 1);

C<clients_status> Returns as two lists with goods and bads. Check for params of new and the below.

=over

=item refresh

(Optional) Refresh for the latest status. Default 0.

=item reboot

(Optional) Reboot the clients and return the status. Default: 0. If set, refresh is forced.

=back

=cut

sub clients_status {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {clients => {type => ARRAYREF, default => undef},
                      refresh => {type => SCALAR,   default => 0},
                      reboot  => {type => SCALAR,   default => 0},
                      tests   => {type => ARRAYREF, default => undef,},
                      timeout => {type => SCALAR,   default => undef},
                      retries => {type => SCALAR,   default => undef},},);
    my (@goods, @bads, %pids, @clients, @tests, $prev);

    if ($opts{clients} && @{$opts{clients}}) {
        push @clients, @{$opts{clients}};
    } elsif (@{$self->{clients}}) {
        push @clients, @{$self->{clients}};
    } else {
        die config_result("clients neither specified nor object created with",
                          stacktrace => 0);
    }

    if ($opts{tests} && @{$opts{tests}}) {
        push @tests, @{$opts{tests}};
    } elsif (@{$self->{tests}}) {
        push @tests, @{$self->{tests}};
    } else {
        die config_result("tests were not specified", stacktrace => 0);
    }

    # Go through the list to see if all clients status are readily available
    my $status_available = 1;
    foreach my $host (@clients) {
        my $pass = 0;
        foreach my $test (@tests) {
            if (   $self->{status}{host}{$test} ne "PASS"
                && $self->{status}{host}{$test} ne "FAIL") {
                $status_available = 0;
            }
            $pass++ if ($self->{status}{host}{$test} eq "PASS");

            if ($pass == scalar(@tests)) {
                push @goods, $host;
            } else {
                push @bads, $host;
            }
        } ## end foreach my $test (@tests)
    } ## end foreach my $host (@clients)

    if (!$status_available || $opts{refresh}) {
        @goods = ();
        @bads  = ();
        foreach my $host (@clients) {
            $Log->debug("starting health check for $host");
            my $tries = 0;
            my $pid;

            # fork() on Solaris will occaisionally fail with ENOSPC, even
            # though the space shortage is very temporary, so if fork fails
            # then retry it a limited number of times.
            while (1) {
                $pid = fork();
                if (defined($pid)) {

                    # parent
                    $pids{$host} = $pid;
                    last;
                }
                if ($tries > 10) {
                    my $error = $!;
                    my @extra_info;
                    if ($^O !~ /win/i) {
                        my $top;
                        my $swap;
                        if ($^O =~ /lin/i) {
                            $top  = "/usr/bin/top -b -n 1";
                            $swap = "/sbin/swapon -s";
                        } else {
                            $top  = "/usr/bin/prstat -a 1 1";
                            $swap = "/usr/sbin/swap -s";
                        }
                        push(@extra_info,
                             ("swap   = " . `$swap`,
                              "ulimit = " . `ulimit -a`,
                              "top    = " . `$top`));
                    } ## end if ($^O !~ /win/i)
                    die config_result(
                          join(
                              "\n",
                              (@extra_info,
                               "Unable to create subprocess with fork(): $error"
                                )));
                } ## end if ($tries > 10)
                $tries++;
            } ## end while (1)

            if (not $pid) {

                # Child process
                $self->on_client(
                              host    => $host,
                              logfile => $self->{logdir} . "/" . $host . ".log",
                              tests   => \@tests,
                              retries => $opts{retries} || $self->{retries},
                              timeout => $opts{timeout} || $self->{timeout},
                              reboot  => $opts{reboot});
            } ## end if (not $pid)
        } ## end foreach my $host (@clients)
        $Log->comment("All client verification processes started");

        while (keys %pids) {
            foreach my $host (keys %pids) {
                my $pid         = $pids{$host};
                my $retpid      = waitpid($pid, WNOHANG);
                my $exit_status = $?;
                if ($retpid > 0) {
                    ### hack. We will have to know how each test performed. May be it's not necessary.
                    foreach my $test (@tests) {
                        $self->{status}{$host}{$test} =
                          (!$exit_status) ? "PASS" : "FAIL";
                    }

                    # The process died. Check if exit 0 or 1
                    $Log->debug("exit status for client $host process: " . $?);
                    if (!$exit_status) {
                        push(@goods, $host);
                    } else {
                        push(@bads, $host);
                    }
                    delete $pids{$host};
                } ## end if ($retpid > 0)
            } ## end foreach my $host (keys %pids)

            # print status in 2 mins interval, not frequent
            my $now = time();
            if (!$prev || (($now - $prev) > 120)) {
                $Log->comment(  "Completed goods:"
                              . scalar(@goods)
                              . ", bads:"
                              . scalar(@bads)
                              . ". Not completed:"
                              . scalar(keys %pids));
                $prev = $now;
            }
            sleep 2;
        } ## end while (keys %pids)
        $Log->comment("All client verification processes returned.");
        $Log->debug(  "Unhealthy clients found: "
                    . scalar(@bads)
                    . ". They are : "
                    . join(" ", @bads))
          if (@bads);
    } ## end if (!$status_available...

    return (\@goods, \@bads);
} ## end sub clients_status

=head2 rsh_check

    $Ha->rsh_check(client => $client);

C<rsh_check> returns 1 if rsh check succeeds for the client. Else returns 0.

=over

=item client

(Required) The client to be checked against.

=back

=cut

sub rsh_check {
    my ($self, %opts) = @_;

    my $fh = Tharn::host($opts{host});
    my $conn =
      Tharn::connect("$opts{host}#rsh-persay;username=root;password=netapp");
    my (@output, $output);
    eval {
        @output = Tharn::say($conn, 'uname -a');
        $output = join(" ", @output);
    };
    if ($@ || ($output !~ /$opts{host}/)) {
        $Log->comment("rsh seems to have failed ($opts{host}):\n$output");
        return 0;
    } else {
        $Log->comment("rsh passed : $output");
        return $output;
    }
} ## end sub rsh_check

=head2 ping_check

    $Ha->ping_check(client => $client);

C<rsh_check> returns 1 if ping check succeeds for the client. Else returns 0.

=over

=item client

(Required) The client to be checked against.

=back

=cut

sub ping_check {
    my ($self, %opts) = @_;

    # Ping the host every 10 seconds until it comes back
    while ($opts{endtime} > time()) {
        my $cmd    = "ping -c 1 " . $opts{host};
        my $output = `$cmd 2>&1`;
        if ($output =~ /(alive|\s0% packet loss)/) {
            $Log->comment("Ping passed");
            return 1;
        } else {
            sleep(10);
        }
    } ## end while ($opts{endtime} > time...

    $Log->comment("ping failed");

    # failed
    return 0;
} ## end sub ping_check

=head2 staf_check

    $Ha->staf_check(client => $client);

C<staf_check> returns 1 if rsh check succeeds for the client. Else returns 0.

=over

=item client

(Required) The client to be checked against.

=back

=cut

sub staf_check {
    my ($self, %opts) = @_;

    my $cmd      = "/usr/software/test/bin/staf " . $opts{host} . " ping ping";
    my $output   = qx($cmd 2>&1);
    my $exitcode = $? >> 8;
    $Log->comment("$cmd:\n$output");

    return ($exitcode == 0) && ($output =~ /PONG/);
} ## end sub staf_check

sub on_client {
    my ($self, %options) = @_;
    my $host    = $options{host};
    my $reboot  = $options{reboot};
    my $retries = $options{retries};
    my $timeout = $options{timeout};
    my $rpower  = 0;
    my $fail    = 1;

    open_as_stdio($options{logfile});

    while ($retries--) {
        my $endtime = time() + $timeout;

        # If either ping or rsh failure means, our only option to rpower
        if ($reboot) {
            if (   $self->ping_check(host => $host, endtime => $endtime)
                && $self->rsh_check(host => $host)) {
                $Log->comment("Trying soft reboot");
                eval { $self->client_reboot(host => $host); };
                if ($@) {
                    $rpower = 1;
                    $Log->comment("reboot failed: $@");
                }
            } else {
                $rpower = 1;
            }
        } ## end if ($reboot)

        # Try rpower
        if ($rpower) {
            $Log->comment("Trying rpower: $@");
            eval { $self->client_rpower(host => $host); };
            $Log->comment("rpower failed: $@") if ($@);
        }
        sleep 60 if ($reboot || $rpower);
        $rpower  = 0;                   # Reset it back
        $endtime = time() + $timeout;

        while ($endtime > time()) {
            $Log->comment(  "Waiting for tests to pass for :"
                          . ($endtime - time())
                          . " seconds.");
            my $pass = 0;
            foreach my $test (@{$options{tests}}) {
                if ($test eq "ping") {
                    $pass++
                      if (
                         $self->ping_check(host => $host, endtime => $endtime));
                } elsif ($test eq "rsh") {
                    $pass++ if ($self->rsh_check(host => $host));
                } elsif ($test eq "staf") {
                    $pass++ if ($self->staf_check(host => $host));
                } else {
                    $Log->comment(
                           "Test $test is not supported. Exiting with failure");
                    $fail = 1;
                    goto END;
                }
            } ## end foreach my $test (@{$options...
            if ($pass == scalar(@{$options{tests}})) {
                $Log->comment("All tests passed. Exiting with success");
                $fail = 0;
                goto END;
            }
            sleep 10;
        } ## end while ($endtime > time())
        $Log->comment("An attempt failed. Retry count remaining: " . $retries);
    } ## end while ($retries--)
  END:
    $Log->comment("Exiting with return code $fail");

    # Make sure the inherited END{} blocks are never used, and
    # inherited objects are never DESTROYed by us.
    POSIX::_exit($fail);
} ## end sub on_client

sub open_as_stdio {
    my ($filename) = @_;
    open(STDOUT, ">>$filename")
      or
      $Log->log("Unable to redirect STDOUT to $filename: $!", level => "error")
      && POSIX::_exit(1);
    binmode(STDOUT, ":utf8");
    open(STDERR, ">&STDOUT")
      or $Log->log("Unable to redirect STDERR to STDOUT: $!", level => "error")
      && POSIX::_exit(1);
    binmode(STDERR, ":utf8");
} ## end sub open_as_stdio

sub client_reboot {
    my ($self, %opts) = @_;
    my @output;

    my $fh = Tharn::host($opts{host});
    my $conn =
      Tharn::connect("$opts{host}#rsh-persay;username=root;password=netapp");

    my $cmd =
      ($fh->hosttype() =~ /sunos/i) ? "reboot" : "/sbin/shutdown -r now";
    eval { @output = Tharn::say($conn, $cmd, '-blind'); };
    my $result = ($@) ? 0 : 1;
    my $output = join(" ", @output);

    $Log->comment("$cmd: $output, $@");

    return $result;
} ## end sub client_reboot

sub client_rpower {
    my ($self, %opts) = @_;
    $opts{host} =~ s/-e0$//g;    # Remove -e0 suffix if any

    my $cmd      = "/x/eng/rtp/scripts/rpower " . $opts{host};
    my $output   = qx($cmd 2>&1);
    my $exitcode = $? >> 8;
    print "$cmd: $output \n";

    return ($exitcode == 0);
} ## end sub client_rpower

1;
