# $Id$
## @summary Common library for SSQA to keep hosts reservation code
## @author gvg@netapp.com
## @pod here

=head1 NAME

Utils::HostManager.pm - Common library for SSQA to manage hosts

=head1 DESCRIPTION

This package defines routines to reserve/unreserve hosts from the pool
of hosts.

=head1 SYNOPSIS

    use Utils::HostManager;

    my $HostM = new Utils::HostManager;
    my @hosts = $HostM->get_clients(); # When CLIENTSPEC is set
     (OR)
    my @hosts = $HostM->reserve_clients(reqid      => "reqid1",
                                        configtype => "linux",
                                        count      => 5);
    $HostM->free_clients();
      (OR)
    $HostM->release_hosts();


=head1 EXAMPLE

    *** EXAMPLE #1 - Get clients and free them in stages ***
    *** Reservation is persistent even across reboots, hence explicit release required ***
    *** Uses CLIENT / CLIENTFILE / CLIENTSPEC (and) NDISPD ***

    use Utils::HostManager; # This brings CLIENT, CLIENTFILE, CLIENTSPEC and
                            # NDISPD into name space. Hence setting
                            # appropriate values will be sufficient like below.
                            # CLIENT - Comma separated clients list (OR)
                            # CLIENTFILE - The file contains host names (OR)
                            # CLIENTSPEC (and) NDISPD
                            # E.g, CLIENTSPEC="hosttype=linux,count=10"

    my $HostM = new Utils::HostManager ();

    # get clients based on CLIENTSPEC (or CLIENT or CLIENTFILE)
    my @clients = $HostM->get_clients(persistent  => 1);

    my @part_clients = ["$clients[0]", "$clients[1]"];
    $HostM->free_clients(hosts => \@part_clients);

    # Free all remaining clients (Otherwise, the reservation stays until cleared)
    $HostM->free_clients();


    *** EXAMPLE #2 - Reserve hosts incrementally and release them in one call ***
          *** Reservation is Persistent only for this process life time  ***

    use Utils::HostManager;

    my $HostM = new Utils::HostManager (ndispd_addr => "172.X.X.X:61000");

    # Reserve 10 of 'windows' hosts with the id 'MyReservation001'
    # persistent is not set, so that reservation automatically released, when
    # this process dies
    my @win_hosts = $HostM->reserve_hosts(reqid       => "MyReservation001"
                                          hosttype    => "windows",
                                          count       => 10,
                                          persistent  => 0);

    # Reserve 5 more Linux hosts
    my @linux_hosts = $HostM->reserve_hosts(reqid       => "MyReservation002"
                                            hosttype    => "linux",
                                            count       => 5,
                                            persistent  => 0);

    # Release all hosts (or On the exit of this process, released automatically)
    $HostM->release_hosts();

    *** EXAMPLE #3 - Reserve hosts persistently and unreserve them ***
    *** Reservation is persistent even across reboots, hence explicit release required ***

    use Utils::HostManager;

    my $HostM = new Utils::HostManager (ndispd_addr => "172.X.X.X:61000");

    # Reserve 10 of 'windows' hosts with the id 'MyReservation001'
    # interactive is set, so that reservation automatically released, when
    # this process dies
    my @win_hosts = $HostM->reserve_hosts(reqid       => "MyReservation001"
                                          hosttype    => "windows",
                                          count       => 10,
                                          persistent  => 1);

    # Release all hosts (Otherwise, the reservation stays until cleared)
    $HostM->relase_hosts(reqid       => "MyReservation001");


=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Created $Date: 2008/04/22 $

=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved

=head1 ROUTINES

=cut

package Utils::HostManager;
use strict;
use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use Params::Common;    # To bring CLIENT and CLIENTFILE Options here
use File::Basename qw(basename);
use Sys::Hostname;

use ndisp::API;
use NATE::Events qw(eventloop_drain);
#use IO::Select;

use Data::Dumper;

#
# Add library for access to ndisp API and NATE packages.
#
BEGIN { push @INC, "/x/eng/localtest/noarch/lib"; }
BEGIN { push @INC, "/x/eng/localtest/noarch/share/nate/latest/lib"; }

use Options;
my $Options = new Options;
my $Opts;    # Hash

use TracePkg;
my $Trace = new TracePkg;

$Options->parameters(
    definitions => [
        {name        => "NDISPD",
         description => "IP:PORT of ndispd that is running to manage hosts",
         spec        => "s",
         default     => undef,},
        {name => "CLIENTSPEC",
         description =>
           "Client spec in the form of type and count. E.g, {hosttype=linux;count=5,hosttype=solaris;count=2}",
         spec    => "s",
         default => undef,},
        {name        => "USER",
         description => "Bring USER from Shell to here",
         spec        => "s",},]);

=head2 new

    my $HostM = new Utils::HostManager(ndispd_addr => "172.16.X.X:61000");

C<new> creates HostManager object with provided ndispc API

=over

=item ndispd_addr

(Required) The IP address and port pair of Ndispd daemon. It can be set in NATE param NDISPD as well.

=back

=cut

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;

    # Process Options
    $Options->process();
    $Opts = $Options->get_hash(
        option => [
            qw(
              NDISPD
              CLIENT
              CLIENTSPEC
              USER
              )]);

    my $self = {reqids      => {},
                ndispc      => undef,
                ndispd_addr => $Opts->{NDISPD},
                @_,};
    $Trace->entry();
    bless($self, $class);

    $self->{ndispc} = new ndisp::API($self->{ndispd_addr})
      if ($self->{ndispd_addr});

    return $self;
} ## end sub new

#################################################################
# Routines that can be used by Tests directly
#################################################################

=head2 get_clients

    my @hosts = $HostM->get_clients();

C<get_clients> returns the list of hosts that requested in one of the following order:
- Returns from ndispd when CLIENTSPEC and NDISPD are set
- Returns clients from CLIENT or CLIENTFILE

=over

=item persistent

(Optional) Whether or not reservation should be persistent across processes.

=back

=cut

sub get_clients {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {hosttype   => {type => SCALAR, default => undef},
                      count      => {type => SCALAR, default => undef},
                      persistent => {type => SCALAR, default => 1},},);

    my @clients;

    # If ndispd_addr is set, then call reserve_hosts, otherwise return CLIENT
    if ($self->{ndispd_addr} && $Opts->{CLIENTSPEC}) {
        my (undef, $file, undef) = caller;
        $file = basename($file);
        $file =~ s/\.\w*//g;
        my $id = 0
          ; # If request has to be made more than once, index it to seperate them

        foreach my $spec (split(/,/, $Opts->{CLIENTSPEC})) {
            $opts{reqid} = $file . "-" . $$ . "." . $id;
            print "SPEC is: $spec ($opts{reqid}) \n";
            my @optsets = split(/;/, $spec);
            foreach my $optset (split(/;/, $spec)) {
                my ($param, $value) = split(/=/, $optset);
                $opts{$param} = $value;
            }
            my @cli = $self->reserve_hosts(%opts);
            push(@clients, @cli) if (scalar(@cli));
            $id++;
        } ## end foreach my $spec (split(/,/...
    } elsif (@{$Opts->{CLIENT}}) {
        @clients = @{$Opts->{CLIENT}};
    } else {
        die
          "Clients could not be found either of CLIENT, CLIENTFILE and CLIENTSPEC/NDISPD";
    }

    return $self->kcsort(@clients);
} ## end sub get_clients

=head2 free_clients

    $HostM->free_clients(hosts => \@hosts);

C<free_clients> releases clients if reserved from ndispd.

=over

=item hosts

(Optional) ARRAYREF. List of clients to be released.

=back

=cut

sub free_clients {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {hosts => {type => ARRAYREF, default => undef},},);

    # If ndispd_addr is set, then call release_hosts, otherwise return CLIENT
    if ($self->{ndispd_addr}) {
        $self->release_hosts(%opts);
    } else {

        # XXX - ???
        # What should we do here?

    }
} ## end sub free_clients

=head2 get_reserved_hosts

    my @hosts = $HostM->get_reserved_hosts();

C<get_reserved_hosts> returns the list of hosts that are reserved so far.

=over

=item reqid

(Optional) To selectively get the hosts reserved by a specific ID.

=item user

(Optional) To selectively get the hosts reserved by a user.

=item reason

(Optional) To selectively get the hosts reserved by a reason

=item strict

(Optional) Match the entire reason string

=back

=cut

sub get_reserved_hosts {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {reqid  => {type => SCALAR|UNDEF, default  => undef},
                      user   => {type => SCALAR|UNDEF, optional => 1},
                      reason => {type => SCALAR|UNDEF, default  => undef},
                      strict => {type => SCALAR|UNDEF, default  => undef},},);

    my @reqids;
    if ($opts{reqid}) {
        push(@reqids, $opts{reqid});
    } else {
        push(@reqids, keys %{$self->{reqids}});
    }

    my (@hosts, %args);

    if ($opts{reason}) {
        $args{reason} = $opts{reason};
    }
    $args{user}   = $opts{user}   if ($opts{user});
    $args{strict} = $opts{strict} if ($opts{strict});

    do {
        $args{reqid} = (@reqids) ? pop @reqids : undef;
        my @reqhosts = $self->_get_reserved_hosts(%args);
        push(@hosts, @reqhosts);
    } while (@reqids);

    return $self->kcsort(@hosts);
} ## end sub get_reserved_hosts

=head2 reserve_hosts

    my @hosts = $HostM->reserve_hosts(reqid      => "someid1",
                                      hosttype   => "windows",
                                      count      => 10);

C<reserve_hosts> reserves and returns the list of hosts that are reserved
with this request.

=over

=item reqid

(Required). Reserve hosts with this ID

=item hosttype

(Required). The config type of the hosts to be reserved

=item hosts

(Required). ARRAYREF. List of hosts to be reserved.

(One of hosttype or hosts to be specified)

=item reason

(Optional). The text to be used as a reason to be reserved.

=item user

(Optional). The username to be used as a owner of the reservation. Default: $ENV{USER}.

=item restime

(Optional). The reservation time to be used for this reservation. Default: localtime().

=item from_host

(Optional). The hostname where this reservation request is made. Default: hostname().

=item count

(Optional). The number of hosts to be reserved. Default 1.

=item persistent

(Optional). Whether reservation should be released automatically. By default they are released on the process exit even in the absense of C<release_hosts> called. Default 0.

=item timeout

(Optional). The length of time to wait to reserve the requested hosts. Default 60 secs.

=back

=cut

sub reserve_hosts {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {reqid    => {type => SCALAR,},
                      hosttype => {type => SCALAR | UNDEF, default => undef},
                      hosts    => {type => ARRAYREF | UNDEF, default => undef},
                      reason => {type => SCALAR, default => "Reason Not Given"},
                      user       => {type => SCALAR, optional => 1,},
                      restime    => {type => SCALAR, optional => 1,},
                      from_host  => {type => SCALAR, optional => 1,},
                      count      => {type => SCALAR,  default => 1},
                      persistent => {type => BOOLEAN, default => 0},
                      timeout    => {type => SCALAR,  default => 60},},);

    my @hosts;
    my $result = 0;

    my $api = new ndisp::API($self->{ndispd_addr});

    # Make sure hosttype is always provided for non-persistent mode, and either one
    # for persistent mode.
    if ($opts{persistent} && (!$opts{hosttype} && !$opts{hosts})) {
        die("persistent mode must need either hosttype or hosts");
    } elsif (!$opts{persistent} && (!$opts{hosttype})) {
        die("non-persistent mode need hosttype");
    }

    if (!$opts{persistent} && ($opts{hosts})) {
        die("non-persistent mode has hosts no-op");
    }

    $opts{restime} = localtime() if (! $opts{restime});
    $opts{user} = $Opts->{USER} if (! $opts{user});
    $opts{from_host} = hostname() if (! $opts{from_host});

    # XXX - ???
    # Ndisp waits indefinitely, if the request can not be fulfilled at this time
    # This is limited by $opts{timeout} now. Should we report back with meaningful
    # message sooner than waiting for $opts{timeout} seconds?

    if ($opts{persistent}) {
        my %req;
        $req{username} = $opts{user};
        $req{id}       = $opts{reqid};
        $req{reason}   = "REQID=" . $opts{reqid} .
            ",RESTIME=" . $opts{restime} .
            ",FROM_HOST=" . $opts{from_host} .
            ",REASON=" . $opts{reason};


        if ($opts{hosttype}) {
            $req{reserve} = "%{res_config(q($opts{hosttype}*$opts{count}))}";
        } else {

            # hosts list is given
            $req{reserve} = join(" ", @{$opts{hosts}});
        }

        $result = $self->_write_request(api        => $api,
                                        req        => \%req,
                                        persistent => 1,
                                        timeout    => $opts{timeout});
    } else {
        my %req;
        $req{username} = $opts{user};
        $req{id}       = $opts{reqid};
        $req{jobname}  = $opts{reqid} . "_" . $Opts->{USER};
        $req{command} =
          "echo %{res_config(q($opts{hosttype}*$opts{count}))}; sleep 31536000";

        $result = $self->_write_request(api        => $api,
                                        req        => \%req,
                                        persistent => 0,
                                        timeout    => $opts{timeout});
    } ## end else [ if ($opts{persistent})

    if (!$result) {
        $self->_delete_request(reqid => $opts{reqid});
        die
          "Request $opts{reqid} is not complete and successful after $opts{timeout} seconds, hence the request is deleted.";
    }

    # Make sure request is successful and complete
    $self->_wait_req_complete(reqid   => $opts{reqid},
                              timeout => $opts{timeout});

    @hosts = $self->_get_reserved_hosts(reqid => $opts{reqid},);
    if (!@hosts) {
        die
          "Could not reserve the requested $opts{count} hosttypes of \"$opts{hosttype}\" for $opts{timeout} seconds\n";
    }
    $self->{reqids}{$opts{reqid}} = $api;

    return $self->kcsort(@hosts);
} ## end sub reserve_hosts

=head2 release_hosts

    my @hosts = $HostM->release_hosts(reqid => "someid1");

C<release_hosts> releases the list of hosts that are reserved so far.
reqid and hosts are mutually exclusive.

=over

=item reqid

(Optional) To selectively release the hosts reserved by a specific ID.

=item hosts

(Optional) To selectively release the hosts. arrayref.

=item user

(Optional) To selectively release the hosts reserved by this user. No default.

=back

=cut

sub release_hosts {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {reqid => {type => SCALAR,           default => undef},
                      hosts => {type => ARRAYREF | UNDEF, default => undef},
                      user  => {type => SCALAR, optional => 1},},);

    die "reqid and hosts are mutually exclusive, but passed options are :". Dumper(\%opts) if ( $opts{reqid} && $opts{hosts} );

    my (@reqids, @hosts);
    @hosts = @{$opts{hosts}} if ($opts{hosts});
    push (@reqids, $opts{reqid}) if ($opts{reqid});
    push(@reqids, keys %{$self->{reqids}}) if (! @reqids);

    if ($opts{user} && @hosts) {
        # If user is specified, make sure hosts specified are
        # reserved only by this user.
        my @thosts = $self->get_reserved_hosts(user => $opts{user});
        my %total_hosts = map { $_ => 1 } @thosts;
        foreach my $rhost (@hosts) {
            die "Host $rhost is not reserved by user $opts{user}. Please verify the hosts list"
                if (! $total_hosts{$rhost} );
        }
        $self->_delete_hosts(hosts => \@hosts);
    } elsif ( @hosts) {
        $Trace->debug(msg => "Releasing hosts : " . join(",", @{$opts{hosts}}));
        $self->_delete_hosts(hosts => \@hosts);
    } elsif ( @reqids ) {
        foreach my $reqid (@reqids) {
            # Delete request with id $reqid
            my $req = $self->{ndispc}->call("list requests", ["$reqid"])
                or die "list requests failed: $@\n";
            if ($req->{$reqid}) {
                $self->_delete_request(reqid => $reqid);
            } else {
                # The corresponding request doesn't exist in
                # non-interactive mode, hence we should remove
                # hosts by hosts
                my @hosts = $self->_get_reserved_hosts(reqid => $reqid,);
                $self->_delete_hosts(hosts => \@hosts);
            }

            # Delete from the reqids list
            delete $self->{reqids}{$reqid};
        } ## end foreach my $reqid (@reqids)
    } else {
        die "!user !reqid !hosts = This condition should not have occured";
    }
    sleep 5;    # Wait for NDISPD to update
} ## end sub release_hosts

#################################################################
# A few reporting routines
#################################################################

=head2 list_hosts

    my @hosts = $HostM->list_hosts(config => "someconfig");
    (or)
    my %hosts = $HostM->list_hosts(config => "someconfig", want_value => "hash");

C<list_hosts> returns a list of hosts returned by ndispd by default.

=over

=item hosttype

(Optional) List hosts only specific to hosttype.

=back

=over

=item want_value

(Optional) "hash" or "list". Default is list.

=back

=cut

sub list_hosts {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {config     => {type => SCALAR, default => undef},
                      want_value => {type => SCALAR, default => "list"},},);

    $Trace->debug(msg => "Looking at hosts list");
    my $hostref = $self->{ndispc}->call("list hosts", [])
      or die "list hosts failed: $@\n";

    my @hosts;
    if ($opts{config}) {
        no warnings 'uninitialized';
        @hosts =
          grep { ($hostref->{$_}{hosttype} =~ /$opts{config}/) || ($hostref->{$_}{CONFIGS} =~ /$opts{config}/) } keys %$hostref;
    } else {
        @hosts = keys %$hostref;
    }

    return map { $_ => $hostref->{$_} } @hosts if ($opts{want_value} =~ /hash/);

    return $self->kcsort(@hosts);
} ## end sub list_hosts

=head2 list_reservations

    my @hosts = $HostM->list_reservations(hosttype => "someconfig");

C<list_reservations> returns an array of hosts returned by ndispd.

=cut

sub list_reservations {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {hosttype => {type => SCALAR, default => undef},},);

    $Trace->debug(msg => "Looking at reservations list");
    my $hostref = $self->{ndispc}->call("list reservations", [])
      or die "list reservations failed: $@\n";

    return $hostref;

    #   my @hosts;
    #   if ($opts{hosttype}) {
    #       @hosts = grep($hostref->{$_}{hosttype} =~ /$opts{hosttype}/, keys %$hostref);
    #   } else {
    #       @hosts = keys %$hostref;
    #   }
    #   return @hosts;
} ## end sub list_reservations

sub kcsort {
    my ($self, @orighosts) = @_;
    my %hosts;

    foreach my $host (@orighosts) {
        if ($host =~ m/kc(\d+)b(\d+)/) {
            $hosts{$host} = sprintf("%04d%02d", $1, $2);
        } else {
            $hosts{$host} = 1;
        }
    }

    return sort { $hosts{$a} <=> $hosts{$b} } (keys %hosts);
} ## end sub kcsort

#################################################################
# Internal routines
#################################################################

sub _get_reserved_hosts {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {reqid  => {type => SCALAR|UNDEF, optional => 1},
                      user   => {type => SCALAR|UNDEF, optional => 1},
                      reason => {type => SCALAR|UNDEF, optional => 1},
                      strict => {type => SCALAR|UNDEF, optional => 1},},);
    my @hosts;

    $Trace->debug(msg => "Looking at reservations list");
    my $res = $self->{ndispc}->call("list reservations", [])
      or die "list reservations failed: $@\n";

    return if (ref($res) ne "HASH");    # Otherwise, ndispd returns a string

    $Trace->debug(msg => "Output List Reservations:\n" . Dumper($res));

    my $patt = "REQID=".$opts{reqid} if ($opts{reqid});
    $patt .= ",REASON=".$opts{reason} if ($opts{reason});

    if (!$opts{user}) {
        @hosts = grep($res->{$_}{reason} =~ /$patt/, keys %$res);
        $Trace->debug(msg => "The hosts reserved for $patt are: \n"
                      . Dumper(\@hosts));
    } else {
	#-------------------------------------------
        # Include EOL in the regex if strict is set
	#-------------------------------------------
        $patt .= '$' if ($opts{'strict'});
        if (defined($patt)) {
            @hosts =
                grep($res->{$_}{ownername} eq $opts{user}
                     && $res->{$_}{reason} =~ /$patt/,
                     keys %$res);
        } else {
            @hosts =
                grep($res->{$_}{ownername} eq $opts{user},
                     keys %$res);
        }
        $Trace->debug(
              msg => "The hosts reserved for $opts{user}:$patt are: \n"
                . Dumper(\@hosts));
    }

    return @hosts;
} ## end sub _get_reserved_hosts

sub _wait_req_complete {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {reqid   => {type => SCALAR,},
                      timeout => {type => SCALAR, default => 60},},);
    my $endtime = time() + $opts{timeout};

    my ($req, $status);
    while (time() < $endtime) {
        $req = $self->{ndispc}->call("list requests", ["$opts{reqid}"])
          or die "list requests failed: $@\n";

        $Trace->debug(msg => "list requests: \n" . Dumper(\$req));

        # Otherwise, ndispd returns a string
        die "No requests exists: $req\n" if (ref($req) ne "HASH");

        if ($req->{$opts{reqid}}) {

            # It means, the request was made in non-persistent mode, hence
            # request is still alive.

            $status = $req->{$opts{reqid}}{status};
            if ($status =~ /no charges left/i) {
                return 1;
            } elsif ($status =~ /^Waiting for/i) {

                # This means request has not gotten adequate hosts
                sleep 5;
            } else {
                $Trace->debug(
                           msg => "Ndispd not up-to-date?: \n" . Dumper(\$req));
                sleep 3;
            }
        } else {

            # Not sure, we have to handle persistent mode case here
            $Trace->debug(msg =>
                "No requests found by id $opts{reqid}. Assuming the request was made in persistent mode and returning success"
                );
            return 1;
        }
    } ## end while (time() < $endtime)

    # We are here - Means, we are timed out.
    if ($status =~ /^Waiting for/i) {
        die "Request couldn't get fulfilled after $opts{timeout} seconds\n"
          . Dumper(\$req);
    }
} ## end sub _wait_req_complete

sub _write_request {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {api        => {type => OBJECT,},
                      req        => {type => HASHREF,},
                      persistent => {type => SCALAR,},
                      timeout    => {type => SCALAR, default => 60},
                     },
                     );

    my $result = 0;

    $Trace->debug(msg => "write request: \n" . Dumper(\${opts { req }}));
    my $api = $opts{api};
    my $req = $api->write_request("add request", $opts{req}, interactive => 1)
      or die "Add request failed: $@";

    my $endtime = time() + $opts{timeout};

    while (time() < $endtime) {    # Wait until the job is complete or timed out
        if (!$api->{peer}->perl_eval_is_complete(key => $req)) {
            eventloop_drain(timeout => 5);
            next if (!$api->{peer}->perl_eval_is_complete(key => $req));
        }
        my $reply = $api->read_reply();
        $Trace->debug(msg => "read_reply: \n" . Dumper(\$reply));

        if ($reply->{error}) {
            die "write_request failed: error: $reply->{error}";
        } elsif ($opts{persistent} && $reply->{final}) {

            # Submitted successfully
            $result = 1;
            $Trace->debug(msg => "Request complete: $reply->{data}");
        } elsif (  !$opts{persistent}
                 && $reply->{data} =~ /^Watching new Job/) {

            # Submitted successfully
            $result = 1;
            $Trace->debug(msg => "Request complete: $reply->{data}");
        } elsif ($reply->{final}) {

            # How to handle this?
            $Trace->debug(msg => "How to handle this?: \n" . Dumper(\$reply));
        } else {
            $Trace->debug(msg => "How to handle this?: $reply->{data}");
        }
        last if ($result);
    }

    return $result;
}

sub _delete_request {
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},], {reqid => {type => SCALAR,},},);

    my $res = $self->{ndispc}->call("delete requests", ["/^$opts{reqid}/"]);
    if ($@) {
        die "Delete requests failed: $@";
    } elsif (!$res) {
        die "Ooops, no matching requests found to delete\n";
    }
    $Trace->debug(msg => "$res $opts{reqid} removed\n");
} ## end sub _delete_request

sub _delete_hosts {
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},], {hosts => {type => ARRAYREF,},},);

    foreach my $host (@{$opts{hosts}}) {
        my $res = $self->{ndispc}->call("delete reservations", ["$host"]);
        $Trace->debug(
                  msg => "Deleting $host from reservation: \n" . Dumper(\$res));
        if ($@) {
            die "Delete hosts failed: $@";
        } elsif (!$res) {
            die
              "Ooops, no matching hosts found to delete. Ndispd doesn't allow the hosts that are reserved by a job/request (not by the user)\n";
        }
        $Trace->debug(msg => "$res $host removed\n");
    } ## end foreach my $host (@{$opts{hosts...
} ## end sub _delete_hosts

################## End #########################################

1;
