# $Id $
## @summary Helper methods for Utils::HostManager
## @author gvg@netapp.com
## @pod here

=head1 NAME

Utils::HostManager::Helper - Helper methods for Utils::HostManager

=head1 DESCRIPTION

This package defines routines which can be used by Utils::HostManager.

=head1 SYNOPSIS

use Utils::HostManager::Helper;

=head1 ROUTINES

=cut

package Utils::HostManager::Helper;
use strict;
use POSIX qw(ceil strftime);
use Tharn;
$Log->comment("Starting " . __FILE__);
use Params::Validate qw(validate UNDEF SCALAR OBJECT HASHREF ARRAYREF);
use Module::Common qw(validate_mixed);
use Options;
use Data::Dumper;
use Utils::HostManager;
use Utils::Dump;
use File::Basename;
use Time::Local;
use Sort::Key::Natural qw(natsort);
use NATE::Inc qw(find_file);

our $VERSION = 1.00;
our @ISA     = qw (Exporter);
our @EXPORT  = qw (
    reserve_clients
    retrieve_clients
    store_and_free_clients
    get_max_clients
);

my $Options = new Options;
$Options->parameters(
    definitions => [
        {   name    => 'FREE_UNUSED_CLIENTS',
            spec    => "s",
            default => 0,
            description =>
                'Free reservations for clients that are not needed',
        },
        {   name => 'MAX_CLIENTS',
            spec => "s",
            description =>
                'How logn HostManager.pm waits while reserving clients',
        },
        {   name    => 'RESERVATION_TIMEOUT',
            spec    => "s",
            default => 60,
            description =>
                'How logn HostManager.pm waits while reserving clients',
        },
        {   name => 'RESERVATION_REASON',
            spec => "s",
            description =>
                'The string to be added to the reason when reserving clients via NDISP',
        },
        {   name        => 'TOP_LOGDIR',
            spec        => "s",
            description => 'Import TOP_LOGDIR from NATE params',
        },
    ]
);

=head2 get_max_clients

 use Utils::HostManager::Helper qw(get_max_clients);
 get_max_clients(clients => \@clients);

=item clients

(Required) Arrayref of clients

=back

=cut

sub get_max_clients {

    $Log->enter();

    $Options->process();
    my $Options_hash = $Options->get_hash();
    my $max_clients  = 0;

    if (  !$Options_hash->{'MAX_CLIENTS'}
        || $Options_hash->{'MAX_CLIENTS'} == 0) {
        $Log->debug("Skipping client calculations due to MAX_CLIENTS=0");
        $Log->exit();
        return $max_clients;
    }

    my @clients;
    if ($Options_hash->{'MAX_CLIENTS'} =~ /(\d+)%/) {
        my $max_client_percent = $1;
        if (-e "$Options_hash->{'TOP_LOGDIR'}/total_clients.dat") {
            our $DATA;
            require "$Options_hash->{'TOP_LOGDIR'}/total_clients.dat";
            @clients = @{$DATA->{'CLIENT'}};
            $Log->comment("Calculating MAX_CLIENTS from " 
                    . @clients
                    . " clients retrieved from "
                    . "TOP_LOGDIR/total_clients.dat");
            $max_clients = ceil((@clients * $max_client_percent) / 100);
        } else {
            die "TOP_LOGDIR/total_clients.dat required to calculate "
                . "MAX_CLIENT percentages";
        }
    } elsif ($Options_hash->{'MAX_CLIENTS'} =~ /,/) {
        my @m = split(/,/, $Options_hash->{'MAX_CLIENTS'});
        $max_clients = 0;
        foreach my $c (@m) { $max_clients += $c; }
        $Log->comment("Using multiple MAX_CLIENTS values, "
                . "$Options_hash->{'MAX_CLIENTS'}, for "
                . " a total of $max_clients");

    } elsif ($Options_hash->{'MAX_CLIENTS'}) {
        $max_clients = $Options_hash->{'MAX_CLIENTS'};
    }

    $Log->comment("Using MAX_CLIENTS value of $max_clients");
    $Log->exit();

    return $max_clients;
} ## end sub get_max_clients

=head2 reserve_clients

 use Utils::HostManager::Helper qw(reserve_clients);
 reserve_clients(script => "cifs_combo");

C<reserve_clients> Retrieve clients from stored files or via HostManager

=over

=item obj

(Required) The HostManager object

=back

=cut

sub reserve_clients {

    $Log->enter();

    $Options->process();
    my $Options_hash = $Options->get_hash();
    my (%opts) = validate(@_, {obj => {type => OBJECT, optional => 0},});

    my $date = strftime("%y%m%d%H%M%S", localtime());
    my (@clients, $user);
    
    if ($Options_hash->{'NDISPD'} && $Options_hash->{'CLIENTSPEC'}) {
        $user = `id`;
        chomp($user);
        $user =~ s/uid=\d+\(([^\)]+)\).*/$1/;

        if ($Options_hash->{'RESERVATION_REASON'}) {
            $Log->comment("Searching for clients reserved by $user "
                    . "with reason \"$Options_hash->{'RESERVATION_REASON'}\""
            );
            @clients = $opts{'obj'}->get_reserved_hosts(
                'user'   => $user,
                'reason' => $Options_hash->{'RESERVATION_REASON'},
                'strict' => 0
            );
            if (@clients) {
                $Log->comment("Using " 
                        . @clients
                        . " clients previously reserved by $user with reason"
                        . " \"$Options_hash->{'RESERVATION_REASON'}\"");
            }
        }

        if (!@clients) {
            $Log->comment("No clients found reserved by $user with a "
                    . "reason \"$Options_hash->{'RESERVATION_REASON'}\""
                    . ", reserving new clients");
            my %host_manager_options;
            my $id = 0;
            $host_manager_options{'persistent'} = 1;
            $host_manager_options{'timeout'}
                = $Options_hash->{'RESERVATION_TIMEOUT'};
            foreach my $spec (split(/,/, $Options_hash->{'CLIENTSPEC'})) {
                $host_manager_options{'reqid'} = $date . $$ . $id;
                $host_manager_options{'reason'}
                    = $Options_hash->{'RESERVATION_REASON'};

                foreach my $optset (split(/;/, $spec)) {
                    my ($param, $value) = split(/=/, $optset);
                    $host_manager_options{$param} = $value;

                }

                my @reserved
                    = $opts{'obj'}->reserve_hosts(%host_manager_options);
                push(@clients, @reserved) if (scalar(@reserved));
                $id++;
            }

            if (@clients) {
                $Log->comment("Using " 
                        . @clients
                        . " clients reserved by $user with reason"
                        . " \"$Options_hash->{'RESERVATION_REASON'}\"");
            }
        }
    } elsif ($Options_hash->{CLIENTFILE}) {

        my $file = find_file($Options_hash->{'CLIENTFILE'});
        if (!$file) {
            die script_result(
                "Cannot find CLIENTFILE ($Options_hash->{'CLIENTFILE'})");
        }
        @clients = read_client_file($file);
    } elsif ($Options_hash->{CLIENTS}) {
	@clients = @{$Options_hash->{CLIENTS}}
    } elsif ($Options_hash->{CLIENT}) {
	@clients = @{$Options_hash->{CLIENT}}
    }

    $Log->exit();

    if (!@clients) {
        die "Clients could not be found either by"
            . " CLIENT, CLIENTFILE and CLIENTSPEC/NDISPD";
    } else {
        return natsort(@clients);
    }

} ## end sub reserve_clients

=head2 retrieve_clients

 use Utils::HostManager::Helper qw(retrieve_clients);
 retrieve_clients(script => "cifs_combo");

C<retrieve_clients> Retrieve clients from stored files or via HostManager

=over

=item options

(Required) The options hashref

=item script

(Required) The script name

=back

=cut

sub retrieve_clients {

    $Log->enter();

    $Options->process();
    my $Options_hash = $Options->get_hash();

    my (%opts) = validate(
        @_,
        {   options => {type => HASHREF | UNDEF, optional => 1},
            script  => {type => SCALAR,          optional => 0}
        }
    );

    my @clients;
    if ($opts{'script'}
        && -e "$Options_hash->{'TOP_LOGDIR'}/$opts{'script'}_clients.dat") {

        # Retrieve previously reserved clients from script_clients.dat
        our $DATA;
        require "$Options_hash->{'TOP_LOGDIR'}/$opts{'script'}_clients.dat";
        $Log->comment("Retrieving "
                . @{$DATA->{'CLIENT'}}
                . " configured clients from $opts{'script'}_clients.dat");
        @clients = @{$DATA->{'CLIENT'}};
    } elsif (-e "$Options_hash->{'TOP_LOGDIR'}/avialable_clients.dat") {

        # Retrieve unused clients from the available_clients.dat
        our $DATA;
        require "$Options_hash->{'TOP_LOGDIR'}/avialable_clients.dat";
        $Log->comment("Retrieving "
                . @{$DATA->{'CLIENT'}}
                . " available clients from TOP_LOGDIR/avialable_clients.dat");
        @clients = @{$DATA->{'CLIENT'}};
    } else {

        # Get clients based on CLIENTSPEC or CLIENT or CLIENTFILE
        my $Host = new Utils::HostManager(
            ndispd_addr => "$Options_hash->{'NDISPD'}");
        @clients = reserve_clients('obj' => $Host);

        if (!-e "$Options_hash->{'TOP_LOGDIR'}/total_clients.dat") {

            # Save the original clients so we can calculate averages
            $Log->comment("Saving " 
                    . @clients
                    . " clients to "
                    . "\$TOP_LOGDIR/total_clients.dat");
            my %data;
            $data{'CLIENT'} = \@clients;
            write_dump(
                'file' => "$Options_hash->{'TOP_LOGDIR'}/total_clients.dat",
                'creator'  => $opts{'script'},
                'var_name' => 'DATA',
                'var_ref'  => \%data
            );
        }
    }

    $Log->exit();

    if (!@clients) {
        die 'Unable to retrieve clients';
    } else {
        return @clients;
    }

} ## end sub retrieve_clients

=head2 store_and_free_clients

 use Utils::HostManager::Helper qw(store_and_free_clients); 
 store_and_free_clients(script => "cifs_combo");

C<store_and_free_clients> Retrieve clients from stored files or via HostManager

=over

=item options

(Required) The options hashref

=item script

(Required) The script name

=back

=cut

sub store_and_free_clients {

    $Log->enter();

    $Options->process();
    my $Options_hash = $Options->get_hash();

    my (%opts) = validate(
        @_,
        {   options => {type => HASHREF | UNDEF, optional => 1},
            script  => {type => SCALAR,          optional => 1},
            total   => {type => ARRAYREF,        optional => 1},
            used    => {type => ARRAYREF,        optional => 1},
        }
    );

    # Check to see if we have unused clients that we can free
    $Log->debug("Used:" . @{$opts{'used'}} . " Total:" . @{$opts{'total'}});

    if (@{$opts{'used'}} < @{$opts{'total'}}) {

        my @not_used;
        for (my $i = 0; $i < @{$opts{'total'}}; $i++) {
            my $c = ${$opts{'total'}}[$i];
            if (!grep(/$c/, @{$opts{'used'}})) {
                push(@not_used, $c);
            }
        }
        my %data;
        if (   $Options_hash->{'NDISPD'}
            && $Options_hash->{'CLIENTSPEC'}
            && $Options_hash->{'FREE_UNUSED_CLIENTS'} == 1
            && @not_used) {

            # Free unused clients. Normally the last script to use clients
            # does this.
            $Log->comment("Freeing " . @not_used . " unused clients");
            my $Host = new Utils::HostManager(
                ndispd_addr => "$Options_hash->{'NDISPD'}");
            $Host->free_clients(hosts => \@not_used);

        } else {

            # Save the unused clients so the next script can use them
            $Log->comment("Saving "
                    . @not_used
                    . " available clients "
                    . "to TOP_LOGDIR/avialable_clients.dat");
            $Options_hash->{'CLIENT'} = \@{$opts{'used'}};
            $data{'CLIENT'} = \@not_used;
            write_dump(
                'file' =>
                    "$Options_hash->{'TOP_LOGDIR'}/avialable_clients.dat",
                'creator'  => $opts{'script'},
                'var_name' => 'DATA',
                'var_ref'  => \%data
            );

        }

    }

    # Save the used clients so the script can use them later
    if (@{$opts{'used'}} && $opts{'script'}) {
        my %data;
        $Log->comment("Saving "
                . @{$opts{'used'}}
                . " used clients "
                . "to TOP_LOGDIR/$opts{'script'}_clients.dat");
        $data{'CLIENT'} = \@{$opts{'used'}};
        write_dump(
            'file' =>
                "$Options_hash->{'TOP_LOGDIR'}/$opts{'script'}_clients.dat",
            'creator'  => $opts{'script'},
            'var_name' => 'DATA',
            'var_ref'  => \%data
        );

    }

    $Log->exit();
    return @{$opts{'used'}};
} ## end sub store_and_free_clients

sub read_client_file {

    my $file = shift;
    $Log->enter();
    my @clients;
    my $ranges = $Options->get(option => "CLIENTFILE_RANGE");
    $Log->debug("Processing CLIENTFILE for ranges");

    my (@orig_lines, @lines, $total);

    # Save original contents of the file in an array
    if (open(IF, $file)) {

        # remove the comment lines so that the @orig_lines array consists
        # solely of client names
        while (<IF>) {
            next if (/^\s*#/);
            push(@orig_lines, $_);
        }
        $total = scalar(@orig_lines);
        close(IF);
    } ## end if ( open( IF, $file ))

    # Filter out the required range of clients
    foreach my $range (@$ranges) {
        my ($start, $end);
        if ($range =~ /%/) {

            # If % is specified, if 'start %' has values greater than 1,
            # substract by 1, so you will not miss out any clients. But
            # increment the real value of 'start', so that the same client
            # will not be reused. This is based on the assumption, that user
            # does not repeat the numbers. If he does, then it's most likely
            # that's what he wants. E.g "1-30%,30%-40" makes 1% of the lines
            # in 30% range will be redundant.

            $range =~ s/%//;
            my ($startp, $endp) = split("-", $range);
            if ($startp <= 1) {
                $startp = 0;
            } else {
                $startp = $startp - 1;
            }
            $start = int($startp * $total / 100);
            $end   = int($endp * $total / 100);
            if ($endp != 0 && $end == 0) {

                # Set atleast one client, if 0 is not intended.
                $end = 1;
            }
            $start = $start + 1;
        } else {

            # No % given, so straight forward to just play with the numbers
            ($start, $end) = split("-", $range);
        }
        my @temp_lines = @orig_lines;
        $start = $start - 1 if ($start >= 1);
        $end   = $end - $start;
        $end   = 0 if ($end < 0);
        if ($end == 0) {
            push(@lines, splice(@temp_lines, $start));
        } else {
            push(@lines, splice(@temp_lines, $start, $end));
        }
    } ## end foreach my $range (@$ranges)

    # Save the filtered clients in to CLIENT
    foreach (@lines) {
        chomp;
        next if ($_ eq '');
        push(@clients, $_);
    }

    $Log->exit();
    return @clients;
} ## end sub read_client_file
1;
