# $Id$
## @summary Common library for SSQA to help manage IP for clients and filers
## @author dascenj@netapp.com
## @pod here

=head1 NAME

Utils::IPManager.pm - Common library for SSQA to administer IP addresses 

=head1 DESCRIPTION

This package defines routines to manage IP address ranges, vlans, netmasks, etc 

=head1 SYNOPSIS

    use Utils::HostAdmin;
    my $ipManager = new Utils::IPManager;

=head1 EXAMPLES 

=head2 Create Network and Address objects for the values in SRCIP and DESTIP
     
=over

=item *

C<< $ipManager->process_command_line(); >>

=back

=head2 Get all the existing network Objects 

=over

=item * 

C<< my @nets = $ipManager->networks(); >>

=item * 

C<< my $netObj = $nets[0]; >>

=back


=head2 Get source networks, or get destination networks

=over

=item * 

C<< my $netObj = $ipManager->source_networks() or my $netObj = $ipManager->source_networks() >>

=back

=head2 Assign the next available IP (returns address object) 

=over

=item *

C<< my $addrObj = $netObj->assign_next_address() >>

=back

=head2 Get all the unassigned (available) address objects on this network

=over

=item *

C<< my @addrs = $netObj->unassigned_addresses() >>

=back

=head2 Get the netmask and gateway values for the given network or IP address 

=over

=item * 

C<<< my $netmask = $netObj->netmask(); >>> I< or > C<<< my $netmask = $addrObj->netmask(); >>>

=item * 

C<< my $gateway = $netObj->gateway(); >> I< or > C<<< my $gateway = $addrObj->gateway(); >>>

=back


=head1 REQUIREMENTS

    Before the IPManager is able to pass out addresses, it must have IP address 
    objects and network objects created to populate it. There are a few options
    avaiable to add this data to the IPManager object. 

    1. Use the SRCIP and DESTIP values on the command line. By default, calling the new 
       constructor on IPManager will import these values.

    2. use the address_add method to supply data in the same format as SRCIP or DESTIP. 
       (When using this method to input, be sure to set the network as a destination or 
        or source network using the network object methods source_network() or
        destination_network()).

    When using SRCIP and DESTIP on the command line, they variables take the following
    potential forms:

        SRCIP="nk02-126-000+511" (i.e. tharnhostFileName);

        DESTIP=3.2.76.0/18[10]:3.2.64.1,<additionalNetworks> (i.e address/cidr[range]:gateway)

    If using tharnhost file names to submit data to the IPManager, the
    networks/IP ranges should have a tharnhost file configured in the testlab 
    location (i.e. /x/eng/ssqa-testbeds/nate_hosts, but this will vary by testbed)
    At the very least, the file must contain a value for GATEWAY,as there is no
    way to determine this value programatically. The value could also contain 
    other useful information, in the future, such as DNS servers, NIS servers. 

    Here is an example of the contents of one tharn host files mentioned above: 

        default_ip=3.2.126.0
        STARTIP=3.2.126.0
        LASTIP=3.2.127.255
        NETIP=3.2.126.0 + 511
        CONFIGS="gdl,iprange,nodes-ipv4-range"
        NAMEPREFIX=nk02
        GATEWAY=3.2.64.1
        IPCOUNT=512
        DHCPSERVER=NA
        VLANID=3955
        NETMASK=255.255.192.0
        LAB=gdl
        hosttype=nodes-ipv4-range
        DNSDOMAIN=ssqa.gdl.englab.netapp.com
        FAMILY=inet4
        

   Network ID's can be determined with the stress-scale/tools/networkID.pl tool.

=head1 Client/Server Operation

    When the IPManager is operating in a client/server mode, the client 
    IPManager will attempt to gather IP addresses from a centralized 
    repository (the server) before reporting to the user that the network 
    is out of addressses. 

    The IP_PERCENT value can be used to determine how many of the 
    available addresses should be allocated per network where a request 
    is made. For example, in order to request an IP address, the user must 
    already have a network object to request from. So, if you were to make 
    a request for an address from network 1.2.3.0 and the IP_PERCENT value 
    was set to 50, then the client making the request would be allocated 
    50% of the available addresses from the network where the call was made. 

    The IP Manager server will mark IP address as reserved and pass them 
    to the client IPManager. On the client side, they will arrive as 
    available (unreserved) adddresses. In order to mark them as assigned, 
    the user should call on assign_next_address(). 

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Jd (dascenj@netapp.com)
  Created 

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=head1 ROUTINES

=cut

package Utils::IPManager;
use strict;
use warnings;

use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use Tharn qw($Log $Staf host config_result script_result);
use Utils::IPManager::Network;
use Utils::IPManager::Address;
use Utils::IPManager::ServerMessage;
use NATE::STAF::Peer;
use Options;
use Net::IP;
use Data::Dumper;

my $Options = new Options;
my @cmdLineAddrInput = ("SRCIP", "DESTIP");

$Options->parameters(
    definitions => [
        {name      => "SRCIP",
         spec      => "s",
         default   => undef,
         description =>
           "The starting IP/cidr[number of IPs] to be used for clients. ex: 192.168.0.1/8[255]'",
        },
        {name      => "DESTIP",
         spec      => "s",
         default   => undef,
         description =>
           "The starting IP/cidr[number of IPs] to be used for storage controllers. ex: 192.168.0.1/8[255]'",
        },
        {name      => "IP_PERCENT",
         spec      => "s",
         default   => "100",
         description =>
           "The percentage of the total IP addresses from a given network that this test will consume",
        },]);

$Options->process();

=head2 start_ip_manager()

    This is the preferred method for starting up the IPManager. 
    It ensures that the client server connections are made, and 
    informs the ancestor of the servers existence.

    use Utils::IPManager;
    my $ipManager = Utils::IPManager->start_ip_manager;

    Do NOT do this: $self->start_ip_manager();

    This function should never be called in object syntax. This method 
    is intended to serve as a wrapper to the new constructor. It helps 
    to make sure that there is never more than one IPManager server 
    running in a given ntest instance. 

    There could be reasons that a user may want to call new directly. 
    for instance, if the user does not want to operate with the client
    server model, then calling new directly would cause the IPManager 
    to operate autonomously.

=cut


sub start_ip_manager {
    $Log->enter();
    my (%args) = validate(@_, 
                 {start_messenger => {type => SCALAR, optional => 1, default => 0},},);

    my $startMessenger = $args{start_messenger};
    my $serverOnly = $args{server_mode_only};
    my $argsForNew = undef;

    if ($startMessenger) {  
        $Log->trace("The messenger service startup was requested.");
        $argsForNew = 'start_messenger => "1"';

        # set the Ancestor var indicating that the server is starting here
        # catch the error if it already exists, this means it has already been created
        # so get the existing one and use it
        # this is better than get & set. this avoids race condition when multiple gets
        # return VariableDoesNotExist and multiple IPManagers are created
        my $result = eval {
            $Tharn::Staf->var_set(endpoint => $Tharn::Ancestor->endpoint(),
                                    handle => $Tharn::Ancestor->handle(),
                            fail_if_exists => 1,
                                       var => {IPMANAGER_STAF_SERVER => 1});
        };

        # Do the server implementation if there isn't one already
        if ($@) {
            if (grep($@, "AlreadyExists")) {
                $Log->trace("The IPManager server has already been registered.");
            } else {
                die config_result("There was an error in STAF communication. Error $@", stacktace => 0);
            }
        } else {
            $Log->trace("No existing IPManager server is found...This is it now.");
    
            # Subtest the IPManager listener. The server script will put more detailed
            # information about its whereabouts into the ancestor.
            my $subtest = eval{&Tharn::subtest( "Utils/IPManagerServer.thpl", -runid => "IPManager Server", -bg, )};
            die config_result("Failed to subtest the IPManager Server process! Error: $@") if (not $subtest);
        }
    } else { 
        $Log->trace("The messenger service startup was NOT requested.");
    }

    # Start the client IP manager software and return it. 
    my $local_ipManager = Utils::IPManager->new(start_messenger => $startMessenger);

    $Log->exit("Return local IPManager");
    return $local_ipManager;
}

=head2 new

    use Utils::IPManager;
    my $ip = new Utils::IPManager;

    creates IPManager object to store and manage IP addresses.

    Addresses to add can be specifed using DESTIP and SRCIP values just like 
    they would be supplied from the command line. The parameter "input" is also 
    availbe to mimic the behavior of calling address_add(input => text), which 
    will NOT mark the network as a source or destination network. 

=cut
 
sub new {
    $Log->enter();
    my ($invocant, %opts) =
        validate_mixed(@_,
                [{type => SCALAR},],
                {input => {type => SCALAR, optional => 1},
                 start_messenger => {type => SCALAR, optional => 1, default => 0},
                 SRCIP => {type => SCALAR, optional => 1},
                 DESTIP => {type => SCALAR, optional => 1},},);

    $Log->debug("Generating a new IPManager object. Options: " . Dumper(%opts));

    my $startMessenger = $opts{start_messenger};

    my $class = ref($invocant) || $invocant;
    my $self = {};
    bless($self, $class);
    @{$self->{networks}} = ();
    $self->{isServer} = 0;

    my $input = $opts{input} if ($opts{input});
    my $srcip = $opts{SRCIP} if ($opts{SRCIP});
    my $destip = $opts{DESTIP} if ($opts{DESTIP});

    my $processInput = 0;

    # In order to run as a server successfully, there must 
    # be a listener in place. Typically this would be the 
    # IPManagerServer.thpl script. Thus, only start as a 
    # "server" if we are called from the listener thpl. 
    my ($pack, $file, $line) = caller;
    if ($file =~ /IPManagerServer/i) {
        $processInput = 1;
        $self->is_manager_server(set => 1);
    }

    if (not $self->is_manager_server()) {
        $Log->comment("Starting IPManager in local mode");


        if ($startMessenger) { 
            $Log->comment("A messenger object will be created for server communication");
    
            # Create a messenger object
            my $messenger = $self->peer_messenger();
            if (not $messenger) {
                $Log->warn("Failed to connect to an IPManager Server");
            }
            # execute the _empty_networks sub on the 
            # server side.
    
            foreach my $net ($messenger->server_empty_networks()) {
                # Create this through the backdoor. User should NEVER do this 
                # while using the IPManager objects.
                my $netObj = $self->_network_add(cidr      => $net->{cidr},
                                                 version   => $net->{version},
                                                 networkID => $net->{networkID},
                                                 netmask   => $net->{netmask},
                                                 gateway   => $net->{gateway},);
    
                # Since we used the backdoor, we have to manually set this as a source 
                # or destination network. Do this based on the information returned from 
                # the server.
                my $netType = "blank";
                if ($net->{dest}) {
                    $netType = "destination";
                    $netObj->destination_network(set => 1);
                }
    
                if ($net->{source}) {
                    if ($netType =~ /destination/i) {
                        $netType = "destination AND source";
                    } else {
                        $netType = "source"
                    }
                    $netObj->source_network(set => 1);
                }
                $Log->trace("Created a new (but empty) $netType network object with network ID " . $netObj->networkID());
            }
        }

    } else {
        $Log->comment("Starting IPManager in server mode");
         # This is configured as an IPManager server. 
         # Inform the ancestor so the other process can find it.

         # Use the network address as the endpoint 
         my @host = split(/\./, `hostname`);
         my $hostName = ($host[0]) ? $host[0] : `hostname`;
         my $hostObj = Tharn::host($hostName);
         $Log->trace("Marking the IPManager server ready to use...");
         # We have more information about how to contact the IPManager
         # server now. The population of the object is complete. 
         # Put this data up there.
         my $stafInfo;
         $stafInfo->{handle} = $Staf->handle();
         $stafInfo->{endpoint} = $hostObj->net_best_address();

         $Staf->var_set(handle => $Tharn::Ancestor->handle(),
                      endpoint => $Tharn::Ancestor->endpoint(),
                           var => {IPMANAGER_STAF_SERVER => $stafInfo});
         die config_result ("Failed to inform ancestor of IPManager Server Details! Error: $@") if ($@);
   
         $Log->comment("Notified ancestor of server at staf endpoint $stafInfo->{endpoint} handle $stafInfo->{handle}.");

         # Because this is the server, pull in SRCIP and DESTIP from the command line.
         $self->process_command_line();
    }

    if ($processInput) {
        $Log->comment("The IPManager will process all input to new.");
        # If the input args are specifed as part of the new call
        # run them through address add
        if ($input) {
            $self->address_add(input => $input);
        }

        if ($srcip) {
             $self->address_add(SRCIP => $srcip);
        }

        if ($destip) {
              $self->address_add(DESTIP => $destip);
        }

    } else {
        $Log->comment("The IPManager will not process any input from SRCIP or DESTIP");
    }


    $Log->exit('New IPManager =' . Dumper($self));
    return $self;
}

=head2 process_command_line 

    calling this method causes the IPManager to parse SRCIP and DESTIP
    from the command line. Thus managing the IP's supplied by SRCIP and 
    DESTIP command line variables. 

=cut

sub process_command_line {
    $Log->enter();
    my ($self) = @_;
    my $Options_hash = $Options->get_hash();

    # Do the same steps for SRCIP and DESTIP
    # Break apart all the options specified on the command list
    foreach my $clo (@cmdLineAddrInput) {
        my $value = $Options_hash->{$clo};
 
        if ($value) {  
            my $netObj = $self->address_add($clo => $value);
            if (not $netObj) { 
                die config_result("Failed to add $clo IPManager object.");
            }
            $Log->comment("IPManager server added the command line value ($value) of $clo to the list of available addresses.");
        }
    }
    $Log->exit(1);
    return 1;
}

=head2 address_add

    Add a new IP address or range to the IPManager. The format is the same 
    as the SRCIP and DESTIP values on the command line

    my $ipManager = new Utils::IPManager;
    $ipManager->address_add(SRCIP => 192.168.1.1/23[15]:192.168.1.1);

    The user may specify any one of three parameters. SRCIP, DESTIP, and 
    input. Using SRCIP marks the corresponding network object as a "source" 
    network, and DESTIP marks the corresponding network object as a 
    "destination" network. Using the input parameter marks neither of
    these options. 

    Adding a new address will automatically create a new corresponding network 
    object, if one does not exist already

    This routine will return the NETWORK OBJECT of which the new address 
    is a member!

=cut

sub address_add {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},], 
                     {input  => {type => SCALAR, optional => 1}, 
                      SRCIP  => {type => SCALAR, optional => 1},
                      DESTIP => {type => SCALAR, optional => 1},},);

    $Log->debug("address_add is processing input: " . Dumper(%opts));
    my @return;
    my $netObj;
    my $checkForSomeInput = 0;
    foreach my $param (keys(%opts)) { 
        my $input = $opts{$param}; 
        next if (! $input);

        my @entryList = $self->_normalize_input(input => $input);
        # Each comma seperater on the input value creates an "entry"
        foreach my $entry (@entryList) { 
            # Each entry contains network data and an array of addresses
            foreach my $addr (@{$entry->{addresses}}) { 
                # add the addresses
                $netObj = $self->_address_add(address => $addr,
                                              version => $entry->{version},
                                              gateway => $entry->{gateway},
                                              netmask => $entry->{netmask},
                                              cidr    => $entry->{cidr});
            }

            # Mark the network as source or destination if necessary
            if ($param =~ /SRCIP/i) { 
                $netObj->source_network(set => 1);
            
            } elsif ($param =~ /DESTIP/i) { 
                $netObj->destination_network(set => 1);
           
            }
            # Save up the netObjs, if we were called as an array, 
            # we will return all the objects we create...otherwise, the last object.
            push(@return, $netObj);
        }

        # Set the checker, so we know we went through the loop at least once.
        $checkForSomeInput = 1;
    
    } 
    if (! $checkForSomeInput) { 
        $Log->warn("IPManager did not receive any valid input on the last call to address_add!");
    }

    $Log->exit();
    return wantarray ? @return : $netObj;
}

=head2 address_version

    Attempts to guess the version of a given address.

=cut

sub address_version {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},], {address => {type => SCALAR},},);

    my $input = $opts{address};
    my $version;

    # Looking for this syntax: 192.168.1.2
    if ($input =~ /^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$/) {
        $Log->trace("Determined IPv4 address");
        $version = "4";

    } elsif ($input =~ /::/) {
        $Log->trace("Determined IPv6 address");
        $version = "6";
    }

    if (! $version) { 
        # Try Net::IP if the above fails
        if (my $netIP = new Net::IP("$input")) {
            $version = $netIP->ip_get_version();
    
        } else {
            $Log->warn("Net::IP error encountered!");
    
        }
    
        if (not $version) {
            die config_result(
                            "Could not determine the IP version of address $input");
        }
    }
    
    return $version;

}

=head2 create_range

    my @addrs = IPManager->create_range(begin => 192.168.1.0, 
                                         range => 20,);

    Creates an array containing a range of IP addresses specified
    by "begin" and "range" values.

=cut

sub create_range {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {begin => {type => SCALAR},
                      range => {type => SCALAR},},);

    my $begin = $opts{begin};
    my $range = $opts{range};

    $Log->enter("begin=$begin, range=$range");

    my @return;

    # Don't bother using netIP if the range is 1
    if ($range == 1) { 
        push (@return, $begin);
        $Log->exit();
        return @return;
    }

    if ($range < 1) { 
        die config_result("You want $range addresses? What the hell are you trying to pull?");
    }

    # Take the range value down by one, because the range 
    # will actually create "range" plus, the starting address;
    $range--;

    # This works for both IPv4 and IPv6
    my $ipObj =
      new Net::IP("$begin + $range") or
          die config_result("Net::IP error encountered!");

    do {
        push(@return, $ipObj->ip());
    } while (++$ipObj);

    $Log->exit("First Address: $return[0], Last Address: $return[-1]");
    return @return;
}

=head2 networks 

    Return an array of existing network objects. 

    Called in scalar context, it returns the number of network objects that exist.

=cut

sub networks {
    $Log->enter();
    my ($self) = @_;

    my @return = @{$self->{networks}};

    $Log->exit();
    return wantarray ? @return : scalar(@return);
}

=head2 network_by_networkID

    Return a reference to the network object with ID of networkID

    my $netObj = $IPManager->network_by_networkID(networkID => 192.168.0.0);

=cut

sub network_by_networkID {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {networkID => {type => SCALAR, optional => 0},},);

    my $id = $opts{networkID};

    foreach my $netObj ($self->networks()) {
        if ($netObj->networkID() =~ /$id/i) {
            $Log->exit();
            return $netObj;
        }
    }

    $Log->exit();
    return {};
}
=head2 network_by_gateway

    Return a reference to the network object with the specified gateway 

    my $netObj = $IPManager->network_by_gateway(gateway => 192.168.0.1);

=cut

sub network_by_gateway {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {gateway => {type => SCALAR, optional => 0},},);

    my $gw = $opts{gateway};

    foreach my $netObj ($self->networks()) {
        if ($netObj->gateway() =~ /$gw/i) {
            $Log->exit();
            return $netObj;
        }
    }

    $Log->exit();
    return {};
}

=head2 network_with_netid

    Return the network object that represents the given network ID.  

    There should never be more than one network object with the same network ID. 

    A network ID is the network address with the host portion substituted by zero's. 

=cut

sub network_with_netid {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},],
                     { net_id => {type => SCALAR},},);

    my $netid = $opts{net_id};
    my $return = undef;

    # Get all the networks
    my @networks = $self->networks();

    foreach my $net (@networks) {
        if ($net->networkID() =~ /^$netid$/i) {
            $return = $net;
            last;
        }
    }
    if (not $return) {
        $Log->warn("No network was found with network ID $netid");
    }
    $Log->exit();
    return $return;
}
=head2 source_networks

    Return an array of references to network objects with the "source" attributed set 

    Called in scalar context, it returns the number of network objects that exist with the source attribute.

=cut

sub source_networks {
    $Log->enter();
    my ($self) = @_;

    my @return;

    # Get all the networks
    my @networks = $self->networks();

    foreach my $net (@networks) {
        if ($net->source_network()) {
            push(@return, $net);
        }
    }
    if (not @return) {
        $Log->warn("No source networks were found in IPManager");
    }
    $Log->exit();
    return wantarray ? @return : scalar(@return);
}

=head2 destination_networks

    Return an array of references to network objects with the "destination" attributed set 

    Called in scalar context, it returns the number of network objects that exist with the destination attribute.

=cut

sub destination_networks {
    $Log->enter();
    my ($self) = @_;
 
    my @return;

    # Get all the networks
    my @networks = $self->networks();
  
    foreach my $net (@networks) {
        if ($net->destination_network()) {
            push(@return, $net);
        }
    }

    if (!@return) {
        $Log->warn("No destination networks were found in IPManager");
    }

    $Log->exit();
    return wantarray ? @return : scalar(@return);
}

#######################INTERNAL SUBS##########################

#
# This internal only routine is designed to help make sure that the
# input is provided in regular manner. This helps to allow the user
# to specifiy both a range and single addresses.
# We shall reduce the given input to array of anonymous hashes
#

sub _normalize_input {
    $Log->enter();
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},], {input => {type => SCALAR,},},);

    my @return;

    foreach my $input (split(/,/, $opts{input})) { 
        # The normalized data we need to entry
        my $entry = { gateway   => undef,
                      version   => undef,
                      netmask   => undef,    # Only obtained through tharnhost files
                      addresses => undef,    # Array to store the addresses
                      cidr      => undef,};
    
        my ($addrVersion, $addrStart, $cidr, $addrRange);
    
        # If the user specified a tharnhost file, get the data from there.
        if (my $tharnData = $self->_check_for_tharn_file(name => $input)) {
            # Make sure that this is an IP address object
            if (not $tharnData->{hosttype} =~ /ipv[4|6]/i) { 
                die config_result("The supplied infput file $input was not of the valid types. That is hosttype !~ [ipv4 | ipv6]");
            }

            $addrStart = $tharnData->{STARTIP};
            if (not $addrStart) { 
               die config_result("Unable to determine the starting address (STARTIP) from the tharn host file $input");
            }

            $addrRange = $tharnData->{IPCOUNT};
            if (not $addrRange) { 
               die config_result("Unable to determine the address range (IPCOUNT) from the tharn host file $input");
            }
    
            # Try to determine the address version from the tharnhost file hosttype value
            if ($tharnData->{hosttype} =~ /ipv4/i) {
                $entry->{version} = 4;
            } elsif ($tharnData->{hosttype} =~ /ipv6/i) {
                $entry->{version} = 6;
            }
    
            if (!$entry->{version}) {
                $Log->warn(
                    "Could not determine the address version from the specified hosttype value in the relevant tharnhost file. Attempting to guess version..."
                    );
    
                $entry->{version} =
                  $self->address_version(address => $addrStart);
        
                # We still couldn't figure out the version? Thats bad. Very bad. 
                if (! $entry->{version}) { 
                    die config_result("Unable to determine address version. Check format...giving up.");
                }
                $Log->trace(
                    "Address $addrStart was determined to be of IP version $entry->{version}."
                    );
            }

            # Get the netmask from the file tharnhost file if we can 
            if ($tharnData->{NETMASK}) {
                $entry->{netmask} = $tharnData->{NETMASK};
            } 
    
            # Try to get the cidr from the tharnhost file, if not, just
            # convert the netmask
            if ($tharnData->{CIDR}) {
                $entry->{cidr} = $tharnData->{CIDR};
    
            # If the CIDR notation is not present in the tharnhost file, use the netmask and convert that to CIDR
            } else {
                if (not $entry->{netmask}) {
                    die config_result(
                        "Unable to determine the CIDR value from the tharnhost file or netmask value "
                        );
                } else {
    
                    # Use the converters from the address object. Just pick up a blank object
                    my $ipObj =
                      Utils::IPManager::Address->_new(version => $entry->{version});
                    if (not $entry->{cidr} =
                        $ipObj->convert_netmask_to_cidr(netmask => $entry->{netmask})) {
                        die config_result(
                             "Unable to determine the CIDR value for address $addrStart"
                             );
                    }
                }
            }
    
            if ($entry->{gateway} = $tharnData->{GATEWAY}) {
                $Log->trace(
                    "Found the gateway for this network as $entry->{gateway} from the host file"
                    );
            } else {
                $Log->warn(
                       "Unable to determine the gateway value from the host file.");
            }
    
        # If the user specified all the data on the command line, get it there and ignore the tharnhost file
        } else {
            my $theRest;
    
            # There must always be "/" to seperate the cidr from the address
            if (not(($addrStart, $theRest) = split(/\//, $input))) {
                die config_result(
                    "The IP address specified ($input) was not in a recognized form."
                    );
            }
    
            $entry->{version} =
              $self->address_version(address => $addrStart);
            $Log->comment(
                "Address $addrStart was determined to be of IP version $entry->{version}."
                );
    
            # Check to see if a range is passed
            # ranges are indicated by [someNumber]
            if ($theRest =~ /\[(\d+)\]/) {
                $addrRange = $1;
    
                # Remove the range from the address
                $theRest =~ s/\[(\d+)\]//;
    
                # If the input line did not include a range, then it is 1
            } else {
                $addrRange = 1;
            }
    
            # With the range removed, the CIDR should now be the last value on the input
            if (($entry->{cidr}, $theRest) = split(/:/, $theRest, 2 )) { 
                $Log->trace("Determined CIDR to be $entry->{cidr}");

            } else {
                die config_result(
                    "Could not determine the CIDR value from the given input ($input)"
                    );
            }
    
            if ($theRest) {
                $entry->{gateway} = $theRest;
                $Log->trace("Determined the gateway to be $entry->{gateway}");
            } else { 
                $Log->trace("The gateway was not determined from the given input");
            }
        }
  
        # add all the addresses to the hashref we will be entering
        # Don't waste our time with Net::IP if we only got one address 
        if ($addrRange == 1) { 
            if ($addrStart =~ /::/) { 
                $addrStart = $self->_expand_colons(input => $addrStart);
            }
            push (@{$entry->{addresses}}, $addrStart); 
        } else {
            @{$entry->{addresses}} =
              $self->create_range(range => $addrRange,
                                  begin => $addrStart);
        }
 
        # Expand the dual colon syntax in IP6 gateways and addresses. For now, 
        # the address portion is handled when the range is generated by 
        # the Net::IP program 
        if ($entry->{gateway} =~ /::/) { 
             $entry->{gateway} = $self->_expand_colons(input => $entry->{gateway});
        }

        # Store the whole entry hash on a return array
        # we end up returning one hash for every comma 
        # seperated entry in the input.
        push (@return, $entry);
    }

    $Log->exit();
    return @return;

}

#
# Use this routine to expand IP6 addresses with "::" in them, into 
# the full form
#
sub _expand_colons {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {input => {type => SCALAR},},);
    $Log->enter("input=" .  $opts{input});

    my $string = $opts{input};
    my $groupsThereShouldBe = 8;
    my $groupNum = scalar(my @temp = split(/:/, $string));

    # add one to make up for the num counted as being 
    # between the ::  chars we just split
    my $needFills = (($groupsThereShouldBe - $groupNum) + 1);
    my $fullAddr = "";
    if (my ($firstHalf, $secondHalf) = split(/::/, $string, 2)) {

        my $filler = "0000";
        $fullAddr = $firstHalf;
        foreach (1 .. $needFills) {
           $fullAddr .= ':' . $filler;
        }
        $fullAddr .= ':' . $secondHalf;

    } else {
        $fullAddr = $string;
    }
    $Log->exit($fullAddr);
    return $fullAddr;
}
#
# A private method to handle new IP objects and populating
# the relevant data that is required.
#
sub _address_add {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {address => {type => SCALAR},
                      version => {type => SCALAR},
                      gateway => {type => SCALAR|UNDEF, optional => 1},
                      netmask => {type => SCALAR|UNDEF, optional => 1},
                      cidr    => {type => SCALAR},},);

    my $addr    = $opts{address};
    my $cidr    = $opts{cidr};
    my $version = $opts{version};
    my $gateway = $opts{gateway};
    my $netmask = $opts{netmask};
    my $netObj;

    # Create the address object.
    my $ipObj = Utils::IPManager::Address->_new(version => $version);

    # convert the cidr notation to the netmask, if we don't have it already
    if ((not $netmask) and ($version == 4)) { 
        $netmask = $ipObj->convert_cidr_to_netmask(cidr => $cidr); 

    } elsif ((not $netmask) and ($version == 6)) { 
        $netmask = $ipObj->convert_cidr_to_netmask(cidr => $cidr, address => $addr); 
    }

    # See if the network object exists already. If so, use that
    # We will use the gateway address to find the network obj
    # if there is no gateway, we must have a flat network...in 
    # which case there can be only one existing network object
    if ($gateway) { 
        $netObj = $self->network_by_gateway(gateway => $gateway);
    } else { 
        my @return = $self->networks();
        $netObj = $return[0];
    }

    if (not UNIVERSAL::can($netObj, 'isa')) {
        # otherwise, create a new network object and populate it

        # The gateway must exist for this to work. 
        if (not $gateway) { 
            $Log->warn("Gateway informaiton was not received by address_add for the network " 
                         . "which would contain address $addr and netmask $netmask.");
        }
        $netObj = $self->_network_add(address   => $addr,
                                      ipObj     => $ipObj,
                                      cidr      => $cidr,
                                      version   => $version,
                                      netmask   => $netmask,
                                      gateway   => $gateway);

        $Log->trace("Created a new networks object with network ID " . $netObj->networkID());
    }


    # populate the IP address object with all the relevant data.
    $ipObj->_populate(address   => $addr,
                      network   => $netObj,
                      IPManager => $self,);

    # add the IP object to the network object
    $netObj->_address_add(add => $ipObj);
    $Log->trace("Added IP address $addr to network ID: " . $netObj->networkID());

    return $netObj;
}

#
# This internal only sub is used to connect the messenger object
# to this IPManager object. It handles retries and retrieving 
# contact information from the IPManager server process
#
sub _connect_peer_messenger {
    $Log->enter();
    my ($self) = @_;

    my $retries=4;
    my $try = 0;
    my $sleepTime = 3;

    while (1) {
        # Exit this loop after three retries if
        if (++$try > $retries) {
           $Log->warn("Failed to connect to an IPManager server in the alloted number of $retries retries");
           $Log->exit(0);
           return 0;
        }

        my $result = $Staf->var_list(endpoint => $Tharn::Ancestor->endpoint(), handle => $Tharn::Ancestor->handle(),);
        my $stafInfo =  $result->{IPMANAGER_STAF_SERVER};

        if ((eval{$stafInfo->{handle}}) and (eval{$stafInfo->{endpoint}})) {
            # Connect to the server STAF instance by creating message object
            $Log->comment("Contacting IPManager server $stafInfo->{endpoint}, known by handle $stafInfo->{handle}");
            $self->{messenger} = Utils::IPManager::ServerMessage->new(handle => $stafInfo->{handle},
                                                                     endpoint => $stafInfo->{endpoint},);

            $Log->exit(1);
            return 1;

        } 

        # If the IPManager value is 1, then the server is starting.
        # Give it some time to get up and going...unless, of course,
        # this happens to BE the server...
        $sleepTime = $sleepTime * 2;
        if ($stafInfo == 1) {
            if ($try == 1) { 
                $Log->comment("The IPManager server is starting, but not yet running. Sleeping $sleepTime seconds");
             } else { 
                $Log->comment("Retry...sleeping $sleepTime seconds...");
             }
             sleep($sleepTime);
        }
    }

    $Log->exit(0);
    return 0;
}
=head2 peer_messenger 

    Creates a messenger object which points to the staf instance at the IPManager 
    server process. If an object already exists, it will be returned. If an object
    does not exists, then we shall attempt to create one and connect it before 
    returning to the user. 

=cut

sub peer_messenger { 
    $Log->enter();
    my ($self) = @_;

    # If the peer object already exists, return that. 
    if ($self->{messenger}) { 
        $Log->trace("Returning existing IPMananger Staf server peer object.");

    } elsif ($self->_connect_peer_messenger()) { 
        # If there is a server out there, all defined, then we will rely on that
        # and not process any input. 
        $Log->trace("Connected an existing IPManager server.");
    } else { 
        $Log->warn("Unable to contact an IPManager server. IPManager is running autonomously.");
    }

    return $self->{messenger};
}

#
# A private method to get a new network object and populate it with
# the relevant data that is required.
#

sub _network_add {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {gateway    => {type => SCALAR, optional => 1},
                      ipObj      => {type => OBJECT, optional => 1},
                      networkID  => {type => SCALAR, optional => 1},
                      netmask    => {type => SCALAR},
                      address    => {type => SCALAR, optional => 1},
                      version    => {type => SCALAR},
                      cidr       => {type => SCALAR},},);
    
    my $ipObj = ($opts{ipObj}) ?  $opts{ipObj} : undef;
    my $networkID = ($opts{networkID}) ? $opts{networkID} : undef;
    
    # if we did NOT get the address then we MUST have 
    # the network ID already
    if (((not $opts{address}) or (not $ipObj)) and (not $networkID)) {
        die config_result ("_network_add requires both networkdID and ipObj if an address is not supplied");
    }
 
    if (not $networkID) { 
        # use the IP object to compute the network ID 
        $networkID =
          $ipObj->compute_networkID(address => $opts{address},
                                    netmask => $opts{netmask});
    }

    my %data = (networkID => $networkID,
                netmask   => $opts{netmask},
                version   => $opts{version},
                cidr      => $opts{cidr},);

    $data{gateway} = ($opts{gateway}) ? $opts{gateway} : undef;

    my $netObj = Utils::IPManager::Network->_new();

    $netObj->_populate(%data);

    $netObj->parent(set => $self);

    # add the network object to the IPManager object
    push (@{$self->{networks}}, $netObj);

    $Log->exit();
    return $netObj;
}

#
# check to see if a tharnhost file exists. Return a hashref of its data
# if the file exists
#
sub _check_for_tharn_file {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},], {name => {type => SCALAR},},);

    my $th; 
    
    eval {$th = host($opts{name})};
    if ($@) {
        $Log->trace("Given input is not an existing tharnhost file in valid form. Going on with syntax determination..."); 
        $Log->exit();
        return 0;
    }

    $Log->exit();
    return $th;
    
}

#
# Returns true if this is the IPManager server
#

sub is_manager_server { 
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {set => {type => BOOLEAN, optional => 1},},);

    $self->{isServer} = $opts{set} if $opts{set};

    return $self->{isServer};
}



1;
