package Utils::IPManager::Address::IP4;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use Tharn qw($Log config_result script_result);
use Data::Dumper;

=head1 NAME

Utils::IPManager::Address::IP4

=head1 DESCRIPTION

This library is meant to support version 4 addresses
Do not access this library directly. use the IPManager object
Once the object has been created using the address_add() method 
on the IP manager object, there are a few bits of useful information
that can be retrieved. See below.

=head1 SYNOPSIS

    my @netObjs = $IPManager->networks();
    my @addrs = $netObj[0]->addresses();

    my $addr = $addrs[0];


=head1 EXAMPLE

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Jd (dascenj@netapp.com)
  Created

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=head1 ROUTINES

=cut

##########################Object Creation Methods###########################

#
# This private routine adds creates the address object and
# sets the required information
#

sub _new {
    $Log->enter();
    my ($invocant, %opts) = @_;
    my $class = ref($invocant) || $invocant;
    my $self = {};

    bless($self, $class);

    # The list of variables availabe on this object
    $self->{address}   = undef;
    $self->{netObj}    = {};
    $self->{ipManager} = {};
    $self->{assigned}  = 0;

    $Log->exit();
    return $self;
}

#
# Internal routine to populate the address/IP object
# with the correct relevant data.
#
sub _populate {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {address   => {type => SCALAR, optional => 0},
                      network   => {type => OBJECT, optional => 0},
                      IPManager => {type => OBJECT, optional => 0},},);

    $self->{address}   = $opts{address};
    $self->{netObj}    = $opts{network};
    $self->{ipManager} = $opts{IPManager};

    $Log->exit();
}

#################Return values specific to the address object######################

=head2 address 

    Returns the address for the given IP object 

=cut

sub address {
    my ($self) = @_;

    return $self->{address};
}

=head2 network 

    Returns the network (parent) object for this address 

=cut

sub network {
    my ($self) = @_;

    return $self->{netObj};
}

=head2 assigned

    This boolean value is used to determine whether or not 
    this address has already been assigned to something for 
    use.  

=cut

sub assigned {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {set => {type => BOOLEAN, optional => 1},},);

    if ($opts{set}) {
        $self->{assigned} = $opts{set};
    } else {
        return $self->{assigned};
    }
}

=head2 convert_cidr_to_netmask 

    Takes a CIDR value, and returns the netmask in octets

=cut

sub convert_cidr_to_netmask {
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},], {cidr => {type => SCALAR,},},);

    my $cidr = $opts{cidr};
    if ((not $cidr > 0) and (not $cidr <= 32)) {
        $Log->warn(
            "Could not convert $cidr to a netmask. Valid CIDR values are from 1 to 32"
            );
        return 0;
    }

    my $totalOctets  = 4;
    my $cidrMaskBits = $cidr;
    my $bitsPerOctet = 8;
    my $netMask;
    foreach my $octet (1 .. $totalOctets) {
        my $counter = 1;
        my $octetBin;
        while ($counter <= $bitsPerOctet) {
            if ($cidrMaskBits > 0) {
                $octetBin .= 1;
            } else {
                $octetBin .= 0;
            }
            $counter++;
            $cidrMaskBits--;
        }
        $counter = 0;
        $netMask .= unpack("N", pack("B32", substr("0" x 32 . $octetBin, -32)));
        if ($octet < 4) {
            $netMask .= ".";
        }
    }
    $Log->debug("Translated cidr $cidr to octets $netMask");
    return $netMask;
}

=head2 convert_netmask_to_cidr

    Takes a netmask value (in octets) and returns the CIDR value.

=cut

sub convert_netmask_to_cidr {
    $Log->enter();
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},], {netmask => {type => SCALAR,optional => 1},},);

    my @octets  = ();
    my $cidr    = 0;
    my $netmask; 
    if ($opts{netmask}) {
        $netmask = $opts{netmask};
    } else {
        $netmask = $self->netmask();
    }

    if (not(@octets = split(/\./, $netmask))) {
        $Log->warn(
            "Could not convert netmask ($netmask) to cidr. Verify netmask format"
            );
        return 0;
    }

    foreach my $octet (@octets) {
        my $bits = unpack("B32", pack("N", $octet));

        # remove the leading zero's
        $bits =~ s/^0+(?=\d)//;

        foreach my $digit (split(//, $bits)) {
            if ($digit) {
                $cidr++;
            }
        }
    }

    $Log->exit($cidr);
    return $cidr;

}

=head2 compute_networkID

    The network ID is a unique identifier for the network. 
    It is determined from a the values of "address" and 
    "netmask". The return is a scalar networkID value

=cut

sub compute_networkID {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {address => {type => SCALAR},
                      netmask => {type => SCALAR},},);

    my $addr = $opts{address};
    my $sub  = $opts{netmask};

    my @addrarr = split(/\./, $addr);
    my ($ipaddress) = unpack("N", pack("C4", @addrarr));

    my @maskarr = split(/\./, $sub);
    my ($netmask) = unpack("N", pack("C4", @maskarr));

    # Calculate network address by logical AND operation of addr & netmask
    # and convert network address to IP address format
    my $netadd = ($ipaddress & $netmask);
    my @netarr = unpack("C4", pack("N", $netadd));

    return join(".", @netarr);
}

1;

###############Return values from the associated network object####################

=head2 The remaining routines actually return data from the related network object

=cut 

=head3 netmask 

    returns the version of IP that this network/address is configured as.t 

=cut

sub version {
    my ($self) = @_;

    my $netObj = $self->network();
    return $netObj->version();
}

=head3 netmask 

    returns the netmask (in octet) attribute 

=cut

sub netmask {
    my ($self) = @_;

    my $netObj = $self->network();
    return $netObj->netmask();
}

=head3 gateway 

    Return the default gateway for the network that this address 
    belongs too. 

=cut

sub gateway {
    my ($self) = @_;

    my $netObj = $self->network();
    return $netObj->gateway();
}

=head3 cidr 

    Return the netmask of this address in cidr notation 

=cut

sub cidr {
    my ($self) = @_;

    my $netObj = $self->network();
    return $netObj->cidr();
}

=head3 networkID 

    Return the networkID of the network that this address
    belongs to.  

=cut

sub networkID {
    my ($self) = @_;

    my $netObj = $self->network();
    return $netObj->networkID();
}

