package Utils::IPManager::Network;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use Tharn qw($Log config_result script_result);
use Data::Dumper;
use Options;

my $Options = new Options;

=head1 NAME

Utils::IPManager::Network  

=head1 DESCRIPTION

Creates a network object, and provides information about said network 
Contains mostly get routines to display data about a given network

=head1 SYNOPSIS

    my @netObjs = $IPManager->networks();
    my $netObj = @netObjs[0];
    
=head1 EXAMPLE

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Jd (dascenj@netapp.com)
  Created

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=head1 ROUTINES

=cut

#
# This internal routine creates a new network object.
# Should be called only by adding a new IP address to the
# IPManager object.
#

sub _new {
    $Log->enter();
    my ($invocant, %opts) = @_;

    my $class = ref($invocant) || $invocant;
    my $self = {};

    bless($self, $class);

    # Still no network object? DEAD.
    if (not UNIVERSAL::can($self, 'isa')) {
        die config_result("Unable to create a new network object!");
    }

    @{$self->{assignedAddresses}} = ();
    @{$self->{unassignedAddresses}} = ();
    $self->{source}    = 0;
    $self->{dest}      = 0;
    $self->{parent}    = undef;

    $Log->trace("Created Network object");

    $Log->exit();
    return $self;
}

#
# Internal routine to add required information
# to our network object to make it specific to a given
# network.
#

sub _populate {
    my ($self, %opts) = @_;
    # reset some values
    $self->{networkID} = $opts{networkID};
    $self->{netmask}   = $opts{netmask};

    # gateway is not required in flat networsk...it may
    # not be specified.
    $self->{gateway}   = $opts{gateway};
    $self->{cidr}      = $opts{cidr};
    $self->{version}   = $opts{version};

    $Log->trace("Populated a network object: " . Dumper($self));
}

#
# Private method to add an address object to the network object
# this method should only be invoked when a new address is added
# to the IPManager using address_add
#

sub _address_add {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {add => {type => OBJECT, optional => 0},},);

    my $addr = $opts{add};
    if ($self->gateway() eq $addr->address()) { 
        $Log->warn("The address " . $addr->address() . " was not added to the network object, because it " .  
                    "matches the gateway address of " . $self->gateway());
    } else { 
        push(@{$self->{unassignedAddresses}}, $addr);
    }

}

=head2 source_network

    A boolean value that indicates that this network was created 
    as a known "source" object. Aka, clients are configured on this 
    network.

=cut

sub source_network {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {set => {type => BOOLEAN, optional => 1},},);

    if (defined($opts{set})) {
        $Log->trace("Marked network " . $self->networkID() . " as a source network");
        $self->{source} = $opts{set};
    }

    return $self->{source};
}

=head2 destination_network

    A boolean value that indicates that this network was created 
    as a known "destination" object. Aka, controllers are configured on this 
    network.

=cut 

sub destination_network {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {set => {type => BOOLEAN, optional => 1},},);

    if (defined($opts{set})) {
        $Log->trace("Marked network " . $self->networkID() . " as a destination network");
        $self->{dest} = $opts{set};
    }

    return $self->{dest};
}

=head2 netmask

    Display the netmask (in octets) of the current network object 

=cut

sub netmask {
    my ($self) = @_;

    return $self->{netmask};
}

=head2 cidr

    Display the netmask (in cidr notation) of the current network object 

=cut

sub cidr {
    my ($self) = @_;

    return $self->{cidr};
}
=head2 version 

    Display the IP version. It will return 4 or 6.

=cut

sub version {
    my ($self) = @_;

    return $self->{version};
}

=head2 assign_next_address 

    This method changes the assigned bit on the next avialable address object 
    and returns said object for use. The assigned bit, keeps this address from 
    being given out again.

=cut

sub assign_next_address {
    $Log->enter();
    my ($self) = @_;

    # Look at all the addresses. Do this so that if 
    # it comes back empty, we will check with the server
    # for more addresses
    $self->unassigned_addresses();
  
    # take an object from the unassigned pool
    my $ipObj = shift(@{$self->{unassignedAddresses}});

    # Set the assigned bit
    $ipObj->assigned(set => 1);
    $Log->debug("Assigned IP address " . $ipObj->address() . ".");

    push (@{$self->{assignedAddresses}}, $ipObj);

    $Log->exit();
    return $ipObj;
}

=head2 networkID

    display the networkID value of the current network 

=cut

sub networkID {
    my ($self) = @_;

    return $self->{networkID};
}

=head2 unassigned_addresses 

    return an array of address objects that are not assigned and 
    are available for use. In scalar context, it returns the number of 
    available addresses on the current network

=cut

sub unassigned_addresses {
    $Log->enter();
    my ($self) = @_;
 
    my @unassigned = (); 

   
    if (not scalar(@{$self->{unassignedAddresses}})) {
        $Log->trace("There were no additional IP addresses available locally. Broadening horizons...");
        @unassigned = $self->_get_address_from_server();

    } else { 
        @unassigned = @{$self->{unassignedAddresses}};
    }

    $Log->exit();
    return wantarray ? @unassigned : scalar(@unassigned);

}

=head2 assigned_addresses 

    return an array of address objects that are assigned already  
    and not available for use. In scalar context, it returns the
    number of available addresses on the current network.

=cut

sub assigned_addresses {
    my ($self) = @_;
 
    my @addrs = ();
 
    if (scalar(@{$self->{assignedAddresses}})) {  
        @addrs = @{$self->{assignedAddresses}}; 
    }
   
    return wantarray ? @addrs : scalar(@addrs);

}

#
# This routine should only be called internally. This should 
# never be called by the user directly. The idea here is to 
# head off to the server and gather up some more IP addresses
# only after we have run out of any addresses that may be 
# available locally
#

sub _get_address_from_server { 
    $Log->enter();  
    my ($self) = @_;
    # See if the manager (parent) has a
    # messenger configured for server communication.
    # If so, get an IP address from there using the messenger
    my $Options_hash = $Options->get_hash();
    my $manager = $self->parent();
    my $percent = $Options_hash->{IP_PERCENT};
    $percent =~ s/\%//;
    my $messenger;
    my @return;

    # If there is a server out there somwhere, then the peer messenger object 
    # will be configured. 
    if ($messenger = $manager->peer_messenger()) { 
        # If this is the server, there is no where else to look. Make sure there is a messenger 
        # configured also 
        if (not $manager->is_manager_server()) {
            my @serverAddrs;
            # Destination networks should always assign one address at a time
            if (($percent) and ($self->source_network())) { 
                $Log->trace("Using the value of IP_PERCENT (" . $percent . ") to determine the number of addresses to get from IPManager server.");
                my @return = $messenger->server_assign_address_percent(networkID => $self->networkID(), percent => $percent);
                my $sc = $return[0];
                my $string = $sc->{address} . '/' . $sc->{cidr} . '[' . $sc->{range} . ']:' . $sc->{gateway};

                $manager->address_add(SRCIP => $string);
                $Log->comment("Added SRCIP $string to the IPManager server.");   
    
            } else {  
                $Log->trace("There are no unassigned addresses locally. Checking with IPManager server.");
                my @return =  $messenger->server_assign_next_address(networkID => $self->networkID());
                my $sc = $return[0];

                $manager->_address_add(gateway => $sc->{gateway},
                                       version => $sc->{version},
                                       netmask => $sc->{netmask},
                                       address => $sc->{address},
                                          cidr => $sc->{cidr},);
            }
            @return = $self->unassigned_addresses();

        } else { 
            $Log->trace("This IS the server, there is no where else to look...");
        }
    } else { 
        $Log->trace("There is no messenger server configured. Cannot obtain additional addresses");
    }

    if (not @return) { 
        my $text = "type";
        if ($self->source_network()) { 
            $text .= ": source";
        }
        if ($self->destination_network()) { 
            $text .= ": destination";
        }
        die config_result(
           "IPManager was asked to assign a new address from the $self->{networkID} of $text, but "
             . "there are no unassigned addresses to give. Check SRCIP, DESTIP and IP_PERCENT values.");
    }
    
    $Log->exit();  
    return @return;
}

=head2 gateway

    Return the default gateway for the current network 

=cut

sub gateway {
    my ($self) = @_;

    return $self->{gateway};
}

=head2 new

    Return an array of all address objects on this network
    In scalar context, the number of addresses is returned.

=cut

sub addresses {
    $Log->enter();
    my ($self) = @_;

    my @addrs =  (); 
   
#    no strict 'refs';
#    no warnings;

    @addrs = (@addrs, $self->assigned_addresses()) if (scalar(@{$self->{assignedAddresses}}));

    # Don't use the get routines, unless we know there are values there, otherwise
    # it will try to contact the server for more addresses
    @addrs = (@addrs, $self->unassigned_addresses()) if (scalar(@{$self->{unassignedAddresses}}));

#    use strict;
#    use warnings;

    $Log->exit();
    return wantarray ? @addrs : scalar(@addrs);
}

sub parent {
    my ($self, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {set => {type => OBJECT, optional => 1},},);

    if (defined($opts{set})) {
        $self->{parent} = $opts{set};
    }

    return $self->{parent};
}

1;

