# $Id$
## @summary Facilitate the communication between IPManager and Server processes 
## @author dascenj@netapp.com
## @pod here

=head1 NAME

Utils::IPManager::ServerMessage.pm - Facilitate communication between IPManager
                                     processes.

=head1 DESCRIPTION

This package defines routines that allow the user to execute code on the IPManager
server process and receive the output. 

=head1 SYNOPSIS

    Generally messenger objects are created automatically by the 
    IPManager client process. However, if needed, messenger objects
    can be created independently.
    
    use Utils::IPManager::ServerMessage;
    my $messenger = new Utils::IPManager::ServerMessage->new();

    To send any of the preset messages, one need only to call the 
    messenger method that corresponds to the server call that the 
    user wishes to execute. In this case, no manual calls to any 
    other method will be required.

=head1 EXAMPLES

=head2 Send the server a request to receive all the empty networks

=over

=item *

C<< $response = messenger->server_empty_networks(); >>

=back

=head1 REQUIREMENTS

    This is meant to be used with the IPManager libraries in a client/server
    type configuration.

    In order to send a message manually, you must populate the messenger 
    object using server_method() and server_code_string() methods. Once this 
    information is supplied you can send the message by calling contact_server().

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Jd (dascenj@netapp.com)
  Created

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=head1 ROUTINES

=cut


package Utils::IPManager::ServerMessage;

use Data::Dumper;
use Tharn qw($Log config_result);
use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use NATE::STAF::Peer;

my $timeout = 9999;

sub new {
    $Log->enter();
    my ($invocant, %opts) =
      validate_mixed(@_,
                     [{type => SCALAR},],
                     {handle => {type => SCALAR},
                      endpoint => {type => SCALAR},},);

    my $class = ref($invocant) || $invocant;
    my $self = {};
    bless($self, $class);

    # Initialize the response array
     $self->{response} = ();
     my $staf = NATE::STAF::Peer->new(handle => $opts{handle},
                                               endpoint => $opts{endpoint}) or
         die config_result ("Unable to create a peer relationship with the IPManager server");

    $self->server_staf(set => $staf);
  
    $Log->exit();
    return $self;
}

=head2 clean_messenger 

    Reinitialize all the pertinent fields on this
    messenger object in order to prepare it for the 
    next usage.
   
=cut
sub clean_messenger { 
    $Log->enter();
    my ($self) = @_; 
 
    $self->{response}  = ();
    $self->{serverMethod} = undef;
    $self->{codeString} = undef;

    $Log->exit();
}
    
=head2 server_empty_networks

    Ask the IPManager server to send information
    on all the networks that it is currently hosting.
    The network data is returned, with the IP addresses 
    removed. The IP addresses must be sepcifically requested
    from the correct network Object.

=cut
sub server_empty_networks { 
    $Log->enter();
    my ($self) = @_;

    $self->clean_messenger();

    # The subroutine has the package attached...if
    # there is a "::" get the last element in the list
    my $subroutine = (caller(0))[3];
    if ($subroutine =~ '::') {
        my @data = split(/::/, $subroutine);
        $subroutine = $data[-1];
    }

    # These are private methods on the 
    # server side.
    $subroutine = '_' . $subroutine;
    $self->server_method(set => $subroutine);

    my $code = 'return &' . $self->server_method() . '();';
    $self->server_code_string(set => $code);

    $self->contact_server();

    $Log->exit($code);
    return $self->server_response();
}

=head2 server_assign_next_address

    Ask the IPManager server to assign it's next
    available address.

=cut

sub server_assign_next_address {
    $Log->enter();
    my ($self, %opts) = 
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {networkID => {type => SCALAR},},);

    $self->clean_messenger();

    # The subroutine has the package attached...if
    # there is a "::" get the last element in the list
    my $subroutine = (caller(0))[3];
    if ($subroutine =~ '::') {
        my @data = split(/::/, $subroutine);
        $subroutine = $data[-1];
    }


    # These are private methods on the 
    # server side.
    $subroutine = '_' . $subroutine;
    $self->server_method(set => $subroutine);

    my $code = 'return &' . $self->server_method() . '(networkID => "' . $opts{networkID} . '");';

    $self->server_code_string(set => $code);
    $self->contact_server();

    $Log->exit($code);
    return $self->server_response();
}

=head2 server_assign_address_percent

    Ask the IPManager server to assign a percentage of
    it's available addresses.

=cut

sub server_assign_address_percent { 
    $Log->enter();
    my ($self, %opts) = 
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {networkID => {type => SCALAR},
                      percent   => {type => SCALAR},},);

    $self->clean_messenger();

    # The subroutine has the package attached...if
    # there is a "::" get the last element in the list
    my $subroutine = (caller(0))[3];
    if ($subroutine =~ '::') {
        my @data = split(/::/, $subroutine);
        $subroutine = $data[-1];
    }

    # These are private methods on the 
    # server side.
    $subroutine = '_' . $subroutine;
    $self->server_method(set => $subroutine);

    my $code = 'return &' . $self->server_method() . '(networkID => "' . $opts{networkID} . '", percent => "' . $opts{percent} . '");';
    $self->server_code_string(set => $code);

    $self->contact_server();

    $Log->exit($code);
    return $self->server_response();
}


=head2 contact_server

    send the completed message object to the 
    server for processing. Before calling this 
    method, the user MUST call server_method()
    server_code_string() and fill in the right values
=cut

sub contact_server { 
    $Log->enter();
    my ($self) = @_;
    my $server = $self->server_staf();
    
    $self->{response} = $server->perl_eval(request_type      => 'MyRPCRequest',
                                           reply_type_prefix => 'MyRPCReply,', 
                                           timeout           => $timeout,
                                           code              => $self->server_code_string(),); 

    if (not $self->{response}) { 
       die config_result("Communication with IPManager server from messenger object failed.");
    }
    $Log->exit(1);
     return 1;
}

=head2 server_response

    Get the array which contains the IPManager 
    server's response to the query performed. 
    Calling this method causes the message to be 
    delete from the object.

=cut

sub server_response { 
    $Log->enter();
    my ($self) = @_;

    # save what we need and clean off the rest
    my @return = @{$self->{response}};
    $self->clean_messenger();

    $Log->exit(Dumper(@return));
    return @return;
}

=head2 server_staf 

    Get or set the object that references the staf instance
    associated with the IPManager server process.

=cut

sub server_staf {
    $Log->enter();
    my ($self, %opts) = 
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {set => {type => OBJECT, optional => 1},},);

    $self->{serverStaf} = $opts{set} if $opts{set};

    $Log->exit();
    return $self->{serverStaf};
}

=head2 server_code_string

    Get or set the code string that will be passed to the server method
    for execution when this messenger contacts IPManagerServer.thpl 

=cut

sub server_code_string {
    $Log->enter();
    my ($self, %opts) = 
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {set => {type => SCALAR, optional => 1},},);

    $self->{codeString} = $opts{set} if $opts{set};

    $Log->exit();
    return $self->{codeString};
}
 
=head2 server_method

    Get or set the name of the method that should receive the code
    code ref specified by this messenger when contacting IPManagerServer.thpl 

=cut
sub server_method {
    $Log->enter();
    my ($self, %opts) = 
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {set => {type => SCALAR, optional => 1},},);

    $self->{serverMethod} = $opts{set} if $opts{set};

    if (not $self->{serverMethod}) { 
        die config_result("The server_method must be defined before asking the server to process this message!"); 
    }

    $Log->exit($self->{serverMethod});
    return $self->{serverMethod};
}

1;
