package Utils::LMR;

use strict;
use warnings;
use Params::Validate
    qw(validate validate_with SCALAR HASHREF BOOLEAN ARRAYREF OBJECT UNDEF CODEREF);
use Tharn;
use DBI qw();
use Module::Common qw(validate_mixed);
use List::MoreUtils qw(uniq);
use Tharn qw($Log);
use Sort::Key::Natural qw(natsort);
use POSIX qw(strftime);
use Data::Dumper;
use NATE::STAF qw(staf_global);
use Time::Local;
use Time::HiRes qw(gettimeofday tv_interval);
use Sys::Hostname qw(hostname);
our $VERSION = 1.00;
our @ISA     = qw (Exporter);

# Creats a new LMR object
sub new {

    $Log->enter();
    my ($invocant, %opts) = validate_mixed(
        @_,
        [{type => SCALAR}],
        {   path     => {type => SCALAR,         optional => 1},
            user     => {type => SCALAR | UNDEF, optional => 1},
            auth     => {type => SCALAR | UNDEF, optional => 1},
            hostname => {type => SCALAR,         optional => 1},
            jobid    => {type => SCALAR,         optional => 1}
        },
    );

    my $class = ref($invocant) || $invocant;

    my $self = {};
    $self->{'path'} = $opts{'path'};
    $self->{'user'} = $opts{'user'};
    $self->{'auth'} = $opts{'auth'};

    if ($opts{'hostname'}) {
        $self->{'hostname'} = $opts{'hostname'};
    } elsif ($NATE::STAF::Peer::Ancestor) {
        $self->{'hostname'} = $NATE::STAF::Peer::Ancestor->endpoint();
    }

    # If a jobid is specified, look for LMR STAF handles and
    # then choose the right handle based on the presence of
    # the jobid in the LMR_PATH
    if ($opts{jobid}) {
        $self->{'staf_handle'} = get_staf_handle(
            jobid    => $opts{jobid},
            hostname => $opts{hostname}
        );
        my $staf = staf_global();
        eval {
            $self->{'path'} = $staf->var_get(
                handle   => $self->{staf_handle},
                var      => 'LMR_PATH',
                endpoint => $self->{hostname}
            );
        };
        if (!$self->{'path'}) {
            die config_result(
                      "Unable to find the LMR_PATH in the STAF variable"
                    . "service. Did you specify the right hostname or "
                    . "jobid?\n");
        }
    } elsif ($NATE::STAF::Peer::Ancestor) {
        $self->{'staf_handle'} = $NATE::STAF::Peer::Ancestor->handle();
    }

    if (-f "$opts{'path'}/lmr.db") {
        $self->{'databases'} = ['single'];
    } else {
        $self->{'databases'} = ['mhost', 'slow', 'fast', 'clients'];
    }

    $self->{'slow'} = {
        'dsn'     => "dbi:SQLite:dbname=$self->{'path'}/lmr_slow.db",
        'options' => {PrintError => 0, RaiseError => 1, AutoCommit => 1},
        'tables'  => [
            {   'source' => 'ontap',
                'type'   => 'statistics',
                'name'   => 'ontap_statistics',
                'create_query' =>
                    'CREATE TABLE "ontap_statistics" (id INTEGER PRIMARY '
                    . 'KEY, date DATE, counter_name TEXT, counter_value TEXT, '
                    . 'subsys TEXT, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'ontap',
                'type'   => 'memory',
                'name'   => 'ontap_memory',
                'create_query' =>
                    'CREATE TABLE "ontap_memory" (id INTEGER PRIMARY '
                    . 'KEY, date DATE, counter_name TEXT, counter_value TEXT, '
                    . 'subsys TEXT, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'counter',
                'type'   => 'lookup',
                'name'   => 'counters',
                'create_query' =>
                    'CREATE TABLE "counters" (id INTEGER PRIMARY '
                    . 'KEY, counter_name TEXT, subsys TEXT, table_name TEXT, unit TEXT, counter_type TEXT)'
            },
            {   'source'       => 'host',
                'type'         => 'lookup',
                'name'         => 'hosts',
                'create_query' => 'CREATE TABLE "hosts" (id INTEGER PRIMARY '
                    . 'KEY, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'ontap',
                'type'   => 'data',
                'name'   => 'node_data',
                'create_query' =>
                    'CREATE TABLE "node_data" (id INTEGER PRIMARY '
                    . 'KEY, node_name TEXT, ontap_release TEXT, node_model TEXT)'
            }
        ]
    };
    ##New database created for MHOST path
    $self->{'mhost'} = {
        'dsn'     => "dbi:SQLite:dbname=$self->{'path'}/lmr_mhost.db",
        'options' => {PrintError => 0, RaiseError => 1, AutoCommit => 1},
        'tables'  => [
            {   'source' => 'MHOST',
                'type'   => 'statistics',
                'name'   => 'mhost_stats',
                'create_query' =>
                    'CREATE TABLE "mhost_stats" (id INTEGER PRIMARY '
                    . 'KEY, date DATE, counter_name TEXT, counter_value TEXT, '
                    . 'subsys TEXT, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'counter',
                'type'   => 'lookup',
                'name'   => 'counters',
                'create_query' =>
                    'CREATE TABLE "counters" (id INTEGER PRIMARY '
                    . 'KEY, counter_name TEXT, subsys TEXT, table_name TEXT, unit TEXT, counter_type TEXT)'
            },
            {   'source'       => 'host',
                'type'         => 'lookup',
                'name'         => 'hosts',
                'create_query' => 'CREATE TABLE "hosts" (id INTEGER PRIMARY '
                    . 'KEY, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'ontap',
                'type'   => 'data',
                'name'   => 'node_data',
                'create_query' =>
                    'CREATE TABLE "node_data" (id INTEGER PRIMARY '
                    . 'KEY, node_name TEXT, ontap_release TEXT, node_model TEXT)'
            }
        ]
    };

    $self->{'fast'} = {
        'dsn'     => "dbi:SQLite:dbname=$self->{'path'}/lmr_fast.db",
        'options' => {PrintError => 0, RaiseError => 1, AutoCommit => 1},
        'tables'  => [
            {   'source' => 'ontap',
                'type'   => 'statistics',
                'name'   => 'ontap_statistics',
                'create_query' =>
                    'CREATE TABLE "ontap_statistics" (id INTEGER PRIMARY '
                    . 'KEY, date DATE, counter_name TEXT, counter_value TEXT, '
                    . 'subsys TEXT, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'counter',
                'type'   => 'lookup',
                'name'   => 'counters',
                'create_query' =>
                    'CREATE TABLE "counters" (id INTEGER PRIMARY '
                    . 'KEY, counter_name TEXT, subsys TEXT, table_name TEXT, unit TEXT, counter_type TEXT)'
            },
            {   'source'       => 'host',
                'type'         => 'lookup',
                'name'         => 'hosts',
                'create_query' => 'CREATE TABLE "hosts" (id INTEGER PRIMARY '
                    . 'KEY, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'ontap',
                'type'   => 'data',
                'name'   => 'node_data',
                'create_query' =>
                    'CREATE TABLE "node_data" (id INTEGER PRIMARY '
                    . 'KEY, node_name TEXT, ontap_release TEXT, node_model TEXT)'
            }
        ]
    };

    $self->{'clients'} = {
        'dsn'     => "dbi:SQLite:dbname=$self->{'path'}/lmr_clients.db",
        'options' => {PrintError => 0, RaiseError => 1, AutoCommit => 1},
        'tables'  => [
            {   'source' => 'client',
                'type'   => 'statistics',
                'name'   => 'client_statistics',
                'create_query' =>
                    'CREATE TABLE "client_statistics" (id INTEGER PRIMARY '
                    . 'KEY, date DATE, counter_name TEXT, counter_value TEXT, '
                    . 'subsys TEXT, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'counter',
                'type'   => 'lookup',
                'name'   => 'counters',
                'create_query' =>
                    'CREATE TABLE "counters" (id INTEGER PRIMARY '
                    . 'KEY, counter_name TEXT, subsys TEXT, table_name TEXT, unit TEXT, counter_type TEXT)'
            },
            {   'source'       => 'host',
                'type'         => 'lookup',
                'name'         => 'hosts',
                'create_query' => 'CREATE TABLE "hosts" (id INTEGER PRIMARY '
                    . 'KEY, hostname TEXT, hosttype TEXT)'
            }
        ]
    };

    $self->{'single'} = {
        'dsn'     => "dbi:SQLite:dbname=$self->{'path'}/lmr.db",
        'options' => {PrintError => 0, RaiseError => 1, AutoCommit => 1},
        'tables'  => [
            {   'source' => 'client',
                'type'   => 'statistics',
                'name'   => 'client_statistics',
                'create_query' =>
                    'CREATE TABLE "client_statistics" (id INTEGER PRIMARY '
                    . 'KEY, date DATE, counter_name TEXT, counter_value TEXT, '
                    . 'subsys TEXT, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'ontap',
                'type'   => 'statistics',
                'name'   => 'ontap_statistics',
                'create_query' =>
                    'CREATE TABLE "ontap_statistics" (id INTEGER PRIMARY '
                    . 'KEY, date DATE, counter_name TEXT, counter_value TEXT, '
                    . 'subsys TEXT, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'ontap',
                'type'   => 'memory',
                'name'   => 'ontap_memory',
                'create_query' =>
                    'CREATE TABLE "ontap_memory" (id INTEGER PRIMARY '
                    . 'KEY, date DATE, counter_name TEXT, counter_value TEXT, '
                    . 'subsys TEXT, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'counter',
                'type'   => 'lookup',
                'name'   => 'counters',
                'create_query' =>
                    'CREATE TABLE "counters" (id INTEGER PRIMARY '
                    . 'KEY, counter_name TEXT, subsys TEXT, table_name TEXT, unit TEXT, counter_type TEXT)'
            },
            {   'source'       => 'host',
                'type'         => 'lookup',
                'name'         => 'hosts',
                'create_query' => 'CREATE TABLE "hosts" (id INTEGER PRIMARY '
                    . 'KEY, hostname TEXT, hosttype TEXT)'
            },
            {   'source' => 'ontap',
                'type'   => 'data',
                'name'   => 'node_data',
                'create_query' =>
                    'CREATE TABLE "node_data" (id INTEGER PRIMARY '
                    . 'KEY, node_name TEXT, ontap_release TEXT, node_model TEXT)'
            }
        ]
    };

    bless($self, $class);

    $Log->exit();
    return $self;
} ## end sub new

# Gets the table name based on source and type
sub get_table_name() {

    $Log->enter();

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   database => {type => SCALAR},
            source   => {type => SCALAR},
            type     => {type => SCALAR}
        }
    );
    my @tables;
    foreach my $table (@{$self->{$opts{'database'}}{'tables'}}) {
        if (   $table->{'source'} eq $opts{'source'}
            && $table->{'type'} eq $opts{'type'}) {
            push(@tables, $table->{'name'});
        }
    }
    $Log->debug("Tables: " . Dumper(\@tables));
    $Log->exit();
    return join(',', @tables);
} ## end sub get_table_name()

# Creates DB tables
sub create_table {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database => {type => SCALAR},
            query    => {type => SCALAR}
        }
    );

    my $database = $opts{database};
    my $query    = $opts{query};
    my $var      = NATE::ParamSet->new(global => 1);
    $self->connect_db($database);
    $Log->debug("SQL Query: $query");
    eval { $self->{$database}->{'handle'}->do($query) || die $DBI::errstr; };
    if ($@) {
        $Log->debug("Unable to create table: $@");
    }
} ## end sub create_table

# Writes data to the specified table and also lookup tables
sub write {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database => {type => SCALAR,          optional => 0},
            stats    => {type => HASHREF,         optional => 0},
            table    => {type => SCALAR,          optional => 0},
            datetime => {type => SCALAR | UNDEF,  optional => 1},
            units    => {type => HASHREF | UNDEF, optional => 1},
        }
    );

    my $database      = $opts{'database'};
    my $counter_units = $opts{'units'};
    $self->connect_db($database);

    my $table = $opts{'table'};
    my %stats;
    my %hosts;
    my %counters;
    my @subsystems;
    my $row_count = 0;
    my @queries;
    my $query    = '';
    my $subquery = '';

    if ($table eq 'node_data') {
        my $data = $opts{'stats'};
        $query
            = "INSERT INTO \"$table\" (node_name, ontap_release, node_model) ";
        $Log->debug("Data received in write is: " . Dumper(\%$data));
        foreach my $node (keys %$data) {
            my $nodename = $node;
            my $version  = $data->{$node}->{'version'};
            my $model    = $data->{$node}->{'model'};
            $query .= " SELECT '$nodename', '$version', '$model' UNION";
        }
        $query =~ s/UNION$//;
        $Log->debug("SQL Query: $query");
        push(@queries, $query);
    } else {
        delete $opts{stats}->{'Timestamp'};

        # Remove unwanted counters and format conter names
        foreach my $cn (keys(%{$opts{stats}})) {

            # Check if there is a value
            if (!defined($opts{stats}->{$cn}) || $opts{stats}->{$cn} eq '') {
                $opts{stats}->{$cn} = '';
                $Log->debug("$cn not defined, skipping");
                next;
            }

            my $counter_value = $opts{stats}->{$cn};

            # Get rid of double colons in memory calls
            $cn =~ s/::/!!/g;

            my ($hostname, $subsys, $instance, $counter);
            my $fields = $self->parse_counter_name($cn);
            if (@{$fields} == 4) {
                $counter  = ${$fields}[0];
                $instance = ${$fields}[1];
                $subsys   = ${$fields}[2];
                $hostname = ${$fields}[3];
            } elsif (@{$fields} == 3) {
                $counter  = ${$fields}[0];
                $subsys   = ${$fields}[1];
                $hostname = ${$fields}[2];
            }

            # Make sure we only write numerical values to the DB
            if ($counter_value =~ /[^\d+\.+\+\-]|\d+\.\d+\.\d+\.\d+/) {
                $Log->debug("The $opts{stats}->{$cn} is an invalid number");
                next;
            }

            if (!$hostname) {
                $Log->comment("Unable to determine hostname from $cn\n"
                        . Dumper($fields));
                next;
            }

            my $hosttype;
            if ($hosts{$hostname}) {
                $hosttype = $hosts{$hostname};
            } else {
                eval { $hosttype = host($hostname)->hosttype(); };
                $hosts{$hostname} = $hosttype;
            }

            if (!$hosttype) { $hosttype = 'filer-ng-cserver'; }

            my $counter_name = $subsys . ':';
            if ($instance && $instance ne $subsys) {
                $counter_name .= $instance . ':';
            }
            $counter_name .= $counter . ':';

            if (!grep { $_ eq $counter_name } keys(%counters)) {
                $counters{$counter_name} = $subsys;
            }

            my $msg
                = "Original Name: $cn\n"
                . "     Hostname: $hostname\n"
                . "     Hosttype: $hosttype\n"
                . "    Subsystem: $subsys\n";
            if ($instance) {
                $msg = "     Instance: $instance\n";
            }
            $msg
                .= "      Counter: $counter\n"
                . "Modified Name: $counter_name\n"
                . "        Value: $counter_value";

            $Log->debug($msg);
            my $h = $hostname;
            my $c = $counter_name;
            if (!$stats{$h}{'hosttype'}) {
                $stats{$h}{'hosttype'} = $hosttype;
            }
            $stats{$h}{'counters'}{$c}{'subsys'} = $subsys;
            $stats{$h}{'counters'}{$c}{'value'}  = $counter_value;
            if ($hosttype eq 'filer-ng-cserver') {
                $stats{$h}{'counters'}{$c}{'counter_type'} = 'cluster';
            } elsif ($hosttype =~ /filer/) {
                $stats{$h}{'counters'}{$c}{'counter_type'} = 'node';
            }
            $row_count++;
        }

        # Create the queries
        my $row            = 0;
        my $rows_per_query = 0;

        my $date = strftime("%Y-%m-%d %H:%M:%S", gmtime());
        if ($opts{'datetime'}) { $date = $opts{'datetime'}; }

        # Start a new query
        $query = '';
        $query
            = "INSERT INTO \"$table\" (date,counter_name,counter_value,subsys,hostname,hosttype)";
        @queries = ();
        foreach my $h (keys(%stats)) {
            foreach my $c (keys(%{$stats{$h}{'counters'}})) {
                my $hosttype = $stats{$h}{'hosttype'};
                my $value    = $stats{$h}{'counters'}{$c}{'value'};
                my $subsys   = $stats{$h}{'counters'}{$c}{'subsys'};

                $query
                    .= " SELECT '$date', '$c', '$value', '$subsys', '$h', '$hosttype' UNION";
                $row++;
                $rows_per_query++;

                if ($rows_per_query > 450) {

                    # The max number of queries are exceeded, save and start a
                    # new one
                    $query =~ s/UNION$//;
                    push(@queries, $query);

                    $query
                        = "INSERT INTO \"$table\" (date, counter_name,counter_value,subsys,hostname,hosttype)";
                    $rows_per_query = 0;
                    next;

                } elsif ($row == $row_count) {

                    # We are at the end of rows save the query
                    $query =~ s/UNION$//;
                    push(@queries, $query);
                }

            }
        }

        # Get the current entries of the hostname lookup table
        $query = "SELECT DISTINCT hostname,hosttype FROM hosts";

        my $rows;
        eval {
            $rows = $self->{$database}->{'handle'}->selectall_arrayref($query)
                || die $DBI::errstr;
        };
        if ($@) { $Log->debug($@); }

        my @db_hostnames;
        foreach my $row (@{$rows}) {
            my ($hostname, $hosttype) = @{$row};
            if (!grep { $_ eq $hostname } @db_hostnames) {
                push(@db_hostnames, $hostname);
            }
        }

        $query          = "INSERT INTO \"hosts\" (hostname,hosttype)";
        $row            = 0;
        $rows_per_query = 0;
        foreach my $hostname (keys(%stats)) {
            my $hosttype = $stats{$hostname}{'hosttype'};

            if (!grep { $_ eq $hostname } @db_hostnames) {
                $query .= " SELECT '$hostname', '$hosttype' UNION";
                $row++;
                $rows_per_query++;

                if ($rows_per_query > 450) {

                # The max number of queries exceeded, save and start a new one
                    $query =~ s/UNION$//;
                    push(@queries, $query);
                    $query = "INSERT INTO \"hosts\" (hostname,hosttype)";
                    $rows_per_query = 0;
                    next;

                }

# elsif ($row <= keys(%hosts)) {
# 	$Log->comment("query is: $query and row is: $row and hosts are: ".keys(%hosts));
# We are at the end of rows save the query
#     $query =~ s/UNION$//;
#     push(@queries, $query);
# }

            }
        }

        if ($row <= keys(%hosts) && $row != 0) {

            # We are at the end of rows save the query
            $query =~ s/UNION$//;
            push(@queries, $query);
        }

        # Get the current entries of the counters lookup table
        $query
            = "SELECT DISTINCT counter_name,subsys,table_name,unit,counter_type FROM counters";

        eval {
            $rows = $self->{$database}->{'handle'}->selectall_arrayref($query)
                || die $DBI::errstr;
        };
        if ($@) { $Log->debug($@); }

        my $db_counters = {};
        foreach my $row (@{$rows}) {
            my ($counter_name, $subsys, $table_name, $unit, $counter_type)
                = @{$row};
            if (!grep { $_ eq $counter_name } keys(%{$db_counters})) {
                $db_counters->{$counter_name}->{'subsys'}     = $subsys;
                $db_counters->{$counter_name}->{'table_name'} = $table_name;
            }
        }

        my $parsed_counter_units = {};

        #parse the counter names in %counter_units to match those in %stats
        if (keys %$counter_units) {
            foreach my $k (keys %$counter_units) {
                my $fields       = $self->parse_counter_name($k);
                my $counter      = ${$fields}[0];
                my $instance     = ${$fields}[1];
                my $subsys       = ${$fields}[2];
                my $hostname     = ${$fields}[3];
                my $counter_name = $subsys . ":" . $counter . ":";
                $parsed_counter_units->{$counter_name} = $counter_units->{$k};
            }
        }

        $query
            = "INSERT INTO \"counters\" (counter_name,subsys,table_name,unit,counter_type)";
        $row            = 0;
        $rows_per_query = 0;
        foreach my $h (keys(%stats)) {
            foreach my $c (keys(%{$stats{$h}{'counters'}})) {
                my $subsys       = $stats{$h}{'counters'}{$c}{'subsys'};
                my $counter_type = $stats{$h}{'counters'}{$c}{'counter_type'};
                my $db_table_name = '';
                my $unit;
                if (defined($db_counters->{$c}->{'table_name'})) {
                    $db_table_name = $db_counters->{$c}->{'table_name'};
                }

                if (!defined($db_table_name)) {
                    $Log->comment("\$c:$c not defined \$db_table_name");
                }

                #get the counter unit or set it to none
                if (keys %$parsed_counter_units
                    && defined $parsed_counter_units->{$c}) {
                    $unit = $parsed_counter_units->{$c};
                } else {
                    $unit = "none";
                }

                my $found = grep { $_ eq $c } keys(%{$db_counters});

                if (!$found || ($found && $table ne $db_table_name)) {

                    $query
                        .= " SELECT '$c', '$subsys', '$table', '$unit', '$counter_type' UNION";
                    $row++;
                    $rows_per_query++;

                    if ($rows_per_query > 450) {

                        # The max number of queries exceeded, save and start
                        # a new one
                        $query =~ s/UNION$//;
                        push(@queries, $query);
                        $query
                            = "INSERT INTO \"counters\" (counter_name,subsys,table_name,unit,counter_type)";
                        $rows_per_query = 0;
                        next;

                    } elsif ($row == $row_count) {

                        # We are at the end of rows save the query
                        $query =~ s/UNION$//;
                        push(@queries, $query);
                    }
                }
            }
        }
    }

    # Execute the SQL queries
    foreach my $query (@queries) {
        eval {
            $self->{$database}->{'handle'}->do($query) || die $DBI::errstr;
        };
        if ($@) {
            $Log->comment($@);
        }
    }

} ## end sub write

# Conenct to the SQL DB
sub connect_db {
    my ($self, $database) = @_;
    my $dsn     = $self->{$database}->{'dsn'};
    my $options = $self->{$database}->{'options'};
    if (!ref($self->{$database}->{'handle'})) {
        my $handle = DBI->connect($dsn, '', '', $options)
            || $Log->debug($DBI::errstr);
        $self->{$database}->{'handle'} = $handle;
    } ## end

} ## end sub connect_db

# Returns a list of counters
sub fetch_counters {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database     => {type => SCALAR | UNDEF, optional => 1},
            subsys       => {type => SCALAR | UNDEF, optional => 1},
            table        => {type => SCALAR | UNDEF, optional => 1},
            counter_type => {type => SCALAR | UNDEF, optional => 1},
        },
    );

    my @databases;
    if (!$opts{'database'}) {
        @databases = @{$self->{'databases'}};
    } else {
        @databases = split(/,/, $opts{'database'});
    }

    my @counter_names;
    foreach my $database (@databases) {
        my %args = (database => $database);
        if ($opts{'subsys'}) { $args{'subsys'} = $opts{'subsys'}; }
        if ($opts{'table'})  { $args{'table'}  = $opts{'table'}; }
        my $counters = eval { $self->_fetch_counters(%args); };
        if (!$@ && $counters) {
            foreach my $c (@{$counters}) {
                if (!grep { $_ eq $c } @counter_names) {
                    push(@counter_names, $c);
                }
            }
        }
    }
    return \@counter_names;
} ## end sub fetch_counters

sub _fetch_counters {
    my $t0 = [gettimeofday];
    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database => {type => SCALAR,         optional => 0},
            subsys   => {type => SCALAR | UNDEF, optional => 1},
            table    => {type => SCALAR | UNDEF, optional => 1},
            type     => {type => SCALAR | UNDEF, optional => 1},
        },
    );

    my $database = $opts{'database'};
    if (!$opts{subsys}) { $opts{subsys} = ''; }
    my @subsystems = split(',', $opts{subsys});
    my $subsys = '(\'' . join("', '", @subsystems) . '\')';

    my $table = $self->get_table_name(
        database => $database,
        source   => 'counter',
        type     => 'lookup'
    );

    my %connect_options = (PrintError => 0, RaiseError => 1, AutoCommit => 0);
    $self->connect_db($database);

    my $query = "SELECT DISTINCT * FROM $table ";

    if ($opts{subsys} || $opts{table} || $opts{type}) {
        $query .= 'WHERE ';

        if ($opts{subsys}) {
            $query .= "subsys IN $subsys ";
            if ($opts{table} || $opts{type}) {
                $query .= 'AND ';
            }
        }

        if ($opts{table}) {
            $query .= "table_name = '$opts{'table'}' ";
            if ($opts{type}) {
                $query .= 'AND ';
            }
        }
        if ($opts{type}) {
            $query .= "counter_type = '$opts{'counter_type'}'";
        }
    }

    my $all;
    eval {
        $all = $self->{$database}->{'handle'}->selectall_arrayref($query)
            || die $DBI::errstr;
    };
    if ($@) {
        $Log->comment("SQL Query: $query");
        $Log->comment($@);
    }

    my $dbh = $self->{$database}->{'handle'};
    my $sth = $dbh->prepare("PRAGMA table_info($table)") || die $DBI::errstr;
    $sth->execute() || die $DBI::errstr;
    my @columns;
    while (my $row = $sth->fetchrow_arrayref()) {
        push(@columns, @$row[1]);
    }

    my @results;
    foreach my $row (@$all) {
        my ($id, $counter_name, $subsys, $table_name, $unit, $counter_type);
        if (grep(/unit/, @columns) && grep(/counter_type/, @columns)) {
            ($id, $counter_name, $subsys, $table_name, $unit, $counter_type)
                = @$row;
        } elsif (grep(/unit/, @columns)
            && !grep(/counter_type/, @columns)) {
            ($id, $counter_name, $subsys, $table_name, $unit) = @$row;
        } else {
            ($id, $counter_name, $subsys, $table_name) = @$row;
        }
        my %c = (
            'counter_name' => $counter_name,
            'subsys'       => $subsys,
            'table_name'   => $table_name,
            'counter_type' => $counter_type,
        );

        #add unit key only if it exists
        if (defined $unit) {
            $c{'unit'} = $unit;
        }

        push(@results, \%c);
    }

    $Log->debug("Counters: " . Dumper(\@results));

    my $elapsed = tv_interval($t0, [gettimeofday]);
    $Log->debug("fetch_counters completed in $elapsed seconds");

    return \@results;
} ## end sub _fetch_counters

# Returns data for a given set of counters
sub fetch_data {

    my $t0 = [gettimeofday];
    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   hostname => {type => SCALAR,         optional => 0},
            database => {type => SCALAR | UNDEF, optional => 1},
            limit    => {type => SCALAR | UNDEF, optional => 1},
            prune    => {type => SCALAR | UNDEF, default  => 2},
            range    => {type => SCALAR | UNDEF, optional => 1},
            table    => {type => SCALAR,         optional => 0},
            samples  => {type => SCALAR | UNDEF, optional => 1},
            subsys   => {type => SCALAR | UNDEF, optional => 1},
            xfield   => {type => SCALAR | UNDEF, optional => 1},
            yfields  => {type => SCALAR | UNDEF, optional => 1},
            yregex   => {type => SCALAR | UNDEF, optional => 1},
            output   => {type => SCALAR,         default  => 'lmr'}
        },
    );

    my $legend = [];
    my $data   = [];
    my @databases;
    if (!$opts{'database'}) {
        @databases = @{$self->{'databases'}};
    } else {
        @databases = split(',', $opts{'database'});
    }
    my @tables = split(/,/, $opts{'table'});

    foreach my $database (@databases) {
        foreach my $table (@tables) {
            my %args = (
                database => $database,
                table    => $table,
                hostname => $opts{'hostname'},
                subsys   => $opts{'subsys'}
            );

            if ($opts{'limit'})   { $args{'limit'}   = $opts{'limit'}; }
            if ($opts{'prune'})   { $args{'prune'}   = $opts{'prune'}; }
            if ($opts{'range'})   { $args{'range'}   = $opts{'range'}; }
            if ($opts{'samples'}) { $args{'samples'} = $opts{'samples'}; }
            if ($opts{'xfield'})  { $args{'xfield'}  = $opts{'xfield'}; }
            if ($opts{'yfields'}) { $args{'yfields'} = $opts{'yfields'}; }
            if ($opts{'yregex'})  { $args{'yregex'}  = $opts{'yregex'}; }

            my ($counters, $counter_data) = $self->_fetch_data(%args);

            foreach my $counter (@{$counters}) {
                if (!grep { $_ eq $counter } @{$legend}) {
                    push(@{$legend}, $counter);
                }
            }

            ($legend, $data) = $self->merge_arrays(
                'legend1' => $legend,
                'data1'   => $data,
                'legend2' => $counters,
                'data2'   => $counter_data
            );
        }
    }

    if ($opts{'output'} eq 'aoh') {
        my $aoh = [];
        for (my $i = 0; $i < @{$legend}; $i++) {
            push(@{$aoh}, {name => ${$legend}[$i], data => $data->[$i]});
        }
        return $aoh;
    } else {
        return ($legend, $data);
    }
} ## end sub fetch_data

# The real fetch data work is done here
sub _fetch_data {

    my $t0 = [gettimeofday];
    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database => {type => SCALAR,         optional => 0},
            xfield   => {type => SCALAR | UNDEF, optional => 1},
            samples  => {type => SCALAR | UNDEF, optional => 1},
            yfields  => {type => SCALAR | UNDEF, optional => 1},
            yregex   => {type => SCALAR | UNDEF, optional => 1},
            subsys   => {type => SCALAR | UNDEF, optional => 1},
            hostname => {type => SCALAR,         optional => 0},
            table    => {type => SCALAR,         optional => 0},
            range    => {type => SCALAR | UNDEF, optional => 1},
            limit    => {type => SCALAR | UNDEF, optional => 1},
            prune    => {type => SCALAR | UNDEF, default  => 2}
        },
    );

    my $hashref;
    my @time_stamp;
    my @counters;
    my @counter_values;
    my $database = $opts{'database'};
    my $table    = $opts{'table'};
    my $prune    = $opts{'prune'};
    my $xfield   = 'datetime';

    $self->connect_db($database);

    my %hosttypes;
    my $results = $self->fetch_hosts();
    foreach my $r (@{$results}) {
        $hosttypes{$r->{'hostname'}} = $r->{'hosttype'};
    }

    my @subsystems = split(',', $opts{subsys});
    my $subsys = '(\'' . join("', '", @subsystems) . '\')';

    my @hostnames = split(',', $opts{hostname});
    my $hostname = '(\'' . join("', '", @hostnames) . '\')';

    my $date = sql_time_conversion('date', 'localtime');
    my $query = "SELECT $date,counter_name,counter_value,hostname FROM "
        . "$table WHERE subsys IN $subsys AND hostname IN $hostname  ";

    # If yfields is specified we can limit the size of the query to only
    # the counters we want assuming we string the hostnames from them
    if ($opts{'yfields'}) {
        my @yfields = split(',', $opts{yfields});
        my @counter_names;
        foreach my $yfield (@yfields) {
            $yfield =~ s/:$//;
            my $counter_name;

            # Parse the counter name and remove the hostname if it exists
            my $fields = $self->parse_counter_name($yfield);
            if (@{$fields} == 3) {
                my $f0 = ${$fields}[0];
                my $f1 = ${$fields}[1];
                my $f2 = ${$fields}[2];

                # The third field could be a hostname or
                # subsys, trim accordingly
                if (grep { $_ eq $f2 } @subsystems) {
                    $counter_name = "$f2:";
                }
                $counter_name .= "$f1:$f0:";

            } else {
                $counter_name = "$yfield:";
            }

            if (!grep { $_ eq $counter_name } @counter_names) {
                push(@counter_names, $counter_name);
            }
        }
        my $yf = '(\'' . join("', '", @counter_names) . '\')';
        $query .= "AND counter_name in $yf";
    }

    if ($opts{'range'}) {
        my ($start, $end) = split(',', $opts{'range'});
        $start = sql_time_conversion($start, 'utc');
        $end   = sql_time_conversion($end,   'utc');
        $query .= "AND date BETWEEN $start AND $end ";
    }

    if ($opts{'limit'}) {
        $query .= "LIMIT $opts{'limit'}";
    }

    # Execute the query
    my $all;
    eval {
        $all = $self->{$database}->{'handle'}->selectall_arrayref($query)
            || die $DBI::errstr;
    };
    if ($@) {
        $Log->debug("SQL Query: $query");
        $Log->debug($@);
    }

    # Get all sample dates
    my @samples;
    foreach my $row (@$all) {
        my ($date, $counter, $counter_value, $hostname) = @$row;
        if (!grep { $_ eq $date } @samples) {
            push(@samples, $date);
        }
    }

    # If samples is specified then figure out the sample dates for the
    # rows we will return
    if ($opts{'samples'} && @samples > $opts{'samples'}) {
        splice(@samples, 0, "-$opts{'samples'}");
    }

    foreach my $row (@$all) {

        # Get the data from the row
        my ($date, $counter, $counter_value, $hostname) = @$row;

        # If the date is not in the array of dates we want then skip
        if (!grep { $_ eq $date } @samples) {
            next;
        }

        # node based counters can be in the form of
        #
        # nblade_cifs:nst-fas3250-2:memory_budget_consumed
        # system:cpu_busy
        #
        # For instances that have node names in them we need to make sure
        # we add the only appropriate node names instead on blindly adding
        # node names to every counter because the following is invalid
        #
        # nst-fas3250-1:nblade_cifs:nst-fas3250-2:memory_budget_consumed
        if (@hostnames > 1) {
            $counter =~ s/:$//;
            my $fields = $self->parse_counter_name($counter);
            if (@{$fields} == 3) {
                my $instance = ${$fields}[1];
                if (   $hosttypes{$hostname} ne 'filer-ng-cserver'
                    && (grep { $_ eq $instance } @hostnames)
                    && $instance ne $hostname) {
                    next;
                }
            }
            $counter = "$hostname:$counter";
        }

        if (!grep { $_ eq $counter } @counters) {
            push(@counters, $counter);
        }
        my $value = 0;
        if (defined($value)) {
            $value = $counter_value;
        }

        if ($value eq '-') { $value = 0; }

        $hashref->{$counter}->{$date} = $value;

        if (!grep { $_ eq $date } @{$hashref->{$xfield}}) {
            push(@{$hashref->{$xfield}}, $date);
        }

    }

    my @keys = keys(%{$hashref});

    # Decide on legend keys based on yfields
    my @legends;
    if ($opts{not_yfields}) {
        $Log->debug('Filtering counters by not_yfields');
        my @not_yfields = split(',', $opts{not_yfields});
        foreach my $counter (@counters) {
            if (!grep { $_ eq $counter } @not_yfields) {
                push(@legends, $counter);
            }
        }
    } elsif ($opts{yfields}) {
        chomp($opts{yfields});
        $Log->debug('Filtering counters by yfields');
        push(@legends, split(',', $opts{yfields}));
    } elsif ($opts{yregex}) {
        $Log->debug('Filtering counters by yregex');
        @legends = grep(/$opts{yregex}/, @counters);
    }

    # If there were multiple hostnames specified we added hostnames to
    # counter names to make them unique. We need to do the same for the
    # legends so they match
    if (@hostnames > 1) {
        my @modified_legends;
        foreach my $counter (@legends) {
            foreach my $hostname (@hostnames) {
                $counter =~ s/:$//;
                my $fields = $self->parse_counter_name($counter);
                if (@{$fields} == 3) {
                    my $instance = ${$fields}[1];
                    if (   $hosttypes{$hostname} ne 'filer-ng-cserver'
                        && (grep { $_ eq $instance } @hostnames)
                        && $instance ne $hostname) {
                        next;
                    }
                }
                my $cn = "$hostname:$counter";
                if (!grep { $_ eq $cn } @modified_legends) {
                    push(@modified_legends, $cn);
                }
            }
        }
        @legends = @modified_legends;
    }

    # Filter out only required data
    my @data;
    my $timezone_offset = strftime("%z", localtime());
    foreach my $key (@legends) {
        $Log->debug("working on data for $key");
        my @dates = natsort(keys(%{$hashref->{$key}}));

        my @redundant_data;
        my $last_value;
        my @trimmed_dates;

        if (!$prune) {
            @trimmed_dates = @dates;
        }

        # Walk through each data point and weed out redundant data
        foreach my $date (@dates) {

            if (!$prune) { last; }

            # Get the data point's value
            my $value = $hashref->{$key}->{$date};

            if (!defined($last_value)) {

                # If this is the first data point, add it to the array
                push(@redundant_data, $date);
                $last_value = $value;
                $Log->debug("Setting initial last_value $value");
            } elsif ($value == $last_value) {

                # If the current data point is the same as the last
                # add it to the redundant_data array
                push(@redundant_data, $date);
                $last_value = $value;
                $Log->debug("Saving redundant value $value");
            }

            # If the current value does not match the last value or we have
            # reached the end of the dates array process the redundant data
            if ($value != $last_value || $date eq $dates[$#dates]) {
                $Log->debug("Reached the end of redundant data or samples\n"
                        . "Current Value:$value Last Value:$last_value\n"
                        . "Current Date:$date Last Date::$dates[$#dates]");

                # If there is more than 3 points of redundant data get the
                # subset of it
                if (@redundant_data > $prune) {
                    $Log->debug("Getting data subset");
                    my $subset = $self->get_array_subset(
                        'source'   => \@redundant_data,
                        'elements' => $prune
                    );
                    @redundant_data = @{$subset};
                }

                # Save the dates of the redundant data
                foreach my $d (@redundant_data) {
                    push(@trimmed_dates, $d);
                }

                # Set the last value to the current value
                $last_value = $value;

                # Reset the redundant data array so we start the process
                # over again
                @redundant_data = ();
                push(@redundant_data, $date);
            }
        }

        if (@redundant_data) {
            foreach my $d (@redundant_data) {
                push(@trimmed_dates, $d);
            }
        }

        my @data_points;
        foreach my $date (@trimmed_dates) {

            # Get the value of the current data point
            my $value = $hashref->{$key}->{$date};

            # Skip if there is no value defined
            if (!defined($value)) { next; }

            # Convert the UTC time to local time
            if ($date =~ /(\d+)(\/|-)(\d+)(\/|-)(\d+)\s+(\d+):(\d+):(\d+)/) {
                $date = $self->get_epoch_time(
                    'date'   => $date,
                    'mode'   => 'gm',
                    'format' => 'Y-m-d H:M:S'
                );

                # Add the timezone offset
                $date += $timezone_offset;

                # Convert to miliseconds for !@#$%^& javascript
                $date *= 1000;
            }

            my @coordinates = ($date, $value);
            push(@data_points, \@coordinates);
        }
        push(@data, \@data_points);
    }

    # Take a final pass on the data and remove the counter name from
    # $legends and the series from $data if the series has no data.
    my @filtered_legends;
    my @filtered_data;
    for (my $i = 0; $i < @legends; $i++) {
        if (!@{$data[$i]}) { next; }
        push(@filtered_legends, $legends[$i]);
        push(@filtered_data,    $data[$i]);
    }
    $Log->debug('Legends:' . Dumper(\@filtered_legends));
    $Log->debug('Data:' . Dumper(\@filtered_data));

    my $elapsed = tv_interval($t0, [gettimeofday]);
    $Log->debug("fetch_data completed in $elapsed seconds");

    return (\@filtered_legends, \@filtered_data);

} ## end sub _fetch_data

# Creates a Highchars.com graph
sub graph {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   container  => {type => SCALAR | UNDEF},
            title      => {type => SCALAR | UNDEF, optional => 1},
            ylabel     => {type => SCALAR | UNDEF, optional => 1},
            data       => {type => ARRAYREF | HASHREF | UNDEF, optional => 1},
            legend     => {type => ARRAYREF | HASHREF | UNDEF, optional => 1},
            width      => {type => SCALAR | UNDEF, default => '100%'},
            height     => {type => SCALAR | UNDEF, default => '100%'},
            type       => {type => SCALAR | UNDEF, default => 'spline'},
            item_width => {type => SCALAR | UNDEF, default => 840}
        },
    );

    my $container = $opts{'container'};
    my $title     = $opts{'title'};
    my $ylabel    = $opts{'ylabel'};
    my $data      = $opts{'data'};
    my $legend    = $opts{'legend'};
    my $type      = $opts{'type'};
    my $height    = $opts{'height'} . 'px';
    my $width     = $opts{'width'} . 'px';
    my $series    = 0;

    my @dates;
    foreach my $name (@{$legend}) {
        foreach my $d (@{$data->[$series]}) {
            my ($date, $value) = @{$d};
            if (!grep { $_ eq $date } @dates) {
                push(@dates, $date);
            }
        }
    }
    natsort(@dates);
    my $subset = $self->get_array_subset(
        'source'   => \@dates,
        'elements' => 15
    );

    my $str = <<END;

<script type='text/javascript'>
    var chart;
    jQuery(document).ready(function() {
        chart = new Highcharts.Chart({
            chart: {
                renderTo: "$container",
                Type: "$type",
                zoomType: 'x'
            },
            title: {
                text: "$title",
                x: -20 //center
            },
            xAxis: {
               type: 'datetime',
	       labels: {
	           rotation: 90,
		   formatter: function() {
		       return Highcharts.dateFormat('\%Y\%m\%d_\%H\%M', this.value);
	           }

	       },
	       style: {
	            fontWeight: 'bold'
	       }

            },
            yAxis: {
                title: {
                    text: "$ylabel"
                },
                plotLines: [ {value: 0,width: 1, color: '#808080'} ] 
            },
            plotOptions: { 
            column: {
                    dataLabels: {
                        enabled: true,rotation: 90,y: -20
                    }
                }
            },
	    tooltip: { formatter: function() { return '<b>'+ this.series.name +'</b><br/>' +Highcharts.dateFormat('\%Y\%m\%d_\%H\%M\%S', this.x) + ': ' + this.y; } },
            series: [
END

    $series = 0;
    foreach my $name (@{$legend}) {
        $str .= tabs(4) . "{name: \'$name\',\n";
        $str .= tabs(4) . " data: [\n";

        my $xvalue = 0;
        foreach my $d (@{$data->[$series]}) {
            my $coordinates = $self->array_to_list($d);
            $str .= tabs(5) . " $coordinates";

            if ($xvalue < $#{$data->[$series]}) {
                $str .= ",";
            }
            $str .= "\n";
            $xvalue++;
        }
        $str .= tabs(4) . " ]\n";
        $str .= tabs(4) . "}";
        if ($series < $#{$legend}) {
            $str .= ",";
        }
        $str .= "\n";
        $series++;
    }

    $str .= <<END;
            ]
        });
    });
</script>

END

    $str
        .= "<div id=\"$container\" style=\"width:$width; height:$height; margin: 0 auto\"></div>\n";
    return $str;

} ## end sub graph

# Adds TAB spacing
sub tabs {
    my $tabs = shift;
    my $spaces = shift || 4;

    my $string;
    for (my $t = 1; $t <= $tabs; $t++) {
        for (my $s = 1; $s <= $tabs; $s++) {
            $string .= ' ';
        }
    }

    return $string;
} ## end sub tabs

# Converts an array to a javascript list
sub array_to_list {
    my ($self, $arr) = @_;
    if (!@{$arr}) { @{$arr} = (); }
    my $str;
    $str = '[';
    $str .= join ",", @{$arr};
    $str .= ']';
    return $str;
} ## end sub array_to_list

# Returns a arrayref of subsystems
sub fetch_subsystems {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database => {type => SCALAR | UNDEF, optional => 1},
            subsys   => {type => SCALAR | UNDEF, optional => 1}
        },
    );

    my $subsys = $opts{'subsys'};

    my @databases;
    if (!$opts{'database'}) {
        @databases = @{$self->{'databases'}};
    } else {
        @databases = split(/,/, $opts{'database'});
    }

    my @subsystem_names;
    foreach my $database (@databases) {
        my $subsystems = $self->_fetch_subsystems(
            database => $database,
            subsys   => $subsys
        );
        if ($subsystems) {
            foreach my $s (@{$subsystems}) {
                if (!grep { $_ eq $s } @subsystem_names) {
                    push(@subsystem_names, $s);
                }
            }
        }
    }
    return \@subsystem_names;
} ## end sub fetch_subsystems

sub _fetch_subsystems {

    my $t0 = [gettimeofday];
    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database => {type => SCALAR,         optional => 0},
            subsys   => {type => SCALAR | UNDEF, optional => 1},
        },
    );

    my $database = $opts{'database'};
    my $subsys   = $opts{'subsys'};
    my $table    = $self->get_table_name(
        database => $opts{'database'},
        source   => 'counter',
        type     => 'lookup'
    );

    $self->connect_db($database);

    my $query = "SELECT DISTINCT subsys,table_name FROM $table ";
    if ($subsys) {
        $query .= "WHERE subsys = '$subsys'";
    }
    $Log->debug("SQL Query: $query");

    my $rows;
    eval {
        $rows = $self->{$database}->{'handle'}->selectall_arrayref($query);
    };
    my @subsystems;
    foreach my $row (@{$rows}) {

        # Get the data from the row
        my ($subsys, $table_name) = @{$row};
        my %s = (
            'subsys'     => $subsys,
            'table_name' => $table_name
        );
        push(@subsystems, \%s);
    }

    my $elapsed = tv_interval($t0, [gettimeofday]);
    $Log->debug("fetch_subsystems completed in $elapsed seconds");

    return \@subsystems;
} ## end sub _fetch_subsystems

sub fetch_hosts {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {database => {type => SCALAR | UNDEF, optional => 1}},
    );

    my @databases;
    if (!$opts{'database'}) {
        @databases = @{$self->{'databases'}};
    } else {
        @databases = split(/,/, $opts{'database'});
    }

    my @hostnames;
    foreach my $database (@databases) {
        my $hosts = $self->_fetch_hosts($database);
        if ($hosts) {
            foreach my $h (@{$hosts}) {
                if (!grep { $_->{'hostname'} eq $h->{'hostname'} } @hostnames)
                {
                    push(@hostnames, $h);
                }
            }
        }
    }
    return \@hostnames;
} ## end sub fetch_hosts

# Returns a arrayref of hosts
sub _fetch_hosts {

    my $t0       = [gettimeofday];
    my $self     = shift;
    my $database = shift;

    my $table = $self->get_table_name(
        database => $database,
        source   => 'host',
        type     => 'lookup'
    );
    $self->connect_db($database);

    # Add the current hostname to the list of subsystems if not
    # already there
    my $query = "SELECT hostname,hosttype FROM $table";
    $Log->debug("SQL Query: $query");
    my $rows;
    eval {
        $rows = $self->{$database}->{'handle'}->selectall_arrayref($query);
    };
    my @hostnames;
    foreach my $row (@{$rows}) {

        # Get the data from the row
        my ($hostname, $hosttype) = @{$row};
        my %h = (
            'hostname' => $hostname,
            'hosttype' => $hosttype
        );
        push(@hostnames, \%h);
    }
    $Log->debug("Hosts: " . Dumper(\@hostnames));
    my $elapsed = tv_interval($t0, [gettimeofday]);
    $Log->debug("fetch_hosts completed in $elapsed seconds");

    return \@hostnames;
} ## end sub _fetch_hosts

sub fetch_node_data {
    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   database => {type => SCALAR | UNDEF, optional => 1},
            nodename => {type => SCALAR | UNDEF, optional => 1},
        },
    );

    my @databases;
    if (!$opts{'database'}) {
        @databases = @{$self->{'databases'}};
    } else {
        @databases = split(/,/, $opts{'database'});
    }

    my @nodedata;
    my %result;
    foreach my $database (@databases) {
        if ($database =~ /clients/) { next; }
        my %args = (database => $database);
        $args{'nodename'} = $opts{'nodename'};
        my $data = eval { $self->_fetch_node_data(%args); };
        if ($data) {
            foreach my $d (@{$data}) {
                if (!defined $result{$database}{$d->{'node_name'}}
                    {'ontap_release'}
                    || !grep { $_ eq $d->{'ontap_release'} }
                    @{$result{$database}{$d->{'node_name'}}{'ontap_release'}})
                {
                    push(
                        @{  $result{$database}{$d->{'node_name'}}
                                {'ontap_release'}
                            },
                        $d->{'ontap_release'}
                    );
                }
                $result{$database}{$d->{'node_name'}}{'node_model'}
                    = $d->{'node_model'};
            }
        }
    }

    return \%result;
}    ##end sub fetch_node_data

#returns an array ref of node data
sub _fetch_node_data {

    my $t0 = [gettimeofday];
    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database => {type => SCALAR,         optional => 0},
            nodename => {type => SCALAR | UNDEF, optional => 1},
        },
    );

    my $database = $opts{'database'};
    if (!$opts{'nodename'}) { $opts{'nodename'} = ''; }
    my @nodenames = split(',', $opts{'nodename'});
    my $nodename = '(\'' . join("','", @nodenames) . '\')';

    my $table = $self->get_table_name(
        database => $database,
        source   => 'ontap',
        type     => 'data'
    );

    my %connect_option = (PrintError => 0, RaiseError => 1, AutoCommit => 0);
    $self->connect_db($database);

    my $query = "SELECT node_name, ontap_release, node_model FROM $table ";

    if ($opts{nodename}) {
        $query .= "WHERE node_name IN $nodename";
    }

    my $all;
    eval {
        $all = $self->{$database}->{'handle'}->selectall_arrayref($query)
            || die $DBI::errstr;
    };

    if ($@) {
        $Log->comment("SQL query failed with: $@");
    }

    my @results;
    foreach my $row (@{$all}) {
        my ($node_name, $ontap_release, $node_model) = @$row;
        my %d = (
            'node_name'     => $node_name,
            'ontap_release' => $ontap_release,
            'node_model'    => $node_model,
            'table'         => $table,
            'database'      => $database,
        );
        push(@results, \%d);
    }

    $Log->debug("Node Data: " . Dumper(\@results));

    my $elapsed = tv_interval($t0, [gettimeofday]);
    $Log->comment("_fetch_node_data completed in $elapsed seconds");

    return \@results;
} ## end sub _fetch_node_data

# Shrink long legend strings to avoid legend overrun
sub shrink_legend {
    my ($self, $legend, $max_length) = @_;

    $Log->debug(Dumper($legend));

    for (my $i = 0; $i < @{$legend}; $i++) {
        my $key            = ${$legend}[$i];
        my $length         = length($key);
        my $adj_max_length = $max_length - 4;
        $Log->debug("Legend key: $key Length: $length");
        if ($length > $max_length) {
            my $end   = int($adj_max_length / 2);
            my $start = $length - int($adj_max_length / 2);
            ${$legend}[$i] = substr($key, 0, $end);
            ${$legend}[$i] .= '...';
            ${$legend}[$i] .= substr($key, $start, $length);
            $Log->debug("Modified legend key: ${$legend}[$i]");
        }
    }
} ## end sub shrink_legend

# This routine normalizes the data that is read from the Databse.
sub normalize_numbers {
    my ($self, $num_in) = @_;
    my @num_out  = ();
    my @curr_row = ();
    my ($n, $i);
    my $d = 0;    # debug

    for my $row_cnt (0 .. (@{$num_in} - 1)) {

        for my $col_cnt (0 .. (@{$num_in->[$row_cnt]} - 1)) {

            $n = $num_in->[$row_cnt][$col_cnt];
            if (!defined($n)) { next; }

            # strip % sign
            if ($n =~ /(\d+)%/) {
                $i = $1;
                push(@curr_row, $i);
            } elsif ($n =~ /([\d\.]+)B/) {

                # BYTES
                $i = $1;
                push(@curr_row, $i);
            } elsif ($n =~ /([\d\.]+)KB?/) {

                # KILOBYTES
                $i = $1;
                push(@curr_row, $i * 1024);
            } elsif ($n =~ /([\d\.]+)MB?/) {

                # MEGABYTES
                $i = $1;
                push(@curr_row, $i * (1024 * 1024));
            } elsif ($n =~ /^(\d+:([0-5][0-9])$)/) {

                # TIME - assume minutes:seconds
                my @time = split ":", $1;
                push(@curr_row, ($time[0] * 60) + $time[1]);
            } else {

                # All else
                push(@curr_row, $n);
            }
        }
        $num_out[$row_cnt] = [@curr_row];
        @curr_row = ();
    }

    return (\@num_out);
} ## end sub normalize_numbers

# Adds to variables to the STAF var space.
# If key is an array, push value(s) onto it.
# If key is scalar add to the current value
sub add_values {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {   key   => {type => SCALAR,            optional => 0},
            value => {type => SCALAR | ARRAYREF, optional => 0}
        }
    );
    my $staf = staf_global();
    my $key  = uc($opts{key});
    my $var;
    eval {
        $var = $staf->var_get(
            handle   => $self->{staf_handle},
            var      => $key,
            endpoint => $self->{hostname}
        );
    };

    if (!$var) {
        die "Variable $key not found\n";
    }

    if (ref($var)) {
        foreach my $value (@{$opts{value}}) {
            if (!grep { $_ eq $value } @{$var}) {
                push(@{$var}, $value);
                $Log->comment("Adding $value to $key");
            } else {
                $Log->comment("$value exists in $key");
            }
        }
    } elsif ($var =~ /^[\-\.0-9]+$/) {
        if ($opts{value} =~ /[\-\.0-9]+/) {
            $var += $opts{value};
        }
    } else {
        return 0;
    }

    natsort(@{$var});
    $staf->var_set(
        handle   => $self->{staf_handle},
        var      => {$key => $var},
        endpoint => $self->{hostname}
    );

} ## end sub add_values

# List all or specific variables in the STAF variable service
sub list_var {
    my $self = shift;
    my (%opts)
        = validate(@_,
        {key => {type => SCALAR, optional => 1, default => 'all'}});
    my $staf = staf_global();
    my $var;
    if ($opts{key} eq 'all') {
        eval {
            $var = $staf->var_list(
                handle   => $self->{staf_handle},
                endpoint => $self->{hostname}
            );
        };
        return $var;
    } else {
        eval {
            $var = $staf->var_get(
                handle   => $self->{staf_handle},
                var      => $opts{key},
                endpoint => $self->{hostname}
            );
        };
        if (ref($var)) { natsort(@{$var}); }
        return $var;
    }
} ## end sub list_var

# Sets variable values in the STAF variable service
sub set_var {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {   key   => {type => SCALAR,            optional => 0},
            value => {type => SCALAR | ARRAYREF, optional => 0}
        }
    );
    my $staf = staf_global();
    my $key  = uc($opts{key});

    $staf->var_set(
        handle   => $self->{staf_handle},
        var      => {$opts{key} => $opts{value}},
        endpoint => $self->{hostname}
    );

} ## end sub set_var

# Removes values from array variables in the STAF variable service
sub remove_values {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {   key   => {type => SCALAR,   optional => 0},
            value => {type => ARRAYREF, optional => 0}
        }
    );

    my $staf = staf_global();
    my $key  = uc($opts{key});
    my $var;
    eval {
        $var = $staf->var_get(
            handle   => $self->{staf_handle},
            var      => $key,
            endpoint => $self->{hostname}
        );
    };

    if (!$var) {
        die "Variable $key not found\n";
    }

    if (!ref($var)) {
        die "Only array variables can have values removed\n";
    }

    my @values = @{$opts{value}};
    my @tmp;
    foreach my $value (@{$var}) {
        if (grep { $_ eq $value } @values) {
            $Log->comment("Removing $value");
        } else {
            push(@tmp, $value);
        }
    }
    $staf->var_set(
        handle   => $self->{staf_handle},
        var      => {$key => \@tmp},
        endpoint => $self->{hostname}
    );

} ## end sub remove_values

# Returns the epoch and iteration number
sub status {
    my $self = shift;
    my $staf = staf_global();

    my ($lmr_fast_status, $lmr_slow_status, $lmr_mhost_status);
    eval {
        $lmr_slow_status = $staf->var_get(
            handle   => $self->{staf_handle},
            var      => 'LMR_STATUS',
            endpoint => $self->{hostname}
        );
    };

    eval {
        $lmr_fast_status = $staf->var_get(
            handle   => $self->{staf_handle},
            var      => 'LMR_STATUS_FAST',
            endpoint => $self->{hostname}
        );
    };
    eval {
        $lmr_mhost_status = $staf->var_get(
            handle   => $self->{staf_handle},
            var      => 'LMR_STATUS_MHOST',
            endpoint => $self->{hostname}
        );
    };

    my $lmr_status = {};
    $lmr_status->{'slow'}  = $lmr_slow_status;
    $lmr_status->{'fast'}  = $lmr_fast_status;
    $lmr_status->{'mhost'} = $lmr_mhost_status;

    return $lmr_status;

} ## end sub status

# Returns the result of var_get of options and handle of LMR in STAF
sub get_staf_handle {
    my (%opts) = validate(
        @_,
        {   jobid    => {type => SCALAR,         optional => 0},
            hostname => {type => SCALAR | UNDEF, optional => 0}
        }
    );

    if (!defined($opts{hostname})) {
        return;
    }

    $Log->comment("Searching for $opts{'jobid'}\'s STAF handle");
    my $staf = staf_global();
    my @handles = @{$staf->handle_list(endpoint => $opts{hostname})};
    my $handle;
    foreach my $h (@handles) {
        my $lmr_var_service;
        eval {
            $lmr_var_service = $staf->var_get(
                handle   => $h->{'handle'},
                var      => "LMR_VAR_SERVICE",
                endpoint => $opts{hostname}
            );
        };

        if ($lmr_var_service) {
            my $top_logdir;
            eval {
                $top_logdir = $staf->var_get(
                    handle   => $h->{'handle'},
                    var      => "TOP_LOGDIR",
                    endpoint => $opts{hostname}
                );
            };

            if ($top_logdir =~ /$opts{'jobid'}/) {
                $Log->comment("Using STAF handle $h->{'name'}");
                $handle = $h;
                last;
            }

        }
    }
    if (!$handle) {
        die config_result("Unable to find the STAF handle for "
                . "$opts{'jobid'}. Did you specify the right "
                . "jobid?\n");
    } else {
        return $handle->{'handle'};
    }

} ## end sub get_staf_handle

# Converts time using SQL
sub sql_time_conversion {

    my ($time, $modifier) = @_;

    # If a time string is passed in we quote otherwise it could be a column
    # name so we don't
    if ($time =~ /\d+-\d+-\d+ \d+:\d+:\d+/) {
        $time = '\'' . $time . '\'';
    }
    my $sql_time = 'datetime(' . $time . ',\'' . $modifier . '\')';
    return $sql_time;
} ## end sub sql_time_conversion

# Returns the time in seconds from the start of the epoch
sub get_epoch_time {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   date   => {type => SCALAR, optional => 0},
            format => {type => SCALAR, optional => 0},
            mode   => {type => SCALAR, optional => 0}
        }
    );

    my ($time, $month, $day, $year, $hour, $minute, $second);
    if ($opts{'date'} =~ /(\d+)(\/|-)(\d+)(\/|-)(\d+)\s+(\d+):(\d+):(\d+)/) {
        if ($opts{'format'} eq 'm/d/y H:M:S') {
            $month = $1;
            $month--;
            $day    = $3;
            $year   = $5;
            $hour   = $6;
            $minute = $7;
            $second = $8;
        } elsif ($opts{'format'} eq 'Y-m-d H:M:S') {
            $year  = $1;
            $month = $3;
            $month--;
            $day    = $5;
            $hour   = $6;
            $minute = $7;
            $second = $8;
        } ## end

        if ($opts{'mode'} eq 'local') {
            $time = timelocal($second, $minute, $hour, $day, $month, $year);
        } elsif ($opts{'mode'} eq 'gm') {
            $time = timegm($second, $minute, $hour, $day, $month, $year);
        }
    }

    return $time;
} ## end sub get_epoch_time

# Given a source array, returns a subset of evenly spaced array keys
sub get_array_subset {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   source   => {type => ARRAYREF, optional => 0},
            elements => {type => SCALAR,   optional => 0}
        }
    );
    my $source   = $opts{'source'};
    my $elements = $opts{'elements'};
    my @subset;
    my $size  = @{$source};
    my $first = ${$source}[0];
    my $last  = ${$source}[$#{$source}];
    my $msg   = "Source - Size:$size Requested:$elements ";
    push(@subset, $first);

    if ($elements > 2) {

        my $divisor = 1;
        my $count   = int(@{$source} / $divisor);
        while ($count > $elements || $elements == 1) {
            $divisor += 1;
            $count = int(@{$source} / $divisor);
        }
        $msg .= "Divisor:$divisor\n";

        for (my $i = $divisor; $i <= @{$source}; $i += $divisor) {
            if (   @subset < ($elements - 1)
                && $i < $#{$source}
                && defined(${$source}[$i])) {
                push(@subset, ${$source}[$i]);
            }
        }
    }
    push(@subset, $last);
    $msg .= "Subset - Size : " . @subset . " \n " . Dumper(\@subset);
    $Log->debug($msg);
    return \@subset;
} ## end sub get_array_subset

sub parse_counter_name {

    my ($self, $counter_name) = @_;

    # We are purposely forcing object names with an : to have the right
    # hand side prepended to the instance name so we account for these
    # two cases:
    #
    # rtpqa-fas6280-1:lif:rtpqa-fas3070-7:mgmt1:recv_data
    # rtpqa-fas6280-1:lif:vs_io_1:a:b:c:d:e:f:0:1:2:3:recv_data
    # rtpqa-fas6280-1:smb1:vserver:vs_io_1:active_searches
    # rtpqa-fas6280-1:smb1:node:rtpqa-fas3070-7:active_searches
    # rtpqa-fas6280-1:nfsv3:node:rtpqa-fas6280-1:nfsv3_throughput
    #

    $Log->debug("Parsing $counter_name");
    my $node_regex     = '[^:]+';
    my $subsys_regex   = '[^:]+';
    my $instance_regex = '[^:]+:[^:]+:[^:]+:[^:]+:[^:]+:[^:]+:[^:]+:[^:]+|';
    $instance_regex .= '[^:]+:[^:]+|[^:]+';
    my @combo_objects;
    if (defined($self->{'combo_objects'})) {
        @combo_objects = @{$self->{'combo_objects'}};

        $Log->debug("Checking for combination objects");
        foreach my $object (@combo_objects) {
            if ($counter_name =~ /^([^:]+:$object:|^$object:)/) {
                $subsys_regex = $object;
                last;
            }
        }
    } ## end

    my @fields;
    if ($counter_name
        =~ /^([^:]+):($subsys_regex):($instance_regex):([^:]+)$/) {
        $Log->debug(
            "Matched ^([^:]+):($subsys_regex):($instance_regex):([^:]+)\$");
        $fields[0] = $4;
        $fields[1] = $3;
        $fields[2] = $2;
        $fields[3] = $1;
        $fields[2] =~ s/:/_/;
    } elsif ($counter_name =~ /^([^:]+):([^:]+):([^:]+)$/) {
        $Log->debug("Matched ^([^:]+):([^:]+):([^:]+)\$");
        $fields[0] = $3;
        $fields[1] = $2;
        $fields[2] = $1;
    } elsif ($counter_name =~ /^([^:]+):([^:]+)$/) {
        $Log->debug("Matched ^([^:]+):([^:]+)\$");
        $fields[0] = $2;
        $fields[1] = $1;
    }

    return \@fields;
} ## end sub parse_counter_name

sub get_array_index {
    $Log->enter();

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   array => {type => ARRAYREF, optional => 0},
            value => {type => SCALAR,   optional => 0}
        }
    );

    for (my $i = 0; $i < @{$opts{'array'}}; $i++) {
        if (${$opts{'array'}}[$i] eq $opts{'value'}) {
            $Log->exit();
            return $i;
        }
    }
    return undef;
    $Log->exit();
} ## end sub get_array_index

sub fetch_subsystem_tables {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database => {type => SCALAR | UNDEF, optional => 1},
            subsys   => {type => SCALAR | UNDEF, optional => 0}
        },
    );

    my $subsys = $opts{'subsys'};
    my @databases;
    if (!$opts{'database'}) {
        @databases = @{$self->{'databases'}};
    } else {
        @databases = split(/,/, $opts{'database'});
    }

    my @table_names;
    foreach my $database (@databases) {
        my $tables = $self->_fetch_subsystem_tables(
            database => $database,
            subsys   => $subsys
        );
        if ($tables) {
            foreach my $t (@{$tables}) {
                if (!grep { $_ eq $t } @table_names) {
                    push(@table_names, $t);
                }
            }
        }
    }
    return \@table_names;
} ## end sub fetch_subsystem_tables

sub _fetch_subsystem_tables {

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {   database => {type => SCALAR, optional => 0},
            subsys   => {type => SCALAR, optional => 0}
        },
    );

    my @tables = ();
    foreach my $s (split(/,/, $opts{'subsys'})) {
        my $results = $self->fetch_subsystems(
            database => $opts{'database'},
            subsys   => $s
        );
        if (ref($results)) {
            foreach my $hashref (@{$results}) {
                my $t = $hashref->{'table_name'};
                if (!grep { $_ eq $t } @tables) {
                    push(@tables, $hashref->{'table_name'});
                }
            }
        }
    }
    if (!@tables) {
        $Log->debug("Unable to find table name(s) database "
                . "$opts{'database'} for $opts{'subsys'}");
        return;
    }

    return \@tables;
} ## end sub _fetch_subsystem_tables

sub merge_arrays {
    $Log->enter();

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   legend1 => {type => ARRAYREF, optional => 0},
            data1   => {type => ARRAYREF, optional => 0},
            legend2 => {type => ARRAYREF, optional => 0},
            data2   => {type => ARRAYREF, optional => 0}
        }
    );

    my $timestamps = {};

    # Get the timestamps for all counters in the data1 array
    my $data1   = $opts{'data1'};
    my $legend1 = $opts{'legend1'};
    for (my $s = 0; $s < @{$legend1}; $s++) {
        my $counter = $legend1->[$s];
        foreach my $coordinates (@{$data1->[$s]}) {
            push(@{$timestamps->{$counter}}, ${$coordinates}[0]);
        }
    }

    # Get the timestamps for all counters in the data2 array
    my $data2   = $opts{'data2'};
    my $legend2 = $opts{'legend2'};
    for (my $s = 0; $s < @{$legend2}; $s++) {
        my $counter = $legend2->[$s];
        foreach my $coordinates (@{$data2->[$s]}) {
            push(@{$timestamps->{$counter}}, ${$coordinates}[0]);
        }
    }

    # Walk through all the counters and build new legend and data arrays
    my $data         = [];
    my $legend       = [];
    my $legend_index = 0;
    foreach my $counter (natsort(keys(%{$timestamps}))) {
        ${$legend}[$legend_index] = $counter;
        my @sorted_ts = natsort(@{$timestamps->{$counter}});

        #parse the sorted array to remove duplicate timestmaps
        my @unique_ts = uniq(@sorted_ts);
        foreach my $ts (@unique_ts) {
            my $legend1_index = $self->get_array_index(
                'array' => $legend1,
                'value' => $counter
            );
            if (defined($legend1_index)) {
                foreach my $coordinates (@{$data1->[$legend1_index]}) {
                    if ($ts eq $coordinates->[0]) {
                        push(@{$data->[$legend_index]}, $coordinates);
                        last;
                    }
                }
            }
            my $legend2_index = $self->get_array_index(
                'array' => $legend2,
                'value' => $counter
            );
            if (defined($legend2_index)) {
                foreach my $coordinates (@{$data2->[$legend2_index]}) {
                    if ($ts eq $coordinates->[0]) {
                        push(@{$data->[$legend_index]}, $coordinates);
                        last;
                    }
                }
            }
        }
        $legend_index++;
    }

    $Log->exit();
    return ($legend, $data);
} ## end sub merge_arrays

sub wait_for_perf_stats_operation {
    $Log->enter();

    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   count           => {type => SCALAR, optional => 0},
            interval        => {type => SCALAR, optional => 0},
            lmr_mode        => {type => SCALAR, optional => 0},
            operation       => {type => SCALAR, optional => 0},
            perf_stats_mode => {type => SCALAR, optional => 0},
            wait            => {type => SCALAR, optional => 0}
        }
    );

    my $count           = $opts{'count'};
    my $interval        = $opts{'interval'};
    my $lmr_mode        = $opts{'lmr_mode'};
    my $operation       = $opts{'operation'};
    my $perf_stats_mode = $opts{'perf_stats_mode'};
    my $wait            = $opts{'wait'};

    $Log->comment("Waiting for $perf_stats_mode "
            . "$operation count to increase by $count");

    # Get the currrent status
    my $status = $self->status();

    # Get the current operation count
    my $starting_op_count
        = $status->{$lmr_mode}->{$perf_stats_mode}->{'operation'}
        ->{$operation};

    my $current_op_count;
    my $timeout = 1;
    for (my $i = 0; $i < $wait; $i += $interval) {
        $current_op_count
            = $status->{$lmr_mode}->{$perf_stats_mode}->{'operation'}
            ->{$operation};
        if ($current_op_count >= ($starting_op_count + $count)) {
            $timeout = 0;
            last;
        }
        sleep $interval;
        $status = $self->status();
    }

    if ($timeout) {
        die config_result("Timeout waiting for $perf_stats_mode "
                . "$operation to increase by $count");
    }

} ## end sub wait_for_perf_stats_operation

sub sort_counters_by_table {
    $Log->enter();
    $Log->debug(Dumper(\@_));
    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   clients  => {type => ARRAYREF, optional => 0},
            counters => {type => SCALAR,   optional => 0},
            nodes    => {type => ARRAYREF, optional => 0}
        }
    );

    my $clients  = $opts{'clients'};
    my $nodes    = $opts{'nodes'};
    my @counters = split(/,/, $opts{'counters'});

    my $results = $self->fetch_counters();
    my %tables;
    foreach my $result (@{$results}) {
        my $counter_name = $result->{'counter_name'};
        my $table_name   = $result->{'table_name'};
        my $subsys_name  = $result->{'subsys'};
        if (!grep { $_ eq $counter_name } @counters) {
            next;
        }

        if (!grep { $_ eq $counter_name } @{$tables{$table_name}{'counters'}})
        {
            push(@{$tables{$table_name}{'counters'}}, $counter_name);
        }

        if (!grep { $_ eq $subsys_name } @{$tables{$table_name}{'subsys'}}) {
            push(@{$tables{$table_name}{'subsys'}}, $subsys_name);
        }
    }

    foreach my $table (keys(%tables)) {
        if (@{$clients} && $table =~ /client/) {
            $tables{$table}{'hostnames'} = $clients;
        }
        if (@{$nodes} && $table !~ /client/) {
            $tables{$table}{'hostnames'} = $nodes;
        }
    }
    return \%tables;
} ## end sub sort_counters_by_table
1;

