# -*- perl -*-
# $Id: //depot/prod/DOT/dev/test/lib/CConfig/NList.pm#1 $

## @summary Utils::NList 
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

NList - a list 

=head1 SYNOPSIS

  use Utils::NList

=head1 DESCRIPTION

An anstraction of a list of values, contains an iterator

    Examples:

    # creating a list
    my @l=qw(a b c d); 
    my $l=Utils::NList->new(@l); # make a list

    $l->add(e); # add element to the list


    Iteration:

    # Java Style Iteration
    my $i=$l->begin();
    while ($i->hasNext()) { my $e=$i->get(); };

    # perl style Iteration
    my $i=$l->begin();
    while (my $e=$i->next()) { print $e; };

    # stl style Iteration
    for (my $i=$l->begin(); $i->pos != $l->end(); $i->next()) {
        print $i->get." ";
    }

    # wrapping 
    my $i=$l->begin(WRAP=>1);
    for (my $k=0; $k<20; $k++) { print $i->get." "; }

    # increment by 
    my $i=$l->begin(INC=>2);
    while ($i->hasNext()) { $i->next(); }

    # filtered lists
    my $i=$l->begin(FILTER=>sub { 
            # iterate over only the even numbers
            return !($_[0] % 2); 
        }
    );
    while ($i->hasNext()) { print $i->get; };
    
    # filter on get - advance interator and return the next value that meets the filter
    my $i=$l->begin();
    # get the next odd number
    $i->get(FILTER=>sub {$_[0]%2});


=cut

# --------------------------------------------------------
# Class: Util::NList
# --------------------------------------------------------

package Utils::NList;
use strict;
use warnings;
use Params::Validate qw(:all);
use Data::Dumper;
use Carp;

use Utils::NList::NListIter;

# -----------------------------------------------
# Function: new
#   constructor
#
# Usage: 
#   Utils::NList->new(@l); 
#
# Parameters:
#       list      - an array to wrap
#
# Return: 
#   blessed reference
# -----------------------------------------------
sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;

    my $self     = {
        DATA=>[@_],
    };
    bless($self, $class);
}

# -----------------------------------------------
# Function: add
#   add element to the list 
#
# Usage: 
#   add(element)
#
# Parameters:
#       element      - element to add 
#
# Return: 
#   list ref
# -----------------------------------------------
sub add {
    my $self=shift;
    push(@{$self->data},@_);
    return $self;
}

# -----------------------------------------------
# Function:  DATA - list data
# -----------------------------------------------
sub data   { my $self = shift; if(@_) { $self->{DATA} = $_[0]; } return $self->{DATA}; }
# -----------------------------------------------
# Function:  LEN  - length of data array
# -----------------------------------------------
sub len    { my $self = shift; return scalar(@{$self->{DATA}}); }
# -----------------------------------------------
# Function:  LASTI- last index of data in list
# -----------------------------------------------
sub lasti  { my $self = shift; return $#{$self->{DATA}}; }

# -----------------------------------------------
# Function: begin
#   return a list iterator 
#
# Usage: 
#   my $i=$l->begin();
#
# Parameters:
#   INC - increment
#   WRAP - 1=wrap at end
#   FILTER - codref to a function to be used to filter the values in the list
#
# Return: 
#   list iterator 
# -----------------------------------------------
sub begin {
    my $self=shift;
    return Utils::NList::NListIter->new($self,@_);
}

# -----------------------------------------------
# Function: end
#   one past the end of the list pos
# Usage: 
#
# Parameters:
#
# Return: 
#   list iterator 
# -----------------------------------------------
sub end {
    return $_[0]->len;
}

# -----------------------------------------------
# Function: get
#
# Usage: 
#   get([POS=>pos],FILTER=>filter)
#
# Parameters:
#   pos(optional) - list index to return value for, by default
#           will use the current list position
#   filter(optional) - a coderef to a filter function
#           if supplied the iterator will check the current
#           value and if it doesn't pass the filter the 
#           list iterator will be advanced to the next value
#           that meets the filter
#
# Return: 
#   return the current list element
# -----------------------------------------------
sub get {
    return  $_[0]->data->[$_[1]];
}


# -----------------------------------------------
# Function: copy
#   copy a list 
#
# Usage: 
#   copy()
#
# Parameters:
#
# Return: 
#   ref to copy of self
# -----------------------------------------------
sub copy {
    my $self=shift;
    my $copy=__PACKAGE__->new(@{$self->data}); 
    return $copy;
}

1;

=head1 FUNCTIONS



=head2 new

    constructor

=head3 Usage

    Utils::NList->new(@l); 

=head3 Return

    blessed reference

=head2 add

    add element to the list 

=head3 Usage

    add(element)

=head3 Return

    list ref

=head2 DATA - list data


=head2 LEN  - length of data array


=head2 LASTI- last index of data in list


=head2 begin

    return a list iterator 

=head3 Usage

    my $i=$l->begin();

=head3 Return

    list iterator 

=head3 Parameters:


=over 4


=item INC

    increment

=item WRAP

    1=wrap at end

=item FILTER

    codref to a function to be used to filter the values in the list
    

=back


=head2 end

    one past the end of the list pos

=head3 Return

    list iterator 

=head2 get


=head3 Usage

    get([POS=>pos],FILTER=>filter)

=head3 Return

    return the current list element

=head3 Parameters:


=over 4


=item pos(optional)

    list index to return value for, by default
    will use the current list position

=item filter(optional)

    a coderef to a filter function
    if supplied the iterator will check the current
    value and if it doesn't pass the filter the 
    list iterator will be advanced to the next value
    that meets the filter
    

=back


=head2 copy

    copy a list 

=head3 Usage

    copy()

=head3 Return

    ref to copy of self


=head1 SEE ALSO 

