# -*- perl -*-
# $Id: $

## @summary Utils::NList::NListIter
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

NListIter - a list iterator for Nlist.pm

=head1 SYNOPSIS

  use Utils::NList

=head1 DESCRIPTION

An STL like list iterator for with various interfaces
Examples:
    my @l=qw(a b c d); 
    my $l=Utils::NList->new(@l); # make a list
    $l->add(e); # add to the list

    # Java Style Iteration
    my $i=$l->begin();
    while ($i->hasNext()) { my $e=$i->next(); };

    # perl style Iteration
    my $i=$l->begin();
    while (my $e=$i->next()) { print $e; };

    # stl style Iteration
    for (my $i=$l->begin(); $i->pos != $l->end(); $i->next()) {
        print $i->get." ";
    }

    # wrapping 
    my $i=$l->begin(WRAP=>1);
    for (my $k=0; $k<20; $k++) { print $i->next(); }

    # increment by 
    my $i=$l->begin(INC=>2);
    while ($i->hasNext()) { print $i->next(); }

    # filtered lists
    my $i=$l->begin(FILTER=>sub { 
            # iterate over only the even numbers
            return !($_[0] % 2); 
        }
    );
    while ($i->hasNext()) { print $i->next; };
    
    # filter on next - advance the iterator and return the next value that meets the filter 
    # get next odd number from list
    $i->next(FILTER=>sub {$_[0]%2});

    # filter on get - return the next value (starting at the current iterator value) that meets the filter 
    # get the current number from the list if its odd, else advance and return the next odd number
    $i->get(FILTER=>sub {$_[0]%2});
=cut

# --------------------------------------------------------
# Class: Utils::NList::NListIter
# --------------------------------------------------------

package Utils::NList::NListIter;
use strict;
use warnings;
use Params::Validate qw(:all);
use Data::Dumper;
use Carp;

# -----------------------------------------------
# Function: new
#   constructor
#
# Usage: 
#   new(list)
#
# Parameters:
#       list      - reference to an array 
#
# Return: 
#   blessed reference
# -----------------------------------------------
sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;

    my $self     = {
        POS=>0,
        LIST=>shift,
        INC=>1,
        WRAP=>0,
        @_
    };

    bless($self, $class);
}

# -----------------------------------------------
# Function:  data - list data at current iterator pos
# -----------------------------------------------
sub data   { my $self = shift; if(@_) { $self->{LIST}->set($self->{POS},$_[0]); } return $self->{LIST}->get($self->{POS}); }
# -----------------------------------------------
# Function:  list - list data
# -----------------------------------------------
sub list   { my $self = shift; if(@_) { $self->{LIST} = $_[0]; } return $self->{LIST}; }
# -----------------------------------------------
# Function:  pos  - current cursor into the list
# -----------------------------------------------
sub pos    { my $self = shift; if(@_) { $self->{POS}  = $_[0]; } return $self->{POS}; }
# -----------------------------------------------
# Function:  inc  - increment
# -----------------------------------------------
sub inc    { my $self = shift; if(@_) { $self->{INC}  = $_[0]; } return $self->{INC}; }
# -----------------------------------------------
# Function:  wrap  - to wrap or not to wrap
# -----------------------------------------------
sub wrap    { my $self = shift; if(@_) { $self->{WRAP}  = $_[0]; } return $self->{WRAP}; }
# -----------------------------------------------
# Function:  filter  - coderef to a filter to apply to the list
# -----------------------------------------------
sub filter  { my $self = shift; if(@_) { $self->{FILTER}  = $_[0]; } return $self->{FILTER}; }


# -----------------------------------------------
# Function: get
#   return value of current list element.  
#
# Usage: 
#   get()
#
# Parameters:
#   FILTER - (optional) coderef to use to filter the value, if current
#            value does not match the filter expression iterator
#            will be advanced to the next value that passes the filter
# Return: 
#   get current element 
#
# -----------------------------------------------
sub get {
    my $self=shift;
    my %args=(@_);
    my $filter=$args{'FILTER'} || $self->filter;

    if ($filter) {
        unless (&{$filter}($self->data)) { 
            return undef unless $self->advance(%args); 
        }
    }
    return $self->data;
}

# -----------------------------------------------
# Function: next
#   return current list value and advance iterator
#
# Usage: 
#   next()
#
# Parameters:
#   FILTER - (optional) coderef to use to filter the values
#       to be returned
# Return: 
#   return current element and advance iterator
#
# -----------------------------------------------
sub next {
    my $self = shift; 
    my %args=(@_);
    my $d=$self->get(%args);
    $self->advance(%args);
    return $d;
}

# -----------------------------------------------
# Function: advance
#   advance iterator to the next value
#
# Usage: 
#   advance()
#
# Parameters:
#   FILTER - (optional) coderef to use to filter the values
#       to be advanced to.
# Return: 
#   advance iterator, return list value after advance
#   undef if nothing matches the filter
#
# -----------------------------------------------
sub advance {
    my $self = shift; 
    my %args=(@_);

    my $filter=$args{'FILTER'} || $self->filter;

    if ($filter) {
        my $nwraps=0; # don't want to loop fovever if the filter doesn't match anything
        my $inc=$self->inc;
        do {
            $self->pos($self->pos+$inc); 
            $inc=1; # subsequent increments will be by 1 until we find something that matches the filter
            if ($self->wrap && $self->pos >= $self->list->len) { 
                $self->pos($self->pos-$self->list->len); 
                $nwraps++; 
                return undef if ($nwraps>2);
            }
        } until ($self->pos >= $self->list->len || &{$filter}($self->data)); 
    } else {
        $self->pos($self->pos+$self->inc);
        if ($self->wrap && $self->pos >= $self->list->len) { $self->pos(0); }
    }

    return $self->data;
}

# -----------------------------------------------
# Function: hasNext
#   see if there is a next element
#
# Parameters:
#
# Return: 
#   true if there is a next element
#   false otherwise
#
# -----------------------------------------------
sub hasNext {
    my $self = shift; 
    return $self->pos < $self->list->len;
}

# -----------------------------------------------
# Function: reset
#   reset iterator position to beginning
#
# Parameters:
#
# Return: 
#   self
# -----------------------------------------------
sub reset {
    my $self = shift; 
    $self->pos(0);
    return $self;
}

1;

=head1 FUNCTIONS



=head2 new

    constructor

=head3 Usage

    new(list)

=head3 Return

    blessed reference

=head2 data - list data at current iterator pos


=head2 list - list data


=head2 pos  - current cursor into the list


=head2 inc  - increment


=head2 wrap  - to wrap or not to wrap


=head2 filter  - coderef to a filter to apply to the list


=head2 get

    return value of current list element.  

=head3 Usage

    get()

=head3 Return

    get current element 

=head3 Parameters:


=over 4


=item FILTER

    (optional) coderef to use to filter the value, if current
    value does not match the filter expression iterator
    will be advanced to the next value that passes the filter

=back


=head2 next

    advance iterator and return next value 

=head3 Usage

    next()

=head3 Return

    return current element and advance iterator

=head2 hasNext

    see if there is a next element

=head3 Return

    true if there is a next element
    false otherwise

=head2 reset

    reset iterator position to beginning

=head3 Return

    self


=head1 SEE ALSO 


