# this is basically -*- perl -*-
# vim: set tabstop=4 shiftwidth=4 expandtab shiftround:
## @summary Common library for client pool management
## @author dl-qa-cst@netapp.com
## @pod here

=head1 NAME

Utils::PoolManager.pm - Common library for client pool management

=head1 DESCRIPTION

This package helps with managing a pool of clients and adds functionality
on top of ndisp.

=head1 SYNOPSIS

    use Utils::PoolManager;

    my $PoolM = new Utils::PoolManager(host_tag => "FIT", 
                                       ndispd_addr => "10.227.198.39:13086");

    my @hosts = $PoolM->lock_clients(locktype   => "shared",
                                     hosttype   => "win",
                                     count      => 10);

    $PoolM->release_my_clients();

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-qa-cst@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/Utils/PoolManager.pm#1 $
  Created $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=head1 ROUTINES

=cut

package Utils::PoolManager;
use strict;
use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use Params::Common;    # To bring CLIENT and CLIENTFILE Options here
use Utils::HostManager;
use Data::Dumper;
use Options;
use Tharn;

my $Options = new Options();
my $Opts;
our $Log;

$Options->parameters(
    definitions => [
        {
            name        => "NDISPD",
            description => "IP:PORT of ndispd that is running to manage hosts",
            spec        => "s",
            default     => undef,
        },
        {
            name        => "USER",
            description => "Bring USER from Shell to here",
            spec        => "s",
        },
    ]);

=head2 new

    my $PoolM = new Utils::PoolManager(host_tag => 'tag value', ndispd_addr => "172.16.X.X:61000");

C<new> creates PoolManager object with hooks into the ndispc API

=over

=item ndispd_addr

(Required) The IP address and port pair of ndispd daemon. It can be set in NATE param NDISPD as well.

=item host_tag

(Required) Set a tag to filter on a specific pool of clients. It will match on HOST_TAGS in NATE_HOST files.

=item USER

(Optional) To override the username set in the environment hash.

=back

=cut

sub new {
    $Log->enter();
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    
    # Process Options
    $Options->process();
    $Opts = $Options->get_hash(
        option => [
            qw(NDISPD USER)
        ]
    );
    
    # Initialize self
    my $self = {ndispd_addr => $Opts->{NDISPD},
                host_tag => undef,
                @_,
               };
    
    # Verify host_tag and ndispd_addr are set
    if (! defined($self->{'host_tag'})) {
        $Log->error("You must specify [host_tag].");
        return undef;
    }
    if (! defined($self->{'ndispd_addr'})) {
        $Log->error("You must specify [ndispd_addr].");
        return undef;
    }
    
    # Override USER
    if (defined($self->{USER})) {
        if ($Opts->{USER} ne $self->{USER}) {
            $Log->warn("You are running as a different username: ".$self->{USER});
        }
        $Opts->{USER} = $self->{USER};
    }
    
    # Format jobname for ndisp webpage
    $Opts->{JOBNAME} = "qa-PoolManager.".$Opts->{USER}.".".time();
    
    # Create HostManager instance to use as an ndisp API wrapper
    $self->{_HostM} = new Utils::HostManager(ndispd_addr => $self->{ndispd_addr});
    
    # Bless self
    bless($self, $class);
    
    $Log->exit();
    
    # Return self
    return $self;
}

#################################################################
# User-friendly wrapper routines
#################################################################

=head2 release_my_clients

    my @clients = $PoolM->release_my_clients();

C<show_clients> unlocks any clients that you own.

=cut

sub release_my_clients {
    $Log->enter();
    my $self = shift;
    my %opts = @_;
    
    # Set reservation seach options
    my %res_opts;
    $res_opts{ownertype} = 'user';
    $res_opts{want_value} = 'hash';
    if (defined($opts{locktype})) {
        $res_opts{reason} = "PoolManager: ".$opts{locktype};
    }
    if (defined($Opts->{USER})) {
        $res_opts{ownername} = $Opts->{USER};
    }
    
    # Find your locked clients
    my %clients = $self->show_reservations(%res_opts);
    
    # Release clients
    my @ids = keys(%clients);
    my @clients;
    if (scalar(@ids) > 0) {
        @clients = $self->release_clients(clients => \@ids);
    } else {
        return;
    }
    
    # Make sure all reserved clients were released.
    if (scalar(@clients) != scalar(@ids)) {
        my $reserved = join(',', @ids);
        my $released = join(',', @clients);
        $Log->error("You had ".scalar(@ids)." clients reserved, but only ".scalar(@clients)." could be released.\nReserved: ".$reserved."\nReleased: ".$released);
        exit(1);
    }
    
    $Log->exit();
    
    # Return released client array
    if (scalar(@clients) > 0) {
        return @clients;
    } else {
        return;
    }
}

=head2 show_available

    my @clients = $PoolM->show_available(hostname => ['client1', 'client2']);
    
    my @clients = $PoolM->show_available(hosttype => 'win', locktype => 'shared');

C<show_clients> returns the list of clients available for reservation that matched the query fields.

=over

=item hosttype

(Optional) A list of hosttypes to look for (ex: hosttype => 'win', or hosttype => ['win', 'unix']).

=item locktype

(Optional). Show reservations available for this lock type ('shared' or 'exclusive').

=item CONFIGS

(Optional) A list of client CONFIGS to look for (ex: CONFIGS => 'vmware', or CONFIGS => ['vmware', 'win03s-32-sp2']).

=item hostname

(Optional) A list of hostnames to look for (ex: hostname => 'csiqa-vmu070', or hostname => ['csiqa-vmu070', 'csiqa-vmw003']).

=item any other key/value from NATE_HOST

(Optional) A case-sensitive key/value pair to search the NATE_HOST for (ex: default_ip => '10.227.X.X', prefer_family => ['inet4', 'inet6']).

=item want_value

(Optional) "hash" or "array". Default is "array".

=back

=cut

sub show_available {
    $Log->enter();
    my $self = shift;
    my %opts = @_;
    my $locktype;
    
    # Determine locktype
    if (defined($opts{locktype})) {
        $locktype = $opts{locktype};
        delete $opts{locktype};
    }
    
    # Find all clients
    my %clients = $self->show_clients(%opts, want_value => "hash");
    
    # Find all the locked clients
    my @reserved;
    if (defined($locktype) && $locktype =~ /share/i) {
        @reserved = $self->show_reservations(locktype => 'exclusive', want_value => "array");
    } else {
        @reserved = $self->show_reservations(want_value => "array");
    }
    
    # Remove locked clients from the client list
    foreach my $client (@reserved) {
        delete $clients{$client}; 
    }
    
    $Log->exit();
    
    # Return available client array
    return sort(keys(%clients));
}

#################################################################
# External routines that can be used by tests directly
#################################################################

=head2 show_clients

    my @clients = $PoolM->show_clients();
    
    my @clients = $PoolM->show_clients(hosttype => 'unix');

C<show_clients> returns the list of clients that matched the query fields.

=over

=item hosttype

(Optional) A list of hosttypes to look for (ex: hosttype => 'win', or hosttype => ['win', 'unix']).

=item CONFIGS

(Optional) A list of client CONFIGS to look for (ex: CONFIGS => 'vmware', or CONFIGS => ['vmware', 'win03s-32-sp2']).

=item hostname

(Optional) A list of hostnames to look for (ex: hostname => 'csiqa-vmu070', or hostname => ['csiqa-vmu070', 'csiqa-vmw003']).

=item any other key/value from NATE_HOST

(Optional) A case-sensitive key/value pair to search the NATE_HOST for (ex: default_ip => '10.227.X.X', prefer_family => ['inet4', 'inet6']).

=item want_value

(Optional) "hash" or "array". Default is "array".

=back

=cut

sub show_clients {
    $Log->enter();
    my $self = shift;
    my %opts = @_;
    my @clients;
    
    # Determine return type
    my $returntype = "array";
    if (defined($opts{want_value})) {
        $returntype = $opts{want_value};
        delete $opts{want_value};
    }
    
    # Find available clients
    my %hosts = $self->{_HostM}->list_hosts(want_value => "hash");
    
    # Filter clients based upon HOST_TAGS
    $self->_filter_on_tags(clients => \%hosts);
    
    # Filter on key/value search items passed in
    $self->_filter_on_key_value(opts => \%opts, clients => \%hosts);
    
    $Log->exit();
    
    # Return clients as an array or hash
    if ($returntype =~ /hash/i) {
        return %hosts;
    } else {
        return sort(keys(%hosts));
    }
}

=head2 show_reservations

    my @clients = $PoolM->show_reservations();
    
    my @clients = $PoolM->show_reservations(locktype => 'exclusive', ownername => 'user', hostype => 'win');

C<show_reservations> returns the list of reserved clients that match the query fields.

=over

=item locktype

(Optional). Show reservations with this lock type ('shared' or 'exclusive').

=item ownertype

(Optional) A list of reservation ownertypes to look for (ex: ownertype => 'user', or ownertype => ['user', 'job']).

=item ownername

(Optional) A list of reservation ownernames to look for (ex: ownername => 'collinm', or ownername => ['collinm', 'JamsRegression']).

=item reason

(Optional) A reservation reason to seach on (ex: reason => 'FIT testing').

=item any other key/value from show_clients()

(Optional) A case-sensitive key/value pair to narror down the client list (ex: CONFIGS => 'vmware', or hosttype => ['win', 'unix']).

=item want_value

(Optional) "hash" or "array". Default is "array".

=back

=cut

sub show_reservations {
    $Log->enter();
    my $self = shift;
    my %opts = @_;
    
    # Pull out ownername
    my $ownername;
    if (defined($opts{ownername})) {
        $ownername = $opts{ownername};
        delete $opts{ownername};
    }
    
    # Separate client options from reservation options
    my %clientOpts = %opts;
    my $returntype = 'array';
    foreach my $key (keys(%opts)) {
        if ($key eq 'ownertype' || $key eq 'reason' || $key eq 'locktype') {
            delete $clientOpts{$key};
        } elsif ($key eq 'want_value') {
            $returntype = $opts{$key};
            delete $opts{$key};
            delete $clientOpts{$key};
        } else {
            delete $opts{$key};
        }
    }
    
    # List all reservations
    my %hosts = %{$self->{_HostM}->list_reservations()};
    
    # Filter reservations based upon ownername
    foreach my $res (keys(%hosts)) {
        
        if (defined($ownername) && $hosts{$res}->{reason} !~ /$ownername/) {
            delete $hosts{$res};
        }
    }
    
    # Filter clients based upon HOST_TAGS plus any other client parameters
    my @clients = keys(%hosts);
    @clients = $self->show_clients(hostname => \@clients, %clientOpts);
    
    # Tag the filtered clients in the hosts hash
    foreach my $tagged (@clients) {
        $hosts{$tagged}->{filtered} = 1 if (defined($hosts{$tagged}));
    }
    
    # Remove the un-tagged clients
    foreach my $key (keys(%hosts)) {
        delete $hosts{$key} if (! $hosts{$key}->{filtered});
    }
    
    # Remove tags
    foreach my $tagged (@clients) {
        delete $hosts{$tagged}->{filtered} if (defined($hosts{$tagged}));
    }
    
    # Filter on lock_type
    if (defined($opts{locktype})) {
        if (ref($opts{locktype}) eq "ARRAY") {
            # Search on locktype values using OR logic
            foreach my $client (keys(%hosts)) {
                my $match = 0;
                foreach my $item (@{$opts{locktype}}) {
                    if ($hosts{$client}->{reason} =~ /$item/i) {
                        $match = 1;
                    }
                }
                if (! $match) {
                    delete $hosts{$client};
                }
            }
        } else {
            # Search on scalar value
            foreach my $client (keys(%hosts)) {
                if ($hosts{$client}->{reason} !~ /$opts{locktype}/i) {
                    delete $hosts{$client};
                }
            }
        }
        delete $opts{locktype};
    }
    
    # Filter on key/value reservation search items passed in
    $self->_filter_on_key_value(opts => \%opts, clients => \%hosts);
    
    $Log->exit();
    
    # Return clients as an array or hash
    if ($returntype =~ /hash/i) {
        return %hosts;
    } else {
        return sort(keys(%hosts));
    }
}

=head2 lock_clients

    my @hosts = $PoolM->lock_clients(locktype   => "shared",
                                     hosttype   => "win",
                                     count      => 10);

C<lock_clients> reserves and returns the list of clients that are specified
with this request.

=over

=item locktype

(Required). Reserve clients with this lock type ('shared' or 'exclusive').

=item hosttype

(Required). The config type of the hosts to be reserved.  Required only if 'clients' is not specified.

=item count

(Required). The number of hosts to be reserved. Required only if 'clients' is not specified.

=item clients

(Required). ARRAYREF. List of clients to be reserved.  Required only if 'hosttype' is not specified.

=item restime

(Optional). The reservation time to be used for this reservation. Default: localtime().

=item from_host

(Optional). The hostname where this reservation request is made. Default: hostname().

=item timeout

(Optional). The length of time to wait to reserve the requested hosts. Default 60 secs.

=back

=cut

sub lock_clients {
    $Log->enter();
    my $self = shift;
    my %opts = @_;
    my @clients;
    
    # Determine the locktype
    my $locktype = undef;
    if (defined($opts{locktype})) {
        $locktype = $opts{locktype};
    }
    delete $opts{locktype};
    
    
    # Determine if clients are defined
    my %lock_opts;
    if (defined($opts{clients})) {
        # Clients were passed in as a parameter
        if (ref($opts{clients}) eq "ARRAY") {
            $lock_opts{hosts} = $opts{clients};
            $lock_opts{count} = scalar(@{$opts{clients}});
        } elsif (ref($opts{clients}) eq "HASH") {
            my @tmpclients = keys(%{$opts{clients}});
            $lock_opts{hosts} = \@tmpclients;
            $lock_opts{count} = scalar(keys(%{$opts{clients}}));
        } elsif (ref($opts{clients}) eq "") {
            $lock_opts{hosts} = [$opts{clients}];
            $lock_opts{count} = 1;
        } else {
            $Log->error("You must specify 'clients' in the correct format.");
            exit(1);
        }
        
        # Verify these clients are in our pool
        my @tmpclients = $self->show_clients(hostname => $opts{clients});
        if (scalar(@tmpclients) != scalar(@{$opts{clients}})) {
            my %tmp_h;
            my %client_h;
            for (my $i = 0; $i < scalar(@{$opts{clients}}); $i++) {
                $tmp_h{$tmpclients[$i]} = 1;
                $client_h{$opts{clients}->[$i]} = 1;
            }
            foreach my $t_key (keys(%tmp_h)) {
                delete $client_h{$t_key};
            }
            $Log->error("Some of the clients specified do not belong to this client pool: \n\t".join("\n\t", keys(%client_h)));
            exit(1);
        }
        
        # Verify these clients are reservable
        @tmpclients = $self->show_available(hostname => $opts{clients}, locktype => $locktype);
        if (scalar(@tmpclients) != scalar(@{$opts{clients}})) {
            my %tmp_h;
            my %client_h;
            for (my $i = 0; $i < scalar(@{$opts{clients}}); $i++) {
                $tmp_h{$tmpclients[$i]} = 1;
                $client_h{$opts{clients}->[$i]} = 1;
            }
            foreach my $t_key (keys(%tmp_h)) {
                delete $client_h{$t_key};
            }
            $Log->error("Some of the clients specified are already reserved: \n\t".join("\n\t", keys(%client_h)));
            exit(1);
        }
    } elsif (defined($opts{hosttype}) && defined($opts{count}) && defined($locktype)) {
        # Verify there are sufficient matching clients to reserve
        my @tmpclients = $self->show_available(hosttype => $opts{hosttype}, locktype => $locktype);
        if (scalar(@tmpclients) < $opts{count}) {
            $Log->error("You specified ".$opts{count}." clients, but only ".scalar(@tmpclients)." are available.");
            exit(1);
        } else {
            @tmpclients = splice(@tmpclients, 0, $opts{count});
            $lock_opts{hosts} = \@tmpclients;
            $lock_opts{count} = scalar(@tmpclients);
        }
    } else {
        $Log->error("You must specify either 'clients' or 'hosttype' and 'count'.");
        exit(1);
    }
    delete $opts{clients};
    delete $opts{hosttype};
    delete $opts{count};
    
    # Set the reservation REQID
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    my $now = sprintf("%4d-%02d-%02d %02d:%02d:%02d", $year+1900, $mon+1, $mday, $hour, $min, $sec);
    if (defined($locktype) && $locktype =~ /share/i) {
        $lock_opts{reqid} = 'PoolManager: shared -- '.$now.' -- username:'.$Opts->{USER};
    } else {
        $lock_opts{reqid} = 'PoolManager: exclusive -- '.$now.' -- username:'.$Opts->{USER};
    }
    
    # Set remaining optional parameters
    foreach my $optional (keys(%opts)) {
        $lock_opts{$optional} = $opts{$optional};
    }
    
    # Set persistent lock
    $lock_opts{persistent} = 1;
    
    # Set reason
    $lock_opts{reason} = "Locked by PoolManager.pm";
    
    # Set jobname
    $lock_opts{user} = $Opts->{JOBNAME};
    
    # Handle 'shared' locks for already locked clients
    if (defined($locktype) && $locktype =~ /share/i) {
        my %unlock;
        my $collisions = 0;
        my @new_hosts_array;
        
        # Find existing 'shared' locks
        my %shared = $self->show_reservations(locktype => 'shared', want_value => "hash");
        
        # Determine if we are trying to clients with existing 'shared' locks
        foreach my $hosts (@{$lock_opts{hosts}}) {
            if (defined($shared{$hosts})) {
                # This client is already locked
                $collisions++;
                my @split_reason = split(",", $shared{$hosts}->{reason});
                my $curr_user;
                # Find the current owner(s)
                foreach my $item (@split_reason) {
                    if ($item =~ /REQID=(.+) username:(.+)/) {
                        $unlock{$hosts}->{curr_user} = $2;
                    }
                }
            } else {
                # This client is not already locked, so proceed
                # with normal locking
                push (@new_hosts_array, $hosts);
            }
        }
        $lock_opts{hosts} = \@new_hosts_array;
        
        # If there were existing 'shared' locks on any of the clients
        if ($collisions > 0) {
            # Unlock currently locked 'shared' clients
            my @tmp_unlock = keys(%unlock);
            $self->release_clients(clients => \@tmp_unlock, relock_override => 1);
            
            # Now relock them with all usernames
            my %tmp_lock_opts = %lock_opts;
            $tmp_lock_opts{count} = 1;
            foreach my $to_lock (keys(%unlock)) {
                $tmp_lock_opts{hosts} = [$to_lock];
                $tmp_lock_opts{reqid} = 'PoolManager: shared -- '.$now.' -- username:'.$unlock{$to_lock}->{curr_user}.'&'.$Opts->{USER};
                push (@clients, $self->{_HostM}->reserve_hosts(%tmp_lock_opts));
            }
        }
    }
    
    # Request the locks
    push (@clients, $self->{_HostM}->reserve_hosts(%lock_opts)) if (defined($lock_opts{hosts}->[0]));
    
    # Eliminate duplicates in return array
    my %uniq;
    foreach my $c (@clients) {
        $uniq{$c} = 1;
    }
    @clients = sort(keys(%uniq));
    
    $Log->exit();
    
    # Return locked clients as an array
    if (scalar(@clients) > 0) {
        return @clients;
    } else {
        return undef;
    }
}

=head2 release_clients

    my @clients = $PoolM->release_clients(clients => ['client1', 'client2', client3']);

C<release_clients> releases the list of clients, as long as you are the owner.

=over

=item clients

(Optional) To selectively release the clients.

=back

=cut

sub release_clients {
    $Log->enter();
    my $self = shift;
    my %opts = @_;
    my @clients;
    
    # Determine if clients are defined
    my %release_opts;
    if (defined($opts{clients})) {
        if (ref($opts{clients}) eq "ARRAY") {
            $release_opts{hosts} = $opts{clients};
        } elsif (ref($opts{clients}) eq "HASH") {
            my @tmpclients = keys(%{$opts{clients}});
            $release_opts{hosts} = \@tmpclients;
        } elsif (ref($opts{clients}) eq "") {
            $release_opts{hosts} = [$opts{clients}];
        } else {
            $Log->error("You must specify 'clients' in the correct format.");
            exit(1);
        }
    } else {
        $Log->error("You must specify a list of clients");
        exit(1);
    }
    delete $opts{clients};
    
    # Make sure users can only delete their own reservations
    my $ownername = $Opts->{USER};
    delete $opts{user};
    
    # Set reservation seach options
    my %res_opts;
    $res_opts{ownertype} = 'user';
    $res_opts{want_value} = 'hash';
    if (defined($opts{locktype})) {
        $res_opts{reason} = "PoolManager: ";
    }
    # Find locked clients
    my %clients = $self->show_reservations(%res_opts);
    
    # Get time
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    my $now = sprintf("%4d-%02d-%02d %02d:%02d:%02d", $year+1900, $mon+1, $mday, $hour, $min, $sec);
    
    # Specially handle releasing shared locks so that we can
    # relock them with new owners -- Make sure all clients 
    # are reserved in the user's name, and handle shared 
    # clients differently.  Allow a relock override to skip 
    # relocking shared clients (used by lock_clients()).
    if (! defined($opts{relock_override}) || ! $opts{relock_override}) {
        foreach my $res (keys(%clients)) {
            my @reasons = split(',', $clients{$res}->{reason});
            foreach my $reason (@reasons) {
                # Find reservation owner(s)
                if ($reason =~ /REQID=(.+) username:(.+)/) {
                    my @users = split('&', $2);
                    my @new_users;
                    my $match = 0;
                    # Check if the owner matches our username
                    foreach my $u (@users) {
                        if ($u =~ /$ownername/) {
                            $match++;
                        } else {
                            push(@new_users, $u);
                        }
                    }
                    
                    if ($match == 0) {
                        # Remove clients from the list if we don't own them
                        delete $clients{$res};
                    } elsif (scalar(@new_users) > 0) {
                        # Check if there are multiple owners.
                        # If there are, then relock them with 
                        # the remaining users
                        delete $clients{$res};
                        
                        # Release the client
                        $self->{_HostM}->release_hosts(hosts => [$res]);
                        
                        # Relock the client with the remaining owners
                        my $user_str = join('&', @new_users);
                        my %tmp_lock_opts;
                        $tmp_lock_opts{count} = 1;
                        $tmp_lock_opts{hosts} = [$res];
                        $tmp_lock_opts{reqid} = 'PoolManager: shared -- '.$now.' -- username:'.$user_str;
                        $tmp_lock_opts{persistent} = 1;
                        $tmp_lock_opts{user} = $Opts->{JOBNAME};
                        push (@clients, $self->{_HostM}->reserve_hosts(%tmp_lock_opts));
                    }
                }
            } # End foreach @reasons
        } # End foreach %clients
    } else {
        delete $opts{relock_override};
    }
    
    # Remove clients that we don't have a reservation on
    my @new_hosts;
    foreach my $c (@{$release_opts{hosts}}) {
        push (@new_hosts, $c) if (defined($clients{$c}));
    }
    $release_opts{hosts} = \@new_hosts;
    
    # Set remaining optional parameters
    foreach my $optional (keys(%opts)) {
        $release_opts{$optional} = $opts{$optional};
    }
    
    # Release locks
    if (defined($release_opts{hosts}->[0])) {
        $self->{_HostM}->release_hosts(%release_opts);
    }
    
    # Merge released clients with specially handled 'shared' releases
    push (@clients, @new_hosts);
    
    $Log->exit();
    
    # Return released clients as an array
    if (defined($release_opts{hosts})) {
        return sort(@clients);
    } else {
        return;
    }
}

#################################################################
# Internal routines to be used by this library only
#################################################################

############################################################################
# Subroutine: _filter_on_tags()
# Parameters: clients (hash ref)
#
# Returns: None
# Purpose: To filter out any clients that don't have the correct HOST_TAGS
############################################################################

sub _filter_on_tags {
    my ($self, %opts) =
      validate_mixed(@_, [{type => OBJECT},], {clients => {type => HASHREF,},},);
    
    my @c = keys(%{$opts{clients}});
    foreach my $client (keys(%{$opts{clients}})) {
        if ($opts{clients}->{$client}->{'HOST_TAGS'} !~ /$self->{'host_tag'}/) {
            $opts{clients}->{$client} = undef;
            delete $opts{clients}->{$client};
        }
    }
}

############################################################################
# Subroutine: _filter_on_key_value()
# Parameters: clients (hash ref)
#
# Returns: None
# Purpose: To filter out any clients that don't match the key/value pairs
############################################################################

sub _filter_on_key_value {
    my $self = shift;
    my %params = @_;
    my %opts = %{$params{opts}};
    
    # Filter on key/value reservation search items passed in
    foreach my $key (keys(%opts)) {
        # Make sure a value is defined
        if (! defined($opts{$key})) {
            next;
        }
        if (ref($opts{$key}) eq "ARRAY") {
            # Search on array values using OR logic
            foreach my $client (keys(%{$params{clients}})) {
                my $match = 0;
                foreach my $item (@{$opts{$key}}) {
                    if ($params{clients}->{$client}->{$key} =~ /$item/i) {
                        $match = 1;
                    }
                }
                if (! $match) {
                    delete $params{clients}->{$client};
                }
            }
        } else {
            # Search on scalar value
            foreach my $client (keys(%{$params{clients}})) {
                if ($params{clients}->{$client}->{$key} !~ /$opts{$key}/i) {
                    delete $params{clients}->{$client};
                }
            }
        }
    }
}

################## End #########################################

1;
