# $Id $
## @summary Common library for SSQA to keep util functions
## @author gvg@netapp.com
## @pod here

=head1 NAME

Utils::SSQA.pm - Common library for SSQA for Util functions

=head1 DESCRIPTION

This package defines routines which can be used by SSQA tests.

=head1 SYNOPSIS

use Utils::SSQA qw(testusers);

=head1 ROUTINES

=cut

package Utils::SSQA;
use strict;
use POSIX qw(ceil);
use Tharn;
$Log->comment("Starting " . __FILE__);
use Params::Validate qw(validate UNDEF SCALAR OBJECT HASHREF ARRAYREF);
use Module::Common qw(validate_mixed);
use Utils::FilerStats qw(getLockStats);
use Options;
use Data::Dumper;
use nfs::GenericExport qw(export_path);
use GenericMount qw( mount umount );
use filer qw(filer_vfiler_command);
use POSIX qw(strftime);
use Utils::HostManager;
use Utils::FilerStats;
use File::Basename;
use Time::Local;
use Sort::Key::Natural qw(natsort);

our $VERSION = 1.00;
our @ISA     = qw (Exporter);
our @EXPORT  = qw (
  testusers
  WaitForLockStats
  check_term_file
  write_term_string
  mount_vol
  failed_results
  write_dump
  reserve_clients
  retrieve_clients
  store_and_free_clients
  get_node_state
  get_delay_value
  wait_for_stats
  process_result
  check_subtest_status
  parse_spec
  convert_to_gb
  check_required_space
  check_capacity
  get_max_clients
  strip_vars
  mem_to_csv
  get_epoch_time
  get_array_index
  escape_string
  sort_hostnames
  );

my $Options = new Options;
$Options->parameters(
    definitions => [
        {name        => 'ALT_LOGDIR',
         spec        => "s",
         default     => undef,
         description => 'Write a copy of the csv file to this dir if defined',},
        {name    => 'CAPACITY',
         spec    => "s",
         default => '99',
         description =>
           'The capacity threshold where we fail for lack of diskspace',},
        {name         => 'CLEANUP',
         spec         => "s",
         default      => 1,
         description  => 'Toggles testbed cleanup',
         valid_values => [0, 1],},
        {name    => 'CLEANUP_CLIENT',
         spec    => "s",
         default => '0',
         description =>
           'Controls client clean up when the script is invoked in parallel (RemoteExecute)',
        },
        {name        => 'CLEANUP_TASKS',
         spec        => "s@",
         default     => 'shares,client_data,client_processes',
         description => 'Controls what gets cleaned up durng the cleanup phase',
        },
        {name    => 'CLIENT_ID',
         spec    => "s",
         default => undef,
         description =>
           'Determines the client hostid when the script is invoked in parallel (RemoteExecute)',
        },
        {name => 'CLIENT_LOG',
         spec => "s",
         description =>
           'Determines the CSV file where we trace client statistics',},
        {name        => 'CLIENT_STARTUP_DELAY',
         spec        => "s@",
         default     => '130',
         description => 'How long to wait between configuring clients',},
        {name         => 'CLIENT_STARTUP_MODE',
         spec         => "s",
         default      => 'sequential',
         valid_values => ['', 'sequential', 'parallel'],
         description =>
           'Controls how clients are started, sequentailly or in parallel',},
        {name    => 'CONFIGURE_CLIENT',
         spec    => "s",
         default => '0',
         description =>
           'Controls client initialization when the script is invoked in parallel (RemoteExecute)',
        },
        {name        => 'CSV_DIR',
         spec        => "s",
         default     => undef,
         description => 'The directory in TOP_LOGDIR where we write csv files',
        },
        {name    => "DATA_FILE",
         spec    => "s",
         default => 'data.pl',
         description =>
           'The name of the data file to pass info between different scripts',},
        {name        => 'TIMESTAMP',
         spec        => "s",
         description => 'Using when passing timestamps to subtests',},
        {name        => 'FREE_CLIENTS',
         spec        => "s",
         default     => 0,
         description => 'Free client reservations',},
        {name        => 'FREE_UNUSED_CLIENTS',
         spec        => "s",
         default     => 0,
         description => 'Free reservations for clients that are not needed',},
        {name        => 'GB_REQ_PER_CLIENT',
         spec        => "s",
         default     => '10',
         description => 'Define the number of GBs per client required',},
        {name        => 'LAB',
         spec        => "s",
         description => 'Import LAB from NATE params',},
        {name        => 'LOGDIR_WINDOWS',
         spec        => "s",
         description => 'Import LOGDIR_WINDOWS from NATE params',},
        {name        => 'MAX_CLIENTS',
         spec        => "s",
         description => 'How logn HostManager.pm waits while reserving clients',
        },
        {name        => 'MAX_CLIENT_LOSS',
         spec        => "s",
         default     => 25,
         description => 'The max percentage of processes that can '
           . 'become inactive before test restart',},
        {name        => 'MAX_RESTARTS',
         spec        => "s",
         default     => 25,
         description => 'The number of times the script will '
           . 'restart testing before failing',},
        {name    => 'MAX_SUBTESTS',
         spec    => "s",
         default => 10,
         description =>
           'The maximum number of subtests that can be active at any one time when SUBTEST_TYPE=local',
        },
        {name        => 'RESERVATION_TIMEOUT',
         spec        => "s",
         default     => 60,
         description => 'How logn HostManager.pm waits while reserving clients',
        },
        {name        => 'POST_FAILURE_CLEANUP',
         spec        => "s",
         default     => 1,
         description => 'Cleanup on failures',},
        {name         => 'PROVISION',
         spec         => "s",
         description  => 'Toggles limtis provising',
         default      => 1,
         valid_values => [0, 1],},
        {name => 'RESERVATION_REASON',
         spec => "s",
         description =>
           'The string to be added to the reason when reserving clients via NDISP',
        },
        {name    => 'REMOTE_SUBTEST_DELAY',
         spec    => "s",
         default => 120,
         description =>
           'How long to wait before configuing the nex MAX_SUBTESTS subtests',},
        {name        => 'SCRIPT_INDEX',
         spec        => "s",
         default     => 0,
         description => 'Index to append to file where used clients are stored',
        },
        {name        => 'SCRIPT_PHASE',
         spec        => "s",
         default     => undef,
         description => 'Pass the script phase to subtests',},
        {name         => 'SETUP_DONE',
         spec         => "s",
         default      => 0,
         description  => 'Toggles testbed initialization',
         valid_values => [0, 1],},
        {name        => 'SHARE',
         spec        => "s",
         default     => 'test',
         description => 'The basename of all shares created by this test',},
        {name         => 'SUBTEST_TYPE',
         spec         => "s",
         default      => "local",
         description  => 'Determine where we subtest to',
         valid_values => ['local', 'remote'],},
        {name        => 'TERMFILE',
         spec        => "s",
         default     => "TermFile.txt",
         description => 'File to write TERMSTRINGS to',},
        {name        => 'TERMSTRING',
         spec        => "s",
         default     => undef,
         description => 'string to write in TERMFILE when complete',},
        {name    => 'TERMSTRINGS',
         spec    => "s@",
         default => "TeRmStRiNg",
         description =>
           'A comma seperated list of strings to look for in TERMFILE',},
        {name        => 'TESTBED',
         spec        => "s",
         description => 'Import TESTBED from NATE params',},
        {name        => 'TOP_LOGDIR',
         spec        => "s",
         description => 'Import TOP_LOGDIR from NATE params',},
        {name        => 'TOP_LOGDIR_WINDOWS',
         spec        => "s",
         description => 'Import TOP_LOGDIR_WINDOWS from NATE params',},
        {name        => 'USE_STAF',
         spec        => "s",
         default     => 1,
         description => 'Use staf for client access',},
        {name        => 'VOLUME',
         spec        => "s@",
         default     => 'vol1',
         description => 'The volume on the filer to use when creating shares',},
        {name         => 'VOLUME_USAGE',
         spec         => "s",
         default      => 'single',
         valid_values => ['', 'single', 'multiple'],
         description  => 'Specifies the use of single or multiple volumes',},
        {name        => 'VOLUME_INDEX',
         spec        => "s",
         default     => 0,
         description => 'Specifies the index in the VOLUME array',},
        {name        => 'WAIT_FOR_STATS_DELAY',
         spec        => "s@",
         default     => '60',
         description => 'Add delay so actual stats catchup to whats configured',
        },
        {name    => 'WAIT_FOR_STATS_MAX_FAILURES',
         spec    => "s",
         default => 9,
         description =>
           'The max number of times a stat can not be increasing during wait_for_stats',
        },
        {name    => 'WAIT_FOR_STATS_MAX_QUERIES',
         spec    => "s",
         default => 30,
         description =>
           'The max number of times stats will be queried during wait_for_stats',
        },
        {name    => 'SKIP_POLLED_NODES',
         spec    => "s@",
         default => undef,
         description =>
           'Polling and archiver stats will not be collected for this param',},

      ]);

#--------------------------------------------------------------------

=head2 testusers

 use Utils::SSQA qw(testusers);
 my $users = testusers(patt => "ssqau");

C<testusers> returns arrayref of test users list matching pattern.

=over

=item patt 

(Optional) The pattern string to be matched against users. By default, it
is "ssqau".

=back

=cut

sub testusers {
    my (%params) =
      validate(@_, {patt => {type => SCALAR, default => "^ssqau"},});

    my @users;
    my %groups;

    my @lists = grep { /$params{patt}/ } `ypcat passwd`;
    foreach my $list (@lists) {
        chomp $list;
        my ($name, undef, $uid, $gid, $desc, $home, $shell) = split(":", $list);
        if (!$groups{$gid}) {
            my $gname = getgrgid($gid);
            $groups{$gid} = $gname;
        }
        my $user = {user_name  => $name,
                    group_name => $groups{$gid},
                    uid        => $uid,
                    gid        => $gid,
                    desc       => $desc,
                    home       => $home,
                    shell      => $shell,};
        push @users, $user;
    }
    return \@users;
}

=head2 WaitForLockStats

C<Wait_For_Locks> waits until the number of lock stats specified is reached
on the filer. Returns 1 if condition is met. Returns 0 if timed out. Dies
on failures.

=over

=item conn

(Required) Conn object of the filer required to be passed.

=item stats

(Required). Hashref of the LockStats keys => value pair

=item timeout

(Optional). Default is 1800 secs. (30 mins)

=back

=cut

sub WaitForLockStats {
    my ($conn, %opts) =
      validate_mixed(@_,
                     [{type => OBJECT},],
                     {stats   => {type => HASHREF},
                      timeout => {type => SCALAR, default => 1800},});

    my $endtime = time() + $opts{timeout};

    while ($endtime > time()) {
        my $lockstats = getLockStats($conn, prefix => 0);

        # Make sure the stats given to wait is valid by verifying the returned
        # results from the filer.
        my $matched = 0;
        foreach my $key (keys %{$opts{stats}}) {
            die "Given stat $key is not valid"
              if (!defined($lockstats->{$key}));

            if ($opts{stats}->{$key} == $lockstats->{$key}) {
                $Log->debug("$key matched for $opts{stats}->{$key}");
                $matched++;
                $opts{stats}->{$key} = "_Matched_"
                  ; # Hopefully, this string will never be the one wanted :-, so we will never double count it
            } else {
                $Log->trace(
                    "$key value: $lockstats->{$key}, but expected: $opts{stats}->{$key}"
                    );
            }

            if ($matched == scalar(keys %{$opts{stats}})) {
                return 1;
            }
        }
        sleep 5;
    }

    # Timed out
    return 0;
}

=head2 check_term_file

 Deprecated: Use Utils::TermFile::check_term_file

 use Utils::SSQA qw(check_term_file);
 check_term_file(term_file => "$TOP_LOGDIR\$TERMFILE", term_strings => \$TERMSTRINGS);

C<check_term_file> Checks the term_file for the presence all term_strings

=over

=item term_file

(Required) The termination file

=item term_strings

(Required) Array of strings to check for in the term_file

=back

=cut

sub check_term_file {

    $Log->warn(
        "check_term_file is deprecated please use Utils::TermFile::check_term_file"
        );

    $Log->trace("Entering check_term_file at " . localtime);

    my (%opts) =
      validate(@_,
               {term_file    => {type => SCALAR,   optional => 0},
                term_strings => {type => ARRAYREF, optional => 0},});

    $Log->comment("Checking termination file $opts{'term_file'}");

    my @termFileContents = ();
    open(TERM, "$opts{'term_file'}");
    while (<TERM>) {

        $Log->warn(
            "check_term_file is deprecated please use Utils::TermFile::check_term_file"
            );

        chomp;
        push(@termFileContents, $_);
    }
    close TERM;

    my $allStringsFound = 1;
    foreach my $s (@{$opts{'term_strings'}}) {
        if (!grep(/$s/, @termFileContents)) {
            $allStringsFound = 0;
            last;
        }
    }

    if (!$allStringsFound) {
        $Log->comment(  "Not all TERMSTRINGS exist in termination "
                      . "file, continuing\n@termFileContents");
    }

    $Log->trace("Exiting check_term_file at " . localtime);

    if ($allStringsFound) {
        die "All TERMSTRINGS found in "
          . "termination file, exiting\n@termFileContents";
    }

}

=head2 write_term_string

 Deprecated: Use Utils::TermFile::write_term_string;

 use Utils::SSQA qw(write_term_string);
 write_term_string(term_file => "$TOP_LOGDIR\$TERMFILE", term_string => \$TERMSTRING);

C<write_term_string> Write the term string to the term file

=over

=item term_file

(Required) The termination file

=item term_string

(Required) the string to write to the term file

=back

=cut

sub write_term_string {

    $Log->warn(
        "write_term_string is deprecated please use Utils::TermFile::write_term_string"
        );
    $Log->trace("Entering write_term_string at " . localtime);

    my (%opts) =
      validate(@_,
               {term_file   => {type => SCALAR, optional => 0},
                term_string => {type => SCALAR, optional => 0},});

    $Log->comment(  "Writing termination string $opts{'term_string'} "
                  . "to termination file $opts{'term_file'}");
    my $status = 0;
    for (my $i = 1 ; $i <= 5 ; $i++) {
        if (open(TERM, ">>$opts{'term_file'}")) {
            print TERM "$opts{'term_string'}\n";
            close TERM;
            $status = 1;
            last;
        } else {
            sleep 10;
        }
    }

    if (!$status) {
        die "Error, Unable to write $opts{'term_string'} to $opts{'term_file'}";
    }

    $Log->trace("Exiting write_term_string at " . localtime);
}

=head2 mount_vol

 use Utils::SSQA qw(mount_vol);
 mount_vol(conn => $filerConnObj, vol => 'vol1');

C<mount_vol> Uses NFS to mount filer volumes

=over

=item conn

(Required) The filer connection object 

=item vol

(Required) The volume to mount

=back

=cut

sub mount_vol {

    $Log->enter();

    my (%opts) =
      validate(@_,
               {conn   => {type => OBJECT, optional => 1},
                path   => {type => SCALAR, optional => 0},
                server => {type => SCALAR, optional => 1},
                client => {type     => SCALAR,
                           optional => 1,
                           default  => 'localhost;prefer_family=inet4'},});

    if ($opts{'conn'}) {

        # Convert command if simulating a vfiler context
        my $cmd = filer_vfiler_command($opts{'conn'},
                                       "exportfs -i -o anon=0 $opts{'path'}");
        $Log->debug("Executing $cmd");
        for (my $z = 0 ; $z < 100 ; $z++) {
            my $output;
            eval {
                (undef, undef, $output) =
                  say($opts{'conn'}, $cmd, -timeout, 300);
            };
            if ($output =~ m/(vfiler cannot be run)/) {
                $Log->debug("Vfiler error sleeping for 5 seconds before retry");
                sleep 5;
                next;
            } else {
                $Log->debug("Exiting loop");
                sleep 5;
                last;
            }
        }
        if (!$opts{'server'}) {
            $opts{'server'} = $opts{'conn'}->hostp()->hostname();
        }
    }

    my ($mountPath, %args);

    $Log->debug("Connecting to $opts{'client'}");
    host($opts{'client'});
    $args{'client'} = connect('#rsh', '-nologin', -timeout, 300);

    $Log->debug("Setting mount params");
    if ($opts{'path'} eq '/') {
        $mountPath          = "/t/$opts{'server'}/slash";
        $args{'mountpoint'} = $mountPath;
        $args{'sharing'}    = 'exclusive';
    } else {
        $mountPath = "/t/$opts{'server'}$opts{'path'}/shared/tcp_v3";
    }

    $Log->debug("Checking mount path $mountPath");
    if (-d $mountPath) {
        $Log->debug("Unmounting $mountPath");
        eval { umount($mountPath); };
        sleep 5;
    }

    $Log->debug("Executing mount - server:$opts{'server'} path:$opts{'path'} "
                . Dumper(\%args));
    my $mount = mount($opts{'server'}, $opts{'path'}, %args);
    sleep 5;
    $Log->exit();
    return $mount;

}

=head2 failed_results

 use Utils::SSQA qw(failed_results);
 failed_results(obj => $subtest, regex => "20070123");

C<failed_results> Search subtest objects for failed results

=over

=item obj

(Required) The the subtest object

=item regex

(Required) The runid regex

=back

=cut

sub failed_results {

    $Log->trace("Entering failed_results at " . localtime);

    my (%opts) =
      validate(@_,
               {obj   => {type => OBJECT, optional => 0},
                regex => {type => SCALAR, optional => 0},});

    my @subtests = $opts{'obj'}->Subtest::get_all();
    foreach my $s (@subtests) {
        if ($s->runid() =~ /$opts{'regex'}/) {
            my @results = $s->own_results();
            if (@results) {
                my $worstResult = Rslt::worst(@results)->type();
                if ($worstResult !~ /(PASS|WARN|INFO)/) { return 1; }
            }
        }
    }
    $Log->trace("Exiting failed_results at " . localtime);
    return 0;
}

=head2 check_subtest_status

 use Utils::SSQA qw(check_subtest_status);
 check_subtest_status(obj => $subtest, regex => "20070123");

C<check_subtest_status> Search subtest objects for failed results

=over

=item obj

(Required) The the subtest object

=item regex

(Required) The runid regex

=back

=cut

sub check_subtest_status {

    $Log->trace("Entering check_subtest_status at " . localtime);

    my (%opts) =
      validate(@_,
               {obj   => {type => OBJECT, optional => 0},
                regex => {type => SCALAR, optional => 0},});

    my $passed   = 0;
    my $failed   = 0;
    my $running  = 0;
    my $complete = 0;

    #-----------------------------------------------------
    # Parse all the subtest objects for the latest runids
    #-----------------------------------------------------
    my @subtests = $opts{'obj'}->Subtest::get_all();
    if (@subtests) {
        foreach my $s (@subtests) {
            if ($s->runid() =~ /$opts{'regex'}/) {

                #---------------------------------------------------------
                # If is_alive returns true then the subtest is considered
                # to still be running. Otherwise we figure out the worst
                # result
                #---------------------------------------------------------
                $Log->debug("Found runid " . $s->runid());
                if ($s->is_alive()) {
                    $Log->debug("is_alive returned true " . $s->is_alive());
                    $running++;
                } else {
                    $Log->debug("is_alive returned false " . $s->is_alive());
                    my @results     = $s->own_results();
                    my $worstResult = Rslt::worst(@results)->type();
                    if ($worstResult !~ /(PASS|WARN|INFO)/) {
                        $failed++;
                        $complete++;
                    } else {
                        $complete++;
                        $passed++;
                    }
                }
            }
        }
    }
    $Log->trace("Exiting check_subtest_status at " . localtime);
    return ($passed, $failed, $running, $complete);
}

=head2 write_dump

 Deprecated: Use Utils::Dump::write_dump

 use Utils::SSQA qw(write_dump);
 write_dump(file => $filepath, creator => "cifs_combo", var_name => $name, var_ref => \%hash);

C<write_dump> Write the contents of a perl data structure to a file using Dumper

=over

=item file

(Required) The path the to file

=item creator

(Required) A string denoting who created the file

=item var_name`

(Required) Variabel name

=item var_ref

(Required) Variable reference 

=back

=cut

sub write_dump {

    $Log->warn("write_dump is deprecated please use Utils::Dump::write_dump");
    $Log->enter();

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) =
      validate(@_,
               {creator  => {type => SCALAR,  optional => 0},
                var_name => {type => SCALAR,  optional => 0},
                var_ref  => {type => HASHREF, optional => 0},
                file     => {
                         type     => SCALAR,
                         optional => 0,
                         default  => $options->{'DATA_FILE'}}});

    my $date = strftime("%m%d%y%H%M%S", localtime());

    #$Data::Dumper::Varname = $opts{'var_name'};
    open(DMP, ">$opts{'file'}") || die "unable to open $opts{'file'}: $!";
    print DMP "# Created by $opts{'creator'} on $date\n";
    print DMP Data::Dumper->Dump([$opts{'var_ref'}], [$opts{'var_name'}]);
    close DMP;
    snooze 1;

    $Log->exit();

}

=head2 reserve_clients

 Deprecated: Use Utils::HostManager::Helper::reserve_clients

 use Utils::SSQA qw(reserve_clients);
 reserve_clients(file => $filepath, creator => "cifs_combo", var_name => $name, var_ref => \%hash);

C<reserve_clients> Reserve clients like Utils::HostManager::get_clients 
but beable to specify a global REASON.

=over

=item obj

(Required) The the HostManager object

=back

=cut

sub reserve_clients {

    $Log->warn(
        "reserve_clients is deprecated please use Utils::HostManager::Helper::reserve_clients"
        );
    $Log->trace("Entering reserve_clients at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) = validate(@_, {obj => {type => OBJECT, optional => 0},});

    my $date = strftime("%y%m%d%H%M%S", localtime());
    my (@clients, $user);

    if ($options->{'NDISPD'} && $options->{'CLIENTSPEC'}) {
        $user = `id`;
        chomp($user);
        $user =~ s/uid=\d+\(([^\)]+)\).*/$1/;

        if ($options->{'RESERVATION_REASON'}) {
            $Log->comment(  "Searching for clients reserved by $user "
                          . "with reason \"$options->{'RESERVATION_REASON'}\"");
            @clients =
              $opts{'obj'}->get_reserved_hosts(
                                   'user'   => $user,
                                   'reason' => $options->{'RESERVATION_REASON'},
                                   'strict' => 0);
            if (@clients) {
                $Log->comment("Using " 
                           . @clients
                           . " clients previously reserved by $user with reason"
                           . " \"$options->{'RESERVATION_REASON'}\"");
            }
        }

        if (!@clients) {
            $Log->comment(  "No clients found reserved by $user with a "
                          . "reason \"$options->{'RESERVATION_REASON'}\""
                          . ", reserving new clients");
            my %hostManagerOpts;
            my $id = 0;
            $hostManagerOpts{'persistent'} = 1;
            $hostManagerOpts{'timeout'}    = $options->{'RESERVATION_TIMEOUT'};
            foreach my $spec (split(/,/, $options->{'CLIENTSPEC'})) {
                $hostManagerOpts{'reqid'}  = $date . $$ . $id;
                $hostManagerOpts{'reason'} = $options->{'RESERVATION_REASON'};

                foreach my $optset (split(/;/, $spec)) {
                    my ($param, $value) = split(/=/, $optset);
                    $hostManagerOpts{$param} = $value;

                }

                my @reserved = $opts{'obj'}->reserve_hosts(%hostManagerOpts);
                push(@clients, @reserved) if (scalar(@reserved));
                $id++;
            }

            if (@clients) {
                $Log->comment(  "Using " 
                              . @clients
                              . " clients reserved by $user with reason"
                              . " \"$options->{'RESERVATION_REASON'}\"");
            }
        }
    } elsif (@{$options->{CLIENT}}) {
        @clients = @{$options->{CLIENT}};
    }

    $Log->trace("Exiting reserve_clients at " . localtime);

    if (!@clients) {
        die "Clients could not be found either by"
          . " CLIENT, CLIENTFILE and CLIENTSPEC/NDISPD";
    } else {
        return natsort(@clients);
    }

}

=head2 get_node_state

get_node_state(conn => $fc, node => $node);

C<get_node_state> Determine the state of the node

=over

=item conn

(Optional) The node's connection object

=item node

(Required) The nodes hostid

=back

=cut

sub get_node_state {

    $Log->trace("Entering get_node_state at " . localtime);

    my (%opts) =
      validate(@_,
               {conn => {type => OBJECT | UNDEF, optional => 1},
                node => {type => SCALAR,         optional => 0}});

    my ($state, $output);
    for (my $i = 1 ; $i <= 3 ; $i++) {
        $Log->comment("Getting node state for $opts{'node'} attempt $i");
        $state = 'normal';
        if ($opts{'conn'}) {

            my $output;
            eval {
                (undef, undef, $output) =
                  say($opts{'conn'}, 'cf status', -timeout, 300);
            };

            if (   $output
                && $output =~ /$opts{'node'}\s+has\s+been\s+taken\s+over/) {
                $state = 'cfo';
            } elsif ($@) {
                $state = 'unavailable';
            }
        } else {
            my $Server;
            my $node_health;
            $Server = Server->new(nodes => $opts{'node'});
            $Server->set_execution_server(-Server => $opts{'node'});
            eval {
                $node_health =
                  $Server->system_node_show(node         => $opts{'node'},
                                            output       => \$output,
                                            '-Noverbose' => 1,);

            };

            if ($@) {
                $Log->comment($output);
                $state = 'unavailable';
            } elsif ($$node_health[0]{'health'} =~ /false/) {
                $state = 'unavailable';
            } else {
                $Log->debug($output);
            }
        }
        if ($state ne 'unavailable') { last; }
    }

    $Log->comment("Node state for $opts{'node'} is $state");
    $Log->trace("Exiting get_node_state at " . localtime);

    return $state;

}

=head2 retrieve_clients

 Deprecated: Use Utils::HostManager::Helper::retrieve_clients
 
 use Utils::SSQA qw(retrieve_clients);
 retrieve_clients(script => "cifs_combo");

C<retrieve_clients> Retrieve clients from stored files or via HostManager

=over

=item options

(Required) The options hashref

=item script

(Required) The script name

=back

=cut

sub retrieve_clients {

    $Log->warn(
        "retrieve_clients is deprecated please use Utils::HostManager::Helper::retrieve_clients"
        );
    $Log->trace("Entering retrieve_clients at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();

    my (%opts) =
      validate(@_,
               {options => {type => HASHREF | UNDEF, optional => 1},
                script  => {type => SCALAR,          optional => 0}});

    my @clients;
    if ($opts{'script'}
        && -e "$options->{'TOP_LOGDIR'}/$opts{'script'}.pl") {

        #-----------------------------------------------------
        # Retrieve previously reserved clients from script.pl
        #-----------------------------------------------------
        our $DATA;
        require "$options->{'TOP_LOGDIR'}/$opts{'script'}.pl";
        $Log->comment(  "Retrieving "
                      . @{$DATA->{'CLIENT'}}
                      . " configured clients from $opts{'script'}.pl");
        @clients = @{$DATA->{'CLIENT'}};
    } elsif (-e "$options->{'TOP_LOGDIR'}/$options->{'DATA_FILE'}") {

        #--------------------------------------------
        # Retrieve unused clients from the DATA_FILE
        #--------------------------------------------
        our $DATA;
        require "$options->{'TOP_LOGDIR'}/$options->{'DATA_FILE'}";
        $Log->comment("Retrieving "
                . @{$DATA->{'CLIENT'}}
                . " available clients from TOP_LOGDIR/$options->{'DATA_FILE'}");
        @clients = @{$DATA->{'CLIENT'}};
    } else {

        #---------------------------------------------------------
        # Get clients based on CLIENTSPEC or CLIENT or CLIENTFILE
        #---------------------------------------------------------
        my $Host =
          new Utils::HostManager(ndispd_addr => "$options->{'NDISPD'}");
        @clients = reserve_clients('obj' => $Host);

        if (!-e "$options->{'TOP_LOGDIR'}/total_clients.pl") {

            #--------------------------------------------------------
            # Save the original clients so we can calculate averages
            #--------------------------------------------------------
            $Log->comment(  "Saving " 
                          . @clients
                          . " clients to "
                          . "\$TOP_LOGDIR/total_clients.pl");
            my %data;
            $data{'CLIENT'} = \@clients;
            write_dump('file'    => "$options->{'TOP_LOGDIR'}/total_clients.pl",
                       'creator' => $opts{'script'},
                       'var_name' => 'DATA',
                       'var_ref'  => \%data);
        }
    }

    $Log->trace("Exiting retrieve_clients at " . localtime);

    if (!@clients) {
        die 'Unable to retrieve clients';
    } else {
        return @clients;
    }

}

=head2 store_and_free_clients

 Deprecated: Use Utils::HostManager::Helper::store_and_free_clients

 use Utils::SSQA qw(store_and_free_clients); 
 store_and_free_clients(script => "cifs_combo");

C<store_and_free_clients> Retrieve clients from stored files or via HostManager

=over

=item options

(Required) The options hashref

=item script

(Required) The script name

=back

=cut

sub store_and_free_clients {

    $Log->warn(
        "store_and_free_clients is deprecated please use Utils::HostManager::Helper::store_and_free_clients"
        );
    $Log->trace("Entering store_and_free_clients at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();

    my (%opts) =
      validate(@_,
               {options => {type => HASHREF | UNDEF, optional => 1},
                script  => {type => SCALAR,          optional => 1},
                total   => {type => ARRAYREF,        optional => 1},
                used    => {type => ARRAYREF,        optional => 1},});

    #---------------------------------------------------------
    # Check to see if we have unused clients that we can free
    #---------------------------------------------------------
    if (@{$opts{'used'}} < @{$opts{'total'}}) {

        my @notUsed;
        for (my $i = 0 ; $i < @{$opts{'total'}} ; $i++) {
            my $c = ${$opts{'total'}}[$i];
            if (!grep(/$c/, @{$opts{'used'}})) {
                push(@notUsed, $c);
            }
        }
        my %data;
        if (   $options->{'NDISPD'}
            && $options->{'CLIENTSPEC'}
            && $options->{'FREE_UNUSED_CLIENTS'} == 1
            && @notUsed) {

            #--------------------------------------------------------------
            # Free unused clients. Normally the last script to use clients
            # does this.
            #--------------------------------------------------------------
            $Log->comment(  "$options->{'SCRIPT_PHASE'} - Freeing " 
                          . @notUsed
                          . " unused clients");
            my $Host =
              new Utils::HostManager(ndispd_addr => "$options->{'NDISPD'}");
            $Host->free_clients(hosts => \@notUsed);

        } else {

            #---------------------------------------------------------
            # Save the unused clients so the next script can use them
            #---------------------------------------------------------
            $Log->comment(  "$options->{'SCRIPT_PHASE'} - Saving " 
                          . @notUsed
                          . " available clients "
                          . "to \$TOP_LOGDIR/$options->{'DATA_FILE'}");
            $options->{'CLIENT'} = \@{$opts{'used'}};
            $data{'CLIENT'} = \@notUsed;
            write_dump(
                   'file' => "$options->{'TOP_LOGDIR'}/$options->{'DATA_FILE'}",
                   'creator'  => $opts{'script'},
                   'var_name' => 'DATA',
                   'var_ref'  => \%data);

        }

    }

    #--------------------------------------------------------
    # Save the used clients so the script can use them later
    #--------------------------------------------------------
    if (@{$opts{'used'}} && $opts{'script'}) {
        my %data;
        $Log->comment(  "$options->{'SCRIPT_PHASE'} - Saving "
                      . @{$opts{'used'}}
                      . " used clients "
                      . "to \$TOP_LOGDIR/$opts{'script'}.pl");
        $data{'CLIENT'} = \@{$opts{'used'}};
        write_dump('file'     => "$options->{'TOP_LOGDIR'}/$opts{'script'}.pl",
                   'creator'  => $opts{'script'},
                   'var_name' => 'DATA',
                   'var_ref'  => \%data);

    }

    $Log->trace("Exiting store_and_free_clients at " . localtime);
    return @{$opts{'used'}};
}

=head2 get_delay_value

 use Utils::SSQA qw(get_delay_value); 
 get_delay_value(obj => $hostObj, values => \@delays);

C<get_delay_value> Determine the state of the node

=over

=item obj

(Optional) The tharn host object

=item values

(Required) An array ref of delay values

=back

=cut

sub get_delay_value {

    $Log->trace("Entering get_delay_value at " . localtime);

    my (%opts) =
      validate(@_,
               {obj    => {type => OBJECT,   optional => 0},
                values => {type => ARRAYREF, optional => 0}});

    my $delay;
    if (@{$opts{'values'}} > 1) {

        #-------------------------------------------------------
        # If multiple delay values are specified in the form of
        # config=seconds and one of those configs matches the
        # clients config then use that seconds value for delay
        #-------------------------------------------------------
        my %hostFileParams = $opts{'obj'}->param_fields();
        my $matched        = 0;
        for (my $i = 1 ; $i < @{$opts{'values'}} ; $i++) {
            my ($config, $seconds) =
              split(/=/, ${$opts{'values'}}[$i]);
            if ($hostFileParams{'CONFIGS'} =~ /$config/) {
                $delay = $seconds;
            }
        }

        #-------------------------------------------------------------
        # If we did not match a config use the default, the 1st array
        # index
        #-------------------------------------------------------------
        if (!$matched) {
            $delay = ${$opts{'values'}}[0];
        }
    }
    else {
        $delay = ${$opts{'values'}}[0];
    }
    $Log->trace("Exiting get_delay_value " . localtime);

    return $delay;

}

=head2 wait_for_stats

 use Utils::SSQA qw(wait_for_stats); 
 wait_for_stats(conn => $conn, client => $hostObj, values => \@delays, configured => $x, 10mode_key => $key1, 7mode_key => $key2);

C<wait_for_stats> Waits until the the current value reaches the configured 
value. For 7G/Mode this is target a single filer for 10mode this is targeted
against the cluster.

=over

=item conn

(Optional) The tharn connection object

=item client

(Required) The hostid of the client

=item values

(Required) An array ref of delay values

=item configured

(Required) The configured value we are waiting for the node/cluster to match

=item 10mode_key

(Required) The 10mode counter to use

=item 7mode_key

(Required) The 7mode counter to use

=item max_failures

(Optional) The max number of stats fail to increase

=item max_queries

(Optional) The max number of stats will be queried

=back

=cut

sub wait_for_stats {

    $Log->trace("Entering wait_for_stats at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();

    my (%opts) =
      validate(@_,
               {conn          => {type => OBJECT | UNDEF,  optional => 1},
                client        => {type => OBJECT | SCALAR, optional => 0},
                values        => {type => ARRAYREF,        optional => 0},
                configured    => {type => SCALAR,          optional => 1},
                '10mode_key'  => {type => SCALAR,          optional => 1},
                '7mode_key'   => {type => SCALAR,          optional => 1},
                'max_queries' => {
                             type     => SCALAR,
                             optional => 1,
                             default => $options->{'WAIT_FOR_STATS_MAX_QUERIES'}
                  },
                'max_failures' => {
                            type     => SCALAR,
                            optional => 1,
                            default => $options->{'WAIT_FOR_STATS_MAX_FAILURES'}
                  }});

    #---------------------------------------------------------
    # If the configured stat is still greater than the current
    # stat then wait some  minutes for the stats to increase
    #---------------------------------------------------------
    my ($stats, $totalSleepTime, $currentValue, $key);

    if ($opts{'conn'}) {
        $key = $opts{'7mode_key'};
    } else {
        $key = $opts{'10mode_key'};
    }

    $Log->comment("Waiting for $key to equal $opts{'configured'}");

    my @history = ();

    if (!ref($opts{'client'})) {
        $opts{'client'} = host($opts{'client'});
    }

    $Log->comment("Getting cluster/node statistics");
    if ($opts{'conn'}) {

        my $sysStats =
          getSysStats('conn'     => $opts{'conn'},
                      'duration' => 60,
                      'prefix'   => 0);

        foreach my $k (keys(%{$sysStats})) { $stats->{$k} = $sysStats->{$k}; }

        my $cifsStats =
          getCifsStats('conn'   => $opts{'conn'},
                       'mode'   => '',
                       'prefix' => 0);

        foreach my $k (keys(%{$cifsStats})) { $stats->{$k} = $cifsStats->{$k}; }

        $key          = $opts{'7mode_key'};
        $currentValue = $stats->{$key};

    } else {
        $stats        = getClusteredStats('prefix' => 0);
        $key          = $opts{'10mode_key'};
        $currentValue = $stats->{$key};
    }

    my $queryId = 1;
    push(@history, $currentValue);
    while ($opts{'configured'} > $history[@history - 1]) {

        my $failures = 0;
        if (@history > 3) {
            for (my $x = @history - 1 ; $x > 0 ; $x--) {
                for (my $y = $x - 1 ; $y >= 0 ; $y--) {
                    if (($history[$x] =~ /\d+/ || $history[$x] =~ /\d+/)
                        && $history[$y] >= $history[$x]) {
                        $failures++;
                        last;
                    }
                }
            }
            if ($failures >= $opts{'max_failures'}) {
                my $msg = "Statistics are not increasing. "
                  . "Failures: $failures History: @history";
                $Log->comment($msg);
                die $msg;
            } elsif ($queryId >= $opts{'max_queries'}) {
                my $msg =
                    "The max number of queries defined "
                  . "by WAIT_FOR_STATS_MAX_QUERIES "
                  . "($opts{'max_queries'}) has been "
                  . "been reached. Failures: $failures "
                  . "History: @history";
                $Log->comment($msg);
                die $msg;
            }
        }

        $queryId++;
        my $sleep = get_delay_value('obj'    => $opts{'client'},
                                    'values' => $opts{'values'});
        $totalSleepTime += $sleep;
        $Log->comment(  "Configured $key ($opts{'configured'}) does not "
                      . "match the current value of $key ("
                      . $history[@history - 1]
                      . "), sleeping for $sleep seconds");
        snooze $sleep;

        $Log->comment("Getting cluster/node statistics");
        if ($opts{'conn'}) {

            my $sysStats =
              getSysStats('conn'     => $opts{'conn'},
                          'duration' => 60,
                          'prefix'   => 0);

            foreach my $k (keys(%{$sysStats})) {
                $stats->{$k} = $sysStats->{$k};
            }

            my $cifsStats =
              getCifsStats('conn'   => $opts{'conn'},
                           'mode'   => '',
                           'prefix' => 0);

            foreach my $k (keys(%{$cifsStats})) {
                $stats->{$k} = $cifsStats->{$k};
            }

            $currentValue = $stats->{$key};

        } else {
            $stats = getClusteredStats('prefix' => 0);
            $currentValue = $stats->{$key};

        }

        push(@history, $currentValue);

    }

    $Log->comment(  "Configured $key ($opts{'configured'}) "
                  . "matches or exceeds the current value of $key ("
                  . $history[@history - 1]
                  . ")");

    $Log->trace("Exiting wait_for_stats at " . localtime);
}

=head2 process_result

 use Utils::SSQA qw(process_result); 
 process_result(type => "error", msg => "Help Ive fallen and can't get up");

C<report_result> Process result type. All type get a log messages. If the type is error then die

=over

=item type

(Optional) The NATE result type

=item msg

(Required) The log message

=back

=cut

sub process_result {

    $Log->trace("Entering process_result at " . localtime);

    #-----------------------------------------------
    # before you change this please see burt 347025
    #-----------------------------------------------
    my (%opts) = @_;

    if (!$opts{msg}) { $opts{msg} = 'No result message specified'; }

    my $result = $opts{type} . "_result";
    no strict 'refs';
    my $res = &$result($opts{msg});
    if ($res->abortive()) {
        $Log->trace("Exiting process_result at " . localtime);
        die $res;
    }
    $Log->log($res);
    use strict 'refs';

    $Log->trace("Exiting process_result at " . localtime);

}

=head2 parse_spec

 use Utils::SSQA qw(parse_spec);
 my @specs = parse_spec("key1=value1,key2=value2 key3=value3,key4=value4");

C<parse_spec> returns a list of hash refs

=over

=item spec

(Required) A specification string consisting of a space-separated list of comma-separated name=value pairs to be parsed. 

=back

=cut

sub parse_spec ( $ ) {

    my $spec = shift;

    my @list;

    # extract attributes and create hash
    my @specs = split(/\s+/, $spec);
    foreach my $s (@specs) {
        my $args = {};
        my @attributes = split(/,/, $s);
        foreach my $attr (@attributes) {
            my ($key, $value) = ($attr =~ /(.*?)=(.*)/);
            $args->{$key} = $value;
        }
        push(@list, $args);
    }
    return @list;
}

=head2 convert_to_gb

 use Utils::SSQA qw(convert_to_gb);
 convert_to_gb('conn' => $fc);

C<convert_to_gb> Convert values to GB.

=over

=item conn

(Required) The filers connection object

=back

=cut

sub convert_to_gb {

    $Log->trace("Entering convert_to_gb at " . localtime);

    my (%opts) = validate(@_, {value => {type => SCALAR, optional => 0},});

    my $convertedValue = 0;

    if ($opts{'value'} =~ /(\d+)(^B$|^KB$|^MB$|^GB$|^TB$|^PB$)/) {
        my $value = $1;
        my $unit  = $2;
        if ($unit =~ /^B$|^KB$|^MB$/) {
            $convertedValue = 0;
        } elsif ($unit eq 'GB') {
            $convertedValue = $value;
        } elsif ($unit eq 'TB') {
            $convertedValue = int($value * 1073741824);
        } elsif ($unit eq 'PB') {
            $convertedValue = int($value * 1073741824000);
        }
    } elsif ($opts{'value'} =~ /\d+/) {
        $convertedValue = int($opts{'value'} / 1073741824);
    }
    $Log->trace("Exiting convert_to_gb at " . localtime);

    return $convertedValue;
}

=head2 check_required_space

 use Utils::SSQA qw(check_required_space);
 check_required_space(conn => $fc, clients => \@clients);

C<check_required_space> Convert values to GB.

=over

=item conn

(Optional) The filers connection object

=item clients

(Required) Arrayref of clients

=back

=cut

sub check_required_space {

    $Log->trace("Entering check_required_space at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();

    if ($options->{'GB_REQ_PER_CLIENT'} == 0) {
        $Log->comment(
                 "Skipping required space checking due to GB_REQ_PER_CLIENT=0");
        $Log->trace("Exiting check_required_space at " . localtime);
        return;
    }

    my (%opts) =
      validate(@_,
               {conn    => {type => OBJECT | UNDEF, optional => 1},
                clients => {type => ARRAYREF,       optional => 0}});

    my $stats;
    if ($opts{'conn'}) {
        $stats = getVolStats('conn'     => $opts{'conn'},
                             'vol_list' => $options->{'VOLUME'},
                             'prefix'   => 0);
    } else {
        $stats = getVolStats('vol_list' => $options->{'VOLUME'}, 'prefix' => 0);
        foreach my $key (keys(%{$stats})) {
            if ($key =~ /total|used|available/) {
                $stats->{$key} = convert_to_gb('value' => $stats->{$key});
            }
        }
    }

    my $maxClients = get_max_clients('clients' => $opts{'clients'});

    foreach my $volume (@{$options->{'VOLUME'}}) {
        my $avail;
        foreach my $key (keys(%{$stats})) {
            if ($key =~ /$volume-avail/) { $avail = $stats->{$key}; }
        }
        if (!$avail) { die "Unable to determine volume key in stats hash"; }

        if (   $options->{'VOLUME_USAGE'} eq 'single'
            && $volume eq @{$options->{'VOLUME'}}[$options->{'VOLUME_INDEX'}]) {

            if ($maxClients > 0
                && $avail < (($maxClients / @{$options->{'VOLUME'}}) *
                               $options->{'GB_REQ_PER_CLIENT'})
              ) {
                die "Error not enough space on volume $volume to start "
                  . "$maxClients clients when GB_REQ_PER_CLIENT="
                  . "$options->{'GB_REQ_PER_CLIENT'}. Current:${avail}GB Need:"
                  . ($maxClients * $options->{'GB_REQ_PER_CLIENT'}) . "GB";

            }
            elsif ($maxClients < 1
                   && $avail < (
                               (@{$opts{'clients'}} / @{$options->{'VOLUME'}}) *
                                 $options->{'GB_REQ_PER_CLIENT'})
              ) {
                die "Error not enough space on volume $volume to start "
                  . @{$opts{'clients'}}
                  . " clients when GB_REQ_PER_CLIENT="
                  . "$options->{'GB_REQ_PER_CLIENT'}. Current:${avail}GB Need:"
                  . (@{$opts{'clients'}} * $options->{'GB_REQ_PER_CLIENT'})
                  . "GB";

            }

        } elsif ($options->{'VOLUME_USAGE'} eq 'multiple') {

            if ($maxClients > 0
                && $avail < (($maxClients / @{$options->{'VOLUME'}}) *
                               $options->{'GB_REQ_PER_CLIENT'})
              ) {
                die "Error not enough space on volume $volume to start "
                  . "$maxClients clients when GB_REQ_PER_CLIENT="
                  . "$options->{'GB_REQ_PER_CLIENT'}. Current:${avail}GB Need:"
                  . ($maxClients * $options->{'GB_REQ_PER_CLIENT'}) . "GB";

            }
            elsif ($maxClients < 1
                   && $avail < (
                               (@{$opts{'clients'}} / @{$options->{'VOLUME'}}) *
                                 $options->{'GB_REQ_PER_CLIENT'})
              ) {
                die "Error not enough space on volume $volume to start "
                  . @{$opts{'clients'}}
                  . " clients when GB_REQ_PER_CLIENT="
                  . "$options->{'GB_REQ_PER_CLIENT'}. Current:${avail}GB Need:"
                  . (@{$opts{'clients'}} * $options->{'GB_REQ_PER_CLIENT'})
                  . "GB";

            }
        }

    }
    $Log->trace("Exiting check_required_space at " . localtime);
}

=head2 check_capacity

 use Utils::SSQA qw(check_capacity);
 check_capacity(conn => $fc, clients => \@clients);

C<check_capacity> Convert values to GB.

=over

=item conn

(Optional) The filers connection object

=item clients

(Required) Arrayref of clients

=back

=cut

sub check_capacity {

    $Log->trace("Entering check_capacity at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();

    if ($options->{'CAPACITY'} == 0) {
        $Log->comment(  "$options->{'SCRIPT_PHASE'} - Skipping "
                      . "capacity checking due to CAPACTIY=0");
        $Log->trace("Exiting check_capacity at " . localtime);
        return;
    }

    $Log->comment("$options->{'SCRIPT_PHASE'} - Checking volume capacities");

    my (%opts) =
      validate(@_, {conn => {type => OBJECT | UNDEF, optional => 1}});

    my $stats;
    if ($opts{'conn'}) {
        $stats = getVolStats('conn'     => $opts{'conn'},
                             'vol_list' => $options->{'VOLUME'},
                             'prefix'   => 0);
    } else {
        $stats = getVolStats('vol_list' => $options->{'VOLUME'}, 'prefix' => 0);
    }

    foreach my $volume (@{$options->{'VOLUME'}}) {
        my $capacity;
        foreach my $key (keys(%{$stats})) {
            if ($key =~ /$volume-(capacity|percent-used)/) {
                $capacity = $stats->{$key};
            }
        }

        if (!$capacity) { die "Unable to determine volume key in stats hash"; }

        if ($capacity > $options->{'CAPACITY'}) {
            die "The capacity of volume $volume exceeds $options->{'CAPACITY'}";
        }

    }
    $Log->trace("Exiting check_capacity at " . localtime);
}

=head2 get_max_clients

 Deprecated: Use Utils::HostManager::Helper::get_max_clients

 use Utils::SSQA qw(get_max_clients);
 get_max_clients(clients => \@clients);

C<get_max_clients> Convert values to GB.

=over

=item clients

(Required) Arrayref of clients

=back

=cut

sub get_max_clients {

    $Log->warn(
        "get_max_clients is deprecated please use Utils::HostManager::Helper::get_max_clients"
        );
    $Log->trace("Entering get_max_clients at " . localtime);

    $Options->process();
    my $options    = $Options->get_hash();
    my $maxClients = 0;
    if ($options->{'MAX_CLIENTS'} == 0) {
        $Log->debug(  "$options->{'SCRIPT_PHASE'} - Skipping "
                    . "client calculations due to MAX_CLIENTS=0");
        $Log->trace("Exiting get_max_clients at " . localtime);
        return $maxClients;
    }

    my @clients;
    if ($options->{'MAX_CLIENTS'} =~ /(\d+)%/) {
        my $maxClientPercent = $1;
        if (-e "$options->{'TOP_LOGDIR'}/total_clients.pl") {
            our $DATA;
            require "$options->{'TOP_LOGDIR'}/total_clients.pl";
            @clients = @{$DATA->{'CLIENT'}};
            $Log->comment(  "Calculating MAX_CLIENTS from " 
                          . @clients
                          . " clients retrieved from "
                          . "TOP_LOGDIR/total_clients.pl");
            $maxClients = ceil((@clients * $maxClientPercent) / 100);
        } else {
            die "TOP_LOGDIR/total_clients.pl required to calculate "
              . "MAX_CLIENT percentages";
        }
    } elsif ($options->{'MAX_CLIENTS'}) {
        $maxClients = $options->{'MAX_CLIENTS'};
    }

    $Log->comment("Using MAX_CLIENTS value of $maxClients");
    $Log->trace("Exiting get_max_clients at " . localtime);

    return $maxClients;
}

=head2 strip_vars

 $clientOptions = strip_vars();

C<strip_vars> Returns a options hashref minus vars that interfere with windows subtests

=over

=back

=cut

sub strip_vars {

    $Log->trace("Entering strip_vars at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();
    my $newOptions;

    foreach my $key (keys(%{$options})) {
        if (!grep { $_ eq $key } @Tharn::Alias_params && $key ne 'CLIENTFILE') {
            $newOptions->{$key} = $options->{$key};
        }
    }

    $Log->trace("Exiting strip_vars at " . localtime);
    return $newOptions;

}

=head2 mem_to_csv

 mem_to_csv(file => string, path => string);

C<mem_to_csv> Converts mem files to CSV format

=over

=item file

(Required) The mem file name

=item path

(Required) The path the directory contain the mem files 

=back

=cut

sub mem_to_csv {

    $Log->trace("Entering mem_to_csv at " . localtime);

    $Options->process();
    my $options = $Options->get_hash();

    my (%opts) =
      validate(@_,
               {file => {type => SCALAR, optional => 0},
                path => {type     => SCALAR,
                         optional => 0,
                         default  => $options->{'TOP_LOGDIR'}},});

    my $file = fileparse($opts{'file'}, qr/\.[^.]*/);
    if (   -e "$opts{'path'}/$file.mem"
        && -s "$opts{'path'}/$file.mem") {

        sleep 2;

        my %mem;
        open(MEM, "$opts{'path'}/$file.mem")
          || die "Cannot open $opts{'path'}/$file.mem\n";

        #---------------------------------------------------
        # Parse the memstats mem file and create stats hash
        #---------------------------------------------------
        # 07/26/09 22:37:21 DIFF
        # 8 17440 maytag.ko::ontap_liaison_main+247
        # 94 5264 maytag.ko::rt_replenish_pool+313

        # '07/26/09 22:37:21' => {
        #                          'ontap_liaison_main+247' => {
        #                                                        'bytes' => 17440,
        #                                                        'allocations' => '8'
        #                                                      },
        #                          'rt_replenish_pool+313' => {
        #                                                       'bytes' => 5264,
        #                                                       'allocations' => '94'
        #                                                     },
        #                          'state' => 4,
        #
        my $date;
        while (<MEM>) {
            chomp;
            if (/^(\d+\/\d+\/\d+\s+\d+:\d+:\d+)\s+(\S+)/) {
                $date = $1;
                my $state = $2;
                push(@{$mem{'dates'}}, $date);
                if ($state eq 'UNAVAILIBLE') {
                    $mem{$date}->{'state'} = 1;
                } elsif ($state eq 'CFO') {
                    $mem{$date}->{'state'} = 2;
                } elsif ($state eq 'BASELINE') {
                    $mem{$date}->{'state'} = 3;
                } elsif ($state eq 'DIFF') {
                    $mem{$date}->{'state'} = 4;
                }
            } elsif (/^(\d+)\s+(\d+)\s+(\S+)/) {
                my $allocations = $1;
                my $bytes       = $2;
                my $call        = $3;
                $call =~ s/maytag\.ko:://g;

                my $escapedCall = escape_string('string' => $call);
                if (!grep(/$escapedCall/, @{$mem{'calls'}})) {
                    push(@{$mem{'calls'}}, $call);
                }

                if (!$mem{'max_bytes'}->{$call}) {
                    $mem{'max_bytes'}->{$call} = $bytes;
                } elsif ($mem{'max_bytes'}->{$call} < $bytes) {
                    $mem{'max_bytes'}->{$call} = $bytes;
                }

                $mem{$date}->{$call}->{'allocations'} = $allocations;
                $mem{$date}->{$call}->{'bytes'}       = $bytes;
            }
        }
        close MEM;

        #----------------------------------------------------------------------
        # Create the max_calls array. This array consists of the top MAX_CALLS
        # calls over the MIN_BYTES threshold
        #----------------------------------------------------------------------
        my @maxByteKeys = reverse(
            sort {
                $mem{'max_bytes'}{$a} <=> $mem{'max_bytes'}{$b}
              } (keys(%{$mem{'max_bytes'}})));
        for (my $i = 0 ; $i < @maxByteKeys ; $i++) {
            if ($i > $options->{'MAX_CALLS'}) { last; }
            unshift(@{$mem{'max_calls'}}, $maxByteKeys[$i]);
        }

        #------------------------
        # Loop through all calls
        #------------------------
        my @calls;
        if ($mem{'calls'}) {
            foreach my $call (@{$mem{'calls'}}) {

                #--------------------------------------------------
                # Escape the call name so we can use it in a regex
                #--------------------------------------------------
                my $escapedCall = escape_string('string' => $call);

                #-------------------------------------------------------------
                # If IGNORED_CALLS are specified and the current call matches
                # any of the IGNORED_CALLS then skip
                #-------------------------------------------------------------
                if (@{$options->{'IGNORED_CALLS'}}
                    && grep { $call =~ /$?/ } @{$options->{'IGNORED_CALLS'}}) {
                    next;
                }

                #--------------------------------------------------------------
                # If MAX_CALLS is specified and the current call is not in the
                # top MAX_CALLS array skip
                #--------------------------------------------------------------
                if ($options->{'MAX_CALLS'}) {
                    if (!grep(/^$escapedCall$/, @{$mem{'max_calls'}})) {
                        next;
                    }
                }

                #----------------------------------------------------------
                # If MIN_BYTES is specified skip all calls whose max bytes
                # are > MIN_BYTES
                #----------------------------------------------------------
                if (   $options->{'MIN_BYTES'}
                    && !grep(/^$escapedCall$/, @calls)
                    && $mem{'max_bytes'}->{$call} > $options->{'MIN_BYTES'}) {
                    push(@calls, $call);
                }

                #-------------------------------
                # Set any missing calls to zero
                #-------------------------------
                for (my $d = 0 ; $d < @{$mem{'dates'}} ; $d++) {
                    my $date = ${$mem{'dates'}}[$d];
                    if (!$mem{$date}->{$call}->{'bytes'}) {
                        $mem{$date}->{$call}->{'bytes'} = 0;
                    }
                }

                my $firstHourEpochTime =
                  get_epoch_time('date' => ${$mem{'dates'}}[0]) + 3600;
                my $lastDateIndex = $#{$mem{'dates'}};

                #------------------------------------------------------------
                # Find all CFO and UNAVAIL states so we know when memory was
                # cleared. We will use these dates as window boundaries. Add
                # the last date also
                #------------------------------------------------------------
                my @windowEndDates;
                push(@windowEndDates, ${$mem{'dates'}}[$lastDateIndex]);

                for (my $d = $lastDateIndex ; $d > 0 ; $d--) {
                    my $date = ${$mem{'dates'}}[$d];
                    if ($mem{$date}->{'state'} =~ /(1|2)/) {
                        unshift(@windowEndDates, $date);
                    }
                }
                my $windowIndex = @windowEndDates;

                #------------------------------------------------------------
                # Walk backwards in time checking each byte value during the
                # DIFF state and Add calls that match certain criteria
                #----------------------------------------------------------
                my $memCleared = 0;
                my $windowStart;
                for (my $d = $lastDateIndex ; $d > 0 ; $d--) {
                    my $date = ${$mem{'dates'}}[$d];

                    #-------------------------------------------------------
                    # If the current date is < 1 hour from start of the run
                    # skip to the next call
                    #-------------------------------------------------------
                    if (get_epoch_time('date' => $date) < $firstHourEpochTime) {
                        last;
                    }

                    if (grep { $_ eq $date } @windowEndDates) {

                        #-------------------------------------------------
                        # If the current date is in the @windowEndDates
                        # Reset the the current window end to the current
                        # date and calculate a new window start
                        #-------------------------------------------------
                        $windowIndex--;
                        $windowStart = get_epoch_time(
                                       'date' => $windowEndDates[$windowIndex]);
                    } elsif (get_epoch_time('date' => $date) == $windowStart) {

                        #-----------------------------------------------------
                        # If the current date is less the window start, Reset
                        # the window start and end dates to the next window.
                        # If this is last window exit the loop, otherwise set
                        # the date index to the next window end.
                        #----------------------------------------------------
                        if ($windowIndex == 0) {
                            last;
                        } else {
                            $windowIndex--;
                            $windowStart = get_epoch_time(
                                       'date' => $windowEndDates[$windowIndex]);
                            $date = $windowEndDates[$windowIndex];
                            $d =
                              get_array_index('array' => $mem{'dates'},
                                              'value' => $date);
                        }
                    }

                    #-------------------------------------------------------
                    # If a 0 byte value is found when the state is DIFF and
                    # this is the last window exit the loop otherwise reset
                    # the window start and end dates to the next window and
                    # set the date index to the end of the the next window
                    #-------------------------------------------------------
                    if (   $mem{$date}->{$call}->{'bytes'} == 0
                        && $mem{$date}->{'state'} == 4) {

                        $memCleared++;
                        if ($windowIndex == 0) {
                            last;
                        } else {
                            $windowIndex--;
                            $date = $windowEndDates[$windowIndex];
                            $d =
                              get_array_index('array' => $mem{'dates'},
                                              'value' => $date);
                        }
                    }

                }

                #------------------------------------
                # Decide of we need to keep the call
                #------------------------------------
                if ($memCleared == 0) {

                    #--------------------------------------------
                    # No 0 byte value found for any window, keep
                    #--------------------------------------------
                    push(@calls, $call);
                } elsif ($mem{$lastDateIndex}->{$call}->{'bytes'} > 0) {

                    #------------------------------------------
                    # The call is currently using memory, keep
                    #------------------------------------------
                    push(@calls, $call);
                } elsif ($memCleared < @windowEndDates) {

                    #-------------------------------------------
                    # Not all windows have a 0 byte value, keep
                    #-------------------------------------------
                    push(@calls, $call);
                }

            }
        }

        #---------------------------
        # Open the memory CSV files
        #---------------------------
        my $path = $opts{'path'};
        if ($options->{'CSV_DIR'}) {
            $path .= "/$options->{'CSV_DIR'}";
        }

        open(CSV, ">$path/$file.csv")
          || die "Cannot open $path/$file.csv\n";

        #------------------------------------------------------------------
        # Sort the calls by bytes used so graph auto scaling works better
        #------------------------------------------------------------------
        my @sortedCalls = reverse(
            sort {
                $mem{'max_bytes'}{$a} <=> $mem{'max_bytes'}{$b}
              } (keys(%{$mem{'max_bytes'}})));

        #-------------------
        # Create CSV header
        #-------------------
        my $line = 'Timestamp,mem-state:Memory Monitoring State,';
        foreach my $call (@sortedCalls) {
            my $escapedCall = escape_string('string' => $call);
            if (grep(/^$escapedCall$/, @calls)) {
                $line .= "mem-call:$call,";
            }
        }
        $line =~ s/,$//;
        print CSV "$line\n";

        #-----------------
        # Create CSV data
        #-----------------
        for (my $d = 0 ; $d < @{$mem{'dates'}} ; $d++) {
            my $date = $mem{'dates'}[$d];
            my $line = "$date,$mem{$date}->{'state'},";
            foreach my $call (@sortedCalls) {
                my $escapedCall = escape_string('string' => $call);
                if (grep(/^$escapedCall$/, @calls)) {
                    $line .= "$mem{$date}->{$call}->{'bytes'},";
                }
            }
            $line =~ s/,$//;
            print CSV "$line\n";
        }

        #---------------------------
        # close the memory CSV files
        #---------------------------
        close CSV;
    }
    $Log->trace("Exiting mem_to_csv at " . localtime);
}

=head2 escape_string

 $escapedString = escape_string(string => string);

C<escape_string> escapes a string for use as a regex

=over

=item string

(Required) The string to escape

=back

=cut

sub escape_string {
    $Log->trace("Entering escape_string at " . localtime);
    my (%opts) = validate(@_, {string => {type => SCALAR, optional => 0}});
    my $string = $opts{'string'};
    $string =~ s/(\/|\+|\$|\.|\*|\^|\(|\))/\\$1/g;
    $Log->trace("Exiting escape_string at " . localtime);
    return $string;
}

=head2 get_epoch_time

 $clientOptions = get_epoch_time(date => string);

C<get_epoch_time> Converts a date string to epoch time

=over

=item date

(Required) The date string

=back

=cut

sub get_epoch_time {
    $Log->trace("Entering get_epoch_time at " . localtime);
    my (%opts) = validate(@_, {date => {type => SCALAR, optional => 0}});

    my $time;
    if ($opts{'date'} =~ /(\d+)\/(\d+)\/(\d+)\s+(\d+):(\d+):(\d+)/) {
        my $month = $1;
        $month--;
        $time = timelocal($6, $5, $4, $2, $month, $3);
    }
    $Log->trace("Exiting get_epoch_time at " . localtime);
    return $time;
}

=head2 get_array_index

 $index = get_array_index(array => \@array, value => foo);

C<get_array_index> Searches an array for value and returns the index

=over

=item array

(Required) Arrayref

=item value

(Required) The value to seach for

=back

=cut

sub get_array_index {
    $Log->trace("Entering get_array_index at " . localtime);
    my (%opts) =
      validate(@_,
               {array => {type => ARRAYREF, optional => 0},
                value => {type => SCALAR,   optional => 0},});

    for (my $i = 0 ; $i < @{$opts{'array'}} ; $i++) {
        if (${$opts{'array'}}[$i] eq $opts{'value'}) {
            $Log->trace("Exiting get_array_index at " . localtime);
            return $i;
        }
    }
    $Log->trace("Exiting get_array_index at " . localtime);
}

sub sort_hostnames {
    my (@orighosts) = @_;
    my %hosts;

    foreach my $host (@orighosts) {

        # if hostname has two digits or single digit use them to
        # sort based on the number. Otherwise, sort by perl.
        if ($host =~ m/\D*(\d+)\D*(\d+)/) {
            $hosts{$host} = sprintf("%04d%03d", $1, $2);
        } elsif ($host =~ m/\D*(\d+)/) {
            $hosts{$host} = sprintf("%04d", $1);
        } else {
            $hosts{$host} = 1;
        }
    }

    return sort { $hosts{$a} <=> $hosts{$b} } (keys %hosts);
}

#--------------------------------------------------------------------
1;
