# $Id$
## @summary Common library for termination file functions
## @author bellizzi@netapp.com
## @pod here

=head1 NAME

Utils::TermFile.pm - Library for termination file functions

=head1 DESCRIPTION

This package is contains routines to hack stest behavior

=head1 SYNOPSIS

    use Utils::TermFile qw(check_term_file write_term_string);

=head1 ROUTINES

=cut

package Utils::TermFile;
use strict;
use Tharn;
$Log->comment("Starting " . __FILE__);
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);
use Utils::Dump qw(write_dump);
use Data::Dumper;

our $VERSION = 1.00;
our @ISA     = qw (Exporter);
our @EXPORT  = qw (
  check_term_file
  write_term_string
  );

=head2 check_term_file

 use Utils::TermFile qw(check_term_file);
 check_term_file(term_file => "$TOP_LOGDIR\$TERMFILE", term_strings => \$TERMSTRINGS);

C<check_term_file> Checks the term_file for the presence all term_strings

=over

=item term_file

(Required) The termination file

=item term_strings

(Required) Array of strings to check for in the term_file

=back

=cut

sub check_term_file {

    $Log->enter();

    my (%opts) =
      validate(@_,
               {term_file    => {type => SCALAR,   optional => 0},
                term_strings => {type => ARRAYREF, optional => 0},
               }
               );

    $Log->comment("Checking termination file $opts{'term_file'}");

    my @term_file_contents = ();
    open(TERM, "$opts{'term_file'}");
    while (<TERM>) {
        chomp;
        push(@term_file_contents, $_);
    }
    close TERM;

    my $all_strings_found = 1;
    foreach my $s (@{$opts{'term_strings'}}) {
        if (!grep(/$s/, @term_file_contents)) {
            $all_strings_found = 0;
            last;
        }
    }

    if (!$all_strings_found) {
        $Log->comment(  "Not all TERMSTRINGS exist in termination "
                      . "file, continuing\n@term_file_contents");
    }

    $Log->exit();

    if ($all_strings_found) {
        die "All TERMSTRINGS found in "
          . "termination file, exiting\n@term_file_contents";
    }

}

=head2 write_term_string

 use Utils::TermFile qw(write_term_string);
 write_term_string(term_file => "$TOP_LOGDIR\$TERMFILE", term_string => \$TERMSTRING);

C<write_term_string> Write the term string to the term file

=over

=item term_file

(Required) The termination file

=item term_string

(Required) the string to write to the term file

=back

=cut

sub write_term_string {

    $Log->enter();

    my (%opts) =
      validate(@_,
               {term_file   => {type => SCALAR, optional => 0},
                term_string => {type => SCALAR, optional => 0},
               }
               );

    $Log->comment(  "Writing termination string $opts{'term_string'} "
                  . "to termination file $opts{'term_file'}");
    my $status = 0;
    for (my $i = 1 ; $i <= 5 ; $i++) {
        if (open(TERM, ">>$opts{'term_file'}")) {
            print TERM "$opts{'term_string'}\n";
            close TERM;
            $status = 1;
            last;
        } else {
            snooze 10;
        }
    }

    if (!$status) {
        die "Error, Unable to write $opts{'term_string'} to $opts{'term_file'}";
    }

    $Log->exit();
}

1;
