# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Utils/Units.pm#1 $

## @summary Storage Unit Conversion Routines
## @author  dl-test-infra@netapp.com
## @status  Shared
## @pod here

package Utils::Units;
use strict;
use warnings;
use Module::Common;
my(%unit2bytes, %abbrev2unit, %unit2abbrev);
our $ALLOWED_UNITS_RE;

BEGIN
{
	# XXX - If you want to add more unit types, simply add it
	# to this array (@bytes). Utils::Units will automagically
	# create the constants for you.
	# NOTE: The *ORDER MATTERS*! If someday we introduce a
	# unit that is not geometrically proportional to 1024,
	# we'd need to change it

	#               0 1  2  3  4
	my @bytes = qw( B KB MB GB TB );
	my @constants;

	# generate the constants and calculate the multiplier
	# for each unit, which should be 1024 ** $idx
	# (B => 1, KB => 1024, MB => 1024 ** 2... etc)
	#
	# so we're basically doing UNIT_B => 'B', UNIT_KB => 'KB',
	# and $unit2bytes{B} = 1, $unit2bytes{KB} => 1024 ..
	require constant; # XXX - voodoo. don't do this at home
	for my $idx (0..$#bytes) {
		my $unit   = $bytes[$idx];
		my $symbol = "UNIT_$unit";

		# create constant...
		constant->import($symbol, $unit);
		push @constants, $symbol;

		# calculate appropriate factor..
		$unit2bytes{ $unit } = 1024 ** $idx;

		# now for convenience, map abbreviation
		# (first letter) to actual constant
		my $abbrev = lc(substr($unit, 0, 1));
		$abbrev2unit{ $abbrev } = $unit;
		$unit2abbrev{ $unit } = $abbrev;
	}

	# XXX - create a regular expression that will be used to
	# match allowed unit specifications (e.g. "b", "k", "kb")

	# XXX - don't include bytes[0], because that's not allowed
	# as a spec. if you want bytes, leave out the spec
	my $ps = join( "", map { lc(substr($_, 0, 1)) } @bytes[1..$#bytes] );
	$ALLOWED_UNITS_RE = qr!(?:(?i)[$ps])!;

	Module::Common::setup_export_tags(
		convert => [ qw(
			guess_unit
			convert_units
			convert_to_bytes
			convert_to_units
			strip_unit_spec
			ontap2standard_unit
			standard2ontap_unit
		) ],
		constants => \@constants
	);
}
use Params::Validate qw(validate validate_pos SCALAR);
use Carp qw(confess);
use Tharn qw(script_result);
BEGIN
{
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
                confess(script_result(@msgs))}
        );
}

=head1 NAME

Utils::Units - Simple Unit Related Routines

=head1 SYNOPSIS

  use Utils::Units qw(:convert);
  my $size_in_bytes     = convert_to_bytes( "100k" );
  my $size_in_megabytes = convert_to_units(
    $size_in_bytes, units => UNIT_MB
  );

  my $size = 1_000_000;
  print "$size in KB => ",
    convert_units( $size, UNIT_B, UNIT_KB ), "\n";
  print "$size in MB => ",
    convert_units( $size, UNIT_B, UNIT_MB ), "\n";
  print "$size in GB => ",
    convert_units( $size, UNIT_B, UNIT_GB ), "\n";
  print "$size in TB => ",
    convert_units( $size, UNIT_B, UNIT_TB ), "\n";

  use Utils::Units qw(:constants);
  # load constants UNIT_B, UNIT_KB, UNIT_MB, UNIT_GB, UNIT_TB

  use Utils::Units qw(:constants guess_unit);
  if( guess_unit("100M") eq UNIT_MB ) {
     print "size is in MB\n";
  }

=head1 DESCRIPTION

Utils::Units provides simple unit conversion routines, and
some associated constants

=head1 FUNCTIONS

=head2 convert_units($size,$from,$to)

This is the most generic unit conversion routine which converts $size
from units denoted by $from to units denoted by $to, and returns the
result. On failure, returns undef, setting $@.

The constants UNIT_* should be used for $from and $to.

=cut

sub convert_units
{
	my($value, $from, $to) = validate_pos(@_,
		{ regex => qr/^\d+$/ },
		{ callbacks => {
			'valid unit' => sub { exists $unit2bytes{$_[0]} }
		} },
		{ callbacks => {
			'valid unit' => sub { exists $unit2bytes{$_[0]} }
		} }
	);
	return eval { # XXX remove the return eval
		if($from eq $to) {
			return $value;
		}

		my $factor_from = $unit2bytes{$from};
		my $factor_to   = $unit2bytes{$to};

		# first convert to bytes, and then convert to whatever
		# the user is asking for...
		my $bytes = $value * $factor_from;

		return $bytes / $factor_to;
	}
}

=head2 guess_unit($string)

Given a string that contains something like "100k", guesses the units
that the string is represented in and returns it. On failure returns
undef, setting $@.

If $string does not end with a known unit character, assumes that 
$string is in bytes.

=cut

sub guess_unit
{
	my($size) = validate_pos(@_, { type => SCALAR });
	return eval{ # XXX remove the return eval
		if ($size eq "" || $size !~ /^\d+($ALLOWED_UNITS_RE)?$/){
			die "Unable to parse string: $size";
		}
		my $matched = $1;

		# XXX - $matched can be from 2 characters to 0 characters.
		# if 2 characters, we need to get the first character only
		if ($matched && length($matched) > 1) {
			$matched = lc( substr($matched, 0, 1) );
		}

		my $unit;
		if(!$matched) {
			$unit = UNIT_B;
		} else {
			$unit = $abbrev2unit{ lc($matched) };
		}

		if(!$unit) {
			die "Invalid unit: '$matched'\n";
		}
		return $unit;
	}
}

=head2 convert_to_bytes($size)

B<NOTE: You shuld probably be using convert_units() instead!>

Returns the value of $size in bytes. This function is a combination of
guess_unit() and convert_units(), so you can use a number followed by
a character denoting the unit:

  $size_in_bytes = convert_to_bytes( "100m" );

=cut

sub convert_to_bytes($){
	my $size = shift;
	return eval{ # XXX remove the return eval
		my $unit = guess_unit($size) or
			die "Could not guess unit: $size\n";

		# strip out non-numbers from $size's end
		return convert_units(
			strip_unit_spec($size),
			$unit,
			UNIT_B
		) or die $@;
	};
}

=head2 convert_to_units($size,units=>$unit)

B<NOTE: You shuld probably be using convert_units() instead!>

Returns the value of $size in the units specified by $unit. The 
value of $size has to be in bytes prior to calling this function. 

  $size_in_kb = convert_to_unit($size, units => UNIT_KB);


=cut

sub convert_to_units($;%){
	my $size_in_bytes = shift;
	my %args = validate(@_, {
		units  => {
			optional  => 1,
			default   => UNIT_KB,
			callbacks => {
				'valid unit' => sub {
					exists $abbrev2unit{$_[0]};
				}
			}
		}
	});

	return eval { # XXX remove the return eval
		my $new_unit = $abbrev2unit{$args{units}};
		convert_units($size_in_bytes, UNIT_B, $new_unit)
			or die $@;
	};
}

=head2 strip_unit_spec($value)

Strips out the trailing unit specification, if any. E.g. "10k" would
be come "10".

=cut

sub strip_unit_spec
{
	my($v) = validate_pos(@_, { type => SCALAR });
	$v =~ s/$ALLOWED_UNITS_RE$//;
	return $v;
}

=head2 ontap2standard_unit($unit)

=head2 standard2ontap_unit($ontap_unit)

These converts units between units that OnTap uses ('m', 'g', 't') and
units that are used in this module, UNIT_MB, UNIT_GB, UNIT_TB.

=cut 

sub ontap2standard_unit
{
	my($v) = validate_pos(@_, { type => SCALAR });
	return $abbrev2unit{ lc($v) };
}

sub standard2ontap_unit
{
	my($v) = validate_pos(@_,
		{ type => SCALAR,
			callbacks => {
				'valid unit' => sub {
					exists $unit2abbrev{$_[0]};
				}
			}
		}
	);
	return $unit2abbrev{ $v };
}

1;


=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-test-infra@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/Utils/Units.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

