# -*- perl -*-
# $Id: $

## @summary Dispatcher
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Dispatcher 

=head1 SYNOPSIS

  use Dispatcher

=head1 DESCRIPTION

Work dispatcher

    # make some work  
    #   work consists of a list of work items,
    #   a work item is a hash of:
    #       method=>"method/subroutine name to call"
    #       args=>"arguments to pass to the method"
    #   work list can include hashes for multiple methods 
    my @work;
    for (my $i=0; $i<10; $i++) {
        push(@work,{
                method=>'_worker_a',
                args=>{
                    id=>$i,
                    text=>"task $i"
                }
            });
    };

    # do the work - spawn 3 worker processes to process the work list
    my $dispatcher=Utils::WorkDispatcher->new( nworkers=>3, work=>\@work);
    $dispatcher->start(); # start workers
    $dispatcher->wait();  # wait until all work is done

=begin html

<div class=wsd wsd_style="rose" ><pre>

    title Dispatch Work (to single worker)

    thpl->WorkDispatcher: new
    thpl->+WorkDispatcher: start
    WorkDispatcher->_worker: main::subtest
    WorkDispatcher->-thpl:

    thpl->+WorkDispatcher: wait

    opt loop
    WorkDispatcher->_worker: SUBTASK::RPC::COMMAND(work)
    _worker->+thpl: callback(args)
    thpl-->-_worker:
    _worker->WorkDispatcher: SUBTASK::RPC::REPLY(id)
    end

    WorkDispatcher->_worker: SUBTASK::RPC::COMMAND(_exit)

    WorkDispatcher->-thpl:

</pre></div><script type="text/javascript" src="http://www.websequencediagrams.com/service.js"></script>

=end html

=cut

# --------------------------------------------------------
# Class: CConfig::List
#   Class to abstract a list of values and include the concept of a current position
# --------------------------------------------------------

package Utils::WorkDispatcher;
use strict;
use warnings;
use Params::Validate qw(:all);
use Data::Dumper;
use Carp;

use NATE::Log qw(log_global);
use NATE::TestInfo;
use NATE::STAF::Peer qw($Parent);

# -----------------------------------------------
# Function: new
#   constructor
#
# Parameters:
#       list      - reference to an array of work to be performed
#
# Return: 
#   blessed reference
# -----------------------------------------------
sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;

    my $self = {@_};
    bless($self, $class);
}

# -----------------------------------------------
# Accessor Function:  workers - list of worker subtest
# -----------------------------------------------
sub workers   { my $self = shift; return @{$self->{workers}}; }

# -----------------------------------------------
# Function: start
#   dispatch workers to do work
#
# Parameters:
#
# Return: 
#  
# -----------------------------------------------
sub start {
    my $self=shift;
    my @workers;

    # start up the workers...
    for (my $i=0; $i<$self->{'nworkers'}; $i++) {
        push(@workers,main::subtest(
            sub { $self->_worker($i) },
            "-runid", "worker_".sprintf("%03d",$i),
            "-bg",
            ));
    }
    $self->{'workers'}=\@workers;
    # now give them something to do
    foreach my $worker (@workers) { $self->dispatcher($worker); }
}

# -----------------------------------------------
# Function: wait
#   wait for workers to complete 
#
# Parameters:
#   none
#
# Return: 
#   none
#  
# -----------------------------------------------
sub wait {
    my $self=shift;
    $main::Log->comment("wait for workers to complete");
    Subtest::wait_finish(subtests => $self->{'workers'},@_);
    $main::Log->comment("all workers complete");
    return;
}

# -----------------------------------------------
# Function: dispatcher
#   send work to workers. when workers are done 
#   they will reply and get more work 
#
# Parameters:
#   worker - subtest to send the next unit of work to
#
# Return: 
#  
# -----------------------------------------------
sub dispatcher {
    my $self=shift;
    my $worker=shift;
    my $work=$self->{'work'};
    if (my $w=shift(@$work)) {
        $main::Log->comment("dispatch work to subtest=".$worker->{'runid'});
        $main::Log->trace("work to dispatch=".Dumper($w));
        $worker->message_put(message=>$w,type=>"SUBTASK::RPC::COMMAND");
        #push($self->@CLEANUP, $w);
        $worker->message_get(
            type=>"SUBTASK::RPC::REPLY", 
            callback=> sub { $self->dispatcher($worker) } 
        );
        return 1;
    } 
    # out of work
    $worker->message_put(message=>{cmd=>'_exit'},type=>"SUBTASK::RPC::COMMAND");
    return 0;
};

# -----------------------------------------------
# Function: worker
#   call "method" with "args"
#
# Parameters:
#   id - unique identifier to use to name the worker 
#
# Return: 
#  
# -----------------------------------------------
sub _worker {
    my $self=shift;
    my $id=shift;
    $main::Log->comment("worker: $id initializing");
    while (1) {
        # get some work
        my $msg = $Parent->message_get( type=>"SUBTASK::RPC::COMMAND" );
        if (defined($msg->{'cmd'}) && $msg->{'cmd'} eq "_exit") { last; }
        $main::Log->trace("_worker $id: working on ".Dumper($msg));
            {
                no strict 'refs';
                &{'main::'.$msg->{'method'}}($msg->{'args'});
            }
        # send done message
        $Parent->message_put(message=>$id, type=>"SUBTASK::RPC::REPLY");
    }
    $main::Log->comment("worker $id exiting");
    return;
}

1;

=head1 FUNCTIONS



=head2 new

    constructor

=head3 Return

    blessed reference

=head2 start

    dispatch workers to do work

=head2 wait

    wait for workers to complete 

=head2 dispatcher

    send work to workers. when workers are done 
    they will reply and get more work 

=head2 worker

    call "method" with "args"


=head1 SEE ALSO 
