# -*- perl -*-

## ---------------------------------------------------
## Copyright (c) 2008 NetApp, Inc., All Rights Reserved
## Permission is granted to use this script on a NetApp Product.
## Any other use, modification, or distribution is prohibited
## without prior written consent from NetApp, Inc.
## ---------------------------------------------------

package UtilsIpv6;

## ---------------------------------------------------
## @summary
##  Routines for IPv6
##
## @description
##
## @usage
##  use nfs::NFSIPv6;
##
## @status      review
##
## @author      susheel@netapp.com
##
## @notes 
##  This works only with staf_ntest
##  Pre-requisite(s):
##   1) It must be possible to do a DNS reverse-lookup of
##   the IP addr that we pass to this routine 
##   This will be possible if we have a DNS entry (and 
##   the host is configured properly to resolve 
##   IP-addr <-> Host-name correctly)
##   OR 
##   there is an entry in the /etc/hosts file
##   2) The host-name thus obtained must match the name
##   of the host-spec. The global var ADDR_TYPE determines 
##   what addr to use - IPv4 or IPv6
##
##   Possibilities:
##   1) When ADDR_TYPE is not specified:
##   Default ADDR_TYPE is inet4
##   If we get an IPv4 address, we reverse-lookup to get
##   the host-name and then the host-spec. Return the 
##   hash with IPv4 info populated
##   If we get hostname, we try to get the hostspec and
##   return the IPv4 addr info from the host-spec
##   If we get an IPv6 addr, we return error
##
##   2) When ADDR_TYPE is inet4:
##   If we get an IPv4 address, the logic is the same
##   as metioned above
## 
##   3) When ADDR_TYPE is inet6:
##   If we get an IPv4 address, we try to get the 
##   hostspec by doing reverse-lookup and return the
##   IPv6 addr info from the hostspec. The same is
##   done if we get a hostname
##   If we get an IPv6 address, we return the hash with
##   IPv6 addr info
##   
## ---------------------------------------------------

use strict;
use Tharn;
use SPARSE::Util::Logger;
use Net::IP qw(ip_get_version ip_is_ipv4 ip_is_ipv6);
use Tharn qw($Log);
use Socket;

use vars qw(
  $VERSION
  @ISA
  @EXPORT_OK
  $ADDR_TYPE
);

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(
  &get_hostaddr
  &get_hostspec
);

param('ADDR_TYPE', '-default', 'inet4');

my $VERSION = '0.1';

## ---------------------------------------------------
## @name
##  _fill_addr_info
##
## ---------------------------------------------------
sub _fill_addr_info ($$%) {

    my ($hspec, $family, $h) = @_;

    # Based on the family type, populate the addr_h hash
    $h->{'hostname'} = $hspec->hostname();
    if ($family =~m /inet4/i) {
        # For IPv4
        $h->{'family'}   = 'inet4';
        $h->{'addr'}     = $hspec->default_ip();
        $h->{'nfs_mnt_addr'} = $hspec->default_ip();  
    } elsif ($family =~m /inet6/i) {
        # For IPv6
        $h->{'family'}   = 'inet6';
        $h->{'addr'}     = $hspec->net_best_address(sort_preference=>{family=>$h->{'family'}});
        # For nfs mounts
        $h->{'nfs_mnt_addr'} = '[' .
                           $h->{'addr'} .
                           ']';
        # TODO: For CIFS shares
    } else {
        # We will probably never come here!
        logDbg("UtilsIpv6: Add family $family not understood");
        return undef;
    }
    return 1;
}


## ---------------------------------------------------
## @name
##  get_hostspec
##
## ---------------------------------------------------
sub get_hostspec ($) {

    my ($h) = @_;

    my (@out, $h_spec);
    my ($name, $alias, $addrtype, $length, @addrs);

    if (ip_is_ipv4($h)) {
        # Try the IP addr as is
        eval {
            $h_spec = host($h);
        };
        if ($@) {
            # We got IPv4 address, do reverse-lookup
            ($name, $alias, $addrtype, $length, @addrs) =
                   gethostbyaddr(inet_aton($h), AF_INET);
        } else {
            # Getting host-spec with IPv4 addr works!
            return $h_spec; 
        }
    } else {
        # We got hostname
        $name = $h;    
    }

    if (defined($name)) {
        eval {
            $h_spec = host($name);
        };
        if ($@) {
            @out = split(/\./, $name);
            return host($out[0]);
        } else {
            return $h_spec;
        }
    } else {
        logDbg("get_hostspec: FAILED - DNS or /etc/hosts
                for $name probably not configured");
        return undef;
    }
    
}


## ---------------------------------------------------
## @name
##  get_hostaddr
##
## @description
##  Routine to get the IPv4 or IPv6 address of the 
##  host. For Filers, the param passed to the subroutine
##  can be an IP address or hostname.
##  For clients, the param must be a host-name
##
## @params
##  $addr - This can be a hostname or IP address
##
##  hashref to return the address info
##
## @return
##  On failure, returns undef
##  On success, this subroutine populates the 
##  hash addr_h with the addr info (IPv4/IPv6 address,
##  hostname, address family and inet6 address that 
##  includes the [] for mount) 
##  The address info depends on the param ADDR_TYPE 
##  passed on the ntest command line
##   
## @notes
##  This works only with staf_ntest
## ---------------------------------------------------
sub get_hostaddr ($%) {

    my ($host_id, $host_h) = @_;
    my ($host, $hspec);

    # Get hostspec     
    $hspec = get_hostspec($host_id);
    if (!defined($hspec)) {
        # Error msg will be thrown by get_hostname
        return undef;
    }

    if ($main::ADDR_TYPE =~m /inet4/) {
        # Either the default case or caller explicitly wants
        # to test with IPv4 addr
        if (ip_is_ipv4($host_id)) {
            return _fill_addr_info($hspec, 'inet4', $host_h);
        } elsif (ip_is_ipv6($host_id)) {
            # Param is an IPv6 address and ADDR_TYPE=inet4
            # This is an error
            logDbg("get_hostaddr: ERROR! ADDR_TYPE=$ADDR_TYPE but 
                    we got an IPv6 address");
            return undef;
        } else {
            return _fill_addr_info($hspec, 'inet4', $host_h);
        }
    } elsif ($main::ADDR_TYPE =~m /inet6/i) {
        # Caller explicitly wants to test with IPv6 addr
        # No checks requied, just return the addr as is
        return _fill_addr_info($hspec, 'inet6', $host_h);
    } else {
        # Caller has used illegal value for ADDR_TYPE
        logDbg("get_hostaddr: ADDR_TYPE=$main::ADDR_TYPE not valid");
        return undef;
    }
    return 1;
}
   

      
1;

