#-*-perl-*-

## @summary Routines for ipspace
##
## @description
##	ipspace functions for the vfiler test automation
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
##
## @dependencies Please use telnet to connect your filer, not rsh
##
## @keywords IPspace
##
## @status public
##
## @pod here
##

package VFILER::IPspace;

use strict;
use Tharn;
use SayWrap qw(sayEOF);
use base qw(Exporter);

BEGIN
{
	our @EXPORT = qw(
		assign_ipspace
		assign_default_ipspace
		create_ipspace
		destroy_ipspace
		exists_ipspace
		ifc_in_ipspace
	);
}

=head1 NAME

IPspace - Util routines for ipspace

=head1 SYNOPSIS

use VFILER::IPspace;

$ret = assign_ipspace($fc, $ipsname, $ifs);

$ret = assign_default_ipspace($fc, $ifs);

$ret = create_ipspace($fc, $ipsname);

$ret = destroy_ipspace($fc, $ipsname);

=head1 AUTHOR

Jae Bae

=head1 FUNCTIONS

All functions, unless otherwise stated, return a true value on success,
and false on failure. On failure C<$@> is set to the appropriate error
value via C<die()>.

=head2 assign_ipspace($fc, $ipsname, $ifs)

assign an ipspace

=over 4

=item fc

Required: hostspec of the filer

=item ipsname

Required: name of ipspace

=item ifs

Required: name of interface

=back

Examples:

	assign_ipspace($fc, 'ips1', 'e10a');

=cut

sub assign_ipspace($$$)
{
	my ($fc, $ipsname, $ifs) = @_;

	return eval {
		my (undef, undef, $after) = say($fc, "ipspace list", -swallow);
		if ($after =~ /$ipsname/) {
			(undef, undef, $after) =
				say($fc, "ipspace assign $ipsname $ifs", -swallow);
			if ($after !~ /interface\s+\"$ifs\"\s+does\s+not\s+exist./){
				if ($after =~ /interface $ifs has IP addresses/) {
					sayEOF($fc, "ifconfig $ifs 0.0.0.0 down");
					sayEOF($fc, "ipspace assign $ipsname $ifs");
				}
				return(1);
			} else {
				die($after);
			}
		} else {
			die($after);
		}
	};
}

=head2 assign_default_ipspace($fc, $ifs)

assign a default ipspace

=over 4

=item fc

Required: hostspec of the filer

=item ifs

Required: name of interface

=back

Examples:

	assign_default_ipspace($fc, 'e10a');

=cut

sub assign_default_ipspace($$)
{
	my ($fc, $ifs) = @_;

	return eval {
		my $ipsname = "default-ipspace";

		my (undef, undef, $after) = say($fc, "ipspace list", -swallow);
		if ($after =~ /$ipsname/) {
			say($fc, "ipspace assign $ipsname $ifs");
			return(1);
		} else {
			die($after);
		}
	};
}

=head2 create_ipspace($fc, $ipsname)

create an ipspace

=over 4

=item fc

Required: hostspec of the filer

=item ipsname

Required: name of ipspace

=back

Examples:

	create_ipspace($fc, 'ips1');

=cut

sub create_ipspace($$)
{
	my ($fc, $ipsname) = @_;

	return eval {
		my (undef, undef, $after) = say($fc, "ipspace list", -swallow);
		if ($after =~ /$ipsname/) {
			die($after);
		} else {
			(undef, undef, $after) = say($fc, "ipspace create $ipsname", -swallow);
			if ($after =~ /Ipspace\s+\"$ipsname\"\s+created/) {
				print(1);
			} else {
				die($after);
			}
		}
	};
}

=head2 destroy_ipspace($fc, $ipsname)

destroy an ipspace

=over 4

=item fc

Required: hostspec of the filer

=item ipsname

Required: name of ipspace

=back

Examples:

	destroy_ipspace($fc, 'ips1');

=cut

sub destroy_ipspace($$)
{
	my ($fc, $ipsname) = @_;

	return eval {
		my (undef, undef, $after) = say($fc, "ipspace list", -swallow);
		if ($after =~ /$ipsname/) {
			(undef, undef, $after) = say($fc, "ipspace destroy $ipsname", -swallow);
			if ($after =~ /Ipspace\s+\"$ipsname\"\s+destroyed/) {
				print(1);
			} else {
				die($after);
			}
		} else {
			die($after);
		}
	};
}

=head2 exists_ipspace($fc, $ipsname)

return non-zero if specified ipsname exists on the filer

=over 4

=item fc

Required: hostspec of the filer

=item ipsname

Required: name of ipspace

=back

Examples:

	exists_ipspace($fc, 'ips1');

=cut

sub exists_ipspace($$)
{
	my ($fc, $ipsname) = @_;

	return eval {
		my (undef, undef, $after) = say($fc, "ipspace list", -swallow);
		if ($after =~ /$ipsname/) {
			return 1;
		} else {
			return 0;
		}
	};
}

=head2 ifc_in_ipspace($fc, $ipsname, $ifc)

return non-zero if specified interface is in specified ipsname, zero otherwise

=over 4

=item fc

Required: hostspec of the filer

=item ipsname

Required: name of ipspace

=item ifc

Required: name of interface

=back

Examples:

	ifc_in_ipspace($fc, 'ips1', 'ns0');

=cut

sub ifc_in_ipspace($$$)
{
	my ($fc, $ipsname, $ifc) = @_;

	return eval {
		my (undef, undef, $after) = say($fc, "ipspace list", -swallow);
		if ($after =~ /$ipsname/) {
			my @lines = split(/\n/, $after);
			foreach my $line (@lines) {
				next if ($line !~ /^$ipsname/);
				if ($line =~ /[\( ]$ifc[\) ]/) {
					return 1;
				} else {
					return 0;
				}
			}
		} else {
			die("specified ipspace does not exists: $after");
		}
	};
}

1;
