# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/VFILER/Setup.pm#1 $

## @summary Library of vfiler setup commands
## @author archana@netapp.com
## @status public
## @pod here


=head1 NAME

Setup - Set up multiple Vfilers

=head1 SYNOPSIS

   use VFILER::Setup qw(:all);

   my $h = host($FILER);
   my $fc = connect("#rsh");

   setup_vfilers_from_string($fc, $FILER_VFILER_LIST);

   setup_vfilers($fc,
                 {vfiler=>"vf1",ip="e1",...},
                 {vfiler=>"vf2",ip="e2",...},
                );

   vfiler_adjust_limit(filer => $fc, count => 2);

   vfiler_find_if(filer => $fc,
                  ip_addr => $vfiler_ip_address,
                  if => $interface_name,
                  ip_exclude => $exclude_ip_address);

   vfiler_cifs_setup(filer => 'filer_name',
                     dns_domainname => $dns,
                     dns_server_list => "$server1, $server2",
                     domain => $domain,
                     domain_user => $user,
                     domain_pass => $pass,
                     mode => 'Multi',
                     vfiler => $vfname);

=head1 ROUTINES

=cut


package VFILER::Setup;
use strict;
use base qw(Exporter);
BEGIN
{
  our %EXPORT_TAGS = (
    all => [ qw(
                setup_vfilers
                setup_vfilers_from_string

                vfiler_list
                vfiler_present
                vfiler_status
                vfiler_create
                vfiler_purge
                vfiler_stop
                vfiler_destroy
                vfiler_adjust_limit

                vfiler_find_if
                vfiler_cifs_setup
                )
           ] );
  Exporter::export_ok_tags('all');
}

use Tharn qw(connect host tstlog logcomment);
use SayWrap qw(sayEOF);
use Utils qw(make_host make_conn);
use Carp qw(croak);
use Params::Validate v0.70 qw(validate validate_pos SCALAR OBJECT HASHREF
                              UNDEF);
use Module::Common qw(validate_mixed);
use QtreeCmds qw(qtreePresent);
use Qtree::Setup qw(setup_qtrees);
use Volume qw(vol_exists vol_root);
use Network::Utils qw(get_ifconfig_info);
use CIFS::Setup qw(cifs_setup);
use VFILER::Vfiler qw(vfiler_limit vfiler_limit_hard_limit vfiler_limit_in_use);
use boot qw(reboot);

# regex for vfilers
my $VFILER_REGEX = qr{[\w_-]+};

=head2 setup_vfilers_from_string

As C<setup_vfilers>, but takes a string description instead of an
array of hashes.  This is intended to be used when the description of
vfilers to (re)create comes from an external source, such as as a
parameter to a test.

Arguments

=over

=item conn

(Required) : connection to filer

=item string

(Required) : A space separated list of vfiler descriptions. Each
vfiler description is a comma separated list of
I<attribute>=I<value> pairs.  For example:

    "vfiler=tada-u3-vf1,ip=e2,ipaddr=172.16.74.103,path=/vol/catvol/qtree_unix"

Refer to C<setup_vfilers> for possible attributes and more details

=back

=cut

sub setup_vfilers_from_string ( $$ ) {

    my ($conn, $vfiler_list) = @_;

    my @list;
    # extract attributes and create hash
    my @vfilers = split(/\s+/, $vfiler_list);
    foreach my $vfiler (@vfilers) {
        my $args = {};
        my @attributes = split(/,/, $vfiler);
        foreach my $attr (@attributes) {
            my ($key, $value) = ($attr =~ /(.*?)=(.*)/);
            $args->{$key} = $value;
        }
        push (@list, $args);
    }
    setup_vfilers($conn, @list);

    return 1;
}


=head2  setup_vfilers

Create several vfilers.

Returns

1 on success

dies otherwise.


Arguments

=over

=item conn

(Required) : connection to filer

=item vfiler descriptions

(Required) : Each argument after the connection is a hash reference
describing a particular vfiler to be created.

Possible elements are:

=over

=item vfiler

(Required) : Name of vfiler to be created.

=item path

(Required) : Path of vfiler
If it does not exist then will be created.

=item ip

(Required) : Vfiler IP

=item ipaddr

(Required) : Vfilers' IP Address

=item netmask

(Optional) : Vfilers' IP netmask

=item gateway

(Optional) : Vfilers' gateway, typically specified when non-default
ipspace is used.

=item ipspace

(Optional) : Vfilers' ipspace

=item qt_security

(Optional) : security style for the qtree if its to be created.

=item qt_oplocks

(Optional) : oplocks for the qtree if its to be created.

=item purge_existing

(Optional) : flag to purge existing vfiler.

=back

Example

   setup_vfilers($conn,
     {vfiler=>"vf1",ip=>"e2",ipaddr=>"172.16.74.113",path="/vol/catvol/qtree_mixed"}
     {vfiler=>"vf2",ip=>"e2",ipaddr=>"172.16.74.123",path="/vol/catvol/qtree_ntfs"}
    );

=back

=cut

sub setup_vfilers ( $@ ) {

    my ($conn, @list) = @_;

    foreach my $item (@list) {
        $item->{conn} = $conn;
        _setup(%$item);
        croak("$@\n") if ($@);
    }

}

sub _setup ( % ) {
    my (%args) = validate (@_, {
                 conn => { type=>OBJECT|HASHREF, isa=>'Connectrec' },
                 vfiler => { type=>SCALAR, regex=>qr/^\S+$/, optional=>0,
                            depends=>['path','ip','ipaddr'] },
                 path => { type=>SCALAR, regex=>qr/^(\/vol\/)?\w+(\/\w+)?/, optional=>0 },
                 ip => { type=>SCALAR, regex=>qr/^\w+$/, optional=>0 },
                 ipaddr => { type=>SCALAR, regex=>qr/^\d+.\d+.\d+.\d+$/, optional=>0 },
                 netmask => { type=>SCALAR, regex=>qr/^\d+.\d+.\d+.\d+$/, optional=>1 },
                 gateway => { type=>SCALAR, optional=>1 },
                 ipspace => { type=>SCALAR, optional=>1 },
                 qt_security => { type=>SCALAR, regex=>qr/^unix|ntfs|mixed$/, optional=>1 },
                 qt_oplocks => { type=>SCALAR, regex=>qr/^enable|disable$/, optional=>1 },
                 purge_existing => { type=>SCALAR, optional=>1, regex=>qr/^(0|1)$/ },
                }
            );

    if (vfiler_present($args{conn}, $args{vfiler}) ) {
        if (!$args{purge_existing}) {
            logcomment("Vfiler '$args{vfiler}' already exists");
            return(1);
        } else {
            vfiler_purge($args{conn}, $args{vfiler});
        }
    }
    my @opts;
    push(@opts, (netmask=>$args{netmask})) if $args{netmask};
    push(@opts, (gateway=>$args{gateway})) if $args{gateway};
    push(@opts, (ipspace=>$args{ipspace})) if $args{ipspace};
    push(@opts, (qt_security=>$args{qt_security})) if $args{qt_security};
    push(@opts, (qt_oplocks=>$args{qt_oplocks})) if $args{qt_oplocks};
    vfiler_create($args{conn}, $args{vfiler},
                $args{path}, $args{ip},
                $args{ipaddr}, @opts);

    return (1);
}

# LOCAL ROUTINES


=head2 vfiler_create

Arguments

=over

=item conn

(Required) : connection to filer

=back

Returns

Notes

Will not work if $path0 is a volume. This routine can croak and the vfiler
could still have been created.

=cut

sub vfiler_create ( % ) {

    my ($filer, $vfiler, $path0, $ip0, $ipaddr0, %args) = validate_mixed(
                    @_,
                    [ { type => OBJECT|HASHREF, optional=>0 },
                      { type => SCALAR, optional=>0, regex=>qr/^\S+$/ },
                      { type => SCALAR, optional=>0, regex=>qr/^(\/vol\/)?\w+(\/\w+)?$/ },
                      { type => SCALAR, optional=>0, regex=>qr/^\w+$/ },
                      { type => SCALAR, optional=>0, regex=>qr/^\d+.\d+.\d+.\d+$/ }
                    ],
                    {
                     gateway => { type=>SCALAR, optional=>1 },
                     ipspace => { type=>SCALAR, optional=>1 },
                     netmask => { type=>SCALAR, optional=>1, default=>"255.255.252.0" },
                     qt_security => { type=>SCALAR, regex=>qr/^unix|ntfs|mixed$/, optional=>1 },
                     qt_oplocks => { type=>SCALAR, regex=>qr/^enable|disable$/, optional=>1 },
                     #more_ip => { type=>SCALAR, optional=>1 },
                     #more_path => { type=>SCALAR, optional=>1, depends=>['more_ip'] },
                    } );

    my $h = make_host($filer);
    my $conn = make_conn($filer);
    my $filer_id = $h->id();

    if ($path0 =~ /^\/vol\/(\w+)\/\w+$/) {
        # create qtree if it does not exist
        if (qtreePresent($conn, $path0)) {
            tstlog('-commline', "Qtree '$path0' exists on $filer_id");
        } else {
            # check if volume specified in qtree path exists on filer
            my ($vol) = $1;
            if (!vol_exists($conn, $vol)) {
                croak("Unable to create qtree $path0 since volume '$vol' ".
                        "does not exist on filer $filer_id\n");
            } else {
                logcomment("Creating qtree $path0 on $filer_id using volume $vol");
                my $qt_args = {qtree=>$path0};
                $qt_args->{security} = $args{qt_security} if ($args{qt_security});
                $qt_args->{oplocks} = $args{qt_oplocks} if ($args{qt_oplocks});
                setup_qtrees($conn, $qt_args);
                croak("$@\n") if ($@);
            }
        }
    } elsif ($path0 =~ /^\/vol\/(\w+)\$/) {
        tstlog('-commline', "Assuming $path0 to be a volume");
        my ($vol) = $1;
        if (!vol_exists($conn, $vol)) {
            croak("Volume '$vol' does not exist on filer $filer_id\n");
        }
    } elsif ($path0 =~ /^\w+$/) {
        tstlog('-commline', "Qtree $path0 must exist in root volume");
        my $root_vol = vol_root($conn);
        my $path = "/vol/$root_vol/$path0";
        if (qtreePresent($conn, $path)) {
            tstlog('-commline', "Qtree '$path' exists on $filer_id");
        } else {
            logcomment("Creating qtree $path on $filer_id using volume $root_vol");
            my $qt_args = {qtree=>$path0};
            $qt_args->{security} = $args{qt_security} if ($args{qt_security});
            $qt_args->{oplocks} = $args{qt_oplocks} if ($args{qt_oplocks});
            setup_qtrees($conn, $qt_args);
            croak("$@\n") if ($@);
        }
    }

    # create the vfiler
    my $cmd = "vfiler create $vfiler -n ";
    $cmd .= " -s $args{ipspace}" if $args{ipspace};
    $cmd .= " -i $ipaddr0";
    # XXX add stuff to create using additional ipaddr
    $cmd .= " $path0";
    # XXX add stuff to create using additional paths

    my $out = sayEOF($conn, $cmd);
    croak("Vfiler $vfiler creation failed: $out")
      if ($out =~ /Vfiler $vfiler not created/i);
    # Verify with vfiler status below

    # verify if the create succeeded
    my %status = vfiler_status($conn, $vfiler);
    croak("Vfiler $vfiler creation failed")
        if ($status{$vfiler}{status} ne 'running');

    # If the network interface on the filer is unconfigured and ipspaces
    # was specified then configure it to the vfiler IP address.
    my $ifconfig_info = get_ifconfig_info($conn);
    my $if_info = $ifconfig_info->{$ip0};
    if ($if_info->{flags} !~ /UP/i) {
        # configure the IP address alias
        $cmd  = "ifconfig $ip0 $ipaddr0";
        $cmd .= " netmask $args{netmask} up";
        $out  = sayEOF($conn, $cmd);
        # Verify with vfiler status below
    }

    # If the vfiler IP address is unconfigured then setup the IP alias.
    if ($status{$vfiler}{ip0} eq 'unconfigured') {
        # configure the IP address alias
        $cmd  = "ifconfig $ip0 alias $ipaddr0";
        $cmd .= " netmask $args{netmask}";
        $out  = sayEOF($conn, $cmd);
        # Verify with vfiler status below
    }

    # Set the vfilers default route if a gateway was specified
    if (defined($args{gateway})) {
        $out = sayEOF($conn,
                      "vfiler run $vfiler route add default $args{gateway} 1");
        croak("Could not set default route on vfiler\n")
            if ($out =~ /network\s+unreachable/i);
    }

    %status = vfiler_status($conn, $vfiler);
    croak("Vfiler $vfiler status is not 'running'")
        if ($status{$vfiler}{status} ne 'running');
    croak("Vfiler $vfiler IP address is not '$ipaddr0'")
        if ($status{$vfiler}{ipaddr0} ne $ipaddr0);
    croak("Vfiler $vfiler IP is not '$ip0'")
        if ($status{$vfiler}{ip0} ne $ip0);
    croak("Vfiler $vfiler path is not '$path0'")
        if ($status{$vfiler}{path0} ne $path0);

    logcomment("Vfiler '$vfiler' creation succeeded");
    return(1);
}


=head2 vfiler_status($filer,$vfiler)

Arguments

=over

=item filer

(Required) : Physical filer

=item vfiler

(Required) : vfiler whose status is required

=back

Notes

Update routine to give status of all vfilers on filer

=cut

sub vfiler_status ( $$ ) {

    my ($filer, $vfiler) = validate_pos(
                    @_,
                    { type => SCALAR|OBJECT|HASHREF, optional=>0 },
                    { type => SCALAR, optional=>0, regex=>qr/^\S+$/ }
                    );

    my $conn = make_conn($filer);

    # check status of vfiler
    my $out = sayEOF($conn, "vfiler status -a $vfiler");

    croak("Vfiler (multistore) support not licensed\n")
        if ($out =~ /vfilers not licensed|Use of vfilers requires a multistore license/i);

    croak("$out\n") if ($out =~ /vfiler not found/i);

    #need to remove lines containing debug/warning messages from filer console
    #burt: 439710
    $out =~ s/.+:(debug|warning|notice)\]:.+\s+//ig ;

    my @lines = split(/\n/, $out);
    croak ("Insufficient output : $out\n")
        if (scalar(@lines) < 13);

    my %status;
    # XXX make this verification easier put the expected values in a hash
    # verify line by line
    my $line = shift(@lines);
    croak("Unexpected line : $line\n")
        if ($line !~ /^$vfiler\s+(\S+)$/);
    $status{$vfiler}{status} = $1;

    $line = shift(@lines);
    croak("Unexpected line : $line\n")
        if ($line !~ /^\s+ipspace:\s+(\S+)$/);
    $status{$vfiler}{ipspace} = $1;

    # process all IP addresses
    $line = shift(@lines);
    croak("Unexpected line : $line\n")
        if ($line !~ /^\s+IP address:\s+(\S+)\s+\[(\w+)\]$/);
    my $idx = 0;
    do {
        $status{$vfiler}{"ipaddr$idx"} = $1;
        $status{$vfiler}{"ip$idx"} = $2;
        $line = shift(@lines);
    } while($line =~ /^\s+IP address:\s+(\S+)\s+\[(\w+)\]$/);

    # XXX have to do this for all paths
    croak("Unexpected line : $line\n")
        if ($line !~ /^\s+Path:\s+([\/\w]+)\s+\[.*\]$/);
    $status{$vfiler}{path0} = $1;

    # XXX currently we dont care about the rest
    # code it later

    #cranford% rsh -l root: tada-u23 "vfiler status -a vf2"
    #vf2                              running
    #   ipspace: default-ipspace
    #   IP address: 172.16.74.124 [e1a]
    #   Path: /vol/vol3 [/etc]
    #   UUID: f6767b50-9bc1-11d8-ba38-000000000000
    #   Protocols allowed: 6
    #   Allowed: proto=rsh
    #   Allowed: proto=nfs
    #   Allowed: proto=cifs
    #   Allowed: proto=iscsi
    #   Allowed: proto=ftp
    #   Allowed: proto=http
    #   Protocols disallowed: 0
    #cranford%

    return(%status);

}

=head2 vfiler_present($filer,$vfiler)

Arguments

=over

=item filer

(Required) : Physical filer

=item vfiler

(Required) : vfiler name to verified if present or not

=back

Returns

1 if vfiler is present

0 if vfiler is absent

on failure dies with error in $@

=cut

sub vfiler_present ( $$ ) {

    my ($filer, $vfiler) = validate_pos(
                    @_,
                    { type => SCALAR|OBJECT|HASHREF, optional=>0 },
                    { type => SCALAR, optional=>0, regex=>qr/^\S+$/ }
                    );

    my $conn = make_conn($filer);

    my $out = sayEOF($conn, "vfiler status -a $vfiler");
    croak("Vfiler (multistore) support not licensed\n")
        if ($out =~ /vfilers not licensed|Use of vfilers requires a multistore license/i);

    return ($out =~ /vfiler not found/i)  ? 0 : 1;

}


=head2 vfiler_purge($filer,$vfiler,%opts)

Arguments

=over

=item filer

(Required) : Physical filer

=item vfiler

(Required) : vfiler to purge

=item die_not_existing

(Required) : flag denoting if routine must fail (die) when asked to destroy
a non-existing vfiler

=back

=cut

sub vfiler_purge ( $$% ) {

    my ($filer, $vfiler, %args) = validate_mixed(
                    @_,
                    [ { type => SCALAR|OBJECT|HASHREF, optional=>0 },
                      { type => SCALAR, optional=>0, regex=>qr/^\S+$/ }
                    ],
                    {
                        verify => { type=>SCALAR, optional=>1, default=>1, regex=>qr/^(0|1)$/ },
                        die_not_existing => { type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/ }
                    } );

    logcomment("Purging vfiler '$vfiler'");

    vfiler_stop($filer, $vfiler, die_not_existing=>$args{die_not_existing});
    croak($@) if ($@);

    vfiler_destroy($filer, $vfiler, die_not_existing=>$args{die_not_existing});
    croak($@) if ($@);

    if ($args{verify} && vfiler_present($filer, $vfiler) ) {
            croak("Could not purge vfiler '$vfiler'\n");
    }
    return(1);
}


=head2 vfiler_stop($filer,$vfiler,%opts)

Arguments

=over

=item filer

(Required) : Physical filer

=item vfiler

(Required) : vfiler to stop

=item die_not_existing

(Required) : flag denoting if routine must fail (die) when asked to destroy
a non-existing vfiler

=back

=cut

sub vfiler_stop ( $$% ) {

    my ($filer, $vfiler, %args) = validate_mixed(
                    @_,
                    [ { type => SCALAR|OBJECT|HASHREF, optional=>0 },
                      { type => SCALAR, optional=>0, regex=>qr/^\S+$/ }
                    ],
                    {
                        die_not_existing => { type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/ }
                    } );

    eval {
    my $conn = make_conn($filer);

    my $out = sayEOF($conn, "vfiler stop $vfiler");
    croak("Vfiler (multistore) support not licensed\n")
        if ($out =~ /vfilers not licensed|Use of vfilers requires a multistore license/i);

    if ($out =~ /does not exist/i) {
        if ($args{die_not_existing}) {
            croak("Vfiler '$vfiler' does not exist");
        } else {
            return(1);
        }
    }

    if ($out =~ /^(?:Vfiler\s+)?$vfiler\s+(?:already\s+)?stopped\.?$/ ) {
        return(1);
    } else {
        croak("Could not stop vfiler '$vfiler'. Got $out\n");
    }
    };
}

=head2 vfiler_destroy($filer,$vfiler,%args)

Arguments

=over

=item filer

(Required) : Physical filer

=item vfiler

(Required) : vfiler to destroy

=item die_not_existing

(Required) : flag denoting if routine must fail (die) when asked to destroy
a non-existing vfiler

=back

=cut

sub vfiler_destroy ( $$% ) {

    my ($filer, $vfiler, %args) = validate_mixed(
                    @_,
                    [ { type => SCALAR|OBJECT|HASHREF, optional=>0 },
                      { type => SCALAR, optional=>0, regex=>qr/^\S+$/ }
                    ],
                    {
                        die_not_existing => { type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/ }
                    } );
    my $conn = make_conn($filer);

    my $out = sayEOF($conn, "vfiler destroy $vfiler -f");
    croak("Vfiler (multistore) support not licensed\n")
        if ($out =~ /vfilers not licensed|Use of vfilers requires a multistore license/i);

    if ($out =~ /does not exist/i) {
        if ($args{die_not_existing}) {
            croak("Vfiler '$vfiler' does not exist");
        } else {
            return(1);
        }
    }

    # this output contains lots of stuff - eg. shutting down cifs etc.
    # so will make the checking not very strict
    if ($out =~ /Resources for vfiler $vfiler moved to hosting filer\./ ) {
        return(1);
    } else {
        croak("Could not destroy vfiler '$vfiler'\n");
    }
}


=head2 vfiler_list($filer)

Arguments

=over

=item filer

(Required) : Physical filer

=back

=cut

sub vfiler_list ( $ ) {

    my ($filer) = validate_pos( @_,
            { type=>SCALAR|OBJECT|HASHREF }
            );

    eval {
    my $conn = make_conn($filer);
    my $out = sayEOF($conn, "vfiler status");

    my @vfilers;
    foreach my $line (split(/\n/,$out)) {
        croak("Unexpected output format : $line")
            if ($line !~ /^($VFILER_REGEX)\s+(?:(?:[a-z]+,?\s?)+)$/);
        push(@vfilers, $1);
    }

    return(@vfilers);
    };
}

=head2 vfiler_adjust_limit(%args)

Checks vfiler limits on a filer and determines if enough unused
slots are available for the specified number of vfilers. If there
are not enough slots available, the vfiler limit on the filer is
adjusted if the new limit will be under the filer's hard limit.

A filer reboot is required to enable the new limit.

Arguments

=over

=item filer

(Required) : Physical filer

=item count

(Required) : Number of vfilers that we need room for

=back

=cut

sub vfiler_adjust_limit ( % ) {

    my %opts = validate(
        @_,
        {
            filer  => { type => SCALAR|OBJECT|HASHREF },
            count  => { type => SCALAR, regex=>qr/^\d+$/ },
        }
    );
    my ($avail, $hard_limit, $inuse, $limit, $new_limit, $text);
    my $conn = make_conn($opts{filer});

    $limit = vfiler_limit($conn)
      or croak "Could not determine current vfiler limit: $@";
    $hard_limit = vfiler_limit_hard_limit($conn)
      or croak "Could not determine vfiler hard limit: $@";
    $inuse = vfiler_limit_in_use($conn)
      or croak "Could not determine vfiler in use count: $@";

    $avail = $limit - $inuse;
    if ($avail < $opts{count}) {
        $new_limit = $opts{count} + $inuse;
        if ($new_limit > $hard_limit) {
            croak "Cannot make room for vfilers: hard limit: $hard_limit " .
                  "Current limit: $limit Inuse: $inuse Requested: $opts{count}";
        }
        # We use sayEOF here instead of VFILER::Vfiler::vfiler_limit_set
        # so that we can parse the output and determine if we need to
        # reboot or not.
        $text = sayEOF($conn, "vfiler limit $new_limit");
        croak "Output from 'vfiler limit $new_limit' did not indicate " .
          "vfiler limit was correctly changed: $text"
            if ($text !~ /maxvfilers is changed to $new_limit/);
        $text =~ s/\r*\n/ /g;
        if ($text =~ /You must reboot .* to take effect/) {
            reboot($conn->hostp());
            croak "Could not reboot for vfiler limit to take effect: $@" if $@;
            if (($limit = vfiler_limit($conn)) < $new_limit) {
                croak "Vfiler limit is not set correctly after reboot: " .
                  "expected at least $new_limit, recieved $limit";
            }
        }
    }
    return;
}

=head2 vfiler_find_if(%args)

Given a filer connection and an IP address that will be used for a
vfiler, finds an interface that can be used for vfiler setup. Returns
the interface name and the netmask currently assigned to that interface.

This function also allows for a specific interface to be named, and will
verify if the named interface can accept an alias of the requested
IP address.

An IP address to be excluded may also be specified, which will cause
the search to ignore an interface that is assigned to that address. This
is useful for determining interfaces available for secondary ip spaces.

The following named arguments are accepted:

=over

=item filer

(Required) : Connectrec or hostrec of filer.

=item ip_addr

(Required) : IP address that is going to be assigned to the vfiler

=item if

(Optional) : Specific interface name to match against. If specified, the
IP address will be matched against the named interface to validate it
can be aliased to that interface.

=item ip_exclude

(Optional) : An IP address which will be used to exclude interfaces. An
interface that is using that IP address will not be used as a valid
interface for the requested IP address.

=back

Returns

An array containing the selected interface name and the netmask
for that interface, values are 'undef' if no suitable interface
was found, or croaks with an error if a specified interface is not
found or is not usable by the requested IP address.

=cut

sub vfiler_find_if ( % ) {
    my %opts = validate(
        @_,
        {
            filer      => { type => SCALAR|OBJECT|HASHREF },
            ip_addr    => { type => SCALAR, regex=>qr/^\d+.\d+.\d+.\d+$/ },
            if         => { type => SCALAR|UNDEF, optional => 1 },
            ip_exclude => { type => SCALAR, regex=>qr/^\d+.\d+.\d+.\d+$/,
                            optional => 1 },
        }
    );

    # Make sure we have a connection to the filer and get the
    # interface information
    my $conn = make_conn($opts{filer});
    my $ifconfig_info = get_ifconfig_info($conn)
      or croak "Could not retrieve ifconfig info: $@";

    my $ip_addr = $opts{ip_addr};
    my $if_ip;
    my $if_netmask;

    # If an interface was specified, make sure it exists on the filer
    # and is active
    if ($opts{if}) {
        croak "Specified interface $opts{if} does not exist on filer"
          if (!exists $ifconfig_info->{$opts{if}});
        croak "Specified interface $opts{if} is not active"
          if ($ifconfig_info->{$opts{if}}{flags} !~ /UP/);
    }

    # Set the interfaces to check to either the specified interface or
    # to the interfaces returned by get_ifconfig_info.
    my @ifs = ($opts{if}) ? $opts{if} : keys %$ifconfig_info;
    
    foreach my $if (@ifs) {

        # Skip the loopback interface
        next if $if eq 'lo';
        
        # Skip the interface if it is not active (UP)
        my $if_info = $ifconfig_info->{$if};
        
	next if ($if_info->{flags} !~ /UP/);

        # Skip the interface if its IP address is the excluded IP
        next if ($opts{ip_exclude} && $if_info->{ip_addr} eq $opts{ip_exclude});

        # Get the current IP address and netmask from the interface
        $if_ip = $if_info->{ip_addr};
	next if !$if_ip;
        my $ip = pack('C4', split(/\./, $if_ip));
        my $netmask = $if_info->{netmask};
        $netmask =~ s/0x//;
        $netmask = pack('H8', $netmask);
        $if_netmask = Socket::inet_ntoa($netmask);

        # Convert the vfiler IP address into a bit string
        my $vip = pack('C4', split(/\./, $ip_addr));
        
        # Compare the IP addresses and-ed with the netmask
        # They should be equal if the IP can be used on the interface
        if (Socket::inet_ntoa($ip & $netmask) eq
            Socket::inet_ntoa($vip & $netmask)) {
            # Return interface name and netmask if the interface
            # will work with the given vfiler ip address
            return $if, $if_netmask;
        }
    }

    # If no active interface on the correct subnet is found, croak
    # with error if an interface was specified, else return undef
    if ($opts{if}) {
        croak "IP address $opts{ip_addr} does not match the netmask " .
          "<$if_netmask> and ip address <$if_ip> on interface $opts{if}";
    }
    return undef, undef;
}

#---------------------------------------------------------------------

=head2 vfiler_cifs_setup

    vfiler_cifs_setup(mode =>$mode);

This function is a wrapper function that calls cifs_setup to setup CIFS on the
filer with the specified mode. This wrapper is used to isolate the details of
setting up in a domain environment or on WORKGROUP.

Will setup cifs in a domain environment or under WORKGROUP. If domain
is an empty string will setup under WORKGROUP otherwise will attempt to
setup under a domain using passed in or default values. Croaks if their
is an error in call to cifs_setup.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filer

(Required) : Name of filer to connect to.

=item dns_domainname

(Optional) : Filer's domain name. Must be defined if 'domain' is defined.

=item dns_server_list

(Optional) : Comma separted list of DNS server IP(s).
Must be defined if domain is defined.

=item domain

(Required) : Domain name to add filer to, used in cifs_setup. Allows a
domain name with just spaces for testing purposes.

=item domain_user

(Required) : Domain user that administers filer in domain,used in cifs_setup.

=item domain_pass

(Required) : Domain user password,used in cifs_setup.

=item mode

(Optional) : Protocol mode to setup cifs. Either 'Multi' protocol or 'Ntfs'
only. Defaults to 'Multi' protocol.

=item vfiler

(Required) : Name of vfiler running on vfiler to perform the cifs setup on.

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub vfiler_cifs_setup(%) {
    my %opts = validate(
        @_,
        {
            dns_domainname  => {type=>SCALAR, optional=>1},
            dns_server_list => {type=>SCALAR, optional=>1,
                                regex=>qr/(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/},
            domain          => { type => SCALAR },
            domain_user     => { type => SCALAR },
            domain_pass     => { type => SCALAR },
            filer           => { type => SCALAR },
            mode            => { type => SCALAR, default => 'multi',
                                 regex => qr/^(ntfs|multi)$/ },
            vfiler          => { type => SCALAR },
        }
    );

    my (%args, $fc, $fh, $err_msg, @proto);

    # Determine whether to use the domain or the workgroup
    $opts{domain} =~ s/\s+//g;
    if ($opts{domain} ne '') {
        %args = (dns_enable      => 1,
                 dns_domainname  => $opts{dns_domainname},
                 dns_server_list => $opts{dns_server_list},
                 domain          => $opts{domain},
                 domain_user     => $opts{domain_user},
                 domain_pass     => $opts{domain_pass});
    } else {
        %args = (workgroup => 'WORKGROUP');
    }

    # If the protocol requested is ntfs then cifs_setup has to be run
    # twice, the first time to multi then the second time to ntfs. This
    # guarantees that the filer is in NTFS-only mode.
    if ($opts{mode} eq 'ntfs') {
        @proto = ( 'multi', 'ntfs' );
    } else {
        @proto = ( $opts{mode} );
    }
    foreach my $p (@proto) {
        my $security_style = ($p eq 'multi') ? 'multiprotocol' : $p;
        cifs_setup(cifs_pass        => $main::CIFS_PASSWD,
                   host             => $opts{filer},
                   passwd_group     => 'yes',
                   'security-style' => $security_style,
                   vfiler           => $opts{vfiler},
                   %args);
        if ($@) {
            # save the cifs_setup error
            $err_msg = $@;

            # cifs_setup should cleanup after itself but it can leave
            # the command 'cifs setup' running and prompting for input.
            # When 'cifs setup' gets converted to ZAPI this problem should
            # go away. For now grab the telnet connection to send a ^C to
            # the filer.
            eval {
                $fh = host($opts{filer});
                $fc = connect("#telnet", -timeout, $main::TIMEOUT);
                $fc->timeout($main::CMDTIMEOUT);
                $fc->send_intr();
            };
            croak("vfiler_cifs_setup: cifs setup failed $err_msg");
        }
    }
    return;
}

1;
