#-*-perl-*-

## @summary Routines for Util
##
## @description
##	util functions for the vfiler test automation
##
## @dependencies Please use telnet to connect your filer, not rsh
##
## @author dl-johny@netapp.com
##         (FSS WAFL Quality)
##
## @keywords Util
##
## @status public
##
## @pod here
##

package VFILER::Util;

use strict;
use Tharn;
use base qw(Exporter);
use SPARSE::Util::Logger;
use SayWrap qw(sayEOF);
use Admin::License qw(is_licensed);
use Filer::Version qw(are_we_ontapng);
use VFILER::IPspace qw(create_ipspace assign_ipspace exists_ipspace
		ifc_in_ipspace);

BEGIN
{
	our @EXPORT = qw(
		add_wins_dns_host
		check_ipspace
		check_vfiler
		clean_vfiler_and_resources
		create_vfiler_non_default_ipspace
		create_passwd_file
		get_vfiler_root_path
		get_vfiler_status_in_dr
		get_vfiler_storage_paths
		host_connect
		sync
		mk_file
		rm_file
		set_timezone
		setup_cifs
		setup_dns
		setup_nis
		vfiler_cleanup
		new_vfiler_resources
		create_vfiler_and_setup
	);
}

use VFILER::IPspace qw(assign_default_ipspace destroy_ipspace);
use VFILER::Vfiler qw(
	vfiler_create
	vfiler_create_non_default_ipspace
	vfiler_limit_in_use
	vfiler_run_vfiler_add_default_route
	vfiler_run_vfiler_exportfs
	vfiler_run_vfiler_setup
	vfiler_run_vfiler_cifs_setup
	vfiler_stop
	vfiler_stop_destroy
	);
use Volume qw(vol_exists vol_status vol_offline vol_destroy);

=head1 NAME

Util - Util routines for vFiler

=head1 SYNOPSIS

use VFILER::Util;

add_wins_dns_host($fc, $WINS_DNS_SERVER, $WINS_DNS_DOMAIN);

$ret = check_ipspace($fc, $ipsname);

$ret = check_vfiler($fc, $vfname);

clean_vfiler_and_resources($hostspec);

$ret = create_vfiler_non_default_ipspace($fc, $vfname, $ipsname, %args);

$ret = host_connect($hostspec, $connspec, $timeout, $cmdtimeout);

$ret = create_passwd_file($fc, $passwd_file);

$ret = get_vfiler_status_in_dr($fc, $vfname);

$ret = sync($timeout);

$ret = mk_file($fc, $size, $filename, [%args]);

$ret = rm_file($fc, $size);

set_timezone($fc, $RDATE_HOST);

setup_cifs($fc, $WINS_DOMAIN, $WINS_USER, $WINS_USER_PASSWD,
		$WINS_IPADDR_1, $WINS_IPADDR_2);

setup_dns($fc, $NAMESERVER, $DNS_DOMAIN);

setup_nis($fc, $NIS_SERVER, $NIS_DOMAIN);

$ret = vfiler_cleanup($fc, %args);

=head1 AUTHOR

Jae Bae

=head1 FUNCTIONS

All functions, unless otherwise stated, return a true value on success,
and false on failure. On failure C<$@> is set to the appropriate error
value via C<die()>.

=head2 add_wins_dns_host($fc, $wins_dns_server, $wins_dns_domainname)

edit WINS DNS server

=over 4

=item fc

Required: hostspec of the filer

=item wins_dns_server

Required: WINS DNS Server IP

=item wins_dns_domainname

Required: WINS DNS domainname

=back

=cut

sub add_wins_dns_host($$$)
{
	my ($fc, $wins_dns_server,$wins_dns_domainname ) = @_;

	my $fh = $fc->{'hostp'};
	my $filer_ip = $fh->default_ip();

	my (undef, undef, $hosts) = say($fc, "rdfile /etc/hosts");
	if ($hosts !~ /(lab.netapp.com)/) {
		$hosts .= "$wins_dns_server     lab.netapp.com\n"
				unless ($hosts =~ /$wins_dns_server/);
		my $cmd = "wrfile /etc/hosts";
		if (! open(RSH, "|rsh $filer_ip -l root: $cmd")) {
			tstlog("-fail",
				"couldn't open rsh socket to $filer_ip for $cmd");
		}

		$hosts =~ s/rdfile \/etc\/hosts//g;
		if (! print(RSH $hosts)) {
			tstlog("-fail", "rsh socket to $filer_ip for $cmd failed");
			close(RSH);
		}

		if (! close(RSH)) {
			tstlog('-warn', "rsh socket close error to $filer_ip for $cmd");
		}
	}
}

=head2 create_passwd_file($fc, $passwd_file)

add the following accounts in the passwd file

root::0:1::/:

pcuser::65534:65534::/:

nobody::65535:65535::/:

ftp::65533:65533:FTP Anonymous:/home/ftp:

labclient:bHDiBExaWDO0Q:11656:30:Lab Login for WIN Clients (DUMMY):/home:

=over 4

=item fc

Required: hostspec of the filer

=item passwd_file

Required: name of the passwd file

=back

=cut

sub create_passwd_file($$)
{
	my ($fc, $passwd_file) = @_;

	my $fh = $fc->{'hostp'};
	my $filer_ip = $fh->default_ip();
	my $after;
	my $key;
	my $value;
	my $first = 1;

	my %passwds = (
		"root"		=> "root::0:1::/:",
		"pcuser"	=> "pcuser::65534:65534::/:",
		"nobody"	=> "nobody::65535:65535::/:",
		"ftp"		=> "ftp::65533:65533:FTP Anonymous:/home/ftp:",
		"labclient"	=> "labclient:bHDiBExaWDO0Q:11656:30:Lab Login for WIN Clients (DUMMY):/home:"
	);

	foreach $key (keys %passwds) {
		$value = $passwds{$key};


		if (-e $passwd_file) {
			(undef, undef, $after) = say($fc, "rdfile $passwd_file");
			$after .= "$value\n" unless ($after =~ /$key/);
		} else {
			if ($first == 1) {
				$after = "$value\n";
			} else {
				$after .= "$value\n";
			}
			$first++;
		}


		my $cmd = "wrfile $passwd_file";
		if (! open(RSH, "|rsh $filer_ip -l root: $cmd")) {
			die("couldn't open rsh socket to $filer_ip for $cmd");
		}

		$after =~ s/rdfile $passwd_file//g;
		if (! print(RSH $after)) {
			close(RSH);
			die("rsh socket to $filer_ip for $cmd failed");
		}

		if (! close(RSH)) {
			die("rsh socket close error to $filer_ip for $cmd");
		}
	}

	return(1);
}

=head2 set_timezone($fc, $rdate_host)

set timezone and rdate

=over 4

=item fc

Required: hostspec of the filer

=item rdate_host

Required: hostname for rdate

=back

=cut

sub set_timezone($$)
{
	my ($fc, $rdate_host) = @_;

	if (ref($fc->hostp) !~ /Simulator/) {
		my (undef, undef, $after) = say($fc, "timezone");
		if ($after !~ /(PST8PDT)/) {
			say($fc, "timezone PST8PDT");
		}
		say($fc, "rdate $rdate_host");
	}
}

=head2 setup_nis($fc, $nis_servername, $nis_domainname)

setup NIS

=over 4

=item fc

Required: hostspec of the filer

=item nis_servername

Required: name of NIS server

=item nis_domainname

Required: name of NIS domain

=back

=cut

sub setup_nis($$$)
{
	my ($fc, $nis_servername, $nis_domainname) = @_;

	my (undef, undef, $after) = say($fc, "nis info");
	if ($after !~ /(NIS domain is)/) {
		say($fc, "options nis.domainname $nis_domainname");
		say($fc, "options nis.servers $nis_servername");
		say($fc, "options nis.enable on");
	}
}

=head2 setup_dns($fc, $nameserver, $dns_domainname)

setup DNS

=over 4

=item fc

Required: hostspec of the filer

=item nameserver

Required: name of nameserver

=item dns_domainname

Required: name of DNS domain

=back

=cut

sub setup_dns($$$)
{
	my ($fc, $nameserver, $dns_domainname) = @_;

	my $fh = $fc->{'hostp'};
	my $filer_ip = $fh->default_ip();

	my (undef, undef, $server) = say($fc, "rdfile /etc/resolv.conf");
	if ($server !~ /(nameserver $nameserver)/) {
		$server .= "nameserver $nameserver\n" unless ($server =~ /$nameserver/);
	
		my $cmd = "wrfile /etc/resolv.conf";
		if (! open(RSH, "|rsh $filer_ip -l root: $cmd")) {
			tstlog("-fail", "couldn't open rsh socket to $filer_ip for $cmd");
		}

		$server =~ s/rdfile \/etc\/resolv.conf//g;
		if (! print(RSH $server)) {
			tstlog("-fail", "rsh socket to $filer_ip for $cmd failed");
			close(RSH);
		}

		if (! close(RSH)) {
			tstlog('-warn', "rsh socket close error to $filer_ip for $cmd");
		}

		say($fc, "options dns.domainname $dns_domainname");
		say($fc, "options dns.enable on");
	}
}

=head2 setup_cifs($fc, %args)

setup CIFS

=over 4

=item fc

Required: hostspec of the filer

=item args

Required: arguments

=back

=cut

sub setup_cifs($%)
{
	my ($fc, %args) = @_;

	my $domainname = $args{domainname} || "";
	my $cifs_name = $args{cifs_name} || "";
	my $wins_user = $args{wins_user} || "";
	my $wins_user_passwd = $args{wins_user_passwd} || "";
	my $wins_ipaddr1 = $args{wins_ipaddr1} || "";
	my $wins_ipaddr2 = $args{wins_ipaddr2} || "";
	my $wins_ipaddr_count = $args{wins_ipaddr_count} || 0;
	my $workgroup_name = $args{workgroup_name} || "";
	my $user_auth_type = $args{user_auth_type} || "";

	my (undef, undef, $after) = say($fc, "cifs shares",-swallow);
	if ($after =~ /(CIFS not running)/gm) {
		say($fc, "cifs setup", '-noprompt', '-wait', "Enable CIFS access");
		say($fc, "", '-noprompt', '-nocr', '-wait', "Do you.*:");
		say($fc, "y", '-blind');

		if ($wins_ipaddr_count > 0) {
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Please enter next WINS IP address.*:");
			say($fc, "$wins_ipaddr1", '-blind');
			if ($wins_ipaddr_count == 2) {
				say($fc, "", '-noprompt', '-nocr', '-wait',
					"Please enter next WINS IP address.*:");
				say($fc, "$wins_ipaddr2", '-blind');
			}
		}

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Please enter next WINS IP address.*:");
		say($fc, "\r", '-nocr', '-blind');

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want to modify this name.*:");
		say($fc, "y", '-blind');

		say($fc, "", '-noprompt', '-nocr',
			'-wait', "Enter the CIFS name for the filer:");
		#say($fc, "$cifs_name", '-blind');
		my (undef, undef, $after) = say($fc, "$cifs_name", '-blind');
		if ($after =~ /Enter\s+the\s+CIFS\s+name/) {
			$cifs_name = "$cifs_name"."1";
			say($fc, "$cifs_name", '-blind');
		}
		
		say($fc, "", '-noprompt', '-nocr', '-wait',
			"What type of authentication will this filer use.*:");

               if ($user_auth_type && $workgroup_name) {
			say($fc, "2\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Do you want to modify this name.*:");
			say($fc, "y", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Enter the Workgroup for the filer.*:");
			(undef, undef, $after) = say($fc, "$workgroup_name", '-blind');
			if ($after =~ /Enter\s+the\s+Workgroup/) {
				$workgroup_name = "$workgroup_name"."1";
				say($fc, "$workgroup_name", '-blind');
			}
			say($fc, "", '-noprompt', '-nocr', '-wait', "Create the.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
                               "^Set the password for.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"^Retype the password :");
			say($fc, "\r", '-nocr', '-blind');
		} else {
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Enter the Windows Domain for the filer.*:");
			say($fc, "$domainname", '-blind');
	
			say($fc, "", '-noprompt', '-nocr', '-wait', "Create.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Set the password for.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait', "Retype the.*:");
			say($fc, "\r", '-nocr', '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Please enter the Windows 2000 user.*:");
			say($fc, "$wins_user", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',	
				"^Password for $wins_user:");
			say($fc, "$wins_user_passwd", '-blind');

			say($fc, "\r", '-nocr', '-blind');

			return(1);
		}
	}
}

=head2 mk_file($fc, $size, $filename, [%args])

make a file

=over 4

=item fc

Required: hostspec of the filer

=item size

Required: size of file [k | b | m]

=item filename

Required: name of file

=item args

Optional: hash of options.
           $args{count} how many files to create
           $args{step_size} step size in blocks

=back


=cut

sub mk_file($$$;%)
{
	my ($fc, $size, $filename, %args) = @_;
	my $count = $args{count} || 1;
	my $step_size = $args{step_size} || "";
	my $timeout = $args{timeout} || 3600;

	logTrc("mk_file fc:$fc, size:$size, filename:$filename "
	       ." count:$count, step_size:$step_size");

	return eval {		
		sleep(1);
		if (my $server = are_we_ontapng($fc)) {
		    return $server->mk_file(file=>$filename, size=>$size, count=>$count,
					   step_size=>$step_size, timeout=>$timeout);
		}
		my (undef, undef, $after) =
			say($fc,
			    "priv set diag; mkfile $size $filename $count $step_size; priv set", -timeout, $timeout);
		if (($after =~ /file\s+creation\s+failed/) ||
		    ($after =~ /write\s+failed/)) {
			die("file creation failed: $after");
		}
		return(1);
	};
}

=head2 sync($fc, $timeout)

sync

=over 4

=item fc

Required: hostspec of the filer

=item timeout

Required: the timeout value

=back

=cut

sub sync($$)
{
	my ($fc, $timeout) = @_;
	
	if (my $server = are_we_ontapng($fc)) {
	    return $server->sync(timeout=>$timeout);
	} 
	say($fc, "priv set test; sync; sync; priv set", "-timeout", $timeout);
}

=head2 rm_file($fc, $filename)

remove a file

=over 4

=item fc

Required: hostspec of the filer

=item filename

Required: name of file

=back

=cut

sub rm_file($$)
{
	my ($fc, $filename) = @_;

	return eval {
  	        if (my $server = are_we_ontapng($fc)) {
		    return $server->rm_file(file=>$filename);
		}
		my (undef, undef, $after) = say($fc,
			"priv set diag; rm $filename; priv set");
		if ($after =~ /No\s+such\s+file\s+or\s+directory/) {
			die($after);
		} else {
			return(1);
		}
	};
}

=head2 check_vfiler($fc, $vfname)

check that the vfiler is running

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=back

=cut

sub check_vfiler {
	my ($fc, $vfname) = @_;

	return eval {
		my $text = sayEOF($fc, "vfiler status $vfname");
		if ($text =~ /$vfname\s+running/s) {
			return (1);
		} elsif ($text =~ /$vfname\s+stopped,\s+inconsistent/s) {
			die("$vfname is stopped, inconsistent");
		} elsif ($text =~ /$vfname\s+stopped/s) {
			die("$vfname is stopped");
		} else {
			return (0);
		}
	};
}

=head2 check_ipspace($fc, $ipsname)

check that there is the ipspace

=over 4

=item fc

Required: hostspec of the filer

=item ipsname

Required: name of ipspace

=back

=cut

sub check_ipspace {
	my ($fc, $ipsname) = @_;

	return eval {
		my $text = sayEOF($fc, "ipspace list");
		if ($text =~ /$ipsname/s) {
			return (1);
		} else {
			return (0);
		}
	};
}

=head2 host_connect($hostspec, $connspec, $timeout, $cmdtimeout)

call host() and connect() with timeout

=over 4

=item hostspec

Required: hostspec of the filer

=item connspec

Required: connspec of the filer

=item timeout

Required: timeout

=item cmdtimeout

Required: timeout for command

=back

=cut

sub host_connect {
	my ($hostspec, $connspec, $timeout, $cmdtimeout) = @_;

	return eval {
		my $fh = host($hostspec);
		my $fc = connect("#rsh-socket", -timeout, $timeout);
		$fc->timeout($cmdtimeout);

		my $text = sayEOF($fc, "vfiler status");
		if ($text =~ /vfiler is being run from another connection/) {
			logStep("===== HUNG IN THE INTERACTIVE PROMPT =====");
#			logStep("Hung in the interactive prompt, so reboot");
#			if (ref($fc->hostp) =~ /Simulator/) {
#				sayEOF($fc, "reboot -t 0");
#			} else {
#				$fh->powercycle();
#				sleep(200);
#			}
		}

		disconnect($fc);

		$fc = connect($connspec, -timeout, $timeout);
		$fc->timeout($cmdtimeout);
		sleep(1);

		return($fc);
	};
}

=head2 create_vfiler_non_default_ipspace($fc, $vfname, $ipsname, %args)

create a vfiler using non-default ipspace, setup, set default route,
setup CIFS, and export FS

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item ipsname

Required: name of ipspace

=item args

Required: arguments

=back

=cut

sub create_vfiler_non_default_ipspace {
	my ($fc, $vfname, $ipsname, %args) = @_;	

	my ($ipaddr_paths) = "$args{ipaddr_paths}" || "";
	my ($if_to_use) = "$args{if_to_use}" || "";
	my ($netmask) = "$args{netmask}" || "";
	my ($dnsdomain) = "$args{dnsdomain}" || "";
	my ($dnsserver) = "$args{dnsserver}" || "";
	my ($dnsserver1) = "$args{dnsserver1}" || "";
	my ($dnsserver2) = "$args{dnsserver2}" || "";
	my ($nisdomain) = "$args{nisdomain}" || "";
	my ($nisserver) = "$args{nisserver}" || "";
	my ($ipaddr_count) = "$args{ipaddr_count}" || 1;
	my ($fs) = "$args{fs}" || "";
	my ($cifs_name) = "$args{cifs_name}" || "";
	my ($wins_domain) = "$args{wins_domain}" || "";
	my ($wins_user) = "$args{wins_user}" || "";
	my ($wins_user_passwd) = "$args{wins_user_passwd}" || "";
	my ($wins_ipaddr1) = $args{wins_ipaddr1} || "";
        my ($wins_ipaddr2) = $args{wins_ipaddr2} || "";
        my ($wins_ipaddr_count) = $args{wins_ipaddr_count} || 0;
        my ($workgroup_name) = $args{workgroup_name} || "";
        my ($user_auth_type) = $args{user_auth_type} || "";
        my ($after_migration) = $args{after_migration} || "";
        my ($gateway) = $args{gateway} || "";

	return eval {
		vfiler_create_non_default_ipspace($fc, $vfname, $ipsname,
			if_to_use=>"$if_to_use",
                        ipaddr_paths=>"$ipaddr_paths") or
			die("can't create $vfname using non-default ipspace: $@");

		vfiler_run_vfiler_setup($fc, $vfname,
			if_to_use=>"$if_to_use",
			netmask=>"$netmask",
			dnsdomain=>"$dnsdomain",
			dnsserver=>"$dnsserver",
			dnsserver1=>"$dnsserver1",
			dnsserver2=>"$dnsserver2",
			nisdomain=>"$nisdomain",
			nisserver=>"$nisserver",
			ipaddr_count=>"$ipaddr_count") or
			die("vfiler run $vfname setup failed: $@");

		vfiler_run_vfiler_add_default_route($fc, $vfname, $gateway) or
			die("can't add a default route for $vfname: $@");

		if (!$args{no_cifs}) {
			vfiler_run_vfiler_cifs_setup($fc, $vfname,
				wins_ipaddr1=>"$wins_ipaddr1",
				wins_ipaddr2=>"$wins_ipaddr2",
				wins_ipaddr_count=>"$wins_ipaddr_count",
				workgroup_name=>"$workgroup_name",
				user_auth_type=>"$user_auth_type",
				cifs_name=>"$cifs_name",
				wins_domain=>"$wins_domain",
				wins_user=>"$wins_user",
				wins_user_passwd=>"$wins_user_passwd"
			) or die("vfiler run $vfname cifs setup failed: $@");
		}

		vfiler_run_vfiler_exportfs($fc, $vfname, $fs) or
			die("vfiler run $vfname exportfs -i -o anon=0 $fs failed: $@");

		return(1);
	};
}

=head2 vfiler_cleanup($fc, %args)

cleanup vfiler related things for the specific vfiler

=over 4

=item fc

Required: hostspec of the filer

=item args

Required: arguments

=back

=cut

sub vfiler_cleanup {
	my ($fc, %args) = @_;

        my ($vfilername) = $args{vfilername} || "";
        my ($vol) = $args{vol} || "";
        my ($ipsname) = $args{ipsname} || "";
        my ($if_to_use) = $args{if_to_use} || "";
        my ($remote_filer) = $args{remote_filer} || "";
        my ($disconnect) = $args{disconnect} || "";
	my @volnames;

	return eval {
		if ($vfilername) {
			if ($remote_filer) {
				vfiler_stop_destroy($fc, $vfilername,
					$remote_filer);
			} else {
				vfiler_stop_destroy($fc, $vfilername);
			}
		}

		if ($ipsname && $if_to_use) {
			if (check_ipspace($fc, $ipsname)) {
				assign_default_ipspace($fc, $if_to_use);
				# Jae commentted out destroy_ipspace() because
				# you can't create an ipspace if you
				# create and destroy ipspace many times.
				# This problem exists with vfiler, so we
				# need to make it, not to destroy ipspace
				# for nightly runs

				# uncomment the following line if vfiler allows
				# to create and destroy many ipspaces
				#destroy_ipspace($fc, $ipsname);
			}
		}

		if ($vol) {
			@volnames = split(" ", $vol);
			foreach my $volname (@volnames) {
				if (vol_exists($fc, $volname)) {
					my $status =
						vol_status($fc, $volname);
					if ($status->state() ne 'offline') {
						if ($volname ne 'vol0') {
							vol_offline($fc, $volname);
						}
					}
					if ($volname ne 'vol0') {
						vol_destroy($fc, $volname);
					}
				}
			}
		}

		if ($disconnect) {
			disconnect($fc);
			sleep(3);
		}

		return(1);
	};
}

sub get_vfiler_root_path {
    my ($h, $vfiler) = @_;
    my $pathstat = say($h, "vfiler status -a $vfiler");
    my @paths = ($pathstat =~ m#Path: (\S+)#mg);
    return $paths[0];
}
sub get_vfiler_storage_paths {
    my ($h, $vfiler) = @_;
    my $pathstat = say($h, "vfiler status -a $vfiler");
    my @paths = ($pathstat =~ m#Path: (\S+)#mg);
    return @paths;
}

sub vfiler_avail_index {
	my ($fc, %args) = @_;

	use VFILER::Setup qw(vfiler_status vfiler_list);

	my $hostrec = $fc->hostp();
	my %param = $hostrec->param_fields();
	my %all_idx = ();

	for (my $index = 1; ; ++$index) {
		my $k = "IP_$index";
		last if (!$param{$k});
		$all_idx{$param{$k}} = $index;
		logDbg("vfiler_avail_index: adding $index index in all_idx");
	}

	my $rsh_conn = connect("#rsh");
	my @vfiler_list = vfiler_list($rsh_conn);
	my @avail_idx = ();

	foreach my $ip (keys %all_idx) {
		logDbg("vfiler_avail_index: outer loop: ip = $ip");
		my $found = 0;
		foreach my $vf (@vfiler_list) {
			my %vf_stat = vfiler_status($rsh_conn, $vf);
			my $ipaddr0 = $vf_stat{$vf}{ipaddr0};
			logDbg("vfiler_avail_index: innner loop: vf = $vf, "
					. "ipaddr0 = $ipaddr0");
			if ($ipaddr0 eq $ip) {
				$found = 1;
				last;
			}
		}
		if (!$found) {
			push @avail_idx, $all_idx{$ip};
			logDbg("vfiler_avail_index: adding ip = $ip, "
			. "index = $all_idx{$ip}");
		}
	}

	return @avail_idx;
}

=head2 create_vfiler_and_setup($fc, %args)

creates a new vfiler using available resources and returns a hash containing
information about the newly created vfiler

=over 4

=item fc

Required: hostspec of the filer

=item args

Required: arguments

=back

=cut

sub create_vfiler_and_setup {
	my ($fc, %args) = @_;
	my $success = 0;

	my $hostrec = $fc->hostp();
	my %param = $hostrec->param_fields();
	my ($idx) = vfiler_avail_index($fc); $idx = 'X' if (!$idx);
	logStep("returned index of resources to use is $idx");

        my ($vfilername) = $args{vfilername} || $param{"HOSTNAME_$idx"} || "";
        my ($if_to_use) = $args{if_to_use} || $param{"IFC_$idx"} || "";
        my ($ip_to_use) = $args{ip_to_use} || $param{"IP_$idx"} || "";
        my ($ipsname) = $args{ipsname} || "ipsname_$idx";
        my ($remote_filer) = $args{remote_filer} || "";
        my ($part_host) = $args{part_host} || $param{part_host} || "";
        my ($vol) = $args{vol} || "/vol/testvol_$idx";

        my ($netmask) = $param{"NETMASK_$idx"} || "";
        my ($gateway) = $param{GATEWAY} || "";
        my ($dns_domain) = $param{DNS_DOMAIN} || "";
        my ($dns_server) = $param{DNS_SERVER} || "";
        my ($wins_server) = $param{WINS_SERVER} || "";
        my ($nis_domain) = $param{NIS_DOMAIN} || "";
        my ($nis_server) = $param{NIS_SERVER} || "";
        my ($active_dir_user) = $param{ACTIVE_DIR_USER} || "";
        my ($active_dir_passwd) = $param{ACTIVE_DIR_PASSWD} || "";
        my ($rdate_server) = $param{RDATE_SERVER} || "";
	my ($do_not_create) = $args{do_not_create} || 0;

	logStep("vfiler creation and setup is using following resources:\n"
		. "\t vfilername=$vfilername\n\t if_to_use=$if_to_use\n"
		. "\t ip_to_use=$ip_to_use\n\t ipsname=$ipsname\n"
		. "\t remote_filer=$remote_filer\n\t part_host=$part_host\n"
		. "\t vol=$vol\n\t netmask=$netmask\n\t gateway=$gateway\n"
		. "\t dns_domain=$dns_domain\n\t dns_server=$dns_server\n"
		. "\t wins_server=$wins_server\n\t nis_domain=$nis_domain\n"
		. "\t nis_server=$nis_server\n"
		. "\t active_dir_user=$active_dir_user\n"
		. "\t active_dir_passwd=$active_dir_passwd\n"
		. "\t rdate_server=$rdate_server");

	die("Available IP not found to create vfiler") if (!$ip_to_use);
	die("Available IFC not found to create vfiler") if (!$if_to_use);
	die("Available vfilername not found to create vfiler")
		if (!$vfilername);

	set_timezone($fc, $rdate_server) if ($rdate_server);

	my @vol_list = split(/\s+/, $vol);
	my $vol_path = "";
	foreach my $v (@vol_list) {
		$v =~ s|^(\w)|/vol/$1|;
		if ($vol_path) {
			$vol_path = "$vol_path $v"
		} else {
			$vol_path = $v;
		}
	}

	my (%setup_args) = (
		ipsname => $ipsname,
		ip_to_use => $ip_to_use,
		vol_path => $vol_path,
		ipaddr_paths => "-i $ip_to_use $vol_path",
        	if_to_use => $if_to_use,
		netmask => $netmask,
		dnsdomain => $dns_domain,
		dnsserver => $dns_server,
		nisdomain => $nis_domain,
		nisserver => $nis_server,
		ipaddr_count => 1,
		fs => $vol_path,
		cifs_name => $vfilername,
		wins_domain => $dns_domain,
		wins_user => $active_dir_user,
		wins_user_passwd => $active_dir_passwd,
		wins_ipaddr1 => $wins_server,
        	wins_ipaddr_count => 1,
        	gateway => $gateway
	);

	if (!$do_not_create) {
		create_ipspace($fc, $ipsname)
			if (!exists_ipspace($fc, $ipsname));
		assign_ipspace($fc, $ipsname, $if_to_use)
			if (!ifc_in_ipspace($fc, $ipsname, $if_to_use));
		create_vfiler_non_default_ipspace($fc, $vfilername, $ipsname,
				%setup_args)
			and $success = 1;
	} else {
		$success = 1;
	}

	return {success => $success, vfilername => $vfilername,
		ipsname => $ipsname, %setup_args};
}

=head2 new_vfiler_resources($fc, %args)

gets a hash of resources which can be used to create a vfiler

=over 4

=item fc

Required: hostspec of the filer

=item args

Required: arguments

=back

=cut

sub new_vfiler_resources {
	my ($fc, %args) = @_;

	my $vfiler = create_vfiler_and_setup($fc, do_not_create => 1, %args);
	die("could not obtain vfiler resources") if (!$vfiler->{success});

	return $vfiler;
}

=head2 get_vfiler_status_in_dr($fc, $vfname)

get a vfiler status during DR

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=back

=cut

sub get_vfiler_status_in_dr {
	my ($fc, $vfname) = @_;

	return eval {
		sleep(3);

		my $status;
		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
		if ($after =~ /stopped, DR backup/) {
			$status = "backup";
		} elsif ($after =~ /stopped, DR destination activated/) {
			$status = "activated";
		} elsif ($after =~ /running/) {
			$status = "running";
		} elsif ($after =~ /stopped, DR checking/) {
			$status = "checking";
		} else {
			$status = "failed";
		}

		return($status);
	};
}

=head2 clean_vfiler_and_resources($hostspec)

clean up vfilers and resources

=over 4

=item hostspec

Required: hostspec of the filer

=item args

Required: arguments

=back

=cut

sub clean_vfiler_and_resources {
	my ($hostspec, %args) = @_;

	my $fh = host($hostspec);	
	my $fc = connect("#rsh-socket");
        my ($ip_addr) = $args{ip_addr} || "";
        my ($if_is_used) = $args{if_is_used} || "";
        my ($qtree) = $args{qtree} || "";

	my $default_volnames =  "vvolVFT0_cloned_clone1 ".
				"vvolVFT0_cloned_clone2 ".
				"vvolVFT0_cloned_clone3 ".
				"vvolVFT0_cloned1_clone1 ".
				"vvolVFT0_cloned2_clone1 ".
				"vvolVFT0_cloned ".
				"vvolVFT0_cloned1 ".
				"vvolVFT0_cloned2 ".
				"vvolVFT0_cloned3 ".
				"vvolVFT0_cloned4 ".
				"vvolVFT0_cloned5 ".
				"vvol1 ".
				"vvol2 ".
				"vvol3 ".
				"vvol4 ".
				"vvol5 ".
				"vvolVFT1 ".
				"vvolVFT2 ".
				"vvolVFT3 ".
				"vvolVFT0";
	my $default_vfnames = "vFiler100 vFiler101 vFiler102";
	my $text;

	if (is_licensed($fc, 'multistore')) {

	sleep(3);

	my $nvfilers = vfiler_limit_in_use($fc);

	if ($nvfilers > 1) {
		my @out1 = split(" ", $default_vfnames);
		foreach my $vfname (@out1) {
			vfiler_stop($fc, $vfname);
			$text = sayEOF($fc, "vfiler destroy $vfname -f");
			if ($text =~ /\"(.*?)\"/s) {
				sayEOF($fc, "$1");
			}
		}

		if (!$qtree) {
			my @out2 = split(" ", $default_volnames);
			foreach my $volname (@out2) {
				if (vol_exists($fc, $volname)) {
					my $status = vol_status($fc, $volname);
					if ($status->state() ne 'offline') {
						vol_offline($fc, $volname);
					}

					vol_destroy($fc, $volname);
				}
			}
		}

		$text = sayEOF($fc, "vfiler status");
		@out1 = split(" ", $text);

		foreach my $vfname (@out1) {
		if ($vfname !~ /vfiler0|running|stopped|stopped,|DR|destination|activated|backup|checking|inconsistent/) {
			if ($vfname) {
				my @out3;
				if (!$qtree) {
					$text = sayEOF($fc, "vfiler status -r $vfname");
					@out3 = split(" ", $text);
				}

				vfiler_stop($fc, $vfname);
				$text = sayEOF($fc, "vfiler destroy $vfname -f");
				if ($text =~ /\"(.*?)\"/s) {
					sayEOF($fc, "$1");
				}

				if (!$qtree) {
				   $text = sayEOF($fc, "vfiler status $vfname");
				   if ($text !~ /stopped|stopped,/) {
					foreach my $vol (@out3) {
					   if ($vol =~ /\/vol\/(\w+)/s) {
						my $volname = "$1";
						if (vol_exists($fc, $volname)) {
							my $status = vol_status($fc, $volname);
							if ($status->state() ne 'offline') {
								if ($volname ne 'vol0') {
									vol_offline($fc, $volname);
								}
							}

							if ($volname ne 'vol0') {
								vol_destroy($fc, $volname);
							}
						}
					   }
					}
				    }
				}
			}
		}

		}
	}

	$text = sayEOF($fc, "ipspace list");
	$text =~ /Number of ipspaces configured: (\d+)/;
	if ($1 > 1) {
		my @out1 = split(" ", $text);

		foreach my $ips (@out1) {
		   if ($ips =~ /\((\w+)/s) {
			if ($1 ne 'no') {
				# just assign the ipspace to default-ipspace, but
				# don't destroy the ipspace because vfiler
				# can't handle if you destroy and create ipspace
				# many times
				if ($1 ne 'e0' || $1 ne 'ns0') {
					$text = sayEOF($fc, "ifconfig $1");
					my @out2 = split(" ", $text);
					my $found = 0;
					foreach my $ip_addr (@out2) {
						if ($ip_addr =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/s) {
							$found = 1;
						}
					}
					if (!$found) {
						assign_ipspace($fc, "default-ipspace", $1);
					}
				}

				# uncomment the following line if vfiler allows
				# to create and destroy many ipspaces
				#destroy_ipspace($fc, $ipsname);
			}
		   }

		}
	}

	}

	if ($ip_addr) {
		my @out1;

		if (ref($fc->hostp) =~ /Simulator/) {
			$text = sayEOF($fc, "ifconfig ns1");
		   	if ($text =~ /$ip_addr/s) {	
		        	sayEOF($fc, "ifconfig ns1 0.0.0.0 down");
			}
		} else {
			$text = sayEOF($fc, "ifconfig -a");
		   	if ($text =~ /$ip_addr/s) {
		      		my @out1 = split(" ", $text);
		      		foreach my $ifname (@out1) {
		         		if ($ifname =~ /^(\w+)\:/s) {
		            		   if ($1 !~ /lo|00/s) {
		     	       		      $text = sayEOF($fc, "ifconfig $1");
		     	       		      if ($text =~ /$ip_addr/s) {
		      		  		 @out1 = split(" ", $text);
		      		  		 foreach my $ifname (@out1) {
		       		     		    if ($ifname =~ /^(\w+)\:/s) {
						       if ($1 !~ /00/) {
		         		   		  sayEOF($fc,
							  "ifconfig $1 0.0.0.0 down");
						       }
		       		     		    }
		      		  		 }
		     	       		      }
			    		   }
			 		}
		      		}
		   	}
		}
	}

	if ($if_is_used) {
		if (ref($fc->hostp) =~ /Simulator/) {
			$if_is_used = "ns1";
		}

		$text = sayEOF($fc, "ifconfig $if_is_used");
		my @out1 = split(" ", $text);
		foreach my $ip_addr (@out1) {
			if ($ip_addr =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/s) {
				sayEOF($fc, "ifconfig $if_is_used 0.0.0.0 down");
			}
		}
	}

	disconnect($fc);
}

1;
