# $Id: //depot/prod/test/Rsierranevada/lib/VFILER/VfilerUsers.pm#1 $

## @summary misc routines for nate script
##
## @description
##      VfilerUsers.pm functions
##
## @status done 
## 
## @author  Smit 
##
## @keywords Vfiler User Group Account

package VFILER::VfilerUsers;
use strict;
use Carp qw(croak);
use base qw(Exporter);
use Params::Validate qw(validate validate_pos 
                       SCALAR OBJECT BOOLEAN ARRAYREF HASHREF);
use Tharn;
use SPARSE::Util::Logger;
use CLI::FileIO qw(rdfile wrfile);
use Utils qw(make_conn);
use Filer::Version qw(get_wafl_version WV_SCRIMSHAW_MAX);
use SayWrap qw(sayEOF);
use Data::Dumper;
use Params::Validate qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF UNDEF);

BEGIN
{
  our @EXPORT = qw(
	createUsersForVfiler
	createGroupsForVfiler
	setPrimaryGroupForVfiler
	addPasswdEntryForVfiler
	delPasswdEntryForVfiler
	delPasswdListForVfiler
	addGroupEntryForVfiler
	delGroupEntryForVfiler
	delGroupListForVfiler
	addPasswdListForVfiler
	addGroupListForVfiler
	getNextUIDForVfiler
	getNextGIDForVfiler
   	readGroupFileForVfiler
    	readPasswdFileForVfiler
	findUserForVfiler
	findGroupForVfiler
	writeDefaultPasswdFileForVfiler
	writePasswdFileForVfiler
	writeGroupFileForVfiler
	appendPasswdFileForVfiler
	appendGroupFileForVfiler
  );
}

##
## @notes
##
##	Convenional hash table representation of a 'user' (/vol/$vfiler_root/etc/passwd entry) is
##	shown by the following snippet from createUsersForVfiler. This convention is used
##	by all functions in this module
##
##	my $user_ref = {
##				login => $login,
##				passwd => "",
##				uid => $uid,
##				gid => $gid,
##				name => "",
##				home => "",
##				shell => "nfs anonymous user",
##				group => $group,
##			};
##
##	Conventional representation of a 'group' (/vol/$vfiler_root/etc/group entry) is:
##
##	my $grp_hash = {
##				group => $group,
##				passwd => "x",
##				gid => $gid,
##				members => "",
##			};
##


## @name ::createUsersForVfiler()
## @synopsis @users = createUsersForVfiler ($filerconn, $nusers, $vfiler_root)
## @arg $filerconn Required Filer Connection Object
## @arg $nusers: Number of unique users to create. One entry for each
##            user is created in the /vol/$vfiler_root/etc/passwd file. Also, a unique
##            group is created for each of the user in the /vol/$vfiler_root/etc/group
##            and that user is placed in this unique group in
##            the /etc/passwd file.
## @arg $vfiler_root vfiler root volume 
## @description creates specified number of user entries in
##            the /vol/$vfiler_root/etc/passwd.
## @returns array to users created
## @status public
#
#
#   Creates $nusers new users, each in it's own new group
#   If 'gid' optional arg is specified, then no new groups are
#   created, and the specified gid is used for all new users
#
#  Args:
#
#	filerspec:		Required
#
#	nusers:			Required. Number of users to create.
#
#	vfiler_root:		Required. root volume for vfiler.
#
#   Optional args:
#
#	login_prefix		Default: mult_user
#
#	group_prefix		Default: mult_group
#
#	group			Group name. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user.
#
#	gid			Group gid. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user. Ignored if 'group'
#				is specified.
#
#	samegroup		If 'group' and 'gid' are not specified, and 'samegroup' is 1,
#				then a single new group is created for all the new users.
#
#	current_groups		HASHREF which represents current groups. If defined,
#			        and the HASHREF is empty, then the HASHREF is updated with
#			        current groups from /vol/$vfiler_root/etc/groups. Otherwise the  ARRARYREF is
#			        used to determine current groups, and /etc/groups is not read.
#			        The HASHREF is also updated with any new groups created
#
#	current_users		HASHREF which represents current users. If defined,
#			        and the HASHREF is empty, then the HASHREF is updated with
#			        current users from /vol/$vfiler_root/etc/passwd. Otherwise the  HASHREF is
#			        used to determine current users, and /etc/passwd is not read.
#			    	The HASHREF is also updated with any new users created
#
#	created_users		HASHREF. If defined, then the created users are added
#				to this hash.
#
#	created_groups		HASHREF. If defined, then the created groups are added
#			    	to this hash.
#
#	min_uid			Default value is set in _getNextUIDForVfiler
#
#	max_uid			Default value is set in _getNextUIDForVfiler
#
#	min_gid			Default value is set in _getNextGIDForVfiler
#
#	min_gid			Default value is set in _getNextGIDForVfiler
#
#	The following options apply only if 'nusers' is 1.
#	There are no defaults.
#
#	login			Login to use for new user
#
#	uid			Uid to use for new user
#
#	new_users		ARRAYREF, where the elements of the array are each a HASHREF
#				If supplied, these hashesare used as 'templates' to create the
#				new users.Defined fields are used, and undefined fields are
#				filled-in. This option is ignored unless 'nusers' is 0.
#
sub createUsersForVfiler ($$$%args)
{

 logcomment("Enter createUsersForVfiler\n\n");

 my ($filerspec, $nusers, $vfiler_root, %opts) = @_;

  return eval {
    my $spec_group_ref = undef;
    my $spec_group;
    my $spec_gid;
    my $group_specified;
    my $next_login;
    my @new_users;
    my $login_prefix = $opts{login_prefix};
    #  'min_gid', 'max_gid', 'group_prefix' are defaulted by createGroup,
    #  which is passed these values

    my $current_users_hash = $opts{current_users} || {};
    my $created_users_hash = $opts{created_users} || {};
    my $current_groups_hash = $opts{current_groups} || {};
    my $created_groups_hash = $opts{created_groups} || {};
    my $next_uid = undef;

    if (!keys(%$current_users_hash)) {
         $current_users_hash = readPasswdFileForVfiler($filerspec,$vfiler_root);
		#or die($@);
    }

    $nusers = 0 
        unless (defined($nusers));

    #  4 cases:
    #	   1. $nusers is 0, and no 'new_users' option given
    #		==> return empty hashtable
    #	   2. $nusers is 0, and 'new_users' option given
    #		==> @new_users array is set from 'new_users option
    #		    Also, $nusers is set to size of @new_users list
    #	   3. $nusers is 1
    #		==> a singleton @new_users list is constructed using
    #		    the 'uid' and/or 'login' options, if they are given.
    #		    This is the only case where these options are used.
    #	   4. $nusers is > 0
    #		==> @new_users list is empty. If 'samegroup' option is 1,
    #		    then a single new group will be created for all new users.
    if ($nusers == 0) {
        if ($opts{new_users}) {
	    @new_users = @{$opts{new_users}};
	    $nusers = scalar(@new_users);
	}
    } elsif ($nusers == 1) {
	@new_users = ( { login=>$opts{login}, uid=>$opts{uid} } );
    } else {
	@new_users = ();
    }

    return( () )
        if (!$nusers);

    my $num_uids_needed = 0;
    my $new_groups_ref = undef;

    #  Calculate number of new UIDs needed for the new users.
    #  Also, for each user in @new_users, create a group in @new_groups,
    #  using the 'group' and 'gid' fields of the user.
    #
    #  2 cases:
    #	   1. @new_users is non-empty
    #		==> check each user in @new_users. Need a new UID
    #		    for each user whose 'uid' field is undefined.
    #	   2. @new_users is empty
    #		==> Need a new UID for each new user, so $nuser new UIDs
    if (@new_users) {
        $new_groups_ref = [];
        foreach my $u (@new_users) {
	    if (!defined($u->{uid})) {
	    	$num_uids_needed++;
	    }
	    push(@$new_groups_ref, { gid=>$u->{gid}, group=>$u->{group} } );
	}
    } else {
	$num_uids_needed = $nusers;
    	delete($opts{new_users});
    }
    if ($num_uids_needed > 0) {
    	$next_uid = _getNextUIDForVfiler($num_uids_needed,
				    $current_users_hash,
				    min_uid=>$opts{min_uid},
				    max_uid=>$opts{max_uid});
        }
	

    if (!keys(%$current_groups_hash)) {
    	$current_groups_hash = readGroupFileForVfiler($filerspec,$vfiler_root)
	    or die($@);
    }
    my $ngroups = 0;
    my @other_matches;
    if ($opts{gid} || $opts{group} || $opts{samegroup}) {
	$group_specified = 1;
	$spec_group_ref =
	    _findOrCreateGroup($filerspec,$vfiler_root,
				group_prefix=>$opts{group_prefix},
			         group => $opts{group}, gid => $opts{gid},
			         min_gid=>$opts{min_gid}, max_gid=>$opts{max_gid},
			         current_groups=>$current_groups_hash,
			         created_groups=>$created_groups_hash,
			         new_groups=>$new_groups_ref);
	$spec_group = $spec_group_ref->{group};
	$spec_gid = $spec_group_ref->{gid};
   } elsif ($new_groups_ref) {
	foreach my $gref (@$new_groups_ref) {
		my $new_grp = 
		    _findOrCreateGroup($filerspec,$vfiler_root,
					group_prefix=>$opts{group_prefix},
					group => $gref->{group}, gid => $gref->{gid},
					min_gid=>$opts{min_gid}, max_gid=>$opts{max_gid},
					current_groups=>$current_groups_hash,
			         	created_groups=>$created_groups_hash);
		my $new_name = $new_grp->{group};
	        $current_groups_hash->{$new_name} = $new_grp;
	}
   } else {
	$ngroups = $nusers;
	createGroupsForVfiler($filerspec, $ngroups, $vfiler_root, group_prefix=>$opts{group_prefix},
			group => $opts{group}, gid => $opts{gid},
			min_gid=>$opts{min_gid}, max_gid=>$opts{max_gid},
			current_groups=>$current_groups_hash,
			created_groups=>$created_groups_hash,
			new_groups=>$new_groups_ref);
    }
    if ($ngroups == 1) {
	my @keys = keys(%$created_groups_hash);
	$spec_group_ref =  $created_groups_hash->{$keys[0]};
	$group_specified = 1;
	$spec_group = $spec_group_ref->{group};
	$spec_gid = $spec_group_ref->{gid};
    }

    my $current_groups_list = _array_from_hash($current_groups_hash);
    my $created_groups_list = _array_from_hash($created_groups_hash);
    for (my $i=0; $i < $nusers; $i++) {
	my $uid = undef;
	my $login = undef;
	my $group = undef;
	my $gid = undef;
	my $grp_ref = undef;

	if (@new_users) {
		$uid = $new_users[$i]->{uid};
		$login = $new_users[$i]->{login};
		$group = $new_users[$i]->{group};
		$gid = $new_users[$i]->{gid};
	}
	if (!defined($uid)) {
		$uid = $next_uid + $i;
	}
	if (!defined($login)) {
		$login = sprintf "${login_prefix}%04d", $uid;
	}
	if ($group_specified) {
		$grp_ref = $spec_group_ref;
	} else {
		$grp_ref = shift(@$created_groups_list);
	}
	if (!defined($group)) {
		$group = $grp_ref->{group};
	}
	if (!defined($gid)) {
		$gid = $grp_ref->{gid};
	}

	my $user_ref = {
				login => $login,
				passwd => "",
				uid => $uid,
				gid => $gid,
				name => "",
				home => "",
				shell => "nfs anonymous user",
				group => $group,
			};
	$$current_users_hash{$login} = $user_ref;
	$$created_users_hash{$login} = $user_ref;
    }

    addPasswdListForVfiler($filerspec, $vfiler_root, $created_users_hash)
	or die($@);
    return %$created_users_hash;
  };
}


#  _findOrCreateGroup($filerspec, $vfiler_root, %opts)
#
#  Args:
#
#	filerspec:		Required
#
#	vfiler_root:		Required
#
#   Optional args:
#
#	group			Group name. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user.
#
#	gid			Group gid. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user. Ignored if 'group'
#				is specified.
#
#	group_prefix		Default: mult_group
#
#	min_gid			passed to _getNextGIDForVfiler
#
#	max_gid			passed to _getNextGIDForVfiler
#
#	current_groups		ARRAYREF which represents current groups. If defined,
#			        and the ARRAYREF is empty, then the ARRAYREF is updated with
#			        current users from /vol/$vfiler_root/etc/passwd. Otherwise the  ARRARYREF is
#			        used to determine current users, and /etc/passwd is not read.
#
#	created_groups		ARRAYREF. If defined, then the created groups are added
#			    to this list.
#
#	new_groups		ARRAYREF, where the elements of the array are each a HASHREF
#				If supplied, these hashesare used as templates to create the
#				new groups.Defined fields are used, and undefined fields are
#				filled-in. This option is ignored unless 'ngroups' is 0.
#
sub _findOrCreateGroup
{
 
  my ($filerspec, $vfiler_root, %opts) = @_;

  return eval {
    #my $filerconn = make_conn($filerspec);
    my ($group_prefix) = $opts{group_prefix} || "mult_group";
    my ($created_groups_hash) = {};
    my ($current_groups_hash) = {};

    $current_groups_hash = $opts{current_groups} || {};
    $created_groups_hash = $opts{created_groups} || {};
    if (!keys(%$current_groups_hash)) {
         $current_groups_hash = readGroupFileForVfiler($filerspec,$vfiler_root)
	     or die($@);
    }
    my $grp_ref;
    if ( defined($opts{gid}) ) {
	my $gid = $opts{gid};
	($grp_ref) = _findGroupForVfiler($current_groups_hash, gid => $gid);
    } elsif ( defined($opts{group}) ) {
	my $group = $opts{group};
	($grp_ref) = _findGroupForVfiler($current_groups_hash, group => $group);
    }
    if ($grp_ref) {
	return ( $grp_ref );
    }
    my %created_group_hash = 
        createGroupsForVfiler($filerspec, 1, $vfiler_root, group_prefix=>$opts{group_prefix},
			group => $opts{group}, gid => $opts{gid},
			min_gid=>$opts{min_gid}, max_gid=>$opts{max_gid},
			current_groups=>$opts{current_groups},
			created_groups=>$opts{created_groups},
			new_groups=>$opts{new_groups});
    my @created_group_list = values(%created_group_hash);
    my $new_grp = $created_group_list[0];
    return ( $new_grp );
  }
}


## @name ::createGroupsForVfiler()
## @synopsis @groups = createGroupsForVfiler ($filerconn, $ngroups,$vfiler_root)
## @arg $filerconn Required Filer Connection Object
## @arg $ngroups: Number of unique groups to create. One entry for each
##            group is created in the /vol/$vfiler_root/etc/group file
## @arg $vfiler_root Required root volume name of vfiler 
## @description creates specified number of groups entries in
##            /vol/$vfiler_root/etc/group.
## @returns array to group created
## @status public
#
#  Args:
#
#	filerspec:		Required. filer spec.
#
#	ngroups:		Required. Number of groups to create.
#
#	vfiler_root:		Required. root volume of vfiler.
#   Optional args:
#
#	group			Group name. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user.
#
#	gid			Group gid. No default (undefined). If defined, then
#				all created users are added to this group. Otherwise,
#				a new group is created for each user. Ignored if 'group'
#				is specified.
#
#	group_prefix		Default: mult_group
#
#	min_gid			passed to _getNextGIDForVfiler
#
#	max_gid			passed to _getNextGIDForVfiler
#
#	current_groups		ARRAYREF which represents current groups. If defined,
#			        and the ARRAYREF is empty, then the ARRAYREF is updated with
#			        current users from /vol/$vfiler_root/etc/passwd. Otherwise the  ARRARYREF is
#			        used to determine current users, and /vol/$vfiler_root/etc/passwd is not read.
#			        The ARRAYREF is also updated with any new users created
#
#	created_groups		ARRAYREF. If defined, then the created groups are added
#			        to this list.
#
#	new_groups		ARRAYREF, where the elements of the array are each a HASHREF
#				If supplied, these hashesare used as templates to create the
#				new groups.Defined fields are used, and undefined fields are
#				filled-in. This option is ignored unless 'ngroups' is 0.
#
sub createGroupsForVfiler ($$$;%)
{
  print "inside the group\n";
  logcomment("Enter createGroupForVfiler\n");


  my ($filerspec,$ngroups,$vfiler_root, %opts) = @_;

  return eval {
    my ($group_prefix) = $opts{group_prefix} || "mult_group";
    my ($created_groups_hash) = {};
    my ($current_groups_hash) = {};

    $current_groups_hash = $opts{current_groups} || {};
    $created_groups_hash = $opts{created_groups} || {};
    if (!keys(%$current_groups_hash)) {
         $current_groups_hash = readGroupFileForVfiler($filerspec, $vfiler_root);
    }
    my @new_groups;
    $ngroups = 0
        unless (defined($ngroups));
    if ($ngroups == 0 && $opts{new_groups}) {
	@new_groups = @{$opts{new_groups}};
	$ngroups = scalar(@new_groups);
    } elsif ($ngroups == 1) {
	@new_groups = ( { group=>$opts{group}, gid=>$opts{gid} } );
    } else {
	@new_groups = ();
    }

    my $num_gids_needed = 0;
    if (@new_groups) {
        foreach my $g (@new_groups) {
	    if (!defined($g->{gid})) {
	        $num_gids_needed++;
	    }
	}
    } else {
	$num_gids_needed = $ngroups;
    }
    my $next_gid = _getNextGIDForVfiler($num_gids_needed, $current_groups_hash, %opts);

    for (my $i=0; $i < $ngroups; $i++) {
	my $group = undef;
	my $gid = undef;

	if (@new_groups) {
		$gid = $new_groups[$i]->{gid};
		$group = $new_groups[$i]->{group};
	}
	if (!defined($gid)) {
		$gid = $next_gid + $i;
	}
	if (! defined($group)) {
		$group = sprintf "${group_prefix}%04d", $gid;
	}

	logDbg("adding grp=$gid $group ");
	my $grp_hash = {
				group => $group,
				passwd => "x",
				gid => $gid,
				members => "",
			};
	$$created_groups_hash{$group} = $grp_hash;
	$$current_groups_hash{$group} = $grp_hash;
    }

    addGroupListForVfiler($filerspec, $vfiler_root, $created_groups_hash)
        or die($@);

    return %$created_groups_hash;
  };
}


## @name ::setPrimaryGroupForVfiler()
## @synopsis setPrimaryGroupForVfiler($filerspec, $vfiler_root, $user, $group)
## @arg $filerconn Required Filer Connection Object
## @arg $user: user name whose group need to be set
## @arg $group: group name of the above user. The user is added to the
##            group list of this group in the /vol/$vfiler_root/etc/group file.
## @description sets the group of the specified user to the specified group
## @returns true if success
## @status public

sub setPrimaryGroupForVfiler ($$$$)
{
  my ($filerspec, $vfiler_root, $input_login, $new_gid) = @_;    

  return eval {
    my $filerconn = make_conn($filerspec);
    my $oldtext = "";
    my $newpasswdfile = "";
 
    my ($login, $passwd, $uid, $gid, $name, $home, $shell);
    $oldtext = rdfile($filerconn, "/etc/passwd");
    my @oldlines = grep(!/^rdfile/, split(/\n/, $oldtext));

    foreach my $line (@oldlines) {
	if ($line =~ /^\s*$/ || $line =~ /^\s*#/) {
		$newpasswdfile .= "$line\n";
		next;
	}
	($login, $passwd, $uid, $gid, $name, $home, $shell)
			= split(/:/, $line);

	if ($login eq $input_login) {
    		$newpasswdfile .= "${login}:${passwd}:${uid}:${new_gid}:".
			"${name}:${home}:${shell}\n";
		next;
	}
	$newpasswdfile .= "$line\n";
    }

    wrfile($filerconn, "/vol/$vfiler_root/etc/passwd", $newpasswdfile)
		or die "can't write to /etc/passwd: $@";
    sleep(4);
  }
}




## @name ::appendPasswdFileForVfiler()
## @synopsis appendPasswdFileForVfiler($filerspec, $vfiler_root, @text_list)
## @arg $filerconn Required Filer Spec or Connection Object
## @arg $vfiler_root Required root voluem for vfiler 
## @arg @text_list Required. Text to be appended to the /vol/$vfiler_root/etc/passwd file.
## @description Append text to /vol/$vfiler_root/etc/passwd file
## @returns true if success
## @status public
sub appendPasswdFileForVfiler($$@)
{
        my ($filerspec, $vfiler_root, @text_list) = @_;
        my $newtext = join("\n", @text_list);
        $newtext .= "\n" unless ($newtext =~ /\n$/);

	my $oldtext = rdfile($filerspec, "/vol/$vfiler_root/etc/passwd");
        my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
        $oldtext = join("\n", @lines);
	croak($@) if ($@);
	$oldtext .= "\n" unless ($oldtext =~ /\n$/);
	wrfile($filerspec, "/vol/$vfiler_root/etc/passwd", $oldtext . $newtext)
		or croak("can't write to /etc/passwd : $@");
        sleep(4);
	return(1);
}


## @name ::writeDefaultPasswdFileForVfiler()
## @synopsis writeDefaultPasswd($filerspec)
## @arg $filerconn Required Filer Spec or Connection Object
## @description Writes short default /etc/passwd file:
##    root::0:1::/:
##    pcuser::65534:65534::/:
##    nobody::65535:65535::/:
##    ftp::65533:65533:FTP Anonymous:/home/ftp:
## @returns true if success
## @status public
sub writeDefaultPasswdFileForVfiler($@)
{
        my ($filerspec) = @_;
        my $text = <<__EOF__;
root::0:1::/:
pcuser::65534:65534::/:
nobody::65535:65535::/:
ftp::65533:65533:FTP Anonymous:/home/ftp:
__EOF__

	my $filerconn = make_conn($filerspec);
	wrfile($filerconn, "/etc/passwd", $text)
			or croak("can't write to /etc/passwd : $@");
	sleep(4);
	return(1);
}


## @name ::writePasswdFileForVfiler()
## @synopsis writePasswd($filerspec, $vfiler_root, @text_list)
## @arg $filerconn Required Filer Spec or Connection Object
## @arg $vfiler_root Required root volume of vfiler. 
## @arg @text_list Required. Text to be appended to the /vol/$vfiler_root/etc/passwd file.
## @description Write text to /vol/$vfiler_root/etc/passwd file. Overwrites previous contents.
## @returns true if success
## @status public
sub writePasswdFileForVfiler($$@)
{
        my ($filerspec, $vfiler_root, @text_list) = @_;
        my $text = join("\n", @text_list);
        $text .= "\n" unless ($text =~ /\n$/);

	wrfile($filerspec, "/vol/$vfiler_root/etc/passwd", $text)
			or croak("can't write to /etc/passwd : $@");
        sleep(4);
	return(1);
}


## @name ::writeGroupFileForVfiler()
## @synopsis writeGroup($filerspec, $vfiler_root, @text_list)
## @arg $filerconn Required Filer Spec or Connection Object
## @arg $vfiler_root Required root volume of vfiler.
## @arg @text_list Required. Text to be appended to the /vol/$vfiler_root/etc/group file.
## @description Write text to /vol/$vfiler_root/etc/group file. Overwrites previous contents.
## @returns true if success
## @status public
sub writeGroupFileForVfiler($$@)
{
        my ($filerspec, $vfiler_root, @text_list) = @_;
        my $text = join("\n", @text_list);
        $text .= "\n" unless ($text =~ /\n$/);

	wrfile($filerspec, "/vol/$vfiler_root/etc/group", $text)
			or croak("can't write to /etc/group : $@");
	return(1);
}


## @name ::appendGroupFileForVfiler()
## @synopsis appendGroupFileForVfiler($filerspec, $vfiler_root, @text_list)
## @arg $filerconn Required Filer Spec or Connection Object
## @arg $vfiler_root Required root volume of vfiler.
## @arg @text_list Required. Text to be appended to the /vol/$vfiler_root/etc/group file.
## @description Append text to /vol/$vfiler_root/etc/group file
## @returns true if success
## @status public
sub appendGroupFileForVfiler($$@)
{
        my ($filerspec, $vfiler_root, @text_list) = @_;
        my $newtext = join("\n", @text_list);
        $newtext .= "\n" unless ($newtext =~ /\n$/);

	my $oldtext = rdfile($filerspec, "/vol/$vfiler_root/etc/group");
        my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
        $oldtext = join("\n", @lines);
	croak($@) if ($@);
	$oldtext .= "\n" unless ($oldtext =~ /\n$/);
	wrfile($filerspec, "/vol/$vfiler_root/etc/group", $oldtext . $newtext)
		or croak("can't write to /vol/$vfiler_root/etc/group : $@");
	return(1);
}


## @name ::readPasswdFileForVfiler
## @synopsis readPasswdFileForVfiler($filerconn, $vfiler_root, %args)
#
#  Args:
#
## @arg	filerconn  Required.
## @arg	 vfiler_root  Required. root volume of vfiler 
## @arg Optional Args:
##	users:		If defined, then this is an HASHREF. The users
##			will be written to the HASHREF
## @arg	vfiler_root vfiler root volume.
## @description	Reads /etc/passwd and returns a hashtable represnting the users found there.
## @returns parsed passwd file as HASHREF
sub readPasswdFileForVfiler($$%)
{
  my ($filerspec, $vfiler_root, %args) = @_;
  my (%users) = ();
  my ($line);
  my $usershash;
  if (defined($args{users})) {
	$usershash = $args{users};
  } else {
	$usershash = {};
  }

  return eval {
        my $oldtext = rdfile($filerspec, "/vol/$vfiler_root/etc/passwd");
	if ($oldtext =~ /No such file or directory/) {
		return(undef);
	} else {
        	my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
        	foreach my $line (@lines) {
                	if ($line =~ /^\s*$/ || $line =~ /^\s*#/) {
                        	#print NEWPASSWDFILE $line;
                        	next;
                	}

                	my ($login, $passwd, $uid, $gid, $name, $home, $shell)
                        	= split(/:/, $line);
                	my $user_hash = { login => $login, passwd => $passwd,
                                uid => $uid, gid => $gid, name => $name,
                                home => $home, shell => $shell };
			$$usershash{$login} = $user_hash;
        	}
        	return $usershash;
	}
    };

}


#
# addPasswdEntryForVfiler
#
#   Appends the entry to /vol/$vfiler_root/etc/passwd for the user with the specified
#   login, uid, and gid
#
# Args
#
#	filerconn:		Required.
#
#	login:			Required.
#
#	uid:			Required.
#
#	gid:			Required
#
# Optional Args:
#
#	passwd:			Optional. Default is "x"
#

sub addPasswdEntryForVfiler($$$$$%)
{
  my ($filerspec,$vfiler_root,$login, $uid, $gid, %opts) = @_;
  my (%user) = ();

  return eval {

    $opts{passwd} = "x" unless ($opts{passwd});
    my $oldtext = rdfile($filerspec, "/vol/$vfiler_root/etc/passwd");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    my $newline = "${login}:x:${uid}:${gid}:::";
    push(@lines, $newline);
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerspec, "/vol/$vfiler_root/etc/passwd", $oldtext)
		or croak("can't write to /vol/$vfiler_root/etc/passwd : $@");

    sleep(4);
    return (login => $login, passwd => "x", uid => $uid,
		gid => $gid, name => "", home => "", shell => "");
  };
}


#
# delPasswdEntryForVfiler
#
#   Delete the /vol/$vfiler_root/etc/passwd entry for the user with the specified
#   login.
#
# Args
#
#	filerconn:		Required.
#
#	login:			Required.
#
sub delPasswdEntryForVfiler($$$)
{
  my ($filerspec, $vfiler_root, $login) = @_;

  return eval {

    my $oldtext = rdfile($filerspec, "/vol/$vfiler_root/etc/passwd");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    @lines = grep(!/^${login}:/, @lines);
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerspec, "/vol/$vfiler_root/etc/passwd", $oldtext)
		or croak("can't write to /vol/$vfiler_root/etc/passwd : $@");
    sleep(4);
  };
}


#
# delPasswdListForVfiler
#
#   Delete specified users from /vol/$vfiler_root/etc/passwd
#
# Args
#
#	filerconn:		Required.
#
#	userhash:		Required. HASHREF for list of users
#				to be deleted
#
sub delPasswdListForVfiler($$$)
{
  my ($filerconn, $vfiler_root, $userhash) = @_;

  return eval {

    my $oldtext = rdfile($filerconn, "/vol/$vfiler_root/etc/passwd");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    my @newlines = ();
    foreach (@lines) {
        my ($login) = (/^([^:]+):/);
        push(@newlines, $_) unless (exists($userhash->{$login}));
    }
    $oldtext = join("\n", @newlines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/vol/$vfiler_root/etc/passwd", $oldtext)
		or croak("can't write to /vol/$vfiler_root/etc/passwd : $@");
    sleep(4);
  };
}


#
# addPasswdListForVfiler
#
#	Adds entries to /vol/$vfiler_root/etc/passwd for the users in userhash
#
# Args
#
#	filerconn:		Required.
#
#	userhash:		Required. HASHREF for list of users
#				to be added
#
# Optional Args:
#
#	cifs_user:		Optional. Default is 0. If '1', then
#				use the 'winlogin' and 'winpasswd' instead of
#				'login', and 'passwd' fields for the users.
#

sub addPasswdListForVfiler($$$%)
{
  my ($filerspec, $vfiler_root, $userhash, %opt) = @_;

  return eval {
    my @lines;
    my $oldtext = rdfile($filerspec, "/vol/$vfiler_root/etc/passwd");
    if ($oldtext !~ /No such file or directory/) {
    	@lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    }
    my @keys = keys %$userhash;
    foreach my $u (@keys) {
	    my $info = $userhash->{$u};
	    my ($login, $passwd);
	    if ($opt{cifs_user}) {
	    	$login = $info->{winlogin};
	    	$passwd = $info->{winpasswd} || $main::CIFS_PASSWD || "cifspass123";
	    } else {
	    	$login = $info->{login};
	    	$passwd = $info->{passwd};
	    }
	    my $uid = $info->{uid};
	    my $gid = $info->{gid};
	    my $name = $info->{name} || "";
	    my $home = $info->{home} || "";
	    my $shell = $info->{shell} || "";

	    my $newline = join(":", $login, $passwd, $uid, $gid, $name, $home, $shell);
	    push(@lines, $newline);
    }

    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerspec, "/vol/$vfiler_root/etc/passwd", $oldtext)
	or croak("can't write to /vol/$vfiler_root/etc/passwd : $@");
    sleep(4);

  }
}


#
#  readGroupFile
#
#	Reads /vol/$vfiler_root/etc/group and returns a hashtable represnting groups found there.
#
#  Args:
#
#	filerconn:		Required
#
#  Optional Args:
#
#	groups:			If defined, then this is an HASHREF. The groups
#				will be written to the HASHREF
#
sub readGroupFileForVfiler($$%)
{

  my ($filerspec, $vfiler_root, %args) = @_;
  my (%groups) = ();
  my ($line);
  my $groupshash;
  if (defined($args{groups})) {
	$groupshash = $args{groups};
  } else {
	$groupshash = {};
  }
  return eval {
	my $oldtext = rdfile($filerspec, "/vol/$vfiler_root/etc/group");
	if ($oldtext =~ /No such file or directory/) {
		return(undef);
	} else {
		my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
		foreach my $line (@lines) {
			if ($line =~ /^\s*$/ || $line =~ /^\s*#/) {
	    			next;
			}
			my ($group, $passwd, $gid, $members)
				= split(/:/, $line);
			my $grp_hash = { group => $group, passwd => $passwd, 
				gid => $gid, members => [split(/,/, $members)]};
			$$groupshash{$group} = $grp_hash;
    		}
    	return( $groupshash );
	}
  }
}


#
# addGroupEntryForVfiler
#
#   Appends the entry to /vol/$vfiler_root/etc/group for the group with the specified
#   name and gid.
#
# Args
#
#	filerconn:		Required.
#
#	name:			Required.
#
#	gid:			Required
#
# Optional args:
#
#	members:		Optional. For "members" field, default is
#				empty string
#
sub addGroupEntryForVfiler($$$$;$)
{
  my ($filerconn, $vfiler_root, $name, $gid, $members) = @_;

  $members = "" if (!defined($members));

  return eval {

    my $oldtext = rdfile($filerconn, "/vol/$vfiler_root/etc/group");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    push(@lines, "${name}:x:${gid}:${members}");
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/vol/$vfiler_root/etc/group", $oldtext)
		or croak("can't write to /vol/$vfiler_root/etc/group : $@");

    return (group => $name, passwd => "x", gid => $gid, members => $members);
  };
}


#
# delGroupEntryForVfiler
#
#   Delete the existing /vol/$vfiler_root/etc/group entry for the group with
#   the specified name
#
# Args
#
#	filerconn:		Required.
#
#	name:			Required.
#
sub delGroupEntryForVfiler($$$)
{
  my ($filerconn, $vfiler_root, $group) = @_;

  return eval {

    my $oldtext = rdfile($filerconn, "/vol/$vfiler_root/etc/group");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    @lines = grep(!/^${group}:/, @lines);
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/vol/$vfiler_root/etc/group", $oldtext)
		or croak("can't write to /vol/$vfiler_root/etc/group : $@");
  };
}


#
# delGroupListForVfiler
#
#   Delete specified groups from /vol/$vfiler_root/etc/group
#
# Args
#
#	filerconn:		Required.
#
#	grouphash:		Required. HASHREF for list of groups
#				to be deleted
#
sub delGroupListForVfiler($$$)
{
  my ($filerconn, $vfiler_root, $grouphash) = @_;

  return eval {

    my $oldtext = rdfile($filerconn, "/vol/$vfiler_root/etc/group");
    my @lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    my @newlines = ();
    foreach (@lines) {
        my ($gname) = (/^([^:]+):/);
        push(@newlines, $_) unless (exists($grouphash->{$gname}));
    }
    $oldtext = join("\n", @newlines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/vol/$vfiler_root/etc/group", $oldtext)
		or croak("can't write to /vol/$vfiler_root/etc/group : $@");
  };
}


#
# addGroupListForVfiler
#
#	Adds entries to /vol/$vfiler_root/etc/group for the groups in grouphash
#
# Args
#
#	filerconn:		Required.
#
#	vfiler_root:		Required.
#
#	grouphash:		Required. HASHREF for list of groups
#				to be added
#
sub addGroupListForVfiler($$$)
{
  my ($filerconn, $vfiler_root, $grouphash) = @_;

  return eval {
    my $oldtext = rdfile($filerconn, "/vol/$vfiler_root/etc/group");
    my @lines;
    if ($oldtext !~ /No such file or directory/) {
    		@lines = grep(!/^rdfile/, split(/\n/, $oldtext));
    }
     
    my @keys = keys %$grouphash;
    foreach my $g (@keys) {
	    my $info = $grouphash->{$g};
	    my $group = $info->{group};
	    my $passwd = $info->{passwd} || "x";
	    my $gid = $info->{gid};
	    my $members = $info->{members} || "";
	    push(@lines, "${group}:${passwd}:${gid}:${members}");
    }
    $oldtext = join("\n", @lines);
    $oldtext .= "\n" unless ($oldtext =~ /\n$/);
    wrfile($filerconn, "/vol/$vfiler_root/etc/group", $oldtext)
		or croak("can't write to /etc/group : $@");
  }
}


# ----------------------------
# Supporting local subroutines 
# ----------------------------


sub _array_from_hash($)
{
    my ($hashref) = @_;
    my $arrayref = [];
    while (my ($key, $val) = each(%$hashref)) {
	push(@$arrayref, $val);
    }
    return($arrayref);
}


#
#   _getNextUIDForVfiler
#
#  Args:
#
#	nusers:		Required. Number of new UIDs to create
#
#   Optional args:
#
#	min_uid		Default is 10000
#	max_uid		Default is 0x7fffffff
#
#  If min_uid is not specified, and max_uid is specified, and is less than
#  10000, then the greatest of 1000, 100, 10 which is less than max_uid
#  is used as the value for min_uid
#
sub _getNextUIDForVfiler($$%)
{
  my ($nusers, $current_users_hash, %args) = @_;
  #if (!$current_users_hash) {
  #        die("_getNextUIDForVfiler  current_users_hash is undefined");
  #}
  if (!$nusers) {
	return(undef);
  }
  my ($max_uid) = (defined($args{max_uid}) ? $args{max_uid} : 0x7fffffff);
  my ($min_uid);
  if (defined($args{min_uid})) {
	$min_uid = $args{min_uid};
  } elsif ($max_uid > 10000) {
	$min_uid = 10000;
  } elsif ($max_uid > 1000) {
	$min_uid = 1000;
  } elsif ($max_uid > 100) {
	$min_uid = 100;
  } else {
	$min_uid = 10;
  }

  my (@uids) = ($min_uid-1, $max_uid);
  return eval {
    foreach my $login (keys %$current_users_hash) {
	my $u = $$current_users_hash{$login}->{uid};
	push @uids, $u if ($u >= $min_uid);
    }
    @uids = sort {$a <=> $b} @uids;

    #  Look for gaps in uid list
    my ($next_uid) = 0;
    my $found = 0;
    for (my $i = 0; $i < $#uids && !$found; $i++) {
	if ($uids[$i+1] - $uids[$i] > $nusers) {
	  $next_uid = $uids[$i] + 1;
          $found = 1;
    	}
    }
    if ($found) {
	return($next_uid);
    } else {
	return(undef);
    }
  }
}

#
#   Exported version of _getNextUIDForVfiler
#
sub getNextUIDForVfiler($$%)
{
	my ($fc, $vfiler_root,%args) = @_;
	my $nusers = $args{nusers} || 1;
	delete($args{nusers});
	return( _getNextUIDForVfiler($nusers, readPasswdFileForVfiler($fc,$vfiler_root), %args) );
}

#
#   _getNextGIDForVfiler
#
#  Args:
#
#	ngroups:	Required. Number of new GIDs to create
#
#   Optional args:
#
#	min_gid		Default is 10000
#	max_gid		Default is 0x7fffffff
#
#  If min_gid is not specified, and max_gid is specified, and is less than
#  10000, then the greatest of 1000, 100, 10 which is less than max_gid
#  is used as the value for min_gid
#
sub _getNextGIDForVfiler($$%)
{
  my ($ngroups, $current_groups_hash, %args) = @_;

  if (!$ngroups) {
	return(undef);
  }
  my ($max_gid) = (defined($args{max_gid}) ? $args{max_gid} : 0x7fffffff);
  my ($min_gid);
  if (defined($args{min_gid})) {
	$min_gid = $args{min_gid};
  } elsif ($max_gid > 10000) {
	$min_gid = 10000;
  } elsif ($max_gid > 1000) {
	$min_gid = 1000;
  } elsif ($max_gid > 100) {
	$min_gid = 100;
  } else {
	$min_gid = 10;
  }

  my (@gids) = ($min_gid-1, $max_gid);
  return eval {
    foreach my $group (keys %$current_groups_hash) {
	my $u = $$current_groups_hash{$group}->{gid};
	push @gids, $u if ($u >= $min_gid);
    }
    @gids = sort {$a <=> $b} @gids;

    #  Look for gaps in gid list
    my ($next_gid) = 0;
    my $found = 0;
    for (my $i = 0; $i < $#gids && !$found; $i++) {
	if ($gids[$i+1] - $gids[$i] > $ngroups) {
	  $next_gid = $gids[$i] + 1;
          $found = 1;
    	}
    }
    if ($found) {
	return($next_gid);
    } else {
	return(undef);
    }
  }
}

#
#   Exported version of _getNextGIDForVfiler
#
sub getNextGIDForVfiler($$%)
{
	my ($fc, $vfiler_root, %args) = @_;
	my $ngroups = $args{ngroups} || 1;
	delete($args{ngroups});
	return( _getNextGIDForVfiler($ngroups, readGroupFileForVfiler($fc,$vfiler_root), %args) ); 
}



#
# _findInList
#
#	Searches a given ARRAYREF for matching members, and returns
#	any matches.
#
# Args:
#
#	list:		Required. ARRAYREF from which matching members
#			are to be chosen
#
#	args:		Key, value pairs which are used to select
#			matching members

sub _findInList($%)
{
	my ($arraylist, %args) = @_;
	my (@keys) = keys(%args);
	my @matches = ();
	for my $key (@keys) {
		delete($args{$key}) unless (defined($args{$key}));
	}
	@keys = keys(%args);
	return() unless (scalar(@keys));
	for my $elt (@$arraylist) {
		my $match = 1;
		my $args_ref = { %args };
		while ( my ($key, $value) = each(%args)) {
			if ($elt->{$key} ne $value) {
				$match = 0;
				last;
			}
		}
		if ($match) {
			push(@matches, $elt);
		}
	}
	return(@matches);
}



#
# _findInHashList
#
#	Searches a given 'hash list' for matching members, and returns
#	any matches. The 'hash list' format is that returned by for example,
#	readPasswdFileForVfiler, readGroupFile, createUsersForVfiler, createGroupsForVfiler, etc.
#
# Args:
#
#	hashlist:	Required. HASHREF from which matching members
#			are to be chosen
#
#	args:		Key, value pairs which are used to select
#			matching members

sub _findInHashList($%)
{
	my ($hashlist, %args) = @_;
	my (@keys) = keys(%args);
	my @matches = ();
	my $arraylist = _array_from_hash($hashlist);
        return ( _findInList($arraylist, %args) );
}


#
# _findUserForVfiler
#
#	Returns users from current users which match criterea determined
#	by %args
#
# Args:
#
#	current_users_hash:	Required. HASHREF from which matching users
#				are to be chosen
#
#	args:			Key, value pairs which are used to select
#				matching users

sub _findUserForVfiler($%)
{
	my ($current_users_hash, %args) = @_;
	return ( _findInHashList($current_users_hash, %args) );
}


#
#  Exported version of _findUserForVfiler
#
#  args:
#	filerspec	required
#
#	either:
#		(1)	name: 'login' (name) of user
#
#	or
#		(2)	%args: list of key-value pairs describing user
#
sub findUserForVfiler($$%)
{
	my $filerspec = shift;
	my $vfiler_root = shift;
        my $fc = make_conn($filerspec);
	if (@_ == 1) {
		my ($name) = @_;
		return ( _findUserForVfiler(readPasswdFileForVfiler($fc,$vfiler_root), login=>$name) );
	} else {
		my (%args) = @_;
		return ( _findUserForVfiler(readPasswdFileForVfiler($fc,$vfiler_root), %args) );
	}
}


#
# _findGroupForVfiler
#
#	Returns groups from current_groups which match criterea determined
#	by %args
#
# Args:
#
#	current_groups_hash:	Required. HASHREF from which matching groups
#				are to be chosen
#
#	args:			Key, value pairs which are used to select
#				matching groups

sub _findGroupForVfiler($%)
{
	my ($current_groups_hash, %args) = @_;
	return ( _findInHashList($current_groups_hash, %args) );
}


#
#  Exported version of _findGroupForVfiler
#
#  args:
#	filerspec	required
#
#	either:
#		(1)	name: name of group
#
#	or
#		(2)	%args: list of key-value pairs describing group
#
sub findGroupForVfiler($%)
{
	my $filerspec = shift;
        my $fc = make_conn($filerspec);
	if (@_ == 1) {
		my ($name) = @_;
		return ( _findGroupForVfiler(readGroupFile($fc), group=>$name) );
	} else {
		my (%args) = @_;
		return ( _findGroupForVfiler(readGroupFile($fc), %args) );
	}
}


1;
