###############################################################################################
## Id: //depot/prod/test/main/lib/VFILER/storagemotion.pm#1$
##
## Copyright (c) 2008-2009 NetApp, Inc.
## All rights reserved.
################################################################################################
##
## Storage motion utilities Package
##
## Provides routines for Storagemotion project i.e vfiler migrate start,prepare and complete zapis 
## 
##
################################################################################################
## @summary Routines for vfiler storage motion
##
## @description
##	 functions for the vfiler storage test automation
## @dependencies
##	 use telnet to connect your filer, not rsh
##
##       sync snapmirror between volumes of HA-pair won't work.
##	 Please don't use HA-pair as input parameters for the below routines.
##
## @author dl-multistore-test
##
##  @pod here
##
package VFILER::storagemotion;

use strict;
use warnings;
use Tharn;
use base qw(Exporter);
use SayWrap qw(sayEOF);
use Admin::License qw(is_licensed);
use SPARSE::Util::Logger;
use Params::Validate qw(:all);
use CLI::FileIO qw(rdfile wrfile);
 BEGIN
 {
	 our @EXPORT = qw(
			 zapi_vfiler_migrate_start
			 zapi_vfiler_migrate_prepare
			 zapi_vfiler_migrate_complete
			 vfiler_snapmirror_conf_change
			 zapi_vfiler_migrate_cancel
			 zapi_vfiler_migrate 
		 );
}

## @name Zapi_vfiler_migrate_start
## @description
## vfiler migrate start zapi is provided through this routine
##
## Returns 0 on failure
## Returns zapi output on success
## Args
## host:		Required: Destination filer of the Vfiler migrate operation
## user:		Required: remote login of the Destination filer
## password:		Required: remote login password of the Destination filer
##
## Optional Args:
##  remote_vfiler:	optional: name of the source vfiler
##  remote_filer :	optional: name of the source filer.
##  remote_login :	optional: remote login of the Source filer
##  remote_passwd:	optional: remote login password of the Source filer.
##  ifname	 :	optional: Interface to be used for vfiler migrate.
##  ipaddr	 :	optional: Ipaddress to be used for vfiler migrate.
##  netmask      :	optional: Netmask to be used for the vfiler migate.
##  use_secure_channel: optional: default is false i.e vfiler migrate happens through rsh.
##				  true i.e vfiler migrate happens through ssl.
##  sm_not_intialized:	optional: default is false i.e baseline tranfer for snapmirror relationship is required.
#				  true i.e baseline tranfer for snapmirror relationship is not required.
##
##@usage:
## zapi_vfiler_migrate_start($host,$user,$password,
##			     remote_vfiler=>"$VFILER_NAME",
##			     remote_filer=>"FILER_SRC",
##			     remote_login=>$username,
##			     remote_passwd=>$passwd,
##			     ifname=>$VFILER_IF_TO_USE,
##			     ipaddr=>$VFILER_IPADDR,
##			     netmask=>$VFILER_NETMASK,
##			     use_secure_channel=>"true",
##			     sm_not_initialize=>"true");
		
sub zapi_vfiler_migrate_start {

	my ($host,$user,$password,%args)=@_;
	
	my $remote_vfiler=$args{remote_vfiler};
	my $remote_filer=$args{remote_filer};
	my $remote_login = $args{remote_login} || "root";
	my $remote_passwd = $args{remote_passwd} || "";

	my $ifname = $args{ifname};
	my $ipaddr=$args{ipaddr};
	my $netmask=$args{netmask};
	my $use_secure_channel=$args{use_secure_channel} ||"false";
	my $sm_not_initialize=$args{sm_not_initialize}||"false";
    
	if(!defined($remote_vfiler or $remote_filer)) {
		logresult("FAIL","Remote vfiler and its location  not specified properly");
		return 0;
	}
  
	my $api = "vfiler-migrate-start";
	logcomment("vfiler-migrate-start zapi execution started"); 
 	my $server = new ZapiTest($host, $user, $password);
 	if ( !defined $server ) {
        logresult ("FAIL","Unable to connect to $host");
        return 0;
	}
	my $e = new NaElement($api);
    
	my $addr = new NaElement("ipaddrs");
	my $unit = new NaElement("ipaddr-info");
	$unit->child_add(new NaElement("interface", $ifname));
	$unit->child_add(new NaElement("ip-address", $ipaddr));
	$unit->child_add(new NaElement("netmask", $netmask));
   	$addr->child_add($unit);
	$e->child_add($addr);

	my $remote_authentication_info=new NaElement("remote-authentication-info");
	$unit = new NaElement("authentication-info");
	$unit->child_add(new NaElement("password",$remote_passwd) );
	$unit->child_add(new NaElement("username", $remote_login));
	$remote_authentication_info->child_add($unit);
	$e->child_add($remote_authentication_info);
    
        my $remote_vfiler_location=new NaElement("remote-vfiler-location");
	$unit = new NaElement("vfiler-location");
	$unit->child_add(new NaElement("filer", $remote_filer));
	$unit->child_add(new NaElement("vfiler", $remote_vfiler));
	$remote_vfiler_location->child_add($unit);
	$e->child_add($remote_vfiler_location);
        $e->child_add(new NaElement("use-secure-command-channel", $use_secure_channel));
	$e->child_add(new NaElement("snapmirror-not-initialize",$sm_not_initialize));
	##@ step to invoke the zapi


	logcomment($e->sprint_pretty());

	my $o = $server->invoke_elem($e,"any");
	my $errno = $o->results_errno();
	my $reason = $o->results_reason();
	my $out = $o->toString();
        logcomment("$out");
        if (defined $reason) {
	return (0,$out);
	}
	else{
	logresult("PASS","vfiler-migrate-start zapi executed successfully");
	}
				
	return (1,$out);
}	


## @name zapi_vfiler_migrate_prepare
## @description
## vfiler migrate prepare zapi is provided through this routine
##
## Returns 0 on failure
## Returns zapi output on success
## Args
## host:                Required: Destination filer of the Vfiler migrate operation
## user:                Required: remote login of the Destination filer
## password:            Required: remote login password of the Destination filer
##
## Optional Args:
##  remote_vfiler:      optional: name of the source vfiler
##  remote_filer :      optional: name of the souce filer.
##  remote_login :      optional: remote login of the Source filer
##  remote_passwd:      optional: remote login password of the Source filer.
##  use_secure_channel: optional: default is false i.e vfiler migrate prepare happens through rsh.
##                                true i.e vfiler migrate prepare happens through ssl.
##
##
##@usage:
## zapi_vfiler_migrate_prepare($host,$user,$password,
##                            remote_vfiler=>"$VFILER_NAME",
##                            remote_filer=>"FILER_SRC",
##                            remote_login=>$username,
##                            remote_passwd=>$passwd,
##                            use_secure_channel=>"true");
                            

		
sub zapi_vfiler_migrate_prepare {

	my ($host,$user,$password,%args)=@_;
	
	my $remote_vfiler=$args{remote_vfiler};
	my $remote_filer=$args{remote_filer};
	my $remote_login = $args{remote_login} || "root";
	my $remote_passwd = $args{remote_passwd} || "";
	my $use_secure_channel=$args{use_secure_channel} ||"false";
    
	if(!defined($remote_vfiler or $remote_filer)) {
		logresult("FAIL","Remote vfiler and its location  not specified properly");
		return 0;
	}
  
	my $api = "vfiler-migrate-prepare";
	logcomment("vfiler-migrate-prepare zapi execution started"); 
 	my $server = new ZapiTest($host, $user, $password);
 	if ( !defined $server ) {
        logresult ("FAIL","Unable to connect to $host");
        return 0;
	}
	my $e = new NaElement($api);
 	my $remote_authentication_info=new NaElement("remote-authentication-info");
	my $unit = new NaElement("authentication-info");
	$unit->child_add(new NaElement("password",$remote_passwd) );
	$unit->child_add(new NaElement("username", $remote_login));
	$remote_authentication_info->child_add($unit);
	$e->child_add($remote_authentication_info);
        
	my $remote_vfiler_location=new NaElement("remote-vfiler-location");
	$unit = new NaElement("vfiler-location");
	$unit->child_add(new NaElement("filer", $remote_filer));
	$unit->child_add(new NaElement("vfiler", $remote_vfiler));
	$remote_vfiler_location->child_add($unit);
	$e->child_add($remote_vfiler_location);
        $e->child_add(new NaElement("use-secure-command-channel", $use_secure_channel));
        ##@step to invoke the zapi

	logcomment($e->sprint_pretty());


	my $o = $server->invoke_elem($e,"any");
	my $errno = $o->results_errno();
	my $reason = $o->results_reason();
	my $out = $o->toString();
        logcomment("$out");
        if (defined $reason) {
	return (0,$out);
	}
	else{
	logresult("PASS","vfiler-migrate-prepare zapi executed successfully");
	}
	return (1,$out);
}

## @name zapi_vfiler_migrate_complete
## @description
## vfiler migrate complete zapi is provided through this routine
##
## Returns 0 on failure
## Returns zapi output on success
## Args
## host:                Required: Destination filer of the Vfiler migrate operation
## user:                Required: remote login of the Destination filer
## password:            Required: remote login password of the Destination filer
##
## Optional Args:
##  remote_vfiler:      optional: name of the source vfiler
##  remote_filer :      optional: name of the souce filer.
##  remote_login :      optional: remote login of the Source filer
##  remote_passwd:      optional: remote login password of the Source filer.
##  use_secure_channel: optional: default is false i.e vfiler migrate happens through rsh.
##                                true i.e vfiler migrate happens through ssl.
## migrate_transparent: optional: default is false i.e transparent-migration=false.
##                                 true i.e transparent-migration=true.
## max_cutover_time:   optional: default value for the cutover time is 120seconds in storagemotion.
##
##@usage:
### zapi_vfiler_migrate_complete($host,$user,$password,
##				remote_vfiler=>"$VFILER_NAME",
##				remote_filer=>"FILER_SRC",
##				remote_login=>$username,
##                              remote_passwd=>$passwd,
##                              use_secure_channel=>"true",
##                              migrate_transparent=>"true",
##				max_cutover_time=>"120");

		
sub zapi_vfiler_migrate_complete {

	my ($host,$user,$password,%args)=@_;
	
	my $remote_vfiler=$args{remote_vfiler};
	my $remote_filer=$args{remote_filer};
	my $remote_login = $args{remote_login} || "root";
	my $remote_passwd = $args{remote_passwd} || "";
	
	my $use_secure_channel=$args{use_secure_channel} ||"false";
	my $migrate_transparent=$args{migrate_transparent} ||"false";
	my $max_cutover_time=$args{max_cutover_time}||"120";
    
	if(!defined($remote_vfiler or $remote_filer)) {
		logresult("FAIL","Remote vfiler and its location  not specified properly");
		return 0;
	}
	logcomment("vfiler-migrate-complete zapi execution started"); 
	my $api = "vfiler-migrate-complete";
  	my $server = new ZapiTest($host, $user, $password);
 	if ( !defined $server ) {
        logresult ("FAIL","Unable to connect to $host");
        return 0;
	}
	my $e = new NaElement($api);

	my $remote_authentication_info=new NaElement("remote-authentication-info");
	my $unit = new NaElement("authentication-info");
	$unit->child_add(new NaElement("password",$remote_passwd) );
	$unit->child_add(new NaElement("username", $remote_login));
	$remote_authentication_info->child_add($unit);
	$e->child_add($remote_authentication_info);
    
	my $remote_vfiler_location=new NaElement("remote-vfiler-location");
	$unit = new NaElement("vfiler-location");
	$unit->child_add(new NaElement("filer", $remote_filer));
	$unit->child_add(new NaElement("vfiler", $remote_vfiler));
	$remote_vfiler_location->child_add($unit);
	$e->child_add($remote_vfiler_location);
    
	$e->child_add(new NaElement("use-secure-command-channel", $use_secure_channel));
	$e->child_add(new NaElement("migrate-transparent",$migrate_transparent));
	$e->child_add(new NaElement("max-cutover-time",$max_cutover_time));

	logcomment($e->sprint_pretty());
    
	##@step to invoke the zapi
	my $o = $server->invoke_elem($e,"any");
	my $errno = $o->results_errno();
	my $reason = $o->results_reason();
	my $out = $o->toString();
        if (defined $reason) {
	   return (0,$out);
	}
	else{
		logresult("PASS","vfiler-migrate-complete Zapi executed successfully");
	}
	return (1,$out);
}
## @name vfiler_snapmirror_conf_change
##
##	 Edit entries of the /vol/$vfilerroot/etc/snapmirror.conf i.e vsm relationships to semi-sync mode
##
##	 this routine will work in conjunction with vfiler-migrate-start routines. You can not run this in
##	 isolation.
##
##	 Returns 0 on failure
##	 Returns 1 on success
 
## Args
## $fc:		Required: connection spec of the remote filer.
##
## $src_filer:  Required: name of the source filer. 
##
## $remote_filer:Required: name of the remote filer.
##
## $vfilervols:	Required: Reference to the array of the volumes of the migrating vfiler. 
##
## Optional Args:
## syncmode:	optional: options for sync,semisync or async
##		default : async mode with '-' option
## visibilityinterval: optional: time interval to check the snapmirror updates
##
## @examples:
##  vfiler_snapmirror_conf_change($fc,$src_filer,$remote_filer,$vfilervols,
##							       syncmode=>"in-sync",
#							       visibilityinterval=>"600m");
## 
sub  vfiler_snapmirror_conf_change{
	my ($fc,$src_filer,$remote_filer,$vols,%args)=@_;
	my $syncmode=$args{syncmode}||"-";
	my $visibiltyinterval=$args{visibilityinterval}||"-";
	
	my ($snapmirrorconf)=rdfile($fc,"/etc/snapmirror.conf");
	if(!$snapmirrorconf){
		logresult("FAIL","Snapmirror.conf for the filer is missing");
		return 0;
	}
	my $newsnapmirrorconf="";
	my @vols=@$vols;
        foreach my $vol(@vols){
		print"volume is $vol\n";
		my(undef,undef,$status)=say($fc,"vol size $vol",'-swallow');
		$status=~/(\w+)$/;
		print"volume size is $status\n";
		my $newentry="$src_filer:$vol\t$remote_filer:$vol - $syncmode";
		 $newsnapmirrorconf.="$newentry\n";
	 }
	wrfile($fc,"/etc/snapmirror.conf",$newsnapmirrorconf); 
	foreach my $vol(@vols){
		my $flag=0;
		my $count=0;
		do{
			my (undef,undef,$status)=say($fc,"snapmirror status $src_filer:$vol",'-swallow');
			logcomment("snapmirror entries for the volume reflects as $status");
			if($status !~ /In-sync/i){
				logcomment("waiting for the $vol snapmirror state change during vfiler migration to In-sync");
				sleep 60;
				$count++;	
				if(($count > 5) & ($count <20)){
					logcomment("Snapmirror updates has taken more than 5 minutes");
					logcomment("check for the following error messages");
					logcomment("Synchronous SnapMirror cannot run between two flexible volumes on the same filer or HA pair");
				}
				elsif($count >20){
					logcomment("snapmirror updates has taken more than 20 minutes");
					logresult("FAIL","Sync snapmirror updates has taken more than usual time");
					return 0;
				}
				else{
					logcomment("waiting for the $vol snapmirror state change during vfiler migration to In-sync");
				}
					
			}
			else{
				logcomment("snapmirror status of $vol changed to  In-sync");
				$flag=1;
			}
				
		}until($flag!=0);
	}
		sleep(60);
	
	foreach my $vol(@vols){
		logcomment("checking for persistent In-sync state for the volumes");
		my (undef,undef,$status)=say($fc,"snapmirror status $src_filer:$vol",'-swallow');
		if($status !~ /In-sync/i){
			logresult("FAIL","In-sync status of $src_filer:$vol is not persistent");
			return 0;
			}
		}
	return 1; 
}


## @name zapi_vfiler_migrate_cancel
## Returns 0 on failure^M
## Returns zapi output on success^M
## Args^M
## host:                Required: Destination filer of the Vfiler migrate operation^M
## user:                Required: remote login of the Destination filer^M
## password:            Required: remote login password of the Destination filer^M
##^M
## Optional Args:^M
##  remote_vfiler:      optional: name of the source vfiler^M
##  remote_filer :      optional: name of the souce filer.^M
##  use_secure_channel: optional: default is false i.e vfiler migrate happens through rsh.^M
##                                true i.e vfiler migrate happens through ssl.^M
##@usage:^M
### zapi_vfiler_migrate_cancel($host,$user,$password,^M
##                              remote_vfiler=>"$VFILER_NAME",^M
##                              remote_filer=>"FILER_SRC",^M
##                              use_secure_channel=>"true"^);M

sub zapi_vfiler_migrate_cancel {

         my ($host,$user,$password,%args)=@_;
         my $remote_vfiler=$args{remote_vfiler};
         my $remote_filer=$args{remote_filer};
         my $use_secure_channel=$args{use_secure_channel} ||"false";
         my $force=$args{force}||"false";
      
        if(!defined($remote_vfiler or $remote_filer)) {
                logresult("FAIL","Remote vfiler and its location  not specified properly");
                return 0;
        }
        logcomment("vfiler-migrate-cancel zapi execution started"); 
        my $api = "vfiler-migrate-cancel";
        my $server = new ZapiTest($host, $user, $password);
        if ( !defined $server ) {
        logresult ("FAIL","Unable to connect to $host");
        return 0;
        }
        my $e = new NaElement($api);       
     
        my $remote_vfiler_location=new NaElement("remote-vfiler-location");
        my $unit = new NaElement("vfiler-location");
        $unit->child_add(new NaElement("filer", $remote_filer));
        $unit->child_add(new NaElement("vfiler", $remote_vfiler));
        $remote_vfiler_location->child_add($unit);
        $e->child_add($remote_vfiler_location);
     
       $e->child_add(new NaElement("use-secure-command-channel", $use_secure_channel));
    
       $e->child_add(new NaElement("force", $force));
    
     
       logcomment($e->sprint_pretty());
        my $o = $server->invoke_elem($e,"any");
        my $errno = $o->results_errno();
        my $reason = $o->results_reason();
        my $out = $o->toString();
        if (defined $reason) {
           return (0,$out);
        }
        else{
                logresult("PASS","vfiler-migrate-cancel Zapi executed successfully");
        }
        return (1,$out);
}

## @name zapi_vfiler_migrate
## @description
### vfiler migrate start zapi is provided through this routine
###
### Returns 0 on failure
### Returns zapi output on success
### Args
### host:                Required: Destination filer of the Vfiler migrate operation
### user:                Required: remote login of the Destination filer
### password:            Required: remote login password of the Destination filer
###
### Optional Args:
###  remote_vfiler:      optional: name of the source vfiler
###  remote_filer :      optional: name of the source filer.
###  remote_login :      optional: remote login of the Source filer
###  remote_passwd:      optional: remote login password of the Source filer.
###  ifname       :      optional: Interface to be used for vfiler migrate.
###  ipaddr       :      optional: Ipaddress to be used for vfiler migrate.
###  netmask      :      optional: Netmask to be used for the vfiler migate.
###  use_secure_channel: optional: default is false i.e vfiler migrate happens through rsh.
###                                true i.e vfiler migrate happens through ssl.
###  force  optional:   default is false
###			true i.e the vfiler should be migrated even under various conditions which would normally prevent a migration are encountered.
###  migration_method optional:  default is snapmirror 
###			         other option is nocopy
###				 
###
###@usage:
### zapi_vfiler_migrate($host,$user,$password,
###                           remote_vfiler=>"$VFILER_NAME",
###                           remote_filer=>"FILER_SRC",
###                           remote_login=>$username,
###                           remote_passwd=>$passwd,
###                           ifname=>$VFILER_IF_TO_USE,
###                           ipaddr=>$VFILER_IPADDR,
###                           netmask=>$VFILER_NETMASK,
###                           use_secure_channel=>"true",
##			      force=>"false",
##			      migration_method=>"snapmirror"
###                           );
sub zapi_vfiler_migrate {
	my ($host,$user,$password,%args)=@_;

	my $remote_vfiler=$args{remote_vfiler};
	my $remote_filer=$args{remote_filer};
	my $remote_login = $args{remote_login} || "root";
	my $remote_passwd = $args{remote_passwd} || "";
	my $ifname = $args{ifname};
	my $ipaddr=$args{ipaddr};
	my $netmask=$args{netmask};
	my $use_secure_channel=$args{use_secure_channel} ||"false";
	my $migration_method=$args{migration_method}||"snapmirror";
	my $force=$args{force}||"false";
	if(!defined($remote_vfiler or $remote_filer)) {
	logresult("FAIL","Remote vfiler and its location  not specified properly");
	return 0;
	}
	if(!defined($remote_vfiler or $remote_filer)) {
	logresult("FAIL","Remote vfiler and its location  not specified properly");
	return 0;
	}
	my $api = "vfiler-migrate";
	logcomment("vfiler-migrate zapi execution started");
	my $server = new ZapiTest($host, $user, $password);
	if ( !defined $server ) {
	logresult ("FAIL","Unable to connect to $host");
	return 0;
	}
	my $e = new NaElement($api);
	my $addr = new NaElement("ipaddrs");
	my $unit = new NaElement("ipaddr-info");
	$unit->child_add(new NaElement("interface", $ifname));
	$unit->child_add(new NaElement("ip-address", $ipaddr));
	$unit->child_add(new NaElement("netmask", $netmask));
	$addr->child_add($unit);
	$e->child_add($addr);
	$e->child_add(new NaElement("migration-method",$migration_method));
	$e->child_add(new NaElement("remote-filer",$remote_filer));
	$e->child_add(new NaElement("vfiler",$remote_vfiler));
	$e->child_add(new NaElement("use-secure-command-channel", $use_secure_channel));
	$e->child_add(new NaElement("force", $force));
	$e->child_add(new NaElement("username",$remote_login));
	$e->child_add(new NaElement("password",$remote_passwd));
	logcomment($e->sprint_pretty());
	my $o = $server->invoke_elem($e,"any");
	my $errno = $o->results_errno();
	my $reason = $o->results_reason();
	my $out = $o->toString();
	if (defined $reason) {
	return (0,$out);
	}
	else{
	logresult("PASS","vfiler-migrate-cancel Zapi executed successfully");
	}
	return (1,$out);
     
}


1;
