#
# VLDB::Client
#
# A VLDB client module for programmatic interaction
#

use strict;
use warnings;

package VLDB::Client;

use TracePkg;
use Carp qw(cluck);
use Data::Dumper;
use Tharn;
use SayWrap qw(sayEOF);

# from vldb/ng_src/vldb/vldb_defs.h (on 7/29/05)
my @VldbErrorCodes = (
    ['VLDB_SUCCESS'                               , 'success'], 
    ['VLDB_VOLNAME_EXISTS_INVSERVER'              , 'vol name already in the vserver'], 
    ['VLDB_INVALID_NAME'                          , 'specified name does not exist'], 
    ['VLDB_INVALID_VOL_TYPE'                      , 'undefined volume type specified'], 
    ['VLDB_MISSING_VOL_TYPE'                      , 'missing volume type'], 
    ['VLDB_INVALID_MASK'                          , 'something wrong with input mask'], 
    ['VLDB_VALUE_REQUIRED'                        , 'required value not supplied'], 
    ['VLDB_INVALID_VALUE'                         , 'something wrong with value'], 
    ['VLDB_INVALID_VIRTUAL_SERVER'                , 'unknown virtual server'], 
    ['VLDB_NO_VSERVER_ROOT'                       , 'virtual server root RW doesn\'t exist'], 
    ['VLDB_MALLOC_ERROR'                          , 'malloc failure'], 
    ['VLDB_INVALID_LOGMSGLEVEL'                   , 'Log message level out of range'], 
    ['VLDB_OFFLINE'                               , 'Vldb is currently offline (not quorum)'], 
    ['VLDB_INTERNAL_ERROR'                        , undef()], 
    ['VLDB_NOT_YET_IMPLEMENTED'                   , undef()], 
    ['VLDB_DBLADE_UNREACHABLE'                    , 'client create to local dblade failed'], 
    ['VLDB_INVALID_DBLADE'                        , 'map deletion for nonexistent dblade invalid */'], 
    ['VLDB_SNAPSHOT_MASTER_MISSING'               , 'volume the snapshot belongs to is missing'], 
    ['VLDB_RECORD_NOT_FOUND'                      , 'No record with given input exists'], 
    ['VLDB_RW_NOT_FOUND'                          , 'RW of the volume not exists'], 
    ['VLDB_CANNOT_MODIFY_MSID'                    , 'Cannot modify MSID'], 
    ['VLDB_CANNOT_MODIFY_DSID'                    , 'Cannot modify DSID'], 
    ['VLDB_CANNOT_MODIFY_SNAPSHOT_TYPE'           , 'Cannot modify snapshot\'s type'], 
    ['VLDB_CANNOT_MODIFY_AGGRID'                  , 'Cannot modify aggregate ID'], 
    ['VLDB_CANNOT_MODIFY_VSERVERID'               , 'Cannot modify vserver ID'], 
    ['VLDB_CANNOT_MODIFY_MASTERID'                , 'Cannot modify master ID'], 
    ['VLDB_CANNOT_MODIFY_CRTIME'                  , 'Cannot modify creation time'], 
    ['VLDB_CANNOT_MODIFY_VOLTYPE'                 , 'Cannot modify volume type'], 
    ['VLDB_INVALID_RECORD_VALUES'                 , 'Invalid values in the record'], 
    ['VLDB_INVALID_MASTERID'                      , 'Invalid master ID'], 
    ['VLDB_INVALID_UUID'                          , 'Invalid UUID'], 
    ['VLDB_DUPLICATE_DSID'                        , 'Duplicate DSID'], 
    ['VLDB_VSERVER_ROOT_ALREADY_EXISTS'           , 'Vserver root volume already exists'], 
    ['VLDB_DUPLICATE_VSERVER_RW'                  , 'Duplicate RW in the vserver'], 
    ['VLDB_NO_READWRITE_VOLUME'                   , 'RW of the volume family not exists'], 
    ['VLDB_NO_MASTER_VOLUME'                      , 'Master volume of the snapshot not exists'], 
    ['VLDB_INVALID_AGGREGATE'                     , undef()], 
    ['VLDB_INVALID_DATAVERSION'                   , undef()], 
    ['VLDB_INVALID_MASTER_ID'                     , undef()], 
    ['VLDB_VOLUME_RECORD_NOT_CREATED'             , undef()], 
    ['VLDB_FAMILY_RECORD_NOT_CREATED'             , undef()], 
    ['VLDB_SNAPSHOT_DELETE_FAILED'                , undef()], 
    ['VLDB_CANNOT_DELETE_RW_HAS_LS_MIRROR'        , undef()], 
    ['VLDB_CANNOT_DELETE_RW_HAS_DR_MIRROR'        , undef()], 
    ['VLDB_RW_HAS_NO_FAMILY_RECORD'               , undef()], 
    ['VLDB_CANNOT_DELETE_ROOT_VSERVER_HAS_VOLUME' , undef()], 
    ['VLDB_VOLUME_DELETE_FAILED'                  , undef()], 
    ['VLDB_FAMILY_REC_DELETE_FAILED'              , undef()], 
    ['VLDB_SNAPSHOT_RECORD_NOT_CREATED'           , undef()], 
    ['VLDB_AGGR_RECORD_NOT_CREATED'               , undef()], 
    ['VLDB_BLADE_RECORD_NOT_CREATED'              , undef()], 
    ['VLDB_AGGR_LOCATION_NOT_FOUND'               , undef()], 
    ['VLDB_FILER_NOT_FOUND'                       , undef()], 
    ['VLDB_MODIFY_FAILED'                         , undef()], 
    ['VLDB_INVALID_NEXTID_TABLE'                  , undef()], 
    ['VLDB_CANNOT_MODIFY_NEXTID'                  , undef()], 
    ['VLDB_AGGR_NOT_ON_DBLADE'                    , undef()], 
    ['VLDB_CANNOT_UPDATE_AGGR_DBLADEID'           , undef()], 
    ['VLDB_CANNOT_CREATE_AGGR_TABLE_ENTRY'        , undef()], 
    ['VLDB_CANNOT_REMOVE_AGGR_TABLE_ENTRY'        , undef()], 
    ['VLDB_NO_MORE_DSIDS'                         , undef()], 
    ['VLDB_NO_MORE_MSIDS'                         , undef()], 
    ['VLDB_JUNCTION_NOT_FOUND'                    , undef()], 
    ['VLDB_JUNCTION_INACTIVE'                     , undef()], 
    ['VLDB_CANNOT_CREATE_JUNCTION_TABLE_ENTRY'    , undef()], 
    ['VLDB_INVALID_JOBID'                         , undef()], 
    ['VLDB_VOLUME_NOT_FOUND'                      , undef()], 
    ['VLDB_HOSTNAME_NOT_FOUND'                    , undef()], 
    ['VLDB_HOSTIP_NOT_FOUND'                      , undef()], 
    ['VLDB_VSERVER_ALREADY_EXISTS'                , undef()], 
    ['VLDB_VSERVER_NOT_CREATED'                   , undef()], 
    ['VLDB_VSERVER_NOT_EXISTS'                    , undef()], 
    ['VLDB_MGMT_VOL_REC_NOT_CREATED'              , undef()], 
    ['VLDB_MALLOC_FAILED'                         , undef()], 
    ['VLDB_ONLINE_UPDATE_FAILED'                  , undef()], 
    ['VLDB_EZAPI_WRAPPER_GET'                     , undef()], 
    ['VLDB_DBLADEIP_NOT_FOUND'                    , undef()], 
    ['VLDB_BLADE_NOT_FOUND'                       , undef()], 
    ['VLDB_MGMT_SNAPSHOT_REC_NOT_CREATED'         , undef()], 
    ['VLDB_CREATE_TXN_FAILED'                     , undef()], 
    ['VLDB_COMMIT_TXN_FAILED'                     , undef()], 
    ['VLDB_CORAL_EPOCH_NOT_FOUND'                 , undef()], 
    ['VLDB_CORAL_VOLUME_RECORD_NOT_CREATED'       , undef()], 
    );

#
# Based on vldbtest's options (as of 7/29/05)
#
my %commands = (
    bulk => {
        options => {
            file_name    => '-f', # (filename, input filename with bulk arguments) 
        },
    },
    create => {
        options => {
            aggr_id      => '-a', # (aggrID, aggrUUID)
            comment      => '-C', # (comment, comment tag)
            crtime       => '-c', # (secs, crtime in secs)
            dsid         => '-d', # (disd, disd)
            vserver_id   => '-f', # (vserverID, vserver)
            infoval      => '-i', # (infoval, info (Temporary record if infoval >= 128))
            dist_vector  => '-l', # (distvector, distvector)
            msid         => '-m', # (msid, msid)
            vol_name     => '-n', # (name, name)
            policy_name  => '-p', # (policy, policy name)
            root_vol     => ['-r', 'bool', 'no'], # (bool, rootVolume)
            master_id    => '-s', # (masterID, srcmsid(masterDSID for snapshotRec, sourceMSID for volumes))
            vol_type     => '-t', # (type, type(1-RW, 2-LS_RO, 3-DR_RO, 4-Snapshots))
            data_ver     => '-v', # (dataversion, dataversion)
        },
        verification => \&VLDB::CommandVerification::_createVerifier,
    },
    createcoralset => {
        options => {
            aggr_id      => '-a', # (aggrIDs, comma seperated aggrUUIDs)
            comment      => '-C', # (comment, comment tag)
            crtime       => '-c', # (secs, crtime in secs)
            epoch        => '-e', # (epoch, striping epoch)
            algorithm    => '-g', # (algorithm, striping algorithm)
            set_name     => '-n', # (name, name)
            vserver_id   => '-v', # (vserverID, vserver)
            width        => '-w', # (width, stripe width)
        },
        #verification => \&VLDB::CommandVerification::_createVerifier,
    },
    deletecoralset => {
        options => {
            msid         => '-m', # (msid, Coral set MSID) 
        },
        #verification => \&VLDB::CommandVerification::_deleteVerifier,
    },
    delete => {
        options => {
            dsid         => '-d', # (disd, disd) 
        },
        verification => \&VLDB::CommandVerification::_deleteVerifier,
    },
    dsidlookup => {
        options => {
            dsid         => '-d', # (disd, disd) 
        },
    },
    dump => {
        options => {
            aggr_table        => ['-a', 'bool', 'yes'], # (bool, dump aggrtable)      # this is busted!
            allocid_table     => ['-a', 'bool', 'yes'], # (bool, dump allocidtable)   # this is busted!
            blade_table       => ['-b', 'bool', 'yes'], # (bool, dump bladetable)
            coralstripe_table => ['-c', 'bool', 'yes'], # (bool, dump coralstripetable)
            coralepoch_table  => ['-e', 'bool', 'yes'], # (bool, dump coralepochtable)
            family_table      => ['-f', 'bool', 'yes'], # (bool, dump familytable)
            junction_table    => ['-j', 'bool', 'yes'], # (bool, dump junctiontable)
            all_tables        => ['-l', 'bool', 'yes'], # (bool, dump all tables)
            mgmtvol_table     => ['-m', 'bool', 'yes'], # (bool, dump mgmtvoltable)
            nextid_table      => ['-n', 'bool', 'yes'], # (bool, dump nextidtable)
            vserver_table     => ['-r', 'bool', 'yes'], # (bool, dump vservertable)
            snapshot_table    => ['-s', 'bool', 'yes'], # (bool, dump snapshottable)
            vol_table         => ['-v', 'bool', 'yes'], # (bool, dump voltable)
        },
        parser => \&VLDB::OutputParser::_dumpParser,
    },
    getaggrlocation => {
        options => {
            aggr_id => '-a', # (aggrID, aggrUUID) 
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    getbladeinfo => {
        options => {
            blade_uuid => '-b', # (bladeUUID, bladeUUID) 
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    getbladelist => {
        options => {
            blade_count => '-c', # (blade count, Number of Blades) 
            prev_blade => '-p', # (previous blade, Previous Blade Cookie) 
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    getnewids => {
        options => {
            dsid_count => '-d', # (dsidCount, dsidCount) 
            job_id => '-j', # (jobID, jobUUID) 
            msid_count => '-m', # (msidCount, msidCount) 
        },
        parser => \&VLDB::OutputParser::_getnewidsParser,
    },
    getrecord => {
        options => {
            dsid => '-d', # (disd, disd) 
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    getsnapshots => {
        options => {
            dsid => '-d', # (disd, disd) 
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    junctionlookup => {
        options => {
            junction_id => '-j', # (junctionID, format is "0xjuncFlags|juncDsid|juncFileID|juncUniquifier") 
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    mirrorsetlookup => {
        options => {
            dsid => '-d', # (disd, disd) 
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    modify => {
        options => {
            aggr_id     => '-a', # (aggrID, aggrUUID)
            comment     => '-C', # (comment, comment tag)
            crtime      => '-c', # (secs, crtime in secs)
            dsid        => '-d', # (disd, disd)
            new_dsid    => '-D', # (newDsid, newdsid)
            vserver_id  => '-f', # (vserverID, vserver)
            info        => '-i', # (info, info provided)
            dist_vector => '-l', # (distvector, distvector)
            msid        => '-m', # (msid, msid)
            vol_name    => '-n', # (name, name)
            policy_name => '-p', # (policy, policy name)
            master_id   => '-s', # (masterID, srcmsid(masterDSID for snapshotRec, sourceMSID for volumes))
            vol_type    => '-t', # (type, type(1-RW, 2-LS_RO, 3-DR_RO, 4-Snapshots)
            data_ver    => '-v', # (dataversion, dataversion)
        },
        verification => \&VLDB::CommandVerification::_createVerifier,
    },
    msidlookup => {
        options => {
            epoch => '-e', # (epoch, striping epoch)
            msid  => '-m', # (msid, msid)
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    releasenewids => {
        options => {
            id => '-i', # (ID, ID to be released) 
            job_id => '-j', # (jobID, jobUUID) 
        },
    },
    rjunctionlookup => {
        options => {
            child_msid => '-m', # (child msid, Child Msid) 
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    rootlookup => {
        options => {
            vserver_id => '-v', # (vserverID, vserver) 
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    setaggrlocation => {
        options => {
            aggr_id    => '-a', # (aggrID, aggrUUID)
            blade_uuid => '-b', # (bladeUUID, bladeUUID)
            aggr_name  => '-n', # (aggrName, aggrName)
        },
    },
    setbladeinfo => {
        options => {
            blade_uuid => '-b', # (bladeUUID, bladeUUID)
            hostname   => '-h', # (hostName, hostName)
            blade_type => '-t', # (bladeType, bladeType(1-Dblade, 2-Nblade, 3-Service))
        },
    },
    setjunctioninfo => {
        options => {
            active      => '-a', # (bool, active)
            junction_id => '-j', # (junctionID, format is "0xjuncFlags|juncDsid|juncFileID|juncUniquifier")
            msid        => '-m', # (msid, msid)
        },
    },
    setvserverinfo => {
        options => {
            vserver_id   => '-i', # (vserverID, Virtual Server ID)
            vserver_name => '-n', # (vserverName, Virtual Server Name)
        },
    },
    snapshotnamelookup => {
        options => {
            snapshot_name => '-n', # (snapshotName, Snapshot Name)
            vserver_name  => '-s', # (vserverName, Virtual Server Name)
            vol_name      => '-v', # (volumeName, Volume Name)
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
    swap => {
        options => {
            dsid1 => '-d', # (dsid, record1 dsid) 
            dsid2 => '-D', # (dsid, record2 dsid) 
            name1 => '-n', # (name, record1 name) 
            name2 => '-N', # (name, record2 name) 
        },
    },
    updateaggrmap => {
        options => {
            aggr_ids   => '-a', # (aggrIDs, comma seperated aggrUUIDs)
            blade_id   => '-b', # (bladeUUID, bladeUUID)
            aggr_names => '-n', # (aggrNames, comma seperated aggrNames)
        },
    },
    volnamelookup => {
        options => {
            vserver_name => '-s', # (vserverName, Virtual Server Name)
            vol_name     => '-v', # (volumeName, Volume Name)
        },
        parser => \&VLDB::OutputParser::_genericParser,
    },
);

#
# VLDB Client constructor
#

sub new {
    
    my $caller = shift();
    my %args = (
        remote_vldb => undef(),
        vldb_test_binary => undef(),
        timeout => 10,
        trace_pkg => undef(),
        client => $::CLIENT || 'localhost',
        filer_conn => $::FILER_CONN,
        errors_fatal => 1,
        @_
    );

    my $class = ref($caller) || $caller;
    my $self = {};
    bless($self, $class);

    $self->{remote_vldb} = $args{remote_vldb};
    $self->{vldb_test_binary} = $args{vldb_test_binary};
    $self->{timeout} = $args{timeout};
    $self->{trace_pkg} = $args{trace_pkg};
    $self->{client} = $args{client};
    $self->{verify} = 1;
    $self->{errors_fatal} = $args{errors_fatal};

    if (defined($self->{trace_pkg}) && (!(ref($self->{trace_pkg}) eq 'TracePkg'))) {
        cluck("\"$args{trace_pkg}\" is not a valid TracePkg reference.");
        $self->{trace_pkg} = new TracePkg();
    } elsif (defined($self->{trace_pkg})) {
        $self->{trace_pkg}->debug(msg => "$class instance $self using user-supplied TracePkg.");
    } else {
        $self->{trace_pkg} = new TracePkg(traceD => '*=all=enabled,STDOUT');
    }
    
    $self->{trace_pkg}->trace(msg => "already underway");

    my $hspec = host($self->{client});

    $self->{client_connection} = connect($args{filer_conn});
    $self->{trace_pkg}->trace(msg => "Connected to $self->{client}.");

    $self->_buildCommands();

    $self->{trace_pkg}->exit(msg => "Created new $class object");
    return $self;
}

sub ignoreNextError()
{
    my $self = shift();
    $self->{trace_pkg}->entry();
    $self->{ignore_next_error} = 1;
    $self->{trace_pkg}->debug(msg => "Next error will be ignored.");
    $self->{trace_pkg}->exit();
}

sub getErrorCode
{
    my $self = shift();
    my %args = @_;

    $self->{trace_pkg}->entry();

    if (0 == $args{error}) {
        return ($VldbErrorCodes[($args{error})][0], $VldbErrorCodes[($args{error})][1]);
    }

    if (1 == $args{error}) {
        return ('vldbtest flagged error', undef());
    }

    $self->{trace_pkg}->exit();
    return ($VldbErrorCodes[($args{error} - 100)][0], $VldbErrorCodes[($args{error} - 100)][1]);
}


sub _execute
{
    my $self = shift();
    my $command = shift();
    
    $self->{trace_pkg}->entry();

    my $output = sayEOF($self->{client_connection}, $command);

    $self->{trace_pkg}->exit();
    return $output;
}

# my $cmd_help = $self->_execute($self->{vldb_test_binary}." 0 ".$self->{client_connection}->{hostp}->{default_ip}." ?");

sub _buildCommands
{
    my $self = shift();
    my @big_option_list;
    
    $self->{trace_pkg}->entry();

    $self->{trace_pkg}->trace(msg => "Building command list...");
    my $cmd_help = $self->_execute($self->{vldb_test_binary}." 0 ".$self->{remote_vldb}." ?");
    
    $cmd_help =~ s/.*Commands are :\n//s;
    my @lines = split(/\n/, $cmd_help);

    foreach my $line (@lines) {
        if ($line =~ m/(\S+)\s+(.*)/) {
            $self->{trace_pkg}->debug(msg => "Adding command \"$1\" ($2)");
            push(@{$self->{command_list}}, $1);
        } else {
            $self->{trace_pkg}->error(msg => "Can't process vldb command: $line.");
            confess();
        }
    }

    foreach my $cmd (@{$self->{command_list}}) {
        my $out = $self->_execute($self->{vldb_test_binary}." 0 ".$self->{remote_vldb}." $cmd");
        $out =~ s/.*Command options:\n//s;
        @lines = split(/\n/, $out);
        foreach my $line (@lines) {
            if ($line =~ m/(\S+)\s+<(.*)>\s+(.*)/) {

                $self->{trace_pkg}->debug(msg => "Command $cmd: adding value option $1 which takes $2 ($3).");

                push(@big_option_list, "$1 ($2, $3) from $cmd.");

                $self->{commands}->{$cmd}->{options}->{$1}->{flag} = $1;
                $self->{commands}->{$cmd}->{options}->{$1}->{value} = 'yes';
                $self->{commands}->{$cmd}->{options}->{$1}->{value_type} = $2;
                $self->{commands}->{$cmd}->{options}->{$1}->{desc} = $3;

            } elsif ($line =~ m/(\S+)\s+(.*)/) {

                $self->{trace_pkg}->debug(msg => "Command $cmd: adding boolean option $1 ($2).");

                $self->{commands}->{$cmd}->{options}->{$1}->{flag} = $1;
                $self->{commands}->{$cmd}->{options}->{$1}->{value} = 'no';
                $self->{commands}->{$cmd}->{options}->{$1}->{desc} = $2;

                push(@big_option_list, "$1 (bool, $2) from $cmd.");

            } else {
                $self->{trace_pkg}->error(msg => "Can't process vldb command description for: $cmd.");
                confess();
            }
        }
    }

    #
    # createmgmtvol?
    #
    
#    $self->{trace_pkg}->debug(msg => "The BIG options list:");
#
#    foreach my $opt (sort(@big_option_list)) {
#        $self->{trace_pkg}->debug(msg => $opt);
#    }

    $self->{trace_pkg}->exit();
    #print Dumper($self->{commands});
    return;
}

sub AUTOLOAD 
{
    my $self = shift();

    #
    # Autoloaded function name
    #
    
    our $AUTOLOAD;
    my $name = $AUTOLOAD;

    $name =~ s/.*:://;

    my $type = ref($self) || confess("$self is not an object");

    $self->{trace_pkg}->entry(msg => $name);

    if (grep(/$name/, qw(DESTROY))) {
        $self->{trace_pkg}->exit(msg => $name);
        return;
    }

    if (defined($self->{commands}->{$name})) {
        $self->{trace_pkg}->debug(msg => "Executing discovered vldb command $name.");
        return $self->_executeVldbCommand($name, @_);
    } else {
        $self->{trace_pkg}->error(msg => cluck("Invalid VLDB command: $name"));
    }

    $self->{trace_pkg}->exit(msg => $name);
}

sub _executeVldbCommand
{
    my $self = shift();
    my $command = shift();
    my %args = @_;
    my $execute = $command." ";

    $self->{trace_pkg}->entry(msg => "entered for $command");

    while (my ($arg, $value) = each (%args)) {
        $self->{trace_pkg}->debug(msg => "Arg $arg has value $value");
        
        if (defined($commands{$command}->{options}->{$arg})) {
            $self->{trace_pkg}->debug(msg => "Found argument $arg in $command argument list.");

            my $append = '';
            if (ref($commands{$command}->{options}->{$arg}) eq 'ARRAY') {
                $self->{trace_pkg}->debug(msg => "Array ref with details about $arg was found.");
                if ($value) {
                    $self->{trace_pkg}->debug(msg => "Boolean $arg is on.");
                    $append .= ${$commands{$command}->{options}->{$arg}}[0]." ";
                } else {
                    $self->{trace_pkg}->debug(msg => "Boolean $arg is off.");
                }
            } else {
                $self->{trace_pkg}->debug(msg => "No details about $arg found. Assuming \"-key value\" type argument.");
                $append .= $commands{$command}->{options}->{$arg}." ".$value." ";
            }
            
            $self->{trace_pkg}->debug(msg => "appending \"$append\" to command.");
            $execute .= $append;
            $self->{trace_pkg}->debug(msg => "Current command: \"$execute\"");

        } else {
            #$self->{trace_pkg}->error(msg => confess("Invalid argument \"$arg\" doesn't translate to a known value.\n"));
            die(fatal_result("Invalid argument \"$arg\" for VLDB command \"$command.\""));
        }
    }

    chop($execute);

    $self->{trace_pkg}->debug(msg => "Executing VLDB command \"$execute\"");

    my $out = $self->_execute($self->{vldb_test_binary}." 0 ".$self->{remote_vldb}." $execute");

    my $result;

    if ($out =~ m/result: (\d+)/s) {
        $self->{trace_pkg}->debug(msg => "Got result $1 from $command");
        $result = $1;
    } else {
        die("Can't match output \"$out\"");
    }

    if (1 == $result) {

        $self->{trace_pkg}->error(msg => "Got nonzero result executing \"$execute\"\nUnmappable error code.  Probably caught by vldbtest itself.");

    } elsif ((0 != $result) && (!$self->{ignore_next_error})) {

        $self->{trace_pkg}->error(msg => "Got nonzero result executing \"$execute\"\n".${$VldbErrorCodes[$result-100]}[0]." / ".
            (defined(${$VldbErrorCodes[$result-100]}[1]) ? ${$VldbErrorCodes[$result-100]}[1] : "")."");

            if ($self->{errors_fatal}) {
                die(fatal_result("\"$execute\" returned $result (".${$VldbErrorCodes[$result-100]}[0]." / ".(defined(${$VldbErrorCodes[$result-100]}[1]) ? ${$VldbErrorCodes[$result-100]}[1] : "no description").")"));
            }
    }

    $self->{ignore_next_error} = 0;

    if (defined($commands{$command}->{parser})) {

        $self->{trace_pkg}->debug(msg => "Executing parser sub at ".$commands{$command}->{parser}." for $command.");

        my $hash = $commands{$command}->{parser}($out);

        $self->{trace_pkg}->exit(msg => "exited for $command");
        return ($result, $hash);
    } 
    
    if (defined($commands{$command}->{verification} && $self->{verify})) {

        $self->{trace_pkg}->debug(msg => "Executing verifier sub at ".$commands{$command}->{verification}." for $command.");

        $commands{$command}->{verification}($self, $command, %args) 
            or die(fail_result("Verification of $command failed."));
    }

    $self->{trace_pkg}->exit(msg => "exited for $command");
    return $result;
}

package VLDB::OutputParser;

use TracePkg;
use Data::Dumper;

my %KeyConverter = (
    'aggrName' => 'aggr_name',
    'bladeID' => 'blade_id',
    'name' => 'vol_name',
    'dsid' => 'dsid',
    'msid' => 'msid',
    'aggrID' => 'aggr_id',
    'vserverID' => 'vserver_id',
    'type' => 'vol_type',
    'masterID' => 'master_id',
    'dataVersion', => 'data_ver',
    'isRoot' => 'root_vol',
);


#  aggr_id      => '-a', # (aggrIDs, comma seperated aggrUUIDs)
#  comment      => '-C', # (comment, comment tag)
#  crtime       => '-c', # (secs, crtime in secs)
#  epoch        => '-e', # (epoch, striping epoch)
#  algorithm    => '-g', # (algorithm, striping algorithm)
#  set_name     => '-n', # (name, name)
#  vserver_id   => '-v', # (vserverID, vserver)
#  width        => '-w', # (width, stripe width)

sub _genericParser
{
    my $output = shift() || confess();
    my $trc = new TracePkg();

    $trc->entry();

    my @lines = split(/\n/, $output);

    my %dump;

    foreach my $line (@lines) {
        foreach my $tag ('aggrName', 'bladeID', 'RW MSID local_dsid', 'MSID Policy', 'MSID Policy Attributes Count', 'Root Lookup Volume Count', 'dsid', 'msid', 'name', 'aggrID', 'vserverID', 'type', 'masterID', 'dataVersion', 'isRoot') {
            if ($line =~ m/$tag\:\s+(\S+)/) {
                if (defined($KeyConverter{$tag})) {
                    $dump{$KeyConverter{$tag}} = $1;
                } else {
                    $dump{$tag} = $1;
                }
            }
        }
    }

    $trc->exit();
    return \%dump;
}

sub _getnewidsParser
{
    my $output = shift() || confess();
    my $trc = new TracePkg();

    $trc->entry();

    my %dump;

    my @lines = split(/\n/, $output);

    for (my $i = 0; $i < scalar(@lines); $i ++) {
        my $type;
        my $count;
        if ($lines[$i] =~ m/msidCount:\s(\d+)/) {
            $type = 'msid';
            $dump{msidCount} = $1;
        } elsif ($lines[$i] =~ m/disdCount:\s(\d+)/) {
            $type = 'dsid';
            $dump{dsidCount} = $1;
        } elsif ($lines[$i] =~ m/jobidCount:\s(\d+)/) {
            $type = 'jobid';
            $dump{jobidCount} = $1;
        } else {
            $type = undef();
        }
        $i ++;
        if (defined($type)) {
            for (my $j = 0; $j < $dump{$type.'Count'}; $j ++) {
                if ($lines[$i] =~ m/$type\:\s(\d+)/) {
                    push(@{$dump{$type}}, $1);
                } else {
                    confess("Can't match $lines[$i]");
                }
                $i++;
            }
        }
    }

    return \%dump;

    $trc->exit();
}

sub _dumpParser
{
    my $output = shift() || die();

    my @lines = split(/\n/, $output);

    my %dump;

    for (my $i = 0; defined($lines[$i]); $i++) {
        print "Index: $i Value: $lines[$i].\n";
        if ($lines[$i] =~ m/volCount: (\d+)/) {
            $dump{volCount} = $1;
            for (my $j = 0; $j < $dump{volCount}; $j++) {
                if ($lines[++$i] =~ m/vsid:\s+(\d+)\s+dsid:\s+(\d+)\s+msid:\s+(\d+)\s+name:\s+(\S+)\s+info:\s(\d+)/) {
                    $dump{volumes}->{$4}->{vsid} = $1;
                    $dump{volumes}->{$4}->{dsid} = $2;
                    $dump{volumes}->{$4}->{msid} = $3;
                    $dump{volumes}->{$4}->{name} = $4;
                    $dump{volumes}->{$4}->{info} = $5;
                } else {
                    die(fatal_result("Cannot match \"$lines[$i]\""));
                }
            }
        } elsif ($lines[$i] =~ m/snapshotCount: (\d+)/) {
            $dump{snapshotCount} = $1;
            for (my $j = 0; $j < $dump{snapshotCount}; $j++) {
                if ($lines[++$i] =~ m/dsid:\s+(\d+)\s+msid:\s+(\d+)\s+name:\s+(\S+)\s+info:\s+(\d+)/) {
                    $dump{snapshots}->{$3}->{dsid} = $1;
                    $dump{snapshots}->{$3}->{msid} = $2;
                    $dump{snapshots}->{$3}->{name} = $3;
                    $dump{snapshots}->{$3}->{info} = $4;
                } else {
                    die(fatal_result("Cannot match \"$lines[$i]\""));
                }
            }
        } elsif ($lines[$i] =~ m/familyCount: (\d+)/) {
            $dump{familyCount} = $1;
            for (my $j = 0; $j < $dump{familyCount}; $j++) {
                if ($lines[++$i] =~ m/dsid:\s+(\d+)\s+msid:\s+(\d+)\s+vsID:\s+(\d+)/) {
                    $dump{families}->{$1}->{dsid} = $1;
                    $dump{families}->{$1}->{msid} = $2;
                    $dump{families}->{$1}->{vsID} = $3;
                } else {
                    die(fatal_result("Cannot match \"$lines[$i]\""));
                }
            }
        } elsif ($lines[$i] =~ m/aggrCount: (\d+)/) {
            $dump{aggrCount} = $1;
            for (my $j = 0; $j < $dump{aggrCount}; $j++) {
                if ($lines[++$i] =~ m/aggrID:\s+(\S+)\s+aggrName:\s+(\S+)\s+bladeID:\s+(\d+)/) {
                    $dump{aggregates}->{$2}->{aggrID} = $1;
                    $dump{aggregates}->{$2}->{aggrName} = $2;
                    $dump{aggregates}->{$2}->{bladeID} = $3;
                } else {
                    die(fatal_result("Cannot match \"$lines[$i]\""));
                }
            }
        } elsif ($lines[$i] =~ m/bladeCount: (\d+)/) {
            $dump{bladeCount} = $1;
            for (my $j = 0; $j < $dump{bladeCount}; $j++) {
                if ($lines[++$i] =~ m/bladeID:\s+(\S+)\s+hostName:\s+(\S+)/) {
                    $dump{blades}->{$1}->{bladeID} = $1;
                    $dump{blades}->{$1}->{bladeName} = $2;
                } else {
                    die(fatal_result("Cannot match \"$lines[$i]\""));
                }
            }
        } elsif ($lines[$i] =~ m/allocIdCount: (\d+)/) {
            $dump{allocIdCount} = $1;
            for (my $j = 0; $j < $dump{allocIdCount}; $j++) {
                die(fatal_result("Fix me! Cannot match \"$lines[$i]\""));
                if ($lines[++$i] =~ m/bladeID:\s+(\S+)\s+hostName:\s+(\S+)/) {
                    $dump{allocids}->{$1}->{bladeID} = $1;
                    $dump{allocids}->{$1}->{bladeName} = $2;
                } else {
                    die(fatal_result("Cannot match \"$lines[$i]\""));
                }
            }
        } elsif ($lines[$i] =~ m/vserverCount: (\d+)/) {
            $dump{vserverCount} = $1;
            for (my $j = 0; $j < $dump{vserverCount}; $j++) {
                if ($lines[++$i] =~ m/vsid:\s+(\d+)\s+name:\s+(\S+)/) {
                    $dump{vservers}->{$2}->{vsid} = $1;
                    $dump{vservers}->{$2}->{name} = $2;
                } else {
                    die(fatal_result("Cannot match \"$lines[$i]\""));
                }
            }
        } elsif ($lines[$i] =~ m/mgmtVolCount: (\d+)/) {
            $dump{mgmtVolCount} = $1;
            for (my $j = 0; $j < $dump{mgmtVolCount}; $j++) {
                die(fatal_result("Fix Me! Cannot match \"$lines[$i]\""));
                if ($lines[++$i] =~ m//) {
                    $dump{mgmtvols}->{$2}->{vsid} = $1;
                    $dump{mgmtvols}->{$2}->{name} = $2;
                } else {
                    die(fatal_result("Cannot match \"$lines[$i]\""));
                }
            }
        } elsif ($lines[$i] =~ m/nextMsid: (\d+)/) {
            $dump{nextMsid} = $1;
        } elsif ($lines[$i] =~ m/nextDsid: (\d+)/) {
            $dump{nextDsid} = $1;
        } elsif ($lines[$i] =~ m/coralStripeRuleCount: (\d+)/) {
            $dump{coralStripeRuleCount} = $1;
        } elsif ($lines[$i] =~ m/coralEpochCount: (\d+)/) {
            $dump{coralEpochCount} = $1;
        }
    }
    return \%dump;
}

package VLDB::CommandVerification;

use TracePkg;

use Data::Dumper; 
use TracePkg();

sub _createVerifier
{
    my $client = shift();
    my $command = shift();
    my %args = @_;
    my $trc = new TracePkg();
    $trc->entry();

    my ($rv, $rec) = $client->getrecord(dsid => $args{dsid});

    if (0 != $rv) {
        $trc->error(msg => "Failed to get appropriate record.");
        return undef();
    }

    while (my ($field, $value) = each(%args)) {
        if (defined($$rec{$field})) {
            $trc->debug(msg => "Comparing $field\: requested=$value, stored=$$rec{$field}.");

            if ($value ne $$rec{$field}) {
                $trc->error(msg => "Field mismatch for $field detected, verification failed.");
                $trc->exit();
                return undef();
            }
        } else {
            $trc->debug(msg => "No matching filed to verify $field with.");
        }
    }

    $trc->trace(msg => "All available fields matched successfully.");
    $trc->exit();
    return 1;
}

sub _deleteVerifier
{
    my $client = shift();
    my $command = shift();
    my %args = @_;
    my $trc = new TracePkg();
    $trc->entry();

    $client->ignoreNextError();

    my ($rv, undef()) = $client->getrecord(dsid => $args{dsid});

    my ($code, $desc) = $client->getErrorCode(error => $rv);

    if (0 != $rv) {
        $trc->debug(msg => "Failed to get appropriate record.");
        if ('VLDB_RECORD_NOT_FOUND' eq $code) {
            $trc->debug(msg => "Record not found, delete was successful.");
            $trc->exit();
            return 1;
        }
    } 

    $trc->error(msg=>"Record wasn't deleted: ($code/$desc). ");

    $trc->exit();
    return undef();
}

1;
