# Library to perform VM Vol operations using ZAPI protocol
# @author canil@netapp.com

package VMVol::ZAPI::Operations;

use strict;
use warnings;

use NACL::C::Node;
use NACL::C::Volume;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use Tharn;
use Data::Dumper;
use NACL::C::Lun;
use NACL::CS::Lun;
use NACL::C::LunIgroup;
use NACL::STask::Lun;
use NACL::CS::LunMapped;
use NACL::CS::LunBind;
use NACL::C::LunBind;
use MIME::Base64 qw(encode_base64);

# San Libraries for SAN VVOL coverage
use San::Task::VVolHelper;

our @ISA = qw(VMVol);

sub new {
    my $class  = shift;
    my $params = shift;

    #my ( $protocol, $node, $volume, $vserver, $uuid, $file_name ) = @_;
    my $self = {};
    #print Dumper($params); 
    $self->{protocol} = $params->{protocol} if ( defined( $params->{protocol} ) );
    
    if ( defined( $params->{node} ) ) {
        $self->{node}      = $params->{node};
        #$self->{node}      = _get_NACL_obj_node( $params->{node} );
        $self->{node_name} = $params->{node};
        if(defined($params->{node_ci})) {
        	$self->{node_ci} = $params->{node_ci};
        } else {
        	$self->{node_ci}   = NACL::C::Node->find();
        }
        
    }
    $self->{volume}  = $params->{volume}  if ( defined( $params->{volume} ) );
    $self->{vserver} = $params->{vserver} if ( defined( $params->{vserver} ) );

    if ( $self->{protocol} =~ /nfs/ ) {
        $self->{uuid} = $params->{uuid} if ( defined( $params->{uuid} ) );
        $self->{file_name} = $params->{file_name} if ( defined($params->{file_name}));
    }
    elsif ( $self->{protocol} =~ /iscsi|fcp|san/ ) {
        $self->{uuid} = $params->{uuid} if ( defined( $params->{uuid} ) );
        $self->{pe_lun} = $params->{pe_lun} if ( defined( $params->{pe_lun} ) );
        $self->{vvol_lun} = $params->{vvol_lun}
          if ( defined( $params->{vvol_lun} ) );
    }
    $self->{vvol_size} = $params->{vvol_size} if ( defined( $params->{vvol_size} ) );
    
    bless( $self, $class );

    return $self;
}

# createMetaVirtualVolume:
# [file-create-directory, file-delete-file, file-get-file-info, file-write-file,
# lun-create-by-size, lun-get-iter, lun-map-get-iter, lun-map-list-info, lun-set-attribute,
# lun-set-space-alloc, qtree-list-iter, quota-report-iter,
# snapshot-get-iter, volume-get-iter]
# /u/canil/p4/test/lib/NACL/UnitTest/C/VolumeFile.thpl
# ~/p4/test/san/vvols/vvols_basic_func.thpl
sub createMetaVirtualVolume {
    my $self = shift;

    $self->{meta} = 1;
    if ( $self->{protocol} =~ /nfs/ ) {
        if ( !defined( $self->{file_name} ) ) {
            $self->{file_name} = $self->{uuid} . "_metadata.vmx";
        }
    }
    createVirtualVolume($self);
}

sub createVirtualVolume {
    my $self = shift;

    if ( $self->{protocol} =~ /nfs/ ) {
        if ( !defined( $self->{file_name} ) ) {
            $self->{file_name} = $self->{uuid} . "_disk.vmdk";
        }

        # Creating metavvol directory
        #print "Creating metavvol directory\n";
        try {
            $self->{node}->file_create_directory(
                'vserver-override-name' => $self->{vserver},
                path => "/vol/" . $self->{volume} . "/" . $self->{uuid},
                perm => "0777",
            );

            # Create file of size 100kb
            #print "Creating metavvol file\n";
            my $params = {
                obj       => $self,
                size      => $self->{vvol_size},
                file_name => $self->{file_name},
            };

            print $params->{file_name} . "\n";

            print "Error" if ( _create_file($params) == -1 );
            
            return 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            print "Error is: " . $exception_object->text();
            my $resp = $exception_object->get_response_object();
            print "Raw output was:\n" . $resp->get_raw_output();
        }
    }
    elsif ( $self->{protocol} =~ /iscsi|fcp|san/ ) {

# For SAN, there have to be additional steps
# 1. Make sure iSCSI service is up and running on the vserver specified.
#    If not, then create one. Also, one of the LIFs should be able to talk iSCSI
# 2. Create PE lun, unless specified as a Command line argument
# 3. Get size of the vvol lun to be created.
# 4. Bind vvol lun to PE lun
# 5. Create igroup with iqn from the user. Map PE lun to the igroup
#$self->{vh_obj} = new San::Task::VVolHelper();

        # Assuming the path is create-bind
        # If PE Lun doesn't exist, create one
        $self->{pe_obj} = NACL::STask::Lun->create(
            command_interface => $self->{node_ci},
            path      => "/vol/" . $self->{volume} . "/" . $self->{pe_lun},
            vserver   => $self->{vserver},
            size      => '35m',
            class     => 'protocol-endpoint',
            ostype    => 'vmware',
            if_exists => 'reuse',
        );
        
        # Create vvol lun
        $self->{vvol_obj} = NACL::STask::Lun->create(
            command_interface => $self->{node_ci},
            path      => "/vol/" . $self->{volume} . "/" . $self->{vvol_lun},
            vserver   => $self->{vserver},
            size      => $self->{vvol_size},
            class     => 'vvol',
            ostype    => 'vmware',
            if_exists => 'reuse',
        );

        return 0;
    }
    else {
        return -1;
    }
}

sub deleteVirtualVolume {
    my $self = shift;
    if ( $self->{protocol} =~ /nfs/ ) {

        # 1. Delete file
        # 2. Delete folder

        try {
            $self->{node}->file_delete_file(
                'vserver-override-name' => $self->{vserver},
                path                    => "/vol/"
                  . $self->{volume} . "/"
                  . $self->{uuid} . "/"
                  . $self->{file_name},
            );
            $self->{node}->file_delete_directory(
                'vserver-override-name' => $self->{vserver},
                path => "/vol/" . $self->{volume} . "/" . $self->{uuid},
            );
            
            return 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            print "Error is: " . $exception_object->text();
            my $resp = $exception_object->get_response_object();
            print "Raw output was:\n" . $resp->get_raw_output();
            return -1;
        }

    }
    elsif ( $self->{protocol} =~ /iscsi|fcp|san/ ) {
        my $pe_path   = "/vol/" . $self->{volume} . "/" . $self->{pe_lun};
        my $vvol_path = "/vol/" . $self->{volume} . "/" . $self->{vvol_lun};
        NACL::C::Lun->delete(
            command_interface => $self->{node_ci},
            path              => $vvol_path,
            vserver           => $self->{vserver}
        );

    }
    else {
        return -1;
    }
}

# clone-create, file-create-directory, file-delete-file, file-get-file-info, file-write-file
sub cloneVirtualVolume {
    my $self = shift;
    my $snapshot_suffix = shift;
    if ( $self->{protocol} =~ /nfs/ ) {

        my $snapshot_params;

        $snapshot_params = {
            protocol => "nfs",
            node     => $self->{node_name},
            volume   => $self->{volume},
            vserver  => $self->{vserver},
            uuid     => $self->{uuid} . ".snapshot_" . $snapshot_suffix,
        };

        my $snapshot = VMVol::ZAPI::Operations->new($snapshot_params);

        # Clone file and put that in the new created directory
        try {

            # Create directory to store
            $snapshot->{node}->file_create_directory(
                'vserver-override-name' => $snapshot->{vserver},
                path => "/vol/" . $snapshot->{volume} . "/" . $snapshot->{uuid},
                perm => "0777",
            );
			
			$snapshot->{file_name} = $self->{uuid} . "_disk.vmdk";
			
            # Create clone of the file in the destination directory
            $snapshot->{node}->clone_create(
                'vserver-override-name' => $snapshot->{vserver},
                'destination-path'      => $snapshot->{uuid} . "/"
                  . $snapshot->{file_name},
                'source-path' => $self->{uuid} . "/" . $self->{file_name},
                volume        => $snapshot->{volume},
            );
            
            return $snapshot;
            
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            print "Error is: " . $exception_object->text();
            my $resp = $exception_object->get_response_object();
            print "Raw output was:\n" . $resp->get_raw_output();
            
            return -1;
        }
    }
    elsif ( $self->{protocol} =~ /iscsi|fcp|san/ ) {
		# Get the name of the clone
        my $clone_params;

        $clone_params = {
            protocol => $self->{protocol},
            node     => $self->{node_name},
            volume   => $self->{volume},
            vserver  => $self->{vserver},
            vvol_lun => $self->{vvol_lun} . "_clone_" . $snapshot_suffix,
            uuid     => $self->{uuid} . "_clone_" . $snapshot_suffix,
            pe_lun => $self->{pe_lun},
		    node_ci => $self->{node_ci},
		    vvol_size => $self->{vvol_size}
        };

        my $clone = VMVol::ZAPI::Operations->new($clone_params);		
		# Create a snapshot of the lun with the name
		#NACL::C::Snapshot->create(
		#			command_interface => $clone->{node_ci},
		#			volume => $clone->{volume},
		#			vserver => $clone->{vserver},
		#			snapshot => $clone->{uuid},
		#		);
		# Create a clone with the snapshot
		NACL::C::Lun->clone(
					command_interface => $clone->{node_ci},
					#volume => $clone->{volume},
					vserver => $clone->{vserver},
					#'snapshot-name' => $clone->{uuid},
					'source-path' => $self->vvol_path(),
					'path' => $self->vvol_path(),
					'destination-path' => "/vol/" . $clone->{volume} . "/" . $clone->{vvol_lun},
				);
        return $clone;
    }
    else {
        return -1;
    }
}

sub bindVirtualVolume {
    my $self = shift;
    if ( $self->{protocol} =~ /nfs/ ) {
        print "Operation successful. Nothing to do here\n";
    }
    elsif ( $self->{protocol} =~ /iscsi|fcp|san/ ) {
        my $pe_path   = "/vol/" . $self->{volume} . "/" . $self->{pe_lun};
        my $vvol_path = "/vol/" . $self->{volume} . "/" . $self->{vvol_lun};

   	        # Create the bind only if it doesnt already exist
        
            my $bind = NACL::C::LunBind->create(
                command_interface        => $self->{node_ci},
                'protocol-endpoint-path' => $pe_path,
                'vvol-path'              => $vvol_path,
                vserver                  => $self->{vserver},
            );
        
    }
    else {
        return -1;
    }
    
    return 0;
}

sub unbindVirtualVolume {
    my $self = shift;
    if ( $self->{protocol} =~ /nfs/ ) {
        print "Operation successful. Nothing to do here\n";
    }
    elsif ( $self->{protocol} =~ /iscsi|fcp|san/ ) {

        my $pe_path   = "/vol/" . $self->{volume} . "/" . $self->{pe_lun};
        my $vvol_path = "/vol/" . $self->{volume} . "/" . $self->{vvol_lun};

        NACL::C::LunBind->destroy(
            command_interface        => $self->{node_ci},
            'protocol-endpoint-path' => $pe_path,
            'vvol-path'              => $vvol_path,
            vserver                  => $self->{vserver},
            force                    => 'true'
        );

    }
    else {
        return -1;
    }
}

# Internal routines
sub _get_NACL_obj_node {
    my $node = $_[0];
    print "Creating NACL APISet object for $node\n";
    my $Host = host($node);

    # Get APISet object of the node under consideration
    my $Apiset_obj = NACL::APISet->new(
        hostobj   => $Host,
        category  => "Node",
        interface => "ZAPI",
        set       => "CMode",
    );

    return $Apiset_obj;
}

sub _create_file {
    my $param = shift;
    my $data;

    # VP creates a file with a 50 MB offset and writes 1 byte at the end
    $param->{size} =~ m/(\d+)(\w)/;
    my $size = $1;
    my $index = $2;
    
    if ( $size == 0 ) {
        $data = "";
    } else {
    	$data = "A";
    }
    
    my $offset = 0;
    if($index eq 'm') {
    	$offset = $size*1024*1024 - 1;
    } elsif ($index eq 'g'){
    	$offset = $size*1024*1024*1024 - 1;
    } 
    
    $param->{obj}->{node}->file_write_file(
        'vserver-override-name' => $param->{obj}->{vserver},
        path => "/vol/"
          . $param->{obj}->{volume} . "/"
          . $param->{obj}->{uuid} . "/"
          . $param->{file_name},
           offset                  => $offset,
           #data                 => MIME::Base64::encode($data, ''),
           data => 30,
           overwrite => 'false',
    );
    return 0;
}

sub vvol_path {
	my $self = shift;
	if($self->{protocol} =~ /nfs/i) {
		return "/vol/" . $self->{volume} . "/" . $self->{uuid} . "/" . $self->{file_name};
	}
	elsif ( $self->{protocol} =~ /iscsi|fcp|san/i ) {
		return "/vol/" . $self->{volume} . "/" . $self->{vvol_lun};
	}
}

1;
