#$Id$
#Copyright (c) 2006 Network Appliance, Inc.
#All rights reserved.
#
## @summary Package of helper (utility) routines when dealing with 
## Client (Host) connections
##
## @description
##   The functions described in this pacakge are helper or utility
##   routines that are platform independent and assist in the execution
##   of the commands to and on client (host) systems.
##
## @usage
##   use VMwareClient.pm;  
##
## @keywords SAN FCP iSCSI Blocks
## @dependencies
##
## @author Thomas Masterson (thomasm@netapp.com)
## @status released
## @change 2002-08-26 thomasm: created
## @change 2002-09-06 thomasm: added HPUX support
## @change 2002-09-23 thomasm: cleanup documetation
## @change 2003-02-12 thomasm: added AIX support
## @change 2003-02-14 thomasm: convert to SanClient approach
## @change 2003-02-26 thomasm: change say to send fail instead of fatal
## @change 2003-03-04 thomasm: add Windows support
## @change 2003-04-16 thomasm: add -noprompt on say
## @change 2004-12-13 hedstrom: add read_remote_file
## @change 2005-03-10 hedstrom: add appropriate timeout to all &execute calls
## @change 2005-03-29 hedstrom: allow timeout in &execute calls
## @change 2005-06-01 archana: Updated author tag
## @change 2005-07-29 pressley: execute(): fix to remove cmd from output
## @change 2006-12-20 harb: modified package to work with VMwareClient


use strict;

# Package Name
##############
package VMwareClient::ClientUtil;

# Package Requirements
######################
use POSIX;
use Config;
use English;
use Data::Dumper;


# Exported global variables
###########################
$ClientUtil::TRUE      = 1;
$ClientUtil::FALSE     = 0;


# Open global file handle for audit.log

my $Alog;
# my $Alast;
$Alog=*ALOG;
# $Alast=*ALAST;

open(ALOG,">> sanaudit.log");
ALOG->autoflush(1);

# open(ALAST,">> sanaudit.last");
# ALAST->autoflush(1);



###############################################################################
# Subroutine Name: execute 
# Objective: Execute command on host and return output     
#                                
###############################################################################

sub execute {
    my ($connection, $cmd, %args) = @_;

    my ($env_output, $output, @output, $status);
    my ($match, $before, $after, $sentinel, $default_timeout);
    my ($timeout,$cmdlog);

    $default_timeout = "600000";
    $cmdlog=1;

    foreach (keys %args) {
	  if (/^-?cmdlog$/) {
	      $cmdlog = $args{$_};
	  }
	  elsif (/^-?timeout$/) {
	      $timeout = $args{$_};
	  }
    }

    # If we are on HP make some settings on the shell that will turn of
    # of command processing in the SHELL and give true output.
    if (($connection->{hostp}->{hosttype} =~ /unix-hpux/) or
        ($connection->{hostp}->{hosttype} =~ /unix-aix/)  or 
        ($connection->{hostp}->{hosttype} =~ /unix-linux/)) {
        &_set_term_setting($connection);
    }
    
    
    my $tmp=$cmd;
    if ($cmdlog==1) {
  	print $Alog "CLIENT(".$connection->{hostp}->hostname().")> ".$cmd."\n";
  	# print $Alast $cmd."\n";
    }

    # add to the command the echo of status to get that information from
    # the unix shell. Also add the sentinel string DoNe to command so that
    # we can match on that string.
    $sentinel = 'DoNe';
    # If we are on windows and running through RSH (default for now),
    # run the the command with cmd /c in front of it so the RSHD knows
    # what to run
    if ($connection->{hostp}->{hosttype} =~ /windows/i) {
        $cmd = $cmd . " & echo rstatus=\%ERRORLEVEL\% & echo $sentinel";
        $cmd = "cmd.exe /c \"$cmd\"";
    } 
    else {
        $cmd = $cmd . ";echo \"rstatus=\$\?\" ;echo $sentinel";
    }

    # clear buffers
    $connection->discardbuff();

    # execute the passed command
    if (defined($timeout) and $timeout ne "") {
        ($match, $output, $after) =  &main::say($connection, $cmd, $sentinel,
                                                '-noprompt',
                                                '-timeout', $timeout, 
                                                '-ignore', "echo $sentinel");
    } else {
        ($match, $output, $after) =  &main::say($connection, $cmd, $sentinel,
                                                '-noprompt',
                                                '-timeout', $default_timeout, 						                     
                                                '-ignore', "echo $sentinel"
		                                        );
    }

    # strip output and remove CR/LF
    chomp($output);
    $output =~ s/\r//g;

    # remove first line 
    $output =~ s/.*\n//; 
    $cmd=~s/"rstatus=\$\?" ;echo DoNe//g;
    $output=~s/"rstatus=\$\?" ;//g;

    # strip sentinel and then status from output
    $output =~ s/\n$sentinel.*//;

    # strip any telnet ctrl characters that get
    # added to the output (this problem seems 
    # to only occur with certain platforms.
    $output =~ s/ \cH//g;

    # locate status.  Remember that commands
	# echo themselves so make sure you don't 
	# get the command line 
    ($status)=$output=~/rstatus=(\d+)/; 

    $output =~ s/rstatus=\d+//;

    # create an array from the output
    @output = split(/\n/, $output);

    # return the status and the output array

    ## We need to determine if command is multiline 
    ## and exists in the output.  IMPORTANT: Not all
    ## commands echo back their input so don't forget to
    ## check before modifying the output.

    my @cmds=split(/\n/,$cmd);  

    ## If output contains second line of cmd 
    ## then cmd was echoed to output and is 
    ## multiline
    if (defined $cmds[1] and length $cmds[1]){
        ## Remove continuation character
        $cmds[1]=~s/\\//g;
        if ($output[0]=~/$cmds[1]/){
          ## determine size of command and strip
          ## from there.
          @output=@output[@cmds-1..@output-1]; 

        }
    }

    
    return($status, @output);

}
###############################################################################
# Subroutine Name: log
# Objective:       log cmds issued to client and filer
###############################################################################

sub log {
  my (@line)=@_;

  print $Alog  join("\n",@line)."\n";
  # print $Alast join("\n",@line)."\n";
}

###############################################################################
# Subroutine Name: remove_remote_file
# Objective:       delete file on remote machine
###############################################################################

sub remove_remote_file {
    my ($connection, $rem_file) = @_;

    my (@output, $status);

    ($status, @output) = &execute($connection, "rm $rem_file");

    if ($status != 0) {
        &main::tstlog(-fatal, "Cannot remove remote file: $rem_file");
    }

}

###############################################################################
# Subroutine Name: create_remote_file
# Objective:       create file on remote machine
###############################################################################

sub create_remote_file {

    my ($connection, $rem_file, @filebuf) = @_;

    my (@output, $status, $line);

    if (not defined($rem_file)) {
        &main::tstlog(-fatal, "Remote file name not specified.");
    }

    &execute($connection, "touch $rem_file");
    
    foreach $line (@filebuf) {
        $line =~ s/\$/\\\$/g;
        #$line =~ s/\"/\\\"/g;
    
      if ($connection->{hostp}->{hosttype} =~ /windows/i) {
            ($status, @output) = &execute($connection, 
                                          "echo $line >> $rem_file");
        } else {
            ($status, @output) = &execute($connection, 
                                          "echo \"$line\" >> $rem_file");
        }
        if ($status != 0) {
            &main::tstlog(-fatal, "Cannot populate remote file: $rem_file");
        }
    }

    return($ClientUtill::TRUE);

}
###############################################################################
# Subroutine Name: _set_term_setting 
# Objective:       This routine will setup the HP and AIX POSIX shell (sh) so
#                  that it does not do processing of commands in terms of
#                  buffering and logical command placement.
###############################################################################

sub _set_term_setting
{
    my ($connection) = @_;

    my ($cmd, $sentinel);
    my ($match, $before, $output);

    # set COLUMN width to 2048 to remove the command line processing 
    # in HP's POSiX SHELL.
    $cmd = "COLUMNS=2048;export COLUMNS;unset EDITOR";

    # clear buffers
    $connection->discardbuff();
    
    # execute the command
    $sentinel = 'DoNe';
    $cmd = $cmd . ";echo $sentinel";
    ($match, $before, $output) =  &main::say($connection, $cmd, $sentinel,
                                             '-noprompt',
                                             '-timeout', 30);

}

1;

