#$Id: //depot/prod/test/Rsierranevada/lib/VMwareClient/Guests/Solaris/SolarisGuest.pm#1 $
#Copyright (c) 2006 Network Appliance, Inc.
#All rights reserved.
#
## @summary Package to represent and manipulate host specific operations
## on Solaris
##
## @description
##   The functions described in this package allow for the platform
##   independent manipulation of data on a Solaris Guest. 
##
## @usage
##   use VMwareClient.pm;   
##
## @keywords SAN FCP iSCSI Blocks Solaris
## @dependencies
##   VMwareClient.pm package;
##
## @author Thomas Masterson (thomasm@netapp.com)
## @author holly harb (harb@netapp.com)
## @note mount method was written by tom. 


use strict;



# Package Name
##############
package VMwareClient::Guests::Solaris::SolarisGuest;


use POSIX;
use Config;
use English;
use Net::Ping;
use Tie::IxHash;
use Data::Dumper;
use File::Basename;
use VMwareClient::ClientUtil;


use Error;
use vars qw(@EXPORT @ISA);
@ISA=qw(Error SanLog);


## @name ::_new
## @description Create a new Solaris Guest Object.
## This routine creates a new host object for Solaris guests.
## It initializes the host data structure,
## gets Solaris specific information (populates some global variables),
## bless(es) and returns the object.
## @synopsis 
## $dmObj = VMwareClient::Guests->new(-connection => $conn, -host => $guest, -timeout =>30,-protocol => $protocol, -hosttype => $hosttype);
## @arg $conn Required: The host connection to manipulate SCSI disks on
## @arg $host Required: The name of the guest os in tharnhost
## @arg $timeout Required: Timeout if a connection cannot be made 
## @arg $protocol Required: Either fcp or iscsi
## @arg $hosttype Required: The host type of the guest os you are passing in. It better be solaris, otherwise fail. 


sub _new (\%)
{
    my ($class, %args) = @_;

    my ($device_str, @devices, %hRaw);
    my ($self,$conn,$nfs_mount,$host,$timeout,$protocol,$hosttype);

    
    if (defined $args{-self}) {		
	$self=$args{-self};
    }
    else  {
	$self={};
	# bless 
	bless $self,$class; 
    }


    # parse arguments
    foreach (keys %args) {
	if (/-?conn/) {
	    if (not length($conn = $args{$_})) {
		$self->_set_err("Error: SolarisGuest::new() Invalid value for -conn: $args{$_}");                 
		return $self;
	    }
	    next;
	}
	if (/-?self/) {
	    if (not length($self = $args{$_})) {
		$self->_set_err("Error: SolarisGuest::new() Invalid value for -self: $args{$_}");                 
		return $self;
	    }
	    next;
	}
	if (/-?host/) {
		    if (not length($host = $args{$_})) {
			$self->_set_err("Error: SolarisGuest::new() Invalid value for -host: $args{$_}");                 
			return $self;
		    }
		    next;
	}
	if (/-?timeout/) {
	            if (not length($timeout = $args{$_})) {
		        $self->_set_err("Error: SolarisGuest::new() Invalid value for -timeout: $args{$_}");                 
			return $self;
	            }
	            next;
	}
	if (/-?protocol/) {
		    if (not length($protocol = $args{$_})) {
		        $self->_set_err("Error: SolarisGuest::new() Invalid value for -protocol: $args{$_}");                 
			return $self;
                    }
		    next;
	}
	if (/-?hosttype/) {
	            if (not length($hosttype = $args{$_})) {
			$self->_set_err("Error: SolarisGuest::new() Invalid value for -hosttype: $args{$_}");                 
			return $self;
		    } 
                    next;
	}
	else {
	    $self->_set_err("Error: SolarisGuest::new() Invalid parameter specified \'$_\'");
	    return $self; 
	}

    }

    unless (defined $conn and length $conn) {
	$self->_set_err("Error: LinuxGuest::new()",
			"Invalid value for argument '-conn'");
	return $self; 
    }

    $self->{conn}=$conn;

    $self->{nfs_mount}=undef;




    return($self);

}

## @name ::vmware_fdisk
## @description Determines how many luns are disks on the solaris guest.
## Because solaris guests do not have an HBA or an iscsi initiator, we need to tell solaris
## which disks are luns. All luns start wtih /dev/sdb and increase. Linux always starts with /dev/sda as the boot disk. /dev/sdb 
## will always be the first lun. 
##
## @synopsis 
## $VMware = new VMwareClient(-connection => $client_connection, -host => $p_client, -timeout =>30,-protocol => $PROTOCOL);
## $theGuest = $VMware->create_guest(-connection => $client_connection, -host => $client, -timeout =>30,-protocol => $PROTOCOL, -type => $type);
## @disk = $theGuest->vmware_fdisk($numLuns);
##
## @arg $num Required: The number of luns allocated to this solaris guest
## @returns an array of the luns to be configured starting with /dev/sdb and increasing

sub vmware_fdisk() {

    my ($this, $num) = @_;
    my  (@string, $letter);
    my $current = 1;
    my $myletter = 0;
 
 
 while($num > $myletter){
 
 
   $string[$myletter] = "c1t"."$current"."d0"; 
 
   $myletter++;
   $current++; 
 
  } 
    
return @string;  
   
}

## @name ::make
## @description Creates a filesystem (with the makefs or newfs command) on
## the current device and partition. User can pass the file system type
## and file system specific options.
## @synopsis 
## $theGuest->make(-device => $device);
## @arg $device Required: The device to be fdisked

sub make (\%)
{
 
    my ($this, %args) = @_;
    my $device = $args{'-device'};
    my (@output, $status,$n_cmd);

  
    # create format partition command
    $n_cmd = "fdisk -B /dev/rdsk/$device"."p0";

    # execute
    
    ($status, @output) = VMwareClient::ClientUtil::execute($this->{conn}, $n_cmd);

    return $ClientUtil::TRUE;

}

## @name ::format
## @description Creates a filesystem (with the makefs or newfs command) on
## the current device and partition. User can pass the file system type
## and file system specific options.
## @synopsis 
## $theGuest->format(-device => $device);
## @arg $device Required: The device to be fdisked
## @arg $bs Optional: The block size to use on the mkfs call (-b flag)
## @arg $fs Optional: The fragment size to use on the mkfs call (-f flag)
## @arg $size Optional: The size of the fileystem to create, defaults to
## the full size of specified device/parition.


sub format (\%)
{
 
    my ($this, %args) = @_;
    my $device = $args{'-device'};
    my (@output, $status, $n_cmd);

   
  
    my $nate = '"nate"';
    
    # create format partition command
    $n_cmd = "echo -e '\np\nm\n1\nyes\n6\n0\n0\n0\n0\n0\n0\nyes\n$nate\nyes\nquit\nquit\n' | format $device";

    # execute
    
    ($status, @output) = VMwareClient::ClientUtil::execute($this->{conn}, $n_cmd);

    
 

    $n_cmd ="newfs -Nv /dev/rdsk/$device"."s6| grep mkfs ";
    ($status, @output) = VMwareClient::ClientUtil::execute($this->{conn}, $n_cmd);

    
    $output[0] =~ s/-o N//g;
  

    $n_cmd ="$output[0]";
    ($status, @output) = VMwareClient::ClientUtil::execute($this->{conn}, $n_cmd);


    return $ClientUtil::TRUE;

}


## @name ::mount
## @description Mounts a created filesystem onto a specified mount point path
## By default NATE will choose a unique path for you under /usr/tmp/
## @synopsis 
## $filesystem = $theGuest->mount(-device => $device);
## @arg $device Required: The device to be manipulated
## @returns $mount_point : Where its mounted


sub mount (\%)
{
 print "ok in mount function now";
 

    my ($this, %args) = @_;
    
    
    my $device = $args{'-device'};
    my $NATE_MNT_CNT = $args{'-num'};
                 
    print "Ok this is my number file NATE: $NATE_MNT_CNT"; 
 
   $device = "$device"."s6";

    my ($mount_point, $read_only);
    my ($perm, $extra_args, $m_dir, $no_cache, $dobj);
    my ($auto, $mount_args, $pass_thru, $loglv);
    my (@output, $status, $m_cmd, $revision);
    
   
   
    $extra_args = "";
    $mount_args = "";

    

    # get the disk object
    $dobj = $this->{dobj};
    
    # get the disk object

    my $type = $this->{volume_type};


    
 
    # check to make sure the user specified a mount over (or mount point)
    # directory (if it does not exist, allow NATE to create directory for
    # them (and return it to the user). If the directory does not exist, 
    # try to create is for the user. If the directory does exist, try to use it.
   
  
   
   if (not defined($mount_point) or $mount_point =~ /\*/) {
        # construct the PATH to the mount point if the user did not 
        # specify a PATH.
        
        
        $m_dir = "/tmp/nate_mnts/". "mnt_" . $PID . "_".  $NATE_MNT_CNT;
      
      
        # create the new directory and check that it was created properly
        $m_cmd = "mkdir -p $m_dir ; chmod 777 $m_dir";
        ($status, @output) = VMwareClient::ClientUtil::execute($this->{conn}, 
                                                  $m_cmd);

        if ($status > 0) {
            &ExitError("Cannot create automatically created mount directory ($m_dir)");
        }

        $mount_point = $m_dir;
    } else { 
        # if the entry does exist, make sure it really is a directory.
        ($status, @output) = VMwareClient::ClientUtil::execute($this->{conn}, 
                                                  "ls -l $mount_point");

        if ($status > 0) {
            &main::logcomment("mount directory does not exist, attempting to create");

            # create the new directory and check that it was created properly
            $m_cmd = "mkdir -p $mount_point ; chmod 777 $mount_point";
            ($status, @output) = VMwareClient::ClientUtil::execute($this->{conn}, 
                                                      $m_cmd);

            if ($status > 0) {
                &ExitError( 
                   "mount point directory ($mount_point) could not be created");
            }
        } else {
            if (grep(/^drwx/,@output)) {
                &main::logcomment("mount directory already exists");
            } else {
                &ExitConfigError("Incompatible file/directory ($mount_point) already exists. Please remove it so it can be used as a mount point");
            }
        }
    }
    # Add mount point to the object
    $this->{mount_point} = $mount_point;

    # check if you want read-only set
    if ($read_only) { 
        $this->{is_readonly} = $ClientUtil::TRUE;
    }

    # change the permissions on the mount directory if asked to
    # only change permission if the mount it NOT read-only
    if (defined($perm) and not $this->{is_readonly} ) {
        $m_cmd = "chmod $perm $mount_point";
        ($status, @output) = VmwareClient::ClientUtil::execute($this->{conn}, 
                                                  $m_cmd);
        if ($status > 0) {
            &ExitError("Error changing permission on $mount_point");
        }
    }
   
    # mount
    if ($auto) {
	# kick out with error for now if we are using hardware initiator
        # Patrick Stub
        if (defined($this->{adapter_info})) {
            &ExitError("kicking out of automount preserve function");
        }

    	# if automount variable is true, use automount file
    	# automount file will differ for different versions
    	if ($this->{hosttype} =~ /sles/i) {
	    # can't use e2label for reiserfs
            if ($type =~ /^reiserfs$/i) {
            	$m_cmd = "echo '$this->{parteddisk} $mount_point $type defaults 0 0' >> /etc/fstab.iscsi";
            } else {
            	$m_cmd = "echo 'LABEL=$this->{fslabel} $mount_point $type defaultse defaults 0 0' >> /etc/fstab.iscsi";
            }
    	} else {
	    # redhat
	    # need to check for the update version, this is RH3U5 and RH4U2
	    $m_cmd = "echo 'LABEL=$this->{fslabel} $mount_point	$type _netdev 0 0' >> /etc/fstab";
        }
        # aren't actually mounting yet so we don't restart the driver one time per lun
        ($status, @output) = VMwareClient::ClientUtil::execute($this->{conn}, $m_cmd);
        if ($status > 0) {
	    &ExitError("Error with echo command: $m_cmd");
        }
    } else {
	# if automount variable is false, mount by hand
        # we are only using labels for ext2 and ext3, not reiser
        if ($type =~ /^reiserfs$/i) {
                $m_cmd = "mount -t $type $this->{parteddisk} $mount_point";
        } else {
                $m_cmd = "mount /dev/dsk/$device $mount_point";
        }
	($status, @output) = VMwareClient::ClientUtil::execute($this->{conn}, $m_cmd);
       
    }

    # mark the filesystem as mounted
    # this is a little out of place here, leaving it for now
    $this->{is_mounted} = $ClientUtil::TRUE;

    # return the mount point to the user
    return($mount_point);

}

1;




