## Package to perform native VMware operations
## This package offers the following
## 1. Cloning
## 2. Deletion
## 3. Snapshots
## 4. Host Migration
## 5. Storage Migration
## @author
## 		dl-nst-virtual@netapp.com

package VMwareOps::Operations::Native::Operations;

use strict;
use warnings;

use Data::Dumper;
use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use VMware::VIRuntime;
use VMware::VILib;

# Define constants for VM state
use constant FRESH => "Non-existent";
use constant CREATE_TRANSIT => "Create-transit";
use constant OFF => "Off";
use constant RUNNING => "Running";
use constant CLONE_TRANSIT => "Clone-transit";
use constant DESTROY_TRANSIT => "Destroy-transit";

# Constructor for creating a new VM object
sub new {
    my $class  = shift;
    my $params = shift;

    my $self = {};

    die "'apiset_obj' is a required NACL apiset object for VM object" if(!defined($params->{apiset_obj}));
    $self->{apiset_obj} = $params->{apiset_obj};

    if(defined($params->{name})) {
        $self->{name} = $params->{name};
    }

    if(defined($params->{source_datastore})) {
        $self->{source_datastore} = $params->{source_datastore};
    }

    if(defined($params->{destination_datastore})) {
        $self->{destination_datastore} = $params->{destination_datastore};
    }

    if(defined($params->{state})) {
        $self->{state} = $params->{state};
    }

    if(defined($params->{vm_container})) {
        $self->{vm_container} = $params->{vm_container};
    }

    if(defined($params->{vi_path})) {
        $self->{vi_path} = $params->{vi_path};
    }
    
    if(defined($params->{vc})) {
        $self->{vc} = $params->{vc};
    }
    
    if(defined($params->{vc_username})) {
        $self->{vc_username} = $params->{vc_username};
    }
    
    if(defined($params->{vc_password})) {
        $self->{vc_password} = $params->{vc_password};
    }
    
    if(defined($params->{template})) {
    	$self->{template} = $params->{template};
    }
    
    if(defined($params->{datastore})) {
    	$self->{datastore} = $params->{datastore};
    }
    
    if(defined($params->{vi_client})) {
    	$self->{vi_client} = $params->{vi_client};
    }
    
    if(defined($params->{host_domain})) {
    	$self->{host_domain} = $params->{host_domain};
    }
    
    bless( $self, $class);

    return $self;	
}

sub create_vm {
	print "Not implemented yet";
	
	return 0;
}

# Routine to destroy VM
sub destroy_vm {
	my $self = shift;
	my $params;
	
	$params->{cmd} = "/usr/software/bin/perl /u/cliffb/p4/test/nst/virtualization/vmware/utils/vm_destroy.pl ";
	$params->{cmd} .= "--url https://" . $self->{vc} ."/sdk/webService ";
	$params->{cmd} .= "--username $self->{vc_username} ";
	$params->{cmd} .= "--password $self->{vc_password} ";
	$params->{cmd} .= "--vm_name $self->{name}";
	
	my $response = $self->_execute_cmd($params);
	
	if($response =~ /deleted successfully/i) {
		print "VM: $self->{name} successfully destroyed\n";
	} elsif ($response =~ /Cannot find VM/i) {
		print "VM: $self->{name} was not found\n";
		return 1;
	} elsif ($response =~ / Power off VM/i) {
		print "VM $self->{name} is not powered off to perform the delete operation\n";
		return 1;
	} else {
		print "VM $self->{name} could not be destroyed. An error occurred\n";
        print "$response\n";
		return 1;
	}
	
	return 0;
}

sub destroy_vm_host {
	my $self = shift;
	my $params = shift;

	$params->{cmd} = "/usr/software/bin/perl /u/cliffb/p4/test/nst/virtualization/vmware/utils/vm_destroy_hosts.pl ";
	$params->{cmd} .= "--url https://" . $self->{vc} ."/sdk/webService ";
	$params->{cmd} .= "--username $self->{vc_username} ";
	$params->{cmd} .= "--password $self->{vc_password} ";
	$params->{cmd} .= "--host $params->{host}";
	
	my $response = $self->_execute_cmd($params);
	
	if($response =~ /deleted successfully/i) {
		print "VM: $self->{name} successfully destroyed\n";
	} elsif ($response =~ /Cannot find VM/i) {
		print "VM: $self->{name} was not found\n";
		return 1;
	} elsif ($response =~ / Power off VM/i) {
		print "VM $self->{name} is not powered off to perform the delete operation\n";
		return 1;
	} else {
		print "VM $self->{name} could not be destroyed. An error occurred\n";
		return 1;
	}
	
	return 0;
}

# Routine to clone VM
sub clone_vm {
	my $self = shift;
	my $params = shift;
	
	# Construct the source-clone command
	$params->{cmd} = "cd $self->{vi_path}; ";
	$params->{cmd} .= "/usr/software/bin/perl ";
	$params->{cmd} .= "$self->{vi_path}/vmclone.pl ";
	$params->{cmd} .= "--url https://" . $self->{vc} ."/sdk/webService ";
	$params->{cmd} .= "--username $self->{vc_username} ";
	$params->{cmd} .= "--password $self->{vc_password} ";
	$params->{cmd} .= "--customize_guest no ";
	$params->{cmd} .= "--customize_vm no ";
	$params->{cmd} .= "--vmname $params->{template} ";
	$params->{cmd} .= "--vmname_destination $params->{name}" . "_" . $params->{vm_container} . " ";
	$params->{cmd} .= "--vmhost $params->{vm_container}.$self->{host_domain} ";
	$params->{cmd} .= "--datastore $params->{source_datastore}";

	# Create temp VM in source datastore
	my $response = $self->_execute_cmd($params);
	print "Output: " . Dumper($response) . "\n";
	if($response =~ /successfully created/i) {
		print "Temporary template successfully cloned";
	} elsif ($response =~ /No virtual machine found/i) {
		print "VM: $params->{template} was not found";
		return 1;
	} elsif ($response =~ /not available/i) {
		print "Datastore $params->{datastore} is not a valid datastore";
		return 1;
	} else {
		print "Problem while creating temporary tempalte. It may already be cloned. Quiting";
		return 1;
	}
	
	# Construct the destination-clone command
	$params->{cmd} = "cd $self->{vi_path}; ";
	$params->{cmd} .= "/usr/software/bin/perl ";
	$params->{cmd} .= "$self->{vi_path}/vmclone.pl ";
	$params->{cmd} .= "--url https://" . $self->{vc} ."/sdk/webService ";
	$params->{cmd} .= "--username $self->{vc_username} ";
	$params->{cmd} .= "--password $self->{vc_password} ";
	$params->{cmd} .= "--customize_guest no ";
	$params->{cmd} .= "--customize_vm no ";
	$params->{cmd} .= "--vmname $params->{name}" . "_" . $params->{vm_container} . " ";
	$params->{cmd} .= "--vmname_destination $params->{name} ";
	$params->{cmd} .= "--vmhost $params->{vm_container}.$self->{host_domain} ";
	$params->{cmd} .= "--datastore $params->{destination_datastore}";

	# Create final VM in destination datastore
	$response = $self->_execute_cmd($params);
	print "Output: " . Dumper($response) . "\n";
	if($response =~ /successfully created/i) {
		print "VM: $self->{name} successfully cloned";
	} elsif ($response =~ /No virtual machine found/i) {
		print "VM: $params->{template} was not found";
		return 1;
	} elsif ($response =~ /not available/i) {
		print "Datastore $params->{datastore} is not a valid datastore";
		return 1;
	} else {
		print "VM $self->{name} was already cloned";
		return 1;
	}
	
	# Destroy the temporary VM
	my $name = $params->{name};
	$self->{name} = $params->{name} . "_" . $params->{vm_container};
	if($self->destroy_vm()) {
		print "Could not destroy temporary VM\n";
		return 1;
	}
	
	$self->{name} = $name;
	$self->{state} = OFF;
	
	return 0;
}

# Routine to perform clone operation
sub clone_vm_single {
	my $self = shift;
	my $params = shift;
	
	# Construct the source-clone command
	$params->{cmd} = "cd $self->{vi_path}; ";
	$params->{cmd} .= "/usr/software/bin/perl ";
	$params->{cmd} .= "$self->{vi_path}/vmclone.pl ";
	$params->{cmd} .= "--url https://" . $self->{vc} ."/sdk/webService ";
	$params->{cmd} .= "--username $self->{vc_username} ";
	$params->{cmd} .= "--password $self->{vc_password} ";
	$params->{cmd} .= "--customize_guest no ";
	$params->{cmd} .= "--customize_vm no ";
	$params->{cmd} .= "--vmname $params->{template} ";
	$params->{cmd} .= "--vmname_destination $params->{name} ";
	$params->{cmd} .= "--vmhost $params->{vm_container}.$self->{host_domain} ";
	$params->{cmd} .= "--datastore $params->{datastore}";

	# Create temp VM in source datastore
	my $response = $self->_execute_cmd($params);
	if($response =~ /successfully created/i) {
		print "Template successfully cloned";
		return 0;
	} elsif ($response =~ /No virtual machine found/i) {
		print "VM: $params->{template} was not found";
		return 1;
	} elsif ($response =~ /not available/i) {
		print "Datastore $params->{datastore} is not a valid datastore";
		return 1;
	} else {
		print "Problem while creating temporary tempalte. It may already be cloned. Quiting";
		return 1;
	}	
}

# Routine to power on VM
sub power_on {
	my $self = shift;
	my $params;
	
	# Construct the power ON command
	$params->{cmd} = "cd $self->{vi_path}; ";
	$params->{cmd} .= "/usr/software/bin/perl ";
	$params->{cmd} .= "$self->{vi_path}/vmcontrol.pl ";
	$params->{cmd} .= "--vmname $self->{name} ";
	$params->{cmd} .= "--username $self->{vc_username} ";
	$params->{cmd} .= "--password $self->{vc_password} ";
	$params->{cmd} .= "--url https://" . $self->{vc} ."/sdk/webService ";
	$params->{cmd} .= "--operation poweron";
	
	my $response = $self->_execute_cmd($params);
	if($response =~ /powered on/i) {
		print "VM: $self->{name} successfuly powered on";
	} elsif ($response =~ /not found/i) {
		print "VM: $self->{name} was not found";
		return 1;
	} else {
		print "VM was already powered on";
	}
	
	$self->{state} = RUNNING;
	return 0;
}

# Routine to migrate VM
sub storage_vmotion_vm {
	my $self = shift;
	my $params = shift;
	
	# Construct the migrate command
	$params->{cmd} = "cd $self->{vi_path}; ";
	$params->{cmd} .= "/usr/software/bin/perl ";
	$params->{cmd} .= "$self->{vi_path}/vmmigrate.pl ";
	$params->{cmd} .= "--vmname $self->{name} ";
	$params->{cmd} .= "--targetpool CopyOffloadPool ";
	$params->{cmd} .= "--username $self->{vc_username} ";
	$params->{cmd} .= "--password $self->{vc_password} ";
	$params->{cmd} .= "--url https://" . $self->{vc} ."/sdk/webService ";
	$params->{cmd} .= "--sourcehost $self->{host} ";
	$params->{cmd} .= "--targethost $self->{host} ";
	$params->{cmd} .= "--state poweredOn ";
	$params->{cmd} .= "--targetdatastore $params->{destination_datastore} ";
	
	my $response = $self->_execute_cmd($params);
	if($response =~ /powered on/i) {
		print "VM: $self->{name} successfuly migrated";
	} elsif ($response =~ /No Virtual Machine found/i) {
		print "VM: $self->{name} was not found";
		return 1;
	} elsif ($response =~ /Datastore not found on target/i) {
		print "$params->{destination_datastore} does not seem to be a valid datastore";
		return 1;
	}
	else {
		print "failed";
		return 1;
	}
	return 0;
}

# Routine to power off VM
sub power_off {
	my $self = shift;
	my $params;
	
	$params->{cmd} = "cd $self->{vi_path}; ";
	$params->{cmd} .= "/usr/software/bin/perl ";
	$params->{cmd} .= "$self->{vi_path}/vmcontrol.pl ";
	$params->{cmd} .= "--vmname $self->{name} ";
	$params->{cmd} .= "--username $self->{vc_username} ";
	$params->{cmd} .= "--password $self->{vc_password} ";
	$params->{cmd} .= "--url https://" . $self->{vc} ."/sdk/webService ";
	$params->{cmd} .= "--operation poweroff";
	
	my $response = $self->_execute_cmd($params);
	if($response =~ /powered off/i) {
		print "VM: $self->{name} successfuly powered off";
	} elsif ($response =~ /not found/i) {
		print "VM: $self->{name} was not found";
		return 1;
	} else {
		print "VM was already powered on";
	}
	
	$self->{state} = OFF;
	return 0;
}

# Routine to get FQDN of host
sub _get_fqdn {
	my $self = shift;
	return 0;
}

# Routine to send commands
sub _execute_cmd {
    my $self = shift;
    my $params = shift;

    my $host = $self->{apiset_obj};
    
    # Giving the operations an hour to finish up
	my $response = $host->execute_raw_command(
                    	'command' => $params->{cmd},
                    	'connectrec-timeout' => 7200
                        );
   	print "Output: " . Dumper($response) . "\n";
    
    return $response;	
}

sub get_state {
	my $self = shift;
	
	return $self->{state};
}

sub set_state {
	my $self = shift;
	my $state = shift;
	$self->{state} = $state;
}

# Verify if a VM exists already
sub exists {
	my $self = shift;
	my $params;
	
	$params->{cmd} = "/usr/software/bin/perl /u/cliffb/p4/test/nst/virtualization/vmware/utils/vm_discover.pl ";
	$params->{cmd} .= "--url https://" . $self->{vc} ."/sdk/webService ";
	$params->{cmd} .= "--username $self->{vc_username} ";
	$params->{cmd} .= "--password $self->{vc_password} ";
	$params->{cmd} .= "--vm_name $self->{name}";
	
	my $response = $self->_execute_cmd($params);
	
	if($response =~ /exists/i) {
		print "VM: $self->{name} exists\n";
		return 1;
	} else {
		print "VM $self->{name} does not exist";
		return 0;
	}
	return 0;
}

1;
