## Package to help Hyper-V opertaions
## This package offers the following
## 1. add vm to ConfigBroker
## 2. delete vm from ConfigBroker
## 3. get vms from ConfigBroker
## 4. print vms from ConfigBroker
## 5. modify vm from ConfigBroker
## 6. get random host from ConfigBroker
## 7. calculate the used percentage
## @author
## dl-nst-virtual@netapp.com

package VMware_Helper;

use strict;
use warnings;

use POSIX qw (ceil floor);
use Data::Dumper;
use ConfigBroker;
use HyperV;

sub new {
    my $class = shift;
    my $cb = shift;

    my $self = {};

    die "'Config Broker' is a required parameter for ConfigBroker Helper" if(!defined($cb));
    $self->{resource_group} = $cb->get_resource_group(name => "vm_rg");;

    bless( $self, $class );
    return $self;
}

sub add_vm {
    my $self = shift;
    my $params = shift;

    my $vm_rg = $self->{resource_group};

    my $row = {
	vm_name    => $params->{name},
	state      => $params->{state},
	datastore  => $params->{datastore},
	host       => $params->{host},
    };

    return $vm_rg->add_resources(
	resource_type    => "vm",
	resourcearrayref => [$row],
    );
}

sub get_vms {
    my $self = shift;
    my (%opts) = @_;

    my $vm_rg = $self->{resource_group};
    my $percent = $opts{percent} || 100;

    my %filter;
    $filter{state} = $opts{state}		if(defined $opts{state});
    $filter{datastore} = $opts{datastore} . '$'	if(defined $opts{datastore});
    $filter{host} = $opts{host}			if(defined $opts{host});

    my $vms;
    if(defined $opts{state} || defined $opts{datastore} || defined $opts{host}) {
	$vms = $vm_rg->get_resources(type => "vm", randomize => 1, regex_filters => { %filter });
    }
    else {
	$vms = $vm_rg->get_resources(type => "vm", randomize => 1);
    }

    my $total = scalar(@$vms);
    my $count = floor($percent / 100 * $total);
    #print("Getting $count VMs out of $total \n");

    my @ret_vms;
    for(my $i = 0; $i < $count; $i++) {
	push( @ret_vms, $vms->[$i] );
    }

    #print("VM list: ". Dumper(\@ret_vms). "\n");
    return \@ret_vms;
}

sub print_vms {
    my $self = shift;
    my $name = shift;

    my $vm_rg = $self->{resource_group};

    my $title = "VM table";
    my $vms;
    if (defined $name) {
	$title .= " on $name";
	my $datastore = $name . '$';
        $vms = $vm_rg->get_resources(type => "vm", regex_filters => { datastore => $datastore });
    }
    else {
	$vms = $vm_rg->get_resources(type => "vm");
    }

    my $i = 1;
    my $ret = sprintf("**************************************** $title ****************************************\n");
    foreach my $vm (@$vms) {
	$ret .= sprintf ("%4d %30s %30s %20s %20s\n", $i++, $vm->{vm_name}, $vm->{datastore}, $vm->{host}, $vm->{state});
    }
    return $ret;
}

sub delete_vm {
    my $self = shift;
    my $params = shift;

    my $vm_rg = $self->{resource_group};
    my $vm_name = $params->{name};

    return $vm_rg->delete_resources(
	type    => "vm",
	regex_filters => { vm_name => $vm_name },
    );
}

sub modify_vm {
    my $self = shift;
    my $params = shift;

    my $vm_rg = $self->{resource_group};

    my %modify_info;
    $modify_info{vm_name} = $params->{name};
    $modify_info{state} = $params->{state}		if(defined $params->{state});
    $modify_info{datastore} = $params->{datastore}	if(defined $params->{datastore});
    $modify_info{host} = $params->{host}		if(defined $params->{host});

    return $vm_rg->modify_resources(
	type    => "vm",
	regex_filters => { vm_name => $params->{name} },
	modify_info => { %modify_info },
    );
}

sub get_random_host {
    my $self = shift;
    my ($hosts, $max_vm_per_host) = @_;

    my @hosts = @$hosts;

    my $vm_rg = $self->{resource_group};

    my $num_host = scalar(@hosts);
    print("All the hosts: @hosts \n");
    while ($num_host > 0) {
	my $host = $hosts[rand($num_host)];
	my $vms = $vm_rg->get_resources(type => "vm", match_filters => {host => $host});

	my $count = scalar(@$vms);
	if($count < $max_vm_per_host) {
	    return $host;
	}

	@hosts = grep { $_ ne $host } @hosts;
	$num_host = scalar(@hosts);
    }

    return 0;
}

sub get_used_percent {
    my $self = shift;
    my ($params, $disk_size) = @_;

    my $percent_used;
    my $need_size = 0;
    my $available_size = 0;
    if(defined $disk_size) {
	if($disk_size =~ /(\d+)\s*gb/i) {
	    $need_size = $1;
	}
	elsif($disk_size =~ /(\d+)\s*mb/i) {
	    $need_size = $1 / 1024;
	}
	elsif($disk_size =~ /(\d+)\s*kb/i) {
	    $need_size = $1 / 1024 / 1024;
	}
	else {
	    print "Disk size '$disk_size' is incorrect format. Check the DISK_SIZE param.\n";
	    return 100;
	}
    }

    my $hv_obj = new HyperV($params);
    my $size_info = $hv_obj->get_used_percent();

    $available_size = ($size_info =~ /Free Space available\:(\d+)/)? $1 : -1;
    $percent_used = ($size_info =~ /UsedPercent\:(\d{1,3})\%/)? $1 : -1;

    if( $percent_used < 0  || $percent_used > 100) {
	print "Invalid used percent: $percent_used.\n";
	return 100;
    }

    if($need_size >= $available_size) {
	print "Require disk size '$disk_size', but only $available_size gb are available.\n";
	return 100;
    }

    return $percent_used;
}

1;
