## @summary     VOPL class for the volume operation tests.
## @author      snider@netapp.com
## @pod here

=head1 NAME

VOPL - volume-operation-package class

=head1 SYNOPSIS

   package Volume;
   use base qw(VOPL);
   
   my $Volume = Volume->new();
  
   my @volumes = $Volume->pick_volumes();

=head1 DESCRIPTION

This package is used as a class for volume operation tests.
This package is uses L<TestManager> as a base class.

=head1 METHODS

none

=cut

package VOPL;
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF BOOLEAN);

use base qw(TestManager);

use Data::Dumper;
use Options;
use Server;
use TracePkg;
my $Trace = new TracePkg;
my $Server = new Server();
my $Options = new Options;

use constant OK => 1;
use constant RETRY => 2;

$Options->parameters(definitions =>
                     [
                      {
                       name => "VOLOP_TIMEOUT",
                       description => "how long to wait on the volume operation[s] to occur",
                       default => 300,
                      },
                      {
                       name => "STRIPE_STYLE",
                       description => "type of striped volume to operate on",
                       valid_values => [ qw(dv mdv all) ],
                       default => "all",
                       spec => "s",
                      },
                      {
                       name => "AFFINITY",
                       description => "node affinity for the volume op",
                       valid_values => [ qw(local remote all) ],
                       default => "all",
                       spec => "s",
                      },
                      {
                       name => "AFFINITY_PREFERENCING",
                       description => "if AFFINITY value cannot be satisfied, try again with AFFINITY=all",
                       spec => "boolean",
                       default => "false",
                      },
                      {
                       name => "PARALLEL",
                       description => "operate on multiple volumes in parallel, else serially. Parallel volume operations will leverage the JobManager",
                       spec => "boolean",
                       default => "true",
                      },
                      {
                       name => "MAX_VOLUMES",
                       description => "number of volumes to randomly choose (or the keyword all). Defaults to 1.",
                       default => 1,
                       subroutine => [\&_max_volumes],
                      },
                      {
                       name => "RANDOM_VOLUMES",
                       spec => "boolean",
                       default => "false",
                       description => "pick any volumes found on the filer",
                      },
                      {
                       name => "STATIC_VOLUME",
                       spec => "boolean",
                       default => "false",
                       description => "Specifies that the volume is static and will turn on additional verification",
                      },
                      {
                       name => "VOLUME_BASE",
                       description => "the base name for volumes - THIS IS EXPERIMENTAL",
                      },
                      {
                       name => "DST_VOLUME_STYLE",
                       description => "The volume style for destination volumes",
                       spec => "s",
                       valid_values => [ qw ( flex striped * ) ],
                      },
                      {
                       name => "DST_AGGR_BLOCK_TYPE",
                       description => "The block type to select for destination Aggregate ",
                       valid_values => [ qw ( 32-bit 64-bit * ) ],
                      },
                      {
                       name => "VOL_PREMOVE_CHECK",
                       description => "Perform volume pre-move check?",
                       spec => "boolean",
                       default => "false",
                      },
                      {
                       name => "VOL_PREMOVE_TARGET_AGGR_CHECK",
                       description => "Perform volume pre-move target aggr check",
                       spec => "boolean",
                       default => "false",
                      },
                      {
                       name => "VOL_PREMOVE_CHECKS_CONTINUE_ON_FAILURE",
                       description => "When a volume pre-move check fails, continue with the move?",
                       spec => "boolean",
                       default => "true",
                      },
                     ],
                    );

# for now, override TestManager to prevent automatic volume creation and mounting 
# since some more work needs to go into TestManager to support passing multiple volumes
sub define_cluster_topology {
    my $self = shift;
    $Trace->entry();
    my $options = $self->get_hash();
    
    # if we have a volume specified, find and set the vserver
    # This, along with many other things will not work with specified volumes in
    #  multiple vservers
    if ($Options->defined(option=>"volume") && !$options->{vserver}) {
        my $volume_array;
        my $volume_name = $options->{volume}[0];
        eval { $volume_array = $Server->volume_show(volume => $volume_name) };
        if ( $volume_array ) {
            if (@$volume_array > 1) {
                $Trace->fatal(msg=>"Found more than one volumes with the name '$volume_name'");
            }
            my $vserver = $volume_array->[0]{vserver};
            $Trace->trace(msg=>"setting vserver to $vserver");
            $Options->set(vserver=>$vserver);
        }
    }
        
    $Trace->exit();
    return {};
}

sub setup_mount_info {
    $Trace->entry();
    $Trace->exit();
    return {};
}

=head2 pick_volumes

   my @volumes = $Volume->pick_volumes();

C<pick_volumes> identifies the list of volumes that the test should work
with. If C<RANDOM_VOLUMES> is true then it will pick random volumes 
discovered in the cluster. The volumes can further be limited by
specifying C<vserver> and C<VOLUME_TYPE>. Otherwise, the user
must specify C<volume> to list the volumes to work with.

This method will also limit the number of volumes to C<MAX_VOLUMES> if
specified.

A striped volume is broken up into its constituent volumes.

The method will return the list of volumes.

=cut
    
sub pick_volumes {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'VOPL' } ],
        {
         get_striped_members => {type => BOOLEAN | UNDEF, default => 1, optional => 1},
        }
    );

    my (@available_volumes);
    $Trace->entry();
    my $options = $self->get_hash();

    # Either RANDOM_VOLUMES, VOLUME_BASE or volume is required.
    if ($options->{RANDOM_VOLUMES}) {
        $Trace->trace(msg=>"picking volumes based on RANDOM_VOLUMES");
        # find all volumes in the cluster that are online, have the correct
        #  volume types and vserver. This will be shortened later.
	my $output = $Server->volume_show(vserver => $options->{vserver},
                                          state => "online",
                                          type => $options->{VOLUME_TYPE},
                                         );
	@available_volumes = grep($_ !~ /^p_/, map($_->{volume}, @{$output}));
    } elsif ($options->{VOLUME_BASE}) {
        $Trace->trace(msg=>"picking volumes based on $options->{VOLUME_BASE}");
        @available_volumes = $self->generate_volume_names(base_name=>$options->{VOLUME_BASE});
    } else {
        $Trace->trace(msg=>"picking volumes based on @{$options->{volume}}");
        # take the list from the volume param
        @available_volumes = @{$options->{volume}};
    }
    $Trace->trace(msg=>"available_volumes=@available_volumes");

    # loop through each of the volumes that have been identified so far
    #  and break down the striped volumes into the constituents
    my @volumes;
    foreach my $volume (@available_volumes) {
        #Get Volume info
        my $volume_hash;
        my $retries = 10;
      VOL_SHOW: {
            $volume_hash = $Server->volume_show(
                                                vserver => $options->{vserver},
                                                volume => $volume,
                                               );
            $volume_hash = $volume_hash->[0];
            if ($volume_hash->{size} eq "-") {
                if (!$retries--) {
                    $Trace->fatal(msg=>"Did not get useful data for volume '$volume':".Dumper($volume_hash));
                } else {
                    $Trace->trace(msg=>"retry lookup of $volume");
                    sleep 1;
                    redo VOL_SHOW;
                }
            }
        }

        if ($volume_hash->{"volume-style"} eq "flex"){
            if ($volume_hash->{"security-style"} eq "-") {
                $Trace->fatal(msg=>"Found bad security-style on the volume:".Dumper($volume_hash));
            }
            push(@volumes, $volume_hash);
        } elsif ($volume_hash->{"volume-style"} eq "striped") {
	    if ($params{get_striped_members} == 1) {
		#Find volume members
		my $volglob;
		if ($options->{STRIPE_STYLE} eq "all") {
		    $volglob = "*";
		} elsif ($options->{STRIPE_STYLE} eq "dv") {
		    $volglob = "_*";
		} elsif ($options->{STRIPE_STYLE} eq "mdv") {
		    $volglob = "";
		} else {
		    $Trace->fatal(msg=>"Invalid stripe style. Exiting.");
		}
		my $vol_members;
		my $retries = 10;
	      VOL_MEM_SHOW: {
		  $vol_members = $Server->volume_member_show(vserver => $volume_hash->{vserver},
							     volume => $volume.$volglob,
							     msid => $volume_hash->{msid},
							     );
		  foreach my $member_hash (@$vol_members) {
		      if ($member_hash->{size} eq "-") {
			  if (!$retries--) {
			      $Trace->fatal(msg=>"Did not get useful data for volume '$member_hash->{volume}':".Dumper($member_hash));
			  } else {
			      $Trace->trace(msg=>"retry lookup of $volume.$volglob");
			      sleep 1;
			      redo VOL_MEM_SHOW;
			  }
		      }
		  }
	      }
		#Pick a constituent to move
		push(@volumes, @$vol_members);
	    } else {
		# here we return the volume has and not the individual member volumes
		push(@volumes, $volume_hash);
	    }
	}
    }
    
    # now that a list of all possible volumes is generated, reduce it down
    # based on the MAX_VOLUMES param
    my @return_volumes; # the list to return
    my $count = ($options->{MAX_VOLUMES} =~ /all/i) ? scalar @volumes : $options->{MAX_VOLUMES};

    # while there are still available volumes in the working list, pick one off 
    while ($count-- && scalar @volumes) {
        # pick a random volume from the working list, add it to the return list
        #  and then remove it from our working list
        my $volume = $volumes[int rand( scalar @volumes)];
        push(@return_volumes, $volume);
        @volumes = grep($_ ne $volume, @volumes);
    }

    $Trace->exit(msg=>"return ".join(" ", map($_->{volume}, @return_volumes)));
    return @return_volumes;
}

=head2 wait_on_jobs

   @volumes = $self->wait_on_jobs(volumes => \@mirrors, 
                                  job_to_volume_hash => \%Jobs, 
                                  sources => $self->{sources});

C<wait_on_jobs> will wait on all currently running jobs to finish.
If the jobs were not run in the background, then it just verifies
that the volumes are back online. This verification is also done
after valid jobs are finished.

The method will return the list of volumes which are not yet online
and need to be retried.

=over

=item volumes

Specifies a reference to a list of volumes. This can be an empty list
if the jobs are running in PARALLEL

=item job_to_volume_hash

Specifies a job to volume list mapping for the PARALLEL case. The hash
has job ids for the keys and a volume list of the values. For move and
copy jobs, the list only has one element. For mirrors, the list contains
the list of mirror hashes. For the list of volumes, the list must contain
a list of volume hashes.

=item sources

Specifies a volume to its source volume mapping. The keys of this hash
contain the names of the volumes being verified. The value contains the 
source volume for the operation. For a move, the values are the same.
For a mirror, the volumes specify the mirrors and the source specifies
the source. For a copy, the volume is the new copy and the source is the
original.

=back

=cut

sub wait_on_jobs {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'VOPL' } ],
        {
         volumes => {type => ARRAYREF},
         job_to_volume_hash => {type => HASHREF | UNDEF, optional => 1},
         sources => {type => HASHREF},
         show_striped_members => {type => BOOLEAN | UNDEF, default => 1, optional => 1},
        }
    );
    
    my $volumes = $params{volumes};
    my $job_to_volume_hash = $params{job_to_volume_hash};
    my $sources = $params{sources};

    my @volumes;
    my @retry;

    $Trace->entry();
    my $options = $self->get_hash();

    # find the list of volumes to verify
    if ($options->{PARALLEL}) {

        if (!$params{job_to_volume_hash} || ref($params{job_to_volume_hash}) ne "HASH") {
            $Trace->fatal(msg=>"job_to_volume_hash should be a hash reference",
                         result=>"SCRIPT");
        }
        
        # in the parallel case, the list of volumes has to be found based upon the
        #  jobs that are finished. The caller has to pass in this mapping.
        $Trace->comment(msg=>"Wait on the jobs to finish");

        # get a list of all jobs that are now completed
        my $job_table = $Server->wait_on_jobs(timeout=>$options->{VOLOP_TIMEOUT});

        # Examine each job to figure out if it needs to be retried. Also find the
        #  list of volumes for the job to verify is online after the op
        foreach my $jobid (keys %$job_table) {
            # get the specific job hash for this job so that it can be verified
            #  and make sure it completed successfully
            my $jobh = $job_table->{$jobid};

            my $volume_list = $job_to_volume_hash->{$jobid};
            if (ref($volume_list) ne "ARRAY") {
                $Trace->comment(msg=>Dumper($job_to_volume_hash));
                $Trace->fatal(msg => "job '$jobid' should point to a list of volume hashes",
                              result => "SCRIPT");
            }

            # Check each volume in the list for this job
            foreach my $volumeh (@$volume_list) {
                if (ref($volumeh) ne "HASH") {
                    $Trace->comment(msg=>Dumper($volume_list));
                    $Trace->fatal(msg => "job '$jobid' should point to a list of volume hashes",
                                  result => "SCRIPT");
                }

                # check the job to see if it needs to be retried
                my $state = $self->check_job(msg=>$jobh->{completion}, 
                                             state=>$jobh->{state}, 
                                             jobid=>$jobid, 
                                             job_table=>$job_table);
                my $volume_name = $self->_volume_name($volumeh);
                if ($state == $self->RETRY()) {
                    my $source = $self->{sources}{$volume_name};
                    if (!grep($source eq $_, @retry)) {
                        $Trace->trace(msg=>"$volume_name needs to be retried");
                        # throw the source volume into the list of volumes to retry
                        push(@retry, $source);
                    }
                } else {
                    $Trace->trace(msg=>"$volume_name was successful");
                    # throw this volume into the list of volumes to check state
                    push(@volumes, $volumeh);
                }
            }
        }
    } else {
        $Trace->trace(msg=>"The jobs were run serially");
        @volumes = @$volumes;
    }

    my $retries = 10;
    my %online;

    # Try 10 times to verify all of the volumes are online
    while ($retries) {
        # @check_volumes is just a copy because @volumes will be modified inside the for loop
        #  and will contain volumes that are not online yet
        my @check_volumes = @volumes;
        foreach my $volume_hash (@check_volumes) {
            my $volume_name = $self->_volume_name($volume_hash);
            $Trace->trace(msg=>"verify that '$volume_name' is online");

            my $vol_state_hash;
            my $retries = 10;
          VOL_SHOW: {
                if ($volume_hash->{"volume-style"} eq "striped" && $params{show_striped_members}){
                    $vol_state_hash = $Server->volume_member_show(
                                                                  vserver => $volume_hash->{vserver},
                                                                  volume => $volume_name,
                                                                 );
                } elsif ($volume_hash->{snapshot}) {
                    $vol_state_hash = $Server->volume_snapshot_show(
                                                                  vserver => $volume_hash->{vserver},
                                                                  volume => $volume_hash->{volume},
                                                                  snapshot => $volume_hash->{snapshot},
                                                                 );
                } else { # mirrors do not have a 'volume-style' so group them with flex volumes
                    $vol_state_hash = $Server->volume_show(
                                                           vserver => $volume_hash->{vserver},
                                                           volume => $volume_name,
                                                          );
                }
                $vol_state_hash = $vol_state_hash->[0];
                if ($vol_state_hash->{size} eq "-") {
                    if (!$retries--) {
                        $Trace->fatal(msg=>"Did not get useful data for volume '$volume_name':".Dumper($vol_state_hash));
                    } else {
                        $Trace->trace(msg=>"retry lookup of $volume_name");
                        sleep 1;
                        redo VOL_SHOW;
                    }
                }
            }
            if ($vol_state_hash->{"security-style"} eq "-" && $vol_state_hash->{"volume-style"} ne "striped") {
                $Trace->fatal(msg=>"Found bad security-style on the volume:".Dumper($vol_state_hash));
            }

            my $source = $sources->{$volume_name};
            if (!$source) { 
                $Trace->comment(msg=>Dumper($sources));
                $Trace->fatal(msg=>"Could not identify the source volume for '$volume_name'",
                              result=>"SCRIPT");
            }

            # Store the volume hash inside the object. This structure will be used
            #  for the comparisons later
            my $source_volume = $self->_volume_name($source);
            push(@{$self->{verify}{$source_volume}{targets}}, $vol_state_hash);

#            $self->{volumes}{$volume_name} = $vol_state_hash;
            if ($vol_state_hash->{snapshot} || $vol_state_hash->{state} eq "online") {
                @volumes = grep($self->_volume_name($_) ne $self->_volume_name($vol_state_hash), @volumes);
                next;
            }
        }
        if (@volumes) {
            $Trace->comment(msg=>"Not all volumes are online. Waiting.");
            $retries--;
            sleep 30;
        } else {
            last;
        }
    }
    if (@volumes) {
        $Trace->comment(msg=>Dumper(\@volumes));
        $Trace->fatal(msg=>"Not all volumes came back online after the volume operation:\n".
                      join("\n", map($_->{volume}, @volumes))."\n");
    }

    $Trace->exit(msg=>"return (".join(" ", @retry).")");
    return @retry;
}

=head2 find_available_aggregates

   @aggregates = $self->find_available_aggregates();

C<find_available_aggregates> will find all available aggregates in the
cluster. If the aggregate is a striped aggregate then the members will
be returned in the list rather than the high level name.

The method will return the list of aggregates.

=over

=item volumes

Specifies a reference to a list of volumes. This can be an empty list
if the jobs are running in PARALLEL

=item job_to_volume_hash

Specifies a job to volume list mapping for the PARALLEL case. The hash
has job ids for the keys and a volume list of the values. For move and
copy jobs, the list only has one element. For mirrors, the list contains
the list of mirror hashes. For the list of volumes, the list must contain
a list of volume hashes.

=item sources

Specifies a volume to its source volume mapping. The keys of this hash
contain the names of the volumes being verified. The value contains the 
source volume for the operation. For a move, the values are the same.
For a mirror, the volumes specify the mirrors and the source specifies
the source. For a copy, the volume is the new copy and the source is the
original.

=back

=cut

sub find_available_aggregates {
    $Trace->entry();

    my @aggregates;
    my $aggregates = $Server->storage_aggregate_show();
    foreach my $aggrh (@$aggregates) {
        if ($aggrh->{voltype} eq "striped") {
            #Find aggregate members but don't add the top-level aggr to the list
            my $aggr_members = $Server->storage_aggregate_member_show(
                                                                      aggregate => $aggrh->{aggregate}."_*",
                                                                     );
            map($_->{voltype} = "striped", @$aggr_members);
            push(@aggregates, @$aggr_members);
        } else {
            # push non-striped aggregates onto the list
            push(@aggregates, $aggrh);
        }
    }

    $Trace->exit(msg=>"return (".join(" ", map($_->{aggregate}, @aggregates)).")");
    return \@aggregates;
}

=head2 pick_destination_aggregates

   $aggregate = $self->pick_destination_aggregate(src_aggr=>$aggr,
                                                  src_node=>$node,
                                                  volume=>$volume_hash,
                                                  aggregates=>\%aggregates);

C<pick_destination_aggregate> will take a list of aggregates and other input
citeria and pick an appropriate aggregate to return.

=over

=item src_aggr

Specifies the name of the current aggregate. If provided, this aggregate
will be removed from the overall list so that it will not be returned.

=item src_node

Specifies the name of the current node in which the volume/aggregate live.
This value is used when C<AFFINITY> is being specified for picking the
new destination.

=item volume

Specifies the full volume hash for the operation. Specifically, the size
field is used from this strucure for verifying how much space is needed
for the new aggregate. In addition, the volume-style field is used to
find the appropriate aggr type.

=item aggregates

Specifies a listing of all possible aggregates to choose from. This must 
be a list of aggregate hashes.

=item aggr_per_node

Specifies that the final list of aggregates should be restricted to this 
number of aggregates per node

=back

=cut

sub pick_destination_aggregate {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'VOPL' } ],
        {
         src_aggr => {type => SCALAR | UNDEF, optional => 1},
         src_node => {type => SCALAR | UNDEF, optional => 1},
         volume => {type => HASHREF},
         aggregates => {type => ARRAYREF},
         aggr_per_node => {type => SCALAR | UNDEF, optional => 1},
         block_type => {type => SCALAR | UNDEF, optional => 1},
        }
    );
    my $dst_voltype;
    my $src_aggr = $params{src_aggr};
    my $src_node = _aggr_nodes_to_array($params{src_node});
    my $volume_hash = $params{volume};

    $Trace->entry(msg=>"$src_node:$src_aggr, aggregates=".join(",", map($_->{aggregate}, @{$params{aggregates}}))."\n");
    my $options = $self->get_hash();

    # eliminate aggregates based on common rules
    # it would be nice if we could leverage the space-guarantee as well as size...
    

    $dst_voltype = $options->{DST_VOLUME_STYLE} || "flex";

    $Trace->comment(msg=>"Destination volume style is set to $dst_voltype");
    my %pick_aggr_parms = (state=>"online",
                           voltype => $dst_voltype,
                           voltype_discovery_mode => 1,
                           avail_size=>$self->size_in_mb(size=>$volume_hash->{size}),
                           aggregate_list=>$params{aggregates},
                          );
    my $dstaggr_block_type = $params{block_type} ? $params{block_type} : $options->{DST_AGGR_BLOCK_TYPE};

    if (defined($dstaggr_block_type)) {
        $Trace->comment(msg=>"Setting aggr block type to $dstaggr_block_type for destination aggr");
        $pick_aggr_parms{block_type} = $dstaggr_block_type;
        $pick_aggr_parms{blocktype_discovery_mode} = 3; #don't let the commandline override us
    }
    $Server->pick_aggregates(%pick_aggr_parms);

    my @aggrh_list;
    if ($src_aggr) {
        # remove the source aggregate from our list so that we don't try to move to it
        @aggrh_list = grep($src_aggr ne $_->{aggregate}, @{$params{aggregates}});
    } else {
        # just use the full list
        @aggrh_list = @{$params{aggregates}};
    }

    $Trace->debug(msg=>"aggrh_list: ".Dumper(\@aggrh_list));
    if (scalar @aggrh_list == 0) {
        $Trace->fatal(msg=>"There are no available aggregates (after removing the current aggregate)\n");
    }

    # now check the AFFINITY for local vs remote - If remote/local is selected 
    # with striped src/destinations, it becomes a bit troublesome as to 
    # what the right thing to do is. This is a first attempt that will need
    # refinement
    my $error;


# Iterate through the aggr list finding aggrs to put the volume onto
    
    my @aggrs;
    aggrloop: for my $aggr (@aggrh_list) {
        my $dst_nodes = _aggr_nodes_to_array($aggr->{nodes});
        $Trace->debug(msg=>"Aggregate:".$aggr->{aggregate}." Nodes :".Dumper($dst_nodes));
        for my $node ( @{$src_node}) {
            if ( $options->{AFFINITY} eq "remote" ) {
                if (grep(/^$node$/,@{$dst_nodes})) {
                    $Trace->debug(msg=>"Skipping local Aggr $aggr->{aggregate}");
                    next aggrloop;
                }
            } elsif ( $options->{AFFINITY} eq "local" ) {
                if (!grep(/^$node$/,@{$dst_nodes})) {
                    $Trace->debug(msg=>"Skipping remote Aggr $aggr->{aggregate}");
                    next aggrloop ;
                }
            }
        }    
        push (@aggrs, $aggr);
        $Trace->debug(msg=>"adding Aggregate :".$aggr->{aggregate});                    
    }
        
    @aggrh_list = @aggrs;
    
    if ( int(@aggrh_list) == 0)  {
        $error = "There are no other available aggregates on the remote nodes to satisfy your AFFINITY=".$options->{AFFINITY}."request\n";
    }


    if ($error) {
        # An error occurred. Before returning the error, check the AFFINITY_PREFERENCING param.
        # If it is set to true then change the AFFINITY to all. Internally, just move on after
        #  the comments since the list is the full list already.
        if ($options->{AFFINITY_PREFERENCING}) {
            $Trace->comment(msg=>"an error occurred with AFFINITIY='$options->{AFFINITY}' so the\n");
            $Trace->comment(msg=>"AFFINITY has been changed to 'all' since AFFINITY_PREFERENCING was true\n");
            # nothing really to do since @aggrh_list is already correct
        } else {
            # AFFINITY must be met and it failed, so return the error
            $Trace->fatal(msg=>$error);
        }
    }

    # create a list of aggregate names
    my @aggr_list = map($_->{aggregate}, @aggrh_list);

    if (my $aggrs = $params{aggr_per_node}) {
        # trim the aggregates down to a list of $aggrs per node
        # This will probably be broken with coral
        my @nodes_seen;
        my @tmp_aggr_list;
        foreach my $aggrh (@aggrh_list) {
            my $node = $aggrh->{nodes};
            if (ref($aggrh->{nodes}) eq "ARRAY") {
                $node = $aggrh->{nodes}[0];
            }
            if (grep($_ eq $node, @nodes_seen) < $aggrs) {
                push(@nodes_seen, $node);
                push(@tmp_aggr_list, $aggrh->{aggregate});
            }
        }
        @aggr_list = @tmp_aggr_list;
    }

    if (scalar @aggr_list == 0) {
        $Trace->comment(msg=>"AFFINITY=$options->{AFFINITY}\nsrc_node=$src_node\n");
        $Trace->comment(msg=>"aggrh_list:".Dumper(\@aggrh_list));
        $Trace->fatal(msg=>"The aggregate list should not be empty.", result=>"SCRIPT");
    }

    if (wantarray) {
        $Trace->exit(msg=>"aggr_list = @aggr_list");
        return @aggr_list;
    } else {
        my $dst_aggr = $aggr_list[int rand(scalar @aggr_list)];
        $Trace->exit(msg=>"return $dst_aggr");
        return $dst_aggr;
    }
}

=head2 find_flex_source

   ($src_node, $src_aggr) = $self->find_flex_source(vserver=>$vserver,
                                                    volume=>$volume,
                                                    aggregates=>$self->{aggregates},
                                                    );

C<find_flex_source> will return the source node and aggregate for the given volume.

=over

=item vserver

Specifies the name of the vserver for the volume

=item volume

Specifies the name of the volume.

=item aggregates

Specifies a listing of all possible aggregates to choose from. This must 
be a list of aggregate hashes.

=back

=cut

sub find_flex_source {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'VOPL' } ],
        {
         vserver => {type => SCALAR},
         volume => {type => SCALAR},
         aggregates => {type => ARRAYREF},
        }
    );
    my $vserver = $params{vserver};
    my $volume = $params{volume};
    my $aggregates = $params{aggregates};

    $Trace->entry(msg=>"$vserver:$volume, aggregates=".join(" ", 
                                                            map($_->{aggregate}, @{$params{aggregates}}))."\n");
    $Trace->debug(msg=>"aggrh_list=".Dumper $aggregates);

    # perform a show on the volume to get all attributes
    my $output = $Server->volume_show(vserver => $vserver,
                                      volume => $volume,
                                      -Timeout => 60,
                                     );

    # identify the aggregate for the given volume
    my $src_aggr = $output->[0]{aggregate};

    # find the aggregate hash for this aggregate from the provided list
    my ($src_aggrh) = $Server->search_topology(cmd_node=>"aggregate", key=>"aggregate", value=>$src_aggr);
    $Trace->debug(msg=>"src_aggr=".Dumper($src_aggrh));

    # find the node for the src aggregate
    my $src_node = $src_aggrh->{filers};
    $Trace->exit(msg=>"$src_node, $src_aggr\n");
    return ($src_node, $src_aggr);
}

=head2 find_striped_source

   ($src_node, $src_aggr) = $self->find_striped_source(vserver=>$vserver,
                                                       volume=>$volume,
                                                       aggregates=>$self->{aggregates},
                                                       );

C<find_striped_source> will return the source node and aggregate for the given volume.

=over

=item vserver

Specifies the name of the vserver for the volume

=item volume

Specifies the name of the constituent volume.

=item aggregates

Specifies a listing of all possible aggregates to choose from. This must 
be a list of aggregate hashes.

=back

=cut

sub find_striped_source {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'VOPL' } ],
        {
         vserver => {type => SCALAR},
         volume => {type => SCALAR},
         aggregates => {type => ARRAYREF},
        }
    );
    my $vserver = $params{vserver};
    my $volume = $params{volume};
    my $aggregates = $params{aggregates};

    $Trace->entry(msg=>"$vserver:$volume, aggregates=".map($_->{aggregate}, @{$params{aggregates}})."\n");

    # perform a show on the volume to get all attributes
    my $output = $Server->volume_member_show(vserver => $vserver,
                                             volume => $volume,
                                             -Timeout => 60,
                                            );

    # find the aggregate of this constituent volume
    my $src_aggr = $output->[0]{aggregate};
    $Trace->trace(msg=>"src_aggr=$src_aggr");


    # find the aggregate hash for this aggregate from the provided list
    my $src_aggrh;
    foreach my $aggrh (@{$aggregates}) {
        if ($aggrh->{aggregate} eq $src_aggr) {
            $src_aggrh = $aggrh;
            last;
        }
    }
    $Trace->debug(msg=>"src_aggr=".Dumper($src_aggrh));

    # find the node for the src aggregate
    my $src_node = $src_aggrh->{node};

    $Trace->exit(msg=>"$src_node, $src_aggr\n");
    return ($src_node, $src_aggr);
}

=head2 check_job

    $self->check_job(msg=>$@);

C<check_job> will check the specified job to see if it was successful
based upon the output message;

=over

=item jobid

Specifies the job ID for the job.

=item state

Specifies the state of the job.

=item msg

Specifies the error message to evaluate. This can be from a fg or bg job.

=back

=cut

sub check_job {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'VOPL' } ],
        {
         jobid => { type => SCALAR | UNDEF, optional => 1 },
         state => { type => SCALAR | UNDEF, optional => 1 },
         msg => { type => SCALAR | UNDEF, default => "", optional => 1 },
         job_table => { type => HASHREF | UNDEF, optional => 1 },
        }
    );
    my $msg = $params{msg};
    my $state = $params{state};

    my $code;
    $Trace->entry(msg=>"state=$state, msg=>$msg");

    # Issue warnings for some specific error conditions, fail for anything else
    if ($state && $state =~ /success/i) {
        $code = OK();
        $Trace->trace(msg => "Successful status reported; message: $msg");
    } elsif ($msg =~ m/^Volume is busy/ ||
             $msg =~ m/^Volume busy/ ||
             #       $msg =~ m/space on aggregate/ || # not enough dest aggr space for vol
             #       $msg =~ m/Queued job/x || # job queues when required node is unavailable (problem)
             #       $msg =~ m/= 524391/x || # volume is restricted
             #       $msg =~ m/= 524425/x || # attempting to move within same aggr
             $msg =~ m/VLDB could not initiate a/ ||
             $msg =~ m/Dumer could not establish/ ||
             $msg =~ /Volume operation Running/i ||
             $msg =~ /Volume is busy with a mirror/i ||
             $msg =~ /Cannot operate on the volume since a snapmirror operation is in progress/i
            ) {
        $Trace->trace(msg => "Retryable message encountered: $msg");
        $code = RETRY();
    } elsif ( $msg && $msg !~ /Success/i) {
        my $job_msg = "";
        if ($params{jobid}) {
            $job_msg = "for job $params{jobid}";
            my $job_table = $params{job_table};
            $Trace->comment(msg=>Dumper $job_table->{$params{jobid}});
        }
        $Trace->fatal(msg=>"Unexpected error $job_msg: $msg", result=>"FAIL");
    }
    $Trace->exit(msg=>"return $code");
    return $code;
}

=head2 compare_volumes

    $self->compare_volumes(source => $volume_hash,
                           targets => $self->{verify}->{$volume}->{targets},
                           specific_fields => $self->{verify}->{$volume}{specific},
                           skip_fields => [qw(dsid)],
                          );


C<compare_volumes> will compare a source volume to a list of destination volumes.
Mirror jobs will have multiple destinations, whereas move and copy will have one.

=over

=item source

Specifies the source volume hash.

=item targets

Specifies a list of target volume hashes to compare against the source.

=item specific_fields

Specifies a hash of target volume hashes which indicate the values that
each destination volume should have. The keys represent the name of the
target volume and the value is a hash of specific attributes that should
exist. These are attributes that are known ahead of time to be changing
to specific values.

=item skip_fields

Specifies a list of fields that are allowed to be different. Additionally,
C<STATIC_VOLUME> can change the fields that are allowed to be different.

=back

=cut

sub compare_volumes {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'VOPL' } ],
        {
            source => {type => HASHREF},
            targets=> {type => ARRAYREF},
            specific_fields => {type => HASHREF},
            skip_fields => {type => ARRAYREF},
        }
    );

    # maybe we should compare source to source as well
    $Trace->entry();

    # if a volume is not static, then add additional attributes which are allowed to be
    # different.
    my @skip = @{$params{skip_fields}};
    if ( ! $self->get(parameter=>"STATIC_VOLUME")) {
        # if the volumes under test are not static, then they are likely being written to so
        #  add a few more size-related keys that are allowed to be changing
        push(@skip, qw(files files-used total used percent-used size available size-used-by-snapshots));
    }

    push @skip, qw(junctionFileHandle);

    # examine each target volume and compare it to the source volume
    my @errors;
    foreach my $target (@{$params{targets}}) {
        # create a temp hash for this target
        my %target = %{$target};
        my %source = %{$params{source}};

        my %comparison;
        if ($params{specific_fields}{$target->{volume}}) {
            # If there were specific fields specified, then make a local copy of
            # the source and override with the specific values we will be comparing
            # against.
            %comparison = (%source, %{$params{specific_fields}{$target->{volume}}});
        } else {
            # just set the local copy to the source
            %comparison = %source;
        }

        my @volume_skip = @skip;
        if ($source{"volume-style"} eq "striped") {
            push(@volume_skip, qw(junction-path junction-path-source parent junction-active user group
                                  language security-style unix-permissions));
            if ($self->get(parameter=>"DST_VOLUME_STYLE") eq "flex") {
                push(@volume_skip, qw(stripe-width volume-style stripe-format properly-striped concurrency 
                                      striping-epoch optimize stripe-count));
            }
        }

        $Trace->trace(msg=>"compare '$target{volume}' to '$source{volume}'");

        # Now that we have a local source and destination hash, compare each field to see
        #  if they are different. Do not fail if the differences are in the diff list.
        my $diffs;
        foreach my $key (keys %comparison) {
            if ( ! grep($_ eq $key, @volume_skip) &&
                 ( $comparison{$key} ne $target{$key} )) {
                $diffs = "Found the following differences comparing volume '$target{volume}' to $source{volume}:\n" if (!$diffs);
                $diffs .= "    $key is '$target{$key}' rather than '$comparison{$key}'\n";
            }
        }
        push(@errors, $diffs) if ($diffs);
    }
    if (@errors) {
        $Trace->fatal(msg=>"Encountered the following errors while verifying the volumes:\n".
                      join("\n", @errors));
    }
    $Trace->exit();
}

# return the volume name for a mirror or a volume
sub _volume_name {
    my ($self, $volume_hash) = @_;

    $Trace->entry(msg=>"");

    my $volume_name;
    if ($volume_hash->{snapshot}) {
        $volume_name = $volume_hash->{snapshot};
    } elsif ($volume_hash->{volume}) {
        $volume_name = $volume_hash->{volume};
    } elsif ($volume_hash->{mirror}) {
        $volume_name = $volume_hash->{mirror};
    } else {
        $Trace->fatal(msg=>"Could not determine the name of the volume:".Dumper($volume_hash),
                      result=>"SCRIPT");
    }
    $Trace->exit(msg=>"$volume_name");
    return $volume_name;
}

# verification routine for the MAX_VOLUMES param
sub _max_volumes {
    my ($max_volumes) = @_;
    $Trace->entry(msg=>"max_volumes=$max_volumes");
    
    if ($max_volumes !~ /^all$|^\d+$/i) {
        $Trace->fatal(msg=>"$max_volumes needs to be either an integer or the word 'all'");
    }
    $Trace->exit(msg=>"max_volumes=$max_volumes");
}


sub _aggr_nodes_to_array {
	my $filer = shift;   
	if (ref($filer) ne "ARRAY"){
	    # If it's not an array, turn it into one
	    my @a = split(/,/, $filer);

	    # Only use unique keys though since a filer can be listed twice and that would screw up the aggr check
	    my %items;
	    foreach my $key (@a){
		$items{$key} = 1;
	    }
	    @a = keys %items;
	    $filer = \@a;
	}
    return $filer;
}    


=head1 PARAMS

=over

=item VOLOP_TIMEOUT

Specifies how long to wait on the volume operation[s] to occur. Defaults to 300 seconds.

=item STRIPE_STYLE

Specifies the type of striped volume to operate on. Defaults to "all".
Possible valuess are: dv mdv all

=item AFFINITY 

Specified the node affinity for the volume op. Defaults to "all". 
Possible values are: local remote all


=item AFFINITY_PREFERENCING

Specifies what to do if the C<AFFINITY> value cannot be satisfied. If set to true,
then try again with C<AFFINITY>=all. Defaults to false.

=item PARALLEL

Specifies that multiple volumes should be operated in parallel rather than serially.
Parallel volume operations will leverage the JobManager. Defaults to true.

=item MAX_VOLUMES

Specifies the number of volumes to act on during an iteration (or the keyword all).
Defaults to 1.

=item RANDOM_VOLUMES

Specifies that random volumes should be specified rather than using C<volumes>,

=item STATIC_VOLUME

Specifies that the volume is static and will turn on additional verification.

=back

=head1 SEE ALSO

L<Options>
L<TestManager>

=head1 BUGS



=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/VOPL.pm#2 $
  Last revised $Date: 2013/02/07 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
