## @summary     VOPL::Clone class for the volume operation tests.
## @author      snider@netapp.com
## @pod here

=head1 NAME

VOPL::Clone - clone volume-operation-package class

=head1 SYNOPSIS

   package CreateClone;
   use base qw(VOPL::Clone);
   
   my $Volume = CreateClone->new();
  
   my @volumes = $Volume->pick_volumes();

=head1 DESCRIPTION

This package is used as a class for clone tests.
This package uses L<VOPL> as a base class.

=head1 METHODS

none

=cut

package VOPL::Clone;
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);

use base qw(VOPL);

use Data::Dumper;
use Options;
use Server;
use TracePkg;
use constant MAX => 3;
my $Trace = new TracePkg;
my $Server = new Server();
my %Jobs;
my $Options = new Options;

# Hashes Used By Lib:
# Used to compare volume attributes in verify_volumes created by wait_on_jobs
#$self->{verify} = {
#            'source_vol' => {
#                                   'targets' => [
#                                                     {<clone_volume1_hash>},
#			           {<clone_volume2_hash>},
#			           ],
#                                    'specific'=> {
#			            'clone_volume1' => {specific_fields_to_verify},
#     			            'clone_volume1' => {specific_fields_to_verify},
#			          },
#		        },
#	}
#
#Created by clone_volumes for wait_on_jobs
#$self->{sources} = (
#	   child_volume =>	{<parent_volume_hash>},
#	);
#
#
#$self->{dest_vols} = {
#		volume => {
#			   parent-volume => volume,
#			   parent-snapshot => snapshot,
#			   junction-path => path,
#			   vserver => vserver
#			}
#		}

# This is the main workhorse routine for this test. There is no test_setup.
# The VOPL library also forces filer_* and client_* to do nothing.
# Basically, each iteration will pick volumes to clone, pass them to the clone_volume code,
#  wait on the copies to finish (retrying until the copies eventually go through) and then
#  verifies that the resulting volume attributes are what they are supposed to be.


$Options->parameters(definitions=>
                  [
                   {
                    name => "NUMBER_OF_CLONES",
                    description => "Number of clone volumes to create",
                    default => "1",
                   },
                   {
                    name => "SNAPSHOT_PER_CLONE",
                    description => "Create a separate snapshot for each clone",
                    default => "false",
                   }, 
                   ]);



sub test_main {
    my $self = shift;
    my @invols = @_;
    my @volumes;
    $Trace->entry();

    $self->{verify} = undef;
    #$self->{sources} = undef;
    #$self->{dest_vols} = undef;
    
    my $release = $Server->get_release(format=>"cooked");
    if ($release =~ /tricky.3/i) {
        $self->{skip_fields} = [qw(dsid msid junction-path junction-path-source junction-active parent create-time crtime vsroot)];
    }
    else { #Assume Boilermaker or higher.
        $self->{skip_fields} = [qw(dsid msid junction-path junction-path-source junction-active parent create-time crtime vsroot uuid clone-volume-name cloned-snapshot)];
	
    }
    if (@invols) {
        $Trace->debug(msg=>"Invols:".Dumper(@invols));
        @volumes = @invols;
    } else {
        @volumes = $self->pick_volumes();
        $Trace->debug(msg=>"Volumes:".Dumper(@volumes));
    }
    my @original_volumes = @volumes;
    while (@volumes) {
        my @copies = $self->clone_volume(@volumes);
        # $self->{sources} is created inside of clone_volume()
        # %Jobs is created inside of clone_volume()
        @volumes = $self->wait_on_jobs(volumes => \@copies, 
                                       job_to_volume_hash => \%Jobs, 
                                       sources => $self->{sources});
        $Trace->comment(msg=>"Waiting on Vols ".@volumes);

    }
    $self->verify_volumes(@original_volumes);
        $Trace->debug(msg=>"Main \n sources:".Dumper(\$self->{sources})."\n verify:".Dumper(\$self->{verify}));    
    
    # we should have already died on a failure...
    $Trace->comment(msg=>"All volumes were successfully cloned.");

    $Trace->exit();
}

# this code will remove the volume that was created during the test_main
sub test_cleanup {
    my $self = shift;
    $Trace->entry();
    my $options = $self->get_hash();
    my @errors;
    
    my %clones = %{$self->{sources}};
    my @vols_sorted = sort(keys %clones);
    @vols_sorted = reverse(@vols_sorted);
    foreach my $volume (@vols_sorted) {
        $Trace->comment(msg=>"delete $volume");

        eval {
            $Server->volume_purge(vserver => $clones{$volume}->{vserver},
                                  volume => $volume,
                                  foreground => 'true');
        };
        push(@errors, $@) if ($@ && $@ !~ /There are no entries/ && $@ !~ /entry doesn't exist/);
    }
    

    if (scalar @errors) {
        $Trace->fatal(msg=>"Failure to delete at least one of the volumes:".
                      join("\n", @errors));
    }

    $Trace->exit();
}

# This code takes and copies the specified volumes to new locations based upon
#  the type of volume, the settings for AFFINITY, AFFINITY_PREFERENCING, etc.
# If the jobs are supposed to occur in parallel, they will be passed as background
#  jobs and handled via the JobManager.
sub clone_volume {
    my $self = shift;
    my @volumes = @_;

    $Trace->entry();
    my $options = $self->get_hash();
    my $src_aggr; 
    my $src_node;
    my $src_aggrh;
    my $clone_cmd;
    my $output;
    my @copies;
    my $have_snapshot = 0;
    #@{$self->{dest_vols}} = (); #clear the clone array for each iteration
    $self->{dest_vols} = {}; #clear the clone hash for each iteration
    # force a refresh of the aggregates during each iteration
    $self->{aggregates} = $Server->storage_aggregate_show();

    foreach my $volume_hash (@volumes) {
        my $vserver = $volume_hash->{vserver};
        my $volume = $volume_hash->{volume};
        my $snapshot;

        # examine the volume hash and determine the correct clone call as well
        #  as its source node and aggregate
        if ($volume_hash->{"volume-style"} eq "striped") {
            $Trace->fatal(result=>"CONFIG", msg=>"Striped volumes are not supported for clone operations");
        } elsif ($volume_hash->{"volume-style"} eq "flex"){
            $clone_cmd = "volume_clone_new";

            ($src_node, $src_aggr) = $self->find_flex_source(vserver=>$vserver,
                                                             volume=>$volume,
                                                             aggregates=>$self->{aggregates},
                                                             );
        } else {
            $Trace->fatal(msg=>"Unknown volume type '".$volume_hash->{"volume-style"}."':".Dumper($volume_hash));
        }
     

         
        my $dst_aggr = $src_aggr; # a clone lives on the src_aggr, and takes no additional space.
        # An aggregate has been picked. Now deduct the size of this volume from it to minimize the risk
        # of not having enough space when multiple volumes are being cloned. (meaning, we shouldn't pick
        # it again for another volume if the resulting size will be too small)
        #        my ($dst_aggrh) = grep($_->{aggregate} eq $dst_aggr, @{$self->{aggregates}});
        #$dst_aggrh->{availsize} -= $self->size_in_mb(size=>$volume_hash->{size});
        $self->{dst_volume} = $volume_hash->{volume}."_clone"; # Template for volume clones
        $self->{dst_vserver} = $vserver;
        
        $Trace->comment(msg=>"$clone_cmd $vserver $volume_hash->{volume} ($src_aggr) to $self->{dst_volume} ($dst_aggr) Parallel $options->{PARALLEL}");
        my $fg = ($options->{PARALLEL}) ? "false" : "true";
        my $output;

       ##Check if a snapshot was taken.   
        unless ($have_snapshot) {
            $snapshot = $self->{testname}."_".$options->{TESTCASE}."_".time;
            eval {
                  $Server->volume_snapshot_new(
	                                        vserver => $vserver,
                                                volume => $volume,
                                                snapshot => $snapshot,
                                            );
	    };
            if ($@) {
                if ($self->check_job(msg=>$@) == $self->RETRY()) {
                    $Trace->comment(msg=>"Volume operation already running on $volume. Sleeping and retrying");
                    sleep 5;
                    redo;
                } else {
                    $Trace->fatal(msg=>"Unexpected error: $@");
                }
            }
            if ($options->{PARALLEL}) {
                $Server->wait_on_jobs();
            }
            $have_snapshot =1;
	}
    for my $idx (1..$options->{NUMBER_OF_CLONES}) {
	$self->{dst_volume} = $volume."_clone_".$idx;
        my $clone_hash = {
                 volume => $self->{dst_volume},
                 vserver => $vserver,
                 'clone-volume' => 'true',                               
                };
        eval {
 
            $output = $Server->$clone_cmd(
                             vserver => $vserver,
                             flexclone => $self->{dst_volume},
                             'parent-volume'  => $volume,
                             'parent-snapshot'  => $snapshot,
                             foreground => $fg,
                             'junction-path' => "/".$self->{dst_volume},
                             'junction-active' => 'true',
                             'comment' => $volume_hash->{'comment'},
                            );    
                
        
            # Create the structures needed for later verification. In the case of clone,
            #  the new volume points to the original source hash has the value. 
            $self->{sources}{$clone_hash->{volume}} = $volume_hash;

            # set the new specific field to indicate that the volume should have the
            #  specific attributes
            $self->{verify}{$volume}{specific}{$clone_hash->{volume}} = $clone_hash;
            push(@copies, $clone_hash);
            
            $self->{dest_vols}{$clone_hash->{volume}} = {
                                                         'parent-volume' => $volume,
                                                         'parent-snapshot' => $snapshot,                                                  
                                                         'junction-path' => "/".$self->{dst_volume},
                                                         'vserver' => $vserver,
                                                        }
            
        };
        if ($@) {
            if ($self->check_job(msg=>$@) == $self->RETRY()) {
                $Trace->comment(msg=>"Volume operation already running on $volume. Sleeping and retrying");
                sleep 5;
                redo;
            } else {
                $Trace->fatal(msg=>"Unexpected error: $@");
            }
        }
        
        if ($options->{PARALLEL}) {
            # the output will be the job number
            push(@{$Jobs{$output}}, $clone_hash);
        }
    }
    $Trace->comment(msg => "dest vols:".Dumper($self->{dest_vols}));    
    }
    $Trace->exit(msg=>"return ".join(" ", @copies)."\n sources:".Dumper(\$self->{sources})."\n verify:".Dumper(\$self->{verify}));
    return @copies;
}

sub split_volumes {
    my $self = shift;
    $Trace->entry();


    my $release = $Server->get_release(format=>"cooked");
    if ($release =~ /tricky.3/i) {
        $self->{skip_fields} = [qw(dsid msid junction-path junction-path-source junction-active parent create-time crtime vsroot)];
    }
    my @volumes;
    # get the list of volumes that we are going to split from the verify hash created
    # During the clone create phase, then reset the verify hash 

    for my $volume (keys(%{$self->{verify}})) {
        push(@volumes,@{$self->{verify}{$volume}{targets}}); 
    }

    #map(push(@volumes,@{$self->{verify}{$_}{targets}}),keys($self->{verify}));
    $Trace->debug(msg=>"Volumes to split:".Dumper(\@volumes));    
    $self->{verify} = undef;
    #$self->{sources} = undef; Might want to reset this. not sure yet.
    
    my @original_volumes = @volumes;
    while (@volumes) {
        my @copies = $self->_split_volume(@volumes);
        # $self->{sources} is created inside of clone_volume()
        # %Jobs is created inside of clone_volume()
        @volumes = $self->wait_on_jobs(volumes => \@copies, 
                                       job_to_volume_hash => \%Jobs, 
                                       sources => $self->{sources});
        $Trace->comment(msg=>"Waiting on Vols ".@volumes);

    }
    #$self->verify_volumes(@original_volumes);
        $Trace->debug(msg=>"Main \n sources:".Dumper(\$self->{sources})."\n verify:".Dumper(\$self->{verify}));    
    
    # we should have already died on a failure...
    $Trace->comment(msg=>"All volumes were successfully split.");

    $Trace->exit();
}

sub _split_volume {
    my $self = shift;
    my @volumes = @_;
    my $output;
    $Trace->entry();
    my $options = $self->get_hash();
    my $src_aggr; 
    my $src_node;
    my $src_aggrh;
    my $clone_cmd;
    my $output;
    my @copies;
    my $have_snapshot = 0;
    #@{$self->{dest_vols}} = (); #clear the clone array for each iteration
    #$self->{dest_vols} = {}; #clear the clone hash for each iteration
    #This should already be set in clone create
    
    # force a refresh of the aggregates during each iteration
    $self->{aggregates} = $Server->storage_aggregate_show();

    foreach my $volume_hash (@volumes) {
        my $vserver = $volume_hash->{vserver};
        my $volume = $volume_hash->{volume};
        my $snapshot;

        # examine the volume hash and determine the correct clone call as well
        #  as its source node and aggregate
        if ($volume_hash->{"volume-style"} eq "striped") {
            $Trace->fatal(result=>"CONFIG", msg=>"Coral volumes are not supported for clone operations");
        } elsif ($volume_hash->{"volume-style"} eq "flex"){
            $clone_cmd = "volume_clone_split";

            ($src_node, $src_aggr) = $self->find_flex_source(vserver=>$vserver,
                                                             volume=>$volume,
                                                             aggregates=>$self->{aggregates},
                                                             );
        } else {
            $Trace->fatal(msg=>"Unknown volume type '".$volume_hash->{"volume-style"}."':".Dumper($volume_hash));
        }
     
        #############################
        #Pick an aggregate to split clone to
        #############################

#        my $dst_aggr = $self->pick_destination_aggregate(src_node=>$src_node,
#                                                         volume=>$volume_hash,
#                                                         aggregates=>$self->{aggregates});
         
        my $dst_aggr = $src_aggr; # a clone lives on the src_aggr, and takes no additional space.
        # An aggregate has been picked. Now deduct the size of this volume from it to minimize the risk
        # of not having enough space when multiple volumes are being cloned. (meaning, we shouldn't pick
        # it again for another volume if the resulting size will be too small)
        #        my ($dst_aggrh) = grep($_->{aggregate} eq $dst_aggr, @{$self->{aggregates}});
        #$dst_aggrh->{availsize} -= $self->size_in_mb(size=>$volume_hash->{size});

        $self->{dst_volume} = $volume_hash->{volume};
        $self->{dst_vserver} = $vserver;
        
        $Trace->comment(msg=>"$clone_cmd $vserver $volume_hash->{volume} ($src_aggr) to $self->{dst_volume} ($dst_aggr) Parallel $options->{PARALLEL}");
        my $fg = ($options->{PARALLEL}) ? "false" : "true";
        my $output;

       ##Check if a snapshot was taken.   

        my $clone_hash = {
                 volume => $self->{dst_volume},
                 vserver => $vserver,
                 aggregate => $dst_aggr,
                 'volume-style' => $volume_hash->{'volume-style'},
                 
                };
        eval {
 
            $output = $Server->$clone_cmd(
                             vserver => $vserver,
                             flexclone => $self->{dst_volume},
                             foreground => $fg,
                            );    
                
        
            # Create the structures needed for later verification. In the case of clone split,
            #  the split volume points to the original source hash has the value. 
            $self->{sources}{$clone_hash->{volume}} = $volume_hash;

            # set the new specific field to indicate that the volume should have the
            #  specific attributes
            $self->{verify}{$volume}{specific}{$clone_hash->{volume}} = $clone_hash;
            push(@copies, $clone_hash);
            #push(@{$self->{dest_vols}},$clone_hash->{volume});
            #$self->{dest_vols}{$clone_hash->{volume}} = {
            #                                             'parent-volume' => $volume,
            #                                             'parent-snapshot' => $snapshot,
            #                                             'junction-path' => "/".$self->{dst_volume},
            #                                            }
            
        };
        if ($@) {
            if ($self->check_job(msg=>$@) == $self->RETRY()) {
                $Trace->comment(msg=>"Volume operation already running on $volume. Sleeping and retrying");
                sleep 5;
                redo;
            } else {
                $Trace->fatal(msg=>"Unexpected error: $@");
            }
        }
        
        if ($options->{PARALLEL}) {
            # the output will be the job number
            push(@{$Jobs{$output}}, $clone_hash);
        }
    
    }
    $Trace->exit(msg=>"return ".join(" ", \@copies)."\n sources:".Dumper(\$self->{sources})."\n verify:".Dumper(\$self->{verify}));
    return @copies;
}

# Take list of volumes that moved and verify that the volume attributes are correct.
sub verify_volumes {
    my $self = shift;
    my @volumes = @_;

    $Trace->entry();

    foreach my $volume_hash (@volumes) {
        my $volume = $volume_hash->{volume};
        # specific_fields gets set during clone_volume.
        # targets get set by wait_on_jobs and is derived from the sources set in clone_volume
        $self->compare_volumes(source => $volume_hash,
                               targets => $self->{verify}->{$volume}->{targets},
                               specific_fields => $self->{verify}->{$volume}{specific},
                               skip_fields => $self->{skip_fields},
                              );
    }
    $Trace->exit();
}

sub mount_in_namespace {
    my $self = shift;
    my @mount;
    $Trace->entry();
    for my $volume (keys(%{$self->{dest_vols}})) {
        #$Server->volume_mount(vserver => $self->{dst_vserver},
        #           volume => $volume,
        #           'junction-path' => "/$volume");
        push(@mount,"/.admin/$self->{dest_vols}{$volume}{'junction-path'}");
    }
    $Trace->exit(msg=>"return ".Dumper(@mount));
    
    return @mount;
}

=head1 PARAMS

=over

=back

=head1 SEE ALSO

L<VOPL>
L<Options>
L<TestManager>

=head1 BUGS



=head1 STATUS

  Shared

=head1 AUTHOR

  Al Straub (astraub@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/VOPL/Clone.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Cloneright 2006, Network Appliance., All Rights Reserved

=cut

1;
