## @summary     VOPL::Copy class for the volume operation tests.
## @author      snider@netapp.com
## @pod here

=head1 NAME

VOPL::Copy - copy volume-operation-package class

=head1 SYNOPSIS

   package CreateCopy;
   use base qw(VOPL::Copy);
   
   my $Volume = CreateCopy->new();
  
   my @volumes = $Volume->pick_volumes();

=head1 DESCRIPTION

This package is used as a class for copy tests.
This package uses L<VOPL> as a base class.

=head1 METHODS

none

=cut

package VOPL::Copy;
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);

use base qw(VOPL);

use Data::Dumper;
use Options;
use Server;
use TracePkg;
use constant MAX => 3;
my $Trace = new TracePkg;
my $Server = new Server();
my %Jobs;

# This is the main workhorse routine for this test. There is no test_setup.
# The VOPL library also forces filer_* and client_* to do nothing.
# Basically, each iteration will pick volumes to copy, pass them to the copy_volume code,
#  wait on the copies to finish (retrying until the copies eventually go through) and then
#  verifies that the resulting volume attributes are what they are supposed to be.

sub test_main {
    my $self = shift;
    $Trace->entry();

    $self->{verify} = undef;
    my $options = $self->get_hash();
    my $release = $Server->get_release(format=>"cooked");

# Remove 'snapshot-space-used' from skip after burt325395 is fixed
    if ($release =~ /tricky.0/i) {
        $self->{skip_fields} = [qw(dsid msid ruleset junction-path junction-path-source junction-active parent crtime vsroot snapshot-space-used )];
    } else {
        if (   $options->{VOLUME_STYLE} ne $options->{DST_VOLUME_STYLE}
            && $options->{VOL_STRIPE_FORMAT} eq "full" ) {
            $self->{skip_fields} = [
                qw(dsid msid junction-path junction-path-source junction-active
                    parent create-time crtime vsroot uuid node filer fsid
                    snapshot-space-used index-dir-enabled instance-uuid
                    provenance-uuid node filer snapshot-count)
            ];
        } else {
            $self->{skip_fields} = [
                qw(dsid msid junction-path junction-path-source junction-active
                    parent create-time crtime vsroot uuid node filer fsid
                    snapshot-space-used instance-uuid provenance-uuid
                    sis-space-saved-percent dedupe-space-saved-percent
                    sis-space-saved dedupe-space-shared dedupe-space-saved
                    compression-space-saved compression-space-saved-percent
                    junction-parent node filer snapshot-count)
            ];
        }
    }

    my @volumes = $self->pick_volumes(get_striped_members => 0);
    my @original_volumes = @volumes;
    while (@volumes) {
        my @copies = $self->copy_volume(@volumes);
        # $self->{sources} is created inside of copy_volume()
        # %Jobs is created inside of copy_volume()
        @volumes = $self->wait_on_jobs(volumes => \@copies, 
                                       job_to_volume_hash => \%Jobs, 
                                       sources => $self->{sources},
                                       show_striped_members => 0);
    }
    $self->verify_volumes(@original_volumes);

    # we should have already died on a failure...
    $Trace->comment(msg=>"All volumes were successfully copied.");

    $Trace->exit();
}

# this code will remove the volume that was created during the test_main
sub test_cleanup {
    my $self = shift;
    $Trace->entry();

    my @errors;
    my @volumes = $self->pick_volumes(get_striped_members => 0);

    my $options    = $self->get_hash();
    my $cp_vol_cnt = 0;

    if ($options->{FILER_CLEANUP_PRECHECK}) {

        $Trace->comment(msg => "The contents of test_topology before pushing: " .
        Dumper $self->{test_topology});

        my %test_topo = %{$self->{test_topology}};
        my @orig_vols = ();

        # Add the copied volume to the test topology so it will be
        # seen by _checkers. 
        foreach my $volume_hash (@volumes) {
            my %copy_vol_hash = ();
            $copy_vol_hash{volume}  = $volume_hash->{volume}."_copy";
            $copy_vol_hash{vserver} = $volume_hash->{vserver};
            $Trace->debug(msg => "Adding the following volume to the test topology: " . Dumper \%copy_vol_hash);
            push (@orig_vols, $volume_hash);
            $cp_vol_cnt++;
        }

        # Call each of the specified filer cleanup prechecks
        foreach my $app (@{$options->{FILER_CLEANUP_PRECHECK}}) {

            my $testcase;
            my $script;
            ($script,$testcase) = split /\s+/,$app;

            if (!defined $testcase) {
                $testcase = "TESTCASE=testmanager_cleanup_check";
            }

            TestManager::_checkers(script=>$script, testcase=>"$testcase");
        }

        # Restore the test topology to its original list of volumes
        $Trace->debug(msg => "Putting the original list of volumes back " .
                             "in the test topology");
        delete $test_topo{cluster}->{volume};
        $test_topo{cluster}->{volume} = \@orig_vols;
    }

    foreach my $volume_hash (@volumes) {
        my $volume = $volume_hash->{volume}."_copy";
        $Trace->comment(msg=>"delete $volume");
        eval {
            $Server->volume_purge(vserver => $volume_hash->{vserver},
                                  volume => $volume);
        };
        push(@errors, $@) if ($@ && $@ !~ /does not exist/);
    }

    if (scalar @errors) {
        $Trace->fatal(msg=>"Failure to delete at least one of the volumes:".
                      join("\n", @errors));
    }

    $Trace->exit();
}

# This code takes and copies the specified volumes to new locations based upon
#  the type of volume, the settings for AFFINITY, AFFINITY_PREFERENCING, etc.
# If the jobs are supposed to occur in parallel, they will be passed as background
#  jobs and handled via the JobManager.
sub copy_volume {
    my $self = shift;
    my @volumes = @_;

    $Trace->entry();
    my $options = $self->get_hash();
    my $src_aggr; 
    my $src_node;
    my $src_aggrh;
    my $copy_cmd;
    my $output;
    my @copies;

    # force a refresh of the aggregates during each iteration
    $self->{aggregates} = $Server->storage_aggregate_show();

    foreach my $volume_hash (@volumes) {
        my $vserver = $volume_hash->{vserver};
        my $volume = $volume_hash->{volume};

        # examine the volume hash and determine the correct copy call as well
        #  as its source node and aggregate
        $copy_cmd = "volume_copy";
        if ($volume_hash->{"volume-style"} eq "striped") {
            ($src_node, $src_aggr) = $self->find_striped_source(vserver=>$vserver,
                                                                volume=>$volume,
                                                                aggregates=>$self->{aggregates},
                                                                );
        } elsif ($volume_hash->{"volume-style"} eq "flex"){

            ($src_node, $src_aggr) = $self->find_flex_source(vserver=>$vserver,
                                                             volume=>$volume,
                                                             aggregates=>$self->{aggregates},
                                                             );
        } else {
            $Trace->fatal(msg=>"Unknown volume type '".$volume_hash->{"volume-style"}."':".Dumper($volume_hash));
        }

        #############################
        #Pick an aggregate to copy to
        #############################

        my $dst_aggr = $self->pick_destination_aggregate(src_node=>$src_node,
                                                         volume=>$volume_hash,
                                                         aggregates=>$self->{aggregates});
        # An aggregate has been picked. Now deduct the size of this volume from it to minimize the risk
        # of not having enough space when multiple volumes are being copyd. (meaning, we shouldn't pick
        # it again for another volume if the resulting size will be too small)
        my ($dst_aggrh) = grep($_->{aggregate} eq $dst_aggr, @{$self->{aggregates}});
        $dst_aggrh->{availsize} -= $self->size_in_mb(size=>$volume_hash->{size});

        $self->{dst_volume} = $volume_hash->{volume}."_copy"; # FIX THIS!!!!!!!!!!!
        $self->{dst_vserver} = $vserver;
        
        $Trace->comment(msg=>"$copy_cmd $vserver $volume_hash->{volume} ($src_aggr) to $self->{dst_volume} ($dst_aggr)");
        my $fg = ($options->{PARALLEL}) ? "false" : "true";
        my $output;
        my $copy_hash = {
                         volume => $self->{dst_volume},
                         vserver => $vserver,
                         aggregate => $dst_aggr,
                         'volume-style' => $volume_hash->{'volume-style'},
                        };
        eval {
            $output = $Server->$copy_cmd(
                                         vserver => $vserver,
                                         volume => $volume,
                                         "destination-volume" => $self->{dst_volume},
                                         "destination-aggregate" => $dst_aggr,
                                         foreground => $fg,
                                        );
            # Create the structures needed for later verification. In the case of copy,
            #  the new volume points to the original source hash has the value. 
            $self->{sources}{$self->{dst_volume}} = $volume_hash;

            # set the new specific field to indicate that the volume should have the
            #  specific attributes
            $self->{verify}{$volume}{specific}{$self->{dst_volume}} = $copy_hash;
            push(@copies, $copy_hash);
        };
        if ($@) {
            if ($self->check_job(msg=>$@) == $self->RETRY()) {
                $Trace->comment(msg=>"Volume operation already running on $volume. Sleeping and retrying");
                sleep 5;
                redo;
            } else {
                $Trace->fatal(msg=>"Unexpected error: $@");
            }
        }
        if ($options->{PARALLEL}) {
            # the output will be the job number
            push(@{$Jobs{$output}}, $copy_hash);
	}
    }
    $Trace->exit(msg=>"return ".join(" ", @copies));
    return @copies;
}

# Take list of volumes that moved and verify that the volume attributes are correct.
sub verify_volumes {
    my $self = shift;
    my @volumes = @_;

    $Trace->entry();

    foreach my $volume_hash (@volumes) {
        my $volume = $volume_hash->{volume};
        # The volume copy will be type RW, so skip this field when copying DP volumes
        if ($volume_hash->{"type"} eq "DP") {
            push(@{$self->{skip_fields}}, 'type');
        }
        # specific_fields gets set during copy_volume.
        # targets get set by wait_on_jobs and is derived from the sources set in copy_volume
        $self->compare_volumes(source => $volume_hash,
                               targets => $self->{verify}->{$volume}->{targets},
                               specific_fields => $self->{verify}->{$volume}{specific},
                               skip_fields => $self->{skip_fields},
                              );
    }
    $Trace->exit();
}

sub mount_in_namespace {
    my ($self, %params) = @_;
    $Trace->entry();

    my $dst_volume = $self->{dst_volume};
    if (defined($params{destination}) && $params{destination} ne "") {
        $dst_volume = $params{destination};
    }
    $Server->volume_mount(vserver => $self->{dst_vserver},
                   volume => $dst_volume,
                   'junction-path' => "/$dst_volume");
    $Trace->exit(msg=>"return '/.admin/$dst_volume'");
    return "/.admin/$dst_volume";
}

=head1 PARAMS

=over

=back

=head1 SEE ALSO

L<VOPL>
L<Options>
L<TestManager>

=head1 BUGS



=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/VOPL/Copy.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
