## @summary     VOPL::Mirror class for the volume operation tests.
## @author      snider@netapp.com
## @pod here

=head1 NAME

VOPL::Mirror - mirror volume-operation-package class

=head1 SYNOPSIS

   package CreateMirror;
   use base qw(VOPL::Mirror);
   
   my $Volume = CreateMirror->new();
  
   my @volumes = $Volume->pick_volumes();

=head1 DESCRIPTION

This package is used as a class for mirror tests.
This package uses L<VOPL> as a base class.

=head1 METHODS

none

=cut

package VOPL::Mirror;
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);

use base qw(VOPL);

use Data::Dumper;
use Options;
use Server;
use TracePkg;
use constant MAX => 3;
my $Trace = new TracePkg;
my $Server = new Server();
my $Options = new Options;

my %Vol_Mirror_Hash;
my %Jobs;

$Options->parameters(definitions=>
                  [
                   {
                    name => "MIRROR_TYPE",
                    description => "type(s) of mirrors to create",
                    valid_values => [ qw(LS DR all) ],
                    default => "all",
                   },
                   {
                    name => "NUMBER_OF_MIRROR_AGGREGATES",
                    description => "How many aggregates should be used for mirrors. Defaults to all aggregates",
                   },
                   {
                    name => "RESTORER_STATS",
                    description => "turn on restore stats",
                    spec => "boolean",
                    default => "false",
                   },
                   {
                    name => "MIRROR_PROTOCOL",
                    description => "Protocol to be used for mirror operation",
                    valid_values => [ qw( VSM LSM ) ],
                    spec => "s",
                   },
                   ]);


sub process_params {
    my $self = shift;
    $Trace->entry();

    my $options = $self->get_hash();

    # determine which type(s) of mirrors to create
    if ( $options->{MIRROR_TYPE} =~ /ls/i ) {
        $self->{mirror_types} = [qw(LS)];
    } elsif ( $options->{MIRROR_TYPE} =~ /dr/i ) {
        $self->{mirror_types} = [qw(DR)];
    } else {
        $self->{mirror_types} = [qw(LS DR)];
    }
    
    if ( ($options->{MIRROR_PROTOCOL} eq 'VSM' || !defined $options->{MIRROR_PROTOCOL}) && ($options->{VOLUME_STYLE} eq 'striped' || $options->{DST_VOLUME_STYLE} eq 'striped')) {
        $Trace->fatal(msg=>"Mirroring striped volumes using VSM protocol is not supported", result => "CONFIG");
    }
        

    if ($options->{RESTORER_STATS}) {
        foreach my $server ($Server->servers()) {
            my $output;
            $Server->dblade(cmd=>"priv set diag; stats show -R incpr_global",
                            -Server=>$server,
                            output=>\$output);
            $Trace->trace(msg=>$output);
        }
    }
    $Trace->exit();
}

# this routine will create the mirrors for the provided volumes
sub test_setup {
    my $self = shift;
    $Trace->entry();

    my $options = $self->get_hash();
    my $src_aggr;
    my $src_node;

    $self->{volumes} = [$self->pick_volumes(get_striped_members => 0)];

    # force a refresh of the aggregates during each iteration
    $self->{aggregates} = $Server->storage_aggregate_show();

    $self->{volume_style} = $options->{VOLUME_STYLE} || "flex"; #
    if ( $options->{DST_VOLUME_STYLE} ne $options->{VOLUME_STYLE}) {
         $self->set(VOLUME_STYLE => undef, aggregate => undef); # Remove the params set from our parent
    }

    foreach my $volume_hash (@{$self->{volumes}}) {
        my $vserver = $volume_hash->{vserver};
        my $volume = $volume_hash->{volume};

        if ($volume_hash->{"volume-style"} eq "striped") {
            ($src_node, $src_aggr) = $self->find_striped_source(vserver=>$vserver,
								volume=>$volume,
								aggregates=>$self->{aggregates},
								);

        } elsif ($volume_hash->{"volume-style"} eq "flex"){
            ($src_node, $src_aggr) = $self->find_flex_source(vserver=>$vserver,
                                                             volume=>$volume,
                                                             aggregates=>$self->{aggregates},
                                                             );
        } else {
            $Trace->fatal(msg=>"Unknown volume type '".$volume_hash->{"volume-style"}."':".Dumper($volume_hash));
        }

        #############################
        # Find available aggregates
        #############################
        my @aggr_list = $self->pick_destination_aggregate(src_node=>$src_node,
                                                          volume=>$volume_hash,
                                                          aggr_per_node => 1,
                                                          aggregates=>$self->{aggregates});

        # Trim the number of aggregates down to match the number of mirrors we should have
        if (my $number = $options->{NUMBER_OF_MIRROR_AGGREGATES}) {
            my @reduced_aggr_list = ();
            while ($number--) {
                my $aggr = @aggr_list[int rand @aggr_list];
                @aggr_list = grep($_ ne $aggr, @aggr_list);
                push(@reduced_aggr_list, $aggr);
            }
            @aggr_list = @reduced_aggr_list;
        }

        $self->create_mirrors($volume_hash, \@aggr_list);
     }

    $Trace->exit();
}

# This is the main workhorse routine for this test. 
# The VOPL library forces filer_* and client_* to do nothing.
# Basically, each iteration will pick volumes to mirror, pass them to the replicate_mirrors code,
#  wait on the replicates to finish (retrying until the replicates eventually go through) and then
#  verifies that the volume attributes are what they are supposed to be.
sub test_main {
    my $self = shift;
    $Trace->entry();
    my $options = $self->get_hash();
    
    if (!$self->{volumes}) {
        # maybe test_setup was skipped
        @{$self->{volumes}} = $self->pick_volumes();
    }
   
    

    my $release = $Server->get_release(format=>"cooked");
    if ($release =~ /^tricky\.0$/i) {
        $self->{skip_fields} = [qw(dsid msid junction-path junction-path-source
                                   junction-active parent comment crtime space-guarantee-enabled)],
    } elsif ($release =~ /^tricky\.(1|2|3)$/i) {
        $self->{skip_fields} = [qw(dsid msid junction-path junction-path-source crtime description
                                   junction-active parent comment create-time space-guarantee-enabled)],
    } elsif ($release =~ /^tricky\.4$/i) {
        $self->{skip_fields} = [qw(dsid msid junction-path junction-path-source crtime description
                                   junction-active parent comment create-time space-guarantee-enabled
				   snapshot-space-used)],
    } else {
        $self->{skip_fields} = [qw(dsid msid junction-path junction-path-source crtime description
                                   junction-active parent comment create-time space-guarantee
                                   space-guarantee-enabled antivirus-on-access-policy node filer
			           uuid filesys-size-fixed snap-mirror-enabled load-sharing-source)],
    }    
    
    

    #Check if we are using striped volumes. If so we need to exclude
    # additional fields to prevent the volume compare from failing
    if  ( ($self->{volume_style} eq "striped" ) |  ($options->{DST_VOLUME_STYLE} eq "striped" )  ) {
        $Trace->trace(msg=>"Source or destination volume styles are striped. Skipping more volume fields ");
        my @tmpskip = qw (lookup-cache stripe-width volume-style max-autosize properly-striped 
                          autosize-increment concurrency striping-epoch stripe-count);
        push(@{$self->{skip_fields}}, @tmpskip);
                
    }
        
    
    
    
    my @volumes = @{$self->{volumes}};

    while (@volumes) {
        my @mirrors = $self->replicate_mirrors(@volumes);
        @volumes = $self->wait_on_jobs(volumes => \@mirrors, 
                                       job_to_volume_hash => \%Jobs, 
                                       sources => $self->{sources});
    }

    $self->verify_volumes(@{$self->{volumes}});
    $Trace->comment(msg=>"All volumes were successfully mirrored.");

    $Trace->exit();
}

# delete the mirrors that were created
sub test_cleanup {
    my $self = shift;
    $Trace->entry();

    # this code is not ready for primetime
    if (!$self->{volumes}) {
        # maybe test_setup was skipped
        $self->{volumes} = [$self->pick_volumes(get_striped_members => 0)];        
    }

    foreach my $volume_hash (@{$self->{volumes}}) {
        my $volume = $volume_hash->{volume};
        my $vserver = $volume_hash->{vserver};
        my $show_cmd;
        my $delete_cmd;

	$delete_cmd = "volume_mirror_delete";
	$show_cmd = "volume_mirror_show";
        my $output;

        $Trace->comment(msg=>"$delete_cmd $vserver $volume *");
        eval {
            $Server->$delete_cmd(
                                 vserver => $vserver,
                                 source => $volume,
                                 mirror => "*",
                                 output => \$output,
                                 -Timeout => 7200,
                                );
        };
        if ( $@ ) {
            $Trace->warn(msg=>"Unexpected error: $output");
        }
    }

    $Trace->exit();
}

# this code is used to actually create the initial mirrors
sub create_mirrors {
    my ($self, $volume_hash, $aggrs) = @_;
    my %mirror_cfg;

    $Trace->entry();
    my $options = $self->get_hash();
    my $volume = $volume_hash->{volume};
    my $vserver = $volume_hash->{vserver};
    my $cmd;

    $cmd = "volume_mirror_new";
    my $output;

    # this can get very ugly, but create each mirror type on each aggregate in the list
    foreach my $aggr (@$aggrs) {
        foreach my $type (@{$self->{mirror_types}}) {
            my $mirror = "${volume}_${type}_$aggr";
            
            if ($self->mirror_exists(volume=>$volume, type=>$type, aggr=>$aggr)) {
                $Trace->trace(msg=>"Already created mirror $mirror");
                next;
            }
            $Trace->comment(msg=>"$cmd $vserver $volume $mirror $aggr");
            %mirror_cfg = (
                              vserver => $vserver,
                              source => $volume,
                              mirror => $mirror,
                              aggregate => $aggr,
                              type => $type,                              
                              output => \$output,
                              -Timeout => 7200,
                             );
            #The comment field is broken when striped volume is the destination.
            # So We'll just skip it.
            $mirror_cfg{comment} = '"created by VolMirror"',if ($options->{DST_VOLUME_STYLE} ne 'striped');
            
            if ( $options->{MIRROR_PROTOCOL} ) {
                $Trace->trace(msg=>"Setting  Protocol to $options->{MIRROR_PROTOCOL}");
                #add the protocol flag into config hash 
                $mirror_cfg{protocol} = $options->{MIRROR_PROTOCOL};
            }
            eval {
                $Server->$cmd(%mirror_cfg
                             );
            };
            my $rc = $self->process_mirror_errors($@, $output);
            if ($rc == $self->RETRY()) {
                $Trace->comment(msg=>"Volume operation already running on $volume. Sleeping and retrying");
                sleep 5;
                redo;
            } elsif ($rc == $self->MAX()) {
                $Trace->warn(msg=>"Expected error: $output");
            }
            $self->mirror_hash_update(volume => $volume,
                                      type => $type,
                                      aggr => $aggr,
                                      mirror => $mirror,
                                      volume_hash => $volume_hash);
        }
    }
    $Trace->exit();
}

# This code takes and moves the specified volumes to new locations based upon
#  the type of volume, the settings for AFFINITY, AFFINITY_PREFERENCING, etc.
# If the jobs are supposed to occur in parallel, they will be passed as background
#  jobs and handled via the JobManager.
sub replicate_mirrors {
    my $self = shift;
    my @volumes = @_;
    my @mirrors;

    $Trace->entry();

    my $options = $self->get_hash();
    my $fg = ($options->{PARALLEL}) ? "false" : "true";

    if (0 == scalar @volumes) {
        $Trace->fatal(msg=>"Did not find any mirrors to replicate");
    }
    foreach my $volume_hash (@volumes) {
        my $vserver = $volume_hash->{vserver};
        my $volume = $volume_hash->{volume};

        # examine the volume hash and determine the correct replicate call as well
        #  as its source node and aggregate
        my $replicate_cmd;
        my $show_cmd;

	$replicate_cmd = "volume_mirror_replicate";
	$show_cmd = "volume_mirror_show";

        # verify/find the list of mirrors
        $Trace->trace(msg=>"$show_cmd $vserver $volume");
        my $mirrors = $Server->$show_cmd(
                                         vserver => $vserver,
                                         source => $volume
                                        );

	# Replicate all (LS and DR) mirrors for this volume
	$Trace->trace(msg=>"replicating $volume to all LS and DR mirrors");
        my $output;
	eval {
            $Trace->comment(msg=>"$replicate_cmd $vserver $volume *");
	    $output = $Server->$replicate_cmd(
                                              vserver => $vserver,
                                              source => $volume,
                                              mirror => "*",
                                              foreground => $fg,
                                              -Timeout => 7200,
                                             );
            foreach my $mirrorh (@$mirrors) {
                my $mirror_hash = {
                                   volume => $mirrorh->{mirror},
                                   aggregate => $mirrorh->{aggregate},
                                   type => $mirrorh->{type},
                                  };
                # these two are only used by the functional tests
                $self->{dst_volume} = $mirrorh->{mirror};
                $self->{dst_vserver} = $vserver;
            
                # Create the structures needed for later verification. In the case of mirror,
                #  the source hash has the mirror as the key and the source as the value.
                $self->{sources}{$mirrorh->{mirror}} = $volume_hash;

                # set the new specific field to indicate that the specific attributes that
                # this mirror should have
                $self->{verify}{$volume}{specific}{$mirrorh->{mirror}} = $mirror_hash;
                push(@mirrors, $mirrorh);
                if ($options->{PARALLEL}) {
                    # the output will be the job number
                    push(@{$Jobs{$output}}, $mirrorh);
                }
            }
        };
        if ($@) {
            if ($self->check_job(msg=>$@) == $self->RETRY()) {
                $Trace->comment(msg=>"Volume operation already running on $volume. Sleeping and retrying");
                sleep 5;
                redo;
            } else {
                $Trace->fatal(msg=>"Unexpected error: $@");
            }
        }
    }
    $Trace->exit(msg=>"return ".join(" ", map($_->{mirror}, @mirrors)));
    return @mirrors;
}

# Take list of volumes that moved and verify that the volume attributes are correct.
sub verify_volumes {
    my $self = shift;
    my @volumes = @_;

    $Trace->entry();
    foreach my $volume_hash (@volumes) {
        my $volume = $volume_hash->{volume};
        # specific_fields gets set during replicate_volumes.
        # targets get set by wait_on_jobs and is derived from the sources set in replicate_volumes
        $self->compare_volumes(source => $volume_hash,
                               targets => $self->{verify}->{$volume}->{targets},
                               specific_fields => $self->{verify}->{$volume}{specific},
                               skip_fields => $self->{skip_fields},
                              );
    }
    $Trace->exit();
}

sub process_mirror_errors {
    my ($self, $error, $output) = @_;
    
    $Trace->entry(msg=>"$error, $output");
    
    # Because the UI adds a bunch of preceeding spaces to subsequent lines of
    # error messages, for pretty alignment on the screen, we need to use \s+
    # when doing regex comparisons with output. When burt 192929 gets fixed,
    # using the \s+ method won't be needed any longer. And the s at the end
    # is needed to translate newlines into white space.
    if ( $error && $output =~ /load\s+sharing\s+mirror\s+exists\s+on\s+aggregate/s ||
         $error && $output =~ /mirror.+already exists/si ||
         $error && $output =~ /volume\s+record\s+not\s+created,\s+duplicate\s+key/s) {
        $Trace->trace(msg=>"Expected error: $output");
        # This failed because the mirror already exists, so we'll still add it
        # to our hash and use it during this test.
    } elsif ($error && $output =~ /Volume operation Running/i ||
            $output =~ /Volume busy/i) {
        $Trace->exit(msg=>"RETRY");
        return $self->RETRY();
    } elsif ( $error && $output =~ /maximum\s+allowable\s+number\s+of\s+mirrors/s ) {
        return $self->MAX();
    } elsif ( $error ) {
        $Trace->fatal(msg=>"Unexpected error: $output");
    }

    $Trace->exit(msg=>"OK");
    return $self->OK();
}

sub mirror_exists {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'VOPL' } ],
        {
         volume => {type => SCALAR},
         type => {type => SCALAR},
         aggr => {type => SCALAR},
        }
    );
    my $volume = $params{volume};
    my $type = $params{type};
    my $aggr = $params{aggr};

    $Trace->entry();
    my $mirror = $Vol_Mirror_Hash{$volume}{$type}{$aggr};
    $Trace->exit(msg=>"return $mirror");
    return $mirror;
}

sub mirror_hash_update {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'VOPL' } ],
        {
         volume => {type => SCALAR},
         type => {type => SCALAR},
         aggr => {type => SCALAR},
         mirror => {type => SCALAR},
         volume_hash => {type => HASHREF},
        }
    );
    my $volume = $params{volume};
    my $type = $params{type};
    my $aggr = $params{aggr};
    my $mirror = $params{mirror};
    my $volume_hash = $params{volume_hash};

    $Vol_Mirror_Hash{$volume}{$type}{$aggr} = $mirror;
    $Vol_Mirror_Hash{$volume}{hash} = $volume_hash;
}

# this is currently only used by the functional tests
sub mount_in_namespace {
    my $self = shift;
    $Trace->entry();

    $Server->volume_mount(vserver => $self->{dst_vserver},
                   volume => $self->{dst_volume},
                   'junction-path' => "/$self->{dst_volume}");
    $Trace->exit(msg=>"return '/.admin/$self->{dst_volume}'");
    return "/.admin/$self->{dst_volume}";
}

# Originally, this test replicated all LS mirrors of a volume with one command, 
# followed by replicating each DR mirror for this volume, one at a time. Narayana
# suggested a better test would be to replicate all the LS mirrors, do some file 
# system activity, allowing the DR mirrors to become out-of-date, then replicating 
# all (LS and DR) mirrors of a volume, making sure that all mirrors get moved to 
# the same data version even though they started at different data versions.

=head1 PARAMS

=over

=item MIRROR_TYPE

The type of mirror to create: LS, DR or all. Defaults to all.

=head1 SEE ALSO

L<VOPL>
L<Options>
L<TestManager>

=head1 BUGS



=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/VOPL/Mirror.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
