## @summary     VOPL::Move class for the volume operation tests.
## @author      snider@netapp.com
## @pod here

=head1 NAME

VOPL::Move - mirror volume-operation-package class

=head1 SYNOPSIS

   package CreateMove;
   use base qw(VOPL::Move);
   
   my $Volume = CreateMove->new();
  
   my @volumes = $Volume->pick_volumes();

=head1 DESCRIPTION

This package is used as a class for mirror tests.
This package uses L<VOPL> as a base class.

=head1 METHODS

none

=cut

package VOPL::Move;
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);

use base qw(VOPL);

use Data::Dumper;
use Options;
use Server;
use TracePkg;
use constant MAX => 3;
my $Trace = new TracePkg;
my $Server = new Server();
my $Options = new Options();
my %Jobs;

use Tharn;

# This is the main workhorse routine for this test. There is no test_setup or test_cleanup.
# The VOPL library also forces filer_* and client_* to do nothing.
# Basically, each iteration will pick volumes to move, pass them to the move_volume code,
#  wait on the moves to finish (retrying until the moves eventually go through) and then
#  verifies that the volume attributes are what they are supposed to be.
sub test_main {
    my $self = shift;
    $Trace->entry();

    my @servers = $Server->servers();
    my $server_name = (@servers)[int rand(@servers)];
    $Server->set_execution_server(-Server=>$server_name);
    $Trace->comment(msg=>"Execution server set to $server_name");

    my @volumes = $self->pick_volumes();
    my @original_volumes = @volumes;

    # Move each member volume serially, if they belong to the same volume
    if ($original_volumes[0]->{"volume-style"} eq "striped") {
        # Create hash key for each volume with an array element of the volume members
        my %vol_base_names = ();
        foreach my $tmp_vol (@original_volumes) {
            if ($tmp_vol->{volume} =~ /(.*)_[0-9]{3}$/) {
                if (! exists $vol_base_names{$1}) {
                    $vol_base_names{$1} = ();
                }
                push(@{$vol_base_names{$1}}, $tmp_vol);
            } else {
                if (! exists $vol_base_names{$tmp_vol->{volume}}) {
                    $vol_base_names{$tmp_vol->{volume}} = ();
                }
                push(@{$vol_base_names{$tmp_vol->{volume}}}, $tmp_vol);                
            }
        }

        my $idx = 0;
        while (1) {
            # Build an array of 1 member from each volume to move in parallel
            # We could have volumes with different member counts here, 
            # so try each volume before exiting
            @volumes = ();
            foreach my $key (keys %vol_base_names) {
                push(@volumes, $vol_base_names{$key}[$idx]) if defined $vol_base_names{$key}[$idx];
            }
            @original_volumes = @volumes;
            $idx++;

            # Exit if there are no more volume members to move
            if (! defined $volumes[0]) {
                last;
            }

            # Move the volume members
            while (@volumes) {
                $self->move_volume(@volumes);
                # $self->{sources} is created inside of move_volume()
                # %Jobs is created inside of move_volume()
                @volumes = $self->wait_on_jobs(volumes => \@volumes, 
                                               job_to_volume_hash => \%Jobs, 
                                               sources => $self->{sources});
            }
            $self->verify_volumes(@original_volumes);
        }
    } else {
        # Move all non-striped volumes in parallel
        while (@volumes) {
            $self->move_volume(@volumes);
            # $self->{sources} is created inside of move_volume()
            # %Jobs is created inside of move_volume()
            @volumes = $self->wait_on_jobs(volumes => \@volumes, 
                                           job_to_volume_hash => \%Jobs, 
                                           sources => $self->{sources});
        }
        $self->verify_volumes(@original_volumes);
    }

    # we should have already died on a failure...
    $Trace->comment(msg=>"All volumes were successfully moved.");

    $Trace->exit();
}

# This code takes and moves the specified volumes to new locations based upon
#  the type of volume, the settings for AFFINITY, AFFINITY_PREFERENCING, etc.
# If the jobs are supposed to occur in parallel, they will be passed as background
#  jobs and handled via the JobManager.
sub move_volume {
    my $self = shift;
    my @volumes = @_;

    $Trace->entry();
    my $options = $self->get_hash();
    my $src_aggr; 
    my $src_node;
    my $src_aggrh;
    my $move_cmd;
    my $output;

    # force a refresh of the aggregates hash during each iteration
    $self->{aggregates} = $self->find_available_aggregates();

    foreach my $volume_hash (@volumes) {
        my $vserver = $volume_hash->{vserver};
        my $volume = $volume_hash->{volume};
        my @filtered_aggr_list = {};

        # examine the volume hash and determine the correct move call as well
        #  as its source node and aggregate
        if ($volume_hash->{"volume-style"} eq "striped") {
            $move_cmd = "volume_member_move";

            ($src_node, $src_aggr) = $self->find_striped_source(vserver=>$vserver,
                                                                volume=>$volume,
                                                                aggregates=>$self->{aggregates},
                                                               );

            # If we are moving coral volumes around, we want to move them from one aggregate member
            # to another, not between aggregates. So we will filter out all aggregate members that
            # are not part of the source aggregate.

            my $striped_aggr = "";
            if ($volume_hash->{aggregate} =~ /(.*)_[0-9]{3}$/) {
                $striped_aggr = $1;
            } else {
                $Trace->fatal(msg=>"Unable to find aggregate name for striped volume $volume");
            }
            @filtered_aggr_list = grep($_->{aggregate} =~ /^${striped_aggr}_[0-9]{3}$/, 
                                       @{$self->{aggregates}});

        } elsif ($volume_hash->{"volume-style"} eq "flex"){
            $move_cmd = "volume_move";

            ($src_node, $src_aggr) = $self->find_flex_source(vserver=>$vserver,
                                                             volume=>$volume,
                                                             aggregates=>$self->{aggregates},
                                                             );

            # In the case of flex vols, we can move to whatever aggregate the automation
            # chooses, so we will simply copy the list
            
            @filtered_aggr_list = @{$self->{aggregates}};
            
        } else {
            $Trace->fatal(msg=>"Unknown volume type '".$volume_hash->{"volume-style"}."':".Dumper($volume_hash));
        }

        #############################
        #Pick an aggregate to move to
        #############################

        my $dst_aggr = $self->pick_destination_aggregate(src_aggr=>$src_aggr,
                                                          src_node=>$src_node,
                                                          volume=>$volume_hash,
                                                          aggregates=>\@filtered_aggr_list);

        # An aggregate has been picked. Now deduct the size of this volume from it to minimize the risk
        # of not having enough space when multiple volumes are being moved. (meaning, we shouldn't pick
        # it again for another volume if the resulting size will be too small)
        my ($dst_aggrh) = grep($_->{aggregate} eq $dst_aggr, @{$self->{aggregates}});
        $dst_aggrh->{availsize} -= $self->size_in_mb(size=>$volume_hash->{size});

        # now issue the command, either in the foreground or background
        $Trace->comment(msg=>"$move_cmd $vserver $volume_hash->{volume} to $dst_aggr (from $src_aggr)");
        my $fg = ($options->{PARALLEL}) ? "false" : "true";
        my $output;

        my $has_premove_checks;

        my $release = $Server->get_release(cooked=>1);
        if ( ($release =~ /rollingrock/i) && ($Server->has_change(change => "volops-directory")) ) {
            $has_premove_checks = 1;
        }

        my $move_precheck;
        my $target_aggr_check;

        my $vol = $volume_hash->{volume};

        if ($has_premove_checks && ($volume_hash->{"volume-style"} eq "flex") ){
            if ( $options->{VOL_PREMOVE_TARGET_AGGR_CHECK} ) {
                eval {
                    $output = $Server->volume_move_target_aggr_show(
                                                 vserver => $vserver,
                                                 volume => $volume_hash->{volume},
                                                );
                };

                if ( $@ ) {
                    my $msg = "VOL_SIMPLE: Volume move target-aggr show for vserver \"$vserver\", ".
                              "volume \"$volume_hash->{volume}\" failed with $@";
                    if ( $options->{VOL_PREMOVE_CHECK_CONTINUE_ON_FAILURE} ) {
                        $Trace->comment(msg=>$msg, result=>"FAIL" );
                    } else {
                        $Trace->fatal(msg=>$msg );
                    }
                } else {
                    foreach my $item (@$output) {
                        if ( $item->{'aggregate'} =~ /^${dst_aggr}$/) {
                            $target_aggr_check = 1;
                            last;
                        }
                    }
       
                    if ( $target_aggr_check ) {

                        my $msg = "******************************\n".
                                  "VOL_SIMPLE Check Passed - ".
                                  "Volume move target-aggr show for vserver \"$vserver\", ".
                                  "volume \"$volume_hash->{volume}\" agrees that \"$dst_aggr\" is a valid ".
                                  "destination aggregate".
                                  "\n******************************";

                        $Trace->comment( msg=>$msg );
                    } else {
                        my $msg = "VOL_SIMPLE: Volume move target-aggr show for vserver \"$vserver\", ".
                                  "volume \"$volume_hash->{volume}\" does not think that \"$dst_aggr\" is a valid ".
                                  "destination aggregate";

                        if ( $options->{VOL_PREMOVE_CHECKS_CONTINUE_ON_FAILURE} ) {
                            $Trace->comment( msg=>$msg, result=>"FAIL" );
                        } else {
                            $Trace->fatal( msg=>$msg );
                        }
                    }
                }

                # Store results in object
                $self->{verify}{$vol}{target_aggr_check}   = $target_aggr_check;
            }

            if ( $options->{VOL_PREMOVE_CHECK} ) {
                eval {
                    $output = $Server->volume_move_start(
                                                 vserver => $vserver,
                                                 volume => $volume_hash->{volume},
                                                 "destination-aggregate" => $dst_aggr,
                                                 'perform-validation-only' => 'true',
                                                 foreground => 'true',
                                                );
                };
                if ( $@ ) {
                    my $msg = "VOL_SIMPLE: Volume move precheck for vserver \"$vserver\" volume \"$volume_hash->{volume}\" ".
                              "dest_aggr \"$dst_aggr\" failed with $@";
                    if ( $options->{VOL_PREMOVE_CHECKS_CONTINUE_ON_FAILURE} ) {
                        $Trace->comment( msg=>$msg, result =>"FAIL" );
                    } else {
                        $Trace->fatal( msg=>$msg );
                    }
                } else {
                    my $msg = "******************************\n".
                              "VOL_SIMPLE Check Passed - ".
                              "Volume move precheck for vserver \"$vserver\", ".
                              "volume \"$volume_hash->{volume}\" \"$dst_aggr\" succeeded".
                              "\n******************************";
                    $Trace->comment( msg=>$msg );

                    # If command did not throw an error, it succeeded
                    $move_precheck = 1;
                }

                # Store results in object
                $self->{verify}{$vol}{move_precheck} = $move_precheck;
            }
        }

        $output = undef;
        eval {
            $output = $Server->$move_cmd(
                                         vserver => $vserver,
                                         volume => $volume_hash->{volume},
                                         "destination-aggregate" => $dst_aggr,
                                         foreground => $fg,
                                        );

            # Create the structures needed for later verification. In the case of move,
            #  the source hash has the volume as the key and the value since we will just
            #  compare the original hash to the new hash after the move.
            $self->{sources}{$vol} = $volume_hash;

            # set the new specific field to indicate that the volume should be on the new aggregate
            $self->{verify}{$vol}{specific}{$vol} = { aggregate => $dst_aggr };
        };
        if ($@) {
            if ($self->check_job(msg=>$@) == $self->RETRY()) {
                $Trace->comment(msg=>"Volume operation already running on $volume. Sleeping and retrying");
                sleep 5;
                redo;
            } else {
                $Trace->fatal(msg=>"Unexpected error: $@");
            }
        }
        if ($options->{PARALLEL}) {
            # the output will be the job number
            push(@{$Jobs{$output}}, $volume_hash);
	    }
    }
    $Trace->exit();
}

# Take list of volumes that moved and verify that the volume attributes are correct.
sub verify_volumes {
    my $self = shift;
    my @volumes = @_;
    my $premove_errors;

    $Trace->entry(msg=>"Number of volumes to verify: ".int(@volumes));

    my $options = $Options->get_hash();
    my $move_member = $options->{MOVE_MEMBER} || 0;
    foreach my $volume_hash (@volumes) {
        my $volume = $volume_hash->{volume};
        $Trace->trace(msg=>"Verifying volume: $volume");
	    # specific_fields gets set during move_volume.
        # targets get set by wait_on_jobs and is derived from the sources set in move_volume
        # Remove 'snapshot-space-used' from skip after burt325395 is fixed
        if ($move_member) {
            $self->compare_volumes(
                source          => $volume_hash,
                targets         => $self->{verify}->{$volume}->{targets},
                specific_fields => $self->{verify}->{$volume}{specific},
                skip_fields     => [
                    qw(dsid space-guarantee-enabled uuid node filer name-ordinal
                    fsid snapdir-access maxdir-size snapshot-space-used
                    junction-path-source flexcache-origin-volume node filer 
                    snapshot-count)
                ],
            );
        } else {
            $self->compare_volumes(
                source          => $volume_hash,
                targets         => $self->{verify}->{$volume}->{targets},
                specific_fields => $self->{verify}->{$volume}{specific},
                skip_fields     => [
                    qw(dsid space-guarantee-enabled uuid node filer name-ordinal
                    fsid snapshot-space-used junction-path-source
                    sis-space-saved-percent dedupe-space-saved-percent
                    sis-space-saved dedupe-space-shared dedupe-space-saved
                    compression-space-saved compression-space-saved-percent
                    node filer snapshot-count)
                ],
            );
        }

        if ( exists $self->{verify}->{$volume}->{target_aggr_check} ) {
            $Trace->debug(msg=>"VOL_SIMPLE target_aggr_check for volume $volume = |$self->{verify}->{$volume}->{target_aggr_check}|");
            if ( !$self->{verify}->{$volume}->{target_aggr_check} ) {
                $premove_errors++;
            }
        }
        if ( exists $self->{verify}->{$volume}->{move_precheck} ) {
            $Trace->debug(msg=>"VOL_SIMPLE move_precheck for volume $volume = |$self->{verify}->{$volume}->{move_precheck}|");
            if ( !$self->{verify}->{$volume}->{move_precheck} ) {
                $premove_errors++;
            }
        }
    }

    # reset the verification structures
    $self->{verify} = undef;
    $self->{sources} = undef;

    if ( $premove_errors ) {
        # The specific errors have already been reported in move_volume()
        $Trace->fatal(msg=>"Volume pre-move errors reported");
    }

    $Trace->exit();
}

=head1 PARAMS

=over

=back

=head1 SEE ALSO

L<VOPL>
L<Options>
L<TestManager>

=head1 BUGS



=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/VOPL/Move.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
