## @summary     VOPL::Snapshot class for the volume operation tests.
## @author      snider@netapp.com
## @pod here

=head1 NAME

VOPL::Snapshot - snapshot volume-operation-package class

=head1 SYNOPSIS

   package Snapshot;
   use base qw(VOPL::Snapshot);
   
   my $snap = Snapshot->new();
  
   my @volumes = $snap->pick_volumes();

=head1 DESCRIPTION

This package is used as a class for snapshot tests.
This package uses L<VOPL> as a base class.

=head1 METHODS

none

=cut

package VOPL::Snapshot;
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);

use base qw(VOPL);

use Data::Dumper;
use Options;
use Server;
use TracePkg;
use constant MAX => 3;
my $Trace = new TracePkg;
my $Server = new Server();
my %Jobs;

sub process_params {
    my $self = shift;

    $Trace->entry();

    my $options = $self->get_hash();
    if ($options->{PARALLEL}) {
        $Trace->comment(msg=>"PARALLEL=1 is not supported by this test. Unsetting it.");
        $self->set(PARALLEL=>0);
    }

    $Trace->exit(msg=>"");
}


# This is the main workhorse routine for this test. There is no test_setup.
# The VOPL library also forces filer_* and client_* to do nothing.
# Basically, each iteration will pick volumes to snapshot, pass them to the snapshot_volume code,
#  wait on the ops to finish (retrying until the copies eventually go through) and then
#  verifies that the resulting volume attributes are what they are supposed to be.

sub test_main {
    my $self = shift;
    $Trace->entry();

    my $release = $Server->get_release(format=>"cooked");
    if ($release =~ /tricky.0/i) {
        $self->{skip_fields} = [qw(dsid msid ruleset junction-path junction-path-source junction-active parent crtime)];
    } else {
        $self->{skip_fields} = [qw(dsid msid junction-path junction-path-source junction-active parent create-time)];
    }
    my @volumes = $self->pick_volumes();
    my @original_volumes = @volumes;
    while (@volumes) {
        my @snapshots = $self->snapshot_volume(@volumes);
        # $self->{sources} is created inside of snapshot_volume()
        @volumes = $self->wait_on_jobs(volumes => \@snapshots, 
                                       sources => $self->{sources});
    }
    # no real volume verification to be done

    # we should have already died on a failure...
    $Trace->comment(msg=>"All snapshots were successfull.");

    $Trace->exit();
}

# this code will remove the volume that was created during the test_main
sub test_cleanup {
    my $self = shift;
    $Trace->entry();

    my @errors;
    my @volumes = $self->pick_volumes();
    foreach my $volume_hash (@volumes) {
        my $volume = $volume_hash->{volume}."_snapshot";
        $Trace->comment(msg=>"delete $volume");
        my $retry=0;
        
      DELETE: {
            eval {
                $Server->volume_snapshot_delete(vserver => $volume_hash->{vserver},
                                                volume => $volume_hash->{volume},
                                                snapshot => $volume);
            };
            if ($@) {
                if ( ($@ =~ /Snapshot is in use/ ||
                      $@ =~ /The volume is quiesced/ ) &&
                     $retry++ < 5) {
                    $Trace->warn(msg=>"retry delete $volume due to error: $@");
                    redo DELETE;
                } elsif ($@ !~ /doesn't exist/) {
                    push(@errors, $@);
                }
            }
        }
    }

    if (scalar @errors) {
        $Trace->fatal(msg=>"Failure to delete at least one of the volumes:".
                      join("\n", @errors));
    }

    $Trace->exit();
}

# This code takes and creates a snapshot of the specified volumes.
sub snapshot_volume {
    my $self = shift;
    my @volumes = @_;

    $Trace->entry();
    my $options = $self->get_hash();
    my $snapshot_cmd;
    my $output;
    my @snapshots;

    foreach my $volume_hash (@volumes) {
        my $vserver = $volume_hash->{vserver};
        my $volume = $volume_hash->{volume};

        $snapshot_cmd = "volume_snapshot_new";
        
        $self->{snapshot} = $volume_hash->{volume}."_snapshot"; # FIX THIS!!!!!!!!!!!
        
        $Trace->comment(msg=>"$snapshot_cmd $vserver $volume_hash->{volume} $self->{snapshot}");
        my $output;
        my $snapshot_hash = {
                             volume => $volume,
                             vserver => $vserver,
                             snapshot => $self->{snapshot},
                            };
        eval {
            $output = $Server->$snapshot_cmd(
                                             vserver => $vserver,
                                             volume => $volume,
                                             snapshot => $self->{snapshot},
                                        );
            # Create the structures needed for later verification. In the case of snapshot,
            #  the new volume points to the original source hash has the value. 
            $self->{sources}{$self->{snapshot}} = $volume_hash;

            # set the new specific field to indicate that the volume should have the
            #  specific attributes
            $self->{verify}{$volume}{specific}{$self->{snapshot}} = $snapshot_hash;
            push(@snapshots, $snapshot_hash);
        };
        if ($@) {
            if ($self->check_job(msg=>$@) == $self->RETRY()) {
                $Trace->comment(msg=>"Volume operation already running on $volume. Sleeping and retrying");
                sleep 5;
                redo;
            } else {
                $Trace->fatal(msg=>"Unexpected error: $@");
            }
        }
    }
    $Trace->exit(msg=>"return ".join(" ", @snapshots));
    return @snapshots;
}

=head1 PARAMS

=over

=back

=head1 SEE ALSO

L<VOPL>
L<Options>
L<TestManager>

=head1 BUGS



=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/VOPL/Snapshot.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2007, Network Appliance., All Rights Reserved

=cut

1;
