#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/VSA/Storage/Filer.pm#1 $
## @summary Filer interface 
## @author dl-vsa-qa
## @status private

package VSA::Storage::Filer;

use strict;

use Data::Dumper qw(Dumper);
use Error qw(_clear_err _set_err);
use IO::File;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR SCALARREF BOOLEAN OBJECT ARRAYREF UNDEF);
use Time::HiRes qw(gettimeofday);
use Tharn qw($Log);
use Utils qw(make_host);
use VSA::UICommon qw(_caller _reformat_err _log _refresh_cache
                     _size_to_mbytes _trim);
use VSA::Storage;
use VSA::Storage::Filer::Aggregate;
use VSA::Storage::Filer::Igroup;

use base qw(VSA::Storage VSA::UICommon Error);

###############################################################################
# Method:    new()
# Objective: Create new instance of Filer
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($class,%args) = @_;
    my (
      $cmd,
      $connObj,
      $errmsg,
      $self,
      $tharnhost,  
    );
    
    ### bless instance into class
    $self= bless{}, $class;
    
    ### Clear error buffer
    $self->_clear_err();
    #$self->{call_stack_on_error}=1;
    
    ### Determine err handle upfront in case validation method
    ### fails before error handler is selected
    $self->{on_err}=(defined $args{on_err} and $args{on_err}=~/die|rtn/)
        ? $args{on_err}
        : "die";
    
    eval {
        
        ### Validate input arguments
        ($class,%args) = validate_mixed (
            @_,
            [ { type => SCALAR} ],
            {
                hostname  => {type => SCALAR},
                conn_type => {type => SCALAR, optional=>1,
                              default=>"rsh-socket",
                              callbacks=>{
                                'legal values'=>
                                    sub{shift()=~/rsh-socket|rsh-persay|ssh|telnet/i
                             }}},
                conn_name => {type => SCALAR, optional=>1},
                loghdl    => {type => SCALARREF, optional=>1 },
                on_err    => {type => SCALAR, optional=>1, default=>"die",
                              callbacks=>{
                                'legal values'=>sub{shift()=~/die|rtn/i
                             }}},
                password  => {type => SCALAR, optional=>1,},
                show_cmd  => {type => SCALAR, optional=>1, default=>0},
                timeout   => {type => SCALAR, optional=>1, default=>10,
                              regex => qr/^\d+$/ },
                username  => {type => SCALAR, optional=>1,},
               
            }
        );
        
        ###  Lets specify what handler to use
        $self->{name}     = $args{hostname};
        $self->{on_err}   = $args{on_err};
        $self->{show_cmd} = $args{show_cmd};
        $self->{timeout}  = $args{timeout};
        $self->{loghdl}   = $args{loghdl};
        $self->{aggrObjs}  = [];
        $self->{igrpObjs}  = [];
        
        $self->on_err($args{on_err});
        
        $self->_log( 
            msg=>"creating ssh connection to ".$args{hostname},
            indent=>3);
        
        $self->_connect(
            name=>(defined $args{conn_name})
                ? $args{conn_name}
                : $args{conn_type},
            type=>$args{conn_type},
            hostname=>$self->get_name(),
            username=>$args{username},
            password=>$args{password},
            timeout=>$args{timeout}
        );

    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }

    return $self;
}

###############################################################################
# Method:    create_igrp()
# Objective: Create igrp
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_igrp {
    my ($self,%args)=@_;
    my (
        $errmsg,
        $igrpObj,
    );
    
    ### Clear error buffer
    $self->_clear_err();
    
    eval {
        
        ### Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                name      => {type => SCALAR},
                type      => {type => SCALAR},
                protocol  => {type => SCALAR},
                iqns      => {type => SCALAR},
            }
        );
        
        $igrpObj=VSA::Storage::Filer::Igroup->_new(
            name=>$args{name},
            type=>$args{type},
            protocol=>$args{protocol},
            iqns=>$args{iqns},
            _create=>1,
            _filer=>$self,
            _parent=>$self,
            _loghdl=>$self->_get_loghdl()
        );
            
        push @{$self->{igrpObjs}}, $igrpObj;  
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    return $igrpObj;
}

###############################################################################
# Method:    discover_aggrs()
# Objective: Discover aggregates on filer
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover_aggrs{
    my ($self,%args) = @_;
    my (
        @outputs,
        $aggr_name,
        $cmd,
        $errmsg,
        @aggrObjs,
        );  
    
    ### Clear error buffer
    $self->_clear_err();

    ### Initialize variables
    $cmd="aggr status -c";
     
    ### Execute command
    eval {    
        @outputs=$self->_execute(cmd=>$cmd,hide=>1);      
        foreach my $line (@outputs[1..scalar(@outputs)-1]) {
           
            ($aggr_name) = $line =~ /\s{0,}(\S+)/;
            
            $self->_log(
                msg=>"creating aggr object '".$aggr_name."'",
                indent=>3
            );
             
            my $aggrObj=VSA::Storage::Filer::Aggregate->_new(
                name=>$aggr_name,
                _filer=>$self,
                _parent=>$self,
                _loghdl=>$self->_get_loghdl()
            );
            
            push  @{$self->{aggrObjs}}, $aggrObj; 
            
        }
        
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    ### Return result
    return (wantarray()) ? @{$self->{aggrObjs}} : $self->{aggrObjs}; 
    
}

###############################################################################
# Method:    discover_igrps()
# Objective: Discover igroups on filer
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover_igrps{
    my ($self,%args) = @_;
    my (
        @outputs,
        $aggr_name,
        $cmd,
        $errmsg,
        @igroup_names,
        @igrpObjs,
        );  
    
    ### Clear error buffer
    $self->_clear_err();

    ### Initialize variables
    $cmd="igroup show";
     

    ### Execute command
    eval {    
        @outputs=$self->_execute(cmd=>$cmd,hide=>1);      
        my @lines=grep(/\(iSCSI|FCP\)/,@outputs);
        foreach my $line (@lines) {
            my ($igroup_name)=$line=~/(\S+)/i;
            push @igroup_names,$igroup_name;
        }
        
        foreach my $igroup_name (@igroup_names) {
           
           $self->_log(
                msg=>"creating igroup object '".$igroup_name."'",
                indent=>3
           );
           
            my $igrpObj=VSA::Storage::Filer::Igroup->_new(
                name=>$igroup_name,
                _filer=>$self,
                _parent=>$self,
                _loghdl=>$self->_get_loghdl()
            );
        
            push @{$self->{igrpObjs}}, $igrpObj; 
           
        }      
        
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    ### Return result
    return (wantarray()) ? @{$self->{igrpObjs}} : $self->{igrpObjs}; 
    
}

###############################################################################
# Method:    get_aggr_cnt()
# Objective: Get aggregate cache count
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_aggr_cnt {
    my ($self) = @_;
    
    ### Clear error buffer 
    $self->_clear_err();

    return scalar (@{$self->{aggrObjs}}); 
}


################################################################################
# Subroutine Name: get_aggr_objs
# Objective:       return array of aggregate objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_aggr_objs {
    my ($self,%args) = @_;
    my (
        $cnt,
        $errmsg,
        $match,
        $value,
        @aggrObjs,
    );
    
    ### Clear error buffer
    $self->_clear_err();   
    
    eval {
        ### Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                cnt       => {type => SCALAR, optional=>1, default=>9999},
                name      => {type => SCALAR, optional=>1},
            }
        );
        
        $cnt=$args{cnt};
    
        ### Remove object from cache that are marked destroyed
        $self->_refresh_cache(name=>"aggrObjs");
        
        ### Apply filter if provided                 
        if (keys %args) {
            foreach my $aggrObj (@{$self->{aggrObjs}}){
                
                $match=0;
                
                foreach my $key (keys %args){
                    ### Assume exact parameter match if
                    ### no regex chars are included
                    $value=($args{$key} !~/\^|\$|\.|\*/)
                        ? '^'.$args{$key}.'$' 
                        : $args{$key};
                        
                    if ($key=~/name/){
                        $match=1 if ($aggrObj->get_name()=~/$value/); 
                    }
                
                    last if (not $match);
                }
                
                push @aggrObjs,$aggrObj if ($match);
            }
        }
        else {
            push @aggrObjs,@{$self->{aggrObjs}}
        }
        
        unless (@aggrObjs) {
            die "No Aggregates Found";
        }
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    return $aggrObjs[0] if ($cnt==1);
    return (wantarray()) ? @aggrObjs[0..$cnt-1] : \@aggrObjs[0..$cnt-1];
    
} 
###############################################################################
# Method:    get_*()
# Objective: Get filer name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_connection { my ($self) = @_; $self->{connection};    }
sub get_ip         { my ($self) = @_; $self->{default_ip};     }
sub get_hostname   { my ($self) = @_; $self->{name};    }
sub get_igrp_cnt   { my ($self) = @_; scalar (@{$self->{igrpObjs}}); }

################################################################################
# Subroutine Name: get_igrp_objs
# Objective:       return array of igrp objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_igrp_objs {
    my ($self,%args) = @_;
    my (
        $cnt,
        $errmsg,
        $match,
        $value,
        @igrpObjs,
    );
    
    ### Clear error buffer
    $self->_clear_err();
       
    ### Validate input arguments
    eval {
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                cnt      => {type => SCALAR, optional=>1, default=>9999},
                name      => {type => SCALAR, optional=>1},
            }
        );
        
        $cnt=$args{cnt};
        
        ### Remove object from cache that are marked destroyed
        $self->_refresh_cache(name=>"igrpObjs");
        
        ### Apply filter if provided                 
        if (keys %args) {
            foreach my $igrpObj (@{$self->{igrpObjs}}){
                
                $match=0;
                
                foreach my $key (keys %args){
                    ### Assume exact parameter match if
                    ### no regex chars are included
                    $value=($args{$key} !~/\^|\$|\.|\*/)
                        ? '^'.$args{$key}.'$' 
                        : $args{$key};
                        
                    if ($key=~/name/){
                        $match=1 if ($igrpObj->get_name()=~/$value/); 
                    }
                
                    last if (not $match);
                }
                
                push @igrpObjs,$igrpObj if ($match);
            }
            
            unless (@igrpObjs) {
                die "No Igroups Found";
            }
        }
        else {
            push @igrpObjs,@{$self->{igrpObjs}}
        }
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    return $igrpObjs[0] if ($cnt==1);
    return (wantarray()) ? @igrpObjs[0..$cnt-1] : \@igrpObjs[0..$cnt-1];
    
} 

###############################################################################
# Method:    get_iqn()
# Objective: Get iscsi qualified name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_iqn {
    my ($self) = @_;
    my (
        @outputs,
        $cmd,
        $errmsg,
        $iscsi_name,
        );  
    
    ### Clear error buffer
    $self->_clear_err();

    ### Initialize variables
    $cmd="iscsi nodename";

    ### Execute command
    eval {    
        @outputs=$self->_execute(cmd=>$cmd,hide=>1);      
        ($iscsi_name)=$outputs[0]=~ /nodename:\s(\S+)/i;
        unless (defined $iscsi_name){
            die "Unable to retrieve iscsi nodename";
        }
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    ### Return result
    return $iscsi_name;   
}

###############################################################################
# Method:    get_name()
# Objective: Get filer name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_name { my ($self) = @_; return $self->{name};    }

###############################################################################
# Method:    get_timeout()
# Objective: Get timeout
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_timeout { my ($self) = @_; return $self->{timeout};}

###############################################################################
# Method:    get_version()
# Objective: Get filer version
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_version
{
    my ($self) = @_;
    my (
        @outputs,
        $cmd,
        $errmsg,
        $version,
        );  
    
    ### Clear error buffer
    $self->_clear_err();

    ### Initialize variables
    $cmd="version";

    ### Execute command
    eval {    
        @outputs=$self->_execute(cmd=>$cmd,hide=>1);      
        ($version)=$outputs[0]=~ /Release\s(.*)\:\s+/i;
        unless (defined $version){
            die "Unable to retrieve version";
        }
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    ### Return result
    return $version;   
}
 
###############################################################################
# Method:    on_err
# Objective: Return errhdlr_type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub on_err {
    my ($self,$value)=@_; 
    my ($errmsg);
    
    if (defined $value) {
        
        unless ($value=~/die|rtn/){
            $errmsg="Invalid value '".$value."'";
            die _caller($errmsg);
        }
        
        $self->{on_err}=$value;
    }
    
    return $self->{on_err}; 
}

###############################################################################
# Method:    show_cmd
# Objective: display_cmd
# Details:   See POD documentation at the end of this file.
###############################################################################
sub show_cmd { my ($self,$value)=@_; return $self->{show_cmd};}


###############################################################################
#  P  R  I  V  A  T  E
###############################################################################
# Method:    _connect
# Objective: Create ssh connection to host machine
###############################################################################
sub _connect {
    my ($self,%args)=@_;
    my (
        $connObj,
        $conn_str,
        $errmsg,
        $host_of,
        $username,
        $password,
        $timeout,
        $type,
       );
    
    ### Validate input
    ($self, %args) = validate_mixed( @_,
        [ { type => OBJECT} ],
        {
            name      =>  { type => SCALAR, }, 
            hostname  =>  { type => SCALAR, }, 
            type      =>  { type => SCALAR, },
            password  =>  { type => SCALAR|UNDEF, optional => 1},
            username  =>  { type => SCALAR|UNDEF, optional => 1},
            timeout   =>  { type => SCALAR, optional=>1, default=>10,
                            regex => qr/^\d+$/ },
        }                
    );
    
    ### Create a host object
    $host_of=Utils::make_host($args{hostname});
    
    $username=(defined $args{username})
        ? $args{username}
        : $host_of->{connection_fields}{rsh}{username};
        
    $password=(defined $args{password})
        ? $args{password}
        : $host_of->{connection_fields}{rsh}{password};
        
    $timeout=(defined $args{timeout})
        ? $args{timeout}
        : $host_of->{connection_fields}{rsh}{timeout};
        
    $type=(defined $args{type})
        ? $args{type}
        : $host_of->{connection_fields}{rsh}{type};
        
    $self->{default_ip}=$host_of->{default_ip};

    ### Construct the connect string
    $conn_str=$args{hostname}."#".$type.";inherit=0;".
        "conntype=".$type.";".
        "username=".$username.";".
        "password=".$password.";".
        "timeout=".$timeout.";";
        
    $self->_log(msg=>"conn_str: ".$conn_str,indent=>3);
     
    ### Create connection object
    $connObj = Tharn::connect(
        $conn_str,    
        -eval=>sub{
            if (@_) {
                $errmsg = (@_)[0]->{messages};
                die @$errmsg;
                return $self;   
            }
        }
    );

    if (not defined $connObj){
        die $self->_caller(
            "Unable to connect to host ".
            $args{hostname}.
            "(".$args{username}."/".$args{password}.")"
            );
    }
    

    
    $self->{connection}=$connObj;
    
    return $connObj;

} # end sub connect;


##############################################################################
# Method:    _execute
# Objective: Issue arbitrary command to host and get results..
###############################################################################
sub _execute
{
    my($self, %args) = @_;
    my(
       $buffer,
       $connection,
       $cmd_fmt,
       $errmsg,
       $t0,
       $t1,
       @arr,
       @rows,
       );

    ### Clear errors.
    $self->_clear_err;

    ### Validate input arguments
    ($self,%args) = validate_mixed (
        @_,
        [ { type => OBJECT} ],
        {   cmd    => {type => SCALAR}, 
            hide    => {type => SCALAR, optional=>1, default=>0},  
            timeout => {type => SCALAR, optional=>1,
                        default=>$self->get_timeout(),
                        regex => qr/^\d+$/},
        }
    );
   
     
    # Print the command unless the user has specified otherwise.  
    if (!$args{hide} or $self->show_cmd()){
       $cmd_fmt=$self->get_name()."(FILER)> ".$args{cmd}; 
       $self->_log(msg=>$cmd_fmt, indent=>5);
    }
    
    # Remove the carriage returns from the cli command
    $args{cmd} =~ s/[\n]/ /g;

    # Init execution timer
    $t0 = gettimeofday;
  
    $connection=$self->get_connection();

    # Issue command and get results.
    (undef,undef,$buffer,undef)   
        = Tharn::say(
            $connection,
            $args{cmd},
            -timeout=>$args{timeout},
            -eval=>sub{
                if (@_) {
                    $errmsg = join("",@_);
                    $errmsg =~ s/\n//g;
                    die ($self->_caller($errmsg),"\n")
                }
            });
    
    $t1=gettimeofday;
    $self->{command_time}=$t1-$t0;

    # Split the buffer into array
    $buffer=~s/\cI//g;
    @arr=split(/\n/,$buffer);
    
    
    
    # Remove command and trailing blank line results array
    @rows=(@arr[0..scalar(@arr)-1]);

    return (wantarray()) ? @rows : \@rows; 

} # end sub cmd


1;
