#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/VSA/Storage/Filer/Aggregate.pm#1 $
## @summary Aggregate interface
## @author dl-vsa-qa
## @status private

package VSA::Storage::Filer::Aggregate;

use strict;
use Data::Dumper qw(Dumper);
use Error qw(_clear_err _set_err);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR SCALARREF BOOLEAN OBJECT ARRAYREF UNDEF);
use Tharn qw($Log);
use VSA::Storage::Filer::Volume;
use VSA::UICommon qw(_caller _reformat_err _log _refresh_cache
                     _size_to_mbytes _trim);

use base qw(VSA::UICommon Error);

###############################################################################
# Method:    create_vol()
# Objective: Create volume on aggregate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_vol {
    my ($self,%args)=@_;
    my (
        $errmsg,
        $volObj,
    );
    
    ### Clear error buffer
    $self->_clear_err();
    
    eval {
        
        ### Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                name      => {type => SCALAR},
                size      => {type => SCALAR},
            }
        );
        
        $volObj=VSA::Storage::Filer::Volume->_new(
            name=>$args{name},
            size=>$args{size},
            _create=>1,
            _filer=>$self->{filer},
            _parent=>$self,
            _loghdl=>$self->_get_loghdl()
        );
            
        push @{$self->{volObjs}}, $volObj;  
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    return $volObj;
}

################################################################################
# Method:    info
# Objective: return general vol info 
# Details:   See POD documentation at the end of this file
################################################################################
sub info {
   my ($self,%args)=@_;
    my (
        $header,
        @bufs,
    );
    
    $header=(defined $args{header}) ? $args{header} : 1;
    
    push @bufs, $self->info_header() if ($header);
    
    push @bufs,
        sprintf("\t%-15.15s %-7.7s %-15.15s %-15.15s %-15.15s",   
        $self->get_name(),
        $self->get_raid_type(),
        $self->get_state(),
        $self->get_size(),
        $self->get_free_space());

    return (wantarray()) ? @bufs : [@bufs];
}

###############################################################################
# Method:    info_header
# Objective: return info header
# Details:   See POD documentation at the end of this file
################################################################################
sub info_header {
    my ($self)=@_;
    
    return
        sprintf("\t%-15.15s %-7.7s %-15.15s %-15.15s %-15.15s\n".
                "\t%-15.15s %-7.7s %-15.15s %-15.15s %-15.15s",
                "Aggregate","Raid","State","Size","Free",
                "="x15,"="x7,"="x15,"="x15,"="x15);
}
###############################################################################
# Method:    discover_vols()
# Objective: Discover volumes on aggregate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover_vols{
    my ($self,%args) = @_;
    my (
        @outputs,
        $vol_name,
        $cmd,
        $errmsg,
        @volObjs,
        );  
    
    ### Clear error buffer
    $self->_clear_err();
     
    $self->{volObjs}=[]; 
     
    ### Execute command
    eval {
        foreach $vol_name (@{$self->{vols}}){
            
            $self->_log(
                msg=>"creating vol object '".$vol_name."'",
                indent=>3
            );

            my $volObj=VSA::Storage::Filer::Volume->_new(
                name=>$vol_name,
                _filer=>$self->{filer},
                _parent=>$self,
                _loghdl=>$self->_get_loghdl()
            );
            
            push @{$self->{volObjs}}, $volObj;      
            
        }
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    ### Return result
    return (wantarray()) ? @{$self->{volObjs}} : $self->{volObjs}; 
    
}

###############################################################################
# Method:    get_*()
# Objective: retrieve object metadata
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_free_space { my ($self)=@_; $self->{free_space};}
sub get_hostname   { my ($self) = @_; return $self->{filer}->get_hostname();}
sub get_name       { my ($self) = @_; return $self->{name};    }
sub get_raid_type  { my ($self)=@_; $self->{raid_type}; }
sub get_size       { my ($self)=@_; $self->{size};      }
sub get_state      { my ($self)=@_; $self->{state};     }
    
###############################################################################
# Method:    get_vols()
# Objective: Get volume names
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vols { my($self)=@_;(wantarray()) ? @{$self->{vols}} : $self->{vols}; }

###############################################################################
# Method:    get_vol_cnt()
# Objective: Get volume cache count
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vol_cnt { my ($self) = @_; scalar (@{$self->{volObjs}}); }

################################################################################
# Subroutine Name: get_vol_objs
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol_objs {
    my ($self,%args) = @_;
    my (
        $cnt,
        $errmsg,
        $match,
        $value,
        @volObjs,
    );
    
    ### Clear error buffer
    $self->_clear_err();
       
    
    eval {     
        ## Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                cnt       => {type => SCALAR, optional=>1, default=>9999},
                name      => {type => SCALAR, optional=>1},
            }
        );
        
        $cnt=$args{cnt};
        delete $args{cnt};
        
        ### Remove object from cache that are marked destroyed
        $self->_refresh_cache(name=>"volObjs");
        
        
        ### Apply filter if provided              
        if (keys %args) {
            foreach my $volObj (@{$self->{volObjs}}){
                
                $match=0;
                
                foreach my $key (keys %args){
                    ### Assume exact parameter match if
                    ### no regex chars are included
                    $value=($args{$key} !~/\^|\$|\.|\*/)
                        ? '^'.$args{$key}.'$' 
                        : $args{$key};
                        
                    if ($key=~/name/){
                        $match=1 if ($volObj->get_name()=~/$value/); 
                    }
                
                    last if (not $match);
                }
                
                push @volObjs,$volObj if ($match);
            }
        }
        else {
            push @volObjs,@{$self->{volObjs}}
        }
        
        unless (@volObjs) {
            die "No Volumes Found";
        }
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    return $volObjs[0] if ($cnt==1);
    return (wantarray()) ? @volObjs[0..$cnt-1] : \@volObjs[0..$cnt-1];
    
}

###############################################################################
# Method:    refresh
# Objective: Initialize metadata for object instance
# Details:   See POD documentation at the end of this file
###############################################################################
sub refresh {
    my ($self)=@_;
    my ($errmsg);
    
    ### Initialize error buffer
    $self->_clear_err();
    
    eval {
        $self->_init();
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    return $self;
}

###############################################################################
#  P  R  I  V  A  T  E
###############################################################################
# Method:    _init
# Objective: Initialize metadata for new object instance
###############################################################################
sub _init {
    my ($self)=@_;
    my (
        $cmd,
        $errmsg,
        $free_space,
        $raid_type,
        $size,
        $state,
        @outputs,
    );
    
    $self->_log(msg=>"Aggregate::_init()");
    
    ### Execute command
    $cmd="aggr status ".$self->get_name();
    @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1);
    ($state,$raid_type)=$outputs[1]=~/(\S+)\s+(\S+),/i;
    $self->{raid_type}=$raid_type;
    $self->{state}=$state;
    
    my ($buffer)="@outputs"=~/Volumes:\s(.*)Plex/;
    $buffer=~s/ |\t|<none>//g; 
    
    #$self->_log(msg=>$buffer);
    @{$self->{vols}}=split(",",$buffer);
   
    $cmd="aggr show_space ".$self->get_name();
    @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1);
    ($size,$free_space)=$outputs[3]=~/(\S+)\s+\S+\s+\S+\s+(\S+)/i;
    $self->{size}=$self->_size_to_mbytes($size)."m";
    $self->{free_space}=$self->_size_to_mbytes($free_space)."m";
    
    #$self->{size}=$self->_size_to_mbytes($size)."mb";
    #$self->{free_space}=$self->_size_to_mbytes($free_space)."mb";       
 
    return 1;
}

###############################################################################
# Method:    _new
# Objective: Create new aggregate object
###############################################################################
sub _new {
    my ($class,%args) = @_;
    my (
      $errmsg,
      $self,
    );
    
    ### bless instance into base class
    $self= bless {}, $class;
        
    ### Validate input arguments
    ($class,%args) = validate_mixed (
        @_,
        [ { type => SCALAR} ],
        {
            name      => {type => SCALAR},
            timeout   => {type => SCALAR, optional=>1, default=>300,
                          regex => qr/^\d+$/ },
            _create    => {type => SCALAR, optional=>1, default=>0 },
            _filer     => {type => OBJECT},
            _loghdl    => {type => SCALARREF|UNDEF, optional=>1},
            _parent    => {type => OBJECT},
        }
    );
    
    ###  Lets specify what handler to use
    $self->{name}    = $args{name};
    $self->{timeout} = $args{timeout};
    $self->{filer}   = $args{_filer};
    $self->{parent}  = $args{_parent};
    $self->{create}  = $args{_create};
    $self->{loghdl}  = $args{_loghdl};
    $self->{volObjs} = [];
    
    if (not $self->{_create}){
        $self->_init();        
    }

    return $self;
}

1;
