#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/VSA/Storage/Filer/Volume.pm#1 $
## @summary Volume interface
## @author dl-vsa-qa
## @status private

package VSA::Storage::Filer::Volume;

use strict;
use Data::Dumper qw(Dumper);
use Error qw(_clear_err _set_err);
use File::Basename;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR SCALARREF BOOLEAN OBJECT ARRAYREF UNDEF);
use Tharn qw($Log);
use VSA::Storage::Filer::Lun;
use VSA::UICommon qw(_caller _reformat_err _log _refresh_cache
                     _size_to_mbytes _trim); 

use base qw(VSA::UICommon Error);



###############################################################################
# Method:    create_lun()
# Objective: Create lun on volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_lun {
    my ($self,%args)=@_;
    my (
        $errmsg,
        $lunObj,
    );
    
    ### Clear error buffer
    $self->_clear_err();
    
    eval {
        
        ### Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                name      => {type => SCALAR},
                size      => {type => SCALAR},
                type      => {type => SCALAR},
            }
        );
        
        $lunObj=VSA::Storage::Filer::Lun->_new(
            name=>$args{name},
            size=>$args{size},
            type=>$args{type},
            _create=>1,
            _filer=>$self->{filer},
            _parent=>$self,
            _loghdl=>$self->_get_loghdl()
        );
            
        push @{$self->{lunObjs}}, $lunObj;  
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    return $lunObj;
}



###############################################################################
# Method:    destroy
# Objective: destroy volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self,%args)=@_;
    my (
        $cmd,
        $errmsg,
        @outputs
    );
    
    ### Clear error buffer
    $self->_clear_err();
    
    eval {
        $cmd="vol destroy ".$self->get_name()." -f";
        
        @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1);
        unless ($outputs[0]=~/volume (.*) destroyed./i){
            die $outputs[0];
        }
        
        $self->{state}="DESTROYED";
        
        
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self;
    }
     
    return $self;
}

###############################################################################
# Method:    discover_luns()
# Objective: Discover volumes on aggregate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover_luns{
    my ($self,%args) = @_;
    my (
        @outputs,
        $name,
        $path,
        $size,
        $access,
        $state,
        $mapped,
        $cmd,
        $errmsg,
        );  
    
    ### Clear error buffer
    $self->_clear_err();
     
    ### Execute command
    eval {
        
        $cmd="lun show";
        
        @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1);      
        use File::Basename;
        my $vol=$self->get_name();
        my @lines=grep(/$vol\//,@outputs);
        
        foreach my $line (@lines) {
            ($path,$size,$access,$state,$mapped)= 
                $line=~/(\S+)\s+(\S+) \(\d+\)\s+\((\S+), (\S+), (\S+)\)/;
            
            $name=basename($path);
          
            $self->_log(
                msg=>"creating lun object '".$name."'",
                indent=>3
            );
             
            my $lunObj=VSA::Storage::Filer::Lun->_new(
                name=>$name,
                size=>$size,
                access=>$access,
                state=>$state,
                mapped=>$mapped,
                path=>$path,
                _filer=>$self->{filer},
                _parent=>$self,
                _loghdl=>$self->_get_loghdl()
            );
            
            push @{$self->{lunObjs}}, $lunObj;
        
        }    
        
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    ### Return result
    return (wantarray()) ? @{$self->{lunObjs}} : $self->{lunObjs}; 
    
}

###############################################################################
# Method:    get_*()
# Objective: retrieve object metadata
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_aggr      { my ($self)=@_; $self->_clear_err(); $self->{aggr};}
sub get_hostname { my ($self) = @_; return $self->{filer}->get_hostname();}
sub get_lun_cnt { my ($self) = @_; return scalar (@{$self->{lunObjs}}); }
sub get_name      { my ($self)=@_; $self->_clear_err(); $self->{name};}
sub get_size      { my ($self)=@_; $self->_clear_err(); $self->{size};}
sub get_state     { my ($self)=@_; $self->_clear_err(); $self->{state};   }
sub get_timeout   { my ($self)=@_; $self->_clear_err(); $self->{timeout}; }
sub get_type      { my ($self)=@_; $self->_clear_err(); $self->{type};}


################################################################################
# Subroutine Name: get_lun_ids
# Objective:       return array of lun ids
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_ids {
    my ($self,%args) = @_;
    my (
        $errmsg,
        @lun_ids,
        @lunObjs,
       );
    
    ### Clear error buffer
    $self->_clear_err();
    
    eval {
        @lunObjs=$self->get_lun_objs();
        foreach my $lunObj (@lunObjs){
            push @lun_ids,$lunObj->get_id();
        }
    };
    if ($@) {
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return (wantarray()) ? () : []; 
    }
		
    return (wantarray()) ? @lun_ids : [@lun_ids];
    
}

################################################################################
# Subroutine Name: get_lun_objs
# Objective:       return array of lun objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_objs {
    my ($self,%args) = @_;
    my (
        $errmsg,
        $match,
        $value,
        @lunObjs,
    );
    
    ### Clear error buffer
    $self->_clear_err();
       
    ### Validate input arguments
    eval {
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                name      => {type => SCALAR, optional=>1},
            }
        );
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return (wantarray()) ? () : []; 
    }
    
    ### Remove object from cache that are marked destroyed
    $self->_refresh_cache(name=>"lunObjs");
    
    
    ### Apply filter if provided              
    if (keys %args) {
        foreach my $lunObj (@{$self->{lunObjs}}){
            
            $match=0;
            
            foreach my $key (keys %args){
                ### Assume exact parameter match if
                ### no regex chars are included
                $value=($args{$key} !~/\^|\$|\.|\*/)
                    ? '^'.$args{$key}.'$' 
                    : $args{$key};
                    
                if ($key=~/name/){
                    $match=1 if ($lunObj->get_name()=~/$value/); 
                }
            
                last if (not $match);
            }
            
            push @lunObjs,$lunObj if ($match);
        }
        
        unless (@lunObjs) {
            $errmsg="No Luns Found";
            ($self->on_err()=~/die/)
                ? die ($self->_caller($errmsg),"\n")
                : $self->_set_err($errmsg);
            return (wantarray()) ? () : [];    
        }
    }
    else {
        push @lunObjs,@{$self->{lunObjs}}
    }
    
    return (wantarray()) ? @lunObjs : \@lunObjs;
    
} 
  
sub get_lun_hsh {
    my ($self,%args) = @_;
    my (
        $hsh_of,
        @lunObjs,
    );
    
    ### Clear error buffer
    $self->_clear_err();
    
    ### Retrieve luns objects
    @lunObjs=$self->get_lun_objs();
    
    foreach my $lunObj (@lunObjs){
        $hsh_of->{$lunObj->get_id()} = {
            iqn      => $self->{filer}->get_iqn(),
            hostname => $lunObj->get_hostname(),
            path     => $lunObj->get_path(),
            id       => $lunObj->get_id(),
        }   
	}

    
    return $hsh_of;
}
  
  
################################################################################
# Method:    info
# Objective: return general vol info 
# Details:   See POD documentation at the end of this file
################################################################################
sub info {
    my ($self,%args)=@_;
    my (
        $header,
        @bufs,
    );
    
    $header=(defined $args{header}) ? $args{header} : 1;
    
    push @bufs, $self->info_header() if ($header);
        
    push @bufs,
        sprintf("\t%-15.15s %-15.15s %-15.15s",   
        $self->get_name(),
        $self->get_state(),
        $self->get_size());

    return (wantarray()) ? @bufs : [@bufs];
     
}

###############################################################################
# Method:    info_header
# Objective: return info header
# Details:   See POD documentation at the end of this file
################################################################################
sub info_header {
    my ($self)=@_;
    
    return
        sprintf("\t%-15.15s %-15.15s %-15.15s\n".
                "\t%-15.15s %-15.15s %-15.15s",
                "Volume","State","Size",
                "="x15,"="x15,"="x15);
}

###############################################################################
# Method:    info_luns
# Objective: return info about all luns
# Details:   See POD documentation at the end of this file
################################################################################ 
sub info_luns {
    my ($self)=@_;
    my (
        @bufs
       );
    
    
    if (defined @{$self->{lunObjs}}){
        push @bufs, @{$self->{lunObjs}}->[0]->info_header();
    }
            
            
    foreach my $lunObj (@{$self->{lunObjs}}){
        push @bufs,$lunObj->info(header=>0);  
    }
    
    return (wantarray()) ? @bufs : [@bufs];
      
}
      
###############################################################################
# Method:    offline
# Objective: Offline volume
# Details:   See POD documentation at the end of this file
###############################################################################
sub offline {
    my ($self,%args)=@_;
    my (
        $cmd,
        $errmsg,
        @outputs
    );
    
    ### Clear error buffer
    $self->_clear_err();
    
    $cmd="vol offline ".$self->get_name();
    eval {     
        @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1);
        unless ($outputs[0]=~/now offline|already offline/i){
            die $outputs[0];
        }
    
        $self->{state}="offline";
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }
    
    return $self;
}

###############################################################################
# Method:    online
# Objective: online volume
# Details:   See POD documentation at the end of this file
###############################################################################
sub online {
    my ($self,%args)=@_;
    my (
        $cmd,
        $errmsg,
        @outputs
    );
    
    ### Clear error buffer
    $self->_clear_err();
    
    $cmd="vol online ".$self->get_name();
    eval {     
        @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1);
        unless ($outputs[0]=~/now online|already online/i){
            die $outputs[0];
        }
    
        $self->{state}="online";
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }
    
    return $self;
}

###############################################################################
# Method:    options
# Objective: Change whether snapshots occur or not
# Details:   See POD documentation at the end of this file
###############################################################################
sub options {
    my ($self,%args)=@_;
    my (
        $cmd,
        $errmsg,
        @outputs
    );
    
    ### Clear error buffer
    $self->_clear_err();
       
    ### Validate input arguments
    eval {
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                nosnap=> {type => SCALAR, default=>0, optional=>1,
                             callbacks=>{
                                'legal values'=>sub{shift()=~
                                         /1|0/i}}},
            }
        );
    
        foreach my $option qw(nosnap){
    
            $cmd="vol options ".$self->get_name()." $option"." ".$args{$option};
           
            @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1);
            if ($outputs[0]=~/usage:/i){
                die $outputs[0];
            }
        
            $self->{$option}=$args{$option};
        }
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }
    
    return $self;
}

###############################################################################
# Method:    refresh
# Objective: Initialize metadata for object instance
# Details:   See POD documentation at the end of this file
###############################################################################
sub refresh {
    my ($self)=@_;
    my ($errmsg);
    
    ### Initialize error buffer
    $self->_clear_err();
    
    eval {
        $self->_init();
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return ""; 
    }
    
    return $self;
}

###############################################################################
# Method:    snap_reserve
# Objective: Change space percentage reserved for snapshot
# Details:   See POD documentation at the end of this file
###############################################################################
sub snap_reserve {
    my ($self,%args)=@_;
    my (
        $cmd,
        $errmsg,
        @outputs
    );
    
    ### Clear error buffer
    $self->_clear_err();
    
    
    $args{pct}=~s/\%//g;   
       
       
    ### Validate input arguments
    eval {
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                pct      => {type => SCALAR,
                             callbacks=>{
                                'Between 0 and 100'=>
                                    sub{shift() =~/([0-100])/}}},
            }
        );
    
        $cmd="snap reserve ".$self->get_name()." ".$args{pct};
       
        @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1);
        if ($outputs[0]=~/snap reserve:/i){
            die $outputs[0];
        }
    
        $self->{snap_reserve}=$args{pct};
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{filer}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }
    
    return $self;
}

###############################################################################
#  P  R  I  V  A  T  E
###############################################################################
# Method:    _create
# Objective: Create new volume
###############################################################################
sub _create {
    my ($self,%args)=@_;
    my (
        $cmd,
        @outputs
    );
    
    $cmd="vol create ".
         $self->get_name()." ".
         $self->get_aggr()." ".
         $self->get_size();
    
    @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1,
                                      timeout=>$self->get_timeout());
    unless ($outputs[0]=~/Creation of volume/){
        die $outputs[0];
    }
    
    1;
}

###############################################################################
# Method:    _init
# Objective: Initialize metadata for object instance
###############################################################################
sub _init {
    my ($self)=@_;
    my (
        $cmd,
        $errmsg,
        $type,
        $size,
        $state,
        $used,
        @outputs,
    );

    ### Execute command
    $cmd="vol status ".$self->get_name();
    @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1,
                                      timeout=>$self->get_timeout());
    ($state,$type)=$outputs[1]=~/\S+\s+(\S+)\s+\S+, (\S+)/i;
    $self->{type}=$type;
    $self->{state}=$state;
     
    $cmd="vol size ".$self->get_name();    
    @outputs=$self->{filer}->_execute(cmd=>$cmd,hide=>1);
    ($size)=$outputs[0]=~/(\S+).$/i;
    $self->{size}=$self->_size_to_mbytes($size)."m";
    #$self->{used}=$used;
    #$self->{size}=
    #$self->{free_space}=$self->_size_to_mbytes($free_space)."mb";       
  
    return 1;
}

###############################################################################
# Method:    _new
# Objective: Create new volume object
###############################################################################
sub _new {
    my ($class,%args) = @_;
    my (
      $errmsg,
      $self,
    );
    
    ### bless instance into base class
    $self= bless {}, $class;
        
    ### Validate input arguments
    ($class,%args) = validate_mixed (
        @_,
        [ { type => SCALAR} ],
        {
            name     => {type => SCALAR},
            size     => {type => SCALAR, optional=>1}, 
            timeout  => {type => SCALAR, optional=>1, default=>30,
                          regex => qr/^\d+$/ },
            _create  => {type => SCALAR, optional=>1, default=>0 },
            _filer   => {type => OBJECT},
            _loghdl  => {type => SCALARREF|UNDEF, optional=>1},
            _parent  => {type => OBJECT},
        }
    );
    
    ###  Initialize object metadata
    $self->{name}    = $args{name};
    $self->{size}    = $args{size};
    $self->{aggr}    = $args{_parent}->get_name();
    $self->{timeout} = $args{timeout};
    $self->{filer}   = $args{_filer};
    $self->{parent}  = $args{_parent};
    $self->{create}  = $args{_create};
    $self->{loghdl}  = $args{_loghdl};
    $self->{lunObjs} = [];
    
    $self->_create() if ($self->{create});
    
    $self->refresh();

    return $self;
}


1;
