#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/VSA/Vmhost/ESXvmfs.pm#1 $
## @summary ESX vmfs interface
## @author dl-vsa-qa
## @status private


# Package Name
##############
package VSA::Vmhost::ESXvmfs;

use strict;

use File::Basename;
use Data::Dumper qw(Dumper);
use Error qw(_clear_err _set_err);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR SCALARREF BOOLEAN OBJECT ARRAYREF UNDEF);
use POSIX;
use Tharn qw($Log);
use VSA::Vmhost::Lvm;
use VSA::UICommon qw(_caller _reformat_err _log _refresh_cache
                     _size_to_mbytes _trim);
                     
use base qw(VSA::Vmhost::Lvm VSA::UICommon Error );

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: create_datastores()
# Objective:       creates all vmfs datastores for every device passed into the 
#                  -vmfs_devices parameter.
# Details:         See POD documentation at the end of this file
################################################################################
sub create_datastores {
    my ($self, %args) = @_;
    my (
        $bsize,
        $cmd,
        $datastore,
        $device,
        $devObjs,
        $errmsg,
        $lun_path,
        $name,
        $partition,
        $status,
        $type,
        @datastores,
        @outputs,
        @vmfs_devices,
    );
    
    ### Initialize error buffer
    $self->_clear_err();

    eval {
        
        ### Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                bsize          =>  {type => SCALAR, optional=>1,
                                    default=>"1m",
                                    callbacks=>{
                                        'legal values'=>sub{shift()=~
                                         /^([1248][mMkK]|auto)$/i
                                    }}},
                name           =>  {type => SCALAR},
                dev_objs        => {type => ARRAYREF|OBJECT },                   
                type           =>  {type => SCALAR, optional=>1,
                                    default=>"vmfs3",
                                    callbacks=>{
                                        'legal values'=>sub{shift()=~
                                        /vmfs3/i
                                    }}},                   
            }
        );
        
        $bsize               = $args{bsize};
        $name                = $args{name};
        $devObjs             = (ref ($args{dev_objs}) eq "ARRAY")
                                ? $args{dev_objs}
                                : [$args{dev_objs}];                        
        $type                = $args{type};
    
		$self->_log(msg=>"sleeping 5s before formatting devices",indent=>3);
		sleep(5);
		
        foreach my $devObj (@$devObjs){
            
            ### Get metadata from device object
            $device    = $devObj->get_name();
            $lun_path  = $devObj->get_filer().":".$devObj->get_lun_path();
            $partition  = $devObj->get_vmkdisk_name().":1";
            $datastore = $name."_" .$devObj->get_lun_name();
            $bsize = $self->_calculate_block_size($devObj,"one-to-one",1) 
                if ($bsize =~ /^auto$/i);
                
            ### Format the device
            $self->_log(msg => "formating device $device ($lun_path)",indent=>3); 
            $cmd = "echo 'n\np\n1\n\n\nt\nfb\nx\nb\n1\n128\np\nw\n'|fdisk ".$device;     
            ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);                              
            if ($status > 0) {
                die "@outputs\n";    
            }
           
            ### Create the vmfs filesystem (datastore)                     
            $self->_log(msg => "creating VMFS datastore '$datastore'",indent=>3);
            $cmd = "/usr/sbin/vmkfstools -C $type -b $bsize -S $datastore ". 
                   "/vmfs/devices/disks/" . $partition;     
            ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
            if ($status > 0) {
               die "@outputs\n";  
            }
            
            ### Refresh storage so its recongnized 
            $self->_log(msg => "refreshing storage",indent=>3);
            $cmd = "vimsh -n -e \"/hostsvc/storage/refresh\"";  
            ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);                              
            if ($status > 0) {
                die "@outputs\n";   
            }
            $self->_log(msg=>"");
			
            push @datastores, $datastore;
        }
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{parent}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }
    
    return (wantarray()) ? @datastores : [@datastores];
}

################################################################################
# Subroutine Name: create_vmdks()
# Objective:       creates vmfs vmdk files for each datastore
# Returns:         a list of vmfs vmdk file paths.
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vmdks {
    my ($self, %args) = @_;
    my (
		$block_size,
        $cmd,
        $errmsg,
        $controller, 
        $name,
        $reserve,
        $status,
        $uuid,
        $vmdk_cnt,
        $vmdk_size, 
        $datastore,
        @outputs,
        @vmdk_files,
    );
    
    ### Initialize error buffer
    $self->_clear_err();
   
    eval {
        ### Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
				block_size     =>  {type => SCALAR, 
                                    callbacks=>{
                                        'legal values'=>sub{shift()=~
                                         /1m|2m|4m|8m/i
                                    }}},
                controller     =>  {type => SCALAR, optional=>1,
                                    default=>"lsilogic",
                                    callbacks=>{
                                        'legal values'=>sub{shift()=~
                                         /lsilogic|buslogic/i
                                    }}},
                datastore      =>  {type => SCALAR },
                reserve        =>  {type => SCALAR, optional=>1, default=>0},
                vmdk_cnt       =>  {type => SCALAR, optional=>1, default=>1},
                vmdk_name      =>  {type => SCALAR, optional=>1 },
                vmdk_size      =>  {type => SCALAR, optional=>1 }, 
                                                      
            }
        );
        
		$block_size  = $args{block_size};
        $controller  = $args{controller};
        
        $name        = (defined $args{vmdk_name})
                        ? $args{vmdk_name}
                        : "data";
        $datastore   = $args{datastore};
        $reserve     = $args{reserve};
        $vmdk_cnt    = $args{vmdk_cnt};
        $vmdk_size   = $args{vmdk_size};
        $uuid        = $self->{parent}->_get_uuid($datastore);
   
        $self->_log(msg=>"create vmdk on datastore $datastore",indent=>3); 
   
        ### Determine vmdk_size based on vmdk_cnt if size was
        ### not passed in
        unless (defined $vmdk_size) {
            $vmdk_size = $self->_calculate_vmdk_size(
                $datastore,
                $vmdk_cnt,
                $reserve,
               );
        }
    
        # create $num_vmfs_vmdks files in this datastore that have size
        # $vmdk_size_in_kbytes
        for (my $i = 1; $i <=$vmdk_cnt; $i++) {
            my $vmdk_name ="/vmfs/volumes/$uuid/$name"."_".$i.".vmdk"; 
            $self->_log(msg=>"creating $vmdk_name ($vmdk_size)",indent=>3); 
            
            $cmd = "/usr/sbin/vmkfstools ".
				   "-b ".$block_size." ".
                   "-c ".$vmdk_size." ".
                   "-a ".$controller." ".
                   $vmdk_name;
               
            ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
            if ($status > 0) {
                die "@outputs\n"; 
            } 
                 
            push(@vmdk_files, $vmdk_name);  
        } 

    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{parent}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }
    
    # store the datastore array and datastore type in the lvm and return
    # the datastore array to the caller.
    $self->{vmdk_files} = @vmdk_files;
    
    $self->_log(
        msg => scalar(@vmdk_files)." vmdk file(s) created in ".$datastore,
        indent=>3,
        );

	$self->_log(msg=>"");

    return (wantarray()) ? @vmdk_files : \@vmdk_files;
}

################################################################################
# Subroutine Name: destroy_datastores()
# Objective:       destroys datastores that are no longer in use.
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_datastores {
    my ($self, %args) = @_;
    my (
        $cmd,
        $cnt,
        $dstore,
        $errmsg,
        $name,
        $status,
        @datastores,
        @outputs,
    );
    
    ### Initialize error buffer
    $self->_clear_err();

    eval {
        ### Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                name         => {type => SCALAR},                 
            }
        );
             
        ### Remove underscore and dash chars from name
        $cnt     =   0;
        $name    =   $args{name};
        $dstore  =   $name;
        $dstore  =~  s/-|_//g; 
        
        ### Retrieve the list of datastores
        $cmd = "ls /vmfs/volumes/ | egrep '^".$dstore."_'";
        ($status, @datastores) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
        if ($status > 0) {
            $self->_log(msg=>"No datastores found for '$name'", indent=>3); 
        }
        
        foreach my $datastore (@datastores) {
            ### Sometimes ESX magically places a \s(\d+) at the end
            ### of a datastore if it detects duplicate datastore names
            ### that have different UUIDs across multiple ESX machines.
            ### Escape the parentheses and spaces.
            my $datastore_orig=$datastore;
            $datastore=~s#(\(|\)|\s)#\\$1#g;
            
            ### Skip datastore containing vmdks
            $cmd = "ls /vmfs/volumes/". $datastore;
            ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
            if ($status > 0) {
                #atastore link is bad and
                # was deleted improperly. Ignore this datastore, ESX will clean it
                # up next time it creates a datastore with that name, or reboots.
                $self->_log(
                    msg => ["WARNING: The datastore '$datastore' ",
                            "is a dangling symbolic link. Its LUN ",
                            "was probably orphaned."],
                    indent=>3,
                    ); 
                next;            
            }
            elsif (scalar(@outputs) > 0) {
                # datastore is not empty, this is very bad
                $self->_log(
                    msg =>
                        ["WARNING: The VMFS Datastore '$datastore' belongs",
                         "to this host and still has VMDKs in it that are",
                         "not associated with any of the virtual machines ",
                         "that are a part of this test. This is probably ",
                         "going to cause severe problems the next time you",
                         "try to configure this host."],
                    indent=>3,    
                        );
                next;
            }
           
            # Remove '/' at the end, chop it off.
            chop($datastore) if ($datastore =~ /\/$/); 

            ### Remove datastore
            $self->_log(msg=>"removing datastore '$datastore_orig'",indent=>3);
            $cmd = "vimsh -n -e \"hostsvc/datastore/remove $datastore\"";
            ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
            if ("@outputs" =~ /is in use/i){
                die "Can't remove datastore $datastore_orig because its still ".
                    "in use";
            }
            elsif ($status > 0) {
                die "@outputs\n"; 
            } 
    
            $cnt++;
        }

    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{parent}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }    
    
    
    $self->_log(
        msg => $cnt." datastore(s) has/have been removed.",
        indent=>3,
        );
    
    return $self;
}

sub get_datastores {
    my ($self, %args) = @_;
    my (
        $cmd,
        $cnt,
        $dstore,
        $errmsg,
        $name,
        $status,
        @datastores,
        @outputs,
    );
    
    ### Initialize error buffer
    $self->_clear_err();

    eval {
        ### Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                name         => {type => SCALAR},                 
            }
        );
             
        ### Remove underscore and dash chars from name
        $cnt     =   0;
        $name    =   $args{name};
        $dstore  =   $name;
        $dstore  =~  s/-|_//g; 
        
        ### Retrieve the list of datastores
        $cmd = "ls /vmfs/volumes/ | egrep '^".$dstore."_'";
        ($status, @datastores) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
        if ($status > 0) {
            $self->_log(msg=>"No datastores found for '$name'", indent=>3); 
        }
    };
    if ($@){
        $errmsg=$self->_reformat_err($@);
        ($self->{parent}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }    
    
    return (wantarray()) ? @datastores : [@datastores];
    
}

################################################################################
# Subroutine Name: destroy_vmdks()
# Objective:       destroys the vmfs vmdk files in use by -vm_name
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vmdks {
    my ($self, %args) = @_;
    my (
        $cmd,
        $delete_cnt,
        $errmsg,
        $status,
        $vm_name,
        @outputs,
        @vmdk_files,
    );

    ### Initialize 
    $self->_clear_err();
    $delete_cnt=0;
    
    eval {
        
        ## Validate input arguments
        ($self,%args) = validate_mixed (
            @_,
            [ { type => OBJECT} ],
            {
                vm_name  => {type => SCALAR},
            }
        );
        
        $vm_name=$args{vm_name};
                    
        ### Retrieve the vmdk files from the vm
        @vmdk_files =
            $self->{parent}->get_vmdks(vm_name=>$vm_name, lvm_type=>'vmfs');
        
        foreach my $vmdk_file (@vmdk_files) {
        
            ### Destroy vmdk file
            $self->_log(msg=>"destroying vmdk file ".$vmdk_file, indent=>3);
            $cmd = "/usr/sbin/vmkfstools -U '".$vmdk_file."'";
            ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
            if ($status > 0) {
                die "@outputs\n"; 
            } 
                     
            ### Sleep time to process command
            sleep(1);
            
            ### verify that the vmdk file was destroyed.
            $cmd = "ls ".$vmdk_file;
            ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
            if ($status==0) {
                # the file was not deleted properly, sleep and try again.
                die "vmdk file $vmdk_file not deleted";
            }
    
            $delete_cnt++;
            
        }
        
    };
    if ($@){
        $self->_log(msg=>"DBG (lvm::on_err): on_err=".$self->{parent}->on_err());
        $errmsg=$self->_reformat_err($@);
        ($self->{parent}->on_err()=~/die/)
            ? die ($self->_caller($errmsg),"\n")
            : $self->_set_err($errmsg);
        return $self; 
    }
    
    $self->_log(
        msg => $delete_cnt ." vmdk file(s) destroyed for $vm_name",
        indent=>3);
        
    return $self;
}

################################################################################
# Subroutine Name: destroy_spare_vmdks()
# Objective:       destroys spare vmdks passed into the function
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_spare_vmdks {
    my ($self, %args) = @_;

    my ($cmd, @output, $status);
    my (@vmfs_vmdk_files);

    # parse arguments
    foreach (keys %args) {
        if (/-?vmfs_vmdk_files/) {
            @vmfs_vmdk_files = @{$args{$_}};
        }
    }
    
    # if there are no nate vmfs vmdk files in use by this guest,
    # then just return.
    if (scalar(@vmfs_vmdk_files) == 0) {
        $self->sanlog(-msg => "No spare VMFS VMDKs to destroy.");
        return $ClientUtil::TRUE;
    }
    
    my $retry_count = 1;
    my $delete_retry_count = 3;
    my $number_of_vmdks_destroyed = 0;

    # destroy each vmfs vmdk file
    foreach my $vmfs_vmdk_file_path (@vmfs_vmdk_files) {
        $cmd = "/usr/sbin/vmkfstools -U $vmfs_vmdk_file_path";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status > 0) {
            # if its a "system cannot find the file specified" error,
            # log a comment concerning it, but dont set an error, just continue.
            if (grep {/cannot find the file specified/i} @output) {
                $self->sanlog(-msg => "Warning: The command '$cmd' attempted " .
                                      "to remove a vmdk that does not exist.");
            } elsif (grep {/Failed to delete virtual disk: Resource temporarily unavailable/i} @output and 
                     $delete_retry_count > 0) {
                # if we failed to delete it, then sleep a bit and try again.
                &SanUtil::nate_sleep(10);
                $delete_retry_count--;
                redo;
            } else {
                $self->_set_err("Error: Lvm::ESX::VMFS::destroy_spare_vmdks() " .
                                "Command ($cmd) returned a bad " .
                                "status ($status)");
                return $ClientUtil::FALSE;
            }
        } 
        
        # give esx some time to process the command
        &SanUtil::nate_sleep(1);
        
        # verify that the vmdk file was destroyed.
        $cmd = "ls $vmfs_vmdk_file_path";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status == 0 and $retry_count == 1) {
            # the file was not deleted properly, sleep and try again.
            &SanUtil::nate_sleep(1);
            $retry_count = 0;
            redo;
        }

        $number_of_vmdks_destroyed++;
        $retry_count = 1;
    }
    
    $self->sanlog(-msg => "($number_of_vmdks_destroyed) Spare VMFS VMDK(s) destroyed.");
        
    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my (
        $hosttype,
        $lvm,
        $self,
        );
 
    ### bless instance into class
    $self= bless{}, $class;
    
    ### Validate input arguments
    ($class,%args) = validate_mixed (
        @_,
        [ { type => SCALAR} ],
        {
            lvm       => {type => SCALAR, 
                              callbacks=>{
                                'legal values'=>sub{shift()=~/vmfs/i
                             }}},
            _parent    => {type => OBJECT},
        }
    );
    
    $lvm=$args{lvm};
    $self->{parent}=$args{_parent};
    $self->{loghdl}=$self->{parent}->_get_loghdl();;
    $hosttype=$self->{parent}->get_hosttype();
    
    ### initilize the object members
    #$self->{conn} = $parent->get_connection();
    #$self->{sanlun} = $parent->get_sanlun();
    $self->{datastore_type} = "one-to-one";
    
    
    return $self;
}

################################################################################
# Subroutine Name: _convert_datastore_to_uuid()
# Objective:       Given the datastore name, convert to the UUID label
# Details:         See POD documentation at the end of this file
################################################################################
sub _convert_datastore_to_uuid {
    my ($self, $datastore) = @_;
    
    my ($cmd, $status, @output, $uuid);
    my $retry_count = 2;
    my $original_datastore = $datastore;

    $self->_log(msg=>"_convert_datastore_to_uuid()", indent=>3);

    while (1) {
        # grab this datastores uuid and return it
        $cmd = "/usr/sbin/vmkfstools -P -h /vmfs/volumes/$datastore | grep UUID | " .
               "awk {'print \$2'}";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if (($status > 0 or grep(/Could not open/i, @output)) and $retry_count > 0) {
            # sometimes the datastore named changed because it is shared across multiple hosts.
            # attempt to discover the new name
            my $ls_cmd = "ls /vmfs/volumes/ | grep $original_datastore";
            my ($status, @datastore_output) = &ClientUtil::execute($self->{parent}->{conn}, $ls_cmd);
            if ($status > 0 or scalar(@datastore_output) != 1) {
                $self->_set_err("Error: Lvm::ESX::VMFS::_convert_datastore_to_uuid() ". 
                                "Could not find the datastore $original_datastore.");
                return ("", $original_datastore);     
            }
            
            # update the datastore with the new name
            $datastore = $datastore_output[0];

            # escape spaces and parenthesis characters
            $datastore =~ s#\(#\\\(#g;  
            $datastore =~ s#\)#\\\)#g;  
            $datastore =~ s# #\\ #g;
            
            # retry the loop
            $retry_count--;
            redo;
        } elsif ($status == 0) {
            # then we successfully ran the command, break out
            last;
        } else {
            # otherwise we could not find this datastore
            $self->_set_err("Error: Lvm::ESX:VMFS::_convert_datastore_to_uuid() ".
                            "Could not obtain the UUID value for the datastore ".
                            "'$original_datastore'"); 
            return ("", $original_datastore);
            
        }
    }
    
    return ($output[0], $datastore);
}

################################################################################
# Subroutine Name: _calculate_vmdk_size()
# Objective:       Given the size of a datastore and the number of vmdks to 
#                  carve out of the datastore, calculate the size of each vmdk 
#                  in kilobytes.
# Details:         See POD documentation at the end of this file
################################################################################
sub _calculate_vmdk_size {
    my ($self, $datastore, $vmdk_cnt, $reserve) = @_;
    my (
        $available,
        $cmd,
        $line,
        $overhead,
        $status,
        $vmdk_size,
        @outputs,
        );
    
    ### metadata overhead of each created
    ### vmdk in datasource (kbytes)
    $overhead=8000;
    #$overhead=110000;
    ### Calculate maximum vmdk size
    $self->_log(
        msg => "calculating size for $vmdk_cnt datastore vmdks",indent=>3);
    
    # sometimes the datastore named changed because it is shared across multiple hosts.
    # attempt to discover the new name
    $cmd = "ls /vmfs/volumes/ | grep \"$datastore\"";
    ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd,hide=>1);

    if ($status) {
        die "@outputs";
    }
    $datastore=$outputs[0];
    
    # sometimes ESX magically places a \s(\d+) at the end of a datastore
    # if it detects duplicate datastore names that have different UUIDs
    # across multiple ESX machines. Escape the parentheses and spaces.   
    $datastore=~s#(\(|\)|\s)#\\$1#g;
   
    $cmd = "/usr/sbin/vmkfstools -P -h /vmfs/volumes/".$datastore;      
    ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
    if ($status > 0 or not @outputs) {
        die "Could not determine datastore '$datastore' size."; 
    } 
    
    ($line)=grep(/available/,@outputs);
    $line =~ /Capacity\s+[0-9\.]+\s*[A-Za-z]+,\s+([0-9\.]+)\s*([A-Za-z]+)/;
    $available=$1.substr(lc($2),0,1);
    $vmdk_size=
      POSIX::floor((($self->_size_to_kbytes($available)-
                    $self->_size_to_kbytes($reserve))/$vmdk_cnt)-
                    ($overhead*8))."k";
    $self->_log(msg => "available: ".$available, indent=>6);
    $self->_log(msg => "vmdk_size: ".$vmdk_size, indent=>6);
    
    return $vmdk_size;
}

################################################################################
# Subroutine Name: _calculate_block_size()
# Objective:       given the size of device, datastore type, and the number
#                  of vmfs luns to be used per datastore, calculate blocksize
# Details:         See POD documentation at the end of this file
################################################################################
sub _calculate_block_size {
    my ($self, 
        $vmfs_device,
        $datastore_type, 
        $num_luns_per_datastore) = @_;

    my $range_block_size_1m = 256 * 1024 * 1024 * 1024;  # 256g
    my $range_block_size_2m = 512 * 1024 * 1024 * 1024;  # 512g 
    my $range_block_size_4m = 1024 * 1024 * 1024 * 1024; # 1024g

    my $datastore_size = 0;

    # determine the size of the datastore that will be created.
    if ($datastore_type =~ /concat/i) {
        $datastore_size = $vmfs_device->{lun_size} * $num_luns_per_datastore;
    } else {
        $datastore_size = $vmfs_device->{lun_size};
    }

    # determine the range and return the appropriate blocksize
    if ($datastore_size < $range_block_size_1m) {
        return "1m";
    } elsif ($datastore_size < $range_block_size_2m) {
        return "2m";
    } elsif ($datastore_size < $range_block_size_4m) {
        return "4m";
    } else {
        return "8m";
    }
    
}


1;
