#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/VSA/Vmhost/Sanlun.pm#1 $
## @summary interface to sanlun tool to map filer to host luns
## @author dl-vsa-qa
## @status private

package VSA::Vmhost::Sanlun;

use Data::Dumper qw(Dumper);
use Error qw(_clear_err _set_err);
use IO::File;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR SCALARREF BOOLEAN OBJECT ARRAYREF UNDEF);
use Time::HiRes qw(gettimeofday);
use Tharn qw($Log);
use Utils qw(make_host);
use VSA::UICommon qw(_caller _reformat_err _log _refresh_cache
                     _size_to_mbytes _trim);
                     
use base qw(VSA::UICommon Error);

################################################################################
# P U B L I C    M E T H O D S
################################################################################
###############################################################################
# Method:    new
# Objective: Sanlun Object constructor.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($class, %args) = @_;
    my (
        $hosttype,
        $self,
        );
 
    ### bless instance into class
    $self= bless{}, $class;
    
    ### Validate input arguments
    ($class,%args) = validate_mixed (
        @_,
        [ { type => SCALAR} ],
        {
            _parent    => {type => OBJECT},
        }
    );
    
    $self->{parent}=$args{_parent};
    $hosttype=$self->{parent}->get_hosttype();
    
    ## load the appropriate Sanlun for the hosttype
    if ($hosttype =~ /linux-vmware/i) {
        require VSA::Vmhost::ESXSanlun;
        $self = VSA::Vmhost::ESXSanlun->_new(parent=>$self->{parent});
    }
    else {
        die "Platform '$hosttype' not currently supported by Sanlun object";
    }

    $self->{version} = $self->_version();

    return ($self);

}


###############################################################################
# Method:    get_* metadata
# Objective: Returns sanlun metadata
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_version { my ($self) = @_; $self->{version};    }


################################################################################
# P R I V A T E   M E T H O D S
################################################################################
###############################################################################
# Method:    _version
# Objective: Queries the Sanlun version and build number on this client.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _version {
    my ($self) = @_;
    my (
        $cmd,
        $status,
        $version,
        @outputs,
    );
    
    ### Initialize error buffer    
    $self->_clear_err();

    $cmd = "/usr/sbin/sanlun version";   
    eval {
        ($status, @outputs) = $self->{parent}->_execute(cmd=>$cmd, hide=>1);
        if ($status > 0) {
            die "sanlun not installed on host";
        }
    };
 
    return $outputs[0];

}

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Sanlun.pm

=head1 Authors

=begin html 

Mike Latif

=end html

=head1 Description

Sanlun provides an API interface to the Sanlun utility.  This package should  
be instantiated through the SanClient package create_sanlun() method. 

=head1 See Also

Package inherits from the following modules. See these modules for further 
interface detail.

B<Error.pm>

=head1 Attributes

=head2 B<version>

Sanlun Version (float e.g. "3.0")

=head2 B<build_number>

Sanlun Build Number (int)

=head1 Public Methods

=head2 B<display_adapters(%args)>

Display adapter information for a particular protocol.

=over 4

=item Inputs

=item * -protocol (default: fcp-(solaris, hpux, aix, windows)  iscsi-(hpux, aix, linux, windows))

Device protocol

=item Outputs

Returns an array or array reference if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving adapter  information");

    $Test->log($sanlunObj->display_adapters(-protocol=>'iscsi'));
    if ($sanlunObj->error){
        $Test->log($sanlunObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<get_lun_path(%args)>

Retrieves the filer lun path for a host device. 

=over 4

=item Inputs

=item * -device 

Device on the host

=item Outputs

Returns a scalar value containing the full lun path for a host device.
If an error occurred, the error flag is set, errmsg is set to the failure reason, and
an empty string "" is passed back as the return value.      

=item Examples

 #Get lun path on solaris host
 $path=$Sanlun->get_lun_path(-device=>'c3t1d5');
 if ($Sanlun->error){
    $Test->nlog($Sanlun->errmsg());
    return 0;
 }

 print "Device 'c3t1d5' maps to lun path '$path'";


=back

=head2 B<lun_show()>

Retrieve a list of all lun information 

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of lun information if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and an empty array is returned.        


=item Examples

    $Test->log("-- Retrieving lun information");
    @output=$Sanlun->lun_show();
    if ($Sanlun->error){
        $Test->log($Sanlun->errmsg());
        return $TCase::FAIL;
    }  
    
    $Test->nlog(@output);


=back

=cut

1;

