# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/VSeries/Brocade_Director.pm#1 $
#
## @summary Brocade_Director - this is a child of the switch_object class
## it inherits all of that functionality and overrides where vendor/switch type
## behavior deviates from standard Brocade support in switch_object.
##
## @description
##   This *.pm inherits functionality from switch_object and overrides
##   that functionality where vendor or switch type deviates from the general
##   support for Brocade switches in switch_object.
##
## @usage
##   use VSeries::switch_object;
##   use VSeries::switch_object qw(...);
##
## @keywords  switch_object test library
##
## @dependencies
##   use strict;
##   use Tharn;
##
## @author Matt Cattel (mathew.cattel@netapp.com )
##
## @status testme
##
##

package VSeries::Brocade_Director;

use strict;
use Tharn;
use Data::Dumper;
use vars ('@ISA','@EXPORT');
require Exporter;


# This is where we create the inheritance from switch_object
use VSeries::switch_object;
our @ISA = qw(VSeries::switch_object);

# Global variables yo!
my $ERR_STATUS = 1;           # Error status is 1
my $GOOD_STATUS = 0;          # Good status is 0

# Example wwn 50:0a:09:83:00:03:4d:0b
my $WWN_REG_EX = '(\w+:){7}(\w){2}';


## @name ::switch_port_disable()
##
## @description Disable a port on a switch, needs to call
## method to translate between "index" values seen for ports as part of
## fcadmin link_state output and the actual port values that the command
## takes in.
##
## @synopsis
## my $status = $switch_obj->switch_port_disable($port);
##
## @arg $port Required: The switch port number to disable
##
sub switch_port_disable($) {
    my $self = shift;
    my($index) = @_;

    # Variables returned by say
    my ($match, $before, $after, $prompt) = "";

    # Translate the index to a slot_port value.
    my $port = $self->index_to_slot_port($index);

    my $cmd = "portdisable $port";

    ($match, $before, $after, $prompt) = Tharn::say($self->{conn}, $cmd, '-swallow');
    $after =~ s/\r//g;
    my @returned_output = split(/\n/, $after);

    if(scalar(@returned_output) > 1 || $returned_output[0] !~ m/^$cmd/){
        $self->{error} = $ERR_STATUS;
        push(@{$self->{errmsg}}, "ERROR: $after");
        return ($ERR_STATUS);
    }

    return $GOOD_STATUS;

}


## @name ::switch_port_enable()
##
## @description Enable port on switch, needs to call
## method to translate between "index" values seen for ports as part of
## fcadmin link_state output and the actual port values that the command
## takes in.
##
## @synopsis
## my $status = $switch_obj->switch_port_enable($port)
##
## @arg $port Required: The switch port number to enable
##
sub switch_port_enable($) {
    my $self = shift;
    my($index) = @_;

    # Variables returned by say
    my ($match, $before, $after, $prompt) = "";

    # Translate the index to a slot_port value.
    my $port = $self->index_to_slot_port($index);

    my $cmd = "portenable $port";

    ($match, $before, $after, $prompt) = Tharn::say($self->{conn}, $cmd, '-swallow');
    $after =~ s/\r//g;
    my @returned_output = split(/\n/, $after);


    if(scalar(@returned_output) > 1 || $returned_output[0] !~ m/^$cmd/){
        $self->{error} = $ERR_STATUS;
        push(@{$self->{errmsg}}, "ERROR: $after");
        return ($ERR_STATUS);
    }

    return $GOOD_STATUS;

}



## @name ::index_to_slot_port($)
##
## @description takes in an index value which for Brocade Directors use to
## to translate to actual port values used for commands like portenable
## and disable.
##
## @synopsis
## my $port_value = $switch_obj->index_to_slot_port($index);
##
## @arg $index Required: The switch index number to translate to an actual
##                       port value.
##
sub index_to_slot_port($){
    my $self = shift;
    my ($index) = @_;
    my $after = "";
    my @output = ();
    my $slot_port = "";

    # Call switchShow to get the output of all the ports with their index
    # values.
    (undef, undef, $after) = Tharn::say($self->{conn}, "switchShow", -swallow, '-timeout', 60);

    # Split each line of output into it's own array element.
    @output = split(/\r*\n\r*/, $after);
    
    # Loop through the output one line at a time, when we hit a line that's
    # equal signs we know the lines we care about are next.
    for (my $i = 0; $i < scalar(@output); $i++){

        if($output[$i] =~ m/^=+/){

            # Once we hit that line, truncate the array and break out.
            @output = @output[++$i .. $#output];
            last;
        }

    }

    # Ok now for each of these lines the first column when we split
    # on whitespace will be the index, check that and if it's not
    # what we wanted then skip ahead.
    foreach my $line (@output){
        # Strip out any leading whitespace.
        $line =~ s/^ *//;
        my @tokens = split(/ +/, $line);

        if($tokens[0] ne $index){
            next;
        }

        # This means the index matched so take the next two columns which are
        # slot and port values and concatenate them with a / and break out so
        # we can return that value.
        $slot_port = $tokens[1] . "/" . $tokens[2];
        last;

    }

    # Return slot_port value.
    return $slot_port;




}


## @name ::switch_reboot
##
## @description It is very difficult to actually reboot a brocade director
## switch so instead we disable the switch, sleep for a delayed time if
## the user specified one, otherwise it's 30 seconds and then re-enable the
## switch.  This does no verification and it's expected that the caller will
## be polling the status of the ports connected to the switch.
##
## @synopsis
## my $status = $switch_obj->switch_reboot()
##
## @arg $reboot_value Optional: Delay in seconds between disabling and then
##                              re-enabling the switch.  Default is 30s because
##                              that's how long it takes a path to be down for
##                              a filer to register the path gone and switch.
##
sub switch_reboot(;$) {
   my $self = shift;
   my ($reboot_value) = @_;
   my $no_verify = 1;
   my $reboot_delay = 30;

   if(defined($reboot_value)){ $reboot_delay = $reboot_value;}


   $self->so_switch_disable($no_verify);

   sleep($reboot_delay);

   $self->so_switch_enable($no_verify);

   return $GOOD_STATUS;


   #********This is the code to attempt to reboot both CPs on thbe switch *****
   #********And is being left in just in case we want to revisit how we   *****
   #********go about forcing a reboot on Brocade Director class switches  *****

   #my (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "reboot", '-prompt', $confirm_prompt, '-swallow', '-timeout', 30);
   #if($prompt =~ m/$confirm_prompt/){
   #    Tharn::say($self->{conn}, "y", '-noprompt', '-prompt', "closed by foreign host", '-swallow', '-timeout', 60);
   #}

   #my $conn_type = $self->{conntype};

   #$self->switch_disconnect();
   #eval{
   # $self->switch_connect($conn_type);
   #};
   #my $retry = 0;
   #while($@ =~ m/Unexpected EOF/ && $retry < 4){
   #    Tharn::logcomment("Connected refused, retrying to log into switch");
   #    eval{
   #        $self->switch_connect($conn_type);
   #    };
   #    $retry++;
   #    sleep(2);
   #}

   #(undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "reboot", '-prompt', $confirm_prompt, '-swallow', '-timeout', 30);

   #if($prompt =~ m/$confirm_prompt/){
   #    Tharn::say($self->{conn}, "y", '-blind', '-swallow', '-timeout', 30);
   #}

   #return $GOOD_STATUS;
}


## @name ::get_port_wwn_seen_on($)
##
## @description takes in a wwn and finds the port that wwn is connected to.
##              This is very useful in going between soft and hard zoning.
##              This method has been overloaded for the Brocade Director
##              type switches so that it actually returns the "index" value
##              in order to be consistent since all Brocade Director methods
##              expect and index value and translate that to a blade and port
##              only when needing to send a command to the switch itself.
##
## @synopsis
## my $index = $switch_obj->get_port_wwn_seen_on($wwn)
##
## @arg $wwn Required: a wwn the caller wants to know what port on the switch
##                     its seen on.
##
sub get_port_wwn_seen_on($){

    my $self = shift;
    my($wwn) = @_;

    # Captures say output.
    my $after = "";
    
    # Array of single lines of output.
    my @info = ();

    # get host type and determine proper method
    my $host_type = $self->{'hosttype'};

    # Value to be returned to the caller.
    my $port_seen_on = "";


    # vgbrdcxs35:admin> switchShow
    # switchName:     vgbrdcxs35
    # switchType:     62.3
    # switchState:    Online   
    # switchMode:     Native
    # switchRole:     Principal
    # switchDomain:   1
    # switchId:       fffc01
    # switchWwn:      10:00:00:05:1e:46:44:00
    # zoning:         ON (vseries)
    # switchBeacon:   OFF
    # FC Router:      OFF
    # FC Router BB Fabric ID: 1
    # 
    # Index Slot Port Address Media Speed State     Proto
    # ===================================================
    #  48    4    0   013000   id    N4   No_Light    FC  
    #  49    4    1   013100   id    N4   No_Light    FC  
    #  50    4    2   013200   id    N4   No_Light    FC  
    #  51    4    3   013300   id    N4   Online      FC  F-Port  50:0a:09:83:00:03:4d:0b 
    #  52    4    4   013400   id    N4   Online      FC  F-Port  50:0a:09:83:00:03:4d:24 
    #  53    4    5   013500   id    N4   Online      FC  F-Port  50:0a:09:83:00:03:4d:2e 
    #  54    4    6   013600   id    N4   Online      FC  F-Port  50:0a:09:83:00:03:4d:11 

    (undef, undef, $after) = Tharn::say($self->{conn}, "switchShow", -swallow, '-timeout', 60);
        
    # Split into single lines.
    @info = split(/\r*\n/, $after);

    # Same technique.  Match on the wwn regex, make sure it's not the
    # switch wwn, but then based on which version of firmware the switch
    # is running stripping out the port value is done differently.  Using
    # regexes to handle those cases.
    foreach my $line (@info)  {

        $line =~ s/\r//g;
            
        # Parse for port information lines 
        if ( $line =~ /($WWN_REG_EX)/ ) {            
            if($line =~ m/^switchWwn\:/){ next;} 
            elsif($1 =~ m/$wwn/ && $line =~ m/^\s*(\d+)\s+\d+/){ 
                $port_seen_on = $1; 
                last;                                              
            } 
        } 



        
    }
            
    return $port_seen_on;
}





1;  # Needed for *.pm's