# $Id: $
# -*- perl -*-
#
## @summary Additional VSeries functionality using NACL libraries
##
## @description
##  These test functions are needed for the VSeries test suites for testing
##  with NACL libraries. Most of the subroutines are not object oriented,
##  as they are meant to provide additional functionality using NACL libraries.
##
## @usage
##   use VSeries::Clustered::nacl_convenience
##   use VSeries::Clustered::nacl_convenience qw(...);
##
## @keywords  util test library
##
## @dependencies
## This pm is not object oriented because it is intended to augment NACL libs.
## Most methods will require a NACL API Set Object be
## passed in.
##
## @author arvind.chandrasekaran@netapp.com
##
## @status testme
##
##

package VSeries::Clustered::nacl_convenience;

use strict;
use Tharn;
use Getopt::Long;
use Data::Dumper;
use VSeries::utils;
use NACL::APISet;
use NACL::Transit;
use NACL::CS::StorageFailover;
use NACL::CS::StorageArrayConfig;
use NACL::C::License;
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::APISet::Exceptions::ResponseException qw(:try);
use Tharn qw($Log);
use NACL::C::Node;
use NACL::STask::Cluster;

use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT = ('check_initiator_target_pair', 'wait_aggr_online', 'execute_api_set', 'wait_for_prompt',
           'setup_api_set_obj', 'setup_transit_obj', 'execute_cmd', 'setup_io_components',
           'delete_io_components','compare_ZAPI_CLI_output', 'ems_event_search', 'add_license',
           'get_active_inactive_config', 'create_4path_config');


# Global variables for testing.
my $GOOD_STATUS = 0;
my $ERR_STATUS = 1;
my $WWN_REG_EX = '(\w+:){7}(\w){2}';


## @name ::check_initiator_target_pair($$$$)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## Checks the sacs output to ensure that a previously inactive initiator
## is now active Or vice versa.
##
## This subroutine has been copied over from vser_lunquiesce.thpl script
## and modified to work with NACL libraries.
##
## @arg $apiset_obj         Required: NACL API Set Object
##
## @arg $node               Required: filer hostname
##
## @arg $initiator_to_check Required: The initiator to be verified
##
## @arg $target_to_check    Required: The target mapped to the above initiator to be verified
##
## @arg $should_find        Required: Should we expect to find this initiator or not in the sacs output
##
## @arg $max_retry          Optional: Maximum retries allowed
##                                    Default: 4
##
## @synopsis
## my $status = &VSeries::nacl_convenience::check_initiator_target_pair($apiset_obj, $node, $initiator_to_check, target_to_check, $should_find, $max_retry);
##
sub check_initiator_target_pair ($$$$$;$) {
    my $apiset_obj = shift;
    my $node = shift;
    my $initiator_to_check = shift;
    my $target_to_check = shift;
    $target_to_check =~ s/\://g;
    my $should_find = shift;
    my $max_retry = shift;

    if (!defined $max_retry) {
        $max_retry = 4;
    }

    my $retries = 0;

    my $sacs_ref = '';
    my $found_flag = 0;

    # Retry for $max_retry times
    while ($retries < $max_retry) {
        sleep 15;
        # Get the sacs output
        my @local_sacs_output = ();
        # Execute storare array config show for the node $node
        try {
            my $response = $apiset_obj->storage_array_config_show('node' => $node);
            Tharn::logcomment("Parsed output:". Dumper($response->get_parsed_output()));

            my $output_ref = $response->get_parsed_output();
            @local_sacs_output = @{$output_ref};
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            Tharn::logcomment("Error is:". $exception_object->text());
            my $exp_response = $exception_object->get_response_object();
            Tharn::logcomment("Raw output was:\n". $exp_response->get_raw_output());
            return $ERR_STATUS;
        };

        $found_flag = 0;

        # Iterate over the sacs output and check if the initator-target pair is found or not as per $should_find
        foreach my $sacs_hash_ref (@local_sacs_output) {
            my %sacs = %$sacs_hash_ref;
            my $initiator = $sacs{'initiator'};
            my $target = $sacs{'target-wwpn'};
            # If should_find is true then check for the initiator and return GOOD_STATUS if found.
            if ($should_find eq "true") {
                if ($initiator eq $initiator_to_check && $target eq $target_to_check) {
                    return $GOOD_STATUS;
                }
            }
            # Else we should not find  this initiator based on the current switch config.
            else {
                if ($initiator eq $initiator_to_check && $target eq $target_to_check) {
                    $found_flag = 1;
                }
            }
        }
        if ($should_find eq "false" && $found_flag == 0) {
            return $GOOD_STATUS;
        }
        $retries++;
    }

    return $ERR_STATUS;
}

## @name ::wait_aggr_online($;$)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## Wait for the aggregate/volume to come online after creation.
## Returns $ERR_STATUS, if aggr does not come online in $timeout seconds
## Returns $GOOD_STATUS, if aggr comes online in $timeout seconds.
##
## @arg   $component_object     Required: Aggregate/Volume NACL component object.
##
## @arg   $timeout  Optional: Timeout for aggregate to come online.
##                            Default 120 seconds.
##
## @synopsis
## my $status = &wait_aggr_online($;$);
##
sub wait_aggr_online ($;$) {
    my $component_object = shift;
    my $timeout = shift;
 
    if (!defined $timeout) {
        $timeout = 120;
    }

    my $sec = 0;
    while ($component_object->state()->state() ne "online") {
        sleep 15;
        $sec+=15;
        if ($sec > $timeout) {
            Tharn::logcomment("The aggregate/volume has not come online after $timeout seconds");
            return $ERR_STATUS;
        }
    }
    return $GOOD_STATUS;
}

## @name ::execute_api_set($$$;$$$)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## This sub routine executes a NACL API Set and returns the output,
## if there are no exceptions. If the call fails, then the 
## exception object is returned.
##
## @arg   $api_set_object   Required: NACL API Set Object
##
## @arg   $api              Required: API to be executed
##
## @arg   $options_ref      Required: Hash reference with options for $api
##
## @arg   $exception        Optional: Exception to be handled
##                                Default is NACL::APISet::Exceptions::ResponseException
##
## @arg   $timeout          Optional: NACL Timeout for the API call.
##
## @arg    $raw             Optional: When defined, this argument forces to send 
##                          output from 'get_raw_output()' command.
##
## @synopsis
## my ($status, $output_ref)  = &execute_api_set($api_obj, $api, $options_ref);
##
sub execute_api_set ($$$;$$$) {
    my ($api_set_object, $api, $options_ref, $exception, $timeout, $raw) = @_;

    if (!defined $exception) {
        $exception = "NACL::APISet::Exceptions::ResponseException";
    }

    my $current_timeout = undef;
    if (defined $timeout) {
        $current_timeout = $api_set_object->get_timeout();
        $api_set_object->set_timeout("connectrec-timeout" => $timeout);
    }

    my %options = %{$options_ref};

    my $output_ref = undef;
    my $exception_object = undef;
    my @return_value = ();
    try {
        my $response = $api_set_object->$api(%options);

        # When we call a transit_object with the reboot api it will
        # fail by throwing an error saying that the parsed_output method 
        # is not a valid one for transit objects.
        if ( defined( $raw ) ){
            $output_ref = $response->get_raw_output();
        } elsif ( $api eq "execute_raw_command" ) {
            if ( ref $response ){
                $output_ref = $response->get_raw_output();
            } else {
                $output_ref = $response;
            }
        } elsif ($response ne q{} && $api ne "reboot") {
            $output_ref = $response->get_parsed_output();
        }

        @return_value = ($GOOD_STATUS, $output_ref);
    }
    catch $exception with {
        $exception_object = shift;
        @return_value = ($ERR_STATUS, $exception_object);
    };

    if (defined $timeout) {
        $api_set_object->set_timeout("connectrec-timeout" => $current_timeout);
    }

    return \@return_value;
}

## @name ::wait_for_prompt($$;$$)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## This subroutine uses NACL transit objects to wait for the node
## to enter a specified state.
## Returns \($GOOD_STATUS) if successful and returns \($ERR_STATUS, $exception_object)
## on failure.
##
## @arg   $transit_object   Required: Transit Object
##
## @arg   $expected_prompt  Required: Expected Prompt. Can be a regular expression.
##
## @arg   $timeout          Optional: Timeout in seconds to wait for the prompt.
##
## @arg   $append_output    Optional: If defined, we will get the console_output.
##
## @synopsis
## my $status  = &wait_for_prompt($transit_object, $expected_prompt, $timeout);
##
sub wait_for_prompt ($$;$$) {
    my ($transit_object, $expected_prompt, $timeout, $append_output) = @_;

    if (!defined $timeout) {
        $timeout = 900;
    }

    my $max_timeout = time() + $timeout;
    my @return_value = ($ERR_STATUS);
    while (time() < $max_timeout) {
        my ($state, $console_output) = (undef, undef);
        try {
            if (defined $append_output) {
                $state = $transit_object->get_state(append_output_to => \$console_output,
                                                    timeout => $timeout);
            } else {
                $state = $transit_object->get_state();
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            @return_value = ($ERR_STATUS, $exception_object);
            last;
        };
        if ($state =~ /$expected_prompt/) {
        # We have reached the expected prompt
            @return_value = ($GOOD_STATUS, $console_output);
            last;
        }
        sleep 5;
    }

    return \@return_value;
}

## @name ::setup_api_set_obj($;$$)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## Routine to create the API Set Objects for a list of filers.
## Returns a ref to array with API Set objects on success. Else,
## returns an array ref with $ERR_STATUS and the error message.
##
## @arg   $filer_list_ref   Required: Ref to list of arrays
##
## @arg   $mode             Optional: Mode - 'CMode' or '7Mode' or
##                                           'Nodescope' or 'Systemshell' or
##                                           'Maintenance' or SP.
##                                    Default is 'CMode'
##
## @arg   $interface	    Optional: Interface - 'CLI' or 'ZAPI'
##				      Default is 'CLI'
##
## @synopsis
## my $APISet_Array_Ref = &VSeries::Clustered::nacl_convenience::setup_api_set_obj($filer_list_ref, $mode, $interface);
##
sub setup_api_set_obj($;$$) {
    my ($filer_list_ref, $mode, $interface) = @_;
    if (!defined $mode) {
        $mode = "CMode";
    }
    
    $interface = "CLI" if (!defined $interface);

    my $cdefs_from_build = 0;
    $cdefs_from_build = 1 if ($mode eq "CMode");
    
    my @filers = @{$filer_list_ref};

    my @api_set_obj = ();

    my $shell_type = $mode;
    $shell_type = "CMode" if ($mode =~ m/Nodescope/i);

    foreach my $node (@filers) {

        my $hostrec_object = Tharn::host($node);
        my $user = $hostrec_object->default_username();
        my $password = $hostrec_object->default_password();

        # Create the C-Mode CLI Set Object
        my $obj = undef;
        eval {
           if ($mode =~ m/(Systemshell|Maintenance)/) {
              #
              # NACL retrieves the username and password from tharnhost file and so
              # we did not provide those values.
              #
              $obj = NACL::APISet->new(hostobj => $hostrec_object,
                                       category => "Node",
                                       interface => "$interface",
                                       set => "$mode"
                                      );
           } elsif ($mode eq 'SP') {
              my $conn_rec = Tharn::connect($node . '#rlm;conntype=ssh-persist' )
                                 or NACL::MTask::MCC::Exceptions::BaseException->throw(
                                        'Failed to get RLM details from NATE Host file!');

              $obj = NACL::APISet->new(connection => $conn_rec,
                                       category   => 'Node',
                                       interface  => 'CLI',
                                       set        => 'SP'
                                      );
           } else {
              $obj = NACL::APISet->new(hostobj => $hostrec_object,
                                       category => "Node",
                                       interface => "$interface",
                                       set => "$mode",
                                       shell_type => "$shell_type",
                                       username => $user,
                                       password => $password
                                      );
           }
        };
        if ($@) {
           return [$ERR_STATUS, $@];
        }
        push @api_set_obj, $obj;
    }
    return \@api_set_obj;
}

## @name ::setup_transit_obj($)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## Routine to create the Transit Objects for a list of filers.
## Returns a ref to array with transit objects on success. Else,
## returns an array ref with $ERR_STATUS and the error message.
##
## @arg   $filer_list_ref   Required: Ref to list of arrays
##
## @synopsis
## my $Transit_Obj_Array_Ref = &VSeries::Clustered::nacl_convenience::setup_transit_obj($filer_list_ref);
##
sub setup_transit_obj($) {
    my $filer_list_ref = shift;
    my @filers = @{$filer_list_ref};

    my @transit_obj_array = ();
    my $trans_obj = undef;
    foreach my $node (@filers) {
        my $hostrec_object    = Tharn::host($node);
        eval {
            $trans_obj = NACL::Transit->new(name => $hostrec_object);
        };
        if ($@) {
            return [$ERR_STATUS, $@];
        }
        push @transit_obj_array, $trans_obj;
    }
    return \@transit_obj_array;
}

## @name ::execute_cmd($$$;$$$)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## Subroutine to execute NACL APIs. Executes the specified API
## and returns the parsed output. If the API call fails, then
## returns $ERR_STATUS.
## This subroutine acts as a wrapper around the execute_api_set
## library routine nacl_convenience.pm
##
## @arg    $api_set_obj     Required: API Set Object
##
## @arg    $api             Required: API to execute
##
## @arg    $options_ref     Required: Ref to Hash with options for the API
##
## @arg    $exception       Optional: Exception to be handled.
##
## @arg    $timeout         Optional: Timeout to be set for the API execution.
##
## @arg    $raw             Optional: When defined, this argument forces to send 
##                          output from 'get_raw_output()' command.
##
## @synopsis
## my $output_ref = &VSeries::Clustered::nacl_convenience::execute_cmd($api_set_obj, $api, $options_ref};
##
sub execute_cmd($$$;$$$) {
    my ($api_set_obj, $api, $options_ref, $exception, $timeout, $raw) = @_;

    my @params = ($api_set_obj, $api, $options_ref);

    if (!defined $exception) {
        $exception = "NACL::APISet::Exceptions::ResponseException";
    }
    push @params, $exception;

    if (defined $timeout) {
        push @params, $timeout;
    } else {
        push @params, undef;
    }

    push @params, $raw if (defined $raw);

    my $output_ref = &execute_api_set(@params);
    my @output = @{$output_ref};

    if ($output[0] eq $GOOD_STATUS) {
        if (defined $raw) {
           $Log->debug("Raw output:". Dumper($output[1]));
        } else {
           $Log->debug("Parsed output:". Dumper($output[1]));
        }
        return $output[1];
    }
    elsif ($exception eq "NACL::Transit::Exceptions::TransitException") {
        $Log->debug("TransitException occurred. Check filer reboot status.");
        return $ERR_STATUS;
    } else {
        my $exception_object = $output[1];
        $Log->debug("Error is:". $exception_object->text());
        my $exp_response = $exception_object->get_response_object();
        $Log->debug("Raw output was:\n". $exp_response->get_raw_output());
        return $ERR_STATUS;
    }
}

## @name ::setup_io_components($)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## This method handles setting up network ip components necessary to
## run IO to a volume.  It presumes that the vserver and other components
## already exist.  
## This routine takes in a reference to a  hash. And the hash
## has details such as APISetobj, vserver, lif name, node name, io interface,
## private ip, ip mask and ip gateway.
## These details are used to create the various IO components.
## This routine calls the execute_cmd routine in this library.
## Returns either $GOOD_STATUS or $ERR_STATUS.
##
## @arg    $io_comp_hash_ref     Required: Reference to hash with details of 
##                                         IO components.
##   Sample Hash:
##   my %io_comp_hash = ('api_set_obj' => $obj,
##                       'node' => $FILER,
##                       'vserver' => $vserver,
##                       'lif' => $lif,
##                       'io_interface' => $io_interface,
##                       'private_ip' => $private_ip,
##                       'ip_mask' => $ip_mask,
##                       'gateway' => $gateway);
## @synopsis
## my $status = &VSeries::Clustered::nacl_convenience::setup_io_components(\%io_comp_hash);
##
sub setup_io_components($) {
    my $hash_ref = shift;
    my %io_comp_hash = %{$hash_ref};

    my $cli_obj = $io_comp_hash{'api_set_obj'};

    # Create a network interface
    my $api = "network_interface_create";
    my $options = {'vserver'   => $io_comp_hash{'vserver'},
            'lif'       => $io_comp_hash{'lif'},
            'role'      => "data",
            'home-node' => $io_comp_hash{'node'},
            'home-port' => $io_comp_hash{'io_interface'},
            'address'   => $io_comp_hash{'private_ip'},
            'netmask'   => $io_comp_hash{'ip_mask'}
            };
    my $nic = &execute_cmd($cli_obj, $api, $options);
    if ($nic == $ERR_STATUS) {
        $Log->debug("network interface create failed");
        return $ERR_STATUS;
    }

    # Get the routing group value
    $api = "network_routing_groups_show";
    $options = {'vserver' => $io_comp_hash{'vserver'}};
    my $nrgs_ref = &execute_cmd($cli_obj, $api, $options);
    if ($nrgs_ref == $ERR_STATUS) {
        $Log->debug("net routing group show failed");
        return $ERR_STATUS;
    }

    my @nrgs_output = @{$nrgs_ref};
    my $routing_group = $nrgs_output[0]->{'routing-group'};

    # Create network routing groups route
    $api = "network_routing_groups_route_create";
    $options = {'server'        => $io_comp_hash{'vserver'},
            'routing-group' => $routing_group,
            'gateway'       => $io_comp_hash{'gateway'}};
    my $nrgrc = &execute_cmd($cli_obj, $api, $options);
    if ($nrgrc == $ERR_STATUS) {
        $Log->debug("network routing groups route create failed");
        return $ERR_STATUS;
    }

    # Create vserver export policy
    $api = "vserver_export_policy_rule_create";
    $options = {'vserver'     => $io_comp_hash{'vserver'},
            'policyname'  => 'default',
            'clientmatch' => "0.0.0.0/0",
            'rorule'      => 'any',
            'rwrule'      => 'any',
            'allow-suid'  => 'false',
            'allow-dev'   => 'true',
            'superuser'   => 'any',
            'anon'        => "0"};
    my $veprc = &execute_cmd($cli_obj, $api, $options);
    if ($veprc == $ERR_STATUS) {
        $Log->debug("vserver export policy rule create failed");
        return $ERR_STATUS;
    }

    # Create the vserver nfs
    $api = "vserver_nfs_create";
    $options = {'vserver' => $io_comp_hash{'vserver'},
            'access'  => 'true',
	    'v3'      => 'enabled',
	    # 'v2'      => 'enabled',
            'udp'     => 'enabled',
            'tcp'     => 'enabled'};
    my $vnc = &execute_cmd($cli_obj, $api, $options);
    if ($vnc == $ERR_STATUS) {
        $Log->debug("vserver nfs create failed");
        return $ERR_STATUS;
    }
    return $GOOD_STATUS;
}

## @name ::delete_io_components($)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## Method to conveniently clean up what was created by calling
## setup_io_components.
## This routine takes in a reference to a  hash. And the hash
## has details such as APISetobj, vserver, lif name
## These details are used to delete the various IO components.
## This routine calls the execute_cmd routine in this library.
## Returns either $GOOD_STATUS or $ERR_STATUS.
##
## @arg    $io_comp_hash_ref     Required: Reference to hash with details of
##                                         IO components.
##   Sample Hash:
##   my %io_comp_hash = ('api_set_obj' => $obj,
##                       'vserver' => $vserver,
##                       'lif' => $lif);
##
## @synopsis
## my $status = &VSeries::Clustered::nacl_convenience::delete_io_components(\%io_comp_hash);
##
sub delete_io_components($) {
    my $hash_ref = shift;
    my %io_comp_hash = %{$hash_ref};

    my $cli_obj = $io_comp_hash{'api_set_obj'};

    # Delete the net interface
    my $api = "network_interface_delete";
    my $options = {'server' => $io_comp_hash{'vserver'},
            'lif'    => $io_comp_hash{'lif'}};
    my $nid = &execute_cmd($cli_obj, $api, $options);
    if ($nid == $ERR_STATUS) {
        $Log->debug("network interface delete failed");
        return $ERR_STATUS;
    }

    # Delete the net routing groups route
    $api = "network_routing_groups_route_delete";
    $options = {'server'        => $io_comp_hash{'vserver'},
            'routing-group' => "*"};
    my $nrgrd = &execute_cmd($cli_obj, $api, $options);
    if ($nrgrd == $ERR_STATUS) {
        $Log->debug("network routing groups route delete failed");
        return $ERR_STATUS;
    }

    # Delete the net routing grps
    $api = "network_routing_groups_delete";
    $options = {'server'        => $io_comp_hash{'vserver'},
            'routing-group' => "*"};
    my $nrgd = &execute_cmd($cli_obj, $api, $options);
    if ($nrgd == $ERR_STATUS) {
        $Log->debug("network routing groups delete failed");
        return $ERR_STATUS;
    }

    # Delete the vserver export policy rule
    $api = "vserver_export_policy_rule_delete";
    $options = {'vserver'    => $io_comp_hash{'vserver'},
            'policyname' => "*"};
    my $veprd = &execute_cmd($cli_obj, $api, $options);
    if ($veprd == $ERR_STATUS) {
        $Log->debug("vserver export policy delete failed");
        return $ERR_STATUS;
    }

    # Delete the vserver nfs
    $api = "vserver_nfs_delete";
    $options = {'vserver' => $io_comp_hash{'vserver'}};
    my $vnd = &execute_cmd($cli_obj, $api, $options);
    if ($vnd == $ERR_STATUS) {
        $Log->debug("vserver nfs delete failed");
        return $ERR_STATUS;
    }
    return $GOOD_STATUS;
}

## @name ::compare_ZAPI_CLI_output(@@%)
## copied over from vser_clustered_zapi_test.thpl
## @author Michelle Han (xmhan@netapp.com)
##
## @description
## This subroutine compares the corresponding data between two data structures. The first
## holds the output from the zapi call and the second contains the output from the equivalent
## cli command. Since the tag/key names could be different, a hash that maps the Zapi tags to
## CLI keys is needed to do the compare.
##
## @arg $zapi_output_ref Required: reference of ZAPI output data structure
##      $cli_output_ref Required: reference of CLI output data structure
##      $map_key_ref Required: ZAPI and corresponding CLI tag/key map reference
##
## @synopsis
## my $ret = &compare_ZAPI_CLI_output($zapi_output_ref, $cli_output_ref, $map_key_ref);
##
## $ret = $ERR_STATUS if there are difference between ZAPI and CLI output
## $ret = $GOOD_STATUS if there are no difference
##
sub compare_ZAPI_CLI_output(@@%){

    my $zapi_out_ref = shift;
    my $cli_out_ref = shift;
    my $key_map_ref = shift;

    my @zapi_out_ref = @{$zapi_out_ref};
    my @cli_out_ref = @{$cli_out_ref};
    my %key_map_ref = %{$key_map_ref};
    my $length = scalar(@zapi_out_ref);
    Tharn::logcomment("length or entries of ZAPI output array is $length\n");

    # loop through all entries (array elements)
    for (my $i=0; $i<$length; $i++) {

        # Compare the value for the same Key in zapi output hash and CLI output hash
        # However, if the key are different for zapi and CLI, use mapped Key
        foreach my $key (keys %{$zapi_out_ref[$i]}) {

            # For array-list-info and array-port-list-info, zapi displays array-id. However, CLI
            # doesn't display array-id any more. So skip comparing this key/value
            # In addition, don't compare io count and io percentage which change dynamically
            next if ($key eq 'array-id' || $key =~ /io/ || $key eq 'port-failover-type');

            # check if the value for $key in zapi output doesn't exist
            if (not ref $zapi_out_ref[$i]->{$key}) {

                # check if the tag/key name are different from ZAPI and CLI
                if (defined $cli_out_ref[$i]->{$key}) {

                    Tharn::logcomment("The value from ZAPI for $key is $zapi_out_ref[$i]->{$key}\n");
                    Tharn::logcomment("The value from CLI for $key is $cli_out_ref[$i]->{$key}\n");
                    if ($zapi_out_ref[$i]->{$key} ne $cli_out_ref[$i]->{$key}) {
                        Tharn::logresult('Error',"the value from zapi and CLI for $key are different \n");
                        return $ERR_STATUS;
                    }

                } else {
                    # compare with cli use mapped key value
                    my $value = $key_map_ref->{$key};

                    Tharn::logcomment("The value from ZAPI for $key is $zapi_out_ref[$i]->{$key}\n");
                    Tharn::logcomment("The value from CLI for $value is $cli_out_ref[$i]->{$value}\n");
                    if ($value eq "errors") {
                        ## the cli output have " " around the value
                        $cli_out_ref[$i]->{$value}=~ s/\"//g;
                    }

                    if ($zapi_out_ref[$i]->{$key} ne $cli_out_ref[$i]->{$value}) {
                        Tharn::logresult('Error',"the value from zapi for $key is different from value from cli for $value\n");
                        return $ERR_STATUS;
                    }

               } # end of if (defined ...)

           } # end of if(not ref ...)

        } # end of for loop for the keys

    } # end of for loop for $i

    return $GOOD_STATUS;

}

## @name ::ems_event_search($$$;$$)
##
## @author Mark Bozen (mbozen@netapp.com)
##
## @description
## Runs event log show command to detect logging of specific EMS events.
## Takes in a hash of EMS event keys to look for and returns the hash
## with updated values consisting of number of occurences of that event
## within the specificed time range
##
## @synopsis
## $ref = VSeries::Clustered::nacl_convenience::ems_event_search($api_set_obj, $event_href, $start_time, $interval, $node);
##
## @arg $api_set_obj  Required: API Set object
## @arg $event_href   Required: hash ref where keys are the event strings that
##                              will be searched in event logger.
##                              key: event message (ie.  cf.fm.partner)
##                              value: 0 (will be updated by this method)
## @arg $start_time   Required: start time timestamp that we begin looking for
##                              logged events (format is the dblade date
##                              command: Tue Jun 23 18:25:31 GMT 2009)
## @arg $offset  Optional: offset added to $start_time to determine end time to
##                         search for for events.
##                         (default: infinite, all ems events found on or
##                          after the $start_time are considered within
##                          time range)
## @arg $node    Optional: If the node value is passed then the event_log_show
##                         will list the events on that particular node. Otherwise
##                         it will be set to local.
sub ems_event_search($$$;$$) {
    my ($api_set_obj, $event_href, $start_time, $offset, $node) = @_;

    my %events = %$event_href;
    my @events_found = ();
    my $end_time = "";
    my %events_hash = ();

    # If the node is not mentioned then the value will be set to 'local'.
    $node = "local" if ($node eq q{});

    #
    # As we are hitting the time synchronization issue consistently with
    # the event_log_show command, we are using the NACL code to perform
    # the synchronization and this fix is wrt the BURT 715651.
    #
    my @nodes = NACL::C::Node->find();
    NACL::STask::Cluster->sync_clocks(command_interface => $nodes[0]);

    # For date conversion
    my @month_vals = qw / filler Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec /;

    # First let's calculate our end time to search for the EMS events
    # only if offset is defined
    if ( defined $offset ) {
        $end_time = VSeries::utils::date_offset($start_time, $offset);
    }

    # Strip out any leading whitespaces from $start_time string before we split
    $start_time =~ s/^ +//;

    # Extract the timezone in the $start_time. We'll need it later for
    # date comparison
    my $time_zone = (split (/ +/, $start_time))[4];

    # Find all occurences of specified events that have occured in the current
    # EMS log
    foreach my $event (sort keys %events) {
        my $output_ref = &execute_cmd($api_set_obj,"event_log_show",{'node' => $node, 'messagename' => $event});
        return $ERR_STATUS if (ref($output_ref) ne "ARRAY");

        @events_found = @{$output_ref};

        # We have gathered all the events found, now determine if they are
        # within the specified time range
        foreach my $ref (@events_found) {
            my %event_occurences = %$ref;

            # Get the time of the event, strip out the timestamp and covert it
            # to unix date command output format
            # Sample output of event log show for time: "6/23/2009 21:00:00"

            my $occurence_day = (split (/ +/, $event_occurences{'time'}))[0];
            my $occurence_time = (split (/ +/, $event_occurences{'time'}))[1];

            # Remove leading " from $occurence_day
            $occurence_day =~ s/^"//;
            # Remove trailing " from $occurence_time
            chop $occurence_time;

            # Split occurence day into month, day, year
            my $month_num  = (split ("/", $occurence_day))[0];
            my $day = (split ("/", $occurence_day))[1];
            my $year = (split ("/", $occurence_day))[2];

            # Convert month number to equivalent name
            my $month = $month_vals[$month_num];

            # Assemble our date/time stamp to send to compare_dates which must
            # be in this format: Tue Jun 23 18:25:31 GMT 2009)
            # compare_dates() does not parse name of day but a field needs to
            # be there for token positioning so we add a junk 'DAY' token to
            # handle this
            my $event_time = "DAY " . "$month " . "$day " . "$occurence_time " . "$time_zone " . "$year";

            my $date_check_begin = 0;
            my $date_check_end= 0;

            # Now check whether the event timestamp falls within the range
            $date_check_begin = VSeries::utils::compare_dates($event_time, $start_time);
            if ( defined $offset ) {
                $date_check_end = VSeries::utils::compare_dates($event_time, $end_time);
            }

            # Check if EMS event timestamp matched or came after the
            # $start_time
            if ($date_check_begin == 1 || $date_check_begin == 0) {

                # Check if the EMS event timestamp matched or came
                # before the $end_time. If no offset was defined then the event
                # is automatically considered within $end_time since we
                # count any events that occur after $start_time
                if ( (defined $offset && ($date_check_end == 2 || $date_check_end == 0)) ||
                    ! defined $offset ) {
                    $events{$event}++;
                    $events_hash{$event} = [$event_occurences{'event'}, $events{$event}];
                }
            }
        }
    }

    return \%events_hash;
}

## @name ::add_license($$$;$)
##
## @author Aditya Siddavatam (aditya1@netapp.com)
##
## @description
## Adds license(s) to the cluster.
## Returns $ERR_STATUS, if fails to add license. Otherwise
## Returns $GOOD_STATUS.
##
## @arg   $component_object  Required: NACL component object of Node.
## @arg   $feature_name      Required: Feature name of the license_code. 
## @arg   $license_code      Required: License code for the feature.
## @arg	  $interface	     Optional: Interface to be used by license add.
##
## @synopsis
## my $status = &add_license($component_object, $feature_name, $license_code, $interface)
##
sub add_license ($$$;$) {
    my ($component_object, $feature_name, $license_code, $interface) = @_;

    my ($exception, $license_present, $license_state) = (0, 0, undef);

    # If license for a particular feature is found then 'license_exists' flag will
    # be set to '1'.
    my $license_exists = 0;
    my %apiset_interface = ();
    if($interface eq q{}) {
	%apiset_interface = (apiset_must => { interface => 'CLI'});
    } else {
	%apiset_interface = (apiset_should => { interface => $interface});
    }
				    
    # 'is_feature_supported' method is used to determine whether a particular feature is
    # supported or not. Returns boolean value: '1' if the feature is supported
    # Otherwise '0'.
    # Check whether the release supports Licensing version 2.
    my $license_v2 = $component_object->is_feature_supported('feature' => 'license_v2');

    my $feature = 'feature';
    $feature = 'package' if ($license_v2);
    my %license_hash = ();
    %license_hash = ('owner' => $component_object->node(),
	            'type' => "license") if $feature_name eq "V_StorageAttach";

    if (($component_object->mode() eq 'CMode') || ($license_v2)) {
       try {
          NACL::C::License->find(
               command_interface => $component_object,
               filter            => { $feature => $feature_name,
				      %license_hash, }
          );
          Tharn::logcomment("License has already been added for the feature '$feature_name'");
          $license_exists = 1;
       }
       catch NACL::Exceptions::NoElementsFound with {
          my $exception_object = shift;

          if ($exception_object->text() !~ m/No matching system license/i) {
             Tharn::logcomment("Error is: " . $exception_object->text());
             $exception = 1;
          }
       };
       return $ERR_STATUS if ($exception == 1);
       return $GOOD_STATUS if ($license_exists == 1);
    } else {
       my @license = NACL::CS::License->fetch(command_interface => $component_object,
                                              filter => {$feature => $feature_name});

       map{$_->command_interface_reset()}@license;
       Tharn::logcomment("Dumper: ".Dumper(\@license));

       if ($license[0]->{'is_licensed'} eq 'true') {
          Tharn::logcomment("License has already been added for the feature '$feature_name'");
          return $GOOD_STATUS;
       }
    }

    try {
       NACL::C::License->add(
            command_interface => $component_object,
            'license-code' => $license_code,
            $feature => $feature_name,
	    %apiset_interface,
       );

       $license_state = NACL::C::License->find(
                             command_interface => $component_object,
                             filter            => { $feature => $feature_name }
                        );
       Tharn::logcomment("Feature of license is " . $license_state->$feature());
    }
    catch NACL::APISet::Exceptions::ResponseException with {
       my $exception_object = shift;
       Tharn::logcomment("Error is: " . $exception_object->text());

       my $resp = $exception_object->get_response_object();
       Tharn::logcomment("Raw output was:\n" . $resp->get_raw_output());

       $exception = 1;
    };

    return $ERR_STATUS if ($exception == 1);
    return $GOOD_STATUS;
}


## @name ::fcadmin_link_state_parser($)
##
## @description
## Takes in a NACL response object for the raw output when invoking a node
## scoped call for the "fcadmin link_state" command and returns a parsed
## output consistent with our NACL formatted output.  Meaning an array
## of hashes.
##
## @synopsis
## $fcadmin_ref = VSeries::Clustered::nacl_convenience::fcadmin_link_state_parser($response_obj);
##
## @arg $response_obj Required: response object with raw fcadmin link_state output.
##
##
## SAMPLE OUTPUT:
##system node run -node vra-v3050-2a -command "priv set advanced; fcadmin link_state"
##Warning: These advanced commands are potentially dangerous; use
##         them only when directed to do so by Network Appliance
##         personnel.
##Link State for adapter channel 0a: UP - N-port --> Port 0 on switch 'vrs-br200e-3' (100000051e03ba6d)
##Link State for adapter channel 0b: OFFLINE (physical)
##Link State for adapter channel 0c: UP - N-port --> Port 0 on switch 'vrs-br200e-4' (100000051e03b684)
##Link State for adapter channel 0d: OFFLINE (physical)
##
## So if you take the line with 0a as the interface the hash that gets returned
## will have key value pairs of
## adapter-name -> 0a
## link-state -> UP
## port-type -> N-port
## switch-port -> 0
## switch -> vrs-br200e-3
## port-wwn -> 100000051e03ba6d

sub fcadmin_link_state_parser($){
    my ($raw_output) = @_;
    my @return_array = ();

    # Split the raw output on new lines.
    my @temp = split(/\n/, $raw_output);
    

    # for each line of output we only care about lines that begin with
    # Link State.  From there we tokenize on whitespace and take
    # specific tokens and put them into the returned hash for a given
    # interface.
    foreach my $line (@temp) {

        if ($line =~ m/^Link State/) {
            my @tokens = split(/\s+/, $line);
            my %tmp_hash = ();

            my $current_slot = undef;
            if (defined($tokens[5])) {
                chop($tokens[5]);
                $current_slot = $tokens[5];
            }
            $tmp_hash{'adapter-name'} = $current_slot;
            if (defined($tokens[6])) {
                $tmp_hash{'link-state'} = $tokens[6];
            }
            if (defined($tokens[8])) {
                $tmp_hash{'port-type'} = $tokens[8];
            }
            if (defined($tokens[11])) {
                $tmp_hash{'switch-port'} = $tokens[11];
            }
            if (defined($tokens[14])) {
                $tmp_hash{'switch'} = substr($tokens[14], 1, -1);
            }
            if (defined($tokens[15])) {
                $tmp_hash{'port-wwn'} = substr($tokens[15], 1, -1);
            }

            push(@return_array, \%tmp_hash);

        }
    }

    return \@return_array;

}

## @name ::get_active_inactive_config($$$$$)
##
## @author Mathew Cattel (Matthew.Cattel@netapp.com)
##
## @description
## This subroutine handles taking in the information about
## connections and paths to storage specified by the user
## and makes appropriate calls to storage array config show
## and get_seen_wwpns on the switches to determine how
## all the paths are constituted at the start of the test.
## It's easy to figure out what is connected, but it's harder
## to know what things aren't connected and this method puts
## all those things together in the %Paths data structure
## which is keyed by the filer hbas.  So this structure
## doesn't expect that you can have fanout where one filer
## hba is connected to multiple sets of storage in multiple
## zones.  It returns GOOD_STATUS or ERR_STATUS to the caller.
## If the status was GOOD_STATUS it also returns the paths data 
## structure otherwise it returns an error message as well.
##
## @arg $node_component       Required: Node component to communicate with the 
##                                      filer through.
##
## @arg $filer_name           Required: Name of filer to determine topology for.
##
## @arg $switch_component_hash_ref Required: reference to a hash of switch 
##                                           components common in most test thpls
##                                           Key is the switch name and the value
##                                           is the switch component.
##
## @arg $filer_hbas Required: Comma separated list of filer hbas
##                            to be considered in this topology.
##                            "0a,0b,0c,..."
##
## @arg $array_ports Required: Comma separated list of array target
##                             ports wwpns to be considered in the
##                             topology.
##
## @synopsis
## my $paths_ref = VSeries::Clustered::nacl_convenience::get_active_inactive_config($node_component, $filer_name, \%switch_component_array, $filer_hbas, $array_ports);
##
sub get_active_inactive_config($$$$$) {

    my ($node_component, $filer_name, $switch_component_hash_ref, $filer_hbas, $array_ports) = @_;
    
    my %switch_component_hash = %$switch_component_hash_ref;  

    # Initialize data structure to be returned to caller.
    my %paths = ();
    
    # First get the storage array config show output.
    my @storage_array_config = ();
    
    @storage_array_config = NACL::CS::StorageArrayConfig->fetch(command_interface => $node_component,
                                                                    filter => {node => $filer_name});
    
    my @active_hbas = ();

    # Temp hashes that will help us fill up sub-hash for %paths
    my %array_target_port = ();
    my %switch_names = ();


    # Get output of storage array config show and put elements into
    # temporary hashes using initiator as key. This will help us
    # populate the %path_hash
    foreach my $sacs_cs_ref (@storage_array_config) {
        my $initiator = $sacs_cs_ref->{'initiator'};
        # Populate temp array with active hbas
        push @active_hbas, $initiator;

        # Get array wwpn associated with initiator and put into a hash
        $array_target_port{$initiator} = $sacs_cs_ref->{'target_wwpn'};

        # Get the switch name associated with initiator
        $switch_names{$initiator} = $sacs_cs_ref->{'target_side_switch_port'};
    }

    my @all_hbas = split (',', $filer_hbas);

    # Define our hba name keys for %paths
    foreach my $hba (@all_hbas) {

        # Define new hash key for each initiator
        $paths{$hba} = 0;
        # Initialize sub-hash for %paths
        my %path_hash = ();
        # Flag to indicate active initiator
        my $active = 0;

        # Fill in the current state of the initiator by checking if
        # it is in temp array @active_hbas
        foreach my $hba_act (@active_hbas) {
            if ( $hba_act eq $hba ) {
                $path_hash{'state'} = "active";
                $active = 1;
                last;
            }
        }
        if ( ! $active ) {
            $path_hash{'state'} = "inactive";
        }

        # Fill in the array wwpn associated with active initiator from
        # temp hash
        if ( defined $array_target_port{"$hba"} ) {
            $path_hash{'wwpn'} = $array_target_port{"$hba"};
            if($path_hash{'wwpn'} !~ /:/){
                $path_hash{'wwpn'} =~ s/(\w{2})/$1\:/g;
                chop $path_hash{'wwpn'};
            }
        }

        # Fill switch name associated with active initiator from temp hash
        if ( defined $switch_names{"$hba"} ) {
            my $switch = $switch_names{"$hba"};

            # Strip out port number from switch name
            $switch =~ s/:(.*)$//;
            my $port = $1;
            $path_hash{'switch'} = $switch;
            $path_hash{'switch-storage-port'} = $port;
        }
        
        # Use super handy feature that gives you back either the 7 mode apiset
	# or if it's a c-mode filer gives you a nodescoped apiset.
        my $node_7m_or_nodescope_API_obj = $node_component->get_7m_or_nodescope_apiset();        

        # Get our filer hba wwpn from fcadmin channels
        my $hba_wwpn_response =  $node_7m_or_nodescope_API_obj->fcadmin_channels('adapter-name' => $hba);

        my $raw_output = $hba_wwpn_response->get_raw_output();
        
        $raw_output =~ s/\n+//g;
        
        Tharn::logcomment("Raw output of fcadmin link_state:\n" . $raw_output);
        
        my $hba_wwpn = (split(/\s+/, $raw_output))[0];
                
        $hba_wwpn =~ s/(\w{2})/$1\:/g;
        chop $hba_wwpn;
        
        # Fill in wwpn of the associated initiator
        $path_hash{'hba_wwpn'} = $hba_wwpn;

        # Attach populated %path_hash to %paths
        $paths{"$hba"} = \%path_hash;
    }

    my $node_7m_or_nodescope_API_obj = $node_component->get_7m_or_nodescope_apiset();      
    my $response = $node_7m_or_nodescope_API_obj->fcadmin_link_state();
    my $raw_output = $response->get_raw_output();
  
    my @parsed_response = @{VSeries::Clustered::nacl_convenience::fcadmin_link_state_parser($raw_output)};
    
    foreach my $hash_ref (@parsed_response){
        my $hba = $hash_ref->{'adapter-name'};
            
        if(defined($paths{$hba})){
            $paths{$hba}{'switch-initiator-port'} = $hash_ref->{'switch-port'};
        }
    }


    # Now loop through and based on what paths were seen as active
    # find the inactive paths and pair up the hba_wwpns and array
    # port WWNs this will serve as a useful reference in later
    # test cases when zoning needs to be modified.
    my @array_of_ports = split(/,/, $array_ports);
  
    foreach my $array_port (@array_of_ports){
        my $found_port = 0;
        my $found_switch = undef;
        my @seen_wwpns = ();


        # If the target-wwpn is defined and matches the current 
        # array port than this wwn was already paired up on an
        # active path up above and there isn't much more 
        # needed to be done with it.
        foreach my $hba (keys %paths){
            if(defined($paths{$hba}{'wwpn'}) && $paths{$hba}{'wwpn'} eq $array_port){
                $found_port = 1;
                last;
            }
        }
  
        # If you get here it means the array port hasn't been matched
        # up yet. We'll loop through each switch object calling get
        # seen wwns and see if this array port is seen on that
        # switch.  If so then we find the inactive path connected
        # to this switch and fill in the array wwn and switch
        # value.
        if($found_port == 0){
            foreach my $key (keys(%switch_component_hash)){
                my $switch_component = $switch_component_hash{$key};
                @seen_wwpns = @{VSeries::Clustered::nacl_convenience::get_seen_wwns($switch_component)};
                foreach my $seen_wwpn (@seen_wwpns){
                    Tharn::logcomment("Matching $seen_wwpn with $array_port");
                    if($seen_wwpn =~ m/^$array_port/i){
                        $found_switch = $key;
                        last;
                    }
                }
                if($found_switch ne ""){ last;}
            }

            # If this is not blank than we matched on the current
            # switch.  Then we'll loop and find the inactive
            # path to this switch and fill in the appropriate
            # values.
            if($found_switch ne ""){
                my $resolved_information = 0;
            
                foreach my $hba (keys %paths){
                    # Skip over hba's that are active or array
                    # ports that have already been defined                    
                    if($paths{$hba}{'state'} =~ m/^active/ ||
                        defined $paths{$hba}{'wwpn'}) {
                        next;
                    }
  
                    foreach my $wwpn (@seen_wwpns){
                        if($paths{$hba}{'hba_wwpn'} =~ m/^$wwpn/i){
                            $paths{$hba}{'wwpn'} = $array_port;
                            $paths{$hba}{'wwpn'} =~ tr/A-Z/a-z/;
                            $paths{$hba}{'switch'} = $found_switch;
                            $resolved_information = 1;
                            last;
                        }
                    }
                    
                    if($resolved_information == 1){ last;}
                }
            } # end if found_switch
        } # end if found_port
    } # end foreach array_port
    
    # Ok so lastly we need to know the switch initiator ports and 
    # target initiator ports by calling get_port_wwn_seen_on for
    # both the hba_wwpn and wwpn (array wwn) for each path 
    # connected to the given switch.
    foreach my $key (keys(%switch_component_hash)){
        my $switch_component = $switch_component_hash{$key};
        foreach my $hba (keys %paths){
            if($paths{$hba}{'switch'} eq $key){
                $paths{$hba}{'switch-initiator-port'} = VSeries::Clustered::nacl_convenience::get_port_wwn_seen_on($switch_component, $paths{$hba}{'hba_wwpn'});
                $paths{$hba}{'switch-storage-port'} = VSeries::Clustered::nacl_convenience::get_port_wwn_seen_on($switch_component, $paths{$hba}{'wwpn'});
            }
        }
    }  
  
    return ($GOOD_STATUS, \%paths);
}


## @name ::get_seen_wwns($)
##
## @description 
## returns an array ref of all WWNs connected to the switch.
## ported from our switch object to work with the NACL
## switch component.  Gets the apiset and checks the type
## from there this behaves exactly as it does inside our
## switch object.
## 
## @arg switch_component Required: NACL switch object used to talk to the 
##                                 switch.
##
## @synopsis
## my $array_ref = VSeries::Clustered::nacl_convenience::get_seen_wwns($switch_component)
##
##
sub get_seen_wwns($){
    my ($switch_component) = @_;
    my $switch_apiset = $switch_component->apiset();
    
    # This is the array we'll return.
    my @seen_wwns = ();
    
    # Array of single lines of output.
    my @info = ();

    if ($switch_apiset->{'set'} =~ m/BrocadeFC/i) {



        # switchShow sample output (v5)
        #
        #switchName:     vns_b200e_8
        #switchType:     34.0
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:05:1e:04:14:73
        #zoning:         ON (Metrocluster_V6030)
        #switchBeacon:   OFF
        #
        #Area Port Media Speed State     Proto
        #=====================================
        #  0   0   id    AN   No_Sync
        #  1   1   id    N4   No_Light
        #  2   2   id    N4   No_Light
        #  3   3   id    N4   Online           F-Port  50:0a:09:83:00:01:71:9d
        #  4   4   id    N4   Online           F-Port  50:0a:09:81:00:01:71:9d
        #  5   5   id    N4   No_Light
        #  8   8   id    N2   Online           LD E-Port  10:00:00:05:1e:04:15:29 "vns_b200e_9" (upstream)
        # 21  21   id    N2   Online           F-Port  50:0a:09:82:00:00:de:9d
        # 22  22   id    N2   Online           F-Port  50:0a:09:82:00:01:91:cf
        # 23  23   --    N4   No_Module

        # switchShow sample output (v3)
        #
        #switchName:     vss_b3800_7
        #switchType:     9.2
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:60:69:51:f1:e4
        #switchBeacon:   OFF
        #Zoning:         ON (Automation1)
        #port  0: id N2 Port_Flt
        #port  1: id N1 No_Sync
        #port  2: id N2 Port_Flt
        #port  3: id N2 Online         F-Port 50:06:04:82:cb:1b:ce:2c
        #port  4: id N2 No_Light
        #port  5: id N2 No_Light
        #port  6: id N2 No_Light
        #port  7: id N2 No_Light
        #port  8: id N2 No_Light
        #port  9: id N2 No_Light
        #port 10: id N2 No_Light
        #port 11: id N2 No_Light
        #port 12: id N2 No_Light
        
        my $response = $switch_apiset->switchshow();

        
        
        # Split buffer into single lines for parsing.
        @info = split(/\r*\n/, $response->get_raw_output());

        # Same technique here match on the WWN regex, however
        # with brocade you need to be sure the line isn't
        # the wwn of the switch itself so handle that
        # case by skipping over that line.
        foreach my $line (@info)  {
            $line =~ s/\r//g;

            # Parse for port information lines
            if ( $line =~ /($WWN_REG_EX)/ ) {
                if($line =~ m/^switchWwn:/){ next;}
                push(@seen_wwns, $1);
            }
        }

        return \@seen_wwns;


    }
    elsif ($switch_apiset->{'set'} =~ m/CiscoFC/i) {
        #vnci9124s52#
        #vnci9124s52# show flogi database
        #--------------------------------------------------------------------------------
        #INTERFACE        VSAN    FCID           PORT NAME               NODE NAME
        #--------------------------------------------------------------------------------
        #fc1/3            1     0xcf0500  20:04:00:a0:b8:21:a9:fa 20:04:00:a0:b8:21:a9:f8
        #fc1/5            1     0xcf0600  50:0a:09:83:00:01:90:74 50:0a:09:83:00:01:90:74
        #fc1/6            1     0xcf0700  50:0a:09:83:00:01:91:be 50:0a:09:83:00:01:91:be

        
        my $response = $switch_apiset->show_flogi_database();

        
        # Split into single lines for parsing.
        @info = split(/\r*\n/, $response->get_raw_output());
        
        # Same technique here except each line has the port and node wwns
        # so match on both but only push the port wwn onto the seen_wwns array.
        foreach my $line (@info)  {
            # Parse for port information lines, we want the port WWN.
            if ( $line =~ m/($WWN_REG_EX)\s+($WWN_REG_EX)/ ) {
                push(@seen_wwns, $1);
            }
        }

        return \@seen_wwns;
    }
    
}

## @name ::get_port_wwn_seen_on($$)
##
## @description 
## takes in a NACL based switch component and a wwn and 
## finds the port that wwn is connected to.  This was
## ported from switch_obj to depend on the NACL switch component.
## It then pulls out the apiset, checks the type and then proceed
## to invoke and parse the command outputs just as the switch_obj
## does.
##              
##
## @synopsis
## my $port = VSeries::Clustered::nacl_convenience::get_port_wwn_seen_on($switch_component, $wwn)
##
## @arg $switch_component Required: A NACL based switch component to talk to 
##                                  the switch with.
## @arg $wwn Required: a wwn the caller wants to know what port on the switch
##                     its seen on.
##
sub get_port_wwn_seen_on($$){

    my ($switch_component, $wwn) = @_;
    my $switch_apiset = $switch_component->apiset();
    
    # Array of single lines of output.
    my @info = ();

    # Value to be returned to the caller.
    my $port_seen_on = undef;


    if ($switch_apiset->{'set'} =~ m/BrocadeFC/i) {



        # switchShow sample output (v5)
        #
        #switchName:     vns_b200e_8
        #switchType:     34.0
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:05:1e:04:14:73
        #zoning:         ON (Metrocluster_V6030)
        #switchBeacon:   OFF
        #
        #Area Port Media Speed State     Proto
        #=====================================
        #  0   0   id    AN   No_Sync
        #  1   1   id    N4   No_Light
        #  2   2   id    N4   No_Light
        #  3   3   id    N4   Online           F-Port  50:0a:09:83:00:01:71:9d
        #  4   4   id    N4   Online           F-Port  50:0a:09:81:00:01:71:9d
        #  5   5   id    N4   No_Light
        #  8   8   id    N2   Online           LD E-Port  10:00:00:05:1e:04:15:29 "vns_b200e_9" (upstream)
        # 21  21   id    N2   Online           F-Port  50:0a:09:82:00:00:de:9d
        # 22  22   id    N2   Online           F-Port  50:0a:09:82:00:01:91:cf
        # 23  23   --    N4   No_Module

        # switchShow sample output (v3)
        #
        #switchName:     vss_b3800_7
        #switchType:     9.2
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:60:69:51:f1:e4
        #switchBeacon:   OFF
        #Zoning:         ON (Automation1)
        #port  0: id N2 Port_Flt
        #port  1: id N1 No_Sync
        #port  2: id N2 Port_Flt
        #port  3: id N2 Online         F-Port 50:06:04:82:cb:1b:ce:2c
        #port  4: id N2 No_Light
        #port  5: id N2 No_Light
        #port  6: id N2 No_Light
        #port  7: id N2 No_Light
        #port  8: id N2 No_Light
        #port  9: id N2 No_Light
        #port 10: id N2 No_Light
        #port 11: id N2 No_Light
        #port 12: id N2 No_Light
        
        my $response = $switch_apiset->switchshow();
	        
	# Split buffer into single lines for parsing.
        @info = split(/\r*\n/, $response->get_raw_output());

        # Same technique.  Match on the wwn regex, make sure it's not the
        # switch wwn, but then based on which version of firmware the switch
        # is running stripping out hte port value is done differently.  Using
        # regexes to handle those cases.
        foreach my $line (@info)  {

            $line =~ s/\r//g;
            
            # Parse for port information lines
            if ( $line =~ /($WWN_REG_EX)/ ) {           
                if($line =~ m/^switchWwn\:/){ next;}
                if($1 =~ m/$wwn/){    
                    if($line =~ m/^port\s+(\d+)/){
                        $port_seen_on = $1;
                        last;
                    }
                    elsif($line =~ m/^\s*\d+\s+(\d+)/){
                        $port_seen_on = $1;
                        last;                                            
                    }    
                }
            }
        }

        return $port_seen_on;


    }
    elsif ($switch_apiset->{'set'} =~ m/CiscoFC/i) {
        #vnci9124s52#
        #vnci9124s52# show flogi database
        #--------------------------------------------------------------------------------
        #INTERFACE        VSAN    FCID           PORT NAME               NODE NAME
        #--------------------------------------------------------------------------------
        #fc1/3            1     0xcf0500  20:04:00:a0:b8:21:a9:fa 20:04:00:a0:b8:21:a9:f8
        #fc1/5            1     0xcf0600  50:0a:09:83:00:01:90:74 50:0a:09:83:00:01:90:74
        #fc1/6            1     0xcf0700  50:0a:09:83:00:01:91:be 50:0a:09:83:00:01:91:be

        my $response = $switch_apiset->show_flogi_database();

        
        # Split into single lines for parsing.
        @info = split(/\r*\n/, $response->get_raw_output());

        # Same technique.  Match on the wwn regex and then check if the wwn 
        # matches the wwn passed in.  If so then strip out the port seen on
        # and convert the slash to a dash and return.
        foreach my $line (@info)  {
        
            # Parse for port information lines, we want the port WWN.
            if ( $line =~ /($WWN_REG_EX)\s+$WWN_REG_EX/ ) {
                 if($1 =~ m/$wwn/){
                     $line =~ m/^fc(\d+\/\d+)/;
                     $port_seen_on = $1;
                     $port_seen_on =~ s/\//-/g;
                     last;
                }
            }

        }
        return $port_seen_on;
    }
}


## @name ::create_4path_conifg($$$)
##
## @description 
## This is a NACL component based routine that depends on the %Paths 
## data structure format established in the MKBY thpls to take in a
## 2 pathed config and produce a resulting fabric configuration that
## has 4 paths.  It takes in the switch components, paths data structure
## for a single filer, and the name of the zones to be created.  It will
## create a zone of the same name on both switches.
## It will create the new zones and add them to the existing config on
## the switch and return components for both the zone and zonesets
## that result.
## 
## @arg switch_component_hash_ref Required: A hash containing NACL switch 
##                                          components used to talk to the 
##                                          switches used to create the config.
##                                          Keys are expected to be the switch
##                                          names.
## @arg paths_ref                 Required: A data structure used to model the
##                                          information about which paths on a
##                                          filer are currently logically 
##                                          connected.
## @arg zone-name                 Required: Name of the zone to be created on
##                                          all the switches, presumed to be 
##                                          unique.
##
## @synopsis
## my ($zoneset_componenes_ref, $created_zones_ref) = VSeries::Clustered::nacl_convenience::create_4path_config($switch_component_hash_ref, $paths_ref, $zone_name);
##
##
sub create_4path_config($$$){

    my ($switch_component_hash_ref, $paths_ref, $zone_name) = @_;
    
    my %switch_component_hash = %$switch_component_hash_ref;
    my %paths = %$paths_ref;

    # key is switch name, value is created zone component.
    my %created_zones = ();
    
    # Key is switch name, value is active zoneset for switch.
    my %zoneset_components = ();

    # Loop over each switch component and add in an additional zone based on the 
    # 'inactive' path in the Paths data structure.
    foreach my $key (keys %switch_component_hash){
        my $switch_component = $switch_component_hash{$key};
        
        # Get the active zone name.
        my $zoneset_component = NACL::C::Switch::Zoneset->find(  command_interface => $switch_component, filter => {'active' => 1,});


                                                                                      
        
        $zoneset_components{$key} = $zoneset_component;        
        my $zoneset_component_state = $zoneset_component->state();
        
        Tharn::logcomment("Here's the members in the active cfg for switch $key " . Dumper($zoneset_component_state->{'zones'}));  

        
        # Now we'll loop over the Paths structure, find the inactive path on 
        # this switch and add that zone to the active config.  We'll keep the
        # created zone and modified zoneset around for later cleanup.
        my $created_zone_component = undef;
        my $zone_counter = 1;
        my $zone_list = undef;
        foreach my $initiator (keys(%paths)){
            my $members = undef;

            if($paths{$initiator}->{'switch'} eq $key && $paths{$initiator}->{'state'} eq 'inactive'){
                Tharn::logcomment("Creating zone $zone_name on switch $key and members " . $paths{$initiator}->{'hba_wwpn'} . " and " . $paths{$initiator}->{'wwpn'});
                $members = $paths{$initiator}->{'hba_wwpn'} . ";" . $paths{$initiator}->{'wwpn'};
                $created_zone_component = NACL::C::Switch::Zone->create(  command_interface => $switch_component,
		                                                             'vsan-id'            =>"1",
		                                                             'zone-name'          =>$zone_name,
		                                                             'member'             =>$members 
                                                                          );   
                $created_zones{$key} = $created_zone_component;   
                
                my $switch_apiset = $switch_component->apiset();
    
                if($switch_apiset->{'set'} =~ m/CiscoFC/i){   
                    NACL::C::Switch::Zone->add_wwpn( command_interface => $switch_component,
                                                     'vsan-id'            =>"1",
		                                     'zone-name'          =>$zone_name,
                                                     'wwpn' => $paths{$initiator}->{'hba_wwpn'}
                                                   );                   
                    NACL::C::Switch::Zone->add_wwpn( command_interface => $switch_component,
                                                     'vsan-id'            =>"1",
   		                                     'zone-name'          =>$zone_name,
                                                     'wwpn' => $paths{$initiator}->{'wwpn'}
                                                   );   
                }
                
                $zoneset_component->add_zone(
		                              'vsan-id'            =>"1",
		                              'zone-name'          =>$zone_name,);
		                              
		$zoneset_component->enable('vsan-id'            =>"1");   
		
		last;
            }
        }
        
        # Show the updated config to the user.
        my $updated_component_state = $zoneset_component->state();
        
        Tharn::logcomment("Here's the udpated zones in the active cfg " . Dumper($updated_component_state->{'zones'}));  
    
    }
    
    return (\%zoneset_components, \%created_zones);

}


1; # Need for perl modules
