# -*- perl -*-
# $Id$
#
## @summary VSeries library functions to be used by VSeries tests
##
## @description
##   These test functions are needed for the VSeries test suites
## @usage
##   use vseries;
##   use vseries qw(vseries_disklist_get vseries_slotlist_get);
##
## @keywords  VSeries test library
##
## @dependencies
## use config_base
## use Tharn
##
## @author dl-sar-qa@netapp.com, dl-step@netapp.com, mgorman@netapp.com
## @status testme
## @change 2004-09-13 sajan: created
## @change 2006-07-16 mgorman: changed from gfiler.pm to VSeries.pm;
##                    modified for agnostic branch
##

package VSeries::VSeries;

use strict;
use vars ('@ISA','@EXPORT');
use Time::localtime;
use Data::Dumper;

#
# Constants
#
use constant POWER_CYCLE_TYPE_POWERSTRIP => 'powerstrip';
use constant POWER_CYCLE_TYPE_RLM        => 'rlm';
use constant POWER_CYCLE_TYPE_CLI        => 'cli';
use constant POWER_CYCLE_TYPE_PANIC      => 'panic';


use constant POWER_CYCLE_ACTION_OFF      => 0;
use constant POWER_CYCLE_ACTION_ON       => 1;
use constant POWER_CYCLE_ACTION_REBOOT   => 2;

my $NTAP_DISK = '^\w{2}\.\d+';               # Regex for OnTap disks.

require Exporter;
@ISA = ('Exporter');
@EXPORT = qw(
    vseries_info_get
    vseries_hw_sw_info_get
    vseries_disklist_get
    vseries_no_redundant_path_get
    vseries_no_primary_path_get
    vseries_local_switchlist_get
    vseries_remote_switchlist_get
    vseries_storage_portlist_get
    vseries_storage_port_find
    vseries_host_portlist_get
    vseries_fclink_state_get
    vseries_fclink_state_display
    vseries_slotlist_get
    vseries_partner_name_get
    vseries_cluster_slot_info_get
    vseries_heartbeat_link_status_find
    vseries_storage_info_get
    vseries_init
    vseries_print_test_setup
    vseries_local_filer_of_switch_get
    vseries_storage_slotlist_connecting_switch_and_filer
    vseries_verify_prim_no_paths
    vseries_verify_sec_no_paths
    vseries_power_cycle_servertech
    vseries_power_cycle_rlm
    vseries_power_cycle
    vseries_object_is_up
    vseries_get_luns_for_vol
    vseries_compare_arrays
    vseries_storage_port_count
    vseries_is_storage_port
    vseries_storage_port_associate
    vseries_switch_slotlist_get
    vseries_hba_port_disable_enable
    POWER_CYCLE_TYPE_POWERSTRIP
    POWER_CYCLE_TYPE_RLM
    POWER_CYCLE_TYPE_CLI
    POWER_CYCLE_TYPE_PANIC
         );

use Expect;
use config_base;
use Tharn;
use VSeries::switch;
use VSeries::cluster;



## @name ::vseries_info_get
## @description Returns hash containing vseries information
##
## @synopsis
## vseries_info_get {$conn}
##
## @arg $conn Required: The gFiler connection to send this operation over
##
##
sub vseries_info_get ($) {
    my ($conn) = @_;

    my (%filer_info) = ();
    my (%disklist);
    my (@local_switchlist, @remote_switchlist, @slotlist);

    %disklist = vseries_disklist_get($conn);
    %filer_info = %disklist;

    if ($disklist{"primary"}) {
        $filer_info{"no_prim_paths"} = @{$disklist{"primary"}};
    } else {
        $filer_info{"no_prim_paths"} = 0;
    }

    if ($disklist{"secondary"}) {
        $filer_info{"no_sec_paths"} = @{$disklist{"secondary"}};
    } else {
        $filer_info{"no_sec_paths"} = 0;
    }

    @local_switchlist = vseries_local_switchlist_get($conn);
    @{$filer_info{"localswitches"}} = @local_switchlist;
    @remote_switchlist = vseries_remote_switchlist_get($conn);
    @{$filer_info{"remoteswitches"}} = @remote_switchlist;
    @slotlist = vseries_slotlist_get($conn, "UP");
    @{$filer_info{"slotlist"}} = @slotlist;

    return (%filer_info);
}

## @name ::vseries_disklist_get
## @description Returns hash having list of primary and secondary
## LUNS connected to the gFiler
## @synopsis
## vseries_disklist_get {$conn}
##
## @arg $conn Required: The gFiler connection to send this operation over
##
##
sub vseries_disklist_get ($) {
    my ($conn) = @_;

    my (%disklist) = ("primary" => undef, "secondary" => undef);
    my ($after, @temp, $line);
    my ($item, $item1, $item2, $disktype);


    (undef, undef, $after) = say($conn, "storage show disk -p", -swallow, '-timeout', 60);
    @temp = split(/\n/, $after);
    @temp = @temp[2 .. $#temp];

    foreach $item (@temp) {
        ($item1, undef, $item2, undef, undef, undef) = split(/\s+/, $item);
        push(@{$disklist{"primary"}}, $item1);

        if (!($item2 =~ /^-$/)) {
            push(@{$disklist{"secondary"}}, $item2);
        }
    }

    if ($disklist{"primary"}) {
        @{$disklist{"primary"}} = sort (@{$disklist{"primary"}});
    }

    if ($disklist{"secondary"}) {
        @{$disklist{"secondary"}} = sort (@{$disklist{"secondary"}});
    }

    return (%disklist);
}

## @name ::vseries_no_redundant_path_get
## @description Returns number of redundant LUNS connected to vseries
##
## @synopsis
## vseries_no_redundant_path_get {$conn}
##
## @arg $conn Required: The gFiler connection to send this operation over
##
##
sub vseries_no_redundant_path_get($) {
    my ($conn) = @_;

    my ($no_secondary_paths) = undef;
    my (%disklist);

    %disklist = vseries_disklist_get($conn);
    $no_secondary_paths = defined(@{$disklist{"secondary"}}) ? @{$disklist{"secondary"}} : 0;

    return ($no_secondary_paths);
}

## @name ::vseries_no_primary_path_get
## @description Returns number of primary LUNS connected to vseries
##
## @synopsis
## vseries_no_primary_path_get {$conn}
##
## @arg $conn Required: The gFiler connection to send this operation over
##
##
sub vseries_no_primary_path_get($) {
    my ($conn) = @_;

    my ($no_primary_paths) = undef;
    my (%disklist);

    %disklist = vseries_disklist_get($conn);
    $no_primary_paths = defined(@{$disklist{"primary"}}) ? @{$disklist{"primary"}} : 0;

    return ($no_primary_paths);
}

## @name ::vseries_local_switchlist_get
## @description Returns an array of switches locally connected to vseries
##
## @synopsis
## vseries_local_switchlist_get {$conn}
##
## @arg $conn Required: The vseries connection to send this operation over
##
##
sub vseries_local_switchlist_get ($) {
    my ($conn) = @_;

    my (%local_switchlist) = ();
    my ($after, @temp, $line, $switch, $switch_name, $length_switch_name);
    my ($state, @switchlist);

    (undef, undef, $after) = say($conn, "priv set -q advanced;
                                 fcadmin link_state", -swallow, '-timeout', 60);
    @temp = split(/\n/, $after);

    foreach $line (@temp) {
        (undef, undef, undef, undef, undef, undef, $state, undef,
           undef, undef, undef, undef, undef, $switch) = split(/\s+/, $line);

        if ($state =~ "UP") {
            if ($switch =~ "switch") {
                (undef, undef, undef, undef, undef, undef, undef, undef,
                   undef, undef, undef, undef, undef, undef,
                     $switch_name ) = split(/\s+/, $line);

                chop($switch_name);

                $length_switch_name = length($switch_name);
                $switch_name = substr($switch_name, 1, $length_switch_name);

                $local_switchlist{$switch_name} = 1;
                @switchlist = keys (%local_switchlist);
            }
        }
    }
    return (keys(%local_switchlist));
}

## @name ::vseries_remote_switchlist_get
## @description Returns an array of remote-switches connected to vseries
##
## @synopsis
## vseries_remote_switchlist_get {$conn}
##
## @arg $conn Required: The vseries connection to send this operation over
##
##
sub vseries_remote_switchlist_get ($) {
    my ($conn) = @_;

    my (%remote_switchlist);
    my (%disklist, @local_switchlist);
    my ($path, $lun, $switch, $local_switch);

    %disklist = vseries_disklist_get($conn);
    @local_switchlist = vseries_local_switchlist_get($conn);

    foreach $path (keys %disklist) {
        foreach $lun (@{$disklist{$path}}) {
             ($switch, undef) = split(/:/, $lun);
             $remote_switchlist{$switch} = 1;
        }
    }
    foreach $local_switch (@local_switchlist) {
        delete ($remote_switchlist{$local_switch});
    }
    return (keys (%remote_switchlist));
}

## @name ::vseries_storage_portlist_get
## @description For a given vseries it returns an array of hash
## (key = "switchname" value = port_num) having info about switch
## and the its port connected to storage array.
##
## @synopsis
## vseries_storage_portlist_get {$conn}
##
## @arg $conn Required: The vseries connection to send this operation over
##
##
sub vseries_storage_portlist_get ($) {
    my ($conn) = @_;

    my (@storage_portlist) = ();
    my (%disklist, $path, $lun, $switch_and_port_num, $switch);
    my ($local_switch, $port_num, $key);
    my (%temp) = ();

    %disklist = vseries_disklist_get($conn);
    foreach $path (keys %disklist) {     #primary and secondary
        foreach $lun (@{$disklist{$path}}) {
             ($switch_and_port_num, undef) = split(/\./, $lun);
             $temp{"$switch_and_port_num"} = 1;
        }
    }
    foreach $key (keys %temp) {
        my %storage_portlist = ();
        next unless ($key =~ /:/);   #if storage is direcltly connected 2 filer
        ($switch, $port_num) = split(/:/, $key);
        $storage_portlist{"switch"} = $switch;
        $storage_portlist{"port_num"} = $port_num;
        push (@storage_portlist, \%storage_portlist);
    }
    return (@storage_portlist);
}

## @name ::vseries_storage_port_find
## @description Returns 1 if the given storage port (switchname and
## port_num) can be seen from vseries
##
## @synopsis
## vseries_storage_port_find {$conn, $switch, $port}
##
## @arg $conn Required: The vseries connection to send this operation over
## @arg $switch Required: the switchname
## @arg $conn Required: the port_num on the given switch
##
##
sub vseries_storage_port_find ($$$) {
    my ($conn, $switch, $port) = @_;
    my (%disklist, $path, $lun, $switch_and_port_num);

    %disklist = vseries_disklist_get($conn);
    foreach $path (keys %disklist) {  #primary and secondary
        foreach $lun (@{$disklist{$path}}) {
             ($switch_and_port_num, undef) = split(/\./, $lun);

             if ($port =~ /\//) {     #it is a Cisco port
                 my ($blade_num, $temp_port) = split (/\//, $port);
                 --$blade_num;
                 $port = "$blade_num" . "-" . "$temp_port";
             }

             if ($switch_and_port_num eq ("$switch" . ":" . "$port")) {
                 return 1; #found the storage port
             }
        }
    }
    return (undef);
}

## @name ::vseries_host_portlist_get
## @description Returns an array of hash (key = "switchname"
## value = port_num) having info about the switch port connected to
## the storage adapters of vseries.
##
## @synopsis
## vseries_host_portlist_get {$conn}
##
## @arg $conn Required: The gFiler connection to send this operation over
##
##
sub vseries_host_portlist_get ($) {
    my ($conn) = @_;

    my (@channel_info);
    my ($slot, $state, $port_num, );
    my ($after, @temp, $line, $switch, $switch_name, $length_switch_name);
    my ($port_type);

    (undef, undef, $after) = say($conn, "priv set -q advanced;
                                fcadmin link_state", -swallow, '-timeout', 60);
    @temp = split(/\n/, $after);
    foreach $line (@temp) {
        (undef, undef, undef, undef, undef, $slot, $state,
          undef, $port_type, undef, undef, $port_num, undef, $switch,
          $switch_name) = split(/\s+/, $line);
        if (($state =~ "UP") && ($port_type =~ "N-port")) {
            my (%channel_info) = ();
            chop($slot);
            $channel_info{"slot"} = $slot;
            if ($switch =~ "switch") {
                chop($switch_name);
                $length_switch_name = length($switch_name);
                $switch_name = substr($switch_name, 1, $length_switch_name);
                $channel_info{"switch"} = $switch_name;
                $channel_info{"switchport"} = $port_num;
            } else {
                $channel_info{"switch"} = undef;
                $channel_info{"switchport"} = undef;
            }
            push(@channel_info, \%channel_info);
        }
    }
    return (@channel_info);
}

## @name ::vseries_fclink_state_get
## @description Returns "OFFLINE" or "DOWN" or "UP" for a given filer and its
## storage slot connected to the switch
##
## @synopsis
## vseries_fclink_state_get {$conn, $slot}
##
## @arg $conn Required: The vseries connection to send this operation over
## @arg $slot Required: Link state of this slot is needed
##
##
sub vseries_fclink_state_get ($$) {
    my ($conn, $slot) = @_;

    my ($state);
    my ($after);

    (undef, undef, $after) = say($conn, "priv set -q advanced;
                         fcadmin link_state $slot", -swallow, '-timeout', 60);

    (undef, undef, undef, undef, undef, $slot, $state) = split(/\s+/, $after);

    return ($state);
}

## @name ::vseries_slotlist_get
## @description Returns array of storage slots having given state on a vseries
##
## @synopsis
## vseries_fclink_state_get {$conn, $state}
##
## @arg $conn Required: The vseries connection to send this operation over
## @arg $state Required: The state a slot should be into
##
##
sub vseries_slotlist_get ($$) {
    my ($conn, $state) = @_;

    my (@slotlist);
    my ($after, @temp, $line, $slot, $current_state);

    (undef, undef, $after) = say($conn, "priv set -q advanced;
                             fcadmin link_state", -swallow, '-timeout', 60);
    @temp = split(/\n/, $after);
    foreach $line (@temp) {
        (undef, undef, undef, undef, undef, $slot,
                          $current_state) = split(/\s+/, $line);
        if ($state eq $current_state) {
            chop($slot);
            push(@slotlist, $slot);
        }
    }
    return (@slotlist);

}

## @name ::vseries_switch_slotlist_get
## @description Returns a reference to an array of storage slots having a given
##              state on a vseries pertaining to the switchname specified.  It
##              will also return 1 on error if the @slotlist is empty.
##
## @synopsis
## vseries_switch_slotlist_get {$conn, $state, $switch_name1, $switch_name2}
##
## @arg $conn Required: The vseries connection to send this operation over.
## @arg $state Required: The state a slot should be in.
## @arg $switch_name1 Required: The switch name you wish to seek adapters for.
## @arg $switch_name2 Optional -- Another switch you like to seek adapters for.
##
## @Author silva@netapp.com
##
sub vseries_switch_slotlist_get {
    my ($conn, $state, $switch_name1, $switch_name2) = @_;
    my (@slotlist, $sw_name, $after, @temp, $line, $slot, $current_state );

    (undef, undef, $after) = say($conn, "priv set -q advanced;
                             fcadmin link_state", -swallow, '-timeout', 60);
    @temp = split(/\n/, $after);

    foreach $line (@temp) {
        (undef, undef, undef, undef, undef, $slot, $current_state, undef, undef,
        undef, undef, undef, undef, undef, $sw_name) = split(/\s+/, $line);

        if ($state eq $current_state) {

            if (defined $switch_name2) {

                if (("'$switch_name1'" eq $sw_name) ||
                     "'$switch_name2'" eq $sw_name){
                    chop($slot);
                    push(@slotlist, $slot);
                }
            }#defined

            else {

                if ("'$switch_name1'" eq $sw_name){
                    chop($slot);
                    push(@slotlist, $slot);
                }
            }#else

        }#$state
   }#foreach $line

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # If @slotlist does not get populated return 1 (error)
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    if (@slotlist == 0){
        return(0);
    }

    return (\@slotlist);
}

## @name ::vseries_partner_name_get
## @description Returns name of the partner filer
##
## @synopsis
## vseries_fclink_state_get {$conn}
##
## @arg $conn Required: The vseries connection to send this operation over
##
sub vseries_partner_name_get ($) {
    my ($conn) = @_;

    my ($partner_name) = undef;
    my ($after, @temp, $line, $name_length);

    (undef, undef, $after) = say($conn, "sysconfig", -swallow, '-timeout', 60);
    @temp = split(/\n/, $after);

    foreach $line (@temp) {
         next unless ($line =~ "partner ID");

         $line =~ s/^\s+//g;
         ($partner_name) = (split (/\s+/, $line))[7];

         chop($partner_name);

         $name_length = length($partner_name);
         $partner_name = substr($partner_name, 1, $name_length);
    }
    return ($partner_name);

}

## @name ::vseries_cluster_slot_info_get
## @description Returns array of hash, each hash has information about
## cluster slot number, its WWN (if it is metro cluster), cluster
## link number and cluster_type (metro | regular-troika | regular-mellanox)
## for a given filer
##
## @synopsis
## vseries_cluster_slot_info_get{$conn}
##
## @arg $conn Required: The gFiler connection to send this operation over
##
sub vseries_cluster_slot_info_get($) {
    my ($conn) = @_;

    my ($after, @temp, $line, @cluster_slotlist);
    my ($WWN) = undef;
    my ($cluster_type) = undef;
    my ($maybe_cluster_slot) = undef;
    my (@heartbeat_slot_num, @link);

    (undef, undef, $after) =
                       say($conn, "sysconfig -v", -swallow, '-timeout', 240);
    @temp = split(/\n/, $after);
    my $i=0;

    # - - - - - - - -  - - - - - - - - - - - - - - - - - -  -
    # FCIV = Qlogic Metrocluster card,
    # Troika VI = Troika regular cluster card
    # Mellanox = Mellanox another regular cluster card
    # NVRAMV
    # Spectre_Stretched
    # - - - - - - - -  - - - - - - - - - - - - - - - - - -  -
    # Tested on GF980 - metro cluster with NVRAM 5
    # Tested on GF960 - metro cluster without NVRAM 5
    # Tested on GF825 - metro cluster without NVRAM 5
    # Tested on GF825 - Regular Troika cluster without NVRAM 5
    # Tested on GF825 - Regular IB cluster without NVRAM 5
    # Tested on GF960 - IB Mellanox cluster without NVRAM 5
    # Tested on GF960 - NVRAMV cluster obviously with NVRAM 5
    # Tested on v3140 - (Spectre) FCVI direct attached (L-Port)
    # Tested on v3170 - (Spectre) FCVI direct attached (L-Port)
    # - - - - - - - -  - - - - - - - - - - - - - - - - - -  -
    foreach $line (@temp) {
        next unless (($line =~ "FCVI Host Adapter ") ||
                     ($line =~ "Troika FCVI") ||
                     ($line =~ "NVRAM V") ||
                     ($line =~ "Cluster Interconnect Port") ||
                     ($line =~ "Mellanox IB") ||
                     ($line =~ "IB Host Adapter"));
        $line =~ s/^\s+//g;
        if ($line =~ "FCVI Host Adapter.*F-port") {
            $cluster_type = "metro";
            ($heartbeat_slot_num[$i]) = (split (/\s+/, $line))[5];
            ++$i;
        } elsif ($line =~ "FCVI Host Adapter.*(L|N)-port") {
            $cluster_type = "stretched";
            $line =~ s/^\s+//;
            ($heartbeat_slot_num[$i]) = (split (/\s+/, $line))[5];
            ++$i;
        } elsif ($line =~ "Troika FCVI") {
            $cluster_type = "regular-troika";
            ($heartbeat_slot_num[$i]) = (split (/\s+/, $line))[1];
            chop($heartbeat_slot_num[$i]);
            # Troika has only 1 link, so don't need to get value for other link
        } elsif ($line =~ "Mellanox IB") {
            $cluster_type = "regular-mellanox";
            ($heartbeat_slot_num[$i]) = (split (/\s+/, $line))[1];
            chop($heartbeat_slot_num[$i]);
            # The sysconfig -v has only slot number output for IB cluster,
            # (no a nd b), so covering up that
            $heartbeat_slot_num[$i + 1] = $heartbeat_slot_num[$i]
        } elsif ($line =~ "NVRAM V") {
            $maybe_cluster_slot = (split (/\s+/, $line))[1];
            chop($maybe_cluster_slot);
        } elsif ($line =~ "Cluster Interconnect Port") {
            if ($line =~ /copper/i) {
                $cluster_type = "NVRAMV";
                ($heartbeat_slot_num[$i]) = $maybe_cluster_slot;
                ++$i;
            }
        } elsif ($line =~ "IB Host Adapter ") {            
                $cluster_type = "IB-NVRAM8";
                ($heartbeat_slot_num[$i]) = (split (/\s+/, $line))[1];
                
            
        }        

    }
    
    # If we could not determine the cluster type due above then
    # check some other commands to determine the type.  For example
    # new 'carnegie' systems do not have an interconnect that shows
    # up in sysconfig output in any way that's distinguishable from
    # their metrocluster equivalent except the presence of an FCVI
    # card for the metro cluster config.  So if we get here in that
    # case we can try to run a specific command for the non-metrocluster
    # carnegie platform.  That's what we do here.  In this case cluster_type
    # should end up as iWARP. and the slot number is always 0.
    if(! defined($cluster_type)){
               
        (undef, undef, $after) = say($conn, "priv set -q advanced; ic config show", -swallow, '-timeout', 240);
	    
	if($after !~ m/"Usage:"/){
	    my @temp2 = split(/\n/, $after);
	 	    
            foreach my $line2 (@temp2){
	        if($line2 =~ m/interconnect type:/i){	    	     	    	        
	            $cluster_type = (split(/\:\s+/, $line2))[1];
	            last;
	        }
	    }
	    ($heartbeat_slot_num[$i]) = 0;
	    ++$i;
	}
    }

    if ($cluster_type =~ /metro$/) {
        (undef, undef, $after) = say($conn, "sysconfig -M", -swallow,
                                                         '-timeout', 240);
        @temp = split(/\n/, $after);
        my $i=0;
        foreach $line (@temp) {
            next unless (($line =~ "FCVI") || ($line =~ "Troika VI") ||
                         ($line =~ "Mellanox IB"));
            if ($line =~ "FCVI") {
                ($WWN) = (split (/!/, $line))[2];
            }

            my %heartbeat_info;
            $heartbeat_info{"cluster_type"} = $cluster_type;
            $heartbeat_info{"slot"} = $heartbeat_slot_num[$i];
            $heartbeat_info{"WWN"} = $WWN;
            $heartbeat_info{"link"} = $i;
            push (@cluster_slotlist, \%heartbeat_info);
            ++$i;
        }
    } elsif ($cluster_type =~ "stretched") {
        (undef, undef, $after) = say($conn, "sysconfig -M", -swallow,
                                                         '-timeout', 240);
        @temp = split(/\n/, $after);
        my $i=0;
        foreach $line (@temp) {
            next unless ($line =~ "FCVI");
            ($WWN) = (split (/!/, $line))[2];

            my %heartbeat_info;
            $heartbeat_info{"cluster_type"} = $cluster_type;
            $heartbeat_info{"slot"} = $heartbeat_slot_num[$i];
            $heartbeat_info{"WWN"} = $WWN;
            $heartbeat_info{"link"} = $i;
            push (@cluster_slotlist, \%heartbeat_info);
            ++$i;
        }
    } elsif ($cluster_type =~ "regular-mellanox") {
        foreach my $i (0,1) {   # IB has 2 cluster links
            my %heartbeat_info;
            $heartbeat_info{"cluster_type"} = $cluster_type;
            $heartbeat_info{"slot"} = $heartbeat_slot_num[$i];
            $heartbeat_info{"WWN"} = undef;
            $heartbeat_info{"link"} = $i;
            push (@cluster_slotlist, \%heartbeat_info);
        }
    } elsif ($cluster_type =~ "regular-troika") {
        # Troika has only 1 cluster link
        my $i=0;
        my %heartbeat_info;
        $heartbeat_info{"cluster_type"} = $cluster_type;
        $heartbeat_info{"slot"} = $heartbeat_slot_num[$i];
        $heartbeat_info{"WWN"} = undef;
        $heartbeat_info{"link"} = $i;
        push (@cluster_slotlist, \%heartbeat_info);
    } elsif ($cluster_type =~ "NVRAMV" ){
        foreach my $i (0,1) {   # NVRAM V has 2 cluster links
            my %heartbeat_info;
            $heartbeat_info{"cluster_type"} = $cluster_type;
            $heartbeat_info{"slot"} = $heartbeat_slot_num[$i];
            $heartbeat_info{"WWN"} = undef;
            $heartbeat_info{"link"} = $i;
            push (@cluster_slotlist, \%heartbeat_info);
        }
    } elsif ($cluster_type =~ "iWARP"){
        my $i = 0;
        my %heartbeat_info;
        $heartbeat_info{"cluster_type"} = $cluster_type;
        $heartbeat_info{"slot"} = $heartbeat_slot_num[$i];
        $heartbeat_info{"WWN"} = undef;
        $heartbeat_info{"link"} = $i; 
        push (@cluster_slotlist, \%heartbeat_info);
    }
    elsif ($cluster_type =~ "IB-NVRAM8") {
        # IB-NVRAM8 has only 1 cluster slot
        # but two ports underneath that, we're
        # just recording slots here.
        my $i=0;
        my %heartbeat_info;
        $heartbeat_info{"cluster_type"} = $cluster_type;
        $heartbeat_info{"slot"} = $heartbeat_slot_num[$i];
        $heartbeat_info{"WWN"} = undef;
        $heartbeat_info{"link"} = $i;
        push (@cluster_slotlist, \%heartbeat_info);  
    }    
    
    return(@cluster_slotlist);
}

## @name ::vseries_heartbeat_link_status_find
## @description Returns status of the cluster link (up/down/undef)
##
## @synopsis
## vseries_heartbeat_link_status_find{$conn, $link_num}
##
## @arg $conn Required: The gFiler connection to send this operation over
## @arg $link_num Required: cluster link number
##
sub vseries_heartbeat_link_status_find($$) {
    my ($conn, $link) = @_;

    my ($after, @temp, $line);
    my ($link_status) = undef;

    (undef, undef, $after)=say($conn, "priv set -q advanced; ic stats error -v",
                                 -swallow, '-timeout', 300);
    @temp = split(/\n/, $after);
    my $link_counter = 0;   #used for metro cluster

    foreach $line (@temp) {
        # Take any multiple \r and \n off line
        $line =~ s/\r+$//;
        $line =~ s/\n+$//;

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Troika (tested on GF825, raid-u11 and raid-u12)
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        if ($line =~ /Link alive/i) {
            $line =~ s/^\s+//g;
            ($link_status) = (split (/:/, $line))[1];
            $link_status =~ s/^\s+//g;

            if ($link_status == 1) {
                $link_status = 'up';
            } else {
                $link_status = 'down';
            }

            chomp($link_status);
            return ($link_status);
        }

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Metro Cluster, (tested on GF825, raid-u32 and raid-u33 without NVRAM5)
        # Tested on GF960 (raid-u30 and raid-u31 without NVRAM5)
        # Tested on GF980 (gf980-svl1 and gf980-svl2 with NVRAM5)
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        if ($line =~ /Nic state:/i) {
            $line =~ s/^\s+//g;
            ($link_status) = (split (/:/, $line))[1];
            $link_status =~ s/^\s+//g;

            chomp($link_status);

            if ($link == $link_counter) {
                return ($link_status);
            } else {
                ++$link_counter;
            }
        }

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Need to insert code for IB cluster type
        # Tested on GF825 without NVRAM V (gf825-svl1 and svl2)
        # Tested on GF960 without NVRAM V (gf960-svl1 and svl2)
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        if ($line =~ /Port $link state/i) {
            $line =~ s/^\s+//g;
            ($link_status) = (split (/:/, $line))[1];
            $link_status =~ s/^\s+//g;

            if ($link_status =~ /ACTIVE/i) {
                $link_status = 'up';
            } else {
                $link_status = 'down';
            }

            chomp($link_status);
            return ($link_status);
        }

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # NVRAM5 (tested on GF960, snake and tooz)
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        my $temp1 = $link + 1;
        if ($line =~ /Link $temp1 state:/i) {
            $line =~ s/^\s+//g;
            ($link_status) = (split (/:/, $line))[1];
            $link_status =~ s/^\s+//g;

            chomp($link_status);

            $link_status = lc($link_status);

            return ($link_status);
        }
        
        if($line =~ /Link $temp1:/){
            $line =~ s/^\s+//g;
            ($link_status) = (split (/:/, $line))[1];
	    $link_status =~ s/^\s+//g;
	    
	    chomp($link_status);
	    
	    $link_status = lc($link_status);
	    
            return ($link_status);
        }
    }

    return (undef);
}

## @name ::vseries_hw_sw_info_get
## @description Returns hash having following keys and their value
## "model name" => <value>
## "firmware release" => <value>
## "software release" => <value>
##
## @synopsis
## vseries_hw_sw_info_get{$conn}
##
## @arg $conn Required: The vseries connection to send this operation over
##
##
sub vseries_hw_sw_info_get ($) {
    my ($conn) = @_;

    my ($after, @temp, $line);
    my ($model_name, $firmware_release, $software_release);
    my (%filer_hw_sw_info) = ();

    (undef, undef, $after) = say($conn, "sysconfig -v 0", -swallow,
                                                    '-timeout', 60);

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # get model name
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @temp = split(/\n/, $after);
    foreach $line (@temp) {
        next unless ($line =~ /Model Name/);
        $line =~ s/^\s+//g;
        ($model_name) = (split (/\s+/, $line))[2];
    }

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # get firmware release
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    foreach $line (@temp) {
        next unless ($line =~ /Firmware release/);
        $line =~ s/^\s+//g;
        ($firmware_release) = (split (/\s+/, $line))[2];
    }

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # get software release
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    (undef, undef, $after) = say($conn, "version", -swallow, '-timeout', 60);
    @temp = split(/\n/, $after);
    foreach $line (@temp) {
        next unless ($line =~ /Release/);
        $line =~ s/^\s+//g;
        ($software_release) = (split (/\s+/, $line))[2];
        chop($software_release);
    }

    $filer_hw_sw_info{"model name"} = $model_name;
    $filer_hw_sw_info{"firmware release"} = $firmware_release;
    $filer_hw_sw_info{"software release"} = $software_release;

    return (%filer_hw_sw_info);
}

## @name ::vseries_storage_info_get
## @description Extract information about storage arrays connected to vseries.
## Returns an hash having following keys and their value
## <vendor_iname> => @ of model
##
## @synopsis
## vseries_switch_info_get{@conn}
##
## @arg $conn Required: Array of vseries connections
##
sub vseries_storage_info_get($) {
    my ($conn_param) = @_;

    my ($after, @temp, $line);
    my ($vendor, $model);
    my (%storage_info) = ();
    my (%seen) = ();

    foreach my $conn (@{$conn_param}) {
        (undef, undef, $after) = say($conn, "storage show disk", -swallow,
                                                    '-timeout', 60);

        @temp = split(/\n/, $after);
        @temp = @temp[2 .. $#temp];

        foreach $line (@temp) {
            $vendor = (split (/\s+/, $line))[4];
            $model = (split (/\s+/, $line))[5];

            if (!exists($storage_info{$vendor})) {
                push (@{$storage_info{$vendor}}, $model);
                $seen{$model} = 1;
            }

            if (!exists($seen{$model})) {
                push (@{$storage_info{$vendor}}, $model);
            }
        }
    }
    return (%storage_info);
}

sub vseries_init {
    my ($tot_filers, $tot_switches, $filer_list, $switch_list,
        $filer_conn_list, $switch_conn_list, $fh_rec, $sh_rec,
        $fc_rec, $sc_rec, $filer_info, $hba_conn_list,
        $storage_conn_list, $sleep_time, $CLUSTERED, $cluster_type,
        $heartbeat_conn_list, $isl_conn_list, $switch_info, $storage_info) = @_;

    my $FAST_ACTION_SLEEPTIME = 6;
    my $SLOW_ACTION_SLEEPTIME = 70;

    # - - - - - - - - - - - - - - - - - - - - - - - - -
    # Populate @switch_list, $tot_switches, @filer_list, $tot_filers,
    # @switch_conn_list, @filer_conn_list
    # - - - - - - - - - - - - - - - - - - - - - - - - -

    if (defined ($main::SWITCH0)) {
        $switch_list->[${$tot_switches}] = $main::SWITCH0;
        $switch_conn_list->[${$tot_switches}] = $main::SWITCH0_CONN;
        ++${$tot_switches};
    }

    if (defined($main::SWITCH1)) {
        $switch_list->[${$tot_switches}] = $main::SWITCH1;
        $switch_conn_list->[${$tot_switches}] = $main::SWITCH1_CONN;
        ++${$tot_switches};
    }

    if (defined($main::SWITCH2)) {
        $switch_list->[${$tot_switches}] = $main::SWITCH2;
        $switch_conn_list->[${$tot_switches}] = $main::SWITCH2_CONN;
        ++${$tot_switches};
    }

    if (defined($main::SWITCH3)) {
        $switch_list->[${$tot_switches}] = $main::SWITCH3;
        $switch_conn_list->[${$tot_switches}] = $main::SWITCH3_CONN;
        ++${$tot_switches};
    }

    # - - - - - - - - - - - - - - - - - - - - - - - - -
    # Total no of switches should not be more than 4
    # - - - - - - - - - - - - - - - - - - - - - - - - -

    if (${$tot_switches} > 4) {
        tstlog(-comment, 'Max number of switches can be 4');
        tstlog(-comment, 'There are more than 4 switches defined');
        tstlog(-fatal, 'Test FAILED');
    }

    # - - - - - - - - - - - - - - - - - - - - - - - - -
    # Check if a switch is specified more than once
    # - - - - - - - - - - - - - - - - - - - - - - - - -

    {
        my %seen = ();
        foreach my $switch (@{$switch_list}) {
            if (exists $seen{$switch}) {
                tstlog(-comment, "Switch $switch is mentioned more than once");
                tstlog(-fatal, 'Test FAILED');
            } else {
                $seen{$switch} = 1;
            }
        }
    }

    # - - - - - - - - - - - - - - - - - - - - - - - - -
    # Work on FILER* parameters
    # - - - - - - - - - - - - - - - - - - - - - - - - -

    if (defined($main::FILERA)) {
        $filer_list->[${$tot_filers}] = $main::FILERA;
        $filer_conn_list->[${$tot_filers}] = $main::FILERA_CONN;
        ++${$tot_filers};
    }
    if (defined($main::FILERB)) {
        $filer_list->[${$tot_filers}] = $main::FILERB;
        $filer_conn_list->[${$tot_filers}] = $main::FILERB_CONN;
        ++${$tot_filers};
    }

    # - - - - - - - - - - - - - - - - - - - - - - - - -
    # At least 1 filer should be mentioned and
    # Total no of filers should not be more than 2
    # - - - - - - - - - - - - - - - - - - - - - - - - -

    if ((${$tot_filers} == 0) || (${$tot_filers} > 2)) {
        tstlog(-comment, 'There should at least be 1 filer defined OR');
        tstlog(-comment, 'There are more than 2 filers defined');
        tstlog(-fatal, 'Test FAILED');
    }

    # - - - - - - - - - - - - - - - - - - - - - - - - -
    # Check if a filer is specified more than once
    # - - - - - - - - - - - - - - - - - - - - - - - - -

    {
        my %seen = ();
        foreach my $filer (@{$filer_list}) {
            if (exists $seen{$filer}) {
                tstlog(-comment, "Filer $filer is mentioned more than once");
                tstlog(-fatal, 'Test FAILED');
            } else {
                $seen{$filer} = 1;
            }
        }
    }

    # - - - - - - - - - - - - - - - - - - - - - - - - -
    # Establish the connections to filers and switches
    # Populate %fh_rec, %fc_rec, %sh_rec, %sc_rec
    # - - - - - - - - - - - - - - - - - - - - - - - - -

    for (my $i = 0; $i < ${$tot_filers}; ++$i) {
        $fh_rec->{$filer_list->[$i]} = host($filer_list->[$i]);
        $fc_rec->{$filer_list->[$i]} = connect($filer_conn_list->[$i]);
    }

    for (my $i = 0; $i < ${$tot_switches}; ++$i) {
        $sh_rec->{$switch_list->[$i]} = host($switch_list->[$i]);
        $sc_rec->{$switch_list->[$i]} = connect($switch_conn_list->[$i]);
    }


    # - - - - - - - - - - - - - - - - - - - - - - - - -
    # Initialize sleeptime between each actions
    # - - - - - - - - - - - - - - - - - - - - - - - - -
    ${$sleep_time} = ($main::ACTION_TYPE =~ /FAST/) ? $FAST_ACTION_SLEEPTIME :
                     $SLOW_ACTION_SLEEPTIME;

    # - - - - - - - - - - - - - - - - - - - - - - - - -
    # If only one filer is given in the command but it is  clustered
    # then then partner's name should also be given in the command
    # line. The main reason for this is; a test may require verification
    # in the partner filer, so its better to capture the partner's
    # console messages while running the test.
    # - - - - - - - - - - - - - - - - - - - - - - - - -
    if (${$tot_filers} == 1) {
        my $partner_name = undef;
        $partner_name = vseries_partner_name_get($fc_rec->{$filer_list->[0]});

        if (defined($partner_name)) {
             tstlog (-comment, "$filer_list->[0] is partnered 2 $partner_name");
             tstlog (-comment, "Give partner filer in command line as well");
             tstlog (-fatal, 'Test FAILED');
        }
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # The filers should be clustered if they are 2
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    my %cluster_partner = ();

    if (${$tot_filers} == 2) {
        # Both should be partnered
        if (($filer_list->[1] ne
                       vseries_partner_name_get($fc_rec->{$filer_list->[0]})) ||
            ($filer_list->[0] ne
                       vseries_partner_name_get($fc_rec->{$filer_list->[1]}))) {
                tstlog (-comment, "Filers $filer_list->[0] and $filer_list->[1]
                                   are not connected. Exiting");
                tstlog (-fatal, 'Test FAILED');
        }

        ${$CLUSTERED} = 1;
        $cluster_partner{$filer_list->[0]} = $filer_list->[1];
        $cluster_partner{$filer_list->[1]} = $filer_list->[0];
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Intitialize switch information like, switchname, firmware, vendor etc..
    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    {
        foreach my $switch (@{$switch_list}) {
            my $conn = $sc_rec->{$switch};
            my %info = switch_hw_sw_info_get($conn);
            my $rec = {};
            $switch_info->{$switch}                        = $rec;
            $switch_info->{$switch}{'vendor'}              = $info{'vendor'};
            $switch_info->{$switch}{'model'}               = $info{'model'};
            $switch_info->{$switch}{'firmware'}            = $info{'firmware'};
            @{$switch_info->{$switch}{'hba_portlist'}}     = ();
            @{$switch_info->{$switch}{'storage_portlist'}} = ();
            @{$switch_info->{$switch}{'cluster_portlist'}} = ();
            @{$switch_info->{$switch}{'isl_portlist'}}     = ();
        }
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Intitialize storage array(s) information
    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    {
        my @conn = ();
        for (my $i=0; $i<${$tot_filers}; ++$i) {
            $conn[$i] = $fc_rec->{$filer_list->[$i]} ;
        }
        %{$storage_info} = vseries_storage_info_get(\@conn);
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Intitialize filer(s) information
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $filer (keys %{$fc_rec}) {
        my $conn = $fc_rec->{$filer};
        my $rec  = {};
        $filer_info->{$filer} = $rec;
        my %temp1 = vseries_hw_sw_info_get($conn);
        my %temp  = vseries_info_get($conn);

        $rec->{'model name'}       = $temp1{'model name'};
        $rec->{'firmware release'} = $temp1{'firmware release'};
        $rec->{'software release'} = $temp1{'software release'};

        if (${$CLUSTERED}) {
            $rec->{'clustered'}  = 'yes';
            $rec->{'partner'}    = $cluster_partner{$filer};
            my @cluster_slotlist =  vseries_cluster_slot_info_get($conn);

            foreach my $entry (@cluster_slotlist) {
                $rec->{'cluster_type'} = $entry->{'cluster_type'};
                push (@{$rec->{'cluster_slotlist'}}, $entry->{'slot'});
                ${$cluster_type} = $entry->{'cluster_type'};
            }
        } else {
            $rec->{'clustered'}           = 'no';
            $rec->{'partner'}             = undef;
            $rec->{'cluster_type'}        = undef;
            @{$rec->{'cluster_slotlist'}} = ();
            ${$cluster_type}              = undef;
        }

        @{$rec->{'localswitches'}}  = @{$temp{'localswitches'}};
        @{$rec->{'remoteswitches'}} = @{$temp{'remoteswitches'}};

        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Check if storage arrays are directly connected
        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        if ((@{$temp{'localswitches'}} == 0) &&
            (@{$temp{'remoteswitches'}} == 0)) {
            $rec->{'storagearrays_are_directly_connected'} = 'yes';
        } else {
            $rec->{'storagearrays_are_directly_connected'} = 'no';
        }

        $rec->{'no_prim_paths'} = $temp{'no_prim_paths'};
        $rec->{'no_sec_paths'}  = $temp{'no_sec_paths'};

        if ($temp{'primary'}) {
            @{$rec->{'prim_luns'}} = @{$temp{'primary'}};
        } else {
            @{$rec->{'prim_luns'}} = ();
        }

        if ($temp{'secondary'}) {
        @{$rec->{'sec_luns'}} = @{$temp{'secondary'}};
        } else {
            @{$rec->{'sec_luns'}} = ();
        }

        @{$rec->{'storage_slotlist'}} = @{$temp{'slotlist'}};
    } #filer

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    #  All switches given in the command line should be connected to each and
    #  every filer(s)
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    my %switch_connected = ();
    foreach my $switch (@{$switch_list}) {
        foreach my $filer (keys %{$filer_info}) {
            my @local_switchlist  = @{$filer_info->{$filer}{'localswitches'}};
            my @remote_switchlist = @{$filer_info->{$filer}{'remoteswitches'}};
            my @tot_switchlist;

            push (@tot_switchlist, @local_switchlist);
            push (@tot_switchlist, @remote_switchlist);

            foreach my $switch1 (@tot_switchlist) {
                if ($switch eq $switch1) { $switch_connected{$switch} = 1; }
            }
        }
    }

    foreach my $switch (@{$switch_list}) {
        if (!exists ($switch_connected{$switch})) {
            tstlog(-comment,
                   "Switch $switch is not connected to any of the given filer");
            tstlog(-fatal, 'Test FAILED');
        }
    }


    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Number of primary and secondary paths should be same before
    # starting any switch related test.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    foreach my $filer (keys %{$fc_rec}) {
        if (($filer_info->{$filer}{'no_prim_paths'}) !=
             ($filer_info->{$filer}{'no_sec_paths'})) {
            tstlog (-comment,
                 "$filer prim luns=$filer_info->{$filer}{\"no_prim_paths\"}");
            tstlog (-comment,
                 "$filer sec luns=$filer_info->{$filer}{\"no_sec_paths\"}");
            tstlog (-comment,
                  "Number of prim and sec paths are NOT equal in $filer");
            tstlog (-fatal, 'Test FAILED');
        }
    }


    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Perform some error checking for cluster configuration
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    if (${$CLUSTERED}) {
        foreach my $filer (keys %{$fc_rec}) {
            # -  - - - - - - - - - - - - - - - - -  - - - - - -
            # Cluster should be UP and running
            # -  - - - - - - - - - - - - - - - - -  - - - - - -
            for (my $retry = 0; $retry < 61; ++$retry) {
                if ('UP' ne get_cf_state($fh_rec->{$filer})) {
                    sleep 7;
                    next unless ($retry == 60);
                    tstlog (-comment, 'Cluster is Not UP');
                    tstlog (-fatal, 'Test FAILED');
                } else {
                    last;
                }
            }

            # -  - - - - - - - - - - - - - - - - -  - - - - - -
            # Perform some error checking for metro cluster
            # -  - - - - - - - - - - - - - - - - -  - - - - - -
            if (${$cluster_type} =~ /metro/) {
                # -  - - - - - - - - - - - - - - - - -  - - - - - -
                # Both cluster links should be UP for metro cluster
                # -  - - - - - - - - - - - - - - - - -  - - - - - -
                for (my $retry=0; $retry < 61; ++$retry) {
                    if ( (vseries_heartbeat_link_status_find(
                                       $fc_rec->{$filer}, 0) ne 'up' ) &&
                         (vseries_heartbeat_link_status_find(
                                       $fc_rec->{$filer}, 1) ne 'up' )) {
                        sleep 7;
                        next unless ($retry == 60);
                        tstlog (-comment,
                                "Both cluster links are Not UP on $filer");
                        tstlog (-fatal, 'Test FAILED');
                     } else {
                         last;
                     }
                }
                # -  - - - - - - - - - - - - - - - - -  - - - - - -
                # Metro cluster should have atleast 2 switches on the local site
                # and 2 switches on the remote site
                # -  - - - - - - - - - - - - - - - - -  - - - - - -
                if ($filer_info->{$filer}{'remoteswitches'} < 2) {
                    tstlog (-comment, "$filer remoteswitches are less than 2");
                    tstlog (-comment,
                    "remoteswitches=$filer_info->{$filer}{\"remoteswitches\"}");
                    tstlog (-fatal, 'Test FAILED');
                }

                if ($filer_info->{$filer}{'localswitches'} < 2) {
                    tstlog (-comment, "$filer localswitches are less than 2");
                    tstlog (-comment,
                    "localswitches = $filer_info->{$filer}{\"localswitches\"}");
                    tstlog (-fatal, 'Test FAILED');
                }
            } #metro
        } #filer
    } #CLUSTERED

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Intitialize switch ports on the host (filer) side
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $filer (keys %{$fc_rec}) {
        my $conn = $fc_rec->{$filer};
        my @temp = vseries_host_portlist_get($conn);

        foreach my $entry (@temp) {
            my %rec            = ();
            $rec{'filer'}      = $filer;
            $rec{'filer_slot'} = $entry->{'slot'};
            $rec{'switch'}     = $entry->{'switch'};

            my ($sw_conn, $sw_host);
            my ($hosttype);

            #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
            #Check if the this switch is mentioned in command line
            #if not then make a temporary connection to get the info
            #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
            if (!exists $sc_rec->{$entry->{'switch'}}) {
                $sw_host  = host($entry->{'switch'});
                $sw_conn  = connect('#telnet');
                $hosttype = $sw_conn->{hostp}->{hosttype};
                disconnect($sw_conn);
            } else {
                $sw_conn  = $sc_rec->{$rec{'switch'}};
                $hosttype = $sw_conn->{hostp}->{hosttype};
            }
            if ($hosttype =~ /switch-cisco/i) {
                my $temp_switchport = $entry->{'switchport'};
                my ($port_num, $blade_num);
                ($blade_num, $port_num) = split (/-/, $temp_switchport);
                ++$blade_num;
                $temp_switchport = $blade_num . "\/" . $port_num;
                $rec{'switch_port'} = $temp_switchport;
            } else {
                $rec{'switch_port'} = $entry->{'switchport'};
            }
            push (@{$hba_conn_list}, \%rec);
        }
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Intialize the data structure used for executing the test
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $temp1 (@{$hba_conn_list}) {
        push (@{$switch_info->{$temp1->{'switch'}}{'hba_portlist'}},
                                              $temp1->{'switch_port'});
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Intialize switch ports on the storage side
    # Storage ports of all switches can be found out by using only 1 filer
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    {
        my $filer = $filer_list->[0];
        my $conn = $fc_rec->{$filer};
        my  @temp = vseries_storage_portlist_get($conn);

        foreach my $entry (@temp) {
            my %rec = ();
            $rec{"switch"} = $entry->{'switch'};
            my ($sw_conn, $sw_host);
            my ($hosttype);

            #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
            #Check if the this switch is mentioned in command line
            #if not then make a temporary connection to get the info
            #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
            if (!exists $sc_rec->{$entry->{'switch'}}) {
                $sw_host = host($entry->{'switch'});
                $sw_conn = connect('#telnet');
                $hosttype = $sw_conn->{hostp}->{hosttype};
                disconnect($sw_conn);
            } else {
                $sw_conn = $sc_rec->{$rec{'switch'}};
                $hosttype = $sw_conn->{hostp}->{hosttype};
            }

            if ($hosttype =~ /switch-cisco/i) {
                my $temp_switchport = $entry->{'port_num'};
                my ($port_num, $blade_num);
                ($blade_num, $port_num) = split (/-/, $temp_switchport);
                ++$blade_num;
                $temp_switchport = $blade_num . "\/" . $port_num;
                $rec{'port_num'} = $temp_switchport;
            } else {
                $rec{'port_num'} = $entry->{'port_num'};
            }

            push (@{$storage_conn_list}, \%rec);
        }
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Intialize the data structure used for executing the test
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $temp1 (@{$storage_conn_list}) {
        push (@{$switch_info->{$temp1->{'switch'}}{'storage_portlist'}},
                                              $temp1->{'port_num'});
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    #Initialize heartbeat ports for switches for Metrocluster
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    if (${$CLUSTERED}) {
        foreach my $filer (keys %{$fc_rec}) {
            next unless ($filer_info->{$filer}{'storagearrays_are_directly_connected'} =~ /no/i);

            my $conn = $fc_rec->{$filer};
            my @cluster_slotlist =  vseries_cluster_slot_info_get($conn);

            foreach my $entry (@cluster_slotlist) {
                my %rec = ();
                $rec{'filer'}        = $filer;
                $rec{'filer_slot'}   = $entry->{'slot'};
                $rec{'link'}         = $entry->{'link'};
                $rec{'cluster_type'} = $entry->{'cluster_type'};

                if ($entry->{'cluster_type'} =~ /metro/) {
                    $rec{'filer_slot_WWN'} = $entry->{'WWN'};

                    # - - - - - - -  - - -  - - -  - - -  - - -  - -
                    # Find the cluster port on a switch, it should be connected
                    # to one of the local switches
                    # - - - - - - -  - - -  - - -  - - -  - - -  - -
                    foreach my $switch
                        (@{$filer_info->{$filer}{'localswitches'}}) {
                        my $switch_port = undef;

                        #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
                        #Check if the this switch is mentioned in command line
                        #if not then make a temporary connection to get the info
                        #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
                        if (!exists $sc_rec->{$switch}) {
                            my $temp_switch_host = host($switch);
                            my $temp_switch_conn = connect('#telnet');
                            $switch_port =  switch_port_num_get(
                                                            $temp_switch_conn,
                                                            $entry->{'WWN'}
                                                            );
                            disconnect($temp_switch_conn);
                        } else {
                            $switch_port =  switch_port_num_get(
                                                            $sc_rec->{$switch},
                                                            $entry->{'WWN'}
                                                            );
                        }

                        if (defined($switch_port)) {
                             $rec{'switch'}          = $switch;
                             $rec{'switch_port_num'} = $switch_port;
                        }
                    }

                    # - - - - - - -  - - -  - - -  - - -  - - -  - -
                    # If the cluster port is not connected to any of the
                    # the local switches, the switch should be mentioned in
                    # the command line as well, then fail the test.
                    # The heartbeat links need to be brought up before
                    # starting the test
                    # - - - - - - -  - - -  - - -  - - -  - - -  - -
                    if (!exists $rec{'switch'}) {
                        tstlog(-comment,
                                 "$filer: The Heartbeat links are not UP");
                        tstlog(-fatal, 'Test FAILED');
                    }
                } else { # metro
                    $rec{'switch'}          = undef;
                    $rec{'switch_port_num'} = undef;
                }

                push (@{$heartbeat_conn_list}, \%rec);
            } #entry
        } #filer
    } #CLUSTER

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Intialize the data structure used for executing the test
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $temp1 (@{$heartbeat_conn_list}) {
        push (@{$switch_info->{$temp1->{'switch'}}{'cluster_portlist'}},
                                              $temp1->{'switch_port_num'});
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Initialize ISL ports for switches for Metrocluster
    # This information can be built by using only 1 filer
    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    {
    my $filer = $filer_list->[0];
    if (${$cluster_type} =~ /metro/) {
        foreach my $switch (@{$filer_info->{$filer}{'localswitches'}}) {
            my @eport_info = ();

            #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
            # Check if the this switch is mentioned in command line
            # if not then make a temporary connection to get the info
            #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
            if (!exists $sc_rec->{$switch}) {
                my $temp_switch_host = host($switch);
                my $temp_switch_conn = connect('#telnet');
                @eport_info = switch_eport_get($temp_switch_conn);
                disconnect($temp_switch_conn);
            } else {
                @eport_info =  switch_eport_get($sc_rec->{$switch});
            }

            if (@eport_info == 0) {
                tstlog(-comment,
                   "ISL Config problem: No E port is configured for $switch");
                tstlog(-fatal, "Test FAILED");
            } else {
                my $entry_number = 1;
                foreach my $eport_info (@eport_info) {
                    my %rec  = ();
                    my %rec1 = ();

                    $rec{'switch'}        = $switch;
                    $rec{'port'}          = $eport_info->{'port'};
                    $rec{'port_WWN'}      = $eport_info->{'port_WWN'};
                    $rec{'peer_port_WWN'} = $eport_info->{'peer_port_WWN'};
                    $rec{'peer_switch'}   = undef;
                    $rec{'peer_port'}     = undef;

                    # Find the remote switch connected to this port
                    foreach my $remote_switch
                            (@{$filer_info->{$filer}{'remoteswitches'}}) {
                        #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
                        #Check if the this switch is mentioned in command line
                        #if not then make a temporary connection to get the info
                        #- - - - - - - - - - - - - - - - - - - - - -  - - - - -
                        my $peer_port = undef;
                        if (!exists $sc_rec->{$remote_switch}) {
                            my $temp_switch_host = host($remote_switch);
                            my $temp_switch_conn = connect('#telnet');
                            $peer_port = switch_eport_num_get(
                                        $temp_switch_conn,
                                        $eport_info->{'peer_port_WWN'},
                                        $entry_number);
                            disconnect($temp_switch_conn);
                        } else {
                            $peer_port = switch_eport_num_get(
                                        $sc_rec->{$remote_switch},
                                        $eport_info->{'peer_port_WWN'},
                                        $entry_number);
                        }

                        if (defined($peer_port)) {
                            $rec{'peer_switch'} = $remote_switch;
                            $rec{'peer_port'}   = $peer_port;
                        }
                    }
                    push (@{$isl_conn_list}, \%rec);

                    # -  - - - - - - - - - - - - - - - - -  - - - - - -
                    # Now reverse the switch and peer switch info in the hash
                    # so that lateron the ISL port disable enable can be
                    # performed  by looking at the "switch" and "port"
                    # combination in the array  of hash. Equivalent of putting
                    # entry for the remote switch.
                    # -  - - - - - - - - - - - - - - - - -  - - - - - -
                    $rec1{'switch'}        = $rec{'peer_switch'};
                    $rec1{'port'}          = $rec{'peer_port'};
                    $rec1{'port_WWN'}      = $rec{'peer_port_WWN'};
                    $rec1{'peer_port_WWN'} = $rec{'port_WWN'};
                    $rec1{'peer_switch'}   = $rec{'switch'};
                    $rec1{'peer_port'}     = $rec{'port'};

                    push (@{$isl_conn_list}, \%rec1);

                    ++$entry_number;
                } #eport_info
            }
        } #switch local

    } #metro
    } #ISL

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Intialize the data structure used for executing the test
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $temp1 (@{$isl_conn_list}) {
        push (@{$switch_info->{$temp1->{'switch'}}{'isl_portlist'}},
                                              $temp1->{'port'});
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Disconnect all switch connections
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    for (my $i = 0; $i < ${$tot_switches}; ++$i) {
        switch_logout($sc_rec->{$switch_list->[$i]});
        disconnect($sc_rec->{$switch_list->[$i]});
    }

    sleep 10; #just to make sure that switch  conns are cleaned up properly

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # print the topology
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    vseries_print_test_setup($fc_rec, $sc_rec, $filer_info, $switch_info, $storage_info,
                             $hba_conn_list, $storage_conn_list,
                             $heartbeat_conn_list, $isl_conn_list);

}   # end of init function

# -  - - - - - - - - - - - - - - - - -  - - - - - -
# Subroutine to verify the current and original number of primary paths
# -  - - - - - - - - - - - - - - - - -  - - - - - -

sub vseries_verify_prim_no_paths($$%) {
    my ($filer, $conn, %filer_info) = @_;

    my $original_no_pri_paths = $filer_info{$filer}{'no_prim_paths'};
    my $current_no_pri_paths  = vseries_no_primary_path_get($conn);

    if ($current_no_pri_paths != $original_no_pri_paths) {
        tstlog (-comment, "Primary LUN count mismatch");
        tstlog (-comment, "The affected filer is: $filer");
        tstlog (-comment, "Current LUN count = $current_no_pri_paths");
        tstlog (-comment, "Original LUN count = $original_no_pri_paths");
        return (1);
    }
}

# -  - - - - - - - - - - - - - - - - -  - - - - - -
# Subroutine to verify the current and original number of secondary paths
# -  - - - - - - - - - - - - - - - - -  - - - - - -

sub vseries_verify_sec_no_paths($$%) {
    my ($filer, $conn, %filer_info) = @_;

    my $original_no_sec_paths = $filer_info{$filer}{'no_sec_paths'};
    my $current_no_sec_paths  = vseries_no_redundant_path_get($conn);

    if ($current_no_sec_paths != $original_no_sec_paths) {
        tstlog (-comment, "Secondary LUN count mismatch");
        tstlog (-comment, "The affected filer is: $filer");
        tstlog (-comment, "Current LUN count = $current_no_sec_paths");
        tstlog (-comment, "Original LUN count = $original_no_sec_paths");
        return (1);
    }
}

# -  - - - - - - - - - - - - - - - - -  - - - - - -
# Subroutine to get local filer of a switch, needed only for Metrocluster
# -  - - - - - - - - - - - - - - - - -  - - - - - -

sub vseries_local_filer_of_switch_get($$) {
    my ($switch, %filer_info) = @_;

    foreach my $filer (keys %filer_info) {
        my @local_switchlist = @{$filer_info{$filer}{'localswitches'}};
        foreach my $local_switch (@local_switchlist) {
            if ($switch eq $local_switch) { return ($filer); }
        }
    }
    return (undef);
}

# -  - - - - - - - - - - - - - - - - -  - - - - - -
# Subroutine to find list of storage slots connecting given switch
# and filer.
# -  - - - - - - - - - - - - - - - - -  - - - - - -

sub vseries_storage_slotlist_connecting_switch_and_filer($$$) {
    my ($filer, $switch, @hba_conn_list) = @_;

    my (@slotlist) = ();
    my $i = 0;
    foreach my $conn (@hba_conn_list) {
        if (($conn->{'filer'} eq $filer) && ($conn->{'switch'} eq $switch)) {
            $slotlist[$i] = $conn->{'filer_slot'};
            ++$i;
        }
    } #foreach
    return @slotlist;
}

# -  - - - - - - - - - - - - - - - - -  - - - - - -
# Subroutine to print the topology (testbed info)
# -  - - - - - - - - - - - - - - - - -  - - - - - -

sub vseries_print_test_setup() {

    my ($fc_rec, $sc_rec, $filer_info, $switch_info, $storage_info,
        $hba_conn_list, $storage_conn_list, $heartbeat_conn_list,
        $isl_conn_list) = @_;

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Filers(s) info: The following prints are just for debugging purpose
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    print "- - - - - - - - - - - - - - - - - - -  - - - \n";
    print "\n";
    print "Filer(s) Information:\n";
    foreach my $filer (keys %{$fc_rec}) {
        print "- - - - - - - - - - - - - - - - - - -  - - - \n";
        print "$filer:model name: $filer_info->{$filer}{\"model name\"}\n";
        print "$filer:firmware release: $filer_info->{$filer}{\"firmware release\"}\n";
        print "$filer:software release: $filer_info->{$filer}{\"software release\"}\n";
        print "$filer:clustered: $filer_info->{$filer}{\"clustered\"}\n";
        print "$filer:partner: $filer_info->{$filer}{\"partner\"}\n";
        print "$filer:cluster_type: $filer_info->{$filer}{\"cluster_type\"}\n";
        print "$filer:cluster slot list: @{$filer_info->{$filer}{\"cluster_slotlist\"}}\n";
        print "$filer:localswitches: @{$filer_info->{$filer} {\"localswitches\"}}\n";
        print "$filer:remoteswitches: @{$filer_info->{$filer} {\"remoteswitches\"}}\n";
        print "$filer:no_prim_paths: $filer_info->{$filer}{\"no_prim_paths\"}\n";
        print "$filer:no_sec_paths: $filer_info->{$filer}{\"no_sec_paths\"}\n";
        print "$filer:storage slot list: @{$filer_info->{$filer} {\"storage_slotlist\"}}\n";
        print "$filer:storage directly connected: $filer_info->{$filer}{\"storagearrays_are_directly_connected\"}\n";
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Switch(s) info: The following prints are just for debugging purpose
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    print "\n";
    print "Switch(s) Information:\n";
    foreach my $switch (keys %{$sc_rec}) {
        print "- - - - - - - - - - - - - - - - - - -  - - - \n";
        print "$switch:vendor: $switch_info->{$switch}{\"vendor\"}\n";
        print "$switch:model: $switch_info->{$switch}{\"model\"}\n";
        print "$switch:firmware: $switch_info->{$switch}{\"firmware\"}\n";
        print "$switch:hba port list: @{$switch_info->{$switch}{\"hba_portlist\"}}\n";
        print "$switch:storage port list: @{$switch_info->{$switch}{\"storage_portlist\"}}\n";
        print "$switch:cluster port list: @{$switch_info->{$switch}{\"cluster_portlist\"}}\n";
        print "$switch:isl port list: @{$switch_info->{$switch}{\"isl_portlist\"}}\n";
    }

    print "\n";
    print "Storage Array(s) Information:\n";
    print "- - - - - - - - - - - - - - - - - - -  - - - \n";
    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Storage(s) info: The following prints are just for debugging purpose
    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    foreach my $vendor (keys %{$storage_info}) {
        print "Vendor and Models $vendor: @{$storage_info->{$vendor}}\n";
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # HBA ports: The following prints are just for debugging purpose
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    print "\n";
    print "Switch(s) HBA ports Information:\n";
    foreach my $temp1 (@{$hba_conn_list}) {
        print "- - - - - - - - - - - - - - - - - - -  - - - \n";
        print "HBA Filer is: $temp1->{\"filer\"}\n";
        print "HBA Filer Slot is: $temp1->{\"filer_slot\"}\n";
        print "Switch is: $temp1->{\"switch\"}\n";
        print "Switch port is : $temp1->{\"switch_port\"}\n";
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Storage ports: The following prints are just for debugging purpose
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    print "\n";
    print "Switch(s) Storage ports Information:\n";
    foreach my $temp1 (@{$storage_conn_list}) {
        print "- - - - - - - - - - - - - - - - - - -  - - - \n";
        print "Storage Switch: $temp1->{\"switch\"}\n";
        print "Storage Switch Port Num : $temp1->{\"port_num\"}\n";
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Heartebat (Cluster) ports: Following prints are just for debugging purpose
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    print "\n";
    print "Switch(s) Cluster ports Information:\n";
    foreach my $temp1 (@{$heartbeat_conn_list}) {
        print "- - - - - - - - - - - - - - - - - - -  - - - \n";
        print "Cluster Filer: $temp1->{\"filer\"}\n";
        print "Cluster Filer Slot: $temp1->{\"filer_slot\"}\n";
        print "Cluster Type: $temp1->{\"cluster_type\"}\n";
        print "Cluster Link: $temp1->{\"link\"}\n";
        print "Switch: $temp1->{\"switch\"}\n";
        print "Switch Port: $temp1->{\"switch_port_num\"}\n";
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # ISL ports: The following prints are just for debugging purpose
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    print "\n";
    print "Switch(s) ISL ports Information:\n";
    foreach my $temp1 (@{$isl_conn_list}) {
        print "- - - - - - - - - - - - - - - - - - -  - - - \n";
        print "ISL Switch: $temp1->{\"switch\"}\n";
        print "ISL Switch Port Num : $temp1->{\"port\"}\n";
        print "ISL Switch Port WWN: $temp1->{\"port_WWN\"}\n";
        print "ISL Peer Switch : $temp1->{\"peer_switch\"}\n";
        print "ISL Peer Switch Port : $temp1->{\"peer_port\"}\n";
        print "ISL Peer Switch Port WWN: $temp1->{\"peer_port_WWN\"}\n";
    }
    print "- - - - - - - - - - - - - - - - - - -  - - - \n";
}

sub vseries_power_cycle_servertech {
    #
    # This sub will power on/off/reboot any plugs attached to
    # ServerTech power strips.  This sub does not care what is on
    # the other end (eg whether it is a filer, switch, blender, whatever).
    #
    # Arguments:
    #
    #        $action - on | off | reboot
    #        $pdu    - name of pdu to use
    #        @ports  - power ports on pdu to use
    #

    my ($action, $pdu, $ports, $username, $password) = @_;

    my $timeout  = 10;
    my $username = defined($username) ? $username : 'admn';
    my $password = defined($password) ? $password : 'admn';

    if(!$pdu || (@$ports == 0)) {
        logresult('SCRIPT', 'abort' => 1, 'You must set up a power strip, and the ports for each filer.');
    }

    my $exp = Expect->spawn("telnet $pdu");

    $exp->expect($timeout,  [ qr/username: /i, sub { my $self = shift;
                             $self->send("$username\n");
                             exp_continue; }
                    ],
                [ qr/password: /i, sub { my $self = shift;
                             $self->send("$password\n");
                             exp_continue; }
                ],
                [ qr/(Switched CDU:|Sentry:)/, sub { my $self = shift;
                                     foreach my $port (@$ports)
                                     {
                                    $self->send("$action $port\n");
                                    $self->expect($timeout, 'Command successful')
                                    or logresult('FATAL', 'abort' => 1,
                                    "Unsuccessful cmd: $action for port: $port on pdu: $pdu");
                                     }
                                   }
                ],
                [ timeout => sub { logresult('FATAL', 'abort' => 1, "Unable to login to pdu: $pdu"); } ]
    );

    $exp->soft_close();
}

sub vseries_power_cycle_rlm {
    #
    # This sub will power on/off/reboot any plugs attached to
    # ServerTech power strips.  This sub does not care what is on
    # the other end (eg whether it is a filer, switch, blender, whatever).
    #
    # Arguments:
    #
    #        $action - on | off | reboot
    #        $rlm    - name of rlm
    #

    my ($action, $rlm, $username, $password) = @_;

    my $timeout  = 10;
    my $username = defined($username) ? $username : 'naroot';
    my $password = defined($password) ? $password : '';
    my @output = () if ($action eq "status");

    if(!$rlm) { logresult('SCRIPT', 'abort' => 1, 'You must supply a filer RLM name.'); }

    if ($action ne "version") {
       if(lc($action) eq 'reboot') { $action = 'cycle'; }

       $action = "system power $action";
    }

    my $cmd = "ssh -n $username" . '@' . "$rlm \"$action\"";

    my $exp = Expect->spawn($cmd);

    $exp->expect($timeout, [ qr/password: /i, sub { my $self = shift;
                            $self->send("$password\n");
                            exp_continue; }
                   ],
                   [ qr/REMOTE HOST IDENTIFICATION HAS CHANGED/, sub { logresult('FATAL', 'abort' => 1,
                       "Unable to ssh into RLM: $rlm because its remote identification has changed.  Check your ~/.ssh/known_hosts file"); } ],
                   [ qr/ssh: .+: host\/servname not known/,
                    sub { logresult('FATAL', 'abort' => 1, "Failure encountered attempting to log into RLM: $rlm"); }
                   ],
                   [ qr/Permission denied/, sub {
                       logresult('FATAL', 'abort' => 1, "Unable to log into RLM: $rlm as user: $username with password: $password"); } ],
                   [ qr/Are you sure you want to continue connecting/, sub { my $self = shift;
                       $self->send("yes\n");
                       exp_continue; }
                   ],
                   [ qr/RLM .+>/, sub { my $self = shift; }
                   ],
                   #[ timeout => sub { logresult('FATAL', 'abort' => 1, "Timeout attempting to log into RLM: $rlm"); } ]
    );

    if ($action =~ m/status/) {
      @output = $exp->before();
    }

    $exp->soft_close();

    return @output if ($action =~ m/status/);
}

sub vseries_power_cycle {
    #
    # Generic wrapper for vseries_power_cycle_servertech
    # and vseries_power_cycle_rlm
    #

    my ($type) = shift(@_);

    if($type eq POWER_CYCLE_TYPE_POWERSTRIP()) {
        vseries_power_cycle_servertech(@_);
    } elsif($type eq POWER_CYCLE_TYPE_RLM()) {
        vseries_power_cycle_rlm(@_);
    } elsif($type eq POWER_CYCLE_TYPE_CLI()) {
        my ($host_conn) = @_;

        say($host_conn, 'reboot -t 0', '-blind');
    } elsif($type eq POWER_CYCLE_TYPE_PANIC()) {
        my ($host_conn) = @_;

        filer::filer_panic($host_conn);
    }
}

sub vseries_object_is_up {
    #
    # If this is just a switch or a single filer, a
    # simple ping should be good enough.  If this is
    # a cluster, then...maybe telnet in and see if state is in takeover mode
    #

    my ($system, $action) = @_;

    my $timeout = 30;
    my $alive   = 1;

    if($action eq 'reboot') { $action = 'on'; }

    tstlog('-comment', "Attempting to determine if $system is $action ...");

    if($action eq 'on') {
        #
        # If we're turning something on, just ping it.
        #
        if(host_is_alive($system, $timeout)) { $alive = 1; }
        else                                 { $alive = 0; }
    } elsif($action eq 'off') {
        #
        # If we're turning something off, try to telnet
        # into it. Everything but one head of a cluster
        # should refuse the connection if it's off. Otherwise
        # look for the cluster head to say that it's running
        # in failover mode.
        #
        my $exp = Expect->spawn("telnet $system");

        $exp->expect($timeout,  [ qr/Running in failover mode/, sub { $alive = 0; } ],
                    [ qr/Unable to connect/, sub { $alive = 0; } ],
                    [ timeout => sub { $alive = 0; } ] );

        $exp->soft_close();
    } else {
        logresult('FATAL', 'abort' => 1, "Incorrect action: $action specified for system: $system");
    }

    if($alive) { tstlog('-comment', "$system is alive");     }
    else       { tstlog('-comment', "$system is NOT alive"); }

    return($alive);
}

sub vseries_compare_arrays {
    #
    # This function will compare two arrays and return
    # true if they are identical, false if they're not.
    #
    # Note that this uses string comparison for now...will
    # update function with numeric comparison later...
    #
    # Pass in references to the 2 arrays you want to be compared
    #

    my ($array1, $array2) = @_;

    if(@$array1 != @$array2) { return(0); }

    for(my $x = 0; $x < @$array1; $x++) {
        if($array1->[$x] ne $array2->[$x]) { return(0); }
    }

    return(1);
}

## @name ::vseries_storage_port_count
## @description Returns the number of storage ports for all filers in \@conn
## It ignores Native Disks which are Netapp Disks that typically look like:
## 0d.27 or 0a.18.  These are not associated with storage ports because
## they are not part of some FC switch connected to some 3rd party array.
##
## @synopsis
## vseries_storage_port_count {\@conn}
##
## @arg \@conn Required: Array of vseries head DNS names in the cluster
##
## @author = silva@netapp.com
##

sub vseries_storage_port_count {

    my ($filer_conn_list) = @_;           # list of filer connections passed in

    my @storage_port_list = ();
    my $port_count = 0;

    my $p_lun = '';                       # A primary lun.
    my $s_lun = '';                       # A secondary lun.
    my $tmp_sw_port = '';                 # A placeholder for switch:port's.

    my @primary_luns = ();
    my @secondary_luns = ();
    my @tmp_prim_luns = ();               # Temporary list of prim luns.
    my @tmp_sec_luns = ();                # Temporary list of sec luns.

    my %seen = ();                        # Hash used to eliminate duplicates
    my @uniq = ();                        # in @storage_port_list which becomes
                                          # @uniq.

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Populate @storage_port_list with the output of the 'storage show
    # disk -p' filer command for all the filers in $filer_conn_list.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $filer_conn (@{$filer_conn_list}){
        my $host = host($filer_conn);
        my $password = $host->default_password();
        my $cmd = "rsh $filer_conn -l root:$password ";
        $cmd .='storage show disk -p';
        # This may need "say" to make that work
        #my (undef, undef, $cmd_output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
        my @output = `$cmd`;

        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Now eliminate the first 2 lines of output produced from the
        # 'storage show disk -p' command.
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        @output = @output[2 .. $#output]; # (Programming Perl (2nd edition) page 49)

        foreach my $tmp (@output){
            push @storage_port_list, $tmp;
        }
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # @storage_port_list is a list of all the primary/secondary
    # LUNS combined for all filers.  It DOES NOT contain Netapp
    # native disks.
    #
    # Example:
    # 0b.19                   A                                  1    3
    # vss_b4100_1:7.126L8     -    vss-m6140-1:3.127L8     -     -    -
    # vss_b4100_1:7.126L16    -    vss-m6140-1:3.127L16    -     -    -
    # vss_b4100_1:7.126L24    -    vss-m6140-1:3.127L24    -     -    -
    # vss_b4100_1:7.126L36    -    vss-m6140-1:3.127L36    -     -    -
    #
    # -  - - - - - - - - - - - - - - - - -  - - - - - -


    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now populate @primary_luns and @secondary_luns.  If we find a
    # Netapp Native disk, then don't add it to primary or secondary
    # lun arrays.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $tmp2 (@storage_port_list){

        ($p_lun, undef, $s_lun, undef, undef, undef) = split(/\s+/, $tmp2);

        if ($p_lun =~ /$NTAP_DISK/) { next; }
        push (@primary_luns, $p_lun);
        push (@secondary_luns, $s_lun);

    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # @primary_luns and @secondary_luns arrays have now been populated
    # with all primary & secondary luns.  Now we need to further carve
    # up each lun down to the switch_name:storage_port_number.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # First lets deal with the primary LUNS:
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $tmp3 (@primary_luns){

        ($tmp_sw_port, undef) = split(/\./, $tmp3);
        push @tmp_prim_luns, $tmp_sw_port;
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now for the secondary LUNS:
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $tmp4 (@secondary_luns){

        ($tmp_sw_port, undef) = split(/\./, $tmp4);
        push @tmp_sec_luns, $tmp_sw_port;
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now we need to clear the old contents of @storage_port_list
    # and repopulate it.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    @storage_port_list = ();

    foreach my $x (@tmp_prim_luns){
        push @storage_port_list, $x;
    }

    foreach my $y (@tmp_sec_luns){
        push @storage_port_list, $y;
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now lets convert @storage_port_list into a unique list of
    # switch:storageports.  Walk the through all entrys in
    # @switch_port_list and eliminate duplicates.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $item (@storage_port_list){
        push (@uniq, $item) unless $seen{$item}++;  #(Recipe 4.6 page 102 Perl Cookbook)
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # port_count is equal to the number of elements in @uniq
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    $port_count = scalar(@uniq);

    return ($port_count);
}

## @name ::vseries_is_storage_port
## @description
## Returns 1 if this is one of $filer's storage ports, otherwise
## it will return 0
##
## @synopsis
## vseries_is_storage_port {$filer, $switch, $port}
##
## @arg $filer: is a given v-series head.
##      $switch: is a given fibre channel switch.
##      $port: is a given & potential fibre channel disk port.
##
## @author = silva@netapp.com
##

sub vseries_is_storage_port {

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # set our variables to what was passed in at the command line
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    my ($filer, $switch, $port) = @_;

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # create the command $cmd to be executed at the command line
    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    my $host = host($filer);
    my $password = $host->default_password();
    my $cmd = "rsh $filer -l root:$password ";
    $cmd .='storage show disk -p';

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # $output is set to the resulting output from `$cmd`
    # which looks something like this (except all one long string):
    #
    # vss_b4100_1:7.126L8     -    vss-m6140-1:3.127L8     -     -    -
    # vss_b4100_1:7.126L16    -    vss-m6140-1:3.127L16    -     -    -
    # vss_b4100_1:7.126L24    -    vss-m6140-1:3.127L24    -     -    -
    # vss_b4100_1:7.126L36    -    vss-m6140-1:3.127L36    -     -    -
    #
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    my $output = `$cmd`;

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # if we find our $switch:$port in $output then we know
    # that this is one of our storage ports, so return true (1),
    # otherwise return false (0).
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    if ($output =~ /$switch:$port/){
        return (1);
    }
    else{
        return (0);
    }
}

## @name ::vseries_storage_port_associate
## @description Returns a reference to a hash of filer names and their
## associated switch:port's.
##
## @synopsis
## vseries_storage_port_associate {\@filer_list}
##
## @arg \@filer_list Required: Array of filer names.
##
## @author = silva@netapp.com
##

sub vseries_storage_port_associate {

    my ($filer_list) = @_;                # List of filer connections passed in
    my $filera = @$filer_list[0];         # Set $filera and $filerb
    my $filerb = @$filer_list[1];
    my %storage_port_list = ();           # Hash used to track associated storage ports
    my @output_filera;                    # Output of storage show disk -p to filera
    my @output_filerb;                    # Output of storage show disk -p to filerb
    my @filera_lun_paths = ();            # List of all paths to luns
    my @filerb_lun_paths = ();            # (secondary/primary) for filera and filerb
    my $sw_port = '';                     # Temporary holder for switch ports used in loops
    my %seen_in_filera = ();              # Hash used to help build an array of uniq
    my @uniq_in_filera = ();              # Storage ports in filera '@uniq_in_filera'
    my %seen_in_filerb = ();              # Hash used to help build an array of uniq
    my @uniq_in_filerb = ();              # Storage ports in filerb '@uniq_in_filerb'


    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Set @output_filera/b with the proper output
    # for future processing.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    my $host = host($filera);
    my $password = $host->default_password();
    my $cmd_a = "rsh $filera -l root:$password ";
    $cmd_a .='storage show disk -p';
    @output_filera = `$cmd_a`;

    my $host = host($filerb);
    my $password = $host->default_password();
    my $cmd_b = "rsh $filerb -l root:$password ";
    $cmd_b .='storage show disk -p';
    @output_filerb = `$cmd_b`;

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now eliminate the first 2 lines of output produced from the
    # 'storage show disk -p' command.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    @output_filera = @output_filera[2 .. $#output_filera]; # (Programming Perl
    @output_filerb = @output_filerb[2 .. $#output_filerb]; # 2nd edition page 49)

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now @output_filera/b is a list of all the primary/secondary
    # LUNS combined for each specific filer.
    #
    # Example:
    #
    # vss_b4100_1:7.126L8     -    vss-m6140-1:3.127L8     -     -    -
    # vss_b4100_1:7.126L16    -    vss-m6140-1:3.127L16    -     -    -
    # vss_b4100_1:7.126L24    -    vss-m6140-1:3.127L24    -     -    -
    # vss_b4100_1:7.126L36    -    vss-m6140-1:3.127L36    -     -    -
    #
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now populate @filera_lun_paths and @filerb_lun_paths
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    my ($p_lun, $s_lun) = '';

    foreach my $tmp1 (@output_filera){
        ($p_lun, undef, $s_lun, undef, undef, undef) = split(/\s+/, $tmp1);
        push @filera_lun_paths, $p_lun;
        push @filera_lun_paths, $s_lun;
    }

    foreach my $tmp2 (@output_filerb){
        ($p_lun, undef, $s_lun, undef, undef, undef) = split(/\s+/, $tmp2);
        push @filerb_lun_paths, $p_lun;
        push @filerb_lun_paths, $s_lun;
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # @filera_lun_paths and @filerb_lun_paths now contain
    # all of the primary and secondary luns from the result
    # of the 'storage show disk -p' command.  Now we need to
    # further carve up each lun path down to the
    # switch_name:storage_port_number level.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Lets clear @output_filera and @output_filerb
    # so we may reuse it and not create additional arrays.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    @output_filera = ();
    @output_filerb = ();

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Splitting on \. splits the switch:storage_port.lun_number
    # so that you are left with two pieces, switch:storage_port
    # and lun_number.  Lets do that for both vseries heads.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $tmp3 (@filera_lun_paths){
        ($sw_port, undef) = split(/\./, $tmp3);
        push @output_filera, $sw_port;
    }

    foreach my $tmp4 (@filerb_lun_paths){
        ($sw_port, undef) = split(/\./, $tmp4);
        push @output_filerb, $sw_port;
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now we need to eliminate all duplicates in @output_filera
    # and @output_filerb.  We will use %seen_in_filera/b to do this,
    # then send the uniq elements to @uniq_in_filera/b
    #
    # The foreach blocks ahead are explained here:
    # (Recipe 4.6 page 102 Perl Cookbook)
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $item1 (@output_filera){
        push (@uniq_in_filera, $item1) unless $seen_in_filera{$item1}++;
    }

    foreach my $item2 (@output_filerb){
        push (@uniq_in_filerb, $item2) unless $seen_in_filerb{$item2}++;
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now construct the %storage_port_list hash and return a reference to it.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    %storage_port_list = (
        $filera => [@uniq_in_filera],
        $filerb => [@uniq_in_filerb],
    );

    return (\%storage_port_list);
}

## @name ::vseries_hba_port_disable_enable
##
## @description
## Disables the list of hba ports for the given filer passed in,
## then enables them.
##
## @synopsis
## vseries_hba_port_disable_enable {$conn, \@hba_portlist}
##
## @arg $filer Required: A filers DNS name.
## @arg \@hba_portlist Required: Array of hba ports.
##
## @author = silva@netapp.com
##

sub vseries_hba_port_disable_enable {

    my ($conn, $hba_list_ref) = @_;
    my @hba_list = @$hba_list_ref;

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # If we find zero hba_ports exit with return code 1
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    if (@hba_list == 0){
        return(1);
    }

    my $after = "";
    my $match = "";
    my $timeout = 0;

    foreach my $hba_port (@hba_list){

        $timeout = time() + 120;

        while($match !~ /disable succeeded/) {
            (undef, undef, $match) = Tharn::say($conn, "storage disable adapter $hba_port", -swallow, '-timeout', 15);

            # -  - - - - - - - - - - - - - - - - -  - - - - - -
            # Check if time expired, if so return(1) for error.
            # -  - - - - - - - - - - - - - - - - -  - - - - - -

            if ((time() > $timeout) && ($match !~ /disable succeeded/)){
                return(1);
            }
            sleep(5);
        }


        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Sleep for 35 seconds in between disable enable so the filer acknowlages the command.
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        sleep(35);   # need 35 seconds here see burt 242330

        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Clear $match so we don't do a false match in the loop.
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        $match = '';

        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Enable the hba port..
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        Tharn::say($conn, "storage enable adapter $hba_port");

        my $secondary = '-';

        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Reset the timeout.
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        $timeout = time() + 120;  # this way we have plenty of time to come back up.

        my @temp = ();
        my $line = '';

        while($secondary =~ /^-$/) {
            (undef, undef, $match) = Tharn::say($conn, "storage show disk -p all", -swallow, '-noverbose');

            # -  - - - - - - - - - - - - - - - - -  - - - - - -
            # $line now equals the first line of paths from storage show disk -p all
            # -  - - - - - - - - - - - - - - - - -  - - - - - -

            @temp = split(/\n/, $match);
            $line = $temp[2];

            (undef, undef, $secondary) = split(/\s+/, $line);

            # -  - - - - - - - - - - - - - - - - -  - - - - - -
            # Check if time expired, if so return(1) for error.
            # -  - - - - - - - - - - - - - - - - -  - - - - - -

            if ((time() > $timeout) && ($secondary =~ /^-$/)){
                return(1);
            }
            sleep(5);
        }
    }
    return(0);
}

## @name ::vseries_fclink_state_display
## @description Prints the output of fcadmin link_state to the NATE_LOGS.
##
## @synopsis
## vseries_fclink_state_display {$conn}
##
## @arg $conn Required: The vseries connection to send this operation over
##
## @author: silva@netapp.com
##
sub vseries_fclink_state_display ($) {
    my ($conn) = @_;

    say($conn, "priv set -q advanced;
               fcadmin link_state", -swallow, '-timeout', 60);

    return (0);
}

1;
