# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/VSeries/client_object.pm#1 $
#
## @summary client object package - creates an OO model of a (UNIX) test client
##  node. Right now we're only supporting Unix OS's. Windows client object 
##  support is TBD
##
## @description
##   This package creates client objects and contains methods to support POSIX 
##   commands executed on clients as well as a general miscellaneous method to
##   execute whatever command is passed to the client.
##   This package is not meant to support _ALL_ known Unix commands but 
##   exists to provide support for commands required by specific tests. New
##   command support via new methods is added in on an 'as needed' basis
##
##   Requirements for adding new methods:
##   1. methods should be one of two types: 
##       a. a method that reads or obtains system information (ie. ls, df)
##          This type of method should return a buffer of the command output
##       b. a method that modifies something on the system (i.e mkdir, rmdir)
##          This type of method must not only execute the command but also 
##          verify it successfuly performed its task(s) (either via command 
##          return status or additional commands to verify data) and return 
##          a status code to the calling program
##    Limitations:
##      Verifying a command with an option performed successfully can get VERY
##      messy and involve a lot of work parsing/tokenizing the input paramters 
##      and making decisions for a command's various options. We suggest that 
##      the verification be VERY simple in the method, leaving the calling 
##      program to make the most of the error parsing and handling decisions.
##
## @dependencies
##   use strict;
##   use Tharn;
##   use Expect;
##
## @usage
##   use VSeries::client_object;
##
## @author Mark Bozen (mbozen@netapp.com), dl-vseries-automation;
##

package VSeries::client_object;

use strict;
use Tharn;
use Expect;

use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT = ('new',                    # Create an object.
           'set_internal_variables', # Set an object's data members.
           'create_dir',             # mkdir command support.
           'remove_dir',             # rmdir command support.
           'get_mount',              # mount command without args
                                     # (return current mount points).
           'set_mount',              # mount command with args to mount
                                     # filesystems.
           'set_unmount',            # umount command with args to unmount
                                     # filesystems.
           'get_client_hostname',  # Returns the hostname of the client
           'send_client_cmd',      # miscellaneous method to execute whatever
                                   #   command is passed in.
           'run_io',       # run i/o tool or exercisser on client
           'parse_client_logs',  # parse specified client logfiles for a 
                                 # specified set of patterns
           'create_dt_patterns', # create io test patterns
           'get_error_state',        # Returns {error} data member.
           'get_error_msg',          # Returns {errmsg} data member.
           'diff_files',              # Diff 2 files - must have access to both.
           'delete_partition',       # Deletes Linux partition of a device.
           );

## GLOBALS
my $GOOD_STATUS    = 0;
my $ERR_STATUS     = 1;
my $COMMAND_TIMEOUT = 60;

## @name ::new()
##
## @description
## Initializes a new client data structure and its data members.
##
## @synopsis
## $client_obj=VSeries::client_object->new();
##
sub new ()
{
    my ($class, %args) = @_;
    my ($self);

    # create a new object only if necessary
    if (defined $args{-self}) {
        $self=$args{-self};
    } else  {
        $self={};
        bless $self,$class;
    }

    # from Tharn::host();
    $self->{default_ip} = "";          # ip address of client
    $self->{hostname} = "";            # dns name of client

    # from Tharn::connect();
    $self->{conntype} = "";            # type of conn i.e. rsh, telnet, etc.
    $self->{connport} = "";            # port num for telnet & serial else undef
    $self->{username} = "";            # username used for login
    $self->{password} = "";            # password used for login
    $self->{timeout} = "";             # timeout found in objects tharnhost file
    $self->{max_reconnect} = "";       # retry say() commands x times
    $self->{conn} = "";                # the full connectrec spec

    # error tracking
    $self->{error} = 0;                # error flag used as a bool
    @{$self->{errmsg}} = ();           # an array that stores cascading errors
    $self->{shell} = "sh";             # specify the shell to execute the
                                       #   command in, default: sh
    $self->{os_info} = "";             # os type and revision

    $self->{umount_flags} = "";        # set flag for umount

    return($self);
}

## @name ::set_internal_variables()
##
## @description
## Parameters are all optional.  Defined as such in order to introduce 
## readability and flexibility moving forward (i.e. you should be able 
## to call this from anywhere with parameters missing and/or in the 
## wrong order.  Errors will occur if a parameter is not labeled 
## correctly.
##
## @synopsis
## $status = client_obj->set_internal_variables(\%args);
##
## @arg %args Required: The actual key value pairs that you wish to 
## change in the $client_obj.
##
sub set_internal_variables(\%)
{
    # pick up our object
    my $self = shift;

    # catch the optional function parameters here
    my %args = @_;

    # parse function parameters that are present, putting into data members
    # could just say "$self->{$arg} = $val", but no check
    while (my ($arg, $val) = each %args)
    {
        if ($arg eq "hostname")
        {
            $self->{hostname} = $val;
        }
        elsif ($arg eq "default_ip")
        {
            $self->{default_ip} = $val;
        }
        elsif ($arg eq "shell")
        {
            $self->{shell} = $val;
        }
        elsif ($arg eq "conntype")
        {
            $self->{conntype} = $val;
        }
        elsif ($arg eq "conn")
        {
            $self->{conn} = $val;
        }
        elsif ($arg eq "connport")
        {
            $self->{connport} = $val;
        }
        elsif ($arg eq "username")
        {
            $self->{username} = $val;
        }
        elsif ($arg eq "password")
        {
            $self->{password} = $val;
        }
        elsif ($arg eq "timeout")
        {
            $self->{timeout} = $val;
        }
        elsif ($arg eq "max_reconnect")
        {
            $self->{max_reconnect} = $val;
        }
        elsif ($arg eq "umount_flags")
        {
            $self->{umount_flags} = $val;
        } else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "Invalid parameter: $arg => $val");
            return ($ERR_STATUS);
        }         
    }
    return ($GOOD_STATUS);
}

## @name ::get_client_hostname()
##
## @description
## Returns the name of a given client object.
##
## @synopsis
## $hostname = client_obj->get_client_hostname();
##
##
sub get_client_hostname(){
    my $self = shift;
    return($self->{hostname});
}


## @name ::create_dir(@)
## @description
## Creates a new directory with specified name
##
## @synopsis
## $status = $client_obj->create_dir(@dir_names)
##
## @arg $dir_names Required: array of one or more directories to be deleted
##
sub create_dir(@)
{
    my ($self, @dir_names) = @_;
    my $after = "";
    my $cmd = "";
    my $dir = "";   

   # Check if directory exists first before we attempt to create it
   foreach $dir (@dir_names) {
     (undef, undef, $after) = Tharn::say($self->{conn}, "file $dir", -swallow, '-timeout', $COMMAND_TIMEOUT);
     if ( $after =~ /:\s+cannot open/ ) {
       # create directory, -p option creates the entire directory root
       $cmd = "mkdir -p $dir";
       (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', $COMMAND_TIMEOUT);

       my $error_stat = _check_for_errs($cmd, $after);
       if  ( $error_stat )  {
          $self->{error} = $ERR_STATUS;
          push(@{$self->{errmsg}}, "ERROR: directory $dir was not created as expected");
          return ($ERR_STATUS);
       }  

       # now verify directory was created
       (undef, undef, $after) = Tharn::say($self->{conn}, "file $dir", -swallow, '-timeout', $COMMAND_TIMEOUT);
       if ( $after =~ /No such file or directory/ ) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: directory $dir does not exist");
            return ($ERR_STATUS);
       }
     }
   }
   return ($GOOD_STATUS);
}

## @name ::remove_dir(@
## @description
## removes a specified directory or file
##
## @synopsis
## $status = $client_obj->remove_dir($dir_name)
##
## @arg $dir_names Required: array of one or more files or directories to be deleted
##
## Assumption: method is not idiot proof and has no conscience about what is 
## deleted
sub remove_dir(@)
{
   my ($self, @dir_names) = @_;
   my $after = "";
   my $cmd = "";
   my $dir = "";   
    
   foreach $dir (@dir_names) {

      # remove directory or file. We don't check that it previously exists since we do a
      # forced remove
      $cmd = "rm -rf $dir";
      (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', $COMMAND_TIMEOUT);

      my $error_stat = _check_for_errs($cmd, $after);
      if  ( $error_stat )  {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: failed to remove directory $dir");
            return ($ERR_STATUS);
      }  

      # now verify directory or file was deleted
      (undef, undef, $after) = Tharn::say($self->{conn}, "file $dir", -swallow, '-timeout', $COMMAND_TIMEOUT);
      if ( $after !~ /\s+cannot open/ ) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: failed to remove directory $dir");
            return ($ERR_STATUS);
      }
   }
   return ($GOOD_STATUS);
}


## @name ::get_mount()
## @description
## Executes the mount command with no options to retrieve current mount points
## 
## @synopsis
## \%mount_cmd_hash=$client_obj->get_mount()
##
## @return: hash reference of mount command
## key: local mount point
## value: array of three elements: source dir or device, mount type, permissions
## 
## command output parsed:
## /dev/sda3 on / type ext3 (rw)
## none on /proc type proc (rw)
## none on /sys type sysfs (rw)
## none on /dev/pts type devpts (rw,gid=5,mode=620)
## automount(pid2218) on /x/sim type autofs (rw,fd=5,pgrp=2218,minproto=2,maxproto=4)
## 10.97.230.68:/vol/vol1 on /mnt/v6040/2a/vol1 type nfs (rw,addr=10.97.230.68)
## 10.97.230.68:/vol/vol2 on /mnt/v6040/2a/vol2 type nfs (rw,addr=10.97.230.68)

sub get_mount()
{
   my $self = shift;
   my $after = "";
   my $line = "";
   my %mount_cmd_hash=();

   # execute mount cmd
   (undef, undef, $after) = Tharn::say($self->{conn}, "mount", -swallow, '-timeout', $COMMAND_TIMEOUT);

   # put Say output (mount command) into an array
   my @mount_cmd_out = split ("\n", $after);

   foreach $line (@mount_cmd_out) {
      my @mount_params = ();
      if ( $line =~ /^(\S+)\s+\w+\s+(\S+)\s+\w+\s+(\S+)\s+\((\S+)\).*\n?$/ ) {
         $mount_params[0] =$1;
         $mount_params[1] = $3;
         $mount_params[2] = $4;
         $mount_cmd_hash{$2} = [ @mount_params ];
      }
   }     

   return (\%mount_cmd_hash);
}   

## @name ::set_mount($)
## @description
## Executes the mount command with options 
##
## @synopsis
## $status = $client_obj->set_mount($)
##
## @arg $cmd_input Required: a string containing ALL options and arguments that 
##                           follow the mount command including disks
##
## Assumption: the syntax and arguments are assumed to be correct (no idiot 
##             proof code checking)
##
sub set_mount($)
{
   my ($self, $cmd_input) = @_;
   my $after = "";
   my $cmd = "";

   # execute mount command
   $cmd = "mount $cmd_input";  
   (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', $COMMAND_TIMEOUT);

   my $error_stat = _check_for_errs($cmd, $after);

   if  ( $error_stat && $after !~ /(already mounted|You have new mail)/i )  {
        $self->{error} = $ERR_STATUS;
        push(@{$self->{errmsg}}, "ERROR: failed to mount filesystem");
        return ($ERR_STATUS);
   }  
           
   return ($GOOD_STATUS);
}   

## @name ::set_unmount(@;)
## @description
## Executes the umount command with options 
##
## @arg @umount_dir_list  Required: an array of all filesystems to be 
##                        unmounted
##
## To set options for the unmount command please set the value using
## set_internal_variables method. 
## Example:-
## -------
## $client_obj->set_internal_variables(umount_flags => '-l');
##              or
## $client_obj->{umount_flags} = '-l';
##
## @synopsis
## $status = $client_obj->set_unmount(@;)
##
## Assumption: the syntax and arguments are assumed to be correct (no idiot 
## proof code here)
##
sub set_unmount(@;)
{
   my ($self, @umount_dir_list) = @_;

   my $after = "";
   my $line = "";
   
   my $cmd_opts = $self->{umount_flags};

   # verify each filesystem was unmounted
   foreach $line (@umount_dir_list) {

      # execute umount cmd, we unmount filesystems in the array one at at time
      (undef, undef, $after) = Tharn::say($self->{conn}, "umount $cmd_opts $line", -swallow, '-timeout', $COMMAND_TIMEOUT);
      
      # Issue the "mount" command and verify the filesystem was unmounted succesfully.
      (undef, undef, $after) = Tharn::say($self->{conn}, "mount", -swallow, '-timeout', $COMMAND_TIMEOUT);

      my @mounted = split ("\n", $after);
      
      foreach my $mount (@mounted) {
          # Sample mount output: 10.227.190.128:/vol/vol0/ on /var/tmp/mounts_vol type nfs (rw,addr=10.227.190.128).
	  # Since either '10.227.190.128:/vol/vol0/' or '/var/tmp/mounts_vol' can be passed by the user as 
	  # the directory-name to be unmounted, we should check for both scenarios.(BURT 560227)
	  if ( ($line =~ m/.*?:\/.*/  && $mount =~ /^$line on/) ||($line !~ m/.*?:\/.*/  && $mount =~ /on $line type nfs/)){
	      $self->{error} = $ERR_STATUS;
	      push(@{$self->{errmsg}}, "failed to unmount directory $line");
	      return ($ERR_STATUS);
	   }
      } 
   }
   return ($GOOD_STATUS);
}   

## @name ::send_client_cmd($;$)
## @description
## Executes whatever command is passed in. Returns command output, whatever
## that may be.
##
## @synopsis
## $status = $client_obj->send_client_cmd($input_cmd)
##
## @arg $input_cmd Required: string containing command to be executed
## @arg $timeout Optional: sets command timeout value Def.aults to $COMMAND_TIMEOUT.
## @return: array output of say
##
## ASSUMPTIONS: since input command can be anything, the parsing of the command 
## output is not done here and is the responsibility of the calling program
## 
sub send_client_cmd($;$)
{
   my ($self, $input_cmd, $timeout) = @_;
   my $after = "";
   my @say_output=();

   my $cmd_timeout = $COMMAND_TIMEOUT;
   $cmd_timeout = $timeout if (defined $timeout);

   # Execute the input command.
   (undef, undef, $after) = Tharn::say($self->{conn}, "$input_cmd", -swallow, '-timeout', $cmd_timeout);

   # Put Tharn::say() method output into an array
   if ( $after =~ /.*\r\n$/ ) {
         @say_output = split ("\r\n", $after);
      } else {
         @say_output = split ("\n", $after);
   }

   # If we're running on an older version of Nate prior to late July '09
   # we need to check if the say() output echoes the command line as part
   # of the output.  If that's the case then strip out the echoed line
   if ( $say_output[0] =~ m/^$input_cmd$/ ) {
       shift @say_output;
   }

   return (@say_output);
}


## @name ::run_io($$)
## @description
## Executes an i/o tool or exerciser
##
## @synopsis
## @pid_array = $client_obj->run_io($io_tool_path, $io_tool_params)
##
## @arg $io_tool_path   Required: absolute path where i/o tool exists
##      $io_tool_params Required: all arguments for i/o tool
##
## ASSUMPTIONS: method won't verify proper syntax or usage of commmand options
##
## @return value: array of pids of all io background processes 
##
# Sample output:
#   89 pts/1    S+     0:00 /u/mnatale/dt/dt.d-WIP/dt of=/mnt/v6040/2a/vol1/dt_file procs=2 bs=16k limit=100m errors=1 runtime=7d enable=compare,lbdata log=2a_vol1
#  514 pts/1    R+    42:41 /u/mnatale/dt/dt.d-WIP/dt of=/mnt/v6040/2a/vol1/dt_file procs=2 bs=16k limit=100m errors=1 runtime=7d enable=compare,lbdata log=2a_vol1
# 9615 pts/1    R+    42:37 /u/mnatale/dt/dt.d-WIP/dt of=/mnt/v6040/2a/vol1/dt_file procs=2 bs=16k limit=100m errors=1 runtime=7d enable=compare,lbdata log=2a_vol1
#49229 pts/2    S+     0:00 /u/mnatale/dt/dt.d-WIP/dt of=/mnt/v6040/2a/vol2/dt_file procs=2 bs=16k limit=100m errors=1 runtime=7d enable=compare,lbdata log=2a_vol2


sub run_io($$)
{
   my ($self, $io_tool_path, $io_tool_params) = @_;
   my $after = "";
   my $cmd = "$io_tool_path $io_tool_params";
   my $line = "";
   my @pid_array = ();

   # ensure io tool or execiser exists on client node and/or is in defined path
   (undef, undef, $after) = Tharn::say($self->{conn}, "which $io_tool_path", -swallow, '-timeout', $COMMAND_TIMEOUT);
   if ( $after =~ /which:\s+no/ ) {
       $self->{error} = $ERR_STATUS;
       push(@{$self->{errmsg}}, "ERROR: $io_tool_path does not exist on client");
       return ($ERR_STATUS);
   }

   # execute the i/o command
   (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd &", -swallow, '-timeout', $COMMAND_TIMEOUT);
   
   # get the pid of the active process(es)
   (undef, undef, $after) = Tharn::say($self->{conn}, "ps axw | grep \"$cmd\"", -swallow, '-timeout', $COMMAND_TIMEOUT);

   my @output_buffer = split ("\n", $after);
   foreach $line (@output_buffer) {
       # skip over any line if it's a process of the command line itself
       if ( $line =~ /grep $cmd/ ) { 
            next;
        } else {
            # We may be running i/o that spawns multiple concurrent processes
            # and/or slices so parse for parent and any child processes and put 
            # them all into the pid array
            if ( $line =~ /^ *(\d+)\s+/ ) {
                push @pid_array, $1;
            }
        }
   }

   return (@pid_array);
}

## @name ::_check_for_errs($$)
## @description
## Internal function only, not exported
## Parses say method output of a unix command and checks if any errors occured
## Kudos to Matt for the parsing logic
##
## @synopsis
## $status = _check_for_errs($$)
##
## @arg $cmd_run    Required: string containing command that was executed
## @arg $say_output Required: string containing say method command output buffer
##
## ASSUMPTIONS: 
##  1) this method supports only commands and options that only
##     return to prompt if the command was successful
##  2) this method is unsuitable for commands run as background processes
##
sub _check_for_errs($$)
{
    my ($cmd_run, $say_output) = @_;
    my $line = "";

    my @output_buffer = split ("\n", $say_output);
    foreach $line (@output_buffer) {
      # we parse for the command executed that is part of any Say method 
      # output buffer 
      if ($line =~ /^${cmd_run}.*\n?$/ ) { 
           next; 
       } else {
         # there is additional text found in the buffer associated with
         # something unexpected occuring so we fail
         return ($ERR_STATUS);
       }
    }
    return ($GOOD_STATUS);
}

# @name ::parse_client_logs($$)
#
# @description
# Parse specified client log files to check for user specified patterns.
# If any patterns are found in the list of log files, that log file
# is specified as a key in the %match_log hash and the hash value will
# be the corresponding logfile absolute path. A reference to %match_log
# hash is returned.
#
# @synopsis
# $log_ref = parse_client_logs($$);
#
# @arg $log_hash_ref    Required: reference to hash of client logs 
#                       that will be parsed. Key is the log file name
#                       the value is the logfile absolute path
# @arg $pattern_arr_ref    Required: reference to array of strings to be 
#                          grep'ed for in log files
#
sub parse_client_logs ($$)
{
    my ($self, $log_hash_ref, $pattern_arr_ref) = @_;

    my @client_cmd = ();
    my $pattern_match = 0;
    my %log_list = %$log_hash_ref;
    my @pattern_arr = @$pattern_arr_ref;    
    my %match_log = (); # a subset of %log_list containing all the
                        # logs found that contained any of the patterns
                        # specified in $pattern_arr

    foreach my $log_file (keys %log_list) {

        foreach my $pattern (@pattern_arr) {
            @client_cmd = $self->send_client_cmd("grep \"$pattern\" $log_list{$log_file}"); 
 
            # Because the @client_cmd array could possibly contain unexpected
            # or asynchronous client console messages captured by say(), we 
            # need to parse the output of send_client_cmd line by line to 
            # match the specific patterns we're looking for. Any errors
            # detected will push the corresponding log file name into a hash.
            foreach my $dtstat (@client_cmd) {             
                if ( $dtstat =~ /$pattern/ ) {
                    $match_log{$log_file} = $log_list{$log_file};
                    $pattern_match = 1;            
                }
            }
        }
    }
    return \%match_log;
}

# @name ::create_dt_patterns($)
#
# @description
# Create 10 static entries for the first 10 volumes. 
# If we have more than 10 then we create unique patterns based 
# on an offset subtracted from a previous pattern. A reference
# to an array of all generated patterns is returned
#
# @synopsis
# $pattern_arr_ref = create_dt_patterns($);
#
# @arg $num_patterns    Required: total number of different test
#                       patterns to create

sub create_dt_patterns ($)
{
    my ($self, $num_patterns) = @_;
    my @pattern_arr = ();
    my $i = 0;

    $pattern_arr[0] = "0xdeadbeef";
    $pattern_arr[1] = "0x0ffabff0";
    $pattern_arr[2] = "0xa0b1c2d3";
    $pattern_arr[3] = "0xffa0bced";
    $pattern_arr[4] = "0xdeaf9876";
    $pattern_arr[5] = "0xfeab7291";
    $pattern_arr[6] = "0x0f0f0f0f";
    $pattern_arr[7] = "0x2729eadf";
    $pattern_arr[8] = "0x83629ad2";
    $pattern_arr[9] = "0xabcdef00"; 

    if ( ($num_patterns + 1) > 10 ) {
        $i = 10;
        while ( $i < ($num_patterns + 1) ) {
            push @pattern_arr, ($pattern_arr[($i-10)] - 0xffff);
            $i++;
        }
    }

    return \@pattern_arr;
}

## @name ::get_error_state()
##
## @description
## Returns the value of the error flag
##
## @synopsis
## $err_state = VSeries::client_object->get_error_state();
##
##
sub get_error_state(){
    my $self = shift;
    return $self->{error};
}

## @name ::get_error_msg()
##
## @description
## Returns the error messages for a given client.  This is useful for reporting
## failures or unexpected behavior for a given test.
##
## @synopsis
## @err_array = VSeries::client_object->get_error_msg();
##
##
sub get_error_msg(){
    my $self = shift;
    return @{$self->{errmsg}};
}

## @name ::diff_files($$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Uses the posix 'diff' to diff one file against another.
## If differences exist simply return $ERR_STATUS, otherwise
## return $GOOD_STATUS. Note, $file1 and $file2 should be
## absolute file paths, a filename alone will not suffice.
## The client used should also have access to both absolute
## paths, otherwise diff will error causing this method to error.
## We don't set the {error} flag or {errmsg} error holder
## because it may not be a bad thing if you have differences.
##
## @synopsis
## $status = VSeries::client_object->diff_files($cat $hat);
##
## @arg $file1  Required: File to diff against $file2 
## @arg $file2  Required: File to diff against $file1
##
sub diff_files($$){
    my ($self, $file1, $file2) = @_;
    my @diff_output = (); # Used to hold output from send_client_cmd().
    

    # Diff (on the client) $file1 against $file2.  If @diff_output
    # contains any elements that is an indication of failure.
    @diff_output = $self->send_client_cmd("diff $file1 $file2");

    # If @diff_output is greater than zero there were differences.
    if (scalar(@diff_output) > 0) {
        return($ERR_STATUS);
    }

    return($GOOD_STATUS);
}

## @name ::delete_partition($)
##
## @author Aditya Siddavatam (aditya1@netapp.com)
##
## @description
## Deletes linux partition of a device.
##
## @args $device       Required: It is required to delete partition of a device.
##
## @synopsis
## &delete_partition($device);
##
sub delete_partition($) {
   my ($self, $device) = @_;

   #
   # Creating a input file for the fdisk.
   #
   my $cmd = 'printf "d\nw\n"' . " > /var/tmp/part.txt";
   my $cmd_exit_status = ($self->send_client_cmd("$cmd; echo \$?"))[-1];
   return $ERR_STATUS if ($cmd_exit_status != 0);

   #
   # Deleting the partition.
   #
   $cmd = "fdisk $device < /var/tmp/part.txt";
   $cmd_exit_status = ($self->send_client_cmd("$cmd; echo \$?"))[-1];
   return $ERR_STATUS if ($cmd_exit_status != 0);

   $cmd = "dd if=/dev/zero bs=512 count=512 of=$device";
   $cmd_exit_status = ($self->send_client_cmd("$cmd; echo \$?"))[-1];

   return $ERR_STATUS if ($cmd_exit_status != 0);
   return $GOOD_STATUS;
}

# Return 1 in .pm's.
1;
