# $Id$ 

## @summary Library of routines to boot clusters (CFO/SFO pairs) and query their states
##
## @synopsis
##   use VSeries::cluster;
##
##   boot_with_giveback($filerA, $filerB, ...options...)
##   get_cf_state($filer[, $rate]);
##   wait_state($filer, $want_state[, $timeout[, $rate]]);
##   cf_enable($filer);
##   cf_takeover($filer,$dasheff);
##   cf_giveback($filer,[boot=&gt;0],[partner=&gt;$filerB]);
##   cf_disable($filer);
##   get_link_state($filer);
##   get_disks($filer);
##   get_paths($filer);
##   get_volume_status($filer);
##   set_cf_options($filer);
##   force_panic($filer);
##   force_halt($filer);
##   force_reboot($filer);
##
## @example
##
##   # a simple repeated takeover and giveback script
##
##   use VSeries::cluster;
##   param('FILERA','-mesg',"hostspec for one filer");
##   param('FILERB','-mesg',"hostspec for its partner");
##   param('TIMEOUT','-default',180);
##   param('SLEEP_BETWEEN','-default',60);
##
##   my $filerA = host($FILERA);
##   my $filerB = host($FILERB);
##   boot_with_giveback($filerA,$filerB) or tstlog('-fatal',$@);
##   cf_enable($filerA) or tstlog('-fatal',$@);
##   cf_enable($filerB) or tstlog('-fatal',$@);
##
##   while (1) {
##       cf_takeover($filerA) or tstlog('-fail',$@);
##       wait_state($filerA,"TAKEOVER",$TIMEOUT) or tstlog('-fail',$@);
##       wait_state($filerB,"DEAD",$TIMEOUT) or tstlog('-fail',$@);
##       sleep($SLEEP_BETWEEN);
##       cf_giveback($filerA,partner=&gt;$filerB) or tstlog('-fail',$@);
##       wait_state($filerA,"UP",$TIMEOUT) or tstlog('-fail',$@);
##       wait_state($filerB,"UP",$TIMEOUT) or tstlog('-fail',$@);
##       sleep($SLEEP_BETWEEN);
##   }
##
## @usage
##
## <HR><center><b>boot_with_giveback</b></center>
##
## <tt><b>boot_with_giveback($filerA, $filerB, ... boot options ...)</b></tt>
##
## Attempt to boot $filerA and $filerB to the normal filer prompt if
## they are not already.  If this is not possible due to one node
## having taken over the other, then ask it to give back, and attempt
## to boot again.
##
## This uses both console (for booting) and rsh (for determining
## current state).
##
## This returns 1 on success or undef on failure (with $@ set to the error)
##
## <b>$filerA,$filerB</b> - host records or connection records for two
## filers which are a CFO pair.
##
## <b>... boot options ...</b> - further options will be passed to
## boot() when booting.  See the boot library (lib/boot.pm) for more
## information.
##

## <HR><center><b>get_cf_state</b></center>
##
## <tt><b>get_cf_state($filer[, $rate])</b></tt>
##
## Use "cf status" and "cf monitor" to determine the state of $filer
## with respect to clustering.
##
## return "UP" if CFO is enabled on filer but it has not taken over or
## been taken over, "TAKEOVER" if a filer has taken over, "DEAD" if a
## filer has been taken over, "DISABLED" if CFO is disabled on this
## host, or "UNREACH" if we cannot successfully talk to $filer.
##
## If takeover or giveback are in progress, this command waits (tries
## again repeatedly).
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).
##
## <b>$rate</b> - how many seconds to wait in between each poll for the
## current state.  The default is 10 seconds.

## <HR><center><b>wait_state</b></center>
##
## <tt><b>wait_state($filer, $want_state[, $timeout[, $rate]])</b></tt>
##
## Wait for the current state of $filer (according to get_cf_state) to
## become $want_state.
##
## In a scalar context, return 1 if the wanted state was reached, or
## undef if not (in which case $@ will contain an appropriate error
## message that wait_state timed out waiting)
##
## In an array context, return the aforementioned as the first value,
## and the current state at the time wait_state finished as the second
## value.
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).
##
## <b>$want_state</b> - one of UP, TAKEOVER, DEAD, DISABLED, or
## UNREACH, as for get_cf_state.
##
## <b>$timeout</b> - how many seconds (or an absolute time_t value) to
## wait for the wanted state to be reached before giving up with an
## error.  The default is -1 (infinite).
##
## <b>$rate</b> - how many seconds to wait in between each poll for the
## current state.  The default is 10 seconds.

## <HR><center><b>cf_enable</b></center>
##
## <tt><b>cf_enable($filer)</b></tt>
##
## Issue "cf enable" to $filer.
##
## Return 1 if the response suggests CFO is now enabled or undef
## otherwise, with an appropriate error in $@.
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).

## <HR><center><b>cf_disable</b></center>
##
## <tt><b>cf_disable($filer)</b></tt>
##
## Issue "cf disable" to $filer.
##
## Return 1 if the response suggests CFO is now disabled or undef
## otherwise, with an appropriate error in $@.
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).

## <HR><center><b>cf_takeover</b></center>
##
## <tt><b>cf_takeover($filer,[$cmd])</b></tt>
##
## Issue "cf takeover -f" (or other cf takeover command) to $filer
##
## Return 1 if the response suggests takeover has started, or undef
## otherwise, with an appropriate error in $@.
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).
##
## <b>$cmd</b> - what command to use for takeover.  Defaults to "cf
## takeover -f", but could be changed to something else (e.g. boring,
## negotiated "cf takeover").

## <HR><center><b>cf_giveback</b></center>
##
## <tt><b>cf_giveback($filer, [boot=&gt;0], [partner=&gt;$filerB])</b></tt>
##

## Issue "cf giveback -f" to a filer, and possibly help its partner to
## reboot.

## <b>$filer</b> - host record or connection record of filer that will
## give back (if a host record, then the connection #rsh will be
## used).

## <b>boot</b> - whether to help $filer's partner to boot.  The
## default is 0 for non-simulators (which generally need no
## assistance) and 1 for simulators (which generally need assistance).

## <b>partner</b> - the argument or arguments to boot() to use when
## booting the partner (see boot() in lib/boot.pm).  Required if boot
## is 1, ignored if boot is 0.

## <HR><center><b>force_panic</b></center>
##
## <tt><b>force_panic($filer)</b></tt>
##

## Issue "panic" to a filer, and make sure that it reboots on its own.

## <b>$filer</b> - host record or connection record of filer that will
## be issued the panic command.

## <HR><center><b>force_halt</b></center>
##
## <tt><b>force_halt($filer)</b></tt>
##

## Issue "halt" to a filer, and reboot from firmware prompt if necessary, in
## order to bring it back up enough to go into wait for giveback state.

## <b>$filer</b> - host record or connection record of filer that will
## be issued the halt command.

## <HR><center><b>force_reboot</b></center>
##
## <tt><b>force_reboot($filer)</b></tt>
##

## Issue "reboot" to a filer, and make sure that it reboots on its own.

## <b>$filer</b> - host record or connection record of filer that will
## be issued the reboot command.

## @bugs 
##   Both the boot library and this library use "UP" to mean slightly
##   different things ("UP" in a CFO and failover monitor sense is
##   that a node has not taken over or been taken over, but is enabled
##   to do so if need be.  "UP" for the boot library means at the
##   normal filer prompt)

## @see-also lib/boot.pm

## @author lincoln@netapp.com
## mnatale

package VSeries::cluster;

use strict;

use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT = ('boot_with_giveback','boot_with_giveback_one','get_cf_state','wait_state','wait_fw_state',
       'cf_enable','cf_disable','cf_takeover','cf_giveback','ic_connected','disks_on_adapter',
       'cf_enabled','get_link_state', 'get_disks', 'get_paths', 'get_volume_status', 'sysstat', 'set_cf_options',
       'force_panic', 'force_reboot', 'force_halt', 'san_licensed', 'set_panic_option', 'ready_for_giveback',
       'get_takeover_enabled_state');

use Tharn;
use boot;

# given a host or connection, return a connection (when just a host is
# given, favor whatever is given in the 'favor' option or #rsh if not
# specified).

sub make_connection ( $@ ) {
    my ($spec, %opts) = @_;    
    if (ref($spec) =~ /Connectrec/) {
    return $spec;
    } elsif (ref($spec) =~ /Hostrec/) {
    my $conn = $opts{'favor'} || "#rsh";
    return connect($spec->{id} . $conn);
    } else {
    my $host = host($spec,'-ok');
    if ($host) {
        my $conn = $opts{'favor'} || "#rsh";
        return connect($spec . $conn);
    } else {
        return connect($spec);
    }
    }
}

# given a host or connection, return a host.

sub make_host ( $ ) {
    my ($spec) = @_;    
    if (ref($spec) =~ /Connectrec/) {
    return $spec->{hostp};
    } else {
    return host($spec);
    }
}

# Bring both filers up (if they're not already), and make sure any
# which are in takeover state have given back to each other.

sub boot_with_giveback ( $$@ ) {
    my ($filerA, $filerB, @options) = @_;
    my @filers = ($filerA, $filerB);

    return eval {
    my @hosts = map { make_host($_); } @filers;
    my @sobj;
    foreach my $h (@hosts) {
        my $id = $h->{id};
        push (@sobj,
          subtest(sub { boot_with_giveback_one($h,\&boot,@options)
                    or tstlog('-fail',$@); }, 
              '-runid', "boot-$id",
              '-bg',
              '-iffatal','-ok','-ifscript','-ok'
             )
         );
    }
        Subtest::waitFinish(subtests => [@sobj]);

    foreach my $test (@sobj) {
        my $result = Rslt::worst($test->rslts());
        next if !$result;
        my $runid = $test->runid();
        # (Want to error for CONFIG or worse definitely, and I
        # think INSPECT and NA too, but not WARN)
        if ($result->severity() > NATE::Result::Warn->severity()) {
        die "Boot failed in subtest \"$runid\": "
                    . join("\n",$result->type,$result->msgs) . "\n";
        }
    }
    return 1;
    };
}

# Do the work of boot_with_giveback on just one filer (each boots up
# and waits for the other to give back, then if it boots up all the
# way it is told to give back.)

# (needed by Filer/boot.thpl, which is a test that performs what
# boot_with_giveback() does, but with a slightly different interface)

sub boot_with_giveback_one {
    my ($host, $boot_sub, @options) = @_;
    return eval {
    host($host);
    my $c = connect("#console");
    my $r = connect("#rsh");
    $boot_sub->($c,iftakenover=>'wait',@options) or die $@;
    my $try = 0;
    while (get_cf_state($r) eq 'TAKEOVER') {
        say($r,"cf giveback -f", -swallow, '-eval_error');
        die $@ if $@;
        if ($try > 10) {
        die("Tried giveback via $r->{id} too many times ($try)\n");
        }
    }
    return 1;
    };
}

# State will either be "UP", "TAKEOVER", or "DEAD" (a subset of the
# strings that "cf monitor" will print) or "UNREACH" if we can't seem
# to talk to a filer to find out its state, or "DISABLED" if
# clustering is disabled.

# If a takeover is in progress or a giveback is in progress, this
# command will keep retrying.

# About what the states mean: We don't accept "DOWN" and "REBOOTING"
# as states from cf monitor, for some reason.  This makes sense to me
# for REBOOTING (we'll keep trying, and it sounds like a transitory
# state) but not DOWN, but perhaps 'cf monitor' can never print DOWN
# since maybe DOWN implies the filer is halted and not taken over and
# hence neither filer has that IP address to be able to reply to rsh
# of "cf monitor".  "DEAD" might be more appropriately termed "UNDEAD"
# or "TAKEN OVER": someone at this filer's IP address was alive enough
# to tell us that this filer is dead, so it must be that the partner
# has taken over.
#
# - Note from silva@netapp.com - needed to add support to catch the
#   case where $retstr =~ /has taken over/ b/c ignoring this causes
#   some of our .thpl's to enter infinite loops.

sub get_cf_state ( $;$ ) {
    my ($filer,$rate) = @_;
    $rate = $rate || 10;
    my $conn = make_connection($filer);

    my $cur_state;

    while (1) {
    
        my $retstr;
        (undef, undef, $retstr) = Tharn::say($conn, "cf status", -swallow, '-eval_error');
    
        return 'UNREACH' if ($@);
        return 'DISABLED' if ($retstr =~ /Cluster disabled/);
        return 'DISABLED' if ($retstr =~ /takeover disabled/);
        return 'DISABLED' if ($retstr =~ /has disabled takeover by/);
        return 'DISABLED' if ($retstr =~ /Controller Failover disabled/);

        if ($retstr =~ /\s+has taken over\s+/) { last; }
        
        last if (length($retstr) &&
            ($retstr !~ /attempting takeover/) &&
            ($retstr !~ /giveback in progress/));

        sleep $rate;

    } # while (1)

    my $out;
    do {
    (undef, undef, $out) = 
        say($conn,"cf monitor", -swallow, '-eval_error');
    if ($out =~ /(UP|TAKEOVER|DEAD)/) {
        $cur_state = $1;
    }
    } while ($out =~ /Try again/ && !defined($cur_state));

    if (!defined($cur_state)) {
    return 'UNREACH';
    } else {
    return $cur_state;
    }
}

## @name ::get_takeover_enabled_state()
##
## @description
## Method to determine if a cluster is in the TAKEOVER_ENABLED
## state.  This exists when logs are synched and we are able
## to do an unforced takeover.  Returns 0 when TAKEOVER_ENABLED
## is true and 1 when TAKEOVER_ENABLED is false.  Had to create this
## method because 'get_cf_state' doesn't allow us to request
## a TAKEOVER_ENABLED state, instead only UP, DEAD, & TAKEOVER.
## Unfortunately UP and TAKEOVER_ENABLED are not mutually 
## exclusive states.
##
## @author James Silva (silva@netapp.com)
##
## @synopsis
## get_takeover_enabled_state($);
##
sub get_takeover_enabled_state ($) {
    my ($filer) = @_;
    my $conn = make_connection($filer);

    my ($cur_state, $after) = '';

    (undef, undef, $after) = Tharn::say($conn,"cf monitor", -swallow, '-timeout', 30);
    
    if ($after =~ /(TAKEOVER_ENABLED)/) {
        $cur_state = $1;
    }

    # return 0 if we matched, 1 otherwise.
    if ($cur_state eq "TAKEOVER_ENABLED"){
        return(0); 
    }
    else { return(1); }
}

## @name ::ready_for_giveback()
##
## @description
## Method to determine if a DEAD cf state partner that has been
## taken over is ready for giveback.  This method MUST be called
## on the head that is in the TAKEOVER state (which is also UP). 
## Returns 0 if the DEAD filer is ready for giveback and 1  
## otherwise.  
## 
## @author James Silva (silva@netapp.com)
##
## @synopsis
## ready_for_giveback($);
##
sub ready_for_giveback ($) {
    my ($filer) = @_;
    my $conn = make_connection($filer);

    my ($cur_state, $after) = '';

    (undef, undef, $after) = Tharn::say($conn,"cf status", -swallow, '-timeout', 30);
    
    # return 0 if we matched, 1 otherwise.
    if ($after =~ /(is ready for giveback)/) {
        return(0);
    }
    else { return(1); }
}


# wait for up to $timeout seconds ($timeout being undefined means
# infinite) for $filer to be in state $target_state.  Whether or not
# it does so, return the last state seen

sub wait_state ( $$;$$ ) {
    my ($filer, $want_state, $timeout, $rate) = @_;
    my $cur_state;
    my $res = eval {
    my $id = make_host($filer)->{id};
    my $end_time = $timeout ? Tharn::timeout2time($timeout) : Tharn::timeout2time(-1);
    my $rate = $rate || 10;
    while (time() <= $end_time) {
        $cur_state = get_cf_state($filer,$rate);
        last if ($cur_state eq $want_state);
        sleep $rate;
    };
    if ($cur_state ne $want_state) {
        die "Timed out waiting for state \"$want_state\", current state is \"$cur_state\"\n";
    }
    return 1;
    };
    return wantarray ? ($res, $cur_state) : $res;
}

sub get_link_state ($) {
    my ($filer) = @_;
    my $fc_link_state;
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);
    (undef, undef, $fc_link_state) = Tharn::say($conn, 'priv set -q advanced;fcadmin link_state', '-ok', -swallow, -timeout, 60);
    return $fc_link_state;
}

sub get_disks ($) {
    my ($filer) = @_;
    my $disks;
    my $id = make_host($filer)->{id}; 
    my $conn = make_connection($filer);
    (undef, undef, $disks) = Tharn::say($conn, 'storage show disk', '-ok', -swallow, -timeout, 60);
    return $disks;
}

sub get_paths ($) {
    my ($filer) = @_;
    my $paths;
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);
        (undef, undef, $paths) = Tharn::say($conn, 'storage show disk -p', '-ok', -swallow, -timeout, 60);
     return $paths;
}

sub get_volume_status($){
    my ($filer) = @_;
    my $vol_status;
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);
    (undef, undef, $vol_status) = Tharn::say($conn, 'vol status', '-ok', -swallow, -timeout, 60);
    return $vol_status;
}
   
sub sysstat($) {
    my ($filer) = @_;
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);
    Tharn::say($conn, 'sysstat -c 6 -x 2', '-eval_error', -swallow, -timeout, 60);
}

sub force_panic($) {
    my ($filer) = @_;
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);
    Tharn::say($conn, 'priv set advanced; panic', '-eval_error', -swallow, -timeout, 60);
    sleep(60);
}

sub force_halt($) {
    my ($filer) = @_;
    my $id = make_host($filer)->{id};
    # we need to use the console here
    my $conn = connect("#console", -timeout, 60);

    HALT: Tharn::say($conn, 'halt -t0', '-eval_error', -swallow, -blind);
    my @prompts1 = ('Please press any key to reboot', 'Waiting for giveback','^CFE> ', '^LOADER\\*?>','^ok ', 'Password', '^LOADER-(A|B)\*?>');

    # wait 180s (if necessary) for something to happen
    my (undef,undef,$after,$prompt) = Tharn::say($conn,"",'-nocr', '-timeout',180,'-maxidle',180, '-noprompt','-prompt', @prompts1);
    Tharn::logcomment("prompt: $prompt");

    # First check and see if we're prompted to reboot, new in 7-mode BR.
    # If this is the case, we need to listen again for other prompts.
    # This is from burt 283097.
    if ($prompt =~ /Please press any key to reboot/){
        Tharn::say($conn, 'y', -swallow, -blind);

        # Again, wait 180s (if necessary) for something to happen
        (undef,undef,$after,$prompt) = Tharn::say($conn,"",'-nocr', '-timeout',180,'-maxidle',180, '-noprompt','-prompt', @prompts1);
        Tharn::logcomment("prompt: $prompt");
    }

    # Otherwise check to see if we are waiting for giveback, if so return.
    elsif ($prompt =~ 'Waiting for giveback') {
        Tharn::logcomment("WAITING FOR GIVEBACK MSG FOUND");
            return 0;
    }
    elsif($prompt =~ /Password/i) {
        Tharn::say($conn, $conn->password(), -swallow, -blind);
        goto HALT;
    }

    # if we aren't waiting for giveback, maybe we fell into the fw prompt
    # if we find one of these, we'll try to bring the head up and re-check
    my @prompts = ('^CFE> ', '^LOADER\\*?>','^ok ', '^LOADER-(A|B)\*?>');
    foreach my $re (@prompts) {
        if ($prompt =~ $re) {
        tstlog(-comment, "FW PROMPT FOUND");
        Tharn::say($conn, 'bye', -swallow, -blind);
        sleep(60);
            return 0;
        }
        elsif($prompt =~ /Password/i) {
            Tharn::say($conn, $conn->password(), -swallow, -blind);
        }
    }
    return 1;
}

sub force_reboot($) {
    my ($filer) = @_;
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);
    Tharn::say($conn, 'reboot -t 0', -swallow, '-eval_error', -timeout, 60);
    sleep(60);
}

sub san_licensed($) {
    my ($filer) = @_;
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);

    # clear out $after buffer to make sure parse is clean
    my ($match, $before, $after, $undef) = Tharn::say($conn, '', -swallow);
    ($match, $before, $after, $undef) = Tharn::say($conn, 'fcp', -swallow);
    logcomment("Verifying if FCP licensed feature");
    if ($after =~ m/fcp: FCP is not licensed./)
    {
        Tharn::logcomment("FCP is not licensed");
    }
    else
    {
        Tharn::logcomment("FCP is licensed");
        # return as soon as we find out that one or the other is licensed
        return 1;
    }

    # clear out $after buffer to make sure parse is clean
    ($match, $before, $after, $undef) = Tharn::say($conn, '', -swallow);
    ($match, $before, $after, $undef) = Tharn::say($conn, 'iscsi', -swallow);
    if ($after =~ /iscsi: service is not licensed/) {
        Tharn::logcomment("iSCSI is not licensed");
    } else {
        Tharn::logcomment("iSCSI is licensed");
        # return as soon as we find out that one or the other is licensed
        return 1;
    }

    # neither one is licensed
    return 0;
}

sub set_panic_option($) {
    my ($filer) = @_;
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);

    Tharn::say($conn, 'options cf.takeover.on_panic on', -swallow, '-eval_error', -timeout, 60);
}

sub set_cf_options($) {
    my ($filer) = @_;
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);
    Tharn::say($conn, 'options cf.giveback.auto.cifs.terminate.minutes 0', -swallow, '-eval_error', -timeout, 60);
    Tharn::say($conn, 'options cf.giveback.auto.enable off', -swallow, '-eval_error', -timeout, 60);
    Tharn::say($conn, 'options cf.giveback.auto.terminate.bigjobs on', -swallow, '-eval_error', -timeout, 60);
    Tharn::say($conn, 'options cf.giveback.auto.check.partner on', -swallow, '-eval_error', -timeout, 60);
    Tharn::say($conn, 'options cf.takeover.detection.seconds 15', -swallow, '-eval_error', -timeout, 60);
    Tharn::say($conn, 'options cf.takeover.on_disk_shelf_miscompare off', -swallow, '-eval_error', -timeout, 60);
    Tharn::say($conn, 'options cf.takeover.on_failure on', -swallow, '-eval_error', -timeout, 60);
    Tharn::say($conn, 'options cf.takeover.on_panic on', -swallow, '-eval_error', -timeout, 60);
    Tharn::say($conn, 'options cf.takeover.on_short_uptime on', -swallow, '-eval_error', -timeout, 60);
}
 
# wait for up to $timeout seconds ($timeout being undefined means
# infinite) for $filer to be in desired fw state.  Whether or not
# it does so, return the last state seen
# The last var $partner means tells the function whether to check its 
# own firmware state or it's partner's fw state. This is because 
# we cannot read the fw state of a node that is down by doing a cf 
# readfw on itself. So we have to go thru its partner.

sub wait_fw_state ( $$;$$$ ) {
    my ($filer, $want_state, $timeout, $rate, $partner) = @_;
    my $cur_state;
    my $res = eval {
    my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);
    my $end_time = $timeout ? Tharn::timeout2time($timeout) : Tharn::timeout2time(-1);
    my $rate = $rate || 10;
    while (time() <= $end_time) {
                (undef, undef, $cur_state) = say($conn,"priv set -q test; cf readfw", -swallow, "-eval_error");
        if ($partner) {
            ($cur_state) = ($cur_state =~ /PARTNER: state (SF[A-Z_]*), progress/);
        } else {
            ($cur_state) = ($cur_state =~ /LOCAL: state (SF[A-Z_]*), progress/);
        }
        last if ($cur_state eq $want_state);
        sleep $rate;
    };
    if ($cur_state ne $want_state) {
        die "Timed out waiting for state \"$want_state\", current state is \"$cur_state\"\n";
    }
    return 1;
    };
    return wantarray ? ($res, $cur_state) : $res;
}

# check for filer sanity by taking an adapter number as input
# and checking if disks exist on that adapter or not
sub disks_on_adapter ( $$$ ) {
    my ($filer, $adapter, $timeout) = @_;
    #tstlog('-comment', "filer=$filer adapter=$adapter timeout=$timeout");
    my $conn = make_connection($filer);
    my $end_time = $timeout ? Tharn::timeout2time($timeout) : Tharn::timeout2time(-1);
    my $rate = 10;    
    my $count = 0;
    my $out;
    while (time() <= $end_time) {
        (undef, undef, $out) = say($conn,"priv set -q advanced;fcadmin device_map $adapter", -swallow, "-eval_error");
            if ($out =~ /Shelf Unknown: [0-9 X]+/) {
                                tstlog('-comment',"disk adapter $adapter back (but SES not) in $count tries");
                                return 1;
                }
        if ($out =~ /Shelf \d+: [0-9 X]+/) {
            tstlog('-comment',"disk adapter $adapter back in $count tries");
            return 1;
        }
        sleep($rate);
        $count = $count + 1;
    }
    tstlog('-comment',"disk adapter $adapter not back in $count tries");
    #die "Timed out but did not see expected disks on adapter\n";
}

# check whether IC is connected or not. Params are 2 filers, primary timeout1. If that 
# timeout triggers, we reset the NIC and try again for another timeout2 period.
sub ic_connected($$$$) {
    my ($filer1, $filer2, $timeout1, $timeout2) = @_;
    my $conn1 = make_connection($filer1);
    my $conn2 = make_connection($filer2);
    my $end_time = $timeout1 ? Tharn::timeout2time($timeout1) : Tharn::timeout2time(-1);
    my $rate = 10;
    my $count = 0;
    my $out1;
    my $out2;
     # skip all this if in takeover state. (this probably doesn't work if we are currently in the middle
    # of a takeover...)
    (undef, undef, $out1) = say($conn1,"cf status", -swallow, "-eval_error");
        if ($out1 =~ /taken over/) {
                  tstlog('-comment',"IC not back - in takeover state");
                  return 1;
    }
    while (time() <= $end_time) {
        (undef, undef, $out1) = say($conn1,"priv set -q advanced;ic stats error", -swallow, "-eval_error");
        (undef, undef, $out2) = say($conn2,"priv set -q advanced;ic stats error", -swallow, "-eval_error");
        # only one side has to say "CONNECTED" for things to be ok. (right?) Hell, no. BOTH
        if ($out1 =~ /RV connection state :\s+CONNECTED/) {
             if ($out2 =~ /RV connection state :\s+CONNECTED/) {
                               tstlog('-comment',"IC back in $count tries");
                            return 1;
                        }
        }
        sleep($rate);
        $count = $count + 1;
    }
    #things are hosed.
    tstlog('-comment',"IC not back in $count tries");
    say($conn1,"priv set -q test; ic dump hw_stats", -swallow, "-eval_error");
    say($conn2,"priv set -q test; ic dump hw_stats", -swallow, "-eval_error");
    say($conn1,"priv set -q test; ic reset nic", -swallow, "-eval_error");
    say($conn2,"priv set -q test; ic reset nic", -swallow, "-eval_error");
        
    #try again after a reset. Really, this shouldn't be here, because these routines
    # should just look at status, but we need to know is reset does any good....
    $end_time = $timeout2 ? Tharn::timeout2time($timeout2) : Tharn::timeout2time(-1);
    while (time() <= $end_time) {
                (undef, undef, $out1) = say($conn1,"priv set -q advanced;ic stats error", -swallow, "-eval_error");
                (undef, undef, $out2) = say($conn2,"priv set -q advanced;ic stats error", -swallow, "-eval_error");
                if ($out1 =~ /RV connection state :\s+CONNECTED/) {
                        if ($out2 =~ /RV connection state :\s+CONNECTED/) {
                                tstlog('-comment',"IC back after resets in $count tries");
                                return 1;
                    }
        }
                sleep($rate);
                $count = $count + 1;
        }

    tstlog('-comment',"IC still not back in $count tries");
    #die "Timed out while expecting to see IC CONNECTED\n";
}

#check whether clustering is enabled correctly after a NIC- or disk-destroying
# event. Enabled is good. Disabled is good if originally
# induced by cf disable; and form of disabling due to unsync log
# or MB disks or some such is bad. 
sub cf_enabled($$$$) {
        my ($filer1, $filer2, $timeout1, $timeout2) = @_;
        my $conn1 = make_connection($filer1);
        my $conn2 = make_connection($filer2);
        my $end_time = $timeout1 ? Tharn::timeout2time($timeout1) : Tharn::timeout2time(-1);
        my $rate = 10;
        my $count = 0;
        my $out;
        while (time() <= $end_time) {
                (undef, undef, $out) = say($conn1,"cf status", -swallow, "-eval_error");
                # look for cf enabled  on filer1
                if ($out =~ /Cluster enabled/) {
                        tstlog('-comment',"clustering on A back in $count tries");
                        return 1;
                }
                # look for has taken over or has been taken over
                if ($out =~ /taken over/) {
                        tstlog('-comment',"clustering not back - in takeover state");
                        return 1;
                }
        if ($out =~ /Cluster disabled/) {
                        tstlog('-comment',"clustering not back on A - disabled");
                        return 1;
                }
        sleep($rate);
                $count = $count + 1;
        }
    
        # this should not be necessary, but just in case...
    $end_time = $timeout2 ? Tharn::timeout2time($timeout2) : Tharn::timeout2time(-1);
    while (time() <= $end_time) {
                (undef, undef, $out) = say($conn2,"cf status", -swallow, "-eval_error");
                # look for cf enabled on filer2
                if ($out =~ /Cluster enabled/) {
                        tstlog('-comment',"clustering on B back in $count tries");
                        return 1;
                }
                if ($out =~ /Cluster disabled/) {
                        tstlog('-comment',"clustering on B disabled");
                        return 1;
                }
                sleep($rate);
                $count = $count + 1;
        }

    tstlog('-comment',"clustering still not back in $count tries");
    #die...
}



# (arbitrarily decide it takes 10 seconds max cf enable to fully take
# effect)
use constant CF_ENABLE_TIMELIMIT => 10;

sub cf_enable ( $ ) {
    my ($filer) = @_;
    my $conn = make_connection($filer);
    say($conn,"cf enable", -swallow, '-eval_error');
    return ($@) ? undef : 1;
}

sub cf_disable ( $ ) {
    my ($filer) = @_;
    my $conn = make_connection($filer);
    say($conn,"cf disable", -swallow, '-eval_error');
    return ($@) ? undef : 1;
}


sub cf_takeover ( $;$ ) {
    my ($filer,$cmd) = @_;
    if (!$cmd) {
    $cmd = "cf takeover";
    }
    my $conn = make_connection($filer);
    say($conn,$cmd, -swallow, '-eval_error');
    return ($@) ? undef : 1;
}

# Issue a command (cf giveback [-f]) to give back $filer.  If boot=>0,
# that's it.  If boot=>1, then the 'partner' option is required and we
# attempt to help the partner to reboot after the giveback.
# Simulators are usually the only hosts which need such help, so this
# option defaults to 1 if $filer has hosttype "filer-simulator" and 0
# otherwise.  If you want to issue a giveback without the -f flag then
# you need to pass in an options hash with 'dont_force' defined as a 
# key in the hash.
# %options = ('dont_force' => 1);
# VSeries::cluster::cf_giveback($filer, \%options);
#    

sub cf_giveback ( $;% ) {
    my ($filer,%options) = @_;
    return eval {
    my $fhost = make_host($filer);
    my $fconn = make_connection($filer);
    my $do_boot = exists($options{boot}) ? $options{boot} :
        ($fhost->hosttype =~ /^filer-simulator/);

    if(defined($options{'dont_force'})){
        Tharn::say($fconn,"cf giveback", -swallow, '-eval_error');
    }
    else{
        Tharn::say($fconn,"cf giveback -f", -swallow, '-eval_error');
    }

    die $@ if $@;
    if ($do_boot) {
        my $partner = $options{partner};
        if (!$partner) {
        die "'partner' option required for giveback() when 'boot' option is true\n";
        }
        my (@partner) = (ref($partner) eq 'ARRAY') ? @$partner : ($partner);
        &boot(@partner) or die $@;
    }
    return 1;
    };
}

