# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/VSeries/utils.pm#2 $
#
## @summary Utility library functions to be used by VSeries tests
##
## @description
##   These test functions are needed for the VSeries test suites
## @usage
##   use utils;
##   use utils qw(...);
##
## @keywords  util test library
##
## @dependencies
## use config_base
## use Tharn
##
## @author cattel@netapp.com
## @status testme
##
##

package VSeries::utils;

use strict;
use Tharn;
use Expect;
use config_base;
use Date::Format;
use Date::Parse;
use Data::Compare;
use Time::Local;            # Will be used for calculating epoch time
use VSeries::client_object;
use Tharn qw($Log);
use Hostrec;
use Connectrec;
use VSeries::Clustered::nacl_convenience;

my $GOOD_STATUS = 0;
my $ERR_STATUS = 1;
my $TRUE = "true";
my $FALSE = "false";

use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT = ('compare_arrays', 'compare_hashes', 'parse_vol_status', 'parse_aggr_status', 'compare_dates', 'date_offset', 'sysconfig_compare', 'sysconfig_parse', 
           'unix_time_to_epoch', 'check_io_completion', 'check_io_status', 'unmount_test_vol', 'start_io_to_volume', 'mount_test_vol',
           'setup_client', 'scsi_fitp_block_lun', 'check_degraded_vols','check_degraded_aggrs','compare_dates_in_different_formats',
           'compare_chksum');

#-------------------------------------------------------------------------------
## @name ::compare_chksum ($$)
##
## @author Arvind Chandrasekaran (arvindc@netapp.com)
##
## @description
## Takes in a hash reference to disk info array (output of discover_disks routine). And 
## checks if the checksums of the disks are the same in both the hashes. Returns $GOOD_STATUS
## if they match and returns $ERR_STATUS along with an array ref of serial number of disks
## for which the chksums have changed and the serial number of disks that do not have matching
## entries in the other hash ref.
## The compare is done based on the disk serial numbers and this is to avoid the 
## switch and primary and secondary paths from affecting the compare.
## Refer BURT 569983 for more information.
##
## @arg $disk_info_first        Required: Disk information in first hash ref
## @arg $disk_info_second       Required: Disk information in second hash ref to compare with first
##
## @synopsis
## $status = VSeries::utils::compare_chksum($disk_info_first,$disk_info_second);

sub compare_chksum($$) {
    my $disk_info_first = shift;
    my $disk_info_second = shift;

    my %disk_info_first_hash = %{$disk_info_first};
    my %disk_info_second_hash = %{$disk_info_second};

    my @mismatch_serials = ();

    # Iterate over all keys (serial numbers) in the first hash and
    # check if
    # (i) an equivalent entry exists in the second hash
    # (ii) the chksum value is the same in both the hashes
    # If the check fails add the serial number to the array @mismatch_serials.
    foreach my $serial (keys %disk_info_first_hash) {
        if (exists $disk_info_second_hash{$serial}) {
            if ($disk_info_first_hash{$serial}{chksum} eq $disk_info_second_hash{$serial}{chksum}) {
                next;
            }
            else {
                push @mismatch_serials, $serial;
            }
        }
        else {
            push @mismatch_serials, $serial;
        }
    }

    # Now check if every entry in second hash has an equivalent entry in first hash.
    # If any serial is missed, then add the entry to the array @mismatch_serials.
    if (scalar (keys %disk_info_first_hash) != scalar (keys %disk_info_second_hash)) {
        foreach my $serial (keys %disk_info_second_hash) {
            if (!exists $disk_info_first_hash{$serial}) {
                push @mismatch_serials, $serial;
            }
        }
    }

    if (scalar @mismatch_serials > 0) {
        return ($ERR_STATUS, \@mismatch_serials);
    }
    else {
        return ($GOOD_STATUS, undef);
    }
}

#-------------------------------------------------------------------------------
## @name ::sysconfig_compare ($$)
##
## @author Arvind Chandrasekaran (arvindc@netapp.com)
##
## @description
## Takes in references to two parsed sysconfig outputs (hashes). Compares each element in 
## the first hash with the second hash. Whenever an error is encountered, stores the 
## respective key in an array, until the comparison is completed. Returns the reference to
## an array, whose first element is either $GOOD_STATUS or $ERR_STATUS. If the first element
## is $ERR_STATUS, then the following elements in the array will have the mismatching keys.
##
## @arg $first_hash_ref          Required: Reference to hash with parsed first sysconfig output
## @arg $second_hash_ref         Required: Reference to hash with parsed second sysconfig output
##
## @synopsis
## $status = VSeries::utils::sysconfig_compare($first_hash_ref,$second_hash_ref);

sub sysconfig_compare ($$) {
    my $first_hash_ref=shift;
    my $second_hash_ref=shift;

    my %first_hash=%$first_hash_ref;
    my %second_hash=%$second_hash_ref;
    my %temp_second_hash=%second_hash;

    my @temp_first_array=();
    my @temp_second_array=();
    my @return_array=();
    my $itr=0;
    my $key="";

    # Iterate over %first_hash and compare every element with %second_hash
    # Check if the number of elements in the nested arrays for every header in sysconfig
    # output matches between the two hashes.
    # When a miscompare is found, collect  the $key in the array @return_array. This array will
    # be returned finally to the user. This array will help in debugging.

    foreach $key (keys %first_hash) {
        # Sort the arrays under each key to make it easier to compare the hashes.
        @temp_first_array = sort {$a cmp $b} @{$first_hash{$key}};

        # If the second_hash does not have the same key defined, it is an error, so
        # push the missing key into @return_array and continue.
        if (defined $second_hash{$key}) {
            @temp_second_array = sort {$a cmp $b} @{$second_hash{$key}};
        }
        else {
            push @return_array, $key;
        }

        # Check if the array size under the key in both the hashes is the same. Else,
        # push the key into the return_array and continue.

        if ($#temp_first_array != $#temp_second_array) {
            push @return_array, $key;
        }

        # Iterate over the array corresponding to $key in both hashes and compare them.
        # If there is a mismatch, then push the $key into @return_array and continue.

        for ($itr=0; $itr < scalar @temp_first_array ; $itr++) {
            if ($temp_first_array[$itr] ne $temp_second_array[$itr]) {
                push @return_array, $key;
            }
        }

        # Delete the corresponding key in temp_second_hash - this will help to ensure
        # first_hash is not a subset of #second_hash.
        delete $temp_second_hash{$key};
    }

    # The next foreach loop will help to check if the first_hash is a subset of second_hash. 
    # If yes, then push the missing $key into @return_array and continue.
    foreach $key (keys %temp_second_hash) {
        push @return_array, $key if (defined $temp_second_hash{$key});
    }

    # At this point, if there were mismatches, then @return_array will have a array size > 0.
    # So, if there were mismatches, then unshift to set the first element in @return_array to '$ERR_STATUS'.
    # Else, push $GOOD_STATUS into @return_array.
    if ($#return_array >= 0) {
        unshift @return_array, $ERR_STATUS;
    }
    else {
        push @return_array, $GOOD_STATUS;
    }

   # Return ref to @return_array
    return \@return_array;
}

#-------------------------------------------------------------------------------
## @name ::sysconfig_parse ($;$$)
##
## @author Arvind Chandrasekaran (arvindc@netapp.com)
##
## @description
## Takes in a reference to an array with sysconfig output (mandatory). Parses the same
## and stores it in a hash. The parsing logic is as explained below
## i) The sysconfig output has some top level headers.
## ii) Some top level headers have a set of nested fields with data under them.
##
## The sysconfig_parse subroutine, parses the top level headers and takes them as the key for
## the nested fields. Then every nested field is stored as an array element under the respective
## key (header). Finally, the subroutine returns a reference to the hash.
##
## Optionally, the subroutine takes in a header and field value. If both are specified
## then the nested field that matches (pattern match) the header and the field is returned.
## If only header is specified, then a ref to the complete array with all fields that are under
## the header is returned.
##
## @arg $sysconfig_array_ref      Required: Reference to array containing sysconfig output
## @arg $header                   Optional: Header Pattern/String
## @arg $field                    Optional: Field Pattern/String under a header
##
## @synopsis
## $hash_ref  = VSeries::utils::sysconfig_parse($sysconfig_array_ref);
## $array_ref = VSeries::utils::sysconfig_parse($sysconfig_array_ref, "Remote LAN Module");
## $status    = VSeries::utils:sysconfig_parse($sysconfig_array_ref, "Remote LAN Module", "Status");

sub sysconfig_parse($;$$) {
    my $sysconfig_ref = shift;
    my $header = shift if (@_);
    my $field = shift if (@_);

    my @sysconfig_array = @$sysconfig_ref;
    my %parse_hash=();

    my ($line, $key,$prev_key,$value,$space)="";
    my $flag=0;

    # Array containing the different power management modules
    # If a new power management module is available, add the exact string as shwon
    # in sysconfig output as an additional element in @power_modules array

    my @power_modules = ("Remote LAN Module" , "Service Processor");
    my $pmm_pattern = join('|',@power_modules);

    # In the sysconfig output we have 4 different types of lines to be handled.
    # The different types are explained below.
    #
    # (a) Headers/keys      -  The lines that contain the main entry are called headers and
    #                         they are used as keys for the parsed hash. Some of these 
    #                         headers have a set of attributes/fields under them.
    #                         Examples:
    #                             NetApp Release 7.3.4RC1: Thu May 27 16:40:57 PDT 2010
    #                             System ID: 0151734425 (vrv3170f13a); partner ID: 0151734695 (vrv3170f13b)
    #                             System Serial Number: 700000139716 (vrv3170f13a)
    #                             System Rev: B0
    #                             System Storage Configuration: Multi-Path HA
    #
    #
    # (b) PMM Header/Key    - The Power Management Module in a system could be 'Remote LAN Module'
    #                        or 'Service Processor'. And this header alone has the attribute 'Status'
    #                        also displayed on the same line as the header/key. This needs to be 
    #                        handled.
    #                        Examples:
    #                            Remote LAN Module           Status: Online 
    #
    # (c) Attributes/Fields - The lines that are under a certain header are called 
    #                        attributes/fields. The list of attributes under a certain key/header 
    #                        will be stored as an array, pointed to by the respective key.
    #                        Examples:
    #                            slot 0: System Board 2.6 GHz (System Board XV A4)
    #                                    Modey Name:         V3170
    #                                    Part Number:        110-00119
    #                                    Revision:           A4
    #                                    Serial Number:      739481
    #                                    BIOS version:       4.4X1
    #                                    Loader version:     1.7
    #                                    Agent FW version:   3
    #
    #
    # (d) Memory Attributes - The 'Memory Attributes' alone is a special attribute under the 
    #                        x key. This is so because the value for Memory Attributes are 
    #                        displayed on multiple lines. This is explained in detail in
    #                        the following code blocks.
    #                        Examples:
    #                                    Memory Attributes:  Node Interleaving
    #                                                        Bank Interleaving
    #                                                        Hoisting
    #                                                        Chipkill ECC

    # Iterate over the sysconfig output to parse the same.
    foreach $line (@sysconfig_array) {
        chomp;


        # The below code block handles the entries of type (d) as explained above.
        #
        # The below code block handles the Memory attributes field that has values without
        # a field on the same line. i.e. some values must use the field from a previous line
        # This will handle lines that look like this
        #
        #        slot 0: System Board 2.6 GHz (System Board XV A4)
        #
        #                Memory Attributes:  Node Interleaving
        #                                    Bank Interleaving
        #                                    Hoisting
        #                                    Chipkill ECC
        #
        # The above will be parsed as
        #     $key="slot 0: System Board 2.6 GHz (System Board XV A4)";
        #     $parse_hash{$key}[0]="Memory Attributes:  Bank Interleaving";
        #     $parse_hash{$key}[1]="Memory Attributes:  Hoisting";
        #
        # This will ensure that the logic works even if the above list is not in order
        #
        # The value for $key is set in code blocks for (a) and (b) parser below.
        # If the current line has 'Memory Attributes', this needs to be handled separately
        # because the values are also available on the following lines. So set a flag, to keep track
        # of the next lines. Now parse the line and store it into %parse_hash.
 
        if ($line =~ /(Memory Attributes:)(\s+)(.+)$/) {
            $flag=1;
            $space=$2;
            $value=$1. $2. $3;
            push @{$parse_hash{$key}}, $value;
            next;
        }
        # The following code block is to handle the lines following 'Memory Attributes' (flag is set to 1).
        # The values in these lines should be appended to the string 'Memory Attributes' and stored
        # in %parse_hash. This will ensure that the parse logic function works even if the 
        # values are displayed in a different order.
        if ($flag == 1) {
            if ($line !~ /:/) {
                if ($line =~ /\s+(.+)$/) {
                    $value="Memory Attributes:". $space . $1;
                    push @{$parse_hash{$key}}, $value;
                    if ((defined $header && $header =~ /$key/) && (defined $field && $line =~ /$field/)) {
                        return $value;
                    }
                    next;
                }
            }
            # If the current line has a ':' and $flag==1, then we have come to a line after ALL the 'Memory
            # Attributes'. So reset the $flag to 0.
            else {
                $flag=0;
            }
        }


        # The below code block handles the entries of type (c) as explained above.
        #
        # The below code block handles the nested fields under a header
        # Example:
        #
        #         slot 0: System Board 2.6 GHz (System Board XV A4)
        #                 Model Name:         V3170
        #                 Part Number:        110-00119
        #
        # The above will be parsed as
        #     $key="slot 0: System Board 2.6 GHz (System Board XV A4)";
        #     $parse_hash{$key}[0]="Model Name:         V3170";
        #     $parse_hash{$key}[1]="Part Number:        110-00119";
        #
        # The value for $key is set in code blocks for (a) and (b) parser below.


        # The attributes under a header are identified by atleast 8+ spaces in the beginning
        # of the line. Parse the line and store it in an array under the key.
        # The Headers have only 4 spaces at the beginning.
        if ($line =~ /^\s{9,}([\S\s]+?:)(.+)$/ || $line =~ /^\s{1}\t{1}([\S\s]+?:)(.+)$/ ) {
            push @{$parse_hash{$key}}, $line;

            # if the user has passed both the Header/key and Attribute/Field. And if the 
            # current line matches both, then we have to return the current line to the
            # user. 
            if ((defined $header && $header =~ /$key/) && (defined $field && $line =~ /$field/)) {
                return $line;
            }
            next;
        }

        # The below code block handles the entries of type (b) as explained above.
        #
        # The below code block specifically handles the header for power modules like
        # Remote LAN Module or Service Processor. The reason for handling it separately is
        # because the header has a attribute/field in the same line (Status).
        # Example:
        #
        #          Remote LAN Module           Status: Online
        #
        # The above will be parsed as
        #     $key="Remote LAN Module";
        #     $parse_hash{$key}[0]="Status: Online";

        elsif ($line =~ /($pmm_pattern)\s+(.+)$/) {

            # This line denotes the start of the PMM header/key.
            # This line also signifies the end of the previous header. So, store the $key in $prev_key.
            # If the user specified $header that matches $prev_key, then we have to return the array
            # pointed by $prev_key to user.
            $prev_key=$key;
            if ((defined $header && $header =~ /$prev_key/) && (!defined $field)) {
                return \@{parse_hash{$prev_key}};
            }

            # Else, set the $key to the current line.
            # This line also has the attribute/field 'Status' which is stored in $2 as per pattern.
            # Store the same in $parse_key.
            $key=$1;
            $value=$2;
            push @{$parse_hash{$key}}, $value;
            if ((defined $header && $header =~ /$key/) && (defined $field && $value =~ /$field/)) {
                return $value;
            }
            next;
        }

        # The below code block handles the entries of type (a) as explained above.
        #
        # The below code block handles the main table entries
        #
        # Example:
        #
        #         slot 0: System Board 2.6 GHz (System Board XV A4)
        #
        # The above will be parsed as
        #     $key="slot 0: System Board 2.6 GHz (System Board XV A4)";
        #
        # The parsed out header will be used as the key for the nested fields
        # For those headers that do not have nested fields, the value will be
        # set to a empty array.
        #

        # This line signifies the start of a header (other than PMM header)
        # A header has one tab space in the beginning of the line.
        elsif ($line =~ /^\t{1}([\S\s]+?:)(.+)$/ || $line =~ /^\t{1}([\S\s]+?:)(\s*)$/) {
            # This line also signifies the end of the previous header. So, store the $key in $prev_key.
            # If the user specified $header that matches $prev_key, then we have to return the array
            # pointed by $prev_key to user.
            $prev_key=$key;
            $key=$line;
            if ((defined $header && $header =~ /$prev_key/) && (!defined $field)) {
                return \@{parse_hash{$prev_key}};
            }

            # Else, initiate the key with the new header and set the value to a empty array.
            # If this Hader/key has atributes/fields under it they will be handled by the 
            # code block for (c) and (d).
            @{$parse_hash{$key}}=();
            next;
        }
    }

    # If the user had passed, $header or $field, we must have returned in one 
    # of the above code blocks. If not it is an error, so return an error status.
    if (defined $header || defined $field) {
        return $ERR_STATUS;
    }

    # Return the parsed hash. 
    return \%parse_hash;
}

# Common routine to compare two arrays.  Takes in references to each array to
# be compared returns 0 if they are not equal 1 if they are.  It makes the
# assumption that the arrays are sorted.
sub compare_arrays(@@){
    my ($first_array_ref, $second_array_ref) = @_;

    #first if the size of the arrays are different, then they are not equal.
    return $ERR_STATUS unless @$first_array_ref == @$second_array_ref;

    #if they are the same size then iterate over the arrays comparing elements
    # of the same index.
    for(my $i = 0; $i < @$first_array_ref; $i++){
        return $ERR_STATUS if $first_array_ref->[$i] ne $second_array_ref->[$i];
    }
    return $GOOD_STATUS;
}

# Common routine to compare two hashes.  Takes in references to each hash to
# be compared returns 0 if they are not equal 1 if they are.  
sub compare_hashes($$){
    my ($first_hash_ref, $second_hash_ref) = @_;
    return (Compare( $first_hash_ref, $second_hash_ref ) ? 1 : 0);
}

# Common util to parse the different forms of output for the vol status command.
# In all cases it will build a return hash where the key is the unique
# identifier of information, in some cases it's disk names, other cases volume
# names.  This is a user beware utility and a caller should know what they're
# getting back and looking for.  The two parameters are the flag which in the
# case of no flag should be called with "" and the array is the output from the
# calling of vol status on a filer.
sub parse_vol_status($@){
    my ($cmd_flag, @output) = @_;
    my %return_hash = ();

    # Adding a toggle flag to help handle cases where options for a volume span
    # multiple lines.
    my $option_line = 0;

    # Handle the simple case where there is no flag and it's just "vol status"
    # Key for the return hash here is the volume name.

    # SAMPLE OUTPUT:
    #   Volume State      Status            Options
    #             vol0 online     raid0, trad       root
    #    asis_test_vol online     raid0, flex
    #       jimmy_trad restricted raid0, trad       
    #                             wafl inconsistent 

    if($cmd_flag eq ""){
        # Trim the header
        @output = @output[1..$#output];
        my ($name, $state, $status, $options) = "";
        foreach my $line (@output){
            # the options are delimited by commas then a whitespace and it's
            # easier to split on just whitespace so make options just
            # delimited by commas
            $line =~ s/, /,/g;
            # Encountered an issue with vol names dynamically growing TOWARDS
            # the beginning of the line, which really affects generic splitting
            # So let's replace the dynamic whitespace at the front of any line.
            $line =~ s/^ *//;

            # Due to increased complexity of status and options both running
            # onto subsequent lines but status information not having the
            # dangling comma the approach here needed to be reworked.  It is
            # the design assumption that state will always be on the first
            # line of vol input with the name so we key in on the
            # offline|online keywords to determine if a line contains the
            # first parts of info about a given volume and then following
            # lines contain extra info.
            if($line =~ m/(online|offline|failed|partial|restricted|creating|unknown)/){
                ($name, $state, $status, $options) = split(/ +/, $line);
                $return_hash{$name}{state} = $state;
                $return_hash{$name}{status} = $status;
                if(defined($options)){
                    $return_hash{$name}{options} = $options;
                }
                else{
                    $return_hash{$name}{options} = "";
                }


            }
            else{
                # Ok this is designed to properly parse out the additional
                # lines of information associated with the last seen "first
                # line" of vol info.  The difference between status and
                # options information is that option information besides the
                # "root" option contains equal signs "=" so we key on those
                # two things.  Any trailing status information is added to
                # the condition key of the hash, this is intended to hold the
                # "mirrored" status if the volume is mirrored.
                my @token_array = split(/ +/, $line);
                foreach my $token (@token_array){
                    if(($token =~ m/\=/) || ($token =~ m/root/)){
                        my $tmp_options = $return_hash{$name}{options} . $token;
                        $return_hash{$name}{options} = $tmp_options;
                    }
                    else{
                        if(defined($return_hash{$name}{condition})){
                            $return_hash{$name}{condition} .= $token . " ";
                        }
                        else{
                            $return_hash{$name}{condition} = $token . " ";
                        }
                    }

                }

            }
        }


    }
    # Handle the case for -s this will be particularly useful for selection
    # of disks that are already zeroed.  This will dump out all the spare
    # disks owned by the given filer.  Due to the funky complications of
    # having the "(not zeroed)" on a given line, we'll actually parse that
    # out first and then handle the rest of the line.

    # SAMPLE OUTPUT:
    #    Spare disks
    #
    #    RAID Disk       Device                  HA  SHELF BAY CHAN Pool Type  RPM  Used (MB/blks)    Phys (MB/blks)
    #    ---------       ------                  ------------- ---- ---- ---- ----- --------------    --------------
    #    Spare disks for block or zoned checksum traditional volumes or aggregates
    #    spare           0b.19                   0b    1   3   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.21                   0b    1   5   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.22                   0b    1   6   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.23                   0b    1   7   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.24                   0b    1   8   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.26                   0b    1   10  FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.27                   0b    1   11  FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.28                   0b    1   12  FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           vns-c9124-3:1-1.126L19  0a    -   -          -  LUN   N/A  9972/20422656     10073/20630400 (not zeroed)
    #    spare           vns-c9124-3:1-1.126L20  0a    -   -          -  LUN   N/A  9972/20422656     10073/20630400 (not zeroed)
    #    spare           vns-c9124-3:1-1.126L7   0a    -   -          -  LUN   N/A  9972/20422656     10073/20630400
    #    spare           vns-c9124-3:1-1.126L8   0a    -   -          -  LUN   N/A  39890/81694720    40293/82521600

    if($cmd_flag eq "-s"){
        my ($name, $ha, $shelf, $bay, $chan, $pool, $type, $rpm, $used, $phys, $zeroed) = "";
        foreach my $line (@output){
            if($line =~ m/^spare/){
                if($line =~ m/\(not zeroed\)/ || $line =~ m/\(zeroing.*\)/){
                    $zeroed = $FALSE;
                    $line =~ s/\(not zeroed\)//;
                    $line =~ s/\(zeroing.*\)//;
                }
                else{
                    $zeroed = $TRUE;
                }
                
                if($line =~ /\s+$/) {
                    $line =~ s/\s+$//;
                }

                my @line_tokens = split(/ +/, $line);

                if(scalar(@line_tokens) == 10){
                    my @first_part = @line_tokens[0..5];
                    my @second_part = @line_tokens[6 .. $#line_tokens];
                    @line_tokens = (@first_part, "-", @second_part);
                }
                if(scalar(@line_tokens) == 11){
                    shift(@line_tokens);
                    ($name, $ha, $shelf, $bay, $chan, $pool, $type, $rpm, $used, $phys) = @line_tokens;
                    $return_hash{$name}{ha} = $ha;
                    $return_hash{$name}{shelf} = $shelf;
                    $return_hash{$name}{bay} = $bay;
                    $return_hash{$name}{chan} = $chan;
                    $return_hash{$name}{pool} = $pool;
                    $return_hash{$name}{type} = $type;
                    $return_hash{$name}{rpm} = $rpm;
                    $return_hash{$name}{used} = $used;
                    $return_hash{$name}{phys} = $phys;
                    $return_hash{$name}{zeroed} = $zeroed;

                }


            }
        }

    } #end if($cmd_flag eq "-s")

    # Handle the case for -c, key in the return hash is the volume name.
    # Here we're looking for checksum info.

    # SAMPLE OUTPUT:
    #         Volume Checksum style Checksum status
    #           vol0 block          Checksums active
    # asis_test_aggr block          Checksums active

    if($cmd_flag eq "-c"){
        # Trim off the header
        @output = @output[1 .. $#output];


        my ($name, $checksum_type1, $checksum_type2, $checksum_status, $checksum_status1, $checksum_status2) = "";
        foreach my $line (@output){
            # remove the dynamic whitespace that could lead at the front of
            # a line and complicate the splitting.
            $line =~ s/^ *//;

            my @tmp_array = split(/ +/, $line);
            if(scalar(@tmp_array) == 4) {
                ($name, $checksum_type1, $checksum_status1, $checksum_status2) = @tmp_array;
                $return_hash{$name}{checksum_type} = $checksum_type1;
            }
            elsif(scalar(@tmp_array) == 5) {
                # aggr status -c
                # Aggr  Checksum style Checksum status
                # aggr1 block          Checksums active
                # aggr2 zoned v2       Checksums active
                ($name, $checksum_type1, $checksum_type2, $checksum_status1, $checksum_status2) = @tmp_array;
                $return_hash{$name}{checksum_type} = $checksum_type1 . " " . $checksum_type2;
            }
            
            $checksum_status = $checksum_status1 . " " . $checksum_status2;
            
            $return_hash{$name}{checksum_status} = $checksum_status;
        }
    }

    # Handle the case for -r, key in the return hash is the disk name.
    # Due to the difference in the output of this flag depending on if
    # a vol is specified it solves on that case right now.  Eventually
    # this will expand to support if no vol name is specified as well.

    # SAMPLE OUTPUT:
    # Volume vol0 (online, raid0) (block checksums)
    #  Plex /vol0/plex0 (online, normal, active)
    #    RAID group /vol0/plex0/rg0 (normal)
    #
    #      RAID Disk Device                  HA  SHELF BAY CHAN Pool Type  RPM  Used (MB/blks)    Phys (MB/blks)
    #      --------- ------                  ------------- ---- ---- ---- ----- --------------    --------------
    #      data      vns-c9124-3:1-1.126L1   0a    -   -          -  LUN   N/A  46114/94441472    46580/95397120

    if($cmd_flag eq "-r"){

        my $current_plex = "";
        my $current_raid_group = "";
        my $disk_next = 0;
        my ($raid, $name, $ha, $shelf, $bay, $chan, $pool, $type, $rpm, $used, $phys) = "";

        # This piece got redesigned to handle the addition of plex and raid
        # group information associated with disks.  Having multiple plexes
        # is part of having mirrored volumes in metro cluster so we need
        # to handle and account for that.
        foreach my $line (@output){

            # If the line has the word plex means it contains plex info and
            # parse it out.
            if($line =~ m/^ *Plex/){
                my @tokens = split(/ +/, $line);
                $tokens[2] =~ s/\/\w*\///;
                $current_plex = $tokens[2];

            }

            # If the line has the words RAID group it contains raid group info
            # and parse it out.
            if($line =~ m/^ *RAID group/){
                my @tokens = split(/ +/, $line);
                $tokens[3] =~ s/\/\w*\/\w*\///;
                $current_raid_group = $tokens[3];

            }

            # Once we started parsing disk lines and we encounter a blank line
            # it's an indicator that we're either at the end of the input or
            # the next line is NOT a disk line.
            if($line =~ m/^ *$/ && $disk_next == 1){
                $disk_next = 0;

            }

            # This is the line that's an indicator that we've already seen the
            # the header indicating that the next line is a disks line.
            if($disk_next == 1){

                # Remove the dynamic whitespace at the start of the line
                $line =~ s/^ +//;
                if($line =~ /\s+$/) {
                    $line =~ s/\s+$//;
                }

                # Split the line into tokens.
                my @line_tokens = split(/ +/, $line);

                # Due to the introduction of MDL, lines with disks or luns may
                # have 10 or 11 tokens.  10 tokens will most likely be due to the
                # channel column being blank for LUNs so the design idea is to
                # inject a dash for that column and then split out the tokens
                # for the pieces of information they correspond to.
                if(scalar(@line_tokens) == 10){
                    my @first_part = @line_tokens[0..5];
                    my @second_part = @line_tokens[6 .. $#line_tokens];
                    @line_tokens = (@first_part, "-", @second_part);

                }
                if(scalar(@line_tokens) == 11){
                    ($raid, $name, $ha, $shelf, $bay, $chan, $pool, $type, $rpm, $used, $phys) = @line_tokens;
                    $return_hash{$name}{raid} = $raid;
                    $return_hash{$name}{ha} = $ha;
                    $return_hash{$name}{shelf} = $shelf;
                    $return_hash{$name}{bay} = $bay;
                    $return_hash{$name}{chan} = $chan;
                    $return_hash{$name}{pool} = $pool;
                    $return_hash{$name}{type} = $type;
                    $return_hash{$name}{rpm} = $rpm;
                    $return_hash{$name}{used} = $used;
                    $return_hash{$name}{phys} = $phys;
                    $return_hash{$name}{raid_group} = $current_raid_group;
                    $return_hash{$name}{plex} = $current_plex;
                }

            }

            # This is the case where the line is a bunch of dashes representing
            # the separation of the header and disk lines of input. We are trying
            # to match for 5 or more dashes in this case.
            if($line =~ m/(\-{5,})/){
                $disk_next = 1;

            }

        }

    } #end if($cmd_flag eq "-r")

    return \%return_hash;

}


# Common util to parse the different forms of output for the aggr status
# command.  In all cases it will build a return hash where the key is the
# unique identifier of information, in some cases it's disk names, other cases
# aggr names.  This is a user beware utility and a caller should know what
# they're getting back and looking for.  The two parameters are the flag which
# in the case of no flag should be called with "" and the array is the output
# from the calling of aggr status on a filer.  Most flags can be parsed and
# tokenized based on the conventional way a command is called however in the
# -r case it must be called with an aggr name even though the command on the
# filer cli can be called without an aggr name.
sub parse_aggr_status($@){
    my ($cmd_flag, @output) = @_;
    my %return_hash = ();

    # Handle the simple case where there is no flag and it's just "aggr status"
    # Key for the return hash here is the aggr name.

    # SAMPLE OUTPUT:
    # Aggr State      Status            Options
    # vol2 online     raid0, trad
    #                 mirrored
    # aggr0 online     raid0, aggr       root, nosnap=on,
    #                 mirrored          snapshot_autodelete=off,
    #                                   lost_write_protect=off
    # aggr1 online     raid0, aggr       snapshot_autodelete=off,
    #                 mirrored          lost_write_protect=off

    if($cmd_flag eq ""){
        # Trim the header
        @output = @output[1..$#output];
        my ($name, $state, $status, $options) = "";
        foreach my $line (@output){
            # the options are delimited by commas then a whitespace and it's
            # easier to split on just whitespace so make options just
            # delimited by commas
            $line =~ s/, /,/g;
            # Encountered an issue with aggr names dynamically growing TOWARDS
            # the beginning of the line, which really affects generic splitting
            # So let's replace the dynamic whitespace at the front of any line.
            $line =~ s/^ *//;

            # Due to increased complexity of status and options both running
            # onto subsequent lines but status information not having the
            # dangling comma the approach here needed to be reworked.  It is
            # the design assumption that state will always be on the first
            # line of aggregate input with the name so we key in on the
            # offline|online keywords to determine if a line contains the
            # first parts of info about a given aggregate and then following
            # lines contain extra info.
            if($line =~ m/(online|offline|failed|partial|restricted|creating|unknown)/){
                ($name, $state, $status, $options) = split(/ +/, $line);
                $return_hash{$name}{state} = $state;
                $return_hash{$name}{status} = $status;
                $return_hash{$name}{options} = $options;

            }
            else{
                # Ok this is designed to properly parse out the additional
                # lines of information associated with the last seen "first
                # line" of aggregate info.  The difference between status and
                # options information is that option information besides the
                # "root" option contains equal signs "=" so we key on those
                # two things.  Any trailing status information is added to
                # the condition key of the hash, this is intended to hold the
                # "mirrored" status if the aggregate is mirrored.
                my @token_array = split(/ +/, $line);
                foreach my $token (@token_array){
                    if(($token =~ m/\=/) || ($token =~ m/root/)){
                        my $tmp_options = $return_hash{$name}{options} . $token;
                        $return_hash{$name}{options} = $tmp_options;
                    }
                    else{
                        $return_hash{$name}{condition} = $return_hash{$name}{condition} . $token . ",";
                    }

                }

            }
        }


    }

    # Handle the case for -c, key in the return hash is the volume name.
    # Here we're looking for checksum info.

    # SAMPLE OUTPUT:
    #       Aggr Checksum style Checksum status\n
    #      aggr3 block          Checksums active\n
    #      aggr1(1) block          Checksums active\n
    #      aggr2(1) zoned          Checksums reinitialized and active for new\n
    #                            updates\n
    #      aggr0(1) block          Checksums active\n
    #      aggr1 block          Checksums active\n
    #      aggr2 zoned          Checksums initializing:\n
    #                     Checksum reinitialization in progress\n
    #      aggr0 block          Checksums active\n

    if($cmd_flag eq "-c"){
        # Trim off the header
        @output = @output[1 .. $#output];

        my ($name, $checksum_type1, $checksum_type2) = (undef, undef, undef);
        my @checksum_type =();
        my @checksum_status = ();

        foreach my $line (@output){
            # remove the dynamic whitespace that could lead at the front of
            # a line and complicate the splitting.
            $line =~ s/^ *//;

            # Skip the line as it is a continuation for the above line i.e. Checksums initializing:
            # Please refer sample output posted above.
            next if ($line =~ m/^Checksum reinitialization in progress$/);

            my @tmp_array = split(/ +/, $line);
            next if($#tmp_array == 0);

            # Stores the first element
            ($name) = @tmp_array;

            # Grepping the strings like 'reinitialized/active' from the tmp_array where the 'aggr status -c'
            # ouput will be stored.
            # If it could able to find the string then it will be stored in the variable otherwise it will be null.
            my ($checksum_status_index) = grep {$tmp_array[$_] =~ m/(initializing|reinitialized|active)/i}0..$#tmp_array;

            if ($checksum_status_index) {
               # Getting the index of 'Checksums' string from 'Checksums reinitialized and active for new' or
               # 'Checksums active'
               $checksum_status_index = $checksum_status_index - 1;

               # Storing the complete checksum status message.
               @checksum_status = @tmp_array[$checksum_status_index .. $#tmp_array];

               # Starting from the first index till the index of 'Checksums' string.
               @checksum_type = @tmp_array[1 .. $checksum_status_index-1];

               # Storing the checksum type. (works for block/zoned/zoned v2)
               $return_hash{$name}{checksum_type} = "@checksum_type";
            } else {
               if(scalar(@tmp_array) == 4) {
                  ($name, $checksum_type1, @checksum_status) = @tmp_array;

                  $return_hash{$name}{checksum_type} = $checksum_type1;
               }
               elsif(scalar(@tmp_array) == 5) {
                  # aggr status -c
                  # Aggr  Checksum style Checksum status
                  # aggr1 block          Checksums active
                  # aggr2 zoned v2       Checksums active
                  ($name, $checksum_type1, $checksum_type2, @checksum_status) = @tmp_array;

                  $return_hash{$name}{checksum_type} = $checksum_type1 . " " . $checksum_type2;
               }
             }            
            $return_hash{$name}{checksum_status} = "@checksum_status";
        }
    }

    # Handle the case for -r, key in the return hash is the disk name.
    # Due to the difference in the output of this flag depending on if
    # an aggr is specified it solves on that case right now.  Eventually
    # this will expand to support if no aggr name is specified as well.

    # SAMPLE OUTPUT:
    #Aggregate asis_test_aggr (online, raid0) (block checksums)
    #  Plex /asis_test_aggr/plex0 (online, normal, active)
    #    RAID group /asis_test_aggr/plex0/rg0 (normal)
    #
    #      RAID Disk Device                  HA  SHELF BAY CHAN Pool Type  RPM  Used (MB/blks)    Phys (MB/blks)
    #      --------- ------                  ------------- ---- ---- ---- ----- --------------    --------------
    #      data      vns-c9124-3:1-1.126L6   0a    -   -          -  LUN   N/A  9972/20422656     10073/20630400
    #      data      vns-c9124-3:1-1.126L5   0a    -   -          -  LUN   N/A  9972/20422656     10073/20630400
    #      data      vns-c9124-3:1-1.126L4   0a    -   -          -  LUN   N/A  9972/20422656     10073/20630400

    if($cmd_flag eq "-r"){
        my $current_plex = "";
        my $current_raid_group = "";
        my $disk_next = 0;
        my ($raid, $name, $ha, $shelf, $bay, $chan, $pool, $type, $rpm, $used, $phys) = "";

        # This piece got redesigned to handle the addition of plex and raid
        # group information associated with disks.  Having multiple plexes
        # is part of having mirrored aggregates in metro cluster so we need
        # to handle and account for that.
        foreach my $line (@output){

            # If the line has the word plex means it contains plex info and
            # parse it out.
            if($line =~ m/^ *Plex/){
                my @tokens = split(/ +/, $line);
                $tokens[2] =~ s/\/\w*\///;
                $current_plex = $tokens[2];

            }

            # If the line has the words RAID group it contains raid group info
            # and parse it out.
            if($line =~ m/^ *RAID group/){
                my @tokens = split(/ +/, $line);
                $tokens[3] =~ s/\/\w*\/\w*\///;
                $current_raid_group = $tokens[3];

            }

            # Once we started parsing disk lines and we encounter a blank line
            # it's an indicator that we're either at the end of the input or
            # the next line is NOT a disk line.
            if($line =~ m/^ *$/ && $disk_next == 1){
                $disk_next = 0;

            }

            # This is the line that's an indicator that we've already seen the
            # the header indicating that the next line is a disks line.
            if($disk_next == 1){

                # Remove the dynamic whitespace at the start of the line
                # Remove the dynamic whitespace at the start of the line
                $line =~ s/^ +//;
                if($line =~ /\s+$/) {
                    $line =~ s/\s+$//;
                }

                # Split the line into tokens.
                my @line_tokens = split(/ +/, $line);

                # Due to the introduction of MDL, lines with disks or luns may
                # have 10 or 11 tokens.  10 tokens will most likely be due to the
                # channel column being blank for LUNs so the design idea is to
                # inject a dash for that column and then split out the tokens
                # for the pieces of information they correspond to.
                if(scalar(@line_tokens) == 10){
                    my @first_part = @line_tokens[0..5];
                    my @second_part = @line_tokens[6 .. $#line_tokens];
                    @line_tokens = (@first_part, "-", @second_part);

                }
                if(scalar(@line_tokens) == 11){
                    ($raid, $name, $ha, $shelf, $bay, $chan, $pool, $type, $rpm, $used, $phys) = @line_tokens;
                    $return_hash{$name}{raid} = $raid;
                    $return_hash{$name}{ha} = $ha;
                    $return_hash{$name}{shelf} = $shelf;
                    $return_hash{$name}{bay} = $bay;
                    $return_hash{$name}{chan} = $chan;
                    $return_hash{$name}{pool} = $pool;
                    $return_hash{$name}{type} = $type;
                    $return_hash{$name}{rpm} = $rpm;
                    $return_hash{$name}{used} = $used;
                    $return_hash{$name}{phys} = $phys;
                    $return_hash{$name}{raid_group} = $current_raid_group;
                    $return_hash{$name}{plex} = $current_plex;
                }

            }

            # This is the case where the line is a bunch of dashes representing
            # the separation of the header and disk lines of input. We are trying
            # to match for 5 or more dashes in this case.
            if($line =~ m/(\-{5,})/){
                $disk_next = 1;

            }

        }

    } #end if($cmd_flag eq "-r")

    # Handle the case for -s this will be particularly useful for selection
    # of disks that are already zeroed.  This will dump out all the spare
    # disks owned by the given filer.  Due to the funky complications of
    # having the "(not zeroed)" on a given line, we'll actually parse that
    # out first and then handle the rest of the line.  For the purpose of
    # the zeroing flag then when a zeroing is in progress and the end of the
    # disk line will have (zeroing ... x%).  Until a disk is 100% zeroed then
    # the disk will not have the zeroed flag set.

    # SAMPLE OUTPUT:
    #    Spare disks
    #
    #    RAID Disk       Device                  HA  SHELF BAY CHAN Pool Type  RPM  Used (MB/blks)    Phys (MB/blks)
    #    ---------       ------                  ------------- ---- ---- ---- ----- --------------    --------------
    #    Spare disks for block or zoned checksum traditional volumes or aggregates
    #    spare           0b.19                   0b    1   3   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.21                   0b    1   5   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.22                   0b    1   6   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.23                   0b    1   7   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.24                   0b    1   8   FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.26                   0b    1   10  FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.27                   0b    1   11  FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           0b.28                   0b    1   12  FC:A   -  FCAL 10000 136000/278528000  137422/281442144
    #    spare           vns-c9124-3:1-1.126L19  0a    -   -          -  LUN   N/A  9972/20422656     10073/20630400 (not zeroed)
    #    spare           vns-c9124-3:1-1.126L20  0a    -   -          -  LUN   N/A  9972/20422656     10073/20630400 (not zeroed)
    #    spare           vns-c9124-3:1-1.126L7   0a    -   -          -  LUN   N/A  9972/20422656     10073/20630400
    #    spare           vns-c9124-3:1-1.126L8   0a    -   -          -  LUN   N/A  39890/81694720    40293/82521600

    if($cmd_flag eq "-s"){
        my ($name, $ha, $shelf, $bay, $chan, $pool, $type, $rpm, $used, $phys, $zeroed, $chksum) = "";
        foreach my $line (@output){
            if( $line =~ /Spare disks for (\w+)/ ){
                $chksum = $1;
            }
            if($line =~ m/^spare/){
                if($line =~ m/\(not zeroed\)/ || $line =~ m/\(zeroing.*\)/){
                    $zeroed = $FALSE;
                    $line =~ s/\(not zeroed\)//;
                    $line =~ s/\(zeroing.*\)//;
                }
                else{
                    $zeroed = $TRUE;
                }
                
                if($line =~ /\s+$/) {
                    $line =~ s/\s+$//;
                }

                my @line_tokens = split(/ +/, $line);

                if(scalar(@line_tokens) == 10){
                    my @first_part = @line_tokens[0..5];
                    my @second_part = @line_tokens[6 .. $#line_tokens];
                    @line_tokens = (@first_part, "-", @second_part);
                }
                if(scalar(@line_tokens) == 11){
                    shift(@line_tokens);
                    ($name, $ha, $shelf, $bay, $chan, $pool, $type, $rpm, $used, $phys) = @line_tokens;
                    $return_hash{$name}{ha} = $ha;
                    $return_hash{$name}{shelf} = $shelf;
                    $return_hash{$name}{bay} = $bay;
                    $return_hash{$name}{chan} = $chan;
                    $return_hash{$name}{pool} = $pool;
                    $return_hash{$name}{type} = $type;
                    $return_hash{$name}{rpm} = $rpm;
                    $return_hash{$name}{used} = $used;
                    $return_hash{$name}{phys} = $phys;
                    $return_hash{$name}{zeroed} = $zeroed;
                    $return_hash{$name}{chksum} = $chksum;
                }


            }
        }

    } #end if($cmd_flag eq "-s")




    return \%return_hash;

}


## @name ::netmask_hex_to_decimal($)
##
## @description
## A utility method to convert a hexidecimal netmask into dotted octet form.
## Expects a hexidecimal formatted string as input ("0xffffffff") and will
## perform up front checking that the passed in value is of that format.  Will
## return ERR_STATUS if input does not match this format otherwise it will
## return a string that is the dotted octet form of the netmask.
##
## @synopsis
## $converted_netmask = VSeries::utils::netmask_hex_to_decimal($)
##
## @arg $mask_string (required): hexidecimal formatted netmask to convert
##
##
sub netmask_hex_to_decimal($){
    my ($mask_string) = @_;
    my $return_mask = "";

    # Check to see if the format of the netmask is valid hex.  Should be in the
    # form of "0xffffffff"
    if($mask_string !~ m/0x([0-9|a-f|A-F]{8})/){
        return $ERR_STATUS;
    }

    # remove the leading 0x for conversion purposes
    $mask_string = substr($mask_string, 2);

    # looping through if there's at least 2 characters remaining in the
    # mask_string then convert the next 2 characters in the string to a decimal
    # value.  We can make the assumption here about the format due to the check
    # up above on the input.  After that, shift off the two converted chars
    # from the string and continue.
    while(length($mask_string) > 1){
        $return_mask = $return_mask . hex(substr($mask_string, 0, 2)) . ".";
        $mask_string = substr($mask_string, 2);

    }

    # this chops of the unnecessary '.' at the end
    $return_mask = substr($return_mask, 0, -1);

    return $return_mask;

}

## @name ::compare_dates($$)
##
## @description
## A utility method to compare two dates to see which of the two is
## the more recent.  The expected format for the date strings are
## what is commonly output by the filer's date command and in other
## command output like storage load balance.  The dates should look like
## this are are expected to be in that form, the method doesn't check format:
## Mon Feb  2 14:26:24 EST 2009
##
## The method will start comparing the dates by year,month,day,hour,minute
## and seconds.  A return value of 1 means the first param is more recent.
## a return value of 2 means the second param was.  A zero means they are
## equal.
##
## It is expected that the dates passed in are in the same timezone and if
## not and error value of -1 will be returned.
##
## @synopsis
## $return_value = VSeries::utils::compare_dates($date1, $date2)
##
## @arg $date1 (required): string form of a date as shown above.
## @arg $date2 (required): string form of a date as shown above.
##
##
sub compare_dates($$){
    my ($date1, $date2) = @_;

    # split on whitespace to tokenize.
    my @date1_tokens = split(/ +/, $date1);
    my @date2_tokens = split(/ +/, $date2);

    # If the two dates aren't in the same timezone return and error.
    if($date1_tokens[4] ne $date2_tokens[4]){return -1;}

    # It's easier to assign numerical values to the months for comparison.
    my %month_vals = ('JAN' => 1, 'FEB' => 2, 'MAR' =>3, 'APR' => 4, 'MAY' => 5, 'JUN' => 6,
                      'JUL' => 7, 'AUG' => 8, 'SEP' => 9, 'OCT' => 10, 'NOV' => 11, 'DEC' => 12);

    # These are the tokens for the dates we'll use for comparison.
    my ($seconds1, $minutes1, $hours1, $days1, $months1, $years1) = 0;
    my ($seconds2, $minutes2, $hours2, $days2, $months2, $years2) = 0;

    # Let's get the numerical values for the months, convert the month in
    # the date strings to all upper case and find them in the hash.
    # It is assumed the month token was a valid month.
    $months1 = $month_vals{uc($date1_tokens[1])};
    $months2 = $month_vals{uc($date2_tokens[1])};

    # Pull out the date tokens.
    $days1 = $date1_tokens[2];
    $days2 = $date2_tokens[2];

    # Pull out the years tokens.
    $years1 = $date1_tokens[5];
    $years2 = $date2_tokens[5];

    # By splitting on whitespace we ended up with hours:minutes:seconds still
    # all in one token and separated by colons.  This match just helps us pull
    # those out.
    $date1_tokens[3] =~ m/(\d+):(\d+):(\d+)/;
    $hours1 = $1;
    $minutes1 = $2;
    $seconds1 = $3;

    # Same here.
    $date2_tokens[3] =~ m/(\d+):(\d+):(\d+)/;
    $hours2 = $1;
    $minutes2 = $2;
    $seconds2 = $3;


    # So here's the real work, basically we pick that if the first date is more
    # recent than the second date we'll return 1, if they are equal then return
    # 0 and if neither of those two things are true then the second date must
    # be more recent.  So we just walk down the dates comparing in this order
    # years, months, days, hours, minutes, seconds...
    if($years1 > $years2){ return 1;}
    elsif($years1 == $years2 && $months1 > $months2){ return 1;}
    elsif($years1 == $years2 && $months1 == $months2 && $days1 > $days2) {return 1;}
    elsif($years1 == $years2 && $months1 == $months2 && $days1 == $days2 && $hours1 > $hours2){return 1;}
    elsif($years1 == $years2 && $months1 == $months2 && $days1 == $days2 && $hours1 == $hours2 && $minutes1 > $minutes2){return 1;}
    elsif($years1 == $years2 && $months1 == $months2 && $days1 == $days2 && $hours1 == $hours2 && $minutes1 == $minutes2 && $seconds1 > $seconds2){return 1;}
    elsif($years1 == $years2 && $months1 == $months2 && $days1 == $days2 && $hours1 == $hours2 && $minutes1 == $minutes2 && $seconds1 == $seconds2){return 0;}
    else{return 2;}

}


## @name ::date_offset($$;$)
##
## @description
## A utility method that adds or subtracts an offset to a date string. 
## The date format should be in the form of the unix 'date' command output
## Ex:   Mon Feb  2 14:26:24 EST 2009
##
## This method will not verify date syntax or format. It assumes the date
## arg has proper syntax as above. Returns modified date string
##
## @synopsis
## $modified_date = VSeries::utils::date_offset($date, $offset, $mode)
##
## @arg $date (required): date string (format as shown above)
## @arg $offset (required): offset value in seconds to add to $date
## @arg $mode (optional): determines whether to add or subtract offset from
##                        the date. Values: 'add','subt' (default: add) 
##
sub date_offset ($$;$) {
    my ($date, $offset, $mode) = @_;

    if ( ! defined $mode ) {
        $mode = "add";
    }
    
    # Strip out any leading whitespaces from $date string before we split
    $date =~ s/^ +//;

    # First we extract the timezone from $date for time conversion method 
    my $zone = (split (/ +/, $date))[4];

    # Convert $date to unix timestamp (elapsed seconds since epoch)
    my $timestamp = Date::Parse::str2time($date, $zone);

    # Add or subtract offset from timestamp
    if ( $mode eq "subt" ) {
        $timestamp -= $offset;
    } else {
        $timestamp += $offset;
    }
	
    # Convert timestamp back to formatted string
    $date = Date::Format::time2str("%a %b %e %X %Z %Y", $timestamp, $zone);

    return $date;
}

#-------------------------------------------------------------------------------
## @name unix_time_to_epoch($)
##
## @author Pranav Ojha (pranav.ojha@netapp.com)
##
## @description
## Converts time in Unix format to seconds since epoch.
##
## @arg $time       Required: Time in Unix 'date' format "Tue Jun  7 06:40:14 EDT 2011"
##
## @synopsis
## $epoch_secs = VSeries::utils::unix_time_to_epoch("Tue Jun  7 06:40:14 EDT 2011");
sub unix_time_to_epoch ($) {
    my ($time) = @_;

    # We define hash to convert days/months to numeric format 
    my %days = ('Sun' => 0,'Mon' => 1,'Tue' => 2,'Wed' => 3,'Thu' => 4,'Fri' => 5,'Sat' => 6);
    my %months = ('Jan' => 0,'Feb' => 1,'Mar' => 2,'Apr' => 3,'May' => 4,'Jun' => 5,
                'Jul' => 6,'Aug' => 7,'Sep' => 8,'Oct' => 9,'Nov' => 10,'Dec' => 11);
    my @date_components = ();
    my $wday = 0;                   # Abbrevated weekday text
    my $mon = 0;                    # Abbrevated month text
    my $mday = 0;                   # Numeric day of the month
    my ($hour,$min,$sec) = (0,0,0); # Time in hh:mm:ss format
    my $timezone = undef;           # Timezone
    my $year = 0;                   # Year
    my $msg_time = 0;

    # Split start time components as received in Unix format
    @date_components = split /\s+/, $time;

    # Adding a condition where it also converts the date in the following formats :
    # 1) 22/Jul/2011 18:14:56 and
    # 2) 9/24/2011 18:07:17
    if ($time =~ m/\//) {
       ($mday, $mon, $year) = split(/\//, $date_components[0]);
       my $month = $months{$mon};

       # If the condition matches to '9/24/2011' format then interchange $mon and $mday values.
       # And decrement $mon by 1 because values of %months starts with 0.
       # As keys of %months are names, $month value will be set to null.
       # As per the example of '9/24/2011', $mon will be set to 8, after decrementing by 1 and
       # $mday will become 24.
       if ($month eq q{}) {
          my $temp = $mon;
          $mon = $mday - 1;
          $mday = $temp;
       } else {
          $mon = $month;
       }

       ($hour, $min, $sec) = split(/:/, $date_components[1]);
    } else {
       $wday = $days{$date_components[0]};
       $mon = $months{$date_components[1]};
       $mday = $date_components[2];
       ($hour,$min,$sec) = split /:/, $date_components[3];
       $timezone = $date_components[4];
       $year = $date_components[5];
    }

    # Return seconds since epoch
    my $date_in_epoch_seconds = Time::Local::timelocal($sec, $min, $hour, $mday, $mon, $year);
    return $date_in_epoch_seconds;
}

#-------------------------------------------------------------------------------
## @name ::setup_client($$;$)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## Method to conveniently sets up the client object for host io. 
## Also verifies that the host has access to the "dt" io generation tool. 
## Returns the Client object if succsessful, else returns $ERR_STATUS.
## 
## @arg: $hostname    Required: Host name of client
##
## @arg: $dt_path     Optional: Path to the DT binary on the host.
##
## @arg  $conn_type   Optional: Host connection Type. Default '#rsh'
##
## @synopsis
## my $client_object = VSeries::utils::setup_client($$;$);
##
sub setup_client($;$$) {
    my ($hostname, $dt_path, $conn_type) = @_;

    if (!defined $conn_type) {
        $conn_type = "rsh";
    } elsif ($conn_type =~ /#/) {
       $conn_type =~ s/#//g;
    }

    # Create our client object (for now we only support one client node
    # specified on command line)
    my $client_object = undef;
    my $hostrec_object = undef;
    my $connectrec_object = undef;
    eval {
        $client_object = VSeries::client_object->new();
        $hostrec_object = Hostrec->new(id => $hostname);
        $connectrec_object = Connectrec->new(host => $hostrec_object, template => $conn_type);
        $client_object->set_internal_variables(
            hostname   => $hostname,
            default_ip => $hostrec_object->default_ip(),
            conntype   => $connectrec_object->conntype(),
            connport   => $connectrec_object->connport(),
            username   => $connectrec_object->username(),
            password   => $connectrec_object->password(),
            conn       => $connectrec_object,
        );
    };
    if ($@) {
        $Log->debug("Client object creation failed ". $@);
        return $ERR_STATUS;
    }

    if (defined $dt_path) {
        # Check that 'dt' tool is installed on the client
        $client_object->send_client_cmd("cd $dt_path");
        my @client_cmd_out = $client_object->send_client_cmd( "which " . $dt_path . "dt" );

        foreach my $line (@client_cmd_out) {
            if ( $line =~ /which: no dt in/ ) {
                $Log->debug("DT not found on the client at $dt_path");
                return $ERR_STATUS;
            }
        }
    }

    return $client_object;
}

#-------------------------------------------------------------------------------
## @name ::mount_test_vol($$$$)
##
## @author Hari Prasad M (Hari.M@NetApp.com), Arvind Chandrasekaran
##
## @description
## Method to make the mount point, set proper permissions, and mount
## the volume. This will also create the mount dir if it does not exist.
## Returns $GOOD_STATUS if passed or $ERR_STATUS on failure.
##
## @arg:    $client_obj    Required: Client Object to be used
##
## @arg:    $mount_dir     Required: Mount dir on client
##
## @arg:    $private_ip    Required: Private IP to be used for mount
##
## @arg:    $volume        Required: Volume to be mounted.     
##
## @synopsis
## my $status = VSeries::utils::mount_test_vol($client_obj,$mount_dir,$private_ip,$volume);
##
sub mount_test_vol($$$$) {
    my ($client_object,$mount_dir,$private_ip,$volume) = @_;

    my @client_cmd_out = $client_object->send_client_cmd( "ls ". $mount_dir );

    # Check if output has the strings 'No such file or directory'
    foreach my $line (@client_cmd_out) {
        if ($line =~ /No such file or directory/) {
            # If the dir does not exist then create it
            my $mkdir_status = $client_object->create_dir( $mount_dir );
            if ($mkdir_status == $ERR_STATUS) {
                $Log->debug("ERROR: could not create client dir $mount_dir");
                return $ERR_STATUS;
            }
        }
    }

    # NFS mount test volume on client
    my $mount_status = $client_object->set_mount("$private_ip:/$volume  $mount_dir");

    if ($mount_status == $ERR_STATUS) {
        $Log->debug("ERROR: could not mount dir $private_ip:/$volume on $mount_dir");
        return $ERR_STATUS;
    }
    return $GOOD_STATUS;
}

#-------------------------------------------------------------------------------
## @name ::unmount_test_vol($$$)
##
## @author Hari Prasad M (Hari.M@NetApp.com)
##
## @description
## Method to unmount volume mounted at the fixed mount point.
## Returns $GOOD_STATUS on success and $ERR_STATUS on failure.
##
## @arg:    $client_obj    Required: Client Object
##
## @arg:    $private_ip    Required: Private IP of Vserver
##
## @arg:    $volume        Required: Volume to unmount
##
## @synopsis
## my $status = VSeries::utils::unmount_test_vol($client_obj,$private_ip,$volume);
##
sub unmount_test_vol($$$) {
    my ($client_object,$private_ip,$volume) = @_;

    # Check if volume is mounted
    my @output = $client_object->send_client_cmd("mount");
    foreach my $line (@output) {
        if($line =~ /$private_ip/) {
            my $unmount_status = $client_object->set_unmount("$private_ip:/$volume");
            if ($unmount_status == $ERR_STATUS) {
                $Log->debug("ERROR: could not unmount client dir $private_ip:/$volume");
                return $ERR_STATUS;
            }
            last;
        }
    }
    return $GOOD_STATUS;
}

#-------------------------------------------------------------------------------
## @name ::start_io_to_volume($$$$;$$$$)
##
## @author Mathew Cattel (Matthew.Cattel@netapp.com)
##         Modified by Arvind Chandrasekaran
##
## @description
## Method to run IO (dt) at the fixed mount point that should have $Volume_Name
## mounted there by calling the appropriate subroutines.  Most of this code was
## taken directly from the CFOD tests.  It returns the process ids of all the
## IO threads kicked off for subsequent monitoring in other subroutines.  We
## do not want to test to pend and just monitor IO, actually we want to do port
## disables and enables during this.
## Returns an array with reference to DT PIDs and the DT Log Files on success
## and returns $ERR_STATUS on failure.
## This sub routine has been copied from vser_mkby_static.thpl and modified.
##
## @arg:    $client_object    Required: Client object
##
## @arg:    $mount_dir        Required: Mount dir on which IO is to be run
##
## @arg:    $disk_pct_fill    Required: Percentage of disk to fill with IO
##
## @arg:    $dt_path          Required: Path to dt binary on client
##
## @arg:    $ref_dt_pattern   Optional: Ref to array with DT Patterns
##
## @arg:    $io_blocksize     Optional: IO Block size to be used
##
## @arg:    $log_suffix       Optional: Suffix to add for logs
##
## @arg:    $runtime          Optional: Runtime for IO
##
## @synopsis
## my ($dt_log, $ref_to_pids) = 
##   VSeries::utils::start_io_to_volume($client_object, $mount_dir, $disk_pct_fill, $dt_path, $ref_dt_pattern, $io_blocksize, $log_suffix, $runtime);
##
sub start_io_to_volume($$$$;$$$$){
    my ($client_object, $mount_dir, $disk_pct_fill, $dt_path, $ref_dt_pattern, $io_blocksize, $log_suffix, $runtime) = @_;

    # Set the default values for optional params
    if (!defined $ref_dt_pattern) {
        $ref_dt_pattern = ["0xdeadbeef", "0x0ffabff0", "0xa0b1c2d3", "0xffa0bced", "0xdeaf9876", "0xfeab7291",
                           "0x0f0f0f0f", "0x2729eadf", "0x83629ad2", "0xabcdef00"];
    }

    if (!defined $io_blocksize) {
        $io_blocksize = "4k";
    }

    if (!defined $log_suffix) {
        $log_suffix = "";
    }
    
    if (!defined $runtime) {
        $runtime = 1000;
    }

    my @dt_pattern_arr = @{$ref_dt_pattern};
    my @dt_job_pids = ();
    my @dtpids = ();

    my $mkdir_status = $client_object->create_dir("/dtlogs");
    if ($mkdir_status) {
        $Log->debug("ERROR: could not create client dir /dtlogs");
        return ($ERR_STATUS, undef);
    }

    my @client_cmd_out = $client_object->send_client_cmd("df $mount_dir");

    my $dt_log = undef;
    my $dt_pattern_counter = 0;

    # Sample df output
    # [root@vgibrh4x86c37 ~]# df
    # Filesystem           1K-blocks      Used Available Use% Mounted on
    # /dev/sda1             20161172  18198368    938664  96% /
    # none                   2537796         0   2537796   0% /dev/shm
    # rtpinf01.rtp.netapp.com:/vol/software/arch/x86_64-redhat-rhel4
    #                      429496736 266979680 162517056  63% /usr/software
    # rtpinf02.rtp.netapp.com:/vol/localtest
    #                      394264576 181676224 212588352  47% /x/eng/localtest

    foreach my $line (@client_cmd_out) {

        # Extract available disk space and mounted filesystem info
        if ( $line =~ /^\s+\d+\s+\d+\s+(\d+)\s+\d+%\s+\/(\w+)\/(\w+)\/(\w+).*\n?$/ ||
             $line =~ /^.*:\/.*\s+\d+\s+\d+\s+(\d+)\s+\d+%\s+\/(\w+)\/(\w+)\/(\w+).*\n?$/) {

            # $1 = Available disk space  /$2/$3/$4 = /var/tmp/mccfoddirX
            my $disk_size = $1 * int(1024 * ($disk_pct_fill / 100));

            my $dt_file = "/$2/$3/$4/" . $4 . "$log_suffix";
            my $test_dir_name = $4;
            # Store various dt parameters in hashes so we can more easliy
            # restart i/o later after failover

            $dt_log = "/dtlogs/" . $test_dir_name . "$log_suffix";

            @dtpids = $client_object->run_io($dt_path . "dt", "of=$dt_file bs=$io_blocksize runtime=$runtime dispose=delete limit=$disk_size pattern=$dt_pattern_arr[($dt_pattern_counter++ / 10)] log=$dt_log enable=compare errors=1");
            # Reinit pids array for next loop pass
            push (@dt_job_pids, @dtpids);
            @dtpids = ();
        }
    }
    #Arvind
    $Log->comment("$dt_log");
    return ($dt_log, \@dt_job_pids);

}

#-------------------------------------------------------------------------------
## @name ::check_io_status($$)
##
## @author Mathew Cattel (Matthew.Cattel@netapp.com)
##         Arvind Chandrasekaran 
## @description
## Method that checks to see if a set of process IDs
## are still running on the host.  Any that aren't found
## are returned to the caller.
## This sub routine has been copied from vser_mkby_static.thpl
## and modified
##
## @arg:    $client_obj    Required: Client object
##
## @arg:    $pid_ar_ref    Required: Ref DT PIDs array to check
##
## @synopsis
## my $unmatched_pids_array_ref = &check_io_status(\@dt_pids);
##
sub check_io_status($$){

    my ($client_object, $pid_arr_ref) = @_;
    my @pids_to_check = @$pid_arr_ref;
    my @pid_checked_flags = ();

    my @client_cmd_out = $client_object->send_client_cmd("ps ax");

    for(my $i = 0; $i < scalar(@pids_to_check); $i++){
        $pid_checked_flags[$i] = 0;
        foreach my $line (@client_cmd_out){

            if($line =~ m/^\s*(\d+)/){
                if($1 == $pids_to_check[$i]){
                    $pid_checked_flags[$i] = 1;
                    last;
                }
            }
        }
    }

    my @unmatched_pids = ();
    for(my $j = 0; $j <scalar(@pid_checked_flags); $j++){
        if($pid_checked_flags[$j] == 0){
            my $checked_pid = $pids_to_check[$j];
            push(@unmatched_pids, $checked_pid);
        }
    }

    return \@unmatched_pids;

}

#-------------------------------------------------------------------------------
## @name ::check_io_completion($$)
##
## @author Mathew Cattel (Matthew.Cattel@netapp.com)
##         Arvind Chandrasekaran
##
## @description
## This is basically a companion subroutine.  If all pids
## are returned from check_io_status then call this subroutine
## to ensure that all IO completed cleanly.  Check the DT log
## and if there are no errors return GOOD_STATUS otherwise return
## ERR_STATUS.
## This sub routine has been copied from vser_mkby_static.thpl
## and modified.
##
## @arg:    $client_object    Required: Client object
##
## @arg:    $dt_log           Required: DT Log to check
##
## @synopsis
## my $io_status = &check_io_completion($client_object, $dt_log);
##
sub check_io_completion($$){
    my ($client_object, $dt_log) = @_;

    my @client_cmd_out = $client_object->send_client_cmd("cat $dt_log");

    foreach my $line ( @client_cmd_out ) {
        if ( $line =~ /Stale NFS file handle/ || $line =~ /Total errors detected: 1\/1/ ) {
            return $ERR_STATUS;
        }
    }

    return $GOOD_STATUS;
}

#-------------------------------------------------------------------------------
## @name :: scsi_fitp_block_lun($$$$)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## This sub routine is used block the specified lun on all the paths
## on a 7 mode filer.
##
## @param $api_set_obj Required: NACL API Set Object (7-mode)
##
## @param $disk        Required: The disk name to be scsi fitp blocked
##
## @param $sasc_ref    Required: Ref to storage array show config output
##
## @param $sasl_ref    Required: Ref to storage array show luns output
##
## @param $ssd_ref     Optional: Ref to storage show disk output. This
##                               will help handling load balance disk name
##                               changes.
##
## @synopsis
## my $return_ref = &scsi_fitp_block_lun($api_set_obj, $disk, $sasc_ref, $sasl_ref);
##
sub scsi_fitp_block_lun ($$$$;$) {
    my $api_set_obj = shift;
    my $disk = shift;
    my $sasc = shift;
    my $sasl = shift;
    my $ssd = shift;

    my $lun = undef;
    if ($disk =~ /L(\d+)$/) {
        $lun = $1;
    }
    my @ret_array = ();

    # Get the storage array show-config output
    my @sasc_array = @{$sasc};

    # Get the storage array show-luns output.
    # This output will have details of the targets that each luns is mapped from
    my @sasl_array = @{$sasl};

    # If storage show disk -p all output is also available,
    # Then get both the primary and secondary paths for the disk
    my $primary = $disk;
    my $secondary = $disk;
    if (defined $ssd) {
        foreach my $tmp_disk (@{$ssd}) {
            if ($tmp_disk->{'primary'} eq $disk || $tmp_disk->{'secondary'} eq $disk) {
                $primary = $tmp_disk->{'primary'};
                $secondary = $tmp_disk->{'secondary'};
            }
        }
    }

    # Get the targets for the specified lun from the sasl output
    my @disk_target_wwpns = ();
    foreach my $sasl_entry (@sasl_array) {
        if ($sasl_entry->{'name'} eq $primary || $sasl_entry->{'name'} eq $secondary) {
            @disk_target_wwpns = @{$sasl_entry->{'wwpns'}};
            last;
        }
    }
    @disk_target_wwpns = sort @disk_target_wwpns;
    if (scalar @disk_target_wwpns == 0) {
        @ret_array = ($ERR_STATUS, "Could not find the lun $disk in storage array show-luns output");
        return \@ret_array;
    }

    # Iterate over the sasc output and identify the lun group that has the same set of
    # target ports as the identified lun. This would provide us with the initiators
    # that are mapped to the identfied targets
    # Then block the lun with the identified target and initiator pair using scsi fitp block
    my @initiators = ();
    my @target_wwpns = ();
    foreach my $sasc_entry (@sasc_array) {
        @target_wwpns = @{$sasc_entry->{'array_target_ports'}};
        my @target_wwpns_for_compare = sort @target_wwpns;
        @initiators = @{$sasc_entry->{'initiator'}};
        if (Compare(\@target_wwpns_for_compare,\@disk_target_wwpns)) {
            my $i=0;

            while ($i <= $#target_wwpns) {
                my $api = "execute_raw_command";
                my $lun_id = undef;
                my $options = {'command' => "priv set test\; scsi fitp block -r $lun -p $target_wwpns[$i] -i $initiators[$i] -c"};

                my $status = VSeries::Clustered::nacl_convenience::execute_cmd($api_set_obj, $api, $options);
                if ($status == $ERR_STATUS) {
                    # Get the ONTAP command name by replacing _ with <space> in the api name
                    my $cmd = $api;
                    $cmd =~ s/_/ /g;
                    @ret_array = ($ERR_STATUS, "$cmd execution failed");
                    return \@ret_array;
                }
                $i++;
            }
            sleep 60;
            last;
        }
    }
    @ret_array = ($GOOD_STATUS, "$disk blocked successfully");
    return \@ret_array;
}

########################################################################
#
# @name ::check_degraded_vols($$;$$)
#
# @author: Pranav Ojha (pranavo@netapp.com)
#
# @description:
# This routine return $GOOD_STATUS when all volumes passed to it are
# in degraded state. Otherwise, it return $ERR_STATUS.
#
# @args: $filer_obj (Required) Filer object.
#
# @args: $saved_mirrored_vols_ref (Required) Reference to a list of 
#           volume names.
#
# @args: $vol_sleep_time (Optional) Defines sleep time between status
#           checks. Defaults to 30.
#
# @args: $max_vol_status_retries (Optional) Defines number of retries
#            to ensure that degraded status is reached. Defaults to 10
#
# @usage: $status = check_degraded_vols($filer_obj,\@vol_names);
#
########################################################################

sub check_degraded_vols ($$;$$)
{
    my $filer_obj = shift;
    my $saved_mirrored_vols_ref = shift;
    my $max_vol_status_retries = shift;
    my $vol_sleep_time = shift;
    $vol_sleep_time = 30 if( not defined $vol_sleep_time);
    $max_vol_status_retries = 10 if( not defined $max_vol_status_retries);
    
    my @saved_mirrored_vols = @{$saved_mirrored_vols_ref};
    my $vol_status_retries = 0;
    my $degraded_vol = 0;
    my $filer = $filer_obj->{'name'};
    
    foreach my $vol_name (sort @saved_mirrored_vols) {
        while ( ! $degraded_vol ) {
            # Do a new vol status and parse status to get the
            # volume condition
            my $vol_status_ref = $filer_obj->vol_status();
            my @vol_status_out = @$vol_status_ref;
    
            if ( $filer_obj->get_error_state() == $ERR_STATUS ) {
                my $errmsg = $filer_obj->get_error_msg();
                Tharn::logcomment("ERROR: volume status error: $errmsg");
                return $ERR_STATUS;
            }
    
            # Need new parse_vol_hash1 reference for inner while
            # loop that won't collide with hash refernce for outer
            # foreach loop
            my $parse_vol_hash_ref1 = VSeries::utils::parse_vol_status("", @vol_status_out);
            my %parse_vol_hash1 = %$parse_vol_hash_ref1;
            my $vol_condition = $parse_vol_hash1{$vol_name}{condition};
            my $vol_state = $parse_vol_hash1{$vol_name}{state};
    
            if ( $vol_condition !~ /degraded/ ) {
                if ( $vol_status_retries > $max_vol_status_retries ) {
                    Tharn::logcomment("Volume $vol_name on filer $filer never went to degraded condition as expected\nCurrent condition: $vol_condition");             
                    return $ERR_STATUS;
                } else {
                    $vol_status_retries++;
                    Tharn::logcomment("Retry vol status waiting for $vol_name to go to degraded state");
                    sleep $vol_sleep_time;
                }
            } else {
                # Check volume state is online 
                if ( $vol_state =~ /online/ ) {
                    # This volume is degraded, go on to 
                    # next one
                    Tharn::logcomment("Volume $vol_name is online and degraded");      
                    $degraded_vol = 1;                                                        
                } else {
                    Tharn::logcomment("Volume $vol_name on filer $filer is degraded but not online");
                    return $ERR_STATUS;                                                        
                }                         
            } # if $aggr_condition !~ /degraded/
    
        } # end while $degraded_vol
    
        # Reset flag and retry count as we check the next vol
        $degraded_vol = 0;
        $vol_status_retries = 0;
    
    } # end foreach $vol_name for mirrored vols
    
    # If we are here, we have all mirrored volume in degraded status
    return $GOOD_STATUS;
}

########################################################################
#
# @name ::check_degraded_aggrs($$;$$)
#
# @author: Pranav Ojha (pranavo@netapp.com)
#
# @description:
# This routine return $GOOD_STATUS when all aggregates passed to it are
# in degraded state. Otherwise, it return $ERR_STATUS.
#
# @args: $filer_obj (Required) Filer object.
#
# @args: $saved_mirrored_aggrs_ref (Required) Reference to a list of 
#           aggr names.
#
#
# @args: $aggr_sleep_time (Optional) Defines sleep time between status
#           checks. Defaults to 30.
#
# @args: $max_aggr_status_retries (Optional) Defines number of retries
#            to ensure that degraded status is reached. Defaults to 10.
#
# @usage: $status = check_degraded_aggrs($filer_obj,\@aggr_names);
#
########################################################################
sub check_degraded_aggrs ($$;$$)
{
    my $filer_obj = shift;
    my $saved_mirrored_aggrs_ref = shift;
    my $max_aggr_status_retries = shift;
    my $aggr_sleep_time = shift;
    $aggr_sleep_time = 30 if( not defined $aggr_sleep_time);
    $max_aggr_status_retries = 10 if( not defined $max_aggr_status_retries);
    
    my @saved_mirrored_aggrs = @{$saved_mirrored_aggrs_ref};
    my $aggr_status_retries = 0;
    my $degraded_aggr = 0;
    my $filer = $filer_obj->{'name'};
    
    foreach my $aggr_name (sort @saved_mirrored_aggrs) {
        while ( ! $degraded_aggr ) {
            # Do a new aggr status and parse status to get the
            # aggregate condition
            my $aggr_status_ref = $filer_obj->aggr_status();
            my @aggr_status_out = @$aggr_status_ref;
    
            if ( $filer_obj->get_error_state() == $ERR_STATUS ) {
                my $errmsg = $filer_obj->get_error_msg();
                Tharn::logcomment("ERROR: aggregate status error: $errmsg");
                return $ERR_STATUS;
            }
    
            # Need new parse_aggr_hash1 reference for inner while
            # loop that won't collide with hash refernce for outer
            # foreach loop
            my $parse_aggr_hash_ref1 = VSeries::utils::parse_aggr_status("", @aggr_status_out);
            my %parse_aggr_hash1 = %$parse_aggr_hash_ref1;
            my $aggr_condition = $parse_aggr_hash1{$aggr_name}{condition};
            my $aggr_state = $parse_aggr_hash1{$aggr_name}{state};
    
            if ( $aggr_condition !~ /degraded/ ) {
                if ( $aggr_status_retries > $aggr_sleep_time ) {
                    Tharn::logcomment("Aggregate $aggr_name on filer $filer never went to degraded condition as expected\nCurrent condition: $aggr_condition");             
                    return $ERR_STATUS;
                } else {
                    $aggr_status_retries++;
                    Tharn::logcomment("Retry aggr status waiting for $aggr_name to go to degraded state");
                    sleep $aggr_sleep_time;
                }
            } else {
                # Check aggregate state is online 
                if ( $aggr_state =~ /online/ ) {
                    # This aggregate is degraded, go on to 
                    # next one
                    Tharn::logcomment("Aggregate $aggr_name is online and degraded");       
                    $degraded_aggr = 1;                                                        
                } else {
                    Tharn::logcomment("Aggregate $aggr_name on filer $filer is degraded but not online");
                    return $ERR_STATUS;                                                        
                }                         
            } # if $aggr_condition !~ /degraded/
    
        } # end while $degraded_aggr
    
        # Reset flag as we check the next aggr
        $degraded_aggr = 0;
        $aggr_status_retries = 0;
    
    } # end foreach $aggr_name for mirrored aggrs
}

########################################################################
#
# @name ::get_day_of_week($)
#
# @author: Pranav Ojha (pranavo@netapp.com)
#
# @description:
# This routine returns day of week for date passed. The format of return
# will be 'Mmm'. Month needs to be supplied in MMM format, see eg below.
#
# @args: $date (Required) Date in DD MMM YYYY format.
#
# @usage: $status = VSeries::utils::get_day_of_week("14 Oct 2011");
#         This would return "Fri"
#
########################################################################
sub get_day_of_week($)
{
    my $date = shift;
    my @days = ('Sun','Mon','Tue','Wed','Thu','Fri','Sat');
    my %months = ('Jan' => 31,'Feb' => 29,'Mar' => 31,'Apr' => 30,'May' => 31,'Jun' => 30,
                'Jul' => 31,'Aug' => 31,'Sep' => 30,'Oct' => 31,'Nov' => 30,'Dec' => 31);
    
    my ($mday,$mon,$year) = split(/ /, $date);
    # Date format validity checks
    if( (!defined($months{$mon})) or ($mday < 1 && $mday > $months{$mon}) or 
        ($year < 1970 && $year > 2100) ){
            return $ERR_STATUS;
    }
    
    my $time = timelocal(0,0,0,$mday,$mon,$year);
    my $wday = (localtime($time))[6];

    return $days[$wday];
}

## @name ::compare_dates_in_different_formats($$)
##
## @description
## Takes two dates in different formats and returns GOOD_STATUS
## if the first one is more recent.It is expected that the dates
## passed belong to the same timezone.
##
## @param    $date1    Required: Date in the format
##                               1/16/2012 10:09:54
## @param    $date2    Required: Date in the format
##                               Mon Jan 16 10:09:04 GMT 2012
## @synopsis
## $compare = &compare_dates_in_different_formats($date1,$date2);

sub compare_dates_in_different_formats($$){
    my ($date1 , $date2) = @_;

    my %month_vals = ('JAN' => 1, 'FEB' => 2, 'MAR' =>3, 'APR' => 4, 'MAY'  => 5,
                      'JUN' => 6,'JUL' => 7, 'AUG' => 8, 'SEP' => 9, 'OCT'  => 10,
                      'NOV' => 11, 'DEC' => 12);

    $date1 =~ m/(\d+)\/(\d+)\/(\d+) (.*)/;

    my ($date1_mm, $date1_dd, $date1_time) = ($1, $2, $4);

    $date2 =~ m/\w+\s+(\w+)\s+(\d+)\s+((\d+|:)*)/;

    my ($date2_mm, $date2_dd, $date2_time) = ($month_vals{uc($1)}, $2, $3);

    my ($hour1, $min1, $sec1)  = (split /:/, $date1_time);
    my ($hour2, $min2, $sec2) = (split /:/ , $date2_time);

    #
    # Modified the logic to address the issue reported in the
    # BURT #706710. Please refer to the BURT for more information.
    # We validate it for the following scenarios:
    # 1) If '$date1_mm == $date2_mm' then we perform the following
    #    steps:
    #       a) If '$date1_dd == $date2_dd' then we perform
    #          the following steps:
    #             i) If '$hour1 > $hour2', return GOOD_STATUS.
    #            ii) Elsif '$hour1 == $hour2 && $min1 > $min2',
    #                return GOOD_STATUS.
    #           iii) Elsif '$hour1 == $hour2 && $min1 == $min2 && $sec1 >= $sec2',
    #                return GOOD_STATUS.
    #       b) Elsif '$date1_dd > $date2_dd', return GOOD_STATUS.
    # 2) Elsif '$date1_mm > $date2_mm', return GOOD_STATUS.
    # 3) Otherwise, return ERR_STATUS.
    #
    if ($date1_mm == $date2_mm) {
       if ($date1_dd == $date2_dd) {
          return $GOOD_STATUS if (($hour1 > $hour2) ||
                                  ($hour1 == $hour2 &&
                                   $min1 > $min2) ||
                                  ($hour1 == $hour2 &&
                                   $min1 == $min2 &&
                                   $sec1 >= $sec2));
       } elsif ($date1_dd > $date2_dd) {
          return $GOOD_STATUS;
       }
    } elsif ($date1_mm > $date2_mm) {
       return $GOOD_STATUS;
    }

    return $ERR_STATUS;
}

								     

1;
