#
# $Id: //depot/prod/test/Rsierranevada/lib/VifMgr/Failover/PortFail.pm#1 $
# $Revision: #1 $ submitted on $Date: 2013/01/22 $ By $Author: build $
#
## @summary     VifMgr PortFail Class
## @author      ewing@netapp.com
## @pod here

=head1 NAME

VifMgr::Failover::PortFail

=head1 SYNOPSIS

use VifMgr::Failover::PortFail;

my $vfm = new Vif(fail_device => "port_foo");

$vfm->failover_event();

=head1 DESCRIPTION

This is a class for VIF PortFail testing.  It combines several features shared
with VifMgr::Failover::Migrate through the parent class: VifMgr::Failover.
These features include ping validation, gratuitous ARP checking and pre/post
failover state checks.

=head1 METHODS

=cut

package VifMgr::Failover::PortFail;

use strict;
use warnings;

use TracePkg;
use Options;
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate SCALAR ARRAYREF OBJECT BOOLEAN UNDEF);
use Carp qw(confess);
use Server;

use base 'VifMgr::Failover';

my $Trace = new TracePkg();
$Trace->set_options(separateNewline => 1);

# Global variables utilized to track which VIFs have been manipulated, as
# well as their original failover rule layout for future comparisons.

my %Vifs_mod;
my %all_failover_vifs;

=head2 new

my $vfm = new VifMgr::Failover::PortFail(vif => 'vif0', vserver => 'vs0');

C<new> is the constructor for VifMgr::Failover::PortFail objects.  This class
provides a simple interface for testing VIF PortFailure failover operations.  
For VIF PortFail testing, a vif and vserver or address is required.  This vif's
configuration can be supplied in the constructor call if the interface is to 
be created.  Aside from interface selection, you can also select the kind of
migrations that will be performed on the interface.

These are the options that can be supplied:

Failure Port Selection:

fail_devide => <failure_device>    # such that can be passed to power_cycle
fail_node   => <fail_node>         # the node whose port will fail
fail_port   => <fail_port>         # the port which will fail

There are additional options that may be specified, but are documented as 
part of VifMgr::Failover's documentation.

=cut

#############################################################################
# Subroutine: new
# Parameters: Defaults
# Returns: Object reference
# Purpose: PortFail Failover Constructor
###############################################################################

sub new {

    $Trace->entry();
    my $class = shift(@_);

    # Call the parent constructor

    my $self = $class->SUPER::new(@_);
    my $node = $self->{_options}->get(option => 'FILER');

    my %p = (@_);
    $self->{_fail_device}  = $p{fail_device};
    $self->{_fail_vif}     = $p{vif};
    $self->{_fail_vserver} = $p{vserver};
    $self->{_fail_role}    = $p{role};
    $self->{_fail_node}    = $p{fail_node};
    $self->{_fail_port}    = $p{fail_port};
    $self->{_fail_type}    = $p{fail_type} || $self->POWER_CYCLE;
    $self->{_check_vifs}   = $p{check_vifs};

    # fail_grp is a special case parameter that will be used if we are
    # constructing this object to manipulate failover groups for all data VIFs
    # on all nodes in the cluster. It's basically just a flag so that we
    # can successfully create the object.

    $self->{_fail_grp} = (defined $p{fail_grp}) ? $p{fail_grp} : 0;

    # If the user specified the VIF and Vserver, we are going to extract the node
    # and port from there

    if (defined($self->{_fail_vif}) && (defined($self->{_fail_vserver}))) {
        my $vif_info = $self->{_node}->network_interface_show(
            vif    => $self->{_fail_vif},
            server => $self->{_fail_vserver}
        );
        $self->{_fail_node} = $vif_info->[0]->{'curr-node'};
        $self->{_fail_port} = $vif_info->[0]->{'curr-port'};
    }

    # check to see if enough info was specified.

    if (
        !(
               defined($self->{_fail_device})
            || !defined($self->{_fail_node})
            || !defined($self->{_fail_port})
        )
        || !defined($self->{_fail_grp})
      ) {
        $Trace->fatal(
            msg => "Not enough information provided for the "
              . "failover port: ",
            result => "CONFIG"
        );
    } ## end if (!(defined($self->{...

    $Trace->comment(msg => "Created a new $class object");
    bless($self, $class);

    $Trace->exit(msg => "Created a new $class object: $self");
    return $self;
} ## end sub new

#############################################################################
# Subroutine: _get_port_vifs
# Parameters: None
# Returns: List of Addresses
# Purpose: List of VIFs on the current node/port
#############################################################################

sub _get_port_vifs {

    $Trace->entry();
    my $self = shift();

    # We are only concerned with vifs on this port with the same role. This is
    # because there are cases where VIFs with different subnets can be on the
    # same port (i.e. a cluster-mgmt VIF can reside on a data port). Pinging 
    # VIFs on different subnets could give us erroneous results. 

    my $vifs = $self->{_node}->network_interface_show(
        'curr-node' => $self->{_fail_node},
        'curr-port' => $self->{_fail_port},
        role        => $self->{_fail_role},
    );

    $Trace->exit();
    return (
        wantarray()
        ? map({$_->{address}} @{$vifs})
        : [map({$_->{address}} @{$vifs})]
    );
} ## end sub _get_port_vifs

#############################################################################
# Subroutine: failport_available
# Parameters: Vifs and vserver to fail
# Returns: Node and port
# Purpose: Returns the next node/port combination that can be failed over
#          in reguards to its failover policy.
#
#          This method is atrocious.  It really ought to be broken out into
#          a separate class(es) and simplified.
#############################################################################

sub failport_available {

    my ($self, %p) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {
            vif     => {type => SCALAR},
            vserver => {type => SCALAR},
        }
    );
    $Trace->entry(msg => "Looking for an available failport for "
          . "$p{vserver}\:$p{vif}");

    my $vif_info = $self->{_node}->network_interface_show(
        vif    => $p{vif},
        server => $p{vserver}
    );
    my $fo_info = $self->{_node}->network_interface_failover_show(
        vif    => $p{vif},
        server => $p{vserver}
    );

    my @priority_sort = sort({$a->{priority} <=> $b->{priority}} @{$fo_info});
    $Trace->comment(msg => "Priority sorted port list: "
          . join(", ", map({"$_->{node}\:$_->{port}"} @priority_sort)));

    # This is difficult to understand.  It would be best if it could be
    # simplified, but it's complexity relies on the complexity of vif vailover
    # rules/policies.

    # Check the local port first if priority is the policy.

    if ($vif_info->[0]->{'failover-policy'} eq 'priority') {
        $Trace->debug(msg => "Trying priority failover policy.");

        # Look through our sorted port list

        foreach my $fp (@priority_sort) {
            $Trace->debug(
                msg => "Checking listed port: $fp->{node}\:$fp->{port}");

            # If we're on the same node as the vif, check to see if the port is up

            if ($fp->{node} eq $vif_info->[0]->{'curr-node'}) {

                $Trace->debug(
                    msg => "Port $fp->{node}\:$fp->{port} is local to "
                      . "vif $p{vserver}\:$p{vif}.");

                if (ref($fp->{'location'}) eq 'ARRAY') {
                    if (grep(/current/, @{$fp->{'location'}})) {
                        $Trace->debug(msg => "Port $fp->{node}\:$fp->{port} is "
                              . "the current port for $p{vserver}\:$p{vif}, skipping."
                        );
                        next();
                    }
                }

                # Get the port information

                my $p_info = $self->{_node}->network_port_show(
                    node => $fp->{node},
                    port => $fp->{port}
                );

                # If the port is up, it's the one to which our vif will move

                if ($p_info->[0]->{'link'} eq 'up') {
                    $Trace->exit(msg => "Found failport ($p_info->[0]->{node},"
                          . "$p_info->[0]->{port}) for vif $p{vserver}\:$p{vif}"
                    );
                    return ($p_info->[0]->{node}, $p_info->[0]->{port});
                }

            } else {
                $Trace->debug(
                    msg => "Port isn't on node local to vif, skipping.");
            }
        } ## end foreach my $fp (@priority_sort)
        $Trace->comment(msg =>
              "Unable to find a priority port, continuing to nextavail policy");
    } ## end if ($vif_info->[0]->{'failover-policy'...

    # Either the policy isn't priority, or local ports to this one aren't
    # available.

    if ($vif_info->[0]->{'failover-policy'} eq 'nextavail') {
        $Trace->debug(msg => "Trying nextavail failover policy.");

        # Look through our sorted port list

        foreach my $fp (@priority_sort) {

            $Trace->debug(
                msg => "Checking listed port: $fp->{node}\:$fp->{port}");

            # Make sure its not the current port

            if (ref($fp->{'location'}) eq 'ARRAY') {
                if (grep(/current/, @{$fp->{'location'}})) {
                    $Trace->debug(msg => "Port $fp->{node}\:$fp->{port} is the "
                          . "current port for $p{vserver}\:$p{vif}, skipping.");
                    next();
                }
            }

            # Make sure that port is up

            my $p_info = $self->{_node}->network_port_show(
                node => $fp->{node},
                port => $fp->{port}
            );

            # If the port is up, it's the one to which our vif will move

            if ($p_info->[0]->{'link'} eq 'up') {
                $Trace->exit(msg => "Found failport ($p_info->[0]->{node},"
                      . "$p_info->[0]->{port}) for vif $p{vserver}\:$p{vif}");
                return ($p_info->[0]->{node}, $p_info->[0]->{port});
            }

            $Trace->debug(
                msg => "Field port-info for $fp->{node}\:$fp->{port}");
        } ## end foreach my $fp (@priority_sort)
    } ## end if ($vif_info->[0]->{'failover-policy'...

    $Trace->exit(msg =>
          "Failed to find an available failport for vif $p{vserver}\:$p{vif}");
    return (undef(), undef());
} ## end sub failport_available

=head2 failover_event

$vfm->failover_event

C<failover_event> is the main testing method for VifMgr::Failover::PortFail 
objects.  The C<failover_event> method will select a port, verify that the
interface is up, listen for gratuitous ARPs, migrate the interface and finally
verify that the interface is once again pingable.

There are no options for this method (yet).

=cut

sub failover_event {

    $Trace->entry();
    my $self = shift();

    # Get a list of vifs on the fail port. This will basically issue a show based
    # on what VIFs are currently on the fail_node/port and return the addresses. Based
    # on those addresses (addresses of all VIFs that live on the node/port in
    # question), they (all of the vif info) will be added to the vif_watch_list

    my @before_vifs = $self->_get_port_vifs();

    # Watch the state of our interfaces

    foreach my $vif (@before_vifs) {
        $self->add_vif_watch(address => $vif);
    }
    $self->store_vif_state();

    # Verify the interface is up before failover

    $Trace->comment(msg => "Pinging vifs on port $self->{_fail_node}\:"
          . "$self->{_fail_port} (pre-migrate)");
    $self->ping_vif_watch_list();

    # Start gratuitous arp checks in the background before migration

    $Trace->comment(msg => "Starting gratuitous ARP check for vifs on port "
          . "$self->{_fail_node}\:$self->{_fail_port}");
    $self->garp_check_watch_list(skip_nonfailable => 1);

    if ($self->{_fail_type} eq $self->POWER_CYCLE) {

        # Cause power failure on the network port

        $Trace->comment(msg => "Interrupting link on port $self->{_fail_node}\:"
              . "$self->{_fail_port}");

        Tharn::subtest("/usr/local/bin/power_cycle", '--',
            $self->{_fail_device});
    } elsif ($self->{_fail_type} eq $self->DOWN_PORT) {

        # Take the port down

        $Trace->comment(
            msg => "Taking down $self->{_fail_node}\:" . "$self->{_fail_port}");
        $self->{_node}->network_port_modify(
            node       => $self->{_fail_node},
            port       => $self->{_fail_port},
            'up-admin' => "false",
        );
        
        sleep 10;
        $self->{_node}->network_port_show();

        # Bring the port back up

        $Trace->comment(
            msg => "Bringing Up $self->{_fail_node}\:" . "$self->{_fail_port}");
        $self->{_node}->network_port_modify(
            node       => $self->{_fail_node},
            port       => $self->{_fail_port},
            'up-admin' => "true",
        );
        $self->{_node}->network_port_show();

    } ## end elsif ($self->{_fail_type...

    # Check reachability following the failover.

    sleep(30);
    $Trace->comment(msg => "Pinging vifs on port $self->{_fail_node}\:"
          . "$self->{_fail_port} (post-migrate)");
    $self->ping_vif_watch_list();

    # Let the gratuitous ARP check finish & check its return value.

    $self->wait_garp_check_watch_list();

    # Check the vif state and make sure there are some differences. Should only
    # do in the case of a POWER_CYCLE. If DOWN_PORT is run, there may
    # be no differences if auto-revert is configured. Leave to calling method
    # to decide.

    my $difference;
    if ($self->{_check_vifs}) {
        $difference = $self->check_vif_state();
    }

    if ((!scalar(keys(%$difference))) && ($self->{_check_vifs})) {
        $Trace->fatal(
            msg => "Port Failure didn't cause any change in VIF state.");
    } else {
        while (my ($k, $v) = each(%{$difference})) {
            $Trace->debug(msg => "\t$k -> $v");
        }
    }

    foreach my $vif (@before_vifs) {
        $self->delete_vif_watch(address => $vif);
    }

    $Trace->exit();
} ## end sub failover_event

########################################################################
# Function:    modify_failover_group
#
# Description: Subroutine to handle disabling failover groups for a
#              particular set of vifs. The topology will be saved in
#              doing so for later comparison.
#
# Parameters:  Role in which to disable, state to modify to, optional
#              node and /or vif to modify
#
# Returns:     Success - 1
#              Failure - 0
#
# Notes:
#   - This function is used to allow the port roles to be modified
#     (i.e. cannot modify a role of a port if its vif is part of a
#     failover group)
########################################################################
sub modify_failover_group {

    $Trace->entry();
    my $self = shift();

    my %params = validate(
        @_,
        {
            role      => {type => SCALAR, optional => 1, default => "data"},
            new_state => {type => SCALAR, optional => 1, default => "disabled"},
            vif  => {type => SCALAR | UNDEF, optional => 1, default => undef},
            node => {type => SCALAR | UNDEF, optional => 1, default => undef},
        }
    );

    # Since vif and node are optional attributes, then we will set these
    # incase they remain undefined

    my $vif_attr  = (defined $params{vif})  ? "vif"           : undef;
    my $vif_val   = (defined $params{vif})  ? $params{vif}    : undef;
    my $node_attr = (defined $params{node}) ? "\'curr-node\'" : undef;
    my $node_val  = (defined $params{node}) ? $params{node}   : undef;

    # If the state that we are modifying to is "disable", then the
    # current state is enabled, and vice-a-versa

    my $curr_state = "enabled";
    if ($params{new_state} eq "enabled") {
        $curr_state = "disabled";
    }

    # Verify that we are running with the correct change, otherwise,
    # we need to exit. No error/warning necessary if this feature
    # doesn't even exist yet.

    if (!$self->{_node}->has_change(change => "failover-groups")) {
        $Trace->exit(msg => "Not running with change: failover-groups. "
              . "Returning without error...");
        return 1;
    }

    # If we are running with BR or beyond, we should be specifying "system-defined"
    # to enable the failover rules (with a default failover list) and not "enabled".
    # The behavior for "disabled" remains the same.

    if (   ($self->{_node}->has_change(change => "failover-groups-2"))
        && ($curr_state eq "enabled")) {
        $curr_state = "system-defined";
    }
    if (   ($self->{_node}->has_change(change => "failover-groups-2"))
        && ($params{new_state} eq "enabled")) {
        $params{new_state} = "system-defined";
    }

    # Get a list of the interfaces we want to modify (will always
    # be on data vifs)

    my @vifs = $self->{_node}->network_interface_show(
        role                 => "data",
        'use-failover-group' => $curr_state,
        $vif_attr            => $vif_val,
        $node_attr           => $node_val
    );

    # Iterate through the VIFs and modify the state

    for (@vifs) {

        # At this point, $_ is actually a hash reference, so get the actual
        # hash out

        my %this_vif = %{$_};
        $self->{_node}->network_interface_modify(
            server               => $this_vif{server},
            vif                  => $this_vif{vif},
            'use-failover-group' => $params{new_state}
        );

        # If we are disabling, we want to delete the failover rules as well
        if ($params{new_state} =~ m/^disabled$/) {
        	
			if ( !$self->{_node}->has_change( change => "lif-failover-cache" ) ) {
				
	            # Save the topology for later comparison	
	            if (!$self->save_failover_list(vif => $this_vif{vif})) {
	                $Trace->fatal(msg => "Unable to save failover list before "
	                      . "deletion for VIF: $this_vif{vif}",);
	            }
	
	            # Now we want to go though and delete all but the lowest priority
	            # failover rule for that VIF
	            my $priority = $self->get_lowest_priority(vif => $this_vif{vif});
	
	            eval {
	                $self->{_node}->network_interface_failover_delete(
	                    server   => $this_vif{server},
	                    vif      => $this_vif{vif},
	                    priority => ">$priority"
	                );
	            };
	            if ($@ =~ /There are no entries matching your query/) {
	                $Trace->comment(msg => "Failover rules > $priority do not "
	                      . "exist for VIF $this_vif{vif} on server "
	                      . "$this_vif{server}. Continuing...");
	            } elsif ($@) {
	                $Trace->fatal(msg => "Unexpected error while deleting "
	                      . "failover rules: $@");
	            }
    		}
        } else {

            # Since we enabled failover groups, that list should be
            # automatically regenerated, so we will compare it with the old
            # list here. No errors/warnings because the list is not always
            # generated the same, this is just for logging purposes.

            $self->compare_failover_list(vif => $this_vif{vif});
        }

        # Save VIF modified for future use (so we don't have to re-issue the
        # show command at a later date.

        $Vifs_mod{$this_vif{vif}} = 1;
    } ## end for (@vifs)

    # Display current configuration in case we need to debug any issues.
    if ( $self->{_node}->has_change( change => "lif-failover-cache" ) ) {
    	$self->{_node}->cli( cmd => 
				"network interface show -fields failover-targets "); 
    } else {
    	$self->{_node}->network_interface_failover_show();
    }
    $self->{_node}->network_interface_show();
    $self->{_node}->network_port_show();

    $Trace->exit(msg => "Modification to $params{new_state} successful");
    return 1;
} ## end sub modify_failover_group

########################################################################
# Function:    get_lowest_priority
#
# Description: Allows the calling subroutine to be aware of the lowest
#              priority for failover rules, incase we enounter a VIF
#              who's lowest priority is not 0.
#
# Parameters:  VIF name
#
# Returns:     Lowest Priority Number
#
# Notes:
########################################################################

sub get_lowest_priority {

    $Trace->entry();
    my $self = shift();

    my %params = validate(@_, {vif => {type => SCALAR},});

    # Perform the show

    my @vifs =
      $self->{_node}->network_interface_failover_show(vif => $params{vif});
    my $priority = -1;
    for (@vifs) {

        # At this point, $_ is actually a hash reference, so get the actual
        # hash out

        my %this_vif = %{$_};

        # Initialize on the first run, otherwise, save the lowest

        if ($priority == -1) {
            $priority = $this_vif{priority};
        } else {
            if ($this_vif{priority} < $priority) {
                $priority = $this_vif{priority};
            }
        }
    } ## end for (@vifs)

    $Trace->exit(msg => "Exiting with lowest priority of $priority");
    return $priority;
} ## end sub get_lowest_priority

########################################################################
# Function:    save_failover_list
#
# Description: Appends on this failover list to all_failover_vifs
#              for future comparison
#
# Parameters:  VIF name to append
#
# Returns:     Success - 1
#              Failure - 0
#
# Notes:
########################################################################

sub save_failover_list {

    $Trace->entry();
    my $self = shift();

    my %params = validate(@_, {vif => {type => SCALAR},});

    # Perform the show to get the list of hashes and save it

    my @pri_list =
      $self->{_node}->network_interface_failover_show(vif => $params{vif});

    if (scalar(@pri_list) <= 0) {
        $Trace->comment(msg => "Unable to save failover information for vif: "
              . $params{vif});
        return 0;
    }

    # Overwrite list for VIF if its existing (could be out of date)

    $all_failover_vifs{$params{vif}} = \@pri_list;

    $Trace->exit();
    return 1;
} ## end sub save_failover_list

########################################################################
# Function:    compare_failover_list
#
# Description: Compares current output for this vif to the one saved in
#              all_failover_vifs to verify that the list was regenerated
#              consistently. The list is usually not regenerated the same,
#              however, we are just logging it in case we ever want to
#              know.
#
# Parameters:  VIF name to compare
#
# Returns:     Success - 1
#              Failure - Comment to log
#
# Notes:
########################################################################

sub compare_failover_list {

    $Trace->entry();
    my $self = shift();

    my %params = validate(@_, {vif => {type => SCALAR},});

    # Perform the show to get the most recent information

    my @pri_list =
      $self->{_node}->network_interface_failover_show(vif => $params{vif});

    # First, check to see if this VIF previously had a failvoer list saved,
    # then we can extract its list.

    my @old_pri_list;
    if (defined($all_failover_vifs{$params{vif}})) {
        @old_pri_list = @{$all_failover_vifs{$params{vif}}};
    } else {
        $Trace->comment(msg => "No list exists for this VIF. Was not "
              . "previously modified by this library. Exiting without comparing."
        );
        return 1;
    }

    $Trace->debug(msg => "Comparing new list: "
          . Dumper(\@pri_list)
          . " to saved list: "
          . Dumper(\@old_pri_list));

    # Iterate through and perform the comparison, they will appear in the same
    # priority order

    my $err_msg   = "";
    my $pri_index = 0;
    for (@pri_list) {

        my %new_pri = %{$_};
        my %old_pri;

        if (defined($old_pri_list[$pri_index])) {
            %old_pri = %{$old_pri_list[$pri_index++]};
        } else {
            $Trace->exit(msg => "Previous priority list for vif "
                  . " $params{vif} did not exist. Exiting without "
                  . "comparing");
            return 1;
        }

        # Iterate through the keys and perform the comparison

        foreach my $key (keys %new_pri) {

            # Location is the only special case because it could be
            # an array reference (in case of home node/port)

            if ($key !~ m/^location$/) {
                if ($new_pri{$key} !~ m/^$old_pri{$key}$/) {
                    $err_msg .=
                        "Mismatched $key for VIF: $params{vif}, "
                      . "Priority: $new_pri{'priority'} Found: "
                      . "$new_pri{$key}, Expected: $old_pri{$key}.\n";
                }
            } else {
                if (ref($new_pri{$key}) eq 'ARRAY') {

                    foreach my $new_loc (@{$new_pri{$key}}) {

                        # Check to see that the old location is also
                        # an array. If not, we can error out due
                        # to mismatch

                        my $old_loc;
                        if (ref($old_pri{$key}) eq 'ARRAY') {
                            $old_loc = shift @{$old_pri{$key}};
                            if ($new_loc ne $old_loc) {
                                $err_msg .= "Mismatched $key for VIF: ".
                                    "$params{vif}, Priority: $new_pri{'priority'}".
                                    "Found: $new_loc, Expected: $old_loc.\n";
                            } 
                        } else {
                            $old_loc = $old_pri{$key};
                            if ($new_loc ne $old_loc) {
                                $err_msg .= "Mismatched $key for VIF: ".
                                    "$params{vif}, Priority: $new_pri{'priority'}".
                                    "Found: $new_loc, Expected: $old_loc.\n";
                            }                             
                        }
                    }
                } else {
                    if ($new_pri{$key} ne $old_pri{$key}) {
                        $err_msg .=
                            "Mismatched $key for VIF: $params{vif}, "
                          . "Priority: $new_pri{'priority'} Found: "
                          . "$new_pri{$key}, Expected: $old_pri{$key}.\n";
                    }
                }
            } ## end else [ if ($key !~ m/^location$/)
        } ## end foreach my $key (keys %new_pri)
    } ## end for (@pri_list)

    # Check if we have encountered any errors and return

    if ($err_msg eq "") {
        $Trace->exit(msg => "Priority list compared successfully");
        return 1;
    } else {
        $Trace->comment(msg => "Failed comparing new list: "
              . Dumper(\@pri_list)
              . " to saved list: "
              . Dumper(\@old_pri_list)
              . "\nFailed with: "
              . $err_msg);
        $Trace->exit();
        return 0;
    } ## end else [ if ($err_msg eq "")
} ## end sub compare_failover_list

########################################################################
# Function:    port_check
#
# Description: Compares the number of ports found in the failover
#              priority list to the number of available ports.
#
# Parameters:  none
#
# Returns:     Success - Nothing
#              Failure - Message via call to warn
#
# Notes:
#   - Gives the calling method a simple way to verify that all available
#     ports are being utilized.
########################################################################

sub port_check {

    $Trace->entry();
    my $self = shift();

    # If %Vifs_mod has not been set, we will generically initialize it
    # here. Otherwise, we already have a list of the VIFs that have
    # been modified

    if (scalar(keys %Vifs_mod) <= 0) {
        my @vifs = $self->{_node}->network_interface_show(role => "data");
        for (@vifs) {
            my %this_vif = %{$_};
            $Vifs_mod{$this_vif{vif}} = 1;
        }
    }

    # Get the output for both show commands to use during comparison.
    # Done on a per filer basis

    my @Filers = $self->{_node}->servers();
    foreach my $filer (@Filers) {

        # First, we want to get ther total number of available ports
        # for this filer

        my @port_list = $self->{_node}->network_port_show(
            role => "data",
            node => $filer
        );
        my %found_ports_total;
        for (@port_list) {
            my %port_output = %{$_};
            $found_ports_total{$port_output{port}} = 1;
        }
        my $orig_port_count = scalar(keys %found_ports_total);
        $Trace->debug(
            msg => "Origional port count for $filer is: " . $orig_port_count);

        # Now that we have the count, we can perform verification for each vif
		my $new_port_count;
		my @fail_list;
		
        foreach my $vif (keys %Vifs_mod) {
        	if ( $self->{_node}->has_change( change => "lif-failover-cache" ) ) {
        		    $Trace->comment(msg => "In path for lif-failover");
        		my $aoh = $self->{_node}->cli( cmd =>
					"network interface show -fields failover-targets -lif $vif"	);
				my @arr;
				$self->{_node}->_cli_output_handler( output_in => $aoh, want_array => \@arr );
				
				my $size = $#{ $arr[0]{'failover-targets'} } + 1;
				my @fail_list;
				
	            # Iterate through the fail_list so that we can get
	            # a count of the ports in the failover rules that exist on this filer.			
				for ( my $i = 0 ; $i < $size ; $i++ ) {
					if ( $arr[0]{'failover-targets'}[$i] =~ m/(.*)?:(.*)/ ) {
						if ($1 eq $filer) {
							push(@fail_list, $2);
						}
					} 
				}
				$new_port_count = $#fail_list + 1;
				
        	} else {
	            @fail_list = $self->{_node}->network_interface_failover_show(
	                node => $filer,
	                vif  => $vif
	            );
	
	            # Iterate through the fail_list so that we can get
	            # a count of the ports
				my %found_ports_priority;
	            for (@fail_list) {
	                my %priority_output = %{$_};
	                $found_ports_priority{$priority_output{port}} = 1;
	            }
	            $new_port_count = scalar(keys %found_ports_priority);
        	} 
            $Trace->debug(msg => "The port count on VIF: $vif, Filer: "
                  . "$filer is $new_port_count");

            # Perform check for this VIF

            if ($new_port_count != $orig_port_count) {
                $Trace->warn(msg => "Failover rules not utilizing all "
                      . "available ports for VIF: $vif, Filer: $filer. "
                      . "Found: $new_port_count, Expected: "
                      . "$orig_port_count. Compared: \n"
                      . Dumper(\@port_list)
                      . "\nto:\n"
                      . Dumper(\@fail_list));
            }
        } ## end foreach my $vif (keys %Vifs_mod)
    } ## end foreach my $filer (@Filers)

    $Trace->exit();
    return;
} ## end sub port_check



=head1 SEE ALSO

L<VifMgr::Failover>
L<VifMgr::Failover::Migrate>
L<Options>
L<TracePkg>
L<TestUtils>

=head1 BUGS

There are no known bugs here.  However, the code isn't well tested.

=head1 STATUS

  In-progress

=head1 AUTHOR

  Steven Ewing (ewing@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/VifMgr/Failover/PortFail.pm#1 $
  Last revised $Date: 2013/01/22 $

=cut

1;
