# this is basically -*- perl -*-
#
# $Id: //depot/prod/test/main/lib/VifMgr/LifMonitor.pm#1 $
#
# Copyright (c) 2009 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc
#
## @summary       Utility Class to check a LIF's position. The intention of this
##                library is to monitor a LIFs location while LIFs are being 
##                dynamically rebalanced throughout a cluster. This occurs when
##                the user enables DNS Load Balancing. 
##
## @author        dl-cluster-qa
##
## @pod           Here
##

=head1 NAME

LifMonitor.pm

=cut 

=head1 AUTHOR

dl-cluster-qa

=cut

=head1 DESCRIPTION

C<LifMonitor> is a utility library that can monitor and check a LIFs position
within a cluster. The intention of this is to verify that LIFs are or are not 
moving as expected when DNS Load Balancing's Auto Rebalance feature has been 
enabled. 

=cut

=head1 SYNOPSIS

use VifMgr::LifMonitor;
my $lMonitor = new VifMgr::LifMonitor();
if (!$lMonitor->save_all_state(key => "address", value_list => \@ips)) {
    $Test->nlog("- Unable to save lif state for ".Dumper(\@ips));
    return 0;
}
...
if (!$lMonitor->did_lifs_move(
            key        => "address", 
            value_list => \@ips,
            threshold  => $params{threshold},
            move       => $params{move},
        )) {
    $Test->nlog("- Unable to save lif state for ".Dumper(\@ips));
    return 0;
}

=cut

=head1 ARGUMENTS

None.

=cut

=head1 PUBLIC METHODS

=over 6

=item new

Parameters - None. Returns - Object Reference. Purpose - Constructor

=item save_state

Parameters - LIF key (name, address, etc.) and its corresponding value.
Returns - 1 on success, 0 on failure. Purpose - To take in a LIF key and 
value, and save that LIF's relevant information for future comparison.

=item save_all_state

Parameters - List of LIF names to save.  Returns - 1 on success, 0 on failure.
Purpose - Wrapper around save_state to handle multiple LIFs at once

=item get_state

Parameters - LIF name.  Returns - Hash reference containing stored information 
on said LIF. Purpose - To take in a LIF name and returned the stored information 
on it.

=item check_state

Parameters - LIF parameters, action (moved, stayed, etc.). Returns: 1 if the LIF 
remains in the expected state and 0 otherwise. Purpose: A utility that will check 
if a particular LIF has changed state (i.e. moved) or not.

=item did_lifs_move

Parameters - List of LIFs to check, threshold of failures to tolerate. Returns - 
 on success, 0 on failure. Purpose - A routine that when given a list of LIFs to 
check, will be able to tell if they have recently changed state (meaning if they 
have recently moved). If you do no plan on passing in a LIF list, you can call 
check_state directly. Note: There is a "threshold", because with DNS Load 
Balancing, it is very hard to give an exact number of how many LIFs we believe 
should remain stationary and/or how many LIFs we believe should have moved. It 
is all up to the calculations done on the node at a particular time.

=back

=cut

=head1 PRIVATE METHODS 

None.

=cut

package VifMgr::LifMonitor;

# Standard Libs

use strict;
use warnings;
use Data::Dumper;
use Switch;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);

# Netapp Specific Libs

use TracePkg;
use Server;

# Class Data, only accessible from this class
{
    my $refCount = 0;

    sub incrRef {
        $refCount++;
    }

    sub getRef {
        return $refCount;
    }
}

# Globals

my $Trace  = new TracePkg();
my $Server = new Server();
my $self;

#############################################################################
# Subroutine: new
# Parameters: None
# Returns: Reference to this object
# Purpose: To instantiate and configure the LifMonitor object
#############################################################################

sub new {

    $Trace->entry();
    my $class = shift(@_);

    # Check to see if this class has already been declared. If so, then we can
    # just return its reference.

    if (getRef()) {
        $Trace->exit(msg => "Package has already been declared. Returning...");
        return $self;
    }

    # Since this library is basically a data structure used to store LIF state,
    # we can just initialize the array used to hold each LIF's information here.

    $self->{_lif_state} = ();

    # Bless and return

    bless ($self, $class);
    incrRef();

    $Trace->exit();
    return $self;

} ## end sub new

###############################################################################
# Subroutine: save_state
# Parameters: LIF key (name, address, etc.) and its corresponding value.
# Returns: 1 on success, 0 on failure
# Purpose: To take in a LIF key and value, and save that LIF's relevant 
#          information for future comparison.
###############################################################################

sub save_state {

    $Trace->entry();
    my $self = shift(@_);

    # Extract any parameters. State can be saved based on uniquifiers such
    # as LIF name, or address.

    my %params = validate (
        @_,
        {
            key => {
                type     => SCALAR,
                optional => 1,
                default  => "lif",
            },
            value => {
                type     => SCALAR,
                optional => 0,
            },
        }
    );

    # Issue a show on the LIF in question. This should only return 1 LIF.

    my @output;

    eval {
        @output  = $Server->network_interface_show(
            $params{key} => $params{value},
        );
    };
    if ($@) {
        $Trace->fatal(msg => "Cannot save state for $params{key} $params{value}. ".
                      "Failed with: $@");
        return 0;
    }
    if (scalar(@output) > 1) {
        $Trace->warn(msg => "Query for $params{key} $params{value} returned more ".
                    "than one LIF. Will only save state for first LIF.");
    }
    
    # Break out the information that we need. Not saving everything, only 
    # what petains to LIF location
    
    my %save_this = (
        lif    => $output[0]->{lif},
        'curr-node'   => $output[0]->{'curr-node'},
        'curr-port'   => $output[0]->{'curr-port'},
        'status-oper' => $output[0]->{'status-oper'},
        'is-home'     => $output[0]->{'is-home'},
    );

    # Save the LIF's information. Overwriting anything pre-existing for this LIF
    # First check if anything is there, as this could be the first state saved.

    my %already_saved;
    if (defined($self->{_lif_state})) {
        %already_saved = %{$self->{_lif_state}};
    }

    $already_saved{$output[0]->{lif}} = \%save_this;
    $self->{_lif_state} = \%already_saved;

    $Trace->exit();
    return 1;

} ## end sub save_state

###############################################################################
# Subroutine: save_all_state
# Parameters: List of LIF names to save
# Returns: 1 on success, 0 on failure
# Purpose: Wrapper around save_state() to handle multiple LIFs at once
###############################################################################

sub save_all_state {

    $Trace->entry();
    my $self = shift(@_);

    # Extract any parameters. LIF list can either be a list of LIF names
    # or a list of LIF IP addresses. "key" will denote what it is.

    my %params = validate (
        @_,
        {
            key => {
                type     => SCALAR,
                optional => 1,
                default  => "lif"
            },
            value_list => {
                type     => ARRAYREF,
                optional => 0,
            },
        }
    );

    # Loop through the list and save

    foreach my $lif (@{$params{value_list}}) {

        if (!save_state($self,
                    key => $params{key}, 
                    value => $lif,
                )) {
            $Trace->fatal(msg => "Unable to save state for LIF $lif");
            return 0;
        }
    }

    $Trace->exit(msg => "LIF state successfully saved for ".
                 Dumper($params{value_list}));
    return 1;

} ## end sub save_all_state

###############################################################################
# Subroutine: get_state
# Parameters: LIF name
# Returns: Hash reference containing stored information on said LIF
# Purpose: To take in a LIF name and returned the stored information on it
###############################################################################

sub get_state {

    $Trace->entry();
    my $self =  shift(@_);

    my %params = validate (
        @_,
        {
            lif => {
                type     => SCALAR,
                optional => 0,
            }
        }
    );

    my %lif_info = %{$self->{_lif_state}->{$params{lif}}};

    $Trace->exit();
    return \%lif_info;

} ## end sub get_state

###############################################################################
# Subroutine: check_state
# Parameters: LIF parameters, action (moved, stayed, etc.)
# Returns: 1 if the LIF remains in the expected state and 0 otherwise
# Purpose: A utility that will check if a particular LIF has changed state 
#          (i.e. moved) or not.
################################################################################

sub check_state {

    $Trace->entry();
    my $self =  shift(@_);

    my %params = validate (
        @_,
        {
            key => {
                type     => SCALAR,
                optional => 1,
                default  => "lif", 
            },
            value => {
                type     => SCALAR,
                optional => 0,
            },
            move => {
                type     => SCALAR,
                optional => 1,
                default  => 1,
            },
        }
    );    

    # Get the current state

    my @curr_state = $Server->network_interface_show(
        $params{key} => $params{value},                               
    );
    if (scalar(@curr_state) > 1) {
        $Trace->warn(msg => "Query for $params{value} returned more than one LIF. ".
                    "Will only compare the first one.");
    }

    # Get the saved state

    my %saved_state = %{get_state($self, lif => $curr_state[0]->{lif})};

    # Perform the comparison. Any mismatches could indiciate a node, port, status,
    # etc. have change. Any change in those indicates a lif state change.
    
    my $mismatch = 0;
    foreach my $key (keys(%saved_state)) {

        if ($saved_state{$key} ne $curr_state[0]->{$key}) {
            $mismatch++;
        } 
    }

    # Check to see if what we expected (movement or no movement) coincides with
    # what we have found (movement or no movement). $mismatch > 0 means that 
    # the LIF has moved since its previous state was saved.

    if (($params{move}) && ($mismatch > 0)) {

        # Case where we expected the LIF to move, and it did (mismatches found)

        $Trace->exit(msg => "Expected LIF to move; $mismatch mismatches found".
                     ". Returning success.");
        return 1;

    } elsif (($params{move}) && ($mismatch == 0)) {

        # Case where we expected the LIF to move, and it did not (no mismatches
        # found)

        $Trace->exit(msg => "Expected LIF to move; $mismatch mismatches found");
        return 0;

    } elsif ((!$params{move}) && ($mismatch > 0)) {
        
        # Case where we expected the LIF NOT to move, and it did (mismatches found)

        $Trace->exit(msg => "Expected LIF not to move, $mismatch mismatches ".
                    "found.");
        return 0;

    } elsif ((!$params{move}) && ($mismatch == 0)) {

        # Case where we expected the LIF NOT to move, and it did not (no mismatches
        # found). 

        $Trace->exit(msg => "Expected LIF not to move, $mismatch mismatches ".
                    "found.");
        return 1; 

    } else {

        # Catch all

        $Trace->exit(msg => "Move: $params{move}, Mismatches: $mismatch. ".
                    "Unexpected state.");
        return 0;
    }

} ## end sub check_state

###############################################################################
# Subroutine: did_lifs_move
# Parameters: List of LIFs to check, threshold of failures to tolerate
# Returns: 1 on success, 0 on failure
# Purpose: A routine that when given a list of LIFs to check, will be able to 
#          tell if they have recently changed state (meaning if they have
#          recently moved).
#          Note: If you do no plan on passing in a LIF list, you can call 
#                 check_state directly.
# 
# Note: There is a "threshold", because with DNS Load Balancing, it is very 
#       hard to give an exact number of how many LIFs we believe should remain 
#       stationary and/or how many LIFs we believe should have moved. It is all 
#       up to the calculations done on the node at a particular time.
###############################################################################

sub did_lifs_move {

    $Trace->entry();
    my $self =  shift(@_);

    # Extract any parameters

    my %params = validate (
        @_,
        {
            key => {
                type     => SCALAR,
                optional => 1,
                default  => "lif", 
            },
            value_list => {
                type     => ARRAYREF,
                optional => 0,
            },
            threshold => {
                type     => SCALAR,
                optional => 1,
                default  => 1, 
            },
            move => {
                type     => SCALAR,
                optional => 1,
                default  => 1,
            },
        }
    );

    my $failures = 0;

    # Iterate and check each LIF 
    
    foreach my $lif (@{$params{value_list}}) {

        # Check the current state

        my $result = check_state($self,
            key   => $params{key},
            value => $lif,
            move  => $params{move},
        );
        
        # If we had a failure (i.e. expected a LIF to move and it did not),
        # we can check to see if its within our threshold.

        if (!$result) {
            
            $failures++;

            if ($failures > $params{threshold}) {
                $Trace->warn(msg => "Exceeded threshold $params{threshold}. ".
                             "Exiting.");
                return 0;
            } else {
                $Trace->comment(msg => "LIF $lif was not in expected state. ".
                                "Current failures at $failures, threshold is ".
                               "at $params{threshold}. Continuing.");
            }
        }
    }

    $Trace->exit(msg => "LIF checks for LIF(s): ".Dumper($params{value_list}).
                " successful");
    return 1;

} ## end sub did_lifs_move

1;

