# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/VolCmds.pm#1 $

## @summary Library of volume commands
##
## @description
##    This package defines routines which use the filer's "volume" command.
##    They gather and return information about volumes, or attempt to
##    change the state of volumes, and return success or failure.
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @cf     false
## @status public

package VolCmds;

our @ISA = qw(Exporter);
our @EXPORT = qw(volOnline volOffline volRestrict volRename
	         volStatus volOptions volExists volSize volDisks
	         getRootVol maxFiles diskFree);

use SayWrap;
use FSR::FSRutils;
use Volume qw(vol_match_plex_raid_info);
use Tharn;
use strict;

## @name ::volOnline
## @usage volOnline($filer, $vol)
## @synopsis
##    Change the state of the volume to "online".
##    Returns success (if the volume was changed or is already in the desired state).
##    or failure otherwise (if the volume couldn't change or didn't exist).
##    On failure, error string is in $@.
sub volOnline($$) {
  my ($filer, $vol) = @_;
  my $volexp = regx_escape($vol);

  return eval {
    my $text = sayEOF($filer, "vol online $vol");
    die "$text\n" unless ($text =~ /Volume ('$volexp'|$volexp) is (now|already) online./);
    return 1;
  };
}

## @name ::volOffline
## @usage volOffline($filer, $vol)
## @synopsis
##    Change the state of the volume to "offline".
##    Returns success (if the volume was changed or is already in the desired state).
##    or failure otherwise (if the volume couldn't change or didn't exist).
##    On failure, error string is in $@.
sub volOffline($$) {
  my ($filer, $vol) = @_;
  my $volexp = regx_escape($vol);

  return eval {
    my $text = sayEOF($filer, "vol offline $vol");
    die "$text\n" unless ($text =~ /Volume ('$volexp'|$volexp) is (now|already) offline./);
    return 1;
  };
}

## @name ::volRestrict
## @usage volRestrict($filer, $vol)
## @synopsis
##    Change the state of the volume to "restricted".
##    Returns success (if the volume was changed or is already in the desired state).
##    or failure otherwise (if the volume couldn't change or didn't exist).
##    On failure, error string is in $@.
sub volRestrict($$) {
  my ($filer, $vol) = @_;
  my $volexp = regx_escape($vol);

  return eval {
    my $text = sayEOF($filer, "vol restrict $vol");
    die "$text\n" unless ($text =~ /Volume ('$volexp'|$volexp) is (now|already) restricted./);
    return 1;
  };
}

## @name ::volRename
## @usage volRename($filer, $oldvol, $newvol)
## @synopsis
##    Renames the name of volume $oldvol to $newvol. Returns success or
##    failure. On failure, error string is in $@.
sub volRename($$$) {
  my ($filer, $oldvol, $newvol) = @_;

  return eval {
    my $text = sayEOF($filer, "vol rename $oldvol $newvol");
    die "$text\n" unless ($text =~ /('$oldvol'|$oldvol) renamed to ('$newvol'|$newvol)/);
    return 1;
  };
}

## @name ::volStatus
## @usage volStatus($filer, $vol)
## @synopsis
##    Returns the state, status, and raid type of the given volume.
##    When called in a scalar context, returns just the state.
##    On error, returns undef with an error message in $@.
## @example
##    my ($state, $status, $type) = volStatus($filer, $vol);
sub volStatus($$) {
  my ($filer, $vol) = @_;
  my $volexp = regx_escape($vol);

  my $text = sayEOF($filer, "vol status $vol", TIMEOUT => 300);

  return eval {
    my ($state, $type, $options);
    die "$text\n" unless ($text =~ s/^Volume\s+(Type\s+)?State\s+Status\s+Options\s*?\n//);
    if ($1) {
      die "$text\n" unless ($text =~ s/^\s*$volexp\s+(?:\w+\s+)(\w+)\s+(\w+)\s*([\s\w,=\(\)]+)?(?:\n|\z)//);
      ($state, $type, $options) = ($1, $2, (defined $3 ? $3 : ""));
    } else {
      die "$text\n" unless ($text =~ s/^\s*$volexp\s+(\w+)\s+(\w+)\s*([\s\w,=\(\)]+)?(?:\n|\z)//);
      ($state, $type, $options) = ($1, $2, (defined $3 ? $3 : ""));
    }

    my @status_values = ();
    my $got_last_status_value = 0;
    while ($text =~ s/\A(?: {27}([\w-]+|parity uninit'd!)\s*| {45})(\w[\s\w,=\(\)]+)?(?:\n|\z)//) {
      # Successive status values always start on column 27,
      # and successive options start on column 45.
      if ($1) {
	if ($got_last_status_value) {
	  # Saw a gap in the status values
	  die "unrecognized status value \"$1\"\n";
	}
	push @status_values, $1;
      } else {
	$got_last_status_value = 1;
      }

      # Stuff all the option strings together.
      $options .= $2 if $2;
    }

    my $status;
    my $status2;
    if (scalar(@status_values) == 0) {
      # At the moment, this is just to provide backward-compatibility for
      # scripts written before 'vol status' started printing the raid type.
      $status = "normal";
    } elsif (contains("snapmirrored", @status_values)) {
      # Again, for backwards-compatibility.
      $status = "snapmirrored";
    } else {
      # Just return the first one we saw.
      $status = $status_values[0];
    }

    # Additional status
    if (contains("degraded", @status_values)) {
      $status2 = "degraded";
    } elsif (contains("reconstruct", @status_values)) {
      $status2 = "reconstruct";
    }
    
    my $match_flex_vol = "'.*'";
    my $match_trad_vol  = "<N/A>"; #containing aggr is not available
    if ($text =~ s/^\s+Containing aggregate: $match_trad_vol\s*//) {
	    # its a traditional volume, so expect RADI/plex info.
	    if ($text =~ /\A\s*\Z/) {
		    die "did not receive RAID/plex information\n"
	    }
	    $text = vol_match_plex_raid_info($volexp, $text);
    } else {
	    # its a flexible volume
	    $text =~ s/^\s+Containing aggregate: $match_flex_vol\s*//;
    } 		
    # 'vol status' changed?
    if ($text ne "") {
	    die "VolCmds::volStatus: unexpected output: $text";
    }
    return (wantarray ? ($state, $status, $type, $status2) : $state);
  };
}

## @name ::volOptions
## @usage volOptions($filer, $vol)
## @synopsis
##    Returns a reference to a hash of option keys to their values
##    for the specified volume on the specified filer.
##    On error, returns undef with an error message in $@.
sub volOptions($$) {
  my ($filer, $vol) = @_;
  my $volexp = regx_escape($vol);

  my $text = sayEOF($filer, "vol status -v $vol");

  return eval {
    die "$text\n" unless ($text =~ s/^Volume\s+State\s+Status\s+Options\s*?\n//);
    die "$text\n" unless ($text =~ s/^\s*$volexp\s+\S+\s+\S+\s+(.*?)\s*Plex.*?RAID group.*$/$1/s);
    $text =~ s/\s//g;
    my @opts = split /,/, $text;
    my %opthash = map { ($_ =~ /^(\S+)\=(\S*)$/) ? ($1 => $2) : ($_ => undef) } @opts;
    return \%opthash;
  };
}

## @name ::volExists
## @usage volExists($filer, $vol)
## @synopsis
##    Returns whether a given volume exists.
##    On error, an error message is placed in $@.
sub volExists($$) {
  my ($filer, $vol) = @_;

  my $opts = volOptions($filer, $vol);
  return eval {
    return 1 if (defined $opts); # volume exists if we got its options

    my $volexp = regx_escape($vol);
    if ($@ =~ /No volume named '$volexp' exists\./) {
      return 0;
    } else {
      die $@;
    }
  };
}

## @name ::volSize
## @usage volSize($filer, $vol)
## @synopsis
##    Returns the blocksize, volume size, and filesystem size
##    of the given volume as returned by "vol status -b".
##    e.g. my ($bs, $volsize, $fssize) = volSize($filer, $vol);
##    In a scalar context, returns just the filesystem size (in blocks).
##    On failure, returns undef with an error message in $@.
sub volSize($$) {
  my ($filer, $vol) = @_;
  my $volexp = regx_escape($vol);

  my $text = sayEOF($filer, "vol status -b $vol");

  return eval {
    die "$text\n"
      unless ($text =~ /^$volexp\s+(\d+)\s+(\d+)\s+(\d+)\s*$/m);
    return (wantarray ? ($1, $2, $3) : $3);
  };
}

## @name ::volDisks
## @usage volDisks($filer, $vol)
## @synopsis
##    Returns an array of disk descriptor for all disks on the volume.  Each
##    disk descriptor is a hash of the form (TYPE NAME SIZE). On failure,
##    return undef with error message in $@.
sub volDisks($$) {
  my ($filer, $vol) = @_;

  my $text = sayEOF($filer, "vol status $vol -r");

  return eval {
    my @disk_list;

    foreach my $line (split(/\n/, $text)) {
      if ($line =~ /\s+parity|\s+data/) {
	 my @da = split(/\s+/, $line);
	 my %disk;

	 $disk{'TYPE'} = $da[1];
	 $disk{'NAME'} = $da[2];
	 $disk{'SIZE'} = $da[7];

         push (@disk_list, {%disk});
      }
    }
    return @disk_list;
  }
}

## @name ::getRootVol
## @usage getRootVol($filer)
## @synopsis
##    Returns the name of the root volume on the given filer.
##    On failure, returns undef with an error message in $@.
sub getRootVol($) {
  my ($filer) = @_;

  my $text = sayEOF($filer, "vol status");

  return eval {
    die "$text\n"
      unless ($text =~ s/^\s*Volume\s+State\s+Status\s+Options\s*?(\n|$)//m);

    # If we can find it quickly, do it
    if ($text =~ /^\s*?(\w+)\s+\w+\s+\w+\s+root/m) {
      my $volume = $1;
      my $opts = volOptions($filer, $volume);
      return $volume
	if ((defined $opts) && (exists $opts->{'root'}));
    }

    # Otherwise, search for it
    my @lines = split /\n/, $text;
    my @volumes = map { die "Error parsing \"$_\"\n"
			  unless ($_ =~ /^\s*?(\w+)\s+\w+\s+\w+.*$/);
			return $1;
		      } @lines;

    foreach my $vol (@volumes) {
      my $opthash = volOptions($filer, $vol);
      return $vol
	if ((defined $opthash) && (exists $opthash->{'root'}));
    }

    die "No root volume found\n";
  };
}

## @name ::maxFiles
## @usage ::maxFiles($filer, $vol, $count)
## @synopsis
##    Gets or sets the maximum file count for the specified volume.
##    The third argument is optional.  If given, it specifies the
##    maximum file count to set on the volume, and the function returns
##    success or failure with an error message in $@ on failure. If the
##    third argument is omitted, and the function returns a tuple of
##    ($max, $used), representing the maximum number and currently used
##    number of files in the given volume, with an error message in $@
##    on failure.
sub maxFiles($$;$) {
  my ($filer, $vol, $count) = @_;

  return eval {

    if (defined $count) {
      # Set maxfiles
      my $text = sayEOF($filer, "maxfiles -f $vol $count");
      die "$text\n" if ($text);
      return 1;

    } else {
      # Get maxfiles
      my $volexp = regx_escape($vol);
      my $text = sayEOF($filer, "maxfiles $vol");
      die "$text\n"
	unless ($text =~ /Volume $volexp\: maximum number of files is currently (\d+) \((\d+) used\)\./);
      return (wantarray ? ($1, $2) : $1);
    }

  }
}

## @name ::diskFree
## @usage ::diskFree($filer, $path, $space_type, $report_type)
## @synopsis
##    Uses the filer's 'df' command to report about diskspace usage
##    for the path on the filer.
##    Returns a tuple of ($used, $free) on success, or failure and
##    an error message in $@.
##    The last two arguments are optional.
##    The $space_type argument is one of 'space', 'inodes', or 'reserve'
##    (default is 'space'), specifying whether to report about free space,
##    free inodes, or free reserved space.
##    The $report_type specifies the type of units to use in the output,
##    and specified as 'KB', 'MB', 'GB', or 'TB'.  (Default is 'KB'.)
sub diskFree($$;$$) {
  my ($filer, $path, $space_type, $report_type) = @_;
  $space_type = 'space' unless (defined $space_type);
  $report_type = 'KB' unless (defined $report_type);

  return eval {
    if ($path =~ m@^/vol/(\w+)$@) {
      # the df output always has a trailing slash for volumes
      $path .= '/';
    }

    my $space;
    if ($space_type eq 'space') {
      $space = "";
    } elsif ($space_type eq 'inodes') {
      $space = " -i";
    } elsif ($space_type eq 'reserve') {
      $space = " -r";
    } else {
      die "Unrecognized space type: $space_type\n";
    }

    my $report = $report_type;
    die "Unrecognized report type: $report_type\n"
      unless ($report =~ s/^([Kk]|[Mm]|[Gg]|[Tt])[Bb]$/$1/);
    $report =~ tr/A-Z/a-z/;
    $report = '-' . $report;

    # Run the df command and snag the output
    my $text = sayEOF($filer, "df $space $report $path");

    my $pathexp = regx_escape($path);
    if ($text =~ /df:\s+$pathexp:\s+(No such file or directory)/) {
      die "$1\n";
    }

    if ($space_type eq 'space') {
      die "Error parsing df output: $text\n"
	unless ($text =~ s/^Filesystem\s+total\s+used\s+avail\s+capacity\s+Mounted on\s*//);
      die "Could not find $path in df output\n"
	unless ($text =~ /^$pathexp\s+\d+$report_type\s+(\d+)$report_type\s+(\d+)$report_type\s+(\d+)\%\s+$pathexp\s*$/m);
      return (wantarray ? ($1, $2, $3) : $1);

    } elsif ($space_type eq 'inodes') {
      die "Error parsing df output: $text\n"
	unless ($text =~ s/^Filesystem\s+iused\s+ifree\s+\%iused\s+Mounted on\s*//);
      die "Could not find $path in df output\n"
	unless ($text =~ /^$pathexp\s+(\d+)\s+(\d+)\s+(\d+)\%\s+$pathexp\s*$/m);
      return (wantarray ? ($1, $2, $3) : $1);

    } elsif ($space_type eq 'reserve') {
      die "Error parsing df output: $text\n"
	unless ($text =~ s/^Filesystem\s+total\s+used\s+avail\s+reserved\s+Mounted on\s*//);
      die "Could not find $path in df output\n"
	unless ($text =~ /^$pathexp\s+\d+$report_type\s+(\d+)$report_type\s+(\d+)$report_type\s+(\d+)$report_type\s+$pathexp\s*$/m);
      return (wantarray ? ($1, $2, $3) : $1);

    } else {
      die "Internal error.\n";
    }
  }
}
