# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/VolCopy.pm#1 $

## @summary vol copy library -- action subroutines
##
## @description
##    This package defines routines which perform, start, abort,
##    complete, and get status on volcopy transfers.
##
## @dependencies
##    lib/VolCmds.pm
##    lib/SayWrap.pm
##    FSR/FSRutils.pm
##
## @bugs
##
## @author dl-dpgqa-automation@netapp.com
## @status public

package VolCopy;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(volcopy_status volcopy_throttle volcopy_abort
		 volcopy_start volcopy_from_source);

use VolCmds;
use FSR::FSRutils;
use SayWrap;
use Tharn;
use strict;

## @name ::volcopy_status
## @usage volcopy_status($filerConn, $operation)
## @synopsis
##    Returns the progress of the volcopy transfer with the given operation
##    number as a percentage.  Returns undef if no volcopy operation with
##    the given operation number is in progress.  On failure, returns an error
##    message in $@.
sub volcopy_status($$)  {
  my ($conn, $operation) = @_;
  return eval {
    my $text = sayEOF($conn, "vol copy status $operation");
    if ($text eq "No operations in progress.") {
      return undef;
    } elsif ($text =~ /^\d{2}:\d{2}:\d{2} \w{3} : vol copy (?:dump|restore) $operation : (\d+) % done. Estimate \d+ minutes remaining\.$/) {
      return $1;
    } elsif ($text =~ /^\d{2}:\d{2}:\d{2} \w{3} : vol copy (?:dump|restore) $operation : begun, \d+ \w+ to be copied\.$/) {
      return 0;
    } else {
      die "$text\n";
    }
  }
}

## @name :: volcopy_throttle
## @usage volcopy_throttle($filerConn, $operation, $throttle)
## @synopsis
##    Modifies the speed at which volcopy transfer is being done.
##    Returns success or failure, with an error mesage in $@ on failure.
sub volcopy_throttle($$$) {
  my ($conn, $operation, $throttle) = @_;

  return eval {
    die "throttle value must be an integer between 1 and 10\n"
      unless (($throttle =~ /^\d+$/) && ($throttle > 0 && $throttle < 11));

    my $text = sayEOF($conn, "vol copy throttle $operation $throttle");
    die "$@" unless ($text =~ /^vol copy operation \d+: Throttle adjusted from (10|[1-9])0% to ${throttle}0% of full speed\.$/);
    return 1;
  }
}

## @name ::volcopy_abort
## @usage volcopy_abort($filerConn, $operation)
## @synopsis
##    Aborts the volcopy transfer with the given operation number.
##    Returns success or failure, with an error message in $@ on failure.
sub volcopy_abort($$)  {
  my ($conn, $operation) = @_;
  return eval {
    my $text = sayEOF($conn, "vol copy abort $operation");
    return 1 if ($text =~ /^image operation $operation is now being aborted$/);
    die "unexpected end of output\n" if ($text eq "");
    die "$text\n";
  }
}

# Internal function which constructs and returns the vol copy command,
# given the source, destination, and an options hash.
sub volcopy_cmd_from_opts($$;%) {
  my ($src, $dst, %param) = @_;

  # Setup volcopy arguments
  my $cmd = "vol copy start" .
    ($param{'COPY_ALL'} ? " -S" : "") .
    ($param{'SNAPSHOT'} ? " -s " . $param{'SNAPSHOT'} : "") .
    ($param{'METADATA'} ? " -M" : "") .
    " $src $dst";

  # Some options only work in diag mode
  if ($param{'METADATA'}) {
    $cmd = "priv set diag ; " . $cmd;
  }

  return $cmd;
}

# The waitstrings for vol copy depend upon the operation number,
# but it's better to only define the strings in one place.
sub volcopy_regx_from_op(;$) {
  my ($op_num) = @_;
  $op_num = "\\d+" unless (defined $op_num);

  # These are a hardcoded regexps which correspond to the ONTAP code.
  # Don't change these unless you know exactly what you're doing.
  my $finish_prog = "^\\s*\\d\\d:\\d\\d:\\d\\d (?:\\w{3}) : " .
    "vol copy restore $op_num : \\d+ % done\\. " .
    "Estimate \\d+ minutes remaining\\.\\s*(?:\\n|\$)";
  my $finish_wait = "(?m)" .
    "(?:$finish_prog)*" .
    "\\s*\\d\\d:\\d\\d:\\d\\d (?:\\w{3}) : vol copy restore (\\d+) : 100% done, \\d+ [KMGT]B copied\\.\\s*(?:\\n|\$)";

  return ($finish_prog, $finish_wait);
}

## @name ::volcopy_start
## @usage volcopy_start($filerConn, $volname, %options)
## @synopsis
##    Starts a vol copy transfer to the specified volume on the given filer.
##    Note, this must be the destination filer.  If the transfer successfully
##    starts, returns a tuple of a subtest object and the operation number.
##    The subtest object is for an active subtest which is waiting for the
##    transfer to complete. On failure, returns undef with error message in $@.
##    If the COMPLETE option is set, then this function returns whether
##    the transfer both started and completed successfully.
##    There is no need to extract results from the returned subtest object,
##    since $@ will contain the error if the transfer failed.
## @param SOURCE Required: the source <volume> or <filer>:<volume> to copy from
## @param RESTRICT Optional: whether to make sure the destination volume is
##                           restricted before starting the transfer
## @param COPY_ALL Optional: whether or not to copy all snapshots
## @param SNAPSHOT Optional: the name of a specific snapshot on the source
##                           volume to copy
## @param METADATA Optional: only copy metadata and indirect blocks,
##                           not level-0 blocks
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub volcopy_start($$;%) {
  my ($conn, $dstvol, %param) = @_;
  return eval {
    die "SOURCE parameter is required\n" unless ($param{'SOURCE'});

    # Setup some convenience variables and escaped strings.
    my $src = $param{'SOURCE'};
    my ($srcvol, $srcfiler) = split_filer_path($param{'SOURCE'});
    if (!$srcfiler) {
	$srcfiler="127.0.0.1";
    }
    my $dstvol_exp = regx_escape($dstvol);
    my $srcvol_exp = regx_escape($srcvol);
    my $srcfiler_exp = regx_escape($srcfiler);

    # Make sure the destination is restricted (if necessary).
    if ($param{'RESTRICT'}) {
      die $@ unless volRestrict($conn, $dstvol);
    }

    my %opts = ();
    $opts{'TIMEOUT'} = $param{'TIMEOUT'}
      if (defined $param{'TIMEOUT'});

    # Construct the vol copy command.
    my $cmd = volcopy_cmd_from_opts($src, $dstvol, %param);

    # This is a hardcoded regexp which corresponds to the ONTAP code.  The
    # parentheses also determine how we determine the volcopy operation number
    # below. Don't change these unless you know exactly what you're doing.
    my $start_wait = "(?m)" .
      "^This can be a long-running operation. Use Control - C \\(\\^C\\) to interrupt\\.\\s+" .
      "Copy Volume: $srcvol_exp on machine $srcfiler_exp to Volume: $dstvol_exp\\s+" .
      "(?:VOLCOPY: Starting on volume \\d+\\.)?" .
      "\\s*\\d\\d:\\d\\d:\\d\\d (?:\\w{3}) : vol copy restore (\\d+) : begun, \\d+ [KMGT]B to be copied\\.\\s*(?:\\n|\$)";

    my ($matched, $text) = sayWAIT($conn, $cmd, $start_wait, %opts);
    die $@ if ($@); # sayWAIT failed
    unless ($matched) {
      # waitstring never matched
      die "$text\n" if ($text);
      die "unexpected end of output\n";
    }

    # Re-match the matched string to determine the volcopy operation number.
    # Note, this should NEVER fail -- else we would have returned already.
    die "waitstring did not match returned string\n"
      unless ($matched =~ /$start_wait/);
    die "error determining volcopy operation number\n" unless (defined $1);
    my $op_num = $1;
    die "invalid operation number\n" unless ($op_num =~ /^\d+$/);

    # Any other text that we got must be an error message.
    die "$text\n" if ($text);

    my ($finish_prog, $finish_wait) = volcopy_regx_from_op($op_num);

    # The transfer successfully started,
    # so now we just need to wait for the completion string.
    my $obj = subtest( sub {
			 my ($matched, $text) = sayWAIT($conn, "",
							$finish_wait, %opts);
			 die $@ if ($@); # sayWAIT failed
			 (undef, undef) = sayWAIT($conn, "", ""); # wait for prompt
			 die $@ if ($@); # sayWAIT failed
			 logcomment("vol copy is now done.");
						
			 if ($text) {
			   # Strip out any progress lines which may have been
			   # printed between syslogs or other stripped text.
			   $text =~ s/$finish_prog//mg;
			 }
			 unless ($matched) {
			   # waitstring never matched
			   die "$text\n" if ($text);
			   die "unexpected end of output\n";
			 }
			 logresult('FATAL', "EVAL_ERROR=$@") if ($@);
			 logresult('FATAL', "EVAL_ERROR=$text") if ($text);
			 logresult('PASS');
		       },
		       '-bg', '-onexit', sub{}, # don't pass results to parent
		     );

    # If specified, wait for the transfer to complete
    # and check the results once it does.
    if ($param{'COMPLETE'}) {
      Subtest::wait_finish(subtests=>[$obj]);
      die $@ unless subtestResults($obj);
    }

    if (wantarray) {
      my @ret = ($obj, $op_num);
      return @ret;
    } else {
      return $obj;
    }
  }
}

## @name ::volcopy_from_source
## @usage volcopy_from_source($filerConn, $volname, %options)
## @synopsis
##    Starts a vol copy transfer from the specified volume on the given filer.
##    Note, this must be the source filer -- most users should call
##    volcopy_start() instead.  This function only exists for scripts that
##    must invoke the vol copy operation on the source filer for some reason.
##    This function waits for the transfer to complete, and returns whether
##    the transfer succeeded or failed, with an error message in $@ if the
##    transfer failed.
## @param DESTINATION Required: the destination <filer>:<volume> to copy to
## @param COPY_ALL Optional: whether or not to copy all snapshots
## @param SNAPSHOT Optional: the name of a specific snapshot on the source
##                           volume to copy
## @param METADATA Optional: only copy metadata and indirect blocks,
##                           not level-0 blocks
## @param TIMEOUT Optional: timeout period waiting for the command to complete
sub volcopy_from_source($$;%) {
  my ($conn, $srcvol, %param) = @_;

  return eval {
    die "DESTINATION parameter is required\n" unless ($param{'DESTINATION'});

    my ($dstvol, $dstfiler) = split_filer_path($param{'DESTINATION'});
    die "invalid destination: " . $param{'DESTINATION'} . "\n"
      unless ($dstfiler);
    my $srcvol_exp = regx_escape($srcvol);
    my $dstvol_exp = regx_escape($dstvol);
    my $dstfiler_exp = regx_escape($dstfiler);

    # Construct the vol copy command.
    my $cmd = volcopy_cmd_from_opts($srcvol, $param{'DESTINATION'}, %param);

    my $text = sayEOF($conn, $cmd, TIMEOUT => $param{'TIMEOUT'});

    # This is a hardcoded regexp which corresponds to the ONTAP code.  The
    # parentheses also determine how we determine the volcopy operation number
    # below. Don't change these unless you know exactly what you're doing.
    my $start_wait = "(?m)" .
      "^This can be a long-running operation. Use Control - C \\(\\^C\\) to interrupt\\.\\s+" .
      "Copy Volume: $srcvol_exp on machine (?:\\S+) to Volume: $dstvol_exp\\s+" .
      "(?:VOLCOPY: Starting on volume \\d+\\.)?" .
      "\\s*\\d\\d:\\d\\d:\\d\\d (?:\\w{3}) : vol copy restore (\\d+) : begun, \\d+ [KMGT]B to be copied\\.\\s*(?:\\n|\$)";

    my ($finish_prog, $finish_wait) = volcopy_regx_from_op();

    # Make sure that the transfer started successfully.
    die "$text\n" unless ($text =~ s/$start_wait//);

    # Make sure that the transfer completed successfully.
    die "$text\n" unless ($text =~ s/$finish_wait//);

    # Strip out all the progress strings.
    $text =~ s/$finish_prog//mg;

    # All other strings are error messages.
    die "$text\n" if ($text);

    return 1;
  }
}


1;
