##  $Id: //depot/prod/test/Rsierranevada/lib/VolOps.pm#2 $

## @summary     Base class for Move.pm and Copy.pm
## @author      dl-volops-qa@netapp.com
## @status      private
## @pod         here

=head1 NAME

VolOps.pm

=head1 SYNOPSIS

    use base;
    use VolOps::BaseException;
    use strict;

=head1 DESCRIPTION

C<VolOps> is a base class for performing volume move and copy testing.

The class throws exceptions on failures.

This class makes use of its own exception class
VolOps::BaseException. While the exception class will be
inherited when using this library, it is recommended to explicitly import
the exception class for clarity.

The current list of possible exceptions are:
    NATE::Exceptions::Argument
    VolOps::BaseException
    VolOps::ConfigError
    VolOps::JobFailed
    VolOps::JobStatusFailure
    VolOps::AttributeComparison
    VolOps::MissingValidationRule
    VolOps::RetriesExhausted
    VolOps::SrcVolumeFailure
    VolOps::SrcAggregateFailure
    VolOps::AggregateSelectionFailure
    VolOps::DestAggregateFailure
    VolOps::PropsNotAvailable
    VolOps::StripingNotSupported
    VolOps::ConnectionFailure
    VolOps::DataStructureStoreFailure
    VolOps::InternalError
    VolOps::Move::MoveDeferred
    VolOps::Move::JobNotDeferred

This class also uses some of the NATE exceptions.  These exceptions include
the following:
 NATE::Exceptions::Argument;

=head1 METHODS

=cut

package VolOps;
use strict;
#use warnings;
use Params::Validate qw(validate validate_with BOOLEAN SCALAR HASHREF ARRAYREF UNDEF OBJECT);
use Tharn qw(fatal_result config_result fail_result pass_result warn_result $Staf $Log $Parent snooze subtest);
use Data::Dumper;
use NATE::BaseException qw(:try);
use VolOps::BaseException;
use DataStructureStore;
use GNOTE;
use Node;
use Date::Parse;
use NATE::STAF qw(unmarshall);
use NATE::Events qw(listener_add listener_remove);
use NATE::Exceptions::Argument;

my $Options = new Options();

=head2 new

C<< VolOps->new(%params) >>

Invocant for the VolOps class.  It is called from the VolOps::Move.pm
and VolOps::Copy.pm B<new> methods.

=over 4

=item Input

=over 4

=item * node => <scalar>

(Required)
Default = undef

Node in the cluster

=item * engine => <scalar>

(Optional)
Default = "BRE"

Indicates what type of replication to use when moving the volume(s).

Valid values:
   BRE - Use physical replication
   LRE - Use logical replication

=item * save_status => <boolean>

(Optional)
Default = false

Indicates whether or not to save the status hash for each iteration.
If this flag is set, the status hash will be saved, by default, as a
text file to the LOGDIR using file name "Vol[Move|Copy]_status_<timestamp>.dd".
The save mode, filename, and/or directory can be modified on the command line
using the DSS_MODE, DSS_DIR, DSS_FILE, and DSS_FILENAME parameters.  See the POD
for lib/DataStructureStore.pm for more information.

=back

=item Output

Class Reference

=item Example

C<< my $vmv = VolOps::Move->new("node"        => $node,
                                "save_status" => 1); >>

=back

=cut

################################################################################
#
# Subroutine:   new()
#
# Objective:    Invocant for the class
#               Sets the paloma.physical flag based on the type of replication
#               selected.
#
# Details:
#
#   Input:
#       node        - Name of a node in the cluster
#                     Required
#       engine      - Paloma engine to use for replication.  Possible values
#                     are 'BRE' or 'LRE'.  Optional; default is 'undef'.
#       save_status - Indicates whether or not to save the status hash for
#                     each iteration.  If this flag is set, the status hash
#                     will be Saved, by default, as a text file to the LOGDIR
#                     using file name "Vol[Move|Copy]_status_<timestamp>.dd".
#                     The save mode, filename, and/or directory can be modified
#                     on the command line using the DSS_MODE, DSS_DIR,
#                     DSS_FILE, and DSS_FILENAME parameters.  See the POD for
#                     lib/DataStructureStore.pm for more information.
#                     Optional; default = 0
#
#
#   Return value:
#       object
#
################################################################################

sub new {
    $Log->enter();
    my $class = shift;
    my $self;

    eval {
        $self = {
                    validate(
                             @_,
                             {
                              node                => {
                                                      type      => SCALAR,
                                                     },
                              engine              => {
                                                      type      => SCALAR|UNDEF,
                                                      optional  => 1,
                                                      default   => "BRE",
                                                      callbacks => {'bre_or_lre' =>
                                                                    sub { shift() =~ /^(BRE|LRE)$/ }
                                                                   },
                                                     },
                              save_status         => {
                                                      type       => BOOLEAN,
                                                      optional   => 1,
                                                      default    => 0,
                                                     },
                             }
                            )
                };

    };

    if ($@) {
        my $ex_text = "Parameter validation failed: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    $Log->debug("Input Params: " . Dumper($self));
    $self->{'src_vols'}              = ();
    $self->{'job_status'}            = ();
    $self->{'total_errors'}          = 0;
    $self->{'options'}               = $Options->get_hash();
    $self->{'nodes'}                 = $self->{'options'}->{'NODES'};
    $self->{'server'}                = new Server();
    #-------------------------------
    # Determine the calling library
    #-------------------------------
    $self->{'caller'}                = caller;
    $self->{'caller'}                =~ s/^.*::(.*)$/$1/;

    #-------------------------------------------------
    # If user requested to save status data to a file
    #-------------------------------------------------
    if ($self->{'save_status'}) {
        $self->{'dss_obj'} = DataStructureStore->new();
    }

    #--------------------------------
    # Establish a cluster connection
    #--------------------------------
    eval {
        $self->{'cluster'} = NOTE::GX::Cluster->connect(Node =>
                                                        $self->{'node'},
                                                       );
    };

    if ($@) {
        my $ex_text = "Unable to connect to cluster: $@";
        $Log->error(fail_result($ex_text));
        VolOps::ConfigError->throw($ex_text,
                                  );
    }

    #---------------------
    # Create a job object
    #---------------------
    eval {
        $self->{'job_obj'} = NOTE::GX::Job->new(Cluster => $self->{'cluster'});
    };

    if ($@) {
        my $ex_text = "Unable to create a job object: $@";

        $Log->error(fail_result($ex_text));
        VolOps::ConfigError->throw($ex_text,
                                  );
    }

    bless($self, $class);

    my @node_list = ();

    my ($maj_rel, $min_rel);

    #------------------------------------------------
    # Check/set the Paloma engine type for each node
    #------------------------------------------------
    foreach my $node (@{$self->{'nodes'}}) {

        ($maj_rel, $min_rel) = $self->_get_release(node => $node);

        #-------------------------------------------------------
        # If the release for this node is RollingRock or beyond
        #-------------------------------------------------------
        if ($maj_rel >= 8.1) {

            my $engine = $self->{'server'}->shell(-Server => $node,
                                                   cmd    => "sudo kenv paloma.physical");

            #-----------------------
            # If user specified BRE
            #-----------------------
            if ($self->{'engine'} =~ /BRE/) {

                #------------------------------
                # If BRE is the default engine
                #------------------------------
                if ($self->{'server'}->has_change(change => 'paloma-bre-default')) {
                    $Log->trace("BRE is the default engine for node $node");

                    #-----------------------------------------------------
                    # If paloma.physical is set to false, it will need to 
                    # be set to 'true' for this node
                    #-----------------------------------------------------
                    if ($engine =~ /false/) {
                        $Log->trace("paloma.physical was set to false for " .
                                    "node $node");
                        $Log->warn ("The Paloma engine was set to LRE" .
                                    " on node $node.  Overriding the " .
                                    "setting with $self->{'engine'}");
                        $self->{'orig_engine_type'} = 'LRE';
                        push (@node_list, $node);
                    }
                } else {  # BRE is not the default for this build
                    $Log->trace("BRE is not the default engine for node $node");
                    if ($engine !~ /true/i) {
                        $self->{'orig_engine_type'} = ($engine =~ /false/i) ?
                                                      'LRE' : 'unset';
                        push (@node_list, $node);
                    }
                }
            #----------------------------------------
            # Assume LRE was selected for the engine
            #----------------------------------------
            } else {

                #----------------------------------
                # If BRE is not the default engine
                #----------------------------------
                if (! $self->{'server'}->has_change(change => 'paloma-bre-default')) {
                    $Log->trace("BRE is not the default engine for node $node");

                    #----------------------------------------------------
                    # If paloma.physical is set to true, it will need to
                    # be set to 'false' for this node
                    #----------------------------------------------------
                    if ($engine =~ /true/) {
                        $Log->trace("paloma.physical was set to true for " .
                                    "node $node");
                        $Log->warn ("The Paloma engine was set to BRE" .
                                    " on node $node.  Overriding the " .
                                    "setting with $self->{'engine'}");
                        $self->{'orig_engine_type'} = 'BRE';
                        push (@node_list, $node);
                    }
                } else {  # BRE is the default for this build
                    $Log->trace("BRE is the default engine for node $node");
                    if ($engine !~ /false/i) {
                        $self->{'orig_engine_type'} = ($engine =~ /true/i) ?
                                                      'BRE' : 'unset';
                        push (@node_list, $node);
                    }
                }
            }

        } else {

            #---------------------------------------------------
            # If user specified BRE for pre-RollingRock release
            #---------------------------------------------------
            if ($self->{'engine'} =~ /BRE/i) {

                $Log->warn("Node $node is running a pre-RollingRock build " .
                           "and BRE was specified for engine.  Physical " .
                           "replication is not valid for pre-RollingRock " .
                           "releases."
                          );
            }
        }
    }

    #---------------------------------------------------------------
    # If any of the nodes need to have the paloma.physical flag set
    #---------------------------------------------------------------
    if (scalar @node_list) {

        NOTE::logcomment("Setting replication engine type to " .
                         $self->{'engine'});
        my %params = ();

        $params{'engine'} = $self->{'engine'};
        $params{'nodes'}  = \@node_list;

        #---------------------------------------------
        # Set the engine type for the specified nodes
        #---------------------------------------------
        subtest(\&VolOps::set_engine_type,
                '-runid',
                ($self->{'engine'} eq 'BRE') ? "set_physical" : "set_logical",
                '--',
                $self,
                %params,
               );
    } else {
        if ($maj_rel >= 8.1) {
            NOTE::logcomment("Current engine type matches requested engine " .
                             "type ($self->{'engine'}).  Engine type will " .
                             "not be set.");
        }
    }

    my @block_types        = split(',', $self->_BLOCK_TYPES);
    $self->{'block_types'} = \@block_types;

    #-------------------------------------------------------------
    # Get the common volume attribute comparison validation rules
    #-------------------------------------------------------------
    ($self->{'match_fields'},
     $self->{'diff_fields'},
     $self->{'skip_fields'},
     $self->{'match_val'})    = $self->_get_common_match_fields();

    $Log->exit();
    return $self;

} ## end sub new

=head2 set_engine_type

C<< VolOps::->set_engine_type(%params) >>

Sets the value of paloma.physical to B<true>, B<false>, its original value, or
unsets it based on the value of I<engine>.

=over 4

=item Input

=over 4

=item * engine => <scalar>

(Required)

Indicates the paloma engine to use for replication.  Possible values are:

 BRE     - Sets paloma.physical to "true", which uses physical replication.
 LRE     - Sets paloma.physical to "false", which uses logical replication.
 unset   - Unsets the paloma.physical environment variable, which uses the
           default enginefor the build.
 restore - Restores paloma.physical to the value that it was originally set
           to when the object was instantiated.  The restore option is only
           valid if the original settings were saved, which is done when
           the object is instantiated.

=item * nodes => <arrayref>

(Optional)
Default = All nodes in the cluster

List of nodes on which to set the specified engine type.

=back

=item Output

None

=item Example

 $vmv->set_engine_type(engine => "BRE");

=back

=cut

################################################################################
#
# Subroutine:   set_engine_type()
#
# Objective:    Set/unset/restore the paloma engine to use for replication
#
# Details:
#
#   Input:
#      engine - Paloma engine type to select ('BRE' or 'LRE' or 'unset' or
#               'restore').
#               Required.
#      nodes  - List of nodes on which to set the specified engine type.
#               Optional; defaults to all nodes in the cluster
#
#   Return value:
#       None
#
################################################################################

sub set_engine_type {

    $Log->enter();

    my $self = shift;
    my $params;

    eval {

        $params = {
                      validate(
                               @_,
                               {
                                 engine => {
                                            type      => SCALAR,
                                            callbacks => {'bre_or_lre' =>
                                                             sub { shift() =~ /^(BRE|LRE|unset|restore)$/ }
                                                         },
                                           },
                                 nodes  => {
                                            type      => ARRAYREF,
                                            optional  => 1,
                                           },
                               }
                              )
                  };

    };

    if ($@) {
        my $ex_text = "Parameter validation failed for set_engine_type: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    my ($engine, $node_obj);

    if (exists $self->{'cluster'}) {

        #--------------------------------------------------------------------
        # If nodes weren't specified, get a list of all nodes in the cluster
        #--------------------------------------------------------------------
        if (! exists $params->{'nodes'}) {
            my @nodes = $self->{'cluster'}->nodes;
            $params->{'nodes'} = \@nodes;
        }

        #--------------------------------------------------------------------
        # Set the Paloma engine to the specified value for each of the nodes
        #--------------------------------------------------------------------
        foreach my $node (@{$params->{'nodes'}}) {

            #-------------------------------------------------------
            # If restoring engine type, only restore for nodes that
            # had the engine specifically set
            #-------------------------------------------------------
            if ($params->{'engine'} =~ /^restore$/i) {
                if (exists $self->{'orig_engine_type'}->{$node}) {
                    $engine = $self->{'orig_engine_type'}->{$node};
                } else {
                    $engine = "";
                }
            } else {
                $engine = $params->{'engine'};
            }

            if ($engine =~ /^(BRE|LRE)$/i) {

                NOTE::logcomment("Setting the Paloma engine to $engine on " .
                                 "$node");
                eval {
                    $node_obj = NOTE::GX::Node->connect(name => $node);
                    $node_obj->transfer_engine($engine);
                };

                if ($@) {
                    my $ex_text = "Unable to connect to node $node: $@";
                    my $status  = $self->_gen_failure(ex_obj => new VolOps::ConnectionFailure($ex_text),
                                                     );

                }

            } elsif ($engine =~ /^unset$/i) {

                NOTE::logcomment("Unsetting Paloma engine on $node");
                $self->{'server'}->shell(-Server => $node,
                                         cmd     => "sudo kenv -u paloma.physical");
            } else {
                if ($params->{'engine'} !~ /^restore$/i) {

                    my $ex_text = "Unknown engine type ($engine) when trying to set Paloma engine";
                    my $status  = $self->_gen_failure(ex_obj => new VolOps::ConfigError($ex_text), 
                                                     );

                } else {

                    $Log->comment("Original engine type was not set for " .
                                  "node $node.  Engine type will not be " .
                                  "restored for that node.");

                }
            }
        }
    } else {
        my $ex_text = "Cluster connection object does not exist when trying to set Paloma engine";
        my $status  = $self->_gen_failure(ex_obj => new VolOps::InternalError($ex_text), 
                                         );
    }

    $Log->exit();

    return;

} ## end sub set_engine_type

################################################################################
#
# Subroutine:   get_engine_type()
#
# Objective:    Get the replication engine for the specified node
#
# Details:
#
#   Input:
#      node - Node for which to check engine type
#
#   Return value:
#       'BRE' or 'LRE'
#
################################################################################

sub _get_engine_type {

    $Log->enter();

    my $self = shift;
    my $params;

    eval {

        $params = {
                      validate(
                               @_,
                               {
                                 node => {
                                          type      => SCALAR,
                                         },
                               }
                              )
                  };

    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _get_engine_type: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    my ($engine, $node_obj);

    eval {
        $node_obj = NOTE::GX::Node->connect(name => $params->{'node'});
    };

    if ($@) {
        my $ex_text = "Unable to connect to node $params->{'node'}: $@";
        my $status  = $self->_gen_failure(ex_obj => new VolOps::ConnectionFailure($ex_text),
                                         );
    } else {

        my ($maj_rel, $min_rel) = $self->_get_release(node => $params->{'node'});
        #---------------------------------------
        # If this is a post-Boilermaker release
        #---------------------------------------
        if ($maj_rel >= 8.1) {

            #--------------------------------------
            # If this build has BRE set by default
            #--------------------------------------
            if ($self->{'server'}->has_change(change => 'paloma-bre-default')) {

                $Log->trace("BRE is the default engine for this build");

                $engine = 'BRE';

                my $output = $self->{'server'}->shell(cmd => "sudo kenv paloma.physical");
                if ($output =~ /false/i) {
                    $Log->trace("paloma.physical is set to false");
                    $engine = 'LRE';
                }
            #---------------------------
            # Assume LRE is the default
            #---------------------------
            } else {

                $Log->trace("BRE is not the default engine for this build");

                $engine = 'LRE';

                my $output = $self->{'server'}->shell(cmd => "sudo kenv paloma.physical");
                if ($output =~ /true/i) {
                    $Log->comment("paloma.physical is set to true");
                    $engine = 'BRE';
                }
            }
        } else {
            $Log->trace("Build is pre-RollingRock.  Engine type is 'LRE'");
            $engine = 'LRE';
        }
    }

    $Log->exit();

    return $engine;

} ## end sub _get_engine_type

=head2 get_status

C<< VolOps->get_status() >>

Return the status of the move/copy operation.  This method can be called while
the move/copy is processing if it is running in the background.

=over 4

=item Input

=over 4

=item * stat_type => <scalar>

(Optional)
Default = "detailed"

Type of status to return.

Valid values:
    detailed - Provides of detailed status of all of the jobs
               that completed
    summary  - Provides a summary status of the move/copy jobs

=back

=item Output

=over 4

=item * Status => <hashref>

Returns a reference to a hash containing the status of the jobs that have completed,
including a formatted string that can be used for printing.  The contents of
the hash include the following:

  {
   # Status for all jobs
   total_moves|total_copies  => <total number of moves/copies>,
   total_passed => <total number of moves that succeeded>,
   total_failed => <total number of moves that failed>,
   errors => [
    # This key only exists if any errors actually occurred
      {
        reason => <reason for the error>,
        [iteration => <iteration during which error occurred>],
      }
      {
        (Previous hash is repeated for each error if an
         exception was not thrown)
      }
   ],
   iteration    => <current iteration number>,
   iterations   => <total number of requested iterations>,
   duration     => <elapsed time since the start of the test>,
   formatted    => <Formatted output containing the status of all move
                    jobs that can be used for printing>

   # Status for each volume
   vserver => {
     <volume> => {
       jobs => [
        {
            # Fields returned from the job status
            priority     => <job priority>,
            restarted    => <job restarted>,
            affinity     => <job affinity>,
            uuid         => <job uuid>,
            filer        => <job filer>,
            node         => <job node>,
            state        => <job state>,
            elapsed      => <job elapsed time>,
            starttime    => <job start time>,
            category     => <job category>,
            endtime      => <job end time>,
            id           => <job ID>,
            code         => <job return code>,
            xqueuetime   => <job queue time>,
            dropdeadtime => <job drop dead time>,
            vserver      => <job vserver>,
            progress     => <job progress>,
            name         => <job name>,
            description  => <job description>,
            completion   => <job completion string>,
            username     => <job username>,
            schedule     => <job schedule>,
            jobtype      => <job type>,
          }
          {
             (Previous hash repeated for each job executed
              for this volume)
          }
       ],
       errors => [
        # This key only exists if any errors actually occurred
          {
            reason => <reason for the error>,
            [iteration => <iteration during which error occurred>],
          }
          {
            (Previous hash is repeated for each error if an
             exception was not thrown)
          }
       ],
      }
      avg_elapsed  => <Average elapsed time, in seconds, of all of the
                       move jobs for this volume>,
      num_passed   => <Total number of passes for this volume>,
      num_failed   => <Total number of failures for this volume>,
      total_moves|total_copies  => <Total move/copy jobs executed for this volume>
      worst_result => <Worst result for all move jobs for this volume>
      dest_aggr    => <Last destination aggregate to which the volume was
                       moved/copied>
   },
   {
     (Previous hash repeated for each volume on which move operations
      were performed)
   }
  }

=back

=item Example

 # Get detailed status
 $vmv->get_status();

=back

=cut

################################################################################
#
# Subroutine:   get_status
#
# Objective:    Return the status of the move/copy operation.  This method can
#               be called while the move/copy is processing if it is running
#               in the background.
#
# Details:
#
#   Input:
#      stat_type - Type of status to return - "detailed" or "summary".
#                  Optional; default = 'detailed'.
#
#   Return value:
#      status - hash ref containing the status of the jobs that have
#               completed.
#
################################################################################

sub get_status {

    $Log->enter();

    my $self = shift;
    my $params;

    eval {

        $params = {
                      validate (
                                @_,
                                {
                                    stat_type => {type => SCALAR,
                                                  optional => 1,
                                                  default  => 'detailed',
                                                  callbacks => {'det_or_sum' =>
                                                                sub { shift() =~ /^(detailed|summary)$/ }
                                                               },
                                                 },
                                }
                               )
                  };

    };

    if ($@) {
        my $ex_text = "Parameter validation failed for get_status: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    my ($status);

    #----------------------------------
    # If detailed status was requested
    #----------------------------------
    my $stat_type;

    if ($params->{'stat_type'} eq 'detailed') {
        $stat_type = "Detailed";
    } else {
        $stat_type = "Summary";
    }

    my $triggered = 0;

    if (exists $self->{'triggered'}) {
        if ($self->{'triggered'}) {
            $triggered = 1;
        }
    }
    
    # This is a code ref that will be used below in the  
    # message_get and message_put routines for "on_death" 
    # behavior 
    # 11/01/2011
    my $_on_death_ref = sub {
        
        my $status;
        
        $Log->trace("Subtest has completed. Getting " . 
                    "results from result objects");
                           
        #---------------------------------------
        # Get the move results from the subtest
        #---------------------------------------
        my @results = $self->{'st_obj'}->results();
        
        #---------------------------------------
        # Find the result object that has the status in it
        #---------------------------------------
        foreach my $result (@results) {
            my @msgs = $result->messages();
                        
            $Log->trace("Current result: " . Dumper \@msgs);
            $Log->debug("Contents of 'misc' from result " .
                        "object" . Dumper $result->{'misc'});
                
            #--------------------------------------------
            # If the subtest object contains the expected
            # string for the final result
            #--------------------------------------------
            if (grep (qr /^(Move|Copy) (passed|failed) - 
                all (moves|copies) complete$/, @msgs)) {
                    $Log->trace("Result object contains the " .
                                "expected completion string." .
                                " Obtaining final result");
                    $status  = unmarshall($result->{'misc'});
                    last;
            }
        }
                
        if (!$status) {
            $Log->warn("Unable to obtain final status " .
                       "from subtest result objects");
        }
                
        $self->{'job_status'} = $status->{'job_status'};

        return $status; 
    };

    #-----------------------------------------
    # If job is/was running in the background
    #-----------------------------------------
    if ((exists $self->{'st_obj'}) && (!$triggered)) {

        #---------------------------------
        # If the subtest is still running
        #---------------------------------
        if ($self->{'st_obj'}->is_alive()) {

            $Log->trace("Subtest is still running.  Requesting results from " .
                        "subtest");

            my $msg_type = "VolOps/" . $self->{'caller'} . "/" . $stat_type;

            #-----------------------------------------
            # Send request for status data to subtest
            #-----------------------------------------
            $Log->trace("Sending $msg_type message to listener");
            
            # There is a possibility that even after entering this
            # "if"(which is executed only when the subtest is alive),
            # the subtest could still exit(race condition).
            # We need to deal with that here. When the subtest 
            # is dead, we want to deal with this like how we would
            # deal when the subtest has finished running(i.e the "else"
            # behavior for this "if") as opposed to throwing an exception.
            # We tackle this in the "on_death" part of message_put.
            # TO DO: Ideally, when we know we have executed the "on_death"
            # part, we want to exit this "if" right away because there is
            # nothing left to accomplish in the conditional.
            $status = $self->{'st_obj'}->message_put(
                                         type     => $msg_type,
                                         on_death => $_on_death_ref,
                                         on_error => sub {
                                                     throw VolOps::BaseException
                                                     "STAF error occurred " .
                                                     "while requesting " .
                                                     "status from subtest"
                                                     },
                                         message => "unused"
                                         );

            #---------------------------------------
            # Read the status data from the subtest
            #---------------------------------------
            $Log->trace("Receiving $msg_type message from listener");

            # We do the same thing here as above.
            # This issue came up in BURT 542695.
            # "on_death" part will be called if the subtest is known 
            # not to be alive. If "on_death" behavior is called and returns
            # normally, then message_get will return that return value.  
            $status = $self->{'st_obj'}->message_get(
                                         type       => $msg_type,
                                         on_death   => $_on_death_ref,
                                         on_timeout => sub {
                                                       throw 
                                                         VolOps::BaseException
                                                       "Attempt to receive " .
                                                       "status information " .
                                                       "from subtest timed " .
                                                       "out"
                                                       },
                                         timeout    => 300,
                                         );

        } else {

            # subtest is not alive.
            # execute the function by calling the code ref defined earlier
            $status = &$_on_death_ref;
        }
    } else {  # Job ran in the foreground

        $Log->trace("Job ran in the foreground.  Obtaining results directly");

        $status = $self->_gen_status(status_type => $stat_type);

        $self->{'job_status'} = $status->{'job_status'};
    }

    $Log->debug("Status information " . Dumper $status);

    $Log->exit();

    return $status;

} ## end sub get_status

#################################################################################
#
# Subroutine:   _normalize_duration()
#
# Objective:    Converts the specified duration value to number of seconds.
#
# Details:
#
#   Input:
#      duration - duration string to normalize.  Can have one of the following
#                 appended:  D|d (days)
#                            H|h (hours)
#                            M|m (minutes)
#                            S|s (seconds)
#                  Default granularity if appendix is not provided is minutes.
#
#   Return value:
#      Scalar - Duration in seconds
#
################################################################################

sub _normalize_duration {

    $Log->enter();

    my $duration = shift;

    $Log->trace("Converting $duration to seconds");

    #------------------------------------------------------
    # Normalize the duration string into numerical seconds
    #------------------------------------------------------
    if ( $duration =~ /^\d+[Hh]$/ ) { # if value is hours
        $duration = $duration * 3600;
    } elsif ( $duration =~ /^\d+[Dd]$/ ) { # if value is days
        $duration = $duration * 86400;
    } elsif ( $duration =~ /^\d+[Mm]$/ ) { # if value is minutes
        $duration = $duration * 60;
    } elsif ( $duration =~ /^\d+[Ss]$/ ) { # if value is seconds
        $duration = $duration;
    } elsif ( $duration =~ /^\d+$/) {
        $duration = $duration * 60; # convert to seconds
    } elsif ($duration !~ /^\s*$/) {
        $Log->error("Could not parse the DURATION parameter. The value was " .
                    "'$duration'.");
    }

    $Log->exit();

    return $duration;

} ## end sub _normalize_duration

=head2 wait

 VolOps::Move->wait(%params)

Waits for the volume move/copy job(s) to complete.

=over 4

=item Input

=over 4

=item * timeout => <scalar>

(Optional)
Default = 0 (wait forever)

Indicates how long to wait before aborting the move.

=back

=item Output

None

=item Example

 $vmv->wait();

=back

=cut

#################################################################################
#
# Subroutine:   wait()
#
# Objective:    Polls the volume move job and blocks until it completes.
#
# Details:
#
#   Input:
#      timeout - Scalar - Indicates how long to wait before aborting the move.
#
#   Return value:
#      None
#
################################################################################

sub wait {

    $Log->enter();

    my $self   = shift;
    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    timeout     => {type => SCALAR,
                                                    optional => 1,
                                                    default  => 0,
                                                   },
                                }
                               )
                  };

    };

    if ($@) {
        my $ex_text = "Parameter validation failed for wait: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    #---------------------------------------
    # If background execution was requested
    #---------------------------------------
    if ($self->{'bg_mode'} =~ /^-bg$/) {

        if (exists ($self->{'st_obj'})) {

            $self->{'st_obj'}->wait_finish(subtest => [Subtest::get_live()],
                                           timeout => $params->{'timeout'});
        }

    } else {
        $Log->warn("Operation was not executed as a background task");
    }

    $Log->exit();

    return;

} ## end sub wait

=head2 status

  VolOps::status()

Returns the status of the move.

=over 4

=item Input

None

=item Output

Boolean; returns 1 if the move is still running or 0 if it has completed

=item Example

 my $alive = $vmv->status();

=back

=cut

#################################################################################
#
# Subroutine:   status()
#
# Objective:    Returns the status of the move
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Returns 1 if the subtest is still alive; returns 0 if the subtest has
#      completed.  If operation was run in the foreground, 0 will be returned.
#
################################################################################

sub status {

    $Log->enter();

    my $self = shift;
    my ($status, $st_obj);

    $status = 0;

    #--------------------------------
    # If the operation was subtested
    #--------------------------------
    if (exists $self->{'subtested'}) {
        if (exists $self->{'st_obj'}) {
            $st_obj = $self->{'st_obj'};
        } else {
            my $ex_text = "Subtest object is not available when trying to " .
                          "obtain status";
            my $status  = $self->_gen_failure(ex_obj => new VolOps::InternalError($ex_text), 
                                             );

        }

        #-------------------------------------------------------
        # Determine whether or not the subtest is still running
        #-------------------------------------------------------
        $status = $st_obj->is_alive();

    }

    $Log->exit();

    return $status;

} ## end sub status

################################################################################
#
# Subroutine:   _monitor_jobs()
#
# Objective:    Wait until the list of specified jobs has completed
#
# Details:
#
#   Input:
#      vols         - Reference to a hash of volumes objects for volumes that
#                     are currently being moved in parallel.
#      retries      - Number of times to poll the list of jobs to determine if
#                     all volume move jobs have completed.
#      sleep_int    - Number of seconds to sleep between polls.
#      on_fail      - Action to take on failure; possible values are 'exit'
#                     (exit immediately when a failure is encountered),
#                     'complete' (exit when the current group of volumes has
#                     finished), or 'continue' (record the failure, but
#                     continue processing).
#                     Optional; default = 'exit'
#      wait_type    - Indicates whether to wait for all active jobs to complete
#                     or to return when at least one job has completed.
#                     Optional; default = 'any'
#
#   Return value:
#       completed   - Reference to a hash containing the volumes whose jobs
#                     have completed
#
################################################################################

sub _monitor_jobs {

    $Log->enter();

    my $self   = shift;

    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    vols         => {type     => HASHREF,
                                                    },
                                    retries      => {type      => SCALAR,
                                                     optional  => 1,
                                                     default   => 0,
                                                    },
                                    sleep_int    => {type      => SCALAR,
                                                     optional  => 1,
                                                     default   => 3,
                                                    },
                                    on_fail      => {type      => SCALAR,
                                                     optional  => 1,
                                                     default   => 'exit',
                                                     callbacks => {'fail_flg' =>
                                                                       sub { shift() =~ /^(continue|complete|exit)$/ },
                                                                  },
                                                    },
                                    wait_type    => {type      => SCALAR,
                                                     optional  => 1,
                                                     default   => $self->_DEF_WAIT_TYPE,
                                                     callbacks => {'all_or_any' =>
                                                                      sub { shift() =~ /^(all|any)$/ },
                                                                  },
                                                    },
                                }
                               )
                   };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _monitor_jobs: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    $self->{'monitoring'} = 1;

    my ($cluster, $count, $job_id, $job_obj, $job_status, $vol_name);

    my $vols          = $params->{'vols'};
    my $max_retries   = $params->{'retries'};
    my $done          = 0;
    my $retries       = 0;
    my @fails         = ();
    my $deferred_vols = "";

    my $num_jobs  = scalar keys %{$vols};

    $Log->debug("The number of jobs running are: " . $num_jobs);
    $Log->debug("The volumes that are moving are: " . Dumper keys %{$vols});

    my %completed = ();
    my %deferred  = ();

    #------------------------------------------------------------
    # Wait for job(s) to complete or until retries are exhausted
    #------------------------------------------------------------
    while (!$done) {
        foreach my $vol_info (values %{$vols}) {
            $Log->debug("Volume info for the current volume: " .
                        Dumper $vol_info);

            my $vol_name = $vol_info->{'volume'};
            my $vserver  = $vol_info->{'vserver'};
            my $key      = "$vol_name:$vserver";

            #--------------------------------------------------------------
            # If copying volume, get the object for the destination volume
            #--------------------------------------------------------------
            my $vol = ($self->{'caller'} =~ /Copy/) ?
                       $self->{'vol_objs'}->{$vol_info->{'curr_dest_volume'}}->{$vserver} :
                       $self->{'vol_objs'}->{$vol_name}->{$vserver};
            $job_id = $vol->{'_job_id'};

            $Log->trace("Looking at $vol_name ($vserver) to see if it " .
                        "completed");

            #---------------------------------------
            # If this job has not already completed
            #---------------------------------------
            if (! exists $completed{$key}) {

                my $job_stat = $self->_get_job_status(volume  => $vol_name,
                                                      vserver => $vserver,
                                                      job_id  => $job_id,
                                                     );

                my $job_state = $job_stat->{'state'};
                my $job_code  = $job_stat->{'code'};

                $job_stat->{'dest_aggr'} = $vol_info->{'dest_aggr'}->name;
                $job_stat->{'src_aggr'}  = $vol_info->{'src_aggr'};

                $job_stat = $self->_update_job_stats(volume     => $vol_name,
                                                     vserver    => $vserver,
                                                     job_status => $job_stat,
                                                     vol_obj    => $vol,
                                                    );

                $Log->debug("The status of job $job_id is " .
                            Dumper $job_stat);

                #----------------------
                # If job has completed
                #----------------------
                if (($job_stat->{'current_status'} =~ /complete/i) ||
                    ($job_stat->{'current_status'} =~ /quit/i)) {

                    $Log->trace("Job $job_id completed.  Getting job " .
                                "statistics");
                    $completed{$key} = $job_stat;

                    if ($job_code != 0) {
                        $Log->trace("Job $job_id did not complete successfully");
                        #-----------------------------------------------
                        # If continue on failure or complete all active
                        # jobs before bailing is set
                        #-----------------------------------------------
                        if ($params->{'on_fail'} eq 'continue' or
                            $params->{'on_fail'} eq 'complete') {
                            $Log->warn("Job did not succeed for volume " .
                                       "$vol_name (vserver $vserver).  " .
                                       "Job status: " .  $job_state);
                            push @fails, $completed{$key};
                        } else { # Exit on failure
                            my $ex_text = "Job did not complete successfully." .
                                          "  Job description: " .
                                          $job_stat->{'description'} . "\n" .
                                          "Job stats: " . $job_state . "\n" .
                                          "Job progress: " .
                                          $job_stat->{'progress'} . "\n" .
                                          Dumper \%completed;

                            $self->{'monitoring'} = 0;
                            my $status = $self->_gen_failure(ex_obj  => new VolOps::JobFailed($ex_text),
                                                             vserver => $vserver,
                                                             volume  => $vol_name,
                                                            );

                        }
                    }

                    #--------------------------------------------------
                    # If move job should have been deferred but wasn't
                    #--------------------------------------------------
                    if ($vol_info->{'cutover_action'} =~ /wait/ &&
                        $vol_info->{'auto_trigger'} =~ /off/) {

                        my $ex_text = "Cutover action was 'wait' and move " .
                                      "job completed with no auto trigger " .
                                      "for $vol_name ($vserver)";
                        $Log->trace($ex_text);

                        $self->{'monitoring'} = 0;

                        my $status = $self->_gen_failure(ex_obj => new VolOps::Move::JobNotDeferred($ex_text),
                                                         fatal  => ($self->{'on_fail'} =~ /continue/i)
                                                                   ? 0 : 1,
                                                        );
                    }

                } else {
                    $Log->debug("Job appears to still be running");
                    $Log->debug("Job description: " .
                                $job_stat->{'description'});
                    $Log->debug("Job state: " . $job_state);
                    $Log->debug("Job progress:  " .
                                $job_stat->{'progress'});

                    my $deferred = 0;

                    #-----------------------------------------------------
                    # If the move job is hard or soft deferred, check the
                    # auto-trigger setting for this volume to see if
                    # trigger-cutover should be executed
                    #-----------------------------------------------------
                    if ($job_stat->{'current_status'} =~ /deferred/i) {
                        $Log->trace("Job $job_stat->{'id'} is deferred");

                        if ($vol_info->{'auto_trigger'} !~ /off/i) {

                            NOTE::logcomment("Automatic trigger-cutover was " .
                                             "requested.  " .
                                             "Triggering cutover for " .
                                             "$vol_name ($vserver)");

                            #-------------------------------------------------
                            # Attempt to trigger the cutover.  This will
                            # remove the deferred_job from the job status
                            # hash and update the trigger_attempts counter.
                            #-------------------------------------------------
                            try {
                                $Log->trace("Triggering cutover for $vol_name ($vserver)");
                                $self->_trigger_cutover(volume  => $vol_name,
                                                        vserver => $vserver,
                                                        force   => ($vol_info->{'auto_trigger'} =~ /noforce/i)
                                                                    ? 0 : 1,
                                                        job_id  => $job_stat->{'id'},
                                                       );
                            } catch VolOps::Move::TriggerFailed with {
                                $Log->trace("Trigger cutover failed");
                                $self->{'monitoring'} = 0;
                                my $ex     = shift;
                                my $status = $self->_gen_failure(ex_obj  => $ex,
                                                                 volume  => $vol_name,
                                                                 vserver => $vserver,
                                                                 fatal   => $self->{'on_fail'} =~ /continue/i
                                                                            ? 0 : 1,
                                                                );
                            };
                        } else {
                            $Log->trace("Job $job_id is hard deferred and " .
                                        "auto-trigger was not specified.");

                            NOTE::logcomment("Job $job_id is hard deferred and " . 
                                             "auto-trigger was not specified.");

                            #---------------------------------------
                            # Add this to the list of deferred jobs
                            #---------------------------------------
                            $deferred{$key} = $job_stat;

                            if ($vol_info->{'exit_on_deferral'}) {

                                #--------------------------------------------
                                # Finish processing the remaining volumes in
                                # this iteration so the status in case more
                                # than one volume is deferred
                                #--------------------------------------------
                                $Log->trace("Move job for volume $vol_name " .
                                            "($vserver) is deferred and " .
                                            "exit_on_deferral was specified");
                                $deferred_vols .= "Move job $job_id for " .
                                                  "$vol_name ($vserver) is " .
                                                  "deferred";

                                $done = 1;

                            }
                        }
                    }

                    $Log->debug("Job status for $vol_name ($vserver): " .
                                Dumper $self->{'job_status'}->{$vserver}->{$vol_name});
                }
            }
        }

        my $num_done = scalar keys %completed;

        if ($self->{'caller'} =~ /Move/i) {
            $num_done += scalar keys %deferred;
        }

        $Log->trace("The number of completed jobs is " . $num_done);

        #-------------------------------------
        # If waiting for all jobs to complete
        #-------------------------------------
        if ($params->{'wait_type'} eq 'all') {
            $Log->trace("Waiting for all jobs to complete");
            if ($num_done == $num_jobs) {
                $done = 1;
            }
        } else {  ## Waiting for at least one job to complete
            $Log->trace("Waiting for any job to complete");
            if ($num_done > 0) {
                $done = 1;
            }
        }

        #--------------------------------------------------------------------
        # If there are any deferred volumes and exit_on_deferral was set for
        # those volumes
        #--------------------------------------------------------------------
        if ($deferred_vols) {
            $Log->trace("Exit on deferral was specified and there were " .
                        "deferred jobs");
            $self->{'monitoring'} = 0;
            VolOps::Move::MoveDeferred->throw($deferred_vols);
        }

        #--------------------------------------------------------------
        # If all jobs are not complete and a max number of retries was
        # specified, see if retries have been exhausted
        #--------------------------------------------------------------
        if (!$done) {
            my $left = $num_jobs - $num_done;
            $Log->comment("There are $left jobs that are not complete");

            $retries++;

            if ($retries == $max_retries) {
                $done = 1;
                my $ex_text = "Still waiting for jobs to complete and the " .
                              "number of retries were exhausted.  Exiting...";
                my $status = $self->_gen_failure(ex_obj => new VolOps::RetriesExhausted($ex_text), 
                                                 fatal  => ($self->{'on_fail'} =~ /continue/i)
                                                           ? 0 : 1,
                                                );

            }
            snooze $params->{'sleep_int'};
        }
    }

    $self->{'monitoring'} = 0;

    #--------------------------------------------------------------------------
    # If complete on failure was specified and at least one of the jobs failed
    #--------------------------------------------------------------------------
    if ($params->{'on_fail'} eq 'complete') {
        my $num_fails = scalar @fails;

        if ($num_fails) {

            $Log->trace("on_failure is set to 'complete' and $num_fails " .
                        "job(s) failed");
            my $ex_text = $num_fails .
                          (($num_fails == 1) ? " job " : " jobs ") .
                          "did not complete successfully." .
                          Dumper \@fails;
            my $status  = $self->_gen_failure(ex_obj => new VolOps::JobFailed($ex_text),
                                             );

        }
    }

    $Log->trace("The contents of the completed hash are " . Dumper \%completed);

    $Log->exit();

    if ($self->{'caller'} =~ /Move/i) {
        return (\%completed,\%deferred);
    } else {
        return \%completed;
    }

} ## end sub _monitor_jobs

################################################################################
#
# Subroutine:   _get_job_status()
#
# Objective:    Get the job status for the specified job.
#
# Details:
#
#   Input:
#      volume  - Volume associated with the specified job.  This is only needed
#                to throw an exception if there is a failure obtaining the
#                status. (Required)
#      vserver - Vserver associated with the specified job.  This is only
#                needed to throw an exception if there is a failure obtaining
#                the status.  (Required)
#      job_id  - ID of the job for which to obtain a status. (Required)
#
#   Return value:
#      None
#
################################################################################

sub _get_job_status {

    $Log->enter();

    my $self = shift;
    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    volume  => {type => SCALAR,
                                               },
                                    vserver => {type => SCALAR,
                                               },
                                    job_id  => {type => SCALAR,
                                               },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _get_job_status: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    my $job_stat = ();

    $Log->trace("Obtaining job status for job $params->{'job_id'}");

    eval {
        $job_stat = $self->{'job_obj'}->status(id     => $params->{'job_id'},
                                               fields => 'status'
                                              );

        $Log->debug("Job status: " . Dumper $job_stat);

        #-------------------------------------------------------
        # If the job is no longer in the active job list, check
        # job history
        #-------------------------------------------------------
        if (!$job_stat) {
            $Log->trace("Unable to locate job status in active job list.  " .
                        "Checking job history...");
            $job_stat = $self->{'job_obj'}->status(id      => $params->{'job_id'},
                                                   fields  => 'status',
                                                   history => 1,
                                                  );
        }
    };

    if ($@) {
        $Log->trace("An error occurred while obtaining job status: $@");
        my $ex_text = "Attempt to obtain status for job $params->{'job_id'} " .
                      "failed: $@";
        my $status  = $self->_gen_failure(ex_obj  => new VolOps::JobStatusFailure($ex_text),
                                          vserver => $params->{'vserver'},
                                          volume  => $params->{'volume'},
                                          fatal   => ($self->{'on_fail'} =~ /continue/i)
                                                     ? 0 : 1,
                                         );
    }

    $Log->exit();

    return $job_stat;

} ## end sub _get_job_status

################################################################################
#
# Subroutine:   _get_status()
#
# Objective:    Get a detailed status and return to the caller
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      None
#
################################################################################

sub _get_status {

    $Log->enter();

    my ($event, $listener) = @_;

    my $self = shift;

    my $stat_type =~ s/^VolOps\/.*\/(.*)$/$1/;


    my $status = $self->_gen_status(status_type => $listener);

    my $msg_type = "VolOps/" . $self->{'caller'} . "/" . $listener;

    $Log->trace("Sending $msg_type response to caller");
    $Parent->message_put(type    => $msg_type,
                         message => $status);
    $Log->trace("Finished sending $msg_type response");

    #-----------------------------------------------------
    # If user requested to save the status data to a file
    # (Only save the data if there was some move status
    #  available)
    #-----------------------------------------------------
    if (($self->{'save_status'}) &&
        ($status->{'formatted'} !~ /No Data Available/)) {
        if (exists $self->{'dss_obj'}) {

            if (! exists($self->{'options'}->{'DSS_FILENAME'})) {

                eval {

                    $self->{'dss_obj'}->dss_update_file(file => 'Vol' .
                                                            $self->{'caller'} .
                                                            "_" . 'status_' .
                                                            time()
                                                       );

                };

                if ($@) {
                    my $ex_text = "Failed to create DSS filename: $@";
                    $Log->error($ex_text);
                    my $status  = $self->_gen_failure(ex_obj => new VolOps::DataStructureStoreFailure($ex_text), 
                                                      fatal  => ($self->{'on_fail'} =~ /continue/i)
                                                                ? 0 : 1,
                                                     );
                }
            }

            $self->{'dss_obj'}->dss_store($status);

        } else {
            my $ex_text = "Storage of status data was selected but " .
                          "DataStructureStore object does not exist";
            my $status  = $self->_gen_failure(ex_obj => new VolOps::DataStructureStoreFailure($ex_text), 
                                              fatal  => ($self->{'on_fail'} =~ /continue/i)
                                                        ? 0 : 1,
                                             );

        }
    }

    $Log->exit();

} ## end sub _get_status

################################################################################
#
# Subroutine:   _gen_status()
#
# Objective:    Generate a detailed or summary status depending what was
#               specified
#
# Details:
#
#   Input:
#       status_type - Type of status to generate - "Detailed" or "Summary"
#
#   Return value:
#      Status hash
#
################################################################################

sub _gen_status {

    my $self = shift;

    my $params;

    $self->{'generating_status'} = 1;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    status_type => {type      => SCALAR,
                                                    optional  => 1,
                                                    default   => "Detailed",
                                                    callbacks => {'det_or_sum' =>
                                                                  sub { shift() =~ /^(Detailed|Summary)$/ }
                                                                 },
                                                   },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _gen_status: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    $Log->enter("_gen_status called with the following parameters:\n" .
                Dumper $params);

    my $status_type     = $params->{'status_type'};
    my @log_str         = ();
    my @header_str      = ();

    $Log->debug("The contents of the job status are: " .
                Dumper $self->{'job_status'});

    #----------------------------
    # Process header information
    #----------------------------
    my ($elapsed, $op_string, $total_ops_key);

    push (@header_str, ('*' x 250));
    push (@header_str, sprintf("%sVolume %s %s Status\n\n", (' ' x 50),
                               $self->{'caller'},
                               $status_type));
    push (@header_str, sprintf("Iteration %d of %d", $self->{'iter_cnt'},
                               $self->{'iterations'}));
    if ($self->{'start_time'}) {
        push (@header_str, sprintf("Start time:   " .  localtime($self->{'start_time'})));
    }

    #----------------------------------
    # If all iterations have completed
    #----------------------------------
    if (exists $self->{'end_time'}) {
        push (@header_str,  sprintf("End time:     " .  localtime($self->{'end_time'})));
        $elapsed = $self->_normalize_time(($self->{'end_time'} - $self->{'start_time'}), 1);
        push (@header_str, "Status:       Completed");
    } else {
        if ($self->{'start_time'} > 0) {
            $elapsed = $self->_normalize_time((time() - $self->{'start_time'}), 1);
            push (@header_str, "Status:       Running");
        } else {
            push (@header_str, "Status:       Not started");
            $elapsed = 0;
        }
    }

    push (@header_str, sprintf("Elapsed time: $elapsed\n\n"));

    my %job_status        = ();
    my @deferred_summary  = ();

    $Log->debug("The job status is " . Dumper $self->{'job_status'});

    #--------------------------------
    # Set operation-specific strings
    #--------------------------------
    if ($self->{'caller'} =~ /Move/i) {
        $op_string = 'moves';
    } elsif ($self->{'caller'} =~ /Copy/i) {
        $op_string = 'copies';
    } else {
        $Log->warn("Unknown caller: $self->{'caller'}");
        $op_string = 'ops';
    }

    $total_ops_key = 'total_' . $op_string;

    #--------------------------------------------
    # If there is any status available to report
    #--------------------------------------------
    if (scalar keys %{$self->{'job_status'}}) {

#        $job_status{'job_status'} = %{$self->{'job_status'}};
        %{$job_status{'job_status'}} = %{$self->{'job_status'}};

        $Log->debug("Processing the job status information: " .
                    Dumper \%job_status);

        #-------------------
        # Initialize totals
        #-------------------
        $job_status{'total_passed'} = 0;
        $job_status{'total_failed'} = 0;
        $job_status{'total_active'} = 0;
        $job_status{$total_ops_key} = 0;

        if ($self->{'caller'} =~ /Move/i) {
            $job_status{'total_hard_deferrals'} = 0; 
            $job_status{'total_deferred'}       = 0;
        }

        #------------------------------------------
        # Process the results from all of the jobs
        #------------------------------------------
        foreach my $vserver (sort(keys %{$self->{'job_status'}})) {

            $Log->trace("Processing status for vserver $vserver");

            push (@log_str, sprintf("\nVserver: %15s$vserver", " "));

            foreach my $vol (sort(keys %{$self->{'job_status'}->{$vserver}})) {

                $Log->trace("Processing status for volume $vol " .
                            "(vserver: $vserver)");

                $job_status{$vserver}{$vol}{$total_ops_key} = 0;
                $job_status{$vserver}{$vol}{'num_passed'}   = 0;
                $job_status{$vserver}{$vol}{'num_failed'}   = 0;
                $job_status{$vserver}{$vol}{'num_active'}   = 0;

                if ($self->{'caller'} =~ /Move/i) {
                    $job_status{$vserver}{$vol}{'num_deferred'} = 0;

                }

                $job_status{$vserver}{$vol}{'dest_aggr'} = $self->{'job_status'}->{$vserver}->{$vol}->{'dest_aggr'};

                #--------------------------------------------------------------
                # If volume copy, add the last known destination volume to the
                # overall volume status
                #--------------------------------------------------------------
                if ($self->{'caller'} =~ /Copy/i) {
                    $job_status{$vserver}{$vol}{'curr_dest_volume'} = $self->{'job_status'}->{$vserver}->{$vol}->{'curr_dest_volume'};
                }

                push(@log_str, sprintf("\nVolume:%17s$vol", " "));

                #--------------------------------------------------------
                # Get the status for the most recent job for this volume
                # If the status has changed, make sure it's up to date
                # in the job status hash.
                #--------------------------------------------------------
                if (! $self->{'monitoring'}) {
                    my $job_stats = $self->_update_job_status(volume   => $vol,
                                                              vserver  => $vserver,
                                                             );
                    #------------------------------------
                    # Update the status with any changes
                    #------------------------------------
                    $job_status{'job_status'}{$vserver}{$vol} = $self->{'job_status'}->{$vserver}->{$vol};
                }

                if ($self->{'caller'} =~ /Move/i) {
                    if (exists ($self->{'job_status'}->{$vserver}->{$vol}->{'deferrals'})) {
                        $Log->comment("Deferrals: " . Dumper $self->{'job_status'}->{$vserver}->{$vol}->{'deferrals'});
                        foreach my $job_id (keys %{$self->{'job_status'}->{$vserver}->{$vol}->{'deferrals'}}) {
                            if (exists $self->{'job_status'}->{$vserver}->{$vol}->{'deferrals'}->{$job_id}->{'deferral_cnt'}) {
                                $job_status{'total_hard_deferrals'} += $self->{'job_status'}->{$vserver}->{$vol}->{'deferrals'}->{$job_id}->{'deferral_cnt'};
                            }
                        }
                    }
                }
 
                #-----------------------------------------
                # Process the data for the completed jobs
                #-----------------------------------------
                if (exists $self->{'job_status'}->{$vserver}->{$vol}->{'jobs'}) {

                    $Log->trace("Processing job information for volume $vol " .
                                "(vserver: $vserver)");
                    $job_status{$vserver}{$vol}{$total_ops_key} = scalar @{$self->{'job_status'}->{$vserver}->{$vol}->{'jobs'}};

                    push(@log_str,
                         sprintf("Total %s:%12s$job_status{$vserver}{$vol}{$total_ops_key}",
                                 $op_string, " "));

                    my @job_str = ();
                    my ($cnt, $end, $start, $tot_elaps) = (0, 0, 0, 0);
                    $job_status{$vserver}{$vol}{'worst_result'} = 'pass';

                    foreach my $job (@{$self->{'job_status'}->{$vserver}->{$vol}->{'jobs'}}) {

                        $Log->trace("Processing job status information for " .
                                    "job $job->{'id'}");

                        $Log->trace("The contents of job are " . Dumper $job);
                        if ($job->{'code'}) {
                            $job_status{$vserver}{$vol}{'worst_result'} = 'fail';
                            $job_status{$vserver}{$vol}{'num_failed'}++;
                        } else {
                            $job_status{$vserver}{$vol}{'num_passed'}++;
                        }
                        $cnt++;

                        if ($status_type eq 'Detailed') {
                            push (@job_str, @{$self->_gen_job_str(job_stat => $job,
                                                                  volume   => $vol,
                                                                  vserver  => $vserver)});
                        }

                        $tot_elaps += $job->{'elapsed'};
                    }

                    $job_status{'total_passed'} += $job_status{$vserver}{$vol}{'num_passed'};
                    $job_status{'total_failed'} += $job_status{$vserver}{$vol}{'num_failed'};

                    $job_status{$total_ops_key} += $job_status{$vserver}{$vol}{$total_ops_key};
                    my $avg = $tot_elaps / $cnt;
                    $job_status{$vserver}{$vol}{'avg_elapsed'} = sprintf("%.2f", $avg);
                    push (@log_str, "Average execution time: " .
                                    $self->_normalize_time($avg, 1) . "\n");

                    if ($status_type eq 'Detailed') {
                        $Log->trace("Processing data for detailed status");
                        push (@log_str, "\n");
                        push (@log_str, "\n");
                        push (@log_str, "Completed Jobs:\n");
                        push (@log_str, "\n");

                        push (@log_str, @{$self->_gen_header_str(header_type => 'complete')});
                        push (@log_str, @job_str);
                        push (@log_str, "\n\n");
                    }
                }

                #---------------------------------------------------------------
                # If executing volume move and the move job for this volume is
                # currently deferred
                #---------------------------------------------------------------
                if ($self->{'caller'} =~ /Move/i && (exists $self->{'job_status'}->{$vserver}->{$vol}->{'deferred_job'})) {
                    $Log->trace("Processing deferred job information for " .
                                "volume $vol (vserver: $vserver)");

                    $job_status{$vserver}{$vol}{'num_deferred'}++;
                    $job_status{$vserver}{$vol}{'deferred_job'} = $self->{'job_status'}->{$vserver}->{$vol}->{'deferred_job'};

                    $job_status{'total_deferred'}       += $job_status{$vserver}{$vol}{'num_deferred'};

                    my %deferred_info = ();

                    #---------------------------------------------------------
                    # Save the deferral information for the deferral summary
                    #---------------------------------------------------------
                    $deferred_info{'volume'}  = $vol;
                    $deferred_info{'vserver'} = $vserver;
                    $deferred_info{'job_id'}  = $self->{'job_status'}->{$vserver}->{$vol}->{'deferred_job'}->{'id'};
                    push (@deferred_summary, \%deferred_info);

                    if ($status_type eq 'Detailed') {
                        push (@log_str, "Deferred Job:\n");
                        push (@log_str, "\n");

                        push (@log_str, @{$self->_gen_header_str(header_type => 'active')});

                        push (@log_str, @{$self->_gen_job_str(job_stat => $self->{'job_status'}->{$vserver}->{$vol}->{'deferred_job'},
                                                              volume   => $vol,
                                                              vserver  => $vserver)});

                        if ($self->{'job_status'}->{$vserver}->{$vol}->{'deferrals'}->{$deferred_info{'job_id'}}->{'was_soft_deferred'}) {
                            push (@log_str, "\n\t\t\t*** Job $deferred_info{'job_id'} was soft deferred ***\n");
                        }
                    }

                } else {

                    #-------------------------------------
                    # Process the data for the active job
                    #-------------------------------------
                    if (exists $self->{'job_status'}->{$vserver}->{$vol}->{'active_job'}) {
                        $Log->trace("Processing active job information for " .
                                    "volume $vol (vserver: $vserver)");
                        $job_status{$vserver}{$vol}{'num_active'}++;

                        $job_status{'total_active'} += $job_status{$vserver}{$vol}{'num_active'};

                        if ($status_type eq 'Detailed') {
                            push (@log_str, "Active Job:\n");
                            push (@log_str, "\n");

                            push (@log_str, @{$self->_gen_header_str(header_type => 'active')});

                            push (@log_str, @{$self->_gen_job_str(job_stat => $self->{'job_status'}->{$vserver}->{$vol}->{'active_job'},
                                                                  volume   => $vol,
                                                                  vserver  => $vserver)});
                        }

                    }
                }

                #--------------------------------------------------------------
                # If there are any volume-related errors associated with this
                # volume
                #--------------------------------------------------------------
                if (exists $self->{'job_status'}->{$vserver}->{$vol}->{'errors'}) {
                    $job_status{$vserver}{$vol}{'errors'} =
                        $self->{'job_status'}->{$vserver}->{$vol}->{'errors'};
                    push (@log_str, sprintf("\n\tErrors for volume $vol ($vserver):"));
                    foreach my $error (@{$self->{'job_status'}->{$vserver}->{$vol}->{'errors'}}) {
                        push (@log_str, sprintf("\t\t$error->{'reason'}%s\n"));
                    }
                    push (@log_str, "\n");
                }
            }
        }

    } else {

        $job_status{'total_passed'} = 0;
        $job_status{'total_failed'} = 0;
        $job_status{'total_active'} = 0;
        $job_status{$total_ops_key} = 0;

        my $found_errors = 0;

        #------------------------------------------------------------------------
        # Check to see if there were any errors generated for any of the volumes
        #------------------------------------------------------------------------
        if ($self->{'total_errors'} > 0) {

            push (@log_str, "\nErrors:\n");

            #------------------------
            # Process general errors
            #------------------------
            if (exists $self->{'errors'}) {
                if (scalar @{$self->{'errors'}}) {
                    push (@log_str, "\n\n\tGeneral errors:\n");
                    foreach my $error (@{$self->{'errors'}}) {
                        push (@log_str, "\t\t$error");
                    }
                    $found_errors = 1;
                }
            }

            #--------------------------------
            # Process volume-specific errors
            #--------------------------------
            foreach my $vol_data (@{$self->{'src_vols'}}) {
                my $volume  = $vol_data->{'volume'};
                my $vserver = $vol_data->{'vserver'};
                if (exists $self->{$vserver}->{$volume}->{'errors'}) {
                    $job_status{$vserver}{$volume}{'errors'} =
                        $self->{$vserver}->{$volume}->{'errors'};
                    push (@log_str, sprintf("\n\n\tErrors for volume $volume ($vserver):"));
                    $found_errors = 1;
                    foreach my $error (@{$self->{$vserver}->{$volume}->{'errors'}}) {
                        push (@log_str, sprintf("\t\tIteration $error->{'iteration'}: $error->{'reason'}%s"));
                    }
                }
            }
        }

        if (!$found_errors) {
            push (@log_str, sprintf("\n%s*** No Data Available ***\n\n\n", (' ' x 28)));
        }

    }

    push (@log_str, ('*' x 250));

    #------------------------------------------------------------
    # If there were any deferred volumes, include a summary at
    # the beginning of the formatted output and add the deferred
    # information to the status hash
    #------------------------------------------------------------
    if (scalar @deferred_summary) {
        push (@{$job_status{'deferred_jobs'}}, @deferred_summary);

        my @deferred_log = ();

        push (@deferred_log, "Deferred Job Summary:\n");
        push (@deferred_log, "\n");

        push (@deferred_log, sprintf("%-8s%-25s%-s", "Job ID", "Vserver", "Volume"));
        push (@deferred_log, sprintf("%-8s%-25s%-s", ('-' x 3), ('-' x 7), ('-' x 6)));
        foreach my $deferred_job (@deferred_summary) {
            push (@deferred_log, sprintf("%-8s%-25s%-s",
                                         $deferred_job->{'job_id'},
                                         $deferred_job->{'vserver'},
                                         $deferred_job->{'volume'},
                                        )
                 );
        }
        unshift (@log_str, @deferred_log);
    }

    unshift (@log_str, @header_str);

    #------------------------------------------------
    # Add a human-readable format to the status hash
    #------------------------------------------------
    $job_status{'formatted'} =  join("\n", @log_str);
    $job_status{'formatted'} =~ s/\n\n\n/\n/g;

    #-----------------------------
    # Add the summary information
    #-----------------------------
    $job_status{'iteration'}    = (exists ($self->{'iter_cnt'}))
                                   ? $self->{'iter_cnt'} : 0;
    $job_status{'iterations'}   = $self->{'iterations'};
    $job_status{'duration'}     = time() - $self->{'start_time'};
    $job_status{'total_errors'} = $self->{'total_errors'};

    #--------------------------------------
    # If any general errors were generated
    #--------------------------------------
    if (exists $self->{'errors'}) {
        $job_status{'errors'} = $self->{'errors'};
    }

    $Log->trace("The status being returned is:\n" . Dumper \%job_status);

    $Log->exit();

    $self->{'generating_status'} = 0;

    return \%job_status;

} ## end sub _gen_status

################################################################################
#
# Subroutine:   _gen_job_str()
#
# Objective:    Generates the printable job information
#
# Details:
#
#   Input:
#       job_stat - Job status for the current job being processed
#       volume   - Volume associated with the job
#       vserver  - Vserver on which the volume resides
#
#   Return value:
#      Formatted job string containing statistics for the job
#
################################################################################

sub _gen_job_str {
    $Log->enter();

    my $self   = shift;
    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    job_stat => {type => HASHREF,
                                                },
                                    volume   => {type => SCALAR,
                                                },
                                    vserver  => {type => SCALAR,
                                                },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _gen_job_str: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    my ($elapsed, @job_str, $job_format);

    my $job_stat = $params->{'job_stat'};
    my $vol      = $params->{'volume'};
    my $vserver  = $params->{'vserver'};

    if ($job_stat->{'current_status'} =~ /complete/i) {
        $elapsed  = $job_stat->{'elapsed'};
    } else {
        $elapsed  = time() - str2time($job_stat->{'starttime'});
    }

    #----------------------------------------------------------
    # Generate the printable job string based on the operation
    #----------------------------------------------------------
    if ($self->{'caller'} =~ /Move/i) {
        my $job_format = "%-6s  %-16s%-16s%-24s%-12s%-6s%-11s%-10s%-13s%-20s%-35s%-35s%-75s%-20s";

        push (@job_str, sprintf($job_format,
                                $job_stat->{'id'},
                                $job_stat->{'starttime'},
                                ($job_stat->{'current_status'} =~ /complete/i) ? $job_stat->{'endtime'} : '-',
                                $self->_normalize_time($elapsed,0),
                                $job_stat->{'state'},
                                $job_stat->{'code'},
                                $self->{'job_status'}->{$vserver}->{$vol}->{'deferrals'}->{$job_stat->{'id'}}->{'deferral_cnt'},
                                $self->{'job_status'}->{$vserver}->{$vol}->{'deferrals'}->{$job_stat->{'id'}}->{'trigger_attempts'},
                                $job_stat->{'name'},
                                $job_stat->{'node'},
                                $job_stat->{'src_aggr'},
                                $job_stat->{'dest_aggr'},
                                $job_stat->{'description'},
                                ($job_stat->{'current_status'} =~ /complete/i) ? $job_stat->{'completion'} : $job_stat->{'progress'},
                               )
             );


    } else {

        my $job_format = "%-6s  %-16s%-16s%-24s%-12s%-6s%-13s%-20s%-35s%-35s%-75s%-20s%s";

        push (@job_str, sprintf($job_format,
                                $job_stat->{'id'},
                                $job_stat->{'starttime'},
                                ($job_stat->{'current_status'} =~ /complete/i) ? $job_stat->{'endtime'} : '-',
                                $job_stat->{'endtime'},
                                $self->_normalize_time($elapsed,0),
                                $job_stat->{'state'},
                                $job_stat->{'code'},
                                $job_stat->{'name'},
                                $job_stat->{'node'},
                                $job_stat->{'src_aggr'},
                                $job_stat->{'dest_aggr'},
                                $job_stat->{'description'},
                                ($job_stat->{'current_status'} =~ /complete/i) ? $job_stat->{'completion'} : $job_stat->{'progress'},
                               )
             );
    }

    $Log->exit();

    return (\@job_str);

} ## end sub _gen_job_str

################################################################################
#
# Subroutine:   _gen_header_str()
#
# Objective:    Generates the header for the printable job information
#
# Details:
#
#   Input:
#      header_type - Indicates whether to generate a header for a completed
#                    job or a running job.
#
#   Return value:
#      Reference to an array containing the header information.
#
################################################################################

sub _gen_header_str {

    $Log->enter();

    my $self    = shift;
    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    header_type => {type  => SCALAR,
                                                    regex => qr/active|complete/,
                                                   },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _gen_header_str: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    my (@header, $header_format, $last_col);

    if ($params->{'header_type'} =~ /complete/i) {
        $last_col = "Completion";
    } else {
        $last_col = "Progress";
    }

    if ($self->{'caller'} =~ /Move/i) {
        my $header_format = "%-8s%-16s%-16s%-24s%-12s%-6s%-11s%-10s%-13s%-20s%-35s%-35s%-75s%-20s";

        push (@header, sprintf($header_format,
                               "Job ID", "Start", "End", "Elapsed",
                               "State", "Code", "Deferrals", "Triggers",
                               "Name", "Node", "Source Aggregate",
                               "Destination Aggregate",
                               "Description", $last_col,
                              )
             );

        push (@header, sprintf($header_format,
                                ('-' x 6), ('-' x 5), ('-' x 3),
                                ('-' x 7), ('-' x 5), ('-' x 4),
                                ('-' x 9), ('-' x 8), ('-' x 4),
                                ('-' x 4), ('-' x 16), ('-' x 21),
                                ('-' x 11), ('-' x length($last_col)),
                              )
             );

    } else {

        my $header_format = "%-8s%-16s%-16s%-24s%-12s%-6s%-13s%-20s%-35s%-35s%-75s%-20s%s";

        push (@header, sprintf($header_format,
                               "Job ID", "Start", "End", "Elapsed",
                               "State", "Code", "Name", "Node",
                               "Source Aggregate", "Destination Aggregate",
                               "Description", $last_col));

        push (@header, sprintf($header_format,
                               ('-' x 6), ('-' x 5), ('-' x 3),
                               ('-' x 7), ('-' x 5), ('-' x 4),
                               ('-' x 4), ('-' x 4), ('-' x 16),
                               ('-' x 21), ('-' x 11), ('-' x length($last_col)),
                              )
             );

    }

    $Log->exit();

    return (\@header);

} ## end sub _gen_header_str

################################################################################
#
# Subroutine:   _normalize_time()
#
# Objective:    Convert the total elapsed time days, hours, minutes and seconds
#
# Details:
#
#   Input:
#      (Positional parameters)
#      object       - self
#      elapsed time - Elapsed time to convert
#      expanded     - Indicates whether to expand the "days", "hours",
#                     "minutes", "seconds", or just use the first letter for
#                     each
#
#   Return value:
#      Converted time
#
################################################################################

sub _normalize_time {

    $Log->enter();

    my $self        = shift;
    my $elapsed     = shift;
    my $expanded    = shift;
    my $hrs_per_day = (3600 * 24);

    my $days  = int($elapsed / $hrs_per_day);
    my $hours = int(($elapsed % $hrs_per_day) / 3600);
    my $min   = int(($elapsed % 3600) / 60);
    my $sec   = int($elapsed % 60);

    my $string = $days ? $days . " d ": "";
    $string   .= ($hours || ($days && !$hours)) ?
                 $hours . (($hours == 1) ? " hr " : " hrs ") :
                 "";
    $string   .= ($min || $days || $hours) ?
                 $min . (($min == 1) ? " min " : " mins ") :
                 "";
    $string   .= $sec ?
                 $sec . (($sec == 1) ? " sec " : " secs") :
                 "";
    if ($expanded) {
        ($days == 1) ?
            $string =~ s/(\d)d /$1 day / :
            $string =~ s/(\d)d /$1 days /;
        ($min == 1) ?
            $string =~ s/(\d)hrs /$1 hour / :
            $string =~ s/(\d)hrs /$1 hours /;
        ($hours == 1) ?
            $string =~ s/(\d)mins /$1 minute / :
            $string =~ s/(\d)mins /$1 minutes /;
        ($sec == 1) ?
            $string =~ s/(\d)secs/$1 second/ :
            $string =~ s/(\d)secs/$1 seconds/;
    }

    $Log->exit();

    return $string;

} ## end sub _normalize_time

################################################################################
#
# Subroutine:   _get_job_stats()
#
# Objective:    Obtains various statistics from the completed job and stores
#               them in the completed hash.
#
# Details:
#
#   Input:
#      job_stat     - Hash containing the job statistics
#
#   Return value:
#       completed   - Reference to a hash containing the statistics for the
#                     specified completed job
#
#################################################################################

sub _get_job_stats {

    $Log->enter();

    my $self   = shift;
    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    job_stat    => {type => HASHREF,
                                                   },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _get_job_stats: $@";
        $Log->trace($ex_text);
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    my ($start_time, $end_time);

    my $job_stat  = $params->{'job_stat'};
    my %completed = ();
    my $sys_time;

    #----------------------------------
    # Get the pertinent job statistics
    #----------------------------------
    %completed = %{$job_stat};

    #----------------------------------------------
    # Remove quotation marks in the various values
    #----------------------------------------------
    foreach my $key (keys %completed) {

        $completed{$key} =~ s/\"//g;
    }

    #---------------------------------------------------------------
    # If the job is still running, get the current time on the node
    #---------------------------------------------------------------
    if ($completed{'endtime'} =~ /-/) {
        my $job_node = $job_stat->{'node'};
        my $cmd      = "ngsh -c system date show -node $job_node";

        my $output = $self->{'server'}->shell(-Timeout => 300,
                                              cmd      => $cmd);


        my @lines = split /\n/, $output;

        #---------------------------------------------------
        # Find the line that contains the current date/time
        #---------------------------------------------------
        foreach my $line (@lines) {
            if ($line =~ /Date And Time: /) {
                $line =~ s/Date And Time: //;
                $sys_time = $line;
                last;
            }
        }

        if (!$sys_time) {
            $Log->trace("Could not obtain system time from node $job_node\n" .
                        "Using system time to calculated elapsed time");
            $sys_time = time();
        }
    }

    #------------------------
    # Calculate elapsed time
    #------------------------
    $start_time               = str2time($completed{'starttime'});
    $end_time                 = ($completed{'endtime'} =~ /-/)
                                ? str2time($sys_time)
                                : str2time($completed{'endtime'});

    $completed{'elapsed'}     = $end_time - $start_time;

    $Log->exit();

    return \%completed;

} ## end sub get_job_stats

################################################################################
#
# Subroutine:   _update_job_status
#
# Objective:    Checks the job_status for an existing active/deferred job
#               (based on the specified job type) and checks to see if it is
#               still running.  If it is no longer running, the statistics
#               are updated in the job status and the active/deferred job
#               key is removed from the job status.
#
# Details:
#
#   Input:
#      volume  - Volume to check for an active job (Required)
#      vserver - Vserver on which the volume resides (Required)
#
#   Return value:
#      None 
#
################################################################################

sub _update_job_status {

    $Log->enter();

    my $self = shift;

    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    volume  => {type     => SCALAR,
                                               },
                                    vserver => {type     => SCALAR,
                                               },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _update_job_status: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    my $status = "";
    my $job_id;

    #--------------------------------------------------------------------
    # If there are any active jobs recorded for this volume (active or
    # deferred), get the job ID for the active job and update the status
    #--------------------------------------------------------------------
    if ($self->{'caller'} =~ /Move/i && (exists $self->{'job_status'}->{$params->{'vserver'}}->{$params->{'volume'}}->{'deferred_job'})) {
        $job_id = $self->{'job_status'}->{$params->{'vserver'}}->{$params->{'volume'}}->{'deferred_job'}->{'id'};
    } elsif (exists $self->{'job_status'}->{$params->{'vserver'}}->{$params->{'volume'}}->{'active_job'}) {
        $job_id = $self->{'job_status'}->{$params->{'vserver'}}->{$params->{'volume'}}->{'active_job'}->{'id'};
    }

    my $job_stats = ();

    #------------------------------------
    # If there is an active job recorded
    #------------------------------------
    if ($job_id) {
        #------------------------------------
        # Get the current status of this job
        #------------------------------------
        $job_stats = $self->_get_job_status(volume  => $params->{'volume'},
                                            vserver => $params->{'vserver'},
                                            job_id  => $job_id,
                                           );

        $self->_update_job_stats(volume     => $params->{'volume'},
                                 vserver    => $params->{'vserver'},
                                 job_status => $job_stats,
                                );
    }

    $Log->exit();

    return $job_stats;

} ## end sub _update_job_status

################################################################################
#
# Subroutine:   _update_job_stats
#
# Objective:    Updates the job information in the job_status hash for the
#               specified volume.
#
# Details:
#
#   Input:
#      volume     - Volume to check for an active job (Required)
#      vserver    - Vserver on which the volume resides (Required)
#      job_status - Reference to a hash containing the job status (Required)
#      vol_obj    - Pointer to a NOTE volume object (Optional)
#
#   Return value:
#      Pointer to a hash containing the job statistics.  Hash will include
#      a key for 'current_status', which indicates if the job is 'complete',
#      'active', or 'deferred'.
#
################################################################################

sub _update_job_stats {

    $Log->enter();

    my $self = shift;

    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    volume     => {type => SCALAR,
                                                  },
                                    vserver    => {type => SCALAR,
                                                  },
                                    job_status => {type => HASHREF,
                                                  },
                                    vol_obj    => {type => OBJECT,
                                                   optional => 1,
                                                  },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _update_job_stats: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    my $volume  = $params->{'volume'};
    my $vserver = $params->{'vserver'};

    my $vol = $self->{'vol_objs'}->{$volume}->{$vserver};

    if (exists ($params->{'vol_obj'})) {
        $vol = $params->{'vol_obj'};
    }

    my $job_id;
    my $status = "";
    my $job_stats = ();

    #---------------------------------
    # If job information was returned
    #---------------------------------
    if (keys %{$params->{'job_status'}}) {

        $job_stats = $self->_get_job_stats(
                                           job_stat  => $params->{'job_status'},
                                          );

        #-----------------------------
        # If job is no longer running
        #-----------------------------
        if ($job_stats->{'state'} !~ /running/i) {

            $status = 'complete';

            #----------------------------------------------
            # Call job_wait to clear out job_id and status
            #----------------------------------------------
            eval {
                $vol->job_wait(interval => 1);
            };

            if ($@) {
                # Don't do anything here -- just called job_wait to clear out
                # job ID and status
            }

            #-------------------------------------------------------------------
            # Add this job to the completed jobs list if it's not already there
            #-------------------------------------------------------------------
            my $num_jobs = 0;
            if (exists $self->{'job_status'}->{$vserver}->{$volume}->{'jobs'}) {
                $num_jobs = scalar @{$self->{'job_status'}->{$vserver}->{$volume}->{'jobs'}};
            }

            #-----------------------------------------------------------------
            # If there are completed jobs recorded for this volume, make sure
            # the status for this job isn't already recorded
            #-----------------------------------------------------------------
            if ($num_jobs) {
                if ($self->{'job_status'}->{$vserver}->{$volume}->{'jobs'}->[$num_jobs-1]->{'id'} !=
                    $params->{'job_status'}->{'id'}) {
                    $Log->trace("Job $params->{'job_status'}->'id'} does not " .
                                "already have completion information.  " .
                                "Adding job status to jobs for $volume ($vserver)");
                    push @{$self->{'job_status'}->{$vserver}->{$volume}->{'jobs'}},
                           $params->{'job_status'};
                } else {
                    $Log->trace("Job information already exists for job " .
                                "$params->{'job_status'}->{'id'} so it won't " .
                                "be added");
                }
            } else {
                $Log->trace("Adding job status to jobs for $volume ($vserver)");
                push @{$self->{'job_status'}->{$vserver}->{$volume}->{'jobs'}},
                       $job_stats;
            }

            #----------------------------------------------
            # Remove this job from the list of active jobs
            #----------------------------------------------
            delete $self->{'job_status'}->{$vserver}->{$volume}->{'active_job'};

            #--------------------------------------------
            # Remove the deferred job ID for this volume
            #--------------------------------------------
            if ($self->{'caller'} =~ /Move/i) {
                if (exists $self->{'job_status'}->{$vserver}->{$volume}->{'deferred_job'}) {
                    delete $self->{'job_status'}->{$vserver}->{$volume}->{'deferred_job'};
                }

                if (exists $self->{'job_status'}->{$vserver}->{$volume}->{'num_deferred'}) {
                    $self->{'job_status'}->{$vserver}->{$volume}->{'num_deferred'}--;
                }
            }

            #------------------------------------------------
            # Save the destination aggregate for this volume
            #------------------------------------------------
            $self->{'job_status'}->{$vserver}->{$volume}->{'dest_aggr'} = $job_stats->{'dest_aggr'};

            #----------------------------------------------------
            # Save the destination volume if this is volume copy
            #----------------------------------------------------
            if ($self->{'caller'} =~ /Copy/i) {
                $self->{'job_status'}->{$vserver}->{$volume}->{'curr_dest_volume'} = $job_stats->{'curr_dest_volume'};
            }

            #------------------
            # If job succeeded
            #------------------
            if ($job_stats->{'code'} == 0) {
                $Log->comment("Job $job_stats->{'id'} completed successfully " .
                              "for volume $volume (vserver $vserver)");
                $self->{'job_status'}->{$vserver}->{$volume}->{'last_status'} = 'pass';
            } else {
                $self->{'job_status'}->{$vserver}->{$volume}->{'last_status'} = 'fail';
            }
        } else {

            my $deferred = 0;

            $status = 'active';

            #------------------------------------------
            # Check to see if the job is hard deferred
            #------------------------------------------
            if ($self->{'caller'} =~ /Move/i) {

                $deferred = $self->is_deferred(volume   => $volume,
                                               vserver  => $vserver,
                                               job      => $job_stats,
                                              );

                #-------------------------
                # If job is hard deferred
                #-------------------------
                if ($deferred == $self->HARD_DEFERRED) {
                    $Log->trace("Job $job_stats->{'id'} is hard deferred");
                    #---------------------------------------------------
                    # Updated the deferred job information for this job
                    #---------------------------------------------------
                    if (! exists ($self->{'job_status'}->{$vserver}->{$volume}->{'deferred_job'})) {
                        $self->{'job_status'}->{$vserver}->{$volume}->{'deferred_job'} = $job_stats;
                        $self->{'job_status'}->{$vserver}->{$volume}->{'deferrals'}->{$job_stats->{'id'}}->{'deferral_cnt'}++;
                        $status = 'deferred';
                    }
                } elsif ($deferred == $self->SOFT_DEFERRED)  {
                    $Log->trace("Job $job_stats->{'id'} is soft deferred");
                    $self->{'job_status'}->{$vserver}->{$volume}->{'deferrals'}->{$job_stats->{'id'}}->{'was_soft_deferred'} = 1;
                }
            }

            #-----------------------------
            # If job is not hard deferred
            #-----------------------------
            if (! $deferred) {
                $self->{'job_status'}->{$vserver}->{$volume}->{'active_job'} = $job_stats;
            }
        }
    } else {
        $Log->comment("Unable to obtain job status for job " .
                      "$params->{'job_status'}->{'id'}");
    }

    $job_stats->{'current_status'} = $status;

    $Log->exit();

    return $job_stats;

} ## end sub _update_job_stats

################################################################################
#
# Subroutine:   _prep_vols()
#
# Objective:    Attaches the volumes if not already attached
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Array containing volume information for each volume, including the
#      selected destination aggregate.
#
################################################################################

sub _prep_vols {

    $Log->enter();

    my $self = shift;

    my ($dest_aggr, $vol);

    my @vols = ();

    #-----------------------
    # Attach to each volume
    #-----------------------
    foreach my $vol_info (@{$self->{'src_vols'}}) {

        $vol = $self->_get_vol_obj(volume  => $vol_info->{'volume'},
                                   vserver => $vol_info->{'vserver'}
                                  );

        #-------------------
        # If striped volume
        #-------------------
        if ($vol->type ne 'flex') {
            my $ex_text = "Striped volumes are not supported yet";
            my $status  = $self->_gen_failure(ex_obj => new VolOps::StripingNotSupported($ex_text), 
                                             );

        }

        push (@vols, $vol_info);

    }

    $Log->exit();

    return @vols;

} ## end sub _prep_vols

################################################################################
#
# Subroutine:   _get_vol_obj()
#
# Objective:    Get the volume object by either attaching to it or obtaining
#               it from the move/copy object.
#
# Details:
#
#   Input:
#      volume  - Volume for which to obtain a volume object.
#      vserver - Vserver associated with the specified volume.
#
#   Return value:
#      Reference to the volume object.
#
################################################################################

sub _get_vol_obj {

    $Log->enter();

    my $self = shift;

    my ($params, $vol);

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    volume  => {type     => SCALAR,
                                               },
                                    vserver => {type     => SCALAR,
                                               },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _get_vol_obj: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    $Log->trace("_get_vol_obj called with the following params:" . Dumper
                $params);

    #---------------------------------------
    # If this volume isn't already attached
    #---------------------------------------
    if (! exists $self->{'vol_objs'}->{$params->{'volume'}}->{$params->{'vserver'}}) {
        $Log->trace("Attaching to volume $params->{'volume'}");

        $Log->debug("Attaching to volume using following parameters: " .
                    "\tVolume: $params->{'volume'}" .
                    "\tVserver: $params->{'vserver'}"
                   );
        eval {


            $vol = NOTE::Volume->attach(name    => $params->{'volume'},
                                        Vserver => $params->{'vserver'},
                                        Node    => $self->{'node'},
                                       );
        };

        if ($@) {
            $Log->trace("Failed to attach to volume");

            my $ex_text = "Failed to attach to volume $params->{'volume'} ($params->{'vserver'}): $@";
            my $status = $self->_gen_failure(ex_obj  => new VolOps::SrcVolumeFailure($ex_text), 
                                             volume  => $params->{'volume'},
                                             vserver => $params->{'vserver'},
                                             fatal   => 1,
                                            );

        } elsif (!$vol) {
            $Log->trace("Unable to attach to volume");
            my $ex_text = "Volume object for volume $params->{'volume'} not " .
                          "returned from NOTE.  Make sure the volume exists";
            my $status = $self->_gen_failure(ex_obj  => new VolOps::SrcVolumeFailure($ex_text),
                                             volume  => $params->{'volume'},
                                             vserver => $params->{'vserver'},
                                             fatal   => 1,
                                            );
        } else {

            $Log->trace("Saving volume object for $params->{'volume'}");

            $self->{'vol_objs'}->{$params->{'volume'}}->{$params->{'vserver'}} = $vol;
        }
    } else {

        $Log->trace("Obtaining the existing volume object from self");

        #-----------------------------------
        # Obtain the existing volume object
        #-----------------------------------
        $vol = $self->{'vol_objs'}->{$params->{'volume'}}->{$params->{'vserver'}};
    }

    $Log->exit();

    return $vol;

} ## end sub _get_vol_obj

################################################################################
#
# Subroutine:   _select_aggr()
#
# Objective:    Determines the destination aggregate for the volume operation
#               based on the block type and location affinity.
#
# Details:
#
#   Input:
#      vol_info
#
#   Return value:
#      aggregate - HASHREF
#
################################################################################

sub _select_aggr {

    $Log->enter();

    my $self = shift();

    my ($aggr, $aggr_obj, $aggr_props, $nodes, $params, @sel_aggrs, $vol_aggr,
        $vol_node, $vol_obj);

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    vol_info => {type     => HASHREF,
                                                },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _select_aggr: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    #---------------------------------------------------
    # Obtain the volume object for the specified volume
    #---------------------------------------------------
    if (exists $self->{'vol_objs'}->{$params->{'vol_info'}->{'volume'}}->{$params->{'vol_info'}->{'vserver'}}) {
        $vol_obj = $self->{'vol_objs'}->{$params->{'vol_info'}->{'volume'}}->{$params->{'vol_info'}->{'vserver'}};
    } else {
        my $ex_text = "Volume object doesn't exist for $params->{'vol_info'}->{'volume'} ($params->{'vol_info'}->{'vserver'})";
        my $status = $self->_gen_failure(ex_obj  => new VolOps::InternalError($ex_text),
                                         volume  => $params->{'vol_info'}->{'volume'},
                                         vserver => $params->{'vol_info'}->{'vserver'},
                                        );
    }

    my $loc_affin  = $params->{'vol_info'}->{'aggr_loc_affinity'};

    #------------------------------------------------------------------------
    # If location affinity is alternate, check to see what the last affinity
    # was and set it to the opposite
    #------------------------------------------------------------------------
    if ($loc_affin eq 'alternate') {
        if (exists $vol_obj->{'last_aggr'}) {
            $Log->debug("loc_affin for the last operation was " .
                        "$vol_obj->{'last_aggr'}");
            $loc_affin = ($vol_obj->{'last_aggr'} eq 'remote') ?
                         'local' :
                         'remote';
            $vol_obj->{'last_aggr'} = $loc_affin;
            $Log->debug("Setting last_aggr for $params->{'vol_info'}->{'volume'} to " .
                        "$loc_affin");
        } else {
            $Log->debug("loc_affin was not set for the last operation.  " .
                        "Setting loc_affin to 'remote'");
            $loc_affin = $vol_obj->{'last_aggr'} = 'remote';
        }
    }

    my $type_affin = $params->{'vol_info'}->{'aggrtype_affinity'};
    my $aggr_list;

    #---------------------------------------
    # If a list of aggregates was specified
    #---------------------------------------
    if (defined $params->{'vol_info'}->{'aggr_list'}) {
        $aggr_list  = $params->{'vol_info'}->{'aggr_list'};
    } else {
        $aggr_list  = undef;
    }

    if ($vol_obj->type ne 'flex') {
        my $ex_text = 'Striped volumes are not supported for this operation';
        my $status = $self->_gen_failure(ex_obj  => new VolOps::StripingNotSupported($ex_text),
                                         volume  => $params->{'vol_info'}->{'volume'},
                                         vserver => $params->{'vol_info'}->{'vserver'},
                                        );
    }
    eval {
        $aggr_props = $vol_obj->Aggregate->status;
    };

    if ($@) {
        my $ex_text = "Error obtaining aggregate properties: $@";
        my $status = $self->_gen_failure(ex_obj  => new VolOps::SrcAggregateFailure($ex_text),
                                         volume  => $params->{'vol_info'}->{'volume'},
                                         vserver => $params->{'vol_info'}->{'vserver'},
                                        );
    }

    #------------------------------------------------
    # If destination aggregate can be any block type
    #------------------------------------------------
    if ($type_affin eq 'any') {
        $type_affin = '*'
    } elsif ($type_affin eq 'alternate') {

        $Log->trace("Aggregate properties: " . Dumper $aggr_props);
        my $block_type = $aggr_props->{'block-type'};

        $Log->trace("The source aggregate block type is: " . $block_type);
        my @block_types = @{$self->{'block_types'}};

        if (grep (/^$block_type$/, @block_types)) {

            #-----------------------------------------------------
            # Find the next block type in the list of block types
            #-----------------------------------------------------
            for (my $i = 0; $i <= $#block_types; $i++) {

                if ($block_types[$i] =~ /^$block_type$/) {
                    $i          = ($i == $#block_types) ? 0 : ++$i;
                    $type_affin = $block_types[$i];
                    last;
                }
            }
        } else {
            my $ex_text = "Unknown block type \'$block_type\'";
            my $status = $self->_gen_failure(ex_obj  => new VolOps::ConfigError($ex_text),
                                             volume  => $params->{'vol_info'}->{'volume'},
                                             vserver => $params->{'vol_info'}->{'vserver'},
                                            );
        }

    } else {
        $type_affin = $type_affin . '-bit';
    }

    my @aggr_array;

    #---------------------------------------------------------
    # Populate the aggregate hashes needed by pick_aggregates
    #---------------------------------------------------------
    if (defined $aggr_list && scalar @{$aggr_list}) {
        my $aggr_info = $self->{'server'}->storage_aggregate_show();

        foreach my $aggr (@{$aggr_info}) {
            if (grep /^$aggr->{'aggregate'}$/, @{$aggr_list}) {
                push (@aggr_array, $aggr);
            }
        }

        $Log->trace("The list of aggregates to pick from are:");
        foreach my $aggr (@aggr_array) {
            $Log->trace("\t" . $aggr_obj->{'aggregate'});
        }
    }

    #-----------------------------------------
    # Get a list of aggregates to select from
    #-----------------------------------------
    my %pick_aggr_params = ();

    $pick_aggr_params{'state'}                    = "online";
    $pick_aggr_params{'voltype'}                  = $vol_obj->type;
    $pick_aggr_params{'voltype_discovery_mode'}   = 3;
    my $space_guar = $vol_obj->status("space-guarantee");

    #----------------------------------------------------------------
    # If volume's space guarantee is none, use the 'used' size
    # and add a fudge factor in case something is actively writing
    # to the volume.  Otherwize use the volume's size.  NOTE returns
    # the values in KB, so we will convert the values to MB and add
    # 'mb' to ensure it is treated as megabytes by pick_aggregates
    #----------------------------------------------------------------
    if ($space_guar =~ /none/i) {
        my $fudge = 5;
        $Log->trace('Volume space guarantee is none.  Requesting an aggregate ' .
                    'with ' . ($vol_obj->size / 24) . 'kb (used size) available ' .
                    'plus ' . $fudge . '% fudge factor');
        my $multiplier = 1 + ($fudge/100);
        $pick_aggr_params{'avail_size'} = (($vol_obj->used / 1024) * $multiplier) .  'mb';
    } else {
        $pick_aggr_params{'avail_size'} = ($vol_obj->size / 1024) . 'mb';
    }

    $Log->comment("Searching for an aggregate with $pick_aggr_params{'avail_size'} available");

    $pick_aggr_params{'block_type'}               = $type_affin;
    $pick_aggr_params{'blocktype_discovery_mode'} = 3; # Use only the value specified by the block_type param

    #------------------------------------
    # If an aggregate list was passed in
    #------------------------------------
    if (defined $aggr_list) {
        if (scalar @{$aggr_list}) {
            $pick_aggr_params{'aggregate_list'}     = \@aggr_array;
        }
    }

    $Log->comment("Calling pick_aggregates with the following parameters: " .
                  Dumper \%pick_aggr_params);

    my @aggrs = ();

    eval {
        #----------------------------------------------
        # Rediscover so aggregate data will be current
        #----------------------------------------------
        $self->{'server'}->discover_topology(cmd_nodes=>["aggregate"]);
        @aggrs  = $self->{'server'}->pick_aggregates(%pick_aggr_params
                                                    );
    };

    if ($@) {
        my $ex_text = "Call to pick_aggregates failed with $@";
        my $status = $self->_gen_failure(ex_obj  => new VolOps::AggregateSelectionFailure($ex_text),
                                         volume  => $params->{'vol_info'}->{'volume'},
                                         vserver => $params->{'vol_info'}->{'vserver'},
                                        );
    }

    $Log->trace("The list of picked aggregates is " . Dumper \@aggrs);

    @sel_aggrs = ();
    $vol_node  = $vol_obj->Node->name;
    $vol_aggr  = $vol_obj->Aggregate->name;
    $Log->trace("Volume:       " . $vol_obj->fullname);
    $Log->trace("Loc affinity: $loc_affin");
    $Log->trace("Volume node:  $vol_node");
    $Log->trace("Volume aggr:  $vol_aggr");

    #--------------------------------------------------------------------
    # Narrow down the list of aggrs to select from based on loc affinity
    #--------------------------------------------------------------------
    foreach my $aggr_info (@aggrs) {

        $Log->trace("Processing aggregate " . $aggr_info->{'aggregate'} .
                    " to determine if it meets affinity specifications");

        my $same_aggr = ($aggr_info->{'aggregate'} =~ /^$vol_aggr$/) ? 1 : 0;

        #--------------------------------------------------------------------
        # If local loc_affin and aggregate is on the same node as the volume
        #--------------------------------------------------------------------
        if (($loc_affin eq 'local') &&
            ($aggr_info->{'nodes'} =~ $vol_node)) {

            if (($self->{'caller'} !~ /Move/i) ||
                ($self->{'caller'} =~ /Move/i && !($same_aggr)))  {
                $Log->trace("aggr_loc_affinity is set to 'local'.  Adding " .
                            "local $aggr_info->{'aggregate'} to the list");
                push (@sel_aggrs, $aggr_info);
            } else {
                $Log->trace("aggr_loc_affinity is set to 'local' but " .
                            "skipping aggregate $aggr_info->{'aggregate'} " .
                            "because it is the volume's aggregate");
            }
        #----------------------------------------------------------
        # If remote loc_affin and aggregate is on a different node
        #----------------------------------------------------------
        } elsif (($loc_affin eq 'remote') &&
                ($aggr_info->{'nodes'} !~ $vol_node)) {
            $Log->trace("aggr_loc_affinity is set to 'remote'.  Adding " .
                        "remote $aggr_info->{'aggregate'} to the list");
            push (@sel_aggrs, $aggr_info);
        #------------------
        # If any loc_affin
        #------------------
        } elsif ($loc_affin eq 'any') {
            if (($self->{'caller'} !~ /Move/i) ||
                ($self->{'caller'} =~ /Move/i && !($same_aggr)))  {
                $Log->trace("aggr_loc_affinity is set to 'any'.  " .
                            "Adding $aggr_info->{'aggregate'} to the list to " .
                            "select from");
                push (@sel_aggrs, $aggr_info);
            } else {
                $Log->trace("aggr_loc_affinity is set to 'any' but skipping " .
                            "aggregate $aggr_info->{'aggregate'} because " .
                            "it is the volume's aggregate");
            }
        #---------------------
        # Skip this aggregate
        #---------------------
        } else {
            $Log->trace("Skipping aggregate $aggr_info->{'aggregate'} " .
                        "because it doesn't meet the criteria");
        }
    }

    #-----------------------------------------------------
    # Randomly select an aggregate from the list returned
    #-----------------------------------------------------
    if (scalar(@sel_aggrs)) {
        $aggr = $sel_aggrs[rand @sel_aggrs];
    } else {
        my $ex_text = "Unable to find a destination aggregate that meets " .
                      "the following specifications using loc affinity of " .
                      "$loc_affin and aggregate type affinity of " .
                      "$type_affin:\n" .
                      Dumper \%pick_aggr_params;
        my $status  = $self->_gen_failure(ex_obj  => new VolOps::AggregateSelectionFailure($ex_text),
                                          volume  => $params->{'vol_info'}->{'volume'},
                                          vserver => $params->{'vol_info'}->{'vserver'},
                                         );
    }

    $Log->trace("Selected $aggr->{'aggregate'} as the destination aggregate");

    $Log->trace("Attaching to aggregate " . $aggr->{'aggregate'} . " using " .
                $aggr->{'nodes'});
    eval {
        $aggr_obj = NOTE::Aggregate->attach(Node  => $aggr->{'nodes'},
                                            name  => $aggr->{'aggregate'},
                                           );
        $Log->trace("Attached to aggregate " . $aggr_obj->name);
    };

    if ($@) {
        my $ex_text = "Failed to attach to destination aggregate: $@";
        my $status  = $self->_gen_failure(ex_obj  => new VolOps::DestAggregateFailure($ex_text),
                                          volume  => $params->{'vol_info'}->{'volume'},
                                          vserver => $params->{'vol_info'}->{'vserver'},
                                         );
    };

    $Log->exit();

    return $aggr_obj;

} ## end sub _select_aggr

################################################################################
#
# Subroutine:   _perform_vol_ops()
#
# Objective:    Move/Copy the list of specified volumes to their selected
#               destination aggregates.  Operations will be performed in
#               parallel if indicated.
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      None
#
################################################################################

sub _perform_vol_ops {

    $Log->enter();

    my $self = shift @_;

    my $det_filt = "VolOps/" . $self->{'caller'} . "/Detailed";
    my $sum_filt = "VolOps/" . $self->{'caller'} . "/Summary";

    listener_add({filter   => sub { $_[0]->{'type'} eq $det_filt },
                  callback => sub { $self->_get_status("Detailed"); },
                 }
                );

    listener_add({filter   => sub {$_[0]->{'type'} eq $sum_filt },
                  callback => sub { $self->_get_status("Summary"); },
                 }
                );
    #---------------------------------------
    # Iterate through the volume operations
    #---------------------------------------
    $self->{'start_time'} = time();

    #-------------------------------------------------------------------------
    # Get volume properties so comparison between iterations can be performed
    #-------------------------------------------------------------------------
    $self->{'last_vol_props'} = $self->_get_vol_properties();

    my $ex;
    my $done = 0;

    #---------------------------------------------------------------------
    # Execute the volume operation for the specified number of iterations
    #---------------------------------------------------------------------
    for (my $i = 1; ($i <= $self->{'iterations'} && !$done); $i++) {
        $self->{'iter_cnt'} = $i;

        #-----------------------------------------------------------------------
        # Determine which method to call based on the operation being performed
        #-----------------------------------------------------------------------
        if ($self->{'caller'} eq 'Move') {
            try {
                $self->_move();
            } catch VolOps::Move::MoveDeferred with {
                $Log->trace("Catching VolOps::Move::MoveDeferred exception");
                $ex = shift;
                $Log->comment("A deferred move job was found.  All remaining " .
                              "iterations will be abandoned.");
                $done = 1;
            } catch NATE::BaseException with {
                $Log->trace("Catching exception from _move");
                $ex = shift;
            }
        } elsif ($self->{'caller'} eq 'Copy') {
            $self->_copy();
        } else {
            my $ex_text = "Unknown invocant: $self->{'caller'}";
            my $status  = $self->_gen_failure(ex_obj => new VolOps::ConfigError($ex_text),
                                             );
        }
    }

    $self->{'end_time'} = time();

    my $job_status = $self->_gen_status(status_type => "Detailed");

    $Log->comment("Job status:\n" . $job_status->{'formatted'});

    #-----------------------------------------------------------------------
    # Note:  This is the final result object that is returned from the
    #        subtest.  The result string ("Move|Copy passed|failed...")
    #        must be unique because the get_status method searches for
    #        this string when searching for the final status of the subtest.
    #        gen_status is expecting to see the following string in the
    #        final result object:
    #        Move|Copy passed|failed - all moves|copies complete
    #-----------------------------------------------------------------------

    #--------------------------------------------
    # Set string to return if any failures occur
    #--------------------------------------------
    my $string = " - all ";
    $string   .= ($self->{'caller'} =~ /Copy/i) ? "copies" : "moves";
    $string   .= " complete";

    if (($job_status->{'total_failed'} == 0) && ($job_status->{'total_errors'} == 0)) {
        $Log->trace("All operations completed successfully.  Returning " .
                    "pass result" . Dumper $job_status);
        $Log->comment(pass_result("$self->{'caller'} passed" . $string,
                                  misc => $job_status));
    } else {
        $Log->trace("At least one operation failed.  Returning fail result " .
                    Dumper $job_status);
        if ($self->{'create_nate_result'} =~ /no/i) {
            $Log->trace("One or more failures was set but no NATE result " .
                        "will be created");
            $Log->comment("*** One or more failures occurred but " .
                          "create_nate_result was set to 'no'.  Generating " .
                          "a warning result instead of a failure");
            $Log->comment(warn_result("$self->{'caller'} failed" . $string,
                                      misc => $job_status));
        } else {
            $Log->error(fail_result("$self->{'caller'} failed" . $string,
                                    misc => $job_status));
        }
    }

    $Log->trace("Removing Detailed filter");

    listener_remove({filter   => sub { $_[0]->{'type'} eq $det_filt },
                     callback => sub { $self->_get_status(type => "Detailed"); },
                    }
                   );

    $Log->trace("Removing Summary filter");
    listener_remove({filter   => sub {$_[0]->{'type'} eq $sum_filt },
                     callback => sub { $self->_get_status(type => "Summary"); },
                    }
                   );

    #---------------------------------------------------
    # If running in the foreground, and executing 
    # volume move, throw the exception.  Volume copy 
    # exception should already be thrown because it is
    # not caught.
    #---------------------------------------------------
    if ($self->{'caller'} =~ /Move/i) {
        if ($ex && ($self->{'bg_mode'} =~ /^-nobg/)) {
            $ex->throw;
        }
    }

    $Log->exit();

} ## end sub _perform_vol_ops

=head2 get_vol_properties

C<< $vmv->get_vol_properties(%params) >>

Obtain volume properties to be used for basic comparison.  The properties are
saved in the object and are used for later comparison.  A subsequent call to
B<get_vol_properties> will overwrite any existing properties already saved in
the object.

=over 4

=item Input

=over 4

=item * volumes => <arrayref>

(Required)

List of volumes for which to obtain properties.  The array consists of hashes
which must include the following keys:
  volume
  vserver

=back

=item Output

Returns a hash ref containing the lists of properties for each specified
volume.  The properties are also saved in the object and will
be used when comparing volume properties.

=item Example

C<< my $props = get_vol_properties(volumes => \@vol_list); >>

=back

=cut

################################################################################
#
# Subroutine:   get_vol_properties()
#
# Objective:    Obtain the list of volume properties for the list of volumes.
#               Properties can be used for basic comparison of
#               source/destination volumes after a volume move has completed.
#               The properties are also stored in the object using the
#               'vol_props' key.
#
# Details:
#
#   Input:
#       volumes - List of volumes for which to obtain properties.  The list
#                 is an array of hashes whose keys must include the following:
#                   - volume
#                   - vserver
#                 Optional; defaults to the list of volumes that were
#                 specified when the object was instantiated.
#
#   Return value:
#       Returns the properties for all of the specified volumes
#
################################################################################

sub get_vol_properties {

    $Log->enter();

    my $self = shift();

    my $params = {
                     validate (
                               @_,
                               {
                                   volumes        => {type     => ARRAYREF,
                                                      optional => 1,
                                                      default  => $self->{'src_vols'},
                                                     },
                               },
                              )
                 };

    $self->{'vol_props'} = $self->_get_vol_properties(volumes => $params->{'volumes'});

    $Log->exit();

    return $self->{'vol_props'};

} ## end sub get_vol_properties

################################################################################
#
# Subroutine:   _get_vol_properties()
#
# Objective:    Obtain the list of volume properties for the list of volumes.
#               Properties can be used for basic comparison of
#               source/destination volumes after a volume move has completed.
#
# Details:
#
#   Input:
#       None
#
#   Return value:
#       Returns 1 on success, throws exception on failure
#
################################################################################

sub _get_vol_properties {

    $Log->enter();

    my $self = shift;

    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    volumes  => {type     => ARRAYREF,
                                                 optional => 1,
                                                 default  => $self->{'src_vols'},
                                                },
                                    get_dest => {type     => BOOLEAN,
                                                 optional => 1,
                                                 default  => 0,
                                                },
                                },
                               )
                 };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _get_vol_properties: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    my ($vol_name, $vol_obj, %vol_props, $vserver);

    #------------------------------------------------------------------------
    # Get the volume properties for each of the volumes stored in the object
    #------------------------------------------------------------------------
    foreach my $vol_data (@{$params->{'volumes'}}) {

        #------------------------------------------------------------------
        # Determine the volume name based on whether or not we are getting
        # the attributes for the destination volume
        #------------------------------------------------------------------
        if (!$params->{'get_dest'}) {
            $vol_name = $vol_data->{'volume'};
        } else {
            $vol_name = $vol_data->{'curr_dest_volume'};
        }

        $vserver  = $vol_data->{'vserver'};

        #--------------------------------------------
        # Get the object associated with this volume
        #--------------------------------------------
        if (exists $self->{'vol_objs'}->{$vol_name}->{$vserver}) {

            $vol_obj = $self->{'vol_objs'}->{$vol_name}->{$vserver};

        } else {

            $vol_obj = $self->_get_vol_obj(volume  => $vol_name,
                                           vserver => $vserver,
                                          );
        }

        #------------------------------------------------------------------
        # Make sure a selection for comparing attributes was made for this
        # volume
        #------------------------------------------------------------------
        eval {
            $vol_props{$vol_name}{$vserver} = $vol_obj->status();
        };

        if ($@) {
            my $ex_text = "Unable to obtain volume properties for volume " .
                          "$vol_name: $@";
            my $status  = $self->_gen_failure(ex_obj  => new VolOps::PropsNotAvailable($ex_text),
                                              volume  => $vol_name,
                                              vserver => $vserver,
                                              fatal   => ($self->{'on_fail'} =~ /continue/i) 
                                                         ? 0 : 1, 
                                             );
        }
    }

    $Log->trace("The content of the volumes properties hash are: " .
                Dumper \%vol_props);

    $Log->exit();

    return \%vol_props;

} ## end sub _get_vol_properties

=head2 compare_vol_properties

C<< $vol_obj->compare_vol_properties(%params) >>

Perform a basic comparison of volume properties after the operation has completed.
Predefined lists are set up when the object is instantiated.  These lists
include those attributes whose validation rules are common between move and
copy.

=over 4

=item Input

=over 4

=item * volumes => <arrayref>

(Optional)
Default = List of volumes supplied when the object was instantiated.

List of volumes to compare.  Array consists of hashes which must include the
following keys for each volume to compare:

 volume
 vserver

=item * match_fields => <arrayref>

(Optional)
Default = Predefined list of matching attributes.

List of volume attributes whose pre- and post-operation values must match.

=item * diff_fields => <arrayref>

(Optional)
Default = Predefined list of attributes whose values must be different.

List of volume attributes whose pre- and post-operation values must be different.

=item * skip_fields => <arrayref>

(Optional)
Default = Predefined list of attributes to skip.

List of volume attributes whose pre- and post-operation values should be skipped
when comparing because it cannot be determined whether or not they will match.

=item * match_val => <hashref>

(Optional)
Default = Predefined hash of attributes that must be set to a specific value.

List of volume attributes whose values must be set to a specific value after a
move.

The format of the hash is:

 (<attribute> => "<match value>",
  ...
 );

=item * default_action => <scalar>

Action to take when an attribute isn't listed in any of the validation lists.

Valid values:

  match - Verify all attributes for which no validation rules were specified
          match.
  diff  - Verify all attributes for which no validation rules were specified
          contain different values.
  skip  - Skip all attributes for which no validation rules were specified.
  die   - Throw an exception when an attribute is encountered for which no
          validation rules were specified.

=back

=item Output

Returns 1 if all volumes compare; returns 0 if there were any miscomparisons.

=item Example

C<< $vmv->compare_vol_properties(volumes => \@vol_list); >>

=back

=cut

################################################################################
#
# Subroutine:   compare_vol_properties()
#
# Objective:    Compares a list of volume properties after a volume operation
#               has completed.
#
# Details:
#
#   Input:
#      volumes        - Array of volumes to use when comparing volumes
#                       (Optional - defaults to all volumes that were moved)
#      match_fields   - List of attributes to ensure match when comparing volume
#                       attributes (Optional)
#      diff_fields    - List of attributes to ensure are different when
#                       comparing attributes (Optional)
#      match_val      - Hash of attributes that should be set to a specific
#                       value when comparing attributes (Optional)
#      skip_fields    - List of attributes to ignore when comparing volume
#                       attributes (Optional)
#      default_action - Action to take when an attribute isn't listed in any
#                       of the validation lists.  Possible values can be
#                       'match', 'diff', 'skip', 'die' (default = 'match').
#                       (Optional)
#
#   Return value:
#      Returns 1 if volume properties compare; throws an exception if there  
#      were any miscomparisons.
#
################################################################################

sub compare_vol_properties {

    my $self = shift();

    $Log->enter();

    my ($vol_obj, %vol_props);

    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    volumes        => {type     => ARRAYREF,
                                                       optional => 1,
                                                       default  => $self->{'src_vols'},
                                                      },
                                    default_action => {type      => SCALAR,
                                                       optional  => 1,
                                                       callbacks => {'def_action' =>
                                                                     sub { shift() =~ /^(skip|match|diff|die)$/ }
                                                                    },
                                                       default   => 'match',
                                                      },
                                    skip_fields    => {type     => ARRAYREF,
                                                       optional => 1,
                                                       default  => ('size','available','total',
                                                                    'used','percent-used','files',
                                                                    'files-used','snapshot-space-used',
                                                                    'move-target','name-ordinal',
                                                                    'junctionFileHandle'),
                                                      },
                                    diff_fields    => {type     => ARRAYREF,
                                                       optional => 1,
                                                       default  => ('dsid','uuid'),
                                                      },
                                    match_fields   => {type     => ARRAYREF,
                                                       optional => 1,
                                                       default  => (),
                                                      },
                                    match_val      => {type      => HASHREF,
                                                       optional  => 1,
                                                       default   => ('clone-volume' => 'false'),
                                                      },
                                }
                               )
                  };

    };

    if ($@) {
        my $ex_text = "Parameter validation failed for compare_vol_properties: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    my @vols = ();
    my $vol_params;
    my $num_vols = scalar(@{$params->{'volumes'}});

    for (my $i = 1; $i <= $num_vols; $i++) {

        $Log->trace("Processing volume comparison params for volume $i of " .
                    "$num_vols");

        my @vol_params  = shift(@{$params->{'volumes'}});

        eval {
            $vol_params = {
                           validate_with (
                                          params      => \@vol_params,
                                          spec        => {
                                                          vserver          => {type      => SCALAR,
                                                                              },
                                                          volume           => {type      => SCALAR
                                                                              },
                                                          curr_dest_volume => {type      => SCALAR,
                                                                               optional  => ($self->{'caller'} =~ /Copy/i)
                                                                                             ? 0
                                                                                             : 1,
                                                                              },
                                                          compare_attrs    => {type      => BOOLEAN,
                                                                               optional  => 1,
                                                                               callbacks => {'0_or_1' =>
                                                                                             sub { shift() =~ /^(0|1)$/ }
                                                                                            },
                                                                               default   => 1,
                                                                              },
                                                          cascade_copies   => {type      => SCALAR,
                                                                               optional  => 1,
                                                                               default   => 0,
                                                                              },
                                                        },
                                          allow_extra => 1,
                                         )
                          };

            $Log->debug("Volume params for volume $i of " .  scalar($params->{'volume'}));

        };

        if ($@) {
            my $ex_text = "Parameter validation failed for the volumes parameter to compare_vol_properties: $@";

            my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                             );
        }

        push (@vols, $vol_params);
    }

    my $errors_found = 0;
    my ($dest_props, $src_props, $src_prop_key);

    $Log->debug("Match fields " . Dumper $params->{'match_fields'});
    $Log->debug("Diff fields " . Dumper $params->{'diff_fields'});
    $Log->debug("Skip fields " . Dumper $params->{'skip_fields'});
    $Log->debug("Match values " . Dumper $params->{'match_val'});

    #-----------------------------------------------------------------
    # Determine which key contains the source properties that we want
    #-----------------------------------------------------------------
    $src_prop_key = 'vol_props';

    #-------------------------------------------------------
    # If comparison is being performed after each iteration
    #-------------------------------------------------------
    if (exists $self->{'internal_compare'}) {
        if ($self->{'internal_compare'}) {
            $src_prop_key = 'last_vol_props';
        }
    }

    $Log->trace("Using \'$src_prop_key\' for source properties key");
    #-------------------------------------
    # Get the list of original attributes
    #-------------------------------------
    $src_props = $self->{$src_prop_key};
    $Log->debug("Source properties: " . Dumper $src_props);

    #------------------------------------
    # Get the list of current attributes
    #------------------------------------
    $dest_props = $self->_get_vol_properties(volumes  => \@vols,
                                             get_dest => $self->{'caller'} =~ /Copy/ ? 1 : 0,
                                            );
    $Log->debug("Destination properties: " . Dumper $dest_props);

    #------------------------------------------------------
    # Compare attributes for each of the specified volumes
    #------------------------------------------------------
    foreach my $comp_data (@vols) {

        my $vol_name      = $comp_data->{'volume'};
        my $vserver       = $comp_data->{'vserver'};
        my $dest_vol_name = $comp_data->{'volume'};
        if (exists $comp_data->{'curr_dest_volume'}) {
            $dest_vol_name = $comp_data->{'curr_dest_volume'};
        }

        my $compare_attrs = 1;

        #-----------------------------------------------------------
        # If no volumes were passed in, check to see if this volume
        # is supposed to be compared
        #-----------------------------------------------------------
        if (exists $comp_data->{'compare_attrs'}) {
            $Log->trace("compare_attrs exists for volume $vol_name");
            if (!$comp_data->{'compare_attrs'}) {
                $compare_attrs = 0;
            }
        }

        #-----------------------------------------------
        # Get the destination aggregate for this volume
        #-----------------------------------------------
        if (exists $self->{'job_status'}->{$vserver}->{$vol_name}) {
            $params->{'match_val'}->{'aggregate'} =
                $self->{'job_status'}->{$vserver}->{$vol_name}->{'dest_aggr'};
        } else {
            $Log->warn("Destination aggregate information is not " .
                       "available.");
        }

        NOTE::logcomment("Comparing volume properties for $vol_name " .
                         "(vserver $vserver)");

        $Log->debug("Value of match_val hash: " .
                    Dumper $params->{'match_val'});

        #---------------------------------------------
        # If attributes were gathered for this volume
        #---------------------------------------------
        if (exists $src_props->{$vol_name}->{$vserver}) {

            #---------------------------
            # If the job was successful
            #---------------------------
            if ($self->{'job_status'}->{$vserver}->{$vol_name}->{'last_status'}
                =~ /pass/i) {

                if ($compare_attrs) {

                    my $vol_data  = $src_props->{$vol_name}->{$vserver};
                    my $dest_data = $dest_props->{$dest_vol_name}->{$vserver};

                    #--------------------------------------------------------
                    # Compare each of the attributes based on the validation
                    # lists
                    #--------------------------------------------------------
                    my $obj_name  = "Volume $vol_name Vserver $vserver";

                    try {
                        my $compared = $self->compare_attrs(obj_name       => $obj_name,
                                                            src_data       => $vol_data,
                                                            dest_data      => $dest_data,
                                                            default_action => $params->{'default_action'},
                                                            match_fields   => $params->{'match_fields'},
                                                            diff_fields    => $params->{'diff_fields'},
                                                            skip_fields    => $params->{'skip_fields'},
                                                            match_val      => $params->{'match_val'},
                                                           );
                    } catch VolOps::AttributeComparison with {

                        my $ex     = shift;
                        my $status = $self->_gen_failure(ex_obj  => $ex, 
                                                         vserver => $vserver,
                                                         volume  => $vol_name,
                                                         fatal   => ($self->{'on_fail'} =~ /continue/i)
                                                                    ? 0 : 1,
                                                        );

                    } catch VolOps::MissingValidationRule with {

                        my $ex     = shift;
                        my $status = $self->_gen_failure(ex_obj  => $ex, 
                                                         vserver => $vserver,
                                                         volume  => $vol_name,
                                                         fatal   => ($self->{'on_fail'} =~ /continue/i)
                                                                    ? 0 : 1,
                                                        );

                    } catch NATE::BaseException with {

                        my $ex     = shift;
                        $Log->trace("Unknown Exception occurred: $ex->text()");
                        my $status = $self->_gen_failure(ex_obj  => $ex, 
                                                         vserver => $vserver,
                                                         volume  => $vol_name,
                                                         fatal   => ($self->{'on_fail'} =~ /continue/i)
                                                                    ? 0 : 1,
                                                        );
                    };

                    #----------------------------------------------------
                    # Reset the properties to be the current destination
                    # properties
                    #----------------------------------------------------
                    if ($self->{'caller'} =~ /Move/i) {
                        $self->{$src_prop_key}->{$vol_name}->{$vserver} = ();
                        $self->{$src_prop_key}->{$vol_name}->{$vserver} =
                            $dest_props->{$vol_name}->{$vserver};
                    } else {
                        if (($self->{'caller'} =~ /Copy/i) &&
                            (exists $comp_data->{'cascade_copies'}) &&
                            ($comp_data->{'cascade_copies'})) {

                                $self->{$src_prop_key}->{$dest_vol_name}->{$vserver} =
                                    $dest_props->{$dest_vol_name}->{$vserver};

                        }
                    }

                } else {

                    $Log->comment("Skipping volume $vol_name ($vserver) because " .
                                  "compare_attrs was not set");
                }

            } else {

                $Log->warn("Skipping volume $vol_name ($vserver) because " .
                           "the last operation failed");
            }
        } else {
            my $ex_text = "Properties not available for property comparison " .
                          "for volume $vol_name ($vserver)";
            $Log->trace($ex_text);

            my $status  = $self->_gen_failure(ex_obj  => new VolOps::PropsNotAvailable($ex_text),
                                              vserver => $vserver,
                                              volume  => $vol_name,
                                              fatal   => ($self->{'on_fail'} =~ /continue/i)
                                                         ? 0 : 1,
                                             );

        }
    }

    $Log->exit();

    return ($errors_found ? 0 : 1);

} ## end sub compare_vol_properties

################################################################################
#
# Subroutine:   compare_attrs
#
# Objective:    Compare a set of attributes between two objects.
#
# Details:
#
#   Input:
#      obj_name       - Name of the object being compared (e.g., volume name,
#                       snapshot name)
#      src_data       - Hash containing the source attributes to compare
#      dest_data      - Hash containing the destination attributes to compare
#      match_fields   - List of attributes to ensure match when comparing volume
#                       attributes (Optional)
#      diff_fields    - List of attributes to ensure are different when
#                       comparing attributes (Optional)
#      match_val      - Hash of attributes that should be set to a specific
#                       value when comparing attributes (Optional)
#      skip_fields    - List of attributes to ignore when comparing volume
#                       attributes (Optional)
#      default_action - Action to take when an attribute isn't listed in any
#                       of the validation lists.  Possible values can be
#                       'match', 'diff', 'skip', 'die' (default = 'match').
#                       (Optional)
#
#   Return value:
#      Returns 1 if volume properties compare; throws an exception if
#      there were any miscompares or if default action is 'die' and there
#      were no comparison rules provided. 
#
################################################################################

sub compare_attrs {

    my $self = shift;

    $Log->enter();

    my $params;

    eval {

        $params = {
                      validate (
                                @_,
                                {
                                    obj_name       => {type      => SCALAR,
                                                       },
                                    src_data       => {type      => HASHREF,
                                                      },
                                    dest_data      => {type      => HASHREF,
                                                      },
                                    default_action => {type      => SCALAR,
                                                       optional  => 1,
                                                       callbacks => {'def_action' =>
                                                                     sub { shift() =~ /^(skip|match|diff|die)$/ }
                                                                    },
                                                       default   => 'match',
                                                      },
                                    skip_fields    => {type     => ARRAYREF,
                                                       optional => 1,
                                                       default  => (),
                                                      },
                                    diff_fields    => {type     => ARRAYREF,
                                                       optional => 1,
                                                       default  => (),
                                                      },
                                    match_fields   => {type     => ARRAYREF,
                                                       optional => 1,
                                                       default  => (),
                                                      },
                                    match_val      => {type     => HASHREF,
                                                       optional => 1,
                                                       default  => (),
                                                     },
                                }
                               )
                  };

    };

    if ($@) {
        my $ex_text = "Parameter validation failed for compare_attrs: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    #----------------------------------------------------------------------
    # Obtain a list of all the fields for which validation rules have been
    # specified
    #----------------------------------------------------------------------
    my @all_fields = (@{$params->{'skip_fields'}},
                      @{$params->{'diff_fields'}},
                      @{$params->{'match_fields'}},
                      keys %{$params->{'match_val'}},
                     );

    my @missing      = ();
    my @errors       = ();

    my $this_method  = (caller(0))[3];

    $Log->debug("Match fields " . Dumper $params->{'match_fields'});
    $Log->debug("Diff fields " . Dumper $params->{'diff_fields'});
    $Log->debug("Skip fields " . Dumper $params->{'skip_fields'});
    $Log->debug("Match values " . Dumper $params->{'match_val'});

    $Log->debug("Comparing the following attributes:" .
                "Source data: " . Dumper $params->{'src_data'},
                "Destination data: " . Dumper $params->{'dest_data'});

    #--------------------------------------------------------
    # Compare each of the attributes based on the validation
    # lists
    #--------------------------------------------------------
    foreach my $key (keys %{$params->{'src_data'}}) {
        $Log->trace("Comparing $key between source and destination");

        my $src_val  = $params->{'src_data'}->{$key};
        my $dest_val = $params->{'dest_data'}->{$key};

        #-------------------------------------------------
        # If this is an attribute that we are supposed to
        # compare
        #-------------------------------------------------
        if (!(grep /^$key$/, @{$params->{'skip_fields'}}) ||
            (!(grep /^$key$/, @all_fields) &&
             $params->{'default_action'} ne 'skip')) {

            #----------------------------------------
            # If values are supposed to be different
            # (If volume moved only once, aggregate
            #  must be different)
            #----------------------------------------
            if ((grep /^$key$/, @{$params->{'diff_fields'}}) ||
                (! (grep /^$key$/, @all_fields) &&
                 $params->{'default_action'} eq 'diff')) {

                if ("$src_val" eq "$dest_val") {
                    $Log->error("Source \'$key\' for $params->{'obj_name'} " .
                                "matches destination and is not " .
                                "supposed to\n" .
                                "Source:      $src_val\n" .
                                "Destination: $dest_val"
                               );
                    push (@errors, $key);
                } else {
                    $Log->trace("Source and destination \'$key\' " .
                                "does not match as expected for " .
                                "$params->{'obj_name'}\n" .
                                "Source:      $src_val\n" .
                                "Destination: $dest_val");
                }

            #----------------------------------------------
            # If attribute must be set to a specific value
            #----------------------------------------------
            } elsif (exists $params->{'match_val'}->{$key}) {

                if ("$dest_val" ne "$params->{'match_val'}->{$key}") {
                    $Log->error("Destination $key does not " .
                                "contain the expected value\n" .
                                "Expected value:  " .
                                 $params->{'match_val'}->{$key} .
                                 "\n" .
                                "Actual value:    $dest_val"
                               );
                    push (@errors, $key);
                } else {
                    $Log->trace("Destination \'$key\' contains " .
                                "the expected value " .
                                "\($params->{'match_val'}->{$key}\)" .
                                " for $params->{'obj_name'}");
            }

            #---------------------------------
            # If values are supposed to match
            #---------------------------------
            } elsif ((grep /^$key$/, @{$params->{'match_fields'}})
              || (!(grep /^$key$/, @all_fields) &&
                  $params->{'default_action'} eq 'match')){

                if ("$src_val" ne "$dest_val") {
                    $Log->error("Source \'$key\' for $params->{'obj_name'} " .
                                "does not match destination\n" .
                                "Source:      $src_val\n" .
                                "Destination: $dest_val"
                               );
                    push (@errors, $key);
                } else {
                    $Log->trace("Source and destination \'$key\' " .
                                "matches as expected for $params->{'obj_name'}");
                }

            #---------------------------------------------------
            # Attribute was not listed in any of the validation
            # lists and default action is 'die'
            #---------------------------------------------------
            } else {
                my $ex_text = "Attribute comparison failed:  " .
                              "Attribute $key was not listed " .
                              "in the validation params";

                push (@missing, $key);

                $Log->warn($ex_text);

            }
        } else { # Skipping this field
            $Log->trace("Attribute $key was listed as a field to " .
                        "skip for attribute comparison, so the " .
                        "field was skipped");
        }
    }

    #------------------------------------------------------------------------
    # If there were attributes for which a validation rule was not provided,
    # throw an exception
    #------------------------------------------------------------------------
    if (scalar(@missing) && ($params->{'default_action'} eq 'die')) {
        my $ex_text = "$this_method failed:  " .
                      "Validation rules were not provided for the " .
                      "following attributes: " . Dumper \@missing;
        $ex_text   .= "This may be due to a newly added volume attribute(s) " .
                      "in Ontap.\nPlease contact dl-volops-qa to get the " .
                      "appropriate validation rule(s) included for the " .
                      "new attribute(s).";

        VolOps::MissingValidationRule->throw($ex_text);
    }

    #--------------------------------------------------------------------
    # If there were any miscomparisons and exit on failure was specified
    #--------------------------------------------------------------------
    if (scalar @errors) {
        my $ex_text = "The following attributes miscompared for " 
                      . $params->{'obj_name'} . ":\n";

        foreach my $error (@errors) {
            $ex_text .= "\t\t$error\n";
        }

        if ($self->{'on_fail'} !~ /continue/i) {
            VolOps::AttributeComparison->throw($ex_text);
        } else {
            $Log->warn($ex_text);
        }
    }

    $Log->exit();

    return 1;

} ## end sub compare_attrs

################################################################################
#
# Subroutine:   _gen_failure
#
# Objective:    Add the failure to the status information and generate a fail
#               result.
#               If the error is not specific to a volume/vserver, the error
#               is saved in the 'errors' key of the status.  If the error is
#               volume/vserver specific, the error is saved in the
#               <vserver>-><volume>->'errors' key of the status.  If iteration
#               count is available, it is also added to the error information.
#
# Details:
#
#   Input:
#      ex_obj   - Exception object to throw, including text which describes
#                 the reason for the failure.
#      volume   - Volume that had the error
#                 Optional; no default specified
#      vserver  - Vserver associated with the volume that had the error
#                 Optional; no default specified
#      fatal    - Indicates whether or not this is a fatal error.  (Default = 1)
#                 If set, it will log a fail result and throw the exception.
#                 If not set, a warning will be generated.
#
#   Return value:
#       $status - Reference to the status information for this object.
#
################################################################################

sub _gen_failure {

    $Log->enter();

    my $self = shift;

    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    ex_obj  => {type     => OBJECT,
                                               },
                                    vserver => {type     => SCALAR,
                                                optional => 1,
                                               },
                                    volume  => {type     => SCALAR,
                                                optional => 1,
                                               },
                                    fatal   => {type     => BOOLEAN,
                                                optional => 1,
                                                default  => 1,
                                               },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _gen_failure: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    $self->{'total_errors'}++;

    my %error_data = ();
    my $job_status = ();

    $error_data{'reason'} = $params->{'ex_obj'}->text();

    #------------------------------------------------------------------
    # If iteration count is available, add it to the error information
    #------------------------------------------------------------------
    if (exists $self->{'iter_cnt'}) {
        $error_data{'iteration'} = $self->{'iter_cnt'};
    }

    #------------------------------------
    # If this is a volume-specific error
    #------------------------------------
    if (exists $params->{'volume'} && exists $params->{'vserver'}) {
        push (@{$self->{'job_status'}->{$params->{'vserver'}}->{$params->{'volume'}}->{'errors'}},
                \%error_data);
    } else {
        push (@{$self->{'errors'}}, \%error_data);
    }

    #--------------------------
    # If this is a fatal error
    #--------------------------
    if ($params->{'fatal'}) {

        $Log->trace("'fatal' parameter was set");

        #--------------------------------------------
        # Set string to return if any failures occur
        #--------------------------------------------
        my $string = " - all ";
        $string   .= ($self->{'caller'} =~ /Copy/i) ? "copies" : "moves";
        $string   .= " complete";

        $job_status = "Status not available";

        #----------------------------------------------------------------
        # Need to make sure we are not generating the error while inside
        # of _gen_status or we could get in a recursive loop
        #----------------------------------------------------------------
        if (!($self->{'generating_status'})) {
            $job_status = $self->_gen_status(status_type => "Detailed");
        }

        $self->{'monitoring'} = 0;

        $params->{'ex_obj'}->throw();
    } else {

        $Log->trace("'fatal' parameter was not set.  Generating warning");

        $Log->warn($params->{'ex_obj'}->text());
    }

    $Log->exit();

    return $job_status;

} ## end sub _gen_failure

################################################################################
#
# Subroutine:   _get_attr_compare_spec
#
# Objective:    Get the validation spec for the attribute comparison methods.
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Validation spec for the params that are passed for attribute
#      comparison.
#
################################################################################

sub _get_attr_compare_spec {

    $Log->enter();

    my $self = shift;

    return {
        volumes        => {type     => ARRAYREF,
                           optional => 1,
                           default  => $self->{'src_vols'},
                          },
        match_fields   => {type     => ARRAYREF,
                           optional => 1,
                          },
        diff_fields    => {type     => ARRAYREF,
                           optional => 1,
                          },
        skip_fields    => {type     => ARRAYREF,
                           optional => 1,
                          },
        match_val      => {type     => HASHREF,
                           optional => 1,
                          },
        default_action => {type     => SCALAR,
                           optional => 1,
                           regex    => qr/die|match|diff|skip/,
                           default  => 'match',
                          },
    };

    $Log->exit();

} ## end sub _get_attr_compare_spec

################################################################################
#
# Subroutine:   _get_common_param_spec
#
# Objective:    Get the validation spec for the common and per volume
#               parameters to the 'new' method.
#
# Details:
#
#   Input:
#      Operation-specific parameters passed to 'new'.
#
#   Return value:
#      Validation spec for the global/per volume params.
#
################################################################################

sub _get_common_param_spec {

    $Log->enter();

    my $self   = shift;
    my $params = shift;

    my $global_params_key  = ($self->{'caller'} =~ /Copy/i)
                             ? 'global_copy_params'
                             : 'global_move_params';

    my $per_vol_params_key = ($self->{'caller'} =~ /Copy/i)
                             ? 'per_vol_copy_params'
                             : 'per_vol_move_params';

    #------------------------------------------------------
    # At least one of the two parameters must be specified
    #------------------------------------------------------
    my $param_spec = {
        $global_params_key  => {type     => HASHREF,
                                optional => (exists $params->{$per_vol_params_key})
                                             ? 1
                                             : 0,
                               },
        $per_vol_params_key => {type     => ARRAYREF,
                                optional => (exists $params->{$global_params_key})
                                            ? 1
                                            : 0,
                               },
    };

    $Log->exit();

    return $param_spec;

} ## end sub _get_common_param_spec

################################################################################
#
# Subroutine:   _get_common_global_param_spec
#
# Objective:    Get the validation spec for the common global params for both
#               move and copy.  These are also the params that are common
#               between global params and per volume params.
#
# Details:
#
#   Input:
#       None
#   Return value:
#      Validation spec for the params that are passed for attribute
#      comparison.
#
################################################################################

sub _get_common_global_param_spec {

    $Log->enter();

    my $self   = shift;

    #----------------------------------------------------------
    # Get the validation spec for all of the common parameters
    #----------------------------------------------------------
    my $common_param_spec = $self->_get_common_validate_spec();

    #--------------------------------------------------------------
    # Combine the list of common parameters and the list of common
    # global parameters
    #--------------------------------------------------------------
    my $common_global_params = {

        %{$common_param_spec},

        vserver            => {type => SCALAR,
                               optional => 1,
                              },
        volumes            => {type      => ARRAYREF,
                              },
        iteration_mode     => {type      => SCALAR,
                               optional  => 1,
                               regex     => qr/^(iteration|duration)$/,
                               default   => 'iteration',
                              },
        iteration_value    => {type      => SCALAR,
                               optional  => 1,
                               default   => 1,
                              },
        parallel_count     => {type      => SCALAR,
                               optional  => 1,
                               default   => 1,
                               regex     => qr/^\d+$/,
                              },
        wait_type          => {type      => SCALAR,
                               optional  => 1,
                               default   => $self->_DEF_WAIT_TYPE,
                               regex     => qr/^(all|any)$/,
                              },
        bg_mode            => {type      => SCALAR,
                               optional  => 1,
                               default   => $self->_DEF_BG_MODE,
                               regex     => qr/^(bg|fg)$/,
                              },
        timeout            => {type      => SCALAR,
                               optional  => 1,
                               default   => 0,
                              },
        on_failure         => {type      => SCALAR,
                               optional  => 1,
                               default   => 'exit',
                               regex     => qr/^(continue|complete|exit)$/,
                              },
        create_nate_result => {type      => SCALAR,
                               optional  => 1,
                               default   => 0,
                               regex     => qr/^(no|yes)$/i,
                              },
    };

    $Log->exit();

    return $common_global_params;

} ## end sub _get_common_global_param_spec

################################################################################
#
# Subroutine:   _get_common_per_vol_param_spec
#
# Objective:    Get the validation spec for the common params for both move
#               and copy.  These are also the params that are common between
#               global params and per volume params.
#
# Details:
#
#   Input:
#      global_params - Hash containing the global params, if provided.
#
#   Return value:
#      Validation spec for the params that are passed for attribute
#      comparison.
#
################################################################################

sub _get_common_per_vol_param_spec {

    $Log->enter();

    my $self   = shift;
    my $params = {};

    if (scalar @_) {
        $params = shift @_;
    }

    $Log->debug("Parameters passed to _get_common_per_vol_param_spec: " .
                Dumper $params);

    #---------------------------------------------------
    # Get the validation spec for the common parameters
    #---------------------------------------------------
    my $common_params = $self->_get_common_validate_spec($params);

    my %vserver_hash      = ();
    $vserver_hash{'type'} = SCALAR;  # Volume is always required

    #-------------------------------------------------------------------
    # Vserver is only required if a global vserver was not specified in
    # the global parameters
    #-------------------------------------------------------------------
    if (exists($params->{'global_params'}->{'vserver'})) {
        $vserver_hash{'optional'} = 1;
        $vserver_hash{'default'}  = $params->{'global_params'}->{'vserver'};
    }

    #-------------------------------------------------------------------
    # Combine the list of common params with the validation for vserver
    # and volume
    #-------------------------------------------------------------------
    my $common_per_vol_params = {
        %{$common_params},
        vserver           => \%vserver_hash,
        volume            => {type      => SCALAR,
                              callbacks => {'in_global_vols' =>
                                            sub { grep(/$_[0]/, @{$params->{'global_params'}->{'volumes'}})
                                                }
                                           },
                             },
    };

    $Log->exit();

    return $common_per_vol_params;

} ## end sub _get_common_per_vol_param_spec

################################################################################
#
# Subroutine:   _get_common_validate_spec
#
# Objective:    Get the validation spec for the common params for both move
#               and copy.  These are also the params that are common between
#               global params and per volume params.
#
# Details:
#
#   Input:
#      When generating per volume param spec, the global parameters are passed
#      in; otherwise there is no input.
#
#   Return value:
#      Validation spec for the params that are passed for attribute
#      comparison.
#
################################################################################

sub _get_common_validate_spec {

    $Log->enter();

    my $self = shift;
    my $params;

    if (scalar @_) {
        $params = shift @_;
    }

    $Log->debug("Parameters passed to _get_common_param_spec: " .
                Dumper $params);

    #-----------------------------------------------------------------------
    # Generate the validate spec for the parameters that are common between
    # global and per_volume parameters.
    #-----------------------------------------------------------------------
    my $common_params =  {
        aggr_list         => {type      => ARRAYREF,
                              optional  => 1,
                              default   => exists($params->{'global_params'})
                                           ? $params->{'global_params'}->{'aggr_list'}
                                           : undef,
                             },
        aggrtype_affinity => {type      => SCALAR,
                              optional  => 1,
                              regex     => qr/^(32|64|any|alternate)$/,
                              default   => exists($params->{'global_params'})
                                            ? $params->{'global_params'}->{'aggrtype_affinity'}
                                            : 'any',
                             },
        aggr_loc_affinity => {type      => SCALAR,
                              optional  => 1,
                              regex     => qr/^(remote|local|any|alternate)$/,
                              default   => exists($params->{'global_params'})
                                           ? $params->{'global_params'}->{'aggr_loc_affinity'}
                                           : 'any',
                             },
        compare_attrs     => {type      => BOOLEAN,
                              optional  => 1,
                              default   => exists($params->{'global_params'})
                                           ? $params->{'global_params'}->{'compare_attrs'}
                                           : 1,
                             },
    };

    $Log->exit();

    return $common_params;

} ## end sub _get_common_validate_spec

################################################################################
#
# Subroutine:   _process_global_per_vol_params
#
# Objective:    Process the volume-specific parameters.
#
# Details:
#
#   Input:
#      global_[move|copy]_params
#      per_vol_[move|copy]_params (if specified)
#
#   Return value:
#       None
#
################################################################################

sub _process_global_per_vol_params {

    my $self   = shift;
    my $params = shift;

    my ($global_param_name, $global_params, $per_vol_param_name, $per_vol_params, $vol_info);

    #---------------------------------------
    # Set the key names based on the caller
    #---------------------------------------
    if ($self->{'caller'} =~ /Move/i) {
        $global_param_name  = 'global_move_params';
        $per_vol_param_name = 'per_vol_move_params';
    } else {
        $global_param_name  = 'global_copy_params';
        $per_vol_param_name = 'per_vol_copy_params';
    }

    $Log->debug("Parameters: " . Dumper $params);
    $Log->debug("Global params " .  Dumper $params->{$global_param_name});

    if (!defined $params->{$global_param_name}->{'iteration_mode'}) {
        $params->{$global_param_name}->{'iteration_mode'} = 'iteration';
    }

    my @global_params_array = ();
    push (@global_params_array, $params->{$global_param_name});

    my $global_validation_spec;

    #--------------------------------
    # Validate the global parameters
    #--------------------------------
    if ($self->{'caller'} =~ /Copy/i) {
        $global_validation_spec = $self->_get_global_copy_param_spec();
    } elsif ($self->{'caller'} =~ /Move/i) {
        $global_validation_spec = $self->_get_global_move_param_spec();
    } else {
        my $ex_text = "Unknown caller ($self->{'caller'}) when validating global parameters";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    eval {
        $global_params = {
                             validate (
                                       @global_params_array,
                                       $global_validation_spec,
                               )
                         };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for $global_param_name: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    #----------------------------------------------
    # Get the validation spec for Params::Validate
    #----------------------------------------------
    my $per_vol_validation_spec = ();

    if ($self->{'caller'} =~ /Copy/i) {
        $per_vol_validation_spec = $self->_get_per_vol_copy_param_spec(global_params => $global_params,
                                                                      );
    } elsif ($self->{'caller'} =~ /Move/i) {
        $per_vol_validation_spec = $self->_get_per_vol_move_param_spec(global_params => $global_params,
                                                                      );
    } else {
        my $ex_text = "Unknown caller ($self->{'caller'}) when validating per volume parameters";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    #------------------------------------
    # Validate the per-volume parameters
    #------------------------------------
    my $per_vol_data = ();
    if (exists $params->{$per_vol_param_name}) {

        $Log->debug("Per volume params " .  Dumper $params->{$per_vol_param_name});

        my $i = 1;

        foreach my $vol_params (@{$params->{$per_vol_param_name}}) {

            $Log->trace("Processing $per_vol_param_name for volume $i " .
                        "of " . scalar @{$params->{$per_vol_param_name}});

            $i++;

            $vol_info = {};

            my @vol_params = ();

            push (@vol_params, $vol_params);

            eval {
                $per_vol_params = {
                                      validate (
                                                @vol_params,
                                                $per_vol_validation_spec,
                                               )
                                  };
            };

            if ($@) {
                my $ex_text = "Parameter validation failed for " .
                              "$per_vol_param_name: $@";
                $Log->error(fail_result($ex_text));
                NATE::Exceptions::Argument->throw($ex_text);
            }

            #------------------------------------------------------------------------------
            # Save the per_volume data so original volume ordering can be maintained below
            #------------------------------------------------------------------------------
            $per_vol_data->{$per_vol_params->{'volume'}} = $per_vol_params;
        }
        $Log->debug("Per volume parameter hash: " . Dumper $per_vol_data);
    }

    #-----------------------------------------------------------------
    # Apply the global/per-volume parameters to each specified volume
    #-----------------------------------------------------------------
    foreach my $vol (@{$global_params->{'volumes'}}) {

        $Log->trace("Processing volume params for volume $vol");

        $vol_info = {};
        my $vol_params = ();
             
        #-------------------------------------------------------
        # If this volume was not provided in the per_vol_params
        #-------------------------------------------------------
        if (! grep (/^$vol$/, keys %{$per_vol_data})) {
            $vol_params = $global_params;
        } else {
            $vol_params = $per_vol_data->{$vol};
        }
                
        #--------------------------------------------------------
        # Use the keys from the per volume validation spec since
        # the global params contains additional keys
        #--------------------------------------------------------
        foreach my $key (keys %{$per_vol_validation_spec}) {

            if ($key =~ /^volume$/i) {
                $vol_info->{$key} = $vol;
                if ($self->{'caller'} =~ /Copy/i) {
                    $vol_info->{'orig_volume'} = $vol;
                }
            } else {
                $vol_info->{$key} = $vol_params->{$key};
            }
        }

        $Log->trace("Adding the following volume data to the list of source " .
                    "volumes: " . Dumper $vol_info);
        push(@{$self->{'src_vols'}}, $vol_info);

    }

    if ($global_params->{'parallel_count'} eq 'all') {
        $self->{'parallel_count'} = scalar(@{$self->{'src_vols'}});
    } else {
        $self->{'parallel_count'} = $global_params->{'parallel_count'};
    }

    $self->{'wait_type'}          = $global_params->{'wait_type'};
    $self->{'bg_mode'}            = ($global_params->{'bg_mode'} eq 'bg')
                                    ? '-bg'
                                    : '-nobg';
    $self->{'timeout'}            = $global_params->{'timeout'};
    $self->{'on_fail'}            = $global_params->{'on_failure'};
    $self->{'create_nate_result'} = $global_params->{'create_nate_result'};

    if ($global_params->{'iteration_mode'} eq 'iteration') {
        $Log->trace("Iteration mode is set to 'iteration'");
        $self->{'iterations'} = $global_params->{'iteration_value'};
        $self->{'duration'}   = 0;
    } else {
        $Log->trace("Iteration mode is set to 'duration'");
        $self->{'iterations'} = 1;
        $self->{'duration'}   = VolOps::_normalize_duration($global_params->{'iteration_value'});
        $Log->trace("Duration is set for $self->{'duration'} seconds");
    }

    $Log->trace("Parallel count: $self->{'parallel_count'}");
    $Log->trace("Wait type: $self->{'wait_type'}");
    $Log->trace("Background mode: $self->{'bg_mode'}");
    $Log->trace("Iterations: $self->{'iteration'}");
    $Log->trace("Duration: $self->{'duration'}");
    $Log->trace("Timeout: $self->{'timeout'}");
    $Log->trace("On failure: $self->{'on_feil'}");
    $Log->trace("Source volume data: " . Dumper $self->{'src_vols'});

} ## end sub _process_global_per_vol_params

################################################################################
#
# Subroutine:   _get_common_match_fields
#
# Objective:    Get the list of match rules for fields (attributes) whose
#               match rules are common for volume operations.
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      match_fields - Reference to an array containing a list of fields whose
#                     values are expected to be the same before and after a
#                     volume operation.
#      diff_fields  - Reference to an array containing a list of fields whose
#                     values are expected to be different before and after a
#                     volume operation.
#      skip_fields  - Reference to an array containing a list of fields whose
#                     values should be ignored when comparing attributes
#                     because they may or may not be the same
#      match_values - Reference to a hash containing a list of attributes and
#                     their expected values when performing attribute
#                     comparison
#
################################################################################

sub _get_common_match_fields {

    $Log->enter();

    my $self = shift;

    #------------
    # Initialize
    #------------
    my @match_fields = ();
    my @diff_fields  = ();
    my @skip_fields  = ();
    my %match_val    = ();

    my ($maj_rel, $min_rel) = $self->_get_release();

    @match_fields = ('vserver',
                     'state',
                     'volume-style',
                     'volume-ownership',
                     'policy',
                     'user',
                     'group',
                     'security-style',
                     'unix-permissions',
                     'vsroot',
                     'comment',
                     'autosize',
                     'max-autosize',
                     'autosize-increment',
                     'maxdir-size',
                     'space-guarantee',
                     'space-guarantee-enabled',
                     'min-readahead',
                     'atime-update',
                     'snapdir-access',
                     'percent-snapshot-space',
                     'snapshot-policy',
                     'language',
                     'stripe-count',
                     'stripe-width',
                     'stripe-format',
                     'striping-epoch',
                     'properly-striped',
                     'concurrency',
                     'optimize',
                     'max-write-alloc-blocks',
                     'is-inconsistent',
                     'in-nvfailed-state',
#                     'block-type',
                     'ruleset',
                     'description',
                     'flexcache-connection-status',
                     'index-dir-enabled',
                     'read-realloc',
                     'extent-enabled',
                     'is-managed-by-service',
                     'enable-snapdiff',
                     'storage-service',
                     'fc-delegs-enabled',
                     'is-volume-in-cutover',
                     'use-new-algorithm',
                     );

    @diff_fields  = ('dsid',
                     'uuid',
                    );

    @skip_fields  = ('size',
                     'name-ordinal',
                     'available',
                     'total',
                     'used',
                     'percent-used',
                     'files',
                     'files-used',
                     'snapshot-space-used',
                     'antivirus-on-access-policy',
                     'move-target',
                     'block-type',
                     'snapshot-count',
                    );

    %match_val    = ('clone-volume'          => 'false',
                     'is-quiesced-in-memory' => 'false',
                     'is-quiesced-on-disk'   => 'false',
                    );

    #--------------------------------
    # If release is post-Boilermaker
    #--------------------------------
    if ($maj_rel >= 8.1) {
        $Log->trace("Major release is RR.0 or later, adding post-BR " .
                    "attributes to the list of match values");

        @match_fields = (@match_fields,
                         'space-mgmt-try-first',
                         'flexcache-origin-volume',
                         'convert-ucode',
                         'create-ucode',
                         'flexcache-cache-policy',
                         'flexcache-fill-policy',
                         'is-sis-logging-enabled',
                         'is-sis-volume',
                         'is-cluster-volume',
                         'snapshot-clone-dependency',
                         'auto-snapshots-enabled',
                         'sched-snap-name',
                         'constituent-role',
                         'is-constituent',
                         'flexcache-min-reserve',
                         'vm-align-sector',
                         'vm-align-suffix',
                         'vbn-zero-space-saved',
                         'vbn-zero-space-saved-percent',
                         'vm-align-suffix',
                         'ols-constituent-size',
                         'ols-constituent-count',
                         'enable-object-store',
                         'is-reserve-constituent',
                         'object-write-sync-period',
                         'ols-aggr-list',
                         'force-nvfail-on-dr',
                         'autosize-mode',
                         'min-autosize',
                         'autosize-grow-threshold-percent',
                         'autosize-shrink-threshold-percent',
                         'qos-policy-group',
                         'external-cache',
                         'namespace-aggregate',
                         'max-namespace-constituent-size',
                         'ns-mirror-aggr-list',
                         'data-aggr-list',
                         'percent-reserve-constituent',
                         'max-data-constituent-size',
                         'space-nearly-full-threshold-percent',
                         'space-full-threshold-percent',
                        );

        @skip_fields  = (@skip_fields,
                         'compression-space-saved-percent',
                         'sis-space-saved-percent',
                         'dedupe-space-saved-percent',
                         'sis-space-saved',
                         'dedupe-space-saved',
                         'compression-space-saved',
                         'dedupe-space-shared',
                         'hybrid-cache-write-caching-ineligibility-reason',
                         'node',
                         'is-moving',
                         'filer',
                         'filesystem-size',
                         'size-used-by-snapshots',
                        );
                        
        %match_val    = (%match_val,
                         'directory-bitwidth' => '-',
#is-moving is going to the skip array temporarily because of burt 579138
#When this burt merges to the sn branches it can be uncommented from this 
#section and removed from the skip field.
#                         'is-moving' => 'false',
                         'hybrid-cache-eligibility' => 'read-write', 
                         'transition-behavior' => 'none',
                        );
    }

    $Log->exit();

    return (\@match_fields,
            \@diff_fields,
            \@skip_fields,
            \%match_val,
           );

} ## end sub _get_common_match_fields

################################################################################
#
# Subroutine:   _get_release
#
# Objective:    Get the numeric release number of the build
#
# Details:
#
#   Input:
#      node - Node for which to obtain the release.  If not specified, it
#             checks the release on the current node.
#
#   Return value:
#      Returns the value of the default wait type
#
################################################################################

sub _get_release {

    $Log->enter();

    my $self = shift;

    my $params;

    eval {
        $params = {
                    validate(
                             @_,
                             {
                              node                => {
                                                      type     => SCALAR,
                                                      optional => 1,
                                                     },
                             }
                            )
                };

    };

    if ($@) {
        my $ex_text = "Parameter validation failed: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    if (exists $params->{'node'}) {
        $self->{'server'}->set_execution_server(-Server => $params->{'node'});
    }

    my $release = $self->{'server'}->get_release(format=>"raw");

    #------------------------------------------
    # Remove the release name from the release
    #------------------------------------------
    $release =~ s/^.*__//;

    my ($maj1, $maj2, $min) = split(/\./, $release);

    my $maj = $maj1 . '.' . $maj2;

    $Log->exit();

    return ($maj, $min);

} ## end sub _get_release

################################################################################
#
# Subroutine:   _DEF_WAIT_TYPE
#
# Objective:    Returns the default wait type for processing jobs
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Returns the value of the default wait type
#
################################################################################

sub _DEF_WAIT_TYPE {

    return 'any';

} ## end sub _DEF_WAIT_TYPE

################################################################################
#
# Subroutine:   _DEF_BG_MODE
#
# Objective:    Returns the default background mode for the volume operation
#               'bg' runs the operation in a subtest, 'fg' runs it as a
#               blocking operation
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Returns the value of the default background mode
#
################################################################################

sub _DEF_BG_MODE {

    return 'bg';

} ## end sub _DEF_BG_MODE

################################################################################
#
# Subroutine:   _BLOCK_TYPES
#
# Objective:    Returns the aggregate block types that are available
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Returns a comma-separated list of currently available block types
#
################################################################################

sub _BLOCK_TYPES {

    my $self = shift;

    return '32-bit,64-bit';

} ## end sub _BLOCK_TYPES

################################################################################
#
# Subroutine:   DESTROY
#
# Objective:    Restore the engine type to its original setting
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      None
#
################################################################################

sub DESTROY {

    $Log->enter();

    my $self = shift;

    if ($@) {
        my $ex_text = "An error was detected when destroying the object\n";
        $ex_text   .= "Error: $@";
        my $status = $self->_gen_failure(ex_obj => new VolOps::BaseException($ex_text),
                                        );
    }

    #---------------------------
    # If not inside the subtest
    #---------------------------
    if (! exists $self->{'subtested'}) {
        $Log->trace("Destroying object");
        NOTE::logcomment("Restoring the Paloma engine type");
        $self->set_engine_type(engine => 'restore');
    } else {
        $Log->trace("Destroy object inside subtest");
    }

    $Log->exit();

    return;

} ## end sub DESTROY

1;
