# $Id: //depot/prod/test/Rsierranevada/lib/VolOps/Copy.pm#1 $

## @summary     Programmatic control of volume copy.
## @author      dl-volops-qa@netapp.com
## @status      private
## @pod         here

=head1 NAME

VolOps::Copy.pm

=head1 SYNOPSIS

  use VolOps::Copy;
  use VolOps::BaseException;
  use strict;
  
  my %global_params = (
          iteration_mode   => "duration",
          iteration_value  => "5m",
          volumes            => [
                                    "testvol1",,
                                    "testvol2",,
                                    "testvol3",,
                                    "testvol4",,
                                    "testvol5",,
                                ],
          vserver          => "vs0",
          parallel_count   => 2,
          on_failure       => "complete",
          );
  
  my @per_vol_params = (
          {
              volume            => "testvol1",
              aggr_loc_affinity => "alternate",
              dest_volume       => "ctestvol1",
              compare_attrs     => 0,
          },
          {
              vserver           => "vs1",
              volume            => "testvol2",
              aggr_list         => ["csiqa_3070_33_aggr2",
                                    "csiqa_3070_34_aggr1",
                                   ],
          },
          {
              volume            => "testvol4",
              aggrtype_affinity => "64",
              keep_vols         => 0,
              cascade_copies    => 1,
          },
          );
  
  my $vcp = new VolOps::Copy(node                => "csiqa-3070-34",
                             engine              => "BRE",
                             per_vol_copy_params => \@per_vol_params,
                             global_copy_params  => \%global_params,
                             delete_existing     => 1,
                             save_status         => 1,
                            );
  
  # Copy volumes
  my $sub_obj = $vcp->copy();
  
  my $status;
  
  # Get status while volumes are copying
  while ($vcp->status()) {
      $status = $vcp->get_status();
      snooze(30);
      $Log->comment("Status: " . $status->{"formatted"});
  }
  
  # Get the status of the copies
  $status = $vcp->get_status();
  
  $Log->comment("Final status: " . $status->{"formatted"});
  
  if (($status->{"total_failed"} == 0) &&
      ($status->{"total_errors"} == 0)) {
      $Log->comment("All copies passed");
  }

=head1 DESCRIPTION

C<VolOps::Copy> is a class for performing volume copy testing.

The class throws exceptions on failures.

This class makes use of its own exception class
VolOps::BaseException. While the exception class will be
inherited when using this library, it is recommended to explicitly import
the exception class for clarity.

The current list of possible exceptions are:
    NATE::Exceptions::Argument
    VolOps::BaseException
    VolOps::ConfigError
    VolOps::JobFailed
    VolOps::JobStatusFailure
    VolOps::AttributeComparison
    VolOps::MissingValidationRule
    VolOps::RetriesExhausted
    VolOps::SrcVolumeFailure
    VolOps::SrcAggregateFailure
    VolOps::AggregateSelectionFailure
    VolOps::DestAggregateFailure
    VolOps::PropsNotAvailable
    VolOps::StripingNotSupported
    VolOps::ConnectionFailure
    VolOps::DataStructureStoreFailure
    VolOps::InternalError
    VolOps::Copy::DestVolumeExists
    VolOps::Copy::DestVolumeFailure
    VolOps::Copy::CopyFailed

=head1 METHODS

=cut

package VolOps::Copy;
use warnings;
use base qw(VolOps);
use Params::Validate qw(validate BOOLEAN SCALAR HASHREF ARRAYREF UNDEF OBJECT);
use Tharn qw(fatal_result config_result fail_result pass_result $Staf $Log $Parent snooze subtest);
use Data::Dumper;
use VolOps::BaseException;
use GNOTE;
use Node;
use Date::Parse;
use NATE::Events qw(listener_add listener_remove);
use NATE::Exceptions::Argument;

my $Options = new Options();

=head2 new

C<< VolOps::Copy->new(%params) >>

Instantiate an object to use for copying volume(s).

=over 4

=item Input

=over 4

=item * node => <scalar>

(Required)

Identifies one of the nodes in the cluster to be used to attach to the
cluster.

=back

=over 4

=item * engine => <scalar>

(Optional)
Default = "BRE"

Paloma engine to be used for the transfer.  Sets the value of
I<paloma.physical> to B<true> if set to "BRE" and sets it to B<false> if set to
"LRE".  The existing value of paloma.physical is saved then later restored
when the object is destroyed.

Valid values:
    BRE - Use physical replication
    LRE - Use logical replication

=back

=over 4

=item * %global_copy_params => <hashref>

(Required)

Parameters to be used for volume copy operations for each of the volumes
specified.  Parameters consist of the following:

=over 4

=item - vserver => <scalar>

(Optional, required if no per_vol_copy_params are defined)
Default = undef

Name of the vserver to use for all volume operations.

=item - volumes => <arrayref>

(Required)
Default = undef

List of volumes on which to operate.  If a destination volume name is
not provided in the per_vol_copy_params, the source volume name is
appended with "_copy" and an ordinal which is incremented for each
subsequent copy.

=item - aggr_list => <arrayref>

(Optional)
Default = undef

List of destination aggregates from which to select when copying volumes.
If not specified, aggregates associated with the nodes specified by the NODES
parameter on the command line will be used.

=item - aggrtype_affinity => <scalar>

(Optional)
Default = "any"

Block type for the destination aggregate.

Valid values:
    32         - Select a 32-bit aggregate as the
                 destination.
    64         - Select a 64-bit aggregate as the
                 destination.
    any        - Select an aggregate with any block type.
    alternate  - Select an aggregate with a block type that
                 does not match the current block type.

=item - aggr_loc_affinity => <scalar>

(Optional)
Default = "any"

Location of the destination aggregate with respect to the source node.

Valid values:
    local      - Select an aggregate on the same node as
                 the aggregate on which the volume resides.
    remote     - Select an aggregate on a different node.
    any        - Randomly select an aggregate on either a
                 local or a remote node.
    alternate  - Alternate between "remote" and "local",
                 starting with "remote".

=item - iteration_mode => <scalar>

(Optional)
Default = "iteration"

Type of continuity to use for execution of the volume copy(s).

Valid values:
    iteration - Iterate through the list of volumes the
                specified number of times.
    duration  - Continue copying the list of volumes for
                the specified duration.

=item - iteration_value => <scalar>

(Optional)
Default = 1 if iteration_mode is I<iteration>; "undef" if iteration_mode is
I<duration>

Length of time or number of iterations to perform the copy operation.  If
iteration_mode is I<duration>, I<iteration_value> can be appended with
H or h for hours, D or d for days, M or m for minutes, S or s for seconds.
Default granularity is minutes.

=item - parallel_count => <scalar>

(Optional)
Default = 1

Number of volumes to copy in parallel.  A value of "1" will result in serial copies.

Valid values:
    all
    <number>

=item - wait_type => <scalar>

(Optional)
Default = "any"

Indicates whether to wait for all volumes in a group to finish copying before
starting the next copy or to just wait until any of the volumes finishes
before starting the next volume copy operation.  The number of items in a
group is defined by the value of parallel_count.

Valid values:
    any - Wait until at least one job completes before
          starting the next copy.
    all - Wait until all jobs in the current group have
          completed copying before starting the next group.
          The size of the group is defined by the
          I<parallel_count> parameter.

=item - bg_mode => <scalar>

(Optional)
Default = "bg"

Indicates whether to run the volume copy(s) in the foreground or background.
If run in the foreground, B<copy> will block until complete.

Valid values:
    fg - Run the copies in the foreground and block until
         complete.
    bg - Start the copies in the background and return
         immediately.

=item - timeout => <scalar>

(Optional)
Default = 0

Timeout value for the copy operation.  A value of 0 will wait forever.

=item - on_failure => <scalar>

(Optional)
Default = "exit"

Indicates what action to perform when an error is encountered while copying a volume.

Valid values:
    continue - Record any non-abortable failures and
               continue with any remaining copies.
    complete - Allow all volumes currently copying to
               complete before throwing an exception.
    exit     - Immediately throw an exception on failure.
               The status of any ongoing copy jobs will be
               lost.

=item - compare_attrs => <boolean>

(Optional)
Default = 1 (true)

Indicates whether or not to compare volume attributes (pre- and post-copy)
after each copy completes.  The comparison will be performed every time a
volume or group of volumes has completed a copy.

=item - cascade_copies => <boolean>

(Optional)
Default = 0 (false)

Indicates whether or not to use the original source volume or to use the
previous destination volume as the source volume for subsequent volume copy
commands.  A value of 1 will always use the previous destination when
performing subsequent copies.

=item - keep_vols => <boolean>

(Optional)
Default = 1 (true)

Indicates whether or not to keep the source volume after a copy has completed
or to delete it.  A value of 1 will retain volumes.

=back

=item * @per_vol_copy_params => <arrayref>

(Optional)
Default = undef

Array of hashes that provides parameters to use for copying each specified
volume.  Any values supplied override those specified in the global copy params hash.
Any volumes specified must also be included in the global_copy_params volumes
param.

=over 4

=item - vserver => <scalar>

(Optional, required if no global_copy_params are defined)
Default = undef

Name of the vserver to use for the specific volume operation.  Overrides any
vserver value set in global_copy_params.

=item - volume => <scalar>

(Required)
Default = undef

Volume to copy

=item - aggr_list => <arrayref>

(Optional)
Default = undef

List of destination aggregates to select from.  If not specified, aggregates
associated with nodes specified by the NODES param are considered.  Overrides
any aggr_list value set in global_copy_params for a specific volume.

=item - aggrtype_affinity => <scalar>

(Optional)
Default = 64

Block type for the destination aggregate.

Valid values:
    32         - Select a 32-bit aggregate as the destination.
    64         - Select a 64-bit aggregate as the destination.
    any        - Select an aggregate with any block type.
    alterate   - Select an aggregate with a block type that
                 does not match the current block type.

=item - aggr_loc_affinity => <scalar>

(Optional)
Default = "remote"

Location of the destination aggregate with respect to the source node.
Overrides any aggr_loc_affinity value set in global_copy_params.

Valid values:
    local      - Select an aggregate on the same node as
                 the aggregate on which the volume resides.
    remote     - Select an aggregate on a different node.
    any        - Randomly select an aggregate on either a
                 local or a remote node.
    alternate  - Alternate between "remote" and "local",
                 starting with "remote".

=item - compare_attrs => <boolean>

(Optional)
Default = 1 (true)

Indicates whether or not to compare volume attributes (pre- and post-copy)
after each copy completes.  The comparison will be performed every time the
volume has completed a copy.

=item - cascade_copies => <boolean>

(Optional)
Default = 0 (false)

Indicates whether or not to use the original source volume or to use the
previous destination volume as the source volume for subsequent volume copy
commands.  A value of 1 will always use the previous destination when
performing subsequent copies.

=item - keep_vols => <boolean>

(Optional)
Default = 1 (true)

Indicates whether or not to keep the source volume after a copy has completed
or to delete it.  A value of 1 will retain volumes.

=back

=item * save_status => <boolean>

(Optional)
Default = false

Indicates whether or not to save the status hash for each iteration.  If this
flag is set, the status hash will be saved, by default, as a text file to the
LOGDIR using file name "Vol[Copy or Copy]_status_<timestamp>.dd".  The save
mode, filename, and/or directory can be modified on the command line using the
DSS_MODE, DSS_DIR, DSS_FILE, and DSS_FILENAME parameters.  See the POD for
lib/DataStructureStore.pm for more information.

=item * delete_existing => <boolean>

(Optional)
Default = false

Indicates whether or not delete any copy destination volume that already exists
before attempting the copy.

=back

=item Output

Reference to the object.

=item Example

 my $sub_obj = $vcp->new(node                => csiqa-3070-99,
                         engine              => "BRE",
                         per_vol_copy_params => \@per_vol_params,
                         global_copy_params  => \%global_copy_params,
                         save_status         => 1,
                        );

=back

=item * create_nate_result => <scalar>

(Optional)
Default = 'yes'

Indicates whether or not to create a nate result when the copy fails.
By default, a NATE fail result will be generated when an error or failure
occurs.  By setting this value to 'no', a NATE warn result will be generated
to provide a warning that something went wrong.  When run in the background,
the NATE result (warn or fail) does not get propagated back to the parent.
This flag is useful for negative testing when a failure is expected.

=back

=cut

################################################################################
#
# Subroutine:   new()
#
# Objective:    Invocant for the class
#
# Details:
#
#   Input:
#       node                - Node in the cluster
#       engine              - Paloma engine to be used for the transfer
#                             (BRE or LRE)
#       global_copy_params  - Copy params that will be applied to all volumes
#                             being copied
#       per_vol_copy_params - Copy params that will be applied only to
#                             individual volumes
#       save_status         - Indicates whether or not to save the status hash
#                             for each iteration.  If this flag is set, the
#                             status hash will be saved, by default, as a text
#                             file to the LOGDIR using file name
#                             "Vol[Copy|Copy]_status_<timestamp>.dd".  The save
#                             mode, filename, and/or directory can be modified
#                             on the command line using the DSS_MODE, DSS_DIR,
#                             DSS_FILE, and DSS_FILENAME parameters.  See the
#                             POD for lib/DataStructureStore.pm for more
#                             information.
#                             Optional; default = 0
#       delete_existing     - Indicates whether or not to delete any copy
#                             destination volumes that exist before attempting
#                             the copy.
#
#   Return value:
#       object
#
################################################################################

sub new {
    $Log->enter();

    my $class = shift;

    my (%common_params, %copy_params);

    #--------------------------------------------------------------------------
    # Split the params into ones needed by base class and copy-specific params
    #--------------------------------------------------------------------------
    my %args = @_;
    foreach my $key (keys %args) {

        if ($key =~ /global_copy_params|per_vol_copy_params|delete_existing/) {
            $copy_params{$key} = $args{$key};
        } else {
            $common_params{$key} = $args{$key};
        }
    }

    my $self  = $class->SUPER::new(\%common_params);

    #--------------------
    # Process parameters
    #--------------------
    $self->_process_copy_params(\%copy_params);

    my ($maj, $min) = $self->_get_release();


    #---------------------------
    # If RollingRock.0 or later
    #---------------------------
    if ($maj >= 8.1) {
        $Log->trace("Build is RollingRock or later");
        #------------------------------------------------------------------
        # Append fields that have common validation rules for physical and
        # logical
        #------------------------------------------------------------------
        @{$self->{'skip_fields'}}  = (@{$self->{'skip_fields'}},
                                      'qos-policy-group',
                                     );
        @{$self->{'match_fields'}} = (@{$self->{'match_fields'}},
                                      'is-inconsistent',
                                      'is-copied-for-transition',
                                      'is-transitioned',
                                     );

        @{$self->{'diff_fields'}}  = (@{$self->{'diff_fields'}},
                                      'volume',
                                      'msid',
                                      'create-time',
                                      'crtime',
                                      'instance-uuid',
                                      'provenance-uuid',
                                     );

        %{$self->{'match_val'}}    = (%{$self->{'match_val'}},
                                      'type'                 => 'RW',
                                      'filesys-size-fixed'   => 'false',
                                      'load-sharing-source'  => '-',
                                      'junction-path'        => '-',
                                      'junction-path-source' => '-',
                                      'junction-active'      => '-',
                                      'junction-parent'      => '-',
                                      'parent'               => '-',
                                     );

        #-------------------------------------------
        # If the physical replication engine is set
        #-------------------------------------------
        if ($self->{'engine'} =~ /BRE/) {
            #----------------------
            # Physical replication
            #----------------------
            @{$self->{'match_fields'}} = (@{$self->{'match_fields'}},
                                          'fractional-reserve',
                                          'physical-replica',
                                          'transition-type',
                                          'transition-state',
                                          'transition-history',
                                          'nvfail',
                                          'is-nvfail',
                                         );

            @{$self->{'skip_fields'}}  = (@{$self->{'skip_fields'}},
                                          'physical-replica',
                                          'overwrite-reserve',
                                         );
        } else {
            #---------------------
            # Logical replication
            #---------------------
            @{$self->{'skip_fields'}} = (@{$self->{'skip_fields'}},
                                         'transition-type',
                                         'transition-state',
                                         'transition-history',
                                         'nvfail',
                                         'is-nvfail',
                                        );

            %{$self->{'match_val'}}   = (%{$self->{'match_val'}},
                                         'physical-replica' => 'false',
                                       );
        }

        $Log->debug("Copy match fields: " . Dumper $self->{'match_fields'});
        $Log->debug("Copy diff fields: " . Dumper $self->{'diff_fields'});
        $Log->debug("Copy skip fields: " . Dumper $self->{'skip_fields'});
        $Log->debug("Copy match value: " . Dumper $self->{'match_val'});

    } else {

        @{$self->{'match_fields'}} = (@{$self->{'match_fields'}},
                                      'is-inconsistent',
                                     );

        @{$self->{'diff_fields'}}  = (@{$self->{'diff_fields'}},
                                      'volume',
                                      'msid',
                                      'create-time',
                                      'crtime',
                                     );

        %{$self->{'match_val'}}    = (%{$self->{'match_val'}},
                                      'type'                 => 'RW',
                                      'filesys-size-fixed'   => 'false',
                                      'load-sharing-source'  => '-',
                                      'junction-path'        => '-',
                                      'junction-path-source' => '-',
                                      'junction-active'      => '-',
                                      'parent'               => '-',
                                     );
    }

    $self->{'loc_affins'} = ['local',
                             'remote',
                            ];

    $Log->exit();

    return $self;

} ## end sub new

################################################################################
#
# Subroutine:   _process_copy_params
#
# Objective:    Process the global_copy_params and per_vol_copy_params
#
# Details:
#
#   Input:
#      global_copy_params
#      per_vol_copy_params
#
#   Return value:
#       None
#
################################################################################

sub _process_copy_params {

    $Log->enter();

    my $self   = shift;

    my $params;

    my $validate_spec = $self->_get_common_param_spec(@_);

    # Add validation spec for 'delete_existing' param
    $validate_spec->{'delete_existing'}->{'type'}     = BOOLEAN;
    $validate_spec->{'delete_existing'}->{'optional'} = 1;
    $validate_spec->{'delete_existing'}->{'default'}  = 0;
 
    eval {
 
        $params = {
                      validate (
                                   @_,
                                   $validate_spec,
                             )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed in _process_copy_params: $@";

        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    #----------------------------------------------
    # Process the global and per volume parameters
    #----------------------------------------------
    $self->_process_global_per_vol_params($params);

    $self->{'delete_existing'} = $params->{'delete_existing'};

    $Log->trace("Parallel_count: " . $self->{'parallel_count'});
    $Log->trace("Wait type: " . $self->{'wait_type'});
    $Log->trace("Save status: " . $self->{'save_status'});
    $Log->trace("BG mode: " . $self->{'bg_mode'});
    $Log->trace("Iterations: " . $self->{'iterations'});
    $Log->trace("Duration: " . $self->{'duration'});
    $Log->trace("Timeout: " . $self->{'timeout'});
    $Log->trace("On failure: " . $self->{'on_fail'});
    $Log->trace("Delete existing: " . $self->{'delete_existing'});

    $Log->exit();

} ## end sub _process_copy_params

=head2 copy 

C<< VolOps::Copy->copy() >>

Copy the volume(s) to the specified aggregate or to an aggregate of a
specified type.

=over 4

=item Input

None

=item Output

If executed in the background, a reference to the subtest object associated
with the copy operation is returned.  If executed in the foreground, nothing
is returned.

=item Example

C<< my $sub_obj = $vcp->copy(); >>

=back

=cut

################################################################################
#
# Subroutine:   copy()
#
# Objective:    Copy the volume(s) per specifications.  If background mode of
#               'bg' was specified during instantiation, copy is done in the
#               background, otherwise; it is a blocking operation.
#
# Details:
#
#   Input:
#      None 
#
#   Return value:
#      Returns the subtest object if copy is being executed in the background; 
#      otherwise, nothing is returned.
#
################################################################################

sub copy {

    $Log->enter();

    my $self = shift();

    $self->{'job_status'} = ();

    #----------------------------------------------------------
    # If running in the background, subtest the copy operation
    #----------------------------------------------------------
    if ($self->{'bg_mode'} eq '-bg') {

        $self->{'subtested'} = 1;
        $self->{'st_obj'}    = subtest(\&VolOps::_perform_vol_ops,
                                       $self->{'bg_mode'},
                                       '-timeout', $self->{'timeout'},
                                       '-runid', "copy_volumes_" . time(),
                                       "-onexit", sub {},
                                       '--', $self,
                                      ); 
        snooze(1);

    } else {

        #-----------------------
        # Run in the foreground
        #-----------------------

        $self->_perform_vol_ops;

    }

    $Log->exit();

    ($self->{'bg_mode'} eq '-bg') ? return $self->{'st_obj'} : return;

} ## end sub copy 

################################################################################
#
# Subroutine:   _copy()
#
# Objective:    Copy the list of specified volumes to their selected
#               destination aggregates.  Copies will be performed in parallel
#               if indicated.
#
# Details:
#
#   Input:
#      None 
#
#   Return value:
#      None 
#
################################################################################

sub _copy {

    $Log->enter();

    my $self = shift;

    #-----------------------
    # Attach to each volume 
    #-----------------------
    my @vols = $self->_prep_vols();

    my @copy_vols     = ();
    my $dest_vol_name = "";
    my $done          = 0;
    my $expired       = 0;
    my %active_jobs   = ();
    my $start_time    = time();
    my $group_size    = $self->{'parallel_count'};
    my $dest_vol;

    #------------------------------------------------------------------
    # Continue copying volumes until duration has elapsed or specified 
    # iterations are complete
    #------------------------------------------------------------------
    while (!$done) {

        $Log->debug("The content of vols is: " . Dumper \@vols);

        my $completed = ();

        #--------------------------------------
        # If there are volumes left to process
        #--------------------------------------
        if ((scalar @vols) && (!$expired)) {
            #--------------------------------
            # Get the next volume(s) to copy 
            #--------------------------------
            $Log->debug("Splicing $group_size volumes from the list of " .
                        "volumes");
            @copy_vols = splice(@vols, 0, $group_size);

            $Log->debug("Copying the following volumes: ");
            foreach my $vol (@copy_vols) {
                $Log->debug("\t$vol->{'volume'}");
            }

        } else {
            NOTE::logcomment("There are no more volumes left to copy");
            @copy_vols = ();

            #-----------------------------------------------
            # If there are no active jobs left, we are done
            #-----------------------------------------------
            my $left = scalar keys %active_jobs;
            $Log->comment("The number of remaining active jobs is: " . $left);
            if ($left == 0) {
                $Log->trace("Iteration complete.");
                $done = 1;
                last;
            }
        }

        #-------------------------------------------------------------
        # Create new destination objects for each volume to be copied 
        #-------------------------------------------------------------
        foreach my $curr_vol (@copy_vols) {

            #------------------------------------------------------------
            # If cascading and this volume has been copied at least once
            #------------------------------------------------------------
            if (($curr_vol->{'cascade_copies'}) && exists $curr_vol->{'curr_dest_volume'}) {
                $Log->trace("Cascading was requested for volume " .
                            "$curr_vol->{'orig_volume'}.  Setting " .
                            "new source volume to $curr_vol->{'curr_dest_volume'}");

                #--------------------------------------------------------------
                # Set the last aggregate (remote or local) for the new source 
                # volume in case loc_affin is set to alternate so 
                # pick_aggregates know whether to pick a local or remote 
                # aggregate for the next destination aggregate
                #--------------------------------------------------------------
                $self->{'vol_objs'}->{$curr_vol->{'curr_dest_volume'}}->{$curr_vol->{'vserver'}}->{'last_aggr'} = $self->{'vol_objs'}->{$curr_vol->{'volume'}}->{$curr_vol->{'vserver'}}->{'last_aggr'};
                $curr_vol->{'volume'} = $curr_vol->{'curr_dest_volume'};
            }

            $Log->trace("Preparing to copy $curr_vol->{'volume'}");

            #------------------------------------
            # Select a new destination aggregate
            #------------------------------------
            my $vol_obj = $self->{'vol_objs'}->{$curr_vol->{'volume'}}->{$curr_vol->{'vserver'}};
            $curr_vol->{'src_aggr'}  = $vol_obj->Aggregate->name;
            $curr_vol->{'dest_aggr'} = $self->_select_aggr(vol_info => $curr_vol);

            #--------------------------------------------
            # If a destination volume name was specified
            #--------------------------------------------
            if (defined $curr_vol->{'dest_volume'}) {
                $dest_vol_name = $curr_vol->{'dest_volume'};

                #---------------------------------------------
                # If this volume has already been copied once
                #---------------------------------------------
                if (exists $self->{'vol_objs'}->{$curr_vol->{'volume'}}->{$curr_vol->{'vserver'}}->{'last_copy'}) {
                    $dest_vol_name .= "_copy";
                }
            } else { # Use the source volume's name and append "_copy<num>"
                
                $dest_vol_name = $curr_vol->{'orig_volume'} . "_copy";
            }

            #--------------------------------------------------------------
            # If this volume was previously copied, increment the counter.
            # Otherwise, start with 1.
            #--------------------------------------------------------------
            if (exists $self->{'vol_objs'}->{$curr_vol->{'orig_volume'}}->{$curr_vol->{'vserver'}}->{'last_copy'}) {
                $self->{'vol_objs'}->{$curr_vol->{'orig_volume'}}->{$curr_vol->{'vserver'}}->{'last_copy'}++;
                $dest_vol_name .= sprintf("%04i", $self->{'vol_objs'}->{$curr_vol->{'orig_volume'}}->{$curr_vol->{'vserver'}}->{'last_copy'});
            } else {
                $Log->trace("No previous copies exist for this volume.  Setting copy counter to 1.");

                #-----------------------------------------------------------
                # If a destination volume was not specified for this volume
                #-----------------------------------------------------------
                if (!defined $curr_vol->{'dest_volume'}) {

                    $dest_vol_name .= sprintf("%04i", 1);
                    $self->{'vol_objs'}->{$curr_vol->{'orig_volume'}}->{$curr_vol->{'vserver'}}->{'last_copy'} = 1;
                } else {
                    #------------------------------------------------------------------------
                    # Set last copy to 0 so it can be incremented to 1 in the next iteration 
                    #------------------------------------------------------------------------
                    $self->{'vol_objs'}->{$curr_vol->{'volume'}}->{$curr_vol->{'vserver'}}->{'last_copy'} = 0;                
                }
            }

            #-------------------------------------------------------
            # Create a new volume object for the destination volume
            #-------------------------------------------------------
            eval {

                #------------------------------------------------
                # Create a new object for the destination volume
                #------------------------------------------------
                $dest_vol = NOTE::Volume->new(name      => $dest_vol_name,
                                              Vserver   => $curr_vol->{'vserver'},
                                              Node      => $self->{'node'},
                                              Aggregate => $curr_vol->{'dest_aggr'},
                                             );

                #-------------------------------------------------------------
                # Check to see if the volume exists.  If so, throw an  
                # exception unless it was specifically requested to delete 
                # pre-existing volumes. 
                #-------------------------------------------------------------
                if ($dest_vol->exists) {

                    #---------------------------------------------------------
                    # If it was not specifically requested to delete existing
                    # volumes
                    #---------------------------------------------------------
                    if (!$self->{'delete_existing'}) {
                        my $ex_text = "Destination volume $dest_vol_name already exists: $@";
                        my $status = $self->_gen_failure(ex_obj  => new VolOps::Copy::DestVolumeExists($ex_text), 
                                                         volume  => $dest_vol_name,
                                                         vserver => $curr_vol->{'vserver'},
                                                        );
                    } else {

                        #-------------------------------------------------------------
                        # NOTE will delete the target volume if it exists when
                        # performing a volume copy, but it will uses the original 
                        # aggregate when copying the volume instead of the one we 
                        # specify, so we need to delete the existing volume
                        # and create a new volume object for the copy method. 
                        #-------------------------------------------------------------
                        $Log->warn("Destination volume $dest_vol_name already exists.  " .
                                   "Deleting volume...");

                        $self->{'server'}->volume_purge(volume  => $dest_vol_name,
                                                        vserver => $curr_vol->{'vserver'},
                                                       ); 

                        $dest_vol = NOTE::Volume->new(name      => $dest_vol_name,
                                                      Vserver   => $curr_vol->{'vserver'},
                                                      Node      => $self->{'node'},
                                                      Aggregate => $curr_vol->{'dest_aggr'},
                                                 );
                    }
                }

                $Log->trace("The aggregate name for the new volume object is: " . 
                            $dest_vol->Aggregate->name);
            };

            if ($@) {
                $Log->debug("Failed to create destination volume object: $@");

                my $ex_text = "Failed to create destination volume object: $@";
                my $status = $self->_gen_failure(ex_obj  => new VolOps::Copy::DestVolumeFailure($ex_text), 
                                                 volume  => $dest_vol_name,
                                                 vserver => $curr_vol->{'vserver'},
                                                );
            } else {
                $self->{'vol_objs'}->{$dest_vol_name}->{$curr_vol->{'vserver'}} = $dest_vol;
                $curr_vol->{'curr_dest_volume'} = $dest_vol_name;
            }
        }

        #----------------------------------------------
        # Copy each of the volumes in the current list
        #----------------------------------------------
        foreach my $curr_vol (@copy_vols) {

            $Log->trace("Preparing to copy $curr_vol->{'volume'} to $curr_vol->{'curr_dest_volume'}");

            my $vol         = $self->{'vol_objs'}->{$curr_vol->{'volume'}}->{$curr_vol->{'vserver'}};
            my $dest_vol    = $self->{'vol_objs'}->{$curr_vol->{'curr_dest_volume'}}->{$curr_vol->{'vserver'}};
            my %copy_params = ();

            $copy_params{'src'}        = $vol;
            $copy_params{'background'} = 1; # Run in the background so we can
                                            # monitor multiple parallel jobs
            
            $Log->trace("Copying $curr_vol->{'volume'} with the" .
                        " following parameters:");
            $Log->trace("Vserver: $curr_vol->{'vserver'}");
            $Log->trace("Volume:  $curr_vol->{'volume'}");
            $Log->trace("Destination volume: $curr_vol->{'curr_dest_volume'}");
            $Log->trace("Destination aggregate: " . 
                        $curr_vol->{'dest_aggr'}->name);
            eval {

                $dest_vol->copy(%copy_params);

            };

            if ($@) {

                $Log->debug("Copy failed: $@");
                $Log->debug("On failure flag is set to " .  $self->{'on_fail'});

                my $ex_text = "Copy failed: $@";
                my $status = $self->_gen_failure(ex_obj  => new VolOps::Copy::CopyFailed($ex_text), 
                                                 volume  => $curr_vol->{'volume'},
                                                 vserver => $curr_vol->{'vserver'},
                                                 fatal   => ($self->{'on_fail'} =~ /exit/i)
                                                            ? 1 : 0,
                                                );
            } else {

                #--------------------------------------------
                # Add this volume to the list of active jobs
                #--------------------------------------------
                $Log->comment("Adding $curr_vol->{'volume'} to the list of " .
                              "active jobs");
                my $key = "$curr_vol->{'volume'}:$curr_vol->{'vserver'}";
                $active_jobs{$key} = $curr_vol;

            }
        }

        #----------------------------------------------------------
        # Monitor the status of the current volume copy operations
        #----------------------------------------------------------
        $Log->trace("Monitoring the following jobs: ");
        foreach my $key (keys %active_jobs) {
            my %job = %{$active_jobs{$key}};
            $Log->trace("  Volume:                $job{'volume'}");
            $Log->trace("  Vserver:               $job{'vserver'}");
            $Log->trace("  Destination volume:    $job{'curr_dest_volume'}");
            $Log->trace("  Destination aggregate: " . $job{'dest_aggr'}->name);
            $Log->trace("  Job ID:                " . $self->{'vol_objs'}->{$job{'curr_dest_volume'}}->{$job{'vserver'}}->{'_job_id'} . "\n");
        }

        if (scalar keys %active_jobs) {

            $completed = $self->_monitor_jobs(vols      => \%active_jobs,
                                              on_fail   => $self->{'on_fail'},
                                              wait_type => $self->{'wait_type'},
                                             ); 

            $group_size = scalar keys %{$completed};
            $Log->comment("The following jobs completed: " . 
                         Dumper keys %{$completed});
        }

        my @vol_list = ();
        my @del_vols = ();

        #-----------------------------------------------------
        # Remove the completed jobs from the active jobs hash
        #-----------------------------------------------------
        foreach my $key (keys %{$completed}) {

            $Log->trace("Processing the completed job for $key");

            my %vol_info = %{$active_jobs{$key}};

            if (exists $active_jobs{$key}) {
                $Log->trace("Deleting job $key from active jobs list");
                delete $active_jobs{$key};
            }

            #----------------------------------------------------
            # If this volume was tagged for comparing attributes
            #----------------------------------------------------
            if ($vol_info{'compare_attrs'}) {                
                $Log->trace("Adding $vol_info{'volume'} to the list of " .
                            "volumes to compare");

                #------------------------------------------------
                # Make a copy of the volume info since we change
                # attributes below
                #------------------------------------------------
                my %comp_vol_info = %vol_info;
                push (@vol_list, \%comp_vol_info);
            }

            #------------------------------------------------------
            # If requested to delete volume after copy is complete
            #------------------------------------------------------
            if (! $vol_info{'keep_vols'}) {

                my %del_vol_info = ();

                #-----------------------------------------------------------
                # If cascading copied, delete the source volume; otherwise, 
                # delete the destination
                #-----------------------------------------------------------
                $del_vol_info{'volume'}  = ($vol_info{'cascade_copies'}) 
                                           ? $vol_info{'volume'}
                                           : $vol_info{'curr_dest_volume'};
                $del_vol_info{'vserver'} = $vol_info{'vserver'};

                push (@del_vols, \%del_vol_info);
            }

            #----------------------------------------------------------
            # If running test for an extended period of time, add this 
            # volume back into the queue
            #----------------------------------------------------------
            if ($self->{'duration'}) {
                if ($vol_info{'cascade_copies'}) {
                    $vol_info{'volume'} = $vol_info{'curr_dest_volume'};
                    delete $vol_info{'curr_dest_volume'};
                }

                $Log->trace("Iteration type is set to 'duration'. " .
                            "Pushing $vol_info{'volume'} back on the queue");

                push @vols, \%vol_info; 
                
            } 
        }

        #------------------------------------------------------------------
        # If any of the volumes are to have attribute comparison performed
        #------------------------------------------------------------------
        if (scalar @vol_list) {

            $Log->trace("Comparing volume properties for volumes " . 
                        Dumper \@vol_list);
            #---------------------------------------------------------
            # Indicate that we are performing a comparison after each
            # iteration so we use the correct key when obtaining the
            # volume attributes hash
            #---------------------------------------------------------
            $self->{'internal_compare'} = 1;

            #---------------------------------------------------------
            # An exception will be thrown if there were any attribute
            # miscomparisons and continue_on_failure is not set
            #---------------------------------------------------------
            my $compared = $self->compare_vol_properties(volumes  => \@vol_list,
                                                            );
            $self->{'internal_compare'} = 0;

        }

        #-----------------------------------------
        # Delete volumes if keep_vols was not set
        #-----------------------------------------
        if (scalar (@del_vols)) {

            foreach my $del_data (@del_vols) {
                
                $Log->debug("Deleting volume $del_data->{'volume'} since " .
                            "keep_vols was set to false");
                $self->{'server'}->volume_purge(volume  => $del_data->{'volume'}, 
                                                vserver => $del_data->{'vserver'},
                                               ); 
            }
        }

        my $end_time = time();
        my $elapsed  = $end_time - $start_time;
        $Log->trace("Elapsed time: $elapsed");
        $Log->trace("Duration: $self->{'duration'}");

        #-------------------------------------------------------
        # If a duration was specified and time is up, finish up
        #-------------------------------------------------------
        if (($self->{'duration'} > 0) && ($elapsed >= $self->{'duration'})) {

            $Log->trace("Iteration type is 'duration' and elapsed time " .
                        "exceeds specified duration.  Waiting for running " .
                        "jobs to complete");

            #----------------------------------------
            # Indicate that the duration has expired
            #----------------------------------------
            $expired = 1;

            my $left = scalar keys %active_jobs;

            $Log->comment("The list of active jobs is " . Dumper \%active_jobs);
            $Log->comment("The number of active jobs left are: $left");
            if ($left == 0) {
                $done = 1;
            } else {
                #----------------------------------------------------
                # Don't pick any more volumes to copy - go back 
                # to the top of the loop and iterate until there are
                # no more active jobs
                #----------------------------------------------------
                next;
            }
        }
    }

    $Log->exit();

} ## end sub _copy

################################################################################
#
# Subroutine:   _get_global_copy_param_spec 
#
# Objective:    Get the validation hash for the global copy parameters 
#
# Details:
#
#   Input:
#       None
#
#   Return value:
#      Validation spec to be used for global parameter validation 
#
################################################################################

sub _get_global_copy_param_spec {

    $Log->enter();

    my $self   = shift;

    #------------------------------------------------------
    # Get the validation hash for the common global params
    #------------------------------------------------------
    my $global_hash = $self->_get_common_global_param_spec();

    #----------------------------------------------------
    # Get the validation hash for the common copy params
    #----------------------------------------------------
    my $common_hash = $self->_get_common_copy_param_spec();

    #-----------------------------------------------------------
    # Combine the two hashes to get the complete list of params
    #-----------------------------------------------------------
    my $global_params = {
                         %{$global_hash},
                         %{$common_hash},
                        }; 

    $Log->debug("Contents of the global params validation hash: " . 
                Dumper $global_params);

    $Log->exit();

    return $global_params;
        
} ## end sub _get_global_copy_param_spec

################################################################################
#
# Subroutine:   _get_per_vol_copy_param_spec 
#
# Objective:    Get the validation hash for the per volume copy parameters 
#
# Details:
#
#   Input:
#      global_params - Global copy parameters, if provided.
#
#   Return value:
#      Validation spec to be used for per volume copy parameter validation 
#
################################################################################

sub _get_per_vol_copy_param_spec {

    $Log->enter();

    my $self   = shift;
    my $params = {};
    
    if (scalar @_) {
        $params = {@_};
    }

    $Log->debug("Params passed to _get_per_vol_copy_param_spec: " . 
                Dumper $params);

    #----------------------------------
    # Get the common per volume params
    #----------------------------------
#    my $per_vol_hash      = $self->_get_common_per_vol_param_spec(%{$params});
    my $per_vol_hash      = $self->_get_common_per_vol_param_spec($params);

    #----------------------------------------------------
    # Get the validation hash for the common copy params
    #----------------------------------------------------
#    my $common_copy_hash  = $self->_get_common_copy_param_spec(%{$params});
    my $common_copy_hash  = $self->_get_common_copy_param_spec($params);

    #----------------------------------------------------------
    # Combine the two hashes and add 'dest_volume'  to get the 
    # complete list of params
    #----------------------------------------------------------
    my $per_vol_params = {
                          %{$per_vol_hash},
                          %{$common_copy_hash},
                          dest_volume       => {type      => SCALAR,
                                                optional  => 1,
                                                default   => undef,
                                               },
                         };

    $Log->debug("Contents of the per volume copy params validation hash " .
                Dumper $per_vol_params);

    $Log->exit();

    return $per_vol_params;

} ## end sub _get_per_vol_copy_param_spec

################################################################################
#
# Subroutine:   _get_common_copy_param_spec
#
# Objective:    Get the validation hash for the common volume copy parameters 
#
# Details:
#
#   Input:
#      global_params - Global copy parameters, if provided.
#
#   Return value:
#      Validation spec consisting of the common volume copy parameters 
#
################################################################################

sub _get_common_copy_param_spec {

    $Log->enter();

    my $self = shift;

    my $params = {};
    
    if (scalar $@) {
        $params = {@_};
    }

    $Log->debug("Parameters passed to _get_common_copy_param_spec: " . 
                Dumper $params);

    #-----------------------------------------------------
    # Generate a hash consisting of the common parameters
    #-----------------------------------------------------
    my $common_params = { 
        keep_vols      => {type      => BOOLEAN,
                           optional  => 1,
                           default   => (exists $params->{'global_params'}) 
                                         ? $params->{'global_params'}->{'keep_vols'}
                                         : 1,
                          }, 
        cascade_copies => {type      => BOOLEAN,
                           optional  => 1,
                           default   => (exists $params->{'global_params'})
                                         ? $params->{'global_params'}->{'cascade_copies'}
                                         : 0,
                          }, 
    };

    $Log->debug("Contents of the common copy params validation hash: " . 
                Dumper $common_params); 

    $Log->exit();

    return $common_params;

} ## end sub _get_common_copy_param_spec

=head2 compare_vol_properties

C<< VolOps::Copy->compare_vol_properties(%params) >>

Perform a basic comparison of volume properties after the copy has completed.
All attributes in each specified volume are compared.  Attributes whose values
do not contain the expected values are logged.  To ensure that all attributes
are properly validated, an exception is thrown when an attribute is encountered for which no validation rule exists.

=over 4

=item Input

=over 4

=item * volumes => <arrayref>

(Optional)
Default = List of volumes supplied when the object was instantiated.

List of volumes to compare.  Array consists of hashes which must include the
following keys for each volume to compare:

 volume
 vserver

=item * match_fields => <arrayref>

(Optional)
Default = Predefined list of matching attributes.

List of volume attributes whose pre- and post-copy values must match.

=item * diff_fields => <arrayref>

(Optional)
Default = Predefined list of attributes whose values must be different.

List of volume attributes whose pre- and post-copy values must be different.

=item * skip_fields => <arrayref>

(Optional)
Default = Predefined list of attributes to skip.

List of volume attributes whose pre- and post-copy values should be skipped
when comparing because it cannot be determined whether or not they will match.

=item * match_val => <hashref>

(Optional)
Default = Predefined hash of attributes that must be set to a specific value.

List of volume attributes whose values must be set to a specific value after a
copy.

The format of the hash is:

 (<attribute> => "<match value>",
  ...
 ); 

=back 4

=item Output

Returns 1 if all volumes compare; returns 0 if there were any miscomparisons.

=item Example

C<< $vcp->compare_vol_properties(volumes => \@vol_list); >>

=back

=cut

################################################################################
#
# Subroutine:   compare_vol_properties()
#
# Objective:    Compares a list of volume properties after a copy has
#               completed.
#
# Details:
#
#   Input:
#      volumes        - List of volumes to compare.  Optional; defaults to
#                       the list of volumes specified during instantiation
#      match_fields   - List of volume attributes whose pre- and post-copy 
#                       values must match
#      diff_fields    - List of volume attributes whose pre- and post-copy
#                       values must differ
#      skip_fields    - List of volume attributes to skip when doing attribute
#                       comparison because it cannot be predetermined if they
#                       will match
#      match_val      - List of volume attributes whose values must be set to
#                       a specific value after a copy
#      default_action - Default action to take when comparison rules are not
#                       provided for an attribute.  Default is 'exit'.
#
#   Return value:
#      Returns 1 if volume properties compare; returns 0 if there were any
#      miscomparisons.
#
################################################################################

sub compare_vol_properties {

    $Log->enter();

    my $self = shift();

    my $params;

    my $validate_spec = $self->_get_attr_compare_spec();

    eval {
        $params = {
                      validate (
                                @_,
                                $validate_spec,
                              )
                  };    
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for compare_vol_properties: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    my $errors      = 0;
    my %comp_params = ();
    my @comp_vols   = (); 
    my $status;

    #---------------------------------------------------------------------------
    # If job_status is not available, call get_status to obtain the destination 
    # aggregates for each of the specified volumes 
    #---------------------------------------------------------------------------
    if (!exists $self->{'job_status'}) {

        $Log->trace("Getting job status to obtain destination aggregate data " .
                    "for volume property comparison");
        $status = $self->get_status(volumes => $params->{'volumes'});

    } else {

        $Log->trace("Job status: " . Dumper $self->{'job_status'});
        $status = $self->{'job_status'};
    }

    #-------------------------------------------------------------------------
    # Go through the list of volume to compare.  If destination volume was
    # not provided, get the information out of the volume objects.  If
    # destination volume is not available, do not include this volume in the 
    # list for comparison
    # NOTE: compare_vol_properties expects the destination volume key to be 
    #       'curr_dest_volume'
    #-------------------------------------------------------------------------
    foreach my $vol_data (@{$params->{'volumes'}}) {

        if (!exists ($vol_data->{'dest_volume'})) {
            if (exists $status->{$vol_data->{'vserver'}}->{$vol_data->{'volume'}}->{'curr_dest_volume'}) {
                $vol_data->{'curr_dest_volume'} = $status->{$vol_data->{'vserver'}}->{$vol_data->{'volume'}}->{'curr_dest_volume'};
                push @comp_vols, $vol_data;
            } else {
                $Log->warn("Destination volume name not available for volume " .
                           $vol_data->{'volume'} . ".  Volume will not be compared");
            }
        } else {
            #-------------------------------------------------------------------
            # If we are doing a compare between copies, use the current value 
            # for curr_dest_volume; otherwise use the destination volume passed 
            # in by the caller
            #-------------------------------------------------------------------
            if (!$self->{'internal_compare'}) {
                $vol_data->{'curr_dest_volume'} = $vol_data->{'dest_volume'};
            }
            push @comp_vols, $vol_data;
        }
    }
$Log->comment("Comparison volumes: "  . Dumper \@comp_vols);

    $comp_params{'volumes'}        = \@comp_vols; 
    $comp_params{'default_action'} = $params->{'default_action'}; 
    $comp_params{'match_fields'}   = (exists $params->{'match_fields'}) ?
                                      $params->{'match_fields'} :
                                      $self->{'match_fields'};
    $comp_params{'diff_fields'}    = (exists $params->{'diff_fields'}) ?
                                      $params->{'diff_fields'} :
                                      $self->{'diff_fields'};
    $comp_params{'skip_fields'}    = (exists $params->{'skip_fields'}) ?
                                      $params->{'skip_fields'} :
                                      $self->{'skip_fields'};
    $comp_params{'match_val'}      = (exists $params->{'match_val'}) ? 
                                      $params->{'match_val'} :
                                      $self->{'match_val'};

    #------------------------
    # Compare the attributes
    #------------------------
    $errors = $self->SUPER::compare_vol_properties(%comp_params,
                                                  ); 

    $Log->exit();

    return $errors;

} ## end sub compare_vol_properties 

1;
