# $Id: //depot/prod/test/Rsierranevada/lib/VolOps/Move.pm#1 $

## @summary     Programmatic control of volume move.
## @author      dl-volops-qa@netapp.com
## @status      private
## @pod         here

=head1 NAME

VolOps::Move.pm

=head1 SYNOPSIS

  use VolOps::Move;
  use VolOps::BaseException;
  use strict;
  
  my %global_params = (
          parallel_count   => "all",
          volumes          => ["testvol1",
                               "testvol2",
                               "testvol3",
                               "testvol4",
                              ],
          vserver          => "vs0",
          cutover_attempts => 5,
          cutover_window   => 30,
          compare_attrs    => 1,
          on_failure       => "continue",
          );
  
  # Override the global parameters with the following
  my @per_vol_params = (
          {
              volume            => "testvol1",
              aggr_loc_affinity => "remote",
              aggr_list         => ["csiqa_3070_33_aggr2",
                                    "csiqa_3070_34_aggr1",
                                   ],
          },
          {
              volume            => "testvol2",
              vserver           => "vs1",
              compare_attrs     => 0,
              cutover_attempts  => 1,
          },
          {
              volume            => "testvol3",
              aggr_loc_affinity => "local",
          },
          );
  
  my $vmv = new VolOps::Move(node                => "csiqa-3070-33",
                             engine              => "BRE",
                             per_vol_move_params => \@per_vol_params,
                             global_move_params  => \%global_params,
                             save_status         => 1,
                            );
  
  # Move a volume
  my $sub_obj = $vmv->move();
  
  my $status;
  
  # Get status while volumes are moving
  while ($vmv->status()) {
      $status = $vmv->get_status();
      snooze(30);
      $Log->comment("Status: " . $status->{"formatted"});
  }
  
  # Get the status of the moves
  $status = $vmv->get_status();
  
  $Log->comment("Status of all runs: " . $status->{'formatted'});
  
  if (($status->{"total_failed"} == 0) &&
      ($status->{"total_errors"} == 0)) {
      $Log->comment("All moves passed");
  }

=head1 DESCRIPTION

C<VolOps::Move> is a class for performing volume move testing.

The class throws exceptions on failures.

This class makes use of its own exception class
VolOps::BaseException. While the exception class will be
inherited when using this library, it is recommended to explicitly import
the exception class for clarity.

The current list of possible exceptions are:
    NATE::Exceptions::Argument
    VolOps::BaseException
    VolOps::ConfigError
    VolOps::JobFailed
    VolOps::JobStatusFailure
    VolOps::AttributeComparison
    VolOps::MissingValidationRule
    VolOps::RetriesExhausted
    VolOps::SrcVolumeFailure
    VolOps::SrcAggregateFailure
    VolOps::AggregateSelectionFailure
    VolOps::DestAggregateFailure
    VolOps::PropsNotAvailable
    VolOps::StripingNotSupported
    VolOps::ConnectionFailure
    VolOps::DataStructureStoreFailure
    VolOps::InternalError
    VolOps::Move::MoveFailed
    VolOps::Move::TriggerFailed
    VolOps::Move::NotDeferred

=head1 METHODS

=cut

package VolOps::Move;
use strict;
use warnings;
use base qw(VolOps);
use Params::Validate qw(validate validate_with BOOLEAN SCALAR HASHREF ARRAYREF UNDEF OBJECT);
use Tharn qw(fatal_result config_result fail_result pass_result $Staf $Log $Parent snooze subtest);
use Data::Dumper;
use NATE::BaseException qw(:try);
use VolOps::BaseException;
use GNOTE;
use Node;
use Date::Parse;
use NATE::Events qw(listener_add listener_remove);

my $Options = new Options();

=head2 new

C<< VolOps::Move->new(%params) >>

Instantiate an object to use for moving volume(s).

=over 4

=item Input

=over 4

=item * node => <scalar>

(Required)

Identifies one of the nodes in the cluster to be used to attach to the
cluster.

=back

=over 4

=item * engine => <scalar>

(Optional)
Default = "BRE"

Paloma engine to be used for the transfer.  Sets the value of
I<paloma.physical> to B<true> if set to "BRE" and sets it to B<false> if set to
"LRE".  The existing value of paloma.physical is saved then later restored
when the object is destroyed.

Valid values:
    BRE - Use physical replication
    LRE - Use logical replication

=back

=over 4

=item * %global_move_params => <hashref>

(Required)

Parameters to be used for volume move operations for each of the volumes
specified.  Parameters consist of the following:

=over 4

=item - vserver => <scalar>

(Optional, required if no per_vol_move parameters are defined)
Default = undef

Name of the vserver to use for all volume operations.

=item - volumes => <arrayref>

(Required)
Default = undef

List of volumes on which to operate.

=item - aggr_list => <arrayref>

(Optional)
Default = undef

List of destination aggregates from which to select when moving volumes.
If not specified, aggregates associated with the nodes specified by the NODES
parameter on the command line will be used.

=item - aggrtype_affinity => <scalar>

(Optional)
Default = any

Block type for the destination aggregate.

Valid values:
    32         - Select a 32-bit aggregate as the
                 destination.
    64         - Select a 64-bit aggregate as the
                 destination.
    any        - Select an aggregate with any block type.
    alternate  - Select an aggregate with a block type that
                 does not match the current block type.

=item - aggr_loc_affinity => <scalar>

(Optional)
Default = "remote"

Location of the destination aggregate with respect to the source node.

Valid values:
    local      - Select an aggregate on the same node as
                 the aggregate on which the volume resides.
    remote     - Select an aggregate on a different node.
    any        - Randomly select an aggregate on either a
                 local or a remote node.
    alternate  - Alternate between "remote" and "local",
                 starting with "remote".

=item - iteration_mode => <scalar>

(Optional)
Default = "iteration"

Type of continuity to use for execution of the volume move(s).

Valid values:
    iteration - Iterate through the list of volumes the
                specified number of times.
    duration  - Continue moving the list of volumes for
                the specified duration.

=item - iteration_value => <scalar>

(Optional)
Default = 1 if iteration_mode is I<iteration>; "undef" if iteration_mode is
I<duration>

Length of time or number of iterations to perform the move operation.  If
iteration_mode is I<duration>, I<iteration_value> can be appended with
H or h for hours, D or d for days, M or m for minutes, S or s for seconds.
Default granularity is minutes.

=item - parallel_count => <scalar>

(Optional)
Default = 1

Number of volumes to move in parallel.  A value of "1" will result in serial moves.

Valid values:
    all
    <number>

=item - wait_type => <scalar>

(Optional)
Default = "any"

Indicates whether to wait for all volumes in a group to finish moving before
starting the next move or to just wait until any of the volumes finishes
before starting the next volume move operation.  The number of items in a
group is defined by the value of parallel_count.

Valid values:
    any - Wait until at least one job completes before
          starting the next move.
    all - Wait until all jobs in the current group have
          completed moving before starting the next group.
          The size of the group is defined by the
          I<parallel_count> parameter.

=item - bg_mode => <scalar>

(Optional)
Default = "bg"

Indicates whether to run the volume move(s) in the foreground or background.
If run in the foreground, B<move> will block until complete.

Valid values:
    fg - Run the moves in the foreground and block until
         complete.
    bg - Start the moves in the background and return
         immediately.

=item - cutover_window => <scalar>

(Optional)
Default = undef

Specifies the value to pass to the volume move operation for the
cutover-window parameter.  If not specified, no cutover-window parameter will
be passed to the volume move command, resulting in a default of 45 seconds
being used.

=item - cutover_attempts => <scalar>

(Optional)
Default = undef

Specifies the value to pass to the volume move operation for the
cutover-attempts parameter.  If not specified, no cutover-attempts parameter
will be passed to the volume move command, resulting in a default of 3 being
used.

=item - cutover_action => <scalar>

(Optional)
Default = undef

Specifies the value to pass to the volume move operation for the
cutover-action parameter. If not specified, no cutover-action parameter will be
passed to the volume move command, resulting in a default of "defer_on_failure"
being used.

=item - auto_trigger => <scalar>

(Optional)
Default = off

Specifies whether or not the trigger-cutover command should be automatically
executed when a volume move job enters a deferred state.  Valid values are:

 off     - Do not automatically trigger a cutover
 force   - Automatically trigger cutover with the force
           option set to true
 noforce - Automatically trigger cutover with the force
           option set to false

=item - timeout => <scalar>

(Optional)
Default = 0

Timeout value for the move operation.  A value of 0 will wait forever.

=item - on_failure => <scalar>

(Optional)
Default = "exit"

Indicates what action to perform when an error is encountered while moving a volume.

Valid values:
    continue - Record any non-abortable failures and
               continue with any remaining moves.
    complete - Allow all volumes currently moving to
               complete before throwing an exception.
    exit     - Immediately throw an exception on failure.
               The status of any ongoing move jobs will be
               lost.

=item -compare_attrs => <boolean>

(Optional)
Default = 1 (true)

Indicates whether or not to compare volume attributes (pre- and post-move) after each
move completes.  The comparison will be performed every time a volume or group of volumes
has completed a move.

=back

=item * @per_vol_move_params => <arrayref>

(Optional)
Default = undef

Array of hashes that provides parameters to use for moving each specified
volume.  Any values supplied override those specified in the global move params hash.
Any volumes specified must also be included in the global_move_params volumes
array.

=over 4

=item - vserver => <scalar>

(Optional, required if no global_move_params are defined)
Default = undef

Name of the vserver to use for the specific volume operation.  Overrides any
vserver value set in global_move_params.

=item - volume => <scalar>

(Required)
Default = undef

Volume to move.  Volume must be included in the global_move_params volumes
array.

=item - aggr_list => <arrayref>

(Optional)
Default = undef

List of destination aggregates to select from.  If not specified, aggregates
associated with nodes specified by the NODES param are considered.  Overrides
any aggr_list value set in global_move_params for a specific volume.

=item - aggrtype_affinity => <scalar>

(Optional)
Default = any

Block type for the destination aggregate.

Valid values:
    32         - Select a 32-bit aggregate as the destination.
    64         - Select a 64-bit aggregate as the destination.
    any        - Select an aggregate with any block type.
    alterate   - Select an aggregate with a block type that
                 does not match the current block type.

=item - aggr_loc_affinity => <scalar>

(Optional)
Default = "remote"

Location of the destination aggregate with respect to the source node.
Overrides any aggr_loc_affinity value set in global_move_params.

Valid values:
    local      - Select an aggregate on the same node as
                 the aggregate on which the volume resides.
    remote     - Select an aggregate on a different node.
    any        - Randomly select an aggregate on either a
                 local or a remote node.
    alternate  - Alternate between "remote" and "local",
                 starting with "remote".

=item - cutover_window => <scalar>

(Optional)
Default = undef

Specifies the value to pass to the volume move operation for the cutover-window
parameter.  If not specified, no cutover-window parameter will
be passed to the volume move command, resulting in a default of 45 seconds
being used.

=item - cutover_attempts => <scalar>

(Optional)
Default = undef

Specifies the value to pass to the volume move operation for the
cutover-attempts parameter.  If not specified, no cutover-attempts parameter
will be passed to the volume move command, resulting in a default of 3 being
used.

=item - cutover_action => <boolean>

(Optional)
Default = undef

Specifies the value to pass to the volume move operation for the cutover-action
parameter. If not specified, no cutover-action parameter will be passed to the
volume move command, resulting in a default of "defer_on_failure" being used.

=item - compare_attrs => <boolean>

(Optional)
Default = 1 (true)

Indicates whether or not to compare volume attributes (pre- and post-move)
after each move completes.  The comparison will be performed every time the
volume has completed a move.

=back

=item * save_status => <boolean>

(Optional)
Default = false

Indicates whether or not to save the status hash for each iteration.  If this
flag is set, the status hash will be saved, by default, as a text file to the
LOGDIR using file name "Vol[Move or Copy]_status_<timestamp>.dd".  The save
mode, filename, and/or directory can be modified on the command line using the
DSS_MODE, DSS_DIR, DSS_FILE, and DSS_FILENAME parameters.  See the POD for
lib/DataStructureStore.pm for more information.

=back

=item Output

Reference to the object.

=item Example

 my $sub_obj = $vmv->new(node                => csiqa-3070-99,
                         engine              => "BRE",
                         per_vol_move_params => \@per_vol_params,
                         global_move_params  => \%global_move_params,
                         save_status         => 1,
                        );

=back

=item * create_nate_result => <scalar>

(Optional)
Default = 'yes' 

Indicates whether or not to create a nate result when the move fails.  
By default, a NATE fail result will be generated when an error or failure
occurs.  By setting this value to 'no', a NATE warn result will be generated
to provide a warning that something went wrong.  When run in the background,
the NATE result (warn or fail) does not get propagated back to the parent.
This flag is useful for negative testing when a failure is expected.

=back

=item Output

Reference to the object.

=item Example

 my $sub_obj = $vmv->new(node                => csiqa-3070-99,
                         engine              => "BRE",
                         per_vol_move_params => \@per_vol_params,
                         global_move_params  => \%global_move_params,
                         save_status         => 1,
                        );

=back

=cut

################################################################################
#
# Subroutine:   new()
#
# Objective:    Invocant for the class
#
# Details:
#
#   Input:
#       node                - Node in the cluster
#       engine              - Paloma engine to be used for the transfer
#                             (BRE or LRE)
#       global_move_params  - Move params that will be applied to all volumes
#                             being moved
#       per_vol_move_params - Move params that will be applied only to
#                             individual volumes
#       save_status         - Indicates whether or not to save the status hash
#                             for each iteration.  If this flag is set, the
#                             status hash will be saved, by default, as a text
#                             file to the LOGDIR using file name
#                             "Vol[Move|Copy]_status_<timestamp>.dd".  The save
#                             mode, filename, and/or directory can be modified
#                             on the command line using the DSS_MODE, DSS_DIR,
#                             DSS_FILE, and DSS_FILENAME parameters.  See the
#                             POD for lib/DataStructureStore.pm for more
#                             information.
#                             Optional; default = 0
#
#   Return value:
#       object
#
################################################################################

sub new {
    $Log->enter();

    my $class = shift;

    my (%common_params, %move_params);

    #--------------------------------------------------------------------------
    # Split the params into ones needed by base class and move-specific params
    #--------------------------------------------------------------------------
    my %args = @_;
    foreach my $key (keys %args) {

        if ($key =~ /global_move_params|per_vol_move_params/) {
            $move_params{$key} = $args{$key};
        } else {
            $common_params{$key} = $args{$key};
        }
    }

    my $self  = $class->SUPER::new(\%common_params);

    #--------------------
    # Process parameters
    #--------------------
    $self->_process_move_params(\%move_params);

    my ($maj, $min) = $self->_get_release();

    #-------------------------------
    # If release is pre-Boilermaker
    #-------------------------------
    if ($maj lt 8) {

        VolOps::ConfigError->throw("Library not supported for " .
                                   "pre-Boilermaker releases");

    } else {

        @{$self->{'match_fields'}} = (@{$self->{'match_fields'}},
                                      'volume',
                                      'msid',
                                      'type',
                                      'junction-path',
                                      'junction-path-source',
                                      'junction-active',
                                      'junction-parent',
                                      'parent',
                                      'create-time',
                                      'crtime',
                                      'load-sharing-source',
                                      'is-inconsistent',
                                     );

        #---------------------------
        # If RollingRock.0 or later
        #---------------------------
        if ($maj ge 8.1) {
            $Log->trace("Build is RollingRock or later");
            #------------------------------------------------------------------
            # Append fields that have common validation rules for physical and
            # logical
            #------------------------------------------------------------------
            @{$self->{'match_fields'}} = (@{$self->{'match_fields'}},
                                          'fractional-reserve',
                                          'filesys-size-fixed',
                                          'is-copied-for-transition',
                                          'is-transitioned',
                                          'instance-uuid',
                                          'provenance-uuid',
                                         );

            @{$self->{'skip_fields'}} = (@{$self->{'skip_fields'}},
                                          'overwrite-reserve',
                                        );

            #-------------------------------------------
            # If the physical replication engine is set
            #-------------------------------------------
            if ($self->{'engine'} =~ /BRE/) {
                $Log->trace("Setting physical replication-specific comparison rules");
                #----------------------
                # Physical replication
                #----------------------
                @{$self->{'match_fields'}} = (@{$self->{'match_fields'}},
                                              'physical-replica',
                                              'transition-type',
                                              'transition-state',
                                              'transition-history',
                                              'nvfail',
                                              'is-nvfail',
                                             );
            } else {
                #---------------------
                # Logical replication
                #---------------------
                $Log->trace("Setting logical replication-specific comparison rules");
                @{$self->{'skip_fields'}} = (@{$self->{'skip_fields'}},
                                             'transition-type',
                                             'transition-state',
                                             'transition-history',
                                             'nvfail',
                                             'is-nvfail',
                                            );

                %{$self->{'match_val'}}  = (%{$self->{'match_val'}},
                                            'physical-replica' => 'false',
                                           );
            }
        }

        $Log->debug("Move match fields: " . Dumper $self->{'match_fields'});
        $Log->debug("Move diff fields: " . Dumper $self->{'diff_fields'});
        $Log->debug("Move skip fields: " . Dumper $self->{'skip_fields'});
        $Log->debug("Move match value: " . Dumper $self->{'match_val'});

    }

    $self->{'loc_affins'} = ['local',
                             'remote',
                            ];

    $self->{'total_deferrals'} = 0;
    $self->{'triggered'}       = 0; # This is used by get_status to determine
                                    # whether or not a cutover was triggered
                                    # If it is set, the status should be obtained
                                    # directly, not from the subtest object.

    $Log->exit();

    return $self;

} ## end sub new

################################################################################
#
# Subroutine:   _process_move_params
#
# Objective:    Process the global_move_params and per_vol_move_params
#
# Details:
#
#   Input:
#      global_move_params
#      per_vol_move_params
#
#   Return value:
#       None
#
################################################################################

sub _process_move_params {

    $Log->enter();

    my $self   = shift;

    my $params;

    my $validate_spec = $self->_get_common_param_spec(@_);

    eval {

        $params = {
                      validate (
                                   @_,
                                   $validate_spec,
                             )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed in _process_move_params: $@";

        $Log->error(fail_result($ex_text));
        VolOps::ConfigError->throw($ex_text);
    }

    $Log->debug("Move parameters: " . Dumper $params);

    #----------------------------------------------
    # Process the global and per volume parameters
    #----------------------------------------------
    $self->_process_global_per_vol_params($params);

    $Log->trace("Parallel_count: " . $self->{'parallel_count'});
    $Log->trace("Wait type: " . $self->{'wait_type'});
    $Log->trace("Save status: " . $self->{'save_status'});
    $Log->trace("BG mode: " . $self->{'bg_mode'});
    $Log->trace("Iterations: " . $self->{'iterations'});
    $Log->trace("Duration: " . $self->{'duration'});
    $Log->trace("Timeout: " . $self->{'timeout'});
    $Log->trace("On failure: " . $self->{'on_fail'});

    $Log->exit();

} ## end sub _process_move_params

=head2 move

C<< VolOps::Move->move() >>

Move the volume(s) to the specified aggregate or to an aggregate of a
specified type.

=over 4

=item Input

None

=item Output

If executed in the background, a reference to the subtest object associated
with the move operation is returned.  If executed in the foreground, nothing
is returned.

=item Example

 my $sub_obj = $vmv->move();

=back

=cut

################################################################################
#
# Subroutine:   move()
#
# Objective:    Move the volume(s) per specifications.  If background mode of
#               'bg' was specified during instantiation, move is done in the
#               background, otherwise; it is a blocking operation.
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Returns the subtest object if move is being executed in the background;
#      otherwise, nothing is returned.
#
################################################################################

sub move {

    $Log->enter();

    my $self = shift();

    $self->{'job_status'} = ();

    #----------------------------------------------------------
    # If running in the background, subtest the move operation
    #----------------------------------------------------------
    if ($self->{'bg_mode'} eq '-bg') {

        $self->{'subtested'} = 1;
        $self->{'st_obj'}    = subtest(\&VolOps::_perform_vol_ops,
                                       $self->{'bg_mode'},
                                       '-timeout', $self->{'timeout'},
                                       '-runid', "move_volumes_" . time(),
                                       "-onexit", sub {},
                                       '--', $self,
                                      );
        snooze(1);

    } else {

        #-----------------------
        # Run in the foreground
        #-----------------------

        $self->_perform_vol_ops;

    }

    $Log->exit();

    ($self->{'bg_mode'} eq '-bg') ? return $self->{'st_obj'} : return;

} ## end sub move

################################################################################
#
# Subroutine:   _move()
#
# Objective:    Move the list of specified volumes to their selected
#               destination aggregates.  Moves will be performed in parallel
#               if indicated.
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      None
#
################################################################################

sub _move {

    $Log->enter();

    my $self = shift;

    #-----------------------
    # Attach to each volume
    #-----------------------
    my @vols = $self->_prep_vols();

    my @move_vols   = ();
    my $done        = 0;
    my $expired     = 0;
    my %active_jobs = ();
    my $start_time  = time();
    my $group_size  = $self->{'parallel_count'};

    #-----------------------------------------------------------------
    # Continue moving volumes until duration has elapsed or specified
    # iterations are complete
    #-----------------------------------------------------------------
    while (!$done) {

        $Log->debug("The content of vols is: " . Dumper \@vols);

        my $completed = ();
        my $deferred  = ();

        #--------------------------------------
        # If there are volumes left to process
        #--------------------------------------
        if ((scalar @vols) && (!$expired) && ($group_size > 0)) {
            #--------------------------------
            # Get the next volume(s) to move
            #--------------------------------
            $Log->debug("Splicing $group_size volumes from the list of " .
                        "volumes");
            @move_vols = splice(@vols, 0, $group_size);

            $Log->debug("Moving the following volumes: ");
            foreach my $vol (@move_vols) {
                $Log->debug("\t$vol->{'volume'}");
            }

            #--------------------------------------------------------------
            # Check to see if there is an active move job that is deferred 
            # then select a new aggregate
            #--------------------------------------------------------------
            foreach my $vol (@move_vols) {

                $Log->trace("Checking deferral status of volume $vol->{'volume'}");

                $self->_precheck_deferral_status(vol_info => $vol,
                                                );

                my $vol_obj = $self->{'vol_objs'}->{$vol->{'volume'}}->{$vol->{'vserver'}};
                $vol->{'src_aggr'}  = $vol_obj->Aggregate->name;
                $vol->{'dest_aggr'} = $self->_select_aggr(vol_info => $vol);
            }

        } else {
            NOTE::logcomment("There are no more volumes left to move");
            @move_vols = ();

            #-----------------------------------------------
            # If there are no active jobs left, we are done
            #-----------------------------------------------
            my $left = scalar keys %active_jobs;
            $Log->comment("The number of remaining active jobs is: " . $left);
            if ($left == 0) {
                $Log->debug("Iteration complete.");
                $done = 1;
                last;
            }
        }

        #----------------------------------------------
        # Move each of the volumes in the current list
        #----------------------------------------------
        foreach my $curr_vol (@move_vols) {

            my $volume  = $curr_vol->{'volume'};
            my $vserver = $curr_vol->{'vserver'};

            $Log->trace("Preparing to move $volume");

            my $vol = $self->{'vol_objs'}->{$volume}->{$vserver};

            my %move_params = ();

            $move_params{'volume'}     = $volume;
            $move_params{'vserver'}    = $vserver;
            $move_params{'to_aggr'}    = $curr_vol->{'dest_aggr'};
            $move_params{'background'} = 1;

            if (defined $curr_vol->{'cutover_attempts'}) {
                $move_params{'cutover-attempts'} = $curr_vol->{'cutover_attempts'};
            }

            if (defined $curr_vol->{'cutover_window'}) {
                $move_params{'cutover-window'} = $curr_vol->{'cutover_window'};
            }

            if (defined $curr_vol->{'cutover_action'}) {
                $move_params{'cutover-action'} = $curr_vol->{'cutover_action'};
            }

            $Log->trace("Moving $volume with the following " .
                        "parameters:" . Dumper \%move_params);

            eval {

                $vol->move(%move_params);
    
            };

            #--------------------
            # If the move failed
            #--------------------
            if ($@) {

                $Log->debug("Move failed: $@");
                $Log->debug("On failure flag is set to " .  
                            $self->{'on_fail'});

                my $ex_text = "Move failed: $@";
                my $status  = $self->_gen_failure(ex_obj  => new VolOps::Move::MoveFailed($ex_text),
                                                  volume  => $volume,
                                                  vserver => $vserver,
                                                  fatal   => ($self->{'on_fail'} =~ /exit/i)
                                                             ? 1 : 0,
                                                 );

                #--------------------------------------------------------------
                # If continue_on_failure is set, select another volume to move
                #--------------------------------------------------------------
                if ($self->{'on_fail'} =~ /continue/i) {
                    my $vol = splice(@vols, 0, 1); 
                    
                    #------------------------------------
                    # Select a new destination aggregate
                    #------------------------------------
                    if (scalar keys %{$vol}) {
                        $self->_precheck_deferral_status(vol_info => $vol,
                                                        );
                        my $vol_obj = $self->{'vol_objs'}->{$vol->{'volume'}}->{$vol->{'vserver'}};
                        $vol->{'src_aggr'}  = $vol_obj->Aggregate->name;
                        $vol->{'dest_aggr'} = $self->_select_aggr(vol_info => $vol);
                        push (@move_vols, $vol);
                    } else {
                        my $ex_text = "Move failed for volume $volume ($vserver) and there are no more " .
                                      "volumes available to select from";
                        my $status  = $self->_gen_failure(ex_obj  => new VolOps::Move::MoveFailed($ex_text),
                                                          volume  => $volume,
                                                          vserver => $vserver,
                                                          fatal   => 0, 
                                                         );
                    }                        
                }

            } else {

                #--------------------------------------------
                # Add this volume to the list of active jobs
                #--------------------------------------------
                $Log->comment("Adding $volume ($vserver) to the list of " .
                              "active jobs");
                my $key = "$volume:$vserver";
                $active_jobs{$key} = $curr_vol;

                #------------------------------------------------------------
                # Initialize the deferral/trigger attempts counters for this
                # job
                #------------------------------------------------------------
                $self->{'job_status'}->{$vserver}->{$volume}->{'deferrals'}->{$vol->{'_job_id'}}->{'trigger_attempts'} = 0;
                $self->{'job_status'}->{$vserver}->{$volume}->{'deferrals'}->{$vol->{'_job_id'}}->{'deferral_cnt'}     = 0;
            }
        }

        #----------------------------------------------------------
        # Monitor the status of the current volume move operations
        #----------------------------------------------------------
        try {
            if (scalar keys %active_jobs) {
                ($completed,$deferred) = $self->_monitor_jobs(vols      => \%active_jobs,
                                                              on_fail   => $self->{'on_fail'},
                                                              wait_type => $self->{'wait_type'},
                                                             );
            }
            $group_size = scalar keys %{$completed};
            $Log->comment("The following jobs completed: " .
                         Dumper keys %{$completed});

            if (scalar keys %{$deferred}) {
                $Log->comment("The following jobs are deferred: " .
                              Dumper keys %{$deferred});
            }

        } catch VolOps::Move::MoveDeferred with {

            $Log->trace("Move job is deferred and exit_on_deferral was " .
                        "specified");
            my $ex = shift;
            $done  = 1;

            #-----------------------
            # Rethrow the exception
            #-----------------------
            $Log->trace("Move job is deferred.  Throwing exception since " .
                        "exit_on_deferral was set");

            $ex->throw();

        } catch VolOps::BaseException with {

            $Log->trace("A VolOps::BaseException occurred.  Throwing the exception");

            my $ex      = shift;
            my $ex_text = "VolOps exception occurred while monitoring jobs: $ex->text()";
            $ex->throw($ex_text);

        } otherwise {

            $Log->trace("An unknown exception occurred.  Throwing the exception");

            my $ex      = shift;
            my $ex_text = "Unknown exception occurred while monitoring jobs: $ex->text()";
            $ex->throw($ex_text);
        };

        my @vol_list = ();

        #-----------------------------------------------------
        # Remove the completed jobs from the active jobs hash
        #-----------------------------------------------------
        foreach my $key (keys %{$completed}) {

            $Log->trace("Processing the completed job for $key");

            my %vol_info = %{$active_jobs{$key}};

            if (exists $active_jobs{$key}) {
                $Log->trace("Deleting job $key from active jobs list");
                delete $active_jobs{$key};
            }

            if ($vol_info{'compare_attrs'}) {
                $Log->trace("Adding $vol_info{'volume'} to the list of " .
                            "volumes to compare");
                push (@vol_list, \%vol_info);
            }

            #----------------------------------------------------------
            # If running test for an extended period of time, add this
            # volume back into the queue
            #----------------------------------------------------------
            if ($self->{'duration'}) {
                $Log->trace("Iteration type is set to 'duration'. " .
                            "Pushing $vol_info{'volume'} back on the queue");
                push @vols, \%vol_info;
            }
        }

        #-----------------------------------------------------
        # Remove the deferred jobs from the active jobs hash
        #-----------------------------------------------------
        foreach my $key (keys %{$deferred}) {

            $Log->trace("Processing the deferred job for $key");

            if (exists $active_jobs{$key}) {
                $Log->trace("Deleting deferred job $key from active jobs list");
                delete $active_jobs{$key};
            }
        }

        #------------------------------------------------------------------
        # If any of the volumes are to have attribute comparison performed
        #------------------------------------------------------------------
        if (scalar @vol_list) {

            $Log->trace("Comparing volume properties for volumes " .
                        Dumper \@vol_list);
            $self->{'internal_compare'} = 1;

            #---------------------------------------------------------
            # An exception will be thrown if there were any attribute
            # miscomparisons and continue_on_failure is not set
            #---------------------------------------------------------
            my $compared = $self->compare_vol_properties(volumes  => \@vol_list,
                                                        );
            $self->{'internal_compare'} = 0;

        }

        my $end_time = time();
        my $elapsed  = $end_time - $start_time;
        $Log->trace("Elapsed time: $elapsed");
        $Log->trace("Duration: $self->{'duration'}");

        #-------------------------------------------------------
        # If a duration was specified and time is up, finish up
        #-------------------------------------------------------
        if (($self->{'duration'} > 0) && ($elapsed >= $self->{'duration'})) {

            $Log->trace("Iteration type is 'duration' and elapsed time " .
                        "exceeds specified duration.  Waiting for running " .
                        "jobs to complete");

            #----------------------------------------
            # Indicate that the duration has expired
            #----------------------------------------
            $expired = 1;

            my $left = scalar keys %active_jobs;

            $Log->comment("The list of active jobs is " . Dumper \%active_jobs);
            $Log->comment("The number of active jobs left are: $left");
            if ($left == 0) {
                $done = 1;
            } else {
                #----------------------------------------------------
                # Don't pick any more volumes to move - go back
                # to the top of the loop and iterate until there are
                # no more active jobs
                #----------------------------------------------------
                next;
            }
        }
    }

    $Log->exit();

} ## end sub _move

################################################################################
#
# Subroutine:   _precheck_deferral_status
#
# Objective:    Check to see if there is an existing volume move job for this
#               volume that is deferred.  If there is and auto_trigger is set
#               for this volume, attempt to trigger cutover and wait for the
#               job to complete.
#
# Details:
#
#   Input:
#       vol_info - Hash containing the volume move parameters for the volume
#
#   Return value:
#       None
#       A NATE::Exceptions::Argument exception is thrown if there is a problem
#       validating parameters.
#
################################################################################

sub _precheck_deferral_status {

    $Log->enter();

    my $self = shift();

    my $params;

    eval {
        $params = {
                      validate (
                                @_,
                                {
                                    vol_info => {type => HASHREF,
                                                },
                                }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for " .
                      "_precheck_deferral_status: $@";
        my $status  = $self->_gen_failure(ex_obj => new NATE::Exceptions::Argument($ex_text),
                                         );
    }

    my $volume       = $params->{'vol_info'}->{'volume'};
    my $vserver      = $params->{'vol_info'}->{'vserver'};
    my $auto_trigger = $params->{'vol_info'}->{'auto_trigger'};
    my $deferred     = 0;
    
    try {
        $deferred = $self->is_deferred(volume => $volume,
                                       vserver => $vserver,
                                     );
    
    } catch VolOps::BaseException with {
        $Log->trace("No active move jobs were found for volume " . 
                    "$volume ($vserver)");
    } otherwise {
        my $ex = shift;
        $ex->throw($ex->text());
    }; 
    
    #-------------------------------------------------------------------
    # Trigger cutover if a move is deferred and auto_trigger is enabled
    #-------------------------------------------------------------------
    if ($deferred) {

        my $job = $self->_get_volume_move_job_data(volume  => $volume,
                                                   vserver => $vserver,
                                                  );
        if ($auto_trigger !~ /off/i) {
            NOTE::logcomment("Volume $volume ($vserver) was found to be " .
                             "deferred before the move.  Triggering cutover " .
                             "because auto_trigger was set");

            #-----------------
            # Trigger cutover
            #-----------------
            try {
                $Log->trace("Triggering cutover for $volume ($vserver) - " .
                            "job $job->{'id'}");
                $self->_trigger_cutover(volume  => $volume,
                                        vserver => $vserver,
                                        force   => ($params->{'vol_info'}->{'auto_trigger'} =~ /noforce/)
                                                   ? 0 : 1,
                                        job_id  => $job->{'id'},
                                       );
    
                my $job_obj;
    
                #------------------------------
                # Wait until cutover completes
                #------------------------------
                eval {
                    $job_obj = NOTE::GX::Job->new(Cluster => $self->{'cluster'},
                                                  id      => $job->{'id'},
                                                 );
                };
    
                if ($@) {
                    my $ex_text = "Unable to create a job object to trigger cutover: $@";
    
                    if ($self->{'on_fail'} !~ /continue/i) {
                        $Log->error(fail_result($ex_text));
                        VolOps::ConfigError->throw($ex_text,
                                                  );
                    } else {
                        $Log->warn($ex_text);
                    }
                } else {
    
                    eval {
                        $job_obj->wait();
                    };
    
                    #------------------------------------------------------
                    # If we didn't timeout waiting for the job to complete
                    #------------------------------------------------------
                    if ($@ !~ /^Timeout/i) {
                        my $ex_text = "Failure while waiting for trigger-cutover " .
                                      "to complete (job $job->{'id'})";
                        if ($self->{'on_fail'} !~ /continue/i) {
                            $Log->error(fail_result($ex_text));
                            VolOps::ConfigError->throw($ex_text,
                                                      );
                        }
                    } elsif ($@ =~ /^Timeout/i) {
                        $Log->trace("Timed out while waiting for trigger-cutover " .
                                    "to complete (job $job->{'id'})");
                    }
                }

            } catch VolOps::BaseException with {
                my $ex = shift;
                my $ex_text = "Volume $volume ($vserver) was found to be in " .
                              "a deferred state before attempting a move.  " .
                              "auto_trigger was set but trigger_cutover " .
                              "failed: $ex->text()";
                $Log->warn($ex_text);
                my $status = $self->_gen_failure(ex_obj  => $ex,
                                                 volume  => $volume,
                                                 vserver => $vserver,
                                                 fatal   => $self->{'on_fail'} =~ /continue/i
                                                           ? 0 : 1,
                                                );
            };
        } else {

            $Log->warn ("Volume $volume ($vserver) was found to be in a " .
                        "deferred state before attempting a move but " .
                        "auto_trigger was not set.  Trigger_cutover will " .
                        "not be attempted"
                       );

            if ($params->{'vol_info'}->{'exit_on_deferral'}) {

                $Log->trace("Volume was found in a deferred before " .
                            "attempting move and exit_on_deferral is set.  " .
                            "Throwing exception");
                my $ex_text = "Move job $job->{'id'} for $volume ($vserver) " .
                              "is deferred";
                VolOps::Move::MoveDeferred->throw($ex_text,
                                                 );
            }
        }
    }

    $Log->exit();

} ## end sub _precheck_deferral_status

=head2 is_deferred

C<< VolOps::Move->is_deferred(%params) >>

Determined whether or not there is a deferred move job for the
specified volume.

=over 4

=item Input

=over 4

=item * volume => <scalar>

(Required)

Volume to check for a deferred move job.

=item * vserver => <scalar>

(Required)

Vserver on which the volume resides.

=item * job => <scalar|hashref>

(Optional)
Default = *

Either the job ID for the volume move job or reference to a hash containing
the job statistics for the move job.  If not specified, all running jobs will
be examined.

=back

=item Output

$self->HARD_DEFERRED if a move job for the specified volume/vserver
is currently hard deferred
$self->SOFT_DEFERRED if a move job for the specified volume/vserver
is currently soft deferred
0 if a move job was found but the move is not currently deferred
Throws exception VolOps::Move::NoMoveJob job progress is not available

=item Example

 $vmv->is_deferred(volume  => 'testvol',
                   vserver => 'vs0',
                   job     => $job_stat,
                  );

=back

=cut

################################################################################
#
# Subroutine:   is_deferred
#
# Objective:    Determine whether or not the specified volume/vserver has a
#               running volume move job that is currently deferred.
#
# Details:
#
#   Input:
#       volume  - Volume to check for a deferred move job (Required)
#       vserver - Vserver on which the volume resides (Required)
#       job     - Job id or job status hash for the volume move job.  If not
#                 specified, will check all running volume move jobs.
#
#   Return value:
#       $self->HARD_DEFERRED if a move job for the specified volume/vserver
#       is currently hard deferred
#       $self->SOFT_DEFERRED if a move job for the specified volume/vserver
#       is currently soft deferred
#       0 if a move job was found but the move is not currently deferred
#       Throws exception VolOps::Move::NoMoveJob job progress is not available 
#
################################################################################

sub is_deferred {

    $Log->enter();

    my $self = shift();

    my $deferred = 0;

    my $params;

    eval {

        $params = {
                      validate (
                                   @_,
                                   {
                                       volume  => {type     => SCALAR,
                                                  },
                                       vserver => {type     => SCALAR,
                                                  },
                                       job     => {type     => SCALAR|HASHREF,
                                                   optional => 1,
                                                   default  => '*',
                                                  },
                                   }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for is_deferred: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    #----------------------------------------------------------------------
    # Find the job associated with current volume move for this volume if
    # it exists
    #----------------------------------------------------------------------
    my $job;

    if (ref $params->{'job'} eq 'HASH') {
        $job = $params->{'job'};
    } else {
        try {
            $job = $self->_get_volume_move_job_data(volume  => $params->{'volume'},
                                                    vserver => $params->{'vserver'},
                                                    job_id  => $params->{'job'},
                                                   );
        } catch VolOps::BaseException with {
            my $ex = shift;
            $Log->comment("An error occurred while trying to get " .
                          "volume move job data: " . $ex->text());
        };
    }

    if (scalar keys %{$job}) {

        my $progress = $job->{'progress'};

        #-------------------------------------------
        # If there is a job for this volume/vserver
        #-------------------------------------------
        if (length ($progress)) {

            $progress =~ s/"//g;

            my $hard_def_string = $self->_HARD_DEFERRAL_STRING;
            my $def_string      = $self->_DEFERRAL_STRING;

            if ($progress =~ /^$hard_def_string/i) {
                $Log->trace("Job is hard deferred");
                $deferred = $self->HARD_DEFERRED;
            } elsif ($progress =~ /^$def_string/i) {
                $Log->trace("Job is soft deferred");
                $deferred = $self->SOFT_DEFERRED;
            }
        } else {
            $Log->comment("Job progress not available for $job->{'id'}");
            my $ex_text = "Job progress not available for job $job->{'id'}";
            VolOps::Move::NoMoveJob->throw($ex_text);
        }
    }

    $Log->debug("Returning a value of $deferred from is_deferred");

    $Log->exit();

    return $deferred;

} ## end sub is_deferred 

################################################################################
#
# Subroutine:   _get_volume_move_job_data 
#
# Objective:    Find a running move job associated with the specified volume/
#               vserver and return the statistics associated with the job
#
# Details:
#
#   Input:
#       volume  - Volume to check for a deferred move job (Required)
#       vserver - Vserver on which the volume resides (Required)
#       job_id  - ID of the job associated with the move.  Default = '*'
#                 (Optional)
#
#   Return value:
#       Returns a reference to a hash containing the job data for the job
#       Throws NATE::Exceptions::Argument if there was a problem with
#       parameter validation.
#
################################################################################

sub _get_volume_move_job_data {

    $Log->enter();

    my $self = shift();

    my $params;

    eval {

        $params = {
                      validate (
                                   @_,
                                   {
                                       volume  => {type     => SCALAR,
                                                  },
                                       vserver => {type     => SCALAR,
                                                  },
                                       job_id  => {type     => SCALAR,
                                                   optional => 1,
                                                   default  => '*',
                                                  },
                                   }
                               )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for _get_volume_move_job_data: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    my $job_stats = ();
    my $job       = ();

    eval {
        $job_stats = $self->{'server'}->job_show(name  => '"Volume Move"',
                                                 state => "Running",
                                                 id    => $params->{'job_id'},
                                                );
    };

    if ($@) {
        $Log->trace("Job show failed to find a running Volume Move job with " .
                    "job ID $params->{'job_id'}");
    } else {
        $Log->debug("Output from job show: " . Dumper $job_stats);

        #-------------------------------------------------------------
        # Check for a running volume move job for this volume/vserver
        #-------------------------------------------------------------
        $Log->trace("Searching for a volume move job for $params->{'volume'} " .
                      "($params->{'vserver'})");
        if (scalar @{$job_stats}) {
            foreach my $job_data (@{$job_stats}) {
                $Log->trace("Job data for current job: " . Dumper $job_data);

                my ($vol,$vs);

                $vol =  $job_data->{'description'};
                $vol =~ s/^"Move (.*) in .*$/$1/;
                $vs  =  $job_data->{'description'};
                $vs  =~ s/^"Move .* in vserver (.*) to .*$/$1/;
    
                $Log->trace("Move Job Volume: $vol");
                $Log->trace("Move Job Vserver: $vs");
    
                if (($vol =~ /^$params->{'volume'}$/) && 
                    ($vs =~ /^$params->{'vserver'}$/)) {
                    #--------------------------------------------
                    # The job has been located -- stop searching
                    #--------------------------------------------
                    $Log->trace("A matching job was located");
                    $job = $job_data;
                    last;
                }
            }
        } else {
            $Log->comment("No running move jobs were found for " .
                          "$params->{'volume'} ($params->{'vserver'}");
        }
    }

    $Log->exit();

    return $job;

} ## end sub _get_volume_move_job_data

################################################################################
#
# Subroutine:   HARD_DEFERRED
#
# Objective:    Returns a value indicating that a job is hard deferred
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Returns a value representing a hard-deferred job
#
################################################################################

sub HARD_DEFERRED {

    $Log->enter();

    my $self = shift;

    $Log->exit();

    return 2;

} ## end sub HARD_DEFERRED

################################################################################
#
# Subroutine:   SOFT_DEFERRED
#
# Objective:    Returns a value indicating that a job is soft deferred
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Returns a value representing a soft-deferred job
#
################################################################################

sub SOFT_DEFERRED {

    $Log->enter();

    my $self = shift;

    $Log->exit();

    return 1;

} ## end sub SOFT_DEFERRED

################################################################################
#
# Subroutine:   _HARD_DEFERRAL_STRING
#
# Objective:    Returns the progress string that is displayed when a volume
#               move job is hard deferred
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Returns the hard deferral progress string
#
################################################################################

sub _HARD_DEFERRAL_STRING {

    my $self = shift;

    return 'Cutover Deferred, Waiting for user intervention';

} ## end sub _HARD_DEFERRAL_STRING

################################################################################
#
# Subroutine:   _DEFERRAL_STRING
#
# Objective:    Returns the progress string that is displayed when a volume
#               move job is deferred 
#
# Details:
#
#   Input:
#      None
#
#   Return value:
#      Returns an the deferral progress string 
#
################################################################################

sub _DEFERRAL_STRING {

    my $self = shift;

    return 'Cutover Deferred';

} ## end sub _DEFERRAL_STRING


=head2 trigger_cutover 

C<< VolOps::Move->trigger_cutover(%params) >>

Execute the volume move trigger-cutover command for the list of 
specified volumes.

=over 4

=item Input

=over 4

=item * volumes => <arrayref>

(Required)

List of volumes for which to trigger cutover.  Array consists of hashes 
which must include the following keys:

 volume
 vserver

=item * force => <boolean>

(Optional)
Default = 0 

Indicates whether or not the cutover should be forced if time guarantee
may not be met.

=back

=item Output

None.  A VolOps::Move::TriggerFailed exception is thrown if the trigger-cutover
command fails.

=item Example

 $vmv->trigger_cutover(volumes => \@vol_list);

=back

=cut

################################################################################
#
# Subroutine:   trigger_cutover 
#
# Objective:    Trigger cutover for one or more specified volumes. A warning
#               is generated if volumes are specified that aren't being 
#               internally tracked.
#
# Details:
#
#   Input:
#       volumes - Array containing hashes that define the volume/vserver for
#                 which to trigger cutover.  (Required)
#       force   - Indicates whether or not force cutover (Optional; default = 0)
#
#   Return value:
#       None
#       A VolOps::Move::TriggerFailed exception is thrown if there is a 
#       failure triggering cutover.
#
################################################################################

sub trigger_cutover {

    $Log->enter();

    my $self = shift;

    $self->{'triggered'} = 1;

    my $deferred_vols;

    #--------------------------------------------------------------
    # If the move was run in the foreground or subtest is complete
    #--------------------------------------------------------------
    if ($self->status == 0) {
        my ($params, $volumes);

        eval {

            $params = {
                          validate (
                                       @_,
                                       {
                                           volumes  => {type     => ARRAYREF,
                                                        optional => 1,
                                                        default  => $self->{'src_vols'},
                                                       },
                                           force    => {type     => BOOLEAN,
                                                        optional => 1,
                                                        default  => 0,
                                                       },
                                       }
                                 )
                      };
        };

        if ($@) {
            my $ex_text = "Parameter validation failed for trigger_cutover: $@";
            $Log->error(fail_result($ex_text));
            NATE::Exceptions::Argument->throw($ex_text);
        }

        #-------------------------------------------
        # Trigger cutover for all specified volumes
        #-------------------------------------------
        foreach my $curr_vol (@{$params->{'volumes'}}) {

            my @vol_params  = shift(@{$params->{'volumes'}});

            eval {

                $volumes = {
                               validate_with (
                                              params      => \@vol_params,
                                              spec        =>    {
                                                                 volume  => {type => SCALAR,
                                                                            },
                                                                 vserver => {type => SCALAR,
                                                                            },
                                                                },
                                              allow_extra => 1,
                                             ),
                           };
            };

            if ($@) {
                my $ex_text = "Parameter validation failed for trigger_cutover volumes parameter: $@";
                $Log->error(fail_result($ex_text));
                NATE::Exceptions::Argument->throw($ex_text);
            }

            my $volume   = $curr_vol->{'volume'};
            my $vserver  = $curr_vol->{'vserver'};
            my $deferred = 0;

            try {
                $deferred = $self->is_deferred(volume  => $volume,
                                               vserver => $vserver,
                                              );
            } catch VolOps::BaseException with {
                $Log->trace("There is no deferred move job associated " .
                            "with volume $volume ($vserver)");
            } otherwise {
                my $ex = shift;
                $ex->throw($ex->text());
            };

            #--------------------------------------------
            # If the move job is deferred - hard or soft
            #--------------------------------------------
            if ($deferred) {

                my $job = $self->_get_volume_move_job_data(volume  => $volume,
                                                           vserver => $vserver,
                                                          );

                my $job_id = $job->{'id'};

                eval {
                    $self->_trigger_cutover(volume  => $volume,
                                            vserver => $vserver,
                                            force   => $params->{'force'},
                                           );
                };

                #-----------------------------------------------
                # If there was a failure triggering the cutover
                #-----------------------------------------------
                if ($@) {

                    my $ex_text = "Trigger cutover failed for volume $volume ($vserver): $@";
                    if ($self->{'on_fail'} !~ /continue/) {
                        VolOps::Move::TriggerFailed->throw($ex_text);
                    } else {
                        $Log->warn($ex_text . "  Continuing because " .
                                   "continue_on_failure was set");
                    }
                } else {

                    my $done = 0;

                    #-------------------------------------------------------
                    # Wait until the move job completes or becomes deferred
                    #-------------------------------------------------------
                    while (!$done) {

                        $Log->trace("Waiting for move job to complete or defer");

                        my $job = $self->_get_job_status(volume  => $volume,
                                                         vserver => $vserver,
                                                         job_id  => $job_id,
                                                        );

                        $Log->debug("Job status: " . Dumper $job);

                        #-------------------------------------
                        # If job has reached a terminal state
                        #-------------------------------------
                        if ($job->{'state'} !~ /running/i) {

                            $done = 1;

                            if ($self->{'total_deferrals'} > 0) {
                                $self->{'total_deferrals'}--;
                            }

                            #--------------------------------------
                            # If there are no remaining deferrals,
                            # indicate that processing is complete
                            #--------------------------------------
                            if ($self->{'total_deferrals'} == 0) {
                                $self->{'end_time'} = time();
                            }
                        } else {

                            #---------------------------------
                            # Check to see if job is deferred
                            #---------------------------------
                            my $deferred = $self->is_deferred(volume  => $volume,
                                                              vserver => $vserver,
                                                              job     => $job,
                                                             );

                            if ($deferred == $self->HARD_DEFERRED) {
                                $done = 1;
                                $self->{'total_deferrals'}++;
                                $deferred_vols .= "Move job $job_id for " .
                                                  "$volume ($vserver) is " .
                                                  "deferred";

                            } else {
                                snooze(5);
                            }
                        }
                    }

                    $self->_update_job_status(volume  => $volume,
                                              vserver => $vserver,
                                             );

                }

            } else {
                $Log->trace("No active move job to trigger cutover for " .
                            "$volume ($vserver)");
                my $ex_text = "Trigger cutover was requested for volume " .
                              "$volume ($vserver) but the job is not deferred";
                if ($self->{'on_fail'} !~ /continue/) {
                    VolOps::Move::NotDeferred->throw($ex_text);
                } else {
                    $Log->warn($ex_text);
                }
            }                 
        }
    } else {

        $Log->trace("VolOps::Move::move appears to still be running.  Volume " .
                    "move cutover cannot be triggered while move is still running");
    }

    #-------------------------------------
    # If any move jobs are still deferred
    #-------------------------------------
    if ($deferred_vols) {
        VolOps::Move::MoveDeferred->throw($deferred_vols);
    }

    $Log->exit();

    return;

} ## end sub trigger_cutover

################################################################################
#
# Subroutine:   _trigger_cutover 
#
# Objective:    Trigger cutover for the specified volume and increment the
#               trigger attempts for this job. 
#
# Details:
#
#   Input:
#       volume  - Volume being moved (Required)
#       vserver - Vserver on which the volume resides (Required)
#       force   - Indicates whether or not to force cutover (Boolean)
#                 (Optional; default = 0)
#       job_id  - Job ID associated with the volume move (Required)
#
#   Return value:
#       None
#       A VolOps::Move::TriggerFailed exception is thrown if there is a 
#       failure triggering cutover.
#
################################################################################

sub _trigger_cutover {

    $Log->enter();

    my $self = shift;

    my $params;

    eval {

        $params = {
                      validate (
                                   @_,
                                   {
                                       volume   => {type     => SCALAR,
                                                   },
                                       vserver  => {type     => SCALAR,
                                                   },
                                       force    => {type     => BOOLEAN,
                                                    optional => 1,
                                                    default  => 0,
                                                   },
                                       job_id   => {type     => SCALAR,
                                                    optional => 1,
                                                   },
                                   }
                             )
                  };
    };

    my ($job_data, $job_id);

    #--------------------------------------------------------------------
    # If a job ID was not specified, find a running move job associated
    # with this volume/vserver
    #--------------------------------------------------------------------
    if (! exists $params->{'job_id'}) {

        $job_data = $self->_get_volume_move_job_data(volume => $params->{'volume'},
                                                     vserver => $params->{'vserver'},
                                                    );
        if (scalar keys %{$job_data}) {
            $job_id = $job_data->{'id'};
        } else {
            my $ex_text = "Unable to locate running move job for $params->{'volume'} " .
                          " ($params->{'vserver'}).";

            my $status  = $self->_gen_failure(ex_obj  => new VolOps::Move::NoMoveJob($ex_text),
                                              volume  => $params->{'volume'},
                                              vserver => $params->{'vserver'},
                                              fatal   => ($self->{'on_fail'} =~ /exit/i)
                                                         ? 1 : 0,
                                             );
        }

    } else {

        $job_id = $params->{'job_id'};
    }

    #---------------------
    # Trigger the cutover
    #---------------------
    eval {
        $self->{'server'}->volume_move_trigger_cutover(volume  => $params->{'volume'},
                                                       vserver => $params->{'vserver'},
                                                       force   => ($params->{'force'}) ? 'true' : 'false',
                                                      );

    };

    if ($@) {
        my $ex_text = "Trigger cutover failed: $@";
        my $status  = $self->_gen_failure(ex_obj  => new VolOps::Move::TriggerFailed($ex_text),
                                          volume  => $params->{'volume'},
                                          vserver => $params->{'vserver'},
                                          fatal   => ($self->{'on_fail'} =~ /exit/i)
                                                     ? 1 : 0,
                                         );
    } else {

        #------------------------------------------------------------
        # Increment the trigger attempts for this job and delete the
        # deferred job information until it becomes deferred again
        #------------------------------------------------------------
        $Log->trace("Updating trigger attempts for job $job_id");
        $self->{'job_status'}->{$params->{'vserver'}}->{$params->{'volume'}}->{'deferrals'}->{$job_id}->{'trigger_attempts'}++;

        delete $self->{'job_status'}->{$params->{'vserver'}}->{$params->{'volume'}}->{'deferred_job'};

        #-------------------------------------
        # If any deferrals have been recorded
        #-------------------------------------
        if ($self->{'total_deferrals'} > 0) {
            $self->{'total_deferrals'}--;
        }
    }

    $Log->exit();

    return;

} ## end sub _trigger_cutover

################################################################################
#
# Subroutine:   _get_global_move_param_spec
#
# Objective:    Get the validation hash for the global move parameters
#
# Details:
#
#   Input:
#       None
#
#   Return value:
#      Validation spec to be used for global parameter validation
#
################################################################################

sub _get_global_move_param_spec {

    $Log->enter();

    my $self   = shift;

    #------------------------------------------------------
    # Get the validation hash for the common global params
    #------------------------------------------------------
    my $global_hash = $self->_get_common_global_param_spec();

    #----------------------------------------------------
    # Get the validation hash for the common move params
    #----------------------------------------------------
    my $common_hash = $self->_get_common_move_param_spec();

    #-----------------------------------------------------------
    # Combine the two hashes to get the complete list of params
    #-----------------------------------------------------------
    my $global_params = {
                         %{$global_hash},
                         %{$common_hash},
                        };

    $Log->debug("Contents of the global params validation hash: " .
                Dumper $global_params);

    $Log->exit();

    return $global_params;

} ## end sub _get_global_move_param_spec

################################################################################
#
# Subroutine:   _get_per_vol_move_param_spec
#
# Objective:    Get the validation hash for the per volume move parameters
#
# Details:
#
#   Input:
#      global_params - Global move parameters, if provided.
#
#   Return value:
#      Validation spec to be used for per volume move parameter validation
#
################################################################################

sub _get_per_vol_move_param_spec {

    $Log->enter();

    my $self   = shift;
    my $params = {};

    if (scalar @_) {
        $params = {@_};
    }

    $Log->debug("Params passed to _get_per_vol_move_param_spec: " .
                Dumper $params);

    #----------------------------------
    # Get the common per volume params
    #----------------------------------
#    my $per_vol_hash      = $self->_get_common_per_vol_param_spec(%{$params});
    my $per_vol_hash      = $self->_get_common_per_vol_param_spec($params);

    #----------------------------------------------------
    # Get the validation hash for the common move params
    #----------------------------------------------------
    my $common_move_hash  = $self->_get_common_move_param_spec(%{$params});

    #-----------------------------------------------------------
    # Combine the two hashes to get the complete list of params
    #-----------------------------------------------------------
    my $per_vol_params = {
                          %{$per_vol_hash},
                          %{$common_move_hash},
                         };

    $Log->debug("Contents of the per volume move params validation hash " .
                Dumper $per_vol_params);
$Log->comment("Contents of the per volume move params validation hash " .
             Dumper $per_vol_params);

    $Log->exit();

    return $per_vol_params;

} ## end sub _get_per_vol_move_param_spec

################################################################################
#
# Subroutine:   _get_common_move_param_spec
#
# Objective:    Get the validation hash for the common volume move parameters
#
# Details:
#
#   Input:
#      global_params - Global move parameters, if provided.
#
#   Return value:
#      Validation spec consisting of the common volume move parameters
#
################################################################################

sub _get_common_move_param_spec {

    $Log->enter();

    my $self = shift;

    my $params = {};

    if (scalar @_) {
        $params = {@_};
    }

    $Log->debug("Parameters passed to _get_common_move_param_spec: " .
                Dumper $params);

    #-----------------------------------------------------
    # Generate a hash consisting of the common parameters
    #-----------------------------------------------------
    my $common_params = {
        cutover_window    => {type      => SCALAR | UNDEF,
                              optional  => 1,
                              default   => (exists $params->{'global_params'})
                                            ? $params->{'global_params'}->{'cutover_window'}
                                            : undef,
                              regex     => qr/^\d+$/,
                             },
        cutover_attempts  => {type      => SCALAR | UNDEF,
                              optional  => 1,
                              default   => (exists $params->{'global_params'})
                                            ? $params->{'global_params'}->{'cutover_attempts'}
                                            : undef,
                              regex     => qr/^\d+$/,
                             },
        cutover_action    => {type      => SCALAR | UNDEF,
                              optional  => 1,
                              default   => (exists $params->{'global_params'})
                                            ? $params->{'global_params'}->{'cutover_action'}
                                            : undef,
                              regex     => qr/^(abort_on_failure|defer_on_failure|wait|force)$/,
                             },
        auto_trigger      => {type      => SCALAR,
                              optional  => 1,
                              default   => (exists $params->{'global_params'})
                                           ? $params->{'global_params'}->{'auto_trigger'}
                                           : 'off',
                              regex     => qr/^(off|force|noforce)$/,
                             },
        exit_on_deferral  => {type      => BOOLEAN,
                              optional  => 1,
                              default   => 0,
                             }, 
    };

    $Log->exit();

    return $common_params;

} ## end sub _get_common_move_param_spec

=head2 compare_vol_properties

C<< VolOps::Move->compare_vol_properties(%params) >>

Perform a basic comparison of volume properties after the move has completed.
All attributes in each specified volume are compared.  Attributes whose values
do not contain the expected values are logged.  To ensure that all attributes
are properly validated, an exception is thrown when an attribute is encountered for which no validation rule exists.

=over 4

=item Input

=over 4

=item * volumes => <arrayref>

(Optional)
Default = List of volumes supplied when the object was instantiated.

List of volumes to compare.  Array consists of hashes which must include the
following keys for each volume to compare:

 volume
 vserver

=item * match_fields => <arrayref>

(Optional)
Default = Predefined list of matching attributes.

List of volume attributes whose pre- and post-move values must match.

=item * diff_fields => <arrayref>

(Optional)
Default = Predefined list of attributes whose values must be different.

List of volume attributes whose pre- and post-move values must be different.

=item * skip_fields => <arrayref>

(Optional)
Default = Predefined list of attributes to skip.

List of volume attributes whose pre- and post-move values should be skipped
when comparing because it cannot be determined whether or not they will match.

=item * match_val => <hashref>

(Optional)
Default = Predefined hash of attributes that must be set to a specific value.

List of volume attributes whose values must be set to a specific value after a
move.

The format of the hash is:

 (<attribute> => "<match value>",
  ...
 );

=back

=item Output

Returns 1 if all volumes compare; returns 0 if there were any miscomparisons.

=item Example

 $vmv->compare_vol_properties(volumes => \@vol_list);

=back

=cut

################################################################################
#
# Subroutine:   compare_vol_properties()
#
# Objective:    Compares a list of volume properties after a move has
#               completed.
#
# Details:
#
#   Input:
#      volumes      - List of volumes to compare.  Optional; defaults to
#                     the list of volumes specified during instantiation
#      match_fields - List of volume attributes whose pre- and post-move
#                     values must match
#      diff_fields  - List of volume attributes whose pre- and post-move
#                     values must differ
#      skip_fields  - List of volume attributes to skip when doing attribute
#                     comparison because it cannot be predetermined if they
#                     will match
#      match_val    - List of volume attributes whose values must be set to
#                     a specific value after a move
#
#   Return value:
#      Returns 1 if volume properties compare; returns 0 if there were any
#      miscomparisons.
#
################################################################################

sub compare_vol_properties {

    $Log->enter();

    my $self = shift();

    my $params = ();

    my $validate_spec = $self->_get_attr_compare_spec();

    eval {
        $params = {
                      validate (
                                @_,
                                $validate_spec,
                              )
                  };
    };

    if ($@) {
        my $ex_text = "Parameter validation failed for compare_vol_properties: $@";
        $Log->error(fail_result($ex_text));
        NATE::Exceptions::Argument->throw($ex_text);
    }

    my $errors      = 0;
    my %comp_params = ();

    $comp_params{'volumes'}        = $params->{'volumes'};
    $comp_params{'default_action'} = $params->{'default_action'};
    $comp_params{'match_fields'}   = (exists $params->{'match_fields'}) ?
                                      $params->{'match_fields'} :
                                      $self->{'match_fields'};
    $comp_params{'diff_fields'}    = (exists $params->{'diff_fields'}) ?
                                      $params->{'diff_fields'} :
                                      $self->{'diff_fields'};
    $comp_params{'skip_fields'}    = (exists $params->{'skip_fields'}) ?
                                      $params->{'skip_fields'} :
                                      $self->{'skip_fields'};
    $comp_params{'match_val'}      = (exists $params->{'match_val'}) ?
                                      $params->{'match_val'} :
                                      $self->{'match_val'};

    my $status;

    #---------------------------------------------------------------------------
    # If job_status is not available, call get_status to obtain the destination
    # aggregates for each of the specified volumes
    #---------------------------------------------------------------------------
    if (!exists $self->{'job_status'}) {
        $Log->trace("Getting job status to obtain destination aggregate data " .
                    "for volume property comparison");
        $status = $self->get_status(volumes => $params->{'volumes'});
    }

    #------------------------
    # Compare the attributes
    #------------------------
    $errors = $self->SUPER::compare_vol_properties(%comp_params,
                                                  );

    $Log->exit();

    return $errors;

} ## end sub compare_vol_properties

1;
