# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Volume/Clone.pm#2 $

## @summary Library of volume clone commands
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @status public
## @pod here


=head1 NAME

Clone - Volume Cloning Functions

=head1 SYNOPSIS

    use Volume::Clone qw(:all);

    my $h = host($FILER);
    my $conn = connect("#rsh");

    $ret = vol_clone_create(conn=>$conn, parent=>vol1, clone=>clonevol
                        [,spcres=>file] [,snapname=>hourly.0]
                        [,purge_existing=>1]);

    $ret = vol_clone_split_start(conn=>$conn, clone=>vol2
                    [,wait_to_complete=>1] [,validate=>1] [,timeout=>1800]);

    $ret = vol_clone_split_stop(conn=>$conn, clone=>vol2);
                    [,wait_to_complete=>1] [,validate=>1] [,timeout=>1800])

    $ret = vol_clone_split_complete(conn=>$conn, clone=>vol2
                    [,validate=>1] [,timeout=>1800]);

    %status = vol_clone_split_status(conn=>$conn [,clone=>vol2] [,timeout=>200])

    @list = get_clones(conn=>$conn, vol=>vol0);

    ($parent, $base_snapshot) = get_clone_info(conn=>$conn, clone=>vol2)

    $ret = verify_clone(conn=>$conn, clone=>vol2 [,expected_snapname=>foo]
                        [,data_compare=>1] [,compare_timeout=>200])


=head1 DESCRIPTION

This module defines routines related to OnTap volume clone creation.

Volume cloning was introduced in R7.0 / Anchorsteam.
So the OnTap software required is Anchorsteam or later.
Also, cloning is a new capability that applies exclusively to
flexible volumes.

=cut


package Volume::Clone;
use strict;
use base qw(Exporter);
BEGIN
{
  our %EXPORT_TAGS = (
    all => [ qw( vol_clone_create 
                vol_clone_split_start 
                vol_clone_split_stop 
		vol_clone_split_complete
                vol_clone_split_status 
                get_clones
                get_clone_info
                verify_clone )
           ] );
  Exporter::export_ok_tags('all');
}

use Tharn qw(say logresult logcomment tstlog timeout2time);
use Params::Validate qw(validate SCALAR OBJECT BOOLEAN);
use Carp qw(croak);
use Volume qw(vol_exists vol_container vol_online vol_offline vol_destroy
            vol_options vol_purge);
use SayWrap qw(sayEOF);
use fileSystem qw(treecompare_vol);
use SnapCmds qw(snapshot_delete snapsched_volume);

use constant MSG_NO_SPLIT => "Clone volume is not being split from its parent";
use constant MSG_GEN_NO_SPLIT => "No clone status";
use constant ERR_NOT_CLONE => "The volume is not a clone";
use constant ERR_NOT_FLEX => "is not a flexible volume";
use constant ERR_NOT_FOUND => "Volume .* was not found";



=head2 vol_clone_create

This routine will create a clone of a flexible volume. The clone volume
will reside in the same aggregate as the parent volume.

    vol_clone_create (conn=>$conn, parent=>vol1, clone=>clonevol
                        [,spres=>file] [,snapname=>hourly.0 ]
                        [,purge_existing=>1]);

=over

=item Arguments

This routine takes in the arguments in the key value pairs.
The arguments and their description are as follows:

=over

=item conn

(Required) : Connection object to the filer.

=item parent

(Required) : Name of parent volume whose clone is to be made. This has
to be a online flexible volume.  If the parent volume is currently 
in a clone split operation another clone creation will fail.

=item clone

(Required) : Name of clone volume.

=item snapname

(Optional) : Name of existing snapshot of parent volume which will serve
as the basis of the clone volume.  If not specified then a new snapshot 
will be created by OnTap which will be as the basis for the clone.

=item spcres

(Optional) : Space reservation setting for the clone volume
Possible values : none / file / volume

=item purge_existing

(Optional) : Purge existing clone and create a new one

=back

=back

=cut

sub vol_clone_create ( % ) {

    my (%args) = validate (@_, { 
                             conn => { type=>OBJECT, isa=>'Connectrec' },
                             clone => { type=>SCALAR },
                             parent => { type=>SCALAR },
                             snapname => { type=>SCALAR, optional=>1 },
                             spcres => { type=>SCALAR, optional=>1, regex=>qr/none|file|volume/},
                             purge_existing => { type=>BOOLEAN, optional=>1 },
                             force_clone => { type=>BOOLEAN, optional=>1, default=>0 },
                            }
                          );

    return eval {

    # if clone already exists then purge if asked to or return
    if (Volume::vol_exists($args{conn}, $args{clone}) ) {
        if ($args{purge_existing}) {
            vol_purge($args{conn}, $args{clone}) or
                croak("Purge of existing clone '$args{clone}' failed :$@");
            logcomment("vol_clone_create: Purged existing clone '$args{clone}'");
        } else {
            # check if the volume is a clone or not
            my ($parent, $snapname) = get_clone_info(conn=>$args{conn}, 
                                                        clone=>$args{clone});
            croak("Volume '$args{clone}' already exists but is not a clone.\n$@\n") 
                if ($@); # die if its not a clone
            croak("Existing clone '$args{clone}' is not derived from required parent.".
                "\nRequired parent '$args{parent}'" . 
                "\nExisting parent '$parent'\n") if ($parent ne $args{parent});
            if ($args{snapname}) {
                croak("Existing clone '$args{clone}' does not use required snapshot.".
                        "\nRequired snapshot '$args{snapname}'" . 
                        "\nExisting snapshot '$snapname'\n") 
                    if ($snapname ne $args{snapname});
            }
            if ($args{spcres}) {
                my %options = vol_options($args{conn}, $args{clone}) or
                     croak("$@\n");
                my ($spcres) = ($options{guarantee} =~ /(none|file|volume)\(?\S*\)?/);
                croak("Existing clone '$args{clone}' does not have required spcres.".
                        "\nRequired spcres '$args{spcres}'".
                        "\nExisting spcres '$options{guarantee}'\n")
                    if ($args{spcres} ne $spcres);
            }
            logcomment("vol_clone_create: Using existing clone '$args{clone}'");
            return(1);
        }
    }

    # build the options
    my $options;
    $options .= "-s $args{spcres} " if (defined($args{spcres}));
    $options .= "-f " if ($args{force_clone});
    $options .= "-b $args{parent} ";
    $options .= "$args{snapname}" if (defined($args{snapname}));
    logcomment("Clone create options are $options");

    my (undef, undef, $after) = say($args{conn}, "vol clone create $args{clone} $options");

    if ($after =~ /Creation of clone volume '$args{clone}' has completed/) {
        # verify if the clone volume really exists
        croak("vol_clone_create: Cannot see the clone '$args{clone}' on filer")
            if (!Volume::vol_exists($args{conn}, $args{clone}));
    } else {
        croak("vol_clone_create: $after");
    }

    return(1);
    };
}


=head2 vol_clone_split_start 

Separate clone from parent

=over

=item Synposis

    my $ret = vol_clone_split_start (conn=>$conn, clone=>vol2
                            [, wait_to_complete=>1] 
                            [, validate=>1] 
                            [, timeout=>1800]);

=item Arguments

This routine takes in the arguments in the key value pairs.
The arguments and their description are as follows:

=over

=item conn

(Required) : Connection Object to filer.

=item clone

(Required) : name of clone volume

=item wait_to_complete

(Optional) : flag indicating if routine is to wait for the clone
separation to complete. Default: 1 (wait to complete)

=item timeout

(Optional) : Time (in sec) to wait for clone separation to finish
This arg will be used iff wait_to_complete is 1.  Default: 1800

=item validate

(Optional) : Will validate if the clone specified is not a clone.
If not set then the string got by vol clone split status is
returned. Validation will be done iff wait_to_complete is set.
This is useful for negative testing.  Default: 1 (set)

=back

=back

=cut

sub vol_clone_split_start ( % ) {

    my (%args) = validate (@_, { 
                             conn => { type=>OBJECT, isa=>'Connectrec' },
                             clone => { type=>SCALAR },
                             wait_to_complete => { type=>BOOLEAN, optional=>1 },
                             timeout => { type=>SCALAR, optional=>1, default=>1800 },
                             validate => { type=>BOOLEAN, optional=>1, default=>1 },
                            }
                          );

    return eval {

    my $expected = "Clone volume .* will be split from its parent";
    my $wait = defined($args{wait_to_complete}) ? $args{wait_to_complete} : 1;

    my (undef, undef, $after) = say($args{conn}, "vol clone split start $args{clone}");

    if ($after =~ /$expected/) { 
        if ($wait) { # wait for the split to complete

            my $ret = wait_for_clone_separation($args{conn}, $args{clone}, $args{timeout}) or
                croak("vol_clone_split_start: $@");

            if ($args{validate}) {
                # get the status of the clone vol; it should not be a clone now
                my $response = get_clone_info(conn=>$args{conn}, clone=>$args{clone});
                croak("vol_clone_split_start: '$args{clone}' is still a clone") 
                    if ($response);
            } else {
                return ($ret); # return what was got from split status
            }
        }
        return(1);
    } else {
        croak("vol_clone_split_start: Expected '$expected'.\nGot '$after'");
    }

    };
}


=head2 vol_clone_split_stop

Stop separation of clone from parent

=over

=item Synopsis

    vol_clone_split_stop (conn=>$conn, clone=>vol 
                            [, wait_to_complete=>1]
                            [, validate=>1] 
                            [, timeout=>1800])

=item Arguments

This routine takes in the arguments in the key value pairs.
The arguments and their description are as follows:

=over

=item conn

(Required) : Connection Object to filer.

=item clone

(Required) : The name of the clone volume.

=item wait_to_complete

(Optional) : Flag indicating if routine is to wait for clone
separation to complete. Default: 1 (yes, wait for completion)

=item timeout

(Optional) : Time (in sec) to wait for clone separation to stop
This arg will be used iff wait_to_complete is 1. Default: 1800

=item validate

(Optional) : Will validate if the clone specified is still a clone.
If not set then the string got by vol clone split status is
returned. Validation will be done iff wait_to_complete is set.
This is useful for negative testing. Default: 1 (set)

=back

=back

=cut

sub vol_clone_split_stop ( % ) {

    my (%args) = validate (@_, { 
                             conn => { type=>OBJECT, isa=>'Connectrec' },
                             clone => { type=>SCALAR },
                             wait_to_complete => { type=>BOOLEAN, optional=>1 },
                             timeout => { type=>SCALAR, optional=>1, default=>1800 },
                             validate => { type=>BOOLEAN, optional=>1, default=>1 },
                            }
                          );

    return eval {

    my $wait = defined($args{wait_to_complete}) ? $args{wait_to_complete} : 1;
    my $expected = 'Clone split operation will stop';

    my (undef, undef, $after) = say($args{conn}, "vol clone split stop $args{clone}");

    if ($after =~ /$expected/) { 
        if ($wait) { # wait for split to stop

            my $ret = _wait_for_split_abort_completion($args{conn}, $args{clone}, $args{timeout}) or
                croak("vol_clone_split_stop: $@");

            if ($args{validate}) {
                # get the status of the clone and it should still be a clone
                my $response = get_clone_info(conn=>$args{conn}, clone=>$args{clone});
                croak("vol_clone_split_stop: '$args{clone}' is not a clone")
                    if (!$response);
            } else {
                return($ret); # return the message from split status
            }
        }
        return(1);
    } else {
        croak("vol_clone_split_stop: Expected '$expected'.\nGot '$after'");
    }

    };
}

=head2 vol_clone_split_complete

Waits for a clone to separate from its parent.

=over

=item Synposis

    my $ret = vol_clone_split_complete (conn=>$conn, clone=>vol2
                            [, validate=>1] 
                            [, timeout=>1800]);

=item Arguments

This routine takes in the arguments in the key value pairs.
The arguments and their description are as follows:

=over

=item conn

(Required) : Connection Object to filer.

=item clone

(Required) : name of clone volume

=item timeout

(Optional) : Time (in sec) to wait for clone separation to finish
This arg will be used iff wait_to_complete is 1.  Default: 1800

=item validate

(Optional) : Will validate if the clone specified is not a clone.
If not set then the string got by vol clone split status is
returned. Validation will be done iff wait_to_complete is set.
This is useful for negative testing.  Default: 1 (set)

=back

=back

=cut

sub vol_clone_split_complete ( % ) {

    my (%args) = validate (@_, { 
                             conn => { type=>OBJECT, isa=>'Connectrec' },
                             clone => { type=>SCALAR },
                             timeout => { type=>SCALAR, optional=>1, default=>1800 },
                             validate => { type=>BOOLEAN, optional=>1, default=>1 },
                            }
                          );

    return eval {
      my $start_time = time();
      wait_for_clone_separation($args{conn}, $args{clone}, $args{timeout}) or
	    croak("vol_clone_split_complete: $@");
      my $wait_time = time() - $start_time;
      logcomment("Waited " . int($wait_time / 60) . " minutes, " .
		 ($wait_time % 60) . " seconds for split to complete.");

      if ($args{validate}) {
	# make sure it is no longer a clone
	if (get_clone_info(conn=>$args{conn}, clone=>$args{clone})) {
	  croak("vol_clone_split_complete: '$args{clone}' is still a clone");
	}
      }
      return 1;
    };
}

=head2 vol_clone_split_status

Get status of clone separation

=over

=item Synopsis

    vol_clone_split_status (conn=>$conn [,clone=>clone] [,timeout=>200])

=item Arguments

This routine takes in the arguments in the key value pairs.
The arguments and their description are as follows:

=over

=item conn

(Required) : Connection object to the filer

=item clone

(Optional) : Clone volume whose clone separation status is required.
Default : will get the status of all clone separations.

=item timeout

(Optional) : Time (in sec) to wait for the say() to complete.

=back

=item Returns

Returns a hash of hashes indexed by clonename
dies if there is no clone status

=item Notes

Expected Output (from OnTap) of "vol clone split status [<vol>]"

    =====================================================================
    Vol Type        Split Status      Error / Status message
    =====================================================================
    non-existing                      Volume .* was not found
    volume
    ---------------------------------------------------------------------
    aggregate                         Volume .* was not found
    ---------------------------------------------------------------------
    traditional                       .* is not a flexible volume
    ---------------------------------------------------------------------
    flex non-clone                    The volume is not a clone
    ---------------------------------------------------------------------
    clone          none               Clone volume is not being
                                      split from its parent
    ---------------------------------------------------------------------
    clone          being split        no message - just status
    ---------------------------------------------------------------------
    clone          split complete     The volume is not a clone
    ---------------------------------------------------------------------
    volume not     no splits in       No clone status
    specified      progress
    ---------------------------------------------------------------------
    volume not     some splits        no message - just status
    specified      in progress
    ---------------------------------------------------------------------

=back

=cut

sub vol_clone_split_status ( % ) {

    my (%args) = validate (@_, { 
                             conn => { type=>OBJECT, isa=>'Connectrec' },
                             clone => { type=>SCALAR, optional=>1 },
                             timeout => { type=>SCALAR, optional=>1, default=>200 }
                            }
                          );

    eval {

    my %status;
    my $expected = "Volume \'\(\\S\+\)\', \(\\d\+\) of \(\\d\+\) inodes processed \\(\(\\d\+%\)\\)\[\\n\|\\r\]\+".
                   "\\s+\(\\d\+\) blocks scanned\. \(\\d\+\) blocks updated\.";

    #
    # Don't use constants in regular expression
    # they are not expanded as expected
    #
    my @patrn_list = (ERR_NOT_FOUND, ERR_NOT_FLEX, ERR_NOT_CLONE,
                        MSG_GEN_NO_SPLIT, MSG_NO_SPLIT);

    my $match_patrn = join('|',@patrn_list);

    my (undef, undef, $after) = say($args{conn}, "vol clone split status $args{clone}", 
                                    '-timeout', $args{timeout});

    if ($after =~ /($match_patrn)/) {
        croak($after);
    } elsif ($after =~ /$expected/) { 
        while ($after =~ /$expected/gm) { 
            $status{$1}{inode_processed}=$2;
            $status{$1}{total_inodes}=$3;
            $status{$1}{percentage}=$4;
            $status{$1}{blks_scanned}=$5;
            $status{$1}{blks_updated}=$6;
        }
        return(%status);
    } else {
        croak("vol_clone_split_status: Expected $expected.\nGot $after");
    }
    };
}

=head2 get_clones

Will return the clones for a volume.

=over 

=item Synopsis

    get_clones (conn=>$conn, vol=>vol0)

=item Arguments

This routine takes in the arguments in the key value pairs.
The arguments and their description are as follows:

=over

=item conn

(Required) : Connection Object to the filer.

=item vol

(Required) : The volume for which the clone info is to be obtained.

=back

=item Returns

list of clones for specified volume

=back

=cut

sub get_clones ( % ) {

    my (%args) = validate (@_, { 
                             conn => { tyoe=>OBJECT, isa=>'Connectrec' },
                             vol => { type=>SCALAR },
                            }
                          );

    eval {

    croak("get_clones: Volume $args{vol} does not exist")
        if (!Volume::vol_exists($args{conn}, $args{vol}));

    # TBD use Volume::vol_status to get this information
    my $out = sayEOF($args{conn}, "vol status $args{vol}");
    my @list;

    foreach my $line (split('\n', $out)) {
        if ($line =~ /Volume has clones: (.*)/) {
            @list = split(', ',$1);
        }
    }

    return(@list);

    };
}



=head2 get_clone_info

Will return the parent and base snapshot for a clone

=over

=item Synopsis

    get_clone_info (conn=>$conn, vol=>vol2)

=item Arguments

This routine takes in the arguments in the key value pairs.
The arguments and their description are as follows:

=over

=item conn

(Required) : Connection object to the filer.

=item clone

(Required) : Name of clone volume

=back

=item Returns

($parent, $base_snapshot)

dies on failure

=back

=cut

sub get_clone_info ( % ) {

    my (%args) = validate (@_, { 
                             conn => { type=>OBJECT, isa=>'Connectrec' },
                             clone => { type=>SCALAR },
                            }
                          );


    eval {

    croak("get_clone_info: Volume $args{clone} does not exist")
        if (!Volume::vol_exists($args{conn}, $args{clone}));

    # TBD use Volume::vol_status to get this information
    my (undef, undef, $out) = say($args{conn}, "vol status $args{clone}");
    
    # Expected CLI
    #tada-u22> vol status foofile                      
    #         Volume State      Status            Options
    #        foofile online     raid_dp, flex     guarantee=file(disabled)
    #                Clone, backed by volume 'src', snapshot 'clone_foofile.1'
    #                Containing aggregate: 'SRC'
    #tada-u22

    # when parent is offline / restricted

    #tada-u22> vol status foofile
    #         Volume State      Status            Options
    #        foofile online     raid_dp, flex     guarantee=file(disabled)
    #                Clone, backed by volume 'src', snapshot '(unavailable: id 11)'
    #                Containing aggregate: 'SRC'
    #tada-u22> 

    if ($out =~ /Clone, backed by volume \'(\S+)\', snapshot \'(.*)\'/) {
        return($1,$2);
    } else {
        croak("get_clone_info: '$args{clone}' is not a clone\n$out");
    }
    };
}

=head2 verify_clone

This routine will verify if the volume is a clone.
It is invoked usually after vol_create_clone()
and will verify the following
    - if parent and clone have same hosting aggregate
    - if base snapshot is un-removable
    - if snap sched is the same
    - if parentvol cannot be destroyed
    - if the data in parent and clone are same

The following verification need to be added
    - if diskfree is same (TBD)
        

Modification to filesystem on parent / clone will result
will result in differences in the volume

=over

=item Synposis

    verify_clone (conn=>$conn, clone=>vol2 [,expected_snapname=>foo]
                    [,data_compare=>1] [,compare_timeout=>200])

=item Arguments

This routine takes in the arguments in the key value pairs.
The arguments and their description are as follows:

=over

=item conn

(Required) : Connection object to the filer.

=item clone

(Required) : Clone volume to be verified.

=item expected_snapname

(Optional) : The name of the base snapshot that we expect the clone to
be created from.

=item data_compare

(Optional) : Compare the data in the clone and its parent. If set
treecompare will be used for this purpose. Default: 0 (do not compare)

=item compare_timeout

(Optional) : time in sec for the treecompare to finish. Will be used iff
data_compare is set.  Default: 200 (sec)

=back

=item Returns

returns 1 on success

undef on failure with $@

=back

=cut 

sub verify_clone ( % ) {


    my (%args) = validate (@_, { 
                             conn => { type=>OBJECT, isa=>'Connectrec' },
                             clone => { type=>SCALAR },
                             expected_snapname => {type=>SCALAR, optional=>1 },
                             data_compare => {type=>BOOLEAN, optional=>1, default=>0 },
                             compare_timeout => {type=>SCALAR, optional=>1, default=>200 },
                            }
                          );

    eval {

    # get parent and base snapshot of clone
    my ($parent, $base_snapshot) = get_clone_info(conn=>$args{conn}, clone=>$args{clone}) 
        or croak("verify_clone: Unable to get clone info: $@");
    tstlog('-commline', "Parent is $parent, Base snapshot is $base_snapshot");

    # Verify clone and parent have same host aggr
    my $parent_host_aggr = vol_container($args{conn}, $parent);
    my $clone_host_aggr = vol_container($args{conn}, $args{clone}); 
    croak("verify_clone: Parent $parent\(host: $parent_host_aggr\) and clone ". 
        "$args{clone} \(host: $clone_host_aggr\) have different hosting aggregate") 
        if ($parent_host_aggr ne $clone_host_aggr);

    # Verify base snap shot on parent is un-removable
    croak("verify_clone: Was able to delete base snapshot $base_snapshot from parent $parent")
        if (snapshot_delete($args{conn}, $parent, $base_snapshot));

    # If user has specified the expected snapname compare it to base snapshot
    if ($args{expected_snapname} && ($args{expected_snapname} ne $base_snapshot)) {
        croak("Base snapshot '$base_snapshot' is not the expected".
                    " snapname. Expected snapname was '$args{expected_snapname}'");
    }

    # Verify if the snap sched on parent and clone are same
    my @snapsched_parent = snapsched_volume($args{conn}, $parent);
    croak($@) if ($@);
    logcomment("Snap sched for '$parent' is @snapsched_parent");

    my @snapsched_clone = snapsched_volume($args{conn}, $args{clone});
    croak($@) if ($@);
    logcomment("Snap sched for '$args{clone}' is @snapsched_clone");

    croak("Snap sched for '$parent' and '$args{clone}' are not same size") 
        if(scalar(@snapsched_parent) != scalar(@snapsched_clone)); 
    foreach my $item (@snapsched_parent) {
        if (!scalar(grep {$_ eq $item } @snapsched_clone)) {
            croak("Snap sched for '$parent' and '$args{clone}' are not same");
        }
    }


    # Verify that the parent cannot be destroyed
    vol_offline($args{conn}, $parent) or
        croak("verify_clone: Unable to offline parent volume '$parent' :$@");

    vol_destroy($args{conn}, $parent);
    if ($@) {
        if ($@ !~ /cannot be destroyed because it has clone volumes/){
            croak("verify_clone: Was not able to destroy parent volume '$parent'".
                " but did not see expected string explaining why the destroy failed");
        }
    } else {
        croak("verify_clone: Was able to destroy parent volume '$parent'")
    }

    vol_online($args{conn}, $parent) or
        croak("verify_clone: Unable to online parent volume '$parent' :$@");


    # Verify that the data in parent and clone are same
    # Note that we expect lun metadata in the clone to change
    # since just creating the clone forces the lun offline,
    # so don't check that.
    #
    if ($args{data_compare}) {
        treecompare_vol($args{conn}, $parent, $args{clone}, args=>"-V", timeout=>$args{compare_timeout})
            or croak("verify_clone: Mismatch in data of parent $parent and clone $args{clone}: $@");
    }
    
    return(1);

    };
}

# Local subroutines

# wait_for_clone_separation

# loops for a while and checks vol clone status for a particular 
# volume. When the clone splitting has stopped / finished
# vol clone status will not display the clone status

# returns string if its what we expected
# when clone split start/stop has completed
# otherwise dies


sub wait_for_clone_separation ( $$$ ) {

    my ($conn, $clone, $timeout) = @_;

    eval {
        my $endtime = timeout2time( ($timeout) ? $timeout : 1800);
        my $msg_no_split_count = 10;
        my $msg_no_split_wait  = 2;
        my $match_msg_no_split = MSG_NO_SPLIT;

        while(time() < $endtime) {
            my (%status) = vol_clone_split_status(conn=>$conn, clone=>$clone);
            if ($@) {
                if ($@ =~ /$match_msg_no_split/ ) {
                    # When we check "clone split status",
                    # We can be here due to 2 reasons
                    # 1. Clone split initiated, but not started,
                    #    => that means, split is taking time to progress
                    #    => and we need to wait for split to start
                    #
                    # 2. clone split not started on clone vol
                    #    => this is a false condition 
                    if ( $msg_no_split_count ) {
                        sleep($msg_no_split_wait);
                        $msg_no_split_count--;
                        logcomment("clone split not started, split status:$@");
                    } else {
                        die("split not started on clone vol, split status:$@");
                    }
                } elsif ($@ !~ m/Expected.*Got/s) {
                    logcomment("Returning from clone split.., looks like done"); 
                    #return($@); # then we got what we expected
                    return 1; # return true instead of message.(see burt724041)
                } else {
                    die("wait_for_clone_separation, error: $@");
                }
            } else {
                # No error is set in $@, that means split is still running
                logcomment("Clone split is running"); 
                sleep(5); 
            }
        }
        die("Timed out waiting for clone status");
    };
}

# Local subroutines

# _wait_for_split_abort_completion

# loops for a while and checks vol clone status for a particular
# volume. When the clone splitting has stopped
# vol clone split status will display the clone status as
# "Clone volume is not being split from its parent"

# returns 1, if clone split is aborted successfully
# else dies with error string

sub _wait_for_split_abort_completion ( $$$ ) {

    my ($conn, $clone, $timeout) = @_;

    eval {
        my $endtime = timeout2time( ($timeout) ? $timeout : 1800);
        my $match_msg_no_split = MSG_NO_SPLIT;

        while(time() < $endtime) {
            my (%status) = vol_clone_split_status(conn=>$conn, clone=>$clone);
            if ($@) {
                if ($@ =~ /$match_msg_no_split/ ) {
                    # after clone split abort, clone is still a clone
                    # so we should always get MSG_NO_SPLIT message
                    logcomment("split abortion is completed, split status:$@");
                    return 1;
                } else {
                    logcomment("Clone split not running, error: $@");
                    die("_wait_for_split_abort_completion(), error: $@");
                }
            } else {
                # No error is set in $@, that means split is still running
                logcomment("Clone split is running");
                sleep(5);
            }
        }
        die("Timed out waiting for clone status");
    };
}


1;
