# $Id: //depot/prod/test/Rsierranevada/lib/Volume/Playlist.pm#1 $
## @summary Library for volume playlist operations
## @author dl-perf-qa@netapp.com
## @status public
## @pod here

=head1 NAME

Volume::Playlist - API for vol playlist commands

=head1 SYNOPSIS

  # Object-oriented interface:
  my $playlist = Volume::Playlist->new($conn, $vol)
  my $str      = $playlist->dump();
  my $hashref  = $playlist->status();
  $playlist->copy_in($path);
  $playlist->copy_out($path);

  # Procedural interface:
  playlist_copy_in(conn => $conn, vol => $vol, path => $path);
  playlist_copy_out(conn => $conn, vol => $vol, path => $path);
  my $str     = playlist_dump(conn => $conn, vol => $vol);
  my $hashref = playlist_status(conn => $conn, vol => $vol);


=head1 DESCRIPTION

Volume::Playlist provides two interfaces for interacting with
a volume's playlist: object-oriented and procedural.

The object-oriented interface is a wrapper for the procedural
one, and provides these methods:

=head1 METHODS

=head2 Volume::Playlist-E<gt>new($conn, $vol)

Constructs and returns a new Volume::Playlist object.

=head2 $obj-E<gt>copy_in($path)

Overwrites the volume's playlist file with the one stored at
the specified path.

=head2 $obj-E<gt>copy_out($path)

Copies the volume's playlist file to the specified path.

=head2 $obj-E<gt>dump()

Returns the volume's playlist contents as a string.

=head2 $obj-E<gt>status()

Returns the volume's playlist status as a hashref.

=head1 PROCEDURES

Alternatively, the playlist can be accessed through the
following procedures:

=head2 playlist_copy_in(conn => $conn, vol => $vol, path => $path)

Overwrites the volume's playlist file with the one stored at
the specified path.

The following keyword parameters are all required:

=over 4

=item conn

The connection to filer host on which the volume resides.

=item vol

The name of the volume whose playlist file is to be overwritten.

=item path

The pathname of the playlist file to copy in.

=back

=head2 playlist_copy_out(conn => $conn, vol => $vol, path => $path)

Copies the volume's playlist file to the specified path.

The following keyword parameters are all required:

=over 4

=item conn

The connection to filer host on which the volume resides.

=item vol

The name of the volume whose playlist file is to be overwritten.

=item path

The pathname of the playlist file to copy in.

=back

=head2 playlist_dump(conn => $conn, vol => $vol)

Returns the volume's playlist contents as a string.

The following keyword parameters are all required:

=over 4

=item conn

The connection to filer host on which the volume resides.

=item vol

The name of the volume whose playlist file is to be overwritten.

=back

=head2 playlist_status(conn => $conn, vol => $vol)

Returns the volume's playlist status as a hashref.

The following keyword parameters are all required:

=over 4

=item conn

The connection to filer host on which the volume resides.

=item vol

The name of the volume whose playlist file is to be overwritten.

=back

=cut

package Volume::Playlist;

use strict;
use warnings;


# Object-oriented interface:


sub new {
    my $class = shift;
    my $conn  = shift;
    my $vol   = shift;

    return bless {
        _conn => $conn,
        _vol  => $vol
    }, $class;
}


sub copy_in {
    my $self = shift;
    my $path = shift;

    playlist_copy_in(conn => $self->{_conn},
                     vol  => $self->{_vol},
                     path => $path);
}


sub copy_out {
    my $self = shift;
    my $path = shift;

    playlist_copy_out(conn => $self->{_conn},
                      vol  => $self->{_vol},
                      path => $path);
}


sub dump {
    my $self = shift;

    return playlist_dump(conn => $self->{_conn},
                         vol  => $self->{_vol});
}


sub status {
    my $self = shift;

    return playlist_status(conn => $self->{_conn},
                           vol  => $self->{_vol});
}


# Procedural interface:


use Carp;
use Params::Validate qw(validate HASHREF SCALAR);
use Tharn;
use base qw(Exporter);
our @EXPORT_OK = qw(playlist_copy_in
                    playlist_copy_out
                    playlist_dump
                    playlist_status);


sub playlist_copy_in {
    my %params = validate(@_,
                          {
                              conn => {
                                  type     => HASHREF,
                                  optional => 0,
                              },

                              vol  => {
                                  type     => SCALAR,
                                  optional => 0,
                              },

                              path => {
                                  type     => SCALAR,
                                  optional => 0,
                              },
                          });


    my $path = shift;
    my $cmd  = 'priv set -q test;'
             . ' sync;'
             . " vol playlist copy-in $params{vol} $params{path};"
             . ' sync';

    my (undef, undef, $out) = say($params{conn}, $cmd);

    confess "Unexpected output from '$cmd': $out"
        if $out ne q() and $out !~ m/$cmd/;
}


sub playlist_copy_out {
    my %params = validate(@_,
                          {
                              conn => {
                                  type     => HASHREF,
                                  optional => 0,
                              },

                              vol  => {
                                  type     => SCALAR,
                                  optional => 0,
                              },

                              path => {
                                  type     => SCALAR,
                                  optional => 0,
                              },
                          });


    my $path = shift;
    my $cmd  = 'priv set -q test;'
             . " vol playlist copy-out $params{vol} $params{path}";

    my (undef, undef, $out) = say($params{conn}, $cmd);

    confess "Unexpected output from '$cmd': $out"
        if $out ne q() and $out !~ m/$cmd/;
}


sub playlist_dump {
    my %params = validate(@_,
                          {
                              conn => {
                                  type     => HASHREF,
                                  optional => 0,
                              },

                              vol  => {
                                  type     => SCALAR,
                                  optional => 0,
                              },
                          });

    confess 'This subcommand has not been implemented yet.';
}


sub playlist_status {
    my %params = validate(@_,
                          {
                              conn => {
                                  type     => HASHREF,
                                  optional => 0,
                              },

                              vol  => {
                                  type     => SCALAR,
                                  optional => 0,
                              },
                          });

    my $cmd  = "priv set -q test; vol playlist status $params{vol}";

    my (undef, undef, $out) = say($params{conn}, $cmd);

    confess "Unexpected output from '$cmd': $out"
        if $out !~ m{((.*?)
                      playlist:\s+<(.*?)>
                      .*?
                      size=(\d+)\s+
                      (?:\(blocks=(\d+)\))? )}sgx;

    my $text = $1;
    my $status = {
        prolog => $2,
        path   => $3,
        size   => $4,
        blocks => $5 || 'n/a',
    };

    confess "Unexpected output from '$cmd': $out"
        if $out !~ m{(.*?
                      type=(\d+)\s+\((.*?)\)
                      .*?
                      perm=(\d+)
                      .*?
                      uid=(\d+),\s+gid=(\d+)
                      .*?
                      fileid=(\d+) )}sgx;

    $text .= $1;

    $status->{typeno}   = $2;
    $status->{typename} = $3;
    $status->{perm}     = $4;
    $status->{uid}      = $5;
    $status->{gid}      = $6;
    $status->{fileid}   = $7;

    confess "Unexpected output from '$cmd': $out"
        if $out !~ m{(.*?
                      atime/mtime/ctime/crtime/now=
                      (\d+)/(\d+)/(\d+)/(\d+)/(\d+) )}sgx;

    $text .= $1;

    $status->{atime}  = $2;
    $status->{mtime}  = $3;
    $status->{ctime}  = $4;
    $status->{crtime} = $5;
    $status->{now}    = $6;
    $status->{text}   = $text;

    return $status;
}


1;


__END__

