# $Id: //depot/prod/test/Rsierranevada/lib/Volume/RndAttr.pm#1 $
## @summary     Library to randomly pick a set of volume attributes for use in creation/modification.
## @author      astraub@netapp.com,dl-striping-qa@netapp.com
## @pod here

=head1 NAME

RndAttr.pm

=head1 SYNOPSIS

 use Volume::RndAttr;;
 use strict;
 my $vol_attrs;
 %vol_attrs = pick_vol_attrs(
                             template => undef, 
                             include_attrs => [qw(size new_attr)],
                             attr_definitions => {"new_attr"=> {type => "list", 
                                                                limits => [qw(value1 value2)],
                                                               },
                                                  },
                             attr_limits => {size => [52428800, 104857600]},
                             );
                             
 $server_obj->volume_create(volume=>"foovol",aggregate=>"myaggr",%vol_attrs);



=head1 DESCRIPTION

C<Volume::RndAttr> provides a mechanism to generate random volume configurations.
The volume attributes defined, and returned by this library are the same as the attributes used in Clustered ONTAP,
and can be used directly in ngsh volume commands. 


=head1 FUNCTIONS

=cut
 

package Volume::RndAttr;
use Params::Validate qw(:all);
use Tharn; qw($Log);
use Data::Dumper;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = 'pick_vol_attrs';

#list of parameters that are not available at create time.
my @post_create = qw(atime-update snapdir-access min-readahead percent-snapshot-space);
my %COTcfghash = (
                    'size'                   => {type => "range", limits => [20971520, 17592186044416]}, #20MB-16TB
                    'type'                   => {type => "list", limits => [qw(RW LS DP TMP)]},
                    'security-style'         => {type => "list", limits => [qw(unix mixed ntfs)]},
                    'autosize'               => {type => "list", limits => [qw(true false)]},
                    'max-autosize'           => {type => "range", limits => [52428800, 104857600]}, #50MB-100MB
                    'autosize-increment'     => {type => "range", limits => [1048576, 20971520]}, #1MB-20MB
                    'maxdir-size'            => {type => "range", limits => [10485760, 52428800]}, #10-50MB 
                    'space-guarantee'        => {type => "list", limits => [qw(none volume)]},
                    'stripe-count'           => {type => "range", limits => [1, 254]},
                    'stripe-width'           => {type => "range", limits => [64,262145], modifier => _stripe_width}, 
                    'concurrency'            => {type => "list", limits => [qw(low medium high peak)]},
                    'optimize'               => {type => "list", limits => [qw(max-performance performance space max-space)]},
                    'stripe-format'          => {type => "list", limits => [qw(full data)]},
                    'max-write-alloc-blocks' => {type => "range", limits => [0, 2048]},
                    'autosize'               => {type => "list", limits => [qw(true false)]},
                    'min-readahead'          => {type => "list", limits => [qw(true false)]}, 
                    'atime-update'           => {type => "list", limits => [qw(true false)]},
                    'snapdir-access'         => {type => "list", limits => [qw(true false)]},
                    'percent-snapshot-space' => {type => "range", limits => [0, 90]},                 
                );

my %templates = ('striped' => [qw(stripe-format stripe-count stripe-width 
                                  concurrency optimize security-style
				  maxdir-size space-guarantee min-readahead
				  atime-update percent-snapshot-space)],
                 'flex'    => [qw(maxdir-size space-guarantee security-style 
                                  min-readahead atime-update percent-snapshot-space)],
                );
         

sub _random_range {
    $Log->enter();
    my ($min,$max) = @_;
    my $_max = $max - $min;
    my $retval = int(rand($_max)) + $min;
    $Log->exit();    
    return $retval;
}

sub _stripe_width {
    $Log->enter();
    my $val = shift;
    return 4096 * ($val);
    $Log->exit();    
}

# template => <templatename> <flex|striped>
# include_attrs => <Arrayref> list of additional attributes to include
# attr_limits => <hashref> parameters to override the default hash
# attr_definitions => additional paramerters to add to the hash

=pod

=head2 pick_vol_attrs

C<pick_vol_attrs(%hash)>


B<Description>
 Randomly select volume attributes

B<Input>

=over 4

=item * B<template> => <striped|flex>

(Optional)
Default = flex

Defines default set of attributes for which to generate random values

B<flex> includes to following attributes: maxdir-size space-guarantee security-style
min-readahead atime-update percent-snapshot-space


B<striped> includes all of the 'flex' items, and additionally: stripe-format stripe-count stripe-width
concurrency optimize 


=item * B<include_attrs> => <arrayref>

(Optional)
Default = undef

list of volume attributes for which to generate random values.
Can be used in combination with the template param to include additional attributes.
The parameters currently defined in this library are:

size type security-style 
autosize max-autosize autosize-increment maxdir-size           
space-guarantee stripe-count stripe-width concurrency           
optimize stripe-format max-write-alloc-blocks autosize
min-readahead atime-update snapdir-access percent-snapshot-space

=item * B<attr_limits> => <hashref>

(Optional)
Default = undef

Allows the ability to control the criteria used in randomization of each attribute
The expectation is that the attributes must be  defined in the attributes list.
The hashref expects the data structure in the form:
    C<{attribute1 => $limits_arrayref}>

The limits are expressed in 2 ways depending on the attribute. Refer to the "limits" value in the attr_definitions parameter for more information.

example::

pick_vol_attrs(attr_limits => {numeric_attribute => [2, 50], 
                               list_attribute => [qw(value1 value2)]
                              }
               );

=item * B<attr_definitions> 

(optional)
hashref to define additional attributes that can be randomized

attr_definitions => {"new_attr"=> {type => "list",
                                   limits => [qw(value1 value2)],
                                  }
                     };


The following items define an attribute:


=over 8

=item * B<type> => <scalar> 

(required)

controls how limits values will be handled
Valid values are <"list"|"range"|"coderef">

=item * B<limits> => <ARRAYREF>

(required)

Arrayref of values to be used for randomization
The limits are expressed in 2 ways depending on the attribute:

B<range> - Arrayref of beginning ending values

B<list> - Arrayref of the items to randomly select from

=item * B<modifier> => &coderef

(optional)
allows modification of the randomized value by a secondary function.
After the value is selected, it will pass the value to the coderef defined here,
and use the value that was returned from it as the final value.

=back

=back

B<Output>
 hash of volume attributes

=cut


sub pick_vol_attrs {
    $Log->enter();
    my %cfghash = %COTcfghash; # make a copy of the global hash so we dont change it inbetween calls
    my %params = validate( @_,
                           { template         => {type => SCALAR|UNDEF},
                             attr_definitions => {type => HASHREF, optional => 1},
                             attr_limits      => {type => HASHREF, optional => 1},
                             include_attrs    => {type => ARRAYREF, optional => 1},
                });

    $Log->debug("InParms:".Dumper(\%params));

    my @attrs; #The list of items to randomly generate values

    ## add the attrubutes listed in the template
    if (defined($params{template})) {
        $Log->debug("Processing template: $params{template}");
        die "Template $params{template} not found" if (!defined($templates{$params{template}}));
        $Log->debug("using template $params{template} : @attrs");
        @attrs = @{$templates{$params{template}}};

    }

    ## add any new definitions provided by the caller
    if (defined($params{attr_definitions})) {
        $Log->debug("Processing attr definititons");
        for $key (keys(%{$params{attr_definitions}})) {
        $Log->debug("processing $key");
        if (exists($cfghash{$key})) {
                $Log->warn("$key already exists. Use of attr_limits parameter to override is recommended");
        }
            $cfghash{$key} = $params{attr_definitions}{$key};
        }
    }
    
    ## process the attr_limits hash to override any of the previously defined limits
    if (defined($params{attr_limits})) {
        $Log->debug("Processing attr_limits");
        for my $key (keys(%{$params{attr_limits}})) {
            $Log->debug("Processing $key");
            die "attribute $key not defined" if (!defined($cfghash{$key}));
            $cfghash{$key}{limits} = $params{attr_limits}{$key};
        }
    }

    ## Iterate through the include_attrs list, and pop the items onto the list
    if (defined($params{include_attrs})) {
        $Log->debug("Processing include_attrs");        
        for $attr (@{$params{include_attrs}}) {
            die "$attr is not defined in the config hash" if (! grep(/$attr/,keys(%cfghash)));
            if (! grep(/$attr/,@attrs)) {
                push(@attrs,$attr);
            }
        }
    }

  
    ## iterate through the @attrs list and generate the random values
    my %volattrs; # The hash of volume attributes we will return
    for $attr (@attrs) {
        my $reftype = ref($cfghash{$attr});
        $Log->debug("Randomizing '$attr' - Type '".ref($cfghash{$attr})."' Data: ".Dumper($cfghash{$attr}));            
        if ($cfghash{$attr}{"type"} eq "list") {
                $Log->trace("Randomly selecting value from list for '$attr'");
                my $index = int(rand(@{$cfghash{$attr}{"limits"}}));
                $volattrs{$attr} = $cfghash{$attr}{"limits"}[$index];
        } elsif ($cfghash{$attr}{"type"} eq "range") {
                $Log->trace("Randomly selecting value from range for '$attr'");
                $volattrs{$attr} = _random_range($cfghash{$attr}{"limits"}[0],$cfghash{$attr}{"limits"}[1] + 1);    
        } elsif ($cfghash{$attr}{"type"} eq "coderef") {
            $Log->trace("Calling limits subroutine for '$attr'");
            $volattrs{$attr} = $cfghash{$attr}{"limits"}();
        }
        if (defined($cfghash{$attr}{"modifier"})) {
            $subname = $cfghash{$attr}{"modifier"};
            $Log->trace("Calling modifier subroutine for $attr");
            $volattrs{$attr} = &$subname($volattrs{$attr});
        }
        $Log->debug("Value selected '$attr' : '$volattrs{$attr}' ");
    
    }
    $Log->debug("Returning hash: ".Dumper(\%volattrs));

    $Log->exit();
    return %volattrs;
}


# Convienence function to verify strip vol size
#
# $_[0] = ref to hash made by &pick_vol_attrs

=pod

=head2 stripe_vol_min_size

B<Description>
   Calculate the minimum value for -size for volume new command from the 
   values returned to the hash from &pick_vol_attrs

   $min_size = &Volume::RndAttr::stripe_vol_min_size(\%vol_attrs);

B<Input>

=over

=item reference to the hash populated with &pick_vol_attrs

=back

B<Ouput>
   Size value that the filer will allow the striped volume being created to use
   for the -size option.  The value returned is not the absolute min value
   that would be caclulated on the filer, but an approximation.


B<Modes>
   Tricky and BR have different min size requirements. This function
   will default to the Tricky approximation, since Tricky requires
   more then BR a Tricky approximation will more than meet the needs on
   a BR install. To set modes, directly define the $SIZE_FACTOR value

=over

=item Tricky  $Volume::RndAttr::SIZE_FACTOR = '1.2G';   # default

=item BR      $Volume::RndAttr::SIZE_FACTOR = '100M';

=back 

When SIZE_FACTOR is set to one of these two strings, the variables used
in calculating the approximation are set.

The user can set SIZE_FACTOR so as to define their own approximation varaibles,
example

=over

=item  $Volume::RndAttr::SIZE_FACTOR     = 'none';

=item  $Volume::RndAttr::SIZE_OVERHEAD   = 1.5;

=item  $Volume::RndAttr::SIZE_MULTIPLIER = 1024 * 1024 * 500;   # 500M

=item  $Volume::RndAttr::SIZE_MDV_TAX    = 0

=item Algorithmn 

=item  min_size = (<stripe_count> + MDV_TAX) * (SIZE_OVERHEAD * SIZE_MULTIPLIER) 

=back

=cut

our $SIZE_FACTOR = '1.2G';    # Tricky 1.2G  BR 100M 
our $SIZE_OVERHEAD;
our $SIZE_MULTIPLIER;
our $SIZE_MDV_TAX;


sub stripe_vol_min_size {
   # 1.2G sizing 
   if ($SIZE_FACTOR eq '1.2G') {
      $SIZE_MDV_TAX    = 1;
      $SIZE_OVERHEAD   = 1.2;
      $SIZE_MULTIPLIER = 1024 * 1024 * 1024;
   }
   # 100M sizing
   if ($SIZE_FACTOR eq '100M') {
      $SIZE_MDV_TAX    = 1;
      $SIZE_OVERHEAD   = 1;
      $SIZE_MULTIPLIER = 1024 * 1024 * 100;
   }
   int( ($_[0]->{'stripe-count'} + $SIZE_MDV_TAX) * ($SIZE_OVERHEAD * $SIZE_MULTIPLIER) );
}



1;
