# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Volume/Setup.pm#2 $


## @summary Library of volume / aggregate setup commands
## @author archana@netapp.com, dl-test-infra@netapp.com
## @status Shared
## @pod here


=head1 NAME

Setup - Set up multiple volumes / aggregates

=head1 SYNOPSIS

   use Volume::Setup qw(setup_volumes_from_string setup_volumes);

   my $host = host($FILER);

   # scalar context, return 1 on success.
   my $status = setup_volumes_from_string($host, $FILER_VOL_LIST);

   # list context, return refs to lists of new vols and new aggrs
   my ($new_vols, $new_aggrs) = setup_volumes_from_string($host, $FILER_VOL_LIST);

   # scalar context, return 1 on success.
   my $status = setup_volumes($host,
                 {vol=>"aggr1",type=>"aggr",...},
                 {vol=>"vol1",type=>"flex",aggr=>"aggr1",...},
                 {vol=>"clone1",type=>"clone",parent=>"vol1",...},
                 {vol=>"sparse1",type=>"sparse",source=>"hostname:vol",...},
                );

   # list context, return refs to lists of new vols and new aggrs
   my ($new_vols, $new_aggrs) = setup_volumes($host,
                 {vol=>"aggr1",type=>"aggr",...},
                 {vol=>"vol1",type=>"flex",aggr=>"aggr1",...},
                 {vol=>"clone1",type=>"clone",parent=>"vol1",...},
                 {vol=>"sparse1",type=>"sparse",source=>"hostname:vol",...},
                );

=head1 DESCRIPTION

Creates aggregates and volumes of different types namely traditional, flexible, 
clones, compressed and a_sis(de-dup). Volumes can also be populated with data via 
different protocols. 

Existing volumes or aggregates can also be purged before new ones are created.
Verification is not done to see if existing aggregate / volume matches user
specified attributes.

In list context, the setup* functions return two additional values: a newly-created
volumes list ref and newly-created aggregates list ref.


=head1 DEPENDENCIES

Will not work for ontap versions below 7.0

nearstore_option/wafl_compression has to be licensed if compression is used

nearstore_option/a_sis has to be licensed if a_sis is used

syncmirror_local must be licensed and set up if raid_mirror is set to mirrored

Filer must be setup for NFS if population of the volume is desired.

Filer must be setup for CIFS if population of the volume via CIFS is desired.

Filer must be setup for FCP if population of the volume via FCP is desired.

snaplock|snaplock_enterprise to be licensed if snaplock is used.
Compliance clock for snaplock needs to be setup in the calling script/library
before calling any function in this library.

=head1 BUGS / ENHANCEMENTS

RAID0 not yet implemented

Purging of clones and volumes with clones do not work very well

Have to add support for using different types of cifs authentication while
populating the volume.

Because ONTAP <7.1 only supports one of snaplock and snaplock enterprise
license at a time, only one type of volume/aggr can be created per call.

Populate volume and not just qtree in a volume

=head1 ROUTINES

=cut


package Volume::Setup;

use strict;
use warnings;

use base qw(Exporter);
BEGIN
{
  our %EXPORT_TAGS = (
    all => [ qw(
                setup_volumes 
                setup_volumes_from_string
                create_aggregate
                create_volume
                )
           ] );
  Exporter::export_ok_tags('all');
}

use Carp qw(croak confess);
use Params::Validate v0.70 qw(validate BOOLEAN SCALAR OBJECT HASHREF UNDEF);
use Tharn qw(logcomment logresult script_result $Log);
BEGIN
{
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
                confess(script_result(@msgs))}
        );
}
use Volume::Clone qw(vol_clone_create);
use Volume qw(vol_status vol_exists vol_purge vol_create vol_create_flex vol_size);
use Aggregate qw(aggr_exists aggr_purge aggr_create aggr_status);
use fileSystem qw(populate_qtree);
use Utils qw(make_conn);
use Utils::Units qw(:convert :constants);
use Dense qw(dense_toggle);
use compression qw(enable_compression_option disable_compression_option);
use RaidInfo;
use HybridBuftree qw(aggr_is_convertible aggr_convert);
use Filer::Version qw(get_wafl_version WV_BOILERMAKER_MAX);

=head2 setup_volumes_from_string

As C<setup_volumes>, but takes a string description instead of an
array of hashes.  This is intended to be used when the description of
volumes to (re)create comes from an external source, such as as a
parameter to a test.

=over

=item Arguments

=over

=item host

(Required) : NATE hostobj of filer

=item string

(Required) : A space separated list of volume descriptions. Each
volume description is a a comma separated list of 
I<attribute>=I<value> pairs.  For example:

   vol=foo,type=trad aggr=foo2,type=aggr vol=foo3,type=flex,aggr=foo2

Refer to C<setup_volumes> for possible attributes and more details

=back

=item Returns

on success:
- in scalar context, 1.
- in list context, return two additional values: newly-created volumes
list ref and newly-created aggregates list ref.

dies otherwise.

=back

=cut

sub setup_volumes_from_string ( $$ ) {

    my ($host, $vol_list) = @_;

    my @list;
    # extract attributes and create hash
    my @vols = split(/\s+/, $vol_list);
    foreach my $vol (@vols) {
        my $args = {};
        my @attributes = split(/,/, $vol);
        foreach my $attr (@attributes) {
            my ($key, $value) = ($attr =~ /(.*?)=(.*)/);
            $args->{$key} = $value;
        }
        push (@list, $args);
    }
    return setup_volumes($host, @list);

}


=head2 setup_volumes

Create (or destroy and then re-create) several volumes and aggregates,
possibly including traditional, flexible, cloned, compressed and/or a_sis
volumes.

=over

=item Returns

on success:
- in scalar context, 1.
- in list context, return two additional values: newly-created volumes
list ref and newly-created aggregates list ref.

dies otherwise.


=item Arguments

=over

=item host

(Required) : NATE hostobj to filer

=item volume descriptions

(Required) : Each argument after the hostobj is a hash reference
describing a particular volume, aggregate, or clone to be created.

Possible elements are:

=over

=item type

type of volume/aggregate.  possible values: C<trad>, C<flex>, C<clone>, C<sparse> or C<aggr>.

=item vol

name of volume. not used if type=aggr. if type=sparse, volume name format is
'remotehost:remotevolume'.

=item aggr

name of aggregate

=item vol_size

size of the volume (used only if type=flex)

=item disks

number of disks to be used (used only if type=trad or aggr)

=item raid_mirror

type of raid mirroring.  possible values: C<mirrored> or C<normal>
(used only if type=trad or aggr)

=item cksum_style

Not implemented.
style of checksum.  possible values: C<block> or C<zoned>
(used only if type=trad or aggr)

=item raid_type 

type of raid.  Possible values: C<raid4> or C<raid_dp> (C<raid0> not yet implemented)
(used only if type=trad or aggr)

=item lang

language for the volume

=item spcres

space reservations for volume.  possible values: C<file>, C<volume>, or C<none>

=item parent

create a clone volume, using the volume with this name as the parent.
(parent specified must exist already) (used only if type=clone).

=item snapname

An optional C<snapname> to be used when creating a clone volume

=item compression

Allowed for volumes only. Cannot compress aggregates.
If set, then compression is enabled for the volume.
Licensing for this feature is not done by this routine. 
It has to be done prior to invoking this routine.
Use Admin::License to license this feature.

=item large_aggr

Allowed for volumes and aggregates. For aggregates, will
cause a "large aggregate" type aggr to be created. For vols,
if the containing aggr for the vol needs to be created, then
this aggr will be a "large_aggr" type. If the containing aggr
already exists, then this option has no effect.

=item a_sis

Allowed for volumes only. Cannot perform on aggregates.
If set, then Advanced Single Instance Storage (SIS) management
is enabled for the volume.
Licensing for this feature is not done by this routine. 
It has to be done prior to invoking this routine.
Use Admin::License to license this feature.

=item snaplock

Optional, applies to aggregates, flex and trad volumes.
If present, should have value "compliance" or "enterprise".
As the flexvol "inherits" the type from the aggregate, 'snaplock'
applied to flexvols is actually redundant.
Licensing for this feature is not done by this routine;
it has to be done prior to invoking this routine.
Use Admin::License to license this feature.
System date and Compliance Clock need to be set/initialized before
calling.

=item purge_existing

if true, and the given volume or aggregate already exists, destroy it
and then create it again.  If false (0), then leave it as-is.

=item populate

Allowed for volumes only. Cannot populate aggregates. If set will
will populate a qtree in the volume with all types of wafl data.

=item nfs_client

An nfs client to populate the volume. Default: 'localhost'
If used then filer and nfs_client must be able to talk via nfs.
(used only if populate=1)
If not specified will use NFS_CLIENT from the environment.

=item fcp_client

A fcp client to populate a lun in the volume with data. If not
specified then only a lun will be created on the volume.
If used then filer and fcp_client must be able to talk via fcp.
(used only if populate=1)
If not specified will use FCP_CLIENT from the environment.

=item cifs_client

A cifs client to populate the volume with windows data.
If used then filer and cifs_client must be able to talk via cifs.
(used only if populate=1)
If not specified will use CIFS_CLIENT from the environment.

=item cifs_user

User to use for cifs authentication when contacting the filer.
(used only if populate=1 and cifs_client is defined)
If not specified will use CIFS_USER from the environment.

=item cifs_password

Password to use for cifs authentication when contacting the filer.
(used only if populate=1 and cifs_client is defined)
If not specified will use CIFS_PASSWORD from the environment.

=item tree_prefix

Prefix of qtree name to  use when qtrees are created for 'populated'
volumes. Used only if populate=1. Mutually exclusive with "tree_name"
option. Prefix is appended with 'random' number (frem timestamp)

=item tree_name

Qtree name to use when qtrees are created in the volume for 'populated'
volumes. Used only if populate=1. Mutually exclusive with "tree_prefix"
option.

=item timeout

time in sec to wait for aggr or volume creation.

=back

If the volume / aggregate does not exist then it will be created.  If
the volume is a flex volume and its containing aggr is not present
that will be created too.

If a clone volume is to be created then specify type=clone
and its parent by using C<parent> attribute.

Parent must exist before clone can be created and can be created
earlier on in the list.  For example, volume C<foo> is created first
and is used as the parent for the clone volume C<boo>:

   my ($new_vols, $new_aggrs) = setup_volumes($host,
     {vol=>'foo',type=>'flex',aggr=>'xxx'},
     {vol=>'boo',type=>'clone',parent=>'foo'}
    );

Verification of pre-existing aggregates/ volumes is not done to see
if they match the specified aggregates.
So the safest way would be to purge existing volumes
which can be acomplished using the C<purge_existing> attribute.

C<purge_existing> will delete the existing volume / aggregate.
When used with flex volumes note that the container aggr is 
not purged. If one desires to purge the container aggregate also
then it can be done as specified in the following example:

   my ($new_vols, $new_aggrs) = setup_volumes($host,
     {aggr=>'foo',type=>'aggr',disks=>2,purge_existing=>1}, 
     {vol=>'boo',type=>'flex',aggr=>'foo',purge_existing=>1}
    );


=back

=back


=cut

sub setup_volumes ( $@ ) {

    my ($host, @list) = @_;

    # info about what gets created
    my (@new_vols, @new_aggrs);
    foreach my $item (@list) {
        $item->{host} = $host;
	my ($vols, $aggrs) = _setup(%$item);
        croak("$@\n") if ($@);
        push @new_vols,  @$vols  if (wantarray && length(@$vols));
        push @new_aggrs, @$aggrs if (wantarray && length(@$aggrs));
    }
    return wantarray ? (\@new_vols, \@new_aggrs) : 1;
}


use Volume qw(volume_special_options_hook);

sub _setup ( % ) {
    my (%args) = validate (@_, { 
                 host => { type=>OBJECT, isa=>'Hostrec' },
                 type => { type=>SCALAR, regex=>qr/flex|trad|aggr|clone|sparse/ },
                 vol => { type=>SCALAR, optional=>1, regex=>qr/\w+/ },
                 aggr => { type=>SCALAR, optional=>1, regex=>qr/\w+/ },
                 vol_size => { type=>SCALAR, optional=>1, regex=>qr/\d+(k|m|g|t)?/ },
                 disks => { type=>SCALAR, optional=>1 },
                 raid_mirror => { type=>SCALAR, optional=>1, regex=>qr/mirrored|normal/ },
                 cksum_style => { type=>SCALAR, optional=>1, regex=>qr/block|zoned/ },
                 raid_type => { type=>SCALAR, optional=>1, regex=>qr/raid4|raid_dp/ },
                 lang => { type=>SCALAR, optional=>1 },
                 spcres => { type=>SCALAR, optional=>1, regex=>qr/volume|file|none/ },
                 parent => { type=>SCALAR, optional=>1, regex=>qr/\w+/ },
                 snapname => { type=>SCALAR, optional=>1, regex=>qr/\w+/, depends=>['parent'] },
                 purge_existing => { type=>SCALAR, regex=>qr/1|0/, optional=>1 },
                 force_clone => { type=>BOOLEAN, optional=>1, default=>0 },
                 populate => { type=>SCALAR, regex=>qr/1|0/, optional=>1 },
                 nfs_client => { type=>SCALAR|HASHREF, optional=>1 },
                 cifs_client => { type=>SCALAR|HASHREF, optional=>1 },
                 fcp_client => { type=>SCALAR|HASHREF, optional=>1 },
                 cifs_user => { type=>SCALAR, optional=>1, default=>'administrator' },
                 cifs_password => { type=>SCALAR|HASHREF, optional=>1 },
                 tree_prefix => { type=>SCALAR, optional=>1 },
                 tree_name => { type=>SCALAR, optional=>1 },
                 timeout => { type=>SCALAR, optional=>1, regex=>qr/^\d+$/ },

                 # Compression, A_SIS, etc.
                 license_special_options => { type=>SCALAR, optional=>1, default=>0 },
                 compression => { type=>SCALAR|UNDEF, optional=>1 },
                 a_sis => { type=>SCALAR|UNDEF, optional=>1 },
                 large_aggr => { type=>SCALAR|UNDEF, optional=>1, default=>$::LARGE_AGGR },
                 snaplock => { type=>SCALAR|UNDEF, optional=>1,
                              regex=>qr/^(|compliance|enterprise)$/, default=>'' },

		 source => { type=>SCALAR, optional=>1, 
			    regex=>qr/^[^:]+:\w+$/ },
                }
            );

    # Validate Compression, A_SIS, etc. options
    my $special_option_actions = {};
    foreach my $opt  (qw( a_sis compression )) {
        if (defined($args{$opt}) && $args{$opt} !~ /1|0|true|false|on|off|yes|no/i) {
            croak("setup_volumes: \"$opt\" option is defined, but fails 'regex' test: \"$args{$opt}\"");
        }
        $special_option_actions->{$opt} = $args{$opt} if (defined($args{$opt}));
    }

    logresult('warn', 'Argument cksum_style is not implemented.') if ($args{cksum_style});
    # copy relevant args to pass down to create_aggregate()
    my %aggr_args = (host=>$args{host}, aggr=>$args{aggr});
    foreach my $x (qw(purge_existing disks raid_mirror cksum_style raid_type timeout large_aggr snaplock)) {
        $aggr_args{$x} = $args{$x} if (defined($args{$x}));
    }
    # copy relevant args to pass down to create_volume()
    my %vol_args = (host=>$args{host}, vol=>$args{vol});
    foreach my $x (qw(aggr purge_existing type vol_size disks raid_mirror cksum_style raid_type lang spcres timeout source snaplock)) {
        $vol_args{$x} = $args{$x} if ($args{$x});
    }

    # make a connection (default's to #rsh) to send to other routines
    my $conn = make_conn($args{host});

    # info about what gets created
    my (@new_vols, @new_aggrs);

    if ($args{type} eq 'clone') {

        # create a clone vol
        croak("Volume name for clone creation is not defined\n") if (!$args{vol});
        croak("Parent volume for creating clone $args{vol} not defined\n") if (!$args{parent});
        logcomment("Creating clone '$args{vol}'");

        my @args = (conn=>$conn, parent=>$args{parent}, clone=>$args{vol});
        foreach my $x (qw(snapname spcres purge_existing force_clone)) {
            push(@args, ($x, $args{$x})) if ($args{$x});
        }
        vol_clone_create(@args) or croak("$@\n");
	push(@new_vols, $args{vol});

    } elsif ($args{type} eq 'aggr') {

        # create an aggregate
        croak("Aggregate name not defined\n") if (!$args{aggr});
        logcomment("Creating aggregate '$args{aggr}'");
        eval {
            create_aggregate(%aggr_args);
	    push(@new_aggrs, $args{aggr});
        };
        croak("$@\n") if ($@);

    } else {

        # create a trad/flex volume
        croak("Volume name not defined\n") if (!$args{vol});
        logcomment("Creating volume '$args{vol}'");

        if (($args{type} eq 'flex') || ($args{type} eq 'sparse')) {

            croak("Hosting aggregate for flex volume '$args{vol}' is not defined\n") 
                if (!$args{aggr});

            # create hosting aggr for flex vol / or check if its online
            delete $aggr_args{purge_existing}; # dont purge aggregate
            eval {
		create_aggregate(%aggr_args);
		push(@new_aggrs, $args{aggr});
            };
            croak("$@\n") if ($@);
        }

        eval {
	    create_volume(%vol_args);
	    push(@new_vols, $args{vol});
        };
        croak("$@\n") if ($@);
    }

    if ($args{type} ne 'aggr') {
	# Handle Compression, A_SIS, etc.
	Volume::volume_special_options_hook(filer=>$args{host},
					    volume=>$args{vol},
					    do_license=>$args{license_special_options},
					    actions=>$special_option_actions);
    }

    # populate volume if asked to
    if ($args{populate}) {
        croak("Volume name for population is not defined\n") if (!$args{vol});

        # get values from environment if not defined
        # Get CIFS_DOMAIN AND CIFS_WORKGROUP XXX - add domain stuff
        foreach my $arg (qw(nfs_client fcp_client cifs_client cifs_user cifs_password)) {
            if (!$args{$arg}) {
                my $param_name = uc($arg);
                my $value = $ENV{$param_name};
                #my $value = param($param_name, '-default', undef);
                if ($value) {
                    $args{$arg} = $value;
                    $Log->debug("Have set $arg to $args{$arg}");
                }
            }
        }

        # populate a qtree in the volume with diff types of WAFL files
        my @opts;
        push(@opts, (uclient=>$args{nfs_client}) ) if ($args{nfs_client});
        push(@opts, (sanclient=>$args{fcp_client}) ) if ($args{fcp_client});
        push(@opts, (wclient=>$args{cifs_client}) ) if ($args{cifs_client});
        push(@opts, (win_user=>$args{cifs_user}) ) if ($args{cifs_user});
        push(@opts, (win_passwd=>$args{cifs_password}) ) if ($args{cifs_password});
        my $qtreepath;
	if ($args{tree_name}) {
            $qtreepath = $args{tree_name};
	} else {
            # get a unique name for each tree - useful when clones are created
            $qtreepath = ($args{tree_prefix} ? $args{tree_prefix} : 'qtree');
            $qtreepath .= time();
	}
        logcomment("Will populate qtree '$args{tree_prefix}' in volume '$args{vol}'");
        $Log->debug("Invoking populate qtree with options @opts");
        populate_qtree($conn, $args{vol}, $qtreepath, @opts)
            or logresult('inspect',"Population of vol '$args{vol}' failed : $@\n");
    }

    return(\@new_vols, \@new_aggrs);

}

=head2 create_aggregate

Creates an aggregate. Basically a wrapper around aggr_create.
If the volume / aggregate is already present and if purge_existing 
is '1' then the exisiting volume / aggregate will be purged and
a new one created. The volume / aggregate creation heavily relies
on the additional args disks, raid_mirror, cksum_style and raid_type.
So depending on the type of volume / aggregate asked for, the correct
spares are chosen and the volume / aggregate created.

    create_aggregate(host=>$host, aggr=>$aggr, 
                    [, purge_existing=>0]
                    [, create_opts=>'']
                    [, disks=>2]
                    [, raid_mirror=>'normal']
                    [, cksum_style=>'']
                    [, raid_type=>'raid4']
                    [, snaplock=>'']
                    [, large_aggr=>'']
                    [, timeout=>3600] );

The routine takes the following named arguments

=over

=item spec

(Required): Hostspec of filer

=item aggr  

(Required) : Name of aggregate to be created

=item purge_existing

(Optional): Flag to denote if existing vol / aggr must be purged and a 
new one created in its place.  Default: 0

=item create_opts

(Optional): Options to create the aggregate. If provided then the arguments
disks, cksum_style, raid_type, raid_mirror will be ignored. These options
will be directly passed to aggr_create or vol_create.
No error checking will be done.  Default: ''
This option will be deprecated soon.

=item disks

(Optional): Number of disks to use for volume / aggregate creation. Default: 2

=item raid_mirror

(Optional): To create a normal or mirrored volume / aggregate.
Possible values: 'normal' / 'mirrored'. Default: 'normal'

=item cksum_style

Not implemented.
Optional: Style of checksum to use for volume / aggregate. 
Possible values: 'zoned' / 'block' / ''. When '' then ontap will decide.
Default: ''

=item raid_type

(Optional): Type of raid group to use for volume / aggregate creation.
Possible values: 'raid4' / 'raid_dp'. Default: 'raid4'

=item snaplock

(Optional): Whether to create a snaplock aggregate,
and which type (compliance or enterprise).

=item large_aggr

Allowed for volumes and aggregates. For aggregates, will
cause a "large aggregate" type aggr to be created. For vols,
if the containing aggr for the vol needs to be created, then
this aggr will be a "large_aggr" type. If the containing aggr
already exists, then this option has no effect.

=item timeout

(Optional): Time (in sec) to wait for aggregate to be created. 
Default: 3600 (sec)

=back

The routine returns 1 if the aggregate was created by this routine
0 if aggregate was not created in this routine
will die with $@ if the routine could not complete for some reason


The following are the dependencies 
- Need equal number of disks in each plex for mirrored volume / aggregate. 
- Its assumed that the number of plexes is 2, so an even number of disks are
required to make a mirrored volume / aggregate.
- Need a min of 2 disks to make a raid4 volume / aggregate
- Need a min of 3 disks to make a raid_dp volume / aggregate

Note that this routine 
- does not take into consideration the size of the spares.
- does not take into consideration the % of zeroing of the spares.
- flex volume is a wafl object and is not created by this routine

=cut

sub create_aggregate ( % ) {

    my (%args) = validate (@_, 
                    { 
                     host => { type=>OBJECT, isa=>'Hostrec' },
                     aggr => { type=>SCALAR, regex=>qr/\w+/ },
                     purge_existing => { type=>SCALAR, regex=>qr/1|0/, 
                                optional=>1, default=>0 },
                     create_opts => { type=>SCALAR, optional=>1, default=>'' },
                     disks => { type=>SCALAR, optional=>1, regex=>qr/^\d+$/, 
                                default=>2 },
                     raid_mirror => { type=>SCALAR, optional=>1, 
                                regex=>qr/^(mirrored|normal)$/, default=>'normal' },
                     cksum_style => { type=>SCALAR, optional=>1, 
                                regex=>qr/^(block|zoned)$/, default=>'' },
                     raid_type => { type=>SCALAR, optional=>1, 
                                regex=>qr/^(raid4|raid_dp)$/, default=>'raid4' },
                     snaplock => { type=>SCALAR, optional=>1,
                                   regex=>qr/^(|compliance|enterprise)$/, default=>'' },
                     large_aggr => { type=>SCALAR, optional=>1,
                                   regex=>qr/^(0|1|true|false|yes|no)$/, default=>undef },
                     timeout => { type=>SCALAR, optional=>1, 
                                regex=>qr/^\d+$/, default=>3600 },
                    }
                );

    logresult('warn', 'Argument cksum_style is not implemented.') if ($args{cksum_style});

    my $conn = make_conn($args{host});

    $Log->debug("Will create aggregate:$args{aggr} disks:$args{disks}". 
                " cksum_style:$args{cksum_style}". 
                " raid_mirror:$args{raid_mirror} raid_type:$args{raid_type}");

    # Check if volume / aggregate already exists
    my $exists = aggr_exists($conn, $args{aggr});

    if ($exists) {
        if ($args{purge_existing}) { 
            $Log->debug("Purging existing aggregate '$args{aggr}'");
            aggr_purge($conn, $args{aggr}) or
                die("Purge of aggregate '$args{aggr}' failed :$@\n");
            $exists=0;
        } else {
            # check if aggregate is online
            my $status_obj = aggr_status($conn, $args{aggr});
            die ("Unable to get status of aggregate $args{aggr} : $@") if ($@);
            die("Aggregate '$args{aggr}' exists but is not online. Will not be able to use it.")
                if ($status_obj->state() ne 'online');

            $Log->debug("Will use existing aggregate '$args{aggr}'");
        }
    }

    my $we_created = 0;
    if (!$exists) {
        if (0) { # TBD implement this part of getting raid options later
            if (!$args{create_opts}) {
                # get raid create options
                eval {
                    my $create_opts = _get_raid_create_opts($args{host}, 
                                disks=>$args{disks},
                                cksum_style=>$args{cksum_style},
                                raid_type=>$args{raid_type},
                                raid_mirror=>$args{raid_mirror}
                                );
                };
                die($@) if ($@);
            }
        }

        my %extra_args = ();
        if (defined($args{large_aggr})) {
            my $large_aggr = ($args{large_aggr} =~ /1|true|yes/);
            $extra_args{indirect_type} = ($large_aggr ? '64' : '32');
        }
	if ($args{snaplock} eq 'compliance') {
	    $extra_args{snaplock} = 1;
        } elsif ($args{snaplock} eq 'enterprise') {
	    $extra_args{snaplock_enterprise} = 1;
	}

        # create an aggregate
        # use aggr_create from Aggregate.pm
# WIP: why set this value here and also after the die below?
        $we_created = aggr_create($conn,
                                    $args{aggr},
                                    $args{disks},
                                    mirror=>($args{raid_mirror} eq 'mirrored') ? 1 : 0,
                                    raidtype=>$args{raid_type},
                                    timeout=>$args{timeout},
                                    %extra_args,
                        );
        die("Aggregate '$args{aggr}' could not be created :$@") if ($@);
# WIP: why do this *and* also use the aggr_create() return value?
        $we_created = 1;
    }
    return($we_created);
}


=head2 create_volume

    create_volume(host=>$host, vol=>$vol, 
                [, aggr=>0]
                [, purge_existing=>0]
                [, type=>'flex']
                [, vol_size=>'100m']
                [, create_opts=>'']
                [, disks=>2]
                [, raid_mirror=>'normal']
                [, cksum_style=>'']
                [, raid_type=>'raid4']
                [, lang=>'']
                [, snaplock=>'']
                [, spcres=>''] );


=over

=item host

(Required): Hostspec of filer

=item aggr

(Optional): Name of aggregate for the volume to live in.
Default: $vol (same name as volume)

=item vol

(Required): Name of volume to be created

=item type

(Optional): Type of volume to be created.  Possible values: 'flex' / 'trad'
Default: 'flex'.

=item vol_size

(Optional): Size of volume to be created.  Will be used iff the type is 'flex'
In 'trad' size of the volume is set to size of hosting aggregate
Default: 100m

=item purge_existing

(Optional): Flag to denote if existing vol must be purged and a 
new one created in its place.  Default: 0

=item create_opts

(Optional): Options to create the volume. Refer create_raid_obj for default
This option will be deprecated soon.

=item disks

(Optional): Number of disks to use for volume / aggregate creation
Will be used iff the type is 'trad' Refer create_raid_obj for default

=item raid_mirror

(Optional): To create a normal or mirrored volume / aggregate. Will be used
iff the type is 'trad'. Refer _get_raid_create_opts for default / possible values

=item cksum_style

Not implemented.
(Optional): Style of checksum to use for volume / aggregate. Will be used
iff the type is 'trad'. Refer _get_raid_create_opts for default / possible values

=item raid_type

(Optional): Type of raid group to use for volume / aggregate creation. Will be
used iff the type is 'trad'. Refer _get_raid_create_opts for default / possible values

=item lang

(Optional): Type of language code to use while creating the volume.  Default: ''

=item spcres Optional: Type of space reservation to use while creating the 
volume. Possible values : none / file / volume Default: ''

=item snaplock

(Optional): Whether to create a snaplock volume,
and which type (compliance or enterprise).
Requires same-type aggregate if flexvol.

=item timeout Optional: Time (in sec) to wait for vol to be created.
Default: 3600 (sec)

=back

Returns 1 if the vol was created by this routine
0 if vol was not created in this routine
will die with $@ if the routine could not complete for some reason

=cut

sub create_volume ( % ) {

    my (%args) = validate (@_, 
                    { 
                     host => { type=>OBJECT, isa=>'Hostrec' },
                     vol => { type=>SCALAR, regex=>qr/\w+/ },
                     aggr => { type=>SCALAR, regex=>qr/\w+/, 
                                optional=>1 },
                     type => { type=>SCALAR, regex=>qr/^flex|trad|sparse$/, 
                                optional=>1, default=>'flex' },
                     create_opts => { type=>SCALAR, optional=>1, default=>'' },
                     vol_size => { type=>SCALAR, regex=>qr/^\d+\w?$/, 
                                optional=>1, default=>'100m' },
                     purge_existing => { type=>SCALAR, regex=>qr/1|0/, 
                                optional=>1, default=>0 },
                     disks => { type=>SCALAR, optional=>1, regex=>qr/^\d+$/, 
                                default=>2 },
                     raid_mirror => { type=>SCALAR, optional=>1, 
                                regex=>qr/^(mirrored|normal)$/, default=>'normal' },
                     cksum_style => { type=>SCALAR, optional=>1, 
                                regex=>qr/^(block|zoned)$/, default=>'' },
                     raid_type => { type=>SCALAR, optional=>1, 
                                regex=>qr/^(raid4|raid_dp)$/, default=>'raid4' },
                     lang => { type=>SCALAR, optional=>1, regex=>qr/^(\w+(\.\w+)?(\.[\w\-]+)?)$/ },
                     spcres => { type=>SCALAR, optional=>1, regex=>qr/volume|file|none/ },
                     snaplock => { type=>SCALAR|UNDEF, optional=>1,
                                   regex=>qr/^(|compliance|enterprise)$/, default=>'' },
                     timeout => { type=>SCALAR, optional=>1, 
                                regex=>qr/^\d+$/, default=>3600 },
                     source => { type=>SCALAR, optional=>1, 
				regex=>qr/^[^:]+:\w+$/ },
                    }
                );

    logresult('warn', 'Argument cksum_style is not implemented.') if ($args{cksum_style});

    my $aggr = defined($args{aggr}) ? $args{aggr} : $args{vol};
    my $conn = make_conn($args{host});

    my $exists = vol_exists($conn, $args{vol});
    if ($exists) {
        if ($args{purge_existing}) { 
            $Log->debug("Will purge existing volume '$args{vol}'");
            vol_purge($conn, $args{vol}) or
                die("Purge of existing vol '$args{vol}' failed :$@\n");
            $exists=0;
            # TBD if its a tradvol and if the aggr did not get purged because
            # it contained other volumes etc... then we will not be able to create
            # the trad vol again since aggr of same name exists
        } else {
            # check if volume is online
            my $status_obj = vol_status($conn, $args{vol});
            die ("Unable to get status of volume $args{vol} : $@") if ($@);
            die("Volume '$args{vol}' exists but is not online. Will not be able to use it.")
                if ($status_obj->state() ne 'online');

            # verify if existing volume / aggregate is what we want
            # if its flex or trad as we required.
            my $ri = new RaidInfo($conn) or
                die("Unable to get raid info\n");
            my $volume = $ri->volume($args{vol}); 
            die("Could not get volume '$args{vol}' information from raid_config")
                if (!$volume);

            die("Existing vol '$args{vol}' is a flexible volume. We require it to be a traditional volume")
                if (($volume->{store_type} eq 'FLEX') && $args{type} eq 'trad');
            die("Existing vol '$args{vol}' is a traditional volume. We require it to be a flexible volume")
                if (($volume->{store_type} eq 'TRAD') && $args{type} eq 'flex');
            
            # Determine if the volumes size is compatible with this request
            # otherwise try and adjust it
            my $vol_size_in_kb = vol_size($conn, $args{vol});
            my $size_in_bytes     = convert_to_bytes( $args{vol_size} );
            my $requested_size = convert_units( $size_in_bytes, UNIT_B, UNIT_KB );
            if ($vol_size_in_kb >= $requested_size) {
              $Log->debug("Will use existing volume size $vol_size_in_kb");
            } else {
              vol_size($conn, $args{vol}, $args{vol_size});
            }
             
            $Log->debug("Will use existing volume '$args{vol}'");
        }
    }

    my $we_created = 0;
    if (!$exists) {
        if (($args{type} eq 'flex') || ($args{type} eq 'sparse')) {
            # Create flex volume
            # use vol_create_flex from Volume.pm for flex volumes
            # copy relevant args to pass down to create_volume()
            my %vol_args = ();
            $vol_args{language} = $args{lang} if ($args{lang});
            $vol_args{guarantee} = $args{spcres} if ($args{spcres});
            $vol_args{source} = $args{source} if ($args{source});
            $vol_args{timeout} = $args{timeout} if ($args{timeout});
	    if ($args{snaplock} eq 'compliance') {
	        $vol_args{snaplock} = 1;
	    } elsif ($args{snaplock} eq 'enterprise') {
	        $vol_args{snaplock_enterprise} = 1;
	    }
            vol_create_flex($conn, $args{vol}, $args{aggr}, $args{vol_size},
                            %vol_args);
        } else {
            # Create trad volume
            if (0) { # TBD implement this part of getting raid options later
            if (!$args{create_opts}) {
                # get raid create options
                eval {
                    $args{create_opts} = _get_raid_create_opts($args{host}, 
                                disks=>$args{disks},
                                cksum_style=>$args{cksum_style},
                                raid_type=>$args{raid_type},
                                raid_mirror=>$args{raid_mirror}
                                );
                };
               die ($@) if ($@);
            }
            }
            # use vol_create from Volume.pm for trad volumes
            my %vol_args = ();
            #$vol_args{force} = TBD
            $vol_args{language} = $args{lang} if ($args{lang});
            $vol_args{mirror} = ($args{raid_mirror} eq 'mirrored' ? 1 : 0) if ($args{raid_mirror});
            $vol_args{raidtype} = $args{raid_type} if ($args{raid_type});
            #$vol_args{randomize} TBD
            #$vol_args{rgsize} = $args{} ???? TBD
	    if ($args{snaplock} eq 'compliance') {
	        $vol_args{snaplock} = 1;
	    } elsif ($args{snaplock} eq 'enterprise') {
	        $vol_args{snaplock_enterprise} = 1;
	    }
            $vol_args{timeout} = $args{timeout} if ($args{timeout});
            #$vol_args{purge} not needed its taken care before reaching here
            vol_create($conn, $args{vol}, 
                                $args{disks}, # vol_create accepts ARRAYREF too TBD
                                %vol_args);
        }
        die("Volume '$args{vol}' could not be created :$@") if ($@);
# WIP: why do this instead of using the vol_create() return value?
        $we_created = 1;
    } else {
	# Check if aggregate should be converted to 64-bit. We are
	# converting the aggregate during volume creation because
	# the goal is to test with as many hybrid buftrees as we
	# can. If the aggregate was converted as part of aggregate
	# creation, all its volumes would have pure 64-bit buftrees.
        if (($args{type} eq 'flex') && $main::CONVERT_TO64) {
            # XXX This check should be changed to
            # >= WV_ROLLINGROCK_MIN when that value is defined
            if (get_wafl_version($conn->hostp()) > WV_BOILERMAKER_MAX) {
                if (aggr_is_convertible($conn, $args{aggr})) {
		    logcomment("Converting aggregate $aggr to a mixed-format (hybrid) buftree");
                    aggr_convert($conn, $args{aggr});
                } else {
                    logcomment("Skipping conversion: aggregate $args{aggr} is root or already 64-bit");
                }
            } else {
                croak("filer version does not support converting to 64-bit aggregates");
            }
        }
    }

    return($we_created);

}


=head2 _get_raid_create_opts()

This routine will query the filer and decide the create options string that 
can be used to create the desired raid object (trad volume / aggregate)

    _get_raid_create_opts($host,
                    [, disks=>2]
                    [, raid_mirror=>'normal']
                    [, cksum_style=>'']
                    [, raid_type=>'raid4']);

=over

=item spec

(Required): Hostspec of filer

=item disks

(Optional): Number of disks to use for volume / aggregate creation. Default: 2

=item raid_mirror

(Optional): To create a normal or mirrored volume / aggregate.
Possible values: 'normal' / 'mirrored'. Default: 'normal'

=item cksum_style

(Optional): Style of checksum to use for volume / aggregate.
Possible values: 'zoned' / 'block' / ''
When '' then ontap will decide.  Default: ''

=item raid_type

(Optional): Type of raid group to use for volume / aggregate creation.
Possible values: 'raid4' / 'raid_dp'. Default: 'raid4'

=back

On success, returns a string of the create options that can be used to create the vol/aggr
Will die with $@ if the routine could not complete for some reason

Dependencies are

Need equal number of disks in each plex for mirrored volume / aggregate. 
We assume number of plexes is 2, so we need even number 
of disks to make a mirrored volume / aggregate
Need a min of 2 disks to make a raid4 volume / aggregate
Need a min of 3 disks to make a raid_dp volume / aggregate

Notes

does not take into consideration the size of the spares.
does not take into consideration the % of zeroing of the spares.
flex volume is a wafl object and is not created by this routine

=cut

sub _get_raid_create_opts ( $% ) {

    my ($spec, %args) = @_;

    my $conn = make_conn($spec);

    # set the defaults for the optional arguments
    my $disks = defined($args{disks}) ? $args{disks} : 2;
    my $cksum_style = defined($args{cksum_style}) ? $args{cksum_style} : '';
    my $raid_mirror = (defined($args{raid_mirror}) && ($args{raid_mirror} eq 'mirrored')) 
                        ?  $args{raid_mirror} : 'normal';
    my $raid_type = (defined($args{raid_type}) && ($args{raid_type} eq 'raid_dp')) 
                        ? $args{raid_type} : 'raid4';

    $Log->debug("Will generate string to create raid object ". 
                "disks:$disks cksum_style:$cksum_style". 
                " raid_mirror:$raid_mirror raid_type:$raid_type");


    #-------------------------
    # do inital error checking
    #-------------------------
    # calculate minumum disks that are required for the volume / aggregate
    my $num_plex = 2;  # for now ontap supports only 2 plexes
    my $min_disks = 2; # min number of disks to create a basic aggr is 2
    $min_disks += 1 if ($raid_type eq 'raid_dp');
    $min_disks *= $num_plex if ($raid_mirror eq 'mirrored');

    croak("Need minimum of $min_disks disks to make this raid object\n")
        if ($disks < $min_disks);

    # exit if asked for mirrored volume / aggregate but # disks is odd
    croak("Need even number of disks ". 
        "\(you asked for $disks disks\) to make a mirrored raid object\n")
        if (($raid_mirror eq 'mirrored') && ($disks % 2));

    my $ri = new RaidInfo($conn) or 
        die("Unable to get raid info\n");

    # check if number of spares are available
    my %spare_disks =  $ri->spare_disks();
    croak("Not enough spare disks. Require $disks in spare pool.\n")
        if (scalar(keys(%spare_disks)) < $disks);

    # check if syncmirror is setup if aggr to be created is to be mirrored
    my (%spare_disks_pool0, %spare_disks_pool1);
    my $disks_pool;
    if ($raid_mirror eq 'mirrored') {
        # check if syncmirror is licensed
        my (undef, undef, $out) = say($conn, "license");
        croak("Raid mirror (syncmirror_local) is not licensed.\n".
                "Pls license, reboot and re-run this test\n")
                if ($out =~ /syncmirror_local\s+not licensed/);

        # check if there are 2 pools on the filer
        %spare_disks_pool0 = $ri->spare_disks_pool0();
        %spare_disks_pool1 = $ri->spare_disks_pool1();
        croak("Disks are not in 2 pools.\nUnable to created a mirrored raid object\n")
            unless (scalar(keys(%spare_disks_pool1)) > 0);

        # check if there are enough spares in both pools
        $disks_pool = $disks / 2;
        croak("Not enough \(require $disks_pool\) disks in both spare pools.\n")
            if ((scalar(keys(%spare_disks_pool0)) < $disks_pool) ||
                (scalar(keys(%spare_disks_pool1)) < $disks_pool));
    }

    #-------------------------
    # start assembling the create options for the volume / aggregate
    #-------------------------

    my $create_opts = "";
    $create_opts .= "-m " if ($raid_mirror eq 'mirrored');
    $create_opts .= "-t $raid_type ";

    #-------------------------
    # branch based on the cksum_style
    #-------------------------
    if ($cksum_style eq '') {

        # let ontap take care and pick the disks
        $create_opts .= "$disks";

    } else {

        # Check if we have required checksum style disks; else exit
        # For 'zoned' all disks (512/520 bps) can be configured in zoned checkum style.
        # whereas 'block' checksum style requires disks of 520bps.

        # get a list of 520 and 512 bps disks
        my (@bps_512_spares, @bps_520_spares,
            @bps_512_spares_pool0, @bps_520_spares_pool0,
            @bps_512_spares_pool1, @bps_520_spares_pool1);

        @bps_512_spares = _get_matching_sub_list ( 'bps', '512', %spare_disks );
        @bps_520_spares = _get_matching_sub_list ( 'bps', '520', %spare_disks );
        $Log->debug("512bps spares are @bps_512_spares");
        $Log->debug("520bps spares are @bps_520_spares");

        if ($raid_mirror eq 'mirrored') {
            @bps_512_spares_pool0 = _get_matching_sub_list ( 'bps', '512', %spare_disks_pool0 );
            @bps_512_spares_pool1 = _get_matching_sub_list ( 'bps', '512', %spare_disks_pool1 );
            @bps_520_spares_pool0 = _get_matching_sub_list ( 'bps', '520', %spare_disks_pool0 );
            @bps_520_spares_pool1 = _get_matching_sub_list ( 'bps', '520', %spare_disks_pool1 );
            $Log->debug("512bps spares in pool0 are @bps_512_spares_pool0");
            $Log->debug("512bps spares in pool1 are @bps_512_spares_pool1");
            $Log->debug("520bps spares in pool0 are @bps_520_spares_pool0");
            $Log->debug("520bps spares in pool1 are @bps_520_spares_pool1");
        }

        if ($cksum_style eq 'block') {

            # TBD need to review this - since this section should check for 520bps disks not 512
            if (scalar(@bps_512_spares)) {
                # randomly pick required number of disks from the list 
                # here we take control and select the disks to be used
                if ($raid_mirror eq 'mirrored') {

                    # TBD we could use -f option to acheive this
                    croak("Not enough \(require $disks_pool\) 520 bps in ".
                        "each spare pool to create $cksum_style raid object\n")
                        if ((scalar(@bps_520_spares_pool0) < $disks_pool) || 
                            (scalar(@bps_520_spares_pool1) < $disks_pool));

                    # could look for zeroed disks
                    my @chosen_disks_pool0 = _get_rand_sublist($disks_pool, @bps_520_spares_pool0);
                    my @chosen_disks_pool1 = _get_rand_sublist($disks_pool, @bps_520_spares_pool1);
                    $create_opts .= "-d @chosen_disks_pool0 -d @chosen_disks_pool1";

                } else {
                    croak("Not enough \(require - $disks\) spares of 520 ". 
                        "bps to create $cksum_style raid object\n")
                        if (scalar(@bps_520_spares) < $disks);
                    my @chosen_disks = _get_rand_sublist($disks, @bps_520_spares);
                    $create_opts .= "-d @chosen_disks";
                }
            } else {
                # there are no 512 bps spare disks on the filer so just go ahead and
                # create the vol / aggr with $disks and ontap will do the disk selection
                $create_opts .= "$disks";
            }

        } elsif ($cksum_style eq 'zoned') {

            if (scalar(@bps_520_spares)) {
                if ($raid_mirror eq 'mirrored') {
                    croak("No spares \(require $disks_pool\) ". 
                        "of 512 bps in both spare pools to create ". 
                        "$cksum_style raid object\n")
                        unless ((@bps_512_spares_pool1) && (@bps_512_spares_pool0));

                    if (scalar(@bps_512_spares_pool0) < $disks_pool) {
                        # not enough disks in 512 pool0 to make the zoned checksum vol/aggr
                        # we can see if there are some in the bps_520_spares pool0 list
                        if (scalar(@bps_520_spares_pool0) > (scalar(@bps_512_spares_pool0) - $disks_pool)) {
                            $Log->debug("Did not get enough spares 512 bps in pool0. ". 
                                        "\nWill look in the 520 bps pool0 spare list");
                            my @additional_spares = _get_rand_sublist(scalar(@bps_512_spares_pool0) - $disks_pool, 
                                                                    @bps_520_spares_pool0);
                            push(@bps_512_spares_pool0, @additional_spares);
                        } else {
                            croak("Not enough \(require $disks_pool\) spares", 
                                        " of required bps in pool0 to". 
                                        " create $cksum_style raid object\n");
                        }
                    }
                    if (scalar(@bps_512_spares_pool1) < $disks_pool) {
                        # not enough disks in 512 pool0 to make the zoned checksum vol/aggr
                        # we can see if there are some in the bps_520_spares pool1 list
                        if (scalar(@bps_520_spares_pool1) > 
                            (scalar(@bps_512_spares_pool1) - $disks_pool)) {
                            $Log->debug("Did not get enough ". 
                                        "spares 512 bps in pool1. \nWill look ". 
                                        "in the 520 bps pool1 spare list");
                            my @additional_spares = _get_rand_sublist(scalar(@bps_512_spares_pool1) 
                                                        - $disks_pool, @bps_520_spares_pool1);
                            push(@bps_512_spares_pool1, @additional_spares);
                        } else {
                            croak("Not enough \(require $disks_pool\) spares", 
                                        " of required bps in pool1 to". 
                                        " create $cksum_style raid object\n");
                        }
                    }
                    my @chosen_disks_pool0 = _get_rand_sublist($disks_pool, @bps_512_spares_pool0);
                    my @chosen_disks_pool1 = _get_rand_sublist($disks_pool, @bps_512_spares_pool1);
                    $create_opts .= "-d @chosen_disks_pool0 -d @chosen_disks_pool1";

                } else {

                    croak("No spares \(require $disks\) of 512 bps ". 
                        " to create $cksum_style raid object\n")
                        unless (@bps_512_spares);

                    if (scalar(@bps_512_spares) < $disks) {
                        # not enough disks in 512 to make the zoned checksum vol/aggr
                        # we can see if there are some in the bps_520_spares list
                        # because the 520 when combined with 512 for an vol/aggr
                        # behaves as a zoned checksum style
                        if (scalar(@bps_520_spares) > (scalar(@bps_512_spares) - $disks)) {
                            $Log->debug("Did not get enough ". 
                                        "spares in 512 bps. ". 
                                        "Will look in the 520 bps list");
                            my @additional_spares = _get_rand_sublist(
                                                        scalar(@bps_512_spares) - $disks, 
                                                        @bps_520_spares);
                            push(@bps_512_spares, @additional_spares);
                        } else {
                            croak("Not enough \(require $disks\) spares ".
                                "of required bps to create ".
                                "$cksum_style raid object\n");
                        }
                    }
                    # randomly pick required number of disks from the list 
                    # here we take control and select the disks to be used
                    my @chosen_disks = _get_rand_sublist($disks, @bps_512_spares);
                    $create_opts .= "-d @chosen_disks";
                }

            } else {
                # there are no 520 bps disks on the filer so just go ahead and
                # create the vol/aggr with $disks and ontap will do disk selection
                $create_opts .= "$disks";
            }
        }
    }

    return($create_opts);

}


# _get_rand_sublist

# Randomly selects $num unique elements from @list
# TBD Improve to pick the ones that are already zeroed

sub _get_rand_sublist ( $@ ) {

    my ($num, @list) = @_;

    eval {
    my @sublist;
    die ("Not enough elements in original list.\n")
        if (scalar(@list) < $num);

    return(@list) if (scalar(@list) == $num);

    while (($num) && (@list))  {
        my $index = int(rand(@list));
        my $element = $list[$index];
        $Log->debug("Adding $element to sub list");
        push(@sublist, $element);
        $num--;
        splice(@list,$index,1);
    }
    return(@sublist);
    };
}

# _get_matching_sub_list

sub _get_matching_sub_list ( $$% ) {

    my ($item, $match, %hash) = @_;

    my @new_list;
    my @keys = keys(%hash);
    foreach my $index (@keys) {
        push(@new_list, $index) if ($hash{$index}{$item} eq $match);
    }
    return (@new_list);
}

1;


=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-waflqa-art, archana@netapp.com, dl-test-infra@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/Volume/Setup.pm#2 $
    Last revised $Date: 2013/04/24 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

