# $Id: //depot/prod/test/Rsierranevada/lib/Volume/Transition.pm#1 $
# Copyright (c) 2009-2010 NetApp, Inc.
# All rights reserved.

## @summary     programmatic control of volume transition.
## @author      astraub@netapp.com
## @status      private
## @pod         here

=head1 NAME

Volume::Transition.pm

=head1 SYNOPSIS

=begin man

    use Volume::Transition;
    use Volume::Transition::BaseException;
    use strict;

    my $vt1 = new Volume::Transition(node=>$node, vserver => 'myvserver');
    $vt1->add_volume(volume=>"vol1", 'junction-path' => '/vol1');

    #Automatic transition to cluster mode
    try {
        $vt1->transition_check();
        $vt1->transition();
    } catch Volume::Transition::BaseException with {
        my $exception = shift;
        $Log->error("Transition failure:".$exception->text());
        $exception->throw();
    };

    my $vt2 = new Volume::Transition(node=>$node);
    $vt2->add_volume(volume=>"vol2", 'junction-path' => "/vol2");

    # Controlled transition to c-mode using zapi calls

    try {
        $vt2->prepare('transition-type'=>'transition');
        $vt2->wait();
        $vt2->commit();
        $vt2->wait();
        $vt2->online();
        $vt2->wait();
        $vt2->cleanup();
        $vt2->wait();
    } catch Volume::Transition::BaseException with {
        my $exception = shift;
        $Log->error("Transition failure:".$exception->text());
        $exception->throw();
    };

=end man


=head1 DESCRIPTION

C<Volume::Transition> is a class for performing volume transition testing.

It provides methods to perform transition via the cli as well as
via zapi calls. The class throws exceptions on failures.

This class makes use of its own exception class
Volume::Transition::BaseException. While the exception class will be
inherited when using this library, it is recommended to explicitly import
the exception class for clarity.

The current list of possible exceptions are:
 Volume::Transition::BaseException;
 Volume::Transition::ZapiException;
 Volume::Transition::CliException;
 Volume::Transition::InvalidOperationType;
 Volume::Transition::NoVolumes;
 Volume::Transition::JunctionParamInvalid;
 Volume::Transition::InvalidState;
 Volume::Transition::JobNotFound;
 Volume::Transition::JobConflict;
 Volume::Transition::TooManyVols;
 Volume::Transition::JobTerminated;

=head1 METHODS

=cut

package Volume::Transition;
use strict;
use Params::Validate
  qw(validate validate_with BOOLEAN SCALAR HASHREF ARRAYREF UNDEF OBJECT);
use Tharn qw($Log host snooze);
use Data::Dumper;
use NATE::BaseException qw(:try);
use Volume::Transition::BaseException;
use ZapiTest;
use NaElement;
use ZapiErrno;
use Server;

=head2 new

C<< Volume::Transition->new(%params) >>

Invocant for the Volume::Transition class.

=over 4

=item Input

=over 4

=item * vserver => <scalar>

(Optional)
Default = undef

Defines vserver name used in transition operations.
For transition, it is the destination vserver, and for
untransition, it is the source vserver.

=item * node => <scalar>

(Required)
Default = undef
Node that is the target of cli commands or zapi calls

=item * source-node => <scalar>

(Optional)
Default = <$node>
Node that is hosting the volume to be transitioned
Defaults to the value defined in the "node" parameter


=back

=item Output

Class Reference

=item Example

C<< my $vt_obj = Volume::Transition->new('node' => $string,
                                      'vserver' => $vserver); >>

=back

=cut

################################################################################
#
# Subroutine:   new()
#
# Objective:    Invocant for the class
#
# Details:
#
#   Input:
#       vserver  - c-mode vserver used in transition
#       node     - target of the cli commands or zapi calls
#       source-node - node hosting volume to be transitioned
#
#   Return value:
#       object
#
################################################################################

sub new {
    $Log->enter();
    my $class  = shift;
    my $defvol = "p_voltrans_" . time();
    my $self = {
                validate(
                         @_,
                         {
                          vserver => {
                                      type     => SCALAR,
                                      optional => 1,
                                     },
                          node => {
                                   type     => SCALAR,
                                   optional => 0,
                                  },
                          'source-node' => {
                                         type     => SCALAR | UNDEF,
                                         optional => 1,
                                         default  => undef,
                                        },
                         }
                        )
               };

    if (!defined($self->{'source-node'})) {
        $self->{'source-node'} = $self->{'node'};
    }
    $Log->comment("Input Params: " . Dumper($self));
    my $host = host($self->{node});
    $Log->comment("HOST $self->{node} info" . Dumper($host));

    #Create the zapi class object
    my $Zserver = new ZapiTest($host->{MGMT_IP}, "admin", 'netapp1!');
    $Zserver->set_legacy_http_conn(0);

    #Create the "volumes" data element for future use
    my $zvolumes = new NaElement("volumes");
    $self->{zvol_elems}     = $zvolumes;
    $self->{trans_state}    = "idle";
    $self->{can_transition} = 1;
    $self->{zapi_srv}       = $Zserver;

    $self->{server_obj} = new Server;
    bless($self, $class);
    $Log->exit();
    return $self;
} ## end sub new

=head2 add_volume

C<< Volume::Transition->add_volume(%params) >>

Adds volumes to the transition class to use for transition operations.
For CLI based transitions, only the most recently added volume will be
used as the cli only supports transition of a single volume.

=over 4

=item Input

=over 4

=item * volume => <scalar>

(Required)
Default = undef
Defines vserver name used in transition operations.
For transition, it is the destination vserver, and for
untransition, it is the source vserver.

=item * junction-path => <scalar>

(Optional)
Default = undef
location in c-mode namespace to mount the volume

=back

=item Output

boolean

=item Example

C<< $vt_obj->add_volume(volume => "foovol",
                    'junction-path' => "/foovol"); >>

=back

=cut

################################################################################
#
# Subroutine:   add_volume()
#
# Objective:    Add a volume to the transition class
#
# Details:
#
#   Input:
#       volume         - Name of volume
#       junction-path  - location in c-mode namespace to mount the volume
#
#   Return value:
#       boolean
#
################################################################################

sub add_volume {
    my $self = shift;
    my $params = validate(
                          @_,
                          {
                           volume => {
                                      type     => SCALAR,
                                      optional => 0,
                                     },
                           'junction-path' => {
                                               type     => SCALAR,
                                               optional => 1,
                                              },
                          }
                         );
    $Log->enter();
    my $zvolumes = $self->{zvol_elems};
    my $volhash;

    #Create the volume-transition-volinfo data structure
    my $zvolume = new NaElement("volume-transition-volinfo");

    #push the volume name, and other child elements if they are defined
    $zvolume->child_add_string("volume-name", $params->{volume});
    $volhash->{volume} = $params->{volume};
    if (defined $params->{'junction-path'}) {
        $zvolume->child_add_string("junction-path", $params->{'junction-path'});
        $volhash->{'junction-path'} = $params->{'junction-path'};
    }

    #For the CLI, we only act on the last volume added
    #since we can only transition one volume at a time.
    #This needs to be improved.
    $self->{volume} = $params->{volume};
    $self->{'junction-path'} = $params->{'junction-path'};
    $zvolumes->child_add($zvolume);

    push(@{$self->{volumes}}, $volhash);
    $Log->exit();
    return 1;
} ## end sub add_volume

################################################################################
#
# Subroutine:   _create_vol_elems()
#
# Objective:    Create a transition volinfo data structure for use in zapi calls
#               from the list of volumes that were added to the class
#
# Details:
#
#   Input:
#       need-junction  - Include junction path in the volinfo structure
#
#   Return value:
#       NaElement object
#
################################################################################

sub _create_vol_elems {
    $Log->enter();
    my $self = shift();
    my $params = {
                  validate(
                           @_,
                           {
                            'need-junction' => {
                                                type     => BOOLEAN,
                                                optional => 1,
                                               },
                           }
                          )
                 };
    my $zvolumes = new NaElement("volumes");
    for my $vol (@{$self->{volumes}}) {
        my $zvolume = new NaElement("volume-transition-volinfo");
        $zvolume->child_add_string("volume-name", $vol->{volume});
        if ($params->{'need-junction'} && defined $vol->{'junction-path'}) {
            $zvolume->child_add_string("junction-path",
                                       $vol->{'junction-path'});
        }
        $zvolumes->child_add($zvolume);
    }
    $Log->exit();
    return $zvolumes;
} ## end sub _create_vol_elems


=head2 transition

C<< Volume::Transition->transition(%params) >>

Perform transition to c-mode on volumes defined in the class.
For CLI based transitions, only the most recently added volume will be
used as the cli only supports transition of a single volume.

=over 4

=item Input

=over 4

=item * method => <scalar>

(Optional)
Default = zapi
Access method to use in transition (cli|zapi|zapi-nonstaged)

=item * %prepare params

refer to the params accepted by the prepare() method

=back

=item Output

returns 1 on success, throws exception on failure

=item Example

C<< $vt_obj->transition(); >>

=back

=cut

################################################################################
#
# Subroutine:   transition()
#
# Objective:    Perform transition to c-mode on volumes defined in the class
#
# Details:
#
#   Input:
#       method          - Access method to use in transition (cli|zapi|zapi-nonstaged)
#       transition-type - Type of transition to perform (transition|conversion)
#       %prepare params - Refer to parameters accepted by the prepare() method
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub transition {
    my $self = shift();
    my $ret;
    $Log->enter();
    my $params = {
        validate_with(
            params => \@_,
            spec   => {
                'method' => {
                    type      => SCALAR,
                    optional  => 1,
                    default   => 'zapi',
                    callbacks => {
                                'zapi, zapi-nonstaged or cli' => sub { shift() =~ /^(zapi|zapi-nonstaged|cli)$/ }
                                 },
                            },
                'operation-type' => {
                    type      => SCALAR,
                    optional  => 1,
                    callbacks => {
                        'copy_based or in_place' => sub { shift() =~ /^(copy_based|in_place)$/ }
                    },
                },
                'affinity-node' => {
                    type      => SCALAR,
                    optional  => 1,
                },
                    },
            allow_extra => 1,
                     ),
                 };
    my $method = $params->{method};
    delete $params->{method};
    if ($method eq 'zapi') {
        $ret = $self->_zapi_trans_to_mode(%$params);
    } elsif ($method eq 'zapi-nonstaged') {
        $ret = $self->_zapi_trans(%$params);
    } else {
        $Log->comment("DOING CLI TRANSITION");
        $ret = $self->_cli_execute(%$params);
    }
    $Log->exit();
    return $ret;
} ## end sub transition

################################################################################
#
# Subroutine:   _zapi_trans_to_mode()
#
# Objective:    Perform complete transition using zapi methods to specified mode
#
# Details:
#
#   Input:
#       see the params defined in prepare()
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub _zapi_trans_to_mode {
    my $self = shift();
    $Log->enter();
    $Log->debug("INVAR:" . Dumper(\@_));
    my $params = {@_};
    $Log->debug("Params:" . Dumper($params));
    try {
        $self->prepare(%$params);
        $self->wait();
        $self->commit();
        $self->wait();
        $self->online();
        $self->wait();
        $self->cleanup();
        $self->wait();
    } ## end try
    catch NATE::BaseException with {
        my $ex = shift;
        $Log->warn("Error occurred during transition: " . $ex->text());
        $Log->warn("Call stack:" . $ex->stacktrace());
        $ex->throw();
    };

    return 1;
} ## end sub _zapi_trans_to_mode

################################################################################
#
# Subroutine:   _zapi_trans()
#
# Objective:    Perform complete transition using zapi nonstaged method
#
# Details:
#
#   Input:
#       see the params defined in prepare()
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub _zapi_trans {
    my $self = shift();
    $Log->enter();
    $Log->debug("INVAR:" . Dumper(\@_));
    my $params = {@_};
    $Log->debug("Params:" . Dumper($params));
    try {
        $self->transition_nonstaged(%$params);
        $self->wait();
    } ## end try
    catch NATE::BaseException with {
        my $ex = shift;
        $Log->warn("Error occurred during transition: " . $ex->text());
        $Log->warn("Call stack:" . $ex->stacktrace());
        $ex->throw();
    };

    return 1;
} ## end sub _zapi_trans

=head2 transition_check

C<< Volume::Transition->transition_check(%params) >>

Perform transition check to verify volumes defined in the class can be transitioned
For CLI based check, only the most recently added volume will be
used as the cli only supports transition of a single volume.

=over 4

=item Input

=over 4

=item * method => <scalar>

(Optional)
Default = zapi
Access method to use in transition (cli|zapi)


=item * %check params

refer to the params accepted by the check() method

=back

=item Output

returns 1 on success, throws exception on failure

=item Example

C<< $vt_obj->transition_check(); >>

=back

=cut

################################################################################
#
# Subroutine:   transition_check()
#
# Objective:    Perform transition check to verify volumes defined in the class
#               can be transitioned
#
# Details:
#
#   Input:
#       method          - Access method to use in transition check (cli|zapi)
#       transition-type - Type of transition to check (transition|conversion)
#       %check params   - Refer to parameters accepted by the check() method
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub transition_check {
    my $self = shift();
    $Log->enter();
    my $params = {
        validate_with(
            params => \@_,
            spec   => {
                'method' => {
                    type      => SCALAR,
                    optional  => 1,
                    default   => 'zapi',
                    callbacks => {
                                'zapi or cli' => sub { shift() =~ /^(zapi|cli)$/ }
                                 },
                            },
                'operation-type' => {
                    type      => SCALAR,
                    optional  => 1,
                    callbacks => {
                        'copy_based or in_place' => sub { shift() =~ /^(copy_based|in_place)$/ }
                    },
                },
                    },
            allow_extra => 1,
                     ),
                 };
    my $ret;
    my $method = $params->{method};
    delete $params->{method};
    if ($method eq 'zapi') {
        $ret = $self->check(%$params);
    } else {
        $Log->comment("DOING CLI TRANSITION CHECK");
        $ret = $self->_cli_execute(%$params, 'check-only'=>'true');
    }
    $Log->exit();
    return $ret;
} ## end sub transition_check

=head2 transition_recover

C<< Volume::Transition->transition_recover(%params) >>

Perform transition recover to recover a failed volume transition.

=over 4

=item Input

=over 4

=item * vserver => <scalar>

(Optional)
Vserver to be used for transition recover operation

=item * source-node => <scalar>

(Optional)
Node hosting volumes

=item * action => <scalar>

(Optional)
go-forward or go-backward

=back

=item Output

returns 1 on success, throws exception on failure

=item Example

C<< $vt_obj->transition_recover(); >>

=back

=cut

################################################################################
#
# Subroutine:   transition_recover()
#
# Objective:    Perform transition recover to recover a failed volume transition
#
# Details:
#
#   Input:
#       vserver                  - vserver involved in transition recover
#       source-node              - node hosting the volumes
#       action                   - go-forward (default) or go-backward
#       affinity-node            - node where the job should be queued
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub transition_recover {
    my $self = shift();
    $Log->enter();
    my $params = {
        validate(
            @_,
            {
                'vserver' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $self->{vserver},
                },
                'source-node' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $self->{'source-node'},
                },
                'action' => {
                    type     => SCALAR,
                    optional => 1,
                    default  => 'rollforward',
                    callbacks => {
                        'rollback or rollforward' => sub {
                            !defined($_[0]) || $_[0] =~ /^roll(?:back|forward)$/;
                        },
                    },
                },
                'affinity-node' => {
                                    type     => SCALAR,
                                    optional => 1,
                                   },
            },
        ),
    };
    $Log->comment("Params:" . Dumper($params));

    # Volume is not passed in, it is stored internally
    my $cli_cmd = "set test;volume transition-recover -volume $self->{volume} ";
    $cli_cmd .= "-vserver $params->{vserver} " if ($params->{vserver});
    $cli_cmd .= "-source-node $params->{'source-node'} " if ($params->{'source-node'});
    $cli_cmd .= "-action $params->{action} ";
    $cli_cmd .= "-affinity-node $params->{'affinity-node'} "
        if ($params->{'affinity-node'});

    $Log->comment("do: " . $cli_cmd);

    my $out;
    eval {
        $out = $self->{server_obj}->cli(cmd => $cli_cmd);
    };
    if ($@) {
        _cli_error_handler($@);
    }

    if ($out =~ /\[Job (\d+)\]/) {
        $self->{job_id} = $1;
        $Log->comment("Job ID: $self->{job_id}");
        $self->{trans_state} = "cli";
    } else {
        $Log->warn("Could not obtain job id");
        throw Volume::Transition::BaseException("Could not obtain job id");
    }

    $Log->exit();
    return 1;
} ## end sub transition_recover

################################################################################
#
# Subroutine:   _cli_execute()
#
# Objective:    execute the specified transition operation
#               using the cli and return the result
#
# Details:
#
#   Input:
#       transition-type          - type of operation (transition/untransition/conversion)
#       non-disruptive           - set non-disruptive flag
#       override-warnings        - set override warnings flag
#       vserver                  - vserver involved in (un)transition
#       source-node              - node hosting the volumes
#       check-only               - don't transition, only check
#       affinity-node            - node where the job should be queued
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub _cli_execute {
    my $self = shift();
    $Log->enter(_wrap_msg(""));

    my $params = {
        validate(
            @_,
            {
             'operation-type' => {
                 type      => SCALAR,
                 optional  => 1,
                 callbacks => {
                     'copy_based or in_place' => sub {
                         shift() =~ /^(copy_based|in_place)$/;
                       }
                 },
             },
             'non-disruptive' => {
                                 type     => BOOLEAN,
                                 optional => 1,
                                },
             'override-warnings' => {
                                     type     => BOOLEAN,
                                     optional => 1,
                                    },
             'vserver' => {
                           type     => SCALAR,
                           optional => 1,
                           default  => $self->{vserver},
                          },
             'source-node' => {
                            type     => SCALAR,
                            optional => 1,
                            default  => $self->{'source-node'},
                           },
             'check-only' => {
                              type     => SCALAR,
                              optional => 1,
                             },
             'affinity-node' => {
                                 type     => SCALAR,
                                 optional => 1,
                                },
            },
        ),
    };

    $Log->comment("Params:" . Dumper($params));
    my $op_type = $params->{'operation-type'};
    my $cli_cmd = "set diag;volume transition ";

    # Volume and junction path are not passed in, they are stored internally
    $cli_cmd .= "-volume $self->{volume} ";
    $cli_cmd .= "-junction-path $self->{'junction-path'} "
        if ($self->{'junction-path'});

    $cli_cmd .= "-vserver $params->{'vserver'} " if ($params->{'vserver'});
    $cli_cmd .= "-source-node $params->{'source-node'} ";

    $cli_cmd .= "-check-only $params->{'check-only'} "
        if ($params->{'check-only'});
    $cli_cmd .= "-operation-type $op_type " if ($op_type);
    $cli_cmd .= "-affinity-node $params->{'affinity-node'} "
        if ($params->{'affinity-node'});

    if ($op_type eq 'in_place' ) {
        $cli_cmd .= "-suppress-warnings $params->{'suppress-warnings'} "
            if ($params->{'suppress-warnings'});
        $cli_cmd .= "-non-disruptive $params->{'non-disruptive'} "
            if ($params->{'non-disruptive'});
    }

    $Log->comment("do: " . $cli_cmd);

    my $out;
    eval {
        $out = $self->{server_obj}->cli(cmd => $cli_cmd);
    };
    if ($@) {
        _cli_error_handler($@);
    }

    if ($params->{'check-only'} !~ /^true$/i) {
        if ($out =~ /\[Job (\d+)\]/) {
            $self->{job_id} = $1;
            $Log->comment("Job ID: $self->{job_id}");
            $self->{trans_state} = "cli";
        } else {
            $Log->warn("Could not obtain job id");
            throw Volume::Transition::BaseException("Could not obtain job id");
        }
    } ## end if ($params->{'check-only'...
    $Log->exit();
    return 1;
} ## end sub _cli_execute

=head2 check

C<< Volume::Transition->check(%params) >>

Perform zapi volume-transition-check to verify volumes defined in the class
can be (un)transitioned. For CLI based check, only the most recently
added volume will be used as the cli only supports transition of a single volume.

=over 4

=item Input

=over 4

=item * transition-type => <scalar>

(Required)
Type of transition operation to perform (transition|untransition|conversion)

=item * non-disruptive => <scalar>

(Optional)
Set non-disruptive flag for transition operation

=item * override-warnings => <scalar>

(Optional)
Set override-warnings flag to force transition

=item * vserver => <scalar>
b

(Optional)
Vserver to be used for (un)transition operation

=item * source-node => <scalar>

(Optional)
Node hosting volumes

=back

=item Output

boolean

=item Example

C<< $vt_obj->check('transition-type'=>'transition'); >>

=back

=cut

################################################################################
#
# Subroutine:   check()
#
# Objective:    Perform zapi check to verify volumes defined in the class
#               can be (un)transitioned
#
# Details:
#
#   Input:
#       transition-type          - transition type to check for
#       non-disruptive           - set non-disruptive flag
#       override-warnings        - set override warnings flag
#       vserver                  - vserver involved in (un)transition
#       source-node              - node hosting the volumes
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub check {
    my $self = shift();
    $Log->enter(_wrap_msg("BEGIN TRANSITION CHECK ZAPI"));

    my $params = {
        validate(
            @_,
            {
             'operation-type' => {
                               type     => SCALAR,
                               optional => 1,
                               # NOTE: We don't have a callback here to
                               # validate the argument as we want to be able
                               # to pass bogus values for testing.
                              },
             'non-disruptive' => {
                                 type     => BOOLEAN,
                                 optional => 1,
                                },
             'override-warnings' => {
                                     type     => BOOLEAN,
                                     optional => 1,
                                    },
             'vserver' => {
                           type     => SCALAR,
                           optional => 1,
                           default  => $self->{vserver},
                          },
             'source-node' => {
                            type     => SCALAR,
                            optional => 1,
                            default  => $self->{'source-node'},
                           },
            },
        ),
    };

    $Log->debug("check parms:" . Dumper($params));
    my $xi = new NaElement("volume-transition-check");
    $Log->trace("added volume-transition-check");
    if (defined $params->{'operation-type'}) {
        $xi->child_add_string('operation-type', $params->{'operation-type'});
        $Log->trace('added operation-type');
    }
    if (defined $params->{'non-disruptive'}) {
        $Log->comment("added non-disruptive");
        $xi->child_add_string("non-disruptive", $params->{'non-disruptive'});
    }
    if (defined $params->{'override-warnings'}) {
        $Log->comment("added override-warnings");
        $xi->child_add_string("override-warnings",
                              $params->{'override-warnings'});
    }
    if (defined $params->{'vserver'}) {
        $Log->comment("added vserver");
        $xi->child_add_string('destination-vserver-name', $params->{'vserver'});
    }
    if (defined $params->{'source-node'}) {
        $Log->comment("added source-node");
        $xi->child_add_string('source-node', $params->{'source-node'});
    }

    #$xi->child_add($self->{zvol_elems});
    my $need_junction = 1;
    my $vol_elems = $self->_create_vol_elems('need-junction' => $need_junction);
    $xi->child_add($vol_elems);

    $Log->comment("added volumes");
    $Log->debug("ZapiElems to send" . Dumper($xi));
    my $xo = $self->{zapi_srv}->invoke_http($xi, "any");
    $Log->debug("ZapiElems received" . Dumper($xo));
    if (!$xo->results_ok()) {
        $Log->comment("Zapi Failure: " . $xo->results_reason());
        _zapi_error_handler($xo);
    }
    $Log->exit(_wrap_msg('FINISHED TRANSITION CHECK'));
    return 1;
} ## end sub check

=head2 prepare

C<< Volume::Transition->prepare(%params) >>

Perform zapi volume-transition-prepare using volumes defined
in the class. Store the job-id internally for use by other
methods.

=over 4

=item Input

=over 4

=item * transition-type => <scalar>

(Required)
Type of transition operation to perform (transition|untransition|conversion)

=item * non-disruptive => <scalar>

(Optional)
Set non-disruptive flag for transition operation

=item * override-warnings => <scalar>

(Optional)
Set override-warnings flag to force transtion

=item * vserver => <scalar>

(Optional)
Vserver to be used for (un)transition operation

=item * source-node => <scalar>

(Optional)
Node hosting volumes

=back

=item Output

boolean

=item Example

C<< $vt_obj->prepare('transition-type'=>'transition'); >>

=back

=cut

################################################################################
#
# Subroutine:   prepare()
#
# Objective:    Perform zapi volume-transition-prepare using volumes defined
#               in the class
#
# Details:
#
#   Input:
#       transition-type          - transition type to prepare
#       non-disruptive           - set non-disruptive flag
#       override-warnings        - set override warnings flag
#       vserver                  - vserver involved in (un)transition
#       source-node              - node hosting the volumes
#       affinity-node            - node where the job should be queued
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub prepare {
    my $self = shift();

    # Create the volumes object
    $Log->enter(_wrap_msg("BEGIN PREPARE"));
    my $params = {
        validate(
            @_,
            {
             'operation-type' => {
                               type     => SCALAR,
                               optional => 1,
                               # NOTE: We don't have a callback here to
                               # validate the argument as we want to be able
                               # to pass bogus values for testing.
                              },
             'non-disruptive' => {
                                 type     => BOOLEAN,
                                 optional => 1,
                                },
             'override-warnings' => {
                                     type     => BOOLEAN,
                                     optional => 1,
                                    },
             'vserver' => {
                           type     => SCALAR,
                           optional => 1,
                           default  => $self->{vserver},
                          },
             'source-node' => {
                            type     => SCALAR | UNDEF,
                            optional => 1,
                            default  => $self->{'source-node'},
                           },
             'affinity-node' => {
                                 type     => SCALAR,
                                 optional => 1,
                                },
            },
        ),
    };

    $Log->comment("Params:" . Dumper($params));

    my $xi = new NaElement("volume-transition-prepare");
    if (defined $params->{'operation-type'}) {
        $xi->child_add_string('operation-type', $params->{'operation-type'});
    }
    if (defined $params->{'non-disruptive'}) {
        $xi->child_add_string("non-disruptive", $params->{'non-disruptive'});
    }
    if (defined $params->{'override-warnings'}) {
        $xi->child_add_string("override-warnings",
                              $params->{'override-warnings'});
    }
    if (defined $params->{'vserver'}) {
        $xi->child_add_string('destination-vserver-name', $params->{'vserver'});
    }
    if (defined $params->{'source-node'}) {
        $Log->comment("added source-node");
        $xi->child_add_string('source-node', $params->{'source-node'});
    }
    if (defined $params->{'affinity-node'}) {
        $Log->comment("added affinity-node");
        $xi->child_add_string('affinity-node', $params->{'affinity-node'});
    }

    #$xi->child_add($self->{zvol_elems});
    my $need_junction = 1;
    my $vol_elems = $self->_create_vol_elems('need-junction' => $need_junction);
    $xi->child_add($vol_elems);

    my $xo = $self->{zapi_srv}->invoke_http($xi, "any");


    if (!$xo->results_ok()) {
        $Log->comment("Zapi Failure: " . $xo->results_reason());
        _zapi_error_handler($xo);
    }
    my $job_id = $xo->child_get_string("job-id", "");
    if ($job_id eq '') {
        $Log->warn("Could not obtain job id");
        throw Volume::Transition::BaseException("Could not obtain job id");
    }
    $self->{job_id} = $job_id;
    $self->{trans_state} = "preparing";

    $Log->debug('Workaround for burt 385773: ' .
               "sleeping for 1 second before looking up job $job_id");
    snooze(1);

    $Log->exit(_wrap_msg("FINISH PREPARE"));
    return 1;
} ## end sub prepare

=head2 commit

C<< Volume::Transition->commit(%params) >>

Perform zapi volume-transition-commit on a given job-id

=over 4

=item Input

=over 4

=item * job_id => <scalar>

(Optional)
Default = <job id stored internally from prepare>
Job id to use for commit zapi


=back

=item Output

boolean

=item Example

C<< $vt_obj->commit(); >>

=back

=cut

################################################################################
#
# Subroutine:   commit()
#
# Objective:    Perform zapi volume-transition-commit on a given job-id
#
# Details:
#
#   Input:
#       job_id         - Job id to use for commit zapi
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub commit {
    my $self = shift();
    $Log->enter(_wrap_msg("BEGIN COMMIT"));
    my $params = validate(
                          @_,
                          {
                           job_id => {
                                      type     => SCALAR,
                                      optional => 1,
                                      default  => $self->{job_id}
                                     }
                          }
                         );
    my $xi = new NaElement("volume-transition-commit");
    $xi->child_add_string("job-id", $params->{job_id});
    my $xo = $self->{zapi_srv}->invoke_http($xi, "any");

    if (!$xo->results_ok()) {
        $Log->comment("Zapi Failure: " . $xo->results_reason());
        _zapi_error_handler($xo);
    }
    $self->{trans_state} = "committing";
    $Log->exit(_wrap_msg("FINISH COMMIT"));
    return 1;
} ## end sub commit

=head2 online

C<< Volume::Transition->online(%params) >>

Perform zapi volume-transition-online on a given job-id

=over 4

=item Input

=over 4

=item * job_id => <scalar>

(Optional)
Default = <job id stored internally from prepare>
Job id to use for online zapi


=back

=item Output

boolean

=item Example

C<< $vt_obj->online(); >>

=back

=cut

################################################################################
#
# Subroutine:   online()
#
# Objective:    Perform zapi volume-transition-online on a given job-id
#
# Details:
#
#   Input:
#       job_id         - Job id to use for online zapi
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub online {
    my $self = shift();
    $Log->enter(_wrap_msg("BEGIN ONLINE"));
    my $params = validate(
                          @_,
                          {
                           job_id => {
                                      type     => SCALAR,
                                      optional => 1,
                                      default  => $self->{job_id}
                                     }
                          }
                         );
    my $xi = new NaElement("volume-transition-online");
    $xi->child_add_string("job-id", $params->{job_id});
    my $xo = $self->{zapi_srv}->invoke_http($xi, "any");

    if (!$xo->results_ok()) {
        $Log->comment("Zapi Failure: " . $xo->results_reason());
        _zapi_error_handler($xo);
    }
    $self->{trans_state} = "onlining";
    $Log->exit(_wrap_msg("FINISH ONLINE"));
    return 1;
} ## end sub online

=head2 cleanup

C<< Volume::Transition->cleanup(%params) >>

Perform zapi volume-transition-cleanup on a given job-id

=over 4

=item Input

=over 4

=item * job_id => <scalar>

(Optional)
Default = <job id stored internally from prepare>
Job id to use for cleanup zapi


=back

=item Output

boolean

=item Example

C<< $vt_obj->cleanup(); >>

=back

=cut

################################################################################
#
# Subroutine:   cleanup()
#
# Objective:    Perform zapi volume-transition-cleanup on a given job-id
#
# Details:
#
#   Input:
#       job_id         - Job id to use for cleanup zapi
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub cleanup {
    my $self = shift();
    $Log->enter(_wrap_msg("BEGIN CLEANUP"));
    my $params = validate(
                          @_,
                          {
                           job_id => {
                                      type     => SCALAR,
                                      optional => 1,
                                      default  => $self->{job_id}
                                     }
                          }
                         );
    my $xi = new NaElement("volume-transition-cleanup");
    $xi->child_add_string("job-id", $params->{job_id});
    my $xo = $self->{zapi_srv}->invoke_http($xi, "any");

    if (!$xo->results_ok()) {
        $Log->comment("Zapi Failure: " . $xo->results_reason());
        _zapi_error_handler($xo);
    }
    $self->{trans_state} = "cleaningup";
    $Log->exit(_wrap_msg("FINISH CLEANUP"));
    return 1;
} ## end sub cleanup

=head2 rollback

C<< Volume::Transition->rollback(%params) >>

Perform zapi volume-transition-rollback on a given job-id

=over 4

=item Input

=over 4

=item * job_id => <scalar>

(Optional)
Default = <job id stored internally from prepare>
Job id to use for rollback zapi


=back

=item Output

boolean

=item Example

C<< $vt_obj->rollback(); >>

=back

=cut

################################################################################
#
# Subroutine:   rollback()
#
# Objective:    Perform zapi volume-transition-rollback on a given job-id
#
# Details:
#
#   Input:
#       job_id         - Job id to use for rollback zapi
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub rollback {
    my $self = shift();
    $Log->enter(_wrap_msg("BEGIN ROLLBACK"));
    my $params = validate(
                          @_,
                          {
                           job_id => {
                                      type     => SCALAR,
                                      optional => 1,
                                      default  => $self->{job_id}
                                     }
                          }
                         );
    my $xi = new NaElement("volume-transition-rollback");
    $xi->child_add_string("job-id", $params->{job_id});
    my $xo = $self->{zapi_srv}->invoke_http($xi, "any");

    if (!$xo->results_ok()) {
        $Log->comment("Zapi Failure: " . $xo->results_reason());
        _zapi_error_handler($xo);
    }
    $self->{trans_state} = "rollingback";
    $Log->exit(_wrap_msg("FINISHED ROLLBACK"));
    return 1;
} ## end sub rollback

=head2 transition_nonstaged

C<< Volume::Transition->transition_nonstaged(%params) >>

Perform zapi volume-transition using volumes defined
in the class. Store the job-id internally for use by other
methods.

=over 4

=item Input

=over 4

=item * transition-type => <scalar>

(Required)
Type of transition operation to perform (transition|untransition|conversion)

=item * non-disruptive => <scalar>

(Optional)
Set non-disruptive flag for transition operation

=item * override-warnings => <scalar>

(Optional)
Set override-warnings flag to force transtion

=item * vserver => <scalar>

(Optional)
Vserver to be used for (un)transition operation

=item * source-node => <scalar>

(Optional)
Node hosting volumes

=back

=item Output

boolean

=item Example

C<< $vt_obj->transition_nonstaged('transition-type'=>'transition'); >>

=back

=cut

################################################################################
#
# Subroutine:   transition_nonstaged()
#
# Objective:    Perform zapi volume-transition using volumes defined
#               in the class
#
# Details:
#
#   Input:
#       transition-type          - transition type to perform
#       non-disruptive           - set non-disruptive flag
#       override-warnings        - set override warnings flag
#       vserver                  - vserver involved in (un)transition
#       source-node              - node hosting the volumes
#       affinity-node            - node where the job should be queued
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub transition_nonstaged {
    my $self = shift();

    # Create the volumes object
    $Log->enter(_wrap_msg('BEGIN TRANSITION'));
    my $params = {
        validate(
            @_,
            {
             'operation-type' => {
                               type     => SCALAR,
                               optional => 1,
                               callbacks => {
                                   'copy_based or in_place' => sub {
                                       shift() =~ /^(copy_based|in_place)$/;
                                   }
                               },
                              },
             'non-disruptive' => {
                                 type     => BOOLEAN,
                                 optional => 1,
                                },
             'override-warnings' => {
                                     type     => BOOLEAN,
                                     optional => 1,
                                    },
             'vserver' => {
                           type     => SCALAR,
                           optional => 1,
                           default  => $self->{vserver},
                          },
             'source-node' => {
                            type     => SCALAR | UNDEF,
                            optional => 1,
                            default  => $self->{'source-node'},
                           },
             'affinity-node' => {
                                 type     => SCALAR,
                                 optional => 1,
                                },
            },
        ),
    };

    $Log->comment('Params:' . Dumper($params));

    my $xi = new NaElement('volume-transition');
    if (defined $params->{'operation-type'}) {
        $xi->child_add_string('operation-type', $params->{'operation-type'});
    }
    if (defined $params->{'non-disruptive'}) {
        $xi->child_add_string('non-disruptive', $params->{'non-disruptive'});
    }
    if (defined $params->{'override-warnings'}) {
        $xi->child_add_string('override-warnings',
                              $params->{'override-warnings'});
    }
    if (defined $params->{'vserver'}) {
        $xi->child_add_string('destination-vserver-name', $params->{'vserver'});
    }
    if (defined $params->{'source-node'}) {
        $Log->comment('added source-node');
        $xi->child_add_string('source-node', $params->{'source-node'});
    }
    if (defined $params->{'affinity-node'}) {
        $Log->comment("added affinity-node");
        $xi->child_add_string('affinity-node', $params->{'affinity-node'});
    }

    my $need_junction = 1;
    my $vol_elems = $self->_create_vol_elems('need-junction' => $need_junction);
    $xi->child_add($vol_elems);

    my $xo = $self->{zapi_srv}->invoke_http($xi, 'any');

    if (!$xo->results_ok()) {
        $Log->comment('Zapi Failure: ' . $xo->results_reason());
        _zapi_error_handler($xo);
    }
    my $job_id = $xo->child_get_string('job-id', '');
    if ($job_id eq '') {
        $Log->warn('Could not obtain job id');
        throw Volume::Transition::BaseException('Could not obtain job id');
    }
    $self->{job_id} = $job_id;
    $self->{trans_state} = 'transitioning';

    $Log->debug('Workaround for burt 385773: ' .
               "sleeping for 1 second before looking up job $job_id");
    snooze(1);

    $Log->exit(_wrap_msg('FINISH TRANSITION'));
    return 1;
} ## end sub transition_nonstaged

=head2 job_get

C<< Volume::Transition->job_get(%params) >>

Return status of a given job-id

=over 4

=item Input

=over 4

=item * job_id => <scalar>

(Optional)
Default = <job id stored internally from prepare>
Job id of which to query status

=item * method => <scalar>

(Optional)
Default = zapi
Access method to use for query (cli|zapi)

=back

=item Output

returns hash of job info on success, throws exception on failure
job info hash is normalized to return the same key/value format
from both access methods.

=item Example

C<< $jobinfo = $vt_obj->job_get(); >>

=back

=cut

################################################################################
#
# Subroutine:   job_get()
#
# Objective:   Return status of a given job-id
#
# Details:
#
#   Input:
#       job_id         - Job id of which to query status
#       method         - Access method to use for query (cli|zapi)
#
#   Return value:
#       returns hash of job info on success, throws exception on failure
#       job info hash is normalized to return the same key/value format
#       from both access methods.
################################################################################

sub job_get {
    my $self = shift();
    my $params = validate(
        @_,
        {
         job_id => {
                    type     => SCALAR,
                    optional => 1,
                    default  => $self->{job_id}
                   },
         method => {
             type      => SCALAR,
             optional  => 1,
             default   => 'zapi',
             callbacks => {
                           'zapi or cli' => sub { shift() =~ /^(zapi|cli)$/ }
                          },
                   },
        },
    );
    my $jobinfo;
    if ($params->{method} eq "zapi") {
        my $xo =
          $self->{zapi_srv}->invoke("job-get", "job-id", $params->{job_id});

        if (!$xo->results_ok()) {
            $Log->comment("Zapi Failure: " . $xo->results_reason());
            _zapi_error_handler($xo);
        }

        #$Log->debug("Job result:".Data::Dumper->Dump([$xo]));
        my $xo_attrs   = $xo->child_get("attributes");
        my $xo_jobinfo = $xo_attrs->child_get("job-info");
        $Log->debug("Job_info:" . Dumper($xo_jobinfo));
        my $zjob;
        for my $child (@{$xo_jobinfo->{children}}) {
            my $key = $child->name();

            #make the params match up the same for both
            #cli/zapi
            #$key =~ s/^job-//;
            #$key =~ s/^is-restarted/restarted/;
            #$key =~ s/-//g;
            $zjob->{$key} = $child->content();
        } ## end for my $child (@{$xo_jobinfo...
        $jobinfo->{id}          = $zjob->{'job-id'};
        $jobinfo->{statuscode}  = $zjob->{'job-status-code'};
        $jobinfo->{node}        = $zjob->{'job-node'};
        $jobinfo->{progress}    = $zjob->{'job-progress'};
        $jobinfo->{completion}  = $zjob->{'job-completion'};
        $jobinfo->{description} = $zjob->{'job-description'};
        $jobinfo->{state}       = $zjob->{'job-state'};
        $jobinfo->{name}        = $zjob->{'job-name'};
        $jobinfo->{category}    = $zjob->{'job-category'};
        $jobinfo->{type}        = $zjob->{'job-type'};
    } else {
        my $jobout = $self->{server_obj}->job_show(id => $params->{job_id});
        $Log->debug("Job_info:" . Dumper($jobout));
        $jobinfo->{id}          = $jobout->[0]{id};
        $jobinfo->{statuscode}  = $jobout->[0]{code};
        $jobinfo->{node}        = $jobout->[0]{node};
        $jobinfo->{progress}    = $jobout->[0]{progress};
        $jobinfo->{progress}    =~ s/^"(.*)"$/$1/;
        $jobinfo->{completion}  = $jobout->[0]{completion};
        $jobinfo->{completion}  =~ s/^"(.*)"$/$1/;
        $jobinfo->{description} = $jobout->[0]{description};
        $jobinfo->{description} =~ s/^"(.*)"$/$1/;
        $jobinfo->{state}       = lc($jobout->[0]{state});
        $jobinfo->{name}        = $jobout->[0]{name};
        $jobinfo->{name}        =~ s/^"(.*)"$/$1/;
        $jobinfo->{category}    = $jobout->[0]{category};
        $jobinfo->{type}        = $jobout->[0]{jobtype};
    } ## end else [ if ($params->{method} ...
    return $jobinfo;
} ## end sub job_get

=head2 wait

C<< Volume::Transition->wait(%params) >>

Wait for next expected state based upon the last method called

=over 4

=item Input

=over 4

=item * job_id => <scalar>

(Optional)
Default = <job id stored internally from prepare>
Job id to wait for

=item * tries => <scalar>

(Optional)
Default = 5
Number of times to check the job before giving up

=item * sleep => <scalar>

(Optional)
Default = 15
Time to wait between checks


=back

=item Output

1 on success, Throws excception on failure

=item Example

C<< $vt_obj->wait(); >>

=back

=cut

################################################################################
#
# Subroutine:   wait()
#
# Objective:    Wait for next expected state based upon the last method called
#
# Details:
#
#   Input:
#       job_id         - Job id to wait for
#       tries          - Number of times to check the job before giving up
#       sleep          - Time to wait between checks
#
#   Return value:
#       returns 1 on success, throws exception on failure
#
################################################################################

sub wait {
    my $self = shift();
    $Log->enter(_wrap_msg("WAITING JOB STATE"));
    my $params = validate(
                          @_,
                          {
                           job_id => {
                                      type     => SCALAR,
                                      optional => 1,
                                      default  => $self->{job_id}
                                     },
                           tries => {
                                     type      => SCALAR,
                                     optional  => 1,
                                     default   => 10,
                                     callbacks => {
                                         '>= 1' => sub {
                                             !defined($_[0]) || $_[0] >= 1;
                                         },
                                     }
                                    },
                           sleep => {
                                     type     => SCALAR,
                                     optional => 1,
                                     default  => 5,
                                    },
                          }
                         );
    my %completions = (
                       preparing => {
                                     state => 'dormant',
                                     progress => 'prepared',
                                    },
                       committing => {
                                      state => 'dormant',
                                      progress => 'committed',
                                     },
                       onlining => {
                                    state => 'dormant',
                                    progress => 'online',
                                   },
                       cleaningup => {
                                      state => 'success',
                                      progress => 'done',
                                     },
                       rollingback => {
                                       state => 'success',
                                       progress => 'done',
                                      },
                       transitioning => {
                                         state => 'success',
                                         progress => 'done',
                                        },
                       cli => {
                               state => 'success',
                               progress => 'done',
                              },
        );

    my $completion = $completions{$self->{trans_state}};
    my $state_match = $completion->{state};
    my $prog_match  = $completion->{progress};
    my $tries = $params->{tries};
    my $try = 0;
    while ($try < $tries) {
        $Log->comment("####Sleeping $params->{sleep} secs and retrying####")
            if ($try);
        $try++;
        snooze $params->{sleep};
        $Log->comment("####Try # $try / $tries: Checking job state####");

        my $jobinfo = $self->job_get(job_id=>$params->{job_id},method=>"cli");
        my $job_prog = _code_string($jobinfo->{statuscode});
        my $job_state = $jobinfo->{state};

        $Log->comment(
            "### Transition state: $self->{trans_state}: waiting for state: $state_match progress: $prog_match, current state: $job_state progress: $job_prog"
            );

        if ($job_state =~ /^(?:error|fail|quit|)$/i) {
            throw Volume::Transition::BaseException(
                "Failure detected in job state: " . Dumper($jobinfo));
        }
        if ($job_prog =~ /^(?:error|recovering)$/) {
            throw Volume::Transition::BaseException(
                "Failure detected in job progress: " . Dumper($jobinfo));
        }

        if ($job_state =~ /$state_match/i && $job_prog eq $prog_match) {
            $Log->comment("####Desired state reached. Continuing..");
            $Log->exit(_wrap_msg("WAIT FINISHED"));
            return 1;
        }
        $Log->comment("####Desired state not reached####");
    } ## end while
    throw Volume::Transition::BaseException("No more retries left");
} ## end sub wait

=head2 wait_for_complete

C<< Volume::Transition->wait_for_complete(%params) >>

Wait for job to complete

=over 4

=item Input

=over 4

=item * job_id => <scalar>

(Optional)
Default = <job id stored internally from prepare>
Job id to wait for

=item * tries => <scalar>

(Optional)
Default = 5
Number of times to check the job before giving up

=item * sleep => <scalar>

(Optional)
Default = 15
Time to wait between checks


=back

=item Output

1 on success, Throws excception on failure

=item Example

C<< $vt_obj->wait_for_complete(); >>

=back

=cut

################################################################################
#
# Subroutine:   wait_for_complete()
#
# Objective:    Wait for job to complete
#
# Details:
#
#   Input:
#       job_id         - Job id to wait for
#       tries          - Number of times to check the job before giving up
#       sleep          - Time to wait between checks
#
#   Return value:
#       returns 1 on success, 0 on no job_id, throws exception on failure
#
################################################################################

sub wait_for_complete {
    my $self = shift();
    $Log->enter(_wrap_msg('WAITING JOB COMPLETION'));
    my $params = validate(
                          @_,
                          {
                           job_id => {
                                      type     => SCALAR,
                                      optional => 1,
                                      default  => $self->{job_id}
                                     },
                           tries => {
                                     type      => SCALAR,
                                     optional  => 1,
                                     default   => 10,
                                     callbacks => {
                                         '>= 1' => sub {
                                             !defined($_[0]) || $_[0] >= 1;
                                         },
                                     }
                                    },
                           sleep => {
                                     type     => SCALAR,
                                     optional => 1,
                                     default  => 5,
                                    },
                          }
                         );

    my $job_id = $params->{job_id};
    if (!defined $job_id) {
        $Log->comment('####No job_id, assuming job never started####');
        return 0;
    }

    my $tries = $params->{tries};
    my $try = 0;
    while ($try < $tries) {
        $Log->comment("####Sleeping $params->{sleep} secs and retrying####")
            if ($try);
        $try++;
        snooze $params->{sleep};
        $Log->comment("####Try # $try / $tries: Checking job state####");

        my $jobinfo = $self->job_get(job_id=>$job_id,method=>"cli");
        my $job_prog = _code_string($jobinfo->{statuscode});
        my $job_state = $jobinfo->{state};

        $Log->comment(
            "### Transition state: $self->{trans_state}: waiting for completion, current state: $job_state progress: $job_prog"
            );

        # Our choices should be limited to dormant (error),
        # running (keep waiting), and complete or error (return)

        if ($job_state eq 'dormant') {
            throw Volume::Transition::BaseException(
                'Failure detected in job state: ' . Dumper($jobinfo));
        }

        if ($job_state ne 'running') {
            $Log->comment("####Desired state reached ($job_state). Continuing..");
            $Log->exit(_wrap_msg('WAIT FINISHED'));
            return 1;
        }

        $Log->comment('####Desired state not reached####');
    } ## end while
    throw Volume::Transition::BaseException('No more retries left');
} ## end sub wait_for_complete

################################################################################
#
# Subroutine: _code_string()
#
# Objective:  Convert a volume transition progress code to a string
#
# Details:
#
#   Input:
#       $code - code to convert
#
#   Return value:
#       string corresponding to the passed code, or 'error' if code is invalid
#
################################################################################

sub _code_string {
    my $code = shift();
    my @codes = (
        'done',
        'initializing',
        'preparing',
        'prepared',
        'committing',
        'committed',
        'onlining',
        'online',
        'cleaningup',
        'rollingback',
        'completing',
        'recovering',
        );

    return ($code < 0 || $code >= scalar(@codes)) ? 'error' : $codes[$code];
}

################################################################################
#
# Subroutine: _wrap_msg()
#
# Objective:  Wrap a box around a scalar containing text
#
# Details:
#
#   Input:
#       $scalar - scalar of text to wrap
#
#   Return value:
#       scalar containing wrapped text
#
################################################################################

sub _wrap_msg {
    my $in     = shift();
    my @lines  = split(/\r?\n/, $in);
    my $border = ("#" x 60) . "\n";
    my $out;
    $out .= "\n" . $border;
    for my $line (@lines) {
        $out .= "# " . $line . "\n";
    }
    $out .= $border;
    return $out;
} ## end sub _wrap_msg

################################################################################
#
# Subroutine: _zapi_error_handler()
#
# Objective:  Check the zapi error, and throw the proper exception
#
# Details:
#
#   Input:
#       $naElem - NaElement object
#
#   Return value:
#       none, throws an exception
#
################################################################################

sub _zapi_error_handler {
    $Log->enter();
    my $naElem = shift;
    my $text = $naElem->results_reason();
    if ($text =~ /Argument 'operation-type' is invalid/) {
        throw Volume::Transition::InvalidOperationType($text);
    } elsif ($text =~ /Argument 'volumes' did not include any volumes/) {
        throw Volume::Transition::NoVolumes($text);
    } elsif ($text =~ /Argument 'junction-path' not valid/)
    {
        throw Volume::Transition::JunctionParamInvalid($text);
    } elsif ($text =~ /Job failed: Invalid State|Job is in invalid state/) {
        throw Volume::Transition::InvalidState($text);
    } elsif ($text =~ /Failed to find specified volume transition job/) {
        throw Volume::Transition::JobNotFound($text);
    } elsif ($text =~ /Too many volumes/) {
        throw Volume::Transition::TooManyVols($text);
    } elsif ($text =~ /Job in unexpected terminal state/) {
        throw Volume::Transition::JobTerminated($text);
    } elsif ($text =~ /Failed to queue .* Exists'/) {
        throw Volume::Transition::JobConflict($text);
    } else {
        throw Volume::Transition::ZapiException($text);
    }

    $Log->exit();
} ## end sub _zapi_error_handler

################################################################################
#
# Subroutine: _cli_error_handler()
#
# Objective:  Check the CLI error, and throw the proper exception
#
# Details:
#
#   Input:
#       $text - text from failed CLI command
#
#   Return value:
#       none, throws an exception
#
################################################################################

sub _cli_error_handler {
    $Log->enter();
    my $text = shift;

    if ($text =~ /Failed to queue .* Exists'/) {
        throw Volume::Transition::JobConflict($text);
    }

    throw Volume::Transition::CliException($text);
} ## end sub _cli_error_handler

## Ignore this function
sub xtransition {
    my $self = shift();
    print "Performing Volume Transition\n";
    my $res = $self->_do_transition();
    if ($res && $self->{can_transition}) {

        # Transition succeeded, verify features
        print "Transition completed succesfully. Verifying features\n";
        return $self->verify();
    } elsif (!$res && $self->{can_transition}) {

        # Transition failed unexpectedly; verify rollback
        print
          "Transition failed unexpectedly. Verifying rollback of features\n";
        $self->verify(type => "rollback");
        return 0;
    } elsif (!$res && !$self->{can_transition}) {

        # Transition failed as expected; verify rollback
        print "Transition failed as expected. Verifying rollback of features\n";
        return $self->verify(type => "rollback");
    } elsif ($res && !$self->{can_transition}) {

        # Transition unexpectedly succeeded
        print "Transition unexpectedly completed.\n";
        return 0;
    } else {
        die "We should not have gotten here";
    }
} ## end sub xtransition

#unused - ignore this
sub add_type {
    my $self = shift;
    my %params = validate(@_, {name => {type => SCALAR, optional => 0}});
    $Log->enter();

    my $lib = "Volume::Transition::$params{name}";
    print "Add $params{name} $lib\n";
    eval "require $lib" or die $@;
    push(@{$self->{types}}, $lib->new());
    $Log->exit();
} ## end sub add_type

#unused Ignore
sub verify {
    my $self = shift();
    my $res  = 1;
    print "Verifying volume\n";
    for (@{$self->{types}}) {
        $res &= $_->verify();
    }
    $self->{trans_state} = "idle";
    return $res;
} ## end sub verify

sub configure {
    my $self = shift();
    $Log->enter();
    my %params = validate(@_, {params => {type => HASHREF, optional => 1}});
    print "Configuring volume properties\n";
    for (@{$self->{types}}) {
        my $ret = $_->configure();
        $self->{can_transition} &= $ret->{can_transition};
    }
    print "Volume Can transition with this config: $self->{can_transition}\n";
    $Log->exit();
} ## end sub configure
1;
