# Hey emacs, this is basically -*- perl -*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/VserverTypes.pm#1 $
#
# Copyright (c) 2011 Network Appliance, Inc.
# All rights reserved.
#
## @summary Get info and setup access to different types of vservers
## @author samuele@netapp.com
##
## @change 2012-11-12  davidlow:  First, revert back to change #1661366.
##  Then change the way vserver info (name, id,
## 	and type) were obtained.  NACL does not covert the raw output of
##	execute_raw_command(debug smdb table...) to a proper APISet 
##	response object.  When I looked, the response object was missing
##	the parsed_output[] array.  This response object could not be 
##	used for searching.  I changed new() to get the vserver info directly
##	from the raw output of execute_raw_command(debug smdb table...).
##
## @pod here

=head1 NAME

VserverTypes

=head1 SYNOPSIS

 use VserverTypes;
 my $ci = NACL::C::Node->find();
 my $vservers = VserverTypes->new(command_interface => $ci);
 $vservers->setup_users("http");
 $vservers->setup_lifs();
 foreach my $type (@{$vservers->{types}}) {
     my $vs = $vservers->{$type};
     $Log->comment("$vs->{name} is a $type vserver.\n".
                   "Its ID is $vs->{id}.\n".
                   "It has a LIF at $vs->{address}\n".
                   "and an http user named $vs->{user}\n".
                   "whose RBAC role is $vs->{auth_role}.\n");
 }

=head1 DESCRIPTION

Get info and setup access to different types of vservers.

This class is for tests that need to access different types of vservers. It
gathers information about one vserver of each of the following types.

=over

=item admin   - The cluster admin vserver. Also known as the Cserver.

=item node    - A node vserver.

=item data - A regular vserver.

=back

For each of these vservers, the VserverTypes object has a hash of information
about the vserver. The calling script tells the object to collect different
kinds of information by calling instance methods. The information hash may
contain the following keys.

=over

=item name

The vserver name. This key always exist.

=item id

The integer vserver ID. This key always exist.

=item user

An arbitrary user in this vserver. The login application is choosen by the
calling script. This key is created by the setup_users() method.

=item auth_role

The RBAC role of this user. This key is created by the setup_users() method.

=item address

The address of an arbitrary LIF on this vserver. This key is created by the
setup_lifs() method.

=item lif_role

The LIF role name. This key is created by the setup_lifs() method.

=item lif_role_code

An integer representation of the LIF role. This key is created by the
setup_lifs() method.

=back

The setup_users() and setup_lifs() methods may modify configuration in order to
make users and LIFs ready for remote access. See their documentation for
details.

=cut

package VserverTypes;

use strict;
use Tharn qw($Log);
use Data::Dumper;
use NACL::APISet;
use NACL::C::Node;
use NACL::C::SecurityLogin;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Response::CLI::CMode;

# Keys from vifmgr/src/types/netport_role.smf
# Values from freebsd/sys/sys/socket.h
my %Lif_Role_Map = (
    "cluster"      => 0x04,
    "data"         => 0x02,
    "node-mgmt"    => 0x10,
    "intercluster" => 0x40,
    "cluster-mgmt" => 0x20,
    "loopback"     => 0x08,
);

# The LIF role to use for each vserver type when creating a LIF
my %Vs_Type_To_Lif_Role = (
    node  => "node-mgmt",
    admin => "cluster-mgmt",
    data  => "data",
);

=head1 CONSTRUCTOR

=head2 new

Create a VserverTypes object and get the names and IDs of one of each type of
vserver.

=over

=item Synopsis

 $vservers = $VserverTypes->new([command_interface => $ci,]
                                [apiset => $cmode_cli_apiset,]
                                [types => \@vserver_types]);

=item Arguments

=over

=item command_interface (optional)

A NACL::C::Node object to use as a NACL command interface. By default, the
constructor will get this by calling NACL::C::Node->find().

=item apiset (optional)

A NACL Cmode CLI APISet object to use for CLI commands. By default, the
constructor will create one from the NACL command interface.

=item types (optional)

A reference to an array of vserver types that the caller requires. The object
will only get information for one vserver of each of these types. By default,
the types are "admin", "node", and "data". The argument must be a non-empty
subset of those types.

=back

=item Returns

A reference to a new VserverTypes object.

=back

=head1 METHODS

=cut

sub new($;%) {
    my ($this, %args) = @_;

    my $class = ref($this) || $this;
    my $self = {};
    $self->{ci}  = $args{command_interface} || NACL::C::Node->find();
    $self->{cli} = $args{apiset} || NACL::APISet->new(
                                        hostobj   => $self->{ci}->hostrec(),
                                        category  => "Node",
                                        interface => "CLI",
                                        set       => "CMode");

    $self->{types} = [qw(admin node data)];
    if (ref($args{types}) eq "ARRAY") {
        $self->{types} = [];
        my %type_hash = map {$_ => 1} @{$args{types}};

        # take only the valid types with no duplicates
        if ($type_hash{node}) {
            # if we have the node type, we need the admin type
            push(@{$self->{types}}, qw(admin node));
        } elsif ($type_hash{admin}) {
            push(@{$self->{types}}, "admin");
        }

        push(@{$self->{types}}, "data") if $type_hash{data};
    }

    $self->{unlocked_users} = [];

    # Find one of each type of vserver. The node vserver must be the same as
    # the APISet node, but we still need to get its ID.

    my $vs_resp = _debug_smdb_table($self->{cli},
                     "vserver_by_name show -fields vserver, id-unsigned, type");

    my @vs_info = @{_get_vs_info($vs_resp, $self->{cli}->{session}->{CLI_separator})};
    $Log->comment("\@vs_info=" . Dumper @vs_info);
    foreach my $type (@{$self->{types}}) {
        my $vs = {};
        $self->{$type} = $vs;
	foreach my $info (@vs_info) {
		if (($type eq 'node') && ($info->{name} ne $self->{ci}->node)) {
			$Log->comment("Skip non-local node: ". $info->{name});
			next;
		}
		if ($type eq $info->{type}) {
			$vs->{name} = $info->{name};
			$vs->{id} = $info->{id};
		}
	}
        $Log->comment("\$type = $type, ". Dumper($vs));
    }

# The $vs_resp returned from _debug_smdb_table() can not be used for
# searching, because it does not contain the parsed_output[] array.
#
#    foreach my $type (@{$self->{types}}) {
#        my $vs = {};
#        $self->{$type} = $vs;
#        my %query = (type => $type);
#        $query{vserver} = $self->{ci}->node if ($type eq 'node');
#        my @vs_info = $vs_resp->search(%query);
#        if (@vs_info) {
#            $vs->{name} = $vs_info[0]->{'vserver'};
#            $vs->{id}   = $vs_info[0]->{'id-unsigned'};
#        }
#
#        $Log->debug("$type = ".Dumper($vs));
#    }

    $self->{localhost} = {
        id            => 0,
        address       => "127.0.0.1",
        lif_role      => "loopback",
        lif_role_code => $Lif_Role_Map{loopback},
    };

    bless($self, $class);
    return $self;
}

=head2 setup_users

This fills out the "user" and "auth_role" keys of the vserver hashes for the
given vserver types or all vserver types set up in the constructor. It takes
any user with the given application. If necessary, the user accounts are
unlocked and the password is set to the default_password in the hostrec of the
object's NACL command interface object. Unlocked user accounts are stored
internally, and can be relocked with the relock_user_accts() method.

=over

=item Synopsis

 $vservers->setup_users([application => $login_application,]
                        [types => \@vserver_types]);

=item Arguments

=over

=item application (optional)

A string indicating the login application that the user must have. The default
is "http".

=item types (optional)

Only setup users on the vservers of these types. The node type implies the admin
type because users are not defined for node vservers. Node vservers are accessed
by admin vserver users. The default is all the types that were setup in the
constructor.

=back

=item Returns

Nothing.

=back

=cut

sub setup_users($%) {
    my $self = shift;
    my %args = @_;

    my $application = $args{application} || "http";
    $args{types} = $self->{types} unless (ref($args{types}) eq "ARRAY");
    my %type_hash = map {$_ => 1} @{$args{types}};

    my @login_types = ();
    if (($type_hash{admin} || $type_hash{node}) && $self->{admin}) {
        push(@login_types, "admin");
    }
    if ($type_hash{data} && $self->{data}) {
        push(@login_types, "data");
    }

    my @login_vs_names = map($self->{$_}->{name}, @login_types);

    my $user_resp = $self->{cli}->security_login_show(
                                    vserver     => join("|", @login_vs_names),
                                    application => $application,
                                    authmethod  => "password");

    foreach my $type (@login_types) {
        my $vs = $self->{$type};
        my @login = $user_resp->search(vserver => $vs->{name});
        if (@login) {
            $vs->{user}      = $login[0]->{username};
            $vs->{auth_role} = $login[0]->{role};

            if ($login[0]->{acctlocked} eq "yes") {
                my %user = (
                    vserver  => $vs->{name},
                    username => $vs->{user},
                );

                # if the password was already set, ignore the exception
                try {
                    $Log->trace("Setting password for ".
                                "user $vs->{user}, vserver $vs->{name}");
                    NACL::C::SecurityLogin->password(
                         command_interface => $self->{ci},
                         %user,
                         newpwd  => $self->{ci}->hostrec()->{default_password},
                         newpwd2 => $self->{ci}->hostrec()->{default_password});
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $ex = shift;
                    if ($ex->text() !~ /password must be different/i) {
                        $ex->throw();
                    }
                    $Log->trace("Password was already set for ".
                                "user $vs->{user}, vserver $vs->{name}");
                }
                finally {
                    $Log->trace("Unlocking account for ".
                                "user $vs->{user}, vserver $vs->{name}");
                    NACL::C::SecurityLogin->unlock(
                                command_interface => $self->{ci},
                                %user);
                };

                push(@{$self->{unlocked_users}}, \%user);
            }
        }
    }

    if ($type_hash{node}) {
        $self->{node}->{$_} = $self->{admin}->{$_} foreach qw(user auth_role);
    }

    $Log->debug("$_ = ".Dumper($self->{$_})) foreach @{$self->{types}};
}

=head2 relock_user_accts

Lock user accounts that were unlocked by setup_users().

=over

=item Synopsis

 $vservers->relock_user_accts();

=item Arguments

None.

=item Returns

Nothing.

=back

=cut

sub relock_user_accts($) {
    my $self = shift;

    while (my $user = shift(@{$self->{unlocked_users}})) {
        $Log->trace("Lock account for ".
                    "user $user->{username}, vserver $user->{vserver}");
        NACL::C::SecurityLogin->lock(command_interface => $self->{ci}, %$user);
    }
}

=head2 setup_lifs

Find or create a LIF on each vserver that can be accessed by test clients, and
set the "address", "lif_role", and "lif_role_code" keys of the vserver hashes.

LIF firewall policies may be changed to "mgmt". LIFs are created where
necessary. Some LIFs may be deleted so their addresses can be used to create
new LIFs. For example, on vsims, gx_boot/deux_init creates the cluster mgmt LIF
with a 172.* address, which can only be reached on the cluster subset. One of
the data LIFs will be deleted and its address will be used to create
a usable cluster mgmt LIF.

=over

=item Synopsis

 $vservers->setup_lifs();

=item Arguments

None.

=item Returns

Nothing.

=back

=cut

sub setup_lifs($) {
    my $self = shift;

    # Find management LIF addresses for the three vservers. A management LIF
    # uses the "mgmt" firewall-policy. Only use LIFs on the command interface
    # node. Addresses begining with 172 are unusable.
    #
    # If a vserver has no management LIFs, but has a LIF with a usable address,
    # change the firewall policy to "mgmt" and use that LIF. If a vserver has no
    # LIFs with usable addresses, store it in the no_lifs hash.

    my $lif_resp = $self->{cli}->network_interface_show(
                                    "curr-node" => $self->{ci}->node);
    my %no_lifs = ();
    foreach my $type (@{$self->{types}}) {
        my $vs = $self->{$type};
        my @lifs = $lif_resp->search(
                        "vserver" => "^$vs->{name}\$",
                        "address" => "^(?!172)");
        if (@lifs) {
            my @mgmt_lifs = $lif_resp->search(
                                "__search-array-ref__" => \@lifs,
                                "firewall-policy"      => "^mgmt\$");
            my $lif;
            if (@mgmt_lifs) {
                $lif = shift(@mgmt_lifs);
                $Log->trace("Using LIF $lif->{lif} for the $type vserver");
            } else {
                $lif = shift(@lifs);
                $Log->trace("Setting firewall-policy to 'mgmt' on the ".
                            "$type vserver LIF, $lif->{lif}");
                $self->{cli}->network_interface_modify(
                        "vserver"         => $lif->{"vserver"},
                        "lif"             => $lif->{"lif"},
                        "firewall-policy" => "mgmt");
            }
            $vs->{address}       = $lif->{"address"};
            $vs->{lif_role}      = $lif->{"role"};
            $vs->{lif_role_code} = $Lif_Role_Map{$lif->{"role"}};
        } else {
            $no_lifs{$type} = $vs;
        }
    }

    # If we couldn't get management LIFs for some vservers, find usable
    # addresses on non-management LIFs. Delete those LIFs to free up their
    # addresses and create new LIFs on the vservers that need them.

    if (scalar(keys(%no_lifs))) {
        $lif_resp = $self->{cli}->network_interface_show(
                                    "curr-node" => $self->{ci}->node,
                                    "role" => "data" );
        my @lifs = $lif_resp->search(
                        "firewall-policy" => "^(?!mgmt)",
                        "address"         => "^(?!172)");
        while (my ($type, $vs) = each(%no_lifs)) {
            my $lif = shift(@lifs);
            if (!$lif) {
                $Log->error("No LIFs to delete and reuse as a $type vserver ".
                            "management LIF");
                return $TCD::FAIL;
            }

            # Get the gateway for this address. It is needed to create a route
            # for the new LIF.
            my $route_resp = $self->{cli}->network_routing_groups_route_show(
                                "vserver"       => $lif->{"vserver"},
                                "routing-group" => $lif->{"routing-group"});
            my $gateway = $route_resp->get_row(index => 0)->{"gateway"};

            $Log->trace("Deleting LIF $lif->{lif}");
            $self->{cli}->network_interface_delete(
                            "vserver" => $lif->{"vserver"},
                            "lif"     => $lif->{"lif"});

            my $new_lif_name = "$vs->{name}_mgmt";
            my $role = $Vs_Type_To_Lif_Role{$type};
            $Log->trace("Creating $type vserver LIF, $new_lif_name");
            $self->{cli}->network_interface_create(
                            "vserver"         => $vs->{"name"},
                            "lif"             => $new_lif_name,
                            "role"            => $role,
                            "home-node"       => $self->{ci}->node,
                            "home-port"       => $lif->{"home-port"},
                            "address"         => $lif->{"address"},
                            "netmask"         => $lif->{"netmask"},
                            "firewall-policy" => "mgmt");

            # Let the create command create the routing group automatically.
            # Get the routing group that was created so we can create a route.
            my $new_lif_resp = $self->{cli}->network_interface_show(
                                                "vserver" => $vs->{"name"},
                                                "lif"     => $new_lif_name,
                                                "fields"  => "routing-group");
            my $rgroup = $new_lif_resp->get_row(index => 0)->{"routing-group"};

            $Log->trace("Creating a route for the $type vserver LIF");
            $self->{cli}->network_routing_groups_route_create(
                            "vserver"       => $vs->{"name"},
                            "routing-group" => $rgroup,
                            "gateway"       => $gateway);

            $vs->{address}       = $lif->{"address"};
            $vs->{lif_role}      = $role;
            $vs->{lif_role_code} = $Lif_Role_Map{$role};
        }
    }

    $Log->debug("$_ = ".Dumper($self->{$_})) foreach @{$self->{types}};
}

################################################################################
# Subroutine:   _debug_smdb_table
# Parameters:   cli  - APISet for c-mode CLI
#               args - A string to append to "debug smdb table ".
# Returns:      An APISet response object.
# Purpose:      Run the "debug smdb table" command with an APISet. This is a
#               special CLI command that is not defined in an SMF schema and is
#               not in the cdefs.
################################################################################
sub _debug_smdb_table($$) {
    my $cli  = shift;
    my $args = shift;

    my $cmd = "debug smdb table $args";
    my $out = $cli->execute_raw_command(command => $cmd);

    $Log->comment("$cmd OUTPUT=" . Dumper $out);
    return $out;

    # The following does not convert the output from execute_raw_command()
    # to a proper APISet response object.  When I looked, the response
    # object was missing the parsed_output[] array.  This response object
    # could not be used for searching.  I am returing only the raw output
    # about and let the caller get the info from the raw output.
    #
    my $resp = NACL::APISet::Response::CLI::CMode->new(
        output         => $out,
        command        => $cmd,
        CLI_separator  => $cli->{session}->{CLI_separator},
        execution_time => 0,
        output_compat_mirror => {},
    );

    return $resp;
}

################################################################################
# Subroutine:   _get_vs_info
# Parameters:   $raw_output  - the output from execute_raw_command(debug smdb table...).
#               $separator - CLI output separator.
# Returns:      A ref to a vs info array containing vserver name, id, and type.
# Purpose:      Search the $raw_output and find the vserver name, ID, and
#               vserver type.
################################################################################
sub _get_vs_info {
        my $raw_output = shift;
        my $separator= shift;
        my @vs;
        my @lines = split(/\n/, $raw_output);
        my $i=0;
        foreach my $line (@lines) {
                next if ($line =~ /${separator}id-unsigned/i);
                my @parts = split("$separator",$line);
                $vs[$i]={
                        'name'=>$parts[0],
                        'id'=>$parts[1],
                        'type'=>$parts[2],
                        };
                $i++;
        }
        return \@vs;
}
