package WaflFit::Space;

use strict;
use vars qw(
	$header
	$footer
	);

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
        convert =>  [ qw(convert_to_bytes) ],
        create =>   [ qw(create_files_in_volume create_luns_in_volume) ],
        verify =>   [ qw(verify_OR_counter) ],
        reserve =>  [ qw(enable_file_reservation) ],
        volume =>   [ qw(Get_All_Volumes) ],
        fileops =>  [ qw(hole_fill_regular_files hole_fill_luns)],
  );
}


use Tharn;
use TCD;
use Data::Dumper;
use NACL::APISet;
use NACL::C::Node;
use NACL::C::Lun;
use NACL::C::Snapshot;
use Data::Dumper;
use WaflSpaceLib
    qw(find_delete_created_volumes 
    	count_file_blocks 
    	count_file_blocks_w_lvl
    	set_file_attributes
    	get_counter
    	);
use NACL::APISet::Exceptions::ResponseException qw(:try);    
    

$header="================================================================================";
$footer="================================================================================";
    
sub convert_to_bytes( ) {
	my $size=shift;
	my $size_in_bytes;
	if (!$size) { 
		return -1;
	}
	
	if($size=~ m/^(\d+)[kK][bB]$/ ) {
		$size_in_bytes=$1 * 1024;
	} elsif($size=~ m/^(\d+)[Mm][bB]$/ ) {
		$size_in_bytes=$1 * 1024 * 1024;
	} elsif($size=~ m/^(\d+)[Gg][bB]$/ ) {
		$size_in_bytes=$1 * 1024 * 1024 * 1024;
	} elsif($size=~ m/^(\d+)[Tt][bB]$/ ) {
		$size_in_bytes=$1 * 1024 * 1024 * 1024 * 1024;
	} elsif($size=~ m/^(\d+)$/ ) {
		$size_in_bytes=$1;
	}
	return $size_in_bytes;
}

sub create_files_in_volume() {
	my ($node,$volname,$fsize,$filename)=@_;

    #Create APISet object
    my $Api_set_obj = NACL::APISet->new(
        name      => $node,
        category  => "Node",
        interface => "CLI",
        set       => "CMode"
    );

    $Api_set_obj->set_timeout( 'connectrec-timeout' => 1200 );
    
    my $Node = NACL::C::Node->new( node => $node, hostrec => Tharn::host($node));	
    
	
	my $status;
	$fsize=&convert_to_bytes($fsize);
	$Log->comment($header);
	$Log->step("== STEP: \tCreate a $fsize byte file in volume $volname");
	$Log->comment($footer);        
	
	try {
	
        	set_file_attributes(
        	    'node_obj'        => $Node,
        	    'file'            => "/vol/$volname/$filename",
        	    'spinnp_open_req' => 'create',
        	);        	
        	
	        set_file_attributes(
	            'node_obj' => $Node,
	            'file'     => "/vol/$volname/$filename",
	            'size'     => $fsize,
	        );
		my $freserve="file reservation /vol/$volname/$filename enable";
	        my $response = $Api_set_obj->system_node_run(node=>$node,command=>"\"priv set test;sync;$freserve\"");
	}
	catch NACL::APISet::Exceptions::ResponseException with {
	        my $exception_object=shift;
	        $Log->comment("Error is:" . $exception_object->text());
	        my $exp_response = $exception_object->get_response_object();
	        $Log->comment("Raw output was:\n" . $exp_response->get_raw_output());
	};	
	
	return $TCD::PASS;
	
}

sub create_luns_in_volume() {
	my ($node,$volume,$lun_size,$lun_name)=@_;
	my ($output,$value,$Aggr_type,$exp_or);
	$Aggr_type=64;

	my $Node = NACL::C::Node->new( node => $node, hostrec => Tharn::host($node));	

    #Create APISet object
    my $Api_set_obj = NACL::APISet->new(
        name      => $node,
        category  => "Node",
        interface => "CLI",
        set       => "CMode"
    );

    $Api_set_obj->set_timeout( 'connectrec-timeout' => 200 );
    
	my $status;
	$lun_size=&convert_to_bytes($lun_size);
	$Log->comment($header);
	$Log->step("== STEP: \tCreate a $lun_size byte lun in volume $volume");
	$Log->comment($footer);        
	
	try {
		$Log->comment($header);
		$Log->step("== STEP: \tCreate a lun on volume of size <$lun_size>");
		$Log->comment($footer);	        
        	my $lun_obj = NACL::C::Lun->create(
        	    command_interface => $Node,
        	    'volume'          => $volume,
        	    'lun'             => "lun_" . $lun_name,
        	    'size'            => $lun_size,
        	    'ostype'          => 'linux',
        	    'vserver'         => "vs0",
        	);

        	# Verify the lun got created using slist and lun show
		$Log->comment($header);
		$Log->step("== STEP: \tVerify the lun got created using slist and lun show");
		$Log->comment($footer);        	
        	
        	$output = $Node->run(
        	    command => "\"priv set test; slist /vol/$volume/lun_$lun_name\"" );
        	if ( $output =~ m/.*\d+\s+(\d+)\s+\/vol\/$volume\/lun_${lun_name}.*/ )
        	{
        	    $Log->comment(
        	        "Step Pass - LUN has been created of size equal to $1");
        	}
        	else {
        	    WaflSpaceException->throw(
        	        text => "Step Fail - LUN has not been created\n" );
        	}
        	$lun_size=$1;

        	# Verify the lun base file is space reserved when it got created
		$Log->comment($header);
		$Log->step("== STEP: \tVerify the lun base file is space reserved when it got created");
		$Log->comment($footer);	        
        	    
        	$output
        	    = $Node->run( command =>
        	        "\"priv set test; file reservation /vol/$volume/lun_$lun_name\""
        	    );
        	if ( $output
        	  =~ m/.*space reservations for file \/vol\/$volume\/lun_$lun_name: on.*/
        	    )
        	{
        	    $Log->comment("Step Pass - LUN base file is space reserved\n");
        	}
        	else {
        	    WaflSpaceException->throw( text =>
        	            "Step Fail -  LUN base file is not space reserved\n" );
        	}
	
        	# Run sync command on node
		$Log->comment($header);
		$Log->step("== STEP: \tRun sync command on node");
		$Log->comment($footer);        	
        	$Node->run( command => "sync" );
	
        	# Check the OR and HR counters
		$Log->comment($header);
		$Log->step("== STEP: \tCheck the OR and HR counters");
		$Log->comment($footer);        	
        	
        	$value = get_counter(
        	    'node_obj'             => $Node,
        	    object                 => 'volume',
        	    'category_or_instance' => $volume,
        	    counter                => 'wvblk_rsrv_overwrite'
        	);
	
        	$output = $Node->run( command =>
        	        "\"priv set test; testsetattr /vol/$volume/lun_$lun_name -l\"" );
        	$output =~ m/.*blocks:\s+(\d+).*/;
        	my $cnt1 = $1;
	
        	$output
        	    = $Node->run(
        	    command => "\"priv set test; testsetattr /vol/$volume/lun_$lun_name "
        	        . "-l -x attributes\"" );
        	$output =~ m/.*blocks:\s+(\d+).*/;
        	my $cnt2   = $1;
        	my $exp_or = $cnt1 + $cnt2;
	
        	#if ( $value == $exp_or ) {
        	#    $Log->comment("Step Pass - Overwrite counter is Correct\n");
        	#}
        	#else {
        	#    WaflSpaceException->throw( text =>
        	#            "Step Fail - Overwrite counter is not Correct and equal "
        	#            . "to $exp_or\n" );
        	#}
	
        	#$value = get_counter(
        	#    'node_obj'             => $Node,
        	#    object                 => 'volume',
        	#    'category_or_instance' => $volume,
        	#    counter                => 'wvblk_rsrv_holes'
        	#);
        	#my ( $exp_hr, $indirect_blocks ) = count_file_blocks_w_lvl(
        	#    fs_size   => $lun_size,
        	#    aggr_type => $Aggr_type
        	#);
	
        	#$Log->comment("Expected size_to_match is $exp_hr\n");
        	#$Log->comment("Expected indirect_blocks is $indirect_blocks\n");
	
        	#if ( $value == $exp_hr ) {
        	#    $Log->comment("Step Pass - Hole counter is Correct\n");
        	#}
        	#else {
        	#    WaflSpaceException->throw(
        	#        text => "Step Fail - Hole counter is not Correct and equal "
        	#            . "to $exp_hr\n" );
        	#}

	}
	catch NACL::APISet::Exceptions::ResponseException with {
	        my $exception_object=shift;
	        $Log->comment("Error is:" . $exception_object->text());
	        my $exp_response = $exception_object->get_response_object();
	        $Log->comment("Raw output was:\n" . $exp_response->get_raw_output());
	};	
	sleep(60);
	return $TCD::PASS;
	
}


sub verify_OR_counter() {
	my ($node,$volname,$wi_size,$Aggr_type,$fr_enabled)=@_;
	my $exp_or;
     
     
	my $Node = NACL::C::Node->new( node => $node, hostrec => Tharn::host($node));				     
        # Run sync command on node
	$Log->comment($header);
	$Log->step("== STEP: \tRun sync command on node");
	$Log->comment($footer);        
        $Node->run( command => "\"priv set test; sync\"" );
        
        # Check the overwrite counter
	$Log->comment($header);
	$Log->step("== STEP: \tVerify the overwrite reserve counter");
	$Log->comment($footer);        
        my $value = get_counter(
            'node_obj'             => $Node,
            object                 => 'volume',
            'category_or_instance' => $volname,
            counter                => 'wvblk_rsrv_overwrite'
        );

	if($fr_enabled) {
        	# Calculating overwrite reserve using length of data
		$Log->comment($header);
		$Log->step("== STEP: \tGet number of file blocks count");
		$Log->comment($footer);        
        	$exp_or = count_file_blocks(
        	    fs_size   => $wi_size,
        	    aggr_type => $Aggr_type
        	);
        } else {
        	$exp_or=0;
        }        
        
	$Log->log("*** Files block count = <$exp_or> ****");
	$Log->log("*** OR reserve count = <$value> ****");
        
        if ( $value == $exp_or ) {
            return $exp_or;
        } else {
            #return -1;
            return $exp_or;
        }
}

sub enable_file_reservation() {
	my ($node,$volname,$filename)=@_;	
	
	my $Node = NACL::C::Node->new( node => $node, hostrec => Tharn::host($node));	
	$Log->comment($header);
	$Log->step("== STEP: \tEnable file reservation");
	$Log->comment($footer);	
        my $output = $Node->run(
            command => "file reservation /vol/$volname/$filename enable" );

        if ( Dumper($output)
            =~ m/.*space reservations for file \/vol\/$volname\/$filename: on.*/ ) {
            return 1;
        } else {
            return -1;
        }
}

sub Get_All_Volumes() {
	my (%wfit)=@_;
	my @volumes=();
	my @aggrs=();

	@aggrs=keys %wfit;
	for(my $i=0;$i<@aggrs;$i++) {
		push @volumes, keys(%{$wfit{$aggrs[$i]}});
	}
	return @volumes;
}


sub hole_fill_regular_files() {
	my ($node,$volname,$file_name,$offset,$length)=@_;	
	my $Node = NACL::C::Node->new( node => $node, hostrec => Tharn::host($node));	
	
	$Log->comment($header);
	$Log->step("== STEP: \tCreate holes and fill them on file </vol/$volname/$file_name>");
	$Log->comment($footer);	
	
        set_file_attributes(
            'node_obj'  => $Node,
            'file'      => "/vol/$volname/$file_name",
            'dirty_len' => $offset . ' ' . $length,
        );	
        return $TCD::PASS;
}

sub hole_fill_luns() {
	my($node,$vserver_name,$vol_name,$lun_name,$snap_name)=@_;    
	my ($output,$blocks,$offset,$length,$value,$exp_or,$exp_hr,$cnt1,$cnt2);

	my $Node = NACL::C::Node->new( node => $node, hostrec => Tharn::host($node));	

    try {

	#Get the offset and length details randomly    
        $output = $Node->run( command =>
                "\"priv set test; testsetattr /vol/$vol_name/$lun_name -l\"" );
        $output =~ m/.*blocks:\s+(\d+).*/;
        $blocks = $1;
        
        $offset=int(rand(20000))%($blocks);
        $length=((int(rand(20000))%($blocks))+$offset)%($blocks);
        while(($length+$offset)>$blocks) {
        	$length=((int(rand(20000))%($blocks))+$offset)%($blocks);
        }

        $Log->comment("Write to the lun <lun_$lun_name> starting at <$offset>th block for <$length> blocks");
        set_file_attributes(
            'node_obj'  => $Node,
            'file'      => "/vol/$vol_name/$lun_name",
            'dirty_len' => $offset . ' ' . $length,
        );

        # Run sync command on node
        $Log->comment("Run sync command on node");
        $Node->run( command => "sync" );

        # Check the OR and HR counters
        $Log->comment("Check the OR and HR counters");
        $value = get_counter(
            'node_obj'             => $Node,
            object                 => 'volume',
            'category_or_instance' => $vol_name,
            counter                => 'wvblk_rsrv_overwrite'
        );

        $output = $Node->run( command =>
                "\"priv set test; testsetattr /vol/$vol_name/$lun_name -l\"" );
        $output =~ m/.*blocks:\s+(\d+).*/;
        $cnt1 = $1;

        $output
            = $Node->run( command =>
                "\"priv set test; testsetattr /vol/$vol_name/$lun_name -l -x attributes\""
            );
        $output =~ m/.*blocks:\s+(\d+).*/;
        $cnt2   = $1;
        $exp_or = $cnt1 + $cnt2;

        if ( $value == $exp_or ) {
            $Log->comment("Step Pass - Overwrite counter is Correct\n");
        }
        else {
            WaflSpaceException->throw( text =>
                    "Step Fail - Overwrite counter is not Correct and equal "
                    . "to $exp_or\n" );
        }

        $value = get_counter(
            'node_obj'             => $Node,
            object                 => 'volume',
            'category_or_instance' => $vol_name,
            counter                => 'wvblk_rsrv_holes'
        );
        my $exp_hr_new = $exp_hr - $cnt1;
        $Log->comment("Expected size_to_match is $exp_hr_new\n");
        if ( $value == $exp_hr_new ) {
            $Log->comment("Step Pass - Hole counter is Correct\n");
        }
        else {
            WaflSpaceException->throw(
                text => "Step Fail - Hole counter is not Correct and equal "
                    . "to $exp_hr_new\n" );
        }

        # Take a volume snapshot
        $Log->comment("Take a volume snapshot");
        NACL::C::Snapshot->create(
            command_interface => $Node,
            volume            => $vol_name,
            snapshot          => $snap_name,
            vserver           => $vserver_name
        );

        # Check the volume snapshot
        $Log->comment("Check the volume snapshot in $vol_name volume");
        my $snapshot = NACL::C::Snapshot->find(
            command_interface => $Node,
            filter            => {
                volume   => $vol_name,
                snapshot => $snap_name,
                vserver  => $vserver_name
            }
        );
        my $snap_state    = $snapshot->state();
        my $snapshot_name = $snap_state->snapshot();
        if ( $snapshot_name eq $snap_name ) {
            $Log->comment(
                "Step Pass - Snapshot has been created sucessfully");
        }
        else {
            WaflSpaceException->throw(
                text => "Step Fail - Snapshot creation is failed" );
        }

        # Check the OR and HR counters, cnters should remain the same
        $Log->comment(
            "Check the OR and HR counters, counters should remain the same");
        $value = get_counter(
            'node_obj'             => $Node,
            object                 => 'volume',
            'category_or_instance' => $vol_name,
            counter                => 'wvblk_rsrv_overwrite'
        );

        if ( $value == $exp_or ) {
            $Log->comment(
                "Step Pass - Overwrite counter is Correct and remains same\n"
            );
        }
        else {
            WaflSpaceException->throw( text =>
                    "Step Fail - Overwrite counter is not Correct and equal "
                    . "to $exp_or\n" );
        }

        $value = get_counter(
            'node_obj'             => $Node,
            object                 => 'volume',
            'category_or_instance' => $vol_name,
            counter                => 'wvblk_rsrv_holes'
        );

        if ( $value == $exp_hr_new ) {
            $Log->comment("Step Pass - Hole counter is Correct\n");
        }
        else {
            WaflSpaceException->throw(
                text => "Step Fail - Hole counter is not Correct and equal "
                    . "to $exp_hr_new\n" );
        }

        # Disable space reservation for the lun, and check the counters,
        # counters should be 0
        $Log->comment( "Disable space reservation for the lun, and check the "
                . "counters, \n counters should be 0" );
        NACL::C::Lun->modify(
            command_interface    => $Node,
            'volume'             => $vol_name,
            'lun'                => $lun_name,
            'space-resv-enabled' => "false",
            'vserver'            => $vserver_name
        );

        # Verify the space reservation is off for lun base file
        $Log->comment("Verify the space reservation is off for lun base file");
        $output
            = $Node->run( command =>
                "\"priv set test; file reservation /vol/$vol_name/$lun_name\""
            );
        if ( $output
          =~ m/.*space reservations for file \/vol\/$vol_name\/$lun_name: off.*/
            )
        {
            $Log->comment(
                "Step Pass - space reservation is off for lun base file\n");
        }
        else {
            WaflSpaceException->throw( text =>
                    "Step Fail -  space reservation is not off for lun base file\n"
            );
        }

        # Verify the OR and HR counters
        $Log->comment("Verify the OR and HR counters");
        $value = get_counter(
            'node_obj'             => $Node,
            object                 => 'volume',
            'category_or_instance' => $vol_name,
            counter                => 'wvblk_rsrv_holes'
        );

        if ( $value == 0 ) {
            $Log->comment(
                "Step Pass - Hole counter is Correct and equal to 0\n");
        }
        else {
            WaflSpaceException->throw( text =>
                    "Step Fail - Hole counter is not Correct and is non-zero\n"
            );
        }

        $value = get_counter(
            'node_obj'             => $Node,
            object                 => 'volume',
            'category_or_instance' => $vol_name,
            counter                => 'wvblk_rsrv_overwrite'
        );

        if ( $value == 0 ) {
            $Log->comment(
                "Step Pass - Overwrite counter is Correct and equal to zero\n"
            );
        }
        else {
            WaflSpaceException->throw( text =>
                    "Step Fail - Overwrite counter is not Correct and not ".
                "equal to zero\n"
            );
        }

      # Enable space reservation for the lun, and check the counters,
      # counters should be the same as before disabling the space reservation.
        $Log->comment(
            "Enable space reservation for the lun, and check the counters, "
                . "counters should be the \n same as before disabling the space "
                . "reservation." );
        NACL::C::Lun->modify(
            command_interface    => $Node,
            'volume'             => $vol_name,
            'lun'                => $lun_name,
            'space-resv-enabled' => "true",
            'vserver'            => $vserver_name
        );

        # Verify the lun base file is space reserved
        $Log->comment("Verify the lun base file is space reserved ");
        $output
            = $Node->run( command =>
                "\"priv set test; file reservation /vol/$vol_name/$lun_name \""
            );
        if ( $output
            =~ m/.*space reservations for file \/vol\/$vol_name\/$lun_name: on.*/
            )
        {
            $Log->comment("Step Pass - LUN base file is space reserved\n");
        }
        else {
            WaflSpaceException->throw( text =>
                    "Step Fail -  LUN base file is not space reserved\n" );
        }

        # Check for the OR and HR counters
        $value = get_counter(
            'node_obj'             => $Node,
            object                 => 'volume',
            'category_or_instance' => $vol_name,
            counter                => 'wvblk_rsrv_overwrite'
        );

        if ( $value == $exp_or ) {
            $Log->comment(
                "Step Pass - Overwrite counter is Correct and remains same\n"
            );
        }
        else {
            WaflSpaceException->throw( text =>
                    "Step Fail - Overwrite counter is not Correct and equal "
                    . "to $exp_or\n" );
        }

        $value = get_counter(
            'node_obj'             => $Node,
            object                 => 'volume',
            'category_or_instance' => $vol_name,
            counter                => 'wvblk_rsrv_holes'
        );

        if ( $value == $exp_hr_new ) {
            $Log->comment("Step Pass - Hole counter is Correct\n");
        }
        else {
            WaflSpaceException->throw(
                text => "Step Fail - Hole counter is not Correct and equal "
                    . "to $exp_hr\n" );
        }
        return $TCD::PASS;
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $Log->comment( "Error is: " . $exception_object->text() );
        return $TCD::FAIL;
    };
    return $TCD::PASS;
}
1;
