# $Id $

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking ProhibitFixedStringMatches)

## @summary MCC Aggregate Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::Aggregate;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use NACL::C::Aggregate ();

use NATE::BaseException qw(:try);
use Tharn qw($Log);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);

use Readonly;
use Data::Dumper;


=head1 NAME

WaflMCC::MCC::Aggregate

=head1 DESCRIPTION

This class is to be used for C-Mode Metrocluster testing.

    use WaflMCC::MCC::Aggregate;
    ...
    my $mcc_aggr = WaflMCC::MCC::Aggregate->new(...);

=cut

=head1 Public Methods

=head2 new()

    my $mcc_aggr = WaflMCC::MCC::Aggregate->new(
                          owner_mcc_node => $owner_mcc_node,
                          aggregate_name => $existing_aggr_name);

WaflMCC::MCC::Aggregate objects are mostly bookkeeping objects with a few
useful methods which are used mostly internally in MCC.  It also gives access
to the associated NACL::C::Aggregate object if needed.

=cut

# Automatically generate the (non-public) constructor (includes a call to
# 'init') and also creates non-public members and non-public accesors for
# them.
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
        { -type => 'NACL::C::Aggregate' },
        '_Aggregate_component',
              ],
    scalar => [
        qw/ _name
            _uuid
            _is_root /,
              ],
                       ];

# validate the parameters passed in and what else?
sub init {
    $Log->enter();
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            owner_mcc_node => { isa   => 'WaflMCC::MCC::Node' },
            aggregate_name => { regex => qr/^\w+[\w\-]+$/ },
        },
                            );

    # set members per the passed constructor arguments.
    $self->_Aggregate_component(NACL::C::Aggregate->new(
          command_interface =>
              $opts{owner_mcc_node}->get_Node_component()->command_interface(),
          aggregate         => $opts{aggregate_name}));

    my $state = $self->_Aggregate_component()->
                      state(requested_fields => ['aggregate', 'uuid', 'root']);
    $self->_name($state->aggregate());
    $self->_uuid($state->uuid());
    $self->_is_root($state->root() =~ /true/i ? 1 : 0);

    $Log->exit();
    return;
}

=head2 get_Aggregate_component()

    my $nacl_aggr_component = $mcc_aggr->get_Aggregate_component();

This utility method returns the NACL::C::Aggregate object associated with
this MCC::Aggregate.

=cut

sub get_Aggregate_component {
    $Log->enter();
    my $self = shift;

    $Log->exit();
    return $self->_Aggregate_component();
}

=head2 get_name()

    my $aggr_name = $mcc_aggr->get_name();

This utility method returns the name of this aggregate.

=cut

sub get_name {
    $Log->enter();
    my $self = shift;

    $Log->exit();
    return $self->_name();
}

=head2 get_uuid()

    my $aggr_uuid = $mcc_aggr->get_uuid();

This utility method returns the uuid of this aggregate.

=cut

sub get_uuid {
    $Log->enter();
    my $self = shift;

    $Log->exit();
    return $self->_uuid();
}

=head2 is_root()

    my $is_a_root_aggregate = $mcc_aggr->is_root();

This utility method returns a boolean (0|1) flagging this aggregate as a root
aggregate or not.

=cut

sub is_root {
    $Log->enter();
    my $self = shift;

    $Log->exit();
    return $self->_is_root();
}

#=head2 L_get_root_aggr_disk_serials()
#
#This internal method can only be called on a node which is in a good state.  It
#queries the node for the current set of root aggregate disks' serial numbers.
#
#This method returns an arrayref containing a list of disk serial numbers.
#
#=cut
#
#sub L_get_root_aggr_disk_serials {
#    $Log->enter();
#    my $self = shift;
#
#    my @root_disk_names = ();
#    my $disk_arrayref =
#                 $self->_root_aggregate()->get_one_state_attribute('disklist');
#
#    # TBD: reported as NACL bug.  the attribute returned is an arrayref
#    # sure enough, but its only member is a single csv list string
#    # instead of an an array of names...  :p
#    if (@{ $disk_arrayref } == 1) {
#        # this works now, but will stop working after NACL bug fix (?)
#        my $root_disks_csv_list = $disk_arrayref->[0];
#        @root_disk_names = split(/\s*,\s*/, $root_disks_csv_list);
#    } else {
#        # this should work after NACL bug fix (?)
#        @root_disk_names = @{ $disk_arrayref };
#    }
#
#    my @serials = ();
#    my $ci = $self->get_Node_component()->command_interface();
#    foreach my $disk_name (@root_disk_names) {
#        push( @serials, NACL::CS::StorageDisk->fetch(
#                                command_interface => $ci,
#                                filter            => { 'disk' => $disk_name, },
#                                requested_fields  => ['serial-number'],
#                                                    )->serial_number() );
#    }
#
#    $Log->exit();
#    return \@serials;
#}

1;
