# $Id: //depot/prod/test/main/lib/NACL/MTask/MCC/Cluster.pm#17 $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Cluster Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::Cluster;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::DR_Coordinator ();

use NACL::C::Metrocluster ();
use NACL::C::Cluster ();
use NACL::C::ClusterIdentity ();
use NACL::C::ClusterPeer ();
use NACL::C::NetworkInterface ();
use NACL::C::Node();

use NATE::BaseException qw(:try);
use NACL::Exceptions::NoElementsFound qw(:try);

use NACL::Transit ();

use Tharn qw($Log subtest);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);
use List::MoreUtils qw(any);
use Data::Dumper;
use Subtest;


=head1 NAME

WaflMCC::MCC::Cluster

=head1 DESCRIPTION

This class is to be used for C-Mode Metrocluster testing.  It has only the
functionality required for MCC testing and should not be confused with
more general implementations of ONTAP cluster functionality such as
NACL::C::Cluster and NACL::STask::Cluster.  Think of it as a filter which in
some sense helps define the functionality needed to test clusters in the MCC
realm.

    # Start test
    use WaflMCC::MCC::Metrocluster;
    ...
    my $cluster_A = $mcc->get_cluster(site => 'A');
    ...
    # Perform test here
    ...
    # End test

=cut

=head1 Public Methods

There is no public constructor for this class.  A WaflMCC::MCC::Cluster
object is obtained via calls such as

    my $cluster_A = $mcc->get_cluster(site => 'A');

=cut

# Automatically generate the (non-public) constructor (includes a call to
# 'init') and also creates non-public members and non-public accesors for
# them.
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
        { -type => 'WaflMCC::MCC::Metrocluster' },
        '_metrocluster',
              ],
    scalar => [
        { -type => 'WaflMCC::MCC::DR_Coordinator' },
        '_DR_coordinator',
              ],
    scalar => [
        qw/ _site
            _name
            _pre_switchover_halted_node_transit_objs
            _pre_switchover_intercluster_network_interfaces_downed
            _pre_switchover_disabled_isl_ports
          /,
              ],
                       ];

# validate the parameters passed in and what else?
sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            metrocluster => { isa   => 'WaflMCC::MCC::Metrocluster' },
            site         => { regex => qr/^(A|B)$/ },
        },
                            );

    # set members per the passed constructor arguments.
    $self->_metrocluster($opts{'metrocluster'});
    $self->_site(        $opts{'site'});

    return;
}

=head2 get_Metrocluster_component()

    my $site_specific_mcc_component = $cluster_X->get_Metrocluster_component();

Returns the NACL::C::Metrocluster object associated with the execution of
metrocluster commands at $cluster_X's site.

=cut

sub get_Metrocluster_component {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_metrocluster()->get_Metrocluster_component(
                                                       site => $self->_site());
}

=head2 get_Cluster_component()

    my $cluster_component = $cluster_X->get_Cluster_component();

Returns the NACL::C::Cluster object associated with this cluster.

=cut

sub get_Cluster_component {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $cluster_component =
            $self->_metrocluster()->get_Metrocluster_component(
                                  site => $self->_site())->command_interface();

    # this method (and a lot else) expects that the Metrocluster components'
    # command interfaces are NACL::C::Cluster objects.
    unless ($cluster_component->isa('NACL::C::Cluster')) {
        confess('Metrocluster CI found to not be a NACL::C::Cluster!!');
    }

    return $cluster_component;
}

=head2 get_site()

    my $cluster_site = $cluster->get_site();

Returns this cluster's site ('A'|'B').

=cut

sub get_site {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_site();
}

=head2 get_name()

    my $cluster_name = $cluster_X->get_name();

Returns the name of this cluster.

=cut

# WARNING: Currently this method relies on this cluster's name being the first
# entry in a list of cluster names returned by
#     metrocluster show -fields cluster!
#
# TBD: use _get_ClusterIdentity_state below to get cluster name more reliably
sub get_name {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $name =
        $self->get_Metrocluster_component()->get_one_state_attribute('cluster');

    return $name;
}

# expects a single argument remote_cluster_name
sub _get_ClusterPeer_component {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = @_;

    my $remote_cluster_name = $opts{'remote_cluster_name'};

    my $mcc_component = $self->get_Metrocluster_component();

    my $ClusterPeer_component = NACL::C::ClusterPeer->find(
                 command_interface => $mcc_component->command_interface(),
                 filter            => { 'rcluster' => $remote_cluster_name },
             );

    return $ClusterPeer_component;
}

=head2 get_peer()

    my $cluster_B = $cluster_A->get_peer();

Returns the other cluster (WaflMCC::MCC::Cluster) in the MCC relationship.

=cut

sub get_peer {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $peer_site = ($self->_site() =~ /A/) ? 'B' : 'A';

    return $self->_metrocluster()->get_cluster(site => $peer_site);
}

=head2 get_DR_Coordinator()

Each cluster in the MCC relationship has a DR Coordinator.

    my $dr_coordinator = $cluster_B->get_DR_Coordinator();

Returns the WaflMCC::MCC::DR_Coordinator associated with this cluster.

=cut

sub get_DR_Coordinator {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    if (! $self->_DR_coordinator_isset()) {
        my $mcc_component = $self->get_Metrocluster_component();
        $self->_DR_coordinator(WaflMCC::MCC::DR_Coordinator->new(
                                        metrocluster => $self->_metrocluster(),
                                        site         => $self->_site()));
    }

    return $self->_DR_coordinator();
}

=head2 prepare_for_switchover_execution()

    $cluster_A->prepare_for_switchover_execution(
                # default values shown here
            do_intercluster_network_interface_disable   => 1,
            do_isl_link_down                            => 0,
            do_peer_metrocluster_site_nodes_halt        => 1,
            do_wait_minutes_for_cluster_appears_down    => 3,);

This method performs various actions often/usually performed before this site's
cluster executes a DR switchover command.

=over

=item Options

=item C<< do_intercluster_network_interface_disable => 0|1 (default = 0) >>

(Optional) This option will disable the intercluster network interface on this
cluster which allows this cluster to see its Metrocluster peer cluster.

=item C<< do_isl_link_down => 0|1 (default = 0) >>

(Optional) set to disable ISL ports of switch to bring down ISL link

If set to 1 do_isl_link_down => 1 all FC ports will be brought down and will
be brought back up by prepare_for_switchback_execution

=item C<< do_peer_metrocluster_site_nodes_halt => 0|1 (default = 1) >>

(Optional) This option will attempt to halt all the nodes in this cluster's
Metrocluster peer cluster.  (I.e., if this called on cluster A, it wiil try
to halt all nodes in cluster B.)

=item C<< do_wait_minutes_for_cluster_appears_down => $minutes (default = 3) >>

(Optional) wait for specified number of minutes

=back

=cut


# mica - if we do external state changes to the nodes in script, set using below method for this class.
sub set_transit_objects {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;
    my %raw_opts = @_;
        
    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => { trans_objs => { type => ARRAYREF } },
    );

    my @trans_objs = delete $opts{trans_objs};
    
    # save this information for rebooting halted nodes
    $self->_pre_switchover_halted_node_transit_objs(
                                                \@trans_objs);
}

sub prepare_for_switchover_execution {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %raw_opts = @_;

    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            do_intercluster_network_interface_disable =>
                                  { regex => qr/^[01]$/,       default => 0, },
            do_isl_link_down                          =>
                                  { regex => qr/^[01]$/,       default => 0, },
            do_peer_metrocluster_site_nodes_halt      =>
                                  { regex => qr/^[01234]$/,       default => 1, },
            do_wait_minutes_for_cluster_appears_down  =>
                                  { regex => qr/^\d+[.]?\d*$/, default => 5, },
                  },
                            );

    if ($opts{do_intercluster_network_interface_disable}) {
        $Log->comment('Disabling intercluster interfaces to disaster site...');
        my @interfaces_to_disable =
             $self->_find_DR_cluster_to_THIS_cluster_intercluster_interfaces();

        foreach my $network_interface (@interfaces_to_disable) {
            $network_interface->modify('status-admin' => 'down');
        }

        # keep track of these for re-enable after switchback
        $self->_pre_switchover_intercluster_network_interfaces_downed(
                                                      \@interfaces_to_disable);
    } else {
        $self->_pre_switchover_intercluster_network_interfaces_downed(undef);
    }

    if ($opts{do_isl_link_down}) {
        my @isl_ports = ( );
        my $dr_groups = $self->_metrocluster()->get_DR_groups();

        foreach my $dr_group (values(% {$dr_groups} )) {
            foreach my $fabric (
            @{ $dr_group->get_network_topology()->get_intersite_fabrics() }) {
                push(@isl_ports,
                  @{ $fabric->get_switch(
                    site => $self->get_peer()->_site())->get_isl_ports() });
            }
        }

        $self->_pre_switchover_disabled_isl_ports( \@isl_ports);

        foreach my $isl_port (@isl_ports) {
            $isl_port->disable();
        }
    }

    # mica 
    # wafl-qa proposed changes in library code 
    # 0 - halt inhibit-takeover , used currently by dpg
    # 1 - power-cycle
    # 2 - panic
    # 3 - reboot inhibit-takeover
    # 4 - halt 
    # 5 - reboot
    # 6 - Do nothing (for planned switchover)

    my (@subtest_handles,@halted_node_objects,@peer_cluster_mcc_nodes); 
    if ($opts{do_peer_metrocluster_site_nodes_halt} == 0) {
    # 0 - halt inhibit-takeover.
        $Log->comment('Halting MCC associated nodes at disaster site...');
        @subtest_handles = ();
        @halted_node_objects = ();
        @peer_cluster_mcc_nodes =
                             @{ $self->get_peer()->L_get_cluster_mcc_nodes() };

        foreach my $peer_mcc_node (@peer_cluster_mcc_nodes) {
            my $node_name = $peer_mcc_node->get_name();
            my $halt_cmd  = 'set -confirmations off;'.
                            'system node halt -inhibit-takeover true -reason '.
                            'prepare_for_switchover_execution -node '.
                            $node_name;
            my $trans_obj = NACL::Transit->new( name => $node_name );
            push @subtest_handles, subtest(sub {
                $trans_obj->change_state( to      => 'FIRMWARE',
                                          haltcmd => $halt_cmd,
                                          timeout => 2000 );
                },
                '-bg',
                '-runid',
                $node_name . '_prepare_for_switchover_execution_down');
            push(@halted_node_objects, $trans_obj);
        }
        Subtest::wait_finish(subtests => \@subtest_handles);

        # save this information for rebooting halted nodes
        $self->_pre_switchover_halted_node_transit_objs(
                                                \@halted_node_objects);
    } elsif ($opts{do_peer_metrocluster_site_nodes_halt} == 1) {
    # 1 - power-cycle all peer nodes.
        #$Log->comment('Powercycle MCC associated nodes at disaster site...');
        #@subtest_handles = ();
        #@halted_node_objects = ();
        #@peer_cluster_mcc_nodes =
        #                     @{ $self->get_peer()->L_get_cluster_mcc_nodes() };
        #
        #foreach my $peer_mcc_node (@peer_cluster_mcc_nodes) {
        #    my $node_name = $peer_mcc_node->get_name();
        #    my $halt_cmd  = 'set -confirmations off;'.
        #                    'system node halt -inhibit-takeover true -reason '.
        #                    'prepare_for_switchover_execution -node '.
        #                    $node_name;
        #    my $trans_obj = NACL::Transit->new( name => $node_name );
        #    push @subtest_handles, subtest(sub {
        #        $trans_obj->change_state( to      => 'SP',
        #                                  powercycle => 1,
        #                                  timeout => 2000 );
        #        },
        #        '-bg',
        #        '-runid',
        #        $node_name . '_prepare_for_switchover_execution_down');
        #
        #    my $mccHost = host($node_name);
        #    my $mcc_Api_Set_Obj = NACL::APISet->new(
        #        hostobj   => $mccHost,
        #        category  => "Node",
        #        interface => "CLI",
        #        connid    => "sp_login",
        #        set       => "SP",
        #    );
        #
        #    try {
        #        my @prompts_answers =
        #            ( '.*This will cause a dirty shutdown of your appliance.*' => 'n', );
        #        my $response_create = $mcc_Api_Set_Obj->system_power( cycle => 1,
        #            'connectrec-match_table' => \@prompts_answers, );
        #        $Log->comment( "Response get_parsed_output:\n"
        #            . Dumper( $response_create->get_parsed_output() ) );
        #    } catch NACL::APISet::Exceptions::ResponseException with {
        #        my $exception_object = shift;
        #        $Log->comment( "Error is: " . $exception_object->text() );
        #        my $resp = $exception_object->get_response_object();
        #        $Log->comment( "Raw output was:\n" . $resp->get_raw_output() );
        #    };
        #    
        #    #push(@halted_node_objects, $trans_obj);
        #}
        ##Subtest::wait_finish(subtests => \@subtest_handles);
        #
        ## save this information for rebooting halted nodes
        #$self->_pre_switchover_halted_node_transit_objs(
        #                                        \@halted_node_objects);
 
    } elsif ($opts{do_peer_metrocluster_site_nodes_halt} == 2) {
        # 2 - panic
        $Log->comment('Panic the MCC associated nodes at disaster site...');
        @subtest_handles = ();
        @halted_node_objects = ();
        @peer_cluster_mcc_nodes =
                             @{ $self->get_peer()->L_get_cluster_mcc_nodes() };

        foreach my $peer_mcc_node (@peer_cluster_mcc_nodes) {
            my $node_name = $peer_mcc_node->get_name();
            my $halt_cmd  = 'set -confirmations off;'.
                            'system node halt -dump true -skip-lif-migration true -reason '.
                            'prepare_for_switchover_execution -node '.
                            $node_name;
            my $trans_obj = NACL::Transit->new( name => $node_name );
            push @subtest_handles, subtest(sub {
                $trans_obj->change_state( to      => 'FIRMWARE',
                                          haltcmd => $halt_cmd,
                                          timeout => 2000 );
                },
                '-bg',
                '-runid',
                $node_name . '_prepare_for_switchover_execution_down');
            push(@halted_node_objects, $trans_obj);
        }
        Subtest::wait_finish(subtests => \@subtest_handles);

        # save this information for rebooting halted nodes
        $self->_pre_switchover_halted_node_transit_objs(
                                                \@halted_node_objects);
    } elsif ($opts{do_peer_metrocluster_site_nodes_halt} == 3) {
        # 3 - reboot inhibit-takeover
        $Log->comment('Reboot MCC associated nodes at disaster site...');
        @subtest_handles = ();
        @halted_node_objects = ();
        @peer_cluster_mcc_nodes =
                             @{ $self->get_peer()->L_get_cluster_mcc_nodes() };

        foreach my $peer_mcc_node (@peer_cluster_mcc_nodes) {
            my $node_name = $peer_mcc_node->get_name();
            my $halt_cmd  = 'set -confirmations off;'.
                            'system node reboot -inhibit-takeover true -reason '.
                            'prepare_for_switchover_execution -node '.
                            $node_name;
            my $trans_obj = NACL::Transit->new( name => $node_name );
            push @subtest_handles, subtest(sub {
                $trans_obj->change_state( to      => 'FIRMWARE',
                                          haltcmd => $halt_cmd,
                                          timeout => 2000 );
                },
                '-bg',
                '-runid',
                $node_name . '_prepare_for_switchover_execution_down');
            push(@halted_node_objects, $trans_obj);
        }
        Subtest::wait_finish(subtests => \@subtest_handles);

        # save this information for rebooting halted nodes
        $self->_pre_switchover_halted_node_transit_objs(
                                                \@halted_node_objects);
    } elsif ($opts{do_peer_metrocluster_site_nodes_halt} == 4) {
        # 4 - halt 
        $Log->comment('Halting without inhibit takeover of MCC associated nodes at disaster site...');
        @subtest_handles = ();
        @halted_node_objects = ();
        @peer_cluster_mcc_nodes =
                             @{ $self->get_peer()->L_get_cluster_mcc_nodes() };

        foreach my $peer_mcc_node (@peer_cluster_mcc_nodes) {
            my $node_name = $peer_mcc_node->get_name();
            my $halt_cmd  = 'set -confirmations off;'.
                            'system node halt -reason '.
                            'prepare_for_switchover_execution -node '.
                            $node_name;
            my $trans_obj = NACL::Transit->new( name => $node_name );
            push @subtest_handles, subtest(sub {
                $trans_obj->change_state( to      => 'FIRMWARE',
                                          haltcmd => $halt_cmd,
                                          timeout => 2000 );
                },
                '-bg',
                '-runid',
                $node_name . '_prepare_for_switchover_execution_down');
            push(@halted_node_objects, $trans_obj);
        }
        Subtest::wait_finish(subtests => \@subtest_handles);

        # save this information for rebooting halted nodes
        $self->_pre_switchover_halted_node_transit_objs(
                                                \@halted_node_objects);
    } elsif ($opts{do_peer_metrocluster_site_nodes_halt} == 5) {
        # 5 - reboot
        $Log->comment('Rebooting without inhibit takeover of MCC associated nodes at disaster site...');
        @subtest_handles = ();
        @halted_node_objects = ();
        @peer_cluster_mcc_nodes =
                             @{ $self->get_peer()->L_get_cluster_mcc_nodes() };

        foreach my $peer_mcc_node (@peer_cluster_mcc_nodes) {
            my $node_name = $peer_mcc_node->get_name();
            my $halt_cmd  = 'set -confirmations off;'.
                            'system node reboot -reason '.
                            'prepare_for_switchover_execution -node '.
                            $node_name;
            my $trans_obj = NACL::Transit->new( name => $node_name );
            push @subtest_handles, subtest(sub {
                $trans_obj->change_state( to      => 'FIRMWARE',
                                          haltcmd => $halt_cmd,
                                          timeout => 2000 );
                },
                '-bg',
                '-runid',
                $node_name . '_prepare_for_switchover_execution_down');
            push(@halted_node_objects, $trans_obj);
        }
        Subtest::wait_finish(subtests => \@subtest_handles);

        # save this information for rebooting halted nodes
        $self->_pre_switchover_halted_node_transit_objs(
                                                \@halted_node_objects);
    } elsif ($opts{do_peer_metrocluster_site_nodes_halt} == 6) {
    # 6 - Do nothing (for planned switchover)
        $self->_pre_switchover_halted_node_transit_objs(undef);
    } else {
        $self->_pre_switchover_halted_node_transit_objs(undef);
    }

    if ($opts{do_wait_minutes_for_cluster_appears_down}) {
        $Log->comment('Sleeping for ' .
              $opts{do_wait_minutes_for_cluster_appears_down} . ' minutes...');
        sleep (60 * $opts{do_wait_minutes_for_cluster_appears_down});
    }
}

=head2 prepare_for_switchback_execution()

    $cluster_A->prepare_for_switchback_execution(
                # default values shown here
            do_peer_metrocluster_site_nodes_boot           => 1,
            do_intercluster_network_interface_enable       => 0,
            do_peer_metrocluster_command_interface_refresh => 1,
            do_wait_seconds_for_cluster_peering_up         => 90,

                # vsims are slow re disk migration
                # so is reserver-id currently in real systems
            timeout_for_slow_disk_SANOWN_updates      => 1800,

            # this is obsolete
            do_disaster_site_aggrs_offline                 => 0,

This method performs various actions often/usually performed before this site's
cluster executes a DR switchback command.

=over

=item Options

=item C<< do_peer_metrocluster_site_nodes_boot => 0|1 (default = 1) >>

(Optional) This option will attempt to boot all the nodes in this cluster's
Metrocluster peer cluster.  (I.e., if this called on cluster A, it wiil try
to boot all nodes in cluster B.)

=item C<< do_intercluster_network_interface_enable => 0|1 (default = 0) >>

(Optional) This option will enable the intercluster network interface on this
cluster which allows this cluster to see its Metrocluster peer cluster.  (See
also the do_intercluster_network_interface_disable option in
prepare_for_switchover_execution().)

=item C<< do_peer_metrocluster_command_interface_refresh => 0|1 (default = 1) >>

(Optional) This option will re-enable the command interface to allow
automation to access the disaster site CLI again.

=item C<< do_wait_seconds_for_cluster_peering_up => 90 (default = 90) >>

(Optional) This option manipulates the amount of time to wait for cluster
peering to be reinstated beteen the metrocluster sites after the disaster site
nodes have been brought back up.

=item C<< timeout_for_slow_disk_SANOWN_updates => 1800 (default = 1800) >>

(Optional) This is the same phenomenon as is addressed in
perform_switchback_execution_recovery().

=item C<< do_disaster_site_aggrs_offline => 0|1 (default = 0) >>

(Optional) This option will offline all the DR aggrs before switchback
operation to avoid panic.

NOTE: sn_warm chunk 5 renders this option obsolete.  The default is now set to
0 (false), and a warning will be logged if it is set to 1 by the user.  Plan to
remove this option at the start of chunk 6.

=back

=cut

sub prepare_for_switchback_execution {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %raw_opts = @_;

    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
              do_peer_metrocluster_site_nodes_boot           =>
                                   { regex   => qr/^[01]$/, default => 1, },
              do_peer_metrocluster_site_after_nodes_boot           =>
                                   { regex   => qr/^[01]$/, default => 0, },
              do_intercluster_network_interface_enable       =>
                                   { regex   => qr/^[01]$/, default => 0, },
              do_peer_metrocluster_command_interface_refresh =>
                                   { regex   => qr/^[01]$/, default => 1, },
              do_wait_seconds_for_cluster_peering_up         =>
                                   { regex   => qr/^\d+$/,  default => 90, },
              timeout_for_slow_disk_SANOWN_updates           =>
                                   { regex   => qr/^\d+$/,  default => 7200, },
              do_disaster_site_aggrs_offline                 =>
                                   { regex   => qr/^[01]$/, default => 0, },
              send_external_transit_objs                 =>
                                   { regex   => qr/^[01]$/, default => 0, optional => 1, },
              transit_objs                                   =>
                                   { type => ARRAYREF,      optional => 1, }, 
              endstate                                       =>
                                   { type => ARRAYREF,      optional => 1, }, 
                  },
                            );

    if ($opts{do_peer_metrocluster_site_nodes_boot}) {
        $Log->comment('Booting MCC associated nodes at disaster site...');
        my @subtest_handles = ();
        my $halted_node_obj;
        my $endstate; my $inc = 0;

        # mica wafl-changes  
        if ($opts{send_external_transit_objs}) {
            $Log->comment("Using external transit objs received from script");
            $halted_node_obj = $opts{transit_objs};        
            $Log->comment("Transit objs :" . NACL::ComponentUtils::Dumper($halted_node_obj));
            $endstate = $opts{endstate};
            $Log->comment("Endstate     :" . NACL::ComponentUtils::Dumper($endstate));
        } else {
            $halted_node_obj =
                           $self->_pre_switchover_halted_node_transit_objs();
            $Log->comment("Transit objs :" . NACL::ComponentUtils::Dumper($halted_node_obj));
            $endstate =['CLI','CLI']
        }
                
        foreach my $transit_obj (@{$halted_node_obj}) {
            my $to = ${$endstate}[$inc];
            push @subtest_handles, subtest(sub {
                $transit_obj->change_state( to => "$to", timeout => 2500);
                },
                '-bg',
                '-runid',
                $transit_obj->{conn}->{connhost}.
                '_prepare_for_switchback_execution_up');
            $inc++;
        }
        Subtest::wait_finish( subtests => \@subtest_handles );
        $self->_pre_switchover_halted_node_transit_objs(undef);
    }

    if ($opts{do_peer_metrocluster_command_interface_refresh}) {
        my $disaster_cluster = $self->get_peer();
        my $refresh_timeout_left = 600;
        my $refresh_timeout_end   = time + $refresh_timeout_left;
		my $refresh_done = 0;

		# It takes a while for ssh server to come up. Keep polling until its up
        do {
			my $disaster_site_ci = $disaster_cluster->
			                get_Metrocluster_component()->command_interface();
			my $session_is_alive = $disaster_site_ci->apiset()->
			                                           api_is_session_alive();
            if (!$session_is_alive){
			    $disaster_site_ci->apiset()->api_reconnect_session();
                $Log->comment('ssh server is not up will try again after 10 sec');
                Tharn::snooze(10);
            } else {
                $refresh_done = 1;
            }
        }while (time < $refresh_timeout_end && !$refresh_done);

		$Log->comment('Refreshing disaster site command interfaces...');

		# cluster mgmt CI
        $disaster_cluster->get_Metrocluster_component()->command_interface()->
		                    refresh_command_interface(no_version_change => 1);
        # node mgmt CIs
        my @mcc_nodes = @{ $disaster_cluster->L_get_cluster_mcc_nodes() };
        foreach my $mcc_node (@mcc_nodes) {
            $mcc_node->get_Node_component()->command_interface()->
                             refresh_command_interface(no_version_change => 1);
        }
    }

    # additional workaround stuff for do_peer_metrocluster_site_nodes_boot
    if ($opts{do_peer_metrocluster_site_after_nodes_boot}) {
        # big workaround after nodes reboot...
        unless ($self->_metrocluster()->has_uichange(
                            uichange =>'cluster-stream-down-on-switchover')) {
            $Log->comment('BURT 554679 + 662064 workaround.  BEGIN to bring up '.
                          'admin-state of crs stream');
            #BURT 662064 mods:
            #survivor SENDER stream up before switcback
            #we are skipping disaster side stuff as dev _thinks_ it's OK.
            #put in if needed
            ##disaster RECEIVER stream up before switchback
            #
            #survivor RECEIVER stream up AFTER switcback
            #we are skipping disaster side stuff as dev _thinks_ it's OK.
            #put in if needed
            ##disaster SENDER stream up AFTER switchback

            my $survivor_site_ci =
                      $self->get_Metrocluster_component()->command_interface();

            my $parsed_output = $survivor_site_ci->apiset()->
                          cluster_identity_show->get_parsed_output();

            my $cluster_uuid = $parsed_output->[0]->{'uuid'};

            $parsed_output = $survivor_site_ci->apiset()->
                        debug_crs_stream_show(
                            'direction' => 'sender',
                            'synopsis'  => '"MetroCluster Cluster Domain Sender Stream"',
                        )->get_parsed_output();

            my $cmd = 'debug smdb table mccStreamDBv2 modify -cluster-uuid '.
                  $cluster_uuid.' -stream-target-admin-state up ';

            foreach my $stream (@{$parsed_output}) {
                my $c_cmd =  $cmd.'-stream-uuid '.$stream->{'stream-uuid'};

                $survivor_site_ci->apiset()->execute_command(
                                                      command => $c_cmd );

                $survivor_site_ci->apiset()->
                    debug_crs_stream_modify_stream(
                                'stream-uuid' => $stream->{'stream-uuid'},
                                'admin-state' => 'up');

            }

            # the stream may not come up immediately, retry for a while...
            my $end_time = time() + (2 * 60);    # time allowed
            my $stream_done = 0;
            while (!$stream_done && (time() < $end_time)) {
                $parsed_output = $survivor_site_ci->apiset()->
                    debug_crs_stream_show(
                        'direction' => 'sender',
                        'synopsis'  =>
                            '"MetroCluster Cluster Domain Sender Stream"',
                    )->get_parsed_output();

                my $check = 1;
                foreach my $stream (@{$parsed_output}) {
                    if ($stream->{'admin-state'} ne 'up') {
                        $check = 0;
                        # iff retry time ran out, throw an exception here
                        if (time() > $end_time) {
                            WaflMCC::MCC::Exceptions::BaseException->
                                throw('Failed to modify stream admin-state to up');
                        }
                    }
                }
                $stream_done = $check;
                if (!$stream_done) {
                    $Log->comment("sleeping 10 to retry stream up check...");
                    Tharn::snooze(10);
                }
            }
            # shouldn't really need this except for corner cases maybe
            if (!$stream_done) {
                WaflMCC::MCC::Exceptions::BaseException->
                    throw('Failed to modify stream admin-state to up');
            }
        }

        unless ($self->_metrocluster()->workaround_expired(
                          uichange => 'mcc_expire_workaround_652320',
                          burt => '645504')) {
            $Log->comment('Part 2/3 of the BURT 645504 workaround.  BEGIN...');
            $Log->comment('Stopping all data vservers (not the special ' .
                          'vservers such as vs0, MCC_Vserver and ' .
                          'CRS_Vserver) at the disaster site...');
            my $survivor_site_ci =
                     $self->get_Metrocluster_component()->command_interface();
            # re "disaster_site_ci": makes sense only after heal roots...
            my $disaster_site_ci = $self->get_peer()->
                            get_Metrocluster_component()->command_interface();

            # subtype seems to be missing from component spec.  must use apiset.
            #my @disaster_site_src_data_vservers = NACL::C::Vserver->find(
            #           command_interface => $disaster_site_ci,
            #           filter            => { type    => 'data|cluster',
            #                                  subtype => 'sync-source', },
            #           allow_empty       => 1 );
            try {
                my $vserver_specs = $disaster_site_ci->apiset()->vserver_show(
                           type    => 'data',
                           subtype => 'sync-source')->get_parsed_output();
                my @disaster_site_src_data_vservers = map {
                        NACL::C::Vserver->new(
                                command_interface => $disaster_site_ci,
                                vserver           => $_->{vserver})
                                                          } @{ $vserver_specs };

                # remove any vservers on node's root aggr (is redundant check?)
                # TBD If really required

                # stop each such vserver found
                foreach my $vserver (@disaster_site_src_data_vservers) {
                    try {
                        $vserver->stop();
                    } catch NACL::APISet::Exceptions::CommandFailedException
                                                                         with {
                        # ignore already stopped vservers
                        my $exception = shift;
                        my $acceptable_error_msg =
                            'Vserver stop requires admin-state of the ' .
                            'Vserver to be running or restricted';
                        unless ($exception->text() =~ /$acceptable_error_msg/) {
                            # rethrow unexpected exception
                            $exception->throw();
                        }
                    };
                }
            } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
                # why this is an exception is beyond me...  :p    ignore it.
            };

            $Log->comment('Sleeping 60 sec for vserver stops...');
            Tharn::snooze(60);

            $Log->comment('Showing dm vserver streams @ survivor site:');
            $survivor_site_ci->apiset()->execute_command(
                                    command => 'debug dm vserver stream show');
            $Log->comment('Showing dm vserver streams @ disaster site:');
            $disaster_site_ci->apiset()->execute_command(
                                    command => 'debug dm vserver stream show');

            $Log->comment('Deleting all sender side streams at the ' .
                          'disaster site...');
            my $cmd = 'debug dm vserver stream delete -stream-uuid * ' .
                      '-abort-pending-requests true';
            $disaster_site_ci->apiset()->execute_command(command => $cmd);

            $Log->comment('Sleeping 120 sec for table to update...');
            Tharn::snooze(120);

            # verification
            $cmd = 'debug dm vserver stream show';
            my $output =
                $disaster_site_ci->apiset()->execute_command(command => $cmd);

            unless ($output =~ /This table is currently empty/i) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                                          'Failed to delete vserver steams.');
            }

            $Log->comment('On survivor site: re-start the switched over ' .
                          'data vservers...');

            # subtype seems to be missing from component spec.  must use apiset.
            #my @SOed_data_vservers = NACL::C::Vserver->find(
            #           command_interface => $survivor_site_ci,
            #           filter            => { type    => 'data|cluster',
            #                                  subtype => 'sync-destination', },
            #           allow_empty       => 1 );
            my @SOed_data_vservers = undef;
            try {
                my $vserver_specs = $survivor_site_ci->apiset()->vserver_show(
                           type    => 'data',
                           subtype => 'sync-destination')->get_parsed_output();
                @SOed_data_vservers = map {
                            NACL::C::Vserver->new(
                                    command_interface => $survivor_site_ci,
                                    vserver           => $_->{vserver})
                                             } @{ $vserver_specs };

                # remove any vservers on node's root aggr (is redundant check?)
                # TBD If really required

                # start each such vserver found
                map { $_->start() } @SOed_data_vservers;
            } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
                # why this is an exception is beyond me...  :p    ignore it.
                @SOed_data_vservers = ();
            };

            $Log->comment('Sleeping 120 sec for table to update...');
            Tharn::snooze(120);

            # convenience variables only
            my $REBASELINE_COMMAND = 'debug smdb table ' .
               'metrocluster_vserver_rebaseline_all_vservers run -drc-job-id 1';
            my $TRACKER_TABLE = 'metrocluster_vserver_rebaseline_tracker';

            # deal with the possibility of old entries in
            # metrocluster_vserver_rebaseline_tracker table
            {
                my $entries = $survivor_site_ci->apiset()->
                        debug_smdb_table_show(table => $TRACKER_TABLE)->
                                                           get_parsed_output();
                if (ref($entries) eq 'ARRAY') {
                    if (scalar(@{ $entries }) > 0) {
                        $Log->comment('Cleaning up old tracker entries.');
                        $survivor_site_ci->apiset()->execute_command(
                                               command => $REBASELINE_COMMAND);
                        $Log->comment('Sleeping 30 sec for table to update...');
                        Tharn::snooze(30);
                    }
                }

                #TODO failing with multiple vservers at this time, retry
                my $end_time = time() + (5 * 60);    # time allowed
                while (time() < $end_time) {
                    # retry better give what we expect now...
                    $entries = $survivor_site_ci->apiset()->debug_smdb_table_show(
                                     table => $TRACKER_TABLE)->get_parsed_output();
                    unless ((ref($entries) eq 'ARRAY') &&
                            (scalar( @{ $entries }) == 0)) {
                        $Log->comment(
                                      "Failed to clean the $TRACKER_TABLE " .
                                      "table! (1), retrying for " .
                                      ($end_time - time) .
                                      " more second...");
                        Tharn::snooze(30);
                    } else {
                        last;
                    }
                }

                $entries = $survivor_site_ci->apiset()->debug_smdb_table_show(
                                 table => $TRACKER_TABLE)->get_parsed_output();
                unless ((ref($entries) eq 'ARRAY') &&
                                               (scalar( @{ $entries }) == 0)) {
                    WaflMCC::MCC::Exceptions::BaseException->throw(
                              "Failed to clean the $TRACKER_TABLE table! (1)");
                }
            }

            $Log->comment('On survivor site: rebaseline vservers and wait ' .
                          'for its completion...');
            $survivor_site_ci->apiset()->execute_command(
                                               command => $REBASELINE_COMMAND);

            $Log->comment('Sleeping 120 sec for table to update...');
            Tharn::snooze(120);

            # mica wafl-fix , if dr site is in takeonover state 
            
            # this is essentially a list of vservers we are waiting for
            # rebaseline success on
            my %vserver_rebaseline_waits = map {
                $_->get_one_state_attribute('uuid') => 1; } @SOed_data_vservers;
            $Log->comment('vserver_rebaseline_waits (1) = ' .
                     WaflMCC::MCC::Utils->dump(%vserver_rebaseline_waits));

            my $all_vserver_rebaselines_were_successful = 0;
            my $end_time = time() + (5 * 60);    # time allowed

            # mica 
            while (time() < $end_time) {
                my $parsed_output = $survivor_site_ci->apiset()->
                        debug_smdb_table_show(table => $TRACKER_TABLE)->
                                                           get_parsed_output();
                foreach my $element (@{ $parsed_output }) {
                    my $success_vserver_uuids = $element->{'success-vservers'};
                    if (ref($success_vserver_uuids)) {
                        my @uuids = @{ $success_vserver_uuids };
                        foreach my $uuid (@uuids) {
                            $Log->comment(
                                    "Deleting success vserver uuid = $uuid\n");
                            delete($vserver_rebaseline_waits{$uuid});
                        }
                    } elsif ($success_vserver_uuids ne q{-}) {
                        # i don't think this case should ever happen.
                        $Log->comment('Deleting success vserver uuids = ' .
                                                       $success_vserver_uuids);
                        delete(
                            $vserver_rebaseline_waits{$success_vserver_uuids});
                    } else {
                        $Log->comment("Ignoring vserver uuid = '-'.");
                    }
                }
            
                # not waiting on any more successes?
                $Log->comment('vserver_rebaseline_waits (2) = ' .
                     WaflMCC::MCC::Utils->dump(%vserver_rebaseline_waits));
                if (scalar(keys(%vserver_rebaseline_waits)) == 0) {
                    $all_vserver_rebaselines_were_successful = 1;
                    last;
                } else {
                    Tharn::snooze(15);
                }
            }

            unless ($all_vserver_rebaselines_were_successful) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                          'Failed to rebaseline vservers: ' .
                          join(q{, }, keys(%vserver_rebaseline_waits)) . q{!});
            }

            $Log->comment('Cleaning up all the tracker entries and ' .
                   're-establishing the streams in the forward direction...');
            $survivor_site_ci->apiset()->execute_command(
                                               command => $REBASELINE_COMMAND);
            
            my $entries = $survivor_site_ci->apiset()->debug_smdb_table_show(
                                 table => $TRACKER_TABLE)->get_parsed_output();
            
            unless ((ref($entries) eq 'ARRAY') &&
                                               (scalar( @{ $entries }) == 0)) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                              "Failed to clean the $TRACKER_TABLE table! (2)");
            }

            $Log->comment('Part 2/3 of the BURT 645504 workaround.  ...END.');
        }
    }

    # enable ISL ports if they are disabled earlier
    my $isl_ports = $self->_pre_switchover_disabled_isl_ports();
    if (defined($isl_ports) && scalar @{$isl_ports}) {
        $Log->comment('Enabling the ISL ports');
        foreach my $isl_port (@{$isl_ports}) {
            $isl_port->enable();
        }
    }

    if ($opts{do_intercluster_network_interface_enable}) {
        $Log->comment('Enabling intercluster interfaces to partner site...');
        my $interfaces_to_enable =
               $self->_pre_switchover_intercluster_network_interfaces_downed();

        foreach my $network_interface (@{$interfaces_to_enable}) {
            $network_interface->modify('status-admin' => 'up');
        }
        $self->_pre_switchover_intercluster_network_interfaces_downed(undef);
    }

    # TBD: reimplement as a more exacting way of doing this.
    if ($opts{do_wait_seconds_for_cluster_peering_up}) {
        my $wait_secs = $opts{do_wait_seconds_for_cluster_peering_up};
        $Log->comment("Waiting $wait_secs seconds for reinstatement of " .
                      "cluster peering...");
        sleep($wait_secs);
    }

    $self->L_workaround_force_reserver_id_update();

    if ($opts{timeout_for_slow_disk_SANOWN_updates}) {
        my $wait_secs = $opts{timeout_for_slow_disk_SANOWN_updates} -
                                 $opts{do_wait_seconds_for_cluster_peering_up};
        if ($wait_secs > 0) {
            my $dr_groups = $self->_metrocluster()->get_DR_groups();
            my $sanown_timeout_end = time + $wait_secs;
            # we don't want to wait more than timeout (not timeout per call)
            $Log->comment(
                 "Verifying SANOWN disk state migrations...");

            foreach my $dr (sort keys %{$dr_groups}) {
                my $dr_group = $dr_groups->{$dr};
                $Log->comment(
                        "Waiting for SANOWN disk state migrations: $dr ...");

                # timeout is total time, not each
                # L_wait_for_slow_SANOWN_state_migration logs warnings if needed
                $dr_group->get_shared_storage()->
                              L_wait_for_slow_SANOWN_state_migration(
                                        timeout => $sanown_timeout_end - time);
            }
        }
    }

    if ($opts{do_disaster_site_aggrs_offline}){
        $Log->warn('prepare_for_switchback_execution(' .
            'do_disaster_site_aggrs_offline => 1) is probably an ' .
            'obsolete usage!');
        my @mcc_nodes = @{ $self->L_get_cluster_mcc_nodes() };
        foreach my $mcc_node (@mcc_nodes) {
            my $node = $mcc_node->get_Node_component();
            my $response = $node->run(command => 'aggr status');
            my @aggrs = split(/\n/,$response);

            foreach my $aggr_names(@aggrs){
                if($aggr_names =~ /offline|online/){
                    my @aggr_name = split(/offline|online/,$aggr_names);
                    my $cmd = '"priv set test;raid_config info listaggr '.
                              $aggr_name[0].'"';
                    $response = $node->run(command => $cmd);

                    if( $response !~ /dr_home_owner_id=0/i) {
                        $cmd = '"priv set test;aggr offline ' . $aggr_name[0] .
                               ' -a -f"';
                        $response = $node->run(command => $cmd);
                        $Log->comment($response);
                    }
                }
            }
        }
    }
}

# runs workaround to try to force reserver ids to update quickly
sub L_workaround_force_reserver_id_update {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            known_live_sites => { type => ARRAYREF,
                                  optional => 1, },
                  },
    );

    my $live_sites_ref = (defined($opts{known_live_sites})) ?
                         $opts{known_live_sites} :
                         $self->_metrocluster()->L_determine_live_sites();

    unless ($self->_metrocluster()->workaround_expired(
                      uichange => 'mcc_expire_workaround_649633',
                      burt => '649633')) {
        my @live_sites = @{ $live_sites_ref };
        foreach my $site (@live_sites) {
            my $cluster = $self->_metrocluster()->get_cluster(site => $site);
            my $sites_nodes = $cluster->L_get_cluster_mcc_nodes();

            foreach my $mcc_node (@{ $sites_nodes }) {
                try {
                    my $ci = $mcc_node->get_Node_component()->command_interface();
                    my $out = $ci->apiset()->execute_command(
                        command => 'node run local disk show -v -z');
                    #print $mcc_node->get_name() . ' node run local disk show:' .
                    #Dumper $out;
                } catch NATE::BaseException with {
                    $Log->comment("Ignoring error in workaround attempt ");
                };
            }
        }
    }

}

=head2 cycle_to_normal_DR_state()

    $site_X_cluster->cycle_to_normal_DR_state( );
    $site_X_cluster->cycle_to_normal_DR_state( verify => {disks => 0} );

Perform necessary operations to bring DR state back to normal.

=over

=item Options

=item C<< verify => { disks }  >>

Verify is a hashref containing the various things you can verify.
Currently disks is the only value.

=over

=item C<< disks => 0|1 (default = 1) >>

(Optional) This option will verify the disks after each state transition
if set to 1. Internally it just calls library methods to verify disks.

=back

=back

=cut

sub cycle_to_normal_DR_state {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my ($self,%opts) = @_;

    $opts{'state'} = 'normal';

    my $state = $self->cycle_to_DR_state(
                                          %opts, #validate in cycle_to_DR_state
                                        );

    return $state;
}

=head2 cycle_to_DR_state()

    $site_X_cluster->cycle_to_DR_state( state => 'normal' );
    $site_X_cluster->cycle_to_DR_state( state => 'normal', verify=> {disks => 0} );

Perform necessary operations to bring DR state to the specified state.

=over

=item Options

=item C<< state => normal | switchover_completed | heal_aggrs_completed |
heal_roots_completed >>

=item C<< disable_cycle_short_circuit => 0 | 1 (default = 0) >>

(Optional)  By default (disable_cycle_short_circuit=0) the DR state will not
change if it is asked to cycle to a state it is already in, it will simply log
that it is already in that state and exit.

If disable_cycle_short_circuit=1, it will force a full cycle back to the
specified state even if that is the start state.

=item C<< verify => { disks } >>

Verify is a hashref containing the various things you can verify.
Currently disks is the only value.

=over

=item C<< disks => 0|1 (default = 1) >>

(Optional) This option will verify the disks after each state transition
if set to 1. Internally it just calls library methods to verify disks.

=back

=back

=cut

sub cycle_to_DR_state {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my ($self,%opts) = @_;

    # validate verify opts (because Params::Validate can't handle
    # recursive hashrefs)
    $opts{verify} = validate_with(
        params => $opts{verify} || {},
        spec   => {
                       disks => { regex => qr/^(0|1)$/, default => 1 },
        },
    );


    # validate rest of opts
    %opts = validate_with(
        params => \%opts,
        spec   => {
            verify                      => HASHREF,
            disable_cycle_short_circuit => { type => BOOLEAN, default => 0},
            state                       => { regex =>
                                 qr/^(normal|switchover_completed|
                                 heal_aggrs_completed|heal_roots_completed)$/x
                                           },
                  },
                         );

    my $next = {
                   'normal'               => 'switchover_completed',
                   'switchover_completed' => 'heal_aggrs_completed',
                   'heal_aggrs_completed' => 'heal_roots_completed',
                   'heal_roots_completed' => 'normal',
    };

    my $verify_disks = delete $opts{verify}->{disks};
    my $target_state = delete $opts{state};

    my $disable_cycle_short_circuit = delete $opts{disable_cycle_short_circuit};

    $Log->comment("Target metrocluster state $target_state");

    my $dr_coordinator = $self->get_DR_Coordinator();
    my $dr_groups = $self->_metrocluster()->get_DR_groups();

    my $state = $dr_coordinator->get_DR_state();
    $Log->comment("Currently in metrocluster state $state");

    my $transitions = 0; # number of transitions

    #if this is the start and disable_cycle_short_circuit=1,
    #transition anyway even if in target state
    while ((!$transitions && $disable_cycle_short_circuit) ||
                                                     $state ne $target_state) {
        my $next_state = $next->{$state};
        $Log->comment("Init transition to metrocluster state $next_state");

        #we could shove these calls into a hash but this allows more flexibility
        if ($next_state eq 'switchover_completed') {
                $self->prepare_for_switchover_execution(
                    do_intercluster_network_interface_disable => 0,
                    do_peer_metrocluster_site_nodes_halt      => 1,
                    #do_wait_minutes_for_cluster_appears_down  => 0,
                );

            $dr_coordinator->DR_switchover();
        } elsif ($next_state eq 'heal_aggrs_completed') {
            $dr_coordinator->DR_heal_aggrs();
        } elsif ($next_state eq 'heal_roots_completed') {
            $dr_coordinator->DR_heal_root_aggrs();

            # this is placed here because it really is part of the recovery
            # after the heal roots phase, and before disk validations can be
            # done in this new state.
            $self->prepare_for_switchback_execution(
                    do_peer_metrocluster_site_nodes_boot           => 1,
                    do_peer_metrocluster_command_interface_refresh => 1,
                                                   );
        } elsif ($next_state eq 'normal') {
            $dr_coordinator->DR_switchback();

            $self->perform_switchback_execution_recovery(
                    do_peer_metrocluster_site_nodes_boot => 0,
                                                        );
        } else {
            # how did we get here?!
            $Log->comment("ERROR: bad next state $next_state (state=$state)");
            return;
        }

        if ($verify_disks) {
            foreach my $dr (sort keys %{$dr_groups}) {
                $Log->comment("validate shared storage for dr group $dr");

                $dr_groups->{$dr}->get_shared_storage()->validate_disk_states();
            }
        } else {
            $Log->comment('skipping disk verification...');
        }

        $transitions++;
        $state = $dr_coordinator->get_DR_state();
    }

    return $state;
}


=head2 L_get_cluster_mcc_nodes()

This library-only (L_) method finds all nodes in this cluster which are in MCC
relationships.  Returns a reference to an array of WaflMCC::MCC::Node
objects.  This method should not find nodes in this cluster's DR partner
cluster, nor should it return nodes in this cluster which are not a part of any
MCC relationship.

=cut

sub L_get_cluster_mcc_nodes {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @nodes = ();
    my @dr_groups = values(%{$self->_metrocluster()->get_DR_groups()});
    foreach my $dr_group (@dr_groups) {
        my @dr_partnerships = @{ $dr_group->get_DR_partnerships() };
        foreach my $partnership (@dr_partnerships) {
            push(@nodes,$partnership->get_partner_node(site => $self->_site()));
        }
    }

    return \@nodes;
}

# This method is used to find what network interfaces on THIS cluster are
# associated with the intercluster addresses that the DR cluster uses to
# communicate to THIS cluster.  It returns an array of NACL::C::NetworkInterface
# objects.  (Usually it will only be one object, but is possible that
# multiple such interfaces could exist (I think.))
sub _find_DR_cluster_to_THIS_cluster_intercluster_interfaces {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $mcc_component = $self->get_Metrocluster_component();
    my $ci = $mcc_component->command_interface();

    my @network_interfaces = ();

    my $this_cluster_name = $self->get_name();

    # be careful, this is confusing.  we are looking for the ClusterPeer
    # element on the DR site which has information about THIS site.  I.e.,
    # don't miss the get_peer() call in here!
    my $dr_site_ClusterPeer_component = $self->get_peer()->
         _get_ClusterPeer_component(remote_cluster_name => $this_cluster_name);

    # these are the interface addresses on THIS cluster which we want
    my @IC_addresses_to_this_site_from_dr_site =
        $dr_site_ClusterPeer_component-> get_one_state_attribute('peer-addrs');

    # find all the intercluster IFs on THIS cluster
    my @interface_components = NACL::C::NetworkInterface->find(
                    command_interface => $ci,
                    filter            => { role => 'intercluster' },
                );

    # if they match any of the DR cluster's remote addresses to THIS cluster
    foreach my $IC_IFs (@interface_components) {
        my $address = $IC_IFs->get_one_state_attribute('address');
        if (any { $_ eq $address } @IC_addresses_to_this_site_from_dr_site) {
            push(@network_interfaces, $IC_IFs);
        }
    }

    return @network_interfaces;
}

=head2 perform_switchback_execution_recovery()

    $cluster_A->perform_switchback_execution_recovery(
            # see BURT 559197
            timeout_for_slow_disk_SANOWN_updates           => 600,
            # probably obsolete use here
            do_peer_metrocluster_site_nodes_boot           => 0,
            do_intercluster_network_interface_enable       => 0,
            do_peer_metrocluster_command_interface_refresh => 0,
                                                            );

As of sn_warm chunk 5, most of this functionality has been replicated (and has
more valid usage) in prepare_for_switchback_execution().  The default values
for all options which were moved to prepare_for_switchback_execution()  in this
method are currently 0 (FALSE) at this point.

This method performs various actions to be performed after this site's
cluster executes a DR switchback command.

=over

=item Options

=item C<< timeout_for_slow_disk_SANOWN_updates => N (default = 600) >>

(Optional) [See BURT 559197] The library now actively checks to see if
disks complete the migration of their SANOWN states to the proper post-SB
states (normal).
This parameter sets the timeout.  If it takes longer than the timeout for
the states to migrate back to normal it will abort the wait and continue
anyway.

=item C<< do_intercluster_network_interface_enable => 0|1 (default = 0) >>

(Optional) [OBSOLETE usage here] This option will enable the intercluster
network interface on this cluster which allows this cluster to see its
Metrocluster peer cluster.

=item C<< do_peer_metrocluster_site_nodes_boot => 0|1 (default = 0) >>

(Optional) [OBSOLETE usage here] This option will attempt to boot all the nodes
in this cluster's Metrocluster peer cluster.  (I.e., if this called on cluster
A, it wiil try to boot all nodes in cluster B.)

=item C<< do_peer_metrocluster_command_interface_refresh => 0|1 (default = 0) >>

(Optional) [OBSOLETE usage here] This option will re-enable the command
interface to allow automation to access the disaster site CLI again.

=back

=cut

#TBD: Look later at this method and remove obsolete junk at some point.
sub perform_switchback_execution_recovery {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %raw_opts = @_;

    # validate opts

    # need to know if this was explicitly passed or falling back on default
    my $timeout_for_slow_disk_SANOWN_updates_opt_was_passed =
       List::MoreUtils::any { $_ eq 'timeout_for_slow_disk_SANOWN_updates'} @_;

    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            timeout_for_slow_disk_SANOWN_updates   =>
                                   { regex   => qr/^\d+$/,  default => 6000, },
            do_peer_metrocluster_site_nodes_boot           =>
                                   { regex   => qr/^[01]$/, default => 0, },
            do_intercluster_network_interface_enable       =>
                                   { regex   => qr/^[01]$/, default => 0, },
            do_peer_metrocluster_command_interface_refresh =>
                                   { regex   => qr/^[01]$/, default => 0, },
                  },
                            );

    unless ($self->_metrocluster()->workaround_expired(
                      uichange => 'mcc_expire_workaround_652320',
                      burt => '645504')) {
        $Log->comment('Part 3/3 of the BURT 645504 workaround.  BEGIN...');
        $Log->comment('Showing switched-back aggr status (re BURT 640067)...');
        my $until_recently_disaster_site_ci = $self->get_peer()->
                            get_Metrocluster_component()->command_interface();
        $until_recently_disaster_site_ci->apiset()->
                                   execute_command(command => 'disk show -v');

        $Log->comment('Part 3/3 of the BURT 645504 workaround.  ...END.');
    }


    if ($opts{do_peer_metrocluster_site_nodes_boot}) {
        $Log->comment('Booting MCC associated nodes at disaster site...');
        $Log->warn('perform_switchback_execution_recovery(' .
            'do_peer_metrocluster_site_nodes_boot => 1) is probably an ' .
            'obsolete usage!');
        my @subtest_handles = ();
        my $halted_node_transit_obj =
                           $self->_pre_switchover_halted_node_transit_objs();
        foreach my $node_transit_obj (@{$halted_node_transit_obj}) {
            push @subtest_handles, subtest(sub {
                $node_transit_obj->change_state(to => 'CLI', timeout => 1500);
                },
                '-bg',
                '-runid',
                $node_transit_obj->{conn}->{connhost}
                . '_perform_switchback_execution_recovery_up');
        }
        Subtest::wait_finish( subtests => \@subtest_handles );
        $self->_pre_switchover_halted_node_transit_objs(undef);
    }

    if ($opts{do_intercluster_network_interface_enable}) {
        $Log->comment('Enabling intercluster interfaces to partner site...');
        $Log->warn('perform_switchback_execution_recovery(' .
            'do_intercluster_network_interface_enable => 1) is probably an ' .
            'obsolete usage!');
        my $interfaces_to_enable =
               $self->_pre_switchover_intercluster_network_interfaces_downed();

        foreach my $network_interface (@{$interfaces_to_enable}) {
            $network_interface->modify('status-admin' => 'up');
        }
        $self->_pre_switchover_intercluster_network_interfaces_downed(undef);
    }

    if ($opts{do_peer_metrocluster_command_interface_refresh}) {
        $Log->comment('Refreshing disaster site command interface...');
        $Log->warn('perform_switchback_execution_recovery(' .
            'do_peer_metrocluster_command_interface_refresh => 1) is ' .
            'probably an obsolete usage!');
        my $disaster_site = $self->get_peer()->_site();
        $self->_metrocluster()->refresh_command_interfaces(
                                                       site => $disaster_site);
    }

    $self->L_workaround_force_reserver_id_update();

    unless ($self->_metrocluster()->has_uichange(
                        uichange =>'cluster-stream-down-on-switchover')) {
        $Log->comment('BURT 554679 + 662064 workaround.  BEGIN to bring up '.
                      'admin-state of crs stream');
        #BURT 662064 mods:
        #
        #survivor RECEIVER stream up AFTER switcback
        #we are skipping disaster side stuff as dev _thinks_ it's OK.
        #put in if needed
        ##disaster SENDER stream up AFTER switchback

        my $survivor_site_ci =
                  $self->get_Metrocluster_component()->command_interface();

        my $parsed_output = $survivor_site_ci->apiset()->
                      cluster_identity_show->get_parsed_output();

        my $cluster_uuid = $parsed_output->[0]->{'uuid'};

        $parsed_output = $survivor_site_ci->apiset()->
                    debug_crs_stream_show(
                        'direction' => 'receiver',
                        'synopsis'  => '"MetroCluster Cluster Domain Receiver Stream"',
                    )->get_parsed_output();

        my $cmd = 'debug smdb table mccStreamDBv2 modify -cluster-uuid '.
          $cluster_uuid.' -stream-target-admin-state up ';

        foreach my $stream (@{$parsed_output}) {
            my $c_cmd =  $cmd.'-stream-uuid '.$stream->{'stream-uuid'};

            $survivor_site_ci->apiset()->execute_command(
                                                  command => $c_cmd );

            $survivor_site_ci->apiset()->
                debug_crs_stream_modify_stream(
                            'stream-uuid' => $stream->{'stream-uuid'},
                            'admin-state' => 'up');

        }

        # the stream may not come up immediately, retry for a while...
        my $end_time = time() + (2 * 60);    # time allowed
        my $stream_done = 0;
        while (!$stream_done && (time() < $end_time)) {
            $parsed_output = $survivor_site_ci->apiset()->
                debug_crs_stream_show(
                    'direction' => 'receiver',
                    'synopsis'  =>
                        '"MetroCluster Cluster Domain Receiver Stream"',
                )->get_parsed_output();

            my $check = 1;
            foreach my $stream (@{$parsed_output}) {
                if ($stream->{'admin-state'} ne 'up') {
                    $check = 0;
                    # iff retry time ran out, throw an exception here
                    if (time() > $end_time) {
                        WaflMCC::MCC::Exceptions::BaseException->
                            throw('Failed to modify stream admin-state to up');
                    }
                }
            }
            $stream_done = $check;
            if (!$stream_done) {
                $Log->comment("sleeping 10 to retry stream up check...");
                Tharn::snooze(10);
            }
        }
        # shouldn't really need this except for corner cases maybe
        if (!$stream_done) {
            WaflMCC::MCC::Exceptions::BaseException->
                throw('Failed to modify stream admin-state to up');
        }

    }

    # see BURT 559197 / 541653

    my $dr_groups = $self->_metrocluster()->get_DR_groups();
    my $sanown_timeout_left = $opts{timeout_for_slow_disk_SANOWN_updates};
    my $sanown_timeout_start = time;
    my $sanown_timeout_end   = time + $sanown_timeout_left;
    #we don't want to wait more than timeout really (not timeout per call)
    $Log->comment("Verifying SANOWN disk state migrations...");

    foreach my $dr (sort keys %{$dr_groups}) {
        # timeout is total time, not each
        # L_wait_for_slow_SANOWN_state_migration logs warnings if needed.

        # if timeout_for_slow_disk_SANOWN_updates was not explicitly passed, and
        # this is a vsim system, we need to extend the timeout a lot.
        my $dr_group = $dr_groups->{$dr};
        my $timeout = $sanown_timeout_end - time;
        if (($dr_group->L_is_vsim_group()) &&
                    (! $timeout_for_slow_disk_SANOWN_updates_opt_was_passed)) {
            $timeout *= 10;
        }

        $dr_group->get_shared_storage()->
                   L_wait_for_slow_SANOWN_state_migration(timeout => $timeout);
    }

    return;
}

1;
