# $Id: //depot/prod/test/main/lib/NACL/MTask/MCC/DR_Coordinator.pm#22 $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)

# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

## @summary MCC DR_Coordinator Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::DR_Coordinator;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::Log ();

use NACL::C::Metrocluster ();
use NACL::CS::MetroclusterOperation ();
use NACL::STask::Vserver ();
use NACL::Exceptions::NoElementsFound qw(:try);
# for burt 667721
#use NACL::C::DebugVreport ();
use Tharn;
use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);
use Readonly;
use Data::Dumper;
use List::MoreUtils qw{ any };

# the values are strings to attempt matching with the dr_state return
# value from 'metrocluster operation show'

Readonly my %DR_STATES => (
       normal =>
                { op => 'Switchback', state => 'Successful', },
       switchover_in_progress =>
                { op => 'Switchover', state => 'In Progress', },
       switchover_completed   =>
                { op => 'Switchover', state => 'Successful', },
       heal_aggrs_in_progress =>
                { op => 'Heal Aggregates', state => 'In Progress', },
       heal_aggrs_completed   =>
                { op => 'Heal Aggregates', state => 'Successful', },
       heal_roots_in_progress =>
                { op => 'Heal Root Aggregates', state => 'In Progress', },
       heal_roots_completed   =>
                { op => 'Heal Root Aggregates', state => 'Successful', },
       switchback_in_progress =>
                { op => 'Switchback', state => 'In Progress', },
       no_op =>
                { op => 'META_no_op', state => 'META_no_state', },
       ignore__metrocluster_modify_unknown =>
                { op => 'Modify', state => 'Unknown', },
       ignore__metrocluster_modify =>
                { op => 'Modify', state => 'Successful', },
       ignore__metrocluster_configure_unknown =>
                { op => 'Configure', state => 'Unknown', },
       ignore__metrocluster_configure =>
                { op => 'Configure', state => 'Successful', },
       ignore__metrocluster_unconfigure =>
                { op => 'Unconfigure', state => 'Successful', },
       ignore__metrocluster_unconfigure_unknown =>
                { op => 'Unconfigure', state => 'Unknown', },
       ignore__switchback_continuation_agent =>
                { op => 'Switchback Continuation Agent', state => 'Successful', },
       ignore__switchback_continuation_agent_unknown =>
                { op => 'Switchback Continuation Agent', state => 'Unknown', },
                          );

Readonly my %DR_STATES__DEPRECATED => (
        # these are deprecated
       normal =>
                { op => 'Switchback', state => 'completed-successfully', },
       switchover_in_progress =>
                { op => 'Switchover', state => 'in-progress', },
       switchover_completed   =>
                { op => 'Switchover', state => 'completed-successfully', },
       heal_aggrs_in_progress =>
                { op => 'Heal Aggrs', state => 'in-progress', },
       heal_aggrs_completed   =>
                { op => 'Heal Aggrs', state => 'completed-successfully', },
       heal_roots_in_progress =>
                { op => 'Heal Roots', state => 'in-progress', },
       heal_roots_completed   =>
                { op => 'Heal Roots', state => 'completed-successfully', },
       heal_aggrs_in_progress_temp =>
                { op => 'Heal Aggregates', state => 'in-progress', },
       heal_aggrs_completed_temp   =>
                { op => 'Heal Aggregates', state => 'completed-successfully', },
       heal_roots_in_progress_temp =>
                { op => 'Heal Root Aggregates', state => 'in-progress', },
       heal_roots_completed_temp   =>
                { op => 'Heal Root Aggregates', state => 'completed-successfully', },
       switchback_in_progress =>
                { op => 'Switchback', state => 'in-progress', },
       ignore__switchback_continuation_agent_unknown =>
                { op => 'Switchback Continuation Agent', state => 'unknown', },
                          );


=head1 NAME

WaflMCC::MCC::DR_Coordinator

=head1 DESCRIPTION

This class is to be used for C-Mode Metrocluster testing.

    # Start test
    use WaflMCC::MCC::Metrocluster;
    ...
    my $site_A_dr_coordinator =
                          $mcc->get_cluster(site => 'A')->get_DR_coordinator();
    ...
    # Perform test here
    ...
    # End test

=cut

=head1 Public Methods

There is no public constructor for this class.  A
WaflMCC::MCC::DR_Coordinator object is obtained via calls such as

    my $site_A_dr_coordinator =
                          $mcc->get_cluster(site => 'A')->get_DR_Coordinator();

=cut

# Automatically generate the (non-public) constructor (includes a call to
# 'init') and also creates non-public members and non-public accesors for
# them.
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
        { -type => 'WaflMCC::MCC::Metrocluster' },
        '_metrocluster',
              ],
    scalar => [
        qw/ _site /,
              ],
    scalar => [
        # this is a flag that is true only when a test is first started and the
        # metrocluster operation table is seen to start out as empty.  this
        # SHOULD signify that no DR operation has ever occurred on this MCC
        # system since it was configured.
        qw/ _dr_operation_table_ok_to_be_empty /,
              ],
                       ];

# validate the parameters passed in and what else?
sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            metrocluster => { isa   => 'WaflMCC::MCC::Metrocluster' },
            site         => { regex => qr/^[AB]$/ },
        },
                            );

    # set members per the passed constructor arguments.
    $self->_metrocluster($opts{'metrocluster'});
    $self->_site(        $opts{'site'});

    # when we start, we do not yet know if any DR op has ever occured on this
    # mcc system.  this flag is set to false after we have seen an actual
    # result after querying for a DR operation state.  once we have seen
    # an entry in this table, it should never be empty again until unconfig'd.
    $self->_dr_operation_table_ok_to_be_empty(1);

    return;
}

=head2 get_Metrocluster_component()

Utility to get the Metrocluster component associated with executing commands
for this cluster's DR Coordinator.  This is private because users probably
don't need to be getting the Metrocluster component from this class.

Returns a NACL::C::Metrocluster object.

=cut

sub get_Metrocluster_component {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_metrocluster()->
                            get_Metrocluster_component(site => $self->_site());
}

=head2 DR_switchover()

    $site_X_dr_coordinator->DR_switchover(timeout  => xxx;);

Execute a 'DR Switchover' command on the site X cluster command interface.

=over

=item Options:

=over

=item C<< timeout => xxx >>

=back

=back

Optional: See wait_for_DR_state() docs for default timeout value.

* Usage of this parameter is highly discouraged

=cut

sub DR_switchover {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            timeout  => {regex => qr/^\d+$/ , optional => 1 },
        },
                            );


    my $mcc_build = $self->_metrocluster()->get_ontap_version_info();

    if($mcc_build->{branch} eq 'mcc_config_repl'){
        $self->_metrocluster()->refresh_command_interfaces(
                                                    site => $self->_site());
    }

    my $mcc_component = $self->get_Metrocluster_component();
    $Log->comment('Initiating metrocluster switchover...');

    # get timing information
    my $time_start = time();

    my $state = 'switchover_completed';
    my $timeout = $self->_get_DR_state_timeout(state => $state );

    #TODO just for testing this out.  We will generalize ui change strings
    #later, similar to workarounds

    if ( $self->_metrocluster->has_uichange(uichange => 'mcc-switchover-ui-change')) {
        $Log->comment("Using newer switchover ui: mcc-switchover-ui-change");
        #cdefs need to be updated for this to work TODO
        #$mcc_component->switchover(
        #'forced-on-disaster' => 'true',
        #'method-timeout' => $timeout,
        #);
        #for now we need to load our own cdefs
        #for APISet
        my $hostobj = $mcc_component->{command_interface}->{version_manager}->{hostobj};


        my $apiset = NACL::APISet->new(
            hostobj   => $hostobj, #Tharn::host($cluster_name),
            category  => 'Node',
            interface => 'CLI',
            set       => 'CMode'
        );

        $apiset->set_timeout( 'connectrec-timeout' => $timeout );

        #mica   
        my $run =  $apiset->metrocluster_switchover(
                         'privilege-level' => 'test',
                         'forced-on-disaster' => 'true',
                         'override-vetoes' => 'true', 
                      );
    } else {
        #this is another build that doesn't support -forced-on-disaster
        $mcc_component->switchover(
            'method-timeout' => $timeout,
        );
    };

    $self->wait_for_DR_state(state => $state,%opts);

    my $execution_time = time() - $time_start;

    unless ($self->_metrocluster()->has_uichange(
                            uichange =>'cluster-stream-down-on-switchover')) {
        $Log->comment('BURT 554679 workaround.  BEGIN to bring down '.
                      'admin-state of crs stream');

        my $survivor_site_ci =
                    $self->get_Metrocluster_component()->command_interface();

        my $parsed_output = $survivor_site_ci->apiset()->
                          cluster_identity_show->get_parsed_output();

        $Log->comment(WaflMCC::MCC::Utils->dump(target => $parsed_output));

        my $cluster_uuid = $parsed_output->[0]->{'uuid'};

        $parsed_output = $survivor_site_ci->apiset()->
                        debug_crs_stream_show()->get_parsed_output();

        my $cmd = 'debug smdb table mccStreamDBv2 modify -cluster-uuid '.
                  $cluster_uuid.' -stream-target-admin-state down ';

        foreach my $stream (@{$parsed_output}) {
            if ($stream->{'synopsis'} =~ /MetroCluster Cluster Domain*/i){

                my $c_cmd =  $cmd.'-stream-uuid '.$stream->{'stream-uuid'};

                $survivor_site_ci->apiset()->execute_command(
                                                          command => $c_cmd );

                $survivor_site_ci->apiset()->debug_crs_stream_modify_stream(
                                    'stream-uuid' => $stream->{'stream-uuid'},
                                    'admin-state' => 'down');
            }
        }

        $parsed_output = $survivor_site_ci->apiset()->
                        debug_crs_stream_show()->get_parsed_output();

        foreach my $stream (@{$parsed_output}) {
            if ($stream->{'synopsis'} =~ /MetroCluster Cluster Domain*/i){
                if ($stream->{'admin-state'} ne 'down'){
                    WaflMCC::MCC::Exceptions::BaseException->
                        throw('Failed to modify stream admin-state to down');
                }
            }
        }
    }

    return;
}

=head2 DR_heal_aggrs()

    %args = {
                timeout  => xxx;
                interval => xxx;
            };

    $site_X_dr_coordinator->DR_heal_aggrs(%args);

Execute a 'DR heal -phase aggr' command on the site X cluster command interface

=over

=item Options:

=over

=item C<< timeout => xxx >>

Optional: See wait_for_DR_state() docs for default timeout value.

=item C<< interval => xxx >>

Optional: See wait_for_DR_state() docs for default timeout value.

=back

=back

=cut

sub DR_heal_aggrs {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            timeout  => {regex => qr/^\d+$/ , optional => 1 },
            interval => {regex => qr/^\d+$/,  optional => 1},
        },
                            );

    my $mcc_component = $self->get_Metrocluster_component();
    $Log->comment('Initiating metrocluster heal -phase aggregates...');

    # get timing information
    my $time_start = time();

    my $state = 'heal_aggrs_completed';
    my $timeout = $self->_get_DR_state_timeout(state => $state );

    try {
        $mcc_component->heal(phase => 'aggregates',
                             'method-timeout' => $timeout);
    } catch NATE::BaseException with {
        #TODO look for unknown command in exception string before retry
        #(or maybe an actual different exception related to invalid params)
        #aggrs -> aggregates interface change.  Support deprecated UI for now...
        if (! WaflMCC::MCC::Utils->workaround_expired(
                          expiration_date => '2013/03/06', burt => '630166')) {
            $Log->comment('trying older UI, new interface likely not on this build...');
            $mcc_component->heal(phase => 'aggrs',
                                 'method-timeout' => $timeout);
        } else {
            #rethrow exception, workaround has expired
            my $exception_object = shift;
            $exception_object->throw();
        }
    };

    $self->wait_for_DR_state(state => $state,%opts);

    my $execution_time = time() - $time_start;

    return;
}

=head2 DR_heal_root_aggrs()

    $site_X_dr_coordinator->DR_heal_root_aggrs(timeout  => xxx);

Execute a 'DR heal -phase root' command on the site X cluster command interface.

=over

=item Options:

=over

=item C<< timeout => xxx >>

=back

=back

Optional: See wait_for_DR_state() docs for default timeout value.

=cut

sub DR_heal_root_aggrs {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            timeout  => {regex => qr/^\d+$/ , optional => 1 },
        },
                            );

    unless ($self->_metrocluster()->workaround_expired(
                      uichange => 'mcc_expire_workaround_652320',
                      burt => '645504')) {
        $Log->comment('Part 1/3 of the BURT 645504 workaround.  BEGIN...');
        $Log->comment('Stopping switched-over data vservers (not the special ' .
                      'vservers such as vs0, MCC_Vserver and CRS_Vserver)...');
        my $site_ci = $self->get_Metrocluster_component()->command_interface();

        # subtype seems to be missing from component spec.  must use apiset.
        #my @SOed_data_vservers = NACL::C::Vserver->find(
        #               command_interface => $site_ci,
        #               filter            => { type    => 'data|cluster',
        #                                      subtype => 'sync-destination', },
        #               allow_empty       => 1 );
        try {
            my $vserver_specs = $site_ci->apiset()->vserver_show(
                           type    => 'data',
                           subtype => 'sync-destination')->get_parsed_output();
            my @SOed_data_vservers = map {
                    NACL::C::Vserver->new(command_interface => $site_ci,
                                          vserver           => $_->{vserver})
                                         } @{ $vserver_specs };

            # remove any vservers on a node's root aggr (is a redundant check?)
            # TBD If really required

            # stop each such vserver found
            map { $_->stop() } @SOed_data_vservers;
        } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
            # why this is an exception is beyond me...  :p    ignore it.
        };

        $Log->comment('Part 1/3 of the BURT 645504 workaround.  ...END.');
    }

    my $mcc_component = $self->get_Metrocluster_component();
    $Log->comment('Initiating metrocluster heal -phase root-aggregates...');

    # get timing information
    my $time_start = time();

    my $state = 'heal_roots_completed';
    my $timeout = $self->_get_DR_state_timeout(state => $state );

    try {
        $mcc_component->heal(phase => 'root-aggregates',
                             'method-timeout' => $timeout);
    } catch NATE::BaseException with {
        #TODO look for unknown command in exception string before retry
        #(or maybe an actual different exception related to invalid params)
        #aggrs -> aggregates interface change.  Support deprecated UI for now...
        if (! WaflMCC::MCC::Utils->workaround_expired(
                          expiration_date => '2013/03/06', burt => '630166')) {
            $Log->comment('trying older UI, new interface likely not on this build...');
            $mcc_component->heal(phase => 'roots',
                                 'method-timeout' => $timeout);
        } else {
            #rethrow exception, workaround has expired
            my $exception_object = shift;
            $exception_object->throw();
        }
    };
    $self->wait_for_DR_state(state => $state,%opts);

    my $execution_time = time() - $time_start;

    # make sure disk states get updated after heal roots
    $self->_metrocluster()->get_cluster(site => 'A')->
           L_workaround_force_reserver_id_update();

    # following code is removed as workaround for burt 667721 is rolled back
    # unless ($self->_metrocluster()->workaround_expired(
                      # uichange => 'mcc_expire_workaround_667721',
                      # burt     => '667721')) {
        # $Log->comment('BEGIN...trying to fix vldb discrepancies');
        # my $site_ci = $self->get_Metrocluster_component()->command_interface();
		# my @DebugVreport_objs = NACL::C::DebugVreport->find(
	                                            # command_interface => $site_ci);
        # foreach my $obj (@DebugVreport_objs){
            # if ($obj->{'object'} !~ /^unknown/i){
			    # $obj->fix();
            # }
        # }
    # }

    return;
}

=head2 DR_switchback()

    $site_X_dr_coordinator->DR_switchback(timeout  => xxx);

Execute a 'DR Switchback' command on the site X cluster command interface.

=over

=item Options:

=over

=item C<< timeout => xxx >>

=back

=back

Optional: See wait_for_DR_state() docs for default timeout value.

* Usage of this parameter is highly discouraged

=cut

sub DR_switchback {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            timeout  => {regex => qr/^\d+$/ , optional => 1 },
        },
                            );

    my $mcc_component = $self->get_Metrocluster_component();
    $Log->comment('Initiating metrocluster switchback...');

    # get timing information
    my $time_start = time();

    my $state = 'normal';
    my $timeout = $self->_get_DR_state_timeout(state => $state );

    $mcc_component->switchback('method-timeout' => $timeout);
    #note switchback changes the state to normal
    $self->wait_for_DR_state(state => $state,%opts);

    my $execution_time = time() - $time_start;

    # make sure disk states get updated after switchback
    $self->_metrocluster()->get_cluster(site => 'A')->
           L_workaround_force_reserver_id_update();

    return;
}

=head2 get_DR_state()

    my $state = $site_X_dr_coordinator->get_DR_state();

Returns one of the following strings, identifying the state of the DR operation
in the DR lifecycle:

    normal
    switchover_in_progress
    switchover_completed
    heal_aggrs_in_progress
    heal_aggrs_completed
    heal_roots_in_progress
    heal_roots_completed
    switchback_in_progress

=cut

sub get_DR_state {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $site_ci = $self->get_Metrocluster_component()->command_interface();

    # get timing information
    my $time_start = time();

    my $operation = undef;
    my $state     = undef;
    my $mcc_op_CS = undef;
    try {
        $mcc_op_CS = NACL::CS::MetroclusterOperation->fetch(
                       command_interface => $site_ci, 'method-timeout'  => 30);

        $operation = $mcc_op_CS->operation();
        $state     = $mcc_op_CS->state();
    } catch NACL::Exceptions::NoElementsFound with {
        # this is expected if no DR op has yet been performed on this setup.
        # but otherwise, it is unexpected.
        if ($self->_dr_operation_table_ok_to_be_empty()) {
            $operation = 'META_no_op';
            $state     = 'META_no_state';
        } else {
            WaflMCC::MCC::Exceptions::BaseException->throw(
                "ONTAP ERROR: metrocluster operation table empty after\n" .
                "             previous DR operations have occured.");
        }
    } catch NACL::Exceptions::InvalidMethod with {
        # allow this to pass through to the workaround below.
        $operation = undef;    # (just reiterating this here.)
        $state     = undef;    # (just reiterating this here.)
    };

    my $execution_time = time() - $time_start;

    my $found_state_string = undef;
    if (defined($operation) && defined($state)) {
        # strip any quotes
        $operation =~ s/^['"]//;    $operation =~ s/['"]$//;
        $state     =~ s/^['"]//;    $state     =~ s/['"]$//;

        foreach my $state_string (keys(%DR_STATES)) {
            if (      ($operation eq $DR_STATES{$state_string}->{op}) &&
                      ($state     eq $DR_STATES{$state_string}->{state})) {
                $found_state_string = $state_string;
                last;
            }
        }

        #aggrs -> aggregates interface change.  Support deprecated UI for now...
        if (! WaflMCC::MCC::Utils->workaround_expired(
                          expiration_date => '2013/03/06', burt => '630166')) {
            #here are deprecated states for aggrs

            # strip any quotes
            $operation =~ s/^['"]//;    $operation =~ s/['"]$//;
            $state     =~ s/^['"]//;    $state     =~ s/['"]$//;

            foreach my $state_string (keys(%DR_STATES__DEPRECATED)) {
                if (      ($operation eq $DR_STATES__DEPRECATED{$state_string}->{op}) &&
                          ($state     eq $DR_STATES__DEPRECATED{$state_string}->{state})) {
                    $found_state_string = $state_string;
                    $Log->comment("WARNING! DR Operation/State [$operation/$state] is DEPRECATED!...");
                    last;
                }
            }
        }

        unless (defined($found_state_string)) {
            WaflMCC::MCC::Exceptions::BaseException->throw(
                 "DR Operation/State [$operation/$state] not expected.");
        }

        if ($found_state_string eq 'no_op') {
            # no_op means no previous DR op has occurred on this setup.  it
            # must be in normal state
            $found_state_string = 'normal';
        } elsif ($found_state_string eq
                                     'ignore__metrocluster_unconfigure_unknown') {
# TBD: do better with this next chunk TODO burt
            $found_state_string = 'normal';
        } elsif ($found_state_string eq
                                     'ignore__metrocluster_unconfigure') {
# TBD: do better with this next chunk TODO burt
            $found_state_string = 'normal';
        } elsif ($found_state_string eq
                                     'ignore__metrocluster_configure_unknown') {
# TBD: do better with this next chunk TODO burt 653221
            $found_state_string = 'normal';
        } elsif ($found_state_string eq
                                     'ignore__metrocluster_configure') {
# TBD: do better with this next chunk TODO burt 653221
            $found_state_string = 'normal';
        } elsif ($found_state_string eq
                                     'ignore__metrocluster_modify') {
# TBD: do better with this next chunk TODO
            $found_state_string = 'normal';
        } elsif ($found_state_string eq
                                     'ignore__metrocluster_modify_unknown') {
# TBD: do better with this next chunk TODO
            $found_state_string = 'normal';
        } elsif ($found_state_string eq
                                     'ignore__switchback_continuation_agent') {
# TBD: do better with this next chunk
            $found_state_string = 'normal';
        } elsif ($found_state_string eq
                                     'ignore__switchback_continuation_agent_unknown') {
# TBD: do better with this next chunk
            $found_state_string = 'normal';
        } elsif ($found_state_string eq 'heal_roots_in_progress_temp' ||
                 $found_state_string eq 'heal_roots_completed_temp' ||
                 $found_state_string eq 'heal_aggrs_in_progress_temp' ||
                 $found_state_string eq 'heal_aggrs_completed_temp'
        ) {
            #this all needs to be upgraded, we need to support this for the short term
            #burt 655889
            $found_state_string =~ s/_temp$//;
        } else {
            # a previous op has been seen, should never see an empty table again
            # _FOR THIS SITE'S_ DR Coordinator
            $self->_dr_operation_table_ok_to_be_empty(0);
        }

    }

    return $found_state_string;
}

=head2 _get_DR_state_timeout()

    my $timeout = $site_X_dr_coordinator->_get_DR_state_timeout(state => 'switchover_completed');

returns the default timeout for the specified dr operation:

                            switchover_completed => 300,
                            heal_aggrs_completed => 600,
                            heal_roots_completed => 600,
                            normal               => 120,

=cut

sub _get_DR_state_timeout {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # constants
    my %DEFAULT_TIMEOUT = (
                            switchover_completed => 600,
                            heal_aggrs_completed => 600,
                            heal_roots_completed => 600,
                            normal               => 120,
                          );

    #building state string for regex needed in option validation
    my $state_regex_string = join('|', keys %DEFAULT_TIMEOUT);

    my %opts = validate_with(
        params => \@_,
        spec    => {
            state   => { regex => qr/^($state_regex_string)$/ }, # or no regex?
            timeout => { regex => qr/^\d+$/, optional => 1 },
        }
                            );

   $Log->comment("$opts{state} timeout: $DEFAULT_TIMEOUT{$opts{state}}");
   return $opts{timeout} || $DEFAULT_TIMEOUT{$opts{state}};
}

=head2 wait_for_DR_state()

    my $state = $site_X_dr_coordinator->wait_for_DR_state(
                                     state  => 'heal_aggrs_completed');

This method waits until expected DR State is reached.DR_SOM check will be
skipped for in_progress dr-state's

=over

=item Options:

=over

=item C<< state => $state_to_wait_for >>

Required: Need to pass one of the following state values:

    normal
    switchover_in_progress
    switchover_completed
    heal_aggrs_in_progress
    heal_aggrs_completed
    heal_roots_in_progress
    heal_roots_completed
    switchback_in_progress

=item C<< wait_for_cluster_recovery => 0/1 >>

(Optional.)  If in a particular test the DR operation itself is expected to
cause a cluster to become inaccessible, then this option may be set to 1 for the
wait to include a wait for the cluster to come back up and be reconnected to.

=item C<< timeout => xxx >>

Optional: The timeout in seconds for this method to wait through.  It is not
recommended that this option be passed for 'switchover_completed' or
'normal', as these are changes in state that have product specified timeouts.

=back

=back

=cut
#mica check here 
sub wait_for_DR_state {

    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    $Log->comment('mica : I came here under wait_for_DR_state');

    my $self = shift;

    #building state string for regex needed in option validation
    my @state_array_regex;
    for my $state_regex (keys(%DR_STATES)) {
        push(@state_array_regex, $state_regex);
    }
    my $state_regex_string = join('|', @state_array_regex);
    my %opts = validate_with(
        params => \@_,
        spec    => {
            state   => { regex => qr/^($state_regex_string)$/ },
            timeout => { regex => qr/^\d+$/, optional => 1 },
            wait_for_cluster_recovery
                    => { regex => qr/^\d+$/, default => 0 },
                   },
                            );

    # constants
    # TODO these are in another method now...BUT we may want to have a shorter wait here,
    # depending on how we deal with jobs now.  So for now, leaving them here
    #my %DEFAULT_TIMEOUT = (
    #switchover_completed => 300,
    #heal_aggrs_completed => 600,
    #heal_roots_completed => 600,
    #normal               => 120,
    #);

    my %INTERVAL = (
                            switchover_completed =>  5,
                            heal_aggrs_completed => 15,
                            heal_roots_completed =>  5,
                            normal               =>  5,
                   );

    my $desired_state = delete($opts{state});
    my $snooze_time = exists($INTERVAL{$desired_state}) ?
                        $INTERVAL{$desired_state} : 5;
    my $timeout       = exists($opts{timeout}) ? delete($opts{timeout}) :
                        $self->_get_DR_state_timeout(state => $desired_state );
                        #$DEFAULT_TIMEOUT{$desired_state};

    my $end_time      = ($timeout > 0) ? Tharn::timeout2time($timeout) : -1;
    my $no_time_limit = ($timeout == 0) ? 1 : 0;

    my $wait_for_cluster_recovery = delete($opts{wait_for_cluster_recovery});

    $Log->comment("Waiting for State: $desired_state\n");
    my $found_desired_state = 0;
    my $current_state       = 'unknown';
    # time this block
    {
        my $tracer = WaflMCC::MCC::Log::TRACELOG->start_timer(
                               message => "WAIT_FOR: $desired_state"); #metrics

        while ($no_time_limit || (time() <= $end_time)) {
            my $ci_connection_is_live = 0;
            try {
                $Log->comment('mica dr_coordinator.pm : I came here under try ');
                $current_state = $self->get_DR_state();
                $Log->comment('mica dr_coordinator.pm : after I came after get_dr_state ');
                $ci_connection_is_live = 1;
            } catch WaflMCC::MCC::Exceptions::BaseException with {
                # rethrow an MCC exception
                my $exception_object = shift;
                $exception_object->throw();
            } catch NATE::BaseException with {
                # but this may signify a node is down
                my $exception_object = shift;
                if (! $wait_for_cluster_recovery) {
                    # unexpected
                    WaflMCC::MCC::Exceptions::BaseException->throw(
                                                    $exception_object->text());
                } else {
                    # expected
                    $Log->comment('Error is:'.$exception_object->text());
                    $Log->comment('Trying to refresh CI if site live');
                    my $live_sites =
                              $self->_metrocluster()->L_determine_live_sites();
                    if ( any { $_ eq $self->_site()} @{$live_sites} ) {
                        $self->_metrocluster()->refresh_command_interfaces(
                                                       site => $self->_site());
                    }

                    $ci_connection_is_live = 0;
                }
            };

            if ($ci_connection_is_live) {
                if ($current_state eq $desired_state) {
                    $Log->comment(
                              "State found - Current State: $current_state\n");
                    $found_desired_state = 1;
                    last;
                }

                $Log->comment(
                          "Sleeping for: $snooze_time seconds\n");
                $Log->comment("Current State: $current_state\n");
                $Log->comment("Waiting for State: $desired_state\n");
            }

            Tharn::snooze($snooze_time);

            unless ($no_time_limit) {
                my $time_remaining = $end_time - time();
                $Log->comment("Time remaining: $time_remaining\n");
            }
        }

        unless ($found_desired_state) {
            WaflMCC::MCC::Exceptions::BaseException->throw(
                 "Timed out waiting for DR State: $desired_state".
                 " , Current State: $current_state\n");
        }
    }

    # DR_SOM check will be skipped for in_progress dr-state

    if ($current_state !~ /_in_progress/i){
        # now check that all nodes' DR_SOM's agree with this.
        # if not DR normal, we only care about nodes on the survivor site.
        my @sites_to_check =
             ($current_state eq 'normal') ? ( qw/ A B / ) : ( $self->_site() );
        foreach my $site (@sites_to_check) {
            my $cluster = $self->_metrocluster()->get_cluster(site => $site);
            my $sites_nodes = $cluster->L_get_cluster_mcc_nodes();

            foreach my $mcc_node (@{ $sites_nodes }) {
                # this will throw an exception if there are any inconsistencies.
                $mcc_node->L_validate_DRSOM_FSM_state(
                                                   dr_state => $current_state);
            }
        }
    }

    return $current_state;
}

1;
