# $Id $

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC DR_Partnership Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::DR_Partnership;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::Node ();

use NATE::BaseException qw(:try);
use Tharn qw($Log);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);

use Readonly;
use Data::Dumper;


=head1 NAME

WaflMCC::MCC::DR_Partnership

=head1 DESCRIPTION

This class is to be used for C-Mode Metrocluster testing.

    # Start test
    use WaflMCC::MCC::Metrocluster;
    ...
    my $the_dr_group = $mcc->get_singleton_DR_group();
    my @dr_parnerships = @{ $the_dr_group->get_DR_partnerships() };
    ...
    # Perform test here
    ...
    # End test

=cut

=head1 Public Methods

There is no public constructor for this class.  WaflMCC::MCC::DR_Partnership
objects are obtained via calls such as

    my $the_dr_group = $mcc->get_singleton_DR_group();
    my @dr_parnerships = @{ $the_dr_group->get_DR_partnerships() };

=cut

# Automatically generate the (non-public) constructor (includes a call to
# 'init') and also creates non-public members and non-public accesors for
# them.
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
        { -type => 'WaflMCC::MCC::DR_Group' },
        '_dr_group',
              ],
    scalar => [
        { -type => 'WaflMCC::MCC::Node' },
        qw/ _site_A_node _site_B_node /,
              ],
                       ];

# validate the parameters passed in and what else?
sub init {
    $Log->enter();
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_group            => { isa   => 'WaflMCC::MCC::DR_Group' },
            site_A_partner_name => { regex => qr/^\w+[\w\-]+$/ },
            site_B_partner_name => { regex => qr/^\w+[\w\-]+$/ },
        },
                            );

    # set members per the passed constructor arguments.
    $self->_dr_group($opts{'dr_group'});
    $self->_site_A_node( WaflMCC::MCC::Node->new(
                                     dr_group  => $opts{dr_group},
                                     node_name => $opts{site_A_partner_name}));
    $self->_site_B_node( WaflMCC::MCC::Node->new(
                                     dr_group  => $opts{dr_group},
                                     node_name => $opts{site_B_partner_name}));

    $Log->exit();
    return;
}

=head2 get_partner_node()

    my $mcc_node = $dr_partnership->get_partner_node(site => 'A');

This method returns the partner WaflMCC::MCC::Node object associated with
the specified site.

=over

=item Options

=over

=item C<< site => 'A'|'B' >>

(Required) Specifies which site's partner node is desired.

=back

=back

=cut

sub get_partner_node {
    $Log->enter();
    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site => { regex   => qr/^(A|B)$/ },
        }, );

    $Log->exit();
    return ($opts{site} eq 'A') ? $self->_site_A_node() : $self->_site_B_node();
}

1;
