# $Id: //depot/prod/test/main/lib/NACL/MTask/MCC/HA_Group.pm#2 $

##no critic (RequireArgUnpacking ProhibitFixedStringMatches)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC HA_Group Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::HA_Group;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::Utils ();

use NATE::BaseException qw(:try);
use NACL::Exceptions::NoElementsFound qw(:try);

use NACL::C::StorageFailover ();
use NACL::CS::StorageFailover ();

use Tharn qw($Log subtest);
use Subtest;
use NATE::STAF::Peer qw($Parent);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);


=head1 NAME

WaflMCC::MCC::HA_Group

=head1 DESCRIPTION

This class is to be used for C-Mode Metrocluster testing.

    # Start test
    use WaflMCC::MCC::Metrocluster;
    ...
    my $ha_group_at_A = $dr_group->get_HA_groups()->{'A'};
    ...
    # Perform test here
    ...
    # End test

Note that HA Group objects will be created for 2-node systems even though there
will only be 1 node in each HA 'Group'.  Some methods in this class will work
for calls in a 2-node system, but others will throw exceptions.  See individual
method documentation for notes on this.

=cut

=head1 Constructor

There is no public constructor for this class.  A
WaflMCC::MCC::HA_Group object is obtained via calls such as

    my $ha_group_at_A = $dr_group->get_HA_groups()->{'A'};

=cut

# Automatically generate the (non-public) constructor (includes a call to
# 'init') and also creates non-public members and non-public accesors for
# them.
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
        { -type => 'WaflMCC::MCC::DR_Group' },
        '_dr_group',
              ],
    scalar => [
        { -type => 'WaflMCC::MCC::Cluster' },
        '_cluster',
              ],
                       ];

sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      dr_group => { isa => 'WaflMCC::MCC::DR_Group' },
                      cluster  => { isa => 'WaflMCC::MCC::Cluster' },
                  },
                            );

    # set members per the passed constructor arguments.
    $self->_dr_group($opts{'dr_group'});
    $self->_cluster( $opts{'cluster'});

    return;
}

=head1 Public Methods

=head2 get_nodes()

    my $ha_pair_at_B = $dr_group->get_HA_groups()->{'B'}->get_nodes();

Returns an array ref of WaflMCC::MCC::Node objects associated with the
nodes present in this HA Group.

In a 2-node MCC system, this call is valid and will return just the single node
in the "group".

=cut

sub get_nodes {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $site = $self->_cluster()->get_site();
    my @ha_pair = ();
    my $dr_partnerships = $self->_dr_group()->get_DR_partnerships();
    foreach my $dr_partnership (@{ $dr_partnerships }) {
        push(@ha_pair, $dr_partnership->get_partner_node(site => $site));
    }

    return \@ha_pair;
}

=head2 get_partner_node()

    my $ha_partner_node = $ha_group->get_HA_partner(node => $mcc_node);

Returns a WaflMCC::MCC::Node objects associated with the HA partner of the
node $mcc_node.

In a 2-node MCC system, this call will throw an exception.

=over

=item Options

=item C<< node => $mcc_node >>

(Required) Given one node ($mcc_node isa WaflMCC::MCC::Node) in an HA group,
this method will return the other node in the HA group.

=back

=cut

sub get_partner_node {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      node => { isa   => 'WaflMCC::MCC::Node' },
                  },
                            );

    my $nodes = $self->get_nodes();
    my $HA_partner = undef;
    foreach my $node (@{ $nodes }) {
        if ($node != $opts{node}) {
            $HA_partner = $node;
            last;
        }
    }

    unless (defined($HA_partner)) {
        my $msg = "HA partner node not found for " . $opts{node}->get_name() .
                  "\nin HA Group: " . WaflMCC::MCC::Utils->dump($nodes) .
                  "\nin DR Group: " . WaflMCC::MCC::Utils->dump(
                                                          $self->_dr_group());
        WaflMCC::MCC::Exceptions::BaseException->throw($msg);
    }

    return $HA_partner;
}

=head1 Private and Library-Only Methods

=head2 L_get_HA_state()

This library-only method should only be called when this HA Group is in a
currently live cluster.  Public availability of the information this method
returns can be found in WaflMCC::MCC::Metrocluster::get_disaster_status().

This method returns the HA state of this HA group.  The returned data is in the
form:

    {
        HA_state      => (takeover|normal|other),
        survivor_node => $survivor_node,
        disaster_node => $disaster_node
    }

survivor_node is the WaflMCC::MCC:Node object associated with the node which
has performed the takeover operation.  disaster_node is the
WaflMCC::MCC:Node object associated with the node which is taken over.

Both survivor_node and disaster_node are undef if HA_state is not 'takeover'.

[NB: In 2-node fake MCC systems, this call always returns

    {
        HA_state      => 'normal',
        survivor_node => undef,
        disaster_node => undef,
    }
]

=cut
# mica - need to change 
sub L_get_HA_state {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $HA_state = undef;

    # short circuit for 2-node fake MCC systems
    unless ($self->_dr_group()->is_real_4_pack()) {
        return { HA_state      => 'normal',
                 survivor_node => undef,
                 disaster_node => undef, };
    }

    # temp workaround to short-circuit HA state queries for 4-node vsim systems
    unless (WaflMCC::MCC::Utils->workaround_expired(
                          expiration_date => '2013/06/12', burt => '669098')) {
        if ($self->_dr_group()->L_is_vsim_group()) {
            return { HA_state      => 'normal',
                     survivor_node => undef,
                     disaster_node => undef,
                 };
        }
    }

    $Log->comment("mica >>> Wafl fix for checking ha state .....");
    # mica - for wafl we need to have different states in HA.
    return {    HA_state      => 'normal',
                survivor_node => undef,
                disaster_node => undef,
            };

    my $cluster_component = $self->_cluster()->get_Cluster_component();

    #sfo states sometimes come back as '-' after switchover...avoid this
    #burt 669942

    my @sfo_CSs;
    my $valid_sfo_state;
    my $timeout           = 600;
    my $poll_time         = 10;
    my $end_time          = time + $timeout;

    while (!$valid_sfo_state && (time < $end_time) ) {
        my $loop_start = time;
        @sfo_CSs = NACL::CS::StorageFailover->fetch(
                                     command_interface => $cluster_component,
                                     requested_fields  => [ 'node', 'state',
                                                            'state-description' ]);
        my $invalid_state;

        my @ha_group_nodes = @{ $self->get_nodes() };
        foreach my $mcc_node (@ha_group_nodes) {
            my $node_name = $mcc_node->get_name();
            foreach my $sfo_CS (@sfo_CSs) {
                if ($sfo_CS->node() eq $node_name) {
                    if ($sfo_CS->state() eq '-') {
                        $invalid_state = 1;
                        $Log->comment("Found invalid storage failover state -");
                    }
                }
            }
        }

        $valid_sfo_state = !$invalid_state;
        if (!$valid_sfo_state) {
            #this gives us a constant loop time (if tuned correctly)
            my $sfo_time = time - $loop_start;
            $Log->comment("SFO show command took $sfo_time seconds");
            my $loop_wait_time = $loop_start + $poll_time - time;
            if ($loop_wait_time > 0) {
                $Log->comment("sleeping $loop_wait_time seconds " .
                              "($poll_time second loop time)");
                Tharn::snooze($loop_wait_time);
            } else {
                $Log->comment("not sleeping, ran over poll time by " .
                              (-1 * $loop_wait_time) .
                              " seconds ($poll_time second loop time)");
            }
        }
    }

    my %sfo_state = ();
    my @ha_group_nodes = @{ $self->get_nodes() };
    foreach my $mcc_node (@ha_group_nodes) {
        my $node_name = $mcc_node->get_name();
        foreach my $sfo_CS (@sfo_CSs) {
            if ($sfo_CS->node() eq $node_name) {
                $sfo_state{$node_name} = $sfo_CS->state();
                last;
            }
        }
    }

    my $status = 'abnormal';
    my $takeover_node = undef;
    my @connecteds = grep { $_ eq 'connected' } values(%sfo_state);
    if (scalar(@connecteds) == 2) {
        # all nodes claim sfo state is 'connected'.  normal mode.
        $status = 'normal';
    } else {
        my @takeovers = grep { $_ eq 'takeover' } values(%sfo_state);
        if ((scalar(@takeovers) == 1) && (scalar(@connecteds) == 0)) {
            foreach my $node_name (keys(%sfo_state)) {
                if ($sfo_state{$node_name} eq 'takeover') {
                     $status = 'takeover';
                     $takeover_node = $node_name;
                }
            }
        }
    }

    my $abnormal_state_message = undef;
    if ($status eq 'normal') {
        $HA_state = { HA_state      => 'normal',
                      survivor_node => undef,
                      disaster_node => undef, };
    } elsif ($status eq 'takeover') {
        my @HA_partners = @{ $self->get_nodes() };
        my ($survivor_node, $disaster_node) = ( undef, undef );
        # currently HA Groups always have only 2 members.  when this
        # changes (if ever), this will fail out to the exception below, and
        # will need work at that point.
        if (scalar(@HA_partners) == 2) {
            if ($HA_partners[0]->get_name() eq $takeover_node) {
                $survivor_node = $HA_partners[0];
                $disaster_node = $HA_partners[1];
            } elsif ($HA_partners[1]->get_name() eq $takeover_node) {
                $survivor_node = $HA_partners[1];
                $disaster_node = $HA_partners[0];
            }
        }

        unless (defined($survivor_node) && defined($disaster_node)) {
            WaflMCC::MCC::Exceptions::BaseException->throw(
                                  "Inconsistent HA Partners:\n" .
                                  WaflMCC::MCC::Utils->dump(@HA_partners));
        }

        $HA_state = { HA_state      => 'takeover',
                      survivor_node => $survivor_node,
                      disaster_node => $disaster_node, };
    } else {
        # $HA_state remains undef
        $abnormal_state_message =
                      'abnormal: ' . WaflMCC::MCC::Utils->dump(%sfo_state);
    }

    unless (defined($HA_state)) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
               "ERROR: Invalid HA states found:\n$abnormal_state_message\n\n");
    }

    return $HA_state;
}

1;
