#
#
## @summary  Trace Log class
## @status   Public
## @pod      here
## @author   dl-dpgqa-automation@netapp.com

=head1 NAME

WaflMCC::MCC::Log

=head1 DESCRIPTION

This is the Log class used to generate a trace log
for call count and timing analysis

You can manually create a Log object and initialize it as described in
Advanced Usage.

However, the
standard usage would be to use the Log object's singleton method to use an
auto initialized object:

    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

You can also pass a message that will be noted in the log for this call.

    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(
                message => 'unique message here'); #metrics

Instrument every method you
want to trace/time (probably all) but putting this at the beginning
of each method:

    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

Try to keep it exactly the same as the above line (except for indentation)
so that later on we can do mindless search and replace to modify if needed.

You don't HAVE to put it at the very beginning, and in general the
policy is to place it right after @_ is processed:

sub time_me {
    my ($self, %args) = @_;                      # or however you want to
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

}

=head Advanced Usage

If you really want to use the Log object directly, here's some info.  (Don't
do this in MCC!)

To use, create a new Log object and initialize it.  Using a BEGIN block is safer:

REDO THIS... don't use a begin block unless really needed...

BEGIN {
use WaflMCC::MCC::Log;
    $tracelog            = WaflMCC::MCC::Log->new(); # sets up trace log file
    $tracelog->init_log();

}


Now that the TRACELOG object is set up, instrument every method you
want to trace/time (probably all) but putting this at the beginning
of each method:

    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

Try to keep it exactly the same as the above line (except for indentation)
so that later on we can do mindless search and replace to modify if needed.

You don't HAVE to put it at the very beginning, and in general the
policy is to place it right after @_ is processed:

sub time_me {
    my ($self, %args) = @_;                      # or however you want to
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

}


Then clean things up in the END block:

END {
    $WaflMCC::MCC::Log::TRACELOG->log_call_count();
    $WaflMCC::MCC::Log::TRACELOG->close_log();
    $WaflMCC::MCC::Log::TRACELOG->write_csv_log();
}


=head1 METHODS

=cut

###############################################################################
# Actual start of code here
###############################################################################
# TODO
# coordinate time in various calls.  having the same timestamp in various
# places makes it easier to coordinate logs and match them
#
# add unique test ID (plan file based?  unique for that suite of tests)
# - though may not be possible as things are now
###############################################################################

package WaflMCC::MCC::Log;

use strict;
use warnings;
no warnings qw/ redefine /;

use WaflMCC::MCC::Log::Timer;

use NATE::Events qw (call_on_exit_late_add) ;

use Carp qw(carp confess);
use File::Path qw(mkpath); #for log dirs
use Data::UUID;

my $THE_ONE_LOG; #one log to rule them all...

# LOGDIR set by ntest, NACL_LOGDIR will override that (or we default to current dir)
#better way to get this?
our $DEFAULT_LOG_DIR  = $ENV{'NACL_LOGDIR'} || $ENV{'LOGDIR'} ||  q{.};

# Turn log on/off?
# set env var NACL_LOG_ACTIVE=1 to turn on, =0 to turn off, unset defaults on
our $LOG_ON = defined $ENV{'NACL_LOG_ACTIVE'} ? $ENV{'NACL_LOG_ACTIVE'} : 1 ;

# To play nicely with smoke tests we will turn off by default during smokes
# unless the user has explicitly turned on logging.  We may want to change this
# to turning off local log but still writing to the central log summary
#$LOG_ON = ($ENV{'SITE'} || q{}) eq 'smoke_bubble' ? $ENV{'NACL_LOG_ACTIVE'} : $LOG_ON ;

# log signifier, can use to grep for our logs with this string
# If a string is passed, see if it's null.  If so, allow logging without
# any additional [] around the null string.  Otherwise, add now so we only
# have to do this once.
our $NACL_GREP_STRING = defined $ENV{'NACL_GREP_STRING'}                ?
    ($ENV{'NACL_GREP_STRING'} ? "[$ENV{'NACL_GREP_STRING'}] " : q{})    :
    '[NACL_TRACE_LOG] ';

#where to write central summary logs
our $CENTRAL_LOG_DIR = defined $ENV{'CENTRAL_LOGDIR'} ?
                       $ENV{'CENTRAL_LOGDIR'} :
                       '/x/eng/rtpeng-testbeds/dpg_automation/nacllogs';

#for csv sorting purposes - should be a set of predefined values ideally
our $NACL_LOG_TYPE    = $ENV{'NACL_TEST_TYPE'} || $ENV{'NACL_LOG_TYPE'} || q{};

our $NACL_LOG_COMMENT = $ENV{'NACL_LOG_COMMENT'} || q{};

#default to the local log file being off
our $NACL_LOGFILE_ON = $ENV{'NACL_LOGFILE_ON'} ||
                         $ENV{'WaflMCC::MCC_LOGFILE_ON'} || q{};

#use Data::Dumper;  print Dumper \%ENV;
#print "FILE:$0\n";

# level "constants"
our $TRACE   = 0b00000001; #1
our $DEBUG   = 0b00000010; #2
our $INFO    = 0b00000100; #4
our $WARN    = 0b00001000; #8
our $ERROR   = 0b00010000; #16
our $REPORT  = 0b00100000; #32
our $TIMER   = 0b01000000; #64
our $REPLAY  = 0b10000000; #128

# level names
our %LEVELNAME = (
  $TRACE  => 'TRACE',
  $DEBUG  => 'DEBUG',
  $INFO   => 'INFO',
  $WARN   => 'WARN',
  $ERROR  => 'ERROR',
  $REPORT => 'REPORT',
  $TIMER  => 'TIMER',
  $REPLAY => 'REPLAY',
);

my $DEFAULT_LOG_LEVEL = defined $ENV{'NACL_LOG_LEVEL'}  ?
                        $ENV{'NACL_LOG_LEVEL'}          :
                        $TRACE + $TIMER + $INFO + $REPORT;

# we tack on .log and .csv to log name by default
my $DEFAULT_LOG_NAME  = $ENV{'NACL_LOG_NAME'} || 'trace'; #=_ for STDOUT
my $DIE_ON_LOG_FAIL   = 0;    #should we die if we can't write to logs?

our $LOG_DEBUGGING        = 0;  # 1=print out some values for testing

###############################################################################
#
#   Sub Name:       new
#
#   Description:    creates new WaflMCC::MCC::Log object.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $class    IN      scalar    this class name
#                   %arg      IN      hash      hash containing initial values:
#                                                 (These are all optional)
#                                               level = log level
#                                               logname =name of log w/o suffix
#                                               logdir = dir of log
#                                               die_on_fail = whether to die
#                                                 or warn on failures
#
#
#   Globals:        $LOG_ON   IN      scalar    default for if logging is on
#        $DEFAULT_LOG_LEVEL   IN      scalar    default for log level
#         $DEFAULT_LOG_NAME   IN      scalar    default for log name
#          $DEFAULT_LOG_DIR   IN      scalar    default for log dir
#          $DIE_ON_LOG_FAIL   IN      scalar    default for dying on write fail
#          $NACL_LOGFILE_ON   IN      scalar    default for trace log being on
#
#   Environment:
#
#   Returns:        Success:    $self (an object of this type)
#                   Failure:
#
###############################################################################
sub new {
    my ($class, %arg) = @_;


    my $self = bless {%arg},$class;


    $self->{level}   = $self->{level} || $DEFAULT_LOG_LEVEL;
    $self->{logname} = $self->{logname} || $DEFAULT_LOG_NAME;
    $self->{logdir}  = $self->{logdir} || $DEFAULT_LOG_DIR;
    $self->{die_on_fail}  = $self->{die_on_fail} || $DIE_ON_LOG_FAIL;
    $self->{on}      = $LOG_ON; #we still need to init but we'll noop calls
    $self->{grep_string} = $NACL_GREP_STRING;
    $self->{uuid}    = $self->{uuid} || new_uuid();
    $self->{log_on} = $NACL_LOGFILE_ON;

    #$self->init_log();  #do we want to let the user do this?

    return $self;
}

###############################################################################
#
#   Sub Name:       TRACELOG
#
#   Description:    returns a singleton logging object, creating it if needed
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#
#   Globals:        $THE_ONE_LOG
#
#   Environment:
#
#   Returns:        Success:    initialized log object
#                   Failure:    initialized log object
#
###############################################################################
sub TRACELOG {

    if ($THE_ONE_LOG) {
        return $THE_ONE_LOG;
    } else {
        $THE_ONE_LOG = __PACKAGE__->new();
        $THE_ONE_LOG->init_log();

        #this deals with NATE messing with DESTROY/END and also the problem of
        #subtests calling the same cleanup code to close the log (we don't
        #want that)

        call_on_exit_late_add({
                callback => \&exit_cleanup,
                keep_on_fork => 0,
        });

        return $THE_ONE_LOG;
    }

}




###############################################################################
#
#   Sub Name:       init_log
#
#   Description:    creates and opens a log file for writing, storing a
#                   file handle for all future log writes
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       object of this class
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    nothing, but sets $self->{filehandle}
#                   Failure:    nothing
#
###############################################################################
sub init_log {
    my ($self, %args) = @_;

    if (!$self->{on}) {return;} #logging turned off, do nothing
    #log file turned off, do nothing (unless forced on, used for central logs)
    if (!$self->{log_on} && !$self->{force}) {return;}

    my $logfile = "$self->{logdir}/$self->{logname}.log";
    #print "LOGFILE:$logfile\n";

    my $filehandle;
    my $mode = $args{append} ? '>>' : '>';

    if ($self->{logname} eq '_' || $self->{logdir} eq '_') { #check later, same as stdout for GNOTE/NATE?
      ##no critic (RequireBriefOpen)
      open($filehandle,"$mode&STDOUT") ||
          ($self->{die_on_fail} &&
	        confess 'could not write to STDOUT') ||
          (!$self->{die_on_fail} &&
	        carp 'could not write to STDOUT');

    }
    else {
        # open logfile OR if we can't open log then die if we want to
        #  OR warn if we don't

        my $umask = umask (0000); #try to force world writeable
        ##no critic (RequireBriefOpen)
        open($filehandle, $mode, $logfile) ||
            ($self->{die_on_fail} &&
	         confess "could not open WaflMCC::MCC::Log file for writing: $logfile") ||
            (!$self->{die_on_fail} &&
	         carp "could not open WaflMCC::MCC::Log file for writing: $logfile");

        umask($umask);  #set it back
    }

    $self->{filehandle} = $filehandle;  # save a reference to the filehandle

    $self->log('OPEN log for writing');

    #print "LOGGING LEVEL:$self->{level}\n";

    foreach my $level (1,2,4,8,16,32,64,128) # loop through all event types
    {
        if ($level & $self->{level}) {
	    #print "MATCH:$LEVELNAME{$level} ($level)\n";
	    # make note in log that we are logging this type
            $self->log("Logging on for: $LEVELNAME{$level}", $INFO);
	}

    }

    return;
}

###############################################################################
#
#   Sub Name:       log
#
#   Description:    logs message(s) passed
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this type
#                   $message  IN      scalar    single string to log
#                   $message  IN      arrayref  multiple strings to log
#                   $type     IN      scalar    the log type of the message(s).
#                                               should be one of the logging
#                                               types shown at the top of the
#                                               file: $TRACE, $DEBUG,$INFO,
#                                               $WARN, $ERROR, $FATAL, $TIMER,
#                                               $REPLAY
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    nothing
#                   Failure:    nothing
#
###############################################################################
sub log {
    my $self    = shift;
    my $message = shift;
    my $type    = shift || $INFO;  #messages default to INFO?
    if (!$self->{on}) {return;} #logging turned off, do nothing
    #log file turned off, do nothing (unless forced on, used for central logs)
    if (!$self->{log_on} && !$self->{force}) {return;}

    # skip this message if we aren't logging this type right now
    if (!($self->{level} & $type)) {
      return;
    }

    my $fh = $self->{filehandle};   # get the filehandle back


    # right now we fail silently if we didn't init or something went wrong
    if (!$fh) {
        return;
    }
    if (! defined fileno $fh) {
        return;
    }

    #for now this is a very basic write.  Later we may do something fancy
    #like spawning children to deal with writes or something
    if (ref($message) eq 'ARRAY') {  #passed more than one message?
      foreach my $item (@{$message})
      {
        print {$fh} $self->format($message,$type);
      }

    }
    else
    {
      print {$fh} $self->format($message,$type);
    }

    return;
}

###############################################################################
#
#   Sub Name:       format
#
#   Description:    formats log message for writing to file (or whatever)
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#                   $message  IN      scalar    string of message to format
#                   $type     IN      scalar    string of the log type
#
#   Globals:        $LOG_DEBUGGING    scalar    prints message also if true
#                   $LEVELNAME        hash      gets printable name of $type
#
#   Environment:
#
#   Returns:        Success:    scalar of formatted message
#                   Failure:
#
###############################################################################
sub format {
    my $self    = shift;
    my $message = shift;
    my $type    = shift;

    my $format = '[' . scalar localtime() .
                 "] $self->{grep_string}$LEVELNAME{$type} $message\n";

    if ($LOG_DEBUGGING) {print $format;}
    return $format;
}


###############################################################################
#
#   Sub Name:       start_timer
#
#   Description:    helper method for creating a WaflMCC::MCC::Log::Timer
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#                   %args     IN      hash      arguments to pass through to
#                                               WaflMCC::MCC::Log::Timer
#
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    WaflMCC::MCC::Log::Timer object
#                   Failure:
#
###############################################################################
sub start_timer {
    my ($self, %args) = @_;
    if (!$self->{on}) {return;} #logging turned off, do nothing

    $args{logger} = $args{logger} || $self;

    #see docs for these in Timer.
    $args{inpackage} = [caller 0]; #pass in package of caller  structure
    $args{incall} = [caller 1]; #pass in caller structure
    $args{calledby} = [caller 2];

    my $timer = WaflMCC::MCC::Log::Timer->new(%args);

    return $timer;
}


###############################################################################
#
#   Sub Name:       increment_call
#
#   Description:    increment call count of call - called by WaflMCC::MCC::Log::Timer
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#                   $incall   IN      scalar    call we are in
#           $incall_package   IN      scalar    package $incall is in
#                   $calledby IN      scalar    who called $incall (unused)
#
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    nothing
#                   Failure:
#
###############################################################################
sub increment_call {
    my ($self, $incall, $incall_package, $calledby) = @_;

    #if (!($calledby && $incall)) { #check this again when counting both

    if (!$incall) {
        return;
    }

    #increment simple called by counter
    #$self->{callcount}->{$calledby}->{$incall}++;

    #increment total call count counter
    $self->{totalcallcount}->{$incall_package}->{$incall}++;

    return;
}

###############################################################################
#
#   Sub Name:       increment_call_timer
#
#   Description:    Sums time in calls.  Note that this is additive, that is,
#                   any call will include the time of any subcalls. Another
#                   metric might be to subtract out the subcalls to analyze
#                   each call directly.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#                   $incall   IN      scalar    call timer is in
#           $incall_package   IN      scalar    package $incall is in
#                   $time     IN      scalar    time in seconds $incall took
#
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    nothing
#                   Failure:
#
###############################################################################
sub increment_call_timer {
    my ($self, $incall, $incall_package, $time) = @_;

    if (!$incall) {
        return;
    }


    #increment total time counter for subroutine call
    $self->{totalcalltime}->{$incall_package}->{$incall} += $time;

    return;
}


###############################################################################
#
#   Sub Name:       log_call_count
#
#   Description:    dumps a total of the call counts and seconds to the log
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#
#
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    nothing
#                   Failure:
#
###############################################################################
sub log_call_count {
    my $self = shift;

    #not sure I'm happy yet with just using REPORT here, will try for now...
    #maybe combine REPORT+TRACE? would need to correct the name though    TODO
    $self->log('[message=Log Report Totals]',$REPORT); #was $TRACE

    #log total call counts for each counter
    foreach my $package (sort keys %{$self->{totalcallcount}}) {
        foreach my $count (sort keys %{$self->{totalcallcount}->{$package}}) {
            $self->log("[call=$count] [package=$package] [count=" .
		           $self->{totalcallcount}->{$package}->{$count} . ']',
	               $REPORT);  # was $TRACE
            #print $count . ' = ' .
	        #$self->{totalcallcount}->{$package}->{$count} . "\n";
	}
    }

    #log total call timers for each counter
    foreach my $package2 (sort keys %{$self->{totalcalltime}}) {
        foreach my $time (sort keys %{$self->{totalcalltime}->{$package2}}) {
	    #use sprintf for time formatting
            $self->log(sprintf('[call=%s] [total seconds=%f]',
	                   $time,$self->{totalcalltime}->{$package2}->{$time}),
	           $REPORT); #was $TIMER
            #print $time . ':' . $self->{totalcalltime}->{$time} . "\n";
	}
    }

    return;
}
###############################################################################
#
#   Sub Name:       write_csv_log
#
#   Description:    opens the csv file, dumps caller metrics (call counts,
#                   times, etc) and closes the file.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    nothing
#                   Failure:    nothing
#
###############################################################################
sub write_csv_log {
    my $self = shift;
    if (!$self->{on}) {return;} #logging turned off, do nothing
    #log file turned off, do nothing (unless forced on, used for central logs)
    if (!$self->{log_on} && !$self->{force}) {return;}

    # TODO:
    # this was thrown in at the last second, should really redo other stuff
    # to allow reuse of those methods (log, init_log, etc...)

    # csv log file format
    # package, call, count, time

    my $logfile = "$self->{logdir}/$self->{logname}.csv";
    #print "LOGFILE:$logfile\n";

    my $filehandle;

    if ($self->{logname} eq '_' || $self->{logdir} eq '_') { #check later, same as stdout for GNOTE/NATE?
      ##no critic (RequireBriefOpen)
      open($filehandle,'>&STDOUT') ||
            ($self->{die_on_fail} &&
	        confess 'could not write to STDOUT') ||
            (!$self->{die_on_fail} &&
	        carp 'could not write to STDOUT');

    }
    else {
        # open logfile OR if we can't open log then die if we want to
        #  OR warn if we don't
        #
        my $umask = umask (0000); #try to force world writeable

        ##no critic (RequireBriefOpen)
        open($filehandle, q{>}, $logfile) ||
            ($self->{die_on_fail} &&
	        confess "could not open WaflMCC::MCC::Log file for writing: $logfile") ||
            (!$self->{die_on_fail} &&
	        carp "could not open WaflMCC::MCC::Log file for writing: $logfile");
        umask($umask);
    }

    #should these be global, or allow ENV arguments or something?
    my $TEST_RELEASE = $ENV{'TEST_RELEASE'} || q{}; #set by ntest
    my $USER         = $ENV{'USER'}         || q{};

    #assuming we want to write a header...
    print {$filehandle}
        "log name $self->{grep_string},package,call,total count," .
        "total seconds,date,time,release,user,log type,comment,uuid\n";
    (my $comment = $NACL_LOG_COMMENT) =~ s/,/_/g; # we don't want commas

    my @time = localtime(time);
    #log total call counts for each counter
    foreach my $package (sort keys %{$self->{totalcallcount}}) {
        foreach my $count (sort keys %{$self->{totalcallcount}->{$package}}) {
	    print {$filehandle}
            sprintf("%s,%s,%s,%i,%f,%i%02i%i,%02i:%02i:%02i,%s,%s,%s,%s,%s\n",
	               $self->{grep_string},            # grep string
	               $package,                             # name of package
		           $count,                               # name of call
		           $self->{totalcallcount}->{$package}->{$count}, # count
	               $self->{totalcalltime}->{$package}->{$count},   # time
                   $time[5]+1900,
                   $time[4]+1,
                   $time[3],
                   $time[2],
                   $time[1],
                   $time[0],
                   $TEST_RELEASE || q{},                    #release name
                   $USER || q{},                            #user running test
                   $NACL_LOG_TYPE || q{},                  #predefined types
                   $comment,                                #user comment
                   $self->{uuid},                           #unique ID
		       );
	    }
    }
    close $filehandle || carp "could not close log $logfile";
    return;
}
###############################################################################
#
#   Sub Name:       close_log
#
#   Description:    writes closing message to log, closes log/filehandle
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    nothing
#                   Failure:
#
###############################################################################
sub close_log {
    my $self = shift;
    if (!$self->{on}) {return;} #logging turned off, do nothing
    #log file turned off, do nothing (unless forced on, used for central logs)
    if (!$self->{log_on} && !$self->{force}) {return;}


    $self->log('CLOSE log');

    if ($self->{filehandle} && defined fileno $self->{filehandle}) {
        close $self->{filehandle} ||
            carp "could not close log $self->{logdir}/$self->{logname}";
    }

    return;

}
###############################################################################
#
#   Sub Name:       get_log_path
#
#   Description:    returns log path based on log path + current date,
#                   creating it if needed (by default).  This is for writing
#                   to the central log location
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#   Globals: $CENTRAL_LOG_DIR IN      scalar    path to fixed central logs
#
#   Environment:
#
#   Returns:        Success:    either path, or (path,time)
#                   Failure:
#
###############################################################################
sub get_log_path {
    my ($self, %args) = @_;
    my $no_path_creation = $args{no_path_creation};

    my $time = $args{time} || time;
    my @date = gmtime($time); #use GMT rather than local for this
    # /year/month/day
    my $subpath = (1900+$date[5]) . q{/} . ($date[4]+1) . q{/} . $date[3];

    my $fullpath = "$CENTRAL_LOG_DIR/$subpath";

    if (! -e $fullpath) #does it exist?
    {
        print "nope doesn't exist...$fullpath\n" if $LOG_DEBUGGING;
        if (!$no_path_creation) {
            my $umask = umask (0000); #try to force world writeable
            if (eval {mkpath($fullpath); 1;}) {
                #now we need to add this new path to the master logpath
                my @restore = ($self->{logdir},$self->{logname},$self->{level},
                      $self->{grep_string},$self->{die_on_fail},);
	            $self->{logname} = 'logpaths';
	            $self->{level} = $TRACE; #or ??
                $self->{grep_string} = '[NACL_CENTRAL_LOG] ';
                my $logdir = $CENTRAL_LOG_DIR;
                $self->{logdir} = $logdir;
	            $self->init_log(append=>1);
	            $self->log("[logdir=$subpath]",$TRACE);
	            $self->close_log();
                #restore vars
                ($self->{logdir},$self->{logname},$self->{level},
                      $self->{grep_string},$self->{die_on_fail}) = @restore;
            }
            else {
                carp "error creating central log dir [$fullpath]:$@";
            }

            umask($umask);
        }
    }

    if (wantarray) {
        return ($fullpath,$time);
    }
    else {
        return $fullpath;
    }

}


###############################################################################
#
#   Sub Name:       write_central_logs
#
#   Description:    writes summary to central location
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#   Globals:        $TRACE    IN      scalar    TRACE log level value
#
#   Environment:    $LOGDIR   IN      scalar    logging dir used by ntest
#
#   Returns:        Success:    nothing
#                   Failure:    nothing
#
###############################################################################
sub write_central_logs {
    my $self = shift;
    if (!$self->{on}) {return;} #logging turned off, do nothing

    #so let's abuse some globals/class vars now...
    #...this is horrible and we need to rewrite it to remove globals... TODO

    if ($self->{on} && $CENTRAL_LOG_DIR)
    {

        $self->log("[message=write report to $CENTRAL_LOG_DIR]",$INFO);

        #save these for later - we could also copy and make a new object,
        #or really it would be better to have functionality to allow this
        #more cleanly
        my @restore = ($self->{logdir},$self->{logname},$self->{level},
                      $self->{grep_string},$self->{die_on_fail},$self->{force});
        #still write central logs even when local logs are off
        $self->{force} = 1;
        my $time;  #we want the same timestamp as the log file itself
        ($self->{logdir}, $time) = $self->get_log_path(); #$CENTRAL_LOG_DIR;

        #create logname now for storing in first log
        my $testname = $ENV{'TESTNAME'} || q{}; #remove warnings
	    #yes there is still some chance for name collision
	    my $logname = (($testname =~ /([^\/]+)$/)[0] || 'unknown') .
                           q{.} .  $time;

	    #first write out logfile location
	    $self->{logname} = 'logpaths';
	    $self->{level} = $TRACE; #or ??
        $self->{grep_string} = '[NACL_CENTRAL_LOG] ';
	    $self->init_log(append=>1);
        my $logdir = $ENV{LOGDIR} || q{};
	    $self->log("[logdir=$logdir] [csvlog=$logname]",$TRACE);
	    $self->close_log();

        #now dump a copy of the summary

	    $self->{logname} = $logname;
        $self->{grep_string} = $ENV{'TESTNAME'} || q{};
	    $self->{die_on_fail} = 0;

        if (eval {$self->write_csv_log(); 1;}) { #just in case something fails
        }
        else {
            carp "error writing central log:$@";
        }

        #restore vars
        ($self->{logdir},$self->{logname},$self->{level},
         $self->{grep_string},$self->{die_on_fail},$self->{force},) = @restore;

    }

    return;

}

###############################################################################
#
#   Sub Name:       new_uuid
#
#   Description:    generate UUID for this session
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    UUID string
#                   Failure:    nothing
#
###############################################################################
sub new_uuid {
    #we will likely switch to UUID::Tiny when it is installed.  For now,
    #a very old version of Data::UUID will be used

    my $generator = Data::UUID->new();
    my $uuid = lc $generator->create_str(); #prefer lowercase
    return $uuid;
}

###############################################################################
#
#   Sub Name:       DESTROY
#
#   Description:    Double check we have cleaned up and closed the log
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     IN      ref       object of this class
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    nothing
#                   Failure:    nothing
#
###############################################################################
sub DESTROY {
    my $self = shift;
    if (!$self->{on}) {return;} #logging turned off, do nothing

    $self->log('[message=got to DESTROY]',$TRACE);

#use Data::Dumper;
#print Dumper $self->{totalcalltime};
#print Dumper $self->{totalcallcount};

    $self->log('CLOSE log (in destroy)');
    if ($self->{filehandle} && defined fileno $self->{filehandle}) {
        close $self->{filehandle} ||
        carp "log already closed, don't need to close log $self->{logdir}/$self->{logname}";
    }

    return;
}

###############################################################################
#
#   Sub Name:       exit_cleanup
#
#   Description:    called by NATE::Events to clean up and close the log
#                   - but ONLY by the parent process, not any subtests
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#
#   Globals:        $THE_ONE_LOG      ref       object of this class
#
#   Environment:
#
#   Returns:        Success:    nothing
#                   Failure:    nothing
#
###############################################################################
sub exit_cleanup {

    if ($THE_ONE_LOG) {
        $THE_ONE_LOG->log_call_count();
        $THE_ONE_LOG->close_log();
        $THE_ONE_LOG->write_csv_log();
        $THE_ONE_LOG->write_central_logs();
    }
}

1;
